(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[     23785,        848]*)
(*NotebookOutlinePosition[     57990,       2080]*)
(*  CellTagsIndexPosition[     57901,       2074]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Complex Variables", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "In this notebook we explore some of the basic techniques for manipulating \
complex variables using ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "."
}], "Abstract",
  CellMargins->{{55.1875, 87.5}, {Inherited, Inherited}}],

Cell[CellGroupData[{

Cell[TextData[StyleBox["Initialization"]], "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(\(Off[General::spell, General::spell1];\)\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " does not automatically assume that variables are real or that real \
variables are positive; its symbolic manipulations are intended to be as \
general as possible unless given specific information about the properties of \
particular variables.  This, as we've seen before, is the reason that it \
often does not perform some of the simplifications you want.  For example, \
although at first glance it might seem that the following expression can be \
simplified, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " declines to produce the obvious result because the replacement ",
  Cell[BoxData[
      \(TraditionalForm\`\@\(\(a\^2\) b\^2\) \[Rule] a\ b\)]],
  " may not be correct if either ",
  Cell[BoxData[
      \(TraditionalForm\`a\)]],
  " or ",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  " is negative or if either is complex."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(expr1\  = \ Sqrt[\ a\^2\ b\^2\ ];\)\), "\n", 
    \(expr1\  // \ Simplify\)}], "Input"],

Cell[BoxData[
    \(\@\(a\^2\ b\^2\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "However, if you are certain that all variables in an expression are real \
and positive, ",
  ButtonBox["PowerExpand",
    ButtonStyle->"RefGuideLink"],
  " will perform the desired simplification.  Of course, you are then \
responsible for any problems that might arise if the variables involved \
violate these assumptions!"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(expr1\  // PowerExpand\)], "Input"],

Cell[BoxData[
    \(a\ b\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Similarly, many other functions do not respond to ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " without making assumptions about their arguments which determine the \
appropriate branch for functions of complex variables."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[expr1]\  // Simplify\)], "Input"],

Cell[BoxData[
    \(1\/2\ Log[a\^2\ b\^2]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Log[expr1]\  // PowerExpand\)\  // Simplify\)], "Input"],

Cell[BoxData[
    \(Log[a] + Log[b]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Similarly, ",
  ButtonBox["ComplexExpand",
    ButtonStyle->"RefGuideLink"],
  " separates an expression into its real and imaginary parts assuming that \
all variables within that expression are real. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(z = x\  + \ \[ImaginaryI]\ y;\)\), "\n", 
    \({Sin[z], Cosh[z\^2], Log[z]} // ComplexExpand\)}], "Input"],

Cell[BoxData[
    \({Cosh[y]\ Sin[x] + \[ImaginaryI]\ Cos[x]\ Sinh[y], 
      Cos[2\ x\ y]\ Cosh[x\^2 - y\^2] + \[ImaginaryI]\ Sin[2\ x\ y]\ Sinh[
            x\^2 - y\^2], \[ImaginaryI]\ Arg[x + \[ImaginaryI]\ y] + 
        1\/2\ Log[x\^2 + y\^2]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "We can still use ",
  StyleBox["ComplexExpand", "InlineInput",
    FontWeight->"Bold"],
  " to simplify expressions involving complex variables by supplying a list \
of complex arguments as an optional argument, such that ",
  StyleBox["ComplexExpand[expr,{z1,z2...}]", "InlineInput",
    FontWeight->"Bold"],
  " can recognize ",
  StyleBox["z1", "InlineInput"],
  " and ",
  StyleBox["z2", "InlineInput"],
  " as complex quantities."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[k, z];\)\), "\n", 
    \(ComplexExpand[\ Exp[\[ImaginaryI]\ k\ z], \ k]\)}], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(\(-z\)\ Im[k]\)\ Cos[
          z\ Re[k]] + \[ImaginaryI]\ \[ExponentialE]\^\(\(-z\)\ Im[k]\)\ Sin[
          z\ Re[k]]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Another useful option is to specify a list of ",
  StyleBox["TargetFunctions", "InlineInput",
    FontWeight->"Bold"],
  " we would like to appear in the result.  Thus, ",
  StyleBox["TargetFunctions\[Rule]{Abs,Arg}", "InlineInput",
    FontWeight->"Bold"],
  " can be used to obtain polar notation."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[k, z];\)\), "\n", 
    \(ComplexExpand[\ Exp[\[ImaginaryI]\ k\ z], \ k, \ 
      TargetFunctions \[Rule] {Abs, Arg}]\)}], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(\(-z\)\ Abs[k]\ Sin[Arg[k]]\)\ Cos[
          z\ Abs[k]\ Cos[
              Arg[k]]] + \[ImaginaryI]\ \[ExponentialE]\^\(\(-z\)\ Abs[k]\ \
Sin[Arg[k]]\)\ Sin[z\ Abs[k]\ Cos[Arg[k]]]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[k, z];\)\), "\n", 
    \(ComplexExpand[\ Exp[\[ImaginaryI]\ k\ z], \ k, \ 
        TargetFunctions \[Rule] {Abs, Arg}]\  /. {Abs[k] \[Rule] \[Kappa], 
        Arg[k] \[Rule] \[CurlyPhi]}\)}], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(\(-z\)\ \[Kappa]\ Sin[\[CurlyPhi]]\)\ Cos[
          z\ \[Kappa]\ Cos[\[CurlyPhi]]] + \[ImaginaryI]\ \[ExponentialE]\^\(\
\(-z\)\ \[Kappa]\ Sin[\[CurlyPhi]]\)\ Sin[
          z\ \[Kappa]\ Cos[\[CurlyPhi]]]\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Built-in Functions: Re, Im, Abs, Conjugate", "Section"],

Cell[TextData[{
  "The built-in functions ",
  ButtonBox["Re",
    ButtonStyle->"RefGuideLink"],
  ", ",
  ButtonBox["Im",
    ButtonStyle->"RefGuideLink"],
  ", ",
  ButtonBox["Abs",
    ButtonStyle->"RefGuideLink"],
  ", and ",
  ButtonBox["Conjugate",
    ButtonStyle->"RefGuideLink"],
  " return the real part, imaginary part, absolute value, and complex \
conjugate of complex numbers, but make no ",
  StyleBox["a priori",
    FontSlant->"Italic"],
  " assumptions about their arguments.  Hence, the evaluation of many \
expressions appears incomplete."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[a, b, z];\)\), "\n", 
    \(\(z\  = \ a\  + \ 2\ \[ImaginaryI]\ b;\)\), "\n", 
    \({Re[z], \ Im[z], \ Abs[z], \ Conjugate[z]}\)}], "Input"],

Cell[BoxData[
    \({\(-2\)\ Im[b] + Re[a], Im[a] + 2\ Re[b], Abs[a + 2\ \[ImaginaryI]\ b], 
      Conjugate[a + 2\ \[ImaginaryI]\ b]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Now suppose that we declare ",
  StyleBox["a", "InlineInput"],
  " and ",
  StyleBox["b", "InlineInput"],
  " to be real by associating rules with those symbols that nullify their \
imaginary parts.  These rules are known as ",
  StyleBox["upvalues",
    FontSlant->"Italic"],
  " because they are associated with the argument of a function, here ",
  StyleBox["Im", "InlineInput"],
  ", instead of the function itself."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[a, b, z];\)\), "\n", 
    \(\(a /: \ Im[a]\  = \ 0;\)\ \), "\n", 
    \(\(b /: \ Im[b]\  = \ 0;\)\), "\n", 
    \(\(z\  = \ a\  + \ 2\ \[ImaginaryI]\ b;\)\), "\n", 
    \({Re[z], \ Im[z], \ Abs[z], \ Conjugate[z]}\)}], "Input"],

Cell[BoxData[
    \({Re[a], 2\ Re[b], Abs[a + 2\ \[ImaginaryI]\ b], 
      Conjugate[a + 2\ \[ImaginaryI]\ b]}\)], "Output"]
}, Closed]],

Cell["\<\
Notice that we could have assigned the upvalue using the more compact syntax \
shown below instead.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?a\)\)], "Input"],

Cell["Global`a", "Print",
  CellTags->"Info3351094705-7946987"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {GridBox[{
                {\(Im[a] ^= 0\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "a"],
      Editable->False]], "Print",
  CellTags->"Info3351094705-7946987"]
}, Closed]],

Cell[TextData[{
  "Better results are obtained, but notice that setting ",
  StyleBox["Im[a]\[Rule]0", "InlineInput"],
  " does not automatically ensure that ",
  StyleBox["Re[a]\[Rule]a", "InlineInput"],
  ".  The following declarations are more specific and give somewhat better \
results."
}], "Text"],

Cell[BoxData[{
    \(\(Clear[a, b, z];\)\), "\n", 
    \(\(a /: \ Im[a]\  = \ 0;\)\ \), "\n", 
    \(\(a /: \ Re[a]\  = \ a;\)\), "\n", 
    \(\(b /: \ Im[b]\  = \ 0;\)\), "\n", 
    \(\(b /: \ Re[b]\  = \ b;\)\), "\n", 
    \(\(z\  = \ a\  + \ 2  \[ImaginaryI]\ b;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({Re[z], Im[z], \ Abs[z], \ Conjugate[z]}\)], "Input"],

Cell[BoxData[
    \({a, 2\ b, Abs[a + 2\ \[ImaginaryI]\ b], 
      Conjugate[a + 2\ \[ImaginaryI]\ b]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Nevertheless, ",
  StyleBox["Abs", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["Conjugate", "InlineInput",
    FontWeight->"Bold"],
  " still do not behave as we would like.  These deficiencies arise because \
this family of functions is defined for numerical arguments and do not have \
rules attached to guide the evaluation of expressions.  Although we have \
supplemented these rules by attaching declarations to some of the symbols, it \
is a rather incomplete solution.  For example, even if we declare ",
  StyleBox["a", "InlineInput"],
  " to be real, ",
  StyleBox["Re", "InlineInput",
    FontWeight->"Bold"],
  " does not recognize that ",
  StyleBox["Cos[a]", "InlineInput"],
  " is then real."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Re[Cos[a]]\)], "Input"],

Cell[BoxData[
    \(Re[Cos[a]]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "For expressions involving real symbols, we can use ",
  StyleBox["ComplexExpand", "InlineInput",
    FontWeight->"Bold"],
  " to obtain the real or imaginary parts of functions."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({Re[Cos[a]], Re[Log[x\  + \ \[ImaginaryI]\ y]]}\  // 
      ComplexExpand\)], "Input"],

Cell[BoxData[
    \({Cos[a], 1\/2\ Log[x\^2 + y\^2]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Similarly, ",
  StyleBox["Conjugate", "InlineInput",
    FontWeight->"Bold"],
  " does not utilize associate or distributive properties."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Conjugate[z]\)], "Input"],

Cell[BoxData[
    \(Conjugate[a + 2\ \[ImaginaryI]\ b]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The obvious way to circumvent this problem is to manually replace ",
  StyleBox["\[ImaginaryI]", "InlineInput"],
  " by ",
  StyleBox["-\[ImaginaryI]", "InlineInput",
    FontWeight->"Bold"],
  ", but it fails!"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(z\  /. \[ImaginaryI] \[Rule] \(-\[ImaginaryI]\)\)], "Input"],

Cell[BoxData[
    \(a + 2\ \[ImaginaryI]\ b\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Why doesn't this work (other than Murphy's Law prohibiting success of the \
obvious)?  If we use ",
  StyleBox["FullForm", "InlineInput",
    FontWeight->"Bold"],
  " to examine our expression"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FullForm[z]\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(Plus[a, Times[Complex[0, 2], b]]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FullForm[\[ImaginaryI]]\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(Complex[0, 1]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[TextData[{
  "we discover that neither the symbol ",
  StyleBox["\[ImaginaryI]", "InlineInput"],
  " nor its internal representation ",
  StyleBox["Complex[0,1]", "InlineInput",
    FontWeight->"Bold"],
  " is present in the expression for ",
  StyleBox["z", "InlineInput",
    FontWeight->"Bold"],
  ".  Thus, the replacement rule has no effect.  Fortunately, this problem is \
easily circumvented by defining our own complex conjugation function as \
suggested by Zimmerman and Olness:"
}], "Text"],

Cell[BoxData[{
    \(\(\(conjugate::usage\  = "\<A simple method for computing the conjugate \
of an object which is explicitly complex.\>";\)\(\n\)
    \)\), "\n", 
    \(\(conjugateRule\  = \ 
        Complex[re_, im_] :> Complex[re, \(-im\)];\)\), "\n", 
    \(\(conjugate[exp__]\  := \ exp /. conjugateRule;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(conjugate[z]\)], "Input"],

Cell[BoxData[
    \(a - 2\ \[ImaginaryI]\ b\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(conjugate[x\  - \ \[ImaginaryI]\ y]\)], "Input"],

Cell[BoxData[
    \(x + \[ImaginaryI]\ y\)], "Output"]
}, Closed]],

Cell[TextData[{
  "However, this function will not work for arbitrary expressions because it \
assumes that all symbols are real and conjugates only the numerical \
coefficients expressed in terms of ",
  StyleBox["Complex", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell[TextData[{
  "Explain why ",
  StyleBox["Re[ ComplexExpand[a + \[ImaginaryI] b] ", "InlineInput"],
  "and ",
  StyleBox["ComplexExpand[ Re[a + \[ImaginaryI] b] ]", "InlineInput"],
  " give different results.  Which should be preferred?"
}], "Problem"],

Cell[TextData[{
  "Determine the real and imaginary parts of ",
  StyleBox["Cosh[k z]", "InlineInput"],
  " for real ",
  StyleBox["z", "InlineInput"],
  " and complex ",
  StyleBox["k", "InlineInput"],
  "."
}], "Problem"],

Cell[TextData[{
  "Plot the real and imaginary parts of ",
  StyleBox["BesselJ[3,z]", "InlineInput"],
  " for complex ",
  StyleBox["z", "InlineInput"],
  "."
}], "Problem"],

Cell[TextData[{
  "Write a function which properly simplifies ",
  StyleBox["Abs[expr]", "InlineInput"],
  " assuming that the variables in ",
  StyleBox["expr", "InlineInput"],
  " are real.  Thus, your function should yield ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(Abs[x + \[ImaginaryI]\ y] \[Rule] Sqrt[x\^2 + y\^2]\),
          "InlineInput"], TraditionalForm]]],
  " and similar results for related expressions.  Use your function to obtain \
a simple expression for ",
  StyleBox["Abs[Cos[x + \[ImaginaryI] y]]", "InlineInput"],
  "  involving only real quantities."
}], "Problem"],

Cell[TextData[{
  "Write a function which declares a variable to be real.  Your function \
should specify values for ",
  StyleBox["Re", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Im", "InlineInput",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Conjugate", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Problem"]
}, Closed]],

Cell[CellGroupData[{

Cell["Using the package ReIm", "Section"],

Cell[CellGroupData[{

Cell["The package", "Subsection"],

Cell[BoxData[
    \(Needs["\<Algebra`ReIm`\>"]\)], "Input"],

Cell[TextData[{
  "The package ",
  StyleBox["ReIm", "InlineInput",
    FontWeight->"Bold"],
  " can also be used to extend the functionality of the built-in functions \
that manipulate complex variables.  Hence, if we declare certain variables to \
be real, expressions involving those variables can be simplified in the \
expected manner."
}], "Text"],

Cell[BoxData[{
    \(\(Clear[a, b, z];\)\), "\n", 
    \(\(a /: \ Im[a]\  = \ 0;\)\), "\n", 
    \(\(b /: \ Im[b]\  = \ 0;\)\), "\n", 
    \(\(z\  = \ a\  + \ \[ImaginaryI]\ b;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({Re[z], Im[z], Abs[z], Conjugate[z], \ z\ Conjugate[z]}\  // 
      Simplify\)], "Input"],

Cell[BoxData[
    \({a, b, Abs[a + \[ImaginaryI]\ b], a - \[ImaginaryI]\ b, 
      a\^2 + b\^2}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Notice that now it is not necessary to give two rules to get a real \
number, simply nullifying the imaginary part is sufficient when ",
  StyleBox["ReIm", "InlineInput",
    FontWeight->"Bold"],
  " is active.  However, the absolute value function still does not operate \
as we would like, but we can work around that with a replacement rule.  \
Furthermore, expressions involving variables of unspecified type are handled \
as if they are complex."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Conjugate[x\  - \ \[ImaginaryI]\ y]\)], "Input"],

Cell[BoxData[
    \(Conjugate[x] + \[ImaginaryI]\ Conjugate[y]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[z]\  // ComplexExpand\)], "Input"],

Cell[BoxData[
    \(\[ImaginaryI]\ Arg[a + \[ImaginaryI]\ b] + 
      1\/2\ Log[a\^2 + b\^2]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The ",
  StyleBox["ReIm", "InlineInput",
    FontWeight->"Bold"],
  " package provides a function ",
  ButtonBox["RealValued",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[f,g,\[CenterEllipsis]]", "InlineInput"],
  " that declares that the functions ",
  StyleBox["f,g,\[CenterEllipsis]", "InlineInput"],
  " produce real-valued output for real-valued arguments.  Thus, if we \
declare ",
  StyleBox["s", "InlineInput"],
  " and ",
  StyleBox["t", "InlineInput"],
  " to be real"
}], "Text"],

Cell[BoxData[{
    \(\(RealValued[f, g];\)\), "\[IndentingNewLine]", 
    \(\(Im[s] ^= 0;\)\), "\[IndentingNewLine]", 
    \(\(Im[t] ^= 0;\)\)}], "Input"],

Cell["\<\
the imaginary part of the logarithm of a product of real-valued functions\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Im[Log[f[s]\ g[t]]]\)], "Input"],

Cell[BoxData[
    \(Arg[f[s]] + Arg[g[t]]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "reduces to the sum of their arguments, as expected.  Similarly, if we also \
declare ",
  StyleBox["s", "InlineInput"],
  " and ",
  StyleBox["t", "InlineInput"],
  " to be negative"
}], "Text"],

Cell[BoxData[{
    \(\(Negative[s] ^= True;\)\), "\[IndentingNewLine]", 
    \(\(Negative[t] ^= True;\)\)}], "Input"],

Cell["then", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Im[Log[s\ t]]\)], "Input"],

Cell[BoxData[
    \(2\ \[Pi]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "is interpreted as the sum ",
  Cell[BoxData[
      \(TraditionalForm\`Arg[s] + Arg[t] = \(\[Pi] + \[Pi] = 2  \[Pi]\)\)]],
  "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Efficiency issues for ReIm", "Subsection"],

Cell[TextData[{
  "Although ",
  StyleBox["ReIm", "InlineInput",
    FontWeight->"Bold"],
  " provides a set of rules which performs many simplifications of \
expressions involving complex variables in a manner that is transparent to \
the user, this package is quite inefficient and can slow down calculations \
enormously.  To illustrate this problem, we borrow the following example from \
Bahder.  Before evaluating the following expression, quit the kernel in order \
to unload the package."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(expr\  = \ 
      Integrate[\ \(a\ x\^2 + b\ x\  + \ c\)\/\(c\ x\  + \ a\), x]\)], "Input"],

Cell[BoxData[
    \(\(\((\(-a\^2\) + b\ c)\)\ x\)\/c\^2 + \(a\ x\^2\)\/\(2\ c\) + \(\((a\^3 \
- a\ b\ c + c\^3)\)\ Log[a + c\ x]\)\/c\^3\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The time needed to evaluate the real part of this result, assuming that \
the coefficients ",
  StyleBox["a", "InlineInput"],
  ", ",
  StyleBox["b", "InlineInput"],
  ", ",
  StyleBox["c", "InlineInput"],
  " are complex while the integration variable ",
  StyleBox["x", "InlineInput"],
  " is real can now be obtained using ",
  StyleBox["Timing", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ComplexExpand[\ Re[expr], \ {a, b, c}];\)\  // Timing\)], "Input"],

Cell[BoxData[
    \({0.015000000000000013`\ Second, Null}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Now we reload the package, declare ",
  StyleBox["x", "InlineInput"],
  " to be real, and evaluate ",
  StyleBox["Re[expr]", "InlineInput"],
  " using the rules from the package.  Note that ",
  StyleBox["ComplexExpand", "InlineInput",
    FontWeight->"Bold"],
  " is not needed when using the package."
}], "Text"],

Cell[BoxData[
    \(Needs["\<Algebra`ReIm`\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(x\  /: Im[x] = 0;\)\), "\n", 
    \(\(Re[expr];\)\  // Timing\)}], "Input"],

Cell[BoxData[
    \({0.5`\ Second, Null}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Thus, similar evaluations take much longer using ",
  StyleBox["ReIm", "InlineInput",
    FontWeight->"Bold"],
  " compared with ",
  StyleBox["ComplexExpand", "InlineInput",
    FontWeight->"Bold"],
  ".  Bahder ascribes this inefficiency to the need to test all of the rules \
associated with ",
  StyleBox["Re", "InlineInput",
    FontWeight->"Bold"],
  " that are contained in the package, whether or not they are actually \
relevant to the expression being evaluated.  Therefore, for most purposes it \
is better to simplify expressions manually, with appropriate replacement \
rules, or to supply a set of user-defined rules pertinent to the expressions \
encountered in one's applications.  Some guidance in constructing those rules \
can be obtained from the package itself because the package is a text file \
which can be edited by the user.  Of course, the user is advised to work on a \
copy rather than the original! "
}], "Text"]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
ShowSelection->True,
ShowCellLabel->False,
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Info3351094705-7946987"->{
    Cell[9931, 338, 63, 1, 34, "Print",
      CellTags->"Info3351094705-7946987"],
    Cell[9997, 341, 428, 13, 35, "Print",
      CellTags->"Info3351094705-7946987"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Info3351094705-7946987", 57707, 2065}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 83, 2, 81, "Title"],
Cell[1862, 57, 455, 15, 85, "Author"],
Cell[2320, 74, 253, 7, 71, "Abstract"],

Cell[CellGroupData[{
Cell[2598, 85, 53, 0, 83, "Section"],
Cell[2654, 87, 122, 2, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2813, 94, 31, 0, 51, "Section"],
Cell[2847, 96, 937, 22, 223, "Text"],

Cell[CellGroupData[{
Cell[3809, 122, 110, 2, 100, "Input"],
Cell[3922, 126, 50, 1, 64, "Output"]
}, Closed]],
Cell[3987, 130, 357, 8, 120, "Text"],

Cell[CellGroupData[{
Cell[4369, 142, 55, 1, 73, "Input"],
Cell[4427, 145, 38, 1, 59, "Output"]
}, Closed]],
Cell[4480, 149, 277, 6, 94, "Text"],

Cell[CellGroupData[{
Cell[4782, 159, 57, 1, 73, "Input"],
Cell[4842, 162, 55, 1, 75, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[4934, 168, 78, 1, 73, "Input"],
Cell[5015, 171, 49, 1, 59, "Output"]
}, Closed]],
Cell[5079, 175, 234, 6, 68, "Text"],

Cell[CellGroupData[{
Cell[5338, 185, 130, 2, 99, "Input"],
Cell[5471, 189, 262, 4, 132, "Output"]
}, Closed]],
Cell[5748, 196, 466, 13, 121, "Text"],

Cell[CellGroupData[{
Cell[6239, 213, 114, 2, 99, "Input"],
Cell[6356, 217, 172, 3, 61, "Output"]
}, Closed]],
Cell[6543, 223, 331, 8, 95, "Text"],

Cell[CellGroupData[{
Cell[6899, 235, 159, 3, 125, "Input"],
Cell[7061, 240, 234, 4, 87, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[7332, 249, 229, 4, 151, "Input"],
Cell[7564, 255, 258, 4, 61, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[7871, 265, 61, 0, 51, "Section"],
Cell[7935, 267, 571, 19, 120, "Text"],

Cell[CellGroupData[{
Cell[8531, 290, 170, 3, 125, "Input"],
Cell[8704, 295, 148, 2, 84, "Output"]
}, Closed]],
Cell[8867, 300, 451, 12, 121, "Text"],

Cell[CellGroupData[{
Cell[9343, 316, 256, 5, 177, "Input"],
Cell[9602, 323, 124, 2, 59, "Output"]
}, Closed]],
Cell[9741, 328, 123, 3, 67, "Text"],

Cell[CellGroupData[{
Cell[9889, 335, 39, 1, 73, "Input"],
Cell[9931, 338, 63, 1, 34, "Print",
  CellTags->"Info3351094705-7946987"],
Cell[9997, 341, 428, 13, 35, "Print",
  CellTags->"Info3351094705-7946987"]
}, Closed]],
Cell[10440, 357, 304, 7, 95, "Text"],
Cell[10747, 366, 281, 6, 203, "Input"],

Cell[CellGroupData[{
Cell[11053, 376, 73, 1, 73, "Input"],
Cell[11129, 379, 116, 2, 59, "Output"]
}, Closed]],
Cell[11260, 384, 755, 19, 174, "Text"],

Cell[CellGroupData[{
Cell[12040, 407, 43, 1, 73, "Input"],
Cell[12086, 410, 44, 1, 59, "Output"]
}, Closed]],
Cell[12145, 414, 209, 5, 68, "Text"],

Cell[CellGroupData[{
Cell[12379, 423, 106, 2, 73, "Input"],
Cell[12488, 427, 66, 1, 75, "Output"]
}, Closed]],
Cell[12569, 431, 168, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[12762, 440, 45, 1, 73, "Input"],
Cell[12810, 443, 68, 1, 59, "Output"]
}, Closed]],
Cell[12893, 447, 242, 7, 68, "Text"],

Cell[CellGroupData[{
Cell[13160, 458, 80, 1, 73, "Input"],
Cell[13243, 461, 57, 1, 59, "Output"]
}, Closed]],
Cell[13315, 465, 224, 6, 68, "Text"],

Cell[CellGroupData[{
Cell[13564, 475, 44, 1, 73, "Input"],
Cell[13611, 478, 208, 6, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[13856, 489, 56, 1, 73, "Input"],
Cell[13915, 492, 189, 6, 59, "Output"]
}, Closed]],
Cell[14119, 501, 505, 12, 68, "Text"],
Cell[14627, 515, 327, 6, 151, "Input"],

Cell[CellGroupData[{
Cell[14979, 525, 45, 1, 73, "Input"],
Cell[15027, 528, 57, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[15121, 534, 68, 1, 73, "Input"],
Cell[15192, 537, 54, 1, 59, "Output"]
}, Closed]],
Cell[15261, 541, 282, 7, 68, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[15580, 553, 33, 0, 54, "ExerciseMain"],
Cell[15616, 555, 256, 6, 58, "Problem"],
Cell[15875, 563, 223, 8, 58, "Problem"],
Cell[16101, 573, 173, 6, 58, "Problem"],
Cell[16277, 581, 599, 14, 107, "Problem"],
Cell[16879, 597, 338, 12, 58, "Problem"]
}, Closed]],

Cell[CellGroupData[{
Cell[17254, 614, 41, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[17320, 618, 33, 0, 68, "Subsection"],
Cell[17356, 620, 59, 1, 73, "Input"],
Cell[17418, 623, 353, 8, 68, "Text"],
Cell[17774, 633, 192, 4, 151, "Input"],

Cell[CellGroupData[{
Cell[17991, 641, 109, 2, 73, "Input"],
Cell[18103, 645, 109, 2, 61, "Output"]
}, Closed]],
Cell[18227, 650, 482, 9, 94, "Text"],

Cell[CellGroupData[{
Cell[18734, 663, 68, 1, 73, "Input"],
Cell[18805, 666, 76, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[18918, 672, 58, 1, 73, "Input"],
Cell[18979, 675, 106, 2, 75, "Output"]
}, Closed]],
Cell[19100, 680, 512, 16, 69, "Text"],
Cell[19615, 698, 154, 3, 125, "Input"],
Cell[19772, 703, 97, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[19894, 709, 52, 1, 73, "Input"],
Cell[19949, 712, 55, 1, 59, "Output"]
}, Closed]],
Cell[20019, 716, 213, 7, 42, "Text"],
Cell[20235, 725, 117, 2, 99, "Input"],
Cell[20355, 729, 20, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[20400, 733, 46, 1, 73, "Input"],
Cell[20449, 736, 42, 1, 59, "Output"]
}, Closed]],
Cell[20506, 740, 158, 5, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[20701, 750, 48, 0, 68, "Subsection"],
Cell[20752, 752, 508, 10, 94, "Text"],

Cell[CellGroupData[{
Cell[21285, 766, 111, 2, 101, "Input"],
Cell[21399, 770, 150, 2, 81, "Output"]
}, Closed]],
Cell[21564, 775, 426, 14, 69, "Text"],

Cell[CellGroupData[{
Cell[22015, 793, 88, 1, 73, "Input"],
Cell[22106, 796, 71, 1, 59, "Output"]
}, Closed]],
Cell[22192, 800, 334, 9, 42, "Text"],
Cell[22529, 811, 59, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[22613, 816, 98, 2, 99, "Input"],
Cell[22714, 820, 54, 1, 59, "Output"]
}, Closed]],
Cell[22783, 824, 962, 19, 146, "Text"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

