(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   2073214,      31366]*)
(*NotebookOutlinePosition[   2108510,      32628]*)
(*  CellTagsIndexPosition[   2108280,      32619]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Using Files", "Title",
  Evaluatable->False,
  TextAlignment->Center,
  TextJustification->0,
  AspectRatioFixed->True],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Off[General::spell, General::spell1]\)}], "Input"],

Cell[BoxData[
    \(Needs["\<Graphics`\>"]\)], "Input"],

Cell["\<\
The following strings identify some of the directories used in this \
lesson.  You will need to modify them for your own system!\
\>", "Text",
  FontSize->14,
  FontColor->RGBColor[1, 0, 0]],

Cell[BoxData[{
    \(\(EssentialMathematica = "\<C:\Documents and Settings\Jim\My \
Documents\My eBooks\EssentialMathematica\>";\)\), "\[IndentingNewLine]", 
    \(\(testdir = "\<C:\Documents and Settings\Jim\My \
Documents\ScratchTest\>";\)\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["File management", "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "All modern computer systems group files in a hierarchical organization of \
folders or directories, but the details of accessing these files depends upon \
the operating system.  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides functions which handle most of the system-dependent aspects of \
file management in a uniform fashion, except that pathnames remain system \
dependent."
}], "Text"],

Cell[CellGroupData[{

Cell["Directories", "Subsection"],

Cell["\<\
First, we need functions which access the directories.  Most of \
these can be found using the information escape with wildcards.\
\>", "Text"],

Cell[CellGroupData[{

Cell["?*Directory*", "Input",
  AspectRatioFixed->True],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["CopyDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`CopyDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["DirectoryStack",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`DirectoryStack"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ResetDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`ResetDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$InitialDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$InitialDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$TopDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$TopDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["CreateDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`CreateDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["HomeDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`HomeDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["SetDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`SetDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$InstallationDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$InstallationDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$UserAddOnsDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$UserAddOnsDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["DeleteDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`DeleteDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["NotebookDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`NotebookDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$AddOnsDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$AddOnsDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$LaunchDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$LaunchDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$UserBaseDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$UserBaseDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["Directory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817303-1111292", "System`Directory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParentDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`ParentDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$BaseDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$BaseDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$PreferencesDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$PreferencesDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""},
              {
                ButtonBox[
                  StyleBox["DirectoryName",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`DirectoryName"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["RenameDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`RenameDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$HomeDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$HomeDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$RootDirectory",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817303-1111292", 
                    "System`$RootDirectory"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1, 1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell["\<\
First, we ask for the current directory \[LongDash] your answer \
will almost certainly be different and provides a template for proper \
specification of other directory locations on your operating system.\
\>", \
"Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Directory[]\)], "Input"],

Cell[BoxData[
    \("C:\\Program Files\\Wolfram Research\\Mathematica\\5.2"\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Next, we choose a directory using ",
  ButtonBox["SetDirectory",
    ButtonStyle->"RefGuideLink"],
  ".  To do this you must use notation for a path that is appropriate to the \
operating system on your computer.  On a Windows system you might enter\n\t",
  StyleBox["SetDirectory[\"C:\\My Notebooks\"]", "InlineInput",
    FontWeight->"Bold"],
  "  \nif that folder exists.  Try setting the directory to the one \
containing the notebooks for this course.  As an example, here is the \
directory on my laptop."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(SetDirectory[EssentialMathematica]\)], "Input"],

Cell[BoxData[
    \("C:\\Documents and Settings\\Jim\\My Documents\\My \
eBooks\\EssentialMathematica"\)], "Output"]
}, Closed]],

Cell["\<\
Changing the working directory may interfere with hyperlinks in \
these course materials.  If so, you can reset the working directory by \
re-evaluating the initialization cell in contents.nb. \
\>", "Caution"],

Cell[TextData[{
  "You can use ",
  StyleBox["ResetDirectory[]", "InlineInput",
    FontWeight->"Bold"],
  " instead to restore the previous definition, which works if the directory \
has not been changed more than once.  Sorry for the inconvenience of \
resetting the working directory, but ",
  StyleBox["Mathematica ",
    FontSlant->"Italic"],
  "does not support symbolic names within hyperlinks."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Files", "Subsection"],

Cell["\<\
To obtain a list of most of the available file-manipulation \
functions, we again request information using wildcards.\
\>", "Text"],

Cell[CellGroupData[{

Cell["?*File*", "Input",
  AspectRatioFixed->True],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ContextToFilename",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`ContextToFilename"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["File",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`File"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileName",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`FileName"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["IncludeFileExtension",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`IncludeFileExtension"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["$PasswordFile",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`$PasswordFile"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["ContextToFileName",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`ContextToFileName"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileByteCount",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`FileByteCount"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileNameDialogSettings",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`FileNameDialogSettings"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["RenameFile",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`RenameFile"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""},
              {
                ButtonBox[
                  StyleBox["CopyFile",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`CopyFile"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileDate",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`FileDate"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileNames",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`FileNames"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["SetFileDate",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`SetFileDate"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""},
              {
                ButtonBox[
                  StyleBox["DeleteFile",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`DeleteFile"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileFormat",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`FileFormat"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileType",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`FileType"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["SetFileLoadingContext",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`SetFileLoadingContext"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""},
              {
                ButtonBox[
                  StyleBox["EndOfFile",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`EndOfFile"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["FileInformation",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`FileInformation"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["GetFileName",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350817488-9420471", 
                    "System`GetFileName"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ToFileName",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350817488-9420471", "System`ToFileName"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1, 1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell[TextData[{
  "Use ",
  ButtonBox["FileNames",
    ButtonStyle->"MainBookLink"],
  " with an empty argument to list all the files in the current directory: "
}], "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames[]\)], "Input"],

Cell[BoxData[
    \({"Additional topics", "Applications", "contents.nb", "Data analysis", 
      "EssentialMathematica.htm", "Physics problems", "Primary topics", 
      "Solutions", "TutorialJJK.nb"}\)], "Output"]
}, Closed]],

Cell["\<\
If a name or pattern is given, only matching files are listed.  The \
wildcard character (*) can be used to identify files by extension or by \
partial names.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames["\<*.nb\>"]\)], "Input"],

Cell[BoxData[
    \({"contents.nb", "TutorialJJK.nb"}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames["\<a*\>"]\)], "Input"],

Cell[BoxData[
    \({"Additional topics", "Applications"}\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["FileNames[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["forms,directories,n", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " searches a list of ",
  StyleBox["directories", "InlineInput"],
  ", including ",
  StyleBox["n", "InlineInput"],
  " levels of subdirectories, for files matching any member of list of ",
  StyleBox["forms", "InlineInput"],
  StyleBox[".",
    FontSlant->"Italic"],
  "  The curly braces may be omitted when a list contains only one element."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames[{"\<*.nb\>", "\<*.txt\>"}, {"\<Applications\>", "\<Additional \
topics\>"}] // ColumnForm\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {"\<\"Additional topics\\\\complex.nb\"\>"},
          {"\<\"Additional topics\\\\dimensional analysis.nb\"\>"},
          {"\<\"Additional topics\\\\files.nb\"\>"},
          {"\<\"Additional topics\\\\FractalsPrototype.nb\"\>"},
          {"\<\"Additional topics\\\\InPolygonQ.nb\"\>"},
          {"\<\"Additional topics\\\\packaging.nb\"\>"},
          {"\<\"Additional topics\\\\popclock.txt\"\>"},
          {"\<\"Additional topics\\\\recurrence relations.nb\"\>"},
          {"\<\"Additional topics\\\\sample data types.txt\"\>"},
          {"\<\"Additional topics\\\\worldhis.txt\"\>"},
          {"\<\"Additional topics\\\\worldpop.txt\"\>"},
          {"\<\"Applications\\\\cicada.nb\"\>"},
          {"\<\"Applications\\\\cobweb.nb\"\>"},
          {"\<\"Applications\\\\collisions.nb\"\>"},
          {"\<\"Applications\\\\duffing.nb\"\>"},
          {"\<\"Applications\\\\FN model for BZ reaction.nb\"\>"},
          {"\<\"Applications\\\\frequency analysis.nb\"\>"},
          {"\<\"Applications\\\\henon.nb\"\>"},
          {"\<\"Applications\\\\Home Run.nb\"\>"},
          {"\<\"Applications\\\\JuliaIIM.nb\"\>"},
          {"\<\"Applications\\\\Julia.nb\"\>"},
          {"\<\"Applications\\\\Koch.nb\"\>"},
          {"\<\"Applications\\\\logistic map.nb\"\>"},
          {"\<\"Applications\\\\markov.nb\"\>"},
          {"\<\"Applications\\\\Paschen.nb\"\>"},
          {"\<\"Applications\\\\pendulum.nb\"\>"},
          {"\<\"Applications\\\\van der Pol.nb\"\>"},
          {"\<\"Applications\\\\Volterra.nb\"\>"}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      ColumnForm[ {"Additional topics\\complex.nb", 
        "Additional topics\\dimensional analysis.nb", 
        "Additional topics\\files.nb", 
        "Additional topics\\FractalsPrototype.nb", 
        "Additional topics\\InPolygonQ.nb", "Additional topics\\packaging.nb",
         "Additional topics\\popclock.txt", 
        "Additional topics\\recurrence relations.nb", 
        "Additional topics\\sample data types.txt", 
        "Additional topics\\worldhis.txt", "Additional topics\\worldpop.txt", 
        "Applications\\cicada.nb", "Applications\\cobweb.nb", 
        "Applications\\collisions.nb", "Applications\\duffing.nb", 
        "Applications\\FN model for BZ reaction.nb", 
        "Applications\\frequency analysis.nb", "Applications\\henon.nb", 
        "Applications\\Home Run.nb", "Applications\\JuliaIIM.nb", 
        "Applications\\Julia.nb", "Applications\\Koch.nb", 
        "Applications\\logistic map.nb", "Applications\\markov.nb", 
        "Applications\\Paschen.nb", "Applications\\pendulum.nb", 
        "Applications\\van der Pol.nb", "Applications\\Volterra.nb"}],
      Editable->False]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames["\<*.m\>", "\<.\>", 2]\)], "Input"],

Cell[BoxData[
    \({".\\Primary topics\\Fractals.m"}\)], "Output"]
}, Closed]],

Cell["Create a new directory and manipulate some files", "Exercise"],

Cell[TextData[{
  "For the purpose of exercising our file-management skills, let's create a \
subdirectory called ScratchTest ",
  ButtonBox["CreateDirectory",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"dir\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ".  The following example applies to my computer \[LongDash] your input \
will vary."
}], "ExerciseText",
  CellDingbat->"\[FilledSmallSquare]"],

Cell[BoxData[
    \(\(CreateDirectory[testdir];\)\)], "Input"],

Cell["\<\
Next, reset the working directory and verify that the new directory \
was created empty.\
\>", "ExerciseText",
  CellDingbat->"\[FilledSmallSquare]"],

Cell[CellGroupData[{

Cell[BoxData[
    \(SetDirectory[testdir]\)], "Input"],

Cell[BoxData[
    \("C:\\Documents and Settings\\Jim\\My Documents\\ScratchTest"\)], \
"Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames[]\)], "Input"],

Cell[BoxData[
    \({}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Then, copy a file into the new directory using ",
  ButtonBox["CopyFile",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"oldfile\",\"newfile\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "where ",
  StyleBox["oldfile", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "is an existing file and ",
  StyleBox["newfile", "InlineInput",
    FontWeight->"Bold"],
  " does not exist."
}], "ExerciseText",
  CellDingbat->"\[FilledSmallSquare]"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CopyFile[
      EssentialMathematica <> "\<\Applications\Julia.nb\>", \
"\<temp.np\>"]\)], "Input"],

Cell[BoxData[
    \("temp.np"\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FileNames[]\)], "Input"],

Cell[BoxData[
    \({"temp.np"}\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["OOPS!",
    FontWeight->"Bold",
    FontColor->RGBColor[1, 0, 0]],
  " The file name is not correct.  Use ",
  ButtonBox["RenameFile",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"oldfile\",\"newfile\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " to correct the problem."
}], "ExerciseText",
  CellDingbat->"\[FilledSmallSquare]"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RenameFile["\<temp.np\>", \ "\<temp.nb\>"]\)], "Input"],

Cell[BoxData[
    \("temp.nb"\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Now open the file and verify that it is correct.  Then close the file and \
use ",
  ButtonBox["DeleteFile",
    ButtonStyle->"RefGuideLink"],
  " to remove it; verify that it disappears.  ",
  StyleBox["Be careful!",
    FontWeight->"Bold",
    FontColor->RGBColor[1, 0, 0]]
}], "ExerciseText",
  CellDingbat->"\[FilledSmallSquare]"]
}, Closed]],

Cell[CellGroupData[{

Cell["Paths", "Subsection"],

Cell[TextData[{
  "Files can often be found without specifying their locations completely if \
the directory containing them is included in ",
  StyleBox["$Path", "InlineInput",
    FontWeight->"Bold"],
  ".  Thus, normally we need not specify the directory for one of the \
standard packages or for one of the extra packages installed with ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " because the relevant directories are usually included in the path.  At \
the time this was written, the path was as follows."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \($Path\)], "Input"],

Cell[BoxData[
    \({"C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\JLink", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\NETLink", 
      "C:\\Documents and Settings\\Jim\\Application \
Data\\Mathematica\\Kernel", 
      "C:\\Documents and Settings\\Jim\\Application \
Data\\Mathematica\\Autoload", 
      "C:\\Documents and Settings\\Jim\\Application \
Data\\Mathematica\\Applications", 
      "C:\\Documents and Settings\\All Users\\Application \
Data\\Mathematica\\Kernel", 
      "C:\\Documents and Settings\\All Users\\Application \
Data\\Mathematica\\Autoload", 
      "C:\\Documents and Settings\\All Users\\Application \
Data\\Mathematica\\Applications", ".", "C:\\Documents and Settings\\Jim", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\StandardPackages", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\StandardPackages\\StartUp", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\Autoload", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\Applications", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\ExtraPackages", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\SystemFiles\\Graphics\\Packages", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\Configuration\\Kernel"}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Suppose that you have created a directory that contains packages you have \
either written or acquired from external sources.  You may then find it \
convenient to append that directory to the path string.  For the purposes of \
illustration, we copy a file into ",
  StyleBox["testdir", "InlineInput"],
  " and add that directory to the path."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(CopyFile[EssentialMathematica <> "\<\Additional topics\Fractals.m\>", 
      testdir <> "\<\Fractals.m\>"]\)], "Input"],

Cell[BoxData[
    \("C:\\Documents and Settings\\Jim\\My \
Documents\\ScratchTest\\Fractals.m"\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(AppendTo[$Path, testdir]\)], "Input"],

Cell[BoxData[
    \({"C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\JLink", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\NETLink", 
      "C:\\Documents and Settings\\Jim\\Application \
Data\\Mathematica\\Kernel", 
      "C:\\Documents and Settings\\Jim\\Application \
Data\\Mathematica\\Autoload", 
      "C:\\Documents and Settings\\Jim\\Application \
Data\\Mathematica\\Applications", 
      "C:\\Documents and Settings\\All Users\\Application \
Data\\Mathematica\\Kernel", 
      "C:\\Documents and Settings\\All Users\\Application \
Data\\Mathematica\\Autoload", 
      "C:\\Documents and Settings\\All Users\\Application \
Data\\Mathematica\\Applications", ".", "C:\\Documents and Settings\\Jim", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\StandardPackages", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\StandardPackages\\StartUp", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\Autoload", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\Applications", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\AddOns\\ExtraPackages", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\SystemFiles\\Graphics\\Packages", 
      "C:\\Program Files\\Wolfram \
Research\\Mathematica\\5.2\\Configuration\\Kernel", 
      "C:\\Documents and Settings\\Jim\\My Documents\\ScratchTest"}\)], \
"Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \($ContextPath\)], "Input"],

Cell[BoxData[
    \({"Graphics`ThreeScript`", "Graphics`SurfaceOfRevolution`", 
      "Graphics`Spline`", "Graphics`Shapes`", "Graphics`Polyhedra`", 
      "Graphics`PlotField`", "Graphics`PlotField3D`", 
      "Graphics`ParametricPlot3D`", "Graphics`MultipleListPlot`", 
      "Graphics`Legend`", "Graphics`InequalityGraphics`", 
      "Graphics`ImplicitPlot`", "Graphics`Graphics`", "Graphics`Graphics3D`", 
      "Graphics`FilledPlot`", "Graphics`ContourPlot3D`", 
      "Graphics`ComplexMap`", "Graphics`Common`GraphicsCommon`", 
      "Graphics`Colors`", "Graphics`Arrow`", "Graphics`ArgColors`", 
      "Graphics`Animation`", "Global`", "System`"}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Any package in that directory can now be loaded without including the \
directory name.  If you know the context, you can use ",
  StyleBox["Needs",
    FontWeight->"Bold"],
  "; otherwise use ",
  StyleBox["Get", "InlineInput",
    FontWeight->"Bold"],
  " (",
  StyleBox["<<", "InlineInput"],
  ") to load the appropriate file and then inspect ",
  StyleBox["$ContextPath", "InlineInput",
    FontWeight->"Bold"],
  " to determine the context."
}], "Text"],

Cell[BoxData[
    \(Needs["\<Fractals`\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \($ContextPath\)], "Input"],

Cell[BoxData[
    \({"Fractals`", "Graphics`ThreeScript`", "Graphics`SurfaceOfRevolution`", 
      "Graphics`Spline`", "Graphics`Shapes`", "Graphics`Polyhedra`", 
      "Graphics`PlotField`", "Graphics`PlotField3D`", 
      "Graphics`ParametricPlot3D`", "Graphics`MultipleListPlot`", 
      "Graphics`Legend`", "Graphics`InequalityGraphics`", 
      "Graphics`ImplicitPlot`", "Graphics`Graphics`", "Graphics`Graphics3D`", 
      "Graphics`FilledPlot`", "Graphics`ContourPlot3D`", 
      "Graphics`ComplexMap`", "Graphics`Common`GraphicsCommon`", 
      "Graphics`Colors`", "Graphics`Arrow`", "Graphics`ArgColors`", 
      "Graphics`Animation`", "Global`", "System`"}\)], "Output"]
}, Closed]],

Cell["\<\
We now find several new functions, some with usage messages.  \
Options are also listed but usage messages were not provided for them.\
\>", \
"Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Fractals`*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["Fractals`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["CloseFigure",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350915520-3336292", 
                    "Fractals`CloseFigure"},
                  ButtonFrame->"None",
                  ButtonNote->"Fractals`"], 
                ButtonBox[
                  StyleBox["DisplayKoch",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350915520-3336292", 
                    "Fractals`DisplayKoch"},
                  ButtonFrame->"None",
                  ButtonNote->"Fractals`"], 
                ButtonBox[
                  StyleBox["juliaRate",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350915520-3336292", 
                    "Fractals`juliaRate"},
                  ButtonFrame->"None",
                  ButtonNote->"Fractals`"], 
                ButtonBox[
                  StyleBox["KochType",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3350915520-3336292", "Fractals`KochType"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Fractals`"], 
                ButtonBox[
                  StyleBox["mandelbrotRate",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350915520-3336292", 
                    "Fractals`mandelbrotRate"},
                  ButtonFrame->"None",
                  ButtonNote->"Fractals`"], 
                ButtonBox[
                  StyleBox["ShowIterations",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3350915520-3336292", 
                    "Fractals`ShowIterations"},
                  ButtonFrame->"None",
                  ButtonNote->"Fractals`"], "", ""}
              },
            RowMinHeight->1.05]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}],

Cell[BoxData[
    \("DisplayKoch[polygon,n,opts] produces a Koch curve of order n based \
upon the list of points in polygon. By default, the triangle rule is applied \
in the outward direction assuming that the polygon is defined \
counterclockwise.  \n  Options: \n   KochType\[Rule]\"sieve\" applies  the \
triangle-building rule in reverse.  \n  ShowIterations\[Rule]True can be used \
to display intermediate iterations. \n  CloseFigure\[Rule]True connects the \
last point to the first to close the figure."\)], "Print",
  CellTags->"Info3350915520-3336292"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayKoch[{{0. , 0. }, {0.5, \@3\/2. }, {1. , 0. }}, 4, 
        ShowIterations \[Rule] False, KochType \[Rule] "\<sieve\>", 
        CloseFigure \[Rule] True];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .86603 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0206197 0.952381 [
[ 0 0 0 0 ]
[ 1 .86603 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .86603 L
0 .86603 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .02062 m
.97031 .0308 L
s
.97031 .0308 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .04098 L
s
.96443 .04098 m
.95855 .05117 L
s
.95855 .05117 m
.9468 .05117 L
s
.9468 .05117 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .05117 L
s
.93504 .05117 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .06135 L
s
.92916 .06135 m
.92328 .07153 L
s
.92328 .07153 m
.93504 .07153 L
s
.93504 .07153 m
.94092 .08171 L
s
.94092 .08171 m
.93504 .0919 L
s
.93504 .0919 m
.92328 .0919 L
s
.92328 .0919 m
.92916 .10208 L
s
.92916 .10208 m
.92328 .11226 L
s
.92328 .11226 m
.91152 .11226 L
s
.91152 .11226 m
.90564 .10208 L
s
.90564 .10208 m
.89976 .11226 L
s
.89976 .11226 m
.88801 .11226 L
s
.88801 .11226 m
.88213 .10208 L
s
.88213 .10208 m
.88801 .0919 L
s
.88801 .0919 m
.87625 .0919 L
s
.87625 .0919 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .0919 L
s
.86449 .0919 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .10208 L
s
.85861 .10208 m
.85273 .11226 L
s
.85273 .11226 m
.84098 .11226 L
s
.84098 .11226 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .11226 L
s
.82922 .11226 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .12245 L
s
.82334 .12245 m
.81746 .13263 L
s
.81746 .13263 m
.82922 .13263 L
s
.82922 .13263 m
.8351 .14281 L
s
.8351 .14281 m
.82922 .15299 L
s
.82922 .15299 m
.81746 .15299 L
s
.81746 .15299 m
.82334 .16318 L
s
.82334 .16318 m
.81746 .17336 L
s
.81746 .17336 m
.82922 .17336 L
s
.82922 .17336 m
.8351 .18354 L
s
.8351 .18354 m
.84098 .17336 L
s
.84098 .17336 m
.85273 .17336 L
s
.85273 .17336 m
.85861 .18354 L
s
.85861 .18354 m
.85273 .19372 L
s
.85273 .19372 m
.86449 .19372 L
s
.86449 .19372 m
.87037 .20391 L
s
.87037 .20391 m
.86449 .21409 L
s
.86449 .21409 m
.85273 .21409 L
s
.85273 .21409 m
.85861 .22427 L
s
.85861 .22427 m
.85273 .23445 L
s
.85273 .23445 m
.84098 .23445 L
s
.84098 .23445 m
.8351 .22427 L
s
.8351 .22427 m
.82922 .23445 L
s
.82922 .23445 m
.81746 .23445 L
s
.81746 .23445 m
.82334 .24464 L
s
.82334 .24464 m
.81746 .25482 L
s
.81746 .25482 m
.82922 .25482 L
s
.82922 .25482 m
.8351 .265 L
s
.8351 .265 m
.82922 .27518 L
s
.82922 .27518 m
.81746 .27518 L
s
.81746 .27518 m
.82334 .28537 L
s
.82334 .28537 m
.81746 .29555 L
s
.81746 .29555 m
.8057 .29555 L
s
.8057 .29555 m
.79982 .28537 L
s
.79982 .28537 m
.79394 .29555 L
s
.79394 .29555 m
.78219 .29555 L
s
.78219 .29555 m
.77631 .28537 L
s
.77631 .28537 m
.78219 .27518 L
s
.78219 .27518 m
.77043 .27518 L
s
.77043 .27518 m
.76455 .265 L
s
.76455 .265 m
.75867 .27518 L
s
.75867 .27518 m
.74691 .27518 L
s
.74691 .27518 m
.75279 .28537 L
s
.75279 .28537 m
.74691 .29555 L
s
.74691 .29555 m
.73516 .29555 L
s
.73516 .29555 m
.72928 .28537 L
s
.72928 .28537 m
.7234 .29555 L
s
.7234 .29555 m
.71164 .29555 L
s
.71164 .29555 m
.70576 .28537 L
s
.70576 .28537 m
.71164 .27518 L
s
.71164 .27518 m
.69988 .27518 L
s
.69988 .27518 m
.694 .265 L
s
.694 .265 m
.69988 .25482 L
s
.69988 .25482 m
.71164 .25482 L
s
.71164 .25482 m
.70576 .24464 L
s
.70576 .24464 m
.71164 .23445 L
s
.71164 .23445 m
.69988 .23445 L
s
.69988 .23445 m
.694 .22427 L
s
.694 .22427 m
.68812 .23445 L
s
.68812 .23445 m
.67637 .23445 L
s
.67637 .23445 m
.67049 .22427 L
s
.67049 .22427 m
.67637 .21409 L
s
.67637 .21409 m
.66461 .21409 L
s
.66461 .21409 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .21409 L
s
.65285 .21409 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .22427 L
s
.64697 .22427 m
.64109 .23445 L
s
.64109 .23445 m
.62934 .23445 L
s
.62934 .23445 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .23445 L
s
.61758 .23445 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .24464 L
s
.6117 .24464 m
.60582 .25482 L
s
.60582 .25482 m
.61758 .25482 L
s
.61758 .25482 m
.62346 .265 L
s
.62346 .265 m
.61758 .27518 L
s
.61758 .27518 m
.60582 .27518 L
s
.60582 .27518 m
.6117 .28537 L
s
.6117 .28537 m
.60582 .29555 L
s
.60582 .29555 m
.59406 .29555 L
s
.59406 .29555 m
.58818 .28537 L
s
.58818 .28537 m
.5823 .29555 L
s
.5823 .29555 m
.57055 .29555 L
s
.57055 .29555 m
.56467 .28537 L
s
.56467 .28537 m
.57055 .27518 L
s
.57055 .27518 m
.55879 .27518 L
s
.55879 .27518 m
.55291 .265 L
s
.55291 .265 m
.54703 .27518 L
s
.54703 .27518 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .28537 L
s
.54115 .28537 m
.53527 .29555 L
s
.53527 .29555 m
.52352 .29555 L
s
.52352 .29555 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .29555 L
s
.51176 .29555 m
.5 .29555 L
s
.5 .29555 m
.50588 .30573 L
s
.50588 .30573 m
.5 .31591 L
s
.5 .31591 m
.51176 .31591 L
s
.51176 .31591 m
.51764 .3261 L
s
.51764 .3261 m
.51176 .33628 L
s
.51176 .33628 m
.5 .33628 L
s
.5 .33628 m
.50588 .34646 L
s
.50588 .34646 m
.5 .35664 L
s
.5 .35664 m
.51176 .35664 L
s
.51176 .35664 m
.51764 .36683 L
s
.51764 .36683 m
.52352 .35664 L
s
.52352 .35664 m
.53527 .35664 L
s
.53527 .35664 m
.54115 .36683 L
s
.54115 .36683 m
.53527 .37701 L
s
.53527 .37701 m
.54703 .37701 L
s
.54703 .37701 m
.55291 .38719 L
s
.55291 .38719 m
.54703 .39737 L
s
.54703 .39737 m
.53527 .39737 L
s
.53527 .39737 m
.54115 .40756 L
s
.54115 .40756 m
.53527 .41774 L
s
.53527 .41774 m
.52352 .41774 L
s
.52352 .41774 m
.51764 .40756 L
s
.51764 .40756 m
.51176 .41774 L
s
.51176 .41774 m
.5 .41774 L
s
.5 .41774 m
.50588 .42792 L
s
.50588 .42792 m
.5 .4381 L
s
.5 .4381 m
.51176 .4381 L
s
.51176 .4381 m
.51764 .44829 L
s
.51764 .44829 m
.51176 .45847 L
s
.51176 .45847 m
.5 .45847 L
s
.5 .45847 m
.50588 .46865 L
s
.50588 .46865 m
.5 .47883 L
s
.5 .47883 m
.51176 .47883 L
s
.51176 .47883 m
.51764 .48902 L
s
.51764 .48902 m
.52352 .47883 L
s
.52352 .47883 m
.53527 .47883 L
s
.53527 .47883 m
.54115 .48902 L
s
.54115 .48902 m
.53527 .4992 L
s
.53527 .4992 m
.54703 .4992 L
s
.54703 .4992 m
.55291 .50938 L
s
.55291 .50938 m
.55879 .4992 L
s
.55879 .4992 m
.57055 .4992 L
s
.57055 .4992 m
.56467 .48902 L
s
.56467 .48902 m
.57055 .47883 L
s
.57055 .47883 m
.5823 .47883 L
s
.5823 .47883 m
.58818 .48902 L
s
.58818 .48902 m
.59406 .47883 L
s
.59406 .47883 m
.60582 .47883 L
s
.60582 .47883 m
.6117 .48902 L
s
.6117 .48902 m
.60582 .4992 L
s
.60582 .4992 m
.61758 .4992 L
s
.61758 .4992 m
.62346 .50938 L
s
.62346 .50938 m
.61758 .51956 L
s
.61758 .51956 m
.60582 .51956 L
s
.60582 .51956 m
.6117 .52975 L
s
.6117 .52975 m
.60582 .53993 L
s
.60582 .53993 m
.61758 .53993 L
s
.61758 .53993 m
.62346 .55011 L
s
.62346 .55011 m
.62934 .53993 L
s
.62934 .53993 m
.64109 .53993 L
s
.64109 .53993 m
.64697 .55011 L
s
.64697 .55011 m
.64109 .56029 L
s
.64109 .56029 m
.65285 .56029 L
s
.65285 .56029 m
.65873 .57048 L
s
.65873 .57048 m
.65285 .58066 L
s
.65285 .58066 m
.64109 .58066 L
s
.64109 .58066 m
.64697 .59084 L
s
.64697 .59084 m
.64109 .60102 L
s
.64109 .60102 m
.62934 .60102 L
s
.62934 .60102 m
.62346 .59084 L
s
.62346 .59084 m
.61758 .60102 L
s
.61758 .60102 m
.60582 .60102 L
s
.60582 .60102 m
.6117 .61121 L
s
.6117 .61121 m
.60582 .62139 L
s
.60582 .62139 m
.61758 .62139 L
s
.61758 .62139 m
.62346 .63157 L
s
.62346 .63157 m
.61758 .64175 L
s
.61758 .64175 m
.60582 .64175 L
s
.60582 .64175 m
.6117 .65194 L
s
.6117 .65194 m
.60582 .66212 L
s
.60582 .66212 m
.59406 .66212 L
s
.59406 .66212 m
.58818 .65194 L
s
.58818 .65194 m
.5823 .66212 L
s
.5823 .66212 m
.57055 .66212 L
s
.57055 .66212 m
.56467 .65194 L
s
.56467 .65194 m
.57055 .64175 L
s
.57055 .64175 m
.55879 .64175 L
s
.55879 .64175 m
.55291 .63157 L
s
.55291 .63157 m
.54703 .64175 L
s
.54703 .64175 m
.53527 .64175 L
s
.53527 .64175 m
.54115 .65194 L
s
.54115 .65194 m
.53527 .66212 L
s
.53527 .66212 m
.52352 .66212 L
s
.52352 .66212 m
.51764 .65194 L
s
.51764 .65194 m
.51176 .66212 L
s
.51176 .66212 m
.5 .66212 L
s
.5 .66212 m
.50588 .6723 L
s
.50588 .6723 m
.5 .68249 L
s
.5 .68249 m
.51176 .68249 L
s
.51176 .68249 m
.51764 .69267 L
s
.51764 .69267 m
.51176 .70285 L
s
.51176 .70285 m
.5 .70285 L
s
.5 .70285 m
.50588 .71303 L
s
.50588 .71303 m
.5 .72322 L
s
.5 .72322 m
.51176 .72322 L
s
.51176 .72322 m
.51764 .7334 L
s
.51764 .7334 m
.52352 .72322 L
s
.52352 .72322 m
.53527 .72322 L
s
.53527 .72322 m
.54115 .7334 L
s
.54115 .7334 m
.53527 .74358 L
s
.53527 .74358 m
.54703 .74358 L
s
.54703 .74358 m
.55291 .75376 L
s
.55291 .75376 m
.54703 .76395 L
s
.54703 .76395 m
.53527 .76395 L
s
.53527 .76395 m
.54115 .77413 L
s
.54115 .77413 m
.53527 .78431 L
s
.53527 .78431 m
.52352 .78431 L
s
.52352 .78431 m
.51764 .77413 L
s
.51764 .77413 m
.51176 .78431 L
s
.51176 .78431 m
.5 .78431 L
s
.5 .78431 m
.50588 .79449 L
s
.50588 .79449 m
.5 .80468 L
s
.5 .80468 m
.51176 .80468 L
s
.51176 .80468 m
.51764 .81486 L
s
.51764 .81486 m
.51176 .82504 L
s
.51176 .82504 m
.5 .82504 L
s
.5 .82504 m
.50588 .83522 L
s
.50588 .83522 m
.5 .84541 L
s
.5 .84541 m
.49412 .83522 L
s
.49412 .83522 m
.5 .82504 L
s
.5 .82504 m
.48824 .82504 L
s
.48824 .82504 m
.48236 .81486 L
s
.48236 .81486 m
.48824 .80468 L
s
.48824 .80468 m
.5 .80468 L
s
.5 .80468 m
.49412 .79449 L
s
.49412 .79449 m
.5 .78431 L
s
.5 .78431 m
.48824 .78431 L
s
.48824 .78431 m
.48236 .77413 L
s
.48236 .77413 m
.47648 .78431 L
s
.47648 .78431 m
.46473 .78431 L
s
.46473 .78431 m
.45885 .77413 L
s
.45885 .77413 m
.46473 .76395 L
s
.46473 .76395 m
.45297 .76395 L
s
.45297 .76395 m
.44709 .75376 L
s
.44709 .75376 m
.45297 .74358 L
s
.45297 .74358 m
.46473 .74358 L
s
.46473 .74358 m
.45885 .7334 L
s
.45885 .7334 m
.46473 .72322 L
s
.46473 .72322 m
.47648 .72322 L
s
.47648 .72322 m
.48236 .7334 L
s
.48236 .7334 m
.48824 .72322 L
s
.48824 .72322 m
.5 .72322 L
s
.5 .72322 m
.49412 .71303 L
s
.49412 .71303 m
.5 .70285 L
s
.5 .70285 m
.48824 .70285 L
s
.48824 .70285 m
.48236 .69267 L
s
.48236 .69267 m
.48824 .68249 L
s
.48824 .68249 m
.5 .68249 L
s
.5 .68249 m
.49412 .6723 L
s
.49412 .6723 m
.5 .66212 L
s
.5 .66212 m
.48824 .66212 L
s
.48824 .66212 m
.48236 .65194 L
s
.48236 .65194 m
.47648 .66212 L
s
.47648 .66212 m
.46473 .66212 L
s
.46473 .66212 m
.45885 .65194 L
s
.45885 .65194 m
.46473 .64175 L
s
.46473 .64175 m
.45297 .64175 L
s
.45297 .64175 m
.44709 .63157 L
s
.44709 .63157 m
.44121 .64175 L
s
.44121 .64175 m
.42945 .64175 L
s
.42945 .64175 m
.43533 .65194 L
s
.43533 .65194 m
.42945 .66212 L
s
.42945 .66212 m
.4177 .66212 L
s
.4177 .66212 m
.41182 .65194 L
s
.41182 .65194 m
.40594 .66212 L
s
.40594 .66212 m
.39418 .66212 L
s
.39418 .66212 m
.3883 .65194 L
s
.3883 .65194 m
.39418 .64175 L
s
.39418 .64175 m
.38242 .64175 L
s
.38242 .64175 m
.37654 .63157 L
s
.37654 .63157 m
.38242 .62139 L
s
.38242 .62139 m
.39418 .62139 L
s
.39418 .62139 m
.3883 .61121 L
s
.3883 .61121 m
.39418 .60102 L
s
.39418 .60102 m
.38242 .60102 L
s
.38242 .60102 m
.37654 .59084 L
s
.37654 .59084 m
.37066 .60102 L
s
.37066 .60102 m
.35891 .60102 L
s
.35891 .60102 m
.35303 .59084 L
s
.35303 .59084 m
.35891 .58066 L
s
.35891 .58066 m
.34715 .58066 L
s
.34715 .58066 m
.34127 .57048 L
s
.34127 .57048 m
.34715 .56029 L
s
.34715 .56029 m
.35891 .56029 L
s
.35891 .56029 m
.35303 .55011 L
s
.35303 .55011 m
.35891 .53993 L
s
.35891 .53993 m
.37066 .53993 L
s
.37066 .53993 m
.37654 .55011 L
s
.37654 .55011 m
.38242 .53993 L
s
.38242 .53993 m
.39418 .53993 L
s
.39418 .53993 m
.3883 .52975 L
s
.3883 .52975 m
.39418 .51956 L
s
.39418 .51956 m
.38242 .51956 L
s
.38242 .51956 m
.37654 .50938 L
s
.37654 .50938 m
.38242 .4992 L
s
.38242 .4992 m
.39418 .4992 L
s
.39418 .4992 m
.3883 .48902 L
s
.3883 .48902 m
.39418 .47883 L
s
.39418 .47883 m
.40594 .47883 L
s
.40594 .47883 m
.41182 .48902 L
s
.41182 .48902 m
.4177 .47883 L
s
.4177 .47883 m
.42945 .47883 L
s
.42945 .47883 m
.43533 .48902 L
s
.43533 .48902 m
.42945 .4992 L
s
.42945 .4992 m
.44121 .4992 L
s
.44121 .4992 m
.44709 .50938 L
s
.44709 .50938 m
.45297 .4992 L
s
.45297 .4992 m
.46473 .4992 L
s
.46473 .4992 m
.45885 .48902 L
s
.45885 .48902 m
.46473 .47883 L
s
.46473 .47883 m
.47648 .47883 L
s
.47648 .47883 m
.48236 .48902 L
s
.48236 .48902 m
.48824 .47883 L
s
.48824 .47883 m
.5 .47883 L
s
.5 .47883 m
.49412 .46865 L
s
.49412 .46865 m
.5 .45847 L
s
.5 .45847 m
.48824 .45847 L
s
.48824 .45847 m
.48236 .44829 L
s
.48236 .44829 m
.48824 .4381 L
s
.48824 .4381 m
.5 .4381 L
s
.5 .4381 m
.49412 .42792 L
s
.49412 .42792 m
.5 .41774 L
s
.5 .41774 m
.48824 .41774 L
s
.48824 .41774 m
.48236 .40756 L
s
.48236 .40756 m
.47648 .41774 L
s
.47648 .41774 m
.46473 .41774 L
s
.46473 .41774 m
.45885 .40756 L
s
.45885 .40756 m
.46473 .39737 L
s
.46473 .39737 m
.45297 .39737 L
s
.45297 .39737 m
.44709 .38719 L
s
.44709 .38719 m
.45297 .37701 L
s
.45297 .37701 m
.46473 .37701 L
s
.46473 .37701 m
.45885 .36683 L
s
.45885 .36683 m
.46473 .35664 L
s
.46473 .35664 m
.47648 .35664 L
s
.47648 .35664 m
.48236 .36683 L
s
.48236 .36683 m
.48824 .35664 L
s
.48824 .35664 m
.5 .35664 L
s
.5 .35664 m
.49412 .34646 L
s
.49412 .34646 m
.5 .33628 L
s
.5 .33628 m
.48824 .33628 L
s
.48824 .33628 m
.48236 .3261 L
s
.48236 .3261 m
.48824 .31591 L
s
.48824 .31591 m
.5 .31591 L
s
.5 .31591 m
.49412 .30573 L
s
.49412 .30573 m
.5 .29555 L
s
.5 .29555 m
.48824 .29555 L
s
.48824 .29555 m
.48236 .28537 L
s
.48236 .28537 m
.47648 .29555 L
s
.47648 .29555 m
.46473 .29555 L
s
.46473 .29555 m
.45885 .28537 L
s
.45885 .28537 m
.46473 .27518 L
s
.46473 .27518 m
.45297 .27518 L
s
.45297 .27518 m
.44709 .265 L
s
.44709 .265 m
.44121 .27518 L
s
.44121 .27518 m
.42945 .27518 L
s
.42945 .27518 m
.43533 .28537 L
s
.43533 .28537 m
.42945 .29555 L
s
.42945 .29555 m
.4177 .29555 L
s
.4177 .29555 m
.41182 .28537 L
s
.41182 .28537 m
.40594 .29555 L
s
.40594 .29555 m
.39418 .29555 L
s
.39418 .29555 m
.3883 .28537 L
s
.3883 .28537 m
.39418 .27518 L
s
.39418 .27518 m
.38242 .27518 L
s
.38242 .27518 m
.37654 .265 L
s
.37654 .265 m
.38242 .25482 L
s
.38242 .25482 m
.39418 .25482 L
s
.39418 .25482 m
.3883 .24464 L
s
.3883 .24464 m
.39418 .23445 L
s
.39418 .23445 m
.38242 .23445 L
s
.38242 .23445 m
.37654 .22427 L
s
.37654 .22427 m
.37066 .23445 L
s
.37066 .23445 m
.35891 .23445 L
s
.35891 .23445 m
.35303 .22427 L
s
.35303 .22427 m
.35891 .21409 L
s
.35891 .21409 m
.34715 .21409 L
s
.34715 .21409 m
.34127 .20391 L
s
.34127 .20391 m
.33539 .21409 L
s
.33539 .21409 m
.32363 .21409 L
s
.32363 .21409 m
.32951 .22427 L
s
.32951 .22427 m
.32363 .23445 L
s
.32363 .23445 m
.31188 .23445 L
s
.31188 .23445 m
.306 .22427 L
s
.306 .22427 m
.30012 .23445 L
s
.30012 .23445 m
.28836 .23445 L
s
.28836 .23445 m
.29424 .24464 L
s
.29424 .24464 m
.28836 .25482 L
s
.28836 .25482 m
.30012 .25482 L
s
.30012 .25482 m
.306 .265 L
s
.306 .265 m
.30012 .27518 L
s
.30012 .27518 m
.28836 .27518 L
s
.28836 .27518 m
.29424 .28537 L
s
.29424 .28537 m
.28836 .29555 L
s
.28836 .29555 m
.2766 .29555 L
s
.2766 .29555 m
.27072 .28537 L
s
.27072 .28537 m
.26484 .29555 L
s
.26484 .29555 m
.25309 .29555 L
s
.25309 .29555 m
.24721 .28537 L
s
.24721 .28537 m
.25309 .27518 L
s
.25309 .27518 m
.24133 .27518 L
s
.24133 .27518 m
.23545 .265 L
s
.23545 .265 m
.22957 .27518 L
s
.22957 .27518 m
.21781 .27518 L
s
.21781 .27518 m
.22369 .28537 L
s
.22369 .28537 m
.21781 .29555 L
s
.21781 .29555 m
.20606 .29555 L
s
.20606 .29555 m
.20018 .28537 L
s
.20018 .28537 m
.1943 .29555 L
s
.1943 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.17666 .28537 L
s
.17666 .28537 m
.18254 .27518 L
s
.18254 .27518 m
.17078 .27518 L
s
.17078 .27518 m
.1649 .265 L
s
.1649 .265 m
.17078 .25482 L
s
.17078 .25482 m
.18254 .25482 L
s
.18254 .25482 m
.17666 .24464 L
s
.17666 .24464 m
.18254 .23445 L
s
.18254 .23445 m
.17078 .23445 L
s
.17078 .23445 m
.1649 .22427 L
s
.1649 .22427 m
.15902 .23445 L
s
.15902 .23445 m
.14727 .23445 L
s
.14727 .23445 m
.14139 .22427 L
s
.14139 .22427 m
.14727 .21409 L
s
.14727 .21409 m
.13551 .21409 L
s
.13551 .21409 m
.12963 .20391 L
s
.12963 .20391 m
.13551 .19372 L
s
.13551 .19372 m
.14727 .19372 L
s
.14727 .19372 m
.14139 .18354 L
s
.14139 .18354 m
.14727 .17336 L
s
.14727 .17336 m
.15902 .17336 L
s
.15902 .17336 m
.1649 .18354 L
s
.1649 .18354 m
.17078 .17336 L
s
.17078 .17336 m
.18254 .17336 L
s
.18254 .17336 m
.17666 .16318 L
s
.17666 .16318 m
.18254 .15299 L
s
.18254 .15299 m
.17078 .15299 L
s
.17078 .15299 m
.1649 .14281 L
s
.1649 .14281 m
.17078 .13263 L
s
.17078 .13263 m
.18254 .13263 L
s
.18254 .13263 m
.17666 .12245 L
s
.17666 .12245 m
.18254 .11226 L
s
.18254 .11226 m
.17078 .11226 L
s
.17078 .11226 m
.1649 .10208 L
s
.1649 .10208 m
.15902 .11226 L
s
.15902 .11226 m
.14727 .11226 L
s
.14727 .11226 m
.14139 .10208 L
s
.14139 .10208 m
.14727 .0919 L
s
.14727 .0919 m
.13551 .0919 L
s
.13551 .0919 m
.12963 .08171 L
s
.12963 .08171 m
.12375 .0919 L
s
.12375 .0919 m
.11199 .0919 L
s
.11199 .0919 m
.11787 .10208 L
s
.11787 .10208 m
.11199 .11226 L
s
.11199 .11226 m
.10024 .11226 L
s
.10024 .11226 m
.09436 .10208 L
s
.09436 .10208 m
.08848 .11226 L
s
.08848 .11226 m
.07672 .11226 L
s
.07672 .11226 m
.07084 .10208 L
s
.07084 .10208 m
.07672 .0919 L
s
.07672 .0919 m
.06496 .0919 L
s
.06496 .0919 m
.05908 .08171 L
s
.05908 .08171 m
.06496 .07153 L
s
.06496 .07153 m
.07672 .07153 L
s
.07672 .07153 m
.07084 .06135 L
s
.07084 .06135 m
.07672 .05117 L
s
.07672 .05117 m
.06496 .05117 L
s
.06496 .05117 m
.05908 .04098 L
s
.05908 .04098 m
.0532 .05117 L
s
.0532 .05117 m
.04145 .05117 L
s
.04145 .05117 m
.03557 .04098 L
s
.03557 .04098 m
.04145 .0308 L
s
.04145 .0308 m
.02969 .0308 L
s
.02969 .0308 m
.02381 .02062 L
s
.02381 .02062 m
.03557 .02062 L
s
.03557 .02062 m
.04145 .0308 L
s
.04145 .0308 m
.04733 .02062 L
s
.04733 .02062 m
.05908 .02062 L
s
.05908 .02062 m
.06496 .0308 L
s
.06496 .0308 m
.05908 .04098 L
s
.05908 .04098 m
.07084 .04098 L
s
.07084 .04098 m
.07672 .05117 L
s
.07672 .05117 m
.0826 .04098 L
s
.0826 .04098 m
.09436 .04098 L
s
.09436 .04098 m
.08848 .0308 L
s
.08848 .0308 m
.09436 .02062 L
s
.09436 .02062 m
.10611 .02062 L
s
.10611 .02062 m
.11199 .0308 L
s
.11199 .0308 m
.11787 .02062 L
s
.11787 .02062 m
.12963 .02062 L
s
.12963 .02062 m
.13551 .0308 L
s
.13551 .0308 m
.12963 .04098 L
s
.12963 .04098 m
.14139 .04098 L
s
.14139 .04098 m
.14727 .05117 L
s
.14727 .05117 m
.14139 .06135 L
s
.14139 .06135 m
.12963 .06135 L
s
.12963 .06135 m
.13551 .07153 L
s
.13551 .07153 m
.12963 .08171 L
s
.12963 .08171 m
.14139 .08171 L
s
.14139 .08171 m
.14727 .0919 L
s
.14727 .0919 m
.15315 .08171 L
s
.15315 .08171 m
.1649 .08171 L
s
.1649 .08171 m
.17078 .0919 L
s
.17078 .0919 m
.1649 .10208 L
s
.1649 .10208 m
.17666 .10208 L
s
.17666 .10208 m
.18254 .11226 L
s
.18254 .11226 m
.18842 .10208 L
s
.18842 .10208 m
.20018 .10208 L
s
.20018 .10208 m
.1943 .0919 L
s
.1943 .0919 m
.20018 .08171 L
s
.20018 .08171 m
.21193 .08171 L
s
.21193 .08171 m
.21781 .0919 L
s
.21781 .0919 m
.22369 .08171 L
s
.22369 .08171 m
.23545 .08171 L
s
.23545 .08171 m
.22957 .07153 L
s
.22957 .07153 m
.23545 .06135 L
s
.23545 .06135 m
.22369 .06135 L
s
.22369 .06135 m
.21781 .05117 L
s
.21781 .05117 m
.22369 .04098 L
s
.22369 .04098 m
.23545 .04098 L
s
.23545 .04098 m
.22957 .0308 L
s
.22957 .0308 m
.23545 .02062 L
s
.23545 .02062 m
.24721 .02062 L
s
.24721 .02062 m
.25309 .0308 L
s
.25309 .0308 m
.25897 .02062 L
s
.25897 .02062 m
.27072 .02062 L
s
.27072 .02062 m
.2766 .0308 L
s
.2766 .0308 m
.27072 .04098 L
s
.27072 .04098 m
.28248 .04098 L
s
.28248 .04098 m
.28836 .05117 L
s
.28836 .05117 m
.29424 .04098 L
s
.29424 .04098 m
.306 .04098 L
s
.306 .04098 m
.30012 .0308 L
s
.30012 .0308 m
.306 .02062 L
s
.306 .02062 m
.31775 .02062 L
s
.31775 .02062 m
.32363 .0308 L
s
.32363 .0308 m
.32951 .02062 L
s
.32951 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.34715 .0308 L
s
.34715 .0308 m
.34127 .04098 L
s
.34127 .04098 m
.35303 .04098 L
s
.35303 .04098 m
.35891 .05117 L
s
.35891 .05117 m
.35303 .06135 L
s
.35303 .06135 m
.34127 .06135 L
s
.34127 .06135 m
.34715 .07153 L
s
.34715 .07153 m
.34127 .08171 L
s
.34127 .08171 m
.35303 .08171 L
s
.35303 .08171 m
.35891 .0919 L
s
.35891 .0919 m
.36479 .08171 L
s
.36479 .08171 m
.37654 .08171 L
s
.37654 .08171 m
.38242 .0919 L
s
.38242 .0919 m
.37654 .10208 L
s
.37654 .10208 m
.3883 .10208 L
s
.3883 .10208 m
.39418 .11226 L
s
.39418 .11226 m
.3883 .12245 L
s
.3883 .12245 m
.37654 .12245 L
s
.37654 .12245 m
.38242 .13263 L
s
.38242 .13263 m
.37654 .14281 L
s
.37654 .14281 m
.36479 .14281 L
s
.36479 .14281 m
.35891 .13263 L
s
.35891 .13263 m
.35303 .14281 L
s
.35303 .14281 m
.34127 .14281 L
s
.34127 .14281 m
.34715 .15299 L
s
.34715 .15299 m
.34127 .16318 L
s
.34127 .16318 m
.35303 .16318 L
s
.35303 .16318 m
.35891 .17336 L
s
.35891 .17336 m
.35303 .18354 L
s
.35303 .18354 m
.34127 .18354 L
s
.34127 .18354 m
.34715 .19372 L
s
.34715 .19372 m
.34127 .20391 L
s
.34127 .20391 m
.35303 .20391 L
s
.35303 .20391 m
.35891 .21409 L
s
.35891 .21409 m
.36479 .20391 L
s
.36479 .20391 m
.37654 .20391 L
s
.37654 .20391 m
.38242 .21409 L
s
.38242 .21409 m
.37654 .22427 L
s
.37654 .22427 m
.3883 .22427 L
s
.3883 .22427 m
.39418 .23445 L
s
.39418 .23445 m
.40006 .22427 L
s
.40006 .22427 m
.41182 .22427 L
s
.41182 .22427 m
.40594 .21409 L
s
.40594 .21409 m
.41182 .20391 L
s
.41182 .20391 m
.42357 .20391 L
s
.42357 .20391 m
.42945 .21409 L
s
.42945 .21409 m
.43533 .20391 L
s
.43533 .20391 m
.44709 .20391 L
s
.44709 .20391 m
.45297 .21409 L
s
.45297 .21409 m
.44709 .22427 L
s
.44709 .22427 m
.45885 .22427 L
s
.45885 .22427 m
.46473 .23445 L
s
.46473 .23445 m
.45885 .24464 L
s
.45885 .24464 m
.44709 .24464 L
s
.44709 .24464 m
.45297 .25482 L
s
.45297 .25482 m
.44709 .265 L
s
.44709 .265 m
.45885 .265 L
s
.45885 .265 m
.46473 .27518 L
s
.46473 .27518 m
.47061 .265 L
s
.47061 .265 m
.48236 .265 L
s
.48236 .265 m
.48824 .27518 L
s
.48824 .27518 m
.48236 .28537 L
s
.48236 .28537 m
.49412 .28537 L
s
.49412 .28537 m
.5 .29555 L
s
.5 .29555 m
.50588 .28537 L
s
.50588 .28537 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .27518 L
s
.51176 .27518 m
.51764 .265 L
s
.51764 .265 m
.52939 .265 L
s
.52939 .265 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .265 L
s
.54115 .265 m
.55291 .265 L
s
.55291 .265 m
.54703 .25482 L
s
.54703 .25482 m
.55291 .24464 L
s
.55291 .24464 m
.54115 .24464 L
s
.54115 .24464 m
.53527 .23445 L
s
.53527 .23445 m
.54115 .22427 L
s
.54115 .22427 m
.55291 .22427 L
s
.55291 .22427 m
.54703 .21409 L
s
.54703 .21409 m
.55291 .20391 L
s
.55291 .20391 m
.56467 .20391 L
s
.56467 .20391 m
.57055 .21409 L
s
.57055 .21409 m
.57643 .20391 L
s
.57643 .20391 m
.58818 .20391 L
s
.58818 .20391 m
.59406 .21409 L
s
.59406 .21409 m
.58818 .22427 L
s
.58818 .22427 m
.59994 .22427 L
s
.59994 .22427 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .22427 L
s
.6117 .22427 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .21409 L
s
.61758 .21409 m
.62346 .20391 L
s
.62346 .20391 m
.63521 .20391 L
s
.63521 .20391 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .20391 L
s
.64697 .20391 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .19372 L
s
.65285 .19372 m
.65873 .18354 L
s
.65873 .18354 m
.64697 .18354 L
s
.64697 .18354 m
.64109 .17336 L
s
.64109 .17336 m
.64697 .16318 L
s
.64697 .16318 m
.65873 .16318 L
s
.65873 .16318 m
.65285 .15299 L
s
.65285 .15299 m
.65873 .14281 L
s
.65873 .14281 m
.64697 .14281 L
s
.64697 .14281 m
.64109 .13263 L
s
.64109 .13263 m
.63521 .14281 L
s
.63521 .14281 m
.62346 .14281 L
s
.62346 .14281 m
.61758 .13263 L
s
.61758 .13263 m
.62346 .12245 L
s
.62346 .12245 m
.6117 .12245 L
s
.6117 .12245 m
.60582 .11226 L
s
.60582 .11226 m
.6117 .10208 L
s
.6117 .10208 m
.62346 .10208 L
s
.62346 .10208 m
.61758 .0919 L
s
.61758 .0919 m
.62346 .08171 L
s
.62346 .08171 m
.63521 .08171 L
s
.63521 .08171 m
.64109 .0919 L
s
.64109 .0919 m
.64697 .08171 L
s
.64697 .08171 m
.65873 .08171 L
s
.65873 .08171 m
.65285 .07153 L
s
.65285 .07153 m
.65873 .06135 L
s
.65873 .06135 m
.64697 .06135 L
s
.64697 .06135 m
.64109 .05117 L
s
.64109 .05117 m
.64697 .04098 L
s
.64697 .04098 m
.65873 .04098 L
s
.65873 .04098 m
.65285 .0308 L
s
.65285 .0308 m
.65873 .02062 L
s
.65873 .02062 m
.67049 .02062 L
s
.67049 .02062 m
.67637 .0308 L
s
.67637 .0308 m
.68225 .02062 L
s
.68225 .02062 m
.694 .02062 L
s
.694 .02062 m
.69988 .0308 L
s
.69988 .0308 m
.694 .04098 L
s
.694 .04098 m
.70576 .04098 L
s
.70576 .04098 m
.71164 .05117 L
s
.71164 .05117 m
.71752 .04098 L
s
.71752 .04098 m
.72928 .04098 L
s
.72928 .04098 m
.7234 .0308 L
s
.7234 .0308 m
.72928 .02062 L
s
.72928 .02062 m
.74103 .02062 L
s
.74103 .02062 m
.74691 .0308 L
s
.74691 .0308 m
.75279 .02062 L
s
.75279 .02062 m
.76455 .02062 L
s
.76455 .02062 m
.77043 .0308 L
s
.77043 .0308 m
.76455 .04098 L
s
.76455 .04098 m
.77631 .04098 L
s
.77631 .04098 m
.78219 .05117 L
s
.78219 .05117 m
.77631 .06135 L
s
.77631 .06135 m
.76455 .06135 L
s
.76455 .06135 m
.77043 .07153 L
s
.77043 .07153 m
.76455 .08171 L
s
.76455 .08171 m
.77631 .08171 L
s
.77631 .08171 m
.78219 .0919 L
s
.78219 .0919 m
.78807 .08171 L
s
.78807 .08171 m
.79982 .08171 L
s
.79982 .08171 m
.8057 .0919 L
s
.8057 .0919 m
.79982 .10208 L
s
.79982 .10208 m
.81158 .10208 L
s
.81158 .10208 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .10208 L
s
.82334 .10208 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .0919 L
s
.82922 .0919 m
.8351 .08171 L
s
.8351 .08171 m
.84685 .08171 L
s
.84685 .08171 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .08171 L
s
.85861 .08171 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .07153 L
s
.86449 .07153 m
.87037 .06135 L
s
.87037 .06135 m
.85861 .06135 L
s
.85861 .06135 m
.85273 .05117 L
s
.85273 .05117 m
.85861 .04098 L
s
.85861 .04098 m
.87037 .04098 L
s
.87037 .04098 m
.86449 .0308 L
s
.86449 .0308 m
.87037 .02062 L
s
.87037 .02062 m
.88213 .02062 L
s
.88213 .02062 m
.88801 .0308 L
s
.88801 .0308 m
.89389 .02062 L
s
.89389 .02062 m
.90564 .02062 L
s
.90564 .02062 m
.91152 .0308 L
s
.91152 .0308 m
.90564 .04098 L
s
.90564 .04098 m
.9174 .04098 L
s
.9174 .04098 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .04098 L
s
.92916 .04098 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .0308 L
s
.93504 .0308 m
.94092 .02062 L
s
.94092 .02062 m
.95267 .02062 L
s
.95267 .02062 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .02062 L
s
.96443 .02062 m
.97619 .02062 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 249.375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001MR000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002P3oool6000000@0oooo1P00
000>0?ooo`H00000103oool6000002d0oooo1P0000040?ooo`H000003P3oool6000000D0oooo1P00
00270?ooo`H000001@3oool6000000h0oooo1P0000040?ooo`H00000;@3oool6000000@0oooo1P00
000>0?ooo`H00000103oool6000000/0oooo000:0?ooo`030000003oool0oooo0080oooo00D00000
0?ooo`3oool0oooo000000060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3o
ool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0:P3oool00`000000oooo
0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo
08D0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3o
ool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo
000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool01@000000oooo
0?ooo`3oool0000000@0oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool0
0P3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool0103oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo02T0oooo00<000000?oo
o`3oool0103oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3oool0P`3o
ool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000103oool0
0`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`020?ooo`030000003oool00000
00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`030000003oool00000
00@0oooo00<000000?ooo`3oool02P3oool000`0oooo00<000000?ooo`3oool00P3oool2000000P0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`800000203oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo00H0oooo0P0000080?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01P3oool00`000000oooo000000080?ooo`030000003oool0oooo0840
oooo00<000000?ooo`3oool01P3oool00`000000oooo000000080?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01P3oool2000000P0oooo00<000000?ooo`3oool09`3oool00`000000
oooo0?ooo`060?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0
oooo0P0000040?ooo`030000003oool0oooo00/0oooo000<0?ooo`H000002P3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`210?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00P0oooo1P00000=0?ooo`00403oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0W0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0@0?ooo`00403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo08<0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0?0?ooo`003`3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo08@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0?0?oo
o`003`3oool00`000000oooo0?ooo`070?ooo`H00000103oool6000000h0oooo1`00000R0?ooo`H0
00003P3oool6000000@0oooo1P00000?0?ooo`H00000O@3oool6000000l0oooo1P0000040?ooo`H0
00003P3oool600000280oooo1`00000>0?ooo`H00000103oool6000000T0oooo00<000000?ooo`3o
ool03P3oool000l0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000040?ooo`040000
003oool0oooo000001X0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0H0?ooo`04
0000003oool0oooo000001X0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0H0?oo
o`050000003oool0oooo0?ooo`0000006P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo01P0oooo00D000000?ooo`3oool0oooo000000040?ooo`800000203oool00`000000oooo0?oo
o`0?0?ooo`00403oool00`000000oooo0?ooo`050?ooo`030000003oool0000000@0oooo00@00000
0?ooo`3oool000006`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01T0oooo00@0
00000?ooo`3oool000006`3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`0000006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`000000103oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo00l0
oooo000@0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000040?ooo`80
0000703oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01X0oooo0P00000L0?ooo`03
0000003oool0oooo07P0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000L0?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool06P3oool00`000000oooo000000040?ooo`04
0000003oool0oooo000000H0oooo00<000000?ooo`3oool0403oool00140oooo1`0000040?ooo`H0
00007@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01/0oooo1P0000040?ooo`L0
00004P3oool00200oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo03T0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0i0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0O0?ooo`00803o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0>03o
ool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01l0oooo000O0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo07T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool07P3oool001l0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0N`3oool00`00
0000oooo0?ooo`0f0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0N0?ooo`007P3oool00`000000oooo0?ooo`0F0?ooo`L000008P3oool6000002d0
oooo1P00001m0?ooo`H00000;@3oool600000280oooo1`00000H0?ooo`030000003oool0oooo01d0
oooo000O0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`0/0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool07P3oool001l0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02d0oooo00<0
00000?ooo`3oool0P`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02P0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0O0?ooo`00803oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`22
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01l0oooo000L0?ooo`H00000603oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`3oool05P3oool6000001d0oooo000K0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo0840oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0W0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool06P3oool001/0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0
P`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`0J0?ooo`006P3oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`230?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo01T0oooo000J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`0[0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06@3oool0
01T0oooo00<000000?ooo`3oool06`3oool7000000@0oooo1P00000>0?ooo`H00000103oool60000
02d0oooo1P0000040?ooo`H00000J@3oool6000000@0oooo1P00000]0?ooo`H00000103oool60000
00h0oooo1P0000040?ooo`L000007@3oool00`000000oooo0?ooo`0H0?ooo`006P3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0000000@0oooo00D000000?ooo`3oool0oooo000000050?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0
000003P0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo06H0oooo00<0
00000?ooo`3oool0103oool010000000oooo0?ooo`00000i0?ooo`040000003oool0oooo000000H0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo0000
00D0oooo0P00000L0?ooo`030000003oool0oooo01T0oooo000J0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`0000001@3oool00`000000oooo000000070?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000j0?ooo`030000003oool0000000L0
oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo0000
03T0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00@0oooo00@000000?ooo`3oool00000103oool010000000oooo0?ooo`00000J0?ooo`030000
003oool0oooo01X0oooo000K0?ooo`030000003oool0oooo01L0oooo00D000000?ooo`3oool0oooo
000000040?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0000003X0oooo00<000000?ooo`0000001`3oool00`000000oooo
0?ooo`1T0?ooo`030000003oool0oooo00H0oooo0P00000k0?ooo`800000203oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000050?ooo`040000003oool0oooo000001X0
oooo00<000000?ooo`3oool06P3oool001`0oooo1P00000>0?ooo`H00000103oool6000000T0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03X0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool6000000@0oooo1P00000>0?ooo`H000007@3oool00200oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo06@0oooo00<0
00000?ooo`3oool0C@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01l0oooo000O
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1U
0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0O
0?ooo`007`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04X0oooo00<000000?oo
o`3oool0IP3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool07P3oool001h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool6000000@0oooo1P00001;0?ooo`H0
0000G`3oool6000004/0oooo1P0000040?ooo`H000002@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0M0?ooo`00
7`3oool00`000000oooo0?ooo`060?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00H0oooo0P0000040?ooo`050000003oool0oooo0?ooo`000000EP3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo05@0oooo00D000000?ooo`3oool0oooo000000040?oo
o`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000080?oo
o`030000003oool0oooo01h0oooo000O0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`00
00001`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`00
0000103oool010000000oooo0?ooo`00001G0?ooo`030000003oool0oooo05/0oooo00<000000?oo
o`3oool0E@3oool010000000oooo0?ooo`0000040?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool07P3oool00200oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`000006
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00
00040?ooo`030000003oool0000005P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?oo
o`1F0?ooo`030000003oool0000000@0oooo00@000000?ooo`3oool000001P3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0O0?ooo`00803oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`00
00001@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3o
ool0oooo000000040?ooo`800000F@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo
05L0oooo0P0000040?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003o
ool0oooo01l0oooo000Q0?ooo`H00000103oool6000000h0oooo1P0000040?ooo`H00000FP3oool0
0`000000oooo0?ooo`1I0?ooo`030000003oool0oooo05P0oooo1P0000040?ooo`H000003P3oool6
000000@0oooo1P00000R0?ooo`00C@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo
05X0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1=0?ooo`00C@3oool00`000000
oooo0?ooo`1H0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0F03oool00`000000
oooo0?ooo`1<0?ooo`00C03oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo05`0oooo
00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo0?oo
o`1C0?ooo`H00000G`3oool6000005D0oooo00<000000?ooo`3oool0B`3oool004`0oooo00<00000
0?ooo`3oool0D`3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0B`3oool004d0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1U0?oo
o`030000003oool0oooo05<0oooo00<000000?ooo`3oool0C03oool004d0oooo00<000000?ooo`3o
ool0D`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3o
ool0C03oool004h0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo05<0oooo00<000000?ooo`3oool0C@3oool004T0oooo1P00001;0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo04T0oooo1P00001:0?ooo`00B03oool00`000000oooo0?ooo`1>0?ooo`030000003oool0
000000L0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`060?ooo`800000D03oool0
0`000000oooo0?ooo`180?ooo`00B03oool00`000000oooo0?ooo`1>0?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo
00000500oooo00<000000?ooo`3oool0A`3oool004L0oooo00<000000?ooo`3oool0CP3oool01@00
0000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo00000500oooo00<000000?ooo`3oool0A`3oool004L0oooo00<00000
0?ooo`3oool0B@3oool6000000@0oooo1P00001Y0?ooo`H00000103oool6000004/0oooo00<00000
0?ooo`3oool0AP3oool004L0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`240?oo
o`030000003oool0oooo04T0oooo00<000000?ooo`3oool0A`3oool004P0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0A`3oool004P0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`220?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool0B03oool004T0oooo1P0000160?ooo`030000003o
ool0oooo0840oooo00<000000?ooo`3oool0A03oool6000004X0oooo001>0?ooo`030000003oool0
oooo04@0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`140?ooo`030000003oool0
oooo04d0oooo001=0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0P`3oool00`00
0000oooo0?ooo`140?ooo`030000003oool0oooo04`0oooo001=0?ooo`030000003oool0oooo04@0
oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo04`0
oooo001<0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo04/0oooo001<0?ooo`030000003oool0oooo04@0oooo1P00
001m0?ooo`H00000AP3oool00`000000oooo0?ooo`1;0?ooo`00C03oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`1<
0?ooo`00C@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo07T0oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`1<0?ooo`00C@3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo07P0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`1=0?ooo`00
?`3oool6000000@0oooo1P00001;0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0
B@3oool6000000@0oooo1P0000100?ooo`00?`3oool00`000000oooo0?ooo`030?ooo`050000003o
ool0oooo0?ooo`000000D03oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo04h0oooo
00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo03h0oooo000n0?ooo`030000
003oool0oooo00@0oooo00@000000?ooo`3oool00000D03oool00`000000oooo0?ooo`1i0?ooo`03
0000003oool0oooo04h0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0m0?oo
o`00?P3oool00`000000oooo0?ooo`050?ooo`030000003oool000000500oooo00<000000?ooo`3o
ool0N@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0000000L0oooo00<000000?ooo`3o
ool0?@3oool003d0oooo00<000000?ooo`3oool01P3oool200000500oooo00<000000?ooo`3oool0
N`3oool00`000000oooo0?ooo`1>0?ooo`800000203oool00`000000oooo0?ooo`0l0?ooo`00?@3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04T0oooo1P00001m0?ooo`H00000B`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03`0oooo000m0?ooo`030000003oool0
oooo05<0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`1C0?ooo`030000003oool0
oooo03d0oooo000n0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0P`3oool00`00
0000oooo0?ooo`1C0?ooo`030000003oool0oooo03d0oooo000n0?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03h0
oooo000j0?ooo`H00000E@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo05<0oooo
1P00000k0?ooo`00>@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0880oooo00<0
00000?ooo`3oool0F@3oool00`000000oooo0?ooo`0h0?ooo`00>03oool00`000000oooo0?ooo`1I
0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0h
0?ooo`00>03oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo08@0oooo00<000000?oo
o`3oool0F@3oool00`000000oooo0?ooo`0g0?ooo`00=`3oool00`000000oooo0?ooo`1I0?ooo`H0
0000103oool6000000h0oooo1P0000040?ooo`H00000;@3oool6000000@0oooo1P00000>0?ooo`H0
0000103oool6000005/0oooo00<000000?ooo`3oool0=P3oool003P0oooo00<000000?ooo`3oool0
E`3oool2000000@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001P3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000
00040?ooo`800000F@3oool00`000000oooo0?ooo`0g0?ooo`00>03oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0000000@0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000oooo
0?ooo`0Y0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000103oool00`00
0000oooo0000001I0?ooo`030000003oool0oooo03L0oooo000i0?ooo`030000003oool0oooo05D0
oooo00@000000?ooo`3oool00000103oool00`000000oooo000000070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000070?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool01@3oool00`000000oooo000000080?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool01@3oool00`000000oooo000000040?ooo`040000003oool0oooo
000005L0oooo00<000000?ooo`3oool0>03oool003T0oooo00<000000?ooo`3oool0E@3oool01@00
0000oooo0?ooo`3oool0000000@0oooo0P0000080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool01P3oool2000000P0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`06
0?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P000004
0?ooo`050000003oool0oooo0?ooo`000000E`3oool00`000000oooo0?ooo`0h0?ooo`00>P3oool6
000004/0oooo1P0000040?ooo`H000002@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00L0oooo1P0000040?ooo`H00000B`3oool6000003/0oooo000n0?ooo`030000003oool0oooo
04/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0n0?ooo`00?P3oool00`000000
oooo0?ooo`1;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:03oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0?@3oool003d0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02T0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo03d0oooo
000m0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool6000000@0oooo1P00
000>0?ooo`H000008`3oool6000000h0oooo1P0000040?ooo`H000002@3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo03`0oooo000m0?ooo`030000003oool0oooo00H0oooo0P00000j
0?ooo`030000003oool0000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`06
0?ooo`800000103oool01@000000oooo0?ooo`3oool0000001X0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0H0?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`000000
1`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo0P00000j0?ooo`030000
003oool0000000L0oooo00<000000?ooo`3oool0?@3oool003h0oooo00<000000?ooo`3oool0103o
ool010000000oooo0?ooo`00000i0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`00
00006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01T0oooo00@000000?ooo`3o
ool00000103oool00`000000oooo000000070?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool01@3oool010000000oooo0?ooo`00000i0?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool0?@3oool003h0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00000h
0?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00@0oooo00@000000?ooo`3oool00000103oool00`000000oooo0000000L0?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool06P3oool2000000@0oooo00D000000?ooo`3oool0
oooo000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool010000000
oooo0?ooo`00000h0?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0n
0?ooo`00?`3oool6000000@0oooo1P00000]0?ooo`H00000103oool6000000h0oooo1P0000040?oo
o`H000007P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01`0oooo1P0000040?oo
o`H000003P3oool6000000@0oooo1P00000]0?ooo`H00000103oool600000400oooo001>0?ooo`03
0000003oool0oooo02/0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo01d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0Z0?ooo`03
0000003oool0oooo02/0oooo00<000000?ooo`3oool0C@3oool004d0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`1<0?ooo`00C@3oool00`000000oooo0?ooo`0]0?ooo`030000003o
ool0oooo02P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0]0?ooo`030000003o
ool0oooo04`0oooo001<0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0B`3oool0
04`0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
01H0oooo1P00000S0?ooo`H00000603oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
02l0oooo00<000000?ooo`3oool0B`3oool004`0oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`0F0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`1<0?ooo`00C@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02P0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo04`0oooo
001=0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0C@3oool004T0oooo1P00000]
0?ooo`H000008P3oool6000001T0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0G
0?ooo`H000008P3oool6000002d0oooo1P00001:0?ooo`00B@3oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo0200oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo04P0oooo00180?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3o
ool07P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3o
ool0A`3oool004P0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`170?ooo`00A`3o
ool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo04H0oooo00170?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0
oooo01/0oooo1P0000040?ooo`H000003`3oool6000000@0oooo1P00000M0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo04H0oooo00170?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000006`3oool00`00
0000oooo0?ooo`0M0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`000000103oool01000
0000oooo0?ooo`0000060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool0
1@000000oooo0?ooo`3oool0000000@0oooo0P00000L0?ooo`030000003oool0oooo01d0oooo00<0
00000?ooo`3oool06P3oool2000001`0oooo00<000000?ooo`3oool0A`3oool004P0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo000000040?ooo`040000003oool0oooo000000H0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000@0oooo00@0
00000?ooo`3oool000006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01P0oooo
00@000000?ooo`3oool000006`3oool00`000000oooo0?ooo`170?ooo`00B03oool00`000000oooo
0?ooo`0H0?ooo`050000003oool0oooo0?ooo`0000006@3oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo01P0oooo00D000000?ooo`3oool0oooo000000040?ooo`800000203oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000103oool01000
0000oooo0?ooo`00000J0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0603oool0
10000000oooo0?ooo`00000J0?ooo`030000003oool0oooo04P0oooo00190?ooo`H000003P3oool7
000000@0oooo1P00000>0?ooo`H000008P3oool6000000l0oooo1P0000040?ooo`H000002@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo1P0000040?ooo`H000003`3oool6
00000280oooo1P00000>0?ooo`H00000103oool7000000h0oooo1P00001:0?ooo`00C@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0Y0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`1=0?ooo`00C@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`0W0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`1<0?ooo`00C03oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`H000001@3oool6000000T0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0B`3oool004`0oooo00<000000?ooo`3oool01P3oool2000000P0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`800000203oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`000000203oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00H0oooo0P0000040?ooo`050000003oool0oooo0?ooo`000000103o
ool2000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`030000003o
ool0000000P0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`060?ooo`800000203o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo0P0000080?ooo`030000003o
ool0oooo04/0oooo001=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000103oool0
1@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`0X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`1<0?ooo`00C@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`00
00001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3o
ool00000103oool00`000000oooo000000040?ooo`040000003oool0oooo000000H0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001P3o
ool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000
1P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool0
00001P3oool00`000000oooo0?ooo`1<0?ooo`00CP3oool00`000000oooo0?ooo`030?ooo`050000
003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0
oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo02X0oooo00<000000?oo
o`3oool00`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0000
00040?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo02X0
oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3oool0
0`000000oooo0?ooo`1=0?ooo`00CP3oool6000000@0oooo1P00000?0?ooo`H00000103oool60000
02`0oooo1P0000050?ooo`H000003P3oool6000000@0oooo2`0000040?ooo`H000003P3oool60000
00D0oooo1P00000/0?ooo`H00000103oool6000000l0oooo1P0000040?ooo`H00000C`3oool00=H0
oooo00<000000?ooo`000000e`3oool00=D0oooo00@000000?ooo`3oool00000e`3oool00=D0oooo
00D000000?ooo`3oool0oooo0000003F0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0=<0oooo003E0?ooo`050000003oool0oooo0?ooo`000000eP3oool00=D0oooo00D0
00000?ooo`3oool0oooo0000003F0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool0
0`000000oooo0000003G0?ooo`00dP3oool;00000=<0oooo003A0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0d03oool00=00oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`3@0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo003?
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0cP3oool00=00oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`3?0?ooo`00d03oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo0=00oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
d03oool00=80oooo2`00003C0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`00
0000oooo0000003G0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`05
0000003oool0oooo0?ooo`000000eP3oool00=@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`3C0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`040000
003oool0oooo00000=L0oooo003F0?ooo`030000003oool000000=L0oooo00380?ooo`H00000103o
ool;000000@0oooo1P0000390?ooo`00a`3oool00`000000oooo0?ooo`030?ooo`050000003oool0
oooo0?ooo`0000002`3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0<H0oooo
00360?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000303oool010000000oooo0?oo
o`0000060?ooo`030000003oool0oooo0<H0oooo00360?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`0000003@3oool2000000P0oooo00<000000?ooo`3oool0a@3oool00<D0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0a03oool00<H0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`350?oo
o`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00370?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool0aP3oool00<L0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`360?ooo`00``3oool6000001d0oooo1P0000340?ooo`00`P3oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo0<40oooo00310?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool0`@3oool00<40oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`300?ooo`00`03oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0;l0oooo0031
0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0`03oool00<40oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`300?ooo`00`P3oool00`000000oooo0?ooo`0W0?ooo`03
0000003oool0oooo0<40oooo00320?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
`@3oool00<<0oooo1P00000M0?ooo`H00000a03oool00<L0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`360?ooo`00aP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
0<H0oooo00360?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0a@3oool00<D0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0a03oool00<H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000
000=0?ooo`800000203oool00`000000oooo0?ooo`350?ooo`00aP3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0000000`0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?oo
o`360?ooo`00a`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000002`3o
ool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0<H0oooo00380?ooo`H00000103o
ool;000000@0oooo1P0000390?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`00
0000oooo0000003G0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`05
0000003oool0oooo0?ooo`000000eP3oool00=@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`3C0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`040000
003oool0oooo00000=L0oooo003F0?ooo`030000003oool000000=L0oooo003B0?ooo`/00000d`3o
ool00=40oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00d@3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo0=00oooo003@0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0c`3oool00=00oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`3?0?ooo`00c`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0<h0oooo003@
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0c`3oool00=00oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`3@0?ooo`00d@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0=00oooo003B0?ooo`/00000d`3oool00=H0oooo00<000000?ooo`000000e`3o
ool00=D0oooo00@000000?ooo`3oool00000e`3oool00=D0oooo00D000000?ooo`3oool0oooo0000
003F0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003E0?oo
o`050000003oool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F
0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00
Z@3oool6000000D0oooo1P00000>0?ooo`H00000103oool;000000@0oooo1P00000>0?ooo`H00000
1@3oool600000:X0oooo002X0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo
000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool0000000/0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0
oooo0:P0oooo002X0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000303oool0
10000000oooo0?ooo`0000060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103o
ool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo0:L0oooo002W0?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`0000001`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`0000003@3oool2000000P0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`060?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0Y`3o
ool00:L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0YP3oool00:L0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo0:L0oooo002X0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`2W0?ooo`00Z03oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool07@3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0Z03oool00:@0oooo1P00
000?0?ooo`H00000103oool6000001d0oooo1P0000040?ooo`H000003`3oool600000:D0oooo002T
0?ooo`030000003oool0oooo01L0oooo00D000000?ooo`3oool0oooo0000000Y0?ooo`050000003o
ool0oooo0?ooo`0000006@3oool00`000000oooo0?ooo`2S0?ooo`00X`3oool00`000000oooo0?oo
o`0I0?ooo`040000003oool0oooo000002T0oooo00@000000?ooo`3oool000006`3oool00`000000
oooo0?ooo`2R0?ooo`00X`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0000002/0oooo
00<000000?ooo`0000006`3oool00`000000oooo0?ooo`2R0?ooo`00XP3oool00`000000oooo0?oo
o`0K0?ooo`800000:`3oool2000001d0oooo00<000000?ooo`3oool0X@3oool00:80oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool0X@3oool00:80oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`2R0?oo
o`00X`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0:80oooo002S0?ooo`030000
003oool0oooo06@0oooo00<000000?ooo`3oool0X`3oool00:@0oooo1P00001K0?ooo`H00000Y@3o
ool00:T0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`2X0?ooo`00Z03oool00`00
0000oooo0?ooo`1K0?ooo`030000003oool0oooo0:L0oooo002X0?ooo`030000003oool0oooo05/0
oooo00<000000?ooo`3oool0Y`3oool00:L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`2V0?ooo`00Y`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo0:H0oooo002W
0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0Y`3oool00:P0oooo00<000000?oo
o`3oool0F`3oool00`000000oooo0?ooo`2W0?ooo`00Z03oool00`000000oooo0?ooo`1J0?ooo`03
0000003oool0oooo0:P0oooo002J0?ooo`H00000103oool6000005/0oooo1P0000040?ooo`H00000
V`3oool009T0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00001V0?ooo`050000
003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`2I0?ooo`00V@3oool00`000000oooo
0?ooo`040?ooo`040000003oool0oooo000006L0oooo00<000000?ooo`0000001`3oool00`000000
oooo0?ooo`2H0?ooo`00V03oool00`000000oooo0?ooo`060?ooo`800000J03oool00`000000oooo
000000070?ooo`030000003oool0oooo09P0oooo002H0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0I`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo09L0oooo002H
0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0U`3oool009T0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`2H0?ooo`00V@3oool00`000000oooo0?ooo`1i0?ooo`03
0000003oool0oooo09P0oooo002J0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0
V@3oool009D0oooo1P00001i0?ooo`H00000UP3oool009@0oooo00<000000?ooo`3oool0PP3oool0
0`000000oooo0?ooo`2D0?ooo`00U03oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo
09<0oooo002C0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0T`3oool009<0oooo
00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`2B0?ooo`00T`3oool00`000000oooo0?oo
o`250?ooo`030000003oool0oooo0980oooo002D0?ooo`030000003oool0oooo08<0oooo00<00000
0?ooo`3oool0T`3oool009@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`2C0?oo
o`00U@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo09@0oooo002E0?ooo`H00000
N@3oool6000009H0oooo002I0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0V@3o
ool009T0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`2H0?ooo`00V03oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo09P0oooo002H0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo09L0
oooo002H0?ooo`030000003oool0oooo00H0oooo0P00001X0?ooo`030000003oool0000000L0oooo
00<000000?ooo`3oool0V03oool009T0oooo00<000000?ooo`3oool0103oool010000000oooo0?oo
o`00001W0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0V03oool009T0oooo00<0
00000?ooo`3oool0103oool010000000oooo0?ooo`00001V0?ooo`050000003oool0oooo0?ooo`00
00001@3oool00`000000oooo0?ooo`2I0?ooo`00VP3oool6000000@0oooo1P00001K0?ooo`H00000
103oool6000009/0oooo002Y0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0Z03o
ool00:P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`2W0?ooo`00Z03oool00`00
0000oooo0?ooo`1K0?ooo`030000003oool0oooo0:L0oooo002W0?ooo`030000003oool0oooo05d0
oooo00<000000?ooo`3oool0YP3oool00:L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`2V0?ooo`00Y`3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0:L0oooo002X
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0Y`3oool00:P0oooo00<000000?oo
o`3oool0FP3oool00`000000oooo0?ooo`2X0?ooo`00Y03oool6000005/0oooo1P00002U0?ooo`00
X`3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0:<0oooo002S0?ooo`030000003o
ool0oooo06D0oooo00<000000?ooo`3oool0XP3oool00:80oooo00<000000?ooo`3oool0IP3oool0
0`000000oooo0?ooo`2R0?ooo`00XP3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2Q0?ooo`00XP3oool00`000000
oooo0?ooo`0K0?ooo`800000:`3oool2000001d0oooo00<000000?ooo`3oool0X@3oool00:<0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0000000[0?ooo`030000003oool0000001/0oooo
00<000000?ooo`3oool0XP3oool00:<0oooo00<000000?ooo`3oool06@3oool010000000oooo0?oo
o`00000Y0?ooo`040000003oool0oooo000001/0oooo00<000000?ooo`3oool0XP3oool00:@0oooo
00<000000?ooo`3oool05`3oool01@000000oooo0?ooo`3oool0000002T0oooo00D000000?ooo`3o
ool0oooo0000000I0?ooo`030000003oool0oooo0:<0oooo002T0?ooo`H000003`3oool6000000@0
oooo1P00000M0?ooo`H00000103oool6000000l0oooo1P00002U0?ooo`00Z03oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool07@3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0Z03oool00:P0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01h0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0:L0oooo002W
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0O
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`2W
0?ooo`00Y`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`2V0?ooo`00Y`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`80
00003@3oool2000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`03
0000003oool0000000P0oooo00<000000?ooo`3oool0YP3oool00:P0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo000000070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0000000=0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
Y`3oool00:P0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000H0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000/0
oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00@0oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo0:L0oooo002Y0?oo
o`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000/0oooo00D0
00000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
0`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0Z03oool00:T0oooo
1P0000050?ooo`H000003P3oool6000000@0oooo2`0000040?ooo`H000003P3oool6000000D0oooo
1P00002Z0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00e@3oool010000000oooo0?ooo`00
003G0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003D0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F
0?ooo`00e@3oool010000000oooo0?ooo`00003G0?ooo`00eP3oool00`000000oooo0000003G0?oo
o`00dP3oool;00000=<0oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
d03oool00=40oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00d03oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo003@0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0c`3oool00<l0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`3>0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo
003@0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0d03oool00=40oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00dP3oool;00000=<0oooo003F0?ooo`03
0000003oool000000=L0oooo003F0?ooo`030000003oool000000=L0oooo003E0?ooo`050000003o
ool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?ooo`00e03o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003E0?ooo`050000003oool0
oooo0?ooo`000000eP3oool00=D0oooo00@000000?ooo`3oool00000e`3oool00=H0oooo00<00000
0?ooo`000000e`3oool00<P0oooo1P0000040?ooo`/00000103oool600000<T0oooo00370?ooo`03
0000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`040000003oool0oooo
000000H0oooo00<000000?ooo`3oool0aP3oool00<H0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0000000<0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool0aP3oool0
0<H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000=0?ooo`800000203oool00`00
0000oooo0?ooo`350?ooo`00a@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`340?ooo`00aP3oool00`000000oooo
0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00360?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool0a@3oool00<L0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`36
0?ooo`00a`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0<H0oooo00330?ooo`H0
00007@3oool600000<@0oooo00320?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
`@3oool00<40oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`310?ooo`00`@3oool0
0`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo00300?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool0_`3oool00<40oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`300?ooo`00`@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo
00320?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0`@3oool00<80oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`310?ooo`00``3oool6000001d0oooo1P0000340?oo
o`00a`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0<H0oooo00360?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0aP3oool00<H0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`350?ooo`00a@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`340?ooo`00aP3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0000000d0oooo0P0000080?ooo`030000003oool0oooo
0<D0oooo00360?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000303oool010000000
oooo0?ooo`0000060?ooo`030000003oool0oooo0<H0oooo00370?ooo`030000003oool0oooo00<0
oooo00D000000?ooo`3oool0oooo0000000;0?ooo`040000003oool0oooo000000H0oooo00<00000
0?ooo`3oool0aP3oool00<P0oooo1P0000040?ooo`/00000103oool600000<T0oooo003F0?ooo`03
0000003oool000000=L0oooo003F0?ooo`030000003oool000000=L0oooo003E0?ooo`050000003o
ool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?ooo`00e03o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003E0?ooo`050000003oool0
oooo0?ooo`000000eP3oool00=D0oooo00@000000?ooo`3oool00000e`3oool00=H0oooo00<00000
0?ooo`000000e`3oool00=80oooo2`00003C0?ooo`00d@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0=00oooo003@0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
d03oool00=00oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3?0?ooo`00c`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0<h0oooo003@0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0c`3oool00=00oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`3?0?ooo`00d@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0=00oooo
003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=80oooo2`00003C
0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00e@3oool010000000oooo0?ooo`00003G0?oo
o`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003D0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?ooo`00
e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003F0?ooo`030000003oool000000=L0oooo
003F0?ooo`030000003oool000000=L0oooo003G0?ooo`030000003oool0oooo0=H0oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {373.063, 0}} -> {-0.0252817, -0.0216552, \
0.00365625, 0.00365625}}]
}, Closed]],

Cell["Create a new directory and add it to the path.", "Exercise"],

Cell["\<\
Create a new directory and add your new directory to the \
path.\
\>", "ExerciseText"],

Cell[TextData[{
  "If you have already written a package of your own, copy it to this \
diectory; otherwise, go to the ",
  ButtonBox["MathSource",
    ButtonData:>{
      URL[ "http://www.mathsource.com"], None},
    ButtonStyle->"Hyperlink"],
  " web site and download an interesting package into your new directory.  \
Use ",
  StyleBox["FileNames", "InlineInput",
    FontWeight->"Bold"],
  " to verify its existence."
}], "ExerciseText"],

Cell["Load the package and ascertain the symbols it exports.", "ExerciseText"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
  "Files using ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " formatting"
}], "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[CellGroupData[{

Cell["Storing data", "Subsection"],

Cell[TextData[{
  "The command ",
  StyleBox["Put[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr, \"filename\"", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " puts the result of evaluating ",
  StyleBox["expr", "InlineInput"],
  " into a file with the name ",
  StyleBox["filename", "InlineInput"],
  ".  If the file does not exist, ",
  ButtonBox["Put",
    ButtonStyle->"RefGuideLink"],
  " creates it in the directory returned by ",
  StyleBox["Directory[]", "InlineInput",
    FontWeight->"Bold"],
  ".  If the file does exist, ",
  StyleBox["Put", "InlineInput",
    FontWeight->"Bold"],
  " ",
  StyleBox["deletes its contents",
    FontSlant->"Italic"],
  " and replaces them with the evaluation of ",
  StyleBox["expr", "InlineInput"],
  ".  If you want to ",
  StyleBox["append",
    FontSlant->"Italic"],
  " the evaluation of ",
  StyleBox["expr", "InlineInput"],
  " to the contents of an existing file without losing the original contents \
of that file, you must use ",
  StyleBox["PutAppend[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr, \"filename\"", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ".  "
}], "Text"],

Cell[TextData[
"Put replaces the contents of an existing file \[LongDash] to add something \
to the end of a file you must use PutAppend."], "Caution"],

Cell[TextData[{
  "A convenient shorthand for ",
  StyleBox["Put", "InlineInput",
    FontWeight->"Bold"],
  " is ",
  StyleBox["expr>>filename", "InlineInput"],
  " and for ",
  StyleBox["PutAppend", "InlineInput",
    FontWeight->"Bold"],
  " is ",
  StyleBox["expr>>>filename", "InlineInput"],
  ".  Note that quotes are not needed for these shortcuts.  To display but \
not read (input) the contents of a file, use the command ",
  StyleBox["!!", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput"],
  ". "
}], "Text"],

Cell["Exercise:  write an unformatted file.", "Exercise"],

Cell["1. Create some data by evaluating the following expression.", 
  "ExerciseText"],

Cell[BoxData[
    \(\(data = 
        Table[Sin[2  \[Pi]\ 220\ t]\  + \ Sin[2  \[Pi]\ 220.5\ t], {t, 0, 5, 
            10\^\(-3\)}];\)\)], "Input"],

Cell["\<\
2. Put this data into a file with the name \"tmp\".
3. Check that such a file exists in your working directory.
4. Change the name of the file to \"Beats\".
5. Check whether this was done.
6. Display the data in the file.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Reading data", "Subsection"],

Cell[TextData[{
  "The command ",
  StyleBox["Get[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " reads a file and provides input for defining a symbol that stores the \
contents of the file.  Evaluation of\n\t",
  StyleBox["data = Get[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  "\nwill store the contents of ",
  StyleBox["filename", "InlineInput"],
  " in the symbol ",
  StyleBox["data", "InlineInput"],
  ".  The shorthand for this expression is\n\t",
  StyleBox["data = <<", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontSlant->"Italic"]
}], "Text"],

Cell["Exercise:  read from an unformatted file.", "Exercise"],

Cell[TextData[{
  "1. ",
  StyleBox["Clear", "InlineInput",
    FontWeight->"Bold"],
  " the symbol ",
  StyleBox["data", "InlineInput",
    FontWeight->"Bold"],
  " used above and then check that it no longer stores any data.\n2. Put the \
data back into ",
  StyleBox["data", "InlineInput",
    FontWeight->"Bold"],
  " by using ",
  StyleBox["data = <<\"Beats\"", "InlineInput",
    FontWeight->"Bold"],
  ".\n3. Verify that ",
  StyleBox["data", "InlineInput",
    FontWeight->"Bold"],
  " now holds this data.  An interesting verification method is illustrated \
below.  Reproduce and interpret this display.  If your machine has a sound \
card, you will be able to hear the data also!"
}], "ExerciseText"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ListPlay[data];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.588604 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of sampled sound
1 string
8000 5001 8
{ currentfile 1 index readhexstring pop }
 false 1 sampledsound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pop
% End of sampled sound
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gQ000`40O003h00Oo`00J@0000000`00
Oomoo`3oOomUOol100000003001oo`0000H0009oo`8007Qoo`<00003Ool0000000D00005Ool00000
001oo`0300000goo0000001iOol200001Goo00000000Ool00`0000=oo`0000001@0000Aoo`000000
03eoo`40000000<007oo00000Woo0P000Woo0P000Woo20000Woo00<007ooOol0H7oo0P0000=oo`00
00001`0000Eoo`00000007oo0080009oo`8000=oo`03001oogoo0080009oo`T00003Ool00000065o
o`800003Ool0000000L0009oo`80009oo`04001oogooOol20002Ool20002Ool200000goo00000007
00000goo0000000bOol100000003001oogoo00<00004Ool00000Ool400000goo001oo`03000017oo
000007oo1P0000=oo`000000DGoo00<007ooOol01@0000Eoo`00000007oo00<00003Ool0000000<0
0005Ool007oo001oo`0300001Goo00000000Ool00P000Woo00H007oo001oo`00Ool300000goo0000
000700000goo0000001AOol200000goo0000000700002Woo00000000Ool007oo001oo`000Woo0P00
00Eoo`00000007oo00<0009oo`800003Ool0000000<00005Ool00000001oo`0300000goo00000004
00000goo001oo`0YOol100000003001oogoo0080009oo`03001oogoo009oo`03001oogoo0080009o
o`80009oo`@00005Ool00000001oo`060017Ool500001Goo00000000Ool010000Woo0P000Woo0P00
0goo00D007ooOomoo`00009oo`8000=oo`05001oogooOol00003Ool01@00Oomoogoo00000goo00T0
07ooOol007oo001oo`00Ool00`0000=oo`0000000`00Agoo1@0000Yoo`00000007oo001oo`00Ool0
009oo`04001oogooOol20002Ool01@00Oomoogoo00000goo00@007ooOomoo`80009oo`05001oogoo
Ool00003Ool20002Ool20002Ool400001Goo00000000Ool01@009Goo0@0000000`00Oomoo`020002
Ool2000017oo000000000Woo0P000Woo0P000Woo0P000Woo0P000Woo1@0000=oo`0000000`00?goo
100000Aoo`00001oo`<00003Ool0000000=oo`03001oogoo00<00005Ool007oo001oo`0300000goo
00000002Ool02000Ool007oo001oo`00Ool20002Ool01000Ool007oo0`0000=oo`0000000Woo0P00
0Woo0P000Woo0`0000=oo`00Ool01@00?goo1@0000=oo`00Ool00`000Woo0P000Woo0P000Woo0P00
00Qoo`00000007oo001oo`000Woo0P0000Qoo`00Ool007oo001oo`000Woo0P0000Qoo`00000007oo
001oo`000Woo0P000Woo00@007ooOomoo`800005Ool00000001oo`0200000goo00000002000QOol1
00000003001oo`00009oo`80009oo`80009oo`80009oo`80009oo`800004Ool007ooOol20002Ool2
0003Ool00`00Ool00009000bOol:00000goo00000002Ool300002goo001oo`00Ool007oo001oo`00
Ool00P000Woo0P000goo0P0000=oo`0000000Woo0P000goo0P0000=oo`0000000Woo0P000Woo00h0
07oo001oo`00Ool007oo001oo`00Ool007oo0P000goo0P0000=oo`00Ool02000<Goo200000=oo`00
Ool00P000goo0P0000Yoo`00Ool007oo001oo`00Ool0009oo`80009oo`80009oo`800004Ool007oo
0002Ool20002Ool2000017oo001oo`000Woo0P000Woo0P0000Yoo`00Ool007oo001oo`00Ool0009o
o`80009oo`800003Ool0000000P001Yoo`40000000<007oo00000goo00@007ooOol000=oo`8000=o
o`03001oogoo0080009oo`80009oo`80009oo`@0009oo`80009oo`X002Qoo`X000=oo`03001oogoo
00<00003Ool007oo0080009oo`8000=oo`05001oogooOol00003Ool01@00Oomoogoo00000goo00D0
07ooOomoo`0000=oo`05001oogooOol00003Ool01@00Oomoogoo00000goo00<007ooOol00P000Woo
0P000Woo00D007oo001oo`00009oo`04001oogooOol9000YOol90003Ool20002Ool2000017oo001o
ogoo0P000Woo0P000Woo00D007ooOomoo`0000=oo`05001oogooOol00003Ool01@00Oomoogoo0000
0goo0P000Woo00D007ooOomoo`0000=oo`05001oogooOol00003Ool20002Ool200000goo001oo`03
0002Ool20002Ool:000EOol100000003001oo`00009oo`80009oo`80009oo`80009oo`80009oo`80
009oo`80009oo`80009oo`80009oo`T00004Ool00000Ool7000POol700000goo00000002Ool00`00
Ool0000200000goo00000003Ool01@00Oomoogoo00000Woo0P000Woo0P000goo00L007ooOol007oo
001oo`020002Ool20002Ool02000Ool007oo001oo`00Ool20002Ool01000Ool007oo0P000Woo0P00
0goo0P000Woo00D007ooOomoo`00009oo`<00003Ool0000000<00005Ool00000001oo`07000OOol7
00001Goo00000000Ool01@0000Aoo`000000009oo`05001oogooOol00003Ool01000Oomoogoo0P00
0Woo0P000Woo0P000Woo0P0000Qoo`00Ool007oo001oo`000Woo0P000Woo0P0000Aoo`00Ool0009o
o`04001oogooOol20002Ool20002Ool01@00Oomoogoo00000goo0P0000Ioo`00000007oo0002Ool2
00000goo00000005000AOol100000003001oogoo008000=oo`04001oogoo0003Ool2000017oo001o
o`000Woo00<007oo00000goo0P000Woo0`0000=oo`0000000Woo0P000Woo0P000Woo10000Woo1P00
67oo1`0000=oo`0000000`0000=oo`0000000goo00<007ooOol00P000Woo0P000Woo0P0017oo00P0
07oo001oo`00Ool007oo0P0017oo00<007oo00000goo0P000Woo00<007oo00000goo00<007oo0000
0goo00/007oo001oo`00Ool007oo001oo`0000=oo`04001oo`00Ool30002Ool01@00Oomoogoo0000
0goo0P0000=oo`0000000`0000=oo`0000001@005goo1`0000=oo`0000000`0000=oo`0000000goo
00D007ooOomoo`00009oo`<00004Ool007oo0003Ool02`00Ool007oo001oo`00Ool007oo00000goo
00<007oo00000goo00<007oo00000Woo0P000goo00<007oo000017oo0P0000Qoo`00Ool007oo001o
o`0017oo0P000Woo0P000Woo0P000Woo00@007ooOomoo`80009oo`@00003Ool0000000D000eoo`40
000000D007ooOomoo`00009oo`80009oo`80009oo`80009oo`80009oo`80009oo`80009oo`80009o
o`<00003Ool0000000H0009oo`80009oo`<00003Ool0000000L000moo`T00003Ool00000009oo`80
009oo`D00005Ool007oo001oo`020002Ool20002Ool01000Ool007oo0P000goo00D007ooOomoo`00
009oo`80009oo`8000=oo`05001oogooOol00002Ool20003Ool01@00Oomoogoo00000Woo0P000goo
00X007ooOol007oo001oo`00Ool0009oo`800007Ool007oo001oo`00Ool00`000Woo00<007ooOol0
0`0000=oo`0000001`003goo1`0000=oo`00Ool00`000Woo00<007ooOol00`0000=oo`00Ool00`00
00aoo`00000007oo001oo`00Ool007oo0002Ool01000Oomoogoo0P000Woo00D007ooOomoo`0000=o
o`80009oo`05001oogooOol00003Ool20002Ool20002Ool01000Oomoogoo0P000Woo0P000Woo0P00
0Woo0P000Woo0P0000=oo`00Ool01`000Woo0P000Woo0P0000=oo`0000001`002Goo0@0000001000
Oomoo`000goo0P000Woo0P001Woo0P000Woo0P000Woo0P000goo00<007ooOol00P000Woo0P000goo
00<007ooOol00`0000=oo`0000001P000Woo200027oo2@0000=oo`0000000`0000Eoo`00000007oo
008000=oo`05001oogooOol00002Ool20003Ool01@00Oomoogoo00000Woo0P000goo00<007ooOol0
17oo0P000goo00D007ooOomoo`00009oo`8000=oo`05001oogooOol00007Ool01@00Oomoogoo0000
0goo00D007ooOomoo`0000=oo`05001oogooOol00003Ool01000Oomoogoo0P0000Moo`00000007oo
001oo`0300000goo000000070007Ool900001goo00000000Ool007oo00<00004Ool000000002Ool0
1@00Oomoogoo00000goo00D007ooOomoo`0000=oo`05001oogooOol00003Ool20006Ool01@00Oomo
ogoo00000goo0P000Woo00D007ooOomoo`0000=oo`8000Ioo`04001oogooOol20002Ool01@00Oomo
ogoo00000goo0P000Woo00D007ooOomoo`0000=oo`800005Ool00000001oo`0500000goo00000007
0005Ool100000005001oogooOol00002Ool20002Ool01000Oomoogoo0P000Woo0P000Woo0P000Woo
0P000Woo0P000Woo0P000Woo0P000Woo0P000Woo0P000Woo0`0000=oo`0000000Woo0P000Woo5@00
00=oo`0000000Woo0P000Woo0P000Woo0`0000=oo`0000000Woo0P000Woo00@007oo001oo`80009o
o`80009oo`08001oo`00Ool007oo001oo`8000Aoo`07001oo`00Ool007oo00001Goo00L007oo001o
o`00Ool00003Ool20002Ool02000Ool007oo001oo`00Ool20002Ool02000Ool007oo001oo`00Ool2
0002Ool01000Ool007oo0`0000Eoo`00000007oo00<00005Ool00000001oo`0E000017oo00000000
0Woo0P0000Eoo`00000007oo00<00004Ool007oo0002Ool2000027oo001oo`00Ool007oo0002Ool2
000027oo001oo`00Ool007oo0002Ool20003Ool01@00Ool007oo00000Woo00<007ooOol00Woo00L0
07oo001oo`00Ool00004Ool2000027oo001oo`00Ool007oo0002Ool20002Ool2000017oo001oo`00
0Woo0P000Woo0P0000Aoo`000000009oo`80009oo`80009oo`800003Ool0000000T0005oo`400000
00<007oo00000Woo00@007ooOomoo`80009oo`80009oo`04001oogooOol20002Ool20002Ool20002
Ool20002Ool20002Ool20002Ool20002Ool20002Ool20002Ool300000goo0000000F0002Ool20002
Ool20002Ool20002Ool20002Ool300000goo00000002Ool20002Ool01000Ool007oo0P000Woo0P00
17oo00H007oo001oo`00Ool20004Ool01`00Ool007oo001oo`0000=oo`80009oo`05001oo`00Ool0
0003Ool20003Ool200001Goo001oo`00Ool00P000Woo00@007oo001oo`<00003Ool00000009oo`<0
009oo`800005Ool00000001oo`0I00001Goo00000000Ool00P000Woo0`000Woo0P0000Qoo`000000
07oo001oo`000Woo0P0000Qoo`00Ool007oo001oo`000Woo0P000goo00D007oo001oo`00009oo`80
00=oo`05001oo`00Ool00002Ool01000Oomoogoo0P0000Ioo`00Ool007oo0004Ool20002Ool20000
17oo001oo`000Woo0P000Woo0P0000Aoo`000000009oo`80009oo`80009oo`80009oo`80009oo``0
005oo`40000000<007oo00000Woo0P000Woo00@007ooOomoo`80009oo`80009oo`8000=oo`03001o
ogoo0080009oo`8000=oo`03001oogoo0080009oo`80009oo`800004Ool00000Ool300000goo0000
00080008Ool;00000goo0000000300000goo00000002Ool20002Ool20003Ool01@00Oomoogoo0000
0Woo0P000goo00D007ooOomoo`00009oo`8000=oo`05001oogooOol00002Ool20003Ool01@00Oomo
ogoo00000goo00D007ooOomoo`0000=oo`05001oogooOol00003Ool01@00Oomoogoo00000goo00D0
07ooOomoo`0000=oo`04001oogooOol200001goo001oo`00Ool007oo00<00003Ool0000000P000Uo
o`X0000:Ool00000001oo`00Ool007oo0002Ool01000Oomoogoo0P000Woo00D007ooOomoo`0000=o
o`05001oogooOol00003Ool01@00Oomoogoo00000goo00D007ooOomoo`0000=oo`05001oogooOol0
0003Ool20002Ool01@00Oomoogoo00000goo0P000Woo00D007ooOomoo`0000=oo`80009oo`05001o
ogooOol00003Ool20002Ool20002Ool200000goo0000000300000goo000000090005Ool100000003
001oo`00009oo`8000=oo`03001oogoo0080009oo`80009oo`80009oo`80009oo`80009oo`80009o
o`80009oo`03001oo`0000800003Ool00000009oo`80009oo`X0011oo`/00003Ool00000009oo`<0
0007Ool00000001oo`00Ool00P000Woo00@007oo001oo`8000=oo`07001oogoo001oo`00Ool00P00
0Woo0P000goo00<007ooOol00P000Woo0P000goo00D007ooOomoo`00009oo`8000=oo`07001oogoo
001oo`00Ool00P000goo00d007ooOol007oo001oo`00Ool007oo001oo`0300001Goo00000000Ool0
0`0000=oo`00000020004Goo2P0000Eoo`00000007oo00<0000>Ool00000001oo`00Ool007oo001o
o`00Ool0009oo`04001oogooOol2000017oo001oo`000Woo00@007ooOomoo`80009oo`05001oogoo
Ool00003Ool20002Ool20002Ool01000Oomoogoo0P000Woo0P000Woo0P000Woo0P000Woo0P0000Ao
o`00Ool0009oo`800003Ool007oo00<0009oo`80009oo`800003Ool0000000T000Uoo`40000000<0
07oo00000Woo00@007ooOomoo`800004Ool007oo0002Ool00`00Ool00003Ool200001Woo001oogoo
001oo`80009oo`80009oo`80009oo`80009oo`D00003Ool0000000D001Yoo`L0009oo`04001oo`00
0002Ool20002Ool20002Ool20002Ool300001Goo001oo`00Ool00P0017oo00P007oo001oo`00Ool0
07oo0P0017oo00L007oo001oo`00Ool00003Ool02`00Ool007oo001oo`00Ool007oo00000goo00P0
07oo001oo`00Ool007oo0P000Woo0`0000=oo`0000000goo00<007ooOol01@0000=oo`0000001P00
6Goo200000=oo`0000000`000Woo00@007ooOomoo`800004Ool000000002Ool2000027oo001oo`00
Ool007oo0003Ool02`00Ool007oo001oo`00Ool007oo00000goo00L007oo001oo`00Ool00004Ool2
000027oo001oo`00Ool007oo0004Ool200001Goo001oo`00Ool00`000Woo0P000Woo0P000Woo0P00
0Woo0P0000Aoo`00Oomoo`L000ioo`40000000D007ooOomoo`00009oo`80009oo`04001oogooOol2
0002Ool20002Ool20002Ool20002Ool20002Ool20002Ool200000goo000000030002Ool7000ROol8
00001goo00000000Ool007oo00<00003Ool00000009oo`80009oo`8000=oo`03001oogoo0080009o
o`8000=oo`08001oogoo001oo`00Ool000Aoo`08001oogoo001oo`00Ool000=oo`8000=oo`80009o
o`03001oogoo008000=oo`03001oogoo0080009oo`<00003Ool007oo00<00003Ool0000000D002=o
o`L00005Ool00000001oo`050002Ool20002Ool01000Oomoogoo0P000Woo00D007ooOomoo`0000=o
o`8000=oo`05001oo`00Ool00002Ool00`00Oomoo`02Ool01@00Ool007oo00000Woo00@007ooOomo
o`80009oo`80009oo`04001oogooOol20002Ool20002Ool200001Goo00000000Ool01@0000=oo`00
00001P004Woo0@0000001000Oomoo`000goo0P000goo00<007ooOol00P000goo00<007ooOol00P00
0Woo0P000Woo00@007oo0000009oo`80009oo`800003Ool0000000H002Yoo`P00003Ool00000009o
o`<00003Ool007oo0080009oo`<00003Ool00000009oo`8000=oo`05001oogooOol00003Ool01@00
Oomoogoo00000goo00D007ooOomoo`0000=oo`05001oogooOol00003Ool01`00Oomoo`00Ool007oo
00<00007Ool007oo001oo`00Ool00P000Woo0`0000=oo`00Ool01@00:goo1@0000=oo`00Ool00`00
0Woo0P0000Moo`00Ool007oo001oo`030002Ool20002Ool01@00Oomoogoo00000goo00D007ooOomo
o`0000=oo`80009oo`05001oogooOol00003Ool01@00Oomoogoo00000goo0P000Woo0P0000Aoo`00
0000009oo`800003Ool007oo00<0009oo`800003Ool0000000H001Ioo`40000000<007ooOol00P00
0Woo0P000Woo0P000Woo0P000Woo00@007oo0000009oo`80009oo`80009oo`80009oo`D00003Ool0
0000008003Aoo`L00005Ool00000001oo`020002Ool20002Ool01P00Ool007oo001oo`80009oo`04
001oo`00Ool20002Ool20003Ool200000goo00000002Ool20002Ool03P00Ool007oo001oo`00Ool0
07oo001oo`00Ool20002Ool01000Ool007oo0P000Woo0`0000=oo`00Ool01000=Goo100000=oo`00
Ool00`000Woo0P0000Aoo`00Ool0009oo`80000>Ool007oo001oo`00Ool007oo001oo`00Ool0009o
o`80009oo`800004Ool007oo0002Ool20002Ool2000017oo001oo`000Woo0P0000Aoo`00Oomoo`80
009oo`80009oo`800005Ool00000001oo`07000KOol100000004001oo`00Ool30002Ool20002Ool2
0002Ool20002Ool20002Ool20002Ool200000goo001oo`09000lOol900000goo001oo`020002Ool2
0003Ool00`00Oomoo`020002Ool20002Ool01000Ool007oo0`0000=oo`0000000Woo00P007oo001o
o`00Ool007oo0P000Woo0`000Woo00D007ooOomoo`00009oo`<00003Ool007oo00P003eoo`P00003
Ool007oo00<0009oo`05001oogooOol00002Ool30002Ool2000027oo001oo`00Ool007oo0002Ool2
000027oo00000000Ool007oo0002Ool20002Ool20002Ool20002Ool20002Ool200000goo001oo`09
000OOol100000003001oo`0000=oo`03001oogoo0080009oo`04001oogooOol20002Ool2000017oo
000007oo0P000Woo1@00Bgoo1@0000Moo`00000007oo001oo`0300000goo00000003Ool00`00Oomo
o`020003Ool01@00Oomoogoo00000goo00D007ooOomoo`0000=oo`03001oogoo00<00003Ool00000
00<00005Ool00000001oo`05001;Ool500001Goo00000000Ool01@0000Aoo`000000009oo`05001o
ogooOol00003Ool01@00Oomoogoo00000goo00@007ooOomoo`80009oo`04001oogooOol200001goo
00000000Ool007oo00<00003Ool0000000<002Moo`40000000@007oo001oo`<0009oo`03001oo`00
00D0009oo`80009oo`P005Aoo`T00003Ool00000009oo`<00007Ool007oo001oo`00Ool00`0000]o
o`00Ool007oo001oo`00Ool007oo00L00003Ool00000009oo`P005Eoo`P0009oo`800003Ool00000
00D00006Ool007oo001oo`000Woo0P0000]oo`00000007oo001oo`00Ool007oo00<0009oo`80009o
o`P002]oo`40000000<007ooOol00P000Woo0P0000=oo`0000002`00I7oo0`0000=oo`0000001`00
00=oo`0000000Woo0`0000=oo`0000000Woo0`0000=oo`0000001`0000Aoo`00000006Aoo`800003
Ool0000000L00004Ool000000002Ool20002Ool20002Ool200000goo00000007000017oo00000000
<goo0@0000000`00Ool00002Ool300000goo0000001mOol00`00Oomoo`0300001Goo00000000Ool0
0`0000=oo`000000OGoo0P0000Eoo`00000007oo00<00004Ool000000002Ool00`00Oomoo`0nOol1
00000003001oogoo0?moofEoo`400000o`00J@0000000`00Oomoo`3oOomUOol100000004001oo`00
0002Ool00`00Oomoo`24Ool200000goo00000003Ool00`00Oomoo`23Ool20002Ool01@00Oomoogoo
0000@goo0@0000001000Oomoo`000Woo0P000Woo00<007ooOol0O7oo00@007ooOol000=oo`04001o
o`00Ool20002Ool2001mOol20002Ool02P00Oomoo`00Ool007oo001oo`000Woo00<007ooOol0?Goo
0@0000000`00Oomoo`05Ool01000Oomoo`00NGoo00X007ooOol007oo001oo`00Ool000Moo`04001o
ogoo001iOol02000Oomoo`00Ool007oo0002Ool00`00Oomoo`04Ool01@00Oomoogoo0000?Woo0@00
00001000Oomoogoo0P00OWoo00<007oo00002goo0P00Ogoo0P0037oo0P00A7oo0@0000000`00Oomo
o`05Ool20002Ool4001aOol01000Oomoo`000Woo00D007ooOomoo`0000=oo`03001oogoo00Eoo`80
0003Ool007oo00<0071oo`04001oogoo0003Ool01000Oomoo`0017oo00<007ooOol017oo0P0000=o
o`00Ool00`00>Woo0@0000000`00Oomoo`1mOol00`00Ool00003Ool00`00Oomoo`02Ool01@00Oomo
ogoo0000Ogoo0P0017oo00D007ooOomoo`0000=oo`03001oogoo04Moo`40000000<007ooOol00Woo
00D007ooOomoo`0000Qoo`03001oogoo07ioo`05001oogooOol00009Ool00`00Oomoo`1mOol01@00
Oomoogoo00002Goo00<007ooOol0=Goo0@0000001@00Oomoogoo00001goo0P000Woo00@007oo0000
06Uoo`05001oogoo001oo`020003Ool00`00Oomoo`05Ool01@00Oomoogoo00001goo00D007ooOol0
07oo008006Yoo`03001oo`00009oo`80009oo`03001oogoo00Eoo`05001oogooOol00007Ool01000
Oomoogoo0P0000=oo`00Ool0=Goo0@0000000`00Oomoo`06Ool00`00Oomoo`05Ool01000Oomoo`00
J7oo00<007ooOol01goo00D007ooOomoo`0000]oo`03001oogoo00Eoo`04001oogoo001XOol00`00
Oomoo`07Ool00`00Oomoo`02Ool00`00Oomoo`09Ool00`00Oomoo`05Ool00`00Ool0000fOol10000
0003001oogoo07Uoo`03001oo`0000Aoo`03001oogoo00Uoo`03001oogoo07Qoo`03001oo`0000Ao
o`03001oogoo00Yoo`03001oogoo04Ioo`40000000@007ooOomoo`80009oo`05001oogooOol00004
Ool00`00Oomoo`02Ool2001WOol01@00Oomoogoo00001Woo00<007ooOol01Goo00D007oo001oo`00
00=oo`03001oogoo009oo`03001oogoo009oo`8006Moo`05001oogooOol00007Ool00`00Oomoo`04
Ool00`00Ool00002Ool01@00Oomoogoo00000goo00<007ooOol00goo0P00<Woo0@0000000`00Oomo
o`0:Ool01000Oomoo`000goo0P00Hgoo00<007oo00007goo00D007ooOomoo`0000=oo`8006=oo`03
001oo`0001moo`04001oogoo0003Ool2000dOol100000003001oogoo07Ioo`8000]oo`03001oogoo
07ioo`8000]oo`03001oogoo051oo`40000000<007ooOol01Woo00<007ooOol02Goo00H007ooOol0
07oo001MOol00`00Ool00008Ool00`00Oomoo`02Ool01`00Oomoogoo001oo`0000Uoo`03001oogoo
00Qoo`06001oogoo001oo`00GGoo00@007ooOol000Qoo`05001oogooOol00003Ool00`00Ool00009
Ool00`00Oomoo`08Ool01000Oomoo`000Woo00<007ooOol0;Goo0@0000000`00Ool0000FOol00`00
Ool0001SOol00`00Oomoo`0>Ool00`00Oomoo`0EOol2001TOol00`00Oomoo`0=Ool00`00Oomoo`0E
Ool2000_Ool100000003001oogoo00Aoo`05001oogooOol00005Ool01@00Oomoogoo00001Goo00<0
07ooOol0GGoo00@007ooOol000]oo`03001oogoo00Uoo`05001oogooOol00005Ool01000Oomoo`00
1Woo00<007ooOol0G7oo00D007ooOomoo`0000Yoo`03001oogoo00Uoo`05001oogooOol00005Ool0
1@00Oomoogoo00001Goo00<007ooOol0;7oo0@0000001000Oomoo`000Woo00<007ooOol01Goo00<0
07oo00000goo00@007ooOomoo`8005]oo`<000=oo`03001oogoo00Aoo`03001oogoo00aoo`04001o
ogoo0007Ool00`00Ool00003Ool00`00Oomoo`02Ool00`00Oomoo`1IOol200001Goo001oogoo0000
1Woo00<007ooOol03Goo00@007ooOol000Ioo`03001oo`0000Aoo`05001oogooOol0000aOol10000
0003001oogoo01Uoo`03001oogoo05Eoo`03001oo`0000eoo`03001oogoo02=oo`03001oogoo05Io
o`03001oo`0000aoo`03001oogoo02Aoo`03001oogoo02]oo`40000000<007ooOol06goo00<007oo
Ool0Ggoo00<007ooOol00Woo00@007ooOol000Aoo`03001oogoo01eoo`03001oogoo05ioo`03001o
ogoo00=oo`04001oogoo0004Ool00`00Oomoo`0LOol00`00Oomoo`0YOol100000003001oogoo00Mo
o`03001oogoo00Qoo`03001oogoo008000Aoo`8005Aoo`04001oogoo0002Ool01@00Ool007oo0000
3Woo00<007ooOol027oo00<007ooOol02Goo00@007oo000000Aoo`8005Aoo`04001oogoo0003Ool2
00000goo001oo`0=Ool00`00Oomoo`09Ool00`00Oomoo`08Ool00`00Oomoo`020003Ool00`00Ool0
000ZOol100000003001oo`0000Qoo`05001oogooOol00002Ool00`00Oomoo`08Ool00`00Oomoo`1G
Ool00`00Oomoo`0GOol00`00Oomoo`05Ool01@00Oomoogoo00000Woo00<007ooOol027oo00<007oo
Ool0Egoo00<007ooOol05goo00<007ooOol01Woo00D007ooOomoo`00009oo`03001oogoo00Qoo`03
001oogoo02Yoo`40000000<007ooOol017oo00<007ooOol02Goo00<007ooOol01goo00<007oo0000
DGoo0`003Goo00<007ooOol01Goo00<007ooOol02Goo00<007ooOol02Woo00<007ooOol01Woo00@0
07ooOol0051oo`<000ioo`03001oogoo00Aoo`03001oogoo00Uoo`03001oogoo00Yoo`03001oogoo
00Ioo`04001oogoo000ZOol100000003001oo`0000]oo`03001oogoo00]oo`03001oogoo055oo`03
001oogoo00Eoo`03001oogoo01Ioo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo051oo`03
001oogoo00Eoo`03001oogoo01Ioo`03001oogoo00Uoo`03001oogoo00]oo`03001oogoo02]oo`40
000000<007ooOol017oo00<007ooOol03Goo00<007ooOol01Woo00<007oo0000D7oo00<007ooOol0
2goo00D007ooOomoo`0000moo`03001oogoo00ioo`03001oogoo00Eoo`03001oo`00051oo`03001o
ogoo00]oo`05001oogooOol0000?Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool2000XOol100000003
001oogoo01=oo`05001oogooOol00002Ool01000Oomoogoo0`00Bgoo00<007ooOol017oo00L007oo
Oomoo`00Ool00008Ool01@00Oomoogoo00006Woo00D007ooOomoo`0000=oo`06001oogoo001oo`00
C7oo00<007ooOol00goo00D007ooOomoo`00009oo`03001oogoo00Eoo`05001oogooOol0000KOol0
1000Oomoo`000goo00@007ooOomoo`<002Ioo`40000000<007ooOol06Goo00<007ooOol0Dgoo0P00
2Goo00<007ooOol03Goo00<007ooOol06Woo00<007ooOol0E7oo0P0027oo00<007ooOol03Goo0P00
6goo00<007ooOol0;7oo0@0000000`00Oomoo`08Ool01000Oomoo`004Goo00<007ooOol0BWoo0P00
2Goo00<007ooOol07Woo00D007ooOomoo`00015oo`03001oogoo04Yoo`03001oogoo00Qoo`03001o
ogoo01eoo`05001oogooOol0000AOol00`00Oomoo`0WOol100000004001oogoo0006Ool00`00Oomo
o`06Ool01@00Oomoogoo00002Goo00@007ooOol004Qoo`<001Qoo`03001oogoo00Ioo`03001oogoo
00Eoo`03001oogoo00Eoo`05001oogooOol00009Ool01000Oomoo`00B7oo00<007oo00006Goo00<0
07ooOol01Woo00<007ooOol017oo00<007ooOol01Goo00<007ooOol00Woo00<007ooOol01Woo00@0
07ooOol002Ioo`40000000<007ooOol00goo00<007ooOol05goo00<007ooOol0DGoo00<007ooOol0
2Woo00<007ooOol047oo00<007ooOol05goo00<007ooOol0DGoo00<007ooOol02Woo00<007ooOol0
4Goo00<007ooOol05goo00<007ooOol09goo0@0000000`00Oomoo`0QOol00`00Ool00015Ool01000
Oomoo`00@goo0P00AGoo00@007ooOol004=oo`03001oo`0002=oo`40000000<007oo000017oo00<0
07ooOol01Goo00<007ooOol027oo0P000Woo00@007ooOol000=oo`03001oo`0004Yoo`04001oogoo
0003Ool01@00Oomoogoo00000goo00D007ooOomoo`0000Yoo`03001oogoo00=oo`03001oogoo00Ao
o`03001oogoo00Uoo`04001oogoo0003Ool00`00Oomoo`03001:Ool01@00Oomoogoo00000Woo00D0
07ooOomoo`0000Aoo`05001oogooOol0000:Ool00`00Oomoo`02Ool00`00Oomoo`05Ool00`00Oomo
o`08Ool20002Ool01000Oomoo`000Woo0`008goo0@0000000`00Oomoo`0DOol00`00Oomoo`0=Ool0
0`00Oomoo`11Ool00`00Oomoo`0;Ool00`00Oomoo`08Ool00`00Oomoo`03Ool00`00Oomoo`0HOol0
0`00Oomoo`0>Ool00`00Oomoo`11Ool00`00Oomoo`0:Ool00`00Oomoo`09Ool00`00Oomoo`02Ool0
0`00Oomoo`0HOol00`00Oomoo`0>Ool00`00Oomoo`0POol100000003001oogoo01aoo`03001oogoo
04Qoo`07001oogooOol007oo00002goo00<007ooOol0;goo00<007ooOol0Agoo00<007ooOol00Woo
0P0037oo00<007ooOol0;Woo00<007ooOol0:7oo0@0000000`00Oomoo`08Ool01000Oomoo`000goo
0P000Woo00<007ooOol02goo0P000Woo00<007ooOol0?Goo0`002Woo00<007ooOol057oo00<007oo
00002goo00<007oo00000goo00<007oo00000Woo00<007ooOol02Woo0P000Woo00<007ooOol0?Woo
00<007oo00002Goo00<007ooOol057oo00<007oo00002goo00<007oo000017oo0P000Woo00<007oo
Ool02goo0P000Woo00<007ooOol07goo0@0000000`00Oomoo`07Ool00`00Oomoo`1NOol00`00Oomo
o`0ZOol00`00Oomoo`1NOol00`00Oomoo`0ZOol00`00Oomoo`0nOol100000005001oogooOol0000E
Ool00`00Oomoo`1SOol00`00Oomoo`0?Ool00`00Oomoo`0DOol00`00Oomoo`1ROol00`00Oomoo`0?
Ool00`00Oomoo`0DOol00`00Oomoo`0]Ool100000003001oogoo009oo`03001oogoo01=oo`03001o
ogoo0080009oo`8000Eoo`<003aoo`08001oogooOol007oo001oo`8000=oo`03001oogoo00Ioo`05
001oogooOol00005Ool00`00Oomoo`09Ool00`00Oomoo`0BOol01@00Oomoogoo00000Woo00<007oo
000017oo0`00?7oo00L007ooOomoo`00Ool00002Ool01@00Oomoogoo000027oo00D007ooOomoo`00
00Eoo`03001oogoo00Uoo`03001oogoo01=oo`04001oogoo0003Ool20004Ool4000NOol100000003
001oogoo00eoo`03001oogoo01=oo`03001oogoo04moo`8000aoo`03001oogoo01Eoo`03001oogoo
01=oo`03001oogoo051oo`8000]oo`03001oogoo01Eoo`03001oogoo01=oo`03001oogoo029oo`40
000000<007ooOol01Goo00<007ooOol037oo00<007ooOol03Woo00@007ooOol003]oo`04001oogoo
000:Ool00`00Oomoo`0HOol00`00Oomoo`03Ool00`00Oomoo`0=Ool00`00Oomoo`0>Ool01000Oomo
o`00>Woo00@007ooOol000]oo`03001oogoo01Moo`03001oogoo00Aoo`03001oogoo00aoo`03001o
ogoo00ioo`04001oogoo000NOol100000003001oogoo021oo`03001oogoo00=oo`03001oogoo03Uo
o`<000Aoo`03001oogoo00eoo`03001oogoo00Eoo`03001oogoo00=oo`03001oogoo02Aoo`03001o
ogoo00Aoo`03001oogoo03Qoo`<000Eoo`03001oogoo00aoo`03001oogoo00Eoo`03001oogoo00Ao
o`03001oogoo02=oo`03001oogoo00Aoo`03001oogoo01ioo`40000000<007ooOol03Woo00<007oo
Ool00Woo00<007ooOol03goo00D007ooOomoo`0004Aoo`03001oogoo02aoo`05001oogooOol0000A
Ool01@00Oomoogoo0000A7oo00<007ooOol0;7oo00D007ooOomoo`00015oo`03001oogoo009oo`03
001oogoo01]oo`40000000<007ooOol02Woo00<007ooOol02Woo00<007ooOol01Woo00<007ooOol0
?Woo00@007ooOol002=oo`03001oogoo00eoo`03001oogoo00Yoo`03001oogoo00Ioo`03001oogoo
03moo`04001oogoo000SOol00`00Oomoo`0=Ool00`00Oomoo`09Ool00`00Oomoo`06Ool00`00Oomo
o`0UOol100000003001oogoo00Ioo`04001oogoo0007Ool00`00Oomoo`07Ool20006Ool00`00Oomo
o`04Ool2000gOol01000Oomoo`000goo00@007ooOol000=oo`03001oogoo01Moo`03001oogoo00Io
o`04001oogoo0007Ool00`00Oomoo`07Ool20007Ool00`00Oomoo`03Ool3000gOol01000Oomoo`00
0Woo00D007ooOomoo`00009oo`03001oogoo01Qoo`03001oogoo00Ioo`04001oogoo0007Ool00`00
Oomoo`07Ool20006Ool00`00Oomoo`04Ool2000KOol100000003001oogoo01Uoo`03001oogoo00Qo
o`03001oogoo009oo`8003Ioo`03001oogoo00Ioo`03001oogoo00moo`03001oogoo02Qoo`03001o
ogoo00Qoo`03001oogoo00=oo`03001oogoo03Aoo`03001oogoo00Ioo`03001oogoo011oo`03001o
ogoo02Moo`03001oogoo00Uoo`03001oogoo009oo`03001oogoo01]oo`40000000D007ooOomoo`00
01eoo`03001oogoo00Eoo`03001oogoo009oo`03001oogoo03Ioo`03001oogoo00moo`05001oogoo
Ool0000DOol00`00Oomoo`0JOol00`00Oomoo`05Ool00`00Oomoo`02Ool00`00Oomoo`0fOol00`00
Oomoo`0?Ool00`00Oomoo`02Ool00`00Oomoo`0AOol00`00Oomoo`0KOol00`00Oomoo`05Ool00`00
Oomoo`02Ool00`00Oomoo`0HOol100000003001oogoo009oo`03001oogoo00Qoo`03001oogoo01Uo
o`05001oogooOol0000aOol20002Ool00`00Oomoo`09Ool01`00Oomoogoo001oo`0000Uoo`03001o
ogoo009oo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo01Qoo`05001oogooOol0000aOol2
0002Ool00`00Oomoo`09Ool01`00Oomoogoo001oo`0000Yoo`05001oogooOol0000;Ool00`00Oomo
o`09Ool00`00Oomoo`0IOol01000Oomoo`006Woo0@0000000`00Oomoo`0EOol00`00Oomoo`16Ool0
0`00Ool0000SOol00`00Oomoo`0LOol00`00Oomoo`17Ool2000SOol00`00Oomoo`0MOol00`00Oomo
o`0_Ool100000003001oogoo00Eoo`03001oogoo01]oo`03001oogoo00Aoo`03001oogoo035oo`03
001oo`0000Eoo`04001oogoo000ROol00`00Oomoo`05Ool00`00Oomoo`0KOol00`00Oomoo`04Ool0
0`00Oomoo`0`Ool01000Oomoo`0017oo00D007ooOomoo`00025oo`03001oogoo00Eoo`03001oogoo
01]oo`03001oogoo00Aoo`03001oogoo01Yoo`40000000<007ooOol03Woo00D007ooOomoo`0000=o
o`05001oogooOol00002Ool01000Oomoo`000goo00<007ooOol00P0017oo00<007ooOol0;Goo00D0
07oo001oo`0000Aoo`05001oogooOol00002Ool00`00Oomoo`08Ool00`00Oomoo`05Ool00`00Oomo
o`05Ool00`00Oomoo`0AOol01@00Oomoogoo00000goo00<007ooOol00Woo00<007oo00000goo00@0
07ooOol000=oo`8000Aoo`03001oogoo02aoo`03001oogoo008000Aoo`03001oogoo009oo`03001o
o`0000]oo`03001oogoo00Aoo`03001oogoo00Eoo`03001oogoo015oo`03001oogoo009oo`05001o
ogooOol00003Ool00`00Ool00003Ool01@00Oomoogoo00000Woo0P0017oo00<007ooOol05goo0@00
00000`00Oomoo`0NOol00`00Oomoo`0=Ool2001mOol00`00Oomoo`0=Ool00`00Oomoo`1kOol00`00
Oomoo`0=Ool2000GOol100000003001oogoo00Uoo`03001oogoo00Eoo`03001oogoo01Yoo`03001o
o`00035oo`03001oogoo00]oo`03001oogoo01Eoo`03001oogoo00eoo`03001oogoo00Eoo`03001o
ogoo01Uoo`03001oo`00039oo`03001oogoo00]oo`03001oogoo01Aoo`03001oogoo00eoo`03001o
ogoo00Eoo`03001oogoo01Uoo`03001oo`0001Moo`40000000<007ooOol01Goo00<007ooOol00Woo
00<007ooOol037oo00<007ooOol01goo00<007ooOol00Woo00D007ooOol007oo00800003Ool007oo
02Yoo`03001oogoo009oo`03001oogoo009oo`03001oogoo00]oo`05001oogooOol0000COol00`00
Oomoo`05Ool01@00Oomoogoo00003goo00<007ooOol01Woo00<007ooOol00Woo00D007ooOol007oo
0080009oo`03001oogoo02Uoo`03001oogoo009oo`03001oogoo009oo`03001oogoo00Yoo`05001o
ogooOol0000COol00`00Oomoo`05Ool01@00Oomoogoo00003goo00<007ooOol01Woo00<007ooOol0
0Woo00H007ooOomoo`00Ool200000goo001oo`0EOol100000003001oogoo02aoo`03001oogoo031o
o`03001oogoo00eoo`03001oogoo009oo`03001oogoo00Eoo`03001oogoo03aoo`03001oogoo02mo
o`03001oogoo00ioo`05001oogooOol00007Ool00`00Oomoo`0lOol00`00Oomoo`0HOol100000003
001oogoo02Eoo`03001oogoo03=oo`03001oogoo00eoo`03001oogoo019oo`03001oogoo031oo`03
001oogoo03=oo`8000moo`03001oogoo019oo`03001oogoo02moo`03001oogoo021oo`40000000@0
07ooOomoo`8000]oo`03001oogoo00Eoo`03001oogoo00Qoo`03001oogoo00=oo`8000Ioo`03001o
o`0002Qoo`04001oo`000004Ool01000Oomoo`000goo00<007ooOol05goo00<007ooOol02Goo0P00
2goo00<007ooOol01Goo00<007ooOol027oo00<007ooOol00goo0P001Woo00<007oo0000:7oo00@0
07oo000000Aoo`04001oogoo0003Ool00`00Oomoo`0HOol00`00Oomoo`09Ool2000;Ool00`00Oomo
o`04Ool00`00Oomoo`09Ool00`00Oomoo`03Ool20005Ool01000Oomoo`005Goo0@0000000`00Oomo
o`0EOol00`00Oomoo`02Ool00`00Oomoo`0?Ool00`00Oomoo`0ZOol00`00Oomoo`0;Ool00`00Oomo
o`02Ool00`00Oomoo`0EOol00`00Oomoo`0MOol01@00Oomoogoo00004Goo00<007ooOol0:goo00<0
07ooOol02Woo00<007ooOol00goo00<007ooOol05Goo00<007ooOol077oo00D007ooOomoo`00015o
o`03001oogoo01Uoo`40000000<007ooOol04Goo00<007ooOol02Goo0P004Woo00<007ooOol0:Goo
00<007ooOol0@7oo00<007ooOol02Goo00<007ooOol04Goo00<007ooOol0:Goo00<007ooOol0?goo
00<007ooOol02Goo0P004Woo00<007ooOol04goo0@0000000`00Oomoo`06Ool00`00Oomoo`04Ool0
0`00Oomoo`0BOol00`00Oomoo`0<Ool3000UOol00`00Ool00002Ool00`00Oomoo`0?Ool00`00Oomo
o`05Ool00`00Oomoo`04Ool00`00Oomoo`07Ool00`00Ool00008Ool00`00Oomoo`04Ool00`00Oomo
o`0BOol00`00Oomoo`0=Ool2000UOol00`00Ool00002Ool00`00Oomoo`0?Ool00`00Oomoo`05Ool0
0`00Oomoo`05Ool00`00Oomoo`06Ool01000Oomoo`0027oo00<007ooOol017oo00<007ooOol04Woo
00<007ooOol037oo0`004goo0@0000000`00Oomoo`0[Ool01`00Oomoogoo001oo`00009oo`03001o
ogoo02aoo`03001oogoo05=oo`03001oogoo009oo`80009oo`03001oogoo02aoo`03001oogoo05Ao
o`04001oogooOol20002Ool00`00Oomoo`0AOol100000003001oogoo00Uoo`03001oogoo00Ioo`03
001oogoo00Eoo`03001oogoo00Uoo`04001oogoo0008Ool00`00Oomoo`0SOol00`00Oomoo`03Ool0
0`00Oomoo`03Ool00`00Oomoo`09Ool01@00Oomoogoo00008Woo00<007ooOol01Woo00<007ooOol0
1Goo00<007ooOol02Woo00<007oo000027oo00<007ooOol08goo00<007ooOol00goo00<007ooOol0
17oo00<007ooOol027oo00<007ooOol00Woo00<007ooOol087oo00<007ooOol01Woo00<007ooOol0
1Goo00<007ooOol02Goo00<007oo00002Goo00<007ooOol04Woo0@0000000`00Oomoo`0:Ool00`00
Oomoo`0HOol00`00Oomoo`04Ool01@00Oomoogoo00000goo00<007ooOol07goo0`0000=oo`00Ool0
1Goo00<007ooOol017oo00D007ooOomoo`0000]oo`03001oogoo009oo`03001oogoo00Moo`03001o
ogoo00=oo`03001oogoo00Uoo`03001oogoo01Qoo`03001oogoo00Aoo`04001oogoo0004Ool00`00
Oomoo`0OOol20002Ool00`00Oomoo`03Ool00`00Oomoo`04Ool01@00Oomoogoo000037oo00D007oo
Oomoo`0000Uoo`03001oogoo00=oo`03001oogoo00Uoo`03001oogoo01Qoo`03001oogoo00Aoo`05
001oogooOol00003Ool00`00Oomoo`0@Ool100000003001oogoo00Eoo`03001oogoo01Yoo`03001o
ogoo039oo`8000Aoo`03001oogoo01=oo`03001oogoo00Eoo`03001oogoo01Eoo`03001oogoo01Yo
o`03001oogoo03=oo`8000Aoo`03001oogoo01=oo`03001oogoo00Aoo`03001oogoo01Eoo`03001o
ogoo01Yoo`03001oogoo02=oo`40000000<007ooOol0:Woo00<007ooOol00Woo0P000Woo00<007oo
000087oo00@007ooOol000Eoo`03001oogoo00Yoo`03001oogoo04Yoo`03001oogoo009oo`06001o
ogoo001oo`0087oo00@007ooOol000Eoo`03001oogoo00Yoo`03001oogoo04Yoo`03001oogoo009o
o`800005Ool007ooOol0000AOol100000004001oogoo0002Ool00`00Oomoo`0?Ool20002Ool01000
Oomoogoo0P000Woo00<007ooOol0<Goo00<007ooOol00Woo00<007ooOol01Woo00<007ooOol06goo
00<007oo00002goo00<007oo00004Goo00D007oo001oo`0000Aoo`05001oogooOol0000cOol01@00
Oomoogoo00002Goo00<007ooOol06Woo00<007oo00002goo00<007oo00004Woo0P000Woo00@007oo
Oomoo`80009oo`03001oogoo02=oo`40000000<007ooOol03goo00<007oo000057oo00<007ooOol0
37oo0`007goo00<007ooOol0@goo00<007oo000057oo00<007ooOol03Goo0P007goo00<007ooOol0
@goo00<007oo000057oo00<007ooOol03Goo0P0047oo0@0000000`00Oomoo`0YOol01`00Oomoogoo
001oo`0000Eoo`03001oogoo01eoo`03001oogoo00Qoo`03001oogoo00aoo`03001oogoo01=oo`03
001oogoo031oo`06001oogoo001oo`001Goo00<007ooOol07Goo00<007ooOol02Goo00<007ooOol0
2goo00<007ooOol04goo00<007ooOol0<7oo00@007ooOomoo`8000Eoo`03001oogoo011oo`400000
00<007ooOol01goo00<007ooOol00Woo00<007ooOol05goo00D007ooOomoo`0000Eoo`03001oogoo
00800004Ool00000000QOol01P00Oomoo`00Ool000Aoo`05001oogooOol00003Ool00`00Oomoo`02
Ool00`00Oomoo`03Ool00`00Oomoo`0AOol00`00Oomoo`07Ool00`00Oomoo`03Ool00`00Oomoo`0F
Ool01@00Oomoogoo00001Woo00@007ooOol0009oo`80029oo`03001oo`00009oo`05001oogooOol0
0004Ool01@00Oomoogoo00000goo00<007ooOol00goo00<007ooOol04Goo00<007ooOol027oo00<0
07ooOol00Woo00<007ooOol05goo00D007ooOomoo`0000Eoo`03001oogoo00800004Ool00000000>
Ool100000003001oogoo01Yoo`03001oogoo01Ioo`03001oogoo01ioo`03001oo`00011oo`03001o
ogoo00]oo`03001oogoo00Moo`03001oogoo00Moo`03001oogoo01eoo`03001oogoo01Ioo`03001o
ogoo01eoo`04001oogoo000@Ool00`00Oomoo`0:Ool00`00Oomoo`07Ool00`00Oomoo`08Ool00`00
Oomoo`0LOol00`00Oomoo`0FOol00`00Oomoo`0AOol100000003001oogoo01=oo`03001oogoo00ao
o`03001oogoo01Aoo`03001oogoo01Moo`8000Aoo`03001oogoo00Qoo`03001oogoo019oo`03001o
ogoo02Eoo`03001oogoo00eoo`03001oogoo01=oo`03001oogoo01Moo`<000=oo`03001oogoo00Qo
o`03001oogoo019oo`03001oogoo02Eoo`03001oogoo00eoo`03001oogoo01=oo`03001oogoo00ao
o`40000000<007ooOol027oo00<007ooOol05Woo00<007ooOol02Woo00<007ooOol00Woo0P000Woo
00<007ooOol067oo0P0000=oo`00Ool017oo00<007ooOol00Woo00<007ooOol027oo00<007ooOol0
1Goo00<007ooOol08Goo00<007ooOol05Goo00<007ooOol02Woo00<007ooOol00Woo0P000Woo00<0
07ooOol06Goo100017oo00<007ooOol00Woo00<007ooOol027oo00<007ooOol01Goo00<007ooOol0
8Goo00<007ooOol05Goo00<007ooOol02goo00<007ooOol00Woo0P000Woo00<007ooOol03Goo0@00
00000`00Ool0000;Ool00`00Oomoo`0?Ool00`00Oomoo`05Ool00`00Oomoo`06Ool00`00Oomoo`08
Ool00`00Oomoo`0FOol00`00Oomoo`0;Ool00`00Oomoo`0JOol00`00Oomoo`06Ool00`00Oomoo`04
Ool00`00Oomoo`09Ool00`00Oomoo`0?Ool00`00Oomoo`05Ool00`00Oomoo`06Ool00`00Oomoo`09
Ool00`00Oomoo`0EOol00`00Oomoo`0;Ool00`00Oomoo`0KOol00`00Oomoo`06Ool00`00Oomoo`04
Ool00`00Oomoo`09Ool00`00Oomoo`0>Ool00`00Oomoo`06Ool00`00Oomoo`05Ool00`00Oomoo`09
Ool00`00Oomoo`0;Ool100000003001oogoo00Aoo`03001oogoo015oo`03001oogoo00ioo`8000Io
o`03001oogoo029oo`03001oogoo00=oo`03001oogoo01]oo`03001oogoo01Eoo`03001oogoo019o
o`03001oogoo00ioo`03001oogoo00Aoo`03001oogoo029oo`8000Aoo`03001oogoo01aoo`03001o
ogoo01Eoo`03001oogoo015oo`03001oogoo00ioo`8000Ioo`03001oogoo01=oo`40000000<007oo
Ool057oo00@007ooOol000Moo`03001oogoo00Aoo`03001oogoo00Uoo`04001oogoo0002Ool4000E
Ool00`00Ool00003Ool20003Ool00`00Oomoo`04Ool00`00Oomoo`0KOol00`00Oomoo`0OOol00`00
Ool00007Ool00`00Oomoo`05Ool00`00Oomoo`09Ool00`00Ool00002Ool4000EOol00`00Ool00003
Ool00`00Ool00002Ool00`00Oomoo`05Ool00`00Oomoo`0KOol00`00Oomoo`0NOol01000Oomoo`00
1Woo00<007ooOol01Goo00<007ooOol02Goo00<007oo00000Woo1@002goo0@0000000`00Oomoo`0]
Ool00`00Oomoo`08Ool00`00Oomoo`0EOol00`00Oomoo`0BOol01@00Oomoogoo00000goo0P004goo
00<007ooOol0=goo00<007ooOol01goo00<007ooOol05Woo00<007ooOol04Goo00D007ooOomoo`00
00=oo`03001oo`0001=oo`03001oogoo03Ioo`03001oogoo00Moo`03001oogoo00eoo`40000000@0
07ooOol0009oo`03001oogoo00Uoo`8002Ioo`03001oogoo009oo`03001oogoo01Ioo`03001oogoo
00moo`03001oogoo00Moo`03001oogoo01Yoo`05001oogooOol0000;Ool00`00Oomoo`0TOol00`00
Oomoo`02Ool00`00Oomoo`0FOol00`00Oomoo`0?Ool00`00Oomoo`08Ool00`00Oomoo`0JOol01000
Oomoo`002goo0P009Woo00D007ooOomoo`0000]oo`40000000<007ooOol06goo00<007ooOol01Goo
00<007ooOol027oo00D007ooOomoo`0000Eoo`03001oogoo0080019oo`@000Aoo`05001oogooOol0
0004Ool00`00Oomoo`0BOol2000?Ool00`00Oomoo`0POol00`00Oomoo`05Ool00`00Oomoo`09Ool0
1000Oomoo`001Woo00D007oo001oo`00019oo`<000Eoo`05001oogooOol00004Ool00`00Oomoo`0B
Ool2000>Ool00`00Oomoo`0QOol00`00Oomoo`05Ool00`00Oomoo`08Ool01000Oomoo`001Woo00D0
07oo001oo`0000Yoo`40000000<007ooOol037oo00<007ooOol05Goo00<007ooOol03goo00<007oo
Ool06Goo00<007ooOol00goo00D007ooOomoo`00009oo`03001oogoo03Qoo`03001oogoo01Ioo`03
001oogoo00moo`03001oogoo01Qoo`03001oogoo00=oo`03001oogoo009oo`03001oo`0003Yoo`03
001oogoo01Ioo`03001oogoo00moo`03001oogoo00ioo`40000000<007ooOol01goo00<007ooOol0
5Woo00<007ooOol01Goo00@007ooOol000Uoo`03001oogoo00=oo`03001oo`00015oo`03001oogoo
02aoo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo01Eoo`03001oogoo00Ioo`03001oo`00
00Yoo`03001oogoo00=oo`80015oo`03001oogoo02aoo`03001oogoo00Uoo`03001oogoo00Uoo`03
001oogoo01Ioo`03001oogoo00Eoo`03001oo`0000Yoo`03001oogoo00=oo`03001oo`0000Uoo`40
000000<007ooOol027oo00<007ooOol027oo00<007ooOol01Woo00<007ooOol03Woo00<007ooOol0
1Woo00<007ooOol017oo00<007ooOol03Woo0P0000=oo`00Ool00P000goo00<007ooOol027oo00<0
07ooOol017oo00<007ooOol01Goo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol02Goo00<0
07ooOol02Goo00<007ooOol01Goo00<007ooOol03Woo00<007ooOol01Woo00<007ooOol01Goo00<0
07ooOol03Goo0P000Woo0P0000Eoo`00Oomoo`0000Yoo`03001oogoo00Eoo`03001oogoo00Aoo`03
001oogoo00Eoo`03001oogoo00eoo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00Ioo`03
001oogoo00eoo`03001oogoo00Ioo`03001oogoo00Eoo`03001oogoo00Moo`40000000<007ooOol0
67oo00<007ooOol017oo00<007ooOol01Goo00<007ooOol01goo00<007ooOol01Goo00<007ooOol0
0`003Woo00<007ooOol00goo00<007ooOol0D7oo00<007ooOol01Goo00<007ooOol017oo00<007oo
Ool01goo00<007ooOol01Goo00<007ooOol00`003goo00<007ooOol00goo00<007ooOol0Cgoo00<0
07ooOol01Goo00<007ooOol01Goo00<007ooOol01goo00<007ooOol01Goo00<007ooOol00`0027oo
0@0000000`00Oomoo`0bOol00`00Oomoo`06Ool2000AOol00`00Oomoo`06Ool00`00Oomoo`07Ool0
1@00Oomoogoo00000goo00<007ooOol04Woo00<007ooOol0@7oo00<007ooOol01Woo00<007ooOol0
47oo00<007ooOol01Woo00<007ooOol01Woo00<007ooOol00Woo00D007ooOomoo`0001=oo`03001o
ogoo041oo`03001oogoo00Ioo`8000Yoo`40000000@007ooOol0009oo`03001oogoo00Ioo`03001o
ogoo00Aoo`03001oogoo009oo`03001oogoo01Aoo`03001oogoo00=oo`8000Aoo`03001oogoo0080
00eoo`05001oogooOol00002Ool00`00Oomoo`02Ool01@00Ool007oo000027oo00<007ooOol01Goo
00<007ooOol02Goo00<007ooOol02Goo00<007ooOol01Goo00D007ooOomoo`0000Moo`03001oogoo
00Eoo`05001oogooOol0000FOol00`00Oomoo`03Ool20004Ool00`00Oomoo`02000=Ool01@00Oomo
ogoo00000Woo00<007ooOol00Woo00D007oo001oo`0000Qoo`03001oogoo00Eoo`03001oogoo00Uo
o`03001oogoo00Uoo`03001oogoo00Eoo`05001oogooOol00007Ool00`00Oomoo`05Ool01@00Oomo
ogoo00005goo00<007ooOol00goo0P000goo00@007ooOomoo`8000Moo`40000000<007ooOol0;7oo
00<007ooOol02goo00<007oo00003goo0`003goo00<007ooOol03Goo00<007ooOol04Goo00<007oo
Ool0=Goo00<007ooOol02goo00@007ooOol000moo`80011oo`03001oogoo00aoo`03001oogoo015o
o`03001oogoo03Ioo`03001oogoo00Yoo`800003Ool007oo00Uoo`40000000<007ooOol097oo00D0
07ooOomoo`00009oo`03001oogoo009oo`03001oogoo00aoo`8000]oo`05001oo`00Ool0000QOol0
0`00Oomoo`0mOol01@00Oomoogoo00000goo00<007ooOol00Woo00<007ooOol02goo0P002goo0P00
00=oo`0000008Goo00<007ooOol0?Goo00<007ooOol00Woo00@007ooOol000Aoo`03001oogoo00]o
o`8000Moo`40000000<007oo000017oo00<007ooOol027oo0P002goo00@007ooOol001=oo`80009o
o`03001oo`0000=oo`04001oo`00000:Ool300000goo00000002Ool20002Ool00`00Oomoo`08Ool0
0`00Oomoo`0AOol00`00Oomoo`09Ool00`00Oomoo`09Ool01@00Oomoogoo00002goo00<007ooOol0
2Goo00D007ooOomoo`00019oo`05001oo`00Ool00002Ool01000Oomoo`000Woo0P002Woo00<007oo
00000P000Woo00D007oo001oo`0000Yoo`03001oogoo015oo`03001oogoo00Uoo`03001oogoo00Uo
o`03001oogoo009oo`03001oogoo00Qoo`8000]oo`04001oogoo000COol200001Goo001oogoo0000
0goo00@007oo000000Ioo`40000000<007ooOol0>Woo00<007ooOol00Woo00<007ooOol02Goo00<0
07ooOol02goo0P000goo00D007ooOomoo`0000Moo`03001oogoo01Uoo`03001oogoo041oo`03001o
ogoo009oo`03001oogoo00Qoo`03001oogoo00aoo`05001oogooOol00003Ool00`00Oomoo`05Ool0
0`00Oomoo`0JOol00`00Oomoo`0oOol00`00Oomoo`02Ool00`00Oomoo`05Ool100000003001oogoo
00]oo`03001oogoo00eoo`03001oogoo00Yoo`03001oogoo00Aoo`03001oogoo00Moo`03001oogoo
009oo`@000Uoo`04001oogooOol2000EOol00`00Oomoo`0`Ool00`00Oomoo`0>Ool00`00Oomoo`09
Ool00`00Oomoo`05Ool00`00Oomoo`06Ool00`00Oomoo`03Ool30009Ool00`00Oomoo`02Ool2000D
Ool00`00Oomoo`0aOol00`00Oomoo`0=Ool00`00Oomoo`09Ool00`00Oomoo`05Ool00`00Oomoo`06
Ool00`00Oomoo`03Ool40005Ool100000003001oogoo00Qoo`03001oogoo00Uoo`03001oogoo00Uo
o`03001oogoo00eoo`05001oogooOol00008Ool2000017oo000000001Woo0P000Woo00<007ooOol0
1Goo00<007ooOol01Woo00<007ooOol02Goo00<007ooOol017oo00<007ooOol01Woo00@007ooOol0
00]oo`03001oogoo00Yoo`03001oogoo00Uoo`03001oogoo00Uoo`03001oogoo00eoo`05001oogoo
Ool00009Ool01@00Ool007oo00001goo0P0000=oo`00Ool01goo00<007ooOol01Goo00<007ooOol0
2Woo00<007ooOol017oo00<007ooOol01Woo00@007ooOol000]oo`03001oogoo00Yoo`03001oogoo
00Uoo`03001oogoo00Uoo`03001oogoo00aoo`03001oogoo009oo`03001oogoo00Ioo`03001oogoo
00<000Aoo`40000000<007ooOol05goo00<007ooOol04Goo00<007ooOol02Goo0P000Woo00<007oo
00000goo00<007ooOol01goo0P000Woo00@007oo000000=oo`03001oogoo00aoo`03001oogoo029o
o`03001oogoo01Qoo`03001oogoo019oo`03001oogoo00Uoo`06001oogoo001oo`000goo00<007oo
Ool01goo0P000Woo00D007oo001oo`00009oo`03001oogoo00eoo`03001oogoo029oo`03001oogoo
01Qoo`03001oogoo015oo`03001oogoo00Uoo`80009oo`03001oo`00009oo`03001oogoo00Aoo`40
000000<007ooOol01goo00<007ooOol07Goo00<007ooOol00Woo00<007ooOol01Goo00<007ooOol0
1goo00@007ooOomoo`8000Eoo`<0029oo`03001oogoo02Aoo`03001oogoo01ioo`05001oogooOol0
0007Ool00`00Oomoo`08Ool01@00Oomoogoo00001Goo10008Woo00<007ooOol097oo00<007ooOol0
7Goo00<007ooOol00Woo00<007ooOol017oo00<007ooOol027oo00D007ooOomoo`0000Aoo`400000
00<007ooOol037oo00<007ooOol017oo00<007ooOol00Woo00<007ooOol027oo0P005Woo0P0017oo
00<007oo000017oo0P0000Aoo`00000000Aoo`05001oogooOol00002Ool20003Ool01000Oomoo`00
2goo00<007ooOol02Goo00<007ooOol00Woo00<007ooOol06Woo00<007ooOol00goo00<007ooOol0
0goo00<007ooOol01goo00<007oo00005Woo0P0017oo0`000goo0P000Woo0P0017oo00D007ooOomo
o`0000=oo`03001oogoo009oo`04001oogoo000;Ool00`00Oomoo`09Ool00`00Oomoo`02Ool00`00
Oomoo`0IOol00`00Oomoo`04Ool00`00Oomoo`02Ool00`00Oomoo`08Ool2000FOol20004Ool00`00
Ool00003Ool100000004001oogoo000_Ool00`00Oomoo`04Ool20006Ool00`00Ool000030003Ool0
0`00Oomoo`03Ool01@00Oomoogoo00003goo00<007ooOol03Woo00<007ooOol03goo00<007ooOol0
1goo00<007ooOol0;7oo00<007ooOol01Goo00<007ooOol017oo00<007oo00000`000goo00<007oo
Ool017oo00@007ooOol0011oo`03001oogoo00eoo`03001oogoo00moo`03001oogoo00Moo`03001o
ogoo02eoo`03001oogoo00Aoo`8000Ioo`03001oo`0000<0009oo`40000000<007ooOol00goo00<0
07ooOol0;Goo00<007ooOol01Woo00T007ooOomoo`00Ool007oo00000Woo0P000Woo00<007oo0000
1goo00<007ooOol0>Goo00<007ooOol0;Woo00<007ooOol01Goo00<007ooOol00Woo0P0000=oo`00
Ool00Woo0P0000Eoo`00Oomoo`0000Moo`03001oogoo03Qoo`03001oogoo02ioo`03001oogoo00Eo
o`03001oogoo009oo`04001oo`000003Ool100000003001oogoo00moo`8000]oo`04001oogoo0008
Ool01000Oomoo`000goo00<007ooOol00Woo00<007ooOol01Goo00D007ooOomoo`00009oo`<00003
Ool00000008000=oo`03001oogoo008000=oo`03001oogoo00Aoo`03001oogoo00Moo`03001oo`00
00]oo`03001oogoo00]oo`03001oogoo01Moo`03001oo`0000]oo`03001oo`0000Uoo`03001oo`00
00=oo`03001oogoo00=oo`03001oogoo00Aoo`05001oogooOol00002Ool01000Ool007oo10000goo
00H007ooOol007oo0003Ool00`00Oomoo`03Ool00`00Oomoo`07Ool00`00Ool0000;Ool00`00Oomo
o`0;Ool00`00Oomoo`0GOol00`00Ool0000;Ool01000Oomoo`0027oo00@007ooOol000=oo`03001o
ogoo009oo`03001oogoo00Eoo`03001oogoo00800004Ool007oo0002Ool100000003001oogoo03]o
o`04001oogooOol20002Ool500000goo0000000@Ool00`00Oomoo`0IOol00`00Oomoo`07Ool00`00
Oomoo`15Ool00`00Oomoo`02Ool20002Ool40002Ool00`00Oomoo`0>Ool00`00Oomoo`0IOol00`00
Oomoo`07Ool00`00Oomoo`16Ool01000Oomoogoo0P000Woo0`000Goo0@0000000`00Ool00004Ool0
0`00Oomoo`0DOol00`00Oomoo`03Ool00`00Oomoo`03Ool00`00Oomoo`03Ool00`00Oomoo`03Ool0
1`00Oomoogoo001oo`0000=oo`80009oo`03001oo`0000L0009oo`03001oo`0000=oo`03001oo`00
00=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo
01=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo01=oo`03001oogoo00=oo`03001oogoo
00=oo`03001oogoo00=oo`03001oogoo00=oo`07001oogooOol007oo00000goo00H007oo001oo`00
Ool700001Woo001oo`00Ool000=oo`03001oo`0000=oo`03001oogoo00=oo`03001oogoo00=oo`03
001oogoo00=oo`03001oogoo00=oo`03001oogoo01=oo`03001oogoo00=oo`03001oogoo00=oo`03
001oogoo01=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo00=oo`03001oogoo00=oo`07
001oogooOol007oo00000goo00<007oo00000Woo1@000Goo0@0000000`00Oomoo`0:Ool00`00Oomo
o`08Ool00`00Oomoo`0IOol00`00Oomoo`0>Ool00`00Ool000040002Ool20003Ool00`00Oomoo`16
Ool00`00Oomoo`07Ool00`00Oomoo`0IOol00`00Oomoo`0>Ool01@00Oomoo`00Ool00`0000Aoo`00
Ool000=oo`03001oogoo04Eoo`03001oogoo00Moo`03001oogoo01Yoo`03001oogoo00eoo`800004
Ool00000Ool100000003001oogoo01Ioo`03001oogoo00aoo`03001oogoo00eoo`03001oogoo009o
o`04001oogoo0003Ool500000goo001oo`06Ool00`00Oomoo`05Ool00`00Oomoo`07Ool00`00Oomo
o`0GOol00`00Ool0000WOol00`00Oomoo`0;Ool00`00Oomoo`0=Ool00`00Oomoo`03Ool01000Oomo
o`000goo10000Woo00<007ooOol01Goo00<007ooOol017oo00<007ooOol01goo00<007ooOol05goo
00<007oo00009goo00<007ooOol037oo00<007ooOol03Goo00<007ooOol00Woo00@007ooOol000=o
o`@0009oo`40000000<007ooOol02Goo00<007ooOol05Woo00<007ooOol02Goo00<007ooOol01Goo
00@007ooOol000Aoo`05001oogoo001oo`020002Ool300000goo00000003Ool00`00Oomoo`05Ool0
1@00Oomoogoo00001Woo00<007ooOol01goo00<007oo00006Goo00<007ooOol02goo00<007ooOol0
5goo00<007ooOol02Goo00<007ooOol017oo00D007ooOomoo`0000=oo`05001oogoo001oo`030000
0goo000000040003Ool00`00Oomoo`06Ool01@00Oomoogoo00001Goo00<007ooOol01goo00<007oo
00006Goo00<007ooOol02goo00<007ooOol05goo00<007ooOol02Goo00<007ooOol01Goo00@007oo
Ool000Aoo`05001oogoo001oo`020002Ool100000003001oogoo00Ioo`03001oogoo011oo`03001o
ogoo00eoo`03001oogoo00ioo`04001oogoo0006Ool00`00Oomoo`0400000goo001oo`05Ool20006
Ool00`00Oomoo`0]Ool00`00Oomoo`07Ool00`00Oomoo`0?Ool00`00Oomoo`0=Ool00`00Oomoo`0>
Ool01@00Oomoogoo00001Goo00@007ooOomoo`@00003Ool007oo00Eoo`03001oogoo00Eoo`03001o
ogoo02aoo`03001oogoo00Moo`03001oogoo011oo`03001oogoo00eoo`03001oogoo00eoo`05001o
ogooOol00006Ool00`00Oomoo`0100000003001oogoo019oo`03001oogoo00Uoo`03001oogoo00Uo
o`04001oogoo0004Ool00`00Oomoo`05Ool00`00Oomoo`02Ool200000goo00000004Ool00`00Ool0
0005Ool00`00Oomoo`0DOol2000:Ool00`00Oomoo`03Ool00`00Oomoo`03Ool00`00Oomoo`0OOol0
0`00Oomoo`09Ool00`00Oomoo`09Ool01000Oomoo`000goo00<007ooOol01Goo00<007ooOol00Woo
00@007oo001oo`8000=oo`<000Eoo`03001oogoo01Eoo`8000Yoo`03001oogoo009oo`03001oogoo
00Aoo`03001oogoo01ioo`03001oogoo00Uoo`03001oogoo00Uoo`04001oogoo0004Ool00`00Oomo
o`05Ool02000Oomoogoo001oo`00Ool20003Ool100000003001oogoo00eoo`03001oogoo00eoo`03
001oogoo015oo`03001oogoo009oo`03001oogoo00Ioo`D000Eoo`07001oogooOol007oo000027oo
00<007ooOol017oo00<007ooOol00Woo00<007ooOol07Goo00<007ooOol057oo00<007ooOol03Goo
00<007ooOol04Woo00D007ooOomoo`0000Uoo`04001oo`000005Ool01@00Oomoogoo00000Woo00<0
07ooOol01Goo00<007ooOol01Goo00<007ooOol00Woo00<007ooOol07Goo00<007ooOol057oo00<0
07ooOol03Goo00<007ooOol04Goo00D007ooOomoo`0000Uoo`D000=oo`40000000D007ooOomoo`00
02Aoo`03001oogoo00aoo`04001oogooOol20002Ool01@00Ool007oo00002Goo00P007ooOomoo`00
Ool007oo0P002goo00<007ooOol04Goo00<007ooOol067oo00<007ooOol08Woo00<007ooOol03Goo
00P007ooOol007oo001oo`000Woo0P002Goo00L007ooOomoo`00Ool00002Ool2000:Ool00`00Oomo
o`0BOol00`00Oomoo`0HOol00`00Oomoo`0ROol00`00Oomoo`0<Ool01000Oomoogoo0P0000Aoo`00
Oomoo`8000Ioo`40000000<007ooOol00Woo00<007ooOol02Goo00@007ooOol000Qoo`03001oogoo
00Aoo`03001oogoo00Yoo`03001oogoo011oo`8000Ioo`8000=oo`03001oogoo00Moo`05001oogoo
Ool0000>Ool00`00Oomoo`0EOol00`00Oomoo`0:Ool00`00Oomoo`09Ool01000Oomoo`0027oo00<0
07ooOol01Goo00<007ooOol02Goo00<007ooOol04Goo00<007ooOol01Goo00D007ooOomoo`0000Uo
o`05001oogooOol0000?Ool00`00Oomoo`0EOol00`00Oomoo`0:Ool00`00Oomoo`09Ool01000Oomo
o`0027oo00<007ooOol017oo00<007ooOol02Goo00<007ooOol04Goo0P0017oo0@0000000`00Oomo
o`0UOol00`00Oomoo`0>Ool00`00Oomoo`02Ool20002Ool30008Ool200000goo00000004Ool00`00
Oomoo`06Ool00`00Oomoo`05Ool00`00Oomoo`09Ool00`00Oomoo`04Ool00`00Oomoo`06Ool00`00
Oomoo`0aOol00`00Oomoo`0>Ool01000Oomoogoo0P000Woo00@007oo000000Moo`D000Aoo`03001o
ogoo00Ioo`03001oogoo00Eoo`03001oogoo00Uoo`03001oogoo00Eoo`03001oogoo00Ioo`03001o
ogoo031oo`03001oogoo00ioo`03001oogoo009oo`80009oo`<000Eoo`40000000<007ooOol01Goo
00<007ooOol06Woo00<007ooOol01Goo00D007ooOomoo`00009oo`8000ioo`03001oogoo00Qoo`03
001oogoo009oo`03001oogoo03moo`03001oogoo01Yoo`03001oogoo00Eoo`05001oogooOol00003
Ool00`00Oomoo`0<Ool00`00Oomoo`08Ool00`00Oomoo`03Ool00`00Oomoo`0oOol00`00Oomoo`0I
Ool00`00Oomoo`05Ool00`00Oomoo`02Ool00`00Oomoo`02000>Ool00`00Oomoo`03Ool100000003
001oogoo01Eoo`03001oogoo015oo`03001oogoo00Uoo`04001oogoo0003Ool01@00Ool007oo0000
2Woo0P0000=oo`00Ool01Goo00@007oo000001=oo`05001oogooOol0000GOol01@00Oomoogoo0000
5goo00<007ooOol04Goo00<007ooOol027oo00D007ooOomoo`00009oo`06001oogoo001oo`002Goo
0P000Woo00<007ooOol00goo00D007oo001oo`00019oo`05001oogooOol0000GOol00`00Oomoo`02
Ool00`00Oomoo`0DOol00`00Oomoo`0AOol00`00Oomoo`09Ool01000Oomoo`000Woo00H007ooOol0
07oo0006Ool100000003001oogoo00Yoo`03001oogoo00Uoo`04001oogoo000MOol01000Oomoogoo
0`0000=oo`0000002goo0P0017oo00<007ooOol01Woo00<007ooOol00Woo00D007ooOomoo`0000=o
o`03001oogoo015oo`03001oogoo01Uoo`03001oogoo00Uoo`05001oogooOol0000LOol00`00Oomo
o`02Ool200000goo0000000;Ool20004Ool00`00Oomoo`06Ool00`00Oomoo`02Ool01@00Oomoogoo
00000goo00<007ooOol04Goo00<007ooOol06Woo00<007ooOol02Goo00@007ooOol001eoo`04001o
ogooOol300000goo00000006Ool100000003001oogoo00Uoo`03001oogoo019oo`03001oogoo00ao
o`03001oogoo00ioo`8000moo`04001oo`00000<Ool00`00Oomoo`0fOol00`00Oomoo`0AOol00`00
Oomoo`0<Ool00`00Oomoo`0>Ool3000>Ool01000Oomoo`003Goo00<007ooOol0=Goo00<007ooOol0
4Goo00<007ooOol03Goo00<007ooOol03Goo0`001goo0@0000000`00Oomoo`0NOol00`00Oomoo`04
Ool00`00Oomoo`0;Ool00`00Oomoo`02Ool00`00Oomoo`0AOol00`00Oomoo`0YOol00`00Oomoo`0=
Ool00`00Oomoo`0QOol00`00Oomoo`05Ool00`00Oomoo`0:Ool00`00Oomoo`02Ool00`00Oomoo`0B
Ool00`00Oomoo`0XOol00`00Oomoo`0=Ool00`00Oomoo`0QOol00`00Oomoo`05Ool00`00Oomoo`0:
Ool00`00Oomoo`02Ool00`00Oomoo`0=Ool100000003001oogoo00Ioo`03001oogoo00Eoo`05001o
ogooOol0000?Ool01@00Oomoogoo00000goo00D007ooOomoo`00009oo`04001oogoo0006Ool01@00
Ool007oo00003Goo00<007ooOol00P000goo0P000Woo00D007ooOomoo`0001Ioo`03001oogoo00Uo
o`03001oogoo00Uoo`03001oogoo00Eoo`03001oogoo00Eoo`05001oogooOol0000?Ool01@00Oomo
ogoo000017oo00L007ooOomoo`00Ool00003Ool00`00Oomoo`03Ool01@00Ool007oo00003Goo0P00
00=oo`00000017oo0P0000=oo`00Ool00Woo00<007ooOol057oo00<007ooOol02Goo00<007ooOol0
2Goo00<007ooOol01Goo00<007ooOol01Goo00D007ooOomoo`00011oo`05001oogooOol00003Ool0
1`00Oomoogoo001oo`0000=oo`03001oogoo00=oo`03001oogoo008000Qoo`40000000<007ooOol0
>7oo00<007ooOol00goo00<007ooOol03Goo0`0000=oo`00Ool01Woo00<007ooOol01goo00<007oo
Ool01Goo00<007ooOol01Goo00<007ooOol0D7oo00<007ooOol00Woo00<007ooOol03Goo0`000Woo
00<007ooOol01Goo00<007ooOol01goo00<007ooOol017oo00<007ooOol01Goo00<007ooOol0D7oo
00<007ooOol00goo00<007ooOol01Woo0@0000000`00Oomoo`0IOol00`00Oomoo`0=Ool00`00Oomo
o`07Ool00`00Oomoo`0HOol00`00Oomoo`05Ool00`00Oomoo`0[Ool00`00Oomoo`0QOol00`00Oomo
o`0<Ool00`00Oomoo`08Ool00`00Oomoo`0GOol00`00Oomoo`05Ool00`00Oomoo`0[Ool00`00Oomo
o`0QOol00`00Oomoo`0=Ool00`00Oomoo`07Ool00`00Oomoo`0AOol100000004001oogooOol2000:
Ool00`00Oomoo`02Ool00`00Oomoo`0<Ool00`00Oomoo`0COol01@00Ool00000Ool00P004Goo0P00
1Goo00<007ooOol01Woo00D007oo001oo`0000Moo`05001oogooOol00008Ool00`00Oomoo`09Ool0
0`00Oomoo`09Ool00`00Oomoo`09Ool01@00Oomoogoo00003Woo00<007ooOol04goo00@007oo0000
009oo`80011oo`03001oo`0000Aoo`03001oogoo00Ioo`05001oo`00Ool00008Ool01@00Oomoogoo
00001goo00<007ooOol02Goo00<007ooOol02Goo00<007ooOol02Goo00D007ooOomoo`0000moo`03
001oogoo01=oo`05001oo`00001oo`020009Ool100000003001oogoo02aoo`04001oogoo0004Ool0
0`00Oomoo`02Ool00`00Oomoo`0IOol00`00Oomoo`0?Ool00`00Oomoo`0FOol00`00Oomoo`0hOol0
0`00Ool00004Ool00`00Oomoo`03Ool00`00Oomoo`0HOol00`00Oomoo`0?Ool00`00Oomoo`0FOol0
0`00Oomoo`0hOol01000Oomoo`000goo00<007ooOol00goo00<007ooOol02goo0@0000000`00Oomo
o`0FOol00`00Oomoo`0GOol01@00Oomoogoo00001Goo0`004Woo00D007ooOomoo`0000Ioo`04001o
ogoo000:Ool00`00Oomoo`05Ool00`00Oomoo`0aOol00`00Oomoo`0GOol01@00Oomoogoo00001Goo
0`004goo00D007ooOomoo`0000Eoo`05001oogooOol0000:Ool00`00Oomoo`05Ool00`00Oomoo`0a
Ool00`00Oomoo`0GOol01@00Oomoogoo00001Goo0`002Woo0@0000000`00Oomoo`06Ool00`00Oomo
o`0<Ool00`00Oomoo`02Ool00`00Oomoo`08Ool00`00Oomoo`03Ool00`00Oomoo`08Ool01000Oomo
o`005Woo0P000Woo00<007ooOol097oo0P002goo00D007ooOomoo`0000Moo`03001oogoo00eoo`05
001oogooOol0000:Ool00`00Oomoo`03Ool00`00Oomoo`09Ool01000Oomoo`005Goo0P000goo00<0
07ooOol097oo0P002goo00@007ooOol000Qoo`03001oogoo00aoo`03001oogoo009oo`03001oogoo
00Moo`03001oogoo00Aoo`03001oogoo00Qoo`04001oogoo000=Ool100000003001oogoo00Uoo`03
001oogoo015oo`8000Aoo`05001oogooOol0000COol00`00Oomoo`0FOol00`00Oomoo`07Ool00`00
Oomoo`0gOol00`00Oomoo`0@Ool00`00Ool00003Ool01@00Oomoogoo00004goo00<007ooOol05Woo
00<007ooOol01goo00<007ooOol0=goo00<007ooOol04Goo0P000goo00D007ooOomoo`0001Aoo`03
001oogoo00Uoo`40000000<007ooOol02Woo00<007ooOol08goo00<007ooOol00Woo00<007ooOol0
6Goo0P0000Eoo`00Oomoo`0000eoo`03001oogoo031oo`03001oogoo02=oo`03001oogoo009oo`03
001oogoo01Uoo`800005Ool007ooOol0000=Ool00`00Oomoo`0aOol00`00Oomoo`0SOol01@00Oomo
ogoo00004Woo0@0000000`00Oomoo`0AOol00`00Oomoo`0DOol00`00Oomoo`04Ool01P00Oomoogoo
001oo`80009oo`8001Ioo`<000Eoo`8000Ioo`8000Yoo`03001oogoo00=oo`80009oo`03001oogoo
00eoo`03001oogoo01Eoo`03001oogoo01Eoo`03001oogoo00Aoo`04001oogoo0002Ool01P00Oomo
o`00Ool001Ioo`8000Ioo`8000Ioo`8000Uoo`03001oogoo00=oo`03001oo`00009oo`03001oogoo
00eoo`03001oogoo01Eoo`03001oogoo01Aoo`03001oogoo00Aoo`06001oogooOol007oo0P0000Ao
o`00Ool000aoo`40000000<007ooOol01Goo00<007ooOol01Woo00<007ooOol06goo00<007ooOol0
2goo00<007ooOol05Goo00<007ooOol02Goo00<007ooOol01Goo00<007ooOol01Woo00<007ooOol0
3Woo00<007ooOol02Goo00<007ooOol017oo00<007ooOol01goo00<007ooOol06Woo00<007ooOol0
2goo00<007ooOol05Woo00<007ooOol027oo00<007ooOol01Woo00<007ooOol01Goo00<007ooOol0
3goo00<007ooOol02Goo00<007ooOol017oo00<007ooOol01Woo00<007ooOol06goo00<007ooOol0
2Woo00<007ooOol02goo0@0000000`00Oomoo`0IOol00`00Oomoo`04Ool00`00Oomoo`09Ool00`00
Oomoo`07Ool01000Ool000006Woo00<007ooOol00Woo00<007ooOol03goo00<007ooOol0>Goo00<0
07ooOol01Goo00<007ooOol027oo00<007ooOol01goo00@007oo000001]oo`05001oogooOol0000A
Ool00`00Oomoo`0iOol00`00Oomoo`05Ool00`00Oomoo`08Ool00`00Oomoo`08Ool01000Ool00000
3Goo0@0000000`00Oomoo`0BOol00`00Oomoo`0BOol00`00Oomoo`06Ool00`00Ool00004Ool2000J
Ool01@00Oomoogoo00001Goo00<007ooOol02Goo00<007ooOol03Goo00<007ooOol01goo00<007oo
Ool06goo00<007ooOol04Woo00<007ooOol01Woo00<007oo00000goo0`006Goo00<007ooOol00Woo
00<007ooOol00goo00<007ooOol02Goo00<007ooOol037oo00<007ooOol027oo00<007ooOol06Woo
00<007ooOol04Woo00<007ooOol01Woo00<007oo00000goo0`003Woo0@0000000`00Oomoo`02Ool0
0`00Oomoo`08Ool00`00Oomoo`07Ool00`00Oomoo`0:Ool00`00Oomoo`0>Ool00`00Ool0000POol0
0`00Oomoo`0FOol00`00Oomoo`0MOol00`00Oomoo`07Ool00`00Oomoo`07Ool00`00Oomoo`0:Ool0
0`00Oomoo`0?Ool00`00Ool0000OOol00`00Oomoo`0FOol00`00Oomoo`0MOol00`00Oomoo`07Ool0
0`00Oomoo`07Ool00`00Oomoo`0;Ool00`00Oomoo`0>Ool00`00Ool0000?Ool100000005001oogoo
Ool0000IOol01@00Oomoogoo00000goo00<007ooOol0;Goo00<007oo00001Goo00<007ooOol017oo
00<007ooOol06Woo00<007ooOol00goo00<007ooOol01goo00<007ooOol05goo00D007ooOomoo`00
00Aoo`03001oogoo02eoo`8000Eoo`03001oogoo00Aoo`03001oogoo01Yoo`03001oogoo00=oo`03
001oogoo00Qoo`03001oogoo01Moo`05001oogooOol00003Ool00`00Oomoo`0QOol100000003001o
ogoo00Ioo`03001oogoo00]oo`03001oogoo00Eoo`03001oogoo00Uoo`03001oogoo0080009oo`03
001oo`0000Eoo`03001oogoo01Yoo`03001oogoo008000Aoo`8000=oo`05001oogooOol0000^Ool0
0`00Oomoo`0;Ool00`00Oomoo`05Ool00`00Oomoo`08Ool01000Oomoogoo0P0000Aoo`00Ool000Eo
o`03001oogoo01Yoo`03001oogoo008000Aoo`03001oo`00009oo`05001oogooOol0000^Ool00`00
Oomoo`0;Ool00`00Oomoo`05Ool00`00Oomoo`08Ool01000Oomoogoo0P0000Eoo`00Oomoo`0000Eo
o`03001oogoo00eoo`40000000<007ooOol0=Goo00<007ooOol07Goo0P003goo00<007ooOol04Woo
00<007oo0000AGoo00<007ooOol07Goo0P003goo00<007ooOol04Woo00<007oo0000AGoo00<007oo
Ool047oo0@0000000`00Oomoo`09Ool00`00Oomoo`0VOol01@00Oomoogoo0000=Woo00<007ooOol0
1goo00<007ooOol03goo00<007oo00002goo00<007ooOol09Woo00<007ooOol00Woo00<007ooOol0
=7oo00<007ooOol01Woo00<007ooOol03goo00<007oo00002goo00<007ooOol09Woo00<007ooOol0
0Woo00<007ooOol03Woo0@0000000`00Oomoo`0;Ool00`00Oomoo`0BOol00`00Oomoo`05Ool00`00
Oomoo`02Ool00`00Oomoo`02Ool01@00Oomoogoo00007goo00@007ooOol0009oo`03001oogoo009o
o`03001oogoo00Ioo`05001oogooOol00003Ool01000Oomoo`008goo00<007ooOol04Woo00<007oo
Ool01Goo00<007ooOol00Woo00<007ooOol00goo00@007ooOol0021oo`03001oo`00009oo`03001o
ogoo009oo`03001oogoo00Ioo`03001oogoo009oo`07001oogooOol007oo00008goo00<007ooOol0
4goo00<007ooOol01Goo00<007ooOol00Woo00<007ooOol00Woo00@007ooOol0015oo`40000000<0
07ooOol047oo00<007ooOol01Goo00<007ooOol04goo00<007ooOol00Woo0P00=Goo00<007ooOol0
6Woo00<007ooOol05Goo00<007ooOol017oo00<007ooOol04goo00<007ooOol00Woo0P00=Goo00<0
07ooOol06Woo00<007ooOol05Goo00<007ooOol01Goo00<007ooOol04goo00<007ooOol00Woo0P00
4Woo0@0000000`00Ool0000?Ool00`00Oomoo`0AOol2000?Ool2000YOol00`00Oomoo`0OOol00`00
Oomoo`09Ool00`00Oomoo`0=Ool00`00Oomoo`0BOol00`00Oomoo`0=Ool2000YOol00`00Oomoo`0O
Ool00`00Oomoo`09Ool00`00Oomoo`0>Ool00`00Oomoo`0AOol2000>Ool3000AOol100000003001o
ogoo00Aoo`03001oogoo00aoo`05001oogooOol00003Ool00`00Oomoo`02Ool00`00Oomoo`04Ool0
1000Oomoo`000Woo00L007ooOomoo`00Ool0000SOol01@00Ool007oo00001goo00@007oo000000]o
o`03001oogoo00Aoo`03001oogoo00Moo`03001oogoo00eoo`03001oogoo00]oo`03001oogoo009o
o`05001oogooOol00003Ool00`00Oomoo`04Ool01@00Oomoogoo00000Woo00@007ooOol0009oo`03
001oogoo021oo`06001oogoo001oo`001Woo00<007ooOol00P002Woo00<007ooOol01Goo00<007oo
Ool01Woo00<007ooOol03Goo00<007ooOol037oo00D007ooOomoo`0000=oo`05001oogooOol00007
Ool01000Oomoo`000Woo00L007ooOomoo`00Ool0000COol100000003001oogoo02Uoo`03001oogoo
02aoo`03001oogoo008000=oo`03001oogoo05Aoo`03001oogoo02aoo`06001oogoo001oo`000goo
00<007ooOol0Dgoo00<007ooOol077oo0@0000000`00Oomoo`03Ool00`00Oomoo`06Ool00`00Oomo
o`0=Ool00`00Oomoo`0iOol00`00Ool0000fOol00`00Oomoo`06Ool00`00Oomoo`0=Ool00`00Oomo
o`0jOol00`00Ool0000fOol00`00Oomoo`06Ool00`00Oomoo`0<Ool00`00Oomoo`0YOol100000004
001oogoo000COol00`00Oomoo`0FOol200001Goo001oogoo00009Goo0P003Woo00<007ooOol00Woo
00<007ooOol02Goo00D007ooOomoo`0000Ioo`03001oogoo00Moo`03001oogoo011oo`03001oogoo
01Moo`06001oogoo001oo`009Goo0P003goo00D007ooOomoo`0000aoo`05001oogooOol00006Ool0
0`00Oomoo`06Ool00`00Oomoo`0AOol00`00Oomoo`0FOol2000017oo001oo`0057oo0@0000000`00
Oomoo`07Ool00`00Oomoo`0EOol00`00Oomoo`02Ool00`00Oomoo`0;Ool00`00Oomoo`0[Ool00`00
Oomoo`0?Ool01@00Oomoogoo00007Woo00<007ooOol05Woo00<007ooOol00Woo00<007ooOol02Woo
00<007ooOol0:goo00<007ooOol03goo00D007ooOomoo`0001moo`03001oogoo01Eoo`03001oogoo
009oo`03001oogoo00]oo`03001oogoo019oo`40000000<007ooOol0:Goo0P0017oo00<007ooOol0
:goo00<007ooOol01Goo00<007ooOol03Woo00<007ooOol017oo00<007ooOol0>7oo00<007ooOol0
0Woo00<007ooOol0:goo00<007ooOol01Goo00<007ooOol03Woo00<007ooOol01Goo00<007ooOol0
=goo0P0017oo00<007ooOol05Goo0@0000000`00Oomoo`08Ool01000Oomoo`0057oo00@007ooOol0
009oo`03001oogoo00Ioo`<002Uoo`03001oogoo00Eoo`05001oogooOol00002Ool00`00Oomoo`0L
Ool2000;Ool01@00Oomoogoo000057oo00<007oo00000goo00<007ooOol01Woo0`00:7oo00<007oo
Ool01Woo00@007ooOol000=oo`03001oogoo01aoo`8000]oo`04001oogoo000DOol00`00Ool00003
Ool00`00Oomoo`06Ool3000EOol100000003001oogoo00moo`8000Ioo`03001oogoo009oo`03001o
ogoo00ioo`03001oogoo02moo`03001oogoo03aoo`03001oogoo00Eoo`05001oogooOol0000@Ool0
0`00Oomoo`0_Ool00`00Oomoo`0lOol00`00Oomoo`05Ool00`00Oomoo`02Ool00`00Oomoo`0>Ool0
0`00Oomoo`0GOol100000003001oogoo01Aoo`05001oogooOol0000AOol00`00Oomoo`0^Ool00`00
Oomoo`0;Ool00`00Oomoo`0JOol00`00Oomoo`0IOol01@00Oomoogoo00004Goo00<007ooOol0;goo
00<007ooOol02Woo00<007ooOol06Woo00<007ooOol06Goo00<007ooOol00Woo00<007ooOol03goo
00<007ooOol06Woo0@0000000`00Ool00004Ool00`00Oomoo`0DOol00`00Oomoo`07Ool01@00Oomo
ogoo00001Goo00<007ooOol0:Woo0`0000=oo`00Ool00Woo00<007ooOol02goo00<007ooOol01Goo
00<007ooOol01Goo00<007ooOol02Goo00D007ooOomoo`0001Ioo`03001oogoo00Moo`03001oogoo
009oo`03001oogoo009oo`03001oogoo02Yoo`06001oo`00001oo`000Woo00<007ooOol037oo00<0
07ooOol017oo00<007ooOol01Goo00<007ooOol02Goo00D007ooOomoo`0001Moo`03001oogoo00Mo
o`05001oogooOol00005Ool00`00Oomoo`0EOol100000003001oogoo05aoo`8000moo`03001oogoo
07]oo`03001oogoo00eoo`03001oogoo06Ioo`40000000<007ooOol02goo00<007ooOol07Goo0P00
00=oo`00Ool0=7oo00<007ooOol027oo00<007oo000037oo00<007ooOol06Goo00<007ooOol077oo
00D007oo001oo`0003Aoo`03001oogoo00Qoo`03001oo`0000aoo`03001oogoo01Uoo`03001oogoo
01aoo`05001oo`00Ool0000HOol100000003001oo`0000=oo`03001oogoo00eoo`03001oogoo00Uo
o`04001oo`000003Ool00`00Oomoo`02Ool01000Oomoo`00;goo00H007ooOol007oo0003Ool20002
Ool00`00Oomoo`07Ool00`00Oomoo`02Ool00`00Oomoo`09Ool00`00Oomoo`05Ool01@00Oomoogoo
00003goo00<007ooOol027oo00@007oo000000=oo`03001oogoo00=oo`03001oo`0002moo`04001o
ogoo0002Ool01@00Oomoogoo00000Woo00<007ooOol027oo00D007ooOomoo`0000]oo`03001oogoo
00Eoo`05001oogooOol0000?Ool00`00Oomoo`08Ool00`00Oomoo`020002Ool00`00Oomoo`03Ool0
1000Oomoo`0067oo0@0000000`00Oomoo`08Ool00`00Oomoo`0QOol20018Ool00`00Oomoo`0MOol0
0`00Oomoo`0QOol20019Ool00`00Oomoo`0LOol00`00Oomoo`0ROol2000IOol100000003001oogoo
01eoo`03001oogoo00Uoo`04001oo`00000bOol00`00Oomoo`0LOol00`00Oomoo`0ZOol00`00Oomo
o`09Ool00`00Oomoo`02000aOol00`00Oomoo`0MOol00`00Oomoo`0YOol00`00Oomoo`09Ool00`00
Oomoo`02000JOol100000003001oogoo00Moo`03001oogoo009oo`05001oogooOol00003Ool00`00
Oomoo`02000=Ool00`00Oomoo`0fOol01`00Oomoogoo001oo`0000Ioo`03001oogoo01]oo`03001o
ogoo00Uoo`05001oogooOol00003Ool00`00Oomoo`02Ool01@00Ool007oo00003Goo00<007ooOol0
=Woo00H007ooOol007oo0007Ool00`00Oomoo`0JOol00`00Oomoo`09Ool00`00Oomoo`02Ool01@00
Oomoogoo00000goo00<007ooOol00P003Goo00<007ooOol07Woo0@0000000`00Oomoo`0?Ool00`00
Oomoo`0@Ool00`00Oomoo`06Ool00`00Oomoo`0dOol00`00Oomoo`02Ool00`00Oomoo`09Ool00`00
Oomoo`0WOol00`00Oomoo`0@Ool00`00Oomoo`06Ool00`00Oomoo`0dOol00`00Oomoo`03Ool00`00
Oomoo`08Ool00`00Oomoo`0XOol00`00Oomoo`0?Ool00`00Oomoo`07Ool00`00Oomoo`0IOol10000
0003001oogoo01Yoo`03001oogoo00Qoo`03001oogoo03Qoo`8002=oo`03001oogoo025oo`03001o
ogoo00Moo`03001oogoo03Qoo`<002=oo`03001oogoo021oo`03001oogoo00Qoo`03001oogoo01mo
o`40000000@007ooOol0009oo`03001oogoo01Qoo`04001oogoo0005Ool200000goo001oo`0lOol0
1@00Oomoogoo00000Woo0P0017oo00<007ooOol00goo00<007ooOol00goo0P002goo00@007ooOol0
01Yoo`05001oogooOol00004Ool20002Ool00`00Oomoo`0kOol01000Oomoo`000goo0P000goo00<0
07ooOol00goo00<007ooOol017oo0P002goo00@007ooOol001Yoo`04001oogoo0005Ool200000goo
001oo`0KOol100000003001oogoo01eoo`03001oogoo049oo`03001oogoo009oo`03001oogoo00mo
o`05001oogooOol0000^Ool00`00Oomoo`12Ool01@00Oomoogoo00004Goo00D007ooOomoo`0002io
o`03001oogoo02Moo`40000000<007ooOol017oo00<007ooOol017oo00<007ooOol01Goo00<007oo
Ool037oo00<007ooOol00goo00<007oo0000>Woo00<007ooOol017oo00<007ooOol08goo00<007oo
Ool017oo00<007ooOol01Goo00<007ooOol03Goo00<007ooOol00Woo00<007oo0000>goo00<007oo
Ool00goo00<007ooOol097oo00<007ooOol017oo00<007ooOol017oo00<007ooOol03Goo00<007oo
Ool00Woo00<007oo00007Goo0@0000000`00Ool0000IOol00`00Oomoo`08Ool00`00Oomoo`03000i
Ool01000Oomoo`0047oo00<007ooOol03Goo00<007ooOol00goo00<007ooOol05goo00<007ooOol0
2Goo00<007ooOol00P00>Goo00D007ooOomoo`0000moo`03001oogoo00eoo`03001oogoo00Aoo`03
001oogoo01Moo`03001oogoo00Qoo`03001oogoo00<001eoo`40000000<007ooOol027oo00<007oo
Ool02Woo0P00DWoo00<007ooOol04goo00<007ooOol05Goo00<007ooOol02Goo00<007oo0000DGoo
00<007ooOol04goo00<007ooOol05Goo00<007ooOol02Woo0P00<7oo0@0000000`00Oomoo`0=Ool0
1@00Oomoogoo00003goo00<007oo0000@7oo00<007oo00002Goo00@007ooOol002Moo`05001oogoo
Ool0000?Ool00`00Ool00010Ool00`00Ool00009Ool01@00Oomoogoo00009Woo00D007ooOomoo`00
011oo`03001oo`0002=oo`40000000<007ooOol01goo00<007ooOol017oo00<007ooOol02Goo00<0
07ooOol00P0027oo00<007ooOol0>Woo0P001Goo0P000goo00D007ooOomoo`0001Aoo`03001oo`00
00Uoo`03001oogoo00Eoo`03001oogoo00Qoo`05001oogooOol00008Ool00`00Oomoo`0kOol00`00
Oomoo`03Ool00`00Ool00002Ool01@00Oomoogoo000057oo00<007oo00002Goo00<007ooOol01Goo
00<007ooOol027oo00@007ooOomoo`8000Moo`03001oogoo01ioo`40000000<007ooOol08goo0P00
H7oo00<007ooOol0:Woo00<007ooOol0GWoo00<007ooOol0:Woo00<007ooOol08Goo0@0000000`00
Oomoo`0VOol00`00Oomoo`0nOol00`00Oomoo`0>Ool01000Oomoo`001Goo00<007oo0000<7oo00<0
07ooOol0?Woo00<007ooOol03goo00@007ooOol000Aoo`03001oo`00035oo`03001oogoo01ioo`40
000000@007ooOomoo`8000moo`03001oogoo00Eoo`03001oogoo009oo`8000=oo`8004=oo`8000Ao
o`03001oogoo00Yoo`03001oogoo011oo`03001oo`0000ioo`03001oogoo00Eoo`03001oogoo009o
o`03001oo`00009oo`<004=oo`8000Aoo`03001oogoo00Yoo`03001oogoo00moo`03001oo`0000mo
o`03001oogoo00Eoo`07001oogooOol007oo00000goo0P008Goo0@0000000`00Oomoo`05Ool00`00
Oomoo`02Ool00`00Oomoo`09Ool00`00Oomoo`0:Ool00`00Oomoo`11Ool00`00Oomoo`0>Ool00`00
Oomoo`0HOol00`00Oomoo`03Ool00`00Oomoo`08Ool00`00Oomoo`0:Ool00`00Oomoo`12Ool00`00
Oomoo`0=Ool00`00Oomoo`0IOol00`00Oomoo`02Ool00`00Oomoo`08Ool00`00Oomoo`0;Ool00`00
Oomoo`0QOol100000003001oogoo01Eoo`03001oogoo059oo`04001oogoo000fOol00`00Oomoo`1A
Ool01000Oomoo`00=Woo00<007ooOol0;goo0@0000000`00Oomoo`09Ool01@00Oomoogoo00000goo
00<007ooOol01Goo00L007ooOomoo`00Ool00002Ool00`00Oomoo`13Ool30006Ool01000Ool00000
2Woo00<007ooOol01Goo00<007ooOol00goo00<007ooOol01goo00<007ooOol00Woo00D007ooOomo
o`0000Ioo`05001oogooOol00002Ool01000Oomoo`00A7oo0`001Woo00@007ooOol000]oo`03001o
ogoo00Eoo`03001oogoo009oo`03001oogoo00Qoo`05001oogooOol00003Ool00`00Oomoo`05Ool0
1000Oomoo`000Woo00@007ooOol002Aoo`40000000<007ooOol03Woo00<007ooOol02Goo00<007oo
Ool0DWoo00<007ooOol05Woo00<007ooOol04Goo00<007ooOol02Woo00<007ooOol0DGoo00<007oo
Ool05goo00<007ooOol047oo00<007ooOol02Woo00<007ooOol0:Woo0@0000000`00Oomoo`06Ool0
0`00Oomoo`0GOol00`00Oomoo`1EOol00`00Oomoo`0@Ool00`00Oomoo`06Ool00`00Oomoo`0GOol0
0`00Oomoo`1EOol00`00Oomoo`0AOol00`00Oomoo`06Ool00`00Oomoo`0FOol2000VOol100000003
001oogoo01Ioo`03001oogoo00Qoo`8004Qoo`04001oogoo000=Ool01@00Oomoogoo00000goo00<0
07ooOol07Woo00<007ooOol01goo0P00B7oo00@007ooOol000eoo`05001oogooOol00003Ool2000O
Ool00`00Oomoo`07Ool3000TOol100000004001oogooOol2000>Ool00`00Oomoo`07Ool2001FOol0
0`00Oomoo`0JOol00`00Oomoo`0=Ool00`00Oomoo`06Ool2001FOol00`00Oomoo`0JOol2000>Ool0
0`00Oomoo`07Ool2000ZOol100000003001oogoo019oo`04001oogoo0002Ool00`00Oomoo`04Ool0
0`00Oomoo`19Ool00`00Ool00003Ool01@00Oomoogoo00000Woo00<007ooOol09Goo00@007ooOol0
00=oo`03001oogoo00=oo`03001oogoo04Yoo`03001oo`0000=oo`04001oogoo0003Ool00`00Oomo
o`0UOol01000Oomoo`000Woo00<007ooOol017oo00<007ooOol09Goo0@0000000`00Oomoo`05Ool0
0`00Oomoo`02Ool01@00Oomoogoo00003Goo00<007ooOol0CGoo0`0027oo00<007ooOol03Woo00<0
07ooOol02Goo00D007ooOomoo`0000=oo`03001oogoo00]oo`03001oogoo04eoo`800003Ool007oo
00Ioo`03001oogoo00ioo`03001oogoo00Uoo`05001oogooOol00004Ool00`00Oomoo`0:Ool00`00
Oomoo`0YOol100000003001oogoo01Moo`03001oogoo00Aoo`03001oogoo055oo`03001oogoo00]o
o`03001oogoo00Uoo`03001oogoo01Ioo`03001oogoo00Eoo`03001oogoo051oo`03001oogoo00]o
o`03001oogoo00Uoo`03001oogoo01Moo`03001oogoo00Aoo`03001oogoo02Ioo`40000000<007oo
Ool03Goo00<007ooOol02goo0P00OWoo00<007ooOol037oo0P00OWoo00<007ooOol02goo0P00:Woo
0@0000000`00Oomoo`05Ool00`00Oomoo`0@Ool00`00Oomoo`02Ool2001@Ool01000Ool0000027oo
0P000Woo00<007ooOol00Woo00D007ooOomoo`0000=oo`03001oogoo00Eoo`03001oogoo00moo`03
001oogoo009oo`80051oo`04001oogoo0009Ool20002Ool01@00Oomoogoo00000goo00D007ooOomo
o`0000Moo`03001oogoo011oo`03001oogoo009oo`03001oogoo02Moo`40000000<007ooOol04Goo
00@007oo0000061oo`05001oogooOol0000:Ool00`00Oomoo`0HOol01@00Ool007oo0000H7oo00@0
07ooOol000]oo`03001oogoo01Moo`05001oo`00Ool0000bOol100000004001oogooOol20003Ool0
1000Oomoo`0017oo00<007ooOol01Woo00D007ooOomoo`0005Aoo`8000Aoo`03001oogoo01Moo`05
001oogooOol00002Ool00`00Oomoo`03Ool00`00Oomoo`06Ool01000Oomoo`00E7oo0P001Goo00<0
07ooOol05Woo0P000goo00@007ooOol000Aoo`03001oogoo00Ioo`04001oogoo000/Ool100000003
001oogoo00Yoo`03001oogoo00]oo`03001oo`0005Moo`03001oogoo02Aoo`03001oogoo00Yoo`03
001oo`0005Qoo`03001oogoo02=oo`03001oogoo00]oo`03001oo`0002aoo`40000000<007ooOol0
5goo00<007ooOol0G7oo00<007ooOol00Woo00<007ooOol00goo00<007ooOol017oo00@007ooOol0
01Yoo`03001oogoo05]oo`03001oogoo009oo`03001oogoo00=oo`03001oogoo00Eoo`04001oogoo
000IOol00`00Oomoo`0]Ool100000003001oogoo00Eoo`03001oogoo00Aoo`05001oogooOol00002
Ool20004Ool00`00Oomoo`1GOol01000Oomoo`000goo00H007ooOol007oo0003Ool01@00Oomoogoo
00002goo00<007ooOol01Goo00@007ooOol000=oo`03001oogoo009oo`8005Qoo`04001oogoo0003
Ool01000Oomoogoo0P000goo00D007ooOomoo`0000]oo`03001oogoo00Eoo`05001oogooOol00002
Ool00`00Oomoo`03Ool00`00Oomoo`0[Ool100000003001oogoo00moo`03001oogoo069oo`8001Mo
o`03001oogoo00eoo`03001oogoo069oo`8001Moo`03001oogoo00ioo`03001oogoo03Eoo`400000
00<007ooOol057oo00<007oo0000GGoo00<007ooOol00goo00<007ooOol027oo00<007ooOol06Woo
00@007ooOol005eoo`03001oogoo009oo`03001oogoo00Qoo`03001oogoo01]oo`03001oo`00031o
o`40000000H007ooOol007oo0004Ool20002Ool00`00Oomoo`07Ool2001POol00`00Oomoo`0EOol0
0`00Ool00003Ool01@00Ool007oo00002Goo0P00H7oo00<007ooOol05Goo00<007oo00000goo00<0
07oo00000Woo00<007ooOol01Woo00<007oo0000<7oo0@0000000`00Oomoo`0COol00`00Ool0001S
Ool20003Ool01000Oomoo`007goo00@007ooOol006=oo`80009oo`05001oogooOol0000OOol00`00
Ool0000aOol100000003001oogoo00aoo`03001oogoo008006Ioo`03001oogoo00Aoo`05001oogoo
Ool00003Ool01P00Oomoo`00Ool000ioo`05001oogooOol0001WOol00`00Oomoo`03Ool00`00Oomo
o`02Ool02@00Oomoogoo001oo`00Ool0000?Ool01000Oomoo`00=goo0@0000000`00Ool00004Ool0
0`00Oomoo`05Ool01`00Oomoogoo001oo`0006Eoo`8001=oo`03001oogoo009oo`03001oogoo00Eo
o`03001oogoo009oo`03001oo`0006Aoo`8001Aoo`03001oogoo009oo`03001oogoo00Aoo`03001o
ogoo009oo`03001oo`0003Aoo`40000000<007ooOol00goo00D007ooOomoo`0000Yoo`03001oogoo
06Ioo`03001oo`0000Qoo`03001oogoo00Qoo`03001oogoo009oo`03001oogoo00Moo`03001oogoo
06Ioo`04001oogoo0007Ool00`00Oomoo`09Ool01@00Oomoogoo00002Goo00<007ooOol0<goo0@00
00001000Oomoo`001goo00<007ooOol017oo00<007oo0000K7oo00D007ooOomoo`0000]oo`03001o
ogoo00Eoo`03001oogoo00Aoo`03001oo`0006eoo`04001oogoo000<Ool00`00Oomoo`04Ool00`00
Oomoo`04Ool01000Oomoo`00=Goo0@0000000`00Oomoo`0;Ool2001_Ool20004Ool01@00Oomoogoo
00000goo00<007ooOol03Goo0P00KWoo0P0017oo00D007ooOomoo`0000=oo`03001oogoo00eoo`80
03Yoo`40000000@007ooOol0009oo`03001oogoo009oo`03001oogoo009oo`8007moo`05001oogoo
Ool00003Ool00`00Oomoo`02Ool2001oOol01@00Oomoogoo000017oo00<007ooOol00Woo0P00>7oo
0@0000000`00Oomoo`04Ool01000Oomoo`000goo00<007oo0000LGoo00<007oo000017oo00<007oo
Ool02Goo00@007ooOol000=oo`04001oogoo001`Ool00`00Ool00005Ool00`00Oomoo`09Ool01000
Oomoo`000Woo00@007ooOol003Uoo`40000000<007oo00002Goo0P00MGoo00H007ooOol007oo0003
Ool20003Ool00`00Oomoo`06Ool2001eOol200001Goo001oogoo00000goo0P000Woo00<007ooOol0
1Woo00<007oo0000?7oo0@0000001000Oomoo`000Woo0P0000Eoo`00Oomoo`0007Uoo`05001oogoo
Ool00007Ool01`00Ool007oo001oo`00009oo`03001oogoo07Moo`04001oogoo0007Ool00`00Oomo
o`0200001Goo001oogoo0000?Goo0@0000000`00Ool00024Ool20007Ool00`00Oomoo`22Ool20007
Ool00`00Oomoo`16Ool100000004001oogooOol20002Ool20020Ool300000goo00000002Ool30000
0goo00000021Ool2000017oo000000000Woo0P000Woo00<007ooOol0?goo0@0000000`00Oomoo`3o
OomUOol100000?l006T00000\
\>"],
  ImageRangeCache->{{{0, 359}, {221.375, 0}} -> {-0.0261857, -0.0250052, \
0.00366425, 0.00592887}},
  ImageSoundCache->{{{0, 359}, {221.375, 0}} -> GraphicsData["WindowsSound", "\<\
DTU6AZdC001G@EI5IVed810000010040@1l0040O000100P0I65dHHTC001ooJhC
:<ccGP1an[/K7/3hK01SmLLU5[?lNP5Ek]<_3ZKoR@A8imdk2ISoU`PlgNI818[n
Y@h`dniE0G_l/QHVb?AS06gh_alL_?Ua0EobbRPD[oao0e;[eC<>XOj=1TGSgcl8
TojK2cWJid`5QObX4Rk?kUT2MoVe6RC3m6L2JOG18a^gn7D3G>o=;ABZnh<5CnOG
>0jMo94:@moPA0V?nih?=mGXD@J1nJ/F;L[^GPEcm[PN8kkcK0EVlL<X6k;fN@II
j/hb5JGhQ`Y<h]Lm3iSiU0i1fN192h[hXA@fd>MF2Gge[Q//aNaS273a^B@T^O5`
2F?[a2dL[O=m2eKUcSPFX?F:3d_Med<BToBG543DgTl?Q_>T6SK:iE/=NO2_8Bfo
jVP=KN^j:RFckG@>H>G4<anWkh4AEMk=?QVKl8hEB]KEBAJ?kiXJ@<gLE1>2kJDQ
=l?QH19fjK0X;kWVK1=Zi;Xa::gXNAAOg/<j8Z;ZQAMDela47YKZT1a:cm=?6h[X
W251a]UJ6GoVYRPh_=eU67?R/30a/^1a6FSM^CP[Z>9l75kG`T4VWN>87eC?bD/S
TN>B94_7ceDQQ^6M:D:oe60OO=jV<3^ef6XPLMZ_=cF/fWDQImBh@32Rg80TGLjo
B2bGg8XWE<O5DRV=fi@/C<3:FbR3f9dbAKO>IBMieJHh?jkAKbQ_dJi0>ZGCNBYV
c;E7=YcDPReNa[]@<i;CS35F_l1H<HWBU3E?]lAQ<83?WC]9/<QZ<7O<Y454YlYc
<Fo7ZdQ0Wl]l<fO2/DllU/^5=Unl]ULjS/Z=>UVf^UliQ/RD?e>__VLiOLFKA4jX
`6liM/:RBTZQ`GLkK[jWDDJI`WlmJ;V/EdBA`HI0H[>`GT>:`8e4G:fdID:3_Y=9
EjNfK49l^iU>Dj2gLd=e]iiCD9ZhNTE_/j=ICY>hPDQZ[jMOCHfgQd]UZZYUC8Je
SDmPY:a[C82cTU=MWjeaCG^`UUQJVJigCWF]VUaHU:imD76YWV5GSZj2DfbUX6MF
RJb7EVVPXVaFQ:Z;FFJLXg5FOjR?GF>GY7IHNZFBHF:BY7YIMZ:EIF2>Y7mKLinG
JV29XX=NL9^IKV25XHIQKYNJLV22WhUTK9>JMV5nWHaWJi2JNF=kVXi[JXbJOFEh
Uhm^JXVHP6MfU91aJXFGPVUeTI5dJX>EQ6acSY5gK82CQVmcRi1jKGjAQg5bRI1l
Kgb?R7AcQXinLG^<RGIcQ8f0LgZ:R7UdPX^1MGZ8R7]ePHZ2MgZ6QgagOhR2NGZ4
QWiiOhJ2Ng^2QGmjOXB2OGb1PgmlOX:1OWj0PH1nOh20Ogn0P7n0P7moP85oOWn1
PGimPH>0O7j3PWekPHF0NWf4Q7eiP8J1NG^5QWegP8R3N7V5R7eeOXZ4MgN5RWid
OH^6MgF5S7mbNhb9MW>4SX5aNHb;Mg23T8=_MXf>N6j1TXE_M8b@NFaoThQ^LHbC
NfYlUH]^KXZFOFQjUHi_JhVHOfMfUI5`J8NJPVEcUIAaIHBLQVE_U9QcHX6NRFE/
Ti]fH7fOSFEXTIiiGWZOTFITSj5lGGJOUFQPS:B0Fg6OVVYMRJJ5FffMWVeJQJN9
FfRLXW1GP:V>G6>IYWAENjVCGEnFZGQCMZVHGeZC[7eALJRNHEJ>[h9AJjNSIE::
/HQAIZFWJDn4/XiBH::/KDao/iACFin`LTYi/iYEEIZdN4Qb/Z1HD9FgOTM//JIL
C92jQDMU[jaQB8ZlS4QO[;5VA8>mTdUHZ;I/@GbmVT]BXk]b?gFlXDm<WKmi?Vjk
Z5=6Ul:0?FJi[UQ1TLB8?Uje]EdmRLJ@?eNa^f@jP/JH@E2/`6/gNLJPADVVaG8e
LLFXBD>PbGXdJL>`CSfHc8<dH;ngE3R@cX`eF;^nFc>8ciDgCkK4HS1ocihkB;3:
JbefcZLo@:W>Lba]c:m4>J7BOB]SbKQ:<iWEQRaJaKmA;Y3GT2iA`<MI:XKHVS58
^/eR9gcGY3Do/m=[97;F[CXhZmQe8fSC]d0aX]]o8eg?_dL[V=j:9E?:b50US]nE
9d[4ceTQPn2O:d2meV8NNMjZ<3RdffdLK]bd=S2[h7LLH]Vn?BRQhh<MEmC7AR:G
iXhOCLk?Caf<iiXR@lOGFAV0iZDV>KkNHaMei;0/<;GSKaEYhK/c::_XNQEMgLDl
8J3[QQMBeli56iG/TaU6d=M?5XW]Wahlb=iJ4Wg/Zb<b_nEU473Y]RXY]NYa3fCV
`C8QZ^in45SPc3/JWo6;4TcJeDDDT_:G5D3BgU0?Q_:T6SG9iE/<NO2`8B^ojfP;
K>fl:2:dl7D:GnW7<AZYm88<D^?B>aBLmXl>A]_KAPj?mi`C>]?TD@Z2mZTH;lW[
GPMem;HO9[kaJ`IWl<8X7K?eN0MJj/da5JKiQPU=i=Ll3YWjT`a1g>572HcjXA4e
d^UD1Wki[QLZb?1Q173f^alP_?E^0f;aabPH/?Yl15G[dS8@Xob:1dSTg3h:U_fH
2ccKiDX6R?fV4C3AkEL3N_Zc62G6m6@1K?O081bjnG81G_7<:QB]o803D>_G=@fP
oXh6@n?P@0NBoi`;=mWYC@>4oZXA;<oaFP5enkLI8/?gJ01Wm/@R6;OkMP1IlM0/
4J[nQ@=<jMXh2YcoT`LohNA41HkoX@`cenaA0Wom[Q<Xc?=N077j^a/N`?Q/06?e
abDF/oaj0EG^dbl>Y_n914SWgC/9V?nG23cMiTP4R_jU3S3CkUD1Nobb5RK8m6<0
KORo7ablnG41Go;::1B_o7l3D^_E<`jQoXd6AN?O?`RCoY/;>M[WC0F5o:PB;/o^
F@9gnKDJ9<?dI`9YmL4S6kOhM@=Lkld]5:[kP`E?imLh3YglT@Y3gn142HokWPlg
eNQA1X7iZaH]b^iN1G?f^1hS__=/1FKa`bPK/_Ii1UWZcS8EYOR72TcRecd?V?VD
3T7Ih4T;R_RQ53K@ieH9OOF^6bc5k6<8L?6i92BilG09Hn_4;Ab]lgd;E^G>>1JP
mHX?BmgG@a:Cm9LD@=CNC`n6lj@J=/[UF`eil:lQ;KoZJ0e]jkXZ9K?]M0iPiL@c
7jO_PA5Eg/dn6I_`SQE:e]E95Xo_VQY0cMaD4h;]YB4g`n5P4WKY/2P_^NI/4f[T
^S4X[NQi55oN`cXRX^Z55eCGc4@NU^Z@74[?ddlKR^RL8D76fEXIOnJV:3RlgFDH
Ln:`<36bh74IJ=fi>2^XhW`LG]O2@BJMhhPOE<o9Bb>Ahi8TBlO?EB66hIdY@[oD
H1mlgZH`>kGHJR1afZlg=JcJMB5We;Q0<:;LP2AMc[m8;9OLRRMDalEB:HgKU2a<
`<YK:8?HWC95]liU9gWEYSPo[]5_:6oA[T0jYM=i:VK<]DLfW=B2;Ek6^e0cT]><
<EJo`5PaRM:D=Dnga64`P<nM>dV`b6X`MlbT@DBWbW<aKlN[B42Obg`cIl:aCcbF
bhDfGkbfEcZ>bXdjFKJjGcV6b9@oDjnnIcUmaI]4CZS0KcUf`Z9:BZ71Mc]^_ZMA
AYW2OceX^JaGA971QT1R/k1N@h[0SDAL[KAU@X>nTdUGYkI/@WbkVDiCX;Mc@gFg
WU=@V[QjAFncXeU>TkR1B6Z_Yem=SKN7BfFZZVE<Q[F=Cf2T[6]<P;>BDefO[G5=
Nk2FF5ZI[WM>MJfJG5RD[We@LJVNHEN>[X9CK:FPIeJ9[8MFJJ2RK5J4ZX]IIYbS
LEIoZ8mMHiNTMUQjYI9QHY:TNUUfXYEUH8jTOe]cWiMZH8VRPei`ViU^H8FQQV5^
UiYbH8:ORFA/TiYfHGjMS6M[T9YiHg^JSV]ZS9YmIGRGSfiZRIR0IgJDT75ZQIN2
JGFATGAZPiF4K7>>TGM/P9>6Kg>;T7Y]OY67LG:9T7a_O8n8M7>6SWiaNhb9MW>4
SH1cNXZ8NGB2Rh5eNXR8NgF1RX9gNXJ7O7MoR89iNXB6OWUoQX9kNh:5OgYnQ89m
O863OganPX5nOX21P7ioP81oOh20Oh20Ogn0PGmnOh61OWf1Ph1lOX>2OG^1QH1j
OHB4OGV0QX5iNhF6OGN0R8=hNHF8OGEnRXAgMhF:OWAmRhIgMHF<Og9kS8UfLhB>
PG5iS8]gL8>@PfmfSHihKX6BQFmdS91iK7nCR6iaS9=kJWbERfi^RYImJ7ZESVm[
RIQoIgJETG1XQiZ2IG>EU75UQ9b6IFnDV7=RPIj9IFbCVgIPOIn=IFRAWWUNNYnA
IVB?XGaMMYnEJ62<Y81KLInJJUf9YXEKKIfNKEZ5YhUKJ9bRL5N0ZHiLHiVVM5Ek
ZI=MGiJYN5=fZIQOFY>/OE5aZ9iQEXj_PU5[Yj=UDXZaR55VYJMYChBbSU9PXZa]
C7ncU5=KWk1bBWVcVUEEV[AhB7:bX5Q@UKMnAfbaYUa<T;Z5AfF_[658R[b<B5n/
/FI4PkfCBERX]Va1O;fJBe:S^g8oMKbQCdbM_gTnK[^XDdJG`X0mI[V^F46Aa8Pn
G[FeGCf9aY0oEk6kI3Z2aYQ1D:c0JcMiaZ15BJK5LSEaaJQ9@j39NSAY`k1>?IS<
PcAP_kMD>93>S3EH^kiK<hS?UCM?]/AR<7o?WS]8/<Y[;GK>Ycm0ZLic;6g<[d@i
XM9m:f?9^4XcVMF6;5[5_e4^T=N@;U70aeTZQ]RJ<DRjcF8WO=NT=Cncdf/TL]J]
>SR[f7DSJ=>g@36RfglSGLnoAb^HgXXUDl[8D2F>giDWB/C?FB63h9l[@;gFHQii
gZX`>;CKKAa^g;@f<:_PMaaRfKhm::7SPaeGe<M68YOVSQm=c/m?7HcWVR93amMI
6H3VYBHi_]iS5gGT/2``]N=_5FWQ^c<XZnQj5EgMaC`QX>^65e;GcTDKUNbC6DK@
edlFRNfO7Sc8gUXBONb[8c:oiFD@L>Vf:RVejW4?I>K1<R6ZkWh@F>3<>aZOlH/B
C=[EAABBlYLE@=;ND0n6lZ@J=LWUF`ail;0Q:ko[J0]/kK`X8[C`M@YOjLLa6ZWd
PPaBhm8k59cfS`i6fm]63XogW1<jdnAA2X;fZAP_bN]N1gGd]QlV__5[1VO``RPM
/oEh1e[ZcC4EY_V62DgTec`>VOZC347LhDL9S?ZQ4CGBjE@6O_V^5b[8l644L?Jk
7b2lmFh3H_77:1R`nW`4EN_B<Q2So8X7B>CL?PZFoIP;?=_UBPJ8oJHA<=7]E`=j
n[<H9LKdI05/ml0P7;[iLP5NlL`Z5:glP0=@jmLe3J3nSPI3hn101i;oW0/gfNU=
0hCnZQ4/co5J0GGk]aTR`oMX06Ofa28H]o]f05Wad2`AZ_j50dcYfSP:W?nC1coQ
i4@5S_nQ33?Gk542Oof^4bS<leh0LOZk6ak0n6`0HoG79AJco7X1ENkC;`jVohT4
B>OM>`VHoiL8?=gVB0B:oZD><=?^E@5ko;8F9/SdH`1]n;lO7;ciL@5Ol/XX5:ol
O`=BjmDc3Z7nS@I5hmlo29?nV`/if^M<1HGlZ18^cniI0WOi]AXT`oAW0VWe`B<K
]oQe0ec_cBdDZ_^31DoWecP>WObA2T?Oh4@9So^N3cOEj546POV[5Rg:kUh5LoJh
7R>nlf`5I_73:1^bmWT6FN[><QFUn8L:C>;G?@nHnI@>@MWPB@^:n:4D=]3WEPUm
mJhK;<G/H`Q`lKTT9;WaL0USjl@]7:gcO@]FiLhh5Z3eRPm;gMM34Y?dUaA0e=i?
3hKcY1Xfb^EK3GW`[b4]_nYX3Fg[^RXU/ned3V3Ua3<OYnn14EGNcChIVo2>5D[F
eDTFSnnJ6T3=g5@CP^fU8CO3hF0BM^V`:2niiV`CJ^Bj<BR]j7TDGmk3>R:RjXDG
E=O<A1jFjY0LB/oCCa^:j9`Q@LKIFQUoiZHX>;cMIAQch[0`<K;PLAUXgKTh:jSR
O1aNel919YgSR1mDclU;8i7STRA;almE8HKQWBU2_mAP7gcNYS0k]MQZ877J[cLe
[=Ye8FOD^40`X]b095g>_dP/Umb:9eC7aE8YSM^D;4c0bU/XPmRM<TFgcVDWNMFV
>3n^dFlXKm6^@3ZUdgTZI/beAcJLe88]G/JkD3>Bdh`aE[o0F369dY@eCkO4HC20
cidkBK38JS1gc:A1A:O:Lc5_aj]8@9o;O3=W`[5??9K;QCIO_;IG>Xk:SCYI][YO
>HK8U3mC[kiW>Gg5VdA>Z<1_>GK2XTY:XL5g>fjnYe56VL9o?FRi[5M4TL66@6:c
/5i3R/2=A5b]]6E2PkjCBENW]Va2O;^ICU>P]g=3MKNNDe2J^7Y5Kk>SFDjC^858
JZnWGdf=]hM;IJZZIDb6]He?H:B/Jdb0/i9CGIn]LDek/9IHFYV^Mdie[IYLF9B^
OE1aZIiQEhj^PU=/YJ1WEXV/QeIYX:9/EXBZReUVW:=aEWnXSeeSUjAfF7ZUTV5R
TZAjFGJRUFEPSZAoFg>OUfYPRJ:3GW2KVFiPQJ66HFjGVW9PPYn9I6bCVWIQOYf<
If^@VWUSNiZ>JfZ<VWeUN9N?KVZ9V81WMYB@LFZ5Uh9YMI6AM6Z3UHA/LhjAMfb0
ThI_Lh^@NVenTHMaLXV@O6mlShQdLhJ>OW5kS8UfLhB=P7=jRXQiM8:;PGEjR8Qk
MH6:PWMjQXMlMgn8PWUjQ8InNGn6PW]kPXEoNWj4PWelPH=oO7j2PGinP860OWn0
P7moP80\>"]},
  ImageCacheContents->"Sound"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Storing definitions", "Subsection",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "The assignment statement of a symbol is not saved with ",
  StyleBox["Put", "InlineInput",
    FontWeight->"Bold"],
  " (",
  StyleBox[">>", "InlineInput"],
  "); only its evaluation is saved.  The assignments of symbols (variables \
and functions) can be saved to a file by this command.\n    \t",
  StyleBox["Save[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput"],
  StyleBox[", ", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["{s1, s2, s3,\[CenterEllipsis]", "InlineInput"],
  StyleBox[" ", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["}", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\n",
    FontWeight->"Bold"],
  "Symbols can represent simple assignments, functions, procedures, graphical \
data, or any other ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " object. Note that the name of the file and the symbol to be saved are \
passed to ",
  ButtonBox["Save",
    ButtonStyle->"RefGuideLink"],
  " in the opposite order from ",
  StyleBox["Put", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["PutAppend", "InlineInput",
    FontWeight->"Bold"],
  ".  Also, note that ",
  StyleBox["Save", "InlineInput",
    FontWeight->"Bold"],
  " ",
  StyleBox["appends",
    FontSlant->"Italic"],
  " definitions to a file; it does not replace the file contents with the \
definitions.  When the file is read into ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " using\n\t",
  StyleBox["<<", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ", "InlineInput",
    FontSlant->"Italic"],
  "\nall these assignments will be activated. If you ",
  StyleBox["Save", "InlineInput",
    FontWeight->"Bold"],
  " a symbol whose definition uses other symbol definitions, then ",
  StyleBox["Save", "InlineInput",
    FontWeight->"Bold"],
  " also saves these definitions.  (Section ",
  ButtonBox["2.11.1",
    ButtonStyle->"MainBookLink"],
  ")"
}], "Text"],

Cell["Exercise: storing and retrieving definitions ", "Exercise"],

Cell["\<\
The following example constructs and stores an interpolating \
function based upon a relatively lengthy numerical computation.  Rather than \
repeat such an evaluation every time the function is needed, we can simply \
retrieve its definition from a file and use it.  \
\>", "ExerciseText"],

Cell[TextData[
"1.  Define the partition function for an axially symmetric rotor in terms of \
a numerical sum.  We also establish definitions for some basic thermodynamic \
relationships.  Do not worry if the physics is unfamiliar \[LongDash] the \
point of this exercise is to store a nontrivial calculation for later use, \
not to study statistical mechanics."], "ExerciseText"],

Cell[BoxData[
    \(Z1rot[y_] := 
      NSum[\((2\ j + 1)\)\ Exp[\(-j\)\ \((j + 1)\)\ y], {j, 
          0, \[Infinity]}]\)], "Input"],

Cell[BoxData[{
    \(\(F[T_, Z_] := \(-N\)\ k\_B\ T\ Log[\ Z[T]\ ];\)\), "\n", 
    \(\(S[T_, Z_] := \(-\[PartialD]\_T F[T, Z]\);\)\), "\n", 
    \(\(U[T_, Z_] := F[T, Z] + T\ S[T, Z];\)\), "\n", 
    \(CV[T_, Z_] := \[PartialD]\_T U[T, Z]\)}], "Input"],

Cell[TextData[{
  "2. Compile numerical data suitable for interpolation.  It is useful to \
divide the temperature range into three nonoverlapping regions with different \
step sizes.  Do not be concerned about the error messages for extremely large \
",
  StyleBox["j",
    FontSlant->"Italic"],
  " and be patient, this could take some time."
}], "ExerciseText"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(TableZ1RotLowT = 
        Table[{T, Z1rot[y] /. y \[Rule] 1\/T}, {T, 0.05, 0.5, 
            0.05}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(NSum :: \\\"nsnum\\\"\\) will be suppressed during this calculation. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], \
"Message"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(TableZ1RotMidT = 
        Table[{T, Z1rot[y] /. y \[Rule] 1\/T}, {T, 0.6, 2.9, 
            0.1}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(NSum :: \\\"nsnum\\\"\\) will be suppressed during this calculation. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], \
"Message"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(TableZ1RotHighT = 
        Table[{T, Z1rot[y] /. y \[Rule] 1\/T}, {T, 3, 10, 0.5}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(NSum::"nsnum"\), \(\(:\)\(\ \)\), "\<\"Summand (or its \
derivative) \\!\\(\\(\\((\\(\\(2\\\\ j\\)\\) + 1)\\)\\)\\\\ \
\[ExponentialE]\\^\\(\\(\\(-j\\)\\)\\\\ \\(\\((j + 1)\\)\\)\\\\ y\\)\\) is \
not numerical at point \\!\\(j\\) = \\!\\(46661\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"NSum::nsnum\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(NSum :: \\\"nsnum\\\"\\) will be suppressed during this calculation. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], \
"Message"]
}, Closed]],

Cell[BoxData[
    \(\(TableZ1Rot = 
        Join[TableZ1RotLowT, TableZ1RotMidT, TableZ1RotHighT];\)\)], "Input"],

Cell["3. Now we construct the interpolating function.", "ExerciseText"],

Cell[BoxData[
    \(\(Z1RotInterpolatingFunction = 
        Interpolation[TableZ1Rot, InterpolationOrder \[Rule] 5];\)\)], "Input"],

Cell[TextData[{
  "4.  Finally, we save the interpolating function in a file named ",
  StyleBox["zrot", "InlineInput"],
  ".  Although we probably would not need additional information, we also \
save the symbolic definition of the partition function and the numerical data \
used for the interpolating function.  We save the thermodynamic relationships \
also.  An interesting feature of ",
  StyleBox["Save", "InlineInput",
    FontWeight->"Bold"],
  " is that the definitions of antecedents are saved automatically.  Thus, \
the definitions of ",
  StyleBox["F", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["U", "InlineInput",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["S", "InlineInput",
    FontWeight->"Bold"],
  " are saved because ",
  StyleBox["CV", "InlineInput",
    FontWeight->"Bold"],
  " depends upon those functions, which are not mentioned explicitly in the \
save command.  ",
  StyleBox["Identify a suitable directory first!",
    FontSlant->"Italic"]
}], "ExerciseText"],

Cell[BoxData[
    \(\(SetDirectory[testdir];\)\)], "Input"],

Cell[BoxData[
    \(Save["\<zrot\>", {Z1rot, CV, TableZ1Rot, 
        Z1RotInterpolatingFunction}]\)], "Input"],

Cell["5. Verify that the file contains the desired information.", \
"ExerciseText"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(!! zrot\)\)], "Input"],

Cell["\<\
Z1rot[y_] := NSum[(2*j + 1)*Exp[(-j)*(j + 1)*y], {j, 0, Infinity}]
 
CV[T_, Z_] := D[U[T, Z], T]
 
U[T_, Z_] := F[T, Z] + T*S[T, Z]
 
F[T_, Z_] := (-N)*Subscript[k, B]*T*Log[Z[T]]
 
S[T_, Z_] := -D[F[T, Z], T]
 
TableZ1Rot = {{0.05, 1.}, {0.1, 1.0000000061834609}, 
     {0.15000000000000002, 1.000004858790377}, {0.2, 1.0001361997897553}, 
     {0.25, 1.0010063880724642}, {0.30000000000000004, 1.0038179117097876}, 
     {0.35000000000000003, 1.009895696708852}, {0.4, 1.020215370509514}, 
     {0.45, 1.0352389833733873}, {0.5, 1.0549776379922287}, 
     {0.6, 1.107248994118675}, {0.7, 1.1732453181499667}, 
     {0.8, 1.2490225591636712}, {0.9, 1.3314785778602378}, 
     {1., 1.4184426386310551}, {1.1, 1.5084740879357856}, 
     {1.2, 1.6006348631665925}, {1.3, 1.6943131150676127}, 
     {1.4, 1.7891037650765185}, {1.5, 1.8847324462389792}, 
     {1.6, 1.9810083295173986}, {1.7, 2.07779502848606}, 
     {1.8, 2.1749925113247643}, {1.9000000000000001, 2.272525665262398}, 
     {2., 2.370336904751403}, {2.1, 2.468381273809074}, 
     {2.2, 2.566623118206253}, {2.3000000000000003, 2.665033769002961}, 
     {2.4, 2.7635898936414214}, {2.5, 2.8622722982879925}, 
     {2.6, 2.9610650421166493}, {2.7, 3.0599547717073863}, 
     {2.8000000000000003, 3.158930213671361}, {2.9, 3.257981782923125}, 
     {3, 3.357101276748311}, {3.5, 3.853500607612988}, 
     {4., 4.350848418439135}, {4.5, 4.848813239207984}, 
     {5., 5.347202053734316}, {5.5, 5.845894796910737}, 
     {6., 6.34481284719952}, {6.5, 6.843902558068493}, 
     {7., 7.343126067236967}, {7.5, 7.842455887575912}, 
     {8., 8.341871580292752}, {8.5, 8.841357631361026}, 
     {9., 9.340902052231089}, {9.5, 9.84049543207012}, 
     {10., 10.340130280194153}}
 
Z1RotInterpolatingFunction = InterpolatingFunction[{{0.05, 10.}}, 
     {2, 0, True, Real, {5}, {0}}, {{0.05, 0.1, 0.15000000000000002, 0.2, 
       0.25, 0.30000000000000004, 0.35000000000000003, 0.4, 0.45, 0.5, 0.6, 
       0.7, 0.8, 0.9, 1., 1.1, 1.2, 1.3, 1.4, 1.5, 1.6, 1.7, 1.8, 
       1.9000000000000001, 2., 2.1, 2.2, 2.3000000000000003, 2.4, 2.5, 2.6, 
       2.7, 2.8000000000000003, 2.9, 3., 3.5, 4., 4.5, 5., 5.5, 6., 6.5, 7., 
       7.5, 8., 8.5, 9., 9.5, 10.}}, {{0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 
       12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 
       29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 
       46, 47, 48, 49}, {1., 1.0000000061834609, 1.000004858790377, 
       1.0001361997897553, 1.0010063880724642, 1.0038179117097876, 
       1.009895696708852, 1.020215370509514, 1.0352389833733873, 
       1.0549776379922287, 1.107248994118675, 1.1732453181499667, 
       1.2490225591636712, 1.3314785778602378, 1.4184426386310551, 
       1.5084740879357856, 1.6006348631665925, 1.6943131150676127, 
       1.7891037650765185, 1.8847324462389792, 1.9810083295173986, 
       2.07779502848606, 2.1749925113247643, 2.272525665262398, 
       2.370336904751403, 2.468381273809074, 2.566623118206253, 
       2.665033769002961, 2.7635898936414214, 2.8622722982879925, 
       2.9610650421166493, 3.0599547717073863, 3.158930213671361, 
       3.257981782923125, 3.357101276748311, 3.853500607612988, 
       4.350848418439135, 4.848813239207984, 5.347202053734316, 
       5.845894796910737, 6.34481284719952, 6.843902558068493, 
       7.343126067236967, 7.842455887575912, 8.341871580292752, 
       8.841357631361026, 9.340902052231089, 9.84049543207012, 
       10.340130280194153}}, {Automatic}]\
\>", "Print"]
}, Closed]],

Cell["\<\
6. Now remove all symbols related to the partition function and the \
thermodynamic relationships.\
\>", "ExerciseText"],

Cell[BoxData[
    \(Remove["\<*Z1*\>", T, S, U, F, CV]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Z1*\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(Information::"nomatch"\), \(\(:\)\(\ \)\), "\<\"No symbol \
matching \\!\\(\\\"*Z1*\\\"\\) found. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, \
ButtonData:>\\\"Information::nomatch\\\"]\\)\"\>"}]], "Message"]
}, Closed]],

Cell["7. Retrieve the stored definitions.", "ExerciseText"],

Cell[BoxData[
    \(\(<< "\<zrot\>";\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Z1rot\)\)], "Input"],

Cell["Global`Z1rot", "Print",
  CellTags->"Info3350830501-3442347"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {GridBox[{
                {\(Z1rot[y_] := 
                    NSum[\((2\ j + 
                            1)\)\ \[ExponentialE]\^\(\(-j\)\ \((j + 1)\)\ \
y\), {j, 0, \[Infinity]}]\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "Z1rot"],
      Editable->False]], "Print",
  CellTags->"Info3350830501-3442347"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?CV\)\)], "Input"],

Cell["Global`CV", "Print",
  CellTags->"Info3350830503-7920564"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {GridBox[{
                {\(CV[T_, Z_] := \[PartialD]\_T U[T, Z]\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "CV"],
      Editable->False]], "Print",
  CellTags->"Info3350830503-7920564"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?F\)\)], "Input"],

Cell["Global`F", "Print",
  CellTags->"Info3350830504-7344348"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {GridBox[{
                {\(F[T_, Z_] := \(-N\)\ "Subscript"[k, B]\ T\ Log[Z[T]]\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "F"],
      Editable->False]], "Print",
  CellTags->"Info3350830504-7344348"]
}, Closed]],

Cell["\<\
8. Check that the interpolating function does its job by using \
basic thermodynamic relationships to compute the heat capacity for this \
system.\
\>", "ExerciseText"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[
        Evaluate[CV[T, Z1RotInterpolatingFunction]\/\(N\ k\_B\)], {T, 0.05, 
          5}, \n\tPlotLabel \[Rule] "\<Rotational Heat Capacity\>", \n\t
        PlotRange \[Rule] {{0, 5}, {0, 1.2}}, \n\tGridLines \[Rule] None, 
        Frame \[Rule] True, \n\t
        FrameLabel \[Rule] {\*"\"\<T/\!\(T\_rot\)\>\"", \*"\"\<\!\(C\_V\)/(N \
\!\(k\_B\))\>\""}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.77556e-017 0.2 -2.18738e-017 0.515028 [
[.2 -0.0125 -3 -9 ]
[.2 -0.0125 3 0 ]
[.4 -0.0125 -3 -9 ]
[.4 -0.0125 3 0 ]
[.6 -0.0125 -3 -9 ]
[.6 -0.0125 3 0 ]
[.8 -0.0125 -3 -9 ]
[.8 -0.0125 3 0 ]
[1 -0.0125 -3 -9 ]
[1 -0.0125 3 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -17.5313 -21.5625 ]
[.5 -0.0125 17.5313 -9 ]
[-0.0125 .10301 -18 -4.5 ]
[-0.0125 .10301 0 4.5 ]
[-0.0125 .20601 -18 -4.5 ]
[-0.0125 .20601 0 4.5 ]
[-0.0125 .30902 -18 -4.5 ]
[-0.0125 .30902 0 4.5 ]
[-0.0125 .41202 -18 -4.5 ]
[-0.0125 .41202 0 4.5 ]
[-0.0125 .51503 -6 -4.5 ]
[-0.0125 .51503 0 4.5 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -30.5625 -27.75 ]
[-0.01875 .30902 -18 27.75 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -74 0 ]
[.5 .63053 74 12.5625 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 0 m
.2 .00625 L
s
[(1)] .2 -0.0125 0 1 Mshowa
.4 0 m
.4 .00625 L
s
[(2)] .4 -0.0125 0 1 Mshowa
.6 0 m
.6 .00625 L
s
[(3)] .6 -0.0125 0 1 Mshowa
.8 0 m
.8 .00625 L
s
[(4)] .8 -0.0125 0 1 Mshowa
1 0 m
1 .00625 L
s
[(5)] 1 -0.0125 0 1 Mshowa
.125 Mabswid
.04 0 m
.04 .00375 L
s
.08 0 m
.08 .00375 L
s
.12 0 m
.12 .00375 L
s
.16 0 m
.16 .00375 L
s
.24 0 m
.24 .00375 L
s
.28 0 m
.28 .00375 L
s
.32 0 m
.32 .00375 L
s
.36 0 m
.36 .00375 L
s
.44 0 m
.44 .00375 L
s
.48 0 m
.48 .00375 L
s
.52 0 m
.52 .00375 L
s
.56 0 m
.56 .00375 L
s
.64 0 m
.64 .00375 L
s
.68 0 m
.68 .00375 L
s
.72 0 m
.72 .00375 L
s
.76 0 m
.76 .00375 L
s
.84 0 m
.84 .00375 L
s
.88 0 m
.88 .00375 L
s
.92 0 m
.92 .00375 L
s
.96 0 m
.96 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -78.5313 -25.5625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20.5625 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 12.813 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 12.813 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(T) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
69.000 12.813 moveto
(\\220) show
75.000 12.813 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(T) show
81.000 14.313 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 7.125 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(rot) show
94.063 12.813 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
[(0.2)] -0.0125 .10301 1 0 Mshowa
0 .20601 m
.00625 .20601 L
s
[(0.4)] -0.0125 .20601 1 0 Mshowa
0 .30902 m
.00625 .30902 L
s
[(0.6)] -0.0125 .30902 1 0 Mshowa
0 .41202 m
.00625 .41202 L
s
[(0.8)] -0.0125 .41202 1 0 Mshowa
0 .51503 m
.00625 .51503 L
s
[(1)] -0.0125 .51503 1 0 Mshowa
.125 Mabswid
0 .02575 m
.00375 .02575 L
s
0 .0515 m
.00375 .0515 L
s
0 .07725 m
.00375 .07725 L
s
0 .12876 m
.00375 .12876 L
s
0 .15451 m
.00375 .15451 L
s
0 .18026 m
.00375 .18026 L
s
0 .23176 m
.00375 .23176 L
s
0 .25751 m
.00375 .25751 L
s
0 .28327 m
.00375 .28327 L
s
0 .33477 m
.00375 .33477 L
s
0 .36052 m
.00375 .36052 L
s
0 .38627 m
.00375 .38627 L
s
0 .43777 m
.00375 .43777 L
s
0 .46353 m
.00375 .46353 L
s
0 .48928 m
.00375 .48928 L
s
0 .54078 m
.00375 .54078 L
s
0 .56653 m
.00375 .56653 L
s
0 .59228 m
.00375 .59228 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -113.031 -10.2813 Mabsadd m
1 1 Mabs scale
currentpoint translate
88.75 10.2813 translate
90 rotate
-88.75 -10.2813 translate
0 20.5625 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 12.813 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
%%BeginResource: font Courier-MISO
%%BeginFont: Courier-MISO
/Courier /Courier-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 12.813 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(C) show
69.000 14.313 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 7.125 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(V) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
73.813 12.813 moveto
(\\220) show
79.813 12.813 moveto
(H) show
85.813 12.813 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(N) show
97.813 12.813 moveto
(k) show
103.813 14.313 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 7.125 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(B) show
%%IncludeResource: font Mathematica2Mono
%%IncludeFont: Mathematica2Mono
/Mathematica2Mono findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
108.500 12.813 moveto
(L) show
114.500 12.813 moveto
%%IncludeResource: font Courier-MISO
%%IncludeFont: Courier-MISO
/Courier-MISO findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .61178 m
0 .61803 L
s
.2 .61178 m
.2 .61803 L
s
.4 .61178 m
.4 .61803 L
s
.6 .61178 m
.6 .61803 L
s
.8 .61178 m
.8 .61803 L
s
.125 Mabswid
.04 .61428 m
.04 .61803 L
s
.08 .61428 m
.08 .61803 L
s
.12 .61428 m
.12 .61803 L
s
.16 .61428 m
.16 .61803 L
s
.24 .61428 m
.24 .61803 L
s
.28 .61428 m
.28 .61803 L
s
.32 .61428 m
.32 .61803 L
s
.36 .61428 m
.36 .61803 L
s
.44 .61428 m
.44 .61803 L
s
.48 .61428 m
.48 .61803 L
s
.52 .61428 m
.52 .61803 L
s
.56 .61428 m
.56 .61803 L
s
.64 .61428 m
.64 .61803 L
s
.68 .61428 m
.68 .61803 L
s
.72 .61428 m
.72 .61803 L
s
.76 .61428 m
.76 .61803 L
s
.84 .61428 m
.84 .61803 L
s
.88 .61428 m
.88 .61803 L
s
.92 .61428 m
.92 .61803 L
s
.96 .61428 m
.96 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .02575 m
1 .02575 L
s
.99625 .0515 m
1 .0515 L
s
.99625 .07725 m
1 .07725 L
s
.99625 .12876 m
1 .12876 L
s
.99625 .15451 m
1 .15451 L
s
.99625 .18026 m
1 .18026 L
s
.99625 .23176 m
1 .23176 L
s
.99625 .25751 m
1 .25751 L
s
.99625 .28327 m
1 .28327 L
s
.99625 .33477 m
1 .33477 L
s
.99625 .36052 m
1 .36052 L
s
.99625 .38627 m
1 .38627 L
s
.99625 .43777 m
1 .43777 L
s
.99625 .46353 m
1 .46353 L
s
.99625 .48928 m
1 .48928 L
s
.99625 .54078 m
1 .54078 L
s
.99625 .56653 m
1 .56653 L
s
.99625 .59228 m
1 .59228 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -135 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20.5625 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 12.813 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 12.813 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Rotational) show
129.000 12.813 moveto
(Heat) show
159.000 12.813 moveto
(Capacity) show
207.000 12.813 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.01 6e-005 m
.01122 9e-005 L
.01233 .0001 L
.01362 .0001 L
.01483 .0001 L
.01595 8e-005 L
.01699 6e-005 L
.01933 0 L
s
.01933 0 m
.01934 0 L
s
.02679 0 m
.02704 1e-005 L
.02834 .00015 L
.02955 .00034 L
.03084 .00063 L
.03224 .00107 L
.03357 .00162 L
.03478 .00227 L
.03709 .00395 L
.03825 .00507 L
.03953 .00652 L
.04218 .0107 L
.04503 .01705 L
.04771 .02489 L
.05016 .03375 L
.05576 .06045 L
.06093 .09295 L
.07265 .18567 L
.09396 .36606 L
.1033 .43014 L
.11336 .4824 L
.11912 .5051 L
.12165 .51529 L
.12441 .52329 L
.12946 .53522 L
.13478 .54473 L
.13741 .54863 L
.13814 .54963 L
.1389 .55068 L
.14027 .55483 L
.14155 .55636 L
.14274 .55766 L
.14546 .5602 L
.1479 .562 L
.15014 .56327 L
.15137 .56383 L
.15272 .56433 L
.15389 .56469 L
.15516 .56498 L
.15593 .56512 L
.15664 .56521 L
.15732 .56528 L
.15804 .56534 L
.15928 .56537 L
.16064 .56537 L
.1614 .56539 L
.16209 .56539 L
.16289 .56537 L
.16365 .56534 L
.165 .56523 L
Mistroke
.16576 .56514 L
.16646 .56505 L
.16888 .56463 L
.17155 .564 L
.17694 .5623 L
.21879 .54459 L
.23803 .53789 L
.25906 .53242 L
.27993 .52852 L
.30188 .52561 L
.3223 .52366 L
.34409 .52212 L
.36632 .52096 L
.38728 .52011 L
.42732 .51894 L
.44803 .5185 L
.46991 .5181 L
.50935 .51755 L
.55134 .51711 L
.59175 .51679 L
.63471 .51654 L
.6562 .51641 L
.6761 .51627 L
.68605 .51618 L
.69092 .51613 L
.69539 .51608 L
.69768 .51606 L
.69882 .51604 L
.70008 .51621 L
.70135 .51621 L
.7027 .5162 L
.70514 .51619 L
.71591 .51616 L
.73658 .51611 L
.75901 .51606 L
.7695 .51604 L
.78061 .51602 L
.7901 .51599 L
.79269 .51599 L
.79549 .51598 L
.79684 .51597 L
.79813 .51597 L
.79928 .51597 L
.80054 .5159 L
.80121 .5159 L
.80192 .5159 L
.8032 .5159 L
.80607 .51589 L
.81128 .51587 L
.821 .51585 L
Mistroke
.84281 .5158 L
.8823 .51573 L
.92434 .51566 L
.96481 .5156 L
1 .51557 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00l`3oool00`3ogi`0B0000000
000200000004001d_`3oool0oooo0;md00<0000000D007Bo0?ooo`3oool0oooo09a8000200000003
07Boo`3oool0oooo0:L0oooo003R0?ooo`030?oo_`1dM;l0oooo0100oooo00<0_gA809cOo`3oool0
0P3oool00`3ogi`0B7Bo0?ooo`020?ooo`090;mdB02Lgol0oooo0?oOW018B9`0oooo0?oOW01804P0
W=oo0:P0oooo003I0?ooo`030?oOW018000000000080000000<0001d0;ooo`3oool00P3oool00`3O
W7@0W=oo0?ooo`050?ooo`030?oOW018000000000080000000<0001d0;ooo`3oool00P3oool00`2o
M4P0W=oo0?ooo`020?ooo`030?noM018W=l0oooo0080oooo00D0gia807Boo`3oool0omnL04Qd_`2/
0?ooo`00f`3oool00`3ogi`0B4RL0?ooo`050?ooo`030?oo_`1dM;l0oooo00L0oooo00<0omnL04Q8
W03oool01@3oool00`2oM000001d0;ooo`020?ooo`0:0?oOW01807@0_ooo0?ooo`3ookl0M01809cO
o`3oool0omnL04Qd_j`0oooo003K0?ooo`030?oOW018B9`0oooo00H0oooo00<0giad09cOo`3oool0
1P3oool00`3ogi`0B4RL0?ooo`040?ooo`0309a8000004P0W9ad0080000000<0M;oo0?ooo`3o_g@0
0`0000000`1d_ol0oooo0?noM0040000000304RLg`3oool0oooo0:L0oooo003K0?ooo`030?oOW018
B9`0oooo00H0oooo00<0omnL07BLg`3oool01P3oool00`3ogi`0B4RL0?ooo`0B0?ooo`030?oOW018
M;l0oooo0:/0oooo003H0?ooo`090?oOW018M;l0oooo0?oOW018B9`0oooo0?oo_`1d07@0_ooo00@0
oooo00<0_gAd0;ooo`3oool00P3oool02@3ogi`0B7Bo0?ooo`3ogi`0B4RL0?ooo`3ookl0M01d0;oo
o`2m0?ooo`00f03oool02@3ogi`0B7Bo0?ooo`3ogi`0B4RL0?ooo`3ookl0M01d0;ooo`040?ooo`03
0?oOW01dW=l0oooo0080oooo00T0omnL04Qd_`3oool0omnL04Q8W03oool0oono07@0M02oool0_@3o
ool00=P0oooo00T0omnL04Qd_`3oool0omnL04Q8W03oool0oono07@0M02oool01@3oool03@2oM7@0
_ooo0?ooo`3oool0omnL04Qd_`3oool0omnL04Q8W03oool0oono07@0M02oool0_@3oool00=P0oooo
00<0omnL04P000000000100000000`0007@0_ooo0?ooo`040?ooo`060?oOW01dW=l0oooo0?ooo`3o
gi`0B0001@0000000`0007@0_ooo0?ooo`2l0?ooo`00i`3oool00`2oW9`0gooo0?ooo`360?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
Q`3oool00`3o_g@000000000000300000003001d_`3oool0oooo03T0oooo00<0oono07@000000000
100000000`1d_ol0oooo0?ooo`0k0?ooo`050;md00000000000000000018W=l0?P3oool01@3ogi`0
B000000000000000M;oo03d0oooo00D0_g@0000000000000000007Boo`0H0?ooo`00RP3oool00`2L
B4P0W=oo0?ooo`0m0?ooo`030;md000007@0_ooo0080oooo00<0giad09cOo`3oool0>`3oool01`2o
M00007Bo0?ooo`3oool0oono07@00018W=l0?`3oool00`2oM000M;oo0?ooo`0l0?ooo`070;md0000
M;l0oooo0?ooo`3ogi`0B00007Boo`0G0?ooo`00RP3oool00`2LB4P0W=oo0?ooo`0n0?ooo`030?no
M00004P0W=oo04@0oooo00<0okmd001d_`3oool0?P3oool00`2oM000M;oo0?ooo`110?ooo`030?no
M000M;l0oooo01H0oooo002:0?ooo`0309a8B02Lgol0oooo03l0oooo00<0omnL04P00018W=l0@P3o
ool00`3ogi`0B01809cOo`0k0?ooo`030=nLB0000000000000<0000000<0M;oo0?ooo`3oool0?`3o
ool00`3o_g@007Bo0?ooo`0F0?ooo`00RP3oool00`2LB4P0W=oo0?ooo`110?ooo`0309a80018W=l0
oooo03h0oooo00@0gia80000000004P0W=oo?03oool01P3o_g@0001d0;ooo`3oool0_g@007Bood80
oooo00<0W4P007Boo`3oool05P3oool008X0oooo00<0W4Q809cOo`3oool0@P3oool00`2LB7@0_ooo
0?ooo`0o0?ooo`030?oOW01804P0W=oo03`0oooo00D0gia8001d_`3oool0_g@007Boo`0m0?ooo`03
0?oo_`1d000000000080000000<0B9cO0?ooo`3oool05P3oool008X0oooo00<0W4Q809cOo`3oool0
?@3oool00`3OW4P0M;oo0?ooo`020?ooo`030;md001d_ol0oooo0400oooo00<0_g@007Boo`3oool0
?03oool0102LB000B9cO0;md001d_olm0?ooo`030?oo_`1d07@0_ooo01/0oooo00270?ooo`050=nL
B000000004RL09a8B02Lgol0?P3oool01`3o_g@0001d0;ooo`3oool0okmd0000B02Lgol0?03oool0
1`3o_g@0001809cOo`3oool0omnL04P0B02Lgol0?@3oool0103ogi`0B01804Q8001d_olm0?ooo`03
0?oo_`1d07@0_ooo01/0oooo00290?ooo`030=nLB00004P0W=oo03l0oooo00D0okmd000000000000
001809cOo`0n0?ooo`050?oOW018000000000000001d_ol0?`3oool00`3o_g@0000007Boo`0m0?oo
o`030?oo_`1d000000000080000000<0001d0;ooo`3oool05P3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool004P0ooooo`00001@000001T0oooo00180?ooo`030000003oool0oooo00P0oooo0`00000=
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0G
0?ooo`00B03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo03<0oooo00<000000?oo
o`3oool0?`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0400oooo00<000000?oo
o`3oool0@03oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo0?l0oooo?`3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo0?l0oooo?P3oool00`000000oooo0?ooo`0G0?ooo`00B03o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0?l0oooo?P3oool00`000000oooo0?oo
o`0G0?ooo`00B03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0oooo?@3oool0
0`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
0?l0oooo?@3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo0?l0oooo?03oool00`000000oooo0?ooo`0G0?ooo`00B03oool2000000h0oooo
00<000000?ooo`3oool0o`3ooolj0?ooo`<000006@3oool004P0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`3o0?oooc/0oooo00<000000?ooo`3oool05`3oool004P0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?oooc/0oooo00<000000?ooo`3oool05`3oool0
04P0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?oooc/0oooo00<000000?oo
o`3oool05`3oool004P0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?oooc/0
oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`3o0?oooc/0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`3o0?oooc/0oooo00<000000?ooo`3oool05`3oool004P0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`3o0?ooocX0oooo00<000000?ooo`3oool05`3oool0
04P0oooo0P00000@0?ooo`030000003oool0oooo0?l0oooo>03oool3000001T0oooo00180?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3ooolj0?ooo`030000003oool0oooo01L0
oooo00180?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3ooolj0?ooo`030000
003oool0oooo01L0oooo00180?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3o
oolj0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0o`3ooolj0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0o`3ooolj0?ooo`030000003oool0oooo01L0oooo00180?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo01L0
oooo00180?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000
003oool0oooo01L0oooo00180?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3o
ooli0?ooo`030000003oool0oooo01L0oooo00180?ooo`8000004@3oool00`000000oooo0?ooo`3o
0?ooocL0oooo0`00000I0?ooo`00B03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
0?l0oooo>@3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0?l0oooo>03oool00`000000oooo0?ooo`0G0?ooo`00B03o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0?l0oooo>03oool00`000000oooo0?oo
o`0G0?ooo`00:`3oool01@3o_g@00000000000000000M;oo00D0oooo00<0_g@000000000M;l00`3o
ool00`3ookl0M000000000040000000307Boo`3oool0oooo00<0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`3o0?ooocP0oooo00<000000?ooo`3oool05`3oool002/0oooo00H0W4Qd
0;ooo`3oool0oono07@0M02oool40?ooo`030;md0000000007Bo00@0oooo00<0_g@00000M02oool0
0P3oool00`3OW7@0W=oo0?ooo`040?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
o`3ooolh0?ooo`030000003oool0oooo01L0oooo000Z0?ooo`030?noM000M;l0oooo0080oooo00<0
gia804RLg`3oool02`3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool0o`3ooolh0?ooo`030000003oool0oooo01L0oooo000Z0?ooo`030=nLB018W=l0
oooo0080oooo00<0okmd001d_`3oool0303oool00`3ogi`0B00004RLg`070?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0o`3ooolh0?ooo`030000003oool0oooo01L0oooo000Z0?oo
o`030=nLB018W=l0oooo0080oooo00<0okmd001d_`3oool03P3oool00`2LB000B9cO0?ooo`050?oo
o`<000004P3oool00`000000oooo0?ooo`3o0?ooocD0oooo0`00000I0?ooo`00:P3oool00`3OW4P0
B9cO0?ooo`020?ooo`030?noM000M;l0oooo00l0oooo00<0W4Qd0;ooo`3oool0103oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo0?l0oooo=`3oool00`000000oooo0?ooo`0G0?ooo`00
:P3oool00`3o_g@0B9cO0?ooo`020?ooo`030=nLB018W=l0oooo00X0oooo00<0gia807Boo`3oool0
0P3oool00`2oM000M;oo0?ooo`040?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
o`3ooolg0?ooo`030000003oool0oooo01L0oooo000[0?ooo`0609a8001d_ol0oooo0?oOW01807@0
_ooo2`3oool01`3o_g@0001d0;ooo`3oool0okmd0000B02Lgol01@3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo0?l0oooo=`3oool00`000000oooo0?ooo`0G0?ooo`00:`3oool01@3o
okl0M0000000000004P0W=oo00d0oooo00D0okmd000000000000001809cOo`060?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0o`3ooolg0?ooo`030000003oool0oooo01L0oooo0018
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0o`3ooolg0?ooo`030000003oool0
oooo01L0oooo00180?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0o`3ooolg0?oo
o`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo0180oooo00<000000?ooo`3o
ool0o`3ooolg0?ooo`030000003oool0oooo01L0oooo00180?ooo`800000503oool00`000000oooo
0?ooo`3o0?oooc@0oooo0`00000I0?ooo`00B03oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo0?l0oooo=P3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo0?l0oooo=P3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo0?l0oooo=P3oool00`000000oooo0?ooo`0G0?oo
o`00B03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0?l0oooo=P3oool00`000000
oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0?l0oooo
=P3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo0?l0oooo=P3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo0?l0oooo=P3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo0?l0oooo=@3oool00`000000oooo0?ooo`0G0?oo
o`00B03oool2000001D0oooo00<000000?ooo`3oool0o`3ooolc0?ooo`<000006@3oool004P0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o0?ooocD0oooo00<000000?ooo`3oool0
5`3oool004P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o0?ooocD0oooo00<0
00000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o
0?ooocD0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`3o0?ooocD0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3o
ool0503oool00`000000oooo0?ooo`3o0?ooocD0oooo00<000000?ooo`3oool05`3oool004P0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o0?ooocD0oooo00<000000?ooo`3oool0
5`3oool004P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3o0?oooc@0oooo00<0
00000?ooo`3oool05`3oool004P0oooo0P00000F0?ooo`030000003oool0oooo0?l0oooo<P3oool3
000001T0oooo00180?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0o`3ooold0?oo
o`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0o`3ooold0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool0o`3ooold0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool0o`3ooold0?ooo`030000003oool0oooo01L0oooo000[
0?ooo`050?noM000000000000000001d_ol01@3oool00`2oM0000000001d_`060?ooo`050?oOW018
000000000000001d_ol01@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo
=03oool00`000000oooo0?ooo`0G0?ooo`00:`3oool01P2LB7@0_ooo0?ooo`3ookl0M01d0;ooo`@0
oooo00<0_g@000000000M;l0203oool00`2oM000M;oo0?ooo`050?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0o`3ooold0?ooo`030000003oool0oooo01L0oooo000I0?ooo`030?oo
g`2LB00000000080000000<0001809cOo`3oool02@3oool00`3o_g@007Bo0?ooo`020?ooo`030=nL
B018W=l0oooo00h0oooo00<0_g@007Boo`3oool01@3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo0?l0oooo<`3oool00`000000oooo0?ooo`0G0?ooo`006@3oool00`2oM4P0W=oo0?oo
o`020?ooo`030?oog`2LB4P0W=oo00T0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?oo
o`0:0?ooo`030=nLB0000000000000<0000000<0M;oo0?ooo`3oool00`3oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo0?l0oooo<`3oool00`000000oooo0?ooo`0G0?ooo`00603oool0
0`3ogi`0M9cO0?ooo`040?ooo`030?oog`2LW;l0oooo00P0oooo00<0gia804RLg`3oool00P3oool0
0`3o_g@007Bo0?ooo`0:0?ooo`060?noM00007@0_ooo0?ooo`2oM000M;oo1P3oool3000001H0oooo
00<000000?ooo`3oool0o`3ooola0?ooo`<000006@3oool001P0oooo00<0okmd07Boo`3oool01@3o
ool00`3OW9`0_ooo0?ooo`070?ooo`030=nLB018W=l0oooo0080oooo00<0okmd001d_`3oool02`3o
ool01@3OW4P007Bo0?ooo`2oM000M;oo00H0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`3o0?oooc<0oooo00<000000?ooo`3oool05`3oool001P0oooo00<0omnL09boo`3oool01@3o
ool00`2oM7@0_ooo0?ooo`070?ooo`030?noM018W=l0oooo0080oooo00<0gia804RLg`3oool0303o
ool0102LB000B9cO0;md001d_ol60?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
o`3ooolc0?ooo`030000003oool0oooo01L0oooo000I0?ooo`030;mdM02oool0oooo00<0oooo00<0
omnL04Qd_`3oool02@3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;ooo`d0oooo00@0omnL04P0B018
B000M;oo1P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0?l0oooo<`3oool00`00
0000oooo0?ooo`0G0?ooo`00603oool0103o_g@000000000B02Lgol20?ooo`030?oo_`1dB9`0gooo
00X0oooo00D0oono07@000000000001809cOo`0?0?ooo`030?noM0000000M;oo00H0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?oooc<0oooo00<000000?ooo`3oool05`3oool0
04P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?oooc<0oooo00<000000?oo
o`3oool05`3oool001`0oooo00<0giad09cOo`3oool0:@3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo0?l0oooo<P3oool00`000000oooo0?ooo`0G0?ooo`00703oool01P3OW7@0B4P0
00000000000004RL0=ooobH0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`3o0?oo
oc80oooo00<000000?ooo`3oool05`3oool00200oooo00@0omnL04P00018M;l0gooo903oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo0?l0oooo<P3oool00`000000oooo0?ooo`0G0?oo
o`00803oool00`3OW4P00000001d_`0U0?ooo`800000603oool00`000000oooo0?ooo`3o0?oooc00
oooo0`00000I0?ooo`007P3oool00`2oM000000004RLg`0W0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0o`3ooolb0?ooo`030000003oool0oooo01L0oooo000L0?ooo`030=nLB000
0000M;oo02T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`3o0?oooc80oooo00<0
00000?ooo`3oool05`3oool001`0oooo00<0gibL0;ooo`3oool0:@3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo0?l0oooo<P3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo0?l0oooo<P3oool00`000000oooo0?ooo`0G0?oo
o`00B03oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0?l0oooo<P3oool00`000000
oooo0?ooo`0G0?ooo`00803oool00`3o_g@0001809cOo`0U0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0o`3ooola0?ooo`030000003oool0oooo01L0oooo000M0?ooo`040?oo_`1d
00000000001d_bL0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`3o0?oooc40oooo
00<000000?ooo`3oool05`3oool001/0oooo00@0gia80000000004P0W=oo:@3oool2000001T0oooo
00<000000?ooo`3oool0o`3oool_0?ooo`<000006@3oool001P0oooo00@0omnL04P000000000B9cO
;03oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0?l0oooo<@3oool00`000000oooo
0?ooo`0G0?ooo`005`3oool00`3o_g@004RL0=ooo`0^0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0o`3ooola0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0o`3ooola0?ooo`030000003oool0oooo01L0oooo00180?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0o`3ooola0?ooo`030000003oool0oooo
01L0oooo00180?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0o`3ooola0?ooo`03
0000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
o`3ooola0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool0o`3oool`0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool0o`3oool`0?ooo`030000003oool0oooo01L0oooo000J0?oo
o`030?oOW0180000000000<0000000<007Bo0?ooo`3oool09@3oool2000001X0oooo00<000000?oo
o`3oool0o`3oool^0?ooo`<000006@3oool001P0oooo00@0gia80000000007@0_ooo103oool0103O
W4P000000000M02ooolT0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0o`3oool`
0?ooo`030000003oool0oooo01L0oooo000G0?ooo`030?noM000M;l0oooo00P0oooo00<0omnL04Q8
W03Oool08`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0?l0oooo<03oool00`00
0000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0?l0
oooo<03oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo0?l0oooo<03oool00`000000oooo0?ooo`0G0?ooo`00:`3oool01@3o_g@000000000
00000000M;oo00D0oooo00<0_g@000000000M;l01P3oool01@2oM00000000000000007@0_ooo00D0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`3o0?oooc00oooo00<000000?ooo`3o
ool05`3oool001P0oooo00<0okmd07Boo`3oool01@3oool00`2oM7@0_ooo0?ooo`080?ooo`0609a8
M02oool0oooo0?oo_`1d07@0_ooo103oool00`2oM0000000001d_`050?ooo`060=nLB000M;l0oooo
0?ooo`3o_g@007Bo1@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0?l0oooo;`3o
ool00`000000oooo0?ooo`0G0?ooo`00603oool00`3o_g@0000000000005000000030000M02oool0
oooo00L0oooo00<0okmd001d_`3oool00P3oool00`3OW4P0B9cO0?ooo`0:0?ooo`030?oo_`1d0000
M;oo0080oooo00<0oono07A8W03oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo0?l0oooo;`3oool00`000000oooo0?ooo`0G0?ooo`006@3oool00`2oM000001809cOo`040?oo
o`030;mdM02oool0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0:0?oo
o`070?noM000B4P0M01d0;ooo`3oool0gia8001d_`050?ooo`<000006P3oool00`000000oooo0?oo
o`3o0?ooobd0oooo0`00000I0?ooo`006P3oool0103ooml0W4P00000B02Lgol<0?ooo`030=nLB018
W=l0oooo0080oooo00<0okmd001d_`3oool02P3oool00`3ogi`0B4RL0?noM002000000030018W03o
ool0oooo00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?ooobl0oooo00<0
00000?ooo`3oool05`3oool001`0oooo00@0oooO09a8000004P0W=oo2P3oool00`3OW4P0B9cO0?oo
o`020?ooo`030?noM000M;l0oooo00X0oooo00<0oono07@0M02oool02@3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo0?l0oooo;`3oool00`000000oooo0?ooo`0G0?ooo`00603oool0
0`3o_g@0M;oo0?ooo`030?ooo`030?oog`2LB000M;oo00T0oooo00<0okmd04RLg`3oool00P3oool0
0`3OW4P0B9cO0?ooo`0;0?ooo`030;md0018W=l0oooo00P0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`3o0?ooobl0oooo00<000000?ooo`3oool05`3oool001P0oooo00<0okmd0000
000000001@0000000`0007@0_ooo0?ooo`080?ooo`0609a8001d_ol0oooo0?oOW01807@0_ooo303o
ool00`3ookl0M000001d_`080?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0o`3o
ool_0?ooo`030000003oool0oooo01L0oooo000H0?ooo`030?noM01d_ol0oooo0100oooo00D0oono
07@000000000001809cOo`0?0?ooo`040;md0000000000000018W0D0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`3o0?ooobl0oooo00<000000?ooo`3oool05`3oool004P0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?ooobl0oooo00<000000?ooo`3oool05`3o
ool004P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`3o0?ooobh0oooo00<00000
0?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`3o0?oo
obh0oooo00<000000?ooo`3oool05`3oool004P0oooo0P00000L0?ooo`030000003oool0oooo0?l0
oooo;03oool3000001T0oooo00180?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
o`3oool^0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool0o`3oool^0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool0o`3oool^0?ooo`030000003oool0oooo01L0oooo00180?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool0o`3oool^0?ooo`030000003oool0oooo
01L0oooo00180?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0o`3oool^0?ooo`03
0000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0
o`3oool]0?ooo`030000003oool0oooo01L0oooo000G0?ooo`030?oo_`1dM;l0oooo00H0oooo00<0
_gAd0;ooo`3oool09@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0?l0oooo;@3o
ool00`000000oooo0?ooo`0G0?ooo`005`3oool00`3ookl0M00000000006000000030000M02oool0
oooo02D0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`3o0?ooobd0oooo00<00000
0?ooo`3oool05`3oool001`0oooo00<0oono07@0M02oool0:@3oool2000001d0oooo00<000000?oo
o`3oool0o`3oool[0?ooo`<000006@3oool001`0oooo00<0_g@000000000M;l0:@3oool00`000000
oooo0?ooo`0L0?ooo`030000003oool0oooo0?l0oooo;@3oool00`000000oooo0?ooo`0G0?ooo`00
6P3oool0203OW7@0W9ad04RLg`3oool0okmd0000M01dM7@0_ooo9P3oool00`000000oooo0?ooo`0L
0?ooo`030000003oool0oooo0?l0oooo;@3oool00`000000oooo0?ooo`0G0?ooo`006P3oool00`3O
W4P004RL0=ooo`020?ooo`030?oo_`1d07@0_ooo02H0oooo00<000000?ooo`3oool0703oool00`00
0000oooo0?ooo`3o0?ooobd0oooo00<000000?ooo`3oool05`3oool001X0oooo00<0giad09cOo`3o
ool00`3oool00`2oM7@0_ooo0?ooo`0U0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0o`3oool]0?ooo`030000003oool0oooo01L0oooo000J0?ooo`030=nLM02Lgol0oooo02/0oooo
00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooob`0oooo00<000000?ooo`3oool0
5`3oool001`0oooo00<0gibL0;ooo`3oool00`3oool00`3OW9`0_ooo0?ooo`0S0?ooo`030000003o
ool0oooo01d0oooo00<000000?ooo`3oool0o`3oool/0?ooo`030000003oool0oooo01L0oooo000L
0?ooo`030=nLB0000000000000<0000000<0001d0;ooo`3oool08`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo0?l0oooo;03oool00`000000oooo0?ooo`0G0?ooo`00703oool0203O
W9`0_ooo0?ooo`2oW9`0gooo0?ooo`3OW9`0_ooo903oool2000001h0oooo00<000000?ooo`3oool0
o`3ooolZ0?ooo`<000006@3oool001`0oooo00P0gibL0;ooo`3oool0_ibL0=ooo`3oool0gibL0;oo
ob@0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooob`0oooo00<000000?oo
o`3oool05`3oool001`0oooo00P0giad09cOo`3ookl0M01d0;ooo`3oool0_gAd0;ooob@0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooob`0oooo00<000000?ooo`3oool05`3o
ool001`0oooo00L0oono07@0000007@0_ono07@000000000M;oo02D0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`3o0?ooob`0oooo00<000000?ooo`3oool05`3oool004P0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooob`0oooo00<000000?ooo`3oool05`3o
ool002/0oooo00D0okmd000000000000000007Boo`050?ooo`030;md0000000007Bo00D0oooo00D0
_g@00000000000000000001d_`060?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
o`3oool[0?ooo`030000003oool0oooo01L0oooo000[0?ooo`0609a8M02oool0oooo0?oo_`1d07@0
_ooo103oool00`2oM0000000001d_`040?ooo`030?noM000M;l0oooo0080oooo00<0_gA809cOo`3o
ool0103oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000
oooo0?ooo`0G0?ooo`005`3oool00`3o_g@007Bo0?ooo`080?ooo`030?oOW018B9`0gooo00D0oooo
00<0okmd001d_`3oool00P3oool00`3OW4P0B9cO0?ooo`0:0?ooo`030=nLB018W=l0oooo0080oooo
00<0_g@007Boo`3oool0103oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0?l0oooo
:`3oool00`000000oooo0?ooo`0G0?ooo`00603oool0103OW4P000000018W03Oool40?ooo`040?no
M0000000001d0;ooo`H0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0:0?ooo`03
0?noM000B9`0oooo0080oooo00<0W4Q809cOo`3oool0103oool00`000000oooo0?ooo`0N0?ooo`03
0000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0G0?ooo`006P3oool00`3ogi`0B000
0000000300000003001d_`3oool0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo
0?ooo`0;0?ooo`050=nLB0000000000000000018W=l01P3oool3000001h0oooo00<000000?ooo`3o
ool0o`3ooolY0?ooo`<000006@3oool002X0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo
0?ooo`0:0?ooo`030?oo_`1dB9`0oooo0080oooo00<0W4Qd0;ooo`3oool0103oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0G0?ooo`00:P3o
ool00`3o_g@0B9cO0?ooo`020?ooo`030=nLB018W=l0oooo00X0oooo00<0okmd04RLg`3oool00P3o
ool00`2oM000M;oo0?ooo`040?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0o`3o
ool[0?ooo`030000003oool0oooo01L0oooo000[0?ooo`0609a8001d_ol0oooo0?oOW01807@0_ooo
2`3oool01`3ogi`0B01d0;ooo`3oool0omnL04P0M02oool01@3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo0?l0oooo:P3oool00`000000oooo0?ooo`0G0?ooo`00:`3oool01@3ookl0
M0000000000004P0W=oo00d0oooo00D0okmd000000000000001809cOo`060?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool0o`3ooolZ0?ooo`030000003oool0oooo01L0oooo00180?oo
o`030000003oool0oooo01l0oooo00<000000?ooo`3oool0o`3ooolZ0?ooo`030000003oool0oooo
01L0oooo00180?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0o`3ooolZ0?ooo`03
0000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
o`3ooolZ0?ooo`030000003oool0oooo01L0oooo00180?ooo`800000803oool00`000000oooo0?oo
o`3o0?ooobP0oooo0`00000I0?ooo`00B03oool00`000000oooo0?ooo`0P0?ooo`030000003oool0
oooo0?l0oooo:@3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0P0?oo
o`030000003oool0oooo0?l0oooo:@3oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo0?l0oooo:@3oool00`000000oooo0?ooo`0G0?ooo`00
B03oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0?l0oooo:@3oool00`000000oooo
0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0?l0oooo:03o
ool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0?l0oooo:03oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0?l0oooo:03oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo0?l0oooo:03oool00`000000oooo0?ooo`0G0?ooo`00
B03oool2000002<0oooo00<000000?ooo`3oool0o`3ooolU0?ooo`<000006@3oool004P0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`3o0?ooobL0oooo00<000000?ooo`3oool05`3o
ool004P0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`3o0?ooobL0oooo00<00000
0?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`3o0?oo
obL0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool08`3oool00`000000
oooo0?ooo`3o0?ooobH0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0
8`3oool00`000000oooo0?ooo`3o0?ooobH0oooo00<000000?ooo`3oool05`3oool004P0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?ooobH0oooo00<000000?ooo`3oool05`3o
ool004P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?ooobH0oooo00<00000
0?ooo`3oool05`3oool004P0oooo0P00000T0?ooo`030000003oool0oooo0?l0oooo903oool30000
01T0oooo00180?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0o`3ooolU0?ooo`03
0000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0
o`3ooolU0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo02@0oooo00<0
00000?ooo`3oool0o`3ooolU0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool0o`3ooolU0?ooo`030000003oool0oooo01L0oooo00180?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool0o`3ooolT0?ooo`030000003oool0oooo
01L0oooo000l0?ooo`030?noM0000000000000<0000000<007Bo0?ooo`3oool00`3oool00`000000
oooo0?ooo`0U0?ooo`030000003oool0oooo0?l0oooo903oool00`000000oooo0?ooo`0G0?ooo`00
?`3oool00`2LB4P0W=oo0?ooo`060?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0
o`3ooolT0?ooo`030000003oool0oooo01L0oooo000o0?ooo`0309a8B02Lgol0oooo00H0oooo00<0
00000?ooo`3oool09@3oool00`000000oooo0?ooo`3o0?ooob@0oooo00<000000?ooo`3oool05`3o
ool003l0oooo00<0W4Q809cOo`3oool01P3oool3000002D0oooo00<000000?ooo`3oool0^@3ooom[
000001T0oooo000o0?ooo`0309a8B02Lgol0oooo00H0oooo00<000000?ooo`3oool09@3oool00`00
0000oooo0?ooo`1N0?oooe/00000JP3oool00`000000oooo0?ooo`0G0?ooo`00?`3oool00`2LB4P0
W=oo0?ooo`060?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0A03ooolI00000<D0
oooo00<000000?ooo`3oool05`3oool003l0oooo00<0W4Q809cOo`3oool01P3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo03d0oooo1`00003N0?ooo`030000003oool0oooo01L0oooo
000l0?ooo`050=nLB000000004RL09a8B02Lgol01`3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo03H0oooo1`00003U0?ooo`030000003oool0oooo01L0oooo000n0?ooo`030=nLB000
04P0W=oo00L0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0]0?ooo`P00000k03o
ool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0
oooo02P0oooo1@00003d0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo
02P0oooo00<000000?ooo`3oool08`3oool400000?T0oooo00<000000?ooo`3oool05`3oool004P0
oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0O0?ooo`<00000o@3oool00`000000
oooo0?ooo`0G0?ooo`00B03oool2000002X0oooo00<000000?ooo`3oool0703oool300000?h0oooo
0`00000I0?ooo`00B03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01P0oooo0`00
003o0?ooo`@0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0E0?ooo`800000o`3oool70?ooo`030000003oool0oooo01L0oooo00180?oo
o`030000003oool0oooo02/0oooo00<000000?ooo`3oool04`3oool200000?l0oooo2@3oool00`00
0000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00l0
oooo0`00003o0?ooo`/0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0
;03oool2000000h0oooo0P00003o0?ooo`h0oooo00<000000?ooo`3oool05`3oool004P0oooo00<0
00000?ooo`3oool0;@3oool00`000000oooo0?ooo`0:0?ooo`800000o`3oool@0?ooo`030000003o
ool0oooo01L0oooo00180?ooo`030000003oool0oooo02h0oooo0P0000060?ooo`@00000o`3ooolB
0?ooo`030000003oool0oooo01L0oooo00180?ooo`800000<@3oool600000?l0oooo503oool30000
01T0oooo00180?ooo`030000003oool0oooo0?l0ooooC03oool00`000000oooo0?ooo`0G0?ooo`00
B03oool00`000000oooo0?ooo`3o0?oood`0oooo00<000000?ooo`3oool05`3oool004P0oooo00<0
00000?ooo`3oool0o`3ooom<0?ooo`030000003oool0oooo01L0oooo00180?ooo`030000003oool0
oooo0?l0ooooC03oool00`000000oooo0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`3o0?oo
od`0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0o`3ooom<0?ooo`03
0000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo0?l0ooooC03oool00`000000oooo
0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`3o0?oood`0oooo00<000000?ooo`3oool05`3o
ool004P0oooo0P00003o0?oood/0oooo0`00000I0?ooo`00B03oool00`000000oooo0?ooo`3o0?oo
od`0oooo00<000000?ooo`3oool05`3oool004P0oooo00<000000?ooo`3oool0o`3ooom<0?ooo`03
0000003oool0oooo01L0oooo00180?ooo`030000003oool0oooo0?l0ooooC03oool00`000000oooo
0?ooo`0G0?ooo`00B03oool00`000000oooo0?ooo`3o0?oood`0oooo00<000000?ooo`3oool05`3o
ool004P0oooo00<000000?ooo`3oool0o`3ooom<0?ooo`030000003oool0oooo01L0oooo00180?oo
o`030000003oool0oooo0400oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`100?oo
o`030000003oool0oooo0400oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0G0?oo
o`00B03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01L0oooo00180?ooool00000
D000000I0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooolV0?ooo`0509a8000000000000
0000B02Lgol0:03oool00`3ookl0M0000000000200000003001d_`3oool0oooo05H0oooo003o0?oo
obL0oooo00<0_g@007Boo`3oool0;03oool00`3o_g@007Bo0?ooo`1H0?ooo`00o`3ooolW0?ooo`03
0;md001d_ol0oooo02d0oooo00<0W4Q809cOo`3oool0E`3oool008<0oooo00`0gia8000000000000
001809cOo`3oool0oono07@00000B9`0oooo0?oo_`1d0003000000030000M02oool0oooo0080oooo
00D0oono07@00000000000000018W0030?ooo`080=nLB0000000000000000000M7@0M0000000B02L
gol20?ooo`050?oo_`1d0000000000000000B9`00`3oool00`2LB000000000000003000000050018
W03oool0oooo0?oo_`1d00000`000000100007@0_ooo0?ooo`3OW4P2000000030018W03oool0_g@0
0080000000<0B9cO0?ooo`3OW4P00`0000001P00M7@0M0000000B02Lgol0oooo09a800D0000000<0
04RL0?ooo`3oool0203oool01P3OW4P0000000000000B9`0oooo09a80080000000@0B9cO0?ooo`3o
ool0_g@00`0000001@0007@0_ooo0?ooo`3oool0gia800<0000000@007Ad07@0000004P0W=oo0P3o
ool01@3ookl0M00000000000000004RL00d0oooo00H0gia800000000000000000000B02Lgol20?oo
o`0;0=nLB0000000000000000000M7@0M0000000B02Lgol0oooo0;md001dM7@00`0000001@0004P0
W=oo0?ooo`3oool0gia800<0000000@007Ad07@0000004P0W=oo0P3oool01P3OW4P0000000000000
0000001809cOo`80oooo00<0W4P00000000000000`0000001P00B9`0oooo0?ooo`3oool0oono07@0
0080000000<004RL0?ooo`3oool00`3oool00`3ookl0M000001d_`1H0?ooo`00Q03oool00`3o_g@0
07Bo0?ooo`020?ooo`060;md0000M;l0oooo0?ooo`2oM000M;oo0`3oool00`3o_g@007Bo0?ooo`02
0?ooo`0:0=nLB018W=l0oooo0?ooo`3OW4P0001d0;ooo`3ookl0M01d0;ooo`80oooo00<0_g@00000
M02oool00`3oool01`3OW4P0B9cO0?ooo`3oool0gia80000M02oool00`3oool00`3ookl0M01d0;oo
o`040?ooo`030;md001d_ol0oooo0080oooo00H0okmd001d_`3oool0oooo0?noM000M;l30?ooo`06
0;md001d_ol0oooo0?oo_`1d07@0_ooo0P3oool00`2oM000001d0;ooo`040?ooo`030?oo_`1d07@0
_ooo00d0oooo00<0okmd001d_`3oool00P3oool01P2oM000M;oo0?ooo`3oool0_g@0001d_`<0oooo
00H0okmd0018W03oool0oono07@0M02oool20?ooo`030;md000007@0_ooo00<0oooo00L0gia804RL
g`3oool0oooo0=nLB00007@0_ooo00X0oooo00<0okmd0018W03oool00P3oool01P3ogi`0B01809cO
o`3ookl0M01d0;ooo`80oooo00<0_g@00000M02oool00P3oool00`2oM000001d0;ooo`020?ooo`06
0?oOW018B9`0oooo0?oo_`1d07@0_ooo0P3oool00`2oM000001d0;ooo`020?ooo`030=nLB000B9`0
oooo0080oooo00<0omnL04P0B02Lgol00`3oool00`3ookl0M01d0;ooo`050?ooo`070=nLB018W=l0
oooo0?ooo`3OW4P0001d0;ooo`030?ooo`040=nLB000M;l0_g@007BooeL0oooo00240?ooo`060?no
M000M;l0oooo0?ooo`2oM00007Bo0P3oool00`3ookl0M4RL0?ooo`040?ooo`0609a8B02Lgol0oooo
0?ooo`3OW4P0B9cO1@3oool00`3ookl0M01809cOo`020?ooo`030?oo_`1d07@0_ooo00<0oooo00<0
gia804RLg`3oool01`3oool00`3ookl0M01d0;ooo`030?ooo`030?oo_`1dB9`0oooo00@0oooo00D0
W4Q809cOo`3oool0okmd001d_`030?ooo`060;md001d_ol0oooo0?oo_`1d04P0W=oo0P3oool00`3o
okl0M01d0;ooo`040?ooo`030?oo_`1d07@0_ooo00d0oooo00<0okmd001d_`3oool00P3oool01P2o
M000M;oo0?ooo`3ookl0M01d0;ooo`H0oooo00<0oono07@0B02Lgol00P3oool00`3ookl0M01d0;oo
o`030?ooo`030=nLB018W=l0oooo00d0oooo00<0oono07@0M02oool01P3oool00`3ookl0M01809cO
o`020?ooo`030?oo_`1d07@0_ooo0080oooo00<0_g@007Boo`3oool00`3oool01@2oM000M;oo0?oo
_`1d04P0W=oo0080oooo00<0oono07@0M02oool00P3oool00`2LB4P0W=oo0?ooo`080?ooo`030?oo
_`1d07@0_ooo00D0oooo00<0gia804RLg`3oool01`3oool0102LB4P0W=oo0?oOW018B9aG0?ooo`00
Q03oool01P3o_g@00000000000000000001809cOo`80oooo00<0oono07A8W03oool0103oool01P2L
B4P0W=oo0?ooo`3oool0gia804RLg`H0oooo00<0omnL04P0000000000P0000000`0007@0_ooo0?oo
o`020?ooo`030=nLB018W=l0oooo00L0oooo00<0oono07@0M02oool00`3oool00`3ookl0M4RL0?oo
o`040?ooo`0509a8B02Lgol0oooo0?noM000M;l00`3oool01P2oM000M;oo0?ooo`3oool0omnL04P0
00<0000000<0001d0;ooo`3oool00`3oool00`3ookl0M01d0;ooo`0=0?ooo`030?noM000M;l0oooo
0080oooo00D0_g@007Boo`3oool0oono07@00005000000050018W03oool0oooo0?oOW01800000`00
00000`0007@0_ooo0?ooo`020?ooo`030=nLB018W=l0oooo00d0oooo00<0omnL04Qd_`3oool01`3o
ool00`3ogi`0B00000000002000000060000M02oool0oooo0?ooo`2oM000B9cO103oool01@2oM4P0
W=oo0?ooo`3ogi`0B00000<0000000H0001d0;ooo`3oool0oooo09a8B02Lgol90?ooo`030?oo_`1d
07@0_ooo00D0oooo00<0gia804RLg`3oool01P3oool01P3o_g@007Bo0?ooo`3oool0_g@007BooeH0
oooo00240?ooo`070?noM000M;l0oooo0?ooo`3ogi`0B01809cOo`020?ooo`030;md0000B9`0oooo
0080oooo00<0_g@0001d_`3oool00P3oool00`3OW4P0B9cO0?ooo`050?ooo`070?oo_`2o_ol0oooo
0?ooo`3ogi`0B01d0;ooo`030?ooo`030=nLB018W=l0oooo00L0oooo00<0oono07@0M02oool0103o
ool00`2oM00004RL0?ooo`020?ooo`070;md0000M;l0oooo0?ooo`3o_g@0000007Boo`020?ooo`06
09a8001d_ol0oooo0?ooo`3ookl0_koo0P3oool00`3ogi`0B01d0;ooo`040?ooo`030?oo_`1d07@0
_ooo00d0oooo00<0okmd0000000000000`0000001@1d_ol0oooo0?ooo`2oM000M;oo00<0oooo00H0
W4P004RLg`3oool0oooo0?oo_`2o_ol20?ooo`030?oOW01807@0_ooo00<0oooo00<0gia804RLg`3o
ool03@3oool00`3ogi`0B7Bo0?ooo`070?ooo`070?oo_`2o_ol0oooo0?ooo`3ogi`0B01d0;ooo`02
0?ooo`030;md00000000M;oo0080oooo00H0gia80018W03oool0oooo0?oo_`2o_ol20?ooo`030?oO
W01807@0_ooo0080oooo00<0gia80000M02oool00P3oool00`2oM000001d0;ooo`030?ooo`030?oo
_`1d07@0_ooo00D0oooo00<0gia804RLg`3oool01P3oool01P2LB4P0W=oo0?ooo`3oool0omnL04Q8
W5H0oooo00240?ooo`030?noM000M;l0oooo0080oooo00<0gia804RLg`3oool00P3oool01@3OW4P0
000000000000000007Bo0080oooo00<0omnL04P0000000000`0000001P0004P0W=oo0?ooo`3oool0
omnL04P00080000000H0001809cOo`3oool0oooo0?oOW0180004000000090000B02Lgol0oooo0?oo
o`3ogi`0B0000000000007@0_ooo00D0oooo00D0gia80000000000000000001d_`020?ooo`040=nL
B000000007Ad07@00080000000H007Bo0?ooo`3oool0oooo0?oOW0180002000000030000B02Lgol0
oooo00@0oooo00<0oono07@0M02oool03@3oool00`3o_g@007Bo0?ooo`020?ooo`060;md001d_ol0
oooo0?ooo`3ookl0M0000`0000001P18W=l0oooo0?ooo`3oool0omnL04P00080000000H0001809cO
o`3oool0oooo0?oOW0180004000000030000B02Lgol0oooo00T0oooo00<0oono07A8W03oool00`3o
ool01P3ookl0M01809cOo`3oool0omnL04P00080000000P0001809cOo`3oool0oooo09a800000000
M;no07@00080000000H004RL0?ooo`3oool0oooo0?oOW0180002000000030000B02Lgol0oooo00<0
oooo00d0okmd0000000000000000001dM01d07@0_ooo0?ooo`3ogi`0B0000000000007@0_ooo00<0
oooo00<0omnL04P0000000000`000000300004P0W=oo0?ooo`3oool0_g@00000000007@0_ooo0?oo
_`1d00000000001d_eD0oooo00240?ooo`070?noM000M;l0oooo0?ooo`3ookl0M00007Boo`0<0?oo
o`030=nLB018W=l0oooo00l0oooo00<0gia804RLg`3oool0:`3oool00`3ookl0M01d0;ooo`0=0?oo
o`030?noM000M;l0oooo0080oooo00<0_g@007Boo`3oool0503oool00`3OW4P0B9cO0?ooo`0>0?oo
o`080=nLB000B9`0oooo0?ooo`3ookl0M0000000B02Lgol_0?ooo`030=nLB018W=l0oooo0680oooo
00230?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool02`3oool00`3OW4P0B9cO0?oo
o`0?0?ooo`030=nLB018W=l0oooo02/0oooo00<0oono07@0M02oool0303oool02@3ogi`0B0000000
0000B9`0oooo09a800000000001809cOo`0D0?ooo`030=nLB018W=l0oooo00l0oooo00L0gia80000
00000000000007Boo`2LB4P0W=oo02l0oooo00<0gia804RLg`3oool0HP3oool00;<0oooo00<0okmd
0000M02oool07`3oool01@3ogi`0B0000000000007@0_ooo06T0oooo00<0okmd0000M02oool0JP3o
ool00;<0oooo00<0okmd0000M02oool0S@3oool00`3o_g@0001d0;ooo`1Z0?ooo`00o`3ooona0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.08381, -0.212377, \
0.0224642, 0.00872347}}]
}, Closed]],

Cell["\<\
Hence, this function is now available for further use here or in \
other notebooks.\
\>", "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["ASCII SDF Files", "Section",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[TextData[{
  "The simplest type of data files are written in plain text (ASCII format) \
and consist of a sequence of lines, called ",
  StyleBox["records",
    FontSlant->"Italic"],
  ", each containing the same number of entries in a common format.  Formats \
of this type are described as ",
  StyleBox["sequential data format",
    FontSlant->"Italic"],
  " (SDF) and can be read easily by a large variety of programs or software \
packages.  Variations of sequential data format might also include titles and \
column headings, often called ",
  StyleBox["headers",
    FontSlant->"Italic"],
  ", and perhaps endnotes, often called ",
  StyleBox["trailers",
    FontSlant->"Italic"],
  ".  Sometimes the sequence of header, data, trailer is repeated several \
times in a ",
  StyleBox["blocked",
    FontSlant->"Italic"],
  " SDF file.  However, these variations entail complications of the reading \
procedures and require more specialized software.  Therefore, in this section \
we consider only the simplest type of sequential data files in which \
numerical data is arranged in columns without headers or trailers.  More \
general file formats are discussed in the next section."
}], "Text"],

Cell[CellGroupData[{

Cell["Writing an SDF file", "Subsection"],

Cell[TextData[{
  "The simplest way to store a data table in an SDF file is to use ",
  ButtonBox["TableForm",
    ButtonStyle->"RefGuideLink"],
  " to format the data and use ",
  ButtonBox["OutputForm",
    ButtonStyle->"RefGuideLink"],
  " to deliver the data to the file as it would appear on the screen.\n\t",
  StyleBox["OutputForm[TableForm[data]] >> ", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["filename\"", "InlineInput"]
}], "Text"],

Cell["\<\
For example, here we create an ASCII SDF file that contains a \
multiplication table\
\>", "Text",
  Evaluatable->False,
  AspectRatioFixed->True],

Cell[BoxData[{
    \(\(data = Table[i\ j, {i, 12}, {j, 12}];\)\), "\n", 
    \(OutputForm[TableForm[data]] >> "\<multiplication.dat\>"\)}], "Input"],

Cell["and examine the contents of that file.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(!! "\<multiplication.dat\>"\)\)], "Input"],

Cell["\<\
1    2    3    4    5    6    7    8    9     10    11    12

2    4    6    8    10   12   14   16   18    20    22    24

3    6    9    12   15   18   21   24   27    30    33    36

4    8    12   16   20   24   28   32   36    40    44    48

5    10   15   20   25   30   35   40   45    50    55    60

6    12   18   24   30   36   42   48   54    60    66    72

7    14   21   28   35   42   49   56   63    70    77    84

8    16   24   32   40   48   56   64   72    80    88    96

9    18   27   36   45   54   63   72   81    90    99    108

10   20   30   40   50   60   70   80   90    100   110   120

11   22   33   44   55   66   77   88   99    110   121   132

12   24   36   48   60   72   84   96   108   120   132   144\
\>", "Print"]
}, Closed]],

Cell[TextData[{
  "To gain more control over the output format, use ",
  ButtonBox["OpenWrite",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\",options", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " to prepare a file to receive data.  Note that if the file already exists, \
",
  StyleBox["OpenWrite", "InlineInput",
    FontWeight->"Bold"],
  " erases the file and prepares a fresh version; if you need to append data \
to the end of an existing file, use ",
  ButtonBox["OpenAppend",
    ButtonStyle->"RefGuideLink"],
  " instead.  Often one chooses ",
  StyleBox["FormatType\[Rule]OutputForm", "InlineInput",
    FontWeight->"Bold"],
  " and then does not need to use ",
  StyleBox["OutputForm", "InlineInput",
    FontWeight->"Bold"],
  " on each expression.  Sometimes it is desirable to increase ",
  StyleBox["PageWidth", "InlineInput",
    FontWeight->"Bold"],
  ", perhaps to \[Infinity], to ensure that each row of table occupies a \
single line or record without undesirable line feeds."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[OpenWrite]\)], "Input"],

Cell[BoxData[
    \({BinaryFormat \[Rule] False, 
      CharacterEncoding \[RuleDelayed] Automatic, 
      DOSTextFormat \[RuleDelayed] True, FormatType \[Rule] InputForm, 
      NumberMarks \[RuleDelayed] $NumberMarks, PageHeight \[Rule] 22, 
      PageWidth \[Rule] 78, TotalHeight \[Rule] \[Infinity], 
      TotalWidth \[Rule] \[Infinity]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Then use one or more ",
  ButtonBox["Write",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\",expr1,\[CenterEllipsis]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " statements to save your data, and finally release the file using ",
  ButtonBox["Close",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\"", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[TextData[{
  "Create a 100\[Times]100 matrix of random integers in the range {1,100}.  \
Open a file using FormatType\[Rule]OutputForm",
  StyleBox[" ",
    FontWeight->"Plain"],
  "and PageWidth\[Rule]\[Infinity], write the matrix to the file using \
TableForm, close the file, and then check its contents and format.  How does \
it differ from a similar file written with default options?"
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Reading an SDF file", "Subsection"],

Cell[TextData[{
  "To read numerical data from a file with each record stored as a sublist, \
you can use the function ",
  ButtonBox["ReadList",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["\"filename\",", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ", "InlineInput",
    FontSlant->"Italic"],
  StyleBox["Number,RecordLists\[Rule]True]", "InlineInput",
    FontWeight->"Bold"],
  ".  Here, ",
  StyleBox["Number", "InlineInput",
    FontWeight->"Bold"],
  " identifies the data entries to be numerical (exact or approximate) and ",
  StyleBox["RecordLists\[Rule]True", "InlineInput",
    FontWeight->"Bold"],
  " indicates that each record is to be returned as a sublist.  Variations of \
",
  StyleBox["ReadList", "InlineInput",
    FontWeight->"Bold"],
  " which handle more general data types are discussed in the next section."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(ReadList["\<multiplication.dat\>", Number, \ 
      RecordLists \[Rule] True]\)], "Input"],

Cell[BoxData[
    \({{1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12}, {2, 4, 6, 8, 10, 12, 14, 16, 
        18, 20, 22, 24}, {3, 6, 9, 12, 15, 18, 21, 24, 27, 30, 33, 36}, {4, 
        8, 12, 16, 20, 24, 28, 32, 36, 40, 44, 48}, {5, 10, 15, 20, 25, 30, 
        35, 40, 45, 50, 55, 60}, {6, 12, 18, 24, 30, 36, 42, 48, 54, 60, 66, 
        72}, {7, 14, 21, 28, 35, 42, 49, 56, 63, 70, 77, 84}, {8, 16, 24, 32, 
        40, 48, 56, 64, 72, 80, 88, 96}, {9, 18, 27, 36, 45, 54, 63, 72, 81, 
        90, 99, 108}, {10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 110, 
        120}, {11, 22, 33, 44, 55, 66, 77, 88, 99, 110, 121, 132}, {12, 24, 
        36, 48, 60, 72, 84, 96, 108, 120, 132, 144}}\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Gently down the stream", "Section"],

Cell[CellGroupData[{

Cell["Streams and data types", "Subsection"],

Cell[TextData[{
  "Perhaps the most flexible file concept is that of a ",
  StyleBox["stream",
    FontSlant->"Italic"],
  " which contains a sequence of data of various types.  A file can be \
attached to an input stream using ",
  ButtonBox["OpenRead",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "with basic syntax\n",
  StyleBox["\t",
    FontWeight->"Bold"],
  StyleBox["streamname = OpenRead[", "InlineInput"],
  StyleBox["\"filename\"", "InlineInput"],
  StyleBox["]", "InlineInput"],
  "\nA single datum is then obtained using ",
  ButtonBox["Read",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["streamname", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ".  Alternatively, ",
  StyleBox["Read[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["streamname,type", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " reads one expression of specified type while  ",
  StyleBox["Read[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["streamname,{type1,type2,\[CenterEllipsis]}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " reads a sequence of expressions of specified type.  A list of the \
available data types is given in the following table.  The default data type \
for ",
  StyleBox["Read", "InlineInput",
    FontWeight->"Bold"],
  " is ",
  StyleBox["Expression", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {
            StyleBox[\(Data\ Types\),
              FontWeight->"Bold"]},
          {GridBox[{
                {
                  StyleBox["Type",
                    "Text",
                    FontWeight->"Bold"], 
                  StyleBox["description",
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Number",
                    "InlineInput"], 
                  StyleBox[\(approximate\ or\ exact\ number\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Real",
                    "InlineInput"], 
                  StyleBox[\(approximate\ real\ number\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Byte",
                    "InlineInput"], 
                  StyleBox[\(single\ byte\ returned\ as\ integer\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Character",
                    "InlineInput"], 
                  StyleBox[\(single\ character\ returned\ as\ string\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Word",
                    "InlineInput"], 
                  
                  StyleBox[\(character\ sequence\ delimited\ by\ word\ \
separators\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Record",
                    "InlineInput"], 
                  
                  StyleBox[\(character\ sequence\ delimited\ by\ record\ \
separators\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["String",
                    "InlineInput"], 
                  StyleBox[\(string\ terminated\ by\ newline\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox["Expression",
                    "InlineInput"], 
                  StyleBox[\(complete\ Mathematica\ expression\),
                    "Text",
                    FontWeight->"Bold"]},
                {
                  StyleBox[\(Hold[Expression]\),
                    "InlineInput"], 
                  StyleBox[\(unevaluated\ Mathematica\ expression\),
                    "Text",
                    FontWeight->"Bold"]}
                },
              ColumnAlignments->{Left},
              GridFrame->True,
              RowLines->{True, False}]}
          }],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "To illustrate these data types, examine the file ",
  StyleBox["\"sample data types.txt\"",
    FontSlant->"Italic"],
  " \[LongDash] you will need to specify the appropriate directory on your \
system."
}], "Text"],

Cell[BoxData[
    \(\(SetDirectory[EssentialMathematica];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(!! Additional\ topics\\sample\ data\ types.txt\)\)], "Input"],

Cell["\<\
 2 2 3.14 4.1e-5 -6.27*^5
 supercalifragilisticexpialodocious Mary Poppins!
Have a nice day!
 q abc
Sqrt[4] 
Sqrt[4]\
\>", "Print"]
}, Closed]],

Cell["Attach this file to an input stream.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(samples = 
      OpenRead["\<Additional topics\sample data types.txt\>"]\)], "Input"],

Cell[BoxData[
    \(InputStream["Additional topics\\sample data types.txt", 17]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "We read the first line using a sequence of data types.  Notice that the \
first datum is interpreted as an integer because it is declared as type ",
  StyleBox["Number", "InlineInput"],
  ", which includes both exact and approximate numbers, while the second \
datum is interpreted as an approximate real number because it is declared as \
type ",
  StyleBox["Real", "InlineInput"],
  " even though the input did not include a decimal point.  Also notice that \
both Fortran and C formats for real numbers are recognized and interpreted \
properly."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, {Number, Real, Real, Number, Number}]\)], "Input"],

Cell[BoxData[
    \({2, 2.`, 3.14`, 0.000041`, \(-627000.`\)}\)], "Output"]
}, Closed]],

Cell["\<\
The next line contains Mary Poppins' favorite word, which is \
returned as a string.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, Word]\  // FullForm\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox["\"\<supercalifragilisticexpialodocious\>\"",
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[TextData[{
  "Additional words on the same line can be read, but care must be taken with \
the delimiters.  The next word does not include the surrounding spaces \
because it is read with Word type, but the final word includes a leading \
space because it is read as a String which is terminated only by the end of \
line.  However, if we had read it as a Word, the end-of-line character would \
have been the next item in the stream.  Also note that reading ",
  StyleBox["Byte", "InlineInput"],
  " and ",
  StyleBox["Character", "InlineInput"],
  " data can be tricky because spaces, tabs, carriage returns, and line-feeds \
are represented as byte data."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, {Word, String}]\  // FullForm\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(List["\<Mary\>", "\< Poppins!\>"]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, String] // FullForm\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox["\"\<Have a nice day!\>\"",
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, Record] // FullForm\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox["\"\< q abc\>\"",
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " expressions are evaluated"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, Expression]\)], "Input"],

Cell[BoxData[
    \(2\)], "Output"]
}, Closed]],

Cell["unless explicitly held unevaluated.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Read[samples, Hold[Expression]]\)], "Input"],

Cell[BoxData[
    \(Hold[\@4]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Functions like ",
  StyleBox["Read", "InlineInput",
    FontWeight->"Bold"],
  ",",
  StyleBox[" ",
    FontWeight->"Bold"],
  StyleBox["Skip", "InlineInput",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Find", "InlineInput",
    FontWeight->"Bold"],
  " operate in a sequential fashion, advancing the stream pointer as data is \
read.  You can position the stream pointer manually using ",
  ButtonBox["SetStreamPosition",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["streamname,n", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " where ",
  StyleBox["n", "InlineInput"],
  " is the position in bytes relative to the beginning of the stream.  Stream \
position 0 indicates the beginning and \[Infinity] the end."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(SetStreamPosition[samples, 0];\)\), "\n", 
    \(Read[samples, Record] // FullForm\)}], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox["\"\< 2 2 3.14 4.1e-5 -6.27*^5\>\"",
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(SetStreamPosition[samples, 75];\)\), "\n", 
    \(Read[samples, Record] // FullForm\)}], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox["\"\<Have a nice day!\>\"",
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Manipulating formatted data", "Subsection"],

Cell["\<\
Often one needs to extract data from a file with an inconvenient \
format, perhaps prepared using a text editor.  Although there is no general \
solution to this problem, we illustrate some possible strategies by \
extracting population data from a file downloaded from the web site of the \
U.S. Census Bureau. \
\>", "Text"],

Cell["First, open an input stream.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(pop = OpenRead["\<Additional topics\popclock.txt\>"]\)], "Input"],

Cell[BoxData[
    \(InputStream["Additional topics\\popclock.txt", 18]\)], "Output"]
}, Closed]],

Cell["Next, read the entire file into a list of records.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(poprecords = ReadList[pop, Record, RecordLists \[Rule] True]\)], "Input"],

Cell[BoxData[
    \({{"Historical National Population Estimates:  July 1, 1900 to July 1, \
1997"}, {"Source:  U.S. Bureau of the Census"}, {"Internet Release date:  \
April 2, 1998"}, {"                          National          Population     \
 Average Annual"}, {"     Date               Population              Change   \
   Percent Change"}, {" July 1, 1997          267,636,061           2,456,650 \
               0.92"}, {" July 1, 1996          265,179,411           \
2,418,772                0.92"}, {" July 1, 1995          262,760,639         \
  2,468,202                0.94"}, {" July 1, 1994          260,292,437       \
    2,539,735                0.98"}, {" July 1, 1993          257,752,702     \
      2,750,875                1.07"}, {" July 1, 1992          255,001,827   \
        2,877,465                1.13"}, {" July 1, 1991          252,124,362 \
          2,684,817                1.07"}, {" July 1, 1990          \
249,439,545           2,620,315                1.06"}, {" July 1, 1989        \
  246,819,230           2,320,248                0.94"}, {" July 1, 1988      \
    244,498,982           2,210,064                0.91"}, {" July 1, 1987    \
      242,288,918           2,156,031                0.89"}, {" July 1, 1986  \
        240,132,887           2,209,092                0.92"}, {" July 1, \
1985          237,923,795           2,098,893                0.89"}, {" July \
1, 1984          235,824,902           2,032,908                0.87"}, {" \
July 1, 1983          233,791,994           2,127,536                0.91"}, \
{" July 1, 1982          231,664,458           2,198,744                \
0.95"}, {" July 1, 1981          229,465,714           2,241,033              \
  0.98"}, {" July 1, 1980          227,224,681           2,169,194            \
    0.96"}, {" July 1, 1979          225,055,487           2,470,942          \
      1.10"}, {" July 1, 1978          222,584,545           2,345,120        \
        1.06"}, {" July 1, 1977          220,239,425           2,204,261      \
          1.01"}, {" July 1, 1976          218,035,164           2,061,965    \
            0.95"}, {" July 1, 1975          215,973,199           2,119,271  \
              0.99"}, {" July 1, 1974          213,853,928           \
1,945,140                0.91"}, {" July 1, 1973          211,908,788         \
  2,012,767                0.95"}, {" July 1, 1972          209,896,021       \
    2,235,344                1.07"}, {" July 1, 1971          207,660,677     \
      2,608,503                1.26"}, {" July 1, 1970          205,052,174   \
        2,375,228                1.17"}, {" July 1, 1969          202,676,946 \
          1,970,894                0.98"}, {" July 1, 1968          \
200,706,052           1,993,996                1.00"}, {" July 1, 1967        \
  198,712,056           2,151,718                1.09"}, {" July 1, 1966      \
    196,560,338           2,257,375                1.16"}, {" July 1, 1965    \
      194,302,963           2,414,172                1.25"}, {" July 1, 1964  \
        191,888,791           2,646,993                1.39"}, {" July 1, \
1963          189,241,798           2,704,061                1.44"}, {" July \
1, 1962          186,537,737           2,846,256                1.54"}, {" \
July 1, 1961          183,691,481           3,020,323                1.66"}, \
{" July 1, 1960          180,671,158           2,841,530                \
1.59"}, {" July 1, 1959          177,829,628           2,947,724              \
  1.67"}, {" July 1, 1958          174,881,904           2,897,774            \
    1.67"}, {" July 1, 1957          171,984,130           3,081,099          \
      1.81"}, {" July 1, 1956          168,903,031           2,971,829        \
        1.78"}, {" July 1, 1955          165,931,202           2,905,348      \
          1.77"}, {" July 1, 1954          163,025,854           2,841,662    \
            1.76"}, {" July 1, 1953          160,184,192           2,631,452  \
              1.66"}, {" July 1, 1952          157,552,740           \
2,674,851                1.71"}, {" July 1, 1951          154,877,889         \
  2,606,472                1.70"}, {" July 1, 1950          152,271,417       \
    3,083,287                2.05"}, {" July 1, 1949          149,188,130     \
      2,556,828                1.73"}, {" July 1, 1948          146,631,302   \
        2,505,231                1.72"}, {" July 1, 1947          144,126,071 \
          2,737,505                1.92"}, {" July 1, 1946          \
141,388,566           1,460,401                1.04"}, {" July 1, 1945        \
  139,928,165           1,530,820                1.10"}, {" July 1, 1944      \
    138,397,345           1,657,992                1.21"}, {" July 1, 1943    \
      136,739,353           1,879,800                1.38"}, {" July 1, 1942  \
        134,859,553           1,457,082                1.09"}, {" July 1, \
1941          133,402,471           1,280,025                0.96"}, {" July \
1, 1940          132,122,446           1,242,728                0.95"}, {" \
July 1, 1939          130,879,718           1,054,779                0.81"}, \
{" July 1, 1938          129,824,939           1,000,110                \
0.77"}, {" July 1, 1937          128,824,829             771,649              \
  0.60"}, {" July 1, 1936          128,053,180             802,948            \
    0.63"}, {" July 1, 1935          127,250,232             876,459          \
      0.69"}, {" July 1, 1934          126,373,773             795,010        \
        0.63"}, {" July 1, 1933          125,578,763             738,292      \
          0.59"}, {" July 1, 1932          124,840,471             800,823    \
            0.64"}, {" July 1, 1931          124,039,648             962,907  \
              0.78"}, {" July 1, 1930          123,076,741           \
1,309,741                1.07"}, {" July 1, 1929          121,767,000         \
  1,258,000                1.04"}, {" July 1, 1928          120,509,000       \
    1,474,000                1.23"}, {" July 1, 1927          119,035,000     \
      1,638,000                1.39"}, {" July 1, 1926          117,397,000   \
        1,568,000                1.34"}, {" July 1, 1925          115,829,000 \
          1,720,000                1.50"}, {" July 1, 1924          \
114,109,000           2,162,000                1.91"}, {" July 1, 1923        \
  111,947,000           1,898,000                1.71"}, {" July 1, 1922      \
    110,049,000           1,511,000                1.38"}, {" July 1, 1921    \
      108,538,000           2,077,000                1.93"}, {" July 1, 1920  \
        106,461,000           1,947,000                1.85"}, {" July 1, \
1919          104,514,000           1,306,000                1.26"}, {" July \
1, 1918          103,208,000             -60,000               -0.06"}, {" \
July 1, 1917          103,268,000           1,307,000                1.27"}, \
{" July 1, 1916          101,961,000           1,415,000                \
1.40"}, {" July 1, 1915          100,546,000           1,435,000              \
  1.44"}, {" July 1, 1914           99,111,000           1,886,000            \
    1.92"}, {" July 1, 1913           97,225,000           1,890,000          \
      1.96"}, {" July 1, 1912           95,335,000           1,472,000        \
        1.56"}, {" July 1, 1911           93,863,000           1,456,000      \
          1.56"}, {" July 1, 1910           92,407,000           1,917,000    \
            2.10"}, {" July 1, 1909           90,490,000           1,780,000  \
              1.99"}, {" July 1, 1908           88,710,000           \
1,702,000                1.94"}, {" July 1, 1907           87,008,000         \
  1,558,000                1.81"}, {" July 1, 1906           85,450,000       \
    1,628,000                1.92"}, {" July 1, 1905           83,822,000     \
      1,656,000                2.00"}, {" July 1, 1904           82,166,000   \
        1,534,000                1.88"}, {" July 1, 1903           80,632,000 \
          1,469,000                1.84"}, {" July 1, 1902           \
79,163,000           1,579,000                2.01"}, {" July 1, 1901         \
  77,584,000           1,490,000                1.94"}, {" July 1, 1900       \
    76,094,000                 ---                 ---"}, {"NOTES:"}, {"(1)  \
National population data for the years 1900 to 1929 are only available"}, {"\t\
rounded to the nearest thousand."}, {"(2)  National population data for the \
years 1900 to 1949 exclude the"}, {"\tpopulation residing in Alaska and \
Hawaii."}, {"(3)  National population data for the years 1940 to 1979 cover \
the resident"}, {"\tpopulation plus Armed Forces overseas.  National \
population data for "}, {"\tall other years cover only the resident \
population."}, {"SOURCES:"}, {"Current Population Reports, Series P-25, Nos. \
311, 917, 1095, and"}, {"Population Paper Listing PPL-91."}}\)], "Output"]
}, Closed]],

Cell["\<\
Examining this file we find that there are 5 records containing \
titles and column headings followed by about 98 records containing data in a \
fixed columnar format.  The file concludes with a set of notes and \
bibliographic information we do not need.  Hence, the next step is to drop \
the records containing uninteresting text.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(popdata = Drop[Drop[poprecords, 5], \(-11\)]\)], "Input"],

Cell[BoxData[
    \({{" July 1, 1997          267,636,061           2,456,650               \
 0.92"}, {" July 1, 1996          265,179,411           2,418,772             \
   0.92"}, {" July 1, 1995          262,760,639           2,468,202           \
     0.94"}, {" July 1, 1994          260,292,437           2,539,735         \
       0.98"}, {" July 1, 1993          257,752,702           2,750,875       \
         1.07"}, {" July 1, 1992          255,001,827           2,877,465     \
           1.13"}, {" July 1, 1991          252,124,362           2,684,817   \
             1.07"}, {" July 1, 1990          249,439,545           2,620,315 \
               1.06"}, {" July 1, 1989          246,819,230           \
2,320,248                0.94"}, {" July 1, 1988          244,498,982         \
  2,210,064                0.91"}, {" July 1, 1987          242,288,918       \
    2,156,031                0.89"}, {" July 1, 1986          240,132,887     \
      2,209,092                0.92"}, {" July 1, 1985          237,923,795   \
        2,098,893                0.89"}, {" July 1, 1984          235,824,902 \
          2,032,908                0.87"}, {" July 1, 1983          \
233,791,994           2,127,536                0.91"}, {" July 1, 1982        \
  231,664,458           2,198,744                0.95"}, {" July 1, 1981      \
    229,465,714           2,241,033                0.98"}, {" July 1, 1980    \
      227,224,681           2,169,194                0.96"}, {" July 1, 1979  \
        225,055,487           2,470,942                1.10"}, {" July 1, \
1978          222,584,545           2,345,120                1.06"}, {" July \
1, 1977          220,239,425           2,204,261                1.01"}, {" \
July 1, 1976          218,035,164           2,061,965                0.95"}, \
{" July 1, 1975          215,973,199           2,119,271                \
0.99"}, {" July 1, 1974          213,853,928           1,945,140              \
  0.91"}, {" July 1, 1973          211,908,788           2,012,767            \
    0.95"}, {" July 1, 1972          209,896,021           2,235,344          \
      1.07"}, {" July 1, 1971          207,660,677           2,608,503        \
        1.26"}, {" July 1, 1970          205,052,174           2,375,228      \
          1.17"}, {" July 1, 1969          202,676,946           1,970,894    \
            0.98"}, {" July 1, 1968          200,706,052           1,993,996  \
              1.00"}, {" July 1, 1967          198,712,056           \
2,151,718                1.09"}, {" July 1, 1966          196,560,338         \
  2,257,375                1.16"}, {" July 1, 1965          194,302,963       \
    2,414,172                1.25"}, {" July 1, 1964          191,888,791     \
      2,646,993                1.39"}, {" July 1, 1963          189,241,798   \
        2,704,061                1.44"}, {" July 1, 1962          186,537,737 \
          2,846,256                1.54"}, {" July 1, 1961          \
183,691,481           3,020,323                1.66"}, {" July 1, 1960        \
  180,671,158           2,841,530                1.59"}, {" July 1, 1959      \
    177,829,628           2,947,724                1.67"}, {" July 1, 1958    \
      174,881,904           2,897,774                1.67"}, {" July 1, 1957  \
        171,984,130           3,081,099                1.81"}, {" July 1, \
1956          168,903,031           2,971,829                1.78"}, {" July \
1, 1955          165,931,202           2,905,348                1.77"}, {" \
July 1, 1954          163,025,854           2,841,662                1.76"}, \
{" July 1, 1953          160,184,192           2,631,452                \
1.66"}, {" July 1, 1952          157,552,740           2,674,851              \
  1.71"}, {" July 1, 1951          154,877,889           2,606,472            \
    1.70"}, {" July 1, 1950          152,271,417           3,083,287          \
      2.05"}, {" July 1, 1949          149,188,130           2,556,828        \
        1.73"}, {" July 1, 1948          146,631,302           2,505,231      \
          1.72"}, {" July 1, 1947          144,126,071           2,737,505    \
            1.92"}, {" July 1, 1946          141,388,566           1,460,401  \
              1.04"}, {" July 1, 1945          139,928,165           \
1,530,820                1.10"}, {" July 1, 1944          138,397,345         \
  1,657,992                1.21"}, {" July 1, 1943          136,739,353       \
    1,879,800                1.38"}, {" July 1, 1942          134,859,553     \
      1,457,082                1.09"}, {" July 1, 1941          133,402,471   \
        1,280,025                0.96"}, {" July 1, 1940          132,122,446 \
          1,242,728                0.95"}, {" July 1, 1939          \
130,879,718           1,054,779                0.81"}, {" July 1, 1938        \
  129,824,939           1,000,110                0.77"}, {" July 1, 1937      \
    128,824,829             771,649                0.60"}, {" July 1, 1936    \
      128,053,180             802,948                0.63"}, {" July 1, 1935  \
        127,250,232             876,459                0.69"}, {" July 1, \
1934          126,373,773             795,010                0.63"}, {" July \
1, 1933          125,578,763             738,292                0.59"}, {" \
July 1, 1932          124,840,471             800,823                0.64"}, \
{" July 1, 1931          124,039,648             962,907                \
0.78"}, {" July 1, 1930          123,076,741           1,309,741              \
  1.07"}, {" July 1, 1929          121,767,000           1,258,000            \
    1.04"}, {" July 1, 1928          120,509,000           1,474,000          \
      1.23"}, {" July 1, 1927          119,035,000           1,638,000        \
        1.39"}, {" July 1, 1926          117,397,000           1,568,000      \
          1.34"}, {" July 1, 1925          115,829,000           1,720,000    \
            1.50"}, {" July 1, 1924          114,109,000           2,162,000  \
              1.91"}, {" July 1, 1923          111,947,000           \
1,898,000                1.71"}, {" July 1, 1922          110,049,000         \
  1,511,000                1.38"}, {" July 1, 1921          108,538,000       \
    2,077,000                1.93"}, {" July 1, 1920          106,461,000     \
      1,947,000                1.85"}, {" July 1, 1919          104,514,000   \
        1,306,000                1.26"}, {" July 1, 1918          103,208,000 \
            -60,000               -0.06"}, {" July 1, 1917          \
103,268,000           1,307,000                1.27"}, {" July 1, 1916        \
  101,961,000           1,415,000                1.40"}, {" July 1, 1915      \
    100,546,000           1,435,000                1.44"}, {" July 1, 1914    \
       99,111,000           1,886,000                1.92"}, {" July 1, 1913  \
         97,225,000           1,890,000                1.96"}, {" July 1, \
1912           95,335,000           1,472,000                1.56"}, {" July \
1, 1911           93,863,000           1,456,000                1.56"}, {" \
July 1, 1910           92,407,000           1,917,000                2.10"}, \
{" July 1, 1909           90,490,000           1,780,000                \
1.99"}, {" July 1, 1908           88,710,000           1,702,000              \
  1.94"}, {" July 1, 1907           87,008,000           1,558,000            \
    1.81"}, {" July 1, 1906           85,450,000           1,628,000          \
      1.92"}, {" July 1, 1905           83,822,000           1,656,000        \
        2.00"}, {" July 1, 1904           82,166,000           1,534,000      \
          1.88"}, {" July 1, 1903           80,632,000           1,469,000    \
            1.84"}, {" July 1, 1902           79,163,000           1,579,000  \
              2.01"}, {" July 1, 1901           77,584,000           \
1,490,000                1.94"}, {" July 1, 1900           76,094,000         \
        ---                 ---"}}\)], "Output"]
}, Closed]],

Cell["\<\
Note that each data record is represented as a list containing a \
single string.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FullForm[popdata\[LeftDoubleBracket]1\[RightDoubleBracket]]\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[\(List["\< July 1, 1997          267,636,061           \
2,456,650                0.92\>"]\),
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FullForm[
      popdata\[LeftDoubleBracket]2, 1\[RightDoubleBracket]]\)], "Input"],

Cell[BoxData[
    TagBox[
      StyleBox[
        "\"\< July 1, 1996          265,179,411           2,418,772           \
     0.92\>\"",
        ShowSpecialCharacters->False,
        ShowStringCharacters->True,
        NumberMarks->True],
      FullForm]], "Output"]
}, Closed]],

Cell[TextData[{
  "The commas used to format dates and numbers are rather inconvenient, but \
can easily be removed by mapping ",
  ButtonBox["StringReplace",
    ButtonStyle->"RefGuideLink"],
  " onto a list of strings.  Note that one level of listing is removed first \
using ",
  StyleBox["Flatten", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(\ \)\(newpopdata = \(\(\((StringReplace[#, "\<,\>" \[Rule] "\<\>"] \
&)\) /@ \ Flatten[popdata]\)\(\ \)\)\)\)\)], "Input"],

Cell[BoxData[
    \({" July 1 1997          267636061           2456650                \
0.92", " July 1 1996          265179411           2418772                \
0.92", " July 1 1995          262760639           2468202                \
0.94", " July 1 1994          260292437           2539735                \
0.98", " July 1 1993          257752702           2750875                \
1.07", " July 1 1992          255001827           2877465                \
1.13", " July 1 1991          252124362           2684817                \
1.07", " July 1 1990          249439545           2620315                \
1.06", " July 1 1989          246819230           2320248                \
0.94", " July 1 1988          244498982           2210064                \
0.91", " July 1 1987          242288918           2156031                \
0.89", " July 1 1986          240132887           2209092                \
0.92", " July 1 1985          237923795           2098893                \
0.89", " July 1 1984          235824902           2032908                \
0.87", " July 1 1983          233791994           2127536                \
0.91", " July 1 1982          231664458           2198744                \
0.95", " July 1 1981          229465714           2241033                \
0.98", " July 1 1980          227224681           2169194                \
0.96", " July 1 1979          225055487           2470942                \
1.10", " July 1 1978          222584545           2345120                \
1.06", " July 1 1977          220239425           2204261                \
1.01", " July 1 1976          218035164           2061965                \
0.95", " July 1 1975          215973199           2119271                \
0.99", " July 1 1974          213853928           1945140                \
0.91", " July 1 1973          211908788           2012767                \
0.95", " July 1 1972          209896021           2235344                \
1.07", " July 1 1971          207660677           2608503                \
1.26", " July 1 1970          205052174           2375228                \
1.17", " July 1 1969          202676946           1970894                \
0.98", " July 1 1968          200706052           1993996                \
1.00", " July 1 1967          198712056           2151718                \
1.09", " July 1 1966          196560338           2257375                \
1.16", " July 1 1965          194302963           2414172                \
1.25", " July 1 1964          191888791           2646993                \
1.39", " July 1 1963          189241798           2704061                \
1.44", " July 1 1962          186537737           2846256                \
1.54", " July 1 1961          183691481           3020323                \
1.66", " July 1 1960          180671158           2841530                \
1.59", " July 1 1959          177829628           2947724                \
1.67", " July 1 1958          174881904           2897774                \
1.67", " July 1 1957          171984130           3081099                \
1.81", " July 1 1956          168903031           2971829                \
1.78", " July 1 1955          165931202           2905348                \
1.77", " July 1 1954          163025854           2841662                \
1.76", " July 1 1953          160184192           2631452                \
1.66", " July 1 1952          157552740           2674851                \
1.71", " July 1 1951          154877889           2606472                \
1.70", " July 1 1950          152271417           3083287                \
2.05", " July 1 1949          149188130           2556828                \
1.73", " July 1 1948          146631302           2505231                \
1.72", " July 1 1947          144126071           2737505                \
1.92", " July 1 1946          141388566           1460401                \
1.04", " July 1 1945          139928165           1530820                \
1.10", " July 1 1944          138397345           1657992                \
1.21", " July 1 1943          136739353           1879800                \
1.38", " July 1 1942          134859553           1457082                \
1.09", " July 1 1941          133402471           1280025                \
0.96", " July 1 1940          132122446           1242728                \
0.95", " July 1 1939          130879718           1054779                \
0.81", " July 1 1938          129824939           1000110                \
0.77", " July 1 1937          128824829             771649                \
0.60", " July 1 1936          128053180             802948                \
0.63", " July 1 1935          127250232             876459                \
0.69", " July 1 1934          126373773             795010                \
0.63", " July 1 1933          125578763             738292                \
0.59", " July 1 1932          124840471             800823                \
0.64", " July 1 1931          124039648             962907                \
0.78", " July 1 1930          123076741           1309741                \
1.07", " July 1 1929          121767000           1258000                \
1.04", " July 1 1928          120509000           1474000                \
1.23", " July 1 1927          119035000           1638000                \
1.39", " July 1 1926          117397000           1568000                \
1.34", " July 1 1925          115829000           1720000                \
1.50", " July 1 1924          114109000           2162000                \
1.91", " July 1 1923          111947000           1898000                \
1.71", " July 1 1922          110049000           1511000                \
1.38", " July 1 1921          108538000           2077000                \
1.93", " July 1 1920          106461000           1947000                \
1.85", " July 1 1919          104514000           1306000                \
1.26", " July 1 1918          103208000             -60000               \
-0.06", " July 1 1917          103268000           1307000                \
1.27", " July 1 1916          101961000           1415000                \
1.40", " July 1 1915          100546000           1435000                \
1.44", " July 1 1914           99111000           1886000                \
1.92", " July 1 1913           97225000           1890000                \
1.96", " July 1 1912           95335000           1472000                \
1.56", " July 1 1911           93863000           1456000                \
1.56", " July 1 1910           92407000           1917000                \
2.10", " July 1 1909           90490000           1780000                \
1.99", " July 1 1908           88710000           1702000                \
1.94", " July 1 1907           87008000           1558000                \
1.81", " July 1 1906           85450000           1628000                \
1.92", " July 1 1905           83822000           1656000                \
2.00", " July 1 1904           82166000           1534000                \
1.88", " July 1 1903           80632000           1469000                \
1.84", " July 1 1902           79163000           1579000                \
2.01", " July 1 1901           77584000           1490000                \
1.94", " July 1 1900           76094000                 ---                 \
---"}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Next, each string can be treated as an input stream using ",
  ButtonBox["StringToStream",
    ButtonStyle->"RefGuideLink"],
  " and then read using ",
  StyleBox["Read", "InlineInput",
    FontWeight->"Bold"],
  ".  Minor problems occur at the end because of the change of format, but \
they can be ignored."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(newerpopdata = \((Read[#, {Word, 
                Table[Number, {5}]}] &)\) /@ \ \((StringToStream\  /@ \ 
            newpopdata)\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(Read::"readn"\), \(\(:\)\(\ \)\), "\<\"Syntax error reading a \
real number from \\!\\(System`Private`StringTurnedIntoStream[\\\" July 1 1900 \
          76094000                 ---                 ---\\\"]\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Read::readn\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({{"July", {1, 1997, 267636061, 2456650, 0.92`}}, {"July", {1, 1996, 
          265179411, 2418772, 0.92`}}, {"July", {1, 1995, 262760639, 2468202, 
          0.94`}}, {"July", {1, 1994, 260292437, 2539735, 
          0.98`}}, {"July", {1, 1993, 257752702, 2750875, 
          1.07`}}, {"July", {1, 1992, 255001827, 2877465, 
          1.13`}}, {"July", {1, 1991, 252124362, 2684817, 
          1.07`}}, {"July", {1, 1990, 249439545, 2620315, 
          1.06`}}, {"July", {1, 1989, 246819230, 2320248, 
          0.94`}}, {"July", {1, 1988, 244498982, 2210064, 
          0.91`}}, {"July", {1, 1987, 242288918, 2156031, 
          0.89`}}, {"July", {1, 1986, 240132887, 2209092, 
          0.92`}}, {"July", {1, 1985, 237923795, 2098893, 
          0.89`}}, {"July", {1, 1984, 235824902, 2032908, 
          0.87`}}, {"July", {1, 1983, 233791994, 2127536, 
          0.91`}}, {"July", {1, 1982, 231664458, 2198744, 
          0.95`}}, {"July", {1, 1981, 229465714, 2241033, 
          0.98`}}, {"July", {1, 1980, 227224681, 2169194, 
          0.96`}}, {"July", {1, 1979, 225055487, 2470942, 
          1.1`}}, {"July", {1, 1978, 222584545, 2345120, 
          1.06`}}, {"July", {1, 1977, 220239425, 2204261, 
          1.01`}}, {"July", {1, 1976, 218035164, 2061965, 
          0.95`}}, {"July", {1, 1975, 215973199, 2119271, 
          0.99`}}, {"July", {1, 1974, 213853928, 1945140, 
          0.91`}}, {"July", {1, 1973, 211908788, 2012767, 
          0.95`}}, {"July", {1, 1972, 209896021, 2235344, 
          1.07`}}, {"July", {1, 1971, 207660677, 2608503, 
          1.26`}}, {"July", {1, 1970, 205052174, 2375228, 
          1.17`}}, {"July", {1, 1969, 202676946, 1970894, 
          0.98`}}, {"July", {1, 1968, 200706052, 1993996, 1.`}}, {"July", {1, 
          1967, 198712056, 2151718, 1.09`}}, {"July", {1, 1966, 196560338, 
          2257375, 1.16`}}, {"July", {1, 1965, 194302963, 2414172, 
          1.25`}}, {"July", {1, 1964, 191888791, 2646993, 
          1.39`}}, {"July", {1, 1963, 189241798, 2704061, 
          1.44`}}, {"July", {1, 1962, 186537737, 2846256, 
          1.54`}}, {"July", {1, 1961, 183691481, 3020323, 
          1.66`}}, {"July", {1, 1960, 180671158, 2841530, 
          1.59`}}, {"July", {1, 1959, 177829628, 2947724, 
          1.67`}}, {"July", {1, 1958, 174881904, 2897774, 
          1.67`}}, {"July", {1, 1957, 171984130, 3081099, 
          1.81`}}, {"July", {1, 1956, 168903031, 2971829, 
          1.78`}}, {"July", {1, 1955, 165931202, 2905348, 
          1.77`}}, {"July", {1, 1954, 163025854, 2841662, 
          1.76`}}, {"July", {1, 1953, 160184192, 2631452, 
          1.66`}}, {"July", {1, 1952, 157552740, 2674851, 
          1.71`}}, {"July", {1, 1951, 154877889, 2606472, 
          1.7`}}, {"July", {1, 1950, 152271417, 3083287, 
          2.05`}}, {"July", {1, 1949, 149188130, 2556828, 
          1.73`}}, {"July", {1, 1948, 146631302, 2505231, 
          1.72`}}, {"July", {1, 1947, 144126071, 2737505, 
          1.92`}}, {"July", {1, 1946, 141388566, 1460401, 
          1.04`}}, {"July", {1, 1945, 139928165, 1530820, 
          1.1`}}, {"July", {1, 1944, 138397345, 1657992, 
          1.21`}}, {"July", {1, 1943, 136739353, 1879800, 
          1.38`}}, {"July", {1, 1942, 134859553, 1457082, 
          1.09`}}, {"July", {1, 1941, 133402471, 1280025, 
          0.96`}}, {"July", {1, 1940, 132122446, 1242728, 
          0.95`}}, {"July", {1, 1939, 130879718, 1054779, 
          0.81`}}, {"July", {1, 1938, 129824939, 1000110, 
          0.77`}}, {"July", {1, 1937, 128824829, 771649, 0.6`}}, {"July", {1, 
          1936, 128053180, 802948, 0.63`}}, {"July", {1, 1935, 127250232, 
          876459, 0.69`}}, {"July", {1, 1934, 126373773, 795010, 
          0.63`}}, {"July", {1, 1933, 125578763, 738292, 
          0.59`}}, {"July", {1, 1932, 124840471, 800823, 
          0.64`}}, {"July", {1, 1931, 124039648, 962907, 
          0.78`}}, {"July", {1, 1930, 123076741, 1309741, 
          1.07`}}, {"July", {1, 1929, 121767000, 1258000, 
          1.04`}}, {"July", {1, 1928, 120509000, 1474000, 
          1.23`}}, {"July", {1, 1927, 119035000, 1638000, 
          1.39`}}, {"July", {1, 1926, 117397000, 1568000, 
          1.34`}}, {"July", {1, 1925, 115829000, 1720000, 
          1.5`}}, {"July", {1, 1924, 114109000, 2162000, 
          1.91`}}, {"July", {1, 1923, 111947000, 1898000, 
          1.71`}}, {"July", {1, 1922, 110049000, 1511000, 
          1.38`}}, {"July", {1, 1921, 108538000, 2077000, 
          1.93`}}, {"July", {1, 1920, 106461000, 1947000, 
          1.85`}}, {"July", {1, 1919, 104514000, 1306000, 
          1.26`}}, {"July", {1, 1918, 
          103208000, \(-60000\), \(-0.06`\)}}, {"July", {1, 1917, 103268000, 
          1307000, 1.27`}}, {"July", {1, 1916, 101961000, 1415000, 
          1.4`}}, {"July", {1, 1915, 100546000, 1435000, 
          1.44`}}, {"July", {1, 1914, 99111000, 1886000, 
          1.92`}}, {"July", {1, 1913, 97225000, 1890000, 
          1.96`}}, {"July", {1, 1912, 95335000, 1472000, 
          1.56`}}, {"July", {1, 1911, 93863000, 1456000, 
          1.56`}}, {"July", {1, 1910, 92407000, 1917000, 2.1`}}, {"July", {1, 
          1909, 90490000, 1780000, 1.99`}}, {"July", {1, 1908, 88710000, 
          1702000, 1.94`}}, {"July", {1, 1907, 87008000, 1558000, 
          1.81`}}, {"July", {1, 1906, 85450000, 1628000, 
          1.92`}}, {"July", {1, 1905, 83822000, 1656000, 2.`}}, {"July", {1, 
          1904, 82166000, 1534000, 1.88`}}, {"July", {1, 1903, 80632000, 
          1469000, 1.84`}}, {"July", {1, 1902, 79163000, 1579000, 
          2.01`}}, {"July", {1, 1901, 77584000, 1490000, 
          1.94`}}, {"July", {1, 1900, 76094000, $Failed, 
          EndOfFile}}}\)], "Output"]
}, Closed]],

Cell["\<\
Finally, we can extract the items we want; namely, the year and the \
total population.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(data = \(Part[#, 2, {2, 3}] &\)\  /@ \ newerpopdata\)], "Input"],

Cell[BoxData[
    \({{1997, 267636061}, {1996, 265179411}, {1995, 262760639}, {1994, 
        260292437}, {1993, 257752702}, {1992, 255001827}, {1991, 
        252124362}, {1990, 249439545}, {1989, 246819230}, {1988, 
        244498982}, {1987, 242288918}, {1986, 240132887}, {1985, 
        237923795}, {1984, 235824902}, {1983, 233791994}, {1982, 
        231664458}, {1981, 229465714}, {1980, 227224681}, {1979, 
        225055487}, {1978, 222584545}, {1977, 220239425}, {1976, 
        218035164}, {1975, 215973199}, {1974, 213853928}, {1973, 
        211908788}, {1972, 209896021}, {1971, 207660677}, {1970, 
        205052174}, {1969, 202676946}, {1968, 200706052}, {1967, 
        198712056}, {1966, 196560338}, {1965, 194302963}, {1964, 
        191888791}, {1963, 189241798}, {1962, 186537737}, {1961, 
        183691481}, {1960, 180671158}, {1959, 177829628}, {1958, 
        174881904}, {1957, 171984130}, {1956, 168903031}, {1955, 
        165931202}, {1954, 163025854}, {1953, 160184192}, {1952, 
        157552740}, {1951, 154877889}, {1950, 152271417}, {1949, 
        149188130}, {1948, 146631302}, {1947, 144126071}, {1946, 
        141388566}, {1945, 139928165}, {1944, 138397345}, {1943, 
        136739353}, {1942, 134859553}, {1941, 133402471}, {1940, 
        132122446}, {1939, 130879718}, {1938, 129824939}, {1937, 
        128824829}, {1936, 128053180}, {1935, 127250232}, {1934, 
        126373773}, {1933, 125578763}, {1932, 124840471}, {1931, 
        124039648}, {1930, 123076741}, {1929, 121767000}, {1928, 
        120509000}, {1927, 119035000}, {1926, 117397000}, {1925, 
        115829000}, {1924, 114109000}, {1923, 111947000}, {1922, 
        110049000}, {1921, 108538000}, {1920, 106461000}, {1919, 
        104514000}, {1918, 103208000}, {1917, 103268000}, {1916, 
        101961000}, {1915, 100546000}, {1914, 99111000}, {1913, 
        97225000}, {1912, 95335000}, {1911, 93863000}, {1910, 
        92407000}, {1909, 90490000}, {1908, 88710000}, {1907, 
        87008000}, {1906, 85450000}, {1905, 83822000}, {1904, 
        82166000}, {1903, 80632000}, {1902, 79163000}, {1901, 
        77584000}, {1900, 76094000}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ListPlot[data];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-18.6311 0.00981836 -0.21912 3.07297e-009 [
[.22018 .07568 -12 -9 ]
[.22018 .07568 12 0 ]
[.41654 .07568 -12 -9 ]
[.41654 .07568 12 0 ]
[.61291 .07568 -12 -9 ]
[.61291 .07568 12 0 ]
[.80928 .07568 -12 -9 ]
[.80928 .07568 12 0 ]
[.01131 .24183 -40.5 -6.25 ]
[.01131 .24183 0 6.25 ]
[.01131 .39548 -28.5 -6.25 ]
[.01131 .39548 0 6.25 ]
[.01131 .54912 -40.5 -6.25 ]
[.01131 .54912 0 6.25 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.22018 .08818 m
.22018 .09443 L
s
[(1920)] .22018 .07568 0 1 Mshowa
.41654 .08818 m
.41654 .09443 L
s
[(1940)] .41654 .07568 0 1 Mshowa
.61291 .08818 m
.61291 .09443 L
s
[(1960)] .61291 .07568 0 1 Mshowa
.80928 .08818 m
.80928 .09443 L
s
[(1980)] .80928 .07568 0 1 Mshowa
.125 Mabswid
.0729 .08818 m
.0729 .09193 L
s
.12199 .08818 m
.12199 .09193 L
s
.17108 .08818 m
.17108 .09193 L
s
.26927 .08818 m
.26927 .09193 L
s
.31836 .08818 m
.31836 .09193 L
s
.36745 .08818 m
.36745 .09193 L
s
.46564 .08818 m
.46564 .09193 L
s
.51473 .08818 m
.51473 .09193 L
s
.56382 .08818 m
.56382 .09193 L
s
.662 .08818 m
.662 .09193 L
s
.71109 .08818 m
.71109 .09193 L
s
.76019 .08818 m
.76019 .09193 L
s
.85837 .08818 m
.85837 .09193 L
s
.90746 .08818 m
.90746 .09193 L
s
.95655 .08818 m
.95655 .09193 L
s
.25 Mabswid
0 .08818 m
1 .08818 L
s
.02381 .24183 m
.03006 .24183 L
s
gsave
.01131 .24183 -40.5 -4.25 Mabsadd m
1 1 Mabs scale
/Courier findfont 10 scalefont setfont
(1.5) show
/Math1Mono findfont 10 scalefont setfont
(\\264) show
/Courier findfont 10 scalefont setfont
(10) show
0 5.25 rmoveto
/Courier findfont 7.5 scalefont setfont
(8) show
grestore
.02381 .39548 m
.03006 .39548 L
s
gsave
.01131 .39548 -28.5 -4.25 Mabsadd m
1 1 Mabs scale
/Courier findfont 10 scalefont setfont
(2) show
/Math1Mono findfont 10 scalefont setfont
(\\264) show
/Courier findfont 10 scalefont setfont
(10) show
0 5.25 rmoveto
/Courier findfont 7.5 scalefont setfont
(8) show
grestore
.02381 .54912 m
.03006 .54912 L
s
gsave
.01131 .54912 -40.5 -4.25 Mabsadd m
1 1 Mabs scale
/Courier findfont 10 scalefont setfont
(2.5) show
/Math1Mono findfont 10 scalefont setfont
(\\264) show
/Courier findfont 10 scalefont setfont
(10) show
0 5.25 rmoveto
/Courier findfont 7.5 scalefont setfont
(8) show
grestore
.125 Mabswid
.02381 .11891 m
.02756 .11891 L
s
.02381 .14964 m
.02756 .14964 L
s
.02381 .18037 m
.02756 .18037 L
s
.02381 .2111 m
.02756 .2111 L
s
.02381 .27256 m
.02756 .27256 L
s
.02381 .30329 m
.02756 .30329 L
s
.02381 .33402 m
.02756 .33402 L
s
.02381 .36475 m
.02756 .36475 L
s
.02381 .4262 m
.02756 .4262 L
s
.02381 .45693 m
.02756 .45693 L
s
.02381 .48766 m
.02756 .48766 L
s
.02381 .51839 m
.02756 .51839 L
s
.02381 .05745 m
.02756 .05745 L
s
.02381 .02672 m
.02756 .02672 L
s
.02381 .57985 m
.02756 .57985 L
s
.02381 .61058 m
.02756 .61058 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.97619 .60332 Mdot
.96637 .59577 Mdot
.95655 .58834 Mdot
.94674 .58075 Mdot
.93692 .57295 Mdot
.9271 .56449 Mdot
.91728 .55565 Mdot
.90746 .5474 Mdot
.89764 .53935 Mdot
.88783 .53222 Mdot
.87801 .52543 Mdot
.86819 .5188 Mdot
.85837 .51201 Mdot
.84855 .50556 Mdot
.83873 .49932 Mdot
.82892 .49278 Mdot
.8191 .48602 Mdot
.80928 .47914 Mdot
.79946 .47247 Mdot
.78964 .46488 Mdot
.77982 .45767 Mdot
.77 .4509 Mdot
.76019 .44456 Mdot
.75037 .43805 Mdot
.74055 .43207 Mdot
.73073 .42589 Mdot
.72091 .41902 Mdot
.71109 .411 Mdot
.70128 .4037 Mdot
.69146 .39764 Mdot
.68164 .39152 Mdot
.67182 .38491 Mdot
.662 .37797 Mdot
.65218 .37055 Mdot
.64237 .36242 Mdot
.63255 .35411 Mdot
.62273 .34536 Mdot
.61291 .33608 Mdot
.60309 .32735 Mdot
.59327 .31829 Mdot
.58346 .30938 Mdot
.57364 .29991 Mdot
.56382 .29078 Mdot
.554 .28185 Mdot
.54418 .27312 Mdot
.53436 .26504 Mdot
.52455 .25682 Mdot
.51473 .24881 Mdot
.50491 .23933 Mdot
.49509 .23147 Mdot
.48527 .22378 Mdot
.47545 .21536 Mdot
.46564 .21088 Mdot
.45582 .20617 Mdot
.446 .20108 Mdot
.43618 .1953 Mdot
.42636 .19082 Mdot
.41654 .18689 Mdot
.40673 .18307 Mdot
.39691 .17983 Mdot
.38709 .17676 Mdot
.37727 .17438 Mdot
.36745 .17192 Mdot
.35763 .16922 Mdot
.34782 .16678 Mdot
.338 .16451 Mdot
.32818 .16205 Mdot
.31836 .15909 Mdot
.30854 .15507 Mdot
.29872 .1512 Mdot
.28891 .14667 Mdot
.27909 .14164 Mdot
.26927 .13682 Mdot
.25945 .13153 Mdot
.24963 .12489 Mdot
.23981 .11906 Mdot
.23 .11441 Mdot
.22018 .10803 Mdot
.21036 .10205 Mdot
.20054 .09804 Mdot
.19072 .09822 Mdot
.1809 .0942 Mdot
.17108 .08986 Mdot
.16127 .08545 Mdot
.15145 .07965 Mdot
.14163 .07384 Mdot
.13181 .06932 Mdot
.12199 .06484 Mdot
.11217 .05895 Mdot
.10236 .05348 Mdot
.09254 .04825 Mdot
.08272 .04347 Mdot
.0729 .03846 Mdot
.06308 .03337 Mdot
.05326 .02866 Mdot
.04345 .02415 Mdot
.03363 .01929 Mdot
.02381 .01472 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0.625}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000
003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool0
0`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0oooo0?l0ooooK@3oool00400
oooo0P00003o0?ooofl0oooo00100?ooo`800000o`3ooom_0?ooo`00@@3oool010000000oooo0?oo
o`00003o0?ooof`0oooo00110?ooo`040000003oool0oooo00000?l0ooooK03oool00440oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?ooofL0oooo00110?ooo`800000103oool0
0`000000oooo0?ooo`3o0?ooofL0oooo00110?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0o`3ooomS0?ooo`00@@3oool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?oo
o`000000o`3ooomQ0?ooo`00@@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0?l0
ooooG`3oool00440oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?oooe`0oooo
00110?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0o`3ooomL0?ooo`00@@3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0200oooo00<0okmd0000000000000`000000
1@00M;l0oooo0?ooo`3ogi`0B0000080000000H007Bo0?ooo`3oool0oooo0?oo_`1d000500000005
07Boo`3oool0oooo0?ooo`3o_g@00`0000000`1d_ol0oooo0?ooo`0W0?ooo`030?noM00000000000
00<0000000D007Bo0?ooo`3oool0omnL04P0000200000003001d_`3oool0oooo00@0oooo00D0omnL
04P000000000000007Boo`030?ooo`050?noM000000000000000001d_ol0:@3oool00`3o_g@00000
0000000300000005001d_`3oool0oooo0?oOW01800000P0000000`00M;l0oooo0?ooo`040?ooo`05
0;md0000000000000000M02oool00`3oool01@3o_g@00000000000000000M;oo02P0oooo00<0okmd
0000000000000`0000001@00M;l0oooo0?ooo`3ogi`0B0000080000000<007Bo0?ooo`3oool00`3o
ool01@2oM000000000000000000007Bo00@0oooo00D0okmd000000000000000007Boo`0i0?ooo`00
@@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo02<0oooo00<0W4Q809cOo`3oool0
1`3oool00`2oM00004RL0?ooo`020?ooo`030;md000007@0_ooo0080oooo00<0giad09cOo`3oool0
0P3oool01P2LB7@0_ooo0?ooo`3ookl0M01d0;ooob/0oooo00<0W4Q809cOo`3oool01`3oool00`2o
M00004RL0?ooo`060?ooo`030;md001d_ol0oooo00<0oooo00H0W4Qd0;ooo`3oool0oono07@0M02o
ool[0?ooo`0309a8B02Lgol0oooo00L0oooo00<0_g@00018W03oool00`3oool01P3OW4P007Bo0?oo
o`3oool0okmd001d_`<0oooo00H0W4Qd0;ooo`3oool0oono07@0M02ooolZ0?ooo`0309a8B02Lgol0
oooo00L0oooo00<0_g@00018W03oool00P3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdB02Lgol0
oooo0080oooo00H0W4Qd0;ooo`3oool0oono07@0M02ooolh0?ooo`00@@3oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo01l0oooo00<0W4Q809cOo`3oool0203oool00`2oM000B9cO0?oo
o`020?ooo`030?noM00004P0W=oo00D0oooo00<0okmd001d_`3oool00P3oool00`3OW4P0B9cO0?oo
o`0Z0?ooo`0309a8B02Lgol0oooo00P0oooo00<0_g@004RLg`3oool01@3oool00`2oM000M;oo0?oo
o`020?ooo`030?noM000M;l0oooo0080oooo00<0gia804RLg`3oool0:P3oool00`2LB4P0W=oo0?oo
o`080?ooo`070;md0018W=l0oooo0?ooo`3ookl0M00007Boo`020?ooo`060?oo_`1dB9`0oooo0?oo
o`3o_g@007Bo0`3oool00`3OW4P0B9cO0?ooo`0Y0?ooo`0309a8B02Lgol0oooo00P0oooo00H0_g@0
04RLg`3oool0oooo0=nLB018W=l30?ooo`060;md001d_ol0oooo0?ooo`3o_g@007Bo0`3oool00`3O
W4P0B9cO0?ooo`0g0?ooo`00@@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01l0
oooo00<0W4Q809cOo`3oool0203oool00`3o_g@004RL0?ooo`030?ooo`030?oOW0180000B9cO00@0
oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0Z0?ooo`0309a8B02Lgol0oooo00P0
oooo00D0okmd0018W03oool0oooo0=nLB0050000000507Boo`3oool0oooo0=nLB018W=l00`3oool0
0`3o_g@007Bo0?ooo`0Z0?ooo`0309a8B02Lgol0oooo00P0oooo00/0okmd0018W03oool0oooo0?no
M000B4P0M01d0;ooo`3oool0gia8001d_`020?ooo`030=nLB018W=l0oooo0080oooo00<0okmd001d
_`3oool0:@3oool00`2LB4P0W=oo0?ooo`080?ooo`060?noM000B9`0oooo0?ooo`3o_g@004RL0`3o
ool01P2LB4P0W=oo0?ooo`3oool0gia804RLg`<0oooo00<0okmd001d_`3oool0=`3oool00440oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0L0?ooo`0309a8B02Lgol0oooo00@0oooo
00H0okmd000000000000000004QdM018B9`60?ooo`0309a80018W=l0oooo0080oooo00<0gia804RL
g`3oool00P3oool00`3o_g@007Bo0?ooo`0Z0?ooo`0309a8B02Lgol0oooo00@0oooo00H0okmd0000
00000000000004QdM018B9`20?ooo`060?noM00007@0_ooo0?ooo`2oM000M;oo0`3oool00`3OW4P0
B9cO0?ooo`020?ooo`030?noM000M;l0oooo02X0oooo00<0W4Q809cOo`3oool0103oool01P3o_g@0
0000000000000000B7Ad04Q8W080oooo00<0omnL04Q8W03o_g@00P0000001P00B9`0oooo0?ooo`3o
ool0gia804RLg`<0oooo00<0okmd001d_`3oool0:@3oool00`2LB4P0W=oo0?ooo`040?ooo`060?no
M0000000000000000018M7@0B4RL0`3oool01@3OW4P00000000000000000B9cO00<0oooo00<0gia8
04RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0g0?ooo`00@@3oool2000001X0oooo00<000000?oo
o`3oool0703oool00`2LB4P0W=oo0?ooo`030?ooo`030?oo_`1d04P0W=oo0080oooo00<0W4P00018
W03oool01P3oool01P2LB7@0_ooo0?ooo`3oool0gia804RLg`<0oooo00<0okmd001d_`3oool0:P3o
ool00`2LB4P0W=oo0?ooo`030?ooo`030?oo_`1d04P0W=oo0080oooo00<0W4P00018W03oool00P3o
ool01@3OW4P007Bo0?ooo`2oM000M;oo00<0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo
0?ooo`0Z0?ooo`0309a8B02Lgol0oooo00<0oooo00<0oono07@0B02Lgol00P3oool01`2LB00004RL
0?ooo`3oool0oono07@0M02oool01P3oool00`3OW4P0B9cO0?ooo`020?ooo`030?noM000M;l0oooo
02T0oooo00<0W4Q809cOo`3oool00`3oool00`3ookl0M01809cOo`020?ooo`0609a80000B9`0oooo
0?ooo`3ookl0M4RL0`3oool01P2LB7@0_ooo0?ooo`3oool0gia804RLg`<0oooo00<0okmd001d_`3o
ool0=`3oool00440oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0H0?ooo`0309a8
B02Lgol0oooo00<0oooo00<0okmd001d_`3oool00P3oool01P3ogi`0M9cO0?ooo`3oool0gia807Bo
o`<0oooo00H0_g@007Boo`3oool0oooo0?noM018W=l30?ooo`030=nLB018W=l0oooo02X0oooo00<0
W4Q809cOo`3oool00`3oool00`3o_g@007Bo0?ooo`020?ooo`030?oOW01dW=l0oooo00<0oooo00@0
W4P004RLg`2oM000M;oo0`3oool00`3o_g@0B9cO0?ooo`020?ooo`030=nLB018W=l0oooo02X0oooo
00<0W4Q809cOo`3oool00`3oool00`3o_g@007Bo0?ooo`020?ooo`030?oOW01dW=l0oooo0080oooo
00<0_g@004RLg`3oool01@3oool00`3o_g@0B9cO0?ooo`020?ooo`030=nLB018W=l0oooo02T0oooo
00<0W4Q809cOo`3oool00`3oool00`3o_g@007Bo0?ooo`020?ooo`060?oOW01dW=l0oooo0?ooo`3o
_g@0B9cO0`3oool01P2oM000M;oo0?ooo`3oool0okmd04RLg`<0oooo00<0gia804RLg`3oool0=`3o
ool00440oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0E0?ooo`050=nLB0000000
04RL09a8B02Lgol0103oool01`3ookl0M01809cOo`3oool0oono07@0B02Lgol00P3oool01`3o_g@0
001d0;ooo`3oool0okmd0000B02Lgol00`3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;ooobP0oooo
00D0gia800000000B9`0W4Q809cOo`040?ooo`070?oo_`1d04P0W=oo0?ooo`3ookl0M01809cOo`04
0?ooo`040?oOW01804P0B4P007Boo`@0oooo00H0W4P007Boo`3oool0omnL04P0M02ooolX0?ooo`05
0=nLB000000004RL09a8B02Lgol0103oool01`3ookl0M01809cOo`3oool0oono07@0B02Lgol00`3o
ool00`3ookl0M000001d_`060?ooo`0609a8001d_ol0oooo0?oOW01807@0_ooo9`3oool01@3OW4P0
00000018W02LB4P0W=oo00@0oooo00L0oono07@0B02Lgol0oooo0?oo_`1d04P0W=oo0080oooo00L0
omnL04P0M02oool0oooo0?oOW01807@0_ooo00<0oooo00H0W4P007Boo`3oool0omnL04P0M02ooolh
0?ooo`00@@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01<0oooo00<0gia80000
B02Lgol01@3oool01@3o_g@00000000000000000M;oo00@0oooo00D0okmd000000000000001809cO
o`040?ooo`050?oo_`1d000000000000B02Lgol0:`3oool00`3OW4P0001809cOo`050?ooo`050?no
M000000000000000001d_ol01P3oool00`3o_g@0000007Boo`040?ooo`050?oo_`1d000000000000
B02Lgol0:`3oool00`3OW4P0001809cOo`050?ooo`050?noM000000000000000001d_ol01P3oool0
102oM000000000000000B9`30?ooo`050?oo_`1d000000000000B02Lgol0:P3oool00`3OW4P00018
09cOo`050?ooo`050?noM000000000000000001d_ol0103oool01@3o_g@000000000000004P0W=oo
00@0oooo00D0oono07@000000000001809cOo`0i0?ooo`00@@3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo0?l0ooooB@3oool00440oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`3o0?ooodH0oooo00110?ooo`030000003oool0oooo02@0oooo00D000000?ooo`3oool0
oooo0000003o0?oood@0oooo00110?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
o`3ooom20?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0
oooo02`0oooo00<000000?ooo`3oool0o`3oooln0?ooo`00@@3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo0?l0oooo?P3oool00440oooo00<000000?ooo`3oool0;`3oool00`000000
oooo0?ooo`3o0?oooc/0oooo000h0?ooool00000N@0000000440oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00h0oooo0P00000A0?ooo`030000003o
ool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo00l0oooo00110?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`160?ooo`00@@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0o`3oool/0?ooo`00@@3oool00`000000oooo0?ooo`0k0?ooo`04
0000003oool0oooo00000?l0oooo;P3oool00440oooo00<000000?ooo`3oool0>`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0?l0oooo:03oool00440oooo00<000000?ooo`3oool0
@P3oool00`000000oooo0?ooo`3o0?ooobP0oooo00110?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool0o`3ooolT0?ooo`00@@3oool00`000000oooo0?ooo`160?ooo`030000003oool0
oooo0?l0oooo903oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool00`000000
oooo0?ooo`190?ooo`030000003oool0oooo0?l0oooo8@3oool00440oooo0P00001:0?ooo`050000
003oool0oooo0?ooo`000000o`3ooolO0?ooo`00@@3oool00`000000oooo0?ooo`1=0?ooo`030000
003oool0oooo0?l0oooo7@3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool0
0`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0?l0oooo6@3oool00440oooo00<000000?oo
o`3oool0D@3oool00`000000oooo0?ooo`3o0?oooaT0oooo00110?ooo`030000003oool0oooo05@0
oooo00<000000?ooo`3oool0o`3ooolF0?ooo`00@@3oool00`000000oooo0?ooo`1D0?ooo`030000
003oool0oooo0?l0oooo5P3oool00440oooo00<000000?ooo`3oool0F03oool00`000000oooo0?oo
o`3o0?oooa80oooo00110?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0o`3ooolB
0?ooo`00@@3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0?l0oooo3P3oool00440
oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`3o0?ooo`h0oooo00110?ooo`030000
003oool0oooo05l0oooo00<000000?ooo`3oool0o`3oool;0?ooo`00@@3oool200000600oooo00D0
00000?ooo`3oool0oooo0000003o0?ooo`T0oooo00110?ooo`030000003oool0oooo06<0oooo00<0
00000?ooo`3oool0o`3oool70?ooo`00@@3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0
oooo0?l0oooo0`3oool00440oooo00<000000?ooo`3oool0I`3oool010000000oooo0?ooo`00003o
0?ooo`80oooo00110?ooo`030000003oool0oooo06X0oooo00D000000?ooo`3oool0oooo0000003m
0?ooo`00@@3oool00`000000oooo0?ooo`1^0?ooo`050000003oool0oooo0?ooo`000000n@3oool0
0440oooo00<000000?ooo`3oool0LP3oool010000000oooo0?ooo`00003f0?ooo`00@@3oool00`00
0000oooo0?ooo`1e0?ooo`050000003oool0oooo0?ooo`000000lP3oool00440oooo00<000000?oo
o`3oool0N@3oool01@000000oooo0?ooo`3oool000000>h0oooo00110?ooo`030000003oool0oooo
07d0oooo00@000000?ooo`3oool00000j`3oool00440oooo00<000000?ooo`3oool0P03oool01@00
0000oooo0?ooo`3oool000000>L0oooo00110?ooo`800000Q@3oool01@000000oooo0?ooo`3oool0
00000><0oooo00110?ooo`030000003oool0oooo08P0oooo00@000000?ooo`3oool00000h03oool0
0440oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`3N0?ooo`00@@3oool00`000000
oooo0?ooo`2?0?ooo`030000003oool0oooo0=X0oooo00110?ooo`030000003oool0oooo08l0oooo
00D000000?ooo`3oool0oooo0000003H0?ooo`00@@3oool00`000000oooo0?ooo`2C0?ooo`030000
003oool0oooo0=H0oooo00110?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0d`3o
ool00440oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`3C0?ooo`00@@3oool00`00
0000oooo0?ooo`2J0?ooo`030000003oool0oooo0<l0oooo00110?ooo`030000003oool0oooo09X0
oooo00<000000?ooo`3oool0c`3oool00440oooo00<000000?ooo`3oool0WP3oool00`000000oooo
0?ooo`3;0?ooo`00@@3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo0</0oooo0011
0?ooo`800000XP3oool00`000000oooo0?ooo`380?ooo`00@@3oool00`000000oooo0?ooo`2Q0?oo
o`050000003oool0oooo0?ooo`000000aP3oool00440oooo00<000000?ooo`3oool0Y@3oool00`00
0000oooo0?ooo`340?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000
003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?oo
o`300?ooo`0000<0oooo0?noM0000000100000000`00M;l0oooo0?ooo`020?ooo`030;md00000000
07Bo00D0oooo00D0_g@0000000000000000007Boo`030?ooo`030;mdB02Lgol0oooo0080oooo00<0
okmG0000000000000`0000001@00M;l0oooo0?ooo`3oool0okmd00<0000000<0M;oo0?ooo`3oool0
3`3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo0<00oooo00040?ooo`0309a8B02L
gol0oooo00D0oooo00<0_g@000000000M;l0103oool01`2oM00007Bo0?ooo`3oool0omnL04P0001d
_ol00P3oool02P3ooml0W4Q809cOo`3oool0oono07@0M02oool0oooo09a8B02Lgol50?ooo`0609a8
M02oool0oooo0?oo_`1d07@0_ooo403oool00`000000oooo0?ooo`2/0?ooo`030000003oool0oooo
0;d0oooo00040?ooo`0309a8B02Lgol0oooo0140oooo00<0okmd001d_`3oool00P3oool01@3ooml0
W4Qd0;oog`2LB7@0_ooo0080oooo00<0W4Q809cOo`3oool00`3oool00`3o_g@007Bo0?ooo`020?oo
o`030=nLB018W=l0oooo00l0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`2m0?oo
o`00103oool00`2LB4P0W=oo0?ooo`0A0?ooo`030?noM000M;l0oooo00<0oooo00<0oono07@0B02L
gol00`3oool00`2LB4P0W=oo0?ooo`030?ooo`030=nLB018W=l0oooo0080oooo00<0okmd001d_`3o
ool03`3oool00`000000oooo0?ooo`3o0?ooofd0oooo00040?ooo`0309a8B02Lgol0oooo0140oooo
00<0W4P007Boo`3oool00P3oool01@3ookl0M01d0;oog`2LB4P0W=oo0080oooo00<0W4Q809cOo`3o
ool00`3oool00`3OW4P0B9cO0?ooo`020?ooo`030?noM000M;l0oooo00l0oooo00<000000?ooo`3o
ool0/03oool00`000000oooo0?ooo`2i0?ooo`00103oool00`2LB4P0W=oo0?ooo`0<0?ooo`030?oo
_`1d000000000080000000h0B9cO0?ooo`3oool0oooo0?oo_`1d07@0_ooo0?ooo`3ooml0W4Q809cO
o`3oool0W4Q809cOo`@0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0?0?ooo`<0
0000/03oool00`000000oooo0?ooo`2i0?ooo`00103oool00`2LB4P0W=oo0?ooo`0<0?ooo`030?oo
_`1d07@0_ooo00H0oooo00<0_gAd0;ooo`3oool00P3oool01@3ooml0W7BL0=ooo`2LB4P0W=oo00@0
oooo00<0okmd04RLg`3oool00P3oool00`3OW4P0B9cO0?ooo`0?0?ooo`030000003oool0oooo0?l0
ooooK@3oool000060?ooo`3OW4P000000018W02LB4P0W=oo3@3oool00`3ookl0M01d0;ooo`0;0?oo
o`050=nLB000000004RL09a8B02Lgol01@3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;oooa00oooo
00<000000?ooo`3oool0o`3ooom]0?ooo`000`3oool00`3OW4P0001809cOo`0=0?ooo`030?oo_`1d
000000000080000000<0001d0;ooo`3oool0203oool00`3OW4P0001809cOo`050?ooo`050?oo_`1d
000000000000B02Lgol00`3oool01@3OW4P00000000000000000M;oo00T0oooo00<000000?ooo`3o
ool0]03oool00`000000oooo0?ooo`2e0?ooo`00<`3oool01@2LM9`0oooo0?ooo`3ookl0M7Bo00T0
oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2e0?ooo`00<`3oool01@2LB7@0_ooo
0?ooo`3ogi`0B7Bo00T0oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00<`3oool01@3o_g@00000
0000000004P0W=oo00T0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`2a0?ooo`00
<`3oool01@2oM7@0_ooo0?ooo`3ogi`0B7Bo00T0oooo00<000000?ooo`3oool0^03oool00`000000
oooo0?ooo`2a0?ooo`00<`3oool01@2oM4P0W=oo0?ooo`3o_g@007Bo00T0oooo00<000000?ooo`3o
ool0o`3ooom]0?ooo`00<`3oool01@3ogi`0B0000000000007@0_ooo00T0oooo00<000000?ooo`3o
ool0^`3oool00`000000oooo0?ooo`2^0?ooo`00@@3oool00`000000oooo0?ooo`2k0?ooo`030000
003oool0oooo0:h0oooo00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo0P000030
0?ooo`030000003oool0oooo0:X0oooo00110?ooo`030000003oool0oooo0;l0oooo00<000000?oo
o`3oool0ZP3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool00`000000oooo
0?ooo`330?ooo`030000003oool0oooo0:H0oooo00110?ooo`030000003oool0oooo0<<0oooo00<0
00000?ooo`3oool0YP3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool00`00
0000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3o
ool0X`3oool00440oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`2S0?ooo`00@@3o
ool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0oooo0<X0oooo00<00000
0?ooo`3oool0W`3oool00440oooo0P00003;0?ooo`030000003oool0oooo09l0oooo00110?ooo`03
0000003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3o
ool00`000000oooo0?ooo`3>0?ooo`030000003oool0oooo09/0oooo00110?ooo`030000003oool0
oooo0<h0oooo00<000000?ooo`3oool0V`3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?oo
o`00@@3oool00`000000oooo0?ooo`3A0?ooo`030000003oool0oooo09P0oooo00110?ooo`030000
003oool0oooo0=40oooo00<000000?ooo`3oool0V03oool00440oooo00<000000?ooo`3oool0o`3o
oom]0?ooo`00@@3oool00`000000oooo0?ooo`3E0?ooo`030000003oool0oooo09@0oooo00110?oo
o`030000003oool0oooo0=D0oooo00<000000?ooo`3oool0U03oool00440oooo00<000000?ooo`3o
ool0o`3ooom]0?ooo`00@@3oool200000?l0ooooKP3oool00440oooo00<000000?ooo`3oool0f@3o
ool00`000000oooo0?ooo`2@0?ooo`00@@3oool00`000000oooo0?ooo`3I0?ooo`030000003oool0
oooo0900oooo00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3o
ool0g03oool00`000000oooo0?ooo`2=0?ooo`00@@3oool00`000000oooo0?ooo`3L0?ooo`030000
003oool0oooo08d0oooo00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo00<00000
0?ooo`3oool0h03oool00`000000oooo0?ooo`290?ooo`00@@3oool00`000000oooo0?ooo`3P0?oo
o`030000003oool0oooo08T0oooo00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo
00<000000?ooo`3oool0i03oool00`000000oooo0?ooo`250?ooo`00@@3oool200000>D0oooo00<0
00000?ooo`3oool0Q@3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool00`00
0000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3o
ool0PP3oool00440oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`220?ooo`00@@3o
ool00`000000oooo0?ooo`3[0?ooo`030000003oool0oooo07h0oooo00110?ooo`030000003oool0
oooo0>/0oooo00<000000?ooo`3oool0OP3oool00180oooo00<0oono07@000000000100000001@1d
_ol0oooo0?ooo`2oM4P0W=oo00<0oooo00<0okmG0000000000000`0000001@00M;l0oooo0?ooo`3o
ool0okmd00<0000000<0M;oo0?ooo`3oool03`3oool00`000000oooo0?ooo`3o0?ooofd0oooo000C
0?ooo`030;md000007@0_ooo0080oooo00h0giad09cOo`3oool0oooo0?oog`2LB4P0W=oo0?ooo`3o
okl0M01d0;ooo`3oool0W4Q809cOo`D0oooo00H0W4Qd0;ooo`3oool0oono07@0M02oool@0?ooo`03
0000003oool0oooo0>l0oooo00<000000?ooo`3oool0NP3oool001@0oooo00<0okmd0000B02Lgol0
1P3oool01@3ooml0W4Qd0;oog`2LB7@0_ooo0080oooo00<0W4Q809cOo`3oool00`3oool00`3o_g@0
07Bo0?ooo`020?ooo`030=nLB018W=l0oooo00l0oooo00<000000?ooo`3oool0k`3oool00`000000
oooo0?ooo`1j0?ooo`005@3oool00`3ogi`0B00004RLg`060?ooo`030?oo_`1d04P0W=oo00<0oooo
00<0W4Q809cOo`3oool00`3oool00`3OW4P0B9cO0?ooo`020?ooo`030?noM000M;l0oooo00l0oooo
00<000000?ooo`3oool0lP3oool00`000000oooo0?ooo`1g0?ooo`005`3oool00`2LB000B9cO0?oo
o`030?ooo`050?oo_`1d07@0_ooO09a8B02Lgol00P3oool00`2LB4P0W=oo0?ooo`030?ooo`030=nL
B018W=l0oooo0080oooo00<0okmd001d_`3oool03`3oool00`000000oooo0?ooo`3b0?ooo`030000
003oool0oooo07L0oooo000H0?ooo`0>09a8M02oool0oooo0?ooo`3ookl0M01d0;ooo`3oool0oooO
09a8B02Lgol0oooo09a8B02Lgol40?ooo`030=nLB018W=l0oooo0080oooo00<0okmd001d_`3oool0
3`3oool300000?l0ooooK@3oool001<0oooo00<0gia807Boo`3oool00P3oool01P2oM000M;oo0?oo
o`3oool0_gAd0;ooo`<0oooo00D0oooO09adW03Oool0W4Q809cOo`040?ooo`030?noM018W=l0oooo
0080oooo00<0gia804RLg`3oool03`3oool00`000000oooo0?ooo`3f0?ooo`030000003oool0oooo
07<0oooo000C0?ooo`070?noM00007@0_ooo0?ooo`3o_g@0001809cOo`070?ooo`050=nLB0000000
04RL09a8B02Lgol01@3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;ooo`80oooo00D0gia800000000
0000000007Boo`090?ooo`030000003oool0oooo0?H0oooo00<000000?ooo`3oool0L`3oool001@0
oooo00D0okmd000000000000001809cOo`0:0?ooo`030=nLB00004P0W=oo00D0oooo00D0oono07@0
00000000001809cOo`030?ooo`0509adW03oool0oooo0?oo_`1dM;l02@3oool00`000000oooo0?oo
o`3j0?ooo`030000003oool0oooo06l0oooo000c0?ooo`0509a8M02oool0oooo0?oOW018M;l02@3o
ool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo06l0oooo000c0?ooo`050?noM0000000
00000000B02Lgol02@3oool00`000000oooo0?ooo`3o0?ooofd0oooo000c0?ooo`050;mdM02oool0
oooo0?oOW018M;l02@3oool00`000000oooo0?ooo`3m0?ooo`030000003oool0oooo06`0oooo000c
0?ooo`050;mdB02Lgol0oooo0?noM000M;l02@3oool00`000000oooo0?ooo`3m0?ooo`030000003o
ool0oooo06`0oooo000c0?ooo`050?oOW018000000000000M02oool02@3oool00`000000oooo0?oo
o`3o0?ooofd0oooo00110?ooo`030000003oool0oooo0?l0oooo0P3oool00`000000oooo0?ooo`1X
0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooo`80oooo00<000000?ooo`3oool0J03oool00440
oooo0P00003o0?ooo`L0oooo00<000000?ooo`3oool0I03oool00440oooo00<000000?ooo`3oool0
o`3oool60?ooo`030000003oool0oooo06@0oooo00110?ooo`030000003oool0oooo0?l0ooooK@3o
ool00440oooo00<000000?ooo`3oool0o`3oool90?ooo`030000003oool0oooo0640oooo00110?oo
o`030000003oool0oooo0?l0oooo2@3oool00`000000oooo0?ooo`1Q0?ooo`00@@3oool00`000000
oooo0?ooo`3o0?ooo`d0oooo00<000000?ooo`3oool0G@3oool00440oooo00<000000?ooo`3oool0
o`3oool=0?ooo`030000003oool0oooo05d0oooo00110?ooo`030000003oool0oooo0?l0oooo4@3o
ool00`000000oooo0?ooo`1I0?ooo`00@@3oool00`000000oooo0?ooo`3o0?oooa40oooo00<00000
0?ooo`3oool0F@3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool00`000000
oooo0?ooo`3o0?oooa@0oooo00<000000?ooo`3oool0EP3oool00440oooo00<000000?ooo`3oool0
o`3ooolD0?ooo`030000003oool0oooo05H0oooo00110?ooo`800000o`3ooolI0?ooo`030000003o
ool0oooo0580oooo00110?ooo`030000003oool0oooo0?l0oooo603oool00`000000oooo0?ooo`1B
0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0oooo0?l0
oooo703oool00`000000oooo0?ooo`1>0?ooo`00@@3oool00`000000oooo0?ooo`3o0?oooa`0oooo
00<000000?ooo`3oool0CP3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool0
0`000000oooo0?ooo`3o0?oooal0oooo00<000000?ooo`3oool0B`3oool00440oooo00<000000?oo
o`3oool0o`3ooolO0?ooo`030000003oool0oooo04/0oooo00110?ooo`030000003oool0oooo0?l0
oooo8`3oool00`000000oooo0?ooo`170?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooob<0oooo
00<000000?ooo`3oool0A`3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00@@3oool2
00000?l0oooo:03oool00`000000oooo0?ooo`130?ooo`00@@3oool00`000000oooo0?ooo`3o0?oo
obL0oooo00<000000?ooo`3oool0@`3oool00440oooo00<000000?ooo`3oool0o`3ooolZ0?ooo`03
0000003oool0oooo0400oooo00110?ooo`030000003oool0oooo0?l0oooo:P3oool00`000000oooo
0?ooo`100?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0
oooo0?l0oooo;P3oool00`000000oooo0?ooo`0l0?ooo`00@@3oool00`000000oooo0?ooo`3o0?oo
obh0oooo00<000000?ooo`3oool0?03oool00440oooo00<000000?ooo`3oool0o`3ooolb0?ooo`03
0000003oool0oooo03P0oooo00110?ooo`030000003oool0oooo0?l0oooo<P3oool00`000000oooo
0?ooo`0h0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000003oool0
oooo0?l0oooo=P3oool00`000000oooo0?ooo`0d0?ooo`00@@3oool00`000000oooo0?ooo`3o0?oo
ocH0oooo00<000000?ooo`3oool0=03oool00440oooo0P00003o0?ooocX0oooo00<000000?ooo`3o
ool0<@3oool00440oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0340oooo
00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3oool0o`3ooolm
0?ooo`030000003oool0oooo02d0oooo00110?ooo`030000003oool0oooo0?l0oooo?@3oool00`00
0000oooo0?ooo`0]0?ooo`00@@3oool00`000000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3o
ool0:@3oool000030?oo_`1d0000000000@0000000<0M;oo0?ooo`3oool00P3oool00`2oM0000000
001d_`050?ooo`050;md0000000000000000001d_ol00`3oool00`2oM4P0W=oo0?ooo`020?ooo`03
0?noE`000000000000<0000000D007Bo0?ooo`3oool0oooo0?noM0030000000307Boo`3oool0oooo
00l0oooo00<000000?ooo`3oool0o`3ooom10?ooo`030000003oool0oooo02T0oooo0000103oool0
_g@00000M02oool20?ooo`030=nLM02Lgol0oooo00<0oooo00<0_g@000000000M;l0103oool01`2o
M00007Bo0?ooo`3oool0omnL04P0001d_ol00P3oool02P3ooml0W4Q809cOo`3oool0oono07@0M02o
ool0oooo09a8B02Lgol50?ooo`0609a8M02oool0oooo0?oo_`1d07@0_ooo403oool00`000000oooo
0?ooo`3o0?ooofd0oooo00020?ooo`030?noM00004P0W=oo01<0oooo00<0okmd001d_`3oool00P3o
ool01@3ooml0W4Qd0;oog`2LB7@0_ooo0080oooo00<0W4Q809cOo`3oool00`3oool00`3o_g@007Bo
0?ooo`020?ooo`030=nLB018W=l0oooo00l0oooo00<000000?ooo`3oool0o`3ooom40?ooo`030000
003oool0oooo02H0oooo00030?ooo`030?oOW0180000B9cO0180oooo00<0okmd001d_`3oool00`3o
ool00`3ookl0M01809cOo`030?ooo`0309a8B02Lgol0oooo00<0oooo00<0gia804RLg`3oool00P3o
ool00`3o_g@007Bo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooA03oool00`000000oooo0?oo
o`0V0?ooo`001@3oool00`2LB000B9cO0?ooo`0@0?ooo`0309a8001d_ol0oooo0080oooo00D0oono
07@0M02ooml0W4Q809cOo`020?ooo`0309a8B02Lgol0oooo00<0oooo00<0gia804RLg`3oool00P3o
ool00`3o_g@007Bo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooK@3oool000H0oooo00<0W4Qd
0;ooo`3oool02P3oool00`3ookl0M000000000020000000>04RLg`3oool0oooo0?ooo`3ookl0M01d
0;ooo`3oool0oooO09a8B02Lgol0oooo09a8B02Lgol40?ooo`030=nLB018W=l0oooo0080oooo00<0
okmd001d_`3oool03`3oool300000?l0ooooB03oool00`000000oooo0?ooo`0R0?ooo`0000<0oooo
0=nLB01d_ol00`3oool00`2oM000M;oo0?ooo`0:0?ooo`030?oo_`1d07@0_ooo00H0oooo00<0_gAd
0;ooo`3oool00P3oool01@3ooml0W7BL0=ooo`2LB4P0W=oo00@0oooo00<0okmd04RLg`3oool00P3o
ool00`3OW4P0B9cO0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooB03oool00`000000oooo0?oo
o`0R0?ooo`0000P0oooo0?noM00007@0_ooo0?ooo`3o_g@0001809cOo`/0oooo00<0oono07@0M02o
ool02`3oool01@3OW4P000000018W02LB4P0W=oo00D0oooo00H0W4P007Boo`3oool0omnL04P0M02o
ool@0?ooo`030000003oool0oooo0?l0ooooK@3oool00080oooo00D0okmd000000000000001809cO
o`0<0?ooo`030?oo_`1d000000000080000000<0001d0;ooo`3oool0203oool00`3OW4P0001809cO
o`050?ooo`050?oo_`1d000000000000B02Lgol00`3oool01@3OW4P00000000000000000M;oo00T0
oooo00<000000?ooo`3oool0o`3ooom<0?ooo`030000003oool0oooo01h0oooo000c0?ooo`0509ad
W03oool0oooo0?oo_`1dM;l02@3oool00`000000oooo0?ooo`3o0?oood`0oooo00<000000?ooo`3o
ool07P3oool003<0oooo00D0W4Qd0;ooo`3oool0omnL04Qd_`090?ooo`030000003oool0oooo0?l0
ooooK@3oool003<0oooo00D0okmd000000000000001809cOo`090?ooo`030000003oool0oooo0?l0
ooooC`3oool00`000000oooo0?ooo`0K0?ooo`00<`3oool01@2oM7@0_ooo0?ooo`3ogi`0B7Bo00T0
oooo00<000000?ooo`3oool0o`3ooom?0?ooo`030000003oool0oooo01/0oooo000c0?ooo`050;md
B02Lgol0oooo0?noM000M;l02@3oool00`000000oooo0?ooo`3o0?ooofd0oooo000c0?ooo`050?oO
W018000000000000M02oool02@3oool00`000000oooo0?ooo`3o0?oooe<0oooo00<000000?ooo`3o
ool05`3oool00440oooo00<000000?ooo`3oool0o`3ooomC0?ooo`030000003oool0oooo01L0oooo
00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo0P00003o0?oooeP0oooo00<00000
0?ooo`3oool04`3oool00440oooo00<000000?ooo`3oool0o`3ooomG0?ooo`030000003oool0oooo
01<0oooo00110?ooo`030000003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3oool0
o`3ooomJ0?ooo`030000003oool0oooo0100oooo00110?ooo`030000003oool0oooo0?l0ooooFP3o
ool00`000000oooo0?ooo`0@0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?oo
o`030000003oool0oooo0?l0ooooGP3oool00`000000oooo0?ooo`0<0?ooo`00@@3oool00`000000
oooo0?ooo`3o0?oooeh0oooo00<000000?ooo`3oool0303oool00440oooo00<000000?ooo`3oool0
o`3ooom]0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooof80oooo00<000000?ooo`3oool0203o
ool00440oooo00<000000?ooo`3oool0o`3ooomR0?ooo`030000003oool0oooo00P0oooo00110?oo
o`800000o`3ooom^0?ooo`00@@3oool00`000000oooo0?ooo`3o0?ooofd0oooo00110?ooo`030000
003oool0oooo0?l0ooooK@3oool00440oooo00<000000?ooo`3oool0o`3ooom]0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {1882.21, 5.63443*^07, \
0.407964, 1.30347*^06}}]
}, Closed]],

Cell["\<\
This entire procedure can now be packaged into a single module in \
case we might need to apply it to a file of similar format.  Additional \
arguments can be added to specify the desired columns.  With relatively minor \
modifications this module can be adapted to handle files of similar \
structure.\
\>", "Text"],

Cell[BoxData[
    \(GetPopData[file_String, nSkipFront_Integer, nSkipBack_Integer] := 
      Module[{pop}, 
        pop = OpenRead[
            file]; \(Part[#, 
              2, {2, 3}] &\)\  /@ \((\((Read[#, {Word, 
                      Table[Number, {5}]}] &)\) /@ \ \((StringToStream\  /@ \ \
\(\((StringReplace[#, "\<,\>" \[Rule] "\<\>"] &)\) /@ \ 
                    Flatten[
                      Drop[Drop[
                          ReadList[pop, Record, RecordLists \[Rule] True], 
                          nSkipFront], \(-nSkipBack\)]]\))\))\)]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(popdata = 
      GetPopData["\<Additional topics\popclock.txt\>", 5, 11]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Read::"readn"\), \(\(:\)\(\ \)\), "\<\"Syntax error reading a \
real number from \\!\\(System`Private`StringTurnedIntoStream[\\\" July 1 1900 \
          76094000                 ---                 ---\\\"]\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Read::readn\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({{1997, 267636061}, {1996, 265179411}, {1995, 262760639}, {1994, 
        260292437}, {1993, 257752702}, {1992, 255001827}, {1991, 
        252124362}, {1990, 249439545}, {1989, 246819230}, {1988, 
        244498982}, {1987, 242288918}, {1986, 240132887}, {1985, 
        237923795}, {1984, 235824902}, {1983, 233791994}, {1982, 
        231664458}, {1981, 229465714}, {1980, 227224681}, {1979, 
        225055487}, {1978, 222584545}, {1977, 220239425}, {1976, 
        218035164}, {1975, 215973199}, {1974, 213853928}, {1973, 
        211908788}, {1972, 209896021}, {1971, 207660677}, {1970, 
        205052174}, {1969, 202676946}, {1968, 200706052}, {1967, 
        198712056}, {1966, 196560338}, {1965, 194302963}, {1964, 
        191888791}, {1963, 189241798}, {1962, 186537737}, {1961, 
        183691481}, {1960, 180671158}, {1959, 177829628}, {1958, 
        174881904}, {1957, 171984130}, {1956, 168903031}, {1955, 
        165931202}, {1954, 163025854}, {1953, 160184192}, {1952, 
        157552740}, {1951, 154877889}, {1950, 152271417}, {1949, 
        149188130}, {1948, 146631302}, {1947, 144126071}, {1946, 
        141388566}, {1945, 139928165}, {1944, 138397345}, {1943, 
        136739353}, {1942, 134859553}, {1941, 133402471}, {1940, 
        132122446}, {1939, 130879718}, {1938, 129824939}, {1937, 
        128824829}, {1936, 128053180}, {1935, 127250232}, {1934, 
        126373773}, {1933, 125578763}, {1932, 124840471}, {1931, 
        124039648}, {1930, 123076741}, {1929, 121767000}, {1928, 
        120509000}, {1927, 119035000}, {1926, 117397000}, {1925, 
        115829000}, {1924, 114109000}, {1923, 111947000}, {1922, 
        110049000}, {1921, 108538000}, {1920, 106461000}, {1919, 
        104514000}, {1918, 103208000}, {1917, 103268000}, {1916, 
        101961000}, {1915, 100546000}, {1914, 99111000}, {1913, 
        97225000}, {1912, 95335000}, {1911, 93863000}, {1910, 
        92407000}, {1909, 90490000}, {1908, 88710000}, {1907, 
        87008000}, {1906, 85450000}, {1905, 83822000}, {1904, 
        82166000}, {1903, 80632000}, {1902, 79163000}, {1901, 
        77584000}, {1900, 76094000}}\)], "Output"]
}, Closed]],

Cell["\<\
Plot the annual change in population, either absolute or relative, \
using these data.\
\>", "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Import and Export", "Section"],

Cell[TextData[{
  "Versatile built-in functions for importing and exporting data using \
standardized file formats became available in Version 4 and more recent \
versions handle a greater variety of file formats.  In fact, there are so \
many formats and conversion options that we must refer you to the reference \
guide for the details of ",
  ButtonBox["Import",
    ButtonStyle->"RefGuideLink"],
  " and ",
  ButtonBox["Export",
    ButtonStyle->"RefGuideLink"],
  ".  The import formats available on your machine are listed by"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \($ImportFormats\)], "Input"],

Cell[BoxData[
    \({"AIFF", "APS", "AU", "Bit", "BMP", "Byte", "Character16", 
      "Character8", "Complex128", "Complex256", "Complex64", "CSV", "DICOM", 
      "DIF", "Dump", "DXF", "EPSTIFF", "Expression", "ExpressionML", "FITS", 
      "GIF", "HarwellBoeing", "HDF", "HDF5", "Integer128", "Integer16", 
      "Integer32", "Integer64", "Integer8", "JPEG", "Lines", "List", "MAT", 
      "MathML", "MGF", "MPS", "MTX", "NB", "NotebookML", "PBM", "PCX", "PGM", 
      "PNG", "PNM", "PPM", "Real128", "Real32", "Real64", "RSS", "SDTS", 
      "SND", "STL", "String", "SymbolicXML", "Table", "TerminatedString", 
      "Text", "TIFF", "TSV", "UnicodeText", "UnsignedInteger128", 
      "UnsignedInteger16", "UnsignedInteger32", "UnsignedInteger64", 
      "UnsignedInteger8", "VCF", "WAV", "Words", "XBitmap", "XLS", 
      "XML"}\)], "Output"]
}, Closed]],

Cell["\<\
and a similar list of export formats can also be obtained in the \
obvious manner.  To illustrate these functions, let us first export the \
population data from the previous section to a data files in our test \
directory.\
\>", "Text"],

Cell[BoxData[{
    \(\(SetDirectory[testdir];\)\), "\n", 
    \(\(Export["\<popdata.dat\>", popdata];\)\)}], "Input"],

Cell["\<\
The default format is based upon the file extension, which here \
simply contains a column formatted table, as shown below.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(!! popdata.dat\)\)], "Input"],

Cell["\<\
1997   267636061
1996   265179411
1995   262760639
1994   260292437
1993   257752702
1992   255001827
1991   252124362
1990   249439545
1989   246819230
1988   244498982
1987   242288918
1986   240132887
1985   237923795
1984   235824902
1983   233791994
1982   231664458
1981   229465714
1980   227224681
1979   225055487
1978   222584545
1977   220239425
1976   218035164
1975   215973199
1974   213853928
1973   211908788
1972   209896021
1971   207660677
1970   205052174
1969   202676946
1968   200706052
1967   198712056
1966   196560338
1965   194302963
1964   191888791
1963   189241798
1962   186537737
1961   183691481
1960   180671158
1959   177829628
1958   174881904
1957   171984130
1956   168903031
1955   165931202
1954   163025854
1953   160184192
1952   157552740
1951   154877889
1950   152271417
1949   149188130
1948   146631302
1947   144126071
1946   141388566
1945   139928165
1944   138397345
1943   136739353
1942   134859553
1941   133402471
1940   132122446
1939   130879718
1938   129824939
1937   128824829
1936   128053180
1935   127250232
1934   126373773
1933   125578763
1932   124840471
1931   124039648
1930   123076741
1929   121767000
1928   120509000
1927   119035000
1926   117397000
1925   115829000
1924   114109000
1923   111947000
1922   110049000
1921   108538000
1920   106461000
1919   104514000
1918   103208000
1917   103268000
1916   101961000
1915   100546000
1914   99111000
1913   97225000
1912   95335000
1911   93863000
1910   92407000
1909   90490000
1908   88710000
1907   87008000
1906   85450000
1905   83822000
1904   82166000
1903   80632000
1902   79163000
1901   77584000
1900   76094000\
\>", "Print"]
}, Closed]],

Cell["\<\
Alternatively, we can export this data to an Excel file as follows. \
 Open the output file with Excel to verify that this operation succeeded.  \
Obviously, this method is easier that stream input when the file formats are \
simple or standard; on the other hand, stream input/output is more \
flexible.\
\>", "Text"],

Cell[BoxData[
    \(\(Export["\<popdata.xls\>", popdata, "\<XLS\>"];\)\)], "Input"],

Cell["\<\
Similarly, we can import pictures or other specialized files either \
with standard filename extensions or by specifying the appropriate format as \
follows.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(me = 
        Import[EssentialMathematica <> "\<\Additional \
topics\jjk.jpg\>"];\)\), "\[IndentingNewLine]", 
    \(\(Show[me];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .94813 
%%ImageSize: 348 330 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
3.27429e-017 0.00288184 -0.00288184 0.00288184 [
[ 0 0 0 0 ]
[ 1 .94813 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .94813 L
0 .94813 L
closepath
clip
newpath
% Start of colorimage (RGB)
p
0 -0.00288 translate
1.00288 .95101 scale
1044 string
348 330 8 [348 0 0 330 0 0] {
\tcurrentfile
\t1 index
\treadhexstring
\tpop
}  false 3 Mcolorimage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P
% End of image
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{348, 330},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008:0001kb000`400?l00000o`00003oo`3ooooo
0?ooo``0oooo000304=:D0<0@T]B0`11Be@303m<E@04045>E`0lBE80?4UB045;E080Ae1G00@0?TE;
02L[;P0W:bh051DG0P0<3@l0100A4A40410@010@40092@T200P8200400`<300820P020P800L71`80
4A4A00@07QhN030`<00`<300;Rh^0P0Z:RX0100T92@09BDU02DU9@0`<30202TY:@0402LW:@0X:2X0
:2PZ02DU9`80:BT[00@0=3@f030`<P0`<380820R0P0H61X0100D51H0<30b030`<P0Y:B/202PX:P04
03Ph>P0j>S`0>SXl03De=`80:b/]00@0;Rh^02PX:00X:2P0:2PX0P0_;bl0100];Bd08b<S02<S8`0P
820202DU9@0402PX:00g=cL0=cLg024Q8@809BDU00@07QhN02@T900T92@0<C4a0P0V9RH302LW9`04
02@T900Z:RX0:RXZ03<c<`80;Bd]00@08b<S02DU9@0U9BD08b<S0P0l?3`0100W9bL0:b/[02/[:`0Y
:BT202`/;00403Lg=`0b<S80<S8b02@T9080:BTY00@0820P02<S8`0S8b<0@D510P0I6AT0100O7al0
?clo03lo?`0P820202@T9004028R8P0W9bL09bLW03<c<`80<C4a00@092@T01lO7`0O7al0:2PZ0P0Y
:B/00`18B4X0820R020P8P0302PX:P0402`/;P0[:bd0:b/]0492A0808R8T00@09BDW02PX:P0X:2X0
820R0P0d=3H0100U9BL092@V02@T9P0O7b4202<S9@0403/k?@0T92H092@V02PX:P807AdO00@0<30b
02@T9P0T92H09RHX0P0U9BL0100V9RP0=cLi03Lg>@0S8bD2024Q8`0402`/;P0V9RP09RHX0410@P80
:b/]00@09BDW02LW:@0W9bT0:b/]0P0U9BL01016ATP09bLY02LW:@0/;2h202PX:P0402LW:@0d=3H0
=3@f02`/;P80820R00@0:BT[02LW:@0W9bT0>3Pj0P0S8bD0100Y:B/09RHX02HV:00/;2h202@T9P04
02HV:00f=SP0=SHh02HV:0805aLI00@04Q8D01hN800N7R00:2PZ0P0R8R@0100G5aT04Q8D018B500G
5aD201lO7@0402TY9`092@L02@T700D50`808R8P00@07AdK01`L6P0L71X09RHV0P0Q8B40100S8b<0
:b/]02/[;@0Z:Bh202DT:@0402PW;@0b<CL0<S4g04=3A@<0:2PZ0P0T92H0100^;S00:RX/02XZ;00]
;Bl203`l?P0402LW:@0_;c40;bla02TY:`80:RX/00@0>CTk03@d=P0d=3H09bLY0P0Z:R`0100m?Cl0
?cm103lo@@10@48203/k?@0403@d=P0^;S00;Rh`02d];`80;bla00@0:b/]04I6B016ATP0:RX/0P0[
:bd0100V9RP0<C4c034a<`0T92H204Y:C00402l];P0`;Rl0<2h_038`<@<0<Bl`0P12@440100_;Bh0
9bDV02LU9P0[:BX2030^;`04040n?`0c<C80<c4b02HT9@80;bd^00@0<2h_02l];P0_;Bh0BTQ90P0`
;Rl0100R8240=C<d03Dc=00[:BX203Hd=@0403lm?P0[:BX0:bTZ030^;`80;R`]00<09R@U04A2@`14
@T<00`0^;2d0100R8240>SPi03Xh>@0U8b@204E3A00403lm?P0X9RL0:2HW02/Y:P80;R`]00@0;bd^
040n?`10?Sl09bDV0P0k>CX0100d<S<09bDV02LU9P0X9RL2034_<00404a:B`0`;Rl0<2h_0490@@80
;R`]00@0<2h_034_<00a;c00<S0a0P0h=SL0100c<C80<S0a038`<@0c<C82044o@00403Le=P0e<c@0
=C<d03Tg>080>3Hg00@0@3ho044o@011?d00;B//0P0b<34303Hd=@0404M5AP15@d@0AD=402d[;080
<2h_00@0;B//038`<@0b<340@d520P0h=SL0100]:b`0=S@e03Hd=@0d<S<2030^;`0404M5AP0d<S<0
=38c02d[;080>3Hg00@0?cdn03Le=P0g=CH0<S0a0P0e<c@0100j>3T0=cDf03Le=P0P7Ql205aJF`04
03Xh>@0W9BH09bDV03Tg>080;2X[00@0@3ho03dk?00m>c`0;bd^0P0^;2d0100c<C80<2h_030^;`18
ATL203Dc=0<0;R`]00@0;B//03Dc=00e<c@0@3ho0P0k>CX0100b<340=C<d03Dc=00g=CH203Xh>@04
03dk?00i=cP0>CLh02d[;0809R@U00@0;B//03Xh>@0j>3T0AD=40P0^;2d0100Z:2T0=C<d03Dc=00e
<c@202h/;@0402`Z:`0Z:2T0:RPY038`<@80?cdn00@0?3Xk030^;`0`;Rl0<Bl`0P0d<S<0100c<C80
=C<d03Dc=00i=cP201hL7@40EE=D0@15@d@10?ooo`000`1:DEL304UBF@<0B59K0`16De`01018EEh0
AE9K04EBF`1;EEh2055JH@0404E<DP0X;2l0:2`_00h?4@800P<50`0=3@d300L71`0400d=3@0:2PX0
2PX:00X:2P805ADE00@0820P02h^;P0^;Rh0:2PX0P0Q8B40100L71`07alO01lO7`0^;Rh202/[:`04
028R900/;2h0;2`^028R90805QHH00@09BDW03Lg>@0g=cT0<S8d0P0U9BL0100@4180;2`^02`/;P0V
9RP202@T9P0403@d=P0f=SP0=SHh038b=0809bLY00@07QhN01/K6`0K6a/08B4Q0`0/;2`2024Q8@04
01XJ6P0M7Ad07AdM02TY:@80=CDe00@06a/K01`L700L71`05QHF0P0O7al0100`<300:2PX02PX:00R
8R8202<S8`04028R8P0Y:BT0:BTY03@d=080;bl_00@09RHV02TY:@0Y:BT092@T0P0o?cl0100Z:RX0
9bLW02LW9`0L71`201hN7P04034a<@0e=CD0=CDe024Q8@809RHV00@071`L020P800P8200?Cdm0P0F
5QH0100K6a/0>c/k03/k>`0L71`2020P800401lO7`0T92@092@T030`<080;Bd]00@0820P01`L700L
71`08B4S0P0R8R@00`11@D<06ATK01TI6`03024Q8`0402DU9`0T92H092@V0492A0808R8T00@09BDW
02PX:P0X:2X0820R0P0d=3H0100U9BL092@V02@T9P0O7b4202<S9@0403/k?@0T92H092@V02PX:P80
7AdO00@0<30b02@T9P0T92H09bLY0`0U9BL203@d=P0401lO8@0K6ad06a/M02DU9`807QhP00@0?Cdo
02PX:P0X:2X0820R0P0Q8B<0100T92H06a/M01/K7@0k>cd301`L7P806a/M00@0820R038b=00b<S@0
:b/]0P0N7R00100T92H07alQ01lO8@0a<C<201`L7P04028R900O7b407alQ02DU9`807AdO00@07alQ
02l_<@0_;c4071`N0P0J6Q`0100N7R009BDW02DU9`0Q8B<201HF600401DE5`0L71h071`N02/[:@80
;bl]00@0=3@b00l?3@0?3`d02PX80P0X:2H0100V9R@0:2PV02PX9P0/;2`202<S8`04020P800S8bD0
8b<U024P9@807A`Q00@08b8X030_=@0`;cD0B4Q:0P0Z:R`0100V9RP07alQ01lO8@0X:2X202LW:@04
02l_<@11@D<0@D5302<S9@80;2`^00@09RHX02LW:@0W9bT0=CDg0P0`<380100S8bD09RHX02HV:00d
=3H203Xj?00403lo@@0m?Cl0?Cdo03De=`80;Bd_00@0:2PZ02HV:00V9RP0:b/]0P16ATP0100Z:R`0
:b/]02/[;@0V9RP2034a<`0402@T9P1:BT`0BTY<02/Y:P80;B//00@0;bd^02h/;@0^;2d0;R`]0P0n
?3d0100/:R/08b4R02<Q8P0X9RL202d[;00403dk?00`;Rl0<2h_028P8@80:bTZ00@0;2X[02/Y:P0[
:BX0BdU:0P0_;Bh0100M6a`0;bd^02l];P0W9BH203Hd=@0404E3A00d<S<0=38c038`<@80<c4b00@0
;bd^051>C`1@CTl0>SPi0P0g=CH0100W9BH0?3Xk03`j>`0]:b`204a:B`0404E3A00/:R/0;2X[02d[
;0<0;bd^0P0o?Ch0100S8B80=cDf03Le=P0b<34202HT9@0402TW:00c<C80<c4b051>C`80=38c00@0
CD]<02l];P0_;Bh09bDV0P0V92D0100/:R/0=S@e03Hd=@0_;Bh202/Y:P0403Dc=00o?Ch0?cdn034_
<080;2X[00@0<2h_03<a<P0c<C80?cdn0P13@D800`0Z:2T0;bd^02l];P03038`<@0404=1@P12@440
@T1102TW:080;B//00@0:BLX02h/;@0^;2d0?cdn0P0e<c@0100Y9bP0<S0a038`<@0`;Rl202`Z:`04
04M5AP0d<S<0=38c02`Z:`80=S@e00@0>cTj038`<@0b<340:bTZ0P0^;2d0100a;c00<S0a038`<@0N
71d205iLG@0403`j>`0V92D09R@U03<a<P80928S00@0AD=4040n?`10?Sl0;B//0P0Y9bP302h/;@04
04Y8B@0k>CX0>cTj02XX:@80:bTZ00@0:RPY034_<00a;c00?3Xk0P0g=CH00`0_;Bh0<Bl`034_<003
02l];P0403<a<P0f=3D0=S@e03<a<P80;B//00@0;2X[030^;`0`;Rl0=38c0P12@440100c<C809R@U
02HT9@0a;c0202/Y:P0401hL7@0X9RL0:2HW04U7B080?3Xk00@0:RPY02HT9@0V92D0;2X[0P0Y9bP0
100[:BX0>CLh03Tg>00n?3d205iLG@40?3Xk0@0I5aP10?ooo`0000@0A51P04=@H013D600@e1P0`11
DF42041AH`<0@55U00@0@e5L049@F`12D5/0AU=L0P1>F640101;E5/0=Se303Hm@`0K82@200P<3`<0
2PX:00<02@T900/;2`0;2`/00`0=3@d0100:2PX01`L700L71`0I6AT202l_;`0402d];@0S8b<08b<S
01/K6`80820P00@0<C4a02d];@0];Bd0820R0P0R8R@0100W9bT092@V02@T9P0T92H203Hf>00403`l
?P0[:bd0:b/]01HF6080>3Pj00@0:RX/020P8P0P8280:RX/0P0_;c40100f=SP0;2`^02`/;P0T92@2
01`L700402h^;P0Y:BT0:BTY030`<0809BDU00@08B4Q02DU9@0U9BD08B4Q0P0j>SX0100P82009RHV
02HV9P0N7Qh502d];@0402TY:@0T92@092@T02<S8`80>3Ph00@0;Bd]034a<@0a<C40;bl_0P0Y:BT0
100/;2`0<C4a034a<@0^;Rh202PX:00402XZ:P0O7al07alO024Q8@80BDU900@06a/K028R8P0R8R80
9BDU0P0L71`0100c<c<0=3@d03@d=00G5aL202l_;`0402XZ:P0T92@092@T01hN7P8071`L00@0;bl_
03<c<`0c<c<07AdM0P0P8200100P828071`N01`L7P1;Bdd201dM7`0402l_<@0Y:B/0:BT[02LW:@80
:RX/00@0D51B03@d=P0d=3H0>3Pj0P0[:bd0100e=CL0>3Pj03Ph>P17AdT202XZ;004030`<P0V9RP0
9RHX0492A0809BDW00@0;Bd_020P8P0P8280?Cdo0P0X:2X0100[:bd0;Rh`02h^<00V9RP203/k?@04
024Q8`0P8280820R02d];`806ATK00@0>3Pj02DU9`0U9BL08B4S0P0S8bD0100U9BL06a/M01/K7@1<
C4h301lO8@807QhP00@08B4S03<c=@0c<cD0;Bd_0P0M7Al302<S9@0403Lg>@0P8280820R02<S9@80
9RHX00@09BDW02LW:@0W9bT08b<U0P0X:2X0100H61X092@V02@T9P0H61X201LG6@0401hN800M7Al0
7AdO020P8P806ATK00@0=CDc02<S8@0S8b40<30^0P020P0010000000=CDc03De<`0N7Q`202PX9P04
03De=@0Y:BT0:BTY02@T90807AdO00@0:bX_01dL8@0M72407ahT0P0W9R`01019BD/0:RX/02XZ;00W
9bT201`L7P04034a<`0G5aT05aLI024Q8`80Bd]=00@09RHX02/[;@0[:bd0820R0P0_;c40100a<C<0
;2`^02`/;P0j>S`202h^<00402PX:P0_;c40;bla03Lg>@80>SXl00@0@T9402l_<@0_;c40=3@f0P0W
9bT0100U9BL0CdmA04m?D@0N7R02034a<`0402<S9@0a<C<0<C4c02/[;@80Bd]=00@0:RPY02TW:00Y
9bP0<2h_0P0]:b`0100^;2d0AdE604M5AP0Y9bP202HT9@0402d[;00P7Ql081hO0490@@80>CLh00@0
:2HW02HT9@0V92D0<c4b0P0[:BX0101JF5T0:BLX02TW:00^;2d2030^;`0402h/;@0c<C80<c4b06AR
H`80=cDf00@0>3Hg03/i>P0k>CX0;R`]0P14@T<0101HEUL0<S0a038`<@0d<S<203Tg>004030^;`0m
>c`0?C/l04I4A@80;R`]00@0<S0a03Tg>00i=cP0;2X[0P18ATL0100W9BH0<2h_030^;`0f=3D202TW
:004038`<@0a;c00<Bl`06ESI080;bd^00@0EE=D02TW:00Y9bP0:BLX0P0X9RL0100R8240=cDf03Le
=P0Z:2T2028P8@0402`Z:`11?d00@Cm002@R8`<0;2X[0P0Y9bP0100c<C80?3Xk03`j>`0`;Rl202HT
9@0402XX:@0]:b`0;B//03hl?@80BDM800@0:2HW02`Z:`0/:R/0;B//0P0X9RL01012@440>SPi03Xh
>@0U8b@203@b<`<0:bTZ00@0BDM802d[;00]:b`0:RPY0P0c<C80100k>CX09bDV02LU9P0T8R<202DS
90<0<Bl`00@0:bTZ03<a<P0c<C80BDM80P0^;2d0100[:BX0;bd^02l];P0]:b`204M5AP0402h/;@0W
9BH09bDV02/Y:P80;bd^00@0ATA503Xh>@0j>3T0;2X[0P0g=CH0100W9BH0<Bl`034_<00f=3D203Dc
=0<0;bd^00@0;B//03@b<`0d<S<09R@U0P0j>3T01010?Sl09bDV02LU9P0X9RL202XX:@04034_<018
ATL0B4I703Tg>080<Bl`00@07adN02PV9`0X9RL0<S0a0P0g=CH01019AdP0=cDf03Le=P0^;2d202PV
9`0402h/;@0Z:2T0:RPY038`<@80=S@e00D0AdE60490@@12@440=cDf02`Z:`010?ooo`0000@0A51P
04=@H013D600@e1P0`11DF42041AH`<0@55U00@0@e5L049@F`12D5/0AU=L0P1>F640101;E5/0=Se3
03Hm@`0K82@200P<3`<02PX:00<02@T900/;2`0;2`/00`0=3@d0100:2PX01`L700L71`0I6AT202l_
;`0402d];@0S8b<08b<S01/K6`80820P00@0<C4a02d];@0];Bd0820R0P0R8R@0100W9bT092@V02@T
9P0T92H203Hf>00403`l?P0[:bd0:b/]01HF6080>3Pj00@0:RX/020P8P0P8280:RX/0P0_;c40100f
=SP0;2`^02`/;P0T92@201`L700402h^;P0Y:BT0:BTY030`<0809BDU00@08B4Q02DU9@0U9BD08B4Q
0P0j>SX0100P82009RHV02HV9P0N7Qh502d];@0402TY:@0T92@092@T02<S8`80>3Ph00@0;Bd]034a
<@0a<C40;bl_0P0Y:BT0100/;2`0<C4a034a<@0^;Rh202PX:00402XZ:P0O7al07alO024Q8@80BDU9
00@06a/K028R8P0R8R809BDU0P0L71`0100c<c<0=3@d03@d=00G5aL202l_;`0402XZ:P0T92@092@T
01hN7P8071`L00@0;bl_03<c<`0c<c<07AdM0P0P8200100P828071`N01`L7P1;Bdd201dM7`0402l_
<@0Y:B/0:BT[02LW:@80:RX/00@0D51B03@d=P0d=3H0>3Pj0P0[:bd0100e=CL0>3Pj03Ph>P17AdT2
02XZ;004030`<P0V9RP09RHX0492A0809BDW00@0;Bd_020P8P0P8280?Cdo0P0X:2X0100[:bd0;Rh`
02h^<00V9RP203/k?@04024Q8`0P8280820R02d];`806ATK00@0>3Pj02DU9`0U9BL08B4S0P0S8bD0
100U9BL06a/M01/K7@1<C4h301lO8@807QhP00@08B4S03<c=@0c<cD0;Bd_0P0M7Al302<S9@0403Lg
>@0P8280820R02<S9@809RHX00@09BDW02LW:@0W9bT08b<U0P0X:2X0100H61X092@V02@T9P0H61X2
01LG6@0401hN800M7Al07AdO020P8P806ATK00@0=CDc02<S8@0S8b40<30^0P020P0010000000=CDc
03De<`0N7Q`202PX9P0403De=@0Y:BT0:BTY02@T90807AdO00@0:bX_01dL8@0M72407ahT0P0W9R`0
1019BD/0:RX/02XZ;00W9bT201`L7P04034a<`0G5aT05aLI024Q8`80Bd]=00@09RHX02/[;@0[:bd0
820R0P0_;c40100a<C<0;2`^02`/;P0j>S`202h^<00402PX:P0_;c40;bla03Lg>@80>SXl00@0@T94
02l_<@0_;c40=3@f0P0W9bT0100U9BL0CdmA04m?D@0N7R02034a<`0402<S9@0a<C<0<C4c02/[;@80
Bd]=00@0:RPY02TW:00Y9bP0<2h_0P0]:b`0100^;2d0AdE604M5AP0Y9bP202HT9@0402d[;00P7Ql0
81hO0490@@80>CLh00@0:2HW02HT9@0V92D0<c4b0P0[:BX0101JF5T0:BLX02TW:00^;2d2030^;`04
02h/;@0c<C80<c4b06ARH`80=cDf00@0>3Hg03/i>P0k>CX0;R`]0P14@T<0101HEUL0<S0a038`<@0d
<S<203Tg>004030^;`0m>c`0?C/l04I4A@80;R`]00@0<S0a03Tg>00i=cP0;2X[0P18ATL0100W9BH0
<2h_030^;`0f=3D202TW:004038`<@0a;c00<Bl`06ESI080;bd^00@0EE=D02TW:00Y9bP0:BLX0P0X
9RL0100R8240=cDf03Le=P0Z:2T2028P8@0402`Z:`11?d00@Cm002@R8`<0;2X[0P0Y9bP0100c<C80
?3Xk03`j>`0`;Rl202HT9@0402XX:@0]:b`0;B//03hl?@80BDM800@0:2HW02`Z:`0/:R/0;B//0P0X
9RL01012@440>SPi03Xh>@0U8b@203@b<`<0:bTZ00@0BDM802d[;00]:b`0:RPY0P0c<C80100k>CX0
9bDV02LU9P0T8R<202DS90<0<Bl`00@0:bTZ03<a<P0c<C80BDM80P0^;2d0100[:BX0;bd^02l];P0]
:b`204M5AP0402h/;@0W9BH09bDV02/Y:P80;bd^00@0ATA503Xh>@0j>3T0;2X[0P0g=CH0100W9BH0
<Bl`034_<00f=3D203Dc=0<0;bd^00@0;B//03@b<`0d<S<09R@U0P0j>3T01010?Sl09bDV02LU9P0X
9RL202XX:@04034_<018ATL0B4I703Tg>080<Bl`00@07adN02PV9`0X9RL0<S0a0P0g=CH01019AdP0
=cDf03Le=P0^;2d202PV9`0402h/;@0Z:2T0:RPY038`<@80=S@e00D0AdE60490@@12@440=cDf02`Z
:`010?ooo`0000@0@e5N045AH011DF00@E5P0`10DF4303mBH`80?E9S00@0@E5N04ABG@14DUd0AeAM
0P1<EUl0101;E5/0>d9803]2B00O92P200L;3P0400/;2`0:2PX02PX:00X:2P802`/;0`0=3@d0100;
2`/020P800P8200C4a<202d];@0402`/;00O7al07alO01PH60807QhN00<0;Rh^034a<@0a<C400`0W
9bT302XZ;00402DU9`0V9RP09RHX02LW:@808b<U00@03@d?03De=`0e=CL0:2PZ0P0R8R@302XZ;004
03<c=@0X:2X0:2PZ024Q8@8071`L00@0:b/[02LW9`0W9bL0<C4a0P0X:2P0100P82009BDU02DU9@0T
92@203/k>`04034a<@0[:b/0:b/[02<S8`80:RXZ00@0<C4a03Ti>@0i>CT0:2PX0P0U9BD0100M7Ad0
:2PX02PX:00X:2P202/[:`0401/K6`0M7Ad07AdM01hN7P809bLW00@0;Rh^028R8P0R8R808b<S0`0V
9RH20451@@0402@T900Q8B408B4Q028R8P806QXJ00@0;Bd]03<c<`0c<c<05QHF0P0L71`0100d=3@0
8b<S02<S8`0Q8B4202<S8`0402l_;`0e=CD0=CDe02LW9`809RHV00@0:b/]02PX:P0X:2X0D51B0P0Y
:B/0100f=SP0;bla02l_<@0^;S02030`<P0404I6B00T92H092@V02HV:0807QhP00@09RHX02DU9`0U
9BL0=cLi0P0Q8B<0100T92H07AdO01dM7`0c<cD202DU9`04028R900K6ad06a/M038b=080:2PZ00@0
9BDW02/[;@0[:bd092@V0P0h>3X0100Q8B<07QhP01hN800/;2h201/K7@0403/k?@0e=CL0=CDg02TY
:`80<30b00@0;Bd_02XZ;00Z:R`0DE5C0P0b<S@0100X:2X06ATK01TI6`0];Bl202h^<00402@T9P0V
9RP09RHX028R9080:b/]00@0=3@f01hN800N7R008R8T0P0U9BL3028R900401`L7P0P8280820R018B
50805QHH00@03Ph@01`L7P0L71h0<30b0P0a<C<0100_;c40:RX/02XZ;00K6aT202`/:P0403De<`07
1`D01`L500<30@80<c<a00@0;Rh/02`/:P0/;2X0;bl_0P0b<S80100W9bL08b<U02<S9@0X9b`2024P
9@0401dL8P0V9B/09RD[04=3A@80;bla00@092@V01/K7@0K6ad0;Bd_0P0U9BL0100V9RP0@d=504=3
A@0N7R02028R900402h^<00Q8B<08B4S01LG6@80=SHh00@0A4A6030`<P0`<380:BT[0P0V9RP0100W
9bT0=3@f03@d=P16ATP203@d=P0403De=`0_;c40;bla02<S9@80BTY<00@05aLI02TY:`0Y:B/08R8T
0P0Y:B/0100S8bD0DE5C055AD`0Z:2T202d[;00402h/;@0c<C80<c4b03<a<P80DU1A00@0<c4b03@b
<`0d<S<0=C<d0P0]:b`01017ADH0B4I704Q6A`0_;Bh2038`<@0403Tg>00c<C80<c4b06=QHP80?3Xk
00@0=C<d03Le=P0g=CH0<2h_0P0]:b`00`1LFU/0;bd^02l];P0302/Y:P0402LU9P0e<c@0=C<d055?
D080:2HW00@0;B//02l];P0_;Bh0:BLX0P0c<C800`12@4409B<T02DS900302`Z:`0402LU9P10?Sl0
@3ho02`Z:`80=38c00@0?3Xk02`Z:`0/:R/0<Bl`0P0_;Bh0101LFU/0;2X[02`Z:`1FE5D202h/;@04
02`Z:`0_;Bh0;bd^02/Y:P80?cdn00@0<c4b02h/;@0^;2d0:bTZ0P14@T<0100Y9bP0;bd^02l];P0`
;Rl202LU9P04034_<01=Bd`0CD]<02l];P80:2HW00@0;R`]030^;`0`;Rl0?S`m0P1>C4d0100]:b`0
<2h_030^;`0Y9bP202DS900403dk?00j>3T0>SPi02@R8`80<Bl`00@0:2HW02LU9P0W9BH0AdE60P0_
;Bh0100[:BX0=S@e03Hd=@11?d0202h/;@0402`Z:`0^;2d0;R`]030^;`80<S0a00@0:2HW02d[;00]
:b`0AdE60P0_;Bh0100X9RL0:bTZ02/Y:P0Y9bP203lm?P04034_<00W9BH09bDV02TW:080;bd^00@0
?cdn03/i>P0k>CX0;2X[0P0b<340100O7Ah0:RPY02XX:@0a;c0203Dc=003034_<00`;Rl0<2h_00<0
;2X[00@0;bd^034_<00a;c00>3Hg0P0g=CH0100[:BX0:RPY02XX:@0X9RL203Xh>@0403dk?010?Sl0
@3ho02/Y:P80=cDf00@0A49303Xh>@0j>3T07adN0P0`;Rl0100a;c00?3Xk03`j>`0W9BH2034_<004
02d[;00h=SL0>3Hg03Tg>080<Bl`0@0d<S<1040n?`40oooo0003045CG`<0@5=Q0`0nE68303eDI004
045AGP15EF80AEER04MEH080B5EN00@0BeAK04=:D013BU009R/_0P062`l0100<30`02PX:00X:2P0:
2PX200/;2`<03@d=00@02`/;00T92@092@T03@d=0P0Y:BT0100[:b/06a/K01/K6`0H61P201lO7`04
02TY:@0b<S80<S8b02DU9`8092@V00@09bLY02d];`0];Bl0:b/]0P0M7Al0100J6Q`09BDW02DU9`0:
2P`203Hf>00402@T9P0P8280820R02DU9`808B4S00<0=3@f02d];`0];Bl00`0K6a/0100W9bL09BDU
02DU9@0f=SH203<c<`0402PX:00_;bl0;bl_020P8080;bl_00@0?Cdm02HV9P0V9RH0820P0P0N7Qh0
100Z:RX0>c/k03/k>`0W9bL2028R8P0401/K6`0M7Ad07AdM02d];@80=CDe00@071`L02DU9@0U9BD0
7AdM0P0U9BD0100f=SH08b<S02<S8`0L71`202PX:004024Q8@0Z:RX0:RXZ03De=@80:BTY00@09bLW
024Q8@0Q8B40;2`/0P0i>CT0100S8b<04Q8B018B4P11@D4302XZ:P80;Rh^00@0=SHf0410@010@400
=cLg0P0a<C40100_;c40;Bd_02d];`19BD/202`/;P0402h^<00U9BL09BDW02<S9@808R8T00@0A4A6
028R900R8R@08B4S0P0N7R00100Q8B<07AdO01dM7`0c<cD2028R9004020P8P0M7Al07AdO02TY:`80
<C4c0`0O7b40100/;2h0<c<e03<c=@0Q8B<202`/;P0402TY:`0k>cd0>c/m02d];`80:BT[00@0>c/m
02l_<@0_;c40@4120P1:BT`0100^;S00=3@f03@d=P0S8bD2028R900403@d=P0Y:B/0:BT[02XZ;080
4a<E00@0>SXl02`/;P0/;2h05aLI0P0Y:B/0100Q8B<0<C4c034a<`14A4H302d];`80:RX/0`0S8bD0
100N7R008R8T028R900E5AL2024Q8`0402HV:00b<S@0<S8d03De=`809RHX00@092@V02/[;@0[:bd0
8b<Q0P0e=C<0100b<S002@T700T91`071`D202@T8P0402d]:`0S8b408b<Q02HV9P80?3`l00@0:2PX
02TY:`0Y:B/092<X0P0Z:Bh0100J6Al08b8X02<R:00^;S0203Hf>00402d];`0Z:R`0:RX/02d];`80
<C4c00@08R8T034a<`0a<C<0>3Pj0P0P8280100f=SP0:BT[02TY:`0P828304Q8BP80<c<e00@0:2PZ
02l_<@0_;c40;2`^0P0_;c401012@T@0>SXl03Xj?00c<cD202XZ;00402l_<@1FEUP0EUIH02<S9@80
<C4c0`0g=cT0100a<C<0K6a^06a/KP0i=cP203hl?@0403<a<P0n?3d0?S`m03Pf=`80FUQI00@0=cDf
03`j>`0l>S/0<2h_0P0^;2d0100l>S/0BdU:04]9BP0Z:2T203@b<`04030^;`0^;2d0;R`]04a:B`80
>SPi00@09B<T02`Z:`0/:R/0:2HW0P0P7Ql0101@CTl0:RPY02XX:@0]:b`202TW:0<0;bd^00@0EE=D
02PV9`0X9RL0;B//0`0[:BX2030^;`0404M5AP0T8R<0928S030^;`809B<T00@0;2X[03lm?P0o?Ch0
;R`]0P0d<S<01010?Sl0;R`]02h/;@0_;Bh202d[;00405=ADP0]:b`0;B//05=ADP80<S0a00@0:BLX
034_<00a;c00;2X[0P0l>S/0100a;c00<2h_030^;`0T8R<204=1@P<0;R`]00@0<c4b02<Q8P0S8B80
:RPY0P1GEEH0100Z:2T09bDV02LU9P0a;c02030^;`0403Tg>01BD540DU1A030^;`80<Bl`00@0:RPY
02TW:00Y9bP0?3Xk0P12@440100Z:2T0=38c03@b<`0[:BX202TW:004044o@00^;2d0;R`]02PV9`80
<S0a00@0?S`m02/Y:P0[:BX0:RPY0P0/:R/0100]:b`0=C<d03Dc=00U8b@202LU9P0404I4A@0h=SL0
>3Hg02/Y:P80;B//00@0:BLX03Le=P0g=CH0>CLh0P0Y9bP0100X9RL0<S0a038`<@0h=SL204=1@P04
038`<@0e<c@0=C<d020N7`80:RPY00@0;bd^03@b<`0d<S<0<c4b0P0a;c00100_;Bh09R@U02HT9@0f
=3D202/Y:P04034_<017ADH0AdE603<a<P80:bTZ00@0=C<d03Tg>00i=cP0A4930P1;BDX0100`;Rl0
=38c03@b<`0k>CX202PV9`<0:bTZ00@0>SPi01lM7P0O7Ah0<c4b0P0]:b`01019AdP0AdE604M5AP0^
;2d202TW:040<2h_0@0o?Ch10?ooo`000`11Del3041CH@<0?UAR0`0mE6@01011DEh0AEER04EEHP17
EF0204QEGP0404]DF`13BU00@dY@02H[;`801P/?00@030`<00X:2P0:2PX02PX:0P0;2`/300d=3@04
00/;2`092@T02@T900d=3@80:BTY00@0:b/[01/K6`0K6a/061PH0P0O7al0100Y:BT0<S8b038b<P0U
9BL202@T9P0402LW:@0];Bl0;Bd_02/[;@807AdO00@06QXL02DU9`0U9BL02PX<0P0f=SP0100T92H0
820R020P8P0U9BL2024Q8`0303@d=P0];Bl0;Bd_00<06a/K00@09bLW02DU9@0U9BD0=SHf0P0c<c<0
100X:2P0;bl_02l_;`0P820202l_;`0403dm?@0V9RH09RHV020P80807QhN00@0:RXZ03/k>`0k>c/0
9bLW0P0R8R80100K6a/07AdM01dM7@0];Bd203De=@0401`L700U9BD09BDU01dM7@809BDU00@0=SHf
02<S8`0S8b<071`L0P0X:2P0100Q8B40:RXZ02XZ:P0e=CD202TY:@0402LW9`0Q8B408B4Q02`/;080
>CTi00@08b<S018B4P0B4Q80@D510`0Z:RX202h^;P0403Hf=P10@400@41003Lg=`80<C4a00@0;bla
02d];`0];Bl0BDU;0P0/;2h0100^;S009BDW02DU9`0S8bD2028R900404A4AP0R8R@08R8T024Q8`80
7QhP00@08B4S01dM7`0M7Al0<c<e0P0R8R@0100P82807AdO01dM7`0Y:B/2034a<`<07alQ00@0;2`^
03<c=@0c<cD08B4S0P0/;2h0100Y:B/0>c/m03/k?@0];Bl202TY:`0403/k?@0_;c40;bla0410@P80
BTY<00@0;Rh`03@d=P0d=3H08b<U0P0R8R@0100d=3H0:BT[02TY:`0Z:R`201<C5@0403Xj?00/;2h0
;2`^01LG6@80:BT[00@08B4S034a<`0a<C<0A4A60`0];Bl202XZ;0<08b<U00@07QhP028R900R8R@0
5ADG0P0Q8B<0100V9RP0<S8d038b=00e=CL202HV:00402@T9P0[:bd0:b/]02<S8@80=CDc00@0<S8`
00T91`092@L01`L50P0T9280100];B/08b<Q02<S8@0V9RH203`l?00402PX:00Y:B/0:BT[02@S:080
:RT^00@06QTO02<R:00S8RP0;Rh`0P0f=SP0100];Bl0:RX/02XZ;00];Bl2034a<`04028R900a<C<0
<C4c03Ph>P80820R00@0=SHh02TY:`0Y:B/0820R0`18B4X203<c=@0402PX:P0_;c40;bla02`/;P80
;bla00@0@T9403Xj?00j>S`0<c<e0P0Z:R`0100_;c40EUIH05IFF00S8bD2034a<`<0=cLi00@0<C4c
06a/KP1/K6h0>CLh0P0n?3d0100c<C80?S`m03hl?@0h=SL205YHF@0403Le=P0l>S/0?3Xk030^;`80
;R`]00@0?3Xk04]9BP1;BDX0:RPY0P0d<S<0100`;Rl0;R`]02h/;@1<BT/203Xh>@0402DS900/:R/0
;2X[02PV9`8081hO00@0D4i?02XX:@0Z:2T0;B//0P0Y9bP302l];P0405ECE00X9RL0:2HW02d[;0<0
:bTZ0P0`;Rl01017ADH0928S02@R8`0`;Rl202DS900402`Z:`0o?Ch0?cdn02h/;@80=38c00@0@3ho
02h/;@0^;2d0;bd^0P0]:b`0101CDE80;B//02d[;01CDE82038`<@0402TW:00a;c00<Bl`02`Z:`80
?3Xk00@0<Bl`030^;`0`;Rl0928S0P13@D8302h/;@0403<a<P0S8B808b4R02XX:@80EeEF00@0:RPY
02LU9P0W9BH0<Bl`0P0`;Rl0100i=cP0DU1A059@D@0`;Rl2034_<00402XX:@0Y9bP0:BLX03`j>`80
@T1100@0:RPY03@b<`0d<S<0:bTZ0P0Y9bP01011?d00;R`]02h/;@0X9RL2038`<@0403hl?@0[:BX0
:bTZ02XX:@80;2X[00@0;B//03Dc=00e<c@09B<T0P0W9BH01016A4D0>3Hg03Pf=`0[:BX202d[;004
02TW:00g=CH0=cDf03Tg>080:BLX00@0:2HW038`<@0b<340>3Hg0P13@D80100b<340=C<d03Dc=00P
7Ql202XX:@0402l];P0d<S<0=38c03<a<P80<Bl`00@0;bd^02HT9@0V92D0=S@e0P0[:BX0100a;c00
AdE604M5AP0c<C8202/Y:P0403Dc=00i=cP0>CLh04A2@`80BdU:00@0<2h_03@b<`0d<S<0>cTj0P0X
9RL302/Y:P0403Xh>@0O7Ah07adN03<a<P80;B//00@0BDM804M5AP17ADH0;R`]0P0Y9bP1030^;`40
?cdn0@3oool00004041DG@10E5l0@5AO041DG`80?UAQ00<0?EEQ03eDHP0mE6800`0kEF801010DV00
AeMT04MGI018EV4204EBF`0404]EGP19DUT0BE9I02dd>P802a0D00@03@d=00/;2`0;2`/02PX:0P0;
2`/300d=3@0400`<300:2PX02PX:00d=3@809BDU00@0:b/[01/K6`0K6a/07QhN0P0W9bL0100U9BD0
;2`/02`/;00U9BL202HV:00402@T9P0[:bd0:b/]030`<P80820R00@061PJ02@T9P0T92H04Q8D0P0k
>cd0100U9BL07QhP01hN800O7b4201`L7P0303Ti>`0j>S`0>SXl00<0:b/[00@0;Bd]02TY:@0Y:BT0
=SHf0P0e=CD0100S8b<0:2PX02PX:00O7al202<S8`0403lo?`0Q8B408B4Q020P80805aLG00@092@T
03Hf=P0f=SH0=SHf0P0T92@0100R8R807alO01lO7`0[:b/203Ph>0<08R8R0`0M7Ad0100e=CD0:RXZ
02XZ:P0P820202d];@0402@T900U9BD09BDU03`l?080;Rh^00@0:b/[02DU9@0U9BD09RHV0P0g=cL0
100a<C405aLG01LG5`0/;2`202@T900402<S8`0N7Qh07QhN02`/;080>SXj00@0;2`/02DU9@0U9BD0
7AdO0P0O7b40100b<S@0:RX/02XZ;00U9BL2024Q8`0402@T9P0S8bD08b<U0410@P80;2`^00@09bLY
02@T9P0T92H092@V0P0R8R@00`0f=SP09RHX02HV:003020P8P04028R900l?3h0?3`n01dM7`808b<U
00@09BDW03/k?@0k>cd08B4S0P0_;c40100X:2X0<S8d038b=00X:2X201`L7P0402XZ;00Q8B<08B4S
024Q8`80?cm100@07AdO02/[;@0[:bd06QXL0P0T92H0100/;2h0>3Pj03Ph>P0Q8B<2010@4P0403Ti
>`0];Bl0;Bd_018B508092@V00@08R8T02PX:P0X:2X0:2PZ0P0H61X0100R8R@09bLY02LW:@0X:2X2
02l_<@0402d];`0a<C<0<C4c02/[;@80<S8d00@0;bla02XZ;00Z:R`09BDW0P0K6ad0100O7b40;Rh`
02h^<00Z:RP201`L6P0402PX9P061P@01PH400<30@805aLE00@092@R01lO7@0O7ad0820P0P10@400
100S8b<0:RX/02XZ;00R8BH203@c>0<07A`R00@08B4S03`l?P0l?3h0=CDg0P0f=SP0100V9RP0;Bd_
02d];`0M7Al202l_<@0403Lg>@0G5aT05aLI020P8P8092@V00@07alQ02`/;P0/;2h0<S8d0P0e=CL0
100I6A/0:b/]02/[;@0S8bD201/K7@0403`l?P1EEEL0EEEG04M7B@80:2PZ00@0=cLi05mOH@1OGf40
;Rh`0P0[:bd0100d=3H0;Rh`02h^<00S8bD205QHFP0403<a<P0e<c@0=C<d02<Q8P80<Bl`00@0:2HW
04Y8B@1:B4T0:BLX0P0_;Bh0100X9RL0:RPY02XX:@0^;2d204Q6A`0402<Q8P0`;Rl0<2h_02DS9080
9R@U00@0=cDf03hl?@0n?3d071XK0`0X9RL201hL7@0404]9BP0a;c00<Bl`03<a<P80;2X[00@0>CLh
02d[;00]:b`0F5IG0P0/:R/00`0_;Bh0:RPY02XX:@0302h/;@0404]9BP0Y9bP0:BLX03@b<`808AlP
00@0<Bl`03dk?00m>c`0;R`]0P0`;Rl0100n?3d0:RPY02XX:@0Y9bP202d[;00404a:B`0e<c@0=C<d
059@D@80?C/l00@0:RPY03<a<P0c<C80;B//0P0i=cP302d[;00402@R8`19AdP0BDM803/i>P80;bd^
00@0>SPi02`Z:`0/:R/0:RPY0P1CDE80100V92D0928S02@R8`0`;Rl202h/;@04034_<01CDE80De5B
030^;`80;bd^0`0[:BX0100h=SL0BDM804U7B00_;Bh203Dc=00402d[;00Y9bP0:BLX044o@080=38c
00@0:RPY034_<00a;c00?S`m0`0X9RL202XX:@0402LU9P0c<C80<c4b024O80807Q`M0`10?Sl302d[
;00402TW:00^;2d0;R`]03`j>`809bDV00@08b4R030^;`0`;Rl0;bd^0P1;BDX0100j>3T0?C/l03dk
?00/:R/203@b<`04030^;`0b<340<S0a03<a<P80<S0a00@0=38c02XX:@0Z:2T0=38c0P0^;2d0100b
<340AdE604M5AP0m>c`202d[;004034_<00]:b`0;B//03@b<`80?cdn00@0<c4b02h/;@0^;2d0<2h_
0P0[:BX0100^;2d0<Bl`034_<00M6a`203dk?00403Pf=`1HEUL0F5IG03`j>`80<c4b00D0:bTZ02d[
;00]:b`0;2X[02PV9`010?ooo`000`12E5h00`11EEh0@EEP045EH00303mEHP<0?UES00@0@59P04MI
I`17FFL0AUIS0P12D5/01018EEh0CeQO04mHG`0e?TD2018I7`0400h>3P0<30`030`<00X:2P802`/;
00@03@d=00h>3P0>3Ph030`<0P0;2`/0100@41007alO01lO7`0/;2`201hN7P0402@T900a<C40<C4a
02<S8`8092@T0`0^;S0302LW:@03030`<P0Y:B/0:BT[00<07AdO00@05aLI03`l?P0l?3h0;Bd_0`0R
8R@2020P8P0403Ph>P10@480@412034a<@80;Bd]00@0:BTY02HV9P0V9RH0;bl_0P0e=CD3028R8P04
02HV9P0P8200820P03`l?0809RHV00@0:2PX01dM7@0M7Ad09RHV0P0b<S80100k>c/07AdM01dM7@0U
9BD302<S8`80<c<c00@0<30`024Q8@0Q8B409bLW0P0N7Qh00`0b<S80=CDe03De=@0302XZ:P0402<S
8`0Q8B408B4Q030`<080;2`/00@09bLW01lO7`0O7al06QXJ0P0X:2P0100d=3@08R8R028R8P0F5QH2
02@T9004024Q8@0B4Q804Q8B02LW9`80>3Ph00@08b<S01hN7P0N7Qh071`N0P0Q8B<0100W9bT0=CDg
03De=`0U9BL202@T9P0402/[;@0X:2X0:2PZ038b=080=CDg00@0:b/]028R900R8R@092@V0P0U9BL0
100c<cD08R8T028R900N7R02020P8P0401/K7@0n?T00?Si001XJ7080820R00@06a/M03Ti>`0i>C/0
7QhP0P0^;S00100U9BL0:RX/02XZ;00X:2X201HF6004024Q8`0K6ad06a/M01TI6`80@D5300@07QhP
02XZ;00Z:R`06a/M0P0V9RP0100W9bT0@D530451@`0I6A/200h>400402LW:@0[:bd0:b/]01TI6`80
8B4S00@08R8T01@D5P0D51H0<C4c0P0W9bT0100c<cD0=CDg03De=`0`<38202l_<@0402DU9`0R8R@0
8R8T020P8P808b<U00@08B4S02<S9@0S8bD0<C4c0P0d=3H0100V9RP07QhP01hN800];B/2028R8004
05eMF`0U9B<09BDS00d=2`808b<Q00@0:RXX02LW9@0W9bD07alO0P0n?Sh0100M7Ad09bLY02LW:@0S
8RL203Xi?P0402@S:@0I61h06APN02HV:080>c/m00@0:RX/02`/;P0/;2h07AdO0P0V9RP0100S8bD0
?cm103lo@@0/;2h203Hf>00402XZ;00U9BL09BDW028R90808b<U00@0>SXl04=3A@13@dD04Q8D0P0I
6A/0100<30h020P:00P82P0];Bl205UIF`0404m?D@0d=3H0=3@f02XZ;080DU9D00@0;Rh`020P8P0P
8280;2`^0P0Y:B/0100M7Al0@4120410@P0`;Rl202XX:@0401/I6P0Y9bP0:BLX02@R8`80@d520`0`
;Rl0100[:BX0;R`]02h/;@0/:R/204Y8B@0302TW:00c<C80<c4b00<0:2HW00@0>cTj05MEEP1GEEH0
:RPY0P0c<C80100b<3409R@U02HT9@16A4D203`j>`0402d[;00U8b@09B<T03@b<`809B<T00@0Cde>
02l];P0_;Bh0;2X[0P0Y9bP0100/:R/0:RPY02XX:@16A4D202h/;@04034_<00O7Ah07adN030^;`80
=S@e00@0<c4b02d[;00]:b`0?3Xk0P0U8b@0100S8B80;B//02d[;011?d0203Tg>004051>C`1<BT/0
C4Y;02l];P80>3Hg00@0<Bl`03lm?P0o?Ch0<c4b0P0b<340100Z:2T0CD]<04e;C01;BDX203@b<`04
040n?`0k>CX0>cTj034_<080AD=400@0?3Xk03Dc=00e<c@0@Cm00P10?Sl00`0n?3d0IfEV06MUIP03
0490@@0403Hd=@0h=SL0>3Hg03dk?080FEMH00@0?S`m03lm?P0o?Ch0>CLh0P0e<c@01014@T<0?S`m
03hl?@0`;Rl203Le=P0404I4A@0_;Bh0;bd^034_<080<S0a00@0>3Hg04A2@`14@T<0<S0a0P0Y9bP0
1015@d@0DU1A059@D@0l>S/203Hd=@0403/i>P0g=CH0=cDf04Y8B@80=C<d00@0;bd^03/i>P0k>CX0
>3Hg0P1PGUl0100e<c@0?C/l03dk?00d<S<203/i>P0302h/;@0]:b`0;B//00<0<c4b00@0>SPi03Dc
=00e<c@0;2X[0P0f=3D0100g=CH0>SPi03Xh>@14@T<2034_<00402DS900S8B808b4R02DS9080<Bl`
00@0?cdn034_<00a;c009bDV0P0e<c@0100Y9bP05aDF01LE5P0/:R/202l];P0405mMGP15@d@0AD=4
03/i>P806ALH00D0;2X[02d[;00]:b`0:BLX02<Q8P010?ooo`000`12E5h00`11EEh0@EEP045EH003
03mEHP<0?UES00@0@59P04MII`17FFL0AUIS0P12D5/01018EEh0CeQO04mHG`0e?TD2018I7`0400h>
3P0<30`030`<00X:2P802`/;00@03@d=00h>3P0>3Ph030`<0P0;2`/0100@41007alO01lO7`0/;2`2
01hN7P0402@T900a<C40<C4a02<S8`8092@T0`0^;S0302LW:@03030`<P0Y:B/0:BT[00<07AdO00@0
5aLI03`l?P0l?3h0;Bd_0`0R8R@2020P8P0403Ph>P10@480@412034a<@80;Bd]00@0:BTY02HV9P0V
9RH0;bl_0P0e=CD3028R8P0402HV9P0P8200820P03`l?0809RHV00@0:2PX01dM7@0M7Ad09RHV0P0b
<S80100k>c/07AdM01dM7@0U9BD302<S8`80<c<c00@0<30`024Q8@0Q8B409bLW0P0N7Qh00`0b<S80
=CDe03De=@0302XZ:P0402<S8`0Q8B408B4Q030`<080;2`/00@09bLW01lO7`0O7al06QXJ0P0X:2P0
100d=3@08R8R028R8P0F5QH202@T9004024Q8@0B4Q804Q8B02LW9`80>3Ph00@08b<S01hN7P0N7Qh0
71`N0P0Q8B<0100W9bT0=CDg03De=`0U9BL202@T9P0402/[;@0X:2X0:2PZ038b=080=CDg00@0:b/]
028R900R8R@092@V0P0U9BL0100c<cD08R8T028R900N7R02020P8P0401/K7@0n?T00?Si001XJ7080
820R00@06a/M03Ti>`0i>C/07QhP0P0^;S00100U9BL0:RX/02XZ;00X:2X201HF6004024Q8`0K6ad0
6a/M01TI6`80@D5300@07QhP02XZ;00Z:R`06a/M0P0V9RP0100W9bT0@D530451@`0I6A/200h>4004
02LW:@0[:bd0:b/]01TI6`808B4S00@08R8T01@D5P0D51H0<C4c0P0W9bT0100c<cD0=CDg03De=`0`
<38202l_<@0402DU9`0R8R@08R8T020P8P808b<U00@08B4S02<S9@0S8bD0<C4c0P0d=3H0100V9RP0
7QhP01hN800];B/2028R800405eMF`0U9B<09BDS00d=2`808b<Q00@0:RXX02LW9@0W9bD07alO0P0n
?Sh0100M7Ad09bLY02LW:@0S8RL203Xi?P0402@S:@0I61h06APN02HV:080>c/m00@0:RX/02`/;P0/
;2h07AdO0P0V9RP0100S8bD0?cm103lo@@0/;2h203Hf>00402XZ;00U9BL09BDW028R90808b<U00@0
>SXl04=3A@13@dD04Q8D0P0I6A/0100<30h020P:00P82P0];Bl205UIF`0404m?D@0d=3H0=3@f02XZ
;080DU9D00@0;Rh`020P8P0P8280;2`^0P0Y:B/0100M7Al0@4120410@P0`;Rl202XX:@0401/I6P0Y
9bP0:BLX02@R8`80@d520`0`;Rl0100[:BX0;R`]02h/;@0/:R/204Y8B@0302TW:00c<C80<c4b00<0
:2HW00@0>cTj05MEEP1GEEH0:RPY0P0c<C80100b<3409R@U02HT9@16A4D203`j>`0402d[;00U8b@0
9B<T03@b<`809B<T00@0Cde>02l];P0_;Bh0;2X[0P0Y9bP0100/:R/0:RPY02XX:@16A4D202h/;@04
034_<00O7Ah07adN030^;`80=S@e00@0<c4b02d[;00]:b`0?3Xk0P0U8b@0100S8B80;B//02d[;011
?d0203Tg>004051>C`1<BT/0C4Y;02l];P80>3Hg00@0<Bl`03lm?P0o?Ch0<c4b0P0b<340100Z:2T0
CD]<04e;C01;BDX203@b<`04040n?`0k>CX0>cTj034_<080AD=400@0?3Xk03Dc=00e<c@0@Cm00P10
?Sl00`0n?3d0IfEV06MUIP030490@@0403Hd=@0h=SL0>3Hg03dk?080FEMH00@0?S`m03lm?P0o?Ch0
>CLh0P0e<c@01014@T<0?S`m03hl?@0`;Rl203Le=P0404I4A@0_;Bh0;bd^034_<080<S0a00@0>3Hg
04A2@`14@T<0<S0a0P0Y9bP01015@d@0DU1A059@D@0l>S/203Hd=@0403/i>P0g=CH0=cDf04Y8B@80
=C<d00@0;bd^03/i>P0k>CX0>3Hg0P1PGUl0100e<c@0?C/l03dk?00d<S<203/i>P0302h/;@0]:b`0
;B//00<0<c4b00@0>SPi03Dc=00e<c@0;2X[0P0f=3D0100g=CH0>SPi03Xh>@14@T<2034_<00402DS
900S8B808b4R02DS9080<Bl`00@0?cdn034_<00a;c009bDV0P0e<c@0100Y9bP05aDF01LE5P0/:R/2
02l];P0405mMGP15@d@0AD=403/i>P806ALH00D0;2X[02d[;00]:b`0:BLX02<Q8P010?ooo`000`15
De`304=DGP03049DH011E680@EAR00<0@EAR00@0@5=R04MII`17FFL0A5IT0P11CeX01018EEh0CeUR
04mIHP0mATd201lX;`0400l?3`0<30`030`<00X:2P802`/;00@03@d=00h>3P0>3Ph03@d=0P0<30`0
100@41005QHF01HF5P0^;Rh2024Q8@04028R8P0f=SH0=SHf02DU9@808R8R00@0;Bd_02`/;P0/;2h0
9BDW0P0S8bD0100^;S00=3@f03@d=P0];Bl202DU9`0401DE5`0e=CL0=CDg03Lg>@80:RX/00@0:2PZ
02LW:@0W9bT0;bla0P0k>cd0100/;2`08B4Q024Q8@0M7Ad201lO7`0402DU9@0g=cL0=cLg02TY:@80
92@T00@09BDU01hN7P0N7Qh0;Rh^0P0[:b/0100X:2P07alO01lO7`0R8R8202HV9P0303`l?00Q8B40
8B4Q00<0:RXZ00@092@T034a<@0a<C40@T920P0Z:RX0100`<3009RHV02HV9P0[:b/203/k>`04038b
<P0Q8B408B4Q020P80806a/K00@09bLW038b<P0b<S809bLW0P0Q8B40100K6a/0820P020P800d=3@2
03<c<`0401<C4`0V9RH09RHV02XZ:P806QXJ00@0:BTY03Xj>P0j>SX0:BTY0P0T92@0100T92H0:b/]
02/[;@0R8R@203lo@@0402<S9@0R8R@08R8T02PX:P808b<U00@0:RX/0492A012@T@0;bla0P0R8R@3
02TY:`04034a<`0V9RP09RHX02<S9@809RHX00@0820R0410@P10@4808B4S0P0U9BL0100M7Al0>SXl
03Xj?00K6ad202d];`0402<S9@0W9bT09bLY034a<`807alQ00@0:BT[02DU9`0U9BL08b<U0P1:BT`0
100Y:B/0:2PZ02PX:P0K6ad201dM7`0401`L7P0i>C/0>CTk01`L7P8030`>00@06ATK02DU9`0U9BL0
7alQ0P0R8R@0100O7b403@d?00d=3`0`<38202<S9@0402PX:P0S8bD08b<U01dM7`808R8T00@0820R
028R900R8R@04a<E0P0O7b4302TY:`0403@d=P0e=CL0=CDg020P8P805QHH00@0820N0492@012@T00
UIFC0P0^;R`010061P@0820N020P7P0T928201/K6@0402LW9`0m?Cd0?Cdm028R8P809RHX00@09B@Y
03He>P0f=CX0;bhd0P0J6Al0100S8bD0;bla02l_<@0N7R02028R900402HV:00X:2X0:2PZ02/[;@80
?cm100@07AdO0492A012@T@0:2PZ0P0I6A/0100S8bD0:b/]02/[;@0^;S02018B500400X:300;2`d0
2`/=00h>4080410B00@04a<E02h^<00^;S00@4120P1DE5H0100U9BL0@d=504=3A@0b<S@202<S9@04
030`<P0d=3H0=3@f030`<P80?Si000@0AdE6030^;`0`;Rl0:bTZ0`0_;Bh3044o@080=38c0`0]:b`0
0`0/:R/0A49304A2@`03034_<00402`Z:`0[:BX0:bTZ02`Z:`80EeEF00@0:2HW02`Z:`0/:R/0;B//
0P0V92D0100f=3D0@3ho040n?`0Z:2T202@R8`04030^;`0V92D09R@U04M5AP80?C/l00@0;bd^038`
<@0b<340<2h_0P0]:b`01014@T<0?S`m03hl?@0d<S<202TW:00403<a<P0d<S<0=38c03dk?080;bd^
00@0?cdn02HT9@0V92D08R0Q0P0_;Bh0100f=3D0>3Hg03Pf=`0o?Ch204m=CP0402/Y:P0a;c00<Bl`
02d[;080@3ho00@0=C<d030^;`0`;Rl0:2HW0P0o?Ch0101<BT/0<Bl`034_<00e<c@203Xh>@0402h/
;@0]:b`0;B//038`<@808AlP00@0:RPY02`Z:`0/:R/09R@U0P1CDE80100[:BX0;B//02d[;00X9RL2
02XX:@0402PV9`1=Bd`0CD]<030^;`80;bd^00@0:bTZ02PV9`0X9RL0>CLh0P0j>3T0100Z:2T0<S0a
038`<@15@d@202h/;@0403<a<P0f=3D0=S@e03@b<`80?cdn00@0;R`]01lM7P0O7Ah0;bd^0P18ATL0
100b<3408b4R02<Q8P0c<C8202`Z:`0403hl?@0^;2d0;R`]02LU9P80;B//00@0:RPY05IDE@1FE5D0
;2X[0P0d<S<0100_;Bh0>3Hg03Pf=`0Z:2T202/Y:P0403@b<`0c<C80<c4b03Tg>080>SPi00@0:BLX
03Pf=`0h=SL0=cDf0P0]:b`01011?d00>3Hg03Pf=`0Y9bP202/Y:P0402`Z:`0^;2d0;R`]04Q6A`80
>3Hg00@0928S02l];P0_;Bh09B<T0P0Z:2T0100T8R<0AdE604M5AP0i=cP204I4A@04030^;`0g=CH0
=cDf02d[;0809R@U0@0/:R/103Tg>040oooo000404QBG080AU9N0`15Df0304=CHP04045DH`14EfH0
A5MV049DHP80?dmL00@0AeEP04mIHP1?FF80@D]D0P0]=Sd00`0@41003@d=00d=3@0300X:2P0400`<
300>3Ph03Ph>00h>3P803@d=00@02`/;00`<300<30`0<30`0P0Q8B40100H61P0<S8b038b<P0Z:RX2
02TY:@0402XZ;00U9BL09BDW028R90808b<U00@0:BT[03De=`0e=CL0>SXl0P0c<cD0100C4aD0:b/]
02/[;@0j>S`202LW:@0402HV:00Z:R`0:RX/028R9080=CDg00@0>CTi02@T900T92@07QhN0P0Q8B40
100M7Ad0=CDe03De=@0[:b/201lO7`04024Q8@0P8200820P028R8P80=CDe00@09bLW01lO7`0O7al0
820P0P0O7al01013@d<0>3Ph03Ph>00b<S8202d];@0402HV9P0S8b<08b<S03hn?P809bLW00@092@T
02<S8`0S8b<06ATI0`0d=3@201hN7P0402TY:@0P8200820P02<S8`80>c/k00@0:BTY02@T900T92@0
9BDU0P0K6a/0100];Bd0>SXj03Xj>P0D51@201DE5@0402HV9P0R8R808R8R01hN7P80;2`/00@0:b/[
02DU9@0U9BD071`N0P0V9RP0100F5QP0@d=504=3A@0R8R@202@T9P0402h^<00Y:B/0:BT[02`/;P80
CTi@00@0;bla02@T9P0T92H0<S8d0P0[:bd0100];Bl0<c<e03<c=@0W9bT202PX:P0402DU9`0m?Cl0
?Cdo02/[;@80:RX/00@092@V03`l?P0l?3h07alQ0P0`<380100Q8B<0820R020P8P0b<S@201XJ7004
01lO8@0J6Q`06QXL01HF6080>c/m00<09RHX020P8P0P82800`0Q8B<0100X:2X0ATI804I6B00P8282
00d=3`0402HV:00S8bD08b<U01PH6P807QhP00@05QHH02<S9@0S8bD0>c/m0P0/;2h0100];Bl08b<U
02<S9@0L71h202HV:00402XZ;00_;c40;bla02PX:P80;Bd_00@0;bla020P8P0P828071`N0P0O7b40
100P8280<S8d038b=017AdD20861O`040:NWY@0Q8Al08B4O01<C4@80:b/Y00@0=CDc02XZ:00Z:RP0
<C4a0P0o?cl0100`<300:RX/02XZ;00V9BX202/Z;`0403Th?P0P7bD081lU01XJ7080:2PZ00@08R8T
01lO8@0O7b40<C4c0P0W9bT0100];Bl0<c<e03<c=@0/;2h202PX:P0401dM7`0S8bD08b<U02l_<@80
:RX/00@06QXL00T92`092@/02PX<0P0820X0100K6ad092@V02@T9P0>3Q0201dM7`0403Hf>01HF5X0
F5QJ03Ti>`80?Cdo00@0=SHh02PX:P0X:2X0:2PZ0`0W9bT202HV:004059@D@0Z:2T0:RPY030^;`80
9bDV00@0;2X[02l];P0_;Bh0@d520P0W9BH0100]:b`0:BLX02TW:00Z:2T203/i>P0403Le=P0[:BX0
:bTZ030^;`80:bTZ00<06QPI051>C`1@CTl00`0U8b@0100Z:2T0;2X[02`Z:`0^;2d204]9BP0403@b
<`0a;c00<Bl`03@b<`80<2h_00@0A49304a:B`1<BT/0<S0a0P0l>S/0100d<S<0<S0a038`<@10?Sl2
04i<C@@0=C<d0P0c<C801010?Sl0;bd^02l];P12@44202/Y:P0402XX:@0i=cP0>CLh03<a<P80>CLh
00@0<Bl`059@D@1BD5409bDV0P0X9RL0100T8R<0?3Xk03`j>`0a;c0202LU9P0402PV9`0]:b`0;B//
04]9BP80<S0a00@0:BLX030^;`0`;Rl0;B//0P0T8R<01018ATL0;2X[02`Z:`0a;c0203Le=P0402l]
;P1NG5d0GUaM038`<@80>CLh00@0:BLX02/Y:P0[:BX08R0Q0P1>C4d0100`;Rl0;2X[02`Z:`0/:R/2
02TW:00402l];P0e<c@0=C<d02<Q8P80:bTZ00@0?cdn02LU9P0W9BH0;R`]0P0`;Rl0100b<340?3Xk
03`j>`0a;c0201lM7P0402@R8`18ATL0B4I703Hd=@808AlP00@0<Bl`02XX:@0Z:2T0>cTj0P0c<C80
100]:b`0:RPY02XX:@0W9BH204m=CP<0<S0a00@0:2HW03<a<P0c<C80:BLX0P0^;2d0100g=CH0<S0a
038`<@0b<34203@b<`@0;bd^0P0[:BX0100d<S<0@3ho040n?`0[:BX202d[;004038`<@0W9BH09bDV
044o@080ATA500@0>CLh038`<@0b<340:2HW0P0/:R/01011?d00>3Hg03Pf=`11?d02030^;`0403`j
>`0m>c`0?C/l030^;`80;B//0@0g=CH104Q6A`40oooo000404QBG080AU9N0`15Df0304=CHP04045D
H`14EfH0A5MV049DHP80?dmL00@0AeEP04mIHP1?FF80@D]D0P0]=Sd00`0@41003@d=00d=3@0300X:
2P0400`<300>3Ph03Ph>00h>3P803@d=00@02`/;00`<300<30`0<30`0P0Q8B40100H61P0<S8b038b
<P0Z:RX202TY:@0402XZ;00U9BL09BDW028R90808b<U00@0:BT[03De=`0e=CL0>SXl0P0c<cD0100C
4aD0:b/]02/[;@0j>S`202LW:@0402HV:00Z:R`0:RX/028R9080=CDg00@0>CTi02@T900T92@07QhN
0P0Q8B40100M7Ad0=CDe03De=@0[:b/201lO7`04024Q8@0P8200820P028R8P80=CDe00@09bLW01lO
7`0O7al0820P0P0O7al01013@d<0>3Ph03Ph>00b<S8202d];@0402HV9P0S8b<08b<S03hn?P809bLW
00@092@T02<S8`0S8b<06ATI0`0d=3@201hN7P0402TY:@0P8200820P02<S8`80>c/k00@0:BTY02@T
900T92@09BDU0P0K6a/0100];Bd0>SXj03Xj>P0D51@201DE5@0402HV9P0R8R808R8R01hN7P80;2`/
00@0:b/[02DU9@0U9BD071`N0P0V9RP0100F5QP0@d=504=3A@0R8R@202@T9P0402h^<00Y:B/0:BT[
02`/;P80CTi@00@0;bla02@T9P0T92H0<S8d0P0[:bd0100];Bl0<c<e03<c=@0W9bT202PX:P0402DU
9`0m?Cl0?Cdo02/[;@80:RX/00@092@V03`l?P0l?3h07alQ0P0`<380100Q8B<0820R020P8P0b<S@2
01XJ700401lO8@0J6Q`06QXL01HF6080>c/m00<09RHX020P8P0P82800`0Q8B<0100X:2X0ATI804I6
B00P828200d=3`0402HV:00S8bD08b<U01PH6P807QhP00@05QHH02<S9@0S8bD0>c/m0P0/;2h0100]
;Bl08b<U02<S9@0L71h202HV:00402XZ;00_;c40;bla02PX:P80;Bd_00@0;bla020P8P0P828071`N
0P0O7b40100P8280<S8d038b=017AdD20861O`040:NWY@0Q8Al08B4O01<C4@80:b/Y00@0=CDc02XZ
:00Z:RP0<C4a0P0o?cl0100`<300:RX/02XZ;00V9BX202/Z;`0403Th?P0P7bD081lU01XJ7080:2PZ
00@08R8T01lO8@0O7b40<C4c0P0W9bT0100];Bl0<c<e03<c=@0/;2h202PX:P0401dM7`0S8bD08b<U
02l_<@80:RX/00@06QXL00T92`092@/02PX<0P0820X0100K6ad092@V02@T9P0>3Q0201dM7`0403Hf
>01HF5X0F5QJ03Ti>`80?Cdo00@0=SHh02PX:P0X:2X0:2PZ0`0W9bT202HV:004059@D@0Z:2T0:RPY
030^;`809bDV00@0;2X[02l];P0_;Bh0@d520P0W9BH0100]:b`0:BLX02TW:00Z:2T203/i>P0403Le
=P0[:BX0:bTZ030^;`80:bTZ00<06QPI051>C`1@CTl00`0U8b@0100Z:2T0;2X[02`Z:`0^;2d204]9
BP0403@b<`0a;c00<Bl`03@b<`80<2h_00@0A49304a:B`1<BT/0<S0a0P0l>S/0100d<S<0<S0a038`
<@10?Sl204i<C@@0=C<d0P0c<C801010?Sl0;bd^02l];P12@44202/Y:P0402XX:@0i=cP0>CLh03<a
<P80>CLh00@0<Bl`059@D@1BD5409bDV0P0X9RL0100T8R<0?3Xk03`j>`0a;c0202LU9P0402PV9`0]
:b`0;B//04]9BP80<S0a00@0:BLX030^;`0`;Rl0;B//0P0T8R<01018ATL0;2X[02`Z:`0a;c0203Le
=P0402l];P1NG5d0GUaM038`<@80>CLh00@0:BLX02/Y:P0[:BX08R0Q0P1>C4d0100`;Rl0;2X[02`Z
:`0/:R/202TW:00402l];P0e<c@0=C<d02<Q8P80:bTZ00@0?cdn02LU9P0W9BH0;R`]0P0`;Rl0100b
<340?3Xk03`j>`0a;c0201lM7P0402@R8`18ATL0B4I703Hd=@808AlP00@0<Bl`02XX:@0Z:2T0>cTj
0P0c<C80100]:b`0:RPY02XX:@0W9BH204m=CP<0<S0a00@0:2HW03<a<P0c<C80:BLX0P0^;2d0100g
=CH0<S0a038`<@0b<34203@b<`@0;bd^0P0[:BX0100d<S<0@3ho040n?`0[:BX202d[;004038`<@0W
9BH09bDV044o@080ATA500@0>CLh038`<@0b<340:2HW0P0/:R/01011?d00>3Hg03Pf=`11?d02030^
;`0403`j>`0m>c`0?C/l030^;`80;B//0@0g=CH104Q6A`40oooo000304Y@G00404UAG019DEh0BE5N
04QBGP<0AU9P0P16DV801010EFH0@eIU04=FI@11Df4203m?G00404IDG`1<FF80C5UR04A>E`80=Sm6
00@0410@00d=3@0=3@d02`/;0P0:2PX0100<30`03Ph>00h>3P0>3Ph200d=3@0400H61P051@D01@D5
034a<@80820P00@03Ph>02`/;00/;2`0;Bd]0P0a<C400`0`<3809RHX02HV:00302HV:00402<S9@0/
;2h0;2`^03Lg>@80=CDg00@05QHH02DU9`0U9BL0>3Pj0P0N7R00100M7Al0:2PZ02PX:P0I6A/203@d
=P0403lo?`0U9BD09BDU020P80809BDU00@07AdM03`l?00l?3`0=SHf0P0U9BD0100T92@0:RXZ02XZ
:P0T92@204E5A@0402`/;00V9RH09RHV02HV9P8092@T00@0=3@d03Xj>P0j>SX0:RXZ0P0R8R80100S
8b<05aLG01LG5`0f=SH202HV9P04020P800X:2P0:2PX018B4P80;Rh^00@0<C4a01HF5P0F5QH09bLW
0P0H61P0100L71`0>SXj03Xj>P0R8R82020P800402LW9`0C4a<04a<C01hN7P80<c<c00@08b<S00d=
3@0=3@d0:BTY0P0b<S80100L71`0:2PX02PX:00f=SH202h^;P04024Q8`0];Bl0;Bd_01HF6080Bd]=
00@09BDW02LW:@0W9bT0<S8d0P0[:bd0100/;2h0CdmA04m?D@0Y:B/2020P8P04038b=00U9BL09BDW
02HV:080>CTk00@0820R024Q8`0Q8B<08B4S0P0c<cD0100Z:R`09RHX02HV:00S8bD203Hf>00402LW
:@0h>3X0>3Pj02LW:@809RHX00@0?Cdo02HV:00V9RP0:BT[0P0T92H0100R8R@0@T940492A00b<S@2
02<S9@0402PX:P0P8280820R02XZ;080@d=500@0820R00l?4@0?3a40?Cdo0P0U9BL0100<30h05aLI
01LG6@0=3@l20451@`0403`l?P0b<S@0<S8d03Ti>`80<S8d00@09bLY02TY:`0Y:B/08b<U0P0P8280
100T92H0820R020P8P0T92H202<S9@0402XZ;00/;2h0;2`^02TY:`80?Cdo00@0UIFC0;no_@2o_kd0
W9bJ0P0<30X0100O7ad0;Bd[02d]:`0k>cT2038b<00403Ti>@12@T80@T9203dm?@80;bla00@09bH[
024P9@0Q82D0?Se30P0V9B/0100J6Q`0;Bd_02d];`0];Bl201TI6`0402`/;P0K6ad06a/M02d];`80
<30b00@0B4Q:00d=3`0=3@l0820R0P15ADL0100j>S`06QXL01XJ700L71h205EEE`0402/[;@0I6A/0
6ATK024Q8`80;2`^00@07QhP03<c=@0c<cD0;bla0P0X:2X0101<C4h0?Cdo03dm?`0j>S`202l_<@04
02LW:@0Q8B<08B4S02DU9`8071`N00@0De5B024O800Q7b00<S0a0P0T8R<0100a;c00;B//02d[;01@
CTl202/Y:P04038`<@0Z:2T0:RPY02h/;@80>3Hg00@0?cdn02/Y:P0[:BX0=cDf0P0`;Rl0100V92D0
H5iO061NG`0h=SL2038`<@0403Tg>00o?Ch0?cdn03Le=P80GUaM00@0?C/l03/i>P0k>CX0=S@e0P0e
<c@0100m>c`0D4i?051>C`0]:b`203Tg>00402l];P0^;2d0;R`]03Hd=@80DDm@00@0;R`]03Dc=00e
<c@0;R`]0P0Z:2T0100n?3d0:bTZ02/Y:P13@D82030^;`04038`<@13@D80@d5203Dc=080?S`m00@0
=38c05eKG01MFe`0;R`]0P0Z:2T0100T8R<0?S`m03hl?@0b<34202@R8`0403@b<`0[:BX0:bTZ05AB
D`80?cdn00@0:RPY034_<00a;c00=cDf0P0^;2d0101?CDh0:RPY02XX:@0]:b`203Hd=@0402d[;01L
FU/0G5YK02l];P80=cDf00@0<S0a03@b<`0d<S<09bDV0P1GEEH0100h=SL0=38c03@b<`0e<c@303<a
<P80>cTj00@09R@U02`Z:`0/:R/0@3ho0P0V92D0100/:R/0;R`]02h/;@0a;c0203Tg>00403@b<`0S
8B808b4R02<Q8P80Cde>00@0@T1102TW:00Y9bP0<S0a0P0^;2d0100m>c`0?3Xk03`j>`0f=3D202d[
;00402TW:01>C4d0CTa=040n?`80=cDf00@09bDV038`<@0b<340:bTZ0P0c<C80100j>3T0;bd^02l]
;P0[:BX202`Z:`0403Hd=@0V92D09R@U02HT9@80;R`]00@0:BLX04E3A015@d@0<2h_0P0^;2d0100e
<c@071XK01`J6`0`;Rl204a:B`0404e;C00e<c@0=C<d030^;`80=cDf00@0<c4b02l];P0_;Bh09bDV
0P0b<340100e<c@0>3Hg03Pf=`0e<c@203`j>`40@d520@14@T<10?ooo`00301:DeX01016E5l0BUQS
04YHH`13DE`304MEH080@DmJ00@0AeEP049@F`12D5/072<[0P0941H0100<4Q8020d700P=1`072P<2
010A2`0400h=2@0@30/040`;00X91@8040l=00@04A0>01XH6@0J61T0:BLX0P0]:bh0100Y9bX0;2X_
02`Z;`0/;2`202LW9`0402HV9P0Y:BT0:BTY02XZ:P80:2PX00@0:RXZ02h^;P0^;Rh07alO0P0/;2`0
100`<300=3@d03@d=00J6QX202<S8`0402TY:@0R8R808R8R04=3A@808B4S00@07AdO02`/;P0/;2h0
9RHX0P0P8280101DE5H0=3@f03@d=P0Y:B/203Lg>@0401XJ700i>C/0>CTk03`l?P809BDW00@07alQ
02LW:@0W9bT08R8T0P0k>cd0100/;2h04a<E01<C5@0T92H302LW:@80>c/m00@08R8T02DU9`0U9BL0
92@V0P0P8280100];Bl0<S8d038b=00R8R@2020P8P<092@T00<0>c/k03Ph>00h>3P00`0U9BD0100/
;2`0>c/k03/k>`0V9RH201XJ6P@09bLW0P0c<c<01018B4P092@T02@T900/;2h202/[;@0402d];`0/
;2h0;2`^0451@`809BDW00@0;Rh`02XZ;00Z:R`092@V0P0i>C/00`0Q8B<08b<U02<S9@0301hN8004
01@D5P10@480@41201PH6P807QhP00@05aLI024Q8`0Q8B<0@4120P0K6ad302DU9`04034a<`0^;S00
;Rh`03Ti>`807QhP00<0ADE702<S9@0S8bD00`0[:bd0100_;c40BDU;04U9B`0h>3X202TY:`0402LW
:@0Z:R`0:RX/01XJ7080BDU;00@071`N01XJ700J6Q`0@D530P0^;S00100=3@l01PH800H6200U9BL2
04=3A@0402HW:@0Z:bd0:R/]024R90806Q/M00<07b0R024R900Q8R@00`0M7R00100X:RT08B<R024S
8P0J71/203Hh=`04034c<P0/;Rd0;2h]04]=C080V9ZI00@0YJNT0;Be/02d]K00J6EP0P0S71H0100M
5A80;b/Z02l[:P0Q8B<202l`=00402P/;@1=CU00CDi@038c=@80<S8d00@09RHX03<a=00c<C@0A495
0P0[9RX0100Q8B<0:BT[02TY:`0W9bT201TI6`0401hN800X:2X0:2PZ030`<P80;Bd_00@0EUIH01DE
5`0E5AL02`/=0P0n?T001014A4H03Ph@00h>400V9RP20492A00404A4AP0I6A/06ATK01`L7P805ADG
00@07AdO020P8P0P82807AdO0P0Y:B/0100j>3/0>cTl03/i?00c<C@202PV:@0402/Y;00e<cH0=C<f
03<a=0809bDX00@0EeEF038`<@0b<340>CLh0P0b<340100n?3d09bDV02LU9P1QGf0202DS900403/i
>P0h=SL0>3Hg02d[;080=cDf00@0D4i?03<a<P0c<C80<Bl`0P0j>3T0100c<C80>CLh03Tg>01=Bd`2
02/Y:P0403Le=P0Z:2T0:RPY02d[;080D4i?00@0;bd^02d[;00]:b`0:RPY0P0^;2d0100Z:2T0F5IG
05QFE`0W9BH202h/;@0402XX:@0Y9bP0:BLX02LU9P80CD]<00@08R0Q038`<@0b<340<Bl`0P0P7Ql0
100j>3T08b4R02<Q8P18ATL202PV9`0403Dc=00k>CX0>cTj02d[;080?S`m00@0>cTj065OH01QGf00
<S0a0P0l>S/0100Q7b00=cDf03Le=P0j>3T202XX:@0402/Y:P0V92D09R@U04E3A080AdE600@09bDV
02/Y:P0[:BX0;2X[0P0Q7b000`15@d@0:RPY02XX:@0302XX:@0402h/;@1MFe`0GE]L030^;`80?cdn
00@0<S0a02d[;00]:b`08AlP0P1?CDh0100^;2d0;2X[02`Z:`0S8B8302XX:@80?C/l00@0;2X[02h/
;@0^;2d0AdE60P0Y9bP00`0/:R/0;bd^02l];P03030^;`0403hl?@0Q7b008AlP020N7`80>SPi00@0
>cTj03@b<`0d<S<0928S0P0W9BH0100S8B80>cTj03/i>P0e<c@302HT9@80<S0a00@0BDM802d[;00]
:b`0?S`m0P0c<C80100[:BX0:RPY02XX:@15@d@202/Y:P0302LU9P0]:b`0;B//00<0<Bl`00@0<2h_
02h/;@0^;2d0<2h_0P0d<S<01017ADH0=C<d03Dc=00X9RL203Pf=`04034_<00e<c@0=C<d051>C`80
>3Hg00@0;R`]038`<@0b<3408AlP0P0X9RL0100h=SL0<Bl`034_<00^;2d2030^;`0504E3A01=Bd`0
CD]<03Tg>00Y9bP00@3oool000`0BU=J00@0AUAO04YHH`1:F6<0@e5L0`17EF02045?FP0404MEH012
D5/0@U1K01`S:`802A0F00@0318B00P=1`083@L01`X30P0@4@/0100>3@T040`;010<2`0:2@D2010?
3@04014@3P0J61T06QPI02TW:080;B/^00@0:BLZ02`Z;`0/:Rl0;2`/0P0W9bL0100V9RH0:BTY02TY
:@0Z:RX202PX:00402XZ:P0^;Rh0;Rh^01lO7`80;2`/00@0<30`03@d=00d=3@06QXJ0P0S8b<0100Y
:BT08R8R028R8P13@dD2024Q8`0401dM7`0/;2h0;2`^02HV:080820R00@0E5AF03@d=P0d=3H0:BT[
0P0g=cT0100J6Q`0>CTk03Ti>`0l?3h202DU9`0401lO8@0W9bT09bLY028R9080>c/m00@0;2`^01<C
5@0C4aD092@V0`0W9bT203/k?@04028R900U9BL09BDW02@T9P80820R00@0;Bd_038b=00b<S@08R8T
0P0P828302@T900303/k>`0h>3P0>3Ph00<09BDU00@0;2`/03/k>`0k>c/09RHV0P0J6QX402LW9`80
<c<c00@0B4Q802@T900T92@0;2`^0P0[:bd0100];Bl0;2`^02`/;P11@D<202DU9`0402h^<00Z:R`0
:RX/02@T9P80>CTk00<08B4S02<S9@0S8bD00`0N7R00100D51H0@4120410@P0H61X201hN800401LG
6@0Q8B<08B4S0410@P806a/M0`0U9BL0100a<C<0;Rh`02h^<00i>C/201hN800304E5A`0S8bD08b<U
00<0:b/]00@0;bla04U9B`19BD/0>3Pj0P0Y:B/0100W9bT0:RX/02XZ;00J6Q`204U9B`0401`L7P0J
6Q`06QXL0451@`80;Rh`00@03@d?00H620061PP09BDW0P13@dD0100V9bT0:R/]02X[;@0Q8R@201XK
7@0301lP8P0Q8R@08B8T00<07AhP00@0:2XY024S8P0Q8b806Q`K0P0f>3L0100a<c80;2h]02`^;@1;
CD`209RJV@040:FWY02d]K00];F`06QUH0808a`F00@07ADB02l[:P0_:bX08B4S0P0_<3@0100X;2d0
CDi@04e>D00b<cD2038b=00402HV:00c<C@0<c4d04A2A@80:bHZ00@08B4S02TY:`0Y:B/09bLY0P0I
6A/0100N7R00:2PZ02PX:P0`<38202d];`0405IFF00E5AL05ADG00/;3@80?Si000@0A4A600h>400>
3Q009RHX0P12@T@01014A4H06ATK01TI6`0L71h201DE5`0401dM7`0P8280820R01dM7`80:BT[00@0
>SPk03/i?00k>C`0<c4d0P0X9RT0100[:B`0=C<f03Dc=P0c<C@202LU:00405MEEP0b<340<S0a03Tg
>080<S0a00@0?S`m02LU9P0W9BH0HEmP0P0U8b@0100k>CX0>3Hg03Pf=`0]:b`203Le=P04051>C`0c
<C80<c4b034_<080>SPi00@0<c4b03Tg>00i=cP0CD]<0P0[:BX0100g=CH0:RPY02XX:@0]:b`2051>
C`0402l];P0]:b`0;B//02XX:@80;R`]00@0:RPY05QFE`1HEUL09bDV0P0^;2d0100Z:2T0:BLX02TW
:00W9BH204e;C004028P8@0b<340<S0a034_<08081hO00@0>SPi02<Q8P0S8B80B4I70P0X9RL0100e
<c@0>cTj03/i>P0]:b`203hl?@0403/i>P1QGf00HEmP038`<@80?3Xk00@08AlP03Le=P0g=CH0>SPi
0P0Z:2T0100[:BX09R@U02HT9@15@d@204M5AP0402LU9P0[:BX0:bTZ02`Z:`808AlP00<0AD=402XX
:@0Z:2T00`0Z:2T0100^;2d0GE]L05eKG00`;Rl203lm?P04038`<@0]:b`0;B//024O8080Cde>00@0
;R`]02`Z:`0/:R/08b4R0`0Z:2T203dk?00402`Z:`0^;2d0;R`]04M5AP80:BLX00<0;2X[02l];P0_
;Bh00`0`;Rl0100n?3d08AlP024O800P7Ql203Xh>@0403/i>P0d<S<0=38c02@R8`809bDV00@08b4R
03/i>P0k>CX0=C<d0`0V92D2038`<@0404U7B00]:b`0;B//03hl?@80<c4b00@0:bTZ02XX:@0Z:2T0
AD=40P0[:BX00`0W9BH0;B//02d[;003034_<004030^;`0^;2d0;R`]030^;`80=38c00@0AdE603Dc
=00e<c@0:2HW0P0h=SL0100a;c00=C<d03Dc=01@CTl203Pf=`0402h/;@0b<340<S0a024O8080:2HW
00@0>3Hg034_<00a;c00;R`]0P0`;Rl01@15@d@0CD]<04e;C00i=cP0:BLX0040oooo000<04UCG004
04IDG`1:F6<0BUQS04=AG0<0AeEP0P12D5/01018EV40@e5L04=AG00Q:30200P?5@0400T?3`083@L0
20d700T<1@80414;00@03@`8010<2`0@30/02PT50P0@3`d0100A40h06QPI01XH6@0Z:2T2030^<@04
02d[;P0_;C80;bdb02/[:`809BDU00@08b<S02PX:00X:2P0;2`/0P0V9RH0100O7al06a/K01/K6`0A
4A42028R8P0402@T900b<S80<S8b01dM7@8071`L00@09bLW024Q8@0Q8B40@d=50P0j>S`0100W9bT0
;bla02l_<@0X:2X2028R900404M7B@0/;2h0;2`^01lO8@809BDW00@07AdO02DU9`0U9BL0@d=50P0Q
8B<0100S8bD0:BT[02TY:`0P828203Hf>00402d];`0I6A/06ATK024Q8`808b<U00@092@V03<c=@0c
<cD0:2PZ0P0M7Al0100W9bT0820R020P8P0[:bd203`l?P0402XZ;00Y:B/0:BT[020P80807alO00@0
<S8b030`<00`<3007alO0P0N7Qh0100U9BD0<S8b038b<P0R8R8201TI6@0401PH600O7al07alO020P
80808b<S00@0?clo01XJ6P0J6QX09RHX0P0R8R@0100U9BL09bLY02LW:@0j>S`201hN80<08b<U00@0
9bLY03/k?@0k>cd0;Bd_0`0U9BL2024Q8`0401`L7P0l?3h0?3`n01lO8@80820R00@061PJ028R900R
8R@0?cm10P0L71h0100T92H09bLY02LW:@0V9RP2024Q8`0402d];`0H61X061PJ03`l?P806ATK0`0M
7Al0100R8R@0=3@f03@d=P0X:2X201hN800402<S9@0/;2h0;2`^020P8P80ATI800@051@F010@4P0@
4180:2PZ0P0N7R00100>3Q004A4C014A4`0e=CL204Q8BP0404I7B@0i>S`0>CXl02X[;@80;Bh`00@0
=SLi02l`<P0_<3809RLY0P0Z:bd0100d=SD0:R`[02X/:`0o@D0201dO7P0402LY:00j?3/0>S`k09>E
U080]KNf00@0Wj6N0:BUX02TYJ00E55<0P0X8A/0100Z8Ql0;b/Z02l[:P0W9bT202d^<P0401/O8016
AdT0ATM902X[;@80:BT[00@09BDW034_<P0a;c80A4950P0e<3@0100W9bT0;Bd_02d];`0`<38202HV
:004028R900Y:B/0:BT[02l_<@80:BT[00@0DU9D00X:300:2P`09BDW0P0Y:B/0100l?3h0BTY<04Y:
C00P828200d=3`0400`<3P0D51H051@F01TI6`803@d?00<03`lA01HF600F5QP00`0];Bl01013@D@0
ATA704I4A`14@TD203Xh>`0403<a=00j>3/0>SPk03dk?P80=C<f00@0E59C040n?`10?Sl0=S@e0P0/
:R/0100d<S<09bDV02LU9P1FE5D202`Z:`04030^;`0]:b`0;B//02HT9@80;bd^00@0B4I7034_<00a
;c00;B//0P0d<S<0100]:b`0=C<d03Dc=01:B4T202LU9P04034_<00X9RL0:2HW02TW:080AD=400@0
<c4b02DS900U8b@0:RPY0P0/:R/0100[:BX0De5B05=ADP0e<c@2030^;`0402l];P0[:BX0:bTZ02`Z
:`80BTQ900@0;R`]034_<00a;c00<Bl`0P0T8R<0100k>CX09B<T02DS9015@d@202@R8`0402d[;00a
;c00<Bl`02@R8`80<S0a00@0=S@e05UGF01IEeP0=cDf0P0e<c@0100S8B80<c4b03<a<P0l>S/202HT
9@0402h/;@0Z:2T0:RPY04=1@P80Cde>00@0;B//03<a<P0c<C80;bd^0P0Z:2T01017ADH0>3Hg03Pf
=`0Z:2T302/Y:P80CD]<00@0<Bl`030^;`0`;Rl0;B//0P0Y9bP0100S8B80AdE604M5AP0g=CH202l]
;P0402TW:00a;c00<Bl`02XX:@80>cTj00@0:2HW02TW:00Y9bP0@Cm00P0W9BH0100Z:2T0;R`]02h/
;@0c<C8203Pf=`04044o@00]:b`0;B//028P8@80;bd^00@0A493034_<00a;c00;B//0P0_;Bh0100Z
:2T0@Cm0044o@00i=cP202`Z:`0402XX:@0b<340<S0a04U7B080:BLX00@0=38c02h/;@0^;2d0<2h_
0P0_;Bh00`16A4D0<Bl`034_<00302d[;00402/Y:P0/:R/0;2X[02h/;@80<Bl`00@0<S0a03<a<P0c
<C80B4I70P17ADH0100c<C80=cDf03Le=P0e<c@202d[;00403/i>P0f=3D0=S@e034_<080;bd^00@0
:2HW02h/;@0^;2d0:BLX0P0U8b@0100h=SL0@3ho040n?`11?d0204M5AP40=S@e0@0W9BH10?ooo`00
3019De`01016E5l0BUQS04YHH`13DE`204QFH@0404MEH012D5/0@U1K04UGHP80A59M00@0:S4i00P?
5@083aD01@/;0P0:3`T0100=40T03a0:00l@2P0:2@D2014=300400X91@0@3`d040l=014@3P8061HG
00@0:RPY038`<`0b<3<0<2ha0P0b<3D0100Z:RX08b<S02<S8`0R8R8202/[:`04034a<@0Z:RX0:RXZ
01dM7@805ADE00@06ATI02XZ:P0Z:RX07QhN0P0a<C40100T92@06QXJ01XJ6P0c<c<203De=@0403<c
=@1:BT`0BTY<02DU9`80:2PZ00@08b<U01`L7P0L71h0<30b0P0R8R@0100Q8B<071`N01`L7P0U9BL2
01DE5`0404=3A@0M7Al07AdO01dM7`808R8T00@08B4S038b=00b<S@0<S8d0P0X:2X0100R8R@08b<U
02<S9@0T92H202`/;P0403dm?`0S8bD08b<U02h^<0809BDW00@092@V03Ph>P0h>3X0:BT[0P0N7R00
100S8b<08R8R028R8P0_;bl202h^;P04024Q8@0N7Qh07QhN02@T9080<30`00@0:RXZ01hN7P0N7Qh0
4a<C0P0O7al0100P82007QhN01hN7P0i>CT201hN7P0402HV:00O7b407alQ02<S9@80:RX/00@0?3`n
020P8P0P82808B4S0P0T92H0100P8280<30b030`<P0c<cD201lO8@0402HV:00O7b407alQ024Q8`80
<S8d00@0:BT[028R900R8R@06QXL0P0T92H0100m?Cl07alQ01lO8@0S8bD202XZ;00402TY:`0P8280
820R02`/;P807QhP00@0?3`n01/K7@0K6ad06ATK0P0L71h0100V9RP0<C4c034a<`0a<C<2024Q8`04
020P8P0X:2X0:2PZ028R9080?cm100@0410B01LG6@0G5aT07QhP0P0I6A/0100<30h0410B010@4P0h
>3X203hn@00402l`<P0e=SP0=CHh03Lh>P80>S/m00@0=SLi02<T9P0S92H061TK0P0P8B<0100W:bX0
<CDd034e=00J7Qd202HZ:@0403<g=P2JWYd0VYjM0;Jj^@80[k>b00@0ZZbY09VJU@2IVYD0?S/f0P0Y
8Q`0100a:BH0:2@S02PT8`0[:bd202DV:P04020T9@1?D580Ce1B02h_<@809bLY00@0:BT[02d[;P0]
:bh0?ce00P0l=c/0100L71h08b<U02<S9@0`<382038b=00402<S9@0V9RP09RHX02LW:@806a/M00@0
820R00@41P0410H0=CDg0P0R8R@0100`<380H61R061PHP0g=cT200d=3`040082100B4Q@04Q8D00d=
3`8051@F00@071`N01@D5P0D51H0=3@f0P0e=CL0100_;C00>CLj03Tg>P15@dH203/i?00302@R9@0S
8B@08b4T00<0;bd`00@0?C/l04M5AP17ADH0;B//0P0U8b@302PV9`0404=1@P0`;Rl0<2h_02l];P80
;B//00@0:BLX030^;`0`;Rl0AD=40P0f=3D0100/:R/0<2h_030^;`0Z:2T203@b<`0404]9BP0Z:2T0
:RPY02h/;@80:bTZ00@0:RPY03`j>`0l>S/0C4Y;0P0Z:2T0100f=3D0<Bl`034_<00`;Rl204U7B004
04=1@P0[:BX0:bTZ030^;`80:RPY00@0;B//03lm?P0o?Ch0?3Xk0P0Y9bP0100/:R/09B<T02DS900i
=cP202PV9`0404=1@P0W9BH09bDV02l];P80=S@e00@0;2X[03Pf=`0h=SL0=cDf0P1DDU<01017ADH0
=38c03@b<`0^;2d203Hd=@0404U7B00Y9bP0:BLX02/Y:P80:BLX00@0>SPi05ECE01EDe@0=S@e0P0o
?Ch0100g=CH0>SPi03Xh>@0n?3d204A2@`0402XX:@0b<340<S0a03@b<`80BTQ900@0C4Y;03Le=P0g
=CH0=C<d0`0b<34204=1@P0404]9BP0g=CH0=cDf03<a<P80>cTj00@0;R`]03/i>P0k>CX09bDV0P0X
9RL0100m>c`0:bTZ02/Y:P0]:b`2030^;`0402/Y:P0h=SL0>3Hg044o@080=C<d00@09bDV02XX:@0Z
:2T0Cde>0P0m>c`0100c<C80<S0a038`<@0_;Bh20490@@0403Tg>00`;Rl0<2h_02/Y:P80;R`]00@0
EE=D03@b<`0d<S<0=C<d0P0c<C80100m>c`0>SPi03Xh>@19AdP203lm?P04034_<00]:b`0;B//02`Z
:`80<Bl`00@0=C<d03Pf=`0h=SL0=S@e0P0a;c00100o?Ch0Cde>04m=CP0f=3D202TW:00402h/;@0X
9RL0:2HW03Dc=080BDM800@0=cDf02/Y:P0[:BX0:RPY0P0e<c@0100X9RL09bDV02LU9P12@44204E3
A00502PV9`0b<340<S0a02d[;00W9BH00@3oool000`0BE=L00@0AUAO04YHH`1:F6<0@e5L0P18EV40
1017EF00@U1K049@F`19Ef8204ABG@0402Xa>@083aD020lE00D;2`802Pl900@03A0900l@2P0?40X0
2PT50P0A3@`0100:2@D040l=010?3@0A40h201PF5`0402XX:@0b<3<0<S0c030^<@80<S0e00@0:RXZ
02<S8`0S8b<08R8R0P0[:b/0100a<C40:RXZ02XZ:P0M7Ad201DE5@0401TI6@0Z:RX0:RXZ01hN7P80
<C4a00@092@T01XJ6P0J6QX0<c<c0P0e=CD0100c<cD0BTY<04Y:C00U9BL202PX:P0402<S9@0L71h0
71`N030`<P808R8T00@08B4S01`L7P0L71h09BDW0P0E5AL01013@dD07AdO01dM7`0M7Al2028R9004
024Q8`0b<S@0<S8d038b=080:2PZ00@08R8T02<S9@0S8bD092@V0P0/;2h0100m?Cl08b<U02<S9@0^
;S0202DU9`0402@T9P0h>3X0>3Pj02TY:`807QhP00@08b<S028R8P0R8R80;bl_0P0^;Rh0100Q8B40
7QhN01hN7P0T92@2030`<00402XZ:P0N7Qh07QhN01<C4`807alO00@0820P01hN7P0N7Qh0>CTi0P0N
7Qh0100V9RP07alQ01lO8@0S8bD202XZ;00403`l?P0P8280820R024Q8`8092@V00@0820R030`<P0`
<380<c<e0P0O7b40100V9RP07alQ01lO8@0Q8B<2038b=00402TY:`0R8R@08R8T01XJ708092@V00@0
?Cdo01lO8@0O7b408b<U0P0Z:R`0100Y:B/0820R020P8P0/;2h201hN800403`l?P0K6ad06a/M01TI
6`8071`N00@09RHX034a<`0a<C<0<C4c0P0Q8B<0100P8280:2PZ02PX:P0R8R@203lo@@04010@4P0G
5aT05aLI01hN80806ATK00@030`>010@4P0@4180>3Pj0P0n?T00100_<380=CHh03Df>00g>3X203Xk
?@0403Hg>@0S92H08b@V01PI6`80824S00@09b/Z034e=00a=C@06QhM0P0V:RT0100c=cH0VYjM09ZN
W@2f^[T20:nc/P040:Z/Z@2IVYD0VIZE03hk=P80:B8L00@0<BTV02PT8`0X92<0:b/]0P0U9RX0100P
92D0Ce1B04m@DP0^;c4202LW:@0402TY:`0]:bh0;B/^03lm@080?3Lk00@071`N02<S9@0S8bD0<30b
0P0b<S@0100S8bD09RHX02HV:00W9bT201/K7@04020P8P0410H010@603De=`808R8T00@0<30b061P
HP1PH680=cLi0P0=3@l010020P@04Q8D018B500=3@l201@D5P0401`L7P0D51H051@F03@d=P80=CDg
00@0;bd`03Tg>P0i=cX0AD=60P0k>C`00`0T8RD08b4T02<Q900302l]<00403dk?017ADH0AdE602d[
;0809B<T0`0X9RL01013@D80<2h_030^;`0_;Bh202d[;00402TW:00`;Rl0<2h_04E3A080=S@e00@0
;2X[030^;`0`;Rl0:RPY0P0d<S<0101;BDX0:RPY02XX:@0^;2d202/Y:P0402XX:@0l>S/0?3Xk04a:
B`80:RPY00@0=S@e034_<00a;c00<2h_0P19AdP01013@D80:bTZ02/Y:P0`;Rl202XX:@0402d[;00o
?Ch0?cdn03`j>`80:BLX00@0;2X[02DS900U8b@0>CLh0P0X9RL01013@D809bDV02LU9P0_;Bh203Hd
=@0402`Z:`0h=SL0>3Hg03Le=P80E59C00@0AdE603@b<`0d<S<0;R`]0P0f=3D01019AdP0:BLX02TW
:00[:BX202TW:00403Xh>@1EDe@0EE=D03Hd=@80?cdn00@0=cDf03Xh>@0j>3T0?S`m0P14@T<0100Z
:2T0<S0a038`<@0d<S<204Y8B@0404a:B`0g=CH0=cDf03Dc=0<0<S0a0P13@D80101;BDX0=cDf03Le
=P0c<C8203/i>P0402h/;@0k>CX0>cTj02LU9P80:2HW00@0?C/l02/Y:P0[:BX0;B//0P0`;Rl0100[
:BX0>3Hg03Pf=`11?d0203Dc=00402LU9P0Z:2T0:RPY04m=CP80?C/l00@0<c4b038`<@0b<340;bd^
0P12@440100i=cP0<2h_030^;`0[:BX202h/;@0405ECE00d<S<0=38c03Dc=080<c4b00@0?C/l03Xh
>@0j>3T0BDM80P0o?Ch0100a;c00;B//02d[;00/:R/2034_<00403Dc=00h=SL0>3Hg03Hd=@80<Bl`
00@0?cdn04m=CP1?CDh0=S@e0P0Y9bP0100^;2d0:2HW02PV9`0e<c@204U7B00403Le=P0[:BX0:bTZ
02XX:@80=C<d00@0:2HW02LU9P0W9BH0@T110P15@d@01@0X9RL0<S0a038`<@0]:b`09bDV0040oooo
000<04MDG@0404ECGP1:F6<0BUQS04=AG0<0B5IQ0P13DE`01019Ef80AE=N04ECGP0d>d<200XA5`04
0088200;40X02a0:010C30803a0:00@02@P4018>3@0B3Pd02PT50P0@3`d0100?3P`0518C01@B4`0W
9BH2034_<P0402l]<00_;C80;bdb02TY:@8092@T00@09BDU02h^;P0^;Rh0=3@d0P0`<300100X:2P0
92@T02@T900A4A4202`/;00402DU9@14A4@0A4A403lo?`80820P00@0<C4a02`/;00/;2`06QXL0P14
A4H0100J6Q`0820R020P8P0S8bD201lO8@0402XZ;00c<cD0<c<e02HV:0807alQ00@09BDW01LG6@0G
5aT0>CTk0P0Y:B/0100H61X08b<U02<S9@0V9RP2030`<P0403Ti>`0j>S`0>SXl02LW:@809RHX00@0
:2PZ02TY:`0Y:B/0?Cdo0P0W9bT0100U9BL09RHX02HV:00P828202d];`0303<c=@0K6ad06a/M00<0
9BDU00@0:RXZ02/[:`0[:b/08b<S0P0N7Qh0100R8R80:b/[02/[:`0d=3@201hN7P0401XJ6P0N7Qh0
7QhN020P80809BDU00@0;bl_02`/;00/;2`09RHX0P0M7Al0100Q8B<0;Bd_02d];`0n?T0202HV:004
028R900Y:B/0:BT[01/K7@809RHX00@0>SXl01lO8@0O7b40:BT[0P0S8bD0100Z:R`0:b/]02/[;@0c
<cD202<S9@0401`L7P0U9BL09BDW03Xj?0808b<U00@08R8T02XZ;00Z:R`0=3@f0P0U9BL0100_;c40
9RHX02HV:00l?3h2028R900401`L7P0S8bD08b<U02TY:`80<S8d00@0@T9402PX:P0X:2X071`N0P0N
7R00100O7b40>CTk03Ti>`0>3Q0302HV:0808B4S00@02`/=00L72@071`T0=CDg0P0i>C/0100d=CL0
?Sm103ho@@0k?3h202l`<P0402PY:`0S92H08b@V02@U9`80;c0b00@05ATH02T];00Y;B`0<CDd0P0l
@3l01029SH`0/[Je0;:f]@2h_;/20:Z^[@040;fo_02BThh0TY>>030]:0809AhH00@0;RHS024M700Q
7A`0;Rh`0P0K7200100Q9BH0CDi@04e>D00Y:R`201hN800402@T9P0Q7b808AlR03@b=@80=c8f00@0
6a/M01`L7P0L71h0:b/]0P0c<cD0100O7b408b<U02<S9@0W9bT201dM7`0400P82P0U9BL09BDW02h^
<0<0;Bd_0P0i>C/01016ATP09BDW02DU9`0e=CL201TI6`0400X:300X:2X0:2PZ0451@`809bLY00@0
;Rh`03Ph>P0h>3X09B<V0P0c<C@01018ATT0@d5404=1A00V92L201hL7`0402XX:`0]:bh0;B/^02@R
8`80C4Y;00@0:bTZ02TW:00Y9bP0:BLX0P0b<340100g=CH0>CLh03Tg>00`;Rl302d[;080<2h_00@0
?cdn03/i>P0k>CX0:RPY0P0[:BX0100V92D0;bd^02l];P19AdP202/Y:P0402TW:00]:b`0;B//02PV
9`80<2h_00@0E59C02<Q8P0S8B80<S0a0P0[:BX0100X9RL0<c4b03<a<P15@d@2020N7`0403Dc=00^
;2d0;R`]034_<080>SPi00@0C4Y;02TW:00Y9bP0;2X[0P0[:BX0100f=3D0:bTZ02/Y:P10?Sl202TW
:004030^;`0h=SL0>3Hg034_<080>SPi00@0=38c04U7B019AdP0De5B0P0b<34303@b<`04051>C`0Y
9bP0:BLX03@b<`80<c4b00@0>3Hg05MEEP1GEEH0>CLh0P0l>S/0100/:R/0<Bl`034_<00_;Bh204U7
B00402HT9@0d<S<0=38c03Tg>080@T1100@0Hf5R03dk?00m>c`0?3Xk0P0j>3T0100m>c`0?3Xk03`j
>`1DDU<203Le=P04038`<@0g=CH0=cDf038`<@80?C/l00@0:bTZ02h/;@0^;2d0@3ho0P0e<c@3034_
<00402TW:00h=SL0>3Hg04U7B080;R`]00@0:BLX02XX:@0Z:2T0@d520P1;BDX0100b<340;B//02d[
;00/:R/203`j>`0403@b<`0a;c00<Bl`02XX:@80:2HW00@0F5IG044o@011?d00>CLh0P0f=3D01011
?d00>CLh03Tg>013@D8204U7B00402d[;00/:R/0;2X[03Dc=0D0?S`m00@0>CLh030^;`0`;Rl0;2X[
0P18ATL0100m>c`09bDV02LU9P0V92D202XX:@0403Le=P1<BT/0C4Y;044o@080;bd^00@09B<T03Le
=P0g=CH0?S`m0P0l>S/01012@440=S@e03Hd=@0K6AX202XX:@40=S@e0@0k>CX10?ooo`003017E5d0
1015Deh0BUQS04YHH`14DUd304QFH@80A59M00@0BeUT04IDG`16E5l0?4=;0P0=51X010032@T02a0:
00/@2P0@4``200h?2@0400P70`0C3`h04`l>00/:1P8040l=00@03@`:00l=3P0?3@h08R0Q0P0^;2l0
100/:Rd0:BL/02TW;00Z:RX202PX:00402TY:@0^;Rh0;Rh^02h^;P80;2`/00@0;Rh^03<c<`0c<c<0
410@0P0Z:RX0100O7al0>c/k03/k>`0o?cl201LG5`0402<S8`0O7al07alO020P8P80@d=500@092@V
02<S9@0S8bD0:BT[0P0P8280100T92H0>c/m03/k?@0X:2X202@T9P0401/K7@0T92H092@V02/[;@80
?cm100@0820R02l_<@0_;c409RHX0P0Z:R`00`0f=SP0@T940492A00302LW:@0402HV:00U9BL09BDW
034a<`80;Rh`00@06ATK02DU9`0U9BL07QhP0P0M7Al0100j>S`06a/M01/K7@0P8202024Q8@04020P
800S8b<08b<S028R8P<06a/K0P0P8200100f=SH061PH01PH600Q8B4201PH600401TI6@0/;2`0;2`/
024Q8@80>3Ph00@08B4S01PH6P0H61X071`N0P0X:2X0100h>3X0:2PZ02PX:P0P828202PX:P0402HV
:00Z:R`0:RX/04M7B@80:b/]00@0;Rh`02/[;@0[:bd0<C4c0P0W9bT0100h>3X092@V02@T9P0O7b42
02@T9P0403@d=P0Y:B/0:BT[024Q8`809bLY00@0=CDg02<S9@0S8bD0:RX/0P0S8bD0100a<C<092@V
02@T9P0J6Q`2028R900401lO8@0U9BL09BDW0492A0808R8T00@07QhP01dM7`0M7Al07alQ0P0b<S@0
100A4A<0;Rh`02h^<00U9BL2024Q8`0400X:300820X020P:03Ti>`80@d=500@0@T=503Pi>`0h>C/0
:b`^0P0_<380100k?3h0=cPj03Lh>P0_<38203<d=P0402dc<@0_=C<0;cDc04A:B080M7Yh00@0_L?1
0;Fk^@2e^kT0Yjf[0P2`][@0102l_[/0NW]f07YkMP0Z9b82028K5@0402`T8@0W8b809b<R038b=080
6AXN00@0;30a04a=C`1<CDl0;2d_0P0U9BL0100];Bl0:2HY02PV:@12@4<2040k?`0402`/;P0V9RP0
9RHX02XZ;080=CDg00@07AdO02DU9`0U9BL0:b/]0P0Y:B/0100C4aD0@4120410@P0U9BL202`/;P04
030`<P0V9RP09RHX0410@P80<c<e00@0B4Q:02@T9P0T92H09RHX0P0b<S@0101;Bdd0@d=504=3A@0_
;c4202d];`0402`Z;@0`;S40<2ha040n@@80ATA700@0>SPk030^<@0`;S40;B/^0P0X9RT0100P7Ql0
EUAE05IDE@0b<342030^;`0402d[;00i=cP0>CLh03@b<`80AdE600@0:RPY02HT9@0V92D0:RPY0P0Y
9bP0100d<S<0?cdn03lm?P0Y9bP202LU9P0402<Q8P0Y9bP0:BLX04=1@P80;2X[00@0928S02`Z:`0/
:R/09R@U0P0X9RL00`1DDU<0928S02@R8`0302h/;@0402`Z:`0^;2d0;R`]059@D@80:bTZ00@0>cTj
03Dc=00e<c@0=S@e0P0j>3T00`1GEEH0<Bl`034_<003034_<00403<a<P0_;Bh0;bd^03/i>P<0:BLX
0P0_;Bh0100Y9bP0;2X[02`Z:`0/:R/203Le=P04055?D00Z:2T0:RPY02d[;080:bTZ00@0BdU:028P
8@0R8240<Bl`0P0b<340100c<C80CD]<04e;C00n?3d203Hd=@0402LU9P0Z:2T0:RPY02PV9`80B4I7
00@09R@U02l];P0_;Bh0<2h_0P0/:R/0101MFe`0;bd^02l];P0f=3D203@b<`0403Pf=`0]:b`0;B//
04e;C080<Bl`00@0:2HW02/Y:P0[:BX0<2h_0P0i=cP0100[:BX0<Bl`034_<00o?Ch203/i>P04034_
<00]:b`0;B//03<a<P80>SPi00<0EeEF028P8@0R82400`0Y9bP0100R8240BdU:04]9BP0c<C8202TW
:00402XX:@0f=3D0=S@e02l];P80=S@e00@0;R`]02LU9P0W9BH0ATA50P12@440100f=3D0;bd^02l]
;P0i=cP202d[;00403Dc=01?CDh0Cde>02PV9`80:BLX00@0=cDf03lm?P0o?Ch0>3Hg0P0g=CH0100i
=cP0<S0a038`<@0Q7b0203dk?00404e;C00i=cP0>CLh02PV9`80;2X[0`0^;2d01018ATL0?3Xk03`j
>`0U8b@203<a<P0404i<C@19AdP0BDM803/i>P80:bTZ00D071XK02DS900U8b@0<S0a03Pf=`010?oo
o`003017E5d01015Deh0BUQS04YHH`14DUd304QFH@80A59M00@0BeUT04IDG`16E5l0?4=;0P0=51X0
10032@T02a0:00/@2P0@4``200h?2@0400P70`0C3`h04`l>00/:1P8040l=00@03@`:00l=3P0?3@h0
8R0Q0P0^;2l0100/:Rd0:BL/02TW;00Z:RX202PX:00402TY:@0^;Rh0;Rh^02h^;P80;2`/00@0;Rh^
03<c<`0c<c<0410@0P0Z:RX0100O7al0>c/k03/k>`0o?cl201LG5`0402<S8`0O7al07alO020P8P80
@d=500@092@V02<S9@0S8bD0:BT[0P0P8280100T92H0>c/m03/k?@0X:2X202@T9P0401/K7@0T92H0
92@V02/[;@80?cm100@0820R02l_<@0_;c409RHX0P0Z:R`00`0f=SP0@T940492A00302LW:@0402HV
:00U9BL09BDW034a<`80;Rh`00@06ATK02DU9`0U9BL07QhP0P0M7Al0100j>S`06a/M01/K7@0P8202
024Q8@04020P800S8b<08b<S028R8P<06a/K0P0P8200100f=SH061PH01PH600Q8B4201PH600401TI
6@0/;2`0;2`/024Q8@80>3Ph00@08B4S01PH6P0H61X071`N0P0X:2X0100h>3X0:2PZ02PX:P0P8282
02PX:P0402HV:00Z:R`0:RX/04M7B@80:b/]00@0;Rh`02/[;@0[:bd0<C4c0P0W9bT0100h>3X092@V
02@T9P0O7b4202@T9P0403@d=P0Y:B/0:BT[024Q8`809bLY00@0=CDg02<S9@0S8bD0:RX/0P0S8bD0
100a<C<092@V02@T9P0J6Q`2028R900401lO8@0U9BL09BDW0492A0808R8T00@07QhP01dM7`0M7Al0
7alQ0P0b<S@0100A4A<0;Rh`02h^<00U9BL2024Q8`0400X:300820X020P:03Ti>`80@d=500@0@T=5
03Pi>`0h>C/0:b`^0P0_<380100k?3h0=cPj03Lh>P0_<38203<d=P0402dc<@0_=C<0;cDc04A:B080
M7Yh00@0_L?10;Fk^@2e^kT0Yjf[0P2`][@0102l_[/0NW]f07YkMP0Z9b82028K5@0402`T8@0W8b80
9b<R038b=0806AXN00@0;30a04a=C`1<CDl0;2d_0P0U9BL0100];Bl0:2HY02PV:@12@4<2040k?`04
02`/;P0V9RP09RHX02XZ;080=CDg00@07AdO02DU9`0U9BL0:b/]0P0Y:B/0100C4aD0@4120410@P0U
9BL202`/;P04030`<P0V9RP09RHX0410@P80<c<e00@0B4Q:02@T9P0T92H09RHX0P0b<S@0101;Bdd0
@d=504=3A@0_;c4202d];`0402`Z;@0`;S40<2ha040n@@80ATA700@0>SPk030^<@0`;S40;B/^0P0X
9RT0100P7Ql0EUAE05IDE@0b<342030^;`0402d[;00i=cP0>CLh03@b<`80AdE600@0:RPY02HT9@0V
92D0:RPY0P0Y9bP0100d<S<0?cdn03lm?P0Y9bP202LU9P0402<Q8P0Y9bP0:BLX04=1@P80;2X[00@0
928S02`Z:`0/:R/09R@U0P0X9RL00`1DDU<0928S02@R8`0302h/;@0402`Z:`0^;2d0;R`]059@D@80
:bTZ00@0>cTj03Dc=00e<c@0=S@e0P0j>3T00`1GEEH0<Bl`034_<003034_<00403<a<P0_;Bh0;bd^
03/i>P<0:BLX0P0_;Bh0100Y9bP0;2X[02`Z:`0/:R/203Le=P04055?D00Z:2T0:RPY02d[;080:bTZ
00@0BdU:028P8@0R8240<Bl`0P0b<340100c<C80CD]<04e;C00n?3d203Hd=@0402LU9P0Z:2T0:RPY
02PV9`80B4I700@09R@U02l];P0_;Bh0<2h_0P0/:R/0101MFe`0;bd^02l];P0f=3D203@b<`0403Pf
=`0]:b`0;B//04e;C080<Bl`00@0:2HW02/Y:P0[:BX0<2h_0P0i=cP0100[:BX0<Bl`034_<00o?Ch2
03/i>P04034_<00]:b`0;B//03<a<P80>SPi00<0EeEF028P8@0R82400`0Y9bP0100R8240BdU:04]9
BP0c<C8202TW:00402XX:@0f=3D0=S@e02l];P80=S@e00@0;R`]02LU9P0W9BH0ATA50P12@440100f
=3D0;bd^02l];P0i=cP202d[;00403Dc=01?CDh0Cde>02PV9`80:BLX00@0=cDf03lm?P0o?Ch0>3Hg
0P0g=CH0100i=cP0<S0a038`<@0Q7b0203dk?00404e;C00i=cP0>CLh02PV9`80;2X[0`0^;2d01018
ATL0?3Xk03`j>`0U8b@203<a<P0404i<C@19AdP0BDM803/i>P80:bTZ00D071XK02DS900U8b@0<S0a
03Pf=`010?ooo`003016E5l01015Deh0BUQS04YHH`14DUd304UGHP80AE=N00@0C5YU04MEH017EF00
@4M?0P0B6Al0100630`02a0:00/@2P0>4@X200d>200400T8100C3`h04`l>00/:1P804A0>00@02`X8
00`:2`0<2P/07adN0P0]:bh0100[:B`09R@Y02HT:@0/;2`302XZ:P80:b/[00@09bLW02<S8`0S8b<0
:2PX0P0`<300100N7Qh0;Bd]02d];@0I6AT202@T900403@d=00E5AD05ADE01lO7`8092@T00@09RHX
03@d=P0d=3H0<c<e0P0R8R@0100X:2X07AdO01dM7`0O7b4203<c=@0403De=`0_;c40;bla01/K7@80
;Bd_00@07alQ04A4AP14A4H09RHX0P0];Bl0100R8R@08B4S024Q8`0[:bd203lo@@0402@T9P0U9BL0
9BDW024Q8`80820R00@09RHX03De=`0e=CL06QXL0P0S8bD0100P82805aLI01LG6@0f=SP202HV:004
028R8P0U9BD09BDU01lO7`809RHV00@0:b/[024Q8@0Q8B407QhN0P0O7al0100g=cL08B4Q024Q8@0Z
:RX201/K6`0401TI6@0`<300<30`020P8080@D5100@09RHX01lO8@0O7b408R8T0P0Y:B/0100f=SP0
<30b030`<P0T92H202/[;@0402`/;P0Y:B/0:BT[04I6B080=3@f00@092@V02TY:`0Y:B/09bLY0P0K
6ad0100h>3X092@V02@T9P0R8R@2024Q8`0402d];`0`<380<30b024Q8`808R8T00@0=SHh02DU9`0U
9BL0:RX/0P0S8bD0100Z:R`0:b/]02/[;@0M7Al202<S9@0402HV:00T92H092@V04=3A@80820R00@0
:b/]02LW:@0W9bT07alQ0P0U9BL0100T92H0<S8d038b=00J6Q`201PH6P0400P82P0=3@l03@d?03De
=`80@D5300@0<c@f02LX:P0W:2X0:2T[0P0o@4801019BT`0<S<e038c=@0U9RP203<d=P0403Ho?00g
@3d0=d0m04eFD`80Z[>`00@0];fj0;W2_`2i`[l0X:VV0P2f_k`0102XZZL0Ce1;04m@B`0V8ah202@M
5`0402dU8P0b;Rd0<Rh]03<c=@808R<W00@0>Cdn0493A@12@dD0:2T[0P0W9bT0100Z:R`09B<V02DS
9P14@TD2038]<@0402`/;P0X:2X0:2PZ02/[;@80?3`n00@09BDW02HV:00V9RP07QhP0P0M7Al0100E
5AL0=cLi03Lg>@0i>C/202/[;@0402TY:`0c<cD0<c<e038b=080@d=500@0;Rh`02l_<@0_;c40AdM9
0P0[:bd0100c<cD0D51B051@DP0n?T02024Q8`04034_<P0X9RT0:2HY02TW:P80>CLj00@0AdE8040n
@@10?T40;2X]0P0N71l0100V92D0EE=D05ECE00b<34202/Y:P0402DS900^;2d0;R`]02h/;@80B4I7
00@0;B//02LU9P0W9BH0;bd^0P0Z:2T0100c<C80C4Y;04a:B`0c<C8202l];P0402d[;00[:BX0:bTZ
04M5AP80>CLh00<0:bTZ03Hd=@0f=3D00`0_;Bh0101UHf@0?cdn03lm?P0l>S/204=1@P0403dk?00i
=cP0>CLh065OH080@T1100@0<S0a034_<00a;c00;B//0P0d<S<0101?CDh0<Bl`034_<00/:R/202/Y
:P0402d[;00a;c00<Bl`03Xh>@80;bd^00@0;R`]03Dc=00e<c@0<Bl`0P0_;Bh0100^;2d0;bd^02l]
;P1BD542030^;`0402TW:00Z:2T0:RPY04Q6A`809R@U00@0928S02PV9`0X9RL0;bd^0P0m>c`01017
ADH0=S@e03Hd=@0a;c0202`Z:`0302TW:013@D80@d5200@0;R`]0P0T8R<0101JF5T0;bd^02l];P0`
;Rl202l];P0403<a<P0/:R/0;2X[04Q6A`80>CLh00@0:bTZ02`Z:`0/:R/0;2X[0P0c<C80100W9BH0
;R`]02h/;@0i=cP203`j>`0402h/;@0V92D09R@U03<a<P80<Bl`00<0FUQI01hL7@0N71d00`0/:R/0
100F51D0AdE604M5AP0i=cP202/Y:P0402`Z:`0d<S<0=38c02h/;@80?cdn00@0=cDf02/Y:P0[:BX0
=C<d0P15@d@0100g=CH0:bTZ02/Y:P0g=CH202d[;004034_<01IEeP0FEMH02TW:0809R@U00@0<2h_
03<a<P0c<C80:2HW0P0]:b`0100h=SL0=C<d03Dc=00U8b@2030^;`0404e;C012@440@T1102/Y:P80
;bd^00@0:bTZ02HT9@0V92D0AD=40P19AdP3030^;`0404A2@`11?d00@Cm003Dc=080;2X[00<0928S
024O800Q7b000P0V92D10?ooo`003016E5l01015Deh0BUQS04YHH`14DUd304UGHP80AE=N00@0C5YU
04QFH@18EV40@TUA0P0G7R@0100;4A402Pl900X?2@0;3PL200d>200400X91@0C3`h04`l>00/:1P80
4A0>00@02`X800X82@0:20T07adN0P0_;C00100^;2l0:2H[02PV:`0[:b/302TY:@80;2`/00@0:RXZ
02@T900T92@08R8R0P0U9BD0100I6AT0:RXZ02XZ:P0Q8B4202PX:00404=3@`0W9bL09bLW01hN7P80
7alO00@06QXL01LG6@0G5aT0?Si00P0N7R00100W9bT092@V02@T9P0Z:R`202d];`0404A4AP0b<S@0
<S8d024Q8`80:2PZ00@07AdO03Hf>00f=SP0;bla0P0Q8B<0100P82807AdO01dM7`0R8R@203`l?P04
02@T9P0V9RP09RHX01lO8@808B4S00@071`N038b=00b<S@0:2PZ0P0O7b4302@T9P0402h^<00o?d40
?cm102DU9@80:RXZ00@08R8R02`/;00/;2`0=SHf0P0Z:RX0100T92@08B4Q024Q8@0c<c<203@d=004
02TY:@0Q8B408B4Q01/K6`809RHV00@09bLW0410@010@400;bla0P0Z:R`0100Y:B/0:RX/02XZ;00d
=3H203Lg>@0402HV:00Z:R`0:RX/02PX:P807alQ00@0>c/m03Lg>@0g=cT05aLI0P0V9RP0100M7Al0
51@F01@D5P0e=CL202<S9@0402@T9P0M7Al07AdO02HV:080=SHh00@08B4S01/K7@0K6ad0=SHh0P0Z
:R`0100`<380:2PZ02PX:P0Z:R`203Lg>@<092@V00@0;Rh`02`/;P0/;2h0CDe?0P0R8R@0100d=3H0
:2PZ02PX:P0G5aT201PH6P0402l_<@0^;S00;Rh`01@D5P805ADG00@02@T;00l?4@0?3a408R8T0P0_
;c40100f=cT0>3Tk03Pi>`0h>C/203Xk?@0403Lh>P0[;2h0:b`^030a<`80B4U;00@07bPU04iGE01>
Ee@0WjRU0P2X/Jh0102g`;d0UYnL09JOW02b^kP20<39aP0409nQWP0Z:bH0:R/V02DR7@808a`F00@0
:20M034];00a;B`09RHX0P0/;C40101?De@0@D940452A00^;c4203@d=P04034a<`0^;2l0;R`_059@
D`80;BP/00@09RHX02LW:@0W9bT09RHX0P0o?d40100];Bl0:BT[02TY:`0D51H201<C5@0402@T9P0^
;S00;Rh`04a<CP80=cLi00@07QhP03<c=@0c<cD08R8T0P0l?3h0100c<cD0=SHh03Hf>012@T@202PX
:P0402@T9P11@D<0@D5304a<CP809bLY00@0<c4d02`Z;@0/:Rd0928U0P0e<cH0101>C4l0ATA704I4
A`0_;C0202`Z;@0402/Y:P1=Bd`0CD]<03Hd=@80;B//0`0X9RL0100_;Bh0BDM804U7B00i=cP2034_
<00403Xh>@0`;Rl0<2h_03@b<`80EUAE00@0>CLh03@b<`0d<S<0=C<d0P0[:BX01017ADH0@T110490
@@0`;Rl203/i>P<0=C<d00@0C4Y;03Le=P0g=CH08R0Q0P0b<340100V92D07adN01lM7P12@442034_
<00402XX:@0_;Bh0;bd^02@R8`80<Bl`00@0@d5203@b<`0d<S<0:RPY0P0U8b@0100W9BH0<c4b03<a
<P0i=cP203Hd=@0403Dc=00m>c`0?C/l03/i>P80=cDf00@0=C<d02h/;@0^;2d0De5B0P0j>3T0100X
9RL0;B//02d[;017ADH202h/;@0402PV9`0/:R/0;2X[03Pf=`80<c4b00@0DDm@038`<@0b<340=38c
0P0T8R<0100Y9bP0>SPi03Xh>@0g=CH3030^;`809bDV00<0FeUJ03/i>P0k>CX00`0a;c0303<a<P<0
ATA50`0a;c00100^;2d0<Bl`034_<00T8R<202`Z:`0403@b<`0n?3d0?S`m02h/;@809R@U00@0;R`]
02DS900U8b@0EE=D0P0Y9bP0100b<340:bTZ02/Y:P0P7Ql203`j>`04040n?`0]:b`0;B//02d[;080
<S0a00@0;2X[04=1@P13@D80?3Xk0`0]:b`204a:B`0403Xh>@0Z:2T0:RPY03Tg>080<Bl`00@0<2h_
05aJF`1LFU/0;bd^0P0T8R<0100[:BX0;R`]02h/;@0U8b@202h/;@0403/i>P0f=3D0=S@e034_<080
9bDV00@0@3ho044o@011?d00=S@e0P0c<C80100[:BX0=cDf03Le=P0k>CX204U7B00303dk?00c<C80
<c4b00<0<2h_00@0=C<d02XX:@0Z:2T0<2h_0P0Y9bP1034_<040<S0a0@3oool000`0AUAO00@0AE=N
04YHH`1:F6<0A59M0`19Ef8204ECGP0404aJI@18EV40B5IQ0499D@805ahT00@02a4A00X?2@0:3`T0
2`h70P0=3PP0100:2@D04`l>01<?3P0;2PH2014@3P0400/:200:20T02PP901lM7P80;bd`00@0;R`_
02PV:`0X9R/0:b/[0`0Y:BT202`/;00402XZ:P0T92@092@T028R8P809BDU00@06ATI02XZ:P0Z:RX0
8B4Q0P0X:2P01013@d<09bLW02LW9`0N7Qh201lO7`0401XJ700G5aT05aLI03hn@0807QhP00@09bLY
02@T9P0T92H0:RX/0P0];Bl01014A4H0<S8d038b=00Q8B<202PX:P0401dM7`0f=SP0=SHh02l_<@80
8B4S00@0820R01dM7`0M7Al08R8T0P0l?3h0100T92H09RHX02HV:00O7b42024Q8`0401`L7P0b<S@0
<S8d02PX:P807alQ0`0T92H0100^;S00?cm103lo@@0U9BD202XZ:P04028R8P0/;2`0;2`/03Hf=P80
:RXZ00@092@T024Q8@0Q8B40<c<c0P0d=3@0100Y:BT08B4Q024Q8@0K6a/202HV9P0402LW9`10@400
@41002l_<@80:RX/00@0:BT[02XZ;00Z:R`0=3@f0P0g=cT0100V9RP0:RX/02XZ;00X:2X201lO8@04
03/k?@0g=cT0=cLi01LG6@809RHX00@07AdO01@D5P0D51H0=CDg0P0S8bD0100T92H07AdO01dM7`0V
9RP203Hf>004024Q8`0K6ad06a/M03Hf>080:RX/00@0<30b02PX:P0X:2X0:RX/0P0g=cT302@T9P04
02h^<00/;2h0;2`^04e=C`808R8T00@0=3@f02PX:P0X:2X05aLI0P0H61X0100_;c40;Rh`02h^<00D
51H201DE5`0400T92`0?3a403`lA028R9080;bla00@0=SLi03Pi>`0h>C/0>3Tk0P0j>cd0100g>3X0
:b`^02//;P0`<C<204Q9B`0401lX9@1>Ee@0CUMD09nXY@80Z;6^00@0]l2m09JOW02FWi`0/[^h0P30
bLH0102OXIh0:R/V02X[9P0U8Qd202<L5P0402PP7@0a;B`0<Bd/02HV:080;2da00@0Ce=D0452A011
@T@0;Rla0P0d=3H0100a<C<0;R`_02h/;`1BD5<202dX;00402HV:00W9bT09bLY02HV:080?cm100@0
;Bd_02TY:`0Y:B/051@F0P0C4aD0100T92H0;Rh`02h^<01<C4h203Lg>@0401hN800c<cD0<c<e028R
9080?3`n00@0<c<e03Hf>00f=SP0@T940P0X:2X0100T92H0@D530451@`1<C4h202LW:@0403<a=00/
:Rd0;2X]02@R9@80=C<f00@0CTa?04I4A`16A4L0;bd`0P0/:Rd0100[:BX0CD]<04e;C00f=3D202d[
;0<0:2HW00@0;bd^04U7B019AdP0>CLh0P0a;c00100j>3T0<2h_030^;`0d<S<205IDE@0403Tg>00d
<S<0=38c03Dc=080:bTZ00@0AdE60490@@12@440<2h_0P0k>CX303Dc=00404a:B`0g=CH0=cDf028P
8@80<S0a00@09R@U01lM7P0O7Ah0@T110P0a;c00100Z:2T0;bd^02l];P0T8R<2034_<00404=1@P0d
<S<0=38c02XX:@809B<T00@09bDV03<a<P0c<C80>CLh0P0f=3D0100e<c@0?C/l03dk?00k>CX203Le
=P0403Dc=00^;2d0;R`]05=ADP80>SPi00@0:2HW02d[;00]:b`0AdE60P0^;2d0100X9RL0;2X[02`Z
:`0h=SL203<a<P04055?D00b<340<S0a03@b<`80928S00@0:BLX03Xh>@0j>3T0=cDf0`0`;Rl202LU
9P0305]IFP0k>CX0>cTj00<0<Bl`0`0c<C8304I4A@<0<Bl`00@0;R`]034_<00a;c00928S0P0/:R/0
100d<S<0?S`m03hl?@0^;2d202HT9@0402h/;@0U8b@09B<T05ECE080:BLX00@0<S0a02/Y:P0[:BX0
81hO0P0l>S/01010?Sl0;B//02d[;00]:b`2038`<@0402`Z:`13@D80@d5203`j>`<0;B//0P1<BT/0
100j>3T0:RPY02XX:@0i=cP2034_<004030^;`1LFU/0G5YK02l];P80928S00@0:bTZ02h/;@0^;2d0
9B<T0P0^;2d0100k>CX0=S@e03Hd=@0a;c0202LU9P04040n?`11?d00@Cm003Hd=@80<c4b00@0:bTZ
03Le=P0g=CH0>cTj0P19AdP00`0m>c`0<c4b03<a<P03030^;`0403Dc=00Z:2T0:RPY030^;`80:BLX
0@0a;c01038`<@40oooo000<04IDG`0404ECGP1:F6<0BUQS04ABG@80BEMR00@0BUQS04IDG`16E5l0
CE]V0P18EV401012BE406b8X01/R:00>51@200X?2@0400T<1@0<3@L030d700/:1P804`l>00@02`X6
014@3P0A40h02`X80P0:20T0100P7Ql0<c4d03<a=00b<3<202`Z;`0402TY:@0V9RH09RHV02PX:080
<30`00@0<S8b02/[:`0[:b/08b<S0P0O7al00`0I6AT09bLW02LW9`03020P80040492@P0^;Rh0;Rh^
020P80808b<S00@0:2PZ01TI6`0I6A/0F5QJ0P0U9BL0100Z:R`0:BT[02TY:`0];Bl201dM7`0404=3
A@0W9bT09bLY02<S9@806a/M00@08R8T02/[;@0[:bd0?Cdo0P0O7b40100S8bD07alQ01lO8@0P8282
03/k?@0402LW:@0[:bd0:b/]024Q8`809RHX00@09bLY03Xj?00j>S`0@4120P0P8280100U9BL0:RX/
02XZ;00K6ad204A4AP04024Q8@0W9bL09bLW01dM7@80:RXZ00@0=SHf02TY:@0Y:BT08B4Q0P0L71`0
100V9RH0?Cdm03dm?@0L71`201lO7`0401DE5@0C4a<04a<C02HV9P80<c<c00@0;Bd_02/[;@0[:bd0
:2PZ0P0R8R@0100Z:R`0<S8d038b=00O7b42020P8P0402LW:@0K6ad06a/M03De=`80?Si000@05ADG
02d];`0];Bl08B4S0P0K6ad0100b<S@08R8T028R900U9BL201/K7@0402<S9@0i>C/0>CTk024Q8`80
5aLI00@0<30b02PX:P0X:2X0<30b0P0X:2X0100W9bT0>c/m03/k?@0T92H2020P8P04028R900W9bT0
9bLY04i>D0807AdO00@0<30b01lO8@0O7b4030`>0P0B4Q@0100T92H08b<U02<S9@0B4Q@201TI6`04
00h>400@4180410B01HF608092@V00@0<34c0493A@12@dD0?Ci00P0[;2h0100];S00>CXl03Tj?00h
>C/2038c=@0402Tb;`2@VIH0T9VF0;>l^@80ZkBa00@0Z[>`0;[3`02j`l00[;Fb0P2f_k`0102XZZL0
6a`G01/L5`0V8ah2020I4`0401dE4P0X92<0:2@S01HF6080<C8f00@0CU9C03<d=P0c=3H09bPZ0P0f
=SP0100b<S@0=38e03@b=@1OGF0202lZ;P04038b=00_;c40;bla02<S9@80>CTk00@0;Bd_02l_<@0_
;c407AdO0P0V9RP0100l?3h09bLY02LW:@0l?3h20410@P04028R900^;S00;Rh`02<S9@80:RX/00@0
G5aN03Ti>`0i>C/09RHX0P0];Bl0100[:bd0;2`^02`/;P1?Ce4203Lg>@0402TW:P0/:Rd0;2X]02HT
9`80<Blb00@0ATA703Xh>`0j>3/0:bT/0P0i=cX0100`;Rl0C4Y;04a:B`13@D8203dk?00403Xh>@0c
<C80<c4b03dk?080DU1A00@0@3ho03Hd=@0f=3D0?C/l0`0]:b`205=ADP04038`<@0[:BX0:bTZ02l]
;P8081hO0`0l>S/0100X9RL0<S0a038`<@0]:b`202l];P0404]9BP13@D80@d5202DS9080?S`m00@0
<c4b02h/;@0^;2d0CD]<0P18ATL0100_;Bh0=cDf03Le=P0Y9bP203/i>P0404E3A011?d00@Cm003<a
<P80;2X[00@09R@U03@b<`0d<S<0>3Hg0P0e<c@0100a;c00=cDf03Le=P0d<S<202/Y:P0403@b<`0X
9RL0:2HW04a:B`80>cTj00@08AlP02XX:@0Z:2T0@3ho0P0^;2d0100V92D0;2X[02`Z:`0m>c`202`Z
:`0405YHF@0d<S<0=38c03dk?0809R@U00@0:RPY034_<00a;c00>SPi0P0[:BX0100Z:2T08R0Q028P
8@1?CDh203Le=P<0;bd^00@0;R`]03@b<`0d<S<0?C/l0P17ADH0100^;2d0;2X[02`Z:`0d<S<203Dc
=00402HT9@0/:R/0;2X[03<a<P80@T1100@0<S0a02/Y:P0[:BX0<c4b0P0U8b@0101CDE80>3Hg03Pf
=`0e<c@202<Q8P<0:BLX00@0@d5202d[;00]:b`0;2X[0P0^;2d0100X9RL0@T110490@@0k>CX202/Y
:P0402TW:01=Bd`0CD]<03Hd=@808R0Q00@0=38c02h/;@0^;2d09R@U0P1ACe00100d<S<09R@U02HT
9@0/:R/203<a<P0402h/;@0g=CH0=cDf040n?`80=38c00@0>CLh02DS900U8b@0>cTj0`18ATL203Hd
=@04020N7`0l>S/0?3Xk030^;`80@Cm000@0AD=403Tg>00i=cP08b4R0P0X9RL0100f=3D08AlP024O
800V92D202DS9040>3Hg0@11?d010?ooo`003015Deh704MEH0D0AUAO00@0@Dm@01d[:`0M:b/01a4B
0P084@h0100;40`02@X500T:1@0<2`H201LD3`04010;1`0<1`<030L3010<2@805A4>00@06APF02PX
:00X:2P0<c<c0P0_;bl0100g=cL0;Bd]02d];@0T92@202HV9P04030`<00f=SH0=SHf030`<0809RHV
00@03Ph>02<S8`0S8b<0;2`/0P0I6AT0100V9RH0ATI604I6AP0N7Qh202TY:@0402/[;@0a<C<0<C4c
02h^<080A4A600@0820R02TY:`0Y:B/06QXL0P0N7R00100Y:B/0=3@f03@d=P0X:2X201/K7@0402@T
9P0U9BL09BDW0451@`80:2PZ00@09bLY01TI6`0I6A/08b<U0P0R8R@01013@dD09BDW02DU9`0/;2h2
02TY:`04028R900];Bl0;Bd_04M7B@80:b/]00@09BDW01`L7P0L71h09BDW0P0[:bd0100a<C406a/K
01/K6`0P820202<S8`03024Q8@0W9bL09bLW00<08b<S00<06ATI02d];@0];Bd00`0U9BD0100M7Ad0
30`<00`<300M7Ad202h^;P04034a<`0O7b407alQ02DU9`<06a/M0P0j>S`0100M7Al09RHX02HV:00Z
:R`202HV:00401dM7`15ADL0ADE7028R9080:b/]00@08b<U01`L7P0L71h0;Rh`0P0j>S`0100O7b40
8B4S024Q8`0T92H203Ti>`0401dM7`0V9RP09RHX028R9080;2`^00@0<30b01lO8@0O7b406a/M0P12
@T@0100J6Q`08B4S024Q8`0R8R@201`L7P0404e=C`0L71h071`N02/[;@80410B00@04A4C02TY:`0Y
:B/0;2`^0P0N7R00100B4Q@05ADG01DE5`0K6ad200`<3P0401/K7@0O7b407alQ030a=@80=cPl00@0
9BHZ02l`=00_<3@0@D960P1DEET0100];S80<S<g038c=`1eMgH20:^][0040;Nk^P2[[jh0Zjn^0;Jl
^P80Yk2]0`2X/Jh0101_LVT0<C8Z034b:P0R7aX202PP7P0401lF5`0M6AX07ATJ01dM7`808B8V00@0
:B`a02@U:P0T9BX05AHJ0P0U92T0100W9bT0;2X[02`Z:`1IEeP2030/;@04034a<`0i>C/0>CTk03<c
=@80ADE700@0<S8d03De=`0e=CL05aLI0P0O7b40100];Bl0<C4c034a<`0U9BL203Hf>00403lo@@0[
:bd0:b/]02`/;P80<C4c00@0DU1C03Xh>`0j>3/0=cDh0P0b<3<0100S8B@0@3i1040n@@0T8RD205ID
E`0403Hd=`0S8B@08b4T0490@`80;R`_00@0=cDh02LU:00W9BP0AD=60P0[:B`0100g=CH0;R`]02h/
;@1CDE8201TG600403@b<`0T8R<0928S02LU9P80=cDf00@0A49302<Q8P0S8B80=S@e0P0Q7b00100f
=3D0?3Xk03`j>`0b<34202l];P0402<Q8P0/:R/0;2X[02<Q8P80?cdn00@0:2HW028P8@0R8240<S0a
0P0d<S<0100O7Ah0EE=D05ECE00S8B82034_<00402d[;00j>3T0>SPi02HT9@80H5iO00@09bDV02l]
;P0_;Bh0;R`]0P0[:BX00`0T8R<0EUAE05IDE@0302@R8`0402d[;00[:BX0:bTZ03Hd=@80@Cm000@0
:bTZ02l];P0_;Bh0;2X[0P0P7Ql0100a;c00>CLh03Tg>016A4D203hl?@<0:RPY00@0=cDf03Tg>00i
=cP0;2X[0P0Z:2T0100f=3D0<2h_030^;`1BD542051>C`0402l];P0h=SL0>3Hg03Hd=@80>SPi00@0
D4i?03Tg>00i=cP0<S0a0P0Z:2T01019AdP0C4Y;04a:B`0a;c0202l];P04034_<00_;Bh0;bd^03@b
<`80DU1A00<08b4R02`Z:`0/:R/00`0c<C80100b<34081hO020N7`0a;c0204U7B00403Dc=00Z:2T0
:RPY038`<@809B<T00@0E59C03Tg>00i=cP0>SPi0P0^;2d0100T8R<0:RPY02XX:@19AdP203`j>`04
02HT9@0^;2d0;R`]02l];P80?3Xk00@0CD]<02DS900U8b@0=38c0P0k>CX0100d<S<0:bTZ02/Y:P0]
:b`202h/;@0403<a<P10?Sl0@3ho0490@@80;B//00@0<c4b02XX:@0Z:2T0;2X[0P0h=SL0100_;Bh0
?3Xk03`j>`0m>c`2030^;`<0?C/l00@0EUAE02l];P0_;Bh0<Bl`0P0Y9bP302HT9@04055?D00]:b`0
;B//03@b<`80:BLX00@0928S03Dc=00e<c@0<Bl`0P0d<S<1040n?`40?C/l0@3oool000`0AE=N1`17
EF0504IDG`04045?D00M:b/07B/[00LA4P80214>00@02a0<00T:1@092PD030/60P0G50l0100@2`L0
30L300`70`0@30T201DA3P0401TH5P0X:2P0:2PX03<c<`80;bl_00@0=cLg02d];@0];Bd092@T0P0V
9RH0100`<300=SHf03Hf=P0`<30202HV9P0400h>3P0S8b<08b<S02`/;0806ATI00@09RHV04I6AP16
ATH07QhN0P0Y:BT0100[:bd0<C4c034a<`0^;S0204A4AP04020P8P0Y:B/0:BT[01XJ70807QhP00@0
:BT[03@d=P0d=3H0:2PZ0P0K6ad0100T92H09BDW02DU9`11@D<202PX:P0402LW:@0I6A/06ATK02<S
9@808R8T00@0@d=502DU9`0U9BL0;2`^0P0Y:B/0100R8R@0;Bd_02d];`17AdT202/[;@0402DU9`0L
71h071`N02DU9`80:b/]00@0<C4a01/K6`0K6a/0820P0P0S8b<00`0Q8B409bLW02LW9`0302<S8`03
01TI6@0];Bd0;Bd]00<09BDU00@07AdM00`<300<30`07AdM0P0^;Rh0100a<C<07alQ01lO8@0U9BL3
01/K7@80>SXl00@07AdO02HV:00V9RP0:RX/0P0V9RP0100M7Al0ADE704E5A`0R8R@202/[;@0402<S
9@0L71h071`N02h^<080>SXl00@07alQ024Q8`0Q8B<092@V0P0i>C/0100M7Al09RHX02HV:00R8R@2
02`/;P04030`<P0O7b407alQ01/K7@80@T9400@06QXL024Q8`0Q8B<08R8T0P0L71h0101=CDl071`N
01`L7P0[:bd2010@4P04014A4`0Y:B/0:BT[02`/;P807QhP00@04Q8D01DE5`0E5AL06a/M0P0<30h0
100K6ad07alQ01lO8@0`<CD203Lh?00402DV:P0_<3@0;c0d0452AP80E5EI00@0;Bhb038c=`0b<cL0
MGMf0P2[[J`0102g^kX0Zjn^0:^_[P2f_;X20:N`[@<0Z;6^00@0Kg9Y034b:P0a<RX08QlJ0P0X81h0
100O5QL07ATJ01dI6P0M7Al2024R9P0402T/<@0T9BX092DZ01DF6P809B@Y00@09bLY02`Z:`0/:R/0
FEMH0P0`;2d0100a<C<0>CTk03Ti>`0c<cD204E5A`04038b=00e=CL0=CDg01LG6@807alQ00@0;Bd_
034a<`0a<C<09BDW0P0f=SP0100o?d40:b/]02/[;@0/;2h2034a<`04059@D`0j>3/0>SPk03Le>080
<S0c00@08b4T040n@@10?T40928U0P1FE5L0100f=3L08b4T02<Q9012@4<202h/;`0403Le>00W9BP0
9bDX04E3AP80:bT/00@0=cDf02h/;@0^;2d0De5B0P0I5aP0100d<S<0928S02@R8`0W9BH203Le=P04
04A2@`0S8B808b4R03Hd=@808AlP00@0=S@e03`j>`0l>S/0<S0a0P0_;Bh0100S8B80;2X[02`Z:`0S
8B8203lm?P0402PV9`0R82408R0Q038`<@80=38c00@07adN05ECE01EDe@08b4R0P0a;c00100]:b`0
>SPi03Xh>@0V92D2061NG`0402LU9P0_;Bh0;bd^02h/;@80:bTZ00<0928S05IDE@1FE5D00`0T8R<0
100]:b`0:bTZ02/Y:P0f=3D2044o@00402/Y:P0_;Bh0;bd^02`Z:`8081hO00@0<Bl`03Tg>00i=cP0
ATA50P0n?3d302XX:@0403Le=P0i=cP0>CLh02`Z:`80:RPY00@0=S@e030^;`0`;Rl0DU1A0P1@CTl0
100_;Bh0>3Hg03Pf=`0f=3D203Xh>@04051>C`0i=cP0>CLh038`<@80:RPY00@0BDM804a:B`1<BT/0
<Bl`0P0_;Bh0100a;c00;bd^02l];P0d<S<2059@D@0302<Q8P0/:R/0;2X[00<0<c4b00@0<S0a020N
7`0P7Ql0<Bl`0P19AdP0100e<c@0:RPY02XX:@0b<34202DS900405ABD`0i=cP0>CLh03Xh>@80;R`]
00@0928S02XX:@0Z:2T0BDM80P0l>S/0100V92D0;R`]02h/;@0_;Bh203`j>`0404e;C00U8b@09B<T
03@b<`80>cTj00@0=38c02/Y:P0[:BX0;B//0P0^;2d0100c<C80@3ho040n?`12@44202d[;00403<a
<P0Z:2T0:RPY02`Z:`80>3Hg00@0;bd^03`j>`0l>S/0?C/l0P0`;Rl303dk?00405IDE@0_;Bh0;bd^
034_<080:BLX0`0V92D0101ACe00;B//02d[;00d<S<202TW:00402@R8`0e<c@0=C<d034_<080=38c
0@10?Sl103dk?040oooo000<04ECGPH0AeEP1P16E5l01012D54082h_020^;`074A8200H?3P0400/@
300;30L02``700d<1`805a@?00@03`X600d8100=20@040`90P0B3P/0100F5A<0:2PX02PX:00e=CD2
030`<00302l_;`0Z:RX0:RXZ00<09RHV00@0:BTY02h^;P0^;Rh0=3@d0P0j>SX0100E5AD09bLW02LW
9`0/;2`202DU9@0402`/;018B4P0B4Q803<c<`80;bl_00@092@V02DU9`0U9BL09bLY0P0i>C/0100S
8bD09bLY02LW:@0R8R@202LW:@04028R9010@480@41202@T9P809BDW00@07QhP02<S9@0S8bD0:RX/
0P0d=3H0100M7Al07alQ01lO8@0U9BL2028R90040492A00_;c40;bla02LW:@80<30b00@08B4S02DU
9`0U9BL0>3Pj0P0/;2h0100S8bD07alQ01lO8@0Q8B<202DU9`0403Xj>P0J6QX06QXJ01lO7`808B4Q
00@092@T02d];@0];Bd08B4Q0P0T92@0100M7Ad0;2`/02`/;00Z:RX202PX:00401lO7`0B4Q804Q8B
01/K6`80;2`/00@0>3Pj024Q8`0Q8B<08R8T0P0N7R00100P8280>3Pj03Ph>P0O7b4202TY:`04020P
8P0R8R@08R8T01dM7`80@D5300@07AdO024Q8`0Q8B<07QhP0P0M7Al0100[:bd0?3`n03`l?P0O7b42
020P8P04028R900e=CL0=CDg02@T9P809BDW00@08R8T02XZ;00Z:R`0;Rh`0P0P8280100M7Al0@D53
0451@`0L71h201hN800402HV:00I6A/06ATK03Xj?08061PJ00@092@V00000P000080<c<e0P0G5aT0
100T92H07alQ01lO8@0J6Q`201/K7@0401dM7`0<30h030`>01TI6`806QXL00@092DW02`];`0/;Bl0
<C8d0`0i>S`203Pi>`0402h_<@1:Bdd0BT]=0;Jh]`80_Knn00@0[k>b0:6UY02QYJ@0Y:ZX0P2OZ:D3
0;6j]`0404mBB@0Y:R80:BXR02TV8@80<b/Y00@0:B0Q01dI6P0M6AX061PJ0P0Q8RH0100U:2d092DZ
02@U:P0L7B42020O9004024Q8`0]:b`0;B//04i<C@809B4R00@09bLY02`/;P0/;2h09RHX0P16ATP0
100b<S@0;bla02l_<@0N7R0202`/;P04034a<`0Z:R`0:RX/02PX:P80<S8d00@0?3`n03<c=@0c<cD0
:b/]0P0e=CL0100^;2l0@d5404=1A013@D@202h/;`0403`j?@0L6Qd071XM038`<`80=C<f00@0BDM:
04I4A`16A4L0=38e0P0e<cH0100]:bh0@T130490@`0f=3L202HT9`0402h/;@0/:R/0;2X[05IDE@80
9B<T00@0=cDf02@R8`0T8R<09R@U0P0e<c@01017ADH09B<T02DS900b<34202DS9004038`<@0h=SL0
>3Hg03/i>P80:bTZ00@0;B//03<a<P0c<C80;R`]0P18ATL0100a;c009B<T02DS900a;c0202d[;004
02XX:@1JF5T0FUQI02d[;080=38c00@0;R`]03Tg>00i=cP09B<T0P1JF5T00`0T8R<0:BLX02TW:003
02`Z:`0402HT9@1EDe@0EE=D02TW:0809B<T00@0;R`]02/Y:P0[:BX0<c4b0P0n?3d0100Z:2T0;R`]
02h/;@0^;2d202@R8`0403`j>`11?d00@Cm004Q6A`80CD]<00@0=38c03Hd=@0f=3D0?C/l0P19AdP0
100d<S<0=S@e03Hd=@0k>CX203Xh>@04055?D01HEUL0F5IG03Le=P80?3Xk00@0:bTZ02`Z:`0/:R/0
?C/l0P0i=cP0100b<340;bd^02l];P19AdP205IDE@<0@d5200@0AD=4044o@011?d00A4930P1THV<0
100j>3T0?C/l03dk?00`;Rl203Le=P04038`<@0T8R<0928S034_<080AD=400@0>3Hg02<Q8P0S8B80
<Bl`0P0X9RL0101DDU<0?3Xk03`j>`0k>CX2038`<@0402PV9`0Z:2T0:RPY04Q6A`80AdE600@0<S0a
030^;`0`;Rl0;R`]0P0d<S<01014@T<0;bd^02l];P0[:BX203Le=P0403`j>`0k>CX0>cTj03hl?@80
>cTj00@0>SPi04=1@P13@D80HEmP0P11?d00100k>CX0;R`]02h/;@0/:R/203Le=P0403@b<`10?Sl0
@3ho04i<C@80;bd^00@0>SPi03@b<`0d<S<0A4930P0n?3d0100`;Rl05A<D01DC500/:R/202h/;@04
03<a<P1;BDX0BdU:02`Z:`80;bd^00@08b4R03Hd=@0f=3D0F5IG0P1:B4T103Tg>0406aTJ0@3oool0
00`0AE=N3016E5l01015DeH09S@e02Hd=@0:51D200<<2`0400X>3@0>3P`03Ph<00l>2P805Q<>00@0
3PT500l:1P0?2PH040`90P0>2PL0100A40h09bLW02LW9`0g=cL2034a<@04038b<P0/;2`0;2`/02PX
:080:RXZ00@0:BTY02PX:00X:2P0;Rh^0P0g=cL0100M7Ad0:2PX02PX:00V9RH202XZ:P0402@T900c
<c<0<c<c03Hf=P807AdM00@09bLY01lO8@0O7b409BDW0P0^;S00100X:2X07alQ01lO8@0Q8B<202LW
:@0401LG6@14A4H0A4A6028R9080;bla00@07QhP02PX:P0X:2X0820R0P1:BT`0100R8R@0;Rh`02h^
<00/;2h202<S9@0403Ti>`0f=SP0=SHh01/K7@80:BT[00@0820R01`L7P0L71h0:2PZ0P0a<C<0100R
8R@092@V02@T9P0N7R02020P8P0403lo?`0R8R808R8R01dM7@807alO00@09bLW030`<00`<3009RHV
0P0S8b<0100R8R80:2PX02PX:00`<30202XZ:P04028R8P0I6AT06ATI01HF5P809bLW00<0?Si002HV
:00V9RP00`0P8280100V9RP0=cLi03Lg>@0T92H202XZ;004020P8P0Y:B/0:BT[02XZ;080BDU;0`0/
;2h0100^;S00=3@f03@d=P0];Bl204E5A`0402LW:@0V9RP09RHX02LW:@<0=CDg0P0[:bd0100S8bD0
:BT[02TY:`0/;2h3024Q8`80@41200@07alQ01/K7@0K6ad0;Bd_0P0P8280100i>C/0:b/]02/[;@0E
5AL200l?4@0404E5A`0L71h071`N01hN80808R8T00@0820R01/K7@0K6ad06ATK0P0;2`d00`0H61X0
5aLI01LG6@03028S9@0403`m?`0o@480?d1203Hg>@80;2d_00@0EEIH08f>T02=SY00[:j]0P2RY:<0
102PY:<0Yj^Z0:N[ZP2Y[jd209jWY0040:ff/`2PZJH0X:VV034d:`80:BXR00@0=34/03Ta;`0i<Bl0
;RDV0P0M6AX0100G5aT092DY02@U:@0D5a`201PI7P0401hO8`0L6b0071/P028R9080=C<d00@0AdE6
02<O800S7b00:2PZ0P0Y:B/0100M7Al0Bd]=04];C@0`<38201hN800401`L7P0^;S00;Rh`03@d=P80
:BT[00@0:RX/02`/;P0/;2h0=cLi0P0k>cd0100^;S00=CDg03De=`0U8bH202TW:P04061NH@0`;S40
<2ha02XX:`80;2X]00@0:BLZ02l]<00_;C00B4I90P1DDUD0100j>3/0<Blb034_<P0[:B`2044o@P04
034_<P0^;2l0;R`_02PV9`80:RPY00@0DDm@02`Z:`0/:R/0<Bl`0P0Q7b00100V92D0<c4b03<a<P18
ATL202LU9P0402/Y:P0Y9bP0:BLX02/Y:P80<2h_00@0AD=402DS900U8b@0:bTZ0P0^;2d0100_;Bh0
BDM804U7B00e<c@202<Q8P0402l];P0S8B808b4R02@R8`80CTa=00@0;R`]02d[;00]:b`0:RPY0P0f
=3D0100W9BH0E59C05ABD`0a;c02030^;`0403Dc=00g=CH0=cDf02l];P80EUAE00@0=S@e02XX:@0Z
:2T0;R`]0P0/:R/0100_;Bh0?3Xk03`j>`0[:BX202h/;@04034_<00Y9bP0:BLX03<a<P80=38c00@0
<S0a04Y8B@1:B4T0:bTZ0`0^;2d204M5AP0402/Y:P0a;c00<Bl`02h/;@80=C<d00@0?S`m05=ADP1C
DE80=S@e0P0c<C80100_;Bh0:RPY02XX:@0_;Bh203lm?P0402h/;@0[:BX0:bTZ03Pf=`80De5B00@0
=38c03Le=P0g=CH0>CLh0P0a;c00100c<C80E59C05ABD`0c<C8202`Z:`0402h/;@0m>c`0?C/l03Hd
=@80<S0a00@0?3Xk04m=CP1?CDh0DU1A0P0d<S<0100^;2d0;2X[02`Z:`1BD542040n?`0403Tg>00f
=3D0=S@e02`Z:`809bDV00@0=38c04=1@P13@D80<c4b0`0Z:2T202d[;00403dk?011?d00@Cm002LU
9P80<Bl`00@0>cTj03`j>`0l>S/0>cTj0P0d<S<0100]:b`0=38c03@b<`1IEeP203Dc=00402XX:@0V
92D09R@U02XX:@80=cDf00@0>cTj04E3A015@d@0CD]<0P0^;2d0100c<C80;B//02d[;00_;Bh204Y8
B@0403Tg>00F51D05Q@E03@b<`80;B//00@09B<T05=ADP1CDE80=cDf0P0_;Bh0100]:b`0<S0a038`
<@0g=CH203Dc=040;bd^0@0G5AH10?ooo`003015Deh<04IDG`0404ECEP0V=3D09S@e00XD5@800``;
00@02Ph=00h>300>3P`03`h:0P0F4`h0100>2@D03`X600l:1P0@30T200h:1`04014@3P0W9bL09bLW
03Lg=`80<C4a00@0<S8b02`/;00/;2`0:2PX0P0Z:RX0100Y:BT0:2PX02PX:00^;Rh203Lg=`0401dM
7@0X:2P0:2PX02HV9P80:RXZ00@092@T03<c<`0c<c<0=SHf0P0M7Ad0100W9bT07alQ01lO8@0U9BL2
02h^<00402PX:P0O7b407alQ024Q8`809bLY00@05aLI04A4AP14A4H08R8T0P0_;c40100N7R00:2PZ
02PX:P0P828204Y:C004028R900^;S00;Rh`02`/;P808b<U00@0>CTk03Hf>00f=SP06a/M0P0Y:B/0
100P828071`N01`L7P0X:2X2034a<`04028R900T92H092@V01hN8080820R00@0?clo028R8P0R8R80
7AdM0P0O7al0100W9bL0<30`030`<00V9RH202<S8`04028R8P0X:2P0:2PX030`<080:RXZ00@08R8R
01TI6@0I6AT05QHF0P0W9bL00`0n?T009RHX02HV:003020P8P0402HV:00g=cT0=cLi02@T9P80:RX/
00@0820R02TY:`0Y:B/0:RX/0P19BD/302`/;P0402h^<00d=3H0=3@f02d];`80ADE700@09bLY02HV
:00V9RP09bLY0`0e=CL202/[;@0402<S9@0Y:B/0:BT[02`/;P<08B4S0P10@480100O7b406a/M01/K
7@0];Bl2020P8P0403Ti>`0[:bd0:b/]01DE5`803`lA00@0ADE701`L7P0L71h07QhP0P0R8R@0100P
82806a/M01/K7@0I6A/200/;3@0301PH6P0G5aT05aLI00<08R<U00@0?3do03m0@P0o@480=SLi0P0/
;Bl0101EEUP0SHj@08f>T02/[Zd20::TX`040:2TX`2WZjX0Yj^Z0:V_[@80WZNT00@0[KJc0:2YYP2P
ZJH0<C@[0P0Y:R80100d<B`0>C4_03Ta;`0^9BH201dI6P0401LG6@0T9BT092DY01@G708061TN00@0
7QlS01`K800L6b008R8T0P0e<c@01017ADH08alP02<O800X:2X202TY:`0401dM7`1;Bdd0Bd]=030`
<P807QhP00@071`N02h^<00^;S00=3@f0P0Y:B/0100Z:R`0;2`^02`/;P0g=cT203/k?@0402h^<00e
=CL0=CDg02DS9P80:BLZ00@0H5iQ030^<@0`;S40:RP[0P0/:Rd0100Y9bX0;bd`02l]<018ATT205AB
E@0403Xh>`0a;c80<Blb02/Y;080@Cm200@0<Blb02h/;`0^;2l0:2HW0P0Z:2T0101ACe00;2X[02`Z
:`0a;c02024O800402HT9@0c<C80<c4b04Q6A`809bDV00@0:bTZ02TW:00Y9bP0:bTZ0P0`;Rl01015
@d@09B<T02DS900[:BX202h/;@0402l];P19AdP0BDM803Dc=0808b4R00@0;bd^02<Q8P0S8B80928S
0P1>C4d0100^;2d0;B//02d[;00Z:2T203Hd=@0402LU9P1DDU<0E59C034_<080<2h_00@0=C<d03Le
=P0g=CH0;bd^0P1FE5D0100f=3D0:RPY02XX:@0^;2d202`Z:`0402l];P0l>S/0?3Xk02/Y:P80;R`]
00@0<Bl`02TW:00Y9bP0<c4b0P0d<S<0100b<340BTQ904Y8B@0[:BX302h/;@80AdE600@0:bTZ034_
<00a;c00;R`]0P0e<c@0100n?3d0De5B05=ADP0f=3D203<a<P0402l];P0Z:2T0:RPY02l];P80?cdn
00@0;R`]02/Y:P0[:BX0>3Hg0P1CDE80100d<S<0=cDf03Le=P0i=cP2034_<00403<a<P1DDU<0E59C
03<a<P80;2X[00@0;R`]03dk?00m>c`0=S@e0P0b<340100l>S/0Cde>04m=CP1BD54203@b<`0402h/
;@0/:R/0;2X[059@D@80@3ho00@0>CLh03Hd=@0f=3D0;2X[0P0W9BH0100d<S<0@d5204=1@P0c<C83
02XX:@80;B//00@0?C/l044o@011?d009bDV0P0a;c00100k>CX0?3Xk03`j>`0k>CX203@b<`0402d[
;00d<S<0=38c05UGF080=C<d00@0:RPY02HT9@0V92D0:RPY0P0g=CH0100k>CX0AD=404E3A01=Bd`2
02h/;@0403<a<P0]:b`0;B//02l];P80BTQ900@0>CLh01HD5@0F51D0=38c0P0]:b`0100U8b@0De5B
05=ADP0g=CH202l];P0402d[;00b<340<S0a03Le=P80=C<d0@0_;Bh101LE5P40oooo000@04ECGPH0
AUAO0P17EF001017EEP0;S`o02hl?`0?6A/200492`0400X>3@0@40h0410>00l>2P80510=00@03`X6
010;1`0@2`L040`90P0:1P<0100=30X09BDU02DU9@0g=cL2038b<P0403Hf=P0Y:BT0:BTY02<S8`80
;Bd]00@0=3@d030`<00`<300:RXZ0P0X:2P0100H61P071`L01`L700K6a/202TY:@0401`L700U9BD0
9BDU03dm?@806ATI00@0:2PZ01/K7@0K6ad092@V0P0Y:B/0100a<C<07alQ01lO8@0S8bD202HV:003
01`L7P0n?T00?Si000<0<30b00@08b<U02LW:@0W9bT09RHX0P18B4X0100U9BL0:RX/02XZ;00V9RP2
024Q8`0402XZ;010@480@41202<S9@809bLY00@08R8T01/K7@0K6ad08R8T0P0i>C/302LW:@0401lO
8@0R8R@08R8T03Lg=`80=3@d00@07AdM020P800P82009RHV0P0Y:BT0100b<S80820P020P800T92@2
028R8P04030`<00Z:RX0:RXZ02<S8`807QhN00@04Q8B01lO7`0O7al0?Cdo0P0^;S00100R8R@07alQ
01lO8@0X:2X203Hf>00402h^<00X:2X0:2PZ02h^<080=SHh00@0<c<e04Q8BP18B4X0=3@f0P0Y:B/0
100X:2X0;Bd_02d];`0P828203/k?@04028R900L71h071`N01/K7@8092@V00@0=CDg020P8P0P8280
9BDW0P0Z:R`0100[:bd09BDW02DU9`0V9RP203hn@00402HV:00L71h071`N024Q8`807alQ00@0>3Pj
02d];`0];Bl000020P0V9RP302@T9P0401`L7P0R8R@08R8T01hN80805QHH00@05ADG00X:300:2P`0
61PJ0P0D51H0100D5QD02`d<00/=300^<2l2038d<`0403<e=00e=cH0=CLf07eoOP80/[Bc00@0X::O
09^MVP2KWIX0VYnK0P2OY:00102OYJ40SiRC08nHT`2FWiX208J?RP0402LZ8@0e=Rh0=CH^03`i=080
=Bd[00@0;bHW02HR8`0V8R<08R8T0P0[;300100>4AH03a0E00l@5@0M7R8201TH7@0402@T9P0d<S<0
=38c03Tg>0808QhO00@0;Bd_02XZ;00Z:R`071`N0P1ADE<0100/;2h04Q8D018B500K6ad202h^<004
03@d=P0e=CL0=CDg02/[;@80:2PZ00@0<S8d03Xj?00j>S`0>SXl0P0^;S00100X9RT0=S@g03Hd=`0h
=ST2051>D@0401dK7P0^;2l0;R`_04A2A@80<Blb00@0;R`_03hl?`0n?3l0CTa?0P0Y9bX0100`;S40
81hQ020N8@0i=cX2040n@@0402`Z:`0Z:2T0:RPY04A2@`80;R`]00@0:BLX028P8@0R8240:RPY0P0c
<C80101:B4T0;bd^02l];P0Y9bP2030^;`0402TW:00]:b`0;B//051>C`809R@U0`0Y9bP0100^;2d0
AD=404E3A00m>c`202PV9`0403Hd=@0V92D09R@U02l];P80CTa=00@0@Cm003Dc=00e<c@0<c4b0P0o
?Ch0100f=3D0F5IG05QFE`14@T<203Hd=@0403Xh>@0k>CX0>cTj030^;`80CTa=00@0?S`m02/Y:P0[
:BX0;B//0P0^;2d0100]:b`0?S`m03hl?@0b<342034_<00403<a<P0]:b`0;B//03@b<`80<S0a00@0
:RPY04i<C@1>C4d0;bd^0P0`;Rl0100Z:2T0BdU:04]9BP0V92D202d[;00402DS900c<C80<c4b030^
;`80Cde>00@0>cTj030^;`0`;Rl0=38c0P0^;2d0100/:R/0AdE604M5AP0/:R/202@R8`0402PV9`1?
CDh0Cde>02h/;@80<Bl`00@0<c4b02TW:00Y9bP0:2HW0P19AdP0100e<c@08b4R02<Q8P0]:b`2044o
@004034_<00_;Bh0;bd^03<a<P80@3ho00@0Cde>02d[;00]:b`0:RPY0P0]:b`0101<BT/0@d5204=1
@P0f=3D203Le=P0402l];P0T8R<0928S02XX:@80@Cm000@0=S@e02LU9P0W9BH0;2X[0P0]:b`00`0g
=CH0C4Y;04a:B`0302/Y:P0403@b<`0a;c00<Bl`02d[;080:BLX00@08b4R02d[;00]:b`0D4i?0P0c
<C80100V92D09bDV02LU9P0Z:2T502h/;@0303dk?00k>CX0>cTj00<0;bd^00@09bDV044o@011?d00
?cdn0P0Z:2T0100d<S<0928S02@R8`0f=3D203Xh>@0404i<C@0c<C80<c4b03Pf=`80:bTZ0`0Y9bP1
02h/;@40928S0@3oool00100A59M0P15Deh304IDG`<0AeEP00@0AeAJ03M5B00gADP05b4S0P022P`0
100:3Pl04A4A014A4@0>3@/2014=30<040/700@03`/800X60`0:1P<02PT70P0Q8B40100e=CD0<c<c
03<c<`0a<C4202<S8`0401hN7P0Y:BT0:BTY03Lg=`80>3Ph00@0;bl_02LW9`0W9bL07alO0P0M7Ad0
100S8b<0;Rh^02h^;P0O7al202DU9@040492@P0T92@092@T01lO8@805aLI00@08B4S02TY:`0Y:B/0
>CTk0`0Z:R`202/[;@0402d];`0b<S@0<S8d04=3A@80:2PZ00@092@V01XJ700J6Q`0:2PZ0P0W9bT0
100c<cD092@V02@T9P0P828201hN800401`L7P0o?d40?cm1034a<`807QhP00@09BDW024Q8`0Q8B<0
9bLY0P0n?T00100`<380:2PZ02PX:P0T92H202PX:P0302LW9`16ATH0ATI600@08R8R0P0N7Qh0100m
?Cd0820P020P800S8b<201`L700402/[:`0Y:BT0:BTY02<S8`80820P00@04A4A01PH600H61P0<S8d
0P0f=SP0100X:2X071`N01`L7P0U9BL203<c=@0403Lg>@0T92H092@V02DU9`80:BT[00@08R8T030`
<P0`<380;Rh`0P0N7R0301PH6P0401`L7P0d=3H0=3@f02HV:0806a/M00@061PJ01/K7@0K6ad0=SHh
0P0N7R00100W9bT0:b/]02/[;@0/;2h202PX:P0402TY:`0j>S`0>SXl02d];`807QhP00@0820R02TY
:`0Y:B/0<c<e0P0D51H0100G5aT08R8T028R900J6Q`202/[;@0401PH6P0M7Al07AdO01`L7P805aLI
00@06ATK00h>400>3Q0061PJ0P0@4180100J71/03Q0?00h@3`0/;Rd203<e=0040494@`1DEUD0E5IE
09BFU@80YJNV00@0WZ2K0:NYY02WZJ@0WZ>M0P2GW9H0102QZ:40TI^C096KT`2:U8`207EoM`0402DX
7`0m?SH0?Chf03`i=080<2PV00@0=2//03Lc=00g<c@0<c<e0P0`<CD0100L7b@04a@I01<D6@0I6Qh2
014@5@0401XJ700Q7b008AlP024O808071PI0`0X:2X0100O7b40DU9D059BE00W9bT201@D5P0402DU
9`0`<380<30b034a<`80A4A600@0;Rh`02XZ;00Z:R`0;Bd_0P0c<cD00`14A4H09RHX02HV:003038`
<`0402DS9P1>C4l0CTa?03Dc=P80>CLj00@0>cTl034_<P0a;c809B<V0P0]:bh0101:B4/0=cDh03Le
>00_;C0201/I700403Pf>@11?d80@Cm203Dc=080;bd^00<0?3Xk03Xh>@0j>3T00`0^;2d0100a;c00
<2h_030^;`19AdP203Tg>00402/Y:P0f=3D0=S@e02/Y:P80;R`]00@0EE=D02`Z:`0/:R/0<Bl`0P0`
;Rl0100b<340@T110490@@1:B4T2030^;`04040n?`0a;c00<Bl`03Le=P80BdU:00@0D4i?03Dc=00e
<c@0<S0a0P0i=cP0100b<340AdE604M5AP19AdP202l];P04038`<@0c<C80<c4b02HT9@80?S`m00@0
@T1102TW:00Y9bP0:bTZ0P0_;Bh0100/:R/0@T110490@@0j>3T303@b<`80;R`]00@0=C<d03@b<`0d
<S<0;B//0P1>C4d0100g=CH0=38c03@b<`0[:BX204Y8B@<0=38c00@0:bTZ03Xh>@0j>3T0<Bl`0P1=
Bd`01018ATL0<S0a038`<@0^;2d302/Y:P80BDM800@0<Bl`02HT9@0V92D0:2HW0P1DDU<0100e<c@0
=cDf03Le=P0h=SL2030^;`0402`Z:`18ATL0B4I70490@@80:BLX00@0;B//0490@@12@440:bTZ0P0Y
9bP0100]:b`0=38c03@b<`1:B4T2034_<00402LU9P0[:BX0:bTZ0490@@80ATA500@0<Bl`03Dc=00e
<c@0<2h_0P0S8B80100c<C80A49304A2@`0m>c`202`Z:`@0;bd^0P17ADH0100c<C80:BLX02TW:00c
<C82034_<00402l];P0a;c00<Bl`02`Z:`80=S@e00@0De5B04A2@`14@T<0<2h_0P0_;Bh0100a;c00
:bTZ02/Y:P0W9BH2028P8@0403@b<`1@CTl0D4i?02l];P80=38c0`0^;2d0100l>S/0>SPi03Xh>@0[
:BX202HT9@03044o@00X9RL0:2HW00<0BdU:00@0=S@e02TW:00Y9bP0<c4b0P0W9BH102XX:@40:2HW
0@3oool00100A59M0P15Deh304IDG`<0AeEP00@0AeAJ03M5B00gADP05b4S0P022P`0100:3Pl04A4A
014A4@0>3@/2014=30<040/700@03`/800X60`0:1P<02PT70P0Q8B40100e=CD0<c<c03<c<`0a<C42
02<S8`0401hN7P0Y:BT0:BTY03Lg=`80>3Ph00@0;bl_02LW9`0W9bL07alO0P0M7Ad0100S8b<0;Rh^
02h^;P0O7al202DU9@040492@P0T92@092@T01lO8@805aLI00@08B4S02TY:`0Y:B/0>CTk0`0Z:R`2
02/[;@0402d];`0b<S@0<S8d04=3A@80:2PZ00@092@V01XJ700J6Q`0:2PZ0P0W9bT0100c<cD092@V
02@T9P0P828201hN800401`L7P0o?d40?cm1034a<`807QhP00@09BDW024Q8`0Q8B<09bLY0P0n?T00
100`<380:2PZ02PX:P0T92H202PX:P0302LW9`16ATH0ATI600@08R8R0P0N7Qh0100m?Cd0820P020P
800S8b<201`L700402/[:`0Y:BT0:BTY02<S8`80820P00@04A4A01PH600H61P0<S8d0P0f=SP0100X
:2X071`N01`L7P0U9BL203<c=@0403Lg>@0T92H092@V02DU9`80:BT[00@08R8T030`<P0`<380;Rh`
0P0N7R0301PH6P0401`L7P0d=3H0=3@f02HV:0806a/M00@061PJ01/K7@0K6ad0=SHh0P0N7R00100W
9bT0:b/]02/[;@0/;2h202PX:P0402TY:`0j>S`0>SXl02d];`807QhP00@0820R02TY:`0Y:B/0<c<e
0P0D51H0100G5aT08R8T028R900J6Q`202/[;@0401PH6P0M7Al07AdO01`L7P805aLI00@06ATK00h>
400>3Q0061PJ0P0@4180100J71/03Q0?00h@3`0/;Rd203<e=0040494@`1DEUD0E5IE09BFU@80YJNV
00@0WZ2K0:NYY02WZJ@0WZ>M0P2GW9H0102QZ:40TI^C096KT`2:U8`207EoM`0402DX7`0m?SH0?Chf
03`i=080<2PV00@0=2//03Lc=00g<c@0<c<e0P0`<CD0100L7b@04a@I01<D6@0I6Qh2014@5@0401XJ
700Q7b008AlP024O808071PI0`0X:2X0100O7b40DU9D059BE00W9bT201@D5P0402DU9`0`<380<30b
034a<`80A4A600@0;Rh`02XZ;00Z:R`0;Bd_0P0c<cD00`14A4H09RHX02HV:003038`<`0402DS9P1>
C4l0CTa?03Dc=P80>CLj00@0>cTl034_<P0a;c809B<V0P0]:bh0101:B4/0=cDh03Le>00_;C0201/I
700403Pf>@11?d80@Cm203Dc=080;bd^00<0?3Xk03Xh>@0j>3T00`0^;2d0100a;c00<2h_030^;`19
AdP203Tg>00402/Y:P0f=3D0=S@e02/Y:P80;R`]00@0EE=D02`Z:`0/:R/0<Bl`0P0`;Rl0100b<340
@T110490@@1:B4T2030^;`04040n?`0a;c00<Bl`03Le=P80BdU:00@0D4i?03Dc=00e<c@0<S0a0P0i
=cP0100b<340AdE604M5AP19AdP202l];P04038`<@0c<C80<c4b02HT9@80?S`m00@0@T1102TW:00Y
9bP0:bTZ0P0_;Bh0100/:R/0@T110490@@0j>3T303@b<`80;R`]00@0=C<d03@b<`0d<S<0;B//0P1>
C4d0100g=CH0=38c03@b<`0[:BX204Y8B@<0=38c00@0:bTZ03Xh>@0j>3T0<Bl`0P1=Bd`01018ATL0
<S0a038`<@0^;2d302/Y:P80BDM800@0<Bl`02HT9@0V92D0:2HW0P1DDU<0100e<c@0=cDf03Le=P0h
=SL2030^;`0402`Z:`18ATL0B4I70490@@80:BLX00@0;B//0490@@12@440:bTZ0P0Y9bP0100]:b`0
=38c03@b<`1:B4T2034_<00402LU9P0[:BX0:bTZ0490@@80ATA500@0<Bl`03Dc=00e<c@0<2h_0P0S
8B80100c<C80A49304A2@`0m>c`202`Z:`@0;bd^0P17ADH0100c<C80:BLX02TW:00c<C82034_<004
02l];P0a;c00<Bl`02`Z:`80=S@e00@0De5B04A2@`14@T<0<2h_0P0_;Bh0100a;c00:bTZ02/Y:P0W
9BH2028P8@0403@b<`1@CTl0D4i?02l];P80=38c0`0^;2d0100l>S/0>SPi03Xh>@0[:BX202HT9@03
044o@00X9RL0:2HW00<0BdU:00@0=S@e02TW:00Y9bP0<c4b0P0W9BH102XX:@40:2HW0@3oool000`0
A59M0`13DE`304ABG@0304ECGP16E5l0AUAO00<0AeEP00@0AU=I045>E011CU@08R/`0P053A001009
3A00410@010@400<2P/200h:2@0401<>2P0?2PH03`X600h:1`8030P500@02`X801`L700L71`0;bl_
0P0b<S800`0a<C40;2`/02`/;00302LW9`0402`/;00_;bl0;bl_02d];@80:BTY00@09RHV01`L700L
71`0;bl_0P0/;2`0100P82009RHV02HV9P0h>3P202h^;P04028R900Q8B<08B4S024Q8`80:b/]00@0
=cLi038b=00b<S@09bLY0P0T92H0100/;2h06a/M01/K7@13@dD2024Q8`04028R900F5QP05QHH02XZ
;0804a<E00@0@41202HV:00V9RP08B4S0P0M7Al0100I6A/0<c<e03<c=@0k>cd201LG6@0402TY:`0X
:2X0:2PZ02d];`80=cLi00@0>c/m02LW:@0W9bT09bLY0P0X:2X0100N7Qh0ADE504E5A@0[:b/2028R
8P0401hN7P0L71`071`L03Ti>@809BDU00@08B4Q01TI6@0I6AT08b<S0P0X:2P0100U9BD0820P020P
800G5aL201@D500402@T9P0i>C/0>CTk02`/;P8071`N00@08R8T02XZ;00Z:R`0?3`n0P0R8R@0100O
7b408B4S024Q8`0J6Q`202LW:@0403Ti>`0V9RP09RHX024Q8`807alQ00@09BDW03<c=@0c<cD0=cLi
0P0V9RP0100P82807QhP01hN800l?3h302HV:0<0:RX/0P0Y:B/0100X:2X0<S8d038b=00a<C<2024Q
8`0402/[;@0a<C<0<C4c02<S9@800@4300@0@T9400h>400>3Q00<S8d0P0X:2X0100A4A<06ATK01TI
6`0N7R02020P8P04024Q8`0@4180410B01XJ70805ADG00@0>S`i02h`;@0^<2d0=3Hc0P0g>CH01017
BDH0KFm/06e_K02LWY/2092BS`0408b>R@2PXYd0X::M09^PVP80UibF00@0[;>/0:>]Y@2S[JD0TYbD
0P1_NG40100W:R40>S/c03Xk<`0e<Rd2030X9P0403/b<`11?Ch0@Cdn03dm?`80<S<g00@09RT^01dN
8`0M7R<05QLK0P0=3140100H61X05aDF01LE5P0I5aP201dI6P0402HV:00[:bd0:b/]02`/;P80CTi@
0`0M7Al0100[:bd0:2PZ02PX:P0b<S@204E5A`0403De=`0a<C<0<C4c02TY:`80;Bd_00@0@T9402XZ
;00Z:R`0<S0c0P0N71l0100k>C`0;B/^02d[;P1KFE`204E3AP0401PF6@0[:B`0:bT/034_<P80<S0c
00@0;B/^04]9C01;BD`0;B/^0P0k>C`0100/:Rd0<Blb034_<P0h=SL203<a<P0403`j>`1>C4d0CTa=
03Pf=`80>SPi00@0=S@e02XX:@0Z:2T0BDM80P18ATL0100f=3D0?C/l03dk?00f=3D203Pf=`0405UG
F00m>c`0?C/l03<a<P80=38c00@0<2h_03Hd=@0f=3D0CTa=0P0/:R/0100h=SL0;B//02d[;00Z:2T2
03Le=P0404m=CP0[:BX0:bTZ02TW:080;bd^00@0:bTZ03@b<`0d<S<0CD]<0P0X9RL0100]:b`0<Bl`
034_<00U8b@203Le=P0404m=CP0`;Rl0<2h_02TW:080<2h_00@0:BLX0490@@12@440?C/l0`0c<C82
02d[;00402/Y:P0^;2d0;R`]02d[;080?cdn00@0=cDf02d[;00]:b`09R@U0P0i=cP0100k>CX0;R`]
02h/;@0X9RL2038`<@0402`Z:`0i=cP0>CLh04M5AP809R@U00@0;bd^02d[;00]:b`0<2h_0P14@T<0
100j>3T0:RPY02XX:@0Y9bP204Q6A`0403@b<`0`;Rl0<2h_034_<080;B//00@09R@U03`j>`0l>S/0
@T110P0U8b@0100]:b`0@Cm0044o@00Z:2T202TW:004038`<@0f=3D0=S@e04Y8B@80A49300@0:BLX
02XX:@0Z:2T0=cDf0P1:B4T0100^;2d0<S0a038`<@0a;c0202DS900402d[;00c<C80<c4b03Xh>@80
;2X[00@0:RPY02d[;00]:b`0:BLX0P0j>3T00`0o?Ch0:bTZ02/Y:P0303Dc=00403<a<P0d<S<0=38c
02PV9`80;R`]00@0AD=404a:B`1<BT/0<S0a0P0Z:2T00`0c<C80;R`]02h/;@0302l];P0302/Y:P1=
Bd`0CD]<00<0<c4b00@0>cTj02XX:@0Z:2T0?S`m0P17ADH0100Y9bP0<Bl`034_<00g=CH302l];P80
GeeN00@0;bd^02`Z:`0/:R/0:BLX0P0L6Q/102@R8`40;2X[0@3oool000`0@e5L0`12D5/01013DE`0
A59M04ABG@15Deh204IDG`<0AeEP00@0AE9J04QEF`18EE/0:c@i0P0841<010093A003`lA00l?4@09
1`P200/7200401LB3P0>2@D03PT500d91P8040`900@03@`:01LG5`0G5aL0:RXZ0P0b<S80100_;bl0
=CDe03De=@0c<c<202XZ:P0402DU9@0X:2P0:2PX02PX:0808b<S00@06ATI00d=3@0=3@d0<S8b0P0P
8200100O7al0;Bd]02d];@0^;Rh20492@P0402PX:P0_;c40;bla020P8P<0;2`^0P0f=SP0100O7b40
71`N01`L7P0Q8B<200l?4@0403Hf>00[:bd0:b/]024Q8`80820R00@09RHX01XJ700J6Q`0=SHh0P0X
:2X0100S8bD071`N01`L7P0U9BL202/[;@0404M7B@0T92H092@V02h^<080:b/]00@0;Bd_02HV:00V
9RP0ADE70P0U9BL0100X:2X0820R020P8P0Q8B42034a<@0403De=@0O7al07alO01`L70<09BDU0P0^
;Rh0100P82006QXJ01XJ6P0L71`202XZ:P0402PX:00Q8B408B4Q020P808051@D00@061PJ03Hf>00f
=SP0;Bd_0P0P8280100Q8B<07AdO01dM7`0k>cd202DU9`0402XZ;00];Bl0;Bd_02DU9`80:b/]00@0
A4A602/[;@0[:bd09BDW0P0T92H0100S8bD09RHX02HV:00j>S`202DU9`0401dM7`0F5QP05QHH038b
=0808R8T00@08b<U02PX:P0X:2X09bLY0P0V9RP0100S8bD0:2PZ02PX:P0a<C<2024Q8`0402<S9@0J
6Q`06QXL00T92`8051@F00@0>CTk018B500B4Q@0:b/]0P0L71h0100C4aD071`N01`L7P0T92H202DU
9`04020P8P0<30h030`>020P8P80:b/]00@0>cdj03Tk>00i>cP0;S0]0P0d=S<0100k?CX0LgEb07=e
LP2KWIX208^=RP0409RKU02OXY/0Wj:K09^PV@80U9VB00@0WJBL09jXW`2NZ9l0WjVP0P1jQ7/0100e
>2l0=SL_03Hg;`0Z9b82034Y9`0403/b<`0g<c@0=c<d03Lg>@80<34e00@0:2/`02l`=@0_<3D09bP/
0P0Q82D0100];Bl08R0Q028P8@0O7Ah201TE5P0401dM7`0Z:R`0:RX/03Hf>080ADE700@04Q8D02XZ
;00Z:R`0<C4c0P0O7b40100h>3X0=3@f03@d=P11@D<203Ph>P0402PX:P0];Bl0;Bd_038b=080>SXl
00<09R@W038`<`0b<3<00`0^;2l0101HEUT0=C<f03Dc=P0a;c8202<Q900402l]<00d<SD0=38e020N
8@80@Cm200@0>CLj04i<C`1>C4l0:bT/0P0X9RT0100e<c@0<c4b03<a<P0h=SL205IDE@04038`<@0e
<c@0=C<d03<a<P809R@U00@0;2X[03`j>`0l>S/0:2HW0P0W9BH0100V92D0:BLX02TW:00m>c`203<a
<P0402`Z:`0a;c00<Bl`02TW:080:2HW00@0Cde>02DS900U8b@0:bTZ0P0T8R<0100Y9bP0<2h_030^
;`1IEeP202h/;@0302l];P0e<c@0=C<d00<0<c4b00@0De5B02LU9P0W9BH0;bd^0P0f=3D0100Y9bP0
=C<d03Dc=01IEeP203<a<P0402PV9`0_;Bh0;bd^028P8@80?C/l00@0>SPi02h/;@0^;2d0<2h_0P0[
:BX0100]:b`0=38c03@b<`0m>c`203Xh>@040490@@0a;c00<Bl`02h/;@80<S0a00@0@3ho02DS900U
8b@0:2HW0P0^;2d0100`;Rl0;R`]02h/;@1?CDh202LU9P0403Tg>00c<C80<c4b03Dc=080>3Hg00@0
A49302h/;@0^;2d0:2HW0P0_;Bh0100l>S/0<S0a038`<@0c<C8203@b<`0402`Z:`0j>3T0>SPi04Y8
B@80:bTZ00@0;B//03lm?P0o?Ch0:BLX0P0U8b@0100b<340;R`]02h/;@0g=CH204=1@P0402d[;00[
:BX0:bTZ030^;`80Cde>00@0;B//030^;`0`;Rl0<S0a0P0Z:2T0100X9RL09B<T02DS900m>c`203Dc
=00402LU9P0]:b`0;B//02XX:@80=38c00@0BdU:02d[;00]:b`0=38c0P0b<340100_;Bh0<Bl`034_
<00P7Ql202<Q8P0403Le=P1FE5D0EUAE03Le=P809bDV00@0<c4b02`Z:`0/:R/0;R`]0P0c<C80100U
8b@0>3Hg03Pf=`17ADH2034_<00403Tg>00^;2d0;R`]03hl?@80CD]<00@0>3Hg030^;`0`;Rl0;R`]
0P0e<c@0100R8240BdU:04]9BP0i=cP202`Z:`0503<a<P0U8b@09B<T02TW:00[:BX00@3oool000`0
@e5L0`12D5/01013DE`0A59M04ABG@15Deh204IDG`<0AeEP00@0AE9J04QEF`18EE/0:c@i0P0841<0
10093A003`lA00l?4@091`P200/7200401LB3P0>2@D03PT500d91P8040`900@03@`:01LG5`0G5aL0
:RXZ0P0b<S80100_;bl0=CDe03De=@0c<c<202XZ:P0402DU9@0X:2P0:2PX02PX:0808b<S00@06ATI
00d=3@0=3@d0<S8b0P0P8200100O7al0;Bd]02d];@0^;Rh20492@P0402PX:P0_;c40;bla020P8P<0
;2`^0P0f=SP0100O7b4071`N01`L7P0Q8B<200l?4@0403Hf>00[:bd0:b/]024Q8`80820R00@09RHX
01XJ700J6Q`0=SHh0P0X:2X0100S8bD071`N01`L7P0U9BL202/[;@0404M7B@0T92H092@V02h^<080
:b/]00@0;Bd_02HV:00V9RP0ADE70P0U9BL0100X:2X0820R020P8P0Q8B42034a<@0403De=@0O7al0
7alO01`L70<09BDU0P0^;Rh0100P82006QXJ01XJ6P0L71`202XZ:P0402PX:00Q8B408B4Q020P8080
51@D00@061PJ03Hf>00f=SP0;Bd_0P0P8280100Q8B<07AdO01dM7`0k>cd202DU9`0402XZ;00];Bl0
;Bd_02DU9`80:b/]00@0A4A602/[;@0[:bd09BDW0P0T92H0100S8bD09RHX02HV:00j>S`202DU9`04
01dM7`0F5QP05QHH038b=0808R8T00@08b<U02PX:P0X:2X09bLY0P0V9RP0100S8bD0:2PZ02PX:P0a
<C<2024Q8`0402<S9@0J6Q`06QXL00T92`8051@F00@0>CTk018B500B4Q@0:b/]0P0L71h0100C4aD0
71`N01`L7P0T92H202DU9`04020P8P0<30h030`>020P8P80:b/]00@0>cdj03Tk>00i>cP0;S0]0P0d
=S<0100k?CX0LgEb07=eLP2KWIX208^=RP0409RKU02OXY/0Wj:K09^PV@80U9VB00@0WJBL09jXW`2N
Z9l0WjVP0P1jQ7/0100e>2l0=SL_03Hg;`0Z9b82034Y9`0403/b<`0g<c@0=c<d03Lg>@80<34e00@0
:2/`02l`=@0_<3D09bP/0P0Q82D0100];Bl08R0Q028P8@0O7Ah201TE5P0401dM7`0Z:R`0:RX/03Hf
>080ADE700@04Q8D02XZ;00Z:R`0<C4c0P0O7b40100h>3X0=3@f03@d=P11@D<203Ph>P0402PX:P0]
;Bl0;Bd_038b=080>SXl00<09R@W038`<`0b<3<00`0^;2l0101HEUT0=C<f03Dc=P0a;c8202<Q9004
02l]<00d<SD0=38e020N8@80@Cm200@0>CLj04i<C`1>C4l0:bT/0P0X9RT0100e<c@0<c4b03<a<P0h
=SL205IDE@04038`<@0e<c@0=C<d03<a<P809R@U00@0;2X[03`j>`0l>S/0:2HW0P0W9BH0100V92D0
:BLX02TW:00m>c`203<a<P0402`Z:`0a;c00<Bl`02TW:080:2HW00@0Cde>02DS900U8b@0:bTZ0P0T
8R<0100Y9bP0<2h_030^;`1IEeP202h/;@0302l];P0e<c@0=C<d00<0<c4b00@0De5B02LU9P0W9BH0
;bd^0P0f=3D0100Y9bP0=C<d03Dc=01IEeP203<a<P0402PV9`0_;Bh0;bd^028P8@80?C/l00@0>SPi
02h/;@0^;2d0<2h_0P0[:BX0100]:b`0=38c03@b<`0m>c`203Xh>@040490@@0a;c00<Bl`02h/;@80
<S0a00@0@3ho02DS900U8b@0:2HW0P0^;2d0100`;Rl0;R`]02h/;@1?CDh202LU9P0403Tg>00c<C80
<c4b03Dc=080>3Hg00@0A49302h/;@0^;2d0:2HW0P0_;Bh0100l>S/0<S0a038`<@0c<C8203@b<`04
02`Z:`0j>3T0>SPi04Y8B@80:bTZ00@0;B//03lm?P0o?Ch0:BLX0P0U8b@0100b<340;R`]02h/;@0g
=CH204=1@P0402d[;00[:BX0:bTZ030^;`80Cde>00@0;B//030^;`0`;Rl0<S0a0P0Z:2T0100X9RL0
9B<T02DS900m>c`203Dc=00402LU9P0]:b`0;B//02XX:@80=38c00@0BdU:02d[;00]:b`0=38c0P0b
<340100_;Bh0<Bl`034_<00P7Ql202<Q8P0403Le=P1FE5D0EUAE03Le=P809bDV00@0<c4b02`Z:`0/
:R/0;R`]0P0c<C80100U8b@0>3Hg03Pf=`17ADH2034_<00403Tg>00^;2d0;R`]03hl?@80CD]<00@0
>3Hg030^;`0`;Rl0;R`]0P0e<c@0100R8240BdU:04]9BP0i=cP202`Z:`0503<a<P0U8b@09B<T02TW
:00[:BX00@3oool000`0@e5L00@0@DmJ049@F`12D5/0@e5L0P14DUd00`15Deh0AUAO04IDG`0304ME
H00404=@F01<FF40C5UQ034j@@802Q4G00@02@d@00h>400>3Q001`D60P091@H0100I510030L300`7
0`0<20D2018>2`0400h=2`0D51@051@D02HV9P80<S8b00@08b<S030`<00`<300=CDe0P0/;2`0100Z
:RX0;bl_02l_;`0[:b/2020P800401dM7@0>3Ph03Ph>03dm?@807AdM00@0820P030`<00`<3008b<S
0P19BDT0100T92H0<30b030`<P0I6A/202PX:P0402DU9`0m?Cl0?Cdo028R9080820R00@08R8T01LG
6@0G5aT0<C4c0P0l?3h0100Q8B<0:RX/02XZ;00J6Q`202<S9@0403De=`0d=3H0=3@f02`/;P807AdO
00@0<30b020P8P0P8280A4A60P0W9bT0100`<380;2`^02`/;P0Z:R`201PH6P0404Y:C00S8bD08b<U
02LW:@805aLI00@0:2PX01hN7P0N7Qh0?3`l0`0L71`2030`<004018B4P0e=CD0=CDe020P808071`L
00@061PH02/[:`0[:b/0:b/[0P0R8R80100X:2P05ADE01DE5@0B4Q@2038b=00402/[;@0T92H092@V
028R908051@F00@0>3Pj02PX:P0X:2X071`N0P0S8bD0100N7R008R8T028R900o?d4202<S9@04020P
8P0S8bD08b<U02TY:`8092@V00@0@d=502/[;@0[:bd08B4S0P0I6A/0100a<C<09RHX02HV:00P8283
02DU9`808b<U00@07QhP024Q8`0Q8B<0<30b0`0P828200P82P0400H6201:BT`0BTY<01hN8080>c/m
00@05aLI02<S9@0S8bD06QXL0P0R8R@0100W9bT092@V02@T9P0I6A/200H6200402PX:P12@T@0@T94
02/]:080>S`g00@0<cD`04A6@@14AT40A4I10P24QX40102VZ:<0W9jI09bNV@2MX9T209jQVP040:R]
YP2UZZ<0YJZS09ZQV@80VjFL00@0[;J]08RBR@28TXT0AdY10P0e=Rh0100S81/0<2PV030X9P0c:R/2
02HR8`0402/[;@0^;c<0;Rlc02ha=P80B4U>00@0@dA803hm@P0n?D80AdM90P0`;Rl0100Q7b002@D6
00T51P0<30h201lO8@0403Hf>00l?3h0?3`n00d=3`80>CTk00@0?Si002<S9@0S8bD0?Cdo0P0Q8B<0
101:BT`0?3`n03`l?P0X:2X2030`<P04028R9019BD/0BDU;038`<`80928U00@0>SPk03Le>00g=CP0
7adP0P1DDUD0100f=3L0=cDh03Le>00N71l202/Y;00402TW:P0T8RD0928U04a:C@80AD=600@0=38e
02d[;P0]:bh0<Bl`0`0`;Rl2051>C`04028P8@0Y9bP0:BLX02h/;@809B<T00@0;2X[04I4A@16A4D0
<Bl`0P0[:BX0100_;Bh0<S0a038`<@0m>c`2040n?`0402PV9`0b<340<S0a02TW:0809B<T00@0EUAE
02PV9`0X9RL0:bTZ0P0V92D0100[:BX0;R`]02h/;@1NG5d202d[;00402`Z:`0^;2d0;R`]02/Y:P80
928S00@0De5B02<Q8P0S8B80;B//0P0d<S<0100V92D0;2X[02`Z:`1DDU<202TW:00402PV9`0^;2d0
;R`]01dK7080=cDf00@0=S@e02TW:00Y9bP0;B//0P0[:BX0100]:b`0>3Hg03Pf=`18ATL203Hd=@04
04Q6A`0c<C80<c4b03<a<P80;B//00@0DDm@02l];P0_;Bh0>CLh0P0n?3d01019AdP0?C/l03dk?01/
JV/2044o@00403`j>`0a;c00<Bl`03<a<P80;R`]00@0BdU:03Dc=00e<c@0:bTZ0P0P7Ql01015@d@0
=cDf03Le=P0h=SL203dk?00403<a<P0m>c`0?C/l059@D@80<c4b00@0;2X[03hl?@0n?3d0;R`]0P0]
:b`01010?Sl0>cTj03/i>P0m>c`205MEEP04038`<@0[:BX0:bTZ02d[;080De5B00@0;R`]030^;`0`
;Rl0=38c0P0_;Bh0100h=SL0;bd^02l];P1ACe0204U7B00402h/;@0a;c00<Bl`02l];P80=38c00@0
DDm@02`Z:`0/:R/0;bd^0`0]:b`203Hd=@0402PV9`0/:R/0;2X[02d[;080GUaM00@0@3ho02`Z:`0/
:R/0>cTj0`0b<34203Pf=`0402XX:@0]:b`0;B//06ARH`80<c4b00@0;R`]02/Y:P0[:BX0=38c0P16
A4D0101<BT/0928S02@R8`0b<34202l];P0402XX:@0U8b@09B<T04a:B`80:BLX00D0=C<d02XX:@0Z
:2T0;B//02TW:0010?ooo`0000@0AE=N04ABG@14DUd0@e5L0P12D5/01013DE`0A59M04ABG@16E5l2
04MEH00304ABG@15Deh0AE=N00<0AE=N0`16E5l304MEH00404QEGP17E5d0AeAM0419D08092T]00@0
1@T<00P8200820P04A0>0P0;1`@0100F5A<03@`:00d<2P0?3P`200P71@0400d<2P0920H02@P602HU
8`80?chl00@0:RXX02<S8@0S8b40:BTW0P0[:bT0100V9R@0<c<a03<c<@0f=S@201hN700402`/:P0B
4Q004Q8@02TY9`80:RXX00@09bLU02DU8`0U9B<0;Bd[0P0X:2H0100f=SH06ATI01TI6@0V9RH201/K
6`04024Q8@0];Bd0;Bd]02h^;P807QhN00@08R8R02LW9`0W9bL08b<S0P0j>SX00`0d=3@06QXJ01XJ
6P0302DU9@0402TY:`15ADL0ADE702<S9@80<S8d00@09bLY01lO8@0O7b409bLY0P0n?T00100P8280
92@V02@T9P0Y:B/201/K7@0402h^<00h>3X0>3Pj01hN8080;2`^00@0:2PZ01dM7`0M7Al0=3@f0P0e
=CL0100L71h09RHX02HV:00J6Q`2034a<`0402<S9@0M7Al07AdO01TI6`808b<U00@0;Bd_028R900R
8R@06a/M0P0T92H0100A4A<07alQ01lO8@0`<38202DU9`0401hN800U9BL09BDW028R9080>c/m00@0
8B4S02<S9@0S8bD092@V0P0Z:R`01011@D<0?Cdo03dm?`0X:2X203Lg>@0303@d=P0[:bd0:b/]00<0
=3@f00@0;2`^02@T9P0T92H09BDW0P0m?Cl0100Q8B<0;bla02l_<@0H61X202/[;@0402DU9`0K6ad0
6a/M03lo@@809BDW00@05QHH00h>400>3Q006a/M0P0h>3X0100V9RP07alQ01lO8@0L71h202<S9@04
01hN800P8280820R01TI6`803Ph@00@03@d?02DU9`0U9BL09BDW0P0^;S00100>5180@TQ60498AP1@
EU@203To?@0406Y`KP2MXj40WJ>Q0:F[Z@80UifK00<0Y:ZV0:J/Z02V[:P00`2X[ZX30:J/Z0040:N]
Z@2Y[j/0ZJn[08V8P`807Q/F00@091`J03H/:`0f;2/0;R@U0P0Y82<0100[9bP0<2h_030^;`0c=C@2
03m1@00402h`;`0U9RP09BHX04E6BP80>CXn00@0=SLl02`]<P0/;C80;bd`0P0D4QD0100>30l06ALJ
01TG6P0D4QD202XX:`0403/i?00e<cH0=C<f03Dc=P80;bd`00@0;R`_03dk?P0m>ch0<Blb0P0/:Rd0
100W9BP0>SPk03Xh>`16ATP203dm?`0402<S9@0k>cd0>c/m02XZ;0809BDW00@0EeMI03@d=P0d=3H0
;Bd_0P0Q8B<0100o?d409bLY02LW:@0];Bl204i>D004034a<`0k>cd0>c/m02@T9P80:RX/00@0>c/m
04E5A`15ADL0:RX/0P0Q8B<0100e=CL0;Rh`02h^<00^;S0204=3A@0402HV:00[:bd0:b/]02`/;P80
:RX/00@0;Rh`03lo@@0o?d40:RPY0P0^;2d0100`;Rl0<Bl`034_<00l>S/203Pf=`0402/Y:P0d<S<0
=38c02LU9P808R0Q00@0CTa=02l];P0_;Bh0;bd^0P0[:BX0100e<c@0<2h_030^;`1CDE8202XX:@04
03Xh>@0_;Bh0;bd^03Hd=@80<2h_00@0>SPi03hl?@0n?3d0;2X[0P0]:b`0100O7Ah0;bd^02l];P13
@D8202d[;00403<a<P0Y9bP0:BLX02TW:080=C<d00@0>CLh038`<@0b<340Cde>0P0`;Rl0100_;Bh0
:2HW02PV9`1EDe@203Hd=@0403Dc=00a;c00<Bl`03Dc=080?3Xk00@0B4I7044o@011?d00:BLX0P0[
:BX0100_;Bh0:BLX02TW:00o?Ch203`j>`0402h/;@0Z:2T0:RPY04]9BP80?cdn00@0>3Hg04=1@P13
@D80?3Xk0P0`;Rl0101FE5D0<S0a038`<@0^;2d203hl?@0402/Y:P0g=CH0=cDf0490@@80@3ho00@0
?cdn06ITI@1VI6D0@3lm0P0^;B/0100R8Al0G5]I05aKF@0^;B/203Pg=@0403Lf=00k>SP0>cXh03Dc
=080>3Hg00@0:RPY05]IFP1KFEX0;R`]0P0Z:2T0100j>3T0:BLX02TW:016A4D204=1@P0402d[;00V
92D09R@U03Dc=080=38c00@0;R`]03Pf=`0h=SL0>3Hg0P1SHF80100l>S/0@Cm0044o@00h=SL202d[
;00402/Y:P0m>c`0?C/l02PV9`80>3Hg00@0A49305UGF01IEeP0:RPY0P0g=CH0100_;Bh0:RPY02XX
:@17ADH20490@@0402`Z:`0e<c@0=C<d02/Y:P80<Bl`00@0@T11040n?`10?Sl0;bd^0P0/:R/103Pf
=`409B<T0@3oool0000404ECGP14DUd0A59M04=AG080@U1K00@0@e5L04ABG@14DUd0AUAO0P17EF00
0`14DUd0AE=N04ECGP0304ECGP<0AUAO0`17EF001018EEh0AeAM04MDG@10BE0202@Y;@0400D93008
20P020P8014@3P802`L400@05QDC00d<2P0=30X03`h<0P081`D0100=30X02@P600T81P0V9B<203ln
?00402XZ:00S8b408b<Q02TY9`80:b/Y00@09RHT03<c<@0c<c40=SHd0P0N7Q`0100/;2X04Q8@018B
400Y:BL202XZ:00402LW9@0U9B<09BDS02d]:`80:2PV00@0=SHf01TI6@0I6AT09RHV0P0K6a/0100Q
8B40;Bd]02d];@0^;Rh201hN7P04028R8P0W9bL09bLW02<S8`80>SXj00<0=3@d01XJ6P0J6QX00`0U
9BD0100Y:B/0ADE704E5A`0S8bD2038b=00402LW:@0O7b407alQ02LW:@80?Si000@0820R02@T9P0T
92H0:BT[0P0K6ad0100^;S00>3Pj03Ph>P0N7R0202`/;P0402PX:P0M7Al07AdO03@d=P80=CDg00@0
71`N02HV:00V9RP06QXL0P0a<C<0100S8bD07AdO01dM7`0I6A/202<S9@0402d];`0R8R@08R8T01/K
7@8092@V00@04A4C01lO8@0O7b40<30b0P0U9BL0100N7R009BDW02DU9`0R8R@203/k?@04024Q8`0S
8bD08b<U02@T9P80:RX/00@0@D5303dm?`0m?Cl0:2PZ0P0g=cT00`0d=3H0:b/]02/[;@0303@d=P04
02`/;P0T92H092@V02DU9`80?Cdo00@08B4S02l_<@0_;c4061PJ0P0[:bd0100U9BL06a/M01/K7@0o
?d4202DU9`0401HF600>3Q003Ph@01/K7@80>3Pj00@09RHX01lO8@0O7b4071`N0P0S8bD0100N7R00
820R020P8P0I6A/200h>400400d=3`0U9BL09BDW02DU9`80;Rh`00@03Q@B0498AP12B4H0D5ID0P0i
?cd0101ZL6h0WJ>Q09fSX@2UZjT209NMV`030:BZYP2V[:P0YZbX00<0Z:jZ0`2V[:P0102W[JT0ZJn[
0:V_Z`29R8<201hK5P0402@L6P0f;2/0=R`[02hT9@80:B0S00@0:bLX030^;`0`;Rl0<cDd0P0o@D00
100^<2l09BHX02DV:015ATX203Tj?P0403Hg?00/;C80;2db02l]<080518E00@03P`?01TG6P0I5aX0
518E0P0Z:2/0100k>C`0=C<f03Dc=P0e<cH202l]<00402h/;`0m>ch0?C/n034_<P80;2X]00@09bDX
03Xh>`0j>3/0ATI80P0m?Cl0100S8bD0>c/m03/k?@0Z:R`202DU9`0405MGF@0d=3H0=3@f02d];`80
8B4S00@0?cm102LW:@0W9bT0;Bd_0P1>CU00100a<C<0>c/m03/k?@0T92H202XZ;00403/k?@15ADL0
ADE702XZ;0808B4S00@0=CDg02h^<00^;S00;Rh`0P13@dD0100V9RP0:b/]02/[;@0/;2h202XZ;004
02h^<00o?d40?cm102XX:@80;R`]00@0<2h_034_<00a;c00?3Xk0P0h=SL0100[:BX0=38c03@b<`0W
9BH2028P8@0404i<C@0_;Bh0;bd^02l];P80:bTZ00@0=C<d030^;`0`;Rl0De5B0P0Z:2T0100j>3T0
;bd^02l];P0f=3D2030^;`0403Xh>@0n?3d0?S`m02`Z:`80;B//00@07adN02l];P0_;Bh0@d520P0]
:b`0100c<C80:BLX02TW:00Y9bP203Dc=00403Tg>00b<340<S0a04m=CP80<2h_00@0;bd^02PV9`0X
9RL0EE=D0P0f=3D0100e<c@0<Bl`034_<00e<c@203`j>`0404Q6A`11?d00@Cm002TW:080:bTZ00@0
;bd^02TW:00Y9bP0?cdn0P0l>S/0100^;2d0:RPY02XX:@1;BDX203lm?P0403Pf=`13@D80@d5203`j
>`80<2h_00@0EUAE038`<@0b<340;R`]0P0n?3d0100[:BX0=cDf03Le=P12@442040n?`0403lm?P1V
I6D0IVAU040o?@80;Rd[00@08R4O05aKF@1LFeT0;Rd[0P0h=cD0100g=S@0>cXh03/j>00e<c@203Pf
=`0402XX:@1KFEX0FeUJ02h/;@80:RPY00@0>SPi02TW:00Y9bP0ATA50P13@D80100]:b`09R@U02HT
9@0e<c@203@b<`0402h/;@0h=SL0>3Hg03Pf=`80Hf5R00@0?3Xk044o@011?d00>3Hg0P0]:b`0100[
:BX0?C/l03dk?00X9RL203Pf=`0404A2@`1IEeP0FEMH02XX:@80=cDf00@0;bd^02XX:@0Z:2T0AdE6
0P12@440100/:R/0=C<d03Dc=00[:BX2034_<0040490@@10?Sl0@3ho02l];P80;2X[0@0h=SL102DS
9040oooo00001014DUd0@e5L04=AG013DE`2049@F`0404=AG014DUd0A59M04ECGP80AUAO1P15Deh3
04IDG`<0AeEP00@0B5EN04MDG@17E5d0@daC0P0Z<CL0100>4QD02`/;00/;2`0A40h200l;2004010?
3@0920H02@P600h=2`802@P600@03@`:00P71@081`D07ahL0P0d<c40100a<Bl0;2`Z02`/:P0`<2h2
02h^;00402DU8`0^;R`0;Rh/030`;P806a/I00@09BDS014A3`0A4@l0:BTW0P0];B/0100X:2H0820N
020P7P0X:2H202TY9`0403Hf=P0O7al07alO02@T90806QXJ00@08R8R02l_;`0_;bl0=3@d0P0P8200
100R8R809RHV02HV9P0N7Qh203Ph>00403dm?@0Q8B408B4Q02HV9P80<30`00@06QXL02/[;@0[:bd0
:b/]0P0U9BL0100O7b408B4S024Q8`0O7b420451@`0401TI6`0M7Al07AdO02LW:@8071`N00@0:b/]
03`l?P0l?3h0:2PZ0P0V9RP0100Z:R`08B4S024Q8`0a<C<203Ti>`0402HV:00Y:B/0:BT[028R9080
=CDg00@0<30b02TY:`0Y:B/08R8T0P0Y:B/0100c<cD0;2`^02`/;P0V9RP202h^<00401/K7@0I6A/0
6ATK02`/;P809bLY00@07AdO02`/;P0/;2h071`N0P15ADL0100Z:R`0;bla02l_<@0^;S0202XZ;004
03/k?@0l?3h0?3`n02HV:080;2`^00@09bLY020P8P0P8280;2`^0P0^;S00100V9RP07AdO01dM7`0O
7b4203Lg>@0401dM7`0/;2h0;2`^02<S9@80<c<e00@08b<U02@T9P0T92H0A4A60P0@418010000080
5ADG01DE5`0T92H2038b=004020P8P0G5aT05aLI01LG6@80;2`^00@07alQ01@D5P0D51H071`N0P0C
4aD0100`<380410B010@4P11@D<201dM7`0402@Z:00i?cd0>Clm03<i=`80?dE300@0MG]i0:6WY@2Q
YjD0YZbZ0P2PYZ@0102UZjL0YjfY0:N]Z@2X[ZX20:J/Z0040:BZYP2SZJD0XjVU0:BZYP80YZbX00@0
YZNQ03Hc;P0f<bh0:B4O0P0j<2l0100b:2T091/N02@K7P0[9bP202l_;`0402XZ:P0i>CT0>CTi038b
<P80<c<e00@0ATE:00X93P0:2@h0=cHl0P0K6R00100U8bH0;B/^02d[;P0a;c8201@B5@0403Hd=`0T
8RD0928U03Tg>P80C4Y=00@0;B/^03Xh>`0j>3/0:RP[0P0g=CP0100_;C00;B/^02d[;P0^;2l2024O
8P0404m?D@12@T@0@T94024Q8`809bLY00@0<30b01hN800N7R00;Bd_0P1?Ce40100_;c40?3`n03`l
?P0Y:B/203@d=P04038b=00X:2X0:2PZ04E5A`80@T9400@09BDW02HV:00V9RP0<c<e0P17AdT0100g
=cT08b<U02<S9@0[:bd202TY:`0402LW:@0k>cd0>c/m02d];`80:b/]00@09RHX02LW:@0W9bT0:RX/
0P12@T@0100/:R/0<2h_030^;`0c<C8202XX:@0403Le=P11?d00@Cm002l];P80;R`]00@0;bd^02LU
9P0W9BH0CD]<0P0n?3d302l];P0403<a<P0`;Rl0<2h_04e;C0<0>cTj0P0d<S<0100k>CX0?3Xk03`j
>`18ATL205ECE004034_<00c<C80<c4b02DS9080<Bl`00@0AD=402d[;00]:b`0;bd^0`0Y9bP203Hd
=@0403Dc=00/:R/0;2X[04U7B080;bd^00@0;B//02TW:00Y9bP0D4i?0P11?d00100e<c@0=cDf03Le
=P0c<C82034_<00403`j>`1:B4T0BTQ903Dc=080<Bl`00@0=38c02LU9P0W9BH0?3Xk0P11?d00100c
<C80;R`]02h/;@0l>S/2030^;`0402LU9P0f=3D0=S@e03Dc=080;R`]00@0D4i?02d[;00]:b`0;2X[
0P0k>CX0100a;c00<c4b03<a<P0h=SL203Le=P0402h/;@1ACe00DDm@03Dd<P80:bXX00@08b8P05aK
F@1LFeT0>CPf0P0n?C/00`0c<S00=3<a03@c<@0302h/;@04028P8@17ADH0AdE602/Y:P808b4R00@0
<S0a02<Q8P0S8B80>SPi0P17ADH302LU9P0403<a<P0Z:2T0:RPY03@b<`80<c4b00@0:2HW04e;C01=
Bd`0?3Xk0P0h=SL0100c<C80:BLX02TW:00Y9bP303@b<`80<S0a00@0?C/l05ECE01EDe@0=S@e0P0d
<S<0100_;Bh0<2h_030^;`0j>3T204U7B00402h/;@0Y9bP0:BLX02d[;080;bd^00@0=S@e05ECE01E
De@0<2h_0P0e<c@10490@@40;R`]0@3oool000L0@e5L0P14DUd00`15Deh0AUAO04IDG`0404ECGP<0
AUAO1P17EF0204ICG00404E?F00e?480=Ca201`P8`803PlA0`0?3P`0100:2@L01`H400L6100<2`T2
00/:200400d<2P071P@01`H401LF50809RDS00@0=cLe03@d<P0d=380>3Pf0P0`<2h0100S8b40:2PV
02PX9P0Z:RP201TI5`0401dM6`0A4@l04A4?02DU8`80<S8`00@0;2`Z01dM6`0M7A/092@R0P0Z:RP0
100f=SH0:BTY02TY:@0U9BD201`L7004024Q8@0_;bl0;bl_03/k>`809RHV00@0;2`/02h^;P0^;Rh0
7QhN0P0b<S801013@d<09BDU02DU9@0L71`202h^;P0401hN800Q8B<08B4S0410@P808B4S00@0820R
02DU9`0U9BL06QXL0P0m?Cl0100];Bl08R8T028R900U9BL201TI6`0401lO8@0k>cd0>c/m03Ti>`80
;2`^00@0=3@f02XZ;00Z:R`0:b/]0P0j>S`0100/;2h0820R020P8P0M7Al202LW:@0402d];`0W9bT0
9bLY01`L7P807alQ00@0:RX/02PX:P0X:2X08R8T0P0V9RP0100M7Al0410B010@4P0P828202TY:`04
01lO8@0e=CL0=CDg01dM7`80DU9D00@08B4S02PX:P0X:2X09RHX0P0O7b40100^;S00=SHh03Hf>00V
9RP3024Q8`807QhP00@0:b/]034a<`0a<C<0:BT[0P0N7R00100Q8B<0=cLi03Lg>@0O7b4302XZ;080
=CDg00@0820R02PX:P0X:2X0@T940P00008010092@/09bLY02LW:@0[:bd202DU9`0401PH6P0?3a40
3`lA00l?4@80:RX/00@0;bla01PH6P0H61X01@D70P0K6ad0100W9bT0:2PZ02PX:P0h>3X202/[;@04
04U?C@0nA480?TA202X`;P80I6YX00@0U9ZH0:bb/02//[00YZbZ0P2UZjT00`2QYj<0XjVU0:>YY@03
0:F[Y`<0Y:ZV00@0YZbX0:R^ZP2X[ZX0/K:/0P1@CdX0100/9b@0?38a03`b<@0l<S<202<M7`0402h/
;@0h>3P0>3Ph020L7@80=S8c00@0920Q04M2AP17@TH0>3<i0P0N6Al0100T7bH08ahU02<N9@0U8bH2
04M5B00400/9300N71l07Q`O02DS9P80@Cm200<081hQ051>D@1@CU400`0a;c80100Y9bX0;2X]02`Z
;@0`;S4203`j?@04034_<P0U8bH09B<V03dm?`80?cm100@0=SHh02<S9@0S8bD0;bla0P0W9bT0100E
5AL0A4A604A4AP19BD/303De=`80;2`^00@0:BT[02l_<@0_;c40<C4c0P18B4X0100d=3H0;2`^02`/
;P0/;2h204E5A`0404=3A@0X:2X0:2PZ02HV:080:b/]00@09BDW03De=`0e=CL0>SXl0P0_;c40100T
92H0:BT[02TY:`0Y:B/204I6B00402h/;@0Z:2T0:RPY030^;`807A/L00@0:bTZ04M5AP17ADH0;R`]
0P0R8240100b<340:2HW02PV9`15@d@205ECE00303@b<`0n?3d0?S`m00<0?S`m00@0B4I705ABD`1D
DU<0>SPi0P0a;c00100_;Bh0<S0a038`<@0g=CH204a:B`0402XX:@0_;Bh0;bd^02DS9080;bd^00@0
BDM803Dc=00e<c@0<c4b0P0f=3D0100X9RL0>3Hg03Pf=`0c<C8202TW:0040490@@0a;c00<Bl`02XX
:@80:bTZ00@0>3Hg04A2@`14@T<0:bTZ0P0g=CH0100`;Rl0:bTZ02/Y:P0g=CH205aJF`0403@b<`0[
:BX0:bTZ03@b<`808b4R00@0=cDf04E3A015@d@0<c4b0P0X9RL0100k>CX0<2h_030^;`0T8R<203@b
<`0403Le=P0c<C80<c4b04i<C@80<2h_00@0;R`]03/i>P0k>CX0@3ho0P0e<c@0100d<S<0>3Hg03Pf
=`0W9BH204A2@`0404A3@@0b<Bl0<S4_02@S8@80E5=A0`0g=S@0100];2X0;2/Y02`[:@0_;Bh202h/
;@0402HT9@0h=SL0>3Hg03@b<`80:2HW00@0<Bl`02DS900U8b@0;R`]0P19AdP0100W9BH0:BLX02TW
:00`;Rl202@R8`0403Tg>00_;Bh0;bd^02DS9080?C/l00@0BdU:03Pf=`0h=SL0=38c0P0[:BX0100^
;2d0;bd^02l];P0n?3d202TW:00403<a<P1;BDX0BdU:04=1@P80;bd^00@0;B//038`<@0b<340;2X[
0P1?CDh00`0m>c`0:RPY02XX:@0303@b<`0402`Z:`1PGUl0H5iO04=1@P<0>cTj0@0^;2d10?ooo`00
1`13DE`204ABG@0304ECGP16E5l0AUAO00@0AE=N0`16E5l604MEH080AU=L00@0ADmH03Dl@P0e?480
720S0P0>3a4300l>300400X91`071P@01`H400`;2@802`X800@03@`:00L610071P@05aHD0P0V9B<0
100g=cD0=3@b03@d<P0h>3H2030`;P0402<S8@0X:2H0:2PV02XZ:0806ATG00@07AdK014A3`0A4@l0
9BDS0P0b<S00100/;2X07AdK01dM6`0T928202XZ:00403Hf=P0Y:BT0:BTY02DU9@8071`L00@08B4Q
02l_;`0_;bl0>c/k0P0V9RH0100/;2`0;Rh^02h^;P0N7Qh2038b<P0404=3@`0U9BD09BDU01`L7080
;Rh^00@07QhP024Q8`0Q8B<0@4120P0Q8B<0100P82809BDW02DU9`0J6Q`203dm?`0402d];`0R8R@0
8R8T02DU9`806ATK00@07alQ03/k?@0k>cd0>CTk0P0/;2h0100d=3H0:RX/02XZ;00[:bd203Xj?004
02`/;P0P8280820R01dM7`809bLY00@0;Bd_02LW:@0W9bT071`N0P0O7b40100Z:R`0:2PZ02PX:P0R
8R@202HV:00401dM7`0@4180410B020P8P80:BT[00@07alQ03De=`0e=CL07AdO0P1BDU@0100Q8B<0
:2PZ02PX:P0V9RP201lO8@0402h^<00f=SP0=SHh02HV:0<08B4S0P0N7R00100[:bd0<C4c034a<`0Y
:B/201hN8004024Q8`0g=cT0=cLi01lO8@<0:RX/0P0e=CL0100P8280:2PZ02PX:P12@T@200000P04
00T92`0W9bT09bLY02/[;@809BDW00@061PJ00l?4@0?3a403`lA0P0Z:R`0100_;c4061PJ01PH6P05
1@L201/K7@0402LW:@0X:2X0:2PZ03Ph>P80:b/]00@0BDm=03i4@P0nA480:S0^0P1TJVP0102DVYP0
[;:`0:bb/02V[:X20:F[Z@030:6WX`2SZJD0XjVU00<0YJ^W0`2TZZH0102V[:P0Z:jZ0:R^ZP2a/Z`2
051?BP0402`W900l<S40?38a03`b<`808adO00@0;R`]03Ph>00h>3P081`M0P0f<S<0100T8240Ad96
04M2AP0h<cT201hI7`0402@O9P0S7RD08ahU02DS9P80AdE800@02`T<01hL7`0N71l09B<V0P11?d80
0`0P7R40D4iA051>D@03034_<P0402TW:P0/:Rd0;2X]030^<@80?3Xm00@0<Blb02DS9P0U8bH0?Cdo
0P0o?d40100f=SP08b<U02<S9@0_;c4202LW:@0401DE5`14A4H0A4A604U9B`<0=CDg0P0/;2h0100Y
:B/0;bla02l_<@0a<C<204Q8BP0403@d=P0/;2h0;2`^02`/;P80ADE700@0@d=502PX:P0X:2X09RHX
0P0[:bd0100U9BL0=CDg03De=`0j>S`202l_<@0402@T9P0Y:B/0:BT[02TY:`80ATI800@0;R`]02XX
:@0Z:2T0<2h_0P0M6a`0100[:BX0AdE604M5AP0^;2d2028P8@04038`<@0X9RL0:2HW04E3A080EE=D
00<0=38c03hl?@0n?3d00`0n?3d01018ATL0E59C05ABD`0j>3T2034_<00402l];P0b<340<S0a03Le
=P80C4Y;00@0:RPY02l];P0_;Bh09B<T0P0_;Bh01019AdP0=C<d03Dc=00c<C8203Hd=@0402PV9`0h
=SL0>3Hg03<a<P80:BLX00@0@T11034_<00a;c00:RPY0P0[:BX0100h=SL0A49304A2@`0[:BX203Le
=P04030^;`0[:BX0:bTZ03Le=P80G5YK00@0=38c02/Y:P0[:BX0=38c0P0S8B80100g=CH0AD=404E3
A00c<C8202PV9`0403/i>P0`;Rl0<2h_02@R8`80=38c00@0=cDf03<a<P0c<C80CTa=0P0`;Rl0100^
;2d0>cTj03/i>P10?Sl203Dc=00403@b<`0h=SL0>3Hg02LU9P80A49300@0A4=1038a;`0b<Bl092<Q
0P1DDe4303Lf=00402d/:P0/:bT0;2/Y02l];P80;R`]00@09R@U03Pf=`0h=SL0=38c0P0X9RL0100a
;c009B<T02DS900^;2d204U7B00402LU9P0Y9bP0:BLX030^;`80928S00@0>CLh02l];P0_;Bh09B<T
0P0m>c`0101;BDX0>3Hg03Pf=`0d<S<202/Y:P0402h/;@0_;Bh0;bd^03hl?@80:BLX00@0<c4b04]9
BP1;BDX0@d520P0_;Bh0100]:b`0<S0a038`<@0/:R/204m=CP0303dk?00Z:2T0:RPY00<0=38c00@0
;2X[061NG`1PGUl0@d520`0k>CX102h/;@40oooo000604=AG0<0A59M1P15Deh304IDG`<0AeEP0`18
EV401017EF00AUAO04IDG`18DU/203m8C`0402P]<@0C51H04a@F00/;2`803Pd;00@02PT700T81P09
20H030/90P0=30X0100<2`T020L500P71@0B4@l201dL6P04038b<00f=S@0=SHd03Xj>080<30^00@0
8B4O02@T8P0T92809bLU0P0K6aT0100L71X05QHD01HF500O7ad203De<`0403@d<P0Q8Al08B4O02<S
8@80:2PV00@0<S8b03<c<`0c<c<0;2`/0P0Q8B40100M7Ad0:BTY02TY:@0k>c/202d];@0402DU9@0[
:b/0:b/[01TI6@808b<S00@0@410030`<00`<3007QhN0P0_;bl0100R8R@07alQ01lO8@14A4H2028R
900402PX:P0U9BL09BDW024Q8`80<c<e00@0>3Pj02LW:@0W9bT0;Bd_0P0_;c40100[:bd0>SXl03Xj
?00j>S`2028R900402TY:`0R8R@08R8T01/K7@80<S8d00@0;bla01`L7P0L71h07alQ0P0R8R@0100[
:bd0:BT[02TY:`0O7b4201dM7`0402TY:`0[:bd0:b/]02<S9@808B4S00@09BDW020P8P0P82807AdO
0P0/;2h0100N7R00;Bd_02d];`0I6A/203hn@00401hN800P8280820R01hN80806a/M00@09bLY030`
<P0`<3809bLY0P0N7R03020P8P0402`/;P0e=CL0=CDg02`/;P807QhP00@08R8T03@d=P0d=3H0:2PZ
0P0Y:B/00`0U9BL0<30b030`<P03020P8P04034a<`030`D00`<502@T9P809RHX00@071`N01HF600F
5QP061PJ0P0J6Q`0100M7Al0;2`^02`/;P0j>S`2010@4P0400D51`0E5AL05ADG01dM7`80@d=500@0
71`N0410@P10@480=3Xh0P0X;R`0100V;2X0QH^908F;R@2X[Z`20:ne/`040:BZZ02UZjT0YJ^Y09fS
W`80X:JR00@0Y:ZV0:J/Z02V[:P0Z:jZ0P2Z/:`0102^];00/KNc0;6g/`2e^;4207YiM00403/f<`0d
;2X0=2`Z040g>0808A/M00@08b4R02d_;P0];bh0<bX]0P0Q61/01012>C`07A@I01dD6@1?ADd203H/
=0<0:R0Y00@0;R`_02TW:P0Y9bX02@L:0P0I5aX0100_;C00<Blb034_<P0Y9bX203Hd=`0404U7BP0U
8bH09B<V02h/;`809B<V00@0<Blb04M5B017ADP0:bT/0P0i=cX0100S8bD0=cLi03Lg>@1<C4h202l_
<@0402<S9@0e=CL0=CDg02XZ;0809BDW00@0BTY<04U9B`19BD/0:RX/0`0];Bl202HV:00403Xj?00`
<380<30b04U9B`80?Cdo00@0;2`^03/k?@0k>cd0@d=50P0^;S00100[:bd0<c<e03<c=@0];Bl203<c
=@0404M7B@0d=3H0=3@f02LW:@80<C4c00@0:b/]04A4AP14A4H0BTQ90P0h=SL0100n?3d0:bTZ02/Y
:P0i=cP205]IFP04044o@00a;c00<Bl`03Pf=`80;2X[00@0>SPi069PH@1RH640;R`]0P0j>3T0100c
<C80=C<d03Dc=00^;2d205MEEP04030^;`0Y9bP0:BLX02LU9P80;bd^00@0<2h_04]9BP1;BDX0:2HW
0P0[:BX0100U8b@0:bTZ02/Y:P19AdP203Xh>@0403@b<`0m>c`0?C/l02LU9P80>SPi00@0=38c02/Y
:P0[:BX0?S`m0P0g=CH302XX:@0402d[;01@CTl0D4i?02/Y:P80<c4b00@0;B//02TW:00Y9bP0;B//
0P1GEEH0100a;c009R@U02HT9@0f=3D202TW:00403Xh>@1;BDX0BdU:03Le=P80928S00@0=C<d038`
<@0b<3409B<T0P0`;Rl0100_;Bh0:RPY02XX:@10?Sl202`Z:`0402/Y:P0e<c@0=C<d04I4A@80<c4b
00@0<2h_03Xh>@0j>3T0:BLX0P10?Sl0101EE580;bh/02l^;00S8R02051?C@04038a;`0^;B/0;Rd[
030_;@80:RTW00@0<2h_02l];P0_;Bh0;R`]0P0_;Bh01011?d00<c4b03<a<P0a;c0202PV9`0402`Z
:`15@d@0AD=403Hd=@<0;R`]0P0Z:2T0100f=3D0<Bl`034_<00[:BX203Dc=00405MEEP0i=cP0>CLh
03<a<P80:bTZ00@0<Bl`02/Y:P0[:BX0?C/l0P0U8b@0100]:b`0?C/l03dk?01:B4T2038`<@0402d[
;00/:R/0;2X[02TW:080BTQ900@0De5B03hl?@0n?3d0=cDf0P0m>c`0100^;2d0BTQ904Y8B@1BD542
03Hd=@40:RPY0@0^;2d10?ooo`0000@0@e5L04ABG@14DUd0A59M0`15Deh904IDG`<0AeEP1@18EV40
1016EV<0B5IQ04QFH@19EUl204I?EP0403<j@00H71l061`O00X:2P802@T900@03Pd;00d<2P0=30X0
30/90P0>3@/0100:2@L030/900`;2@0A40h201XI5`0402LW9@0`<2h0<30^03Hf=080;Bd[00@08B4O
02@T8P0T92809bLU0P0O7ad0100Q8Al0820N020P7P0H61H203<c<@0403/k>@0Y:BL0:BTW02HV9080
92@R00@0;2`/03Hf=P0f=SH0=SHf0P0Y:BT0100I6AT08R8R028R8P0b<S8203<c<`04020P800X:2P0
:2PX01/K6`805aLG00@0=3@d03Lg=`0g=cL08R8R0P0X:2P301/K7@0402h^<00S8bD08b<U02d];`80
820R00@0;Rh`02TY:`0Y:B/0@d=50P0[:bd0100Y:B/0<c<e03<c=@0[:bd202d];`0403Lg>@0[:bd0
:b/]024Q8`807QhP00@05ADG02/[;@0[:bd0<c<e0P0M7Al0100R8R@07alQ01lO8@0T92H202TY:`04
028R900M7Al07AdO02LW:@80;Bd_00@08b<U01/K7@0K6ad07alQ0P0];Bl0100M7Al0:2PZ02PX:P0P
828202@T9P04020P8P0X:2X0:2PZ02h^<0809RHX00@08B4S02DU9`0U9BL0:b/]0P0_;c400`0/;2h0
820R020P8P0301lO8@0402LW:@0c<cD0<c<e02/[;@806QXL00@07AdO02XZ;00Z:R`0:b/]0P0V9RP0
100M7Al0;Bd_02d];`0X:2X2018B500401DE5`0>3Q003Ph@03dm?`80820R00@05ADG01TI6`0I6A/0
7alQ0P0S8bD0100W9bT08R8T028R900[:bd200000P0402DU9`0J6Q`06QXL02l_<@80:RX/00@09RHX
03@d=P0d=3H0728P0P0P9R@01011AdD0WjFS09nUX`2_]K<20:bb/0<0Yjf[00@0XJNS0:BZYP2TZZH0
Z:jZ0P2Z/:`0102//Zh0[kFa0:ne/@2c^KD20;Jl^0040;Nj/`2LWIL0W9fG04I2?`807QTF00@0>S@d
02XV9`0Z9RL08R8R0P0U9bH0100j<C@0=B`_03D/;`0g;S42038Y;P04020F7P1?ADd0CdE=02`R:`80
;B</00@0928U00`:3@0<2Pd08AlR0P1;BD`0100]:bh081hQ020N8@0e<cH202PV:@0404a:C@0g=CP0
=cDh030^<@80:2HY00@0<Blb03Pf>@0h=ST0<2ha0P0_;C00100S8bD0<c<e03<c=@0m?Cl203Ti>`04
02PX:P0^;S00;Rh`03`l?P80:BT[00@0<S8d0451@`11@D<0@4120P0_;c40100V9RP0;Bd_02d];`0b
<S@202`/;P0404];C@18B4X0B4Q:030`<P80<C4c00@0>c/m034a<`0a<C<0;bla0P0`<380100d=3H0
;bla02l_<@18B4X203<c=@0402TY:`0e=CL0=CDg02/[;@80=cLi00@0BTQ902TW:00Y9bP0;2X[0P0T
8R<0100_;Bh0BTQ904Y8B@0h=SL202TW:004038`<@0W9BH09bDV02TW:080GeeN00@09R@U030^;`0`
;Rl09bDV0P0X9RL0100T8R<0EeEF05MEEP0d<S<202d[;00402h/;@0g=CH0=cDf03<a<P80C4Y;00@0
=C<d02d[;00]:b`0:2HW0P0Y9bP01012@440=cDf03Le=P0/:R/203Dc=00402LU9P0j>3T0>SPi03Pf
=`80<c4b00@0>SPi03lm?P0o?Ch0;B//0P0X9RL0100[:BX0EUAE05IDE@0e<c@2030^;`04038`<@0e
<c@0=C<d034_<080Cde>00@0@d5203<a<P0c<C80?3Xk0P0c<C80100n?3d0C4Y;04a:B`14@T<202XX
:@0403Pf=`10?Sl0@3ho03Dc=080?C/l00@0=C<d030^;`0`;Rl0@T110P0n?3d0100W9BH0;2X[02`Z
:`13@D8202`Z:`0402TW:00e<c@0=C<d02TW:080=cDf00@0DE1>028Q7`0R8Al08B0N0P1@Cdd0100e
=380;bh/02l^;00k>SP202d/:P0402`Z:`0^;2d0;R`]038`<@80:RPY00@0A49303`j>`0l>S/0;2X[
0P0W9BH0100a;c00?C/l03dk?01:B4T203@b<`0402d[;00g=CH0=cDf02h/;@80=cDf0`0[:BX0101C
DE80=S@e03Hd=@0]:b`202XX:@04038`<@0[:BX0:bTZ03Tg>080;R`]00@0<2h_03Hd=@0f=3D0CD]<
0P10?Sl0100e<c@0:RPY02XX:@0b<34203`j>`0405UGF01?CDh0Cde>02/Y:P80?3Xk00@0=C<d02HT
9@0V92D0BdU:0P0h=SL102XX:@40<S0a0@3oool0000404=AG014DUd0A59M04ABG@<0AE=N2@16E5l3
04MEH0D0B5IQ00@0AUIS04QFH@18EV40BEIO0P16CeH0100c>T0061`O01PL7`0:2PX200T92@0400h=
2`0=30X03@`:00`;2@803Pd;00@02PT700`;2@0<2`T04A0>0P0J6AL0100W9bD0<30^030`;P0f=S@2
02d]:`04024Q7`0T928092@R02LW9@807alM00@08B4O020P7P0P81h061PF0P0c<c40100k>cT0:BTW
02TY9`0V9R@202@T8P0402`/;00f=SH0=SHf03Hf=P80:BTY00@06ATI028R8P0R8R80<S8b0P0c<c<0
100P8200:2PX02PX:00K6a/201LG5`0403@d=00g=cL0=cLg028R8P80:2PX0`0K6ad0100^;S008b<U
02<S9@0];Bl2020P8P0402h^<00Y:B/0:BT[04=3A@80:b/]00@0:BT[03<c=@0c<cD0:b/]0P0];Bl0
100g=cT0:b/]02/[;@0Q8B<201hN800401DE5`0[:bd0:b/]03<c=@807AdO00@08R8T01lO8@0O7b40
92@V0P0Y:B/0100R8R@07AdO01dM7`0W9bT202d];`0402<S9@0K6ad06a/M01lO8@80;Bd_00@07AdO
02PX:P0X:2X0820R0P0T92H0100P8280:2PZ02PX:P0^;S0202HV:004024Q8`0U9BL09BDW02/[;@80
;bla00<0;2`^020P8P0P82800`0O7b40100W9bT0<c<e03<c=@0[:bd201XJ700401dM7`0Z:R`0:RX/
02/[;@809RHX00@07AdO02d];`0];Bl0:2PZ0P0B4Q@0100E5AL03Ph@00h>400m?Cl2020P8P0401DE
5`0I6A/06ATK01lO8@808b<U00@09bLY028R900R8R@0:b/]0P000080100U9BL06QXL01XJ700_;c42
02XZ;00402HV:00d=3H0=3@f01`R808082HT00@0@DM509nUX`2OYJ<0[kFc0P2//[030:N]Z`040:6W
X`2TZZH0Y:ZV0:R^ZP80Z[2/00@0[;:^0:ne/@2_]K40/kVe0P2f_;P0102g^[<0W9fG09bMU`16@Sl2
01hI5P0403Xd=00Z9RL0:RHW028R8P809BLV00@0>S4d03D/;`0e;2l0=bha0P0b:Bh0100P5Qh0CdE=
04m5C@0/8R/202dS;00402@R9@0<2Pd030X=024O8P80BdU<00@0;B/^020N8@0P7R40=C<f0P0X9RT0
101<BTd0=cDh03Le>00`;S4202PV:@04034_<P0h=ST0>3Hi030^<@80;bd`00@08b<U03<c=@0c<cD0
?Cdo0P0i>C/0100X:2X0;Rh`02h^<00l?3h202TY:`04038b=011@D<0@D530410@P80;bla00@09RHX
02d];`0];Bl0<S8d0P0/;2h0101;Bdd0B4Q:04Q8BP0`<382034a<`0403/k?@0a<C<0<C4c02l_<@80
<30b00@0=3@f02l_<@0_;c40B4Q:0P0c<cD0100Y:B/0=CDg03De=`0[:bd203Lg>@0404Y8B@0Y9bP0
:BLX02`Z:`80928S00@0;bd^04Y8B@1:B4T0>3Hg0P0Y9bP0100b<3409bDV02LU9P0Y9bP205mMGP04
02HT9@0`;Rl0<2h_02LU9P80:2HW00@0928S05MEEP1GEEH0=38c0P0]:b`0100^;2d0=cDf03Le=P0c
<C8204a:B`0403Dc=00]:b`0;B//02PV9`80:BLX00@0@T1103Le=P0g=CH0;2X[0P0e<c@0100W9BH0
>SPi03Xh>@0h=SL203<a<P0403Xh>@0o?Ch0?cdn02d[;080:2HW00@0:bTZ05IDE@1FE5D0=C<d0P0`
;Rl0100b<340=C<d03Dc=00a;c0204m=CP0404=1@P0c<C80<c4b03`j>`80<c4b00@0?S`m04a:B`1<
BT/0A4930P0Z:2T0100h=SL0@3ho040n?`0e<c@203dk?00403Dc=00`;Rl0<2h_0490@@80?S`m00@0
9bDV02`Z:`0/:R/0@d520P0/:R/0100Y9bP0=C<d03Dc=00Y9bP203Le=P04055@CP0R8Al08R4O024P
7P80D4m=00@0=C@b02l^;00_;R`0>cXh0P0];2X0100/:R/0;R`]02h/;@0b<34202XX:@0404A2@`0l
>S/0?3Xk02`Z:`809bDV00@0<Bl`03dk?00m>c`0BTQ90P0d<S<0100]:b`0=cDf03Le=P0^;2d203Le
=P<0:bTZ00@0De5B03Hd=@0f=3D0;B//0P0Z:2T0100b<340:bTZ02/Y:P0i=cP202h/;@04030^;`0f
=3D0=S@e04e;C080@3ho00@0=C<d02XX:@0Z:2T0<S0a0P0l>S/0101IEeP0Cde>04m=CP0[:BX203`j
>`0403Dc=00V92D09R@U04]9BP80>3Hg0@0Z:2T1038`<@40oooo00001014DUd0AE=N04ECGP16E5l2
04MEH00304QFH@17EF00AeEP00<0AeEP0`16E5l304MEH0@0B5IQ0P19Ef801014EV@0B5QU04QHI@19
Ef8204UCG00403m6C00U:B`09BT/00h?4@802PX:0`0A40h0100<2`T03`h<00l>300:2@L2010?3@04
014@3P0G5Q@05aHD01dM6`80:b/Y00@0<c<a02`/:P0/;2X092@R0P0V9R@0100W9bD08R8P028R800S
8b4202/[:@0401@D4P0^;R`0;Rh/03`l>P80;Bd[00@0:2PV028R800R8R009RHV0P0`<300100i>CT0
;Bd]02d];@0H61P2020P800302DU9@0f=SH0=SHf00<09bLW00@08R8R01HF5P0F5QH0:BTY0P0j>SX0
100W9bL07alO01lO7`0V9RP202LW:@0402DU9`0b<S@0<S8d030`<P807AdO00@0<c<e024Q8`0Q8B<0
@d=50P0^;S00100K6ad08R8T028R900N7R0201/K7@0403Lg>@12@T@0@T9402@T9P808b<U00@07QhP
02XZ;00Z:R`0>3Pj0`0T92H201hN800401dM7`0Z:R`0:RX/02HV:08071`N00@092@V02d];`0];Bl0
92@V0P0I6A/0100F5QP0:2PZ02PX:P0T92H201`L7P0402@T9P0S8bD08b<U02/[;@8092@V00@0<C4c
02@T9P0T92H07QhP0P0X:2X0100[:bd0<C4c034a<`0e=CL202DU9`0402LW:@0V9RP09RHX02PX:P80
=SHh00@0<30b01hN800N7R00820R0P0X:2X0100Y:B/08b<U02<S9@0N7R02030`<P0402/[;@092@/0
2@T;00H620805aLI00@0?3`n01hN800N7R007QhP0`0U9BL201lO8@04024Q8`0C4aD04a<E018B5080
2@T;00@092@V0451@`11@D<07QhP0P0E5AL0100i>C/09bLY02LW:@0O9B<202`b<00406idLP2^];80
[[Bb0:jd/P80Z:j/00@0[kFc0:bb/02//[00Z[2/0P2//Zh60:ne/@040;:h]02d^[H0];Zf0;Rm]P80
/;6[00@0G5]G01/F4`0K5Q<0;bTY0P0b<340100b<S809RXY02HZ:@0^:R/202TU9P0403hj>`0Z9BT0
:RDY01dH7P806QDK00@0EE1G02PS:P0X8bX08b4T0P0<2Pd0100L6Qd0HV1S069PH`0R82<202PV:@04
03Pf>@0V92L09R@W02/Y;080BdU<0`0]:bh0100`;S409R@W02HT9`17ADP2028P8`0402`/;P0`<380
<30b01lO8@80@D5300@0BTY<02DU9`0U9BL0;Rh`0P0k>cd0100K6ad09RHX02HV:01JFU`203@d=P04
024Q8`0j>S`0>SXl02<S9@80?Cdo00@0>CTk04A4AP14A4H0=3@f0P0_;c40100e=CL0<30b030`<P0]
;Bl202DU9`04038b=00Y:B/0:BT[0492A080;bla00@09bLY034a<`0a<C<0:RX/0P0V9RP0101?CDh0
:2HW02PV9`0T8R<202PV9`0402`Z:`0m>c`0?C/l03Tg>080:RPY00@0;bd^02TW:00Y9bP09B<T0P1M
Fe`00`0]:b`0<c4b03<a<P0302h/;@04038`<@1GEEH0EeEF044o@080;bd^00@0<S0a03<a<P0c<C80
;2X[0P0j>3T01010?Sl0:BLX02TW:00X9RL202HT9@0403`j>`0j>3T0>SPi02h/;@80=38c00<0:2HW
03Pf=`0h=SL00`0j>3T0100b<340A49304A2@`0a;c0202DS900402/Y:P1=Bd`0CD]<04=1@P80<2h_
00@0?C/l044o@011?d00>SPi0P19AdP0101EDe@0?C/l03dk?00e<c@2034_<00403<a<P0j>3T0>SPi
04Q6A`80;bd^00@08AlP03<a<P0c<C80:RPY0P0a;c00100Y9bP09bDV02LU9P0f=3D203hl?@04030^
;`0_;Bh0;bd^04A2@`80=38c00@0<2h_03Dc=00e<c@0<Bl`0P0d<S<0101CDU00;Rd[02h]:`0W9R@2
0450?P04038a;`0^;B/0;Rd[03dl>P80<2l]00@0<Bl`03@b<`0d<S<0>CLh0P0_;Bh01011?d00BDM8
04U7B00[:BX202d[;004038`<@0c<C80<c4b055?D080>3Hg00@0;bd^03dk?00m>c`0:2HW0P0j>3T0
100]:b`0;2X[02`Z:`1:B4T2040n?`0403<a<P0d<S<0=38c03/i>P80>CLh00@0=cDf03lm?P0o?Ch0
=S@e0P0d<S<01018ATL0D4i?051>C`0o?Ch2030^;`0403Xh>@0_;Bh0;bd^04]9BP80Cde>00@08R0Q
034_<00a;c00=cDf0P0H5QL01@0m>c`0B4I704Q6A`0h=SL0:bTZ0040oooo00001015Deh0AUAO04ID
G`18EV4204UGHP0404YHH`19Ef80BEMR04UGHP80B5IQ1017EF0504QFH@<0BEMR00@0@eES04YJI`1:
FVL0BUQS0P1<EUl01019DUT0=3Tm03@i?@0J6ad200h?4@04010?3@0A40h04A0>00/:20804A0>00@0
2`X801<B400C4Q003`h<0P0C4Q00100K6aT0:b/Y02/[:@0c<c4202h^;00402TY9`0Z:RP0:RXX02LW
9@<08B4O0P0d=380100D5180:BTW02TY9`0e=C<202XZ:00302TY9`0R8R008R8P00<08R8R00@0=SHf
02h^;P0^;Rh06a/K0P0W9bL0100J6QX0=SHf03Hf=P0[:b/201`L700402HV9P0N7Qh07QhN02LW9`80
@d=300@0>3Ph02/[:`0[:b/0<c<e0P0^;S00100W9bT0@D530451@`0Z:R`201`L7P0402TY:`0Q8B<0
8B4S02LW:@80;bla00@07alQ02TY:`0Y:B/0:BT[0P0K6ad0100_;c40?cm103lo@@0M7Al2020P8P04
02@T9P0W9bT09bLY03/k?@80<30b00@0;2`^02HV:00V9RP08b<U0P0e=CL0100b<S@08b<U02<S9@0Y
:B/203De=`04034a<`0V9RP09RHX02HV:08092@V00@0?3`n010@4P0@41808b<U0P0R8R@3020P8P04
02h^<00T92H092@V01lO8@809RHX00@0:2PZ03Hf>00f=SP0@D530P0V9RP0100^;S00:RX/02XZ;00W
9bT203Lg>@0403Hf>00S8bD08b<U02DU9`80:RX/00@0:b/]02@T9P0T92H0:2PZ0P0^;S00100L71h0
1PH800H6200=3@l201lO8@04028R900K6ad06a/M024Q8`807alQ00@09bLY01hN800N7R007AdO0P0D
51H0100;2`d09bLY02LW:@0F5QP204I6B004014A4`0R8R@08R8T030`<P80:RX/00@082HT02X`;P0Z
<2h0R8j<0`2X[Z`20:BZZ0040;Bj^02//[00[;:`0:fc[`80[kFa0`2a]k<0102`][80[kFa0:ne/@2a
]k<20;>i]@040;k3_02o`[/0_l:k09:AS@80?c/h00@081`K02XX:@0Z:2T0>C/j0P0W:bX0100Q8B40
;Bd]02d];@0L71`203Ph>P0401DD6@0M72407A`Q03dl@P80<c8h00<0;B/^018@4`0B41<00`0U8bH0
101=Bdh08AlR024O8P13@D@202<Q900401`J7@11?d80@Cm203`j?@80;B/^00@0:bT/02PV:@0X9RT0
B4I90P0d<SD0100Z:R`0<C4c034a<`0M7Al203hn@00405YJG00b<S@0<S8d02@T9P80<C4c00@06ATK
03@d=P0d=3H0:b/]0P15ADL0100o?d409BDW02DU9`0n?T0203<c=@0402PX:P0k>cd0>c/m03Ph>P80
=3@f00@0<C4c02d];`0];Bl0;bla0P0V9RP0100/;2h09RHX02HV:010@482030`<P0402HV:00/;2h0
;2`^02h^<080820R00@0E59C034_<00a;c009R@U0P0`;Rl0100]:b`0=cDf03Le=P14@T<202l];P04
03Hd=@0d<S<0=38c02/Y:P80FEMH00@0=S@e038`<@0b<340<S0a0P0^;2d0100d<S<0@T110490@@12
@44202TW:00403@b<`0e<c@0=C<d038`<@80>3Hg00@0C4Y;02HT9@0V92D0:2HW0P0V92D0100h=SL0
?cdn03lm?P0e<c@203@b<`0402/Y:P0d<S<0=38c03Tg>080>cTj00@09R@U04E3A015@d@0=C<d0P0S
8B80100c<C80@3ho040n?`1CDE82034_<00403lm?P0g=CH0=cDf02l];P80=C<d00@0EE=D03`j>`0l
>S/0:2HW0P0_;Bh0100[:BX0:2HW02PV9`1<BT/2038`<@0402h/;@11?d00@Cm003Hd=@80?cdn00@0
>cTj03hl?@0n?3d0AdE60P1EDe@0100f=3D0<c4b03<a<P16A4D2044o@00403lm?P0i=cP0>CLh03dk
?080=38c00@0CTe;04]:B01;BTP0<c8`0P0Y:2H0100b<Bl0<C0^034`;P0f=C<202l^;00403Le=P0i
=cP0>CLh03Xh>@80<c4b00@0=S@e059@D@1BD540:bTZ0P0d<S<0100Y9bP0;2X[02`Z:`16A4D203Xh
>@04034_<00f=3D0=S@e02XX:@80>3Hg00@0<S0a03<a<P0c<C80@Cm00P1=Bd`0100j>3T0?C/l03dk
?00n?3d203lm?P0403<a<P1:B4T0BTQ903@b<`80<2h_00@0>cTj05MEEP1GEEH0@Cm00P0g=CH0100e
<c@0;B//02d[;00g=CH204e;C004034_<00[:BX0:bTZ034_<080:BLX00D0<2h_051>C`1@CTl0@Cm0
02/Y:P010?ooo`0000@0AE=N04IDG`16E5l0B5IQ0P19Ef80101:F6<0BEMR04UGHP19Ef8204QFH@@0
AeEP1@18EV4304UGHP0404=EH`1:FVL0BUYW04YHH`80C5IO00@0BE9I03@i?@0d>Cd06Q/M0P0>3a40
100@3`d04A0>014@3P0;2PP2014@3P0400/:200C4Q004a8@00l>30804a8@00@06a/I02/[:@0[:bT0
<c<a0P0^;R`0100Y:BL0:RXX02XZ:00W9bD3024Q7`80=3@b00@051@B02TY9`0Y:BL0=CDc0P0Z:RP0
0`0Y:BL08R8P028R8003028R8P0403Hf=P0^;Rh0;Rh^01/K6`809bLW00@06QXJ03Hf=P0f=SH0:b/[
0P0L71`0100V9RH07QhN01hN7P0W9bL204=3@`0403Ph>00[:b/0:b/[03<c=@80;Rh`00@09bLY0451
@`11@D<0:RX/0P0L71h0100Y:B/08B4S024Q8`0W9bT202l_<@0401lO8@0Y:B/0:BT[02TY:`806a/M
00@0;bla03lo@@0o?d407AdO0P0P8280100T92H09bLY02LW:@0k>cd2030`<P0402`/;P0V9RP09RHX
02<S9@80=CDg00@0<S8d02<S9@0S8bD0:BT[0P0e=CL0100a<C<09RHX02HV:00V9RP202@T9P0403`l
?P0@4180410B02<S9@808R8T0`0P8280100^;S0092@V02@T9P0O7b4202HV:00402PX:P0f=SP0=SHh
0451@`809RHX00@0;Rh`02XZ;00Z:R`09bLY0P0g=cT0100f=SP08b<U02<S9@0U9BL202XZ;00402/[
;@0T92H092@V02PX:P80;Rh`00@071`N00H620061PP03@d?0P0O7b40100R8R@06a/M01/K7@0Q8B<2
01lO8@0402LW:@0N7R007QhP01dM7`8051@F00@02`/=02LW:@0W9bT05QHH0P16ATP0100A4A<08R8T
028R900`<38202XZ;004020V900Z<2h0:S0^08R>S0<0Z:j/0P2TZZP0102d^[P0[;:`0:bb/02]/jl2
0:ne/@<0/KNc00@0/;Jb0:ne/@2_]K40/KNc0P2c^KD0102n`k`0_l:k0;o2^`2BTHd203lk>004020L
6`0Z:2T0:RPY03Tk>P809b/Z00@08B4Q02d];@0];Bd071`L0P0h>3X0100E51T07A`Q01dL8@0m?482
03<b>00302d[;P0B41<04Q0C00<09B<V00@0CD]>024O8P0Q7b80@d540P0S8B@0100L6Qd0@Cm2044o
@P0l>Sd202d[;P0402/Y;00X9RT0:2HY04Q6B@80=38e00@0:RX/034a<`0a<C<07AdO0P0n?T00101J
FU`0<S8d038b=00T92H2034a<`0401TI6`0d=3H0=3@f02/[;@80ADE700@0?cm102DU9`0U9BL0?Si0
0P0c<cD0100X:2X0>c/m03/k?@0h>3X203@d=P04034a<`0];Bl0;Bd_02l_<@809RHX00@0;2`^02HV
:00V9RP0@4120P0`<380100V9RP0;2`^02`/;P0^;S02020P8P0405ABD`0a;c00<Bl`02HT9@80<2h_
00@0;B//03Le=P0g=CH0A4930P0_;Bh0100f=3D0=38c03@b<`0[:BX205UGF00403Hd=@0b<340<S0a
038`<@80;R`]00@0=38c0490@@12@440@T110P0Y9bP0100d<S<0=C<d03Dc=00b<34203Pf=`0404a:
B`0V92D09R@U02PV9`809R@U00@0>3Hg03lm?P0o?Ch0=C<d0P0d<S<0100[:BX0=38c03@b<`0i=cP2
03/i>P0402HT9@15@d@0AD=403Dc=0808b4R00@0<c4b040n?`10?Sl0De5B0P0a;c00100o?Ch0=cDf
03Le=P0_;Bh203Dc=00405ECE00l>S/0?3Xk02PV9`80;bd^00@0:bTZ02PV9`0X9RL0C4Y;0P0b<340
100^;2d0@Cm0044o@00f=3D203lm?P0403/i>P0n?3d0?S`m04M5AP80EE=D00@0=S@e03<a<P0c<C80
ATA50P11?d00100o?Ch0>CLh03Tg>00m>c`203@b<`0404i=B`1;BTP0BdY803<b<080:BPV00@0<S4_
034`;P0a<2h0=SDc0P0_;R`0100g=CH0>CLh03Tg>00j>3T203<a<P0403Hd=@1BD540DU1A02/Y:P80
=38c00@0:BLX02`Z:`0/:R/0ATA50P0j>3T0100a;c00=S@e03Hd=@0Z:2T203Pf=`04038`<@0c<C80
<c4b044o@080CD]<00@0>SPi03dk?00m>c`0?S`m0P0o?Ch0100c<C80BTQ904Y8B@0d<S<2030^;`04
03/i>P1GEEH0EeEF044o@080=cDf00@0=C<d02d[;00]:b`0=cDf0P1=Bd`0100a;c00:bTZ02/Y:P0a
;c0202TW:005030^;`1@CTl0D4i?044o@00[:BX00@3oool0000404IDG`17EF00AeEP04UGHP80BUQS
00<0BeUT04YHH`1:F6<00`19Ef8404MEH0<0B5IQ1@19Ef801012E680BUaZ04YLJP1;FF@204aIHP04
051IH010ADT0@4E9024U:0804a@F00@03@`:010?3@0@3`d02`X80P0C4Q00100<2`T051<A01@C4@0>
3@/200l>300401dM6`0];B/0;Bd[03Hf=080<C4_00@0;Bd[02`/:P0/;2X09bLU0P0O7ad0100L71X0
>SXh03Xj>00E5A<202DU8`0402h^;00U9B<09BDS02PX9P808b<Q00@0820P01LG5`0G5aL0<30`0P0/
;2`0100O7al0;Bd]02d];@0D51@203De=@0403/k>`0M7Ad07AdM02d];@80:RXZ00@0:2PX04=3@`13
@d<0?3`l0P0/;2`0100Z:R`08B4S024Q8`0R8R@20451@`0401hN800J6Q`06QXL01hN80809RHX00@0
7alQ03Ti>`0i>C/09RHX0P0X:2X0100/;2h06a/M01/K7@0^;S020492A004020P8P0T92H092@V02`/
;P8092@V00@0>3Pj030`<P0`<3809BDW0P0O7b40100M7Al0<C4c034a<`0^;S0201XJ700401lO8@0^
;S00;Rh`02h^<0809BDW00@0;bla01PH6P0H61X0Bd]=0P061PP0100T92H09bLY02LW:@0K6ad202XZ
;00403`l?P0f=SP0=SHh02h^<080;2`^00@0:RX/03`l?P0l?3h0B4Q:0P0S8bD0100Z:R`092@V02@T
9P0N7R02030`<P04034a<`0P8280820R028R908092@V00@0<30b02PX:P0X:2X0<C4c0P0X:2X01008
20X01@D700D51`0L71h202DU9`0401XJ700T92H092@V02<S9@804Q8D00@08B4S01LG6@0G5aT04A4C
0P0=3@l0100A4A<0:BT[02TY:`0Y:B/201PH6P0403@d=P0a<C<0<C4c01lO8@80:RX/00@0>Clm03]1
?`0k@Cl0YJ^Y0P2`][@0102a]kD0ZJn]0:V_[@2f_;X20:J/ZP040:Z`[02]/jl0[K>_0:ne/@<0/;Jb
0P2a]k<0102c^KD0][bh0;Jl^02`]Jh20;bo^0040;nn^P1YIF80JFER01/G5P808R0Q00@0ATQ703@h
=`0d>3L0<S@c0P0Z;2/0100N81l07b0R01lP8P0h>Cd202DV:P0401<D6@0d=CX0=3Dj01lM8080=S@g
00@07adP020N8@0P7R40F5II0P0k>C`0100b<3<0<2ha030^<@0a;c8202h/;`0405MEF00]:bh0;B/^
02<Q9080<Blb00@0;R`_04i<C`1>C4l09RHX0P0g=cT00`0];Bl0;bla02l_<@0304M7B@0402l_<@0J
6Q`06QXL03/k?@809BDW00@05QHH03hn@00n?T00DU9D0P0k>cd0100d=3H0:b/]02/[;@0U9BL203Ph
>P0403Xj?00h>3X0>3Pj030`<P80;2`^00@0=CDg034a<`0a<C<0:BT[0P0X:2X01012@T@0=CDg03De
=`0Y:B/202XZ;00403De=`0S8bD08b<U04e;C080<Bl`00@08R0Q02d[;00]:b`09B<T0P0]:b`01015
@d@0:RPY02XX:@0W9BH202XX:@04024O801;BDX0BdU:03Hd=@80;bd^00@0=38c02l];P0_;Bh0=cDf
0P0d<S<01014@T<0:2HW02PV9`0j>3T203dk?004040n?`11?d00@Cm005eKG080;2X[00@0;R`]02TW
:00Y9bP0=C<d0P0n?3d0100b<340;2X[02`Z:`0]:b`2030^;`0403Le=P0j>3T0>SPi01`J6`80@d52
00@0=cDf024O800Q7b00:bTZ0P0Y9bP0101CDE80;B//02d[;00m>c`202d[;00402TW:00_;Bh0;bd^
051>C`80>SPi00@0928S03Le=P0g=CH0<S0a0P0X9RL0101FE5D0>SPi03Xh>@0O7Ah2034_<00402<Q
8P0/:R/0;2X[02d[;080<S0a00@0=C<d0490@@12@440;bd^0P0/:R/0100n?3d0@d5204=1@P14@T<2
03Pf=`04040n?`0`;Rl0<2h_03Dd<P80FEQF00@0?3/i01lN700O7Q`0?chl0P0n?C/0100b<Bl0;2/Y
02`[:@0d<S<203Dc=004034_<00/:R/0;2X[02@R8`80CD]<00@08R0Q034_<00a;c0081hO0P0X9RL0
100g=CH0>cTj03/i>P0c<C8202d[;00402h/;@0d<S<0=38c030^;`80=38c00@0=S@e055?D01ACe00
>3Hg0P0i=cP0100d<S<0=cDf03Le=P0_;Bh204]9BP0402`Z:`0Z:2T0:RPY02d[;080E59C00@0?3Xk
03Xh>@0j>3T0;2X[0P0a;c00100]:b`0Cde>04m=CP19AdP202`Z:`0402/Y:P0o?Ch0?cdn01lM7P80
AdE60@12@44103Tg>040oooo00001016Df<0AE9R04EBHP14DF46049?G`80@e1P0`11Df43049DHP03
04=EH`14EV@0A5IT00<0A5IT00@0B5YX04ADH@14E640AeEP0P1:Ef00101:E5d0Ae1G04M@E`0f?D<2
01XQ9`04010A4`0?41803a0B00l@4P80414C00@04A8D010A4`0@4A<03@h@0P0:2`d0100B4Q008B4O
024Q7`0_;bd203<c<@04038b<00`<2h0<30^02XZ:0808R8P00@0820N02TY9`0Y:BL061PF0P0L71X0
100S8b40:2PV02PX9P0e=C<201dM6`0402PX:00T92@092@T02PX:080<S8b00@0;Rh^028R8P0R8R80
8B4Q0P0/;2`01012@T80:b/[02/[:`0^;Rh202l_;`04020P800_;bl0;bl_03dm?@808R8R00@07AdO
02XZ;00Z:R`03`lA0P0/;2h0100^;S008B4S024Q8`0S8bD202@T9P0402DU9`0g=cT0=cLi03<c=@80
8R8T00@0;Bd_01hN800N7R00:b/]0P10@480100b<S@09bLY02LW:@0g=cT201DE5`0402`/;P0k>cd0
>c/m01`L7P808R8T00@09bLY01/K7@0K6ad0;Bd_0P0T92H0100H61X071`N01`L7P0d=3H2028R9004
01/K7@0O7b407alQ03Ph>P8071`N00@051@F02HV:00V9RP0<S8d0P0M7Al0100k>cd092@V02@T9P0Y
:B/202HV:0<071`N00@0ADE7018B500B4Q@092@V0P0Q8B<0100L71h09RHX02HV:00k>cd201`L7P04
024Q8`0J6Q`06QXL02DU9`80<c<e00@07QhP00/;3@0;2`d030`>0P0S8bD0100Q8B<0820R020P8P0J
6Q`201lO8@0402<S9@0I6A/06ATK01lO8@805aLI00@00@4301PH6P0H61X06QXL0P0O7b400`0Y:B/0
=cLi03Lg>@0302XZ;00402h^<00i>C/0>CTk03@j>080GVAR00@0Zk6_0;>i]@2c^KD0/[Vb0P2//j`3
0:ZaZ@040:V`Z02Z/JT0Z[6Y0:^bZP80[KFZ00@0[kN/0;2hZ`2`^:/0/KV/0P2b^Z/0102na;X0Z[2V
0:Z`YP2l`KX209ZOV00403e2?00F61<05QPC034c<080DE=@00@0@41002LW9`0W9bL0;bla0P0P8280
100Q82D0@3m4040oA00A41H201TH7P0400D92P0R8bD08R<U02`];`806a/M00@0=3@f04Y8B`1:B4/0
;R`_0P0]:2`0100X9RT0:bT/02/Y;01>C4l203lm@00402`Z;@0]:bh0;B/^02<Q9080@Cm200@0>c/m
03@d=P0d=3H09RHX0P0m?Cl0100I6A/0DU9D059BE00`<38203<c=@0402`/;P0`<380<30b02PX:P80
8B4S00@0FEUK04Y:C01:BT`09RHX0P0b<S@0100X:2X0;2`^02`/;P11@D<20410@P0402/[;@12@T@0
@T94034a<`80:b/]00@0:RX/02<S9@0S8bD08B4S0P1@D580100T92H0:b/]02/[;@0Z:R`202PX:P04
03Xh>@0i=cP0>CLh034_<080=38c0`0^;2d0101ACe00;bd^02l];P0b<34203Dc=00402h/;@0[:BX0
:bTZ06=QHP80=C<d00@0<c4b03dk?00m>c`0?S`m0P0k>CX0101YIfP0?C/l03dk?00k>CX203Dc=004
034_<00c<C80<c4b04m=CP80<c4b00@06QPI02TW:00Y9bP0;2X[0P1=Bd`0100f=3D0>3Hg03Pf=`0^
;2d2030^;`0402`Z:`0l>S/0?3Xk03Dc=0809B<T00@0AD=402PV9`0X9RL0:2HW0P0U8b@0101=Bd`0
>CLh03Tg>00Z:2T203dk?00402d[;00g=CH0=cDf03hl?@80AdE600@0=38c030^;`0`;Rl0=38c0P0j
>3T0100e<c@0FEMH05UGF00^;2d202HT9@0402<Q8P0b<340<S0a034_<080;2X[00@0:BLX04Q6A`18
ATL0:bXX0P0_;R`0100/:bT0D4m=051?C@0e=383038a;`80>cXh00@0;B`Z055@CP1AD4h0>STg0P0_
;R`01011@3h0?3/i03`k>@0h=cD203Xi=`0403dk?00]:b`0;B//034_<0<0;bd^0P0]:b`00`0o?Ch0
;B//02d[;00302XX:@0402<Q8P19AdP0BDM802l];P80;R`]00@0:RPY030^;`0`;Rl0<S0a0P0j>3T0
100]:b`0BDM804U7B00a;c0203Le=P0402PV9`0/:R/0;2X[03Tg>080=S@e00@0@3ho02d[;00]:b`0
;B//0P0m>c`01010?Sl0=S@e03Hd=@0d<S<203Hd=@0402l];P0`;Rl0<2h_05IDE@80;2X[00@0>cTj
03hl?@0n?3d0:2HW0P0i=cP104I4A@40AdE60@3oool0000404ICH`15DV80AE9R04AAH@H0@TmO0P13
D603045CH@<0@UAR00<0@eES04AFI014EV@00`14EV@01018FVP0A5AQ04ADH@17EF0204YGH00404YD
G@17D5L0Ae1G03Hm@`806R4W00@0414C00l@4P0?41803a0B0P0@4A<0100A4Q@0414C010A4`0=3Q02
00X;3@04018B400Q8Al08B4O02l_;@80<c<a00@0<S8`030`;P0`<2h0:RXX0P0R8R00100P81h0:BTW
02TY9`0H61H201`L6P0402<S8@0X:2H0:2PV03De<`807AdK00@0:2PX02@T900T92@0:2PX0P0b<S80
100^;Rh08R8R028R8P0Q8B4202`/;0040492@P0[:b/0:b/[02h^;P80;bl_00@0820P02l_;`0_;bl0
?Cdm0P0R8R80100M7Al0:RX/02XZ;00?3a4202`/;P0402h^<00Q8B<08B4S02<S9@8092@V00@09BDW
03Lg>@0g=cT0<c<e0P0R8R@0100];Bl07QhP01hN800[:bd20410@P04038b=00W9bT09bLY03Lg>@80
5ADG00@0;2`^03/k?@0k>cd071`N0P0R8R@0100W9bT06a/M01/K7@0];Bl202@T9P0401PH6P0L71h0
71`N03@d=P808R8T00@06a/M01lO8@0O7b40>3Pj0P0L71h0100D51H09RHX02HV:00b<S@201dM7`04
03/k?@0T92H092@V02TY:`809RHX0`0L71h01015ADL04Q8D018B500T92H2024Q8`0401`L7P0V9RP0
9RHX03/k?@8071`N00@08B4S01XJ700J6Q`09BDW0P0c<cD0100N7R002`/=00/;3@0<30h202<S9@04
024Q8`0P8280820R01XJ70807alQ00@08b<U01TI6`0I6A/07alQ0P0G5aT010010@<061PJ01PH6P0J
6Q`201lO8@0302TY:`0g=cT0=cLi00<0:RX/00@0;Rh`03Ti>`0i>C/0=3Xh0P1NI680102[/Jl0/kVe
0;>i]@2b^K820:bc[0<0Z[6Y00@0ZK2X0:ZaZ@2Z/JT0Zk:Z0P2]]JX0102_]j`0/;R[0;2hZ`2a^J`2
0;:jZ`040;k4^P2Z/:H0Z[2V0;c1^P80VYnH00@0?D8l01HH4`0F61<0<C<`0P1ADe001010@4009bLW
02LW9`0_;c42020P8P04024P9@10?d@0@3m4014@5P806APN00@01@T:028S9@0R8bD0;2d_0P0K6ad0
100d=3H0BTQ;04Y8B`0^;2l202dX;00402PV:@0[:B`0:bT/04i<C`80?ce000@0;2X]02d[;P0]:bh0
8b4T0P11?d80100k>cd0=3@f03@d=P0V9RP203dm?`0401TI6`1BDU@0DU9D030`<P80<c<e00@0;2`^
030`<P0`<380:2PZ0P0Q8B<0101IFE/0BTY<04Y:C00V9RP2038b=00402PX:P0/;2h0;2`^0451@`80
@41200@0:b/]0492A012@T@0<C4c0P0[:bd0100Z:R`08b<U02<S9@0Q8B<2051@DP0402@T9P0[:bd0
:b/]02XZ;080:2PZ00@0>SPi03Tg>00i=cP0<Bl`0P0d<S<302h/;@04055?D00_;Bh0;bd^038`<@80
=C<d00@0;R`]02/Y:P0[:BX0Hf5R0P0e<c@0100c<C80?C/l03dk?00n?3d203/i>P0406UWJ00m>c`0
?C/l03/i>P80=C<d00@0<Bl`03<a<P0c<C80Cde>0P0c<C80100J61T0:BLX02TW:00/:R/204e;C004
03Hd=@0h=SL0>3Hg02h/;@80<2h_00@0;2X[03`j>`0l>S/0=C<d0P0U8b@01015@d@0:2HW02PV9`0X
9RL202DS900404e;C00i=cP0>CLh02XX:@80?C/l00@0;B//03Le=P0g=CH0?S`m0P17ADH0100d<S<0
<2h_030^;`0d<S<203Xh>@0403Dc=01IEeP0FEMH02h/;@809R@U00@08b4R038`<@0b<340<Bl`0P0/
:R/0100Y9bP0B4I704Q6A`0[:RP202l^;00402`[:@1@Cdd0D4m=03Dd<P<0<S4_0P0k>SP0100];2X0
DE1>055@CP0j>CL202l^;0040450?P0l>cT0?3/i03Pg=@80>STg00@0?C/l02d[;00]:b`0<Bl`0`0_
;Bh202d[;00303lm?P0]:b`0;B//00<0:RPY00@08b4R04U7B019AdP0;bd^0P0^;2d0100Z:2T0<2h_
030^;`0b<34203Xh>@0402d[;019AdP0BDM8034_<080=cDf00@0:2HW02`Z:`0/:R/0>CLh0P0f=3D0
1010?Sl0;B//02d[;00]:b`203dk?004040n?`0f=3D0=S@e03@b<`80=S@e00@0;bd^030^;`0`;Rl0
EUAE0P0/:R/0100k>CX0?S`m03hl?@0X9RL203Tg>040ATA50@17ADH10?ooo`0000@0A55Q04=@H013
D600@TmO0P11CUh3041=G@<0@DiN00<0@59P045CH@11Df401P11Df43049DHP04045AGP0nCU/0?TiK
04=AG080B5EN0`19De`0100oB4l0:S<j02Xc>P0E6A`2010D5`<02`lB00@03a0B014B500A4Q@03PlA
0P0;30h0100=3@/06ATG01TI5`0V9R@202`/:P<0<30^00@0:BTW020P7P0P81h08R8P0P0/;2X0100H
61H06a/I01/K6@0U9B<202PX9P0403De<`0X:2H0:2PV02XZ:P809bLW00@0:BTY030`<00`<300;bl_
0P0U9BD0100Q8B4092@T02@T900g=cL203<c<`0302<S8`0[:b/0:b/[00<08R8R00@0>c/k028R8P0R
8R807alQ0P0R8R@0100K6ad0820R020P8P0h>3X201dM7`0402@T9P0S8bD08b<U020P8P80<C4c00@0
>3Pj02DU9`0U9BL0:b/]0P0V9RP0100];Bl0>CTk03Ti>`0e=CL201hN800402`/;P0H61X061PJ02HV
:080>3Pj00@0:BT[01/K7@0K6ad0:2PZ0P0M7Al0100];Bl09RHX02HV:00I6A/201`L7P03030`<P0P
8280820R00<071`N00@0<S8d01TI6`0I6A/03Ph@0P0K6ad0100W9bT06QXL01XJ700g=cT202d];`04
02TY:`0X:2X0:2PZ01lO8@808b<U00@0@41201lO8@0O7b409BDW0P0T92H0100R8R@09BDW02DU9`0j
>S`202LW:@04024Q8`0O7b407alQ02XZ;080:BT[00@0410B00T92`092@/051@F0P0[:bd0100Q8B<0
6QXL01XJ700N7R0201XJ70<06a/M00@06QXL00h>400>3Q001`L90P0I6A/0100S8bD09RHX02HV:00Y
:B/2034a<`0402LW:@0c<cD0<c<e03Lg>@80>CTk00@0:S0^06Y`KP1ZL6h0[;:`0P2a]k<0102c^[<0
Zk:[0:^bZ`2Y/:P20:R_Y`<0[[F]00@0[kJ^0:ng[02_]j`0[kN/0`2_]jX20:ngZ0<0/KN]00@0_<:h
0;bo^02l_kP0KW5Z0P0N81/0100T9R<0?d4n03m1?P0]:b`203Dc=00401/K6`0^;S00;Rh`02TX;@80
;B`a00@0>CXo024R9`0Q8RL04QHG0P0b<cD0100M7R00;Bd_02d];`0^;S0202h/;`04055?DP0_:Rh0
;bX^02h/;`80>cTl00@09bDX03lm@00o?D00?S`o0P0U8bH0100e<cH09B<V02DS9P1@D58203dm?`04
03@d=P0L71h071`N03Lg>@80:b/]00@0@d=50410@P10@4808B4S0P0Y:B/0100_;c40@D530451@`0V
9RP205ADEP0403hn@00U9BL09BDW02`/;P80<S8d00@0=cLi03dm?`0m?Cl0ADE70P0a<C<0100n?T00
8B4S024Q8`0];Bl202XZ;00402/[;@1BDU@0DU9D02HV:0809bLY00@0:BT[02XZ;00Z:R`0=38c0P10
?Sl00`0Z:2T0<Bl`034_<00302/Y:P0404M5AP0a;c00<Bl`03Dc=080>3Hg00@0<c4b030^;`0`;Rl0
Hf5R0P0k>CX0100d<S<0?3Xk03`j>`0`;Rl202d[;00405IDE@0b<340<S0a02h/;@<0;2X[0P0_;Bh0
0`1=Bd`0<S0a038`<@0302HT9@0402h/;@1GEEH0EeEF03Dc=080?S`m00@0<Bl`038`<@0b<340:2HW
0P0`;Rl0100^;2d09bDV02LU9P15@d@202DS900402d[;00X9RL0:2HW04U7B080?cdn00@0;R`]03Le
=P0g=CH0;B//0P0a;c000`0d<S<0BDM804U7B00302TW:00403@b<`0[:BX0:bTZ02XX:@80De5B00@0
<2h_02PV9`0X9RL09bDV0P0[:BX0100W9BH0:bTZ02/Y:P0Z:2T204e;C00402HU8`0Z:BL0:RTW02PW
9@80BdY800@0=C@b034`;P0a<2h0<2l]0P0g=S@0100U9280@T4o0491?`0e=38202HU8`03038a;`0e
=380=C@b00<0<S4_00@0<c4b02l];P0_;Bh0:RPY0P0`;Rl0100c<C80;2X[02`Z:`18ATL203Le=P04
02h/;@0Z:2T0:RPY02d[;080?C/l00@0@d5202h/;@0^;2d0<Bl`0P0b<340100[:BX0?C/l03dk?00Y
9bP2044o@00303Xh>@0b<340<S0a00<0;R`]00@0>3Hg030^;`0`;Rl0BDM80P0_;Bh0100e<c@0<c4b
03<a<P17ADH203lm?P0402`Z:`0]:b`0;B//03Hd=@80;bd^00@0C4Y;04=1@P13@D80?3Xk0P0j>3T3
034_<040?3Xk0@1;BDX10?ooo`0000@0@e1P049?G`12Cel0@DiN0P10CEd0100oC5`0@4eM041=G@10
CEd2045>GP<0@E=Q1`10DV0203mAG`0403i>F`0kBeP0>d]H049@F`80AE=N00@0AE9K04YDG@1:E5d0
Ae5J0P0jA4d0100R9b/05a`P01LL800:3a<200L;3P0400/?4P0A4Q@04A8D010A4`803Ph>00@020P6
00l?3@0?3`d061PF0P0R8R00100[:bT0;bl]02l_;@0Y:BL2020P7P0402@T8P0^;R`0;Rh/01PH5P80
6QXH00@09bLU02HV900V9R@0<30^0P0a<Bl0100W9bL09BDU02DU9@0U9BD202PX:00402/[:`0Z:RX0
:RXZ02@T90807alO00@0:2PX03Xj>P0j>SX06a/K0P0V9RH0100T92@05QHF01HF5P0f=SH202/[:`04
02LW:@0N7R007QhP02TY:`806ATK00<0@T94028R900R8R@00`0V9RP0100W9bT0<C4c034a<`11@D<2
02`/;P0402DU9`0V9RP09RHX028R908092@V00@0?Cdo01lO8@0O7b4092@V0P0O7b40100P8280<30b
030`<P0c<cD2018B500402LW:@0M7Al07AdO02d];`80:RX/0`0M7Al0100/;2h0820R020P8P0Y:B/2
02DU9`0403Lg>@0U9BL09BDW01LG6@8071`N00@09BDW02@T9P0T92H0:2PZ0P0b<S@0100Q8B<08R8T
028R900K6ad202DU9`04030`<P0/;2h0;2`^020P8P808B4S00@08R8T01`L7P0L71h0<C4c0P0^;S00
100H61X07QhP01hN800a<C<201lO8@04008210092@/02@T;01hN8080<30b00@07AdO00h>400>3Q00
8B4S0P0G5aT0100F5QP07AdO01dM7`0@418200@41P0401<C5@0L71h071`N02PX:P80:RX/00@0:b/]
02l_<@0_;c409RHX0`0f=SP202`/;P0402@Z:022R8H0PXR60;2f]080[K>_00@0]Kbe0:V`Z@2Y/:T0
ZK2X0P2W[ZH30:je[@040:nf[P2_]j`0[kN/0:ng[0<0[kNZ0P2_]jP0102^]:P0][b`0;Jl/02_]J/2
0<G8_`0409BGT00Y:RD0:BXU01HG4P80<c8^00@0;2PW040l?@10?3d08R0Q0P0`;S40100P82809RDZ
02HU:P16Ad/2018C5`0402<W:00m?T00?Ci0020Q8`80<C4c00@0;2`^02@R9@0T8RD0EE=F0P12?D40
100h=ST08R0S028P8`0^;2l2038`<`0403/i?00n?3l0?S`o02@R9@80:2HY00@0BTY<051@DP1@D580
9bLY0P0^;S00100_;c409BDW02DU9`0d=3H204I6B00403`l?P0L71h071`N04A4AP80<30b00@0:b/]
034a<`0a<C<0EEEG0P0Z:R`0100];Bl0<C4c034a<`0i>C/202d];`0405UIF`0b<S@0<S8d03Lg>@80
:RX/00@0>SXl03`l?P0l?3h0?Si00P1NGV00100h>3X0=3@f03@d=P0i>C/203`l?P0403Le=P1DDU<0
E59C02d[;080>CLh0`0c<C801015@d@0@3ho040n?`0_;Bh3034_<080;B//00@0EUAE03Tg>00i=cP0
;B//0P0a;c00100`;Rl0;B//02d[;01>C4d203Hd=@0402/Y:P0^;2d0;R`]02h/;@80<S0a00@0@d52
03<a<P0c<C80;bd^0P0N71d0100]:b`0FEMH05UGF00c<C8203lm?P04034_<00d<S<0=38c02/Y:P80
;B//00@0<Bl`02l];P0_;Bh0A4930P0V92D0100^;2d0:2HW02PV9`0o?Ch204M5AP0403Le=P0a;c00
<Bl`034_<080;bd^00@0=S@e05ABD`1DDU<0=S@e0P0e<c@01014@T<0=S@e03Hd=@0g=CH2061NG`04
03@b<`0X9RL0:2HW02`Z:`80:2HW00@08b4R02h/;@0^;2d0:2HW0P1@CTl0100S8R009RDS02HU8`0U
928204A3@@0403Th=P0b<Bl0<S4_030_;@80<c8`00@0;bh/040o?@10?cd0@D0n0`0[:RP203Pg=@04
03@c<@0`;bd0<2l]02l];P80=C<d00@0:bTZ038`<@0b<340=C<d0P0[:BX01016A4D0?S`m03hl?@0[
:BX202DS900402`Z:`0Y9bP0:BLX04i<C@80:bTZ00@0;bd^02h/;@0^;2d09bDV0P0n?3d0100Z:2T0
=S@e03Hd=@15@d@3030^;`<0<S0a0P0W9BH01019AdP0=cDf03Le=P0d<S<202DS900404E3A017ADH0
AdE602XX:@80:bTZ00@0?3Xk034_<00a;c00?S`m0P1GEEH0100m>c`0<2h_030^;`0j>3T202`Z:`40
=38c0@1=Bd`10?ooo`0000@0@e1P049?G`12Cel0@DiN0P10CEd0100oC5`0@4eM041=G@10CEd2045>
GP<0@E=Q1`10DV0203mAG`0403i>F`0kBeP0>d]H049@F`80AE=N00@0AE9K04YDG@1:E5d0Ae5J0P0j
A4d0100R9b/05a`P01LL800:3a<200L;3P0400/?4P0A4Q@04A8D010A4`803Ph>00@020P600l?3@0?
3`d061PF0P0R8R00100[:bT0;bl]02l_;@0Y:BL2020P7P0402@T8P0^;R`0;Rh/01PH5P806QXH00@0
9bLU02HV900V9R@0<30^0P0a<Bl0100W9bL09BDU02DU9@0U9BD202PX:00402/[:`0Z:RX0:RXZ02@T
90807alO00@0:2PX03Xj>P0j>SX06a/K0P0V9RH0100T92@05QHF01HF5P0f=SH202/[:`0402LW:@0N
7R007QhP02TY:`806ATK00<0@T94028R900R8R@00`0V9RP0100W9bT0<C4c034a<`11@D<202`/;P04
02DU9`0V9RP09RHX028R908092@V00@0?Cdo01lO8@0O7b4092@V0P0O7b40100P8280<30b030`<P0c
<cD2018B500402LW:@0M7Al07AdO02d];`80:RX/0`0M7Al0100/;2h0820R020P8P0Y:B/202DU9`04
03Lg>@0U9BL09BDW01LG6@8071`N00@09BDW02@T9P0T92H0:2PZ0P0b<S@0100Q8B<08R8T028R900K
6ad202DU9`04030`<P0/;2h0;2`^020P8P808B4S00@08R8T01`L7P0L71h0<C4c0P0^;S00100H61X0
7QhP01hN800a<C<201lO8@04008210092@/02@T;01hN8080<30b00@07AdO00h>400>3Q008B4S0P0G
5aT0100F5QP07AdO01dM7`0@418200@41P0401<C5@0L71h071`N02PX:P80:RX/00@0:b/]02l_<@0_
;c409RHX0`0f=SP202`/;P0402@Z:022R8H0PXR60;2f]080[K>_00@0]Kbe0:V`Z@2Y/:T0ZK2X0P2W
[ZH30:je[@040:nf[P2_]j`0[kN/0:ng[0<0[kNZ0P2_]jP0102^]:P0][b`0;Jl/02_]J/20<G8_`04
09BGT00Y:RD0:BXU01HG4P80<c8^00@0;2PW040l?@10?3d08R0Q0P0`;S40100P82809RDZ02HU:P16
Ad/2018C5`0402<W:00m?T00?Ci0020Q8`80<C4c00@0;2`^02@R9@0T8RD0EE=F0P12?D40100h=ST0
8R0S028P8`0^;2l2038`<`0403/i?00n?3l0?S`o02@R9@80:2HY00@0BTY<051@DP1@D5809bLY0P0^
;S00100_;c409BDW02DU9`0d=3H204I6B00403`l?P0L71h071`N04A4AP80<30b00@0:b/]034a<`0a
<C<0EEEG0P0Z:R`0100];Bl0<C4c034a<`0i>C/202d];`0405UIF`0b<S@0<S8d03Lg>@80:RX/00@0
>SXl03`l?P0l?3h0?Si00P1NGV00100h>3X0=3@f03@d=P0i>C/203`l?P0403Le=P1DDU<0E59C02d[
;080>CLh0`0c<C801015@d@0@3ho040n?`0_;Bh3034_<080;B//00@0EUAE03Tg>00i=cP0;B//0P0a
;c00100`;Rl0;B//02d[;01>C4d203Hd=@0402/Y:P0^;2d0;R`]02h/;@80<S0a00@0@d5203<a<P0c
<C80;bd^0P0N71d0100]:b`0FEMH05UGF00c<C8203lm?P04034_<00d<S<0=38c02/Y:P80;B//00@0
<Bl`02l];P0_;Bh0A4930P0V92D0100^;2d0:2HW02PV9`0o?Ch204M5AP0403Le=P0a;c00<Bl`034_
<080;bd^00@0=S@e05ABD`1DDU<0=S@e0P0e<c@01014@T<0=S@e03Hd=@0g=CH2061NG`0403@b<`0X
9RL0:2HW02`Z:`80:2HW00@08b4R02h/;@0^;2d0:2HW0P1@CTl0100S8R009RDS02HU8`0U928204A3
@@0403Th=P0b<Bl0<S4_030_;@80<c8`00@0;bh/040o?@10?cd0@D0n0`0[:RP203Pg=@0403@c<@0`
;bd0<2l]02l];P80=C<d00@0:bTZ038`<@0b<340=C<d0P0[:BX01016A4D0?S`m03hl?@0[:BX202DS
900402`Z:`0Y9bP0:BLX04i<C@80:bTZ00@0;bd^02h/;@0^;2d09bDV0P0n?3d0100Z:2T0=S@e03Hd
=@15@d@3030^;`<0<S0a0P0W9BH01019AdP0=cDf03Le=P0d<S<202DS900404E3A017ADH0AdE602XX
:@80:bTZ00@0?3Xk034_<00a;c00?S`m0P1GEEH0100m>c`0<2h_030^;`0j>3T202`Z:`40=38c0@1=
Bd`10?ooo`000`13D6000`12Cel0@DiN045>GP03045>GP03049?G`13D600@e1P00<0@eES0`12E680
0`11Df40@59P041BH003041BH004049BG`11CeX0@DmJ04ABG@80AUAO00@0@DmJ04ABG@14DUd0B5EN
0P12CeP0100c?4<08R/b028[<P0?5Q`200H;3`0400P<3`0>3a403PlA010@40803`l?00@02PX800T9
1`092@L03@d;0P0G5aD0100T9280;2`Z02`/:P0[:bT302DU8`80;bl]00@06QXH01/K6@0K6aT0:BTW
0P0S8b400`0X:2H0<C4_034a;`03028R8P<07alO00@09RHV02l_;`0_;bl0:b/[0P0R8R800`0M7Ad0
>CTi03Ti>@0302@T900402DU9@0D51@051@D02`/;080?Shn00@0;bla02<S9@0S8bD0;Bd_0P0P8280
100n?T00<C4c034a<`0U9BL202/[;@0402DU9`0Q8B<08B4S03Ti>`80;bla00@08b<U02LW:@0W9bT0
820R0P0T92H0100h>3X0:2PZ02PX:P0Q8B<202HV:00402<S9@0/;2h0;2`^03/k?@807QhP00@092@V
01`L7P0L71h0:BT[0P0^;S00100P82807alQ01lO8@0Z:R`202<S9@0402LW:@0P8280820R034a<`80
:RX/00@06QXL01LG6@0G5aT06a/M0P0S8bD00`0P8280>SXl03Xj?003024Q8`0401dM7`0[:bd0:b/]
02DU9`80?3`n00@0:RX/02`/;P0/;2h0;Rh`0P0R8R@0100d=3H0@D530451@`0O7b4202TY:`04038b
=00F5QP05QHH00000P803`lA00@092@V02d];`0];Bl05ADG0P071`T0100N7R0071`N01`L7P0L71h2
01lO8@0300@41P000080000200<08b<U00@09BDW02HV:00V9RP0;Bd_0P0e=CL0100V9RP0;Rh`02h^
<00[:bd201hN800402`b<02MXj40WJ>Q0;Bj^080Z:jZ00<0/kZc0:J]YP2V[JH01P2X[jL0102Z/JT0
Zk>X0:^cZ02]]JX20:ngZP040;2hZ`2a^JX0/KVZ0;>k[P80][b`00@0Z:jR0;^n]@2k_[D0[k2X0P1>
CDP0100E4@h0;RXW02hZ9`0g;Rl203<];@04040j>P0Z9RL0:RHW014?4P80<30b00@0;Rlc00@52@04
1@T0;S8c0P0h>C/0100g>3X0:2PZ02PX:P0`<38203@b=@0402l]<01GDUH0Ee9F02d[;P80:RP[00@0
9R@W03Hd=`0f=3L0>cTl0P0a;c8303@b=@04038b=00i>C/0>CTk03hn@080;bla00@0?3`n01DE5`0E
5AL0;Bd_0P0d=3H0101IFE/0?3`n03`l?P0V9RP203<c=@0402PX:P0/;2h0;2`^038b=080Bd]=00@0
;bla02/[;@0[:bd0@T940P0R8R@0101=CDl0Bd]=04];C@0T92H203hn@00402TY:`0[:bd0:b/]02`/
;P80ATI800@0<30b02TY:`0Y:B/0;Rh`0P0_;c40100/:R/0EE=D05ECE00V92D2034_<00402`Z:`0[
:BX0:bTZ03<a<P80@3ho00@0:RPY02/Y:P0[:BX0;bd^0P0[:BX01019AdP0>cTj03/i>P0X9RL202TW
:00403Le=P0d<S<0=38c04]9BP80AD=400@0<2h_03@b<`0d<S<0<Bl`0P0b<340100e<c@0?3Xk03`j
>`0/:R/201PF5`0402d[;019AdP0BDM803Le=P80=S@e00@0;B//038`<@0b<340<c4b0P0h=SL0100n
?3d0=C<d03Dc=011?d02034_<00402TW:00V92D09R@U038`<@80C4Y;00@0@Cm002l];P0_;Bh0>3Hg
0P0c<C80100m>c`0CTa=04i<C@1<BT/203Pf=`0403Le=P0c<C80<c4b02PV9`80A49300@0?S`m02PV
9`0X9RL0;2X[0P0[:BX0100Y9bP0<c4b03<a<P0U8b@204m=CP0402DT8P0V9B<09RDS02LV9080@3lm
00@0@D0n03Lf=00g=S@0=C@b0P0c<S001010?cd0ADA204E4@P1@Cdd203He<`0402TX9P0k>SP0>cXh
03He<`80<2l]00@0;2X[02l];P0_;Bh0<c4b0P0b<340100d<S<0<c4b03<a<P0m>c`204Y8B@0403Dc
=00b<340<S0a02h/;@80;bd^00@0E59C03dk?00m>c`0=38c0P0i=cP0100a;c00?cdn03lm?P0i=cP2
02d[;00404Q6A`0d<S<0=38c02HT9@80<2h_00@0<c4b02d[;00]:b`0A4930P1=Bd`0100_;Bh09bDV
02LU9P0m>c`204m=CP0403@b<`0e<c@0=C<d03Tg>080=S@e00@0=38c05ABD`1DDU<0?C/l0P0V92D0
1@0j>3T0;2X[02`Z:`0a;c00A4930040oooo00001015DV80A55Q04AAH@14DF4304=@H080A55Q00<0
AE9R04ICH`16Df<01015EfD204AFI0<0@eES0`12E6801017EF00@U1K049@F`16E5l204MEH00403m?
G010D5d0@51M04IFH`80AEER00@0@e1I03@nA`0d?TL07BH]0P0=51X010093A0030d?00`=3`0>3Ph2
010?3@0400l?3@092@L02@T700L71@803Ph<00@06a/I02DU8`0U9B<0:RXX0P0/;2X0100U9B<0;Rh/
02h^;00M7A/201`L6P0402TY9`0Q8Al08B4O01lO7@80:BTW0`0Z:RX0100T92@07alO01lO7`0U9BD2
034a<@04038b<P0Z:RX0:RXZ01dM7@80<30`00@0<S8b02HV9P0V9RH09BDU0P0L71`0100T92@0CTi>
04i>CP0a<C<202LW:@04024Q8`0[:bd0:b/]02/[;@80?Si000@0820R02XZ;00Z:R`09RHX0P0H61X0
100[:bd0<c<e03<c=@0T92H2028R900401hN800X:2X0:2PZ02TY:`80=3@f00<08b<U02TY:`0Y:B/0
0`0[:bd0100l?3h0=cLi03Lg>@0Q8B<201XJ700402DU9`0`<380<30b02@T9P808R8T00@0:BT[02XZ
;00Z:R`092@V0P0O7b40100_;c40<c<e03<c=@0T92H201`L7P0401PH6P0U9BL09BDW02@T9P80@d=5
00@0;Bd_02PX:P0X:2X0:BT[0P0e=CL0100W9bT0C4a>04a<CP0/;2h302d];`8092@V00@0;Bd_04I6
B016ATP07QhP0P0X:2X0100U9BL02`/=00/;3@00008201TI6`0402HV:00W9bT09bLY01<C5@802`/=
00@06ATK02DU9`0U9BL09RHX0P0L71h0100000801PH800H6200_;c4202TY:`0402@T9P0O7b407alQ
02/[;@80>SXl00@09RHX02PX:P0X:2X0:b/]0P0W9bT01014BTP0/[Rf0;:h]P2f_;X20:J/Z0040:je
[P2SZZ<0XjZS0:N^YP80Zk:Z0`2TZj<0102V[JD0YjnT0:N_Y02Y/JH20:^cYP040:bdY`2]]JH0[KFV
0;2hZ@80Zk>T00@0/KN[0;Jh[@2f^:d0a<Bl0P28QGh0100O6QH08a/H02<K600e:bX202<I600404a3
A00c;bh0<bl^01LE5P80:BT[00@071dO01DI6P0E6AX0<3@e0P0^;c401014ADL0<30b030`<P0e=CL2
03Tg>P0401dK7P1BCE40DTeA03Pf>@80:RP[0`0^;2l0100e<cH0?3Xm03`j?@0d<SD203`j?@04024Q
8`0X:2X0:2PZ04E5A`80@41200@0:2PZ02DU9`0U9BL09BDW0P0a<C<0101:BT`0CdmA04m?D@0Y:B/2
02PX:P0402l_<@0Y:B/0:BT[020P8P80BTY<00@0?Si002h^<00^;S00=SHh0P0`<380100[:bd0EEEG
05EEE`0b<S@203Ph>P0402h^<00];Bl0;Bd_02TY:`80>c/m00@0=cLi02d];`0];Bl0;Rh`0P0Y:B/0
100Z:2T0DDm@055?D00^;2d2030^;`0402`Z:`0Z:2T0:RPY02d[;080AD=40`0]:b`0100e<c@0<2h_
030^;`13@D8204I4A@0302h/;@0/:R/0;2X[00<0<2h_00@0?cdn04i<C@1>C4d0;bd^0P0d<S<0100]
:b`0;2X[02`Z:`0_;Bh204i<C@0402TW:00L6Q/071XK030^;`80<c4b00@0AD=402l];P0_;Bh0:bTZ
0P0]:b`0100d<S<0?3Xk03`j>`12@442030^;`0403Xh>@14@T<0A49302PV9`809bDV00@0:BLX04Y8
B@1:B4T0B4I70P0_;Bh0100g=CH0<c4b03<a<P0h=SL203@b<`0405YHF@0h=SL0>3Hg02DS9080=C<d
00@07adN02PV9`0X9RL0D4i?0P0^;2d0100V92D0:bTZ02/Y:P0^;2d203@b<`0402PV9`1ACe00DDm@
02`[:@<0:bXX0P0m?3X0101;BTP0?C`j03dl>P0m?3X203Dd<P0403hm>`0m?3X0?C`j04i=B`80>3Le
00@092<Q03@c<@0d<c40<c8`0P0];2X0100/:R/0928S02@R8`0l>S/203@b<`04034_<00n?3d0?S`m
038`<@80DDm@00@0>SPi03Pf=`0h=SL08R0Q0P0d<S<01011?d00B4I704Q6A`0^;2d203Xh>@0402XX
:@0[:BX0:bTZ03dk?0809bDV00@0ATA504Q6A`18ATL0;2X[0P12@440100h=SL0>cTj03/i>P0k>CX2
06=QHP0402h/;@0d<S<0=38c03@b<`80D4i?0`0n?3d0100^;2d0>cTj03/i>P0b<342040n?`030490
@@0Z:2T0:RPY00<0;R`]0@0`;Rl1038`<@40oooo00001015DV80A55Q04AAH@14DF4304=@H080A55Q
00<0AE9R04ICH`16Df<01015EfD204AFI0<0@eES0`12E6801017EF00@U1K049@F`16E5l204MEH004
03m?G010D5d0@51M04IFH`80AEER00@0@e1I03@nA`0d?TL07BH]0P0=51X010093A0030d?00`=3`0>
3Ph2010?3@0400l?3@092@L02@T700L71@803Ph<00@06a/I02DU8`0U9B<0:RXX0P0/;2X0100U9B<0
;Rh/02h^;00M7A/201`L6P0402TY9`0Q8Al08B4O01lO7@80:BTW0`0Z:RX0100T92@07alO01lO7`0U
9BD2034a<@04038b<P0Z:RX0:RXZ01dM7@80<30`00@0<S8b02HV9P0V9RH09BDU0P0L71`0100T92@0
CTi>04i>CP0a<C<202LW:@04024Q8`0[:bd0:b/]02/[;@80?Si000@0820R02XZ;00Z:R`09RHX0P0H
61X0100[:bd0<c<e03<c=@0T92H2028R900401hN800X:2X0:2PZ02TY:`80=3@f00<08b<U02TY:`0Y
:B/00`0[:bd0100l?3h0=cLi03Lg>@0Q8B<201XJ700402DU9`0`<380<30b02@T9P808R8T00@0:BT[
02XZ;00Z:R`092@V0P0O7b40100_;c40<c<e03<c=@0T92H201`L7P0401PH6P0U9BL09BDW02@T9P80
@d=500@0;Bd_02PX:P0X:2X0:BT[0P0e=CL0100W9bT0C4a>04a<CP0/;2h302d];`8092@V00@0;Bd_
04I6B016ATP07QhP0P0X:2X0100U9BL02`/=00/;3@00008201TI6`0402HV:00W9bT09bLY01<C5@80
2`/=00@06ATK02DU9`0U9BL09RHX0P0L71h0100000801PH800H6200_;c4202TY:`0402@T9P0O7b40
7alQ02/[;@80>SXl00@09RHX02PX:P0X:2X0:b/]0P0W9bT01014BTP0/[Rf0;:h]P2f_;X20:J/Z004
0:je[P2SZZ<0XjZS0:N^YP80Zk:Z0`2TZj<0102V[JD0YjnT0:N_Y02Y/JH20:^cYP040:bdY`2]]JH0
[KFV0;2hZ@80Zk>T00@0/KN[0;Jh[@2f^:d0a<Bl0P28QGh0100O6QH08a/H02<K600e:bX202<I6004
04a3A00c;bh0<bl^01LE5P80:BT[00@071dO01DI6P0E6AX0<3@e0P0^;c401014ADL0<30b030`<P0e
=CL203Tg>P0401dK7P1BCE40DTeA03Pf>@80:RP[0`0^;2l0100e<cH0?3Xm03`j?@0d<SD203`j?@04
024Q8`0X:2X0:2PZ04E5A`80@41200@0:2PZ02DU9`0U9BL09BDW0P0a<C<0101:BT`0CdmA04m?D@0Y
:B/202PX:P0402l_<@0Y:B/0:BT[020P8P80BTY<00@0?Si002h^<00^;S00=SHh0P0`<380100[:bd0
EEEG05EEE`0b<S@203Ph>P0402h^<00];Bl0;Bd_02TY:`80>c/m00@0=cLi02d];`0];Bl0;Rh`0P0Y
:B/0100Z:2T0DDm@055?D00^;2d2030^;`0402`Z:`0Z:2T0:RPY02d[;080AD=40`0]:b`0100e<c@0
<2h_030^;`13@D8204I4A@0302h/;@0/:R/0;2X[00<0<2h_00@0?cdn04i<C@1>C4d0;bd^0P0d<S<0
100]:b`0;2X[02`Z:`0_;Bh204i<C@0402TW:00L6Q/071XK030^;`80<c4b00@0AD=402l];P0_;Bh0
:bTZ0P0]:b`0100d<S<0?3Xk03`j>`12@442030^;`0403Xh>@14@T<0A49302PV9`809bDV00@0:BLX
04Y8B@1:B4T0B4I70P0_;Bh0100g=CH0<c4b03<a<P0h=SL203@b<`0405YHF@0h=SL0>3Hg02DS9080
=C<d00@07adN02PV9`0X9RL0D4i?0P0^;2d0100V92D0:bTZ02/Y:P0^;2d203@b<`0402PV9`1ACe00
DDm@02`[:@<0:bXX0P0m?3X0101;BTP0?C`j03dl>P0m?3X203Dd<P0403hm>`0m?3X0?C`j04i=B`80
>3Le00@092<Q03@c<@0d<c40<c8`0P0];2X0100/:R/0928S02@R8`0l>S/203@b<`04034_<00n?3d0
?S`m038`<@80DDm@00@0>SPi03Pf=`0h=SL08R0Q0P0d<S<01011?d00B4I704Q6A`0^;2d203Xh>@04
02XX:@0[:BX0:bTZ03dk?0809bDV00@0ATA504Q6A`18ATL0;2X[0P12@440100h=SL0>cTj03/i>P0k
>CX206=QHP0402h/;@0d<S<0=38c03@b<`80D4i?0`0n?3d0100^;2d0>cTj03/i>P0b<342040n?`03
0490@@0Z:2T0:RPY00<0;R`]0@0`;Rl1038`<@40oooo00000`16Df<0AE9R04EBHP0404AAH@80AE9R
00@0AeAT04QEI@18EFD0AUQV2015EfD304AFI00404EBF`0oCEP0?deH04ABG@80AeMT00@0@E5N045C
H@11Df40AUUX0P17FVT0101:F6<0@4iI041>F@0a>d@201hW;P04014F6P0<3@l030d?00`<30803`h<
00@04A4?00/;2@0;2`T020P60P0;2`T0100B4Q006QXH01XJ600U9B<202h^;00402LW9@0];B/0;Bd[
020P7P8071`J00@09BDS02<S8@0S8b407alM0P0Q8Al303De=@0402l_;`0W9bL09bLW02LW9`80;Rh^
00@0<30`02h^;P0^;Rh09BDU0P0X:2P0100d=3@0:BTY02TY:@0V9RH202<S8`04020P801;Bd/0Bd];
034a<`809BDW00@05ADG02d];`0];Bl06ATK0P0l?3h0100P82808b<U02<S9@0[:bd201dM7`0402<S
9@0h>3X0>3Pj02TY:`<07QhP0P0Y:B/00`0W9bT0@T940492A00302`/;P0402/[;@0S8bD08b<U02d]
;`80>SXl00@0820R01TI6`0I6A/08B4S0P0a<C<0100U9BL08b<U02<S9@0X:2X2030`<P0402DU9`0Q
8B<08B4S02l_<@80>CTk00@0;Bd_02DU9`0U9BL06ATK0P0X:2X0100S8bD0>SXl03Xj?00b<S@202<S
9@0402`/;P0a<C<0<C4c028R9080BDU;00@09bLY02<S9@0S8bD07QhP0P0P8280100N7R00>c/m03/k
?@0F5QP201`L7P04018B50020P@00P8400<31@808R8T00<09BDW024Q8`0Q8B<00`0D51H0100H61X0
:b/]02/[;@0Z:R`201@D5P0400<31@0F5QP05QHH034a<`80;2`^00@09RHX01/K7@0K6ad09RHX0P0h
>3X0100S8bD0:b/]02/[;@0h>3X203/k?@04065WI@2h_[`0^;jl0;Bj^080YZbX00@0Z:nX0::YXP2R
ZJ80YJbT0P2//j/30:F/Y0040:J]Y@2V[Z<0YZjS0:J^X`<0YZjQ0P2V[Yl0102W/Z40WjNH09nWV02b
^:X20:NYWP040;nl/`2]YZ00[JJP03P_:P806a0>00@0<bDU02XN7P0Z7Qh0@3He0P17@D40100];2X0
3@d=00d=3@0H71/202dc<`0402hb<`0_<380;c0b03Xk?@80B4Q:00@0>SXl02XX:`0Z:2/0;R`_0P0f
<CD0101>C4l09R@W02HT9`0j>3/2020N8@0402l]<01DDUD0E59E02XX:`80=cDh00@08b<U02/[;@0[
:bd0<c<e0P1?Ce40100B4Q@0?cm103lo@@0S8bD203`l?P0401hN800n?T00?Si004];C@80820R00@0
=cLi02DU9`0U9BL0=3@f0P0Z:R`0101:BT`0>c/m03/k?@0P828204=3A@0401HF6012@T@0@T9405=C
E@809bLY00@0=SHh03<c=@0c<cD0;2`^0P0c<cD0100l?3h0:b/]02/[;@0Y:B/2024Q8`0402XX:@10
?Sl0@3ho03/i>P80;B//00@0;2X[02LU9P0W9BH0:BLX0P16A4D0100Y9bP09bDV02LU9P0c<C8202h/
;@0403Le=P1<BT/0C4Y;034_<080;B//00@0;2X[02h/;@0^;2d0=C<d0P1GEEH0100`;Rl0>SPi03Xh
>@0a;c0202h/;@04034_<01IEeP0FEMH02`Z:`808b4R00@0=C<d02LU9P0W9BH0E59C0P0b<340100]
:b`0:RPY02XX:@0_;Bh2038`<@0403Xh>@0W9BH09bDV038`<@80DU1A00@0:RPY02d[;00]:b`09R@U
0P13@D80101<BT/0;bd^02l];P0^;2d2030^;`0403Dc=00V92D09R@U05iLG@80@Cm000@0:BLX040n
?`10?Sl0<2h_0P0/:R/0101HEUL0>SPi03Xh>@0T8R<202PV9`0402/Y:P0_;Bh0;bd^02d[;080D4i?
00@0<S4_02h]:`0^;B/0;B`Z0P0g=S@0101AD4h0@D0n0450?P11@3h203He<`<0=3<a00@0A4=103`k
>@0l>cT0:2LU0P0_;R`0100g=S@0<c8`03<b<00i=cP202XX:@0404E3A00m>c`0?C/l02d[;080>CLh
00@09B<T044o@011?d00?S`m0P0a;c00100G5AH0;B//02d[;00W9BH2044o@00402TW:00a;c00<Bl`
02@R8`807adN00@0=S@e02/Y:P0[:BX0?3Xk0P1ACe00100e<c@0ATA504I4A@0d<S<203Tg>00402/Y
:P1LFU/0G5YK034_<080=C<d00@09B<T03hl?@0n?3d0@d520P0l>S/0100Y9bP0=cDf03Le=P0e<c@2
02l];P0404I4A@0k>CX0>cTj02HT9@80;bd^0@0`;Rl102LU9P40oooo000304AAH@<0@e1P00@0A55Q
04EBHP15DV80AU=S0P17E6@304=EH`L0A5IT0P15EfD01015DU/0?TaG03i<E`14DUd204UIIP0404=E
H`10Df80@5=R04=HJ@80@UMX00@0A5IT04IFH`16EV<0A55J0P0e?TD0100O92P0414C010A4`0<2`T2
00h=2`0300l?3@0=3@/03@d;00@030`:0P0?3`d0100K6aT0:BTW02TY9`0/;2X202d]:`04028R800K
6aT06a/I024Q7`80:2PV00@09RHT01lO7@0O7ad0<30`0P0d=3@0100c<c<0;2`/02`/;00U9BD202@T
900402PX:00[:b/0:b/[030`<08092@T00@09BDU02/[:`0[:b/0:BTY0P0Q8B40100S8b<0<c<c03<c
<`0h>3X2024Q8`0401PH6P0X:2X0:2PZ01/K7@80<c<e00@0;Rh`01lO8@0O7b408B4S0P0R8R@0100L
71h0>c/m03/k?@0b<S@202HV:00402`/;P0`<380<30b02LW:@80?cm100@0;Rh`02d];`0];Bl08b<U
0P0L71h0100Q8B<0<c<e03<c=@0S8bD201/K7@0401lO8@0b<S@0<S8d02DU9`808b<U00@09BDW038b
=00b<S@07QhP0P0J6Q`0100R8R@0;Bd_02d];`0S8bD201lO8@0400h>400L71h071`N01hN8080:BT[
00@0<C4c01TI6`0I6A/09bLY0P0S8bD0100K6ad0>SXl03Xj?00`<38202LW:@0401`L7P0/;2h0;2`^
01dM7`80>SXl00@07QhP01dM7`0M7Al02@T;0P010@<0100;2`d0:RX/02XZ;00S8bD201`L7P04018B
500D51H051@F01`L7P80:BT[00@08b<U00X:300:2P`0410B0P0W9bT0100Z:R`0:b/]02/[;@0W9bT2
01XJ7004028R900a<C<0<C4c020P8P80<C4c00@0?cm103Ti>`0i>C/0MG]i0P2_]K<0102]/k40ZJn[
0:V_Z`2SZZ<20::YXP030:6XX02Z/JT0Z[6Y00<0YJbT00@0YZfU0:J^X`2V[Z<0YZjS0`2V[Z420:J^
W`040:^fY@2T[I`0Y:fL0;>iZ`80XjFH00@0/k2W0;RaZ@2h/JT0GU=?0P0S5A80100]7Ad0=bTX03LY
:00f:RX203de<`0403Hb<@0<30`030`<01dQ8080<SPf00@0;30a03Xk?@0j>cd0<34c0P1>CU001013
@dD08b4T02<Q900e<cH202PS9`0403`j?@1=Bdh0CD]>02@R9@80;R`_00@0=S@g034_<P0a;c80BTQ;
0P0W9BP0100Z:R`071`N01`L7P0j>S`202h^<00404M7B@0[:bd0:b/]03Ph>P80<S8d00@051@F03Lg
>@0g=cT0=SHh0P19BD/0100T92H0>3Pj03Ph>P0a<C<202XZ;00403hn@014A4H0A4A602@T9P80>CTk
00@08R8T02`/;P0/;2h0DE5C0P0i>C/0100f=SP0>3Pj03Ph>P0b<S@2034a<`0403hn@00W9bT09bLY
02LW:@808R8T00@0<S0a03@b<`0d<S<0DDm@0P0a;c00100c<C80;2X[02`Z:`0a;c0204a:B`0402DS
900R82408R0Q030^;`80:RPY00@0;B//051>C`1@CTl0<S0a0P0^;2d0100c<C80=38c03@b<`0b<342
05mMGP04034_<00n?3d0?S`m03Dc=080<S0a00@0=38c05=ADP1CDE80<S0a0P0T8R<0100b<3408b4R
02<Q8P1HEUL203Pf=`04038`<@0^;2d0;R`]030^;`80:RPY00@0=cDf02XX:@0Z:2T0;R`]0P1ACe00
100W9BH0;bd^02l];P0T8R<203Tg>00404e;C00d<S<0=38c02PV9`80<c4b00@0=C<d02XX:@0Z:2T0
CD]<0P11?d03030^;`0402l];P0U8b@09B<T04]9BP80AD=40`0[:BX0100W9BH0:BLX02TW:00`;Rl2
04=1@P0403@c<@0];2X0;B`Z02XY9`80;Rd[00@0DE1>03ln?00o?S`0@D0n0P0c<S00100_;R`0<c8`
03<b<00m?3X204=2@004034`;P0Z:BL0:RTW03Xi=`80>CPf00@0?S`m03Hd=@0f=3D0@3ho0P18ATL0
100_;Bh0:bTZ02/Y:P0V92D202l];P0405ECE00b<340<S0a02PV9`80;2X[00@0:BLX03dk?00m>c`0
?3Xk0P0e<c@0100a;c00;R`]02h/;@0_;Bh203Tg>00402`Z:`10?Sl0@3ho038`<@80:bTZ00@0<S0a
02h/;@0^;2d0:2HW0P15@d@01014@T<0;bd^02l];P0T8R<202/Y:P0404E3A00c<C80<c4b03@b<`80
;bd^00@0=S@e02PV9`0X9RL0@Cm00P1;BDX01@0`;Rl0=38c03@b<`0d<S<0<2h_0040oooo000304AA
H@<0@e1P00@0A55Q04EBHP15DV80AU=S0P17E6@304=EH`L0A5IT0P15EfD01015DU/0?TaG03i<E`14
DUd204UIIP0404=EH`10Df80@5=R04=HJ@80@UMX00@0A5IT04IFH`16EV<0A55J0P0e?TD0100O92P0
414C010A4`0<2`T200h=2`0300l?3@0=3@/03@d;00@030`:0P0?3`d0100K6aT0:BTW02TY9`0/;2X2
02d]:`04028R800K6aT06a/I024Q7`80:2PV00@09RHT01lO7@0O7ad0<30`0P0d=3@0100c<c<0;2`/
02`/;00U9BD202@T900402PX:00[:b/0:b/[030`<08092@T00@09BDU02/[:`0[:b/0:BTY0P0Q8B40
100S8b<0<c<c03<c<`0h>3X2024Q8`0401PH6P0X:2X0:2PZ01/K7@80<c<e00@0;Rh`01lO8@0O7b40
8B4S0P0R8R@0100L71h0>c/m03/k?@0b<S@202HV:00402`/;P0`<380<30b02LW:@80?cm100@0;Rh`
02d];`0];Bl08b<U0P0L71h0100Q8B<0<c<e03<c=@0S8bD201/K7@0401lO8@0b<S@0<S8d02DU9`80
8b<U00@09BDW038b=00b<S@07QhP0P0J6Q`0100R8R@0;Bd_02d];`0S8bD201lO8@0400h>400L71h0
71`N01hN8080:BT[00@0<C4c01TI6`0I6A/09bLY0P0S8bD0100K6ad0>SXl03Xj?00`<38202LW:@04
01`L7P0/;2h0;2`^01dM7`80>SXl00@07QhP01dM7`0M7Al02@T;0P010@<0100;2`d0:RX/02XZ;00S
8bD201`L7P04018B500D51H051@F01`L7P80:BT[00@08b<U00X:300:2P`0410B0P0W9bT0100Z:R`0
:b/]02/[;@0W9bT201XJ7004028R900a<C<0<C4c020P8P80<C4c00@0?cm103Ti>`0i>C/0MG]i0P2_
]K<0102]/k40ZJn[0:V_Z`2SZZ<20::YXP030:6XX02Z/JT0Z[6Y00<0YJbT00@0YZfU0:J^X`2V[Z<0
YZjS0`2V[Z420:J^W`040:^fY@2T[I`0Y:fL0;>iZ`80XjFH00@0/k2W0;RaZ@2h/JT0GU=?0P0S5A80
100]7Ad0=bTX03LY:00f:RX203de<`0403Hb<@0<30`030`<01dQ8080<SPf00@0;30a03Xk?@0j>cd0
<34c0P1>CU001013@dD08b4T02<Q900e<cH202PS9`0403`j?@1=Bdh0CD]>02@R9@80;R`_00@0=S@g
034_<P0a;c80BTQ;0P0W9BP0100Z:R`071`N01`L7P0j>S`202h^<00404M7B@0[:bd0:b/]03Ph>P80
<S8d00@051@F03Lg>@0g=cT0=SHh0P19BD/0100T92H0>3Pj03Ph>P0a<C<202XZ;00403hn@014A4H0
A4A602@T9P80>CTk00@08R8T02`/;P0/;2h0DE5C0P0i>C/0100f=SP0>3Pj03Ph>P0b<S@2034a<`04
03hn@00W9bT09bLY02LW:@808R8T00@0<S0a03@b<`0d<S<0DDm@0P0a;c00100c<C80;2X[02`Z:`0a
;c0204a:B`0402DS900R82408R0Q030^;`80:RPY00@0;B//051>C`1@CTl0<S0a0P0^;2d0100c<C80
=38c03@b<`0b<34205mMGP04034_<00n?3d0?S`m03Dc=080<S0a00@0=38c05=ADP1CDE80<S0a0P0T
8R<0100b<3408b4R02<Q8P1HEUL203Pf=`04038`<@0^;2d0;R`]030^;`80:RPY00@0=cDf02XX:@0Z
:2T0;R`]0P1ACe00100W9BH0;bd^02l];P0T8R<203Tg>00404e;C00d<S<0=38c02PV9`80<c4b00@0
=C<d02XX:@0Z:2T0CD]<0P11?d03030^;`0402l];P0U8b@09B<T04]9BP80AD=40`0[:BX0100W9BH0
:BLX02TW:00`;Rl204=1@P0403@c<@0];2X0;B`Z02XY9`80;Rd[00@0DE1>03ln?00o?S`0@D0n0P0c
<S00100_;R`0<c8`03<b<00m?3X204=2@004034`;P0Z:BL0:RTW03Xi=`80>CPf00@0?S`m03Hd=@0f
=3D0@3ho0P18ATL0100_;Bh0:bTZ02/Y:P0V92D202l];P0405ECE00b<340<S0a02PV9`80;2X[00@0
:BLX03dk?00m>c`0?3Xk0P0e<c@0100a;c00;R`]02h/;@0_;Bh203Tg>00402`Z:`10?Sl0@3ho038`
<@80:bTZ00@0<S0a02h/;@0^;2d0:2HW0P15@d@01014@T<0;bd^02l];P0T8R<202/Y:P0404E3A00c
<C80<c4b03@b<`80;bd^00@0=S@e02PV9`0X9RL0@Cm00P1;BDX01@0`;Rl0=38c03@b<`0d<S<0<2h_
0040oooo0003049?G`<0@DiN00@0@TmO04=@H013D600AE9R0P16Df<4045CH@80@UAR0`13EF<304AF
I00404QEGP11CUL0@DiG04IDG`80Be]X00@0A5IT03mBH@0oDV40?eAU0P0lDfD0100nD5h0BEUV04UI
IP1@GVT204A>E`0402X_<`0D5AL051DG00`;2@803PX70`0<30X300l?3@0400T91`071`D01`L501<C
4@808b<Q00@0<30^02d]:`0];B/08b<Q0P0I6AL0100M7A/0;Bd[02d]:`0^;R`2024Q7`0402<S8`0Y
:BT0:BTY02h^;P80:b/[00@08R8R01/K6`0K6a/07alO0P0W9bL0100g=cL092@T02@T900C4a<302/[
:`806a/K00@09bLW01`L700L71`0@T940P0O7b40100R8R@08b<U02<S9@0U9BL202/[;@0403lo@@0O
7b407alQ02<S9@80<30b00@08B4S0410@P10@480=cLi0P0X:2X0100^;S00:BT[02TY:`0L71h202`/
;P0402DU9`0Z:R`0:RX/01hN8080820R00@09BDW03De=`0e=CL09RHX0`0N7R0203<c=@0402DU9`0R
8R@08R8T02<S9@80<c<e00@092@V01lO8@0O7b408R8T0P0/;2h0100R8R@08B4S024Q8`0?3a4201`L
7P04020P8P0R8R@08R8T03Hf>08061PJ00@0:BT[01lO8@0O7b4071`N0P0e=CL0100a<C<092@V02@T
9P0E5AL202h^<00401PH6P0c<cD0<c<e01dM7`805aLI00@02`/=00D51`051@L04Q8D0P0^;S000`0R
8R@05aLI01LG6@0300d=3`04028R900U9BL09BDW01TI6`800P8400@06QXL038b=00b<S@08R8T0P0X
:2X0100T92H06ATK01TI6`0Q8B<202d];`0401lO8@0d=3H0=3@f03`l?P809bLY00@0OhF30:F[Z@2U
ZjT0Z:j/0P2Z/:`0102PYj00XZVR0::YXP2NYId20:R_Y`040::YX@2SZZ80XjZR0:B[X`80YZjS00@0
Z;2U0:VaY02Y/J@0Zk>V0P2[/j@0102d_jh0/kj]0;>n[@2oa[H20;JhZ`040;bh[`34^k@0a;^d08Af
L`80<24N00@08a4A03@T900d92@0<bLW0P0L5180100^:RT0:2PV02PX9P0T:2L202hd<P0402/_<013
A4H0@dA6038c=@80@D5300@0C4a>02`Z;@0/:Rd09R@W0P0/9b/0100Z:2/0BdU<04]9C00h=ST203<a
=00402XX:`0X9RT0:2HY04I4A`80=38e0`0P8280100R8R@0=cLi03Lg>@1NGV02038b=004030`<P0S
8bD08b<U03Lg>@808B4S00@0;Rh`04I6B016ATP0ADE70P0R8R@0100e=CL0?Cdo03dm?`0W9bT204E5
A`0403Xj?00O7b407alQ03Lg>@809BDW00@0<S8d05mOH@1OGf40=3@f0P0k>cd0100h>3X0<c<e03<c
=@12@T@202TY:`0402`/;P0[:bd0:b/]02TW:08071XK00@0D4i?02DS900U8b@0:2HW0P0O7Ah0100W
9BH0@Cm0044o@00Y9bP202DS900403Dc=00^;2d0;R`]02h/;@80FEMH00<0>cTj03Hd=@0f=3D00`0h
=SL0100^;2d0H5iO061NG`0/:R/203Tg>004030^;`0]:b`0;B//03@b<`80ATA500@0=S@e01lM7P0O
7Ah0:bTZ0P0R82400`1CDE80?3Xk03`j>`0303@b<`0403Hd=@0/:R/0;2X[03hl?@80=38c00@0;bd^
04Y8B@1:B4T08AlP0P0]:b`0100R8240<S0a038`<@1>C4d203Xh>@0302LU9P0h=SL0>3Hg00<0<S0a
00@0>SPi0490@@12@440?S`m0P0V92D0100f=3D0;R`]02h/;@0g=CH204Y8B@0403Dc=00c<C80<c4b
02PV9`809R@U00@0<2h_03<a<P0c<C80<c8`0P0[:RP302LV900404i=B`0k>SP0>cXh03hm>`80;bh/
00@0:bXX034`;P0a<2h0=SDc0P14@d40100b<Bl08B0N024P7P0d<c4203Dd<P04034_<00e<c@0=C<d
030^;`80Cde>00@0=38c02LU9P0W9BH0>3Hg0P0^;2d0101PGUl0:RPY02XX:@0d<S<2028P8@0402/Y
:P0`;Rl0<2h_04M5AP80<c4b00@0:RPY038`<@0b<3407Q`M0P11?d00100T8R<0=S@e03Hd=@0m>c`2
028P8@0403/i>P0^;2d0;R`]03@b<`80>CLh00@0HV1Q038`<@0b<340<S0a0P0Y9bP01018ATL0;B//
02d[;012@44202LU9P0403Dc=00W9BH09bDV03Pf=`80DDm@00D0?cdn03/i>P0k>CX0>cTj03lm?P01
0?ooo`003016E5l304ABG@<0AE=N0`16E5l404MEH080AUAO00@0AE=N04ABG@14DUd0A59M0P15Deh0
1016E5l0AeEP04MEH011Df4204ADH@0404YHH`1BG6D0DUaU03Hm@`804aPL00@03a<F00`=3`0<3@l0
1`XA0P0@4AD0100E5aH04A8=014B3@092P@200/:1P0401HD5@0S7R808ahR03<c<@8092@R00@0;Rh/
02<S8@0S8b406ATG0P0a<Bl0100d=3807alM01lO7@0U9282020O7@04028Q7`0[:RP0:bXX02XY9`80
8B0N00@08R4O02/Z:00[:RP0<C0^0P0g=S@0100/:bT081lM020O7@0V9B<202PW9@04024P7P0N7A/0
7QdK03Hf=P80;Rh^00@06ATI02DU9@0U9BD0;2`/0P0V9RH0100n?Sh0:2PX02PX:00W9bL202<S8`04
03<c<`0S8b<08b<S04I6AP80:2PX00<06a/K020P800P82000`0N7R00100g=cT092@V02@T9P0Z:R`2
01dM7`0402DU9`0/;2h0;2`^03Ti>`807QhP00@08B4S02@T9P0T92H0>SXl0P0F5QP0100P8280=CDg
03De=`0N7Qh202<S8`0402/[:`0Y:BT0:BTY030`<080;2`/00@06ATI01hN7P0N7Qh06QXJ0P0O7al0
100l?3`0820P020P800O7al201hN7P0402d];@0X:2P0:2PX040oA@807QdS00@08B0U02LV:`0W9R/0
8B4S0P0^;Rh0100Z:RX02`/;00/;2`061`820082000400h>300T92@092@T01/K6`803`lA00@05aLI
00d<4@0=31405aLG0P0/;2`0100K6a/02`T:00/92P0[9bP202HR8`<0:bDW00@081XL03@^<00d;S00
920Q0P0M6a`0100W:BP0<SHe038f=@0g?C/2020V900408jGU02S[:L0XjbW0:V`Z@80XZVQ00@0XjVO
0::UVP2RYIX0Wj>E0P2UYiX0102NYj80WJNO09fWW`2NZ9l20:F]XP040:bbYP2_]JL0[kFW0;6eYP80
[k>R00@0Zk>V0:J/XP2V[:80/;>Z0P2l^kH0102i][40a;bj0<Bl^P2OUI@203P/;004048X:@0R2PX0
8PX:04@`;`80<28Q00@030@203/j=P0k>SH0DUAA0P0d>CD0100b<S@0<C4c034a<`0k>cd202/[;@04
0451@`13@dD0@d=502`/;P80;Bd_00@08B4S03De=`0e=CL0GemQ0P0_;c40100N7R00<C4c034a<`0J
6Q`204E5A`0403<a=00T8RD0928U02l]<080@d5400@0=cDh05ABE@1DDUD0:2HY0P0`;S40100b<3<0
<2ha030^<@0U8bH203lm@00404a:C@0m>ch0?C/n03Xh>`809B<V00@0;B/^02l]<00_;C00Cde@0P0Q
7b80100m>ch0:2HY02PV:@0c<C@205aJG@0404i<C`0T8RD0928U03/i?080?ce000@0C4Y=038`<`0b
<3<0=S@g0P0g=CP0100d<SD0?3Xm03`j?@1[JF`303`j?@80ATA700@0@T1303`j?@0l>Sd0G5YM0P0/
:Rd0100n?3l0;2X]02`Z;@15@dH204m=D00403Hd=`0d<SD0=38e03<a<P80<2h_00@0;bd^05ABD`1D
DU<0=S@e0P0Y9bP0100X9RL0<2h_030^;`0U8b@203@b<`0404E3A00S8B808b4R02XX:@80:BLX00@0
AD=404=1@P13@D80=38c0P0/:R/0100]:b`0=C<d03Dc=00b<34204A2@`0402TW:01=Bd`0CD]<03hl
?@80;R`]00@0:2HW02XX:@0Z:2T0?S`m0P12@440100`;Rl0=cDf03Le=P0g=CH203Tg>00403Le=P1R
H640HV1Q0490@@80>CLh00@0A493044o@011?d00?Sdk0P1IF5H0100f=C<0>cXh03/j>00[:RP203Dd
<P0402XY9`0m?3X0?C`j03Pg=@808R4O00@0<2l]02`[:@0/:bT0>CPf0P1MG5X0100b<Bl0>CPf03Th
=P0W9R@2034`;P0403/j>019B4H0BDQ602d/:P806aXH00@0<c8`02HU8`0V9B<09R@U0P0g=CH0100[
:BX0AD=404E3A012@44202h/;@0402`Z:`0]:b`0;B//04M5AP80A49300@0;2X[028P8@0R8240<c4b
0P0Q7b00101>C4d0<c4b03<a<P0b<34202TW:0<0<S0a00@0?3Xk02TW:00Y9bP0A4930P0a;c00100f
=3D0;R`]02h/;@0[:BX202TW:004059@D@0[:BX0:bTZ02d[;080;bd^00@0;R`]03Pf=`0h=SL0<Bl`
0P0W9BH0100^;2d0=cDf03Le=P0m>c`204E3A00504i=B`13@T00@d9003He<`0i>3H00@3oool000`0
AUAO0`14DUd304ECGP<0AUAO1017EF0204IDG`0404ECGP14DUd0A59M04ABG@80AE=N00@0AUAO04ME
H017EF00@E=Q0P14E640101:F6<0DUaU059LI@0f?D<201<H700400lC5P0<3@l030d?00L:4@80414E
00@05ALF014B3@0A4Pd02@X40P0;2PH0100F51D08ahR02<N8P0c<c4202@T8P0402h^;00S8b408b<Q
01TI5`80<C4_00@0=3@b01lO7@0O7ad09B@R0P0P7ad0100R8Al0:bXX02/Z:00Z:BL2024P7P04028Q
7`0[:RP0:bXX034`;P80=cHd00@0;2/Y020O7@0P7ad09RDS0P0X9bD0100Q81h07QdK01hM6`0f=SH2
02h^;P0401TI6@0U9BD09BDU02`/;0809RHV00@0?Shn02PX:00X:2P09bLW0P0S8b<0100c<c<08b<S
02<S8`16ATH202PX:00301/K6`0P8200820P00<07QhP00@0=cLi02@T9P0T92H0:RX/0P0M7Al0100U
9BL0;2`^02`/;P0i>C/201hN8004024Q8`0T92H092@V03Xj?0805QHH00@0820R03De=`0e=CL07QhN
0P0S8b<0100[:b/0:BTY02TY:@0`<30202`/;00401TI6@0N7Qh07QhN01XJ6P807alO00@0?3`l020P
800P82007alO0P0N7Qh0100];Bd0:2PX02PX:010?dD201hM8`04024P9@0W9R/09bH[024Q8`80;Rh^
00@0:RXZ00/;2`0;2`/01PL20P020P00100>3P`092@T02@T900K6a/200l?4@0401LG6@0=31403@`A
01LG5`80;2`/00@06a/K00/92P0;2@X0:bLX0P0V8R<302/U9`04020J700d;S00=2h`02@P8@807A/L
00@09bTX038f=@0b=SD0=cdk0P0P9R@0102>Ui@0XjbW0:>/Y`2Y/:T20::YX@040:>YW`2RYIX0XZFJ
09nSU@80YJNJ00@0WZNR09fWW`2MYil0WZRO0P2U[J80102//ZH0[kFW0:neY`2a]JH20:ncXP040:^c
YP2V[:80YZbR0;2cZP80_;^f00@0^KJa0<Bl^P34_;X0WiFD0P0h;2`01012:2T08PX:028:2P14<2l2
030R8@0400`40P0k>SH0>cXf059DD@80=3Te00@0<S8d034a<`0a<C<0>c/m0P0[:bd01011@D<0@d=5
04=3A@0/;2h202d];`04024Q8`0e=CL0=CDg05mOH@80;bla00@07QhP034a<`0a<C<06QXL0P15ADL0
100c<C@0928U02@R9@0_;C0204=1A00403Le>01DDUD0E59E02PV:@80<2ha00@0<S0c030^<@0`;S40
9B<V0P0o?D00101<BTd0?C/n03dk?P0j>3/202DS9P0402d[;P0_;C00;bd`04m=D0808AlR00@0?C/n
02PV:@0X9RT0<c4d0P1LFUd0101>C4l0928U02@R9@0k>C`203lm@00404a:C@0b<3<0<S0c03Hd=`80
=cDh00@0=38e03`j?@0l>Sd0JfU/0`0l>Sd204I4A`040490@`0l>Sd0?3Xm05aJG@80;2X]00@0?S`o
02`Z;@0/:Rd0AD=60P1?CE00100f=3L0=38e03@b=@0c<C82030^;`0402l];P1DDU<0E59C03Hd=@80
:BLX00@0:2HW030^;`0`;Rl09B<T0P0d<S<01015@d@08b4R02<Q8P0Z:2T202TW:00404E3A013@D80
@d5203@b<`80;2X[00@0;B//03Dc=00e<c@0<S0a0P14@T<0100Y9bP0CD]<04e;C00n?3d202h/;@04
02PV9`0Z:2T0:RPY03hl?@80@T1100@0<2h_03Le=P0g=CH0=cDf0P0i=cP0100g=CH0HV1Q069PH@12
@44203Tg>00404A2@`11?d00@Cm003hm>`80FEQF00@0=SDc03/j>00k>SP0:bXX0P0e=380100Z:BL0
?C`j03dl>P0h=cD2028Q7`04030_;@0/:bT0;2/Y03Th=P80GEaJ00@0<S4_03Th=P0i>3H09bHT0P0a
<2h0100k>SP0BDQ604U8AP0];2X201/J600403<b<00V9B<09RDS02HT9@80=cDf00@0:bTZ04E3A015
@d@0@T110P0^;2d0100/:R/0;B//02d[;017ADH204A2@`0402`Z:`0R82408R0Q03<a<P808AlP00@0
CTa=03<a<P0c<C80<S0a0P0Y9bP3038`<@0403`j>`0Y9bP0:BLX04A2@`80<Bl`00@0=S@e02h/;@0^
;2d0:bTZ0P0Y9bP0101BD540:bTZ02/Y:P0]:b`202l];P0402h/;@0h=SL0>3Hg034_<0809bDV00@0
;R`]03Le=P0g=CH0?C/l0P15@d@01@1>CD/0@d9004=2@00f=C<0>CPf0040oooo000<04IDG`0304AB
G@15Deh0AE=N00<0AE=N00<0AUAO04MEH017EF001017EF0204IDG`0404ECGP14DUd0A59M04ABG@80
AE=N00@0AUAO04MEH017EF00AEMU0P14E6401015Deh0D5YS051JH`0nAdh202H[;`0401lS9P0A5AP0
4ADH00L:4@802``@00@030h=00/<1`0;30L02@X40P0<2`L0100B41406A@H01TD600S8b42028R8004
02TY9`0N7Q`07QhL01TI5`809RHT00@0:2PV02TY9`0Y:BL09B@R0P0P7ad0100U9280<c8`03<b<00g
=S@202d/:P0402<R800R8Al08R4O02HU8`80;2/Y00<0:RTW02HU8`0V9B<00`0X9bD0100W9R@0:BPV
02TX9P0/;2`203Hf=P04024Q8@0^;Rh0;Rh^02DU9@80;Rh^00@0>CTi03Xj>P0j>SX08b<S0P0R8R80
100[:b/07QhN01hN7P0i>CT202XZ:P0401`L700Q8B408B4Q028R90809bLY00@0>c/m02`/;P0/;2h0
9BDW0P0T92H0100V9RP09BDW02DU9`0h>3X502<S9@0403Ph>P0L71h071`N028R9080<C4c00@08R8R
01TI6@0I6AT08B4Q0P0U9BD0100X:2P09RHV02HV9P0H61P201HF5P0401DE5@0M7Ad07AdM03Xj>P80
:BTY00@06QXJ028R8P0R8R80;2`/0P0R8R801013@TP0;2/a02`[<@0V9BX202/Z;`0402DU9`0`<300
<30`038b<P803Ph>00@030d800P81P0820H04Q8@0P0V9RH0100J6QX03@d?00d=3`0F5QP200d<4@04
01@D500X:2P0:2PX014A4@802`T:00@0920Q02LS900W8b@09R0R0P0X8R@0100T7R00<2X/030Z;00V
8R<201`J6`0402@V9@0^<S40;S8a038h=P80;C<a00@0XZb[0::]Z@2R[JT0WZNR0P2IXi/0102JXIT0
V9jD09RNU02KWY<20:6UU`0409NISP2NX9D0WZ2E0:RZW@80[JjP00@0ZjbL0:ZYU@2ZZID0Z:NB0P2Y
Yi80102KW8`0SXam08j<O@2<R7`208j8O004089hK`2:OGD0RWee07]YI@80<1hJ00@0;Q@C04L_;@17
;bd0>BDT0P0Z71/0100S6aT0>cXf03/j=P1HFUL202hc;`04038b=00V9RP09RHX03Ph>P80:2PZ00@0
<c<e0492A012@T@0;bla0P0J6Q`0100f=SP0;bla02l_<@0j>S`203hn@00402d];`0V9RP09RHX02d]
;`80:BT[00@0@d5402h/;`0^;2l0@3i10P0U8bH0100^;2l0?3Xm03`j?@1@CU4202@R9@0402/Y;00]
:bh0;B/^02PV:@80:RP[00@0@Cm204Q6B@18ATT0=38e0P0[:B`0100b<3<0;2X]02`Z;@0m>ch503Dc
=P04034_<P10?T40@3i104U7BP80>3Hi00@0:BLZ0490@`12@4<0>SPk0P13@D@0100U8bH0<Blb034_
<P0S8B@202@R9@0404a:C@0Y9bX0:BLZ02/Y;080<S0c00@0;2X]02DS9P0U8bH0CTa?0P0S8B@0100b
<3<09bDX02LU:00f=3L203dk?P0403Pf>@0]:bh0;B/^02`Z:`80:RPY00@0;B//04]9BP1;BDX0@3ho
0P0a;c00100b<340>SPi03Xh>@0c<C8203`j>`04059@D@0X9RL0:2HW02XX:@80;R`]00@0?3Xk04Q6
A`18ATL0;bd^0P0X9RL0100V92D0=C<d03Dc=00b<342040n?`0402LU9P0g=CH0=cDf04E3A080:BLX
00@0<S0a02l];P0_;Bh0A4930P1JF5T00`11?d00C4Y;04a:B`03040n?`0403`j>`1HEUL0F5IG0490
@@80;B//00@0=38c02d[;00]:b`0=SDc0P1=C4X0100g=S@0=C@b03Dd<P0W9R@202d/:P0402XY9`0e
=380=C@b03Lf=0807QdK00@0:bXX02PW9@0X9bD0<S4_0P1FEE<3038a;`0403Dd<P0m?3X0?C`j0450
?P80D4m=00@0?3/i02TX9P0Y:2H0>cXh0P0d<c40100^;2d0=cDf03Le=P0Z:2T203Xh>@0404=1@P0]
:b`0;B//03<a<P80>CLh00@0=C<d055?D01ACe00>CLh0P0n?3d0100l>S/0<S0a038`<@15@d@2059@
D@0403Hd=@0d<S<0=38c03`j>`80=cDf00@0@T1102PV9`0X9RL0=S@e0P0i=cP0100^;2d0:2HW02PV
9`0d<S<202`Z:`0404i<C@19AdP0BDM803Xh>@80>3Hg00@0=C<d03dk?00m>c`0>SPi0P0e<c@0100j
>3T0?3Xk03`j>`0m>c`204A2@`05059AC`0e=380=C@b02<R800i>3H00@3oool000`0AUAO0`15Deh4
04IDG`H0AeEP0P16E5l01015Deh0A59M04ABG@14DUd204ECGP0404IDG`17EF00AeEP04UIIP80AEER
00@0@U1K04]HH@1;F640BE9I0P0l@dT0100e>Sh0724U01`Q9@0<3aH200X;3`0400P:2@0:2`H02P/6
00d>20803`h:00@03`d>018=4@0B3A4040l=0P0L6aT0100U928081lM020O7@0P7ad2024P7P0402@S
8@0i>3H0>CPf02PW9@8081lM00@08R4O030_;@0`;bd0>CPf0P0a<2h0100T8b407A`J01dL6P0P7ad2
024P7P0402TX9P0^;B/0;Rd[02XY9`809bHT00@0:bXX02h]:`0^;B/09bLW0P11@D40100c<c<0;Rh^
02h^;P0K6a/202PX:00402@T900m?Cd0?Cdm01lO7`808R8R00@08b<S01`L700L71`0:BTY0P0`<300
100P82008b<S02<S8`0L71h202LW:@<0=CDg00@08b<U030`<P0`<380;Rh`0P0Y:B/0100f=SP0:BT[
02TY:`0U9BL2024Q8`0403<c=@0T92H092@V02DU9`80:RX/00@0>SXj02<S8`0S8b<0:2PX0`0a<C42
03@d=00402d];@0P8200820P01HF5P806a/K00@0=SHf03Xj>P0j>SX06ATI0P0W9bL0100^;Rh0820P
020P8013@TP203dl@P0402TX;@0X9b`0:2L/01lO8@808R8R00@0:b/[0000000000003@h90P0;2`T0
100G5aD0:b/[02/[:`0M7Ad200h>400401PH6P0@3a@040lD01DE5@80:2PX00@02@T901<A4P0C4A80
81`M0P0[9bP0100P6Q`08adO02<M7`0W8B<202TS9@0402TU9P0L6Q/071XK020R8@80:Bd/00@0;38`
0498AP12B4H0YK>c0P2IZ:D0102HY:00VZFO09ZUW`2GWYL209>HT@0409RNU02PXiP0X:>H0;>ZV@80
]JbK00@0/j^H0:fRSP2]XXh0WYAk0P2?PfT01022Me/0OW1E07i`E@2?Pfd208QlIP0408ioJP2GQW@0
UhId08ejJ`80U7e_00@0Tg]_06E:?`1UBSl0=a/H0P0`5a<0100j9b<09AHC02DF4`0j<Rl204i:A`04
02la;P1@E5<0D5AC02`/;P80;Rh`00@09bLY034a<`0a<C<0;bla0P0^;S001019BD/04a<E01<C5@0l
?3h203<c=@0401lO8@17AdT0AdM903dm?`806ATK00@0;bla01XJ700J6Q`0BDM:0P12@4<0100V92L0
<Blb034_<P0P7R4203@b=@0404E3AP0e<cH0=C<f02LU:0809R@W00@0>CLj02d[;P0]:bh0=S@g0P1=
Bdh0100g=CP0;B/^02d[;P0[:B`202HT9`0402h/;`1:B4/0BTQ;02h/;`80>cTl00@0<2ha02TW:P0Y
9bX0?S`o0P10?T40100Y9bX0=S@g03Hd=`0g=CP204A2A@04028P8`0Z:2/0:RP[02PV:@808b4T00@0
@Cm2030^<@0`;S40<2ha0P0a;c80100/:Rd09B<V02DS9P1:B4/202HT9`0402h/;`0]:bh0;B/^030^
<@80<c4d00@0BDM:02h/;`0^;2l0<2h_0P0]:b`0100a;c00?S`m03hl?@19AdP2030^;`0402l];P0d
<S<0=38c02HT9@80:BLX00@0BdU:01hL7@0N71d08AlP0P0b<340100c<C80EeEF05MEEP0`;Rl202`Z
:`0402TW:00o?Ch0?cdn03`j>`80ATA500@0=cDf02`Z:`0/:R/0EUAE0P0^;2d0100j>3T0;2X[02`Z
:`0i=cP205ECE00403Hd=@0l>S/0?3Xk034_<080<S0a00@0<c4b04A2@`14@T<0BDM80P0`;Rl0100f
=3D0;B//02d[;00_;R`2040o?@0303ln?00d<c40=3<a00<0:RTW00@0=3<a03Dd<P0e=380C4]90P0/
:bT0100^;B/09RDS02HU8`0[:RP205=BD00404E4@P12@Cl0@T4o03Pg=@80?C`j00@0>CPf04Q7A@18
AdD0@D0n0P0/:bT303Dd<P04038`<@0e<c@0=C<d02h/;@80<c4b00@0CD]<038`<@0b<340=cDf0P0n
?3d0100h=SL0HEmP065OH010?Sl2044o@00402d[;00/:R/0;2X[02HT9@80De5B00@0;R`]02`Z:`0/
:R/0=38c0P0`;Rl01010?Sl0:BLX02TW:00W9BH2040n?`0402l];P0]:b`0;B//03dk?080<S0a00@0
A493065OH01QGf00AD=40P0f=3D0100c<C80=cDf03Le=P0h=SL203Xh>@0403dk?00g=CH0=cDf03<a
<P80>3Hg00D0@T4o03ln?00o?S`0<c8`03ln?0010?ooo`003016E5l304ECGP@0AUAO1P17EF0204ID
G`0404ECGP14DUd0A59M04ABG@80AE=N00@0AUAO04MEH017EF00BEUV0P15EF801012D5/0BeQQ04]H
H@19DUT203a3B@0403Dj?P0L8BD0724U00`?5P802P/?00@020X900X;1P0:2`H03@h80P0?3PX0100?
3@h04PdA018=4@0@3`d201`K6@0402DT8P0P7ad081lM020O7@808B0N00@092<Q03Th=P0i>3H0:2LU
0P0P7ad0100R8Al0<2l]030_;@0i>3H2034`;P0402@S8@0M71X07A`J020O7@808B0N00@0:BPV02h]
:`0^;B/0:RTW0P0W9R@0100[:RP0;Rd[02h]:`0W9bL20451@@0403<c<`0^;Rh0;Rh^01/K6`80:2PX
00@092@T03dm?@0m?Cd07alO0P0R8R80100S8b<071`L01`L700Y:BT2030`<004020P800S8b<08b<S
01`L7P809bLY0`0e=CL0100S8bD0<30b030`<P0^;S0202TY:`0403Hf>00Y:B/0:BT[02DU9`808B4S
00@0<c<e02@T9P0T92H09BDW0P0Z:R`0100j>SX08b<S02<S8`0X:2P3034a<@80=3@d00@0;Bd]020P
800P82005QHF0P0K6a/0100f=SH0>SXj03Xj>P0I6AT202LW9`0402h^;P0P8200820P04=2B080?Ca2
00@0:BP]02PW;00X9b`07alQ0P0R8R80100[:b/000000000000=3PT200/;2@0401LG5@0[:b/0:b/[
01dM7@803Ph@00@061PJ010?500@3a@05ADE0P0X:2P010092@T04a4B01<A4P0P71d202/W:004020J
700S7Al08adO02LQ8`80:B<U00@0:BDV01`J6`0L6Q/0828Q0P0Y;B`0100/<S00@TQ60498AP2U/k<2
09VXY@0409RTX02JYIl0VZFO09NNU`80TiRA00@0V9jD0:2SV02PXiP0/jZI0P2e[9/0102cZiP0[J:>
0:fRSP2NU7/208n3J@04089gF`1nL5D0OW1E08n3K@80R7aV00@0SWmZ09N6M02GQW@0SGY[0P2DOFl0
102CNfl0IDXo06E:?`0g6aP2030G4`0403XW8`0U5Q<09AHC03Xb;`80CTY700@0;c4^051DD`1@E5<0
;2`^0P0^;S00100W9bT0<C4c034a<`0_;c4202h^<00404U9B`0C4aD04a<E03`l?P80<c<e00@07alQ
04M7B@17AdT0?Cdo0P0I6A/0100_;c406QXL01XJ7019AdX20490@`0402HT9`0a;c80<Blb020N8@80
=38e00@0AD=603Dc=P0e<cH09bDX0P0V92L0100i=cX0;B/^02d[;P0f=3L204e;CP0403Le>00]:bh0
;B/^02/Y;0809R@W00@0;R`_04Y8B`1:B4/0;R`_0P0k>C`0100`;S40:BLZ02TW:P0n?3l2040n@@04
02TW:P0f=3L0=S@g03Le>080A49500@08R0S02XX:`0Z:2/0:2HY0P0S8B@01011?d80<2ha030^<@0`
;S42034_<P0402`Z;@0U8bH09B<V04Y8B`809R@W00@0;R`_02d[;P0]:bh0<2ha0P0c<C@01019AdX0
;R`_02h/;`0`;Rl202d[;004034_<00n?3d0?S`m04U7B080<2h_00@0;bd^03@b<`0d<S<09R@U0P0Y
9bP0101;BDX07Q`M01hL7@0Q7b02038`<@0403<a<P1GEEH0EeEF030^;`80;2X[00@0:BLX03lm?P0o
?Ch0?3Xk0P16A4D0100g=CH0;2X[02`Z:`1FE5D202h/;@0403Xh>@0/:R/0;2X[03Tg>080EE=D00@0
=S@e03`j>`0l>S/0<Bl`0P0b<340100c<C80A49304A2@`19AdP2030^;`0403Hd=@0]:b`0;B//02l^
;080@3lm00<0?chl03@c<@0d<c400`0Z:BL0100d<c40=C@b03Dd<P1<BdT202`[:@0402h]:`0V9B<0
9RDS02/Z:080De9@00@0ADA20491?`12@Cl0>3Le0P0m?3X0100i>3H0B4M504Q7A@11@3h202`[:@<0
=C@b00@0<S0a03Dc=00e<c@0;R`]0P0c<C80101=Bd`0<S0a038`<@0g=CH203hl?@0403Pf=`1QGf00
HEmP040n?`80@Cm000@0;B//02`Z:`0/:R/09R@U0P1CDE80100^;2d0;2X[02`Z:`0d<S<2030^;`04
040n?`0Y9bP0:BLX02LU9P80@3ho00@0;bd^02d[;00]:b`0?C/l0P0b<3401014@T<0HEmP065OH015
@d@203Hd=@0403<a<P0g=CH0=cDf03Pf=`80>SPi00@0?C/l03Le=P0g=CH0<c4b0P0h=SL01@12@Cl0
?chl03ln?00c<S00?chl0040oooo000B04IDG`L0AeEP0P16E5l01015Deh0A59M04ABG@14DUd204EC
GP0404IDG`17EF00AeEP04QHI@80B5IQ00@0@U1K04YGH01:Ef00C5IO0P19DUT01013C5<09bhd02L^
=00C5Qd200h?4`0400/=300?40/03a0;014B30804A0<00@040h?018=4@0B3A4020L50P0@3`d0100P
7ad0:bXX02/Z:00X9bD2028Q7`0402XY9`0m?3X0?C`j02h]:`808R4O00@07A`J028Q7`0R8Al0:BPV
0P0X9bD402DT8P807QdK00@09bHT038a;`0b<Bl0;2/Y0P0W9R@0100Z:BL0:2LU02PW9@0J6QX203<c
<`0403Xj>P0M7Ad07AdM01dM7@807alO00@07QhN03`l?00l?3`08B4Q0P0U9BD0100R8R80820P020P
800R8R8203Ti>@0402PX:00V9RH09RHV02DU9`80;2`^00@0<C4c03lo@@0o?d408b<U0P0];Bl0100T
92H08B4S024Q8`0a<C<202h^<00302@T9P0M7Al07AdO00<0:RX/00@09BDW024Q8`0Q8B<0>c/k0P0R
8R80100S8b<0:BTY02TY:@0V9RH2038b<P0403<c<`0K6a/06a/K01hN7P806QXJ00@0;Rh^04M7A`17
AdL07AdM0P0Y:BT0100];Bd09bLW02LW9`0_;S@203`k@@04024P9@0R8BH08R4V01dM7`807QhN00@0
<C4a00<30`030`<00@800P030`40100B4Q00:BTY02TY:@0K6a/200`<3P0401@D5P0<2a0030/@01PH
6080:RXZ00@02@T9020N7`0P7Ql08alP0P0b;Rl0100M5aT08A/M024K7@0V8282028L7P0402dY:P0O
7Ah07adN01lQ80809BTX00@0:2h/05YPGP1JH5h0WZj]0P2CXj00102EY:40VZFO09ZUW`2DVi@209FJ
T`040:6TV`2UYYh0YJJN0:VGP@80VXQ`00@0RGEL07m[D01oJe00Q6mB0P2@NEX0102JPf<0WhQV09n8
IP29LED208eeF`0409IlH`2OQFh0WhE^09QlIP80WGiZ00@0WWeZ07QGAP1hEdH0>1dF0P0e71L0100S
3PT0=BHQ03DV8@16?3X2030/:`0403@d=01?DE00Ce5@0451@`80;Rh`00@0:2PZ02/[;@0[:bd0:BT[
0P0];Bl0101=CDl0?3`n03`l?P0W9bT2038b=004028R900i>C/0>CTk04=3A@808b<U00@09BDW02`/
;P0/;2h0=S@g0P1<BTd0100Q7b80=C<f03Dc=P0W9BP203Dc=P0402LU:013@D@0@d5403Xh>`80:bT/
00@0@3i103`j?@0l>Sd0;2X]0P0n?3l01014@TD0>SPk03Xh>`0O7B02028P8`04030^<@1;BD`0BdU<
034_<P80<S0c00@0<Blb02PV:@0X9RT0<S0c0P0h=ST01013@D@0:bT/02/Y;011?d82034_<P04030^
<@0U8bH09B<V030^<@80:RP[00@0@Cm204=1A013@D@0>3Hi0P0b<3<0100d<SD0<Blb034_<P14@TD2
02PV:@04028P8`0Z:2/0:RP[02PV:@809R@W00@0DU1C02HT9`0V92L0=S@e0P0a;c00100d<S<0<c4b
03<a<P1@CTl2038`<@0402`Z:`0_;Bh0;bd^03Dc=080<Bl`00@0EE=D02PV9`0X9RL08R0Q0P0f=3D0
100X9RL0FeUJ05]IFP0X9RL202LU9P0402<Q8P0l>S/0?3Xk03Dc=0<0?3Xk0P0R8240101<BT/0;2X[
02`Z:`0/:R/202DS900402h/;@16A4D0ATA503Xh>@80<S0a00@0<2h_034_<00a;c00<c4b0P0f=3D0
101>C4d0<Bl`034_<00c<C8202PV9`0402TX9P0c<S00<c8`04=2@080=3<a00@0;B`Z02PW9@0X9bD0
>STg0P0c<S000`15A480;bh/02l^;00303Xi=`0403`k>@1GEU@0EeID04a;B@80>CPf00@0=C@b03He
<`0f=C<0;2/Y0P0j>CL0100o?S`0:RTW02XY9`0Z:BL202l^;003030^;`0a;c00<Bl`00<0=C<d00@0
FEMH03/i>P0k>CX0=S@e0P0g=CH0100`;Rl0CTa=04i<C@0k>CX202l];P0402LU9P0/:R/0;2X[028P
8@80ATA500@0>SPi02HT9@0V92D0:RPY0P0[:BX0100i=cP0<2h_030^;`0T8R<203/i>P0403Xh>@0m
>c`0?C/l03Xh>@80>3Hg00@0>CLh05]IFP1KFEX0B4I70P0[:BX0100_;Bh0;R`]02h/;@0b<34203Tg
>00403Xh>@0a;c00<Bl`02`Z:`80<2h_00D0?3/i04I5@`16AD<0=C@b02PW9@010?ooo`003016E5l=
04MEH080AUAO00@0AE=N04ABG@14DUd0A59M0P15Deh01016E5l0AeEP04MEH016E5l204UGHP0304EC
GP1:Ef00BUMP00<0B5EN00@0BE=L034kA00a>d@05QTP0P0A4QH0100>40l04A8=014B3@0@4@/200d<
200400d;300C3Q804`hB010<2`8020@300@07ATH03Da<00e<C00:bLV0P0S7ah302dY:00402l^;00X
9bD0:2LU024P7P807QdK00@07A`J020O7@0P7ad09bHT0P0`;bd0100[:RP07QdK01hM6`0S8R0202l^
;00402h]:`0/:bT0;2/Y02XY9`808b8P00@07AdM02@T900T92@0@4100P0@4100100Y:BT05QHF01HF
5P0P8202034a<@@0:2PX0P0V9RH0100S8b<0?Shn03hn?P0a<C4202LW9`0402XZ;00Y:B/0:BT[02PX
:P80@T9400@0:BT[02HV:00V9RP06QXL0P0O7b40100Z:R`0<30b030`<P0S8bD201/K7@04028R900^
;S00;Rh`02@T9P806a/M00@0;Rh^024Q8@0Q8B407QhN0P0K6a/0100H61P0:RXZ02XZ:P0a<C4201HF
5P0402DU9@0F5QH05QHF01hN7P80@d=30`0S8b<0100U9BD0;2`/02`/;00R8BL203dl@P04028Q9P0S
8RL08b8W028R90808R8R00@0>3Ph010@400@41001`P30P092@L0100L71X0=CDe03De=@0Y:BT201LG
6@0401`L7P0@3a@040lD01@D50808R8R00@01`L702DS900U8b@09b<T0P0c;c00100O6A/08Q`N028L
7P0R71h201dG6@0402dY:P0Q7b008AlP020R8@808bLV00@0:c4_079hMP1bN7H0VZZW0P2DXih0102F
Xi`0UZ2H09JPV02DVI8209jOU`040:>SV@2JV8`0VYR<08mgG`80QFeC00@0Nf5807YQ@`1jHD<0QFY<
0P2DMeL0102LOed0Wh9P09n2H02JO60209n1I@0409]lH02HN5l0V7QO09AdF`80WGYT00@0WW]U07aI
A@1lFDD0>QhC0P0f6a80100b6aD0?bdY03l]:@0c:2H202XT900403Xh>@0/;Bl0;2d_05eMG`809bLY
00@0=CDg01`L7P0L71h0:BT[0P0m?Cl0100`<380G5aN05aLGP0Q8B<2028R900402l_<@0V9RP09RHX
03Lg>@80@d=500@0:BT[03Hf>00f=SP09B<V0P0o?D001019AdX07A/N01dK7P0b<3<202`Z;@0402XX
:`0e<cH0=C<f05=AE080A49500@0;R`_034_<P0a;c80:bT/0P0X9RT01011?d80BTQ;04Y8B`0S8B@2
02LU:00403Dc=P0g=CP0=cDh03dk?P80:BLZ00@0<2ha02/Y;00[:B`0:bT/0P0g=CP0101EDeH0?S`o
03hl?`0j>3/202HT9`0403Hd=`0U8bH09B<V02<Q9080928U00@0<c4d04I4A`16A4L0;R`_0P0R82<3
02`Z;@0404a:C@0m>ch0?C/n02PV:@80<S0c00@0=38e02d[;P0]:bh0HV1S0P0/:Rd0100]:b`0:bTZ
02/Y:P0a;c0202l];P0405MEEP15@d@0AD=403dk?080@Cm000@0?3Xk03Dc=00e<c@0De5B0P0b<340
100P7Ql0<Bl`034_<00N71d2059@D@04030^;`0^;2d0;R`]02TW:080>cTj00@0;bd^034_<00a;c00
?C/l0P0S8B80100o?Ch0=cDf03Le=P0Q7b0202TW:004030^;`0k>CX0>cTj04i<C@80=S@e00@0=C<d
03Hd=@0f=3D0=C<d0P0b<340101?CDh0=S@e03Hd=@0c<C8202HT9@0402d/:P0`;bd0<2l]04E4@P80
>3Le00@0;2/Y02@S8@0T8b40<S4_0P0];2X0100g=S@0;2/Y02`[:@0e=38203He<`0403Lf=018AdD0
B4M504U8AP80<2l]0`0g=S@0100/:bT0=cHd03Lf=012@Cl2038a;`0402XY9`0_;R`0;bh/034_<080
;bd^00@0=cDf03Dc=00e<c@0EeEF0P15@d@0100e<c@0;bd^02l];P0[:BX203Le=P0404Q6A`0_;Bh0
;bd^03Hd=@80;bd^0`0b<340101HEUL0<S0a038`<@0[:BX2030^;`04038`<@0m>c`0?C/l02`Z:`80
:bTZ00@0?3Xk04E3A015@d@0:RPY0P0i=cP0100c<C80@Cm0044o@01>C4d202/Y:P04038`<@0]:b`0
;B//030^;`80>CLh00@0>3Hg03<a<P0c<C80<S0a0P0d<S<01@0k>SP0?Sdk03hm>`0j>CL0<C0^0040
oooo000<04IDG`d0AeEP0P16E5l01015Deh0A59M04ABG@14DUd204ECGP0404IDG`17EF00AeEP04ID
G`80BEMR00<0AE=N04YGH01:Ef000`18EEh01019De`0<C]4034kA00F6B02014B5P0400h@3`0A4Pd0
4A8=010A2`803@`800@03@/<01<>4P0C3Q8040`;0P0810<0100M6AP0=C4`03Da<00[9bH202<O7P<0
;BTX00@0;bh/02PW9@0X9bD08B0N0P0N7A/0100M71X081lM020O7@0W9R@2030_;@0402/Z:00N7A/0
7QdK02<R8080;bh/00@0;Rd[02`[:@0/:bT0:RTW0P0S8R00100M7Ad092@T02@T9010@402010@4004
02TY:@0F5QH05QHF020P8080<C4a100X:2P202HV9P0402<S8`0n?Sh0?Shn034a<@809bLW00@0:RX/
02TY:`0Y:B/0:2PZ0P12@T@0100Y:B/09RHX02HV:00J6Q`201lO8@0402XZ;00`<380<30b02<S9@80
6a/M00@08R8T02h^<00^;S0092@V0P0K6ad0100^;Rh08B4Q024Q8@0N7Qh201/K6`0401PH600Z:RX0
:RXZ034a<@805QHF00@09BDU01HF5P0F5QH07QhN0P13@d<302<S8`0402DU9@0/;2`0;2`/028Q9`80
?Ca200@08R4V02<R9`0S8RL08R8T0P0R8R80100h>3P0410@010@400720<200T91`0401`L6P0e=CD0
=CDe02TY:@805aLI00@071`N010?500@3a@051@D0P0R8R8010071`L09B<T02DS900W8b@203<_<004
01lI6`0R71h08Q`N028L7P807ALI00@0;BTZ024O800Q7b00828Q0P0S9bH0100[<Bl0LWQf079hMP2J
ZZL209BSWP0409JSW02FX9P0UZ2H09BITP80WYnG00@0Xj>I09ZHS02JV8`0SgMO0P25KE<0101kHDP0
NV5307YQ@`25JT`209AgE`0409aoG@2OPV00Wh9P09YlH080Wh5U00@0VgaP09QhG`2HN5l0U7AK0P2M
NV@0102NNfD0O5U507aIA@0j7Q<203HK4P04038K5@0o;BT0?bdY03<X9P80:R@T00@0>SPi02`];`0/
;Bl0GEeO0P0W9bT0100e=CL071`N01`L7P0Y:B/203dm?`04030`<P1LG5h0G5aN024Q8`808R8T00@0
;bla02HV:00V9RP0=cLi0P13@dD0100Y:B/0=SHh03Hf>00U8bH203lm@00404U7BP0M6ah07A/N038`
<`80;2X]00@0:RP[03Dc=P0e<cH0De5D0P14@TD0100^;2l0<Blb034_<P0[:B`202PV:@04044o@P1:
B4/0BTQ;02<Q90809bDX00@0=C<f03Le>00g=CP0?C/n0P0Y9bX0100`;S40:bT/02/Y;00[:B`203Le
>00405ECEP0n?3l0?S`o03Xh>`809R@W00@0=S@g02DS9P0U8bH08b4T0P0T8RD0100c<C@0ATA704I4
A`0^;2l2028P8`<0;2X]00@0C4Y=03dk?P0m>ch0:2HY0P0b<3<0100d<SD0;B/^02d[;P1RH6<202`Z
;@0402d[;00[:BX0:bTZ034_<080;bd^00@0EeEF04E3A015@d@0?C/l0P11?d00100l>S/0=C<d03Dc
=01CDE82038`<@04020N7`0a;c00<Bl`01hL7@80DU1A00@0<2h_02h/;@0^;2d0:BLX0P0k>CX0100_
;Bh0<Bl`034_<00m>c`202<Q8P0403lm?P0g=CH0=cDf024O8080:BLX00@0<2h_03/i>P0k>CX0CTa=
0P0f=3D0100e<c@0=S@e03Hd=@0e<c@2038`<@0404m=CP0f=3D0=S@e03<a<P809R@U00@0;B`Z030_
;@0`;bd0ADA20P0h=cD0100/:bT092<Q02@S8@0b<Bl202d/:P0403Lf=00/:bT0;2/Y03Dd<P80=SDc
00@0=cHd04Q7A@18AdD0BDQ60P0`;bd303Lf=00402`[:@0g=S@0=cHd0491?`80<S4_00@0:RTW02l^
;00_;R`0<Bl`0P0_;Bh0100g=CH0=C<d03Dc=01GEEH204E3A00403Dc=00_;Bh0;bd^02/Y:P80=cDf
00@0B4I702l];P0_;Bh0=S@e0P0_;Bh3038`<@0405QFE`0b<340<S0a02/Y:P80<2h_00@0<S0a03dk
?00m>c`0;2X[0P0[:BX0100l>S/0AD=404E3A00Z:2T203Tg>00403<a<P11?d00@Cm004i<C@80:bTZ
00@0<S0a02d[;00]:b`0<2h_0P0i=cP0100h=SL0<c4b03<a<P0b<34203@b<`0503/j>00n?C/0?Sdk
03Xi=`0a<2h00@3oool000`0AUAO1P18EV4404MEH080AUAO00@0AeEP04IDG`16E5l0AE=N0`14DUd2
04ECGP0404IDG`17EF00AeEP04=AG080B5IQ00<0AUAO04UGHP19Ef800`14DUd01019Ef80>4IA03Q6
D@0I72<2010A5@0400/=300>3`X03Pl:00d>20802@P400@02@L8010;3`0@2`l05a<B0P060P40100F
4Q40<Rh]038^;@0W8b8201hJ6@0402HR8@0J5QD06QHE02XY9`<0;2/Y0P0X9bD0100R8Al081lM020O
7@0X9bD2034`;P0402/Z:00P7ad081lM01lN70809bHT00@0;B`Z030_;@0`;bd0;Rd[0P0V9B<0100[
:b/08R8R028R8P0o?cl201dM7@04030`<00H61P061PH02@T90809RHV00@0=3@d02PX:00X:2P0;Rh^
0`0V9RH203Xj>P0403De=@0V9RH09RHV02HV:0808b<U00@0820R03Xj?00j>S`0<c<e0P0R8R@0100L
71h09bLY02LW:@0T92H2038b=004028R900M7Al07AdO01dM7`80<C4c00@08b<U01`L7P0L71h0:RXZ
0P0^;Rh0100V9RH07QhN01hN7P0N7Qh202TY:@0403<c<`0Y:BT0:BTY02<S8`8051@D00@04Q8B034a
<@0a<C40:b/[0P0L71`0100M7Ad0:b/[02/[:`0L6b4203`k@@0402DT:@0S8RL08b8W02<S9@807QhN
00@0;Rh^00l?3`0?3`l000400P000000100A4@l0;2`/02`/;00R8R82014A4`0401<C5@040`P010<8
014A4@8061PH00@01PH6024O800Q7b00:2@U0P0[9bP0100Q6ad091hP02@N800N61X201dG6@0402TU
9P0P7Ql081hO020R8@808BDT00@0=S`j08V?S@29Shd0T:2F0P2DXIP0102JY9/0WZBJ09jTVP2QXID2
0:2LT00408n7NP1fK600MVaP07UPBP80OfE>00@0QVaC08icF02>LeP0TWMJ0P2CMUP0102@Le<0SW5A
08iaD@2GMeh209emI004099bF@2=KEH0SFeF08m_F080UWIQ00@0UGEP07QGA01hEd@0B2XO0P0M0P00
100j91T0@2lX040_:00^8al203La<@0402d[;00Z:R`0:RX/04Y:C080?Si000@0:b/]02d];`0];Bl0
@d=50P0^;S00100];Bl0?3`n03`l?P10@482024Q8`0403Hf>00Q8B<08B4S02PX:P80DE5C00<0=3@f
02@T9P0T92H00`0]:bh01012@4<0>3Hi03Pf>@0W9BP202XX:`0402HT9`0d<SD0=38e04Y8B`80DDmB
00@08R0S02<Q900S8B@0>3Hi0P0U8bH0100]:bh0@Cm2044o@P0l>Sd2030^<@04038`<`0T8RD0928U
04=1A080<Blb00@0;2X]02XX:`0Z:2/0;B/^0P12@4<0101:B4/0FUQK05YHF`0Z:2/202h/;`0402`Z
;@0Y9bX0:BLZ020N8@809bDX00@0;bd`04m=D01?CE00;2X]0P0O7B00100_;C00;B/^02d[;P0i=cX2
03/i?00401dK7P0S8B@08b4T02/Y;0807adP00@0E59E02DS9P0U8bH0@T110P10?Sl0100m>c`0>CLh
03Tg>01BD54204M5AP0403<a<P0d<S<0=38c02/Y:P80:2HW00@0?C/l03Tg>00i=cP081hO0P0/:R/0
100V92D0D4i?051>C`15@d@203Xh>@0403<a<P0l>S/0?3Xk030^;`80;bd^00@0?3Xk02`Z:`0/:R/0
>cTj0P1:B4T0100V92D0;B//02d[;00_;Bh202d[;00404e;C00a;c00<Bl`02l];P80<S0a00@0;bd^
038`<@0b<340BdU:0P16A4D0100o?Ch0=C<d03Dc=00m?3X203/j>00404U8AP16AD<0ATE3038a;`80
:bXX00@0;B`Z034`;P0a<2h0=SDc0P0b<Bl0100];2X092<Q02@S8@0X9bD203Lf=00404a;B@0k>SP0
>cXh03Lf=080>3Le00@0;bh/03He<`0f=C<0@T4o0P0i>3H0100^;B/0;2/Y02`[:@0e<c@2030^;`04
038`<@0_;Bh0;bd^04=1@P80CTa=00@0>3Hg02d[;00]:b`0>CLh0P0h=SL0101BD540=C<d03Dc=00d
<S<201lM7P04030^;`0R82408R0Q05ECE080>CLh00@0:BLX02l];P0_;Bh0:BLX0P13@D80100g=CH0
928S02@R8`0e<c@203hl?@0401hL7@0c<C80<c4b03@b<`80;B//00@0EE=D03/i>P0k>CX0<S0a0P0/
:R/0100^;2d0<c4b03<a<P0b<34203<a<P0403Hd=@0d<S<0=38c02`[:@80?3/i0@1DDe4103ln?040
oooo000<04IDG`@0BEMR0P18EV4404MEH080AUAO00@0AeEP04IDG`16E5l0AE=N0`14DUd204ECGP04
04IDG`17EF00AeEP04EBF`80AU=L00@0@e5L04UGHP19Ef80AE=N0P11DEh01019FFH0?DeJ03e=FP0S
9Rd201<D600400L9200;30L02``700l@2P8030/700@02@L800h93@0>2@d05a4A0P0=1`L0100A2`/0
:28R02PR8P0X8R82020J6P0402HP800O6AT07aTI02@S8@80:RTW00@0<2l]02h]:`0^;B/09bHT0P0S
8R000`0V9B<0:RTW02XY9`0302PW9@04024P7P0P7ad081lM02TX9P80;bh/00@0;Rd[02`[:@0/:bT0
:2PX0P0M7Ad0100/;2`0>3Ph03Ph>00Y:BT202LW9`0402`/;00[:b/0:b/[03lo?`809BDU00@0<30`
01hN7P0N7Qh09BDU0P0];Bd00`0e=CD08R8R028R8P0302PX:P0402DU9`0_;c40;bla03/k?@807QhP
00@071`N02@T9P0T92H08R8T0P0e=CL0100T92H08b<U02<S9@0L71h203@d=P0402DU9`0Q8B<08B4S
028R8P80<S8b00@09RHV024Q8@0Q8B409BDU0P0N7Qh0100Y:BT0?3`l03`l?00O7al201`L7004018B
4P0P8200820P03Ti>@80820P00@071`L02LW9`0W9bL05QDK0P0f=C/0100W9R/08b8W02<R9`0U9BL2
01hN7P0402DU9@0D51@051@D00830080000000@0410>02d];@0];Bd0:BTY0P0N7R00100Q8B<04Q4F
018A5P0O7al201`L700400l?3`0L6Q/071XK02PT9@807QXK00@081XL02<M7`0S7Al07ALI0P0P6Q`0
100R7Ql06QPI01XH6@0N81l2020T8`0404E;B@2MXj40WJ>Q08NCQ@80U9nA00@0X:JH0:BUU`2TYIL0
X9Z<0P2:PG80101`He<0HU92069B@P1nIeD207eVE00407eWDP21J580PFQB08M]EP80SG=H00@0U7UN
09MlG`2GO5l0TgAO0P2INVD0102=LE`0Sg=N08mcGP2CMV@208ebG`0408I[FP1bEdH0LUM605Tl;080
@BDI00@05`0003`Y8P0l:B80@S@a0P0W7Ql0100o>Sh0:bX_02/Z;`0Z:R`204e=C`0403/k?@0j>S`0
>SXl0451@`8092@V00@0<30b024Q8`0Q8B<0BDU;0P0i>C/0100`<380:BT[02TY:`0Y:B/203`l?P04
03lo@@0W9bT09bLY02d[;P80:2HY00@07Q`O05IDE`1FE5L0;R`_0P0b<3<0100U8bH0?S`o03hl?`0/
:Rd20490@`0403Pf>@0]:bh0;B/^03@b=@80<2ha00@0:2HY02XX:`0Z:2/0D4iA0P0f=3L0100X9RT0
9R@W02HT9`0h=ST204E3AP0402`Z;@0_;C00;bd`03<a=080@Cm200@0>CLj059@D`1BD5<0=S@g0P0e
<cH0100[:B`0;bd`02l]<00[:B`203<a=004034_<P1KFE`0FeUL03Hd=`80;2X]00@0?C/n03<a=00c
<C@0>cTl0P1DDUD0100e<cH0<Blb034_<P0n?3l202h/;`0405mMH011?d80@Cm203Pf=`80>CLh00@0
=38c03Tg>00i=cP0AD=40P1<BT/0100c<C80=C<d03Dc=00`;Rl2034_<00303Le=P1>C4d0CTa=00<0
:RPY00@0<S0a04M5AP17ADH0BdU:0P0c<C800`0[:BX0<2h_030^;`0302d[;00403<a<P0]:b`0;B//
030^;`80BDM800<0;R`]02`Z:`0/:R/00`0]:b`01019AdP0>cTj03/i>P0d<S<203Le=P0402l];P0h
=SL0>3Hg044o@080C4Y;00@0@3ho03Le=P0g=CH0?chl0P0k>SP0100o?S`0C4]904a;B@0d<c4203<b
<00402LV900i>3H0>CPf02TX9P80=cHd00@0=3<a02`[:@0/:bT0<c8`0P0e=380101<BdT0>STg03Xi
=`0e=38203Lf=00403@c<@0g=S@0=cHd040o?@80?Sdk00@0<S4_02LV900W9R@0=38c0P0]:b`0100Z
:2T0;2X[02`Z:`0_;Bh205UGF00303`j>`0[:BX0:bTZ00<0=S@e00@0?C/l03hl?@0n?3d0<S0a0P0Q
7b0302h/;@0303Tg>014@T<0A49300<0;R`]00@09B<T04=1@P13@D80?C/l0P0Z:2T0100d<S<0=cDf
03Le=P0Y9bP202`Z:`0403Hd=@0T8R<0928S055?D080AdE600@0=C<d030^;`0`;Rl0<S0a0P0c<C80
100a;c00=cDf03Le=P0k>CX2038`<@0502`[:@0f=C<0=SDc05EDDP0W9R@00@3oool000`0AUAO1019
Ef8204QFH@@0AeEP0P16E5l01017EF00AUAO04IDG`15Deh304ABG@80AE=N00@0AUAO04MEH017EF00
AE9K0P16De`01013DE`0BEMR04UGHP15Deh2045AGP0404UIIP0mCEX0?DeJ02<V;@804a@H00@01`T8
00/<1`0;30L03a0:0P0<2`L010091`P03PT=00h93@0G4A4200d71`04014;2`0X8R80:28R02PR8P80
81XJ00@09R0P01lI6@0O6AT092<Q0P0Z:BL0100`;bd0;Rd[02h]:`0W9R@202<R800302HU8`0Z:BL0
:RTW00<0:2LU00@08B0N020O7@0P7ad0:BPV0P0_;R`0100^;B/0;2/Y02`[:@0X:2P201dM7@0402`/
;00h>3P0>3Ph02TY:@809bLW00@0;2`/02/[:`0[:b/0?clo0P0U9BD0100`<3007QhN01hN7P0U9BD2
02d];@0303De=@0R8R808R8R00<0:2PZ00@09BDW02l_<@0_;c40>c/m0P0N7R00100L71h092@V02@T
9P0R8R@203De=`0402@T9P0S8bD08b<U01`L7P80=3@f00@09BDW024Q8`0Q8B<08R8R0P0b<S80100V
9RH08B4Q024Q8@0U9BD201hN7P0402TY:@0l?3`0?3`l01lO7`8071`L00@04Q8B020P800P8200>CTi
0P0P8200100L71`09bLW02LW9`0F5A/203He>`0402LV:`0S8RL08b8W02DU9`807QhN00@09BDU01@D
500D51@00P<00P000000100@40h0;Bd]02d];@0Y:BT201hN8004024Q8`0B4AH04Q4F01lO7`8071`L
00@03`l?01`J6`0L6Q/0:2@U0P0N6Q/0100P6Q`08adO02<M7`0M5aT2020J7004028N7`0J61T06QPI
01hP7`8082@S00@0AD]909fSX@2MXj40Qi>50P2DWi40102PYYP0Y:FG0:BUU`2PVX`208Z1LP04071S
D`1RDT80HU9207iWE@80OFID00@0OFMB085XDP21J580QfeF0P2=LeP0102DNEh0UgaO09MlG`2CM5l2
09UjI@0408eaG02?Leh0Sg=N09=fI080SG9O00@0QV]J079GAP1bEdH0FC`/0P119AT0100G0000?2TR
03`Y8P12=34202LN7`0403lj?P0[:Rl0:bX_02XZ;080CDe?00@0>c/m03Xj?00j>S`0@D530P0T92H0
100`<3808B4S024Q8`19BD/203Ti>`04030`<P0Y:B/0:BT[02TY:`80?3`n00@0?cm102LW:@0W9bT0
;B/^0P0X9RT0100N71l0EUAG05IDE`0^;2l2038`<`0402DS9P0n?3l0?S`o02`Z;@80@T1300@0>3Hi
02d[;P0]:bh0=38e0P0`;S40100X9RT0:RP[02XX:`1@CU4203Hd=`0402PV:@0V92L09R@W03Pf>@80
AD=600@0;2X]02l]<00_;C00<c4d0P11?d80100i=cX0DU1C059@D`0f=3L203Dc=P0402/Y;00_;C00
;bd`02/Y;080<c4d00@0<Blb05]IG01KFE`0=S@g0P0/:Rd0100m>ch0<c4d03<a=00k>C`205ABE@04
03Dc=P0a;c80<Blb03hl?`80;R`_00@0GeeP044o@P11?d80>3Hg0P0i=cP0100d<S<0>CLh03Tg>015
@d@204a:B`0403<a<P0e<c@0=C<d030^;`80<Bl`00<0=cDf04i<C@1>C4d00`0Z:2T0100b<340AdE6
04M5AP1;BDX203<a<P0302/Y:P0`;Rl0<2h_00<0;B//00@0<c4b02d[;00]:b`0<2h_0P19AdP00`0^
;2d0;2X[02`Z:`0302d[;00404U7B00k>CX0>cTj03@b<`80=cDf00@0;bd^03Pf=`0h=SL0@Cm00P1<
BT/01010?Sl0=cDf03Le=P0o?S`203/j>00403ln?01<BdT0C4]903@c<@80<c8`00@09bHT03Th=P0i
>3H0:BPV0P0g=S@0100d<c40;2/Y02`[:@0c<S0203Dd<P0404a;B@0j>CL0>STg03Dd<P80=cHd00@0
=3<a03Lf=00g=S@0@3lm0P0n?C/0100b<Bl09bHT02LV900d<S<202d[;00402XX:@0/:R/0;2X[02l]
;P80FEMH00<0?3Xk02/Y:P0[:BX00`0f=3D0100m>c`0?S`m03hl?@0b<342024O80<0;R`]00<0>CLh
04A2@`14@T<00`0^;2d0100U8b@0@d5204=1@P0m>c`202XX:@0403@b<`0g=CH0=cDf02TW:080;2X[
00@0=S@e02@R8`0T8R<0DDm@0P17ADH0100e<c@0<2h_030^;`0b<34203<a<P04034_<00g=CH0=cDf
03/i>P80<S0a00D0;2/Y03He<`0f=C<0EEAB02LV90010?ooo`003016E5l0101:F6<0BEMR04UGHP19
Ef8304QFH@<0AeEP0P16E5l01017EF00AUAO04IDG`15Deh304ABG@80AE=N00@0AUAO04MEH017EF00
AeAM0P14DEX0100oCEP0BEMR04UGHP15EF8204=CH00404]MJ`10DV00@59P02ha>0805aPL00@01PP7
00`=200<3@P05AH@0P0D4`l0100>30d03PT=00h93@0B30`201D?3`0400l92@0P6QX081XJ02dW9`80
9R0P00@0;2HV034[:`0a:b/081lM0P0V9B<0100/:bT0:bXX02/Z:00V9B<202<R8004028Q7`0S8R00
8b8P02PW9@80<2l]00@09bHT01`K6@0L6aT092<Q0P0[:RP0100Z:BL0;Rd[02h]:`0Z:RX202<S8`04
01lO7`1@D500D51@01TI6@80;Bd]00@092@T02DU9@0U9BD0ADE50P0Q8B400`0_;bl05ADE01DE5@03
028R8P0403<c<`0O7al07alO01hN80809RHX00@09BDW02@T9P0T92H0@T940P0T92H0100W9bT0;Bd_
02d];`0R8R@203Lg>@0402HV:00X:2X0:2PZ01dM7`80=SHh00@09bLY02DU9`0U9BL08B4Q0P0g=cL3
02PX:004030`<00J6QX06QXJ02@T9080CDe=00@07QhN02LW9`0W9bL06QXJ0P0H61P01016ATH0:BTY
02TY:@0R8R8202LW9`0402LV;013@TP0@d9803Lf>`80;Rdb00@0;2`^020P800P82007AdM0P0A4A40
10082@@010@200@40P0B4Q02038b<P0403@d=00_;c40;bla03Lg>@80:RT^00@0<c<c02PX:00X:2P0
71`L0P0L6Q/0100Y9BH0510A01@@4@0O6A/2028L7P0401hH6P0S7Al08adO01dI6P805A<D00@06adL
01lS8P0O8b80CeEC0P2Z/:h01028Th<0VZ6A09ZQT@2QY9<209^HQ`0408QoKP1ZGD`0JUe<061??@80
KeY700@0MfAE07EOD@1eGe40MF1?0P1nIeD0102;MF00Tg]S09=kH`2@N60208UaE`0408Y_G02<LEh0
S75N08=XE`80RW1O00@0SgEV085ZFP21JUX0OFIH0P1gH580101hFd/0Gd<e05m3=@0/50P201D20004
034S8010=cP0@3Lh02lZ;P80:RP]00@0;Rh`03@d=P0d=3H0Lg=e0P0T92H0100E5AL0?cm103lo@@0J
6Q`203Xj?00402l_<@1=CDl0CDe?024Q8`80;Rh`00<0=SHh028R900R8R@00`1;Bdd0100I5aX0;2X]
02`Z;@0S8B@203Tg>P0405=AE00c<C@0<c4d03lm@080>3Hi00@07A/N02`Z;@0/:Rd0EeEH0P0m>ch0
100J61/0=C<f03Dc=P0h=ST2020N8@0405IDE`0f=3L0=S@g01lM8080<Blb00@0:2HY05IDE`1FE5L0
;R`_0P0h=ST0100i=cX0<c4d03<a=00b<3<203<a=00405=AE00a;c80<Blb03Hd=`80=C<f00@0;2X]
03<a=00c<C@0:bT/0P1JF5/0100f=3L0;bd`02l]<00o?D0202d[;P0401hL7`17ADP0AdE802XX:`80
7adP00@0;bd`01/I700K6A`0BdU<0P0l>Sd0100c<C80=C<d03Dc=00/:R/203Le=P0403Dc=016A4D0
ATA502LU9P80:BLX00@0;R`]03<a<P0c<C80;B//0P1CDE80100W9BH07A/L01dK700^;2d203Dc=004
059@D@0b<340<S0a02XX:@80;bd^00@0>3Hg03Xh>@0j>3T0>SPi0P0i=cP0100`;Rl0BTQ904Y8B@0m
>c`302l];P80=S@e00@0AdE604Q6A`18ATL0=cDf0P0k>CX0100_;Bh0?C/l03dk?00i=cP204m=CP04
03lm?P0h=SL0>3Hg02h]:`80:RTW00@0:2LU04=2@013@T00:RTW0P0`;bd0100L6aT0>CPf03Th=P0V
9B<203/j>003034`;P0U92809B@R00<0;Rd[00@0CDa:0450?P11@3h0>3Le0P0k>SP0100l>cT0?C`j
03dl>P13@T0204I5@`0403Xi=`0Z:BL0:RTW02d[;080:BLX00@09R@U02h/;@0^;2d09B<T0P1QGf00
100n?3d09R@U02HT9@0^;2d203Hd=@0402XX:@1CDE80De5B044o@080=S@e00@0;2X[03Tg>00i=cP0
;R`]0P1LFU/01010?Sl0=cDf03Le=P0Y9bP20490@@0403`j>`0a;c00<Bl`03/i>P80>3Hg00@0?3Xk
02TW:00Y9bP0=cDf0P0R82401017ADH0BDM804U7B00o?Ch203`j>`0403lm?P0m>c`0?C/l03Tg>080
@T1100@0AD=403Pf=`0h=SL0?C`j0P0Y:2H1051?C@40=C@b0@3oool000`0AUAO0`18EV401017EF00
AUAO04IDG`15Deh204ABG@<0@e5L1012D5/304=AG0D0A59M1016E5l504MEH0<0B5IQ00@0>d=6024Y
:`0Q:B/02PlB0P042PX0100:3Pl03Q0?00h@3`0;3@`200X<2@04014<200D3`/050l;014:10805Pl9
00@0:R4J038Y8P0b:B80<2HM0P0e:b8302<R800302DT8P0X9bD0:2LU00<0;2/Y00@0:BPV02HU8`0V
9B<08b8P0P0_;R`0100d<c40:bXX02/Z:00U928202TX9P04034`;P0d<c40=3<a02d];@80:2PX00@0
:b/[03@d=00d=3@0<30`0P0R8R80100M7Ad08B4Q024Q8@0T92@203/k>`04028R8P0M7Ad07AdM028R
8P805QHF00@0:2PX03Ph>00h>3P09RHX0P0B4Q@0100`<380:b/]02/[;@0/;2h203lo@@0402LW:@0T
92H092@V02HV:08092@V00@0=cLi02HV:00V9RP07alQ0P0T92H0100Z:R`09RHX02HV:00V9RH202<S
8`0403De=@0a<C40<C4a01@D508071`L00@0<30`02@T900T92@0=CDe0P0P8200100O7al07QhN01hN
7P0a<C4202TY:@0402<S8`0d=3@0=3@d02`[<@80<c8h00@0ADA902LV:`0W9R/0:2PZ0P0K6a/0100R
8R802`/;00/;2`0720<200H6100400d=2`0U9BD09BDU02HV9P80820R00@0:b/]02TX;@0Y:2d0@3i1
0P0J61T0100M71P06aXE01/J5@0J6AD2020N7`0401TG700Y9Rl0:BH_00h<3`809bDV00@05Q@E01LG
5@0G5aD0824L0P0I6aH0101WJV<0WjBM09nTW@2MU8D20:nST`040:>FP`24MEh0Q7EN07=SBP80KUa6
00@0K5M406mHB01_F4P0L54m0P1hG4L01025JE@0Rg1K08]`F`2;L5/208M^F0<0QFaF00<0Q6mN081[
FP20JeX00`1mJ5D0101eH4/0JUHn06YF?P1UDCT206MC>`04065<=`1GAC40EdDa04<b8P808Q@700@0
4`/002TT7P0Y91h0=C@_0P0V9b80100n?cT0=CH`03Df<00i>SD204m>C00402LV900Z9RL0:RHW03Xf
=`80;RX[00@09b<T04Q6A`18ATL0?3Xk0P0k>cd0100X9b`0=cPl03Lh?00U9R/2059EFP<0;B/^0`0Y
9bX01018ATT0De5D05=AE00`;S4201hL7`04034_<P0W9BP09bDX02LU:080CTa?00@0:BLZ038`<`0b
<3<0;B/^0P0k>C`0100[:B`0>SPk03Xh>`0a;c8202PV:@04038`<`0^;2l0;R`_05MEF080>cTl00@0
>SPk038`<`0b<3<08R0S0P0W9BP00`10?T40@d5404=1A00303Dc=P0402PV:@0W9BP09bDX02TW:P80
=cDh00@0@T1302PV:@0X9RT0:BLZ0P0X9RT0100O7B00BdU<04]9C00`;S42024O8P0402`Z;@0]:bh0
;B/^02XX:`80EE=F0`0`;Rl0100Y9bP0;B//02d[;00]:b`205YHF@0402XX:@0k>CX0>cTj03Hd=@80
=38c00<0:bTZ05YHF@1JF5T00`0S8B80100Z:2T0928S02@R8`1=Bd`203<a<P0402LU9P0Z:2T0:RPY
04i<C@80=cDf00@0?S`m03@b<`0d<S<0<S0a0P13@D801019AdP0:RPY02XX:@0o?Ch203@b<`0402d[
;01THV<0I69S03`k>@80<C0^00@0>CPf03@c<@0d<c40:2LU0P14@d40101@Cdd0:bXX02/Z:00[:RP2
03Dd<P<0=cHd00@0Cdi<02HU8`0V9B<0;B`Z0P0_;R`0100];2X0<c8`03<b<00i>3H201`K6@0402XY
9`0l>cT0?3/i038a;`80I6=Q0`17AT@0100i>3H0@d9004=2@00h=cD204e<BP0403/j>00[:RP0:bXX
02l];P80<2h_00@0<c4b030^;`0`;Rl08b4R0P0l>S/0101CDE80;B//02d[;00]:b`202XX:@0403hl
?@1KFEX0FeUJ04=1@P80?cdn00@0>CLh03<a<P0c<C80<2h_0P13@D801015@d@09bDV02LU9P0i=cP2
02HT9@0403@b<`0n?3d0?S`m02`Z:`80<Bl`00@0@3ho02h/;@0^;2d0:bTZ0P0i=cP0100j>3T0EUAE
05IDE@12@Cl2034`;P04038a;`0i>3H0>CPf030_;@80:bXX00@0<S4_03He<`0f=C<0;2/Y0P0R8Al1
02`[:@40EUEC0@3oool000`0AUAO0`18EV401017EF00AUAO04IDG`15Deh204ABG@<0@e5L1012D5/3
04=AG0D0A59M1016E5l504MEH0<0B5IQ00@0>d=6024Y:`0Q:B/02PlB0P042PX0100:3Pl03Q0?00h@
3`0;3@`200X<2@04014<200D3`/050l;014:10805Pl900@0:R4J038Y8P0b:B80<2HM0P0e:b8302<R
800302DT8P0X9bD0:2LU00<0;2/Y00@0:BPV02HU8`0V9B<08b8P0P0_;R`0100d<c40:bXX02/Z:00U
928202TX9P04034`;P0d<c40=3<a02d];@80:2PX00@0:b/[03@d=00d=3@0<30`0P0R8R80100M7Ad0
8B4Q024Q8@0T92@203/k>`04028R8P0M7Ad07AdM028R8P805QHF00@0:2PX03Ph>00h>3P09RHX0P0B
4Q@0100`<380:b/]02/[;@0/;2h203lo@@0402LW:@0T92H092@V02HV:08092@V00@0=cLi02HV:00V
9RP07alQ0P0T92H0100Z:R`09RHX02HV:00V9RH202<S8`0403De=@0a<C40<C4a01@D508071`L00@0
<30`02@T900T92@0=CDe0P0P8200100O7al07QhN01hN7P0a<C4202TY:@0402<S8`0d=3@0=3@d02`[
<@80<c8h00@0ADA902LV:`0W9R/0:2PZ0P0K6a/0100R8R802`/;00/;2`0720<200H6100400d=2`0U
9BD09BDU02HV9P80820R00@0:b/]02TX;@0Y:2d0@3i10P0J61T0100M71P06aXE01/J5@0J6AD2020N
7`0401TG700Y9Rl0:BH_00h<3`809bDV00@05Q@E01LG5@0G5aD0824L0P0I6aH0101WJV<0WjBM09nT
W@2MU8D20:nST`040:>FP`24MEh0Q7EN07=SBP80KUa600@0K5M406mHB01_F4P0L54m0P1hG4L01025
JE@0Rg1K08]`F`2;L5/208M^F0<0QFaF00<0Q6mN081[FP20JeX00`1mJ5D0101eH4/0JUHn06YF?P1U
DCT206MC>`04065<=`1GAC40EdDa04<b8P808Q@700@04`/002TT7P0Y91h0=C@_0P0V9b80100n?cT0
=CH`03Df<00i>SD204m>C00402LV900Z9RL0:RHW03Xf=`80;RX[00@09b<T04Q6A`18ATL0?3Xk0P0k
>cd0100X9b`0=cPl03Lh?00U9R/2059EFP<0;B/^0`0Y9bX01018ATT0De5D05=AE00`;S4201hL7`04
034_<P0W9BP09bDX02LU:080CTa?00@0:BLZ038`<`0b<3<0;B/^0P0k>C`0100[:B`0>SPk03Xh>`0a
;c8202PV:@04038`<`0^;2l0;R`_05MEF080>cTl00@0>SPk038`<`0b<3<08R0S0P0W9BP00`10?T40
@d5404=1A00303Dc=P0402PV:@0W9BP09bDX02TW:P80=cDh00@0@T1302PV:@0X9RT0:BLZ0P0X9RT0
100O7B00BdU<04]9C00`;S42024O8P0402`Z;@0]:bh0;B/^02XX:`80EE=F0`0`;Rl0100Y9bP0;B//
02d[;00]:b`205YHF@0402XX:@0k>CX0>cTj03Hd=@80=38c00<0:bTZ05YHF@1JF5T00`0S8B80100Z
:2T0928S02@R8`1=Bd`203<a<P0402LU9P0Z:2T0:RPY04i<C@80=cDf00@0?S`m03@b<`0d<S<0<S0a
0P13@D801019AdP0:RPY02XX:@0o?Ch203@b<`0402d[;01THV<0I69S03`k>@80<C0^00@0>CPf03@c
<@0d<c40:2LU0P14@d40101@Cdd0:bXX02/Z:00[:RP203Dd<P<0=cHd00@0Cdi<02HU8`0V9B<0;B`Z
0P0_;R`0100];2X0<c8`03<b<00i>3H201`K6@0402XY9`0l>cT0?3/i038a;`80I6=Q0`17AT@0100i
>3H0@d9004=2@00h=cD204e<BP0403/j>00[:RP0:bXX02l];P80<2h_00@0<c4b030^;`0`;Rl08b4R
0P0l>S/0101CDE80;B//02d[;00]:b`202XX:@0403hl?@1KFEX0FeUJ04=1@P80?cdn00@0>CLh03<a
<P0c<C80<2h_0P13@D801015@d@09bDV02LU9P0i=cP202HT9@0403@b<`0n?3d0?S`m02`Z:`80<Bl`
00@0@3ho02h/;@0^;2d0:bTZ0P0i=cP0100j>3T0EUAE05IDE@12@Cl2034`;P04038a;`0i>3H0>CPf
030_;@80:bXX00@0<S4_03He<`0f=C<0;2/Y0P0R8Al102`[:@40EUEC0@3oool000`0AUAO0`18EV40
1017EF00AUAO04IDG`15Deh204ABG@<0@e5L0`12D5/404=AG0D0A59M1016E5l504MEH0<0B5IQ00@0
A4a?02`d=`0/=3L04QXL0P083A0010083Ph02Ph?00X>3`0;3`h200lA400401<?300F4@d05Q4=00l:
1P804P/500@08AXD02/R6`0[8Q/0;bHO0P0j<2L0100P7ad07ahL01lN700Q81h202PW9@0402l^;00b
<Bl0<S4_02h]:`80:BPV00@08B0N02PW9@0X9bD0;B`Z0P0Z:BL302HU8`0402XY9`0/:bT0;2/Y02/[
:`809RHV00@0:BTY038b<P0b<S80<C4a0P0V9RH0100P82008b<S02<S8`0Q8B4203Ti>@0402HV9P0T
92@092@T02PX:0807AdM00@0<30`03hn?P0n?Sh09BDW0P0Q8B<0100a<C<08b<U02<S9@0_;c4204A4
AP0302TY:`0W9bT09bLY00<08R8T00@0<30b02XZ;00Z:R`071`N0P0T92H0100V9RP0:RX/02XZ;00U
9BD201lO7`<0<C4a00@06ATI01hN7P0N7Qh0;2`/0P0R8R80100`<300820P020P800M7Ad201PH60<0
;bl_00@0:BTY02l_;`0_;bl07QdS0P0U92X0100g=S/081lT020O900R8R@201XJ6P0402@T900?3`l0
3`l?004200801PH400@04Q8@02TY:@0Y:BT09BDU0P0K6ad0100V9RP09RDZ02HU:P0m>ch201TG6004
01`K5`0J6A@06QTD01TH50807adN00@06ALL02LT;@0W92d061HI0P0N71d0100@3Pl07AdK01dM6`0S
91l201`M5`0406Y]IP2HVi80V9^B0:^OT@80Vhmo00@0PgIS07MXD@1gJ540N6Q?0P1hIU00101aG4T0
JeA406]DA01/CCT206Y>>@0406mC?P1hGDP0N5e808=XD`80QfiH00@0PfYD07eTCP1mI4h0M5m>0P1f
HE00101iI540M5m<07AOC01SCST2058n9P0404hj8P1B?RH0DShV040Z5@80@2/H00@0>bXJ02XL3`0Z
70l05P`20P0F3`L0100W91d0=SD`03He<00i>3<2034`:`0402HU8@1:BDD0BTU5040l>`80>CDd00@0
;b//03<_<00c;c00;BTZ0P0f=3D0101FE5D0=3@f03@d=P0^;C8201`M8@0402//<@0]<3D0;C0e044o
@P80<c4d00@0928U030^<@0`;S40>3Hi0P10?T401013@D@071XM01`J7@0_;C0202d[;P04038`<`0^
;2l0;R`_05ABE@80=38e00@0;B/^03Dc=P0e<cH08AlR0P0_;C00100g=CP0:RP[02XX:`0Z:2/203Pf
>@0404]9C00k>C`0>cTl02XX:`80;R`_00@0;B/^02PV:@0X9RT0<S0c0P1:B4/01013@D@08AlR024O
8P0Z:2/202TW:P04030^<@0e<cH0=C<f04M5B080:2HY00@0;2X]02XX:`0Z:2/0:bT/0P1;BD`0100h
=ST0:BLZ02TW:P0`;S4202h/;`0402PV:@1JF5/0FUQK03Le=P80>CLh00@0<Bl`030^;`0`;Rl0:bTZ
0P1@CTl0100Q7b009bDV02LU9P0^;2d202`Z:`0402XX:@1ACe00DDm@02TW:0807adN00@0:RPY024O
800Q7b00C4Y;0P0i=cP0100Y9bP09R@U02HT9@15@d@203<a<P0403Tg>00]:b`0;B//02d[;080>cTj
00@0A49302LU9P0W9BH0=38c0P0/:R/0100V92D0DDm@055?D00n?C/202h]:`0403<b<00d<c40=3<a
02d/:P80A4=100@0DE1>03Lf=00g=S@0<2l]0`0d<c4202l^;00404Y9A`0];2X0;B`Z02d/:P80=SDc
00@0=C@b03hm>`0n?C/0AdI40P0];2X0100f=C<0@T4o0491?`0i>3H206QWI@0403Th=P0j>CL0>STg
03Lf=080=C@b00@0;2/Y040o?@10?cd0>cXh0P0T8b40100_;Bh0;R`]02h/;@0`;Rl202l];P0402HT
9@0n?3d0?S`m05ABD`80=38c00@0?3Xk03<a<P0c<C80?cdn0P1@CTl00`16A4D0<S0a038`<@0302/Y
:P0402l];P0o?Ch0?cdn04Q6A`80:bTZ00@0=cDf02LU9P0W9BH0<Bl`0P10?Sl0100a;c00=38c03@b
<`13@D8203Pf=`0403Le=P0n?3d0?S`m03`j>`80EUAE00@0BdY803He<`0f=C<0:RTW0P0^;B/0100`
;bd0;2/Y02`[:@0Z:BL202TX9P05034`;P0/:bT0;2/Y03Pg=@1LFeT00@3oool000`0AUAO0`18EV40
1017EF00AUAO04IDG`16E5l204ECGP<0A59M00<0@U1K04=AG013DE`00`13DE`304ABG@<0AE=N1016
E5l504MEH0<0B5IQ00@0CEIK03Y3B00j@dP08B/]0P0A6A`0100:3a801`d=00L=3@0;3a02010D5@04
01<?300E4@h05A4>00l;20803@P400@050l;01lH4P0O6180:b@N0P0j<BX0100];2X0:BPV02TX9P0U
928202HU8`0402XY9`0];2X0;B`Z02/Z:080:2LU00@092<Q02DT8P0U9280:2LU0P0[:RP0100Z:BL0
9RDS02HU8`0U928202PW9@0402/[:`0U9BD09BDU02HV9P80;Rh^00@0<C4a02/[:`0[:b/09BDU0P0T
92@0100Q8B40=cLg03Lg=`0/;2`202d];@0402`/;00R8R808R8R038b<P80?Shn00@092@V02/[;@0[
:bd0;2`^0P0I6A/0100];Bl0@T940492A00Z:R`202HV:00401lO8@0Q8B<08B4S02HV:080<30b00@0
6a/M02<S9@0S8bD0820R0P0^;S00100V9RH06a/K01/K6`0[:b/2034a<@04028R8P0S8b<08b<S02PX
:0808B4Q00@0;Rh^02HV9P0V9RH08B4Q0P0D51@0100X:2P0<C4a034a<@0/;2`202@T900402HU:`0[
:S00:bX`03Lf>`80:RT^00@09bLY01lO7`0O7al0:2PX0P0B4Q8010000@001`L500L71@0F5Q@202`/
;00402HV9P0H61X061PJ02<S9@8092<X00@0>CLj01PF5`0H5QL06APD0P0I61<0100H5a<07Q`M01hL
7@0H5Q/202<P:@0401lM800C4A@04a4D00d;30808b8P00@06Q/F01/L500K71@0MgQ`0P2MWi@0102d
Z9X0R7a/08QlK01WFTL206MG@00406mMA@1aG4L0LEa706]D@P80HT/k00@0K4di06]?>P1[CcX0LUI1
0P1kH4/01023J5<0P6MA081WD@1dFdD206U@>P0406M@@01VCcl0ITlo069;>@80FD8`00@0C3HQ04<]
5P13;AH0AS0I0P1>>240101<=R40B3<P04Pc8018=B@204Lg:004034U6@0C2@004`T0014:0P80:R<M
00@0@d0k03Db;@0e<Rd0;BXU0P0a;BX0101LF5D0=C4`03Da<00Y9B@202/W:00403di>P0e<c@0=C<d
04I4A@80DU9D00@0;2/`02d^<P0];S80;2db0P0[;S<01010?T40@d5404=1A00/:Rd203Pf>@0403<a
=00a;c80<Blb04a:C@80:bT/00@09bDX03Hd=`0f=3L09B<V0P0d<SD01018ATT0BDM:04U7BP0Z:2/2
03@b=@0402PV:@0Y9bX0:BLZ03Le>080=38e00@09bDX03Pf>@0h=ST0>CLj0P12@4<0100c<C@0:RP[
02XX:`0]:bh202DS9P04024O8P15@dH0AD=605ABE@80<Blb00@0=38e02TW:P0Y9bX0<Blb0P0Y9bX0
1018ATT08b4T02<Q900^;2l202l]<00402/Y;00n?3l0?S`o03/i?080:RP[0`0/:Rd0100T8RD0FeUL
05]IG00U8b@202/Y:P0402`Z:`0a;c00<Bl`038`<@80FEMH00@0>cTj03Hd=@0f=3D0AdE60P12@440
1014@T<0F5IG05QFE`15@d@202LU9P0403Hd=@0X9RL0:2HW04a:B`80ATA500@0<S0a02HT9@0V92D0
?3Xk0P0c<C80100g=CH0:RPY02XX:@0]:b`203Hd=@0404A2@`0]:b`0;B//02d[;080:RPY00@09bDV
040n?`10?Sl0BdY80P0c<S00100_;R`0<c8`03<b<00^;B/203Pg=@040491?`0i>3H0>CPf02PW9@80
:bXX00@0<c8`02`[:@0/:bT0BDQ60P0f=C<0100N7A/09bHT02LV900/:bT203@c<@0403hm>`0Z:BL0
:RTW02d/:P80<c8`00@0;2/Y059AC`1BDDl0>3Le0P0d<c40100m?3X0;bh/02l^;00[:RP203/j>004
04I5@`0W9R@09bHT02l];P80;R`]00@0<S0a03Dc=00e<c@0<Bl`0P18ATL0101PGUl0CTa=04i<C@0j
>3T202l];P0403Le=P0k>CX0>cTj04i<C@80;bd^00@0:BLX02l];P0_;Bh0928S0P0a;c001014@T<0
;B//02d[;00b<34202XX:@04030^;`16A4D0ATA503@b<`80<Bl`00@0>CLh03Dc=00e<c@0=38c0P0^
;2d0100W9BH0?3Xk03`j>`1?CT`203/j>00402LV900X9bD0:2LU03@c<@80<c8`00@0:RTW02LV900W
9R@081lM0P0i>3H10450?P40?C`j0@3oool000`0AUAO0`18EV401017EF00AUAO04IDG`16E5l204EC
GP<0A59M00<0@U1K04=AG013DE`00`13DE`304ABG@<0AE=N1016E5l504MEH0<0B5IQ00@0CEIK03Y3
B00j@dP08B/]0P0A6A`0100:3a801`d=00L=3@0;3a02010D5@0401<?300E4@h05A4>00l;20803@P4
00@050l;01lH4P0O6180:b@N0P0j<BX0100];2X0:BPV02TX9P0U928202HU8`0402XY9`0];2X0;B`Z
02/Z:080:2LU00@092<Q02DT8P0U9280:2LU0P0[:RP0100Z:BL09RDS02HU8`0U928202PW9@0402/[
:`0U9BD09BDU02HV9P80;Rh^00@0<C4a02/[:`0[:b/09BDU0P0T92@0100Q8B40=cLg03Lg=`0/;2`2
02d];@0402`/;00R8R808R8R038b<P80?Shn00@092@V02/[;@0[:bd0;2`^0P0I6A/0100];Bl0@T94
0492A00Z:R`202HV:00401lO8@0Q8B<08B4S02HV:080<30b00@06a/M02<S9@0S8bD0820R0P0^;S00
100V9RH06a/K01/K6`0[:b/2034a<@04028R8P0S8b<08b<S02PX:0808B4Q00@0;Rh^02HV9P0V9RH0
8B4Q0P0D51@0100X:2P0<C4a034a<@0/;2`202@T900402HU:`0[:S00:bX`03Lf>`80:RT^00@09bLY
01lO7`0O7al0:2PX0P0B4Q8010000@001`L500L71@0F5Q@202`/;00402HV9P0H61X061PJ02<S9@80
92<X00@0>CLj01PF5`0H5QL06APD0P0I61<0100H5a<07Q`M01hL7@0H5Q/202<P:@0401lM800C4A@0
4a4D00d;30808b8P00@06Q/F01/L500K71@0MgQ`0P2MWi@0102dZ9X0R7a/08QlK01WFTL206MG@004
06mMA@1aG4L0LEa706]D@P80HT/k00@0K4di06]?>P1[CcX0LUI10P1kH4/01023J5<0P6MA081WD@1d
FdD206U@>P0406M@@01VCcl0ITlo069;>@80FD8`00@0C3HQ04<]5P13;AH0AS0I0P1>>240101<=R40
B3<P04Pc8018=B@204Lg:004034U6@0C2@004`T0014:0P80:R<M00@0@d0k03Db;@0e<Rd0;BXU0P0a
;BX0101LF5D0=C4`03Da<00Y9B@202/W:00403di>P0e<c@0=C<d04I4A@80DU9D00@0;2/`02d^<P0]
;S80;2db0P0[;S<01010?T40@d5404=1A00/:Rd203Pf>@0403<a=00a;c80<Blb04a:C@80:bT/00@0
9bDX03Hd=`0f=3L09B<V0P0d<SD01018ATT0BDM:04U7BP0Z:2/203@b=@0402PV:@0Y9bX0:BLZ03Le
>080=38e00@09bDX03Pf>@0h=ST0>CLj0P12@4<0100c<C@0:RP[02XX:`0]:bh202DS9P04024O8P15
@dH0AD=605ABE@80<Blb00@0=38e02TW:P0Y9bX0<Blb0P0Y9bX01018ATT08b4T02<Q900^;2l202l]
<00402/Y;00n?3l0?S`o03/i?080:RP[0`0/:Rd0100T8RD0FeUL05]IG00U8b@202/Y:P0402`Z:`0a
;c00<Bl`038`<@80FEMH00@0>cTj03Hd=@0f=3D0AdE60P12@4401014@T<0F5IG05QFE`15@d@202LU
9P0403Hd=@0X9RL0:2HW04a:B`80ATA500@0<S0a02HT9@0V92D0?3Xk0P0c<C80100g=CH0:RPY02XX
:@0]:b`203Hd=@0404A2@`0]:b`0;B//02d[;080:RPY00@09bDV040n?`10?Sl0BdY80P0c<S00100_
;R`0<c8`03<b<00^;B/203Pg=@040491?`0i>3H0>CPf02PW9@80:bXX00@0<c8`02`[:@0/:bT0BDQ6
0P0f=C<0100N7A/09bHT02LV900/:bT203@c<@0403hm>`0Z:BL0:RTW02d/:P80<c8`00@0;2/Y059A
C`1BDDl0>3Le0P0d<c40100m?3X0;bh/02l^;00[:RP203/j>00404I5@`0W9R@09bHT02l];P80;R`]
00@0<S0a03Dc=00e<c@0<Bl`0P18ATL0101PGUl0CTa=04i<C@0j>3T202l];P0403Le=P0k>CX0>cTj
04i<C@80;bd^00@0:BLX02l];P0_;Bh0928S0P0a;c001014@T<0;B//02d[;00b<34202XX:@04030^
;`16A4D0ATA503@b<`80<Bl`00@0>CLh03Dc=00e<c@0=38c0P0^;2d0100W9BH0?3Xk03`j>`1?CT`2
03/j>00402LV900X9bD0:2LU03@c<@80<c8`00@0:RTW02LV900W9R@081lM0P0i>3H10450?P40?C`j
0@3oool000`0AeEP0`18EV4304MEH0<0AUAO0`15Deh404=AG0<0A59M1@15Deh404IDG`D0AeEP0`18
EV40101@Ff40AE1F04E@EP0d?d<2028[<004018L7P094A@02A4D00L?4@802PlB00@030`<010?3@0@
3`d03Pd;0P0=2@H0100<20D04`h:01<>2P0Q71P2030Y8`0403Dd<P0a<2h0<C0^02/Z:08092<Q00@0
8B0N02<R800S8R00:2LU0P0];2X0100_;R`0:2LU02PW9@0W9R@202d/:P0402h]:`0Y:2H0:BPV02HU
8`80:BPV00@0;2`/02LW9`0W9bL09RHV0P0[:b/0100_;bl0;Rh^02h^;P0X:2P202<S8`04028R8P0c
<c<0<c<c030`<080<S8b00@0:b/[020P800P8200;Bd]0P0b<S8302PX:P04028R900H61X061PJ02@T
9P80=cLi00@0;Rh`028R900R8R@0820R0P0T92H0100R8R@0=SHh03Hf>00M7Al202<S9@0401`L7P0a
<C<0<C4c02TY:@806a/K00@0:BTY034a<@0a<C40:b/[0P0Z:RX0100U9BD08R8R028R8P0[:b/202l_
;`0402`/;00F5QH05QHF01lO7`<0:RXZ0P0J6QX00`0S8RP09bH/02LV;00302l^<`0402DU9`0M7Ad0
7AdM024Q8@802@T900@00@8000L71@071`D051@B0P0/;2`0100X:2P071`N01`L7P0V9RP202@S:004
03@b=@0G5AH05aDF01LF4P8061LB00@05aHB01`J6`0L6Q/061HK0P0P7BH0100N7B803Ph@00h>400@
4102028Q7@0400/:1@0K60l06aP?08Z8O080Z:JI00@0RG]^06ePD01]H500IeQ50P1aHDX0101_GDD0
Ie8m06MB?@1MAS@2058h:@0406]<>P1bED<0LUE307aOC@80Nf1=00@0LUM406A;=`1TBcL0Fd8^0P1H
?b/0101I@S80Dc`/05<l;01==R@204Tb800404Xd7`1==b00CCLP054k9080EClX00@0GdHb05Dn;01E
?R`0CcXY0P1GA3D0101HB3T0>B/P03T[800M40L201`B2004038[9@0m=S00?CH`02/S8080<2PU00<0
BDA104]6@`1;AT<00`0e;bl01015@D80;2X[02`Z:`0O7Ah204e=C`0403Dd>@0a<SH0<C8f01/L8@80
;c8g00@0<Blb05=AE01CDE@0?C/n0P0a;c80100`;S40;2X]02`Z;@13@D@204M5B00402DS9P0Z:2/0
:RP[02PV:@80AdE800@0;R`_04U7BP19AdX0>3Hi0P0[:B`0100f=3L0;B/^02d[;P0`;S420490@`04
02/Y;00Y9bX0:BLZ02d[;P80AD=600@0B4I902PV:@0X9RT09B<V0P0]:bh0100Y9bX0=cDh03Le>01@
CU42055?DP0403/i?00T8RD0928U02d[;P809B<V00@0DDmB02@R9@0T8RD0:RP[0P0[:B`0100W9BP0
=38e03@b=@11?d8202XX:`0402TW:P0b<3<0<S0c02/Y;080FeUL00@0<2h_03<a<P0c<C80=38c0P0g
=CH0100d<S<0EE=D05ECE01:B4T203Hd=@0402l];P0Y9bP0:BLX02l];P80=C<d00@0?S`m01@B4`0D
4Q<0:BLX0P0L6Q/0101:B4T0De5B05=ADP0k>CX202TW:00403Le=P0i=cP0>CLh03Xh>@80;bd^00@0
<Bl`03@b<`0d<S<0ATA50P0i=cP0100]:b`0;bd^02l];P0_;Bh203Hd=@0404m>C00g=S@0=cHd02h]
:`80=SDc00@0>3Le03Xi=`0j>CL0@d900P1<BdT0100l>cT0>STg03Xi=`16AD<203hm>`0405QGE@1B
DDl0DU5?02TX9P80<2l]00@0<c8`03@c<@0d<c40>CPf0P0^;B/0100`;bd0<S4_038a;`0/:bT204I5
@`0404M6A00c<S00<c8`0450?P80<S4_00@0=3<a03Xi=`0j>CL0Cdi<0P0_;R`302d[;004038`<@0`
;Rl0<2h_02LU9P80<S0a0`17ADH0100]:b`0:RPY02XX:@0a;c0202PV9`0305=ADP0f=3D0=S@e00<0
<S0a00@0;2X[034_<00a;c00AdE60P0f=3D0100[:BX09R@U02HT9@0S8B8203/i>P0403Dc=00^;2d0
;R`]02l];P80=S@e00@0>SPi02h/;@0^;2d0:RPY0P0n?3d01014@d40?Sdk03hm>`0_;R`202`[:@04
03Th=P0j>CL0>STg03<b<080=C@b00D0;bh/0491?`12@Cl0?Sdk02XY9`010?ooo`003017EF0304QF
H@H0AeEP0`16E5l304=AG0<0A59M0`15Deh704IDG`D0AeEP0`18EV401019F5l0BEIN04UFGP11CUH2
03E0AP0402<^<P0D7B8051dR00TC5@800`d?00@01`P:00X;3@0:2`d03`l?0P0=3@d0100920H03@T6
00d91P0F4Pl201`H5@0402DT8P0Z:BL0:RTW02h]:`80:BPV0`0Q81h0100Z:BL0=3<a03@c<@0f=C<2
02/Z:00402DT8P0Y:2H0:BPV02/Z:080:2LU00@09RDS02TX9P0Y:2H0;Bd]0P0[:b/0100X:2P0:BTY
02TY:@0];Bd202l_;`0402TY:@0R8R808R8R01lO7`80:BTY00@0;Rh^034a<@0a<C409RHV0P0N7Qh0
100W9bL09RHV02HV9P0c<cD202@T9P0401hN800R8R@08R8T01lO8@80;2`^00@0>3Pj028R900R8R@0
9BDW0P0Y:B/0100R8R@0=SHh03Hf>00U9BL2028R900401`L7P0_;c40;bla02d];@807alO00@0:BTY
02h^;P0^;Rh0<30`0P0a<C40100T92@09RHV02HV9P0S8b<2034a<@0403@d=00K6a/06a/K01TI6@80
820P00@0:b/[01dM7@0M7Ad07ahT0P0S8RP0100Y:2d0>STn03Xi?P0Z:R`202@T900402DU9@0;2`/0
2`/;00T:1@8020P600@0410>02PX:00X:2P0:b/[0P0S8bD0100[:bd09B@Y02DT:@0_;C0201LE5P04
01DD400G5Q405aHA01LF4P806aTJ00@06ALL01`I8P0L6B807QdR0P0A4A<0100E5AD08B0L024P7008
1@02020L4@0408^8N@2ESh40UHn105e?@P80GDdn00@0JEQ6075OB@1aGdT0JEDm0P1SCCP0101RB3L0
FT0a05Y0<@1J>R/20654=00406M:>P1OA3<0Gd@c054f9@80BBlN00@0CS@S05Lm;01G?B`0DCLX0P1C
>BX0101F?2/0FCl^05To;P1NAC420617<@0405i5;`1K@R`0Fd8/05m6<P80GTDa00@0Dc`Z05@o;P1D
?bh0Gd`m0P1FA3H0100j:1`0:1P;02PH2`0J3@L204@f<`0402/P7015>cT0AC/i028J5`80IEeK00@0
@Sdj03dh=@0m>3D0@Sho0P0]:b`0100Z:2T0:2PZ02PX:P1GEU/202DV:P0401lP9@0]<3D0;C0e02d[
;P80B4I900@0ATA702XX:`0Z:2/09B<V0P0]:bh0100f=3L0B4I904Q6B@0m>ch202@R9@040490@`0a
;c80<Blb03Dc=P80<Blb00@0C4Y=02@R9@0T8RD0;R`_0P0c<C@0100X9RT0@d5404=1A00`;S42020N
8@04030^<@0c<C@0<c4d04I4A`80<2ha00@0:bT/03`j?@0l>Sd0?C/n0P0^;2l0100e<cH0CTa?04i<
C`12@4<202HT9`0402`Z;@0]:bh0;B/^05iLG`80<Blb00@09R@W02TW:P0Y9bX0:bT/0P0i=cX0101@
CU40<S0c038`<`0/:Rd203/i?00403Tg>P1DDUD0E59E04]9BP80@d5200@0?C/l03Xh>@0j>3T0;R`]
0P0o?Ch01016A4D09R@U02HT9@0g=CH2038`<@0403@b<`0c<C80<c4b04i<C@8081hO00@0=38c02/Y
:P0[:BX0?cdn0P1GEEH0100m>c`0:bTZ02/Y:P0a;c0203`j>`0403/i>P0e<c@0=C<d038`<@80;R`]
00@0@Cm00490@@12@440;R`]0P0b<340100c<C80<2h_030^;`1EE58204E4@P0403Pg=@0l>cT0?3/i
0491?`80?C`j00@0A4=105YIE`1JFEL0>3Le0P0_;R`0100d<c40;B`Z02d/:P0n?C/204e<BP0402HU
8`0/:bT0;2/Y03`k>@80=3<a00<0<S4_034`;P0a<2h00`0d<c40100b<Bl0?Sdk03hm>`1FEE<203<b
<00403`k>@0e=380=C@b03Th=P80=cHd00@0C4]903Dd<P0e=380:RPY0P0]:b`0100e<c@0<Bl`034_
<00U8b@202HT9@0403Pf=`1;BDX0BdU:02`Z:`80<Bl`00@0=S@e028P8@0R8240Cde>0P0n?3d00`0g
=CH0:2HW02PV9`03038`<@0404A2@`13@D80@d5202l];P80;2X[00@09bDV03Xh>@0j>3T0?C/l0P0d
<S<0100[:BX0=S@e03Hd=@0j>3T202XX:@0402/Y:P0l>S/0?3Xk03<b<080?C`j00<0>STg03<b<00c
<S000`0g=S@0100h=cD0@T4o0491?`1AD4h203@c<@409B@R0@0Y:2H10?ooo`003017EF0304QFH@H0
AeEP0`16E5l304=AG0<0A59M0`15Deh704IDG`D0AeEP0`18EV401019F5l0BEIN04UFGP11CUH203E0
AP0402<^<P0D7B8051dR00TC5@800`d?00@01`P:00X;3@0:2`d03`l?0P0=3@d0100920H03@T600d9
1P0F4Pl201`H5@0402DT8P0Z:BL0:RTW02h]:`80:BPV0`0Q81h0100Z:BL0=3<a03@c<@0f=C<202/Z
:00402DT8P0Y:2H0:BPV02/Z:080:2LU00@09RDS02TX9P0Y:2H0;Bd]0P0[:b/0100X:2P0:BTY02TY
:@0];Bd202l_;`0402TY:@0R8R808R8R01lO7`80:BTY00@0;Rh^034a<@0a<C409RHV0P0N7Qh0100W
9bL09RHV02HV9P0c<cD202@T9P0401hN800R8R@08R8T01lO8@80;2`^00@0>3Pj028R900R8R@09BDW
0P0Y:B/0100R8R@0=SHh03Hf>00U9BL2028R900401`L7P0_;c40;bla02d];@807alO00@0:BTY02h^
;P0^;Rh0<30`0P0a<C40100T92@09RHV02HV9P0S8b<2034a<@0403@d=00K6a/06a/K01TI6@80820P
00@0:b/[01dM7@0M7Ad07ahT0P0S8RP0100Y:2d0>STn03Xi?P0Z:R`202@T900402DU9@0;2`/02`/;
00T:1@8020P600@0410>02PX:00X:2P0:b/[0P0S8bD0100[:bd09B@Y02DT:@0_;C0201LE5P0401DD
400G5Q405aHA01LF4P806aTJ00@06ALL01`I8P0L6B807QdR0P0A4A<0100E5AD08B0L024P70081@02
020L4@0408^8N@2ESh40UHn105e?@P80GDdn00@0JEQ6075OB@1aGdT0JEDm0P1SCCP0101RB3L0FT0a
05Y0<@1J>R/20654=00406M:>P1OA3<0Gd@c054f9@80BBlN00@0CS@S05Lm;01G?B`0DCLX0P1C>BX0
101F?2/0FCl^05To;P1NAC420617<@0405i5;`1K@R`0Fd8/05m6<P80GTDa00@0Dc`Z05@o;P1D?bh0
Gd`m0P1FA3H0100j:1`0:1P;02PH2`0J3@L204@f<`0402/P7015>cT0AC/i028J5`80IEeK00@0@Sdj
03dh=@0m>3D0@Sho0P0]:b`0100Z:2T0:2PZ02PX:P1GEU/202DV:P0401lP9@0]<3D0;C0e02d[;P80
B4I900@0ATA702XX:`0Z:2/09B<V0P0]:bh0100f=3L0B4I904Q6B@0m>ch202@R9@040490@`0a;c80
<Blb03Dc=P80<Blb00@0C4Y=02@R9@0T8RD0;R`_0P0c<C@0100X9RT0@d5404=1A00`;S42020N8@04
030^<@0c<C@0<c4d04I4A`80<2ha00@0:bT/03`j?@0l>Sd0?C/n0P0^;2l0100e<cH0CTa?04i<C`12
@4<202HT9`0402`Z;@0]:bh0;B/^05iLG`80<Blb00@09R@W02TW:P0Y9bX0:bT/0P0i=cX0101@CU40
<S0c038`<`0/:Rd203/i?00403Tg>P1DDUD0E59E04]9BP80@d5200@0?C/l03Xh>@0j>3T0;R`]0P0o
?Ch01016A4D09R@U02HT9@0g=CH2038`<@0403@b<`0c<C80<c4b04i<C@8081hO00@0=38c02/Y:P0[
:BX0?cdn0P1GEEH0100m>c`0:bTZ02/Y:P0a;c0203`j>`0403/i>P0e<c@0=C<d038`<@80;R`]00@0
@Cm00490@@12@440;R`]0P0b<340100c<C80<2h_030^;`1EE58204E4@P0403Pg=@0l>cT0?3/i0491
?`80?C`j00@0A4=105YIE`1JFEL0>3Le0P0_;R`0100d<c40;B`Z02d/:P0n?C/204e<BP0402HU8`0/
:bT0;2/Y03`k>@80=3<a00<0<S4_034`;P0a<2h00`0d<c40100b<Bl0?Sdk03hm>`1FEE<203<b<004
03`k>@0e=380=C@b03Th=P80=cHd00@0C4]903Dd<P0e=380:RPY0P0]:b`0100e<c@0<Bl`034_<00U
8b@202HT9@0403Pf=`1;BDX0BdU:02`Z:`80<Bl`00@0=S@e028P8@0R8240Cde>0P0n?3d00`0g=CH0
:2HW02PV9`03038`<@0404A2@`13@D80@d5202l];P80;2X[00@09bDV03Xh>@0j>3T0?C/l0P0d<S<0
100[:BX0=S@e03Hd=@0j>3T202XX:@0402/Y:P0l>S/0?3Xk03<b<080?C`j00<0>STg03<b<00c<S00
0`0g=S@0100h=cD0@T4o0491?`1AD4h203@c<@409B@R0@0Y:2H10?ooo`003017EF0404QFH@P0AeEP
1014DUd304ECGPT0AUAO1@17EF0304QFH@0404AEG@17F600AeQP04QFG`80@4mF00@0=D9:02Pc>@0X
<cT05b8X0P0:5AT00`0;3a802@d@00T=400300h?4@<02@T900@03`h<00h=2`0>3@/051<A0P0R8Al0
100_;R`0<2l]030_;@0X9bD202<R800402LV900^;B/0;Rd[03Dd<P80;2/Y00<092<Q02<R800S8R00
0`0U9280100T8b409B@R02DT8P0[:b/202d];@0402/[:`0W9bL09bLW02XZ:P80;bl_00@0;2`/02@T
900T92@07AdM0P0O7al0100Y:BT0<30`030`<00T92@2024Q8@0402PX:00P8200820P03`l?P80:RX/
00@08R8T02h^<00^;S008B4S0P0X:2X01013@dD0:2PZ02PX:P0X:2X202XZ;00402DU9`0_;c40;bla
02h^<0<0820R0P0Z:R`0100_;bl092@T02@T900Z:RX202LW9`0402l_;`0d=3@0=3@d028R8P80:2PX
00@071`L02/[:`0[:b/0<c<c0P0R8R80100J6QX06a/K01/K6`0_;bl202LW9`0402TX;P0[:S00:bX`
02XY;P80AdI;00@0;bla02TY:@0Y:BT09RHV0P0;2`/0100=3PT02@T700T91`0>3P`202LW9`0402`/
;00U9BL09BDW02XZ;0808b8W00@0:bT/01LE5P0G5AH051<?0P0G5Q40100G5Q806QPI01XH6@0K6Ah2
01XG800401hO900E51T05A@I01DE5@808QhK00@04Ph502LQ5@0W8AD0NG9R0P1YHU00101@@3<0GDdn
05e=?P1/FdT206]I@`04061<=01MA300GD@`05]0;`80E3PZ00@0@b<D04P[6`18:a/0D3<S0P1A=RD0
101@=B@0DCLV054g9P1G?B`205i4<`0405a0<P1TB3X0I4Pj06U>?@80J4dl00<0IDXg06E:=@1UBSD0
0`1VBcH0101UBSL0KeA106mD@@1YD3`205]1<00405U2<P1LACD0G4De05<n;@80B38T00@080h:02`J
5P0/6QH0BCXe0P0h:RL0100b:2H0Bd=004]3@01GDTl203Td<@0402LS900l>S/0?3Xk03Dc=080:BT[
00@0CDaA03dn@P0m?T80<S<h0P0a=3T0100_;C00:RP[02XX:`10?T4203Xh>`0402XX:`0`;S40<2ha
02/Y;0809R@W00@0FEMJ04M5B017ADP0>CLj0P0O7B00100W9BP0<Blb034_<P12@4<203Pf>@0401/I
700`;S40<2ha02XX:`80=S@g00@0=C<f02<Q900S8B@0>SPk0P0K6A`0100j>3/0@T130490@`0k>C`2
03Tg>P0403`j?@0[:B`0:bT/02<Q9080=cDh00@0CTa?03Pf>@0h=ST0<2ha0P0[:B`0101BD5<0?3Xm
03`j?@0]:bh203Le>00402d[;P0j>3/0>SPk055?DP80=38e00@0:RP[03@b=@0d<SD0=cDh0P0m>ch0
1018ATL0=cDf03Le=P0b<34203Dc=00402TW:00c<C80<c4b055?D080:bTZ00@0=38c03<a<P0c<C80
;B//0P0/:R/0101=Bd`07Q`M01hL7@0Y9bP202@R8`04030^;`1DDU<0E59C03Tg>0<0:RPY0P0j>3T0
100g=CH0>3Hg03Pf=`11?d0203Le=P0404Q6A`1EDe@0EE=D03hl?@80@3ho00<0@Cm003/i>P0k>CX0
0`1;BTP0100m?3X0=SDc03He<`0k>SP203<b<00403He<`1@Cdd0D4m=03<b<080:2LU00@0:bXX02h]
:`0^;B/0=3<a0P1@Cdd0100`;bd0;2/Y02`[:@0c<S0202PW9@0402DT8P0];2X0;B`Z02h]:`80;2/Y
00@0<2l]038a;`0b<Bl0H5mM0P0n?C/0100l>cT0?C`j03dl>P0k>SP203Xi=`0304Y9A`12@Cl0@T4o
00<0;2X[00@0=S@e03Pf=`0h=SL0<S0a0P0]:b`0100j>3T0FUQI05YHF@0^;2d203Dc=00403Pf=`0W
9BH09bDV04I4A@80B4I700@0@Cm002PV9`0X9RL0<S0a0P0`;Rl0100i=cP0Cde>04m=CP0i=cP203Pf
=`0403Hd=@0n?3d0?S`m04A2@`80?S`m00@0;2X[03@b<`0d<S<0=C<d0P0R8240100W9BH0=38c03@b
<`0/:bT203/j>00403hm>`0c<S00<c8`02h]:`80<C0^00@0>CPf0491?`12@Cl0Cdi<0P0`;bd102<R
80409B@R0@3oool001P0B5IQ0`14DUd304ECGPX0AUAO1@17EF0304QFH@0404=EG`16Ef40AUMQ04IG
H@80AUAM00@0@e9I03e:DP0mBU80;CY00P0M:S00100J7b<03a@H00lD600<41<200X>4@0400L82P0;
2`/02`/;010@40802PX:00@03Pd;01TH5P0I61H09RDS0P0/:bT0100Z:BL0:2LU02PW9@0Y:2H202`[
:@0403<b<00`;bd0<2l]02TX9P<092<Q0P0W9R@0100U92808B0N024P7P0W9bL302d];@809RHV00@0
9bLW030`<00`<300<30`0P0X:2P0100T92@07QhN01hN7P0Z:RX2034a<@<09BDU00@0:2PX01/K6`0K
6a/0>3Pj0P0g=cT0100V9RP0;Bd_02d];`0S8bD202LW:@0304I6B00[:bd0:b/]00@09RHX0P0S8bD0
100f=SP07QhP01hN800V9RP202@T9P0402l_;`0X:2P0:2PX02/[:`807QhN00@0:2PX03<c<`0c<c<0
820P0P0X:2P0100O7al09RHV02HV9P0];Bd202LW9`0401lO7`0I6AT06ATI030`<080:b/[00@0:RT_
02LV;00W9R`07QdR0P11@4D0100R8R@07alO01lO7`0L71`200410@0400X;1P092@L02@T7014A3`80
:BTY00@0:RXZ01lO8@0O7b408b<U0P0L6b00100Y9bX061HG01PF5`0D4`l201PG4P0401PG4`0K6AX0
6aTJ01`J7`8061DN00@06a`Q01<D600C51P04A4A0P0P7AP0100G4`X081P;020H2`1OETD204e2<004
04<c9P1LBS`0G4Xl06iKBP80KEQ300@0GdTb04dd801==200?B8A0P0h70h010118A@0ARTK04HY6`1A
=2H205Xn<0040655=`1SBCX0HdTj06=9>P80HTPi00@0Id/m06a@@P1/D480KE910P1WC3/0101UBSL0
JDhi06U>>@1^Dch206mD?`0406Q;>@1]D3h0KE0n06mD@@80Je0m00@0IDXg06Q=?01XCC`0JU0o0P1U
BcX01011;2L05`@001L40014<Rh202lP7@0404Hk=`0b:2H0<RPV05MBCP80Ad8n00@06QHG03@b<`0d
<S<0:BLX0P0a<C<0100T8bP0B4U=04Q9C@13A4T202T/<@04038`<`0S8B@08b4T02l]<080BTQ;00@0
@3i102/Y;00[:B`0:2HY0P0Q7b80101<BTd0E59E05ABE@0_;C0202PV:@0401hL7`0f=3L0=S@g02l]
<080@T1300@0:BLZ02HT9`0V92L0<2ha0P0Z:2/0100l>Sd09bDX02LU:00i=cX201`J7@0403<a=01>
C4l0CTa?04U7BP80<Blb0`0^;2l0100Z:2/0;R`_02h/;`16A4L204U7BP0403Le>00W9BP09bDX03Xh
>`80@Cm200@0<2ha03dk?P0m>ch0:BLZ0P0b<3<01016A4L0=38e03@b=@0X9RT202HT9`04034_<P0Y
9bX0:BLZ04I4A@80;R`]00<0:bTZ03Hd=@0f=3D00`0[:BX0101IEeP0<2h_030^;`0[:BX203@b<`04
02PV9`0`;Rl0<2h_05ECE080:RPY0`0Y9bP0100X9RL0DU1A059@D@0g=CH202h/;@0402PV9`0j>3T0
>SPi03@b<`80?C/l00@0=cDf02TW:00Y9bP0=38c0P1;BDX0100c<C80<Bl`034_<00_;Bh202`Z:`04
034`;P16AD<0ATE303/j>080;2/Y00@0=3<a030_;@0`;bd0<S4_0P1>CD/0100h=cD0;Rd[02h]:`0]
;2X203Th=P04034`;P1BDDl0DU5?03/j>080:RTW00@0;B`Z02PW9@0X9bD0:bXX0P0j>CL00`0d<c40
;2/Y02`[:@0303Lf=00405QGE@18AdD0B4M503dl>P80@D0n00@0<C0^03Xi=`0j>CL0ADA20P1<BdT0
100c<C80:RPY02XX:@0]:b`2034_<004038`<@0W9BH09bDV02LU9P80B4I700@0=C<d03@b<`0d<S<0
=38c0P0c<C80100l>S/0D4i?051>C`1=Bd`2038`<@0403/i>P0f=3D0=S@e02h/;@80De5B00@0>CLh
038`<@0b<340<Bl`0P0Z:2T0100i=cP0?3Xk03`j>`0Z:2T203Le=P0403`j>`0]:b`0;B//03/i>P80
@d5200@0=3<a03Th=P0i>3H0>CPf0P0_;R`0100Z:BL0=cHd03Lf=013@T020491?`0503Pg=@0j>CL0
>STg03`k>@0[:RP00@3oool001P0B5IQ0`14DUd304ECGPX0AUAO1@17EF0304QFH@0404=EG`16Ef40
AUMQ04IGH@80AUAM00@0@e9I03e:DP0mBU80;CY00P0M:S00100J7b<03a@H00lD600<41<200X>4@04
00L82P0;2`/02`/;010@40802PX:00@03Pd;01TH5P0I61H09RDS0P0/:bT0100Z:BL0:2LU02PW9@0Y
:2H202`[:@0403<b<00`;bd0<2l]02TX9P<092<Q0P0W9R@0100U92808B0N024P7P0W9bL302d];@80
9RHV00@09bLW030`<00`<300<30`0P0X:2P0100T92@07QhN01hN7P0Z:RX2034a<@<09BDU00@0:2PX
01/K6`0K6a/0>3Pj0P0g=cT0100V9RP0;Bd_02d];`0S8bD202LW:@0304I6B00[:bd0:b/]00@09RHX
0P0S8bD0100f=SP07QhP01hN800V9RP202@T9P0402l_;`0X:2P0:2PX02/[:`807QhN00@0:2PX03<c
<`0c<c<0820P0P0X:2P0100O7al09RHV02HV9P0];Bd202LW9`0401lO7`0I6AT06ATI030`<080:b/[
00@0:RT_02LV;00W9R`07QdR0P11@4D0100R8R@07alO01lO7`0L71`200410@0400X;1P092@L02@T7
014A3`80:BTY00@0:RXZ01lO8@0O7b408b<U0P0L6b00100Y9bX061HG01PF5`0D4`l201PG4P0401PG
4`0K6AX06aTJ01`J7`8061DN00@06a`Q01<D600C51P04A4A0P0P7AP0100G4`X081P;020H2`1OETD2
04e2<00404<c9P1LBS`0G4Xl06iKBP80KEQ300@0GdTb04dd801==200?B8A0P0h70h010118A@0ARTK
04HY6`1A=2H205Xn<0040655=`1SBCX0HdTj06=9>P80HTPi00@0Id/m06a@@P1/D480KE910P1WC3/0
101UBSL0JDhi06U>>@1^Dch206mD?`0406Q;>@1]D3h0KE0n06mD@@80Je0m00@0IDXg06Q=?01XCC`0
JU0o0P1UBcX01011;2L05`@001L40014<Rh202lP7@0404Hk=`0b:2H0<RPV05MBCP80Ad8n00@06QHG
03@b<`0d<S<0:BLX0P0a<C<0100T8bP0B4U=04Q9C@13A4T202T/<@04038`<`0S8B@08b4T02l]<080
BTQ;00@0@3i102/Y;00[:B`0:2HY0P0Q7b80101<BTd0E59E05ABE@0_;C0202PV:@0401hL7`0f=3L0
=S@g02l]<080@T1300@0:BLZ02HT9`0V92L0<2ha0P0Z:2/0100l>Sd09bDX02LU:00i=cX201`J7@04
03<a=01>C4l0CTa?04U7BP80<Blb0`0^;2l0100Z:2/0;R`_02h/;`16A4L204U7BP0403Le>00W9BP0
9bDX03Xh>`80@Cm200@0<2ha03dk?P0m>ch0:BLZ0P0b<3<01016A4L0=38e03@b=@0X9RT202HT9`04
034_<P0Y9bX0:BLZ04I4A@80;R`]00<0:bTZ03Hd=@0f=3D00`0[:BX0101IEeP0<2h_030^;`0[:BX2
03@b<`0402PV9`0`;Rl0<2h_05ECE080:RPY0`0Y9bP0100X9RL0DU1A059@D@0g=CH202h/;@0402PV
9`0j>3T0>SPi03@b<`80?C/l00@0=cDf02TW:00Y9bP0=38c0P1;BDX0100c<C80<Bl`034_<00_;Bh2
02`Z:`04034`;P16AD<0ATE303/j>080;2/Y00@0=3<a030_;@0`;bd0<S4_0P1>CD/0100h=cD0;Rd[
02h]:`0];2X203Th=P04034`;P1BDDl0DU5?03/j>080:RTW00@0;B`Z02PW9@0X9bD0:bXX0P0j>CL0
0`0d<c40;2/Y02`[:@0303Lf=00405QGE@18AdD0B4M503dl>P80@D0n00@0<C0^03Xi=`0j>CL0ADA2
0P1<BdT0100c<C80:RPY02XX:@0]:b`2034_<004038`<@0W9BH09bDV02LU9P80B4I700@0=C<d03@b
<`0d<S<0=38c0P0c<C80100l>S/0D4i?051>C`1=Bd`2038`<@0403/i>P0f=3D0=S@e02h/;@80De5B
00@0>CLh038`<@0b<340<Bl`0P0Z:2T0100i=cP0?3Xk03`j>`0Z:2T203Le=P0403`j>`0]:b`0;B//
03/i>P80@d5200@0=3<a03Th=P0i>3H0>CPf0P0_;R`0100Z:BL0=cHd03Lf=013@T020491?`0503Pg
=@0j>CL0>STg03`k>@0[:RP00@3oool000`0B5IQ00<0AeEP04QFH@18EV402@18EV4304ABG@<0AE=N
1016E5l204MEH0@0AUAO1@17EF0304QFH@<0AEMS00@0@eEO04AEG`14EEl0BUQQ0P1:FF00100oC5@0
;ca402llA00W;30201DJ7P0400/@50072`h01`/>00D6208030d?00@04a<C00d=3@0=3@d01`H40P0=
30X0100E51807ahL01lN700Y:2H2034`;P0403He<`0i>3H0>CPf03<b<080=C@b00@0<C0^02XY9`0Z
:BL0:BPV0P0/:bT0100Y:2H08B0N024P7P0S8b<202`/;00402d];@0V9RH09RHV02HV9P80<30`00@0
<c<c02`/;00/;2`0;bl_0P0S8b<00`0^;Rh0=CDe03De=@0402LW9`805ADE00@0;Rh`0451@`11@D<0
9bLY0P0U9BL0100S8bD09RHX02HV:012@T@202XZ;00402<S9@0Q8B<08B4S02DU9`806ATK00@0>c/m
01dM7`0M7Al0:RX/0P0Q8B<0100^;Rh0:RXZ02XZ:P0Z:RX201HF5P04028R8P0a<C40<C4a01dM7@<0
9bLW0P0U9BD0100Y:BT0:RXZ02XZ:P0S8b<201LG5`0402d];@0Y:BT0:BTY02TX;P8092<Y00@06APM
0450A@11@4D08b<U0P0U9BD0100W9bL03`l?00l?3`051P4200X:200401HF500/;2`0;2`/02LW9`80
61PJ00@071`N01HE6P0F5AX0:2HY0P0H5QL0100D4`l061LB01PG4P0H5a<201/I6P0401dK800H5Ah0
61DN01<F6`804A8F00@030`<01`I500L6A@05Pl50P0C3000101@AC<0BChZ04Tn:P1:>Rd205a:?004
05a9>01?>RD0CcXU040Z4`80<aX600@0=1T804DX6P15:1X0Dc<V0P1E>2X0101J?Bl0GT8d05i2=01P
A3H20616=`040698>@1SBCX0HdTj0696>080IDTk00@0IT/j06M<>`1WC3/0KU=00P1fFdH0101cF4<0
K54l06aA?01/Ccd20613<@0406M:>01eF4H0MEQ607AGA@80L5=100@0K4lm0656<`1QAS<0KEEA0P0c
7QT0100G1000Dd@o05=4?`0U6QH203/a;`0402TT801BCDT0DTe903lk?0809B<T00@0?3Xk02TY:`0Y
:B/0:RT^0P0e=SX0101QHVL07b8W01lR9`0k>C`203Hd=`0401dK7P14@TD0A49504i<C`808AlR00@0
;2X]0490@`12@4<0928U0P0a;c801019AdX0;bd`02l]<010?T4202@R9@0402l]<00/:Rd0;2X]04]9
C0807adP00@0=38e02LU:00W9BP0AD=60P0V92L0100a;c80;R`_02h/;`0^;2l204a:C@0404m=D00f
=3L0=S@g02d[;P80=C<f00@0=S@g030^<@0`;S40:RP[0P1<BTd01011?d80;bd`02l]<00a;c8204I4
A`0402PV:@0]:bh0;B/^02XX:`80;R`_00@0@3i103/i?00k>C`0;bd`0P0T8RD0100d<SD0:2HY02PV
:@1>C4d2034_<00402h/;@0m>c`0?C/l02l];P809bDV00@0FeUJ02l];P0_;Bh0:RPY0P0h=SL0100W
9BH0=38c03@b<`1DDU<202XX:@<071XK00@09bDV05ECE01EDe@0>SPi0P0d<S<0100[:BX0?C/l03dk
?00e<c@20490@@0403dk?00[:BX0:bTZ03@b<`80Cde>00@0>3Hg034_<00a;c00;R`]0P0]:b`0100c
<S00De9@05=BD019B4H2038a;`0403Lf=00c<S00<c8`038a;`80C4]900@0@T4o038a;`0b<Bl09bHT
0P0b<Bl0100N7A/0ADA204E4@P0k>SP302DT8P809RDS00@0<C0^04A3@@14@d40=cHd0P0Z:BL303Xi
=`0404E4@P16AD<0ATE303Lf=080>CPf00@07ahL034`;P0a<2h0>CPf0P1:BDL0100i=cP0:BLX02TW
:00X9RL2030^;`0403/i>P0`;Rl0<2h_02LU9P80ATA500@0@T1103Le=P0g=CH0<c4b0P0m>c`0100e
<c@0D4i?051>C`1?CDh203Hd=@0403Le=P0`;Rl0<2h_01lM7P80Cde>00@0=C<d02`Z:`0/:R/0;B//
0P0M6a`0100V92D0;bd^02l];P0O7Ah202d[;00403<a<P0V92D09R@U03Hd=@80?3Xk00@0?chl03Pg
=@0h=cD0<c8`0P0[:RP0100];2X0@d9004=2@01AD4h204I5@`0501lN700Y:2H0:BPV03Th=P0c<S00
0@3oool000`0AeEP00@0AUAO04MEH017EF00B5IQ1@19Ef800`18EV40AeEP04MEH00<04IDG``0AeEP
3015EfD0100h?D40;C8f02db=P0M8RH2018G6`0300dB5P0<4AD0314E00<02a0D00@00@X700D;2@05
2`T03@d;0P0F5A<0100T81d0;RdX02h]:00e=S0203Tl=@0403Da<00g<c00=c<`03Pd<@80>CHa00@0
=c@_03<`:@0c<2T0;R/T0P0[:240100/;2`09bLW02LW9`0Q8B4201lO7`04024Q8@0V9RH09RHV02TY
:@80:RXZ00@0;b//02dY:P0]:BX08QhO0P0Y9BH0100i=CH0:bLX02/W:00O6a`202dY:P0402HV:00e
=CL0=CDg038b=0807QhP00@09BDW02TY:`0Y:B/08b<U0P0i>C/0100N7R00:2PZ02PX:P0C4aD201lO
8@0403Hf>00X:2X0:2PZ024Q8`808R8T00@09bLY038b=00b<S@0:2PZ0P0R8R@0100Q8B<0:2PZ02PX
:P0d=3H202/[;@0402HV:00O7b407alQ02@T9P80;Rh`00@0:2PZ01TI6`0I6A/07AdO0P0_;c40100Q
8B406ATI01TI6@0V9RH202<S8`0403<c<`0J6QX06QXJ01`L70806ATI00@020P800D51@051@D03`l?
0P0^;Rh0100W9bL08R8R028R8P0V9RH201XJ6P0402HQ7P0K5AD06aDE01<?3P805A4B00@05aDF01LG
6@0G5aT05aLI0P0I61d0100N6Ad03`/<00l;300E3`l2028J5`0401@;1P0R5@d08QD=04@g;P80?2hS
0`14=RT0100j:Q/0:1P802PH200R4@42030M300404Hc8P1E@S40ED8a05Lk;@80F3`^00@0FCd_05/o
<@1K?c40GT8d0P1QACL0101SAcT0I4Pj06A8>P1^CT42071@@`0407ADA@1gEdP0MeM807QIA`80MeQ4
00@0MEI207AE@@1dED40KE4k0P1_D3/0101aDSd0Le<n07=C?P1cDch207AA?@0407=@?01bCc/0LTlk
06YDAP80K5I900@0>2HJ01d<100M30@0>bhX0P0o=S40100]9B<0>C@a03Td<@1MFe`2034_<004030^
<@10@480@41202l^<`80>3Tm00@0=STn04A7C014Ad`0:RPY0P0c<C80100X9RL05A<D01DC5019AdP2
03hl?@04034_<01<BT/0C4Y;02PV9`808R0Q00@0@T1104=1@P13@D80<c4b0P0b<340100X9RL0;B//
02d[;012@4<203Dc=P0402PV:@0T8RD0928U034_<P80>SPk00@081hQ03/i?00k>C`0=S@g0P0Y9bX0
100j>3/0GUaO05iLG`0/:Rd203<a=004038`<`0W9BP09bDX03@b<`80?S`m00@0IF=T03Dc=00e<c@0
<Bl`0P1?CDh0100S8B809B<T02DS900k>CX2030^;`0402d[;01@CTl0D4i?02PV9`80<c4b00@0:BLX
02h/;@0^;2d0EUAE0P0f=3D0100_;Bh0;R`]02h/;@0[:BX202`Z:`0404m=CP12@440@T1103lm?P80
AD=400@0@T1103`j>`0l>S/0EUAE0P1?CDh0100U8b@0<2h_030^;`0d<S<203Le=P0404U7B00c<C80
<c4b038`<@80:bTZ00@0@Cm003Tg>00i=cP0;2X[0P0X9RL0100i=cP0;bd^02l];P1=Bd`2038`<@04
02@R8`0/:R/0;2X[034_<080@d5200@0@3ho03Le=P0g=CH0:bTZ0P0_;Bh303@b<`0405IDE@0V92D0
9R@U02/Y:P80;R`]00@0<2h_03/i>P0k>CX0BDM80P0`;Rl0100_;R`0;Rd[02h]:`0f=C<202l^;004
03Dd<P0^;B/0;Rd[03Lf=080=SDc00@0=3<a05aKF@1LFeT0;Rd[0P0j>CL0100Z:BL0=3<a03@c<@0V
9B<205EDDP04030_;@0^;B/0;Rd[02PW9@80;bh/00@0?Sdk03/j>00k>SP0<S4_0P0f=C<0101CDU00
B4M504Q7A@0`;bd203Pg=@0403dl>P0c<S00<c8`051?C@80<c8`00@09bHT02XY9`0Z:BL08R4O0P0`
;bd0100l>cT0<S4_038a;`0];2X202TX9P0402/Z:00W9R@09bHT02PW9@80;Rd[00@0;bh/02XY9`0Z
:BL0:BPV0P0];2X0100f=C<0=cHd03Lf=00f=C<203Th=P0404M6A01@Cdd0D4m=04I5@`80=3<a00D0
<2l]02LV900W9R@09bHT034`;P010?ooo`003017EF001016E5l0AeEP04MEH018EV4504UGHP0304QF
H@17EF00AeEP00`0AUAO3017EF0<04EGI@0403Pm@@0]<SH0;C8f01dR9P804QLK00<03A8F00`A5@0<
4AD00`0;41@010012PL01@/900D;2@0=3@/201HE4`0402@P7@0^;BP0;RdX03Df<080>C`e00@0=C4`
03Lc<00g<c00>3@a0P0i=S40100g=2l0<c0Y03<`:@0^:b@202/X8@0402`/;00W9bL09bLW024Q8@80
7alO00@08B4Q02HV9P0V9RH0:BTY0P0Z:RX0100_:b`0;BTZ02dY:P0R7Ql202TU9P0403Te=P0[9bP0
:bLX01lK7080;BTZ00@09RHX03De=`0e=CL0<S8d0P0N7R00100U9BL0:BT[02TY:`0S8bD203Ti>`04
01hN800X:2X0:2PZ01<C5@807alQ00@0=SHh02PX:P0X:2X08B4S0P0R8R@0100W9bT0<S8d038b=00X
:2X2028R9004024Q8`0X:2X0:2PZ03@d=P80:b/]00@09RHX01lO8@0O7b4092@V0P0^;S00100X:2X0
6ATK01TI6`0M7Al202l_<@04024Q8@0I6AT06ATI02HV9P808b<S00@0<c<c01XJ6P0J6QX071`L0P0I
6AT0100820P01@D500D51@0?3`l202h^;P0402LW9`0R8R808R8R02HV9P806QXJ00@09R4N01/E5@0K
5AD04`l>0P0E4A80100G5AH05aLI01LG6@0G5aT201TH7@0401hI7@0?2``03`/<01D?3`808QXG00@0
50/6028E3@0R5@d0A3L^0P0l;R<304@f:@0403XZ6`0X60P0:1P8028A0@80<1d<00@0AS<R05E2<@1E
@S40Ec/]0P1H?2h0101I?Bl0Fcla05/o<@1N@S@20655=`0406=7>@1TB3X0I4Pj06i>@@80L51300@0
M5A507MGB01gEdP0N5U70P1gF4@0101eET80M5E107AE@@1]DC/206m@>`04075B?@1cDch0Le<n07=C
?P80M54m00@0Le0l079?>`1bCc/0JUA60P1/ETT0100h9QX07@`401d<100k;RP203lf<@0402dU8`0i
=340>C@a05eKG080<Bl`00@0<2ha0410@P10@480;bhc0P0h>Cd0100f>Ch0A4M<04A7C00Z:2T203<a
<P0402PV9`0E4a@05A<D04U7B080?S`m00@0<Bl`04a:B`1<BT/0:2HW0P0R82401012@440@d5204=1
@P0c<C82038`<@0402PV9`0]:b`0;B//0490@`80=C<f00@0:2HY02@R9@0T8RD0<Blb0P0j>3/0100P
7R40>cTl03/i?00f=3L202TW:P0403Xh>`1NG5l0GUaO02`Z;@80<c4d00@0<S0c02LU:00W9BP0=38c
0P0n?3d0101UHf@0=C<d03Dc=00a;c0204m=CP0402<Q8P0U8b@09B<T03/i>P80<2h_00@0;B//051>
C`1@CTl0:2HW0P0c<C80100Y9bP0;R`]02h/;@1FE5D203Hd=@0402l];P0^;2d0;R`]02/Y:P80;2X[
00@0Cde>0490@@12@440?cdn0P15@d@01012@440?3Xk03`j>`1FE5D204m=CP0402DS900`;Rl0<2h_
03@b<`80=cDf00@0BDM803<a<P0c<C80<S0a0P0[:BX01011?d00>CLh03Tg>00/:R/202PV9`0403Tg
>00_;Bh0;bd^04e;C080<S0a00@0928S02`Z:`0/:R/0<Bl`0P13@D801010?Sl0=cDf03Le=P0[:BX2
02l];P<0=38c00@0EUAE02HT9@0V92D0:bTZ0P0^;2d0100`;Rl0>cTj03/i>P19AdP2030^;`0402l^
;00^;B/0;Rd[03He<`80;bh/00@0=C@b02h]:`0^;B/0=cHd0P0f=C<0100d<c40G5]I05aKF@0^;B/2
03Xi=`0402XY9`0d<c40=3<a02HU8`80EEAB00@0<2l]02h]:`0^;B/0:2LU0P0_;R`0100n?C/0>cXh
03/j>00b<Bl203He<`0405=BD018AdD0B4M5030_;@80>3Le00@0?C`j03<b<00c<S00D4m=0P0c<S00
100W9R@0:RTW02XY9`0R8Al2030_;@0403`k>@0b<Bl0<S4_02d/:P80:BPV00@0:bXX02LV900W9R@0
:2LU0P0^;B/0100_;R`0:RTW02XY9`0Y:2H202d/:P0403He<`0g=S@0=cHd03He<`80>CPf00@0AdI4
051?C@1@Cdd0ATE30P0d<c401@0`;bd09bHT02LV900W9R@0<C0^0040oooo000?04MEH00304QFH@19
Ef80BEMR00<0BEMR00<0B5IQ04MEH017EF003016E5l<04MEH0`0AEMU00@0ADiE03a5C00lAD`0;CHm
0P0N9bh00`0A6R402Q<J00XC6P0300H?5P0400XC400;3`h02`l>00/;2`8030P700@040/801/G500K
5a@0:2LR0P0c=2h0100n?C/0>3Le03Pg=@0b<Bd202h]:@0402h]:00_;bL0;blW02h^9P80;BdU00<0
;Rh^02XZ:P0Z:RX00`0W9bL0100X:2P0:RXZ02XZ:P0Y:BT202PX:00402d[;00^;2d0;R`]02DS9080
:BLX00@0=cDf02h/;@0^;2d08R0Q0P0]:b`0100K6ad09RHX02HV:00Z:R`201dM7`04028R900X:2X0
:2PZ02DU9`80=SHh00@0:2PZ02HV:00V9RP06QXL0P0T92H0100c<cD0;bla02l_<@0Y:B/302HV:080
<S8d00@0:b/]02DU9`0U9BL08b<U0P0W9bT0100c<cD0;Bd_02d];`0Z:R`2024Q8`0402<S9@0];Bl0
;Bd_02XZ;0807AdO00@07QhP02XZ;00Z:R`08b<S0P0P8200100V9RH09bLW02LW9`0_;bl202<S8`04
01lO7`0I6AT06ATI00L71`801@D500@03Ph>02l_;`0_;bl0:2PX0P0U9BD0100Z:RX07alO01lO7`0V
8Ah201dG5`0401HB4@0I5AH06ADF01TG60805ADG00@04a<E01<B5`0C4QL06ALL0P0G5AX0100I5AH0
6aDE01/E5@0C2`P202DJ500403`b:@0Y71<0:A`C03DS5@80;Q`>00@0:1D702/H2@0[60T0?RPJ0P1>
>BP0101A>RX0BC8P04Tb801K?c4205a0<P0405m3=@1QACL0HDDg06A8>P80ITXl00@0Id/m06Q<?P1X
C3h0Jd/n0P1/C3l0101]CCh0Kdm006m?@01`DCl5071A?@04075E?`1cE3l0Le@o07=D?`<0Le<n1@1e
DSh0101VCcl0KEM906eGB@1NBcd203<R600401hA2@0g;2P0=b`X04@l>@80;bLU00@0C4Q903`j>`0l
>S/0=C<d0P0g=cT0100i>3d0;c0d02l`=00K724205MHG@0403/i>P0P7Ql081hO02PV9`80?3Xk00@0
:2HW05eKG01MFe`0?3Xk0P0X9RL0100P7Ql0;B//02d[;00d<S<204Y8B@0403Tg>00P7Ql081hO03@b
<`80<c4b00@0:RP[03Dc=P0e<cH08R0S0P0S8B@0100a;c80@d5404=1A00n?3l203`j?@0403Xh>`0P
7R4081hQ02l]<080=cDh00@0De5D03Dc=P0e<cH0=S@g0P0_;C00100g=CH0@3ho040n?`0o?Ch204a:
B`0402HT9@1BD540DU1A02XX:@80<S0a00@0:RPY02DS900U8b@0:2HW0P1=Bd`0100[:BX0<2h_030^
;`0T8R<202TW:00404U7B00j>3T0>SPi03<a<P80=38c0`0i=cP00`1EDe@0De5B05=ADP0303Pf=`04
02l];P0X9RL0:2HW03Xh>@80ATA500@07Q`M02l];P0_;Bh09R@U0P0h=SL0101BD540>CLh03Tg>00d
<S<2030^;`04044o@00e<c@0=C<d034_<080:2HW00@0=S@e03Dc=00e<c@0DU1A0P0e<c@0100T8R<0
;bd^02l];P0/:R/203Xh>@0404a:B`0d<S<0=38c038`<@80<Bl`00@0?3Xk038`<@0b<340IfEV0P13
@D80100m>c`0?3Xk03`j>`0l>S/203dk?00304]9BP0g=CH0=cDf00<0;B`Z00@0=3<a02h]:`0^;B/0
<c8`0P0/:bT0100c<S00<S4_038a;`0b<Bl205MFE0<0=3<a00@0=C@b038a;`0b<Bl0;Rd[0P15A480
100g=S@0<S4_038a;`0/:bT2030_;@0403He<`0b<Bl0<S4_02/Z:080;2/Y00@0EeID03Xi=`0j>CL0
9B@R0`0d<c42034`;P0405QGE@0l>cT0?3/i03Dd<P80=3<a00@09B@R02d/:P0];2X0=3<a0P0Z:BL0
100`;bd0>CPf03Th=P0T8b4202d/:P<0=SDc00@0<2l]02TX9P0Y:2H092<Q0P0Q81h0100/:bT0:2LU
02PW9@0`;bd204M6A00405=BD019B4H0BDQ603He<`80:RTW00D0?3/i038a;`0b<Bl0<2l]03Pg=@01
0?ooo`003@17EF0204QFH@@0BEMR0P18EV4304MEH0`0AUAO3017EF0<04IFH`0404YGH017E5d0AeAM
041=EP80=45:00@0934j01DR:`0E8R/02QLP0P044AX0100;4A403PlA00h?4@0>30d200/51`0400X1
0P0>20P03PP801/G50809B@P100c<S02034`;00302l^:@0];2L0;B`W00<0;BdU00@0;2`/02/[:`0[
:b/0:b/[0`0/;2`202XZ:P0402HV9P0R8R808R8R02LY:080;Bl^00@09BLV02@V9@0T9RD0<38a0P0]
;bh0100T9RD0:B/Z02T[:P0Q8B<202DU9`0402l_<@0V9RP09RHX024Q8`809BDW00@0820R02LW:@0W
9bT0>SXl0`0U9BL202PX:P0402`/;P0b<S@0<S8d02`/;P<092@V0P0_;c40100];Bl0:2PZ02PX:P0S
8bD202DU9`0402l_<@0^;S00;Rh`02/[;@808B4S00@07alQ02LW:@0W9bT0:BT[0P0Q8B<0100M7Al0
820R020P8P0S8b<202LW9`0402DU9@0/;2`0;2`/02TY:@80<30`00@08b<S01LG5`0G5aL020P80P05
1@D0100<30`0;Rh^02h^;P0T92@202<S8`0402HV9P0M7Ad07AdM024L6@806Q@D00@05Q8A01XF5`0J
5QL06aTJ0P0F5QP0100C4aD04a8G01<B5`0E5Ad201TH7P0401TG700O6a`07a/L01/F4`807AH@00@0
9QdF01D<0`0E30<08@T00P0Y4@D0100d7@l0@BLJ044W6P1?<bD205Tm;`0405`o;`1I?2`0FC`/0695
=@80I4Lg00@0IdXj06Y=?@1ZCCd0K4lo1@1]D40307ADA`<0Le=400@0M5E307IG@`1fEd<0MeQ40P1i
FTH0101fFT@0MeQ307MH@`1fEd8307IF@@D0MeA000@0JDln071FA`1`ETL0Mf5C0P1E@CH0100S4PX0
:A`D02TL500o=30203@[9P0402l[:P1^JV/0KVY[04M5AP80:2HY00@0:BT[02PW;00X9b`0:R/_0P0_
<3@01019AdP08AlP024O800j>3T202XX:@0403@b<`0i=cP0>CLh059@D@80:RPY00@0:bTZ030^;`0`
;Rl081hO0P0e<c@01019AdP0=38c03@b<`0]:b`2038`<@04028P8`0n?3l0?S`o02LU:080:2HY00@0
;2X]040n@@10?T40CD]>0P0c<C@0100a;c80<S0c038`<`0b<3<202LU:004059@D`0g=CP0=cDh02l]
<080=S@g00@0>SPi03Pf=`0h=SL0:bTZ0P1?CDh0100a;c00BDM804U7B00e<c@2030^;`<0;2X[00@0
;bd^051>C`1@CTl0=S@e0P0c<C80100[:BX0<2h_030^;`0i=cP204E3A0<0@Cm000@0C4Y;03`j>`0l
>S/0@T110P18ATL0100^;2d0;B//02d[;00X9RL202`Z:`0403<a<P1@CTl0D4i?028P8@80<S0a00@0
>SPi04i<C@1>C4d0IF=T0P13@D80100f=3D0=38c03@b<`0l>S/202XX:@0403Dc=00X9RL0:2HW03<a
<P80>cTj00@0EeEF0490@@12@440<2h_0P0j>3T0100m>c`0@T110490@@1RH642044o@00403Xh>@0g
=CH0=cDf03lm?P80;bd^00<0Cde>04A2@`14@T<00`0`;Rl0100c<C80:bTZ02/Y:P0o?Ch203Dc=004
02/Z:00/:bT0;2/Y034`;P80;Rd[00@0<S4_02`[:@0/:bT0;bh/0P0^;B/0100b<Bl0CDa:04e<BP13
@T0202/Z:0040450?P0a<2h0<C0^03He<`80=C@b00@0@d9003Lf=00g=S@0<2l]0P0a<2h0100`;bd0
;bh/02l^;00];2X202PW9@0402XY9`1;BTP0BdY803Lf=080<2l]00@0?Sdk01lN700O7Q`0=cHd0P1M
G5X0100Y:2H0<C0^034`;P0Z:BL2038a;`0403@c<@0X9bD0:2LU038a;`80?chl00@0=3<a03ln?00o
?S`0@3lm0P0a<2h0100S8R008R4O028Q7`0W9R@202TX9P04020O7@0U92809B@R03He<`80BdY800@0
D4m=0450?P11@3h0<C0^0P0/:bT01@0e=380=SDc03He<`0l>cT0ADA20040oooo000=04MEH080B5IQ
1019Ef8204QFH@<0AeEP3016E5l<04MEH0`0AUIS00@0BUMP04MDG@17E5d0@4eF0P0d@DX0100T<CX0
5B8[01DR:`0:5b0200@A6P0400/A4@0>3a403PlA00h<3@802`D700@02P4200h8200>20P06aLD0P0U
920403<b<080<C0/00<0;bhY02d/9`0];2L00`0];BD0100/;2`0:b/[02/[:`0[:b/302`/;080:RXZ
00@09RHV028R8P0R8R809bTX0P0];bh0100U9bH092HU02@V9@0`<S4202d_;P0402@V9@0Y:bX0:B/Z
024Q8`809BDW00@0;bla02HV:00V9RP08B4S0P0U9BL0100P82809bLY02LW:@0j>S`302DU9`80:2PZ
00@0;2`^038b=00b<S@0;2`^0`0T92H202l_<@0402d];`0X:2X0:2PZ02<S9@809BDW00@0;bla02h^
<00^;S00:b/]0P0Q8B<0100O7b409bLY02LW:@0Y:B/2024Q8`0401dM7`0P8280820R02<S8`809bLW
00@09BDU02`/;00/;2`0:BTY0P0`<300100S8b<05aLG01LG5`0820P200D51@0400`<300^;Rh0;Rh^
02@T90808b<S00@09RHV01dM7@0M7Ad08A`I0P0J51@0100F4Q406QHG01XF5`0K6AX201HF600401<C
5@0C4QL04a8G01DE7@806APN00@06ALL01lK700O6a`06aHC0P0M5Q00100V7AH05@`301D<0`0Q2@02
02TA1@0403@M3`119aX0@BLJ04lc9@80FCd_00@0G3l_05Tl;01I?2`0HTDe0P1TAcL0101WBSX0JTdm
06Y=?@1/Ccl506e@@0<0M5A70`1cDd@0101dED<0MUM307IG@`1gF4@207UJAP0407IJA01gF4<0MeQ3
07IG@P<0MUI11@1gE400101YCch0L5I7071FA`1gHE<205E1=P0402<B2P0Y71@0:A`D03ld<080=2/V
00@0;b/Z06iZJ`1^JV/0AdE60P0X9RT0100Y:B/0:2L/02PW;00Z:bl202l`=00404U7B00Q7b008AlP
03Xh>@80:RPY00@0=38c03Tg>00i=cP0DU1A0P0Z:2T0100[:BX0<2h_030^;`0P7Ql203Dc=00404U7
B00d<S<0=38c02d[;080<S0a00@08R0S03hl?`0n?3l09bDX0P0X9RT0100/:Rd0@3i1040n@@1=Bdh2
03<a=004034_<P0b<3<0<S0c038`<`809bDX00@0DU1C03Le>00g=CP0;bd`0P0f=3L0100j>3T0>3Hg
03Pf=`0[:BX204m=CP04034_<019AdP0BDM803Dc=080<2h_0`0/:R/0100_;Bh0D4i?051>C`0f=3D2
03<a<P0402/Y:P0`;Rl0<2h_03Tg>080AD=40`11?d00101<BT/0?3Xk03`j>`12@44204Q6A`0402h/
;@0]:b`0;B//02PV9`80;2X[00@0<c4b051>C`1@CTl08R0Q0P0b<340100j>3T0CTa=04i<C@1UHf@2
04=1@P0403Hd=@0d<S<0=38c03`j>`80:RPY00@0=C<d02PV9`0X9RL0<c4b0P0k>CX0101GEEH0@T11
0490@@0`;Rl203Xh>@0403dk?012@440@T11069PH@80@Cm000@0>SPi03Le=P0g=CH0?cdn0P0_;Bh0
0`1?CDh0A49304A2@`03030^;`0403<a<P0[:BX0:bTZ03lm?P80=C<d00@0:bXX02`[:@0/:bT0<C0^
0P0^;B/0100b<Bl0;2/Y02`[:@0_;R`202h]:`04038a;`1=C4X0CDa:04=2@080:bXX00@0@D0n034`
;P0a<2h0=SDc0P0e=3801013@T00=cHd03Lf=00`;bd2034`;P04030_;@0_;R`0;bh/02d/:P80:2LU
00@0:RTW04]:B01;BTP0=cHd0P0`;bd0100n?C/07ahL01lN700g=S@205eLFP0402TX9P0a<2h0<C0^
02XY9`80<S4_00@0=3<a02PW9@0X9bD0<S4_0P0o?S`0100d<c40?chl03ln?010?cd2034`;P0402<R
800R8Al08R4O02LV9080:BPV00@081lM02DT8P0U9280=SDc0P1;BTP0101@Cdd0@D0n0450?P0a<2h2
02`[:@0503Dd<P0f=C<0=SDc03`k>@15A4800@3oool000`0AeEP0`18EV4404UGHP80B5IQ00<0AeEP
04IDG`16E5l03016E5l<04MEH0`0AUIS00@0AEER04IFH`16EV<0AeMT0P16EV<0100nCU/0;Si;02hn
B`0K:cP200hN:`0400T=400=3Q803@hB01<>50805@`A00@04`X=014;3@0A2`d04Ph=0P0D4a40100C
4a<08B4Q024Q8@0b<S0203Xj>00403Lh<`0_<2X0;c0Z02//9P80:R/U00<0:BTW02XZ:00Z:RP00`0Z
:RP0100Y:BL09BDS02DU8`0Q8Al201hN700401lT800W;2P09b`X028W8`80724M00@09BXV02P]:@0X
;BT08RLS0P0Q9R80100S8bD08B4S024Q8`0c<cD2034a<`0402LW:@0[:bd0:b/]02PX:P80:BT[00@0
AdM902HV:00V9RP0;2`^0P0X:2X0100Q8B<0;Rh`02h^<00W9bT201/K7@04024Q8`0Z:R`0:RX/02d]
;`80:BT[00@08b<U028R900R8R@0:BT[0P0^;S00100Z:R`08B4S024Q8`0L71h2020P8P0402DU9`0S
8bD08b<U01`L7P805QHH00@07AdM02TY:@0Y:BT092@T0P0`<300100T92@0>SXj03Xj>P0U9BD2010@
400400X:2P0820P020P800/;2`80;Bd]00@07alO01hN7P0N7Qh07AdM0P0D51@0100K5Q<05@l?01D?
3`0C3`h201PD5@0401XH6@0H61X061PJ01PH6P806QTN00@04QDN01<C6`0C4a/04a8G0P0V92L0100S
7ah030L300`70`0>1`4201<:0`0402dA1@0m8AD0?B4E04h`9@80ECPZ00@0EcLZ05`k;01L>b`0IdHg
0P1`Cch0101VBCT0JD`l06U<?01/Ccl206mB@P03075DA01bEDD0LUE500<0LEA40`1bDTD307=CA003
07AE@`1dED40M5E100<0MEI200@0NEe707]LA`1kG4L0O5e80P1mGDP0101lG4L0O5U507aIA@1iET82
07QE@@0407AH@`1aET<0LEI3075GAP80KUQ:00@0Dcld02LF3P0W5Ph081<;0P11=2h0100o>CT0Ad=2
04M3@P13?ch203`j>`0402h/;`0S8bD08b<U02XY;P80<c8g00@0<Bl`05UGF01IEeP0:BLX0P0c<C80
100_;Bh08b4R02<Q8P0e<c@205IDE@0403Tg>00Z:2T0:RPY02HT9@807Q`M00@0B4I7051>C`1@CTl0
7A/L0P0b<340100Y9bX0?3Xm03`j?@0d<SD2038`<`0402@R9@0f=3L0=S@g04E3AP80=C<f00@07adP
03/i?00k>C`0:BLZ0P0`;S40100c<C@0DU1C059@D`0k>C`203Xh>`0403Le=P0[:BX0:bTZ03Tg>080
=C<d00@0BTQ903hl?@0n?3d0>SPi0P0Q7b0303Pf=`0403@b<`1<BT/0C4Y;03`j>`80<S0a00@0<2h_
03Pf=`0h=SL0?S`m0P1EDe@0100l>S/0;2X[02`Z:`0h=SL202/Y:P0403<a<P1?CDh0Cde>03Xh>@80
=S@e00@0<S0a03Pf=`0h=SL0;bd^0P1DDU<0100P7Ql09B<T02DS900[:BX203Dc=00404Y8B@0b<340
<S0a02/Y:P80<c4b00@0@3ho03/i>P0k>CX0?S`m0P0c<C80100g=CH0>cTj03/i>P1ACe0204]9BP03
03Pf=`0j>3T0>SPi00<0;R`]00@0BTQ9040n?`10?Sl09R@U0P0a;c00100/:R/0:BLX02TW:00k>CX2
04Q6A`<0:bTZ00@0<S0a02HT9@0V92D0>cTj0P0n?3d0100Y:2H0:bXX02/Z:00_;R`2030_;@04038a
;`0`;bd0<2l]030_;@80;bh/00@0=SDc03dl>P0m?3X0FUUG0P0V9B<01013@T00=3<a03@c<@0f=C<2
034`;P0304i=B`0i>3H0>CPf00<0;bh/00@0<2l]03Lf=00g=S@0>CPf0P0_;R`00`0b<Bl0=SDc03He
<`0303/j>004034`;P0W9R@09bHT04M6A080ATE300@0:2LU03Xi=`0j>CL0=C@b0P0b<Bl0100/:bT0
8B0N024P7P0X9bD202d/:P04059AC`17AT@0AdI403<b<0808b8P00@08B0N02PW9@0X9bD0;2/Y0P0Z
:BL0100S8R00=C@b03Dd<P11@3h203hm>`0403Th=P0j>CL0>STg03/j>080=cHd00D0;Rd[03Xi=`0j
>CL0@D0n040o?@010?ooo`003017EF000`18EV40BEMR04UGHP0404UGHP80B5IQ00<0AUAO04ECGP15
Deh03016E5l<04MEH0`0B5IQ00@0AeMT04EEHP15EF80AEER0P1:FVL0101=GFX0AEER04EEHP0dA542
02De@P0401LN900E61d05APM018A5`804PdC00@050dD018=4@0B3A403P`?0P0:2PX010071`L04A4A
014A4@0P820202d]:`0403@d<P0c=2l0<c@_02l`:`80;2dW100[;2L202TZ9@0402HW8P0T9B0092DP
028S7P808B8M00@082DO02P]9`0X;BL092TS0P0M8Q`0100R9b409b`V02L/9P0T:B<2020U7`0402TY
:`0R8R@08R8T03De=`80=SHh00@0:RX/02/[;@0[:bd09bLY0P0U9BL01014A4H09BDW02DU9`0Z:R`2
02<S9@0401TI6`0X:2X0:2PZ02@T9P806ATK00@07alQ02<S9@0S8bD0:RX/0P0Y:B/0100S8bD0820R
020P8P0S8bD202/[;@0402h^<00X:2X0:2PZ024Q8`80820R00@09RHX02TY:`0Y:B/08R8T0P0G5aT0
100F5QH092@T02@T900T92@202l_;`04028R8P0k>c/0>c/k02LW9`8030`<00@02`/;00X:2P0:2PX0
30`<0P0b<S80100P82008B4Q024Q8@0M7Ad201@D500401`G500F41005Q0@018>3@805A4B00@05Q@E
01HF600F5QP06ATK0P0N7B80100D51`05QDK01HE6`0A41D201hJ6`0401TD4@0600001P0001070080
81L@00@0CBlT04l_901?;b@0DS8U0P1K>Bd0101T@c@0K4Ph06a8>01_Bc/2071;>`0406a=>`1^Ccd0
KTlm071A?`80LeA21P1eET@0101eEDP0MeM:07MGBP1jFT/207]KC00407aMB`1jFdL0NU]707UJAP80
MeQ400@0NEe707eNB@1mGTT0P65<0P23Hdh01022HTd0P5e9081MB@1lFDD207UF@P0407YJ@`1eET40
MEI106]?>P80ME]:00@0M5i@04<_9013;b@08a880P0h9al01019A4408A`I024L6@1@BTX204E1@004
02l];P0c<C@0<c4d01dM7`80?Si000@09B<T05iLG@1NG5d0?S`m0P0i=cP0100_;Bh0:2HW02PV9`0P
7Ql2059@D@0403/i>P0[:BX0:bTZ03Hd=@80:BLX00@0;bd^04=1@P13@D80<2h_0P0]:b`0100]:bh0
:2HY02PV:@0h=ST203Tg>P0402<Q900_;C00;bd`03Hd=`80BDM:00@0<Blb02l]<00_;C00928U0P0e
<cH0100]:bh0F5II05QFF@1;BD`202d[;P0402h/;@0Z:2T0:RPY03lm?P807Q`M00@0AdE604Q6A`18
ATL0=cDf0P0Q7b00100^;2d0<S0a038`<@0/:R/203lm?P0403dk?00Z:2T0:RPY02XX:@<0<Bl`0P1A
Ce00100c<C808AlP024O800^;2d202LU9P04034_<01IEeP0FEMH03/i>P80<S0a00@0;B//034_<00a
;c009bDV0P1?CDh0100]:b`09B<T02DS900V92D202LU9P04044o@00f=3D0=S@e02TW:080:bTZ00@0
>SPi04i<C@1>C4d0A4930P0o?Ch0100i=cP0<c4b03<a<P0m>c`204Y8B@0403@b<`0/:R/0;2X[03Tg
>080=S@e00@0?C/l05QFE`1HEUL0:BLX0P0m>c`0100[:BX0>CLh03Tg>00f=3D2059@D@0403<a<P0`
;Rl0<2h_03Dc=080:2HW00@0>3Hg04M5AP17ADH0:RTW0P0/:bT0100^;B/0<S4_038a;`0c<S0203He
<`<0=C@b00@0>cXh030_;@0`;bd0J6MU0P0/:bT0100l>cT0=cHd03Lf=00`;bd203Pg=@04059AC`0i
>3H0>CPf02XY9`80:2LU00@0;B`Z03Xi=`0j>CL0?C`j0P0_;R`0100U9280D4m=051?C@0i>3H201`K
6@0403hm>`0f=C<0=SDc03Dd<P80De9@00@0;bh/04=2@013@T00=SDc0P0V9B<0100Q81h08b8P02<R
800a<2h203<b<00404m>C00k>SP0>cXh02TX9P809bHT00@0;B`Z02`[:@0/:bT09B@R0P0P7ad0100g
=S@0?chl03ln?00m?3X2030_;@0402XY9`0`;bd0<2l]03Pg=@80>CPf00D0@3lm04U8AP19B4H0@D0n
02`[:@010?ooo`003017EF000`18EV40BEMR04UGHP0404UGHP80B5IQ00<0AUAO04ECGP15Deh03016
E5l<04MEH0`0B5IQ00@0AeMT04EEHP15EF80AEER0P1:FVL0101=GFX0AEER04EEHP0dA54202De@P04
01LN900E61d05APM018A5`804PdC00@050dD018=4@0B3A403P`?0P0:2PX010071`L04A4A014A4@0P
820202d]:`0403@d<P0c=2l0<c@_02l`:`80;2dW100[;2L202TZ9@0402HW8P0T9B0092DP028S7P80
8B8M00@082DO02P]9`0X;BL092TS0P0M8Q`0100R9b409b`V02L/9P0T:B<2020U7`0402TY:`0R8R@0
8R8T03De=`80=SHh00@0:RX/02/[;@0[:bd09bLY0P0U9BL01014A4H09BDW02DU9`0Z:R`202<S9@04
01TI6`0X:2X0:2PZ02@T9P806ATK00@07alQ02<S9@0S8bD0:RX/0P0Y:B/0100S8bD0820R020P8P0S
8bD202/[;@0402h^<00X:2X0:2PZ024Q8`80820R00@09RHX02TY:`0Y:B/08R8T0P0G5aT0100F5QH0
92@T02@T900T92@202l_;`04028R8P0k>c/0>c/k02LW9`8030`<00@02`/;00X:2P0:2PX030`<0P0b
<S80100P82008B4Q024Q8@0M7Ad201@D500401`G500F41005Q0@018>3@805A4B00@05Q@E01HF600F
5QP06ATK0P0N7B80100D51`05QDK01HE6`0A41D201hJ6`0401TD4@0600001P000107008081L@00@0
CBlT04l_901?;b@0DS8U0P1K>Bd0101T@c@0K4Ph06a8>01_Bc/2071;>`0406a=>`1^Ccd0KTlm071A
?`80LeA21P1eET@0101eEDP0MeM:07MGBP1jFT/207]KC00407aMB`1jFdL0NU]707UJAP80MeQ400@0
NEe707eNB@1mGTT0P65<0P23Hdh01022HTd0P5e9081MB@1lFDD207UF@P0407YJ@`1eET40MEI106]?
>P80ME]:00@0M5i@04<_9013;b@08a880P0h9al01019A4408A`I024L6@1@BTX204E1@00402l];P0c
<C@0<c4d01dM7`80?Si000@09B<T05iLG@1NG5d0?S`m0P0i=cP0100_;Bh0:2HW02PV9`0P7Ql2059@
D@0403/i>P0[:BX0:bTZ03Hd=@80:BLX00@0;bd^04=1@P13@D80<2h_0P0]:b`0100]:bh0:2HY02PV
:@0h=ST203Tg>P0402<Q900_;C00;bd`03Hd=`80BDM:00@0<Blb02l]<00_;C00928U0P0e<cH0100]
:bh0F5II05QFF@1;BD`202d[;P0402h/;@0Z:2T0:RPY03lm?P807Q`M00@0AdE604Q6A`18ATL0=cDf
0P0Q7b00100^;2d0<S0a038`<@0/:R/203lm?P0403dk?00Z:2T0:RPY02XX:@<0<Bl`0P1ACe00100c
<C808AlP024O800^;2d202LU9P04034_<01IEeP0FEMH03/i>P80<S0a00@0;B//034_<00a;c009bDV
0P1?CDh0100]:b`09B<T02DS900V92D202LU9P04044o@00f=3D0=S@e02TW:080:bTZ00@0>SPi04i<
C@1>C4d0A4930P0o?Ch0100i=cP0<c4b03<a<P0m>c`204Y8B@0403@b<`0/:R/0;2X[03Tg>080=S@e
00@0?C/l05QFE`1HEUL0:BLX0P0m>c`0100[:BX0>CLh03Tg>00f=3D2059@D@0403<a<P0`;Rl0<2h_
03Dc=080:2HW00@0>3Hg04M5AP17ADH0:RTW0P0/:bT0100^;B/0<S4_038a;`0c<S0203He<`<0=C@b
00@0>cXh030_;@0`;bd0J6MU0P0/:bT0100l>cT0=cHd03Lf=00`;bd203Pg=@04059AC`0i>3H0>CPf
02XY9`80:2LU00@0;B`Z03Xi=`0j>CL0?C`j0P0_;R`0100U9280D4m=051?C@0i>3H201`K6@0403hm
>`0f=C<0=SDc03Dd<P80De9@00@0;bh/04=2@013@T00=SDc0P0V9B<0100Q81h08b8P02<R800a<2h2
03<b<00404m>C00k>SP0>cXh02TX9P809bHT00@0;B`Z02`[:@0/:bT09B@R0P0P7ad0100g=S@0?chl
03ln?00m?3X2030_;@0402XY9`0`;bd0<2l]03Pg=@80>CPf00D0@3lm04U8AP19B4H0@D0n02`[:@01
0?ooo`003017EF0304UGHP<0BUQS00@0BEMR04MEH017EF00AUAO0P15Deh<04IDG``0AeEP3019EUl0
101?Gf`0B5QU04QHI@12DUl204ADH@0404YJI`1;FfP0Be]X04ADH@80?4aI00@0=D1602P_=@0X;cD0
61/R0P0@3aD0100=31803`hC00l>4`0;3a8200T>4@04014B500;30h02``>00P:2@804A<B00@08B<P
02d_;00];b`0<C<^0P0_<B`0100^;bX0;c0[02l`:`0^;bX202`]:00402TZ9@0W:2<09bPS02PY9080
:BXU00@0:B/V02d_:P0];bX0:bdX0P0U9b80100V:2<0:R`W02X/9`0Y:bH202HX8`0402`/;P0S8bD0
8b<U038b=080<C4c00<09RHX02@T9P0T92H00`0N7R00100d=3H09RHX02HV:00T92H201dM7`0401XJ
700V9RP09RHX02LW:@808R8T00@07alQ01hN800N7R00:2PZ0P0Z:R`0100V9RP08B4S024Q8`0N7R02
02TY:`0403<c=@0b<S@0<S8d02/[;@808b<U00@09bLY02l_<@0_;c40;2`^0P0P8280100D51@07QhN
01hN7P0U9BD202TY:@04024Q8@0e=CD0=CDe02`/;0803`l?00@01`L700X:2P0:2PX030`<0P0h>3P0
100U9BD0:RXZ02XZ:P0S8b<201XJ6P04024L6@0J51@06Q@D01@@3`805A4B00@04a4B018B500B4Q@0
5ADG0P0K6Ql0100F51T07adP01lM800D414200/60`0400T0000A1P004@H002`O5P80=2LN00@0EcX/
05De:01E=BP0EcLZ0P1R@C80101/BcX0LDdm075=?@1_C3T206e:=`04071A?@1aDSh0LE8n079C?`80
LeA000<0MEI207IG@`1fEd<00`1hFDD0101nGU40P61C081PD`23He@208IVE`0408IWE@24IE40Q6EA
085RCP80Of1<00@0Nem907mPB`1oH4/0PV=>0P25IE001024I4l0Pf1<08=PC01oG4P207aIA@0407]G
?`1kFD00NeU007QH@`80LEE000@0KeE406]EA`1[EDL0Cc`^0P0Z60`0100]9B<0@c/i04<k>@0o>SL2
04@n?P04048n?@0V92D09R@U03Dc=080=38e00@0?C/l034_<00a;c00GeeN0P0c<C80100Y9bP0=cDf
03Le=P0]:b`202/Y:P0403lm?P0f=3D0=S@e02XX:@80>cTj00@08R0Q02XX:@0Z:2T0EeEF0P0J61T0
100`;S407Q`O01hL7`0f=3L203dk?P0402HT9`0Y9bX0:BLZ02PV:@80C4Y=00@0AdE802DS9P0U8bH0
>CLj0P0f=3L0100j>3/0;bd`02l]<018ATT203Dc=P0402`Z:`0a;c00<Bl`034_<0808b4R00@0;B//
05ABD`1DDU<0>3Hg0P0_;Bh0100Q7b00:RPY02XX:@0V92D203/i>P0404U7B00/:R/0;2X[02XX:@80
:bTZ00@08R0Q04]9BP1;BDX0>CLh0P0`;Rl0100e<c@0;R`]02h/;@0`;Rl2055?D00402l];P0Y9bP0
:BLX02/Y:P80<c4b00@0<S0a059@D@1BD540BDM80P0[:BX0100b<340<2h_030^;`1ACe02051>C`04
03@b<`0T8R<0928S02d[;080DDm@00@0=C<d03Tg>00i=cP0<2h_0P0[:BX0100]:b`0BdU:04]9BP0f
=3D202TW:00403Xh>@0g=CH0=cDf030^;`80EeEF00@0;R`]03Hd=@0f=3D0928S0P0d<S<0100Y9bP0
CD]<04e;C00e<c@2030^;`0403@b<`0/:R/0;2X[03Dc=080CTa=00@0;Rd[030_;@0`;bd0;B`Z0P0c
<S00100b<Bl0>CPf03Th=P0f=C<203Pg=@0403Xi=`0/:bT0;2/Y05YIE`80?Sdk00@0=C@b03<b<00c
<S00;B`Z0P0k>SP0101:BDL0>CPf03Th=P0Y:2H202@S8@0402TX9P0c<S00<c8`03<b<0809bHT00@0
8B0N02l^;00_;R`0BDQ60P0g=S@0100W9R@0=3<a03@c<@1>CD/204A3@@0402XY9`0m?3X0?C`j03<b
<080:RTW00@0;bh/03He<`0f=C<0@3lm0P0k>SP0100X9bD0;bh/02l^;00i>3H203Lf=00402HU8`0G
5Q@05aHD01lN7080<C0^00<0ATE303Lf=00g=S@00`0];2X0100/:bT0:2LU02PW9@0[:RP2038a;`05
04e<BP1@Cdd0D4m=0491?`0[:RP00@3oool000`0AeEP1P1:F6<01019Ef80AeEP04MEH015Deh204AB
G@`0AUAO3017EF0<04UFG`0404mLI@1<FF80C5UR04QEGP80AeAM00@0BEIO04]HH@1;F640C5UR0P1;
F6401013E5/0>dQ>03]8CP0^=ch202<X;P0401`O9P0E6Qh05AXN00dG6@802ADE00@051DG00d>400=
3Q001PL90P072@P0100A4a8071hK01`N6`0U9b@202PZ9@0402//9P0];RP0;BhX02h_:@80;BhX00@0
:R/U02TZ900Y:R@0:R/U0P0/;BL0100`;bX0;bhY02l^:@0^;BP202/Z9@0402TX8`0Z:B@0:RTT02XY
9080:BPS00@08b<U01/K7@0K6ad0:2PZ0P0Y:B/0100U9BL09RHX02HV:00R8R@302PX:P80;bla00@0
8B4S01/K7@0K6ad08R8T0P0U9BL0100Z:R`0;2`^02`/;P0S8bD201dM7`0402PX:P0];Bl0;Bd_02/[
;@809BDW00@07QhP02TY:`0Y:B/0<C4c0P0g=cT0100`<3808R8T028R900R8R@202l_<@04030`<P0V
9RP09RHX01PH60806ATI00@0:2PX024Q8@0Q8B408B4Q0P0];Bd0100d=3@061PH01PH60020P8200L7
1`0400X:2P0j>SX0>SXj02DU9@80:b/[00@0820P01LG5`0G5aL07QTF0P0I4a<0100F4Q405a<D01LC
500D4Q<2014A4`0401<C5@0H5a`061LL01XD5P808API00@04`T800/0000;000040<00P0V60d01016
=RT0AcDW04Le9`1E>B/205Tm;`040695=@1/C3d0K4`m06e>?080JdXg00<0KD`i07AC@01dDd000`1f
ED8307ED@@0407IE@P1iF4D0NEQ507aKB080OEa900@0O5a?07eMD01mGE00OUi?0P1oGe00101oH4h0
OUm;07iOB`1mGTX207aMB@0407iRC021HTd0PF9=085RC@80Pf=>00@0PV9=08AQC@24HDd0Pf1<0P22
Gd/01021G440OETo07eI?`20GTD207AE@00406U>>`1jH540NV1A071JC080@bdP00@091`I03Xb;`0j
<Rl05`l=0P1<Ad@0101RGUd071PI01`H6@0n?3d2040n?`0403hl?@0`;Rl0<2h_03Le=P80EUAE00@0
5A<D04E3A015@d@0;B//0P0W9BH0100l>S/0AdE604M5AP0P7Ql203@b<`0402/Y:P0T8R<0928S05=A
DP808AlP00@0;R`_02LU:00W9BP0;R`_0P0m>ch0100]:bh09bDX02LU:00U8bH203Le>004044o@P0d
<SD0=38e040n@@80=38e00@0>SPk01dK7P0M6ah0>CLj0P1DDUD0100h=SL0;bd^02l];P0/:R/203<a
<P0402h/;@11?d00@Cm004M5AP80<c4b00@09B<T02h/;@0^;2d09bDV0P0i=cP0101DDU<0<S0a038`
<@0_;Bh302d[;080Cde>00<0>cTj038`<@0b<3400`0/:R/0100e<c@0DU1A059@D@10?Sl203Pf=`04
03Xh>@0k>CX0>cTj03lm?P80BDM800@0EE=D01lM7P0O7Ah08b4R0P0K6AX01010?Sl0CD]<04e;C00h
=SL202/Y:P0402d[;01BD540DU1A02d[;080<c4b00@0;B//034_<00a;c00;R`]0P1@CTl0100j>3T0
;bd^02l];P0/:R/202/Y:P0402d[;016A4D0ATA504I4A@80;bd^00@0=C<d03<a<P0c<C80:bTZ0P1?
CDh00`11?d00>CLh03Tg>00303Xh>@0403Tg>01GEEH0EeEF03@c<@80<c8`00@0;2/Y03<b<00c<S00
;bh/0P0h=cD0100d<c40=SDc03He<`0d<c42038a;`0403Lf=01DDe40E5=A038a;`80:RTW00@0;Rd[
03@c<@0d<c40>CPf0P0j>CL0100a<2h0:BPV02TX9P0[:RP202d/:P0402TX9P0W9R@09bHT024P7P80
;Rd[00@0>3Le02XY9`0Z:BL0?chl0P18AdD0100m?3X0<C0^034`;P0W9R@203He<`0403@c<@0o?S`0
?chl04Y9A`80@3lm00@0=3<a02@S8@0T8b4061LE0P0/:bT0100n?C/0>cXh03/j>00Y:2H2020O7@04
02`[:@0n?C/0?Sdk03Lf=080;bh/00@0;2/Y02l^;00_;R`0;bh/0P0[:RP0100`;bd0>STg03Xi=`10
?cd203hm>`40>cXh0@0i>3H10?ooo`003017EF0604YHH`0404UGHP17EF00AeEP04ECGP80A59M3016
E5l<04MEH0`0BEIO00@0CeaU04aIHP1<FF80B5EN0P17E5d01019EUl0BeQQ04]HH@1<FF8204]HH@04
04=DF`0kB4h0>dQ>02hg?P808bP^00@071lV01DJ7P0E6Qh03ALI0P095AD0100D5AL03@h@00d>4006
1`T200L92004014C4P0L7Q/071hK02DW9080:2XU00@0:b`V02d^:00];RP0;RlY0P0];RP0100Z:bD0
:BXT02TZ900Z:bD202`]9`04030_:P0_;RT0;bhY02h]:080:bXU00@0:BPS02XY900Z:B@0:RTT0P0Y
:2<0100S8bD06a/M01/K7@0X:2X202TY:`0402DU9`0V9RP09RHX028R90<0:2PZ0P0_;c40100Q8B<0
6a/M01/K7@0R8R@202DU9`0402XZ;00/;2h0;2`^02<S9@807AdO00@0:2PZ02d];`0];Bl0:b/]0P0U
9BL0100N7R00:BT[02TY:`0a<C<203Lg>@04030`<P0R8R@08R8T028R9080;bla00@0<30b02HV:00V
9RP061PH0P0I6AT0100X:2P08B4Q024Q8@0Q8B4202d];@0403@d=00H61P061PH00820P801`L700@0
2PX:03Xj>P0j>SX09BDU0P0[:b/0100P82005aLG01LG5`0N6AH201TC4`0401HB4@0G4a@05a<D01@B
4`804A4C00@04a<E01PG700H5a`06Q@F0P0Q61T0100C2@P02`0000/0000@0`0202HH3@0404Hf:@17
=BL0AcDW05Di:`80FCd_00@0HTDe06a<?@1/C3d0KDhl0P1[BSL00`1]C3T0M5=007AC@00307IE@P<0
MEA100@0MUE207UHA@1iF4D0O5]80P1mG4T0101lG4l0OEe@07eMD01nGTl207mOD00407mPCP1nGd/0
OUm;07eNBP80O5e900@0OV9<085RC@21HTd0PF9=0P23Hdh01022HTd0Q65=08AQC@23H4`2089OB`04
085L@@1mFCl0OETo081NA@80M5E000@0JDhk07YPD@1jH540L5Y<0P13;B00100T71T0>S8_03Xb;`0G
3`d204a7A004069NG@0L61T071PI03hl?@80@3ho00@0?S`m030^;`0`;Rl0=cDf0P1FE5D0100E4a@0
AD=404E3A00]:b`202LU9P0403`j>`17ADH0AdE6020N7`80=38c00@0:bTZ02@R8`0T8R<0De5B0P0Q
7b00100^;2l09bDX02LU:00^;2l203dk?P0402d[;P0W9BP09bDX02DS9P80=cDh00@0@Cm203@b=@0d
<SD0@3i10P0d<SD0100j>3/07A/N01dK7P0i=cX205ABE@0403Pf=`0_;Bh0;bd^02`Z:`80<c4b00@0
;R`]044o@011?d00AdE60P0c<C80100U8b@0;R`]02h/;@0W9BH203Tg>00405ABD`0b<340<S0a02l]
;P<0;B//0P1?CDh00`0k>CX0<S0a038`<@0302`Z:`0403Dc=01BD540DU1A040n?`80>3Hg00@0>SPi
03/i>P0k>CX0?cdn0P19AdP0101EDe@07adN01lM7P0S8B8201/I6P04040n?`1=Bd`0CD]<03Pf=`80
:bTZ00@0;B//059@D@1BD540;B//0P0c<C80100]:b`0<Bl`034_<00^;2d2051>C`0403Xh>@0_;Bh0
;bd^02`Z:`80:bTZ00@0;B//04I4A@16A4D0ATA50P0_;Bh0100e<c@0<c4b03<a<P0[:BX204m=CP03
044o@00i=cP0>CLh00<0>SPi00@0>CLh05MEEP1GEEH0=3<a0P0c<S00100/:bT0<c8`03<b<00_;R`2
03Pg=@0403@c<@0f=C<0=SDc03@c<@80<S4_00@0=cHd05ACD@1DDe40<S4_0P0Z:BL0100^;B/0=3<a
03@c<@0i>3H203Xi=`04034`;P0Y:2H0:BPV02/Z:080;B`Z00@0:BPV02LV900W9R@08B0N0P0^;B/0
100h=cD0:RTW02XY9`0o?S`204Q7A@0403dl>P0a<2h0<C0^02LV9080=SDc00@0=3<a03ln?00o?S`0
BTU70P10?cd0100d<c4092<Q02@S8@0H5aD202`[:@0403hm>`0k>SP0>cXh02TX9P8081lM00@0;2/Y
03hm>`0n?C/0=cHd0P0_;R`0100/:bT0;bh/02l^;00_;R`202/Z:004030_;@0j>CL0>STg040o?@80
?Sdk0@0k>SP103Th=P40oooo000<04MEH0H0BUQS00@0BEMR04MEH017EF00AE=N0P14DUd<04IDG``0
AeEP3019EUl00`15CeP0BUAM04YDG@0304mIHP0404aFG`1:E5d0BUAM04eGH080DE]T00@0@EIK04AE
G014EE`0AE9H0P11BU40100d>d408R/`028[<00C7R0200PH600400P<3`0=4A@03A4D014E5P803a<B
00@02@d<00X?2`0:3`/04QLC0P0K81`0100T9Al09bPR02LX8P0Z:bD202//9P0402TZ900W:2809bPR
02LX8P80:BXT00@0<RlZ02h[9P0^:bH0;BXU0P0/:B@00`0X9B009b@O02LT7`0302PU800402LW:@0P
8280820R02PX:P809bLY00@09RHX02DU9`0U9BL0820R0P0Z:R`0100U9BL0>c/m03/k?@0T92H201dM
7`0402PX:P0T92H092@V02/[;@80<C4c00@09RHX01dM7`0M7Al0:BT[0`0_;c4202PX:P0401lO8@0Z
:R`0:RX/02`/;P80=CDg00@0<30b01hN800N7R006a/M0P0Y:B/0100^;S009RHX02HV:00N7Qh201PH
600402XZ:P0K6a/06a/K024Q8@809bLW00@0>c/k024Q8@0Q8B4000000P0410@010071`L0>3Ph03Ph
>00P820202@T900401LG5`0=3@d03@d=01PC40805@l?00@05A4@01TE5P0I5AH05aDF0P0D51H0100E
5AL06APM01TH7@0P5QD201H<2P0400`1000I30@06@`402DD2P80=B<G0`1@?Bh0101QAcP0Gd@c05m4
<`1RAcH206]>?00406mB@01`DCd0L54m075B?P80MUE200@0NUU607UHA@1iF4D0MeI30P1fED80101g
ET<0NUU607YIAP1nGDX2085PC@0408ATE`23HeH0Pf=F089RD`80PF5B00@0PF9@089SC`22Hdl0PV=?
0P23I5001022IU00PV=>089SCP20HD`207mOBP04081PB`23H4`0Pf1<08ISC`80R6EA00@0S6A:07iI
?P1nFCh0O5Q00P1kFd@0101cEd80L5I5071FA@1aFTX206aFB00402`T8@0Q6AH08ATF03de<`80;RTV
00@0BD=3059>C@1BCTd0<bl`0P0^;2d0100Z:2T0=cDf03Le=P0P7Ql205ECE00404M5AP0H5QL061HG
03Hd=@80<c4b00@0:BLX05ECE01EDe@0<S0a0P0U8b@0100d<S<08R0Q028P8@0[:BX204E3A004028P
8`0^;2l0;R`_02<Q9080>SPk00@0=C<f02XX:`0Z:2/0;R`_0P0W9BP01017ADP0FEMJ05UGFP0b<3<2
034_<P0402h/;`0`;S40<2ha024O8P80EE=F00@0ATA502<Q8P0S8B80=cDf0P0k>CX01017ADH07adN
01lM7P1JF5T202TW:00402l];P0c<C80<c4b02DS9080<S0a00@0DU1A02h/;@0^;2d0;bd^0P0]:b`0
100T8R<0ATA504I4A@0m>c`203/i>P<0<2h_00@0>3Hg04]9BP1;BDX0DDm@0P13@D80100l>S/0<S0a
038`<@0j>3T203/i>P0405aJF`0M6a`07A/L03Tg>080928S00@0?C/l04Y8B@1:B4T0?C/l0P0e<c@0
100Z:2T0@d5204=1@P0h=SL203`j>`0403Hd=@0o?Ch0?cdn03@b<`80Cde>00<0=C<d02d[;00]:b`0
0`0f=3D01014@T<0@T110490@@1VI6D202l];P0404U7B00a;c00<Bl`03Dc=080EE=D00@0C4Y;03dk
?00m>c`0=cDf0P0i=cP0100^;2d0CD]<04e;C00g=S@203He<`0402/Z:00a<2h0<C0^02/Z:080=C@b
00@0<C0^03<b<00c<S00;Rd[0P0h=cD0100H5aD0I6=Q06ASH@0d<c42024P7P04038a;`0Z:BL0:RTW
02XY9`80>STg00@0>cXh038a;`0b<Bl0<C0^0P0];2X0100X9bD0;B`Z02d/:P0Y:2H2028Q7`0404=2
@00i>3H0>CPf02PW9@80<S4_00@0?chl034`;P0a<2h07ahL0P0X9bD0100/:bT0BDQ604U8AP1KFUP2
04E4@P0402`[:@0I61H06APF02/Z:080:2LU00@09bHT038a;`0b<Bl0ATE30P1@Cdd01011@3h0:RTW
02XY9`0N7A/202l^;00403Pg=@0`;bd0<2l]02XY9`80=3<a00@0@d9004a;B@1<BdT0;bh/0P0X9bD1
02d/:P40?3/i0@3oool000`0AeEP0`18EV4404MEH0D0AUAO3017EF0@04IDG`D0AeEP0`18EV4<04UG
HP0304QJJ015EfD0AEMU00<0@UAR00@0?e5O03I8EP0fB5H0:C]90P0N<3h0100G7bX03aPQ00lH8@0:
4aX200dE600400`E50083P`020h<00H<20802@h800@04aDD01XL6@0J71T092HS0P0];RT0100a<2`0
;RdX02h]:00Z9b8202DR7@0402HV900Y:RD0:BXU02`[9`80;2/V00@0:bPS02XW800Z9b00;2LQ0P0]
:28302DU8`0402XZ:00^;R`0;Rh/02TY9`80820N00@07AdK024Q7`0Q8Al06QXJ0P0];Bd0100/;2`0
:BTY02TY:@0b<S8202/[:`0402DU9@0c<c<0<c<c01lO8@808R8T00@092@V02HV:00V9RP0:RX/0P0[
:bd0100R8R@05QHH01HF600O7b4202XZ;00402TY:`0L71h071`N01PH6P808b<U00@0:b/]02TY:`0Y
:B/071`L0P0K6a/0100>3Ph08B4Q024Q8@0R8R8202DU9@030410@00[:b/0:b/[00<0000000@00`<3
038b<P0b<S808B4Q0P0O7al0100H61P03`l?00l?3`0E5aH201TJ700401@C6@0@3aL040lG01LF7P80
5aHL00@04a<E01DE4`0E5A<060d70P0I30@0100O3PH0:aL<02/G300m9QH204db7`0405Tj9@1N?RL0
GShW06A5<@80ITLc00@0JT/g06i?>`1^Cc/0LE8n1@1cE400101jFdL0NEY607UJAP1iFTH207YKA`04
07]LB01nGd/0OUm;085RCP80PV=?00<0Q6EA089SC`22Hdl00`1oH4`01022Hdl0QFIB08EVDP26Ie<2
08EVDP0408UVD`28IE80R6EB08ERC`80Pf1<0`22GdT308=QB00408]N@@29GD00REe008IL@`80Q5a3
00@0P5a607aKB01lFdP0NU]90P1gFTX0101IB4009A@=02DD3@11=2h202lT8P0402hU9P1UH6@0IF1T
03Hd>@809bH/00@0=Bla03La<`0g<C<09R8S0P0o>c`0101?CDh08R8R028R8P0b<S8203/k>`0402l]
;P0X9RL0:2HW04]9BP80>CLh00@081hO02TW:00Y9bP0;2X[0P1;BDX0100Z:2/0;2X]02`Z;@0^;2l2
02XX:`0403lm@00^;2l0;R`_02HT9`80<S0c00@0?C/n05eKGP1MFeh0<S0c0P0a;c80100]:bh0=S@g
03Hd=`0]:bh201lM800405=ADP1BD540DU1A038`<@80>3Hg00@0>SPi01lM7P0O7Ah0;R`]0P19AdP0
100`;Rl08b4R02<Q8P0i=cP202LU9P0404m=CP0k>CX0>cTj02h/;@80;2X[0`0[:BX0101HEUL0=cDf
03Le=P14@T<203dk?0040490@@0j>3T0>SPi05aJF`809R@U00@0>SPi02XX:@0Z:2T0<Bl`0P0g=CH0
101JF5T08R0Q028P8@0g=CH203Dc=00403hl?@1MFe`0GE]L03dk?0809bDV00@0;2X[03<a<P0c<C80
DDm@0P0b<340100W9BH0:BLX02TW:00Y9bP203Tg>00405QFE`13@D80@d520490@@80?C/l00@0@T11
038`<@0b<340DU1A0P0n?3d0100[:BX0;B//02d[;00c<C82034_<00405MEEP0_;Bh0;bd^03hl?@80
;B//00@0928S03Pf=`0h=SL0A4=10P0X9bD0100];2X0:RTW02XY9`0d<c4303Pg=@80<S4_00@0<C0^
02/Z:00[:RP0=SDc0P13@T00100k>SP0;Rd[02h]:`0Z:BL202TX9P0403Lf=00k>SP0>cXh02l^;080
<C0^00@0=C@b03Th=P0i>3H0?chl0P0d<c40100`;bd0?3/i03`k>@0f=C<202TX9P<0:2LU00@09bHT
02`[:@0/:bT0>cXh0P0_;R`01011@3h0BdY804]:B00b<Bl202d/:P<0?3/i00@092<Q02TX9P0Y:2H0
8B0N0P0g=S@01014@d40BDQ604U8AP0Z:BL2020O7@0402LV900/:bT0;2/Y02l^;080<2l]00@0=SDc
040o?@10?cd0A4=10P12@Cl01@0V9B<0;2/Y02`[:@0`;bd0;bh/0040oooo000<04MEH0<0B5IQ1017
EF0504IDG``0AeEP4016E5l504MEH0<0B5IQ3019Ef800`18FVP0AEMU04EGI@03049DHP0403mAG`0f
B5H0=TQF02TkB@807S0n00@05alZ00lH8@0?62402Q<J0P0=5AP0100<5A@020h<00P>300630P200T>
200401<E500J71T06Q`I02@V8`80;BhY00@0<C0/02h]:00^;BP0:RLR0P0U8Qd0100V9R@0:BXU02TZ
9@0/:bL202`[9P0402/X8`0Z9b00:RLP02`W8@80;BPR0`0U9B<0100Z:RP0;Rh/02h^;00Y:BL2020P
7P0401dM6`0Q8Al08B4O01XJ6P80;Bd]00@0;2`/02TY:@0Y:BT0<S8b0P0[:b/0100U9BD0<c<c03<c
<`0O7b42028R900402@T9P0V9RP09RHX02XZ;080:b/]00@08R8T01HF600F5QP07alQ0P0Z:R`0100Y
:B/071`N01`L7P0H61X202<S9@0402/[;@0Y:B/0:BT[01`L70806a/K00@03Ph>024Q8@0Q8B408R8R
0P0U9BD00`10@400:b/[02/[:`030000000400<30`0b<S80<S8b024Q8@807alO00@061PH00l?3`0?
3`l05ALF0P0I6Q`0100D4aT040lG010?5`0G5Qh201LF700401<C5@0E5A<05ADC01P=1`806@`400@0
7`h602/G300[5``0?BHF0P1=<Ql0101I>RD0GShW05hn9`1TAC4206I7<`0406Y;=`1^Cc/0KTlk075B
?PD0LeA000@0NU]707UJAP1iFTH0NEY60P1jFdL0101kG4P0OUm;07iOB`21HTh2089SC`0308AUD@22
Hdl0PV=?00<0Of1<00@0PV=?08EVDP25IU80QVMC0P25IU801029IU<0R6EB08QUDP25HTl208=PC0<0
PUm90`23HDP0102;GT40REe008UM@026G4<208AL@`04081LAP1lFdP0O5]807YKB@80MeY:00@0FDQ0
02DD3@0U50d0@C@^0P0_9280100^9BH0IF1T06EPI00f=3T202LV;00403D_<@0g<C<0=c4c02HR8`80
?c/l00@0Cde>028R8P0R8R80<S8b0P0k>c/0100_;Bh0:2HW02PV9`1;BDX203Tg>004020N7`0Y9bP0
:BLX02`Z:`80BdU:00@0:RP[02`Z;@0/:Rd0;R`_0P0Z:2/0100o?D00;R`_02h/;`0V92L2038`<`04
03dk?P1MFeh0GE]N038`<`80<Blb00@0;B/^03Hd=`0f=3L0;B/^0P0O7B00101CDE80DU1A059@D@0b
<34203Pf=`0403Xh>@0O7Ah07adN02h/;@80BDM800@0<2h_02<Q8P0S8B80>CLh0P0W9BH0101?CDh0
>cTj03/i>P0^;2d202`Z:`<0:bTZ00@0F5IG03Le=P0g=CH0A4930P0m>c`01012@440>SPi03Xh>@1L
FU/202HT9@0403Xh>@0Z:2T0:RPY034_<080=cDf00@0FUQI028P8@0R8240=cDf0P0e<c@0100n?3d0
GE]L05eKG00m>c`202LU9P0402`Z:`0c<C80<c4b055?D080<S0a00@09bDV02TW:00Y9bP0:BLX0P0i
=cP0101HEUL0@d5204=1@P12@44203dk?0040490@@0b<340<S0a059@D@80?S`m00@0:bTZ02d[;00]
:b`0<c4b0P0a;c00101GEEH0;bd^02l];P0n?3d202d[;00402@R8`0h=SL0>3Hg04A3@@80:2LU00@0
;B`Z02XY9`0Z:BL0=3<a0`0h=cD2038a;`04034`;P0[:RP0:bXX03He<`80@d9000@0>cXh02h]:`0^
;B/0:RTW0P0Y:2H0100g=S@0>cXh03/j>00_;R`2034`;P0403Dd<P0i>3H0>CPf03ln?080=3<a00@0
<2l]03`k>@0l>cT0=SDc0P0Y:2H302PW9@0402LV900/:bT0;2/Y03/j>080;bh/00@0@D0n04]:B01;
BTP0<S4_0P0];2X303`k>@0402@S8@0Y:2H0:BPV024P7P80=cHd00@0A4=104U8AP19B4H0:RTW0P0P
7ad0100W9R@0;2/Y02`[:@0_;R`2030_;@0403He<`10?cd0@3lm04A3@@80@T4o00D09RDS02`[:@0/
:bT0<2l]02l^;0010?ooo`003017EF0304QFH@@0AeEP8@16E5l304MEH0D0B5IQ3019Ef800`18FVP0
AEMU04EGI@0304=EH`0404EGI@12E680@UAR03a>G080=DME00@0;SQ2028/=@0R;3D05QlV0P0@5ad0
100;4aD020h>00P>3P083PX200/A3@0400<71P0:30/02P`;01<E50807AdK00@09RLR02h]:@0^;BT0
<S4/0P0c<Rd302`/:P0402//9`0Z:BD0:RTU02TV8@80:BHO00@0;RTS034/9P0a;2H0:BTW0P0X:2H0
100/;2X0<S8`038b<00b<S0202d]:`0402`/:P0_;bd0;bl]02PX:080<c<c00@0:RXZ02<S8`0S8b<0
:BTY0P0V9RH0100T92@0<c<c03<c<`0f=SP203<c=@0402/[;@0S8bD08b<U028R90809BDW00@08b<U
01dM7`0M7Al07alQ0P0Y:B/0100/;2h09BDW02DU9`0Q8B<202DU9`0302LW:@0U9BL09BDW00<07QhN
00@03@d=01/K6`0K6a/09bLW0P0S8b<3038b<P0400<30`000000000000D51@80<c<c00@092@T028R
8P0R8R806a/K0P0A4A40100A4a805QHH01HF600E51X201HE7@0401/J800F5AX05QDJ00l=3P804Q4=
00@04PD001X;100J2`@0:AP>0P0l:A/0101?=BH0FSd[05Xm:`1O@2/20610:@0406E6<P1XBCD0J4Te
06a=>@80LE8n00<0MEI207MHA01gF4@00`1gF4@0101oH4`0P65=081QC@20HDd2089SC`0408AUD@26
Ie<0QVMC08QYE@80RFYF00@0PfA@085RCP21HTh0Of1<0P21HTh01023I500QFIB08EVDP25IU8208AU
D@0408YWE029IU<0RFIC08QUD@80QfA@0`26Hdd308ITB`0408iQA02<H4D0S61508UOAP80Pei400@0
P5a607YJA@1jFTD0N5U70P1gF4H0101]Fe40IE9;06EBB`0Y6Q<202XL6@0403/a<016@480AT1204@o
@`80;bdb00@0;bT[03<];`0c;Bl0>SHg0P0Y9BH0101:B4T0AdM704M7A`0W9bL203/k>`0402DS900[
:BX0:bTZ03dk?080@Cm000@0=38c02HT9@0V92D0:bTZ0P10?Sl0100l>Sd09B<V02DS9P0d<SD201hL
7`0403Dc=P0k>C`0>cTl01lM8080>CLj00@0=38e02`Z;@0/:Rd0EUAG0P0^;2l0100i=cX0;B/^02d[
;P0/:Rd2020N8@04038`<@1;BDX0BdU:04]9BP80<Bl`00@0:BLX02/Y:P0[:BX0<c4b0P1;BDX0100o
?Ch09bDV02LU9P0b<34202l];P0404I4A@1=Bd`0CD]<03Le=P80<S0a00@0=S@e03<a<P0c<C80FeUJ
0P0h=SL0100k>CX0;bd^02l];P0a;c0202XX:@0405YHF@0X9RL0:2HW03Pf=`80;2X[00@0<Bl`038`
<@0b<340Cde>0P0Q7b00100W9BH0;2X[02`Z:`0d<S<205=ADP04040n?`0X9RL0:2HW028P8@809bDV
00@0G5YK03Tg>00i=cP0?C/l0P0h=SL01010?Sl0?3Xk03`j>`1LFU/203lm?P0402l];P0d<S<0=38c
03Xh>@80<c4b00@0CD]<04M5AP17ADH0;R`]0P0Y9bP0100`;Rl0;2X[02`Z:`1?CDh2034_<00402l]
;P0X9RL0:2HW02@R8`80=S@e00@0@T4o02HU8`0V9B<0:bXX0P0U9280100Y:2H0;Rd[02h]:`0e=382
034`;P0403Pg=@0/:bT0;2/Y030_;@80?3/i00@0>3Le02l^;00_;R`0:RTW0P0V9B<0100k>SP0>3Le
03Pg=@0[:RP2030_;@0403`k>@11@3h0@D0n040o?@80<2l]00@092<Q02PW9@0X9bD09B@R0P0W9R@0
100`;bd0;bh/02l^;00/:bT203<b<00402TX9P0o?S`0?chl04i=B`80A4=100@0<c8`02l^;00_;R`0
=3<a0P0f=C<0100Y:2H0;Rd[02h]:`11@3h202TX9P040491?`0Z:BL0:RTW038a;`80:RTW00@0<C0^
02XY9`0Z:BL09B@R0P0];2X01011@3h0BdY804]:B00o?S`202`[:@05034`;P0^;B/0;Rd[02TX9P0U
92800@3oool000`0AeEP0`18EV4404MEH0D0AUAO3015Deh?04IDG`<0AeEP1018EV4>04UGHP0404UK
I`16F6@0AUQT049DH080@eEQ00@0AeUU04YLJ01:G6P0BE]W0P17FFD01017DE/0>TA=03Y4C@0Z=3d2
020Y;P0301PP8`0@61X041PJ00<03aDE00@020d@00P<3@0830d020`=0P0=3`h0100D5QD07QhL01hN
700X:2H202l^;00402h`;@0`<2h0<30^030`;P80;bhZ00@0;B`W02h[9P0^:bH0;R/T0P0_;2D0100X
:2H09RHT02HV900Y:BL202l_;@@0=3@b0P0f=S@0100Z:RX0;bl_02l_;`0V9RH2024Q8@<0:RXZ00@0
:BTY03@d=00d=3@0<S8d0P0c<cD0100a<C<0;bla02l_<@0c<cD203`l?P040492A011@D<0@D5302h^
<080;bla00@0;Bd_02PX:P0X:2X08R8T0P0Q8B<0100T92H0:2PZ02PX:P0V9RH202@T900401@D500J
6QX06QXJ02/[:`809RHV00@0:2PX03/k>`0k>c/01`L70P030`<010071`L0<S8b038b<P0X:2P202@T
900401dM7@0D51@051@D014C4P804Q8D00@04A0E01TF7@0I5Qd081hS0P0G4QH0100;1`@03@X500d:
1@0J2@4202LF300403/W701>>2X0CSPZ05]1<080HDD`00@0Hd@_06A4;@1TA2d0IdPc0P1ZBcH0101_
D3/0M5E007AE@01hFD@307]LA`80O5e800@0OEi907iOBP1nGdX0Of1;2020HD`01021HTd0P65<081Q
C020HD`2089SCP0408EVD@26Ie80QVMB08EVD@80PfA?00@0Pf1=08=PC023H4`0Q65=0`24HD/308=Q
B080PV1500@0SF5608YQA@2:HDD0Qem50P24GdD01021GDL0OU]707iKA`1jFDH207UJAP0407eWG01c
Ge@0LemD048_:080;AhG00@0<R@Q03`b<@0l<S40@STj0P1FD580100S7Al0?CLi03dg>@0`;2d202`X
:@0404I4A@17AdL0AdM704=3@`809RHV00@0;2X[03@b<`0d<S<09bDV0P0j>3T01013@D80:BLX02TW
:00Z:2T203@b<`0403lm@00]:bh0;B/^02d[;P807A/N00@0<c4d03/i?00k>C`0:2HY0P0f=3L00`0Y
9bX09B<V02DS9P0304U7BP0402`Z;@0e<cH0=C<f02d[;P80=38e00@0>3Hg04U7B019AdP0E59C0P0_
;Bh0100X9RL0=38c03@b<`0R824202h/;@04051>C`0h=SL0>3Hg02d[;080=38c00@0<Bl`05UGF01I
EeP0<c4b0`0Z:2T202TW:00404m=CP0f=3D0=S@e03@b<`80;B//00@0<c4b030^;`0`;Rl0EE=D0P0/
:R/0100c<C80;bd^02l];P0c<C82034_<00404E3A00/:R/0;2X[01lM7P<0:2HW0`15@d@202h/;@04
01lM7P0S8B808b4R05aJF`80?3Xk00@0@d5203Pf=`0h=SL0@T110P0b<340101?CDh0?C/l03dk?00Y
9bP203Dc=00403@b<`0c<C80<c4b03lm?P80CD]<00@0;B//020N7`0P7Ql0>3Hg0P0]:b`0101;BDX0
@d5204=1@P0^;2d203<a<P0402/Y:P0]:b`0;B//03@c<@808B0N00@0;Rd[02d/:P0];2X0:BPV0P0V
9B<0100[:RP08b8P02<R800c<S0203Dd<P0404E4@P1:BDL0BTU703Pg=@80:RTW00@0;2/Y038a;`0b
<Bl0:bXX0P0d<c40100h=cD0@T4o0491?`17AT@204=2@00403Xi=`0Z:BL0:RTW038a;`80:2LU00<0
7A`J020O7@0P7ad0100Z:BL203<b<00403/j>01;BTP0BdY803hm>`80;B`Z00@0=cHd03He<`0f=C<0
;2/Y0P0b<Bl0100m?3X0CDa:04e<BP0d<c4203He<`0401PG5@0_;R`0;bh/02DT8P80=C@b00@0:2LU
02XY9`0Z:BL0;B`Z0P0a<2h0100h=cD0?C`j03dl>P0k>SP203Dd<P0502`[:@0_;R`0;bh/03<b<00f
=C<00@3oool000`0AeEP0`18EV4404MEH0D0AUAO3015Deh?04IDG`<0AeEP1018EV4>04UGHP0404UK
I`16F6@0AUQT049DH080@eEQ00@0AeUU04YLJ01:G6P0BE]W0P17FFD01017DE/0>TA=03Y4C@0Z=3d2
020Y;P0301PP8`0@61X041PJ00<03aDE00@020d@00P<3@0830d020`=0P0=3`h0100D5QD07QhL01hN
700X:2H202l^;00402h`;@0`<2h0<30^030`;P80;bhZ00@0;B`W02h[9P0^:bH0;R/T0P0_;2D0100X
:2H09RHT02HV900Y:BL202l_;@@0=3@b0P0f=S@0100Z:RX0;bl_02l_;`0V9RH2024Q8@<0:RXZ00@0
:BTY03@d=00d=3@0<S8d0P0c<cD0100a<C<0;bla02l_<@0c<cD203`l?P040492A011@D<0@D5302h^
<080;bla00@0;Bd_02PX:P0X:2X08R8T0P0Q8B<0100T92H0:2PZ02PX:P0V9RH202@T900401@D500J
6QX06QXJ02/[:`809RHV00@0:2PX03/k>`0k>c/01`L70P030`<010071`L0<S8b038b<P0X:2P202@T
900401dM7@0D51@051@D014C4P804Q8D00@04A0E01TF7@0I5Qd081hS0P0G4QH0100;1`@03@X500d:
1@0J2@4202LF300403/W701>>2X0CSPZ05]1<080HDD`00@0Hd@_06A4;@1TA2d0IdPc0P1ZBcH0101_
D3/0M5E007AE@01hFD@307]LA`80O5e800@0OEi907iOBP1nGdX0Of1;2020HD`01021HTd0P65<081Q
C020HD`2089SCP0408EVD@26Ie80QVMB08EVD@80PfA?00@0Pf1=08=PC023H4`0Q65=0`24HD/308=Q
B080PV1500@0SF5608YQA@2:HDD0Qem50P24GdD01021GDL0OU]707iKA`1jFDH207UJAP0407eWG01c
Ge@0LemD048_:080;AhG00@0<R@Q03`b<@0l<S40@STj0P1FD580100S7Al0?CLi03dg>@0`;2d202`X
:@0404I4A@17AdL0AdM704=3@`809RHV00@0;2X[03@b<`0d<S<09bDV0P0j>3T01013@D80:BLX02TW
:00Z:2T203@b<`0403lm@00]:bh0;B/^02d[;P807A/N00@0<c4d03/i?00k>C`0:2HY0P0f=3L00`0Y
9bX09B<V02DS9P0304U7BP0402`Z;@0e<cH0=C<f02d[;P80=38e00@0>3Hg04U7B019AdP0E59C0P0_
;Bh0100X9RL0=38c03@b<`0R824202h/;@04051>C`0h=SL0>3Hg02d[;080=38c00@0<Bl`05UGF01I
EeP0<c4b0`0Z:2T202TW:00404m=CP0f=3D0=S@e03@b<`80;B//00@0<c4b030^;`0`;Rl0EE=D0P0/
:R/0100c<C80;bd^02l];P0c<C82034_<00404E3A00/:R/0;2X[01lM7P<0:2HW0`15@d@202h/;@04
01lM7P0S8B808b4R05aJF`80?3Xk00@0@d5203Pf=`0h=SL0@T110P0b<340101?CDh0?C/l03dk?00Y
9bP203Dc=00403@b<`0c<C80<c4b03lm?P80CD]<00@0;B//020N7`0P7Ql0>3Hg0P0]:b`0101;BDX0
@d5204=1@P0^;2d203<a<P0402/Y:P0]:b`0;B//03@c<@808B0N00@0;Rd[02d/:P0];2X0:BPV0P0V
9B<0100[:RP08b8P02<R800c<S0203Dd<P0404E4@P1:BDL0BTU703Pg=@80:RTW00@0;2/Y038a;`0b
<Bl0:bXX0P0d<c40100h=cD0@T4o0491?`17AT@204=2@00403Xi=`0Z:BL0:RTW038a;`80:2LU00<0
7A`J020O7@0P7ad0100Z:BL203<b<00403/j>01;BTP0BdY803hm>`80;B`Z00@0=cHd03He<`0f=C<0
;2/Y0P0b<Bl0100m?3X0CDa:04e<BP0d<c4203He<`0401PG5@0_;R`0;bh/02DT8P80=C@b00@0:2LU
02XY9`0Z:BL0;B`Z0P0a<2h0100h=cD0?C`j03dl>P0k>SP203Dd<P0502`[:@0_;R`0;bh/03<b<00f
=C<00@3oool000l0B5IQ1017EF0504IDG``0A59M4017EF0504QFH@l0BEMR00<0BUaX04MII@17FFD0
0`13EF401015Ef<0AUQT04IHI016F6@204EGH`0404aIHP16De`0AU=L041:D`80>D9900@0<CXo02L^
=00W;S@072@W0P0F7R40100H82<051TM01@I7@0<4AD200L;3P0400H:3@0<3@l030d?01<E508061XI
00@09BLV02T[:P0Y:bX0;S0]0P0c<c40100d<bl0<C0[034`:`0_;2L202`Y900402DU8`0S8b408b<Q
02<S8@809bLU00@0;2`Z02h^;00^;R`0;Rh/0P0];B/0100_;bl0<c<c03<c<`0^;Rh202/[:`04038b
<P0];Bd0;Bd]02<S8`809BDU00@09BDW02LW:@0W9bT09RHX0`0Q8B<202LW:@<0:RX/00@0=3@f030`
<P0`<380;Rh`0P0];Bl0100W9bT0820R020P8P0Q8B<202LW:@0402PX:00Q8B408B4Q01TI6@805ADE
00@07QhN028R8P0R8R808b<S0P0d=3@010092@T00`<300<30`051@D202d];@0402/[:`0T92@092@T
01hN7P8051@D00@05QHF018B4P0B4Q803P`A0P0E41H0100L5QX050/<01@;300<1@0201<:0@04038N
4`0l:A/0?2TK04`f:080F3h_00@0GD8_0652;P1Q@Rh0ITHa0P1YBC80101/CCP0KTlj06i?>P1bDch2
07IG@P0407YKAP1kG4L0Nea707aMB080Nea700<0QFIA08IWDP26Ie800`26Ie801025IU40PfA?08=T
C`20HD`207iOBP<0Of1;00@0P65<08=TC`23I4l0QVMB0P27J5<01025IU40PfA?08=TC`21GTX2089O
B`0408AQB`25HT`0QF9<08ESBP80Q69900@0Pf56089PA@22H4D0QUa30P25GD<01026GTD0Q5i708AN
A`23GdT2089OB@<0PUm;00@0NV1A06mEB01_EDP0KeU<0P1A?C@0100Y6140<28O030R7`0k<2h206EK
FP0402lY:`0e;c40=Bla02LS9080?SXk00@0=S@e03De=@0e=CD0EUIF0P0Q8B40100a;c00=38c03@b
<`0O7Ah2030^;`0404a:B`0j>3T0>SPi02LU9P809B<T00@0;2X]0490@`12@4<08R0S0P0Y9bX0100f
=3L0;B/^02d[;P10?T4202/Y;00402HT9`0m>ch0?C/n02LU:080C4Y=0`0f=3L00`0k>C`0BDM:04U7
BP0303@b<`0403hl?@0k>CX0>cTj03Xh>@80<2h_00@07Q`M02l];P0_;Bh0ATA50P17ADH0100Y9bP0
<2h_030^;`0O7Ah205iLG@0402/Y:P0T8R<0928S02`Z:`80:BLX00@0BTQ903/i>P0k>CX0;bd^0P0X
9RL0100_;Bh0;R`]02h/;@1@CTl203@b<`0402l];P0b<340<S0a03Le=P80=S@e00@0@T1104=1@P13
@D80928S0P0_;Bh0100T8R<0>CLh03Tg>01=Bd`203Tg>00402LU9P0]:b`0;B//051>C`80BDM800@0
>3Hg030^;`0`;Rl0;2X[0P0[:BX0100h=SL0Cde>04m=CP0/:R/203Pf=`04034_<00c<C80<c4b03@b
<`80E59C00@0=C<d02DS900U8b@0<2h_0P0T8R<0100n?3d0BTQ904Y8B@0/:R/203Dc=00402HT9@0G
5AH05aDF03<b<0808b8P00@0;B`Z02XY9`0Z:BL092<Q0P0N7A/0100Y:2H09B@R02DT8P0Y:2H203@c
<@0404]:B01AD4h0DE1>03`k>@80:bXX00@0;2/Y030_;@0`;bd07ahL0P0d<c401019B4H0D4m=051?
C@17AT@203Th=P0403<b<00`;bd0<2l]03He<`80;2/Y0`0Q81h0100X9bD0;2/Y02`[:@0`;bd203Lf
=00404]:B015A480ADA202h]:`80:BPV00@0>3Le030_;@0`;bd0:BPV0P0k>SP0101DDe40ADA204E4
@P0i>3H202<R800402/Z:00T8b4092<Q02XY9`809bHT00@0<S4_03/j>00k>SP0?Sdk0P0d<c40100V
9B<08R4O028Q7`0Z:BL303@c<@80?3/i0@0n?C/103/j>040oooo000?04QFH@@0AeEP1@16E5l<04AB
G@l0AeEP1018EV4A04UGHP0304YLIP18FV@0B5YT00<0AUQR00@0AEMQ04AFH014EV00@UAN0P10DU`3
04]HH@0404YGH019De`0BE=L04E?F080?TM@00@0=Ca402`c>`0/<c/09Bhe0P0O:2l0100G7RH03QDK
00hE6`0:3a<200T<4@0400/?4P0<41<0310C010D4`8051PG00@07AlN02DW900U9b@0:b/Y0P0^;BT0
100];2P0:bXU02/Z9@0Y:BL202PX9P0402HV900W9bD09bLU02XZ:080;2`Z00@0:RXX02HV900V9R@0
;2`/0P0_;bl0100^;Rh0<30`030`<00f=SH203<c<`0302XZ:P0W9bL09bLW00<0:RX/00@09RHX020P
8P0P8280820R0P0Y:B/0100`<380<S8d038b=00V9RP202DU9`0402`/;P0e=CL0=CDg038b=0809bLY
00@08B4S02@T9P0T92H09BDU0`0L71`2014A4@0400X:2P0I6AT06ATI02<S8`808R8R00@02@T90082
0P020P800P820P0V9RH0100[:b/08R8R028R8P0L71`201<C4`0401@D4P0G5aL05aLG01<A508040X>
00@03`H900/1000;0@004`T00P0Y7A401014;Ad0CcDV04le9P1H?Rd205e2;`0405i2;@1Q@Rd0HD8]
06I6;`80Jd/d00@0L58j079D?01bE3`0MELo0P1hFT8307YLA0<0NE]300@0PF=;089TC022I4`0PfE=
0P24ITh01023IDd0PVA<089TC020HTX507iPB004081RBP23IDd0PfE=08IXD080R6YB00@0QVQ@08=U
C@23IDd0RFI@0P2:Ie40102<JE<0SF]B08e[DP2=Je8208aZC`0408]YC@29Id/0RFM;08INA@80Q5m5
0`24GTL00`23GdL0Pem908=OB@0308APBP0407IGA@26JET0QVUI07]OD@80MemC00@0AC4V02DB2`0U
4P/0=BHO0P12<bh0101BC4h0:B<U02TS9@0g<c@203Xf=`0402LU9P0f=SH0=SHf03dm?@80ADE500@0
9R@U02XX:@0Z:2T0;B//0P0b<340101=Bd`0BdU:04]9BP0V92D201lM7P0401lM8015@dH0AD=602XX
:`<0;B/^0P0Z:2/01016A4L0;bd`02l]<00W9BP203Pf>@0402XX:`0Y9bX0:BLZ061NH@80=S@g00@0
?S`o04=1A013@D@07adN0P0X9RL0100_;Bh0BDM804U7B014@T<202<Q8P0402/Y:P0o?Ch0?cdn02LU
9P80C4Y;0`0[:BX0100O7Ah0GE]L05eKG00c<C8202h/;@04034_<00]:b`0;B//04M5AP80A49300@0
;bd^02TW:00Y9bP0;R`]0P0/:R/0101?CDh0A49304A2@`0`;Rl203Dc=00403Pf=`0j>3T0>SPi03dk
?080EUAE00@08b4R03<a<P0c<C809R@U0P0e<c@0101FE5D0@Cm0044o@00]:b`2030^;`04044o@01K
FEX0FeUJ03<a<P80=S@e00@08AlP03Dc=00e<c@0;2X[0P1YIfP0100Z:2T0<S0a038`<@0a;c0203Tg
>00403Le=P1KFEX0FeUJ04=1@P80<S0a00@0;B//02LU9P0W9BH0?cdn0P1?CDh0100`;Rl0=cDf03Le
=P0b<34202@R8`0403He<`0Y:2H0:BPV02LV90<08B0N0P0N7A/0100a<2h0=C@b03Dd<P0[:RP202HU
8`04030_;@0m?3X0?C`j03`k>@80=SDc00@0;B`Z024P7P0Q81h0<2l]0P0g=S@01011@3h0@T4o0491
?`0m?3X203Lf=00403Dd<P0h=cD0>3Le02HU8`80:RTW00@0;B`Z02h]:`0^;B/0<C0^0P0h=cD0100l
>cT0>CPf03Th=P0c<S0203@c<@0403Lf=00i>3H0>CPf034`;P80:bXX00@0=C@b04M6A017AT@0B4M5
0P0g=S@0100e=380;B`Z02d/:P0f=C<202TX9P0402/Z:00U92809B@R03hm>`80>STg00@0=3<a02h]
:`0^;B/0:2LU0P0V9B<0100W9R@0:BPV02TX9P0g=S@204=2@040A4=10@0i>3H10?ooo`003`18EV44
04MEH0D0AUAO3014DUd?04MEH0@0B5IQ4@19Ef800`1:G6H0B5YT04QJI00304IHHP0404EGH@14EV00
A5IP049DGP80@59L0`1;F640101:Ef00BE=L04UCG015CeP203i7D00403DlA00/<c/0;3<k02D^=@80
7bP_00@05ahV00hE6`0>5A/02PlC0P093140100;3a80310C00`@4`0@51<201@H5`0401dO7P0U9b@0
9BLT02/[:@80;RdY00@0;B`X02/Z9@0[:RD0:BTW0P0X:2H0100V9R@09bLU02LW9@0Z:RP202`/:P04
02XZ:00V9R@09RHT02`/;080;bl_00@0;Rh^030`<00`<300=SHf0P0c<c<00`0Z:RX09bLW02LW9`03
02XZ;00402HV:00P8280820R020P8P80:BT[00@0<30b038b=00b<S@09RHX0P0U9BL0100/;2h0=CDg
03De=`0b<S@202LW:@04024Q8`0T92H092@V02DU9@<071`L0P0A4A40100:2PX06ATI01TI6@0S8b<2
028R8P0400T92@020P800P8200820P809RHV00@0:b/[028R8P0R8R8071`L0P0C4a<0100D51805aLG
01LG5`0C4A@2010:3P0400l62@0;0@002`4001<90080:AdA00@0A2dM04le9P1?=BH0F3h]0P1M@Rl0
101N@Rd0HD8]0652;@1VARl206];=004071B>P1bE3`0LU@l07EG?`80N5Y20`1jG4@307UK@`04085S
B`22I4`0PVA<08=UC@80Q6I>00@0PfE=089TC022I4`0P69:1@1nH4P01020HTX0PfE=08=UC@26J502
08QZDP0408IXD023IDd0PfE=08UVD080RVMA00@0S6UC08e[DP2=Je80SF]B0P2<JTl0102;JDd0RFM;
08UWB`26GTD208AOA@<0Q5i700<0Pem708=OB@23GdT00`24H4X0101fEdD0QVUI08IYF@1kGe4207MO
D`0404Da9P0U4P/09A8;03DV7`80@S<^00@0DTa>02TS9@0Y8bD0=c<d0P0j=SL0100W9BH0=SHf03Hf
=P0m?Cd204E5A@0402HT9@0Z:2T0:RPY02d[;080<S0a00@0CD]<04]9BP1;BDX09R@U0P0O7Ah0100O
7B00AD=604E3AP0Z:2/302d[;P80:RP[00@0ATA702l]<00_;C009bDX0P0h=ST0100Z:2/0:BLZ02TW
:P1PGV4203Hd=`0403hl?`13@D@0@d5401lM7P80:2HW00@0;bd^04U7B019AdP0A4930P0S8B80100[
:BX0?cdn03lm?P0W9BH204a:B`<0:bTZ00@07adN05eKG01MFe`0<c4b0P0^;2d0100a;c00;B//02d[
;017ADH204A2@`0402l];P0Y9bP0:BLX02h/;@80;2X[00@0Cde>04A2@`14@T<0<2h_0P0e<c@0100h
=SL0>SPi03Xh>@0m>c`205IDE@0402<Q8P0c<C80<c4b02HT9@80=C<d00@0EUAE044o@011?d00;B//
0P0`;Rl01011?d00FeUJ05]IFP0c<C8203Hd=@04024O800e<c@0=C<d02`Z:`80JFMX00@0:RPY038`
<@0b<340<Bl`0P0i=cP0100g=CH0FeUJ05]IFP13@D82038`<@0402d[;00W9BH09bDV03lm?P80Cde>
00@0<2h_03Le=P0g=CH0<S0a0P0T8R<0100f=C<0:BPV02TX9P0W9R@3024P7P807QdK00@0<C0^03Dd
<P0e=380:bXX0P0V9B<0100`;bd0?C`j03dl>P0l>cT203He<`0402d/:P0Q81h08B0N030_;@80=cHd
00@0@D0n0491?`12@Cl0?C`j0P0g=S@0100e=380>3Le03Pg=@0V9B<202XY9`0402d/:P0^;B/0;Rd[
034`;P80>3Le00@0?3/i03Th=P0i>3H0<c8`0P0d<c40100g=S@0>CPf03Th=P0a<2h202/Z:00403Dd
<P17AT@0AdI404Q7A@80=cHd00@0=C@b02d/:P0];2X0=SDc0P0Y:2H0100[:RP09B@R02DT8P0n?C/2
03Xi=`0403@c<@0^;B/0;Rd[02PW9@809RDS00@09bHT02TX9P0Y:2H0=cHd0P13@T0104A3@@40>CPf
0@3oool000`0BEMR0`18EV4404MEH0D0AUAO3015Deh=04MEH0D0B5IQ0`19Ef8304YHH``0BEMR00<0
AeYQ04QKHP18Ff801019G6<204QKHP0404EHG`13EUd0@eIM04YHH@80BeUR00@0BeQQ04QEGP18EEh0
BeEO0P1=EF001019DE`0@d]F04=;EP0g@D/203<mA`0402df?`0T;CH092df01/R:P804aXR00@03a@J
00`A5`0<4AL01`/<0P062P/010072`X03@l>00d?3P0E5a@201lO7@0402HW8P0[:RH0:bXV02/[:@80
;Bd[00@0;Rh/02d]:`0];B/0;bl]0P0a<Bl0100_;bd0:RXX02XZ:00X:2P202LW9`0402HV9P0Y:BT0
:BTY030`<0<0=CDe0P0f=SH0100g=cT0>3Pj03Ph>P0d=3H202`/;P0402TY:`0Z:R`0:RX/02PX:P80
8b<U00@07QhP01dM7`0M7Al08b<U0P0/;2h0100^;S00:BT[02TY:`0Y:B/202d];`0402/[:`0X:2P0
:2PX02HV9P806a/K00@030`<01TI6@0I6AT0;2`/0P0P8200100:2PX00@4100410@00000201lO7`04
02`/;00Q8B408B4Q01/K6`804A4A00@0410>01/K6`0K6a/06QHG0P0@1`X0100:000040<00103000W
5P`2048b8`0404/`7@1E>2H0ECPV05m3;P80ITLc00@0ITLb06M7<P1WAc80JTXc0P1^CSL0101dESh0
MUQ007IH@01hFT8507]MA@0407UK@`1hFT80N5Y207MI@@80N5Y200@0NUa407aNAP1lGTH0OEm70P1n
H4P407mQB@80OV1800@0Of59089TC022I4`0QFM?0`27JE4208EWC`0408YWD@2;J580RfQB08]YD080
S6YA00@0RfU>08]YC@2;JDd0RVQ<0P29Id/308YTC@0408QRB`26H4T0QV1908ANA`80Pee600@0PE]4
081J@`20FT<0O5U50P26IU40101nGdd0NUm>07YOCP1_EDP2058l<@0402LC2P0[6140:aPA05QBE080
C4I800@0;b//038^;`0b;Rl0;bd^0P0W9bL00`0c<c<0FUYJ05YJFP0302TW:00403Le=P0c<C80<c4b
03/i>P80AdE600@0<S0a02l];P0_;Bh0928U0P0d<SD0100l>Sd0:bT/02/Y;00O7B02038`<`0403<a
=00n?3l0?S`o02@R9@80:BLZ00@0<2ha02XX:`0Z:2/0EeEH0P1DDUD0100W9BP0<S0c038`<`0W9BH2
038`<@0402/Y:P11?d00@Cm004E3A080<2h_00@0=S@e02/Y:P0[:BX07adN0P1>C4d0100f=3D0<Bl`
034_<00[:BX204e;C0040490@@0c<C80<c4b02XX:@80928S00<0?3Xk04]9BP1;BDX00`0d<S<0100i
=cP0=cDf03Le=P1<BT/205ECE00403<a<P0f=3D0=S@e03Hd=@80>cTj00@0<Bl`05YHF@1JF5T06ALH
0P0`;Rl0100Z:2T0=S@e03Hd=@1JF5T20490@@0402/Y:P0X9RL0:2HW02h/;@80EeEF00@0>3Hg03dk
?00m>c`09B<T0P0m>c`0100Y9bP0I69S06ARH`0i=cP2038`<@0403Xh>@10?Sl0@3ho03`j>`80Cde>
00@0@Cm0038`<@0b<340<c4b0P0f=3D01017ADH0BdU:04]9BP0g=CH203Dc=00404A2@`13@D80@d52
02DT8P80:BPV00@0;2/Y034`;P0a<2h0?Sdk0P0];2X0100[:RP0:2LU02PW9@0/:bT202DT8P0402XY
9`0c<S00<c8`03Dd<P80>STg00@0>cXh03@c<@0d<c40?chl0P0a<2h0100^;B/0<c8`03<b<00n?C/2
04=2@00403Pg=@0c<S00<c8`03`k>@80?C`j00@0?3/i03<b<00c<S00:bXX0P0`;bd0100c<S00:bXX
02/Z:00U928202TX9P<0=3<a00@0;Rd[03dl>P0m?3X0BdY80P14@d40100c<S00=cHd03Lf=00Y:2H2
03ln?004028Q7`0h=cD0>3Le034`;P80?chl00@0;bh/02`[:@0/:bT0;2/Y0P0_;R`0100`;bd0;Rd[
02h]:`0_;R`2034`;P0503Dd<P0o?S`0?chl03dl>P0a<2h00@3oool000`0BEMR0`18EV4404MEH140
AUAO4018EV4304UGHPD0BUQS3019Ef801015F5l0AUUP04IIH017FV4504QKHP0404MJH@16FF00AUUP
04QFG`80BEMP00@0AeEP04EAG@15DEd0B5AP0P1@FFP0101AF6P0CEAT04eDI018E60204IBGP0404A>
FP0lAU00?4I@03<lA@809c0i00@07B@/01LN9P0G7RH041HF0P0;4A4010052@X00`L600<71P092`X2
014A3`0401PH5P0K71L06a`G01dM6`809BDS0`0[:bT0100^;R`0<S8`038b<00c<c42030`;P0403<c
<`0^;Rh0;Rh^02PX:0<09BDU0P0X:2P0100/;2`0;Bd]02d];@0_;c42034a<`03038b=00`<380<30b
00<0<C4c00@0:RX/020P8P0P82808b<U0P0O7b40100L71h07AdO01dM7`0P828202@T9P0402/[;@0a
<C<0<C4c034a<@80>CTi00@0:BTY02@T900T92@06QXJ0P0I6AT0100];Bd09RHV02HV9P0<30`20082
0P040000000L71`071`L02l_;`808R8R00@06a/K018B4P0B4Q804Q8@0P0J61T0100D414030<400`3
100I2`X202hM5P0403h/7P1<>BP0C3TX05Df8@80Fc`W00@0Hd@_06Q8<`1XB3<0JTXc0P1/C3D0101_
CcP0LU8k079B>`1gFCl207UK@@0307]M@`1nH4H0OV1600<0Of570`1mGdD6085SB@03089TBP24IT`0
Q6I<00<0QVQ>00@0P69807iPAP1nH4H0OV160P20HTP01024IT`0QfU?08MYC`28JU0208MYC`0408UW
CP28ITd0R6I=08MUC080QVA900<0QF=808ESA`25HdL00`25I4D0102;Ie40RVI@08YVD029IDl208IR
BP0408EOB022G4D0PUa5085L@P80P5]100@0Q5i707]G@@1kEd40OEY60P1lGD/0101nHe80N5iA07QN
D@0k9AP2038L4@0403`f>01RG5h0HUaN02PT9@80<2`]00@0=C<d024Q8@0Q8B40=SHf0P11@D401015
@d@0=C<d03Dc=00`;Rl2038`<@0402/Y:P0e<c@0=C<d0490@@80=C<d00@0:RP[02/Y;00[:B`0>cTl
0P0g=CP0100T8RD0;2X]02`Z;@0U8bH203lm@00402l]<00[:B`0:bT/02h/;`80?ce000@0;R`_05UG
FP1IEeX0<Blb0P0X9RT303@b<`0402/Y:P0`;Rl0<2h_0490@@80CD]<00@0?C/l01`J6`0L6Q/0:2HW
0P13@D80100k>CX0>SPi03Xh>@0h=SL203<a<P0405=ADP0`;Rl0<2h_02`Z:`80928S00@0=cDf055?
D01ACe00=cDf0P0h=SL0100i=cP0=38c03@b<`0n?3d205YHF@04034_<00d<S<0=38c03<a<P80>SPi
00@09B<T05MEEP1GEEH05A<D0P0^;2d0100]:b`0=38c03@b<`1FE5D203dk?00402XX:@0T8R<0928S
02XX:@80@3ho00@0ATA503Tg>00i=cP0<2h_0P0h=SL0100c<C80B4I704Q6A`1CDE8203Hd=@0303lm
?P0j>3T0>SPi00<0=C<d00@0>3Hg034_<00a;c00=C<d0P0h=SL0100l>S/0<2h_030^;`0f=3D202HT
9@0403Xh>@0i=cP0>CLh02XY9`80=cHd00@0<S4_03Pg=@0h=cD0CTe;0P0a<2h0100M71X05QDC01HE
4`0X9bD202XY9`0403<b<00h=cD0>3Le03He<`80?Sdk00@0C4]9051?C@1@Cdd0>STg0`0Z:BL202h]
:`0403Th=P11@3h0@D0n03Th=P80?3/i00@0EeID04M6A017AT@0>cXh0P0^;B/0100R8Al08b8P02<R
800V9B<201hM6`0402/Z:00W9R@09bHT02TX9P80<2l]00@0?3/i04e<BP1=C4X0BdY80P0g=S@0100b
<Bl0:RTW02XY9`0m?3X202@S8@0403/j>00f=C<0=SDc04i=B`80>STg00@0:BPV02`[:@0/:bT0<2l]
0P0b<Bl0100/:bT0:2LU02PW9@0^;B/203Th=P050491?`0k>SP0>cXh02d/:P0P7ad00@3oool000`0
BEMR0`18EV4404MEH140AUAO4018EV4304UGHPD0BUQS3019Ef801015F5l0AUUP04IIH017FV4504QK
HP0404MJH@16FF00AUUP04QFG`80BEMP00@0AeEP04EAG@15DEd0B5AP0P1@FFP0101AF6P0CEAT04eD
I018E60204IBGP0404A>FP0lAU00?4I@03<lA@809c0i00@07B@/01LN9P0G7RH041HF0P0;4A401005
2@X00`L600<71P092`X2014A3`0401PH5P0K71L06a`G01dM6`809BDS0`0[:bT0100^;R`0<S8`038b
<00c<c42030`;P0403<c<`0^;Rh0;Rh^02PX:0<09BDU0P0X:2P0100/;2`0;Bd]02d];@0_;c42034a
<`03038b=00`<380<30b00<0<C4c00@0:RX/020P8P0P82808b<U0P0O7b40100L71h07AdO01dM7`0P
828202@T9P0402/[;@0a<C<0<C4c034a<@80>CTi00@0:BTY02@T900T92@06QXJ0P0I6AT0100];Bd0
9RHV02HV9P0<30`200820P040000000L71`071`L02l_;`808R8R00@06a/K018B4P0B4Q804Q8@0P0J
61T0100D414030<400`3100I2`X202hM5P0403h/7P1<>BP0C3TX05Df8@80Fc`W00@0Hd@_06Q8<`1X
B3<0JTXc0P1/C3D0101_CcP0LU8k079B>`1gFCl207UK@@0307]M@`1nH4H0OV1600<0Of570`1mGdD6
085SB@03089TBP24IT`0Q6I<00<0QVQ>00@0P69807iPAP1nH4H0OV160P20HTP01024IT`0QfU?08MY
C`28JU0208MYC`0408UWCP28ITd0R6I=08MUC080QVA900<0QF=808ESA`25HdL00`25I4D0102;Ie40
RVI@08YVD029IDl208IRBP0408EOB022G4D0PUa5085L@P80P5]100@0Q5i707]G@@1kEd40OEY60P1l
GD/0101nHe80N5iA07QND@0k9AP2038L4@0403`f>01RG5h0HUaN02PT9@80<2`]00@0=C<d024Q8@0Q
8B40=SHf0P11@D401015@d@0=C<d03Dc=00`;Rl2038`<@0402/Y:P0e<c@0=C<d0490@@80=C<d00@0
:RP[02/Y;00[:B`0>cTl0P0g=CP0100T8RD0;2X]02`Z;@0U8bH203lm@00402l]<00[:B`0:bT/02h/
;`80?ce000@0;R`_05UGFP1IEeX0<Blb0P0X9RT303@b<`0402/Y:P0`;Rl0<2h_0490@@80CD]<00@0
?C/l01`J6`0L6Q/0:2HW0P13@D80100k>CX0>SPi03Xh>@0h=SL203<a<P0405=ADP0`;Rl0<2h_02`Z
:`80928S00@0=cDf055?D01ACe00=cDf0P0h=SL0100i=cP0=38c03@b<`0n?3d205YHF@04034_<00d
<S<0=38c03<a<P80>SPi00@09B<T05MEEP1GEEH05A<D0P0^;2d0100]:b`0=38c03@b<`1FE5D203dk
?00402XX:@0T8R<0928S02XX:@80@3ho00@0ATA503Tg>00i=cP0<2h_0P0h=SL0100c<C80B4I704Q6
A`1CDE8203Hd=@0303lm?P0j>3T0>SPi00<0=C<d00@0>3Hg034_<00a;c00=C<d0P0h=SL0100l>S/0
<2h_030^;`0f=3D202HT9@0403Xh>@0i=cP0>CLh02XY9`80=cHd00@0<S4_03Pg=@0h=cD0CTe;0P0a
<2h0100M71X05QDC01HE4`0X9bD202XY9`0403<b<00h=cD0>3Le03He<`80?Sdk00@0C4]9051?C@1@
Cdd0>STg0`0Z:BL202h]:`0403Th=P11@3h0@D0n03Th=P80?3/i00@0EeID04M6A017AT@0>cXh0P0^
;B/0100R8Al08b8P02<R800V9B<201hM6`0402/Z:00W9R@09bHT02TX9P80<2l]00@0?3/i04e<BP1=
C4X0BdY80P0g=S@0100b<Bl0:RTW02XY9`0m?3X202@S8@0403/j>00f=C<0=SDc04i=B`80>STg00@0
:BPV02`[:@0/:bT0<2l]0P0b<Bl0100/:bT0:2LU02PW9@0^;B/203Th=P050491?`0k>SP0>cXh02d/
:P0P7ad00@3oool000`0BEMR0`18EV4404MEH0D0AUAO3017EF0?04QFH@@0BEMR1@1:F6<<04UGHP@0
A5MN1@13EUd01014Eeh0AEQO04EHG`12De/204MEGP0404MEH016DUh0AU9N04QDHP80D5UX00<0CUMX
04QAHP18DF800`1<F6H0101<F6@0BeEQ04]EH@14CUP203]3CP04034j@`0/<cd0;3<m020U:0806AhQ
00@04ADF00X>3@0:3Pd02P`;0P0:30T0100:2PP02@T700T91`092@L201DE4`0401lO7@0S8b408b<Q
02LW9@80;Rh/00@0<C4_030`;P0`<2h0<S8b0P0];Bd0100Y:BT09BDU02DU9@0Q8B42028R8P<09BDU
0`0T92H0100S8bD092@V02@T9P0^;S0203Ti>`0403Xj?00e=CL0=CDg02DU9`808b<U00@07alQ01`L
7P0L71h071`N0P0Q8B<0100U9BL09RHX02HV:00Z:RX203`l?0@07alO0P0>3Ph0100O7al09BDU02DU
9@0>3Ph200@410040000000K6a/06a/K034a<@808b<S00@071`L01<C4`0C4a<06ATG0P0F51D01007
0`@0200000P0000[7A`204dl=004054n;`1:=B80BSDR0692:`80Hd</00@0IDD^06I6;`1VARl0J4Pa
0P1/C3D0101bDS/0MUHo07IF?`1hFT0207YL@P0407eOA@20HTP0P698089TBP80PfE;00@0PVA:085S
B@21HdT0Q6I<0P22I4X00`20HTP0OV1607iPAP0307eOA@<0OV1600@0PF=907mQA`1oHDL0OEm50P1o
HDL01023ID/0QfU?08MYC`28JU0208U[D@0408i/D`2=Je80SF]B08YXC@80R6I;0`27IDT01028IdP0
RFQ908UXB@28IDl208MTCP0408MSC@26HTX0QV9:08IPB@80QF1600@0QUi508EMA025GD@0Q5a30P2=
Ie00101[Ac40U7=P09AcH01oHU020696>004079KC@0a6@d0<AT=02LQ8`80A3i000@0@Cdn02dY:P0]
:BX09bDV0P0k>c/0100U9BD07alO01lO7`1MFe`2044o@00402HT9@0j>3T0>SPi02h/;@80:RPY00@0
AdE602@R8`0T8R<09bDX0P0_;C00100[:B`0BTQ;04Y8B`0e<cH201`J7@0402DS9P0c<C@0<c4d04I4
A`80;R`_00@0>cTl03Le>00g=CP0928U0P0_;C00101THVD08b4T02<Q900h=SL203<a<P0403dk?00_
;Bh0;bd^03Tg>080D4i?00@0=S@e028P8@0R8240:bTZ0P0]:b`0100e<c@0?S`m03hl?@0n?3d2028P
8@0406ARH`0c<C80<c4b02h/;@80928S00@0=C<d05ECE01EDe@0>3Hg0P0j>3T0100h=SL0<Bl`034_
<00^;2d205IDE@0402d[;00b<340<S0a034_<080>SPi00@07Q`M05ECE01EDe@06ALH0P0a;c00100_
;Bh0<2h_030^;`1?CDh203Tg>00402d[;00Y9bP0:BLX03Dc=080<Bl`00@0EUAE03Dc=00e<c@0?3Xk
0P0d<S<01015@d@0=38c03@b<`1RH64203<a<P0403Xh>@0/:R/0;2X[02/Y:P80928S00@0>CLh03dk
?00m>c`0@Cm00P0n?3d0100g=CH08b4R02<Q8P11?d0202HT9@03030^;`0U8b@09B<T00<0DU5?00@0
;B`Z01lN700O7Q`0>3Le0P0S8R00100I61H081lM020O7@0V9B<202<R800402XY9`0e=380=C@b03dl
>P80BTU700@0DU5?04e<BP1=C4X0<C0^0P0[:RP0100`;bd0:RTW02XY9`0W9R@2030_;@0403`k>@1E
E580EEAB04a;B@80=3<a00<0:RTW02d/:P0];2X00`0Z:BL0100[:RP092<Q02@S8@0U928202LV9004
034`;P15A480ADA2051?C@80A4=100@0=3<a02l^;00_;R`0;B`Z0P0c<S00100_;R`0>CPf03Th=P0n
?C/205mNG00404Y9A`0O7Q`07ahL02`[:@809RDS00@09B@R02PW9@0X9bD0:bXX0P0^;B/0100h=cD0
A4=104A3@@0l>cT2030_;@40:RTW0@0/:bT10?ooo`003019Ef8404QFH@<0AeEP1@16E5l@04MEH0<0
B5IQ7@19Ef8<04UHG@`0BeUT00@0AUAO04MEH017EF00B5IQ0P19Ef801018EV40AeEP04MEH015Deh2
04ABG@0403m0A@0c=3T0<c@i02<T:@806AXO00@05QLL01@E6P0D5AX04A8G0P0=3Q<010072@P01PP7
00H81`082PL2018B4004020O6`0Y:2<0:BPS02`Y9080:RLR00@09b`X02X_:@0Z;bT0:RlY0P0[;RL0
100_<R/0=3D]03@e;@0`<BT202TY8@0402HR8@0U92809B@R02<U8P80720O00@061`K01/L7P0K71h0
928W0P0]:2h0100n>Cd0=S@g03Hd=`0_;C0202PX:P0402@T9P0P8B<0824S01`M7`805a/L00@07B4R
02P/:`0X;2/0:Bd/0P0>4`l0100I7QX09BXT02DZ900@5@l201dR700400dC3`052PH01@X600@81`80
51HE00@0>CTk02@T9P0T92H0518G0P0G5AX0100=1`L030L400`7100D3@L202DK4P0403d_901>?Rl0
CSh_05E2<@80E48^00@0HdDY06E7:`1UAb/0J4X^0`1YBbl206]=<@0406mA=@1bE3P0LU@h07QH@`80
NEU400@0Ne]607iNB@1nGTT0P61;0P22HTd308ATC`0408EVD@23I4l0PfA?081QC080OUm:0`1lGDP0
101mGTT0OUm:07iOBP21HDP208=SBP0408EUC028J4l0R6Q?08UYD080R6Q?00<0QfM>08IVC@26ITd0
0`2<K5<0102:JU40RFU@08UYD027Idh208IVC@<0QFE<1024I4/208=SBP04089RB@20H4L0P61707iN
A@80O5a300@0PE`o07eJ?P1mFSh0QV980P1nFT@0101nFdP0KD`k06e<>`1fEDH207YHC00401dK3`0R
7aH08QlF05AABP80?S/f00@0;bXW03`f>00l=SP091hR0P0m=c/0100m?480CTeC04i=D`0^;C8202DT
:@04030`<P0];Bd0;Bd]03Ti>@80>SXj00@0;B//02/Y:P0[:BX08AlP0P15@d@0100f=3D0=38c03@b
<`0U8b@203<a<P0404]9BP0h=SL0>3Hg03Pf=`809R@U00@09B<T02l];P0_;Bh0<S0a0P1LFU/0100^
;2d0;bd^02l];P12@44302PV9`80E59C00@0=38c02TW:00Y9bP0;bd^0P0^;2d0100`;Rl0=cDf03Le
=P0e<c@2020N7`0406YXJ@0`;Rl0<2h_030^;`80:RPY00@0;R`]05UGF01IEeP0=cDf0P0]:b`0100k
>CX0:bTZ02/Y:P0`;Rl2051>C`0402`Z:`0`;Rl0<2h_03Dc=080=38c00@0=C<d04E3A015@d@0:bTZ
0P0Q7b00100c<C80<Bl`034_<00[:BX205QFE`0402TW:00]:b`0;B//03Lf=080=SDc00@0G5]I03Lf
=00g=S@0=C@b0P0h=cD0100b<Bl0:bXX02/Z:01FEE<202`[:@04034`;P0K6QP06aXH02l^;080:RTW
0`18AdD01016AD<0=3<a03@c<@0[:RP2038a;`0403<b<00[:RP0:bXX030_;@80?Sdk00@0FeYH04i=
B`1>CD/0=3<a0P0O7Q`0100P7ad0;B`Z02d/:P0^;B/202@S8@0401LF500[:RP0:bXX02h]:`80<c8`
00@0BTU704]:B01;BTP0>3Le0P0c<S001014@d40;2/Y02`[:@0[:RP203<b<00402@S8@0m?3X0?C`j
04U8AP80<S4_00@0>CPf03dl>P0m?3X0?3/i0P0_;R`0100O7Q`071/I01`K6@0Y:2H303Th=P<0=cHd
0P0i>3H0100k>SP0>CPf03Th=P0c<S0202h]:`0403Lf=00Z:BL0:RTW03@c<@80?chl00@0DE1>0450
?P11@3h0?C`j0P0i>3H0100[9BD0<R`/038/;00_:BT202LQ8@0403<];@19@d<0BD=304Y4A080>cDe
00D0:BPV02d/:P0];2X0<S4_03@c<@010?ooo`003019Ef8404QFH@<0AeEP1@16E5l@04MEH0<0B5IQ
7@19Ef8<04UHG@`0BeUT00@0AUAO04MEH017EF00B5IQ0P19Ef801018EV40AeEP04MEH015Deh204AB
G@0403m0A@0c=3T0<c@i02<T:@806AXO00@05QLL01@E6P0D5AX04A8G0P0=3Q<010072@P01PP700H8
1`082PL2018B4004020O6`0Y:2<0:BPS02`Y9080:RLR00@09b`X02X_:@0Z;bT0:RlY0P0[;RL0100_
<R/0=3D]03@e;@0`<BT202TY8@0402HR8@0U92809B@R02<U8P80720O00@061`K01/L7P0K71h0928W
0P0]:2h0100n>Cd0=S@g03Hd=`0_;C0202PX:P0402@T9P0P8B<0824S01`M7`805a/L00@07B4R02P/
:`0X;2/0:Bd/0P0>4`l0100I7QX09BXT02DZ900@5@l201dR700400dC3`052PH01@X600@81`8051HE
00@0>CTk02@T9P0T92H0518G0P0G5AX0100=1`L030L400`7100D3@L202DK4P0403d_901>?Rl0CSh_
05E2<@80E48^00@0HdDY06E7:`1UAb/0J4X^0`1YBbl206]=<@0406mA=@1bE3P0LU@h07QH@`80NEU4
00@0Ne]607iNB@1nGTT0P61;0P22HTd308ATC`0408EVD@23I4l0PfA?081QC080OUm:0`1lGDP0101m
GTT0OUm:07iOBP21HDP208=SBP0408EUC028J4l0R6Q?08UYD080R6Q?00<0QfM>08IVC@26ITd00`2<
K5<0102:JU40RFU@08UYD027Idh208IVC@<0QFE<1024I4/208=SBP04089RB@20H4L0P61707iNA@80
O5a300@0PE`o07eJ?P1mFSh0QV980P1nFT@0101nFdP0KD`k06e<>`1fEDH207YHC00401dK3`0R7aH0
8QlF05AABP80?S/f00@0;bXW03`f>00l=SP091hR0P0m=c/0100m?480CTeC04i=D`0^;C8202DT:@04
030`<P0];Bd0;Bd]03Ti>@80>SXj00@0;B//02/Y:P0[:BX08AlP0P15@d@0100f=3D0=38c03@b<`0U
8b@203<a<P0404]9BP0h=SL0>3Hg03Pf=`809R@U00@09B<T02l];P0_;Bh0<S0a0P1LFU/0100^;2d0
;bd^02l];P12@44302PV9`80E59C00@0=38c02TW:00Y9bP0;bd^0P0^;2d0100`;Rl0=cDf03Le=P0e
<c@2020N7`0406YXJ@0`;Rl0<2h_030^;`80:RPY00@0;R`]05UGF01IEeP0=cDf0P0]:b`0100k>CX0
:bTZ02/Y:P0`;Rl2051>C`0402`Z:`0`;Rl0<2h_03Dc=080=38c00@0=C<d04E3A015@d@0:bTZ0P0Q
7b00100c<C80<Bl`034_<00[:BX205QFE`0402TW:00]:b`0;B//03Lf=080=SDc00@0G5]I03Lf=00g
=S@0=C@b0P0h=cD0100b<Bl0:bXX02/Z:01FEE<202`[:@04034`;P0K6QP06aXH02l^;080:RTW0`18
AdD01016AD<0=3<a03@c<@0[:RP2038a;`0403<b<00[:RP0:bXX030_;@80?Sdk00@0FeYH04i=B`1>
CD/0=3<a0P0O7Q`0100P7ad0;B`Z02d/:P0^;B/202@S8@0401LF500[:RP0:bXX02h]:`80<c8`00@0
BTU704]:B01;BTP0>3Le0P0c<S001014@d40;2/Y02`[:@0[:RP203<b<00402@S8@0m?3X0?C`j04U8
AP80<S4_00@0>CPf03dl>P0m?3X0?3/i0P0_;R`0100O7Q`071/I01`K6@0Y:2H303Th=P<0=cHd0P0i
>3H0100k>SP0>CPf03Th=P0c<S0202h]:`0403Lf=00Z:BL0:RTW03@c<@80?chl00@0DE1>0450?P11
@3h0?C`j0P0i>3H0100[9BD0<R`/038/;00_:BT202LQ8@0403<];@19@d<0BD=304Y4A080>cDe00D0
:BPV02d/:P0];2X0<S4_03@c<@010?ooo`003`19Ef8304QFH@@0AeEP0P16E5l@04MEH0<0B5IQ7@19
Ef8<04UHG``0BeUT00<0AeEP04QFH@18EV401018EV4204MEH00404IDG`15Deh0AE=N04U=EP80@TI?
00@0=SY302L[=00W:c@06AdV0P0A5Ah0100>4Q/03a<L00lC700;3`h300d?3P803@d;00@03a0;01@C
3`0D4`l06aXE0P0O7QT0100T9R<0;2h[02`^:`0b<bh2034b;@0403@c;P0k>c<0>c/c04=0>@80A44j
00@0<BdZ02PW9@0X9bD08B<P0P0S9bH0100[;bh0<34c030a<`0]:bh202LR9P0402hY;@0Z:2/0:RP[
02TW:P80:BT[00@0:RX/02LX:P0W:2X08b@V0P0N8R<0100N7b<08b@X02<T:00[;2h201`M7`0402DW
9P0_<C00;c4`01XL6@808R@Q00@06AhJ00X?2`0:3`/010H30P0A4A40100g=CH09R@U02HT9@0C3Q82
014<400400d30@0H3`X060l:02/N5P80>R`Q00@0AcDW04ll:`1??2/0F4<`0P1MAc80101P@RX0I4H^
06A6;P1YBc<206a>=P<0KDlg00@0KU0h071B>P1`DSX0KThi0P1`D3/0101dE3l0NEU407UIA01lG4L2
07mOBP<0P61;00@0PV=>08=TC`23I4l0Q6E@0`25IU4208IWDP<0QFIA00@0Oem6085QB021HDP0Pf=:
0P25ID`00`28J4l0RFU@08UYD00308YZD@0408IVC@25ID`0QFE<08=SBPD0PV990`24I4/408EUC080
Q6A;00@0PV99081PA`20H4L0OEe40P1lG4<0101jEcT0Me@h07MD>020G48207aI@`04085NBP1fDT80
MU9207eIB`80Oei?00@0DD<j02XM5@0Z7AD0;b8L0P1UG5L01016?S`06Q@D01XD500m>CX203`g>`04
03Lf?011@4H0@D1603dl@@80;2/`00@0:BT[02/[:`0[:b/0;Rh^0P11@D40100m>c`08AlP024O800b
<34202`Z:`04044o@00Y9bP0:BLX02PV9`80;2X[00@0>cTj04E3A015@d@0;2X[0P0_;Bh0100a;c00
9R@U02HT9@0j>3T203hl?@0404Q6A`0[:BX0:bTZ044o@080=S@e00@0?cdn01/I6P0K6AX0CTa=0P0j
>3T0100[:BX0:BLX02TW:00b<34203`j>`0403<a<P0T8R<0928S05IDE@80>3Hg00@0;2X[02TW:00Y
9bP0:BLX0P1>C4d0100g=CH0;2X[02`Z:`0i=cP302l];P80CD]<00@0;bd^02h/;@0^;2d0<S0a0P0_
;Bh0100^;2d0>cTj03/i>P0c<C8202DS900403Hd=@0d<S<0=38c02h/;@80EE=D00@0:BLX02LU9P0W
9BH0>3Le0P0i>3H0101IF5H0CTe;04i=B`11@3h203Xi=`04030_;@0T8b4092<Q04I5@`80@d9000@0
7QdK03Lf=00g=S@0<c8`0P1?CT`0101>CD/0=3<a03@c<@0X9bD203<b<00403Lf=00];2X0;B`Z02LV
9080<c8`00@0BTU705YIE`1JFEL0:bXX0P0b<Bl0100a<2h0:2LU02PW9@0U928202TX9P0402DT8P0L
6aT071/I01lN7080<S4_00@0>3Le03Xi=`0j>CL0@D0n0P0k>SP0100_;R`0<C0^034`;P0n?C/2040o
?@040450?P0g=S@0=cHd05]JF080BDQ600@0>STg02d/:P0];2X0<C0^0P0_;R`0100a<2h0=cHd03Lf
=00i>3H203Dd<P0403<b<00e=380=C@b03Xi=`80<S4_00@0:BPV02HU8`0V9B<0:RTW0P0`;bd0100d
<c40=C@b03Dd<P0c<S0203Th=P0404m>C01AD4h0DE1>04I5@`80:2LU00@0;Rd[03hm>`0n?C/0Ad51
0P0Y8b<0100Q6a/0>S@d03Xd=01:A4@203dg=`04034[:`0d;Rh0=2h^040o?@80<S4_0@0Z:BL102d/
:P40oooo000<04UGHP0304YHH`19Ef80BEMR00<0BEMR0`18EV4C04MEH0<0B5IQ7@19Ef8<04YIH0`0
BeUT0`19Ef800`18EV40AeEP04MEH00304IDG`<0AeEP00@0CUQR04aFH01<EV00ADmI0P0i@dd0100[
=Cl07RPb01hX<P0E7bT2010J9004010F5P0?4a@03a<D00`@3`802P`;00@020X700/;2@0;2`T0414<
0P0D4`l0100O7Q`09RDS02HU8`0Y:2H202TU8P0402PS7`0Z9B40:RDQ02hW8@80;bPR00@0A40m03/j
=P0k>SH0<38_0P0W;2P0100T:2L09RPW02HX9`0Y9bX202/V:P0401lJ7P0P7R4081hQ02@R9@80:2PZ
00@0;2`^02`];`0/;Bl0:BX/0P0U:BX0100_;cL092@/02@T;00W:2d201`M8P0401dN8P0T9BL092DW
01HG6@8071dO00@05ALB00d?2P0=3`X02``70P0H5aD0100g<C409ahO02LN7`0A1`P200/10P0401d?
1P0_7Q@0;ahD04@b9P80DCh_00@0ED0]05M1;01G@B`0GDD]0P1RB340101VAc<0J4Te06Q9=@1ZBcL2
06]<>00406a=>@1^Cc/0KTlk07=D@080MeQ4101gEd8207QH@`0407]KAP1nGTT0OUi9089RC@80Q6A?
00@0Q6E@08=TC`23I4l0PV=>1@21HTd3089SCP@0Oem60P20H4L01021HDP0Pf=:08=SBP25ID`208IV
C@04089RB@20H4L0P61707iNA@80O5a300@0OEe407mOAP1oGdH0PV990P24I4/308IVC@0408EUC024
I4/0Q6A;089RB@80Oem600@0OEe407]K@P1kFd80MU@h0P1cDCD0101iEc`0NeU007]I@022Gd/207aI
AP04081LC022GTh0PUi>07iMEP80BRdW00@0ARdX04Xg<`1:=c<0ASPg0P16?S`0100f=3D0<S8b038b
<P0_;S@2034`=P0404M6B`0g=S/0=cHk02PX:P80:2PX00@08b<S0492@P12@T80@d520P0T8R<0100b
<3408AlP024O800j>3T202l];P@0:bTZ0P10?Sl0100`;Rl0<c4b03<a<P0b<34202HT9@0403Dc=00b
<340<S0a06]YJP80:BLX00@0>SPi034_<00a;c009R@U0P0f=3D0100]:b`0De5B05=ADP0c<C8202@R
8`0402TW:00e<c@0=C<d02h/;@80<S0a00@0@T1104]9BP1;BDX0<2h_0P0_;Bh0100X9RL0@d5204=1
@P0k>CX202TW:004038`<@0^;2d0;R`]02l];P80B4I700<0>SPi03Hd=@0f=3D00`0l>S/0100k>CX0
AD=404E3A011?d0202DS900402d[;00Y9bP0:BLX02PV9`80DU1A00@0>SPi03Pf=`0h=SL0>3Le0P0c
<S00100k>SP0CDa:04e<BP0f=C<2038a;`0403He<`0a<2h0<C0^02`[:@80=3<a00@0DE1>03Lf=00g
=S@0BdY80P18AdD0100V9B<0;2/Y02`[:@0T8b42030_;@0403<b<00^;B/0;Rd[038a;`80@T4o00@0
BTU704M6A017AT@0:2LU0P0a<2h0100b<Bl0:RTW02XY9`0T8b4202LV900402TX9P0V9B<09RDS02h]
:`80<S4_0`0c<S00100`;bd0:bXX02/Z:00[:RP3038a;`8092<Q00@0?3/i075`KP1aL6h0=3<a0`0V
9B<203/j>00402`[:@0Z:BL0:RTW030_;@80?Sdk00@0ATE303hm>`0n?C/0;Rd[0P0T8b40100Z:BL0
:bXX02/Z:00/:bT202/Z:00402PW9@0W9R@09bHT02PW9@80:RTW00@0>STg04I5@`16AD<0C4]90P0h
=cD00`0`;bd0;B`Z02d/:P0303He<`0403Ha;P0h<c00>3<`03hi=P80@Sdj00@0?3Ld034/:@0a;2T0
;BPU0P0a;2T01@0`;bd0<c8`03<b<00l>cT0AdI40040oooo000<04UGHP0304YHH`19Ef80BEMR00<0
BEMR0`18EV4C04MEH0<0B5IQ7@19Ef8<04YIH0`0BeUT0`19Ef800`18EV40AeEP04MEH00304IDG`<0
AeEP00@0CUQR04aFH01<EV00ADmI0P0i@dd0100[=Cl07RPb01hX<P0E7bT2010J9004010F5P0?4a@0
3a<D00`@3`802P`;00@020X700/;2@0;2`T0414<0P0D4`l0100O7Q`09RDS02HU8`0Y:2H202TU8P04
02PS7`0Z9B40:RDQ02hW8@80;bPR00@0A40m03/j=P0k>SH0<38_0P0W;2P0100T:2L09RPW02HX9`0Y
9bX202/V:P0401lJ7P0P7R4081hQ02@R9@80:2PZ00@0;2`^02`];`0/;Bl0:BX/0P0U:BX0100_;cL0
92@/02@T;00W:2d201`M8P0401dN8P0T9BL092DW01HG6@8071dO00@05ALB00d?2P0=3`X02``70P0H
5aD0100g<C409ahO02LN7`0A1`P200/10P0401d?1P0_7Q@0;ahD04@b9P80DCh_00@0ED0]05M1;01G
@B`0GDD]0P1RB340101VAc<0J4Te06Q9=@1ZBcL206]<>00406a=>@1^Cc/0KTlk07=D@080MeQ4101g
Ed8207QH@`0407]KAP1nGTT0OUi9089RC@80Q6A?00@0Q6E@08=TC`23I4l0PV=>1@21HTd3089SCP@0
Oem60P20H4L01021HDP0Pf=:08=SBP25ID`208IVC@04089RB@20H4L0P61707iNA@80O5a300@0OEe4
07mOAP1oGdH0PV990P24I4/308IVC@0408EUC024I4/0Q6A;089RB@80Oem600@0OEe407]K@P1kFd80
MU@h0P1cDCD0101iEc`0NeU007]I@022Gd/207aIAP04081LC022GTh0PUi>07iMEP80BRdW00@0ARdX
04Xg<`1:=c<0ASPg0P16?S`0100f=3D0<S8b038b<P0_;S@2034`=P0404M6B`0g=S/0=cHk02PX:P80
:2PX00@08b<S0492@P12@T80@d520P0T8R<0100b<3408AlP024O800j>3T202l];P@0:bTZ0P10?Sl0
100`;Rl0<c4b03<a<P0b<34202HT9@0403Dc=00b<340<S0a06]YJP80:BLX00@0>SPi034_<00a;c00
9R@U0P0f=3D0100]:b`0De5B05=ADP0c<C8202@R8`0402TW:00e<c@0=C<d02h/;@80<S0a00@0@T11
04]9BP1;BDX0<2h_0P0_;Bh0100X9RL0@d5204=1@P0k>CX202TW:004038`<@0^;2d0;R`]02l];P80
B4I700<0>SPi03Hd=@0f=3D00`0l>S/0100k>CX0AD=404E3A011?d0202DS900402d[;00Y9bP0:BLX
02PV9`80DU1A00@0>SPi03Pf=`0h=SL0>3Le0P0c<S00100k>SP0CDa:04e<BP0f=C<2038a;`0403He
<`0a<2h0<C0^02`[:@80=3<a00@0DE1>03Lf=00g=S@0BdY80P18AdD0100V9B<0;2/Y02`[:@0T8b42
030_;@0403<b<00^;B/0;Rd[038a;`80@T4o00@0BTU704M6A017AT@0:2LU0P0a<2h0100b<Bl0:RTW
02XY9`0T8b4202LV900402TX9P0V9B<09RDS02h]:`80<S4_0`0c<S00100`;bd0:bXX02/Z:00[:RP3
038a;`8092<Q00@0?3/i075`KP1aL6h0=3<a0`0V9B<203/j>00402`[:@0Z:BL0:RTW030_;@80?Sdk
00@0ATE303hm>`0n?C/0;Rd[0P0T8b40100Z:BL0:bXX02/Z:00/:bT202/Z:00402PW9@0W9R@09bHT
02PW9@80:RTW00@0>STg04I5@`16AD<0C4]90P0h=cD00`0`;bd0;B`Z02d/:P0303He<`0403Ha;P0h
<c00>3<`03hi=P80@Sdj00@0?3Ld034/:@0a;2T0;BPU0P0a;2T01@0`;bd0<c8`03<b<00l>cT0AdI4
0040oooo000<04QFH@@0BUQS0`19Ef8504QFHA00AeEP0`18EV4M04UGHP`0BUQQ3@1;FF@204YHH`03
04QFH@16E5l0AUAO00<0AUAO0`18EV400`16EV<0A5AQ04ADH@0304ADH@04049BG`0fAU<0=TIC02@d
@@805bLd00@05alQ018G6P0B5aX02a0C0P0;3a00100>4Q404Q@C018D4`0?4@h200/=2P0401PD5@0J
5QL06QHG01lI6@809AlO00@0;bLU03@Z:00d:RP0<RPV0P0^9280100Q7QT0:2LS02PW8`0a<bh203Dj
=P0403<h=00_<C00;c4`02XX:@80:2@U00@08ahR028P8`0R82<0928U0P0U9BL0100X:2X0:BX/02TZ
;00Y:R`202L[;004051>F@0h=T40>3I103Le@080<2lg00@09bH/02HU:`0V9B/06QTN0P0I61d0100<
3@P03`h:00l>2P0A3PT201TD400402PN700Q4a008A<@01P91P8070d:00@0?2HI04H`8P16<280Dc`/
0P1NA3<0101RBC<0HdTb06=9<P1TB30206E:;`0306Y:>`1/C3d0K4`m00<0KDdn00@0K4`m06e=?P1]
CCh0L5110P1bDT<00`1hF4<0MEE007EE@003075A?00407AD?`1kFdH0Ne]608ATC`80RFUD00@0RfaG
08MXD`27J5<0PV=>0P1mGTT307]LA`0407eNB@1nGdX0OUm:08QXC`80QVI=00@0Q6A;085QB021HDP0
P6170P21HDP01023HdX0Q6A;08ATB`27Idh208ATB`04081PA`1mGD@0OEe407aL@`80OUi500<0PV99
08ATB`24I4/00`26ITd01025ID`0Q6A;08ATB`22HTT207mOAP0407aL@`1kFd80Ne]207IG>P80LU<g
00@0MU@i07]I@01kFD00PEi80`1oFTL2081KB@0409IWG`27FU@0QeYD03`G4@80?b8N00@0I55=04<i
=`13>CL0:bXX0P0b<S80100Y:2h0:2L]02PW;@0o?T<2040oA00403<c=@0Y:BT0:BTY02<S8`80=CDe
0`0j>3T0100P7Ql0:bTZ02/Y:P0T8R<2044o@00403@b<`0_;Bh0;bd^02@R8`80:bTZ00@0@Cm0038`
<@0b<340;2X[0P0`;Rl0100V92D0@3ho040n?`0k>CX2051>C`04030^;`0k>CX0>cTj034_<080>SPi
00@0:BLX038`<@0b<340EeEF0P0c<C80100U8b@09bDV02LU9P0S8B8203Tg>00402PV9`1?CDh0Cde>
02HT9@80:bTZ00@08b4R03/i>P0k>CX0BDM80P0c<C8403/i>P80BTQ900@0CD]<03`j>`0l>S/0?cdn
0P0m>c`0100i=cP0?S`m03hl?@0k>CX201dK700402h/;@0d<S<0=38c03Le=P80De5B00@0@3ho034_
<00a;c00;B`Z0P0/:bT0100Y:2H0FEQF05UHEP0m?3X2034`;P0403@c<@0[:RP0:bXX02<R8080>cXh
00@0;bh/03Pg=@0h=cD0KVe[0P0d<c40100f=C<09B@R02DT8P0/:bT202/Z:004030_;@0k>SP0>cXh
0491?`80?Sdk00<0<C0^02HU8`0V9B<00`0g=S@0100`;bd09RDS02HU8`0U928202`[:@<0<c8`00@0
92<Q01dL6P0M71X07ahL0P0Z:BL0100_;R`0<2l]030_;@0d<c4203Pg=@0402<R8018AdD0B4M506AS
H@80<S4_00@0?C`j02LV900W9R@0=cHd0P0a<2h0100_;R`0=SDc03He<`0m?3X20450?P04040o?@0j
>CL0>STg02d/:P808R4O00@092<Q02XY9`0Z:BL0;Rd[0P0Z:BL0100S8R0092<Q02@S8@0^;B/203Th
=P0404A3@@0n?C/0?Sdk038a;`809B@R00@0:BPV034`;P0a<2h0=C@b0P0`;bd0100/9b@0CDQ504e8
A@1HDe0203hi=P0402LR7`0Y9240:B@Q030[:080;BPU00D0:bXX03ln?00o?S`0BTU704I5@`010?oo
o`003018EV4404YHH`<0BEMR1@18EV4@04MEH0<0B5IQ7@19Ef8<04]IHPd0BeUT0P1:F6<01018EV40
AeEP04MEH016E5l204MEH00304QFH@19Ef80BEMR00<0@59N00@0@UAP04MII@17FFD0BUaX0P14EV80
100hBUH0;Cm;02doB`0Y<SL201lY:`0401@L7`0>5QP03QHH00lD5`803a<D00@02Ph?00H:2@062PT0
4@d>0P0@30d0100D3Q007QPH01hH600/8b@203H/:`0403X`;`0l<S00?38`02h[9P808R4L00@061/D
01TN600I7QP09b`X0P0d=S<0100j>3T0>SHg03Xf=`0b;C4202d[;P0402PV:@0S8bD08b<U024Q8`80
8B8T00@092DW02DY:P0U:BX0<Blj0P0O7BP0100/:SD0=cHn03Lf?P0c<SP2038a=`0302TX;@0S8RL0
8b8W00<04Pl:00@03@H000l4000?100060T40P0S40X0100_6a@0@BXT044Z901A=bH205Dk:P0405a1
;P1QAS40HDHa06E;=080J4`d00@0K50h06iC>01^DcP0Jd/l0P1/C3d0101]CCh0Jd/l06];?01VASL2
0692<`030610<@1O?c00Gcl`00@0H40[0P1S@bh0101YBC@0L50k071@>`1hF4<207aLA`0408AUD023
I4l0PfA?089SCP80PF9=0`20HD`3085RC@0408a/D`2:JU40RVYA08IVC@80Pf=:00@0PV9908=SBP23
HdX0QFE<0P27Idh0102=KE@0RFU@08UYD024I4/2081PA`0407iNA@1oGdH0Oem6085QB080Pf=:1024
I4/208=SBP04085QB01nGTD0OUi507aL@`80Ne]200@0MELk071B=P1`DSH0Ke0d0P1jF3l0101nFT80
Oe]307mK@`1nF4<2081JA@0409IPE02IIe`0VFML06Xo=P80@B0I00@0BS8^04i0?@1>@3d0=S4^0P0U
9280100X9bd0:BP^02TX;P0[:Rl204A3B00404=3A@0];Bd0;Bd]02/[:`8092@T00@0<Bl`04Y8B@1:
B4T08AlP0P0]:b`0100S8B80?C/l03dk?014@T<202d[;00402DS900V92D09R@U03hl?@80>cTj00@0
>3Hg03<a<P0c<C808b4R0P0n?3d0100O7Ah0D4i?051>C`1GEEH202d[;00403Le=P0S8B808b4R03Tg
>080:RPY00@0FEMH03Tg>00i=cP09R@U0P0R8240100P7Ql0?3Xk03`j>`0T8R<205aJF`0402PV9`0]
:b`0;B//02HT9@80>CLh00@0EUAE03Pf=`0h=SL0?S`m0P0l>S/0100f=3D0>cTj03/i>P1?CDh2038`
<@0403@b<`0c<C80<c4b02d[;080;R`]00@0BDM8024O800Q7b00<2h_0P0e<c@0100f=3D0ATA504I4
A@12@44202d[;00403/j>00j>CL0>STg02XY9`80G5]I00@0@d90030_;@0`;bd0=C@b0P0Z:BL01016
AD<0=C@b03Dd<P0o?S`204i=B`0403Lf=01AD4h0DE1>02<R8080=cHd00<0:RTW03Dd<P0e=3800`11
@3h0100f=C<0:RTW02XY9`0X9bD202d/:P0402HU8`0/:bT0;2/Y030_;@80;Rd[00@0;2/Y02d/:P0]
;2X0;Rd[0P0/:bT0100X9bD08R4O028Q7`0U9282030_;@0303Dd<P0b<Bl0<S4_00<0<2l]00@0EeID
04Y9A`1:BDL08b8P0P15A4801011@3h0ATE304I5@`0Q81h203@c<@0403Xi=`14@d40A4=104E4@P80
>CPf00@0=3<a03Pg=@0h=cD0=cHd0P0_;R`0100P7ad09RDS02HU8`0];2X2034`;P0403<b<00e=380
=C@b03Xi=`80?Sdk00@0?3/i02d/:P0];2X0:RTW0P0g=S@0100c<S009bHT02LV900Y:2H203`k>@04
04a7@`16@Cd0AT4m040k=`80>SDa00@0<RdY02/V8P0[9R80:RDQ0P0]:2@01@0g=S@0@D0n0450?P0n
?C/0<C0^0040oooo000<04QFH@@0BUQS0`19Ef8504QFHA00AeEP0`18EV4M04UGHP`0BeUR3@1;FF@2
04YHH`0404QFH@17EF00AeEP04IDG`80AeEP00<0B5IQ04UGHP19Ef800`10DUh01012E600AeUU04MI
I@1:G6P204AFHP0403Q:EP0]?d/0;Cm;02Tb=`807bT[00@051`O00hF600>5QP03a@G0P0?4a@0100:
3Pl01PX900H:2@0A3@h2010<3@0401@>400N61P07QPH02`S9080=R`[00@0>S0_03`b<00l<S00;R/V
0P0R8A`0100H6a@06AhH01TN600W;2P203@f<`0403Xh>@0j=SL0>SHg038]<@80;B/^00@0:2HY02<S
9@0S8bD08B4S0P0Q8R@0100T9BL09BTZ02DY:P0a;cX201lM:00402`Z=@0g=Sh0=cHn03<b>080<S4g
00<0:BP]02<R9`0S8RL00`0B3`X0100=1P003`@000l4000H2@@202<@2P0402lK5011:R@0@BXT054g
9P80EC/Z00@0G44^0656<@1QAS40ID/d0P1XC3@0101/D3P0KU<h06iC>01[Bc`206a<?@0406e=?P1[
Bc`0Jd/l06I6=`80HT8c00<0H40a05lo<01O?c00101P@2/206=3;P0406U9=01`D3/0L50k07QH@`80
O5a700@0Q6E@08=TC`23I4l0PV=>0P21HTd3081QC0<0PF9=00@0S6aC08YZD@2:JU40QVI=0P23HdX0
1022HTT0Pf=:08=SBP25ID`208MWCP0408e]E029JE00RFU@08ATB`80P61700@0OUi507mOAP1oGdH0
PF580P23HdX408ATB`80Pf=:00@0PF5807iNA@1nGTD0O5a30P1kFd80101eEc/0L58f071B=P1_D3@2
07YH?`0407iJ@P1oFd<0Oe]307iH@`80P5Y500@0UV1D09UWG02IIe`0JSlf0P1181T0101:<Rh0CT0m
04i0?@0f<Bh202DT8P0402PW;@0Y:2h0:BP^02/Z;`80A4=800@0@d=502d];@0];Bd0:b/[0P0T92@0
100a;c00BTQ904Y8B@0Q7b0202d[;00402<Q8P0m>c`0?C/l04A2@`80;B//00@09B<T02HT9@0V92D0
?S`m0P0k>CX0100h=SL0<c4b03<a<P0S8B8203hl?@0401lM7P1@CTl0D4i?05MEEP80;B//00@0=cDf
02<Q8P0S8B80>CLh0P0Z:2T0101IEeP0>CLh03Tg>00V92D2028P8@04020N7`0l>S/0?3Xk02@R8`80
G5YK00@0:2HW02d[;00]:b`09R@U0P0i=cP0101FE5D0>3Hg03Pf=`0n?3d203`j>`0403Hd=@0k>CX0
>cTj04m=CP80<S0a00@0=38c03<a<P0c<C80;B//0P0^;2d01019AdP08AlP024O800`;Rl203Dc=004
03Hd=@16A4D0ATA50490@@80;B//00@0>cXh03Xi=`0j>CL0:RTW0P1LFeT01013@T00<2l]030_;@0e
=38202XY9`0404I5@`0e=380=C@b03ln?080CTe;00@0=cHd055@CP1AD4h08b8P0P0g=S@00`0Z:BL0
=C@b03Dd<P030450?P0403He<`0Z:BL0:RTW02PW9@80;B`Z00@09RDS02`[:@0/:bT0<2l]0P0^;B/0
100/:bT0;B`Z02d/:P0^;B/202`[:@0402PW9@0R8Al08R4O02DT8P80<2l]00<0=C@b038a;`0b<Bl0
0`0`;bd0101GEU@0BTU704Y9A`0S8R0204E4@P040450?P16AD<0ATE3024P7P80=3<a00@0>STg04A3
@@14@d40ADA20P0i>3H0100d<c40>3Le03Pg=@0g=S@202l^;004020O7@0V9B<09RDS02d/:P80<C0^
00@0<c8`03Dd<P0e=380>STg0P0n?C/0100l>cT0;B`Z02d/:P0Z:BL203Lf=00403<b<00W9R@09bHT
02TX9P80?3/i00@0C4M304I1?@16@Cd0@3/g0P0j=C40100b;BT0:bHR02/V8P0Z9B4202dX900503Lf
=011@3h0@D0n03hm>`0a<2h00@3oool000`0AeEP00<0BUQS04UGHP19Ef800`19Ef8304QFHA<0AeEP
0`18EV4M04UGHQT0BeUT0P1:F6<00`19Ef80B5IQ04QFH@0604QFH@03049DH013EF40@eEQ00<0A5IR
1012E60204AFHP04041;D@0h@dL0>4=702/d>@806R@V00@03QHI00L?4@073a4020d@0P093Q40100@
3Q403P`?00h<3`0<2P/200l;3004014;2`0F41005Q0@02DM6`80<2PV00@0;R/T02/[8`0[:b<0:B`U
0P0X;BH0100X;BL0:B/X02T[:00Y:2H202TU900403Xe>@0d<SD0=38e02h/;`809RHX00@08B4S01lP
8P0O82808b@V0P0T:2T0100]:S<07Q/T01hK900S82L202LT:`0401hL8@0H5QT061HI014?4P802PP;
00@04@`601D>200E3PP040L00P0J2`@0100T4@X0?2@J03`T6P19;R<205Di;P0405Ll:@1M@Rl0GD8_
0697=0<0I4Td0P1VC3D0101/DS/0LUPo079H?`1_Cd0206Y:>`040692<`1H>2T0F3PY058b8`80Dc<T
00@0F3PY05dm;P1M?Bh0G3`W0P1O?bX0101S@bh0ITHa06I6<@1XB3<206M7<P0406I6<@1UAC00IDD`
06Q9=080KDhi00@0M5E007]LA`1kG4L0OUm:0P1mGTT0101jFdH0N5U407QIA01iFD0207MG?P0407IF
?@1eEC`0MEDl07EE?080MeLn00@0NUY107aL@`1lG4<0QfM>0P25ID`00`21HDP0OUi507iNA@0307eM
A00307mOAP20H4L0P61700H0PF5800@0P61707iNA@1nGTD0O5a30P1kFd80101bEcX0Ke4e06mA=@1Z
C30207ME>P0407UE>`1oFT00OeY007mG?P80QEe400@0SU]:085B@P21DT80Zh5c0P1K>Bl0100O1P00
GD]705e;A`1?A482020H5P0402/Z<00^;C<0;Rdc024P9@80?Se200@0Bd]=03@d=00d=3@0<30`0P0O
7al0100_;Bh0A49304A2@`0j>3T202DS900403<a<P0V92D09R@U04Y8B@80:bTZ00@0:BLX03<a<P0c
<C80:RPY0P15@d@0101=Bd`0<Bl`034_<00^;2d202LU9P0403`j>`0i=cP0>CLh055?D080>3Hg00@0
7adN03Dc=00e<c@0?3Xk0P0m>c`0100f=3D0<c4b03<a<P0/:R/202DS900402HT9@0j>3T0>SPi030^
;`80IfEV0`0d<S<0100/:R/0=cDf03Le=P1IEeP2038`<@0403Hd=@0_;Bh0;bd^02h/;@80:bTZ00@0
D4i?02XX:@0Z:2T0;R`]0P0`;Rl302/Y:P0405UGF00S8B808b4R02TW:0809bDV00@0:2HW03Le=P0g
=CH0CD]<0P0l>S/01010?cd0?3/i03`k>@0T8b4204=2@00403`k>@0^;B/0;Rd[03dl>P80=C@b00@0
;B`Z03Pg=@0h=cD0@D0n0P1JFEL0100T8b40:bXX02/Z:01EE58203@c<@0403hm>`1:BDL0BTU704Q7
A@80<S4_00@08B0N02HU8`0V9B<0<2l]0P0e=380100S8R00<C0^034`;P0l>cT203Xi=`0402h]:`0T
8b4092<Q028Q7`8092<Q00@08B0N02<R800S8R0092<Q0P0Z:BL0100d<c40?3/i03`k>@15A482055@
CP0404]:B00[:RP0:bXX03Xi=`808B0N00@09RDS02l^;00_;R`0De9@0P0b<Bl304I5@`0403Xi=`0X
9bD0:2LU02DT8P80;Rd[00<0<C0^02/Z:00[:RP00`0];2X0100a<2h0>STg03Xi=`11@3h203ln?004
03<b<00X9bD0:2LU02LV9080;B`Z00@0;2/Y034`;P0a<2h0<2l]0P0m?3X01010?cd0@d9004=2@019
A3h203Ha:`0402/V800c;RP0<bhX03/f<080>3<]00@0=c8/03`g<@0l=c40?Sdk0P0c<S0102/Z:040
:BPV0@3oool000`0AeEP0`19Ef8304QFH@@0AeEP0P16E5l@04MEH0<0B5IQ7@19Ef8<04aJI@`0BeUT
00<0BEMR04YHH`1:F6<0101:F6<204UGHP0404QFH@17EF00AeEP04EFGP80AUMO00@0AEIN04AEG@14
EEd0@eAL0P14EEd01017F600BEYR04UJHP1;F60204ICF@0403i9C`0a?400<Ca002@]<P806R@V00@0
5AdP01<K7@0C6ad051DG0P0@4A<0100>3Q003Ph>00h>3P0<2P/200d<2P0401PD4`0S7a`08alL02@Q
6P80:b/S00@0;S4X02P]9P0X;BH07B8L0P0L7QT0100W9R@0<bl^03<_;P0a;302030^<@04034_<P0]
;Bl0;Bd_02TY:`809RLY00@09bPZ02HZ:`0V:R/0<2hc0P0V92T0100O7B0061HI01PF6@0?3@h200/9
2P0400X91`081`D020L500X300806a8;00@0:1/C03/Z8P0k:R80@2XO0P1F>Rl0101L?340HClc064o
<`1RAcH206=8=`0406E:=`1VCCT0ITdi06M>>P80ID`h00@0H4Xc05a6;`1LARl0I4Da0P1O@2`0101G
>2@0DC8N054b7P1@<Ad205Pi9@0406A5<@1]CSX0KDhj06];=P80K4`g0`1^CST0101/C3L0JDTd06U9
=01VAS4206=3;P0405m0:`1R@bh0HT<^06M8<`80JT/f00@0K4dh06Y;=P1ZBcH0IdPc0P1UAS4306E5
;0<0I4@[00@0IDD/06M7;P1WAbh0JDT`0P1[Bc8607EE?00407IF?@1hF3l0N5Po07]K@P80O5a30`1n
GTD307mOAP0407iNA@1mGD@0OEe407aL@`80Ne]200@0M5Xo071E>P1`ECX0JT``0P1cE3P0101aCC<0
Ne<i07]C>@1kDCP208AJ@@0408]QB022FD<0PUU3085LB@80UGEV00@0Ec`c02hI500^6A@0ETI60P0l
;Rh0100_;S@0<C0f034`=P0Y:2d203@c>00304A4AP0i>CT0>CTi00<0;Bd]00@0:BLX034_<00a;c00
ATA50P0]:b`0100b<3408b4R02<Q8P0f=3D203Le=P0402`Z:`0d<S<0=38c02DS9080?S`m00@0CD]<
03Tg>00i=cP0<c4b0P0R8240100d<S<0?cdn03lm?P0J61T206ESI0<0>SPi00@0AdE602/Y:P0[:BX0
:RPY0P0m>c`0100i=cP0:2HW02PV9`0Y9bP202XX:@04030^;`1BD540DU1A02l];P80:BLX00@0928S
02`Z:`0/:R/0EE=D0P0^;2d0100i=cP0;R`]02h/;@0i=cP202d[;00405eKG00]:b`0;B//02l];P80
<S0a00@0;2X[02TW:00Y9bP0CD]<0P0J61T0100Z:2T0;R`]02h/;@0_;Bh2038`<@0404m=CP0f=3D0
=S@e02d/:P80<c8`00@09bHT03@c<@0d<c40ATE30P0e=3801010?cd0<S4_038a;`0n?C/202XY9`04
04e<BP0`;bd0<2l]02l^;080<S4_00@0?C`j030_;@0`;bd0DU5?0P19B4H0100h=cD0:2LU02PW9@0V
9B<202h]:`04030_;@0];2X0;B`Z03`k>@<0?chl0P0d<c40100U92807QdK01hM6`0Q81h302PW9@80
;B`Z00@09B@R02<R800S8R00=3<a0P0o?S`01015A480Cdi<04m>C00b<Bl203`k>@0402/Z:00e=380
=C@b02/Z:080<C0^00@0>CPf06]ZJ01[JVP0BTU70P0n?C/0100_;R`09bHT02LV900V9B<202TX9P04
02d/:P0`;bd0<2l]04E4@P80>cXh00@0<2l]02/Z:00[:RP0;B`Z0P0_;R`0100];2X0:RTW02XY9`0T
8b42030_;@0402/Z:00/:bT0;2/Y02l^;080BTU700@0ATE303/j>00k>SP0;RTS0`0a;2H2030[9@04
03Pc;@14?cT0A3li04Q3?@80A3li00D0@3lm030_;@0`;bd09B@R02LV90010?ooo`003017EF0304UG
HP<0B5IQ1017EF0204IDGa00AeEP0`18EV4M04UGHP`0C5YU301;FF@00`19Ef80BUQS04YHH`0404YH
H`80BEMR00@0B5IQ04MEH017EF00AEIN0P16Eel01015EUh0A5EM04AEG@13E5`204AEG@0404MHH019
FV80BEYR04]HH080AU=I00@0?TU?034l@00a?40092db0P0J92H0100E7B004a/M01<K7@0D5AL2010A
4`0400h>400>3Ph03Ph>00`:2`803@`:00@061@C02<O700S7a`0924J0P0[:b<0100^<BP0:2dV02P]
9P0M8Q`201`N6@0402LV900c;bh0<bl^034/<080<2ha00@0<Blb02d];`0];Bl0:BT[0P0V9bT0100W
:2X09RX[02HZ:`0`;S<202HT:@0401lM800H5QT061HI00l=3P802`T:00@02PT700P71@081`D02P<0
0P0K4P/0100X6a<0>bXR03/Z8P10:Ql205Hj;`0405`l<@1Q?c<0HClc0697=P80HdPg00@0IDXg06I=
>@1VCCT0Idhj0P1UC3P0101PBS<0G4H_05a6;`1TAC4205m0;00405Lh901A<Qh0DC8N050a7@80F3TU
00@0I4Da06e>>P1]CSX0Jd/f0P1/C3L306i>>@0406a<=`1YBC@0JDTd06I6<@80Hd<^00@0Gd0[0693
;P1R@bh0IdPc0P1ZBcH0101/CCP0JT/f06Y;=P1WB3<206E6<@<0IDD/0`1TA2/0101UAB`0IdL^06M7
;P1YBC0206];<PH0MEDl00@0MUHm07QH?`1hF3l0Ne]20P1lG4<307iNA@<0Oem600@0OUi507eMA01m
GD@0O5a30P1kFd80101dFSl0L5Dj071E>P1ZC30207=D>004075=<`1kDcT0Ne<i07]A>080Q5Y100@0
Rf58089I@`22FD<0PEa90P2EMFH0101G?3<0;QTD02hI501FATH203`^;P0402l^=00a<3H0<C0f02TX
;@80=3<h00<0A4A603Ti>@0i>CT00`0];Bd0100Y9bP0<Bl`034_<016A4D202d[;004038`<@0S8B80
8b4R03Hd=@80=cDf00@0;2X[03@b<`0d<S<09B<T0P0n?3d0101=Bd`0>CLh03Tg>00c<C82028P8@04
03@b<`0o?Ch0?cdn01XH6@80IF=T0`0j>3T01017ADH0:bTZ02/Y:P0Z:2T203dk?00403Tg>00X9RL0
:2HW02TW:080:RPY00@0<2h_059@D@1BD540;bd^0P0Y9bP0100T8R<0;2X[02`Z:`1EDe@202h/;@04
03Tg>00^;2d0;R`]03Tg>080;B//00@0GE]L02d[;00]:b`0;bd^0P0b<340100/:R/0:BLX02TW:01=
Bd`201XH6@0402XX:@0^;2d0;R`]02l];P80<S0a00@0Cde>03Hd=@0f=3D0;B`Z0P0c<S00100W9R@0
=3<a03@c<@16AD<203Dd<P04040o?@0b<Bl0<S4_03hm>`80:RTW00@0CDa:030_;@0`;bd0;bh/0P0b
<Bl0100m?3X0<2l]030_;@1BDDl204U8AP0403Pg=@0X9bD0:2LU02HU8`80;Rd[00@0<2l]02d/:P0]
;2X0?3/i0`0o?S`203@c<@0402DT8P0N7A/07QdK024P7P<0:2LU0P0];2X0100U92808b8P02<R800d
<c4203ln?00404E4@P1?CT`0Cdi<038a;`80?3/i00@0:bXX03Dd<P0e=380:bXX0P0a<2h0100i>3H0
JfYX06]ZJ01:BDL203hm>`0402l^;00W9R@09bHT02HU8`80:BPV00@0;B`Z030_;@0`;bd0ADA20P0k
>SP0100`;bd0:bXX02/Z:00];2X202l^;00402d/:P0Z:BL0:RTW02@S8@80<2l]00@0:bXX02`[:@0/
:bT0;bh/0P1:BDL01016AD<0>cXh03/j>00^:B<3034/9P80<2/U00@0>3<]04@o>@14?cT0B4<m0P14
?cT01@10?cd0<2l]030_;@0U92809bHT0040oooo000<04MEH0@0B5IQ0`17EF0504IDGa00AeEP0`18
EV4M04UGHP`0C5YU301;FF@304UGHP@0BeUT0P19Ef801017EF00AUAO04IDG`17EUd204EDF`0404AC
FP19F5l0BEQO04iMI080CeiU00@0BEQO04=BF@13DUT0BeQP0P18EEd01015DUP0AU5G04IAE`13CU82
03U2A`0402Tc=@0M9BP07BDX01DI70803Q8E00@02Ph?010B4@0@4Q404A<B0P0=3@/010092@L02PT7
00X91`0=2P<201LG3`0402DX7`0_=2d0;c@]030e;P80:bdX00@08b8N01hJ6@0N6QT09B0T0P0Y9bX0
100a;c80=3@f03@d=P0c<cD202l`<P0402`];`0Y;Bh0:Bd^01/F6P806QHG00@05Q8C00l;2P0?2`X0
2PH50P070`0010060P001@4000D100090P0202HM5P0403d`:01@?S@0D3hd04`d:080Gd4f00@0J4Lh
075=?`1aCCl0L5E40P1WCC`0101SBCP0ITlm06I??@1/ED<206E@?@04058m:P11;1T0@B`I04TZ5@80
Cc0K00@0F3TT05m0:`1O@2/0ITLb0P1/CCP0101bDch0MUM207IG@P1^CST206e=>00406Y:=@1YBC@0
JDTd06];=P80Kdlj00@0Le<n07IF@@1fET40L54l0P1]CST0101WB3<0HD8]0652;@1M?RT505/l9`<0
IDD/00<0I4@[06=3:P1S@bX0101S@bX206A4:`0406E5;01VARd0ITH]06Q8;`80K4`c00@0L50g07AD
>`1dE3/0MeLn0P1iFD0307aL@`0407eMA01nGTD0OUi507iNA@80OEe400@0O5a307]K@P1kFd80NV15
0P1eFd00101]Cc<0Le@h07=D>01/BBd207E=<`0407A;;`20E3T0P5@i07iK?@80SFU?00@0OEe608MZ
FP27JUX0P6EJ0P17;b/0100k9bH0Ed=505M3A@0b<CL2034`=P0403He>P0/:c00;2/`03Ti>`80?3`l
00@09bLW03`l?00l?3`081hO0P0S8B80100n?3d0@T110490@@0Q7b0203<a<P0401`J6`1:B4T0BTQ9
02l];P80:2HW00@0<c4b02d[;00]:b`0?3Xk0P18ATL0100]:b`0<Bl`034_<00_;Bh202DS900403<a
<P1?CDh0Cde>05IDE@80=S@e00@0=38c03Hd=@0f=3D0:BLX0P13@D80100i=cP08R0Q028P8@0Z:2T2
02<Q8P0403<a<P11?d00@Cm0034_<0809B<T00@07adN02<Q8P0S8B80CD]<0P0V92D0100e<c@0:2HW
02PV9`0d<S<202DS900405eKG00[:BX0:bTZ030^;`80=cDf00@0=38c038`<@0b<340EUAE0P0O7Ah3
02d[;00402XX:@0X9RL0:2HW04]9BP80<S0a00@0?C`j03ln?00o?S`0<2l]0P0U92801014@d40=cHd
03Lf=017AT@203dl>P04038a;`1IF5H0FEQF038a;`80=C@b00@0:2LU02/Z:00[:RP09bHT0P1WIV@0
1017AT@0<S4_038a;`0Y:2H2038a;`0403He<`0/:bT0;2/Y02TX9P80<2l]00@0ATE303dl>P0m?3X0
;bh/0P0S8R00100N7A/081lM020O7@0W9R@202h]:`0402<R800^;B/0;Rd[02DT8P809RDS00@0@3lm
04I5@`16AD<0=SDc0P0a<2h0100j>CL0<2l]030_;@0k>SP2030_;@0402l^;01=C4X0CDa:05aKF@80
;bh/00@0ATE303Lf=00g=S@0<C0^0P0g=S@0100h=cD0=3<a03@c<@0l>cT204a;B@<0=C@b00<0<S4_
02`[:@0/:bT00`0U9280100Z:BL0<2l]030_;@0b<Bl202PW9@0402`[:@1<BdT0C4]9040o?@80:RTW
00@07QdK03Dd<P0e=380=2lY0P0`:bD0100_:R@0=S4[03Ha:`11?3H204E0>P04040k=@0i=2h0>C@^
02d/:P80<C0^0@0e=38103Lf=040oooo00001017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ3P17EF0<
04IDG``0AeEP0`1:F6<904UGHRP0BUQS1@19Ef8304QFH@0404IDG`17EF00AeEP04UGHP<0BUQS0P19
Ef801017EF00AUAO04IDG`18EFD204UFIP0404UGI018EV<0B5IS04IDG`80@U1I00@0?TaE03]9DP0k
BE809Blh0P0N9bh0100C72<031<I00`C6@0941H200`A5@0400hC5`0?51P03a@H01HA5`8040/?00@0
3PT=01DA4P0E4A80920Q0P0`;2/0100e<C00=30]03@`;@0^;B/202LV900401lN700K6aT06a/I020P
7P80:bdZ00@0>3Xg03e2?P0m@Sh0<Bd^0P0J5QD0100;1P<02P8000X20008000200X0000401P90P0c
91d0<b@M04Xg<080BSP^00@0CSP]058k;@1B>bd0G44^0P1XBC@0101aDCP0MULk07IG>`1[CcT206Q<
=P0406A8<P1O@bd0Gd<]05/o:@80Ec/U00@0E3PR05<g8@1C=b40EcLR0P1UAC00101_CcX0LTlk079?
>`1cD3`207]F@`0407mJA`1nF4D0OUQ507]HA0<0OEY60P1jEd<0101dDCd0K4Te06a9=@1U@Rh2068o
:`04060n9@1/BS40K4Xa075?=P80LE4h00@0LU@j06]=<`1[CC<0HDH[0P1TBBh0101[CC40Le@h07=D
>01hFCd307YH?@80OETo00<0NEDk079>=01bCS@00`1]Cc<0101XBRh0HT@X0694:01Q@bL206M9;@04
06]=<@1ZC300JT``07E?>080OEU100@0P5a407YH?`1jF3l0N5Hm0P1iFD00101jFT40MUHm07IF?@1g
G44207MI?`0407=E>`1_CcH0Kdlf06a<<`80K4Xa0`1[BC001024Ed80OE<k07eC>`2?Idd2089NA004
08e]E025IU80QFIB04HZ7080CS8W00@0>bdZ03`a;@0l<Bd0?S@c0P0c;Bl0100[:Bh0>3Tn03Pi?P0m
@4T202da>P0402LU:P0f=3L0=S@g01hL7`<0?3Xk0P0R8Al0100j>CL08b8N02<R7P0n?3d2030^;`04
02d[;00`;Rl0<2h_01PF5`80EeEF00@0=C<d03Pf=`0h=SL0518C0P11?d00100m>c`0<S0a038`<@1M
Fe`203<a<P04034_<00`;Rl0<2h_034_<080:bTZ00@0<Bl`0490@@12@4406QPI0P0/:R/0100f=3D0
=38c03@b<`19AdP201lM7P0402`Z:`0_;Bh0;bd^03hl?@80@Cm000@0=C<d024O800Q7b00=38c0P0^
;2d01018ATL0<Bl`034_<00d<S<2044o@00403dk?00e<c@0=C<d05ACD@80>cXh00@0<2l]020O7@0P
7ad0:2LU0P0c<S00100^;B/0Cdi<04m>C00n?C/204a;B@0404Y9A`0_;R`0;bh/02@S8@80>CPf00@0
AdI403/j>00k>SP0FeYH0P16AD<0100j>CL0<c8`03<b<00Y:2H2030_;@0404A3@@1>CD/0CTe;034`
;P80>STg00@0?3/i03<b<00c<S00;Rd[0P0e=380100o?S`0ADA204E4@P0e=38202`[:@0402<R800Q
81h08B0N02HU8`80:bXX00@0:RTW02PW9@0X9bD061LE0P0j>CL0101DDe40BTU704Y9A`0^;B/201hM
6`04028Q7`0Z:BL0:RTW034`;P80=SDc00@0=C@b0491?`12@Cl0DU5?0P12@Cl0100b<Bl0?chl03ln
?00S8R0203He<`0403dl>P0e=380=C@b03dl>P80CDa:00@0ATE302h]:`0^;B/092<Q0P0V9B<0100/
:bT0;bh/02l^;00[:RP202@S8@0402PW9@0a<2h0<C0^03`k>@<0@d900P0i>3H0100_;R`0;2/Y02`[
:@0[:RP202PW9@0403D`;00h<bl0>3<_03/f<P80?CPd00@0>cHb03Pc;`0h<bl0<bhZ0P0`:bL01@0/
:bT0<c8`03<b<00n?C/0ATE30040oooo00001017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ3P17EF0<
04IDG``0AeEP0`1:F6<904UGHRP0BUQS1@19Ef8304QFH@0404IDG`17EF00AeEP04UGHP<0BUQS0P19
Ef801017EF00AUAO04IDG`18EFD204UFIP0404UGI018EV<0B5IS04IDG`80@U1I00@0?TaE03]9DP0k
BE809Blh0P0N9bh0100C72<031<I00`C6@0941H200`A5@0400hC5`0?51P03a@H01HA5`8040/?00@0
3PT=01DA4P0E4A80920Q0P0`;2/0100e<C00=30]03@`;@0^;B/202LV900401lN700K6aT06a/I020P
7P80:bdZ00@0>3Xg03e2?P0m@Sh0<Bd^0P0J5QD0100;1P<02P8000X20008000200X0000401P90P0c
91d0<b@M04Xg<080BSP^00@0CSP]058k;@1B>bd0G44^0P1XBC@0101aDCP0MULk07IG>`1[CcT206Q<
=P0406A8<P1O@bd0Gd<]05/o:@80Ec/U00@0E3PR05<g8@1C=b40EcLR0P1UAC00101_CcX0LTlk079?
>`1cD3`207]F@`0407mJA`1nF4D0OUQ507]HA0<0OEY60P1jEd<0101dDCd0K4Te06a9=@1U@Rh2068o
:`04060n9@1/BS40K4Xa075?=P80LE4h00@0LU@j06]=<`1[CC<0HDH[0P1TBBh0101[CC40Le@h07=D
>01hFCd307YH?@80OETo00<0NEDk079>=01bCS@00`1]Cc<0101XBRh0HT@X0694:01Q@bL206M9;@04
06]=<@1ZC300JT``07E?>080OEU100@0P5a407YH?`1jF3l0N5Hm0P1iFD00101jFT40MUHm07IF?@1g
G44207MI?`0407=E>`1_CcH0Kdlf06a<<`80K4Xa0`1[BC001024Ed80OE<k07eC>`2?Idd2089NA004
08e]E025IU80QFIB04HZ7080CS8W00@0>bdZ03`a;@0l<Bd0?S@c0P0c;Bl0100[:Bh0>3Tn03Pi?P0m
@4T202da>P0402LU:P0f=3L0=S@g01hL7`<0?3Xk0P0R8Al0100j>CL08b8N02<R7P0n?3d2030^;`04
02d[;00`;Rl0<2h_01PF5`80EeEF00@0=C<d03Pf=`0h=SL0518C0P11?d00100m>c`0<S0a038`<@1M
Fe`203<a<P04034_<00`;Rl0<2h_034_<080:bTZ00@0<Bl`0490@@12@4406QPI0P0/:R/0100f=3D0
=38c03@b<`19AdP201lM7P0402`Z:`0_;Bh0;bd^03hl?@80@Cm000@0=C<d024O800Q7b00=38c0P0^
;2d01018ATL0<Bl`034_<00d<S<2044o@00403dk?00e<c@0=C<d05ACD@80>cXh00@0<2l]020O7@0P
7ad0:2LU0P0c<S00100^;B/0Cdi<04m>C00n?C/204a;B@0404Y9A`0_;R`0;bh/02@S8@80>CPf00@0
AdI403/j>00k>SP0FeYH0P16AD<0100j>CL0<c8`03<b<00Y:2H2030_;@0404A3@@1>CD/0CTe;034`
;P80>STg00@0?3/i03<b<00c<S00;Rd[0P0e=380100o?S`0ADA204E4@P0e=38202`[:@0402<R800Q
81h08B0N02HU8`80:bXX00@0:RTW02PW9@0X9bD061LE0P0j>CL0101DDe40BTU704Y9A`0^;B/201hM
6`04028Q7`0Z:BL0:RTW034`;P80=SDc00@0=C@b0491?`12@Cl0DU5?0P12@Cl0100b<Bl0?chl03ln
?00S8R0203He<`0403dl>P0e=380=C@b03dl>P80CDa:00@0ATE302h]:`0^;B/092<Q0P0V9B<0100/
:bT0;bh/02l^;00[:RP202@S8@0402PW9@0a<2h0<C0^03`k>@<0@d900P0i>3H0100_;R`0;2/Y02`[
:@0[:RP202PW9@0403D`;00h<bl0>3<_03/f<P80?CPd00@0>cHb03Pc;`0h<bl0<bhZ0P0`:bL01@0/
:bT0<c8`03<b<00n?C/0ATE30040oooo00001017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ0P17EF0H
04IDG``0AeEP00<0B5IQ04UGHP19Ef801P19Ef8[04YHH`D0BEMR0`18EV4304MEH0H0BEMR0`17EF00
0`16Df<0AeAT04MDI00304QFH`0404MEH015De`0AE=L049@F@80@4iG00@0>TA=038k@P0b>d809Rlf
0P0M92X0100F7B<04aPL01<H700>4aL200/@500400`:3`0<2Pd030X=00`:3@803P`=00@0518C01/J
600K6QP08R4O0P0V9B40100[:RP0;B`Z02d/:P0];2X202XZ:00402@T8P0O7ad07alM01`N6`806adJ
00@040X:00D0000500001P000P0<0`00100H3@L08QD=028E3@0a8aX2040_9`0404dg;01A>bh0DC/^
05Q1<`80HTLf00@0JDdh071@>@1`D3T0M58g0P1eDSH0101/CSH0JD/c06U;<`1RA2`205Pj8P04058d
701F>200ESPP0653:`80K4hf00@0JDTb06m?>01_CcP0M5@m0P1dDC/0101eDC/0NEDo07UE?`1oFD@2
089LA`0407eJA01nFdD0OU]5081MA`<0PUm90P1oG4H0101kF480N5Do07QE?`1`CSD207QF?@0407ME
?01cDcX0Le<j07EE?080KU0f00@0J4X`06]@=@1[D3D0NEXn0P1nGd<0101oH4@0Oee207mM@P21GD<2
08APAP0408AOA@20Fd40P5]107]L@080NU/o00@0MEHj06]<<01[C300IdP/0`1XBBd206E6:P0406i8
<@1dCSL0M4hg07QD?080NEDm00@0NELn07QH?`1hF3l0MeLn0P1eEC`0101hGD80N5Y007QJ@01dES`2
071@=`0306a<<`1[BC00JdT`00<0JDL^00@0PeU308aRBP2<HTX0Qem50P23H4@0102?KE@0PV=?089S
C`1K?S0204<W7004048c;00j;BL0>RdW03H[:@80<BPY00@0;RT]03Dd>P0e=3X0=CPo0P0Z;CH0100a
;c@0<c4d03<a=00/:Rd203Xh>@0403`j>`0[:RP0:bXX02/Z:080:BPT00@0?S`m03Le=P0g=CH0?3Xk
0P0M6a`0100h=SL0:bTZ02/Y:P1;BDX202HT9@04030^;`0m>c`0?C/l03`j>`80?C/l00@0?3Xk05iL
G@1NG5d0=38c0P0i=cP0100[:BX0928S02@R8`0]:b`204e;C00402h/;@0X9RL0:2HW02<Q8P80;2X[
00@0EeEF02`Z:`0/:R/0;B//0P0[:BX0100d<S<0AdE604M5AP0d<S<202d[;00403hl?@0m>c`0?C/l
05aJF`80C4Y;00@0ATA504=1@P13@D80=C<d0P0[:BX01017AT@0:bXX02/Z:00Y:2H202/Z:00403Th
=P12@Cl0@T4o03Th=P80CTe;00@0?chl0450?P11@3h0;Rd[0P0f=C<0101;BTP0;Rd[02h]:`0O7Q`2
04]:B00404a;B@17AT@0AdI403Pg=@80:2LU00@0<2l]04a;B@1<BdT0AdI40P0S8R00100`;bd0:RTW
02XY9`0];2X2040o?@0405=BD01BDDl0DU5?03/j>0809B@R00@08R4O02XY9`0Z:BL0<2l]0P0Z:BL0
100O7Q`06APF01TH5P0O7Q`202PW9@0403hm>`0k>SP0>cXh03Dd<P80;Rd[00@0;B`Z02l^;00_;R`0
;B`Z0P0Z:BL0100a<2h0>3Le03Pg=@0g=S@203`k>@0404E4@P0k>SP0>cXh02d/:P80<2l]00@0>STg
02TX9P0Y:2H0<c8`0P1GEU@0101RHEl0ADA204E4@P0Y:2H202DT8P0403Pg=@0`;bd0<2l]02XY9`80
:bXX00@0;2/Y02/Z:00[:RP0;B`Z0P0`;bd0100l>cT0@3lm040o?@0l>cT202d/:P04028Q7`0T8b40
92<Q02h]:`80=SDc00@0>C@`04I1?@16@Cd0D4]70P18@cl0100g<Rh0:bHR02/V8P0_:RH203Pc;`05
02TX9P0c<S00<c8`0450?P1<BdT00@3oool0000404MEH018EV40B5IQ04QFH@<0BEMR0`18EV4204ME
H1P0AUAO3`17EF0304QFH@0404UGHP1:F6<0BUQS04YHH`80BeUT:01:F6<504UGHP<0B5IQ00@0AeEP
04QFH@18EV40B5IQ0`19Ef8304QFH@80AeEP00@0A55Q04EBHP15DV80AeER0P18EV<01019Ef80B5IO
04QFG`16E5d204ECG00404eGH016CeH0ATmF03a5C080=3]100@0:c8h028W:`0R9b/05AXN0P0;41@0
100:20d03P`?00h<3`0@3Q4200h<3@0400T7200:2@L02PT7018A3`806APD00@08B0L02TX900Y:2@0
<2l[0P0_;RX0100S91l05AHA01DF4@092PD200<400<01P0000@02@0001L<1P0G30H0;b4H0P12<BL0
101:>2`0BcT]04/i;@1J?S0205m4<`0406M:>01_D3/0Ke0k07ID>`80Me<i00@0Me8e07E@<`1eD3<0
JT`d0P1TARh0101J?2@0DcDM05<e7@1D=Qh205i0:00406i@>01jG4@0NUa407EE?P80MUHo00@0O5U3
07mLAP1oG4H0OUY40P1nF4<01021FdH0R69=08QRC@2=JE<208]WD@0408YVD02;Ie40RfMA08aXDP80
RVI@00@0Q61:07mKA@1oFdD0NUHn0P1nFT80101jF3l0N5Hm07QF?@1kFd8207YL@P0407UK@@20HTP0
P69808MUBP80RFM<00@0R6I;08APAP24H4H0QV570P2;IT`0102?Idh0SFE<08eUC027IDX208QVB`04
08=QAP1iEc`0NELl07=A=P80LDld00<0KD/`06Q6:`1XAR/00`1Y@b`0101^B340ME4i07EA>@1hESd2
07ME?0<0Le<j00@0N5e207QJ@01hFT00M5Hl0P1`D3L0101/C3<0JTP_06Y8;`1WAB`206I4:`0406Lm
9`24G4<0Q5a307Q@=P80Pf1400@0SVaA08QXD`28J5<0R6QI0P1M?C00100k:Al0<B0H034P600a8Qd2
038X9P0402dW:@0^;340;R`a03Lg?`80?3m800@0>CLl02TW:P0Y9bX0<S0c0P0a;c00100e<c@0=SDc
03He<`0S8R0202h]:@0402HT9@17ADH0AdE6034_<080928S00@0:bTZ038`<@0b<340?cdn0P10?Sl0
1011?d00<2h_030^;`0`;Rl203Dc=00402TW:01KFEX0FeUJ044o@080<Bl`00@0<S0a02l];P0_;Bh0
:bTZ0P10?Sl0100a;c00:2HW02PV9`0O7Ah2030^;`0405MEEP0c<C80<c4b02`Z:`80:RPY00@0;R`]
05ECE01EDe@0<S0a0P0g=CH0100b<340;B//02d[;010?Sl203Hd=@<0;bd^00@0<2h_03<a<P0c<C80
@T4o0P0V9B<0100Z:BL0;Rd[02h]:`0c<S0203Th=P0402l^;00j>CL0>STg05=BD080@T4o00@0<2l]
02XY9`0Z:BL0;bh/0P0i>3H01014@d40CDa:04e<BP0c<S0203hm>`0404=2@00j>CL0>STg03Th=P80
ATE300@0@3lm02LV900W9R@0>CPf0P0a<2h0100d<c40ADA204E4@P1?CT`204E4@P0403<b<00X9bD0
:2LU02`[:@80;B`Z00@0;2/Y02PW9@0X9bD09B@R0P0X9bD0100b<Bl0?3/i03`k>@0_;R`302d/:P<0
<2l]0P0];2X0100/:bT0;Rd[02h]:`0l>cT203ln?00403Th=P0a<2h0<C0^030_;@80;Rd[00@0:2LU
028Q7`0R8Al0<S4_0P11@3h0101:BDL0AdI404M6A013@T020450?P0403Pg=@0/:bT0;2/Y02d/:P80
:2LU00@09bHT02l^;00_;R`0>3Le0P0l>cT0100m?3X0?chl03ln?00Y:2H202l^;004038a;`0`;bd0
<2l]02h]:`80;bh/0`0b<Bl01015@3`0?cXf03lj=P0h<bl203Ha;@<0=c8^00<0=2l[02lZ9P0_:RH0
100g=S@103Pg=@40oooo00001017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ0P17EF0H04IDG`l0AeEP
0`18EV401019Ef80BUQS04YHH`1:F6<204]II2P0BUQS1@19Ef8304QFH@0404MEH018EV40B5IQ04QF
H@<0BEMR0`18EV4204MEH00404AAH@15DV80AE9R04MEHP80B5IS00@0BEMR04QFG`18EUl0AUAM0P15
De`0101=Ef00ATmF04I?EP0lAD`203@k@@0402/b>00R9b/08RL[01DJ7P802a0D00@02PP=00h<3`0>
30l040hA0P0>30d010091`P02PT700X91`0B4@l201TH5004024P700Y:2@0:BPT030_:`80;bhZ00@0
8b@O01DF4@0E5Q402@X50P03100300H0000400T0000G30H05``602lQ6080@S4W00@0BSP/04/i;@1;
>Bd0FSh`0P1OA3<0101WBSP0Ke0k06m@>`1fE3/207MC>@0407MB=@1eD3<0ME0c06Y<=080I4H^00@0
FS`T05<e7@1C=Ad0E3HN0P1N@2P0101^D3P0NUa407YLA01eECh207IF?`0407aI@`1oG4H0Oea607iJ
A080OUQ300@0PE]608QRC@28HTd0SFUC0P2;Ie40102:IU00RfMA08]WD@2<J58208YVD00408APBP1o
FdD0Oe]507YF?P80OUY200@0NUPo07QF?@1hESd0Ne]20P1jG480101iFd40P698081RB027IDX208UW
C00408QVB`24H4H0Q61608IQA`80RfI<00@0SfM>08eUC02=ID`0QfE:0P28IT/01023HDH0NELl07UG
?01cDCH2075?=00306e;<01XAR/0J4H[00<0JD</00@0KTPa07EA>@1eDCT0N5Hm0P1gEC`307=C>P04
07QM@P1hFT00N5Y007AF?080L50g00@0K4`c06Y8;`1ZB2l0IdD/0P1VA2/0101W?BL0Q5a308AL@`1h
D3H208=PA00408i/D@28J5<0R6QC08QXF@80GCd`00@0>bTO034P600a81P0<B8M0P0b:2H0100]9bT0
;R`a02h/<@0g=cl203`oB00403Tg?00Y9bX0:BLZ038`<`80<Bl`00@0=C<d03He<`0f=C<08b8P0P0^
;BT0100V92D0AdE604M5AP0a;c0202@R8`0402/Y:P0b<340<S0a03lm?P80@3ho00@0@Cm0030^;`0`
;Rl0<2h_0P0e<c@0100Y9bP0FeUJ05]IFP11?d02034_<004038`<@0_;Bh0;bd^02/Y:P80@3ho00@0
<Bl`02PV9`0X9RL07adN0P0`;Rl0101GEEH0<c4b03<a<P0/:R/202XX:@0402h/;@1EDe@0EE=D038`
<@80=cDf00@0<S0a02d[;00]:b`0@3ho0P0f=3D302l];P04030^;`0c<C80<c4b0491?`809RDS00@0
:RTW02h]:`0^;B/0<c8`0P0i>3H0100_;R`0>STg03Xi=`1CDU020491?`04030_;@0Z:BL0:RTW02l^
;080>CPf00@0A4=104e<BP1=C4X0<c8`0P0n?C/01013@T00>STg03Xi=`0i>3H204I5@`04040o?@0W
9R@09bHT03Th=P80<C0^00@0=3<a04E4@P15A480Cdi<0P15A480100c<S00:2LU02PW9@0/:bT202d/
:P0402`[:@0X9bD0:2LU02DT8P80:2LU00@0<S4_03`k>@0l>cT0;bh/0`0];2X3030_;@80;B`Z00@0
;2/Y02h]:`0^;B/0?3/i0P0o?S`0100i>3H0<C0^034`;P0`;bd202h]:`0402PW9@0R8Al08R4O038a
;`80@D0n00@0BTU704M6A017AT@0@d900P11@3h0100h=cD0;2/Y02`[:@0];2X202PW9@0402LV900_
;R`0;bh/03Pg=@80?3/i00@0?C`j03ln?00o?S`0:BPV0P0_;R`0100b<Bl0<2l]030_;@0^;B/202l^
;0<0<S4_00@0AD0l03lj=P0o>SH0>3<_0P0f<Bd303Lb;P0303@_:`0_:RH0;bXV00@0=cHd0@0h=cD1
0?ooo`0000@0AeEP04QFH@18EV40B5IQ0`19Ef8304QFH@80AeEP3015DehH04MEH00404ECGP16E5l0
AUAO04MEH080B5IQ00<0BEMR04YHH`1:F6<00`1;FF@X04YHH`D0BEMR3`18EV401015DV80AU=S04IC
H`18EV<204UGI00404UGHP19Ef00BEMP04QFG`80AeEN00@0D5YS04eFG@1=EUd0B55H0P15C580100n
AD/0<CHj034f>P0M8RH200hC5`0400l>4`0C4aD04a<E01@D5P80410@00@02PX:00P81P0820H03@d;
0`0C50l201PI500401dN6@0P7a/081lK01dL608061@A00@04@d:00d91P0=2@H05@X40P0E2`80100N
4@P0;R0E02hP5@13<bH20590<P0405Y7>01OBC/0GdTk06e<>@80Kdhk00@0Le0l07EC>P1eDcX0Me<i
0P1gDSD0101hDC80Me0a07M@<@1WAbh205hn9@0405Hf7@1J>R40FSXQ06M7;P80LU8i00@0MUHm07EE
?01eEC`0NELn0P1jF3l01022H4L0SFUA08eYD@2>J54208MQBP0408UPBP2>IDl0SVE?09IbFP80TFeE
0`2<J500102=JE40RVI>08YVCP23GdL207aH@004089NAP25HDT0QF5907mK@`80O5Y100@0Q69908AT
B`24I4/0Q6A;0P29Je40102;Idd0SVY@08iZD02<J4h208QSB@0408YRB@2?Idh0SfM>099XD080T6I>
00@0RfI<08iYC`2>JDl0S6M=0P25H4H01021G480P5]1081K@@1lEcd207IA=`0406i5;`1X?bT0J3lY
06M1:P80L4Xc00@0ME4i075?=P1aCcH0L4he0P1cDCP0101dFCh0MELm07EG?@1cEC/2071@=`0406a<
<`1ZB2l0JTP_06M5;080ID<Z00@0GSDQ07eD?P1mE3h0O5@j0P28HdP01025HDL0Q65;08AQB`2ELUl2
075@@@0404D]8@0m9a`0?BLL03`Y8P80>bhX00@0;2@R024L800Q7200;Rdc0P15ADd0100l>Sl08b4T
02<Q900/:Rd202PV9`0402TW:00l>cT0?3/i03<b<080;2/W00@07adN03@b<`0d<S<0BDM80P0?3@h0
100_;Bh0;R`]02h/;@0e<c@205QFE`0404I4A@0^;2d0;R`]02LU9P80:2HW00@0>3Hg02h/;@0^;2d0
E59C0P0b<340100h=SL0@3ho040n?`0_;Bh202/Y:P0402TW:00a;c00<Bl`02TW:080=38c00@0@d52
02h/;@0^;2d0:BLX0P0/:R/0100^;2d0H5iO061NG`0`;Rl203Pf=`04038`<@0Z:2T0:RPY03@b<`80
=S@e00@0;R`]02PV9`0X9RL0:BLX0P0[:BX0100i>3H09RDS02HU8`0];2X202XY9`0402HU8`0b<Bl0
<S4_03Dd<P<0>cXh0P10?cd01012@Cl0;bh/02l^;00[:RP2055@CP0405QGE@0Z:BL0:RTW040o?@80
:RTW00@0:bXX04Q7A@18AdD0EUEC0P19B4H0100i>3H0=C@b03Dd<P0`;bd203He<`0403ln?012@Cl0
@T4o03/j>080<c8`00@0=SDc03ln?00o?S`0;B`Z0P0^;B/0100_;R`0<C0^034`;P0`;bd202`[:@04
02HU8`0Q81h08B0N01/J60808R4O00@0;bh/03Th=P0i>3H0=SDc0P0_;R`0100d<c40@3lm040o?@0n
?C/203Xi=`0403<b<00Z:BL0:RTW02LV9080;2/Y00@0<C0^02l^;00_;R`0<C0^0P1<BdT0101?CT`0
<S4_038a;`0U928203He<`04040o?@0i>3H0>CPf02DT8P80;bh/00@0?3/i04A3@@14@d40@d900P0j
>CL0100b<Bl0;B`Z02d/:P0/:bT202LV900402@S8@0Z:BL0:RTW03Dd<P<0?Sdk0P0i>3H0100d;b/0
<bhZ03<^:P0b;BT2030[9`04034/:00f<Bd0=S4]03lj=P80AT4m00D0@3lm03Lf=00g=S@0:bXX028Q
7`010?ooo`0000@0AeEP04QFH@18EV40B5IQ0`19Ef8304QFH@80AeEP3016E5lH04MEH0<0AE=N00@0
AUAO04MEH017EF00B5IQ0P19Ef800`1:F6<0BeUT04]II00T04UGHP@0BUQS1@19Ef8?04QFH@<0B5EU
0`19Ef@00`19Ef80B5IO04QFG`0304MEGP0404aFG`1;E5/0BeAK04YCFP80BE1F00@0@TU?03@i?@0d
>Cd07R<W0P0>4aL0100E51T05ADG01DE5`0C4aD2018B4P04010@400?3`d03`l=00l?3@803a0;00@0
30d700/<1P0;30H02``60P0@3`X0100H5A0071LC01`G4`0L5a<201`D4@0401dA1@0R50T08Q@903<S
5P80BCLY00@0ED8c05]6<`1KAS<0I4hi0P1bFDD0101fDcd0N5@n07QD?P1iECd207YF?00407UD>@1f
DC@0MU4d079=<080Kd/[00@0IDD/05lo9P1O?bH0GShU0P1WAbh0101eEC`0O5a307aL@`1hF3l2079B
>@04085OAP21GDD0PEe508MSB`80TF]D00@0TVaE08iUC`2>IDl0SFA>0P2CJU@0102?JE80RfE>08]U
CP28HT/208YTC@0408mYDP2AJe@0TF]D08eWD080R69;0`2AJe@01027Hd/0Pem708=OA`28ITd208MU
C004081PA`23HdX0Pf=:08mZD080TVeC00@0UFeD099ZD@2BJU40TfUA0P2HJe@0102GJU<0TfI?09=V
C`2@ITh309AZDP80T6I>0`2>I4`0102:H4P0Q5Y208AJ@P1iD3X206m6<00406]2;01]Ac00KDL`071:
<`80KDTa00@0KTXb079@=`1bD3L0Ke@i0P1aDcT0101`DSP0L50g071@=`1]CC@206a:<@0406Q6;@1V
A2/0IT@[06U0;080L4La00@0P5Pn08AO@P24Gd80OUTn0P23GdD01029I540OEQ607eHAP1QA3@204hb
900403lW6`0k:R80>bXR03H[:@80;2HX00@0<2hc03m0A@0o@4D0@T150P0_;C00100W9BP0:2HW02PV
9`0S8B8203Xi=`0404a;B@0Y:2@0:BPT02/Y:P80;bd^00@0@T1102l];P0_;Bh081hO0P0b<340100a
;c00@Cm0044o@01@CTl203`j>`<0;R`]00@0?C/l024O800Q7b00CD]<0P1CDE80100d<S<0>cTj03/i
>P0]:b`202PV9`0402h/;@0d<S<0=38c02XX:@80;2X[00@0<S0a02PV9`0X9RL09bDV0`0Y9bP205aJ
F`0402l];P0a;c00<Bl`03Le=P80;R`]00@0;bd^03/i>P0k>CX0=38c0P0^;2d0100`;Rl0;B//02d[
;00W9R@2020O7@0402PW9@0V9B<09RDS02`[:@80ATE300@0D4m=04i=B`1>CD/08R4O0P0b<Bl0100c
<S00?C`j03dl>P1EE58204Q7A@0402XY9`0W9R@09bHT03Lf=080:bXX00@0<2l]04Y9A`1:BDL0E5=A
0P13@T00100f=C<0>CPf03Th=P0c<S0203hm>`0404A3@@0o?S`0?chl03He<`80<c8`00@0<S4_034`
;P0a<2h0;bh/0P0d<c40100g=S@0<c8`03<b<00/:bT202HU8`0402LV900[:RP0:bXX03Dd<P809RDS
00@07ahL02/Z:00[:RP0>cXh0P13@T001015A480AdI404M6A00a<2h302`[:@80;bh/00@0<C0^03Dd
<P0e=380?C`j0P13@T001015A480;Rd[02h]:`0[:RP203`k>@0403Xi=`0W9R@09bHT02TX9P80?chl
00@0BDQ604M6A017AT@0?chl0P0d<c40100Z:BL09bHT02LV900Z:BL302d/:P80:RTW00@0;2/Y03He
<`0f=C<0@3lm0P13@T00100m?3X0=C@b03Dd<P0b;BT2034/:004038]:@0g<Rh0=c8^03lj=P80A3lk
0`16@Cd01@0c<S00<S4_038a;`0`;bd0;Rd[0040oooo00001017EF00B5IQ04QFH@18EV4304UGHP<0
B5IQ0P17EF0<04IDGaP0AeEP0`15Deh01016E5l0AeEP04MEH018EV4204UGHP0304YHH`1;FF@0BeUT
02@0BEMR101:F6<504UGHPl0B5IQ0`18EFD304UGI00304UGHP18EUl0B5IO00<0AeEN00@0C5IO04]D
F`1;E5/0BU=J0P19D5H01012BDl0=3Tm03@i?@0N8bL200hC5`0401DD6@0E5AL05ADG01<C5@804Q8B
00@0410@00l?3@0?3`d03`l=0P0?40/0100<3@L02``600/<1P0;30H2010?2P0401PE400L5a<071LC
01`G4`8071@A00@07A45028D2@0R50T0<b<F0P19=bT0101E@S<0FdHc05]6<`1TCST2079IA@0407IC
?@1hE3h0N5@n07UE?@80NUHl00@0NE@i07IA=01fDC@0LTd`0P1_Bb/0101UAB`0GclV05lo9P1N?RD2
06M7;P0407EE?01lG4<0O5a307QH?`80LU8i00@0PEm6085MA@21GDD0Qf=;0P2AJe@0102BK5D0SVE?
08iUC`2=I4h209=ZE00408mYDP2;IDh0RfE>08QRB`80RVA=00@0SfUB095[E02AJe@0SFM@0P28HT/3
095[E00408MSB`23GdL0Pem708QVC@80QfE<00@0P61708=SBP23HdX0SfY@0P2BKE<0102EKE@0TVYA
099ZD@2CJE4209Q[E00409MZD`2CITl0TfI?091VCP<0U6YB0P2@ITh308iTC00408YPB024FT80Q5Y2
07U@>P80KdH`00@0Jd8/06e7<01]Ac00L4Xc0P1]BC40101^BS80LU0g079@=`1_E3T2075C>@04071B
>01`D3L0L50g06e==080K4Xa00@0J4H]06I4:`1VA2/0JD0/0P1`Ac401020F3h0Q5m208AO@P1nFCh2
08=OA@0408UTD@1mF4H0OEQ60654=080CS8T00@0?bLK03/Z8P0k:R80=R/Y0P0/9RP0100`;S<0?d15
03m0A@12@4D202l]<00402LU:00X9RL0:2HW02<Q8P80>STg00@0C4]902TX900Y:2@0:bTZ0P0_;Bh0
1012@440;bd^02l];P0P7Ql2038`<@04034_<011?d00@Cm0051>C`80?3Xk0`0^;2d0100m>c`08AlP
024O801=Bd`205=ADP0403@b<`0k>CX0>cTj02d[;080:2HW00@0;R`]03@b<`0d<S<0:RPY0P0/:R/0
100b<340:2HW02PV9`0W9BH302TW:080G5YK00@0;bd^034_<00a;c00=cDf0P0^;2d0100_;Bh0>cTj
03/i>P0d<S<202h/;@04030^;`0]:b`0;B//02LV908081lM00@0:2LU02HU8`0V9B<0;2/Y0P16AD<0
101@Cdd0CTe;04i=B`0R8Al2038a;`0403<b<00m?3X0?C`j05EDDP80B4M500@0:RTW02LV900W9R@0
=cHd0P0[:RP0100`;bd0BTU704Y9A`1DDe4204=2@00403He<`0i>3H0>CPf03<b<080?Sdk00@0A4=1
03ln?00o?S`0=SDc0P0c<S00100b<Bl0<C0^034`;P0_;R`203@c<@0403Lf=00c<S00<c8`02`[:@80
9RDS00@09bHT02/Z:00[:RP0=C@b0P0V9B<0100O7Q`0:bXX02/Z:00k>SP204=2@00404E4@P17AT@0
AdI4034`;P<0;2/Y0P0_;R`0100a<2h0=C@b03Dd<P0m?3X204=2@00404E4@P0^;B/0;Rd[02/Z:080
?3/i00@0>STg02LV900W9R@0:BPV0P0o?S`01019B4H0AdI404M6A00o?S`203@c<@0402XY9`0W9R@0
9bHT02XY9`<0;B`Z0P0Z:BL0100/:bT0=SDc03He<`10?cd204=2@00403dl>P0e=380=C@b038]:@80
<B`X00@0<RdY03Lb;P0g<Rh0?cXf0P14?c/304I1?@0503<b<00b<Bl0<S4_030_;@0^;B/00@3oool0
000404MEH018EV40B5IQ04QFH@<0BEMR0`18EV4V04MEH00404ECGP16E5l0AUAO04IDG`80AeEP0`18
EV4W04UGHP@0BUQS1@19Ef8304QFH@0404UGHP18EV40B5IQ04QFH@<0AeEP0`18EV4204UGHP<0BEIV
0`19Ef@00`18EV40AeEN04MEGP0304IDG@0404YDG@1:DeX0BU=J04I?EP80@DQ>00@0=Ca202L/<00W
;3005AXN0P0:3a<0100C51P0414C010A4`0>3a4200lA4004014C4P0A4a004A<@00h@3@802`d800@0
30l600/<100;30@02P/30P0?30D00`0E40X071D?01`E3`03020G4P0402PH2@0c8Q80<b8B04Tf9`80
GdXg00@0HT`g05e4;P1MA2h0HdTb0P1_ECh0101dD3X0N5@n07QD?P1mFD42081LA00407eI?`1fDSP0
MU8h06e:;P80Id@X00@0JTP]06]9;P1[BBh0Kddb0P1dDSL00`1iEc`0O5Xo07aJ?`0307aJ?`0408eY
C`2:IT`0RVI<08UTBP80RVE;0`2=ID`0102CJE40V6iF09Q^EP2@J4l208eUC00408aTB`2>ITd0SVI=
09=[DP<0UViE0P2DK5<0102DJeD0TF]D095[E027HDX208APB00408aXD02;JE00RfU@08ESBP80QVA;
00@0UViE09]cFP2KLeX0WWAL0P2LLUX0102NLEX0X79K0:1bF`2NL5T209QZD`0409U[E02MKeP0WFmH
09e_F080VF]D00@0UVQA09EWD02EIe00TF=<0P2;GDH01026G4H0OU@n07iD?P1eC3H20758<P<0KTPa
00@0L4`d07=?=`1cCcL0Je0e0P1^D3H0101_DCL0L50g071@=`1^CSD206a:<@0406Q6;@1VA2/0IT@[
06U3<080FS@M00@0L4/`07MB=@1gDSD0OULj0P2=ID/01028GdT0RV5=08YQC@21GTX20691;P0403hR
500e8AH0=B4F048c;P80@STj00@0=c8f034`=P0a<3H0@d560P12@4<0100/:Rd0;bd^02l];P0Y9bP2
03@c<@04055@CP0^;BT0;RdY02d[;080<Bl`00@06aTJ069PH@1RH64061HG0P0i=cP0100l>S/07adN
01lM7P1>C4d204A2@`<0=S@e00@09bDV03Tg>00i=cP0;B//0P1WIFH0100n?3d0;B//02d[;00R8242
02XX:@0402l];P0/:R/0;2X[02LU9P80:RPY00@0;B//02TW:00Y9bP0:BLX0`0R824204i<C@0303Hd
=@0/:R/0;2X[00<09B<T00<08R0Q03Pf=`0h=SL00`0c<C80100l>S/0>CLh03Tg>00Z:BL2028Q7`04
024P7P0S8R008b8P03@c<@80BdY800@0B4M503Pg=@0h=cD0=cHd0P0U9280100_;R`0Cdi<04m>C01B
DDl203Dd<P0402XY9`0h=cD0>3Le01hM6`80?chl00@0C4]903hm>`0n?C/0=cHd0P0j>CL0100g=S@0
;bh/02l^;01=C4X204m>C00404E4@P0a<2h0<C0^02HU8`80:BPV00@0<2l]034`;P0a<2h0EUEC0P1:
BDL0100h=cD0:2LU02PW9@0P7ad2024P7P0402DT8P0Y:2H0:BPV03<b<0809bHT00@08B0N02d/:P0]
;2X0?chl0P13@T00100i>3H0;B`Z02d/:P0];2X2030_;@0403@c<@0j>CL0>STg03dl>P80>3Le00@0
=C@b03Xi=`0j>CL0=3<a0P0[:RP0100V9B<0:RTW02XY9`0[:RP202`[:@0403Th=P19B4H0BDQ6051?
C@80ATE300@0=C@b02DT8P0U92808B0N0P0X9bD0100a<2h0=C@b03Dd<P0/:bT203@c<@0403ln?017
AT@0AdI404=2@080>3Le00@0;bh/02d/:P0];2X0>SDa0P0c;RX0100a;2P0>cHb03/f<P18@cl204U4
@00403Td<00Y9200:B@P02`[:@80<S4_0@0j>CL103ln?040oooo00001017EF00B5IQ04QFH@18EV43
04UGHP<0B5IQ0P17EF0H04QFH@`0AeEP0`16E5l704MEH2H0B5IQ101:F6<504UGHP<0B5IQ0`19Ef86
04MEH0<0BEMR0`18EFD01018EV<0AeER04MEHP17EF0504MEGP0404aFG`18DEP0B55H03i7CP80<CPn
00@08BP^01HK7`0F6al03A8F0P093Q80100?41@03PlA00h?4@0=3Q0200lA4004014C4P0@4Pl0418?
00h@3@802`d800@03A0700lB2@0?4PT04A8:0P0A3PL0100@2`D05`h901L>2@0R5a<202`N6`0403`[
6@1:=bH0BSLV05e8=@80IU0k00@0IDde0698;`1RB2l0IT/`0P1]CcD0101aCSX0MU<o07IC?`1kF4@2
07eMAP0407]KA01eECh0MEDn06i>=`80J4X`00@0LU0e07AB=`1dDSL0MeDj0P1jF3d00`1lFSl0OUa1
07iL@@03081N@`0408]WC@2<Idd0S6M=08QSB@80QUi500@0R617095WC`2AIdl0UfeE0P2IKeL00`2D
K5<0Tf]B09=[DP03095YD00309=[DP2DK5<0U6aC00<0UViE00<0TVUC08mVD02?IU000`23GDH0102>
JU80TViF099^EP2=Je8208m]E00409abFP2QMel0XGMO0:AgH080XWEN00@0XWAM0:EeG`2UMEl0XW9L
0P2MKEL0102OKeT0Xg=M0:=cG@2RLU`209a/EP0409QXDP2GIe40UfMA09=SC@80SEe700@0S69<08YP
BP2:H4X0PeU30P1hCcT0101cBS@0LT`e079<=@1dCSL207A@>00406]@=@1^D3H0KU0f071B>080L50g
00@0KThe06]9<01[BC00IT@[0P1S@BP0101X@Rl0G3HO05`f7`1W@RL20719;00407UB<`24GD00Q5e0
08=I@@80R5i800@0UFmH08=OB@23GdT0FS/Y0P0l91P0100l:b@0>S0_03X`;`0^:Bd202HT:@0403@b
=`19AdX0BDM:03<a=080<Bl`00@0=38c02`[:@0/:bT0?3/i0P10?c/0100Z:2T07adN01lM7P0Z:2T2
040n?`0404Y8B@0R82408R0Q04=1@P8081hO00@0=S@e04A2@`14@T<0?3Xk0P0b<340100X9RL0A493
04A2@`0X9RL204M5AP04055?D00_;Bh0;bd^02@R8`809bDV00@09B<T02XX:@0Z:2T0<c4b0P0/:R/0
100W9BH0:BLX02TW:00^;2d2028P8@0402DS9013@D80@d5204E3A080;2X[00@081hO03<a<P0c<C80
=S@e0P1>C4d0100m>c`0<c4b03<a<P0h=SL202h/;@0403Lf=00Y:2H0:BPV01`K6@808R4O00@0=C@b
03/j>00k>SP0;B`Z0P0R8Al0100m?3X0:bXX02/Z:016AD<205MFE00403Xi=`0f=C<0=SDc04E4@P80
=SDc00@0=C@b04I5@`16AD<0?3/i0P0X9bD0100c<S00A4=104A3@@0k>SP202LV900405=BD019B4H0
BDQ603Pg=@80:BPV00@09RDS02l^;00_;R`0?Sdk0P18AdD0100j>CL0<C0^034`;P0W9R@202<R80<0
9B@R00@08B0N01`K6@0L6aT07ahL0P0_;R`0100m?3X0@3lm040o?@0i>3H2030_;@0402TX9P0T8b40
92<Q034`;P80?chl00@0?Sdk03Th=P0i>3H0>cXh0P0b<Bl0100W9R@0:2LU02PW9@0R8Al303Pg=@80
8B0N00@07QdK03Th=P0i>3H0B4M50P12@Cl0100a<2h0>CPf03Th=P0m?3X203/j>00403He<`0c<S00
<c8`02d/:P809RDS0`0o?S`0100n?C/0>CPf03Th=P0];2X202HU8`0402`[:@0g=S@0=cHd034/:080
?CPd00@0B4<o04<n>P13?SX0=2l[0P0Y9200100Z9B40<2/W030[9`0f=C<203Lf=040>CPf0@0j>CL1
0?ooo`0000@0AeEP04QFH@18EV40B5IQ0`19Ef8304QFH@80AeEP6018EV4<04MEH0<0AUAO1`17EF0V
04QFH@@0BUQS1@19Ef8304QFH@<0BEMR1P17EF0304UGHP<0B5EU00@0B5IS04MEHP17EF80AeEP1@17
EEh0101<EUl0B55H04QAF00nAdh2034h?P04024X;P0F6al05Q/O00dB5P802@hB00@03a0D00h?4@0>
3a403@h@0P0?4A00100A4a80418?010B3`0>40d200/=200400d@1`0?4PT03a89014B2P804@h700@0
40/501L>2@0G3PT08QLC0P0/7Q/0100l:aT0BSLV04Xg9P1MB3D206I@>`0406E==@1RB2l0HTP_06I;
<080KDle00@0LDhj07IC?`1fDcl0NeQ40P1mGDH0101kFd@0MEDn07EE?P1^CSL206Q:<004079@=@1d
DSL0M58g07ME>P80NUPm00<0O5Xo07iL@@1nG4400`20GT<0102;Idd0S6M=08aWC@28HdT208INA@04
08QPA`2AIdl0TFM?09M]E@80VFmG00<0U6aC09=[DP2CJe800`2AJE000`2CJe80U6aC09A/D`0309I^
E@03099YD`2?IU00SfI@00<0Pee600@0SVYB099^EP2BKUH0SF]B0P2?KE@0102LLUX0XGMO0:5gG`2T
Mf020:9eGP040:9dG@2UMEl0YGEO0:9bG080WFeG00@0WfmI0:=cG@2SLed0XW9L0P2LK5H0102HJ580
UfMA09MWD@2CHdd208eMA`0408aRC02:H4X0RV1:08=I@`80N4li00@0LdXd079<=@1bC3D0M4hg0P1d
D3P0101[D3D0KU0f06i@=P1`DSP2071@=`0406i>=@1[BC00JdT`06I4:`80Hd4X00@0J48_05`f7`1L
=Ql0Id8W0P1`BB`0101iDS<0Q5e008AM@023FD4208QNB00409E_F023GdT0Pem905Xk:@80?2@H00@0
?2/T03X`;`0j<2l0;RT]0P0V92T0100d<SL0BDM:04U7BP0c<C@2034_<00403@b<`0/:bT0;2/Y03`k
>@80@3lk00@0:RPY01lM7P0O7Ah0:RPY0P10?Sl0101:B4T08R0Q028P8@13@D82020N7`0403Hd=@14
@T<0A49303`j>`80<S0a00@0:2HW04A2@`14@T<0:2HW0P17ADH0101ACe00;bd^02l];P0T8R<202LU
9P0402DS900Z:2T0:RPY03<a<P80;2X[00@09bDV02TW:00Y9bP0;R`]0P0R8240100U8b@0@d5204=1
@P15@d@202`Z:`04020N7`0c<C80<c4b03Hd=@80CTa=00@0?C/l03<a<P0c<C80>3Hg0P0^;2d0100g
=S@0:BPV02TX9P0L6aT2028Q7`0403Dd<P0k>SP0>cXh02d/:P808R4O00@0?C`j02/Z:00[:RP0ATE3
0P1GEU@0100j>CL0=SDc03He<`15A48203He<`0403Dd<P16AD<0ATE303`k>@80:2LU00@0<c8`04A3
@@14@d40>cXh0P0W9R@0101CDU00BDQ604U8AP0h=cD202TX9P0402HU8`0_;R`0;bh/03hm>`80B4M5
00@0>STg034`;P0a<2h09bHT0P0S8R0302DT8P04024P7P0L6aT071/I01lN7080;bh/00@0?C`j040o
?@10?cd0>CPf0P0`;bd0100Y:2H092<Q02@S8@0a<2h203ln?00403hm>`0i>3H0>CPf03/j>080<S4_
00@09bHT02PW9@0X9bD08R4O0`0h=cD2024P7P0401hM6`0i>3H0>CPf04Q7A@80@T4o00@0<C0^03Th
=P0i>3H0?C`j0P0k>SP0100f=C<0<c8`03<b<00];2X202HU8`<0?chl00@0?Sdk03Th=P0i>3H0;B`Z
0P0V9B<0100/:bT0=cHd03Lf=00a;2P203dh=00404Q3?`13?SX0@chj03@_:`80:B@P00@0:RDQ030[
9`0`:bL0=SDc0P0g=S@103Th=P40>STg0@3oool0000404MEH018EV40B5IQ04QFH@<0BEMR0`18EV42
04MEH0`0BEMR3018EV4E04MEH0<0AUAO9018EV4404YHH`D0BEMR0`18EV40101:F6<0BEMR04UGHP17
EF0304IDG`80AeEP00@0BEMR04YHH`1:F6<0AeAT0P16Df<304IDH@0304MEH017EEh0AeEN00<0B5IO
00@0C5IO04I?EP16CeH0=Sm60P0T:c40100B6Al02a0D00/@500;41@200dB5P0400/?4P0=4A803A4B
00lC508041@C00@03a<B00hC3`0>4`l0314=0P0<4@/0100:40H04A@;014D2`0D5@d2010=1P0400`5
000C2PD04`X502TK6080>R`Y00@0BSPT05Y5<P1JAC80IDlj0P1TBcD0101UBc@0KE8g06eB=`1fF3`2
07MI?@0407=C?P1dE3l0M5@o07IF@@<0MUM20P1cEd40101aECl0L5@n071D?P1iEc`207ME>P0407QF
>`1mFd00OE]008ARA`80QVA900@0PV1507aJ?`1lFSl0PEa20P26HDL01027HTP0QUi508INA@2;HdX2
09M]E@0309e`F@2KKUL0VfiG00H0T6Q?00@0TFU@09=[DP2CJe80UfmF0P2JLUT0102JLE/0UFaF09E/
EP27GTP2089LA@0408iXD@2AKED0TFeE08e[DP80SVaC00@0Wg9K0:=fG`2SMUl0XgIO0P2QLe`0102P
LU/0Xg=M0:=cG@2RLU`20:1]F0040:9_FP2ULUd0YG9M0:AaG080WV]F00@0VfQC09YWDP2JIe80UfA?
0P2AGTT0102?HTd0TfIA09=VD@2<HT`207iE?`0407I==`1fD3T0MU0i07MA>P80MDlh00@0K54f06mA
=`1_DCL0L58h0P1aDCP0101]CC@0JTP_06Y8;`1T@RT2060n9@0405hh9@1R?2L0HS`W06@o9080KTLZ
00@0Ld`]07Q?<`1hCc<0PeU00P29G4L0102:HTP0TF]D095[E01fED8204P^8@0402hM5@0Y7ad0:AlM
02`X:@80<c4f00@07adR04=1A013@D@0=C<f0P0/:R/0100k>CX09B@R02DT8P0R8Al2051?B`0402<Q
8P0]:b`0;B//030^;`80;2X[00@0CTa=02l];P0_;Bh05Q@E0P10?Sl0100O7Ah0B4I704Q6A`17ADH2
02`Z:`0404M5AP0j>3T0>SPi04=1@P806QPI00@0FEMH03Xh>@0j>3T0<S0a0P0X9RL0100P7Ql0=C<d
03Dc=013@D8202HT9@0401dK700V92D09R@U03<a<P809bDV00@0;B//044o@011?d00De5B0P0]:b`0
100`;Rl0A49304A2@`0m>c`204M5AP0402TW:00Q7b008AlP034_<080;R`]00@0=cHd02PW9@0X9bD0
6APF0P0U9280100j>CL0=SDc03He<`0];2X203<b<00403He<`18AdD0B4M504M6A080=SDc00@0;Rd[
03<b<00c<S00?C`j0P16AD<01017AT@0A4=104A3@@0g=S@202l^;004034`;P0_;R`0;bh/03Lf=080
C4]900@0D4m=03/j>00k>SP0;B`Z0P0f=C<01015A480AdI404M6A00n?C/203Dd<P0402PW9@0^;B/0
;Rd[034`;P80:RTW00@07ahL01dL6P0M71X09RDS0P0a<2h0100^;B/0A4=104A3@@1AD4h20491?`04
02d/:P0X9bD0:2LU03<b<080?Sdk00@0;Rd[04A3@@14@d40?C`j0P0/:bT0100b<Bl0<C0^034`;P0W
9R@202PW9@0403Th=P0Y:2H0:BPV02h]:`80BDQ600@0Cdi<03Lf=00g=S@09B@R0P0W9R@0100Z:BL0
=cHd03Lf=00m?3X203@c<@0402/Z:00];2X0;B`Z03Lf=080?Sdk00@0?chl03Xi=`0j>CL0=cHd0P0c
<S00100[:RP09B@R02DT8P0Y:2H2038a;`04048m>@16@Cd0AT4m04I1?@80?CPd00@0<B`X02dX900]
:2@0<RdY0P0j=C401@0n?C/0>STg03Xi=`0c<S00;bh/0040oooo000H04QFH@0404MEH018EV40B5IQ
04QFH@<0BEMR0`18EV4204MEH0`0AUAO1018EV4204MEH0<0AUAO0`15Deh<04QFH@@0AeEP0`18EV45
04UGHPd0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00`0B5IQ00@0@eES04IHIP16F6H0AUIS0P16
E5l0101;EEh0CeQO04mHG`1<DeT304=:D080=ci400@09R/a01DJ800E6R0030lF0P0:3A@0100>3QH0
4A4I014A6@0A3@h200l=3P0400/?3P084@h0214>00PA3P802Q0>00@02`l>00d?3P0=3`h0418=0P0C
50l0100E4Pd050`901@<2@0I2`P202XH5004044/:@1B>SP0DSXh06A7>@80Gd8b00@0I4Dc071A?@1`
DCd0NEU20P1lFSl0101kFCd0OE`m07eL?@1nFT8207iL@`0407eMA01nH4H0OV1607aQAP80NEm400@0
MEe107=K?`1cFcl0Nea00P1lGD401020GT80QF=708ESA`29ITP208QUA`0408AP@01oFc/0Oe/k08ER
@P80Pelo00@0T6a<08eVB@2=ITT0Qem50P2CJE00102CJE40TVQ@099XD02GKeH208iVC@04099ZD@2G
KED0UfeE08mUC@80TVE>00@0W6iG09i`F@2NL5T0VW5K0`2EK5H209U`FP0409Q_F@2AJ580TFQB08iU
C`80TVUC00@0WVYE0:McGP2WLeh0[WYT0P2]NF<0102/N600[GYO0:ejG`2/NEh20:UfF`<0ZGEO00@0
ZWIN0:YgG02ZMe`0Z7EJ0P2SL5<0102NJdh0VVM:09YWBP2PK5@209ITC@0409ATC@2AI4d0TFA=085I
@080NE<l00@0NeLo07YH?`1jF3l0LE4h0`1fESd2075A>0<0KThe00@0JTXa06A4:`1TA2/0GS/W0P1J
?2@0101C=al0CcDJ04le6P1J?b@2075B=P0407mK@@23Fd80Pe]208ELB080RVAA00@0R6=C06m=@`1_
CD<0ARTS0P0]4a80100d7R00B3@f04Pd=P0_9bD202TU9004059BE00g>3`0=cPl02l`=@806QXL00@0
:2@S03Xe<P0j=C80?d=60P0V9bT0100a<S@0<C4a034a<@0j>SX203Dd<P04034`;P0/:2D0;2PU038a
;`808R4O00@0Cdi<04U8AP19B4H0BDQ60P0`;bd0100X9bD0=SDc03He<`0V9B<204i=B`04034`;P0k
>SP0>cXh01/J6080>cXh00@0>STg02d/:P0];2X081lM0P0W9R@0100_;R`0:bXX02/Z:00Y:2H203Xi
=`0404=2@00f=C<0=SDc04I5@`80@d9000@0?3/i02d/:P0];2X08B0N0P0T8b40100_;R`0=SDc03He
<`0d<c4203<b<00402LV900J6AL06QTG02DT8P80?Sdk00@0A4=103Th=P0i>3H0>3Le0P10?cd01015
A480?3/i03`k>@0];2X202TX9P0403Pg=@1;BTP0BdY8055@CP80<c8`00@0=3<a03/j>00k>SP0;B`Z
0P0f=C<0101;BTP0AdI404M6A00`;bd2038a;`04030_;@0l>cT0?3/i059AC`80CDa:00@0<S4_028Q
7`0R8Al0;R`]0P0[:BX0100X9RL09B<T02DS900U8b@202HT9@0402PV9`0Y9bP0:BLX04U8AP80>cXh
00@0=3<a03Pg=@0h=cD0>CPf0P0a<2h302/Z:00304i=B`0h=cD0>3Le00<0;B`Z00@09bHT02TX9P0Y
:2H0=cHd0P11@3h01013?Cd0>S@d03Xd=00_:RL202XU8P0302dX900b;BT0<RdY00<0<bhX00@0<RdY
02hY9@0^:BD0;RTU0P0g<Rh01015@3`0BdI204]6@P16@Cd203hi=@0402lZ9P0]:2@0;BPT02`W8`80
<2/W00@0=c8^03lj=P0o>SH0@chj0P15@3`01012>SL0>C4^03Ta;P0`:2D202lW900403H^:`0n=S<0
?SHc04<k>080A3`i00D0?3Ld03Pc<00h<c00<RdZ02lZ9`010?ooo`006018EV401017EF00B5IQ04QF
H@18EV4304UGHP<0B5IQ0P17EF0<04IDG`@0B5IQ0P17EF0304IDG`<0AE=N3018EV4404MEH0<0B5IQ
1@19Ef8=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0<04QFH@0404=EH`16F6H0AUQV04IFH`80
AUAO00@0BeEN04mHG`1?F5l0C5=I0`13BU0203LnA00402H[<@0E6R005AXP00`?5P802PdD00@03PhF
014A6@0A4AT04@d>0P0?3@h0100;3`h0214>00PA3P084@h200X@3P0400/?3P0=3`h03@l>010B3@80
4a@?00@05A8=01@<2@0D30T06@/80P0Z61@01011;2T0DSXh058j>01TAcT205m2<P0406A5<`1`DCd0
L54m07UI@P80O5Xo00@0NeTm07eL?@1mG3d0OUY20P1nG4<0101mGD@0OV1607iPAP1lHDH207UOA004
07EM@@1cFcl0Le/o07]L@080O5e100@0P5i208ESA`25HdL0RFI80P28IDL01024H400Oe/k07mK>`25
HT8208=O?`04091/C02=ITT0SFI908MOA@80TfU@00@0TfUA099XD02BJ500UfmF0P2>ITd0102BJU40
UfeE09M]E@2?IDd2099UCP0409a^E`2NL5T0WW1I09YaF`<0UFaF0P2IL5X0102HKeT0TFQB095XDP2>
IDl2099YD`0409iZE@2WLeh0Yg=N0:ijI080[GUS00@0[7QP0:ejG`2]NUl0[7UN0P2YMU/30:UeG`04
0:YfGP2ZMe`0ZWML0:QeFP80Xg1C00@0WV]>09YWBP2JIdX0X6aD0P2FI4d0102DI4d0TFA=095TC@21
FD0207UC?00407]G?`1jF3l0NUPo075A>0<0MUHm0P1aDCP306i>=@0406Y:<@1TA2/0I4@[05hk9`80
FS`T00@0DcLO04le6P1?=AX0FSlT0P1aDSH0101oFd40Pe]208=K@P25G4P208YTD@0408QSD`1_CD<0
Kde304HY8`80;A<B00@0=1hP04Pd=P18=3H0;bLU0P0Y9B@0101BDU@0=cPl03Lh?00_<3D201XJ7004
02PT8`0j=C80>SDb03m3AP809RLY00@0<C8d034a<@0a<C40>SXj0P0e=380100a<2h0;2PU02`X9@0b
<Bl2028Q7`0404m>C019B4H0BDQ604U8AP80<2l]00@0:2LU03He<`0f=C<09RDS0P1>CD/0100a<2h0
>cXh03/j>00K6QP203/j>00403Xi=`0];2X0;B`Z020O7@809bHT00@0;bh/02/Z:00[:RP0:BPV0P0j
>CL01013@T00=SDc03He<`16AD<204=2@00403`k>@0];2X0;B`Z024P7P8092<Q00@0;bh/03He<`0f
=C<0=3<a0P0c<S00100W9R@06QTG01XI5`0U928203hm>`0404A3@@0i>3H0>CPf03Pg=@80@3lm00@0
ADA203`k>@0l>cT0;B`Z0P0Y:2H0100h=cD0BdY804]:B01AD4h203<b<00403@c<@0k>SP0>cXh02d/
:P80=SDc00@0BdY804M6A017AT@0<2l]0P0b<Bl0100`;bd0?3/i03`k>@1BDDl204e<BP04038a;`0R
8Al08R4O02h/;@80:bTZ00@0:2HW02DS900U8b@09B<T0P0V92D0100X9RL0:BLX02TW:019B4H203/j
>00403@c<@0h=cD0>3Le03Th=P80<C0^0`0[:RP00`1>CD/0>3Le03Pg=@0302d/:P0402LV900Y:2H0
:BPV03Lf=080@D0n00@0@cdm03Xd=00j=3@0;bXW0P0Z9B800`0]:2@0<RdY038]:@0303<^:004038]
:@0^:BD0;RTU02hY9@80=c8^00@0AD0l04]6@P1;AT80AT4m0P0n>CD0100_:RH0;BPT02dX900/9b<2
030[9`0403Lb;P0o>SH0?cXf04<n>P80AD0l00@0@SXg03Ta;P0i<Bh0<2PU0P0_9b@0100f;R/0?SHc
03hf<`13>cP204@l>@0503`g=00h<c00>3<`038]:P0_:RL00@3oool001P0B5IQ00@0AeEP04QFH@18
EV40B5IQ0`19Ef8304QFH@80AeEP3016E5l404QFH@80AeEP1016E5l204ECGQ00AeEP0`18EV4504UG
HPd0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00`0B5IQ00@0@eES04QHI@18F6D0BEUV0P18EEh0
1019De`0C5EL04aEG01;DUP204I=D`0403/o@P0]<C@0;C4d01/O8P804A8F00@03a0D014@5@0A41D0
4A0E0P0@3Q<0100B3Pl03`l?00l?3`0<40l200/A3`0400TB3`0;4@l02a4?00h@3`803`l?00@04A0<
010<2@0@30T04PX70P0J3`/0100Z61@0>bHQ03/V8@1==30205Xn>P0406A7=`1WAcP0IdLh06m@?P80
O5a700@0PV9;085OA021Gd@0OE/o0P1mG3d01022G4D0PUi6089NAP23GdL208ARB@03089RB@22I4X0
PVA:00<0OfA900@0RFE;08QTBP28I4X0RFI:0P2;IT/0102;ITT0S6E608aUAP29HT<208MN@004091/
C02;IdL0RfM709M`D@80SfQ;00<0R61609E[DP2EJe800`2FK5@0102FKUD0SVI=08iVC@2CJE4209Q^
EP0409AWD02EJ540UFQA09maFP80XWAM00@0UVeG09M^F02GKUP0Vg9L0P2PMf40102QN680WgIP09mf
H02OMV020:5hHP0309i[EP2ULUd0YG9M00<0ZgUR00@0[7YQ0:mmHP2_OF80[WaQ0P2[NEh0102aOfH0
/7iU0;1nI@2^O6<20:ajG`040:UgG02VMEL0YWEG0:AcE@80Xg9B00@0X6iE09QVC@2HITd0UVI?0P2D
J4l01026G4@0O5Ho07aF?`1nFT8207aJ@@0407IF?@1jFT40NUY107UI@080M5@k0`1bDST0101`D3L0
Jd/b06];<P1V@bl205i0:00405@h801>=1T0CS@I05Dj7`80IdP/00@0N5@j085I@021FD00P5Po0P22
G4D0102:IE80Pf9C08=RD`1O@cP203LN5`04034I5@11;2/0@B`[03@/:@80?3Ld00@0>3Hg04Y;C`1:
Bdl092DY0P0Y:B/0100Y9BH0;bXW02lZ9`1:CU4202TZ;00403dn@00V9RH09RHV03De=@80BdY800@0
<C0^03@`;@0d<2d09RDS0P0i>3H0100j>CL0JfYX06]ZJ00`;bd2028Q7`0402/Z:00g=S@0=cHd02TX
9P80>3Le00@0CDa:034`;P0a<2h0:bXX0P0Z:BL0100i>3H0=C@b03Dd<P0T8b4202LV9004030_;@0_
;R`0;bh/02<R80809bHT00@0=SDc03/j>00k>SP0GEaJ0P0a<2h0100];2X0=3<a03@c<@0P7ad202@S
8@0403Pg=@0a<2h0<C0^051?C@80<C0^00@061LE01/J600K6QP0:2LU0P0^;B/0100a<2h0=C@b03Dd
<P15A4820491?`0403He<`0[:RP0:bXX03<b<080ADA200@0BDQ6040o?@10?cd0:2LU0P10?cd0100g
=S@0:BPV02TX9P0m?3X204Q7A@0403Xi=`0a<2h0<C0^038a;`80=3<a00@0>3Le04=2@013@T00B4M5
0P0g=S@00`0[:RP0<S4_038a;`0302LU9P0402HT9@0X9RL0:2HW02/Y:P80<2h_00@0=C<d03Pf=`0h
=SL08B0N0P0/:bT0100b<Bl0;Rd[02h]:`0a<2h203hm>`0404M6A018AdD0B4M503Lf=0809B@R00@0
92<Q034`;P0a<2h0=SDc0P0i>3H0100n?C/0?chl03ln?00f<302038/;00402lZ9`0^:BD0;RTU030[
9`80<B`V00@0;bXT02`W8@0/9b40=C0/0P0g<Rh0100m>3@0A3lk04@o>`17@Sh204<n>P0403Td<00`
:bL0<2/W03<^:P80;bXV00@0;RTU03Lb;P0g<Rh0A3lk0P19A4001014?c/0?3Lc03`g<`0]9R0202`U
7`0402dV800c;2H0<b`V03Xc;@<0?SLa0P0l=Bl01@0g<Rh0>3<_03Pc;`0i=300>SDa0040oooo000H
04QFH@0404MEH018EV40B5IQ04QFH@<0BEMR0`18EV4204MEH0`0AUAO0`18EV4404MEH140AUAO1017
EF0304QFH@D0BEMR3@18EV4204UGHPH0BUQS00<0BEMR04QFH@18EV403018EV401014E640BEUV04UI
IP1<FVD204YGH00404QBF`18DEP0B55H04I=D`80ADY>00@0;2`/01lO7`0O7al04Q0A0P0=2``0100C
3a005a<D01LC500F4182018<3P04014?400@4100410@00dA40<0318@0P0=4A00100@41004@l@014?
400H40d20106100401@60`0X5Q80:1HB044Y9@80D3@`00@0FCXe060o>01P?cP0J4Tg0P1]CS`0101g
F4@0PF5<085QC024HTT2081N@`0407mL@020GD40P5e108]RC080S6==00<0SFA>08]UCP2;IDh00`29
IDd308QVC@0408INA028H4H0R61608aRB@80SFA800@0TFE:09=WBP2CIdX0UfY=0P2JKDh0102ILU<0
TFY;095ZB`2HLE8208iWBP0308QNA@2CJE00TfU@00<0UF]C00@0Tf]B08iTC02>I4`0TVQ@0P2KKUL0
102IK5D0VfeF09]]EP2RM5d20:IhH@0409]aF`2RN680XWQR0:MmI`80YWaV00@0YG]U0:IlIP2VO6H0
YWaV0P2UNfD0102QKUT0YW=N0:IcGP2YMf020:]iHP040:ilH`2cPFH0/h5V0;:0I@80[G]P00@0]H=Z
0;>1J02cPFP0[geR0P2[NEh30:QgF@040:YiF@2/Ne/0[7]K0:9`E`80V6U?00@0VF]A09U]E02IKE@0
S69:0P24G4<01023GDH0P5a4081LA01mGD@207iNA@0407eMA01jFT40NUY107UI@080NUY100@0N5Po
07AD>`1dE3/0KT/g0P1SABd0101G>b<0DCLL054g701A=Q/205Xk7`0406U5:`1hD3L0N50g07eF>@80
O5Ll00@0R6A<091`F`2@L5/0N5];0P1<<RD0100f7Q@0>b@L03/T7012=c<2040h=@0402HR8`1:BT`0
BTY<034b=P80:BT[00@0:bLX02TS8`0Y8b<0<3@g0P0m?T00100U9RP09BDU02DU9@0h>3P203Lf=003
040o?@0R7Q/08QhK00@0=3<a0P1:BDL01017AT@09B@R02DT8P0b<Bl202TX9P0403<b<00];2X0;B`Z
05ACD@80>CPf00@0;Rd[02HU8`0V9B<0;2/Y0P0g=S@0100a<2h0:2LU02PW9@0[:RP202l^;00402LV
900T8b4092<Q03@c<@80A4=100@0@D0n02PW9@0X9bD0<c8`0P0k>SP0100R8Al092<Q02@S8@0n?C/2
0491?`0403/j>00U92809B@R01XI5`808R4O00@09RDS02<R800S8R00<2l]0P15A480101BDDl0@D0n
0450?P0^;B/202`[:@0403dl>P1:BDL0BTU703ln?080;2/Y00@0<c8`02DT8P0U9280=SDc0P1BDDl0
101>CD/0>cXh03/j>00d<c42038a;`0403Th=P16AD<0ATE304M6A080?3/i00@0=C@b034`;P0a<2h0
<C0^0P0g=S@0100Z:2T0:bTZ02/Y:P0]:b`202h/;@0302d[;00/:R/0;2X[00<0:bTZ00@08R4O02/Z
:00[:RP0;bh/0P0];2X0100c<S00@3lm040o?@16AD<204A3@@0403Pg=@0[:RP0:bXX02h]:`80?C`j
00@0?Sdk03Pg=@0h=cD0;bh/0P0T8b40100f<Bh0<RdZ038]:P0]:2D202XU8@0402`W8@0`:bD0<2/U
03Da:080>3@[00@0Ad8n04@o>`14?c/0?STe0P0g<Rh0100b;BT0<2/W030[9`0b;BT203D`;004030[
9`0f<Bd0=S4]03dh=080?STe00@0>C@`03D`;00e<2`0=2l[0P0e<2`0100k=2X0?3D[03`e:`0n=bd2
040i;`0403hg;@0i<RP0>C8X03</8P80;bPN00D0=c@[03Tf;@0i=Rd0?CXa040m=0010?ooo`006018
EV401017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ0P17EF0<04IDG`<0B5IQ1017EF0A04IDG`@0AeEP
0`18EV4504UGHPd0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00`0B5IQ00@0A5AQ04UIIP19FFH0
C5YU0P1:Ef001018DU/0B55H04QAF016CE<204E:CP0402`/;00O7al07alO018@4@803@/<00@04`l@
01LC500G4a@05Q0B0P0B30h0100A3a00410@010@400=4A0300`B40803A4@00@0410@014?400A3a00
610=0P0@1P@0100D1P<0:1HB02PF4P11:BD2050d<00405Tj=@1P?cP0H3lh06Q9=`80KDhl00@0MeQ4
085QC021HD`0Q6990P20GT<0101oG400P5e1081M@@2;HT`208aSC@0308eTCP2;IDh0RfE>00<0RFE=
0`28ITd01026GT@0R61608QPAP2<HTT208eTB004095UBP2CIdX0TfM:09MZC@80VVe>00@0VG9C095Z
B`2AJT/0V75B0P2>IdX00`28GTD0TfU@09=YD00309E[D`0409=[DP2>I4`0SVA<099XD080VfiG00@0
VFaE09]]EP2KKEH0XWAM0P2VN640102KLE/0XWQR0:9hHP2WOFL20:IlIP040:EkI@2VO6H0YWaV0:Il
IP80YG]U00@0XFiI0:IcGP2VLeh0ZGMP0P2[NF80102^O6<0/h5V0;>1IP2bP6D20:ekH0040;F3JP2c
PFP0/h5X0:mmHP80ZgUN0`2XMeT0102ZNET0[7]K0:akF`2RL5L209QYC`0409U[D@2IKE@0VFeD08aR
BP80Q5a300@0Pee6081LA020G4@0OEe40P1nGTD0101mGD@0NUY107YJ@@1iFD0207YJ@@0407QH?`1d
E3/0M5@k06i;=`80HdD]00@0Ec/S054g701A=a`0DCHK0P1J>al0101YAB/0N50g07Q@=`1mEST207aG
?00408QTC02@L5/0T71K07QKB`80C38U00@0=QhD03/T700k91`0@SLc0P10>3D0100V8R<0BTY<04Y:
C00a<SH202TY:`0402/W:00Y8b<0:B<S030d=`80?Ci000@09BHX02DU9@0U9BD0>3Ph0P0g=S@00`10
?cd08QhK028N6`0403@c<@80BTU700@0AdI402DT8P0U9280<S4_0P0Y:2H0100c<S00;B`Z02d/:P1D
De4203Th=P0402h]:`0V9B<09RDS02`[:@80=cHd00@0<C0^02PW9@0X9bD0:bXX0P0_;R`0100W9R@0
92<Q02@S8@0d<c4204A3@@040450?P0X9bD0:2LU03<b<080>cXh00@08R4O02@S8@0T8b40?Sdk0P12
@Cl0100k>SP09B@R02DT8P0J6AL2028Q7`0402HU8`0S8R008b8P030_;@80ADA200@0DU5?0450?P11
@3h0;Rd[0P0/:bT0100m?3X0BTU704Y9A`0o?S`202`[:@0403<b<00U92809B@R03He<`80DU5?00@0
CTe;03/j>00k>SP0=3<a0P0b<Bl0100i>3H0ATE304I5@`17AT@203`k>@0403Dd<P0a<2h0<C0^034`
;P80=cHd00@0:RPY02/Y:P0[:BX0;B//0P0^;2d00`0]:b`0;2X[02`Z:`0302/Y:P04028Q7`0[:RP0
:bXX02l^;080;B`Z00@0<c8`040o?@10?cd0ATE30P14@d40100h=cD0:bXX02/Z:00^;B/203dl>P04
03hm>`0h=cD0>3Le02l^;08092<Q00@0=S4^038]:P0b;BX0;BPU0P0Z9B40100/9b40<2/U030[9@0e
<BP203Pd:`0404M2?P14?c/0A3lk03hi=@80=c8^00@0<RdY030[9`0`:bL0<RdY0P0e<2`0100`:bL0
=S4]03Ha;@0m>3@203hi=@0403Td<00e<2`0=C0/03@_:`80=C0/00@0>c@Z03`e:`0l=B/0?SL]0P10
>Bl0100n=bd0>C8X03Tb:00c;28202lX7P0503Ld:`0i=Rd0>CH]03dj<@10?C@00@3oool001P0B5IQ
00@0AeEP04QFH@18EV40B5IQ0`19Ef8304QFHAL0AeEP3`16E5l404MEH0<0B5IQ1@19Ef8=04QFH@80
BEMR1P1:F6<00`19Ef80B5IQ04QFH@0<04QFH@0404IFH`19Ef80BEMR04YHH`80CEMP00@0C5EL04M>
E017CU@0?dA80P0g?400100N6QL05a<@01LC400@30T2010;200401H>300I4@l06A4?01T?3P805`d<
00@04Q0A014A4@0A4A403Q8A0P0=4a40100>4Q40418A010B4@0B414201D?4@0401d?3P0H2@H060T6
020=2@80>B4M00@0DC@^05/j<@1K>S40HC`c0P1U?cH0101]CSX0M5@o07AD?`1jFTD207mLAP0407eK
@P1mFCl0OETo081L@P80QF9600@0T6I>099XD02BJ500U6YB0P2CJE40102>ITd0RV9908YRB@27GdH2
08INA@0408YNA@2=HDP0SF58091TB@80TVA:00@0TF96095RA02AHT@0UVE70P2IJ4X0102JKU40TVI9
099VB@2HK4l208eTB00408IL@`2@J4h0T6Q>091XC`80SfM>00@0SVA<08]QB@2;HDT0TFA=0P2IK5D0
102LKUL0WW1I09i`F@2TM5h20:QhHP040:5dG@2YO6D0ZGaU0:]nI`80YGQQ00@0XGAM0:IiHP2VNF80
Z7]T0P2VNF800`2TM5h0YgMQ0:MgH@030:QhH@040:enI02cQ6P0/hAX0;62IP80ZgaP00@0[7eQ0:]l
H02[O600ZGYN0P2VMeT0102TMEL0YGMF0:EgEP2YNeX20:anG0040:9eF02JKE00VVe@09e_E@80WW9G
00@0TfU@08]SBP2;HdX0RV990P25H4H3089PA@<0Pf5600@0Q69708=QAP23HDH0Oee20P1kFCh0101a
CSX0IDL_06E7;`1M@BT205To900405@i7P1A<QH0DC8F05Xf7080IclV00@0N5@d07YE>01jECP0PEi2
0P26ITd01020HD`0IT/h06I;>01:<24203TO400404Dh<00b9b<0<RLS02XU8P80<2ha00@0CDaA02<S
9@0S8bD0:RPY0P0^:RT0100X;2l0BDY<04U:C00X:B/202h^;P0403Ph>00_;R`0;bh/04U8AP80=C4^
00@0?chl02LV900W9R@0=3<a0P0Q81h0101BDDl0ADA204E4@P0/:bT202PW9@0403hm>`0g=S@0=cHd
03hm>`80D4m=00@0:2LU02l^;00_;R`07A`J0P0_;R`0100o?S`0:bXX02/Z:00R8Al202/Z:00403<b
<00h=cD0>3Le040o?@80BTU700@0;Rd[030_;@0`;bd0:RTW0P0[:RP0100l>cT0ADA204E4@P0k>SP2
02d/:P03020O7@0O7Q`07ahL00<07A`J00@08b8P030_;@0`;bd0@D0n0P1<BdT0100n?C/0;Rd[02h]
:`0];2X203ln?00404I5@`0j>CL0>STg030_;@80<C0^00@0;2/Y02h]:`0^;B/0B4M50P1<BdT0100f
=C<0?Sdk03hm>`1;BTP203Pg=@<0?3/i00@0=C@b02/Z:00[:RP0:2LU0P0];2X0100d<c40>STg03Xi
=`0n?3d203/i>P0403Dc=00]:b`0;B//02HT9@<08b4R0P0U8b@0100j>CL0;B`Z02d/:P0Z:BL203Lf
=0040450?P0l>cT0?3/i03<b<080;Rd[00@0<c8`030_;@0`;bd0>cXh0P15A480100m?3X0<c8`03<b
<00/:bT202DT8P0402XU8P0`:bP0<2/X03Pc;`<0?CPb0P0i=B`0100d<2L0<BdT034]900o>38203/d
;P0403D^:00_:280;bPR02dV8080<R/U00@0>S<]040i<`10>C<0@3/e0P16@C/01019A3h0@Sdg048m
=`0f<B/202lZ9004034/9P0g<R`0=c8/03Pd:0<0>CDY0P0h=2P00`0g<bL0=C4U03Da9@0303Da9@05
03hk<P0m>S40?CXa03dj<@0l>C000@3oool001P0B5IQ00@0AeEP04QFH@18EV40B5IQ0`19Ef8304QF
H@h0AeEP0`16E5l904MEH0`0AUAO1017EF0304QFH@D0BEMR3@18EV4204UGHPH0BUQS00<0BEMR04QF
H@18EV403018EV40101:F6<0AUAO04IDG`16De`204aFG`0404eFG@15BTh0ADY>038g>`8092P[00@0
514<01D@300E40`05A0<0P0F3P/0100E3@X05P`:01H<2P0H3P`201/@3P03014A4@0@4Q40418A00@0
3Q8A0P0B4140100E3a4060dA01P=4@0M2`T2024>2P04038J5P19;RL0BBhW05Xi<080HCda00@0Id0a
06m5=`1_ACL0LE4j0P1iFD801022H4L0Pf5808=QB021GDD2085M@`0408IQA`2:ID/0RVE;08]QB080
T6I=0`2DJU40102?ID`0Rem608]OAP26FT4208AH?`0408YN@`2?HDL0Sf5709AVC080UfQ<00@0UVM9
09MVB02GITP0V6M70P2KJ4T0102IJTh0TFA7095TA`2IK4l2095UBP0408YPA`2CJe40Tf]A08mWCP80
S6A;00@0S5m808]NA`2;GTL0SV5:0P2GJE80102NL5T0WfmI09m_F@2QLE/20:UfH@040:=eG@2YNf<0
ZG]S0:UkH`80XWAL00@0XG=K0:QjHP2XNV80[X1X0P2]OfL0102ZNV@0Zg]U0:]kI@2ZNV<20:UiHP04
0:amH`2aPVH0/H9V0;21I@80ZW]O00<0Y7EI0:MhG02WN5`00`2YNU`0102WNEP0Z7YH0:QjF02ZO5X2
0:eoG@040:EiG02LL5<0W71C09ibE@80XWIK00@0V6mC099XC`2BJ4l0TFM>0P2<I4/308AQA@0408MT
B02;J4`0RfQ<08eZCP80RFI:00@0PUm307eJ?P1mFSh0LDhj0P1YBc<0101VBS80ID/`06E;<01M@RL2
058c5`04054]4`1I<AP0FC4H06a8:080MU<c00@0N5Di07ME>P1gECX0Oem80P22Hdh0101SAc80>ah<
03/N300f9A/2034T700403<[:00L61T071PI04Q8BP80=cLi00@0:2HW03Da<P0e<C80<cLj0P0^;c40
1011@T@08b<S02<S8`0N7Qh203Lf=00402h]:`1MFEH0GEUF03<b<080;Rd[00@0:bXX03<b<00c<S00
:bXX0P1FEE<0100d<c40=cHd03Lf=00k>SP203`k>@04034`;P1AD4h0DE1>03<b<080:2LU00@092<Q
02XY9`0Z:BL0@3lm0P0`;bd0100Q81h09B@R02DT8P0f=C<20491?`0404A3@@13@T00@d9002/Z:080
>CPf00<0>3Le03@c<@0d<c400`0n?C/0100^;B/08R4O028Q7`0^;B/202@S8@0401TH5P0J6AL06QTG
02d/:P<0@T4o0P0f=C<0100^;B/0;bh/02l^;00k>SP204U8AP040450?P0];2X0;B`Z02`[:@80>cXh
00@0@D0n05EDDP1EE580@T4o0P0Y:2H0100_;R`0;Rd[02h]:`0/:bT203ln?00405]JF013@T00@d90
03Dd<P80=SDc00@0<C0^02LV900W9R@0;2/Y0P0i>3H0101>C4d0ATA504I4A@0g=CH202LU9P0401dK
700N71d07Q`M02HT9@80;R`]00@0<C0^02d/:P0];2X0<c8`0P10?cd01014@d40>CPf03Th=P0_;R`2
02`[:@0402d/:P0c<S00<c8`0491?`80ADA200@0=C@b02h]:`0^;B/0=SDc0P0l>cT0100e<2`0>3<_
03Pc;`0k=S0203`g<@0403Xf;@0e<BP0=C4X034]8P80;RXO00@09AhH02/T7P0[91h0=2dW0P0m=S00
1012>cD0@CXd044j=00l=Bl203L`:P04034/9P0[9R00:bHP02PS7@80;bXT00@0>SD_040k=@10>cD0
?CPb0P0f<B/01011?c<0@3hb040n<P0o?C4203dk;`<0>cT]00@0?C/_03hl<00n?300?Cdc0P0l?381
03Xj<040>CT_0@3oool001P0B5IQ00@0AeEP04QFH@18EV40B5IQ0`19Ef8304QFH@h0AeEP0`16E5l9
04MEH0`0AUAO1017EF0304QFH@D0BEMR3@18EV4204UGHPH0BUQS00<0BEMR04QFH@18EV403018EV40
101:F6<0AUAO04IDG`16De`204aFG`0404eFG@15BTh0ADY>038g>`8092P[00@0514<01D@300E40`0
5A0<0P0F3P/0100E3@X05P`:01H<2P0H3P`201/@3P03014A4@0@4Q40418A00@03Q8A0P0B4140100E
3a4060dA01P=4@0M2`T2024>2P04038J5P19;RL0BBhW05Xi<080HCda00@0Id0a06m5=`1_ACL0LE4j
0P1iFD801022H4L0Pf5808=QB021GDD2085M@`0408IQA`2:ID/0RVE;08]QB080T6I=0`2DJU40102?
ID`0Rem608]OAP26FT4208AH?`0408YN@`2?HDL0Sf5709AVC080UfQ<00@0UVM909MVB02GITP0V6M7
0P2KJ4T0102IJTh0TFA7095TA`2IK4l2095UBP0408YPA`2CJe40Tf]A08mWCP80S6A;00@0S5m808]N
A`2;GTL0SV5:0P2GJE80102NL5T0WfmI09m_F@2QLE/20:UfH@040:=eG@2YNf<0ZG]S0:UkH`80XWAL
00@0XG=K0:QjHP2XNV80[X1X0P2]OfL0102ZNV@0Zg]U0:]kI@2ZNV<20:UiHP040:amH`2aPVH0/H9V
0;21I@80ZW]O00<0Y7EI0:MhG02WN5`00`2YNU`0102WNEP0Z7YH0:QjF02ZO5X20:eoG@040:EiG02L
L5<0W71C09ibE@80XWIK00@0V6mC099XC`2BJ4l0TFM>0P2<I4/308AQA@0408MTB02;J4`0RfQ<08eZ
CP80RFI:00@0PUm307eJ?P1mFSh0LDhj0P1YBc<0101VBS80ID/`06E;<01M@RL2058c5`04054]4`1I
<AP0FC4H06a8:080MU<c00@0N5Di07ME>P1gECX0Oem80P22Hdh0101SAc80>ah<03/N300f9A/2034T
700403<[:00L61T071PI04Q8BP80=cLi00@0:2HW03Da<P0e<C80<cLj0P0^;c401011@T@08b<S02<S
8`0N7Qh203Lf=00402h]:`1MFEH0GEUF03<b<080;Rd[00@0:bXX03<b<00c<S00:bXX0P1FEE<0100d
<c40=cHd03Lf=00k>SP203`k>@04034`;P1AD4h0DE1>03<b<080:2LU00@092<Q02XY9`0Z:BL0@3lm
0P0`;bd0100Q81h09B@R02DT8P0f=C<20491?`0404A3@@13@T00@d9002/Z:080>CPf00<0>3Le03@c
<@0d<c400`0n?C/0100^;B/08R4O028Q7`0^;B/202@S8@0401TH5P0J6AL06QTG02d/:P<0@T4o0P0f
=C<0100^;B/0;bh/02l^;00k>SP204U8AP040450?P0];2X0;B`Z02`[:@80>cXh00@0@D0n05EDDP1E
E580@T4o0P0Y:2H0100_;R`0;Rd[02h]:`0/:bT203ln?00405]JF013@T00@d9003Dd<P80=SDc00@0
<C0^02LV900W9R@0;2/Y0P0i>3H0101>C4d0ATA504I4A@0g=CH202LU9P0401dK700N71d07Q`M02HT
9@80;R`]00@0<C0^02d/:P0];2X0<c8`0P10?cd01014@d40>CPf03Th=P0_;R`202`[:@0402d/:P0c
<S00<c8`0491?`80ADA200@0=C@b02h]:`0^;B/0=SDc0P0l>cT0100e<2`0>3<_03Pc;`0k=S0203`g
<@0403Xf;@0e<BP0=C4X034]8P80;RXO00@09AhH02/T7P0[91h0=2dW0P0m=S001012>cD0@CXd044j
=00l=Bl203L`:P04034/9P0[9R00:bHP02PS7@80;bXT00@0>SD_040k=@10>cD0?CPb0P0f<B/01011
?c<0@3hb040n<P0o?C4203dk;`<0>cT]00@0?C/_03hl<00n?300?Cdc0P0l?38103Xj<040>CT_0@3o
ool001P0B5IQ00@0AeEP04QFH@18EV40B5IQ0`19Ef8304QFH@h0AeEP1016E5l504MEH0<0B5IQ3016
E5l404MEH0<0B5IQ1@19Ef8=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0<04QFH@0404]II016
De`0AU=L04I@F@80BU=J00@0B4mE03Xo@`0j?d<08bLZ0P0C5aX0100A3@X05A4>01DA3P0G4a0201HA
3P0401D=2`0C2`T04`/901L=30806a4@00<0410@00lA400?4A000`0=4A00100?4A004`l@01<?400G
310201P<400401`91@0/5a80;1LB048[9@80E3P]00@0GCd`06I1<@1V@C40KdTf0P1hCc/0101cDCH0
Oee207mM@P2;JDh208m[D@0408iZD02>JDl0SVU?08iYC`80SfM>00@0Pei308UTB@29I4T0TFU?0P2B
JU000`2?ID`0RV1708YPA`0308UMA00408]OA02?HdH0Sf=609EYC080VVe>00@0WFi@09i`C`2NL4l0
XG1@0P2RLE40102NK540UfE:09MUBP2MKU8209IXCP04091TB`2FKU@0UViD095YD080RVE;00@0RUe6
08aOB02<GdP0SV190P2FJ540102QLE/0XVmJ0:9_FP2PKEP20:IcGP<0/X1Y00@0[WaU0:QfG`2XMUl0
Z7IO0P2/NV<30:mmIP040:YlI@2/OVL0[7iW0:]mI@80Z7YR00@0ZG]Q0:j1I02^PF@0[X5T0P2ZOF00
102UN5T0ZGaM0:UlG@2^PF820:n2H@040:f1GP2[Oe`0ZgmL0:f1G080/8AO00@0YgiN09edE02MM5@0
WWEG0P2QN5X0102KLUH0UVeA09I]D@2FK5<2099XC`0408YWB`28IDT0R6E908UVBP80T6eA00@0TVmC
08eZCP2=JTh0QF960P21GT80101fDcl0LeDm07=E?@1aECd206iD>@0406I;<01L?B40G3dQ05Hb6080
ERhE00@0FcPJ06]8;01[B2`0LDhb0P1_CC@0101mGDH0SFeH08e]F01gET<204d^6P0402PD2@0n;BD0
?RdU030W8P808alN00@0:RX/05EEE`1EEEL0:bT/0P0d<S<0100V:Rd09RLY02HW:@0g>3X202h^;P04
01hN7P0X9bD0:2LU02PW9@80E51=00@0=cHd02d/:P0];2X09bHT0P0i>3H0100V9B<0?C`j03dl>P1N
GE/203@c<@03030_;@0a<2h0<C0^00<0>3Le00@0BTU701`K6@0L6aT0=cHd0P0^;B/303Dd<P0402XY
9`0T8b4092<Q02h]:`<0=SDc0P0e=380100^;B/0<S4_038a;`19B4H204Y9A`0402XY9`0N7A/07QdK
02d/:P80=3<a00@0;2/Y028Q7`0R8Al092<Q0P0e=3801012@Cl0?C`j03dl>P0`;bd202TX9P0403@c
<@11@3h0@D0n04M6A080?C`j0`0b<Bl0100h=cD0?C`j03dl>P1BDDl203`k>@0402HU8`0[:RP0:bXX
03Lf=080<S4_00<0>CPf05ACD@1DDe40100l>cT203He<`0403@c<@10?cd0@3lm04M6A080@3lm00@0
:BLX02PV9`0X9RL09R@U0P0R8240100P7Ql09B<T02DS900^;2d203Dc=00402TX9P0o?S`0?chl04e<
BP80@d9000@0=3<a02d/:P0];2X0;B`Z0P0/:bT0100g=S@0>cXh03/j>013@T0203hm>`0402`[:@0[
:RP0:bXX03Xi=`80A4=100@0B4<o040k=@10>cD0=2lY0P0Z9Qd0100W8aX0:bLL02/W700`;24203@`
900403Xa:P0k<R/0>c8[03dd;@<0?cH_0P0k<R/00`0e;2D0<BPQ034X8@0302dY8004034]900j=Rd0
>SH]04<o=P80AT8i00@0@clf03hj<@0n>S40<BhW0P0b;bP0100d<BX0>3D^03Pe;P0k>34203`i<P04
03/h<@0j=c00>SL`03Pg<P80>STd0@0m?3L103hm>040oooo000H04QFH@0404MEH018EV40B5IQ04QF
H@<0BEMR0`18EV4204MEH0`0B5IQ00<0AE=N04IDG`16E5l00`16E5l00`17EF00B5IQ04QFH@0304QF
HA00AeEP0`18EV4504UGHPd0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00`0B5IQ00@0BeQQ04MD
G@17E5d0Ae5J0P16CeH0100k@TP0:Bd`02T]<00E6A`200h?4@03010@400A4A404A4A00<04Q0A00<0
4`l@01<=3`0C3@l00`0D3Q000`0?3`l03Q0?00h@3`0300`@3`0400l?3`0C3@l04`d?01L;3`806PT?
00@0918>03/X8P0k:280DS/c0P1J@3<0101O@2h0JdPd06]8=01fD3T207]C>P0407AB=P20GT80P5i2
08eZCP80T6aB00@0TfiD09E`EP2EL5H0UVeG0P2CJU@01026IDH0RfY;08]ZB`2@KDl2091[CP0408aW
BP2;I4L0RfA708aSA`80SfI:00@0UFi?09A]CP2DKDh0UFa>0P2FKDd0102IKT`0Vfi=09]^C@2JKT/2
09U]BP040:I`EP2KJ4d0VfQ=09m]DP80TfE;00@0SF5809=[D@2CJe40S6M=0P27HTP0102=GdP0T69;
091RB`2@HT/209QXDP040:AdGP2SL5/0Xg1K09i[EP80YW9M00<0^HM^0;:0I`2bP6L00`2]Nf80102^
O6<0ZgUP0:]iH02XMUd20:MeG0040:QjH`2/OVL0[7iW0:anIP80Z7YR00@0YgUO0:aoHP2/Of80[h9U
0P2]P6<0102[OUl0[h9S0:n2H`2cQVD20;:5I0040:j2G`2[OeX0ZgmJ0:f1G080/8AO00@0ZX5Q09mf
EP2OMUH0WWEE0P2QN5X0102NLUD0VfmD09]_E02LL5D209Q/D`0409E`D`2?JTd0SfY=08eXB`80TVe@
00@0UG1C095/C`2AK4l0RfI90P29I4L01024HDd0P69:081RBP1jGTH2079H?@0406aA=P1YBRh0JDX^
06A09P80H3PO00@0DblE05lk8@1O>b40JTH^0P1aCSP0101mFTH0R6EA08QUD@22Gd`2079A?P0403LQ
4`0i9ad0>BLM02XO6@80=C0]00@0:bT/04a<CP1<C4h0>SXl0P0`;S40100N8RD0=cPj03Lh>P0X:B/2
04a<C00403`l?00U92809B@R03/j>080=S8_00@0B4M502d/:P0];2X0;2/Y0P0U9280100j>CL0<2l]
030_;@1PGed203Th=P0402`[:@0Z:BL0:RTW03Dd<P809bHT00@0C4]902d/:P0];2X0=SDc0P0_;R`0
100[:RP0=3<a03@c<@0a<2h202PW9@0402XY9`0Y:2H0:BPV02HU8`80:bXX00@0?chl03<b<00c<S00
>3Le0P0k>SP0100^;B/0;2/Y02`[:@0a<2h202/Z:0<07ahL00@0;Rd[04E4@P15A480ATE30P0c<S00
100[:RP0<S4_038a;`0h=cD203ln?00403/j>00^;B/0;Rd[02h]:`80?C`j0`17AT@01012@Cl0>CPf
03Th=P0g=S@2030_;@0402h]:`1:BDL0BTU7055@CP80<S4_00<092<Q02d/:P0];2X00`0h=cD0100f
=C<0@3lm040o?@0m?3X202PW9@0402DS900X9RL0:2HW02XX:@80:2HW00@08R0Q01hL7@0N71d07Q`M
0P0Q7b00100i>3H0ATE304I5@`1:BDL203/j>00402d/:P0/:bT0;2/Y02l^;080<2l]00@0?chl03Lf
=00g=S@0=C@b0P0c<S00100];2X0=3<a03@c<@11@3h204=2@00403/f<00g<R`0=c8/038^9@80<BdT
00@0<blT03Da9P0e<BH0=30T0P0b;R80101:@CX0ASdf04Hm=P0m=2d2038Y8P<0:R4J00@0<BPQ03P_
:00h;bP0=c<Z0P0o>c801015@CP0@Cdd044m=00f<RT202hZ8@0402l[8P0d<2L0=30W040o>`80?Sdi
00@0?C`h03hm>@0n?CT0?chj0P0m?3P0100g=S80<S4]038a;@0f=SH203Xj>P40?clo0@12@T810?oo
o`006018EV401017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ0P17EF0<04QFH@0304ECGP16E5l0AUAO
00<0AUAO00<0AeEP04QFH@18EV400`18EV4@04MEH0<0B5IQ1@19Ef8=04QFH@80BEMR1P1:F6<00`19
Ef80B5IQ04QFH@0<04QFH@0404]HH@17E5d0AeAM04MAFP80ATmF00@0>d9802T]<00Y;C005ATL0P0>
3a400`0@41004A4A014A4@03018@4@0301<?400C3@l04`d?00<050h@00<03`l?00h@3`0>40l00`0<
40l0100?3`l04`d?01<=3`0G2`l201X93`0402@B3P0k:280>bPR058k<`80FT0c00@0Gd0^06]8=01[
B3@0MU0i0P1kDcX0101dDSH0P5i2081N@P2=JTh2091/DP0409=^E02EL5H0UG1F09I]E`80TfYD00@0
QVE608]ZB`2;JT/0T6e?0P2@Jdh0102<IdX0RfA708]TA`2<HdL208mVBP0409E^C`2DKDh0U6e>09E/
CP80UVe=00@0VFi<09]^C@2KKTd0VVi;0P2IKDX0102VL5H0VfQ=09]XC@2OKE8209=UB`0408eQB02C
Je40Tf]A08aWC@80Qf9800@0SEm8091RB`2@HT/0T69;0P2HJ580102TM5h0Xg1K0:=`F`2NJeH20:Ib
G@030;V7KP2bP6L0/X1W00<0[G]R00@0[WaS0:]iH02[NF00Z7IM0P2WME`0102XNV<0[7iW0:anI`2/
OVH20:QjHP040:MiG`2/Of80[7mR0:n2I@80[H1S00@0ZgiO0:n2H`2_PV<0/hIU0P2bQF@0102^PUl0
ZgmJ0:]oFP2]PE`20;24G`040:Z1H@2OMUH0WgIF09ieE@80XGQJ00@0WW9E09]_E02KKe@0W71E0P2H
K5<0102EL5<0SfY=08mZC@2=J4/2099]D00409E`D`2AK4l0TFa?08]VB@80RFA700@0Q65=081RBP20
HTX0NUi60P1bF3d0101/DCH0JDX^06U:;P1T@2H2060h7`0405<_5@1O>b40Gc/Q06Y6;P80LDhh00@0
OEY608QUD@28IE40PUm<0P1bDCh0100g8A<0>BLM03TW7@0Z7aT203D`;@0402/Y;01<C4h0C4a>03Xj
?080<2ha00@07R8U03Lh>P0g>3X0:2T[0P1<C4`0100l?3`09B@R02DT8P0k>SP203Hb;`0404Q7A@0]
;2X0;B`Z02`[:@809B@R00@0>STg030_;@0`;bd0H5mM0P0i>3H0100/:bT0:RTW02XY9`0e=38202LV
900404a;B@0];2X0;B`Z03He<`80;bh/00@0:bXX03@c<@0d<c40<C0^0P0X9bD0100Z:BL0:BPV02TX
9P0V9B<202/Z:00403ln?00c<S00<c8`03Pg=@80>cXh00@0;Rd[02`[:@0/:bT0<C0^0P0[:RP301lN
700402h]:`15A480ADA204I5@`80<c8`00@0:bXX038a;`0b<Bl0>3Le0P0o?S`0100k>SP0;Rd[02h]
:`0^;B/203dl>P<0AdI400@0@T4o03Th=P0i>3H0=cHd0P0`;bd0100^;B/0BTU704Y9A`1AD4h2038a
;`0302@S8@0];2X0;B`Z00<0>3Le00@0=SDc040o?@10?cd0?C`j0P0X9bD0100U8b@0:2HW02PV9`0Z
:2T202PV9`04028P8@0N71d07Q`M01hL7@808AlP00@0>CPf04I5@`16AD<0BTU70P0k>SP0100];2X0
;2/Y02`[:@0_;R`2030_;@0403ln?00g=S@0=cHd03Dd<P80<c8`00@0;B`Z03@c<@0d<c40@D0n0P13
@T00100k=S00=c8/03Lb;00b;RD2034]900403<_900e<BH0=C4V03@`9080<RhR00@0BT4j04Hm=P16
?CH0?C@]0P0b:B8302XQ6P04034X8@0h;bP0>2lX03Lc:P80?c/b00@0AD4h044m=011?C@0=S8Y0P0^
:R40100_:b80=30W03@`9`10?c/203hm>@0403dl>00n?CT0?Sdi03ln>P80?C`h00@0=cHb038a;@0b
<Bd0=SHf0P0j>SX103lo?`40@T920@3oool001P0B5IQ00@0AeEP04QFH@18EV40B5IQ0`19Ef8304QF
H@80AeEP3018EV4304ECGP<0AUAO00<0AeEP04QFH@18EV403`18EV4404MEH0<0B5IQ1@19Ef8=04QF
H@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0<04QFH@0404]HH@19EUl0BEIO04YDG@80A4eD00@0<cPl
01/O8P0K7b803a0B0P0=3Q00100A4Q@03PlA00h?4@0<30h200d=3`04018@4`0E41@05A0D01<>4P80
40/?0`0=3`h300/?3P0400h>3P0B30h04P`>01H:3P806@P>00@0<AlK04Th<019>300FdLl0P1NA3L0
101O@S00K4`e06a<=@1fDSP207E@<`0407MF=`1oGSl0Oeho08MTB080RFA900@0S6M=09E]E02EKE@0
VG1J0P2GKUP0102>LE40T7=C091cD`2ALU<208m^C`0408YWB@2:HdH0RV=608iWBP80TfY>00@0Sf];
08aXB02<J4P0RfA30P2=ITD00`2CJdP0VFi<09U^C00309]`C@040:QbF02LIT`0W6I<09]YCP80SUm5
00@0QEU008eUB`2=ID/0Qf980P21GD<0102@HT/0TfE>09=UCP2BI4d209YZE0040:UfH@2ULUd0YG9M
09m[EP80YW9M00@0[g]S0:McF`2WLe/0YG5I0P2^NV80102dP6P0[g]S0:mkH`2ZMUh20:UeG@040:Eg
H02ZO6D0ZWaU0:anIP80YgUQ00@0YWQN0:]nH@2[OV40/8=V0P2aQ6L0102_PV<0/XET0;:5I02dQfH2
0;65HP040:]oG02XO5L0Z7aG0:YnF@80[H5L00@0ZhEQ09miE@2ONED0WWEE0P2QN5X0102NLUD0W71E
09a`E@2OLEL209]]D`0409iiG02DKe80U6mB08iYC080TVe@00@0UG1C099]D02BKE00SfY=0P2@Jdh0
102?K5P0RfeE08]]E@1oHd/2079H?@0406iC>01cE3P0Le@h075=<`80Jd<Z00@0EC4I05Lc7@1G<ad0
I40Z0P1cD3`0101jEd<0NeQ507]HA@26HU8209EaH@0405<l;P0T4PH0918602XM5`80?STf00@0AD=6
02XZ;00Z:R`0Bd]=0P0/:Rd0100T:2/09bPZ02LX:P0W:2X204=3@`0403Lg=`0c<S00<c8`02d/:P80
8AdJ00@0B4M503ln?00o?S`0;B`Z0P0Y:2H0100];2X0ADA204E4@P0_;R`205MFE004030_;@0^;B/0
;Rd[02HU8`80:2LU00@0?3/i04m>C01?CT`09RDS0P0/:bT0100W9R@0<S4_038a;`0a<2h202`[:@04
02d/:P0V9B<09RDS024P7P809bHT00@0=cHd04i=B`1>CD/0A4=10P0^;B/00`0_;R`0;Rd[02h]:`03
02@S8@0402h]:`0V9B<09RDS02HU8`80<C0^00@0>3Le03Dd<P0e=380=3<a0P0g=S@01013@T00>STg
03Xi=`0`;bd202`[:@04030_;@0f=C<0=SDc03hm>`80@d9000@0;B`Z03He<`0f=C<0;Rd[0P0f=C<0
101=C4X0?C`j03dl>P0T8b4202d/:P0404=2@00S8R008b8P034`;P<0H5mM0P0k>SP0100[:RP0<C0^
034`;P0X9RL302TW:0809R@U00@09B<T02XX:@0Z:2T0=C<d0P0n?3d01014@d40<S4_038a;`0W9R@2
02h]:`0403Pg=@0l>cT0?3/i03ln?080ADA200<0BDQ603He<`0f=C<00`0[:RP0100`;bd0?3/i03`k
>@10?cd203Lf=00403Ha:`0a;2H0<B`V02dY8080<2`S00@0>CDZ040l<010?300@clc0P12?S80100/
8QT0<bTP03<Y800g;B@203<Y800402`R6@0[8AP0:b4H03D[8P80?cD/00@0>cL/03`h;@0l>2d0?3P]
0`0j=R/203`h;@04048n<`16@SL0AT8g03lm?P80>cTj00@0>3Hg03Xh>@0j>3T0@Cm00P14@T<01013
@D80@3ho040n?`0j>S`203`l?P40@4120@12@T@10?ooo`001016E5l504MEH0<0B5IQ101:F6<504UG
HP<0B5IQ3017EF0404UGHP<0B5IQ1@17EF0L04QFH@80BEMR0`1:F6<604]II0<0BUQS00<0BEMR04QF
H@18EV400`18EV401016E5l0AeEP04MEH019Ef8304YHH`80BEMR00@0AeEP04IDG`16E5l0B5IQ0P13
DE`01012D5/0BEMR04UGHP1<FVD204YHH`0404UGHP1<FVD0C5YU04UJH@80CEaS00@0BEIL03U2A`0i
@TL08BHY0P0?4a@00`0?3`l05ADE01DE5@0301H>2`0401H>300D3``050l<01<?3P804`l@0`0A3a00
100D3A@04@`B014<4P0?3A02014?4P04010<3@0A1`P04@L801`<3`80:A@I00@0CSX_050j;@1@>Rd0
EC//0P1M@B`0101/BS40Me8e07MB=@1nECD2089G=@04089O@@1mFS`0OEXl08ERA080SFY<00@0R6E7
08UVB029ITP0TVmA0P2FLeD0102MLeX0VW5E09YaE@2GKU8309M^D080UVe=00@0TVU908iV@`2>IT<0
PUTk0P2BJD/0102FJTd0T6E5091UA@2EJ4L209IYB00409MYA`2MKdd0WFm=09a^C080XWAC00@0WW5@
095UB02AIDP0SVE90P2AK540102=J4h0Oe]107mK@@2;G482091QA`0409UWCP2MJe80WF]B0:1/E080
Y6iF00@0Yg5I0:UcF`2YLe/0YW1H0P2SKED0102XLUX0/G]S0;5kH`2bO6@20:YdG0040:E_E`2VL5P0
YW1H0:MnH080YGaN00@0Y7]M0:=jG02SNU`0Y7]M0P2WOV00102ZPF<0[8=U0:b3I@2YPV<30;29JP80
YGiO00@0WGIG09igF02NMeP0Y7eN0P2VOf030:b3H`040:QoG`2RNET0XWUI0:AkF`80ZH1P00@0XgYJ
09U`D02IL500XW]N0P2ILUD0102@Jdh0TFa?095/C`2HMEL209]hFP0409AcE02=K4d0SFa=091YC080
TFU?00@0SVU>08UTBP29I4X0P5a20P1hE3`0101cDCP0LDlf075?=P1V@RX205Lc6`0405Xf7P1YAbh0
JDL^079@=`80OEe400@0QfM>08ATB`24I4/0M5U60P14:QT0100f8180@c8X04<b:00i;RP202lZ9`04
03Dc=00j>S`0>SXl02XU:@809R@Y00@0=C@j02PX<00X:300?d150P0e<cH0100T6a`0>S0_03X`;`0c
;Bd2044k>`0403lk>P0e<C00=C4`03dl>P80<C4_00@08B4O0410?P10@3h0BDQ60P0R8Al0100U9280
=C@b03Dd<P0b<Bl204M6A004038a;`0/:bT0;2/Y02LV9080;2/Y00@0>CPf03Lf=00g=S@09RDS0P0R
8Al302XY9`0404E4@P10?cd0@3lm03ln?080?Sdk00@0<c8`02<R800S8R008R4O0P0[:RP01011@3h0
:bXX02/Z:00P7ad2028Q7`0403@c<@0f=C<0=SDc03/j>080>STg00@0BTU703He<`0f=C<0;bh/0P0e
=380100k>SP0@T4o0491?`0i>3H2024P7P0402l^;00Y:2H0:BPV03Xi=`80De9@00@0CDa:030_;@0`
;bd0;bh/0P15A480100J6AL0?C`j03dl>P1<BdT203Th=P0402h]:`0k>SP0>cXh0450?P80>3Le00@0
9B@R024P7P0Q81h081lM0P0V9B<0100c<S00?chl03ln?012@Cl2040o?@0402HU800];2L0;B`W02h]
:080:RXR00@0<30X0410=P10@3H0Bd]10P1<C400100_;RT0;B`W02d/9`0_;RT203Th<`0404=2?@10
?cX0@3lj030_:P808B0K00@07ahI02d/9`0];2L0?C`g0P12@C`0100k>SD0<C0[034`:`0/:bH202/Z
9@0402dY7P0^:Ql0;RXO03<_9080>SH[0`10?340100l>2d0>3@Y03Pd:@0f<RT203di<00404I1>`1:
ACl0BTDo04I1?@80?cXg00@0>3<`03D`;@0e<2d0>SXh0P0m?Cd0100n?T00>cXo03/j?`0f=C/203He
?@0403dkAP15@e00AD=@04e=E`80BTYD0@16AU0104=3C@40oooo000404IDG`D0AeEP0`18EV4404YH
H`D0BEMR0`18EV4<04MEH0@0BEMR0`18EV4504MEH1`0B5IQ0P19Ef8304YHH`H0BeUT0`1:F6<00`19
Ef80B5IQ04QFH@0304QFH@0404IDG`17EF00AeEP04UGHP<0BUQS0P19Ef801017EF00AUAO04IDG`18
EV4204=AG004049@F`19Ef80BEMR04aJI@80BUQS00@0BEMR04aJI@1<FVD0BEYQ0P1=G6<01019EU`0
>D9703U2A`0Q9RT200lC500300l?3`0E5AD05ADE00<05Ph;00@05Ph<01@?300D3``04`l>0P0C3a03
014?400401@=500A31804@`B00l=40804@lB00@040`=0147200A1`P070`?0P0Y51T0101>>Rl0D3X]
050j;@1E>b`205e1;00406a:<@1gDSD0Me8e07iE=@80PULe00@0PUm107eJ?01mFS`0QF940P2=JT`0
1028IDL0RFI808UVB02BKe4209IcE@0409ecFP2JLED0VW5E09M^DP<0Ufi@0P2FKDd0102BJDT0SVI3
08iV@`22FC/2099YB`0409IZC@2@IDD0T6E509EXA`80UVU800@0UfU709e_C@2MKdd0W6i<0P2RM5<0
102NLE00TFE8095UB02>IDT2095/D@0408eXCP1oFd40Oe]108]L@P80T65700@0VFM>09e[DP2MJe80
X6aD0P2TKUH0102WLET0ZG=K0:UcF`2VL5P20:=]E@040:QbFP2aNf<0/G]S0;9lI080ZWAL00@0YFmG
0:I`F02VL5P0YgiP0P2UO5h0102TNed0XgYL0:=jG02TNed20:MnH0040:Z1H`2/PfD0[8=U0:V2H`<0
/8UZ0P2UOUl0102MMUL0WWMH09igF02TOEh20:IoH0<0[8=S00@0Z7mO0:9iF@2RNET0Y7]K0P2YP600
102SNUX0VG1@09U`D02RNeh209UbE@04091[CP2AK4l0TFa?09QeE`80VgQJ00@0U7=D08e/C@2=K4d0
T6U<0P2AJDl0102>JDh0RFA:08UTBP20G48207QD?00407=A>01aCcH0LDlf06I2:P80Ec<K00@0FSHN
06U7;P1YAbh0LU0g0P1mGD@01027Idh0Q6A;08ATB`1dFDH204@Z6@0403HP4P13<RP0@c8X03T^:080
;bXW00@0=C<d03Xj?00j>S`0:RDY0P0V92T0100e=3X0:2P`02PX<00o@4D203Dc=P0402@K700j<2l0
>S0_03<];@80@C/k00@0?c/j03Da<00e<C00?C`j0P0a<Bl0100Q8Al0@40n0410?P19B4H2028Q7`04
02DT8P0e=380=C@b038a;`80AdI400@0<S4_02`[:@0/:bT09bHT0P0/:bT0100i>3H0=cHd03Lf=00V
9B<2028Q7`<0:RTW00@0ADA2040o?@10?cd0?chl0P0n?C/0100c<S008b8P02<R800R8Al202/Z:004
0450?P0[:RP0:bXX020O7@808R4O00@0=3<a03He<`0f=C<0>cXh0P0j>CL0101:BDL0=SDc03He<`0_
;R`203Dd<P0403/j>012@Cl0@T4o03Th=P808B0N00@0;bh/02TX9P0Y:2H0>STg0P1CDU00101=C4X0
<2l]030_;@0_;R`204E4@P0401XI5`0m?3X0?C`j04a;B@80>CPf00@0;Rd[03/j>00k>SP0@D0n0P0h
=cD0100U92808B0N024P7P0P7ad202HU8`0403<b<00o?S`0?chl0491?`80@3lm00@09RDP02d/9`0]
;2L0;RdX0P0Z:R80100`<2P0@40f0410=P1;Bd4204a<@00402l^:@0];2L0;B`W02l^:@80>CPc00@0
@d8m040o>P10?cX0<2lZ0P0Q81/0100O7QT0;B`W02d/9`0m?3L20491?00403/j=@0a<2/0<C0[02`[
9P80:bXU00@0;BTN02hZ7`0^:Ql0<blT0P0j=R/3040l<@0403`h;@0h=2T0>3@Y03Hb:@80?CT`00@0
AT4k04Y5?`1:ACl0AT4m0P0o>SL0100h<c00=C0]03D`;@0j>SP203dm?@0403hn@00k>Sl0>cXo03He
>`80=SDm00@0?C]604E3D015@e00CDeG0P1:BU@104I6D040@d==0@3oool000@0AUAO0`17EF0504QF
H@@0BUQS1@19Ef8304QFH@`0AeEP1019Ef8304QFH@D0AeEP7018EV4204UGHP@0BUQS0P1;FF@404YH
H`<0BEMR1@18EV401017EF00B5IQ04QFH@18EV4304UGHP<0B5IQ0P17EF00101;FF@0AUAO04IDG`14
DUd204UGHP0404]II019Ef80BEMR04QFH@80BUQS00@0AUMN04QGG018Ee`0?de@0P0]=cT0100J7b80
414C010A4`0@410201@B4`<05Ph;00@05Ph<01@?300D3``04`l>0P0C3a03014?400401@=500A3180
4@`B00l=408040h?00@040X:01<92P0C2@X08Q8C0P0a7Al0101D?S<0ECh`05Dn<01K@2l206E6<P04
075?=P1kEST0NeHi08=J>P80Qe`j00@0QF13085L?`21G3l0RVE80P2DKe80102@Jdh0SfY=08mZC@2C
KU42099]D00409I/D`2EK500UFa@09A[C`80UFa>00@0UVe?09E/C02EK4`0TVU90P2>IT<01027GT00
U6]=09A[C@2FJTd209=VA`0409Q[BP2JK4/0VVa;09U[B@80WW1>00@0Wfi=0:1bD02PLU00W6m>0P2D
JDT0102@Id/0TFU?095YC`29I4X207iJ@00408]L@P2AHTP0TF9809]YD080X6iE00@0Y71H0:IbFP2V
LUX0ZgEM0P2]Mel0102XLUX0ZWAL0:YdG02`NV830;J0J080[WQP00<0Z79J0:E_E`2UKeL00`2UO5h4
0:AkG@D0YGaN00@0Z85R0:b5IP2/QFH0ZhAU0P2TOEh0102QNU/0XgaM0:=lG@2VOf020:EnG`040:Ak
F`2SNUX0XgYJ0:1gE`80WgIF00@0XGQH0:=jFP2SNUX0WGAD0P2FKDd0102KM5L0V75D09QaE02EKU42
099]D004099_D@2CL580Tg1B095`D@80Sfi?00@0RV=608iWBP2>IdX0TFU?0P2>JDh01027HdT0P5a4
081LA01hESd207EC>P0406e9<@1J=Qh0FSHN05Da6@80GC/R00@0ID<Z07=A>01cDCP0Oee40P1mGD@0
101nHTd0HTLd0697=00j8aD202lM4@0403X]9`0d;2T0=2`Y030/;@80@3i100@0:RDY02<Q9P0S8BH0
=3<i0P0Z:S80100o@4D0@3i1040n@@0^9BH202hT8`0403@^;P0h<S80>38b04Q4@`80?c/j00<0;bh/
038b<00b<S000`0^;R`0101AD4h0<c8`03<b<00V9B<2038a;`0402l^;00o?S`0?chl03<b<0807QdK
00<0:2LU02XY9`0Z:BL00`0d<c40100W9R@09RDS02HU8`0^;B/202l^;00403/j>00f=C<0=SDc02/Z
:0809RDS00@0<S4_040o?@10?cd0>STg0P0Y:2H0100H5aD08b8P02<R800O7Q`202d/:P04028Q7`0f
=C<0=SDc03Lf=080A4=100@0>cXh034`;P0a<2h0>cXh0P18AdD01012@Cl0=C@b03Dd<P0a<2h202h]
:`0403dl>P15A480ADA20450?P80<c8`00@0<2l]03Xi=`0j>CL0?3/i0P0c<S00101@Cdd0AdI404M6
A00g=S@302`[:@80<S4_00@0=3<a034`;P0a<2h0:RTW0P0[:RP0100`;bd0<S4_038a;`0[:RP3024P
7P80:2LU00@0=3<^02LV8@0W9R409RDP0P0f=Rh01013@c/0?cle03lo=@0b<RP202XZ7P0402`[9P0e
=2l0=C@_040o>P80A4<n00@0@3lj03Pg<P0h=c80<C0[0P0];2L0100e=2l0>STd03Xi=00o?ST203`k
=P0403Dd;`0_;RT0;bhY02l^:@80<S4/0`0g<bX00`0h=2/0>SH]03Xf;@0303/g;P0403Xf;@0h=2/0
>3@[03`g<@80?STc0`10>cL0100o>SL0?STf03hi=P10>SX2048l?00403Xj>P10@400@41004M7B@80
BTU>00@0BDQ>04I4C`16A4l0ATA?0P17AE801@1:BUH0BDUE04U9E@16AU80A4A@0040oooo000304ID
G`<0AeEP1018EV4204UGHP@0BUQS1@19Ef8304QFH@`0AeEP1019Ef8304QFH@D0AeEP6`18EV4404UG
HPH0BUQS1@19Ef8604QFH@0404UGHP18EV40B5IQ04QFH@<0AeEP0`18EV4204UGHP0404mMJ019Ef80
BEMR04IDG`80B5IQ00@0BEMR04QFH@18EV40B5IQ0P19Ef801017EU/0@U5F049AEP0d?d<201`V:004
00lD5`0?4A003a4@014A4@804A0>0`0F3P/0100F3P`050l<01@?300C3`h201<?40<04@l@00@04`hD
018=4`0B3A<040hA0P0>30d0100@2`P06@h<01T>300^71X2044Y:@0405To<P1K?c40Fcla05m2<P80
JDTd00@0M50h07mH>`1oF3/0QE`n0P29GSh01029I4L0QF1308EP@`2>JD`209UdE`0409QcEP2FLE@0
UW5D09A_DP80SFQ;00@0T6I=095XC02AJ4`0TVU=0P2EK4h0102FKDl0UFa<09E/C02BJDT208mWA004
08iRA@2FJTd0UVY=09IZC@80UFQ900@0Vfe<09a^C@2LKTd0WFa;0P2NKD`0102PL4`0Wfi=09m^C@2K
KD`209Q]C@0409A[C@2?J4/0SfQ;08IQAP80PEa200@0SV1609=TBP2CI4X0VV]A0P2QKeH0102SLEP0
Y71H0:A`F02TL5P20:MaF@040:IbFP2]NF40[GUQ0;B0J080]X9Z00@0]81X0:mkH`2_Nf<0Yg=K0P2P
K5@0102PM5L0XWII0:9fF@2UNE`20:MkGP040:UmH02XO5l0Z7aO0:MkGP80YGUL0`2WOV00102UO5h0
Y7]M0:AkG@2WOV020:^2I0040:V0HP2TNed0Y7]M0:MlG0<0XgQH0P2WO5`0102YOUh0YW]K0:IkF`2P
MED209ebDP0409U`E02KM5L0VgAG09]dE`80UW5D00@0TVe@08m/CP2?K4h0TVmA0P2FLeD0102?ITX0
T6U<091YC02BJdh208mZC`0408YUB`23GdD0Pem507mK@`80OEU100@0NEDm06E1:@1U@BT0Ec<K0P1F
<QX0101J=Qh0Id<[06M3:`1fDSX207eI@@0408=SCP21I580PFAB04`b8`809Q8700@0>BXS03P^;00h
;R`0:bDU0P11?Ch0100i=3P09bDZ02LU:P0c<SP202TY<@0403Df>`16A4L0ATA703Xa<P809Q`K00@0
=Bl_030Z:P0`:RX0CDU80P17@d80100V9B<0<C4_034a;`0f=S@201hN700404=2@010?cd0@3lm02TX
9P80;bh/00@0:BPV03Dd<P0e=380@3lm0P0S8R000`0^;B/0;2/Y02`[:@03030_;@0402TX9P0X9bD0
:2LU02h]:`80<2l]00<0:bXX030_;@0`;bd00`0^;B/0100l>cT0ATE304I5@`0e=38201PG5@0401/J
600L6aT071/I02l^;0806QTG00@0=3<a02l^;00_;R`0BDQ60P13@T00100b<Bl0<C0^034`;P12@Cl2
051?C@04040o?@0[:RP0:bXX030_;@80@D0n00@0?Sdk0491?`12@Cl0>CPf0P0Y:2H0100U9280<c8`
03<b<013@T0204Q7A@0403ln?00e=380=C@b02d/:PD0:bXX00@0;2/Y02l^;00_;R`0C4]90P0f=C<0
100U92808B0N024P7P0P7ad201`K6@04028Q7`0^;B/0;Rd[038a;080<C0[00@0=cHa0410>010@3P0
?Cde0P0^;R@0100V9Q`09bLK02LW6`0d<B`203Xg<P0403ll=`0i=S40>CHa02l/9`80:bPS00@0<RlZ
03/h<`0k>3<0@T4l0P0o?ST0100i>3<0<S4/038a;00^;BP202l^:@0403@c;P0i>3<0>CPc03ll<`80
?S/b00@0?3T`03Tf;@0i=Rd0=S<Z0P0e<RT00`0g=2/0>CH]03Tf;@0303`k=P0403/j=P0j>CD0>STe
03/j>080?cdn00@0AD=404Y8B@1:B4T0@D=20P16AdT0101<CDl0D55F051AEP1ADET204i>F00404Y:
EP16AU80ATIB04I8E080ADMC0@14AU8104=5D@40oooo000304IDG`<0AeEP1018EV4204UGHP@0BUQS
1@19Ef8304QFH@`0AeEP1019Ef8304QFH@D0AeEP6`18EV4404UGHPH0BUQS1@19Ef8604QFH@0404UG
HP18EV40B5IQ04QFH@<0AeEP0`18EV4204UGHP0404mMJ019Ef80BEMR04IDG`80B5IQ00@0BEMR04QF
H@18EV40B5IQ0P19Ef801017EU/0@U5F049AEP0d?d<201`V:00400lD5`0?4A003a4@014A4@804A0>
0`0F3P/0100F3P`050l<01@?300C3`h201<?40<04@l@00@04`hD018=4`0B3A<040hA0P0>30d0100@
2`P06@h<01T>300^71X2044Y:@0405To<P1K?c40Fcla05m2<P80JDTd00@0M50h07mH>`1oF3/0QE`n
0P29GSh01029I4L0QF1308EP@`2>JD`209UdE`0409QcEP2FLE@0UW5D09A_DP80SFQ;00@0T6I=095X
C02AJ4`0TVU=0P2EK4h0102FKDl0UFa<09E/C02BJDT208mWA00408iRA@2FJTd0UVY=09IZC@80UFQ9
00@0Vfe<09a^C@2LKTd0WFa;0P2NKD`0102PL4`0Wfi=09m^C@2KKD`209Q]C@0409A[C@2?J4/0SfQ;
08IQAP80PEa200@0SV1609=TBP2CI4X0VV]A0P2QKeH0102SLEP0Y71H0:A`F02TL5P20:MaF@040:Ib
FP2]NF40[GUQ0;B0J080]X9Z00@0]81X0:mkH`2_Nf<0Yg=K0P2PK5@0102PM5L0XWII0:9fF@2UNE`2
0:MkGP040:UmH02XO5l0Z7aO0:MkGP80YGUL0`2WOV00102UO5h0Y7]M0:AkG@2WOV020:^2I0040:V0
HP2TNed0Y7]M0:MlG0<0XgQH0P2WO5`0102YOUh0YW]K0:IkF`2PMED209ebDP0409U`E02KM5L0VgAG
09]dE`80UW5D00@0TVe@08m/CP2?K4h0TVmA0P2FLeD0102?ITX0T6U<091YC02BJdh208mZC`0408YU
B`23GdD0Pem507mK@`80OEU100@0NEDm06E1:@1U@BT0Ec<K0P1F<QX0101J=Qh0Id<[06M3:`1fDSX2
07eI@@0408=SCP21I580PFAB04`b8`809Q8700@0>BXS03P^;00h;R`0:bDU0P11?Ch0100i=3P09bDZ
02LU:P0c<SP202TY<@0403Df>`16A4L0ATA703Xa<P809Q`K00@0=Bl_030Z:P0`:RX0CDU80P17@d80
100V9B<0<C4_034a;`0f=S@201hN700404=2@010?cd0@3lm02TX9P80;bh/00@0:BPV03Dd<P0e=380
@3lm0P0S8R000`0^;B/0;2/Y02`[:@03030_;@0402TX9P0X9bD0:2LU02h]:`80<2l]00<0:bXX030_
;@0`;bd00`0^;B/0100l>cT0ATE304I5@`0e=38201PG5@0401/J600L6aT071/I02l^;0806QTG00@0
=3<a02l^;00_;R`0BDQ60P13@T00100b<Bl0<C0^034`;P12@Cl2051?C@04040o?@0[:RP0:bXX030_
;@80@D0n00@0?Sdk0491?`12@Cl0>CPf0P0Y:2H0100U9280<c8`03<b<013@T0204Q7A@0403ln?00e
=380=C@b02d/:PD0:bXX00@0;2/Y02l^;00_;R`0C4]90P0f=C<0100U92808B0N024P7P0P7ad201`K
6@04028Q7`0^;B/0;Rd[038a;080<C0[00@0=cHa0410>010@3P0?Cde0P0^;R@0100V9Q`09bLK02LW
6`0d<B`203Xg<P0403ll=`0i=S40>CHa02l/9`80:bPS00@0<RlZ03/h<`0k>3<0@T4l0P0o?ST0100i
>3<0<S4/038a;00^;BP202l^:@0403@c;P0i>3<0>CPc03ll<`80?S/b00@0?3T`03Tf;@0i=Rd0=S<Z
0P0e<RT00`0g=2/0>CH]03Tf;@0303`k=P0403/j=P0j>CD0>STe03/j>080?cdn00@0AD=404Y8B@1:
B4T0@D=20P16AdT0101<CDl0D55F051AEP1ADET204i>F00404Y:EP16AU80ATIB04I8E080ADMC0@14
AU8104=5D@40oooo000404MEH0D0B5IQ0`19Ef8404YHH`D0BEMR3`18EV4404UGHP<0B5IQ1@17EF0H
04QFH@T0BEMR0`1:F6<304UGHPT0B5IQ00@0BUQS04UGHP19Ef80AeEP0`16E5l204MEH00404UGHP1:
F6<0BUQS04iLI`80BUQS0`17EF001018EV40BEMR04UGHP19Ef8204YHH`0404IEFP0mBdh0?D]>02Lb
=0803aLI00@02@d>010@400@41004a8@0P0A3@`301H>2`0401H>300D3``050l<01<?3P804`l@0`0A
3a00100C3Q@04@lB014?4P0@3Pl2010<2`0401@<2@0S5A808aDB03`W8P80CcHb00@0FC`^05dm;P1M
?Bh0Hd8_0P1[B380101eD3H0OULj07iG>P24Fcd208UN?P0408iWBP29HTD0RF95095ZC@80VgAG00@0
W7EH09efF@2MMUT0VG9E0P2>IdX0102AIdh0TVU=099YC@2EK50309M^D080UFa<00@0TVU908mWA02?
Id@0TFE80P2DJ4/0102EJ4/0UfQ:09MXBP2JK4/209e/C00409a[BP2KJTT0VfY90:5_C080Vf]700@0
VF]909e`C`2ML4l0VfmB0P2@Id/01028H4H0QF1508EPA@2BI4X209EWC@0409UZD02LKE<0W6eC09e[
DP80VfU@00@0VVI>09QTC02HI4`0XViF0P2]NF400`2cOfL0/7aT0;1lI0030:ijHP040:YfGP2RKUH0
XViF09Q]C@80W75A00<0XgQH0:UnGP2YOUh00`2]PV80102/PF40ZWmO0:YoG`2VOEl20:9iF`0409mf
F02SNU`0XgYL0:b3I@80/8MY00@0Zh9T0:=jG02SNU`0ZGiN0`2TNET20:b1H@030:f2HP2UNUX0YGYJ
00<0WgAD00@0WWEI0:5hG02QN5`0XGQL0P2MMUT0102GLUD0UW5D09IaE02GM5H209]hFP0409YaD`2G
KU00Ufi@099YC@80RfA700@0QF15089M@`22GD<0PEe30P22GT@01020GTD0LU0g079@=`1U@BT205`h
800405Tc701P>R<0H3XS079<=@80PE]400@0OEY408YZE@2:JUD0Jdhn0P10:A/0100h9Ah0=bXT03LZ
900c:BP203`c=00404e8C00_;C80;bdb034`=P809RH^00@09BH[03hl?`0n?3l0@STj0P0_9B@0100e
;bl0;bTY02lY:@14@3l204<o?P0402h]:`0`<2h0<30^02d]:`809BDS00@0;Rd[04I5@`16AD<0<C0^
0P0`;bd0100S8R00:BPV02TX9P1:BDL2034`;P0303He<`0a<2h0<C0^00<0;Rd[00@0:bXX02PW9@0X
9bD0:BPV0P0[:RP0100S8R00<2l]030_;@0o?S`204A3@@0403`k>@0/:bT0;2/Y01hM6`805aHD00@0
:bXX02h]:`0^;B/061LE0P0e=380100_;R`0D4m=051?C@0g=S@2034`;P0403Th=P0h=cD0>3Le03dl
>P80?chl00@0=SDc034`;P0a<2h0>cXh0P19B4H0100j>CL0<C0^034`;P0/:bT202h]:`0403<b<00h
=cD0>3Le0491?`80BdY800@0<C0^03He<`0f=C<0=3<a0P0^;B/0100a<2h0?C`j03dl>P16AD<204M6
A00403hm>`0^;B/0;Rd[028Q7`8092<Q0`0Z:BL0100Y:2H0:2LU02PW9@0`;bX203ln>@0404A3?P0g
=bl0=cL_02XZ8P80;2`R00@0=SH/03/k;`0k>bl0>cPa0P0i=Rl0100c<2T0;R/T02h[900[:24202h[
900403Db:`0k>340>cPa03`k=P80=cHa00@0<S4/02h]:00^;BP0;RdX0P0b<B`0100h=c80>cXe03/j
=@0l>cH203dl=`0403`k=P0i>3<0>CPc03Dd;`80=3<^00@0=SD`03Xi=00j>C@0>c/i0P0m?C/01010
@400@d=304=3@`15ADL204M7B@0404U8C@1:BDh0BTU>04]?D080BDe>00@0ATY=04I9CP16BDh0AdYA
0P16B5@00`14AU80@D=@0453D00304=5DP80A4IC0@3oool000<0AeEP1018EV4504UGHP@0BUQS1@19
Ef8?04QFH@@0BEMR0`18EV4504MEH1P0B5IQ0`1:F6<904UGHP<0AeEP2@18EV40101:F6<0BEMR04UG
HP17EF0304IDG`80AeEP00@0BEMR04YHH`1:F6<0BeUT0P19Ef8304IDG`0404QFH@1;FF@0BeUT04]I
I080BUQS00@0AE=F03M2AP0g@TH07BLY0P0;41<010072`X0410>010@3P0E4@h2010<2@<05Ph;00@0
5Ph<01@?300D3``04`l>0P0C3a03014?4004014?500B41<04Q0C010>3`804@d:00@06a8=034P6@0a
81T0BS4Z0P1K?SH0101P?2`0Hd0]06=0;@1ZAC82079<=`0407]C>P22FCd0PUTm08MN@080S61300@0
T6U<08YSAP2:HdH0SfQ;0P2GL5<0102JLeH0X7UL0:1iG02NMeX209=/C`0409M]E02HKe<0V6mC09Ya
E@80VW5C00@0V6mA09E/C02EK4`0TVU90P2@J4D0102BIDP0T6=6091SAP2CI4P209IWB@<0V6M700@0
VFM609QVA@2HITD0Wfe80P2IId@0102HIdH0WW1?09i`C`2MLU82095XBP0308YSAP2;I4L0RfA700<0
TfM<00@0UVQ>09QYC`2HJDl0V6U?0P2HITd0102EHdX0Tf5809=QB02PKUD20:UgGP040:ilH`2ZN5l0
ZWQO0:UgGP80[G]R00@0[WaS0:YhG`2ZN5l0Vfi=0P2OLU400`2VNEP0[7mN0:aoGP030;23HP040:j1
H02]P5l0[H1O0:UmH080XWII00@0Wg=F0:IjG@2VNUd0[h=V0P2bQVT0102[Of80Y7QK0:AhF`2WNU/2
0:9eEP040:=fE`2YO5d0ZGaM0:UlG@80X7=D00@0VVe>09]^C`2KKTl0YGUN0P2SMe`0102NMET0WGAH
09edF02MMUT209]fF@0409YeF02IM5L0VGAG09eaE080VFe@00@0TFQ:08]RAP2;HTH0Qf130`24Gd@2
08AOA@0407iL@`1jF3l0NUPo07=?=`80J48[00@0GSPQ05he7`1N=Al0KT@^0P23FD<0101gDCX0P5a6
081LAP27IU<206I:?00403DO500b8AX0<R4J048d<@80>Bd]00@0CTU=038`=@0b<3D0<2le0P0Z:S80
100N7b@0=C<f03Dc=P15?3d2044g=P0403H`<00c;Bd0<bd]03@`;`80=c<b00@0?chl02l_;@0_;bd0
7AdK0P0g=cD0100Z:BL0ADA204E4@P0h=cD202l^;00402@S8@0U92809B@R04Q7A@80=SDc00@0=C@b
02l^;00_;R`0:bXX0P0];2X0100_;R`0:RTW02XY9`0W9R@302/Z:080=SDc00@0?Sdk03Xi=`0j>CL0
:BPV0P0K6QP0100N7A/0:2LU02PW9@0S8R0201/J600403He<`0a<2h0<C0^05=BD080=C@b00@0=SDc
028Q7`0R8Al0@3lm0P0o?S`0100c<S00:BPV02TX9P0b<Bl204=2@00404I5@`0o?S`0?chl03dl>P80
=cHd00@0<S4_03Dd<P0e=380@3lm0P15A480100f=C<08b8P02<R800e=38203He<`0403Lf=00j>CL0
>STg0450?P80ADA200@0?3/i02h]:`0^;B/081lM0P0X9bD0100/:bT09bHT02LV900S8R0202DT8P04
02PW9@0W9R@09bHT03Xi=080>CPc00@0<S4/02/[8`0[:b<0<C4Y0P0n?S@01010@3H0>CT]03Ti;@0k
=S0203Td;P0403Pc;@0j=Bl0>SD_03dh<P80?cXd00<0?3La03Xe;`0j=Bl00`0b<B`0100c<Rd0=SD`
03He<00i>3<203Xi=00403Th<`0h=c80>3Lb03@f<@80>3Xe0`0k?CP0100i>cH0>3Xe03Pj=@0j?3L2
03do>P040414A@12ATL0@TI704E9BP80B4a?0`1;CU<01019C5<0B4]B04Q;DP1=De<204M<C`040456
BP10AD/0@4E;0455CP80@dMB00@0@TIB0413DP10@e80@DEA0P12AU8104=7D`40A4QD0@3oool000<0
AeEP1018EV4504UGHP@0BUQS1@19Ef8?04QFH@@0BEMR0`18EV4504MEH1P0B5IQ0`1:F6<904UGHP<0
AeEP2@18EV40101:F6<0BEMR04UGHP17EF0304IDG`80AeEP00@0BEMR04YHH`1:F6<0BeUT0P19Ef83
04IDG`0404QFH@1;FF@0BeUT04]II080BUQS00@0AE=F03M2AP0g@TH07BLY0P0;41<010072`X0410>
010@3P0E4@h2010<2@<05Ph;00@05Ph<01@?300D3``04`l>0P0C3a03014?4004014?500B41<04Q0C
010>3`804@d:00@06a8=034P6@0a81T0BS4Z0P1K?SH0101P?2`0Hd0]06=0;@1ZAC82079<=`0407]C
>P22FCd0PUTm08MN@080S61300@0T6U<08YSAP2:HdH0SfQ;0P2GL5<0102JLeH0X7UL0:1iG02NMeX2
09=/C`0409M]E02HKe<0V6mC09YaE@80VW5C00@0V6mA09E/C02EK4`0TVU90P2@J4D0102BIDP0T6=6
091SAP2CI4P209IWB@<0V6M700@0VFM609QVA@2HITD0Wfe80P2IId@0102HIdH0WW1?09i`C`2MLU82
095XBP0308YSAP2;I4L0RfA700<0TfM<00@0UVQ>09QYC`2HJDl0V6U?0P2HITd0102EHdX0Tf5809=Q
B02PKUD20:UgGP040:ilH`2ZN5l0ZWQO0:UgGP80[G]R00@0[WaS0:YhG`2ZN5l0Vfi=0P2OLU400`2V
NEP0[7mN0:aoGP030;23HP040:j1H02]P5l0[H1O0:UmH080XWII00@0Wg=F0:IjG@2VNUd0[h=V0P2b
QVT0102[Of80Y7QK0:AhF`2WNU/20:9eEP040:=fE`2YO5d0ZGaM0:UlG@80X7=D00@0VVe>09]^C`2K
KTl0YGUN0P2SMe`0102NMET0WGAH09edF02MMUT209]fF@0409YeF02IM5L0VGAG09eaE080VFe@00@0
TFQ:08]RAP2;HTH0Qf130`24Gd@208AOA@0407iL@`1jF3l0NUPo07=?=`80J48[00@0GSPQ05he7`1N
=Al0KT@^0P23FD<0101gDCX0P5a6081LAP27IU<206I:?00403DO500b8AX0<R4J048d<@80>Bd]00@0
CTU=038`=@0b<3D0<2le0P0Z:S80100N7b@0=C<f03Dc=P15?3d2044g=P0403H`<00c;Bd0<bd]03@`
;`80=c<b00@0?chl02l_;@0_;bd07AdK0P0g=cD0100Z:BL0ADA204E4@P0h=cD202l^;00402@S8@0U
92809B@R04Q7A@80=SDc00@0=C@b02l^;00_;R`0:bXX0P0];2X0100_;R`0:RTW02XY9`0W9R@302/Z
:080=SDc00@0?Sdk03Xi=`0j>CL0:BPV0P0K6QP0100N7A/0:2LU02PW9@0S8R0201/J600403He<`0a
<2h0<C0^05=BD080=C@b00@0=SDc028Q7`0R8Al0@3lm0P0o?S`0100c<S00:BPV02TX9P0b<Bl204=2
@00404I5@`0o?S`0?chl03dl>P80=cHd00@0<S4_03Dd<P0e=380@3lm0P15A480100f=C<08b8P02<R
800e=38203He<`0403Lf=00j>CL0>STg0450?P80ADA200@0?3/i02h]:`0^;B/081lM0P0X9bD0100/
:bT09bHT02LV900S8R0202DT8P0402PW9@0W9R@09bHT03Xi=080>CPc00@0<S4/02/[8`0[:b<0<C4Y
0P0n?S@01010@3H0>CT]03Ti;@0k=S0203Td;P0403Pc;@0j=Bl0>SD_03dh<P80?cXd00<0?3La03Xe
;`0j=Bl00`0b<B`0100c<Rd0=SD`03He<00i>3<203Xi=00403Th<`0h=c80>3Lb03@f<@80>3Xe0`0k
?CP0100i>cH0>3Xe03Pj=@0j?3L203do>P040414A@12ATL0@TI704E9BP80B4a?0`1;CU<01019C5<0
B4]B04Q;DP1=De<204M<C`040456BP10AD/0@4E;0455CP80@dMB00@0@TIB0413DP10@e80@DEA0P12
AU8104=7D`40A4QD0@3oool0000304MEH018EV40B5IQ00<0B5IQ0`19Ef8704YHH`D0BEMR0`18EV4@
04UGHP<0B5IQ1@17EF0H04QFH@@0BUQS1@19Ef8304QFH@0304IDG`17EF00AeEP00<0AeEP1P18EV40
1019Ef80B5IQ04QFH@18EV4304MEH0<0B5IQ0P19Ef801017EF00B5IQ04QFH@17EF0204ECGP0404QF
H@1;FF@0BeUT04YHH`80AUAO00@0?4Y=02Xe=`0Z=CL04adN0P093`l0100<3Pd04Q4?018A3`0C3``2
018=2@<05Ph;00@05Ph<01@?300D3``04`l>0P0C3a03014?4004014?500B41<04Q0C010?3@805A0<
00@091XA03d[8@0m:b40ECX_0P1R@SL0101W@Bh0K4Hc06a6<`1eC3P207aC?@04089H@028GTD0R5i5
08mSB080TfM:00@0TfY>08mVBP2?ITX0TfY>0P2HKe<0102JLED0XWUM0:9iG@2SNUh209Q_D`0409ec
FP2NMET0WWEI09ieF@80W7=E00@0V6mA09A[B`2DJd/0TfY:0P2CJdP0102BIDP0SUm308iO@`2BHdL2
09UXBP0409ETA02DHD80U65209IT@`80Uf9200@0WFU409IT?`2FI3l0UFE10P2IJdT0102IK4/0TFE8
095UB02:HD<208YQA@0408mSB02>HTL0SV9708iRA`80TF=900<0U6E;09IWC@2FIdd00`2GID`0102N
K5<0Y79I0:AbF@2YMeh20:QfG@040:MeG02[NF00ZgUP0:mmI080/7iU00@0YGMF0:MiF02WNEP0ZWaK
0P2]Oeh0102^P5l0[GmN0:eoGP2/OUd20:]mG0040:]oHP2TN5/0Y7QK0:9fF@80Yg]N0`2]PF@0102W
Neh0Y7QK0:AhF`2YNU`20:QiF`040:UjG02/OEl0[7eO0:]lGP80YWMI00@0XgAF0:9cE@2RLeD0YGUN
0P2OLeP0102IKE80V6mC09Q_D`2LLeL209efF@0409UbE@2CKU40TfiA09I[B`80UFY:00@0U6Q;099Y
B`2BJD/0TFQ<0P2=ITT01028H4H0Pei308=N@`1hF3l207aJ@@0407QD?01_BC80KdTb06Ln:080GcDO
00@0J3/V07]<>01kC3P0Ne4i0P1hCcT0102?K5P0P61A081PD@0n9QX203<O600404Lh<`0l;R/0?2h[
03/f>P80<c4f00@0<S4g034a>@0a<CT092DZ0P0d<SD01013>S/0C49104a2@@0h<S8203H`<00402XV
9@0^:RT0;RXY04I5@`80=CDc00@071`J03hn?00n?S`0:RTW0P0d<c40100f=C<0:bXX02/Z:00];2X2
038a;`040491?`0k>SP0>cXh02/Z:080:BPV00@09bHT02h]:`0^;B/0=3<a0P0];2X0100W9R@0;B`Z
02d/:P0h=cD203`k>@0403Lf=00W9R@09bHT01hM6`808B0N00@09B@R02<R800S8R004a8@0P0^;B/0
100i>3H0EUEC05IED`0e=38203Lf=00302d/:P15A480ADA200<0?3/i00@0;Rd[02PW9@0X9bD0>STg
0P1;BTP01012@Cl0<2l]030_;@0j>CL203Th=P040450?P1;BTP0BdY804A3@@80<C0^00@092<Q02<R
800S8R00;Rd[0P0_;R`0100k>SP0B4M504Q7A@13@T0202d/:P0401`K6@0H5aD061LE02TX9P80;2/Y
00@09B@R01XI5`0J6AL07QdK0P0`;bd0100k>SP0>STg03Xi=`0g=S4202`[9P0402/Z9@0h>300>3P`
0451>@80?3`b00@0=3@Z038b9P0b<RH0>34Y0P0i<RX0100k=2`0?CH^03df;P0n=bl303lh<080?SL_
00@0<c8]03Lf<@0g=S40?3/f0P10?cX01012@C`0?chi03ln>@0j>C@203He<004038g<`0f>cL0=S/g
03]0?080?dA00`10AD430456@P0404Q@DP16CU40ATiA04E=D080ATeC00@0B4mE04U@F019D5P0BE1H
0P19D5X01013CDh0@D]<045;C011Bdd204=<D@0404E>E`16CUT0ATiI04=;F080@4QE0`11AeD20498
EP40oooo000404QFH@<0BEMR2@1:F6<504UGHP<0B5IQ4019Ef8304QFH@D0AeEP6018EV400`1;FF@0
BUQS04YHH`0304YHH`0304UGHP18EV40B5IQ00<0B5IQ1016E5l304MEH0D0B5IQ00@0AeEP04QFH@18
EV40B5IQ0`19Ef8304QFH@80AeEP00@0AE=N04QFH@18EV40B5IQ0P16E5l01018EV40BUQS04YHH`17
EF02041>F@04030k?@0K9BL06bDW00TA4`802Ph?00@0418?014@300A40`04`h:0P0D3`/301H>2`04
01H>300D3``050l<01<?3P804`l@0`0A3a00100@3a@0410B010@4P0@3`d201PC3@0402`R6018=2T0
B3@Y05a0=080I4<d00@0K44^0756<`1aAS<0NDli0P20ESh01025Fd80S61708aPA`2AIDX209EYCP04
09M^DP2EK500UFa@09YaE@<0W7=G0P2SNUh0102TNel0VG1D09U`E02NM5/209mfFP0409ieF@2JLE<0
VW5C09I]C`80TfY:00@0U6];09E]BP2EKDX0TfA80P2>Gd<0102GIDX0VfY<09]ZC02DHD82091M?P04
09MR@P2HHd<0V6=309YV?`80Uf<n00@0Tf4n09=R@@2CHT40T6=20P2;H4001024Fcd0PUTk089I>`27
GT8208AK?`0408AH?@26F3h0QUPn08UK@@80SUm50`2?H4H0102CI4X0UfQ>09MXCP2MKU@20:9cF@04
0:5bF02OL5H0Wg1F0:=dFP80ZGYP00<0ZWaJ0:UkF@2YNeT01P2YNeT30:YlFP040:j1I02YO5l0ZGaO
0:MjG@80ZWeP00@0ZgiQ0:IiG02VNE`0Y7MJ0P2UN5/0102ZNed0[WmQ0:ioH@2_P6830:ioH@80/85S
00@0[h1R0:amG`2/OEl0Y7IL0P2PLUP309Y^D`0409]bEP2LMEP0W7EH09UbE@80UVmB00@0VVe>09U/
C@2IK4d0Ufa<0P2HK4l0102EK4h0T6U<091YC02:HdH208EM@`0407aL@`1oGD@0Oee407QD?080Lddf
00@0LTPb06Pk9P1X>bH0Hc@P0P1^?RX0101mDCP0MT`d07I<=025GdX208IUE0040614>010:B40@2TQ
03/X8P80?2XX00@0;BP/03dk@00m>d00=cHl0P0`<3P0100Y:Rl0=cDh03Le>00m=3D204Y0?`0403/e
=@0b;2`0<R`/02l[:P80;BTX00@0>STg03lo?@0o?cd0;bl]0P0^;R`0100];2X092<Q02@S8@0j>CL2
02`[:@0403Dd<P0j>CL0>STg03Dd<P80@T4o00@08b8P02DT8P0U92809RDS0P0a<2h0100h=cD0;B`Z
02d/:P0T8b4202/Z:0<0>3Le00@0=3<a02`[:@0/:bT092<Q0`0Q81h2028Q7`0402XY9`0b<Bl0<S4_
04Q7A@80>CPf00@0>cXh02/Z:00[:RP0@T4o0P1=C4X303<b<004034`;P0i>3H0>CPf04A3@@80?Sdk
0`0a<2h0100c<S00>3Le03Pg=@15A48204Y9A`0403Th=P0U92809B@R02XY9`80?Sdk0`0l>cT0100m
?3X0>cXh03/j>00`;bd202HU8`0402PW9@0b<Bl0<S4_02TX9P809bHT00@08b8P02HU8`0V9B<0=3<a
0`13@T0203Th=P04030_:P0a<2/0<C0[03Xi=080@T8j00@0?Shf038b:00b<RP0<C4W0P0i>Bd0100i
<RX0>S<[03Xc:`0j<b/303Tb:P80>S<[00@0?CH^03lh<00o>300=cHa0P0j>C@0100n?CP0?chi03ln
>@0o?ST203dl=`0403`k=P0k>SD0>cXe03]1?`80?D=100@0?dE304A:B014BTP0Ade;0P18CT`01017
CD/0ATa:04I<BP1:DeP204M@E@0404E>D`15CUD0ADiE04M@F@80BE9K00@0BE5L04Q@F`18D5/0@D]<
0P12C4h01015CU<0B55H04QAF019DU/204Q@F`0404E=FP12BET0@TUI04=9E`@0@TQF0@3oool000@0
B5IQ0`19Ef8904YHH`D0BEMR0`18EV4@04UGHP<0B5IQ1@17EF0H04QFH@0304]II01:F6<0BUQS00<0
BUQS00<0BEMR04QFH@18EV400`18EV4404IDG`<0AeEP1@18EV401017EF00B5IQ04QFH@18EV4304UG
HP<0B5IQ0P17EF001015Deh0B5IQ04QFH@18EV4204IDG`0404QFH@1:F6<0BUQS04MEH080@4iI00@0
<3/m01/U9`0K9BL02A4C0P0:3Pl0100@4Pl04A0<014@300C3PX201@?2`<05Ph;00@05Ph<01@?300D
3``04`l>0P0C3a03014?4004010?500@4180410B010?3@8061<=00@0;28H04Pd:@18=2T0G40d0P1T
@c@0101/@Bh0LDHc0756<`1iCcT2081F?P0408EK@P2<H4L0S617095UBP80UFU>00@0UfiB09E/D02E
K500VW5E0`2LLeL20:=jGP040:AkG`2IL5@0VG1D09idF`80WgIJ00@0WWEI09YaD`2JLE<0UVe?0P2C
JTX0102DJd/0UFe:09E]BP2CI4P208iO@`0409MUBP2KJT`0VfY<09AQ@P80T5dn00@0Uf9209QS@`2H
Hd<0VVHo0P2GHch0102CHCh0Tf9109=R@@2@Hd8208]P@00408AK?@22FC/0PUTk08MN@P80Q5/o00@0
Q5Pm08IH?P26F3h0RE]10P2>GdD308mPAP0409=TBP2GJ4h0UfQ>09e^E080XW=I00@0XG9H09m`EP2O
L5H0XgAJ0P2YNV000`2ZO5X0ZG]I0:UkF@060:UkF@<0ZWaJ00@0[X5T0:UlG`2YO5l0YgYM0P2ZOF00
102[OV40YWUL0:IiG02TMeX20:EhF`040:YkG@2^Of40[WmQ0:n0HP<0[WmQ0P2`PF<0102_P680[7eO
0:amG`2TMU`20:1bF0<0VViC00@0Vg9F09aeF02LMEP0VG9E0P2FKe80102JKDh0VFa=09U/C@2GK4`2
09Q/C`0409E/CP2@JD`0T6U<08YSAP80QEe300@0O5a307mMA01oGD@0N5@l0P1cCCH0101bB380J3/V
06Pk9P1S=20206hn:P0407eA>01fC3@0MT`d08EOBP80QVED00@0H4@h040Y8@10:B40>bPR0P0l:RP0
100]:2`0?C]003dk@00g=S`2030`>00402TZ;`0g=CP0=cDh03dd=@80BT0o00@0>cDe038/;00b;2`0
;b/Z0P0]:BP0100j>CL0?clm03lo?@0_;bd202h^;00402d/:P0T8b4092<Q03Xi=`80;2/Y00@0=C@b
03Xi=`0j>CL0=C@b0P12@Cl0100S8R009B@R02DT8P0V9B<2034`;P0403Pg=@0];2X0;B`Z02@S8@80
:bXX0`0h=cD0100d<c40;2/Y02`[:@0T8b43024P7P808R4O00@0:RTW038a;`0b<Bl0B4M50P0i>3H0
100k>SP0:bXX02/Z:012@Cl204e<BP<0<c8`00@0<C0^03Th=P0i>3H0A4=10P0n?C/3034`;P0403<b
<00h=cD0>3Le04E4@P80BTU700@0>CPf02DT8P0U9280:RTW0P0n?C/303`k>@0403dl>P0k>SP0>cXh
030_;@809RDS00@0:2LU038a;`0b<Bl0:BPV0P0W9R@0100S8R009RDS02HU8`0d<c4304=2@080>CPf
00@0<2lZ034`:`0a<2/0>STd0P12@SX0100n?SH0<S8X038b:00a<BL203Ti;@0403Tb:P0j<b/0>S<[
03Xc:`<0>C8Z0P0j<b/0100m=Rh0?cP`03lh<00g=S4203Xi=00403hm>00o?ST0?chi03ln>@80?C`g
00@0?3/f03/j=@0k>SD0>d4o0P0m@d40100oAD<0A4Y804A:B017CD/204Q>C00404M=B`16C4X0ATa:
04YCF080Ae1E00@0ADiC04E>E@15CUD0Ae1I0P19DU/01019DE`0B51K04Q@F`11Bd`2049<CP0404E>
D`18DEP0B55H04UBF`80B51K00@0ADeJ0499F@12BET0@dUG1012B5H10?ooo`000`18EV4404UGHPT0
BUQS1@19Ef8304QFHA00BEMR0`18EV4504MEH1P0B5IQ0`1;FF@304YHH`0304UGHP18EV40B5IQ00<0
B5IQ0`15Deh304IDG`0304MEH018EV40B5IQ00<0B5IQ00@0AUAO04MEH017EF00BEMR0`1:F6<204UG
HP0404MEH016E5l0AUAO04ECGP<0BEMR0`17EF0204UGHP0404ABG@0kBE@0>dUD02D`<P803aTJ00@0
0@T;00X>3@0:3Pd04a<A0P0A40`00`0B3@T05a8>01LB3P0301H>2`0401H>300D3``050l<01<?3P80
4`l@0`0A3a00100@3a@0410B010@4P0A40`201XE3`0403<V7@1>>bd0CS/]0613=@80Hd8c00@0JchY
0754;`1aA2l0ND`g0P20Dc`01025FD00RUi508YNA@2?HdP209=WC00409]_E02LL5D0W71E0:9fF`80
XgML00@0XGEJ0:MkH02WNf00Yg]P0P2KKe@0102NM5/0WWEI09ieF@2LLeL209Q_D@0409=ZC02BJDT0
TVU909A[B`80Ufm<00@0TfA808iO@`2>Gd<0VFM<0P2NKDl0102CH440TE`m095L?@2HHd<209UTA004
09UU?P2GHc`0Uf<l095O>P80SEdi00@0RE/j08II>P26FCX0PEDh0P1kDS@01022FCd0OE@h07eD>01k
Cc@207Y><`0307e?=@21DcT0PE<i00<0Q5Dk00@0RU]108aM@`2<GD<0U6E;0P2LKE<0102JJe40U6E;
09AUB`2FIdd209m`EP040:QjF02VN5H0YWQF0:EgE@80Y7ID00@0YGME0:QjF02XNUP0ZgeK0P2]Oed0
102`PfH0ZgiQ0:]nH@2ZOF020:aoHP040:QkGP2QM5L0XGAG0:5dE`80YGQK00@0XgAF0:YkG@2ZNed0
[7eO0P2XNE/0102YNU`0/85S0;21H`2aPV@20:enH0040:EgG@2SME/0XgEK0:1dF@80Wg=H0`2MM5P0
102LMEP0VgAG09]dE`2VNEP20:9eE00409]`D02GJdh0Uf]>09=ZC080T6M;00@0RfA708MP@`27H4<0
QVI=0P22HTT0101hE3`0MDlh07E?>01kDC/20793;`0406@d801X=B80J3DR07a>=080Ndhg00@0N58m
08MSD`27He<0PVAI0P1?=Rl0100Y5Q00=R@P03HT800]:2`204]9CP0403/j@00Z:S80:RXb02LX;@80
>3Hi00@0=bh_04@j>@14>ST0?CLg0P0]9bL0100h=3<0<Bd/034];00Z:BL204Q8AP0404=3@@0K6aT0
6a/I03He<`808b8P00@0BDQ603<b<00c<S00=3<a0P0c<S00100Q81h0?chl03ln?00R8Al202HU8`04
02TX9P0c<S00<c8`03Xi=`80:RTW00@07QdK02HU8`0V9B<0<2l]0P0Z:BL0100_;R`0>3Le03Pg=@0Z
:BL201DD4P0401hM6`0i>3H0>CPf03dl>P80>STg0`0];2X0100a<2h0?chl03ln?00i>3H2030_;@04
02h]:`0/:bT0;2/Y03@c<@<0B4M50P0[:RP0100S8R00?3/i03`k>@0d<c4204E4@P0403ln?00W9R@0
9bHT02LV9080@3lm00@0AdI403He<`0f=C<0>cXh0P0l>cT0100`;bd08b8P02<R800X9bD203He<`04
03@c<@0T8b4092<Q020O7@809B@R00@0;B`Z03<b<00c<S00=C@b0`0d<c4203Dd<P0403He<010?cX0
@3lj0491?080>CTa00@0<c<[03Ti;`0i>Bl0?cle0P0n?S80100m=R`0?SL]03hg;@10>Bl204@m<`04
04Ho=@14?C<0A3dc03lh;P80>C8X00@0=cHa03Pg<P0h=c80>3Lb0P0g=S40100f=C00>3Lb03Pg<P0m
?3L20450>`0404I<BP14BTP0A4Y804A:B080Ade;00@0Be5?04eCD@1=De40Be5?0P19Cdd01014CeD0
@diD04=>E014CeD204I@F@0404UCG019Ded0BE=M04I@FP80@deI0`15D5801015D5@0AE1F04E@EP14
CUL2049<EP03045;E`11BUT0@DYI00<0A4YH0@12B5H10457E@40oooo000X04UGHP<0B5IQ1@17EF0=
04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0=04QFH@80AeEP1P16E5l00`17EF00B5IQ04QFH@04
04QEGP80B5AP0`18E68304QDI0<0BEES00@0AU9N04=@F@13D5T0AU=L0P17E5`0100lBE40;S]102hk
@@0;4aD200XC4P0400/A3`0=3``03@l<00l>2P804`h:00@05Pl901L@2P0G40X03`l30P0D50X0100D
50`04A0<014@300A3a0201@B5@0401<A5P0=2Q403@XA00dC4`803Q0;00@06AD<038T5`0b91L0BcPY
0`1JAC@205M2<@0406i09P1dAR`0M4H/07Y<<P80OU0f00@0QUPn095SB@2AHdT0UfU?0P2FJ4h0102K
KEH0W6iG09a^E`2KKeH209ibF@040:ElH02XPF80Z85R0:5jF`80UFi?00@0Vg5H09=WC02CId`0VFeB
0P2HJdh01028Fch0Sf1208mP@P2NKe4209QZB@0409ES@P2JJ4L0VVQ709e[BP80VfU800@0U69108iL
>`2>G3/0Sedl0P2AGch01029FcL0RU`h08YL>026F3H207eA;P0407U<:`1hCB/0N4d[07A9:@80Kd@T
00@0MdLa07a<=P1lC3H0LD4[0P1`@R/0101dARl0ND`e07U<=@23ESl207U?=`<0RE]100@0SEm5099V
B`2BIT/0SFA80P28Gd<0102<IDP0V75D09QaE02NKe420:AeE`040:UkFP2XNUT0Z7YI0:AfE080XG=?
00@0XgEA0:EgD`2UMe<0[7iL0P2[OE`0102ZO5/0ZW]O0:YkG`2[O6820:QhH@040:9bG02NKUP0WViH
0:QjF@80[h5P00@0/H=R0:]nG@2[OUd0YGYH0P2WO5X0102XP5d0ZH5N0:V1GP2ON5T20:5jF`0409ig
F02ILU<0VG9C09UbD`80WWMH00@0WgQI09efE`2MMUL0Vg9D0P2JLE<0102HKe40UVeA09I]D@2DJU42
09=YD0<0TfUA00@0RVQ?08QTC028I4`0PE]40P1jDC/0101hBcH0MDDa07E5<@1]>RL2064/6P0406Tb
4`23D3<0Pe0c08mQB@80PE]800@0PV9E06eDC@1]E4d0?2TU0P0^7Qh0100m;b`0<RPW038X9`18AT/2
03@g@00403Dh@@0T8RL0928W03L^;`80=RPU00@0CTQ802DO7`0U7al0?cTi0P0`:RX0100a:b/0@C/k
044k>`0a:b/202hX:00402TX9P0d<c40=3<a03<b<080AdI400@0=cHd02XY9`0Z:BL0:BPV0P0`;bd0
1010?cd0<2l]030_;@0];2X202LV900402XY9`0d<c40=3<a02/Z:080:2LU00@0<c8`02`[:@0/:bT0
92<Q0P0X9bD0100];2X09bHT02LV900W9R@203Lf=004030Z:P0X8R80:28R02TS8`80=c4a00@0?cTi
03La<@0g<C40;2HV0P0W8B40100f<300@cdm04<m?@18@T8203li>@0303@^;P0c;Bd0<bd]00<0=S0`
00@0ADA2034`;P0a<2h0<S4_0P0e=380100Y:2H0<C0^034`;P13@T0204E4@P0402XY900W9R409bHQ
02HU8080:2LR00@0;2/V02h]:00^;BP0;2/V0P0Y:2<0100Y:2@0<c8^03<b;P0k>SH203Pg<`0402h]
:@0X9b<0:2LS02/Z9P80<C0/00@0>3@Y03/g;00k=b`0@3`a0`14@3D2044m<P0403di;P0j=R/0>SH[
03`b:080@CL]00@0ASdd04Ln=@17?SD0@c`d0P0l>2l303Xe;`<0>3Lb00@0>CPd03`k=`0l>cL0@3lm
0P16A4D00`1<BT/0Cde>04m=CP0404QBD`<0Ae5B1@16D54<04AAFP0404=<D`13C5D0@daE04=<E@<0
@d]F1@13BeP304A:EP40@dUE0@12B5@10?ooo`00:019Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS
00<0BEMR04QFH@18EV403@18EV4204MEH0H0AUAO00<0AeEP04QFH@18EV401018EEh204QDH0<0B5AR
0`18E6@304UEH`0404IBGP13D5T0@e1I04ICG080AeAL00@0?4UA02hk@@0^>d402a<E0P0:4a80100;
4@l03@l<00d?300?3PX201<>2P0401H?2@0G40X05a0:00l?0`8051@:00@051@<014@300A40`04@l@
0P0D4QD0100C4AH03@XA00d:4@0=4a<200h@2`0401TE300b91L0<R@G04/h:@<0FTDd0P1G@S40101^
@2H0M4H/07A6;01jC38207i@=P0408IH?P2AHdT0TF=909MYC`80UVQ>00@0VfeF09a^E`2LKUL0VfmF
0P2NLUT0102UO600Z85R0:R1HP2QNU/209E^C`0409]aF02CId`0TfM<09U]DP80V6]>00@0R5/n08mP
@P2?H480WVmA0P2HJTT0102EHd80VVQ709YXA`2MJdX209]YB00409AR@@2>G3/0SU`k08mM?080TEln
00@0RE/g08YL>02:G3P0QUPf0P1mDBh0101iC2/0N4d[07Q=:`1dBBT206m4900407M7<@1lC3H0O4`f
0751:`80L48[00@0M4H_07U<=@1iC3D0PeHo0P1iCcL308UK@@0408eOA@2BIT/0TVI;08eTB080R5m3
00@0S6E809QaE02HLE@0WVmA0P2TMEL0102YNeX0Z7YI0:QjF@2TMU@20:5cC`040:=eD@2UMe<0YGMC
0:anG080ZgeL00@0ZWaK0:YkG`2ZNel0ZgaR0P2XN640102RLU`0WViH09i^F02XNUT20:n1H0040;63
HP2[OUd0ZgiM0:EjF080YgaJ00@0Z81M0:V1GP2YPEh0WgQI0P2QNU/0102NMeP0VG9C09UbD`2ILU<2
09igF00409mhF@2MMUL0WGIG09]bE080VW5C00@0V6mA09I]D@2FKE40U6YA0P2CJE0309=YD@0408YX
C`28I4`0R6A<085KA080NU4k00@0N4/f07E5<@1eAC40KCXW0P1Q;1X0101Y<Q<0Pe0c08=@<`2?HDT2
085KB004089RE@1]E4d0KEA=03`Y9@80;QhN00@0?Bl/038X9`0b:2L0B4I;0P0d=d00100e>440928W
02@R9`0g;Rl203HX9@0404i8B00U7al09AlO03li>@80<2XZ00@0<B/[044k>`11>c/0<B/[0P0^:2P0
100Y:2H0=3<a03@c<@0c<S0204M6A00403Lf=00Z:BL0:RTW02TX9P80<2l]00@0@3lm030_;@0`;bd0
;B`Z0P0W9R@0100Z:BL0=3<a03@c<@0[:RP202PW9@0403<b<00/:bT0;2/Y02@S8@80:2LU00@0;B`Z
02LV900W9R@09bHT0P0g=S@0100`:RX0:28R02PR8P0Y8b<203La<@0403li>@0g<C40=c4a02`V9P80
9b4Q00@0=S0`04<m?@13?Cd0B4920P0o>CT00`0d;Rh0<bd]03<];@0303H`<00404E4@P0a<2h0<C0^
038a;`80=C@b00@0:BPV034`;P0a<2h0@d900P15A480100Z:B@09bHQ02LV8@0V9B0202PW8P0402`[
9P0^;BP0;RdX02`[9P80:BPS00@0:BPT03<b;P0c<Rh0>cXf0P0h=c<0100^;BT0:2LS02PW8`0[:RH2
034`;00403Pd:@0k=b`0>cL/040l<@<0A40e0P11?C80100m>Bh0>SH[03Xf:`0l<RP2044g;@0404Hm
=017?SD0Ache04<l=080?3P_0`0j=Bl303Pg<P0403Th=00l>cL0?3/g040o?@80ATA500<0C4Y;04m=
CP1?CDh01018DU<304MADPD0AU1A3014DEX01013C5<0@daE04=<E@13C5D304=;EPD0@d]H0`14BUH1
04=9E@40@TQD0@3oool002P0BEMR0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ
00d0B5IQ0P17EF0604IDG`0304MEH018EV40B5IQ00@0B5EN0P18E60304QDHP<0B5AT00@0B5IS04UE
H`19EF<0Ae=O0P16De`01019EUh0B59K04QBF`0iA4X202Pc>@0400/D4`0<4Q80318B00`@3`803@l<
00@03`h:018?2P0B3`X050l;0P0G40X0100=3@<04A47014A1`0C4`/2010?2`04014@3P0D4QD0518E
01<A5P803``C00@03@l>018A3@0B4@d08QT@0P0j:1`0101@>BT0GD<b05e3<P1OA3<205]1<P0406m1
9`1gBBl0MdT_07mA=`80Q5Hl00@0S5i409EWC@2EIdd0V6Y@0P2EIdd0102KKEH0WW1I09i`F@2PM5/2
0:=gGP040:QoH`2[Q6D0ZhAU0:=lG@80V75B00@0VFmF08iUB@2>IDT0UFU>0P2EJD`01026FC`0REXl
08UJ?02GJ4X209AVA@0409ET@`2HIdH0V6M609UXA`80UFA300@0SUdl08]J>@2;FST0SE`k0P2@Gch0
101oDS40P5<b081C<P1lCbh207A7:00406i38`1`A2L0L4@W06m39P80K40U00@0L40Z0751:`1a@B/0
JcdV0P1a@b`0101dAc00Ndhg07]>=`26G4@2081F?P04089D>P23Ec`0PeLl08IJ?`80R5m300@0RF14
08URA@29HTD0SfQ;0P2FKe800`2LKDl0XG9D0:5bE0030:EgEP040:5cD@2NL4`0WW1<0:1bCP80XgEA
00@0ZG]I0:QjF02XNUP0Z7YI0P2ZNed0102[O680ZWYS0:YjH`2VMV020:9bG0040:MhFP2^Of40[WmQ
0;:3I@80ZgiO00@0Y7UI0:=hF02SN5P0XGQH0P2OMUH0102LMEH0X7UJ0:1iFP2QNU/209igF0040:1i
FP2TOEh0Y7eN0:AmGP80XGYK0`2JLE<0102HKe40UfiB09M^DP2EJe8209=YD004095WC`2@ITh0T6I>
08QVC@80Qf=;00@0PE]407]A>`1kDC/0NTdh0P1iBCD0101a?R/0IS<P06Hc801Q<Q@207M:;@0408QN
A@21FdH0PE]6085PC`80Me]=00@0D3P/03XT6@0j91T0=bPS0P0g;B/0100m>ch0ADQ?04E8C`0`<c`2
02TW;00402`S900b9240<R@Q04M1@@80=2h^00@0>C<c03@^;P0d;Rh0<bd]0P0e;bl0100l=SH0;2HV
02`V9P0a<2h203Dd<P04038a;`14@d40A4=103<b<0809bHT00@0:2LU02`[:@0/:bT0@d900P0_;R`0
100W9R@092<Q02@S8@0];2X203He<`0402`[:@0Z:BL0:RTW02`[:@80:bXX00@0:RTW02l^;00_;R`0
<C0^0P0U9280100O7Q`0:RTW02XY9`0h<S8202hX:00403@^;P14?Sh0A3hn03li>@80:bDU00@0:B<S
03Xd=00j=3@0=S0`0P0l=SH0100o>CT0>S@d03Xd=00f<30203/e=@0404Do?`1=AdL0CDM70491?`80
;B`Z00@0:RTW038a;`0b<Bl0<c8`0P0h=cD0100j>CL0<2l]030_;@0g=S4203Xi=00403He<00Y:2<0
:BPS024P6`80:bXU00<0?C`g04Y9A01:BD@00`0e<Rd00`0d<B`0<BhY034^:@0302h[9P04038_:P0f
<bh0=S<^03di<080?c/b00@0@She04<o=P13?cH0@clf0P12?SD0100o>c80?SXa03hj<@0o>30203di
<00403`g<@0j=c00>SL`03/h<`80>cXe00<0@3lk04=2?P13@Sh00`12A4401013AD@0ADM604E7AP19
BT`204e>D003059CE`1EEUX0EEIJ00<0BE=E1018DU@504MAD``0A55J00@0@daC04=<E@13C5D0@daE
0`13BeH504=;F0<0A4YF0@13BED10498E040oooo000X04UGHP<0B5IQ1@17EF0=04QFH@80BEMR1P1:
F6<00`19Ef80B5IQ04QFH@0=04QFH@80AeEP1P16E5l00`17EF00B5IQ04QFH@0404QEGP80B5AP0`18
E68304QDI0<0AeER00@0BEEQ04YGH01:Ef00C5IO0P14CeD0100`>Ch07BLY01dW:@0=4a4200`B4004
00`@3`0=3``03@l<00l>2P804Ph;00@050l;01D@300E40`030`20P0?3`L0100A40/040l;010?2`0@
3`d2018@4@04018@4`0@3Q<040hC00l;2P8071@A00@0<R4I04H^8P16;R80F3/[0P1Q@S00101S@S40
HD0a0650<@1cAB/207a>=00408EG?@2:G480RUa2095SB@80V6Y@00@0VF]A09EWC@2EIdd0W6iG0P2Q
Le`0102UNF00Z7aS0:QlH`2[PVH20:f6I`040:N0H@2NMeP0WWMH0:1fG@80U6]?00@0V6mC09a`D`2L
L5<0SV540P2;GSl00`2GJ4X0V6Y909QZB@0309=VA@04095T@`2<Gch0S5ln08II>080Q5Lf00@0QUTh
08UL>`29G3/0M4P[0P1gBbh0101dB2d0JclT06/o901V>R4206Tm900406e0:@1_@R/0Kd8[07Q8<P80
L40Z00@0KT0Y07I8<@1fB340LTD^0P1cARl0101lDSX0MT`d07I<=01lD3D2081D>@0407iE>@1lDcL0
O5<g07iG>P80QUm20`2:IDP0102IJ4X0WFa>09e/CP2RLE420:=bDP040:5`C`2OKd/0Wfm;0:1`C080
XW9>00@0YGMC0:EgE@2UMeD0YgUH0P2YNU`0102[O600[7eS0:amH`2ZNV<20:MgH@040:IgF@2/OEl0
[7eO0:enH080YWUJ00@0X7EE0:5fEP2QMUH0X7MG0P2OMUH0102FKe00WGIG09efE`2QNU/20:1iFP04
0:9kG02UOUl0YGiO0:EnG`80XGYK102JLE<209Q_D`0409I/D`2BJ4l0TVQ?08iTC080S69:00@0Qf=;
08IPB@26H4T0PeY40P1mDcd0101mD3/0ODhj07e>>P1fAS8206ll:@0405`g701V@BL0IT4W07aH?P80
PUi400@0Q61808QVC@28ITd0LDhh0P1B;aT0100f9Ad0?34]03`a;@0a;Bh204m@E@0402X]=00[:Bh0
:bT^02HN7080<bPT00@0?3Hf048l?012?3`0=2h^0P0f<300100b;2`0:B<S02TS8`18@T8202`V9P04
03@c<@0a<2h0<C0^02h]:`80@3lm00@0<2l]02LV900W9R@0;2/Y0P0Z:BL00`14@d40;bh/02l^;003
02<R8004030_;@0f=C<0=SDc02d/:P80;2/Y0`0a<2h0100[:RP0:BPV02TX9P0X9bD201hM6`0401/J
600V9B<09RDS03La<@80=Bl_00@0?CLg04Do?`15?cl0?SPh0P0`:RX0100e;bl0AT1004I0@010>SX2
03Tc<`<0<bd]00@0=c4a03dg=`0m=cL0AClo0P1<ATH0100g<c80<bl^03<_;P0f<S4303di>080=S8a
00@0;2PW02<O7P0S7ah0;bhY0P0/:bH0100Z:B@0;B`W02d/9`0j>C@204E4?`040491?00i>3<0>CPc
038^9@80:bLN00@09R8I02PT6`0X91/0<RhU0P0m>C001014@3L0AT8i04I2>@14@CP204=0=`04048o
=P10?C@0@3dd03ll<`80?S/b0`0m>S40100m?3P0>CXe03Tj=@0f=S@203Hh=@0403Xl>`0o@d@0?d=4
04E9BP80Ade=00@0Be1C04Y?DP1:Ce80BTmB0P1;D5@0101=DUP0D5EK051EF`1CEf0205EIHP<0BU=L
0`19DU/604QAFP`0A55J00@0A4eD04A=EP14CEH0A4eF0`14C5L504A<F@0504A<E`13BeH0@d]F049:
E@11BE@00@3oool002P0BEMR0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00d0
B5IQ0P17EF0604IDG`0304MEH018EV40B5IQ00@0B5EN0P18E60304QDHP<0B5AT0`17EF801019EF40
BUMP04YGH01<EUl204A?E@04030i?P0M9bT07BLY00dC4@80318@00@0310?00d?300=3``03`h:0P0B
3P/0100D3`/05A0<01D@300<308200l?1`04014@2`0@3`/040l;010?3@804Q0A00@04Q0C010>4`0@
3Q<03`/:0P0L5140100b8AT0ARhR04H^8P1H>b/20652<00406=2<@1Q@340HD0a07=5:`80O4hd00@0
QELm08YL@P2:G480TF=90P2HJU00102IJe40UFM=09EWC@2LKUL20:5cG0040:EiH02XO6<0Z7aS0:^2
IP80[HIW00@0Yh1Q09igF02NMeP0X7IM0P2DJdl0102HKe<0W71C09a`D`2>HD@208]N?`0309MXBP2H
JTT0V6Y900<0TfI500@0TFA308aO?P2<Gch0QUTh0P24EcH01026FCP0RE`k08UL>`1dB2/207M;;P04
07A8;@1[?b@0JclT06Hj8@80JCdT00@0KD0Y06m2:`1_@R/0N4Pb0P1`@2X0101^@2T0MTPa07I8<@1b
ABh207=6;`0407aB>P1fC3@0MT`d07a@=@80P5@i00@0OUDi07aC=`1lDcL0OULj0P26Gd8308YUB004
09UXBP2MK4h0WFa>0:9aD@80Xg9B00@0XG1?09m_B`2OKd/0X71<0P2RLTh0102UMe<0YGME0:EgE@2W
NEP20:UjG0040:]lH02/OF<0[7eS0:YjH`80YgMQ00@0YWMI0:amG`2/OEl0[GiP0P2VNEX0102PMED0
XGIF0:5fEP2PMeL209mfEP0409I_D02MMUL0WGIG0:5jF`80X7UJ00@0XW]L0:EnG`2UOUl0YGiO0P2Q
NU/409YaD`80V6mC00@0UVaC099XC`2BJ4l0SVA<0P2<HTX01027Hd/0QV1908IPB@23FT@207eC?@04
07e@>`1mCSX0ODhj07I6<P80Kc`Y00@0G3LL06I19`1V@BL0O5Pn0P22GT@01024H4P0R6I=08QVC@1a
CSP2058_6@0403HU7@0l<Bd0?34]034];P80Ce1E00@0:Rdd02/Y;P0[:Bh09QhL0P0c:2@0100l=SH0
@S`l048l?00d;Rh203H`<004038/;00Y8b<0:B<S04Q2@P80;2HV00@0=3<a034`;P0a<2h0;Rd[0P10
?cd0100`;bd09bHT02LV900/:bT202XY9`0304A3@@0_;R`0;bh/00<08b8P00@0<2l]03He<`0f=C<0
;B`Z0P0/:bT3034`;P0402/Z:00Y:2H0:BPV02PW9@807QdK00@06aXH02HU8`0V9B<0=c4a0P0e;bl0
100m=cL0AClo04Do?`0n>3P2030Z:P0403D_;`16@400AT10040j>P80>C<c0`0c;Bd0100g<C40?CLg
03dg=`15?cl204a6AP0403Lc<P0c;bh0<bl^03Hb<@<0?CTh0P0f<S40100/:2L08alN02<O7P0_;RT2
02`[9P0402XY900];2L0;B`W03Xi=080AD@o00@0@T4l03Th<`0i>3<0<RhU0P0[9ah0100V8QT0:2@K
02PT6`0b;RD203di<00404A0=`16@ST0AT8i04A1>080@d0g00@0@Slf040m=010?C@0?c`c0P0n>c83
03dj<@0403dl>00i>SD0>CXe03Hf=080=SPe00@0>S`k03m3A00o@d@0ADU:0P17CDd0101;D5<0BTmB
04Y?DP1:Ce8204]@E00404eBF01@EE/0D5EK05=GH080EEUR0`1:De`304UBF`H0B55J3014DEX01014
CE@0A4eF04A=EP14CEH304A<E`D0A4aI00D0A4aG04=;EP13BeH0@TYE0459E0010?ooo`006019Ef8<
04QFH@@0BEMR0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00d0B5IQ0P17EF06
04IDG`0304MEH018EV40B5IQ00@0B5EN0P18E60304QDHP<0B5AT00@0AU=S04MEHP17EF80BUIR0P1>
F680101<EE`0?TI903i6B@0V;S0201@I700400hC3`0=4Ph03A8>00h@3@<03Ph<0P0@3`/0100C3``0
510=01@@3@0=30L200l>2@04014@2`0@3`/040l;00l>30803`d>00@040h?010>4@0@3Q4050X90P0X
6AH01012:b<0DcDZ05<e:P1P?2`206I1;`0406U2<@1W@340Id0a07Q:<080P58h00@0QeTo08aNA02<
GT@0TVA:0P2IJe40102KKE<0UfU?09MYC`2NL5T20:AfG`040:IjH@2WNf80Yg]R0:QoH`80ZhAU00@0
ZH9S0:AmGP2TOEh0Xg]Q0P2HLE@0102LLeL0WWEG09ieE`2BITT208eP@@0409AWB02HJdX0V6]:095V
A080T6E300@0S64o08ML>P27G3X0PEHd0P1nDc4307aA;`0406a2:@1aAbh0LDL^0755;080JC`U00@0
I3LR06Xk9`1Z>bL0LT<a0P1fAcD01022CcX0MD8]07E2;@1dA2h207a<=P040794;@1/?bP0K3lX0795
;P80Kd8[00@0Ndld081D>@20E3T0OE4f0P1cBRh0101eC300OePk07mH>`23G3l207aG>P0408mN@02E
I4H0UFA609e/C080XW5A00<0XW5@0:5aC@2QLDd00`2QLDd0102TMU80YGMC0:EgD`2VN5H20:UkFP03
0:]lH02[O680ZgaR00<0ZWYS00@0YWMK0:UjGP2YNUh0Z7UM0P2PLeH0102MLE@0XWII0:9fF@2UO5h2
0:MnH00409I_D02LMEH0W7EF0:1iFP80WWMH00@0WGIG09mhF@2ON5T0WgQI0P2LMEH0102IL580VW5C
09YaD`2KLU@209YaE@0409M]E02AIdh0TFM>08aRBP80R5i600@0QV5708QPA`28H4L0Q5a30P1oECd0
101oDS/0PE<l085C?01lCSL207E7<00405e1:@1G>B40EcTQ06i?<`80QV=700@0RVI609A[B`2DJd/0
SF900P1_A240100n:Ql0>b`U03//9@0/9b@20491AP0402h_=00W9bT09bLY02dU8`80?C8^00@0<bd]
048l?012?3`0>38b0P0e;bl0100_:BT0:28R02PR8P1:A4@203<];@0402l^;00X9bD0:2LU02XY9`80
?3/i00@0=3<a02XY9`0Z:BL0<C0^0P0Z:BL00`0o?S`0<c8`03<b<00302PW9@<0<S4_00@0:RTW02d/
:P0];2X0;2/Y0P0^;B/0100V9B<081lM020O7@0P7ad201hM6`04028Q7`0^;B/0;Rd[02hX:080>38b
00@0?3Hf03Pb<P0h<S80>cDe0P14?Sh01015?cl0?SPh03hh>013?Cd203<];@0402TS8`0`:RX0<2XZ
03Xd=080>cDe00@0=c4a03H`<00f<300:bLV0P0l>3L0101:ATD0BDE404U5A00j=SD202XV9@0402DQ
800[9bH0:bLV02d/9`80=C@_00@0?chi04=2?@13@Sd0?3/f0P0];2L0100R8A`07QdH01hM600e;RH2
02lX800402/T700^9al0;RLO03Pa:@80@S/c00@0ASlg04M0>017@3P0Ad@o0P14@C`00`10?CP0?3Td
03`i=00303Xg<P0303/h<`0l>C@0?3Td00<0=cdm00@0>Ci103a4AP0lA4H0@d]>0`16Ce@204=>DP03
04UDF018DeL0B5=G00<0Ae9H00@0B59K04UCG@19Ded0BeEO0P1<EV000`1;Deh0BU9M04YBG@0304YB
G@<0BE5L0`18D5/<04AAFP0404A=E014CEH0A4eF04A=EP<0A4aG1@14C5T01@14C5L0@d]F04=;EP13
BeH0@TYE0040oooo000H04UGHP`0B5IQ1019Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS00<0BEMR
04QFH@18EV403@18EV4204MEH0H0AUAO00<0AeEP04QFH@18EV401018EEh204QDH0<0B5AR0`18E6@0
1014E6<0AeAT04MDI01;Ef<204eGH00404I=D`0d>C`0=3Tl01dQ8P803A4@00@0418=00lA300?4@`0
3Q0;0`0>3P`2010?3@04014@3P0B4@l04Q4?00l>2P804A0<00@04Q4=014@300A40`03Pd;0P0<2`T0
100>3@/04@l@014?400M3`h203@R7P0404de:`1L>b`0G3//06@m;080JT4_00@0KD8a06]2<01[@S00
ODle0P23EC/01028FT00Ree308]M@`2AHdT209U[D@0409e_E@2KKE<0VfeC0:9dG@80YGMP00@0YGUP
0:=gGP2SMeh0Y7]O0P2YPV<0102ZPf@0YWmP0:IoH02MME/209I_DP0409QaE02IL580VG1B08eTAP80
R5dm00@0S651095T@`2AI4<0T6E50P2=HT801029GSh0Q5Ti08AI>@1oE3@207Q=;@0407979`1^@b<0
KT<S06Tm9080Kd8[00@0L48[06/k9`1[>bL0JCHS0P1_?2/0101h@c<0O4Li07a7>@1mAS8207@m:@04
07]4<021CCP0PDdh07a8<`80NDHa00@0O4Td07];=@1kBcD0O4dc0P22DcT0101nCcD0LT@Z0794:P1b
AR/207iB=`0407mF>P1hCc<0N4lc08IC=P80SU/n00@0VFI70:5^C`2QKTl0YG=B0P2ULe030:=aCP04
0:AgD02UMe<0YGMC0:MiE`80Z7YI00<0Z7UK0:QiG@2XNEd00`2YNV00102TMET0ZGYN0:UjGP2YNUh2
0:AgFP040:9fF@2VNUd0YWYM0:ImG`80YGaN00@0WGIG0:9kG02RNe`0XgaM0P2NMeP0102JLe@0W7EF
09aeEP2MMUL209]dE@0409YaD`2KLU@0Vg9D09acE@80Vg9F00@0V6iE099XC`2BJ4l0Rf590P27GDD0
1028H4H0RV1708YPA`26G4<2085G?P04085E?022ESd0PUHm081D>`80NThe00@0IT`e05@h801D>200
HdDY0P25HT@0102?J4L0V6a709Q/A`2JKD@208MH;P0404Xc8`0d8aT0=2<I03@[9P80<2h_00@0?Ci2
02PX:P0X:2X0=2l/0P0o=S40100]9bL0=c4a03La<@10>SX2030Z:P0402/U9@0`:RX0<2XZ048l?080
?SPh00@0:bXX02HU8`0V9B<0:RTW0P0k>SP0100j>CL0:bXX02/Z:00`;bd202PW9@0403Dd<P0j>CL0
>STg034`;P80;B`Z00@0<C0^02`[:@0/:bT0:2LU0P0];2X0100^;B/0<S4_038a;`0[:RP3024P7P<0
81lM0P0W9R@0100b;2`0>S@d03Xd=00k=CD203D_;`0403Tc<`12?3`0@S`l03hh>080<2XZ00@0=S0`
02lY:@0_:BT0<2XZ0P0j=3@0100n>3P0=c4a03La<@0_:BT202dW9`0403D_;`12?3`0@S`l04Q2@P80
@C/k00@0=2h^02lY:@0_:BT0=S0`0P11>c/0100f=C00;bhY02l^:@0`;bX203He<004034`:`0W9R40
9bHQ02TX8`80=C@_00@0;BTP02l[8P0_:b80<blV0P0h=2/00`0n>S40@Cdd044m=00304<o=P0403hm
>@0l>cL0?3/g03Th=080=SDa00@0=cHb03Th=00i>3@0?C`h0P0o?SX0100lA4L0?4I803a6B00oB4d2
04A?D`0404UDFP19EU`0BEIL04EBFP80@4mF00@0AU=I04EBFP15DUX0A55I0P14DEX01013D5T0A51L
04A@G014D5`204EAG`0304]CH01:DUl0BU9O00<0BU9O0`19DEh304Q@G@`0A55J00@0ADiE04E>E`15
CUL0ADiG0`15CEP504E=FP0504A>F013CEL0@deG049<EP11BeD00@3oool001P0BEMR3018EV4404UG
HP<0B5IQ1@17EF0=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0=04QFH@80AeEP1P16E5l00`17
EF00B5IQ04QFH@0404QEGP80B5AP0`18E68304QDI00404ADH`17E6@0AeAT04]GH`80CEMP00@0ATeC
03@i?00d>C`07B4R0P0=4A00100@4Pd03a4<00lA300>40/300h>308040l=00@04A0>018A3`0B4@l0
3`h:0P0A40`0100B4@d04A0<014@300>3@/200`;2@0400h=2`0A3a004@l@01d?3P80=28N00@0CCD[
05`k;01L>b`0I3d/0P1Z@Bl0101]@S40Jd8`06]2<01mCcD208=E>`0408QJ@02;GD<0Ree3095SB@80
VF]A00@0WFmE09]]D`2KKE<0XWAM0P2UMf00102UNF00XgMN0:=gGP2TNel20:V2H`040:Z3I02VOf00
YWmP09eeF`80UVmB00@0V75D09U`DP2IL580SFA60P28GCd0102<HD40TFA3095T@`2@IDD208eR@P04
08UN?P24FCT0Q5Ti07mD=080N4d]00@0LTLW06i38`1^@b<0JCdT0P1_@R/0101`@R/0Jc/W06/k9`1Y
=R<206ll:`0407Q3<`1lAcT0O4Li07e6<P80M3dY00@0Nd@`085=>021CCP0O4Pc0P1iAS40101lBC@0
Nd/e07];=@1lCC<2089C>@0407i?=@1bA2X0LT@Z0796:`80OU8g00@0OeHj07Q?<`1hCc<0QU<f0P2>
Fch0102IITL0XFi?0:5^C`2ULe820:EcD0<0Xg5>00@0Y7M@0:EgD`2UMe<0YgUG0P2XNUT00`2XNE/0
Z7UM0:QiG@030:UjH0040:AeF@2YNUh0ZGYN0:UjGP80Y7MJ00@0XWII0:IjG@2VNUd0YWeO0P2UO5h0
102MMUL0XW]L0:9kG02SO5d209igF00409YcE02LMEH0W7EF09efE`80VgAE00@0VW5C09]bE02KLU@0
W7=E0P2KLUH0102HKUD0TVQ?099XC`2;HDT208MMA@0408QPAP2:H4L0RV1708IL@`80PELn00@0PEDl
089F?@22ESd0P5@k0P1jCSD0101VC3D0E3PP05@h801SABT208ERA00408mXA`2HK4L0V6a709Y]A080
QeP^00@0BS<S03@S6@0d8aT0=2/V0P0`;Rl0100m?T80:2PZ02PX:P0d;b`203lf<@0402dW9`0g<C40
=c4a040j>P80<2XZ00@0:bDU030Z:P0`:RX0@S`l0P0n>3P0100[:RP09RDS02HU8`0Z:BL203/j>004
03Xi=`0[:RP0:bXX030_;@80:2LU00@0=C@b03Xi=`0j>CL0<C0^0P0];2X0100a<2h0;2/Y02`[:@0X
9bD202d/:P0402h]:`0b<Bl0<S4_02/Z:0<08B0N0`0P7ad202LV9004038/;00j=3@0>S@d03/e=@80
=Bl_00@0>C<c048l?012?3`0?SPh0P0`:RX0100f<300;bTY02lY:@0`:RX203Xd=00403hh>00g<C40
=c4a02lY:@80;BLW00@0=Bl_048l?012?3`0B4920P11>c/0100d;Rh0;bTY02lY:@0f<302044k>`04
03He<00_;RT0;bhY030_:P80=SD`00@0<C0[02LV8@0W9R40:BPS0P0e=2l0100]:B00;b/R02l[8P0c
;bH203Pd:`0303hj<@11?C@0@Cdd00<0@clf00@0?Sdi03`k=`0l>cL0>CPd0P0f=C40100g=S80>CPd
03Th=00m?3P203ln>P0403a4A`0lATP0?4I803m8C@80A4mC00@0BEAJ04UFG019EU`0AE9J0P10CeH0
1016DeT0AE9J04EBFP14DET204AAFP0404=@F@14D5`0A51L04A@G080AE5O00<0Be=P04YBG`1:DUl0
0`1:DUl304UAGP<0B51M3014DEX01015CUD0ADiG04E>E`15CUL304E=F0D0ADeJ00D0A4iH04==E`13
CEL0@TaF045;E@010?ooo`006019Ef8<04MEH0@0BEMR0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`03
04UGHP18EV40B5IQ00d0B5IQ0P17EF0604IDG`0304MEH018EV40B5IQ00@0B5EN0P18E60304QDHP<0
B5AT00@0AEEU04EEI015EF@0BUIR0P18DU/0100lA4L0:2`]02P/;@0F61D200h?2P04014B300@4@/0
414;00l@2`803Pl:00@03Ph<00l?3`0?3`l0410@0P0A4A43010?3@<04Q4?00@03`h<00/:200;2PP0
3Pd;0P0D4a40100W6AH0?b`U03l/9@1G?34205ln;@0406@n:`1/@Bh0K44^06m4<@80KT<`00@0P58h
08IH?P26F3h0RUa20P2<GT@0102BI4X0VfeC09]]D`2PLUP209maE`040:EgH02VN640YWQQ0:AhG`80
XWIM00@0Y7]O0:V2H`2YPV<0ZH9S0P2TOEh0102IM5T0VWEH09YeF02LMEP209M`D@0408mVB02;HT80
Rf9208iS@`80SfA200@0Sf9508]N@@2;GT40QEPk0P21E3L0101lCc80MDP[07E8:`1/?b8206Dh6`04
06Hf7`1]?BH0KCdV0790:@80LS/W00@0LSTX07Pm;@1h?Bd0Od4d0P20@SD30811<00408Q;>@2;CS`0
Rdhl08aA?`80SE9000@0R4ll08mF@`2?ET<0PD/c0P28DSX01027DCT0NdL_07]7;`1kAbl208AB>@04
08ME?020DCL0P54g089?<P80R5Dh00@0TEho09YWB02JIdP0X6i=1@2SLDh0102SMTl0Y7M@0:AgD02V
N5@20:IhEP040:EgEP2UMUX0YGIJ0:MhG080Z7UO00@0XgAJ0:YkH@2ZNf40[WmU0P2[OF<0102YOF80
Z7aQ0:QlH@2RNEd209acE`040:=lG@2YPV<0ZH9S0:Z3I080XgaM00@0WWMH09mhF@2ON5T0XGYK0P2P
NEX0102KLU@0W7=E09acE@2MM5H209acE`0409Q^E@2CJE00TfU@08eSB`80REm700@0S69908eSBP2=
HdX0REm60P22F3l0101oEC`0P5Hm081F?@1nE3/207Y@=`0406a@>01N@2P0GT0X05ho8`80NULi00@0
SVE509EYAP2EJDH0VVe60P2CHcd0101J@Bd0=R<E03HS5@0m<R`202LV900404Q9B`0g=cL0=cLg03@_
;080>2lZ00@0:B<S02dW9`0]9bL0AClo0P0_:BT0100[9BD0<bd]03<];@0h<S8204<m?@<0:bXX00@0
;B`Z03Th=P0i>3H0?Sdk0P0Y:2H0100[:RP09bHT02LV900];2X20450?P0403Lf=00/:bT0;2/Y02h]
:`<0:bXX0P0_;R`0100b<Bl0=C@b03Dd<P0X9bD201XI5`0401hM6`0V9B<09RDS02XY9`80;bh/00@0
@3Xj03`f=P0l=SH0?CLg0P10>SX0100i<c<0;BLW02dW9`0Y8b<202dW9`0403<];@0k=CD0>cDe04Do
?`80Ad5100@0?CLg034[:`0a:b/0;BLW0P0a:b/01018?d00?SDf03he=P0d:b`2030W:00403@[;012
>CX0@STj04e4A@80C4=400@0@D0k02h]:00^;BP08B0K0P0W9R40100];2L0:2LR02PW8P0T8ah202HU
800402DT7`0/:bH0;2/V03He<080?Sdh00<0@D0k0491?012@C`00`12@C`403Pg=@80>STg00@0?C`j
0491?`12@Cl0AdI40P19B4H01016D580ADiC04E>D`15CU<204E@EP0404EBF017E5`0AeAL04IEG080
AE=L00@0AU=I04ICF`16De/0AE9J0`15DU/504EAG@<0BU9M0`19DE`604Q@F``0A55J00@0ADiE04E>
E`15CUL0ADiG0`15CEP504E=FP0304A>F013CEL0@deG0080@TaF0@3oool001P0BEMR3017EF0404UG
HP<0B5IQ1@17EF0=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0=04QFH@80AeEP1P16E5l00`17
EF00B5IQ04QFH@0404QEGP80B5AP0`18E68304QDI00404AEI@16EVD0AUIU04QDH080A4eF00@0<SLk
01dO7P0M7ah04Q4=0P0@3`X0100B4PX04A0;014@2`0?40X200h?2P0400h>300?3`l03`l?00l@4P80
414C0`0>3@/0100A40h04a8@01<B400@3`/200d<2004014@300J6AD06QTE034T7P80AcH^00@0FT<e
060o;01P?b`0HcdX0P1[@R`0101`Ac40KdH`06m6<021DcT208QJ@00408iPAP2AHdT0TF=909QZD080
X79H00@0XgEK0:5cF@2QLeT0YgUR0P2XNV<0102UNF00YWYQ0:IjH@2ZPFD20:f6I`040:R1HP2ON5T0
WgQI09MbE`<0X7]N0P2GL540102@JDX0TFQ8095XB02AJ4P208mT@P04091N@`29Ec`0RELl089@=@80
OD/`00@0NdT^07E3:01e@bP0KC/P0P1V=1T0101Z=a`0LcdS07<m8`1j@R/207i1;0040852<028ACD0
R4De08Y5=P80QT4b00@0S4@f09E@@@2ED440W5M80P2FDd<0102OG4`0XEm?0:5OC`2FE4@20:1NCP04
08i@>`2DEd80U5M209EH@`80Re4k00@0Qddg08YB>`2:DS/0SEDn0P2:E3`01025Cc<0Qe4e08MA=@2<
EcP2095L?@0409MR@P2MJ4H0WFQ60:9]B`80Y6m=00@0X7=:0:9eCP2RMDh0Y7IB0P2TMU@0102SME@0
Y7EG0:AeE`2VMe/20:UjGP040:EfG02ZNf40ZW]Q0:amH`80Z7YP00@0Yg]P0:QlH@2XO640XWUM0P2M
M5P0102MMUL0YGiO0:EnG`2ZPf@20:N0H@040:=lG@2TOEh0Y7eN0:IoH080YGiO102MM5H209acE`04
09U_EP2DJU40U6YA08mUC@80SF=;00@0TFE:099VB`2BIT/0SV970P25FCh307mF>P0407eD>01iD3@0
NE0d06m==080L4`b00@0Gc/Q06U4:@1YA2T0Pe`o0P2=I4H0102CJ4P0T6=2091S@P1cEd4204Ha8004
03lb:@0Y9B80:BDR0410@080C4a<00@0:RDQ034Z900a:R@0:28R0P0]9bL01011>c/0<R`/038/;00_
:BT202dW9`0403<];@10>SX0@3Xj02TX9P80<2l]00@0;Rd[03<b<00c<S00@D0n0P0X9bD0100Y:2H0
;Rd[02h]:`0Z:BL204E4@P0403Dd<P0S8R008b8P02XY9`80<2l]00@0=3<a038a;`0b<Bl0:2LU0P0Z
:BL0100M71X04a8@01<B400R8Al203He<`04040o?@15A480ADA2044k>`80?CLg00@0?cTi048l?012
?3`0=c4a0P0Y8b<0100Z92@0>38b03Pb<P0k=CD204Do?`0404Y4A012?3`0@S`l03La<@80<bd]00@0
=c4a03/e=@0k=CD0A3/l0P0e;2d0100`9bP0<BPY034X:@0f;Bh204Ln?`0404i5AP11>3T0@CPi02LV
8@80;bhY00@0<c8]02h]:00^;BP092<N0P0R8A`0100X9b80;bhY02l^:@0c=C@203Hh=`0403Xl>`0n
@3l0?T0o0412@@80?d5000@0?3hm03Xl>`0j?3/0?Cdm0P0o?cl01011@D40A4A404A4A017AdL204Y:
BP0304a<C01=CDd0CDe=00<0BE5D00@0Ae1E04E>D`15CU<0@diD0P12CeD00`14DET0AeAL04MDG003
04IAE@0404MBF017DEX0Ae5J04MAFP80Ae5K00<0B59L04MAF`17DE/00`19DU/404QAFPD0Ae1I3014
DEX01016CeH0ATmH04I?F016CeP304I>F@D0ATiK0`12CeP1045>E`40@4eF0@3oool001P0BEMR3017
EF0404UGHP<0B5IQ1@17EF0=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0=04QFH@80AeEP1P16
E5l00`17EF00B5IQ04QFH@0404QEGP80B5AP0`18E68304QDI00404AEI@16EVD0AUIU04QDH080A4eF
00@0<SLk01dO7P0M7ah04Q4=0P0@3`X0100B4PX04A0;014@2`0?40X200h?2P0400h>300?3`l03`l?
00l@4P80414C0`0>3@/0100A40h04a8@01<B400@3`/200d<2004014@300J6AD06QTE034T7P80AcH^
00@0FT<e060o;01P?b`0HcdX0P1[@R`0101`Ac40KdH`06m6<021DcT208QJ@00408iPAP2AHdT0TF=9
09QZD080X79H00@0XgEK0:5cF@2QLeT0YgUR0P2XNV<0102UNF00YWYQ0:IjH@2ZPFD20:f6I`040:R1
HP2ON5T0WgQI09MbE`<0X7]N0P2GL540102@JDX0TFQ8095XB02AJ4P208mT@P04091N@`29Ec`0RELl
089@=@80OD/`00@0NdT^07E3:01e@bP0KC/P0P1V=1T0101Z=a`0LcdS07<m8`1j@R/207i1;0040852
<028ACD0R4De08Y5=P80QT4b00@0S4@f09E@@@2ED440W5M80P2FDd<0102OG4`0XEm?0:5OC`2FE4@2
0:1NCP0408i@>`2DEd80U5M209EH@`80Re4k00@0Qddg08YB>`2:DS/0SEDn0P2:E3`01025Cc<0Qe4e
08MA=@2<EcP2095L?@0409MR@P2MJ4H0WFQ60:9]B`80Y6m=00@0X7=:0:9eCP2RMDh0Y7IB0P2TMU@0
102SME@0Y7EG0:AeE`2VMe/20:UjGP040:EfG02ZNf40ZW]Q0:amH`80Z7YP00@0Yg]P0:QlH@2XO640
XWUM0P2MM5P0102MMUL0YGiO0:EnG`2ZPf@20:N0H@040:=lG@2TOEh0Y7eN0:IoH080YGiO102MM5H2
09acE`0409U_EP2DJU40U6YA08mUC@80SF=;00@0TFE:099VB`2BIT/0SV970P25FCh307mF>P0407eD
>01iD3@0NE0d06m==080L4`b00@0Gc/Q06U4:@1YA2T0Pe`o0P2=I4H0102CJ4P0T6=2091S@P1cEd42
04Ha800403lb:@0Y9B80:BDR0410@080C4a<00@0:RDQ034Z900a:R@0:28R0P0]9bL01011>c/0<R`/
038/;00_:BT202dW9`0403<];@10>SX0@3Xj02TX9P80<2l]00@0;Rd[03<b<00c<S00@D0n0P0X9bD0
100Y:2H0;Rd[02h]:`0Z:BL204E4@P0403Dd<P0S8R008b8P02XY9`80<2l]00@0=3<a038a;`0b<Bl0
:2LU0P0Z:BL0100M71X04a8@01<B400R8Al203He<`04040o?@15A480ADA2044k>`80?CLg00@0?cTi
048l?012?3`0=c4a0P0Y8b<0100Z92@0>38b03Pb<P0k=CD204Do?`0404Y4A012?3`0@S`l03La<@80
<bd]00@0=c4a03/e=@0k=CD0A3/l0P0e;2d0100`9bP0<BPY034X:@0f;Bh204Ln?`0404i5AP11>3T0
@CPi02LV8@80;bhY00@0<c8]02h]:00^;BP092<N0P0R8A`0100X9b80;bhY02l^:@0c=C@203Hh=`04
03Xl>`0n@3l0?T0o0412@@80?d5000@0?3hm03Xl>`0j?3/0?Cdm0P0o?cl01011@D40A4A404A4A017
AdL204Y:BP0304a<C01=CDd0CDe=00<0BE5D00@0Ae1E04E>D`15CU<0@diD0P12CeD00`14DET0AeAL
04MDG00304IAE@0404MBF017DEX0Ae5J04MAFP80Ae5K00<0B59L04MAF`17DE/00`19DU/404QAFPD0
Ae1I3014DEX01016CeH0ATmH04I?F016CeP304I>F@D0ATiK0`12CeP1045>E`40@4eF0@3oool001P0
BEMR3017EF0404UGHP<0B5IQ1@17EF0=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0=04QFH@80
AeEP1P16E5l00`17EF00B5IQ04QFH@0404QEGP80B5AP0`18E68304QDI00404EFJ016EVD0AUIU04MC
G`80@4U@00@0:c0c01HH5`0F61L03`h:0P0D4@X0100B4PX04A49014A2@0@3`X200h?2P0400h>300>
40l03Q0?00l@4P80414E0`0<2P/0100?3P`04a8@01<B400A40`200l>2P0401DD3`0O7QT07ahI03H/
8`80C3/a00@0G4De05ho:`1N?b/0HCdU0P1/A2/0101aB380L4La0717<@21DcT208UK@@04099TBP2G
JDl0UfU?09i`EP80YGMM00@0YWQN0:9dFP2RM5X0Z7YS0P2YNf@0102XO6<0ZWiU0:YnI@2_QVX20;6:
J`040:R1HP2JLe@0VW=D08i[C`80W7UK00@0W7MJ091YBP2@JDX0RV=40P2?ITH0102=I4@0Qell08MO
?02@GD8208QE>P0407i;<01iAR/0NDH[07Q5:P80MD8W00@0Kc`Q06Tf6`1Y=Q/0LSXO0P1k@RL01024
B300RTTc08Y9<`2?C3T209I>?`0409A;?02?@cD0Sd<e08i2=@80XUI900@0Yee@09iDA`2NE4L0[FEG
0P2cJed0102RGDh0[fYK0:mZF`2LFDH20:5PC0040:9QC@2HFTD0V5Y5091B?@80SE0k00@0Se8m091C
?P2@Dch0R58f0P27DCD01026DC80R5<d08QC=02>FCT209IQ?`0409iYA`2SKT`0Xfi<09e`A`80Wg99
00@0XWE>0:=eD@2SME40XWAC0P2SM5H0102WN5`0ZW]O0:YkG`2WMf020:QhH@040:AdG@2NL5P0WW1H
09ibF@80YWYQ00@0Z7iU0:MmI02WOF@0TV]<0P2MMUL30:IoH0040:AmGP2VOf00YWmP0:IoH080YGiO
0`2NMEL0102MM5H0W7=G09acE`2IKeH209E[DP04095WC`2?IDd0SfE=09AVC080UFM=00<0Sf=808IJ
?`26FSl00`1nECT0101kE3L0Me0c07M@<`1_Abd207eE>`04068j801M=A/0GCDK07YB>080RV1700@0
SFA808UPA029H4@0R6aF0P1G@2h0100k;RD0:RLR02XW8P0_;bl205YJF00401lL5`0a:R@0<BXT02LQ
8@80<B/[00@0>cDe03H`<00f<300<bd]0P0U7al0100b;2`0>S@d03Xd=00U9282030_;@0402`[:@0^
;B/0;Rd[0491?`80:BPV00@0;2/Y03Pg=@0h=cD0:bXX0P16AD<0100`;bd06aXH01/J600W9R@203Dd
<P0403/j>00e=380=C@b02TX9P80;bh/00@09B@R01dL6P0M71X0:bXX0P0j>CL0100k>SP0=cHd03Lf
=00h<S8203`f=P0403hh>00i<c<0>C<c03D_;`80=c4a00@0?SPh048l?012?3`0?3Hf0P0n>3P0100j
=3@0<R`/038/;00d;Rh2040j>P0404Q2@P19@d<0BD=3038Y:P80<2LX00@0?cHg04<j>`13>S/0=bh_
0P0m=3D01012>CX0<2LX030W:00];2L202l^:@0402PW8P0N7AP07QdH02PW8P80@3lj00@0BdY504M6
@@17AT40=CTj0P0c=cP0100d>3T0>3`m03Pl?@0n@T<20496A`040455AP0o@d@0?d=404Y:BP80Bd];
00@0CDe=04m?C`1?Cdl0Cdm?0P1>CTh0101<C4`0Bd];04];B`15CDl204Q@DP0404YBE@17DE<0Ae5C
04A=DP80@TeC00@0@diD04=@EP13D5H0A4i@0P14CE801015CU<0ATmF04I?EP17D5L304M@F@80AdmJ
1018DEP304M@E`D0ATmF3014DEX01016CeH0ATmH04I?F016CeP304I>F@D0ATiK00D0@e1I049?F012
CeP0@DiG041=EP010?ooo`003018EV4404UGHP<0B5IQ2@17EF0304QFH@T0BEMR0`18EV4504MEH0d0
B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00`0B5IQ00@0BeUT04YHH`1:F6<0BUQS0P19Ef801017
EF00AUAO04IDG`16E5l204ECGP0404ABF`15De`0AE=L04MEGP80B5IO00@0BEMP04QFG`18EUl0AeEN
0P16E5d01015E5/0CEYR04eJHP15D5H2034h?P0401lS9P0=3@l03@d?00L51P805A4B00@04@h900l>
2@0?3PT03`h:0P0>3`X0100>3P`03@l>00d?3P0=3`h200/?3P0400l?3@0F5A<05QDC01DA408050l<
00@0610=01L<2P0G30X07a4>0P0b9240101H?Rl0FSh`05Xn<01L?bl20650;`0406M2<01]A300KD@`
0756<`80MDPc00@0REXn08iO@`2>Gd<0U6M:0P2IJe40102LL5L0X7AK0:1dF`2RN6020:EkH`030:Mk
HP2WNF40YgUQ00<0YgMP00@0YGMO0:1hGP2PN5h0W7UM0P2HNE`0102FLE@0V7=F09QcEP2GLUD2099]
D00408YUB024Gd80Q5m2089M@080Pei100@0Qee507e@>@1mD3T0LD</0P1^>bH0101d?BT0NCh/07Tn
;01i?2X207Lh9`<0OC/[00@0PC`]08Y0<`2:@3<0UDDj0P2MBCl0102QB400XDLo0:57?`2NACd20:99
@@0409]4>`2ZET`0ZUI<0;5OE080Ze]@00@0]fUM0;=WFP2cIeX0^fmR0P2^I5L0102VGDh0Y5a=0:AL
C@2NFdP309IE?`80VE]400@0UU8m095?>@2ACcT0RTXa0P26B2l0102:CS80T5Lj091G>P2IHT<20:1Y
BP0409mXB@2UKTl0YFi?0:M`DP80Y6m@00@0XVmB0:UfF`2YMU/0[G]P0P2^O640102SM5X0X75G0:1a
E`2QLUP20:IhGP040:IjG`2RMU/0XWIK09mfFP80XWUM00@0XWIK0:=jGP2SNUh0YWeQ0`2VOf820:Ao
HP040:J1I02WQ6H0YhAV0:ElH080YgiR00@0WWEI09M^DP2GKU80UfiB0P2CJTh0102@Id/0UFa@09E/
D02?ITX2091WB`0408aSA`23FSh0PeXn07iE>@80O5<g00@0N4lc07=:;P1cBRh0LDlf0P1nFT80101U
@BL0FC@I05Td6@1`Ab/208UM@00409I[B`2>HD00SV5008UUC@80NU]600@0E3d]04<b:P13<RX07a@>
0P16>cL00`12=c40;R4I02hQ6@0302`[:@0402TX9P0g=S@0=cHd02`[:@80:bXX00@09bHT03@c<@0d
<c40>STg0P0W9R@0100d<c409B@R02DT8P0`;bd204Q7A@0402l^;00b<Bl0<S4_030_:P80<C0[00@0
=C@_024P6`0Q81/0<2lZ0P0_;RT0100i>3<0<2lZ030_:P0Y:2H202h]:`0402/Z:00Z:BL0:RTW02l^
;080:RTW00@08b8P02DT8P0U9280?STe0P0h<bl0100^:BD0;BPT02dX900i=30204E0?00404<n>P0i
=300>C@`03<];@80;2HV00@09b4Q02hX:00^:2P0?3Hf0P14?Sh01013?Cd0?CLg03dg=`0c:BL203@Z
:00403`b<018?S`0B3hl04Pn?080?C<a0`0e:bT0100`;2<0=c<Z03Lc:P13?SP204Q5?P0403ln>@0`
;b/0<2l[02PY9080:BXU00@0;Rlc03Df>`0e=S/0?D150P14Adh00`15BU00ADUB04E9DP0304A;D`04
04UBF@17D5L0Ae1G04I?EP80A4eD00<0@daC049;DP12Be800`13C5<4049?F0D0@e1I0`14DEX404I>
F@<0AdmJ1@18D5/404QAFP<0Ae1I1@16CeP<04M@F@0404M>EP17CUP0AdiH04M@F@80AdmJ00@0AU1J
04E?F`15Ce/0@DeI0P10C5P01@11CeP0A4iH04A>F015CEP0B4aH0040oooo000<04QFH@@0BEMR0`18
EV4904MEH0<0B5IQ2@19Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS00<0BEMR04QFH@18EV403018
EV40101;FF@0BUQS04YHH`1:F6<204UGHP0404MEH016E5l0AUAO04IDG`80AE=N00@0A59K04ECG015
De`0AeEN0P18EUl01019Ef00B5IO04QFG`17EEh204IDG@0404EDF`1=FV80CEYR04E@EP80<CPn00@0
7b<V00d=3`0=3@l01`D60P0E4A80100A3PT03`h900l>2@0?3PX200h?2P0400h>300=3`h03@l>00d?
3P802`l>00@03`l=01HE4`0F5A<05A4@0P0D3``0100H40d05``:01L<2P0O4@h2038T8@0405Pn;`1J
?S00FSh`05`o;`80HD0_00@0Id8`06e4<01]A300LDHc0P1eB3<01029FSh0SUm308iO@`2DIdX209U[
D@0409a`E`2PM5/0X7AK0:9hH080YG]S00<0Yg]R0:MiH@2WNF400`2WMf00102UMel0X7QN0:1hGP2L
NEd209QiG00409IaE02HLeH0V7=F09MbE@80TVe@00@0RVE808AO@P24Gd80PUe00P23GT401027GDD0
OE0i07e@>@1a@b`206hk9P0407@m:@1i?R`0NCh/07Tl:P80McPW0`1m>b/01021?2d0RT0c08Y0<`2E
ACX209e9?`040:58@02QAcl0XDLo09i5?@80XTU100@0Vd@k0:YFC02ZET`0/EmD0P2[Fe00102gJEd0
/fMJ0;=WFP2kKf820:iTE`040:IMCP2TG4d0Y5a=09iKB0<0UUDo0P2IFd@0102FDSd0TDli095?>@2:
BS4208I8;`0408Y><P2@EcX0T5Lj09UR@`80X6U:00@0WfQ90:E^C`2UKTl0Yg1B0P2TKe00102RKe80
ZGIK0:UfF`2]Nf020:ilH@040:=dFP2PLEL0X75G0:5bF080YWQN00@0YWYO0:9fF`2RMU/0WgIJ0P2R
NEd0102RMU/0XgYN0:=jGP2VOF430:IoHP80Y7mR00@0YX5T0:N4IP2WQ6H0YGaP0P2WOV80102NMET0
UfiB09M^DP2GKU8209=ZCP04091WB`2EK500UFa@08mVBP80T6M;00@0S6=708=J?P23FSh0OUDi0P1l
DcL0101hCc<0LdX^07=:;P1aCcH207iJ@P0406E19`1I=1T0FC@I0717:`80REe000@0UV];08iQ@02>
HD00RFE=0P1jFdH0101D?Bd0@c8Z04<b:P0O50h204Hk=`03048g<@0^8AT0;R4I00<0;2/Y00@0:BPV
03Lf=00g=S@0;2/Y0P0[:RP0100W9R@0=3<a03@c<@0j>CL202LV900403@c<@0U92809B@R030_;@80
B4M500@0;bh/038a;`0b<Bl0<2lZ0P0a<2/0100e=2l08B0K024P6`0`;bX202l^:@0403Th<`0`;bX0
<2lZ02TX9P80;Rd[00@0:bXX02XY9`0Z:BL0;bh/0P0Z:BL0100S8R009B@R02DT8P0n>CD203Pc;`04
02hY9@0]:2@0;BPT03Td<080AD0l00@0@chj03Td<00i=300<bd]0P0/9RH0100W8B40;RPX02hX:00l
=SH204@n?P0404<m?@0m=cL0?CLg03<Y9`80=2XX00@0?38`04Pn?018?S`0B3hl0P0m<c4303D[:@04
030/8`0g<bX0=c<Z04<n>080B4Dn00@0?chi030_:`0`;b/0:2TT0P0Y:RD0100^;c<0=CHk03Df>`0m
@4D204A7CP0304E:D015BE80ADUB00<0A4]C00@0BE9I04M@E`17D5L0ATmF0P14CE@00`13C5<0@T]B
049;DP0304=<D`@0@TmH1@13D5T304AAFP@0ATiI0`17CeX504Q@F`@0B55J0`17D5T504I?F0`0Ae1I
00@0AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11CET2041<F005045?F014CUP0A4iH
04E=F018C5P00@3oool0000304MEH018EV40B5IQ00L0B5IQ1P19Ef8304QFH@T0AeEP0`18EV4904UG
HP<0B5IQ1@17EF0=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0<04UGHP0304]II01:F6<0BUQS
00<0BEMR0`17EF001016E5l0AE=N04ECGP15De`204IDG@0304MEGP19Ef00BEMP00<0BEMP00@0B5IO
04MEGP17EEh0BUMO0P1=FV001010BDh09blb02L_<P0H71l200d=3`0400/92P0F4Q<05Q8C014>2@80
3`h900@03`h:00h?2P0>3`X03Ph<0`0=3`h200/?3P04014A3`0C4Q004a8@018>2`805Q4>00@0718@
01P=2@0H3@T091LA0P0o<2/0101K?c40Fd0_05]0;`1O?c020691;P0406Q3<01^AC40KTDa0798<P80
MDPc00@0QULk08aM@@2<GD40TfA80P2HJU00102KKED0WG5H09eaF02NM5`20:5gG`040:IjG`2XNV00
Z7YP0:YkH@80ZGYP00@0YWQN09mfFP2OMUX0UW=E0P2@LE80102CKU40U6mB09A_DP2CKU42091[CP04
08]VB@27HTD0Qf9508AO@P80Pei100@0PUQ007Q>=P1hCSH0Kd4Z0P1[>bD0101a?BP0MSdZ07Hm:P1f
>bT207@g9@<0MC8R00@0NS8T088f:@22=RT0S3`a0P2F@SP0102L@c/0WT@l09i4?02OASh20:=:@P04
0:9;@P2WDdT0Ye=90:ABA`80VdTn00@0WDe209e?@`2MCd<0^6YN0P2fJ5`0102cIeT0/6MH0;1WF02[
I5820:=OBP040:1LA`2PGTP0X5i809MC?P80U50k00@0Sddg08a<<`2<C3<0S50d0P2@EcX0102FGd00
VfA509]TA@2MITL20:A]CP040:M`DP2TKe00Y6m@0:I`E080ZWML00@0[gaQ0:ilH@2^O640XgAJ0`2P
LEL20:=eF`040:9fF`2OLeP0Wg=H09edF080X7MK00@0XgML0:ElH02UO600Z7mS0P2YPVD0102XPF@0
Yh9U0:N2I@2YQ6L20:Z7J@040:^2IP2]Q6P0[HAX0:ElH080WWEI00@0X7MK09acE`2LLeL0VG1D0P2N
MET3095XC00408]RAP22FCd0PUTm07eD>080OUDi00@0OE@h07U@=01iD3@0ME4i0P1hE3X0101[AR`0
IShT06Hn901Y@RD207eD=P0409AYB@2<HD40S65108m[D@80NUY300@0KE=203DS5`0e8aL0<b@M0P0i
;RP0101?A3h0;24K02`Q6`0/:bT202/Z:00402PW9@0e=380=C@b02`[:@80:bXX00@09bHT03<b<00c
<S00A4=10P0W9R@0100];2X09B@R02DT8P0b<Bl204I5@`0402l^;00d<c40=3<a02DT7`80<2lZ00@0
?chi02h]:00^;BP0=C@_0P0/:bH00`0e=2l0;B`W02d/9`0302/Z:004024P7P0O7Q`07ahL02TX9P80
;Rd[00@0;bh/03He<`0f=C<0>cHb0P0h<bl0100d;b/0=c8^03Lb;P12?CT204U4@004044l>00d;b/0
=2l[034[:`80<R`/00@0=S0`03Xd=00j=3@0?CLg0P0l=SH0100g<C40<R`/038/;00n=S<204Hn>`04
04m7A01>AT<0CTI3040h=@80<b/X00@0<BTV03L_;00g;b`0?cXd0P0k=S00100h=C00=S<^03Hc;P0^
;BT202TZ9@0402`/:P0d=380=3@b03/l@@80@D9700@0ATU>04U>E019CU@0BTmE0P18CeL01017CUP0
B4mI04Q?F@17DEX204I@F@0304E?F013CEH0@deF00H0@TaE1012CeP504=@F@<0A55J1016CUT304M?
FPD0B51K0`19DU/404QAFQ40Ae1I00@0AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11
CET2041<F005045?F014CUP0A4iH04E=F018C5P00@3oool000<0AUAO00@0AeEP04QFH@18EV40B5IQ
0P19Ef8304YHH`@0BEMR0`18EV4904MEH0<0B5IQ2@19Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS
00<0BEMR04QFH@18EV403`1:F6<01019Ef80B5IQ04QFH@18EV4204MEH0<0AUAO00@0AUAM04MEGP17
EEh0B5IO0P19Ef0304YHH@<0BEMP00@0CEYP04]FG01;EU`0=Sm40P0L8BD0100A4QH03`l?00l?3`0B
3Pl201DA4004018?2P0@3`X040l:010?2`803a0;00@03`l=00h@3`0>40l03Q0?0P0<40l0100B4Q00
3`h<00l>300?2`P201TA3P0401hC3`0N3`X07Pl:034P6@80Cc`e00@0FC`/05Xm;@1J?Bd0Gch]0P1U
@Rl0101[AC00LdTc07=9<`1iC3D207a>=`0408UG?02?GD80See209EVBP80W6eC00@0WW1H09mcFP2O
LeX0X7=L0P2QM5d00`2UMed0ZGYP0:UjH0030:ekHP040:IgG@2JLED0VW5E08i[C@80QVM800@0Q5m2
089M@022GD00P5/n0P1nFC`0101mF3/0NUDh07YE>01fDC@2079=<00407YB>@1`ARh0L4H^06Hi8P80
HS8L00@0IS<N06`e8@1/=B40KcHS0P1_=280101m>2T0OCDW07de9`1n=2L208<e:@0408Xh;@2@?380
T3`b09Hm=@80Uchf00@0VT4i09e4?02MA3`0ZeA;0P2_F4l0102YED/0YU980:IBB02TDTL20:iLD@04
0:QFB`2XF4d0Z5Q=0:IGBP80We5400@0Vdhn09iC@02NDd00YU]80P2XGd`0102HDCl0V54m09QA?@2E
DC`209AB>P04095C>P2CEc/0TeLk09AK?P80Uei100@0Vf940:9YB`2RJD/0YFi@0P2VKe40102XLUH0
[7UN0:aiGP2_O6420:ikH0040:n0IP2/OF<0[7eS0:YkH@80ZWaR00@0Z7aQ0:IjG`2VNUl0YWeQ0P2[
PVH0102NLUL0WgIJ09mfFP2SNUh30:AmH080XWeP00@0Y7mR0:B1H`2TPF<0YgiR0P2YP6@0102RNEd0
WGAH09edF02PMe/209edF00409Q_D`2MM5P0WGAH09A[C`80TfY>00@0S6=708=J?P23FSh0OeHj0P21
F3`01022FCd0P5Lk081G>`1gDcT207A?=@0407A?=01cBc40Ld/a06Ho8P80KdHX00@0RF1208aP@`2<
H4<0T6]A0P24HTT0101mGTX0@BXJ044Z6P0g9AT203TZ8`040515?`0e;2L0=B`W02`[:@80:2LU00@0
9bHT03<b<00c<S00;bh/0P0/:bT0100X9bD0<2l]030_;@16AD<202PW9@0402/Z:00Z:BL0:RTW03Pg
=@80ADA200@0;Rd[034`;P0a<2h08B0K0P0a<2/01018Ad80=C@_03Dd;`0c<Rd202LV8@0403@c;P0^
;BP0;RdX02h]:@80:bXV00@08B0L01lN6P0O7QX0;2/W0P0a<2`0100`;b/0=3<_03@c;`0l>C@203dj
=@0303`i=00k>3<0>cPc00<0>cPc00@0=34/02dZ9@0]:RD0=C4`0P0k=cH01010?3/0?c/j03lk>P0h
=3<2034];00402dY:00/:2L0;2PW03Lb;`80@Sdj00@0B4=0048m>P12?CX0=c8_0P0d;b`0100i=340
?cXg03lj=`11?ST203Ld;`0402/Z9P0W9R809bHR02HV9080:b/[00@0=CLf0412@@10@T40ADQ=0P18
Be00101;CUD0C55G04aAE`19D5P204Q?F@<0Ae1I00@0Ae5J04I@F@16D5T0A4iG0P13CEH6049<E@@0
@TmH1@13D5T304AAFP@0ATiI0`17CeX504Q@F`<0BU=L0`19DU/604QAFP`0Ae1I00@0AdiF04M>F017
CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11CET2041<F005045?F014CUP0A4iH04E=F018C5P00@3o
ool000<0AUAO00@0AeEP04QFH@18EV40B5IQ0P19Ef8304YHH`@0BEMR0`18EV4904MEH0<0B5IQ2@19
Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS00<0BEMR04QFH@18EV403`1:F6<01019Ef80B5IQ04QF
H@18EV4204MEH0<0AUAO00@0AUAM04MEGP17EEh0B5IO0P19Ef0304YHH@<0BEMP00@0CEYP04]FG01;
EU`0=Sm40P0L8BD0100A4QH03`l?00l?3`0B3Pl201DA4004018?2P0@3`X040l:010?2`803a0;00@0
3`l=00h@3`0>40l03Q0?0P0<40l0100B4Q003`h<00l>300?2`P201TA3P0401hC3`0N3`X07Pl:034P
6@80Cc`e00@0FC`/05Xm;@1J?Bd0Gch]0P1U@Rl0101[AC00LdTc07=9<`1iC3D207a>=`0408UG?02?
GD80See209EVBP80W6eC00@0WW1H09mcFP2OLeX0X7=L0P2QM5d00`2UMed0ZGYP0:UjH0030:ekHP04
0:IgG@2JLED0VW5E08i[C@80QVM800@0Q5m2089M@022GD00P5/n0P1nFC`0101mF3/0NUDh07YE>01f
DC@2079=<00407YB>@1`ARh0L4H^06Hi8P80HS8L00@0IS<N06`e8@1/=B40KcHS0P1_=280101m>2T0
OCDW07de9`1n=2L208<e:@0408Xh;@2@?380T3`b09Hm=@80Uchf00@0VT4i09e4?02MA3`0ZeA;0P2_
F4l0102YED/0YU980:IBB02TDTL20:iLD@040:QFB`2XF4d0Z5Q=0:IGBP80We5400@0Vdhn09iC@02N
Dd00YU]80P2XGd`0102HDCl0V54m09QA?@2EDC`209AB>P04095C>P2CEc/0TeLk09AK?P80Uei100@0
Vf940:9YB`2RJD/0YFi@0P2VKe40102XLUH0[7UN0:aiGP2_O6420:ikH0040:n0IP2/OF<0[7eS0:Yk
H@80ZWaR00@0Z7aQ0:IjG`2VNUl0YWeQ0P2[PVH0102NLUL0WgIJ09mfFP2SNUh30:AmH080XWeP00@0
Y7mR0:B1H`2TPF<0YgiR0P2YP6@0102RNEd0WGAH09edF02PMe/209edF00409Q_D`2MM5P0WGAH09A[
C`80TfY>00@0S6=708=J?P23FSh0OeHj0P21F3`01022FCd0P5Lk081G>`1gDcT207A?=@0407A?=01c
Bc40Ld/a06Ho8P80KdHX00@0RF1208aP@`2<H4<0T6]A0P24HTT0101mGTX0@BXJ044Z6P0g9AT203TZ
8`040515?`0e;2L0=B`W02`[:@80:2LU00@09bHT03<b<00c<S00;bh/0P0/:bT0100X9bD0<2l]030_
;@16AD<202PW9@0402/Z:00Z:BL0:RTW03Pg=@80ADA200@0;Rd[034`;P0a<2h08B0K0P0a<2/01018
Ad80=C@_03Dd;`0c<Rd202LV8@0403@c;P0^;BP0;RdX02h]:@80:bXV00@08B0L01lN6P0O7QX0;2/W
0P0a<2`0100`;b/0=3<_03@c;`0l>C@203dj=@0303`i=00k>3<0>cPc00<0>cPc00@0=34/02dZ9@0]
:RD0=C4`0P0k=cH01010?3/0?c/j03lk>P0h=3<2034];00402dY:00/:2L0;2PW03Lb;`80@Sdj00@0
B4=0048m>P12?CX0=c8_0P0d;b`0100i=340?cXg03lj=`11?ST203Ld;`0402/Z9P0W9R809bHR02HV
9080:b/[00@0=CLf0412@@10@T40ADQ=0P18Be00101;CUD0C55G04aAE`19D5P204Q?F@<0Ae1I00@0
Ae5J04I@F@16D5T0A4iG0P13CEH6049<E@@0@TmH1@13D5T304AAFP@0ATiI0`17CeX504Q@F`<0BU=L
0`19DU/604QAFP`0Ae1I00@0AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11CET2041<
F005045?F014CUP0A4iH04E=F018C5P00@3oool0000404ECGP16E5l0AUAO04IDG`80AeEP00@0BEMR
04YHH`1:F6<0BUQS0P1;FF@404UGHP<0B5IQ2@17EF0304QFH@T0BEMR0`18EV4504MEH0d0B5IQ0P19
Ef8604YHH`0304UGHP18EV40B5IQ00d0BUQS0`19Ef8304QFH@<0AeEP0P16E5l304QFG`<0BEMP1P1:
F640101=F5h0@diB04=>DP0[<cH2018G6P0400`=3`0B41404Q0A01DA40804`d=00@04Pl:010?2P0@
3`X040l;0P0?40/0100?3`d03Q0?00h@3`0>40l200`@3`04018A3`0?3P`03`h<014<20805Pd800@0
7A0802LF3P0W5Ph0?BTP0P1F@3D0101H>BL0FS/W05Xk9`1Q?R/206M3;@040717<@1jCCH0NTdf085C
>`80QEDn00@0SUa109ARA`2DHTL0WF]@0P2RLeT0102UMEh0Y7IN0:AfGP2SMUl20:AgH0040:AgFP2W
N5`0YgQL0:]iGP80ZGML00@0X75E09IZC@2FJTd0RVI60P25I4<0101kEST0Me8e07MB=@1aC2l3071;
;P80KDP[00@0IT4T060k7P1P>ah0LD/d0P1X@R/0101P=R00FbhI05/^6@1M;AT206@_7@0306Xb8@1/
<b80K3<R00<0MRhP00@0N2hQ07h`901n<2@0PS8W0P27=BX0102:=R`0SCH]08df;@2B>c82090i<004
0:9>A02UDDL0YE570:IBB080ZeM=00@0YE570;=OE@2cGeD0[eeB0P2kJEh01031Kf<0]fEI0;MUF@2Z
F4X20:=B@@0409]<>`2EASD0UDHe09I=?080VE0o00@0VeA209YF@@2JET40VUQ00P2GFD00102GFcl0
VEe109UM@@2HGd4209mVB0040:A[CP2UKU00YFi@0:UaEP80[gUO00@0/G]Q0:ejG`2]NUl0[7eS0P2Z
Nf40102VMed0XWAJ0:9dFP2OLeP20:1dF@040:9iG@2WOV80YgiR09a`E@80WWEI00@0XWUM0:=lG`2S
O5l0XgaO0P2QO5l0102ROF00Xh1R0:>0HP2SNUh20:ElH00409ieF@2IL5@0VG1D09edF080VG1D00<0
TfY>09I]D@2FKE400`2FKE40102@Id/0Qei208MN@P23FSh308EL@080PeXn00@0Me8g07UA=`1iDCL0
O5@j0P1jDcH0101Z@bH0I3dP06@m801gD3<208eTB00408UQA`2AK540TFaA07aH@080ITLb00@0<AL8
044_9@11;bD0@CH`0P16?S/0100/:bT09RDS02HU8`0W9R@203<b<00403@c<@0];2X0;B`Z02PW9@80
:bXX00@0?C`j02h]:`0^;B/0<C0^0P0`;bd0100j>CL0@T4o0491?`0`;bd202d/:P0402TX8`0d<bh0
=3<^04M6@@80<C0[00@0;2/V02@S7P0T8ah0>3Lb0P0d<bh00`0_;RX0:bXV02/Z9P03020O6`0402h]
:@0d<bl0=3<_038a;@80=C@`00@0@Chi04=0>`13@3/0@3dh0P0f<bh0100[:2<09R<N02HS7P0Y9R42
02dZ9@0403di>00o>cX0?c/j03lk>P80>CDd00@0<Bd/02hZ:@0^:RT0<Bd/0P0e<C00100e<Bh0>3@a
03Pd<@0f<Rl2038^:`0403Hb;`11?CX0@Cdj04I2?`80@cll00@0<S4_02/Z:00[:RP09RHV0P0/;2`0
100b<cD0=c/l03Lk?00n@TD204E9C00404M:D@18Be80B4]B04Q=D`<0ATeE0P14CEH304A<E`0404AA
FP13D5T0@e1I049?F080@DiG0`10CEH3045>E`@0@TiJ1@13Ce/304A@G0@0ATiI0`17CeX504Q@F`03
04]DG@1:De`0BU=L00<0BU=L0`19DU/304QAFP`0Ae1I00@0AdiF04M>F017CUP0Ae1I0P17CeX01016
D5X0ADmK04E?F`11CET2041<F005045?F014CUP0A4iH04E=F018C5P00@3oool0000404ECGP16E5l0
AUAO04IDG`80AeEP00@0BEMR04YHH`1:F6<0BUQS0P1;FF@404UGHP<0B5IQ2@17EF0304QFH@T0BEMR
0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ00`0BUQS0`19Ef8604QFH@<0AeEP
1018EUl304UGH0<0BUQQ0P1;FF801016DED0=d1503M0A@0P:2/200lC5P0400`<300E4A005A4@01HA
3P804@T700@04a0;014@2`0A40/04A0<0P0@4@`0100@40h03a4@00lA400?4A0200dA400400l>300D
40d0510=01H>2`804`P200@07@`403<O500c7a@0Bc@V0P1E>bh0101M?2T0H3dY060m:@1U@B/206e4
;P0407I9<P1mCcL0ODlg08EC>P80R5Hk00@0T5e209MTB@2GI4T0WfeB0P2ULeX30:EeGP<0Y7IO00@0
Y7EI0:MeFP2WMEX0YgAI0P2RKe@0102JJ4d0T6A7091TA`28I4@208MTA00408AO@P1nFC`0OUTl07MB
=@80M4lb00@0Ldha06m:;@1_BRd0IT4T0P1N>A`0101F=1/0EC4I05Da6@1C;AH205<Y4`0405@U4@1F
8a00ER<@05DP3P80EAd<00@0IB0A06TT5@1Y91D0LBTK0P1i;b801021=BP0QcT]08Li;@2;>c0208dm
<P0409U9?P2??c@0Scld09Y:?`80VDTn00@0Vd]00:9@A@2RD4D0V4Hk0P2RD4D0102B@3D0XDm40:5?
A02YEd/20:=AA@0409e:?02MBSX0WDXj09m<?080WT/k00@0U4Pj09Q<?P2HC3h0W5=20P2NEdD0102M
FD@0W5Y209aJ@P2MGD@209eOAP0409QM?`2NHdD0WV=50:5XB`80Y6]>00@0Z71E0:mgGP2_Meh0/7YP
0P2]Med0102YNV00Z7UO0:QiG`2TME/209i`EP0409Y^D`2LL5D0W71E0:5hG080YGaP00@0YGUN0:Mn
HP2WOV80Zh9V0`2/QFP30:V4I`80ZHIX00@0YWeQ0:MnHP2WOV80X7MK0P2LLeL00`2OMUX0VW5E09Ya
E@03099YC@0409M^DP2HKe<0V6mC09E/D080SFA80`29H4@01027GT80PUTm089I?@1iD3@2089K?P04
07mH>`1jDcH0NU<f07E><@80HCTO00@0I3`R08YUBP2:IDX0Q5/o0P2>IDL01023FS`0O5Ll07aG?011
8Pd203/U600403L/9P1>B4P0CTQ802h]:`809B@R00@09bHT038a;`0b<Bl0>CPf0P0_;R`0100W9R@0
9B@R02DT8P0e=38203Lf=00403Pg=@0^;B/0;Rd[030_;@80>cXh00@0=cHd034`;P0a<2h0<2lZ0P0c
<Rd01013@Sd0;RdX02h]:00Z:B@202HU800403/j=@0c<Rd0<c8]02d/9`80:2LR00@06aXE01XI500J
6A@0;2/V0P0h=c80100m?3L0A4<n04A3?P0l>cH203Xi=00403Dd;`0_;RT0;bhY02XY9080:BPS00@0
;RdX03@c;P0d<bh0>cXh0P0i>3H0100f=C<0<c8`03<b<00b<Bl203@c<@0403Th=P0m?3X0?C`j0491
?`80?C`j00@0=3<a02h]:`0^;B/0=SDc0P11@3h01010?cd0=cHd03Lf=00/;2h202TZ;00402d^<00g
>ch0=c/n0454B@80@dQ<00@0A4U?04A;D@14Be40AT]A0P15BU001013BU80A4]E04A;E@14CEH204A<
E`<0@deG00@0A55J04=@F@13D5T0@TmH0P11CUL01010CEH0@DiG045>E`11CUL2049?F0@0@TiL1@13
Ced304A@GP@0ATiI0`17CeX504Q@F`0304]DG@1:De`0BU=L00<0BU=L0`19DU/304QAFP`0Ae1I00@0
AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11CET2041<F005045?F014CUP0A4iH04E=
F018C5P00@3oool0000404ECGP16E5l0AUAO04IDG`80AeEP00@0BEMR04YHH`1:F6<0BUQS0P1;FF@4
04UGHP<0B5IQ2@17EF0304QFH@T0BEMR0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`0304UGHP18EV40
B5IQ00`0BUQS0`19Ef8604QFH@<0AeEP1018EUl304UGH0<0BUQQ0P1;FF801016DED0=d1503M0A@0P
:2/200lC5P0400`<300E4A005A4@01HA3P804@T700@04a0;014@2`0A40/04A0<0P0@4@`0100@40h0
3a4@00lA400?4A0200dA400400l>300D40d0510=01H>2`804`P200@07@`403<O500c7a@0Bc@V0P1E
>bh0101M?2T0H3dY060m:@1U@B/206e4;P0407I9<P1mCcL0ODlg08EC>P80R5Hk00@0T5e209MTB@2G
I4T0WfeB0P2ULeX30:EeGP<0Y7IO00@0Y7EI0:MeFP2WMEX0YgAI0P2RKe@0102JJ4d0T6A7091TA`28
I4@208MTA00408AO@P1nFC`0OUTl07MB=@80M4lb00@0Ldha06m:;@1_BRd0IT4T0P1N>A`0101F=1/0
EC4I05Da6@1C;AH205<Y4`0405@U4@1F8a00ER<@05DP3P80EAd<00@0IB0A06TT5@1Y91D0LBTK0P1i
;b801021=BP0QcT]08Li;@2;>c0208dm<P0409U9?P2??c@0Scld09Y:?`80VDTn00@0Vd]00:9@A@2R
D4D0V4Hk0P2RD4D0102B@3D0XDm40:5?A02YEd/20:=AA@0409e:?02MBSX0WDXj09m<?080WT/k00@0
U4Pj09Q<?P2HC3h0W5=20P2NEdD0102MFD@0W5Y209aJ@P2MGD@209eOAP0409QM?`2NHdD0WV=50:5X
B`80Y6]>00@0Z71E0:mgGP2_Meh0/7YP0P2]Med0102YNV00Z7UO0:QiG`2TME/209i`EP0409Y^D`2L
L5D0W71E0:5hG080YGaP00@0YGUN0:MnHP2WOV80Zh9V0`2/QFP30:V4I`80ZHIX00@0YWeQ0:MnHP2W
OV80X7MK0P2LLeL00`2OMUX0VW5E09YaE@03099YC@0409M^DP2HKe<0V6mC09E/D080SFA80`29H4@0
1027GT80PUTm089I?@1iD3@2089K?P0407mH>`1jDcH0NU<f07E><@80HCTO00@0I3`R08YUBP2:IDX0
Q5/o0P2>IDL01023FS`0O5Ll07aG?0118Pd203/U600403L/9P1>B4P0CTQ802h]:`809B@R00@09bHT
038a;`0b<Bl0>CPf0P0_;R`0100W9R@09B@R02DT8P0e=38203Lf=00403Pg=@0^;B/0;Rd[030_;@80
>cXh00@0=cHd034`;P0a<2h0<2lZ0P0c<Rd01013@Sd0;RdX02h]:00Z:B@202HU800403/j=@0c<Rd0
<c8]02d/9`80:2LR00@06aXE01XI500J6A@0;2/V0P0h=c80100m?3L0A4<n04A3?P0l>cH203Xi=004
03Dd;`0_;RT0;bhY02XY9080:BPS00@0;RdX03@c;P0d<bh0>cXh0P0i>3H0100f=C<0<c8`03<b<00b
<Bl203@c<@0403Th=P0m?3X0?C`j0491?`80?C`j00@0=3<a02h]:`0^;B/0=SDc0P11@3h01010?cd0
=cHd03Lf=00/;2h202TZ;00402d^<00g>ch0=c/n0454B@80@dQ<00@0A4U?04A;D@14Be40AT]A0P15
BU001013BU80A4]E04A;E@14CEH204A<E`<0@deG00@0A55J04=@F@13D5T0@TmH0P11CUL01010CEH0
@DiG045>E`11CUL2049?F0@0@TiL1@13Ced304A@GP@0ATiI0`17CeX504Q@F`0304]DG@1:De`0BU=L
00<0BU=L0`19DU/304QAFP`0Ae1I00@0AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11
CET2041<F005045?F014CUP0A4iH04E=F018C5P00@3oool000<0AUAO00@0AeEP04QFH@18EV40B5IQ
0P19Ef8304YHH`@0BEMR0`18EV4904MEH0<0B5IQ2@19Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS
00<0BEMR04QFH@18EV40301:F6<<04QFH@<0B5IO0`17EEh01018EUl0BEMP04UGH01:F64204]IHP04
03m8C@0Y<C@0:C4d01LL7`80414C00@03@d=01@@3`0D40l05A0=0P0@20D0100C40/04A0;014@2`0A
40`2010A3004010@3P0?4A003a4@00lA40803A4@00@03Pd;01@@3@0D40d0614;0P0F2@40100R40@0
@R/M048[7@1G?2/205Df900406I3;@1XA2h0J4@^06]5;P80LDL_00@0N4Xb07e>=01mCS@0Q54f0P27
E3L0102AFd40UF9709ERA`2LJDh20:1^E@040:9`F@2PL5T0X71I0:1`FP80X79K00@0XgAF0:EdEP2U
M5H0Xg1C0P2NJ4`0102CHT@0Reho08]N?`28HD0208QV@`0408QSAP22GD00PUe007aG>P80NE@g00@0
N5<f07A?<P1dCc80K4LZ0P1U@2<0101J>R<0HChX064n:01W@bd206a3;`0406/n:`1U=2<0IC@S05h[
6P80FB@D00@0H1h>06@R4P1T8Q80JRLG0P1_;1`0101c;Ql0Mc8S07Lb8`1l=2H207hf:00408Y2=022
>R`0PSX/08Hn<080Rd4d00@0TDLj09I:?@2FBSd0TdDi0P2IBcl0102EAc/0UdTm09M9?@2DACP208Xk
;P0408Hd9P2=>RX0SCXZ09Y7=`80Y55100@0T44d09E7>P2EAcX0VTi00P2KDT40102JDd40VUI109YF
@@2MFdD20:1PA`0409UL?`2LHD<0W65309mSA`80X6M:00@0YVeB0:eeG02]ME`0[gMN0P2/M5/0102X
NEl0ZW]Q0:YkH@2WN5h209maE`0409Y^D`2NLUL0WW9G0:AkG`80Z7mS00@0Yg]P0:V0I02YP6@0[HAX
0P2^QfX0102/QFP0Z8=V0:R3IP2WPVD20:J3I@040:AkG`2UO600YGaP09edF080VW5E00@0WWEI09Q_
D`2HKe<0SVE90P2=I4P0102FKE40V6mC09Q_D`2FKE4208mVBP<0S6=700@0RV5508EL@025G400OU8e
0P27GT001021F3X0NE8e07UB=@1mEC/2068m8P0405Td6P24H4H0Q61608QL?`80PeHe00@0UVQ607iA
<01nDC00JD@Z0P0b60T0100h;BL0A41104A0@@0f=C<302PW9@80;bh/00@0>cXh030_;@0`;bd0:BPV
0P0T8b40100c<S00?C`j03dl>P0h=cD202PW9@0402HU8`0d<c40=3<a03dl>P80>3Le00@0;B`W02h]
:00^;BP0@D0k0P0a<2/0100^;BP0:RTT02XY900k>SD202d/9`<0:BTQ00@08R8J02@T700T91`0<c<[
0P0j>S80100i>C40?Cde03dm=@0e=2l202h]:00402`[9P0d<bh0=3<^03dl=`80?chi00@0?C`g03/j
=@0k>SD0=C@b0P0d<c40100e=380>3Le03Pg=@0l>cT303ln?080?C`j00@0?Cdm03/k>`0k>c/0=SHf
0`0a<C4303<c<`80<30`00@0=c/n03@g?00d=c`0=cXo0P0oA4X01015BU00@dYB04=:DP12BE4204=<
E@0404I:D`14Be<0A4]C04A;D`80ADiG00@0AdmJ04QBG018DU`0Ae5M0P14D5`01012D5/0@DmJ045?
FP10CUT303m=F080@4iI0`11CeX4049>G0D0@dmM0`14D5h404I>F@<0AdmJ1@18D5/304YCG0<0BE9K
1P18DEX<04M@F@0404M>EP17CUP0AdiH04M@F@80AdmJ00@0AU1J04E?F`15Ce/0@DeI0P10C5P01@11
CeP0A4iH04A>F015CEP0B4aH0040oooo00000`17EF00B5IQ04QFH@0704QFH@H0BEMR0`18EV4904ME
H0<0B5IQ2@19Ef8304QFH@D0AeEP3@18EV4204UGHPH0BUQS00<0BEMR04QFH@18EV403019Ef8<04QF
H@0304QFG`17EEh0AeEN00@0AUAM0P18EUl01019Ef00BUQQ04YHH@0jA4H201hV:@0400hC5P0?4180
3a0B00h=2`804Pd:00@05`l<01@:200D2PP0514<0P0B4@`0100B4@d04A8=014B3@0A4@l3010B4@80
3Q8A00@04A0>018>2`0B3P/060l:0P0N40L0100^6`d0CS@U04hd9@1Q@S0205Th9@0406M3;@1Y@b`0
JD</06e5;080M4P_00@0Nd`b089@=@22D3D0RE<g0P2<EcP0102CGD<0Uf5709MQA`2KJ4d209m[D`04
09m]EP2PKUL0X6iG09m_F@80XG5K00@0Y7=E0:EbE@2ULUD0Y6iB0P2KIDT0102@GD00QeXk08MJ>`24
GC`208EQ?`04089M@01nFC`0OUTl07]F>@<0NUDh0P1gDSD0101bCC00KTT/06i9;01WBC4206a<=@04
07=@>P1dCST0M4hi0797=080K3d[00<0IcHU06Hc8P1V<b800`1W:QP0101V:AL0I2LE06@W5@1Q9182
06<T4`0406LX5`1Z:aX0JR/J06X[6P80KBhM00@0KbdM07Xh:01j>2P0OCXZ0P1j=BH01022>R`0Pc/]
08<k;@27?C0208To<P0408e1<`2=?c80SClb08Pi:P80QC@S00@0QcHU08hk:`2>>b/0SS`a0P2A@CH0
102DAST0UDTk09E9>`2BBCP209A=>@0409UE@02NG4H0WUa609UL?`80W5m200@0W61409iRAP2NHTH0
XVU>0P2[LET0102]ME`0ZW9I0:YbF@2MKU@20:5bF0040:1aE`2GJDl0UfU?099VB`80V6aA00@0WGAH
0:5hG02QN5`0YGUN0P2WOV80102[PVH0ZhAW0:^4I`2YPVD20:AoHP040:9mH02POEl0X7eO0:9iG@80
XgYN00@0W7=G09YaE@2JLED0X7MK0P2KLUH0102?ITX0SFA808eTB02DJdl209I]D@0409A[C`2>IDT0
SVE908aSA`80SfI:00<0T6M;08aSA`2<HdL00`26Fc/01022FC/0OEHi07eF>@1nES`206U4:P0405Xf
701jESh0NUHn08iS@`80QELe00@0VVI108UG<P29Ec80OE@f0P119Q<0100e:R@0<Blb034_<P11@3h2
02`[:@0402LV900Y:2H0:BPV03Th=P80;Rd[00@0;2/Y02LV900W9R@0<C0^0P0j>CL0100b<Bl0:bXX
02/Z:00Z:BL2038a;`0403dl>P0f=C<0=SDc02PW8P80:RTT00@0@T4l03<b;@0c<Rd0;B`W0P0Y:2<0
100k>SD0;2/V02`[9P0T91`202/[8`0402`/900b<RX0<S8Z03`l=080>SXb00@0<C4Y030`:00`<2P0
<C8/0P0^;bT0100a<R`0>c`f03/l=P11@S`203ho>@0403Xk=@0i>S@0>CXd038b<080=CDc00@0>SXh
03hn?00n?S`0@40n0P0o?cd0100m?C/0>c/i03/k>@0d=SD203Hh=`0403Xl>`0l?Sd0?3hm03Tk>P80
=SPg00@0>C/j03i0?`0n@3l0AdYA0P11AT`01010AD/0@TUA0499D@13BU@20419DP030418D`12C5H0
@TaF00<0A4]C00@0@daE04I>F@16CUT0B59L0P19Del00`15DEd0@TiJ049>FP03045?FP04041>F@0o
CEP0?deH03m=F080@4iI00<0@DmJ049@F`12D5/01012CUh504=?G`<0A51P1016CUT304M?FPD0B51K
0`19DU/404QAFQ40Ae1I00@0AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11CET2041<
F005045?F014CUP0A4iH04E=F018C5P00@3oool0000304MEH018EV40B5IQ00L0B5IQ1P19Ef8304QF
H@T0AeEP0`18EV4904UGHP<0B5IQ1@17EF0=04QFH@80BEMR1P1:F6<00`19Ef80B5IQ04QFH@0<04UG
HP`0B5IQ00<0B5IO04MEGP17EEh01016E5d204QFG`0404UGH01:F640BUQQ03Y4AP807RHY00@03Q<F
00l@4P0?41803Pd;0P0B3@X0100G3``050X801@:200D4@`2018A3004018A3@0A4Pd04A8=014A3`<0
418A0P0>4Q40100A40h04Ph;018>2`0H3`X201h@1`0402hK3@1>=2D0CS@U0652<080FCPU00@0Id<]
06U3;01Y@b`0KDD/0P1dB2l0101kC380PU0e089@=@29DcL208aG>00409=M@`2GHDL0Uf5709]XC@80
Wf]C00@0WfeF0:1^E`2PKUL0WfmI0P2QLE/0102TLeD0YG9E0:EbE@2TKU8209]UB@04091M@027FS/0
QeXk08AM?080QF4o00@0PUe007iI?01nFC`0NeHi0`1jECP207MB=@04079=<01^BB`0KTT/06M9<@80
K4`e00@0Le0j07A>>@1dCST0LTLd0P1/?B/00`1W=RD0IS<R06Hc8P0306LZ600406HY5`1T9aD0I2LE
064T4P80Hb@C00@0IbPG06X[6P1Z:aX0JR/J0P1];Qd0101_;Ad0NSPX07Xh:01m>RX207Xe9P04088j
;023>bd0Pc/]08Lm<080RClb00@0SD4c08do<P2=?c80R3TZ0P25=2<01027=RD0SS/[08hk:`2>?342
0951=P0409A6>@2EBC/0UDTk0999>080U4di00@0VEE009iLAP2NG4H0VE`o0P2LGd80102LH4@0WV96
09iRAP2RJDh20:]aF@040:eeG02ZLUT0ZW9I09e^E080XG9H00@0X75G09MYC`2GJDl0TVI;0P2HK540
102MM5P0XGQL0:5hG02UNEh20:MnHP040:^2IP2[Q6L0ZhAW0:V2I@80Y7mR00@0XWeP0:1mG`2POEl0
XWUM0P2SNUh0102LLeL0VW5E09YaE@2PMe/209]bEP0408mVBP2=I4P0SFA809A[C`80UVeA00@0U6]?
08iUB@2>IDT0S6=70P2?ITX00`2@Id/0S6=708aSA`0308IK>`04089I>`1mEST0OEHi07iF?080JD@Z
00@0FSHL07YF?P1jESh0SV=30P25EcD0102JIT40RELb08UG<P1mE3H2044V4`0403DZ900a;c80<Blb
0450?P80;2/Y00@09bHT02TX9P0Y:2H0>CPf0P0^;B/0100/:bT09bHT02LV900a<2h203Xi=`04038a
;`0[:RP0:bXX02XY9`80<S4_00@0?C`j03He<`0f=C<0:2LR0P0Z:B@01012@C`0<c8]03<b;@0];2L2
02TX8`0403/j=@0/:bH0;2/V02@T7080:b/S00@0;2`T038b:P0b<RX0?3`d0P0j>S80100a<BT0<30X
030`:00a<R`202h_:@04034b;00k?3H0>c`f0452?080?Sli00@0>S/e03Tj=00i>S@0<S8`0P0e=C<0
100j>SP0?Shl03hn?010@3h203lo?@0403dm>`0k>cT0>c/i03@f=@80=SPg00@0>S`k03`n?@0l?Sd0
>C/j0P0f>3L0100i>cX0?T0o03i0?`17BU420456C0040415B`12BE40@TUA04=:E080@4UB00<0@4QC
049<EP12C5H00`14Be<01013C5D0ATiI04I>F@18DU`204UCG`0304EAG@12CUX0@TiJ00<0@DmJ00@0
@4iI03m=F00oCEP0?deH0P10CUT00`11CeX0@U1K049@F`04049>GPD0@dmO0`14D60404I>F@<0AdmJ
1@18D5/304UBF`@0B55J4@17D5T01017CUH0AdiH04M>F017D5T204M?FP0404I@FP15Ce/0ADmK045=
F@80@4aH00D0@DmH04A>F014CUP0ADeH04Q<F0010?ooo`003018EV4404UGHP<0B5IQ2@17EF0304QF
H@T0BEMR0`18EV4504MEH0d0B5IQ0P19Ef8604YHH`0304UGHP18EV40B5IQ01P0B5IQ00<0AeEN04ID
G@16E5d01015De`204MEGP0404UGH01:F640BUQQ03U1A0805alQ00@02PhA00lA400?4A003@`:0P0?
2PL0100F3P/060h<01P>300D4@`2018A3004018A3@0A4Pd04A8=014A3`<0418A0P0>4Q40100E5180
3PX700h:1`0F3@P202DG3P0403XT5`1F>bX0ES/Z06I5<P80Gc`X00@0HShV06Do:01U?bP0Jd<Z0P1d
B2l0101mCS80Qe@g08MD=`2?FS/209AM?`0409QRB02KID/0VfE;0:1ZD080XFeE00@0XFmH0:9`F@2R
L5T0XW9L0P2UMEl0102TLe<0YG9C0:EbD`2TKe0209]VA`0408mL?@24EcH0Q5Lf085I=P80PEdi00@0
OUTl07eH>`1mF3/0O5Lj0P1mF3/307iI?00407aG>P1iE3L0NE@g07EF@@80MEE000@0M54m071:=`1`
BSL0K44`0P1[?Bd0101^?Bh0LD0a0750<@1mA34207a3<00407Ln:`1b>BH0LSTV06lf8`80LSTV00@0
ND0]07m6<`1oAS<0LSTV0P1l@c00101h?B/0PD@b0854<P1h>BP206PU5@0406l/701^:AX0KRTJ07Xe
9P80NS8T00@0OC@U08<g:@23=bT0PC@T0P1l;A`01021<1l0RSTX08Xi:02;>Bh208lm<P040911=02>
@S@0ST8d08]2<@80SDHd00@0Tdlj09]G@P2KEd80VUe00P2LGd80102LGd<0Vem309]O@`2RIT`20:U_
E`040:]cFP2XL5L0Z71G0:1aE`80YWMM00@0YGIL09a^E02LKU@0Uf]@0P2MLEH0102SNUh0YWeQ0:Im
H@2ZOV<20:b3I`040;27J`2`RF`0/8U/0:f6J@80Yh9U00@0Y7mR0:9oH@2ROf40Z7mS0P2ZPFD0102T
Nel0XgYN0:=jGP2ZPFD20:ElH00309YaE@2HKe<0V6mC00<0U6]?00@0TFQ<08]RAP2;HTH0Rf960P2A
J4`0102DJdl0TfY>09=ZCP2>HD02081E=@0408AH>`22FCd0PUTm07YE>P80K4P^00@0H3`T075?=P1a
CcH0TFA50P2EID40102<ES00WfU309mY@`1jCB`205i2;@0402/P6P0S8bD08b<U04U8AP80<2l]00@0
9RDS02@S8@0T8b40=C@b0P0/:bT0100^;B/0:RTW02XY9`0/:bT203<b<00402h]:`0c<S00<c8`03Lf
=080=C@b00@0>CPf02h]:`0^;B/09bHQ0P0Z:B@01013@Sd0<S4/038a;00X9b8202@S7P0403dl=`0a
<2/0<C0[020P6080:2PP00<0:b/S030`:00`<2P00`0k>c<0100f=Rh0>CTa03Ti<@0e=S0203Tj=004
03ho>@0o@3X0?d0j03Hg<@80:b`V00@0;2dW03@e;`0d=Bl0=cLe0P0k>cT30410?P0403dm>`0k>cT0
>c/i03/k>@80?Cdk0`0m@D001014B4L0CE5@04eAD01>DU4204Q<B`0404U=C01?De80Ce=B04Y>E`80
ATYC00@0@dYB04E<EP15C5H0@T]D0P0oAe80100nB580@TaH049<F012BE<20458DP04045:D`14C5L0
A4aG04I@FP80A51L00<0@DeI03e9E`0mBEL00`11CeX01010CUT0?deH03m=F00oCEP2041>F@03045?
FP12D5/0@U1K00@0@TiN1@13Cel304A@H0@0ATiI0`17CeX504Q@F`@0B55J0`17D5T504I?F0`0Ae1I
00@0AdiF04M>F017CUP0Ae1I0P17CeX01016D5X0ADmK04E?F`11CET2041<F005045?F014CUP0A4iH
04E=F018C5P00@3oool000`0B5IQ1019Ef8304QFH@H0AeEP0P18EV401019Ef80BUQS04YHH`1;FF@3
04YHH`h0BEMR3017EF0404QFH@80BEMR0`1:F6<304]II0@0BUQS0`19Ef8604QFH@H0BEMR0P18EV43
04IDG`0404MII@17FF<0AeUS04]LI080BUMO00@0Ae9H051IGP1@FEh0DEUL0P0oAdT0100U;Bl05AdO
01DM7`0:410200d?3P0301<A4P0E3`l05@l?00<050/<0`0G3Pl0100E3`l0510?01@@3`0B4@l2014A
3`04010B3`0>4`l03Q<?014@3P804Pd900@05@`5034P5P0a81H0D3X/0P1I?2X00`1M>RH0JDD_06U5
;`0306e4:0040755:P1eBBh0MDT^07a>=080Q5Dk00@0REY008]L@P2;G480Uf590P2MIdl30:1ZDP04
0:9/E02UKeL0YFmG0:A^EP80X6YB00@0XFmD09m]DP2OKE80VVQ=0P2AHTH0102<GD40RE`o08UL?`2;
GT4208eQA00408IO@P2:HdH0RV=608QSB080Q5m400@0OUY007aH@01lF400NUPo0P1iEch0101lG4L0
NEU407UIA01eED0207IC?`0407QE@@1iE440NEA107MB?`80MU0m00@0Qd`l091GAP2@EdH0U5a;0P2<
EdD01028ED80Re]708]KA`2?H4`208iQC00408mLA`2MJUD0WFYE09eYE080WfQD00@0Uei;08]>?02;
CS`0QTLf0P1d=B@01021>B/0Qcd`08Lm<02;?c4208Lj:P04080a8P20;ah0P2lN08Hc8`80SCXZ00@0
L2P@07`d7P1l=1h0Qch]0P28?S401022>2/0PCPW084h9`2:@R`209E==@@0UUe00P2IH4<0102NIDP0
XVU<0:9YC02TJdh20:E/C`040:AbF@2MKU@0WFiD09E/D0<0TFa?0P2FKe80102QN5`0[h=X0:n3J02/
QFH20:j7J0040;29JP2_R6T0[hQY0:^4I@D0Yh1Q00@0[H=Z0:UoIP2YOfH0ZGmV0P2KLEP0102TNV40
XGMN0:5gGP2TNV4209]aF00409Q^E@2EJe80UF]B099XC`<0T6I=0P2BJ4l0102EJe80V6iE09Q^E@2B
IDP208iQA00408YN@@20EcT0P5Li07YA<`80O5Df00@0Ld`]05/d5@1K=1D0VFU30P2HI3`0102MI3L0
QD`Q08E<8@27ESD207=C?P0403<T7`0T7b<091lS03Tg:P80?d4f00@08RTR02</:@0S;2T07R8U0P0o
>T00100j:C40?bP`03lX<00X92D2038]:P04040h=@0^9B00;RDP03H/:P80;BLW00@0=cLi03<d>@0c
=3T0=C@_0P0`;bX01010?cX0=SD`03He<00V9B0203<b;@0403@c;P0X9b80:2LR02@Q6P808QlH00@0
9B8K02h[900^:b@0>CH_0`11?SL203dj<`<0?CPe00@0>3@a034];00a;B`0:2LU0P0X9RL0100`<300
>3Pj03Ph>P0n>CD203lj=P04040l>@13?c`0@cll04E4@P80B4Q800@0C4a<04i>CP1>CTh0ATm<0P18
DDh0101:De00BeAA04]DD@1;E54204UBC`0404I?C014CDX0A4e:0459E080@4QC00@0?dMB03i6D@0n
AU40?TIA0P0oAe800`10B5<0@DUD0459E00304=;EP<0A4aG0`15CEP304I>F@0404AAFP13D5T0@e1I
049?F080@DiG0`10CEH3045>E`@0ADaN1@16CEl304M>H0`0AdmJ3017D5T404I?F0<0Ae1I1@18DEX3
04U@EP0404QAF017D5L0Ae1G04I@F@80ADmH0`13D5T01@11CeX0A4iG04A>E`15CUD0B4eA0040oooo
000<04QFH@@0BEMR0`18EV4604MEH080B5IQ00@0BEMR04YHH`1:F6<0BeUT0`1:F6<>04UGHP`0AeEP
1018EV4204UGHP<0BUQS0`1;FF@404YHH`<0BEMR1P18EV4604UGHP80B5IQ0`16E5l01017FFD0AeUS
04MIH`1;G6@204YGG`0404MBF01@FEh0D5UN055IG080?dM900@09Bd_01DM7`0E7Al02Q0@0P0=3`h0
0`0C4A805@l?01D?3`0301@;30<05`h?00@05@l?01@@3`0D40l04Q4?0P0A4@l0100@4Pl03Q<?00hC
3`0A40h2018=2@0401D<1@0a81H0<B0F050j;080FC`Z00<0GCXV06U5;`1YABl00`1]A2P0101aABX0
MDT^07E9;P1lCS@208AE>`0408UJ@02;G480Rea209MQB@80WFM?0`2PJU80102RK5@0YFmG0:E_E`2T
KUH20:1ZDP040:5_E02OKE80WfeB09YXC@80TF9600@0S5e108UL?`29G3l0Rei10P2=HD@01026Gd80
RV=608YSAP28HdP208AOA00407iJ@01lF400O5Q007YH?`80NELn00@0O5a707UIA01iFD@0MEE00P1f
Dcl0101hED40NEA107UD@@1gDSl207I@?@0408M<?02@EdH0T5M609ALB`80S5M500@0R5E208]KA`2;
FdL0Sf1<0P2>HD`0102?G4L0WFYE09eZE@2MJE@209mXE00409MNB`2;CS`0Rdhl08I7=P80M3DT00@0
PCT[08Lm<027?C00Rcla0P27>RX01020<B80P2lN080_7P26<b<208dj:P04070X401l=1h0O3@N08Ln
;@80R3ha00@0PSP[084h9`21>2L0RT8/0P2ECCD409IM@080VF1300@0WVE80:9YC02RJD`0Y6]>0P2U
K4l0102TLUT0WFiD09e^E02EK503095/C`80UVmB00@0XGQL0:n3J02_PfP0[8EV0P2^QfP0102`RFX0
[hQY0:n8J@2[Q6D50:N0H@040:f3JP2YOfH0ZGmV0:UoIP80Vg5H00@0Y7YQ0:5gGP2QMeh0Y7YQ0P2K
LEP0102HKUD0UF]B09E[DP2BJ4l3091VC@80TVQ?00@0UF]B09Q^E@2HKUD0TVE80P2>HD@0102:GT40
P5Li081G>@1jDC<207aE=P0407=<;@1K=1D0Fc@E09UY@`80V6@l00@0WF@g08E<8@25C240QeHe0P1c
Dch0100c91l091lS02@O8`0i=bX203m1=P04028Y8P0S;2T08b`Y01hR9@80?cY000@0>RTa03lX<00o
:300:2@U0P0b;BX01010>3D0;RDP02hU800f;2X202dW9`0403Lg>@0c=3T0<c@i03Dd;`80<2lZ00@0
@3lj03He<00f=C009RDP0P0c<Rd0100d<bh0:2LR02PW8P0T8AX2028O600402DR6`0^:b@0;R/T03Tf
;`<0@Chg0P0m>S<303dh=@0403Pd<@0a;B`0<Bd/02PW9@80:2HW00@0<30`03Ph>P0h>3X0?STe0P0o
>SH01010?3T0@cll04<o?015A48204Q8B00404a<C01>CTh0CTi>04I?C080B55>00@0BU=@04]DD@1;
E540BeAA0P19DTl01016Cd`0A4e:04A=BP11BE@20418D`0403m7DP0nAU40?TIA03i6D@80?dMB00<0
@4QC0459E011BE@00`13BeH304A<E`<0ADeH0`16CUT01014DEX0@e1I04=@F@12CeP2045>E`<0@4eF
0`11CUL404E<GPD0ATeO0`17CV0<04M?FP`0Ae1I1016CeP304M@F@D0B55J0`19D5H01018DEP0Ae1G
04M@E`16D5T204E?F0<0@e1I00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`003018EV4404UGHP<0
B5IQ1P17EF0204QFH@0404UGHP1:F6<0BUQS04]II0<0BUQS3P19Ef8<04MEH0L0BEMR1P1:F6<504UG
HP<0B5IQ0`17EF000`18EV40BEMR04UGHP0404UGHP80B5IQ0`16E5l01017FFD0AUQR04IHHP19FV82
04UFGP0404QCF@1@FEh0D5UN04eEF080>T9400@07RLV01@J6P0D6QX0310A0P0>40l0100C4A80510?
01@@3`0C3@d201D=2`0401L>3`0E3`l05@l?01D?3`80510?00<04Q4?014A3`0A4@l00`0@4Pl0100D
40d04Pd9018=2@0L4@/203LV700405<l;P1L?bd0G3l]064n:P80JDD]00@0K4<W06e4:01]A2P0LTH[
0P1gBc00101mCcD0Q5Dk08AE>`28FCl208]L@P0309IPB02LITh0W6I>00<0WfUA00@0XF]C0:A^EP2T
KUH0Y6iF0P2OJE40102NJe00W6U>09aYCP2GIDX2091QA@0408]L@029G3l0RE`o08]N@@80SV540`28
HD@01029HDL0QV5608IQAP20G48207]G?`0407UG?P1kFD00NeU007YJ@`80N5Q100@0MUHo07UF@01i
ET00O5Q20P1oFdD0101oFD@0OUQ307iH@`29DD0208mJB00409AQCP2AHDd0TF5=08mPC080T6=>00@0
TVQB09=ZE02CJU@0TfUA0P2IK5D0102DITl0VFUC09UYD`2OK5L209aXD`0409QQC@28Cc`0R4ll08A1
<@80Oc`/00@0N3<S078Z6`1b:Q/0KRDD0P1_8Q80101a9180Lb@C07<T4`1f;a/207hg9@0408Dm;`25
?C40QCda084i;@80P3PZ00@0QSl]08a5<@2<AC40T5@j0P2EFCl0102JGT@0WF5709eQA`2NHTP20:1T
BP040:AXCP2XK580Z6aB0:5_EP80VFY@00@0TFQ<091[CP2@Jdh0TFa?0P2GL5<0102QN5`0[H5V0:f1
IP2^QfP20;29JP<0/X]/00@0[hQY0:f6I`2]QVL0[HIW0P2^QfP0102^QFT0ZX5U0:Z1I@2YP6@209]b
EP040:5hG02NMET0WWEI0:AkG`80WgIJ0`2DJdl0102EK500UVeA09I]D@2FKE4209E/D0<0U6]?00@0
TFA708iQA02>HD@0RUi10P23EcX0101lDcD0OUDg07iE=`1eCRl205lh6@0407mD<P2MKDL0WFe709IQ
>@80Qe8Z00@0RUTh07EA>@1eDCT0CSP]0P0[7AX0100k;b<0Ad0f04M0=P0W9R42024U900402HW:`0l
=cd0?3Lm02lR:@80?BX`00@0:2@S03D`;@0e<2d0@cTg0P0Y81/0100a9bD0;bXW02lZ9`0l>Sd2038c
=`0402d/9`0/:bH0;2/V040o>P80>3Lb00@0:2LR03<b;@0c<Rd0=3<^0P0Z:B@0100N7QH0820H020P
600V9Qh202l_9`0403Ph<00k>c<0>c/c03Ti<@80=CD]00@0<RdY034/:00a;2P0<BhY0P0e<Bh00`0h
=cD0>c/i03/k>@0303dm?@0403Th=P0m?3X0?C`j04=3@`80BDU900@0CDi@04m@DP1?D580CE5D0P1<
D5<01017Ce40B51B04Q@DP1:DU@204]CE@0404YBE018D580B51B04E=C`80A4a>00<0@DUD0418D`10
B5<01P0oAe800`10B5<0@DUD0459E00304=;EP<0A4aG0`15CEP304I>F@0404AAFP13D5T0@e1I049?
F080@DiG0`10CEH3045>E`@0ADaN1@16CEl304M>H0`0AdmJ3017D5T404I?F0<0Ae1I1@18DEX304U@
EP0404QAF017D5L0Ae1G04I@F@80ADmH0`13D5T01@11CeX0A4iG04A>E`15CUD0B4eA0040oooo000@
04UGHP<0B5IQ1P17EF0204QFH@0404UGHP1:F6<0BUQS04]II0<0BUQS3P19Ef8<04MEH0@0BUQS2`19
Ef8304QFH@@0AeEP0P16E5l00`18EV40BEMR04UGHP0404UGHP80B5IQ0`16E5l01017FFD0A5IP04AF
H015EUh204MDG00404YEF`1?F5d0CeQM04M?DP80<SXl00@05Q`L010F5P0@5QH03A4@0P0@4Q40100C
4Q004`l>01<?3P0D3``201D@3@<05@l?0`0D40l00`0B4@l04A4?014A3`03014A3`0401D@300D3@L0
50d702DH4080@S0T00@0ES`]0611;@1P@Bd0J4@^0P1ZA2d0101[@RH0KTDY06i5:@1eBBh207Y><`04
07mA=`24EC/0Q5Dk08QI?`80RU]100<0U5i609YTC02JI4`00`2MIdl0102OJE40XfeE0:=]E@2SKED2
09mYD@0409YWC02IIT/0VFI;09MTB@80TV1500@0S5e108]L@02;G400S5m20P2?HTD0102<HdL0QUm2
08IO@P26GT@208QSB00408AOA@1jESh0NUHn07QD?080OUY200@0OEe607]KA01kFd@0O5U30P1mFT@0
0`20G4H0Pee808=MB00308ANB@0408QHA02:FdL0RU]7091QC@80TVQB00@0TFQB08iXD@2>J540T6aD
0P2ELET0102=KU80S6Y?08aZC`23GdD208MRB00409=[DP2IKeL0VFmG09]^E`80TFA=00@0W65?08mD
@P2?E480PD@b0P1e=R@0101_;Q`0JBHC06TV4`1O6`P205XC000406@Q401X9AD0J2DE06l[8080MS8Y
00@0NcL^084m<P21?C80QT<c0P28AC@01025Ac80T58m091B?@2MGdX20:=UD0040:9TC`2PHTd0X69=
0:5SCP80Y6IA00@0W6YA09AUB`2DID/0SFA80P2>JD`0102CKU40VG9E09UbE@2QN5`20:]oI0040:f6
I`2_R6T0[hQY0;6:J`80/X]/00@0/HY[0;29JP2`RFX0/HY[0P2bRf`0102^QFT0Zh9V0:^2IP2YP6@2
09edF00409ieF@2LLeL0W7=G0:=jGP80XWUM00@0Uf]@09Q/D@2HK540VViC0`2KKe@209Y^D`0409Q/
D@2GJe00Uf]@095TA`80SF1300@0S5m208EI?025FC`0P5@g0P21F3X0101jDC<0IChO06Dn7`1P?B42
09acD`0408eO>`2<FSD0S5Xe08aL>080NU4e00@0KTdj03dQ4`0m8A<0?R4I0P1B>cD0100d9R<09B4P
02DQ800[:bd203@b=`0402HP900i;S80>Bhb02XT9080>C4_00@0A3/f02PM6@0X7AT0;B@O0P0a;2T0
1010?Sl0<C8f034b=P0Z:B@202/Z9@040450>`0k>SD0>cXe02`[9P80<S4/00@0;RdX02@S7P0T8ah0
8B0K0P0W9R40100_;RT0=SD`03He<00j>C@203Th<`0403Dd;`0a<2/0<C0[034^9`80<RlX00@0=C8[
03Lf<@0g=S40>cXf0P0m?ST0100m?C/0?3hk03`n>`10A4L204=7BP0304Q<C`1<DED0C55E00<0CEAJ
00@0BeAK04YCFP1:DeX0B55H0P19DUT304YCFP0404QAF016CeH0ATmF04A=E080@T]B0`11BE@01010
B5<0?dMB03m7DP0oAe820418D`<0@DUD0`13BeH304A<E`<0ADeH0`16CUT01014DEX0@e1I04=@F@12
CeP2045>E`<0@4eF0`11CUL404E<G0D0ATeM0`17CUh<04M?FP`0Ae1I1016CeP304M@F@D0B55J0`19
D5H01018DEP0Ae1G04M@E`16D5T204E?F0<0@e1I00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`00
4019Ef8304QFH@H0AeEP0P18EV401019Ef80BUQS04YHH`1;FF@304YHH`h0BEMR3017EF0404YHH`/0
BEMR0`18EV4404MEH080AUAO00<0B5IQ04UGHP19Ef801019Ef8204QFH@<0AUAO00@0AeUU04AFH014
EV00AEIN0P17E5`0101:EE/0CeQM04mHG@17Ce82038j?00401HL700@5QH041HF00dA4080418A00@0
4a8@01<?3P0C3`h050l<0P0E40d301D?3`<0510?00<04Q4?014A3`0A4@l00`0A4@l0100E40`050d7
01@=1`0U6102048`900405Hl;@1P@Bd0H44]06Q4;P80JT@]00@0Jd8V06i5:@1^ABT0MDT^0P1jCS<0
101oDCL0Q5Dk08AE>`28FCl208YK@@0309ANAP2JI4`0VVA<00<0WFM?00@0WfUA0:=]E@2SKED0XfeE
0P2OJE40102JId`0VFI;09UVB`2GI4T2099PA@0408aM@@2;G400Rea008aO@P80Sf9500@0S6=708IO
@P26Gd80QUi40P28HdP01024GdD0NUHn07YF?P1hE3`207iJ@P0407eMAP1kFd@0Ne]407aI@`80OEY4
00<0P5a608=MB023GDP00`24GTT01028F4@0RU]708YKA`2@HDd2099XDP04095XDP2>J540SVQA091/
E080UG5I00@0SFiB08aZC`2<JTl0Pem50P27HTP0102CJe80VFmG09U_E`2KKUL2095TC@0409aQC`2?
E480SeA20854<P80MCHT00@0KbhL06TV4`1Y9Q<0Ga/80P1J4`00101T8A00J2DE06PU5@1_:b0207Hb
:@0407/g;P21?C80PCdb08I3<`80R4Dd00@0QDLb091B?@2@DSd0WEm:0P2SIE00102RI4l0X69=0:1R
C@2QHdh20:AVD@0409aZD@2DID/0U6E;08eTB080SVU<00@0TfiA09UbE@2ILUD0XGQL0P2[Of@0102]
QVL0[hQY0:n8J@2aRV/20;:;K0040;6:J`2`RFX0/8UZ0;6:J`80/X]/00@0[XEY0:^2IP2[PVH0ZH1T
0P2MM5P0102NMET0W7=G09acE`2SNUh20:9iG@0409M[D02HK540V6aA09Y^D`<0VfmD0P2JKU<0102H
K540Uf]@09M[D02AI4L208eP@`0408aO@P25FC`0QETl081D=`80PEPj00@0NU4c06Dn7`1U?Ql0H3dQ
0P2LLe<0102=Gc/0S5Xe08aJ=@2<G3P207YA=@0406i=>P0m8A<0?B4C03hQ6@80DS/e00@0=2HS02DQ
800U8B00:b/]0P0d<SL0100V82@0>Bhb03T^<P0Z92@203Ta;`0404@k=P0X7AT0:1dI02dT7`80<B`Y
00@0@3ho034b=P0a<SH0:RTT0P0[:RD01011@3/0>cXe03/j=@0/:bH2038a;00402h]:00T8ah092<N
024P6`809bHQ00@0;bhY03He<00f=C00>STd0P0i>3<0100e=2l0<C0[034`:`0a;RL2038_:00403Db
:`0g=S40=cHa03/j=P80?Chi00@0?Cdk03`n>`0l?S/0@4A70P13AdX00`18C4l0C55E04aAE@0304eD
FP0404]DF`1:DeX0BU=J04QAF080BE9I0`1:DeX01018DEP0ATmF04I?EP14CE@2049;DP<0@DUD00@0
@4QC03m7DP0oAe80?dMB0P10B5<30459E0<0@d]F0`14C5L304E=F0<0ATiI00@0A55J04=@F@13D5T0
@TmH0P11CUL3041=EP<0@DiG1015C5`504I=G@<0AdiN3017CeX<04M@F@@0ATmH0`17D5T504QAFP<0
BE1F00@0B55H04M@E`17D5L0AU1I0P15CeP304=@F@05045?FP14CUL0A4iG04E>E@18CE400@3oool0
0100BEMR0`18EV4604MEH080B5IQ00@0BEMR04YHH`1:F6<0BeUT0`1:F6<>04UGHP`0B5IQ0`1;FF@3
04YHH`0304UGHP18EV40B5IQ00L0B5IQ0`17EF0504IDG`0304QFH@19Ef80BEMR00@0BEMR0P18EV43
04IDG`0404MII@14EV00A5IP049CF`80AU=K00@0CEQN04eFF`1=EU/0?DE80P0W;c4300hB4@04010B
4@0B4Q004Q8@014A3`804A0<00<0510=01DB3@0E4Pd0100D40l8018A3`0401H?2@0E30D05@`502hP
5`80BSLY00@0F3d/0692;@1R@Rd0JdD^0P1Z@RP0101[@RH0KdHZ06m6:P1gBc0207a@=@04085C>@25
ES`0QEHl08QI?`80RU]100@0Tee509QRBP2HHTX0VfE=0P2JI4`0102MIdl0XF]C0:5[D`2RK5@209iX
D0<0WFM=00@0VVM<09ESB02EHdP0TEm40P2>Gd<0102?H4@0TF96095RAP2?ITX208MN@P0408IL@`2<
I4X0S6A:08MRB080NeDn00@0N5@l081LA020G4@0PV170P21GdH00`20GTD0P5a4081LA003089LA@<0
QEa600@0R5i808AJA024FT@0QUe70P2;IDh0102<J500R6I=08QVC@29JE02091bF00409=kG`2>MUX0
SWIJ08I/D@80Q6U>00@0RVYA099`E`2BL5L0VgMO0P2MNF40102_Of/0Z7ER0:QeHP2MJE@209ENBP04
08aC?`20AS80P4Hb074d7`80IRPC00@0Fa`=05XJ3P1J6Ph0G1`C0P1R8A/0101Y:280KRhU06h^9@1^
;R8206`]7P0406l`7`1l?B`0O3d/08m@?`80WEi=00<0Y6ED0:IWEP2VIeH00`2UIUD0102EHdX0SUm5
08iOA@2:HDD208iYC00409E`D`2KM5L0VgAG0:9iG@80ZgmT00@0ZhAS0:b5I02/QF@0[HIU1@2^QfH3
0:n8I`<0Zh9T00@0ZH1R0:9iF`2RNE/0WWEG0P2LLeD0102PMeT0XWUK0:9iF`2NLUD209a`D`0409Y^
D@2HK4l0V6a?09Q/C`80VViA00@0W71C09ibE@2NLUD0TfA80P2>Gd<0102=H4<0RE`o08UL?`23EcX2
08EI?00407iE=`1^ABL0KTDW05@f7P80T6eA00@0R5lo08eO?@2=Gcd0RUTh0P23EC@01020E3T0G3@K
05`d6`146A<2054^:P04048]:P0b:2T0<RPY02XX:`80:bX_00@0:RP[03Td>00i=3P0;2LT0P0e;BX0
1013>3@0;B8L02dR700b9b<203<[:003044m?P0b<S@0<S8d00<0;RdX00@0@3lj03/j=@0k>SD0;bhY
0P0a<2/0100U91l06QTD01XI500[;2L2038c;P0403Tj=@0l?CP0?3dh03Xk=P80=SLb00@0<c@_038c
;P0b<bh0=S</0P0j=c00100l?3@0>STd03Xi=00e=S0203Dg<P0403`n>@13AD80@dE204mFG080CEAJ
00@0BU=J04QAF018DEP0Ae5J0P15DU/00`17E5d0B5EN04QEGP0404YBG@80BE5L00@0B51K04E=F015
CEP0@d]F0P11BE@01012BUD0@DUD0459E010B5<303m7DP80@4QC00<0@DUD049:E@12BUD00`13BeH3
04A<E`<0ADeH0`16CUT01014DEX0@e1I04=@F@12CeP2045>E`<0@4eF0`11CUL404E<G0D0ATeM0`17
CUh<04M?FP`0Ae1I1016CeP304M@F@D0B55J0`19D5H01018DEP0Ae1G04M@E`16D5T204E?F0<0@e1I
00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`00301:F6<404UGHP<0B5IQ1P17EF0204QFH@0404UG
HP1:F6<0BUQS04]II0<0BUQS3P19Ef8<04QFH@<0BeUT0`1:F6<00`19Ef80B5IQ04QFH@0704QFH@<0
AeEP1@16E5l00`18EV40BEMR04UGHP0404UGHP80B5IQ0`16E5l01016F6@0A5IP04AFH011DUX204IC
F`0404iIG`18DEH0B55F038j?@807BDW00@02`d<00h@3`0>40l04A<@0P0B4Q000`0A4Pd04Q4<018A
3003018A30L04Q4?1@0D40l0100F3@H06Q0701X@1`0g9aX2058k;@0405Xn:@1U@B/0ID4[06e5:`80
K40U00@0Jd8V0758;01aB2`0NDdb0P1nDSL01021DcT0Q5Dk08AE>`28FCl208]L@P03099LA02GHDT0
Uf5900<0VF=;00@0VfE=0:5[D`2QJe<0XVaD0P2OJE40102RK580XfeC0:=]D`2RK58209aYCP0409MT
B@2CHDH0Tf56095RAP80TV=700@0U6Q=08YN@`2:GT<0Qee40P2?ID`0102<I4/0OUQ107iH@@1kECh2
089LA@04081NA@21GDD0PEe5085MA@80PUa500@0PE]408AKA@24FdD0QUe70P27GTP0102>IDl0QUe7
08IMA`22G4D208MSB`0408IVC@24IT`0Q6I<08I[D080SG9G00@0U85S08YgF@2:MeT0PfiA0P20IT/0
1022Id`0RfeC08]]D`2CLeX209]kHP040:^0K@2/OfX0[7mZ0:YkI`80YWIP00@0XFiI09UUC`2IIDl0
TU]60P2>ED401029D3l0PdTm08=9?@1l@CT207Ll=P0407<h<P1]<RX0KC8Z06<Y7@80G2<B00@0IRPI
06/]7P1[;Ah0MCLX0P25AcP00`2DETL0Wf5B09mQDP030:9TE@0408aJ@@29FT00REY008UPA080SfY=
00@0Ug9E09aeF02LMEP0XgYN0P2[Of@30:f7H`040:b6HP2]Qf<0[HMS0:f7H`D0[XQT00@0YgiP0:V0
HP2YP680ZH1R0P2XOf40102RNE/0WgIH09mfF02MM5H209mfF0040:1cEP2NLE@0WW5D09a_DP<0VfiA
0P2LKe80102NLE@0X7=F0:1cEP2FI4T308mPA080S5m200@0QeXm08MK?P27Fch0Q5Pk0P1gBbh0101O
?bH0NeTn07]I?P24Gd8208aQ@@0408EG=P2;FST0ReXi07m>;@80OTd]00@0BQTB04XQ6`1:8A/0D3Lc
0P0n<S80100U8bH08R<W028S9`0f=c/203Ph>P04034Y9`0_9B<0;bDS044f<080>2/S00@0>BhX038X
9P0b:2H0?SXi0P0b<3<0100b<B`0;bhY02l^:@0l>cH203Lf<@<0<C0[00@08b8M01XI500J6A@0=CDc
0P0i>CL0100l?3X0>c/i03/k>@0g=cD303<c<@80=3@b00@0=C@_03/j=@0k>SD0@3lj0P10@C`0100n
@3d0@D=00453@017BdX204mCDP04059KHP1>F640CUQQ04]EGP80AU=L00@0A55J04=AG013DE`0A59M
0P13Df0404UCG`80B59N00@0AU1L04A>FP14CUX0@TaH0P11BeL01012BUD0@DUD0459E011BE@30418
D`<0@DUD0P12BUD304=;EP<0A4aG0`15CEP304I>F@0404AAFP13D5T0@e1I049?F080@DiG0`10CEH3
045>E`@0ADeJ1@16CU/304M?G0`0AdmJ3017D5T404I?F0<0Ae1I1@18DEX304U@EP0404QAF017D5L0
Ae1G04I@F@80ADmH0`13D5T01@11CeX0A4iG04A>E`15CUD0B4eA0040oooo000<04YHH`@0BEMR0`18
EV4604MEH080B5IQ00@0BEMR04YHH`1:F6<0BeUT0`1:F6<>04UGHP`0B5IQ0`1;FF@304YHH`0304UG
HP18EV40B5IQ00L0B5IQ0`17EF0504IDG`0304QFH@19Ef80BEMR00@0BEMR0P18EV4304IDG`0404IH
I014EV00A5IP045BFP80AU=K00@0CUUO04QAEP18DEH0<SXm0P0M9BL0100;3@`03Q0?00h@3`0A4a02
018B4003014B3@0B4@`04Q4<00<04Q4<1`0B4@l501@@3`0401H=1P0J40L06Q0703LW6P80DS/]00@0
FShY06E1:`1U@B/0KDD[0P1/@2D0101[@RH0LDP/0758;01iCC8207iB=`04085C>@24EC/0Q5Dk08QI
?`80Rea200<0TUa409MQB@2GHDT00`2IHd/0102KIDd0XF]C0:5[D`2RK5@209mYD@040:9/DP2SKE<0
XfeC0:9/DP80W6U>00@0UfA909=QAP2CHDH0TF960P2BHdL0102DJ4d0RUi308YN@`27GD@208mUC004
08aTB`1nF440OUQ107]E?P80PUa500@0P5i5085MA@21GDD0PEe50P22G4D01021Fd@0Q5]508AKA@26
GDL208MNB00408iUC`26GDL0QUe7089LA@80Qf=;00@0QVI=08AVC024IT`0QV]@0P2=LUL0102DPF<0
RWMI08YgF@23KU42081VB`04089WC02;KE<0RfeC09=cFP80Vg]R00@0Zh1]0:aoJP2/OfX0ZW]W0P2V
MV00102QKUT0VFE?09UUC`2BFdH208iE@@0408U@?`23BCd0PdTm07a1>@80Mc`f00@0LcPb06db:P1]
<RX0HbTM0P1L8a80101V:1T0JbdN06/]7P1e=bP208E7>00309AFA`2OHE80Wf5B00<0XVAE00@0S5Y1
08UJ@029FT00RF140P2?JTd0102GLUD0W7EH09aeF02SNUh20:]oI0<0[HMS00@0[8IR0:f7H`2]Qf<0
[HMS1@2^R6@0102WOV00ZH1R0:V0HP2YP6820:QoH@040:9iF`2OMUP0WgIH09edEP80WgIH00@0X7=F
09iaE02NLE@0W6mB0`2KKU4209a_DP0409iaE02PLeH0X7=F09ITB@<0Sf140P2<Gd801027FSd0Qe/n
08MK?P24F3/207M;;P0405lo9P1kFCh0NeTn08AO@P80S65100@0QELf08]J>@2;FST0Odh]0P1nCBd0
101:6A80BR4K04XQ6`1@=c<203hb<P0402DS9P0R8bL08R<W03Hg>`80>3Pj00@0<BTW02lU8`0_9B<0
@CH`0P0h:b<0100i;RP0<RPV038X9P0n>ST2038`<`04038a;00_;RT0;bhY03`k=P80=cHa0`0a<2/0
100S8Qd06QTD01XI500e=C<203Ti=`0403`l>P0k>cT0>c/i03Lg=@<0<c<a0P0d=380100e=2l0>cXe
03/j=@10?cX20411?00403i0?@11@d00@D=004M;BP80Ce=B00@0DU]R04iHH@1>F640BeEN0P16De`0
1014DEX0@e5L04=AG014DUd204=CH0@0BE=O0P18DUh01016D5`0A4iJ04A>FP12C5P2045;E`04049:
E@11BE@0@DUD0459E0<0@4QC0`11BE@2049:E@<0@d]F0`14C5L304E=F0<0ATiI00@0A55J04=@F@13
D5T0@TmH0P11CUL3041=EP<0@DiG1015CEX504I>F`<0AdmL3017CeX<04M@F@@0ATmH0`17D5T504QA
FP<0BE1F00@0B55H04M@E`17D5L0AU1I0P15CeP304=@F@05045?FP14CUL0A4iG04E>E@18CE400@3o
ool000`0BUQS1019Ef8304QFH@H0AeEP0P18EV401019Ef80BUQS04YHH`1;FF@304YHHaX0BEMR101:
F6<;04UGHP<0B5IQ1017EF0204IDG`0304QFH@19Ef80BEMR00@0BEMR0P18EV4304IDG`0404EGH`16
F680AUQR049CF`80AeAL00@0CUUO049;D012Be009Rha0P0C6ad0100=3@/04A4?014A3`0C4a42014B
3@04014B300C50h04a@>018C2`803a08100A4@l2018A3`<0510?0`0E3`l0100H3PD09QP=02HH3@10
;R0205Lm;0040611;01W@RP0Id8X06e4:080L4<V00@0KD@X07=:;P1cBRh0Ndld0P1nDSL01020DSP0
Q5Dk08AE>`28FCl208eNA004099LA02FH4P0UV1809USB`80V69:00@0VfE=0:1ZDP2PJU80XVaD0P2P
JU80102VKUD0YfmF0:M_EP2VL5H20:=]D`0309aYCP2GIDX0UfE:00<0UVA900@0UFU>08iRA`2>HTL0
S6170P2?ID`0102>I4`0QEa608ELAP21F482089LA@0407aH?P1mFCl0OETo081K@@80PEa200@0Q5a3
08MOAP27GdH0S69:0P2?IDd0102CJU@0SFA>08eTCP27HDX208UUC@0408aZD@2<K5<0S6aC08e_E@80
SG9G00@0T7]P085/D@21K540O6A:0P1jH4L01024ITh0SFeF08e]EP2@KEL209IcG@<0YGaX00@0YGYW
0:5dG`2QM5l0VfeF0P2KJ5<0102NJU@0X6aF0:1/EP2WLV020:=^GP0409aVFP2EGeD0UEmE08iHCP80
QU1400@0ODPh07M2<01g@S00Nd@`0P1c?2P0101]=R80KcPT06lh901h@Bd2089;=`<0Qe0l00@0PDlf
085B>021DSP0QUe10P2>JD`0102EL5<0VG9E09UbE@2QN5`20:]oI0030;6;I`2`RVH0/8YV00<0[XQT
00@0/8YV0;:<J02bS6P0/XaX0P2aRfL0102WOUh0Z7mO0:QoG`2VOEd20:b3H`040:AkF`2QN5P0XGQH
09]bDP80WGAD00@0W6e?09i_D@2NKe40X75C0`2SM5H20:1aD`0409i_D@2LKDl0W6e?09UWC0<0Tf56
0P2@HDD308YM@00408MK?P1lD3<0O50c06m:;@80JdHY00@0OULj08YQ@`2:HD<0PEHf0P2>H3l0101l
BRT0VFA209UT@P1B91L2044L4`04054j=00o=C<0?cDc02@T90807b0T00@0?3e103<b=`0c<SL0=2`Y
0P0[8Qd01011=2h0?RlX03h_:00o<RX2038Y900403hh>00b<340<S0a030_:P80;bhY00@0>CPc03<b
;@0c<Rd0;RdX0P0_;RT0100V9B009bHQ02LV8@0f>3L203Li>00303Pj>@0f>3L0=SPg00<0=3He00@0
=SPg03Tk>P0i>cX0@D0n0P12@T001017AdL0C4e?04a=C`1ADU@2055EF004051DE`1?DeH0Ce=F04UB
F@80BE=L00@0BUAM04UFG`19EUl0BEIO0P16E5l00`15Deh0A59M04ABG@0304UCG`0404QBGP17DEd0
Ae5M04I@G080A4iJ0`13CET01013BeH0@TYE049:E@11BE@30418D`80@DUD00<0@TYE04=;EP13BeH0
0`13BeH304A<E`<0ADeH0`16CUT01014DEX0@e1I04=@F@12CeP2045>E`<0@4eF0`11CUL404E=F0D0
ATiI3`17CeX<04M@F@@0ATmH0`17D5T504QAFP<0BE1F00@0B55H04M@E`17D5L0AU1I0P15CeP304=@
F@05045?FP14CUL0A4iG04E>E@18CE400@3oool000`0BUQS1019Ef8304QFH@H0AeEP0P18EV401019
Ef80BUQS04YHH`1;FF@304YHHb40BEMR1P1:F6<504UGHP<0B5IQ0`17EF000`18EV40BEMR04UGHP04
04UGHP80B5IQ0`16E5l01015Ef<0AeUS04MIH`14EEd204QEG@0404iIG`0lADX0?4E:01dU:0802a<E
00@04Q4?01@C4@0D4a404a@?0P0@4@`0100A4P`051D=01@E3@0A50/200`?1P<0418?00@04A4?018A
3`0B4@l0510?0`0E3`l201L>3`0401hA200b91T0<R@I04Tf9`80G44`00@0ITH_06]3:P1[@bX0Kd<V
0P1fAbT0101`Ab/0MD``07E<<01lD3D207iB=`0407mA=`23E3X0Pe@j08UJ@080Sf1600@0TUa409IP
B02FH4P0V69:0P2GHDT0102JI4`0XF]C0:5[D`2SKED20:5[D`040:A/D`2VKUD0YViE0:M_EP80YFmE
00<0XF]A09]XC@2KJ4d00`2IId`0102GJDl0TfE;09=UB`2@I4/2095UC00408iTC02:HD/0RV5;08IM
A`80Q5]500@0O5Pn07mJ@01oFT00PUe30P25GD@01026GTD0RV1808YPB02>I4`2099XD004091VD02?
IDl0SfE?08eTCP80SFM@00@0TViF09MeG02GME`0U7AK0P2=KeD0102ENf80RViF08Y^EP2>L5P208i^
E`0409=`FP2JMV00VWIP09QbG@80XgYV00@0XgeZ0:AkI`2TNfL0XGIS0P2NLE`0102IJUH0V6QB09QX
DP2LJE@209m[EP0409ATC@2FIU80UVIB09MVE@80UFAF00@0T5mA08aKBP2<FdX0Re]70P2<G4D0102@
GTL0QeDn08ME?P1kBC8207=1:P<0LClX0`1`?RL0101d@RT0NDX`07U:<022FCd208]VB@04091[CP2C
K4l0Tfa?09acE`80ZGeR00@0[hUT0:b6H@2/QV40ZhEP0P2/QV400`2`RVD0/heX0;>=J0030;>=J004
0:Z1H@2XOel0Z7mO0:9iF@80[8=S0`2SNUX0102JLE40WGAD09edE02KK4h209e^D0040:5bE02SM5H0
XgAF0:=dEP80XG9D00@0WFi@09]/CP2KK4h0W6U>0P2DHTL0102EHdP0TfA809=TB02<GD4208YM@004
08UL?`21E3L0PE@g07Y<:080JCdJ00@0LDPX08YSAP2:HdH0PUTm0P2=HD@01020Cbl0XW1?0:9`C`1F
<b0203hR500404He:`0a;2H0<B`V02/[:`808R<W00@0>3Ll02XX;@0Z:2d0=R`Z0P0^8al01012=Bd0
>bdT03/]900k;RH203DZ9P04048m>P0c;c00<bl`02h]:080<c8]00@0?chi03<b;@0c<Rd09RDP0P0U
91l0100V9B00=SD`03He<00f=cT203Lh>P0403Pi>`0j>cd0>S/m03`m?`80?d1200<0@dA604I7B@16
AdT00`1>CE800`1?CU@0D55F051AEP03051CFP0404eAFP1:CUL0BTiG04YAE`80Be9H00@0BeAK04]E
GP1;EEh0BeEN0P18EEh00`17E5d0AU=L04ICG00304ICG00304EBF`14DEX0A55J00@0@e1I0P12CeP0
0`13BeH0@TYE049:E@060459E003049:E@13BeH0@d]F00<0@d]F0`14C5L304E=F0<0ATiI00@0A55J
04=@F@13D5T0@TmH0P11CUL3041=EP<0@DiG1015CEP504I>F@l0AdmJ3017D5T404I?F0<0Ae1I1@18
DEX304U@EP0404QAF017D5L0Ae1G04I@F@80ADmH0`13D5T01@11CeX0A4iG04A>E`15CUD0B4eA0040
oooo000<04YHH`@0BEMR0`18EV4604MEH080B5IQ00@0BEMR04YHH`1:F6<0BeUT0`1:F6<Q04UGHPH0
BUQS1@19Ef8304QFH@<0AeEP00<0B5IQ04UGHP19Ef801019Ef8204QFH@<0AUAO00@0AEMS04MIH`17
FF<0A5EM0P18EEd0101>FEl0?4E:03a5BP0M9BP200/C5@04018A3`0D4a4051<A01<D3`80414<00@0
4A8<01@E3@0D5@d04A@;0P0<3`H3010B3`04014A3`0B4@l04Q4?01@@3`<05@l?0P0G3Pl0100N4@P0
<R@I038T6@19=RL205a1<00406I6;`1[@bX0Jd<Z06m39P80MTLY00@0L4L[07E<<01eC300O50e0P1n
DSL0101oDCL0Pe@j08=D>P29FT0208mPAP04099LA02FH4P0UV1809QRBP80Uf5900@0VVA<0:5[D`2Q
Je<0XfeE0P2QJe<0102TK5<0YViE0:I^E@2WKeH20:E_E@030:5[D@2KJ4d0VfQ=00<0VFM<00@0UfU?
09=UB`2CID/0T6A;0P2AID`0102>I4`0RV5;08YQB`26GDL208AKA@0407aH?P1oFT00OeY0089M@`80
QEe400@0QUi508YPB02:H4P0SVA<0P2BJ500102@IU00SfE?08mUC`2=I4h208eWD004099^EP2GME`0
UgEL09AdF`80SFmE00@0UG]R08Y^EP2:KUH0SW1H0P2>KUL0102CL5X0VWIP09YfH02HLUd20:=jIP04
0:=mJP2TNfL0Y7]W0:5fH`80WW5L00@0VFYF09QXDP2HJ580W6UD0P2OJeH0102DI4d0UVIB09IVDP2G
IUD209ETEP04091OD@2<FdX0S5]:08]KA`80S5a500@0T5i708ME?P27ECh0NdTb0P1c@BX3074o:0<0
L3hW00@0M48Y07U:<01iBS00PUTm0P2;ITT0102@Jdh0Tfa?09=/C`2LLeL20:UmHP040:n9I02/QV40
[8IQ0:^5H080[8IQ00<0/8YU0;>=J02cSFP00`2cSFP0102ZPF40Z7mO0:QoG`2RNET20:b3H`<0XgYJ
00@0VW5A09edE02MM5@0Vfa>0P2MKU00102QLU@0XgAF0:=dEP2SM5H20:5bE00409e^D02KK4h0Vfa>
09aYCP80U69700@0UF=809=TB02CI4P0S5e10P2:GD001029G3l0PE@g085D=`1jC2P206Tm6P040758
:02:HdH0RV=6089I?@80SF5400@0P4l_0:9`C`2RL4l0ES<P0P0n8Q@01016=B/0<B`V034/9P0[:b/2
028S9`0403Pg?00Z:2d0:RP]03H/:P80;R<O00@0@SD]03/]900k;B@0>bhV0P0e:RH01012?CX0<bl`
03<_<00^;BP203<b;@0403ln>@0c<Rd0<c8]02HU80809B@O00@09RDP03He<00f=C00=SLi0P0g>3X0
100h>C/0>S/m03Xk?@0l?Cl203m0@P0304=4AP16AdT0ATM900<0CTeB00<0CdiD051AEP1@DEH00`1@
DeX0101=DEX0BTiG04Y>E`1:DEL204]BF00404]DF`1;EEh0BeEN04]EGP80B5EN00<0AeAM04ICG016
De`00`16De`00`15DU/0A55J04AAFP0404=@F@80@TmH00<0@d]F049:E@12BUD01P11BE@00`12BUD0
@d]F04=;EP0304=;EP<0A4aG0`15CEP304I>F@0404AAFP13D5T0@e1I049?F080@DiG0`10CEH3045>
E`@0ADeH1@16CUT?04M?FP`0Ae1I1016CeP304M@F@D0B55J0`19D5H01018DEP0Ae1G04M@E`16D5T2
04E?F0<0@e1I00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`00301:F6<404UGHP<0B5IQ1P17EF02
04QFH@0404UGHP1:F6<0BUQS04]II0<0BUQS6P19Ef8404QFH@80BEMR0`1:F6<304]II0@0BUQS0`19
Ef8604QFH@H0BEMR0P18EV4304IDG`0404AFHP19FfD0BE]U04IGG`80BEIN00@0CEQN03Q1AP0h@DH0
5alR0P084180100E51805QDA01HE4@0D4`l200l@2P04014B300D5`h051L>018E30802`h500@03Q<?
010B3`0@4Pl04A4?0P0B4@l00`0D40l05@l?01D?3`0301L>3`0402<E2P0l;1l0?2`O054k;@80GT<`
00@0K4Tc06i6;01^AR`0L4<V0P1kC2h0101aB2`0MTda07I=<@1lD3D207eA=P0407i@=P22DcT0PU<i
08YK@@80TF9800@0TUa409MQB@2GHDT0V69:0P2GHDT0102KIDd0XF]C0:5[D`2TKUH20:9/E0040:9X
D02TK5<0Y6aC0:I^E@80YFmE00<0XVaB09eZC`2MJTl00`2LJDh0102GJDl0UVQ>09IXCP2CIdh2091T
B`0408mUC@2>I4h0SVA>08UPBP80PeY400@0PUe308AOA@24GdD0QV570P28H4L01027GdH0REm708UO
A`2?HT/2099UCP0408YMB02>HD`0SV5<08mUC`80SfI@00@0UW1I09]iH02KNF00UgEL0P2:JU40102>
KeX0S6aG08a/E`2IMV8209MbG`0409=]FP2DJeT0U6]I09EZF@80YgUY00@0XWUW0:5hIP2QN6H0XGIS
0P2SMV<0102TMF40Xg=M0:=cG@2SL5/20:9^F@0409MYC`2KKED0VfeE09i_F`80VV]K00@0TfAD08eN
BP2=GTX0SV180P2AHdT0102;GD<0RUa208YL@P28FT02089D>P0407]=<`1eAbd0MDL]0753:@80Kd4W
00@0K3XQ07=4:P1cA2X0OeHj0P29I4L0102<IdX0SVM:08iWBP2HKe<20:IjG`040:R2G@2VP5/0YX1K
0:EoFP80Yh5L00@0[8IQ0;2:I@2`RVD0/XaW0P2aRfH0102]QF80ZH5N0:V1GP2OMe@20:Z2G`040:9j
E`2TO5T0Y7aI09]cD080WgMD00@0X79A09maD02OLE00Wg5@0`2NL4l309maD080X79A00<0WV]@09IS
B02FHdP00`2FI4T0102=GT80Rei108]N@@2:GD0208=F>@0407e98@1[>aD0Jc/E06Pm6`80RfA700@0
Q5a208]RAP2;HTH0QUTj0P2TLe<0101I?BL0?bXG03lZ5`0k;b4202DR6`04034c<P0U9RX09BHZ030^
<`808a`T00@0=RdX034V800a9R00ACH_0P0g:B00100g:240=b`X03L/:017@Sl2038^;`04030_:P0j
>C@0>STd04M6@@80=3<^00@07QdH01XI500J6A@08b8M0P0n?CP0100f>Sd0=c/n03Lk?P0j?T4203m3
AP0404E9C01:CU40BTiA04iBE@80DEEH00@0CTeE051?E`1@CeL0D51J0P1;BeD01014AU80ADMC04E7
D`1;Ce/2059FHP0405IKG`1AF5h0DEQN04aCF@80B55H00@0ATmF04MAFP17DEX0BE=L0P19EUl304EB
FP<0A55I1013D5P204AAF@0304=;EP12BUD0@TYE00H0@DUD00<0@TYE04=;EP13BeH00`13BeH304A<
E`<0ADeH0`16CUT01014DEX0@e1I04=@F@12CeP2045>E`<0@4eF0`11CUL404E=F0D0ATiI3`17CeX<
04M@F@@0ATmH0`17D5T504QAFP<0BE1F00@0B55H04M@E`17D5L0AU1I0P15CeP304=@F@05045?FP14
CUL0A4iG04E>E@18CE400@3oool000<0BeUT0`1:F6<00`19Ef80B5IQ04QFH@0304QFH@@0AeEP0`18
EV4504UGHP@0BUQS1@19Ef8304QFHA`0BUQS1@19Ef8?04QFH@0404aJI@1;FF@0BeUT04YHH`H0B5IQ
0P19Ef8304MDG@0404aFG`1=EUd0CEIM0417C@80:Bhb00<051PK00h?4@0>3a400`0E40X0100D4@X0
514<01@A300B4@d2014B3@<04A4?00@05a8>01HB3`0F4Pl03`h<0P0?3`l0100F5A<05A0<01D@300E
30D2024D2`0403@P5@17;b<0AblS05Xn<080HT4^00@0J48[0717:`1`Ab/0MdX[0P1kBRX0101iBc40
Nddc07]=<`1nD3H2081A=`04085B>027EC`0QeDl08aJ@@80TEe500@0T5i509=QB02CHDP0Uf=;0P2H
I4`0102KIDd0WfM@09mWD02RJU<20:E]EP040:9ZD@2SJe80Xf]B0:A/D`<0Y6iD0P2RKe@0102QKU<0
X6iC0:1^D`2DJ4d2095UBP0409=WC02GJe00Uf]@09=WC080RUi300@0QUXo08QL@@28G440Pe`o0P28
HD@01028H4H0QE]208EK@P23FD4208]NB@04091SCP2BHdl0TV=?08QJ@`80R5]400@0SF1909=YD`2C
JE<0UFmH0P2CKeT0102=JE<0QF9>08ERCP2BKe`209=`G@0308m/F02AKUX0TFiJ00<0QfA>0`2JN5l0
102PO6@0V7AL09QdG02IMEd209]gG@0409QeF@2KN5`0VgQL09ikG@80V7EG00@0UViE09E]E02EKE@0
TVYA0P2@J4l0102AJE00Tf]B09=[DP2FKUD209Q`E`0409=YD@2<HTX0S69:08QPA`80Rf=:00@0RfI<
08QSB@28HdT0QF570P26HTP0101mGT80OETo07eI?`22FT0208eQB00409AXC@2GJdh0Uf]>09UbD`80
WGUI00@0XWUK0:=jG02SNU`0YgiP0P2]Q6@0102_Qf@0[XIS0:j6H`2^QV820:n7H`040:j5I@2[PV80
Zh9R0:QoG`80YWeM00@0Y7]K0:1gE`2PMeL0Vg9B0P2FKDd0102RMeL0WW5B09iaDP2KKTl309i_D@80
WVe?00@0X6mA0:IeE`2VMEL0XVa@0P2HHTH0102GI4L0V6E809QUB02@Gd42095P@P0408mP@P24ECL0
Q5Dg085D=@80OU4b00@0EbX;08iO@@2>Gd40REXl0P25E3H01027ESP0WV]>09i[CP1bDD02048V6P04
030L5@18?C/0B3dk02XV9`80:BP]00@092DY03@e>@0d=CT0;BTV0P0a;2P0100l<b`0?C0W03d`9`0f
:B0203P_:00403dh=00h=340>3@a03Lc:P80>cL^00@0>SD_034^9`0a;RL09bHQ0P0X9b<0100c=2l0
@44l0411?010B4X20459C004049<CP15CU<0ADiC04IAE@80BEAJ00<0B5EK04UFG019EU`01019C5l2
04Y=H00404]>H@1<Cf80C4mR04iAI080Ce9U301;Deh304UAG00404Q@F`16CUT0ATiI04E=F080A4aG
00@0@d]F049:E@12BUD0@d]F0`12BUD30459E0<0@TYE0P13BeH404E=F0D0ATiI1017CeX204I>F@03
04E=F014C5L0A4aG00<0@d]F0`14C5LH04M?FP`0B55J1016CeP304M@F@D0B55J0`18CeD304I?EP03
04E?F014CUL0A4iG00<0@TmH00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`000`1;FF@304YHH`03
04UGHP18EV40B5IQ00<0B5IQ1017EF0304QFH@D0BEMR101:F6<504UGHP<0B5IQ701:F6<504UGHPl0
B5IQ00@0C5YU04]II01;FF@0BUQS1P18EV4204UGHP<0AeAM00@0C5IO04eFG@1=EUd0@4M=0P0Y;S80
0`0D61/03PlA00h?4@0301D@2P0401@A2P0D4@`0514<018A3@804A8=0`0A4@l0100G4Ph05Q8?01HB
3`0?3P`200l?3`0401HE4`0E40`05A0<01D<1@808A@;00@0=20E04L_8`17;b<0FSh`0P1R@Bh0101X
@R/0L4L[0717:`1gBR/207]::P0407U;<@1kCC<0Nddc07i@=P80P54g00@0PE8h08ME?027EC`0S5Y1
0P2AGDD0102@GTD0Tf5809=QB02GHd/209QTC00409]UC@2OIe00WfM@0:9ZD`80YFeF00@0XVYA0:=[
DP2SJe80Y6aC0`2TKU@20:9_E0040:5^D`2PKU<0X6iC09AXC@80TFE:00@0TfM<09M[D02GJe00TfM<
0P2:GT<01026FSl0R5a108QL@@23G3l208QQA00408QPAP25Fd80QE]208=I@@80Rei900@0T6=>099S
C`2BHdl0R5Y30P28Fd@0102=H4T0TfUC09=YD`2EKeP209=_F@0408eYD`25HTh0QF9>099_G080Tg1M
00<0SfaH095^FP2AKUX00`27I4h309YhG`040:1lI02HM5`0V7AL09UeG@80VgMM00@0V7EI09]hG02K
N5`0WW]M0P2HMEL0102FKUD0UFeD09E]E02BJU42091XC`04095YD02CJe80Tf]B09I^E@80V71G00@0
TfUA08aRBP2<HTX0R6170P2;HdX0102;IT`0R6=908QSB@25HDL208IRB00407eN@P1mFCl0OETo089J
@080SF5800@0U6Q=09M[CP2GJdh0VG9C0P2MNET0102RNE/0XgYL0:=jG02WOV020:f4I0040:n7I02^
QV<0[XIS0:j6HP80[hMS00@0[XEU0:^2HP2[PV80Z7mO0P2VOEd0102TNe/0X7MG0:1gE`2KLU8209I]
C@040:9gE`2NLE80WW5B09]^C`<0WVmA0P2NKDl0102PKe40YWEG0:IeE`2RK50209QRAP0409MTA`2H
IDP0V6E8091O@@80TF1200@0Sf1208AE=`24ECL0PE@e0P1nDC80101G:P/0SUm108iO@@29FS`208ED
=P0408MF>02NJdh0WV]>079A@080@RHJ00@0<1`E04Pm>`18?C/0:RHW0P0Y:2d0100T9BT0=3Di03@e
>@0]:BH2034/:00403`c;00m<2L0?C0W03HY8080>2lX00@0?CPd03Pd<@0h=340=c<Z0P0k=bh0100j
=Bl0<BhW034^9`0W9R4202PW8`0403<d;`10@C`0@44l0418BP80@DU<00@0@Ta>04E>D`15CU<0AU5E
0P19E5X00`18EE/0BEIL04UFG00404U<G`80BTeP00@0BdiQ04a?HP1<Cf80CU5T0P1?DVD<04]CGP<0
BE5L00@0B51K04I>F@16CUT0ADeH0P14C5L01013BeH0@TYE049:E@13BeH3049:E@<0@DUD0`12BUD2
04=;EP@0ADeH1@16CUT404M?FP80ATiI00<0ADeH04A<E`14C5L00`13BeH304A<EaP0AdmJ3018DEX4
04I?F0<0Ae1I1@18DEX304Q?E@<0ATmF00<0ADmH04A>E`14CUL00`12CeP01@11CeX0A4iG04A>E`15
CUD0B4eA0040oooo00000`1;FF@0BUQS04YHH`0304YHH`0304UGHP18EV40B5IQ00<0B5IQ1017EF03
04QFH@D0BEMR101:F6<504UGHP<0B5IQ701:F6<504UGHPl0B5IQ00@0BeUT04YHH`1:F6<0BEMR0P18
EV4304MEH0<0B5IQ00@0AU=L04QEGP18EEh0C5IO0P1:DeX0100j?d<08RHY028V:@0B4aD200d>40<0
5A0:00@0514:01@A300D4@`04Q4=0P0A4Pd3014A3`0401PC3`0E4@h05A4>010?3@8051@D00@06APF
01<>2P0C3PX05P`30P0W6Q40100k9AX0Bc@V04/d9P1L@C0206A3<00406U3;01_ARX0KdHZ07I9:P80
N4XY00@0ND/a07]=<`1kCC<0OU0f0P21DSP01022DcT0QeDl08ME?02<FT42095MA@0408mMA02CGdL0
Tem709MSB`<0W6I>0P2OIe00102RJU<0YFeF0:E]EP2QJE020:A/D`040:Q`E`2YLeT0ZG=I0:UcF@80
Y75F00@0X6eB09e[D02MJe00VVaB0P2FJ4h0102EIdd0V6Y@09QZD02FJ4h208mQA`<0Ree300@0R5m3
08]RAP2;HTH0RV170P25Fd801027FT<0S5m:08aOBP2?HTd208iQC00408mRB`2>HDX0SV5:08iTC080
U6]E00@0Vg9L09]eH02KMF00UG5K0P2?JUL0102FLf00U75N09AaGP2=JUH2091]F@0408YWD@2<JE<0
S6UC09mmI080W7YQ00@0WWQQ09QbF`2HLU/0VWAM0P2JME/0102EL5D0Ug9G09MbE`2IM5L209A_DP04
09Q`E`2GKeH0UfmF09I^E@80UFeD00@0UViE09Q`E`2HL5L0Vg=J0P2LM5/0102BJU40SFE<08eUC029
I4X308]WC@80R6I;00@0RVQ=091^D`2@KU<0Rfa@0P2:IT`0102=ID/0UFU@09EYD02IJe4209Q/C`04
09QaDP2LN5P0W7QH0:5hFP80YGaN00@0ZH1R0:f4I02]Q6@0/8QU0P2cRfP0102cRfL0/HUU0;69I@2b
RFT20:n6IP040:^2HP2XOel0Z7mO0:ElG080XGQH00@0W7=C09Q_C`2HKdl0Y7MH0P2OLU<0102LKe00
Wg1B09m`DP2NKe4209e/CP040:1]D02ULUD0YG9E09iXC080VVA80`2KJ4/0102AH480Sei008mN@02?
H48208YK?@04085D=@1lCc00O4l`05lb4`80Pe@f00@0Re`n08QG>@28EcT0QeHh0P2KJ4/01024HDh0
F3P[05Ph:`0h8AT204`n>`0402dW:@0W9bT09bLY02X[;`80>3Tm00@0:RHS02lZ9P0_:RH0>c8[0P0k
<BP0100d:R40>C0Y03T`:@0m>3@203Pd<@0403Ha:`0c;RP0<bhX02dZ9@80:RLR00@0;2/W03Hg<P0f
=c80ATI40P1BDU0304M?DP0404I@DP17D5D0Ae1E04IAE@80AU5G0`15DUX0101:CF00BDaO04U<G`19
C5l204Y=H00404]>H@1<Cf80C4mR04iAI080Ce9U301;Deh304UAG00404Q@F`16CUT0ATiI04E=F080
A4aG00@0@d]F049:E@12BUD0@d]F0`12BUD30459E0<0@TYE0P13BeH404E=F0D0ATiI1017CeX204I>
F@0304E=F014C5L0A4aG00<0@d]F0`14C5LH04M?FP`0B55J1016CeP304M@F@D0B55J0`18CeD304I?
EP0304E?F014CUL0A4iG00<0@TmH00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`00101:F6<504UG
HP<0B5IQ1017EF0304QFH@D0BEMR101:F6<504UGHP<0B5IQ6019Ef8404YHH`D0BEMR3`18EV40101:
F6<0BEMR04UGHP18EV4204MEH00404IDG`17EF00AeEP04MEH080B5IQ00@0AU=L04aFG`1<EUl0CUMN
0P14Be40100_=3P061`O01PL7`0>3a4200d>40<05A0:00@0514:01@A300D4@`04Q4=0P0A4Pd3014A
3`0401PC3`0E4@h05A4>014@3P805aLG00@06aLD01<<1P0C30H06Pd40P0c8aH01014;Al0DcTZ05<i
:P1P@c<206I6<@0406Y4;@1_ARX0KdHZ07E8:`80MdPZ00@0ND/a07a>=01lCS@0Oe4g0P22DcT01023
E3X0R5Hm08QF?@2=Fd82095MA@0408mK@`2CGdL0Tem709QTC080WFM?00@0WfM@0:1XD@2PJ540XfUC
0P2UJeD0102SJe80YViE0:I^E@2[LeX30:ihGP80YgAI00@0X6eB09YXC@2JJ4d0X75G0P2JJe40102F
Idd0V6U?09QYC`2IJU0209IWC@04099SB@2@HDL0T65708mSB080T6A900@0SF5808QL@`28G4<0REa5
0`2>HD`208YMB00409EXD@2?IDd0SfE=08iTC080U6]E00@0WGAN0:1jI@2PNVD0W7IQ0P2EL5d0102P
NfT0W7MU09agI@2BKEX209A_G00408eYD`2>JU@0SVYD09eiH@80Ug=K00@0YGmX0:1jH`2PNV<0XgeV
0P2TOfD309agG00309iiG02IM5L0VGAG00<0Vg5H00@0W79I09ecFP2MLeX0WWAK0P2OME`30:1fG@04
08m[D@2<J4h0S6Q>08YXC@80SF]@00@0S6eA08Y/D02:K500SW1D0P2EMe/0102IMe`0UW9H09IbF02H
KUD209eaF00409e_E@2KKe80VfmB09acE@80X7UJ00@0XgMJ0:b0H`2/P6<0/8AW0P2_Q6@0102bQfD0
^Xm]0;Z?K@2kT6d20;F:I`040;J;J@2cR6H0/hQV0:j3H@80Zh1N00@0Z7eK0:EjF02UNUP0XGID0P2O
M580102UN5L0X7=B0:1cDP2OLE020:1bD@040:5`D02NJd`0WV]<09m/C@80Xg1A00@0VVE609]VA`2K
ITL0WFY;0P2LJDX0102BHD40Sehn08mN?P2?HD0208iP?`04089E=01jCB`0NTd/06Xm7080LdDT00@0
Redl08aK>`2<Fc/0QUDe0P2HIDH0102>J5<0JDPg06U8=`0n8aP204/l=`04034X:@0S8B@08b4T030`
<P80?3`n00@09bHT02hY9@0^:BD0>S4Z0P0k<BP0100e:b80>S4Z03Xa:P0n>CD203He<`0403Pe<00b
;bX0<RlZ02/Z9P80;B`X00@0=cLe04A4A014A4@0CU1?0P1CEE@304YBE@<0B55F00@0Ae9H04MAFP17
DEX0AE9J0P14DET604Y=GP0404]>G`1<Cf00C4mP04iAHP80Ce9S301;Deh304UAG00404Q@F`16CUT0
ATiI04E=F080A4aG00@0@d]F049:E@12BUD0@d]F0`12BUD30459E0<0@TYE0P13BeH404E=F0D0ATiI
1017CeX204I>F@0304E=F014C5L0A4aG00<0@d]F0`14C5L<04I>F@`0AdmJ3018DEX404I?F0<0Ae1I
1@18DEX304Q?E@<0ATmF00<0ADmH04A>E`14CUL00`12CeP01@11CeX0A4iG04A>E`15CUD0B4eA0040
oooo000404YHH`D0BEMR0`18EV4404MEH0<0B5IQ1@19Ef8404YHH`D0BEMR0`18EV4H04UGHP@0BUQS
1@19Ef8?04QFH@0404YHH`19Ef80BEMR04QFH@80AeEP00@0AUAO04MEH017EF00AeEP0P18EV401016
De`0C5IO04aFG`1>Eeh204A;D@0402ld>00H71l061`O00h?4@803@h@0`0E40X0100D4@X0514<01@A
300B4@d2014B3@<04A4?00@061<?01DA3P0E4@h04A0>0P0G5aL0100K5a@04``601<<1P0J3@@203<S
5P0404@]7`1C>BX0DcTZ0613<`80ITHa00@0JT@]06m6:P1_ARX0MDP[0P1gB2X0101iBc40O4hd07a>
=01oDCL2089C>@0408=D>P28ESd0R5Hm08eK@P80TEe500@0Se]309=OA`2CGdL0V6A<0P2MIdl0102O
Ie00X6QA0:1XD@2SJE<20:E[E@040:=[DP2VKUD0YViE0:]cFP<0[WQN0P2WM5T0102PKE80VVQ=09YX
C@2PLEL209Y[D@0409IWC@2HJDl0V6U?09UZD080UVM=00@0TV=9091QA`2@HDL0Sf=80P2@I4T0102=
HDP0R5a308QL@`29G4D308iQC080RUe800@0UFQA08mUC@2?IDd0SVA<0P2DJeD0102MM5h0X7YU0:1j
I@2LMV4209E`G@040:1kJ@2LMfD0W7MU099]FP80U6mL00@0SFUC08iZE02>JU@0WGUQ0P2GLe/0102U
OfP0X7YS0:1jH`2SOFH20:AoI@<0W7ML00<0WWUL09UdE`2IM5L00`2KLEP0102LLUT0WG=J09ecFP2N
M5/209meG0<0X7IM00@0Sf]A08aXCP2<J4h0RVQ=0P2=Je00102<KE40RVa@08Y/D02>L5@209EgF`04
09UgG02FLUP0UW9H09Q^E@80WG5H00@0WFmE09]_DP2KKe80W7=E0P2PNEX0102SMeX0[81S0:b0H`2`
Q6L20:n4I0040;:7I@2jSfd0^Xm]0;^@K@80]HYW00@0]X]Y0;>8IP2cR6H0[X=Q0P2[P5h0102XOE/0
YGYH0:EjF02QMU@209mdDP040:EhE`2PLe80X7=B09maD080X79A00@0XG1@09i[C02NJd`0Wfa=0P2S
L540102JIDH0VfI709]VA`2MJT/209aYBP04099Q@@2?GSh0Sehn08mQ@080SV0o00@0PUDd07Y=;01j
CB`0JSdL0P1cAB@0102;GC`0S5/k08aK>`26ECD209QUAP0408iXD`1YB3L0JDPg03hS6080Bc`g00@0
<BPY02<Q900S8B@0<30b0P0l?3h0100W9R@0;RTU02hY9@0j<BX203/a:00403D[8P0j<BX0>S4Z03hi
=@80=SDc00@0>3D`038_:P0b;bX0:bXV0P0];2P0100g=cD0A4A404A4A01>D4l205=EE0<0BU9E0`18
DEH01017DUP0Ae5J04MAFP15DUX204AAF@H0BTeN00@0BdiO04a?H01<Cf00CU5R0P1?DV<<04]CGP<0
BE5L00@0B51K04I>F@16CUT0ADeH0P14C5L01013BeH0@TYE049:E@13BeH3049:E@<0@DUD0`12BUD2
04=;EP@0ADeH1@16CUT404M?FP80ATiI00<0ADeH04A<E`14C5L00`13BeH304A<E``0ATiI3017CeX<
04QAFP@0ATmH0`17D5T504QAFP<0B4mE0`16CeH00`15CeP0A4iG04A>E`03049?F005045?FP14CUL0
A4iG04E>E@18CE400@3oool000<0BUQS2@19Ef8404MEH0<0B5IQ1@19Ef8404YHH`D0BEMR6`18EV44
04YHH`D0BEMR4P18EV400`17EF00AUAO04IDG`0304IDG`0404MEH018EV40B5IQ04YDG@80CUQQ00@0
C5EL03a3B@0l@dT09BT/0P0B4aD00`0=3@d0410@010@400301D@2P0401@A2P0D4@`0514<018A3@80
4A8=0`0A4@l0100D40d05A@B01DD4P0C4Q02018A3`0401DA3P0F3@H05Pd602<E2P80>RPJ00@0CS8T
05Pk:`1H>b/0Hd@`0P1ZAc40101/Abd0L4L[0717:`1fBB`207Y;;@0407Y<<P1mCcD0ODle081B>080
Pe@j00@0QEHl08UG?P29Ech0SE]20P2BGTH0102?Fd<0UEm709EOA`2KIDd209mWD0040:1XD@2RJ580
XVQB0:=YD`80YF]E00@0YfmD0:UaEP2YLEH0[GEJ0`2`NUh20:YgFP040:9_DP2KJT`0VfY<0:9`E`80
VfU@00@0UVA;09MUC02GID`0VfU@0P2LJU40102IIdh0UVA;09ITB`2EIdd209AVC004091RBP2;GDD0
Ree508]NA`80Sf9=00@0SF1;08EKA@25FdD0TFM?0P2<I4/0102;HdX0TfYD09=ZE02NMEl20:=jIP04
09mfHP2FL5d0UW1M09A_G@80UG1N00<0TfiK09QcH02HLf000`2BKUP0102ROVH0WWYR09ijHP2[PV`2
0:QoJ@030:b3K@2]QF`0[HE/00<0YGeS00@0Yh1S0:=lG`2SO5l0WG=J0P2NM5/0102PMUd0XGMN0:5g
GP2RN5l20:5gGP040:1fG@2OME`0WgEL095_E080RV]?00@0RFY>08e_D`2=Ke<0TG=G0`2?M5L2091g
F@0409YhG@2HLeT0V7=I09Y`E`80X79J00@0X75G09iaE02NLE@0X7MI0P2VOf00102WNeh0/HEX0;65
J02dR6/20:n4I0040;:7I@2lTFl0_95_0;fBK`80]8UV00@0/hQV0;66I02aQV@0[H9P0P2ZOed0102X
OE/0YW]I0:IkF@2TNEL20:=hEP040:MiF02SME@0XgED0:5cDP80Y7=C00@0XW5A09i[C02NJd`0WfY;
0P2RKDh309UTA@<0VFI700@0U6=3095P@02AH400SV0o0P2<GSd01023ESD0ND`[07U<:`1eB2L206Li
600408QJ>@2>GCd0SUdm08ED=080Tf1100@0SV9907==>01cCCP0?b8D0P18=Bh0100c:BL08AdN024M
7P0b<3<203Xj?00402TX9P0_:bP0;b/X03/d;P80?3</00@0=B`U03Tb;00i<R`0?3Pe0P0f=C<303<b
<00403De=@0n?Sh0?Shn04M8BP80Bdm@00<0C51C04]?DP1;Ce800`18CeD01018DEH0BE9I04UBF@19
E5X204UCG0<0B5EN00<0BDmO04Q>GP18CUh01018CUh204Y@H00304]AH@1<DV80C59R00`0Be=N0`19
DE`01018D5/0ATiI04I>F@15CEP204A<E`0404=;EP12BUD0@TYE04=;EP<0@TYE0`11BE@3049:E@80
@d]F1015CEP504I>F@@0AdmJ0P16CUT00`15CEP0A4aG04A<E`0304=;EP<0A4aG3016CUT<04M?FP`0
Ae1I1016CeP304M@F@D0B55J0`18CeD304I?EP0304E?F014CUL0A4iG00<0@TmH00D0@DmJ04A>E`14
CUL0ADiE04Q=D@010?ooo`002@19Ef8304YHH`@0AeEP0`18EV4504UGHP@0BUQS1@19Ef8K04QFH@@0
BUQS1@19Ef8@04QFH@80AeEP1016E5l204MEH0<0B5IQ00@0C5IO04iGGP1>Eeh0AdiD0P0c>3`0100L
82<03a0B00l@4P0?3`l201@C4@<05A0:00@0514:01@A300D4@`04Q4=0P0A4Pd3014A3`0400h=2`0H
5aD061LE01<B408020L500@040/7020F3@0P5Pd0<AlA0P10:aX0101F=bD0Fc`X05/l:01V@bl206a8
<P040718;`1aB2`0LDP/07M;;P80O4lb00@0NT`b07e?=@1mCcD0PE<i0P25ES`01026Ecd0RUPo08YH
?`2>G4<2099NAP04099LA02FH4P0UV1809aTC@80WfM@00@0XFMA0:9XDP2RJ580YVUD0P2XJeH0102/
MEL0ZgAF0:]dEP2/MEL20:iiFP040;1kG02/NEX0[7UJ0:MdE@80XG1@00<0WVYB09UUC@2IIDd00`2E
HDT0102JITh0WVYB09iZDP2MJE4209UUC@<0V6I=00@0TV9;08aNAP2<GTH0SEm80P2?HTd0102;HD/0
Q5Y408AJA02;IT`208]SBP0408iVC@2FKEL0UVeG0:5hHP80Y7]W00@0WgIR09U^G@2IKUd0TFYI0P2H
LF00102KMF80X7YW0:1jI`2EKeX2091ZE@0409miHP2LMUl0W7IO0:V0JP80Y7]U00@0Z7mY0:Z2J@2Z
PVT0Y7aR0P2UOF<0102XPF@0XgaO0:=lG`2QMEX20:9fF`@0Y7QM0P2RMU/0102OLeP0WW9G09ibE`2G
MEX208a]D@0408UZCP2@LUH0T79F09QjGP80UGYM00@0SW=F08U`DP29L580VWIL0P2HLeT0102LL5L0
WfmH09m_F02NKeD209e^DP040:5eF02WOV00YgiP0:aoHP80]XU/00@0]hY]0;23I02`Pf@0/HAS0P2k
SVd00`2kSf`0/XIS0;:6H`030:j2G`040:b0G@2[Oe`0ZgmL0:UmFPD0Yg]H00@0YgUG0:=eD`2SME<0
Y7=B0P2WME@0102ULe80XFa<0:5/C02OJTX20:5/C00409]VAP2HHd<0V6=309=Q@080U69100@0UfI5
09IUA02FID@0S5hl0P26F3H01025FCH0O50]07a@;@1jCR/206@f5004085C<@2>GC`0SUdl08AC<P80
SU`k00@0T5m107mC>P1oDcX0BRTF0P1<=R/0100l;bT091`J02@L6P0`;2d203Dc=00402d];@0a<2h0
<C0^03`g<`80?C@]00@0=RdV03Pc;`0h<bl0?3/i0P0g=cL0100^;S00=CHh03Df>011@T@204Q<C`04
04e@E@1=DUH0CE9F04eBF080Be9H0`19D5H01019DUT0BE9K04UBF`19De`204UCG@<0B5EN00<0BU1N
04U?G@19Ced01018CU`204U?G@0304Y@GP1;DEl0Be5O00`0Be=N0`19DE`01018D5/0ATiI04I>F@15
CEP204A<E`0404=;EP12BUD0@TYE04=;EP<0@TYE0`11BE@3049:E@80@d]F1015CEP504I>F@@0AdmJ
0P16CUT00`15CEP0A4aG04A<E`0304=;EP<0A4aG3016CUT<04M?FP`0Ae1I1016CeP304M@F@D0B55J
0`18CeD304I?EP0304E?F014CUL0A4iG00<0@TmH00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`00
2@19Ef8304YHH`@0AeEP0`18EV4504UGHP@0BUQS1@19Ef8K04QFH@@0BUQS1@19Ef8@04QFH@80AeEP
1016E5l204MEH0<0B5IQ00@0C5IO04iGGP1>Eeh0AdiD0P0c>3`0100L82<03a0B00l@4P0?3`l201@C
4@<05A0:00@0514:01@A300D4@`04Q4=0P0A4Pd3014A3`0400h=2`0H5aD061LE01<B408020L500@0
40/7020F3@0P5Pd0<AlA0P10:aX0101F=bD0Fc`X05/l:01V@bl206a8<P040718;`1aB2`0LDP/07M;
;P80O4lb00@0NT`b07e?=@1mCcD0PE<i0P25ES`01026Ecd0RUPo08YH?`2>G4<2099NAP04099LA02F
H4P0UV1809aTC@80WfM@00@0XFMA0:9XDP2RJ580YVUD0P2XJeH0102/MEL0ZgAF0:]dEP2/MEL20:ii
FP040;1kG02/NEX0[7UJ0:MdE@80XG1@00<0WVYB09UUC@2IIDd00`2EHDT0102JITh0WVYB09iZDP2M
JE4209UUC@<0V6I=00@0TV9;08aNAP2<GTH0SEm80P2?HTd0102;HD/0Q5Y408AJA02;IT`208]SBP04
08iVC@2FKEL0UVeG0:5hHP80Y7]W00@0WgIR09U^G@2IKUd0TFYI0P2HLF00102KMF80X7YW0:1jI`2E
KeX2091ZE@0409miHP2LMUl0W7IO0:V0JP80Y7]U00@0Z7mY0:Z2J@2ZPVT0Y7aR0P2UOF<0102XPF@0
XgaO0:=lG`2QMEX20:9fF`@0Y7QM0P2RMU/0102OLeP0WW9G09ibE`2GMEX208a]D@0408UZCP2@LUH0
T79F09QjGP80UGYM00@0SW=F08U`DP29L580VWIL0P2HLeT0102LL5L0WfmH09m_F02NKeD209e^DP04
0:5eF02WOV00YgiP0:aoHP80]XU/00@0]hY]0;23I02`Pf@0/HAS0P2kSVd00`2kSf`0/XIS0;:6H`03
0:j2G`040:b0G@2[Oe`0ZgmL0:UmFPD0Yg]H00@0YgUG0:=eD`2SME<0Y7=B0P2WME@0102ULe80XFa<
0:5/C02OJTX20:5/C00409]VAP2HHd<0V6=309=Q@080U69100@0UfI509IUA02FID@0S5hl0P26F3H0
1025FCH0O50]07a@;@1jCR/206@f5004085C<@2>GC`0SUdl08AC<P80SU`k00@0T5m107mC>P1oDcX0
BRTF0P1<=R/0100l;bT091`J02@L6P0`;2d203Dc=00402d];@0a<2h0<C0^03`g<`80?C@]00@0=RdV
03Pc;`0h<bl0?3/i0P0g=cL0100^;S00=CHh03Df>011@T@204Q<C`0404e@E@1=DUH0CE9F04eBF080
Be9H0`19D5H01019DUT0BE9K04UBF`19De`204UCG@<0B5EN00<0BU1N04U?G@19Ced01018CU`204U?
G@0304Y@GP1;DEl0Be5O00`0Be=N0`19DE`01018D5/0ATiI04I>F@15CEP204A<E`0404=;EP12BUD0
@TYE04=;EP<0@TYE0`11BE@3049:E@80@d]F1015CEP504I>F@@0AdmJ0P16CUT00`15CEP0A4aG04A<
E`0304=;EP<0A4aG3016CUT<04M?FP`0Ae1I1016CeP304M@F@D0B55J0`18CeD304I?EP0304E?F014
CUL0A4iG00<0@TmH00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?ooo`000`18EV4404UGHPD0BUQS1017
EF0304QFH@D0BEMR101:F6<504UGHP<0B5IQ6019Ef8404YHH`D0BEMR4018EV4304MEH080AUAO00@0
AeEP04QFH@18EV40BEMR0P1:F6<0101=EUd0B55H04QAF00k@TP202HZ;@0401HG6@0@4100410@01<B
40805QDC0`0E40X0100D4@X0514<01@A300B4@d2014B3@<04A4?00@03@`:01PG5@0H5aD0410@0P06
1@<0100G40X0<B<H034S600m:Q/204@[5`0405`l9`1P?BT0H3dY06M3;@80KTPa00@0LDT`0757;P1a
Abh0MTX_0P1lCc80101jC380OU0f07i@=P22E3X208IG?@0408QI?`2<FT40S5Y108mMA080TUi600@0
UUi709QPB@2HH4T0W6A=0P2OIDl0102PIU00Y6MB0:AWDP2WJUD20:]]F0040:efF02[M5H0ZgAF0:Yc
E@80ZgIG00@0[gYK0:mlG@2_O5d0[7UJ0P2XMeL0102MIdl0W6I>09aVCP2IHd/209MQB@0409YTC02O
JE40WfUA09iXD080VVA<00@0V69:09MSB`2GHd/0U69;0P2>GTL0102=GdP0Sf9=08mRC@2;HD/208EL
AP0408YVC02<Idd0S6M=095/DP80Vg9L00@0XWUS0:IkJ02VNfP0XWMT0P2OLF40102NM6@0XgUY0:=i
J@2TNfT20:ImJ`0409U`G02@Ie<0T6MC09]bG080UFaF00@0[XA^0:IlIP2VO6H0YgeW0P2YOfL0102V
O6<0ZGmV0:UoIP2/PfL20:MnHP030:5eFP2RMU/0XWIK00@0XgML0P2QMEX0102OLeP0WW9G09ibE`2M
NEl209=_E@0408m]DP2FM5T0UWAI09amH@80VG]O00@0TG=G08e_D`2=Ke<0WWYP0P2NMUd0102OLeX0
X71I0:1`F@2NK5<209UZCP0409e`D`2UNE`0YGUL0:n0I080]HIZ00@0]XM[0;21H`2`PF<0/H=R0P2g
RFP0102gRFL0/H=Q0;63H@2[OET20:eoF`<0[X1L00<0ZgeI0:UkE`2YNeL00`2XNUH0102XN5@0YGEA
0:EeD@2UME420:UgE0040:QcD@2RKD/0XVe;0:1ZB080XVa:00@0WFQ609YU@`2JID<0TEll0P2CHCh0
102KJdL0V6Q409QXA02:G3P208AF<P0408MK=P20E2l0P5@_07Y>:@80JcdI00@0MDLS08YJ=P2:FSH0
Q5@`0P29Ec@0102KID<0SUdo08iM?`1P>RD205@j;@0404Pg<00/8R00;28P02`V9P80=30_00@0<S8b
03Dd<P0e=380?STe0P0m=S00100f;bT0>cHb03/f<P11@3h203dm?@0403/o@P12ADX0@TE:04Y=DP80
BTmE00@0AdaB04A;D`14Be<0AdiF0P19DU/304eDG00404]DF`1:De`0BU=L04QBG080Ae5K0`14D5`3
04UAG00404Q@F`17CeX0AdmJ04M?FP80B51K00<0BE5L04YBG@1:DUd0301;Deh304UAG00404Q@F`16
CUT0ATiI04E=F080A4aG00@0@d]F049:E@12BUD0@d]F0`12BUD30459E0<0@TYE0P13BeH404E=F0D0
ATiI1017CeX204I>F@0304E=F014C5L0A4aG00<0@d]F0`14C5L<04E=F0`0AdmJ4016CeP304M@F@D0
B55J0`18CeD304I?EP0304E?F014CUL0A4iG00<0@TmH00D0@DmJ04A>E`14CUL0ADiE04Q=D@010?oo
o`001018EV4204UGHP@0BUQS0P1;FF@404MEH0<0B5IQ1@19Ef8404YHH`D0BEMR0`18EV4L04YHH`D0
BEMR4P18EV4304MEH00404QFH@19Ef80BEMR04YHH`80BeUT00@0BU=J0458CP11B4h0;S<g0P0H71l0
100>3a404A4A014A4@0E518201LC40<05A0:00@0514:01@A300D4@`04Q4=0P0A4Pd3014A3`0400h=
2`0E5AD05ADE00l?3`803`/800@09ahG044a9011<B@0Ac8Q0P17;1L306E1:`0406U3;01^BBl0KTT_
0759;`80KdD/00@0M4P]07U=<P1iCC80NT`b0P1nD3H01023EC/0QePn08MH?P29FT0208aJ@@0408mM
A02BGTH0TUi609QPB@80VV9;00@0W69<09eSC@2MHdd0X6=>0P2SIU40102YJeH0[FmJ0:e_FP2[M5D2
0:QaDP040:M`D@2ZMED0ZWEE0;5lG080/X1O00@0/7iM0:]jF@2[NUT0X6QA0P2RJU<0102PJ540Vf=<
09]SC02KHd`209iVC`0409eUCP2IHDX0VF5:09ALA@80Uf5900@0UV9<095OB02AGdP0S5i70P2=H4/0
102<HT`0R5m908QOB@29ID/208]WC@04095/DP2HKeT0V6mI09acG@80WgAQ00<0XGAQ0:5cH`2QLf<0
0`2OMFD0102LLf40XgYX0:=jJ02OMV8209ieH@040:Z1J`2QN680XGQR0;>9L`80YgeW00@0Y7YT0:Ek
H`2UNf<0YG]R0P2ZP6L0102/PfL0Y7]O0:AkG`2QM5L20:9eF0030:=fF@2TMeX0Y7MJ00H0Y7MJ00@0
XWYQ09eeG02MME`0V7=I0P2KMed0102LN5h0VGML09UgG02IMe`209ajG`040:AoI@2SNf80Xg]R0:Eg
G`80YG=L00@0X6iE09aZC`2LJTl0WW5D0P2WNeh0102^Of<0/85U0;21I@2`PFD20:n0HP040;22H@2b
Q6<0/XAS0;:4HP80/H=Q00@0[GmK0;22GP2`PUh0/hEQ0P2bQ600102_PEd0ZWaH0:YlF02XNUH20:Mi
E@040:MgD`2TM500Y7A@0:IdD@80[7ME00@0ZWEC0:A^C02TKT`0XF]90P2RK4X0102NJDL0WfY809mZ
B02GID8209IT@@0409e]B@2HJ4@0V6Q408YL>080QeTe00@0REdh08EI=025FC@0MTXU0P1gBBD0101Z
?1P0QEDa08EE<@23Dbl208ID<@040:5W@P2DGd00U5m006e5;080ESTY00@0CS/d03@Y9@0d:BD0:B@Q
0P0g<c80100d=CL0=SHf03Hf=P0n>SL203hg<@0403Pa:`0n>SL0?SXg04Q8B080ATM900@0BTeD04aA
E`1<DEL0CE9H0P18CeL01013BU@0?dQA03m8D@11BE@204A>F0<0Be9J00@0BU=L04UAG019DE`0Ae5K
0P17DEd304A@G00404YBG@19DE`0BE5L04Q@F`<0AdmJ0P18D5/304UAG0`0Be=N0`19DE`01018D5/0
ATiI04I>F@15CEP204A<E`0404=;EP12BUD0@TYE04=;EP<0@TYE0`11BE@3049:E@80@d]F1015CEP5
04I>F@@0AdmJ0P16CUT00`15CEP0A4aG04A<E`0304=;EP<0A4aG3015CEP<04M?FQ00ATmH0`17D5T5
04QAFP<0B4mE0`16CeH00`15CeP0A4iG04A>E`03049?F005045?FP14CUL0A4iG04E>E@18CE400@3o
ool000@0B5IQ0P19Ef8404YHH`80BeUT1017EF0304QFH@D0BEMR101:F6<504UGHP<0B5IQ701:F6<5
04UGHQ80B5IQ0`17EF001018EV40BEMR04UGHP1:F6<204]II00404YCFP11B4h0@DQ>02hc=`8061`O
00@03PlA014A4@0A4A405A@B0P0G4a0301D@2P0401@A2P0D4@`0514<018A3@804A8=0`0A4@l0100>
3@/05ADE01DE5@0?3`l200l;200402LN5`11<B@0@C4T04Lb8@80Ab`G0`1U@B/0101Y@b`0KTT_06i9
;`1aBBl206m5;00407A8;@1iCC80NDdb07Y<<P80OU0f00@0PeDk08MH?P27F3h0REY00P2<FT40102?
GD@0TUi6099NAP2HH4T209YRB`0409aRC02MHdd0WF==0:1SCP80XfIA00@0ZF]F0:e_FP2]KeX0ZgAE
0P2XLE80102WL540ZWEE0:YeE@2aO5`20;:0G`040;1nG@2[NUT0ZgYI0:1XD@80XVYC00@0X6QA09]S
C02KHd`0Vf=<0P2NITl0102MIDh0VF5:09UQBP2DG4D209MQB@0409IRC02AGdP0TEm808aNA`80SF1;
00@0S69<08QOB@28GdT0RFE;0P2;Idd0102AK580V6mI09Q_F@2LLed209mdH@030:5dH@2QLf<0XG=S
00<0WgEU00@0W7=Q0:=jJ02SNVP0WgIR0P2NMF40102ZPF/0XGQR0:5hHP2cRG<20:MmI`040:AjI02U
Nf<0YG]S0:EkHP80ZX1W00@0[8=W0:AkG`2TNel0XGAG0P2RMEP00`2SMUT0Y7MJ0:AgFP060:AgFP04
0:9jH@2MME`0WGEL09QcF@80VgMM00@0W7QN09UgG02IMe`0VGML0P2LNUl0102TOfD0Xg]R0:=kHP2U
Mel20:EcG0040:1^E@2LJTl0W6Y?09iaE080Yg]N00@0[WmS0;21I@2`PFD0/85U0P2_P680102`PV40
/XAS0;:4H`2bQ6820;63H@040:eoF`2`PUh0/89N0;>5H@80/XAP00@0[h5M0:YlF02ZO5P0Z7YF0P2W
NED0102WMe<0Y7A@0:AdD02VM5420:agE@040:YeD`2TKT`0Y6i<0:5[B@80XVa:00@0WVU709mZB02O
JTP0UfE20P2FI440102MKDT0V6Q409QXA02:G3P208MI=@0408UM>025FC@0QETd07I:9@80MdTU00@0
JS`H08EE<@25EC40Pe<_0P26E340102QId80U5m009AO@01]AB`205Hi:@0404hk=00d:BD0=2TU02TT
8@80=c<b00@0=3Dg03Hf=P0f=SH0?SXg0P0n=c40100h<B/0?SXg03hj=`18B4P204I7B@0404Y=E01<
DEL0C55G04eBF080B4mG00@0@dYD03m8D@0oB540@DUD0P14CUP304]BFP0404YCG019DE`0BE5L04MA
F`80Ae5M0`14D5`0101:DUd0BE5L04UAG018D5/304M?FP80B51K0`19DE`<04]CGP<0BE5L00@0B51K
04I>F@16CUT0ADeH0P14C5L01013BeH0@TYE049:E@13BeH3049:E@<0@DUD0`12BUD204=;EP@0ADeH
1@16CUT404M?FP80ATiI00<0ADeH04A<E`14C5L00`13BeH304A<E``0ADeH3017CeX@04I?F0<0Ae1I
1@18DEX304Q?E@<0ATmF00<0ADmH04A>E`14CUL00`12CeP01@11CeX0A4iG04A>E`15CUD0B4eA0040
oooo000404QFH@80BEMR0`1:F6<304]II0@0AeEP0`18EV4504UGHP@0BUQS1@19Ef8304QFHA`0BUQS
1@19Ef8?04QFH@0304UGHP18EV40B5IQ00@0B5IQ0P1:F6<0101;FF@0C5YU04aJI@18CeD203]2B004
02@Y;@0>4QD03Q8E00X;3@804Q4?00<05A@B01DA3P0E4@h00`0E40X0100D4@X0514<01@A300B4@d2
014B3@<04A4?00@0410@018B4P0B4Q803@d=0P0H5140100e;2D0BSX]04Xj;@1;=28204X^600406Y6
;P1X@R/0J48[06Y4;@80LDT`00@0LDL^06e3:P1]@bX0LDDZ0P1gBc00101kCC<0Oe4g07mA=`24ES`2
08QI?`0408UJ@02=Fd80SE]208mMA080Tem700@0VF5:09YRB`2JHT/0W69<0P2NHD`0102OHTd0Y6IA
0:AVD@2ZK5L20:maG0040:M`D@2VKe00YVm@0:I_D080ZgIF00@0/giN0;B2H@2dPV40/GmN0P2/NeX0
102TJU@0Z6iH0:Q^F02WKEL20:1VD00409aRC02NI4h0WVA>09aRC080Uee700@0TEM109MOB02GGdP0
V65<0P2AGdP0102;GDH0S5m:08aOBP2<HT`208YQB`0408ESB028I4X0R6A:08]WC@80SfUB00@0U6YD
09U/F@2IK5T0W6mL0P2PLF40102YOfl0X7IV0:1fIP2FKE/209mfI0040:=jIP2XOf/0Z7m[0;:9L`80
YGaV00@0/XE`0:=fH@2SMV40W6mJ0P2ML5T0102MLEP0XgMN0:=gGP2TN5d209]_E0040:1cEP2QM5L0
XGAG0:9eF080XgII00<0YGQK0:MjG@2WNUd00`2YO5l0102TNV80XWQP0:9hH02QNF020:1hG`0409Ye
F`2HLeT0V7=I09mkH@80ZXI/00@0YX5W0:IlI02VO6@0YgUQ0P2XMUl0102TL5P0X6iC0:1^D`2SMUT2
0:f1I0040:amH@2[O600ZgaP0:]lH080[WmQ0`2_PF00102`PV00/hES0;>5H`2`PUh20;B6HP040;R:
IP2gRFD0]hUU0;63G`80ZgeI00@0YgUE0:EgD`2UMe<0YgEB0P2ULe00102VM540[7ME0:agE@2[MU@2
0:E_C@040:5[B@2SJdX0Xf]:09eXAP80Y6m=00@0WVa909]YAP2KJDH0WVi:0P2GId<0102:G3P0S5hj
08aN>P2:GST208QL=`0407968@1oDBd0Oe4]068d4080PE4]00@0Pe<_08AB;`24DRl0WV8m0P2AFcT0
101`ARd0DC@T054d901>?38203T^:P0402/S8@0l=SH0?3Hf03@e=`80=SHf00@0?CTf03`g<`0l=c<0
=c8^0P12?S/0101=CDd0CDi@04e>D016BU<204M;E00404M>EP18CeT0B4mI04I?F080ATiI00@0ADmI
04I@G016D5`0ATeG0P17CUP01016CeP0B51K04Q@F`18DU`204UCG`0404QDH019EF<0BEES04UCG@80
B59L00@0Ae5K04I@FP16D5X0AU1J0`17DE/204QBG0`0Be=N0`19DE`01018D5/0ATiI04I>F@15CEP2
04A<E`0404=;EP12BUD0@TYE04=;EP<0@TYE0`11BE@3049:E@80@d]F1015CEP504I>F@@0AdmJ0P16
CUT00`15CEP0A4aG04A<E`0304=;EP<0A4aG3015CEP<04M?FQ00ATmH0`17D5T504QAFP<0B4mE0`16
CeH00`15CeP0A4iG04A>E`03049?F005045?FP14CUL0A4iG04E>E@18CE400@3oool000`0BEMR0`18
EV4404UGHPD0BUQS0`1<FV<304]IHP<0BUQQ3`19Ef0<04UGHP`0B5IQ101;F64504YGH0<0BEIO00@0
AeER04MEH017EF00BEMR0P1<FF80101=FV<0C5IO04aFG`19De`204QAFP0404Y?DP0Z;c80:Rlb01PL
7@805QXI00@04aDD014C400A4a0051@B0P0C4a40100D4`h04a8=01<B3@0C4Pd3018A30<04a8=0P0D
4`h0100B4`h04a8=01<B3@0H4`d202`S6P0404Dg;P18=RX0B3HZ04La9080DS/]00@0HT0W06I4:`1V
A2/0K4L]0P1aBBl0101dB2d0MdP/07M8;01kBR`207e::`0408=@=@24DCH0Q54f08EB=`80R5Dj00@0
S5Tn091M@P2@GD80U6560P2FHdP0102JI4X0VfE;09]UB`2LIT`209mWCP040:=YD@2VK5@0YVaD0:]_
E`80[G5I00@0Z6aD0:U]E@2YKED0ZW1H0P2_MEd0102cNf80]7iT0;AnI02cOF<20;5kH@040:AaEP2L
JDh0W6U>09i[D080X6Y@00@0UV1609UQB02IHDP0WVI=0P2IGdL0102FH4P0TUa4099LA02CGdT209=Q
BP0408eMA`2@Hdh0T6=>09E[E@80TVQB00@0SFQE08IPB`26H4/0QEm:0P2?IU00102EK5H0UF]C09E[
D`2BJ50209AWD0040:5hHP2SNV@0XgYT0:ImI`80ZH1Z00@0[8=]0:Z1J`2ZPF/0YWeW0P2RNF<0102`
Pf`0Y7IO0:AfG`2KKEH209e]E`0409m_F@2OK5L0WfaG09i[EP80WfaG00@0[WaQ0:YgG02ZMe`0Yg5G
0P2YLEP0102]Med0[G]P0:ekH02YO5l20:=jG00409eeG02SNf80Xg]R0:1hG`80WgMN00@0YgmV0:Em
I02UOF@0WgMN0P2TO6<0102_PVD0ZGaO0:UlG`2SMUT20:1cEP040:=fF@2YO5l0ZGaO0:aoHP80[X5T
00@0[GiP0:ioH@2^Of40[h5P0P2aPf80102cQF<0]HMS0;F7H`2fR6@20;N9I@040;N9I`2aPf40/H=Q
0;N9I`80^XaZ00@0[h5O0:UkF@2YNeT0[7iL0P2[OE/0102RL4l0Y79A0:AbD@2YMeH20:MeE00409m]
C02RL4l0XW1?0:EcDP80WVa;00@0W6Y909e[BP2MJdX0WF]:0P2JJ4L0102EHd80TEln095O?P2@GSd2
095O?P0408MI=@2;GCT0Redi07E78`80ODl]00@0MDLV06Li601W>AP0OTla0P1iBR`0102DGcl0QE0`
08E@<025D30206Df6P0405Hc7@0l:1d0?2PM02lY:@80<3<h00@0<c8h03He>`0f=C/0>3Lm0P0g>3d0
0`0n?d@0AdY?04M:C`0304eBEP@0@e1I0`14DEX504EBF``0BU1L3019DUT404YBG@D0Be=N0`1<E5l0
101;Deh0BU9M04YBG@19DE`204M?FP0404E=F013BeH0@d]F049:E@H0@DUD0`12BUD504=;EP0404A:
EP15BeL0AD]G04E;E`80ATaH00<0AdeI04Q>FP18CUX00`19Ce/304I@FP0404E?F@14CUP0A4iH04==
E`80@TaF0`11BeD3049<EP<0@deG1014CUP204E?F@`0AU1J1P14CUL604==EP@0ADmH0`16D5T504MA
FP<0Be1F00@0BTiG04Q?E`18CeL0B4mI0P17D5T01016CUT0ADmI04E?F@11CeX204A>E`40ADiE0@18
CE410?ooo`003019Ef8304QFH@@0BEMR1@1:F6<304aJH`<0BeUR0`1:F64?04UGH0`0BEMR3018EV44
04]HH@D0BUMP0`19EUl01017EF80AeEP04MEH019Ef8204aIHP0404eJH`1<EUl0C5IO04UCG080B55J
00@0BTmB02X_<P0Z;c8061`M0P0F6QT0100C5A@04A<@014C400D518201<C4@0401@C3P0C4Pd04a8=
01<B3@<04Q4<0`0C4Pd201@C3P04018C3P0C4Pd04a8=01PC3@80;2<J00@0ACL^04Pf:P18=RX0Ac4T
0P1B>bd0101R@2L0IT@[06I4:`1/Abd20759;`0407A8;@1gB2`0MdP/07]:;080ODX[00@0Pe0e08AA
=P24DCH0QE8g0P28ECX0102<FCh0T5e2091M@P2DHDH209ISB00409YTBP2KID/0VfE;09aVC080WfM>
00@0XfUA0:I/E02VK5@0ZfmG0P2]LET0102XK5@0ZFeE0:U]E@2ZL5P20:meG@040;=kHP2dOV@0]7iT
0;=mH`80/G]Q00@0Y75F09aYCP2LJDh0WV]@0P2PJU00102FH4H0VF5809UQB02NITd209UOA`0409IP
B02BG4@0TUa409=OB@80Tf5:00@0SEe7091SCP2@Hdh0UF]E0P2BJ580102=J5D0QV1;08IPB`25GdX2
08mVD00409E/EP2EJe<0UF]C099XD080U6M@00@0XGQR0:=jI02SNV@0YWeW0P2YP6X0102/Pfd0ZX5[
0:Z1J`2VOFL20:9iH`040;23K02TMUl0Y7IO09]]EP80WFeG00@0WfmI09m/E`2OK5L0WV]F0P2OK5L0
102^O640ZWML0:YgG02WLEL20:UaF0040:egG@2]Nf00[G]P0:UlG`80XgYL00@0WGEL0:=kHP2SNf80
X7QO0P2OMeh0102WOfH0YGeT0:EmI02OMeh20:AlH`040:n2I@2YO5l0ZGaO0:=fF@80X7=F00@0XgII
0:UlG`2YO5l0[7mR0P2^PF@0102]OV00[WmQ0:ioH@2_PF020;63HP040;>5H`2eQf<0]HMS0;J8I080
]hUU00@0]hUW0;63H@2aPf40]hUW0P2jS6X0102_PEl0ZG]I0:UkF@2/OU`20:]mF`040:9`C`2TLU40
Y79A0:UgEP80YgED00@0Wfe<0:9`C`2RL4l0YG=B0P2NK4/0102LJTT0WF]:09e[BP2MJdX209YXA`04
09ES@P2AGch0TEln091N?@80TEln00@0QeTe08]M>@2;GCT0MDLS0P1mCbd0101eAbH0IcTH06Li601n
Cc4207U:;00409AO?`25D300QE0`08E@<080ICHJ00@0ES<M03`X7@0l:1d0;bTY0P0`<cP0100c<SP0
=SDk03He>`0h=cd203Lh?@0303hoA017BTl0AdY?00<0CE9F1013D5T304AAFPD0AE9K301:D5`<04UB
F@@0BU9M1@1;Deh304aDG`0404]CGP1:DUd0BU9M04UAG080AdmJ00@0ADeH04=;EP13BeH0@TYE1P11
BE@3049:E@D0@d]F00@0A4YF04E;E`15BeL0AD]G0P16C5P00`17CET0B4iJ04Q>FP0304U?F`<0AU1J
00@0ADmI04A>F014CUP0@deG0P12C5H3045;E@<0@TaF0`13CEL404A>F080ADmI3016D5X604A>E`H0
@deF1015CeP304I@F@D0Ae5J0`1;D5H0101:CUL0B4mG04Q?E`18CeT204M@F@0404I>F@15CeT0ADmI
045?FP80A4iG0@15CUD104Q=D@40oooo000<04UGHP<0B5IQ1019Ef8504YHH`<0C5YS0`1;FF8304YH
H@l0BEMP3019Ef8<04QFH@@0BeQQ1@1:Ef0304UFG`@0B5IQ0P1;F640101<FF80CEMP04eGH01:EE/2
04M@E`0403e3@`0Q9bL08BLW01<G60804aLF00@04Q@C014A3`0A4@l051@B0P0B4`h0100D4`h04a8=
01<B3@0C4Pd3018A30<04a8=0P0D4`h0100B4@d04a0;01<@2`0L5a4203@Z8@0404Hh;@16=2H0AS@V
04Pb9080Dc`/00@0Hd4X06I4:`1VA2/0KDP^0P1aBBl0101dB2d0MdP/07M8;01kBR`207i;;0<0Pe0e
00@0QE8g08QE>P28ECX0S5Tn0P2@GD80102DHDH0UV=809ISB02IHdT209]UB`0409eWC@2QJE00XFU@
0:I/E080Z6iF00@0[71H0:eaF@2]LET0[G5I0P2/L5P0102ZL5P0[G=K0:ecF`2`N5l20;9lHP040;=m
H`2bO680/WaR0:ekH080Xg5F00@0WfaA09aVC02LIT`0Tee30P2GGdH0102OIdh0WFE<09eUC02EGdL2
08mI@@0408iJA02=Fd@0SE]408IF@080QeY500@0RV1:08ILAP26G4H0Qf9?0P23GDP01021FdH0RF1:
08UPBP2>IDl2091VCP0409=YD@2IK5D0VFaE0:=jI080Y7]U00@0YGaV0:QoJ@2XOfT0ZH1Z0P2WOVP0
102RNF<0WWEO09ieG`2^P6T20:EgH00409i`F@2PL5X0X71J0:9bG080XFiI00@0XFeH0:=_FP2SKeX0
ZgUN0P2ZMe`0102[ME/0[gMN0:mgGP2dO6<20;9oI0040:YmH02TN5/0Y7QK09]cFP<0Y7aS0P2UOF@0
102ZPVT0YgmV0:MoIP2QNF020:AlH`040:n2I@2ZOF00ZWeP0:EhF`80XgII00@0YWUL0:YmH02ZOF00
[X5T0P2`PfH0102/OEl0[GiP0:enH02^P5l20:n1H0030;63H@2bQ600/XAP00<0/hEQ00@0/XAR0:Yl
FP2ZO5X0[7iL0P2_PEl00`2WNEL0Y7ID0:AfE0030:MiE`040:AbD@2ULe80YG=B0:YhE`80YgED00@0
Wfe<0:9`C`2RL4l0YWAC0P2OKD`0102KJDP0WF]:09e[BP2NK4/209aZB@0309MUA02CHD00Tf5000<0
TEln00@0QeTe08YL>02:G3P0N4XV0P1kCB/0101hBRT0KClN06do7P1jBbd207e><00409MR@P28Dc<0
R5<c08MB<P80JCXN00@0F3DO03dY7P0m:Ah0<B/[0P0b=CX0100a<3H0=SDk03He>`0i>Sl203/l@@04
03m2A`18Be00B4]@04aAE@80Be1D1013D5T504AAFP<0AE9K301:D5`<04UBF`@0BU9M1@1;Deh304aD
G`0404]CGP1:DUd0BU9M04UAG080AdmJ00@0ADeH04=;EP13BeH0@TYE1P11BE@3049:E@D0@d]F0`14
BUH01015BeL0ATaH04I<F017CET204Q>FP<0BDmK0`17CeX01016CUT0ADeH04E=F014C5L204=;EP<0
@TYE0`13BeH304A<E`H0ADeH3017CeX404I?F080ADiG00<0A4eF04=<E@13C5D00`13C5D404I?F0<0
Ae1I1@18DEX304aAE`0404]?F018CeL0B4mG04Q?F@80ATmH00@0ADeH04A>F014CUP0@DmJ0P14CUL1
04E>E@40B4eA0@3oool000`0BEMR0`18EV4404UGHPD0BUQS0`1<FV<304]IHP<0BUQQ3`19Ef0H04UG
HP@0BeQQ1@1:Ef0304UFG`0404UGHP18EV40B5IQ04QFG`80BEIO00@0C5UQ04iIG`1>FEl0BUEK0P15
CUD0100]<C805a/L01LK700=4A02018D4`04010B3`0A4@l04A4?01@E40804a8>00@051<>01<B3@0C
4Pd04a8=0`0B4@`301<B3@8051<>00@04A0;018?200B3`P09b0H0P10=R`01019>B`0A34S04@a8`1;
=2@205Lm;00406A2:@1WAB`0IdD/06e8;P80LDT_00@0M4P]07M8;01gB2`0NdX/0P1oC2d01022Cc@0
Pe0e08=@=@25DSL208QE>P0408aI?P2@GD80T5e209AQAP80UV=800@0VF=909]UB`2KID/0WfU?0P2U
KE@0102YKeL0Zg5I0:]aF@2^LUX20:mcF`030;=fH@2`Leh0/7=N00<0Zg5K00@0[WIO0;1jHP2`NV80
/WaT0P2cOFD0102cPFP0ZGMN0:UgGP2RL5L209aXD00409MQB@2HHTX0V69:09iVC`80W6A=00@0V69:
095K@`2AFd<0T5a60P2AGdP0102;FdD0Rei908]NB@2?IDl208aRC00408EPC@21FdH0PE]607iH@`80
PeY400@0Qei808eSB`2=Hd/0TfUA0P2KKUL0102RN680XGMQ0:5gH@2QMf420:AjI0040:MmI`2VO6H0
YWaV0:5gH@80WG=M00@0ZG]S0:IhH02VN600YGEN0P2VMUl0102WMEh0YW9L0:IbG02VLU`20:QdGP04
0:MeFP2YMU/0ZGIK0:mgGP80]WaT00@0^H5X0;F2I`2ePVL0[7eQ0P2TN5/0102LLUT0YWaS0:IlH`2[
PFP20:f3JP040:j4J`2YOfH0ZGmV0:EkHP80Z7iU00@0[X5T0:]nH@2[OV40Z7]N0P2WNUd0102ZOF00
[H1S0:f0H`2`PfH20;64I`030:imG`2_OV00[giP00<0[giN00<0/7mN0;20G02`P5`00`2aPEd0102f
QF@0ZgYI0:]jF@2[NUT20:mnG@<0ZWUH00<0[GaK0:akFP2/NeX00`2WME@0102ZN5L0YgED0:MeE02O
KD`20:=aD0040:MeE02QKdh0XFm>09]YB080WF]:00@0Wfe<09i/B`2NK4/0VfU80P2FI4<0102CHD00
TV0o099P?`27FCD208YL>00407e?:`1iBbT0ND/Y07e?;P80M4HU00@0M4DW081A<`20DC<0VfI60P2<
EcL01029E3@0L44U07119@1I=R0203hZ7`0403<];@0c=S/0<cHk02l^=080=SLl00@0?Ci303i1AP0n
@DH0@dI;0P19CU800`1;D5@0B4eA04Q=D@0304A>F0<0ADmI1016D5X204MAF``0BU1L3019DU/404YB
G@D0Be=N0`1<E5l0101;Deh0BU9M04YBG@19DE`204M?FP0404E=F013BeH0@d]F049:E@H0@DUD0`12
BUD504=;EP<0A4YF00@0AD]G04I<F016C5P0AdeI0`18CUX204U?F`<0AdmJ00@0ATiI04E=F015CEP0
A4aG0P13BeH3049:E@@0@d]F0`14C5L504E=F0`0ATiI00@0B55J04M@F@17D5T0ATmH0P15CUL01014
CEH0@daE04=<E@12Be@2045:D`@0ATmH0`17D5T504QAFP<0CE9H00@0C51I04U@F019D5P0AdiH0P15
CUL01014C5L0@deG04==E`11CeX204A>E`40ADiE0@18CE410?ooo`003019Ef8304QFH@@0BEMR1@1:
F6<304aJH`<0BeUR0`1:F64?04UGH1P0BEMR101;F64504YGH0<0BEIO00@0BEMR04QFH@18EV40B5IO
0P19EUl0101<FF40CUUO04iIG`1:EE/204E>E@0402da<P0G6a`05a/L00dA40804Q@C00@0418?014A
3`0A4@l051D@0P0C4Ph0100D4`h04a8=01<B3@0C4Pd3018A30<04a8=0P0D4`h0100A40/04Pl8018?
200W81P2040f;00404Ti;014<B<0A34S04/d9080Ecd/00@0I48Y06M5;01WAB`0KDP^0P1aBBl0101d
B2d0MdP/07M8;01kBR`207m<;@04089?=023D3D0Pe0e08EB=`80R5Dj00@0S5Tn091M@P2@GD80U656
0P2FHdP0102IHdT0VfE;09]UB`2OJDl20:E]E0040:U_E`2[LET0Zg5I0:ibFP80[g=K00<0/gIQ0;1c
GP2`Leh00`2[LE/0102^MUl0/7YR0;1jHP2bO6@20;=mI@040;>1J02YMeh0ZGMN0:9`E`80W6Q@00@0
Uf5909QRBP2HHTX0WVI?0P2LI4d0102HHTX0TE]3095K@`2@G4H2095OB00408]KA@2;GTT0Rei908mU
C`80S69<00@0QF1=085KAP21FdH0OUQ30P23FT@01027GTP0SF=;08eSB`2CJE4209]^E`040:9hHP2Q
Mf40XGMQ0:5gH@80Y7YT00@0YgeW0:IlIP2VO6H0XGMQ0P2MLed0102YNf<0YWQP0:IhH02UMEh20:If
G`040:MeGP2VLU`0YW9L0:IbG080Z7AN00@0YgEJ0:UfF`2YMU/0[gMN0P2fO6@0102iPFP0]H9W0;F2
I`2/OF420:AhF`0409abF@2VO6<0YWaS0:^1J080[H=Z00@0[XA[0:UoIP2YOfH0YG]R0P2XOVD0102^
PF@0ZgiQ0:]nH@2XNeh20:MjG@040:YmH02]P6<0[H1S0;23IP80/HAW00<0[WeO0:mnH02_OV000`2_
OUh00`2`Oeh0/81L0;20G0030;61G@040;J5I02[NUT0ZgYI0:]jF@80[giM0`2ZNEP00`2]O5/0[7]J
0:akFP030:MeE0040:YhE`2WME@0YgED09m]C080Xg5@00@0YgED0:5_CP2QKdh0VfU80P2MJdX0102O
KD`0WVa;09i/B`2KJDP209IT@`0409=Q@02BH3l0TV0o08MI=@80RU`h00@0ODl[07U;:@1iBbT0ODl^
0P1dARD0101dABL0P54c081A<`2KITH208aG=`0408UD=01`@BD0L44U05Tf8080?RXO00@0<bd]03<f
>`0c=S/0;bhd0P0f=c`0100m?T<0?T5603i1AP13AT/204U>DP0304]@E018CE40B4eA00<0A4iH0`15
CeT404I@FP80Ae5K301:D5`<04UBF`@0BU9M1@1;Deh304aDG`0404]CGP1:DUd0BU9M04UAG080AdmJ
00@0ADeH04=;EP13BeH0@TYE1P11BE@3049:E@D0@d]F0`14BUH01015BeL0ATaH04I<F017CET304Q>
FP80BDmK0`17CeX01016CUT0ADeH04E=F014C5L204=;EP<0@TYE1013BeH304A<E`D0ADeH3016CUT0
1018DEX0Ae1I04M@F@16CeP204E>E`0404A=EP13C5D0@daE049;E080@DYC1016CeP304M@F@D0B55J
0`1=DUP0101<D5T0BE1H04U@F017CUP204E>E`0404A<E`13CEL0@deG045?FP80A4iG0@15CUD104Q=
D@40oooo000<04UGHP<0B5IQ1019Ef8504YHH`<0C5YS0`1;FF8304YHH@l0BEMP6019Ef8404]HH@D0
BUMP0`19EUl0101:F640BEMP04UGH017EEh204QEG@0404aIH@1?FV00CeYP04UDF080@4U>00@082@S
00lC4P0?4a803@l>0P0B5140100A4@l04a8>01<B3P0E5102018A300401@C3P0C4Pd04a8=01<B3@<0
4Q4<0`0C4Pd201@C3P04018?200E4@P05A48034Z8080CD4e00@0BCTZ048_7P12;ah0CCHT0P1I@2`0
101U@bX0J4H]06Q6;@1]B2h20759;`0407A8;@1gB2`0MdP/07a;;@80P4d^0`23D3<01025DSD0R5Dh
08QE>02<FC`2091M@00409AQA02FHdH0UV=609USB@80WFM=00@0XVaB0:Q`E`2XL5L0[G=K0P2^M5`3
0;1dG0030;EhH`2bMF00/WEP00<0[G=M00<0[gMP0;5kH`2aNf<00`2aNf<0102_P6H0ZW]Q0:YkH@2X
MUd20:A`F00409m[D`2NJ500WVQ@09YTC080UV1800@0SELo08A>=P24CSH0Q50j0P29Ed001025ECl0
QUU408IIA02:H4X208UOB@0408IQCP21FdH0PE]607aF@@80P5M100@0R5m9095WC`2AIdl0UF]C0P2H
Je@0102KLET0VW1H09Y`F02JL5P209idG0040:AjHP2VO6@0YWaT0:=iH@80X7IN00@0YWIO0:MgH02W
Mf00ZWYS0P2/NV<0102ZN640ZGEO0:UeG`2]MV420:mhH`040:IcF02ZM5X0ZWAJ0:mgGP80]WaT00@0
^X1X0;N1I`2gPFL0[GiR0P2WNUd0102QME`0ZWiU0:YnI@2aQF`20;:6K@040;24J`2/P6L0[81W0:]o
IP80[X9Y0`2]OV80102/OF40[GiR0:enHP2^Of<20;21I@030;:3I`2cQ6P0/hAX00<0/7mQ0`2_OUh0
0`2_OUd0[WiJ0:inFP030:inFP040;F4H`2ZNEP0ZWUH0:]jF@80/H1O00@0[giM0:imG02^OE`0/7mN
0P2]O5/0102YMe@0Z7IC0:QfD`2[NEH20:MeDP0409m]BP2SLDh0Xg5>0:QfD`80XW1=00@0W6Y709i/
B@2NK4T0X6i;0P2QKd`0102NK4T0VFM409UWA02EHd02099P?@0408YL>02<GSX0S5hj089D<080ND/Y
00@0PE<b07U;:P1iBbX0KSlQ0P1oD380102OJTX0T5/k091K>`2:ECD207M8;00405Tf800l:1d0?2PM
038/;080<C@i00@0;blg03Ti@@0i>D40@4=:0P14Adh01016Be40BTmE04Y?E@17CU@204A;D@@0A4iH
0`15CeT504I@FP`0BU1L3019DE`404YBG@D0Be=N0`1<E5l0101;Deh0BU9M04YBG@19DE`204M?FP04
04E=F013BeH0@d]F049:E@H0@DUD0`12BUD504=;EP0404=9E@14BUH0A4YF04A:EP80AD]G00<0ATaH
04M=F@17CET01P18CUX01017CET0ATaH04I<F015BeL204A:EPH0@dUE0`14BUH304E;E`<0ATaH3017
CET0101:DE/0BE1J04U@FP18CeT204I=E`0404E<EP13BU@0@dYD0499D`80@DQB1017CUP304Q?F@D0
BE1J00@0CU=I04eBF01=DUP0C51I0P19D5P01017CUP0ADiG04E>E`14C5L2049<EP05045?FP14CUL0
A4iG04E>E@18CE400@3oool000`0BEMR0`18EV4404UGHPD0BUQS0`1<FV<304]IHP<0BUQQ3`19Ef0<
04UGHP`0BUQS101;F64504YGH0<0BEIO0`1:F6401017EUd0AeAJ04MDFP1<FEl2051KG`0404E@E00h
@DH0>45601/M7080418A00@0418?01DE4`0E5A<04a8>0P0C4Pd0100G50l04a0;01<@2`0D4`h301<B
3@<04Q4<0`0C4Pd201@C3P0401@B1P0M6@h07AT>03`d9`80DDDg00@0BCPX04@_7014;a`0DCPT0P1L
@B`0101VA2/0J4H]06Q6;@1]B2h20718;P0407A8;@1hBBd0N4T]07e<;P80PDh_0`24DC801026Dc@0
REHg08UF=`2=FS/2095N?`0409ER@`2GI4D0UfA509aVC080WfU?00@0Y6iD0:UaF02YLEP0[WAL0P2_
MEd30;5eG@040;AgI02bMF80/WER0:meH@80/7IR00@0/WUU0;9kIP2bNfH0/W]V0P2`NF@0102_OfP0
[GeV0:emIP2XN6420:IdG@040:AbF`2NJU@0WVYD09QQC080Tea700@0SeU1089<=022C3@0P4`f0P25
Dc`01020D3X0OU4l07iA?020ET0207iD?P04085LB@1jE3l0NU@o07E?>P80OUDo00@0SFA>09Y`F02J
L5P0Vg5I0P2IK5D0102LL5L0VFeD09U]E02HK5<209a`E`040:9fG@2UNF00YGUP0:=gGP80Wg=J00@0
XgAJ0:QiG`2XNEl0[WaS0P2^NV80102/N600[WQP0:ihH02`NV820;9lI0040:MdF@2[LeX0Zg=J0:id
G080]7QP00@0]G]S0;=mH`2cOF<0[WaQ0P2YO5l0102YNf40[GmU0:eoI@2cQF/20;B6K0040:j0IP2/
OV@0[7iT0:n1I`80/H=Y00@0ZgaP0:enHP2]OV80[h1T0P2`PFD60;62IP040;9oHP2aOV40/GiQ0;5n
G`80/7eN00<0[geL0:ilF@2^O5T00`2]NeP0102]NeX0Y79A0:AbD@2XMUD20;5oGP040:mmG02]NeX0
[G]J0:ajF@80YgED00@0ZGMD0:QfD`2XMU<0[7YG0P2XMU<0102OKDX0Xg5>0:=aCP2WME820:1^B`03
09i/B@2OKDX0Wfe:00<0XFm<00@0Wfe:09]YAP2KJDH0UF=00P2BH3d0102=Gc/0Sf4m08mQ?@25Ec<2
07a>;00408=E=01jC2/0NT`[06`m7`80Nd`^00@0X6];095L?02AG3`0R5<c0P1mCS80101G=1h0>BDJ
03TU6P0a:b/202lb=`04034d>`0l?dH0?3m604E8C`80AT]A00@0AdaB04M>E017CU@0ATeC0P11BTl3
04A<E`<0ADeH0`16CUT304M?FP`0BDmK3018D5/404YBG@D0Be=N0`1<E5l0101;Deh0BU9M04YBG@19
DE`204M?FP0404E=F013BeH0@d]F049:E@H0@DUD0`12BUD504=;EP<0@dUE00@0A4YF04E;E`15BeL0
ATaH0`17CET504Q>FP0404M=F@16C5P0ATaH04E;E`80A4YF1`13BED304A:EPD0AD]G3017CET0101:
DE/0BE1J04U@FP18CeT204I=E`0404E<EP13BU@0@dYD0499D`80@DQB1017CUP304Q?F@D0BE1J00@0
CU=I04eBF01=DUP0C51I0P19D5P01017CUP0ADiG04E>E`14C5L2049<EP05045?FP14CUL0A4iG04E>
E@18CE400@3oool000`0BEMR0`18EV4404UGHPD0BUQS0`1<FV<304]IHP<0BUQQ3`19Ef0<04UGHP`0
BUQS101;F64504YGH0<0BEIO0`1:F6401017EUd0AeAJ04MDFP1<FEl2051KG`0404E@E00h@DH0>456
01/M7080418A00@0418?01DE4`0E5A<04a8>0P0C4Pd0100G50l04a0;01<@2`0D4`h301<B3@<04Q4<
0`0C4Pd201@C3P0401@B1P0M6@h07AT>03`d9`80DDDg00@0BCPX04@_7014;a`0DCPT0P1L@B`0101V
A2/0J4H]06Q6;@1]B2h20718;P0407A8;@1hBBd0N4T]07e<;P80PDh_0`24DC801026Dc@0REHg08UF
=`2=FS/2095N?`0409ER@`2GI4D0UfA509aVC080WfU?00@0Y6iD0:UaF02YLEP0[WAL0P2_MEd30;5e
G@040;AgI02bMF80/WER0:meH@80/7IR00@0/WUU0;9kIP2bNfH0/W]V0P2`NF@0102_OfP0[GeV0:em
IP2XN6420:IdG@040:AbF`2NJU@0WVYD09QQC080Tea700@0SeU1089<=022C3@0P4`f0P25Dc`01020
D3X0OU4l07iA?020ET0207iD?P04085LB@1jE3l0NU@o07E?>P80OUDo00@0SFA>09Y`F02JL5P0Vg5I
0P2IK5D0102LL5L0VFeD09U]E02HK5<209a`E`040:9fG@2UNF00YGUP0:=gGP80Wg=J00@0XgAJ0:Qi
G`2XNEl0[WaS0P2^NV80102/N600[WQP0:ihH02`NV820;9lI0040:MdF@2[LeX0Zg=J0:idG080]7QP
00@0]G]S0;=mH`2cOF<0[WaQ0P2YO5l0102YNf40[GmU0:eoI@2cQF/20;B6K0040:j0IP2/OV@0[7iT
0:n1I`80/H=Y00@0ZgaP0:enHP2]OV80[h1T0P2`PFD60;62IP040;9oHP2aOV40/GiQ0;5nG`80/7eN
00<0[geL0:ilF@2^O5T00`2]NeP0102]NeX0Y79A0:AbD@2XMUD20;5oGP040:mmG02]NeX0[G]J0:aj
F@80YgED00@0ZGMD0:QfD`2XMU<0[7YG0P2XMU<0102OKDX0Xg5>0:=aCP2WME820:1^B`0309i/B@2O
KDX0Wfe:00<0XFm<00@0Wfe:09]YAP2KJDH0UF=00P2BH3d0102=Gc/0Sf4m08mQ?@25Ec<207a>;004
08=E=01jC2/0NT`[06`m7`80Nd`^00@0X6];095L?02AG3`0R5<c0P1mCS80101G=1h0>BDJ03TU6P0a
:b/202lb=`04034d>`0l?dH0?3m604E8C`80AT]A00@0AdaB04M>E017CU@0ATeC0P11BTl304A<E`<0
ADeH0`16CUT304M?FP`0BDmK3018D5/404YBG@D0Be=N0`1<E5l0101;Deh0BU9M04YBG@19DE`204M?
FP0404E=F013BeH0@d]F049:E@H0@DUD0`12BUD504=;EP<0@dUE00@0A4YF04E;E`15BeL0ATaH0`17
CET504Q>FP0404M=F@16C5P0ATaH04E;E`80A4YF1`13BED304A:EPD0AD]G3017CET0101:DE/0BE1J
04U@FP18CeT204I=E`0404E<EP13BU@0@dYD0499D`80@DQB1017CUP304Q?F@D0BE1J00@0CU=I04eB
F01=DUP0C51I0P19D5P01017CUP0ADiG04E>E`14C5L2049<EP05045?FP14CUL0A4iG04E>E@18CE40
0@3oool000`0BEMR0`18EV4404UGHPD0BUQS0`1<FV<304]IHP<0BUQQ3`19Ef0<04UGHP`0BUQS101;
F64504YGH0<0BEIO00@0BEQO04YIH01:FF00B5MN0P17E5X0101<FUd0CeYN04mJGP10Bdd202hh>P04
01LI5P0@40h0410>01<C4@805aHB00@04Q4<01DB3@0E4Pd061D>0P0C40T0100D4`h04a8=01<B3@0C
4Pd3018A30<04a8=0P0D4`h0100H50P0:bDI02/U6@15?2d204m2<P0404Pg9@17<A`0Ac4L05<j9080
GD</00@0IdD/06U7;P1YAbh0KDP^0P1`B2h0101dB2d0N4T]07Q9;@1nCBl2089?<00408EB<`26Dc@0
QU<d08QE=P80RePi00@0SU/l09=P@@2CH440UV=40P2IITL0102OJDl0XF]A0:5[D@2UKeD20:UaF003
0:ecF`2_MEd0[gEM00<0/GEM00@0/WER0;5dH@2aM640[gEQ0P2aMf<0102dNfL0]GiY0;EnJ@2dOFP2
0;9kIP040;>5K@2aPf/0/H=[0:QhH@80XG5J00@0Xg5J09eYD`2MJE<0TUi80P2@G4H0102FH4P0QU0h
08I@>020C3H208=A>P0407];=@1dAc80M4Lb07=9<`80LDLa00@0M4ll071:=@1`BSD0KDLb0P1gCSP0
102:HD/0W79J09abFP2PMUh209e`F@040:5eG02LL5L0W71G09M[DP80VFeD00@0WW9I0:1dF`2PM5/0
WG5H0P2IKE@0102TLUT0ZWQO0:YhG`2]Nf820:ahH0030:egG`2`NV80/7YR00<0/g]T00@0ZgEK0:ad
F`2/M5/0/7AL0P2bMUh0102dN600/WYQ0;9jH@2^O6420:]lH0<0/GmV0`2fQ6/0102_OF@0[WaS0:il
H`2bP6L20;5oIP040:]iGP2^O640[WaQ0;>1IP80]89W00<0/h5V0;5oI02aOf@00`2_OF80102`OF00
[gaO0:mlG`2_O5d20:ikG0030:ekFP2/NUL0[7YG00<0ZgUF00@0ZgUH0:AbD@2TLU40ZgUH0P2dPV40
102aOeh0[WaK0:ilF`2]NeX20:QfE@<0YgE@00@0[7YE0:UgDP2YMe80X6i90P2RL4/0102ULdh0WF]6
09e[AP2RL4/30:5_BP80X6i900@0Wfe809]YA02KJD@0UF<n0P2AGcX0102=Gc/0Sf4m08mQ?@23EC42
085C<@0408EG=P1hBRT0N4XY07118`80MDHX00@0WVU9091K>`2@Fc/0QU4a0P22DcL0101G=1h0>RHK
03XV6`0e;bl203<f>`0303PkA012ADh0@TE>00<0B4aE00@0A4]C04A=E014CE@0A4eD0P11BU4404A<
E`<0ADeH1@16CUT<04U?F``0B51M101:DUd504]CGP<0C5AO00@0Be=N04YBG@1:DUd0BE5L0P17CeX0
1015CEP0@d]F04=;EP12BUD60459E0<0@TYE1@13BeH01012B5@0@dUE04=9E@13BED204A:EP0304E;
E`16C5P0ATaH00<0AdeI0`1:CUX01019CET0B4aH04Q<F017BeL204I:EP<0ADUE00<0A4QD04E9E@15
BED00`15BED304I:EP<0Ad]G3018C5P0101;CeX0BTiI04Y>F@19CEP204Q<E`0404M;EP16BUD0ATYE
04E9E080A4QC1019CEP304Y>F@D0BdmJ0`1=DUP0101<D5T0BE1H04U@F017CUP204E>E`0404A<E`13
CEL0@deG045?FP80A4iG0@15CUD104Q=D@40oooo000<04UGHP<0B5IQ1019Ef8504YHH`<0C5YS0`1;
FF8304YHH@l0BEMP3019Ef8<04YHH`@0BeQQ1@1:Ef0304UFG`0404QGGP1:FF00BUUP04QGG080AeEH
00@0CE]N04mJG01?FU`0>dI80P0T;S00100C4a403Ph<00h>300B4`h201DD4004018?2P0E4P/05A8;
01TD3P805A0:00@051<>01<B3@0C4Pd04a8=0`0B4@`301<B3@8051<>00@06aP903Pb900h<R@0Bd8a
0P19?2/01016=200BS@M04Xd7@1E>b@205m3:`0406Q6;@1ZB2l0JTP_06i9;`80L4P^00@0M4P]07Q9
;@1hBBd0OTd_0P23D3401027EC@0R5He08QF=@29EcH208aJ>@04091N?@2DHT40U69109QVA@80VVQ7
00@0XVaB0:=]D`2SKE<0YFmE0P2XL5L0102[LET0[G=K0:ecF`2`M5`20;5eG@040;=fI02aM680/GAR
0:ebH080[gAR00<0/gYW0;IoJ`2fOf/00`2hPFd30;F7L0040:YlI@2TM5h0Y7AN0:=cG@80VVMB00@0
RUI108IB?@26DSd0PT`d0P1`>R80101/>280LT0Y0790:@1/?2H206Hi900406Dk9@1T>R@0I3XT06Q3
<080JD<^00@0J48]06i5;`1^ABl0Ne8l0P2?IDd0102IKeL0W6mH09a_F02TMU`209e_E@0409MYC`2H
JU00V6Y@09e_E@80X79H00@0W6iD09QZD02HJU00YG=H0P2ZN5d0102/NEh0ZgEK0:]eF`2/MU`20;5i
H0040;AlH`2bNV40/WYQ0:egG@80/7IN00@0/gMO0;IhH@2fN640]GUQ0P2bNV40102_O640[7eQ0:am
H@2cPFH20;1nH`040;F3J02hQV/0^8I[0;1nH`80[geR00@0/7iS0:UgG02YMe`0ZGML0P2^O640102c
PFH0]H=X0;F3J02cPFH20:mmHP030:ajG`2ZN5d0ZWQM00<0[GMK0`2/MeP00`2[MUH0ZWEC0:YeD`03
0:UdDP040:iiF@2YM5@0ZGAD0;1kF`80]h9R00@0/GaL0:iiF@2^NET0/7]K0P2^NET0102TLTd0YWA?
0:IdC`2/NUD20:YhD`040:5_BP2RL4/0XW1;0:=aC080VVQ300@0YG=>0:=aC02SLD`0X6i90P2NK4L0
102MJdH0VFM209UW@P2DHSd2091N>@0408YL>02=Gc/0SElk07i@;080QUPf00@0QUPg07I89`1fB2L0
NDX/0P1b@bD0102JIDD0SEPh08eH>024Cbl208QI?@0405Xg8@0o:b00?b/P03dg=`80>ci300@0?T9;
04I:D`16BU<0BdmH0P16CED01012BE40@T]B049;DP13C5<2045<DPH0AD]G0`16C5P304M=F@`0B4iJ
3017Ce`404YBG@D0Be=N0`1<E5l0101;Deh0BU9M04YBG@19DE`204M?FP0404E=F013BeH0@d]F049:
E@H0@DUD0`12BUD504=;EP<0@TQD00@0@dUE04A:EP14BUH0AD]G0P16C5P304M=F@<0BTiJ00@0BDeI
04Q<F018C5P0Ad]G0P16BUH304E9E@<0A4QD1015BED504I:EP`0B4aH1019CEP204Q<E`0304M;EP16
BUD0ATYE00<0ATYE1019CEP304Y>F@D0BdmJ0`1<DEL0101;CeP0B4mG04Q?E`18CeT204I?F00404E=
F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000<04UGHP<0B5IQ1019Ef8504YHH`<0C5YS
0`1;FF8304YHH@l0BEMP3019Ef8<04YHH`@0BeQQ1@1:Ef0304UFG`0404QGGP1:FF00BUUP04QGG080
AeEH00@0CE]N04mJG01?FU`0>dI80P0T;S00100C4a403Ph<00h>300B4`h201DD4004018?2P0E4P/0
5A8;01TD3P805A0:00@051<>01<B3@0C4Pd04a8=0`0B4@`301<B3@8051<>00@06aP903Pb900h<R@0
Bd8a0P19?2/01016=200BS@M04Xd7@1E>b@205m3:`0406Q6;@1ZB2l0JTP_06i9;`80L4P^00@0M4P]
07Q9;@1hBBd0OTd_0P23D3401027EC@0R5He08QF=@29EcH208aJ>@04091N?@2DHT40U69109QVA@80
VVQ700@0XVaB0:=]D`2SKE<0YFmE0P2XL5L0102[LET0[G=K0:ecF`2`M5`20;5eG@040;=fI02aM680
/GAR0:ebH080[gAR00<0/gYW0;IoJ`2fOf/00`2hPFd30;F7L0040:YlI@2TM5h0Y7AN0:=cG@80VVMB
00@0RUI108IB?@26DSd0PT`d0P1`>R80101/>280LT0Y0790:@1/?2H206Hi900406Dk9@1T>R@0I3XT
06Q3<080JD<^00@0J48]06i5;`1^ABl0Ne8l0P2?IDd0102IKeL0W6mH09a_F02TMU`209e_E@0409MY
C`2HJU00V6Y@09e_E@80X79H00@0W6iD09QZD02HJU00YG=H0P2ZN5d0102/NEh0ZgEK0:]eF`2/MU`2
0;5iH0040;AlH`2bNV40/WYQ0:egG@80/7IN00@0/gMO0;IhH@2fN640]GUQ0P2bNV40102_O640[7eQ
0:amH@2cPFH20;1nH`040;F3J02hQV/0^8I[0;1nH`80[geR00@0/7iS0:UgG02YMe`0ZGML0P2^O640
102cPFH0]H=X0;F3J02cPFH20:mmHP030:ajG`2ZN5d0ZWQM00<0[GMK0`2/MeP00`2[MUH0ZWEC0:Ye
D`030:UdDP040:iiF@2YM5@0ZGAD0;1kF`80]h9R00@0/GaL0:iiF@2^NET0/7]K0P2^NET0102TLTd0
YWA?0:IdC`2/NUD20:YhD`040:5_BP2RL4/0XW1;0:=aC080VVQ300@0YG=>0:=aC02SLD`0X6i90P2N
K4L0102MJdH0VFM209UW@P2DHSd2091N>@0408YL>02=Gc/0SElk07i@;080QUPf00@0QUPg07I89`1f
B2L0NDX/0P1b@bD0102JIDD0SEPh08eH>024Cbl208QI?@0405Xg8@0o:b00?b/P03dg=`80>ci300@0
?T9;04I:D`16BU<0BdmH0P16CED01012BE40@T]B049;DP13C5<2045<DPH0AD]G0`16C5P304M=F@`0
B4iJ3017Ce`404YBG@D0Be=N0`1<E5l0101;Deh0BU9M04YBG@19DE`204M?FP0404E=F013BeH0@d]F
049:E@H0@DUD0`12BUD504=;EP<0@TQD00@0@dUE04A:EP14BUH0AD]G0P16C5P304M=F@<0BTiJ00@0
BDeI04Q<F018C5P0Ad]G0P16BUH304E9E@<0A4QD1015BED504I:EP`0B4aH1019CEP204Q<E`0304M;
EP16BUD0ATYE00<0ATYE1019CEP304Y>F@D0BdmJ0`1<DEL0101;CeP0B4mG04Q?E`18CeT204I?F004
04E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000<04UGHP<0B5IQ1019Ef8504YHH`<0
C5YS0`1;FF8304YHH@l0BEMP3019Ef8<04YHH`@0BeQQ1@1:Ef0304UFG`0404MFF`1:FEh0BUUN04QG
G080B5II00@0CE]N04iIF`1>FE/0>4=50P0N:2T0100?3`d030d800`=200B4@d201<B3@04014>2@0D
4@X0514:01XE3`805Q4;00@051<>01<B3@0C4Pd04a8=0`0B4@`301<B3@8051<>00@07aT;048k:`12
>b/0CTDd0P17>2D01015<ad0C3HO04`f7`1E>b820614;00406Q6;@1ZB2l0JTP_06i9;`80L4P^00@0
M4P]07Q9;@1hBBd0OTd_0P23D3401028ESD0RELf08UG=P2:F3L208eK>P04095O?P2EHd80UF=209UW
AP80VfU800@0Y6iD0:E_E@2UKeD0YFmE0P2WKeH0102ZL5P0[79J0:abFP2_Le/20;5eG@040;AgI@2a
M680/GAR0:]`GP80[75O00@0/7MT0;IoJ`2fOf/0^hA`0P2mQW80102aQ6d0]XQa0;J8L@2_PFX30:Yj
I080VVMB00@0PDhi07Q5<01hAC00LS`T0P1R;1@0101Q;AL0K3XS06`j8`1]?BL206/n:@0406e3;@1_
ABl0KdD_06@o;080JD<^00@0J48]06Ln:01W?RP0KD@^0P1nE3`0102>I4`0UfYC09MZD`2QLeT209Y/
DP0409AVC02FJ4h0UVQ>09e_E@80XG=I00@0Wg5G09]]D`2KKE<0YgAI0P2[N5d0102ZMe`0Z79H0:Qb
F02ZM5X20;9jH@040;AlH`2aMel0/GMO0:mgGP80/WQP00@0]g]S0;YlI@2jO6D0^7aT0P2eNf<0102`
OF80[WaQ0:ilH@2cP6D20:mlH@040;J3J02jQf`0^XM/0;>0I@80/GiS00@0[gaQ0:AaEP2TLEH0ZGIK
0P2^Nf00102ePVL0]hAY0;N4J@2cP6D20:ikH0040:YgG02XMEX0Z7EJ0:YdF080ZG=G00@0ZGAE0:Qc
E02XLe@0Yg9B0P2WLU030:IaC`040:UdE02TKdl0Y6m?0:YeE@80[gYJ00@0Yg9B0:A_C`2TKdl0ZGAD
0P2[MUH0102RLDT0YGA<0:EdC02/Ne<20:]jDP<0XW5900@0XG1809MV?P2GISh0Z7M?0P2TLd/0102P
KdL0WFa409e/A02KJT8209QW?`0409=R>P2?GSH0Sehf08IH=080Redi00@0ND/W08UK>@29FcT0QUPg
0P1dARD01020DC<0LD8T0752902GHT8208]F=P0408=>;P2;G400Rea005dj9080A30U00@0AClo04=6
B`13AT/0@dMB0P1:CUT0101=DE`0ADaF04E<EP0oB5420419DP<0@TaE00<0A4YF04E;E`15BeL00`15
BeL304I<F0<0AdeI3018CUX<04M?G0@0BU9M1@1;Deh304aDG`0404]CGP1:DUd0BU9M04UAG080AdmJ
00@0ADeH04=;EP13BeH0@TYE1P11BE@3049:E@D0@d]F0`12B5@01013BED0A4YF04A:EP15BeL304I<
F080AdeI0`1:CUX01019CET0B4aH04Q<F017BeL204I:EP<0ADUE1014B5@304E9E@D0ATYF3018C5P6
04Q<E`H0Ad]F1019CEP304Y>F@D0BdmJ0`1;D5H0101:CUL0B4mG04Q?E`18CeT204M@F@0404I>F@15
CeT0ADmI045?FP80A4iG0@15CUD104Q=D@40oooo000<04UGHP@0B5IQ0P19Ef8304YHH`<0BeUT101:
F64D04UGH0<0B5IQ1019Ef8504YHH``0BEMR00@0CUIQ04eGH@1=Ef40CEMQ0`1:Ef0204UGH0<0B5IO
00@0AEMS04MIH`17FF<0B5UQ0P1;F600101AG680CEIK04eFF`0c>ch201@L7P@04a<C0P0C4a4301<D
3`<04a@>00@04A0>01<B400C4Q0051<A0P0C4Q03014@3P0401@C4@0H5aD061LE01/M6P80B3``00@0
F40X05<i801C>B00AS<R0P13<b@0101KAC00K4da06a=<@1/B2h206e8;P040718;P1cBC00LdT`07M;
<080NTd`00@0ODhb07i?<`1nCc<0Pe0c0P25DSD01028ECP0SEXm08eJ?@2BGd8209ISAP0409YWBP2K
J4/0VfQ;0:=^C`80YW1D00@0ZG=G0:]cFP2[LeX0[79J0P2/LU`00`2_LUd0/GAQ0;5dH@030;5dG`04
0:idGP2]Mel0[GMO0;1lI080]XA[00@0]hQ^0;J7K@2fQfd0/XAZ0P2fR6h0102gRFl0[GiT0:enI02H
ITd207]9<00406di8@1^>R80KSXR06H`5P80KcTO00@0NDH[08=?=`23CcL0QU@m0P27ECh308AD?P04
07YF?P1lF400O5Q007UC?080Ld/b00@0LDT_07U?=P1iCcH0PUTm0P26GD40102AGd@0V6E:09QUBP2C
H4D209EOA@0409iXCP2OIdh0WfM>09]SBP80XFU@00@0Y6YF0:]^F`2[KU/0[G1K0P2/KUL0102]KeP0
]7AK0;AdF`2eME`20;5_E`040;IhH@2dNV80]7YR0;AnI080]H9W00@0]85V0;AnI02dOV@0/gUQ0P2e
Mf00102aNf40^89X0;R2J02iPfT20;1jH0040:QbF02ZM5X0ZWAJ0:ihGP80[gUO00@0ZGIK0:aiGP2/
NEh0/GiS0`2ePVL20;5nH`040:]hG@2XMEX0Z7EJ0:elGP80[7UL00@0ZgQI0:ehF@2]N5T0[WUI0P2^
N5H0102[ME<0Z79@0:QbD02]MUL20:UbD`<0Y6e>00<0Yg1A0:YcE02ZLe@00`2[M5D0102MJD@0[WYE
0:ijE@2[Me820:5]B0040:EaC02XM4l0Z7A?0:A`B`80Xfm:00@0YW5?0:=^C02SKT`0X6]90P2NJTD0
102MJD80VVHo09YV?`2EHCT2095M=@0408aM<`25ER`0QEH/07m?9`80TVDn00@0RE/i07a?;P1lCbh0
R5/l0P1`ABD0102:F3<0T5hk091N>`26ECD207Y@>00405m2<P15>300ACP`04=7AP80@4iA00@0AE1F
04A?E@14CeD0@diD0`0oC58303i=DP80?DmC00@0@deG04A<E`14C5L0@d]F0P14BUH01015BeL0B4aH
04Q<F019CET204Y>FP@0AdeI0`18CUX504U?F`0304M?FP16CUT0ATiI00@0ATiI0P17CeX304UAG0@0
BU9M1@1;Deh304aDG`<0Be=N00@0BU9M04UAG019DE`0AdmJ0P15CEP00`12BUD0@DUD0459E00<049:
E@0404=9E@12B5@0@TQD04=9E@80A4YF00@0AD]G04M=F@17CET0BDmK0P1;DEd304Q>FP0404M=F@16
C5P0ATaH04E;E`80A4YF0`13BED70498E080A4YF00<0AD]G04I<F016C5P00`14BUH404=9E@80A4YF
00@0ATaH04M=F@17CET0AdiH0P16CEL01015C5H0A4]E04A;E@14BeD204E<EP0304I=E`17CUP0AdiH
00`0B4mI0`19CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80A4iG0@15
CUD104Q=D@40oooo000<04UGHP@0B5IQ0P19Ef8304YHH`<0BeUT101:F64D04UGH0<0B5IQ1019Ef85
04YHH``0BEMR00@0CUIQ04eGH@1=Ef40CEMQ0`1:Ef0204UGH0<0B5IO00@0AEMS04MIH`17FF<0B5UQ
0P1;F600101AG680CEIK04eFF`0c>ch201@L7P@04a<C0P0C4a4301<D3`<04a@>00@04A0>01<B400C
4Q0051<A0P0C4Q03014@3P0401@C4@0H5aD061LE01/M6P80B3``00@0F40X05<i801C>B00AS<R0P13
<b@0101KAC00K4da06a=<@1/B2h206e8;P040718;P1cBC00LdT`07M;<080NTd`00@0ODhb07i?<`1n
Cc<0Pe0c0P25DSD01028ECP0SEXm08eJ?@2BGd8209ISAP0409YWBP2KJ4/0VfQ;0:=^C`80YW1D00@0
ZG=G0:]cFP2[LeX0[79J0P2/LU`00`2_LUd0/GAQ0;5dH@030;5dG`040:idGP2]Mel0[GMO0;1lI080
]XA[00@0]hQ^0;J7K@2fQfd0/XAZ0P2fR6h0102gRFl0[GiT0:enI02HITd207]9<00406di8@1^>R80
KSXR06H`5P80KcTO00@0NDH[08=?=`23CcL0QU@m0P27ECh308AD?P0407YF?P1lF400O5Q007UC?080
Ld/b00@0LDT_07U?=P1iCcH0PUTm0P26GD40102AGd@0V6E:09QUBP2CH4D209EOA@0409iXCP2OIdh0
WfM>09]SBP80XFU@00@0Y6YF0:]^F`2[KU/0[G1K0P2/KUL0102]KeP0]7AK0;AdF`2eME`20;5_E`04
0;IhH@2dNV80]7YR0;AnI080]H9W00@0]85V0;AnI02dOV@0/gUQ0P2eMf00102aNf40^89X0;R2J02i
PfT20;1jH0040:QbF02ZM5X0ZWAJ0:ihGP80[gUO00@0ZGIK0:aiGP2/NEh0/GiS0`2ePVL20;5nH`04
0:]hG@2XMEX0Z7EJ0:elGP80[7UL00@0ZgQI0:ehF@2]N5T0[WUI0P2^N5H0102[ME<0Z79@0:QbD02]
MUL20:UbD`<0Y6e>00<0Yg1A0:YcE02ZLe@00`2[M5D0102MJD@0[WYE0:ijE@2[Me820:5]B0040:Ea
C02XM4l0Z7A?0:A`B`80Xfm:00@0YW5?0:=^C02SKT`0X6]90P2NJTD0102MJD80VVHo09YV?`2EHCT2
095M=@0408aM<`25ER`0QEH/07m?9`80TVDn00@0RE/i07a?;P1lCbh0R5/l0P1`ABD0102:F3<0T5hk
091N>`26ECD207Y@>00405m2<P15>300ACP`04=7AP80@4iA00@0AE1F04A?E@14CeD0@diD0`0oC583
03i=DP80?DmC00@0@deG04A<E`14C5L0@d]F0P14BUH01015BeL0B4aH04Q<F019CET204Y>FP@0AdeI
0`18CUX504U?F`0304M?FP16CUT0ATiI00@0ATiI0P17CeX304UAG0@0BU9M1@1;Deh304aDG`<0Be=N
00@0BU9M04UAG019DE`0AdmJ0P15CEP00`12BUD0@DUD0459E00<049:E@0404=9E@12B5@0@TQD04=9
E@80A4YF00@0AD]G04M=F@17CET0BDmK0P1;DEd304Q>FP0404M=F@16C5P0ATaH04E;E`80A4YF0`13
BED70498E080A4YF00<0AD]G04I<F016C5P00`14BUH404=9E@80A4YF00@0ATaH04M=F@17CET0AdiH
0P16CEL01015C5H0A4]E04A;E@14BeD204E<EP0304I=E`17CUP0AdiH00`0B4mI0`19CU@01019CEH0
AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80A4iG0@15CUD104Q=D@40oooo000<04UG
HP@0B5IQ0P19Ef8404YHH`80BeUT6019Ef0404QFH@<0BEMR1@1:F6<<04UGHP0404iFH@1=Ef40CEMQ
04eGH@<0BUMP0P19Ef0304QFG`0404AFHP17FF<0AeUS04MHH080BUMO00@0D5]Q04aEFP1<EEX0<SXm
0P0D71h401<C4`804a<A0`0C50l301<D3P04018A3`0C4Q004a8@01@C4@804a8@0`0A40h0100D4a40
5QDC01HE4`0P81h204Xn<00405To:01C>B00DcTP04Pf8P80ACDU00@0FdD`06]<<01[C300K4P^0P1]
B2h0101aBBl0LdT`07=9<01gBc0207Y=<00407e><P1nCc<0OTlc08=@<`80QE8e00@0R5Dh08eJ?@2=
FSd0TUm20P2FHdH0102JIdX0VfQ;09]XB`2OJT/20:9/D0040:I`E02XL5L0Z71G0:U_E`80ZFmI00@0
[G1K0:mbG`2_LUl0/gIQ0P2`Leh0102^M5h0/7YR0;1jHP2bOVH20;>1J0040;:3J@2cQ6X0/hAZ0;F7
K@80/H=Y00@0YgUO09IWC@2FIdd0OTle0P1[>B00101U<aX0JCDM06Te7@1kAB/2089<<P0408YG?02B
GTH0TUi609ARB`80UF=<0`2DI4h0102AJe@0T6YC091ZD`29Hd`207iF?@0407M?=@1hCSD0N4he07Q?
<`80N4`a00@0SUa109ISB02FHdP0U6560P2GHDL0102OJDl0WVI=09iVC@2KHdX209mWCP0409mUC`2U
J5<0YFQC0:EXD`80Y6I?00@0YVQA0:i^E@2^KUD0/71G0P2`KU@0102dMUl0/WQP0;9hH02cOF<30;>0
I@80]GmU00@0]WaT0;QjH`2hNV<0]7iT0P2hPVP0102gPFL0[gUO0:miG`2ZM5X20:egG@<0/G]Q00@0
[7UN0:ejG`2]NUl0/7eR0`2aOV<20:mlH@040:ejG`2/NEh0[7UN0:mnH080[W]N00@0[7UJ0:iiFP2^
NEX0[GQH0P2]MeD0102YLe40YW1>0:I`CP2YLU<20:M`D@040:E^C`2VKe00YVm@0:UbD`<0ZW=D0P2X
LE80102PK4D0[WYC0:ijD`2[Me020:9^A`040:EaBP2YMDh0ZGE>0:A`B@80XVi700@0YW5?0:A_C@2T
Kdd0XFe80P2NJTD0102LJ440VFDn09UU?P2DH3P2091L=00408iM=028Ebh0R5L^085A:@80TVDn00@0
RU`h07e@;`1mD2l0RE`k0P1cARL01025Dbh0SU`i08iL>@27ESH207U?=`040654=01:?CD0BSde04Q:
B@80@4iA00@0ADiG04A=EP14CEH0@TaE0`0oC5@203i=E00403m>E@10CeH0@4mF04==E`80A4aG00@0
@d]F04A:EP14BUH0A4YF0P17BeL304U=F@@0AdeI1@18CUX304U?F`0304M?FP16CUT0ATiI00@0ATiI
0P17CeX304UAG0@0BU9M1@1;Deh304aDG`<0Be=N00@0BU9M04UAG019DE`0AdmJ0P15CEP00`12BUD0
@DUD0459E00<049:E@@0@dUE0P14BUH01015BeL0AdeI04M=F@19Ce/204]AG@<0B4iJ00@0AdeI04I<
F016C5P0AD]G0`14BUH304=9E@H0@TQD0P14BUH00`15BeL0ATaH04I<F00304A:EP@0@dUE0P14BUH0
1016C5P0AdeI04M=F@17CUP204I=E`0404E<EP14BeD0A4]E04A;E@80ADaF00<0ATeG04M>F017CUP0
3018CeT304U>E00404U=EP17CUH0AdiF04M>F080ATmH00@0ATiI04E?F@15CeT0@DmJ0P14CUL104E>
E@40B4eA0@3oool000`0BEMR0`18EV4404UGHPD0BUQS2P19Ef0204YHH@`0BEMP00<0AeEP04QFH@18
EV400`18EV4304UGHP<0BUQS3019Ef80101>EV40CEMQ04eGH@1=Ef4304YGH080BEMP0`18EUl01013
EF40AUQR04IHHP16Eel204UFGP0404mJH01;E5T0BeAI034i?0804a/M100C4a<201<C4@<04a@?0`0C
50h401@C4@804a8@0`0A40h0100C4Q005A@B01DD4P0V9R@204do<P0405Pn9@1E>B40ECTQ04dh9@80
B3LW00@0FT@]06Q9;01XBB`0K4P^0P1^BBl0101aBBl0M4Xa07A:<@1hC34207]><@<0ODhb00@0Pe0c
08EB=@25DSD0R5Dh0P2=FSd0102BGd80UV=609ISAP2JIdX209]XB`0409]VA`2OJDd0WfU=0:9/D080
YFeD00@0YVaD0:M]E`2WKEL0ZfiI0P2]L5d0102dMf80[W5L0:iaG02]Led20;=mI@040;B0J02_OF@0
[geT0:amH`80/85W00@0]HU^0:IjG`2VNUl0TF=90P1hBS00101V=ad0Hc4H06<a601Y=ah206lm9004
08mI?`2EGdD0UEm509UVB`80WVYB00@0X6iG0:5_F02QKeP0XG5K0P2SLed309UcG00409E/EP2;HdX0
Rf=:08IL@`80PeU000@0Oe<h07M;<01gBc00OD/`0P29ES/0102BGd@0VfE;09]UB`2RK58209mWCP<0
V61700@0U5Y409YMB02JGDP0WF590P2OHDX0102SID`0[6aC0:a/D`2_Ke@20;1^E0040;EgH02dNV80
]7YR0;=mH`<0/GiS0P2eOfD0102gOFD0^G]T0;UkI02hPVP20;N1I`040;=mH`2]Med0[GMM0:afG080
[gUO00<0/WaR0;5kH@2aNf400`2_O640102^Nf00[GYO0:ejG`2]NUl20:ikH0040:mlH@2`OF80/7eR
0;60HP80[gaO00@0[GYK0:iiFP2^NEX0[GQH0P2[ME<00`2WLDl0Y6i<0:A^C0030:E^C`030:I_D02W
L540Yg1A00<0ZG9C00@0Z75B0:I_D02VKe00YG5:0P2^NU<0102[Me00Xfm80:=_B02VLT/20:UeCP04
0:A`B@2PK4D0X6a50:IaC`80Xfm:00@0X6a709iZ@`2NJT<0W6Q10P2IICd0102DH3P0T5db091M<P2>
GC8208YI<00408ED;02DI3h0U6@n08YL>080PE<a00@0RE`k07I9:01fBBP0P4hY0P2<FSL01029F3P0
Ndhg07]>=`1UACH205=4?@0304a>C@11C500@Da@00<0@daE00@0@T]D041:D`10BU<0?daD0P10CED3
041?EP04049<EP13BeH0@d]F049:E@<0@dUE0P17BeL00`18C5P0BDeI04U=F@0304I<F0<0AdeI1018
CUX204U?F`0304M?FP16CUT0ATiI00@0ATiI0P17CeX304UAG0@0BU9M1@1;Deh304aDG`<0Be=N00@0
BU9M04UAG019DE`0AdmJ0P15CEP00`12BUD0@DUD0459E00<04=;EP@0@dUE0P14BUH01015BeL0AdeI
04M=F@19Ce/204Y@G0<0B4iJ0`17CET00`16C5P0AD]G04E;E`0404A:EP<0@dUE0P12B5@01013BED0
A4YF04A:EP15BeL204I<F0<0A4YF1013BED204A:EP0404I<F017CET0AdeI04M>F080ATeG00@0ADaF
04A;E@14BeD0A4]E0P15C5H00`16CEL0AdiH04M>F00<04M>F0<0BDiD00@0BDeF04M>EP17CUH0AdiH
0P16CeP01016CUT0ADmI04E?F@11CeX204A>E`40ADiE0@18CE410?ooo`003019Ef8304QFH@@0BEMR
1@1:F6<:04UGH080BUQQ3019Ef000`17EF00B5IQ04QFH@0304QFH@<0BEMR0`1:F6<<04UGHP0404iF
H@1=Ef40CEMQ04eGH@<0BUMP0P19Ef0304QFG`0404=EH@16F680AUQR04IGG`80BEIN00@0CeYP04]D
F@1;E5T0<CTl0P0C6ad401<C4`804a<A0`0C50l301<D3P@051<A0P0C4Q03014@3P0401<B400E5180
5A@B02HV9080CClb00@0F3hU05Di8@1E>B40CCPU0P18=bL0101JA2d0J4T/06Q9;01/B2h206i9;`04
0759;`1dBS40M4Xa07Q<<@80Ndha0`1mCS801023D3<0QE8e08EB=@28ECP208eJ?@04099O@P2FHdH0
UV=609YWBP80VfQ;00@0VfI709mYC@2OJDd0XVa@0P2UKE@0102VK5@0YfeG0:M]E`2[KUT20:e`G@04
0;AgHP2^LE`0[W5L0:ecG@80/geU00@0]81X0:mmI02_OF@0[7eS0P2`PFL0102eRFh0YWYO0:IjG`2A
HdT207Q:<00406Hg7@1S<AP0Hc4H06Tg7P80KcdT00@0SeTo09EOA@2EGdD0VFI;0P2NJU80102PKUL0
XFmH0:5_F02QLE/20:=cG@<0VG=L00@0UFaF08]SBP2;HdX0QUa30P23FD00101oDcP0Md/`07M;<01m
Bc0208UF>`04099OA02KID/0VfE;0:9/DP80WfM>0`2HH4L0102DFT@0VUe809YMB02MHDT209mQBP04
0:=UC02/K5<0[6aC0:m_E080/6iD00@0]GMP0;AjHP2dNV80/geS0`2aOV<20;EoI@040;MmI@2iNf@0
^G]T0;R2J080]h5W00@0/geS0:egG@2]Med0[7IL0P2_NEl00`2bO680/G]Q0;5kH@030:mlH@040:ik
H02]NUl0[GYO0:ejG`80[W]P00@0[gaQ0;1mHP2`OF80/H1R0P2_O5l0102]NU/0[WUJ0:iiFP2]N5P2
0:]eD`030:MaC`2TKT`0Y6i<00<0YFi?00<0YVm@0:M`D@2WL5400`2YLU<0102XLE80YVm@0:I_D02U
LDX20:ijD`040:]gD02SKdP0Xfm80:IbB`80ZGE>00@0Y7190:1/A@2PK4D0YW5?0P2SKdX0102PK4L0
WVY309iZ@`2LJ44209UU?@0409AP>02@GC80T5db08iM<P80RUT`00@0QE@/09AT?P2DI3h0RU`h0P21
Dc401029G3/0MTTX07I9:020CRT208aJ=`0408UH>01kCSL0Ndhg06E5=P80Dd@m00<0C4i=045<D011
C5000`13C5D01012Be@0@4YC041:D`0oC5@2041=E@<0@4mF00@0@TaF04=;EP13BeH0@TYE0`13BED2
04M;E`0304Q<F019CET0BDeI00<0ATaH0`17CET404Q>FP80BDmK00<0AdmJ04I>F@16CUT01016CUT2
04M?FP<0BE5L101:DUd504]CGP<0C5AO0`1;Deh0101:DUd0BE5L04UAG017CeX204E=F003049:E@11
BE@0@DUD00`0@d]F1013BED204A:EP0404E;E`17CET0AdeI04U?F`80BU1L0`18CUX304M=F@0304I<
F015BeL0AD]G00@0A4YF0`13BED20498E00404=9E@14BUH0A4YF04E;E`80ATaH0`14BUH404=9E@80
A4YF00@0ATaH04M=F@17CET0AdiH0P16CEL01015C5H0A4]E04A;E@14BeD204E<EP0304I=E`17CUP0
AdiH00`0AdiH0`19CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80A4iG
0@15CUD104Q=D@40oooo000E04UGHP<0BUQS00<0B5IO04UGH019Ef001019Ef0504YHH@`0BEMP0`17
EF0404QFHA40BEMR00@0CUIQ04eGH@1=Ef40CEMQ0`1:Ef0204UGH0<0B5IO00@0@eEQ04EGH@15Ef40
AUMO0P18EEd0101>FEl0BU=H04YCF00`>3/2018J70@04a<C0P0C4a4301<D3`<04a@>0`0E51800`0D
4a404Q4?018A3`03014@3P04018A3`0C4Q004a8@02d/:080Cd8b00@0F3`T05Hk801F>b00CcXU0P19
>2H0101K@R`0J4HZ06Q6:P1/B2h206i9;`04079:<01eBc80MD/b07U=<P80Ndha0`1mCS801023D3<0
QE8e08EB=@28ECP208eJ?@04099O@P2FHdH0UV=609YWBP80VfQ;00@0VfI709iXC02NJ4`0XVa@0P2U
KE@0102VK5@0YfeG0:M]E`2[KUT20:e`G@040;=fH@2ZKEP0ZVeH0:Y`FP80]GmW00@0]X9Z0:ekHP2]
Nf80ZGYP0P2_P6H0102`Q6T0V6aA09Q/D@1mDCH206Pj800405h`5P1Y>R00JCXP07]<<P80QU@k00@0
Uf5709YTBP2JI4X0W6U>0P2OJe<0102PKUL0XW1I0:9`F@2TM5h20:MgH@0409acG@2MM5h0WGAN09]b
G080UfeE00@0U6YA09=WCP2CIdh0RUi30P20DSP0101kBBh0Qe@i08MD>@2DHDH209iXCP040:1ZD02L
I4/0W6A;09MOAP80UUi500@0WF=;09eQB@2MHDT0V5a40P2FF3l0102HFT40XF560:5QAP2XJ4d20:i]
D@040;EgH02eNf<0]G]S0;AnI0<0/7eR0P2dOV@0102fO6@0]WQQ0;IhH@2kPfX20;InI@040;1hG`2[
LeX0Zg=J0:adF`80[gMN00<0/7QO0:mgGP2_Meh00`2_O6440:ikH080[gaQ00@0/7eR0;5nH`2aOV<0
/7mQ0P2_O5l0102]NU/0[WUJ0:iiFP2]N5P20:afE0030:QbD02UKdd0YFm=00@0YFm=0P2VL4h00`2W
LDl0Z79@0:QbD0030:MaC`040:UdC02]N500[GQ@0:]fCP80YW5900@0Yg9:0:UdC02YM4`0YG180P2P
Jd<0102TKTX0XVa80:9/B02OJD<209iX@P0409eX@02KISh0VfHn09IQ=`80Tehd00@0T5db08iK<02>
Fc00REP_0P2BHSX0102=GCL0Q5Hd08AF=02;GC/207]=;004081<9`2=F3H0SEPf08aI>P80Ndhg00@0
IdLh05Y;A01JBd@0CTi>0P0nBDd01013BED0@TQD0498E011BE@3041:E080@4eF00@0@DiG049?F012
CeP0@D]E0P12BUD01011BE@0@TQD0498E013BED204I:EP0304M;E`18C5P0B4aH00@0ATaH0`17CET5
04Q>FP0304M?FP16CUT0ATiI00@0ATiI0P17CeX304UAG0@0BU9M1@1;Deh304aDG`<0Be=N00@0BU9M
04UAG019DE`0AdmJ0P15CEP00`12BUD0@DUD0459E00<04=;EP<0A4YF00@0@dUE04A:EP14BUH0AD]G
0P17CET00`18CUX0BDmK04U?F`0304U?F`0404Q>FP17CET0AdeI04I<F080AD]G0`14BUH00`15BeL0
A4YF04A:EP0404=9E@80A4YF0`15BeL304A:EP@0@dUE0P14BUH01016C5P0AdeI04M=F@17CUP204I=
E`0404E<EP14BeD0A4]E04A;E@80ADaF00<0ATeG04M>F017CUP03017CUP304U>E00404U=EP17CUH0
AdiF04M>F080ATmH00@0ATiI04E?F@15CeT0@DmJ0P14CUL104E>E@40B4eA0@3oool000`0BEMR0`1:
F6<904UGHP@0B5IO0P19Ef0304YHH@<0BeUR3019Ef0404MEH0D0B5IQ3`19Ef80101>EV40CEMQ04eG
H@1=Ef4304YGH080BEMP0`18EUl01013EF40AEMQ04EGH@16Eel204QEG@0404iIG`1:DeP0BU=H030h
>`804QXL100C4a<201<C4@<04a@?0`0C50h0100F5A<05A@B01DD4P0C4Q08018A3`04030_:P1A@C80
DD4b05Pm8P80FC/Q00@0DS`W04Xg9P1:=bH0Fd4Z0P1XABL0101[Abd0KTT_06i9;`1cBc4207I<<`03
07Y><`1lCc80O4lb00<0ODhb00@0Pe0c08EB=@25DSD0R5Dh0P2=FSd0102BGd80UV=609ISAP2JIdX2
09]XB`0409iYBP2QJdl0XF]?0:E_D`80Z71G00@0ZFmG0:U_F@2YKeT0[G1K0P2_LUl0102`Leh0Z6]F
0:Q[EP2YKeT20;EoI`040;N3J`2^O6<0[WaS0:YkH@80[h1V00@0YgiR08IM@@26GD40KT8W0P1S=a`0
101R=1X0LdD[07=5:`2;G48209=TBP0409QRB02JI4X0VVA:09]XC@80WFUA00@0WVaE0:9`F@2RL5T0
YWIP0P2ZNV@0102VOFL0Z7mY0:QoJ@2WOFL20:9hH00409mcFP2KKED0VfeE08mQA`80PeDk00@0P4hc
089?=022Cc@0QE8g0P28DSP01029DcT0U5a309AL@`2TK5<20:mgGP040:abFP2SIdl0XfM?09AH?P80
RT`c00@0QdX^08i><P2>CS80VETm0P2RHDD0102/KUL0/7IN0;1fGP2bO6820:mlH@040;5nH`2gPFL0
]h5W0;QnIP80]WQQ00@0^X1X0;=iH@2cNF40[79J0P2ZL5P00`2/LUX0[G=K0:ecF`030:abFP040:ej
G`2_O640[gaQ0;1mHP80/WmT00@0/h1U0;9oI02bOf@0/GiS0P2`OF80102^OEl0[W]N0:ikGP2/NEX2
0:iiFP040:mjFP2^N5H0[WQF0:]eD`80Z79@00@0Yg5?0:I`CP2VL4h0YFm=0`2TKT`20:I`CP040:Qb
D02ZM580ZWAB0:YcC0<0[7E>0P2YLT/0102WL4T0Z75:0:QaBP2VKdP20:5Z@`040:5YA02PJ4<0X6Q3
09iV@@<0WVM00P2MISd0102IHcL0UV0d09IP=02?G2l2091M<P0408iK<02AH3P0TF0h08iL=`80RETe
00@0RU`j07mA;`1oDBl0PDdX0P2<EcD0102=FS/0Nddf07]==P1YB3T205m>A`0304e=C@0lADX0?4E:
00<0@TQD00@0@DMC0418D`10B5<0@DUD0P11BeD3045>E`04041:E011BE@0@DUD0418D`80@DMC00@0
@TQD04E9E@15BED0ATYF0P17BeL304E;E`<0ATaH0`17CET304Q>FP0304M?FP16CUT0ATiI00@0ATiI
0P17CeX304UAG0@0BU9M1@1;Deh304aDG`<0Be=N00@0BU9M04UAG019DE`0AdmJ0P15CEP00`12BUD0
@DUD0459E00<04A<E`0304E;E`14BUH0A4YF00<0A4YF00@0AD]G04I<F016C5P0B4iJ1@19Ce/01018
CUX0AdeI04M=F@17CET204I<F0<0AD]G00@0ATaH04E;E`15BeL0A4YF0`13BED204A:EP<0AD]G0`14
BUH404=9E@80A4YF00@0ATaH04M=F@17CET0AdiH0P16CEL01015C5H0A4]E04A;E@14BeD204E<EP03
04I=E`17CUP0AdiH00`0AdiH0`19CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI
045?FP80A4iG0@15CUD104Q=D@40oooo000<04UGHP<0BUQS2@19Ef8404QFG`80BEMP0`1:F64304]I
HP`0BEMP1017EF0504QFH@l0BEMR00@0CUIQ04eGH@1=Ef40CEMQ0`1:Ef0204UGH0<0B5IO00@0@eEQ
04EGH@15Ef40AUMO0P18EEd0101>FEl0BU=H04YCF00`>3/2018J70@04a<C0P0C4a4301<D3`<04a@>
00@05QDC01DD4P0E51804a8@200B4@l0100`;bX0DD4b0551<P1H?B8205Tk8@04058l9`1:=bH0BSLV
05]1:P80J4DW00@0JdL]06i9;`1^BBl0Ld/a0P1fC3<00`1jCS<0O4lb07a?<P0307e><P0408=@<`25
DSD0QE8e08QE>080SEXm00@0TUm209ISAP2FHdH0VVM:0P2KJ4/0102NJDX0XF]?0:5[C`2UKe<20:Q`
E`040:U_E`2YKeT0ZFmI0:e`F`80[g9O00@0/7=N0:Q[EP2XJeH0ZFmI0P2eOfL0102gPf/0[WaS0:il
H`2ZNf420:n0IP040:MnHP26GD40QUe106i29`80HcLL00@0HS@J07=5:`1cAB/0Rea20P2CI4X0102H
HTP0VVA:09YTBP2KJ4d209eYD@0409i/E@2RL5T0XW1I0:IfH080ZWYT00@0YWeW0:QoJ@2XOfT0YgeW
0P2RN600102OLeX0VfeE09]]E@2?HDL208=E>`04081><`22Cc@0PTld08EB=`80R58h00@0RE<i09AL
@`2DG4<0Y6aC0P2_Meh0102/LUX0XfM?0:=WC`2DF3h208Y<<`0408M:;P2>CS80SThb09UI?@80XV55
00@0[6iG0;1fGP2`MUh0/WaR0P2_O640102aOV<0]h5W0;N1I`2hOVH20;IhH@040;Z0J02cNF40/gUQ
0:abFP80ZW1H00<0[79J0:ecF`2]Le/00`2/LUX0102]NUl0[gaQ0:mlH@2`OF820;9oI0040;>0I@2b
Of@0/WmT0;5nH`80/7eR00@0[WeO0:ikGP2^Neh0[7UJ0P2^NEX0102_NUX0[WQF0:ihEP2[ME<20:Qb
D0040:MaC`2VL4h0YW1>0:E_C@<0Y6i<0P2VL4h0102XLU00ZWAB0:YdDP2ZLd`30:aeCP80ZG9;00@0
Yg190:QaBP2XLDX0YVm80P2QJT<0102QJD@0X6Q30:1X@`2NIT4309iW@080WFHm00@0VF<g09IP=02F
H3@0Se`_0P2@GC80102>Fc00TF0h095P>02>G3L208UI=@0408YL>P1oDBl0Oe4_085=:080S5Le00@0
SEXk07]==P1kCCH0JDPi0P1OCTL00`1=CDd0?4E:03a5BP030498E0040457D`10B5<0@4QC0459E080
@D]E0`11CUL01010BU@0@DUD0459E010B5<20457D`040498E015BED0ADUE04I:EP80Ad]G0`15BeL3
04I<F0<0AdeI0`18CUX00`17CeX0ATiI04I>F@0404I>F@80AdmJ0`19DE`404YBG@D0Be=N0`1<E5l3
04]CGP0404YBG@19DE`0BE5L04M?FP80ADeH00<0@TYE0459E011BE@03014C5L00`15BeL0A4YF04A:
EP0304A:EP0404E;E`16C5P0ATaH04Q>FPD0BDmK00@0B4iJ04M=F@17CET0AdeI0P16C5P304E;E`04
04I<F015BeL0AD]G04A:EP<0@dUE0P14BUH304E;E`<0A4YF1013BED204A:EP0404I<F017CET0AdeI
04M>F080ATeG00@0ADaF04A;E@14BeD0A4]E0P15C5H00`16CEL0AdiH04M>F00<04M>F0<0BDiD00@0
BDeF04M>EP17CUH0AdiH0P16CeP01016CUT0ADmI04E?F@11CeX204A>E`40ADiE0@18CE410?ooo`00
3019Ef8404YHH`D0BEMR0`18EV4304MEGP0304QFG`19Ef00BEMP00<0BUQQ0`1;FF8<04UGH0<0AUAO
0`17EF0404QFH@h0BEMR00@0CUIQ04eGH@1=Ef40CEMQ0`1:Ef0204UGH0<0B5IO00@0@eEQ04IHHP16
F680AUMO0P19EUh0101?FV00BeAI04]DF@0a>C`201<K7@@04a<C0P0C4a4301<D3`<04a@>00@05QDC
01@C4@0D4a404Q4?0P0A40h0100B4@l04a8@01<B400C4Q02018A3`0403<`:`1A@C40DD4a05Xl8P80
Fc`P00@0DS`U04Pf8P18=R80FClV0P1YARH0101[Abd0KTT_06i9;`1cBc4207M==00307]?=01mD3<0
OE0c00<0ODhb00@0Pe0c08EB=@25DSD0R5Dh0P2=FSd0102BGd80UV=609ISAP2JIdX209]XB`040:9]
CP2UKe<0YFmC0:QbEP80ZW9I00@0Zg5I0:]aF`2[LE/0[W5L0P2`Lf00102_LUd0ZVeH0:Y]F02ZL5X2
0;=mI@040;F1J@2`OVD0/7iU0:amH`80[h1V00@0XgaO07U@=01iD3@0HCPL0P1U>Ah0101W>b00MTP^
07I8;P2;GD<2099TBP0409USB@2LIT`0W6I<09eZC`80X6aD00@0Xg5J0:UgH02YMf00[gmY0P2dQ6h0
102_QFl0/XQb0;:8LP2aQg420:aoJ0040:QjHP2RM5`0XWAL09UZD080Sf1600@0S5Xo081=<P20CC80
LSlT0P1V<1H0101Q:a40MclV07Lo9P2KHdX20;5iH00409EL@@2BES`0TUHl08iB>080Se8f00@0SE0c
08i><P2>CS80Se0a0P2DDcD0102MGdP0YfeE0:M]E@2^N5h20:mlH@040;B1IP2lQV`0_8I/0;^1J@80
]GMP00@0/gUQ0:ecF`2]Le/0ZFmG0P2ZL5P0102/LUX0Zg5I0:]aF@2[LET20:abFP040:aiGP2_O640
[gaQ0;>0I@80]X=X00@0]hAY0;F2I`2ePVL0/h1U0P2aOV<0102_OV00[W]N0:ikGP2]NU/20;1kG004
0;1kF`2aNeT0/G]I0:egE@80ZgEC0`2XLTh0102VL4`0Y6i:0:A^BP2SKDT20:E_B`030:UcC`2]Me<0
[GMC00<0Z75800<0[7E<0:]dB`2[M4/00`2TKD@0102VKdH0Xfa30:=/@`2PJ4<209mW@P<0WFHo00@0
WVLn09eW>`2MIc/0VF<g0P2FH3@0102>FBd0TUda099M<@2@GC82095M=@0408eL=02;FcD0Re/e08]K
=`80Pe8a00@0Q4hZ08QB<028DS00ReHg0P1kCCH0101]BC/0HTm9069?B@1;BDX203i5B`0304E8E`14
AeH0A4MF00<0@DME00@0@4QE041:EP10BUH0@D]G0P12C5P01010BU@0@4QC0418D`0oAe820416DP04
0457D`14B5@0A4QD04E9E@80ATYF1015BeL304I<F0D0AdeI00<0AdmJ04I>F@16CUT01016CUT204M?
FP<0BE5L101:DUd504]CGP<0C5AO0`1;Deh0101:DUd0BE5L04UAG017CeX204E=F003049:E@11BE@0
@DUD00`0A4aG0`15BeL304A:EP0404E;E`16C5P0ATaH04M=F@80B4iJ00@0BU1L04U?F`19Ce/0BDmK
0P18CUX00`17CET0ATaH04I<F00304E;E`0404M=F@16C5P0ATaH04E;E`D0A4YF0`15BeL304A:EP@0
@dUE0P14BUH01016C5P0AdeI04M=F@17CUP204I=E`0404E<EP14BeD0A4]E04A;E@80ADaF00<0ATeG
04M>F017CUP03016CEL304U>E00404U=EP17CUH0AdiF04M>F080ATmH00@0ATiI04E?F@15CeT0@DmJ
0P14CUL104E>E@40B4eA0@3oool000`0BEMR00<0BeUT04YHH`1:F6<00`1:F6<00`19Ef80B5IQ04QF
H@0304QFH@<0AeEN00@0B5IO04UGH019Ef00BUQQ0`1;FF8204aJH``0BEMP1016E5l304MEH0D0B5IQ
3019Ef80101>EV40CEMQ04eGH@1=Ef4304YGH080BEMP0`18EUl01014EV80AeUS04MIH`17F60204YG
G`04051KH@1<EEX0C5EJ038j?@8051`N100C4a<201<C4@<04a@?0`0C50h0100E51804a8@01<B400@
3`d2014@3P0401<B400E51805A@B01@C4@804a8@00@0<BhW050o;`1@?bl0FS`P0P1L?B40101C>R@0
B3<P04Pc801I?BD206]79`0406]7;@1^BBl0KTT_07A<<P80N4he00@0O50e07e@<`1mD3<0OTlc0P1m
CS801023D3<0QE8e08EB=@28ECP208eJ?@04099O@P2FHdH0UV=609YWBP80VfQ;00@0YG1A0:MaE@2W
LED0ZWAH0P2/M5/0102[LET0ZW1J0:Y`FP2]L5/20:mbG`040;5dG`2_LUd0[g9M0:idGP80/G]S00@0
/WiV0;5oIP2aOfH0[GiT0P2/OF<0102SO5l0KdP[06m8:`1K<QH206U0900406m3:01gBc00Md/`08QJ
@080Sf5700@0VfE;09mYC`2OJDl0XVmD0P2WLe/0102/NV<0/h5Z0;>1JP2jRW@20<2@NP040;F;M@2j
T7X0^Y1j0;^>N@80]8M`00@0[X1X0:YjH`2ZNV<0Y7EK0P2MKU@0102PKU<0Tf1509=PA@23D3D206li
7`0405XT2P1R:Q40HRXA07a4:`80RU8i00@0OdH[08E9;`25BBl0Se<g0P2LGd<0102NHD@0V5Tj09QI
>P2ADS<2091?<@0409IH@@2SJE40XfUA0:miG`80/WmT00@0]X=X0;b6K02lQV`0]geU0P2]KeP0102]
LET0ZFeE0:U]E@2ZKUH20:mcF`040;5eG@2`M5`0/7AL0;1dG080]7QP00@0[W]P0;5nH`2aOV<0]H9W
0P2hQFX0102iQV/0^8EZ0;R5JP2ePVL20;B1IP040;:1H`2aOV40/GiQ0;1mGP80/WeN00@0/WeM0;5k
F@2aNeT0[WQF0P2ZM580102WLDd0Z79>0:QbCP2XLTh20:MaC@<0YW1<00<0ZG=?0:afDP2/MU800`2V
K4@0102^M4`0[gE=0:meC@2SJD420:9X@0040:Q^AP2WKDD0Yfe50:E[A@80XfU300@0X6I00:1V?P2P
ISh0WfDm0P2MI3T0102IH3<0UUd`09IM<02=FBX3099M<@80Se/c00@0SUXc08eK=P2=FcH0RUXf0P25
EC401024CRX0PD/Y085;:@26DC8207Y:=00406i:?01RCdT0HTm904Y8B@80@dY@00@0ATUH04E8E`15
B5L0A4MF0`11AeD20418E@<0@4YF00@0?dUC03m7DP0oAe80?dMB0`10AU8204=7D`0304A8E015BED0
ADUE00H0AD]G0`16C5P304M=F@0304M?FP16CUT0ATiI00@0ATiI0P17CeX304UAG0@0BU9M1@1;Deh3
04aDG`<0Be=N00@0BU9M04UAG019DE`0AdmJ0P15CEP00`12BUD0@DUD0459E00<04A<E`0304I<F015
BeL0AD]G00<0A4YF00@0AD]G04I<F016C5P0AdeI0P18CUX304Y@G00404U?F`18CUX0B4iJ04M=F@<0
ATaH0P15BeL01018CUX0AdeI04M=F@16C5P504A:EP<0AD]G0`14BUH404=9E@80A4YF00@0ATaH04M=
F@17CET0AdiH0P16CEL01015C5H0A4]E04A;E@14BeD204E<EP0304I=E`17CUP0AdiH00`0ATeG0`19
CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80A4iG0@15CUD104Q=D@40
oooo000<04UGHP0304]II01:F6<0BUQS00<0BUQS00<0BEMR04QFH@18EV400`18EV4304MEGP0404QF
G`19Ef00BEMP04YHH@<0BeUR0P1<FV<<04UGH0@0AUAO0`17EF0504QFH@`0BEMR00@0CUIQ04eGH@1=
Ef40CEMQ0`1:Ef0204UGH0<0B5IO00@0A5IR04MIH`17FF<0AeQP0P1:Eel0101@Ff40C5EJ04aEFP0b
>Sd201@L7P@04a<C0P0C4a4301<D3`<04a@>00@05A@B01<B400C4Q0040l=0P0A40h0100C4Q005A@B
01DD4P0D4a4201<B4004034^9`1@?bl0D3l_05Xl8080G3dQ00@0DcXT04Pc8018<b00FCdU0P1[AbL0
101[Abd0KTT_06i9;`1dC38207Q>=@0407a@=@1mD3<0OE0c07i?<`80ODhb00@0Pe0c08EB=@25DSD0
R5Dh0P2=FSd0102BGd80UV=609ISAP2JIdX209]XB`040:E`D@2WLED0Yg5E0:YdF080[7AK00@0Zg5I
0:Y`FP2ZL5X0[G1K0P2_LUl0102aM5l0[g9M0:mbG@2^M5h20;5kH`040;9nIP2aOfH0/GmV0:enI080
[7eS00@0XgaO06m8:`1_B2/0Fc8F0P1Y@2@0101_@bP0Md/`07M;<028FT0208mQA`0409]UB`2OJDl0
WfU?0:9_E080Yg=K00@0[7YS0;>1JP2cPFX0^XYd0P30T7X0102eRgD0^Y1j0;Z@NP2kSWT20;B7L004
0:j0J02ZNV<0ZWYS0:AeF`80WFiD00@0X6iC09=PA@2CH4D0Pe0e0P1_>Al0101J90X0HRXA068Z4@1l
A2/208YB>@0407m6:`25BBl0QDT_08mC=`80W5m300@0WV5409QI>P2HFCX0TE8c0P2@Cc40102FF440
XfUA0:=YD@2_NEl20;9oI0040;J3J02lQV`0_8I/0;MmI@80[FmH00@0[G5I0:U]E@2YKED0ZViF0P2_
Le/0102aMEd0/7AL0;1dG02`M5`20;AhH0040:ikH02aOV<0/GiS0;F2I`80^8EZ00@0^HI[0;R5JP2h
QFX0]H9W0P2dPFH0102bPF<0/GiQ0;5nH@2`OEh20;9mGP040;9mG@2aNeT0/G]I0:ihEP80ZWAB00@0
Yg5=0:QbCP2XLTh0Z79>0P2WLDd30:I`C0030:UcC`2/MU80[7IB00<0YVa400@0[WA<0:meC@2_MDd0
XfU10P2RJ400102XKTH0Yfe50:M]A@2UJdD20:=Y@`040:1V@02PISh0X6Hn09mU?@80WF@i00@0VF0c
09IM<02FGC00SETZ0`2BGC4208mK<`0408iJ<`2=FcH0SE/f08YJ=P80QEDa00@0Q4hZ085;:@21BbT0
QU4b0P1jBS@0101^BS`0HTm9069?B@1:B4T204=:D00404I9F015B5L0ADQG04A7EP<0@DME0P10B5D3
041:EP0403m9D`0oAe80?dMB03m7DP<0@4IB0P13Ae<00`14B5@0ADUE04E9E@0604E;E`<0ATaH0`17
CET00`17CeX0ATiI04I>F@0404I>F@80AdmJ0`19DE`404YBG@D0Be=N0`1<E5l304]CGP0404YBG@19
DE`0BE5L04M?FP80ADeH00<0@TYE0459E011BE@03014C5L00`16C5P0AD]G04E;E`0304A:EP0404E;
E`16C5P0ATaH04M=F@80B4iJ0`1:D5`01019Ce/0B4iJ04Q>FP17CET304I<F080AD]G00@0B4iJ04M=
F@17CET0ATaH1@14BUH304E;E`<0A4YF1013BED204A:EP0404I<F017CET0AdeI04M>F080ATeG00@0
ADaF04A;E@14BeD0A4]E0P15C5H00`16CEL0AdiH04M>F00<04I=E`<0BDiD00@0BDeF04M>EP17CUH0
AdiH0P16CeP01016CUT0ADmI04E?F@11CeX204A>E`40ADiE0@18CE410?ooo`003019Ef8304]II0<0
BUQS00<0BEMR04QFH@18EV400`18EV401016E5d0AeEN04MEGP18EUl204UGH00304YHH@1;FF80BeUR
00<0C5YS3019Ef0404IDG`D0AeEP0`18EV4<04UGHP0404iFH@1=Ef40CEMQ04eGH@<0BUMP0P19Ef03
04QFG`0404EGH`17FF<0AeUS04QIH@80BeQP00@0DEaR04eFF`1=EU/0<c/n0P0D71h401<C4`804a<A
0`0C50l301<D3P0401@C4@0B4@l04Q4?00l>308040l=00@04a8@01HE4`0F5A<05A@B0P0D4a40100`
;BH0Cch^04ln;P1K?20205`m800405<j9016<A`0AS4L05Tm9@80KDTY00@0JdL]06i9;`1^BBl0M4`b
0P1hCSD0101lD3D0OU4d07iA=01nCc<207e><P0408=@<`25DSD0QE8e08QE>080SEXm00@0TUm209IS
AP2FHdH0VVM:0P2KJ4/0102VLE80Z79F0:QbEP2[MET20:adF`040:]aF@2YKeT0ZFmI0:a_FP80[G1M
0`2cMV40102aMf40[gUQ0:miH@2`O6@20;5oIP040:enI02ZNf40ZW]Q0:EnH@80K4DX00@0FC8E06i5
:@1^ABT0MD``0P1jCS<0102:G480TF=9095SB@2LIT`20:1ZD0040:AaEP2[Mel0ZgMO0;5oJ080^HM`
00@0`I5k0<NGP@37Uh40`IN10P37VXD01035V8<0^hef0;^=MP2`PVX20:]kI0040:IgG@2PLEL0X75G
09i/D@80VfQ=00@0W6U>099L@P2BG480NT@Z0P1e?B@0101mAB`0OTH]07i6;@2>ECX2091D>00409IJ
?P2OHTD0Wf9509mRA@80VEXk00@0Te@e09AC=@2DDcD0UeU20P2VK5@0102cOF<0]H9W0;F2I`2gQ6T2
0;Z4JP040;1fGP2SIDh0XfE>0:Q/E080YVYB00@0ZViF0;9fGP2bMUh0]GUQ0P2cMel0102eNF40^WiV
0;YnIP2aOV<20;9oI0030;F2I`2gQ6T0]hAY00<0^8EZ0`2gQ6T0102fQFL0]85T0;B1I02bOf020;=n
G`040;9mG@2aNeT0/G]I0:egE@80ZG=A00<0YFm90:QbC02XLT`00`2[MDl0102YLdd0Z79<0:QbC02Y
Ldd20:]eC`<0Y6Y20`2_MDd00`2QIcl0WfDm09mU?@030:Q^AP040:M]A`2UJdD0YF]50:9X@080X6Hn
00@0WVDj09]R=`2KHSL0UUd`0P2BFB`0102=EbT0TUh_099N;`2CGS8208eJ;`0408eI<P2=FcH0SE/f
08YH=@80QeDb00@0Q4`Y07e5901mAB@0Pd`^0P1jBS@0101_Bcd0He1:06=@BP1=BDX204M>E00404Q:
F@16BEP0ATUH04A7EP80@DME00<0@4ID03m7E00oAe@00`10B5D0100nB580?dMB03m7DP0nAU4203m5
D@040416DP13Ae<0@dMC04A8E080ADUE00<0A4YF04E;E`15BeL00`15BeL304I<F0<0AdeI00<0AdmJ
04I>F@16CUT01016CUT204M?FP<0BE5L101:DUd504]CGP<0C5AO0`1;Deh0101:DUd0BE5L04UAG017
CeX204E=F003049:E@11BE@0@DUD00`0A4aG00@0ATaH04E;E`15BeL0AD]G0P14BUH304E;E`<0AdeI
0`1:D5`00`19Ce/0B4iJ04Q>FP0304M=F@<0ATaH00@0B4iJ04M=F@17CET0ATaH0P15BeL304A:EP<0
AD]G0`14BUH404=9E@80A4YF00@0ATaH04M=F@17CET0AdiH0P16CEL01015C5H0A4]E04A;E@14BeD2
04E<EP0304I=E`17CUP0AdiH00`0ATeG0`19CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15
CeT0ADmI045?FP80A4iG0@15CUD104Q=D@40oooo00000`19EUl0BUMP04YGH00604]HH@0304YGH019
EUl0BEIO00<0BEIO101:Ef0504]HH@`0BUIR00<0CEUU04aHI01<F6@00`1<F6@304]GH`<0BUIR0`18
EV401017EF00AUAO04IDG`17EF0204QFH@0304YHH`1;FF@0BeUT00`0BEMR00@0C5EL04]EGP1;EEh0
C5IO0`1:Ef0504YHH`04041BH013EF<0@eES04IFH`80BeQQ00@0DUaU04iGGP1>Eeh0<cY00P0D6b44
01<C4`804a<A0`0C50l301<D3P@04Q4?0P0E5180100G5Q@05QDC01HE4`0B4@l200l>300403<U8P1B
@c00DT<`05M28`80EC/J00@0ES<M05Ha7P1F<Ah0Hd0T0P1VAQl0101aBR/0LT//079;;01dCBh207E<
;P0407I:;@1gBbh0Md/^07]><@80OU4d00@0Oe0b08AE=`24ECL0S5/m0P2@Gd40102BGd80UF9509ER
A@2KIDT209mYC@040:I_D@2XL5D0Z71E0:UaEP80Zg=J00@0[7AK0:adG@2/M5d0Zg=L0P2ZLEd0102W
Led0ZW=N0:YcGP2^M6020;9hI0040;IlJ02fOfX0]WmZ0;>1JP80/89Z00@0Zh9V07I=<@1fCC40Gc<H
0P1]@BH0101`@RP0QUPn08IH?P2>GdD208mPAP040:QbF02MIdd0WFM=0:1]DP80ZWML00@0ZgUN0;62
IP2aPVH0_i1d0P33U7P0102kV7`0^YEj0;ZENP2iTWD20;F<KP040;65J02/Of00[7mP0:QiF`80YgIH
00@0WVmC09]/D02KK500VFM<0P2JJ4d0102LJDh0WFM=09eWC@2GHDL2099L@P@0XVM70P2UJ4T0102T
IdP0XF930:5R@`2IFS/209=D=@0409EP@@2SKE40XfeA0;1jGP80]7aS00@0/gUQ0;1fH02`MV00[W5L
0P2ZKEX0102VJ540YfUB0:MYDP2^L5T20;1dG0040:mcF`2eNf<0]G]S0;Z2J@80/g]R00<0]h5W0;R2
J02hPVP00`2iPfT0102hPVP0]X1V0;J0IP2dOV@20;=mH`040:imG02fQ6<0]XAS0;V7IP80]H1N00@0
/7]I0;5kE`2aNeL0[WQD0P2ZM4h0102VK4H0Z6i80:Q^B02ZL4X20:U_B@<0Z6i800@0Zg5;0:idCP2^
M4h0ZW1:0P2/LT`0102ZL4X0Y6Y40:AZA02SJD<30:I/AP80XfU300@0WFQ00:E`B02UL4P0YW590P2M
J400102DGcL0TUde099M=@2BGCD208mJ<P04091K<@2DGcD0U5le09=N=080SUXb00@0SETa08eL=02=
G3@0S5/c0P27Ec401026BRH0L3DE070e5@1c?BD207=5=@0406U8@@1OCD/0Gde;04Q2AP80>CXo00@0
@d1A04]7F01;AeP0CTMI0P1:@UL01016@ED0ADIJ04E6FP11B5X203e6E`0403i3EP0nA5@0?TAD03m5
E@80@4IB00<0@DQB0499D@12BE400`13BU0304A:EP<0AD]G1016C5P204M=F@<0ADmI00@0A4iH04==
E`13CEL0A4iH0P15CeT00`16D5X0Ae5K04MAF`0=04YDGP80BE=M00@0B59L04MAF`17DE/0ADmI0P14
CUP00`13CEL0@TaF049<EP0304A>F00404==E`12C5H0@TaF049<EP80@deG0`14CUP604I<F0<0AdeI
1018CUX204Y@G00404aBGP1;DEd0Be5M04Q>FP<0AD]G0P16C5P304Q>FP<0AdeI4P16C5P304I=E`<0
ADaF00<0A4]E04=:E013BU@00`13BU@<04I=E`<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015
CEP0A4iH04A>F00oCEP204==EP40ADiE0@19CU810?ooo`0000<0BEIO04YGH01:Ef001P1;F6400`1:
Ef00BEIO04UFG`0304UFG`@0BUMP1@1;F64<04YFHP0304eII@1<F6@0C5QT00<0C5QT0`1;Ef<304YF
HP<0B5IQ00@0AeEP04IDG`16E5l0AeEP0P18EV400`1:F6<0BeUT04]II00<04UGHP0404aEG01;EEh0
BeEN04aFG`<0BUMP1@1:F6<01010DV00@eES04=EH`16EV<204]HH@04059LI@1>Eeh0CUMN03<j@080
51/Q100C4a<201<C4@<04a@?0`0C50h4018A3`805A@B00@05aHD01HE4`0F5A<04Q4?0P0?3P`0100c
9B80DT<`0593<01G@R<205Dk6P0405Hc7@1F<Ah0ES4N06=09080ITHO00@0LDX[079;;01bBb`0M4d^
0P1eC2h0101fBRd0Md/^07M;;P1kCS4207iA=00407m@<P24ECL0Q5Dg08aK?@80T5m100@0TUm209ER
A@2EHTD0VfE90P2OJDd0102VKe40Z71E0:Q`E@2YLEH20:]cFP040:adF`2/M5d0[7AM0:]cG080ZW5M
00@0Yg=M0:YcGP2ZLeh0[WAP0P2bN6@0102fO6P0]WmZ0;IoJP2cPFX20;22JP040:^2IP1fCC40MTda
05lc6080KD4V00@0L48X08IH?P26F3h0SUm50P2?H4H0102XLUP0WFM=09eWC@2PKE820:YgG0040:]i
GP2aPVH0/H9V0;n@M080`iAh00@0^iQl0;ZENP2jUGX0^I9e0P2eS6h0102aQFP0[7mP0:aoH02XNE/2
0:MfF00409i_D`2KK500Vfa@09UWC080VVQ=00@0W6U>09eWC@2MIdd0Uf570P2BG4840:9WA`80YFQ9
00@0Y6M80:5R@`2QHT<0VEXk0P2CE3D0102EH440XfeA0:=]D@2`NUh20;AlH`040;=iH@2`MV00/7IP
0:iaG080ZVeJ00@0YVQA0:MYDP2WJE80[W1I0P2`M5`0102_Le/0]G]S0;EkH`2jPVT20;=kHP030;N1
I`2hPVP0^89X00<0^H=Y00@0^89X0;J0IP2fP6H0]7iT0P2cOF<0102^OE`0]XAS0;J4H`2iQfH20;F0
GP040;1kF@2aNeL0/G]G0:ihE080ZWA>00@0YVa60:Q^B02XKTP0ZW1:0P2YKdT30:Q^B0040:]aB`2^
M4h0[WA>0:Y`BP80[79<00@0ZW1:0:AZA02TJT@0XfU30`2VK4H20:=Y@`0409eX@02UL4P0YG180:Ia
B@80WFQ000@0U5lg099M=@2BGCD0TUde0P2?FS80102@Fc40U5le09AO=@2CGS@208iJ<P0408eI<@2=
G3@0SE`d08aK<`80QeLa00@0QTXV070e5@1`=AD0LcdU0P1cACD0101YB440Gde;05m=B`18@TH203Tj
?`0404=0D@1;AeP0BdMH04i7F@80BT9G00@0AT5E04E6FP15AUX0@DQJ0P0mAUL0100n@eH0?TAD03i4
E00oAED20416DP030458DP12BE40@TUA00<0@dY@0`14BUH304E;E`@0ATaH0P17CET304E?F@0404A>
F013CEL0@deG04A>F080ADmI00<0AU1J04MAF`17DE/03@1:E5h204UCG@0404QBG017DE/0Ae5K04E?
F@80A4iH00<0@deG049<EP12C5H00`14CUP01013CEL0@TaF049<EP12C5H204==E`<0A4iH1P16C5P3
04M=F@@0B4iJ0P1:D5`0101<DUh0Be5M04]AG@18CUX304E;E`80ATaH0`18CUX304M=FA80ATaH0`16
CEL304E<EP0304A;E@13BU@0@dYD00<0@dYD3016CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG
00@0ADeH04A>F014CUP0?deH0P13CEH104E>E@40BDiB0@3oool0000304UFG`1:Ef00BUMP00H0BeQQ
00<0BUMP04UFG`19EUl00`19EUl404YGH0D0BeQQ301:EV8604aHI0<0BeMS0`1:EV8304QFH@0404ME
H016E5l0AUAO04MEH080B5IQ00<0BUQS04]II01;FF@03019Ef80101<EE`0BeEN04]EGP1<EUl304YG
H0D0BUQS00@0@eES04EGI@15EfD0AEER0P18EEh0101?FF80CEIM04eFG@0e?48201LN90@04a<C0P0C
4a4301<D3`<04a@>00@04a8@018A3`0B4@l04a8@0P0E518301LF500401<B400@3`d040l=034S8P80
DD4a00@0ET<S05Dk6P1E>aX0Ec@N0P1F<Ah0101R@2@0ITHO06I67`1_B2T2075::`0407A=;P1fCBl0
MTd_07Q<;`80NDd`00@0O4lb07iA=01nDC@0PE8d0P24ECL0102;FS`0Sei008mN@02CH4<209ISAP04
09YTB02LITX0W6I:0:5ZC080Xf]@00@0YfmD0:]cFP2[LeX0[GEL0`2]MEh20:acG`040:YfGP2ZM5`0
ZWAL0:]aF`80[WAN00@0/WQR0;=mI@2cOFD0/7iU0P2[OF<0102UO600Q5/o08AK?`1M<AH206Tk8@04
0794:P26Ecd0QULm095RB080RU]100@0Y6iD09aVC02LIT`0WfaA0P2XMEX0102YMe`0[geR0:mmHP2i
RVh20;b=L@040;J?LP2fSG40]Xea0;N;L080]HQ[00@0/hAX0;60HP2aP680[W]N0P2/NEX0102WM5D0
Y75B0:AaDP2QKTl20:=^C`040:M`DP2VKe40YVmA0:5XB`80W6=60`2RIdL0102TIdP0YVU:0:IYBP2V
JDX20:9UAP0409eN?`2GF3T0UePi095L?@80WfU=00@0[GMK0;9jH@2bNV40/WQP0P2_MEl0102]L5/0
Z6]H0:Q[F02dMUl20;9dG@<0]GUQ00@0/7IN0;QnIP2hOVH0_XI]0P2gOfH00`2gQFX0^8I[0;R6J`04
0;V7K080]hEZ00@0]H=X0;>1IP2cPFH0/H1O0P2gQF@0102hQVD0]H1N0;F0GP2bOE/20;9lF0040:mi
E@2ZM4h0ZWA>0:]aB`80[79<00@0Zg5;0:U_B@2YKdT0Z6i80P2ZL4X0102^M4h0/WQB0;9hDP2YKdT2
0:]aB`040:U_B@2TJT@0Y6Y40:=Y@`80Yfe700@0Z6i80:AZA02TJT@0WfY20P2VLDT0102WLTX0WfY2
09mZ@P2GHSX209AO=`0409=N=P2BGCD0TUde08mJ<0<0Tehd0P2?Fc<0102>FS80SUde08iM=@2=G3@2
08MG<@04091F<@25CB`0QDd/09IN@`80VFUC00@0T6UJ08M[H027Jf00Lf1J0P1[GEX0101LCE@0Fd]E
05];E@1EAE0204doCP0404Y1DP1:BE/0BTUK04=:G080>dIH0`0nA5@0100oAE<0@4IB0416DP11B582
0499D@0304=:DP13BU00@dY@00<0A4YF0`15BeL604I<F00404I>F@15CEP0ADeH04E=F080A4aG00@0
ADeH04I>F@16CUT0AdmJ0P18D5/=04]CGP80BU9M00@0BE5L04Q@F`18D5/0ATiI0P15CEP01014C5L0
@d]F04=;EP15CEP204A<E`0404=;EP12BUD0@TYE049:E@80@d]F00@0A4aG04E=F015CEP0AD]G1@16
C5P304M=F@@0B4iJ0P1:D5`0101;DEd0BU1L04Y@G017CET304E;E`80ATaH0`18CUX304M=F@@0ATaH
0P15BeL<04I<F0@0ADaF2014BeD<04I=E`<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015CEP0
A4iH04A>F00oCEP204==EP40ADiE0@19CU810?ooo`0000<0BEIO04YGH01:Ef001P1;F6400`1:Ef00
BEIO04UFG`0304UFG`@0BUMP1@1;F64<04YFHP@0C5QT0`1;Ef<504YFHP0304QFH@17EF00AeEP00@0
AeEP0P18EV400`19Ef80BUQS04YHH`0<04UGHP0404aEG01;EEh0BeEN04aFG`<0BUMP1@1:F6<01017
FFL0B5YX04QJJ016EV<204MDG@0404eGH01<EE`0C5EL03LnA08072<Y100C4a<201<C4@<04a@?0`0C
50h401@C4@805QDC0`0G5Q@0100D4a404Q4?018A3`0/8Al204e0;`0405E28P1F>ad0ES/M05De8080
EC8O00@0HT0U06M7801WAb00K4DV0P1`BBX00`1dCBh0N4la07Q?<@0307Y><@0407e@<`1nDC@0OU4d
08AE=`80QEHh00@0RUTk08iM?`2>GCl0U6540P2GI4L309YTB00409]TAP2OId`0WfM<0:E]DP80Zg=J
00@0[WIM0:mgH02_Mf00[WIO0P2]M600102/N600ZG=K0:UcF`2XKUP20:Y`FP040;1fH02aNf<0/G]S
0:ajH@80YGMM00@0XWIK09Q/D@2HK540ICTN0P1S=A/0101b@bT0QePn08MH?P2DHTT208aJ@@040:1X
C`2KID/0VfE;0:5[D@80YgAI00@0Z7EJ0:]iGP2[NEh0/X1U0P2bPfL30;B6K0040;B5J@2ePfP0]H=X
0;EoH`80/g]P00@0/WUL0;1gFP2`MeX0[GME0P2YLe40102XL4l0ZG5@0:UaD02]LU820:iaDP030:Y]
CP2UJ4T0YFQ900<0Y6M800@0YVU:0:MZB`2WJT/0YfY;0P2TIdP0102NHD80VE`m09UL?@2?FS/209eW
B`040:YdF02_Meh0[gMN0;1fGP80[gEO00@0[G1K0:Q[F02XJeP0^G]T0P2fNV80102hO6@0]7YR0;Aj
HP2aMel20;V1J0040<69L02lQ6/0_8A[0;J9K080]hY]00@0^Ha_0;Z=L02jSG00^Xe`0P2hRfh0102f
RF`0]8MZ0;B7JP2dPf820;N5I0040;J4H`2eP5h0]H1N0;=nG080]7iJ00@0/7YF0:]eC`2[MDl0/GMA
0P2_MDl0102]Ldd0ZW1:0:Y`BP2YKdT20:]aB`040;1fD02dNU@0]7YD0:M]A`80ZW1:00@0ZFm90:E[
A@2UJdD0YF]50`2YKdT20:I/AP040:5/A02VLDT0YW590:MbBP80X6]300@0VF@l09IQ>@2FHCT0UF0h
0P2CGSH0102<Ebd0TE`b095L<P2DGcD3095M=@80T5lg00@0SE`d08MG<@27Ec40Q4`W0P28D2d0102S
JT`0[GAI0:edF@2WLe/20:9cG`0409=]F02BKe/0TVmK089OF@80MeAB00@0IDE805LnA01G?T@0DD5<
0P1=BEP01015BUd0>dQI03]8F@0nA5@303m5D`80@4IB00@0@DMC0499D`12BE<0@TUA0P13BU8404A:
EP<0AD]G1@16C5P304E=F0<0A4aG00@0ADeH04I>F@16CUT0AdmJ0P18D5/=04]CGP80BU9M00@0BE5L
04Q@F`18D5/0ATiI0P15CEP00`14C5L0@d]F04=;EP0304=;EP04049:E@11BE@0@DUD0459E080@TYE
0`13BeH404E;E`<0ATaH1P17CET204U?F`0404]AG@1:D5`0BU1L04M=F@D0AD]G1017CET304I<F0D0
AD]G3016C5P704A;E@D0ADaF3016CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>
F014CUP0?deH0P13CEH104E>E@40BDiB0@3oool0000304UFG`1:Ef00BUMP00H0BeQQ00<0BUMP04UF
G`19EUl00`19EUl404YGH0D0BeQQ301:EV8404aHI0<0BeMS1@1:EV800`18EV40AeEP04MEH00404ME
H080B5IQ00<0BEMR04YHH`1:F6<03019Ef80101<EE`0BeEN04]EGP1<EUl304YGH0D0BUQS00@0AeUW
04QJJ018FVP0AUIS0P17E5d0101=Ef00C5EL04aEG00g?T@201`S:@@04a<C0P0C4a4301<D3`<04a@>
100D4a4201HE4`<05aHD00@051<A018A3`0B4@l0;24O0P1=@2l0101E@R80ES/M05Hk7@1E=B0205Db
7`0406909@1WAb00IdLP06a59P80L4TZ00<0M4d^07Q?<@1hCc400`1jCS40101mD3<0OU4d07iA=024
ECL208EF>00408YI>`2>GCl0SUdo09AQA080UfA70`2JI4P0102KI4H0WfM<09mWC02UKE820:]cFP04
0:ifG@2_Mf00[gMP0:ifG`80[GAP00@0[7QP0:UcF`2YLe/0Z6iH0P2ZL5X0102`MV00/G]S0;5kH`2/
NV420:EgG@040:9fF`2HK540V6aA06Di7P80HcDK00@0LT<Y08MH?P27F3h0U6990P2<FT40102PJ4l0
VfE;09]UB`2QJe420:MdF@040:QeFP2[NEh0ZgUN0;:0I@80/X=W0`2dQV`0102dQFT0]H=X0;F3J02e
Of<20;=kH0040;9iG02`MeX0/7MJ0:egE@80ZG=A00@0Z71?0:UaD02YLE00[G9B0P2^LE800`2ZKDh0
YFQ90:EXB@030:AWB0040:IYBP2WJT/0YfY;0:MZB`80Y6M800@0WV5209UL?@2IG3d0SeXk0P2MId/0
102ZM5P0[gMN0:mgGP2`MUh20:meG`040:e`F`2XJeP0Z6]H0;UkI080]WYR00@0^7aT0;AjHP2dNV80
/GMO0P2iPFP01031RG00_8A[0;b4J`2fRF`20;N:K@040;V<K`2jSG00^Xe`0;Z=L080^8]^00@0]XU/
0;B7JP2dQfX0]8=R0P2gQF@0102fQ6<0]H1N0;F0GP2cOU`20;AnFP040;1jEP2[MDl0ZgE?0;5gD@80
[gE?00@0[G==0:Y`BP2ZL4X0ZFm90P2[LD/0102`MU00]7YD0;AjE02WKDL20:Y`BP040:U_B@2UJdD0
YF]50:E[A@<0ZFm90P2VK4H0102QK4@0YW590:IaB@2WLTX20:1[@`0409UT?02FHCT0UV4i09EP>080
Tehf00@0S5L]095L<P2AG380U5le0`2AGCD2091O=`0408eL=027Ec40QeLa08A<9`80R50]00@0XfY<
0:edF@2]M5T0Yg=K0P2RLel0102CKEP0TVmK099_F`22GeT207MDDP0406E5B01G?T@0Eci40551C080
CDUH00@0ADYM03]8F@0kB5T0?TAD0`0oAE<20416DP040457D`12BE<0@TUC0499D@80@dYB1014BUH3
04E;E`D0ATaH0`15CEP304A<E`0404E=F016CUT0ATiI04M?FP80B51K3@1;Deh204YBG@0404UAG018
D5/0B51K04I>F@80ADeH00<0A4aG04=;EP13BeH00`13BeH01012BUD0@DUD0459E011BE@2049:E@<0
@d]F1015BeL304I<F0H0AdeI0P19Ce/0101;DEd0BU1L04Y@G017CET504E;E`@0AdeI0`16C5P504E;
E``0ATaH1`14BeD504E<EP`0ATeG0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0
A4iH03m=F080@deF0@15CUD104U>DP40oooo00000`19EUl0BUMP04YGH00604]HH@0304YGH019EUl0
BEIO00<0BEIO101:Ef0504]HH@`0BUIR0`1<F6@304]GH`<0BUIR0`19EF4604MEH00404QFH@19Ef80
BEMR04UGHP80BUQS3019Ef80101<EE`0BeEN04]EGP1<EUl304YGH0D0BUQS00@0AUQV04QJJ018FVP0
AeMT0P18EEh0101>F640CEIM04eFG@0i@4H201lV;0@04a<C0P0C4a4301<D3`<04a@>00@051<A01DD
4P0E51805QDC0`0G5Q@201HE4`0401DD4P0D4a4051<A02PL7080BSd]00@0Dd8T05Hm7P1F?Ah0ESHQ
0P1F<b00101Q@RH0ITPR06I88P1[A2D206m8:@0307E>;`1iD380NE0b00<0O50c00@0OE0c07iA=01n
DC@0QePj0P26EcT0102:FC/0SUdo08iM?`2EHTD209UVB@0409YTB02HHTH0V69609UTA@80WFM;00@0
Y6iB0:YdFP2ZM5X0[WQN0P2^N600102/MUh0ZW=N0:YcGP2ZMe`20:MaE`040:I/E02[LET0Zg5I0;=i
H@80/geS00@0ZgUN0:5dE`2QM5L0XWIK0P2SMe`0101oDCL0Hc@J06<d6P1Y>R0208YH?`0408iL@`2F
I4/0UVA;09aTB`80WfM>00@0Y6iD0:UcF@2YLeT0ZGIK0P2ZMe`0102/NUl0[G]P0:ekH02fQ6T20;N4
J@040;N1I`2fOV<0]WiS0;AkH080/gMK00@0/GAH0;5aE@2aLED0[G=>0P2YKdX0102WJdL0ZFe90:U]
B@2^Kd`20:m`C@040:e]BP2YJDH0ZFU60:UZB`80ZV]<00@0Zfa=0:U/C@2YK4d0Z6]<0P2SJ4P0102M
HT80VUlo09YO?`2DGd0209iXC0040:QbEP2/M5/0[7AK0:idG080/GMQ00@0/WEP0:mbG`2_LUl0/GEM
0P2`M5`30;5gG`040:mgGP2iPFP0^H5X0<2:L080^hE[00@0]8]]0;J=K`2fSFl0^8ma0P2jTG<0102i
T780]hi`0;N>L02eS6h20;>:K0040;J5I02gQF@0]hET0;F3HP80]X5O00@0]H1N0;AnFP2dOUX0/7YF
0P2/MU00102aMe40/7I@0;1fD02]Ldd30:]aB`80[79<00@0[WA>0;5gD@2aMe40YF]50`2YKdT20:E[
A@040:I/AP2ZL4X0ZW1:0:]aB`80Z6i800@0XFa20:A_A@2TKdD0Y6m50P2PJd40102JIC/0UV4g09IQ
=`2DGcD209=N=00408YE:`2@Fc40T5/a09AO=@<0TUhf0P2AH3P0102>GCD0QeLa08MG<@24CRX208aD
<@040:U]B`2dME@0]7ED0;MfEP80^7]L00@0/GIF0;AnG02dOU`0YfmN0P2GHU@01020CdP0JD=206U3
@P1J@DL20516D@0304I9F00nBUX0?TYJ00<0?dEA0`10AU830457D`<0@TUC0`13BED304A:EP<0AD]G
1016C5P504E;E`0404I<F017CET0AdeI04U?F`80BU1L3@1<DUh204]AG@0404Y@G019Ce/0BDmK04M=
F@80ATaH00<0AD]G04A:EP14BUH00`13BED01012B5@0@DMC0457D`11Ae<20498E0<0@dUE1015BeL5
04I<F0@0AdeI0P19Ce/0101;DEd0BU1L04Y@G017CET304A:EP80AD]G0`17CET404I<F0D0AD]G3016
C5P404=:E080A4]E0`15C5H?04I=E`<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015CEP0A4iH
04A>F00oCEP204==EP40ADiE0@19CU810?ooo`0000<0BEIO04YGH01:Ef001P1;F6400`1:Ef00BEIO
04UFG`0304UFG`@0BUMP1@1;F64<04YFHP@0BeMS0`1:EV8504UEH@0404IDG`17EF00AeEP04MEH080
B5IQ4P19Ef80101<EE`0BeEN04]EGP1<EUl304YGH0D0BUQS00@0A5IT04QJJ018FVP0BEUV0P1;F640
101AFf@0D5UP051IH00k@TP2028Y;`@04a<C0P0C4a4301<D3`<04a@>00<051<A01DD4P0E5180100F
5A<501DD4P04028H5`17>b/0Ac/[05=29080EChO00@0ECHR05@c8P1D<b80HD4X0P1VBRD0101[A2D0
L4TZ0719:P1eCRl207YA<`<0O50c00<0OU4d07mB=@1oDSD00`28FC/0102;FS`0T5m1091O@@2GI4L2
09YWBP0409]UB@2IHdL0VF=709aWB080X6Y>00@0YW1D0:]eF`2[ME/0[GMM0P2/MUh0102YLe/0YVmJ
0:I_FP2VLeP20:E_E@040:M]E@2^M5`0[WAL0;MmI@80]X1V00<0[7YO0:5dE`2QM5L00`2WNEl0102M
KU@0Jc`R06/l8P1O;A@208ID>`0408EA>@2SKeL0XfmG09aRBP80XfUA00@0ZG5H0:]eF`2[ME/0[GMM
0`2[N5d20:ejG`040;J0IP2eOfD0]GmU0;InH`80]GaQ00@0]GUM0;IfFP2fMUX0/g=G0P2cLUH0102]
LD`0ZFe80:U]B02VJTD20:U[AP040:i^BP2_Kd/0[fm;0:e[B080ZVQ500@0[Ve?0:m^D02_KU00[Vm@
0P2ZKDh0102VJDX0X6E50:1UA@2IHD0209MO?P0409eXB@2TKU80Y6iB0:UcE`80Zg=J00@0[WAL0;Aj
I02dNV@0^GaW0P2hNfP30:abFP040;9hH02bNV40/WYQ0;1hG`80^89X00@0_HM]0;B1IP2dPFH0/ha]
0P2eSVl0102hTG80^I9c0;VBL`2iTW<20;J?L0040;><K@2`RFX0/8UZ0;N6I@80]XAS00@0]H=R0;J1
G`2fPEl0]7mM0P2bO5P0102_NED0[WQB0:ihDP2^M4h50:ecC@<0[79<00@0ZW1:0:U_B@2YKdT0Y6Y4
0P2XKTP0102YKdT0YVa60:I/AP2WKDL30:]aB`80Z6i8102PJcl209iY?@0309UT>02DGc<0U5lc00<0
TUda00@0ReH/095L<P2AG380UF0f0`2CGcL2099Q>@0408mN=P28F380R5Pb08YF<@80T5Pc00@0Y6Q4
0:e[B02]JdP0^WA@0P35Oe/0102oNe@0a8=K0<B3F`2oNF020;5^F`0409UND020CdP0P4m806M5A080
Dd9800<0B4M?04A:EP14BUH00`10Ae430416DPH0@DMC1013BED304A:EP/0AD]G00@0ATaH04Q>FP18
CUX0BU1L0P1;DEd=04aBGP80Be5M00@0BU1L04U?F`19Ce/0AdeI0P16C5P01015BeL0A4YF04A:EP13
BED20498E0040457D`10AU80@4IB0416DP80@DMC00<0@TQD04=9E@13BED00`14BUH404E;E`H0ATaH
0P18CUX0101:D5`0BDmK04U?F`16C5P504A:EP@0ATaH1@15BeL304A:EP`0ATaH1013BU@204A;E@<0
ADaF3`16CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>F014CUP0?deH0P13CEH1
04E>E@40BDiB0@3oool0000304UFG`1:Ef00BUMP00H0BeQQ00<0BUMP04UFG`19EUl00`19EUl404YG
H0D0BeQQ301:EV8404]GH`<0BUIR1@19EF401016E5l0AeEP04MEH017EF0204QFHA80BEMR00@0C5EL
04]EGP1;EEh0C5IO0`1:Ef0504YHH`0404AFI018FVP0B5YX04UIIP80BeQQ00@0DE]T051IH01@FF00
>d980P0R:Bl401<C4`804a<A0`0C50l301<D3P0301@C4@0E51805A@B00@05QDC1@0E5180100R61L0
Ac/[04Lk:`1C@R@205Dn7`0405Df8P1D<b80E3<R0651:080ITXU00@0Jd@U0719:P1`BBX0MDh_0P1j
DC<307a@<`0307iA=01oDSD0Oe8e00<0R5Tk00@0ReXl091O@@2@Gd40UfA70P2JIdX0102KIDT0VF=7
09USA`2LIdP20:1ZCP040:I`E02[ME/0ZgEK0:egG@80[7IN00@0ZG=K0:I_FP2VKeX0YW=H0P2UKeD0
102WKED0[WAL0:idG02gOFD20;J0IP030:ajG`2QM5L0XGAG00<0YgUO00@0WFiD06/l8P1[?280GbdD
0P26E3/01025DCT0XfmG0:=_E`2LHTX20:=YD@040:UaF02[ME/0ZgEK0:egG@<0ZgQM0P2]NUl0102f
P6H0]GmU0;EoI@2fOV<20;ElH@040;EiG@2fMUX0]WIJ0;=cE`80/g9F00@0[G5<0:U]B02YKDP0YVY5
0P2YJdH0102^KTX0[fm;0:m_B`2]JdP20:YXA@040:i]C`2_KU00[fi@0:i_D080ZVe>00@0YVU:0:1U
A@2PIDD0VF500P2GGch0102MJ4T0Y6iB0:A^DP2YLeL20:]cFP040:idG02dNV@0]7YT0;UlI`80^7]X
0`2/LUX0102bN600/WYQ0;9jH@2`N5l20;R2J0040;f7K@2dPFH0]85V0;><K@80]Hi_00@0^95b0;VB
L`2iTW<0^I9c0P2fSg00102cS6d0/8UZ0;29JP2gQVD20;J4H`040;F3HP2fPEl0]X5O0;AoG@80/WaH
00@0[gUE0:ihDP2^N580[WA>1@2]Ldd30:abC0040:Y`BP2YKdT0ZFm90:AZA080Z6i800@0ZFm90:I/
AP2VK4H0Yfe70`2[LD/20:Q^B0@0X6/o0P2NJCd00`2II3P0U5lc09AO<`03099M<@0408]F;02AG380
TE`b09EP=P<0Telg0P2BHCT0102?GSH0R5Pb08QH<P2:ES42091H<`040:AXA02]JdP0[F]80;YdD080
aGmK00@0_g]D0<B3F`34Pe/0_gUP0P2aKU/0102IGU00P4m8081?B01WAD@205=2B00304Q7C`14BUH0
A4YF00<0@4MA0`10AU860457D`@0@dUE0`14BUH;04E;E`0404I<F018CUX0B4iJ04Y@G080Be5M3@1<
DUh204]AG@0404Y@G019Ce/0BDmK04M=F@80ATaH00@0AD]G04A:EP14BUH0@dUE0P12B5@01011Ae<0
@4IB0416DP10AU820457D`030498E013BED0@dUE00<0A4YF1015BeL604I<F080B4iJ00@0BU1L04U?
F`19Ce/0ATaH1@14BUH404I<F0D0AD]G0`14BUH<04I<F0@0@dYD0P14BeD304E<EPl0ATeG0`18CE<0
1018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH03m=F080@deF0@15CUD104U>DP40oooo
00000`19EUl0BUMP04YGH00604]HH@0304YGH019EUl0BEIO00<0BEIO101:Ef0504]HH@`0BUIR0`1;
Ef<304YFHP@0BEEQ0P18E6001016E5l0AeEP04MEH018EV4604UGHP80B5IQ3019Ef80101<EE`0BeEN
04]EGP1<EUl304YGH0D0BUQS00@0@UAR04MII`17FFL0BUYW0P1=FV<0101CGFH0DU]R059KHP0oAT`2
02L^=0@04a<C0P0C4a4301<D3`<04a@>00<04a8@01DD4P0E51800`0F5A<401@C4@805QDC00@07A@E
04Di:`15>B/0DT<V0P1F?R80101D>2<0Dc<T05<c901Q@bT206M;9P0406i7:01aBR/0LDX[07I?<080
NE0b00@0Ndlb07a@<`1lD3<0OU4d0P1oDSD01028FC/0REXl08UJ?02>GCl2099Q@`0309QUB02KJ4/0
VfQ;00<0WVQ<00@0X6e>0:=`D`2SL5<0YgAG0P2YMU/0102ZMe`0Z7AL0:QdG02TL5P20:9^F0040:Aa
E02TKU80Y6iB0:Q_E080[WEJ00@0]GaQ0;AnHP2dOV80ZgYL0P2RMEH0102XNV00YWMM0:IgG@2]OV@2
081>=@0405/Y401iABd0NDD]085==@80XfeE00@0VEm70:I/E02VK5@0[79J0P2/M5/30:miG`040:ai
GP2_O640[gaQ0:mlG`80/G]O00@0/g]P0;AkGP2dNeh0]W]M0P2fNE`0102gN5T0^7MI0;QgF@2_ME02
0:]aC0040:Q/B02YKDT0ZFe90:i_C080/75>00@0[fm<0:][B02[JdP0[fa?0P2`KE00102`Ke40[6e>
0:a]CP2SJ4P209]S@P0409IP?P2DGS`0U5hl0:A_D080ZG=G00@0[7IJ0:eeG02]ME`0/GMO0P2iOfT0
102nPF`0_H1]0;f0K@2bN6020;=iH@<0]gmV00@0/geS0;J3J02fPfP0^XM/0P2aOV<0102eS6h0]hi`
0;N>L02jTG<20;bCM@040;^BM02gSW00]hi`0;>:K080/8MY00@0]XET0;J4H`2fQ6<0]XAS0P2gPV00
102cOU`0[WQD0:ihE02^N5@20;5kE@@0[79<0`2]Ldd20:]aB`040:M]A`2TJT@0Y6Y40:AZA080ZFm9
00@0ZW1:0:Q^B02XKTP0Z6i80P2/LT`0102[LD/0Z6i80:Q^B02PJcl209iY?@<0WFPl00@0VVDi09AO
<`2DGc<0TE``0P2BGC40102=F2h0Tehd09=N=02EH3H3099N=P80TF0h00@0Sehf08QH<P28F380Q5<[
0P2;EBl0102JGSX0XEll0:5O?02bKT/20;mkF0040;YfD`2nOEL0_WeG0<MlI@80^gAP00@0ZfUI09=M
D`2CGE<0Me1;0P1OAdL00`1?ATT0BdY?04]:C`030417C`040417D@10AU80@4IB0457D`<0@DME0P11
AeL00`12B5@0@dUE04=9E@0304=9E@<0A4YF0`15BeL01017BeL0ATYF04I:EP16BUH204M;E`0404Q<
F01:CUX0BTiJ04a@G080CE5M3@1>DUh204eAG@0404a@G01;Ce/0BdmK04U=F@80B4aH00@0Ad]G04I:
EP16BUH0ADUE0P14B5@604=7D`0304A8E015BED0ADUE00@0A4YF0`15BeL604I<F080B4iJ00@0BU1L
04U?F`19Ce/0ATaH0`13BED204A:EP@0ATaH0`15BeL504A:EP`0ATaH1`14BeD504E<EP`0ATeG0`18
CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH03m=F080@deF0@15CUD104U>DP40
oooo00000`19EUl0BUMP04YGH00604]HH@0304YGH019EUl0BEIO00<0BEIO101:Ef0504]HHA00BUIR
1@19EF4304QDH00404ECGP16E5l0AUAO04QFH@80BEMR00<0BUQS04UGHP19Ef800`18EV4<04UGHP04
04aEG01;EEh0BeEN04aFG`<0BUMP1@1:F6<01012E680AeUW04MII`19FFH204]HH@04055KI01CG6<0
DeaS04=:D080;SDk100C4a<201<C4@<04a@?0`0C50h0100B4@l051<A01@C4@0F5A<201DD4P0401<B
400B4@l04Q4?01@C4@805QDC00@06Q4B044h:@11>2T0DT<V0P1E@2<0101E>2H0Dc<T05<c901Q@b/2
06I<:@04075::`1cC2d0Ld`]07I?<080N4la00@0NTha07]?<P1kCc80OU4d0P20DcH01027F3X0Re`n
08]L?P2AH48209ETAP0409UVB@2LJD`0W6U<0:5[C`80XfeA00@0YG9C0:IcEP2VLeH0Z7EH0`2YMU/2
0:McF`040:A`F02SKeT0XfmI0:EbE@80YW1D00@0YfiC0:YaEP2ZLEH0[WEJ0P2]Me/0102XMeT0XWEF
0:9eEP2VMed20:EfG0040:YhG`2HITd0V6I=064]5@80IC4I00@0QU0h095K@`2AFd<0V5a40P2XK5@0
102/LUX0ZW9I0:YbF@2`N5l20;1jH0040:ihGP2cOF<0/geS0:imG`80/7eP00@0/giO0;AmG`2dOEl0
]GaN0P2fNe/0102fNEX0^7UJ0;QiFP2]MeD20:YbD@040:E]C02XKDd0Z6e=0:]`D080[W5B00@0Zfi?
0:Y[C02ZJd`0[VU<0P2^Jdh0102^KDl0ZV]<0:Y[C02QITH209YR@@0409IP?P2DGcd0U5lm0:E`D@80
ZgEI00@0[gUM0;5iH02aNF00]WaT0P2lPV`0102oPVd0^gi[0;]nJ`2eOF@20;AlH`040;R0I`2fP6H0
]X1V0;9lHP80]X=X00@0^hQ]0;>1IP2cPFH0^Xia0P2lT7<0102oTgH0`9Ag0<2DM`2nTWD20;Z>L@04
0;F9K02aQFP0/HEX0;B3HP80]H=R00@0]hET0;R3H@2hPf40/7]I0P2ZM500102/MU80/geG0;=mE`2`
MU020:idCP040:abC02[LD/0Zg5;0:abC080Zg5;00@0Z6i80:E[A@2UJdD0YF]50P2ZL4X0102[LD/0
Z6i80:Q^B02YKdT20:abC0040:]aB`2WKDL0Yfe70:9^?`80WVXk00@0WFTj09m[?02OJc`0W6Pi0P2E
HC80102BGRl0U60a09AP<@2AG38309EP=P80TEde00@0T5`d091O=`2@GcL0Sehf0P29FC<01024DbX0
TU`f099L=P2SJD@20:QYB0040;QgF@32Pf@0`X=T0;QkGP80_X=U00@0a7]X0;]dHP2kM680[feO0P2Q
J5l0102:GeP0Ke9>06mBCP1IBTL20518A@040458CP11B500@DQ@0458D080@DQB00<0@4IB0416E010
AU@00`10AUH30498E0@0@dUE1@14BUH404I:EP80Ad]G00@0B4aH04Y>FP1:CUX0C51L3`1>DUh204eA
G@0404a@G01;Ce/0BdmK04U=F@80B4aH00@0Ad]G04I:EP16BUH0ATYF0P15BED01014B5@0@dMC04=7
D`13Ae<204A8E00304E9E@16BUH0ATYF00<0@dUE0`14BUH404E;E`<0ATaH0P18CUX01019Ce/0B4iJ
04Q>FP15BeL304=9E@80A4YF00<0ATaH04E;E`15BeL00`15BeL304A:EP<0@dUE3016C5P404E<EPP0
A4]E3016CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>F014CUP0?deH0P13CEH1
04E>E@40BDiB0@3oool0000304UFG`1:Ef00BUMP00H0BeQQ00<0BUMP04UFG`19EUl00`19EUl404YG
H0D0BeQQ401:EV8504UEH@<0B5AP00@0AE=N04IDG`16E5l0B5IQ0P19Ef800`1:F6<0BEMR04UGHP03
04QFH@`0BEMR00@0C5EL04]EGP1;EEh0C5IO0`1:Ef0504YHH`04049DHP17FFL0AeUW04UIIP80BeQQ
00@0DE]T05=LH`1CG6<0@dY@0P0^=C/401<C4`804a<A0`0C50l301<D3P04018A3`0D4a4051<A01HE
4`805A@B00@04a8@018A3`0B4@l051<A0P0F5A<0100J4A80@CPY044h:@1B@bH205E08`0405Dh9P1C
<b@0Dc<T0653:`80IT`Y00@0LDX[07=<;@1cC2d0MTl`0P1hCc40101jCS40Ndlb07]?<P1nDC@2081C
=P0408MH>P2;G3h0Re`n095P@P80UFA600@0VFI909aYC02LJD`0XF]?0P2SKE40102ULU<0YW=F0:Ic
EP2XMEP30:UfF`80Yg=K00@0Y71H0:=_F@2SKeT0YG9E0P2VL5@0102WKU<0ZW5F0:YaEP2^MEX20:eg
F`040:QgF@2RMEH0XWEF0:IgG@80YGIL00@0ZWQO09QVC@2HITd0HBdE0P1U<AT01026D3P0TE]3095K
@`2HG4@20:Q/E0040:abFP2ZLUT0ZW9I0;1hG`80/7YP00@0[WQN0;=mH`2cOF<0[WeO0P2`OF00102c
OUl0]7eO0;AmG`2eO5h20;IkF`040;IiFP2hNEX0^7UJ0:egE@80ZW9A00@0YFe<0:Q]C@2XKDd0Zg1@
0P2^LE80102[KTl0ZV]<0:Y[C02^JD`20:i[CP040:i]C`2ZJd`0ZV]<0:5VAP80VV9100@0UV0n09AO
?@2DGcd0YG1A0P2[MET0102_NEd0/GUP0;5iH02fO6@20;b2K0040;n2K@2kOV/0^gi[0;EmI080]7aS
00@0^81W0;J0IP2fP6H0/WaR0P2fPfP0102kR6d0/h5V0;>1IP2jSW420;b@L`040;nCMP30U7L0`9Ag
0;jBM@80^Xia00@0]HU/0;65J02aQFP0]8=R0P2ePf80102gQF@0^8=Q0;R3H@2`NeT20:YdD0040:af
DP2cOEL0/geG0;1fD080[WA>00@0[79<0:]aB`2[LD/0[79<0P2[LD/0102XKTP0YF]50:E[A@2UJdD2
0:Y`BP040:]aB`2XKTP0Z6i80:U_B@80[79<00@0Zg5;0:M]A`2WKDL0XVho0P2NJS/0102MJCX0Wf/l
09m[?02LJ3T209EQ<P04099N;`2DH340U60a095L<P<0UF0f0P2AGCD0102@G3@0T5lg091O=`2?GSH2
08UI<`0408AC:P2BG3H0TU`f0:=YA080Z6U800@0^7MI0<:3I032Pf@0^7]N0P2nPfD01034NfP0^gAR
0;]dHP2_KEl20:5XG`0408YOF01_DTh0Ke9>05U:A`80D4Q500@0@DQ>0458D011B500@DQ@0P11B580
0`10AU80@4ID0416E0030416EP<0@TQD1013BED504A:EP@0ATYF0P17BeL01018C5P0BTiJ04Y>FP1<
D5`?04iBGP80CE5M00@0C51L04]?F`1;Ce/0BDeI0P18C5P01017BeL0ATYF04I:EP16BUH204E9E@04
04A8E013Ae<0@dMC04=7D`80A4QD00<0ADUE04I:EP16BUH00`13BED304A:EP@0AD]G0`16C5P204Q>
FP0404U?F`18CUX0B4iJ04E;E`<0@dUE0P14BUH00`16C5P0AD]G04E;E`0304E;E`<0A4YF0`13BED<
04I<F0@0ADaF2014BeD<04I=E`<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015CEP0A4iH04A>
F00oCEP204==EP40ADiE0@19CU810?ooo`0000<0BEIO04YGH01:Ef001P1;F6400`1:Ef00BEIO04UF
G`0304UFG`@0BUMP1@1;F64@04YFHP<0BEEQ1@18E6001015Deh0AUAO04IDG`18EV4204UGHP0304YH
H`19Ef80BEMR00<0B5IQ3019Ef80101<EE`0BeEN04]EGP1<EUl304YGH0D0BUQS00@0A5IT04MII`17
FFL0AeMT0P18EEh0101?FF80DeaS05=LH`16CE<203<j@0@04a<C0P0C4a4301<D3`<04a@>00@04Q4?
01@C4@0D4a405A@B0P0D4a43014@3P0401<B400E51805A@B01P?4P80@3HZ00@0DD4W05E08`1E@2<0
ECPV0P1D=2D0101Q@b/0IT`Y06I<:@1cC2d207E>;`0407M@<@1hCc40N4la07U=<080NTha00@0OU4d
081C=P20DcH0QULi0P2;G3h0102CHT@0UfI809MVB02JIdX209aYC0040:9/D02VL5@0YW1D0:EdE080
YWEG00@0YgIH0:QfF`2XMU/0YgEJ0P2VM5/0102TLUT0Xg5J0:=aFP2XMEH20:MbD`040:I]D02UK4l0
YFa?0:I]D080YW5B00@0YGAD0:=fE@2SMUD0XW=I0P2VMed0102OKE@0Z7IM0:QfG@1X=1`205PR2P04
08]E?@20BS80P4Xb09EI@@80Z6aD00@0Zg5I0:U_E`2YKeL0[gMN0P2aNF00102_NEl0]7iT0;AnI02c
PV@20;>2HP040;B1HP2eP600]H1P0;AmGP80/g]J00@0/WMG0;9eEP2bMEH0ZgIF0P2WL540102RJd`0
XfY<0:=ZC02VKDl20:U^D0040:M/CP2TJD/0Y6U;0:]VB@80[FQ;00@0[FY=0:QYBP2XJDX0X6E50P2H
HT00102EH3h0UV4o09IQ?`2SKTl20:]eF@040;5kG`2eOF@0]GeT0;Z0J080_H=]00@0_H1[0;QkJ02h
NfP0/WYQ0P2`N5l0102bNV40/7YP0;1jH02/NEh20;F2I`040;f;L02gQFX0]hEZ0;jAM080`9=f00@0
`iIi0<BGNP34UgX0`IAg0P2lSg80102gRVd0/hIY0;>6J@2bPF020;B2H@040;N5I02hPf40^8=Q0:mj
F080Yg5=00@0ZgEA0;EoF@2eOeT0]7YD0P2`MU000`2/LT`0ZW1:0:Y`BP030:]aB`040:U_B@2WKDL0
Yfe70:I/AP80ZW1:00@0[79<0:U_B@2YKdT0ZFm90P2/LT`0102ZL4X0YVa60:I/AP2ULD820:1/?@04
09iZ>`2QKCh0XFdn09m[?080Uf<d00@0U60a09IR<`2FHS<0Tehd0`2FHCL2091L=00408mK<`2@GcL0
T5lg08iM=@80RETc00@0OD`S099N=P2BGSH0Z71=0P2^LE80102kOV80`HE[0<65J`2dO6<20;Z4K004
0<1kJ`2eLV80]G9R0:a/H080YVeT00@0UfYT07eLE@1mG5D0HDe60P1BACd30458CP040458D011B580
@DQB0416DP<0@4IF0P10AEP40498E0<0@dUE1@14BUH404I:EP80Ad]G00@0B4aH04]?F`1;Ce/0CE5M
3`1>DUh204eAG@0404a@G01;Ce/0BdmK04U=F@80B4aH00@0Ad]G04I:EP16BUH0ATYF0P15BED604A8
E00304E9E@16BUH0ATYF00<0@dUE0`14BUH704E;E`80AdeI00@0BDmK04Q>FP18CUX0AD]G1@13BED6
04E;E`<0A4YF0`13BED<04I<F0<0ATeG0`15C5H00`14BeD0@dYD04=:E00304=:E0`0ATeG0`18CE<0
1018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH03m=F080@deF0@15CUD104U>DP40oooo
000404]HH@D0BUMP0`19EUl<04YGH0@0BeMS1@1:EV8604UEH@@0B5AP1@17Del<04UGHP@0AeEP0`18
EV4504UGHP0404]DF`1:E5d0BUAM04YDG@<0BEIO1@19Ef801017FF<0B5YT04QJI01=GVP204aIHP04
04QBG01<EEh0C5EN04M>F080<cY200@06QXJ01LG5`0G5aL051@D0P0C4a4301@E400301<D3P0A4P`0
4A8<00`051<A00@0514:038Y800b:B00Bch^0P1B?BX0101A=b00DcTP05<i801L@RT206E;=00406m:
;@1dCc80M4lb071;;P80LDX]00@0NE8e07]B=P1kDSH0OE4f0P24F3d01024EcP0RE`m08UL?@2?HT<2
09AUA`0409IWB@2KJT`0VfY<0:1_D@80Y75D00@0YgUH0:EfF02UMUP0Y7EG0P2VMe/0102XNEd0YgQN
0:MhGP2TME/20:1`F@040:YdFP2WLEL0Yg5G0:=]D`<0XVaB0P2UKeD0102VL5H0Yg5G0:MaE`2OLUD2
09]^D@0409Q[CP26Ecd0QULm07e==P80E28;00@0LChY08iKAP2>FdH0T5i30P2LJDh0102VKUD0YVeB
0:I]DP2WKU420:aeE`040:ilF`2^OE`0[WeL0;:1H`80/GiQ00@0[W]L0;5jG02aNU`0/gYL0P2dNE/0
102aM5D0[G1A0:e`D@2^Ke020:MZB`0409mWAP2MID@0WFE40:1XA`80YVU:00@0ZVU;0:eXB`2]J4/0
[Ve?0P2ZJD/0102RHd@0Veho09]N?`2IG3d209YO?`0309iS@`2QITH0XFI600<0/WaR00@0/g]R0;Al
H`2dO6<0]geU0P2hOVH30;b0J0040;AjHP2hP6L0^81W0;AnI080ZgQM00@0[geR0;b=L@2lSG40_i9e
0P2fRVd01030V7h0_iQk0;nHN`2nUgX20;nFN0040<:FN@30UGD0`9Ee0;b?L080]hYY00@0/X5P0;N5
I02gQF@0^hUX0P2lQV@0102gOeh0]W]I0;IkF@2hO5X20;]oF`<0/7I@00@0[gE?0:idCP2^M4h0Zg5;
0P2YKdT00`2WKDL0YVa60:I/AP030:]aB`@0[79<0P2ZL4X0102XKTP0Yfe70:M]A`2WKT420:E/?`04
09mY>`2KICL0VfDg09MS=080Tf0a00@0TUl`095N;`2AGRl0Tehd0P2HHcT0102II3X0UF4i09EQ>@2D
H3P209=R>P04095P>02:FS@0RUXd07m79P80TUXi00@0YFe<0;1eE@2`MED0]7UI0P2hNe`0102jOEh0
^GaM0;UlG@2lNEl20;ijH`040;9^F@2`Ked0/6mM0:aaH@80WV]J00@0S65@075:>@1aBST0FT<k0P19
@dL0100lAeT0>TYQ03Y:H@10BEX20454D`040419FP0oCfH0?dmV03m8F@80@4MI00@0@DMG04A8E014
B5@0ADMC0P17Ae401019B500BTM@04Y7D012B5@204=9E@0404A:EP16C5P0ATaH04Q>FP80BU1L00<0
Be5M04aBGP1<DUh00`1;DEd604Y@G0<0Be5M00@0CE=O04aBGP1<DUh0BU1L0P18CUX00`16C5P0AD]G
04E;E`0304E;E``0A4YF1012B5@304=9E@D0A4YF1P16C5P01015BeL0@dUE04=9E@12B5@20457D`<0
@dYD1014BeD504E<EP0404A;E@15C5H0ADaF04E<EP80ATeG00@0ADaF04A;E@14BeD0@dYD0P12BE<0
1016CEL0ADaF04E<EP13BU@60499D`80@dYD1`17CUP504I=E`<0B4eC00@0B4aE04I=E@16CED0ATeG
0P15CUL01015CEP0A4iH04A>F012D5/204E?F040ATmF0@19CU810?ooo`00101;F64504YGH0<0BEIO
301:Ef0404]GH`D0BUIR1P19EF4404QDH0D0Ae=O3019Ef8404MEH0<0B5IQ1@19Ef80101;E5/0BUAM
04YDG@1:E5d304UFG`D0BEMR00@0AeUS04QJI018FV@0CEiX0P1<FF801018DU`0C5EN04aEGP17CUP2
03<j@P0401XJ6P0G5aL05aLG01@D50804a<A0`0D5A000`0C50h04A8<014B300<01@C4@0401@A2P0b
:B00<RTP04/n;P80DSdZ00@0DCLP05<i801C>B00G48Y0P1UBc@0101_BRd0M4lb07A?<P1`Bbh2075:
;@0407UB=@1kDSH0Ne8f07eA=P80Q5Pm00@0Q5Lh08UL?@29G3d0Sf930P2DIDL0102FIdT0VfY<09]Z
C02PKe420:AaE0040:MiF02UMUP0YGIH0:AeE`80YWMK00@0Z7UM0:MhGP2WN5h0Y7EK0P2PL5T0102Z
M5X0Yg5G0:MaE`2SKE<30:9/DP80YFmE00@0YW1F0:MaE`2WLEL0Wg9E0P2KKU40102HJdh0QULm08IG
?@1mCCH205@R2`04074n:@2>FdH0SU]6091N@`80W6U>00@0YViE0:I]DP2VKE80YfiA0P2/MEL0102^
O5/0[WeL0:imG02bPF<20;5nH@040:ikG02aNU`0/GYL0;=jG080]7UK00@0/GAE0:e`D@2]L540[Vm@
0P2WJT/0102OIdH0WFE409eUA02PJ4L20:IYBP040:YYB`2]J4/0[FQ;0:i]C`80ZVU;00@0XV=409]N
?`2KGSl0VE`m0P2JGcl00`2NHd<0XFI60:5VAP030;9lHP040;=kHP2dO6<0]7aS0;MmI@80^7iV0`2l
P6P0102dNV80^81W0;R0I`2dOV@20:]hG@040:mmHP2lSG40_8ea0;nBM@80]XY]00@0`9Qn0;nHN`2o
V7/0_YMj0P2oUWP01032UWT0`9Ee0<2EM@2lSg020;N:J@040;:1H02gQF@0]hET0;^9J080_8IT00@0
]gmN0;IkF@2fNeT0^7aJ0P2kOe/30;1fD0040:meC`2^M4h0[WA>0:]aB`80ZFm900<0Yfe70:I/AP2V
K4H00`2[LD/40:abC080ZW1:00@0Z6i80:M]A`2WKDL0Yfi10P2UK3l0102OJC/0VfDg09]U=`2GHc@2
09=P<@04099O<02AGRl0TEh_09=N=080V6<i00@0VF@j09EQ>@2EHCT0U60h0P2CHSX0102AH3P0RUXd
08YJ=01oAbH2099J>@040:E]C02`MED0/7EE0;AiF@80^7]L00@0^WeN0;UlG@2iO5d0_7UO0P2nNV<0
102bKUT0/6mM0;1_G@2/LF4209i[FP0408aQD01aBST0LDXi05Y3>`80BD=700@0?4MI03Y:H@0jBV40
@4UJ0P11A5<01010BEX0?dmV03m?IP0oB5T20417F@040457E`14B5@0A4QD04E7D`80AdMA00@0BDQ@
04Y7D01:Ae00@TQD0P13BED01014BUH0ATaH04I<F018CUX204Y@G00304]AG@1<DUh0C59N00<0Be5M
1P1:D5`304]AG@0404eCG`1<DUh0C59N04Y@G080B4iJ00<0ATaH04E;E`15BeL00`15BeL<04A:EP@0
@TQD0`13BED504A:EPH0ATaH00@0AD]G04=9E@13BED0@TQD0P11Ae<304=:E0@0A4]E1@15C5H01014
BeD0ADaF04E<EP15C5H204I=E`0404E<EP14BeD0A4]E04=:E080@TUC00@0ATeG04E<EP15C5H0@dYD
1P12BE<204=:E0L0AdiH1@16CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>F014
CUP0@U1K0P15CeP104I?EP40BDiB0@3oool000@0BeQQ1@1:Ef0304UFG``0BUMP101;Ef<304YFHPP0
BEEQ1018E60504MCG``0BEMR1017EF0304QFH@D0BEMR00@0C5EL04]EGP1;EEh0BeEN0`1:Ef0204UG
HP<0BUQS00@0AeUS04QJI018FV@0C5eW0P1;F6401019Ded0CUMP04iGH019D5X203DlA00401XJ6P0G
5aL05aLG01@D50804a<A0`0D5A000`0C50h04Q<=018C3@0<01@C4@0401@A2P0`9ah0<2LN04Tl;080
DSdZ00@0DCLP05<i801C>B00G48Y0P1UBc@0101]BR`0Ldha07=><@1aC2l207=<;`0407]D=`1mE3P0
OE@h07aC=`80PUTm00@0QEPi08UL?@29G3d0SF110P2AHT@0102CI4H0V6M909QWB@2MK4h20:9_DP04
0:EgEP2TMEL0Y7EG0:=dEP80YGIJ00@0YgQL0:MhGP2WN5h0YGIL0P2RLU/0102YLeT0Z79H0:QbF02V
L5H20:A^E0040:9/DP2SKE<0XfeC0:E_E@80Yg5G00@0Wg9E09]^D@2KKU40UfQ<0P25ES`0101kBc@0
E28;05@R2`1`?BP208eJA@0408iL@@2JId`0VVM<0:E]E080YfiC00@0ZG1C0:igF@2^MeT0/7iM0P2`
Oeh0102^OEl0[W]N0:ikGP2]NU/20:mhF@040:mfF02_M5@0[gAD0:a_D080ZVe>00@0ZV]<0:AWB02T
IdP0WFE40P2JI480102MID@0XfI70:=VA`2XIdT20:]VB@040:=TA@2PHD80X65209aM?P80V5/l00@0
VE`m09aQ@@2LHD40XVM70P2TK4/40;J0IP<0]gmV0P2hOVH30;YnIP040;QnIP2iPFP0^H5X0;J0IP80
/7eR00@0]89W0;f>LP2mSW80`9=f0P2kSg801032VX00`IYm0<6JO@31VWd20<>JO0040<NKOP36Vg/0
aY]k0<BGN080`IAc00@0]8=R0;R6I@2hQVD0^hUX0P2lQV@0102gOeh0]GYH0;EjF02gNeT20;YnFP04
0;1fD02_MDl0[gE?0:ecC@80[79<00@0ZW1:0:U_B@2YKdT0Z6i80P2WKDL00`2[LD/0[79<0:abC003
0:abC0040:]aB`2ZL4X0ZW1:0:Q^B080Yfe700@0YVe00:A[?P2TJch0WfTk0P2KICL0102GHc@0Tf0a
09=P<@2BGc02095N;`04099M<`2GHSP0Uf8h09QS>@<0U60h0P2CHSX0102AH3P0Re/e08]K=@22BRT2
09=K>P040:E]C02^MUD0[WIE0;EjFP80^WmO00@0_85Q0;^0H02kP600_GYO0P31Nf40102fKeT0]W9M
0;IbG@2cM6820:I_F`0409MWD`1mDSl0OE8o05e5>`80CDA700@0?TMH03a;H00lBf00@4UJ0P11AE40
100nAeP0?4aS03a<H`0oB5T20417F@040417E`14B5@0A4QD04E7D`80AdMA0`19B5001012B5@0@dUE
04=9E@14BUH204I<F00404Q>FP1:D5`0BU1L04]AG@80C59N301;DEd0101=Del0C59N04aBGP19Ce/2
04M=F@0304I<F015BeL0AD]G00<0AD]G3014BUH30498E0@0@dUE1@14BUH604I<F00404E;E`14BUH0
A4YF0498E080@DMC0`13BU@404A;E@P0ADaF0`16CEL01015C5H0A4]E04A;E@13BU@20499D`0404E<
EP14BeD0A4]E04=:E0D0@TUC0`13BU@604M>F0H0ATeG0`18CE<01018C5D0ATeE04I=E@16CEL204E>
E`0404E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000404]HH@D0BUMP0`19EUl<04YG
H0<0BeMS0`1:EV8404UEH@80B5AP0`19EF4404QDH0D0Ae=O3019Ef8404MEH0<0B5IQ1@19Ef80101=
EUd0C5IO04aFG`1<EUl204YGH00304]HH@1:F6<0BUQS00<0BeUT0`17FF<0101:FfD0BUMP04YGH019
Ded2051IHP0404aCG@0i@4P0>D1801TI6@805QHF00@051@D01<C4@0C4a4051D@0P0E5Q400`0C50h0
4Q<=018C3@0<01DD4P0401@A2P0[8QT0:b8I04Pk:`80DC`Y00@0D3HO05@j8@1D>R40G48Y0P1VC3D0
101[B2X0LDh`075><01bCC0207=><@<0OEHi00@0O5<g089I?@22FCd0R5/l0P2:GCh0102;GSl0SUm1
08iO@@2@HD<209IUA`0409]ZC02OK4l0Wfa?0:5cDP<0XG9D0P2SM5P0102VMe/0YWMM0:IgG@2UMU`2
0:AdG@040:MdF@2YMU/0ZGIK0:QeFP80YG9G00@0X6eB09m/D@2OK540XVmD0P2ULUL0102PLED0WFiB
09e^DP2DIDT208=D>P0407U9<P1D90d0E2@=06hn:080RUY400@0SE]009QUBP2HIDX0Y6aC0P2YL5D0
102[LUD0/7UK0;1iF`2aOeh20;:1H0040:imG@2`OEh0/7eN0;1mGP80/7UJ00@0[GED0:aaD@2/LE40
Zfm=0P2ZKT`0102UJ4T0XFI60:5VAP2KHd8209US@@0409aT@`2PIDD0X6E50:IUA`80YfA700@0WUm0
09aM?P2LGCh0VE`m0P2JGCh0102NHd<0YV];0:I[B`2/M5<20;5iF0<0^HI[00@0^XAZ0;V3J@2iPfT0
^H5X0P2iOfL0102hOVH0]geU0;MmI@2lPVX20;^3JP040;R2J02fPfP0]X=X0;V7K080_Hib00@0`9=f
0;nCMP2oTgH0`i^10P31VWd01030VG`0`iYl0<>JO037Vgh20<JKN`040<FHN@35V7L0aIQg0;J5I080
^XQW00@0^hUX0;b6I02lQV@0]gmN0P2eNUP0102fNUP0]g]G0;MkE`2`MU020:idCP030:]aB`2YKdT0
ZFm900<0Z6i800<0ZFm90:Y`BP2ZL4X01P2/LT`0102[LD/0ZFm90:U_B@2XKTP20:M]A`040:I]@02T
Jch0Y6/n09iX>P80VfDg00@0UV8c09=P<@2CH340TUl`0P2AGRl0102AG380UV4g09IQ=`2GHSP309AP
>080Tf8j00@0TF0h08]K=@2;FcD0QDd/0P2GGch0102XL4l0/7QG0;1hE`2cNeX20;MoGP040;Z2H@2k
Pf80^h=R0;mjF`80`W]M00@0^G9F0;ifG@2nMUd0_GUR0P2aLe`0102UKeL0See408mMA01UB3X20555
A@040416E00oBU`0?dYL04=9E`80@dEA00<0?TIC03]:G`0kBUl00`0oB5T01010AeL0@TQD0498E014
B5<204M7D@0404Q8D019B500BDQ@0498E080@dUE00@0A4YF04I<F016C5P0B4iJ0P1:D5`00`1;DEd0
C59N04aBGP0<04]AG@0404aBGP1;DEd0Be5M04U?F`80AdeI00<0ATaH04E;E`15BeL00`15BeL<04A:
EP030498E013BED0@dUE00<0@dUE0`14BUH304E;E`<0ATaH00@0AdeI04I<F016C5P0AD]G0P14BUH0
0`12B5@0@DMC0457D`0304=:E0@0A4]E2015C5H404I=E`80ADaF00<0@dYD0499D`12BE<00`13BU@4
0499D`80@dYD00<0A4]E04E<EP15C5H03016CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0
ADeH04A>F014CUP0@4iI0P13CEH104A=E040Ada@0@3oool000@0BeQQ1@1:Ef0304UFG``0BUMP0`1;
Ef<304YFHP@0BEEQ0P18E60304UEH@@0B5AP1@17Del<04UGHP@0AeEP0`18EV4504UGHP0404eFG@1<
EUl0C5IO04aFG`80BUMP00<0BeQQ04YHH`1:F6<00`1;FF@304MIH`0404YKI@1:Ef00BUMP04UCG@80
D5UR00@0C5=M03U0B00i@4P06ATI0P0F5QH0100D51@04a<A01<C4@0D5A0201DF4@0301<D3P0B4`d0
4Q<=00`05A@B00@0514:02/R6@0[8QT0B3/[0P1A?2T0101@=Ql0E3XQ05@j8@1L@RT206I<=@0406]8
:P1aCS00LDh`079=<080Ldha0`1mEST0101lDcL0PUTm089I?@28Fc`208YM?P0408]N?`2>Gd40SUm1
091Q@`80UVE700@0VfY<09m/C`2OK4l0XG=B0`2QLU@20:=dF0040:IgF`2VMed0YWMM0:EfG080Y7AM
00@0YgAI0:UfF`2YMU/0Z7EJ0P2ULUL0102PKE80WfaA09m/D@2RKe@20:EbE`040:1aE@2MKU80WFiB
09AUB@80Pe@j00@0NDTb05@T3@1D90d0KShX0P2:FT@0102=Fd00V6E:09QUBP2TK5<20:U`E@040:]b
E@2`NE/0/7UK0;5oGP80/X5P00@0[WeM0;1mGP2`OEh0/7eN0P2`NEX0102]ME@0[75A0:aaD@2[Kdd2
0:Y^C0040:EXB@2QITH0XFI609]S@P80VF=100@0W6A30:1UA@2PIDD0YVE70P2WI4L0102NGd00W5dn
09aM?P2IG3d209YM?P0409iS@`2VJd/0YV];0:adD`80/GUH0`2iQV/0102jQ6X0^H=Y0;V3J@2iPFP2
0;UoI`040;QnIP2gOFD0]geU0;b2JP80^h=Z00@0^89X0;J3J02fPfP0^HM/0P2mSW801030TgH0_i=f
0;nCMP33Vh420<6JO@040<2IO033VW`0`iYl0<NKOP80aY]k00@0aIQi0<FHM`35V7L0]XET0P2jR6L0
102kRFP0_8IT0;b6I02gOeh20;EjF0040;IjF02gNeL0]g]G0;1fD080[WA>00<0Zg5;0:U_B@2YKdT0
0`2XKTP00`2YKdT0ZW1:0:Y`BP060:abC0040:]aB`2YKdT0ZFm90:Q^B080Yfe700@0YVe00:A[?P2T
Jch0WVPj0P2KICL0102FHS<0Tf0a09=P<@2BGc02095N;`04095L<P2FHCL0UV4g09MR>0<0U60h0P2C
HSX0102AH3P0Re/e08]K=@25CB`209MO?P040:Q`C`2`N5L0/7QG0;=kFP80]gmN00@0^X9Q0;^3HP2k
Pf80_gYK0P32Ned0102iLUH0_WIM0;ifG@2mNF820;5cG0040:E_E`2?GD@0See406E8>P80DDE500@0
@4ID03m:G00oBU`0@dUG0P13AE400`0nAU<0>dYO03]:G`0303m8F@040417E`12B5@0@TQD04A8D`80
AdMA00@0B4Q@04U8D019B500@TQD0P13BED01014BUH0ATaH04I<F018CUX204Y@G00304]AG@1<DUh0
C59N00`0Be5M00@0C59N04]AG@1;DEd0BDmK0P17CET00`16C5P0AD]G04E;E`0304E;E``0A4YF00<0
@TQD04=9E@13BED00`13BED304A:EP<0AD]G0`16C5P01017CET0ATaH04I<F015BeL204A:EP030498
E011Ae<0@DMC00<0@dYD1014BeD804E<EP@0ATeG0P15C5H00`13BU@0@TUC0499D`0304=:E0@0@TUC
0P13BU@00`14BeD0ADaF04E<EP0<04I=E`<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015CEP0
A4iH04A>F010CUT204==EP40A4eD0@17C5010?ooo`00101;F64504YGH0<0BEIO301:Ef0404YFHPD0
BEEQ0`18E60304UEH@@0B5AP1@17Del<04QFH@@0AeEP0`18EV4504UGHP0404eFG@1<EUl0C5IO04eG
H080BeQQ00@0C5UR04]II01;FF@0BeUT0P1<FVD01018FV@0AUQR04IHHP18FF<204UFG`0404UCG@1A
FV<0DEYS04iEG`80>T5900@06ATI01HF5P0F5QH04a<C0P0C4a40100D5A005AHA01DF4@0D5@l201<D
3P`05A@B00@0514:02DL4`0U71<0ASTY0P1A?2T0101?=Ah0EC/R05Dk8P1L@RT206M==P0406U8:@1a
CS00LDh`079?<@80ME0c00@0OEPk07mH>`1oF3/0O5Dh0P21FSd0102=H440S5m008aO@02<Gd0208eN
@00408mP@P2EI4H0UFA609YYB`80WV]>00@0WW1?09i_D@2NKe40Wg1B0P2QLUH0102TMET0YGIL0:Ef
G02VMed20:IfG`040:MdF@2[N5d0ZgQM0:aiGP80Z7EJ00@0XFiC09i[D02NJe00XFiC0P2ULUL0102Q
Ke@0WfeB09m]DP2CHDH2089C>@0407I6;`1F9Pl0ERH?06`n9`80QUQ100@0SUa109MTB@2GI4T0Xf]B
0P2ZLEH0102^MEP0/7UK0;1iF`2`OUd20;60G`040;60H02dPF80]85R0;F3HP80]GmM00@0/7QG0:eb
D02]LU00[G5?0P2^LU00102TIdP0XVM70:9WA`2LIT@209YU@`0409aVA02QITH0XFI60:IWB080ZFI9
102PHdH20:=XBP040:Y_D@2aN5X0/GQJ0;Z1H`80_XMX102kR6d20;b6K0040;^5J`2jPVT0^X9Y0;Z0
J080^GmW00@0_XA/0;Z2J@2jPVT0^H=Y0P2iQV/0102kRFh0_8ea0;b=L@2nTG@20<6EN0040<2HOP2l
UGP0_9Eh0;bEN080_iIh00@0`IEh0;jCL`2nTg<0_Y5b0P30Tg80102hQfH0^hUX0;^9J02lRVT20;b6
I0040;MoGP2eNUP0]GYH0;EiE`80]GUE00@0[gE?0:ecC@2]Ldd0ZW1:0P2XKTP0102WKDL0Z6i80:Q^
B02ZL4X20:]aB`@0[G==0P2[LD/0102ZL4X0ZFm90:U_B@2XKTP20:M]A`040:E/?`2SJSd0XfXm09iX
>P80VV@f00@0UV8c09=P<@2CH340TUl`0P2AGRl0102@Fc40UF0f09EP=P2GHSP309=O=`80Tf8j00@0
TF0h08]K=@2;FcD0QU0^0P2KID<00`2^N5H0/G]I0;5kF@030:agE@040;9mF`2gPV00]h9P0;ahE@80
`7UI00@0_7=C0<=iG033NE`0`gaP0P2hMEX0102aM5L0WfI909mVB@1/BcP205Q7@00404=6C`11BEH0
@DUF04E9E080ATI>00@0@4IB03e:F`0mBU/0?DUI0P0oB5T0100oB5L0@TQD0498E014B5<204E8D@<0
B4Q@00@0@TQD04=9E@13BED0A4YF0P16C5P01018CUX0BU1L04Y@G01;DEd204aBGP<0Be5M1P1<DUh4
04]AG@80BU1L00<0B4iJ04I<F016C5P01P15BeL<04A:EP@0@dUE0`14BUH504E;E`H0AdeI00@0ATaH
04A:EP14BUH0@dUE0P12B5@304=:E0@0A4]E1P15C5H304I=E`80AdiH00@0ATeG04E<EP15C5H0A4]E
0`13BU@60499D`80@dYD00<0ADaF04I=E`16CEL01015C5H604I=E`80AdiH0`18CE<01018C5D0ATeE
04I=E@16CEL204E>E`0404E=F014CUP0A4iH03m=F080@TaE0@13C5<104I;C`40oooo000404]HH@D0
BUMP0`19EUl<04YGH0<0BUIR1019EF4504QDH0<0BEEQ1018E60504MCG``0B5IQ1017EF0304QFH@D0
BEMR00@0CEIM04aFG`1<EUl0CEMP0P1;F640101<FF80BeUT04]II01;FF@204aJI@0404QJI016F680
AUQR04QIH`80B5EN00@0BUAN055JH`1AFV<0C5=M0P0h?dL0100H61P05ADE01DE5@0C4a<201<C4@04
01@E400E5Q405AHA01DF40804a@>300F5A<0100E4P/07aH=01lF3@15>2P2058m:P0404le7P1G?B@0
EcdT05i4:`80JDlh00@0JdX[079A<P1bDC80LU4b0P1eDS@0101nFC`0OEPk07eH>`1lECP2085J?@04
095TA@2?HT<0Sf9308iQ@P80Sf1200@0TV=509MVB02GITP0VfY<0P2NJdh0102KKD`0W6e?09a]C`2N
Ke420:1aE@040:5bEP2TME/0Y7EK0:IgG@80YgMP00@0YgEJ0:]iGP2[NEh0[G]P0P2YMe`0102RL5D0
WVaA09i/D@2OKE820:=aEP040:1]DP2OK540WfaA091N@`80Pe4h00@0M4@]05PZ4P1H:Q80JC`U0P20
Dc`0102>G440UF9709ERA`2RJU420:]bE`040;1gFP2`NE/0/7UK0:ilF`80[giM00@0/H1O0;B2H@2d
PV40]H=P0P2eOed0102`N5D0[G9@0:ebD02]LDd20:ibCP040:EZBP2SJdX0Xf]:0:1[B@<0WfY80P2S
JdX00`2ZJd`0[Fa>0:e/CP030:MZCP040:M[C`2[Ke<0ZfmC0:mfF@80]WmQ00@0_XMY0<><KP33S6h0
^8I[0P2iQf`0102kR6d0_8U^0;b9KP2nR6h20;f7K@040;f5K02lQ6/0_8A[0;b2JP80^X9Y00@0^H=Y
0;Z7K02jQf`0^XQ]0P2jRfl0102lSg80`9Ag0<2DM`2jTWP20;N@L`040;RAM02nUGL0_YEg0<2DM`80
^i1`00@0_8m`0<6DL`31U7<0^HQW0`2kRFP20;^5H`040;R0G`2eNUP0]GYH0;AhEP80]7QD00@0/7I@
0:idCP2^M4h0Zg5;0P2XKTP0102WKDL0Z6i80:Q^B02ZL4X20:abC0040:meC`2^M4h0[WA>0:ecC@80
Zg5;00<0ZFm90:Q^B02XKTP00`2WKDL0102SJSd0XVTl0:9Y?02MIcT209US=@0409IR<`2CH340Tf0a
099O<080TEh_00@0T5/a09EP=P2EH3H0Uf8h0`2CGcL2099Q>@04091O=`2:FS@0RUXd08A>;080WFM5
0`2bOE/0102ULe00XG5=0:5aC@2VN5@20:j0G0040;EeD@2jN5D0^WQE0;QaD@80`gYJ00@0aG]N0;]d
EP2kM5H0^7IF0P2/KD`0101hDcT0H4Xo061:?`17ADX204=7DP0404Q8D019B4d0BDQ=04A8D`80?daL
00@0?4UI03e9F@0mBET0?dQG0P11BE@01012BE<0A4QA04A8D@16BE0204Q8D0040498E013BED0@dUE
04A:EP80ATaH00@0B4iJ04Y@G01:D5`0Be5M0P1<DUh304]AG@H0CE=O0`1;DEd0101:D5`0BDmK04U?
F`17CET204I<F0@0AD]G0P16C5P<04A:EP<0@dUE0`14BUH304E;E`<0ATaH0`17CET01018CUX0AdeI
04M=F@16C5P204E;E`0304=9E@12B5@0@TQD00<0@dYD1014BeD504E<EP<0ATeG1017CUP204I=E`04
04A;E@13BU@0@dYD04=:E0H0@TUC0P13BU@00`15C5H0ATeG04I=E`0404A;E@80ADaF0`16CEL304M>
F0<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015CEP0A4iH04A>F00oCEP2049<E@40@daC0@16
Bdl10?ooo`00101;F64504YGH0<0BEIO301:Ef0304YFHP@0BEEQ1@18E60304UEH@@0B5AP1@17Del<
04QFH@@0AeEP0`18EV4504UGHP0404eFG@1<EUl0C5IO04eGH080BeQQ00@0C5UR04]II01;FF@0BeUT
0P1<FVD01018FV@0AUQR04IHHP18FF<204QEGP0404YDGP1AFV<0DEYS04aCG@80>3m700@061PH01DE
5@0E5AD04a<C0P0C4a40100D5A005AHA01DF4@0E5Q0201<D3P`05QDC00@05A8;01lF3@0O5Pd0ACPX
0P1B?BX0101?=Ah0EcdT05Lm901NA2/206U?>00406]::`1bDC80LU4b079A<P80ME8d00@0OUTl07eH
>`1mF3/0O5Dh0P21FSd0102AI4D0Sf9308mR@`2>HD8208mP@P04099SA@2GITP0UfI809]ZC080WV]>
00@0Vfe<09a]C`2LKDl0WVmA0P2PLED0102QLUH0Y7EK0:AeF`2VMed20:MgH0040:MeFP2[NEh0ZgUN
0:ekH080ZGML00@0XW1E09i/D@2NK540WfeB0P2SLEH0102PKE80WfaA09m/D@2@GT<208=A>00407A4
;@1H:Q80F2XB06Tl9@80P5<l00@0SUa109ERA`2EHTL0XVYA0P2[LUL0102`MeX0/7UK0;1iF`2^O5/2
0:mnG@040;60G`2dPV40]89Q0;F3H080]GmM00@0/7QE0:ebD02]LU00[G5=0P2^LTh0102UJTX0Xf]:
0:=[BP2PJdT309mZB080Xf]:00<0ZV]<0:e/CP2]K4h00`2WJTh0102WJdl0ZfmC0:]_D`2_MUT20;Io
H@040;j7J@33S6h0`ha^0;R6J`80^HM/00@0^hQ]0;b9KP2lRFh0_XQ^0P2mQfd0102mQF`0_8A[0;b4
J`2lPVX20;Z2J@040;V3J@2jQf`0^XM/0;Z8K@80^X]_00@0_8mb0<2DM`30U7L0^Y9h0P2gT7<0102h
TG@0_YEg0;jEM`30U7L20;^@L0040;b?L031U7<0`IAc0;V8I`<0^hUX0P2kQF<0102hP5l0]GYH0;Ej
F02dN5H20;AhE0040;1fD02^M4h0[WA>0:]aB`80Z6i800@0Yfe70:Q^B02XKTP0ZW1:0P2/LT`0102_
MDl0[WA>0:idCP2]Ldd20:]aB`030:U_B@2XKTP0Z6i800<0Yfe700@0XfXm0:9Y?02RJC`0WFLi0P2I
HcD0102FHS<0Tf0a09=P<@2BGc02095N;`04091K<@2EH3H0UF0f09MR>0<0Telg0P2BHCT0102@GcL0
RUXd08YJ=024CR`209eWA@<0/WeK00@0YG=@0:5aC@2QLDd0YWQD0P2^P5`0102eME40^WQE0;YhE@2h
LE420<=jFP040<EkGP2kM5H0^gAF0;QfEP80[6e<00@0N5<i061:?`1PBSl0AdE:0P13Ae801018B500
BDQ=04U8C@14B5<203m<G00403a9F@0mBET0?DUI03m8E`80@DUD00@0@TUC04A8D@14B540ATU@0P18
B5001012B5@0@dUE04=9E@14BUH204I<F00404Q>FP1:D5`0BU1L04]AG@80C59N0`1;DEd604eCG`<0
Be5M00@0BU1L04U?F`19Ce/0AdeI0P16C5P404E;E`80ATaH3014BUH304=9E@<0A4YF0`15BeL304I<
F0<0AdeI00@0B4iJ04M=F@17CET0ATaH0P15BeL00`13BED0@TQD0498E00304=:E0@0A4]E1@15C5H3
04I=E`@0AdiH0P16CEL01014BeD0@dYD04=:E013BU@60499D`80@dYD00<0ADaF04I=E`16CEL01014
BeD204E<EP<0ATeG0`17CUP304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>F014CUP0
?deH0P12C5D104=<D`40AT]?0@3oool000@0BeQQ1@1:Ef0304UFG``0BUMP00<0BUIR04UEH@19EF40
0`19EF4304QDH0<0Ae=O0`19EF4404QDH0D0Ae=O4017EF0304QFH@D0BEMR00@0CEIM04aFG`1<EUl0
C5IO0P1:Ef000`1;F640BUQS04YHH`0304]II00404QJI016F680AUQR04QIH`80BEIO00@0BUAN051I
HP1@FF80BE1J0P0b>D40100G5aL05ADE01DE5@0B4Q82018B400401@E400F5a805QLB01DF408051D?
300F5A<0100G50d06a8901/B2@14=bL205<n:`0404le7P1J@2L0FT0W05m5;080Je4j00@0KDh_07=D
=@1cE3D0Le8c0P1dDc@0101lFC/0OEPk07eH>`1jECP2081K?P04095TA@2@Hd@0T6=408mR@`80TF94
00@0U6E709UXBP2IJ4X0VfY<0P2MJTd0102IJdX0Vfa>09]/CP2MKU0209i_D`0409m`E02QLUP0XG9H
0:EfG080Z7QQ00@0YgEJ0:YhG@2ZN5d0[G]P0P2[NEh0102ULeP0XFmD0:5_E02PKU<20:5_E00409iX
CP2PJU00X6Y@08iK@080Q58i00@0LdD]05X^5@1J;QD0IC/S0P1hCSH0102;FCh0TEi3095N@`2NITd2
0:]bE`040;9iG02aNU`0/GYL0:ilF`<0[WeL0`2`OU/20;1jEP040:eeDP2/LTd0[79=0:]_B`80ZVi9
00@0YFe<0:E_C@2UKdd0Y6m=0P2RL4d0102SKT`0YVi=0:I^C@2/Ke020;9aD`@0ZViB0P2ZLE@0102^
MeT0]7eO0;AmG`2jQFH20;j9JP040;F6JP2iQf`0^HM/0;^9KP80_HY_00@0_X]`0;n9K`2oRFl0_XQ^
0P2nQVd0102jP6P0^h=Z0;^3JP2lQV`20;V6J`040;R6J`2jRfl0^X]_0;b?LP80_I5d00@0]hme0;B=
L02dSG00^95d0P32VG/01034V7/0_Y=c0;jCL`2oTW<20<JIN0040;Z9J02kRFP0^hUX0;Z8I`80^hES
00@0^81O0;IkF@2fNeT0]GUG0P2cMe<0102aMe40[gE?0:meC`2]Ldd20:]aB`<0ZFm900@0ZW1:0:]a
B`2[LD/0/7I@0P2_MDl0102]Ldd0Zg5;0:]aB`2YKdT20:Q^B0<0Yfe700@0XVTl0:5X>`2QJ3/0W6Hh
0P2IHcD00`2EHC80Tf0a09=P<@03099O<004099M<`2FHCL0UV4g09MR>080Telg00@0TUhf095P>02A
H3P0SUde0P28F3801021BbT0VVA209YT@P2`NeT20;1nF`040:AfDP2PM4l0X7A?0:AjE080ZX1J00@0
/7EC0;MhE`2gN5L0]G=C0P32Ned01035Neh0^W=C0;YcD`2jME@20;9`CP0408UM@01ZD440JU1104a6
AP80ADI;00@0B4M<04Y8B`1:B4/0ATU@0P0oBeT303a9F@0403e9E`10BU@0@4YD0499D`80A4QA0`16
BE001012B5@0@dUE04=9E@14BUH204I<F00404Q>FP1:D5`0BU1L04]AG@80C59N00@0BU1L04aBGP1<
DUh0CE=O0`1>E60204eCG`0404aBGP1:D5`0BU1L04U?F`80B4iJ00<0ATaH04E;E`15BeL01015BeL2
04I<F100A4YF0`15BeL504I<F00304M=F@18CUX0B4iJ00<0B4iJ00@0AdeI04E;E`15BeL0A4YF0P13
BED304=:E0@0A4]E1@15C5H01016CEL0AdiH04M>F017CUP204Q?F@0404M>F016CEL0ATeG04E<EP80
A4]E0`13BU@40499D`80@dYD00<0A4]E04E<EP15C5H00`13BU@304A;E@0304E<EP16CEL0ATeG00<0
AdiH0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH041>F@80@deF0@14CE@1
04M<D040oooo000404]HH@D0BUMP0`19EUl<04YGH0@0BEEQ0`18E60504MCG`<0BEEQ1018E60504MC
Ga00AeEP0`18EV4504UGHP0404aEG01;EEh0BeEN04]EGP<0BUMP0P19Ef8304YHH`0404UKI@16F680
AUQR04UJI080BUMP00@0BeEO04mHH@1?F640A4]E0P0[<SX0100G5aL051@D01@D500B4Q82018B4004
01@E400F5a805QLB01HG4@8051D?300G5Q@0100H5@h060l601P?1P14=bL205@o;004050f7`1M@bX0
GD<Z0657;P80KE<l00@0LE8c07EF=`1eESL0LU<d0P1cDS<307]H>P0407YE>020Fch0P5/n08mR@`<0
SF110P2@HD<0102EITP0VFQ:09UXBP2JJD/209YWBP0409QZB@2JJdd0VV]=09a]C`<0WFiB0P2OL5H0
102TME/0YgMP0:MgH02VMe/20:QiG@<0ZGYN00@0YWMK0:=dF02SM5P0X75E0P2NKe<0102LI4/0X6Q?
0:1XC`2<FCh208ID>`0407=5;@1M<AP0GC4H06<i8@80Ld/b00@0QU@i08]H?@2;F3d0VV990P2ZLEH0
102cNUd0/W]M0;9kG@2_OE`20;60G`040:inFP2^O5T0[WaI0:ekEP80[gUE00@0[gMB0:meD02_ME00
[71;0P2ZKTT0102VKTd0YW1>0:I`CP2ULe020:AbC`040:=aCP2WLDl0Yg5?0:iaDP80/gAE00@0[W9H
0:eaE`2]LEL0Zg9G0P2/LeP0102^MU/0/G]O0;5kG`2fP6@20;V3I`040;J7J`2gR6`0]hQ/0;b:K`80
_H]`00@0_haa0;j;L02nRg00_XQ^0P2mQfd0102hOVH0_hM^0;n7KP30RW020;V6J`040;N5JP2mSW80
_Hib0;nBM@80_91c00@0^Y9h0;N@L`2gT7<0_9Eh0P36WGl01036VWd0_95a0;bAL@2kSVl20<6DL`04
0;Z9J02jR6L0^XQW0;V7IP80^XAR00@0^H5P0;MlFP2gO5X0]GUG0P2dN5@0102bN580/GMA0;5gD@2`
MU020:idCP040:abC02[LD/0Zg5;0:Y`BP80ZFm900@0/GMA0:meC`2_MDl0[G==0P2ZL4X00`2XKTP0
Yfe70:M]A`030:M]A`040:5X>`2PIcX0X6Lj09]U=`80V68d00<0UF4b09=P<@2CH3400`2BGc00102C
GS@0Uf8h09MR>02HHcT209=O=`04095M=@2@GcL0T5lg08aK<`80QUH`00@0OTPV099L>P2BG3X0Y79?
0P2]OET0102/P5/0[H=M0:f3G@2/QEh20:V3G0040:agE@2bNUT0/WYI0;EfE`80`GiQ00@0a7eO0;Ub
E02iLU@0^gIG0P2fM5@0102IJdT0MUY507IJA@1CBDL204E5A`0404Q6A`1:ATD0BTI504A7C080?dUE
00@0>TYI03a9F@0lBET0?DUG0P10BU@01011BU<0@TUA0499D@15BU0204I9D0040498E013BED0@dUE
04A:EP80ATaH00@0B4iJ04Y@G01:D5`0Be5M0P1<DUh0101:D5`0C59N04aBGP1>E60304mEH@80CUAP
00@0C59N04Y@G01:D5`0B4iJ0P17CET01016C5P0AD]G04E;E`14BUH304E;E`80ATaH3`14BUH304E;
E`H0ATaH1P18CUX01017CET0ATaH04I<F014BUH204=9E@<0@dYD1014BeD504E<EP<0AdiH0`18CeT0
1017CUP0ATeG04I=E`15C5H204A;E@0404E<EP14BeD0A4]E04=:E0D0@TUC0`13BU@30499D`0404=:
E014BeD0A4]E04E<EP80ATeG0`17CUP304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>
F014CUP0@DmJ0P14CUL104E>E@40B4eA0@3oool000@0BeQQ1@1:Ef0304UFG``0BUMP1019EF4304QD
H0D0Ae=O0`19EF4404QDH0D0Ae=O4017EF0304QFH@D0BEMR00@0C5EL04]EGP1;EEh0BeEN0`1:Ef02
04UGHP<0BUQS00@0BE]U04IHHP16F680BEYT0P1:Ef00101;EEl0CeQQ04mHH@14BeD202/b>P0401LG
5`0D51@051@D018B4P804Q8@00@051D@01HG4P0F5a805QLA0P0D5@l<01LF500401PE3P0H3`H060l6
04@g9`80E3l/00@0D3HO05e3:P1M@bX0HDL^0P1]Dc`0101aDS<0MEHg07EF=`1bDc@207=B<`<0NePj
00@0NUDh081K?P20Fch0Sf930`2=H442091Q@`0409EVB02IJ4X0VFQ:09YYB`80VVM:00@0V6Y909Y[
C@2JJdd0W6e?0`2MKU8209m`EP040:AeF`2WMf00YgMP0:IgF`80Z7UM0`2YNUh0102VMe/0XgAH0:=d
F02PLED209i_D`0409aTB`2PJ4l0X6Q?08aI?P80QU@k00@0LdD]05da601M<AP0HcTQ0P1cBc801026
E3T0RePm08]H?@2JHTT20:YaEP040;=jG@2bNed0/W]M0:mmG080/H1O00@0[WiJ0:ilF@2^O5T0[G]F
0P2_NED0102_Me80[gE@0:meD02/L4/20:Y^B@040:I^C@2VL4h0YW1>0:EcD080Y79?00@0Xg5>0:Ma
C`2WLDl0[W5B0P2cM5D0102^LUP0[G5G0:eaE`2[LUL20:acF0040:ifF`2aNel0/G]O0;J0I080^H=W
00@0]XM[0;N8K02gR6`0_8Y_0P2mRg00102oS740_X]`0;j;L02nR6h20;f7K@040;QnIP2oQfh0_hM^
0<2:L080^HI[00@0]hEZ0;f>LP2mSW80_i9e0P2lT7<0102jTWP0]i1c0;N@L`2lUGP20<JMO`040<JJ
O@2lTG40_95a0;^>K`80`IAc00@0^XUX0;Z8I`2jR6L0^HMV0P2jQ680102iPF00]gaJ0;MlFP2eNEL2
0;AhE0040;9hDP2aMe40/GMA0;1fD080[WA>00@0[79<0:]aB`2[LD/0ZW1:0P2YKdT0102aMe40[gE?
0:meC`2]Ldd20:Y`BP030:Q^B02WKDL0Yfe700<0Yfe700@0XFPk0:1W>P2PIcX0VfDg0P2HHS@00`2E
HC80Tf0a09=P<@03099O<00409=N=02GHSP0Uf8h09QS>@80Telg00@0TEde091O=`2@GcL0S5/c0P26
ES00101nB2H0TU`j099L>P2TLTl20:emF@040:b0F`2]Ped0[H=M0:b5GP80ZH=L00@0[7ME0;9jF@2b
NUT0]GIG0P31OV401034OEl0^G9D0;UbE02kMUL20;IdE00409U[B@1fFTD0MUY505=9A`80ADE700@0
B4I704Y6A@1:ATD0A4M<0P0oBED0100jBUT0?4UI03a9F@0mBEL2041:E004045:D`12BE40@TUA04E:
D080ATU@00@0@TQD04=9E@13BED0A4YF0P16C5P01018CUX0BU1L04Y@G01;DEd204aBGP0404Y@G01<
DUh0C59N04iDH0<0CeEQ0P1>E600101<DUh0BU1L04Y@G018CUX204M=F@0404I<F015BeL0AD]G04A:
EP<0AD]G0P16C5P?04A:EP<0AD]G1P16C5P604Q>FP0404M=F@16C5P0ATaH04A:EP80@dUE0`13BU@4
04A;E@D0ADaF0`17CUP304Q?F@0404M>F016CEL0ATeG04E<EP80A4]E00@0ADaF04A;E@14BeD0@dYD
1@12BE<304=:E0<0@TUC00@0@dYD04A;E@14BeD0ADaF0P16CEL304M>F0<0B4eC00@0B4aE04I=E@16
CED0ATeG0P15CUL01015CEP0A4iH04A>F011CeX204A>E`40ADiE0@18CE410?ooo`00101;F64504YG
H0<0BEIO301:Ef0304UEH@@0B5AP1@17Del304UEH@@0B5AP1@17Del@04MEH0<0B5IQ1@19Ef80101;
E5/0BUAM04YDG@1:E5d304UFG`D0BEMR00@0BE]U04MIH`17FF<0BEYT0P1;F640101;EEl0CUMP04iG
H011B58202H]=@0401LG5`0D51@051@D018B4P804Q8@00@051D@01HG4P0F5a805QLA0P0E5Q0<01LF
500401TF3`0F3@@05Pd404Dh:080ED0]00@0DCLP05i4:`1NA2/0HTP_0P1_ECh0101cE3D0MePi07MH
>@1cE3D207=B<`0407UH>@1jEcT0NULi07UF>080P5/n00@0S5m008]N?`2;GSl0Reho0P2?H480102D
IDL0VFQ:09UXBP2HIdT209MTA`0409QZB@2JJdd0VV]=09a]C`<0W6eA0P2NKeD0102SM5X0YgMP0:Mg
H02VMe/20:MhG0040:QiG@2YNUh0ZGYN0:QiG@80YGIJ00@0XG9F09i_D`2NKe<0Vf590P2OIdh0102>
F3h0QeDl08ME?01dARh205lc6P04064i801`B2l0L4P_085?=080QU<h00@0UUi50:U`E@2YL5D0]7]N
0P2dOEl0102aOeh0/h9Q0;>2H@2bPUh20;5oG0040:ilE`2aNeL0/G]G0;=kEP80]7YD00@0/GE@0:ea
C02]LD`0YFe<0P2VL4h0102ULe00Y7A@0:AdD02TLTl20:MaC`040:ebDP2dMEH0]7EF0;9fG080[gIK
00@0[WEJ0:eeFP2]MEX0/7QM0P2cOF40102gPFD0^H=W0;V3I`2gR6`20;R9K@040;Z;K`2nS740_Xaa
0;n<L@80_HY_00@0_HM]0;^5J`2kQF/0^7iV0P32RW401034SW@0^HI[0;V6J`2gQFX20;n@M0040<:E
N02kSg80^hmb0;nGO@80^iAg00@0`9Ul0<ROP@38Wh40a9Qk0P2eRVX0102aQ6D0]XUX0;J9J02iR6L2
0;V7IP040;R6I@2jQ680^XAR0;V1H080^7eK00@0]WYH0;AhE02dN5@0/gUC0P2bN580102aMe40/7I@
0;1fD02^M4h20:]aB`040:U_B@2XKTP0Z6i80;5gD@80/7I@00@0[G==0:Y`BP2ZL4X0Z6i81@2WKDL0
102QJ3/0WfHi09mV>@2KICL209QR=00309EQ<P2CH340Tf0a00<0TUl`00@0U5le09QS>@2HHcT0V6<i
0P2CGcL0102AGCD0Sehf08mN=P2;FS8208EE;`0407a6902;EC<0ReDc09]YAP80ZWYF00@0]HUT0;VB
J`2iTV/0]8iW0P2ZQel0102[NUX0/WeN0;9mGP2cN5X20<:1I@040<AoHP2hLeH0^7=F0;YhF080^7IF
00@0Y7A@07mPB`1oH4/0Eda60P16A4D01018A4<0BD=304U3@`13A4T203a6D00403Y:F@0lBET0?4UI
03e9E`80@4YD00<0@DYC0499D@12BE400`15BU001012B5@0@dUE04=9E@14BUH204I<F00404Q>FP1:
D5`0BU1L04]AG@80C59N00@0BU1L04aBGP1<DUh0CUAP0`1@EV8204iDH00304aBGP1:D5`0BU1L00<0
AdeI0`15BeL01014BUH0AD]G04E;E`15BeL204I<F0l0A4YF0`15BeL404I<F080AdeI0`18CUX01019
Ce/0B4iJ04Q>FP17CET204I<F00304A:EP13BED0@dUE00<0@dYD1014BeD504E<EP<0AdiH1018CeT2
04M>F00404E<EP14BeD0A4]E04I=E`80ADaF00<0@dYD0499D`12BE<01012BE<204=:E0040458DP12
BE<0@TUC04=:E080A4]E00<0ADaF04I=E`16CEL00`17CUP304Q=D`0404Q<E@16CED0ATeE04I=E`80
ADiG00@0ADeH04A>F014CUP0@U1K0P15CeP104I?EP40BDiB0@3oool00100BUMP0`19EUlA04QEGP03
04UFG`18EEh0B5EN00H0AeAM00<0B5EN04UFG`19EUl01019EUl204QEGP<0AeAM0`16De`404QEGP<0
BEIO1@1:Ef0304EBF`0404ICG017E5d0AeAM04QEGP80BEIO00@0BUMP04]HH@1;F640BE]W0P18FV@0
1016Eel0BEQO04UHG`1@Ff4204eFF`0303U1A00R:Rd08RX]00<04a8@0`0D4a4601DD4P`05aHA00@0
5Q<:01dG2`0M5`/0ACPX0P1IAc40101G@b/0G4H^05a6;P1PC3@206]G?`04079C=`1aDSH0LE8f075B
=P80MeDj00@0O5Pn07eI?`1mFCl0NeHl0P1hDcT01022F3l0Q5/o08AK?`28G44208]O@P04091SAP2D
IDL0U6E709IWB@80UfU800@0UVA=09MUCP2GIDh0UfM@0P2JJU<0102LKUH0X79J0:1bFP2SMeh20:Ah
G`040:5hFP2VOEl0YWeO0:b0H`80ZWiQ00@0Y7MJ0:1aE@2PLED0XG9F0P2SM5P0102SKE<0UEm509EO
A@2;FCh208QI?`0407U==01R>R00HSXP05dh7P80J4@Z00@0O4Xa08IC>026DcP0UEho0P2QJDH0102[
Le00/geI0;=mF@2gQF@20;Z9J@040;:2GP2[NEH0ZgUF0:QfD`80/G]I00<0]GeL0;AiF@2dNET00`2`
Le@0102[Le80ZgEC0:]eD`2YMe@20:EgE@040:AfE@2WM5D0YgAE0:adF@80[gIK00@0]WIJ0;EhF`2e
N5/0]7QL0P2`MeX0102^MU/0/G]Q0;5kH@2fPfP20;^9KP040;n9K`31Rg40`H]a0<:<LP80_hU_00@0
^XAZ0;R2J02hPVP0^H=Y0P2lQV`0102hP6D0^X9W0;Z2I`2lQVX20;^8J`040;Z9J`2iRV`0^HY/0;b?
L080_Y5b00@0aiEj0<FFNP35UWX0ai]n0P2nUgH0102lUG@0^8e[0;R=J`2`PV020<6@K`040;F7I@2l
RfX0_8]Z0;^9J080]7mM00@0/WaJ0;InF`2fOU/0]W]I0P2`MU40102gO5X0[gAB0:mdDP2/LDl20:md
DP040;5fE02]LU00[G9@0:]`CP80[W=A00@0]G]E0:Q^B02XKTP0YVa60P2YKdT0102UJdD0Y6Y40:AZ
A02WKDL20:E[A@0409m/@@2KJ3d0VfPm09IS>0<0U64f0P2EHSL0102DHCH0TUld099O=02CGcL209IR
>P0409EQ>@2@G3@0T5`d08mK<`80TEde00@0T5`d08aH<02<F300PU8^0P1cAB401023Ec80WG==09ec
C@2WP5T20;B>I`030;VFKP2]RV80[HYR00<0ZWEC00@0/G]I0;^3HP2kPf80`XEV0P2mOUl0102gMUP0
]GAF0;EdEP2JNel208a`F00406mIA01B@SD0DT8e04Hm>080AdE800@0A4M>03Y1B`0j@D/0>D]E0P0l
BUD0100oBED0@TQD0498E014B5@20499D`03045:D`11C580@DaB00@0ADaF0P16CEL01017CUP0BE1J
04U@FP1;DU`204eDGP0404aBGP1=Del0CE=O04eCG`<0CUAP0`1=Del204aBGP0404Y@G019Ce/0BDmK
04M=F@80AD]G00<0A4YF04=9E@13BED01P13BED404A:EPD0AD]G0`14BUH304E;E`<0ATaH0`17CET0
0`18CUX0BDmK04U?F`0404U?F`80B4iJ0`16C5P304A;E@<0ADaF0`16CEL304M>F0@0ATeG1@17CUP3
04Q?F@0404Y>F@19CEP0BDeH04M;EP80ATYE00<0ADUD04A8D`14B5<00`15BE@01013Ae80A4QC04A8
D`15BE@204I:E@0304M;EP18C5L0B4aG00<0BDeH0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`04
04E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000@04YGH0<0BEIO4@18EEh00`19EUl0
B5EN04QEGP0604MDG@0304QEGP19EUl0BEIO00@0BEIO0P18EEh304MDG@<0AU=L1018EEh304UFG`D0
BUMP0`15DU/01016De`0AeAM04MDG@18EEh204UFG`0404YGH01;F640BeQQ04UKI`80B5YT00@0AUMO
04UHG`19F5l0D5]Q0P1=EU/00`0i@D@08RX]028Z;@0301<B40<051<A1P0E518<01LF4@0401HC2P0M
5`/07AL;04Dh:080FDLa00@0Ed<[05a6;P1LARh0H4`d0P1[Ecl0101bDcL0LE8f075B=P1aDSH207ME
>P0407aH?P1mFCl0OETo07]F?080N5<i00@0PUPo08AK?`24Fcl0R5a10P2;Gd80102@HdH0U6E709AU
A`2FIdT209MYB00409ITC@2GIDh0UfE>09MWD080VVYC00@0W6iF0:1bFP2PLUX0XgMN0P2TN5l0102Q
N5X0YWeO0:ImG`2/P6<20:YnH@040:AgFP2PLED0X75E0:5bEP80XgAH00@0XfeC09EOA@2EGdD0ReTn
0P28FCl0101iCC@0HSXP068j801M>1h206Q4:P0407a:<@26DcP0QU<h09EN?`80XFU600@0Zg=@0;=m
F@2cOET0]hET0P2jRFT0102bPUh0ZgUF0:]iEP2XMU<20;5kF@030;EmG02dNET0]7UI00<0/7=D00@0
Zg=B0:]eD`2[ME<0ZGMD0P2UMeD0102TMUD0YgAE0:MdE@2/M5T20:mfF`040;IfFP2eN5/0]GQK0;Ah
G080/7MJ00@0[WIK0;5kH@2aNf40]X=X0P2kRFh0102oRFl0`H]a0<6;L@32S7820;n9K`040;Z4JP2h
PVP0^89X0;V3J@80_8I/00@0^81U0;Z2I`2jPVL0_8IZ0P2kR6/0102jRF/0^HY/0;V:K02lSg020;jA
LP040<NENP35UWX0aIIj0<NKOP80_YMf00@0_9Ed0;R=J`2hSF/0/89P0P31T6l0102eQfD0_8]Z0;b;
JP2kRFP20;AoG@040;9lFP2fOU/0]WiK0;IkF@80/7IA00@0]gaJ0:mdDP2_M580[75?0P2_M580102a
MU@0[G9@0:ebD02[L4h20:icD@040;EkE@2XKTP0Z6i80:I/AP80ZFm900@0YF]50:AZA02TJT@0Yfe7
0P2UJdD0102OK440VfPm09]X?@2FHcP309AQ=P80UF8g00@0U64f099O=02BGc@0Telg0P2FHSX0102E
HCT0T5`d091L=02?Fc<2095M=@04091L=02<F300S5P`089B;P80LdDQ00@0PeLb09ecC@2MLdd0Yh1I
0P2dSVL00`2iUVh0[HYR0:f:HP030:YeD`040;5kF@2kPf80^h=R0<:5IP80_GiO00@0]gIH0;EdEP2e
M5H0VW]O0P2<L5P0101_FD@0DT8e0592=@16?CP204M5B00404A7CP0j@D/0>T5;03U;E@80?4YE00@0
?dUE0498E012B5@0A4QD0P12BE<00`11BU<0@DaB045<DP0404E<EP80ATeG00@0AdiH04U@FP19D5X0
Be9L0P1=E5h0101<DUh0CE=O04eCG`1=Del304iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17CET2
04E;E`0304A:EP13BED0@dUE00H0@dUE1014BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0B4iJ
04U?F`19Ce/01019Ce/204Q>FP<0ATaH0`14BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18CeT0
101:CUT0BDeH04U=F017BeH204I:E@0304E9E014B5<0A4QC00<0ADUD00@0@dMB04A8D`14B5<0ADUD
0P16BUD00`17BeH0B4aG04Q<E`0304U=F0<0B4eC00@0B4aE04I=E@16CED0ATeG0P15CUL01015CEP0
A4iH04A>F011CeX204A>E`40ADiE0@18CE410?ooo`00401:Ef0304UFGa40B5EN00<0BEIO04QEGP18
EEh01P17E5d00`18EEh0BEIO04UFG`0404UFG`80B5EN1017E5d204ICG0@0B5EN0`19EUl504YGH004
04EBF`16De`0AU=L04ICG080AeAM00@0B5EN04UFG`19EUl0BUMP0P1;F6401017FFD0AeQR04MHHP15
EUh204YGG`0404mJH019DUL0BE9G03@l?`807BDW0`0C4Q0301@C4@@05A@B0P0F5A<<01LF4@0401PD
2`0R6Pd08QX=04Pk:P80FdTc00@0FT@/05e7;`1MAbl0HD/d0P1ZESh3075B=P0407=D>01gECX0MeDj
07]G?@80O5Pn00@0O5Lm07YE>`1jEC/0P5Hm0P21F3`01026FSl0REe008UM@02>HD@2099SA@0309EV
B02FJ4L0UVQ700<0UFI<00@0UFM=09MYC`2GJDl0VViC0P2LLeL0102OMUX0XGQL0:5hG02SNU`20:Im
G`040:YnH@2WNeh0Yg]N0:9eF080X75E00@0XW=G0:EfFP2UMUX0XfeC0P2CH4D0102;FCh0R5To08QI
?`1iCC@2068j800405`g7@1V@BL0IT4W07]9<080Q54f00@0TU/l09iV@`2NIT<0Yfm<0P2]Me<0102a
Oeh0]8=S0;B3H`2fQV820:mmFP040:]iEP2bO5X0/WaJ0;MoGP80]W]K0`2dMeP0102[Le80ZgEC0:]e
D`2YMe@20:MfE@040:EdE02VLe@0YW=D0:adF@80[gIK0`2_LUD0102]LED0Zg9E0:]bE@2/M5T20:ik
H0030;F2I`2iQf`0^HM/00@0_H]`0P2kRFh00`2iQf`0^8I[0;R6J`030;R6J`040;f5JP2nQV/0_XI[
0;f7J`80^HIY00@0]XEW0;F6J02eQVP0]XUZ0P2iS6d01030SW<0_Xmc0;j?L`32UgL20;^BLP040;^C
L02hSF/0^8e[0;63H@80`i9a0`2hQVD0102gPV80]7iL0;AnG02cNeX20;EjF0040;=gE@2aLTl0/G9?
0:mdDP80ZVm=00@0Z6e;0:aaC`2/LDl0[G9@0P2YKT`0102VJdT0Yfa:0:M/BP2_MDl30:=Y@`80Yfe7
0`2SJD<0102UJdD0XfU30:=Y@`2NJd0209YW?00309ER=`2CH3D0Tf0e00@0U64f0P2BGc@0102BGSH0
UV8j09IR>P2FHSX2099N=P04091L=02BGSH0TUhf091L=080ReL_00@0Pe4^07E58@1eAB40Pe<_0P2K
KDT0102UNE@0/XQR0;:8HP2hTFX20:j7H0040:egE@2WKdh0Yfm>0:Y_C`80]gYK00@0a8EV0<B3I@34
PfD0_gaO0P2lMeX0102RMUd0TfeF09=]EP1gF4H205Q2=@0404Tn>P18B4X0B4Q:04=:D080>TA=00@0
>DYD03]9E00kBE@0@4QE0P13Ae<01014B5@0@dYD04=:E012Be@2049=D`H0ADaF00@0AdiH04U@FP19
D5X0Be9L0P1<Ded0101<DUh0CE=O04eCG`1=Del304iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17
CET204E;E`0304A:EP13BED0@dUE00H0@dUE1014BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0
B4iJ04U?F`19Ce/01019Ce/204Q>FP<0ATaH0`14BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18
CeT0101:CUT0BDeH04U=F017BeH204I:E@0304E9E014B5<0A4QC00<0ADUD0`14B5<01015BE@0ATYE
04I:E@17BeH304Q<E`80BDeH0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH
045?FP80A4iG0@15CUD104Q=D@40oooo000@04YGH0<0BEIO4@18EEh00`19EUl0B5EN04QEGP0604MD
G@0304QEGP19EUl0BEIO00<0BEIO1018EEh504MDG@@0B5EN0`19EUl504YGH0<0AU=L00@0AeAM04QE
GP18EEh0BEIO1@1:Ef001015EV00A5EO04AEG`17EEh204]HH00404iIG@15Ce40ADmA02hf>0806AhQ
100D4a4301DD4PD05QDC300G5Q40100L5@/0:R0D02XP501>?Rh205m:=@0405m7;`1QBC40HDTa06E=
=@80JE<l00@0L54e079C=`1bDcL0MEHj0P1hES/0101iEC/0NUHl07YF?01lEcd207eH?P0407eC>P1n
ECT0OUDi08=G?080Qe/n00@0S5m2091Q@`2@HD<0TfA60P2EIdH0102EJDh0U6Q=09AXC@2CId`2099Y
C@0409E/D02HLE@0V75D09aeF080W7MJ00@0XgYN0:AkG`2TNel0Yg]P0P2TN5d0102QLeT0X75G0:1a
E`2RLeT20:EfG0040:1]DP2BH4D0TV1508]I?P80QeTo00@0Ndlf06Dk8P1U>b80GCDL0P1S>b80101c
@BP0O4T^07a9;P2<ECH209YR?`040:=[B02ZM500ZWA@0:ilF`80/H1P00@0]XIR0;1nF`2`OU/0[7YG
0P2`NUP00`2cNeX0]GYJ0;EjFP030;AgF0040:aaD@2ZM580ZWAB0:QfD`80YWED00@0YGAD0:IcE02V
Le@0[7AI0P2bMU`0102_LUD0[75C0:aaD`2ZKU820:QaD`040:afFP2aOV<0/GiS0;F3J080^8I[00<0
^hia0;V<K`2iS6l00`2hRfh0102iS6l0^8]^0;R;KP2gRVd20;F8J`040;n9K@2nR6`0_XQ/0;^8J`80
]hAW0`2dPfD0102fQfT0^HY/0;V:K02nRg020;b=K`040;nDM02hT6d0^91]0;VAKP80^8eZ00@0/H=Q
0<:@K`32T6l0^hIV0P2fOf00102dNed0]W]K0;IkF`2gNU/20;EfE@040;9aD@2aKdd0/Fm=0:YbC`<0
Yfm<0P2ZLTl0102[Le00Yfm<0:M_C02TK4T20:=[B0040:M`B@2NId00WVM00:1Y@P80Y6e60`2OJ440
102QJT<0WVM009iW@02LJCh209QU>P0409ER=`2CH3D0Tf0e09=P=@<0U64f0P2CH3D0102@G3@0UF4i
09EQ>@2GHc/209AP>0<0Telg00@0Se/c08UE;@29EBd0Q4l]0P1g@R001020CR/0UfE209MU@P2QLDd2
0;20G0040;N9I@2`PUh0/89N0;9eEP80Zfi?00@0Zfa=0;EdEP2eM5H0_gaO0P33O6001031NUh0`gUN
0<=iGP2]L5/20:9[E`0408EJBP1SAC/0HdDk04i1>`80BDU;00@0@TeA03M8D00gB500>4UC0P0kAe<0
100oAe@0@dMC04=7D`15Ae@204E9E00304=<E@13CU@0@diD00H0A4]E00@0ATeG04Q?F@18CeT0BU5K
0P1;DU`0101<DUh0CE=O04eCG`1=Del304iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17CET204E;
E`0304A:EP13BED0@dUE00H0@dUE1014BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0B4iJ04U?
F`19Ce/01019Ce/204Q>FP<0ATaH0`14BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18CeT0101:
CUT0BDeH04U=F017BeH204I:E@0304E9E014B5<0A4QC00<0ADUD0`14B5<00`15BE@0ATYE04I:E@03
04M;EP<0B4aG0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH045?FP80A4iG
0@15CUD104Q=D@40oooo000@04YGH0<0BEIO4@18EEh00`19EUl0B5EN04QEGP0604MDG@0304QEGP19
EUl0BEIO00<0BEIO1018EEh504MDG@@0B5EN0`19EUl504YGH0<0AU=L00@0AeAM04QEGP18EEh0BEIO
1@1:Ef001015EV00A5EO04AEG`17EEh204]HH00404iIG@15Ce40ADmA02hf>0806AhQ100D4a4301DD
4PD05QDC300G5Q40100L5@/0:R0D02XP501>?Rh205m:=@0405m7;`1QBC40HDTa06E==@80JE<l00@0
L54e079C=`1bDcL0MEHj0P1hES/0101iEC/0NUHl07YF?01lEcd207eH?P0407eC>P1nECT0OUDi08=G
?080Qe/n00@0S5m2091Q@`2@HD<0TfA60P2EIdH0102EJDh0U6Q=09AXC@2CId`2099YC@0409E/D02H
LE@0V75D09aeF080W7MJ00@0XgYN0:AkG`2TNel0Yg]P0P2TN5d0102QLeT0X75G0:1aE`2RLeT20:Ef
G0040:1]DP2BH4D0TV1508]I?P80QeTo00@0Ndlf06Dk8P1U>b80GCDL0P1S>b80101c@BP0O4T^07a9
;P2<ECH209YR?`040:=[B02ZM500ZWA@0:ilF`80/H1P00@0]XIR0;1nF`2`OU/0[7YG0P2`NUP00`2c
NeX0]GYJ0;EjFP030;AgF0040:aaD@2ZM580ZWAB0:QfD`80YWED00@0YGAD0:IcE02VLe@0[7AI0P2b
MU`0102_LUD0[75C0:aaD`2ZKU820:QaD`040:afFP2aOV<0/GiS0;F3J080^8I[00<0^hia0;V<K`2i
S6l00`2hRfh0102iS6l0^8]^0;R;KP2gRVd20;F8J`040;n9K@2nR6`0_XQ/0;^8J`80]hAW0`2dPfD0
102fQfT0^HY/0;V:K02nRg020;b=K`040;nDM02hT6d0^91]0;VAKP80^8eZ00@0/H=Q0<:@K`32T6l0
^hIV0P2fOf00102dNed0]W]K0;IkF`2gNU/20;EfE@040;9aD@2aKdd0/Fm=0:YbC`<0Yfm<0P2ZLTl0
102[Le00Yfm<0:M_C02TK4T20:=[B0040:M`B@2NId00WVM00:1Y@P80Y6e60`2OJ440102QJT<0WVM0
09iW@02LJCh209QU>P0409ER=`2CH3D0Tf0e09=P=@<0U64f0P2CH3D0102@G3@0UF4i09EQ>@2GHc/2
09AP>0<0Telg00@0Se/c08UE;@29EBd0Q4l]0P1g@R001020CR/0UfE209MU@P2QLDd20;20G0040;N9
I@2`PUh0/89N0;9eEP80Zfi?00@0Zfa=0;EdEP2eM5H0_gaO0P33O6001031NUh0`gUN0<=iGP2]L5/2
0:9[E`0408EJBP1SAC/0HdDk04i1>`80BDU;00@0@TeA03M8D00gB500>4UC0P0kAe<0100oAe@0@dMC
04=7D`15Ae@204E9E00304=<E@13CU@0@diD00H0A4]E00@0ATeG04Q?F@18CeT0BU5K0P1;DU`0101<
DUh0CE=O04eCG`1=Del304iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17CET204E;E`0304A:EP13
BED0@dUE00H0@dUE1014BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0B4iJ04U?F`19Ce/01019
Ce/204Q>FP<0ATaH0`14BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18CeT0101:CUT0BDeH04U=
F017BeH204I:E@0304E9E014B5<0A4QC00<0ADUD0`14B5<00`15BE@0ATYE04I:E@0304M;EP<0B4aG
0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=
D@40oooo000<04UFG`@0BUMP0`19EUlA04QEGP0304UFG`18EEh0B5EN00H0AeAM00<0B5EN04UFG`19
EUl02@18EEh304MDG@@0B5EN0`19EUl504YGH0<0AeAM0`18EEh00`19EUl0BUMP04YGH00304YGH004
04EFH016E5d0AUAM04QGGP80D5]Q00@0D5UN04=;CP13Bdh0:Rlb0P0D6QX301@C4@@05A@B1@0F5A<<
01PG4P0401lF3@0f:Q`0=RXL05A3<@80Hddf00@0I4Xa06E;<P1UBc80JDlh0P1[Dc/0101`DCD0Le@h
07=D>01eESX207ME>P<0Me<i00@0O5Lm07mJ@01oFT00Ne4h0P1mE3P01022ES/0QUXm08IJ?@2;GT42
091Q@`0409=TAP2DITD0U6I509EYC080TVU;00@0T6M908mXB@2?J4T0TFY;0P2CKdl0102GLe<0VWIF
09YfEP2OMUX20:5hG0<0Y7QM00<0XgEK0:9cF@2RLeT00`2QLUP0102NK540TV15099PA@2:Fcl208QJ
@00407eA>01Z?RD0JShU05hd7080H3HN00@0JCLN07A19P1d@BH0QTl`0P2FGS/0102SJdP0[7IB0:af
DP2aOeh20;F4I0040;>3G`2_OEX0[geJ0:UgE080ZgEC00<0[GED0;1eE@2`MED00`2aM5D0102/LE40
Zg=B0:]cDP2YM5820:IdD`040:EbD`2XLe@0Z7=D0:edF@80/gMM00@0]7UK0;1eE`2`MEL0Zg9E0P2[
M5H0102`NUh0]89W0;B2I`2hQV/20;R9K@040;R?L@2fSFl0]Xe_0;B;K@80]Xe_0`2hSg400`2eS6h0
/XU[0;:9J`030;b9K0040;Z7JP2gQ6L0]hAW0;B3I@80]XEW00@0^hY/0<2?L@30Sg40`Xmb0P32TG<0
1034UgH0^i1^0;^@KP2kT6d20;^?K0040;>3G`2oSFX0_heZ0;j5I`80]WeO00@0]7UK0;UlG@2iO5d0
_GaN0P2fM5@0102cKTl0]6m>0;A_CP2_Me@20:eeDP040:adD@2]ME80[GEB0:eeDP80[7AA0`2[Le00
102VKdP0X6U20:1Y@P2SK4D20:E^A`<0WVM000@0WfQ109YS?02JHc`0VFDm0P2GHc/0102DH3P0TUhf
099N=P2BGSH209=O=`<0U60h00@0Se/c09EQ>@2EHCT0V6@l0P2FHSX0102EHCT0U60h09AP>02?Fc<2
08QD;00408E?;@1i@b40ND<Q081::080Tedk00@0WfY80:ehEP2]N5H0^8=Q0P2eP5h0102gMUP0]GAF
0;EdEP2gM5L20;YeF0040;M`E02fK540]VaA0;a`EP80aGQN00@0^FaH0;1/F@2`K5T0UUiA0P1_BT40
101D@Sh0B4U;04Q9B`0nCE8203=:D00403U7DP0jAU80>TIB03m5D`80A4IC00@0ATQE04I:E@16BUD0
ADaF0P15CUD404=:E080A4]E00@0ADaF04M>F017CUP0BE1J0P1:DE/0101<DUh0CE=O04eCG`1=Del3
04iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17CET204E;E`0304A:EP13BED0@dUE00H0@dUE1014
BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0B4iJ04U?F`19Ce/01019Ce/204Q>FP<0ATaH0`14
BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18CeT0101:CUT0BDeH04U=F017BeH204I:E@0304E9
E014B5<0A4QC00L0ADUD0P16BUD304M;EP<0B4aG0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`04
04E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000<04UFG`@0BUMP0`19EUlA04QEGP03
04UFG`18EEh0B5EN00H0AeAM00<0B5EN04UFG`19EUl00`17E5d=04QEGP<0BEIO1@1:Ef000`18EEh0
BEIO04UFG`0404UFG`D0BUMP00@0BEMP04QGGP18Eeh0C5UQ0P1CGV80101AFed0@TM:0497BP0V;2`2
018F5`@05A@B1@0F5A<301LF50`061LB00@08aT?044c9P11<bH0FTLf0P1VCCL0101XC3@0K4hd06a>
=01^DSX206iD?@04079C=`1dECT0M5Di07EF>P80ME<h00@0M50f07EA=`1eDCL0NUDk0P1nFCl0101l
DST0OE@h07eD>022ES/208EI?00408YM@02?H480Sf12099SA@80TfE400@0TfM:099VB@2BITT0TFE8
0P2?ITP0102@JDX0Tfa=09=/C@2DL50209IbDP0409Y`E`2LLUT0W79I0:9fG@80YGUP00@0YgUQ0:Ee
GP2UMEh0WfmH0P2KJe@0102JJdl0T655091QA@29G3l208YL@P04089D?01_@bX0Kd<Z068e7P80Gc8K
00@0IS@K078o901b?b@0QDh_0P2GGc`0102TK4T0[WQD0:ihE02cPF020;N6IP040;>3G`2aOe`0/GmL
0:ajE`80ZgEC00@0[7AC0;1eE@2`MED0/GAE0P2`Le@0102/Ke00[75A0:aaD@2YLe420:EcDP040:Aa
DP2YLU@0ZG9D0;1dFP80]GMN00@0]gaN0;5hFP2aN5X0[7EG0P2/MeP0102_O5l0]89W0;B2I`2eQVX2
0;F8J`040;N@L@2fSg00]Xm`0;F>K`80]Xm`0`2iTW<00`2eSVl0/X]/0;:;K0030;f<KP040;^:K02g
QVP0]hIX0;J3IP80^8EX00@0_HY]0<6>L@31SW40aI1a0P34Tg<01035V7L0_95^0;bAKP2nTg020<2D
K`040;N7H`33SfX0`hmZ0;n6J080]giP00@0]gaN0;aoH02lOf00_giP0P2hMUH0102eL540]W5@0;Ia
D02cOET20;5kE`040:ihE02ZM500ZWA@0:UcC`80[7IB00@0/7YF0;9lF02bO5P0ZGA<0P2UL4P0102W
LTX0YW590:IaB@2MJ40209aW?`0409eX@02FHCT0UV4i09IR>P80UF4i00@0Telg095M=@2AGCD0TEde
0P2BGSH0102DH3P0UF4i09EQ>@2?Fc<209EQ>@0409UU?@2GHc/0Uf<k09IR>P80UF4i00@0Se/c08QD
;028E2`0QDl]0P1kAB<0101oBBL0SUPf08iH=P2KID<20:QbD0040;=kFP2eOE`0]GeL0;MfF080^GII
00@0_GQK0;aeF@2lMET0]F]@0P2`I4X0102fJDl0`7=I0<1cF@2oKEL20;Y_G0040:1UE@1hD4H0N516
05Q5@@80ATM900@0>Ta@0319D00`BE00>DMB0P0lAU80100oAE<0A4IC04A6D`18AeD204M9E@0304M;
EP15CUD0ADiE00@0@dYD0P14BeD01015C5H0AdiH04M>F019D5X204YAF`0404aBGP1=Del0CE=O04eC
G`<0CUAP0`1=Del204aBGP0404Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13BED01P13BED4
04A:EPD0AD]G0`14BUH304E;E`<0ATaH0`17CET00`18CUX0BDmK04U?F`0404U?F`80B4iJ0`16C5P3
04A;E@<0ADaF0`16CEL304M>F0@0ATeG1@17CUP304Q?F@0404Y>F@19CEP0BDeH04M;EP80ATYE00<0
ADUD04A8D`14B5<01015BE@604I:E@D0Ad]F0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=
F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000<04UFG`@0BUMP0`19EUlA04QEGP0304UF
G`18EEh0B5EN00H0AeAM00<0B5EN04UFG`19EUl00`17E5d=04QEGP<0BEIO1@1:Ef000`18EEh0BEIO
04UFG`0404UFG`D0BUMP00@0BEMP04QGGP18Eeh0C5UQ0P1CGV80101AFed0@TM:0497BP0V;2`2018F
5`@05A@B1@0F5A<301LF50`061LB00@08aT?044c9P11<bH0FTLf0P1VCCL0101XC3@0K4hd06a>=01^
DSX206iD?@04079C=`1dECT0M5Di07EF>P80ME<h00@0M50f07EA=`1eDCL0NUDk0P1nFCl0101lDST0
OE@h07eD>022ES/208EI?00408YM@02?H480Sf12099SA@80TfE400@0TfM:099VB@2BITT0TFE80P2?
ITP0102@JDX0Tfa=09=/C@2DL50209IbDP0409Y`E`2LLUT0W79I0:9fG@80YGUP00@0YgUQ0:EeGP2U
MEh0WfmH0P2KJe@0102JJdl0T655091QA@29G3l208YL@P04089D?01_@bX0Kd<Z068e7P80Gc8K00@0
IS@K078o901b?b@0QDh_0P2GGc`0102TK4T0[WQD0:ihE02cPF020;N6IP040;>3G`2aOe`0/GmL0:aj
E`80ZgEC00@0[7AC0;1eE@2`MED0/GAE0P2`Le@0102/Ke00[75A0:aaD@2YLe420:EcDP040:AaDP2Y
LU@0ZG9D0;1dFP80]GMN00@0]gaN0;5hFP2aN5X0[7EG0P2/MeP0102_O5l0]89W0;B2I`2eQVX20;F8
J`040;N@L@2fSg00]Xm`0;F>K`80]Xm`0`2iTW<00`2eSVl0/X]/0;:;K0030;f<KP040;^:K02gQVP0
]hIX0;J3IP80^8EX00@0_HY]0<6>L@31SW40aI1a0P34Tg<01035V7L0_95^0;bAKP2nTg020<2DK`04
0;N7H`33SfX0`hmZ0;n6J080]giP00@0]gaN0;aoH02lOf00_giP0P2hMUH0102eL540]W5@0;IaD02c
OET20;5kE`040:ihE02ZM500ZWA@0:UcC`80[7IB00@0/7YF0;9lF02bO5P0ZGA<0P2UL4P0102WLTX0
YW590:IaB@2MJ40209aW?`0409eX@02FHCT0UV4i09IR>P80UF4i00@0Telg095M=@2AGCD0TEde0P2B
GSH0102DH3P0UF4i09EQ>@2?Fc<209EQ>@0409UU?@2GHc/0Uf<k09IR>P80UF4i00@0Se/c08QD;028
E2`0QDl]0P1kAB<0101oBBL0SUPf08iH=P2KID<20:QbD0040;=kFP2eOE`0]GeL0;MfF080^GII00@0
_GQK0;aeF@2lMET0]F]@0P2`I4X0102fJDl0`7=I0<1cF@2oKEL20;Y_G0040:1UE@1hD4H0N51605Q5
@@80ATM900@0>Ta@0319D00`BE00>DMB0P0lAU80100oAE<0A4IC04A6D`18AeD204M9E@0304M;EP15
CUD0ADiE00@0@dYD0P14BeD01015C5H0AdiH04M>F019D5X204YAF`0404aBGP1=Del0CE=O04eCG`<0
CUAP0`1=Del204aBGP0404Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13BED01P13BED404A:
EPD0AD]G0`14BUH304E;E`<0ATaH0`17CET00`18CUX0BDmK04U?F`0404U?F`80B4iJ0`16C5P304A;
E@<0ADaF0`16CEL304M>F0@0ATeG1@17CUP304Q?F@0404Y>F@19CEP0BDeH04M;EP80ATYE00<0ADUD
04A8D`14B5<01015BE@604I:E@D0Ad]F0`18CE<01018C5D0ATeE04I=E@16CEL204E>E`0404E=F014
CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000<04QEGP@0BUMP0`19EUlA04QEGP0304UFG`18
EEh0B5EN00H0AeAM00<0B5EN04UFG`19EUl01017E5d504QEGP<0BEIO1018EEh304UFG`D0BUMP0`19
EUl904YGH00404]JH@1:Eel0BUMO04mJH080EEiS00@0DEUL03e3@`0m@d<08RHW0P0?4a8401DD4P<0
5QDC1@0G5Q@<01TH4`0402LK3`1;>b`0Bc//05m:=`80Id/e00@0K4hd071@=`1`D3L0MELm0P1dESh0
101eESX0M5Di07AE>@1bDcL2079@=@04079>=01cCcD0Ldle07QC>@80NUDk00@0Ne4h07aC=`1lDcL0
PEDj0P24F3/01028Fch0S5do08aM?`2>Gd4208mQ@0<0TF9400@0TV=5099UAP2BIDH0TfQ80P2EJTX0
102EK4`0UVe=09I]C@2FK5<209Q^E@0409mcFP2UNF00YGUP0:UkH`80YgMP00@0WfmH09QXD@2HJ540
V6]>0P2?HTD0102:GD00S5i408aNA026F40207I8<00406Hh8@1N<1T0GS0I06Tg7P80MD8W00@0R54b
09QP?@2HH3d0Y6a90P2/MU80102`OUd0/h9R0;>2HP2eQF420;J4H@040;>1GP2aNeT0/G]I0;5iF080
]GYJ00@0]WUJ0;=fE`2cMUL0[6m@0P2[L500102XLU00YG1@0:E`D02UL5420:QaD`040;1dFP2eMeh0
]GMN0;=jG080/7MI00@0[7EG0:]fE`2[MUL0ZgYL0P2^Of<00`2aQ6L0]8MZ0;B7JP030;R?L@040;V@
LP2jTG<0^Y5c0;ZAL`80^I1b00@0]Xe_0;>:K02cRV`0_Xma0P2mSW00102mS6h0^HQZ0;V8JP2gQ6L2
0;J3IP040;^5J@2nR6`0_XQ/0;j9JP80`8e^00@0`YAb0;V=JP2iSFX0_Y9]0P34Ug00102mRfH0aY9]
0<JBK@2jQFD20;Z3I0040;Z1H`2mPV80_H9R0;]nG`80]gQG00<0]7=C0;EcD@2eLe400`2`O5L0102[
Me80YG5<0:EaC02SKdX20:UeD0040:mkEP2bOUT0/WiI0:]gC`80Z7A<00<0ZGE=0:A`B02TL4P00`2J
ISh00`2LJ400UF4i09EQ>@0309=O>004099N=`2AGCH0TEdf091L=@80TEdf00@0U60i09IR>`2FHS/0
T5`d0P2FHSX0102IICd0Uf<k09MS>`2FHSX209EQ>@04091L=029EBd0RED]08AB;`<0O4XW0P29E380
102FH3h0WfU709mYA`2XL4l20:adD`040;=dE@2cLU@0/g9D0;MdE`80^gII00@0^75E0;9XC@2bJ4d0
/fM=0P2hJe40102lK5<0^W9J0;YbFP2VJeT207iFBP0405Y7@@17AdT0AdM903]:D@80<DQ@00@0>dMC
03e7D`0mAe<0@TED0P16AE<01017AU@0B4QD04Q8E016BUD204E<E0H0A4]E00@0ATeG04Q?F@18CeT0
BU5K0P1;DU`0101<DUh0CE=O04eCG`1=Del304iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17CET2
04E;E`0304A:EP13BED0@dUE00H0@dUE1014BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0B4iJ
04U?F`19Ce/01019Ce/204Q>FP<0ATaH0`14BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18CeT0
101:CUT0BDeH04U=F017BeH204I:E@0304E9E014B5<0A4QC00<0ADUD2P16BUD204M;EP<0B4eC00@0
B4aE04I=E@16CED0ATeG0P15CUL01015CEP0A4iH04A>F011CeX204A>E`40ADiE0@18CE410?ooo`00
3018EEh404YGH0<0BEIO4@18EEh00`19EUl0B5EN04QEGP0604MDG@0404QEGP19EUl0BEIO04ICG0D0
AeAM00<0B5EN04UFG`19EUl00`19EUl404QEGP<0BEIO4@1:Ef00101<FF40BEIL04UFG01=F5h205=M
G`0404eEE`0i?Ch0>Cdn01dQ80803Q0?00<05A@B01HE4`0F5A<00`0F5A<301LF50<061LE300I61<0
100[7A80Dd4c05=1<`1QBSP206I:=00406i>=@1eDcP0ME<h07]K@P80NE]300@0N5Tm07AE>@1dECT0
L54e0P1_CC80101`C380LThd079>=01eD3H207IA=`0407U?=P1jDCD0NU4e07iB=`80P5@g00@0PeHi
08MH>P27F3X0REXl0P2:G3/0102BG400TEi1095N@@2DHD@209ETAP0409MVB02GJ4X0UfQ:09MXBP80
UfY;0`2GKED0102LKeP0XWEN0:9eGP2XNV<20:MgH@0409m_F@2HJ580V6QB09EYC080SV9500@0RUi1
08eOA@2=GdD0RUa40P1kCCD0101Y>b@0GblI05l_6@1X=Qd207A19P0408M@<@2HH3d0V60m0:A/B@80
[7IB00@0[geL0;:1H@2bPF40]8AP0P2hQV<0102fQ640]7iL0;AnG02cNeX20;IkF`040;IiFP2bMEH0
/WEF0:e^C`80[6m@00@0Z71?0:E`D02UL500Y6m@0P2YL5<0102aLeX0^7QO0;QhG`2aN5X20;1iFP04
0:iiFP2[N5T0ZgQI0:akG@80[WmS00@0]8MZ0;R<K`2hS6l0^Xia0P2lT7<30;jBM@040;b@L`2iSG00
^He`0;N;KP80]XY]00@0]hY[0;R;K02hRf`0^8U[0P2hQfT0102fPfH0]h5U0;N1I@2iPFH20;b4J@04
0;V2H`2jR6L0^XQW0;b>K080/XIQ00@0^8aW0<2CK030Tf`0^HMR0P31RfD0102cPF00^HMV0;V7IP2m
R6P20;V3H@040;AlF`2cN5H0/gQF0;9fE080/W=@00@0[GUD0:mkEP2_NeH0[7QC0P2ULD`0102SKdX0
ZGE@0:UeD02_NeH20;1lE`040:YfCP2XM4`0Z7A<0:UeC@80XVi600@0Uf<k09UU?@2IICd0WFU10P2G
Hc/0102AGCH0TUhg099N=`2BGSL2091L=@0408mK=02@G3D0T5`e09AP>@80Uf<l00@0TEde09MS>`2G
Hc/0VFDm0P2FHSX309EQ>@04095M=@2:ERh0RUH^089F<@80O4hZ00@0ND/W089B;P22DRh0S5Xg0P2C
GS`0102HHT00WFM509eWA@2YK4d20:e`D@040;MhF@2nNeh0_W]N0;UdE`80[fQ<00@0[VA90;5WC02a
Id`0]6];0P2eM5P0102TL5X0Oe]=07mKC@1KBT<204Q8BP0403e:DP0fBU<0=TYC03a8E080?dMD00@0
@dIE04I5D`16AE<0B4ID0P18B5@00`16B5@0A4]C04A;D`0604E<EP0404M>F019D5X0BE1J04]BG080
C5=M00@0C59N04eCG`1=Del0CE=O0`1>E60304eCG`80C59N00@0BU1L04U?F`19Ce/0AdeI0P15BeL0
0`14BUH0@dUE04=9E@0604=9E@@0A4YF1@15BeL304A:EP<0AD]G0`16C5P304M=F@0304Q>FP19Ce/0
BDmK00@0BDmK0P18CUX304I<F0<0A4]E0`15C5H304I=E`<0AdiH1016CEL504M>F0<0B4mI00@0BTiI
04U=F019CEP0Ad]F0P16BUD00`15BE@0A4QC04A8D`0304E9E0`0ATYE0`18CE<01018C5D0ATeE04I=
E@16CEL204E>E`0404E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000<04QEGP@0BUMP
0`19EUlA04QEGP0304UFG`18EEh0B5EN00H0AeAM00@0B5EN04UFG`19EUl0AU=L1@17E5d00`18EEh0
BEIO04UFG`0304UFG`@0B5EN0`19EUlA04YGH00404aIH@19EU`0BEIL04eHGP80DeeO00@0CEEG03Tm
?P0i?Ch07B4P0P0>40l00`0E51805QDC01HE4`0301HE4`<05aHD0`0H5aD<01TH4`0402/M4P1C@C<0
Dd4c065:>080ITXd00@0KThe07EC>01eDcP0Ne]20P1iFd<0101hFCd0M5Di07AE>@1`DCD206m=<P04
071<<P1bCS@0LThd07E@=P80MU4g00@0NDlf07YA=@1jDCD0OU8g0P20E3L01023EST0QePj08MH>P29
FS`208YL>`04099L@02AGT40TEi109AQA080UFA600@0UfI809MXBP2GJ4X0UfQ:0P2GJT/309M]E@04
09a_F02RMEh0XWEN0:QjH`80YgMQ00@0WfmI09QXDP2HJ580UFU<0P2>HTD0102:GT40SEm508eOA@2:
G4@207]==@0406Tk901O;aT0GblI06Pf7@80M44V00@0Qe0a09QP?@2HH3d0Y6a90P2/MU80102_OE`0
/X5Q0;:1H@2dQ6020;R6H`040;J4H@2dOU`0]7iL0;=kFP80]W]K00@0]WUJ0;9eEP2bMEH0[Fi?0P2/
Ke00102XL4l0YG1@0:E`D02TKe020:U`D`040;5cFP2hN5l0^7QO0;5hFP80/7UJ00@0[WUJ0:]hF@2[
N5T0[7]M0P2^Of<0102dQfX0^8a_0;R<K`2jSW420;b@L`<0_Y9e00@0_91c0;V=L02iSG00]h]^0P2f
RVd0102gRV/0^8]/0;R;K02hRF/20;R7J@040;J3IP2gPFD0]h5U0;V1IP80_8AY00@0^H9S0;Z8I`2j
R6L0_8i/0P2bQV40102hS6L0`9=/0<2CK02iQf820<6;I@040;>1H02iQfH0^HMV0;f8J080^H=Q00@0
]7aK0;=hEP2cN5H0/WID0P2bLe00102]NE@0[g]F0:mkEP2/N5<20:EaC0040:=_BP2YME00ZGE@0:mk
EP80/7aG00@0ZWI>0:QdC02XM4`0ZGE=0P2RKTH0102GHc/0VFDm09UU?@2MJD4209MS>`04095M=P2B
GSL0TUhg099N=`80T5`e00@0Se/d091L=@2@G3D0U60i0P2GHc`0102AGCD0Uf<k09MS>`2IICd209IR
>P<0UF4i00@0TEde08YF;P2:ERh0PUHa0P1lCRX0101iBbL0PU8^089B;P2<FSL209=N?00409QR@02M
IdD0WFM50:U/C@80[G1A00@0]gQI0;ikGP2nNeh0^GAG0P2_J4`0102^I4T0/FM<0;5WC02dJd/20;Ed
F0040:A`FP1oFdd0Oe]=05]:@`80B4Q:00@0?DYB03I:D`0fBU<0?4QD0P0oAe@01013AUD0ATEC04I5
D`18AU@204Q8E00304I8E014Be<0A4]C00H0ADaF00@0AdiH04U@FP19D5X0Be9L0P1<Ded0101<DUh0
CE=O04eCG`1=Del304iDH0<0CE=O0P1<DUh0101:D5`0BDmK04U?F`17CET204E;E`0304A:EP13BED0
@dUE00H0@dUE1014BUH504E;E`<0A4YF0`15BeL304I<F0<0AdeI00<0B4iJ04U?F`19Ce/01019Ce/2
04Q>FP<0ATaH0`14BeD304E<EP<0ATeG0`17CUP404I=E`D0AdiH0`18CeT0101:CUT0BDeH04U=F017
BeH204I:E@0304E9E014B5<0A4QC00<0ADUD3016BUD304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG
00@0ADeH04A>F014CUP0@DmJ0P14CUL104E>E@40B4eA0@3oool000`0B5EN101:Ef0304UFGa40B5EN
00<0BEIO04QEGP18EEh01P17E5d00`18EEh0BEIO04UFG`0304ICG0<0AeAM00<0B5EN04UFG`19EUl0
0`19EUl404QEGP<0BEIO1@1:Ef0304]HH@T0BUMP00@0C5UQ04QEF`18EE/0BeIJ0P1@FU`0101<DE@0
=3Pi03@h>@0K7A`200/=2PH05QDC0`0G5Q@301PG5@`06APC00@0;QdC05M4=P1GA3H0I4/g0P1WBC40
101aCc@0MU@i07ID>@20GTD207aL@`0407YK?`1dECT0M5Di06i?<`80KD/`00@0Kd/a079>=01bCS@0
Ldhd0P1bCC<0101gCC@0Mdhb07M><P1kCc@207eA=004081C=P23E3H0Pe@f08AE=`80QELf00@0T5Tk
09=L?P2CG3h0UF110P2II4D0102IITL0VVM809YWB02IJ4P209QWA`0409Q^EP2GKED0UfeE09U/E@80
Wg9K00@0YGMP0:MgH@2WMf40X71J0P2IJE<0102BJD/0SV9508iRA@2:GT4208eQAP0408aNAP1oCcP0
Odlh06`l9P80H30J00@0Hc4H06ll8@1_?240Q4d^0P2FGS/0102TK4T0[GMC0:egD`2aOeh20;F4I004
0;20G02ePf00]H=P0;F3H080/WaJ00@0/7QG0;AiF@2dNET0/gIG0P2^LE80102]KTl0Zfi?0:]^C`2X
L4l20:E^C`040:E^D02YL5<0ZG1C0;9dF`80^7QO00@0/GYK0;=lG@2cO5d0/WeN0P2`OEh0102`Of40
/hAX0;>4J02jSG020<2DM`040;^>L@2oTWD0_i9e0<:EN080`IAg00<0_8mb0;R;KP2hRfh00`2gRVd0
102_PV<0/HAU0;64I@2cQ6H20;F4IP040;F2I@2gPFD0]h5U0;Z2I`80_HEZ00@0]h1Q0;Z5I@2jQFD0
^XUX0P2]PE`0102bQUl0^h]U0;^;I@2bOUL20;N1F`040:anG02hQfH0^8MV0;f;JP80]h9P00@0/7YH
0;1hE@2`N5D0/GID0P2^M4l0102/N5<0/7aG0;1lE`2`O5L20:UeD0040:IbC@2/N5<0[7QC0;1lE`80
/GeH00@0Yg=;0:IbBP2VLTX0Yg=;0P2OJd<0102DH3P0V6@l09QT?02NJT8209QT?004091L=`2AGCP0
TEdh095M>080T5`g00@0SUXe08mK=P2?FcH0U60k0P2HI3l0102CGcL0Uf<k09MS>`2IICd209IR>P04
09AP>02EHCT0UF4i095M=@80ReL_00@0PeTc07aB;01lDR`0N4`W0P1mCb/01026ES80R5Hc08QF<`2<
EcD2099L>P040:1UA@2^Le<0[W=C0;n2H`80`hAU00@0]W=F0:IQA02VHD@0YUm30P2]ITX0102[JDL0
[gAD0:mdE02QLEX207iMCP0405a;A01:B4/0BTQ;045;E080>TYG00@0?dUE0418E@10B5D0A4IE0P17
AU@01018AU@0AdMC04M7D`15Ae<204E9DP@0ADaF0P16CEL01017CUP0BE1J04U@FP1;DU`204eDGP04
04aBGP1=Del0CE=O04eCG`<0CUAP0`1=Del204aBGP0404Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF
04=9E@13BED01P13BED404A:EPD0AD]G0`14BUH304E;E`<0ATaH0`17CET00`18CUX0BDmK04U?F`04
04U?F`80B4iJ0`16C5P304A;E@<0ADaF0`16CEL304M>F0@0ATeG1@17CUP304Q?F@0404Y>F@19CEP0
BDeH04M;EP80ATYE00<0ADUD04A8D`14B5<00`15BE@404M;EPP0ATYE0`18CE<01018C5D0ATeE04I=
E@16CEL204E>E`0404E=F014CUP0A4iH045?FP80A4iG0@15CUD104Q=D@40oooo000H04UFG``0B5EN
4017E5d504QEGP<0BEIO0`1;F640101:Ef00BEIO04UFG`19EUl204YGH0<0BeQQ00@0BUMP04]HH@1;
F640C5UR0`1=FV<204aIHP0404]HH@1:Ef00BUMP04QEGP80BUMP00@0CEMP051IH01@FF00BE1F0P0]
<C@0100D5AL0414C010A4`0I5Q4201XE4@0401dC4@0N51807Q@B01hD4P8071LC00@061TC01LJ4`0G
6Q<05aHD0P0H5aD0100M6AH081`I020L6@0M61@201XB3`0401lH4P0Y8Q`0:B8L03LV5P80FTDb00@0
ID/d06a=<@1/CC40Le0b0P1dDC40101oGSl0PV14089PA023Gc/207aH=00407EA;`1dD300M50`07A?
=080Ldhd00@0LD/d06m9<P1_BC80KDHU0P1^ABD0101bAbL0MDPW07E89`1jC2/207m>;@04089@;`23
DBh0Pe4^089@;`80UVA300@0U5lo09QQ@P2HHD80X6U:0P2NIDL0102OITP0WFA609eTAP2IHdT209YT
BP040:=`E@2XMUd0Z7IM0:5bF080XG=K00@0Y7QO09idG02NM5`0TfI90`2<Gd82095SB@0408iPB01n
D3P0OU0h06/m9P80HS@M00@0I38K07<o9`1c?bL0RU8g0P2MI4H0102YLE00/7UJ0;1iFP2cP6420;J5
I`040;22GP2_PEd0[h5M0:emF@80ZGMD00@0[7ME0;9jF@2bNUT0]7UI0P2_M5@0102_K540[V]@0:i[
D02/Jdl20:][C`040:Y]D02YL580ZG1B0:]bE080ZgAE00@0[gYJ0;9mG@2bOEd0[GaL0P2WNEP0102Z
Oel0]8e^0;B=KP2lV7P20;jKN`040;N>L02mU7H0_IAf0<2GN@80^i9d00@0/XU[0:n6J02_QVP0]Ha^
0P2lTgD0102cQ6P0/85U0;21I@2]OV820:enH0040:n1H02`PV40/89Q0;22H080[h5O00@0^H=W0:mi
G@2_NEd0]h9S0P2fPF80102dOel0]H1N0;F0GP2_NUP20;F0GP040;B2G`2aOe`0/GmL0;:0G@80^8IS
00@0]hER0;:0G@2bP5d0/7iK0P2cPEh0102[Me80[g]F0:mkEP2bOUT20;5kE`040:mgE02^MU<0[WIC
0;5fE080/WME00@0Z79>0:MaC@2WLDd0Y6i:0P2QJdL00`2MId<0VfE109]U@@0309YT@004095M=@2B
GSH0TUhf095M=@80Se/c0`2=FC400`2@G3@0Telg09=O=`0309EQ>@<0UV8j00@0U60h095M=@2AGCD0
SUXb0P2<F300102;FS80QUH`08IF<01mD2T207iB;`0408]M?02CH4<0Tf1308iF?@80QDTa00@0Rdha
0:1SAP2PHdH0]7AH0P31P6@01034OV@0[fM>0:mWCP2PESd20:EK@P040:ANA@2`KUH0/6iF0:abGP80
RUm?00@0I4U204a3A01<@d@0@4MA0P0iBEP0100iBU80?TQB03i8DP13AE8204I4DP0404M5D`16B5D0
ATQE04==E`80?e1H00@0A4iG04E>E`15CUL0A4eF0P16CEL01017CUP0BdmJ04]?FP1>DUd204mCGP@0
CUEO0`1=E5h504aCG@0404Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13BED00`13BED01011
Ae<0@TQD0498E013BED204A:EP0304E;E`16C5P0ATaH00<0AdeI1016CEL504M>F0<0B4mI0`16CEL0
1017CUP0B4mI04Q?F@18CeT204M>F0L0ATeG1@17CUP?04Q?F@<0Ad]F00<0ADUD04A8D`14B5<01014
B5<204E9E0<0BDeH0`18C5L404M;EP80ATYE0`16Be401016BU<0A4]C04A;D`14BeD204=<E@0404=;
EP12C5H0@TaF045?FP80A4iG0@16CeH104Y?D`40oooo000H04UFG``0B5EN4017E5d504QEGP<0BEIO
0`1;F640101:Ef00BEIO04UFG`19EUl204YGH0<0BeQQ00@0BUMP04]HH@1;F640C5UR0`1=FV<204aI
HP0404]HH@1:Ef00BUMP04QEGP80BUMP00@0CEMP051IH01@FF00BE1F0P0]<C@0100D5AL0414C010A
4`0I5Q4201XE4@0401dC4@0N51807Q@B01hD4P8071LC00@061TC01LJ4`0G6Q<05aHD0P0H5aD0100M
6AH081`I020L6@0M61@201XB3`0401lH4P0Y8Q`0:B8L03LV5P80FTDb00@0ID/d06a=<@1/CC40Le0b
0P1dDC40101oGSl0PV14089PA023Gc/207aH=00407EA;`1dD300M50`07A?=080Ldhd00@0LD/d06m9
<P1_BC80KDHU0P1^ABD0101bAbL0MDPW07E89`1jC2/207m>;@04089@;`23DBh0Pe4^089@;`80UVA3
00@0U5lo09QQ@P2HHD80X6U:0P2NIDL0102OITP0WFA609eTAP2IHdT209YTBP040:=`E@2XMUd0Z7IM
0:5bF080XG=K00@0Y7QO09idG02NM5`0TfI90`2<Gd82095SB@0408iPB01nD3P0OU0h06/m9P80HS@M
00@0I38K07<o9`1c?bL0RU8g0P2MI4H0102YLE00/7UJ0;1iFP2cP6420;J5I`040;22GP2_PEd0[h5M
0:emF@80ZGMD00@0[7ME0;9jF@2bNUT0]7UI0P2_M5@0102_K540[V]@0:i[D02/Jdl20:][C`040:Y]
D02YL580ZG1B0:]bE080ZgAE00@0[gYJ0;9mG@2bOEd0[GaL0P2WNEP0102ZOel0]8e^0;B=KP2lV7P2
0;jKN`040;N>L02mU7H0_IAf0<2GN@80^i9d00@0/XU[0:n6J02_QVP0]Ha^0P2lTgD0102cQ6P0/85U
0;21I@2]OV820:enH0040:n1H02`PV40/89Q0;22H080[h5O00@0^H=W0:miG@2_NEd0]h9S0P2fPF80
102dOel0]H1N0;F0GP2_NUP20;F0GP040;B2G`2aOe`0/GmL0;:0G@80^8IS00@0]hER0;:0G@2bP5d0
/7iK0P2cPEh0102[Me80[g]F0:mkEP2bOUT20;5kE`040:mgE02^MU<0[WIC0;5fE080/WME00@0Z79>
0:MaC@2WLDd0Y6i:0P2QJdL00`2MId<0VfE109]U@@0309YT@004095M=@2BGSH0TUhf095M=@80Se/c
0`2=FC400`2@G3@0Telg09=O=`0309EQ>@<0UV8j00@0U60h095M=@2AGCD0SUXb0P2<F300102;FS80
QUH`08IF<01mD2T207iB;`0408]M?02CH4<0Tf1308iF?@80QDTa00@0Rdha0:1SAP2PHdH0]7AH0P31
P6@01034OV@0[fM>0:mWCP2PESd20:EK@P040:ANA@2`KUH0/6iF0:abGP80RUm?00@0I4U204a3A01<
@d@0@4MA0P0iBEP0100iBU80?TQB03i8DP13AE8204I4DP0404M5D`16B5D0ATQE04==E`80?e1H00@0
A4iG04E>E`15CUL0A4eF0P16CEL01017CUP0BdmJ04]?FP1>DUd204mCGP@0CUEO0`1=E5h504aCG@04
04Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13BED00`13BED01011Ae<0@TQD0498E013BED2
04A:EP0304E;E`16C5P0ATaH00<0AdeI1016CEL504M>F0<0B4mI0`16CEL01017CUP0B4mI04Q?F@18
CeT204M>F0L0ATeG1@17CUP?04Q?F@<0Ad]F00<0ADUD04A8D`14B5<01014B5<204E9E0<0BDeH0`18
C5L404M;EP80ATYE0`16Be401016BU<0A4]C04A;D`14BeD204=<E@0404=;EP12C5H0@TaF045?FP80
A4iG0@16CeH104Y?D`40oooo000H04UFG``0B5EN4017E5d504QEGP<0BEIO0`1;F640101:Ef00BEIO
04UFG`19EUl204YGH0<0BeQQ00@0BUMP04]HH@1;F640C5UR0`1=FV<204aIHP0404]HH@1:Ef00BUMP
04UFG`80BeQQ00@0CEMP051IH01@FF00BE1F0P0/<3<0100C51H0414C010A4`0I5Q4201XE4@0401dC
4@0N51807Q@B01hD4P8071LC00<061TC01LJ4`0G6Q<00`0H5aD0100L61D07Q/F01hK5P0L5a<201/D
3P04024H4`0Y81/0:B0K03LV5P80FTDb00@0IT`e06i?<`1^Cc<0ME8d0P1fDcD01020Gd00P654081Q
A029ID4207mK>@0407=?;`1^BB`0KTT/075<<@80M4le00@0L4Xc06U3;P1Y@bh0Jd@U0P1/@bD0101]
A2H0L4DU07159@1dAbP207Q::@0407e<;01oCRd0Odh]07m>;@80TV5000@0Tf5009QVA@2HITD0W6M7
0P2HHD80102KI4D0UEho09EN?`2EHDT209APB00409e[DP2SLEP0Xg5H09m_F080XG=K00@0Y7QO0:1d
F`2PM5/0UfY=0P2=H4<01028Fch0SV1608iPAP2@HTX208=E?@0406m1:P1P<Q/0H38K06Hd7@80MD4Y
00@0Re<h09eTAP2MI4H0Z71?0P2_N5T0102cP640]hIX0;N6J02_PEd20;22GP040;20G02/NUL0[7YG
0:ehEP80/GUH00@0/GIF0:aaD@2/LE40[V]@0P2]JTl0102ZJDd0Z6Q<0:QXC02VJdd20:Q_D@040:]d
E@2]MUL0[GIG0:ehF080[WaK00@0[WeM0:YlF`2ZO5/0[H9R0P2eSVl0102jUWH0^iMg0;^GM`2gSW02
0;^BM0040;bCM@2hSg40^8ma0;27J@80[8=U00@0[hIX0;>:K02cRV`0]HIZ0P2cQ6P00`2aPVH0/X=U
0;:3I@030;B6I@040;>5H`2aPf40/H=Q0;N1I@80/G]O00@0^8=T0;J1HP2fPF80/giN0P2bOE/0102]
N5H0]X5O0;J1G`2dPUl20;:0G@040;B2G`2hQV<0^8IS0;J4H@80/7iK00@0[G]H0:mmFP2_OEX0/gmJ
0`2^NUD20;9lF0040;5iEP2[Le00Zg=@0:ebD080/gQF00@0Z79>0:I`C02VL4`0Xfe90P2OJDD00`2L
IT80VfE109]U@@0309YT@004095M=@2BGSH0TUhf099N=P80T5`d0`2=FC40102@G3@0TUhf099N=P2C
GcL209AP>0<0UF4i00@0Telg091L=02@G3@0SETa0P2<F3001023DRX0P50Z081@:P1kCRL207a@;@04
08UL>`2JJD/0VVU;0:1XC`80WF=;00@0Se8f09UL@02IG400YVI;0P2gMUX01033P6D0^WAJ0;YdFP2X
H4L20:MMA00409iF?@2[JE40ZfUA0:Y`G080SEm@00@0Ida304i5AP1>ADH0@DQB0P0iBEP0100iBU80
?DMA03e7D@12A54204I4DP0404M5D`15Ae@0ADMD049<EP80?TmG00@0A4iG04A=EP14CEH0A4eF0P16
CEL01017CUP0C51K04a@F`1>DUd204mCGP@0CEAN1@1<Ded304]BG00404Y@G019Ce/0BDmK04M=F@80
AD]G00<0A4YF04=9E@13BED00`13BED30457D`040498E013BED0@dUE04A:EP80AD]G0`16C5P404I=
E`D0AdiH0`18CeT304I=E`0404M>F018CeT0B4mI04Q?F@80AdiH1`16CEL504M>F0l0B4mI00@0B4aG
04M;EP17BeH0ADUD1@14B5<304E9E0<0BDeH0`18C5L604M;EP<0AT]A00@0ATYC04A;D`14Be<0A4]E
0P13C5D01013BeH0@TaF049<EP11CeX204A>E`40ATmF0@1:Ce<10?ooo`006019EUl<04QEGQ00AeAM
1@18EEh304UFG`<0BeQQ00@0BUMP04UFG`19EUl0BEIO0P1:Ef0304]HH@0404YGH01;F640BeQQ04aI
HP<0CEYS0P1<FF80101;F640BUMP04YGH01;F64204eJH`0404iHH@1@FF00D5UP04M>E080:blb00@0
4a@F010A4`0@4A<06AHA0P0J5A40100M4a407Q@B01hD4P0N518201`G4`0401PI4`0G6Q<05aXC01TJ
5@806APD00@06aLD01dH500M61@07AH@0P0M50l0100S6180:QlI02XO6@0h9AL205a5=@0306U==`1a
DSH0LE8f00<0MUDf00@0OUm207iQ@`1nHD<0S6Y70P21GSh0101`CBl0I44U06A19@1W@bT206m;<`04
06a8<P1S?bT0HclY06Q09P80IclU00@0ISlR06Ln8P1W?R80JShQ0P1]@B@0101bABH0MDPY07E8:@1_
@R<208AG>004099SA@2IJT`0VFY<099Q@`80S5/m00@0VVM:09QUB02HIDP0UF=<0P2AGdP0102GIDh0
Vf]D09][E02JK5@20:1bFP040:EiH02QME`0XGEL09a_DP80Sf9500@0PUDh08YL@P2:G480TfE=0P2;
GDD0101eAc00G2hG05`^5`1Y=b0207M3:`0408eE>P2MI4H0WFA60:I^C@80[GIG00@0/WmP0;N6J02g
QVP0[h5M0P2cQF40102dQ600[geJ0:mmFP2]N5H20:mgEP040:mdE02ZKdl0ZVm?0:aYCP80ZFQ<00@0
YFE90:9UB02RIDP0XfQ:0P2WKU00102/MEH0/7]K0;1kF`2[NEP20:elF`040:imG@2/Oeh0[7mN0;25
I@80]Xe_0`2iTW<0102iSG00^Xia0;Z>L@2kSg820;R<K`030;B8J`2aQFP0/HEX00<0/HEX00@0]hQ/
0;J7J`2fQf/0]8EY0P2dQFL0102eQfH0]8IU0;B6I@2aPf420:n1G`040;1mH02^Neh0[W]N0;>0H@80
/WmP00@0/7iM0:ajE`2/NUL0YgEB0P2_OEX30;F3H0040;N5HP2jR6D0^XQU0;R6H`80/GmL00@0[G]H
0:ajE`2/NUL0^XIQ0P2^NUD0102[Me80/WaH0;9lF02bNUL20:YbC`040:]`CP2cN5H0/gQF0:UcC`80
YW1<00@0XF]709eW@`2MId<0VfE11@2JI400102AGCD0TUhf099N=P2CGcL2095M=@0408mK<`2>FS80
SUXb08mK<`80TEde00<0TUhf09=O=`2CGcL00`2DH3P0102CGcL0T5`d091L=02=FC4208aH<00407i=
9@1nCRP0OThX07Y>9`80NTl/00@0QUTh09m^D02OKU00/WaR0P2hOVH0102GFd40U5Hm09AF?@2DESd2
0:QVC0040<5nH`34OV@0a7iT0;A/D`80ZF9600@0UThe0:MTBP2WI4X0Z6iH0P2>H500101ZCDD0DDM8
0557B013BU8203Y:E`0403U:DP0mAe40?DMA0494D@80AD=A00@0ATAB04A6D`14AU<0@D]E0P0mCUH0
1013CEH0A4eF04A=EP15CUL204I=E`0404Q?F@1<D5/0C51K04eAG080Ce=N1P1<Ded304]BG0<0BU5K
00@0BU1L04U?F`19Ce/0AdeI0P15BeL00`14BUH0@dUE04=9E@0304=9E@040416DP11Ae<0@DMC0457
D`80@TQD00@0@dUE04A:EP14BUH0A4YF0P15BeL404I=E`D0AdiH0`18CeT00`16CEL0AdiH04M>F006
04Q?F@0304M>F016CEL0ATeG00@0ATeG1@17CUP?04Q?F@0404Q<E`17BeH0Ad]F04I:E@80ADUD0`14
B5<304E9E0@0BDeH0`18C5L504M;EP<0AdaB00@0Ad]D04E<E015C5@0ADaF0P14CEH01014C5L0@deG
04==E`11CeX204A>E`40ATmF0@1:Ce<10?ooo`006019EUl<04QEGQ00AeAM1@18EEh304UFG`<0BeQQ
00@0BUMP04UFG`19EUl0BEIO0P1:Ef0304]HH@0404YGH01;F640BeQQ04aIHP<0CEYS0P1<FF80101;
F640BUMP04YGH01;F64204eJH`0404iHH@1@FF00D5UP04M>E080:blb00@04a@F010A4`0@4A<06AHA
0P0J5A40100M4a407Q@B01hD4P0N518201`G4`0401PI4`0G6Q<05aXC01TJ5@806APD00@06aLD01dH
500M61@07AH@0P0M50l0100S6180:QlI02XO6@0h9AL205a5=@0306U==`1aDSH0LE8f00<0MUDf00@0
OUm207iQ@`1nHD<0S6Y70P21GSh0101`CBl0I44U06A19@1W@bT206m;<`0406a8<P1S?bT0HclY06Q0
9P80IclU00@0ISlR06Ln8P1W?R80JShQ0P1]@B@0101bABH0MDPY07E8:@1_@R<208AG>004099SA@2I
JT`0VFY<099Q@`80S5/m00@0VVM:09QUB02HIDP0UF=<0P2AGdP0102GIDh0Vf]D09][E02JK5@20:1b
FP040:EiH02QME`0XGEL09a_DP80Sf9500@0PUDh08YL@P2:G480TfE=0P2;GDD0101eAc00G2hG05`^
5`1Y=b0207M3:`0408eE>P2MI4H0WFA60:I^C@80[GIG00@0/WmP0;N6J02gQVP0[h5M0P2cQF40102d
Q600[geJ0:mmFP2]N5H20:mgEP040:mdE02ZKdl0ZVm?0:aYCP80ZFQ<00@0YFE90:9UB02RIDP0XfQ:
0P2WKU00102/MEH0/7]K0;1kF`2[NEP20:elF`040:imG@2/Oeh0[7mN0;25I@80]Xe_0`2iTW<0102i
SG00^Xia0;Z>L@2kSg820;R<K`030;B8J`2aQFP0/HEX00<0/HEX00@0]hQ/0;J7J`2fQf/0]8EY0P2d
QFL0102eQfH0]8IU0;B6I@2aPf420:n1G`040;1mH02^Neh0[W]N0;>0H@80/WmP00@0/7iM0:ajE`2/
NUL0YgEB0P2_OEX30;F3H0040;N5HP2jR6D0^XQU0;R6H`80/GmL00@0[G]H0:ajE`2/NUL0^XIQ0P2^
NUD0102[Me80/WaH0;9lF02bNUL20:YbC`040:]`CP2cN5H0/gQF0:UcC`80YW1<00@0XF]709eW@`2M
Id<0VfE11@2JI400102AGCD0TUhf099N=P2CGcL2095M=@0408mK<`2>FS80SUXb08mK<`80TEde00<0
TUhf09=O=`2CGcL00`2DH3P0102CGcL0T5`d091L=02=FC4208aH<00407i=9@1nCRP0OThX07Y>9`80
NTl/00@0QUTh09m^D02OKU00/WaR0P2hOVH0102GFd40U5Hm09AF?@2DESd20:QVC0040<5nH`34OV@0
a7iT0;A/D`80ZF9600@0UThe0:MTBP2WI4X0Z6iH0P2>H500101ZCDD0DDM80557B013BU8203Y:E`04
03U:DP0mAe40?DMA0494D@80AD=A00@0ATAB04A6D`14AU<0@D]E0P0mCUH01013CEH0A4eF04A=EP15
CUL204I=E`0404Q?F@1<D5/0C51K04eAG080Ce=N1P1<Ded304]BG0<0BU5K00@0BU1L04U?F`19Ce/0
AdeI0P15BeL00`14BUH0@dUE04=9E@0304=9E@040416DP11Ae<0@DMC0457D`80@TQD00@0@dUE04A:
EP14BUH0A4YF0P15BeL404I=E`D0AdiH0`18CeT00`16CEL0AdiH04M>F00604Q?F@0304M>F016CEL0
ATeG00@0ATeG1@17CUP?04Q?F@0404Q<E`17BeH0Ad]F04I:E@80ADUD0`14B5<304E9E0@0BDeH0`18
C5L504M;EP<0AdaB00@0Ad]D04E<E015C5@0ADaF0P14CEH01014C5L0@deG04==E`11CeX204A>E`40
ATmF0@1:Ce<10?ooo`006019EUl<04QEGQ00AeAM1@18EEh304UFG`<0BeQQ00@0BUMP04UFG`19EUl0
BEIO0P1:Ef0304]HH@0404YGH01;F640BeQQ04aIHP<0CEYS0P1<FF80101;F640BUMP04YGH01<FF82
04iKI00404mIHP1?F5l0CeQO04E<DP80:Bd`00@04Q<E014B500A4Q@06AHA0P0J5A40100M4a407Q@B
01hD4P0N518201`G4`0401PI4`0G6Q<05aXC01XK5P806AXE00@061LB01/F400K5Q007QD@0P0P5@l0
100V6A<0:a`G02/L5`0m9aX20657>00406a@>`1cDcX0Le<j07ME>@80ME<g00@0NUdo07QM@01hGD00
QVE40P20Gd00101_CC40GC/P05dk801J>1l206=0:P0406E2;P1O?2T0Gc`Y064k9080H3XS00@0GSPQ
05`g7@1L=ad0GCDL0P1O=ad0101T?280J44T06Q1901F;Q@206i6;00408IL@`2@ITd0T6I=085E?080
OU8i00@0VVaD0:1bFP2PLUX0Vf]E0P2CHdd0102CIDh0UFM@09EWD02DITh209e_E`040:IhH02SMe`0
XgML0:5dE`80TfI900@0OU4d08EG?@25Ecd0U6I>0P2@HTX0101kCCH0FR`E05X/5@1[>B8207Y6;P04
08mG?02MI4H0WFA60:I^C@80[7EF00@0/GiO0;N6J02gQVP0/hEQ0P2fR6@0102fQV80[geJ0:mmFP2/
MeD20:eeE0040:ebDP2YKTh0ZFi>0:MVBP80YFA800@0XF5509mRA@2OHTD0X6M90P2WL540102^NET0
]7mO0;AoG`2^OE`20:elF`040:]mG02/Oeh0[7mN0;25I@80]XY]0`2hSg400`2iSG00^8a_0;R<K`03
0;R<K`040;N;KP2eRF`0]HU/0;>7JP80/XIY00@0]HIZ0;:3I`2bPfL0[h1T0P2]OV030:j0G`040:an
G02ZO5X0ZWaJ0:ejG@80[7UL00@0[GYK0;1mGP2`OEh0]H=R0P2]NeP0102SLDh0YWAA0:IdD@2dPUl2
0;F3H0040;R6H`2jR6D0^XQU0;R6H`80/h5N00@0[geJ0:ajE`2/NUL0^HEP0P2`O5L0102ZMU40[GMC
0:egD`2`N5D20:eeDP040:mdDP2aMU@0/GID0:YdD080YW1<00@0X6Y609aV@P2LIT80VfE11@2JI400
102AGCD0Telg09=O=`2DH3P209=O=`04091L=02>FS80SUXb08mK<`80T5`d00<0Telg09AP>02DH3P0
0`2EHCT0102DH3P0TUhf099N=P2?Fc<208iJ<P04089B:P20Db`0P5</07iB:`80NDh[00@0OE8`09]/
CP2KK4h0^hE[0P33Rg80102VJE@0Ue]309MK@`2=CcP209]K@P040;EcF@32Of@0`WmT0;YeF080[fQ<
00@0Uddb0:ERB02UHTP0ZFaG0P2>Gdl0101YCD80DDM60557AP15BE8203e;F00403U:DP0nB580?TQB
0494D@80AD=A00@0ATAB04A6D`14AU<0@4YD0P0lCED01013CEH0A4eF04A=EP15CUL204M>F00404Q?
F@1<D5/0C51K04eAG080CU9M0`1<Ded304]BG0@0BU5K0P19D5X0101:D5`0BDmK04U?F`17CET204E;
E`0304A:EP13BED0@dUE00<0@dUE00@0@4IB0457D`11Ae<0@DMC0`12B5@504=9E@@0ATeG1@17CUP3
04Q?F@<0AdiH00@0B4mI04U@FP19D5X0BE1J0P18CeT304M>F0@0ATeG1@17CUP?04Q?F@<0B4aG00<0
ATYE04E9E015BE@01015BE@204I:E@<0BTiI0`19CEP304Q<E`<0Ad]F0`17C5801017Be@0ADaD04E<
E015C5H204A=EP0404A<E`13CEL0@deG045?FP80A4iG0@16CeH104Y?D`40oooo000H04UFG``0B5EN
4017E5d504QEGP<0BEIO0`1;F640101:Ef00BEIO04UFG`19EUl204YGH0<0BeQQ00@0BUMP04]HH@1;
F640C5UR0`1=FV<204aIHP0404]HH@1:Ef00BUMP04eJH`80CeaU00@0CeUR04mHG`1?F5l0A4]A0P0X
;2l0100B4aD04a@F01<D5P0I5Q4201XE4@0401dC4@0N51807Q@B01hD4P8071LC00@061TC01LJ4`0G
6Q<06A`E0P0I6Q@0100H61006A@>01TD3P0N5@h202DH4P0402XI4P0/6A<0;1TC048Z8080IDTm00@0
L54o07EC>P1eDcX0ME<h0P1`DC@0101cF3/0LEPj075H>P1iG3h207MJ?00406Y<<01F>1h0ESPN04/]
5@80CRlK00<0ESLS05Tj:01I>RP00`1I=280101H<b00ES4N05Ha7P1D<1X205Dc6P0405/i801P?RD0
H3hU04`Z4@80G3XQ00<0KTXb07]G?`1kEcl00`1aBc@0102?IU00UFaF09E/EP2RMF8209Q[F00409EX
D`2CITl0TfI?091RB`80V6YB0`2TMU`0102QM5L0UfY=09MZC@1nDC@208=E>`04095SB`2@HTX0T69:
07i@>@80FB/D00@0JcTR07Y6;P1jARh0TETn0P2OITP0102VKTd0ZgAE0:]dE@2aOUl20;N6J0040;R:
IP2iRfL0^H]W0;J6HP80[WaI00@0ZgID0:eeE02]ME@0[G9B0P2YKTh0102RHTH0XF550:5QA@2NHD@2
09mTAP040:=ZC02ZLe@0ZW=D0;5lG080]H=R00@0/H=Q0:anG02/OU`0Z7]J0P2ZOE`0102_PV<0]8Q[
0;B8J`2hS6l20;Z>L@040;R;KP2fRF`0]XU/0;B7JP80/hIY00@0/XEX0;23IP2`PfH0[H1S0P2[OV40
102YNUh0YWMK0:IgF`2RLeL20:9cE@030:EgEP2XNUT0Z7YI00<0ZWaJ0`2_OV00102YN5P0/X5Q0;:1
H@2oSVd20;J6HP040:IfDP2RLTh0XW9>0:ajE`80[geJ00@0/X1M0;>1GP2cPEh0/X1M0P2`OU/00`2]
NeP0ZWQE0:YhE@030;9nF@040:eiE02YLdl0ZG=?0:adD@80/g]H00@0]GYH0;1eD`2`ME<0ZgEA0P2V
L4`00`2PJTH0WFM309eW@`0309aV@P0409YT@02IHcl0VF<o095M=@80Telg00@0UV8j09EQ>@2EHCT0
TUhf0P2?Fc<308iJ<P0309AP>02EHCT0UF4i00<0Uf<k00@0UV8j09AP>02DH3P0TUhf0P2@G3@01026
ERh0PUH_089F;`1oEBd207E=:@0407Q=:`2HJD/0V6U;0;^8K@80a8ac00@0/WMU09mUD@2OIE40TE@o
0P2DESl0102WID/0^7EJ0;QeFP2lMeX20;A_D00409Y@=@2YHdT0ZF=90:Q/E080S5e;00@0J4Y00556
A01AAT@0ADY@0P0nC5L0100jBe<0?dUC03m9D`13AE8304I4DP80A4IC00@0@4YD03a=E@0lCED0@TaE
0P14CEH01015CUL0B4mI04Q?F@19D5X204a@F`<0CE5L0`1<Ded304]BG0@0BU5K0P19D5X0101:D5`0
BDmK04U?F`17CET204E;E`0304A:EP13BED0@dUE00<0@dUE1011Ae<50498E0<0@dUE1016CEL504M>
F0<0B4mI00<0AdiH04Q?F@18CeT01P19D5X00`18CeT0AdiH04M>F00404I=E`D0AdiH3`18CeT01019
CEP0B4aG04Q<E`17BeH204I:E@<0ADUD0`16BUD404Y>F@<0BDeH1@18C5L304Q=D`0404Q<E@16CED0
ATeE04I=E`80ADiG00@0ADeH04A>F014CUP0@DmJ0P14CUL104I?EP40BTmC0@3oool001P0BEIO3018
EEh@04MDG@D0B5EN0`19EUl304]HH@0404YGH019EUl0BEIO04UFG`80BUMP0`1;F640101:Ef00BeQQ
04]HH@1<FF8304eJH`80C5UR00@0BeQQ04YGH01:Ef00CEYS0P1?G6D0101?FF80CeQO04mHG`14Be42
02P/;`04018C5@0C51H04a@F01TF4@806QDA00@07A<A01hD4P0N51807Q@B0P0L5a<0100H6A<05aXC
01LJ4`0I71D201TJ500401PH400I50h06A@>01hE3P809APB00@0:QTB02`I4`0/6A<0@RXP0P1UBCd0
101`DCl0ME<j07EC>P1eDcP2071A=00407=H>`1aF3X0LEPj07UL?P80MeXl00@0JT``05Hh7P1F>1h0
BbdE0P1>;a/00`1F=b<0FCXX05Tj:00305Td8P0405Pc801F<Ah0ES4N05@`6P80EC<J00@0FcTP060n
9@1P?RD0C2XA0P1L>R400`1^BS80NeLo07]G?`03075;=00408mVD02EK5H0UFaF0:9eHP80V6]H00@0
UFQC09=VC`2CITl0T69;0P2HJU830:AfG0040:5dE`2GJTd0UfY=07iA=080PeDk00@0TF=;091RBP2@
HTX0OU0i0P1I:a@0101[>B80NTH^07Y6;P2AFCh209mVB0040:I^C@2[M5D0ZgAE0;5nG`80]hIX00@0
^8YV0;V;I`2iRfL0]XIR0P2^O5T0102[MU@0[GED0:eeE02]LU820:U^CP040:9RAP2QHDD0XF5509iQ
A080WfA600@0XfY<0:YcE02ZLe@0/GaL0P2ePf80102aPf40[7iL0:anG02XNeX20:YmG0040:n2H`2d
R6/0]8Q[0;R<K`80^Xia00@0^8]^0;J9K02fRF`0]8MZ0P2cQVT0102bQFP0/8=V0;23IP2]P6<20:]n
H@040:UjGP2VMe/0YWMK0:9cE`80XW=E00<0YGMF0:QjF@2XNUT00`2ZO5X30:mnH0040:UhF02bPF40
/X5Q0;n>K@80]XIR00@0YWIB0:9bCP2RLTh0[7YG0P2_OEX0102bP5d0/h5N0;>1GP2bP5d20;1nF`03
0:ekF02ZN5D0ZWQE00<0/WiI00@0[GUD0:UcC`2YLdl0[7AA0P2cNeP0102eNUP0/7EC0;1eD`2[ME42
0:I`C0030:1ZAP2MId<0WFM300<0W6I200@0VVA009US?`2IHcl0TEde0P2CGcL0102FHSX0UF4i09EQ
>@2BGSH208mK<`<0SUXb00<0U60h09EQ>@2EHCT00`2GHc/0102FHSX0U60h09AP>02BGSH2091L=004
08IF;P22ERl0PUH_07mE;@80MDdY00@0N4d[09QYB`2HJD/0^hQ]0P34S7<0102bMfD0WfEA09mUD@2A
E3l209AF?`040:MUB`2hMEX0^7EJ0;agFP80]6m@00@0VU0e0:USB@2YHdT0Z6aD0P2<GD/0101XBT00
DDI40556A015BU0203i<E`0403Y;D`0oBE<0?dUC04=5DP<0ATAB0P14AU<01010BU@0?4eE03a=E@12
C5D204A=EP0404E>E`18CeT0B4mI04U@FP80C51K0`1=DE`304aCG@<0Be9L101:DE/204U@FP0404Y@
G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13BED00`13BED40457D`D0@TQD0`13BED404I=E`D0
AdiH0`18CeT00`17CUP0B4mI04Q?F@0604U@FP0304Q?F@17CUP0AdiH00@0ATeG1@17CUP?04Q?F@04
04U=F018C5L0B4aG04M;EP80ATYE0`15BE@304I:E@@0BTiI0`19CEP504Q<E`<0B4eC00@0B4aE04I=
E@16CED0ATeG0P15CUL01015CEP0A4iH04A>F011CeX204A>E`40ATmF0@1:Ce<10?ooo`006019EUl<
04QEGQ00AeAM1@18EEh304UFG`<0BeQQ00@0BUMP04UFG`19EUl0BEIO0P1:Ef0304]HH@0404YGH01;
F640BeQQ04aIHP<0CEYS0P1<FF80101;F640BUMP04YGH01<FF8204iKI00404mIHP1>Eeh0CUMN0499
C`809b/^00@051DG01HG6@0F5aT06AHA0P0J5A40100M4a407Q@B01hD4P0N518201`G4`0401PI4`0G
6Q<05aXC01PK50806AXB00@05aL?01/D300K50`08AH@0P0[6Q<0100/6A<0:AD>02TE3P14:B0206I8
?@0406i??@1bCcT0LTli079@=@80Jdda00@0KeHh06eE>@1]ECT0Ke@g0P1]DSD0101TBBh0DSHN058f
7P1090l203XM2`04048U5@1=<200CC0P054[7P80DRhP00@0EC4S05<b8@1C<R40Dc8O0P1F=b<0101M
?RT0HTH`0696<01I?BL205a0:P0405m0:`1[C3L0Jd`g06a<=`80Jd/f0`1lFDD0102OMFD0V6mM09Q_
G@2HKEX209MZE@0408mRB`2DITh0U6I>09maE`80XG9F00@0Wg9E09e`D`2ML5<0PeHi0P23EC/0102<
GTH0RUa408YLA01oDCX205/]5P0406Tg801jARh0NTH^099J?`80XVU;00@0Z71?0:aeEP2/MEH0/GiO
0P2gQVP0102jS6P0^H]W0;V;I`2eQF420:mmFP040:iiE`2_MeH0[gMF0:aaD@80YFY:00@0W5a009aO
@P2LGd80Wf950P2QJ4X0102WL540[WUI0:iiF@2cPF020;N5I0040;>7I02/P5d0[81M0:MjF@80ZGaK
00@0[h9S0;>6J@2cQVT0]XU/0P2iS6l0102dQFT0/H9V0;62IP2]OV820:UjGP040:IgF`2SM5P0XgAH
0:5bEP80X75E00@0WVmC09a]D@2LKE40W6eA0P2OL580102TMUD0ZWaK0:YlF`2]Oed20:j0GP040;>2
I02eQ6H0]HAV0:YiF@80]8=S00@0aIAc0;^;I`2kRfL0[7aH0P2YNED0102WME80ZWQE0:YhE@2/NUL3
0:]iEP80[7YG00@0ZWQE0:IdD@2VM540[GUD0P2cOeX0102`O5L0Z79>0:QbCP2ZLTl20;AlF@040;Ml
FP2_M580[gAB0:]eD@80YW1<00<0XF]709mYA@2OJDD00`2OJDD0102KID40Uf4m09MQ?@2AGCD209AP
>0<0Uf<k00@0Telg091L=02@G3@0SUXb0P2=FC400`2DH3P0UF4i09EQ>@0409MS>`80UF4i00@0Telg
095M=@2AGCD0QEP_0P20E2d0101lEB`0MDdY07E=:@1fCR/209]^C`040;j;L030RW00`8Y`0;QoKP80
Z6eM00@0Vei;09IH@@2FF440WEe40P2_KU80102iMeL0]g9C0;MbD`2MDSD20:aTB`040:Y/D`2=GDT0
SEe906Q:?`80DDI400@0ATU@041<F010C5P0?4eE0P10BU@01014AU<0AdEC04M5D`17AE<204E7E004
045;E@0lCED0?4eE045;E080A4eF00@0ADiG04Q?F@18CeT0BU5K0P1<D5/304eAG0H0C5=M0`1;DU`3
04YAF`0404Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13BED00`13BED60498E0H0@dUE1016
CEL504M>F0H0B4mI00@0BE1J04YAF`1:DE/0BU5K0P19D5X304Q?F@@0ATeG1@17CUP?04Q?F@0404U=
F018C5L0B4aG04M;EPH0ATYE0P17BeH00`1;CeX0BTiI04Y>F@0304Y>F@<0BDeH0`18C5L304Q=D`04
04Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH04A>F014CUP0@DmJ0P14CUL104I?EP40BTmC0@3oool0
01P0BEIO3018EEh@04MDG@D0B5EN0`19EUl304]HH@0404YGH019EUl0BEIO04UFG`80BUMP0`1;F640
101:Ef00BeQQ04]HH@1<FF8304eJH`80C5UR00@0BeQQ04YGH01:Ef00BeQQ0P1>Ff@0101?FF80CEIM
04eFG@11B4h202L[;P0401DF600I6Q`06AXL01TF4@806QDA00@07A<A01hD4P0N51807Q@B0P0L5a<0
100H6A<05aXC01LJ4`0D6Q0201LJ4@0401PH3P0L5@d071D=02DH4080;QdF00@0;ATB02P?2P0X3`X0
@bHP0P1SAC/0101[BS/0KT/e06i;=@1^C3<206U;;`0406iD>@1/EcX0K5Lj06eC>080ID/`00@0Fd4X
04hd7@1>=1d0>b0;0P0]4P40100`5@@0?20B03`P4P188aX204dX7`0405<a9@1F=RT0ESHY05Pk:`80
GT<b00@0Id`i06iE@@1^ED40KUE10P1cFTH0101[DSh0LeQ507=HA@1eF4H206]>?004079C@@1dED<0
M5E309A]GP80TfaK00@0Vg9P09]`G@2KL5d0TFA=0P2AHd/0102HJU00VFY>09UZCP2JKE020:9eF004
08UL?`24ES`0Q5Hl08EG?`80PU@l00@0Oe4j05h`6@1N<1T0IS@M0P1iABd0102DG440Y6]=0:A[C@2Z
LU420:efE`040;5nG`2fQFL0]XEW0;R:IP80]hUU00@0]HEQ0;:0G@2bP5d0]7mM0P2dO5/0102[L500
WV=309iS@`2EF3/209YM@0040:1UA`2WKU00Yfi@0:igF080]7mO00@0]XAS0;N6I@2gQVD0/XMT0P2Z
Oe`0102VNeT0[7mN0:aoGP2bQFH30;>6J@80]XM[00@0ZgaP0:UjGP2YNUh0YGIJ0P2QLUH0102MKU80
W6eA09a]D@2MKU8209m`E0040:5bEP2RLeL0XW=G0:AeF@80Z7UK00@0[GmN0:n1H02_PF00[X1N0P2/
OU`0102/OEl0]8EW0;B5I`2UMeH20:]mG0040;Z<JP2_PEd0[h5M0:YlF080[h5M00@0[G]H0;1nF`2`
OU/0/7iK0P2/NUL0102[NEH0[WaI0:ilF@2]NeP20:UgE0040:eiE02bOUT0/WiI0;9nF@80[GMC00@0
[7AA0;1hE@2`N5D0/gQF0P2_M5800`2[ME40Yg5=0:MaC@030:9/B0040:=]B@2QJdL0XF]709aV@P80
UV0l00<0TEde09AP>02DH3P00`2HI3`0102DH3P0T5`d091L=02=FC4208aH<004095M=@2BGSH0TUhf
09AP>080UF4i00@0U60h09=O=`2CGcL0TEde0P2?Fc<01027FS40OU@/07iD;01lEB`207I@:`0407YB
;`2PLe@0X7=D0;n=LP80^XAZ00@0]ge_0:U_H@2YKf40WfAB0P2KGTT0102NGTD0[VeA0:i]D@2jN5P2
0;AbD00409Y?<P2ZHdL0ZV=70:]]E080T61<00@0KDe205I8A@1FB4D0AdY?0P10CEH0100nCeL0@TaF
049<EP16B5D304Q6E080ATQE00@0@TaF03e>EP0mCUH0@D]D0P13C5D01015CUL0BE1J04U@FP1:DE/3
04eAG080C51K101=E5h504aCG@<0Be9L00@0BU1L04U?F`19Ce/0AdeI0P15BeL00`14BUH0@dUE04=9
E@0304=9E@H0A4YF1P13BED404I=E`D0AdiH1018CeT204U@FPH0BU5K00<0BE1J04Q?F@18CeT01016
CEL504M>F0l0B4mI00@0BTiI04U=F019CEP0Ad]F1@16BUD304M;EP<0BdmJ101:CUT504U=F0<0BDiD
00@0BDeF04M>EP17CUH0AdiH0P16CeP01016CUT0ADmI04E?F@11CeX204A>E`40ATmF0@1:Ce<10?oo
o`006019EUl<04QEGQ00AeAM1@18EEh304UFG`<0BeQQ00@0BUMP04UFG`19EUl0BEIO0P1:Ef0304]H
H@0404YGH01;F640BeQQ04aIHP<0CEYS0P1<FF80101;F640BUMP04YGH01;F64204iKI00404mIHP1=
EUd0CEIM0458CP809b/^00@05AHH01TJ700I6Q`06AHA0P0J5A40100M4a407Q@B01hD4P0N518201`G
4`0401PI4`0G6Q<05aXC01@J40805aXA00@061P>01`E3@0L5@d09AP@0P0^7AH0100]6A80:0l:02P?
2P139R0206=5>`0406]:>`1^BcD0KT/e06i<<`80JD/_00@0KU@i06aG>P1/EcX0KE<h0P1UBc00101K
@BP0CS@M04hd7@0k80/202dB0@04030E100l8180?20B04PS6P80CBPO00@0Dc4U05Hf:@1F=RT0F3/[
0P1N@c80101WC3T0KUE106iE@@1^ED4207=JAP0406]B?P1cF4D0LeQ507EHAP80Jdhl00@0LU=107AE
@`1dED<0U6eN0P2CK5/0102KLV00Vg1M09]`G@2AI4d2095SB`0409QZD02IJTh0VFY>09Y]D080XWEH
00@0RE`o08AF?024ES`0QELo0P22E3`0101oDCX0GS0I05h`6@1V=1d207U5;@0409AL@@2TJdd0Y6]=
0:YbD@80[GIG00@0/GiO0;J5I`2fQFL0^8YV0P2gRFD0102eQF40/X1M0;:0G@2dOed20;AlF`040:]`
D02NHd<0WV=309EH>`80VUe000@0X6E70:M^D02WKU00[WMH0P2dOel0102fQ6<0]hIU0;N6I@2bQf@2
0:YoG0040:IkF@2/Oeh0[7mN0;:5IP<0/hIY0P2fQf/0102[O600ZGYN0:UjGP2UMUX20:5bEP0409e^
DP2LKE40W6eA09e^DP80Wg1D00@0XG9F0:9cE`2RLeL0Y7EI0P2XNE/0102]Oeh0[h5P0:n1H02^P5h2
0:anG0040:amG`2dQFL0]8EW0:EgEP80ZgeL00@0^XaZ0:n1G@2_PEd0ZWaH0P2_PEd0102]NeP0/7iK
0;1nF`2`OU/20:ajE`040:]iEP2^O5T0[WaI0:ekF080ZGMD00@0[GUD0;9nF@2bOUT0/WiI0P2]Me<0
102/M540/7QE0;1hE@2cN5H20:mdDP030:]eD@2WLDd0Yg5=00<0XVa800@0Xfe90:5[A`2QJdL0W6I2
0P2FH3`00`2AGCD0U60h09AP>00309QT?00409AP>02@G3@0T5`d08eI<@80S5P`00@0TEde099N=P2B
GSH0U60h0P2EHCT0102DH3P0Telg09=O=`2AGCD208mK<`0408MJ<@1nE2`0OU@/07aE;080MU0[00@0
NU8_0:1cE02PLe@0_heb0P2jQ6X0102gOFl0ZFmQ0:U_H@2OI58209]NB@0409iNA@2^KE40[VeA0;Yh
F080]79@00@0VTlb0:YSA`2ZHdL0ZfeD0P2@H4`0101]CD80ETQ505I8A@17BTl2041=EP0403i?E`12
C5H0@TaF04I8E@<0B4ID0P16B5D01012C5H0?DiF03e>EP11Be@204=<E@0404E>E`19D5X0BE1J04YA
F`<0CE5L0P1<D5/404eDGPD0C5=M0`1;DU`0101:D5`0BDmK04U?F`17CET204E;E`0304A:EP13BED0
@dUE00<0@dUE1P14BUH604=9E@@0ATeG1@17CUP404Q?F@80BE1J1P1:DE/00`19D5X0B4mI04Q?F@04
04I=E`D0AdiH3`18CeT0101:CUT0BDeH04U=F017BeH504I:E@<0Ad]F0`1;CeX404Y>F@D0BDeH0`19
CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80A4iG0@16CeH104Y?D`40
oooo000H04UFG``0B5EN4017E5d504QEGP<0BEIO0`1;F640101:Ef00BEIO04UFG`19EUl204YGH0<0
BeQQ00@0BUMP04]HH@1;F640C5UR0`1=FV<204aIHP0404]HH@1:Ef00BUMP04]HH@80CEYS00@0CUQQ
04aEG01<EE`0@DQ>0P0W:bh0100F5aT06Q/M01XK7@0I5Q4201XE4@0401dC4@0N51807Q@B01hD4P80
71LC00@061TC01LJ4`0G6Q<04aT?0P0G6Q40100H60h07AH>01dF3P0W6Q82030M5P0402lH4P0V3@P0
9Pd8044R7@80HD0g00@0J4Lh06Y7<`1ZAc<0JdT`0P1XBRh0101^ESX0KUTn06iI?P1_ECX20616;@04
05@j8@1<<A`0C34L03/P3@809``000@090P002lC1@0_4`D0@ahF0P199ah0101C<BL0Fc/`05/k<01N
@S@206I<>`0406mHAP1hHDl0N65?07YSD@80RG9P00@0Q6eK08I/F`26K5/0P6IE0P1]DT40101bEDD0
OF1@07ePD029HeH208mXF@0409ecH`2OM640WgAQ09AWD080T69:00@0U6E;09=TB02CI4P0UfY=0P2U
N5/0102>HD@0QELm08EG?@21Dc/207a>=P0407i@>@1P<Q/0H38K06<a6P80N4@/00@0U5a10:I]C`2V
KDl0[7AC0P2^MeP00`2aOUl0]XEW0;J5I`030;F7H`040;F5H@2ePf00]H=P0;V4HP80^81O00@0ZVm?
09UN?P2IGSh0TeHi0P2IG3l0102RIdT0Zg9D0:]bE02bOEd20;F3HP<0]hIU00@0[XIR0:UnF`2YOU/0
YgaJ0P2^Pf400`2eR6T0]8EY0;B5J@030;:3I`040:AeF@2SM5P0XgAH0:5bEP80WVmC00@0W6eA09i_
D`2NKe<0Y7EI0P2YNUh0102/OF40[GiR0:enHP2`PFD20;>4IP040;F7IP2bQ6<0/XAS0:anG080YWQF
00@0XgAF0:n0HP2_P680X79A0P2QLe80102[OE/0XG=?0:5cC`2UMe<20;:4H0040;R6H`2kRFH0^hUV
0;R6H`80/X1M00@0/7iK0;B2G`2dPUl0/h5N0P2_OEX0102`O5L0/GeH0;5mF02cOeX20;9lF0040:mg
E02[Le00Zg=@0:ebD080[gAB00<0ZgEA0:MaC@2WLDd00`2TKTX0102UKd/0Xfe90:=]B@2LIT8209EO
>`04095M=@2EHCT0UF4i09QT?080VFDm00@0UF4i091L=02@G3@0SETa0P2;Ebl0102>FS80T5`d091L
=02AGCD3099N=P80T5`d00@0SUXb08eI<@2=FC40Rf0f0P21Ebl0101nEbh0NE<^07UC;P1mEC820:=h
F0040;n=LP2bOf@0/WmT0;AjKP80YfeO00@0XFIF09mRC@2OHTd0Wf580P2aL5@0102kNUX0/W1>0;9`
CP2GBRh20:QQA@040:i^E@2CHdl0Tf=?075AAP80F4Y700@0BdaA045;E@11BeD0?e1H0P13CEL01017
BEH0BDME04U7E@19AeD204I8E@04049<EP0nCeL0?TmG045;E080@daE00@0ATmH04U@FP19D5X0BU5K
0P1=DE`304a@F`@0CUEO0`1=E5h504aCG@0404Y@G019Ce/0BDmK04M=F@80AD]G00<0A4YF04=9E@13
BED00`13BED304E;E`T0A4YF1016CEL504M>F0@0B4mI0P19D5X604YAF`0304U@FP18CeT0B4mI00@0
ATeG1@17CUP?04Q?F@0404Y>F@19CEP0BDeH04Q<E`80Ad]F0`16BUD304M;EP@0BdmJ0`1:CUT504U=
F0<0BDiD00@0BDeF04M>EP17CUH0AdiH0P16CeP01016CUT0ADmI04E?F@11CeX204A>E`40ATmF0@1:
Ce<10?ooo`003019EUl<04YGH0<0BEIO1018EEhA04MDG@`0B5EN301:Ef0304UFG`@0BUMP1@1;F640
101:EV@0BUIR04YFHP1@F6<204eDG00403/n@`0Q82D08B0U01LE5P807QXK300J5A40100M6Q4061@;
01PD2`0U7QH2028G4@0401l@2@0g91h0=b@N034L5`8080P400@0AR/H06=8=@1SB3D0KE8o0P1TBC@0
101SBC80IT`e06I<=@1VC3<206E;<P04065><01RCS<0HThc04hj8@80?BPE00@0>B<E02PB1`0X4PL0
7`P20P0`61@0101;9Qd0H3he060n=@1C<RT204lb:P04069:@01]ETh0KEI>07=QE`80N6IL00@0OgAR
07iaH01nLF00NfiN0P1aHE80101LC3l0F4Hj05Q6>P1JB3h206ECB@04081JC`2;HeT0Rf=I08mUE`80
S6=?00@0T6=<08eQAP2=HDH0S5m00P2DITD0102MJE40VFE=09UUC@2BH4T209QXD@040754;@1hCSP0
N4hh0717<@80J3lY00@0I38K08A@>024D3P0Te]00P2OITP0102WKdh0Yg1A0:M`D@2aOUl20;N6J004
0;>5H`2gRFL0]hUW0;b;JP80^hUX00@0]H1P0:]bE02[LU@0XVM90P2KH480102FESX0XfQ:0:=XBP2X
Le<20:5cD@040:5fD`2/PEh0[85N0;:6H`80[h=P00@0ZGiK0:=gE02SMe@0[7mN0P2^P5l0102]OV00
[G]P0:ekH02ULeP20:ilH@040:=`F`2TLE`0Y75L0:9bG080Y7AM00@0WFmG0:EiGP2UNEh0XWIK0P2Y
OF80102ZOVD0[H=Z0:f3JP2TOF020:R1I0040:1gF@2]OV80[GiR0:YdF080ZW9G00@0Z7]N09Y]CP2J
KDh0YgYK0P2[OUd0102VNEP0ZgmL0:]oG02[Oe`20;24H@040:YjE02[NeD0Zg]E0:alEP80[GeG00<0
[WiH0:moF@2_OeT00`2`P5X0102`NUH0]GmK0;EoF`2cOET20:]eD@040:egD`2eOe/0]GmK0;=mF@80
ZWA@00@0X6a40:9^AP2RKTH0Y7180P2QKDD0102LJ400VVHn09YV?P2KIcl209eY@@0409IR>P2IICd0
VFDm09]W?`80VFDm00@0U60h08mK<`2?Fc<0SETa0P2<F300102;Ebl0Se/c08mK<`2CGcL209EQ>@03
09=O=`2@G3@0T5`d00<0SUXb00@0QeXa08=F;`23ERl0SF0i0P1oDBl01026F3L0^ha^0;^<KP2iRVh2
0:n0I0040;=eJ02YJe`0ZF]L0:Y]FP80YVQA00@0WF130;9cDP2bLe80`X9O0P2eME40102?BR/0Z6=6
0:QSAP2/JU8209AMB@0407E=@@1GA3h0Ed@n04E5A`80A4mC00@0@deI04E;F`15Be/0BTQM0P1<Aeh0
101>AUd0BdQI04]8F@18BUH204M<DP@0AdiH0P18CeT0101:DE/0CEAN04eDGP1?EV02051GH@0304iE
G`1?EV00CeIP00H0D5MQ00@0CeIP04iEG`1>EEl0BDmK0P18CUX01017CET0ATaH04I<F015BeL204I<
F0<0AdeI0`18CUX304M=F@0304I<F015BeL0AD]G00<0AD]G3018CeT01019D5X0BU5K04YAF`1;DU`3
04aCG@80Be9L00<0BU5K04U@FP19D5X01017CUP304Q?F@D0BE1J301:DE/<04U@FP@0BU5K0`19D5X5
04Q?F@<0BTmE00@0BTiG04Q?E`18CeL0B4mI0P17D5T01017CeX0AU1J04I@FP11CeX204E?F040B55H
0@1<DED10?ooo`003019EUl<04YGH0<0BEIO1018EEhA04MDG@`0B5EN301:Ef0304UFG`@0BUMP1@1;
F640101:EV@0BUIR04YFHP1@F6<204eDG00403/n@`0Q82D08B0U01LE5P807QXK300J5A40100M6Q40
61@;01PD2`0U7QH2028G4@0401l@2@0g91h0=b@N034L5`8080P400@0AR/H06=8=@1SB3D0KE8o0P1T
BC@0101SBC80IT`e06I<=@1VC3<206E;<P04065><01RCS<0HThc04hj8@80?BPE00@0>B<E02PB1`0X
4PL07`P20P0`61@0101;9Qd0H3he060n=@1C<RT204lb:P04069:@01]ETh0KEI>07=QE`80N6IL00@0
OgAR07iaH01nLF00NfiN0P1aHE80101LC3l0F4Hj05Q6>P1JB3h206ECB@04081JC`2;HeT0Rf=I08mU
E`80S6=?00@0T6=<08eQAP2=HDH0S5m00P2DITD0102MJE40VFE=09UUC@2BH4T209QXD@040754;@1h
CSP0N4hh0717<@80J3lY00@0I38K08A@>024D3P0Te]00P2OITP0102WKdh0Yg1A0:M`D@2aOUl20;N6
J0040;>5H`2gRFL0]hUW0;b;JP80^hUX00@0]H1P0:]bE02[LU@0XVM90P2KH480102FESX0XfQ:0:=X
BP2XLe<20:5cD@040:5fD`2/PEh0[85N0;:6H`80[h=P00@0ZGiK0:=gE02SMe@0[7mN0P2^P5l0102]
OV00[G]P0:ekH02ULeP20:ilH@040:=`F`2TLE`0Y75L0:9bG080Y7AM00@0WFmG0:EiGP2UNEh0XWIK
0P2YOF80102ZOVD0[H=Z0:f3JP2TOF020:R1I0040:1gF@2]OV80[GiR0:YdF080ZW9G00@0Z7]N09Y]
CP2JKDh0YgYK0P2[OUd0102VNEP0ZgmL0:]oG02[Oe`20;24H@040:YjE02[NeD0Zg]E0:alEP80[GeG
00<0[WiH0:moF@2_OeT00`2`P5X0102`NUH0]GmK0;EoF`2cOET20:]eD@040:egD`2eOe/0]GmK0;=m
F@80ZWA@00@0X6a40:9^AP2RKTH0Y7180P2QKDD0102LJ400VVHn09YV?P2KIcl209eY@@0409IR>P2I
ICd0VFDm09]W?`80VFDm00@0U60h08mK<`2?Fc<0SETa0P2<F300102;Ebl0Se/c08mK<`2CGcL209EQ
>@0309=O=`2@G3@0T5`d00<0SUXb00@0QeXa08=F;`23ERl0SF0i0P1oDBl01026F3L0^ha^0;^<KP2i
RVh20:n0I0040;=eJ02YJe`0ZF]L0:Y]FP80YVQA00@0WF130;9cDP2bLe80`X9O0P2eME40102?BR/0
Z6=60:QSAP2/JU8209AMB@0407E=@@1GA3h0Ed@n04E5A`80A4mC00@0@deI04E;F`15Be/0BTQM0P1<
Aeh0101>AUd0BdQI04]8F@18BUH204M<DP@0AdiH0P18CeT0101:DE/0CEAN04eDGP1?EV02051GH@03
04iEG`1?EV00CeIP00H0D5MQ00@0CeIP04iEG`1>EEl0BDmK0P18CUX01017CET0ATaH04I<F015BeL2
04I<F0<0AdeI0`18CUX304M=F@0304I<F015BeL0AD]G00<0AD]G3018CeT01019D5X0BU5K04YAF`1;
DU`304aCG@80Be9L00<0BU5K04U@FP19D5X01017CUP304Q?F@D0BE1J301:DE/<04U@FP@0BU5K0`19
D5X504Q?F@<0BTmE00@0BTiG04Q?E`18CeL0B4mI0P17D5T01017CeX0AU1J04I@FP11CeX204E?F040
B55H0@1<DED10?ooo`003019EUl<04YGH0<0BEIO1018EEhA04MDG@`0B5EN301:Ef0304UFG`@0BUMP
1@1;F640101;EfD0BUIR04YFHP1>EV4204YAF@0403Lj?`0O7b407alQ01DC508071PG300J5A40100K
60l06AD<01TE300S71@202<H4P0402LH4@0e8Q`0=B8L02XE40807`L300@0AR/H0656<`1QAS<0Je0m
0`1SB3<206M==P0406I<=@1UBc80ID/b05a9:`80HDdb00@0DShU048]6012;AP0>R@F0P0W4@H0100N
1`00;QLA02hG4@1F>30206Q;@`0405Hk<P1<=Bd0C3D]05]7?P80J5M?00@0NV]T08MjLP27NW80NgEU
0P1]IUH0101ECSh0CdDi04m5>@0d:R0202dP5`0404<f;P0f9b00=RLP05m4>@80MUY>00@0Q6MI08QU
DP28IE80QEm:0P28GTD0102@I4T0UfY=09MZC@2HI4`209e[DP0409e]EP2PLUX0X79J07M:<`80NDli
00@0LDPb06U0:P1Y@2X0Hc4J0P23CcL0102BFSl0WVE709iUA`2VKTd20:I_D0040;1mGP2eQ6H0]HAV
0;>5H`80]HMU00@0^8MV0;R6I@2hQVD0]7mO0P2[LU@0102QITP0VEi009UN@02HF3`20:AYB`040:Ud
E02TMU@0Y7ID0:EjE`80[X=P00@0/XIS0:n1G`2_PEl0Yg]H0P2QME80102WNUT0Z7YI0:QjF@2WN5X2
0:MeFP0409aYCP2MJTl0WFY?0:9^F080XFeG00@0WF]B0:1^E@2PKUD0YGIL0P2VMed0102XNeh0YGQK
0:EhF`2PM5T20:9iG@0409efF@2UOUl0YGiO09edE080YWMI00@0YW=D0:mhFP2_N5X0[GaL0P2TLe<0
102[NUX0Z7MF0:QgEP2SLU420:]kE`040:YjE02XN580Z7QB0:]iE080[WaG00@0/GmJ0;>1G02cPE`0
/X1K0P2_OEP0102ZN5<0Z7IA0:QfD@2^N5@20;=mF@040;9lF02[ME40ZgEA0:egD`80]GmK00<0]7iJ
0:]eD@2[ME40102TL4P20:9^AP0309iZ@P2KIcl0VfLo00<0VFDm0`2LJ400102KIcl0UV8j09IR>P2A
GCD308eI<@80SUXb00@0S5P`091L=02@G3@0U60h0P2EHCT00`2CGcL0T5`d091L=00308mK<`0408QK
<P22EBh0PUD^08]N=`80Oe4_00@0Sf500<6BM031TW@0_i1d0P2aPVH0102`MFD0ZVeK0:Y]F`2[KU/2
0:MYDP040:1SA`2eMUD0]GIE0<B5HP80]gUD00@0U4l`0:UTA`2YI4L0[6Y@0P2GH4`0101kDdL0GTU4
05i9A018ATL204A=DP04045=F@14Be/0A4]K04M:G@80BTQN00@0BTQM04Y9F@1:BET0Ad]F0P15C583
04I=E`0404M>F018CeT0B4mI04YAF`80C5=M00@0CeIP051GH@1@Ef40CEAN0P1>EEl0101?EV00D5MQ
051GH@1@Ef4204mFH00404iEG`1=E5h0CEAN04Y@G080BDmK00<0AdeI04I<F016C5P00`15BeL304I<
F0<0B4iJ1`17CET204I<F0`0B4mI00<0BE1J04YAF`1:DE/01P1;DU`00`1:DE/0BE1J04U@FP0304M>
F0<0B4mI1019D5X>04YAF``0BE1J101:DE/304U@FPD0B4mI0`1:CeD0101:CUL0B4mG04Q?E`18CeT2
04M@F@0404M?FP16D5X0AU1J045?FP80ADmH0@17D5L104]@E040oooo000<04UFG``0BUMP0`19EUl4
04QEGQ40AeAM3018EEh<04YGH0<0BEIO101:Ef0504]HH@0404aHI01:EV80BUIR04aEGP80ADaD00@0
<C@i01/K7@0K6ad04a4B0P0J5QD<01XE4@0401TF3`0I50h06A@>01hG3`808aPB00@0=2DN03<P6P0c
81X08@d60P0O208010109A@0Fd0_05]0;`1SB3D205]0;@0405a1;01PARl0H4H_05i4;@80Fd4Z00@0
Ed8U061;<01PBc00F48Z0P1;=B001011:QX0;1@802`D200O1`0202T@2@0403<L5P1==S00CCH`04Xg
<080DT4j00@0I5M?06eTG@1]I5d0NG9Z0P1mN780101KGE00?Clb03do<P0^;R82038b:004038_9P0`
:bD0<2/U03Xe;`80>3<_00@0GEE:05eCA`1MDdL0I5A50P1dGDd0101fFDL0MU8l07IB?021Ed4208UL
A@0409ESBP2NK5<0WVaC0:=cG080YWQP00@0Oe8k07Q?>@1hCcT0KdH`0P1W@BX0101S<a`0Pe4h08=A
>02AFcl209iUA`040:I^C@2UL500YG1@0:imG@80/X=U00@0]8IT0;>5H`2cQF<0]8=R0P2dPV40102c
OUh0Zg9D0:]bE02PIDL209IK?@0409YM@02SJT`0XfY<0:QfE@80YWQF00@0Z7aI0:n3H02_Pf00/89P
0P2/NeX0102RMU<0Wg=@09mcD02TMUD20:9dD`040:AcE@2SL5D0Xg1E09MTB@80TEi30`2NITd0102H
HTP0W6I<09aVC02_O6420:=`D`040;60HP2QL580XG1B0:ImG`80Z7mQ00@0Yh1Q0:b8IP2/R6H0Z81M
0P2]P5l0102`OUd0^hAU0;^4I@2^NEL20:YeD`040:UdDP2SKdX0Xfm:0:9^B@80ZgM@00@0[GUB0:A`
B02TL4P0[7YE0P2]NeH0102_OEP0/7iI0;1nF@2_OEP20:ekEP040:YhD`2XMU40Z7IA0:afDP<0/7YF
0P2/MU80102^N5@0]X1L0;J0G02eOe/20:egD`040:QdC02ULDT0YG590:9^AP80Wf]300@0WVY209]W
?`2KIcl0V6@l0P2EHCT0102SKdL0X6a40:1/A02JISh209=O=`0408iJ<P2<F300S5P`08eI<@80T5`d
00@0ReXb08mN=P2?GSH0Tf8j0P2DHc/00`2BHCT0Sehf08mN=P0308iM=@0408UL<`20Db`0P5</08QK
=080Oe4_00@0WFm>0<ZKO@3:Vgd0aYMk0P2bPfL0102/LEl0YV]I0:I[F@2ZKEP20:Q/E0040:IYC@2h
Ne`0^7]L0<F6I@80^7UF00@0VE@c0:YUAP2ZIDH0[6Y@0P2MI5001024FTh0HTe8069=B01:ATL204A;
D@0403i<E`0oBeT0?d]I045:F`80@TUL00@0@TUK04=9E`13BEL0@T]D0P12Be0404I=E`80AdiH00@0
BE1J04aCG@1<Ded0CUEO0P1?EV000`1<Ded0CEAN04eDGP0604iEG`0404eDGP1<Ded0C5=M04]AG@80
BU1L00<0B4iJ04I<F016C5P01P15BeL304I<F0L0AdeI0P18CUX<04M>F00404Q?F@19D5X0BE1J04YA
F`<0Be9L0P1:DE/01019D5X0B4mI04Q?F@17CUP504Q?F@<0BE1J0`1:DE/<04U@FP`0B4mI1019D5X3
04Q?F@D0AdiH0`19CU@01019CEH0AdiF04M>EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80ADmH
0@16CeH104Y?D`40oooo000<04UFG``0BUMP0`19EUl404QEGQ40AeAM3018EEh<04YGH0<0BEIO101:
Ef0504]HH@0404aHI01:EV80BUIR04aEGP80ADaD00@0<C@i01/K7@0K6ad04a4B0P0J5QD<01XE4@04
01TF3`0I50h06A@>01hG3`808aPB00@0=2DN03<P6P0c81X08@d60P0O208010109A@0Fd0_05]0;`1S
B3D205]0;@0405a1;01PARl0H4H_05i4;@80Fd4Z00@0Ed8U061;<01PBc00F48Z0P1;=B001011:QX0
;1@802`D200O1`0202T@2@0403<L5P1==S00CCH`04Xg<080DT4j00@0I5M?06eTG@1]I5d0NG9Z0P1m
N780101KGE00?Clb03do<P0^;R82038b:004038_9P0`:bD0<2/U03Xe;`80>3<_00@0GEE:05eCA`1M
DdL0I5A50P1dGDd0101fFDL0MU8l07IB?021Ed4208ULA@0409ESBP2NK5<0WVaC0:=cG080YWQP00@0
Oe8k07Q?>@1hCcT0KdH`0P1W@BX0101S<a`0Pe4h08=A>02AFcl209iUA`040:I^C@2UL500YG1@0:im
G@80/X=U00@0]8IT0;>5H`2cQF<0]8=R0P2dPV40102cOUh0Zg9D0:]bE02PIDL209IK?@0409YM@02S
JT`0XfY<0:QfE@80YWQF00@0Z7aI0:n3H02_Pf00/89P0P2/NeX0102RMU<0Wg=@09mcD02TMUD20:9d
D`040:AcE@2SL5D0Xg1E09MTB@80TEi30`2NITd0102HHTP0W6I<09aVC02_O6420:=`D`040;60HP2Q
L580XG1B0:ImG`80Z7mQ00@0Yh1Q0:b8IP2/R6H0Z81M0P2]P5l0102`OUd0^hAU0;^4I@2^NEL20:Ye
D`040:UdDP2SKdX0Xfm:0:9^B@80ZgM@00@0[GUB0:A`B02TL4P0[7YE0P2]NeH0102_OEP0/7iI0;1n
F@2_OEP20:ekEP040:YhD`2XMU40Z7IA0:afDP<0/7YF0P2/MU80102^N5@0]X1L0;J0G02eOe/20:eg
D`040:QdC02ULDT0YG590:9^AP80Wf]300@0WVY209]W?`2KIcl0V6@l0P2EHCT0102SKdL0X6a40:1/
A02JISh209=O=`0408iJ<P2<F300S5P`08eI<@80T5`d00@0ReXb08mN=P2?GSH0Tf8j0P2DHc/00`2B
HCT0Sehf08mN=P0308iM=@0408UL<`20Db`0P5</08QK=080Oe4_00@0WFm>0<ZKO@3:Vgd0aYMk0P2b
PfL0102/LEl0YV]I0:I[F@2ZKEP20:Q/E0040:IYC@2hNe`0^7]L0<F6I@80^7UF00@0VE@c0:YUAP2Z
IDH0[6Y@0P2MI5001024FTh0HTe8069=B01:ATL204A;D@0403i<E`0oBeT0?d]I045:F`80@TUL00@0
@TUK04=9E`13BEL0@T]D0P12Be0404I=E`80AdiH00@0BE1J04aCG@1<Ded0CUEO0P1?EV000`1<Ded0
CEAN04eDGP0604iEG`0404eDGP1<Ded0C5=M04]AG@80BU1L00<0B4iJ04I<F016C5P01P15BeL304I<
F0L0AdeI0P18CUX<04M>F00404Q?F@19D5X0BE1J04YAF`<0Be9L0P1:DE/01019D5X0B4mI04Q?F@17
CUP504Q?F@<0BE1J0`1:DE/<04U@FP`0B4mI1019D5X304Q?F@D0AdiH0`19CU@01019CEH0AdiF04M>
EP17CUP204I?F00404I>F@15CeT0ADmI045?FP80ADmH0@16CeH104Y?D`40oooo000K04UFG`@0B5EN
4@17E5d<04QEGPl0BEIO101:Ef0504]HH@0404aHI01:Ef00BUMP04YCG080?dI<00@0:blb01PH600H
61P04Q4?0P0I5A8<01XE4@0401LD3`0J5A406QDA01XC3@808aPB00@0@C8[038O600b7aP06`L00P0Q
2P8010109QD0Ecd/05Lm;01L@S42058i9@0405<j9P1G?RP0EchX05<j9080CcLO00@0DC`Q05]6:`1K
AR/0FT@/0P1D?RT0101>=bD0>R<E03XS5@0V3P8202@<000401h;1@0c8Q/0<b8K03PY9080C44k00@0
JV=M071]IP1`KFH0KfiY0P1TIEl01015CD80:S8W02Xb9`0f>c@202L/9@0402PZ9@0N81/07R0K00`<
2P809bLU00@0>T8g03Hi;P0f>Bh0CDTm0P21LfH0102=MVH0MEI407EFA01cCCX207U>>`0408mMA02;
G480Rea2099TC080XgMN00@0Rf5907]B?01kDS`0J48[0P1R?2D0101W=b00QU@k08ID>`2CGD4209mX
B@040:QbD02WLU80Yg9B0;1oG`80/hAV00@0]8IT0;63H@2aPf40/7mN0P2aOeh0102aO5`0Zg9D0:]b
E02PIDL209IK?@0409eRA02TKDh0Y6e>0:QfE@80YgUG00@0Yg]H0:YnF`2ZOU/0ZGQG0P2TLU40102M
Kdd0X79@0:1bD02SME@20:5`D00409i]C`2NJe00WV]@09]UB`80UV1600@0X6M:0:=ZC@2SJTd0W6=6
0P2KHdP0102^N5`0U5i209AN@P2XMEX2099OA0040:ElGP2YPV<0ZH9S0:F1G`80XWiL00@0XWaH0:ao
GP2/Oeh0[WeL0P2/MeL0102ZLTl0YVi;0:I^B`2NIT420:5YA0040:I_B02[M4d0ZgA=0;5jD@80ZgA;
00@0[WYE0:ahD`2/N5<0ZWIA0P2YME00102ZMU40[GUD0:eiE02`O5L20;9nF@040:UcC`2]Me<0[GMC
0:egD`80[7IB00@0/7YF0;J0G02fP5`0]GmK0P2_NED0102[Mdl0Xfm70:=_A`2KIcl209QT?00409UU
?@2KIcl0VfLo09UU?@80UV8j00@0Y7180:1/A02PK4@0VVHn0P2CGcL0102?Fc<0SUXb08iJ<P2?Fc<2
095M=@0408aK<`2?GSH0Sehf09AS>`80UF@l00@0Tf8j091O=`2@GcL0Sehf0P2>GCD0102:GC@0OU4Z
07iA:P26FC8207mA;`040:]mG03?X880cj220<ZKO`80[WmS00@0XFQD0:9XE02RJ5@0Z6iH0P2[KeD0
102/L5@0^giQ0;]nH@30Pf@20;5eD`0409EC<@2WIDD0YfE50:m^DP80XfYF00@0REmA06A;AP1TBdH0
BD=30P17BTl0100mBeH0?D]H03e;F00kBeX203];F`0403e:FP0nBUH0?TYF041:D`80@DY?1015C5H2
04I=E`0404Q?F@1;DU`0Be9L04eDGP80CUEO00<0Be9L04aCG@1<Ded01P1=E5h0101<Ded0Be9L04]B
G01<DUh204]AG@0404U?F`16C5P0ATaH04E;E`D0A4YF0`13BED304A:EP<0AD]G0`16C5P<04I=E`04
04M>F018CeT0B4mI04U@FP<0BU5K0P19D5X00`18CeT0AdiH04M>F00304M>F0<0B4mI1019D5X204YA
F``0B4mI3017CUP404Q?F@D0AdiH0`16CEL304Q=D`0404Q<E@16CED0ATeE04I=E`80ADiG00@0ADeH
04A>F014CUP0@DmJ0P14CUL104E>E@40B4eA0@3oool001/0BEIO1018EEhA04MDG@`0B5EN3`19EUl4
04YGH0H0BeQQ0P1;F6001019DUT0?4A703a4A`0X;2l201LG5`03018A3@0H514061@A00`06QDA00@0
5a<@01/F4`0K5Q<05`l<0P0R5a40101<?CH0<b4G03<Q5`0M2@0202H@1@0404Pa8@1KA3@0Fd@d05]4
<P80CSLU00@0CSPS058l9`1B?2L0CCLP0P17<AX0101>=Q`0EShV05Hn9P1K@b/205m6<0040617<`1@
=RD0D3HU03PN3`80:Q0100@05P<0020?200P3`P091D@0P15>S@0101^If40NGI_07UfK`1`KfX205YK
E@0402`f;@0W;RH09bhV02he;P8061dG00@01@X600T;2P092`X07R0O0P0:30/0100X=2X0=Shc03Hn
<`1DE4P208^1M@0409b9NP25J5P0QFQH085NB`80P5Y700@0N4T_06dn901]?R@0Ndde0P2SMeh0102P
N5l0Q5i708ANA`1T@2P205Xf7P0406`n9P29FT00REY009ERA@80XVe=00@0ZgID0:UgEP2YMeH0/H=R
0P2cQVL0102eQfD0/H=Q0;63H@2_OUd20;1nG@040;1kF`2[LU@0Zg9D0:5VB080V5do00@0WVE70:A_
C`2TKdl0YgIE0P2UNEH0102SMe@0XWAB0:9dDP2QKdh209iYB@0409]]B`2PLU00X79@0:9aD@80WV]<
00@0UF9509=M@`2CGD<0Uf570P2IHdT0102>ECL0V5m209QO@P2DFch209MOA0040:E_E@2?FD40SeU1
0:McF`80Uf==00@0VgAG0:AoHP2TOf80WgaN0P2?K4`0102FLU00Zh1P0:^0H02/Ne/209MTA@040:I^
B`2PJ4D0X6Q5095I=080XFU400@0ZW=<0:9[A02RJd@0[GI=0P2^Mdh00`2ZMU40Z7A?0:QdC`030:Ib
C@040:McCP2[Me80ZgMB0:ijE@80/GeH00@0Z79>0:YdD02ZM500[7IB0P2]Me<0102aNeL0]X1L0;J0
G02eOe/20;1jEP040:]gC`2PK4@0X6a409AP>080T5`d00<0U60h09YV?P2JISh00`2LJ400102PK4@0
WFU109eY@@2JISh209IR>P0309AP>02BGSH0TUhf00<0TEde00@0Re/e08iN>02>GSP0Tf<m0P2DI3h0
102BHS`0Seli08mO>@2>GSP208eM=`0408YM=01oDR/0Oe8[08QK=080Oe4_00@0]HMV0<^LOP3;W7h0
aiQl0P2VMe/0102HH4T0WFE>09eUCP2VKUD20:acF0040;=gF`2lPF<0_85S0;UlG@80YfY;00@0TDl/
0:AR@02THT00/71D0P2YLEX0102=HU80HdQ106=8@@19?d0204a;D004041:EP0mBeP0?D]H03Y<FP80
=deK00@0>DaJ03]<EP0kC5H0@DYC0P14BDl404A;E@80ADaF00@0AdiH04YAF`1:DE/0C5=M0P1>EEl0
101:DE/0Be9L04]BG01<Ded304eDGP80C5=M00@0Be9L04YAF`1:DE/0C59N0P1;DEd01019Ce/0ATaH
04I<F015BeL504A:EP<0@4IB0`11Ae<00`12B5@0@dUE04=9E@0304=9E@`0ADaF00@0ATeG04M>F017
CUP0B4mI0`19D5X204Q?F@0304M>F016CEL0ATeG00<0ATeG0`17CUP404Q?F@80BE1J3018CeT<04I=
E`H0AdiH0`16CEL304E<EP<0AdaB00@0Ad]D04E<E015C5@0ADaF0P14CEH01014C5L0@deG04==E`10
CUT204==EP40A4eD0@17C5010?ooo`006`19EUl404QEGQ40AeAM3018EEh?04UFG`@0BUMP1P1;F642
04]HH00404UBF@0lA4L0?4A702P/;`805aLG00<04Q4=01PD4@0H5140300J5A40100G4a006aHC01/F
4`0G3``2028G4@0404`m=P0c8AL0<b4G01d900809Q0500@0B34Q05]4=01KA3@0Fd@b0P1>=bD0101>
>2<0DS`W058l9`1==b0204La6P0404hf701F?RH0EShV05]3:`80GdH`00@0H4Lc050f9@1@=RD0>1h?
0P0Z4040100F0`0080l8020?200T5A0204Dj=00406iWH@1iMVl0NGI_071_JP80FU]E00@0;3H]02L^
9P0W;RH0;SD^0P0H7AL010052PH02@/:00T;2P0N81l200X<2`0402Pd:P0f?S<0=Shc05ADB080Rh5e
00@0W8Uj08EXF025J5P0PEi;0P20FTL0101hBBl0KChT06dn901kCCD20:=gGP040:1hG`24GTL0Q5i7
06A0:080FSHN00@0K3hV08UJ@029FT00UF950P2RKDd0102[MU@0ZGMF0:UgEP2aPf820;>6I`040;F7
I@2aPf40/H=Q0:mnG@80/7iM00@0/7]K0:]bE02[LU@0XFI80P2HGCl0102NIDL0Y6m?0:A_C`2WMUD2
0:EiEP040:=gE02RM580XWAB0:5_CP80WVU900@0Vfe;0:1bD02PLU00XW5A0P2NJd`0102EHTD0Tee3
09=M@`2GHDL209USB@0408iE=`2HGd80V5m209AK?P80Uem400@0YFmE08mI@@2?FD40Yg=K0P2GHdd0
102KM5L0Y7mR0:AoHP2OO5h208m/C00409IbD02[P600Zh1P0:akF`80UfA500@0YVi;0:1XA@2PJ4D0
TETd0P2QJD@0102ZLd`0XV]40:9[A02]MTd20:igCP030:YfD@2XM4l0Z7A?00<0YW9=00@0Yg=>0:]g
DP2[Me80[WYE0P2aOEP0102XLTh0ZWA@0:YdD02/MU820:egD`040;5kE`2fP5`0]X1L0;EoF`80/7YF
00@0ZgM?0:1/A02PK4@0U60h0P2@G3@00`2DH3P0VVHn09YV?P0309aX@0040:1/A02MJD40WFU109YV
?P80UV8j00<0U60h099N=P2BGSH00`2AGCD0102;FcD0SUhh08iN>02CHcd209AT?P04099R?02?GcT0
Seli08iN>080SEdg00@0RUdd07mB:`1oDR/0R5/d0P1oDBl0102eQfH0bian0<^LOP37V7`20:IgF`04
09QPB@2MIDh0WFE>0:I^E@80[7=H00@0/gMK0;b1H`2lPF<0^GaM0P2WJT/0102ACb`0Y6900:AR@02`
L5@20:UaFP0408eRDP1SB440HdQ104To@080C4]@00@0@4YF03e;F00mBeP0>TaJ0P0gCE/0100iC5X0
>daF03]<EP11BU<204A9C`@0A4]E0P15C5H01017CUP0BU5K04YAF`1<Ded204iEG`0404YAF`1;DU`0
Be9L04aCG@<0CEAN0P1<Ded0101;DU`0BU5K04YAF`1<DUh204]AG@0404U?F`16C5P0ATaH04E;E`D0
A4YF0`10AU830457D`030498E013BED0@dUE00<0@dUE3015C5H01016CEL0AdiH04M>F018CeT304U@
FP80B4mI00<0AdiH04I=E`16CEL00`16CEL304M>F0@0B4mI0P19D5X<04Q?F@`0ATeG1P17CUP304I=
E`<0ADaF0`17C5801017Be@0ADaD04E<E015C5H204A=EP0404A<E`13CEL0@deG041>F@80@deF0@14
CE@104M<D040oooo000<04UFG``0B5EN0`19EUl404QEGQ40AeAM6018EEh304UFG`@0BUMP1@1;F640
101:Ef00BeQP04]HH01:DeX203a4A`0402P/;@0I6AL06ATG01@C3`806AHA300J5A40100H51<071HF
01`F5P0E3@X2020E3`040593?00h9Q`0>2HL02HC1@80;QP;00@0@b`N058k:`1B>b/0C3DU0P0k9180
100l9A<0@B/F044[5P0l9Q4203HP2`0404/a6P1D>R<0E3XS05e2;@80IDXe00@0JTll06=8=@1SB3D0
D3DT0P109A@0100[50h09@h802D>200F0`0202`K500404i1>@1RFE80HUUB06aUG@80H5]E00@0<CH_
02d`:@0]<2T0418=0P0@4@`0100=3@/0>3Le03Pg=@2:R8T201LC500402HY8P0k>c<0>c/c04Ho=@80
MFUM00@0YY>40;2EQ02`UH@0ZhYg0P2ELUh0101S=1X0G2hD05`^501fBS420:MmI0040;69L02<ITl0
S6I?06E1:@80EcDL00@0L4@[08aM@`2<GD<0UfA70P2TKdl0102]N5H0[7YI0:ajF@2bQ6<20;>8J004
0;F7I@2bQ680/XAR0;60G`80/7iM00@0/7]K0:]bE02[LU@0XfQ:0P2LHD<0102MITL0XW1?0:9`C`2W
MUD20:AhE@0409mcD02NKD`0WVe<09mZBP80WfQ90`2QL4l0102KJTX0VVM809YWB02BG40208MA=`04
08mG?P2EGD@0UEe409URA080XFU>00@0UEm309MQB@2GHDT0T5a60P22CST0102:Ed@0PTln089??P2>
IDl209afG`0409YfG020GD40P5e108]VB`80XgYN00@0YWMM08YF?P2:ESh0XFU80P2OIdH0102;Dc80
X6Q50:1XA@2YLDh209IN>@040:=[AP2XLDX0Z75:0:=]B@80YFm;00@0Z79>0:YdD02ZM500ZgEA0P2Z
M500102XLTh0YW1<0:I`C02WLDd20:QbCP040:YdD02^N5@0[WQD0;9lF080]GmK00@0]7iJ0;1jEP2`
NUH0ZGE=0P2OJd<0102CGcL0SUXb08iJ<P2CGcL209]W?`040:1/A02RKTH0XVi609iZ@P80WFU100@0
VVHn09UU?@2IICd0V6@l0P2EHCT0102BGSH0T5`d091L=02:FS@208iN>004099R?02CHcd0Tf<m095Q
>`80SUhh0`2=GCL0102;GSD0PUD^089E;P2<GcP207mA;`040;R:J@30TG<0`95c0;b=L@80VFY>00@0
Tee309aVC02LIT`0YfmD0P2/M5T0102eO5l0_81T0;b0I02bMeT209mTAP0408m?:`2PH3d0X60m0:i_
D080[gEO00@0UVQI06M:A01WBT@0C3hn0P1=Be001012B5H0?TYH03i:F00jC5X203I=F`0403M=FP0k
C5H0>daF0499D`80AdM?1013BU@204E<EP0404I=E`19D5X0BE1J04]BG080CEAN0`1;DU`0101<Ded0
CEAN04eDGP1=E5h204aCG@<0Be9L00@0Be5M04Y@G01:D5`0B4iJ0P16C5P604E;E`@0@4IB2011Ae<<
04A;E@0404E<EP16CEL0ATeG04M>F0<0B4mI0P17CUP00`16CEL0ADaF04E<EP0404E<EP<0ATeG4@17
CUP<04E<EP0304M>F016CEL0ATeG00<0ATeG0`15C5H304A;E@<0AT]A00@0ATYC04A;D`14Be<0A4]E
0P13C5D01013BeH0@TaF049<EP10CUT2049<E@40@daC0@15BTh10?ooo`003019EUl<04QEGP<0BEIO
1018EEhA04MDGAP0B5EN0`19EUl404YGH0D0BeQQ00@0BEIN04]HGP1;F5h0BeAI0P0mADL0100Z;Rl0
71`J01`L6P0G5Q4201XG4P`06QDA00@06QHG01`F5P0L5QH05@d;0P0N4`l0101CAC`0?2XN03`Z7P0b
7a4203DO4@0404li:`1JA3H0FT@f050k:P80?RTH00@0?bXG04Ha7P16<Ah0@bhI0P0n:A@01019;aP0
E3XS05@j8`1QAS4206Q=>00406aA?01]DSd0KE8m06Q=>P80Gd@a00@0ITQ0058e;@1B=Bd0;A890P0M
1P00100O2`80;ahF02lN5P1<?CH205I9@@0404M4?@0j=c00>SL`024N6@805a8>00@07aLE04]3@@1;
@d40S8:30P0K4A80100`9B40ASTc04Hi<`13=B`207APE@040:fFQP2WS7T0Yhai08Y^F080K4hf00@0
I3HL06i09P1^@2H0S6170P2/PVT0102ZQF/0QF5908EQB@1VA2/205hl8`040755;02<GT@0S5i409IU
A`80XW1?00@0[7YG0:]jF@2[NUT0/HAS0P2bQfL0102fR6H0]HMU0;F7I@2eQ6<20;:0G`040:mjFP2[
LU@0Zg9D0:I[C@80XFI800@0VVE50:5_CP2QKdh0YWQF0P2TN5D0102MLDh0VfY909]ZB@2PJd/20:I]
C`040:IeE02NKD`0WVe<09AQ@P80WVU:00@0VF=708YB>@2:DST0SUHm0P2BFT40102BGd@0UV=809IS
B026E3/2091MB004081@?028EdP0R5M807i>@080Odm300@0O5E408iYEP2>JEH0TFiJ0P1cD3X0101i
ECl0Rf5;08]QB`2CHdl207]8<`0409=N?`2KITL0VfI708E@<@80V6=300@0X6];08]F=02;ES@0VfM2
0P2LJ4<0102LIT80WfU509mYA@2TKTX20:QbCP040:YdD02YLdl0ZG=?0:I`C080Y6i:00@0Yg5=0:Qb
CP2XLTh0ZWA@0P2^N5@0102cOET0]7iJ0;AnFP2bO5P20;1jEP040:QdC02QKDD0XFe509QT?080UF4i
00@0V6@l09iZ@P2NJT80Xfm70P2ULDT0102RKTH0Wf]309m[@`2KIcl209UU?@0409QT?02FHSX0UV8j
099N=P80Se/c00@0QeTe08]M>@2;GCT0Sf4m0P2@HSh00`2>H3`0Redi08]M>@0308YL>00408YM=025
F340QEPa095T?@80OU0^00@0^8YY0;62I02aPV@0[h1T0P2<GD40102GHDD0XF]?0:5[C`2WLED20:ad
F@040;ImH02kPVD0^h9U0;1dF080WV9600@0UED_09eM>@2MGCT0ZFY;0P2aMel0102QLV80LUA<079D
C01@@44204e8C00404M6EP12B5P0@TQH03Y<FP80=TeK00@0=deJ03e;EP0mBeH0ADMC0P1;ADl30499
D`0404=:E014BeD0A4]E04I=E`80B4mI00@0Be9L04aCG@1<Ded0Be9L0P1<Ded0101=E5h0CUEO04iE
G`1>EEl204eDGP0404aCG@1;DU`0Be9L04Y@G080BDmK00<0AdeI04I<F016C5P00`15BeL304I<F0@0
@dUE2012B5@<04=:E0<0ADaF00@0ATeG04M>F017CUP0AdiH0P16CEL304E<EP<0@dYD1014BeD504E<
EP`0ATeG3015C5H304I=E`<0ADaF1014BeD204=:E0<0ADY@00@0ADUB04=:DP13BU80@dYD0P12Be@0
1012BUD0@D]E045;E@10CUT2049<E@40@T]B0@14BDd10?ooo`003019EUl<04QEGP<0BEIO1018EEhA
04MDGAP0B5EN0`19EUl404YGH0D0BeQQ00@0BEIN04]HGP1;F5h0BeAI0P0mADL0100Z;Rl071`J01`L
6P0G5Q4201XG4P`06QDA00@06QHG01`F5P0L5QH05@d;0P0N4`l0101CAC`0?2XN03`Z7P0b7a4203DO
4@0404li:`1JA3H0FT@f050k:P80?RTH00@0?bXG04Ha7P16<Ah0@bhI0P0n:A@01019;aP0E3XS05@j
8`1QAS4206Q=>00406aA?01]DSd0KE8m06Q=>P80Gd@a00@0ITQ0058e;@1B=Bd0;A890P0M1P00100O
2`80;ahF02lN5P1<?CH205I9@@0404M4?@0j=c00>SL`024N6@805a8>00@07aLE04]3@@1;@d40S8:3
0P0K4A80100`9B40ASTc04Hi<`13=B`207APE@040:fFQP2WS7T0Yhai08Y^F080K4hf00@0I3HL06i0
9P1^@2H0S6170P2/PVT0102ZQF/0QF5908EQB@1VA2/205hl8`040755;02<GT@0S5i409IUA`80XW1?
00@0[7YG0:]jF@2[NUT0/HAS0P2bQfL0102fR6H0]HMU0;F7I@2eQ6<20;:0G`040:mjFP2[LU@0Zg9D
0:I[C@80XFI800@0VVE50:5_CP2QKdh0YWQF0P2TN5D0102MLDh0VfY909]ZB@2PJd/20:I]C`040:Ie
E02NKD`0WVe<09AQ@P80WVU:00@0VF=708YB>@2:DST0SUHm0P2BFT40102BGd@0UV=809ISB026E3/2
091MB004081@?028EdP0R5M807i>@080Odm300@0O5E408iYEP2>JEH0TFiJ0P1cD3X0101iECl0Rf5;
08]QB`2CHdl207]8<`0409=N?`2KITL0VfI708E@<@80V6=300@0X6];08]F=02;ES@0VfM20P2LJ4<0
102LIT80WfU509mYA@2TKTX20:QbCP040:YdD02YLdl0ZG=?0:I`C080Y6i:00@0Yg5=0:QbCP2XLTh0
ZWA@0P2^N5@0102cOET0]7iJ0;AnFP2bO5P20;1jEP040:QdC02QKDD0XFe509QT?080UF4i00@0V6@l
09iZ@P2NJT80Xfm70P2ULDT0102RKTH0Wf]309m[@`2KIcl209UU?@0409QT?02FHSX0UV8j099N=P80
Se/c00@0QeTe08]M>@2;GCT0Sf4m0P2@HSh00`2>H3`0Redi08]M>@0308YL>00408YM=025F340QEPa
095T?@80OU0^00@0^8YY0;62I02aPV@0[h1T0P2<GD40102GHDD0XF]?0:5[C`2WLED20:adF@040;Im
H02kPVD0^h9U0;1dF080WV9600@0UED_09eM>@2MGCT0ZFY;0P2aMel0102QLV80LUA<079DC01@@442
04e8C00404M6EP12B5P0@TQH03Y<FP80=TeK00@0=deJ03e;EP0mBeH0ADMC0P1;ADl30499D`0404=:
E014BeD0A4]E04I=E`80B4mI00@0Be9L04aCG@1<Ded0Be9L0P1<Ded0101=E5h0CUEO04iEG`1>EEl2
04eDGP0404aCG@1;DU`0Be9L04Y@G080BDmK00<0AdeI04I<F016C5P00`15BeL304I<F0@0@dUE2012
B5@<04=:E0<0ADaF00@0ATeG04M>F017CUP0AdiH0P16CEL304E<EP<0@dYD1014BeD504E<EP`0ATeG
3015C5H304I=E`<0ADaF1014BeD204=:E0<0ADY@00@0ADUB04=:DP13BU80@dYD0P12Be@01012BUD0
@D]E045;E@10CUT2049<E@40@T]B0@14BDd10?ooo`003019EUl<04QEGP<0BEIO1018EEhA04MDGAP0
B5EN0`19EUl404YGH0D0BeQQ00@0B5EM04aIG`1<FEl0C5EJ0P0nATP0100[;bh07QhL01hN700I61<2
01/H4``06QDA00@071LK01dG6@0M5aT05@d;0P0M4Ph0101CAC`0?bdQ03l]8@0j9aP203TT4`0404Tc
9@1B?2h0DS`^04H`8P80<ah=00@0=B0?03hY5P0n:AH0?2LB0P0h8`h01017;AH0ECXU05Dj9@1TBC@3
06U>>@80L5E000@0MU]607MLA`1gG4L0LDa30P1]Bd80101J>C00ARTQ04HY8@0]5@/202090@0402hL
4P0e8aT0=B<I02HO5`809b0H00@0ASdh01D<1`0E30L03@800P0O4A40100X6QX06P`<01X<300f8Ah2
04/f<@0404Tb:P1eGE40MEeA09AjJ@80Je0k00@0F3`T05m4:@1OA2T0MDL]0P29GD80102UNF00Zh=Y
0:^3J@2KMU`207UE?@0406M5;01VA2/0IT@[0755;080Ree300@0UFA60:5_CP2QKdh0ZgUF0P2ZNEP0
0`2_PV40/8EU0;25I@030;J8IP040;N6I@2dPV40]89Q0;1kF`80Zg9D00@0Yfa>0:AYB`2TJD/0UfE4
0P2OKTd0102WNEL0YGUF0:EiEP2MLDh209]ZB@040:E^C`2^MEL0[WEG0:]iF080VfU800@0SU/l0:9]
CP2RKDh0YW1D0`2EGD@209MMA@0408ME?028ESd0R5Hm07E5;`80REU500@0NDXj09=UF02CIEP0Oe16
0P26Edl01024GDh0UW5Q09IaH@2LN6P207aIAP0407]F@`22EdH0PUM608eNC080Od`k00@0Q54f09MT
A`2GI4L0P4d`0P2@GCh0102HIDH0QU@c08ID<`2HITD209AR@@0409US?`2KID40VfE109iXA080XF]7
00@0YFm;0:QbCP2XLTh0ZgEA0P2/MU830:MaC@030:YdD02_NED0[gUE00<0/geI00@0/G]G0:miE@2_
NED0ZGE=0P2TL4P0102NJT80W6Q009aX@02MJD420:1/A0040:A`B02ULDT0YG590:IbBP80XVi600@0
W6Q009QT?02HI3`0Uf<k0P2DH3P0102AGCD0SUXb08iJ<P26F3@208YL>00408iP?02@HSh0T68n08iP
?080Redi0`29FcL0102:GC@0QeXc08MJ<`2EJ44207i@;P040;J8I`2XNE/0Z7UK0:IgF`80Q5Di00@0
WFM;0:E_D`2UKe<0Z79F0P2[LeP0102eOF80^h9W0;^2I`2_MUT20:1TB00409]K=@2MGCT0WEdi0:EV
A`80/GMO00@0ZGYZ07]ME@1kGED0E4A40P1<ATX0101:AUL0ADQI04E8F@0lC5/203M>G00403Q>F`0o
BeL0?d]G04U7E080CdE@1012BE<204=:E00404E<EP18CeT0B4mI04YAF`D0C5=M00@0CEAN04iEG`1>
EEl0CUEO0P1=E5h304aCG@0404U?F`18CUX0B4iJ04M=F@80ATaH00<0AD]G04I<F016C5P00`17CET3
04I<F0<0AD]G0`14BUH304=9E@`0@dYD00@0A4]E04E<EP15C5H0ATeG0`17CUP204I=E`0304E<EP14
BeD0A4]E00@0@TUC0`13BU@504A;E@`0ATeG3014BeD00`16CEL0ADaF04E<EP0304E<EP<0A4]E0`13
BU@304E:D00404E9DP13BU80@dYB04=:E080@T]D00@0@TYE045;E@11BeD0@4iI0P12C5D1045:D@40
@dQ<0@3oool000@0AeAM1@18EEh304UFG`<0B5EN0`17E5d404ICG0h0AE9K3018EEh<04MDGAP0BEIO
00@0AeAL04QEF`18EE/0BU=H0P0oAdT0100Y;B`05QHD01HF500D4`h201dJ5@@07ADB0P0K5Q8301XG
4P<061LB00@06QDB01XE4@0J5A40614;0P0O5Pl0100c:B00?2lV03`_9P0j;24203d]800404<f9@1>
@C00CT4`04Xj:P80@30P00@0@2lO03dZ6`0m:Q/0=R<D0P0h9AH01012:a/0Bc4R04/a8P1J?bh206U=
>00406m@>`1aDCX0LE4j07YH?`80QF5900@0P5U807mJB01oFTP0O5U60P1dED40101WC3L0DcXT05<j
900k9`l202dI0@0401l9000P2P0080X002<=0P80:A8:00@0<1TC03<L5P0c71H0<QXF0P0`61@0100o
;Al0>RLH03XW600i91<203lY500404lg7`1SBC00HdT`07AI?080OF9500@0SVM:09UbE@2ILUD0Ug9E
0P2JMEP0102JMeT0NePj07]H>P1P?b0206M69`0407iA=02:GD00RUe009Q[CP80Y7EI00@0ZGYN0:il
H@2^O640/h5V0P2gQFX0102aQF80]HUV0;F9IP2eQfD20;:1H0040:mmG02_NUX0[gYJ0:]dE@80YFa>
00@0UfU709]]B`2KKD/0YWQF0P2UMeD0102IJdT0VVa:09Y/BP2XNUP20:j0GP040;F2I`2QKe@0XFmD
081C=P80V6a?00@0U6M:08eJ?`2=FSl0U5Y20P2OHDX0102;GDD0SVA<08iTC01mFd8207]OB@0407QN
C@23K5h0PfaN07QPEP80MU]D00@0O61B08ibI02>LV@0RViP0P2@M6H0101hFdd0P61C081PD`1kFDd2
08MUF@0407=Q>`1oGCX0Oedj081;;080QE0a00@0PUTi081QA020HD@0Rf]B0P2:I540102>GSX0TElj
095O>P2II3`20:=Y@@040:M]A@2VLDT0YW590:AbC@80Xg9A00@0Y7=;0:EdC02UM4`0Z7M?0P2]O5@0
102aP5P0/X5I0;:1F@2`OeL20:imE@040:MgC`2UM4`0YGA<0:5`B0<0XFe50`2RKD<20:9]@@040:Q^
B02SK4D0Xfa509mX@@80VF@l00<0UV4i09=O=`2CGcL00`2BGSH0102:FC40Re/c08]K<`2>GSH308iQ
>080S64g00@0RUle08QM<`28GC<0QUXe0P2AID00102;GcX0S5hl08aN?02UMeH20;:1H@0408QG>@2;
F3/0RePk0:5YD080]7aS00@0]7YR0;^2I`2kPVL0_X9V0P2`LeL0102ZKE00XfI90:=VB@2_Idh209mH
?00409iI>P2dMEH0]7EF0;B2J@80QFAA00@0FTPn055:A01ABT@0B4aG0P15CEP01013C5/0@4aL041<
G010BEP20457D`0304=6C`15AT/0ADI;00@0@dYD0P15C5H01017CUP0BE1J04U@FP1<Ded204eDGP<0
C5=M2@1=E5h0101=Del0C59N04aBGP1:D5`204U?F`H0AdeI00<0ATaH04E;E`15BeL00`15BeL304A:
EP<0@dUE3014BeD404U@FP80B4mI00@0AdiH04I=E`16CEL0A4]E0P13BU@304E<EP0404A;E@13BU@0
@dYD04=:E080A4]E0`15C5HT04M>F0<0AT]A00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF
049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`001017E5d504QEGP<0BEIO0`18EEh304MDG@@0AU=L
3P15DU/<04QEGP`0AeAM6019EUl01017E5`0B5EK04QEF`1:DeP203m7B@0402T];00F5Q@05QHD01@C
3P807AXE100M5A8201/F4P<06QLB0`0H5a80100J5A806QDA01XE4@0H4@/201lF3`0403<Y800l;bH0
?2lV03X/8@80?BdP00@0@cHU04i1<01>@C00BSXZ0P10<2001010;al0?BXK03dZ6`0f8a@203PU5P04
048[6`1;<B80Bc4R05Xo;P80JDdh00@0Ke0k075A>P1aDCX0NUPo0P25HDT01020FDP0OeY807mJB01l
FDH207AE@@0406M<=`1C>R@0DcXT03/W3`80;AT100@07`T0020:000P2P008`d20P0Y4PX0100`6A<0
<a`F03<L5P0b6QH2030H500403l]7`0j9aP0>RLH03TT4`80?bTD00@0CcLO06=9<01SBC00M5Tl0P1m
HTD0102>IdX0VG9E09UbE@2GLUD209YeF00409YgF@1kF3X0NePj060o8080IdHW00@0OU4d08YM@02:
GD00V6]>0P2TMET0102YNUh0[WaQ0:ilH@2cPFH20;N5JP040;65HP2eRFH0]HUV0;F7I@80/X5P00@0
[geL0:mjFP2_NUX0ZgAE0P2UK4h0102GJDL0Vfe;09]]B`2VN5H20:EgE@0409U[B@2JK4X0VVa:0:Qj
F080[X1N00@0]H9W0:5_E02QKe@0P5<f0P2HK4l0102DIdX0SEXo08eJ?`2DFT8209mQBP0408]MA@2>
I4`0SVA<07eK@P80Nem900@0N5i=08=/GP23K5h0N61F0P1fFe@0101lH580SW9T08ibI02:KV02091d
IP0407QKC@20H5<0P61C07]IC@80QfEI00@0Lf4k07mM>P1oGCX0P4//0P25D3401022FCT0P654081Q
A02;Je8208YTD@0408iN>P2AGcX0TElj09UT?080XfU100@0Yfe50:IaB@2VLDT0Y79=0P2SLU40102T
Ld/0YGA<0:EdC02XMdl20:elE0040;60F02bPET0/X5I0;1oE`80[WeE00@0YgM?0:EdC02UM4`0XG18
0`2QKDD30:9]@`80XVe100@0Z6i80:=/A@2SK4D0WfQ10P2II3`00`2FHCT0Telg09=O=`03099N=P04
08YI<@2;Fc<0Re/c08iN=P<0SV4h0P2<HCL0102:GcD0R5dc08QM<`26FSD2095U@00408]O>P2<GS`0
S5hl0:EgEP80/X5Q00@0R5Li08]H>`2;F3/0XFU@0P2dO6<0102dNV80^h9W0;^2I`2nPVH20;1cE`04
0:Y]D02SITT0XfI90:mWCP80WePl00@0WUTj0;AeEP2dMEH0]89Y0P25I540101JB3h0DDY4055:A018
C5L204E=F00404=<F`10C5`0@4aL0419F080@DMC00<0@dI?04E6B`15AT/01013BU@204E<EP0404M>
F019D5X0BE1J04aCG@80CEAN0`1<Ded904eDGP0404eCG`1<DUh0C59N04Y@G080BDmK1P17CET00`16
C5P0AD]G04E;E`0304E;E`<0A4YF0`13BED<04A;E@@0BE1J0P18CeT01017CUP0ATeG04I=E`14BeD2
04=:E0<0ADaF00@0A4]E04=:E013BU@0@dYD0P14BeD304E<ER@0AdiH0`16Be401016BU<0A4]C04A;
D`14BeD204=<E@0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000404MDG@D0B5EN
0`19EUl604QEGP<0AeAM0`16De`<04EBF``0B5EN3017E5dH04UFG`0404QEG@19EU`0BEIL04YCF080
?TI800@0:Bd/01LG5@0G5aD05A@?0P0N6aH401dE4P806aHB0`0J5a8301PG4P0401TE4P0K5Q806aHB
01TB30807A@=00@0;B<J03@W7P0d9ah0<R@I0P0g9aX0101:?B`0DTDd0595=01>?Rh2048b8P04044`
800l:AX0?2TJ03<P4@80<Ql@00@0ARlO04hd9@1>=2D0GD8a0P1]DC`0101gF4<0NEU207UI@P1nG4<2
08EQB@0408YTC`29IDl0RFE?08UVD080QFM?00@0OF59071H?P1`F3h0H4/`0P1F@BH0101;=2@0@BXJ
044Z6P0h8A<203TR5004040X7014;280A2`R040X7P80>b<I00@0;a`=03`W5P0l9aH0D3TW0P1SCCP0
101aFD40LeU007=I@01^DcH206M<;`0407A=<01oFSd0OeXm08=N@@80S6M:00@0T6]>07EB=01eDS@0
GS/M0P1U@R@01022ECP0RUe008YM@02FId/209i_D`040:9cE`2XMU/0Z7IK0:ilH@80/h5V00@0/HIS
0;F9IP2eRFH0]HMU0P2aP5l0102^O5/0[gYJ0:mjFP2/MEH20:M^D0<0V6Y800@0Wg5?09i`CP2NL4h0
U6I40P2HJTP0102VN5H0ZgeK0:]mF`2hQFP20:=bE00407a=<@1mD3<0OE0c085B=P80T5e200@0S5@m
0898<P22B380TVA=0P21F480101lG4D0QfiH08M^F01cHU0207UYFP04085cIP2?OW@0Sgid08]eI`80
OVQJ00@0R79T08=/GP23K5h0NF9D0P1fGe40101dGDl0MUmA07IOD@1YG3X206U:;@0406@e6`1`@BL0
L44W07EB=P80KU@i00@0LeM206m>?`1_CSl0QV570P2:HDD0102AHd80VFM209UW@P2PJd<20:9]A@04
0:1/A@2MJdH0WF]609a/AP80Vf]500@0W6a60:9bC02RLT`0Z7QB0P2YNE<0102UMDl0X71:0:1`BP2R
LT`20:IdC`040:MfCP2SKdL0Xfm709]W?`80V6<i00@0W6Lm0:1[@@2PJd40ZFm90P2VK4H0102PJD80
Vf@m09]T?@2GHSX209AP>0<0Telg00@0SE/f08iN>02>GSP0T60j0`2?HS/208]O>00408QL=@26FS<0
QUXc08IJ=@80T6@o00@0Relj08QJ>028FSP0W6i=0P2FIDD01020Cc40S5Tl08aI?02UKE@20;AlH`04
0;EkH`2jPFH0^X5V0;f1I@80[G1D00@0XVE809]N@@2KGT40_7EI0P2SG3h0102IE3D0[g1A0:m`D@2f
Q6T208m/F@0405i<@P1?AT40CdI104M;E080AD]G00@0@TaH03m;F@0oBeT0@DYI0P11BE@00`14Ae00
ATI>04I6CP0404A;E@80ADaF00@0AdiH04U@FP19D5X0Be9L3P1<Ded0101<DUh0Be5M04]AG@19Ce/2
04Q>FP@0AdeI0P18CUX304M=F@0404I<F015BeL0AD]G04E;E`80A4YF0`13BED<04A;E@0404Q?F@19
D5X0BE1J04U@FP80B4mI00@0AdiH04I=E`16CEL0A4]E0P13BU@304E<EP0404A;E@13BU@0@dYD04=:
E080A4]E0`15C5HT04M>F0<0AT]A00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0o
CEP2049<E@40@daC0@16Bdl10?ooo`001017E5d504QEGPL0BEIO1@18EEh304MDG@`0AU=L3018EEh<
04MDG@`0BEIO301:Ef00101:Eel0BUMM04YGG@19DUL203a4AP0402L[:P0H61H061PF01HE40807Q/F
100M5A8201/F4P<06QLB0`0H5a80100I5A806aPC01/H4`0H4`d201/D300402LM500/7aH0;1lF02XL
4@80;R0C00@0C3l^0554<`1AA3<0C3`/0P12<R801011<200?BXK03dZ6`0d8A82034N3`04040Y6@15
:a`0AB/L050e9080GT8]00@0JT/f075A>P1aDCX0NELn0P1oFd<0102?JTl0Sfa@08m/D02@KU82095b
E@04095dEP2<LE<0S75C08E/C@80P6U900@0O6A<06eE?@1]ECd0GTD_0P1L@bd0101QB3@0HTTe0699
=@1G?B`204`b8@0405Lm;01K@C00Fd4`0658=080J4dh00@0L5@l07ML@@1gG440P6960P25Id/00`2;
ITT0UG1C09E`D`0309IaE004099]D01fDC@0MU4d064l7`80IT4T00@0Rea0095RAP2AHTH0VFY>0P2N
Ke<0102QLUH0YGIJ0:EfFP2/OF420;62IP040;>8IP2fRfT0]X]Y0;B7IP80[X1O00@0[W]L0:mjF`2_
NU/0[GIH0P2XLE<0102LKTd0V6Y909QZB@2KKD`209Y/B`0409EWAP2KKD`0Vfe<0:MiF080ZG]J0`2R
LE<0102BHdL0P54e081A=@21DSP209QYC`0408YH@@1e@b`0MD</08AGA080PUa900@0SgAO09>2KP2C
PVh0Q7eZ0P2JVHD0102KW8`0QHIf08F6MP2;PW4205a@@00406YMC@1OCd00Gdm00551<P80EDPh00@0
NfmO07YaH01jLF00GEPk0P1nJ500101lET40HSTU068i9@17;AH204Ti8P0405e:>`1K@3L0Fd0g06EB
A080KeQ600@0OUi708]RA02;HT@0Tf<o0P2EHCX0102CGSH0SeXb08mJ<P2>GSP2099R?00409a/AP2Z
NU@0ZWYD0;F5G`80]hMQ00@0/81J0:QhDP2XN580Z7QB0`2/NUD20:QdC@040:1/A02NJD40WVU10:9]
@`80Yg9800@0ZW1;0:M]B02WKDP0XfU40P2LI3l00`2HHS`0UV0j09IP>P0309EQ>P04091N>@2@H3X0
T60j099R?080TF@m00@0Sf8k08]O>02;GcP0Qe/d0P25FC801025FC@0T6@o091T?`2=HC`208EG=@04
09IXA`1oCRh0Odh^08IE=`80VVM:00@0[WIM0;EmI02eOF@0]WaT0P2hOf@0102iOF40ZFa@0:U/D02I
G3l209MJ?@040<f6JP2ZHdD0ZV=509A?;P80Z6U:00@0^HI[09YeH`2JMF<0IU970P1>@cl01018B500
ADUD04E9E012BUD3045:F@80@TYE00<0ADUB04Q8D018B5001015C5H204I=E`0404M>F019D5X0BE1J
04YAF`P0Be9L0`1:DE/304]BG0<0BDmK0`17CET00`16C5P0AdeI04M=F@0304Q>FP0404U?F`18CUX0
B4iJ04Q>FP80AdeI00@0AD]G04A:EP14BUH0A4YF0P13BED<04A;E@H0B4mI00@0AdiH04I=E`16CEL0
A4]E0P13BU@304E<EP0404A;E@13BU@0@dYD04=:E080A4]E0`15C5HT04M>F0<0AT]A00@0ATYC04A;
D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`001017E5d5
04QEGPL0BEIO1@18EEh304MDG@`0AU=L3018EEh<04MDG@`0BEIO301:Ef00101:Eel0BUMM04YGG@19
DUL203a4AP0402L[:P0H61H061PF01HE40807Q/F100M5A8201/F4P<06QLB0`0H5a80100I5A806aPC
01/H4`0H4`d201/D300402LM500/7aH0;1lF02XL4@80;R0C00@0C3l^0554<`1AA3<0C3`/0P12<R80
1011<200?BXK03dZ6`0d8A82034N3`04040Y6@15:a`0AB/L050e9080GT8]00@0JT/f075A>P1aDCX0
NELn0P1oFd<0102?JTl0Sfa@08m/D02@KU82095bE@04095dEP2<LE<0S75C08E/C@80P6U900@0O6A<
06eE?@1]ECd0GTD_0P1L@bd0101QB3@0HTTe0699=@1G?B`204`b8@0405Lm;01K@C00Fd4`0658=080
J4dh00@0L5@l07ML@@1gG440P6960P25Id/00`2;ITT0UG1C09E`D`0309IaE004099]D01fDC@0MU4d
064l7`80IT4T00@0Rea0095RAP2AHTH0VFY>0P2NKe<0102QLUH0YGIJ0:EfFP2/OF420;62IP040;>8
IP2fRfT0]X]Y0;B7IP80[X1O00@0[W]L0:mjF`2_NU/0[GIH0P2XLE<0102LKTd0V6Y909QZB@2KKD`2
09Y/B`0409EWAP2KKD`0Vfe<0:MiF080ZG]J0`2RLE<0102BHdL0P54e081A=@21DSP209QYC`0408YH
@@1e@b`0MD</08AGA080PUa900@0SgAO09>2KP2CPVh0Q7eZ0P2JVHD0102KW8`0QHIf08F6MP2;PW42
05a@@00406YMC@1OCd00Gdm00551<P80EDPh00@0NfmO07YaH01jLF00GEPk0P1nJ500101lET40HSTU
068i9@17;AH204Ti8P0405e:>`1K@3L0Fd0g06EBA080KeQ600@0OUi708]RA02;HT@0Tf<o0P2EHCX0
102CGSH0SeXb08mJ<P2>GSP2099R?00409a/AP2ZNU@0ZWYD0;F5G`80]hMQ00@0/81J0:QhDP2XN580
Z7QB0`2/NUD20:QdC@040:1/A02NJD40WVU10:9]@`80Yg9800@0ZW1;0:M]B02WKDP0XfU40P2LI3l0
0`2HHS`0UV0j09IP>P0309EQ>P04091N>@2@H3X0T60j099R?080TF@m00@0Sf8k08]O>02;GcP0Qe/d
0P25FC801025FC@0T6@o091T?`2=HC`208EG=@0409IXA`1oCRh0Odh^08IE=`80VVM:00@0[WIM0;Em
I02eOF@0]WaT0P2hOf@0102iOF40ZFa@0:U/D02IG3l209MJ?@040<f6JP2ZHdD0ZV=509A?;P80Z6U:
00@0^HI[09YeH`2JMF<0IU970P1>@cl01018B500ADUD04E9E012BUD3045:F@80@TYE00<0ADUB04Q8
D018B5001015C5H204I=E`0404M>F019D5X0BE1J04YAF`P0Be9L0`1:DE/304]BG0<0BDmK0`17CET0
0`16C5P0AdeI04M=F@0304Q>FP0404U?F`18CUX0B4iJ04Q>FP80AdeI00@0AD]G04A:EP14BUH0A4YF
0P13BED<04A;E@H0B4mI00@0AdiH04I=E`16CEL0A4]E0P13BU@304E<EP0404A;E@13BU@0@dYD04=:
E080A4]E0`15C5HT04M>F0<0AT]A00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0o
CEP2049<E@40@daC0@16Bdl10?ooo`001017E5d504QEGP<0BEIO101:Ef0304UFG`D0B5EN3016De`H
04QEGP`0BEIO301:Ef00101;F600BeQN04]HGP18DEH203Q0@P0402DY:00I6AL06ATG01PG4P807Q/F
100M5A8201/F4P<06QLB0`0H5a80100G5Q8071TD01`I500I50h201XC2`0402<J4@0V71809Q`B02LK
3`80;R0C0`19?2/01015=BD0?2`L03`/700l:a/203h[700403/X6@0g91D0=b@E03`U5@80C38S00@0
HDHe07AH@`1dF4<0P65<0P23Hd`01025HdX0Qf=;08MSB`2=JDT208eZBP04091]C@2BLE00TW5@09=e
D`D0U7ME00@0UGQJ08e`DP2=L580QfY<0P27JDd01029Je40PfE;08=UB`1dESh206I8<00407IJA@1h
G4L0N5a707aMB080OV1800@0P69808IXCP26J4h0Sg5E0P2GN5`0102GLUD0WWUL09iiG02KMUT2095/
C`0408IO@P1aBRd0LDX]06Y19@80Mdhb00@0UF=809]YCP2KJDh0XW1E0P2UMUX0102YNUh0[GiR0:en
HP2bQFP20;F8J`040;B<J@2gS6X0]haZ0;F8I`80[X1O00@0[7]K0:ikG02^Ne`0[WUJ0P2ZLeD0102R
M5<0VVa;09Y/B`2KKD`209e_CP0409]]C02RM5<0XWAC0:UkFP80YWQG00@0W6e?0:EdEP2UM5H0YgEJ
0P25DcX0101eABh0RE]408UKA025Fd<208MNB00409A]GP2KNVT0VgYY0:NBO`80WIJ300@0UiZ509n[
U`2OZiL0O8Ug0P19F4D0100e=RH0B4Df04Q5=P0_9aX203@X700404m3=`11>bd0@C/]04Q9>@80Kg=R
00@0V9b508AgI024Mf@0MUU90P22HU<0101WE4<0=bhO03L^7`0e:b4204/i=@0403Pg=@1AB440DDQ1
06iKC080P61700@0REhn08]J<P2;FS80ReH/0P29DbL0102@HSh0UfU509MYA@2RM5020:anFP<0/H=O
00@0[7iJ0:UkE`2YNeL0[gmK0P2[NEH0102ULdh0Xfm80:=_B02RKTL20:E`B0040:MbBP2XLd/0Z7=;
0:]_B`80ZFe900@0Y6Y509mU@02OID00VV8m1@2HHS`0102AGc`0TF4m095Q?@2BHSh2095S?`04091R
?P2<H3/0S60k08QL=`80QUXe00@0Q5Pc09=W@P2CId80T6@o0P24ES@0102HJTT0PE0`085@<02NKDl2
0;1mH0040;R0I`2eOF@0]GeT0;IlI080/gYO00@0]7QL0:MZCP2WJTh0UUTl0P2NHD@0103GTW<0/fa<
0;=/C02FDC020:ETA0040;Z4J02PNVL0X7YW06mGC@80Dd@o00@0B4M=04M7C`17Adl0A4QC0P12BUL0
1011BeL0@d]F04=;EP16BUD204M:D`0304M>F016CEL0ATeG00<0ATeG00@0AdiH04Q?F@18CeT0BU5K
1@1;DU`00`1:DE/0BE1J04U@FP0404U@FP80BU5K00@0AdeI04I<F016C5P0ATaH0`15BeL204I<F004
04M=F@18CUX0B4iJ04Y@G080BDmK00@0B4iJ04M=F@17CET0ATaH0P15BeL304A:EP`0A4]E00<0AdiH
04Q?F@18CeT01018CeT204I=E`0304E<EP14BeD0A4]E00<0ADaF00@0A4]E04=:E013BU@0@dYD0P14
BeD304E<ER@0AdiH0`16Be401016BU<0A4]C04A;D`14BeD204=<E@0404=;EP12C5H0@TaF03m=F080
@TaE0@13C5<104I;C`40oooo000404MDG@D0B5EN0`19EUl404YGH0<0BEIO1@18EEh<04MDGAP0B5EN
3019EUl<04]HH@0404aIH@1<FEl0C5UO04I?E080=3`n00@08RHU01TI5`0I6AL061LB0P0N6aH401dE
4P806aHB0`0J5a8301PG4P0401LF4P0J6A@06QTD01PE3P806AD<00@091dC02PN500X7Q@0:1hB0P0a
9AL01017>RT0A3LV04@g9P10<20203LW5`0403@S4`0k:1T0>bPI03h[7080?2TJ00@0=b0@04P^7`18
;Ql0G44`0P1YCCP0101bDch0Ne]407]KA027ID`2091/E00408m[A`2AKDT0TFe9099`B`80Tg5<00@0
UG=>09IfD02FMU00UgMA0P2INE<0102JNEH0W7]H09akF02POe`20::1H00409mnG`2HMeP0V7MH08m]
D@80RFM;00@0RFM>095_EP2AKeH0VgUP0P2POVD0102POV<0VgaP09]lH02HNEd209MhG00409]hFP2M
N5/0WGQK09A_DP80Pe`o00@0LdX^06Xn8`1Z?R<0LdL/0P27Fd00102GI4T0WFY?09eZC`2ULeP20:ek
H0040;21I@2cQVT0/hIY0;F9K080]h]^00@0]8][0;J=K@2fSFd0]HYZ0P2^PF80102[O5h0[gaO0:ml
G`2]NUd20:YdF0040:5bE02IJT`0VFY<09Y[C@<0WVmA0P2SM5H0102UMUP0WFi@09e^D02FJDP209MX
BP040:AaEP2=FD40SEU107e;=080MD/e00@0JdPb079D?01bE3`0QfUN0P28LF<0102ERGT0VIR609VH
QP2KY94207b=NP0404ePC013F4D0@eQ503]1<`80@T8f00@07QD<03ha:@0n<BT0;b8J0P0/:Qh0100`
>2T0:3HU02Pf9@1SKe/208V2LP0408eeJ@1gGe<0MemC05iBA080BdTm00@0>3<]01@:2@0D2PT061lU
0P0j>SX0101PDdX0LUU3079I@`1jECP208=E<@0408eI<@2BGC40TUda08MI=@80TF<o00@0W6i:0:5c
C`2QLdl0XG=?0P2SME40102XNUH0[X1L0:j0G02[NeL20:EcD00409e[B02IID00VFE009UU@080W6I0
00@0X6]30:9]A@2RKDD0Zfm=0P2XK4X0102UJDL0X6E30:1U@`2MHT0309YR?`80VF<o00@0TEll091P
?02@H3`0TF4m0P2?HCd0102>H3`0S60k08aP>`2;GcX208YN>@0408EI=02FJTD0UVY5095U@080PeDc
00@0UVQ708iM?@2>GCd0[WeO0P2iQVT01030R6l0]7aS0;AlH`2fO6@20:ieFP040:eaE@2VJDd0YVU=
09UL?`80[W5D00@0fIAe0;QcDP2hLe80W5Pe0P2UID80102gOf@0XWUU0:9iI@1dFDh205Y7@`0404Y5
B@17AT/0AdI;04E8D@80@dUE00@0@TYG04=;EP13BeH0Ad]F0P18Be@604M>F00404Q?F@19D5X0BE1J
04YAF`80Be9L00@0C5=M04]BG01;DU`0BE1J0P18CeT01017CUP0B4mI04Q?F@18CeT204U@FP0304I<
F015BeL0AD]G00@0A4YF0P15BeL304I<F0<0BDmK00@0B4iJ04M=F@17CET0ATaH0`15BeL204A:EP`0
A4]E00<0ATeG04M>F017CUP01018CeT204M>F00304I=E`15C5H0ADaF00<0ADaF00@0A4]E04=:E013
BU@0@dYD0P14BeD304E<ER@0AdiH0`16Be401016BU<0A4]C04A;D`14BeD204=<E@0404=;EP12C5H0
@TaF03m=F080@TaE0@13C5<104I;C`40oooo000404MDG@D0B5EN0`19EUl404YGH0<0BEIO1@18EEh<
04MDGAP0B5EN3019EUl<04]HH@0404aIH@1<FEl0C5UO04I?E080=3`n00@08RHU01TI5`0I6AL061LB
0P0N6aH401dE4P806aHB0`0J5a8301PG4P0401LF4P0J6A@06QTD01PE3P806AD<00@091dC02PN500X
7Q@0:1hB0P0a9AL01017>RT0A3LV04@g9P10<20203LW5`0403@S4`0k:1T0>bPI03h[7080?2TJ00@0
=b0@04P^7`18;Ql0G44`0P1YCCP0101bDch0Ne]407]KA027ID`2091/E00408m[A`2AKDT0TFe9099`
B`80Tg5<00@0UG=>09IfD02FMU00UgMA0P2INE<0102JNEH0W7]H09akF02POe`20::1H00409mnG`2H
MeP0V7MH08m]D@80RFM;00@0RFM>095_EP2AKeH0VgUP0P2POVD0102POV<0VgaP09]lH02HNEd209Mh
G00409]hFP2MN5/0WGQK09A_DP80Pe`o00@0LdX^06Xn8`1Z?R<0LdL/0P27Fd00102GI4T0WFY?09eZ
C`2ULeP20:ekH0040;21I@2cQVT0/hIY0;F9K080]h]^00@0]8][0;J=K@2fSFd0]HYZ0P2^PF80102[
O5h0[gaO0:mlG`2]NUd20:YdF0040:5bE02IJT`0VFY<09Y[C@<0WVmA0P2SM5H0102UMUP0WFi@09e^
D02FJDP209MXBP040:AaEP2=FD40SEU107e;=080MD/e00@0JdPb079D?01bE3`0QfUN0P28LF<0102E
RGT0VIR609VHQP2KY94207b=NP0404ePC013F4D0@eQ503]1<`80@T8f00@07QD<03ha:@0n<BT0;b8J
0P0/:Qh0100`>2T0:3HU02Pf9@1SKe/208V2LP0408eeJ@1gGe<0MemC05iBA080BdTm00@0>3<]01@:
2@0D2PT061lU0P0j>SX0101PDdX0LUU3079I@`1jECP208=E<@0408eI<@2BGC40TUda08MI=@80TF<o
00@0W6i:0:5cC`2QLdl0XG=?0P2SME40102XNUH0[X1L0:j0G02[NeL20:EcD00409e[B02IID00VFE0
09UU@080W6I000@0X6]30:9]A@2RKDD0Zfm=0P2XK4X0102UJDL0X6E30:1U@`2MHT0309YR?`80VF<o
00@0TEll091P?02@H3`0TF4m0P2?HCd0102>H3`0S60k08aP>`2;GcX208YN>@0408EI=02FJTD0UVY5
095U@080PeDc00@0UVQ708iM?@2>GCd0[WeO0P2iQVT01030R6l0]7aS0;AlH`2fO6@20:ieFP040:ea
E@2VJDd0YVU=09UL?`80[W5D00@0fIAe0;QcDP2hLe80W5Pe0P2UID80102gOf@0XWUU0:9iI@1dFDh2
05Y7@`0404Y5B@17AT/0AdI;04E8D@80@dUE00@0@TYG04=;EP13BeH0Ad]F0P18Be@604M>F00404Q?
F@19D5X0BE1J04YAF`80Be9L00@0C5=M04]BG01;DU`0BE1J0P18CeT01017CUP0B4mI04Q?F@18CeT2
04U@FP0304I<F015BeL0AD]G00@0A4YF0P15BeL304I<F0<0BDmK00@0B4iJ04M=F@17CET0ATaH0`15
BeL204A:EP`0A4]E00<0ATeG04M>F017CUP01018CeT204M>F00304I=E`15C5H0ADaF00<0ADaF00@0
A4]E04=:E013BU@0@dYD0P14BeD304E<ER@0AdiH0`16Be401016BU<0A4]C04A;D`14BeD204=<E@04
04=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000404MDG@D0B5EN1`19EUl504QEGPl0
AeAM6018EEh<04UFG``0BeQQ00@0C5UQ04aIG`1<FEl0A4eB0P0`>3X0100N8R4061PF01PH5P0I61<2
01`I50@07ADB0P0K5Q8301XG4P<061LB00@05aPC01XI500J6A@05AD=0P0H50/0100V7aD0:b<H02/S
600/8QH2038Y6P0404Xm;016>BP0ASTX04@d9080>2PH00@0;ah>03DR4`0e8Q<0?2TJ0P0j9aP0100c
70`0@BLH044W601?=2<205Pl9`040652;@1`D3T0L50i085OAP80RfM?00@0UG5?09McD@2GLe40VWID
0P2KMeD309YfE00409]gE@2LN5H0W7QF09]hDP80XGiH0`2VPUh0102ROU`0XGeM0:5mG@2UPF420:Z6
IP030:MoI@2YPFL0ZH5W00<0ZHAY00<0YhAX0:J3I`2VPfL00`2VQ6P0102/RF/0Xh1R0:>0HP2@Jdh2
07]D=`0406i29`1]?bD0KClU07]<<P80S5e300@0UV1609]UB`2KID/0XVmD0P2ZN5d0102^PF@0/HEX
0;65J02bRF/20;>:K0040;><K@2fSFl0]Xe_0;B;K@80/8=V00@0[GiR0:mmHP2_OF80[GYO0P2XMEX0
102NKe<0UFI:09EVBP2EITX309Y[C`80WVmC00@0WFiB09=TB02CI4P0Rf0m0P25EcH0102EGdD0UV18
09IPB02GI4l2081G@`040698<@1B@2P0DT0X06IDBP80NfmS00@0VIJ709>GR02CUhP0RIJ40P1JJ5L0
100Z>bT0>4Ti03Q9>@0V;28201DB2`0400/0000R3Pd08Ph=01400080=BhX00@0CEE:02i0<P0^@380
DEm>0P1lN6`0102DP7L0SGU^08eiKP1RFTl2034c:00401lP6`0B30h04P`>00<<4@809BLV00@0Acdd
05U3;P1I@bh0IT@X0P1hCBX01029FC40TF0g095P=`2BIT<209mcD0040:b0G@2_Pf00[h=P0:QlF@80
WW9?00@0VVi;09]_C02KKd`0YWED0P2XMUD0102ULe00WfY809mZB02GHch209IP>P0409aV@02SKDL0
Xfe70:a]C080ZV]:00@0Y6Q60:1T@P2PI480WF900P2LHCl309YR?`0409=Q@02AH3l0TF0o08mN?@<0
Redk0P2:GS/0102;Gc`0S60m08aP?@28G3L209M[AP0408mS?P25EcD0QELe08aN?@80UFA400@0YWEG
0;5nH@2aOV40`h]b0P2`N5l0102eNf<0ZW5F0:YaEP2WJdl20:AWB`0409YM@02mP6<0_H1S0=BBL080
_7QE00@0XUhi0:IV@`2VIT<0]GiP0P2RMf@0101hFU00GDQ305e8@`1=A4D204]6BP0404M7C`13BU@0
@dYD049:E`80@d]F00<0A4]E04M;EP17BeH01P17CUP01018CeT0BU5K04YAF`1<Ded204eDGP0404iE
G`1<Ded0C5=M04YAF`80B4mI0`17CUP01018CeT0BE1J04U@FP15BeL204A:EP0404=9E@12B5@0@TQD
0498E0<0@dUE0P14BUH304M=F@@0ATaH1@15BeL<04A;E@<0ATeG00@0AdiH04Q?F@18CeT0B4mI0P17
CUP00`16CEL0ADaF04E<EP0304E<EP0404A;E@13BU@0@dYD04=:E080A4]E0`15C5HT04M>F0<0AT]A
00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?oo
o`001017E5d504QEGP<0BEIO1P18EEh304MDG@<0AU=L6018EEhH04UFG``0BeQQ00@0C5UQ04]HGP1;
F5h0@T]@0P0/=3H0100K7ah05aLE01LG5@0H5a8201/H4`@07ADB0P0K5Q8301XG4P<061LB00@06AXE
01TJ500I6Q@04Q8:0P0F4`X0100U8AH0;RHK02hV6`0/91L2034X6@0404Xm;019?2/0BC`[04/k:`80
@30P00@0<B0@03@Q4P0d8A80?2TJ0P0i9QL0100j8a<0BS0Q04X`8@1M@S4206a@>`0407]LA`25IDh0
QFE>08ITB`80PEe500@0T6a<09A`D02DL500VWIF0P2ON5T0102NMeP0WGIG09efE`2MM5H209ieE`04
09mgD`2SNeL0Xg]G0:InF`80Y7aI00@0XGQH0:ElG02UO5`0/8MY0P2jTG<0102bQV/0/hM/0;>7K02c
RVh20;B=L0040;:=LP2^Rfl0[X]_0:V6JP80Y89V00@0X7mP08i[C@2>Jdd0N5<f0P1]A2P0101_@bP0
NT/a07Y;<@26E3/208]I@00409UQB02LI4/0W6A;09m/D@80YWAI00@0ZWeP0:n3IP2_PfH0/8MY0P2a
RV/0102bRf`0]Hi_0;F>K`2eS6h20;65J0040:n2I@2`OV<0/7iS0:ajG`80YgAI00@0X75E09IWB`2F
Id/0UFI:0P2GJ4`0102FId/0VFY>09UZCP2IJTh2091QA@04091XA02=Gch0SEln08]E>`80QTdi00@0
Sea9081KB020FdP0Mf5:0P1NE3/0100j<bT0DDm3055?@`26RW`208ZBP`0408F@P01II5H0FFAF02`d
9`80<3HZ00@01PL102<K600S6aP0ET120P0J0000100X3Q408aPF02<H5P0E6Q<2020b9P0403]<>P1i
MfX0NGMZ09B2N080Pg5W00@0If5E05AGC01DEd`0@44l0P0H4Q@010000`@06QTE01XI5@0j;R8204li
900406E39`1iDbl0NE<_08AJ=080PeLb00@0REdj09AXA@2DJ4D0XWIC0P2[Oe`0102/P5d0YGUF0:Ei
EP2LL4d209IZA`0409m^C@2TLU40Y79A0:QfE@80Yg9@00@0XFa:09mYA@2OJDD0XVa60P2WLD/0102[
K4d0ZFY;0:UZB`2UITL20:1SA00409eP@@2LHD40W65109aQ@@80VV9100@0UfE409=R@@2CHT40SUdl
0P28FSP01026F3H0Qe/h08MK>02:GS/208aP?@0408]O>P2EJD@0UFU408eQ?080Sf4o00@0PU@c09IU
A@2FIDD0Tf940P2ZMeX0102oQfh0Zg=J0:]cFP2fO6@20:U`E@040:=WB`2PHdL0X6=709MJ?@80`hIY
00@0dY1^0;ahE@2lN5D0Xelj0P2WID80102fOF00YWUT0:IiI01jFTl205e4@00404m5A01<ATP0C4I8
04U8CP80ADUD0`12BUD00`14BUH0ATYE04I:E@0404I=E`80AdiH00@0BE1J04]BG01;DU`0CEAN0P1>
EEl0101?EV00CEAN04eDGP1:DE/204Q?F@<0AdiH00@0BE1J04YAF`1:DE/0AD]G0P14BUH00`13BED0
@TQD0498E0040457D`80@TQD0`14BUH404E;E`D0ATaH3014BeD01015C5H0ATeG04I=E`17CUP304Q?
F@80AdiH0`16CEL304E<EP0404A;E@13BU@0@dYD04=:E080A4]E0`15C5HT04M>F0<0AT]A00@0ATYC
04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`001017
E5d504QEGP<0BEIO1P18EEh304MDG@<0AU=L6018EEhH04UFG``0BeQQ00@0C5UQ04]HGP1;F5h0@T]@
0P0/=3H0100K7ah05aLE01LG5@0H5a8201/H4`@07ADB0P0K5Q8301XG4P<061LB00@06AXE01TJ500I
6Q@04Q8:0P0F4`X0100U8AH0;RHK02hV6`0/91L2034X6@0404Xm;019?2/0BC`[04/k:`80@30P00@0
<B0@03@Q4P0d8A80?2TJ0P0i9QL0100j8a<0BS0Q04X`8@1M@S4206a@>`0407]LA`25IDh0QFE>08IT
B`80PEe500@0T6a<09A`D02DL500VWIF0P2ON5T0102NMeP0WGIG09efE`2MM5H209ieE`0409mgD`2S
NeL0Xg]G0:InF`80Y7aI00@0XGQH0:ElG02UO5`0/8MY0P2jTG<0102bQV/0/hM/0;>7K02cRVh20;B=
L0040;:=LP2^Rfl0[X]_0:V6JP80Y89V00@0X7mP08i[C@2>Jdd0N5<f0P1]A2P0101_@bP0NT/a07Y;
<@26E3/208]I@00409UQB02LI4/0W6A;09m/D@80YWAI00@0ZWeP0:n3IP2_PfH0/8MY0P2aRV/0102b
Rf`0]Hi_0;F>K`2eS6h20;65J0040:n2I@2`OV<0/7iS0:ajG`80YgAI00@0X75E09IWB`2FId/0UFI:
0P2GJ4`0102FId/0VFY>09UZCP2IJTh2091QA@04091XA02=Gch0SEln08]E>`80QTdi00@0Sea9081K
B020FdP0Mf5:0P1NE3/0100j<bT0DDm3055?@`26RW`208ZBP`0408F@P01II5H0FFAF02`d9`80<3HZ
00@01PL102<K600S6aP0ET120P0J0000100X3Q408aPF02<H5P0E6Q<2020b9P0403]<>P1iMfX0NGMZ
09B2N080Pg5W00@0If5E05AGC01DEd`0@44l0P0H4Q@010000`@06QTE01XI5@0j;R8204li900406E3
9`1iDbl0NE<_08AJ=080PeLb00@0REdj09AXA@2DJ4D0XWIC0P2[Oe`0102/P5d0YGUF0:EiEP2LL4d2
09IZA`0409m^C@2TLU40Y79A0:QfE@80Yg9@00@0XFa:09mYA@2OJDD0XVa60P2WLD/0102[K4d0ZFY;
0:UZB`2UITL20:1SA00409eP@@2LHD40W65109aQ@@80VV9100@0UfE409=R@@2CHT40SUdl0P28FSP0
1026F3H0Qe/h08MK>02:GS/208aP?@0408]O>P2EJD@0UFU408eQ?080Sf4o00@0PU@c09IUA@2FIDD0
Tf940P2ZMeX0102oQfh0Zg=J0:]cFP2fO6@20:U`E@040:=WB`2PHdL0X6=709MJ?@80`hIY00@0dY1^
0;ahE@2lN5D0Xelj0P2WID80102fOF00YWUT0:IiI01jFTl205e4@00404m5A01<ATP0C4I804U8CP80
ADUD0`12BUD00`14BUH0ATYE04I:E@0404I=E`80AdiH00@0BE1J04]BG01;DU`0CEAN0P1>EEl0101?
EV00CEAN04eDGP1:DE/204Q?F@<0AdiH00@0BE1J04YAF`1:DE/0AD]G0P14BUH00`13BED0@TQD0498
E0040457D`80@TQD0`14BUH404E;E`D0ATaH3014BeD01015C5H0ATeG04I=E`17CUP304Q?F@80AdiH
0`16CEL304E<EP0404A;E@13BU@0@dYD04=:E080A4]E0`15C5HT04M>F0<0AT]A00@0ATYC04A;D`14
Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`001017E5d504QE
GP<0BEIO0`18EEh304MDG@@0AU=L0P15DU/H04QEGQP0BEIO301;F640101;F600BUMM04YGG@11BTl2
02Xb=00401TM700F5Q@05QHD01PG4P806QLB100M5A8201/F4P<06QLB0`0H5a80100J71L061TC01PI
4`0@40P201@A200402DQ5P0]9a/0;BLK02`T5`80;bHG00@0AcXY04Tl:`19?2/0D40`0P17=bL0100f
9AD0=R<D03HS500n:a`203XW600403hW5`109QL0@2HG044V5@80BBdH00@0GChY075A>P1aDCX0N5Hm
0P1dD3P01025H4<0RfI908]VB@2EKU4209YcEP0309acE`2JLED0VW5E00<0W71E00@0XGIC0:=hE@2S
N5D0YGYH0P2TNEL0102SN5P0Z7eM0:QmG@2bQVT20;Z>L@040;B7JP2eR6/0]HQ[0;N;KP80^I1d00@0
^I9e0;:=LP2bSG80Z8EY0P2POF40102IN5T0PEi0085N@01[ART206U0900407U=<P2;G480Rea2095O
AP80S5Q000@0WfE=0:1XC`2PJ4l0XVaB0P2TLUL0102XNeh0[X9U0:j2I@2`RFX20;><K@040;2<K02e
SVl0]Hi_0;J=K`80/hMZ00@0/8=V0:n0I02_P6@0[7YO0P2VLeP0102VMe/0VV]?09Y[C`2GJ4`209QY
C@0409IWB`2IJTh0VFY>09]/D080TfA800@0R60l09Y/B`2JK4/0Uem60P24BSH01021CS/0Ldhk07=>
>`1gIDl205]E?@04020R5`0=40D03A0504YBA@80OXUk00@0SIR:05aREP1LHUH0=3H[0P0m?C<0100Q
81`03P0000h0003C^;d2079?EP0402T83`0N3Pl07Ph?04A6@@807RhT00@0;3d[061ND@1PGU40S7Y`
0P2DPWP0101PFTh07B0E01dP5@0>3`X2018<3P0400L80`0Q7A@08AdD040c8`80G4<]00@0MULj08UU
@`29ID<0R60m0P1kDc001020E340Oe<`07mC<020E34208EI=P0408eQ?P2AID80TFE208mS@080Rell
00@0TV5109IT@`2FI4<0VfU80P2SKT`0102WLU00Z79>0:QbCP2UKd/20:9/B0040:]ZC02YJ4X0ZFQ:
0:EVA`80Wf9300<0WF1109]P@02KH4000`2LHD40102KJDP0UFA309ET@`2>GC`208IH=P0408=E<`24
F3D0Q5Pe08QL>@80Rell00@0SF4l09=W@P2CId80SF4l0P2JK4X0101oDC00UfI609MVAP27ESP20:ai
G0040;^3JP2WKeH0YfmF0;IlI080ZG1E00@0XFE909ePA02MH4@0TeHi0P34QfX0103@SV/0^gMB0;]g
DP2RGST20:IT@@040;MnH02ZNfL0ZW]W07aJD080FT4l00@0DDI404m6A`1?ATL0BDQ=0P15BE801013
BED0@TYE049:E@13BED204E9E0<0ADaF00@0ATeG04M>F017CUP0BE1J0P1<Ded0101>EEl0D5MQ051G
H@1@Ef4204iEG`0404]BG018CeT0B4mI04M>F080B4mI00@0BE1J04YAF`1:DE/0ATaH0P15BeL00`13
BED0@DMC0457D`060416DP040498E013BED0@dUE04=9E@80A4YF0`15BeL304I<F0`0A4]E00@0ADaF
04I=E`16CEL0AdiH0`18CeT304M>F080ATeG0`15C5H01014BeD0@dYD04=:E013BU@204A;E@<0ADaF
9017CUP304I;D@0404I:D`14Be<0A4]C04A;E@80@daE00@0@d]F049<EP12C5H0?deH0P12C5D104=<
D`40AT]?0@3oool000<0AU=L0`17E5d00`18EEh0BEIO04UFG`0304UFG``0AeAM7018EEh304UFGad0
BUMP00@0AeQO04UHG@19F5d0>dU<0P0Q:b`0100B61H051TE01@I5@0J6aD201TJ50@07QLA0P0L5a<3
01/G50<06APF00@05aTD01TJ500I6Q@04Q8:0P0J5`h0100O6a00?3HZ03`f:P0]9AP203<Z6`0404Xm
;@1=?2`0CC`/04/f8`80AB`H00@0@RLD048X5`12:1L0?RLI0P0g91H01013:1d0?bDH03lU60109QT2
04T_7P0405Ln:P1UC3H0ID`f075I@@80NF5900@0TFeO08aXF02<J5P0S6MG0P2AK5X0102AJUT0RfEB
08]UDP2:HDl208aSD@0409M]E@2BJ500TVQ@09iaFP80Uf]B00@0U6I<0:EfG02UMU`0[7eQ0P2aPVH0
102`P6`0/X9^0;:2KP2aPFd20:alIP040:YjH`2VMUl0YWIO09]/DP80Sf1600@0NEHf07IB<P1fDS80
MTd_0P1kCRl01026ECD0U5do09AM?`2LHdD20:5VB00409UVB`2QKe@0XFmD0:MeFP80YGIJ00@0Z7UM
0:j1I02^PF@0/HAW0P2`PfH0102cQ6P0]8EY0;B5J@2eQVX20;B5J@040;>4J02_P6@0[h1T0:amH@80
ZW]O00@0YGIL09i_E@2NKeD0UVM=0P2CI4X00`2EIT`0V6U?09QYC`0309Y[D@0409M`D`2IJe40VF]A
0:1/E`80W6AE00@0PE=606U<?P1YC3h0F4`l0P17B3H0100i?SX03Q<=00hC3@042@8203e3>@0405mU
F@1@ETP0D5I80417=`80<3LW00@0;a/M01l42@0O10T0=a8J0P0d2A8010177bL0:0h?02P>3`0W71H2
03Hf;00402/`;01YJfP0JF]X06i`J`80=cHb00@04Pl:00`70@0<1`4050d70P0P6A<0101D>2`0GD0b
05e0<P1]CCh207eJAP0408IRBP2:HTP0RV9808QO@@80QU/k00@0OeHf08eTA02=I4@0TFI40P2GK4X0
102IKDX0UVQ409IXA02EId<208MI=@0408]J>@2BH3l0TV0o09YXA`80XVe;00@0YW5?0:I`C02VL4`0
Y6i:0P2SKDL0102UJ4`0ZVeA0:Y]D@2SITX209mRA@040:=VA`2NHD80WV5209YN?080XVI400@0Uf<k
09IU?@2FICd0S60m0P1kE3D0101aC2l0M4lb07A?<P1oF3T208MN?P0408YM=P2;GCT0Redi09Y/BP80
Sf5000@0RU/m09=TB02CI4P0T6570P2PKUD0102mQFX0^81U0;R0I@2ZLeD20:QaD`040:9[C025Cbd0
QDl]0:I`CP80/WaJ00@0b8YQ0<9dCP32M4h0/UPf0P2eHT80102nPfD0X7YS0:1jH`1bET/206=5@`04
04U8C@19B4h0BDQ>04Q8D080ATUB00@0ADUD04=9E@13BED0@TYG0P11BUT01016C5P0AD]G04E;E`15
BeL204I<F00404Q>FP1:D5`0BU1L04aBGP80CE=O00@0D5MQ04iEG`1>EEl0C5=M0P1:DE/604Q?F@04
04M=F@16C5P0ATaH04A:EP80@TQD1011Ae<60498E0<0@dUE1@14BUH<04=9E@L0AD]G1@16C5P304M;
EPT0ATYE1019CEP304Y>F@D0BdmJ6019CEP304I;D@0404I:D`14Be<0A4]C04A;E@80@daE00@0@d]F
049<EP12C5H0?deH0P12C5D104=<D`40AT]?0@3oool000<0AU=L0`17E5d00`18EEh0BEIO04UFG`03
04UFG``0AeAM7018EEh304UFGad0BUMP00@0AeQO04UHG@19F5d0>dU<0P0Q:b`0100B61H051TE01@I
5@0J6aD201TJ50@07QLA0P0L5a<301/G50<06APF00@05aTD01TJ500I6Q@04Q8:0P0J5`h0100O6a00
?3HZ03`f:P0]9AP203<Z6`0404Xm;@1=?2`0CC`/04/f8`80AB`H00@0@RLD048X5`12:1L0?RLI0P0g
91H01013:1d0?bDH03lU60109QT204T_7P0405Ln:P1UC3H0ID`f075I@@80NF5900@0TFeO08aXF02<
J5P0S6MG0P2AK5X0102AJUT0RfEB08]UDP2:HDl208aSD@0409M]E@2BJ500TVQ@09iaFP80Uf]B00@0
U6I<0:EfG02UMU`0[7eQ0P2aPVH0102`P6`0/X9^0;:2KP2aPFd20:alIP040:YjH`2VMUl0YWIO09]/
DP80Sf1600@0NEHf07IB<P1fDS80MTd_0P1kCRl01026ECD0U5do09AM?`2LHdD20:5VB00409UVB`2Q
Ke@0XFmD0:MeFP80YGIJ00@0Z7UM0:j1I02^PF@0/HAW0P2`PfH0102cQ6P0]8EY0;B5J@2eQVX20;B5
J@040;>4J02_P6@0[h1T0:amH@80ZW]O00@0YGIL09i_E@2NKeD0UVM=0P2CI4X00`2EIT`0V6U?09QY
C`0309Y[D@0409M`D`2IJe40VF]A0:1/E`80W6AE00@0PE=606U<?P1YC3h0F4`l0P17B3H0100i?SX0
3Q<=00hC3@042@8203e3>@0405mUF@1@ETP0D5I80417=`80<3LW00@0;a/M01l42@0O10T0=a8J0P0d
2A8010177bL0:0h?02P>3`0W71H203Hf;00402/`;01YJfP0JF]X06i`J`80=cHb00@04Pl:00`70@0<
1`4050d70P0P6A<0101D>2`0GD0b05e0<P1]CCh207eJAP0408IRBP2:HTP0RV9808QO@@80QU/k00@0
OeHf08eTA02=I4@0TFI40P2GK4X0102IKDX0UVQ409IXA02EId<208MI=@0408]J>@2BH3l0TV0o09YX
A`80XVe;00@0YW5?0:I`C02VL4`0Y6i:0P2SKDL0102UJ4`0ZVeA0:Y]D@2SITX209mRA@040:=VA`2N
HD80WV5209YN?080XVI400@0Uf<k09IU?@2FICd0S60m0P1kE3D0101aC2l0M4lb07A?<P1oF3T208MN
?P0408YM=P2;GCT0Redi09Y/BP80Sf5000@0RU/m09=TB02CI4P0T6570P2PKUD0102mQFX0^81U0;R0
I@2ZLeD20:QaD`040:9[C025Cbd0QDl]0:I`CP80/WaJ00@0b8YQ0<9dCP32M4h0/UPf0P2eHT80102n
PfD0X7YS0:1jH`1bET/206=5@`0404U8C@19B4h0BDQ>04Q8D080ATUB00@0ADUD04=9E@13BED0@TYG
0P11BUT01016C5P0AD]G04E;E`15BeL204I<F00404Q>FP1:D5`0BU1L04aBGP80CE=O00@0D5MQ04iE
G`1>EEl0C5=M0P1:DE/604Q?F@0404M=F@16C5P0ATaH04A:EP80@TQD1011Ae<60498E0<0@dUE1@14
BUH<04=9E@L0AD]G1@16C5P304M;EPT0ATYE1019CEP304Y>F@D0BdmJ6019CEP304I;D@0404I:D`14
Be<0A4]C04A;E@80@daE00@0@d]F049<EP12C5H0?deH0P12C5D104=<D`40AT]?0@3oool0000304IC
G017E5d0AeAM00<0AeAM00<0B5EN04UFG`19EUl00`19EUl<04MDGAP0B5EN0`17E5d404QEGPD0BEIO
601:Ef001017F5l0B5ML04QGG00jB4/2024[;00401<I5`0F61D05QPE01XK5@806APC100M5Q0201/F
4P<06QHC0`0H5aD0100F5a806a`F01/L5P0F5Ph201dJ4@0401lK400k<bP0>c<X02dU6080>C0Q00@0
BSX[04/j:P1;>RX0B34O0P109a40100n8`h0@2HE040V5@12:ad2044[7P0404P]8P19;R<0BBhS04X`
8`80CC<T00@0DcTZ05To;P1I?bh0FT4]0P1I@2`0101N?S40ECDX05De:01C<BD205Xi:P0405lk:`1Q
?2`0HC`/06Y5<`80N5=100@0NE<n07Q?>`1hCc/0QUe70P25FdD01025Fd<0Uf]B09M[DP2IKE@209Y^
E@<0UfYE00@0TfIA08]NA`2;GTL0QUY10P22ESd0101lD3D0LdL/07=7;01dD30207=?;`0407I=;`1m
D340OE0a08UH>080UV5200@0X6M90:I[C@2VJdd0XVmD0P2XMU/0102[NEh0YgQL0:MhG02VMe/20:Qk
GP040:UlG`2WNUd0YgYM0:enHP80/85U00@0/hAX0;B5J@2dQFT0/H9V0P2YNUh0102QLUH0W6eA09a]
D@2SM5P209e^DP0409EVBP2@HDD0T655099SA`80UfQ<00@0WVmC0:5bEP2QLUH0VgAC0P2KKU40102R
KUP0XVeK0:9]F`2?HE4207QIA`0405e@?@16A2l0AT@_03Th<`8071`D00@05AD=03Xj<00j>S00ATHj
0P0e=RP0100c=2@0;c0P02l`8011;bl2030F6@0404<P9P0a20h0<@P>038>4080:0h=00@0;B<J02DU
6@0U9AT0;b`S0P0e<BP0100T7A<04`X101<:0@0H3P@2020D200402`N4`14=R/0A3H[07=CA080N5M6
00@0PEi;08IRC026HT`0RV990P29H4@01028GCd0QeXi08MJ>@25G3`208mVAP0408mT@P2EJTP0UFY8
09Y^B`80Vfe900@0X79>09IXA02FJ4@0W6]:0P2MJdX0102OKD`0Y6m=0:A_C@2WLU020:QbCP040:A^
BP2QJdD0XF]50:=VB@80YfY=00@0XfI709mR@`2OHT<0XVI40P2NHT00102IGCT0WF4m09eQ?@2GHc/2
099R?00408QM>`1jDc@0NU<d06e:;P80KDP[00@0NE8c08EN?`25GSl0SF4l0P2>HSl0102FJ4H0TF=2
095S@P2;G3h208eN@P0409ARB@2QKeH0XFmF0<69KP80^X9W00@0[7EG0:]dEP2[M5H0Xfa=0P28DS00
102XLU00]h5O0;N1G`36Rf<20;ibCP040:iH=`2dHdD0]6=50;f4J@80WW]W00@0KeI?05m5AP1OADH0
BTU>0P19BTl01019BE40AdYC04M:D`16BUD204A:EP0404=;F013BUX0@dYJ04I<F0<0AD]G0P16C5P0
1018CUX0BU1L04Y@G01<DUh204eCG`0404mFH01>EEl0CUEO04aCG@80BU5K1P18CeT01017CET0ATaH
04I<F014BUH20498E0@0@DMC1P12B5@304=9E@D0A4YF3013BED704E;E`D0ATaH3016BUD404U=F0D0
BTiI0`1;CeXH04U=F0<0AT]A00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2
049<E@40@daC0@16Bdl10?ooo`001017E5d504QEGP<0BEIO3017E5dH04QEGP<0AU=L0`17E5d604QE
GQP0BUMP00@0BEQO04IEFP16EEX0>DA80P0P:R/0100C6AL05QPE01HH5@0I6Q@201TH4`@07AH@0P0K
5Q8301XF4`<061LE00@051D@01dL5`0M71L061P@0P0O6a80100R6a40>C4V03Ta9P0^91P203he9P04
04/k:`1=>RT0CCXY04Xd7`80@RLB00@0?24<040U4P109A80@b`L0P13;Al01019;RD0Bc0W04/`9`17
;24203lT6@04040V6@18;R40B2hQ04T_8080A2XK00@0@bPO048V6`129Q/0ABTN0P1;;B801016:1d0
?QhA03hN4@0n7Q4204HV6@04054^6`1=:QL0CBXG05Tf8P80EC0M00@0DblI060j9@1P>RD0HC/T0P1Q
>280101Z@B/0KTD_06i5;`1_AS0206]3:P0306a4:P1bBS00LTX`00<0MTlb00@0Me<c07YC=01jDc@0
OUDg0P25F3T0102<Fc/0UF1109EP@@2LIDH20:9YB`040:5^D`2ULeP0YG=H0:QfF`80YGIJ00<0Y7EI
0:IiG02VNE`00`2WNUd40:]lH080ZGYN00@0YGIJ09m`E02OL5@0V6U=0P2EITX0102MKdh0Vfe<09]]
C02GJDP209=UA004099T@`2FJ4L0UVQ709i`C`80Y7IE00@0XG]G09icD`2NLe<0XgAH0P2[NF80102R
N680TgEM09=eG@1lJU8206AJ?`0403/a9`0h;R@0>2hT04<i;@80H5I:00@0FE1103P_800h;b00:1l>
0P0G3P00100m:RH090`:02@<2P0h6AL202d;2P0402@50P0U3PP09@h802dQ5@8071T:00@0:1P801h>
000N3P005PD00P0[60L0101@>bP0GDHd05e6=01RBCD207=HA@0408QSD02<IU40S6IA08mVD080SVA;
00@0SF5408]N?`2;GSl0SElm0P2AH3l0102CJTX0VG1@09U`D02DJDL209Q]B`0409]_C02KKDT0Vfe9
09maC@80UVQ400@0SUdl08mM?02?GC`0TEln0P2HHd40102OJTP0YFm;0:E_B`2UKd/20:A^B0040:=W
A@2VJTP0YVY80:5U@@80WV8n00<0XFE009iR?@2NHSd00`2IGSH0102CHcd0REdj08UM>P20EcL207EB
=P0406I4:@1M>b00GC/P06Y7:`80OeXm00@0SfA2091U@`2@ID<0Sf930P2CITL0102?H4@0QePn08MH
?P2EHdX20:5_EP030<F<L@2mQ6T0_HAY00<0[WEH00@0Y6]=08eE=02=EC@0Zg=B0P2oQfH01033RF<0
^W5@0;YaD02/FC/20;9WBP040;^7K`2IO6`0VGa/06UFDP80FTE:00@0BT]?04Y;D01:Be00B4]B0P18
Be@01017BeH0AD]G04E;E`15BeT204E;F`0404I<F015BeL0AD]G04E;E`80ATaH00@0B4iJ04Y@G01:
D5`0C59N0P1=Del0101?EV00CUEO04iEG`1<Ded204YAF`0304U@FP18CeT0B4mI00<0B4mI00@0AdeI
04I<F016C5P0A4YF0P12B5@40457D`H0@TQD0`13BEDA04A:EP0404=9E@14BUH0A4YF04A:EP<0AD]G
0`16C5P204M=F@<0ADUD2@16BUD304Q<E`<0BDeH101:CUT204]?FQP0BDeH0`16Be401016BU<0A4]C
04A;D`14BeD204=<E@0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000404MDG@D0
B5EN0`19EUl<04MDGAP0B5EN0`16De`304MDG@H0B5EN601:Ef001019F5l0AUEJ04IEFP0iA4P2020Z
:`0401<I5`0F61D05QPE01TJ50806APC100M5Q0201/F4P<06QHC0`0H5aD0100D5A007A`G01dL5`0H
610201lK4P04028K4@0i<BH0>C4V02hT6080?SDV00@0Bc/[04dj:@1=>RT0BS@O0P129a80100l8@`0
@2DB040U4P13;1`204<]7`0404T^9@1;<2L0Bc0W04L/8@80?b@I00@0@2HI04P^8@18;R40BBlP0P14
:Q/01013:1l0@RHK048V6`15:Ah204/]8P0404HX7@0n7Q40?QhA03hN4@80ARHI00@0DBhK04dZ5`1=
:QL0FCHR0P1E<1d0101C;aT0H3XU060j9@1Q>b@2064h8P0406Y1:`1^ABl0KTD_06m6<080Jd<Z00<0
K4@Z079:<01bBS000`1fCc80101gDc<0NU<d07YC=01nECL208EH>@0408aK>`2EH440UF1109aUAP80
XVU;00@0XFiC0:EcF02ULeP0Z7IK0P2UMUX00`2TMET0YWUL0:IiG0030:MjG@@0ZgaP0P2YNUh0102U
MUX0Wg1D09m`E02HJDd209EVBP0409e_CP2KKD`0Vfe<09MYB080TfE400@0TVA309IXA`2FJ4L0WW1?
0P2TMUD0102QNeL0WW=C09icD`2SM5P20:]iHP040:9hHP2CMEd0TgEM07aZDP80I5Xo00@0>c4W03P^
900h;R@0@cT]0P1PETX0101ID440>2lP03P_800X7`h201L>000403dZ9P0T30X090`:03PI5`80;@/:
00@090D202D>200U3PP0;B4E0P0L6@X0100X60P07Ph001h>000F1@0202/H1`04050k:01MAS@0GDHd
0699=@80LeQ500@0R6=@08aVD@2<IU40SfI@0P2>I4/0102=HD@0Reho08]N?`2=Gcd2095P?`0409=Z
BP2IL500VG1@09AYA`80V6e;00@0Vfm<09]]B@2KKDT0Wg5=0P2FJ4@0102>GC`0Sedl08mM?02AGch2
09QS@@0409mZB02UKd/0YFm;0:E_B`80Y6i800@0XfM50:IZB02VJTP0XFE10P2NHSh00`2QID00WV8m
09iR?@0309UN=P0409=S?@29GCX0REdj081G=`80ME8f00@0IT@Y05dk801M>b00JTL[0P1oFSd0102?
I480T6E3091U@`2?HT<209=VA`0408mPA027F3h0QePn09ESBP80XFmF00<0aHaa0;f4J@2mQ6T00`2^
MEP0102TJdd0SEDd08eE=02[Le820;n7IP040<>9H`2jLE00^W5@0:aI>`80/VM:00@0^hM_09UlK02I
O6`0JEIB0P1JADX0101:Bdl0BT]@04Y;D018Be8204Q;E00404M;EP15BeL0AD]G04E;F@80AD]K00@0
ATaH04E;E`15BeL0AD]G0P16C5P01018CUX0BU1L04Y@G01<DUh204eCG`0404mFH01>EEl0CUEO04aC
G@80BU5K00<0BE1J04Q?F@18CeT00`18CeT01017CET0ATaH04I<F014BUH20498E0@0@DMC1P12B5@3
04=9EA40A4YF00@0@dUE04A:EP14BUH0A4YF0`15BeL304I<F080AdeI0`15BE@904I:E@<0B4aG0`19
CEP404Y>F@80BdmJ6019CEP304I;D@0404I:D`14Be<0A4]C04A;E@80@daE00@0@d]F049<EP12C5H0
?deH0P12C5D104=<D`40AT]?0@3oool000<0AeAM2@18EEh<04MDGAP0B5EN00<0AE9K04ICG016De`0
0`16De`304MDG@<0B5EN601:Ef001018Eeh0@e1F04=@EP0d?d<201lW:@0401DI600G5aD05aLE01XI
50806QLB100L5@l201XE4@<06ADB0`0G5Q@0100E510071/F01`K5P0F4``201hJ4@0402HO5@0m<bT0
?C<Y02`R5P80@3@V00@0CC`/058o;P1B?bh0EC`X0P1<<Q/01016:Q@0A2TF04@Y5P14:Q/2044Z7004
044V7@17;2<0Ab`S04@Y8080>b0E00@0?24F04T^8`19;R<0CC8W0P17;240100n9b40@R/S048[8`13
:R<203/S6@0403@I400b5``0<QL<03TM4P80?R8F00@0@B@D03dP400m8100BR/I0P128a40100o7P/0
CRdJ04h]6P1H=B4205hk9`0406<o9`1XA2`0J4@/06Y6;P80Id<Y00@0Id@X06m<<01_C300ME8d0P1g
E3H0101jDc@0P5Li081G>@28G3l208aO@004099Q@@2FHd@0UV=409iYB@80Xfa=00@0X6eB0:9`E@2R
L5D0YG=H0P2RLeL0102QLUH0XGAG0:5dE`2SMUT20:AgFP040:enHP2WN5`0YgQL09m`E080V6U=0`2E
ITX0102HJDd0Vfa@09]/D02IJdL209a^BP0309e_B`2KKDT0Vfe900<0UfU500@0W6i:0:5cC`2QLdl0
/HUU0P2ZOed0102[OUl0[X5T0:j1I02VOV@209amH@0408mgF`1lJdd0O6]=07USE@80KeU;00@0IE0o
06aGAP1/EdH0IE0o0`1H@c8204Xe900403XV7@0O20007`P0038G3P80?AlG00@0=a/@02dG2P0]5`X0
:AT90P0T70T01012:0l0DCLN054g7P1K?bL206]=<`0408ATB`2BKU@0TViD091/DP80U6mD00@0TfUA
09IYDP2FJE80UFU>0P2AI4L0102;G3h0RUTh08YI>02;FcL208eM>@0408QO?`2<Hd<0S6=308IK>@80
Rf0n00@0S60m08UK=`29FcL0SElk0P24ES801024Dc80RELf08UG=P2>G3/209EP?P0409MR@02KID40
VfE109mYA@80XF]50`2VK4L0102PIT00Vf4k09]Q>`2MHc/209mV>`0409aS>02GGS<0Uehc08mQ?@80
PULe00@0ME0c06e;<01]Bc00FC/S0P1;;AD0101I>B00MU@i07ID>@2;HT@208eTAP0408QL?`2HJU00
V6Y@09EVC080QeDl00@0UVA;09m[D`2OJe<0`XU^0P2kPVL0102]M5L0[7=F0:acEP2OITP208mG=P04
0:]cDP32RVT0`XYY0;n7HP80]Vm?00@0ZEPk0;9XC`2bJ4l0^8Qb0P2ENfh0101TE5@0EDE?05E5C`19
CE0204U<D@0404U<D`18C5D0B4aE04Q<E`80B4aH00@0B4]J04Q;G018Be`0ATaH0`15BeL204I<F004
04Q>FP1:D5`0BU1L04aBGP80CE=O0`1>EEl0101<Ded0Be9L04]BG019D5X304Q?F@80AdiH00@0AdeI
04I<F016C5P0A4YF0P12B5@40457D`H0@TQD0`13BEDA04A:EP@0@dUE0P14BUH00`15BeL0ATaH04I<
F00304I<F00304A8D`15BE@0ADUD00@0ADUD1@16BUD404Q<E`<0BDeH1@1:CUTH04U=F0<0AT]A00@0
ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`00
2@18EEh?04MDGAP0B5EN00<0AE9K04ICG016De`00`16De`304MDG@<0B5EN601:Ef001017E5`0?dY@
03m:D00^=c`201hS9P0401DI600G5aD05aLE01/H4`806QLB100L5@l201XE4@<06ADB0`0G5Q@0100F
5A406aPC01/H4`0@2`D201hG3`0402hU7013>Bl0@cT_02dQ5@80?S0S00@0BCPV05@o;01D?b`0G4<]
0P1J?RP0101G>2<0DCDP054e801:<1l2048[6`0403/Q5018;R40B2hQ04hd9`80BS0S00@0CC8W05Dj
;`1E>Rl0E3T^0P1;<2D0100/6140?BTR03dY8P14<2L203hW7`0303`T6P18<2H0B30V00<0DcP]00@0
B2`N04H[6P16:aX0DcPW0P1;;Q`010159Q@0Dc@P05<d801M?2T206E4<@0406I4:`1/BS40K4Xa06i<
<`80JdT^00<0JTP/06i<<01^C3000`1bDC80101kCc80PEDh085E>02;GT4209=TAP0409]ZBP2QKTl0
XFi?0:QfE@80[gYJ00@0Z7EJ0:MeFP2WMEX0YWAI0P2RLeL0102MKU80VFa?09U/C`2HJdh209Y]D004
0:YkG`2SM5P0XgAH09Y[C`80TV=700@0T65509AUB@2DIDT0Vfa@0P2PLED0102JK4P0W6i:09a^BP2L
KTX209MYA@04091R?P2>H3`0SV0l09=U@@80VF]700@0WWIC09adD@2LM540Y7]K0P2YP680102SOV40
Xh5U0:>1I@2SQVP209^0H`0409]lI`2CM5l0TgAO08MXD`80RfaG00@0RV]F08Y[E`2:JeL0TG9N0P23
I500101eFdh0JU1306Y@@`1`E4H207UMC`0407MLB`1]Dd80KE=205I1;080GTTd00@0QFI909AbEP2D
LUH0Tg1D0P2;ITT0102?J4T0V6mA09Q_D@2MLU820:1cE00409e_E@2LKU@0W6iD09U/C`80TV=500@0
RE/j08AC<P24Dc80PE0_0P21DBd01021F3P0Q5/k08AK>`1nDc4208=H=P0408IJ=`26F3@0QUPd095S
?`80T68n00@0QE@c095O?P2AGch0W6Y90P2QK4X0102NJDL0WVQ409iXA02TKTX20:YdCP040:]cCP2U
KDP0YFe809eV?`80UUlh00@0V64h09iX?02NJ3`0WVPl0P2EGc<01026FcT0MDh_07E>;`1XAR/205m1
:@0404/`6`0m8@/0?B4;04h`6080Jdde00@0PEa208=N@`23GT<0PETo0P2HK5<0102KKED0SUa508iL
A@2HI4h209YVCP040;UmH`2hO680^7aR0:a`E080Z6a@00@0V5do09=H>02CF3P0Zg1@0P32QfL0102m
QF00]Fi>0;E^CP2ZFC`20;5YD@040;F6LP2<MV/0S7I[05a@DP80DTIB00@0BDiA04U>DP19CU80BDiD
0P19CEH01019CEP0BTaI04Y<F@1:C5/204Y;G@0404I<F015BeL0AD]G04E;E`80ATaH00@0B4iJ04Y@
G01:D5`0C59N0P1=Del0101>EEl0CEAN04eDGP1=E5h204aCG@0304YAF`19D5X0BE1J00<0AdiH00@0
AdeI04I<F016C5P0A4YF0P12B5@40457D`H0@TQD0`13BED504A:EP`0AD]G1013BED304A:EPD0AD]G
0`13Ae8304A8D`0304E9E016BUD0ATYE00<0ATYE0`17BeH304Q<E`<0BDeH0`1:CUTH04U=F0<0AT]A
00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?oo
o`002@18EEh?04MDGAP0B5EN00<0AE9K04ICG016De`00`16De`304MDG@<0B5EN601:Ef001017E5`0
?dY@03m:D00^=c`201hS9P0401DI600G5aD05aLE01/H4`806QLB100L5@l201XE4@<06ADB0`0G5Q@0
100F5A406aPC01/H4`0@2`D201hG3`0402hU7013>Bl0@cT_02dQ5@80?S0S00@0BCPV05@o;01D?b`0
G4<]0P1J?RP0101G>2<0DCDP054e801:<1l2048[6`0403/Q5018;R40B2hQ04hd9`80BS0S00@0CC8W
05Dj;`1E>Rl0E3T^0P1;<2D0100/6140?BTR03dY8P14<2L203hW7`0303`T6P18<2H0B30V00<0DcP]
00@0B2`N04H[6P16:aX0DcPW0P1;;Q`010159Q@0Dc@P05<d801M?2T206E4<@0406I4:`1/BS40K4Xa
06i<<`80JdT^00<0JTP/06i<<01^C3000`1bDC80101kCc80PEDh085E>02;GT4209=TAP0409]ZBP2Q
KTl0XFi?0:QfE@80[gYJ00@0Z7EJ0:MeFP2WMEX0YWAI0P2RLeL0102MKU80VFa?09U/C`2HJdh209Y]
D0040:YkG`2SM5P0XgAH09Y[C`80TV=700@0T65509AUB@2DIDT0Vfa@0P2PLED0102JK4P0W6i:09a^
BP2LKTX209MYA@04091R?P2>H3`0SV0l09=U@@80VF]700@0WWIC09adD@2LM540Y7]K0P2YP680102S
OV40Xh5U0:>1I@2SQVP209^0H`0409]lI`2CM5l0TgAO08MXD`80RfaG00@0RV]F08Y[E`2:JeL0TG9N
0P23I500101eFdh0JU1306Y@@`1`E4H207UMC`0407MLB`1]Dd80KE=205I1;080GTTd00@0QFI909Ab
EP2DLUH0Tg1D0P2;ITT0102?J4T0V6mA09Q_D@2MLU820:1cE00409e_E@2LKU@0W6iD09U/C`80TV=5
00@0RE/j08AC<P24Dc80PE0_0P21DBd01021F3P0Q5/k08AK>`1nDc4208=H=P0408IJ=`26F3@0QUPd
095S?`80T68n00@0QE@c095O?P2AGch0W6Y90P2QK4X0102NJDL0WVQ409iXA02TKTX20:YdCP040:]c
CP2UKDP0YFe809eV?`80UUlh00@0V64h09iX?02NJ3`0WVPl0P2EGc<01026FcT0MDh_07E>;`1XAR/2
05m1:@0404/`6`0m8@/0?B4;04h`6080Jdde00@0PEa208=N@`23GT<0PETo0P2HK5<0102KKED0SUa5
08iLA@2HI4h209YVCP040;UmH`2hO680^7aR0:a`E080Z6a@00@0V5do09=H>02CF3P0Zg1@0P32QfL0
102mQF00]Fi>0;E^CP2ZFC`20;5YD@040;F6LP2<MV/0S7I[05a@DP80DTIB00@0BDiA04U>DP19CU80
BDiD0P19CEH01019CEP0BTaI04Y<F@1:C5/204Y;G@0404I<F015BeL0AD]G04E;E`80ATaH00@0B4iJ
04Y@G01:D5`0C59N0P1=Del0101>EEl0CEAN04eDGP1=E5h204aCG@0304YAF`19D5X0BE1J00<0AdiH
00@0AdeI04I<F016C5P0A4YF0P12B5@40457D`H0@TQD0`13BED504A:EP`0AD]G1013BED304A:EPD0
AD]G0`13Ae8304A8D`0304E9E016BUD0ATYE00<0ATYE0`17BeH304Q<E`<0BDeH0`1:CUTH04U=F0<0
AT]A00@0ATYC04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl1
0?ooo`000`19EUl404QEGQ40AeAM6018EEh304ICG0<0AeAM1P18EEhH04YGH004049?E`0iA4X0>DA:
02L`=@806QlR00@05QPG01PG5@0H5aD06aPC0P0K5Q8401dF40806aHB0`0J5Q<301PG5@0401PD4@0L
6A@071TD010;1@808AXB00@0>2hU04/n=@1;?SD0;b4F0P0n<2<01016<b80D3/V050k9P1L@R/20653
:`040694;01O@bd0Gd<]05Pm:P80CS@U00@0@bXF04`c7`1<<al0D3HU0P1=<b80101<<R<0D3HY050f
:@1>=2L204P^8@0403PT6`1;=bh0BcL^05Dn=P80BS@Y00@0A2`R04P`9018<2@0BRlT0P12:1/01011
9QD0?b@C03lT4`1=<1h204HY5`04040Q3@1::aH0BR/F04d]6080CRhI00@0Ec<K05hj8P1N>R80I40X
0P1V@RP00`1[B2`0M54e07AA=@0307QE=`0408UL?`2=H4<0SF1309=VB@80Vfa>00@0X79A0:IeE@2V
MED0ZWUH0P2^O5/0102WM5T0YG=H0:EcF02TLUL20:9cE`0409m`E02JKE00VVe@09]^D@80Wg9E00@0
YWMK0:=dF02SM5P0Wg1D0P2LKE4309Y[C`0409]/D02MKU80WFiB0:9dD`80X79A00@0Vfe<099T@`2B
I4<0RU`k0P29FcX0102?HD00UFM609EWAP2IKTh209YaD`040:IoH02ZQFP0ZXEX0:=nH@80XGiR00@0
YX5V0:AoI02TOf@0VWMI0P2ROf<0102XQFT0/Xid0;:>M02]RFl20:F1J@040:=oI`2?JeD0Sf]E0:R9
M`80/I:000@0[Xmk0:R<M`2XS7L0XXI`0P2`TG`0102HNV80W7aU09alI@2YQf/20:9oH`0409UdE`2H
LE@0V75D09acE@80W71C00@0VVe>09Q[C02HJd`0VF]A0P2GJDl0102CITT0SV5408iQA029G3d208AG
>004081C<P1nDC00OU4`07mF=P80P5Lg00@0N4d[07Y?;@1jCbd0NTh[0P1jC2P0102=Gc/0UFM309EW
@`2;FST209MUA0040:=aD02XLe40Z7=A0:9]B`80XF]700@0Yg5=0:ihDP2^N580[7AA0P2TK4T0102L
I3l0UEdh09EM>02FGcP209mX@@0409mX?`2AFS40TEXa07I?<080I44U00@0ESPP04db7@1=<Qd0?bHB
0P0e70P01017;1L0I4Pb06A8<P1hECl207MC?@0407YD?@2@ITh0T6I>09e_F080VVQA00@0VfMA09YS
CP2JHdh0ZfeD0P2dMUd0102[KU80XVE90:9UB@2@DcH209MJ>`040:a_D02oPV<0_h9S0;j3F`80^6i;
00@0[UTl0;=YD02cJE00/81Z0P23KF00101GBdd0DTIB0596DP17Ce4204M?DP0404M>E019CEH0BDeF
04Y<F080BTaI00@0C4]K04a;G@1<Bed0ATaH0`15BeL204I<F00404Q>FP1:D5`0BU1L04aBGP80CE=O
101=E5h204aCG@0404]BG019D5X0BE1J04M>F080ATeG00@0AdeI04I<F016C5P0A4YF0P12B5@40457
D`H0@TQD0`13BED504A:EP`0AD]G1`14BUH504=9E@<0@TIA00@0@dMB04A8D`14B5<0ADUD1@16BUD4
04M;EP<0B4aG7@19CEP304I;D@0404I:D`14Be<0A4]C04A;E@80@daE00@0@d]F049<EP12C5H0?deH
0P12C5D104=<D`40AT]?0@3oool000@0BEIO0P18EEh404MDG@80AU=L3017E5dH04QEGP<0AeAM1018
EEh504UFGaP0BUMP00@0@4YC03<l@`0c?4<08RT_0P0H71l00`0E5aH061LE01PG5@0301/F4P@07AH@
0P0K5Q8301XF4`<061LE00@05a<@020K5`0P6aL06A8<0P0Z8AX0100o=B`0Cd8i04m2>@0a8aP204Hf
:@0404Xh901A>bH0DC/V05Lm9P80GClW00@0I4@]06Q9=01XBC@0HDHc0P1H?Rd0101>=Q`0DCTO054i
7`1B>R8204lf800404`c7`1<<R40C38Q04dc8P80CC<T00@0ET0e05Do=01E?c@0CCDY0P0m9AT0100b
60/0<QP9038H2@0h70h203`P4P0404DY500k7`X0>al:04DV4P80A2D@00@0BBTD05Ti8P1I>B80Ec@N
0P1@;QD0101J<A/0GcHP05lf801T>bD206U1:004079:<01nES`0OUHl08=L?`80PU/n00@0VfiA09e^
DP2MKU80WVmC0P2QLU@0102UMeH0ZGQH0:UhF02YN5L20:QgEP040:MdF@2SLEH0Xg5F0:AbE`80YGIJ
00@0Y7EI0:1cEP2PLeH0XgII0P2ZOF040:YkG`80Z7UM00@0Y7EI09m`E02OL5@0VV]?0P2FId/308aM
@@0408UJ?P22DcL0PU<g07Y;;`<0MTL[0P1hBBd01024F3l0QUa308IL@`2AK54209QeF@0409IbF02M
N5h0WGQN0:Z0J080[H1Y00@0/8YV0:f7H`2]Qf<0XgaK0P2SO5d0102LMEP0WGEK09eeF`2]QF/20:Mo
IP040:J0K@2/Qg@0[8Md0:b9L`80[Xig00@0W7aS0;JBN02fTWP0ZX9Y0P2[PFT0102TQ6/0VWQM09Yh
G@2CLEH209UdF@0409mgG@2MM5P0WGAH09M[D080TFE:00@0TfE=08eQB02=HDP0Qe]00`25FCh2089I
?@0407mF>01lDcD0O5<e089I>@80Qehn00@0QEXh08UN?029GS`0QETf0P20DRh0102BI400Vfe909]]
B@2IJ4L209i/B`040:9`C`2SKT`0Xfi<09mZB080WFM300@0WVQ40:5[A@2QJdD0YVm@0P2NIdP0102J
I480UElm09EO?@2DGSX209iXA00409]U?`27DB/0Qe4[06<n8@80EC<H00@0ABTC03hU4@0n9A40=1l>
0P0d7@/01019<1`0HTTe0699=@1aDC`206e:=P0407EA>`25G4P0QEa809]/F080YgAO00@0WVYD09]T
C`2KI4l0Wf580P2aLeX0102[KU80WF1409ePA029C2l209aO@0040:e`D@2kOUl0^giO0<21EP80_6i8
00@0/E/j0;AWB`2dId/0[7YS0P1kI5H0101DATH0DdMC05=7D`15Ce0204E?D@0404I=D`18C5D0B4aE
04U;E`80BdYH00@0BdYJ04a9G01<BE`0ATaH0`15BeL204I<F00404Q>FP1:D5`0BU1L04aBGP80CE=O
1P1=E5h0101;DU`0BE1J04U@FP17CUP204I=E`0404M=F@16C5P0ATaH04A:EP80@TQD1011Ae<60498
E0<0@dUE1@14BUH?04E;E`<0A4YF00<0@dUE0498E012B5@00`11Ae<01011AE00@TIA0496D@12AU42
04=7DP0404A8D`15BE@0ADUD04I:E@P0Ad]F0`18C5LK04U=F0<0AT]A00@0ATYC04A;D`14Be<0A4]E
0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`001019EUl204QEGP@0AeAM
0P16De`<04MDGAP0B5EN0`17E5d404QEGPD0BEIO601:Ef001010BU<0<ca303<l@`0R:Bl201PL7`03
01DG5P0H5aD061LE00<06aHB100M5Q0201/F4P<06QHC0`0H5aD0100G4a0081/G020K5`0I4P`202XQ
6P0403le;01?@ST0Cd8i034S6080ASHY00@0BSPT054k9P1A>bH0EcdV0P1M?bL0101TA2d0J4Td06Q9
=01QAS<205Pn;@0404hf701A>Al0DCTO058j8P80CcHP00@0C3<O04`b8@1<<R40CC<R0P1=<b@0101F
@3D0ECld05Do=01==BT203dU6@04038H2`0b60T0<QP903PL3P80?20B00@0ABTD03/O2P0k7`X0ABHB
0P149A001019:A@0FCTR05Ti8P1G=1h2050^5@0405Xa6`1O=R00GcHP06@k9@80JD4X00@0LTX`07iF
?01nES`0Pe`o0P22Fch0102KKU40WFiB09e^DP2NKe<20:5bE0040:EgEP2YN5P0ZGQH0:UhE`80Z7MF
00@0YgAI0:=aEP2SLEH0Y79G0P2UMUX0102TMET0X7=F0:1cEP2SMUT20:YmH0@0ZW]O0P2XNEd0102T
MET0Wg1D09m`E02JJdl209IWB`<0S5e100@0REXn089C=`22DcL0NT/_0`1fAb/207Q9;@0408AH?`26
G4<0QUa3095/D@80V7EI00@0UW9H09ehGP2MN5h0ZX1X0P2]P6T0102`RVH0[HMS0:f7H`2SO5/20:=l
G@0409aeF02MME/0WGEK0:f5J`80YgmV00@0YX1]0:b7M02/Qg@0[8Uc0P2^SWL0102LO6<0]Y9h0;JB
N02ZPVT20:^1J@040:B4J`2JN5d0VWQM09=aEP80VGAI00@0WgMM09edF02MM5P0Uf]@0P2AIDX0102C
IDd0SF5808eQB027Fd0308EI?P80PUTm00@0OeHh07aC=@1lDcD0PUTi0P27GSh01025FSP0REhl08UN
?025FCH2081B;P04099T@02KKDT0Vfe909UXA`80WVa;00@0XW1?0:=^C02SKT`0WfY80P2MId<0102N
J4@0XF]50:5[A@2VKe0209iWB00409YT@P2EGcd0UElm09AN>P80WVQ400@0VfDo08MA:`27DB/0HchQ
0P1E<aP01015:A<0?RDA03hU4@0d7`h203@M2`0404T`701RBCD0HTTe075A?080KDXf00@0ME4k08EL
B025G4P0VfaH0P2WM5l0102NJU@0VfA?09]TC`2OHDP20;5cFP040:]^DP2MH4@0WF1408U<;`80W5m0
00@0[G1A0;]nG`2kOUl0`85F0P2lKTP0102aFcX0]6M;0;AWB`2/NV<207]TEP0405A6AP1CAe<0DdMC
04E?D080ADmA00@0ATeC04Q<E@18C5D0BD]G0P1;BUP0101;BUX0C4UL04a9G016C5P304E;E`80ATaH
00@0B4iJ04Y@G01:D5`0C59N0P1=Del604eDGP0404]BG019D5X0BE1J04M>F080ATeG00@0AdeI04I<
F016C5P0A4YF0P12B5@40457D`H0@TQD0`13BED504A:EPl0AD]G0`14BUH00`13BED0@TQD0498E003
0457D`040455D012AU40@TIA0496D@80@dMB00@0A4QC04E9E015BE@0ATYE2017BeH304Q<Ea/0BDeH
0`16Be401016BU<0A4]C04A;D`14BeD204=<E@0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;
C`40oooo000404UFG`80B5EN0`17E5d304ICG0`0AeAM7018EEh304UFGad0BUMP00@0?TQA030i@00`
>D007RD[0P0F6Qd00`0F5QH061LE01PG5@0301/F4P@07QLA0P0L5a<301/G50<06APF00@05a8?02@O
6`0T7a/081TC0P0b:B801014>S40D4<j0513>P0c9AX204`l;`0405=1;@1D?RT0E3hY05@j8`80EcTO
00@0GclV06M8<`1WB3<0I4Le0P1L@C00101@>AX0EChO05Dn7`1MABT2069:<0040618<01J@B/0FT4[
05Hm:@80ES`[00@0CCDY04L_8`17;b<0ARlQ0P1:<2<01018;1h0ABXI04DZ6@1><B4205/n;P040693
;P1<;AP0C2dH04TY5080AbL@00@0Dc4H06Q6;@1XARd0HShT0P1E<AL0101R>280IS`V06Hl9P1Z@2X2
0716;P0407eC>P2<HTT0S69909=ZCP80TVU=00@0XG9F09m`E02OL5@0Wg1D0P2SM5H0102YNeX0[7iM
0:anG@2/OU`20:]mF`040;1mHP2[NEh0ZgUN0:YhG@80ZW]O00@0YgQL0:5dE`2QM5L0XWEH0P2YO5l3
0;>4J0040;62IP2^Of<0[WmS0:QiG@80X75E00@0VFY>09AUB@2DIDT0Sf160P2CI4X0102GJ4h0UVM=
09IWC@2?H4H208EF?00407a=<`1gB2h0MdP^07U<=@80Ld/b00@0MU8h07YH?@1jF3d0NeU00P2:HD/0
0`2KKUT0X71L0:1`G00309YcC00409UaC@2UOEX0YGeJ0:ImG@80YWeQ00@0[HAX0:1fG@2PMUd0[H1]
0P2NMF40102VPVX0_I^00;fKP02LNUl20;28KP040:QlH`2^OVL0[WiW0:F7K`80XH=[00@0UWIO08MU
C027ID`0PUe30P2<I4/0102CJE00T6I=091VC@2EJE0208aPA`04081F?@1kDCP0Ne4h07YB>080NE4g
00@0MDdc06m:;`1_BRl0KdHV0P1lDc<01026FcT0TVM5099WA@2?Hd0208IH=00409AV@P2LKTX0W6i:
09]ZB@80VVQ700@0VFM609eXAP2MJ4H0XVe;0P2VL4`0102XLTh0Z79<0:QbC02PJD/209YSA@0409QQ
@`2EGSl0UEho09EO?@80WVQ600@0Uf4m07m99@1oBBD0Ec@H0P19:A00100k80/0<a`:03<L2P0`6`X2
03DP3`0404/d8P1SBSH0HdXf06m@?080ID@a00@0Ldhk07eD@01mE400UfYE0P2_O6L0102QKEL0WVMB
09iWDP2HFT420;1bF@040:a_D`2KGT80Vei208I9;080X6=400@0[g9C0;UlG@2iO5d0`h1C0P2mKTD0
102eFcT0]VM90;IWB@2YMEd207QND@040553@`1EB580EDQB04A@D080ADmA00@0ATeC04Q<E@18C5D0
BD]G0P1;BUP0101<BEX0C4UL04a9G016C5P304E;E`80ATaH00@0B4iJ04Y@G01:D5`0C59N0P1=Del6
04eDGP0404]BG019D5X0BE1J04M>F080ADaF00@0AdeI04I<F016C5P0A4YF0P12B5@40457D`H0@TQD
0`13BED504A:EP`0AD]G0`16C5P01015BeL0A4YF04A:EP12B5@20457D`<0@4IB0`11AE001012AU40
@dMB04=7DP14B5<204E9E00404I:E@17BeH0Ad]F04I:E@D0Ad]F0`18C5LK04U=F0<0AT]A00@0ATYC
04A;D`14Be<0A4]E0P13C5D01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`006017
E5dH04YBG@@0B5EN0`19EUl504YGH0`0BEIO00@0AeMT04MEH017EF00C5UR0P1>F640101;E5/0CeIL
04mFG01AEUX204M;CP0403@mA00T:c4092/a01LL80805QLI00@06APF01XF4`0J5Q<071D?0P0M5Q00
100K6QD06QTD01XI500I61<301PG4P806APC00@06QTD01/J5@0K6QD04A0<0P0K6QD0100V8a`0ASle
04Ho=@19@C@20599>P0403P]6`18?B/0B3d[05E1:080ET8W00@0Ed8W05To901I?b@0G44V0P1RA2P0
101ZBbl0L54e071A=@1L@RL305To9080GT<X00@0HTL/0694:P1RA2X0FS`R0P1C<aX01014:1<0F3`W
05Pl9`1QAC0205Pl9`0404lc7P1B=R40DSHQ058f8@80C30K00@0ABL?05Hh801F>200DcDM0P17:A40
101@<QX0FS`T05Xl901I>b<205Li8@0406Hl901^A2`0KT@/07A:<P80NDlg00@0Q5Y2095WC`2AIdl0
U6YB0P2>I4`30:5bEP040:1aE@2/OF40[7eQ0:5bEPD0[GiR0`2cQ6P0102aPVH0[7eQ0:amH@2VMe/2
0:9cE`040:5bEP2RLeL0XW=G0:=dF080/H9V00@0_Hib0;V:KP2iRVh0Z7UM0P2GJ4`0102BHdL0TfA8
09=TB02HJDd209i_D`0407i?<`1nD3H0OU0f07iB>@80I3PO00@0KdD]089H@022F400HSXQ0P1_Abh0
101lE3/0NU@m07YD?@1ZARh205`h8P0405`i8`1U@Rh0ID8^06M08@80LDX[00<0PEXm08mXB`2?J4/0
0`2BJU00102ILEP0Xg]R0:=kHP2VOVD20:V1J0040:Z2J02WOfD0YgmU0:N0H`80YWmR00@0WWMH09E^
C`2EKTl0WWQQ0P2CJE<0102GJ5@0WV]H09i[F02YM6820:QeHP04095RCP2<GdX0S5m:08=I@@80Q5Y2
00@0PUPo07]A>01kDCP0LdX^0P1_ARX0101bBB/0Mdh`07M><01jDC4207I=;@0407979@1/@Al0K44O
06a07@80MTXW00<0Redi09a^BP2LKTX00`2?GcT0102EHch0W6Y509aZA@2OJdH209aV@P0409eW@`2S
KDT0Xfe909U/C`80VVU900@0VfI409IP?02FH3`0V68n0P2PKTd0102CI4H0MDP[07E8:`0l80P203@N
2@04030P4@0T6Q0091X@02TP5`80;20D00@0@2lM04`f8@1<=R40G3/X0P1[B3D0101O>RL0HCHU064f
9@2:FdT20:]hI@040:eeI02/Lf80[7=R0:5SC080Vee600@0YfM>09aJ@02LFT00U4hd0P2UGT00102d
JTd0_GAD0;edE02nN5@20:eS?P040:QM=@2`Jd@0/6]40:A[CP80O5Q:00@0FDU:04i;EP1>BeH0AdQ=
0P19BTl01019C5<0BTeF04Y=EP1:CUT204U=F@0404I<FP15BeT0AD]I04I=E`<0ADaF0P16CEL01018
CeT0BU5K04YAF`1<Ded604eDGP80C5=M00@0Be9L04U@FP19D5X0AdiH0P16CEL304A:EP0304=9E@12
B5@0@TQD00<0@DMC00@0@4IB03m5D@0oAE40@dUE1P12B5@204=9E@<0AD]G1011Ae<30498E0D0@dUE
1016C5P204E;E`0404=9E@11Ae<0@DMC03m5D@80?TA@0`11AE001012AU40@dMB04=7DP14B5<204E9
E00304I:E@17BeH0Ad]F00<0ADUD1016BUD504M;EQP0BDeH0`19CU@01018C5D0ATeE04I=E@15C5H2
04A=EP0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000H04MDGAP0BU9M1018EEh3
04UFG`D0BUMP3019EUl01017Ef@0AeEP04MEH01<FF8204iHH@0404]DF`1?EU`0CeIL055FFP80Ad]>
00@0=3e402@[<@0T:c405a`P0P0F5aT0100I61H06QHC01XF4`0L5@l201dF400401/J5@0J6A@06QTD
01TH4`<061LB0P0I61<0100J6A@06aXE01/J5@0A40`201/J5@0402HS7016?cD0ASle04U1=080DTTj
00@0>2dK04Pm:`18?B/0ED4X0P1F@RL0101G@RL0FClT05To901L@BH20694:00406Y;;`1`DCD0L54e
05a29`<0FClT0P1N@bP0101RAb`0HT@Z0694:P1J?28205<c6P0404@X4`1H?2L0F3`W0655<080F3`W
00@0Cc<N058f8@1B=R40DSHQ0P1<<1/010159`l0ESPP05Hh801C=Ad204LY4@04050b6P1J?2@0FS`T
05Tk8`80EcTQ00@0IS`T06i4;01^A2`0M4Xb0P1iCcL01024FT80TFM?095WC`2DJU8208iTC0<0XG9F
00@0X75E0:amH@2/OF40XG9F1@2]OV830;>4J0040;62IP2/OF40[7eQ0:IgF`80XW=G00@0XG9F0:9c
E`2RLeL0XgAH0P2aPVH0102mSW80^HY^0;V:KP2XNEd209MXC004099SA`2CI4P0TfA809QYC@80WVmC
00@0OTlc07i@=P1nD3H0OU8i0P1T>1l0101_ABd0PUQ0089H@01R>R4206m7;P0407aD>`1jE3d0NU@m
06Y6;P80G3PR00@0G3TS06E2;P1U@Rh0Id0Q0P1aBR/00`21FSd0SfQ;08mXB`03099ZD00409UaF02S
Nf80Xg]R0:InI@80ZH5X00@0ZX9X0:MoI@2WOfD0Yh1S0P2VOf80102NMeP0UFi?09E^C`2NN64209=Y
D`0409MXE02NJeP0WV]H0:UdHP80Z7ER00@0TF9>08aOBP2<GdX0PeU10P24FT801022F3l0Ne4h07]A
>01cBRh206m6:P040799:`1gCS00Mdh`07YA<@80MTd]00@0LTLU06a17`1/@Al0K40M0P1fBRL00`2;
GCT0W6i:09a^BP0308mO>@0409ES?P2LJTD0W6Y509m[AP80W6I200@0WFM30:=]B@2SKDT0VFa?0P2J
JDT0102KIT@0UV0l09IP?02HHSh20:1^C@0409=TAP1eB2/0MDP[03`P2080=1h900@0<20A02@J400T
6Q00:B0G0P0/81@01010;ad0C3HQ04`f8@1L>bP206]8=@0405lj9`1Q=RD0HCHU08YKB@80ZgQU00@0
[GET0:acHP2/Lf80XF=<0P2KGDH0102WIdh0W5Y009aJ@02DCS@20:EN@0040;AZC@2mM5@0_GAD0;ih
E080[F<n00@0Z5de0;1[A02`Jd@0Y6]>0P1lF4X0101IBDX0CT]F04i;EP17B4d204U:C`0404U<D`1:
CEH0BTeF04Y>F@80BDeI00@0ATaJ04E;F@15BeT0ATeG0`15C5H204I=E`0404Q?F@1:DE/0BU5K04aC
G@H0CEAN0P1<Ded0101;DU`0BE1J04U@FP17CUP204I=E`<0A4YF00<0@dUE0498E012B5@00`11Ae<0
1010AU80?dEA03m5D@13BED60498E080@dUE0`15BeL40457D`<0@TQD1@13BED404I<F080AD]G00@0
@dUE0457D`11Ae<0?dEA0P0nA5030455D0040496D@13Ae80@dMB04A8D`80ADUD00<0ATYE04M;EP17
BeH00`15BE@404I:E@D0Ad]F6019CEP304U>E00404Q<E@16CED0ATeE04E<EP80A4eF00@0@d]F049<
EP12C5H0?deH0P12C5D104=<D`40AT]?0@3oool001P0AeAM601:DUd404QEGP<0BEIO1@1:Ef0<04UF
G`0404IFH`17EF00AeEP04aIHP80CUQQ00@0BeAK04mFG01?EU`0D5EI0P16BTd0100^=C/08BHZ024V
:P0F6Qd201HG6@0401XI5`0L61D071PE01`G4`807QLA00@06aXE01XI500J6A@06APC0`0H5a8201TH
4`0401XI500K6QD06aXE01/J5P807A`G00@0:RLP044j<011>S00BT8e0P1=A3<01010=B<0DDHb0556
<P1;=ah204di7P04054l8@1G?B80EcdR05a19P80I4HZ00@0KDhb07=D>01cE3P0Iddb0P1SBBh0101P
ABX0HDH[0656:`1WBBl206]=<`0406i>=@1]CC@0KDdd06M8=080K4di00@0IdPd05Pi9@1H>BD0Cc0L
0P1E=R80101M?RX0H44]0611;@1M?B@205lo9P0405hn9@1P@2L0H40W06Y:<@80Kdlf00@0L50g079B
>@1bDST0MT`d0P1kDCT01021Ecl0Qee508MMA@2@ITh209Q^EP0309U_E`2EJe<0UF]C00<0XW=G00@0
XgAH0:n0I02_P6@0YgQL0P2bPfL0102cQ6P0/X=W0;:3I`2cQ6P30;F6JP80/X=W00@0[GiR0:MhG02W
N5`0YGIJ0P2TMET0102]OV80/H9V0;62IP2dQFT20:n0I0040:AeF@2JJdl0VV]?09IWB`80UFI:00@0
TfA809i_D`2NKe<0Q5Di0P27FCl0102;GdH0Md/b07M;<P1jD3P2085G?`0406Q09`1W?bH0IclV06U1
:080JdD^00@0I40X05Hb701F<Q`0CBXD0P1=:QH0101A;Q80C2T=04`Y3@1;:0`2054]4`0405Xf7P1V
@RX0IT8Z07QD?P80Qf==00@0[XYd0:F1J`2UPF/0W7QP0P2MNF40102WPfT0ZhQ/0:^8K02MNUh208]X
C00409miHP2CJU@0TfYD09QYE@80VVMD00@0WV]H09m/F@2OK5T0SEi:0P2=H4/0102;HDT0R5i608QN
AP21Ech207Q>=@040717:`1_ARP0KdHX07A;;@80NE0b0`1eC2`0101eBRP0MTXW07I:9`1iCBX208IH
=00409IXA02RM500XWA@099R>P80U6<k00@0VFQ00:5]AP2QKDH0XFe60`2MId420:9/AP0409Y[C@2J
JDT0VVU909]VA0<0UV0l0P2LJTT0102=GT00Kd8U06m29@0j7PH2034K1P0402`L3@0R6Pl08QX?02`S
6P80:alC00@0?2/I04dg8P1==b80C2dI0P1D<b00101:9A80BR4?04XQ3`1_@Rl209ISDP040:E]G02X
L5l0Z71O0:AVC`80W5i700@0Y6A;09UG?@2IEcd0Te0e0P2UH440102bJdd0^W5A0;YaD@2lMU820:YR
?0040:IN=P2]JdP0[F]80:1ZD080ME=700<0EDM704i<E`1>C5L00`1:Be001019C5<0B4aE04Q<E@19
CEP204I<F00304I<FP15BeT0AD]I00H0ATeG00@0B4mI04YAF`1:DE/0C5=M1P1=E5h204aCG@0404]B
G019D5X0BE1J04M>F080ATeG0`15BeL01014BUH0@dUE04=9E@12B5@30457D`80@4IB00<0@dUE0498
E012B5@01012B5@204=9E@<0AD]G1012B5@504=9E@<0A4YF1016C5P204E;E`0404=9E@11Ae<0@DMC
03m5D@80?TA@00@0@DE@0496D@12AU40@TIA0P13Ae800`14B5<0ADUD04E9E00304I:E@<0ADUD1016
BUD504M;EQP0BDeH0`19CU@01018C5D0ATeE04I=E@15C5H204A=EP0404=;EP12C5H0@TaF03m=F080
@TaE0@13C5<104I;C`40oooo000H04MDG@`0BU9M301;Deh404QEGP<0BEIO1@1:Ef0<04UFG`0404IF
H`16E5l0AUAO04]HH@80CUQQ00@0BeAK04mFG01?EU`0CeAH0P13AdX0100W;300720S01`P8`0E5QP2
01LG5`0401`K6@0M6AH07ATF01dH50807aPB0`0J6A@0100I61<061LB01PG4P0H5a8201TH4`<06QTD
00@081lJ01/J5@0K6QD0;2TP0P0i<RP0101:@SD0@CPW044h9`15>RP205I;=`0404Hb6@19=AX0BCDJ
04lj7`80EcdR00@0G44V06A6:P1TARX0KDhb0P1bDcL00`1^DcP0K54f06aA=P0306]=<`0406i>=@1c
DcX0Le<j07]I@080Oee400@0NEU207IF?`1fESl0LE4j0P1ZBS<0101VARl0IdL`06M7<01[Bc@206i>
=`0407YH?@1]Bc00KD/`06e;<080NELl00@0O5Xo07UG?01iEc`0O5Xo0P20GT<308mSBP0409AXC`2L
L5L0W71G0:9fG@80XGEL0`2PM5/30:AeF@040:UjGP2dQFT0]8EY0:ioH`80]hQ/00@0^HY^0;N8K02g
R6`0/H9V0P2cQ6P0102eQVX0]8EY0;B5J@2_P6@20:UjGP040:EfFP2TMET0Y7EI0;>4J080/85U00@0
[GiR0:YkG`2ZNel0YgQL0P2QLUH0102JJdl0UFI:09EVBP2?H4@20:=dF00308]L@020DSP0P58h00<0
R5a300@0PeU10757;`1aAbl0JT8Y0P1N=Qd0101L=1/0IClX06Do:01XA2`205lk9@0405Hc7@1G=200
Ec@P058d7080BR`D00@0ARP@04X[5P1::aH0Cc0K0`1?<1`2054b800405Df901^Cc/0KTlk08AUD@80
PV=>00@0MEI1071B>P1`DSX0MUQ00P1nH4P0102<J500QEa608ELAP2>HD`208iNBP0408iKB02;FdL0
Re]7085B?P80QEQ300@0Qe]208AH?`24F3l0OU8i0P1hC340101dB2d0LdLZ07=7:P1gBbh207Y?;`04
06i38@1`AB<0L4DS07979@80MDTV00@0NddY08=E<@23EC40Re/e0P2>GSP0102HIcl0VFQ009UX@02N
KDD20:A`B@040:=]A`2NJ480WVQ209iV@@80X6Q300@0VFQ809UWA02IId@0VVI10P2FH3X0102DH3/0
UVA309IT@`25F3T206Hj7`0403LM100[5P40:aH102<E20808QT@00@0<2TO02LM4@0W7A40=B@D0P1<
=b8010139Q@0B2TE04PY5@128@h204DO3004068g9P29FTP0REY80:=`G`80ZgIT00@0[71H0:AVC`2T
ITl0XF=:0P2GEC/0102DDCH0Y5m00:AO@02^IdT20;5ZBP040;I`CP2TGSX0Y5hj0:9N>`80ZFa=00@0
UfMA06]=@`1[CD<0Cd=50P1?CEP304Y=E00404U<E@17Be@0Ad]D04E<EP80ATeG0`15CEP404I=E`80
AdiH00@0B4mI04YAF`1:DE/0C5=M1P1=E5h204aCG@0404]BG019D5X0BE1J04M>F080ATeG0`16C5P3
04E;E`0304A:EP13BED0@dUE00<0@TQD00<0@dUE0498E012B5@01012B5@204=9E@<0AD]G1013BED3
04A:EP/0AD]G00@0@dUE0498E012B5@0@4IB0P0nA5030496D@<0@dMB00<0A4QC04E9E015BE@00`16
BUD304E9E0@0ATYE1@17BeHH04U=F0<0BDiD00@0B4aE04I=E@16CED0ADaF0P14CEH01013BeH0@TaF
049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`006017E5d<04YBG@`0Be=N1018EEh304UFG`D0BUMP
3019EUl01016EV<0AUAO04IDG`1;F64204iHH@0404]DF`1?EU`0CeIL04mDF080@dM:00@09b``01`P
8`0L82<05AHH0P0G5aL0100L6aT07ATF01dI5P0M61@201lH4P<06QTD00@06APC01PG4P0H5a8061LB
0P0I61<301XI5004020O6P0K6QD06aXE02`Y8080>C8X00@0BT8e044h9`11>2L0ACXX0P1FBcL01016
<QT0BCDJ04Te6P1?>Ql205Lm8P0405a19P1TARX0I4HZ06e><P80LU<g00<0KU<h06aA=P1/DCH00`1[
CC<0101^CSD0Le<j07=C>P1kFD0207mMA00407UI@P1fESl0MUHo075A>P80JTXc00@0ITH_06M7<01W
Ac00Jd/d0P1^CSL0101jF3d0KD/`06e;<01]Bc0207UG?00407aJ?`1iEc`0NELl07aJ?`80P5i30`2?
HdX0102DJ4l0W71G09a`E`2RMUd20:5eG0<0X7AK0`2TMET0102YNUh0]8EY0;B5J@2^Of<20;N8K004
0;V:KP2gR6`0]hQ/0;62IP80/hAX00@0]HIZ0;B5J@2dQFT0[h1T0P2YNUh0102UMUX0Y7EI0:AeF@2c
Q6P20;21I@040:enHP2ZNel0ZW]O0:MhG080XG9F00@0VV]?09EVBP2EITX0Sf140P2SM5P00`2;G400
P58h081B>00308QL@`0408=I@@1aAbl0LDL_06Y2:@80GSHM00@0G3@K06Do:01U?bP0J4@/0P1O>bD0
101F<ad0Ec@P05Ld801B=1`204X/500404HX401::aH0BR/F04l`6`<0Cc0L0P1A<R00101E=R@0KTlk
06i?>`24IE42089SCP0407EF@@1`DSX0L58j07IH@080OV1800@0S6Q@08ELAP25G4H0SV5<0P2>GTX0
102>FdP0Re]708]KA`21DSh208EH@`0408MK@P24F3l0Q5Po07iB>@80N4`a00@0M4P]07=7:P1cAbX0
Md/^0P1jCbl0101^@b40L4DS07158`1bAbD207E99P0407]=:@23EC40PeDa08]K=@80SUhh00@0V6Lo
09UX@02IJ400WVe50P2TL4T0102SKDL0WVQ209iX@P2NIT420:1X@`0409UXB02IId@0VFM409YV@@80
UV0j00@0U60k09IT@`2FI4<0QEPi0P1V>Ql0100g7@@0:aH102/F0@0S5@P2028I4004030Y7`0W7A40
9adA03DT5080C3LR00@0@bHD04PY5@18:AD0@R4>0P157``0101R=bH0REY808UJB02SL5l20:]fI004
0:a`F02TITl0Y6I?0:5SBP80UeDk00@0U54f0:AO@02TGd00[VM90P2aJTX0102fL4h0Y5hj0:AN>P2R
GS/20:U/C@0409MWD@1[CD<0Jde304m3A@80CdeH0`1:CE@01019C5D0Ad]D04M;E015C5H204I=E`<0
ADeH1016CEL204M>F00404Q?F@1:DE/0BU5K04aCG@H0CEAN0P1<Ded0101;DU`0BE1J04U@FP17CUP2
04I=E`<0ATaH0`15BeL00`14BUH0@dUE04=9E@030498E00304=9E@12B5@0@TQD00@0@TQD0P13BED3
04E;E`@0@dUE0`14BUH;04E;E`0404=9E@12B5@0@TQD0416DP80?TA@0`12AU4304=7DP0304A8D`15
BE@0ADUD00<0ATYE0`15BE@404I:E@D0Ad]F6019CEP304U>E00404Q<E@16CED0ATeE04E<EP80A4eF
00@0@d]F049<EP12C5H0?deH0P12C5D104=<D`40AT]?0@3oool000`0B5EN3017E5d<04YBG@`0C5AO
1018EEh304UFG`D0BUMP3019EUl01016EV<0AUAO04IDG`1;F64204iHH@0404aEG01?EU`0CeIL04aA
E@80?T9500@08b@V01XK7@0J6ad05ADE0P0I61H0100L6aT07ATF01dI5P0K5Q8201`G4`0301XI500I
61<06APC00L06APC0P0J6A@0100K6QD05QH>01HF3P0^:b8203H`900404U2<P0f;Q/0=RhK04Dj9P80
D4D_00@0BSHM04hj7`1>>Ql0DchS0P1J@2D0101M@RL0HdDY06=5:@1ZBbl206e><P0406]=<@1_DCD0
Ke4e07=D>080MEHj00@0MU@i07MC>@1gDcT0NEDk0P1lF3h0101jESh0Me<k07MC>`1iECd2089NAP04
08EQB@21GDD0PEe507iJ@P80Oe]300@0RFA907]F>`1kES/0P5]00P2=J4d01028HdP0Q5m408AOA02<
Id`2091[D0040:1dF`2MLEP0WG5H0:5eG080ZGeT00@0ZgmV0:AhG`2TN5l0XWIM0P2UNF00102XNEd0
YgQL0:MhG02]OV820;F6JP040;:3I`2hRFd0^8U]0;Z;K`80]XM[00@0[WmS0;21I@2`PFD0/H9V0P2_
P6@00`2ZNel0YWMK0:IgF`030:AeF@040;21I@2_P6@0[h1T0:n0I080/85U00@0[WmS0:UjGP2YNUh0
Wg1D0P2HJDd0102AHTH0Z7UM0:QiG@2@HDD207U;<@0408IJ@@2IKE@0VFeD09=YD@80LTP`00@0KdL^
05de701M=A`0E2`C0P1G<AX0101E<AT0BBD?04TU3`127`T204HS3`03054f8@18;AP0B2dH00<0@2DB
00@0@2DD03dR4@0m8Q40>alA0P0l8180100h70h0@bLI04<W6@1?=2<205Hk:P0405e2;`1TBCH0I4Tf
06E:=@80HTLb00@0Ec<I05@_5@1D;aD0KD0Y0`1gAc4207e==`0407a?>022F400PUQ006e1:080Kd<Z
00@0LDDZ0796:`1bAR/0LDDX0P1`ABD306i38`0406m48P1eBRP0MDXX07eA;P80QUPd00@0T68n09YZ
A02JJT@0VVY40P2FIT00102IJ3l0WVY209iZ@P2RKTH20:A_A`030:5/A02NId00WVM000<0Vf@m00@0
UVA109MS?P2GHch0VF<m0P2FHCT0102CGcP0TV50099Q@01oE3@2064h700403PN1`0X4`00:1<001h@
0`808AX@00@0<b`T02DK4@0U6a40;Qh>0P19=2401011:1@0@bPE04<X5@18:AD204l/6@04068l:@27
FDT0QeU90:9cH@80[7]Z00@0[g=K0:Q/E02XK5@0X6990P2HF3d0102IF3`0YF=30:ES@`2^JDX20;1Y
B@040:]WA02KFCL0VeTg09aM?P80Y6iD00@0SFIE06=:A@1SBTD0C4570P1?CEP0101<CeP0BdiG04]>
E`18C5D204E<E00404A;E@14CEH0A4eF04I?F080Ae1I00@0AdiH04I=E`16CEL0ATeG0P17CUP01019
D5X0Be9L04]BG01=E5h204iEG`@0CEAN0P1<Ded0101;DU`0BE1J04U@FP17CUP204I=E`0404Q>FP17
CET0AdeI04M=F@80ATaH00<0AD]G04A:EP14BUH01013BED60498E080@dUE0`15BeL404A:EP<0AD]G
1@16C5P604E;E`0404A:EP12B5@0@TQD0416DP80?dEA0`12AU4304=7DP0304A8D`15BE@0ADUD00H0
ADUD1016BUD504M;EQP0BDeH0`19CU@01018C5D0ATeE04I=E@15C5H204A=EP0404=;EP12C5H0@TaF
03m=F080@TaE0@13C5<104I;C`40oooo000<04QEGP`0AeAM301:DUd<04aDG`@0B5EN0`19EUl504YG
H0`0BEIO00@0AeMT04IDG`16E5l0BeQQ0P1>F640101<EE`0CUEK04iEF`19CU8203Tm@00402<R800K
6QP06aXH01LF50806QTG00@07QXG01/G500K5a@06A@@0P0J5A4401TH4`<06QTD1@0I61<0100F5Ph0
61P>01PH3P0e<bL203dg:@0404Y3<`0e;AX0=BdJ04Hk9@80BChX00@0Cc/R058n8`1B?R<0Ed8W0P1M
@bP0101PABX0IDL[06E7:`1[C30206m@=00406M8:`1`CS80L4hb07ME>@80O5Tm00@0NePl07]F>`1k
ES/0O5Ll0P1mF3d01024G480O5@j07aD>P1mEC/208QPAP04095YC`2@J4h0T6Q>095YC`80UViD00@0
SVE908aSA`2<HdL0V6mC0P2OMUX0102IL5@0WWEI09ieF@2[PVH20:n6JP040:QjH02TMU`0Y7IL0:Mi
G`80[GmU00@0[7iT0:EgG@2UMed0XgEK0P2WNEl0102^Of<0ZW]O0:YkG`2aPVH20;B5J@040;:3I`2c
Q6P0/hAX0;R9K@D0/X=W00@0/85U0:amH@2/OF40YgQL0P2UMUX0102VMe/0Z7UM0:QiG@2]OV820;21
I@040;>4J02dQFT0]8EY0;62IP80ZgaP00@0YgQL0:AeF@2TMET0X75E0P2TMET0102CI4P0SV1608iP
AP2NLUT20:MkHP040:1fGP2;HDT0Rf5907]C>P80I3`S00@0Db/B04lY4P1?:A80BbL?0P0o6`D303DB
000403HM2@0k8Ph0>b8>04T_7P80FCl^00@0GT@e05Pn<@1H?S40D3HY0P1>=2L0101B>2/0AbdP04L]
800n91L204L]7P040698=`1lHU40O69A089YE@80Nf9>00@0JDL[068m8P1R?B80MD/b0P1kCCD0101e
ABh0N4Xb07Q:<P1hCSD207mG?@0407M9;`1jC380NT`b07i@=P80Oe8e00<0OU4b07a?<01lCc000`1l
Cbh0101iCBX0PUHc089F<`2=Gc/209IXA0040:9bC02YNE<0ZGUC0:EdC080VVU100@0VfLo09iZ@P2N
JT80XFa40P2PJd<0102MISl0Vf@m09]T?@2KHC/209YP>P0309AP>@2DGcL0U5lg00<0V64j00@0Telh
08mQ?`2?HCl0OE@f0P1R>R00100m90h09A<002DC000L40@2028K4`04034/9P0R6Pl08QX?02dM3P80
@c0O00@0>RDB03`U4`0l9A<0@RLD0P1;;1P0101M>2H0Ne9007]B@02JK5`20:UjJP040:abFP2[LET0
Zg5I0:9VC080Vei200@0Wei20:ITA02VI4@0[FQ90P2]J4L0102SGSd0UeHf09MF=P2FGD8209Q[F004
08=RF@1LBTX0G4Y:04U2BP80CTiJ00@0BdmJ04Y>F@1:CUT0ATeG0P15C5H01014CE@0ADiE04E>E@16
DEL204MBF0@0AdiH0P18CeT01019D5X0Be9L04]BG01=E5h204iEG`@0CEAN0P1<Ded0101;DU`0BE1J
04U@FP17CUP204I=E`<0B4iJ00@0AdeI04I<F016C5P0AD]G0`14BUH304=9E@H0@TQD0P13BED304E;
E`0304A:EP15BeL0AD]G00<0AD]G0`16C5P304M=F@<0A4YF0`15BeL01014BUH0@TQD0498E011Ae<2
0416DP@0@dMB1P14B5<504E9E0@0ATYE1@17BeHH04U=F0<0BDiD00@0B4aE04I=E@16CED0ADaF0P14
CEH01013BeH0@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`003018EEh<04MDG@`0BU9M301<
E5l404QEGP<0BEIO1@1:Ef0<04UFG`0404MGI016E5l0AUAO04]HH@80CUQQ00@0C5EL04iEF`1>EE/0
BDiB0P0i?D00100S8R006aXH01/J600G5Q@201XI5`0401hJ5`0K5a@06aLD01TD40806QDA100I61<3
01XI50D06APC00@05QH>01PH3P0H60h0=C<W0P0m=bT0101:@c<0=BdJ03D]6P16>bD204Tn:00404lk
8P1B?R<0DShS05M29`80GD<X00@0H4DZ06E7:`1UAb/0Jd``0P1_D3@0101WB2/0L4hb071><P1gECT2
07aI?@0407]H?01kES/0NeHk07aG?080OEPm00@0Q5a207aD>P1lE3X0OEDk0P28H4H0102AJDl0T6Q>
091XCP2AJDl209I^E00408iUB@2<HdL0S6=709Q_D`80WgIJ00@0VG1D09ieF@2NMET0Zh9V0P2_QVX0
102XNV00Y7IL0:AfG02WNEl20:eoI@040:anI02UMed0YGMM0:=eF`80YgUO00@0[WmS0:YkG`2ZNel0
/H9V0P2dQFT0102bPfL0/hAX0;>4J02hRFd50;:3I`040;21I@2/OF40[7eQ0:MhG080YGIJ00@0YWMK
0:QiG@2XNEd0[GiR0P2`PFD0102cQ6P0]8EY0;B5J@2aPVH20:]lH0040:MhG02TMET0Y7EI0:1aE@80
Y7EI00@0TfA808iPAP2>H4H0WW9I0P2WNf80102PMUh0Rf5908]QB@1kDcX206@l8`0405<[4P1?:A80
CbTB04/W3`80?a/50`0e4P00100f7@T0>b8>03/R3P19;ah205To;P0405i4=@1H?S40F3ha050f:@80
CS@W00@0DSP[04L]8017;B00?R@G0P17;Ah0101RB3L0O69A07aRD@22JED207]RCP0406U7:`1R?B80
HSdR07E;<P80Ndde00@0MDD^07Q:<P1hBS80N4he0P1oEcd0101gBBl0NT`b07Y<<P1nD3H207mB=@03
07iA<P1lCc00O4l`00<0O4l^00@0NDdZ089F<`22ES<0SElk0P2FJ4@0102RLT`0ZGUC0:UiD`2UM4`2
09YY@@0409]W?`2NJT80WVY20:5/A080X6]300@0WFHo09]T?@2KI3d0Vf4k0P2JH3X00`2DH3T0U5lg
09AO=`0309QQ>P0409=O>02?HCl0Sf4o07eD=P80HSXP00@0?B@>02DC000U4`0071040P0R6a<0100a
;2H08QX?028J3`0]7@h204<`7`0403XU4P0l9A<0?2DC048W5080Bb`H00@0GCPV07]B@01kDT00VVaL
0P2YNVX0102/LUX0Zg5I0:]aF@2RIT`209]N@P0409mN@P2VI4@0YVA40:eXB@80[FQ700@0Xehm09MF
=P2GESH0UUe20P2HJeP01023HUT0G4Y:05a:BP19@TX204i>FP0404]?FP1:CUT0BTiI04I=E`80ADaF
00@0A4eD04E>E@15CUD0AU5G0P17DUP404M>F080B4mI00@0BE1J04]BG01;DU`0CEAN0P1>EEl404eD
GP80C5=M00@0Be9L04U@FP19D5X0AdiH0P16CEL304Q>FP0404M=F@16C5P0ATaH04E;E`<0A4YF0`13
BED60498E080@dUE0`15BeL00`14BUH0AD]G04E;E`0304E;E`<0ATaH0`17CET304A:EP<0AD]G00@0
A4YF0498E012B5@0@DMC0P10AU8404=7DPH0A4QC1@15BE@404I:E@D0Ad]F6019CEP304U>E00404Q<
E@16CED0ATeE04E<EP80A4eF00@0@d]F049<EP12C5H0?deH0P12C5D104=<D`40AT]?0@3oool000`0
B5EN3017E5d<04YBG@`0Be=N1018EEh304UFG`D0BUMP3019EUl01019FFH0B5IQ04QFH@1<FF8204mI
HP0404aEG01<DeT0C5=I04E:CP80<cLj00@0920M01dI5P0M6AH06QHC0P0L61D0100M6AH06QHC01XF
4`0H514201TE4P0301PG4P0I61<06APC00H06QTD00@06APC01PG4P0H5a805aL?0P0T91X0100m>bl0
Ad4c04M1<`16?bd203/c7P0404Xo:@19?bH0BClV04hj8@80DShS00@0Ed8W05e3:01M@bP0HTL/0P1Y
Bbl0101aDSH0MULk07IG>`1]Bbl2079?<`0407MD>01lEc`0O5Ll07iI?P80Pe]100@0R61608iTB`2>
I4/0SFA60P25G3h01024Fcd0S6=508aSA@2DJdd209I]C`0409M^D02KLU@0Vg9D09IYC080WW5D00@0
YWUL0:EhF`2UN5/0XGAG0P2YO5l0102cQVT0]HQ[0;F8J`2YNV020:QiG`040:UjH02ZNf40ZW]Q0:Uj
H080YWMM00@0YgQN0:YkH@2ZNf40/hAX0P2]OV80102cQ6P0/H9V0;62IP2aPVH20:n0I0040;J7J`2`
PFD0/85U0;V:KP80^8U]00@0]8EY0:ioH`2^Of<0ZGYN0P2XNEd0102ZNel0[7eQ0:amH@2/OF420;21
I@040;>4J02aPVH0/H9V0:ioH`80[7eQ00@0[WmS0;:3I`2bPfL0/hAX0P2NKe<0102EITX0ZgeS0:]m
H`2dR6l20:5eG00409AZDP2FK5@0UVaD095YD080OEDl00@0Jd<Z06Q2:`1X@R/0KTXb0P1[Ac40101N
>bD0E34M05@a7@19<1`2044X500403/Q400n91<0?R@C04@Z6`80C38S00@0F3ha06A:?@1TBSd0FT0c
0P1NA3L0101J@340CC<T04dc9016;1/204hd8`0405e4<01WCSX0Idhj07YH?080JDHZ00@0MT`c07I:
<@1fBS40K3hV0P1bARd0101eBc80NE4g07UA=`2>GdD208mPA00408iO@`2:Fcd0RU/m08AE=`80PeDd
00@0QeTh08eO?P2=Gch0Ndl/0P26F3@0102?HCd0U6@n09AT?P2JJT@209m^AP0409MV?P2=FC40SETa
09]V?0<0WfY00P2LIC`0102IHST0Vf4i09]Q>@2KHCT209]P>00409EP>02DGC@0U5dd09QR=P80VF8i
00@0U60i08mQ?`2?HCl0OeHh0P1T@BD01016;AL09a@302LD0`0M4`L2028M5`0402`Y8P0R6A008QT@
02dO4P80>bXH00@0>RTE03/Y5@0k:AD0>R<A0P109A80101B<Ah0JTDc06Y5<`27GT`209icHP040:if
G@2hOVH0^7iV0:ibF080Y6M;00@0Wem309YI>@2JFCT0W5Xj0P2LEcH0102QGcd0V5/n09QK?P2AGdP2
08QRE@0407=JEP1FBTh0ETY>04U6C`80CDeI00@0BDeI04I<F016C5P0ATeG0P15CUL01016CeH0AU5G
04IAE`17DUH204IDE`@0AdiH0P18CeT0101:DE/0C5=M04aCG@1=E5h204mFH0@0CEAN0P1<Ded0101;
DU`0BE1J04U@FP17CUP204I=E`0404Q>FP17CET0AdeI04M=F@80ATaH00<0AD]G04A:EP14BUH01013
BED60498E080@dUE0`15BeL304A:EP<0AD]G1016C5P204M=F@0304=9E@14BUH0A4YF00@0A4YF0P13
BED01011Ae<0@4IB0416DP13Ae8;04A8D`<0ADUD1016BUD504M;EQP0BDeH0`19CU@01018C5D0ATeE
04I=E@15C5H204A=EP0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000<04UFG``0
AeAM601:DUd404QEGP<0BEIO1@1:Ef0<04UFG`0404]KJ019Ef80BEMR04eJH`80CeUR00@0C5EL04]B
F01;DUP0@DI:0P0]<C@0100V7aT07APD01dH500J5A4201dH500401hI5@0J5Q<06QHC01XF4`8071PE
00@061LB01TH4`0I61<06QTD0`0K6QD201XI500401TH4`0H5a8061LB01XJ4080<C4U00@0?C/^04U3
<`19@c<0=c0N0P0n=R40101=@bX0BT0W04Y09`1>>R42054m8P0405E09@1L@RL0G48W0656:`80JT``
00@0M5Di07YK?`1jFcl0ME8d0P1gDSD0101iE3L0OULj07iG>P22Fch208QO@`0408iUB@2BJDd0TVU=
09=VA`<0T6=40P2EJ4T0102JKDh0Vfi?09]^C`2JKDh209U/C@0409i_D@2SM5H0XgAF0:5bE080WVmA
00<0XgAF0:QiF`2XNE/00`2XNE/0102VMe/0ZGYN0:UjGP2ZNel30:IgF`80ZGYN00@0[GiR0:n0I02_
P6@0]hQ/0P2^Of<0102eQVX0[h1T0:n0I02aPVH20:enHP040;N8K02aPVH0/H9V0;V:KP80^8U]00@0
]XM[0;21I@2`PFD0ZgaP0`2XNEd20:YkG`040:MhG02ZNel0ZW]O0:ioH`D0/85U00@0/H9V0;>4J02c
Q6P0]hQ/0P2QLUH0102NKe<0/XAZ0;:4JP2fRW420:=gGP0409M]E@2LLUX0W79J095YD080Qem600@0
O5@k07I@>@1fD3T0N5@l0P1fDS`0101U@R`0E34M05@a7@1I?RT204`a700304<X5@14:AH0A2TF00<0
Ab`K00@0ARXL04L[7@17:ad0BBdO0P1=<B<0101<<B00AR/J04H[6P14:AH204/`7@0405<h8`1E>RD0
ECXU06m@<@80H3dO00@0Ld`_07mC>01oDcP0P58h0P2=HDH0102AJTd0T6]>091[CP21DSH308EF>P80
ODh`00@0M4HU07948`1bA2<0ND/Y0P21Dc40101nD2`0Redi08]M>@2EICl209MW@@0409a[@`2RLDT0
XW590:1/A080U60h00@0VfHl09mZ@02OJT00WfPo0P2IHST00`2HGSH0W68j09aR>P0309mT?00409US
=`2FH3@0UV0d09aS=P80VfDi00@0UV8j08iR?`2>HSl0OePi0P1XABT0101==b80:AP602TH1P0R60h2
02@O6@0402HS7P0Q6Q008AX@034S5P80=B@D00@0@34N048a7`12<Al0>bHC0P0k8Ph0101;;1X0ES<P
05Hc801V?bh2085GA`040:9ZD@2eOF@0]GeT0:mfF`80Yf]?00@0X61409=B<P2CDS80U58b0P2EDc40
102[JDL0XFE90:5UB@2>Gdd207MFC@04061>CP1AAdl0DDM?04Y8D`80CDaJ00@0A4YH04E;F@15BeT0
ADeH0P17D5T00`17DUP0B5=G04QCE`0304IDE@H0B4mI00@0BU5K04aCG@1<Ded0CUEO0P1?EV0404eD
GP80C5=M00@0Be9L04U@FP19D5X0AdiH0P16CEL304M=F@0404I<F015BeL0AD]G04A:EP<0@dUE0P12
B5@00`13BED0@TQD0498E0040498E080@dUE0`15BeL404A:EP<0AD]G1@16C5P304=9E@@0A4YF0P13
BED00`12B5@0@DMC0457D`0<04A8D`<0ADUD1016BUD504M;EQP0BDeH0`19CU@01018C5D0ATeE04I=
E@15C5H204A=EP0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000<04UFG``0AeAM
601:DUd404QEGP<0BEIO1@1:Ef0<04UFG`0404]KJ019Ef80BEMR04eJH`80CeUR00@0C5EL04]BF01;
DUP0@DI:0P0]<C@0100V7aT07APD01dH500J5A4201dH500401hI5@0J5Q<06QHC01XF4`8071PE00@0
61LB01TH4`0I61<06QTD0`0K6QD201XI500401TH4`0H5a8061LB01XJ4080<C4U00@0?C/^04U3<`19
@c<0=c0N0P0n=R40101=@bX0BT0W04Y09`1>>R42054m8P0405E09@1L@RL0G48W0656:`80JT``00@0
M5Di07YK?`1jFcl0ME8d0P1gDSD0101iE3L0OULj07iG>P22Fch208QO@`0408iUB@2BJDd0TVU=09=V
A`<0T6=40P2EJ4T0102JKDh0Vfi?09]^C`2JKDh209U/C@0409i_D@2SM5H0XgAF0:5bE080WVmA00<0
XgAF0:QiF`2XNE/00`2XNE/0102VMe/0ZGYN0:UjGP2ZNel30:IgF`80ZGYN00@0[GiR0:n0I02_P6@0
]hQ/0P2^Of<0102eQVX0[h1T0:n0I02aPVH20:enHP040;N8K02aPVH0/H9V0;V:KP80^8U]00@0]XM[
0;21I@2`PFD0ZgaP0`2XNEd20:YkG`040:MhG02ZNel0ZW]O0:ioH`D0/85U00@0/H9V0;>4J02cQ6P0
]hQ/0P2QLUH0102NKe<0/XAZ0;:4JP2fRW420:=gGP0409M]E@2LLUX0W79J095YD080Qem600@0O5@k
07I@>@1fD3T0N5@l0P1fDS`0101U@R`0E34M05@a7@1I?RT204`a700304<X5@14:AH0A2TF00<0Ab`K
00@0ARXL04L[7@17:ad0BBdO0P1=<B<0101<<B00AR/J04H[6P14:AH204/`7@0405<h8`1E>RD0ECXU
06m@<@80H3dO00@0Ld`_07mC>01oDcP0P58h0P2=HDH0102AJTd0T6]>091[CP21DSH308EF>P80ODh`
00@0M4HU07948`1bA2<0ND/Y0P21Dc40101nD2`0Redi08]M>@2EICl209MW@@0409a[@`2RLDT0XW59
0:1/A080U60h00@0VfHl09mZ@02OJT00WfPo0P2IHST00`2HGSH0W68j09aR>P0309mT?00409US=`2F
H3@0UV0d09aS=P80VfDi00@0UV8j08iR?`2>HSl0OePi0P1XABT0101==b80:AP602TH1P0R60h202@O
6@0402HS7P0Q6Q008AX@034S5P80=B@D00@0@34N048a7`12<Al0>bHC0P0k8Ph0101;;1X0ES<P05Hc
801V?bh2085GA`040:9ZD@2eOF@0]GeT0:mfF`80Yf]?00@0X61409=B<P2CDS80U58b0P2EDc40102[
JDL0XFE90:5UB@2>Gdd207MFC@04061>CP1AAdl0DDM?04Y8D`80CDaJ00@0A4YH04E;F@15BeT0ADeH
0P17D5T00`17DUP0B5=G04QCE`0304IDE@H0B4mI00@0BU5K04aCG@1<Ded0CUEO0P1?EV0404eDGP80
C5=M00@0Be9L04U@FP19D5X0AdiH0P16CEL304M=F@0404I<F015BeL0AD]G04A:EP<0@dUE0P12B5@0
0`13BED0@TQD0498E0040498E080@dUE0`15BeL404A:EP<0AD]G1@16C5P304=9E@@0A4YF0P13BED0
0`12B5@0@DMC0457D`0<04A8D`<0ADUD1016BUD504M;EQP0BDeH0`19CU@01018C5D0ATeE04I=E@15
C5H204A=EP0404=;EP12C5H0@TaF03m=F080@TaE0@13C5<104I;C`40oooo000<04UFG``0AeAM601:
DUd404QEGP<0BEIO1@1:Ef0<04UFG`0404aLJ@1:F6<0BUQS04eJH`80D5YS00@0C5EL04YAE`1:DEL0
?dA80P0Z;S40100T7AL07QLA01hG4@0K50h201`G4`0401hI5@0K5a@06aLD01`H5@8081`I00@061LB
01TH4`0I61<06QTD0`0K6QD201XI500401TH4`0H5a8061LB01/K4@80=SHZ00@0>3HY04@n;P14?Rh0
9b0=0P0l=1l0101<@RT0Bd4X04]1:01@?2<2058n8`0405E09@1J@2D0FT0U05m4:@80J4X^00@0LU<g
07UJ?P1iFSh0NE@g0P1jECP0101lEcX0PEXm085J?@25G40208QO@`0408]OA02<H4D0S61509MZB@80
UFQ70`2CITD0102FJDP0W6m>09a_CP2NLE0209e`C`040:5`D02RLE40XW5A09YYB@80W6];00@0ZgYJ
0:mnGP2_OUh0ZGQH0P2ZNET0102UMUX0ZgaP0:]lH02[O6020:EfFP030:AeF@2]OV80[GiR00<0/hAX
00@0^HY^0:n0I02_P6@0]XM[0P2^Of<0102aPVH0[WmS0:ioH`2jRfl30;B5J@80]HIZ00@0]8EY0;21
I@2`PFD0ZW]O0P2UMUX30:=dF00409m`E02SM5P0XgAH0:YkG`80/X=W00@0]XM[0;F6JP2eQVX0/85U
0P2[O600102]OV80ZW]O0:YkG`2YNUh20:]mH`040;65K02iSG@0^Hed0;N=M@80/HM_00@0YWiU0:^3
JP2[PfX0ZX9Y0P2ONF80102GLe/0T6aF091/EP20GDL206m<>00406Q<=P1H?2H0F3`V04X^6@80C30K
00@0E3LU05@g9`1D=bL0CC0P0P18:a/01019;1`0Cc8R04lb8P1B=BD205@g9@0405a0:`1]DC`0KE4l
07aPBP80PfMA00@0Q6E607=@<P1cD380QUm20P2BIT/0102<H4D0T6A9091TB@26Gd8207UD=`0407I7
:`1nCc<0OTlc08AE=`80P54c00@0MTPW07538P1a@b80MdTW0P1oDBl0102:G3P0UVQ409IXA02NKTP2
09YZA00409]Z@P2QKDD0XFe509aX@080TEde00@0W6Lm0:1[@@2PJd40WfPo0P2IGcL0102GGCD0WV<k
09iS>`2SJ4020:9W?`0409eW>`2KHSD0Vf8e09iU>080WFLk00@0UV8j08iR?`2>HSl0OUXj0P1YAb`0
101C?BP0:aX:02/J2P0U6a4202DR6`04028O6P0Q6Q808AXB03@X6P80<B0@00@0?S4N04<d8@13=240
>BLC0P0f7`d010139Q@0?ah=03lN3@127@/205Tb8@0408A<<`2QJE00XFU@0:E/D@80YFU=00@0XFA7
09QG=`2HEcL0WUdm0P2UHd40102dLU00YfiC0:M^D`2=Gdl206]>B00405A4A`1<ADd0C4E=04a:E`80
C4]I00@0@TQF04=9E`13BEL0ADeH0P18DEX00`19E5X0AeEH04MEF00304MEEP@0B4mI0P19D5X0101:
DE/0C5=M04aCG@1>EEl204mFH0@0CEAN0P1<Ded0101;DU`0BE1J04U@FP17CUP204I=E`0404M=F@16
C5P0ATaH04I<F080AD]G00<0A4YF04=9E@13BED00`12B5@00`13BED0@TQD0498E0040498E080@dUE
0`15BeL304=9E@<0A4YF1015BeL204I<F0<0@dUE1014BUH204=9E@030498E011Ae<0@DMC00L0A4QC
1@13Ae8304E9E0@0ATYE1@17BeHH04U=F0<0BDiD00@0B4aE04I=E@16CED0ADaF0P14CEH01013BeH0
@TaF049<EP0oCEP2049<E@40@daC0@16Bdl10?ooo`003`18EEh304MDG@<0AU=L0`15DU/<04YBG@`0
Be=N3018EEh<04UFG`0404EAG`18E600B5AP04YDGP80B55J00@0B4mG04I9D016BE00>CXo0P0X:Bh<
01`G4`0401dG5`0P6QX081XJ01XF4`8061@A00@06QTD01@E3@0D5@d04a@<0P0Q8QX0100X:1`0>3HY
03Pf:@15@S<2059;>@0402XR3@0n<QX0?S8J04/m8`80CclU0`1I?RT0101I?bP0GD4Y05e1:@1UAbd2
06a><P0307EF>@1jFc`0NU/l00@0OUHl0P20ESd01024F3l0QUY108IJ@@29Fd<208]MA@0409UVB@2H
IDP0V6E809QUB080W6U<00@0XFiA0:AaE02TLE@0Y75D0P2SL5<0102NJdh0X6e@0:1]D02SL5<20:Md
E`040:YgFP2/NE`0[7UL0:ejG@80[W]N00<0YW=F0:UfF@2YMUT0102]NUd20:ikGP030;9oHP2ePVD0
]H9U00@0]8A]0P2eQFh0102fQVl0^8Qa0;R8L@2jRW<20;b<M@040;21I@2lSG40_8ea0;:3I`80XG9F
00@0Y7EI0:UjGP2YNUh0Y7EI0P2RLeL0102ZNel0[GiR0:enHP2`PFD20;B5J@@0]HIZ0P2dQFT0102/
PfD0/XU[0;:9J`2cRV`20;68JP040;B;K@2kTW@0^i9d0;V@LP80/XU[00@0/Xeb0;B?M02dSg@0/Xmc
0P2XQVX0102GNU`0R6e?08Q]C`21J4T2081WB00407IF?`1bDS/0LU8k06Y:<`80HD4Z00@0FSXS05Ti
8P1I>B80Fc/T0P1N?RL0101XA2h0JDD_06U5;`1Y@bh206Ln:P0406Dl:01/@Bh0K44^07U>>`80QEY7
00@0Q6=007=B;`1cDRl0OU`g0P2:J4<0101gE2h0JT@O06Y47`1Y@a`2068l5@0407I89020DRh0P58^
08]M>@80SEdi00@0RUPe08aJ=`2<FSL0VVE30P2VLDl0102TM4h0Xg==0:=cC@2FIT0208iN>00409IV
@02IJD<0VFU3099R?080SEdg00@0Uf4m09iX@P2NJ480W6@o0P2HHCX0102JH3X0V5de09QM=@2HGCD2
0:1U?@040:5Q>`2OHCP0Wf4h09UN=080TUda00@0TF0g08iP?02>H3`0PULg0P1cBRh0101A?b/0:1T6
02PI1P0a9AL203li;@0401dI3@0^:a`0;R/L01dG1`80?CHT00@0?RdK038S400b8a00@34L0P12<Q/0
100e7`P0>1`603PL1P1<:18205Xa6`04078e8`2LGTT0W5i90:EUC080ZVU;00@0]79B0:mZC@2_JTd0
[fM>0P2aJE40102[KeD0Tem909=OB@1fDCl2065:@P040517B013AE40@dEA0459GP80AE5Y00@0AT]N
04I<G016C5`0ATaJ0P17CUP01017CUH0AdmB04M?DP18D58204QAD0<0B4mI00@0BE1J04YAF`1:DE/0
Be9L0P1<Ded604eDGP0404aCG@1;DU`0Be9L04YAF`80BE1J0`18CeT304Q>FP0404M=F@16C5P0ATaH
04E;E`80A4YF3`13BED304E9E@@0ATYF2017BeL01016BUH0ADUE04E9E@15BED204A8E0<0@dMC00<0
ADUD04A8D`14B5<01P13Ae800`14B5<0ADUD04E9E00304E9E0@0ATYE4@17BeH<04Q<E`<0AdmB00@0
ATiA04I=D`16CE<0ADaB0P15C5@01014Be<0A4]E04A;E@13BEL204A;E@40ADaD0@15CE010?ooo`00
3`18EEh304MDG@<0AU=L0`15DU/<04YBG@`0Be=N3018EEh<04UFG`0404EAG`18E600B5AP04YDGP80
B55J00@0B4mG04I9D016BE00>CXo0P0X:Bh<01`G4`0401dG5`0P6QX081XJ01XF4`8061@A00@06QTD
01@E3@0D5@d04a@<0P0Q8QX0100X:1`0>3HY03Pf:@15@S<2059;>@0402XR3@0n<QX0?S8J04/m8`80
CclU0`1I?RT0101I?bP0GD4Y05e1:@1UAbd206a><P0307EF>@1jFc`0NU/l00@0OUHl0P20ESd01024
F3l0QUY108IJ@@29Fd<208]MA@0409UVB@2HIDP0V6E809QUB080W6U<00@0XFiA0:AaE02TLE@0Y75D
0P2SL5<0102NJdh0X6e@0:1]D02SL5<20:MdE`040:YgFP2/NE`0[7UL0:ejG@80[W]N00<0YW=F0:Uf
F@2YMUT0102]NUd20:ikGP030;9oHP2ePVD0]H9U00@0]8A]0P2eQFh0102fQVl0^8Qa0;R8L@2jRW<2
0;b<M@040;21I@2lSG40_8ea0;:3I`80XG9F00@0Y7EI0:UjGP2YNUh0Y7EI0P2RLeL0102ZNel0[GiR
0:enHP2`PFD20;B5J@@0]HIZ0P2dQFT0102/PfD0/XU[0;:9J`2cRV`20;68JP040;B;K@2kTW@0^i9d
0;V@LP80/XU[00@0/Xeb0;B?M02dSg@0/Xmc0P2XQVX0102GNU`0R6e?08Q]C`21J4T2081WB00407IF
?`1bDS/0LU8k06Y:<`80HD4Z00@0FSXS05Ti8P1I>B80Fc/T0P1N?RL0101XA2h0JDD_06U5;`1Y@bh2
06Ln:P0406Dl:01/@Bh0K44^07U>>`80QEY700@0Q6=007=B;`1cDRl0OU`g0P2:J4<0101gE2h0JT@O
06Y47`1Y@a`2068l5@0407I89020DRh0P58^08]M>@80SEdi00@0RUPe08aJ=`2<FSL0VVE30P2VLDl0
102TM4h0Xg==0:=cC@2FIT0208iN>00409IV@02IJD<0VFU3099R?080SEdg00@0Uf4m09iX@P2NJ480
W6@o0P2HHCX0102JH3X0V5de09QM=@2HGCD20:1U?@040:5Q>`2OHCP0Wf4h09UN=080TUda00@0TF0g
08iP?02>H3`0PULg0P1cBRh0101A?b/0:1T602PI1P0a9AL203li;@0401dI3@0^:a`0;R/L01dG1`80
?CHT00@0?RdK038S400b8a00@34L0P12<Q/0100e7`P0>1`603PL1P1<:18205Xa6`04078e8`2LGTT0
W5i90:EUC080ZVU;00@0]79B0:mZC@2_JTd0[fM>0P2aJE40102[KeD0Tem909=OB@1fDCl2065:@P04
0517B013AE40@dEA0459GP80AE5Y00@0AT]N04I<G016C5`0ATaJ0P17CUP01017CUH0AdmB04M?DP18
D58204QAD0<0B4mI00@0BE1J04YAF`1:DE/0Be9L0P1<Ded604eDGP0404aCG@1;DU`0Be9L04YAF`80
BE1J0`18CeT304Q>FP0404M=F@16C5P0ATaH04E;E`80A4YF3`13BED304E9E@@0ATYF2017BeL01016
BUH0ADUE04E9E@15BED204A8E0<0@dMC00<0ADUD04A8D`14B5<01P13Ae800`14B5<0ADUD04E9E003
04E9E0@0ATYE4@17BeH<04Q<E`<0AdmB00@0ATiA04I=D`16CE<0ADaB0P15C5@01014Be<0A4]E04A;
E@13BEL204A;E@40ADaD0@15CE010?ooo`003`18EEh304MDG@<0AU=L0`15DU/<04YBG@`0Be=N3018
EEh<04UFG`0404MCH@1:EV80BUIR04aFH080BE9K00@0AdiF0495C012AD`0<c@i0P0S92P<01`G4`04
01XD500M5aL07ALG01XF4`806aLD00@07A`G01HG3`0F5`l04Q<;0P0L7AD0100T91P0>CLZ03Tg:P13
@34204a5<`0402lW4P16>R80ASXR04do9@80CShT00@0Ec`W05Pm:01H?BP0FShX0P1M@BT0101UAb/0
KE0b06e@<P1fEcP207]L?@<0OEHi00@0OULj081G>`20Ec/0Q5Pm0P27Fd00102:G480S5i408aNA02I
ITT309QUB080W6U<00<0X6e@0:AaE02TLE@00`2TLE@0102OK4l0XFiA0:5^D@2TLE@20:MdE`040:Yg
FP2/NE`0[7UL0:ejG@80[W]N00@0ZGII0:YgFP2ZMeX0ZgQK0P2/NE`0102]NUd0/7eP0;1mH02dPF@2
0;J3IP@0/hAZ0P2dQF/0102fQfd0^8U_0;R9K`2kS7820;b=L`040;>4J02iRVh0^HY^0:n0I080XG9F
00@0Y7EI0:QiG@2XNEd0YgQL0P2XNEd0102[O600[GiR0:enHP2aPVH20;>4J0<0]HIZ00@0]8EY0;>4
J02cQ6P0/HQZ0P2cRV`00`2dRfd0]Xe_0;J=K`030;V@LP040;B;K@2]Q6H0[HAV0:f4J080/8U/00@0
/X]^0:^8JP2[R6X0XH1Q0P2HNEX309EhF00408i^E@2=KE@0SFeD08YZD@80QFE<00@0O5a3075A>01a
DCP0ITH]0P1N?RD0101K=Ah0IT0Y06I0:@1aBc@207I==`0407=:=01cBC<0LdTc07Y@>P80PeI100@0
PUhl07A@;P1dD2h0OUPd0P26H3/0101jDRh0NU</07YC;028GSP208UO>@0408UK=`2>H3`0SV0l09EU
@@80Tf<o0`2?GCX0102GHT00WVU709iYA`2PKTT20:5_BP0409MU@02BH3/0TV0k09]YA080WF]600@0
UF<n091N>@2@GST0UV0l0P2OJD<0102NIT40VF8k09UR>`2IGcT209EJ<P0409=H<02KH3P0Vf0h0:5Q
>`80X68k00@0VV4f09EP=02EH3@0UF@k0P2AID001026Fc/0N4he07Q>=@1C?BP202lL2`0403@W5`0l
<RH0?38V024K3`80=blR00@09ah?03lc8`0o<b<0?RhN0P0b9A@01015>2D0?S8L03hb700`7PP204<Z
5004050a7@1E<1d0EC0M05L/6@80N4hf00@0Oe<h08MJ>`27FS/0VVa;0P2LJdd0102IId`0V6A<09QT
C02EI5D2085GB@0406Y<@P1GB4<0EdQ304]6BP80@dEA00<0@TUK04I>H`16CV<00`16C5`01016C5X0
AdiH04M>F017CUH204M?DP<0B51B0`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@H0CEAN00@0C5=M
04]BG01;DU`0BU5K0P19D5X304Q?F@<0B4iJ00@0AdeI04I<F016C5P0AD]G0P14BUH?04=9E@<0ADUE
1016BUH804M;E`0404I:EP15BED0ADUE04E9E@80A4QD0`13Ae<01015BE@0A4QC04A8D`13Ae830496
D@80@dMB00<0A4QC04E9E015BE@01015BE@304I:EA40Ad]F3018C5L304M?DP0404I>D@16CE<0ATeC
04E<DP80ADaD00@0A4]C04A;E@14BeD0@dUG0P14BeD104E<E040ATiA0@3oool000l0B5EN0`17E5d3
04ICG0<0AE9K301:DUd<04]CGP`0B5EN3019EUl01019EF40BeMS04]GH`1=Ef4204YCG00404A;D@0l
?d@0?3m402`]<@807AhR300L5a<0100M5aL071HF01`F5P0H514201XF4`0401`K5P0G61005aP@01@E
3@806a`D00@0:2PL0490<`12@3<0@Ch_0P0o>2H0100_9a80BclW04/o9`1?@BL204lo9@0405Hk9P1H
?2H0F3`V05Xn9P80H48X00@0IdT]06m@<`1_D3<0MULh0P1mG3d307aE>00407eF>@20Ec/0P5Lk08EI
?P80R5a100<0S5i408eOA@2=GdD00`2GI4L0102HIDP0VfQ;09]XB`2PKE020:=`D`<0Y75D00@0X6e@
0:9_DP2RKe80YG9E0P2XMEP00`2[N5/0[GYM0:ejG@030:ikGP040:ejF`2[N5T0ZgQI0:YgF080ZgQI
00@0[W]L0;>0H@2cP640]X=T0P2hQFH40;:3I@80/hAV00@0]XMY0;R9J`2hRF/0^ha^0P2lSFl0102h
QV/0]hEZ0;N5JP2^O6420:IdF@040:QfF`2YMe`0ZGML0:]iGP80/GmT00@0[7eQ0:enHP2]OV80/85U
0P2cQ6P30;B5J@<0/hAX00@0]h]^0;>7JP2cQfX0/hMZ0P2iSG00102kSg80]XY]0;J:K@2/P6<20:Ij
G@040:enI02`PVP0/89X0;B8K@80/XU]00@0[XMZ0:Z5J02ZQFP0ZXMY0P2[RV/0102XQV/0XX1U0::0
I@2IMe`209=aEP0408e[D025HdP0QF=807]I?P80Le4f00@0N5<i089M@`22GD<0S6A;0P2:HTT01022
F400O58j07aB>P20Dc`208II@P0408YO?@22EcD0PULe08ML>P80REdj00@0Oe4]08QJ=P28FSH0W6a6
0P2NKTP0102KKTL0W6a609a/AP2KJdD209YX@`0409QV@@2IID00VFE009YV@@80W6I200@0VVQ309i/
A`2NK4L0VFM20P2GID030:5_BP0409IT?`2BH3/0TV0k09AN>P80WfU300@0X6Q309YS?02JHc`0V5hh
0P2AERh0102>Db/0UU/c09IK<`2RHSh20:5S?00409aS>02II3X0VF@j09QX@080UFU600@0R5m107YB
>01jDSP0ECXW0P0h8A40100j9aT0=bTN03LY7P0X7102040d:00403HY6@10<200@30P03X^8080<bXI
00@0BT<a03Xb7`0j<Ql0<2<B0P1=>RT0101@=RL0CS4Q04ha8@19=Ad2061<<`04069=<01ZECH0JUDf
089[B`80R71D00@0PfUB07aQC01lHD`0M5ID0P1XCTd0101GAdL0C4I804a6B016AD/20495CP0404A8
E018BeX0B4]J04I<G080ATaJ00@0ATaH04M>F017CUP0AdiF0P17CU@00`18D5<0B51B04Q@DP0304Q?
F@0404U@FP1:DE/0BU5K04]BG080C5=M1P1=E5h0101<Ded0Be9L04]BG01:DE/204U@FP<0B4mI0`18
CUX01017CET0ATaH04I<F015BeL204A:EPl0@dUE0`15BED404I:EPP0Ad]G00@0ATYF04E9E@15BED0
ADUE0P14B5@304=7D`<0A4QC00@0@dMB0496D@12AU40@TIA0P13Ae8404A8D`D0ADUD0`16BUD?04M;
EP`0B4aG0`17Ce801016CU40ATeC04I=D`15C58204E<E00404A;D`14BeD0A4]E04A:F080A4]E0@15
C5@104I>D@40oooo000?04QEGP<0AeAM0`16De`304EBF``0BU9M301;Deh<04QEGP`0BEIO00@0BEEQ
04]GH`1;Ef<0CEMQ0P1:De`01014Be40?3m403`oA00/;C4201dN8P`071LC00@07ALG01`F5P0L5QH0
61@A0P0J5Q<0100L6aH05aP@01LH400D5@d201/L500402PX7012@3<0@T0c044n;`80?cPV00@0;bLB
04/o9`1;?bL0Cd4W0P1??bD0101F>bH0F3`V05Pl9P1J?RH20612:00406M9;@1_D3<0Ke0c07IG>080
OE`m0`1lECP0101mEST0P5Lk081G>`25FCh208QL@@0308aNA02=GdD0SEm500<0UfA700@0V6E809]X
B`2KJ4/0X6e@0P2SL5<30:AaE0040:1]D02RKe80XVmB0:EbE@80Z7EH00<0ZgQK0:ejG@2]NUd00`2^
Neh0102]NU/0ZgQI0:]hF@2ZMeP20:]hF@040:ikG02cP640/h1Q0;J3I080^8EV102bPfD20;>4IP04
0;J7J@2hRF/0^8U[0;^<KP80_8e_00@0^8I[0;N5JP2gQFX0[WaQ0P2VM5T0102XMU/0ZGML0:UgG02[
NEh20;5oI0040:amH@2]OV80[GiR0;21I@80/hAX0`2dQFT30;>4J0040;N;KP2cQfX0/hMZ0;>7JP80
^He`00@0^hmb0;J:K@2fRVd0[81S0P2VNUd0102]OV@0/89X0;22J02dR6d20;:9K@040:j7JP2ZQFP0
ZXEX0:Z7J@80ZhY[00@0Z8I[0::0I@2RP6D0VGML0P2CLEH0102=Je00QF=808ESB01kFCh207=A=P04
07QC>@22GD<0PUe308aTB`80RV9900@0PUQ007aB>P1lDSX0P5<l0P26FD80102:Gcd0PULe089G=@27
G3X208UM>P0407mA;@28FSH0R5Xf09a/AP80WVi800@0Vfi709a/AP2LK4H0Vf]50P2JJ4<0102HIT40
VFE009UU@02JIT4209aV@P0409YX@`2NK4L0WVa709UW@P80UfE00`2QKdX0102FI3l0TV0k099P>`2D
GSX209mY@`040:1X@`2JHc`0VV<l09QN>080TEH^00@0SU<[09IK<`2FFc<0XV8n0P2QHc`0102LHcP0
VF@j09UT>P2HJ40209EYAP0408QO@@1jDSP0NU8h05Dj9`80>24A00@0>RLI03LY7P0g:Ah0:1`@0P10
=2P0100f:AT0@30P040`800j;R0203<Z6@0404Y3<@0j<Ql0>S8O030S4P80CCXY00@0D3HW04ha8@1>
<B40BCDM0P1PC3<0101RCC00JUDf06YE=P22Jd/208Q`E00408=YDP1lHD`0O65<07AFE080J4i=00@0
EdM704a6B01<ATP0ATE;0P12ADh01014B5@0B4]J04Q;FP16C5`204I<FP0404I<F017CUP0AdiH04M>
EP80AdiD00<0B51C04Q@DP18D5800`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@H0CEAN00@0C5=M
04]BG01;DU`0BU5K0P19D5X304Q?F@<0B4iJ00@0AdeI04I<F016C5P0AD]G0P14BUH?04=9E@<0ADUE
1016BUH804M;E`0404I:EP15BED0ADUE04E9E@80A4QD0`13Ae<304A8D`0404=7DP12AU40@TIA0496
D@80@dMB1014B5<504E9E0<0ATYE3`17BeH<04Q<E`<0AdmB00@0ATiA04I=D`16CE<0ADaB0P15C5@0
1014Be<0A4]E04A;E@14BUP204A;E@40ADaD0@16CU410?ooo`003`18EEh304MDG@<0AU=L0`15DU/<
04YBG@`0Be=N3018EEh<04UFG`0404QDH01:Ef00BUMP04eGH080BU=J00@0@DQ>03Dh?@0e>3d09RL[
0P0J6ad<01`G4`04028L700L5QH071HF01DA3P805Q8?00@05aHA01HG3`0F5`l061TA0P0O81P0100a
<BD0C4Xm04a:?@11?Rl2034Z600402dU401<@2P0C40X0512:080D40V00@0FCdW05Xn:01J?RP0GT0X
0P1RA2X0101YBRd0Ke0a06m@<@1gESD207YI>00407]D=@1lECH0O5Df07eF=`80P5Li00@0QETl08UM
@029GD00SF130P2?HTD0102EHTD0UV=609ISAP2HIDP209]XB`0409m/C`2RKe80XVmB0:AaE080YG9E
00@0XFiA0:=`D`2SL5<0YW=F0P2YMUT00`2[N5/0[GYM0:ejG@030:ikGP040;5nG`2]NU/0[GYK0:Uf
E`80ZWMH00@0[gaM0;F2H`2ePV<0^8EV0P2iQVL30;22H0040;63H@2bQ680/XAR0;F7I@80]hUW00@0
^XaZ0;^=J`2kSF/0^XU[0P2cPV@30:elGP040:akG@2XMeT0Z7MI0:UhFP80/7mQ00@0ZGYL0:]lGP2[
O5h0[WmQ0P2aPV@00`2cQ6H0]HIX0;F6J0030;F6J0040;V=L02`Q6L0/8AW0:j2I@80]HU/00@0]h]^
0:n3IP2_PfH0Yg]N0P2TN5/0102bOf@0]X=X0;J3J02iQf`20;V:KP040;J:K@2dRfd0]8]]0;F>K`80
]i1a00@0/Xmc0:N4J02WQ6P0VWMK0P2DLED0102FLeL0UgAH09MdF02DLED2091]D@0409UaE`2FKU@0
UViD08eUB`80PUPo00@0N4he07U==01iCC@0Oe<j0P27FD40102:G3/0SEln08eO?P2FID@209=S?`04
08]I=P2FHSd0UV8m0:9^B@80Wf]600<0XW9<09i^B02NKTP00`2LJTD0102QKDP0Xfm:0:=_BP2RK4P2
0:1ZAP0409]W@P2OJdH0Wf]609eYA080WVY500@0Yg=>0:A`B`2TL4/0V6@o0P2DH3/0102@FSH0W6I0
09aV@02PJ4<209]T?@0409QN>02@EBd0T5D]08aA:@80U5Ta00@0XV@o0:1T?`2PI3l0W6Dl0P2IIS/0
102IJD40U6U609AYAP26GCl207M?=P0405Df8P0o91<0?b@C03/U5`80<R0D00@0;ahD04Lg:P17=bX0
@S0R0P0o;1d0100g;b80>38R03Pb8P1<ASH203Pb8P0403Xc8`1BA3L0DT@g04<`8P80C3@X00@0A3PR
05=7;`1CAbl0DDD[0P1DAR/0101UEcd0K5e806aMB01TE4D205a<?`0405U=E`1DBU<0E4YC04Y7CP80
ATI>00@0@dI=04=6C`13ATl0AdMA0P1:B5<01016C5X0ATaH04I<F016C5P204M>F00404M>EP17CU@0
AdiD04Q?E@80B51C0`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@H0CEAN00@0C5=M04]BG01;DU`0
BU5K0P19D5X304Q?F@<0B4iJ00@0AdeI04I<F016C5P0AD]G0P14BUH?04=9E@<0ADUE1016BUH804M;
E`0404I:EP15BED0ADUE04E9E@80A4QD0`13Ae<00`14B5<0@dMB04=7DP060496D@0304=7DP14B5<0
A4QC00<0A4QC1015BE@504I:E@`0Ad]F3018C5L304M?DP0404I>D@16CE<0ATeC04E<DP80ADaD00@0
A4]C04A;E@14BeD0ATUH0P17BeH104Q<E@40B4eA0@3oool000l0B5EN0`17E5d304ICG0<0AE9K301:
DUd<04]CGP`0B5EN3019EUl01016De`0B5EM04QEG@1;EEh204M@E@0403e5B00_<cH0;c<f028S9@80
6Q/M300L5a<0100P6QX06A<C01TC4`0G4a0201XF4`0401PG4P0H6A4061TA01`M5@80824I00@0;blS
04i<?`1>C3l0A44b0P0`:AL0100b:QD0CT8Z04i2:P1A@bT204lo9@0405`m:01N@2P0GT0X0653:@80
ITL[00@0JT/^075@<@1aD340MU<c0P1hES<307aE=P0407iG>021F3X0PEPj08IJ?@80REe000@0SF13
08mRA@2?HTD0U6540P2EHTD0102HIDP0VfQ;09]XB`2NJdh20:5^D@040:AaE02VLeH0YW=F0:=`D`80
YG9E00@0YgAG0:YgFP2ZMeX0[7UL0`2^Neh20:mlG`040;>1H02_OE`0[geL0:YhE`80ZgUH00@0[geL
0;B2H@2dPV40]hET0P2hQVD30;22GP040;63G`2bQ600/XAP0;B6HP80]XQT00@0^8YV0;Z<J02jS6P0
^XMX0P2bOf00102aOUl0]85R0;B1HP2`OEh30:QeEP80[7UJ00@0YWQG0:QjF@2XNUT0ZgeL0P2_PF00
102cQF@0]XQW0;J8I`2hRVT20;V;JP040;f@L`2cQVT0/hIY0:n2I@80]HQ[00<0]hY]0;:5J02bQFP0
0`2]P6<0102fP6H0^H=Y0;V3J@2jQf`20;^9KP040;Z;K`2jSG00^Xe`0;Z>L@80^Y5c00<0]I1c0:j9
K02^RF`00`2XPfH30:f8J`040:F0H`2MN5/0WGQK0:5jG@80Ug1C00@0RV55085H?021F3`0PUHk0P28
FT00102>H4H0TVA:099TBP2;FST209ES@P0409i/B@2NJDL0WVU709YV@@80YFm;00@0[GMA0:YbC@2Z
LTd0XW590P2NKDD0102JJD40WVY309iZ@`2SKdP20:MaB`040:I^B@2SJdH0Xf]609eYA080XFe800@0
Wf]60:5]B02QKDP0Z7A?0P2QKDP0102EHC`0TUhi099N>@2<ES8209US?@0409iV@@2KI3d0Vf@m09YP
>P80TeP`00@0Se@/09IK<`2FFc<0Wf<o0P2OICl0102LICh0UfHm09MV?@2EJ442091U@`04081I?01_
BS00KdX`05Dd8@80@B@B00@0>R<E02lK400_6a00=bLJ0P16=2P01018=BH0?R/J03h[6P0j<b<2040j
:P0404Q5=00j=bH0>SLV04Y4=P80CdDi00@0=bTN04hl<P1>?380D3l]0`1JBCD20591;@0405M6=P1L
Bd40G4]105M5@@80DT1000@0BdYJ04U;FP19BeX0Ad]G0P11BU<01011B580@dM@04=7D017Adl204U6
C`@0ATeG0P17CUP304M>EP<0B4mG0`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@H0CEAN00@0C5=M
04]BG01;DU`0BU5K0P19D5X304Q?F@<0B4iJ00@0AdeI04I<F016C5P0AD]G0P14BUH?04=9E@<0ADUE
1016BUH804M;E`0404I:EP15BED0ADUE04E9E@80A4QD0`13Ae<304=7DP040496D@11AE00@DE@0455
D080@TIA0`13Ae8404A8D`D0ADUD3`16BUD<04Q<E`<0AdmB00@0ATiA04I=D`16CE<0ADaB0P15C5@0
1014Be<0A4]E04A;E@18BUT204Q:EP40BDaE0@1:CE810?ooo`003`18EEh304MDG@<0AU=L0`15DU/<
04YBG@`0Be=N3018EEh<04UFG`0404ICG018EEd0B5EM04]EGP80Ae1E00@0?DE802lc=P0_<cH08R<U
0P0J6ad<01`G4`04020J6P0I4a<06A<C01LC40806QHC00@061LB01PI4@0H6A4071dE0P0P8AT0100_
;b<0CT`o04i<?`14@C82030Y5`04038Z5@1>@RX0CT8Z0553:@80CclU00@0G3dX05i0:01N@2P0HD<Y
0P1VAb/0101ZBbh0LE0a075@<@1fDc<207QF<`<0O5Df00@0OULh085H>P21F3X0QUXm0P29GD00102=
H4<0Sf9508mRA@2DHD@209ERA@0409QUB02KJ4/0VfQ;09i[CP80XFiA00@0Y75D0:IcEP2VLeH0Xg1C
0P2ULUD0102WM5L0ZWMJ0:YgFP2/NE`30:ikGP80[gaO00@0/h5P0:mmG02_OE`0ZWQG0P2[NEP0102_
OE`0]89Q0;B2H@2gQF@20;R6I@<0/89N00@0/H=O0;:4H02bQ600]8IR0P2fR6@0102hRVH0^XaX0;Z<
J02jQfP20;9oH0040;5nG`2dPF80]85R0;1mGP<0Z7EF0P2/NEX0102VN5L0Z7YI0:QjF@2[OE`20:n1
H0040;>5I02fR6L0]XQW0;R:J@80^H]Z00@0_I1c0;>6J@2cQVT0[h9U0P2eR6/00`2gRVd0/XEX0;:5
J0030:f0H`040;J0IP2iPfT0^H=Y0;Z7K080^hU^00@0^X]_0;Z=L02jSG00^Xia0P2jTG<00`2eT7<0
[XU/0:j9K0030:R3IP<0[HQ[00@0YH1S09ehF`2MN5/0XGYM0P2GL5<0102:HDD0PEPl085H?022ES/2
08QJ@00408iPAP2BI4X0TVA:08]J>@80UF=200@0WVa909iYA`2NJDL0VVI10P2UKd/0102]Me40ZW9=
0:YbC@2RLDT209i]A@0409YY@@2NJT<0WVY30:=_B080Yg5;00@0YVi90:=[AP2SJdH0WFU40P2QKDP0
102OJdH0XFe80:5]B02XM4l20:5]B00409EQ?02BGST0TUhi08aF<P80VF<m00@0WVI109]T?@2KI3d0
VV0j0P2CF300102?E2`0UU/c09IK<`2OHcl209mU?`0409aU?P2GISd0UfHm09EX@@80T6E300@0P5Tl
06m:<01_BS00EC@Q0P119180100j8aD0;a/@02lK400g9aX204Hd:00404Pe9P0n:aX0?R/J03Xc8`80
@3XZ00@0B4Dd03Xg9P0j=bH0BT@f0P1?ACT0100g:Ah0CS`b04hl<P1@?bd305Y9=@80DT4]00@0EdHf
05a;@@1LBd40EdE10P1B@400101;BUX0BD]J04U;FP17BeL2045:D`040458DP13Ae00@dM@04M7C`80
BDI?1016CEL204M>F0<0AdiF0`18CeL304Q?F@0404U@FP1:DE/0BU5K04]BG080C5=M1P1=E5h0101<
Ded0Be9L04]BG01:DE/204U@FP<0B4mI0`18CUX01017CET0ATaH04I<F015BeL204A:EPl0@dUE0`15
BED404I:EPP0Ad]G00@0ATYF04E9E@15BED0ADUE0P14B5@304=7D`<0@dMB00@0@TIA0455D011AE00
@DE@0P12AU4304=7DP@0A4QC1@15BE@?04I:E@`0B4aG0`17Ce801016CU40ATeC04I=D`15C58204E<
E00404A;D`14BeD0A4]E04Q:F@80B4YF0@19C5D104Y=DP40oooo000?04QEGP<0AeAM0`16De`304EB
F``0BU9M301;Deh<04QEGP`0BEIO0`16De/01018DeT0A4eB04A=DP0g?d8202T];P0301lP8P0L7Qd0
71hM00`071LC00@06A<C01H@400F41006aLD0P0R7Q/0100L6aH06Q/C01XK4`0O81P2028S6`0402LW
6`19AcX0BDLj04]8>@80>34O00@0>C4L04e1:@1=@BT0DD<Y0P1>?R@0101L?RH0GT0X05i0:01S@bX2
06M8:`0406i=;P1cD300Le0`07ID<@80MeDb00@0OEHe07iG=P1nEcH0OePg0P21F3P01026Fc/0REhn
08UN?P2=H44208mR@`04099O@P2EHTD0UF9509QUB080VfQ;00@0WV]>0:1]D02PKE00Y75D0P2WM5L0
102ULUD0YW=F0:IcEP2XMEP20:]hF`030:ejG@2^Neh0[W]N00<0[gaO00<0/X1M0;1nF`2`OU/00`2]
NeP0102_OEX0/X1M0;:0G@2dPUl20;F3H0H0/H=Q00@0/XAR0;B6I02dQV@0]XQV0P2gRFL0102hQFH0
/7eN0;1mGP2bOf020;N4I@040;1mGP2YMUL0ZGIG0:QeEP80YgAE00@0YWQG0:MiF02WNEP0ZWaK0P2^
P5l0102bQ6<0]HMV0;F7IP2hRVT20;V;JP040<2AM@2kS700^ha`0;R9K@<0^X]_0P2iRVh30;R9K@04
0;N1I@2hPVH0^89V0;V6J@80^XU[00@0^Ha]0;Z?K`2jSfl0^Y5a0P2kTW80102kU7D0]Xm`0;J?L02b
Rf`20;B=KP040;J?L02cS6d0/ha]0:R1HP80WWMH00@0XGQJ09acE@2LLeD0V6mA0P2IKE00102KKU40
W6mB09a_DP2KK50209QYC@0409a[BP2SLDh0Xg5>0:IdD@80YW9=00@0YG5<0:egD@2]Me40/WaF0P2`
Ne<0102SLTX0XG180:5`B02PK4D209m[A0040:5[A@2SJdH0Xf]60:9ZA@80X6Q300@0XVa80:=]B@2S
KDT0X6Y60P2RK4P00`2WLDd0WfU509mYA@0309AN>P0408iH=02HHS`0V68l09]S?P80VV<l00@0W68l
09ML=02GG3@0TeP`0P2HGCD0102MHCd0WF<n09eS?P2II3`209IU?@04099V?`2;Hd00Rf=007UD>@80
J4@/00@0ESLR040U4P109A80>2<B0P0b81801014=2L0@30Q040`8@18=bL2044_6`0404Dm:P16?b`0
ASl/04M2;`80@d0_00@0CdTi04Y2=@1:@SD0?c4V0P1GAS`0101OBST0IE0m06E@?@1TDSh205I5<`04
05A2=P1KBT<0FdY305Q5A`80EdA:00@0ATMI04Q;FP18BeX0ATiK0P11BeD01010BE80@dYB04=:DP19
BE4204Q7C`@0ATeE0P17CUH304M>F0H0B4mI00@0BE1J04YAF`1:DE/0Be9L0P1<Ded604eDGP0404aC
G@1;DU`0Be9L04YAF`80BE1J0`18CeT304Q>FP0404M=F@16C5P0ATaH04E;E`80A4YF3`13BED304E9
E@@0ATYF2017BeL01016BUH0ADUE04E9E@15BED204A8E0<0@dMC00<0@dMB0496D@12AU401P11AE00
0`12AU40@dMB04=7DP0304=7DP<0A4QC1015BE@>04I:E@`0B4aG0`17Ce801016CU40ATeC04I=D`15
C58204E<E00404A;D`14BeD0A4]E04Q:F@80BD]G0@1:CEH104Y=DP40oooo000?04QEGP<0AeAM0`16
De`304EBF``0BU9M301;Deh<04QEGP`0BEIO00@0B5EM04MDFP17E5X0Ae9H0P11Bdd00`0b>S`08RHW
028V9`0301/M70`071LC00@06aDE01H@400F410071PE0P0Q7AX0100G5Q4061TA01PI4@0U9Qh202`]
9@0402h^8P1<BSd0C4Xm059?@080?CHT00@0=RhI04<g7`13=al0CT0V0P1>?R@0101I>B80G3`S05`l
8`1R@bL206Y8;004071?<01eDS80ME8b07UE<`80NeLc0`1oF3L01020FCP0PUTi089I>@26Fc/208UN
?P0408aO@02>HD80SV52091M@080U65400@0V6E809]XB`2KJ4/0WFY=0P2PKE00102TLE@0YgAG0:Md
E`2VLeH20:MdE`040:UfF@2/NE`0[7UL0:ikGPD0[gaO102aOe`20:mmFP040:ilF@2_OEX0[geJ0;5o
G080/h5N00<0/XAS0;63HP2aPf80102aPf820;>5I0040;B6I@2eQfH0]HMV0;N2HP80/7]K00@0/giN
0;F0H02eP600[WUI0P2/MeL0102^NET0ZGAD0:UdE02XNUP20:UkF@040:]mF`2^P5h0[X1N0;63H@80
]8IT00@0]XQV0;N9I`2gRFL0^X]_0`2nSg<20;^<L0040;V:KP2jRfl0^X]_0;Z;K`80^8U]0`2kR6/0
102jRF/0^HY/0;V:K02hSFd20;N>KP<0]i1_00@0_9=c0;J=K@2fSFd0/HQX0P2`QfL0102aR6P0/8MW
0;27I`2ZPF420:ElG0040:1gE`2RNET0XWUI0:EjFP80XgQH00<0Wg9C09]/CP2KK4h00`2IJT`0102Z
O5P0ZgeI0:]mF@2[NeD20:YhD`040:]jDP2]NE80[GUB0:YfCP<0YW9:0P2WLd/0102ULDT0Xfi60:=^
AP2NJD4209aU?P0309mU?`2QId40XFM100<0Xfe900@0WVQ40:1ZAP2PJTH0Y6i:0P2LIT80102DGSX0
V68n09QR?P2CGCT209YT?P0409YR?@2IHS/0VF8k09mU?`80Vf0h00@0UU/c09UN=P2IGSH0VEhl0P2I
HC`0102HHS`0U6@l09AT?02CId8208aS@`0407UD>@1XA2`0J4@/05Pn9`80@BPB00@0=b@C03LZ6P0g
:QX0CT8b0P0i;Ad01016>BH0ACHQ04Df8@1EBC<204Pm9`0404M1:`1@BcL0D4/g04M2;`80Ach_00@0
Fde0069@AP1RD4H0GTlh0`1KCcL204Xo:`0404Dl;@1@Ad80D4M204]2A@80BT1800@0AD=@04Y:EP1:
BUH0BDeH0P12Be@01010BE80@daC04=<D`17Be@204M:D`@0ATeE0P17CUH304M>F0H0B4mI00@0BE1J
04YAF`1:DE/0Be9L0P1<Ded604eDGP0404aCG@1;DU`0Be9L04YAF`80BE1J0`18CeT304Q>FP0404M=
F@16C5P0ATaH04E;E`80A4YF3`13BED304E9E@@0ATYF2017BeL01016BUH0ADUE04E9E@15BED204A8
E0<0@dMC00@0@dMB0496D@12AU40@DE@0`10A4l20455D0030496D@13Ae80@dMB00@0@dMB0`14B5<A
04E9E0`0B4aG0`17Ce801016CU40ATeC04I=D`15C58204E<E00404A;D`14BeD0A4]E04Y9F@80Bd]G
0@1<C5H104e>D`40oooo000?04QEGP<0AeAM0`16De`304EBF``0BU9M301;Deh<04QEGP`0BEIO00@0
B5EM04MDFP17E5X0Ae9H0P11Bdd00`0b>S`08RHW028V9`0301/M70`071LC00@06aDE01H@400F4100
71PE0P0Q7AX0100G5Q4061TA01PI4@0U9Qh202`]9@0402h^8P1<BSd0C4Xm059?@080?CHT00@0=RhI
04<g7`13=al0CT0V0P1>?R@0101I>B80G3`S05`l8`1R@bL206Y8;004071?<01eDS80ME8b07UE<`80
NeLc0`1oF3L01020FCP0PUTi089I>@26Fc/208UN?P0408aO@02>HD80SV52091M@080U65400@0V6E8
09]XB`2KJ4/0WFY=0P2PKE00102TLE@0YgAG0:MdE`2VLeH20:MdE`040:UfF@2/NE`0[7UL0:ikGPD0
[gaO102aOe`20:mmFP040:ilF@2_OEX0[geJ0;5oG080/h5N00<0/XAS0;63HP2aPf80102aPf820;>5
I0040;B6I@2eQfH0]HMV0;N2HP80/7]K00@0/giN0;F0H02eP600[WUI0P2/MeL0102^NET0ZGAD0:Ud
E02XNUP20:UkF@040:]mF`2^P5h0[X1N0;63H@80]8IT00@0]XQV0;N9I`2gRFL0^X]_0`2nSg<20;^<
L0040;V:KP2jRfl0^X]_0;Z;K`80^8U]0`2kR6/0102jRF/0^HY/0;V:K02hSFd20;N>KP<0]i1_00@0
_9=c0;J=K@2fSFd0/HQX0P2`QfL0102aR6P0/8MW0;27I`2ZPF420:ElG0040:1gE`2RNET0XWUI0:Ej
FP80XgQH00<0Wg9C09]/CP2KK4h00`2IJT`0102ZO5P0ZgeI0:]mF@2[NeD20:YhD`040:]jDP2]NE80
[GUB0:YfCP<0YW9:0P2WLd/0102ULDT0Xfi60:=^AP2NJD4209aU?P0309mU?`2QId40XFM100<0Xfe9
00@0WVQ40:1ZAP2PJTH0Y6i:0P2LIT80102DGSX0V68n09QR?P2CGCT209YT?P0409YR?@2IHS/0VF8k
09mU?`80Vf0h00@0UU/c09UN=P2IGSH0VEhl0P2IHC`0102HHS`0U6@l09AT?02CId8208aS@`0407UD
>@1XA2`0J4@/05Pn9`80@BPB00@0=b@C03LZ6P0g:QX0CT8b0P0i;Ad01016>BH0ACHQ04Df8@1EBC<2
04Pm9`0404M1:`1@BcL0D4/g04M2;`80Ach_00@0Fde0069@AP1RD4H0GTlh0`1KCcL204Xo:`0404Dl
;@1@Ad80D4M204]2A@80BT1800@0AD=@04Y:EP1:BUH0BDeH0P12Be@01010BE80@daC04=<D`17Be@2
04M:D`@0ATeE0P17CUH304M>F0H0B4mI00@0BE1J04YAF`1:DE/0Be9L0P1<Ded604eDGP0404aCG@1;
DU`0Be9L04YAF`80BE1J0`18CeT304Q>FP0404M=F@16C5P0ATaH04E;E`80A4YF3`13BED304E9E@@0
ATYF2017BeL01016BUH0ADUE04E9E@15BED204A8E0<0@dMC00@0@dMB0496D@12AU40@DE@0`10A4l2
0455D0030496D@13Ae80@dMB00@0@dMB0`14B5<A04E9E0`0B4aG0`17Ce801016CU40ATeC04I=D`15
C58204E<E00404A;D`14BeD0A4]E04Y9F@80Bd]G0@1<C5H104e>D`40oooo000?04QEGP<0AeAM0`16
De`304EBF``0BU9M301;Deh<04QEGP`0BEIO00@0BeQP04QEF`18EE/0Ae9F0P10BT`0100_=cT07R8S
01hR8`0H6QT201XL6@`071LC00@08adM01TC4`0I4a<06QHC0P0I5A80100>3@P05AH>01DF3P0/;BD2
03Xk<`0403lo<`1DDTD0E59505MDA@80?3DS00@0;2@?03HZ4P0f:Q80BC/Q0P1??bD0101E=A`0FCTP
05Ti801R@2D206Y8;004079?<@1hES<0N5Hc07aH=P80OETe1020FSH2089J=`0408IK>@29GS`0REhl
08aO?P80SF0o00@0T5e009AQA02DHD@0V6E80P2KJ4/0102MJTd0Wfa?09m/C`2TLE@20:QeF0040:Ic
EP2XMEP0Z7EH0:YgFP80[7UL00<0[W]N0:mlG`2_O5l00`2_O5l0102_OEX0/X1M0;:0G@2cPEh20;5o
G0040:ilF@2/NUL0[7YG0:ilF@80/GmL00<0/hAV0;:3I@2bPfD0102aPV@20;:3I@<0/hAV00@0]H1P
0:mjFP2_NUX0/WeM0P2aO5`0102ZMED0[GQH0:ehF02bOEd20:]fEP040:]mF`2/OU`0[7iL0:eoG@80
[X1N00@0/89P0;:4HP2bQ680/hES0P2dQV@0102bPfL0^ha`0;^<L02oT7@20;V:KP040;F6JP2fQf/0
]XM[0;J7J`80/H9V00@0`I1`0<2?K`30Sfl0_8i]0P2jSF`0102hSF/0]8iZ0;B>JP2cSFT20;2<J004
0;N<JP2dRFL0]8UW0;:7I@80]HYX00@0^8e[0;V>K02iSV`0]X]Y0P2bQfD0102/Q640/XMU0;:7I@2b
QfD20:aoGP040:9eE02QLe80XG=B0:UkFP80/hET00@0YWYE0:MiE@2WNED0YgYC0P2]OEL0102`P5P0
ZgYA0:]jD@2QL4L209m[@`040:EaB@2XM4`0Z7A<0:YeC@80Y6m700@0WVM009YS?02JHc`0WfDo0P2S
JD<0102TKTX0XVa80:9/B02LIT8209eW@`040:9/B02KID40VfE109EO>`80W6I200@0V68n09aV@02L
IT00VV8m0P2IHS/0102PIT00WV<k09iS>`2GG3@209UN=P0409ML>P2GGcX0Uelj09MS?080UFDm00@0
U6Q308aUA02<ID@0NeHk0P1ZARh0101K@b/0?b/C03l[4`0h9aD203db800405I=?00e;AX0=BdJ04Dj
9P80B3`V00@0HE<i04Lk8`17>b<0BD4Z0P1IDcd0100m=R<0Ac/[04Lk:`1gIeX206QFBP0406AP@`1M
FC`0GETl05UH?080AdL_00@0@d@b04m@B01?D4P0BD]:0P12ADX01015?dT0BTM@04Y7D018Be820458
D00403m8C`13C5D0@daE04M>F080Ad]F0`16CE<00`16CED0AdiF04M>EP0304M>F0<0B4iJ0`18CeT0
1019D5X0BU5K04YAF`1;DU`204aCG@H0CEAN00@0C5=M04]BG01;DU`0BU5K0P19D5X304Q?F@<0B4iJ
00@0AdeI04I<F016C5P0AD]G0P14BUH?04=9E@<0ADUE1016BUH804M;E`0404I:EP15BED0ADUE04E9
E@80A4QD0`13Ae<30496D@040455D010A4l0@4A?0414C`80@DE@0`12AU4404=7DPD0A4QC3`15BE@<
04Q<E`<0AdmB00@0ATiA04I=D`16CE<0ADaB0P15C5@01014Be<0A4]E04A;E@1;BUX204];E`40C4aF
0@1=CU<10?ooo`003018D5/0101;Deh0BU9M04YBG@19DE`304Q@F`80BE5L00@0BU9M04]CGP1;Deh0
Be=N0P1:DUd01019DE`0B51K04Q@F`18D5/204UAG00304YBG@1;Deh0Be=N00`0BU9M2@18EEh304MD
G@`0B5EN00@0BE=L04eFG@1=EUd0BU=J0P0oA4P0100Z;S407AhP01dN800J6QX201dM7@<06aHB100L
5a<601dH508081/G00@05a@?01TF3`0I5Pl061P@0P0O8AH0100g>Bh0?Clb03do<P1<D48204e?@@04
04M5=P0g<Ql0=c8O03@Y5@80AcPQ00@0E48Z04hj8@1>>R40FCTP0P1N?RD0101VAb/0KD/_06e;;`1a
D34207EB<P0407QF<`1lF3@0O5Pd07iF<P80P5Pd00@0QEXh08EJ>P25FSX0QEPi0P27F3`0102;G400
SUm308iO@`2CH4<209ERA@0409UVB@2LJD`0W6U<0:1]D080XVmB0`2TLE@0102ZMeX0[7UL0:aiG02^
Neh20:ejG@<0ZgQK00@0[7UL0:ejG@2]NUd0[geL0P2^O5/0102]NeX0[7YI0:ajF@2^O5/20;1nG@04
0;B2H@2fQ6<0]XAS0;>1H080]XAS00@0]H=R0;5oGP2aOeh0/h5P0P2iQfH0102kRFP0^8IU0;R6I@2d
PUl20;:0G@030:ilF@2[NEH0ZgUF00<0ZWQE00<0ZgUF0:ajE`2/NUL00`2/O5P0102]OET0/H5M0;61
G@2eQF420;N7H`030;F5H@2cPel0/h=O00<0]XYW00@0]8QU0;>5H`2cQF<0/XAR0P2gQVD0102lRfX0
`I1_0<6@K`33TG020<6@K`040;f?K@2jSV/0^Xi[0;V>J`80]hm[00@0]i5/0;RBK@2hTVd0^HeX0P2g
RfH0102dR6<0/hMR0;>7HP2dR6<20;V=J0040;jBK@31UG00`IE`0;N=I`80]8QS00@0/8AO0:n1G@2_
PEd0[h5M0P2`P5`0102/O5P0ZGUE0:UiE@2YME030:QdC`80ZWIA00@0ZgMB0:UeD02YME00YG5<0P2R
KTT0102VLTd0ZWI?0:YfC`2]NE820:]fCP040:=^AP2NIch0WVLn09iW?P80XVQ000@0YF]50:5W@@2Q
Id40WfQ10P2TKDH0102UL4P0WVY209iZ@P2HI3`209MS>`<0U5hj00<0UV0l09US?`2IHcl00`2KID40
102HHSh0UElk09EO>`2FH3X209AP>@0409][A`2IK4/0VFa;08QO?`80RfQ:00@0Le@g06Q9;@1XBBd0
IUA00P10;A`0100g91<0@C0P044`801??Rh2048b8`04034T501@@SD0D48e06eRC080?34M00@0D4Da
04e2<01=@S00@3DS0P1AACD0101ZGTh0DTHh0596>01VGU<205]EA@0404A1;P0m?R`0?Ch/04iAAP80
E5YJ00<0BDmK04A;GP14Beh00`12AE@01013AUD0A4QD04A8E015BED204E9E00304A8D`14B540A4QA
00@0BDeH0`18CeT304QAFP80Ae5J0`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@<0CEAN0`1<Ded3
04]BG0<0BU5K0`19D5X304Q?F@0404M>F016CEL0ATeG04E<EP80A4]E3`13BU@<04Q<E`<0Ad]F00@0
ATYE04E9E015BE@0ADUD0P14B5<304=7DP<0A4QC0`13Ae830496D@<0@DE@1P13Ae8304A8D`<0ADUD
3014B5<01015BE@0ATYE04I:E@17BeH204Q<E`<0BDeH0`18C5L404A<C`<0A4]A0`14Be<204A;E@05
04];E`1;BeD0Bd]E04]<D@1;C5000@3oool000`0B51K00@0Be=N04YBG@1:DUd0BE5L0`18D5/204UA
G00404YBG@1;Deh0Be=N04]CGP80BU9M00@0BE5L04Q@F`18D5/0B51K0P19DE`00`1:DUd0Be=N04]C
GP0<04YBG@T0B5EN0`17E5d<04QEGP0404UCG01=EUd0CEIM04YCFP80?dA800@0:Rha01dN800M7R00
6QXJ0P0M7Ad301/F4P@071LC1P0M61@2020K5`0401LD3`0I5Pl06AH?01PH40807b4F00@0=cT^03do
<P0m?c80C5120P1=Cd401017ACH0=c8O03Lb7`0d:AD204Lh8@0405A2:P1>>R40CSXQ05Ti8080GShU
00@0ITL[06e;;`1]Bbl0LE0a0P1eDS80101hES<0O5Pd07aH=01nES82081H=00408EJ>025FSX0QEXj
08EH>@80QePl00@0Rea008iO@`2>Gd<0Tf130P2EHTD0102IITT0W6U<09aYC02PKE020:9_DP<0Y75D
00@0ZWMJ0:aiG02/NE`0[W]N0P2]NUd30:]hF`040:aiG02]NUd0[GYM0:mmG080[WaK00@0[G]J0:aj
F@2/NUT0[WaK0P2`OUd0102dPV40]XAS0;J4H`2cPF020;J4H`040;F3HP2aOeh0/GmN0;>1H080^HMV
00@0^hUX0;R6I@2hQVD0]89O0P2bP5d00`2^O5T0ZgUF0:]iEP030:YhE@030:]iEP2/NUL0[7YG00<0
[7aH00@0[GeI0;61G@2aPEd0]HEQ0P2gQf<00`2eQF40/h=O0;>3G`030;J:I`040;B8I@2cQF<0/hES
0;:4HP80]hIU00@0_8]Z0<6@K`31T6l0`i5`0P31T6l0102mSfd0^Xi[0;Z>J`2iSV/20;N?J`040;NA
K02hTVd0^99]0;V=J080]h]V00@0]8QS0;>7HP2cQf80]8QS0P2iSFP0102nTVd0`IE`0<6EL02gSFL2
0;B8H`040;24G`2_PEd0[h5M0:n1G@80/81L00@0[7aH0:UiE@2YNED0ZGE@0`2XM4l20:YfD@040:]g
DP2YME00ZGE@0:EaC080XVi900@0YW9=0:YfC`2ZMTl0[GUB0P2[MTh0102SKTH0WVLn09iW?P2NIch2
0:9X@0040:E[A@2QId40XFM109mX@@80Y6e600@0YG1809iZ@P2NJT80V6@l0P2GHc/309AN>P0309IP
?02IHcl0VF<o00<0VfE100@0V68n09EO>`2EGc/0UV0j0P2DH3T0102KJdL0VFa;09U/B`28Gcl208]X
BP0407=D=`1XBBd0J4T]06ID@080@2dL00@0=b@C044`8011<200Cch^0P12<R<0100a91@0D48e0512
=@1]HT`203`a7@040515<@1=@S00CD8`040e8`80DDDe00@0JUi>0596>01BASP0IUiC0P1KEDD01014
@Bh0?Ch/03dn;01>DDH205AJFP0304U?F`14Beh0A4]N00<0@TED00@0@dIE04A8E014B5@0ADUE0P15
BE@00`14B5<0A4QA04A8D@0404U=F0<0B4mI0`18DEX204MAFP<0B4mI00@0BE1J04YAF`1:DE/0Be9L
0P1<Ded304eDGP<0C5=M0`1;DU`304YAF`<0BE1J0`18CeT01017CUP0ATeG04I=E`15C5H204A;E@l0
@dYD3018C5L304M;EP0404I:E@15BE@0ADUD04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU430455
D0H0@dMB0`14B5<304E9E0`0A4QC00@0ADUD04I:E@16BUD0Ad]F0P18C5L304U=F0<0B4aG1014C4l3
04A;D@<0A4]C0P14BeD01@1;BeL0Bd]E04];E@1;C540Bda@0040oooo000<04UAG00404]CGP1:DUd0
BU9M04UAG0<0B51K0P19DE`0101:DUd0Be=N04]CGP1;Deh204YBG@0404UAG018D5/0B51K04Q@F`80
BE5L00<0BU9M04]CGP1;Deh0301:DUd404MDGA@0B5EN00@0BE=L04eFG@1=EUd0BU=J0P0n@dL0100Z
;S4071dO01`M7`0I6AT201XJ6P<06aHB100L5a<501dH500401`G4`0P6aL081/G01PE40806aPA00@0
6a/C020R5`0P8QL0=SP]0P0m?c801013AcT0?Cla03do<@10?Rl2044l:@0404<h901B@b`0DT</05a:
<P80DclV00@0FSXQ05lo9P1O?bH0ITL[0P1^C300101bCc40ME8b07EB<P1iEC<207aH=00307eE<@20
F3@0P5Pd00<0QEXh00@0QUTj08QI>`28FC/0Rea00P2>Gd<409ISAP80V6E800@0VfQ;0:1]D02PKE00
Y75D0P2WM5L0102[N5/0[GYM0:ejG@2^Neh20:ejG@040:]hF`2YMUT0ZGII0:YgFP80[7UL00@0[geL
0:ilF`2^O5/0[G]J0P2/NUT0102]NeX0[geL0:mmG02bP5l20;B2H@040;5oGP2ePf80]H=R0;F3HP<0
/GmN0P2fQ6<0102hQVD0]XAS0;J4H`2cPEh20;5oG0030:ilF@2[NEH0ZgUF00<0ZWQE0`2[NEH0102_
OEX0[G]H0:ekF02[NEH20:ekF0030;:0G@2fQ640]XAQ00<0^8IS0`2eRFH0102eQfD0/hES0;>5H`2c
QF<20;N6I@040;b;JP31T6l0`I1_0;n>K@80_Xe/00@0_8i/0;Z>J`2jSV/0^Xm/0P2hT6`0102iTFd0
^I=^0;VCKP2kSfX20;Z>J@030;R<I`2fRVD0]XYU00<0]XYU00@0^8aW0;V=J02iSFP0/hMR0P2aQF00
102_Peh0[h5M0:n1G@2]Oe/20:inFP040:ajE`2ZN5D0ZWQE0:ahD`80ZWIA0`2YME00102ZMU40ZGE@
0:UeD02VLTd20:=_BP040:A`B`2WLd`0Yg=<0:McC080YW5900@0Xfa50:5Z@@2QJT40Y6Y20P2WKDD0
102RJ480WfDo09mU?`2MISl20:9[A0030:9]A@2KIcl0VfLo00<0UF4i00@0Tedi09AN>P2DGSX0UElk
0P2HHSh309aV@P0409US?`2FH3`0UV0l09US?@80UV8m00@0VVY609U/B`2IK4/0R6520P2;J4X0101d
ECT0IdT]06M9;@1NC3P204Pe9004040_7@19>2P0BCPX04lo;`80@c<T00@0?2lO05Q:?@1HBSd0IEY4
0P13>2@0101HCCT0DDHd0556=012=bD204i2<P0405mC@`1:?S00BSh`04i6>@80B48b00@0>SLT03Pi
9`0h>BL0A4Pj0P14BTX00`0m@dl0?TEH03i5F00404=6E@80@dMC00@0A4QD04E9E015BE@0ADUD0P16
BU<404U=F0<0B4mI0`18DEX204MAFP<0B4mI00@0BE1J04YAF`1:DE/0Be9L0P1<Ded304eDGP<0C5=M
0`1;DU`304YAF`<0BE1J0`18CeT01017CUP0ATeG04I=E`15C5H204A;E@l0@dYD3018C5L304M;EP04
04I:E@15BE@0ADUD04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU430455D0030496D@13Ae80@dMB
00<0@dMB0`14B5<@04E9E080ATYE00@0Ad]F04Q<E`18C5L0BDeH0`18C5L204M;EP@0ADe@0`15C583
04E<E080ADaF00D0Bd]G04];E@1;BeD0BdaA04]<D0010?ooo`00301:DUd0101;Deh0BU9M04YBG@19
DE`304Q@F`80BE5L00@0BU9M04]CGP1;Deh0Be=N0P1:DUd01019DE`0B51K04Q@F`18D5/204UAG003
04YBG@1;Deh0Be=N00`0BU9M0`16De`304MDG@0404QEGP19EUl0BEIO04UFG`80BUMP3018EEh0101:
E5d0CEIM04eFG@19DUT203i3A`0302X^<@0K71h06a`N00<05aLG0`0K5Q8401`G4`D07APD00@06aHB
01lJ5P0O6QH06AHA0P0K6140100M7AD08R@I028T6@0f>2d20494=`0403m3=@0X:Q`0:2XL02h/7@80
?cXW00@0@cPT0511:P1@@BX0FdTa0P1A?B@0101J>R40H40W06109`1YAb`206i<<004079?<@1fDS80
MU8b07UE<`80O5Pd00@0OEDa081H=020F3@0Q5Tf0P26FcT01027FST0RU/m08YK?@2<GCl208iO@`03
09QUAP2FHd@0UV=400<0U65200@0UfA509eZB`2MJT/0Y75B0P2YMUL0102[N5T0[7UJ0:aiFP2]NU/2
0:aiFP040:UfE`2XMEH0Z7EF0:UfE`80ZgQI00@0/7iM0:mmG02_OE`0[G]J0`2/NUT20:ekFP040:mm
G02`OUd0/7iM0:ilF`80]89Q00@0]XAS0;5oGP2aOeh0[WaK0P2aOeh30;B2H@040;:0G@2`OU/0/7iK
0:ekF080ZgUF0`2YMe@0102ZN5D0ZgUF0:]iEP2_OEX20:ajE`040:QfD`2YMe@0ZGMD0:ekF080]89O
00<0^XQU0;f;J02mRfP00`2hQfH0102gQVD0]XET0;J5I02dPf820;J5I0030;^:J@2oSVd0_hi]00@0
_8i/0P2kSf`0102jSf`0^Hi[0;V>J`2fSVX20;F=J@040;Z>J@2iSFP0^HeX0;V=J080^8aW00<0]h]V
0;J:I@2fRVD00`2dR6<00`2/P5/0[X9M0:j2G@030;22GP040:moF`2]NeP0[G]H0:]iEP80ZWQE00@0
/7aG0:eiE02]NE@0ZWIA0P2XM4l0102YME00Z7A?0:QdC`2WLdh20:EaC0040:McCP2VLT/0YW9;0:A^
B080XVe500@0XFY30:=Y@@2SJD40YF]30P2WKDD0102PIT00W68l09aR?02KI3d20:1Y@P0309mZ@P2G
Hc/0Uf<k00<0TUhf0`2CGCT0102DGSX0V68n09QR?P2LIT8209eW@`0409]U@@2IHcl0VF<o09aX@@80
UV@o00@0UfU509Q]B`2HKD/0Qf=30P29Id/0101cECT0I4T^06A9;P1E@bl204dl:P0404Dd8P1>?Rh0
CSh^04hn;P80?C0P00@0AcXZ05e?@P1MCd80GE8l0P18?BT0101DBCD0BCh/04Tn;014>BL205Q<?004
06EIB@1GBcd0Ed/m0559?080A3h^00@0=c@Q03Xl9`0j?2L0ADTk0P12B4H0100m@dl0@DQK0458F`15
B5L204A7EP0404=6E@12AU80@TIB04=7D`80A4QC00<0ATYE04M;E017Be@01019CEP304Q?F@<0B55J
0P17DEX304Q?F@0404U@FP1:DE/0BU5K04]BG080C5=M0`1=E5h304aCG@<0Be9L0`1:DE/304U@FP<0
B4mI00@0AdiH04I=E`16CEL0ADaF0P14BeD304=:E0`0A4]E3018C5L304M;EP0404I:E@15BE@0ADUD
04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU430455D0@0@TIA0`13Ae8504A8D`l0ADUD00<0Ad]F
04Q<E`18C5L00`18C5L304M;EP@0ATiA0`16CE<304I=E@80ATeG00D0Bd]G04];E@1;BeD0BdaA04]<
D0010?ooo`00301:DUd0101;Deh0BU9M04YBG@19DE`304Q@F`80BE5L00@0BU9M04]CGP1;Deh0Be=N
0P1:DUd01019DE`0B51K04Q@F`18D5/204UAG00304YBG@1;Deh0Be=N00`0BU9M0`16De`304MDG@04
04QEGP19EUl0BEIO04UFG`80BUMP3018EEh0101:E5d0CEIM04eFG@19DUT203i3A`0302X^<@0K71h0
6a`N00<05aLG0`0K5Q8401`G4`D07APD00@06aHB01lJ5P0O6QH06AHA0P0K6140100M7AD08R@I028T
6@0f>2d20494=`0403m3=@0X:Q`0:2XL02h/7@80?cXW00@0@cPT0511:P1@@BX0FdTa0P1A?B@0101J
>R40H40W06109`1YAb`206i<<004079?<@1fDS80MU8b07UE<`80O5Pd00@0OEDa081H=020F3@0Q5Tf
0P26FcT01027FST0RU/m08YK?@2<GCl208iO@`0309QUAP2FHd@0UV=400<0U65200@0UfA509eZB`2M
JT/0Y75B0P2YMUL0102[N5T0[7UJ0:aiFP2]NU/20:aiFP040:UfE`2XMEH0Z7EF0:UfE`80ZgQI00@0
/7iM0:mmG02_OE`0[G]J0`2/NUT20:ekFP040:mmG02`OUd0/7iM0:ilF`80]89Q00@0]XAS0;5oGP2a
Oeh0[WaK0P2aOeh30;B2H@040;:0G@2`OU/0/7iK0:ekF080ZgUF0`2YMe@0102ZN5D0ZgUF0:]iEP2_
OEX20:ajE`040:QfD`2YMe@0ZGMD0:ekF080]89O00<0^XQU0;f;J02mRfP00`2hQfH0102gQVD0]XET
0;J5I02dPf820;J5I0030;^:J@2oSVd0_hi]00@0_8i/0P2kSf`0102jSf`0^Hi[0;V>J`2fSVX20;F=
J@040;Z>J@2iSFP0^HeX0;V=J080^8aW00<0]h]V0;J:I@2fRVD00`2dR6<00`2/P5/0[X9M0:j2G@03
0;22GP040:moF`2]NeP0[G]H0:]iEP80ZWQE00@0/7aG0:eiE02]NE@0ZWIA0P2XM4l0102YME00Z7A?
0:QdC`2WLdh20:EaC0040:McCP2VLT/0YW9;0:A^B080XVe500@0XFY30:=Y@@2SJD40YF]30P2WKDD0
102PIT00W68l09aR?02KI3d20:1Y@P0309mZ@P2GHc/0Uf<k00<0TUhf0`2CGCT0102DGSX0V68n09QR
?P2LIT8209eW@`0409]U@@2IHcl0VF<o09aX@@80UV@o00@0UfU509Q]B`2HKD/0Qf=30P29Id/0101c
ECT0I4T^06A9;P1E@bl204dl:P0404Dd8P1>?Rh0CSh^04hn;P80?C0P00@0AcXZ05e?@P1MCd80GE8l
0P18?BT0101DBCD0BCh/04Tn;014>BL205Q<?00406EIB@1GBcd0Ed/m0559?080A3h^00@0=c@Q03Xl
9`0j?2L0ADTk0P12B4H0100m@dl0@DQK0458F`15B5L204A7EP0404=6E@12AU80@TIB04=7D`80A4QC
00<0ATYE04M;E017Be@01019CEP304Q?F@<0B55J0P17DEX304Q?F@0404U@FP1:DE/0BU5K04]BG080
C5=M0`1=E5h304aCG@<0Be9L0`1:DE/304U@FP<0B4mI00@0AdiH04I=E`16CEL0ADaF0P14BeD304=:
E0`0A4]E3018C5L304M;EP0404I:E@15BE@0ADUD04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU43
0455D0@0@TIA0`13Ae8504A8D`l0ADUD00<0Ad]F04Q<E`18C5L00`18C5L304M;EP@0ATiA0`16CE<3
04I=E@80ATeG00D0Bd]G04];E@1;BeD0BdaA04]<D0010?ooo`003@1;Deh204YBG@0404UAG018D5/0
B51K04Q@F`80BE5L00@0BU9M04]CGP1;Deh0Be=N0P1:DUd01019DE`0B51K04Q@F`18D5/204UAG003
04YBG@1;Deh0Be=N00`0BU9M0`15DU/01016De`0AeAM04MDG@18EEh204UFG`0304YGH01;F640BeQQ
00`0B5EN00@0BUAM04eFG@1=EUd0BE9I0P0m@TH0100Z;S406a`N01/L7P0E5AD201@D50<06aHB100L
5a<501dH500401XE4@0M61@07APD01PE40806AH?00@07AdE02@V6`0T9Q/0=SP]0P1;CD00101?DdD0
9BLI02DW6@0S8A8203Pc800403/`7018>B80B3TR05Q6;P80D3`S00@0GS`S06=1:01S@BP0JDL/0P1`
CC40101dCc80MU8b07IB<P1kE3<207eG<`0407eF;`1oF340OePa08AI=P80QU/h00@0RE`k08]M?02;
GC`0SEi00P2>Gd400`2FHd@0U65209AQ@P0309=P@@0409MTA@2LJDX0W6U:0:=`D@80YgAE00@0Z7EF
0:UfE`2YMUL0ZgQI0P2ZMeP0102XMEH0ZGIG0:UfE`2[N5T20:ikG0040;1nG@2_OE`0[geL0:ekFPD0
[7YI00@0[G]J0:ilF`2^O5/0[7YI0P2cPF00102fQ6<0/GmN0;5oGP2/NUT20:ekFP040;5oGP2bP5l0
/X1O0:mmFP80[WaI00<0[7YG0:YhE@2ZN5D0102YMe@20:YhE@030:mjF02/MeD0[7ME00<0ZWEC00@0
[WUG0;AoG@2dOed0^hIT0P2oRVP30;^9IP040;Z8I@2fQV80]XIR0;B4H080]HEQ00<0]XQT0;V;I`2i
RfL01P2lT6/0102jSVT0]XaV0;J<IP2cRF<20;:8HP<0^8YV2@2gRFD0102ZO5P0[X1L0:j0G02bQ602
0;B4H0040;20G02]NeP0[G]H0:iiE`80[gYH00@0]GmK0;1jEP2`NUH0ZgEA0`2XLTh20:UcC`040:Qb
CP2WLDd0Yg5=0:]eD@80Z79<00@0YFm90:=/A@2SK4D0XfU30P2RJ400102RIcl0XFHn0:5V?P2OICl2
09eS?@0409aU?P2PJD80X6U209iY@@80UV8j00<0TEde099N=P2BGSH00`2BGST0102CGcX0Uf<n09MS
?P2KId8209eYA00409]W@P2JIT40VVI109][A@80UfM300@0U6Q509I_CP2FKdh0QF940P22HdH0101`
ECX0HdT^06=9;P1FAC4204ln;004040a7P1>?Rh0CSh^04Tl:`80=RTI00@0BCd]05]?@@1KCd40CD8/
0P14>BD01019?RX0?c@R03ld8P14>BL205eA@@04065EA@1GBcd0Ed/m06=LC080BTDb00@0>3HQ03i0
:`0n@2/0Bdm00P19Cdd01013BU@0A4]M04A;G@17BUT204E8E`0404=6E@11AE40@DEA0496DP80A4QC
00<0Ad]F04U=EP19CEH01019CEP304Q?F@<0B55J0P17DEX304Q?F@0404U@FP1:DE/0BU5K04]BG080
C5=M0`1=E5h304aCG@<0Be9L0`1:DE/304U@FP<0B4mI00@0AdiH04I=E`16CEL0ADaF0P14BeD304=:
E0`0A4]E3018C5L304M;EP0404I:E@15BE@0ADUD04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU43
0455D0@0@TIA1@13Ae8304A8D``0ATYE00@0A4QC04E9E015BE@0ATYE0P17BeH304Q<E`<0Ad]F1017
Ce8304M>E0<0AdiF0P17CUP01@1;BeL0Bd]E04];E@1;C540Bda@0040oooo000=04]CGP80BU9M00@0
BE5L04Q@F`18D5/0B51K0P19DE`0101:DUd0Be=N04]CGP1;Deh204YBG@0404UAG018D5/0B51K04Q@
F`80BE5L00<0BU9M04]CGP1;Deh0301:DUd304EBF`0404ICG017E5d0AeAM04QEGP80BEIO00<0BUMP
04]HH@1;F6403018EEh0101:E5d0C5EL04aEG018DEP203e2AP0402/_<P0L7Al071dO01DE5@804a<C
0`0K5Q8401`G4`D07APD00@06aHB01`G4`0L5a<05a@?0P0E4P/0100L71@092HK02@V6`0e=b`2059D
A`0405eQD`0_<B<0;c4S02PV5`80>SDR00@0?34M04Xk901:>b@0FdTa0P1D@2L0101O?B@0ID4Y06E1
:@1[Abd2075<<@0407A?<P1hDC80N54b07]D<`80OELc00@0OEH]07mH<@1oF340PeTc0P27G3T0102;
Gc`0SEln08eO?P2>H3l208eO?P<0TV0o00@0U69109IT@`2FI4<0VFM60P2MJdX0102QKdh0Xg5@0:=a
D02SLE020:EcDP<0YgED00@0Z7IE0:YhE`2ZN5L0[WaK0P2bP5l0102`OUd0[geL0:mmG02^O5/20:ek
FPH0[7YI00@0ZWQG0;>1H02cPF00]XAS0P2aOeh0102/NUT0[G]J0:ekFP2`OUd20;5oGP040:ekF02/
NUL0[7YG0:]iEP80ZWQE1P2YMe@40:]fE080[7ME00@0[WUG0;9mF`2bOE/0]h9P0P2jQF<30;n9I`04
0;f8IP2hQV<0^8IS0;B4H080/h=O00@0/hEQ0;J8I02fR6@0]heW0P2hSVP0102iSfT0^XiY0;Z>J@2i
SFP20;N;IP040;F9I02dR6<0]8QS0;Z<J080^8YV00@0]XQT0;B6HP2dQV80]8IR0P2eQf<0102fR6@0
]hUU0;N9I@2/OUX20;22GP040;J6HP2gQF80]hER0;>1GP80/WeK0`2aO5X0102fOU/0/WYG0;9jE`2/
M5430:UaCP80ZW9?0`2YLDh0102WLDd0Yfm:0:M_BP2VKTT20:Q^B0040:I/AP2UJT80YFY20:9W?`80
Wf@l00@0XFM109mU?`2OICl0WfQ10P2RJd@0102NJD40UV8j09IR>P2BGSH209=O=`@0Telj0P2GHch0
102KId80WFU409eYA02LJ4<209]W@P0409U[A`2EJD@0UFU4095VA080Tfm=00@0PF1107QM@01hGD00
K58g0P1PB2h0101QDSd0D44^0511;P0n;a`204`o;P0404Tl:`0f:QX0=RXJ04/o;`80F4m000@0=b`F
03ld800o=200BSl[0`13>2H205=7=`0404a0<018?2h0B3`^05]DA080AD0]00@0>cTR04A6<014AS00
C5110P18CTX01012BE40@4MG0417E`17BUT204E8E`0404=6E@12AU80@TIB0496DP80ADUD00<0B4aG
04Y>E`1:CUL01019CEP304Q?F@<0B55J0P17DEX304Q?F@0404U@FP1:DE/0BU5K04]BG080C5=M0`1=
E5h304aCG@<0Be9L0`1:DE/304U@FP<0B4mI00@0AdiH04I=E`16CEL0ADaF0P14BeD304=:E0`0ADaF
3018C5L304M;EP0404I:E@15BE@0ADUD04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU460455D0@0
@TIA1@13Ae8<04I:E@<0A4QC00<0ATYE04M;EP17BeH01017BeH204I:E@@0AdmB0`17CU@304M>EP80
AdiH00D0Bd]G04];E@1;BeD0BdaA04]<D0010?ooo`003@1;Deh204YBG@0404UAG018D5/0B51K04Q@
F`80BE5L00@0BU9M04]CGP1;Deh0Be=N0P1:DUd01019DE`0B51K04Q@F`18D5/204UAG00304YBG@1;
Deh0Be=N00`0BU9M0`15DU/01016De`0AeAM04MDG@18EEh204UFG`0304YGH01;F640BeQQ00`0B5EN
00@0BUAM04aEG01<EE`0B55H0P0m@TH0100[;c8071dO01`M7`0E5AD201<C4`<06aHB100L5a<501dH
500401/F4P0L5a<071LC01LD3`805A8;00@071`D02@V6`0T9Q/0=CL/0P1BE4L0101MHE<0;c4S02la
8`0X9QL203Xe8P0403`a7@1:>b@0BS/T05]9<@80E40W00@0GcdT06E1:@1U@BT0JdL]0P1aC340101d
Cc80N54b07QA<P1kE3<207eG<`0407eF;@1oF340OePa08=I<`80Qe`i00@0Rell08eO?P2=Gch0SV0o
0P2=Gch3099P?`0409AR@@2FI4<0UVA309UWAP80WF]:00@0XFm>0:=aD02SLE00Xg5@0P2ULe830:Me
E0040:QfE@2ZN5L0ZWQG0:ilF`80/X1O00@0/7iM0:mmG02_OE`0[WaK0P2]NeX60:ajF@040:YhE`2c
PF00/h5P0;J4H`80/GmN00@0[7YI0:ekFP2]NeX0/7iM0P2aOeh0102]NeP0[7YG0:ajE`2[NEH20:Yh
E@H0ZGMD102[MU@20:agE@040:iiE`2bOE/0/WeK0;N2H080^XES0`2oRFL0102mR6H0^8IS0;R6H`2d
Q6020;>3G`040;>5H@2fR6@0]XQT0;N=I`80^8iX00@0^HmY0;Z>J@2jSVT0^HeX0P2gRfH0102eRF@0
]8QS0;B8H`2jS6P20;R:IP040;J8I02dQV80]8IR0;B6HP80]HMS00@0]XQT0;N9I@2gRFD0[7iJ0P2`
PUh0102fQV80]hER0;N5HP2cPEh20;9mF`<0/GaJ00@0]WiK0;9jE`2bNUL0[7AA0`2YLDh20:YbC`<0
ZG5>00@0Yg5=0:M_BP2WKdX0YVi90P2XKTP0102VK4H0YFY20:EZ@P2RIcl209mT?0040:5W@@2OICl0
WfDo09mX@@80XV]400@0WVU109IR>P2FHSX0TUhf0P2CGcL409=O>P80Uf<n00@0VfM209eYA02MJD@0
W6Q30P2KId80102IJdL0UFU409EYA02AIT@209=_C@04085P@@1hGD00N5e006aB=`80H4P^00@0HE8m
0511;P1@@Bh0?RlL0P1<?bh01019?2/0=RXJ03HZ6P1;?bl205Q?@00403L/5P0o=200?c@P04Xo:`<0
@cPV0P1CAcL0101<@300B3`^04Pl;P1KE4@204E0;@0403/i8P14AS00A4H`04a@@@80B4i:00@0@TUA
0417E`10AeL0AdYI0P15B5L01013AUD0@TIB0496DP12AU8204E9E00304Q<E`1:CUL0BTiG00@0BDeH
0`18CeT304QAFP80Ae5J0`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@<0CEAN0`1<Ded304]BG0<0
BU5K0`19D5X304Q?F@0404M>F016CEL0ATeG04E<EP80A4]E0`13BU@<04E<EP`0B4aG0`17BeH01016
BUD0ADUD04E9E015BE@204A8D`<0@dMB0`14B5<304=7DP<0@TIA1P11AE040496D@D0@dMB3016BUD3
04A8D`0304I:E@17BeH0Ad]F00@0Ad]F0P16BUD404M?DP<0AdiD0`17CUH204M>F00504];E`1;BeD0
Bd]E04]<D@1;C5000@3oool000`0BU9M00@0Be=N04YBG@1:DUd0BE5L0`18D5/204UAG00404YBG@1;
Deh0Be=N04]CGP80BU9M00@0BE5L04Q@F`18D5/0B51K0P19DE`00`1:DUd0Be=N04]CGP0<04YBG@<0
AU=L0`17E5d01018EEh0BEIO04UFG`19EUl204YGH0`0B5EN00@0BUAM04aEG01<EE`0B55H0P0n@dL0
100]<C@07QlQ01hO8@0F5QH201<C4`<06aHB100L5a<801dH500401TF4@0D4@X0514:01`L50808R@I
00@0;c4V059DA`1BE4L0D5A60P0`<R@0100[:AX0=S4N03Ha7P0j;a/204Xk900405Q6;P1?>b80Cc/R
064m9@80IT8Z00@0KDP^079=<P1bCC80MTlb0P1hDC80101lDc<0OUHc07iF<`1mERd207iG;P0408=I
<@27GCL0Qedg08aP?@80Sf4o00@0Sf5008eO?P2=Gch0Sedj0P2BH3d0102EHd00VFM409UWA02MJdP2
09m]BP@0X6i;0P2RL4d0102TLTl0YG=@0:EcD02WME820:YhE@040;1nF`2dPUl0]89O0;1nG@<0[geL
0P2^O5/60:ekFP040:]iF02bP5l0/X1O0;F3HP80/GmN00@0[WaK0;1nG@2`OUd0/X1O0P2aOeh30:]i
EP040:YhE@2YMe@0ZGMD0:UgE0D0Z7IC00@0ZgEC0:afE02/MU@0[WQF0`2_NEL20;1jF0040;=mF`2e
Oed0]GmM0<25H`80_hMT00@0_HMS0;Z6H@2jQV40]H=N0P2cPed0102cQUl0]8QQ0;B8H@2_R5l20;29
H0040;B:HP2eRf<0]H]S0;R<I@D0^HaU00@0_hm[0;f=J@2mSFT0^HUU0P2fQV80102dQ600/h=O0;>3
G`2dQ6020;F5H@040;61G@2cPel0/h=O0;F5H@80]XAQ00<0]X5O0;=nG02cOU`00`2cOE/0102dO5T0
/7QE0;1hE@2[Le030:UaCP80ZW9?0`2YLDh0102RJTL0Y6a70:A/A`2XKTT20:U_B@040:U]B02WK4@0
Yfa40:QZ@`80YfU200@0XVQ20:5W@@2QId40XFY30P2TKDH0102OJT80UV8j09IR>P2BGSH209AP>004
09=Q?02BH3/0TV0k099P>`80UF<n00@0V6I109YX@`2JJ4<0VVQ30P2IId80102FJTD0U6U609AYAP2=
Id<2095^CP0407YK?P1]DcP0KE<h06M?=@80GTP`00@0JU]605E6<`1EAS<0ACPU0P1=@2l01019?R`0
?cHU03lf9@1=A3<205I=?P0403T^6015>RH0ACXV0556<P80Bd0^00@0@SLU0514=01@A3@0Cd<c0P1I
CCl01015?R`0?cXV03lj9P17ABh2055C?@0404iB@@16CDH0ATe604E<E080A4]K00@0ATUH04E8E`15
B5L0A4MF0`14B5@204I:E@0304Q<E`19CEH0BDeF00@0BDeH0`18CeT304QAFP80Ae5J0`18CeT01019
D5X0BU5K04YAF`1;DU`204aCG@<0CEAN0`1<Ded304]BG0<0BU5K0`19D5X304Q?F@0404M>F016CEL0
ATeG04E<EP80A4]E0`13BU@<04E<EP`0B4aG0`17BeH01016BUD0ADUD04E9E015BE@204A8D`<0@dMB
0`14B5<304=7DP<0@TIA1`11AE030496D@D0@dMB3015BE@01013Ae80A4QC04A8D`15BE@204I:E@<0
Ad]F0`16BUD404I>D@<0ATeC0`16CED204I=E`0504];E`1;BeD0Bd]E04]<D@1;C5000@3oool000`0
BE5L00@0Be=N04YBG@1:DUd0BE5L0`18D5/204UAG00404YBG@1;Deh0Be=N04]CGP80BU9M00@0BE5L
04Q@F`18D5/0B51K0P19DE`00`1:DUd0Be=N04]CGP0<04YBG@@0AeAM5018EEh0101:E5d0BeAK04]D
F`17D5L203m4B00402lc=P0Q8R@08B8T01PH60805ADE0`0K5Q8401`G4`D07APD00@07QTE020K5`0P
6aL07AXE0P0H5@h0100N7QH07b4F01lQ5P0T9Q/204U;?P0403i2=00`<R@0<38T02/Y6P80:bHC00@0
<BHB04Lh8@17>240Dd4Y0P1:=Qd0101R?RH0Id4Z06M1:P1^BBl207A<<P0407I?<P1iD380NE0b07aC
<`80OUHc00@0OEH[07iG;P1nEbh0PUP`0P27GCL0102>HSd0TF=1095S@@2?HCl208eO?@04091N>`2B
H3d0TV0m09IT@@80VVQ500<0WF]809m]BP2OKDX00`2QKd`0102PKT/0XW1=0:9`C@2SLDh20:AbC`04
0:IdD@2YMe@0ZGMD0:mmFP80]89O0`2`OUd60:mmG0<0[WaK00@0[G]J0;:0G`2bP5l0]89Q0`2bP5l3
0;F3HP80/X1O1`2YMe@30:QfD`80YgEB00@0[GME0:miE`2_NEL0[gUG0P2^N5H30:egE@040:miE`2b
O5X0/WaJ0;b0GP80_85O00@0_8AQ0;Z4H02jQ600]XAO0P2fQV00102eRF80^8aU0;R<I@2aS6830;6:
H@80/hUQ00<0]HUR0;N:H`2gRV<00`2jRV@0102kRfL0^XYV0;Z:IP2iRFD20;N7H`030;J6HP2dQ600
]8AP00<0/h=O00@0]8AP0;:2GP2bPUh0/X1M0`2dOed20;AnG0040;9lFP2`NUP0/7YH0;=hEP80[gAB
00@0[75?0:Y_C@2ZKdd0Zg1>0P2/LDl0102[L4h0ZVm=0:Y_C@2SJdP20:Q^B@040:Y`B`2ZKTT0ZVi9
0:M[AP80Z6Y300@0ZF]40:a^A`2/KTL0XFM10P2PIT00102QJT<0Xfa50:=/A@2NJD4209AP>004095M
=@2CGcL0Telg09ES?P80U68m00@0Tf4l09ES?P2EHch0V6I10P2JJ4<0102IId80V6I109QV@@2AIT<2
099ZAP0408aUA02?KTd0Sfi=07=F>080Hd/_00@0H4/`05a8;`1LB2l0Ge<m0P1CAS<0101DAc@0BSl]
04Xo;@13>2H204Y1<00404i5=01>AcL0CTLg04m4;P80C44]00@0Bd0/04Tn;019?R`0@CHT0P1EBCT0
101GBc/0IEU;06EIB`11>RP204<n:P0404i<<`1CECd0DeDm04Y>?@80ATe600<0BU5G04Q@G@18D5d0
1015B5L204E9E@0404I:EP17BeH0Ad]F04M;EP80B4aE1019CEP304Q?F@<0B55J0P17DEX304Q?F@04
04U@FP1:DE/0BU5K04]BG080C5=M0`1=E5h304aCG@<0Be9L0`1:DE/304U@FP<0B4mI00@0AdiH04I=
E`16CEL0ADaF0P14BeD304=:E0`0ADaF3018C5L304M;EP0404I:E@15BE@0ADUD04E9E080A4QC0`13
Ae8304A8D`<0@dMB0`12AU430455D0<0@4A?0`11AE040496D@80@dMB3015BE@01013Ae80A4QC04A8
D`15BE@204I:E@0404M;EP16BUD0ATYE04I:E@80ADUD1015CE0304E<DP<0ADaD0P15C5H01@1;BeL0
Bd]E04];E@1;C540Bda@0040oooo000<04UAG00404]CGP1:DUd0BU9M04UAG0<0B51K0P19DE`0101:
DUd0Be=N04]CGP1;Deh204YBG@0404UAG018D5/0B51K04Q@F`80BE5L00<0BU9M04]CGP1;Deh0301:
DUd404MDGA@0B5EN00@0BUAM04]DF`1;E5/0Ae1G0P0oA4P0100_<cH08B8T024R900H61P201DE5@<0
6aHB100L5a<501dH500401hI5@0P6aL081/G01dJ5@8061D>00@07QhF01lQ5P0O8AH092HK0P19Bch0
100n@S@0<38T030b900[:AX202/V4`04034V4P17>240AcPQ05=1:@80BSHM00@0HShV06M1:P1W@BX0
KTT_0P1dC380101fCc80NE0b07U@<P1lDc<207iF<`0407eF:`1nEbh0OUL^089H<080Qedg00@0SV8m
095S@@2AHd40Sf4o0P2=Gcd0102@GS/0TV0m099P?@2FI44209YXA@0309e[B02OKDX0Wfe:00<0XFm<
00@0X6i;0:9`C@2RL4d0Xg5>0P2TLTl0102VM540ZGMD0:UgE02_OEX20;B2G`<0/7iM1P2_OE`30:il
F`040:ekFP2bP5l0/X1O0;B2H@<0/X1O0`2ePf820;:0G`L0ZGMD0`2XMU<20:MeDP040:egE@2_NEL0
[gUG0:miE`80[WQF0`2]MeD0102_NEL0/WaJ0;9lFP2lP5h20;b1G`040;b4H@2jQ600^XAP0;J4G`80
]XIP00@0]HUR0;R<I@2hS6D0/HaR0`2aRV420;>9H@030;F9HP2gRV<0]hYS00<0^XYT00@0^h]W0;Z:
IP2jRVH0^HUU0P2gQf<00`2fQV80]8AP0;B4H0030;>3G`040;B4H02bPUh0/X9N0;:0G@<0]7mM0P2d
OU`0102bO5X0/7YH0;1jF02cN5H20:mdDP040:aaC`2ZKdd0ZVm=0:]`CP80[75?00@0Zg1>0:Y_C@2Z
Kdd0Xf]80P2XKTT0102ZL4/0ZVi90:Y^B@2WJdH20:QZ@`040:U[A02/KTL0[6i70:5W@@80X6I000@0
XFY30:=/A@2SK4D0WVU10P2DH3P0102AGCD0Telg09=O=`2EHch209AR?@0409=Q?02EHch0UF<n09QV
@@80VVQ300@0VFM209QV@@2HIT40TFI30P2BJTH0102<ID@0Sfi=08m^C@1cESP206=;;`04061;<01L
B2l0G4P_05mC?@80DdHc00@0E4Ld04Xo;@1:?bd0@cPV0P1:@C00101>AC@0CTLg04i7=`1?A2h204a1
;@0404]0;019?R`0BCh/044f9080EDTi00@0Ed/k06EIB`1UFD/0@CXX0P13?RX0101>C3<0DeDm05=E
?@1:CSd204I=AP0304YAE`18D5d0B51M00@0ADQG0P15BED01016BUH0Ad]F04M;EP17BeH204Q<E@@0
BDeH0`18CeT304QAFP80Ae5J0`18CeT01019D5X0BU5K04YAF`1;DU`204aCG@<0CEAN0`1<Ded304]B
G0<0BU5K0`19D5X304Q?F@0404M>F016CEL0ATeG04E<EP80A4]E0`13BU@<04E<EP`0B4aG0`17BeH0
1016BUD0ADUD04E9E015BE@204A8D`<0@dMB0`14B5<304=7DP<0@TIA0`11AE030414C`<0@DE@1012
AU4204=7DP`0ADUD00@0@dMB04A8D`14B5<0ADUD0P16BUD01017BeH0ATYE04I:E@16BUD204E9E0@0
ADe@0`15C58304E<E080ADaF00D0Bd]G04];E@1;BeD0BdaA04]<D0010?ooo`003018D5/0101;Deh0
BU9M04YBG@19DE`304Q@F`80BE5L00@0BU9M04]CGP1;Deh0Be=N0P1:DUd01019DE`0B51K04Q@F`18
D5/204UAG00304YBG@1;Deh0Be=N00`0BU9M2@18EEh304MDG@`0B5EN00@0BE=L04]DF`1;E5/0Ae1G
0P0oA4P0100`=3L08b@V02<T9P0J6QX201HF5P<06aHB100L5a<501dH500401lJ5P0R7AT08QdI028O
6P8071TB00@0820H01`N4`0L7Q<071hC0P11@cH0100n@S@0>S`^03Xl;P0b<24202LR3`0402hS3`18
>B80B3TR05E3:`80CSXQ00@0HcdV06Q2:`1X@R/0KTT_0P1dC380101gD3<0NE0b07U@<P1nDc<2081E
<`0407eF:`1nEb`0OUL/089H<080R5hf00@0SV8m095S@@2AHd40T6900P2=Gcd0102BH3d0Tf4n09=Q
?P2FI44209UWA00409aZA`2OKDX0Wfe:0:9`C@80Xg5>00<0XFm<0:=aCP2SLDh00`2TLTl0102ULe00
Z7IC0:QfD`2^O5T20;:0G@030:mmG02`OUd0/7iM00H0/7iM0`2_OE`0102^O5/0/X1O0;:0G`2cPF02
0;:0G`040;B2H@2hQVD0^8IU0;N5I080/X1O1`2YMe@30:QfD`80YgEB00@0/7YH0;5kF@2aNeT0/7YH
0P2]MeD30:YdDP040:egE@2`NUP0/7YH0;UjE`80^WiJ00@0^h5L0;V3G@2iPed0]hIN0P2hR600102h
SF<0_95W0;bAI`2hTfL20;JAI@040;B=HP2cREl0/hUO0;:7G@80/XEL00@0]8AL0;F4G02eQ5`0]89O
0P2fQ640102hQV<0^XQU0;Z8I@2jR6D20;V7I0040;N5HP2fQ640]XAQ0;J4H@80/X1M00@0[geJ0;1k
F@2`NeT0/GaJ0P2cOE/0102`NUP0[gMF0:mgEP2`ME<20:icD@030:]`CP2ZKdd0ZVm=00<0[75?00@0
Zg1>0:U^C02YKT`0ZVm=0P2]Ldh0102^M4l0ZVi90:Y^B@2TJ4<20:AV?`030:UY@`2^KTP0[Vi800<0
WfDo00@0X6U20:9[A02RJd@0W6Lo0P2BGSH0102?Fc<0TUhf099N=P2FIT0209AT?P0409=S?@2DI3h0
U6@n09MW@@<0V6Q20P2GId40102?I440TFU6095YAP29ID<208i]CP0406iC=@1MABT0GDDY05e9<080
FDP^00@0DDD_0513<01@@c00G4ll0P16>bT0100h<1d0CTDd04i5=01<AC<204Ho;`0405iC?@1:?b/0
BSl[040e8@80A3TW00@0@CHT05A8>01DB3P0BCd]0P1;?c40101:@c40AD0Z04E0:P18ARd204E7;`04
03i2<@12BD80@TU204Q?E@80@d]H00@0A4MF04E8E`15B5L0ATUH0`17BeL304M;EP80Ad]D1019CEP3
04Q?F@<0B55J0P17DEX304Q?F@0404U@FP1:DE/0BU5K04]BG080C5=M0`1=E5h304aCG@<0Be9L0`1:
DE/304U@FP<0B4mI00@0AdiH04I=E`16CEL0ADaF0P14BeD304=:E0`0ADaF3018C5L304M;EP0404I:
E@15BE@0ADUD04E9E080A4QC0`13Ae8304A8D`<0@dMB0`12AU430455D0<0@4A?0`11AE060496D@`0
A4QC00@0@dMB04A8D`14B5<0ADUD1P16BUD204E9E0@0A4a?0`14Be4304A;D`80A4]E00D0Bd]G04];
E@1;BeD0BdaA04]<D0010?ooo`00601:DUd304]CGP<0BU9M0`19DE`304Q@F``0BE5L3018EEh304MD
G@0404QEGP19EUl0BEIO04UFG`80B5EN0`17E5d0101;EEh0BU=J04YCFP18DEP20456BP0402lc=P0L
7Al071dO01DE5@805aLG0`0M61@0100L5a<06aHB01/F4P0K5Q8201`G4`<07APD00@07aHG01XA4P0J
4A8091`J0P0M61@0100T8AP08R0D028P500L6Pd204Q9>`0404M3=`1MFdh0GE]>02h/80807A/>00@0
=BlQ04@i9@14>BD0EdL^0P1<>1d0101YARH0KDTY06e9:@1_Bb/207=<:`0407E?:`1jDRl0NU8_07mG
<`80QEXg0`25FC@01026FSD0RE/g08UK=`2;GCT208iN>P0408mO>`2@GS/0T5hk09IT@`80UF=200@0
TV0o095O?P2AGch0TV0o0P2GID@0102MJdX0XFm>0:5_CP2YMeH20:IdD`040:9`C`2OKD`0Wfe<0:1^
C@80Y79A00<0ZWQG0:ilF`2^O5/00`2XNUP00`2ZO5X0ZgeK0:]mF`030:anG0040:]mF`2ZO5X0ZWaJ
0:YnF`80Yg]H00@0Z7aI0:b0G@2/P5d0ZWiK0P2UNEH0102XO5T0/8AQ0;24H@2`OU/20:ekF0040:Yh
E@2XMU<0Z7IC0:IdD@80YgEB00@0Z7IC0:UgE02YMe@0/7QC0P2cNeH0102aNE@0ZW9=0:YbC@2VKTT2
0:YbC@<0[WIA00@0/gQF0;AlF@2dO5T0]gmL0P2fPUd0102ePeh0]HEO0;F5G`2fRF820;V<I@040;J;
H02gS640]haQ0;N<H@80]X]P0`2dREh00`2fRf00^HiS0;V>H`090;R6H@<0]hEP00@0^XES0;J1G`2f
PEl0/GaJ0P2`NeT00`2aO5X0/WeK0;9mF`030;9mF`040;1jEP2^N5@0[WQD0:afDP80ZG=?00<0Yg5=
0:I`C02VL4`00`2VL4`0102YKdL0[gE=0:meC@2]Ld/20:=Y@@040:5W?`2XKTH0Z6i60:Y`B080Y6Y2
00@0QE0V095L<P2AG380WVTo0P2RKD<00`2MJ3h0UF0f09EP=P03091K<@0409AT=P2DHcP0U6<h09ET
>`80UfHn00@0VFM209QV@`2HIT<0U6910P2@Gch01029ITP0QFA508ETA@1mGT4208EWB`0407=G?`1M
@b`0GD</06=:=080F3lY00@0GU8l05e@?@1MD3d0C3l/0`18?B/204Hm;00404Ln;@17@300Ad0`0559
=080DTXe00@0?3@O03`d8@0l=240A3`Y0P0o=RD01014>bX0Ach]04Ln;@12?BT205=>>P0404=1;00o
?RX0?chZ04];?`80BT]600<0BDY>04I9D016BE000`1:CUT01019CEP0Ad]F04M;EP17BeH504I:E@@0
BDeH0`1:CUT804]?FP<0C51K0`1=DE`604iBG@<0CE5L0`1<D5/304]?FP0404a@F`1;CeX0BdmJ04Y>
F@80BDeH00<0B4aG04M;EP17BeH00`16BUD01015BE@0ATYE04I:E@17BeH304Q<E`80Ad]F00<0ATYE
04E9E015BE@03016BUD01017BeH0B4aG04Q<E`19CEP204M;EP0404=7DP11AE00@DE@0496D@80@dMB
0`15BE@01014B5<0@dMB04=7DP12AU420455D0<0@4A?3013Ae8=04A8D`80ADUD00@0ATYE04M;EP17
BeH0B4aG0`17BeH204I:E@@0A4a?0`14Be4304A;D`80A4]E00D0B4QD04U9D`19BE<0BT]@04Y;C`01
0?ooo`00601:DUd304]CGP<0BU9M0`19DE`304Q@F``0BE5L3018EEh304MDG@0404QEGP19EUl0BEIO
04UFG`80B5EN0`17E5d0101;EEh0BU=J04YCFP18DEP20456BP0402lc=P0L7Al071dO01DE5@805aLG
0`0M61@0100L5a<06aHB01/F4P0K5Q8201`G4`<07APD00@07aHG01XA4P0J4A8091`J0P0M61@0100T
8AP08R0D028P500L6Pd204Q9>`0404M3=`1MFdh0GE]>02h/80807A/>00@0=BlQ04@i9@14>BD0EdL^
0P1<>1d0101YARH0KDTY06e9:@1_Bb/207=<:`0407E?:`1jDRl0NU8_07mG<`80QEXg0`25FC@01026
FSD0RE/g08UK=`2;GCT208iN>P0408mO>`2@GS/0T5hk09IT@`80UF=200@0TV0o095O?P2AGch0TV0o
0P2GID@0102MJdX0XFm>0:5_CP2YMeH20:IdD`040:9`C`2OKD`0Wfe<0:1^C@80Y79A00<0ZWQG0:il
F`2^O5/00`2XNUP00`2ZO5X0ZgeK0:]mF`030:anG0040:]mF`2ZO5X0ZWaJ0:YnF`80Yg]H00@0Z7aI
0:b0G@2/P5d0ZWiK0P2UNEH0102XO5T0/8AQ0;24H@2`OU/20:ekF0040:YhE@2XMU<0Z7IC0:IdD@80
YgEB00@0Z7IC0:UgE02YMe@0/7QC0P2cNeH0102aNE@0ZW9=0:YbC@2VKTT20:YbC@<0[WIA00@0/gQF
0;AlF@2dO5T0]gmL0P2fPUd0102ePeh0]HEO0;F5G`2fRF820;V<I@040;J;H02gS640]haQ0;N<H@80
]X]P0`2dREh00`2fRf00^HiS0;V>H`090;R6H@<0]hEP00@0^XES0;J1G`2fPEl0/GaJ0P2`NeT00`2a
O5X0/WeK0;9mF`030;9mF`040;1jEP2^N5@0[WQD0:afDP80ZG=?00<0Yg5=0:I`C02VL4`00`2VL4`0
102YKdL0[gE=0:meC@2]Ld/20:=Y@@040:5W?`2XKTH0Z6i60:Y`B080Y6Y200@0QE0V095L<P2AG380
WVTo0P2RKD<00`2MJ3h0UF0f09EP=P03091K<@0409AT=P2DHcP0U6<h09ET>`80UfHn00@0VFM209QV
@`2HIT<0U6910P2@Gch01029ITP0QFA508ETA@1mGT4208EWB`0407=G?`1M@b`0GD</06=:=080F3lY
00@0GU8l05e@?@1MD3d0C3l/0`18?B/204Hm;00404Ln;@17@300Ad0`0559=080DTXe00@0?3@O03`d
8@0l=240A3`Y0P0o=RD01014>bX0Ach]04Ln;@12?BT205=>>P0404=1;00o?RX0?chZ04];?`80BT]6
00<0BDY>04I9D016BE000`1:CUT01019CEP0Ad]F04M;EP17BeH504I:E@@0BDeH0`1:CUT804]?FP<0
C51K0`1=DE`604iBG@<0CE5L0`1<D5/304]?FP0404a@F`1;CeX0BdmJ04Y>F@80BDeH00<0B4aG04M;
EP17BeH00`16BUD01015BE@0ATYE04I:E@17BeH304Q<E`80Ad]F00<0ATYE04E9E015BE@03016BUD0
1017BeH0B4aG04Q<E`19CEP204M;EP0404=7DP11AE00@DE@0496D@80@dMB0`15BE@01014B5<0@dMB
04=7DP12AU420455D0<0@4A?3013Ae8=04A8D`80ADUD00@0ATYE04M;EP17BeH0B4aG0`17BeH204I:
E@@0A4a?0`14Be4304A;D`80A4]E00D0B4QD04U9D`19BE<0BT]@04Y;C`010?ooo`00601:DUd304]C
GP<0BU9M0`19DE`304Q@F``0BE5L3018EEh304MDG@0404QEGP19EUl0BEIO04UFG`80B5EN0`17E5d0
101;EEh0BeAK04]DF`18DEP20497B`0402lc=P0M7R007AhP01HF5P805aLG0`0M61@0100L5a<06aHB
01/F4P0K5Q8201`G4`<07APD00@091/L01`G500L5a@08ahK0P0M6180100U8AP09B<G02DS5`0R81<2
04e;?P0404]7>`1BD4<0DU1302d[7`807A/?00@0<B/M04Dj:015>RP0ETH_0P1?>b00101YARH0KDTY
06e9:@1`C2`207=<:`0407E?:`1iDBh0NE4^07iF<P80Q5Tf1025FC@208MI=@0408UK=`2=GCT0SEdi
08iN>P80T5hk00@0UF=209=Q@02CHD00TEln0P2@GSd0102BH3l0UfE409MUA02MJdX20:5_CP030:Id
D`2ULe80YG=B00<0Y79A00@0YG=B0:MeE02WME@0ZWQG0P2/NUT30:]mF`H0ZWaJ0`2[OE/0102YOEX0
Z7aI0:QlF@2YOEX20:]oG0040:YnF`2XO5T0Z7aI0:]oG080/8AQ00@0/GmL0:mmFP2_OEX0[7YG0P2Y
Me@30:QfD`<0ZGMD00@0/geI0;EoF`2eOe/0/WaH0P2/MU80102ZM500[gUE0:miE@2bO5P20;5kE`04
0;1hE@2cNeP0/g]H0;AnFP80]81K00@0/X1K0;61F`2aPE/0]8AN0P2dQf00102cQ5X0]8EK0;B5F`2e
QU`20;B5F`040;>4FP2dQE/0]8EK0;J7G@80^HYP00@0]H=N0;J4G`2fQ5l0]XAO0`2gQF030;R6H@80
^HMR00@0_8MU0;R3H@2hPf40]7mM0P2cOU`0102dOed0]H1N0;F0GP2eP5h20;AoG@040;=mF@2bO5P0
/WaH0:miE@80[GMC1P2[ME40102ZL4P0/7I>0;1fCP2^M4`20:E[@`040:AZ@P2YKdL0ZFm70:I/A080
WF<k00@0Qe8X09=N=02CGS@0WfY00P2RKD<0102LIcd0U5le09AO=@2AG38209=N=00409IV>02EI3T0
UF@i09AS>P80UVDm00@0UVI009EU@@2EID40TV500P2?GSh01024HD<0R6I:08QVBP1mGT4207EG?@04
06mC>`1ZD3T0JU0i06A;=@80DcdV00@0Ge<m05e@?@1MD3d0D4<`0P10=B<0100o=280AC`[04Dl:`15
?2/204a5=@0404Lo:P19@B`0BD4/04@l:@80CdLd00@0C4@a03L_700g;a`0@CPW0P1JDD00101>BCD0
GEQ405eHA01<BSD204Y9=`0405=CA`1<CDP0C4e804U:CP80AdYC2@17BeH304Q<E`@0BDeH0`1:CUT8
04]?FP<0C51K0`1=DE`604iBG@<0CE5L0`1<D5/604]?FP0404Y>F@19CEP0BDeH04Q<E`<0Ad]F0P16
BUD01015BE@0ATYE04I:E@17BeH304Q<E`80Ad]F00<0ATYE04E9E015BE@03016BUD01017BeH0B4aG
04Q<E`18C5L204I:E@0404=7DP11AE00@DE@0455D080@TIA0`15BE@01014B5<0@dMB04=7DP12AU42
0455D0<0@4A?3013Ae8=04A8D`80ADUD00@0ATYE04M;EP17BeH0B4aG0`17BeH204I:E@@0A4a?0`14
Be4304A;D`80A4]E00D0B4QD04U9D`19BE<0BT]@04Y;C`010?ooo`007P1:DUdB04UAG0`0B5EN0`17
E5d01018EEh0BEIO04UFG`19EUl204QEGP<0AeAM00@0BeEN04]DF`1;E5/0B55H0P12Ad/0100`=3L0
7QlQ01hO8@0G5aL201PH60<07APD00@071LC01/F4P0K5Q806aHB0P0L5a<301dH500402<N6`0N6AH0
7QTF028M6@806aH@00@08AdD02DQ5P0U8AH09B4E0P1:AcP0101@C400A48f04A2=P0]:al201hL4004
02`V6016>bT0AS/Y0592:`80DShU00@0JTLW06e9:@1]BBT0L4`/0P1dCB`0101eCb/0NE4^07UA;P1m
EC4208=H=@@0Q5Pc0P25Ec<01026F3@0RUXf08YJ=P2=GCT2091N>`0309AR@@2BH3l0TV0o00<0T5hm
00@0Tf5009QVA@2HITD0WF]:0P2PKTd0102RL4l0Xg5@0:=aD02ULe820:MeE0030:UgEP2ZN5L0ZWQG
00<0ZWQG00@0[GmM0:anG02/OU`0ZgeK0P2YNeT0102XNUP0ZG]I0:UkF@2[OE/20:eoG@040:QlF@2[
Oe`0ZgmL0:b0G@<0ZGeJ0P2]PEh30:n3H0040;:0G@2aOe`0/GmL0:ilF@80[7YG0`2ZN5D0102[NEH0
[7YG0:ajE`2aPE/20;:2G0040:inF02XN580Z7QB0:UiD`80[WiH00@0/81J0:inF02^OUP0/7QE0P2a
NeL0102cOET0/GeH0;5mF02`O5L20:ekEP040:inF02`P5X0/81J0;=oE`80]H5I0`2fPUX0102ePET0
]X9J0;J2FP2iQEd20;b8H0`0]hEP00@0_8QS0;R4G`2hQ5l0]H5L0P2cOeX0102dP5/0]H5L0;F1G02e
PE`20;B0F`040;9lF02aNeL0/G]G0;1jEP80[WQD00<0[GMC0:afDP2/MU800`2/MU80102/LTX0/7I>
0;1fCP2]Ld/20:M]A@040:I/A02YKdL0ZFm70:9X@080UU`d00@0ReH/09MR>02GHSP0XVe30P2SKT@0
102JIC/0Tehd09=N=02CGS@209IQ=`0409IU>P2DHcP0U6<h09=R>@80Tf<k00@0U6A0099T@P2BI480
T6910P2=Gch0101oGD40R6I:08QVBP1nGd<206A6;00406]A>01gGdL0Mem705m9<P80ET0[00@0G50j
05=6<`1CAS<0DD@a0P0k<1h0100o=280BT4`04Y1<014>bX204e6=P0404m7=01=AC80CDDb04Lo;080
CTHc00@0C4@a03/c800k<b00@c/X0P1ECCX01013?RX0CdXf04m:=P11?R/204M6=00405=CA`18BD@0
B4U404E6B`80B4]D00@0A4QC04E9E015BE@0ATYE0P17BeH304Q<E`L0BDeH0`1:CUT804]?FP<0C51K
0`1=DE`604iBG@<0CE5L0`1<D5/604]?FP0304Y>F@19CEP0BDeH00<0B4aG0`17BeH00`15BE@0ATYE
04I:E@0604M;EP0304I:E@15BE@0ADUD00`0Ad]F00@0ATYE04M;EP17BeH0B4aG0P16BUD01012AU40
@4A?0414C`10A4l20496D@0404E9E014B5<0A4QC04A8D`80@dMB00@0@TIA0455D011AE00@DE@0P10
A4l<04=7DPd0A4QC0P15BE@01016BUD0Ad]F04M;EP18C5L304M;EP80ATYE1014C4l304A;D@<0A4]C
0P14BeD01@18B5@0BDUC04U9D`1:Be00BT]?0040oooo000N04YBGA80BE5L3018EEh304MDG@0404QE
GP19EUl0BEIO04UFG`80B5EN0`17E5d0101;EEh0BeAK04]DF`18DEP20497B`04030d=`0N7b407QlQ
01LG5`8061PH0`0M61@0100L5a<06aHB01/F4P0K5Q8201`G4`<07APD00@08ahK01hI5P0N6AH08QdI
0P0K5Q00100Q7A@09B4F02DQ5P0U8AD204Y7>004051<@014@SH0A48f02d[7`807Q`@00@0;2HH04Hk
:@16>bT0DT8[0P1B?RD0101ZAbL0KDTY06e9:@1`C2`207A=;00407E?:`1iDBh0NE4^07eE<@80PePe
1024F3<208EG<`0408IH=02:FSH0RUXf08eM>@80T5hk00<0U691099P?`2BH3l00`2@GSd0102CHD00
V6I509QVA@2MJdX20:1^C@040:9`C`2SLE00Xg5@0:EcDP80YgED00<0ZGMF0:YhE`2ZN5L00`2ZN5L0
102]Oed0[7iL0:anG02[OE/20:UkF@040:QjF02YNeT0ZG]I0:]mF`80[GmM00@0Z7aI0:]oG02[Oe`0
[81M0`2YOEX20:f1GP<0[h=P00@0/X1M0;5oG02aOe`0[WaI0P2/NUL30:YhE@040:]iEP2/NUL0[7YG
0;61F`80/X9L00@0[WiH0:QhDP2XN580ZGUC0P2^OUP0102`P5X0[WiH0:inF02`N5D20;5kE`040;=m
F@2aOEP0/GeH0;1lE`80[G]F00@0[WiH0;20FP2`P5X0/gmG0P2ePET30;J2FP040;F1F@2fPUX0]X9J
0;V5G@80_8QP302gQF00102lR6<0^8AO0;R4G`2ePE`20;=oFP040;B0F`2ePE`0]H5L0;F1G080]81K
00@0/WaH0;5kE`2aNeL0/7YF0P2^N5@00`2]Me<0[7IB0:afDP030:afDP040:abBP2`MTh0/7I>0:ec
B`80Yfe500@0YVa40:U_A`2YKdL0XVQ00P2FG3@0102;ER`0Uf8h09MR>02RKD<20:=^A00409YU>`2C
GS@0Tehd09=N=080UV4g00@0UVDj09AS>02DHcP0Tf8i0P2CHc/0102DI400TVA2099T@P2@HT4208eO
?P0407mM@@28ITX0R6I:07iO@`80I4H/00@0Je4h07MOA`1gGdL0GdTb0P1F@2/0101LD3X0DdHc05=6
<`1AA34203/`7P0403ld8P1:@C00BT4`04@k:P80CDHf00@0CdLd04e5<P1=AC80Acl/0P1>AS<0101<
A340>c<P03/c8013>bP205E=>P0404<n:P1?BSH0CdXf044n:`80AdHd00@0De=704Q9A018BD@0ADI;
0P18Be@01014B5<0ADUD04E9E016BUD204M;EP<0B4aG1`19CEP304Y>F@P0BdmJ0`1<D5/304eAG0H0
CU9M0`1=DE`304a@F`H0BdmJ00<0BTiI04U=F019CEP00`18C5L304M;EP0304E9E016BUD0ATYE00H0
Ad]F00<0ATYE04E9E015BE@03017BeH01016BUD0Ad]F04M;EP18C5L204I:E@040496D@10A4l0@4A?
0414C`80@TIA00@0ADUD04A8D`14B5<0A4QC0P13Ae801012AU40@DE@0455D011AE020414C``0@dMB
3@14B5<204E9E00404I:E@17BeH0Ad]F04Q<E`<0Ad]F0P16BUD404A<C`<0A4]A0`14Be<204A;E@05
04Q8E019BE<0BDUC04Y;D01:Bdl00@3oool001h0BU9M4P19DE`<04QEGP<0AeAM00@0B5EN04UFG`19
EUl0BEIO0P18EEh304MDG@0404]EGP1;E5/0BeAK04QAF080@TM;00@0<3@g01hO8@0N7b405aLG0P0I
6AT301dH500401`G4`0K5Q806aHB01/F4P8071LC0`0M61@0100O6aP071TD01`I500N6aH201TE3004
01hJ3`0Q7A408AdA02<M3`80?SPZ00@0EE5603Tg:`0i=b/0;b`S0P0Q7QD0100W8AD0A3PX04@h:01=
?2P205<o9P0406Y79`1^BRX0KTXZ075=;@80M4d/00@0MDl[07Q@;@1hD2d0Ne<_0P20EC8308=G<P04
08AH<`24ES80Q5Hb08AF<P80R5Pd00@0S5`h091N>`2@GS/0TV0o0P2@GSd0102>G3/0Sedl08mM?02D
HT4209YXA`0409i/B`2PKTd0X6i=0:5_CP80XW1?00@0Xg5@0:EcDP2ULe80YgED0P2XMUD0102YMeH0
ZWQG0:YhE`2YNeT20:YlFP<0ZgeK00@0ZWaJ0:]mF`2[OE/0[GmM0P2^P5h0102XO5T0[H5N0:f1GP2]
PEh30:UmFP80/8AQ00@0/XIS0:n3H02_Pf00/GmL0P2`OU/0102^O5T0[G]H0:ekF02]NeP20:ilF@03
0:mmFP2`OU/0/7iK00<0[h5M00@0ZgeI0:EgD`2UMe<0YGMC0P2ZO5P0102[OET0Z7YF0:QjEP2^NUD2
0;1lE`040;5mF02_NeH0[g]F0:eiE080[7QC00@0ZWQC0:ajE@2/NUD0^H9K0P2jPe`0102kQ5d0^X=L
0;Z3G02iPU/20;Z3G0030;b5GP2oR640_hQQ00<0_8YU00@0^hUT0;V7HP2iQf80^8IQ0P2gQF00102f
Q5l0]H=N0;F3GP2iQF020;J2G@040;=oFP2bOUT0/WiI0;=oFP80]81K00<0/gmJ0;9nF@2bOUT00`2^
N5@0102/MU80ZgEA0:]eD@2ZM5030:UcC`80Z79>00@0[WA<0:ecB`2]Ld/0ZFm70P2TJT80102VK4@0
Z6i60:Q^AP2RJ40209MM=@0408mJ<02JIC/0VVDk0:A_A@80Xfi400@0VF@j099M<`2BGC<0U5le0P2I
I3X0102EHSL0U60h09AP>02AH3P2095Q>`04095S?`2BIT<0TVI308mT@P80SV=300@0P5i2081QA@20
HDD0O5i20P1PA2`0101^E3d0N5m907QOB@1F@2/205m:=@0405A8<P10<b00@3<P04e0;@80?c@R00@0
BCh/0599>01BBCP0AC`[0P1>AcL0101UG4/0F4ln05Q??P10=bH203P`7@0404<k:01<A2l0C4@_04Dm
:080=RhI00@0=c8O044l:@11?2T0=C8O0P11?c00101?CdD0@d=104=3@@11@TL204M9E@<0A4QC00<0
ATYE04M;EP17BeH01P18C5L404U=F0<0BTiI201;CeX304a@F`<0CE5L1P1>DUd304eAG0<0C51K0`1;
CeX404Y>F@80BDeH0`18C5L304M;EP@0ATYE0`17BeH504I:E@`0Ad]F00@0ATYE04M;EP17BeH0Ad]F
0P15BE@01012AU40@4A?0414C`10A4l20455D0<0A4QC0`13Ae830496D@<0@DE@3013Ae8=04A8D`80
ADUD00@0ATYE04M;EP17BeH0B4aG0`17BeH204I:E@@0@d]>0`13BU0304=:DP80@dYD00D0B4QD04U9
D`19BE<0BT]@04Y;C`010?ooo`00601:DUd604UAG0H0BU9M3018D5/<04QEGP<0AeAM00@0B5EN04UF
G`19EUl0BEIO0P18EEh304MDG@0404UCG01:DeX0BU=J04M@E`80@TM;00@0<3@g01hO8@0N7b405aLG
0P0I6AT301dH500401`G4`0K5Q806aHB01/F4P8071LC0`0M61@301`K5P04024N5`0L6A0071T@020L
4@809AlA00@09Ad@03</700c;1`0EE560P0f<bX0100c<2T0924H02@Q600T7Q82040d900404Xi9@1C
?bL0DclW06Y79`80KTXZ00@0LDd]07A=;01dCB`0MDl[0P1gCb`0101iDBd0OU<`07iC<021EC0208=G
<P0408AH<`25Ec<0QELc08AF<P80QeLc00@0Re/g091N>`2@GS/0TV0o0P2?GC`0102=FcX0T5hm091N
?@2FI4<209aZB@0409m]C02PKTd0X6i=0:=aD080XW1?0`2PKTd0102QKdh0Y79A0:AbD@2XMUD20:]i
F0040:=eD`2XNUP0Z7YH0:eoG@80[h5O00@0[X1N0:eoG@2]Oed0[X1N0P2_PEl0102ZOU/0[X9O0:j2
G`2]PEh20:UmFP040:YnF`2aQF80/HER0;:6H`80[X9O102^O5T20:mmFP040;1nF`2bP5d0/X1M0;B2
G`80]H=P00@0/XIQ0;>7HP2cQf80[X9M0P2WNeH0102UNE@0ZGeH0:UmF02ZOUT20:MkEP040:ekEP2_
OEP0[geH0;9nF@80/7aG00@0[GUD0:ahD`2/N5<0[7QC0P2^NUD0102lQEh0_HIO0;f6G`2mQUl20;^4
G@040;R1FP2gP5T0]h1I0;V2F`80^hAM102kRF@60;Z8H`80^HMR00@0^XIO0;R4G@2hQ5d0]H5J0P2d
P5T0102ePEX0]X9K0;J2F`2dP5T20;9nE`040:ihE02]Me<0[GMC0:egD`80ZgEA00@0ZWA@0:UcC`2Y
Ldl0Yg5=0P2VL4`0102^M4`0ZW180:Y`B02TJT820:5W?`040:AZ@P2WKDD0Yfe50:AZ@P80WF<k00@0
TUdc09aW?@2LIcd0YG160P2SKT@0102JIC/0Tehd09=N=02DGcD209UT>P04095M=@2AGCH0TEdf08mN
=P80TF4m00@0TVI309=XAP2CJ4H0T6M70P2=ITD01023I4P0MELk07EG>`1jG48206Y@=`04075I@@1/
ET40K5I104`g8P80Ie8o00@0C40Z030S400`8a00CT4^0P18?B/0101;@2h0C4<b04a3<P1=A3<205aE
A@0406=JB`1@AcP0D4Lh03/b8@80<RXG00@0?cLT0518<`1@B3<0A3`U0P0/90d01010>RX0BD<c04U3
<`0m>RT204I4=@0405=CB@15ADD0ADE50451B@80AdUE00@0A4QC04E9E015BE@0ATYE0P17BeH01018
C5L0Ad]F04M;EP16BUD204E9E0@0BDeH0`1:CUT804]?FP<0C51K0`1=DE`604iBG@<0CE5L0`1<D5/3
04]?FP0304Y>F@19CEP0BDeH00@0BDeH1@18C5L00`17BeH0ATYE04I:E@0704I:E@80Ad]F3018C5L0
1016BUD0B4aG04Q<E`18C5L204I:E@040496D@10A4l0@4A?0455D080@TIA1P13Ae860496D@`0@dMB
3@14B5<204E9E00404I:E@17BeH0Ad]F04Q<E`<0Ad]F0P16BUD404=;CP<0@dY@0`13BU8204=:E005
04Q8E019BE<0BDUC04Y;D01:Bdl00@3oool001P0BU9M1P19DE`604YBG@`0B51K3018EEh304MDG@04
04QEGP19EUl0BEIO04UFG`80B5EN0`17E5d01019De`0BU=J04YCFP17D5L20497B`04030d=`0N7b40
7QlQ01LG5`806ATI0`0M61@0100L5a<06aHB01/F4P0K5Q8201`G4`<07APD0`0L6aH0100Q7QL071T@
01`I400P714202DO4@0402DM400c;1`0<b`L05EAAP80=S<Z00@0<c0Y02@Q600T8AP091hB0P10=2@0
101:>BD0DclW05<o9`1ZAbL206i::P04075=;@1dCB`0M4d/07E?:`80Mdl/00@0NE4]07iC<01nDc00
PED`0P23Ec801024F3<0QELc08EG<`24ES8208MG<`0408]K=`2@GS/0T5hk099P?`80Sedl00@0SE/j
091N?@2@GSd0UVA30P2LJTT0102OKD`0X6i=0:1^C@2SLE020:9`C`<0X6i=00@0XFm>0:AbD@2TLU40
Z7IE0P2[NEP0102SME<0Z7YH0:QjF02]Oed20:n1G`040:j0GP2]Oed0[GmM0:j0GP80[h5O00@0ZWiK
0:j2G`2^PUl0[H5N0P2YOEX0102ZOU/0/HER0;65HP2bQV<20:j2G`@0[WaI0P2_OEX0102`OU/0/X1M
0;:0G@2dPUl20;F3H0040;:6H@2cQf80/hMR0:j2G@80Yg]F00@0YGUD0:UmF02YOEP0ZWiI0P2WNeH0
102]NeH0[geH0:mmF02bOUT20;1lE`040:eiE02/N5<0[7QC0:ahD`80[WYE00@0_8EN0;f6G`2mQUl0
_HIO0P2kQ5d0102hPEX0]h1I0;N0F@2iPU/20;^4G@@0^hUT1P2jR6<20;V7HP040;Z6G`2hQ5d0^8AM
0;F1FP80]81I00@0]H5J0;J2F`2fPU/0]81I0P2bOUL0102^N5@0[GMC0:egD`2]Me<20:]eD@040:Yd
D02YLdl0ZG=?0:MaC@80YW1<00@0[WA<0:Y`B02ZL4P0Y6Y20P2QIcl0102TJT80Yfe50:M]A@2TJT82
09eS>`04099M<`2LIcd0W6Lm0:E`AP80Xfi400@0VVDk09=N=02CGS@0U5le0P2II3X0102AGCD0TEdf
095M=P2?GSH2095Q?@04099V@`2CJ4H0TfQ6091WA`80SFI500@0PfA807EG>`1eEc/0NUa20P1ZD3L0
101aFD40K5I106aF@@1<=b8206MB?`0404a0:P0`8a00<2<@04i1;P80B3d[00@0Bd0^04a3<P1<@c80
CD@c0P1LEDD0101SFT/0D4Lh0517>00k<R42038Z5`0403lg901@B3<0D4Pc04@l9@80;2@=00@0@3XZ
04U3<`19@c<0?CXY0P16A3D0101CDdT0ADE504E5A@11@DT204M9E@0404A8D`15BE@0ADUD04I:E@80
Ad]F00@0B4aG04M;EP17BeH0ATYE0P15BE@404U=F0<0BTiI201;CeX304a@F`<0CE5L1P1>DUd304eA
G0<0C51K0`1;CeX00`1:CUT0BDeH04U=F00404U=F0D0B4aG00<0Ad]F04I:E@16BUD01`16BUD204M;
EP`0B4aG00@0ATYE04Q<E`18C5L0B4aG0P16BUD01012AU40@4A?0414C`11AE020496D@H0@dMB1P12
AU4<04=7DPd0A4QC0P15BE@01016BUD0Ad]F04M;EP18C5L304M;EP80ATYE1013Bdh304=:D0<0@dYB
0P13BU@01@18B5@0BDUC04U9D`1:Be00BT]?0040oooo000H04YBG@H0BE5L1P1:DUd<04Q@F``0B5EN
0`17E5d01018EEh0BEIO04UFG`19EUl204QEGP<0AeAM00@0B59K04QAF018DEP0ATmF0P10ADT0100_
<cH07AhP01dN800G5aL201TI6@<07APD00@071LC01/F4P0K5Q806aHB0P0L5a<301dH500401dN600O
81P07b0H028R60808R0D00@09B4E02dU600]9AP0;BHF0P0b:AX0101<B3l0>3D^03Pe;P0e<R/202HS
700402@M4`0m<B<0?C4S04/j:080E3lZ00@0JTLW06i::P1^BRX0LTh^0P1eCRd0101eCb/0MTh[07I>
:`1hD2`207aA;P0407mC;P23Ec80PeLb08IJ=@80QeTe00@0QELc08MG<`27Ec<0Re/g0P2?GCX0102B
H3l0Sedl08mM?02=FcX2095O?P0309QVA@2OKD`0Wfe<00<0XFm>00@0YgED0:AbD@2TLU40X6i=0P2M
JdX0102NK4/0XW1?0:9`C`2XMUD20:ajF@040:5cD@2WNEL0YgUG0:n1G`80/XAR00@0/89P0:eoG@2]
Oed0[GmM0P2^P5h30:f1GP040:b0G@2ZOU/0ZWiK0:b0G@<0[h=P0P2]PEh30:ekF0040:ilF@2`OU/0
/7iK0;:0G@80]89O00@0]H=P0;J4H@2fQ640]hUW0P2iRfT0102fR6H0[X1N0:j0GP2ZO5X20:eoG@04
0;22H02_PEl0[h5O0:inF080/81J00@0/X1K0;=oFP2cOeX0/GeH0P2`NUH0102bO5P0]7aI0;AlF@2j
Pe`20;^4G@040;Z3G02hPEX0^85J0;AmEP80/gaE00@0]GiG0;N0F@2gP5T0]H=N0P2fQ5l0102gQF00
^8IQ0;R6H@2jR6<20;b:I@<0_H]V00@0_8QQ0;Z6G`2jQUl0^8AM0P2gPe`30;R4G@030;J2F`2dP5T0
]81I00<0/G]G00@0/7YF0:miE@2_NED0[GMC0P2ZM500102XLTh0Yg5=0:MaC@2[LDT20:M]A@040:9X
@02PISh0X6Hn0:9X@080YF]300@0Y6Y20:5W?`2QIcl0UF0f0P2MJ3h0102TKdD0Xfi40:=^A02LIcd2
09EP=P0409AO=@2EH3H0UF0f095K=@80T5`e00@0T5hi095S?`2AHcl0TfQ50P2BJDT0102=JDL0RVM7
08YWA`25Id/206e?=@0407IJ@P1eFd@0ME]4071G@@80HD`g00@0BC@Q05]9=@1KBCD0BShX0P0_8Pl0
101EB3D0Cd@b04m4<P11=R@2040g9P0405aC@P1WH500If1@04To<`80?S@X00@0@STZ04@k:P14>bX0
@c/V0P16?RL01012>b40@CXP044j8012?2`204i8>004040m;016A3L0AT@g059BBP80BDU900@0ATI>
04U;F019BeP0A4QC0P15BE@01016BUD0B4aG04Q<E`18C5L204I:E@0304E9E014B5<0A4QC00@0BDeH
0`1:CUT804]?FP<0C51K0`1=DE`604iBG@<0CE5L0`1<D5/304]?FPX0BDeH0P18C5L304M;EP0404I:
E@15BE@0ADUD04E9E080ATYE0`17BeH<04Q<E`0404M;EP19CEP0BDeH04U=F080Ad]F00@0A4QC0455
D011AE00@TIA2013Ae860496D@`0@dMB3@14B5<204E9E00404I:E@17BeH0Ad]F04Q<E`<0Ad]F0P16
BUD4049:C@<0@TU?0`12BE420499D`0504Q8E019BE<0BDUC04Y;D01:Bdl00@3oool001P0BU9M0`18
D5/304UAG0<0BU9M0`1;Deh<04M?FP`0B5EN0`17E5d01018EEh0BEIO04UFG`19EUl204QEGP<0AeAM
00@0AU1I04I?EP16CeH0A4eD0P0oA4P0100^<SD071dO01`M7`0F5QH201PH60<07APD00@071LC01/F
4P0K5Q806aHB0P0L5a<301dH500401XM500M81L07B0G01lQ5P8081hB00@09R0D034Z6P0a:QX0=RdN
0P0g:a/0100o>c80>cPa03/h<@0f<bh202PU7P0402HO5@0k;b40>blQ050o;@80F4<^00@0JTLW06m;
:`1_Bb/0LTh^0P1eCRd0101dCRX0MDdZ07E=:P1fCRX207Y?;00407eA;022ES40PUHa08QL=`80RU`h
00@0R5Xf08QH=028F3@0RUXf0P2>G3T0102BH3l0Sedl08mM?02=FcX2099P?`0309YXA`2QKdh0XFm>
00<0Xg5@00@0[7YI0:QfE@2XMUD0Xg5@0`2PKTd20:AbD@040:UgEP2/NUT0[7YI0:IhEP80ZgeK0`2a
Pf40102^P5h0ZgeK0:]mF`2[OE/20:j0GP040;65HP2/P5d0[81M0:UmFP80ZgmL00@0[H5N0:b0G@2/
P5d0ZgmL0P2/P5d0102^O5T0[geJ0:mmFP2`OU/20;5oG0040;:0G@2cPEh0/h5N0;>1GP80]89O00@0
]HAS0;^:J@2kRVT0^hYY0P2bPF00102/NeX0[giM0:mnG@2eQ6<20;R7IP040:j1FP2bPU`0/X9L0;B4
GP80]H=N00@0]H5L0;J0G02fP5`0^H5N0P2jPUl0102gPe/0^8AL0;R4G02hQ5`20;J2FP040;=oE`2b
OUH0/WiF0;B0F080]X9J0`2cPE`0102dPUd0]H=N0;F3GP2fQ5l20;N5H0<0^8IQ00<0^8AL0;J2FP2f
PUX00`2dP5P30;F1F@030;=oE`2`O5@0/7aD00<0/G]G00@0/7YF0:ihE02^N5@0ZgEA0P2XLTh0102U
Kd/0Xfe90:=]B@2WKDD20:E[@`030:AZ@P2SJD40XfU100<0XfU100@0XVQ00:5W?`2QIcl0Uf8h0P2L
Icd0102RKD<0Xfi40:=^A02NJCl209MR>00409=N=02@Fc40T5/a099L=P80Tedi00@0Tf4l09EWA@2E
IdD0UV]90P2BJdX0102:IdL0QFA308ET@`22I4X206M<<@0407AH@01eGDD0MEe506E?>080GTTf00@0
C3XV044^7@11;Qd0Cd<]0P0f:AH0101KCS/0E4Tg05A9=`0m<R02040g9P0406YQD01HDD40F55103dc
:@80?S@X00@0D4Lh04i5=01>AC@0@SXU0P16?RL01018@BL0CDHZ04e6:P11>bd2051:?004040m;P11
?c80@Clb04a<A080AdM900@0ADE?04E7E015Ae@0@DE@0P12AU400`15BE@0Ad]F04M;EP0304Q<E`03
04I:E@15BE@0ADUD00@0BDeH0`1:CUT804]?FP<0C51K0`1=DE`604iBG@<0CE5L0`1<D5/304]?FP`0
BDeH00@0B4aG04M;EP17BeH0ATYE0`15BE@204I:E@0304M;EP18C5L0B4aG00d0BDeH0`1:CUT204Q<
E`0404E9E013Ae80@dMB04=7DP80A4QC1P12AU4B04=7DPd0A4QC0P15BE@01016BUD0Ad]F04M;EP18
C5L304M;EP80ATYE1012BTd30499C`<0@TUA0P12BE<01@18B5@0BDUC04U9D`1:Be00BT]?0040oooo
000H04YBG@<0B51K0`19DE`304YBG@<0Be=N3017CeX<04QEGP<0AeAM00@0B5EN04UFG`19EUl0BEIO
0P18EEh304MDG@0404I@F@16CeH0ATmF04A=E080?dA800@0;S8e01`M7`0L7Al05QHF0P0H61P301dH
500401`G4`0K5Q806aHB01/F4P8071LC0`0M61@0100J7A@07B0G01dP5`0O8AH2020N4P0402HP500a
:QX0<BXJ03H]7P80=b/K00@0?c/b03/h<@0k>340=S<^0P0X9Ah0100V7aD0>blQ03/_8@1@?bd205Q3
;P0406Y79`1_Bb/0Kd/[079>;P80MDh]00@0M4hZ07E=:P1eCBX0MThZ0P1jCb`0101mDB`0PUHa089F
<@28G3L208YL>00408QJ=P28F3@0R5Pd08YJ=P80SU`i00@0TV0o08mM?02?GC`0SE/j0P2BH3l00`2J
J4L0XFm>0:5_CP030:=aD0040:ajF@2XMUD0Z7IE0:=aD0<0X6i=0P2TLU40102YMeH0[7YI0:ajF@2V
N5H20:]mF`<0/H=Q00@0[X1N0:]mF`2[OE/0ZgeK0P2^P5h0102aQF80[81M0:b0G@2YOEX20:]oG004
0:f1GP2/P5d0[81M0:]oG080[81M00@0[WaI0:mmFP2_OEX0/7iK0P2aOe`0102bP5d0/h5N0;>1GP2c
PEh20;B2G`040;F4H`2kRVT0^hYY0;^:J@80/X5P00@0[7]J0:mnG@2_OUd0]HAS0P2hQfH0102^PEX0
/X9L0;:2G02dQ5h20;F3GP040;F1G02fP5`0]X1L0;V1GP80^X9O00@0]h=K0;R4G02hQ5`0^8AL0P2f
PUX0102cOeL0/WiF0;9nEP2dP5P20;J2FP<0/h5L00@0]89M0;F3GP2ePeh0]XAO0P2gQF030;R6H@03
0;R4G02fPUX0]X9J00<0]81H0`2ePET00`2cOeL0/7aD0;1lE0030;5kE`040;1jEP2^N5@0[WQD0:]e
D@80Z79>00@0YFm;0:=]B@2SKDT0Yfe50P2UJd<00`2TJT80XfU10:=Y@@030:=Y@@040:9X@02QIcl0
XFLo09MR>080W6Lm00@0XVe30:=^A02SKT@0WVTo0P2GHSP0102CGS@0T5/a091K<@2BG3H209=M>@04
09=Q?02EIdD0UFM509I[B@80TV]:00@0RVM708ET@`25I4<0PVA:0P1WC340101dF400MEe507EMA@1U
CcP205i9=P0404`j9P11;Qd0@BhM04m3;@80=RTF00@0Fdhk05A9=`1DBCL0?C8P0P10=bH0101ZHE00
F55105QA@@0m<bT203hd:0040517>01>AC@0CTDd048j9@80AShW00@0B44W04e6:P1=ARX0@C/]0P1@
BS`01010?Bh0@Clb044o<P1<C4@204M7B@0404E5C`15Ae@0ADMD0455D080@TIA00<0ADUD04M;EP17
BeH00`18C5L00`16BUD0ADUD04E9E00404U=F0<0BTiI201;CeX304a@F`<0CE5L1P1>DUd304eAG0<0
C51K0`1;CeX<04U=F00404Q<E`17BeH0Ad]F04I:E@<0ADUD0P16BUD00`17BeH0B4aG04Q<E`0=04U=
F0<0BTiI0P18C5L01015BE@0@dMB04=7DP13Ae8204A8D`H0@TIA4P13Ae8=04A8D`80ADUD00@0ATYE
04M;EP17BeH0B4aG0`17BeH204I:E@@0@TY=0`12BDl30499D@80@TUC00D0B4QD04U9D`19BE<0BT]@
04Y;C`010?ooo`00601:DUd304Q@F`<0BE5L0`1:DUd304]CGP`0AdmJ3018EEh304MDG@0404QEGP19
EUl0BEIO04UFG`80B5EN0`17E5d01015CeP0ADiE04E>E@13C5<203i3A`0402da=00L7Al071dO01DE
5@8061PH0`0M61@0100L5a<06aHB01/F4P0K5Q8201`G4`<07APD00@051X@01PK4P0H6a806A/@0P0J
60`0100R70h0<2TI030Y6@0k;al203P/700403D`:P0m>S<0?CXc03Hc;P80:2DP00@0:24I03/_8`0k
;b<0E4<a0P1LAc80101[B2P0Kd/[06m;:`1bCRh207E>;@0407A>:P1eCBX0MDdZ07I>:P80NDh[00@0
O50[089F<@22ES40REdh0P2<GSX01029FcL0RETe08UI=@2:FSH208iL>@04099P?`2?GC`0Sedl08iL
>`80Tf5000@0W6Y90:=aD02SLE00YG=B0P2TLU40102^O5/0ZgUH0:]iF02WME@20:AbD@040:=aD02V
M5<0YWAC0:YhE`80[7YI00@0ZgeK0:n1G`2_PEl0/XAR0P2`PV00102[OE/0Z7YH0:QjF02ZO5X20:eo
G@040;>7I02[Oe`0ZgmL0:QlF@80[81M00@0[X9O0:YnF`2ZOU/0Z7aI0P2/P5d<0;5oG0040:ilF`2f
Q6<0]XAS0;V7IP80/GmN00@0ZWQG0:ekFP2]NeX0]XAS0P2kRFP0102`Pe`0/XEN0;:5GP2fQV020;N5
H0040;R4G`2kPf00^h=P0;f5HP80`HIT00@0]hIM0;R7GP2hQeh0^HQO0P2hQeh30;J5G0030;R7GP2j
RF00^XUP00<0]XAO00@0]89M0;>1G02cPE`0/X1K0P2`OUT0102_OEP0[WaG0:ilE`2aOED20:mkD`<0
[GUA0`2^NU80102/N500ZGE=0:UeC@2^N5@20:egD`040:afDP2ZM500ZWA@0:MaC@80Xfe900@0X6Y6
09eW@`2MId<0Y6Y20P2UJd<0102VK4@0Yfe50:M]A@2TJT820:5W?`0409mU?@2NI3`0WV@l09QS>@80
W6Lm00@0XFa20:9]@`2RKD<0WfY00P2II3X0102BGC<0SEP^08eH;P2FGST209IP?00409IT@@2GJDL0
UfU709E]BP80Sf]900@0QfI5089Q@P22HD80OV160P1VBc00101bF3l0LEU1075I@@1LAS42061;>004
058o;P0Y60H0:AP605E9<`80>bhK00@0GE0m05I;>@1FBcT0@SLU0P1:@C00101bJEP0?SLW03hg9`17
?C<204e3>@0405Y@A017?Rd0Ach]03`d7`80DDX`00@0DT/_04Dn8P15?R80CTPj0P1NF4X0101:AcP0
@d4e04=1=@1:BD@204A4AP040410BP0k?D`0>ce<03i2C@80@4A?00<0@dMB04M;EP17BeH00`19CEP0
0`18C5L0Ad]F04M;EP0404U=F0<0BTiI201;CeX304a@F`<0CE5L1P1>DUd304eAG0<0C51K0`1;CeX4
04Q<E`P0BDeH00@0B4aG04M;EP17BeH0ATYE0`15BE@204I:E@0304M;EP18C5L0B4aG00`0BDeH00@0
BTiI04]?FP1;CeX0BdmJ0P19CEP01016BUD0A4QC04A8D`14B5<204E9E0H0@TIA4P13Ae8=04A8D`80
ADUD00@0ATYE04M;EP17BeH0B4aG0`17BeH204I:E@@0@TY=0`12BDl30499D@80@TUC00D0B4QD04U9
D`19BE<0BT]@04Y;C`010?ooo`00301;Deh<04Q@F`l0BE5L00@0B51K04M?FP17CeX0AdmJ0P18D5/3
04UAG0`0B5EN3017E5d01012C5D0A4eD04A=E012Be8203]0A00402da=00P8B<0824S01TI6@805QHF
300N6AD0100H61H061TD01PI500T8ah202LT6`0401PD2@0]9AP0;BDH03df9P80=2/J00@0;BPR03`i
<P0l>C80=C8[0P0Z9ah0100Y8aL0>BhL03T^701EABh205U5:P0406U69P1/B2P0K4PX06m;:`80LT/Z
00@0LT`X07A<:@1dC2T0Mdl[0P1kD2d0101lD2d0PUHc089F<`28G3T208aN>P0408]M>@2:FS@0RUXd
08MG<@80QeH^00@0SE/j08aJ>@2<FST0SU`k0P2BH3l0102IIdH0WF]:09e[BP2OKD`209i/B`040:Id
D`2ZN5L0ZWQG0:ilF`80[7YI00@0YgED0:=aD02SLE00YG=B0P2YMeH0102YNeT0ZgeK0:]mF`2^P5h3
0;22H080[X1N00@0ZgeK0:QjF02XNUP0]8IT0P2`PV030:eoG@040:n1G`2`PV00/89P0;22H080[X1N
00@0/X1O0:]iF02[NEP0/h5P0P2iQfH0102aOeh0/7iM0;1nG@2gQF@20;V7IP040:mmG02dPV40]89Q
0;N5I080/h5P00@0[7YI0:YhE`2ZN5L0/7iM0P2gQF@00`2bP5d0/h5N0;>1GP040;B2G`80/h5N102b
P5d20;F3H0040;V7I02jR6D0^XQU0;V7I0<0^8IS0P2iQf@0102cPE`0/X1K0;:0F`2aOeX20;1nF@04
0:mmF02`OUT0/7iI0;1nF@80/GmJ00@0/X1K0:mmF02_OEP0ZWQC0P2VM4l0102ULdh0YgE@0:MeD02[
NE@20:ilE`040;1kD@2_NU00[gY@0:igCP80ZgA;00@0Zg590:U^AP2YKTH0Yfa40P2VJd<0102YKDP0
Z6a70:Q/A`2VJTD20:AY@@040:9W?`2OI3X0Wf@j09iS>@80WF8f00@0Vf8e09aS=P2LHcH0WfTm0P2T
KT80102OJT00Telg09=O=`2>FS82091L=00409AS>`2<Fc<0S5/c09QZAP80T6E300@0SfI608mZC@2?
JTd0NEHj0P1kFCh0101_ET80FD0/05U0;01fH4/205U3;P0405m:=@1A?2L0DC`W04hl9P80AC<M00@0
?C0M05U<>@1IC3T0BSdZ0P17?2P0101BAc<0AclZ04Lo:P1QFD@203/e7`0404Pm:`1DBCL0E4Tg04i3
<@80@cPT00@0BChX04m4;P1?A2h0CT@[0P1@ARd0101FCST0CdPe04m8=@12?2`204=1=@0404m@BP1>
D4l0CU1?04M<C`80CeAJ1017BeH504I:E@<0ADUD0`1:CUT404]?FPD0C51K0`1;CeX304a@F`<0CE5L
2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019CEP604Q<E`0404M;EP15BE@0ADUD04A8
D`80@dMB0`12AU430455D0040496D@13Ae80@dMB04A8D`80ADUD1016BUD304E9E0D0A4QC3013Ae80
0`15BE@0ATYE04I:E@0604M;EP0304I:E@15BE@0ADUD00@0@d]>0`13BU0304=:DP80@dYD00D0BDUE
04U9D`19BE<0BDY?04U:CP010?ooo`00301;Deh<04Q@F`l0BE5L00@0B51K04M?FP17CeX0AdmJ0P18
D5/304UAG0`0B5EN3017E5d01012C5D0A4eD04A=E012Be8203]0A00402da=00P8B<0824S01TI6@80
5QHF300N6AD0100H61H061TD01PI500T8ah202LT6`0401PD2@0]9AP0;BDH03df9P80=2/J00@0;BPR
03`i<P0l>C80=C8[0P0Z9ah0100Y8aL0>BhL03T^701EABh205U5:P0406U69P1/B2P0K4PX06m;:`80
LT/Z00@0LT`X07A<:@1dC2T0Mdl[0P1kD2d0101lD2d0PUHc089F<`28G3T208aN>P0408]M>@2:FS@0
RUXd08MG<@80QeH^00@0SE/j08aJ>@2<FST0SU`k0P2BH3l0102IIdH0WF]:09e[BP2OKD`209i/B`04
0:IdD`2ZN5L0ZWQG0:ilF`80[7YI00@0YgED0:=aD02SLE00YG=B0P2YMeH0102YNeT0ZgeK0:]mF`2^
P5h30;22H080[X1N00@0ZgeK0:QjF02XNUP0]8IT0P2`PV030:eoG@040:n1G`2`PV00/89P0;22H080
[X1N00@0/X1O0:]iF02[NEP0/h5P0P2iQfH0102aOeh0/7iM0;1nG@2gQF@20;V7IP040:mmG02dPV40
]89Q0;N5I080/h5P00@0[7YI0:YhE`2ZN5L0/7iM0P2gQF@00`2bP5d0/h5N0;>1GP040;B2G`80/h5N
102bP5d20;F3H0040;V7I02jR6D0^XQU0;V7I0<0^8IS0P2iQf@0102cPE`0/X1K0;:0F`2aOeX20;1n
F@040:mmF02`OUT0/7iI0;1nF@80/GmJ00@0/X1K0:mmF02_OEP0ZWQC0P2VM4l0102ULdh0YgE@0:Me
D02[NE@20:ilE`040;1kD@2_NU00[gY@0:igCP80ZgA;00@0Zg590:U^AP2YKTH0Yfa40P2VJd<0102Y
KDP0Z6a70:Q/A`2VJTD20:AY@@040:9W?`2OI3X0Wf@j09iS>@80WF8f00@0Vf8e09aS=P2LHcH0WfTm
0P2TKT80102OJT00Telg09=O=`2>FS82091L=00409AS>`2<Fc<0S5/c09QZAP80T6E300@0SfI608mZ
C@2?JTd0NEHj0P1kFCh0101_ET80FD0/05U0;01fH4/205U3;P0405m:=@1A?2L0DC`W04hl9P80AC<M
00@0?C0M05U<>@1IC3T0BSdZ0P17?2P0101BAc<0AclZ04Lo:P1QFD@203/e7`0404Pm:`1DBCL0E4Tg
04i3<@80@cPT00@0BChX04m4;P1?A2h0CT@[0P1@ARd0101FCST0CdPe04m8=@12?2`204=1=@0404m@
BP1>D4l0CU1?04M<C`80CeAJ1017BeH504I:E@<0ADUD0`1:CUT404]?FPD0C51K0`1;CeX304a@F`<0
CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019CEP604Q<E`0404M;EP15BE@0ADUD
04A8D`80@dMB0`12AU430455D0040496D@13Ae80@dMB04A8D`80ADUD1016BUD304E9E0D0A4QC3013
Ae800`15BE@0ATYE04I:E@0604M;EP0304I:E@15BE@0ADUD00@0@d]>0`13BU0304=:DP80@dYD00D0
BDUE04U9D`19BE<0BDY?04U:CP010?ooo`00301;Deh<04Q@F`l0BE5L00@0B51K04M?FP17CeX0AdmJ
0P18D5/304UAG0`0B5EN3017E5d01013CEH0ADiE04E>E@11BU4203Tn@P0402/_<P0O82807b0R01TI
6@805aLG300N6AD0100J6QP06a`G01/L5`0W9R4202TV7@0401XF2`0T70l091`?03L`8080?C@S00@0
<2/U03Tf;`0i=Rl0<c0Y0P0[:1l0100W8AD0=BXH03DZ601DA2d205Q4:@0406Q59@1[AbL0JdLW06i:
:P80LDXY00@0LD/W07=;:01cBbP0MThZ0P1jCb`0101nDRl0PUHc089F<`28G3T208]M>@0408YL>02:
FS@0RUXd08QH<P80REP`0`2=FcX0102>G3/0Tf5009=Q@02IIdH209e[BP0409m]C02NK4/0WVa;0:Id
D`80ZGMF00@0[7YI0:]iF02[NEP0Z7IE0P2VM5<0102WME@0ZGMF0:UgEP2ZO5X20:]mF`<0[7iL00@0
[GmM0:anG02/OU`0[7iL0P2[OE/00`2bQ680[h5O0:n1G`030:eoG@040:n1G`2aPf40/H=Q0;22H080
[h5O00@0/X1O0:]iF02[NEP0/7iM0P2fQ6<00`2aOeh0/7iM0;1nG@030;J4H`040;F3HP2hQVD0^8IU
0;V7IP80]89Q00@0[G]J0:]iF02[NEP0/GmN0P2hQVD0102jR6D0^HMT0;V7I02gQF820;J4H@<0]H=P
00@0]XAQ0;N5HP2gQF80/GmL0P2dPUl00`2fQ640]hER0;N5HP030;F3H0040;J4H@2hQV<0^8IS0;B2
G@80/h5L102bP5/20;>1G0040;B2G@2ePeh0]H=N0;B2G@80/X1K00@0[geH0:ajE@2/NUD0ZgUD0P2/
NUD0102^O5L0/7iI0;1nF@2^NDl20:ehCP040:efC@2[M4/0ZgA;0:abBP80Zg1800@0ZVm70:U^AP2Y
KTH0ZFe80P2XK4L0102VJTD0Y6U10:AY@@2QIS`209mT>P0409eR=P2LHCD0W64e09]R=@80W6<f00@0
WfTm0:9/@02RK400W6Lm0P2BGSH0102?Fc<0TUhf099N=P2KJT8209=S?@0408aN>P2@ID<0T6E308YS
A080PUe000@0Oee207YH?@1jF3d0LEQ40P1UC3P0101^F4<0GDLb05e7<P1TCcX204/i8`0405Q6<016
=1h0AS@N03lb7`80DD@a00@0BC`Y04e2;P1=@Rh0F4di0P17?bX0101?Ac80;BLA02dW4@1FBcT20515
<`0404Tn;01=@Rh0CD8^05=8<P80CT<]00@0D4H]065G?P1QEch0DTXe0P1ABSL0101;ACD0C4Xn04a:
?P1DEDl204mAD00304=8B`14BDl0A4U?00<0A4QC00@0ADUD04I:E@16BUD0Ad]F1@18C5L304Y>F@@0
BdmJ1@1<D5/304]?FP<0C51K0`1=DE`904iBG@0404eAG01<D5/0C51K04Y>F@80BDeH6018C5L<04U=
F0H0B4aG00@0Ad]F04E9E015BE@0A4QC0P13Ae830496D@<0@DE@00@0@TIA04=7DP13Ae80A4QC0P15
BE@304I:E@@0ADUD1@14B5<<04=7DP0304E9E016BUD0ATYE00H0Ad]F00<0ATYE04E9E015BE@01013
Bdh304=:D0<0@dYB0P13BU@01@19BED0BDUC04U9D`19BTl0BDY>0040oooo000<04YBG@`0B51K3`19
DE`01018D5/0AdmJ04M?FP17CeX204Q@F`<0BE5L6018EEh01015CeP0ADiE04E>E@10BE0203Ll@004
02P/;`0N7b407QlQ01TI6@8061PH300N6AD0100H61H06Q/F01XK5P0U91l202XW7P04028N4`0O5`X0
7aL:034Z6P80BT4`00@0<bhX03Db:`0e<R/0<2dV0P0[:1l0100S7A40<2DC030U4`1B@R/205M3:004
06I38`1ZARH0JTHV06e9:@80L4TX00@0L4XV079:9`1bBRL0M4`X0P1iCR/0101oDc00PeLd08=G=027
FcP208YL>00408UK=`2:FS@0RUXd08YJ=080S5/c0`2>G3/0102@GSd0U69109AR@@2IIdH209e[BP04
09i/B`2OKD`0Wfe<0:IdD`80YgED00<0ZGMF0:YhE`2ZN5L00`2YMeH0102XMUD0ZGMF0:UgEP2ZO5X2
0:UkF@<0Z7YH00@0ZG]I0:YlFP2ZO5X0[7iL0P2]Oed0102_PEl0[X1N0:j0GP2/OU`20:eoG@040:n1
G`2aPf40/H=Q0;:4HP80/H=Q00@0/GmN0:]iF02[NEP0[G]J0P2bP5l0102`OUd0/GmN0;5oGP2cPF02
0;5oGP040;V7IP2lRVT0_8YY0;b:J@80]hET00@0/7iM0:ilF`2^O5/0/h5P0P2hQVD0102nS6T0_8YW
0;b:I`2hQV<20;F3H0040;>1GP2dPUl0]89O0;J4H@80^8IS00@0/X1M0;B2G`2dPUl0]H=P0P2dPUl3
0;:0G@030;B2G`2gQF80]hER00L0]89M0P2fQ5l0102gQF00^8IQ0;R6H@2fQ5l20;F3GP030;B2G@2b
P5/0/X1K00H0/GmJ0`2]N4h0102]MTd0[7E<0:aeC02]Ld/20:ebBP<0[75900@0ZVa70:U[AP2YJdH0
Z6Y30P2UId00102SIC`0XF@h0:5T>02OHSH209iQ=@0309aS=P2MI3L0WF@g00<0WfTm00@0V6<i095M
=@2AGCD0T5`d0P2DH3P0102KJT80UVI009IV@024ES82099WA@0408QQ@P1gE3P0Me@h08=QAP80Kdlf
00@0JeA206M@?P1WD3h0HDXh0P1KAS<0101UD3d0AS@P04Hd801NC3P204Dd80040554<@1<?b`0C3l/
04@g9080B3dY00@0DDHb04Dm:015?BP0AclZ0P0j=1h0101OE480HeQ606=HAP1ICS`204a1;@0404e2
;01AAS00DDH`04m5;080CD<Z00@0DDTd05=<>@1CC3T0CdTi0P1=Bcl0101@DD/0CDm>04e?CP12AdX2
03m4BP040455D012AU40@TIA04=7DP80ADUD00<0Ad]F04U=F019CEP00`1;CeX304Y>F@@0BdmJ1@1<
D5/304]?FP<0C51K0`1=DE`904iBG@0404eAG01<D5/0C51K04Y>F@80BDeH6018C5L<04U=F0H0B4aG
00@0Ad]F04E9E015BE@0A4QC0`13Ae820496D@<0@DE@00@0@TIA04=7DP13Ae80A4QC0P15BE@00`16
BUD0ADUD04E9E00304E9E0<0A4QC0`13Ae8<04A8D`<0ADUD00@0ATYE04M;EP17BeH0Ad]F0P16BUD3
04E9E0@0@d]>0`13BU0304=:DP80@dYD00D0BDUE04U9D`19BE<0BDY?04U:CP010?ooo`00301:DUd<
04Q@F`l0BE5L00@0B51K04M?FP17CeX0AdmJ0P18D5/304UAG1P0B5EN00@0ADmH04E>E@15CUD0@4U@
0P0g?400100X;2l07QlQ01hO8@0I6AT201PH60`07QTE00@061PF01XK5P0J6aH09B@O0P0Z9ah0100R
7Q<07aL:01lG2P0a:QX204Y1<00403<^:00e<R/0=C8[030]9P80:bPO00@08adA030U4`0`9A<0DT8[
0P1G@bP0101V@b<0JTHV06Y69P1]BBT20719:004071:9P1bBRL0LTXW07A<:080NDh[00@0Oe<`08=G
=023Ec@0Qe/h0P2:G3P01029FcL0RUXd08YJ=02:FS@208aK<`<0SU`k00@0T5hm09AR@@2DHT40VFM6
0P2MJdX0102NK4/0Wfe<09m]C02VM5<20:MeE0030:UgEP2ZN5L0ZWQG00<0ZGMF00@0Z7IE0:UgEP2Y
MeH0ZWaJ0P2YNeT30:QjF0040:UkF@2ZO5X0ZWaJ0:anG080[GmM00@0[h5O0:j0GP2^P5h0[7iL0P2]
Oed0102_PEl0/H=Q0;63H@2bQ6820;63H@040;5oGP2[NEP0ZgUH0:ekFP80/X1O00@0/7iM0;5oGP2a
Oeh0/h5P0P2aOeh0102iQfH0_8YY0;b:J@2lRVT20;N5I0040;1nG@2^O5/0[WaK0;>1H080^8IU00@0
_XaY0;b:I`2lRVL0^8IS0P2ePf00102cPEh0]89O0;B2G`2fQ6420;R6H`040;:0G@2dPUl0]89O0;F3
H080]89O0`2bP5d00`2dPUl0]hER0;N5HP070;B2G@80]XAO00@0]hEP0;R6H@2hQV40]XAO0P2ePeh0
0`2dPUd0/X1K0;:0F`060;5oFP<0[GQ>00@0[GI=0:aeC02/MD`0[G=;0P2]LTX30:aaB@040:Y/A`2Y
JdH0ZF]60:QZ@`80YFM000@0XfDl0:5T>02QI3P0Wf8f0P2NHCD00`2LHcH0WF@g09eT=`0309mY?@04
09QS>@2AGCD0TEde091L=080U60h00@0VfY209IV@02FIT00Q5Hb0P2BIdD01028HD80Me@h07MD>023
HDH206m?=P0406]D@P1WD3h0Ie0n065:>080FdHc00@0IE0m04Hd8016=200GT`h0P15=200101AA340
C3l/04`o;014=b@204Pm:@040556<P15?BP0ACdX04Lo:P80>S@N00@0GeA206=HAP1SF4H0FDhl0P1<
@Bd0101=@R`0DDH`0556<01?AB`204e3:P040559=01CC3T0Dd`i04m9>@80CD/o00@0D55;04e?CP1=
Cdh0@TM:0P0oA4X01011AE00@TIA0496D@13Ae8204E9E00304M;EP19CEP0BDeH00<0BdmJ0`1:CUT4
04]?FPD0C51K0`1;CeX304a@F`<0CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019
CEP604Q<E`0404M;EP15BE@0ADUD04A8D`<0@dMB0P12AU430455D0040496D@13Ae80@dMB04A8D`80
ADUD00<0ATYE04E9E015BE@00`15BE@304A8D`<0@dMB3014B5<304E9E00404I:E@17BeH0Ad]F04M;
EP80ATYE0`15BE@404=;CP<0@dY@0`13BU8204=:E00504U9E@19BE<0BDUC04U:C`19BTh00@3oool0
00`0BU9M3018D5/?04UAG00404Q@F`17CeX0AdmJ04M?FP80B51K0`19DE`<04QEGP`0BEIO00@0Ae5J
04I?EP16CeH0?dQ?0P0e>Sh00`0V:Rd071dO01`M7`0301TI6@`07QTE00@05QHD01HG4P0F5a8081lJ
0P0Y9Qd0100_:b008AT<024I300_:1P204m6=@0403D`:P0a;RL0<BhW02`Y8P80;2TP00@08Q`@02/P
3P0[80h0CShW0P1F@RL0101U@R80JDDU06U59@1/B2P206m89`0406m99@1aBBH0LDTV07=;9`80N4dZ
00@0P5@a08=G=023Ec@0Qe/h0`29FcL208YJ=00408]K=@2>GCD0SUde08eK>P80SU`k00@0T5hm09AR
@@2DHT40V6I50P2KJDP00`2MJdX0WVa;09i/B`030:EcDP030:IdD`2XMUD0Z7IE00<0ZWQG00@0ZGMF
0:QfE@2XMUD0YgUG0`2VN5H20:MiE`040:QjF02YNeT0ZG]I0:]mF`D0[7iL00@0[GmM0:j0GP2^P5h0
/89P0P2bQ6830;>5H`040;5oGP2/NUT0[7YI0:ajF@80[WaK00@0/7iM0;5oGP2aOeh0/7iM0P2]NeX0
102gQF@0^hUX0;^9J02nS6/20;b:J@040;J4H`2cPF00/h5P0;F3HP80^8IU00@0^hUV0;V7I02iQf@0
]XAQ0P2cPEh30;:0G@030;>1GP2dPUl0]89O00@0]XAQ0P2dPUl30;:0G@040;F3H02gQF80]hER0;J4
G`<0]H=N0`2dPUd20;F3GP030;J4G`2gQF00]hEP00<0]XAO0`2gQF00102fQ5l0]89M0;B2G@2bP5/2
0;5oFP040:mjD02^NDl0[WU?0:igCP80[GI=00<0[G=;0:ebBP2]LTX00`2/LDT0102ZK4D0ZF]40:U[
A02WJD820:EW?P040:=V>P2PHcL0X6<g09mR=@80WV4d00@0W6<f09iU>02NICP0WVPl0P2KICT0102E
H3H0T5`d091L=02BGSH209IR>P04095Q>`2=Gc/0SElk08QL>@80SVE500@0Pei107aI?@1lFCd0OEe4
0P1VARd0101UCS`0GDHd05e6=01H@c02058m:P0405a:=P16=B40ASDQ05I5<@80BCPT00@0G4ll04Xm
:P1:?BX0ACPU0P16>bL01019?RX0@c/V04<k9P17?bX205===`0405iC@@1QET@0HEI405Q=>`80C44]
00@0CD8/05A9<`1DBC<0DDL^0P19?bH0101ABC@0D4Tf0519=P19@c<204=1=@0404E6@019BdX0BD]:
04M<C`80ADY@00@0@DE@0496D@12AU40@dMB0P15BE@00`17BeH0BDeH04U=F00304]?FP<0BTiI101;
CeX504a@F`<0BdmJ0`1<D5/304eAG0T0CU9M00@0CE5L04a@F`1<D5/0BTiI0P19CEPH04Q<E``0BDeH
1P18C5L01017BeH0ADUD04E9E014B5<504=7DP040496D@11AE00@DE@0455D080@TIA00<0@dMB04A8
D`14B5<01015BE@304A8D`D0@dMB3015BE@00`14B5<0ADUD04E9E00604I:E@0304E9E014B5<0A4QC
00@0@d]>0`13BU0304=:DP80@dYD00D0BDUE04U9D`19BE<0BDY?04U:CP010?ooo`003019DE`H04Q@
F`<0BE5L00@0B51K04M?FP17CeX0AdmJ0P18D5/304UAG0`0B5EN3019EUl01019De`0Ae1G04M@E`10
BE0203Dj?P0302HZ;@0L7Al071dO00<06ATI300N6AD0100J6QP05AHA01DF4@0L6aH202HS6P0403Pd
:@0W7a809alB02dV5P80AC`[00@0=S4[02l/9@0_;2D0:BHO0P0]:R40100U7a<09a`:02LL2P17=b02
05I29`0406E28P1XA2@0J4@T06]79`80KTLV00@0KTPT07189@1`B2D0Ld/W0P1gC2T0101oDc00PUHc
089F<`26FSL208UK=`0408QJ=P2:FS@0RUXd08]K=@80SUde00@0S5Xi08iL>`2>G3/0TEln0P2DHT40
102FI4<0VFM609UWAP2LJTT209e[BP<0Xg5@00@0Y79A0:IdD`2VM5<0Z7IE0P2YMeH30:QfE@040:=e
D`2UMeD0YGME0:IhEP80Z7YH1P2YNeT0102ZO5X0[7iL0:anG02]Oed20:n1G`040;22H02aPf40/H=Q
0;>5H`80]8IT00@0/GmN0;1nG@2`OUd0[G]J0P2/NUT0102`OUd0/GmN0;5oGP2^O5/20:ajF@040;1n
G@2fQ6<0]XAS0;f;JP80_he/00<0_8YY0;N5I02gQF@00`2fQ6<40;R6H`80]hER00@0]XAQ0;F3H02e
Pf00]89O0P2cPEh40;Z8I@80]hER0`2dPUl0102fQ640^HMT0;V7I02hQV420;N5H0040;F3GP2dPUd0
]89M0;>1G0<0/X1K0P2cPE`00`2dPUd0]H=N0;F3GP030;N5H0040;J4G`2dPUd0]89M0;5oFP80[geH
00@0/GaB0;1kD@2`Ne40/7U@0P2]MTd0102]Ld/0[7590:aaB@2ZKdL20:U^AP040:YZA02YJT40ZFY1
0:QY@080YFHk00@0Xf@i0:5R=P2QHSH0Wf0d0P2NGc<0102MI3L0WVDh09iU>02KICT209QR=P0409AO
=@2BGSH0TUhf09AP>080UF4i00@0R5Xf08EG<`25Ec<0T6E30P22FCT0101iE3L0PV15089PA@1dE3/2
06I8<004061;>P1?>RT0CcXY05Y5=080BcPW00@0ETDc04ln;01??R`0Bc`Y0P1B@c00101>@Bh0ACPU
04Dh9@1;?R/20515<@0404Tn:P10>2<0@3PS040h8`80EDli00@0KV=A0597=@1BAcD0B3d[0P1GC3P0
101ICSP0C44[04a1:`1>A2/205mE?00404a4;`15?R/0ACh[040j:P80?S``00@0@44k04Q:B@18BTT0
Be1C0P1:CeD01013Ae80A4QC04A8D`15BE@204I:E@0304M;EP18C5L0B4aG00<0BDeH0`1:CUT404]?
FPD0C51K0`1;CeX304a@F`<0CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019CEP6
04Q<E`0404M;EP15BE@0ADUD04A8D`80@dMB00@0A4QC04=7DP13Ae80@TIA0`11AE020496D@<0@dMB
0`15BE@304A8D`<0@dMB0`12AU4<04E9E0<0A4QC00@0ADUD04I:E@16BUD0ATYE0P15BE@304A8D`@0
@d]>0`13BU0304=:DP80@dYD00D0BDUE04U9D`19BE<0BDY?04U:CP010?ooo`003019DE`H04Q@F`<0
BE5L00@0B51K04M?FP17CeX0AdmJ0P18D5/304UAG0`0B5EN3019EUl01019De`0Ae1G04M@E`10BE02
03Dj?P0302HZ;@0L7Al071dO00<06ATI300N6AD0100J6QP05AHA01DF4@0L6aH202HS6P0403Pd:@0W
7a809alB02dV5P80AC`[00@0=S4[02l/9@0_;2D0:BHO0P0]:R40100U7a<09a`:02LL2P17=b0205I2
9`0406E28P1XA2@0J4@T06]79`80KTLV00@0KTPT07189@1`B2D0Ld/W0P1gC2T0101oDc00PUHc089F
<`26FSL208UK=`0408QJ=P2:FS@0RUXd08]K=@80SUde00@0S5Xi08iL>`2>G3/0TEln0P2DHT40102F
I4<0VFM609UWAP2LJTT209e[BP<0Xg5@00@0Y79A0:IdD`2VM5<0Z7IE0P2YMeH30:QfE@040:=eD`2U
MeD0YGME0:IhEP80Z7YH1P2YNeT0102ZO5X0[7iL0:anG02]Oed20:n1G`040;22H02aPf40/H=Q0;>5
H`80]8IT00@0/GmN0;1nG@2`OUd0[G]J0P2/NUT0102`OUd0/GmN0;5oGP2^O5/20:ajF@040;1nG@2f
Q6<0]XAS0;f;JP80_he/00<0_8YY0;N5I02gQF@00`2fQ6<40;R6H`80]hER00@0]XAQ0;F3H02ePf00
]89O0P2cPEh40;Z8I@80]hER0`2dPUl0102fQ640^HMT0;V7I02hQV420;N5H0040;F3GP2dPUd0]89M
0;>1G0<0/X1K0P2cPE`00`2dPUd0]H=N0;F3GP030;N5H0040;J4G`2dPUd0]89M0;5oFP80[geH00@0
/GaB0;1kD@2`Ne40/7U@0P2]MTd0102]Ld/0[7590:aaB@2ZKdL20:U^AP040:YZA02YJT40ZFY10:QY
@080YFHk00@0Xf@i0:5R=P2QHSH0Wf0d0P2NGc<0102MI3L0WVDh09iU>02KICT209QR=P0409AO=@2B
GSH0TUhf09AP>080UF4i00@0R5Xf08EG<`25Ec<0T6E30P22FCT0101iE3L0PV15089PA@1dE3/206I8
<004061;>P1?>RT0CcXY05Y5=080BcPW00@0ETDc04ln;01??R`0Bc`Y0P1B@c00101>@Bh0ACPU04Dh
9@1;?R/20515<@0404Tn:P10>2<0@3PS040h8`80EDli00@0KV=A0597=@1BAcD0B3d[0P1GC3P0101I
CSP0C44[04a1:`1>A2/205mE?00404a4;`15?R/0ACh[040j:P80?S``00@0@44k04Q:B@18BTT0Be1C
0P1:CeD01013Ae80A4QC04A8D`15BE@204I:E@0304M;EP18C5L0B4aG00<0BDeH0`1:CUT404]?FPD0
C51K0`1;CeX304a@F`<0CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019CEP604Q<
E`0404M;EP15BE@0ADUD04A8D`80@dMB00@0A4QC04=7DP13Ae80@TIA0`11AE020496D@<0@dMB0`15
BE@304A8D`<0@dMB0`12AU4<04E9E0<0A4QC00@0ADUD04I:E@16BUD0ATYE0P15BE@304A8D`@0@d]>
0`13BU0304=:DP80@dYD00D0BDUE04U9D`19BE<0BDY?04U:CP010?ooo`003019DE`H04Q@F`<0BE5L
00@0B51K04M?FP17CeX0AdmJ0P18D5/304UAG1P0B5EN00@0BUAM04UBF@19DUT0@DYA0P0f>cl00`0W
:bh07AhP01dN800301TI6@`07QTE00@08B4O01TJ5@0I6QD07A`G0P0O71<0100h=2T0:b<F02/S5P0^
9aL203L^7@0403D`:P0`;BH0<2dV02HS7080;BXQ00@0:bDI02DJ200U6PP0?RhG0P1G@bP0101U@R80
JDDU06U59@1[AbL206i79P0406i8901`B2D0L4PU07=;9`80N4dZ00@0O50]081D<@20E340QETf0P29
FcL01028FSH0RUXd08YJ=02:FS@208aK<`0408]I>02=FcX0SE/j091N?@80Tf5000@0U69109IT@`2F
I4<0VFM60P2LJTT0102PKTd0XFm>0:5_CP2TLU430:EcDPD0YWAC00@0Y7ID0:EgE@2UMeD0YgUG0P2X
NUP30:UkF@040:QjF02WNEL0YgUG0:UkF@80[7iL00@0[h5O0;22H02`PV00/89P0P2aPf40102bQ680
/hES0;>5H`2`OUd20;F3HP040;5oGP2]NeX0[G]J0;5oGP80/X1O00@0[WaK0:ekFP2]NeX0[7YI0P2b
P5l0102jR6L0_H]Z0;f;JP2lRVT20;R6I@040;F3HP2dPV40]89Q0;R6H`80^HMT00@0^hUV0;b:I`2l
RVL0_8YW0P2kRFH30;Z8I@040;b:I`2mRfP0_H]X0;f;J080^hUV00@0]hER0;J4H@2fQ640]hER0P2i
Qf@0102jR6<0^HMR0;V7HP2fQ5l20;B2G@030;:0F`2aOeX0/GmJ00<0/GmJ00@0/h5L0;F3GP2ePeh0
]XAO0P2gQF00102fQ5l0/h5L0;>1G02`OUT20:mmF0040;5lDP2`Ne40/7]A0:mhC`80[GI=00@0[79:
0:Y_A`2ZKdL0ZFi60P2WK4@0102ZJd80ZFY10:UZ@@2WJ3d20:EV>`040:=T>02PHCD0X64e09mQ<P80
WV0a00@0WVDh09eT=`2MI3L0V68f0P2EGc<0102DGcD0UV8j09IR>P2EHCT2099N=P0408MI=@24F3D0
Q5Pe091U@`80MTl`00@0Kd``07aL@`1lG4<0Jdde0P1^DS`0101PBS`0CCX[04dj:`1LBCX204dl;004
05M6=P1B@S80DT8b04Lg9`80FDTi00@0@SDR04Dh9@15>2D0BSdZ0P1>@bl0100o=200>30K03P`6`0k
<ah204i8<P04075VE01GC3X0Ed`j04e2<080EDXf00@0F4dg05M<=P1GC3H0GE<j0P1SFD001017?bX0
>c@Q03/d8@0k=BD204A2=P0404Q9@`1;CD`0Bde<04U>D@80ADY@0`15BE@404I:E@D0Ad]F0`1:CUT4
04]?FPD0C51K0`1;CeX304a@F`<0CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019
CEP604Q<E`0404M;EP15BE@0ADUD04A8D`80@dMB00@0ADUD04A8D`14B5<0@dMB0P12AU430455D003
0496D@13Ae80@dMB00@0A4QC0`13Ae850496D@`0A4QC00<0@dMB04A8D`14B5<01P15BE@00`14B5<0
@dMB04=7DP0404=;CP<0@dY@0`13BU8204=:E00504U9E@19BE<0BDUC04U:C`19BTh00@3oool001P0
B51K3017CeX304UAG00404Q@F`17CeX0AdmJ04M?FP80B51K0`19DE`<04QEGP`0AeAM00@0BeEN04YC
FP1:DeX0@daC0P0i?T80100Y;C007b0R01lP8P0J6QX201TI6@`07QTE00@07alM01TJ5@0I6QD07QdH
0P0F4`X0100_:b00;RHI02hV6@0h<B4203H]700403<^:00c<2T0<c0Y02DR6`80;BXQ00@0<R`P02@I
1`0T6@L0=bL@0P1HA2T0101V@b<0JDDU06U59@1/B2P206m89`0406m99@1aBBH0LDTV07A<:080N4dZ
00@0NDdZ07iB;`1nDRl0Q5Pe0`29FcL308UI<`80RUTa00@0RELf08aJ>@2<FST0T5hm0P2AGch0102B
H3l0U69109AR@@2GID@209YXA`0409e[BP2QKdh0XFm>0:AbD@80YG=B00@0XW1?0:5_CP2QKdh0Xg5@
0P2ULe80102XNUP0YgUG0:MiE`2WNEL20:IhEP<0YgUG00@0Z7YH0:UkF@2YNeT0ZG]I0P2/OU`00`2`
PV00/H=Q0;63H@030;22H0040;63H@2cQF<0/hES0;1nG@80^XQW00<0]XAS0:mmG02_OE`00`2bP5l0
102^O5/0/7iM0;1nG@2^O5/20;:0G`040;J4H`2hQVD0^8IU0;J4H`80]89Q0`2bP5l40;R6H`80^XQU
00@0^hUV0;f;J02mRfP0_heZ0P30SV/0102lRVL0_XaY0;j<J@2nS6T20;f;J0040;V7I02fQ640]XAQ
0;J4H@80]hER00@0^hUT0;Z8H`2jR6<0^8IQ0P2fQ5l0102ePeh0]89M0;B2G@2dPUd50;F3GP<0]XAO
00@0]H=N0;>1G02cPE`0/GmJ0P2`OUT0102_NU00[WU?0:iiC`2^Mdh20:aeC0040:]aB@2[L4P0Zg18
0:U^AP80Z6e500@0ZfY20:YY?`2ZJCl0Z6Lm0P2VISX0102SHcL0XF4c0:5Q<`2OGc4209iN<00409iU
>02LHcH0W6<f09EO<`80Teda00@0UV4g09YV?P2JISh0UV8j0P2>FS801023Ec@0RUhk08YN>`24Fc/2
07QA<P0407EB=P1[Bc80Jd/b06A6;P80LEDo00@0FDHg05E2<`1E@S<0F4Df0P1B@C40101FASH0B3PX
04Ph:014=bH20595=004044d8@1;?R/0Bch[044d8@80@CHR00@0<RLC03L_6P0g;aX0AclZ0P1CCCL0
101HCC/0FDhl05U>?01ICS`205A9=@0405U>>01/HD/0K65;06]QB080DTP_00@0C4@_03Pa7P0h<Ah0
>S@T0P1<BSh0101@DD/0C4i=04a>C@15BTd20415B`@0ADUD0P16BUD404M;EP80B4aG0`1:CUT404]?
FPD0C51K0`1;CeX304a@F`<0CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019CEP6
04Q<E`0404M;EP15BE@0ADUD04A8D`80@dMB00@0ADUD04A8D`14B5<0@dMB0P12AU430455D0<0@TIA
0`14B5<304=7DPH0@TIA3`13Ae801014B5<0ADUD04E9E015BE@204A8D`<0@dMB1013Bdh304=:D0<0
@dYB0P13BU@01@19BED0BDUC04U9D`19BTl0BDY>0040oooo000H04Q@F``0AdmJ0`19DE`01018D5/0
AdmJ04M?FP17CeX204Q@F`<0BE5L3018EEh<04MDG@0404]EGP1:DeX0BU=J04=<D`80>Ci200@0:Bd`
01lP8P0O82806QXJ0P0I6AT<01hI5@0401lO7@0I6QD06AXE01hM60805Q<:00@0;b/P02hV6@0^9QT0
>34Q0P0f;A`0100c;RP0<c0Y03<`:@0U8Q/202dZ8@04038/800T6@L091T703LW4080F4@Y00@0IT<S
06U59@1YABD0K4PX0P1_B2L0101_BBD0LDTV07599P1dC2P207Q=:P0407U=:P1nDRl0OU8_08AH=@<0
RE/g0`29FC<208YI<@0408UG=P2<FST0S5Xi091N?@80TEln00@0TV0o09AR@@2DHT40UfE40P2JJ4L0
102MJdX0XFm>0:5_CP2TLU420:EcDP040:9`C`2QKdh0XFm>0:=aD080YG=B00@0Z7YH0:MiE`2WNEL0
YgUG0P2VN5H30:MiE`040:QjF02YNeT0ZG]I0:UkF@80[7iL00<0/89P0;63H@2aPf400`2`PV00102a
Pf40/hES0;>5H`2`OUd20;Z8I`030;J4H`2_OE`0[geL00<0/X1O00@0[WaK0;1nG@2`OUd0[WaK0P2b
P5l0102fQ6<0^8IU0;R6I@2fQ6<20;B2H@<0/X1O102hQV<20;Z8I@040;^9IP2mRfP0_H]X0;n=JP80
`8i[00@0_8YW0;j<J@2nS6T0_XaY0P2mRfP0102iQf@0]XAQ0;J4H@2fQ6420;N5HP040;^9I02jR6<0
^XQS0;R6H@80]XAO00@0]H=N0;B2G@2dPUd0]89M1@2ePeh30;J4G`040;F3GP2cPE`0/h5L0;5oFP80
/7iI00@0[gY@0:iiC`2^NDl0[WM>0P2/MD`0102[LDT0Zg180:]`B02YKTH20:Q]A@040:]Z@P2ZJCl0
ZVTo0:QW?@80YVHj00@0Xf<g0:5Q<`2QHC<0Wela0P2NGS00102NICP0W6<f09aS=P2EGc<209=M<@04
09IQ=`2JISh0VVHn09IR>P80SUXb00@0PeLd08YN>`2:GS/0Q5/k0P1hDC80101eDSH0Jd/b06];<P1T
ARh2075E?`0405U6=`1E@S<0ED8c05Q5=P80DT4a00@0ETHf04Ph:018>2P0A3LV0P1BAC@01011=240
Bch[04/n:`11=242044f8P04038W4`0g;aX0=blJ04Lo:P80Dddg00@0F4dk05U>?01ICS`0FDhl0P1D
BCD0101ICSP0K65;06aQB`1[HDP20598;`0404a4;`0h<Ah0>34N03Xd9080C4Xn00@0D55;04a>C@1<
CTd0ADY=0P10AD/404E9E080ATYE1017BeH204Q<E`<0BTiI101;CeX504a@F`<0BdmJ0`1<D5/304eA
G0T0CU9M00@0CE5L04a@F`1<D5/0BTiI0P19CEPH04Q<E``0BDeH1P18C5L01017BeH0ADUD04E9E014
B5<204=7DP0404E9E014B5<0A4QC04=7DP80@TIA0`11AE030496D@<0A4QC0`13Ae860496D@l0@dMB
00@0A4QC04E9E015BE@0ADUD0P14B5<304=7DP@0@d]>0`13BU0304=:DP80@dYD00D0BDUE04U9D`19
BE<0BDY?04U:CP010?ooo`006018D5/<04M?FP<0BE5L00@0B51K04M?FP17CeX0AdmJ0P18D5/304UA
G0`0B5EN3017E5d0101;EEh0BeAK04]DF`15CUD203]0A00402/_<P0P8B<0824S01XJ6P806ATI300N
6AD0100H61H05AHA01DF4@0M71L200l<0`0402LS600`:1/0<2PK04<l;080?C@S00@0<RdW03Db:`0e
<R/09B8K0P0^:b80100g<BD091T702@I1`0b8P/205U5:P0406I38`1ZARH0JTHV06a8:080KdPW00@0
KdTU07599P1aBBH0M4`X0P1iCR/0101gBbP0O50]07a@;@24F3D308UK=`80RETc00@0R5Pb08QG;`28
Ebl0R5He0P2;FCP0102?GC`0TEln095O?P2AGch2099P?`0409IT@`2IIdH0VFM609]YB080X6i=00@0
YG=B0:AbD@2TLU40X6i=0P2NK4/0102QKdh0YG=B0:EcDP2/OU`20:YlFP040:IhEP2TMU@0Y7ID0:Af
E080YWQF00@0ZG]I0:]mF`2[OE/0ZG]I0P2]Oed0102aPf40/XAR0;:4HP2`PV020:n1G`040;22H02b
Q680/XAR0;1nG@80_H]Z00<0^HMV0;1nG@2`OUd00`2bP5l0102^O5/0/X1O0;:0G`2aOeh50;>1H004
0;5oGP2_OE`0[geL0;1nG@80/GmN00<0]H=P0;B2G`2dPUl00`2cPEh0102fQ640^XQU0;Z8I@2oSFX2
0<:@K@040;^9IP2mRfP0_H]X0;j<J@80_H]X00<0^HMT0;J4H@2fQ6400`2dPUl0102kRF@0^XQS0;Z8
H`2iQf820;R6H@040;N5H02hQV40^8IQ0;R6H@80^HMR00<0]XAO0;N5H02gQF000`2gQF00102fQ5l0
]89M0;B2G@2bP5/20;5oFP040:agC@2[MT`0ZgI<0:aeC080ZW=:00<0Zg590:]`B02[L4P00`2ZKdL0
102ZJCl0ZFPn0:UX?P2XIcd20:EU>@040:=S=`2QHC<0XF4c09mO;`80WUh^00@0WfHi09]R=@2KHSD0
Teda0P2AFbl0102GHSP0WFU109eY@@2GHc/208]G;`0407iB;`2?I480SfA207]B<P80PE`o00@0PEi2
05`l8`1L?2<0G40Z0P1]DC/0101B?c00G4Tj05a9>P1>?Bd205A3<`0405A4=00i:AT0>BTI048e9080
BC`[00@0?S4N04i1;P1>@Bh0>bhK0P0l<Ad0100b9a<0@3PS040h8`1ECCP205QB?00405M<>P1DBCL0
E4Tg05eB@080IU]700@0IU]506AI@`1TFD<0E4Xa0P0i;aH0101GCcX0>c@Q03/d8@0k=BD204m=@@04
059CC@1;CD`0Bde<04=8B`80?dA:0`14B5<01015BE@0ATYE04I:E@17BeH204Q<E`<0BDeH0`1:CUT4
04]?FPD0C51K0`1;CeX304a@F`<0CE5L2@1>DUd0101=DE`0C51K04a@F`1:CUT204U=F1P0B4aG3019
CEP604Q<E`0404M;EP15BE@0ADUD04A8D`80@dMB00@0ADUD04A8D`14B5<0@dMB0P12AU430455D0<0
@TIA0`14B5<304=7DP@0@TIA0P11AE0?04=7DP0404A8D`15BE@0ADUD04E9E080A4QC0`13Ae8404=;
CP<0@dY@0`13BU8204=:E00504U9E@19BE<0BDUC04U:C`19BTh00@3oool001P0B51K3017CeX304Q@
F`@0AdmJ0P18D5/00`1:DUd0Be=N04]CGP0<04MDG@`0B5EN00@0BeQP04QEF`18EE/0A4mE0P0mAdT0
100b>S`0:2`]02P/;@0M7ah201DG5P`07QTC00@061LE01@C3`0D4`l0924L0P0F4`X0100J5PX0=2hP
03@^800e;al204I1;P0403ln?00X9240:2@Q02hY9@808QT@00@0:AlC03HY6@0f:AT0:1T60P1<>bL0
101[CB/0K4TY06a9:@1^AbH207569P0407A79P1eBRL0MDXW07A<:080LddV00@0NddY07a>:P1lCRX0
OU0/0P20DRh01022E300Q5Hb08AF<P25Ec<208IH=00408]I=P2=FcP0SE/h08mM>P<0T5hk0P2AGc`0
102DHSl0UfE209MU@P2LIdD209eXAP0409mZB02QK4X0XFa:0:A_C@80YW5?00@0Z7=A0:UdDP2YM580
YgED0`2VM5<20:EcDP040:IdD`2WME@0YgED0:QfE@80ZGMF00<0YgUG0:YlFP2ZO5X00`2/OU`0102[
OE/0ZWaJ0:YlFP2ZO5X20:anG0030;1nF`2aOe`0/GmL00<0/h5N00@0/GmL0:ekF02]NeP0ZGMD0P2V
M5400`2^O5T0/X1M0;:0G@030;F3H0040;>1GP2aOe`0/GmL0;5oG080/h5N00@0]XQT0;B6HP2dQV80
/XAP0P2`PUh0102aPel0]8IR0;B6HP2hRVH20;Z<J0040;N9I@2eQf<0]HMS0;B6HP80]HMS00@0]XQT
0;F7H`2eQf<0/XAP0P2^P5`40;N5H0<0]XAO1@2ePeh30;:2G0<0/h=M00@0/X9L0;61F`2aPE/0[gmI
0P2^OUP0102/Mdd0ZWE;0:YeB`2YLTT20:YcBP040:idC02_M4`0[gA<0:ebBP80Zg1800@0XfPl0:EZ
?P2UJSh0Yfa00P2WJSh0102SI3T0X64f0:1Q=P2PGcD20:5P=P040:1X>@2IHC80VF4b095K;080TE/]
00@0UF0d09MT=`2GI3L0U64f0P2AGS<0102:FS@0RE/g08UK=`2;H3h207iG>00406M5:P1RA2`0HT@/
0697<P80F3l[00@0FTLh05a9>P1LBCX0@BhO0P1GAS@01013=1l0=2D@03@U400b9Ph204Xn9P0403/a
601=@bX0CD<Z04e2;080>BhJ00@0A3TU05Q=>`1HCC/0GE900P1bIeD0101LC3D0GDdf05e==P1]GTL2
07UZD`0406EI@@14>R40A3XQ03/a6080@cTP00@0>3DT03Hc900f<b@0@T0d0P1>CTH01018BTD0@DE6
0455AP14BDd204I;C`<0A4QC00@0ADUD04I:E@16BUD0Ad]F0`18C5L204U=F00404I:E@17BeH0Ad]F
04Q<E`80BDeH00@0BdmJ04a@F`1<D5/0CE5L0P1>DUd304]?FP<0C51K0`1=DE`404iBG@<0CE5L0P1<
D5/00`1;CeX0BTiI04Y>F@0604U=F0<0B4aG1017BeH204I:EAP0B4aG1P17BeH304I:E@<0ADUD0`14
B5<304=7DP@0@TIA3P11AE0?04=7DP0404A8D`15BE@0ADUD04E9E080A4QC0`13Ae8404A9C@<0A4U?
0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool001P0B51K3017CeX304Q@F`@0AdmJ
0P18D5/00`1:DUd0Be=N04]CGP0<04MDG@`0B5EN00@0BeQP04QEF`18EE/0A4mE0P0mAdT0100b>S`0
:2`]02P/;@0M7ah201DG5P`07QTC00@061LE01@C3`0D4`l0924L0P0F4`X0100J5PX0=2hP03@^800e
;al204I1;P0403ln?00X9240:2@Q02hY9@808QT@00@0:AlC03HY6@0f:AT0:1T60P1<>bL0101[CB/0
K4TY06a9:@1^AbH207569P0407A79P1eBRL0MDXW07A<:080LddV00@0NddY07a>:P1lCRX0OU0/0P20
DRh01022E300Q5Hb08AF<P25Ec<208IH=00408]I=P2=FcP0SE/h08mM>P<0T5hk0P2AGc`0102DHSl0
UfE209MU@P2LIdD209eXAP0409mZB02QK4X0XFa:0:A_C@80YW5?00@0Z7=A0:UdDP2YM580YgED0`2V
M5<20:EcDP040:IdD`2WME@0YgED0:QfE@80ZGMF00<0YgUG0:YlFP2ZO5X00`2/OU`0102[OE/0ZWaJ
0:YlFP2ZO5X20:anG0030;1nF`2aOe`0/GmL00<0/h5N00@0/GmL0:ekF02]NeP0ZGMD0P2VM5400`2^
O5T0/X1M0;:0G@030;F3H0040;>1GP2aOe`0/GmL0;5oG080/h5N00@0]XQT0;B6HP2dQV80/XAP0P2`
PUh0102aPel0]8IR0;B6HP2hRVH20;Z<J0040;N9I@2eQf<0]HMS0;B6HP80]HMS00@0]XQT0;F7H`2e
Qf<0/XAP0P2^P5`40;N5H0<0]XAO1@2ePeh30;:2G0<0/h=M00@0/X9L0;61F`2aPE/0[gmI0P2^OUP0
102/Mdd0ZWE;0:YeB`2YLTT20:YcBP040:idC02_M4`0[gA<0:ebBP80Zg1800@0XfPl0:EZ?P2UJSh0
Yfa00P2WJSh0102SI3T0X64f0:1Q=P2PGcD20:5P=P040:1X>@2IHC80VF4b095K;080TE/]00@0UF0d
09MT=`2GI3L0U64f0P2AGS<0102:FS@0RE/g08UK=`2;H3h207iG>00406M5:P1RA2`0HT@/0697<P80
F3l[00@0FTLh05a9>P1LBCX0@BhO0P1GAS@01013=1l0=2D@03@U400b9Ph204Xn9P0403/a601=@bX0
CD<Z04e2;080>BhJ00@0A3TU05Q=>`1HCC/0GE900P1bIeD0101LC3D0GDdf05e==P1]GTL207UZD`04
06EI@@14>R40A3XQ03/a6080@cTP00@0>3DT03Hc900f<b@0@T0d0P1>CTH01018BTD0@DE60455AP14
BDd204I;C`<0A4QC00@0ADUD04I:E@16BUD0Ad]F0`18C5L204U=F00404I:E@17BeH0Ad]F04Q<E`80
BDeH00@0BdmJ04a@F`1<D5/0CE5L0P1>DUd304]?FP<0C51K0`1=DE`404iBG@<0CE5L0P1<D5/00`1;
CeX0BTiI04Y>F@0604U=F0<0B4aG1017BeH204I:EAP0B4aG1P17BeH304I:E@<0ADUD0`14B5<304=7
DP@0@TIA3P11AE0?04=7DP0404A8D`15BE@0ADUD04E9E080A4QC0`13Ae8404A9C@<0A4U?0`14B542
04A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool001P0B51K3017CeX304Q@F`@0AdmJ0P18D5/0
0`1:DUd0Be=N04]CGP0<04MDG@`0B5EN00@0C5UQ04YGG@1:Eed0Ae9H0P0nB4X0100b>S`0:2`]02P/
;@0O8B0201TK6P`07QTC00@06QTG01HE4@0F5A409B8M0P0G50/0100J5P/0<B/M034[7@0b;1`2048l
;004040l>`0]:BH0;BTV038]9`808aXA00@09adA03<V5P0c9QH09aP50P15=R<0101VBBT0JDPY06U8
:@1]ARL207569P0407A79P1eBRL0MDXW07E=:@80M4hW00@0NddY07a>:P1lCRX0ODl[0P1nD2`01020
DRh0PU@`089D<024ES8208EG<`0408UG=02<FSL0S5Xg08mM>P<0T5hk0P2BH3d0102EHd00V6I309QV
@`2KIT@209aWA@0409iYA`2PJdT0X6]90:9]B`80YG1>00<0YW5?0:MbD02WLU001`2ULe820:MeE004
0:QfE@2YMeH0ZGMF0:AfE080YgUG0`2ZO5X30:UkF@030:YlFP2[OE/0ZgeK00<0/7iK0`2aOe`0102`
OU/0[G]H0:ekF02[NEH20:UgE0040:ajE`2^O5T0[WaI0;:0G@<0]89O0P2ePf00102fQ640]hER0;N5
HP2cQF420;:4H0040;63G`2`PUh0/89N0;63G`80/XAP00@0]HMS0;J8I02fR6@0^8YV0P2gRFD30;J8
I0040;R:IP2gRFD0]hUU0;>5H@80/89N302fQ5l0102fQEd0]HAL0;F4G02dPe/20;>2FP030;:1F@2a
P5P0/H1H00<0/7mG00<0[GQ>0:]fC02[MT`00`2[M4/0102_MDd0/7E=0;1eC@2^Ld/20:aaB@@0YFXn
0P2UJ3`00`2SI3T0XF8g0:5R=`030:9Q=`0409eU=P2IHC80VF4b099L;@80TU`^00@0U5lc09ER=@2E
HSD0UF8g0P2CH3D0102CHcd0RU`h08YL>024FCL207QA<P0406U7;01VB300ITP`0656<@80DSTU00@0
HTm004dj:`1=>R/0;a`=0P18=bD01015=R40ASLR04Hg8P0h;1@2048f7P0404Hl8`19?bH0BClV03`a
6`80Ac`V00@0F4di06QMB@1XGDT0N6eK0P1YGT`0101KBc@0I5@m06AD?@27N64208AeGP0404Lk8`0`
9Pd0<2H=04Pn9@80ET`c00@0:bPG02l/7@0_;1d0?cda0P19BD400`13AD00@DE60455AP0304M<D0<0
A4QC00@0ADUD04I:E@16BUD0Ad]F0P18C5L304U=F00404I:E@17BeH0Ad]F04Q<E`80BDeH00@0BdmJ
04a@F`1<D5/0CE5L0P1>DUd304]?FP<0C51K0`1=DE`404iBG@<0CE5L0P1<D5/00`1;CeX0BTiI04Y>
F@0604U=F0<0B4aG1P17BeHI04Q<E`D0Ad]F0`16BUD304E9E0<0A4QC0`13Ae860496D@`0@DE@3`13
Ae801014B5<0ADUD04E9E015BE@204A8D`<0@dMB1014BDd304A9C`<0A4QA0P14B5<01@17Ae<0B4QB
04Q8DP18BDh0BDY>0040oooo000H04Q@F``0AdmJ0`18D5/404M?FP80B51K00<0BU9M04]CGP1;Deh0
3017E5d<04QEGP0404]HH01;F5h0BeQN04YEF`80@4Y<00@0<SXl02L[;00W:b`0828Q0P0L7Qd<01hI
4`0401`K5`0H5a<061LC02@Q70806AH?00@06QH;02hX700^:1`0<2XL0P0o>B/0100l>3D0<B`X034/
:00f;bT202@K4P0402@J3P0`91@0<2@D02PK2P80@34N00@0GT<V06=49`1SA2L0JDHX0P1`AbT0101c
B2P0M4TW07A99`1dC2P207E?:P@0NddY0P1lCRX0101mCb/0P58^081B;P23EC4208AF<P0408QF<`2;
FCH0ReTf08iL>@<0T5hk0P2BH3d0102EHd00V6I309QV@`2II48209YU@`0409]VA02NJDL0WVU70:1[
B@80XVe;00<0Y6m=0:E`CP2UL4h01P2SLE00102TLU40YWAC0:IdD`2XMUD20:UgEP040:AcDP2WMUD0
YgIE0:YiF080ZgYI0`2ZNEP00`2[NUT0[7]J0:akFP060:mmFP040:ilF@2]NeP0[G]H0:ekF080[7YG
0`2^O5T0102`OU/0/X1M0;:0G@2ePf030;N5HP80]XAQ1P2`PUh40;63G`80/XAP00<0^HaU0;R;I02h
Rf@00`2gRV<0102hRf@0]hYS0;N:H`2dQf020;:5GP`0]H=N00@0^8MO0;N6GP2gQUh0]HAL0P2cPUX0
0`2aP5P0/7mG0;1oE`030:mnEP040:iiC`2/Mdd0[7M=0:]dB`80[7E<00@0[gE=0;1eC@2`MDd0[W=;
0P2/LDT0102VJcl0Y6Tm0:AY?@2RIc/20:=V>P040:AU>P2SI3T0Xf@i0:=R>080XV4g00@0VV8c09QP
<@2HH340U5h_0P2CGBl0102CGS80Tf0c09=P<`2EHSL209MT>@0409a/AP2;GCT0Redi07aA;`80M4d^
00@0KT`a06a>=P1/CSH0Gd@_0P1<<al0101SD440@2dN040]7P0X5@H203`[6@04048c7P1>?bX0CSlZ
03T]5@80>blG00@0Bd4X04<i8013>B00<RP?0P1HCCL0101ZGdT0L6EA071UD@27O6P205mD@P0407AT
C@1[Fd@0Je]4075RB`80KEi700@0D4@/04Dk8P15>b80AC/R0P0f;1<0100a;Qd0>cPY03/h:@1;BCd2
04i>AP0404E7@P15BDX0ADU:04U>DP80ADY>00@0A4QC04E9E015BE@0ADUD0P16BUD00`17BeH0B4aG
04Q<E`0304U=F0<0Ad]F00@0B4aG04U=F019CEP0BdmJ0P1<D5/304eAG0<0BdmJ0`1<D5/304eAG0@0
CU9M0`1=DE`204a@F`0304]?FP1:CUT0BTiI00L0BDeH0`18C5L504M;EQ`0B4aG0`17BeH504I:E@@0
A4QC0`13Ae850496D@`0@DE@3`13Ae801014B5<0ADUD04E9E015BE@204A8D`<0@dMB1014BDd304A9
C`<0A4QA0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo000H04Q@F``0AdmJ0`18D5/404M?
FP80B51K00<0BU9M04]CGP1;Deh03017E5d<04QEGP0404]HH01;F5h0BeQN04YEF`80@4Y<00@0<SXl
02L[;00W:b`0828Q0P0L7Qd<01hI4`0401`K5`0H5a<061LC02@Q70806AH?00@06QH;02hX700^:1`0
<2XL0P0o>B/0100l>3D0<B`X034/:00f;bT202@K4P0402@J3P0`91@0<2@D02PK2P80@34N00@0GT<V
06=49`1SA2L0JDHX0P1`AbT0101cB2P0M4TW07A99`1dC2P207E?:P@0NddY0P1lCRX0101mCb/0P58^
081B;P23EC4208AF<P0408QF<`2;FCH0ReTf08iL>@<0T5hk0P2BH3d0102EHd00V6I309QV@`2II482
09YU@`0409]VA02NJDL0WVU70:1[B@80XVe;00<0Y6m=0:E`CP2UL4h01P2SLE00102TLU40YWAC0:Id
D`2XMUD20:UgEP040:AcDP2WMUD0YgIE0:YiF080ZgYI0`2ZNEP00`2[NUT0[7]J0:akFP060:mmFP04
0:ilF@2]NeP0[G]H0:ekF080[7YG0`2^O5T0102`OU/0/X1M0;:0G@2ePf030;N5HP80]XAQ1P2`PUh4
0;63G`80/XAP00<0^HaU0;R;I02hRf@00`2gRV<0102hRf@0]hYS0;N:H`2dQf020;:5GP`0]H=N00@0
^8MO0;N6GP2gQUh0]HAL0P2cPUX00`2aP5P0/7mG0;1oE`030:mnEP040:iiC`2/Mdd0[7M=0:]dB`80
[7E<00@0[gE=0;1eC@2`MDd0[W=;0P2/LDT0102VJcl0Y6Tm0:AY?@2RIc/20:=V>P040:AU>P2SI3T0
Xf@i0:=R>080XV4g00@0VV8c09QP<@2HH340U5h_0P2CGBl0102CGS80Tf0c09=P<`2EHSL209MT>@04
09a/AP2;GCT0Redi07aA;`80M4d^00@0KT`a06a>=P1/CSH0Gd@_0P1<<al0101SD440@2dN040]7P0X
5@H203`[6@04048c7P1>?bX0CSlZ03T]5@80>blG00@0Bd4X04<i8013>B00<RP?0P1HCCL0101ZGdT0
L6EA071UD@27O6P205mD@P0407ATC@1[Fd@0Je]4075RB`80KEi700@0D4@/04Dk8P15>b80AC/R0P0f
;1<0100a;Qd0>cPY03/h:@1;BCd204i>AP0404E7@P15BDX0ADU:04U>DP80ADY>00@0A4QC04E9E015
BE@0ADUD0P16BUD00`17BeH0B4aG04Q<E`0304U=F0<0Ad]F00@0B4aG04U=F019CEP0BdmJ0P1<D5/3
04eAG0<0BdmJ0`1<D5/304eAG0@0CU9M0`1=DE`204a@F`0304]?FP1:CUT0BTiI00L0BDeH0`18C5L5
04M;EQ`0B4aG0`17BeH504I:E@@0A4QC0`13Ae850496D@`0@DE@3`13Ae801014B5<0ADUD04E9E015
BE@204A8D`<0@dMB1014BDd304A9C`<0A4QA0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo
000H04Q@F``0AdmJ0`18D5/404M?FP80B51K00<0BU9M04]CGP1;Deh03017E5d<04QEGP0404QEG@1:
Eed0BUMM04YEF`80@D]=00@0<SXl02HZ:`0V:R/07b4P0P0M7ah<01hI4`0401/J5P0H5a8061LB028O
6P806QL@00@06aL>02/T6P0[91X0;bPN0P0m=b/0100h<bl0=2dW03@]9`0g<2P202LN5@0402@J3P0_
8aD0;b<E02`P4080>B`K00@0ED0U05e3:01M@bP0IdP/0P1^BBh0101`AbT0L4PU07189@1`BRD2079<
9`0407]=:@1jC2P0NT`X07U;9`80NT`X00@0NddY07i@;01nD2`0PE<_0P23EC401028ES<0ReTf08]I
=P2>G3T5091N>`04099P?@2DHSl0U68o09IQ?`80Uf9000@0VFA209]VA02KIT@0WVU70P2PJdT00`2Q
K4X0XVe;0:9]B`060:5_CP040:9`C`2TLU40Y79A0:IdD`80YgED00<0YWED0:UhE`2YN5L00`2[NUT0
102ZNEP0ZGQG0:UhE`2YN5L20:]jF@L0[G]H0P2^O5T30:mmFP040;B2G`2bP5d0/X1M0;1nF`80/GmL
00@0]89O0;F3H02ePf00/h5N0P2aOe`0102_PEd0/89N0;22GP2aPel20;:4H0040;>5H@2bQ600/XAP
0;63G`80/89N00@0^8]T0;N:H`2gRV<0]hYS0`2fRF820;F8H@030;>6G`2bQEh0/XEN00<0]hEP00@0
]XAO0;F3GP2ePeh0]H=N0P2dPUd30;>1G0@0^8AL0P2gPe/0102ePET0/WiF0;9nEP2`O5@20:ijDP03
0:ehCP2[MT`0ZgI<00<0ZW=:00@0[G=;0:icB`2^Ld/0[7590P2ZKdL0102UJSh0XfPl0:=X?02QISX2
0:AW>`040:MX?@2VIc`0YVLl0:=R>080Wehd00@0Uel`09QP<@2HH340UV0a0P2DGS00102BGC40TUlb
099O<P2FHcP209UV>`0409a/AP2<GSX0S5hj07eB<080Me0a00@0ME<h06mA>@1_DCT0GT<^0P1<<al0
101GA3D0>RLH03XW600d8A8203`[6@0403X[5P16=b80ASLR038V3P80>RhF00@0ACTO03hb600n<QP0
>blG0P1QECd0101`I4h0Kf==06mSC@1kKU/205iA?P0406QH@@1_GdP0Kem8061A>P80DT</00@0Cd<[
040f7@10=Qd0=2XA0P0n=1/0100e<R40AD8c04E2<`1DDTH2051@B00404E7@P18C4d0B4a=04aAE@80
@dQ<0`15BE@01016BUD0Ad]F04M;EP18C5L304U=F080BTiI00@0Ad]F04Q<E`18C5L0B4aG0P19CEP0
101;CeX0C51K04a@F`1<D5/204eAG0<0BdmJ0`1<D5/304eAG0@0CU9M0`1=DE`204a@F`0304]?FP1:
CUT0BTiI00<0BDeH0`1:CUT304U=F0<0B4aG0`17BeHL04Q<E`D0Ad]F0`16BUD304E9E0<0A4QC0`13
Ae830496D@`0@DE@3`13Ae801014B5<0ADUD04E9E015BE@204A8D`<0@dMB1014BDd304A9C`<0A4QA
0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo000H04Q@F``0AdmJ0`18D5/404M?FP80B51K
00<0BU9M04]CGP1;Deh03017E5d<04QEGP0404AAF@18EE/0B5EK04YEF`80@Ta>00@0<c/m02HZ:`0V
:R/07R0O0P0K7A`<01hI4`0401TH4`0F5A005QD@01hK5P806aPA00@07APB02PQ6`0X8A/0;bPP0P0m
=Rh0100o=S40>S4Z03Xa:P0h;bP202`S6P0402DK4@0Y7a<0:AlC02XQ4P80;RDF00@0CC/S05Q2:P1H
@RX0IdTa0P1^BS00101^AbX0K4<S06a38`1]Ab<2071:9@0307Y<:01iBbL0ND/W00<0ND/W00@0NddY
07e?:`1mCb/0Oe4]0P21Dbl01028ES<0ReTf08]I=P2>G3T208mM>P0408iL>@2=FcP0SE/h08iL>@80
Sedj00@0U5lm09EP?P2EH3h0Uf900P2II4800`2LIdD0WVU709iYA`030:1[B@030:1^C@2OKD`0Wfe<
00<0Wfe<00@0X6i=0:9`C`2RL4l0Y79A0P2ULe80102ZN5L0[7YI0:ajF@2^O5/20:ekFP040:]iF02Y
MeH0ZGMF0:UgEP80ZWQG00@0[7YG0:]iEP2[NEH0ZgUF0P2/NUL0102]NeP0[WaI0:ilF@2_OEX20;1n
F`040;B2G`2aOe`0/GmL0:mmFP80/7iK00@0/h5N0;B2G`2dPUl0/h5N0P2`OU/0102_PEd0/H=O0;63
G`2cQF430;F7H`80]8IR00<0/hEQ0;63G`2aPel0102eR5l20;>6G@030;:5G02aQ5/0/HAK00<0/HAK
00@0^XQS0;V7HP2iQf80^8IQ0P2fQ5l0102cPE`0/GmJ0;5oFP2`OUT20:mmF0040;B0F02fPUX0]X9J
0;R4G080^HEM00@0]h=K0;=oE`2cOeL0[GUA0P2ZMTh0102/Mdd0ZWE;0:YeB`2YLTT20:QaB0040:]a
B@2/LDT0[7590:Y_A`80Z6e500@0XfPl0:9W>`2RIc/0XfPl0P2WJSh0102ZJd00YfPm0:MX?@2QH3H2
09]J<00409IN;`2HH340V60a09MQ<P80UEla00@0TE``095N<@2AGS40UF8g0P2JIc`0102EICl0SElk
08eO>`23F3H207mH>@0407]I?P1^D3P0KU0h05Xo:P80CSDQ00@0@c0Q03PU5P0h9AH0?b`M0P0o;Q`0
100i:QD0?RlJ03h_6P0_8`/203da6@0403la5P0k;A<0>bdC04Tk8@80HU<l00@0MFI?07IWDP1fIe80
K5e80P1YFTD3065A>P0404dn9`11<Q/0@C8K040d7080=bdD00@0>c4H05M==01GCC@0;b`K0P0n>b`0
101=Bcl0BDU104U9@@10@Sd204M;C00404aAE@12Ad/0@TM;04E9E0<0ATYE0P17BeH00`18C5L0BDeH
04U=F00304Y>F@<0B4aG00@0BDeH04Y>F@1:CUT0BTiI0P1;CeX304a@F`<0BdmJ0`1<D5/304eAG0@0
CU9M0`1=DE`204a@F`0304]?FP1:CUT0BTiI00<0BDeH101:CUT304U=F1d0B4aG0`19CEP404Q<E`D0
Ad]F1015BE@304A8D`D0@dMB3011AE0?04=7DP0404A8D`15BE@0ADUD04E9E080A4QC0`13Ae8404A9
C@<0A4U?0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool001P0B51K3017CeX304Q@
F`@0AdmJ0P18D5/00`1:DUd0Be=N04]CGP0<04MDG@`0B5EN00@0A55I04QEF`18EE/0BUEK0P12C4h0
100c>cd09RX[02HZ:`0N81l201/M70`07QTC00@06APC01HE400F5A007Q/F0P0K6140100M6180:24K
02PQ6`0_:20203df;P0403lf<@0j<BX0>S4Z03P_:080;2<J00@09A/A02TO4`0Y7a<0:R4B0P0^9AH0
101=>b<0F48Z05Q2:P1WBC4206i:<00406i7:P1/@b<0K4<S06e78`80L4XU00<0NT`X07U;9`1iBbL0
0`1iBbL0101kCBT0ODl[07e?:`1oDBd2085C;`0408QF<`2;FCH0ReTf08iL>@80Sedj00@0SU`i08eK
>02=FcP0SU`i0P2?GCX0102DGcd0UF0n09EP?P2GHT0209UT@P0309aWA@2NJDL0WVU700<0X6]900<0
X6i=09m]C02OKD`00`2OKD`0102PKTd0XW1?0:9`C`2TLU420:EcDP040:YhE`2/NUT0[7YI0:ilF`80
[G]J00@0ZgUH0:UgEP2YMeH0ZGMF0P2ZN5L0102/NUL0ZgUF0:]iEP2[NEH20:ajE`040:ekF02^O5T0
[WaI0:mmFP80/7iK00@0]89O0;5oG02aOe`0[geJ0P2`OU/0102cPEh0]89O0;B2G`2cPEh20;1nF`04
0:n1G@2aPel0/H=O0;>5H@<0]HMS0P2dQV800`2cQF40/H=O0;63G`040;F8G`80/hIM00<0/XEL0;64
F`2aQ5/00`2aQ5/0102jR6<0^HMR0;V7HP2hQV420;J4G`040;>1G02aOeX0/GmJ0;1nF@80[geH00@0
]81H0;J2FP2fPUX0^8AL0P2iQEd0102gPe/0/gmG0;=oE`2]NE420:YfCP040:agC@2ZMD/0ZWE;0:Ub
B@80Z75800@0Zg590:aaB@2/LDT0ZVm70P2XKDD0102SJ3`0XVLk0:9W>`2SJ3`20:MZ?P040:Y[@02W
J3d0YfPm0:5P=P80VeX`00@0UUh_09QP<@2HH340Uf4b0P2EGc40102AG300TEha095N<@2EHSL209YW
?00409EU?`2=Gc/0SElk08=H=P80OePi00@0NeTn06i@>01^D3P0FSlZ0P1>=B401013<240>2DF03PU
5P0o;1d203l^700403TZ5@0n;aX0?RlJ02lS2`80?C4I00@0?c4F03/]4`0k;A<0BC/Q0P1RDc`0101e
ITl0MVMB07IWDP1/GDP206UJA@<0HE4j00@0CChW044b6`11<Q/0@3@L0P0g;A@0100k<AP0Eddd05M=
=00_;1/203hk;00404e;?`19BD40BDU10412?@80Ad]<00@0C55E0497B`12Ad/0ADUD0`16BUD204M;
EP0304Q<E`19CEP0BDeH00<0BTiI0`18C5L01019CEP0BTiI04Y>F@1:CUT204]?FP<0C51K0`1;CeX3
04a@F`<0CE5L101>DUd304eAG080C51K00<0BdmJ04Y>F@1:CUT00`19CEP404Y>F@<0BDeH7@18C5L3
04U=F0@0B4aG1@17BeH404E9E0<0A4QC1@13Ae8<0455D0l0@dMB00@0A4QC04E9E015BE@0ADUD0P14
B5<304=7DP@0A4U=0`14BDl304A8D@80A4QC00D0AdMC04Q8DP18B580B4U>04U:CP010?ooo`006018
D5/<04M?FP<0B51K1017CeX204Q@F`0304YBG@1;Deh0Be=N00`0AeAM3018EEh01013D5P0AU=I04IC
F@1:EE/204E?D@0403Lo@@0Z;Rl0:Rh_01lQ808061XI300N6A<00`0H5a805A@?01DD3`0301`I5004
01hI5@0V7Q/09QhK02hV8`80>c<`00@0B3dg03le;00o=B`0>2hU0P0d:R00100Y8AH09Ad@02DM400W
7a8202@L3`0404Df8@1A?b/0DCl[06=8<`80Jd/b00@0JdH[06Y3901Z@b@0JdDQ0P1_BB@307U;9`04
07Y<:01kCBT0NddY07a>:P80ODl[0`1nD2`01025Dc00R5Hc08QF<`2<FSL208eK>00408aJ=`2;FCH0
ReTf08]I=P80S5Xg00@0Tehl09AO?@2DGcd0UV4o0P2HHd40102JID<0WFQ609eXAP2NJDL209mZB0<0
Wfe<102NK4/209m]C0040:1^C@2QKdh0XFm>0:UgEP80ZgUH00@0[G]J0:ajF@2/NUT0ZGMF0`2WME@2
0:QfE@<0ZWQE00@0ZgUF0:ajE`2/NUL0[G]H0P2^O5T30:mmFP040:ekF02/NUL0[7YG0:ajE`80[WaI
00@0/GmL0;>1GP2cPEh0]89O0P2ePf00102^P5`0/89N0;22GP2bQ6020;B6HP<0]HMS0`2dQV80102d
Qeh0]HQO0;F8G`2dQeh20;:5G0040;23FP2_PUT0[h9I0;23FP80/HAK00@0^XQS0;V7HP2iQf80^8IQ
0P2ePeh0102cPE`0/GmJ0;5oFP2_OEP20:ilE`040;=nE02eP5H0]H1F0;N2F080^8=I00@0]h9H0;=n
E02cOU@0[WU?0P2ZMD/0102]N4h0ZWE;0:YeB`2YLTT20:QaB0040:]aB@2/LDT0[7590:U^AP80Yfa4
00@0XfPl0:AY?@2TJCd0YFXn0P2XJcl0102YJSl0Y6Dj0:AU>P2OGS@209]J<00409MO<02IHC80VF4b
09MQ<P80UEla00@0TE``095N<@2AGS40UF8g0P2IIS/0102?GcT0SElk08eO>`28GC/208=L?@0407aJ
?`1XBS80J4Xb05@i9080CSDQ00@0=R<D03TV5`0i9QL0?2TJ0P0l:aT01012<ah0@C8M044b7@0e:A42
03hb6P0403h`5@0k;A80>bdB05129`80HU@j00@0MfQA07i_F01nKeP0KUm:0P1aHTd0101bHT/0C3`U
04`l9@0o<1T204Hg8004048f7P17?B@0AcdT04e3:P80@CLN00@0?3TX04E2<`15@S<0Cde10P19BD40
1010@Sd0Ad]<04M;C01:Ce<20415B@<0ATYE00@0Ad]F04Q<E`18C5L0BDeH0`1:CUT204]?FP@0BDeH
0`1:CUT804]?FP<0C51K0`1=DE`404iBG@<0CE5L0P1<D5/00`1;CeX0BTiI04Y>F@0304U=F00304]?
FP1:CUT0BTiI00<0BTiI0`19CEPK04Q<E`@0BDeH0`18C5L504M;EP0304I:E@15BE@0ADUD00<0ADUD
0`14B5<304=7DP`0@DE@3`13Ae801014B5<0ADUD04E9E015BE@204A8D`<0@dMB1014BDd304A9C`<0
A4QA0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo000H04Q@F``0AdmJ0`18D5/404M?FP80
B51K00<0BU9M04]CGP1;Deh03017E5d<04QEGP0404AAF@17E5X0AeAJ04aGG@80BE=E00@0?TI8030d
=@0`=3D08B<R0P0G6AP<01hI4`0401TI4@0G5`l05aL?01/H4`807AXE00@07QTE028J600R6QP0:b<Q
0P0h<2h0101;@CP0?C<Z03dc:P0c:B0203P^900402lW700V7Q<09QhC02XR5`8091`?00@0>BhL04/j
:01;>RP0H4Lc0P1ZC3@0101ZAR`0JT<T06Y3901]Ab<2071=9`0407M99@1iBbL0ND/W07]=:@80ODl[
0`1nD2`0101mCb/0O4hZ07a>:P1oCBX208=A;P0408QF<`2;FCH0ReTf08]I=P80RUPe00@0ReTf08aJ
=`2<FSL0Tehl0P2DGcd0102EH3h0V6=109QS@@2JID<209aWA@0409iYA`2OJTP0WfY80:1^C@80Wfe<
00<0WVa;09e[BP2MJdX00`2MJdX309i/B`030:A_C`2WLU80Yg9B00<0ZWEE00@0ZGAD0:QcD`2XLe<0
ZGAD0P2ZMED0102XMU<0ZWQE0:YhE@2/NUL20:ilF@030:mmFP2^O5T0[WaI00<0[G]H00@0ZWQE0:]i
EP2[NEH0[7YG0`2]NeP20:mmFP040;>1GP2ePf00]H=P0:anFP80[GmK00@0[h5M0;63G`2aPel0/XAP
0P2dQV830;F7H`040;F8GP2fREl0]XUO0;J9G`80/hIL00@0/8=I0:n2F02_PUP0/HAJ0P2cQU`30;N5
H0040;J4G`2dPUd0]89M0;>1G080/X1K00<0/GmJ0;1nF@2`OUT0102dOeD20;=nE0040;9mD`2`Ne40
/7]A0:mjD080[WU?00@0[gY@0:ehCP2]N4h0[7E<0P2[M4/0102]Ld/0[W=;0:icB`2[L4P20:U^AP@0
Yfa00P2VJCd00`2SI3T0X64f0:1Q=P0309mN=00409YR<`2IHC80VF4b09IP<@80U5h`00@0TUda099O
<P2BGc80U64f0P2FHcP0102?GcT0Sf4m08mQ?@27G3X2081I>P0407ME>P1P@RX0H48Z04/`6`80BC0L
00@0=R<D04<`8@13<240=R<D0P0g9Q@0101:>bH0@c@O04<d7`0i;AD203hb6P04044b5@12<QP0@S8H
0592:080Je]200@0L61707EUCP1eIDh0O6aE0P1^GTL0101IBC80B3PQ04Ph8@1A@R/205I7<004040d
7012>1l0@SPO04e3:P80?S@K00@0CDXi04i;?01>Bc`0DDm30P1<C4@01013AD00ATY;04I:B`17C502
03e2AP<0ATYE00@0Ad]F04Q<E`18C5L0BDeH0P1:CUT304]?FP@0BDeH0`1:CUT804]?FP<0C51K0`1=
DE`404iBG@<0CE5L0P1<D5/00`1;CeX0BTiI04Y>F@0304U=F0<0BdmJ101:CUT504U=F1P0B4aG0`1:
CUT304U=F0@0B4aG0P17BeH304I:E@@0ADUD1@14B5<<0455D0l0@dMB00@0A4QC04E9E015BE@0ADUD
0P14B5<304=7DP@0A4U=0`14BDl304A8D@80A4QC00D0AdMC04Q8DP18B580B4U>04U:CP010?ooo`00
6018D5/<04M?FP<0B51K1017CeX204Q@F`0304YBG@1;Deh0Be=N00`0AeAM3018EEh01014DET0AeAJ
04MDFP1<Eed204UCE@0403i6B00`=3D0<3@e024S8P805aTH300N6A<0100I6A405aL?01LG3`0K61<2
01dJ5@0401hI5@0R6QP08QXH02/S8@80>30^00@0Bd4h03dc:P0m<bX0<bTP0P0h;R@0100_9a`09QhC
02HN4`0Z8QL202@L3`0403T^701;>RP0BcXX0617<`80JT`d00@0JTH/06Y3901Z@b@0KDLS0P1`CBL0
101gBBD0ND/W07U;9`1kCBT207e?:`<0OU0/00@0ODl[07a>:P1lCRX0OddZ0P23DBh01028ES<0ReTf
08]I=P2;FCH208YH=@0408]I=P2<FSL0S5Xg09=N?080U5lm00@0UF0n09QS@@2HHd40VVE30P2LIdD0
102NJDL0WfY809mZB02PKTd209m]C00309i/B`2MJdX0WF]:00<0WF]:0`2NK4/00`2TKdl0Yg9B0:Mb
DP030:YeE@040:UdE02XLe<0Z7=C0:UdE080ZWEE00@0Z7IC0:YhE@2ZN5D0[7YG0P2^O5T00`2_OEX0
[WaI0:ilF@030:ekF0040:YhE@2[NEH0ZgUF0:ajE`<0[G]H0P2_OEX0102cPEh0]H=P0;F3H02/OUX2
0:eoF`040:n1G@2aPel0/H=O0;:4H080]8IR0`2eQf<0102eR5h0]XUO0;J9G`2fREl20;>6G0040;23
F@2_PUP0[h9H0;64FP80/hIL0`2gQF00102fQ5l0]89M0;B2G@2cPE`20;:0F`030;5oFP2`OUT0/7iI
00@0]7mE0P2cOU@0102bOE<0/7]A0;1kD@2_NU020:iiC`040:mjD02]N4h0[GQ>0:aeC080ZgA;00@0
[G=;0:icB`2^Ld/0Zg180P2YKTH40:M/@080YVTm00<0Xf@i0:1Q=P2PHCH00`2OGS@0102JHS<0VF4b
09UQ<P2FH34209AN<004099M<@2BGc80TUlb09AQ=P80UV<h00@0Seli08mQ?@2?HCd0Qe`j0P20FCX0
101gECX0H48Z0612:P1;<1/204T`700403HS5013<240@c0Q03HS5080=bHD00@0BS/V04<d7`13=1l0
>BdE0P0n<QX01011<QD0@S8H048b601B@RP206]K@P04071PA`1eIDh0MFE>07a/E@80KUi700@0FDTb
04Ph8@18>240DD8[0P1FAc001010=1`0@SPO048h7`1=@bX203hd6`0404e:>@1>Bc`0CT/l055?@`80
C4a400@0@dE004I:B`16BT/0Ada@0P0m@TH304I:E@0404M;EP18C5L0B4aG04U=F080BTiI0`1;CeX4
04U=F0<0BTiI201;CeX304a@F`<0CE5L101>DUd304eAG080C51K00<0BdmJ04Y>F@1:CUT00`19CEP3
04]?FP@0BTiI1@19CEPH04Q<E`<0BTiI0`19CEP404Q<E`80Ad]F0`16BUD404E9E0D0A4QC3011AE0?
04=7DP0404A8D`15BE@0ADUD04E9E080A4QC0`13Ae8404A9C@<0A4U?0`14B54204A8D`0504M7D`18
B580B4QB04Q9CP19BTh00@3oool001P0B51K3017CeX304Q@F`@0AdmJ0P18D5/00`1:DUd0Be=N04]C
GP0<04MDG@`0B5EN00@0AU=K04UFG019EU`0CUUO0P1=EeT01014C4h0=CTj03Di>P0T9RD201LI60`0
7QTC00@071`D01TI4@0I6A406aPC0P0N6aH0100N6AH07aLE01lG5@0Y824203@[;00404Tl<`0h:b80
>2/R02`R6080>S0V00@0=BdR02XR5`0Z8QL0<2XN0P0Y8aL0100a:1L0ASHV04Hf9P1LAC<206U:=@04
06]7;@1YABD0JDDU06m99@80Le0Z00@0MdTU07U;9`1iBbL0O4hZ0P1oDBd01020DRh0Oe4]07mA;@1l
CRX207Y<:00407Y89@1oCBX0OddZ08EC<080RELd00@0RUPe08]I=P2;FCH0S5Xg0P2>G3T0102CGS`0
U5lm09AO?@2EH3h209QS@@0409YU@`2LIdD0W6M509iYA`80WfY800@0X6i=09m]C02OKD`0WVa;202L
JTT0102NJDT0XFa<0:5/C02UL5030:IaD@80Yg9B00@0Z7=C0:YeE@2ZMED0YgEB0P2YMe@0102/NUL0
[geJ0:mmFP2_OEX20:ilF@040:ajE`2[NEH0ZgUF0:ekF080[WaI00<0[geJ0:ekF02]NeP00`2[NEH0
0`2^O5T0/GmL0;5oG0030:YlF0040:anFP2^P5`0[X1L0;22GP80/XAP00@0]8IR0;F7H`2eQf<0]hYP
0P2hRf40102gRV00]HQN0;F8GP2aQ5X20;23F@030;>6G02fREl0]XUO00H0]89M1P2cPE`0102eP5H0
/giD0;=nE02`Ne420:iiC`040:ehCP2^NDl0[WU?0;1kD@80/GaB00@0/WeC0;1kD@2`Ne40[WM>0P2]
MTd0102`MTh0/7E=0;1eC@2]LTX20:]`B0<0ZVm300@0Yfa00:AW>`2TIc/0Wf0e0P2LGC80102OGS@0
XV4g0:9Q=`2LI3D209YR<`0409EO<02DGS00U5h`099M<@80TUlb00@0Tf0e09AQ=P2DHCH0Tf<m0P2?
HCd01024FCL0Ne@e07]D=@1cDCH205/m9@0404@Y5013:QH0@bXF03XW6080CS//00@0=R<D03@S4@0d
8a40C3dX0P10<A`0100i;AD0?30H03``6010<A@204Pi700405A4:P1dI4X0M6A:06=C>P80H50i00@0
Q7AM06EE?P1UECh0?BdF0P17=b00101=?RL0BCXS04Tj8`12=Qh204Pn9@040516;@1?AB`0CdD/04Q5
=080AD8c00@0AT@h04A4?014A3`0?d4l0P13AdP01015BTh0?D9603e2AP16BUD304M;EP80B4aG00<0
BDeH04Y>F@1:CUT00`1;CeX304U=F0H0BTiI1P1;CeX304a@F`<0CE5L101>DUd304eAG080C51K00<0
BdmJ04Y>F@1:CUT00`19CEP404]?FP<0BTiI1@19CEPH04Q<E`<0BTiI0`19CEP604Q<E`@0ATYE0`15
BE@504A8D``0@DE@3`13Ae801014B5<0ADUD04E9E015BE@204A8D`<0@dMB1014BDd304A9C`<0A4QA
0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo000H04M?FP@0B55J0`17D5T504I?F0<0Ae1I
1018DEX504UBF``0Be=N301:DUd01016D5X0A55J04AAFP16EE`204QGG00404ABE@0h@T<0>49302T]
;P807QhP00@08A`H01`G4`0L5a<07QTE0P0M61@0100G4Ph06QDA01XE4@0O6QH201/F4P0401/J600I
61H06APF01/G50807QXG00@07aXF01hG4@0N5a408Q/E0P0X8A/0100i<RX0@3T_040i;`0h;bH203<[
800403X`9P0`9QX0<2HJ030T6080ASX^00@0<R8B03TX600i:1P0D3hZ0P1QBcH0101PARl0I4H/06A6
;01]CS8206i<<004079:9P1eBRL0MDXW07I;:080NDdX00<0NdlX07e@:@1mD2T00`1oDRT0101mDB`0
OU8]07iB;@1nDRd207mC;P03081D;`22ES40PUHa00<0PeLb00@0RELf08eK>P2=FcX0TEln0`2FHD42
09QQ@P0409UR@`2KHT@0Vf9409aWA`80WFQ800@0WVU909eXB02MJ4P0VfI60P2JIDD0102LIdL0WVU9
09iYB@2XKDl30:9WB@80ZVmA00@0[gAF0:ebE02]LU@0[G9D0P2`MEL<0:ehEPH0/7]I1P2_NUP0102X
Le@0[7MG0:agE`2aO5`20;=nG0040;9mF`2aOEP0/GeH0;9nF@80/gmJ00@0]X9M0;N3GP2gPeh0^HEP
0P2jQV40102iQF00]h=N0;N3GP2dP5/20;=oFP<0]H9G0`2fPeP0102ePUL0]85F0;B1EP2bOe@20;5n
D`040;B0F02aOED0/GeE0:eiD@80ZgI<00@0ZGA:0:YdB02ZM4P0ZgE90P2^MDP30;1eC@040:mdC02^
Ld/0[W=;0:icB`<0[G9:0P2/LDT30:M]?`040:E[?@2UJ3/0YFPk0:AU>@80XV<g00<0XV8f0:5Q=@2Q
HCD00`2EFbd0102DG2d0TU`^099L;P2AG302091P<P04099Q=P2BHcL0TV<g099U?P80Relj00@0Q5hj
079?<@1bCc40JT`b0P1RBC<01011;a/0<B0>034P3P0n<2D202hP5@0402lQ5011=2@0@C@T048e8P80
=BTC00@0>RhF04i2:P1>@RX0BCXM0P1GAbd0101SDcT0EDD/05E5;01MCC@205];=004069B>`1MCCH0
GDdf038R2`80D40Y00@0ACHQ03PY500h:A@0@C@Q0P1?A380101FBcT0C44_04a1;`12?bh2044n;`04
04I4>013@c/0@d<k03Xl=`80@4A500@0Ada?0415B@10ADT0@dMB0P15BE@304Q<E`<0Ad]F00<0BTiI
04a@F`1<D5/0301;CeX404Y>F@80C51K00@0Ce=N051DG`1@E5l0CU9M0`1<D5/204eAG00404iBG@1;
CeX0BdmJ04Q<E`80ATYE00<0Ad]F04U=F019CEP00`1=DE`304a@F`0304]?FP1:CUT0BTiI00<0BTiI
3019CEP<04Q<E`<0C4aH0`1;BeL304Y:EPL0BDUE0`18B5@504M7D`0404U9E@17Ae<0AdMC04E5D@<0
@d=?0P14A5000`15AE40AdMC04M7D`0304M7D`<0ATIB0`15AE4304A4D0<0ATIB00@0AdMC04Q8E018
B5@0B4QD0P17Ae<304I6DP@0A4U=0`14BDl304A8D@80A4QC00D0AdMC04Q8DP18B580B4U>04U:CP01
0?ooo`006017CeX404QAFP<0Ae1I1@16CeP304M@F@@0B55J1@19DU/<04]CGP`0BU9M00@0AU1J04AA
FP14DEX0AUEL0P18Ee`01014DUD0>49303Q2@`0Y;Bh201hN8004024L600L5a<071LC01hI5@807APD
00@05a8>01XE4@0J5A407aXF0P0K5Q80100K6QP06APF01TH5P0K5a@201hJ5`0401lJ5P0N5a407QLA
028K5@80:24K00@0>C8Z040i;`10>Bl0>2lV0P0c:b00100j<2H0<2HJ030V6P0`91P204Hj;P04038R
4P0i:1P0>BPH050n:P80HD/f00@0H4H_06A6;01TAR`0KDhb0P1^C300101bBRH0MDXW07E:9`1fBbP2
07U=:00307]?:01mD2T0OE0Y00<0Oe8Y00@0OE4/07iB;@1nDRd0OU8]0P1oDbh00`20E2l0PUHa089F
<@0308=G<P0408UG=P2=FcX0SE/j095O?P<0UV510P2HHD80102IHT<0Vf9409]RA02LIdL209eXB004
09iYB@2MJ4P0WFQ809]VAP80VVE500@0W6M709iYB@2NJDT0Z6e?0`2RIdT20:Y_D@040:mdEP2]LU@0
[G9D0:ebE080/7EG302]N5H60;1kF@H0[gYH00@0Z7=D0:agE`2/MeL0/GaL0P2cOU`0102bOE/0/GeH
0;5mF02bOUT20;=oFP040;J2G@2gPeh0]h=N0;V5H080^XIQ00@0^HEP0;N3GP2gPeh0]81K0P2cOeX3
0;F2E`<0]X=H00@0]H9G0;B1EP2dPEH0/WmD0P2aOU<0102dP5P0/GeE0;5mE@2]NE420:]fC0040:Ud
BP2ZM4P0ZWA80:]eB@80[WE80`2`MDd0102_M4`0[W=;0:icB`2^Ld/30:ebBP80[7590`2WKCl0102U
Jcd0YFPk0:EX>`2TICT20:9S=`030:9R=P2QHCD0XF4e00<0UE/]00@0U5`]099L;P2BG2h0TE``0P2@
H380102BHCH0TV<g099S=`2BICh208]O>P0408AN>P1bCc40LTla06Y<<P80HTTc00@0@BlK034P3P0a
80h0?S0U0P0^81D0100_8A@0@C@T044d9012=B8203DY4`0403X^5P1>@RX0CT8Z04Tj7@80EdL]00@0
He<i05E5;01EAB`0GDdd0P1KBc@0101RDS/0GDdf05e==P0b8P/20510:@0404Df8@0h:A@0>2TD044d
8@80Cd@b00@0ET/i04a1;`1<@Bl0@Sl^0P11?Rl01016A3P0@d<k04=3>`0j?3L20414A@0404M<C`10
ADT0@4E904=7DP80ADUD0`18C5L304M;EP0304Y>F@1<D5/0C51K00`0BdmJ101:CUT204a@F`0404mC
GP1@E5l0D5AO04iBG@<0C51K0P1=DE`0101>DUd0BdmJ04]?FP18C5L204I:E@0304M;EP19CEP0BDeH
00<0CE5L0`1<D5/00`1;CeX0BTiI04Y>F@0304Y>F@`0BDeH3018C5L304a<F0<0Bd]G0`1:BUH704U9
E@<0B4QD1@17Ae<01019BED0AdMC04M7D`15AE4304=3C`80A4A@00<0ADEA04M7D`17Ae<00`17Ae<3
04I6DP<0ADEA0`14A50304I6DP0404M7D`18B5@0B4QD04Q8E080AdMC0`16AU8404A9C@<0A4U?0`14
B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool001P0B51K1018DEX304M@F@D0ATmH0`17
D5T404QAFPD0BE9K301;Deh<04YBG@0404MAF`14DEX0A55J04EDF`80AeIK00@0AE=F03]5AP0kADH0
;30a0P0P8280100Q71P06aHB01/F4P0M61@201hI5@0401PC3`0J5A406QDA01lJ5P806aHB00@06aXH
01TH5P0I61H071PE0P0O6aP0100O6QH07QLA01hG4@0Q6Q@202PQ6`0403@]9@13?380@c`b03lf;@80
=blT00@0=bdS02hT600^91P0<BDI0P17>bl0100e9AH0<B0@034P4015<R4205m9=00406A:<`1PABX0
H4DZ06U:;P80Ke0c00@0LTXV07=;9`1cBbL0MT/X0P1hCRP0101kCbX0OU4Z07iA:P20Db`2085D:`04
07a@:`1mDB`0OE4/07mC;P80PED`1022ES42085E<00408QG=P2;FST0ReXi095O?P80UF1000@0Uf92
09UR@`2IHT<0W6E60P2OJ4T00`2JIDD0WFQ809eXB00309mZBP0409eXB02LIdL0W6M709aWA`80WFQ8
00@0WfA60:5VB02QITP0YFY<0P2YKU00102XKDl0Y6U;0:AYB`2RIdT20:=XBP<0Yg9@00@0ZGAB0:]f
E02[MU@0[GQF0P2^NEL00`2`NeT0/GaJ0;5lFP030:iiE`040:mjF02`NeT0/7]I0:mjF080[GQF00@0
ZgID0:UdDP2YM580[7MG0P2_NUX0102bOEd0/WeK0;9mF`2`NeT20;1lE`040;=oF02fPU/0]X9K0:mk
E080/7aE00@0/WiG0;=oF02cOeP0/WiG0P2`O5D0102^NU<0[7QA0:ahD@2cP5D20;B1EP@0]H9G0P2d
PEH0102cP5D0/WmD0;9oE02cOeL20;9nEP040:mkD`2]N4h0[GQ>0:afBP80ZgE900@0[GA70:ieB02^
MDP0]7UA0P2aMTh00`2^Ld/0ZVm70:Y_A`030:Q]A@<0ZFi600@0Y6Xl0:=Y>`2SJC/0XVPj0P2SIST0
0`2RHcL0X64e0:1Q=@030:1P=0<0V5h`00@0UUh_09=M;`2CGBl0TUda0P2?Gc40102@Gc@0T64e091Q
=@2FJD8208aR?00408EO>`1_CRl0Kdh_0615:P80ET0Y00@0?2XF038Q4@0b8A40>bdR0P0/7Q<0100[
7A00>RdM03X]7@10<b0203T]5`0403da6@1;?bL0BclW04<d5`80G4`b00@0ETH/04@d6`14=1/0H50g
0P1HB340101=?BH0E4@]05A4;@19>B8204hn9`0404@e801<?BP0C3dX0513<0<0AS/Y0P13>2H01011
?Rd0@d0a04=0<@1:B3`204U9@@040453@015BDX0ADU:04U>DP80@TM=00<0A4QC04I:E@16BUD00`18
C5L304M;EP0304Y>F@1<D5/0C51K00`0BdmJ101:CUT204a@F`0404mCGP1@E5l0D5AO04iBG@80C51K
00@0BdmJ04a@F`1<D5/0CE5L0P1;CeX01018C5L0Ad]F04M;EP18C5L204Y>F@@0C51K0`1;CeX504Y>
F@`0BDeH3018C5L304a<F0<0Bd]G0`1:BUH704U9E@<0B4QD1@17Ae<01019BED0AdMC04M7D`15AE43
04=3C`80A4A@00<0ADEA04M7D`17Ae<00`17Ae<304I6DP<0ADEA0`14A50304I6DP0404M7D`18B5@0
B4QD04Q8E080AdMC0`16AU8404A9C@<0A4U?0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh0
0@3oool001P0BE5L1018DEX304M@F@D0ATmH0`17D5T404QAFPD0BE9K601:DUd01018DU`0AE9K04EB
F`13DUT204EDF@0404ECEP10BT/0@4Y;038f=`809BDW00@081dH01TF4@0I5Q4071TD0P0N6aH0100I
5Q406QLB01XG4P0N6aH201XG4P0401/J600I61H06APF01`H5@807a/H00@07aXF01hG4@0N5a408AXD
0P0W81X0100]9Qh0@c`b04<l<P16?C@203Ta9P0403<Y7`0/8QH0;28F038V6P80ACT]00@0=bLH02PH
200X60P0=R<B0P1H@bh0101TC3@0Gd<[05m3:`1TARX206m@=00407599P1bBRL0LTXW07A<:080NDh[
00@0O50[07mC;01oDb`0PUD^0P23ERl0101lD2/0OE4/07eA;01oDbh2085E<0<0PUHa00@0PED`081D
;`20E2l0QeHf0P23DS801022DC40R5Df08QE=P2BGCh209QSA00409URA02FGd40UUm109]YB0<0W6Y9
0P2KJDP309UWAP0409aZB@2NK4/0WVa;0:9YB`80Y6]=00<0YVe?0:E/CP2UK4h00`2UK4h0102TJdd0
XVU;0:9YB`2TKdd20:E`CP040:MbD02YM580ZGAB0:agE@80[WUG00@0/7]I0;5lFP2aO5X0ZgID0P2/
MeD0102^NEL0[gYH0:mjF02^NEL20:]fE0040:QcD@2UL4h0YG1>0:agE`80[WUI00@0[gYH0:ehEP2]
N5H0ZWIA0P2[Me00102`O5D0]X9K0;J2F`2cOeP20;B0F@040;F1FP2fPU/0]X9K0;J2F`80]81I00@0
/WiG0;5mEP2aOEH0/GiC0P2bOe@00`2cP5D0]85F0;B1EP030;B1EP<0/h1E00@0/gmG0;9nEP2bOUH0
/GaB0P2_NU00102^N4`0[gI90:mfB@2^MDP20:edA`040;9gC@2_M4X0[gA:0:Y_A@80YFY000@0Y6To
0:EZ@02UJT00Z6e30P2ZKdD30:AZ?0040:=Y>`2TIcX0Y6Lj0:AU>@80Xf@h00@0Xf<g0:9R=P2RHSH0
W68d0P2JH380102GGc00U5h`09AN<02CGS82091P<P0408mN<`2>Gc<0SUlc09=W@080QU/h00@0OePg
06Y8;01ZB2`0Ec/S0P1==b80100i9QD0<b8B03<R4P0c9AX202TK400402LI300a91@0<B@D03X]6P80
?30J00@0@3@L04Hj8P16>R80>BX=0P1GAbd0101HB2h0CCdT04dm901RDST205=3;00403l_6011<AX0
@C4J05=3;080A3@M00@0?BhI0552;@1A@Rd0Cd8_0P10=B<01012=bD0?C8P03db8013@34204M4=@04
04m=@@1>CDP0CTe804Q:A`80BTi?00@0Be1D04E:D015BU00A4QC0P17BeH304U=F0<0B4aG00<0BTiI
04a@F`1<D5/0301;CeX404Y>F@80C51K00@0Ce=N051DG`1@E5l0CU9M0P1<D5/00`19CEP0BTiI04Y>
F@0304]?FP@0BDeH0P1:CUT00`1<D5/0BdmJ04]?FP0304]?FPH0BTiI3019CEP<04Q<E`<0C4aH0`1;
BeL304Y:EPL0BDUE0`18B5@504M7D`0404U9E@17Ae<0AdMC04E5D@<0@d=?0P14A5000`15AE40AdMC
04M7D`0304M7D`<0ATIB0`15AE4304A4D0<0ATIB00@0AdMC04Q8E018B5@0B4QD0P17Ae<304I6DP@0
A4U=0`14BDl304A8D@80A4QC00D0AdMC04Q8DP18B580B4U>04U:CP010?ooo`006019DE`404QAFP<0
Ae1I1@16CeP304M@F@@0B55J1@19DU/H04YBG@0404QBG015DU/0AE9K04=BF@80AEAI00@0AE=F041:
B`10BT/0<SHg0P0U9BL0100P7AP06AHA01TF4@0L6A@201hK5P0401TF4@0J5a806QLB01hK5P806QLB
00@06aXH01TH5P0I61H071PE0P0O6aP0100O6QH07QLA01hG4@0Q6Q@202LP6P0402dV7P13?380@c`b
04Hm=080>C4V00@0<bTO02`R5P0/8QH0<RHJ0P15>Bd0100g9aP0:1P802PH200f8a8205Q3;P0406A<
=01O@b/0Gd<[06A6:P80Ke0d00@0LDTV079:9`1bBRL0M4`X0P1iCR/0101lD2/0Oe</07mC;022EBh2
08=F;`0407a@:`1mDB`0OE4/07mC;P80PED`0`22ES401021EC00P5@_081D;`27ESH208=B<P04089A
<@28ECH0R5Df099M?P80V6=400@0VF9409IO@@2FGd40VfU80`2LJTT209]YB0<0VFM600@0W6Y909i/
B`2NK4/0XVU;0P2TJdd00`2VKDl0YFa>0:E/CP030:E/CP040:A[C@2RJD/0XVU;0:A_C@80YG1>00@0
Yg9@0:UdDP2YM580[7ME0P2^NEL0102`NeT0/GaJ0;5lFP2[MU@20:agE@040:iiE`2_NUP0[gYH0:ii
E`80ZgID00@0Z7=A0:E`CP2UL4h0[7MG0P2^NET0102_NUP0[GQF0:ehEP2ZMU420:]gD0040;1lE@2f
PU/0]X9K0;=oF080]81I00@0]H5J0;J2F`2fPU/0]X9K0P2dP5T0102bOUL0/GeF0;5mEP2aOU<20;9o
E0030;>0E@2dPEH0]85F00<0]85F0`2cP5D0102cOeL0/WiF0;9nEP2aO5820:mjD0040:ihC02_MTT0
[gI90:ieB080[GA700@0/WM=0:mdBP2_M4X0ZVm50P2UJT00102TJCl0YFY00:EZ@02XKD<20:Y_A@<0
Y6Xl00@0XfTk0:AW>P2TIcX0Y6Di0P2SI3P0102SHcL0XV8f0:9R=P2LHS@209YP<P0409MO<02DGS00
U5h`09=N<P80T60b00@0Sehc08iO<`2>Gc<0TfM00P26FcP0101oF3L0JTP/06Y8;01G>b<204dg8P04
03TV5@0c8Q80<b8B03<U6P80:A/@00@09aT<034T500a91@0>RdJ0P0l<1X01010=1`0ASXR04Hj8P0i
:Pd205M7;@0405Q8;P1=?B@0CCdT069B>@80Dd</00@0?blH044a6P11<AX0Dd</0P14=1d0100m;QT0
DD8]0552;@1?@Rl2040e8`04048g9@0m<R00?C8P04=0<@80Ad@e00@0Cde104i=B01>CDP0B4Y70P1:
CTl0101;D5@0ADY@04E:D014B5<204M;EP<0BDeH0`18C5L00`1:CUT0C51K04a@F`0<04]?FP@0BTiI
0P1<D5/0101?Deh0D5AO051DG`1>DUd204a@F`0304U=F01:CUT0BTiI00<0BdmJ1019CEP204Y>F@03
04a@F`1;CeX0BdmJ00<0BdmJ1P1:CUT<04U=F0`0B4aG0`1<C5P304];E`<0BTYF1`19BED304Q8E0D0
AdMC00@0BDUE04M7D`17Ae<0ADEA0`13@dl204A4D00304E5D@17Ae<0AdMC00<0AdMC0`16AU8304E5
D@<0A4A@0`16AU801017Ae<0B4QD04Q8E018B5@204M7D`<0ATIB1014BDd304A9C`<0A4QA0P14B5<0
1@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo000H04UAG0@0B55J0`17D5T504I?F0<0Ae1I1018DEX5
04UBFaP0BU9M00@0BE=M04EBF`15DU/0@U5H0P13DUL01016E5L0ADm@04E?D00i?Ch202`/;P0402<P
6`0J5a806QLB01/H4`807a`G0`0K61<0100N6aH06aPC01/H4`0K6QP201TH5P0401dI5P0P71T081`I
020K5`807QLA00@081TC02HO6@0V7aT0:24I0P0n=bd01013>S40=RhS03H^8`0`9Q`202dS5`04034U
6@0l<2@0?30T03DW6P809QH700@0;1/;04/f8`1;=R<0GdH`0P1NA2/0101RA2X0Jdda06]=<@1_BBD2
075;9`0407A<:@1gCb/0Mdl[07aA;P80Oe<^00<0PUHa08=G<023Ec001P1oDc000`20E340PEDb085E
<P03089F<`0408AF=@1iBbX0ND/Z07=28P80NTLX00@0REHg09AO@02DGd00SeXk0P26DC80102LJTT0
VfU809]YB02IIdH209ES@P0409AR@@2GID@0UfE409i/B`80Xg5@00@0ZG1B0:I]C`2VKDl0XFQ:0P2N
IDL0102RJD/0ZW5C0:YaD`2/LeD20:YaD`040:MbD02XLe40Z7=A0:QcD@80ZWEC00<0ZgID0:agE@2/
MeD00`2]N5H00`2ZME<0ZgID0:]fE0040:]fE080ZWEC0`2YM580102XLU00ZgEC0:]eD`2/MU@20:Yd
D0040:MaC@2WLD/0Yg5;0:agC`80/WeE00<0[gYB0;1kD`2`Ne<0102aO5@20;1kD`040:mjDP2^NE40
[WUA0;1kD@80/GaB00<0/giD0;AoE@2dOeD00`2eP5H30;AoE@<0/giD00@0/WeC0;9lD02bO500/WU<
0P2`MdX0102_MDL0[WA60:idAP2[L4H20:U^A0040:I[@@2UJT00YFY00:EZ@080Z6e300<0Zg160:ic
B@2^LdT00`2XKSh00`2WKCd0Z6/l0:Q[?0030:QZ>`<0Z6Pj00@0WF<e09]Q<`2KHC<0Uel`0P2DGS00
102DGc<0TF4c095Q<`2AH3D208mP=00408aP>`1eCBT0MDdY06m;:@80Gd0S00@0DCLN04Pc7P18<ah0
=bHD0P0a8A80100Z7140:A/@02TK400X6Pd202/N3P04038U4P0i;AL0>BdG03lc6`80@SHN00@0?c0C
04lo9@1??bD0He<i0P1QDCP0101C@bX0BCTR04Ti8P14=1d203XZ4`0404@d7@0h:140>2PA03DV4@80
@c@O00@0?2lL03/`7P0k<1h0Bd0^0P10=B<01017A3D0CDTm04e9?@1ACTD204i=B00404U;BP1:CU40
BTiA04Y?E@80ATYC00<0ADUD04Q<E`18C5L00`19CEP304Q<E`0304U=F01<D5/0C51K00`0BdmJ101:
CUT204a@F`0404mCGP1@E5l0D5AO04iBG@80C51K00@0B4aG04U=F019CEP0BTiI2`1;CeX904Y>F@`0
BDeH3018C5L304a<F0<0Bd]G0`1:BUH704U9E@<0B4QD1@17Ae<01019BED0AdMC04M7D`15AE4304=3
C`80A4A@00<0ADEA04M7D`17Ae<00`17Ae<304I6DP<0ADEA0`14A50304I6DP0404M7D`18B5@0B4QD
04Q8E080AdMC0`16AU8404A9C@<0A4U?0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3o
ool001P0BE5L1018DEX304M@F@D0ATmH0`17D5T404QAFPD0BE9K3019DE`<04YBG@0404QBG017E5d0
AeAM04=BF@80@E1E00@0AE=F04QBD`18DU<0@DE60P0e=CL0100W9R406QTD01XI500J6A@201lN6@04
01/J5@0J6A@06QTD01dL5`806aXE00@06aXH01TH5P0I61H07ATF0P0Q7AX0100P6aL07QLA01hG4@0O
618202DN600402DN5P0e;R@0=BhT03T`9`80<RXO00<0<2HL02lU6@0_9AT00`0_8aL0100b9QP0:ad@
02/M400Z6PX203XX5004050j9@1M@b`0GD</0614;080IDL]00@0KdPW0719:01`BBP0LT`X0P1fCR/0
101iDBd0OU<`07iC<020ES02089F<@04089E=021E3<0PE@c07iA<080O4l^00@0OE0_07mB<@1oDS40
PUDd0P24EcH01026EcT0OTla07i?<@1hBB/207e<;P0408MF>02>Fch0SU/n08eG>`80R58f00@0TV50
09AS@P2DHd80UFA30P2FID@0102EI4<0UfI509MVA@2KJTT209i]C00409iWB02OJ4T0WfQ909mXB@80
W6E600@0VfA509iWB02NIdP0Y6e>0P2WL5430:MbD0040:QcD@2YM580ZGAB0:YeD`<0[7ME0P2]N5H0
102[MU@0ZWEC0:YeD`2XLe420:IaC`040:MbD02YM580ZGAB0:ehEP80[gYH00@0Z71?0:YbD@2ZLU40
[7AA0P2/M4l0102[Ldh0ZgA=0:]dC@2]MTl30;1iDP80/GYC1P2bNe@30;5jD`<0/GYA00<0/gaC0;Am
E02dOE@00`2eOUD30;AmE0040;9mD`2cOE40/geA0;9lD080/gY=00@0/GQ;0;1fB02`MTP0[WA60P2]
LdD30:U^@P<0Z6e100<0ZFi20:Y_@`2ZKd<00`2[L4@30:M]>`040:I/>P2XK3X0Z6`j0:QZ>@80YfTh
00@0Z6Ph0:MW=`2WIcL0W68d0P2IGc40102EGBh0Ted_09=M;`2EH3@209=S=@04099Q=P2@HCD0T64e
08ML>@80J40M00@0HShN05Tj7P1I>Qh0CC<L0P15<1d0100c8Q80:1/;02PK2`0T5P/202/M4P0402hP
4`0/7`l0;1l?02hQ3P80=RXD00@0?C4I040d7010=1`0CChQ0P1B@RP0101JBS00ETH]05I6;@13<aX2
048b6`0404De7P12<Q/0@S8K03PX4@80<20900@0=2D@04<d7`13=1l0=BPE0P0e:QP0101<@Bl0ACXX
04Dj:01;Ac/2055=@P04051=A019B4@0BDQ404I8A`80ATY=00@0AT]A04E9DP15BE80Ad]F0P19CEP3
04Y>F@<0B4aG00<0BDeH04]?FP1;CeX0301;CeX404Y>F@80C51K00@0Ce=N051DG`1@E5l0CU9M0P1<
D5/01017BeH0B4aG04Q<E`19CEP204]?FP<0C51K00<0BdmJ04Y>F@1:CUT00`19CEP904Y>F@`0BDeH
3018C5L304a<F0<0Bd]G0`1:BUH704U9E@<0B4QD1@17Ae<01019BED0AdMC04M7D`15AE4304=3C`80
A4A@00<0ADEA04M7D`17Ae<00`17Ae<304I6DP<0ADEA0`14A50304I6DP0404M7D`18B5@0B4QD04Q8
E080AdMC0`16AU8404A9C@<0A4U?0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool0
01P0BE5L1018DEX304M@F@D0ATmH0`17D5T404QAFPD0BE9K3019DE`<04YBG@0404QBG017E5d0AeAM
04=BF@80@E1E00@0AE=F04QBD`18DU<0@DE60P0e=CL0100W9R406QTD01XI500J6A@201lN6@0401/J
5@0J6A@06QTD01dL5`806aXE00@06aXH01TH5P0I61H07ATF0P0Q7AX0100P6aL07QLA01hG4@0O6182
02DN600402DN5P0e;R@0=BhT03T`9`80<RXO00<0<2HL02lU6@0_9AT00`0_8aL0100b9QP0:ad@02/M
400Z6PX203XX5004050j9@1M@b`0GD</0614;080IDL]00@0KdPW0719:01`BBP0LT`X0P1fCR/0101i
DBd0OU<`07iC<020ES02089F<@04089E=021E3<0PE@c07iA<080O4l^00@0OE0_07mB<@1oDS40PUDd
0P24EcH01026EcT0OTla07i?<@1hBB/207e<;P0408MF>02>Fch0SU/n08eG>`80R58f00@0TV5009AS
@P2DHd80UFA30P2FID@0102EI4<0UfI509MVA@2KJTT209i]C00409iWB02OJ4T0WfQ909mXB@80W6E6
00@0VfA509iWB02NIdP0Y6e>0P2WL5430:MbD0040:QcD@2YM580ZGAB0:YeD`<0[7ME0P2]N5H0102[
MU@0ZWEC0:YeD`2XLe420:IaC`040:MbD02YM580ZGAB0:ehEP80[gYH00@0Z71?0:YbD@2ZLU40[7AA
0P2/M4l0102[Ldh0ZgA=0:]dC@2]MTl30;1iDP80/GYC1P2bNe@30;5jD`<0/GYA00<0/gaC0;AmE02d
OE@00`2eOUD30;AmE0040;9mD`2cOE40/geA0;9lD080/gY=00@0/GQ;0;1fB02`MTP0[WA60P2]LdD3
0:U^@P<0Z6e100<0ZFi20:Y_@`2ZKd<00`2[L4@30:M]>`040:I/>P2XK3X0Z6`j0:QZ>@80YfTh00@0
Z6Ph0:MW=`2WIcL0W68d0P2IGc40102EGBh0Ted_09=M;`2EH3@209=S=@04099Q=P2@HCD0T64e08ML
>@80J40M00@0HShN05Tj7P1I>Qh0CC<L0P15<1d0100c8Q80:1/;02PK2`0T5P/202/M4P0402hP4`0/
7`l0;1l?02hQ3P80=RXD00@0?C4I040d7010=1`0CChQ0P1B@RP0101JBS00ETH]05I6;@13<aX2048b
6`0404De7P12<Q/0@S8K03PX4@80<20900@0=2D@04<d7`13=1l0=BPE0P0e:QP0101<@Bl0ACXX04Dj
:01;Ac/2055=@P04051=A019B4@0BDQ404I8A`80ATY=00@0AT]A04E9DP15BE80Ad]F0P19CEP304Y>
F@<0B4aG00<0BDeH04]?FP1;CeX0301;CeX404Y>F@80C51K00@0Ce=N051DG`1@E5l0CU9M0P1<D5/0
1017BeH0B4aG04Q<E`19CEP204]?FP<0C51K00<0BdmJ04Y>F@1:CUT00`19CEP904Y>F@`0BDeH3018
C5L304a<F0<0Bd]G0`1:BUH704U9E@<0B4QD1@17Ae<01019BED0AdMC04M7D`15AE4304=3C`80A4A@
00<0ADEA04M7D`17Ae<00`17Ae<304I6DP<0ADEA0`14A50304I6DP0404M7D`18B5@0B4QD04Q8E080
AdMC0`16AU8404A9C@<0A4U?0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool001P0
BE5L1018DEX304M@F@D0ATmH0`17D5T404QAFPD0BE9K3019DE`<04YBG@0404I@FP18EEh0B5EN04ED
F`80@E1E00@0A59E04YDE@1:E5D0Ad]<0P0n?T00100`;bX07ahI01lN6@0K6QD2020O6P0301`K5P0J
6A@06QTD00<07QdH00@06aXH01XI5`0J6AL07QXG0P0Q7AX0100P6aL07QLA01hG4@0O618202<L5P04
02@M5@0[91X0:b@J02lV7@80<2PM00@0<bTO034W6`0a9a/0;B4E0P0Y7A40100b9QX0=2HI03@V6@0^
7Pl202hK2P04044/5`1F?RH0EShV05e3:P80H4DZ00@0KDTY06i::01^BRP0LT/Z0P1dCRX0101hD2d0
NU8^07YB;P1nDc0207mD<@0408=F=@21E3<0PE@c07iA<0<0OE0_0P1oDS401022EC@0Q5Lf08AG=P26
FC`208MJ?@<0REXn00@0ReTn08eJ?`2=FSl0Sea10P2AGT<01022DC00RUTh08YI>02CHT4209YYB004
09]ZB@2IJ4L0VFQ709MVA@80UFA300@0U5lo09aWA`2LIdL0YW5A0P2YM5@0102PJd/0UV5109IQ@@2H
Hd<20:5/C0040:5/BP2RKD/0XVe;0:=^C080YW5?00@0Z7=A0:]fE02[MU@0[7ME0P2]N5H0102/MeD0
ZGAB0:UdDP2UL4h20:9]B`040:=^C02VLDl0YW5?0:]fE080[gYH00@0Zg=@0:YbC`2ZLTl0Zg=@0P2]
ME00102`NE80/W]D0;9kE02bNe820;5jD@030;AmE02eOUD0]GiE00<0]GiE1P2fOeH30;5jD@040;9k
DP2cO5<0/gaC0;AmE080/gaC0`2bNe80102bO500/G]?0;5kC`2bND`20;1gBP030:meA`2]LdD0[G=5
00<0[W520`2/LDD0102[L4@0ZVm30:Y_@`2XKD420:M/@0<0YFXn0`2SJCL0102RJ3H0XfLe0:=W=@2R
I3<20:5S<P<0XF4a00@0W68d09QN<02HGS00U5`]0P2BG2h0102DGc<0Tf<e09=S=@2BHCH208mP=004
08EM>@1S?AT0HcdI05dj7080EcTM00@0AbhH03`Z5P0l:QH0;Qh?0P0V60/0100V60d0;AlD02dO500b
91L2038U5@0303DX5@0l<1X0?30J00<0@CDM00@0BCXM05A4:P1DA2X0AcLM0P15=A`01019>B00AcLP
04Lg800g9a0203h^5`0403/[500b8P/0<R8;03//5`80D44/00@0@3<P03<X5P0c:1H0ACXX0P16>bT0
101;Ac/0DTi3059>@`1?C4D204I5@@0404E6B014Ad`0A4M<04=7D080ADUD00@0B4aG04Y>F@1:CUT0
BdmJ0P1:CUT01019CEP0B4aG04Q<E`19CEP>04]?FP@0BTiI0P1<D5/0101?Deh0D5AO051DG`1>DUd2
04a@F`0404M;EP18C5L0B4aG04U=F080BdmJ0`1=DE`0101;CeX0BDeH04U=F018C5L504U=F0H0BTiI
3019CEP<04Q<E`<0C4aH0`1;BeL304Y:EPL0BDUE0`18B5@504M7D`0404U9E@17Ae<0AdMC04E5D@<0
@d=?0P14A5000`15AE40AdMC04M7D`0304M7D`<0ATIB0`15AE4304A4D0<0ATIB00@0AdMC04Q8E018
B5@0B4QD0P17Ae<304I6DP@0A4U=0`14BDl304A8D@80A4QC00D0AdMC04Q8DP18B580B4U>04U:CP01
0?ooo`006018D5/404QAFP<0Ae1I1@16CeP304M@F@@0B55J1@19DU/<04Q@F``0BU9M00@0A4iH04UF
G`19EUl0B5MN0P11D5D01012D5<0BeEF04]EEP1<D54204I6B00403Lh<P0R8ad08R<M01/L5P807b0J
00@06a`F01TJ500I6Q@07QlI0P0O81X0100K6QP06QTG01XI5`0N6QL2028N6`04024L600N5a407QLA
01hG4@808Q/E00@08Q/C02@M4`0T7A<0;2<J0P0d;240100f;280<2HJ030V6P0/81@202XN4P0403@Z
7P0k;b<0>blS02lR4P809aH400@0=b8?04dd7P1==1h0F3hW0P1PA2`0101]BBT0KTXZ06i::P1_BbT2
07=<:`0407E?:`1hD2d0N50]07YB;P80OE8_2P20Dc@207mB<`<0PUDh00@0Q5Lj08MH?027F3`0REXn
0P2=Fd00102>Fd00SEXo08eJ?`1iBbT2085C<@0408aN?02DIT@0U6I409MYA`80UFM500@0U6I409=U
@`2CID<0U5lo0P2CGSh0102NJDT0ZgIF0:]fEP2YM5@209UTA00409AO?`2KITH0VfI609eXAP80WVU7
00@0WfY80:5/BP2QK4X0Xfi<0P2UL4h0102VLDl0Yg9@0:MbD02YM5820:MbD0040:=^C02PJdT0X6]9
09mZB080XFa:00@0Y6m=0:MbD02WLU00[W=A0P2YKT`0102UJdH0Z6i80:Q^B02_MDl20;EkD`030;Aj
DP2aMdl0/GM?00<0[G=;0`2^M4`30:meC@<0/7I>00<0/gUA0;AjDP2dNU80102eNe<20;AjDP040;=i
D@2bN500/WQ@0;5kC`80/GQ;00@0[WE80:ecA@2]LdD0Zg530`2/Kd020:e`@@040:ecA@2[LD<0Zg53
0:Q^@080YF/m1P2SJC/0102RJC@0XFPc0:5X<`2PIc820:5U<@040:1R;`2NH2d0WV0]09iO;080WEh[
00@0WF<e09UO<@2IGc40Te//0P2AFbd0102CGS80TV8d099R=02?GS<208]L<00408AL>@1Q>QT0HCXI
05ll7P80FC/Q00@0@RTC03HS4P0f8a80<R8C0P0c9aT0100_8AH0<28G030R5`0c9AP203X]7@0404@g
901;?bT0BclY04Xn9P80ACTQ00@0?c0C04Xj801:>R00@S8H0P19>B00101>?RD0Bc/T04/k900c8``2
03@T3@0403PX4@0i:A80>BTB03h_6P80C3dX00@0@cHS03hc8@0n<b40CD8`0P16>bT01016@SL0DDe4
055=A01>Bd@204I5@`0404Q9B`17BTl0AdY?04=7D080Ad]F00@0BDeH04Y>F@1:CUT0C51K0P1;CeX0
1019CEP0B4aG04Q<E`19CEP>04]?FP@0BTiI0P1<D5/0101?Deh0D5AO051DG`1>DUd204a@F`<0B4aG
00@0BDeH04]?FP1;CeX0CU9M0P1=DE`00`1:CUT0Ad]F04M;EP0404Q<E`<0BDeH1@1:CUT<04U=F0`0
B4aG0`1<C5P304];E`<0BTYF1`19BED304Q8E0D0AdMC00@0BDUE04M7D`17Ae<0ADEA0`13@dl204A4
D00304E5D@17Ae<0AdMC00<0AdMC0`16AU8304E5D@<0A4A@0`16AU801017Ae<0B4QD04Q8E018B5@2
04M7D`<0ATIB1014BDd304A9C`<0A4QA0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>0040oooo000H
04Q@F`@0B55J0`17D5T504I?F0<0Ae1I1018DEX504UBF``0B51K301:DUd01014CUP0BEIO04UFG`18
Eeh2045@E@04049@D`1;EEH0BeEF04a@D@80ATI800@0=cPb028S7@0R8ad06a`F0P0O81X0100K71H0
6AXD01TJ500N7aT201lP6P0401/J600J6AL06QTG01hJ5`808QhK00@08A`H01hG4@0N5a407QLA0P0R
6aD0100R6a<091dC02@M4`0/8aX203@/8@0403H/8P0`9QX0<2HJ02`P5080:QhB00@0=2XN03/_8`0k
;b<0;b8B0P0W5P@0100g8Pl0CC@N04dd7P1H?RL20614;00406e9:@1^BRX0KTXZ06m;:@80Ld`[00@0
MDl[07Q@;@1hD2d0NU8^0P1mDRl:081C=080Oe8c0`22ECP01024EcX0QePl08MH?029FSh208eK@004
08iK@02=FSl0SEXo07U;:@80PE<a00@0S5hl09AVA02DIT@0UfU70P2EIdD0102DIT@0TfE309=U@`2D
Gcl209=N?P0409iYB@2[MUH0ZgIF0:UdE080VFA400@0U5lo09]VAP2KITH0WFQ60P2NJDL0102OJTP0
XFa:0:5/BP2SKT`20:E`CP040:IaC`2WLU00Yg9@0:UdDP80Yg9@00@0Xfi<0:1[B@2PJdT0WfY80P2Q
K4X0102TKdd0Yg9@0:MbD02^Le420:U^C0040:E[AP2XKTP0Z6i80:meC`80]G]C00<0]7YB0;5gC`2a
Mdl00`2]Ld/30:idC0<0[gE=0`2`MTh00`2cNE40]7YB0;AjDP040;EkD`80]7YB00@0/gUA0;9hD02b
N500/G]?0P2aN4/0102^MDP0[G=50:ecA@2[LD<30:a_@080[G1100@0[G=50:]a@`2[LD<0Z6i00P2U
Jcd60:=Y>`040:9Y=02QJ3<0XFPc0:1W<P80XFDa00@0X68_09iP;@2NH2d0WUl/0P2MGR/0102MHcD0
VEla09UO<@2CFb`2095K;@0409=N<P2BHS@0TV8d08mN<`80Re``00@0Q5`i064j6@1Q>QT0Gc`N0P1I
>b401012:A<0=R<B03HS4P0b8Q<203<W6@0402lQ5P0`8QL0<28G03<U6080>RdM00@0A3LT04/o:@1;
?bT0BShV0P15>B40100o<1<0BSXP04Xj8012<QP204Ti800404hn9@1;>b@0Bc/T03<S3080=2@=00@0
>2PA03TY4P0i:A80?RlJ0P1<?BP01013=R<0?S<Q03hc8@1=@S0204Hk:@0404I2=`1ACD@0DDe404i;
A080ATE300@0B4U;04M:C`17BTl0@dM@0P17BeH01019CEP0BTiI04Y>F@1<D5/204]?FP0404U=F018
C5L0B4aG04U=F0h0BdmJ101:CUT204a@F`0404mCGP1@E5l0D5AO04iBG@80C51K0`18C5L01019CEP0
BdmJ04]?FP1>DUd204eAG00304Y>F@17BeH0Ad]F00@0B4aG0`19CEP504Y>F@`0BDeH3018C5L304a<
F0<0Bd]G0`1:BUH704U9E@<0B4QD1@17Ae<01019BED0AdMC04M7D`15AE4304=3C`80A4A@00<0ADEA
04M7D`17Ae<00`17Ae<304I6DP<0ADEA0`14A50304I6DP0404M7D`18B5@0B4QD04Q8E080AdMC0`16
AU8404A9C@<0A4U?0`14B54204A8D`0504M7D`18B580B4QB04Q9CP19BTh00@3oool001P0AdmJ1018
DEX304M@F@D0ATmH0`17D5T404QAFPD0BE9K3018D5/<04YBG@0404==E`1:Ef00BUMP04YIH080@U5F
00@0@DmB04YDE@1:E5D0Ce=D0P1;Bdd0100l?CL09BHP02DV800L7AL201lP6P0401/L5P0I6Q@06AXD
01hO6@808B8L00@06aXH01XI5`0J6AL07a/H0P0R7Q/0100Q71P07QLA01hG4@0N5a42028K5@04024J
4P0Q6Q008AX@02dT6`80>30U00@0=bdS02hT600^91P0;B4E0P0^8QH0100i;b<0?c<W03lc9`0a91@2
02HE1@0403<N2`14;QL0A2hG05@j8`80HTP_00@0K4TY06e::P1]BRX0Kd/[0P1cC2/0101dCB`0Mdl/
07M?;01hD2`207UA;@0407a?<01oDS<0Oe8c089E=P<0Q5Lh0P21E3D0101nDC80Ndh_07]>;`1kCc82
07A7:P0407139P1hBBd0N4T]08EF>P80SE]000@0QE<h07]9;P1kBBh0Ndd[0P20DS001026F3H0Redk
08]M>`2?HCl2099T@P0409IXAP2IJdT0VF]909EP@080Pdh^00@0Q4l_09YUA@2JIDD0Yg9B0P2LIdL0
0`2AG3`0TUdm099M?@0<09iYA`040:MbD02UL4h0YG1>0:9]B`80WfY8102MJ4H209iYA`040:]aC02R
J4<0XVQ309]Q?080WF<m00@0Z6i80;1fCP2`MTh0/GM?0P2]M4T60;1fCP040;5gC`2bN500/WQ@0;9h
D080/gUA0`2dNU830;EkD`030;AjDP2cNE40/gUA00<0/GM?00@0/gY=0;1gBP2`MdX0[7=60P2ZL480
102XKT00ZVdn0:Y]?P2/KSd20:e_?P040:Y`@P2WKCl0Yfdo0:=Y>`80X6Hh00@0WfDg0:5W>@2QIcT0
Y6Xl0P2VK3h0102UK3L0Y6/f0:A[=P2SJSD20:=W<`030:9T<@2PHRl0X68_00<0Wf0]00@0X6Hh09YP
<P2JH380U5`]0P2AFbd0102BGC40T60b091P<P2=G34208MH;00407mG=01O>1L0GcPG060o8080F3dR
00@0?2H?03<P3`0c80l0=bXJ0P14>2X0100f:1d0<B<H034S600a8aH203ha8@040513<01HC3H0F4`f
0596;P80BCdU00@0A3DH03l_5@0o;aD0ASHL0P1EAB`01010<1L0ASHO04Hf7`11<AX203HV3`0402dM
1P0n;QL0?RhG03`]6080=bPC00@0>B`I04i3<@1>@c40GeA20P18?B/01011?C80CTY104i:@@1>BdH2
04U8AP0404e>D01;CUD0BdiE04E9DP80B4aG00@0BDeH04]?FP1;CeX0C51K0P1;CeX01019CEP0B4aG
04Q<E`19CEP>04]?FP@0BTiI0P1<D5/0101?Deh0D5AO051DG`1>DUd204a@F`0404U=F018C5L0B4aG
04U=F080BdmJ00@0CU9M04eAG01=DE`0BDeH0P16BUD304M;EP<0B4aG00<0BDeH04Y>F@1:CUT00`1:
CUT<04U=F0`0B4aG0`1<C5P304];E`<0BTYF1`19BED304Q8E0D0AdMC00@0BDUE04M7D`17Ae<0ADEA
0`13@dl204A4D00304E5D@17Ae<0AdMC00<0AdMC0`16AU8304E5D@<0A4A@0`16AU801017Ae<0B4QD
04Q8E018B5@204M7D`<0ATIB1014BDd304A9C`<0A4QA0P14B5<01@17Ae<0B4QB04Q8DP18BDh0BDY>
0040oooo000<04Q>FP@0Ae5K0`16D5X504E?F@`0AU1I3017DEX<04QBG0`0AU1J00@0B51M04YBG`1:
DUl0C5AQ0P1:DUd01016CUT0@daE04=<E@13C5D204A=E00403/k>`0`<300<30`02@T908071`J00@0
71dH01hO6P0N7aX07AhH0P0L7AL0100R7Q<08QhE028N5@0Q7A@3020K5@808A`H0`0R7AT0100Q71H0
71LA01`G4@0W8aX203D^9004030Z700_:1P0;bPH03</6P80<RTH00@0;2<N034V800a9R00>B`T0P0c
8AL0100Y4`D0>Al>03TO3P1B=R4205/l9`0406I59P1ZAbT0JTLY079>;P80NU<b00@0NE4]07I<9P1f
C2H0NDdV0P21E2d0101lDBh0NTl/07Y?;01jCb`207iC<@04089G=@21ESH0PEHf07U>;P80LdLZ00@0
KS/N078o8P1b?b80Ld0S0P1`?B00101e@RD0P4d`081=<024DC@2085><@<0OE8b00@0S651091U@`2@
ID<0QEXh0P2AIT<0102MLTl0TFM1095W@@2AI4<208QK>P0407128@1X>QP0J3XH08EE<@80WFe900@0
UfE008ME<027EC00ReHf0P2@Fc/0102BGCd0Uf9209MR@P2LIdL209AO?`04095L?02KITH0VfI609iW
B080XV]<00@0YVe?0:9YB`2RJD/0V5m10P2EGC`0102OIdH0Zg=B0:]cDP2aLdh20:e_BP040:5S?P2J
G3L0VU`g0:1R?@80/gE@00<0YFM20:U[AP2YJdH00`2_M4`00`2`MDd0/GI>0;5fCP060;5fCP040;9k
DP2cO5<0/gaC0;AmE080]GiE00@0/gaC0;1iD02`NE00[7E<0P2ZLdX0102_MDL0[G=50:ecA@2ZL482
0:M]?`<0YV`n00@0Z6i00:U_@@2YKd40ZGA20P2XL3l0102TK3/0Y6Xj0:AZ>P2VK3`30:Y]?P80ZV`m
00<0Xf@a0:MX=@2WJ3D00`2YJcP00`2UJCD0WfHa09mV<@0309eT;`0409aR<P2IGbd0VEl]09EI9@80
T5@P00@0TUXW09=P<@2CH340RE`b0P1kE2T0101bBBd0GcLM05lg7@1G<aT2058d700403lV4P0d8A00
=24@03d]7P80@cDX00@0?C4U02/O4@0[7a40=BTK0P14>2P0101BASH0FDhl05U>?01;@2h204Dj:004
04Tj7@1=?R40CChQ04`l8P80CShT00@0@C4H03h^5`0n;QL0Bc/T0P0o;aP0100f9A/0>2HJ03PV6P0m
:Q/204Db8@0404hm:`1=@C40CD4a048j;@80=30T00@0<cDZ03m0>00o@3P0BDY50P19BdX01016BTd0
B4]D04Q;E019CEP204M;E`0404a@F`1;CeX0BdmJ04U=F080B4aG0`17BeH304Q<E`@0BDeH0`1:CUT5
04]?FP0404eAG01;CeX0BdmJ04Y>F@80BdmJ00@0CU9M04mCGP1?Deh0CU9M0P1<D5/01019CEP0BTiI
04Y>F@1:CUT304]?FP<0BTiI0`19CEP304Y>F@<0BdmJ0`1:CUT>04U=F0`0Bd]G301:BUH<04M7D`04
04U9E@18B5@0B4QD04I6DP<0A4A@0P16AU800`18B5@0BDUE04U9E@0304Q8E0@0AdMC4@16AU8404A9
C@<0A4U?0`14B54204A8D`0504M7D`17Ae40AdMA04Q9CP19BTh00@3oool000`0B4iJ1017DE/304I@
FPD0ADmI3016D5T<04MAFP`0B59L3016D5X01018D5d0BU9O04YBG`1<E64204YBG@0404I>F@13C5D0
@daE04=<E@80A4eD00@0>c/k030`<00`<30092@T0P0L71X0100L7AP07QlJ01hO6P0M7QP201`M5`04
028N4`0R7QD08QhE024M50<081/E0P0Q71P3028M6@04024L5P0L5a4071LA02LS6P80=BhT00@0<2XL
02lX600_:1P0<b`J0P0b:AP0100/8ah0<BHP034V800i;2@203<Q5`0402TC1@0i7`h0>Al>058f8@80
Fc`W00@0ITDV06Y7:@1ZAbT0LTh^0P1jDc80101iDBd0MT`V07I<9P1iCBH2085D;@0407aA;P1jCb`0
NTl/07Y?;080OU<a00@0PULe085F=P21ESH0NDh^0P1cAbX0101^>ah0LSlR078o8P1c@2<2070m8004
07E29@20CC00P4d`08AA=080PDha0`1mDS80102<HD40T6E3091U@`25FSP2095V@`0409ebC`2AId40
TFM1095T@`80R5/j00@0L48Q06Pj601X>QP0QEDa0P2MKDT0102GID00QeD`08ME<02;ESH2091K>`04
099M?@2GHT80Uf9209aWA`80U5lo00@0TE`l09]VAP2KITH0WVM80P2RJd`0102VKDl0XVU;0:9YB`2H
Gd4209EM?00409mWAP2[Le80Zg=B0;5cCP80[Fm:00@0XF<n09YL=`2JG3L0X68m0P2cME000`2UId80
ZF]60:U[AP030:mdC0030;1eC@2aMTh0/GI>00H0/GI>00@0/W]B0;=lD`2cO5<0]7eD0P2eOUD0102c
O5<0/7U@0;1iD02/MD`20:YcBP040:meA`2]LdD0[G=50:Y`@P80Yfdo0`2VK3h0102XKT00ZFm10:U_
@@2YM4820:Q`?`040:A/>`2TJSX0Y6Xj0:I/?0<0ZVdn0P2ZK3d00`2SI340YfPe0:MX=@030:U[>003
0:EY=@2OIS40WfHa00<0WF@_00@0W68b09UO;@2IGbd0UETU0P2@E200102BFRL0Tf0a09=P<@29G382
07]D:@040799;@1O=ad0GcLM05Lc6@80DS@L00@0?bHB03@Q400d8A00?BdN0P13=BP0100m<BD0:alA
02/O4@0e:A/204@h:0040596=P1ICS`0FDhl04]0;P80ACXX00@0BCXM04dn8@1=?R40C3`R0P1>?R@0
1011<AP0?RhG03h^5`1;>b@203l_600403HU6`0h9QX0>2HJ03dZ6`80AC8Q00@0CSd[04e1<@1=@C40
@SX]0P0d<2@0100c=BX0?d0h03m0>019BTD204U;BP0404I:C@18Be@0B4]D04U=F080Ad]G00@0C51K
04]?FP1;CeX0BDeH0P18C5L304M;EP<0B4aG1019CEP304Y>F@D0BdmJ00@0CE5L04]?FP1;CeX0BTiI
0P1;CeX0101>DUd0Ce=N04mCGP1>DUd204a@F`0404U=F01:CUT0BTiI04Y>F@<0BdmJ0`1:CUT304U=
F0<0BTiI0`1;CeX304Y>F@h0BDeH301;BeL<04Y:EP`0AdMC00@0BDUE04Q8E018B5@0ATIB0`14A502
04I6DP0304Q8E019BED0BDUE00<0B4QD1017Ae<A04I6DP@0A4U=0`14BDl304A8D@80A4QC00D0AdMC
04M7D@17Ae40B4U>04U:CP010?ooo`003018CUX404Q@F`<0AdmJ1@16CUT<04M@F@`0B55J3019DE`<
04Q@F`0404Q@G@1:DUl0BU9O04]CH080BU9M00@0ATiI04=<E@13C5D0@daE0P14CE@0100o@480=3Dg
03@e=`0W9bL201dM6`0401/L5`0K71H06a`F01`K5P806a/C00@08QhC028N5@0R7QD08AdD0`0P6aD2
024L60<08QdI00@08A`F01`G4@0L5a409R8I0P0d;B<0100b;200<BXJ034Z6P0c;1`202lV5@0402hW
8@0`:B40<2TQ03L]9080<bDH00@0:AP803<N2`0c7P/0BS4M0P1C>2<0101T@RH0IdHW06M69`1aCS02
07QD<P0407QB;P1dBBH0M4TV07=98`80NDdX00@0OeLc07YB;P1jDRh0MDdZ0P1cBbP0101cBRX0LdX/
07=:;01`AbT206e49P0406Lh7P1X>Al0J3TO06Hg7@80HS<I00@0Hc@J06`m8`1/?B<0MDH/0P1kC380
101nECL0N4la07Q?<@1nECL207mF=P0407E<;01kDc00Ne<`08]S@080SVI200@0Oe@d07e@<@1mD340
JchM0P1R=1<0101a@b40QeHe08MF=@2@H3`209IT@@0408eH>028Dc<0R5<c08A?;`80T5/k00@0XVe=
0:5/C02QK4`0Tehn0P2?FSX0102FHCl0WFM509eWA@2PJTP209eWA@0409US?`2JHSl0VV8o09aT@@80
WVI30`2`M4l0102VJTD0Velj09]O>P2HG3L20:M[AP0409iR?@2YKDP0ZFe80;5fCP80/7E=00@0[gA<
0:icB`2^Ld/0[gA<0P2`MDd0102bMdl0/gQ@0;=hD02cNE420;AjDP<0]G]C00@0]7YB0;5gC`2aMdl0
[WA<0P2[LDT0102]LdD0Zg530:]a@`2XKT030:I/?P80Z6i000@0Zg530:ecA@2]LdD0ZWE30P2ZLT40
102WKch0Yfdm0:M]?@2XKSh20:]^?`040:U/?@2YJc`0ZF/l0:EV<`80YfPe00@0ZF/h0:QZ=`2XJSL0
XfLc0P2MI2l309]R;@0409aR<P2JH2h0VV0^09IJ9P80TEPS00@0Te/Z091P<02@H300QEX_0P1gD2L0
101^AbX0G3LL05`g701=:a8204<W4@0403hU4@0j9aH0>RLF03`/7@80>B`L00@0=RHG04<c9013<b@0
BSXZ0P18>2P0101IBSL0F4Tf05Q9=P10<A`204Pi900403l`4`1<?280C3`R04Lg7@80A3@K00@0@c<J
03l_600o;aP0ACDN0P13<a`01013<RP0?B/O03d[7`0k:1T204Db8@0404ln;01<@300C40`03lg:P80
<RhR00@0@44i04Q9@`18BD<0C4e80P18BTT01014Ad`0B4]D04Q;E01;CeX204a@G00404]?FP1:CUT0
BTiI04U=F080B4aG0`17BeH304Q<E`@0BDeH0`1:CUT504]?FP0404a@F`1:CUT0BTiI04U=F080BTiI
00@0CE5L04iBG@1>DUd0CE5L0P1;CeX01019CEP0BTiI04Y>F@1:CUT304]?FP<0BTiI0`19CEP304Y>
F@<0BdmJ0`1:CUT>04U=F0`0Bd]G301:BUH<04M7D`0404U9E@18B5@0B4QD04I6DP<0A4A@0P16AU80
0`18B5@0BDUE04U9E@0304Q8E0@0AdMC4@16AU8404A9C@<0A4U?0`14B54204A8D`0504M7D`17Ae40
AdMA04Q9CP19BTh00@3oool000`0B4iJ1018D5/304M?FPD0ATiI3017D5T<04QAFP`0B51K3019DE`0
1018D5d0BU9O04YBG`1;Df0204YBG@0404M?FP14CEH0A4eF04=<E@80A4eD00@0ATM;03Xk?@0j>cd0
:R`[0P0O7ad00`0J6aH06aXE01/J5@0301XJ4P04028N4`0R7QD08QhE024M50<081/E0P0Q71P4028M
6@807APB00@0920G034Z800a:R00=2hR0P0e;B00100b:a/0;2<D02`S500^:b8202hZ7`0403D_8`0d
:a`0=2/L02XM3080;1/900@0>RPB04D`6`15<1/0DS@H0P1H>Ad0101VA2P0LTla079?<@1eDBl207==
:@0407A<:01jCb`0NTl/07QA<080MTl^00@0KdPW06@m7P1T?Ah0G3DF0P1N=aX0101Z@bH0MDdc07E=
<`1P=b4205he7`0305`c7@1J<A/0FS4K00<0EbhH00@0GSDO06Hm9`1V?BL0IclU0P1Y@BL0101hDC@0
QUm208IO@P23G3d207aE=P0407eF=@22FcX0PU/j09=WBP80UVY=00@0REe007M::`1gBR/0KD0O0P1[
?A`0101dAR@0QeHe08MF=@26ECD2085@<00307U8:020Cbl0P4l_00<0TV5100@0R5Lg08IE=@26ECD0
SE/h0P2DHSd0102FI3l0V6@m09QT?@2MId420:E`B0040:=/A@2JHc`0VV<l0:M`B@80ZgA=00@0YVm8
09]T?@2KI3d0Te`e0P2MISl0102IHS/0[GI?0:efC`2bMdl20;1eC@040:icB`2/LDT0[7590:ebBP80
[gA<00@0/WM?0;EjDP2eNU80/GQ=0P2bNDh30;=jC`040;5hC@2_MT/0[gI;0:acB080ZW5600@0Zg53
0:Y`@P2ZL480Yfdo0P2VK3h0102WKCl0ZW120:Y`@P2^M4H20;1fB0040:Ud@P2ZLT40ZW910:Q`?`<0
ZW100P2/Kd00102YK3d0YfTj0:MY>P2UIS<20:MX=@040:MY=P2UIc@0YFLd0:5U<@80Vf8]0`2IH2/0
102KHc00VV0^09YP;P2GGRT209EL9`0409=N;02=GBl0SEd_081E:`80LT/T00@0Id8X05Pd6P1H=1X0
@B4:0P0i7@L01012;1L0Ac@S04Ld8`10<20203P[6`0403dZ6@1;>2L0BcPW05E3;`80BcTS00@0CcdW
050n9P1@?RH0?bdE0P11;aL0100`80H0ASHL04Hf7011<AL203XZ4@0403d]5P0i:A80>BTB03`[5`80
A3<O00@0ACDX03XX6P0j:1X0>RLH0P1:=bH0101FAC<0C40`04a0<00k<bP202hZ7`040493?@18BD@0
B4U404];B@80B4U;00@0ADQ=04Q;E018Be@0BdmJ0P1=DEd0101:CUT0BDeH04U=F018C5L504M;EP<0
B4aG1019CEP304Y>F@H0BdmJ0P19CEP01018C5L0BTiI04Y>F@1<D5/204eAG00404a@F`1;CeX0BdmJ
04U=F0<0BTiI0`1;CeX304Y>F@<0BDeH0`1:CUT304]?FP<0BTiI3P19CEP<04];E``0BTYF3017Ae<0
1019BED0B4QD04Q8E016AU8304E5D@80ATIB00<0B4QD04U9E@19BED00`18B5@404M7Da40ATIB1014
BDd304A9C`<0A4QA0P14B5<01@16AU80AdMA04M7D@18BDh0BDY>0040oooo000<04Q>FP@0B51K0`17
CeX504I>F@`0Ae1I3018DEX<04Q@F``0BE5L00@0B51M04YBG`1:DUl0Be=P0P1:DUd01017CeX0A4eF
04A=EP13C5D204A=E00404I7B`0j>cd0>S/m02X/:`807alM00<06Q/F01/J5@0K6QD00`0J6Q80100R
7Q<08QhE028N5@0Q7A@3020K5@808A`H100R7AT201dH4P0402@P5`0a:R00<BXP03@^8P80=BdP00@0
<R/K02`S500/8a@0;R/R0P0^:Ql0100e;b<0=2/L03@[700Z7@`202`K2@0403XX4P15<1/0AC0K058d
6080F3TM00@0IT@X079?<@1bCc40ME4_0P1cCBT0101dC2P0NTl/07Y?;01hDC0207I?;P0406m89`1T
?Ah0I3dN05`e5P80GSLJ00@0JT<V07E=<`1eCC<0H3LQ0P1N=Al00`1L<ad0FS4K05Xa6`0305L^6004
05he7`1V?BL0ISdW06Lo9@80JD4W00@0N54d08IO@P26Gd80Pe`m0P1lECH0101mESD0PU/j089K>P2C
IdX209IZC@0408UM@01gBR/0MdX[06e07`80JcdL00@0M4HT08MF=@27ESD0QUDe0P21D3000`1iB2P0
P4l_081?;`03099Q@@0408QG=`26ECD0QUDe08eK>080U68m00@0UV@o09QT?@2HI3d0WFM10P2UL4P0
102SK4D0VV<l09YS?02WL4T20:]dC@040:I_B02KI3d0Vf@m09=L=@80WFHo00@0VF8k0:efC`2]MTl0
/WM?0P2`MDd0102^Ld/0[7590:aaB@2]LTX20:mdC0040;9gC`2eNU80]GYB0;5hC@80/WU>0`2cNTl0
102aN4d0[gI;0:mfB`2/LdP20:YaAP040:]a@`2ZL480ZW120:M]?`80YV`n00@0Yfdo0:Y`@P2ZL480
[WA60P2`MTP0102YM480ZW910:Yb@@2XL3l30:Y`@080[6m000@0ZF`m0:MY>P2WJCX0YFHc0P2WJ3D0
102WJCH0YFLd0:EW=02QIC4209]R;@<0VF0[00@0Vf<`09YP;P2JH2h0UehY0P2EG2L0102CGR`0SEd_
08eM;`20EB/2079;900406M2:01H=1X0F3@J044Q2P80>Ad700@0@R`G04Ld8`17=2<0@30P0P0h:a/0
100m:QT0BcPW04/h9`1E@bl204/i8`0404lm9`1@?RH0D3hV03l]5@80@BlG00@0<20604Hf7016=Q`0
@C4G0P0j:Q40100m;AH0>BTB03TY4P0l:aL204@c7`0404De:00j:1X0>RPJ03XW6080BSLV00@0ETDc
04a0<01<@300>c<X0P0^:Ql01012@cd0B4U404Q9A01;BdT204Q9B`0404E8C@18Be@0B4]D04]?FP80
CE5M00@0BTiI04U=F019CEP0B4aG1@17BeH304Q<E`@0BDeH0`1:CUT604]?FP80BDeH00@0B4aG04Y>
F@1:CUT0C51K0P1=DE`0101<D5/0BdmJ04]?FP19CEP304Y>F@<0BdmJ0`1:CUT304U=F0<0BTiI0`1;
CeX304Y>F@h0BDeH301;BeL<04Y:EP`0AdMC00@0BDUE04Q8E018B5@0ATIB0`15AE4204I6DP0304Q8
E019BED0BDUE00<0B4QD1017Ae<A04I6DP@0A4U=0`14BDl304A8D@80A4QC00D0ATIB04M7D@17Ae40
B4U>04U:CP010?ooo`003018CUX404U?F`<0B4iJ1@17CET<04Q?F@`0BE1J3018CUX<04Y@G00404Q@
G@19DEh0BE5N04UAGP80BE5L00@0AdmJ04E>E`15CUL0A4eF0P13C5<01017C500?41303a0@`0^;c42
024S800401`K5`0K6a<06a/C01dJ4@807QXA00@08QhC028N5@0R7QD08AdD0`0P6aD2024L60@08QdI
0P0O6QH0100R7AL0;RLO02hW7`0e;R@203L_8P04030X6`0Y8140:B0A02d]8@80;B/N00@0<BhO03<]
7@0c;Ad0:R4@0P0S60@0100Z7PP0=RLB03HW4P18;1@204`a5P0405Tk8@1UART0IDHY06]::`80JdLW
00@0KTXX07E>;@1eCRd0L4d]0P1cD300101^Bbd0G3TM05`i7@189@T204HR200405Pd6P1/B2h0K4P^
05Th9@80EC@Q00@0ESDR05`k:01L>bP0FCPU0P1?;Q/01019:1D0BbXG04/Z5`1D<1P205Pd7004060l
901[Abd0JdL]075=<`80KT/_00@0K4T]06m<;P1_C2h0KDD[0P1gCC@0101iD3@0Md/`07M;<01_@bH2
068e6004068e5P1cA2H0Ld@V07118`80MDHX00@0LT<U06m08P1_@280MDHX0P1iBR`01021DS@0TF94
095RA021Dbl208EH<@0408=C;@1oCRH0OdhV08MC:`80UF0f00@0Tehb08MA9@27DBD0VFDm0P2MJD40
102NJT80VfLo09]W?`2EHCT209aX@00409QT?02_Ne<0[g]C0;1eC@80[gA<102^Ld/20:mdC0040;5f
CP2bMdl0/WM?0;1eB`D0/GI<00@0/7E;0:icB@2^LdT0[7570P2[L4H60:U_@@040:Y`@P2[LD<0Zg53
0:abA080[G=500@0Y6lm0:E]?02UKC`0YVhm0P2YKcl0102[LD40Zfho0:]^?`2WJS/20:EW>0<0Xf@a
00@0XV@a0:5S<02QHc00WV8^0P2KHRd309UP:`0409YR;`2HHB/0V64[09QO:P80Uel/00@0U5h_08QI
;@28FBd0N4hX0P1ZA200101P?RD0DRlI058_6@0j6`H203PM200404/d8P1=>RT0CCXY04Dd9080ACDV
00@0DS`W048/5`12;1L0E3hW0P1A>b@0100o:A40BSDJ04Xe6P1<=a`203dX3@0403d]4`17=ah0AcLN
048b6@80=BD>00@0<B4:03TX500i:1@0AC@P0P1;>RH0100g9aP0<b8B03<R4P0m:QT205=0;`0405Y9
>@18?2h0B3`^03D]8P80;BTP00@0=cLe03hn?00n?S`0ATI60P19BTh01018Be00BDaE04U<E@19CEP2
04Y>F@<0BDeH1P17BeH00`18C5L0BDeH04U=F00404U=F0<0BTiI1@1;CeX0101:CUT0BDeH04U=F018
C5L204U=F00404]?FP1=DE`0CE5L04]?FP80BTiI00@0BDeH04Y>F@1:CUT0BTiI0`1;CeX304Y>F@<0
BDeH0`1:CUT304]?FP<0BTiI3P19CEP<04];E``0BTYF3017Ae<01018B5@0AdMC04M7D`16AU8304E5
D@80ATIB00<0AdMC04Q8E018B5@00`18B5@404M7Da40ATIB1014BDd304A9C`<0A4QA0P14B5<01@16
AU80AdMA04M7D@17B4d0B4U=0040oooo000<04Q>FP@0BDmK0`18CUX504M=F@`0B4mI3019D5X<04Q>
FP`0BU1L1018D5d304Q@F`80Ae1I00@0ADiG04=<D`13C5<0ATeC0P0m@TH0100_<cH092HU02@V9@0O
7QX201lL5@04024M4P0S7A408adA028N4`808QhE00@08AdD020K5@0P6aD081/E0P0Q71P3028M6@04
02<N6`0R7AT08QdI024L6080:B8J00@0=2dS03P`9@0h<2D0;bLJ0P0X7Q8302d[7P0402l/7@0a:a/0
<B/K02/T4P808AT600@081P302/P300[80`0BS4K0P1:<QX0101?=A`0Ec`Q05Ll8@1K?20205/i7@04
05dj701R?b40HSlQ06a=<080LE8e00@0Ke0d06119@1P@BD0BRXA0P0l70D0100n7PL0ABD>04DU3P1?
<R8204T/700404/^7P1F>BT0ESTY05Tl;080Cc8R00@0@bHF03dP400m8100@R4>0P169A8010179a80
BRXE04XZ5@1D=1d205/k8P0405Ph7`1C<aX0Dc<J04lY4P80E2hG00@0GCDL06Y2:@1Z@RT0LdT`0P1Z
?R<0101O<aP0I3PM06@h7@1N<QL206Lk800406Hj7`1T>1d0I3PM06/o9080M4P]00@0PEDj095UBP2A
IDX0Qell0P2;H3d01022ES40M4LP07A7801fAA`208EB9`0408YE:@25Cb40QDlQ095M=@80TUhf00@0
UF4i09]W?`2KIcl0VVHn0P2NJT80102DH3P0Z7A<0:QdC02/LDT20:ebBP040:mdC02`MDd0/7E=0;5f
CP80/7E=00<0[gA<0:icB`2^Ld/01P2_M4P0102^LdL0[G960:ebAP2[L4@20:Y_@`040:U_@@2ZL480
ZW120:abA080[G=500@0[7940:Y`@P2ZL480Z6i00P2VK3h0102MJ3H0WfLf09mW=P2QJCP20:I/?004
0:Q^?P2ZKCh0ZVdn0:IY>P80XfDf0`2NGb`309aN:`0309aP;02KHRd0Vf8]00<0Vf8]00@0UV4[09MP
:P2GH2X0Uf0Z0P2HH2d0102BGRl0Q5@/08AD;01bAb@206Dn7`0405dm:01::aH0BR/F03LK1P80?B@@
00@0C3DS04@a8014<B00?RdM0P1;>b/0101F?BL0@RTC048Y4`1F?BL205U1:@0404@/4P19<AL0BC4G
04hf6P80BC4E00@0FDT`04Ph7`18>1l0A3@M0P0j:Q<0100b8@d0B3LS04Pg8`1D@c4204@c8@0402lO
3`0d8a40=2<A04@b7P80Dd4]00@0Cch^03T]8@0i;B40;bHM0P0b;BL0100l?3h0?cm103lo@@15A4T2
04Y;D00404U<D`19C5D0BDaE04Q<E@80BDeH00@0B4aG04M;EP17BeH0Ad]F0`16BUD204M;EP0304Q<
E`19CEP0BDeH00@0BDeH0`1:CUT504]?FP0404Y>F@19CEP0BDeH04Q<E`80BDeH00@0BdmJ04eAG01=
DE`0BdmJ0P1:CUT01019CEP0BTiI04Y>F@1:CUT304]?FP<0BTiI0`19CEP304Y>F@<0BdmJ0`1:CUT>
04U=F0`0Bd]G3019BED<04M7D`0304Q8E017Ae<0AdMC00H0ATIB00<0AdMC04Q8E018B5@00`18B5@4
04M7Da40ATIB1014BDd304A9C`<0A4QA0P14B5<01@15AE40ATI@04I6D017B4d0B4U=0040oooo000<
04Q>FP@0BDmK0`18CUX504M=F@`0B4mI3019D5X<04Q>FP`0BU1L1018D5d304Q@F`80Ae1I00@0ADiG
04=<D`13C5<0ATeC0P0m@TH0100_<cH092HU02@V9@0O7QX201lL5@04024M4P0S7A408adA028N4`80
8QhE00@08AdD020K5@0P6aD081/E0P0Q71P3028M6@0402<N6`0R7AT08QdI024L6080:B8J00@0=2dS
03P`9@0h<2D0;bLJ0P0X7Q8302d[7P0402l/7@0a:a/0<B/K02/T4P808AT600@081P302/P300[80`0
BS4K0P1:<QX0101?=A`0Ec`Q05Ll8@1K?20205/i7@0405dj701R?b40HSlQ06a=<080LE8e00@0Ke0d
06119@1P@BD0BRXA0P0l70D0100n7PL0ABD>04DU3P1?<R8204T/700404/^7P1F>BT0ESTY05Tl;080
Cc8R00@0@bHF03dP400m8100@R4>0P169A8010179a80BRXE04XZ5@1D=1d205/k8P0405Ph7`1C<aX0
Dc<J04lY4P80E2hG00@0GCDL06Y2:@1Z@RT0LdT`0P1Z?R<0101O<aP0I3PM06@h7@1N<QL206Lk8004
06Hj7`1T>1d0I3PM06/o9080M4P]00@0PEDj095UBP2AIDX0Qell0P2;H3d01022ES40M4LP07A7801f
AA`208EB9`0408YE:@25Cb40QDlQ095M=@80TUhf00@0UF4i09]W?`2KIcl0VVHn0P2NJT80102DH3P0
Z7A<0:QdC02/LDT20:ebBP040:mdC02`MDd0/7E=0;5fCP80/7E=00<0[gA<0:icB`2^Ld/01P2_M4P0
102^LdL0[G960:ebAP2[L4@20:Y_@`040:U_@@2ZL480ZW120:abA080[G=500@0[7940:Y`@P2ZL480
Z6i00P2VK3h0102MJ3H0WfLf09mW=P2QJCP20:I/?0040:Q^?P2ZKCh0ZVdn0:IY>P80XfDf0`2NGb`3
09aN:`0309aP;02KHRd0Vf8]00<0Vf8]00@0UV4[09MP:P2GH2X0Uf0Z0P2HH2d0102BGRl0Q5@/08AD
;01bAb@206Dn7`0405dm:01::aH0BR/F03LK1P80?B@@00@0C3DS04@a8014<B00?RdM0P1;>b/0101F
?BL0@RTC048Y4`1F?BL205U1:@0404@/4P19<AL0BC4G04hf6P80BC4E00@0FDT`04Ph7`18>1l0A3@M
0P0j:Q<0100b8@d0B3LS04Pg8`1D@c4204@c8@0402lO3`0d8a40=2<A04@b7P80Dd4]00@0Cch^03T]
8@0i;B40;bHM0P0b;BL0100l?3h0?cm103lo@@15A4T204Y;D00404U<D`19C5D0BDaE04Q<E@80BDeH
00@0B4aG04M;EP17BeH0Ad]F0`16BUD204M;EP0304Q<E`19CEP0BDeH00@0BDeH0`1:CUT504]?FP04
04Y>F@19CEP0BDeH04Q<E`80BDeH00@0BdmJ04eAG01=DE`0BdmJ0P1:CUT01019CEP0BTiI04Y>F@1:
CUT304]?FP<0BTiI0`19CEP304Y>F@<0BdmJ0`1:CUT>04U=F0`0Bd]G3019BED<04M7D`0304Q8E017
Ae<0AdMC00H0ATIB00<0AdMC04Q8E018B5@00`18B5@404M7Da40ATIB1014BDd304A9C`<0A4QA0P14
B5<01@15AE40ATI@04I6D017B4d0B4U=0040oooo000<04Q>FP@0BdmK0`1:CUX504U=F@`0BTiI301;
CeX<04]?F``0C51L0`18D5d01017Ce`0B51K04Q@F`19DE`204QAFP0404I?F013C5<0@daC04A=E080
?DA:00@0<SLj02PZ:@0X:RT08R4L0P0R7QD0100T7Q809QhA02HN4@0R7Q<2028N5@04024M500P6aD0
81/E020K5@808A`H0`0R7AT0100S7Q/09B0M02DP7@0Q71P202DN6004034Z8P0f;B@0=RdT02lW7080
:R0D00@0;RPJ038/7P0b;1h0<B/K0P0`:AT0100]9Q@09A`;02DL2`0Q6@H202TN3004040[600o:A@0
?bTD044[5080BS0I00@0CC4I04h`5P1><1H0DC8F0P1F=aX00`1L?R@0HD<Y0653:@0306I8<00405`m
:01;;1L0Bb`G03PI1@80;0d000@0A2TH03lT4`0o91<0?R<B0P16:aX304`a800404H[6P119QD0@BHE
03lO4080A2@E00@0BRXK04//6P1;;1X0DS<O0P1P@Bd0101VAc80Gd0[05m0:`1R?BX205hj900405dg
8P1X@R/0J48[07U@>P80LTXa00<0HcTQ060f7@1P=Qd00`1U?B<0101N=Q`0GcLM05lg7@1_Abd207M?
=@04079:<01aBBl0LDT_085L?`80RV=400@0Qehn07Y>:`1jCR/0M4@N0P1oCRD0102BGC<0W6Lk09aW
>`2CG3D2091I<P0409=L=@2MISl0WFHo09eV?`80Vf@m00@0Re@]09eV?`2MISl0ZVm70P2/LDT0102_
M4`0/WM?0;9gC`2bMdl20;5fCP030:icB`2/LDT0[75900@0/7=70`2_LTH20:iaA@<0[G1400@0ZFm1
0:]a@`2[LD<0[7940P2]LdD0102[LD<0Z6i00:Q^@02TJS`20:9X>P0409aW=@2NISD0WVHe09mW=P80
XfTi00@0YV`l0:MZ>`2WJS/0Y6Lh0P2RI3D0102LGBX0VU/X09YK:02GFBH209QJ9`0309UM:@2JHB`0
VV4/00<0Vf8]00@0U64Z099M9`2BGBL0UEhX0P2EH2`0102?G2l0OE0Y07e@:@1Z@B4205lj7@0405Pk
:@129a@0@RLD03LL2@80?RLE00@0@bhK03DR4@0e8Q40=B@B0P17=RH01016<1/0CcTT04li901MAc02
05Y4;@0405Do9`18<QX0B38J03hY3P80E3lT00@0DD4Z03TY4P0i:A80@C4J0P13<Qh0100m;1P0DT4_
0591;`1:>BL201l>000403DV4`0n;AT0?RdI04Xh9080BcTU00@0>bXJ02TM4@0Y7A40;2<L0P0i=300
101=C540BTU>04Y9CP19B4h204U:C`<0AdYA00@0B4aE04eAFP1=DEX0Ad]F0`16BUD204E9E00404I:
E@17BeH0Ad]F04Q<E`H0BDeH0`1:CUT604]?FP80BDeH00@0B4aG04Y>F@1:CUT0C51K0P1=DE`0101<
D5/0BdmJ04]?FP19CEP304Y>F@<0BdmJ0`1:CUT304U=F0<0BTiI0`1;CeX304Y>F@h0BDeH301;BeL<
04U9EAP0AdMC0`18B5@404M7Da40ATIB1014BDd304A9C`<0A4QA0P14B5<01@15AE40ATI@04I6D017
B4d0AdQ<0040oooo000<04Q>FP@0BdmK0`1:CUX504U=F@`0BTiI301;CeX<04a@G0`0BdmK00@0B51M
04M?G017Ce`0ATiK0P17CeX01019DE`0BE9K04UBF`16CeP204=<D`0404E?F00oB4l0?dQ?03Hk?`80
;Bl^00@09B@O02<O5P0S7aH091`A0P0U6`l0100R7Q<08QhE028N5@0Q7A@3020K5@808A`H0`0R7AT0
100S7Ad0:2<P02PS800R7AX2028K5@0402hW7`0d:b80=2/R02hV6`80;B<I00@0<2LH03T`8@0i<240
=bhO0P0a:1T0100d:1P0<2@D030T500Z7Ph202`P4004030M300_7@T0;ad903HQ3080?bTB00@0AR`E
04L[4`17:a<0BR`B0P1><1H01015:a80CS@K04hd6`1J@2T20698<@0405m4;`1B=b@0DSLT044V4`80
=AX700@0>b0?03hS4P0n8a80?b@C0P0n8a800`119QD0AR/J04H[6P0304P]700404/_8`18;200B2`P
04da8`80BblQ00@0@RLF050e901@=B@0JTll0P1bEd@0101J=b@0F3DR05Pe8P1C;Q/205`g9004071:
=@1fCCL0MTdg06e4;P80K4<]00@0ID0V06M2:01W@RP0GSTO0P1K=Q`0101U@2H0J4<Y06Q3:@1O>R02
05Xe6`04060n9@1aCC<0LDdc07aE>080LdPX00@0HcDC06Le401W=A00PdhV0P2KISh0102EFC@0TED`
095E<02DF3<209iR?@0409YN>@2DF3<0U5Pc08A88`80VUhi00@0Zg180:aaB@2/LDT0[W=;1@2`MDd3
0:icB`<0/7=60`2_LTD60:iaA0<0Zg5300@0ZW120:Q^@02XKT00Yfdo0P2UJcd30:AZ?0040:1[>@2Q
JCP0XFTh0:5Y>080XfTi00@0Y6Xj0:EX>@2UJ3T0XVDf0P2PHS<00`2LGBX0VEXW09UJ9`0309EG9003
09MK9`2HGbX0V5lZ00<0VF0[00@0TUlV091L8`2@G2<0TE`V0P2DGbd0102<FB`0N4XV07Q:9P1V?Al2
05dh7P0405@i:00n91<0?R@C03PN3@80?bPF00@0?BPG03DR4@0e8Q40>RLF0P13<R00100o;AT0DT0/
0590;01LBSH30590:P80AS@L00@0=R@<04/j801;>R00?2`E0P0e9@h01015=200@S4M048a7@0j:AL2
04Lf900403TX600H1`0060L003TZ5@80AC@P00@0CC/U04Dc7`15<al0<B0@0P0V6A00100`9b80?STf
03hi=P1AD5H204a;D@0404U8D01:BU80BTYB04Q;DP80ATU@00<0Ad]D04a@F@1<D5T00`16BUD304E9
E00404I:E@17BeH0Ad]F04Q<E`H0BDeH0`1:CUT504]?FP0404a@F`1:CUT0BTiI04U=F080BTiI00@0
CE5L04iBG@1>DUd0CE5L0P1;CeX01019CEP0BTiI04Y>F@1:CUT304]?FP<0BTiI0`19CEP304Y>F@<0
BdmJ0`1:CUT>04U=F0`0Bd]G3018B5@H04M7D`<0B4QD1017Ae<A04I6DP@0A4U=0`14BDl304A8D@80
A4QC00D0ADEA04E5C`15ADl0ATM<04M8C0010?ooo`003018CUX404]?F`<0BTiJ1@19CET<04Y>F@`0
BdmJ301<D5`<04]?F`0404Q@G@17Ce`0AdmL04I>F`80AdmJ00@0BE5L04UBF`19DU/0ATmH0P13C5<0
1015CeP0?dQ?03m8C`0f>cl202d_;P0402DT7`0S7aH08alF02@L4@809A/?00@08QhC028N5@0R7QD0
8AdD0`0P6aD2024L60<08QdI00@08adM02PS800X8b008QdJ0P0R6aD0100^9al0=2/R03@[8P0^9Q/2
02dS6@04030W600i<240>C0Q03L^7`80<BPI00@0=2PH030T500`91@0:Qh>0P0/8100100`7@`0;ad9
02lM2@0f8@`203lY4P0404H/5@17:a<0Ab/C04X/4P80CS0F00@0AB/B04hd6`1>=1/0FT0Y0P1RB340
101OA2l0DSLT058g90119Q<203DJ1`0403/P3`0n8a80?R<B03lT4`80?R<B00<0@BHE04H[6P16:aX0
0`18;A`0101;;b<0B2`P04P/801=<B<204/_8@04048W5P1@=B@0D3DT06Y??080LUM400@0FSLT05Pe
8P1H=B80DbhK0P1L=b@0101`BSD0MTdg07I==`1]A2h206a3;@0406E09P1W@RP0Id8X05hi7`80FcHL
00@0ID0V06Q3:@1X@bT0GcXP0P1J=A/0101P?RD0LDdc075=<`1lECP207=8:00406<e4`1W=A00IcD@
08=>9P80VfHn00@0UETd095E<02AEC00U5Pc0P2NHSd0102JGST0U5Pc09AH<`24B2<209YN>@040:]`
B02/LDT0[7590:icB`D0/7E=0`2^Ld/30;1cAP<0[g951P2^LD@30:]a@`040:Y`@P2XKT00Z6i00:M]
?`80YF/m0`2TJS`0102PJcT0XFTh0:5Y>02QJCP20:=Y>@040:AZ>P2UJ3T0YFPi0:9U=P80X68c00<0
W5dZ09UJ9`2IFRL00`2EEb@00`2GFbL0V5lZ09QO:P0309UP:`04099O9P2@G2<0T5`S095L9P80U5l]
00@0S5T/07Q:9P1hBRH0ISdO0P1M>1h0101D>BP0?R@C03hT4`0h7Pd203lX5P0403dX5`0e8Q40=B8A
03XW5P80@c8P00@0?bdI0590;01B@2`0G4Xf0`1B@2X204Hd700403HT301;>R00BcXP03`/5@80=BD>
00@0AC@P048a7@12<Ad0>RTG0P17=R@0100i:1P060L001P7000i:QD204Dd800404dk9@15<al0AC<O
034P40809QT@00@0<2LR03hi=P0n>CH0DE1F0P1<Be401019B500BTYB04Y:DP18Be8204I9D00304M;
E01<D5T0C51I00<0ATYE0`15BE@01016BUD0Ad]F04M;EP18C5L604U=F0<0BTiI1@1;CeX0101<D5/0
BTiI04Y>F@19CEP204Y>F@0404eAG01>DUd0CU9M04eAG080BdmJ00@0BDeH04Y>F@1:CUT0BTiI0`1;
CeX304Y>F@<0BDeH0`1:CUT304]?FP<0BTiI3P19CEP<04];E``0B4QD6017Ae<304Q8E0@0AdMC4@16
AU8404A9C@<0A4U?0`14B54204A8D`0504E5D@15ADl0ADE?04I7C017B4`00@3oool000`0B4iJ101;
Ce/304Y>FPD0BDeI301:CUT<04]?FP`0C51L301:CUX01018D5d0AdmL04M?G016CU/204M?FP0404UA
G019DU/0BE9K04I?F080@daC00@0AU1I045:D@11BU40>Ci20P0`<S40100W9R408alD02<O500S6`h2
02<J2`04028N4`0R7QD08QhE024M50<081/E0P0Q71P3028M6@0402<M7@0Z92@0:R@T02<N6`8081PE
00@0;2DM038Y800b:B00;RHK0P0_9A/0100c9aT0@3@V040d9P0l<20203DX600303T/700h:a/0>2/K
00<0<B@D00@0;Qd=02hK2P0^6`X0<b4=0P0m9a801012:A<0?bD>03lU3P0n8PX203lT2@04040V3`1=
<a`0CC<L05]0:`80GT<^00@0ECXW04`a7P1<<Ah0B2dL0P19;Qd0100h7@X0A2TF04@Y5P19;Q/204<X
5@0403hS40129a@0@RLD04L/6@80BBhK00@0?B4F03`P5@0l81D0Cc<W0P1B=RX0100g6`d0<1D4030E
1014:AP204db8@0405di:@1N>bP0GS/X05Tf8`80FcHS00<0J48_06a6<@1/AS400`1X?b/0101A;A<0
GSXP05hj801N>R0205@`5P04058^501D<1H0E30F05Te6`80HShT00@0Dc<N06]8<P1[B380OEPn0P1c
BRh0101I:P`0DR00058P001/>1<208eG<@0409AF<@2@DRd0T58]09AF<@80WV0k00@0UeTd091B;@2@
DRd0PdDP0P2NH3/60:ebBP<0[W=;00<0[gA<0;1eC@2`MDd00`2aLTH00`2`LDD0[g140:m`A0060:m`
A0040:ecA@2ZL480ZW120:M]?`80Y6Xl00@0XfTk0:AZ?02TJS`0YV`n0P2XKT00102UL3h0YFdl0:E]
?02SJcX30:AZ>P80YFPi00@0XF@e09mQ<P2OHC80WUl/0P2JFbP0102EEb@0U5HS09AF8`2EFBD209IM
:0<0UehY00@0TEhU08eJ8@2=FR40SeXT0P2BGB/0102:EbX0MTPT07I8901S?1l205/g7@0404le9P0l
8Q<0?28C03PQ4@80@2TI00@0?RTH03l/6`0o;1/0Ac@S0P17=R@01019>2P0BSTW04Xi9`1EA38204Xi
9@0403l^6P13<a`0@c<L03/[5080=bL>00@0@C4J04Ti8P19>B80DD0/0P0h9aD0100Z6@L0>2LG03PW
5`0k:QX203`[6`0403DV4@14=1d0A3@M04hl9P80AC<O00@0<b8B02/N5@0[7QD0=B`W0P10>SX01017
ATh0A4=;04A3B`16ADd204];D`0404Y=E016BE00ATU@04=8CP80ATYC00<0ATYE04E9E015BE@01015
BE@204M;EP0304Q<E`19CEP0BDeH00@0BDeH0`1:CUT504]?FP0404eAG01;CeX0BdmJ04Y>F@80BdmJ
00@0CU9M04mCGP1?Deh0CU9M0P1<D5/01019CEP0BTiI04Y>F@1:CUT304]?FP<0BTiI0`19CEP304Y>
F@<0BdmJ0`1:CUT>04U=F0`0Bd]G3018B5@<04M7D`0304I6DP17Ae<0AdMC00L0AdMC0P16AU8304Q8
E0@0AdMC4@16AU8404A9C@<0A4U?0`14B54204A8D`0504A4D015ADl0ADE?04I7C017B4`00@3oool0
00`0AdiH1019D5X304Q?F@D0AdiH0`18CeT404U@FPD0BU5K601;DU`404U@FP<0B4mI1@17CUP304U?
F`<0B51K00<0Ae5K04EBF`15DU/00`15DU/0101<DEL0ATU@04I9D00k?T<2034`=@0402LU:00R7Qd0
8QhM01hH608071LD0`0T7aT302<N60<08QdG0`0Q71H0100W8Ql08QdI028M6@0V8A/202/T700402TR
600_9a`0;bLL03H^8@80=B`M00@0=2PH03``800l<200@SHV0P0o<b<0100j;Qh0>2`L03P/700j;Qh2
03``800402`Q3@0j;a/0>RlK03hc8@80;b@B00@081@4024E1@0Q5@D0;R8D0P0h;1h0100k:A/0@c4S
04<a8`1=>bd305A2=080CS`^00@0A38T03`Z700l:Q`0>2HH0P0a7a40100k:A/0AC<U04Dc9@0l:Q`2
03DS5@0403@R500`7Q00<1h@03HX7@80>bdR00@0=BLL03`^8`0l;R<0>2XO0P0f:1d0100a8aP0@C<X
044c:01E?S0204D^800404H_7`1?=BH0CcDV04hd8`80E3TX00<0G44^05e0;P1M@2h00`1?;Q/0101B
<Ah0EC@Q05Dd8@1A<1d204XY5P0404/Z5`1B<Ah0DS4N05@a6`80J4@^00@0QEm807]C>P1kDcX0HSPO
0P1J;Q40101E:0/0Gc0B05l`4P2=DS8209QM>`0409aP?02OHch0Wf<n09MI<P80Se0W00@0NS/@09EF
:P2EERX0ZVY40P2ZK4D0102ZK4<0ZFi20:U^@P2ZKd<20:]cA0030:adA@2]MDH0[GE600`0[g9500@0
[6m20:]^@@2[KT40ZVe00P2XJch00`2VJC`0YFPk0:EX>`030:=V>@040:AZ?02UJcd0YF/m0:I/?P80
XfTk00<0WF<e09UO<@2IGc400`2IGc40102PH300WEd]09eM;@2GFBX2095D9@0408aB9028DR@0R58T
08I@8P80Q50Q00@0ReDW08iI;@2>FBd0PThV0P22EBh01023F3H0LT//079;;01U@RH205Dc6004040Z
4P0n:140?RPA03hY6080?b`M00@0?R/M03HU5@0f9AD0>RTG0P16=Ql0100`80T0D40Y0510:@1=?2P2
03TX500403l^7011<1h0@C0N03TX6080?2/K00@0Ac`Z04Hk:@16>bT0?34O0P0/8@l00`0X7@/0<2DC
030U4`0303L/6P0403DV4@15>2D0ACPU04Tn;080=2/L00@08a/>02XS6@0Z8aT0?3La0P1:ACl404Y;
D0<0BDUA0P19BE<00`18B580B4QD04Q8E00@04a<F0<0CDeI1@1>CUX<04eAG0@0BdmJ0`1:CUT504U=
F0`0BTiI3`1<C5P304];E`<0BTYF0`19BED304Q8E0@0AdMC4@16AU8?04M7D`<0ATIB0`15AE4604A4
D0<0ADEA0`16AU8304M7D`@0@dQ<0`13B4h304=7D080@dMB00D0ATIB04I6D016AU00AdQ=04Q9C@01
0?ooo`003017CUP404U@FP<0B4mI1@17CUP304Q?F@@0BE1J1@1:DE/H04]BG0@0BE1J0`18CeT504M>
F0<0BDmK0`18D5/00`17DE/0AE9K04EBF`0304EBF`0404aAE`16BE00ATU@03/n@`80<C0e00@09bDX
028N7@0R7Qd07QPH0P0L5a@302@O6@<08ahH0`0R7AL3024L5P0402LR7`0R7AT08QdI02HQ6`80:b@L
00@0:B8H02lW700_9a`0=RhQ0P0e;1d0100d:1P0?30P03``8012=RH203lc8`0403X^7P0h;1`0>2`L
03X^7P80?30P00@0;24=03X_6`0j;a/0?S<Q0P0_9180100P50@08AD5024E1@0^8Q@203P/7P0403/Y
6`13<B<0@c4S04dk;@<0E48d0P1>?2h01014<R@0?2XL03`Z700h9QP2034O4@0403/Y6`15<bD0AC<U
03`Z7080=B<E00@0=28D030N400`7Q00=RPM0P0k;B80100e9a`0?2hS03`^8`0h:Ql203HX7@04034S
6011<bP0@C<X05Dn<080ABhP00@0ARlO04le9P1?=BH0CS@S0P1D>BP00`1L@Bh0GD0^05e0;P0304l^
6`04058a7P1E=240EC@Q054`7@80BRTF00@0BbXG058a7P1B<Ah0E34K0P1XA2h01025GdP0Ne<j07]C
>P1R>1l205X^4@0405DX2`1O<180Gc0B08eB<P80V5dk00@0W60l09mS?P2OHch0UeTb0P2?D2L0101j
>a00UEHZ09EF:P2ZJT@20:Y/A@040:Y/@`2YKT80ZFi20:Y_@`80Zg=400<0[7A50:eeAP2]MDH0302_
LTD0102/Kd80Zfi10:]^@@2ZKD020:Q[?P030:IY?02UJ3/0YFPk00<0XfHi00@0Y6Xl0:E[?@2UJcd0
YV`n0P2SJC/00`2MHcD0VEla09UO<@0309UO<@040:1P<02MGBd0WEd]09MI:P80TE@U00@0S58T08QB
9028DR@0QU0R0P24D240102;EBL0SUT]08iI;@22CRH2089E;P0408=H=P1bBb`0LT//06E29P80EC<H
00@0@2XB03hX4@0n:140?RTH0P0o;1d0100n:ad0=RDE03HU5@0j:AL204Hf7`04030P2@1@@2T0D40Y
04dl:080>BPD00@0?bhL044`7P11<1h0>BPH0P0l:a/01017?2X0AS/Y04Hk:@0l<Al202`Q3`0302PM
2`0`9A<0<2DC00<0=b`J00@0=BHA04Dh9@15>2D0BCh/0P0d:a`0100S6`h0:R<I02XS6@0l=c4204Y5
?`@0BT]@0`19BE4204U9D`0304Q8DP18B5@0B4QD0100C4aH0`1=CET504i>FP`0CE5L101;CeX304Y>
F@D0BDeH301:CUT?04a<F0<0Bd]G0`1:BUH304U9E@<0B4QD1017Ae<A04I6DPl0AdMC0`16AU8304E5
D@H0A4A@0`15AE4304I6DP<0AdMC1013B4`304=8CP<0@dM@0P13Ae801@16AU80ATI@04I6D017B4d0
B4U=0040oooo000<04Q?F@@0BE1J0`18CeT504M>F0<0B4mI1019D5X504YAFaP0Be9L1019D5X504Q?
F@<0AdiH0`19Ce/304Q@F`<0Ae5K0`15DU/0101=DEX0AdaB04M<DP0m@4D203@e>@0402/Y;00U8B80
9B4R024K6`807QTF0`0T7aT302<N60<08QdG0`0Q71H0100T7a`081/G020K5`0V8A/202hW7`0402hW
7@0e;B80=BdR03/c9P80>2lP00@0;b<C03DY6@0e:AT0?C4Q0P0n<R80100i;Ad0=RXJ03HZ6P0i;Ad2
040d900402hS4@0`9A<0<2DC030T5080;R8D00@0:Ad?02PL400X7100<2@H0P0i;B401015<bD0=bDG
03LU5`0`7Q0203d[7@0405=1<`1HASP0F4Hh04Dc9@80;ad?00@0AC<U030N400`7Q00:QP:0P0b8180
100k:A/0@S0R048`8P10;R0203<Q4`0402dO500d9Q/0=2HK02/M4P80;AlD00@0;b4F03T[800i:b00
=RPM0P0o<BH01014;R00?BLI03dW6@12:ad204Pa8@0404H/6`1:;ah0BRlN05Hk:080FSl/00@0DC0M
04PW50189a@0AB@A0P1@;a`0101K>RL0FCPU05Th9@1>;AX204DT4@0404l/601J=b<0FSLS075==`80
Lddf00@0IclU05P/4@1H;140CB440P1N<A@01028D2l0TUXi099J>@2FFcT209QN>@04099F<@2;CBH0
RddV07Tk4P80U5L[00@0Z6Y30:U[A02YJd@0Z6e30P2YKT@0102[L4@0[7=60:acAP2]MDH20:ifA``0
/7=600@0[6m20:]^@@2[KT40ZVe00P2XJch00`2WJSd0YFPk0:EX>`030:AW>P0409iT=P2PISP0X6Hh
0:5W>@80X6Hh00@0WF<e09]Q<`2KHC<0VV0b0P2KHC<309eM;`0409YL;@2EF2/0UEP[08eE9P80Qe4S
00@0PdlP089=8@22CB40RE<W0P26DBL01024Db/0QeTe08MI=@1dBb/206@o8P0405lm8P18:0l0B2P?
048^5P80>B@?00@0>bHE04Db9015<R@0A34S0P0k:QX0100k:QH0?RlH03h_600h:14204`l9@0404`k
9`11<1`0@C0L03l^7080>RTI00@0=bHF040_7`10;al0@3DS0P0d:AL0100X7@/09Q/902HK2@0]8Q02
03DZ600403/`7P0n<b40?S<Q04@e8080CD0]00@0BSl]03@[700d:a`09AdB0P0/9A/00`0l=c40AT<l
04I3?00404Q9CP80B4Q@00@0BDUA04U9D`19BE<0BDUC0P19BED@04a<F0<0CDeI1@1>CUX<04eAG0@0
BdmJ0`1:CUT504U=F0`0BTiI3`1<C5P304];E`<0BTYF0`19BED404Q8E0<0AdMC4@16AU8?04M7D`<0
ATIB0`15AE4304A4D0H0ADEA1P16AU8404=8C0<0@dQ>0`13Ae0204=7DP0504E5D@15ADl0ADE?04I7
C017B4`00@3oool000`0B4mI1019D5X304Q?F@D0AdiH0`18CeT404U@FPD0BU5K601;DU`304YAF`<0
BE1J1018CeT204M>F00404]?F`19Ce/0BDmK04U?F`<0B51K0`17DE/204EBF`0404]BFP18C5D0B4aE
0413BP80>3Tm00@0;bla02LU9P0W9BH08alN0P0P6QX302@O6@<08ahH0`0R7AL3024L5P0402<N6`0N
6AD07QTE02@O6@80;BHN00@0;RLM03H^8`0f;R<0>S8U0P0f;Ah0100_8a<0<2@D030T500g:a/203hb
8P0403/_7`0c9aL0<bLG03<W5`80>blO00@0A3PZ03P/7P0h;1h0<2@H0P0b9QX0100b9A`0:ahF02/N
5P0X6a<202/N5P0402XJ3@0e9AP0=BDH03l_8P80@30S00@0>BTL03HV6@0f9QT0?blR0P1:>Rd01011
<B80>RXK03XZ6`0f9QL203DU5P0403HV5`0k:a`0>b/L03XZ6`80;al@00@0;b4F03T[800i:b00;AlD
0P0Y6a000`0X6Pl0<R@I038T6@0302PJ3`0403dZ700i9QL0>BHG03XT5P80?BPG00<0>b@D03lU500o
9A@00`19;ah00`1:;A/0ARTG04HY5`0604@W5@0404HY5`1:;A/0BRdK04LX5080BRTF00@0F3DQ06M3
;@1W@bd0JD</0P1B:Q0010136@00FblD05/_5020Bb`208]D=@0408]E<`2@F3D0T5Pe08eE<080QT`V
00@0N3hF099G;@2BEbd0YFY20P2VJd<0102WK4@0ZFi40:U^A02[LTL20:edA`030:ieB02_MdP0[gM8
00`0/GA70`2[KT40102ZKD00Z6/n0:Q[?P2WJSd20:IY?0<0YFPk00<0W68d09iT=P2NI3H00`2OICL0
0`2LHS@0VV0b09YP<P0309UO<@0409YJ;02LGC40W5da09aO<P80UU/_00@0SE@W08A>8P24CR80P4/O
0P1oC2401020Bb40NTHN07Y67P21DB/208AH=@0406Ln801H=AT0F3DI05Lg7P80@B4:00@0B3<N03@O
300d7``0=b@E0P19=RP01014<R@0>bXJ03/Z6P0h:A@203HW4004040a7012<ah0@S<N04Df8@80ASLT
00@0@C8O03TY6@0i:AT0>BTI0P10<240100/8a807aH501lF1@0I4002028I200402/R4@0/8a80;2<B
02lV5@80=B`K00@0?c8O04Dh9@15>2D0@3@T0P0b:AX0100^:1`0>CD/03Te;018ACh2051=B00404E6
B`16Ad`0ATM<04I7C080AdM?00@0B4Q@04U9D`19BE<0BDUC0P1:BUH@04a<F0<0CDeI1@1>CUX<04eA
G0@0BdmJ0`1:CUTA04U=F0l0C4aH0`1;BeL304Y:EP@0BDUE1@18B5@304M7D`l0ATIB3`17Ae<304I6
DP<0ADEA0`14A50604I6DPH0ADEA1013B4`304=8CP<0@dM@0P13Ae801@14A500ADE?04E5C`16Ad`0
ATM;0040oooo000<04Q?F@@0BE1J0`18CeT504M>F0<0B4mI1019D5X504YAFaP0Be9L0`1:DE/304U@
FP@0B4mI0P17CUP0101;Ce/0BDmK04U?F`19Ce/304Q@F`<0Ae5K0P15DU/0101;DUX0B4aE04Q<E@10
@dX203Pi?@0402l_<@0W9BH09bDV02<O7P8081XJ0`0T7aT302<N60<08QdG0`0Q71H0100S7Q/07QTE
01hI5@0T7aT202dV7P0402hW7@0f;R<0=RhS03Xb9@80=RdN00@0;b<C030T500`91@0=b/K0P0n<R80
100k;al0<bLG03<W5`0c9aL203/_7`0404@h:P0h;1h0>2`N030T6080<RHJ00@0<RDL02/N5P0[7QH0
:1/C0P0[7QH0100Z6Pd0=BDH03DU600o;b82040`8`0403TY700f9QT0=RHI03l_8P80BSX]00@0@C4R
03XZ6`0j:Q/0=RHG0P0e9AH0100f9QL0>b/L03/[700j:Q/202lO400402lQ5P0i:b00>B/P02dO5080
:A/@00<0:1X?038T6@0b91T00`0X6Pl0100m:Q`0>BHG03TV5`0j91H203dX5`0303/T500o9A@0?bDD
00<0BBlN00<0BRdK04HY5`16:AL01P149aD01016:AL0BRdK04X]6`17:1@204XY5P0405Pe8@1W@bd0
Id<]06U3;080DRX@00@0@aT005/_501K;a@0P4//0P2;E3D0102;EC<0T5Pe091H=@2=EC0208I<9P04
07Pn5P2BEbd0TUL]0:EZ@P80YV]300@0Yfa40:U^A02YKT@0Zg970P2]M4L00`2^MDP0[gM80:mgB00<
0;5dA`<0Zfi100@0ZVe00:Q[?P2XJch0YfXm0P2VJC`30:EX>`0309aR=02NI3H0WV@f00<0WfDg00<0
W68d09YP<P2JH3800`2IGc40102JFR`0W5da09aM<@2LGc8209IK;`0408eD9`24CR80Q4hR081;7`80
Od`Q00@0P4/Q07Y67P1jAQh0PE4[0P24F3D0101W?R00F3DI05Pe6@1G=ah2044Q2P0404Pc7P0d7``0
=1l<03LT5@80BCHX00@0A38T03/Z6P0k:QX0>2TD0P0f9a001010<A`0@S<N048c7P15=R4204Hg9004
044b7`0i:AT0>BTI03TY6@80@30Q00@0;2<B01lF1@0O5PD06A000P0R6@P0100[8Q40;2<B02`S4P0_
9QD203D/6`0403lb7`15>2D0ACPU040d9080<RTJ00@0;RPL03Te;00i=B`0B4Dn0P1@CDP01015AT/0
ATM<04I7C016Ad`204M7C`0404Q8D019BE<0BDUC04U9D`80BTYF401<C5P304e=F@D0CTiJ301=DE`4
04]?FP<0BTiI4@19CEP?04a<F0<0Bd]G0`1:BUH404U9E@D0B4QD0`17Ae<?04I6DPl0AdMC0`16AU83
04E5D@<0A4A@1P16AU8604E5D@@0@dQ<0`13B4h304=7D080@dMB00D0A4A@04E5C`15ADl0ATM<04I7
B`010?ooo`004019D5X304Q?F@D0AdiH0`18CeT404U@FPD0BU5K601;DU`404YAF`<0BE1J1@18CeT3
04]?F`<0BDmK0`18D5/304MAF`0404U@FP16CED0ATeE0456C080>Se200@0<c8g02XX:`0Z:2/0920Q
0P0P71/302@O6@<08ahH0`0R7AL3024L5P0402@O700N6AD07QTE024L5P80:24I00@0:B8H034Y7P0a
:Ah0>30S0P0f;Ah0100g:a/0;b<C02lS4`0`91@203da8@04040d900d:1P0=2PH02/O3`80;20@00@0
?C4U03<V7@0c9Qd0;R4H0P0b9Ad0100g:R80=bXT03LZ900d9b4203<U8P0402hN4@0b8QD0<R8E03PX
6`80?2`O00@0>BTL03LW6P0g9aX0?RhQ0P18>2/0100n;Ql0@30Q040`8@0k:a`2030P4@0302`L3@0e
9AH0=BDF00<0?RhO00@0>B`S044d:`11=2/0=BPO0P0a91/0100/7aH0<RDL038U700[7QD2030S6P04
040^800k:A/0>bTK02lL3P80<Ah?00@0>B@C03lX600o:1P0ARlM0P0l9A<00`0f6`X0@BHE044V5@03
04/`7`04040U500f6`X0=Q/:03@I2080=a`;00@0@RDE044T501191@0@b@B0P1D<b00101R?bT0CBL@
04dW400n6@0205H^5@0407E49P1oCS00Odh`085>;`80QeDd00@0SEPf089<:022C2P0N48L0P2;ERh0
102OJ440XFY30:5Z@`2SK4<20:Q^AP030:]bA`2^LdT0[W=900<0[gA8302aM4L0102[KT40ZVe00:Y]
@02YK3l20:Q[?P030:MZ?@2VJC`0YVTl00<0YFPk102QIcT20:1V>00409aR=02IGc40VEla09EK;@80
U5X/00@0V5T^09aM<P2LGC80W5lc0P2DFBl01027DBD0OdXP07m:801mBQl207i:8P0407M3701b@1/0
LT0K07E79@80M4TY00@0GcXO05Dc6P1E<aX0D34L0P129Q40101;>BD0<1d>030M3P0e8Q@204Dc9`04
03XX700`8100<20@03DV4@80<RH<00@0@S<N03`]600l;AP0@S<P0P18>BH01011<B40?RhO03h^7`0o
;b0203XZ6`0401XA000H3`0060l001lF1@80:b8A00@0;RDD02XQ400Z8A00:b8A0P0c:QT0100c9Q<0
=R/I03H[6@0e;1/203D]800403`e:`19AC`0BDDl055>B@80DE1;00@0@dA904A5BP14ADX0ADI;0P16
ATh01018B500BDUC04U9D`1:BU@204];Ea00C4aH0`1=CET504i>FP`0CE5L101;CeX304Y>FA40BDeH
3`1<C5P304];E`<0BTYF1P19BED404Q8E140AdMC3018B5@304M7D`<0ATIB0`15AE4304A4D0<0AdMC
0`16AU8304E5D@<0A4A@1013B4`304=8CP<0@dM@0P13Ae801@13@dl0A4A>04A4CP15AT/0ADI:0040
oooo000@04U@FP<0B4mI1@17CUP304Q?F@@0BE1J1@1:DE/K04]BG0<0BU5K0`19D5X304Q?F@<0C4iK
00@0BdmK04U?F`19Ce/0BDmK1@18D5/01015CUL0ADaF04E<EP13Ae0203i1B00403Hg>`0/;2h0;2`^
02@R8`807Q`M0`0T7aT302<N60<08QdG0`0Q71H0100U81d07aXF01lJ5P0P6aD202DN5P0402DN500a
:Ah0<BTN03hf:@80@CPY00@0?c<S02hR4P0^8Q80:Ad=0P0j;Qh01016>RX0?C4Q03da8@0/810202<G
1`0402TM4@0[7QD0:ahE02/N5@80:ahF00@0<2<K03P[9@0h:bD0>bhX0P0h:RL0100c9AX0:adB02/M
4P0Z714203@V6`0403`^8`0k;B80>bdR03HX7@80=BLL0`19?2`00`10<b<0=BPH03DX600303DX6004
02`O3`0Q50@08A@404/n=@80CD0g00@0?c8Y03`_9P0l;bH0<B@K0P0^8AP0100/7aH0?C0W03d`9`0n
;R42044a900402lO400_7@l0;ad?03PU5P80?R/J00@0BC@S03lZ6@0o:QT0@BXJ0P0l9AD0100i8Q80
>b@D03/T5010:AT203lX600403@M3@0Y4P80:A8203dS5P80@2HG00@0>Ad?048U4`129A<0Dc8O0P17
910010117@L0DRhH058^601X>R020794:P0407I9;01oD380Oe0b08eL?080ODdY00@0MdDR08EC;P25
Dbh0VfM00P2NJD40102QK4@0YFi50:E^A@2ZL4P20:aaA`<0[G98302_LTD30:Y]@0040:U/?`2WJSd0
YfXm0:IY?080YFPk00<0Y6Lj0:=V>@2SIST0102QIcT20:1V>00409iT=P2JH380VV0b09IL;P80TeT[
00@0UeT`09QJ<@2HFS40UEX`0P2=Db/01021BR<0NTDM07Y57@1jAQl207a;8`0407A27`1eA2<0MD@S
06Xm7P80GcHJ00@0GSXR05@d7`1D=1l0A2LE0P17;1/0101:=bP0<QlA038O4@0c8AD203`Z800402XJ
3@0Q50<08A@302hR3080=2P>00@0?2lL040c8010<b00Bch[0P16>BP0100f:AT0>2/K03P[6`0k;B02
02lQ500401XC0`0N5`L07QL702DN3P80;2DE0`0`:AT0100b:a/0=2dM03@]7@0_9182038Z5`0403P_
800m=b/0?CL[04A0=@80BTM000<0AdI10450?011@3`00`13A4T01014ADX0ATI>04I6CP18B50204U9
D`0304];E@1;BeL0Bd]G0100C4aH0`1=CET504i>FP`0CE5L101;CeX304Y>FA40BDeH3`1<C5P304];
E`<0BTYF1`19BED504Q8E0l0AdMC3018B5@304M7D`<0ATIB0`15AE4304A4D0<0AdMC0`16AU8304E5
D@<0A4A@1013B4`304=8CP<0@dM@0P13Ae801@13@dl0A4A>04A4CP15AT/0ADI:0040oooo000@04U@
FP<0B4mI1@17CUP304Q?F@@0BE1J1@1:DE/K04]BG0<0BU5K0`19D5X304Q?F@<0C4iK00@0BdmK04U?
F`19Ce/0BDmK1@18D5/01015CUL0ADaF04E<EP13Ae0203i1B00403Hg>`0/;2h0;2`^02@R8`807Q`M
0`0T7aT302<N60<08QdG0`0Q71H0100U81d07aXF01lJ5P0P6aD202DN5P0402DN500a:Ah0<BTN03hf
:@80@CPY00@0?c<S02hR4P0^8Q80:Ad=0P0j;Qh01016>RX0?C4Q03da8@0/810202<G1`0402TM4@0[
7QD0:ahE02/N5@80:ahF00@0<2<K03P[9@0h:bD0>bhX0P0h:RL0100c9AX0:adB02/M4P0Z714203@V
6`0403`^8`0k;B80>bdR03HX7@80=BLL0`19?2`00`10<b<0=BPH03DX600303DX600402`O3`0Q50@0
8A@404/n=@80CD0g00@0?c8Y03`_9P0l;bH0<B@K0P0^8AP0100/7aH0?C0W03d`9`0n;R42044a9004
02lO400_7@l0;ad?03PU5P80?R/J00@0BC@S03lZ6@0o:QT0@BXJ0P0l9AD0100i8Q80>b@D03/T5010
:AT203lX600403@M3@0Y4P80:A8203dS5P80@2HG00@0>Ad?048U4`129A<0Dc8O0P17910010117@L0
DRhH058^601X>R020794:P0407I9;01oD380Oe0b08eL?080ODdY00@0MdDR08EC;P25Dbh0VfM00P2N
JD40102QK4@0YFi50:E^A@2ZL4P20:aaA`<0[G98302_LTD30:Y]@0040:U/?`2WJSd0YfXm0:IY?080
YFPk00<0Y6Lj0:=V>@2SIST0102QIcT20:1V>00409iT=P2JH380VV0b09IL;P80TeT[00@0UeT`09QJ
<@2HFS40UEX`0P2=Db/01021BR<0NTDM07Y57@1jAQl207a;8`0407A27`1eA2<0MD@S06Xm7P80GcHJ
00@0GSXR05@d7`1D=1l0A2LE0P17;1/0101:=bP0<QlA038O4@0c8AD203`Z800402XJ3@0Q50<08A@3
02hR3080=2P>00@0?2lL040c8010<b00Bch[0P16>BP0100f:AT0>2/K03P[6`0k;B0202lQ500401XC
0`0N5`L07QL702DN3P80;2DE0`0`:AT0100b:a/0=2dM03@]7@0_9182038Z5`0403P_800m=b/0?CL[
04A0=@80BTM000<0AdI10450?011@3`00`13A4T01014ADX0ATI>04I6CP18B50204U9D`0304];E@1;
BeL0Bd]G0100C4aH0`1=CET504i>FP`0CE5L101;CeX304Y>FA40BDeH3`1<C5P304];E`<0BTYF1`19
BED504Q8E0l0AdMC3018B5@304M7D`<0ATIB0`15AE4304A4D0<0AdMC0`16AU8304E5D@<0A4A@1013
B4`304=8CP<0@dM@0P13Ae801@13@dl0A4A>04A4CP15AT/0ADI:0040oooo000<04Q?F@@0BE1J0`18
CeT504M>F0<0B4mI1019D5X504YAFa`0Be9L0`1:DE/504U@FP@0C4iK0P1;Ce/304U?F`<0B51K00@0
ADeH04E>E`15CUL0A4]C0P12Add0100j?D80<C0e034`=@0V9RP2020N8@<091lI0`0S7QP3028M5`<0
8A`F00@08ahK01lJ5P0O6QH08QdG0P0V7aL0100U7Q@0<b/P03<[8017?c8204m6=`0404@h:00`91@0
<2@D02DI2@80<bLG00@0ACTY04@h:014>2P0=2PH0P0W6`/0100Y7@l0;b<E02lS5@0^8QH202DI3@04
024D2`0W6Q809aXB02XM5@80:1/C00@09AP?02TL4`0Y71<0<bHM0P0i;2<0100b9A`0:1/B02PK4P0Z
7A@203@W7P0403da8@13=bL0@cLW040d9080>blO00@0>RhN03@X600d:1P0;20@0P0Y7@d0100V71<0
;B<J02dS6P0/8QT203L]900403D[8P0b:1l0<RPO02dS6P80?S@[00@0?BlR0512=@1@@SD0@38U0P0i
:AX0100g9AL0=2<C03@S4`10;Ah203PU500404Xg:00n:a`0?R/L02lL3@80:QL800@0<Ql@03dZ6`0m
:Q/0?R/L0P0i9QL0100c7A80?BDI03dU6@0d7@l203TM3`04044T500m7P`0?Ah<04<R3`80D30K00@0
FS8I06Hl901V?2@0J3hU0P1dB2d01026Fc/0MTTX07I9:01dB2D2081B<00409QW?`2JJD40VVU109m[
A080Xfi600@0YVm60:Y_A`2ZKdL0ZVm50P2/KTD<0:a_@P040:Y]@02YK3l0ZF`o0:Q[?P80YVTl00@0
Y6Lj0:9U>02RICP0XF@g0P2PHcH309aR=00409eS=@2NI3H0WV@f09mU=`80W68d00@0V5h`09AJ;02D
FR`0U5H_0P2AERh0102>DRd0Q4dV08A=9P1lAR0207M3700407I47`1hAR40N4HQ07948`80MTLY00@0
I3PK054/4P1A;180FCHP0P1@<Al0100k80l0AR`M04H/7@11;b<2034O5@04038O600b8AT0<R4I028D
2@807a8100@0;28903<Z3@0c:Pd0>BhJ0P1;@2h0101FBcT0@CDU044e9@0W6`/202XN400403<W6@0Y
7@l0:Ad?02TR4P8091d=00@07aP8024J2P0Q6PX09b0@0P0Z8a<0100W81008a`<02<L300/9142038Y
600403Xc8`13?C40@cda04Q4>`80AdI100<0@D0l03Ti=`0i>CL00`14ADX01015AT/0AdM?04M7C`18
B50204Y:E00304];E@1<C5P0C4aH0100C4aH0`1=CET504i>FP`0CE5L101;CeX304Y>F@D0BDeH3018
C5L?04a<F0<0Bd]G0`1:BUH304U9E@<0BTYF0`19BED404Q8E0h0AdMC3019BED304M7D`<0ATIB0`15
AE4304A4D0H0ATIB1P15AE4404=8C0<0@dQ>0`13Ae0204=7DP0504A4D015ADl0ADE?04I7C016Ad/0
0@3oool000`0B4mI1019D5X304Q?F@D0AdiH0`18CeT404U@FPD0BU5K7P1;DU`304YAF`<0BE1J0`1>
CE/0101<CU/0BdmK04]?F`1;Ce/504U?F`0404I@FP18D5/0B51K04U@FP80B4aE00@0@DA903Hg>`0f
=c/0:b/]0P0S8bD302@O6@<08ahH0`0R7AL3024L5P0401lJ5`0N6AD07QTE02@O6@809QlG00@08AX@
02`T6@0/91T0@CT/0P1<@c@0101<@300>BdM03T]7@0Y7@d202dQ4@0403``8013=bL0@cLW03/_7`80
<2@D00@0:Ql=02/P3P0[80h0:Ad=0P0S5`T0100O4`D08AD9024E2@0U6@d202TM4@04034T6`0X6a80
:1/B02PK4P80=BPO00@0>RdT03<V7@0c9Qd0:QdD0P0Y71<304Lk:`0403hb8P0e:AT0=BTI03<W5`80
;20@00@0:al?03HZ6P0f:QX08QP?0P0X7QD0100T6Q40;B<J02dS6P0`9Qd2034W7P0402DK4P0Y7aH0
:AlF03L[7`80EDTm00@0B3`^040b9@10<RD0>b/L0P0b8Q80100j:AT0;Qd=02hM3@0_7@l203TW6@04
03l]7`0i9aT0>BLI03<Q4`80<R0B00@0=B<E03HT5P0f91H09a<:0P0a7A80100b7140=b0B03LP4P0f
6P`2034D100404<T4P1A<R00DC8P058/5@80FS@M00@0FC@J06<k8@1S>b40O5<g0P1[@R80101cB2P0
OE8`07eB<02CISl209QX@00409a[@`2RKDD0XVe50:E^A`80Z6e500<0ZF]40:QZ@`2XJT<0302YK3l0
102ZKD00ZF`o0:U/?`2WJSd20:EX>`040:9U>02PHcH0X6<f09iQ=080WF0c102IGc4209]Q<`0409aR
=02JH380VV0b09EK;@80T5HX00@0S50[08]?:P2;CbX0QddX0P21BB@0101kAB40MT8M07I27@1b@1d2
0710700406/l7P1[?R40JchQ06<i8080DR`E00@0CbhK04T/7019;1`0>R0C0P14;200100c8AL0;a`E
02lL5@0^7AH202hM5@0402TK400Y7P`0:Ah<02hT2`80<2L:00@0?c@R04i3<@1>@c40CT<a0P0d:1P0
100P50H0:1`>02PL3P0`91P202TM4@0402LQ4`0M5`T07AL901XD1P808Q`>00<09b4C028L3P0R70h0
0`0L5PP0100_9QD0<b`J03</6P0l=2L2044m<P0404E2>@15A3l0AD@o0493?P80@D4o00@0ADI;04I7
C016Ad`0AdQ=0P18B5001019BE40BTYD04Y:E01;BeDB04a<F0<0CDeI1@1>CUX<04eAG0@0BdmJ0`1:
CUT504U=F0`0B4aG3`1<C5P304];E`<0BTYF0`19BED404Y:EP<0BDUE4@18B5@<04U9E@<0AdMC0`16
AU8304E5D@<0A4A@1P15AE4604I6DP@0@dQ<0`13B4h304=7D080@dMB00D0ADEA04E5C`15ADl0ATM<
04M8C0010?ooo`003018CeT404U@FP<0B4mI1@17CUP304Q?F@@0BE1J1@1:DE/N04]BG0<0BU5K0`19
D5X304i=F`0404a>F`1;Ce/0BdmK04]?F`D0BDmK00@0AU1J04Q@F`18D5/0BE1J0P18C5D01011A4T0
=SLk03Hg>`0[:bd202<S9@<091lI0`0S7QP3028M5`<08A`F00@07aXG01hI5@0N6AD091lI0P0V7aL0
100Q6Q00;2@I02`T6@11>B`204a3=00404a0<00i;Ad0>BdM02TM3@80;B4A00@0?30P04<g9`13=bL0
>blO0P0`91@0100Z7`d0:b0>02/P3P0Y7@d202<G2@0401lC1@0Q5@T08AD902DI3@80:AdA00@0<B@K
02PK4P0X6a80:1/B0P0e:1l0100j;B@0<bHM03<V7@0Z7A@202TL4`<0Ac/[00@0?S8R03DY6@0e:AT0
<bLG0P0/8100100[7`l0=RXJ03HZ6P0R60l202PN5@0402@J4@0]8aX0;B<J030V7@80<BLN00@09A/B
02TO5P0Y7aH0=b/O0P1EBCd01018?2h0@38U040b9@0k:a`2038R4P0403XY6@0^7@d0;Qd=02lM3`80
>BLI00@0?bdO03TW6@0i9aT0<b4C0P0b8180100e8aD0=R@F03HT5P0W4`X2034M4P04038L4@0g8180
=b0B03HJ3080<A@400@0@b@B054b801A<R00DR`E0P1J=1d0101I=1X0Hc/Q06<k8@1lDcL206]28P04
07=8:01mDS00OE8`09=V?`80V6Q000@0W6]30:9]A@2RKDD0YFi70P2XKDD00`2YJd@0Z6Y30:QZ@`0<
0:U/?`040:Y]@02YK3l0ZF`o0:MZ?@80YFPk00@0XVDh0:1S=P2PHcH0WV4d0P2MH3<409UO<@80Vf4c
00@0W68d09YP<P2JH380UE/]0P2@ERP0102<D2/0RdlZ08]?:P27CBP20859900407]58@1f@Qd0MT8M
07907@80L40L00@0Jc`N06/n8@1[?R40HcTP0P1B;1D0101?;Q/0BB`L04T/700j81<204@/800403<Q
5`0_71D0;a`E02hM5P80;QdE00@0:A/@02TN300Y7P`0;R@;0P0`9`X0100o=280CT<a04i3<@1>@c42
03@X6004020D1P0X70h0:1`>030T6080:AdA00@09b4C01dG2@0M5`T06Q@60P0R70h00`0W8A<08Q`>
028L3P0301`F200402lV5@0c;1X0<b`J03`d9`80@Cdb00@0AD8i04E4?`15A3l0@T<n0P11@Cl01015
AT/0ATM<04I7C017B4d204Q8D00404U9D@1:BU@0BTYD04];EA80C4aH0`1=CET504i>FP`0CE5L101;
CeX304Y>F@D0BDeH3018C5L?04a<F0<0Bd]G0`1:BUH304U9E@@0BTYF0`19BEDA04Q8E0`0BDUE0`17
Ae<304I6DP<0ADEA0`14A50604E5D@H0ATIB1013B4`304=8CP<0@dM@0P13Ae801@15AE40ADE?04E5
C`16Ad`0AdQ<0040oooo000<04M>F0@0BE1J0`18CeT504M>F0<0B4mI1019D5X504YAFaP0Be9L00<0
C5=M04]BG01;DU`00`1;DU`304YAF`<0BE1J0`1>CE/304a>F`<0BdmK0`19Ce/01018DU`0BUAN04YD
GP1<EEh204]BFP0404E8C`0k?440>ca102l^<`809RHX0`0T7aT302<N60<08QdG0`0Q71H0100L5a@0
7QTE01hI5@0U81X202DN5P0401XC2@0Q6@h08AT>03@/7`80?cHW00@0EDTi04@h:014>2P0<BDE0P0Z
7Ph0100c9aL0?S8R03hb8P0m<B4203HZ6P0402lT400X7@T0:1d902DJ20809Q/900@09AT9028F1P0R
5PH0:1`>0P0a9AL0100h:b80:ahE02/N5@0Z7A@203LZ8@04040c:P0i;2<0>B`S02lR6@80;B0G00@0
=RXJ03<W5`0c9aL0<RHF0P12=RH0101FBSX0BCd]04Tm;@0[7`l201hB0P0403`c;00j<BX0>S4Z02@K
508081L@00@09ahG03P_:00h;bP0<BPQ0P0`9b00100V7100BShb04Xn<P0o<bD203la9004048e9@0n
;Ql0?RhO040`8080;Qh>00@0:QX=03HV6@0f9QT0?BdP0P0e9AP0100[6`h0:AT<02TI300/70l202dM
400401h;100Y5@`0:AD<02hJ3`80>R8F00@0<QP;02P<000X3000@B@D0P1A=280101>:Q@0DblI05<_
6@1?:a<205Lb600407=<;`1T?Ah0I3dN07179`80NU4a00@0TfHo09EX@@2EJ440VfY20P2PK4D0102S
K4D0YV]30:I[@`2WJD820:IV@0`0Z6/n00@0Zfi10:U/?`2YK3l0YfXm0P2TIcX0102QI3L0WV4d09iQ
=02LGc8209YM<00409YP<P2IGc40VEla09QN<0<0VEla0P2EFbd0102>E2H0R4hP08Q>8027BbH208I:
9P0408=99020B2D0P4PU07]78P80M48O00@0KShJ06/k5`1[>aL0H3<F0P1L<1D0101S>b80ES8L05Hb
70179Q<204@X6P04040V6@15;B40ABdQ02TF3`80:aXB00@0;1/D02`M5P0/7AH0=2HK0P0f:aT0100a
9`h0:b8502/R1@17?2X204Tn;00403``800V6PX09QX:028F2080;b<E00@0<bLK02/O4`0[7a<05Q02
0P0B3000100O6@/0=S0R03H`8P0i<bD202/U5`0402PR500c;Al0<bdO03T`7`80>c@T00@0?cT[040l
<@10?340@3df0P0o?ST00`11@Cl0A4I304A6@`0304M8C@0404Q9CP19BE40BDUA04Y:DP<0Bd]E4P1<
C5P304e=F@D0CTiJ301=DE`404]?FP<0BTiI1@19CEP<04Q<E`l0C4aH0`1;BeL304Y:EP<0BDUE101:
BUH504U9E@l0B4QD3019BED304M7D`<0ATIB0`15AE4604A4D0<0ADEA0`16AU8304M7D`@0@dQ<0`13
B4h304=7D080@dMB00D0ATIB04I6D016AU00AdQ=04Q9C@010?ooo`003018CeT01017CUP0B4mI04Q?
F@18CeT304U@FP<0B4mI0P17CUP304Q?F@@0BE1J1@1:DE/<04U@FP`0BU5K301;DU`<04U@FP0404M>
F018CeT0B4mI04M>F080A4]E00@0ADaF04=:E013BU@0>3m90P0[<S`0100S8bD07alM01lO7@0L6aH2
01dM4`04028P500S81L08b0G01lJ5P806aHC00@06a/C01`L500L71@07a`C0P0P7A@00`0R7Q<09AlC
02DO4`0302HP5004048g9@14>BL0A3TW03`a7`80<2DC00@0=2TG04<h9P13>2H0ACXX0P0m<R00100d
:180<2<@030S400[7Pd202PK2`0402TK3P0]7a@0;AlD038T6`80=RPO00@0<bHM034T6`0a91/0;b8I
0P0b9A`0100m<2L0@SD/048e;00f:B0202DH3`0402LK3`0c9a/0<bLK04<g:`80?C4U00<09a/?02DI
3@0U6@d00`0`91P01015>2l0@cH]04<f;@0f:B0203T/8`04038U700k;RD0>bhU03/^9@80;b8I00@0
8QD5030S4`0`8a<0;R4A0P0i;1`01011=2@0=RTI03HY6@0`8a<202TL300403LU6`0f91X0=R@J03DS
6@<0=R@J0P0d8QP0100_7A<0:aT?02/I3`0X5`d202PF300402LC2P0X3`P0:0l802/>2080<Q<>00@0
?QdH04HS7`168al0B2dH0P17:aD01014:100CRhE04h^5@1H=Q/205Ld600406@o8P22GSh0PUhn08eK
@080Tf1500@0VFI90:1YB`2PJD/0XVY90P2VJdT0102YKDT0Zfm;0:]_B`2XKTP20:U`A@040:Ua@P2W
Kc`0Yfll0:=/=P80XVXg00@0YFdl0:Q_@P2XKd80ZFa00P2XJcl0102UJ3`0XVDi0:9U>@2OHSH509iQ
=@0409iW;`2JHbd0VV<]09EK:@80TUPX00@0TeHY08mB9P2?DRH0RT/R0P25AAl307]55`04081;7`24
DBH0Q54V07i=9080LD@K00@0JchG06/o6P1[?aX0JShS0P1P=Qd0101A;180B2D?04PU3`19:QD204X_
700403lX5P0a70/0<A`;02/N5P8060/300@0:QdD044g:`11=b/0=2dM0P0U7P`0100Y9100;BPD02dX
5019?RX204Dj9P0402lT4P0N4P807Q8202@H2P80:alC00@091L>01`?1`0L3`L03PP00P0L5PH0100/
9QH0=2hN03@^7P0c;Ad203@^7P0403Xd9011>b/0@C/[04U4?P80@chh00@0>cPa03/h<`0k>3<0?chi
0P15A4001016Ad80ADE304E5@`17B4d204Q9CP0404U:C`1:BU80BTYB04];D`<0BTYD0P19BED0101>
CUX0CDeI04e=F@1=CET204a<F0<0Bd]G0`1:BUH404m?F`D0CTiJ0`1=CET01019CEP0BTiI04Y>F@1;
CeX204a@F`0404eAG01<D5/0C51K04a@F`80BdmJ0`1<D5/0101;CeX0BTiI04Y>F@1:CUT204U=F0l0
B4aG0`1;BeL0101<C5P0CDeI04e=F@1=CET204a<F0<0Bd]G301<C5@00`1<C5P0Bd]G04];E`0304];
E`<0BTYF0`19BED<04Q8E0`0BTYF0`18B5@404M7Da40ATIB00@0@DI<0497C@12Add0@dQ>0`14BDl2
04=8CP030497C@11AT`0@DI<00D0ATI>0@3oool000`0B4mI00@0AdiH04Q?F@18CeT0B4mI0`19D5X3
04Q?F@80AdiH0`18CeT404U@FPD0BU5K3019D5X<04YAF``0Be9L3019D5X01017CUP0B4mI04Q?F@17
CUP204A;E@0404E<EP13BU@0@dYD03PoB@80:c8l00@08b<U01lO7@0O7ad071/F0P0M7A<0100R81@0
8b0G02<P5`0O6QH201/F4`0401/K4`0L71@071`D01lL4`8081dD00<08QhC02DO4`0U7a<00`0V81@0
1012=bD0A3TW04@i9`0l<Al2030U4`0403@Y5`13>2H0@cPV04Dj:080?C8P00@0=2PB030S400`8a00
:ah=0P0X6`/0100Y6`h0;AlD02dO500b91/203HX7`0403<V7@0a91/0<B@K02lR6@80<RDL00@0?C0W
048e;012=B`0=RTP0P0U60l0100W6`l0<bLK03<W6`13=b/203da9@0302LK3`0U6@d09AT=00<0<2@H
00@0ACP_04<f;@13=Rd0=RTP0P0i;2<0100b9A`0>bhU03/^9@0k;RD202lR6@04028E1@0`8a<0<2<C
02hQ4@80>B`L00@0@C@T03HY6@0f:AT0<2<C0P0Y70`0100g9A/0=R@J03HT6P0e8aT303HT6P80=28H
00@0;adC02/I3`0[6@l0:1L=0P0X5P`0100W4`X0:0l802P?200[3PP2038C3P0403hM60168al0AR<O
04P]6080Ab/E00@0A2P@04h^5@1>;QD0F3HK0P1G=1P0101T?b80PUhn089N?P2=Fd0209=PA@0409UV
B@2PJD/0X6U;0:9ZB@80YV]900@0ZFe90:]_B`2[Kd/0Z6i80P2YL4D0102YLD80Yfll0:M_?02SK3H2
0:9Z=`040:E]?02XKd80Z6m20:U/@080Z6/o00@0YFPl0:9U>@2RICT0Wf8f1@2NHCD0102NIbl0VV<]
09YS;@2EFbT2099H:00409=F:@2?DRH0Se8V08Y;8P80QDDO0`1kAAL01020Bal0Q54V08AA9P1nCB@2
07546`0406/n5`1[?aX0JclJ06Xn8`80H3HM00@0DB`B04PU3`189@l0BBXE0P1:;a`0100o:1H0<A`;
034L2`0[7QH201P;0`0402XM5011=b/0@CL[03@]7@809Ah<00@0:B@@02dX500]:1@0BChZ0P15>RH0
100_91807Q8201hB0P0T60X202/O4`0402@G3P0L3`L070l700h8008071H600@0;2HF03@^7P0d;Qh0
<bdM0P0d;Qh0100j=2@0@C/[044k:`19A3h204<n>00403/h<@0k>3<0>cPc03ln>@80ADA000@0ATM2
04E5@`15AD<0AdQ=0P18BDh01019BTl0BTYB04Y:DP1;Be<304Y:E080BDUE00@0CTiJ04e=F@1=CET0
CDeI0P1<C5P304];E`<0BTYF101?Ce/504i>FP<0CDeI00@0BDeH04Y>F@1:CUT0BdmJ0P1<D5/0101=
DE`0C51K04a@F`1<D5/204]?FP<0C51K00@0BdmJ04Y>F@1:CUT0BTiI0P19CEP?04Q<E`<0Bd]G00@0
C4aH04e=F@1=CET0CDeI0P1<C5P304];E``0C4aD00<0C4aH04];E`1;BeL00`1;BeL304Y:EP<0BDUE
3018B5@<04Y:EP<0B4QD1017Ae<A04I6DP040456C012Add0@TM=04=8CP<0A4U?0P13B4h00`12Add0
@DI<0456C00504I6CP40oooo000=04M>F0<0B4mI0`19D5X304Q?F@80AdiH1018CeT304U@FPD0BU5K
3019D5X<04YAF``0Be9L3@19D5X204YAF`0304Q?F@16CEL0ATeG00<0ATeG00@0?DA>038i@`0b>D<0
:RX/0P0U9BD0100O81X07QhD01hN500R7aH202<P5`03024L600N6AH07QTF00<071`D00@07a`C020M
500P7A@08QhC0P0U7Q@0100U7a<09R0D02HP5012=RH204@h:00403hc8@0_9180;b@B02/P3P80=BXH
00@0@CHR04Dj9P15>RH0>RhH0P0i;1T0100e:1L0;b8B02lR4P0Y6`h202DG300402HH3`0X6Q40:1XA
02<F3@80<2<J00<0>2/R03DX7`0e:1l00`0_8QT0100d9ah0=RTP03HY800_8aL2020D200401D9000I
3@406@d102/O4`80@CDY00@0ASX^03/_8`0k;b<0A3L^0P0k;RD0100j;B@0=bXQ03LZ8@0d9ah203<V
7@0401D8000Y71<0:A`C03LZ6P80>RdM00@09aX:02HI2@0V6@T0;1l?0P0Y70`0100`8a<0<RDE038U
5@0i9ad203@R6003034O5@0c8AL0<b4G00<0>BLM00@0<AlE02TG3@0Y5`d08A<80P0S5@X0100W5@/0
:AD<02TE300]50d2038G400403PK5@0m7QT0?AhI048W4P80B2dH00@0Ab/E04HX4016:100CBdD0P1E
<AL0101Q?R80M4lb07A?<P26FC`208]L@00409=RA02KJ4T0VfQ90:=^CP80YW1>00@0Z71=0:M_BP2W
KdX0ZFe80P2YKT@0102YKd40YV`j0:I/>P2SJSD20:9X=P040:AZ?02VJcl0YV/o0:QY?@80YVLk00@0
Xf@h0:1Q=@2PHCD0WEhb0P2KG30309YK;`0409aU;`2MIC80WFDb09]S<080UE/]00<0S54U08E:8025
BR000`25Ab00101`>a40LSdE078m5@1hA1`207]9900407M78`1]?ad0KClM06Xm7080KD8P00@0I3lU
05`f7`1L=Ql0CR/E0P149A401014:AH0A2dM04@]7@0g91D202TH200402LJ4P0H2`<060/302LM4`80
>c4U00@0:b@D01lH1P0O60H0:RDB0P0d;a/01013?2X0=RlM03H_7@0Q6PX201XB1@04024I3P0O5Pd0
7aH=01P?1P806Q4:0`0Q6`d0100P6P`07AL901dG2@0M5`T202@N4004034[7@0l=RP0?3HX03lj=080
>3<]00@0;b`U02l/9`0_;2L0=3<_0P0m?ST01014A480ATI404I6A017B4d204Q9CP0404U:C`1:BU80
BTYB04];D`<0BTYD0P19BED00`1=CET0C4aH04a<F00404a<F0D0Bd]G101?Ce/504i>FP<0CDeI00@0
BDeH04Y>F@1:CUT0BdmJ0P1<D5/0101=DE`0C51K04a@F`1<D5/204]?FP<0C51K00@0BdmJ04Y>F@1:
CUT0BTiI0P19CEP?04Q<E`<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];E``0C4aD1P1;BeD3
04Y:E0<0BDUC3018B58<04Y:E0<0B4QB1017Ae4A04I6D0<0@TM=00@0@dQ>04A9C`14BDl0A4U?0P13
B4h30497C@D0ATI>0@3oool000`0ATeG00@0AdiH04Q?F@18CeT0B4mI0`19D5X304Q?F@<0AdiH1@18
CeT304U@FP<0BU5K3019D5X<04YAF``0Be9L3019D5X0101;DU`0C5=M04aCG@1:DE/204M>F00404Q?
F@19D5X0BE1J04=:E080>T5;00@0=3Di02h^;P0^;Rh092DP0P0P81P0100P81H08b0I02<P6@0R7aX2
020L6@0401`K5P0L71@071`D01lL5@8081dD00@08QhE02@M4`0T7A<09AhD0P0U7a<0100a:1T0=bhO
03L^7`0l<b8203Xa800403@/6@0`:1D0<2PE03<[5P80=blJ00@0=2PB03HY5P0f:AH0=RTH0P0c9QH0
100/7Q40:1X?02PJ3`0V60l202LI400401X=100[7QD0:ahE03P[8P80=bXQ00@0;b8I02hQ600^8AP0
=BPO0P0k;RD0100a9AT0;20D02`P500S5`/202@H300403HZ7P18?300B3``0514>080DdLk00@0?S4X
04Hi<016>C00;B0G0P10<bX0100_8QT07Q4801hA200d9ah203DX7`0402@G1`0]8100;B0@02<F1P80
:A`<00@0<B@D02/N3P0[7Ph0;1l?0P0Z7@d0100a81H0;QdC02hM4`0]718202lN50<0<R4G00@0;A`B
02PG3@0X5`d07aD90P0P5PX0100S5`/091H;02@F2`0X5P`202`H3`04034J4P0e7QH0=AhF03XP3`80
AbhJ00@0ABXE03TM1`0i7@L0@2090P1B<1L0101Q?B<0JTL[06Y7:`1nEcX208EL?P0408mSAP2HJdX0
V6]:09i`CP80X71<00@0WF]609aX@`2LJ4<0ZVY60P2[JdD0102ZJcl0YfTj0:MY>P2TISD20:9T=@04
0:=T>02TICX0Y6Dj0:AT=P80Xf<e00@0XF4c09mO<@2OGc40WEd_0P2LG2h309]K;@0409MO;02GGbh0
Uel^09EM;P80SE@W00@0PdXO07a26P1l@QX0NT0J0P1k@A`0101f@Al0Ld4P07=1801c@R8207118`04
06Tl7`1P=1T0H3@I05/a6080Fc<I00@0Fc`X05@e8@1D=B40Ab`I0P0n91D0100k9AL0>2HH03PV600`
81<202@F2@0402DH400I3`H06@l602LM4`80<RXO00@08QX=01XD100J50@0:bDE0P0f<b80100g<bL0
920D02@P500C3`@201LC200401hJ4@0B3PD04Ph500/600805Q4;00<07AH>01TB2P0I4PX00`0F3`L0
100K50`091dE02@M5@0/9Ad2034Z8P0402TV7`0U8Q/09B8K02<P6`809RDP00@0<S4]03lo?@0o?cd0
BDU70P1<CT/01017B4d0B4U>04Q9CP19BTl204Y:DP0404];D`1:BU@0BTYD04Y:E080BDUE101;BeL5
04a<F0<0CDeI101?Ce/504i>FP<0CDeI00@0BDeH04Y>F@1:CUT0BdmJ0P1<D5/0101=DE`0C51K04a@
F`1<D5/204]?FP<0C51K00@0BdmJ04Y>F@1:CUT0BTiI0P19CEP?04Q<E`<0Bd]G00@0C4aH04e=F@1=
CET0CDeI0P1<C5P304];E``0Bd]C101;BeD304Y:E0D0BDUC3018B58<04Y:E0<0B4QB1017Ae4A04I6
D00304=6C@14Adh0A4M>00H0ADQ?00<0A4M>04=6C@13ATd01@16ATh10?ooo`003016CEL01017CUP0
B4mI04Q?F@18CeT304U@FP<0B4mI0`17CUP504Q?F@<0BE1J0`1:DE/<04U@FP`0BU5K301;DU`<04U@
FP0404]BG01<Ded0C5=M04YAF`80AdiH00@0B4mI04U@FP19D5X0@dYD0P0j@D/0100d=CT0;Rh^02h^
;P0T9B02020P6004020P5P0S81T08b0I028O6P8081`I00@071/F01`L500L71@07a`E0P0P7A@0100R
7QD091dC02@M4`0U7Q@202DO4`04034X6@0g;Ql0=bhO03`c8P80>S4P00@0=2`I030X5@0`:1D0<b/F
0P0g;aX0100d:180=RTF03HY5P0f:AP203<V5P0402`N4@0X6Pl0:1X?02HH3`809aT@00@06Pd402/N
5@0[7QD0>2/R0P0g:R40100_8QT0;R4H02hQ600e:1l203/^9@04034U6@0/81@0;20D02<G2`8091P<
00@0=RXN04Pl<018?300D4@h0P1CAc/0100n<BP0AST`04Hi<00]81L2040c:P0402lR6@0N4@P07Q48
03@W7P80=BPO00@091L702dP400]81008aH60P0Y70`0100a91@0:ah>02/N3P0/7`l202XM3@04034P
5P0^7A<0;QdC02dL4P80;ahD0`0b8AL0100]7180:1L=02PG3@0O5@T2020F2P0402<G2`0T5P/091H;
02PF3080;1P?00@0<AXB03DN5P0e7QH0>R0?0P17;QX01015:QD0>Ad703TM1`1080T2058`5`04064m
8`1ZAb/0JTL[07iG>P80QE`n00@0Sf=609Q[BP2HJdX0WW1>0P2PL4`0102MJdH0W6Q309aX@`2ZJTH2
0:][A@040:Y[?`2WJCX0YfTj0:AV=@80XV@e00@0Xf@h0:AU>P2TICX0Y6@f0P2SHcD0102QHC<0Wela
09mO<@2MGBl209aL;P<0Ve/]00@0Uel/09MO;P2GGbh0UEd^0P2=E2L01023BQl0O48J07a26P1j@1X2
07]1700407I17`1c@B00Ld4P07=28P80L44S00@0JC`O060d6@1P=1T0Fc4H0P1K<aT0101K?2P0E3DQ
05@e8@17;1T203hT5@0403/U5`0h9QP0>2HH030P4`8091H900@09AP@01T?1P0I3`H09adC0P0b:Ql0
100R6Pd06Q@401XD100[9AD203Hc8P0403Lc9`0T81@0920D01<?10805a<800@07QXA018>1@0B3PD0
2`H00P0F4@/00`0M5Ph06A8:01TB2P0301H?1`0401/D300T7AD091dE02`U7@80<BXR00@0:BHO02DR
6`0U8Q/08b0K0P0V9B00100b<Bd0?clm03lo?@19BDL204a>B`0404M8C@18BDh0B4U>04U:C`80BTYB
00@0Bd]C04Y:E01:BU@0BTYD0P19BED404];E`D0C4aH0`1=CET404m?F`D0CTiJ0`1=CET01019CEP0
BTiI04Y>F@1;CeX204a@F`0404eAG01<D5/0C51K04a@F`80BdmJ0`1<D5/0101;CeX0BTiI04Y>F@1:
CUT204U=F0l0B4aG0`1;BeL0101<C5P0CDeI04e=F@1=CET204a<F0<0Bd]G301;Be<404];E@<0BTYD
1@19BE<<04Q8DP`0BTYD0`18B58404M7DA40ATI@00<0@dI=04A7CP14Adh01P15B4l00`14Adh0@dI=
04=6C@0504I6CP40oooo000<04I=E`0404M>F018CeT0B4mI04Q?F@<0BE1J0`18CeT504M>F0@0B4mI
4@19D5X<04YAF``0Be9L3019D5X0101;DU`0C5=M04aCG@1:DE/204I=E`0404Q?F@1;DU`0Be9L04M>
F080@DQB00@0?Ci203Df>00e=SP0:B/V0P0R8a/301lO5`03028N6`0Q7A`08AdL00<07A`G00@07a`E
020M5P0P7AH08QhE0P0T7AD302DN500402DK4@0X7Q80:1hB03@Z7P80@CPY00@0@STX03L_700g;a`0
;2@?0P0Y8@`0100c9a40<B@A034T4@0`8a8202lR4P04030R5@0^81D0;R0E02/M5080:1XA00@08A@;
02@G3P0T5`h0:QdD0P0d9ah0100l;bH0?S4X03ha:00g:R42030S6P0403L[7`0o<bL0?c<W03@X7080
7a<700@05PX001L;000G2`00:1`@0P13=b/0101gJel0@SHZ048f:P0o<bL202hR5P0401lC1`0l<2@0
?30T02DI3@8091P<00@09QX:02lS4`0_8a<09QX:0P0Z7Ph0100_8a<09QX:02HJ2P0S5`L201lC0`04
02<B200Y60h0:AP>02hM4`80;A`B00@09aH<02@C2@0T4`T09aH<0P0/6a40100U7a408a/>02<K3P0O
5`X201dC1`0401lC1`0T5P/091H;02TK4080;ahD00@0<a`<03hW5`0n9aL0?bDD0P0d6@H0100f6PD0
C2`G04`/5`1M?BH206I4:`0406]8:P1gE3H0Me@f08QTA080UFi=00@0VFi<09M[B02GJdP0UFM30P2C
ISl0102YIdD0Z6M10:QW@@2WIS`20:EU=`040:9R<P2PH380X60b09mN=080Wehf102MGBd809aL;004
09AL;@2>EBP0SUDX08I@8P80PT/R00@0PDXS07e78@1mAb40N48N0P1c?AT0101[?200ISLM06Hg7@1R
=1`2068e7P04068h8P1P>RD0H3XU05hh9@80G3LU00@0C3DW04Tb9019<R@0@R`N0P0f91H0100^7Q40
:Q`A02XL4@0Y7A4202PN500402LM4`0P5P`081H<02HN4`80:b<H00@07QP<01/E1`0K5@L09b@E0P0`
;Ah0100T91`05aL?01LG3`0=3@D201DE3@0401`K5P0=30L03@`700@300803`h900@07QTC01/F400K
5Q006A@>0P0L5a40100S7QP09B0J02DP6P0Q71H201`G4@0401lL5`0P7AP081dH028Q7080:bXV00@0
=cLe0453@011@d00B4Y90P1:C4/01017B4d0B4U>04Q9CP19BTl204Y:DP0404];D`1:BU@0BTYD04Y:
E0<0BDUE0`1:BUH204];E`0304a<F01=CET0CDeI00<0CTiJ101?Ce/504i>FP<0CDeI00@0BDeH04Y>
F@1:CUT0BdmJ0P1<D5/0101=DE`0C51K04a@F`1<D5/204]?FP<0C51K00@0BdmJ04Y>F@1:CUT0BTiI
0P19CEP?04Q<E`<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];E`l0Bd]C0`1:BU8304U9D@l0
B4Q@301:BU8304Q8D0@0AdM?4@16ATh01013ATd0A4M>04A7CP15B4l304I9D080ADQ?00<0A4M>04=6
C@13ATd01@16ATh10?ooo`003016CEL01017CUP0B4mI04Q?F@18CeT304U@FP<0B4mI1P17CUP504Q?
FA/0BE1J301;DU`<04U@FP0404Q?F@1:DE/0BU5K04U@FP80ADaF00@0ATeG04YAF`1:DE/0BE1J0P14
BeD01012ADX0>Cdn03Tm?P0_<Bh202DX8@04020Q6@0O7QT07ahI01lN70807adN00@07A`H01hM6@0N
7AT081dH0P0P7AH00`0Q71H091dE02@M5@0302@M5@0402DL4`0S6Q408aXA02TQ5P80=BhN00@0>C8P
03</6@0c;1T0<2XD0P0b;1@01011=Al0=bXG03LZ5`0]80l202/N3P0402lQ500_8AH0;b4F02PJ4@80
818900@09AP?020C2P0P4`X081<:0P0Z7A@0100g:R40>RdT03X]900c9Qd202XM500401hB1P0`91P0
<2@H03/_8`80?c<W00@0@CDY03<W6`0c9a/09a/?0P0[7a<0100k;b40GE5305eA@`0/818203X^8004
04Pl;P0S5`T08aL901lC1@80=RXL00@0;R8B03HZ6P0f:QX0;20@0P0_8a<0100d:1P0:al?02/O3`0Z
7Ph202HJ2P04020B1`0W6@h09aT>02`N4`80:Q`A00@091H;024C200Q4`P09AL<0P0[7A80100W8A<0
91h@02@N400R6Pd201lG2P0401lE2@0P5PX081H:028H30808aT=00@0;AX<03HP4`0f81<0?2DG0P0m
8a@0100l8A00@bLB04<W4P1@<Ad205Lg8P0405Hd601ZB2`0JTP/089O@@80SVY:00@0SVQ408aQ?P2<
HCh0SV@n0P2CId80102THd<0Xf<o0:=S?`2QHST20:1Q=@0409iP<@2LGC40W5da09YK<080VEXa00@0
UeP/09IG:`2FEb/0UEHZ0P2DEBT0102CE2P0TE8V095B9P2@DBD208m@900408aC:025Cb<0QDlS07m8
7`80ODPP00@0O4PS07Y89@1jB2D0MdDT0P1d@R40101X>b@0GS@N05hd7P1F:aP205@^6`0405/f901P
?2`0H3`/05/j:`80EC<W00@0>bTM03h/800n;200?BdP0P0b91T0100T60`07A<701dC1`0V7182034Y
7P0402dU6P0W7a@09alD02LO50809AlC00@07aP>01dI3P0M6@h081hB0P0O7A40100C50l04A8=014B
3@0>3`T201@E3`0401XK5@0@4@/0414;00H70@802``600@08b0K01`I500L6A@05Q<>0P0H5A00100O
71L08b0K02<P6`0P7AP201`I5004028Q700T8al092<O02LV8P80;Rh/00@0=SPe03i0?`0n@3l0@DE4
0P13AdH01017B4d0B4U>04Q9CP19BTl204Y:DP0404];D`1:BU@0BTYD04Y:E0D0BDUE00@0BTYF04];
E`1;BeL0C4aH0`1>CUX604m?F`D0CTiJ0`1=CET01019CEP0BTiI04Y>F@1;CeX204a@F`0404eAG01<
D5/0C51K04a@F`80BdmJ0`1<D5/0101;CeX0BTiI04Y>F@1:CUT204U=F0l0B4aG0`1;BeL0101<C5P0
CDeI04e=F@1=CET204a<F0<0Bd]G401:BU8304U9DA40B4Q@3019BE4304Q8D0@0AdM?4P16ATh304M7
C`<0B4Q@0`17Adl704I6CP40oooo000<04I=E`0404M>F018CeT0B4mI04Q?F@<0BE1J0`18CeT604M>
F0D0B4mI6`19D5X<04]BG0`0BE1J00@0B4mI04YAF`1:DE/0BE1J0P15C5H01016CEL0BU5K04YAF`19
D5X204A;E@040495BP0i?Ch0>Cdn02la;P809BPQ00@0824I01lN6@0O7QT07ahL0P0O7Ah0100M71P0
7QdI01hM6@0P7AP2020M5P03024L5P0T7AD091dE00<091dE00@09A`C02<J4@0S6Q40:B4F0P0e;Qh0
100i<R00<b`I03</6@0`:Q@2038/5004044e7`0g:QL0=bXG02dP3`80:ah>00@0;b4D02lQ5P0_8AH0
:1XA0P0P4PT0100U60l081<:020C2P0P4`X202XM500403LZ8@0j;B@0>RdT03<V7@80:QdD00@07Q86
030T600`91P0>blS0P0o<bL01011=BT0<bLK03<W6`0W6`l202/O4`0403/_8@1MDD<0GE5302`P4P80
>RhP00@0B3`^02<G2@0S5`T07a<50P0f:Q`0100^8Q80=RXJ03HZ6P0/810202lS4`0403@X600[7`l0
:al?02XN3P809QX:00@0818702LI3P0W6@h0;1hC0P0Z7140100T5P/08A<8024C200U5``202/M4P04
02LQ4`0T7Q0091h@028J3@807aL:00@07aD9020F2P0P5PX08QP<0P0S6@d0100]6P`0=R0C03HP4`0l
9AL203dS500403`Q40139a80@bLB050a7@80EcLR00@0ES@H06Y8;01ZB2`0PUm10P2>JTX0102>J4@0
S64n08aQ?P2>I3h209=W@P040:AS@`2SHcl0Xf<o0:5R>@80X64e00@0WV0a09aM<@2LGC40VU/`0P2I
FS40102GF2`0UUL[09IG:`2EERX209AE:@0409=D:02ADRH0TE8V091A9@80Se0T00@0S5<X08E?8`25
Cb<0OdPO0P1mB200101lB2<0NTPU07Y89@1gAB@207A28@0406Pk901N=1h0GS@N05H[6080E2hK00@0
FcHT060l;01P?2`0FcX[0P1E<bL0100k:Ad0?R`P03h/800m;B02038T6@0402@H300M4`L07A<702HL
4P80<BTN00@0;BDJ02LO500W7a@09alD0P0U7a<0100O60h07AT>01dI3P0P7Q8201lM4@0401<D3`0A
4Pd04A8=00h?2@8051D?00@06Q/E010A2`0@4@/01PL10P0;30H0100S81/071TD01`I500F4`h201PE
400401lL5`0S81/08b0K020M608071TD00@08R4L02@S7`0T8al09bHR0P0^;R`0100f>3D0?T0o03i0
?`11AD@204=7AP0404M8C@18BDh0B4U>04U:C`80BTYB00@0Bd]C04Y:E01:BU@0BTYD1@19BED0101:
BUH0Bd]G04];E`1<C5P304i>FPH0CdmK1@1>CUX304e=F@0404U=F01:CUT0BTiI04]?FP80C51K00@0
CE5L04a@F`1<D5/0C51K0P1;CeX304a@F`0404]?FP1:CUT0BTiI04Y>F@80BDeH3`18C5L304];E`04
04a<F01=CET0CDeI04e=F@80C4aH0`1;BeL@04Y:DP<0BDUA4@18B50<04U9D@<0B4Q@1017AdlB04I6
CP<0AdM?0`18B50304M7C`L0ATI>0@3oool000`0ATeG00@0AdiH04Q?F@18CeT0B4mI0`19D5X304Q?
F@80AdiH0`16CEL304M>F0@0B4mI6P19D5X<04]BG0`0BE1J00@0ATeG04U@FP19D5X0BE1J0P16CEL0
1017CUP0BU5K04YAF`19D5X204E<EP0404I9CP0o@dH0?d=603Dj=P80;2hY00@09BPQ024R7@0Q8Qd0
7alM0P0O7Ah301hM6@04020L6@0P7AP081dH024L5P808a`F00@08a`D02@M5@0T7AD08QTD0P0S6Q<0
100X7aH0<2PK030X6`0a:QP202lX5@0403@^5P0m=ad0?CLM04<g8@80=bXG00@0:ah=02PK2`0X6`/0
;AlB0P0_8AH0100Y6a808Q@;028D2`0R5@`2024D2`03028E300T5`h091L>00<08aH=00@0:1/B02lR
6@0_8QT0=BTM0P0a9AT0100[7a<0;R8F02hR5P0d:1`2038V6P0402hR5P0_8aL0;b<G03<W6@802`00
00@0DdLi07M[G@1gJed0>blQ0P0_8aD0100Y7@l0<2@F030T5P0c:QT203T`7`0402`S4P0^9A@0;RDD
03D/6`80<BPG00@0=B`K03@[6P0d:aX0:1`@0`0W6`l202PL40<0:AdA0`0X7100100R60`091X>02@J
3P0V710302TO4`8091`?00@07QH901TA100I4@@0:AL=0P0]6A00100m9a`0AblS04L_8`12:1T203dR
4@04044T4P149aD0A2LE05Tg7080JTP/00@0O5Tk089N?P22GSh0OUDe0P1kD2h0101oDc00QUXe08IJ
=@2MHCl209]O>`0409YO=@2IGc40VEla09UO<@80Ued_00@0U5T_099G;`2BEbl0SUDZ0P2=E2T0102;
DRL0QdhS08M>8`23BQl207m66`0407]25`1i@1D0ND0E07Lm5`80MSlH00@0MClI070l5`1`?1L0JCLD
0P1V=AD0101W>AP0JSdN06Xm7P1aB3@206I0:`0405Pc801?:a/0Cb/K04`[7080Bb/N00@0@R@I03PL
400h7100<1lE0P0e91X0100g:Ah0<R@I038T6@0T60`201hB1P0402DK3`0c:Ad0<bTM034Y7080;bLL
00@0:28F020I3`0P6@l06QH;0P0K60l0100I5Pd04@h7014>1`0:2@D2014@3004010?2P0A40/04A0;
01PH408051@:00@030`200d=0`0=3@<06aPA0P0H5@h0100D4@X05Q<<01HC300L6A83020M5P807Q/D
00@081lK024R7@0Q8Qd09bLU0P0^<2d0100h>ST0@4A30414@`17Bd`204]?D00404M8C@18BDh0B4U>
04U:C`80BTYB00@0Bd]C04Y:E01:BU@0BTYD1@19BED0101:BUH0Bd]G04];E`1<C5P204e=F@0304i>
FP1?Ce/0CdmK00@0CdmK1@1>CUX304e=F@0404U=F01:CUT0BTiI04]?FP80C51K00@0CE5L04a@F`1<
D5/0C51K0P1;CeX304a@F`0404]?FP1:CUT0BTiI04Y>F@80BDeH3`18C5L304];E`0404a<F01=CET0
CDeI04e=F@80C4aH0`1;BeL<04Y:DP<0BT]@0`19BTl404Q9CP80AdQ=3018BDh<04U:C`<0B4U>1017
B4dA04I7C00404M6CP18Adl0B4M?04U8D0<0BTUA0P19B5000`18Adl0AdI>04M6CP0504I6CP40oooo
000=04M>F0<0B4mI0`19D5X304Q?F@80AdiH1016CEL304M>F0D0B4mI3019D5X<04Q?F@`0Be9L3019
D5X01015C5H0BU5K04YAF`1<Ded304U@FP80Be9L00@0BU5K04I=E`16CEL0B4eC0P14BD`0100m@D00
=3Tc03@i<`0]<2T202LX8`03028R8P0P8280820R00<07QdK00@081`I020M600P7AP08A`H1@0S71H0
100O5a@08AXD024J500W81P202lY7@0403<]7@0d;a/0=2lK03@`5`80=C4F00@0>2`F038U4P0b9A80
:ah=0P0Z7@d0100/7Q40;b4F02lQ5P0^81L202`N5@0302@G3P0U60l09AP?00<0:1/B00@08A@;01`?
1P0L3`H081<:0P0Y71<0100S5`/07A4501dA1@0Q5@T202TM4@<0:alC00@0;20D02/O4`0[7a<081@4
0P0g:a/0100I3@00@CDU044e9@1/H50204Xn;P0302DI2@0c9aL0<bLG00<0@3LV00@0:1l>028I200R
6@P09ah=0P0V7@`0100a:1L0=RdL03H]700X710202HJ3P0402@H300U6@d09AT=02LK3`80:1`@00@0
9QX>02@H300T60`07a<70P0R5PX0100X7100;R8F02hR5P0^91P202XP5004028H300K4@D06a4502@C
2`809Q<<00@0=B4H040Z7`10:Ql0>B4E0P0h7Pl0101091H0A2LG04@W5`1W@bT206a8;P04075<<@1a
BRd0LDX]06m49080Kd8Q00@0MTPV07a>;01lCR`0UUhm0P2DG3L0102CG3<0Ted_09=M;`2DGRl209=M
;`04091J;P2=ERd0SEH]08IB:`80QE4Z00@0PdlX081<9@20C2D0O4PQ0P1g@a`0101c?aP0L3`E070l
5@1`=Q4206lg4P0406`f4P1S<@h0Hc4>05`[2`80ERT:00@0Eb/>05/_4P1K;a80G3PR0P1K>2@0101C
<Ql0ABHD04DV500k7Ph203PL3P0403HL3`0c60d0<aP=02`K4@80;QhA00@0<B<F03<U600c9AP0;AlB
0P0U60P0100U6@T0:al?02/O3`0^:1X2038/7P0402XT600H50T061@9014=10804a0900@03`l700T8
0`0920<02`P30P0@3@P0100?30D03``300l<0`0E4`L201@B1@04014?0P0F50D05Q@501`L4P807AdC
00@07QhD01dM4`0M7A<06QX@0P0F5P`0100C4`T04106010@1P0K71L202<S8@04030`;P0m?ch0?Cln
04E9B080Bdm@00@0C59B04eCD`1=De<0AdQ=0P18BDh01019BTl0BTYB04Y:DP1;Be<304Y:E080BDUE
0`1:BUH304];E`0404a<F01=CET0CDeI04e=F@80CTiJ101?Ce/504i>FP<0CDeI00@0BDeH04Y>F@1:
CUT0BdmJ0P1<D5/0101=DE`0C51K04a@F`1<D5/204]?FP<0C51K00@0BdmJ04Y>F@1:CUT0BTiI0P19
CEP?04Q<E`<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];E``0BDUA1019BTl504Q9CP<0AdQ=
6`18BDh404M8CA40ATM<0`18Adl01019B500BTUA04Y9D@1:BE4204U8D0<0B4M?1@16ATh10?ooo`00
3@17CUP304Q?F@<0BE1J0`18CeT204M>F0@0ATeG0`17CUP504Q?F@`0BE1J3018CeT<04]BG0`0BE1J
00@0ADaF04YAF`1:DE/0C5=M0`19D5X204]BG00404YAF`16CEL0ATeG04Q=D`80A4U<00@0?D5003@i
<`0d>C<0;C0Y0P0W:2<00`0R8R80820R020P8P0301hM6`04020L6@0P7AP081dH024L60D08a`F00@0
7aLD024J500Q6Q@09b0H0P0_:Ad0100c;Ad0=2lK03@_6`0d<1L203Da5P0403P/5P0b9A80<RDB02/N
3@80:Qd=00@0;1hA02lQ5P0_8AH0;R0G0P0/7QD00`0T5`h09AP?02DH3`0302PK4P04024D2`0L3`H0
70l6020C2P80:A`C00@08aL;01dA1@0M4@D08AD90P0Y7A4302/O4`0402`P500[7a<0:alC020D1080
=b/K00@06@d0044e9@11=BD0K61@0P1:?Rh00`0U6@T0<bLG03<W5`03040g9P0402PO3P0R6@P08QT8
02LN3@809Qd<00@0<BPG03H]700f;A`0:1`@0P0V6Ph0100T60`09AT=02DI3@0W6`l202PL400402HJ
3P0T60`091P<01lC1`808QH:00@0:1`@02hR5P0^8QH0;R@H0P0Z81@0100R60`06a4501/A1@0T4`/2
02HC300403DQ6010:Ql0@2XO03TQ5@80>1h?00@0@2@F04@W5`149aL0Id<Y0P1/B2h0101aC340LDX]
075:;@1_A2@206m28@0407I89P1lCR`0O4h/09IN?@80U5`g00@0Te`c09=M;`2CGBl0U5h_0P2CGBl0
102@FRh0SEH]08eF;@26DR/208EA:P0408=?:020C2D0P4`U07a88@80Md<L00@0LclH070l5@1`?1D0
L3HA0P1_=a80101/=Q80Hc4>06<a3P1L:`/205HY2P0405L[3P1K;a80FblB05`h8P80FcPT00@0Dc8O
04DV50159Q@0>ah>0P0h70h0100f70l0<aP=03<H3@0/6a4202hN4@04034S5P0c9AP0<bDH02dO4P80
9AP800@09AT902/O3`0[7`l0;RPJ0P0b;1h0100Z91P061@901PD2@0A3@@201<@2@0400l?1`0920<0
2@P300/80`8040d800@03``500l<0`0?30<05A<70P0D4PD0100A3`805Q@501HD1@0L718201dM4`04
01hN500M7A<07AdC01XJ40805QH<00@04a<9010@1P0@40H06a`G0P0S8b40100`<2h0?Cln03do?P15
BDP204]?D00404aBDP1=De<0CE=C04M8C@80B4U>00@0BDY?04Y:DP1:BU80Bd]C0`1:BU@204U9E@<0
BTYF0`1;BeL0101<C5P0CDeI04e=F@1=CET204i>FP@0CdmK1@1>CUX304e=F@0404U=F01:CUT0BTiI
04]?FP80C51K00@0CE5L04a@F`1<D5/0C51K0P1;CeX304a@F`0404]?FP1:CUT0BTiI04Y>F@80BDeH
3`18C5L304];E`0404a<F01=CET0CDeI04e=F@80C4aH0`1;BeL<04U9D@@0BDY?1@18BDh304M8CA/0
B4U>1017B4dA04I7C0<0B4M?00@0BDQ@04Y9D@1:BE40BTUA0P19B50304Q7C`D0ATI>0@3oool000`0
B4mI00@0AdiH04Q?F@18CeT0B4mI0`19D5X304Q?F@80AdiH1016CEL504M>F0<0B4mI3019D5X<04Q?
F@`0Be9L3019D5X01016CEL0C5=M04aCG@1>EEl204aCG@0404]BG01=E5h0CEAN04]BG080AdiH00@0
Be1F04Q=D018CE00@dM60P0k@3X0100c=C00:R`W02X/9`0U9BD2024Q8`<07ahL00@08AdJ020L6@0P
71T08A`H0P0S6aP302<L5P0402LO7@0S6aP08a/H028K4`80:B<G00@0=S0P03`g8`0l=b<0=S8I0P0]
:@h0100b9Q00<RDB038U4P0a91<202hQ4@0402/M400Z7140:Q`A02`N5@80;b4H00@0;1lF02PK4P0X
6a80:QdD0`0a91/202PK4P0401lB2@0K3PD06`h502@H30805PX000@060`002<G2`0S5`/09AT=0`0V
6Ph201hB1P0402PL300M4@407A4102@H2080:Qh>00@08aL703lc8`0o<b<0A3PX0P0i;Ad0100Q6PP0
:b@B02/T4P0S70X202TR4004030Y5`0W80h09b0>02HO3@808Q/900@081@8028F2P0R5PX08aL;0P0R
5PX0100N4PH07A4501dA1@0O4`L2028F2P04028A1`0T4`T091<902HH3@80;1hC00@0;b<G02hR5P0^
8QH09adA0P0R60`0100S4P/08A09024@2@0[614202lK4P0402`D200g7A00=ad@04T]7`80CS8T00@0
IT4W06<n901S?R@0HCTO0P1Q>1`0101X>ah0LT@S07948`1mC2`208AC<P04095L>P2@G3D0T5`e08mL
<@80T5d^00@0TEl^095N;`2AGRl0SEX_0P2:ERh4081>:`80OddZ00@0OD/X07Y89@1jB2D0MdDR0P1f
A240101oA280MShK07Hn6`1]=aD206Dc4P04060b4@1M<1<0GC0C05H]3`80DbX>00@0@Ql904XZ5@1:
:QD0CB`I0P0m80h0100`5@@0;aD602lE1P0d7@l203TR500402`L3`0Z6Pd0:QX=02hN4@<0=bLH0P0/
7`h0100S5PD081<2020C0P0Z91H203<]7`0402TU6@0C3`@04`l400P500802`P100@02`X500T81009
20@03PT50P0?2P@0100;1`0030P000`8000B3P82014>0004018?000N6`X07Q/:01dM4@80820D00@0
8B4E01lO4`0O7a<06QX>0P0G5`/0100H60`06AT=01TI3@0R8R02030`;P0404=5@P1CEE@0DeED05MK
FP80DUIG00@0Ada?03m4A`0oA4L0AdQ=0P18BDh01019BTl0BTYB04Y:DP1;Be<304Y:E080BDUE00<0
BTYF04];E`1;BeL00`1;BeL304a<F0<0CDeI101?Ce/504i>FP<0CDeI00@0BDeH04Y>F@1:CUT0BdmJ
0P1<D5/0101=DE`0C51K04a@F`1<D5/204]?FP<0C51K00@0BdmJ04Y>F@1:CUT0BTiI0P19CEP?04Q<
E`<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];E``0BDUA1019BTl304Q9CPD0AdQ=6`18BDh4
04M8CA40ATM<0`18Adl01019B500BTUA04Y9D@1:BE4204U8D0<0B4M?1@16ATh10?ooo`003019D5X<
04Q?F@<0ADaF00@0ATeG04M>F017CUP0AdiH0`18CeT>04U@FP<0B4mI1019D5X504YAF`l0Be9L0`1:
DE/00`19D5X0B4mI04Q?F@0304Q?F@`0Be9L00@0CE9H04aAE`1<DEL0C4mD0P1;C5401013A4P0>3Ll
03Pg?00Z:R`2020P8P04020O7@0O6aP07a/H01`H5@807APD00@07QTE024J500Q6Q@08AXD0P0O6180
100U6A/09ATI02DI6@0U6QP202LL5P0402/Q5`0_9AT0;bDI03@[7080=bhM00@0=blL034Y5P0a:AH0
<2LF0P0Z8A80100M4`L0718801`B200R60l2024G3P0402DH4P0]81X0;B0J02`O6@808aH@00@09APB
030S7@0`8ad0<2<M0P0V6A<00`0I30<07Q4801hA2004024D2`8091L>00@0:1/B02/N5@0[7QD08aH=
0P0W6Q40100S5Pd08QD<028E300Z7A@202dP5`0402lR6@0g:R40=bXQ03DX7`80;B0G00@08A@;02XM
500Z7A@0:ahE0P0V6A00100_8QT0;B0G02dP5`0N50P201dC1`0301`B1P0K4@D06a4500<06a450`0L
4PH0100N4@P081<:020C2P0S5Pd202HI400402LJ4@0V6A009QT@02@G3P808aH=00@06a4801T>200I
3PP0714=0P0U5aH0100Z71T0:a`E02/L5@0^81D203DU5P0405@j:@1;<1d0Bc0M04h_6`80DBhH00@0
DB`A06<j701S>Q`0NTd/0P20DS001026FC80PUD^089E;P1nDBX207mB:`<0PUD^00@0OE0Y07M:8`1g
BR<0Nd`b0P1hBS00101fB2h0Md/b07M;<P1hC3<207A:<P0406m5;`1Y?bT0JClY06Y0:080JT0Z00@0
IC`V05lf8P1O=R80G3HQ0P1M>2D0101I=240DbhK05<^6`1;:Ad204DU6P0403dO5@0b6Q00<QX@02/G
3P80:1L?00@09QTA02DK4P0U6a808AP90P0U70d00`0Z8A80:b8C02/R4`0302TP4@0402`S500_9QL0
;bHG02DO4`80;RPL00@0:bDI01PB1P0H4PH03PP00P0B3000100E3`<04@/0014;000;1@0201<=0@04
01H@100@304040`100h;0P805Q<:0`0J6Q80100S7a@0920E02@P5@0U8QT202HS6P0402HV7P0Z:R80
:RXR030a:`80=CH`00@0@dE404U;BP19BdX0DE=B0P1DEEL0101ADU@0C4eA04a=D@19BTl204M8C@<0
BDUA0`1:BU8304];D`<0C4aD0`19BED0101:BUH0C4aH04a<F01>CUX2051@G0<0DE5M301>CUX304Y>
F@@0BdmJ2@1<D5/504]?FP<0BTiI0`16BUD304M;EP<0B4aG0`19CEP304];E`0404a<F01=CET0CDeI
04e=F@80C4aH0`1;BeL<04U:C``0AdQ=6@18BDh204M8C@0304I7C015AT/0ADI;00<0A4E:0`15AT/<
04I7C0`0AdM?1@17B4d10?ooo`003019D5X<04Q?F@<0ADaF00@0ATeG04M>F017CUP0AdiH0`18CeT>
04U@FP<0B4mI1019D5X504YAF`l0Be9L0`1:DE/00`19D5X0B4mI04Q?F@0304Q?F@`0Be9L00@0CE9H
04aAE`1<DEL0C4mD0P1;C5401013A4P0>3Ll03Pg?00Z:R`2020P8P04020O7@0O6aP07a/H01`H5@80
7APD00@07QTE024J500Q6Q@08AXD0P0O6180100U6A/09ATI02DI6@0U6QP202LL5P0402/Q5`0_9AT0
;bDI03@[7080=bhM00@0=blL034Y5P0a:AH0<2LF0P0Z8A80100M4`L0718801`B200R60l2024G3P04
02DH4P0]81X0;B0J02`O6@808aH@00@09APB030S7@0`8ad0<2<M0P0V6A<00`0I30<07Q4801hA2004
024D2`8091L>00@0:1/B02/N5@0[7QD08aH=0P0W6Q40100S5Pd08QD<028E300Z7A@202dP5`0402lR
6@0g:R40=bXQ03DX7`80;B0G00@08A@;02XM500Z7A@0:ahE0P0V6A00100_8QT0;B0G02dP5`0N50P2
01dC1`0301`B1P0K4@D06a4500<06a450`0L4PH0100N4@P081<:020C2P0S5Pd202HI400402LJ4@0V
6A009QT@02@G3P808aH=00@06a4801T>200I3PP0714=0P0U5aH0100Z71T0:a`E02/L5@0^81D203DU
5P0405@j:@1;<1d0Bc0M04h_6`80DBhH00@0DB`A06<j701S>Q`0NTd/0P20DS001026FC80PUD^089E
;P1nDBX207mB:`<0PUD^00@0OE0Y07M:8`1gBR<0Nd`b0P1hBS00101fB2h0Md/b07M;<P1hC3<207A:
<P0406m5;`1Y?bT0JClY06Y0:080JT0Z00@0IC`V05lf8P1O=R80G3HQ0P1M>2D0101I=240DbhK05<^
6`1;:Ad204DU6P0403dO5@0b6Q00<QX@02/G3P80:1L?00@09QTA02DK4P0U6a808AP90P0U70d00`0Z
8A80:b8C02/R4`0302TP4@0402`S500_9QL0;bHG02DO4`80;RPL00@0:bDI01PB1P0H4PH03PP00P0B
3000100E3`<04@/0014;000;1@0201<=0@0401H@100@304040`100h;0P805Q<:0`0J6Q80100S7a@0
920E02@P5@0U8QT202HS6P0402HV7P0Z:R80:RXR030a:`80=CH`00@0@dE404U;BP19BdX0DE=B0P1D
EEL0101ADU@0C4eA04a=D@19BTl204M8C@<0BDUA0`1:BU8304];D`<0C4aD0`19BED0101:BUH0C4aH
04a<F01>CUX2051@G0<0DE5M301>CUX304Y>F@@0BdmJ2@1<D5/504]?FP<0BTiI0`16BUD304M;EP<0
B4aG0`19CEP304];E`0404a<F01=CET0CDeI04e=F@80C4aH0`1;BeL<04U:C``0AdQ=6@18BDh204M8
C@0304I7C015AT/0ADI;00<0A4E:0`15AT/<04I7C0`0AdM?1@17B4d10?ooo`003019D5X<04Q?F@<0
ATeG0`17CUP304Q?F@<0BE1J301:DE/304Q?F@@0BE1J1@1:DE/=04]BG0D0BU5K00<0BE1J04Q?F@18
CeT00`18CeT<04]BG00404eBF01<DEL0C55G04e@E`80BdiC00@0B4U>03dn@P0m?T80<S8d0P0Y:B/0
100T8b4081lM020O7@0O6aP201hJ5`04020K5`0S71H08a`F02<L5P808Q/E00@08aTJ028H5`0R61L0
8QPF0P0T6a@0100V7A@0:b<F02/S5P0_:1P2034Z6P0403L_700c:QT0<bXI03D/6`80;bHG00<08AL;
01hD2P0N50X00`0T6Q40100K3PP09APB02DH4P0Z7AL202HI4`0402DH4P0Y71H0:A`F02/N60809aXD
00@06`h501P;0P0H2`805`X10P0K3PD0100R5@`0:1/B02PK4P0/7aH202lR6@0403@W7P0k;RD0>bhU
03P[8P80<2<J00@0;B0G02PK4P0X6a80:A`C0P0d9ah0100`8aX0<B@K034T6`0V6A0302TL4`8081<:
00@08aH=020C2P0P4`X09Q`@0P0T6Ph0100P5PX07A<701dC1`0L4PH201/A1@0401`B1P0M4`L07A<7
01D8008060/200@06`h501hA200N4@P081<:0P0Q50/3020C2P0401`C2P0I40T06A0901`A3@8091HE
00@09aTH02LH4`0W61<0:A/@0P0^81<01017<200@BLF044W5P14:1<204PX4@0404XV301J<aH0FS<F
07149`80MTTZ00@0NDdX07I:9@1fBRD0MDTT0P1hC2L0101lD2/0OU8]07iB;@1lD2/207U=:00407M9
;`1eBC00MDT`07A8;`80LTP`00@0L4H^06a3;@1/@bd0IchX0P1S>R@0101P=b40HSTS068i8`1O>B@2
05`f8@0405`g901N>BL0GSTW05Tf8`80DRlL00@0CS0U04L[7`17:al0>b0E0P0]5``0100R4@L070l7
01`?1`0K4@P201/B2`04024I300R6Pd08QX=02DM40809alB00@0:B4D02LO4P0W7a8091`?0P0Q6@`0
100P6Ph09b4E02LQ5@0X8QH2024K3`0401dG2`0N60`07QP<01XD20804`d100@03PP0018<000B3000
61470P0I5@X0100G50/061D<01PE300M7AD202@T700402dY7P0a;B@0<BdT03Db:@80>SL`00@0?3`d
0452?011@S`0AdQ20P1<CDP01017BDP0Bde<04]=C01?D582051AD`0404e>DP1;C500Bda@04Y;D080
BdaA0`19BE4304Y:DP<0Bd]C0`1<C5@304Y:EP0404];E`1<C5P0C4aH04i>FP80CdmK0`1@D5`<04i>
FP<0BTiI101;CeX904a@F`D0BdmJ0`1:CUT304I:E@<0Ad]F0`18C5L304U=F0<0Bd]G00@0C4aH04e=
F@1=CET0CDeI0P1<C5P304];E``0BDY?3017B4dI04Q9CP80AdQ=00<0ATM<04E6B`15AT/00`14ADX3
04E6B``0ATM<3017Adl504M8C@40oooo000<04U@FQ80B4mI1P19D5X<04YAF`<0B4mI1019D5X504YA
F``0Be9L101:DE/504U@FP<0B4mI301;DU`304a@F@0404aAE`1=D5L0CE1G04Y=DP80ADI:00@0>S/o
03@c>00d<cP0:BPV0P0U9280100Q81h081`I020L6@0R7AT202<N6P<09AhH0`0Q5aH0100Q5aD08APC
024H4`0S6Q4202HN4`0302TQ500[8aH0:b<F00<0=2/L00@0>S4R03H/800f;2009Q`B0P0R60h0100W
7A@09Q`C02HL4`0D2@<201dB300402HK5@0W71H09a`F028G4@807Q<=00@07a@>024F400Q5Q009A/B
0P0M4`X0100I3`H06a4801/A200L4PT201/A200401hD2`0T6Q4091XA02LJ4@80>2/R00@0@3<Z03d`
9`0m<2L0=bXQ0`0^8AP203T/8`0402PK4P0f:B00=RTP02LJ4@808QD<00@09QT@020C2P0P4`X08aH=
0P0V6A00100Y7a<09A/?02DK3`0P5PX201X@100301H<000C2@004`T000<04`T000@04PP001D;0P0E
2`8060h50P0L4PT00`0O5@`08AL>024G3P03024G3P0401hE300K4P/06a8;01/A3`808ADE00@091HE
024D3P0Q50h08aD<0P0W6@`0100k91H0>24A03PQ4@0m8Q42040T3`0404@T3@1B;Q@0DRhD064l8@80
JD8U101/A24206i68`0307189@1bBRL0LTXW00<0Ld/X00@0KdD]0716;P1`ARh0KdL^0P1/@bd0101T
?RT0GcTT05li901J=B8205Td8@0405Lc7@1J=B80FSDR05Xe8P80ES<P00@0EC@Q05Dd8`1E=2<0Cc0N
0P179aP0100m9aX0=b4D03LQ500[6@d2024@1P0401X=100H3`H060l601XC2`806aH@00@07QH;01dE
2P0M5@X07QH;0P0T7140100]9AX0<RXO038Z7`0b:Ql202lW700401H@100U7a<09AlC034[7`80;BLK
00@07AL;01<=0@0C3@4061860P0Q6`l0100M5`/07AH<01dF300Q6Q0202<O5P0401`I400F4``05Q<<
01XJ4P809bHQ00@08QlF02TV7`0Y9Ql0<c0Y0P0i>3<01010?cX0A4E004A5@018BD@204]=B00404a>
C@1=CU00CDi@04a=C`80Bda@00@0BT]?04Y;D01:Be00C4eB0P1>Ce@304U9D@<0BTYB0`1;Be<304a<
E00404];E`1<C5P0C4aH04a<F0<0CDeI0`1>CUX204m?F``0CTiJ0`1:CUT404]?FPT0C51K1@1;CeX3
04Y>F@H0Ad]F1P18C5L304];E`0404a<F01=CET0CDeI04e=F@80C4aH0`1;BeL<04U:C``0AdQ=6@18
BDh204M8C@0304I7C015AT/0ADI;00<0A4E:0`15AT/<04I7C0`0AdM?1@17B4d10?ooo`003019D5XB
04Q?F@H0BE1J301:DE/304Q?F@@0BE1J1@1:DE/<04]BG0@0BU5K1@19D5X304Q?F@`0Be9L0`1<D5T0
101<DEL0CE1G04e@E`1:CE8204E6BP0403Xk?`0d<cP0=3<h02TX9P809B@R00@08B0N020L6@0P71T0
8QdI0P0S7QX302DN60<08ALF00@08ALE024H4`0Q61<08aXA0P0V7Q<00`0Y8A@0:b<F02/S5P0303@[
700403Xa8P0f;200=R`P02HL4P808QP>00@09adD02HL4`0V71<050T30P0M4P`0100V6aD09a`F02LL
5P0R5a4201hC3@0401lD3P0Q5Q008AH@02DK4P807A<:00@06@l601/A200K4@P071890P0K4@P0100N
50/091XA02@J4@0W6Q4203P[8P04040c:P0m<2L0?C0W03LZ8@<0;R4H0P0i;2<0100X6a80=RTP03HY
800W6Q42028E300402HI400P4`X081<:02<F3@809QT@00@0:AlC02DK3`0U6`l081H:0P0J40@00`0F
30004`T001<9000301<900040188000E2`805@/201P>1@80718900<07aD<024G3P0Q5`h00`0Q5`h0
100N5@`06a8;01/B2`0K4@l2024E5@0402@F5@0Q50h08A@>02<E30809aT<00@0>b@F03PQ4@0h8A40
?B8A0P1090l0101490d0DRhD058^501Q?24206U29@@0K4@Q0P1^AR<00`1`B2D0LTXW079:9`0307=;
:00406m5;@1`ARh0L4H^06m7;P80K4<]00@0I3hY05li901O>B@0FSDR0P1I=240101G<ad0FSDR05Xe
8P1J=B8205Hc800405Dd8@1E=2<0EC@S04l`7P80AbLH00@0?BLJ03LQ500g8A@0:aT=0P0Q40H0100J
3@@060l601P?1P0J4`/201/F400401hF2`0M5@X07AD:01hF2`8091`A00@0;BDJ038Z7`0b:Ql0<RXO
0P0_9a`0100F40@09AlC02DO4`0a:al202dW6`0401dG2`0C3@404`d101PB1P808A/?00@07AL;01dF
300M5P`08AX@0P0S7aH0100L6A005Q<<01HC300J6Q8202LV8@04028O5P0Y9Ql0:BHO03<`:@80>CPc
00@0@3lj04A5@014AD00B4U40P1;CDP0101<CTd0CDi@04e>D01<CDl204]<D00404Y;C`1:Be00BT]@
04a=DP80CTmD0`19BE4304Y:DP<0Bd]C0`1<C5@0101;BeL0C4aH04a<F01<C5P304e=F@<0CTiJ0P1?
Ce/<04i>FP<0BTiI101;CeX904a@F`D0BdmJ0`1:CUT604M;EPH0B4aG0`1;BeL0101<C5P0CDeI04e=
F@1=CET204a<F0<0Bd]G3019BTl<04M8CAT0B4U>0P17B4d00`16Ad`0ADI;04E6B`0304A5BP<0ADI;
3016Ad`<04M7C`D0AdQ=0@3oool001P0B4mI2P19D5X204YAF``0Be9L0`18CeT404U@FPD0BU5K301;
DU`304YAF`T0BE1J301;DU`304YAF@0404a@F@1=DUP0CE9H04a?EP80AdY?00@0@D9603Xk?`0j>cl0
;Rh^0P0Y:BT0100S8R007ahL01lN700Q7AX302<N6P808QdI00@08ATG024I5P0Q6AH081PE0P0P6A<0
100Q6Q808Q/A028K4@0T7A<202DO4`0402lU6@0b:1`0<RPL03/a9@80>BlU00@0:AlE02@J400T6Q00
9adC0P0V718301/@2P04020E3`0V6aD09Q/E028G4@806Pl900@060d701dB300M4P`09adD0`0T6Q42
028H3`0401P>1@0=0`003@<0018800807Q@;00@06@`302XM500Z7A@0=2LN0P0f:B00100g:R40<bHM
03<V7@0e:1l2040c:P0402hQ6010<bX0@3<Z02`O5P807Q4800@08aH=020C2P0P4`X09aXA0P0_8QT0
100b:1h0;bDK02lU6`0[8AL202HL4P04024G3@0M4`T07A<901X@1P806@l500@050X101H<0`0F30<0
6@l60P0L4PT00`0N50/081H=020F3@03020F3@0401`E3@0H4@/0614;01T?3@<07Q8B0P0M3``0100M
40L08AD9024E2@0b7a4203@N400403PQ4@0l8Q40?28A044T4P80C2dH00<0EcLP05ll9P1O?2H01P1W
@RD306E08`<0IT4T00@0IClX06M3;@1W@bd0J4@^0P1R?b/0101I=R80D2lL050_701>;AX204l^6`04
05<b7`1D<b00E3<P050a7`80BR/I00@0ABPH04<W6@139aT0?B4C0P0e6@/0100W5`X08Q@7028D1`0L
40@201L=0`0401D>100F4@/05Q4;01TF4@8071TD00@06A8801L@1P0G40H05a060P0L5@/0100U7Q@0
;bPN02lX7P0e;R@203L`9P0403La9@0Y8aL0:B<G020J3P809b4E00@0;RPL02XT600Z91P081X>0P0I
4`L302DN500402HO5`0Q7A@08AdD01TF3`805a@=00@07ahI02d/9`0];2L0:2DN0P0`;BP0100i>3<0
@d8n04=2?P17B4<204U;B00304]=BP1<CTd0C4i=00<0CDi@00@0Bda@04Y;C`1:Bdl0BT]@0P1;C540
0`1<C5@0CTiF04i>EP0304U9D@<0BTYB0`1;Be<304a<E1P0CDeI0`1:CUT404]?FPT0C51K1@1;CeX3
04Y>F@H0Ad]F1P18C5L304];E`0404a<F01=CET0CDeI04e=F@80C4aH0`1;BeL<04U:CbD0B4U>0P17
B4d00`16Ad`0ADI;04E6B`0304A5BP<0ADI;3016Ad`<04M7C`D0AdQ=0@3oool001P0B4mI00<0BU5K
04U@FP19D5X02@19D5X<04YAF`<0B4mI1019D5X504YAF``0Be9L2@19D5X304YAF``0Be9L00@0BU=L
04YAF`1:DE/0BU5I0P1<D5T0101;D5H0B4]B04Q;DP11A4T203`oA00403<c<`0];Bd0;Bd]02@T9080
7ahL00@07A`J01lK600O6aP07a/H0P0P6aL302<K6003028J5`0R6aD08Q/E00<08Q/E0`0S71@0100X
7QD0:b4H02/Q600f;2<203L]900402XP5P0U6a409A/A02LM4`8091X@00@08aXC01XA2P0J4@X060l8
0P0O5Pl0100R6A807A@=01dD3@0K4P/201hE3P<06a8;00@07aH?024H4@0Q61406a8;0P0C2P<0100F
3@H07aH?01lF3`0S5Pd502XM50<0<2<J00@0;b8I03DX7`0e:1l0=BPO0P18>c80100c9Qd08aH=02<F
3@0S5Pd201/>1@0401lB2@0V6A009QT@02DK4@809Q`B0`0W7A<0100V71808aT?02<I3`0Q5`d201lE
2`<04`X300<05@`501H=1P0F3@H01P0G3PL0100I4PX05@d:01D=2P0F3P`201/A4P0401/A400I3PX0
6@h:01/>1P807a<700@0:AL;02dK3`0]6`l0<1d?0P0d7Q00100j8a<0@RPG048X5`1:;ah2050e8P04
05hn9@1O?bH0GclV05lo9P80GShU00@0G3`S05Xj8@1J>R40FCTP0P1H>1l0101H>2<0Fc/V05/k9P1L
>bP205Df900404T/6P139QH0@bHF044V5@80A2TH00@0BRlN04T^7@19;Qd0@RHH0P0h7Pl0100e6`h0
<a`>03<L3P0_5`/202PB1@04020D200M4`L07A<701TA1P805Pl700@04Pd7014<200A30P040`90P0>
3@T00`0D3@D05Pl701H?1`0301L@200401H?1`0G40P05a0801/D30807aP@00@0>C<W02hX700^:1`0
:R@H0P0a:al0100e;b<0;bTM02lY7@0W8AD202DO4`0401hG3`0R6a<08Q/C020I4@805A0:00@04a09
01lL5`0O71L0;bhZ0P0k>SH0100g=S80?Sdi03hm>@17AT8204e=B`<0CU1?00<0Bdm@04Y>C`1:CTl0
0`1<CE40101;C500C4eB04a=DP1<CE8304a<E080Bd]E0`19BE4304Y:DP<0Bd]C0`1<C5@304i>FPH0
CDeI0`1<C5P<04e=F@<0BTiI101;CeX904a@F`D0BdmJ0`1:CUT604Q<E`H0Ad]F0`1;BeL0101<C5P0
CDeI04e=F@1=CET204a<F0<0Bd]G3019BTlU04Q9CP80AdQ=00<0ATM<04E6B`15AT/00`14ADX304E6
B``0ATM<3017Adl504M8C@40oooo000H04Q?F@0304YAF`19D5X0BE1J00T0BE1J301:DE/304Q?F@@0
BE1J1@1:DE/<04]BG0T0BE1J0`1:DE/<04]BG00404YCG01:DE/0BU5K04YAF@80C51I00@0Be1F04Q;
DP18Be80@DA90P0l?d@0100c<c<0;Bd]02d];@0T92@201lN700401dL6P0O6aP07a/H01lK608081/G
0`0S6aP00`0R6QL08Q/E028K5@03028K5@<08a`D00@0:1hE02/Q600[8AP0=R`S0P0g;B@0100Z81H0
9A/A02DK4@0W7A<202@J400402<J4`0J4@X06Q4:01P?20807aH?00@08QTB01dD3@0M50d06a8;0P0N
5@h301/B2`0401lF3`0Q61408APA01/B2`804`X300@05Pd601lF3`0O5Pl08aH=1@0Z7A@3030S6P04
02lR6@0e:1l0=BPO03DX7`80B3/b00@0<bHM02<F3@0S5Pd08aH=0P0K3PD0100O4PT09QT@02HI400U
6a4202HL4P<09adC00@09Q`B02<I3`0S6@l08AL=0P0O5@/301<:0`0301D<1@0F3@H05Pd600H05`h7
00@06A8:01D=2P0E3@X05Ph<0P0K4A80100K4A006@h:01T>2P0K3PH201lC1`0402TG2`0]6`l0;A/?
030M3`80=1h@00@0>R<C048X5`12:1L0BRlN0P1@=B80101N?RD0GclV05lo9P1O?bH205hn9@0405`l
8`1J>R40FSXQ05Ti8080F3PO00@0F3PS05/k9P1K>bH0G3/X0P1E=R@01019;1X0@bHF04<V5P119QD2
04@Y600404X_7P19;Qd0BBhM048V6080>1h?00@0=A/>03<L3P0c70h0;aL;0P0X4PD0100P50P07A<7
01dC1`0I4@H201H?1`04018=1`0A30P04@`8010<2@803Pd900<050d501H?1`0F3`L00`0G40P0100F
3`L05a0801L@200K50`201lH400403Tc9`0^:1`0;RPL02XT6080<B/O00@0=BlS02lY7@0_:Ad09b4E
0P0U7a<0100N5`l08Q/C028K4`0P6A4201D@2P0401<@2@0O71L07a`G02l^:P80>cXf00@0=cHb03hm
>@0n?CT0AdI20P1=CD/304i@C`0304]?D01:CTl0BTi?00<0C4eA00@0Bda@04a=DP1<CE80C4eB0`1<
C5@204];E@<0BDUA0`1:BU8304];D`<0C4aD0`1>CUX604e=F@<0C4aH301=CET304Y>F@@0BdmJ2@1<
D5/504]?FP<0BTiI1P18C5L604M;EP<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];E``0BDY?
9@18BDh204M8C@0304I7C015AT/0ADI;00<0A4E:0`15AT/<04I7C0`0AdM?1@17B4d10?ooo`003017
CUP<04Q?F@H0BE1J1P18CeT<04U@FP<0B4mI1019D5X504YAF``0Be9L0`18CeT404U@FPD0BU5K301;
DU`0101:DUd0BE9K04UBF`19DU/204YAF@0404a@F@19CU@0BDiD04E8C`80@4=800@0>S/m038c=@0b
<cD0:2PX0P0P82000`0M71X071/I01`K6@0301dI5P04028M6@0Q71P08A`H024L60P081/G00@08aPD
02DJ500U6Q@0;b@N0P0c:B00100Z81L09adC02LM4`0X7Q@202<I3@0402DL5@0J4@X06Q4:01<:0`80
6A0900@081L@024H4@0Q61408APA0P0R6A80100J4@X060l801P?200H3`P201`C3004020G400Q6140
8APA01dD3@806A0900@081<:02HI400V6A00:1/B0P0]81L0100g:R40<RDL038U700U60l2024D2`04
02PK4P0j;B@0>RdT030S6P80;1lF00@0:QdD01hA200N4@P07A070P0N4@P0100@1P004PP00188000F
30<201X@1`0301dC2P0O5@`07aD<00<07aD<1`0J4@X301T@2@8060l800@05A0:018=2@0B3@T050/<
0P0H3a00100J41405``:01L<2P0H3PD201/A1`0401h=1@0T4`T091<902DD2P809aD900@0;QX?038O
4@0b7a40=R<E0P10:Qd0101=<A`0CS8M04hb7@1@=1l2058f8@0405<g8P1A=B00DCDP04hb7@80BblJ
00@0AbXH04X]6`1:;A/0BRlN0P15:QT0100l8Q<0=Ql?03HO3`0f7a4203TR500403/T5P0j8aD0>R<E
038L3P80:Q@700@09a<802XF2`0Z5P/0:1H<0P0S4@L0100J40H06A0701T@1`0I40T201L@2P0401D=
2P0A30T04@`900l92@803PP800@03PT301@?2@0D3`T06A@>0P0H4`d00`0B3@L030L100`70@0300/6
000400l9000Q6`l08A/?03H`9080>38V00@09R0D01LA1@0G4@D08A/?0P0e;b<0100Q6Q808a`D02<L
500K50h200d80P0400h;1P0Q7AX08AdJ03@c;`80?Sdi00@0@3lm04A4@P14A480BTY:0P1<CTd0101;
C4h0Ad]<04M;C015BD`204A8B`<0Bda@00@0BdaA04e>D`1=CU<0CTiF0P1>CUP00`1;BeD0BDUC04U9
D`0304U9D@<0BTYB0`1;Be<304a<E0`0CDeI301<C5P304Y>F@@0BdmJ2@1<D5/504]?FP<0BTiI1P18
C5L604M;EP<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];EaP0BDY?6@18BDh204M8C@0304I7
C015AT/0ADI;00<0A4E:0`15AT/<04I7C0`0AdM?1@17B4d10?ooo`003017CUP<04Q?F@<0BE1J0`18
CeT304M>F0<0ATeG3`18CeT404U@FPD0BU5K301;DU`404Q?F@80BE1J101:DE/>04]BG00404UCG@19
DE`0BE5L04UBF`80Be9L00@0C5=K04eAFP1=DEX0BDiD0P16Be401012@dD0>CXl03Tj?00/;Bl202<S
8`0401dM7@0L6aT071/I01/J60807ATF0`0P6aD0100P6aL07aXF01lJ5P0O6QH501hI5P04024F500P
5A4081DA02XO6`80<2DO00@0:b4H02/Q5`0[8AL0:b4E0P0T6Ph00`0T7AD07QL?01hG3`0301PA2@04
01XC2`0L5@d071D=01dF3P807QL?00@07QLA01dF400M5Q006A8<0P0F3`T0100L5@l091dG02@M5`0N
5a42018;1@040144000J3@@06Pd4024D2`80:ahE00@0=bXQ038U700b9A`08aH=0P0M40L0100P4`X0
:ahE02/N5@0[7QD303<V7@809AP?00<09QT@02@G3P0T5`h00`0H3PD0100I3`H06Q0701X@1`0M4`X2
020F3@0402<I400U6a809A/B024J50<081TC0P0O618601hG4@0401HA2`0C3P/04`h;01<=3@8060lB
00@05`h?01D;2@0E2`T05@X40P0I3`D00`0G20407Q0701h@1`0301`>1@0402DD300W5Ph09aH>02XI
3`80<b8H00@0>b4@03`R4@0l8Q40?bDD0P14:QT01018;Qd0AR`K04H/6`109QD203XP3`0403HL2`0h
7Pl0>1h?03LP4080=1d?00@0;aT;02`F2@0/5PT0:aP:0P0]6P`302lL3P0402/G300T4PH0918602@B
20809aH<00@09QD=01lA200O4@P05P`30`0F2`D201H<2P0401H<2`0F30d05P`=01L<40805Pd@00<0
3`X6014<200A30P00`0C3PX0100A30P03`X600l:1P0>2@D200d8100400`6000?2@003`T001H@1080
7aT=0`0U7a<0100S7A408Q`@028L400Y8Q`2028K5@0401L@2P0>2@D03PT5014>2@808AdJ00@0<C0^
03dl>P0m?3X0Bd];0P1>CTh0101ADE40D55C051AD`1<D5<204Y>D@<0BDiB00@0C4eB04]<D@1;C540
BT]@0P1<C5@0101>CUH0CTiH04i>F01<C5H204U9E@<0BDUA0`1:BU8304];D`<0C4aD301=CET<04a<
F0<0BTiI101;CeX904a@F`D0BdmJ0`1:CUT304U=F0<0B4aG0`17BeH304I:E@<0Bd]G00@0C4aH04e=
F@1=CET0CDeI0P1<C5P304];EaP0BDY?6@18BDh204M8C@0304I7C015AT/0ADI;00<0A4E:0`15AT/<
04I7C0`0AdM?1@17B4d10?ooo`003017CUP<04Q?F@<0BE1J0`18CeT304M>F0<0ATeG3`18CeT404U@
FPD0BU5K301;DU`404Q?F@80BE1J101:DE/>04]BG00404UCG@19DE`0BE5L04UBF`80Be9L00@0C5=K
04eAFP1=DEX0BDiD0P16Be401012@dD0>CXl03Tj?00/;Bl202<S8`0401dM7@0L6aT071/I01/J6080
7ATF0`0P6aD0100P6aL07aXF01lJ5P0O6QH501hI5P04024F500P5A4081DA02XO6`80<2DO00@0:b4H
02/Q5`0[8AL0:b4E0P0T6Ph00`0T7AD07QL?01hG3`0301PA2@0401XC2`0L5@d071D=01dF3P807QL?
00@07QLA01dF400M5Q006A8<0P0F3`T0100L5@l091dG02@M5`0N5a42018;1@040144000J3@@06Pd4
024D2`80:ahE00@0=bXQ038U700b9A`08aH=0P0M40L0100P4`X0:ahE02/N5@0[7QD303<V7@809AP?
00<09QT@02@G3P0T5`h00`0H3PD0100I3`H06Q0701X@1`0M4`X2020F3@0402<I400U6a809A/B024J
50<081TC0P0O618601hG4@0401HA2`0C3P/04`h;01<=3@8060lB00@05`h?01D;2@0E2`T05@X40P0I
3`D00`0G20407Q0701h@1`0301`>1@0402DD300W5Ph09aH>02XI3`80<b8H00@0>b4@03`R4@0l8Q40
?bDD0P14:QT01018;Qd0AR`K04H/6`109QD203XP3`0403HL2`0h7Pl0>1h?03LP4080=1d?00@0;aT;
02`F2@0/5PT0:aP:0P0]6P`302lL3P0402/G300T4PH0918602@B20809aH<00@09QD=01lA200O4@P0
5P`30`0F2`D201H<2P0401H<2`0F30d05P`=01L<40805Pd@00<03`X6014<200A30P00`0C3PX0100A
30P03`X600l:1P0>2@D200d8100400`6000?2@003`T001H@10807aT=0`0U7a<0100S7A408Q`@028L
400Y8Q`2028K5@0401L@2P0>2@D03PT5014>2@808AdJ00@0<C0^03dl>P0m?3X0Bd];0P1>CTh0101A
DE40D55C051AD`1<D5<204Y>D@<0BDiB00@0C4eB04]<D@1;C540BT]@0P1<C5@0101>CUH0CTiH04i>
F01<C5H204U9E@<0BDUA0`1:BU8304];D`<0C4aD301=CET<04a<F0<0BTiI101;CeX904a@F`D0BdmJ
0`1:CUT304U=F0<0B4aG0`17BeH304I:E@<0Bd]G00@0C4aH04e=F@1=CET0CDeI0P1<C5P304];EaP0
BDY?6@18BDh204M8C@0304I7C015AT/0ADI;00<0A4E:0`15AT/<04I7C0`0AdM?1@17B4d10?ooo`00
3017CUP<04Q?F@0404U@FP18CeT0B4mI04Q?F@<0AdiH0P16CEL304E<EP`0AdiH0`18CeT404U@FPD0
BU5K301;DU`404Q?F@80BE1J0`1:DE/?04]BG00404UCG@18DU`0B59L04UAG080BeAM00@0CUEM055E
GP1AEEh0CU=I0P1<DEL01015BD`0?41303a0@`0`<C<202DU9@0301lO7`0M71X07A`J00<071/I0`0N
6A<301hI5@0301hI5P0M61D07APE00<071HF00@08ADE01lD4P0O51809a`H0P0^8ad0100]8aX0;R@J
02hT6P0^91P202HL400402HO5`0U7QH09AhF024J4P806Q<;00@05Pl701D>1P0E3PH05@h60P0F3`L0
100I4P`081TC020I4`0N5a4201@=1`0401H?2@0R6aD08Q/E028K5@80614;00@04PD001L:0@0G2P40
5PT00P0J3@@0100U60l09aXA02LJ4@0S5Pd202LJ4@0402/N5@0]81L0;B0G02`O5P80>B`S00@0=BPO
02LJ4@0W6Q40;1lF0P0Y71<0100P5@l07A8<01dB300H3@L201D:100401@90`0F2`D05P/501X?2@80
7A8<0`0M5Q0901`E3`0401LD3`0D3``050l<01<=3@8060lB00@05`h?01<9200C2@P04PT20P0E30<0
100F2@407Q4901hA2@0J3@D201T<1004024B2`0S50d08a@=02DF3`80;b0I0`0a6P`0100d7@l0>R<E
03XS5@0o:1X203`U5`04038K3@0Z4`D0:Q<502XC0`80:a@400@0;1D702TC1@0Y4`D09Q<50P0U4P@0
100U4`L09aD902LE2@0[6@d202`J3P0402XH300U50X09A@:02DD30809AL>00@08Q<<01/>1P0K3PH0
6a0:0P0I3PP0100G30P05@X801D:200E2@T201H:300401P<400I31<06@`C01HA3P803PT600@01P40
00@0000400001@000P09104300/60`0401H@100B30004P`0010:00805a4500@08adA02dW6`0]9a/0
;2HJ0P0V81@0100W81X06A8<01TB300@20D2014<200401PD4@0T81d0920M03@c<@80A4=100<0Bd];
04e=C@1=CDd00`1=CU001019CE00B4]@04Q;D018CE4204Y?D`0404i?E01<CE80C4eB04U:C`80BTYB
00@0C4aF04i>F01>CUP0C4aH0P1;BeL304U9D@<0BTYB0`1;Be<304a<E0@0C4aH0`1=CET504i>FP`0
C4aH0`1:CUT404]?FPT0C51K1@1;CeX304Y>F@<0BDeH0`18C5L304M;EP<0ATYE0`1;BeL0101<C5P0
CDeI04e=F@1=CET204a<F0<0Bd]G6019BTlI04Q9CP80AdQ=00<0ATM<04E6B`15AT/00`14ADX304E6
B``0ATM<3017Adl504M8C@40oooo000T04M>F0`0B4mI301:DE/<04U@FP`0BU5K3019D5X404]CH0<0
Be=N0`1;E5d204]DF`0404Y@D013BDT0@dU903Tm?P80;c4`00@09RPW020P7P0P81h07AhI0P0K71L0
100K6AX071PI01`H6@0L61T201dG5`0401dH5@0P61H081PF020H5@808ATF0`0O5a@0100R61H09ahI
02LN6@0/8Ad202hQ6`0402/N600W6Q809aXB02PQ6@809QlG00@091dE020I4@0P6A4071D=0P0G40P0
100B2`<03`P000l8000D304201D=0P0401XA200N5@`07QD<01XA2P805Pd600@06Q4<028I500R6A@0
7QL?0P0M5Ph0100O5Pl08QTB028I4P0V6aD202TL5P0402LJ500T5a4091LA02/L5`80;AhI00@0<B8M
038S7P0b8ah0;AhI0P0Y6QD0100b8ah0?RlZ03h_:P0`9Al202DJ500401/@2P0J3`T06Pl901/@2P80
6@h800@05P/501@90`0D2@<06Q4<0P0L4`h0100J4@`05`h901L>2@0J4@`2020G4P04024H4`0L4`h0
71<>01`A3@806a0<00<06Pl;01P=2@0H3@T0100H3@T201T>2P0401L<200H3@T060d901T>2P806a0<
00@07A8>01hC3`0N4`l07Q<?0P0O5100100S40T0:1D>02PE3P0^6a@202lL5@0402/H4@0U4P/09A8;
028?20808@h700<07Q05020B1`0P4PL00`0W6@h0100U5``0;b4F02lQ5P0e9a`202`N4`0402dQ5@0Z
7Q80:QhB02@J3P8091X@00@091`A02DL4`0U71<08a`B0P0Q6Q80100I4P`0614;01PA2`0F3P/201D=
2P0401D=2`0G3`d05`l=01T@4@806Q4B00@05a8>01@?2@0D3`T03`X40P0;1`00100:1P0030P000`8
000@3042018>0`0401@=0`0K50X06a@:01`E2`<05a450P0O6@/0100Y8aD0;2HH02`V600L5a@201HA
3P04014=300G4a@05a<D02@R8`80=C<f00<0@41204M7B@17AdT0101=CET204a<F0<0Bd]G0`1:BUH0
1019BED0BTYF04Y:EP1;BeL304a<F080Bd]G00<0BTYF04U9E@19BED00`19BTl404Q9CP80BDY?00<0
BdaA04a=DP1<CE803@1=CET604a<F080CDeI0`1?Ce/<04]?FP0404iBG@1=DE`0CE5L04eAG080C51K
00<0BdmJ04Y>F@1:CUT00`19CEP0101<C5P0Bd]G04];E`1:BUH304U9E@80BTYF00<0Bd]G04a<F01<
C5P0101=CET204a<F0<0Bd]G0`1:BUH<04U:C`0304Q9CP19BTl0BDY?00H0BT]@00<0BDY?04Q9CP18
BDh03018BDh704M8C@D0ATM<00@0B4U>04M8C@17B4d0ATM<0P15AT/304A5BP<0ADI;0`16Ad`01015
AT/0A4E:04A5BP14ADX204E6B`0304I7C017B4d0AdQ=00<0ATI>00@0AdM?04Q8D018B500B4Q@0P17
Adl304I6CPD0AdQ=0@3oool002@0AdiH3018CeT<04YAF``0BE1J301:DE/<04U@FP@0Be=P0`1;Deh3
04]DG@80BeAK00@0BU1@04=9B@13BDT0>Cdn0P0_<C00100V:2L0820N020P7P0M7QT201/L5`0401/I
6P0L61T071PI01`H6@807ALG00@07APE020H5P0P61H081PE0P0Q6AH301lG5004028H5P0W7QT09ahI
02`Q7@80;R4K00@0:ahH02LJ4P0W6Q80:24I0P0V7aL0100T7AD081TA020I4@0L5@d201L@2004018;
0`0?20003`P001@<0@805@d200@06Q4801hE300N5@`06Q4:0P0F3@H0100J4@`08QTD028I500N5`l2
01dF3P0401lF3`0R6A808QTB02HK5@80:A`F00@09aXD02@G4@0T5a40:a`G0P0]7QT0100a8Qd0<R<N
038S7P0]7QT202TJ5@04038S7P0n;bX0?RlZ030U7`809AXD00@06a0:01X?2@0J3`T06a0:0P0I3PP0
100F2`D050T301@90`0J4@`201`C3P0401XA300G3PT05`h901XA308081LB00@08APC01`C3P0L4`h0
714=0P0K40`00`0J3`/060d901P=2@0401P=2@806@h:00@05``801P=2@0H3@T06@h:0P0K40`0100M
4Ph07Q<?01hC3`0N4`l201lD400402<@2@0X5@h0:1D>02hK5080;a`E00@0:aPA02DB2`0U4P/08Pl8
0P0Q3PL00`0N40D08187020B1`0302LI3P0402DG300_8AH0;b4F03DW7080;1hC00@0;B4E02XN4P0Z
7Q8091X>0P0T6Q00100T71409A`C02DL4`0S7182024J4P0401TB300H4@/0614;01H>2`805@d:00@0
5@d;01L?3@0G3`d06A0A0P0J4A80100G4Ph050l901@?2@0?2P@200/7000400X6000<200030P0010<
0@804Ph300@050d301/D2P0K50X071D;0`0G4@D201lI2`0402TS5@0/9QP0;2HH01`G50805Q4>00@0
4@d<01LC500G4a@0928S0P0e<cH00`10@480AdM904M7B@0404e=F@80C4aH0`1;BeL304Y:EP0404U9
E@1:BUH0BTYF04];E`<0C4aH0P1;BeL00`1:BUH0BDUE04U9E@0304U:C`@0B4U>0P19BTl00`1;C540
C4eB04a=DP0=04e=F@H0C4aH0P1=CET304m?F``0BdmJ00@0CU9M04eAG01=DE`0CE5L0P1<D5/00`1;
CeX0BTiI04Y>F@0304U=F00404a<F01;BeL0Bd]G04Y:EP<0BDUE0P1:BUH00`1;BeL0C4aH04a<F004
04e=F@80C4aH0`1;BeL304Y:EP`0BDY?00<0B4U>04U:C`19BTl01P1:Be000`19BTl0B4U>04Q9CP0<
04Q9CPL0AdQ=1@16Ad`01018BDh0AdQ=04M8C@16Ad`204E6B`<0A4E:0`15AT/304I7C00404E6B`14
ADX0A4E:04A5BP80ADI;00<0ATM<04M8C@17B4d00`16ATh01017Adl0B4Q@04Q8D018B50204M7C`<0
ATI>1@17B4d10?ooo`009017CUP<04Q?F@`0BU5K3019D5X<04YAF``0B4mI101:DUl304YBG@<0BU=L
0P1:DeX0101;D5<0Ade=04M=C@0o@d@203<g=P0402T[:P0Q8b008B<P01dM6`806a/I0`0J61T0100K
5aP06aLF01/G5P0L5a@201lG5@03020H5@0Q6AH08ATF00<07aLD00@08ALE02DL5`0U71L0:AhJ0P0Z
7aT0100Z7AL0:1/C02PK4`0W81H202HO5@0402@M4`0Q6Q008AX@01hG3@806Q<900@05a0601D>100E
3P@03@L00P0@2P00100F3`D06Q<901XC2@0F3`L201090@04014:100G40X05a0:02XS6`80:B8J00@0
:R4J02/R6`0[8Q/0;24K0P0Y7QP0100V6A<08QD?028E3`0X6A@202HG4P<091D@00@08a@?02LH4`0W
61<0=BHQ0P14=C00100g;B@0;R@K02hT6`0Y7aH202`R6@0402dS6P0X7QD0:1hE02DK4P809Q`C00@0
8AL>02<I400S6A0081H=0P0N50/0100P5Pd09Q`C02HL4`0V71<2028H3`0401dB3P0L4@d0714=01X?
2`806@h:00<060d901T>2P0I3PX00`0I3PX601H;1`0401L<200I3PX06@h:01/@3080714=00@09A8;
02LD3@0W50d0:1D>0P0W50d0100T4@X08@h7024>1`0P3@H201l<1@0402PJ3`0T5P/091H;02PJ3`80
;1hC00@0;AlD03HX7@0f:1d0>bdR0P0d9Q/0100]8AD0:AdA02TM4@0S6@d2024G3@04020H3@0Q60l0
8AP?01hG3@807AH>0`0N5a40100M5Q006a<@01/C400I4@h201L?3@0401D<3@0D2``050/<018>2`80
4@d:00@03``700d:1@0=2PD03@X30P0=2P40100?30<040d4010=100D3@<201XC2@0401`E2`0G4@D0
5a4501D?0`806Q@800@081X<028L3P0R70h07ALG0P0H4Q80100F4Q4081`M020L7@0`;Rl203lo@@03
04M7B@19B4d0BDQ=00<0CDeG0`1<C5H304];E@<0BTYD00@0BDUC04Y:E01:BU@0Bd]E0`1<C5H204];
E@0304Y:E019BE<0BDUC00<0BDY?1018BDh204U:C`0304]<D@1<CE80C4eB00d0CDeG1P1<C5H204e=
E`<0CdmI301;CeP0101>DU/0CE5J04eAFP1=DEX204a@F@0304]?F01:CUL0BTiG00<0BDeF0`1;BeD0
101:BU@0BDUC04U9D`19BE<204Y:E0<0Bd]E0`1=CEL704a<EP80Bd]E3019BTl00`18BDh0BDY?04U:
C`0604Y;D00404U:C`18BDh0B4U>04M8C@T0B4U>0P19BTl=04M8C@80ATM<00@0ADI;04A5BP14ADX0
@dA90`14ADX204E6B`<0ATM<00@0ADI;04A5BP14ADX0A4E:0P15AT/00`16Ad`0AdQ=04M8C@0304I6
CP0404M7C`18B500B4Q@04Q8D080AdM?0`16ATh504M8C@40oooo000T04M>F1P0B4mI3019D5XH04Q?
F@@0B51M0`18D5/304QAFP80B55H00@0C55D04Y?DP1:Ce80A4Y:0P0j?Sl0100/<2l08R@S028T8`0L
7Q/201XJ600401XH6@0I5aP06ALH01TG60806QHE00@071LD01dH5@0M61D08ATF0P0R6QL0100N6AD0
7aLD01lG500O5a@2028I500402DL5`0W71H09a`F02LL5P80:1/C0`0W7QD0100U71<08QT@028I400P
5`h201hE30<07A@;00@030X000d;000=2`004Q040P0G5@T0100D4@P030T000`9000:1`0200d:0`04
014=100E3PH05@h601TB2P8081L@00@09QdF02/P6P0[81X0:b0J0P0/7aT302TJ5@0402`M600_81/0
;b0K02dN6@809aPC00@09QLB02XK5P0Z6aH0;bDI0P0X7Q80100[8AD0=R`P03H/800h;R83038X7080
>BlS00@0;bDI02hT600^91P0:b4E0`0V710202LM4@0402DK3`0P5PX081H:01hC3@807A8<00<06a0:
01T>200I3PP01P0H3@L00`0G30H05P/501H;1@0301D:100401H;1@0G30H05``601T>20806a0:00@0
:AP@024@200Q40P060L00P0C0P00100G1P008Q49028A2@0]71@203DT700403DW700_8AH0;b4F03@V
6`80>2XO00@0<B<H02lQ5P0_8AH0<B<H0P0]7a@0100[7a<09a/?02LK3`0Q5`/201hD2P0401`D2@0L
4`X071<:01XC2@806Q<;00@05@h601H?1`0F3`L0614;0P0J4`d0100J4Pl06A4?01TA3`0G3`d201D=
2`<030/900<03@`:00d<200=30P00`0>3@P300h>1P0401D>100L5@/071D;01lH3P8071D;00@06147
01XC2@0J4`T07AH<0P0N5`d0100L61L07QXI01hJ6@0S7b0202d[;00403hl?`1;Bdd0Bd]=04i=DP80
Bda@101<C5H504];E@H0BTYD1P1;BeD304Y:E0<0BDY?1018BDh204U:C`0304]<D@1<CE80C4eB00d0
CDeG1P1<C5H204e=E`<0CdmI301;CeP0101>DU/0CE5J04eAFP1=DEX204a@F@0304]?F01:CUL0BTiG
00<0BDeF00<0Bd]E04Y:E01:BU@01P19BE<00`1:BU@0Bd]E04];E@0704a<EPD0CDeG3019BTl00`18
BDh0BDY?04U:C`0604Y;D00304U:C`18BDh0B4U>00<0ATM<00@0AdQ=04Q9CP18BDh0B4U>0P19BTl3
04Y;D00304I7C017B4d0AdQ=00T0AdQ=00@0ATM<04E6B`15AT/0A4E:0P13A4T01012@dP0@dA904=4
B@13A4T204A5BP0404M8C@16Ad`0ATM<04E6B`<0A4E:0P15AT/304I7C0<0ATI>00@0AdM?04Q8D018
B500B4Q@0P17Adl304I6CPD0ATM<0@3oool002@0AdiH6018CeT<04U@FQP0B4mI1018D5d304Q@F`<0
B55J0P18DEP0101<DE@0BTmB04Y?DP14BTX203Xn?`0402``;`0R92<08R@S01`N6`806QXH00@06QPI
01TG600I5aP06ALH0P0J5QD0100L5a@07APE01dH5@0Q6AH2028J5`0401hI5@0O5a@07aLD01lG5080
8QTD00@09A`G02LL5P0W71H09a`F0P0X6a<302LN5@0402DL4`0R6A008QT@020G3P807QD<0`0M50/0
100<2P003@/000d;000B40@201LE2@0401@A200<2@0030T000X700803@X300@04@d401D>1P0E3PH0
6A8:0P0P5a00100V7AH0:b0J02/P6P0[81X202`O6@<0:AXE00@0;1dH02lP6`0_81/0;AhI0P0W61<0
100V5a80:Q/F02XK5P0_9AT202PN4P0402/Q5@0f;200=R`P03P^8P<0<RPL0P0i;b<0100_9AT0;R@H
02hT600[8AD302HL40809adA00@09A/?020F2P0P5PX07Q<=0P0M4P`00`0K40X06@h801T>200601P=
1`0301L<1P0F2`D05P/500<05@X400@05P/501L<1P0G30H06@h80P0K40X0100Y61008A08024@200H
1`0201<2000401L6000R4@T08Q4902dL5080=B@L00@0=BLL02lQ5P0_8AH0=2HK0P0h:Ql0100a8aP0
;b4F02lQ5P0a8aP202dO500402/O4`0W6`l09a/?024G2`807Q@:00@071@901`C2P0L4`X06Q<90P0J
4`/0100E3PH05Pl701H?1`0H4@/201XC3@0401XB3`0I4@l06A4?01L?3@805@d;0`0<2`T00`0=30X0
3@`800d<200300h=20<03Ph600@05@h401`E2`0L5@/07aP>0P0L5@/0100H4@L06Q<901XC2@0M5P`2
01hG3@0401`H5`0N6QT07QXI02<O8080;B//00@0?S`o04];C@1;Bdd0CTeB0P1;C50404a<EPD0Bd]E
1P1:BU@604];E@<0BTYD0`19BTl404Q9CP80BDY?00<0BdaA04a=DP1<CE803@1=CEL604a<EP80CDeG
0`1?CeT<04]?F00404iBF`1=DEX0CE5J04eAFP80C51I00<0BdmH04Y>E`1:CUL00`19CEH00`1;BeD0
BTYD04Y:E00604U9D`0304Y:E01;BeD0Bd]E00L0C4aF1@1=CEL<04U:C`0304Q9CP19BTl0BDY?00H0
BT]@00<0BDY?04Q9CP18BDh00`16Ad`01017B4d0B4U>04Q9CP18BDh204U:C`<0BT]@00<0ATM<04M8
C@17B4d02@17B4d01016Ad`0ADI;04E6B`14ADX204=4B@040493B013A4T0@dA904=4B@80A4E:00@0
AdQ=04I7C016Ad`0ADI;0`14ADX204E6B`<0ATM<0`16ATh01017Adl0B4Q@04Q8D018B50204M7C`<0
ATI>1@16Ad`10?ooo`009017CUP<04Q?F@`0AdiH3018CeTH04M>F0@0AdmL0`17CeX304M@F@80Ae1G
00@0BE5D04YBE01:DU@0BDiA0P0oADD0100a=CH08bLV02<W9P0L7Qd201TK6P0401XJ6P0I6AT06ATI
01TG608061LE00@06QHC01dH5@0M61D07aXF0P0P6aL0100N6AD07APD01dH500N5Q<201lH4P04024H
4`0S6Q<08aXC02HK5@809adD00@0:20E02LO500W7a@09QhC0P0S6a00100O5``07AD:01dE2P0M5@X2
01hF2`0401HD1`0E4`H05A<601HD20806QP<00@05a@;010=100@3@@03@X30P0@3@H0100:1P002`L0
00/7000?200201<<100401T@2@0O5Pl07aH?02@I4`809Q/E00@0:ahH02HI4`0V6A<091LA0P0U6180
100T5a4081<=020C3@0O4P`202<F400402lV5`0X7a00:1l@02XQ4P80<bXK00@0<2LH02@K300T6``0
91/<0P0_9QL0100Z7Ph0:Ad=02TM3@0Y7@d202XN3P0402hR4P0a9AD0<BDE034U5@80;b<C00@09a`F
02DJ500U6Q@08aPB0P0P5@l00`0N4`d07A8<01dB300301`A2`L06@h80`0J3`T201/@2P0401P7000I
20006@P001h=1@809QD=00@0;ahF03@S6`0d8a/0=B@L0P0c8QX0100d9Q/0<28G030R5`0i:b02040b
9`0403HX7@0^81D0;R0E038T6@80=2HK00@0<2@H02hR5P0^8QH0:1hB0P0U6a40100S6a0091/B02@K
4P0S718202@M5@<05a0800@05Pl701D>200E3PP04``60P0@20D0100>1P@030@200`40P0:2PX200/;
2`0400/;2@0=3@/03@d;00h?2P803a0;00@0414;014B300A4P`06A<70P0O6@d302<L4P0401lH3P0N
5`l07QL?020I4@808AXB00@071PI02PT9@0X92D0<c4b0P0m>ch01015ADL0C4eA04a=D@1>Ce<204a=
DPT0Bd]C1P1:BU8604];D`<0BTYB0`19BTl404Q9CP80BDY?00<0BdaA04a=DP1<CE803@1=CED604a<
E080CDeE0`1?CeL<04]@EP0404iCF@1=DUP0CE9H04eBF080C55G00<0Be1F04Y?E@1:CeD00`19CU@0
101;Be<0BTYB04Y:DP19BE4304Q8D080BDUA00<0BTYB04];D`1;Be<0101;Be<304a<E0<0CDeE0P1>
CUH<04U:C`0304Q9CP19BTl0BDY?00H0BT]@00@0BDY?04Q9CP18BDh0ADI;0`16Ad`204M8C@0404U:
C`1:Be00BT]@04Y;D080BdaA1016Ad`604M8C@80B4U>0`15AT/00`13A4T0@T=80493B0030493B0<0
@dA900@0B4U>04M8C@17B4d0ATM<0P15AT/01014ADX0ADI;04E6B`15AT/204I7C0<0ATI>00@0AdM?
04Q8D018B500B4Q@0P17Adl304I6CPD0ATM<0@3oool002@0AdiH3018CeT<04M>F0`0B4mI6017CUP4
04M?G0<0AdmJ0`17D5T204M@E`0404I@DP19DE@0BE5D04M?DP80@DI900@0=C/k02P/;@0X;2d07R8Q
0P0I7A`0100M7Ad06a/K01/K6`0I6AT201TH5P0401PG4`0L61D071PE01hK5P8081/G0`0M6QD0100M
61@07QLA01hG4@0O6182024H4@0302@K500V7A@09QdD00<0:R0F00@0:AlE02DK4@0U6a4081H<0`0L
4PP201dC2@0401lM400J60/06QP;01LE2@806AL;00@061D<01@A200D4@P04Pl80P0D4@X0100E4PT0
4a0701<@1`0B3PD201<<100401@=1@0J4@X06Q4:020G408091/D00@0<B@N02dP6P0]81X0;B0J0P0a
91h00`0`8ad0:QdG02XM5`0302HI4`0402TP4@0S6P/08aX;02<J2`80:1l@00@08aX;01T@0@0I4040
6a830P0V7@h0100V6`T09AX802DJ200U6PP202TN300402hS4@0`9A<0<2DC030U4`80;b@B00@0;28I
02/Q600[8AP0:1hE0P0U6a80100R60l081H=020F3@0P5Pd201lE300401L=100H3PD060h501X@1`<0
71890P0K4@P0100J40L06@l601T?1P0H2P4201d?1P0402PJ4@0e9ah0=BLN03h`9`80?c4X00@0>B/R
03<U700c9A`0>R`Q0`0^81D203@V6`03030R5`0^81D0;R0E00<0<R@I0`0S5`/0100P5PX07aD;01lE
2`0M5@X201dD2`0401dF300N5`l07QL?01XC2@80614700@05@h6018;0`0B2`<040T30P0?208300l7
100400h=2`0=30X03@`:00d<2P803@`800@03`h:018A300B4@`05A@?0P0G5`l0100O6@d08Q`@028L
400U7Q@202@M4`04020I4@0L5@l071D?01dE4P808ATF00@07A/L038`<@0b<340A4950P18B4X01015
ATX0AdQ=04M8C@18Be0204U<D@T0BTYB1P1;Be<604Y:DP<0Bd]C0`19BTl404Q9CP80BDY?00<0BdaA
04a=DP1<CE803@1=CED604a<E080CDeE0`1?CeL<04]@EP0404iCF@1=DUP0CE9H04eBF080C55G00<0
Be1F04Y?E@1:CeD00`19CU@0101:BU80BDUA04U9D@19BE4304Q8D0<0BDUA1P1:BU8304];D`80C4aD
0`1=CED<04U:C`0304Q9CP19BTl0BDY?00H0BT]@00@0BDY?04Q9CP18BDh0ADI;0`16Ad`204M8C@04
04U:C`1:Be00BT]@04Y;D080BdaA0`15AT/304I7C00304M8C@18BDh0B4U>00<0B4U>0`15AT/00`13
A4T0@T=80493B0030493B0<0@dA900@0B4U>04M8C@17B4d0ATM<0P15AT/304A5BP<0ADI;0`16ATh0
1017Adl0B4Q@04Q8D018B50204M7C`<0ATI>1@15AT/10?ooo`009017CUP<04Q?F@`0AdiH3018CeTH
04M>F0@0AdmL0`17CeX304M@F@80Ae1G00@0AU1B04UAE019DE@0AdmB0P11ATT0100e>c/0:2`]02P/
;@0N8R4201TM700401dM7@0K6a/06a/K01TI6@806APF00@061LC01`H5@0L61D07Q/F0P0P6aL301dJ
5@0401dH500N5a407QLA01lH4P808APA00<091/D02HM500V7A@00`0Z81H0100Y7aD09A/A02DK4@0P
5P`301`B20807A<900@07ad@01XH2`0J60/05aD90P0I5`/0100H5@`0514801@A200B3`P201@A2P04
01DB2@0C40L04a07018>1@804``400@050d501XA2P0J4@X081L@0P0T6a@0100a91h0;B0J02dP6P0]
81X2034T7P03030S7@0Z7AL0:QdG00<09QTC00@0:B0A02<J2`0S6P/08aX;0P0X7a00100S6P/06A01
01T@0@0K4P<202HM3P0402HK2@0U6PP09AX802DJ2080:Ah<00@0;R<A030U4`0`9A<0<2DC0P0_9180
100/8QT0:b4H02/Q600X7QD202DK4P04028H3`0P5Pd081H=020F3@807aD<00@05`d401P>1@0H3PD0
6Q070`0L4PT201/A200401X@1`0I3`H06@l601P:0@807@l600@0:1XA03DW7P0e9ah0?S0W0P0o<BP0
100i:b80<bDL03<U700j;24302hP5@80=2HK00<0<28G02hP5@0^81D00`0b91T302<G2`04020F2P0O
5@/07aD;01dE2P807A@;00@07AH<01hG3`0N5`l06Q<90P0H4@L0100E3PH04P/3018;0`0@2@<200l8
0P<03`L400@03Pd;00d<2P0=30X03@`:0P0=30P0100?3PX04Q4<018A300E50l201LG3`0401lI3@0R
71008Q`@02DN508091dC00@081TA01`E3`0L5@l07ADB0P0Q6AH0100M6a`0<S0a038`<@14@TD204Q8
BP0404E6BP17B4d0AdQ=04Q;D080BDaA2@1:BU8604];D`H0BTYB0`1;Be<304U:C`@0B4U>0P19BTl0
0`1;C540C4eB04a=DP0=04e=E@H0C4aD0P1=CED304m?E``0Be1F00@0CU=I04eBF01=DUP0CE9H0P1<
DEL00`1;D5H0BTmE04Y?E@0304U>E00404Y:DP19BE40BDUA04U9D@<0B4Q@0`19BE4604Y:DP<0Bd]C
0P1<C5@304e=E@`0BDY?00<0B4U>04U:C`19BTl01P1:Be001019BTl0B4U>04Q9CP15AT/304I7C080
AdQ=00@0BDY?04Y;D01:Be00BT]@0P1;C54304E6B`<0ATM<00<0AdQ=04Q9CP18BDh00`18BDh304E6
B`0304=4B@12@dP0@T=800<0@T=80`13A4T01018BDh0AdQ=04M8C@16Ad`204E6B`<0A4E:0`15AT/3
04I6CP0404M7C`18B500B4Q@04Q8D080AdM?0`16ATh504E6B`40oooo000T04M>F0`0B4mI9017CUP<
04Q?F@@0AdmL0`17CeX304M@F@80Ae1G0`16Ce@01014CU00@DU<0459C00k@4<2030f=P0402HZ:`0O
8b@07b<T020R8@807R0O00@071`L01TI5`0I6AL06APD0P0J6AD0100N6aH07a`G01lL5`0M6QD301hK
5P807QTC00@07aPB020I4@0P6A4091/D0P0V7A@0100[8AD0;B<G02dS5`0^91P202TO4`04028H300L
4PH0718601/A1@80718600@08adA01/E2@0K5@T05Pl50P0H4@L0100J4`/05a0801L@200F3`T201PA
2`04018?1P0A3PD04@h500l<0`803PX100@04P/301PA2@0H4@T08QTB0P0X7aP0100X7AL0:AhH02TN
600_91h203L/9P0403H[9@0/8A/0;24K02@I4`808AH@00@0:1hB02LM4@0W7A40:1hB0P0[8AD0100Y
7a<09adA02LM4@0[8AD2038X700402dP400Y70`0:A`<02LJ2P80:A`<00<0:Qd=02LJ2P0W6PX00`0S
5PH0100P5Pd07aD<01lE300M4`X201/A200301X@1`0I3`H06@l600<06@l600@060h501T?1P0I3`H0
6Q070P0K4@P401`B2@806a4800@08A@;02@G3P0T5`h0:QdD0P0b9A`00`0j;B@0?S4X03ha:003040c
:P0403T[800S5@X08aD:01L900806P`100@07a4602LI3P0W6@h0:A/@0P0Q4`P0100H30006Ph201X>
0P0J40@201X@1P0401P@1@0G3PD05`h501H?1@805a0800@06Q@801PA1`0H4@L05Pl50`0E3PH201L@
2P0401TB300J4Pl06Q8?01@?2`804`h:00@04Pd9018=1`0B3@L050l90P0G4`X0100K5`h07QX?01hJ
3`0R70h2028L4004028K4@0P6A4081TA01/D3P805Ph;00@06A4?01lF5`0O5QL09R@W0P0l?3h0101>
CU00Bda@04]<D013A4T20454B@0304E8C`17BU40AdYA00@0BDY?1@1:Be0604]<D@H0BT]@0`1;C543
04U:C`@0B4U>0P19BTl00`1;C540C4eB04a=DP0=04e>D`H0C4eB0P1=CU<304m@E@`0Be1D00@0CU=G
04eBEP1=DUH0CE9F0P1<DED00`1;D5@0BTmC04Y?D`0304U>DP0404Y;D019BTl0BDY?04Q9CP<0AdQ=
0P18BDh00`19BTl0BT]@04Y;D00604U:C`H0BT]@3019BTl00`18BDh0BDY?04U:C`0604Y;D00304U:
C`18BDh0B4U>00<0ATM<00@0AdQ=04Q9CP18BDh0B4U>0P19BTl304Y;D0<0ADI;0`16Ad`00`17B4d0
B4U>04Q9CP0304U:C`0404I7C015AT/0ADI;04A5BP80@dA900@0@T=804=4B@13A4T0@dA90P14ADX0
1019BTl0B4U>04Q9CP16Ad`204E6B`H0A4E:0`16ATh01017Adl0B4Q@04Q8D018B50204M7C`<0ATI>
1@15AT/10?ooo`009017CUP<04Q?FB@0AdiH3018CeT404Q@G@<0B51K0`18DEX204QAF00404MBEP16
Ce@0ATmD04A=DP80@d]>00@0@4Q:03]0@`0k@4<0<3Hf0P0W;Bd0100U9bH08R@S028T8`0M7ah201/K
6@0401TJ5@0J6AD06QTE01`K5P807a`G00@06QTD01`K5P0L6aH07a`G0P0P6aD0100N6A<081TA020I
4@0R6a<202DN500402hR5P0b9QX0<RHJ03DY7@80<BDI00<0:QhB02<G2`0S5`/00`0P50P0100S6@l0
6a4701/A1`0G3@@201/A200301lD3P0M4P`07A8<00<06Pl;00@0410600l?1@0?3`D040d40P0>2P40
100<20003`P000l8000D3@D201PA2@0401X?2@0E2P@05@X401<80P805@X400@05``601T>200I3PP0
8AH@0P0Z7aT0100Q5`h091XA02@J4@0T6Q42020F3@<07Q@;00@0718901X@1`0J40L09aT<0`0R50L2
02HH2`0402PJ3@0U5`X09AL:028D1`808aD800@060h401L=0`0G3@<05`d30P0H3P@0100I3`D06a47
01/A1`0M4`T201hD2P04024G3@0P5P`081H<01lE2`807Q@:00@07aD;028H3P0R60h09A/A0P0W7A<0
100P4`X0:A`C02TL4`0d9ah303`_9P80=RTP00@0;R4H02TL4`0Y71<08aD:0P0I2`00100G2@0070h3
01`>0`0S5@X2030R5`04038T6@0W6@h09aT>028F2P809AT=00@0:1hB02TO5@0Y7aD09AdB0P0R6A00
100P6@l081TA020I4@0O6@d201hH300401`E2`0J4`T06Q<901PA2@805Pl70`0E3PP301L>1`0401P?
1P0J4@P06Q4801`D2@807QH;00@07aL:020H2`0P60/08Q`>0P0O6@/0100N5`d07AH>01dF3P0I4@h2
01L?3@0401lF5`0Y82<0:B0S038b=080A4A600@0CdiC04U:CP19BTh0@DA90P14Adh00`16Be40A4U?
04A9C`0304Q9CP<0BDY?0`1:Be0304]<D@0404a=DP1;C540BdaA04Y;D0<0BDY?0P1:Be000`1;C540
C4eB04a=DP0304U:C`@0B4U>0P19BTl00`1;C540C4eB04a=DP0=04e>D`H0C4eB0P1=CU<304m@E@`0
Be1D00@0CU=G04eBEP1=DUH0CE9F0P1<DED00`1;D5@0BTmC04Y?D`0304U>DP<0BDY?00@0B4U>04M8
C@17B4d0AdQ=0P18BDh604U:C`H0B4U>0`17B4d<04U:C`0304Q9CP19BTl0BDY?00H0BT]@00@0BDY?
04Q9CP18BDh0AdQ=2@18BDh204U:C`0404A5BP15AT/0ADI;04E6B`80ATM<00<0AdQ=04Q9CP18BDh0
0`19BTl01017B4d0ATM<04I7C015AT/204A5BP0404=4B@14ADX0A4E:04A5BP80ADI;00@0BDY?04Q9
CP18BDh0AdQ=0P15AT/604A5BP<0ATI>00@0AdM?04Q8D018B500B4Q@0P17Adl304I6CPD0A4E:0@3o
ool002@0AdiH3018CeTT04M>F0`0B4mI1018D5d304Q@F`<0B55J0P18DEP01017DUH0ATmD04I?E014
CE8204=;CP040418BP0k@4<0>d13030f=P809bd]00@09BLV028T8`0R92<07AlN0P0K6aT0100I6QD0
6QTE01XI5@0L6aH201lL5`0401XI500L6aH071/F01lL5`8081/E00@07QTC020I4@0P6A408Q/C0P0U
7Q@0100^8QH0<RHJ038V6P0e:Ad2034U6@0302XN4P0S5`/08aL;00<081@800@08aT?01/A1`0K4@L0
5`d40P0K4@P00`0O50h07A8<01dB300301X?2`04010@1P0?3`D03`l5010=10803PX100@030P000l8
000?200050d50P0H4@T0100J3`T05@X401D:100C208201D:100401L<1P0I3PP06@h8024F4080:QlI
00@08AL>02@J4@0T6Q4091XA0P0P5Pd301hD2`0401`B2@0J40L06Q0702LI30<08Q@70P0V60/0100X
6Pd09AL:02DG2P0R50L202<E200401P>100G3@<05`d301L=0`8060h400@06@l501/A1`0K4@L07A<9
0P0N50X0100Q5`d081H<020F300O5@/201hD2P0401lE2`0R60h08QP>02DK4@809adC00@081<:02TL
4`0Y71<0=2LN0`0l;bH203HY800402hQ600Y71<0:A`C02<E2P806@/000@05`T001`>0`0L3P<08aD:
0P0`8QL0100b91T09aT>02LI3P0R5PX202DI3@0402PN4P0Y7aD0:AlE02DM4P808QT@00@081T?020I
4@0P6A407aT=0P0N60`0100L5@/06Q<901XC2@0H4@T201H?1`<05@h80`0G3PL0100H3`H06Q4801XA
200L50T201hF2`0401lG2P0P60/081P;028L3P807aT;00@07QL=01dF3P0M5Ph06A4>0P0G3`d0100O
5QL0:B0S02TP8`0b<S@204A4AP0404m>D`19BTh0BDY>0454B@80A4M>00<0AT]A04A9C`14BDl00`18
BDh304U:C`<0BT]@0`1;C540101<CE80BdaA04]<D@1:Be0304U:C`80BT]@00<0BdaA04a=DP1<CE80
0`19BTl404Q9CP80BDY?00<0BdaA04a=DP1<CE803@1=CU<604a=DP80CDiC0`1?D5D<04]@E00404iC
E`1=DUH0CE9F04eBEP80C55E00<0Be1D04Y?D`1:Ce<00`19CU8304U:C`0404Q9CP17B4d0AdQ=04M8
C@80B4U>1P19BTl604Q9CP<0AdQ=3019BTl00`18BDh0BDY?04U:C`0604Y;D00404U:C`18BDh0B4U>
04M8C@T0B4U>0P19BTl01014ADX0ADI;04E6B`15AT/204I7C00304M8C@18BDh0B4U>00<0BDY?00@0
AdQ=04I7C016Ad`0ADI;0P14ADX01013A4T0A4E:04A5BP14ADX204E6B`0404U:C`18BDh0B4U>04M8
C@80ADI;1P14ADX304I6CP0404M7C`18B500B4Q@04Q8D080AdM?0`16ATh504A5BP40oooo000T04M>
F1P0B4mI3017CUP<04Q?F@`0BE1J1018D5d304Q@F`<0B55J0P18DEP01019E5P0AE1D04E@E013C542
04A>D00404I>D@13B4/0@dQ;03Ll?`80;S@d00@09BTX028V9@0R9RD07b4P0P0L71X0100I6QD06QTE
01XI5@0L6aH201dL5`0401XI500M71L07A`G024N6@8081dF00@07aXD01hJ4@0N6Q408Q/C0P0T7A<0
100^8QH0<bLK03<W6`0h;20203HZ7P0302hR5P0W6`l09a/?00<091P<00@09QP?01lA200O4@P07@h7
0P0S50d0100X6A@09AHA02DF4@0Q4Pl201l@3@04018B200C4`T04a<901DB2@804Pl600@03`/200`8
000<20003PL00P0@2@40100M50d06Q4:01XA2P0I40T201XA2P0401P?200F3@H05Pd601XA2P808QTB
00@0:QlI034V800a9R00<bPR0`0^8ad2030U7`0402`Q6`0S61808aPB02TK40808aD:00@08Q@902HH
3@0V60d09aT>0P0Q4`P301d?100401lE2`0P5P`081H<024G3@8091X@00@09adC02/Q5`0[8AL0;bDK
0P0a9ad0100]8aT0:R0F02XP5P0U6a4202<I3`0402@J400Z81H0:R0F034W7@80=B/Q00@0>B`S03`_
9P0l;bH0?C0W0P0h:b80100/7aH07a8901lB2@0F2@0201440004024C200Z7140:Q`A03T[8080?BlT
00@0>B/P03d_900m;b@0?S0U0P0b91T0100N4PH08aL;02<G2`0V710202LM4`0402<K400N5@`07QD<
01XC2@806A8:00@05@l301H@100F40@061860P0I4PP0100J4`/0614901PA2@0G40X201H?2@0401T<
100K3PD06`h501hA20808AD900@08aL;02<G2@0S5`T08QH80P0Q5@D0100P6P`071H801`F200K50X2
01dF3P0401dE4P0O5QL07aHG02/R9@80>C0e00@0?3`n04Q7C018Ad`0C4eA0P17B4d01013AT/0BDaC
04U<D`19CU@204=8CP@0B4U>0P19BTl304Y;D0<0BdaA00@0C4eB04]<D@1;C540BT]@0`19BTl204Y;
D00304]<D@1<CE80C4eB00<0BDY?1018BDh204U:C`0304]<D@1<CE80C4eB00d0CDiC1P1<CE8204e>
D`<0Ce1E301;D5@0101>DeL0CE9F04eBEP1=DUH204aAE@0304]@E01:Ce<0BTmC00<0BDiB0`19BTl0
1018BDh0AdQ=04M8C@17B4d204Q9CP@0BDY?0`18BDh304M8C@<0ATM<0P15AT/<04U:C`0304Q9CP19
BTl0BDY?00H0BT]@00<0BDY?04Q9CP18BDh03018BDh304A5BP0404E6B`16Ad`0ATM<04M8C@80B4U>
00@0BDY?04Y;D01:Be00B4U>0P17B4d00`16Ad`0ADI;04E6B`0304A5BP<0ADI;00@0BT]@04Q9CP18
BDh0AdQ=0P15AT/00`14ADX0@dA904=4B@0304A5BP<0ATI>00@0AdM?04Q8D018B500B4Q@0P17Adl3
04I6CPD0A4E:0@3oool001P0B4mI1017CUP304Q?FA40BE1J3017CUP<04U@FP`0AdiH3018CeT<04Q@
F`0404QCE`17D5D0Ae1E04I?E0<0AdiD0P17C5801013B4h0@4E;0415B`0h>SD2038d;`0402T[9P0R
9Ah08RDN01lR6`807b8I00@07R4H01dP5@0M81D061TC0P0K71H301hN5P0401lL4`0N6a807Q/B024M
4P80920E100e;1d203HZ6P0403<W5`0a91<0<B@C02hQ3P80;Ql<00<0;R8F02XN4P0Z7Q800`0V6A00
100W6Q80:1/C02PK4`0Y71@202TL5P0401TF3@0N6a807Q/B01HC2P803P/200@040d400h;0P0>2`80
3@X10P0E4PT0100F4PT071P?01`H3`0M6A0301PD2`807a/B00@08AdD01hJ4@0N6Q409AXF0P0Z7a/0
100[81`09a`H02LL600X7AT302`Q7@80:1dI00@09alD02DK4@0U6a4081H<0P0M4@D0100J3P8070h1
01`>0@0O4@@2024D100401D=000P60/081P;02HN4@808a/>00@08AT<02DM400U7A009QhA0P0S6`h0
100H40<07aL:01lG2P0U7A0202XR5@0403<[7P0j<RD0>S8U03Ta9080<RXM00@0>2XO02/M4P0[7A80
81870P0O4@H0100P4PL08Q@9028D2@0/7Q<203T[80<0?S4Q00<0@cHV04@g9`14=bL00`0m<200100h
:a/0:A`<02TL300U7A0302HN4@808a/>00@07QH901/C1P0K4`H06a<60P0L50L0100L4`X060l601P?
1P0J4@P201lF3@04024H3`0N5@`07QD<01hE30808aXA00@0<B4B02HI2@0V6@T08QD50P0W6`/0100/
81009ah=02LN3@0Q60L201dD0`<07QL=00@071D?01XE4@0J5A407QPH0P0Y92P0100h<cT0@Ci5044n
A@17B4d204U:C`0404]<D@1:Be00BT]@04Q9CP80AdQ=00<0BDY?04a=DP1<CE800`18BDh404U:C`D0
BT]@3019BTl0101<CE80BdaA04]<D@1:Be0304U:C`80BT]@00<0BdaA04a=DP1<CE80301=CU<404a=
DP<0CDiC5@1>Ce@304e>D`P0C4eB0`1;C54304Y;D0<0BDY?1017B4d204I7C0<0ADI;0`14ADX0101<
CE80BdaA04]<D@1;C54204Y;D00304U:C`18BDh0B4U>00<0AdQ=3019BTl<04M8C@<0ADI;0`16Ad`0
0`17B4d0B4U>04Q9CP0304Q9CP`0ATM<00@0AdQ=04I7C016Ad`0A4E:1@13A4T304A5BP<0ATI>00@0
AdM?04Q8D018B500B4Q@0P17Adl304I6CP<0ATM90@17B4X104Q9B`40oooo000H04Q?F@@0AdiH0`18
CeTA04U@FP`0AdiH3019D5X<04M>F0`0B4mI3018D5/01018DeL0Ae1E04M@E@16Ce@304M>E080AdaB
00@0@dQ>0415B`10AD/0>3Xe0P0b=2l0100Y:bH08RDN028U7P0O8Q/201lR6@0401hQ600M81D07B0E
01PI4`806a`F0`0N7QH0100O71<07Q/B01hK4P0Q7A8202@P5@@0=B`M0P0f:QX0100c9aL0<B@C034T
4`0^8@h202hO300302hR5P0Z7Q80:QhB00<09QT@00@09aXB02PK4`0X6a<0:A`D0P0Y71H0100I5Pd0
7Q/B01hK4P0F4`X200h;0P04010=100>2`803P/200d:0@805A8900@05Q8901`H3`0L60l07AT@0`0H
50/201lK4P04024M500N6Q407QXA02DJ5P80:QlK00@0:b0L02LL600W71P0:1dI0`0/8Ad202PM6@04
02LO500U6a409A/A020F30807A4500@06Ph201`>0@0L3P407a440P0Q50@0100E3@0081P;020H2`0V
7Q4202<K3P04024I300U7A009Ad@02HN4@808a/>00@0610301lG2P0O5`X09Ad@0P0Z8QD0100c:ah0
>S8U03Xb9@0i<B@2038Z7@0403PZ7`0[7A80:adB020B1`807a4600@08187028D2@0R50T0;1hC0P0i
:b0303ha8@0304<f9P14=bL0A3LW00<0?C0P00@0>2/K02TL300Y70`09Ad@0`0V7Q4202<K3P0401hF
2@0K4`H06a<601/C1P8071@700@071<:01P?1P0H3`H06Q480P0O5Pd0100Q60l07QD<01hE300N5@`2
02<J4@04034Q4P0V6@T09QT9028E1@809a/;00@0;20@02LN3@0W7Pd08AP70P0M50<301hG3@0401`E
3`0J5A406QDA01hH6080:B@X00@0>3<i044nA@11?TD0AdQ=0P19BTl0101;C540BT]@04Y;D018BDh2
04M8C@0304U:C`1<CE80C4eB00<0B4U>1019BTl504Y;D0`0BDY?00@0C4eB04]<D@1;C540BT]@0`19
BTl204Y;D00304]<D@1<CE80C4eB00`0CDiC101<CE8304e>DaD0CTmD0`1=CU<804a=DP<0BdaA0`1:
Be0304U:C`@0AdQ=0P16Ad`304E6B`<0A4E:00@0C4eB04]<D@1;C540BdaA0P1:Be000`19BTl0B4U>
04Q9CP0304M8C@`0BDY?3017B4d304E6B`<0ATM<00<0AdQ=04Q9CP18BDh00`18BDh<04I7C00404M8
C@16Ad`0ATM<04A5BPD0@dA90`14ADX304I6CP0404M7C`18B500B4Q@04Q8D080AdM?0`16ATh304I7
B@40AdQ:0@18BD/10?ooo`006018CeT404M>F0D0B4mI3`19D5X<04M>F0`0BE1J6018CeT<04Q@F`04
04QCE`17D5D0Ae1E04I?E080AdiD00@0B4mE04U>E019CU@0ADY@0P12Add0100o@Ch0>C/h03Tk>00`
<Rd202PZ9@0402<V7`0P8aX082<J01dP5`806QdD00@05QLB01PI4`0H6A<06Q/E0P0K6a<0100L6A80
71T@01`I400N6Q42020L4@0402LO4P0X81<0:20C02/R4`80;b<E00<0<2@D038U5@0b9AD00`0a91<0
100g:al0<2@H030T600W6Q43024D2`D08aH>00@05A4601`H3@0L60d061@90P0E4@H0100H50T05A46
01DA1P0C3`@301TE2P8061@900@05105014=0P0A3@804`l40P0J5P/301hJ3`0401hC3`0R5a<08QLC
02<H50808AHB00@08aPD02LL600W71P0:1dI0P0U6QH0100f;R<0<RPN038X7P0[8AL202TM4@0302TM
3`0[7A00:ad@00<0:ah>00@0;B<G02/Q5@0[8AD0:R0D0P0Y7a<00`0V71008QP<028H3003020F2P04
02@J3P0_9AT0;bDI03T_8`80>c4U00@0>2hR03D[7`0e:al0;bDI0P0X7Q80100Q4`P08aD:02<E2P0U
5``202HH3@0402XL4@0`8QL0<28G03LY7P80>R`Q00<0?C0P03LZ6P0g:QX00`0f:AT0100g:QX0@C@T
044d9015>2P203`_7`0302XP500^91P0;R@H00<0<RPL00@0;R@H02PN4P0X7Q8091X>0P0R60`0100U
6a408AL=024G3@0P5P`202<I3`04028H3P0N50X07Q@:01hD2P808AL=00@0;1l?02PK2`0X6`/0:A`<
0P0^8Q800`0a9AL0;2<D02`S500302LP400401/D300M5Ph07AH>01lH4P808QdJ00@0:RHW03Ha=@0f
<CD0@Cm40P19ATd01017B4d0BDY?04U:C`1;C54204Y;D0<0B4U>00<0BDY?04]<D@1;C5400`18BDh4
04U:Ca40BT]@0`1;C540101:Be00BDY?04U:C`19BTl204Y;D0<0BdaA301=CU<404a=DP<0CDiC5@1>
Ce@304e>D`P0C4eB0`1;C54304Y;D0<0BDY?0`18BDh304M8C@<0ATM<0`15AT/0101<CE80BdaA04]<
D@1;C54204Y;D00304U:C`18BDh0B4U>00<0AdQ=3019BTl<04M8C@0304E6B`16Ad`0ATM<00<0ATM<
00<0AdQ=04Q9CP18BDh00`18BDh<04I7C00404M8C@16Ad`0ATM<04E6B`80A4E:0`13A4T304A5BP<0
ATI>00@0AdM?04Q8D018B500B4Q@0P17Adl304I6CP0504E6B016AdT0ATM904M8BP18BD/00@3oool0
01P0AdiH0`16CEL304M>F0@0B4mI3P19D5X<04M>F0`0BE1J6018CeT<04QAFP0404MBEP17D5D0Ae1E
04I?E080AdiD00@0BE1F04Y?E@1:CeD0B4eC0P15BU001016BTT0@4A30414@`0h?CT2030e<@0402T^
:00S:2408bPQ01`Q6P805a`E00@061XE01XK5P0J6aH06a`G0P0K6QD0100J6A@071TB01`I4P0M6Q<2
01lL4`04020H3@0R6Pl08QX?02DM4P80:R0D00@0;b<G038V600b9QP0=BLJ0P0e:1P0100g:R40<RDL
038U700]81L302XM50809aXA00@081<:01P;0P0H2`8071D;0`0Q6Q03020I3`806A8800@04``201L@
1P0G40H07QX?0P0I5@X0100F4PL061@901PD2@0J5P/201/G300401dI3P0O6a007a/@028G4@<091TC
0`0Q5Q0302@I4`808QLA00@091`A024G3@0Q5`d07Q@80P0P50P0100T60X0:1/;02PK2`0Z7@d202TL
300402TP4@0Q60T08AP901lF1`809Qd>00@0:R4B02TP4@0Y8140;2<D0P0b:AX01013>R/0ASd^04Hm
;P14>b`203`c900403@[700^9AH0;RDF02DL3@807A@500@0;1hA03LY700g:A`0?BlR0P0i:ah0100j
;1l0@S@W048d9`14=RT2040b9@0303LZ6P0a91@0<B@D00<0<B@D00@0;b8B03DX600e:1P0=2LG0P0W
6PX0100P5PX09A/?02DK3`0Z81@302dS5`80:R0D00@0:1hB02LM4@0W7A406a450P0G3@40100F3000
60h201P>0P0H3P8201L=0@0401P>0P0L4PH0718602XM3@80;1l?00@0<BDG03HZ700f:Q`0<2HJ0P0V
7Q40100S6a009AdB02DM4P0N5a42024I5P0402DP7@0`:RX0<2XZ03`h>@80AD=600@0CD]@055>E@1A
CUD0B4U>0P1:Be00101;C540BT]@04Y;D019BTl204Q9CP0304U:C`1:Be00BT]@00<0B4U>1019BTl5
04Y;D0d0BdaA0P1:Be0604U:C`0304Y;D01;C540BdaA0100C4eB0`1=CU<E04i?E0<0CDiC201<CE83
04]<D@<0BT]@1P19BTl404Q9CPD0AdQ=00@0C4eB04]<D@1;C540BdaA0P1:Be000`19BTl0B4U>04Q9
CP0304M8C@`0BDY?3017B4d404I7C0D0AdQ=0`18BDh<04I7C00404M8C@16Ad`0ATM<04E6B`80A4E:
00@0@dA904A5BP14ADX0A4E:0P15AT/304I6CP0404M7C`18B500B4Q@04Q8D080AdM?0`16ATh00`15
ATP0ATM904I7B@0204M8BP40oooo000H04M>F0<0ATeG0`17CUP404Q?F@h0BE1J3017CUP<04U@FQP0
B4mI3018DEX01017DUH0Ae1E04M@E@16Ce@204M>E00404U@EP1:CeD0BTmE04Q=D`80ADY@00@0ATY9
0414@`10A4<0>3di0P0`=C40100Y;RP08bPQ02<X8@0L8AX201LL5@0401PJ5@0J6aH06Q/F01/L5`80
6aXE00@06QTD01`I4P0L6A807AXC0P0O71<0100P60d08QX?028J3`0U7A8202XP500402lS5`0b9QP0
<RHH03DW6P80=BPH00@0=bXQ038U700b9A`0;B0G0`0Z7A@202LJ4@04020C2P0H2`8060/201`E2`<0
8AX@0`0P6@l201TB200401<<0P0G40H05a0601hJ3`806AD:00@05Q8701PD2@0H50T06QH;0P0K5``0
100M6@h07a/@01lK400R5a4302@I4`<08AH@0`0T6A<2028G4@0402@L4@0Q5`d08AL=01hD208081@8
00@091P:02PK2`0X6`/0:Qd=0P0Y70`0100Y81408AP9024H2@0O5PL202HM3P0402XQ4P0Y8140:B0A
02`S5080<RTJ00@0@cX[04Hm;P16?Bh0A3//0P0l<b@0100d:a`0;RDF02hU5P0U70d201dD1@0402`N
4@0g:A`0=bTL03d_8P80>B/N00@0>R`O048d9`12=2L0A3HY0P10<RD00`0g:QX0<B@D034T5003034T
500402lR4P0e:1P0=BPH03@W5`809aX:00@081H:02DK3`0U6`l0:R0D0`0]8aL202XP500402PN4P0W
7A409adA01/A1@805`d100@05P`001P>0P0H3P8060h20P0G3@40100H3P80718601`B1P0Z7@d202`O
3`04034U5`0f:Q`0=RXL030V6P809QhA00@08a/@02DM4P0U7A807QLA0P0Q6AH0100U81d0<2XZ030Z
:P0l>3T204E3AP0404e;D01ACUD0DDiE04Q9CP80BT]@00@0BdaA04Y;D01:Be00BDY?0P18BDh00`19
BTl0BT]@04Y;D00304Q9CP@0BDY?1@1:Be0=04]<D@80BT]@1P19BTl00`1:Be00BdaA04]<D@0@04a=
DP<0CDiC5@1>Ce@304e>D`P0C4eB0`1;C54304Y;D0H0BDY?1018BDh504M8C@0404a=DP1;C540BdaA
04]<D@80BT]@00<0BDY?04Q9CP18BDh00`17B4d<04U:C``0AdQ=1016Ad`504M8C@<0B4U>3016Ad`0
1017B4d0ATM<04I7C015AT/204A5BP0404=4B@14ADX0A4E:04A5BP80ADI;0`16ATh01017Adl0B4Q@
04Q8D018B50204M7C`<0ATI>00<0ADI804I7B@16AdT00P17B4X10?ooo`006017CUP404I=E`<0AdiH
4@18CeT<04M>F0`0B4mI3019D5X<04Q?F@`0Ae1I00@0Ae9F04I?E016Ce@0ADiC0P17CU@01019D5H0
C55G04aAE`1:CeD204Q=D`0404a@D@18C4d0B4a=0455A080>Shm00@0<cPd02/`:P0[<2X08RLQ0P0L
8A/0100O8B007b4N01lQ7P0N81d201hO6P0401hM6@0N7AP07QdH020M60808AhG00@091dE02DN5P0U
7QH0:1lF0P0/8QT0100^91X0<RDL038U700d:1`203DY7@0403DX7`0d9ah0=2LN038U7080<BDI00@0
=BTM03HZ700f:Q`0;R8D0P0T60X0100N5PT08AT<024I300T70l302PP4`808a/>00@08AT<02DM400U
7A007QP:0P0M5`T0100R70h0:28D02PR500U7a4201`F200401XD1P0N60X07QP:02@J4@<09A/B0P0T
6Q400`0U6a809adD02LM500302PN5@0402LO4P0Y7a<0:AlC02TO4`80:1`>00@09QX:02HI2@0V6@T0
9aX90P0Y70/0100R5PP081@6020D1P0S5`T202/O4@0403<W6@0k;b40>blQ04Hj;080D4@f00@0C40b
044e9`11=BL0<BDG0P0V6P`0100X70h0;B4C02dQ4`0Z7Q02024E1`0403<U600i:ah0>B/N03LY7080
:ad@00@091H902PJ3@0X6Pd0:Q`?0P0W6@`0100`8a<0<bHF03<V5P0m<202040c8`0403`_7`0n<B40
?S4Q03d`8080<B@D00@0;20D02/O4`0[7a<0:1`@0P0T60`00`0Q5@T07a<701lC1`0301hB1P04020D
1P0N4P@07Q8401dA0`807Q8400@07a<5020D1P0P50H08aL90P0U6@/0100Z7Ph0;b<C02lS4`0f:Q`2
03@Z7P0402TO5@0K4PT06a8901L@208071D=00@0:28R02`V9P0/9RH0=2h`0P0o>c`01019AdX0DDmD
055?E01AD5H2051?E@H0BT]@00@0BDY?04Q9CP18BDh0B4U>0P19BTl304Q9CP@0BDY?1@1:Be0<04a=
DP0404]<D@1:Be00BT]@04U:C`<0B4U>0P19BTl00`1:Be00BdaA04]<D@0@04a=DP<0CDiC5@1>Ce@3
04e>D`P0C4eB0`1;C54304Y;D0l0BDY?00@0C4eB04]<D@1;C540BdaA0P1:Be000`19BTl0B4U>04Q9
CP0304M8C@`0B4U>3017B4d304I7C0T0AdQ=3016Ad`01018BDh0AdQ=04M8C@15AT/504A5BP<0ADI;
0`16ATh01017Adl0B4Q@04Q8D018B50204M7C`<0ATI>0`15ATP104I7B@40AdQ:0@3oool001P0AdiH
0`15C5H304I=E`<0AdiH3`18CeT<04M>F0`0B4mI3019D5X<04Q?F@`0Ae1G00@0Ae9F04I?E016Ce@0
ADiC0P16CE<01019D5H0C55G04aAE`1;D5H204U>E00404aAE019CU40BDiA04E:C@80@DM700@0?490
03Dk>@0e>cT0;C<_0P0W;BT0100T:2T08RHW028V9`0Q8b8201lQ800301hN700O7QX07ahJ00<07QdI
0`0V7Q/0100W81X0:R4L02XQ700[8Q/202hS7@04034T700b9Ad0<RDM03T/9080=bXQ00@0<RDL02lS
5`0_8aL0=b/M0P13=bL01016>RX0@SLU048g9@0d:a`2030W6004038Y6P0f;Ah0=RdN034X6@80:b8C
00@0;B@E034X6@0a:1T08Q`<0P0T7Ph0100Z91@0;RPH02hX600W8A4201/E1@0401XD100P6PX081X:
01dC2@80718800@07A<9020F300P5P`08aT?0P0V7180100Y7aD0;28H02`R600W7a82030W600403H]
7P0f:QX0=RXJ02lS4`80;B0?00@0<bHE03T/6`0i;1/0<2@D0P0g:a/0100j;Qh0=RXJ03HZ6P0f:QX2
03hb8P0404Tm;@1?@c<0Cd<c02hR4P809a/;00@07Q8201dA0@0M4@409a/;0`0c9aL202`P40<0;B0@
00@0;1l?02XM3@0Z7@d0:ah>0P0b9AD0100k;Qh0@C@T044d900c9QH203/^7P<0B3/[00@0?c8R040c
8`10<b<0A3LW0P0n<B40100c9a/0<2@H030T600/81@202LK3`0402@H300R5PX08QH:020D20807a<7
00<091P802DI2@0U6@T00`0T60P00`0U6@T0:1`<02PL300302XN3P0402hR4P0_8aD0;b<E02lU6@80
:R8G00@08AP?01XC3@0J4`d08QXG0P0_9b@0100k=CT0?CPl03dh?012?D4204U7C004051>D`1BDEL0
DU5G04m>E080BdY@00@0BdaA04Y;D01:Be00BT]@1@19BTl00`18BDh0AdQ=04M8C@0304Q9CP@0BDY?
1@1:Be0<04a=DP0404Y;D019BTl0BDY?04U:C`<0B4U>0`19BTl204Y;D100C4eB0`1=CU<E04i?E0<0
CDiC201<CE8304]<D@<0BT]@2P19BTl504Y;D00404a=DP1;C540BdaA04]<D@80BT]@00<0BDY?04Q9
CP18BDh00`17B4d<04Q9CQD0AdQ=3`16Ad`01018BDh0AdQ=04M8C@16Ad`204E6B`0404A5BP15AT/0
ADI;04E6B`80ATM<0`16ATh01017Adl0B4Q@04Q8D018B50204M7C`<0ATI>00<0A4E704E6B015ATP0
0P16AdT10?ooo`006017CUP304E<EP<0ATeG0`17CUP?04Q?F@`0AdiH3018CeT<04U@FP`0B4mI3017
D5L01017DUH0ATmD04I?E015CU<204I=D`0404U@EP1<DEL0C55G04]@EP80BDiD00@0C55D04U>D@19
CU40ADY=0P11AdL0100l@T00=C/i03Dk>@0]<bl202L]:@0402@X:@0R9RL08RHW024S8P807b4P00<0
7QhL01lN6P0O7QX00`0N7AT302HN6`0402LP6P0Z8A`0:R4L02/R6`80;R<M00@0<B@L038U7@0b9Ad0
>B`T0P0g:R40100b9A`0;b<G02lS5`0g:ad204<g9`0404Hj:P12=bD0@SLU03@[7080<2LH00@0<RTJ
03H]7P0f;Ah0<BPI0P0[8Q<0100]91D0<BPI034X6@0R70`202@N3P0402XT500^:1P0;RPH02LQ4@80
6aD500@06Q@4020J2P0P6PX07A<90P0L4PP0100M4`T081H<020F300S6@l202HL4P0402TO5@0/8QP0
;28H02LO4P80<2LH00@0=RdN03HZ6P0f:QX0;b<C0P0]80l0100c9QD0>B`K03T/6`0`91@203L[6`04
03X^7P0f:QX0=RXJ03HZ6P80?S8R00@0BCd]04m3<`1?@c<0;R8B0P0W6`/0100N4P807A4101dA0@0W
6`/303<W5`80;20@0`0]8100100/7`l0:Qd=02XM3@0[7Ph2038U5@0403/^7P11=2@0@C@T03<V5P80
>bhN0`18>b/0100o<R80@3<S040c8`14=bL203ha8@0403<W6`0`91P0<2@H02`P50809a/?00@091P<
028F2P0R5PX081@80P0O4`L00`0T60P09AT902DI2@0302@H200302DI2@0X70`0:1`<00<0:Qh>00@0
;R8B02lS5@0_8aD0;bDI0P0Z8QL0100Q60l06Q<=01XC3@0R6QL202lW900403/e>@0m>3`0?CPl048m
@@80BDM<00@0D4iC059AE`1BDEL0CdiD0P1;BU00101;C540BT]@04Y;D01:Be0504U:C`0304Q9CP17
B4d0AdQ=00<0B4U>1019BTl504Y;D0`0C4eB00@0BT]@04U:C`19BTl0BDY?0`18BDh304U:C`80BT]@
401<CE8304e>DaD0CTmD0`1=CU<804a=DP<0BdaA0`1:Be0:04U:C`D0BT]@00@0C4eB04]<D@1;C540
BdaA0P1:Be000`19BTl0B4U>04Q9CP0304M8C@`0B4U>5@17B4d?04I7C00404Q9CP17B4d0AdQ=04I7
C080ADI;00@0A4E:04E6B`15AT/0ADI;0P16Ad`304I6CP0404M7C`18B500B4Q@04Q8D080AdM?0`16
ATh00`14ADL0ADI804E6B00204I7B@40oooo000H04I=E`@0ADaF0`16CELY04M>F1P0B4mI3016CeH0
1017DUH0ATmD04I?E014CE8204E<DP0404M>E01;D5H0Be1F04Y?E@80BDiD00@0BU9E04Q@D`18D5<0
ATiA0P14C4h01012BdX0?TM603i7AP0h@Ch203<l>@0402``<`0X;2l0:2`_02<W:0808B8T00@07b4P
01hN7P0N7Qh07AdK0P0L6aT0100P6QX08A`I024L6@0S6aT202@L6@0402LM6`0Z7a/0:QlK02`Q7@80
;B8N00@0<RDM03HY8@0f:B40=RTP0P0c9a/0100j;Qh0B3d[04Pm:`1=@Rh204Xo:`0404e1<@12=RH0
@SHV048f9P80@cLW00@0>2`L030T500`91@0<RHF0P0e:AT0100e;Q`0<b`J03</6P0c;1X2038[6@04
02hW5@0Z8a40:R<A02/T4P80;bPF00@08AP901dD1@0M50D06a830`0L4`@201/B0`0401`C100P5`P0
81L801XC0`809ah?00@0<RTH03L[6`0g:a/0<RLE0P0b9A@0100i;1T0@C@Q044d8@0g:QX303lb8P80
=BPH00<0<2<C03DX600e:1P00`0i;1`0100U60P0;R4A02hQ4@0f:AT203T/700403`_7`0m<200?C0P
03T/7080<RDE00@0?S4Q03/^7P0k;Qh0=RTI0P0b9AD0100a91@0<bHF03<V5P0g:QX203T/700403ha
8@15>2P0ACPX0513<`80C3l_00@0>bhN03HY6@0f:AT0=BPH0P0]810302TK400402XL4@0/7Q<0;1hC
02hP5@80;AlD00@0:A/@02HH3@0V60d07Q400P0S5PD0100W6PT0:A`;02TL2`0_8Q4203HY600403X]
700i;1/0>B`K034X5`80;RDF00@0:b4E02PO5P0X7aH09b0J0P0^9R@0100n>3P0D4Y<051:C019A4/2
04Y5C00404Y7CP1=BU40CDYA04m>E080CTeC00@0BdaA04Q9CP18BDh0C4eB0P1;C54304U:C`0404Y;
D019BTl0BDY?04Q9CP80ATM<0`18BDh404U:C`D0BT]@301;C540101:Be00BDY?04U:C`18BDh304M8
C@80B4U>00<0BDY?04Y;D01:Be00301;C54404a=DP<0CDiC5@1>Ce@304e>D`P0C4eB0`1;C54304Y;
D0<0BDY?0`18BDh404U:C`D0BT]@00@0C4eB04]<D@1;C540BdaA0P1:Be000`19BTl0B4U>04Q9CP0K
04M8C@<0B4U>1017B4dA04I7C00404U:C`18BDh0B4U>04I7C0D0ADI;0`16Ad`304I6CP0404M7C`18
B500B4Q@04Q8D080AdM?0`16ATh304A5A`40ADI80@16AdT10?ooo`006016CEL604E<EP<0ATeG9`17
CUPH04Q?F@`0ATmD00@0B5=G04I?E016Ce@0@daA0P14Be401016CE<0BDiD04U>E019CU@204Q=D`04
04U@EP18CeD0B4mE04M>E080AdmB00@0AdmA04E=C`15CDl0@DY90P0mATD0100h?D40<cPl03<h?00]
<C@202L[;P0402HW:@0U9BD09BDU028R8P80820P0`0S7Al0100R71`091/L02@K700U7A/202PN7@04
02/P7P0/8Al0;24O02TL5080=2LO00@0>RdT03L[7@0g:ad0>2`L0P0o=200100o=1h0>2hE03P^5@10
<b8203DX5`0403T/6`11=2<0@C@S03T/6`80=2LF00@0>bhM03lb8@0o<R40?3DR0P0f;a`302hW5004
03D^6`0k=240>c@Q03L`7@80;bPE00<0;bHE02TP3`0Y80l00`0U70/0100R6@P07A@301dD0`0M50<2
024H1`0402XS4`0a:1L0<BPG03L^7@80>RlM00@0=b`J03@W500d9a@0=2LD0P0e:1D302lR4@0402hQ
400_8Q40;b8A038U5080=BPG00@0<bHE030S4P0`8a80>2/J0P12=B@01019?2/0B3/Z04Pk:P13=RD2
03lb8@0403P[6P0b9A@0<RDD044d8`80@3<R00@0>B`K038U500b9A@0=RTH0P11=2<01015>2L0@3<R
040c8P14=bL204Dh:00404/n;P19?2`0BC`/03lb8P80?S4Q00@0?C0P03<V5P0c9QH0;1hC0`0[7A83
02`N4`80:Q`A00@09QP=028D2@0R50T0:ah;0P0b9A80100h:aP0?2lL03`_7012=B8204Xm:P0404e0
;@1;?R/0Bch[03@[6P80<BPI00@0<RXO03L`:00g<2P0?SHc0P12?3`0101>B4X0FE=G05UCE`1=Ae<3
04a9E0<0BdYB0P1:BU80101<CE80CTmD04i?E01=CU<204]<D@<0BDY?0`1:Be000`17B4d0ADI;04E6
B`0304Q9CP@0BDY?4@1:Be0304U:C`0404Q9CP17B4d0AdQ=04M8C@80B4U>0`19BTl<04]<D@@0C4eB
0`1=CU<E04i?E0<0CDiC201<CE8304]<D@<0BT]@0`19BTl304M8C@<0B4U>0`19BTl304Y;D00404a=
DP1;C540BdaA04]<D@80BT]@00<0BDY?04Q9CP18BDh06`17B4d404Q9CP80AdQ=1016Ad`204E6B``0
ATM<00@0BDY?04Q9CP18BDh0AdQ=0P16Ad`304E6B`<0ATM<0`16ATh01017Adl0B4Q@04Q8D018B502
04M7C`<0ATI>00D0@dA604A5A`14ADL0ADI804I7B@010?ooo`006016CEL604E<EP<0ATeG9`17CUPH
04Q?F@`0ATmD00@0B5=G04I?E016Ce@0@daA0P14Be401016CE<0BDiD04U>E019CU@204Q=D`0404U@
EP18CeD0B4mE04M>E080AdmB00@0AdmA04E=C`15CDl0@DY90P0mATD0100h?D40<cPl03<h?00]<C@2
02L[;P0402HW:@0U9BD09BDU028R8P80820P0`0S7Al0100R71`091/L02@K700U7A/202PN7@0402/P
7P0/8Al0;24O02TL5080=2LO00@0>RdT03L[7@0g:ad0>2`L0P0o=200100o=1h0>2hE03P^5@10<b82
03DX5`0403T/6`11=2<0@C@S03T/6`80=2LF00@0>bhM03lb8@0o<R40?3DR0P0f;a`302hW500403D^
6`0k=240>c@Q03L`7@80;bPE00<0;bHE02TP3`0Y80l00`0U70/0100R6@P07A@301dD0`0M50<2024H
1`0402XS4`0a:1L0<BPG03L^7@80>RlM00@0=b`J03@W500d9a@0=2LD0P0e:1D302lR4@0402hQ400_
8Q40;b8A038U5080=BPG00@0<bHE030S4P0`8a80>2/J0P12=B@01019?2/0B3/Z04Pk:P13=RD203lb
8@0403P[6P0b9A@0<RDD044d8`80@3<R00@0>B`K038U500b9A@0=RTH0P11=2<01015>2L0@3<R040c
8P14=bL204Dh:00404/n;P19?2`0BC`/03lb8P80?S4Q00@0?C0P03<V5P0c9QH0;1hC0`0[7A8302`N
4`80:Q`A00@09QP=028D2@0R50T0:ah;0P0b9A80100h:aP0?2lL03`_7012=B8204Xm:P0404e0;@1;
?R/0Bch[03@[6P80<BPI00@0<RXO03L`:00g<2P0?SHc0P12?3`0101>B4X0FE=G05UCE`1=Ae<304a9
E0<0BdYB0P1:BU80101<CE80CTmD04i?E01=CU<204]<D@<0BDY?0`1:Be000`17B4d0ADI;04E6B`03
04Q9CP@0BDY?4@1:Be0304U:C`0404Q9CP17B4d0AdQ=04M8C@80B4U>0`19BTl<04]<D@@0C4eB0`1=
CU<E04i?E0<0CDiC201<CE8304]<D@<0BT]@0`19BTl304M8C@<0B4U>0`19BTl304Y;D00404a=DP1;
C540BdaA04]<D@80BT]@00<0BDY?04Q9CP18BDh06`17B4d404Q9CP80AdQ=1016Ad`204E6B``0ATM<
00@0BDY?04Q9CP18BDh0AdQ=0P16Ad`304E6B`<0ATM<0`16ATh01017Adl0B4Q@04Q8D018B50204M7
C`<0ATI>00D0@dA604A5A`14ADL0ADI804I7B@010?ooo`006016CEL00`14BeD0ADaF04E<EP0304E<
EP<0ATeG<`17CUP<04Q?F@`0ATmD00@0B5=G04I?E016Ce@0@daA0P13BU001015C580B4eC04Q=D`18
CE<204M<DP0404YAF@18CeD0B4mE04Q?E@80B51C00@0BE5D04UAD`19DE<0ATi@0P13Bdd01014BDd0
?T=703i3A`0g>Sl2034e>004030a<`0];S00;Bh`02XZ:P809bLW00@0:bHZ02/U:@0[9BT0:B<U0P0Y
8240100X7b00:AlN02TO7P0Y7ah202XP7`0402`O6@0g:R80=bXR03T/8`80;b<E00@0:b0>034V4P0a
9Q80<RP?0P0]8`X0100c8a<0:1P802PH200`810203/[6`0403DU5@0a8A40<B4A03PX6080?2`L00@0
<2PE02PP3@0X80d081P50P0S6`P0100e;AX0@3PU040h9@0d;1T201lG100403@[6P0_9QD0;bHE02dT
4`80<2LF00@0<BPG02lV5@0_9QD0<RTH0P0i<1l0100/9A<0;B@C02dT4`0_9QD203@Y5`0403DZ600c
9Q<0<bHC02dQ2`80:1`600@0<2<B02HI200V6@P09AP70P0c9QD01010<b80@C@S044d8`0l;ah203P[
6P0404<f9@11=2<0@C@S03/^7@<0=BPG0P0i;1/0100h:aX0<bHE03<V5@0k;Qd203lb8@0403LZ6@0/
7`h0;1l>03T/6`80EDPg00@0H5=205Q;>P1HBcX0@cHV0P0i;1`0100b9AD0;B0@02dP400/7`l203T/
7004048e9@0l;al0?2lO034S6080;AlD0`0Y6a00100/7Q<0;b4F02lQ5P0`8QL202lQ5P0403X^6010
=1h0@3@N04@h8P80@SHP00@0@cLQ04Hj9016>R@0A3PR0P0n<Q`0100c;1X0=BhN03D^7P0o=b`204]4
?004059:A`1>B4X0CTQ:04Y4B080BDA:00@0BDEC04Y6E01:AU@0BTMB0P18AU401017ATh0BDUA04U9
D@1>DEH205=FF`0304i?E01;C540BdaA00<0BDY?0`1:Be000`17B4d0ADI;04E6B`0304Q9CP@0BDY?
1@1:Be0?04U:C`0404Q9CP17B4d0AdQ=04M8C@80B4U>0`19BTl<04]<D@@0C4eB0`1=CU<E04i?E0<0
CDiC201<CE8304]<D@<0BT]@0`19BTl01016Ad`0AdQ=04M8C@17B4d304Q9CP<0BDY?0P1:Be00101<
CE80BdaA04]<D@1;C54204Y;D00304U:C`18BDh0B4U>01/0AdQ=1018BDh204M8C@<0ATM<0`15AT/<
04I7C00404U:C`18BDh0B4U>04M8C@80ATM<00@0ADI;04I7C016Ad`0ATM<0P17B4d304I6CP0404M7
C`18B500B4Q@04Q8D080AdM?0`16ATh00`13A4H0A4E704A5A`0204E6B040oooo000`04I=E``0AdiH
3016CELH04M>F0`0AE1D101:CUT204U=F0<0B4aG0`17BeH304I=E`0404M>F018CeT0B4mI04Q?F@80
AdiH0`16CEL0100oAT`0?TE;03i5B`0mA4X203]2B00403PoA@0f?D<0=Se303@k@@80<cY00`0d<C`0
100c<3/0<Rlh038_>00_;3D202dZ<@0402XW;P0X9R/0:2H[02XV8`80:B@P00@0:2@K02`V6P0/9QX0
<B`I0P0f<1X0100h<QX0>C8H03Tb6012;ah204Ld8`0404dk9`1?>RL0CcXW04/f8@80BC<N00@0BS@M
04he7`1>=Al0@CTT0P0h<1/0100b:Q<0>c0J03/`6P17?B@204a09P04044e6`0g:@l0=bT?03La5`80
<R`B00@09Al702<M1`0S7@L0:b@A0P0W80h0100S70`0:b<F02/S5P0[8`h203Ta700403hc7`0c:1@0
<bPD03lb7`80@3<P00@0;b0=03LX5@0g:1D0?CDP0P0h<1/0100g;aX0>S8M03Xb7@0h<1/303<[5P<0
>30K0P0a:A@0100/90l0<2PC030X4`0g;aX203Ta700403D]600`:1<0<2PC03@R5080>BLI00@0AS@V
0590<P1B@380Ccd_0P12<280100i9aT0>RPJ03XX6P12<28204Le9`0404Hd9P0m:ad0?B/M03DS5@80
<1h@00@09aD701d;000M2`00:adB0P0]7a@0100V60d0:Q`C02XL4`0m;bH2048c;00403h_:P11<Rd0
@C8]05M><@80DdT^00@0CT@Y04M09P17@2H0A3dS0P11>b<01012?2@0@C`V044l9P14@SD204E3=`04
04E5>`16Ad40ATM104U:A@80Bde<00<0CDi@04i?D`1>Ce<00`1<C5P304];E`<0BTYF0`19BED404Y;
D0D0BDY?0`18BDh@04U:C`<0B4U>1@17B4d00`18BDh0AdQ=04M8C@0604I7C00404M8C@18BDh0B4U>
04Y;D0D0BdaA0`1<CE8304e>DaP0CTmD00<0CDiC04i?E01>Ce@0101>Ce@204e>D`<0BdaA0`1<CE83
04]<D@<0BT]@0`19BTl404M8C@<0B4U>2@19BTl304Q9CQ40AdQ=0`16Ad`904M8C@@0ATM<1@17B4d3
04Q9CP`0AdQ=00@0BDY?04Q9CP18BDh0AdQ=0P16Ad`01015AT/0ATM<04I7C016Ad`>04M8C@0504A5
A`15ATP0ADI804I7B@17B4X00@3oool00300ATeG3017CUP<04I=EaP0AdiH3015D5@404Y>F@80BDeH
0`18C5L304M;EP<0ATeG00@0AdiH04Q?F@18CeT0B4mI0P17CUP304I=E`0403m6C00nAD/0?TE;03e4
BP80>d9800@0>3m503Hm@`0f?D<0=3]10P0c>T0303@a?00403<`>`0b;cP0<Rlh02l/=@80;BXa00@0
:RL^02PV:`0X9R/0:RHS0P0Y9200100X91/0;2HJ02`V6P0a;1T203H`6P0403Pb6P0i<QP0>C8H048_
7P80Ac@S00@0CC/W04lj9`1?>RL0BcHQ0P19<ah0101:=1d0CSDO04he7`11>B@203P`6`04038Z4`0k
<1X0>c0J04Lm9080C40V00@0@CDK03LY3`0g:@l0=c4G0P0b;180100U7`L08ad702<M1`0[914202LP
3P0402<L300[8aH0:b<F02/S3P80>C4L00@0?S<O03<X500c:1@0?c8O0P10<b00100_80d0=bPE03LX
5@0m=B0203P`6`0403L_6P0j<Qd0>S8M03P`6`<0<b/F0`0h<1/2034Y500402`T3`0`:1<0<2PC03L_
6P80>C4L00@0=BdH030X4`0`:1<0=28D0P0i9aT01016=2H0DT0b0590<P1??Bl2048`8P0403TW6@0j
:1X0>RPJ048`8P80AcDW00@0AS@V03d[7@0m:ad0=B<E0P0`7Q00100W5@L07@/001d;000[7A8202dO
500402HH3@0Z71<0:Q`C03d_9P80@S</00@0?RlZ044b;@11<Rd0Edha0P1CBBh0101>A2T0Ad0V04M0
9P14?B<2044k8`04048l9011?2H0@C`V04A2=@80AD<g00@0ADDk04I7@@16Ad40BDY50P1;CD`00`1=
CU00CTmC04i?D`0304a<F0<0Bd]G0`1:BUH304U9E@@0BT]@1@19BTl304Q9CQ00BDY?0`18BDh504M8
C@0304Q9CP17B4d0AdQ=00H0ATM<00@0AdQ=04Q9CP18BDh0BT]@1@1;C54304a=DP<0CDiC601>Ce@0
0`1=CU<0CTmD04i?E00404i?E080CDiC0`1;C54304a=DP<0BdaA0`1:Be0304U:C`@0AdQ=0`18BDh9
04U:C`<0B4U>4@17B4d304I7C0T0AdQ=1016Ad`504M8C@<0B4U>3017B4d01019BTl0B4U>04Q9CP17
B4d204I7C00404E6B`16Ad`0ATM<04I7C0h0AdQ=00D0A4E704E6B015ATP0ATM904M8BP010?ooo`00
<016CEL<04M>F0`0ATeG6017CUP<04I?E0@0BTiI0P19CEP404Q<E`80Ad]F0`16CEL01017CUP0B4mI
04Q?F@18CeT204M>F0<0ATeG00@0@dY@0499C`12BDl0@DQ>0P0oAT`0100mA4X0>d9803]2B00j@DL2
03U0AP0403XjAP0i>DD0>CU503Ti@`80>3Q000@0=CDm03<d>@0c=3T0<34f0P0_<3D0100^;B/0:RTU
02XY9@0W9R4202TV7@0402dY7@0b;ah0<RlN03Lb7P80>3<M00@0?BdM044a8@11<B40AC@T0P12<Al0
100l:AP0>2<@03PS400g8Pl203PS3P0404Ll:01;?bT0BclY05=7<@80H54j00@0JEY306eMA01]GD@0
J5Po0P1SDcX0101ECCH0CTH_04i6;`13>bH203de8P0403T`7`0b:AP0<RTH02lV5`80=B`M00@0@SLS
04e2;P1=@Rh0BSdZ0P0i;1T0100k;1T0=RDC03HU4`0X5`D203@S4@04040e7`18?BL0B3dW04i3;@80
Ac`V00@0<bPB02@I0`0T6@<0:1d70P0e:Q@0100Z7`T0>RlI03X_6@19?RP204]0:P0404Pm9`14>B<0
A3TS03db7080=BXD00@0CSd]04`k:`1<>b/0BcXZ0P19>2P01014<b<0?B`L03d/700k:QX203`[6`04
044`8010;al0@2lO03XY6@80<R4A00@0;ah>038Q4@0b8A40<R4A0P0_7Ph0100/7`h0=bXJ03LZ6P0n
<B4204Li;00405E7>P1GBCh0EdTn05A6>`80FD]200@0CD8/04m4;P1?A2h0D4Pa0P1CBcH0101CC3T0
DT/h059;>01@BcP204m:=`0404a<@P1<C4@0C4a404a;AP80Bda700@0BTY:04U:C019BT`0BDY>0P17
BTl304a<EP<0Bd]E0`1:BU@304U9D`@0BT]@1@19BTl304Q9CQ00BDY?0`18BDh504M8C@0304Q9CP17
B4d0AdQ=00H0ATM<00<0AdQ=04Q9CP18BDh0101:Be0304]<D@D0C4eB301=CU<<04i?E00304e>D`1>
Ce@0CTmD00@0CTmD0P1=CU<304]<D@<0C4eB0`1;C54304Y;D0<0BDY?1018BDh504U:C`<0BT]@1019
BTl304Q9CQP0AdQ=2@16Ad`504M8C@<0B4U>3017B4d01019BTl0B4U>04Q9CP17B4d204I7C00404E6
B`16Ad`0ATM<04I7C0h0AdQ=00D0A4E704E6B015ATP0ATM904M8BP010?ooo`00<016CEL<04M>F0`0
ATeG6017CUP<04I?E0<0BTiI1019CEP504Q<E`0304I=E`17CUP0AdiH00H0B4mI00<0AdiH04I=E`16
CEL00`16CE<00`15C580A4]A04A;D@0304=:D0030499C`11B4h0@DQ>00<0@DE@00@0@4A?0414C@10
A4d0?T=90P0l@DL0100i?T80>3e103Pm@@0c=cX202hb<`0402X/:`0W:B@09bTT02//9080;c4T00@0
<c@V03Df9P0e=RH0?c<W0P11=BT01012=RP0?c4T03la900g:QX203<S5004030P400a8100<B0@03h_
7080BcXX00@0F4Lc05m=>@1OCCT0FTPb0P1D@2P0101?>b<0CCTQ04di8@1=@Rh204Ll:00404@i9`0n
<b40?S<Q034U5@80:Qh>00@0;R8D03@X6P0d:1X0<b@A0P0a8Pl0100X6@H09A@202DD0P0Y60H202XG
1P0403dZ6@1XDd80J5=203@U3P80<24:0`0]7PL0100^7`P0<B8;034R2`0h:A82040a6P0404/l9@1F
Ac00ETL`05E6;`80A3DN00@0>BXC03`]5P0l;AH0@C8K0P10<AX0100i:QL0>b`I03//6@0k;1T303X[
6080>BXG00@0=BHC030Q3P0`8@h0<B8?0P0f9a@00`0k;1T0?2dJ03`]6P0303d^6`0403TZ5`0c9140
<b@A03X^6080ACTS00@0BShX04Tl:@19?2T0ASTV0P0n<B00100j;A`0?c8R03lb8P1CBST20558>@04
04i7=`1;@cH0Bd<f04U3=@80BD<g0`18A3P0101CDTh0DU5?059AC`1@D50204i>D00404a=D@1:Be00
BT]@04M:D@80AdYC0`1<C5H304];E@<0BTYD0`19BE<404Y;D0D0BDY?0`18BDh@04U:C`<0B4U>1@17
B4d00`18BDh0AdQ=04M8C@0604I7C00304M8C@18BDh0B4U>00@0BDY?0`1:Be0504]<D@`0CDiC301>
Ce@00`1=CU<0CTmD04i?E00404i?E080CDiC0`1;C54304a=DP<0BdaA0`1:Be0904U:C`<0BT]@0`1;
C54404U:C`<0B4U>4@17B4d01019BTl0B4U>04Q9CP18BDh204M8C@<0ATM<0`15AT/404I7C0D0AdQ=
0`18BDh<04M8C@0404U:C`18BDh0B4U>04M8C@80ATM<00@0ADI;04I7C016Ad`0ATM<3P17B4d01@14
ADL0ADI804E6B016AdT0AdQ:0040oooo000`04I=E``0AdiH3016CELH04M>F0`0ATmD0`1:CUT404U=
F0D0B4aG00<0ATeG04M>F017CUP01P18CeT00`17CUP0ATeG04I=E`0304I=D`0304E<DP14Be40A4]A
00<0@dY@00<0@TU?0458CP11B4h00`11AE001010A4l0@4A=0414C@0n@dT203a1A`0403Tn@P0h?D40
>3e103<g>P80;S8c00@0:R`[02LY900W:B@0:b`T0P0_<B@0100c=2H0=CHV03Df9P0o<bL2044e:@04
048f:00o<B@0?c4T03LZ6P80<b<D00@0<20@034P400a8100?RlL0P1;>RP0101HAc<0Gddi05m=>@1J
B38205A0:00404lk8`1=>B40CCTQ04e2;P80Ac`X00@0A3TW03hc8@0n<b40<BDE0P0Z7Ph0100^8Q@0
=2PJ03@X6P0c9142034R3`0402PI1P0U50809A@202TH1P80:QL600@0?BXI06QC@P1XDd80=2D>0P0`
8@X302dN1`0402hO200a8P/0<B8;03PY4P80@34J00@0Bc`U05I7<01FAc00EDH_0P14=Ah0100i:Q<0
?2dF03`]5P11<Q/2040a6P0403TZ5`0k;1T0>b`I03//6@<0>R/H0P0i:QL0100e9Q<0<24>030Q3P0a
8Pl203HW500303//6@0l;AX0?2dJ00<0?BhK00@0>BXG03<T4@0c9140>RhH0P15>B<0101:?RP0BC`Y
04Tl:@16>BH203ha800403X]700o<R80?c8R05=:>@80DDPi00@0CTLg04]3=P1;@cH0BD<e0P19@cL3
04Q4>00405=BCP1BDDl0DU5?051@D080CTi@00@0C4eA04Y;D01:Be00AdYA0P17BU<304a<EP<0Bd]E
0`1:BU@304U9D`@0BT]@1@19BTl304Q9CQ00BDY?0`18BDh504M8C@0304Q9CP17B4d0AdQ=00H0ATM<
00<0AdQ=04Q9CP18BDh01019BTl304Y;D0D0BdaA301=CU<<04i?E00304e>D`1>Ce@0CTmD00@0CTmD
0P1=CU<304]<D@<0C4eB0`1;C54304Y;D0T0BDY?0`1:Be0304]<D@@0BDY?0`18BDhA04M8C@0404U:
C`18BDh0B4U>04Q9CP80AdQ=0`16Ad`304E6B`@0ATM<1@17B4d304Q9CP`0AdQ=00@0BDY?04Q9CP18
BDh0AdQ=0P16Ad`01015AT/0ATM<04I7C016Ad`>04M8C@0504A5A`15ATP0ADI804I7B@17B4X00@3o
ool00300ATeG3017CUP<04I=EaP0AdiH3017CU@904U=F0<0B4aG0`17CUP01018CeT0BE1J04U@FP19
D5X204Q?F@<0AdiH1P18CeD604M>E0<0A4iH00@0A4iG04A?E@14CeD0@diD0P11C500100oBTh0?TU=
03i9C@0h@DH203@l?`0402lg>@0^=3@0;S@d02le<@80<SLa00@0=3Tb03Hi<00f>C00>cH`0P0m>C00
1010>C40?CH/03df;00i<2L203D[8@0403<Y7`0d:Qh0=2XN03/Z6P80BCPX00@0F4Lg05e:>@1MBST0
F4<b0P1>>BH0101;=280C3HQ04`f8@12=B@203/^7P0403d`800g:QX0=bXJ028E1@806Pd000@08Q@7
02HH2`0V60/0;b0=0P0_7P`0100]70X0@BhM044^7@18=B@2048]7004065<>`2BOF`0TWe/04<d7@80
>b`E00<0=BH?03TZ4`0i:Q<00`11<Q/00`0g:140;Ah702dN1`0303@U3P0402hO200Y6P<0:AX302dN
1`80=bPA00@0>b`E03PY4P0h:A80BCdW0P1<@2X0101;?bT0A3PR04@h8P12=R02044e7`0403P/5P0[
7`T0:al902DI0`80:al900@0<BD?03<W4@0c9a40<bLA0P0g:aD0100m<A/0@CDO044e7`0m<aP2040f
6`0303le700i;aH0>BlF00<0=BXD00@0?34M04Dj9P15>RH0B3lf0P16?C@01013?3@0@S/c048k<`12
?CL204I3?00404]8@`1>BdH0CT]604m?D@<0CdiC0P1=CU<0101=CED0BdiG04]>E`1;CUL204Y>E`<0
C4aD0`1;Be<304Y:DP<0BDUA101:Be0504U:C`<0B4U>4019BTl304Q9CPD0AdQ=00<0B4U>04M8C@17
B4d01P16Ad`00`17B4d0B4U>04Q9CP0304Q9CP<0BDY?101:Be0204]<D@`0C4eB301>Ce@00`1=CU<0
CTmD04i?E00404i?E080CDiC0`1;C54304a=DP<0BdaA0`1:Be0604U:C`<0BT]@101;C54204a=DP@0
BDY?0`18BDhA04M8C@0404Y;D019BTl0BDY?04Q9CP80AdQ=00<0ATM<04E6B`15AT/00`14ADX404I7
C0D0AdQ=0`18BDh<04M8C@0404U:C`18BDh0B4U>04M8C@80ATM<00@0ADI;04I7C016Ad`0ATM<3P17
B4d01@14ADL0ADI804E6B016AdT0AdQ:0040oooo000`04I=E``0AdiH3016CELH04M>F0`0AdiD0`18
C5L904U=F00304M>F018CeT0B4mI00H0BE1J00<0B4mI04M>F017CUP01017CU@504Q?E@<0BE1F0`15
DU/01016De`0AU=K04ICF`15DUX204=@EP04049@D`11Ce80@DmB03i9C@80>dI:00@0>D=503Q2@`0h
@T<0>D910P0k@Cl0100i?cd0=cdi03Lm>@0d=S<303Hf=080=SDa00<0<S4/038_:P0b;bX00`0a;RL0
100e:1l0?c4V03la9P1:?3430592=@80D3h`00@0Ccd_04ln;P1??Rh0>blQ0P0g:ad303da8`0402hR
500[7`l0:al?03HZ6P80?30P00@0<b@A04Tj9`19>RL0CSl/0P1JBCL0101E@S40@2dL040]701:=B@2
050k:P0402TJ1@0j:aH0>R/F04`m:080DD8]00@0BCXU03`]600l;AP0<B8=0P0/7@P0100h:A@0:Q/6
02XK1P0S500202lP2`0404Df8@1HBC@0F4Td06MH@`80LF9=00@0CTHc04Y2;`1:@Rl0?CDR0P0^9Q<0
100Z8Pl0;RHC02hV4`0]9A8302LO3080<BTF00<0>c<P03hf8`0n=R<00`0k<b001011>BH0BT8_04Y2
;`11=ah203/a600403DZ500a9Q80<BHB034V4P80>BhL00@0@cPV04Pm:`18?B/0@S/e0P12?CT01017
@Sh0BdM404]7A01@C4T2051?C@03055@CP1BD540DU1A00<0C4]A00@0C4eB04a<E01<C5@0BdiE0P1:
CUL304]?F0<0C4aD0`1;Be<304Y:DP<0BDUA101:Be0504U:C`<0B4U>4019BTl304Q9CPD0AdQ=00<0
B4U>04M8C@17B4d01P16Ad`00`17B4d0B4U>04Q9CP0304Q9CP<0BDY?101:Be0204]<D@`0C4eB301>
Ce@00`1=CU<0CTmD04i?E00404i?E080CDiC0`1;C54304a=DP<0BdaA0`1:Be0604U:C`<0BT]@101;
C54204a=DP@0BDY?0`18BDhA04M8C@0404Y;D019BTl0BDY?04Q9CP80AdQ=00<0ATM<04E6B`15AT/0
0`14ADX404I7C0D0AdQ=0`18BDh<04M8C@0404U:C`18BDh0B4U>04M8C@80ATM<00@0ADI;04I7C016
Ad`0ATM<3P17B4d01@14ADL0ADI804E6B016AdT0AdQ:0040oooo000`04I=E``0AdiH3016CELH04M>
F0`0AdiD0`18C5L904U=F00304M>F018CeT0B4mI00H0BE1J00<0B4mI04M>F017CUP01017CU@504Q?
E@<0BE1F0`15DU/01016De`0AU=K04ICF`15DUX204=@EP04049@D`11Ce80@DmB03i9C@80>dI:00@0
>D=503Q2@`0h@T<0>D910P0k@Cl0100i?cd0=cdi03Lm>@0d=S<303Hf=080=SDa00<0<S4/038_:P0b
;bX00`0a;RL0100e:1l0?c4V03la9P1:?3430592=@80D3h`00@0Ccd_04ln;P1??Rh0>blQ0P0g:ad3
03da8`0402hR500[7`l0:al?03HZ6P80?30P00@0<b@A04Tj9`19>RL0CSl/0P1JBCL0101E@S40@2dL
040]701:=B@2050k:P0402TJ1@0j:aH0>R/F04`m:080DD8]00@0BCXU03`]600l;AP0<B8=0P0/7@P0
100h:A@0:Q/602XK1P0S500202lP2`0404Df8@1HBC@0F4Td06MH@`80LF9=00@0CTHc04Y2;`1:@Rl0
?CDR0P0^9Q<0100Z8Pl0;RHC02hV4`0]9A8302LO3080<BTF00<0>c<P03hf8`0n=R<00`0k<b001011
>BH0BT8_04Y2;`11=ah203/a600403DZ500a9Q80<BHB034V4P80>BhL00@0@cPV04Pm:`18?B/0@S/e
0P12?CT01017@Sh0BdM404]7A01@C4T2051?C@03055@CP1BD540DU1A00<0C4]A00@0C4eB04a<E01<
C5@0BdiE0P1:CUL304]?F0<0C4aD0`1;Be<304Y:DP<0BDUA101:Be0504U:C`<0B4U>4019BTl304Q9
CPD0AdQ=00<0B4U>04M8C@17B4d01P16Ad`00`17B4d0B4U>04Q9CP0304Q9CP<0BDY?101:Be0204]<
D@`0C4eB301>Ce@00`1=CU<0CTmD04i?E00404i?E080CDiC0`1;C54304a=DP<0BdaA0`1:Be0604U:
C`<0BT]@101;C54204a=DP@0BDY?0`18BDhA04M8C@0404Y;D019BTl0BDY?04Q9CP80AdQ=00<0ATM<
04E6B`15AT/00`14ADX404I7C0D0AdQ=0`18BDh<04M8C@0404U:C`18BDh0B4U>04M8C@80ATM<00@0
ADI;04I7C016Ad`0ATM<3P17B4d01@14ADL0ADI804E6B016AdT0AdQ:0040oooo000`04I=E``0AdiH
3016CELH04M>F0`0BDiD1018C5L504U=F0<0BTiI0`18CeT01019D5X0BU5K04YAF`1:DE/204U@FP<0
B4mI0`16CE<304M>E0<0B4mE0`19D5H01014D5`0AE5M04EAG@15DU/204ICG00404ICF`15DUP0AE9H
04AAE`80@e1F00@0@diB049<CP12C4h0@D]=0P12BTd01013Bdd0A4Y:04A:BP10ATH203i4A0<0>413
00@0>Ci103Lm?@0g?Cd0=CTj0P0c=cH0100d=SD0<cDd03<e=00m>SD2040k=`0404<n>018@C/0B44k
04]4?080Bd8i00@0BT4h04Po=P18?cH0B3hd0P13>Bl01014>Rh0@cT]04<i;@0j<B8203D/6`0403Xa
8010=bH0@3LV04Dj9P80Iea800@0Fe0l04i1;P1>@Bh0ASLT0P14=B80101=?2X0<B0>034P3P1?@c<2
05I:>P0405A8>013=bL0@cLW02lS4`809QX:00@09a/;02XN3P0Z7Ph0;20@0P0^8Q801010=2@0Fdlo
05]??`1UFDT205Q<?00404Pl;013=bL0@cLW03df9P80=c0P00@0;BHF02DN3P0U7Ph09b0@0P0`:AT0
100i<R80?3DU03`e9@0k=2@2044j:P0404Y3<`1=ASH0CDHf04Dn;P80>C8R00@0=2dM03H_7`0f;al0
9ah=0P0S6PT0100X7a00<RTJ038Y6P0l<RH204Y0=P0405=9?`1BB3h0DTPn05=>BP80DTi;00@0DDe:
04e<BP1=C4X0CDa:0P1;Bd/00`1:BTX0BDY<04U:C00404e>D`80BTeB00@0BDiB04Q=D@18CE40ATiA
0P15CE0304a=DP<0BdaA0`1:Be0304U:C`@0BT]@1@19BTl304Q9CQ00BDY?0`18BDh504M8C@0304Q9
CP17B4d0AdQ=00H0ATM<00<0AdQ=04Q9CP18BDh01019BTl304Y;D0D0BdaA301=CU<<04i?E00304e>
D`1>Ce@0CTmD00@0CTmD0P1=CU<304]<D@<0C4eB0`1;C54304Y;D0T0BDY?0`1:Be0304]<D@@0BDY?
0`18BDhA04M8C@0404U:C`18BDh0B4U>04Q9CP80AdQ=0`16Ad`304E6B`@0ATM<1@17B4d304Q9CP`0
AdQ=00@0BDY?04Q9CP18BDh0AdQ=0P16Ad`01015AT/0ATM<04I7C016Ad`>04M8C@0504A5A`15ATP0
ADI804I7B@17B4X00@3oool00300ATeG3017CUP<04I=EaP0AdiH3019CU@00`17BeH0B4aG04Q<E`03
04Q<E`0304U=F01:CUT0BTiI00<0BTiI00<0B4mI04U@FP19D5X01P1:DE/00`19D5X0B4mI04Q?F@06
04M>E0@0B4mE0P19D5H01015Ce/0AU1L04I@G017DEd204MAF`0304QBG017DEX0Ae5J00<0AU5G00@0
B59C04I>D016CU00A4a>0P16Bdh01017C4l0BDe>04U=CP17Bd`204E9BP04045>E010CE<0@4eC041;
D@80@4U>00<0?dQ=03i6B@0nATT00`0mADL303]1@@0403]1?`0n@T40?T910494@@80@dE000@0@d@o
0493?P12@ch0@3`a0P0o>c00100l>2`0@3d^040m;P15@S4204A1;P0404E2;`1=BcH0CD/f05IA>`80
If5;00@0Ad4[03lg8P0o=b80?cLR0P19?RX0101OE400ASTV04Hi9P15?C8203/c:00402XR5`0M5@X0
7AD:020H3@80<2PM00@0?CDZ048j;`12>Rl0@c/`0P11>Bh01015?C80C4@i04a4>@19@CH203Xb9`04
02hV6`0Y8AH0:B4F03</8P<0<R/Q0P0d;B<0100e;R@0=c0V03L`9P0l=B/2048k<@0404a5>`10>Bl0
@3T_03D^9080<R/Q00@0<BXP02dV700]9Q`0:b@J0P0]9Q`0100e:R@0=2TS03@Y8`12=c420515@@04
05E:B01LDDl0G55?05iBDP80Ed];00@0De1;04i=B@1>CDT0BTU50`16AT@204Q:B@0404a>C@1?DE00
Ce5@04m@E@80Ce1D00@0C51C04]?DP1;Ce80BTmB0P17D4l304I?CP<0C4eB0`1;C54304Y;D0<0BDY?
101:Be0504U:C`<0B4U>4019BTl304Q9CPD0AdQ=00<0B4U>04M8C@17B4d01P16Ad`00`17B4d0B4U>
04Q9CP0404Y;D0<0BdaA1@1<CE8<04e>D``0CTmD00<0CDiC04i?E01>Ce@0101>Ce@204e>D`<0BdaA
0`1<CE8304]<D@<0BT]@0`19BTl404Q9CPD0BDY?0`1:Be0404U:C`<0B4U>6017B4d904I7C0D0AdQ=
0`18BDh<04M8C@0404U:C`18BDh0B4U>04M8C@80ATM<00@0ADI;04I7C016Ad`0ATM<3P17B4d01@14
ADL0ADI804E6B016AdT0AdQ:0040oooo000`04I=E``0AdiH3016CELH04M>F0`0BDiD00<0Ad]F04Q<
E`18C5L00`18C5L00`19CEP0BTiI04Y>F@0304Y>F@0304Q?F@19D5X0BE1J00H0BU5K00<0BE1J04Q?
F@18CeT01P17CU@404Q?E@80BE1F00@0ADmK04I@G016D5`0Ae5M0P17DE/00`18DU`0Ae5J04MAFP03
04IAE`0404QBD`16CU00ATi@04A<CP80AT]>00@0Ada?04U=CP19CDh0Ad]<0P15BDX01011CU@0@4eC
041=D`10Be420419CP0303m8C@0nATT0?TI900<0?DE70`0k@D40100k@Cl0?T9103i2@@12A44204=5
@00404=4?`12@ch0@T<n040l<@80?c/`00@0?3P/040m;P10?Bh0AD8a0P14@Bh01015@Rl0CD/f04e;
=P1FDC/206MQB`0404M1:`0o=b80?cLR03lg8P80BChZ00@0GeA004Hi9P16>BH0ACdb0P0k<bP0100Z
8QL07AD:01dE2P0P60d2030X7@0403de:P12>Rl0@SX_04<k<080@CT^00@0ACdb04a4>@1<A3T0BD4f
0P0j<RL0100^9Q/0:B4F02TQ5P0c;283038[8@80=2dS00@0=BhT03L`9P0g<2H0?3D[0P12>c40101<
AC/0@3T_040i;`0e;R@2038[8@04034Z800]9Q`0;BHL02/T6P80;BHL00@0=BXT03@Y8`0d:B<0@SLa
0P1@AD40101EBTP0G55?05aAC`1NDU8205M;B`0405=@B`1>CDT0CTe904Y9A@<0ATI40P18BTT0101<
CTd0Ce5@04mAD01?D5D204m@E00404a@D`1;Ce80BdmB04Y?DP80Ae1?0`16Cdh304a=DP<0BdaA0`1:
Be0304U:C`@0BT]@1@19BTl304Q9CQ00BDY?0`18BDh504M8C@0304Q9CP17B4d0AdQ=00H0ATM<00<0
AdQ=04Q9CP18BDh0101:Be0304]<D@D0C4eB301=CU<<04i?E00304e>D`1>Ce@0CTmD00@0CTmD0P1=
CU<304]<D@<0C4eB0`1;C54304Y;D0<0BDY?1018BDh504U:C`<0BT]@1019BTl304Q9CQP0AdQ=2@16
Ad`504M8C@<0B4U>3017B4d01019BTl0B4U>04Q9CP17B4d204I7C00404E6B`16Ad`0ATM<04I7C0h0
AdQ=00D0A4E704E6B015ATP0ATM904M8BP010?ooo`00<016CEL<04M>F0`0ATeG6017CUP<04U>E0<0
Ad]F0`18C5L00`19CEP0BTiI04Y>F@0304Y>F@0304Q?F@19D5X0BE1J00H0BU5K00<0BE1J04Q?F@18
CeT01`18CeD504U@EP0404E<G016CU/0ATiK04M?G080B51K00@0BE5L04QAFP18DEX0B55J0P17D5T0
101;E5<0Ae1?04M@C`15Bd/204A:BP0304M;C01;C4h0Bda>00<0C4e?0`11CeP01010CeH0@DiF045>
EP14CeD204A?D`<0ADiC0`11Ce001013CU00@De=045=C@10BT/203m8A`0403e6A@0lAD80?4E204]8
?`80BTPl00@0A48f04U7>P19AcX0EeI40P1GET80101DDch0GEa705eLA`1RH4T205aJ@`0403hi8`1A
C3H0DD`f05E?>@80@C/U00@0D4Pc03hf8@0n=R40=C4V0P0f<RL0100d<2D0<RhS038^8`0i=BX2048n
<`0404A0=@10?340@3`a044m<P80?CT^00@0>CDZ03Xf:`0j=R/0@Cdb0P17@cP01018A3T0AT8g04I2
=`0e;RP203</9P0403L`:P0n=c40?SLa03df<080>C8/00@0?CH`04Ho>@16?cT0B44k0P0h<B/0100^
9b40=c0Z03L`:P14?CL204M0>P03044j=00m=S00?CH`00<0DDM600@0GUAE06EKG01UFe`0G55E0P1G
C500101CB4h0B3e304Pm@`18Ad8204U8@`0404a;A`1=CTT0CDi904mACP80DE=@00@0Ce=B051DD`1@
E5<0CTmC0P1<D5<0101<D540Be5A04]AD@1<DU0304YCD080BU=>0`1<CE8304]<D@<0BT]@0`19BTl4
04Y;D0D0BDY?0`18BDh@04U:C`<0B4U>1@17B4d00`18BDh0AdQ=04M8C@0604I7C00404M8C@18BDh0
B4U>04Y;D0D0BdaA0`1<CE8304e>DaP0CTmD00<0CDiC04i?E01>Ce@0101>Ce@204e>D`<0BdaA0`1<
CE8304]<D@<0BT]@0`19BTl404M8C@<0B4U>2@19BTl304Q9CQ40AdQ=0`16Ad`904M8C@@0ATM<1@17
B4d304Q9CP`0AdQ=00@0BDY?04Q9CP18BDh0AdQ=0P16Ad`01015AT/0ATM<04I7C016Ad`>04M8C@05
04A5A`15ATP0ADI804I7B@17B4X00@3oool001P0ADaF9017CUPH04I=E``0AdiH1018C5L504U=F0H0
BTiI1019CEP504Q<E`<0ATeG00@0AdiH04Q?F@18CeT0B4mI0P17CUP304I=E`<0AdiH1018CeT204M>
F00304E<EP14BeD0A4]E00`0ADmH3015Ce40100kD5D0?55F03aAEP0mDUL303iCF080?E9G00<0?55F
03]@E@0kD5D00`11D5L01012DEP0@e9I04=BF@13DUT2049AF0<0@E1G00@0CEA<04aCB`1<Dd/0BE18
0P17CTH01014Bd<0@DQ00458@00oASh203i5?@0403e0>@13ASl0@dHo04Y=AP80DE9:00@0FUY@05QH
CP1HF4h0BDLk0P0f=2P0100l?CL0?Chh03dn>011@S`204I7@@0404]<AP1@DD/0D55;05=DCP80EEI@
00@0ADE504Y:BP1:BTX0DE5A0P1DE5@0101ADE40B4Q804Q8B00l?3`203De=@0403Ph=P0n?S`0?Shl
04E5@`80AdM50`15AD<0101:BTP0D51>051@CP1:BTP204U9A`0404Q8AP1;BdT0Bd]904m?C@<0DE5?
0P1?Cdd<04a;D0`0Be1D00<0BTmC04]@E01;D5@01P1<DED00`1;D5@0BTmC04Y?D`0304e>D`@0C4eB
1@1;C54<04U:C`@0BT]=1@19BT`304Q9B``0B4U>3017B4d604]<D@<0C4eB0`1=CU<H04i?E0`0CDiC
0`1<CE8304]<D@<0BT]@3`19BTl<04Q9CP`0ATM<3017B4d304A5BP<0ADI;0`16Ad`304M8C@<0B4U>
00<0AdQ=04I7C016Ad`00`16Ad`304M8C@0404U:C018BD/0B4U;04M8BP80ATM900@0ADI804I7B@16
AdT0ATM90P17B4X<04M8C@0504A5A`15ATP0ADI804I7B@17B4X00@3oool001P0ADaF9017CUPH04I=
E``0AdiH1018C5L504U=F0H0BTiI1019CEP504Q<E`<0ATeG00@0AdiH04Q?F@18CeT0B4mI0P17CUP3
04I=E`<0AdiH1018CeT204M>F00304E<EP14BeD0A4]E00`0ADmH3015Ce40100kD5D0?55F03aAEP0m
DUL303iCF080?E9G00<0?55F03]@E@0kD5D00`11D5L01012DEP0@e9I04=BF@13DUT2049AF0<0@E1G
00@0CEA<04aCB`1<Dd/0BE180P17CTH01014Bd<0@DQ00458@00oASh203i5?@0403e0>@13ASl0@dHo
04Y=AP80DE9:00@0FUY@05QHCP1HF4h0BDLk0P0f=2P0100l?CL0?Chh03dn>011@S`204I7@@0404]<
AP1@DD/0D55;05=DCP80EEI@00@0ADE504Y:BP1:BTX0DE5A0P1DE5@0101ADE40B4Q804Q8B00l?3`2
03De=@0403Ph=P0n?S`0?Shl04E5@`80AdM50`15AD<0101:BTP0D51>051@CP1:BTP204U9A`0404Q8
AP1;BdT0Bd]904m?C@<0DE5?0P1?Cdd<04a;D0`0Be1D00<0BTmC04]@E01;D5@01P1<DED00`1;D5@0
BTmC04Y?D`0304e>D`@0C4eB1@1;C54<04U:C`@0BT]=1@19BT`304Q9B``0B4U>3017B4d604]<D@<0
C4eB0`1=CU<H04i?E0`0CDiC0`1<CE8304]<D@<0BT]@3`19BTl<04Q9CP`0ATM<3017B4d304A5BP<0
ADI;0`16Ad`304M8C@<0B4U>00<0AdQ=04I7C016Ad`00`16Ad`304M8C@0404U:C018BD/0B4U;04M8
BP80ATM900@0ADI804I7B@16AdT0ATM90P17B4X<04M8C@0504A5A`15ATP0ADI804I7B@17B4X00@3o
ool001P0ADaF9017CUPH04I=E``0AdiH1018C5L504U=F0H0BTiI1019CEP504Q<E`<0ATeG00@0AdiH
04Q?F@18CeT0B4mI0P17CUP304I=E`<0AdiH1018CeT204M>F00304E<EP14BeD0A4]E00`0ADiG3015
Ce4303i@E00403mAE@10DUH0@59F041BEP80?e5E0`0nD5@01014CeD0AE1F04E@EP16DEL304MBF080
AU5G00@0AE1F04A?E@14CeD0BU1<0P19Cd/01018CTX0Ade904M=B@15BdL204A:AP0404=9A@12B4@0
@TQ40497@`80B4Y700@0Bde:051AC01@DD`0EEIA0P1FEE00101<BdH0?Shf03hn=P1:BTP204];B@04
04a<BP1>CT`0CTi<04m?C@80DE5?0`1BDU00101=CDd0CTi>04i>CP1>CTh204e=C@0404];B`19BDT0
BDU904I6AP80ADE500@0A4A404Q8B018B4P0CDe=0P1>CTh304a<C00404i>CP1ADE40DE5A04];B`80
BTY:00@0BDU904];B`1;Bd/0CTi>0`1@D50204m?C``0C4]A3@1:Ce<204]@E0H0C55E00<0Be1D04Y?
D`1:Ce<00`1=CU<404a=DPD0BdaA3019BTl404Y;C@D0BDY<0`18BD/<04Q9CP`0AdQ=00<0BT]@04]<
D@1;C5400`1;C54304a=DP<0CDiC601>Ce@<04e>D`<0C4eB0`1;C54304Y;D0l0BDY?3018BDh<04I7
C0`0AdQ=0`14ADX304E6B`<0ATM<0`17B4d00`18BDh0AdQ=04M8C@0604I7C00404M8C@18BDh0B4U>
04U:C080B4U;00@0AdQ:04I7B@16AdT0ADI80`16AdT204M8BP`0AdQ=00D0A4E704E6B015ATP0ATM9
04M8BP010?ooo`006015C5H`04I=EaP0AdiH1018C5L504U=F0H0BTiI1019CEP504Q<E`<0ATeG00@0
AdiH04Q?F@18CeT0B4mI0P17CUP304I=E`<0AdiH1018CeT204M>F00304E<EP14BeD0A4]E00`0A4eF
3015Ce400`13CU@0A4mE04A?E@0604E@EP0404A?E@13CU@0@diD04M?DP80B51C00@0BE5D04YBE@1:
DUD0BU9E0P19DE@00`18D5<0AdmB04M?DP0404Q=D0P0Ada?00@0BDiB04]?DP1;Ce80BdmB0P1=CU00
101?D580DU9B059BDP1?Cdl204Y9A`<0DU5F0`1AD5D6051?E004055AD`1?Ce40CdmA04e=C`<0Bd]=
0P1<C4h0101?Ce40D51B051@DP1>CE<204m>E004051?E@1AD5H0DE1F055@EP<0D4mE0P1?CU@0101<
Be40BdY@04]:D01;BU0204a;D@0404i=D`1?CU@0CdiD04i=D`80CDaB301<BUD=04Y?D`80Be1D1P1<
DED00`1;D5@0BTmC04Y?D`0304e>D`@0C4eB1@1;C54<04U:C`@0BT]=1@19BT`304Q9B``0B4U>3017
B4d404Y;D0<0BdaA1@1<CE8H04i?E0`0CDiC0`1<CE8304]<D@<0BT]@3`19BTl<04Q9CQP0AdQ=0`14
ADX304E6B`<0ATM<0`17B4d00`18BDh0AdQ=04M8C@0604I7C00404M8C@18BDh0B4U>04U:C080B4U;
00@0AdQ:04I7B@16AdT0ADI80`16AdT204M8BP`0AdQ=00D0A4E704E6B015ATP0ATM904M8BP010?oo
o`006015C5H`04I=EaP0AdiH1018C5L504U=F0H0BTiI1019CEP504Q<E`<0ATeG00@0AdiH04Q?F@18
CeT0B4mI0P17CUP304I=E`<0AdiH1018CeT204M>F00304E<EP14BeD0A4]E00`0A4eF3015Ce400`13
CU@0A4mE04A?E@0604E@EP0404A?E@13CU@0@diD04M?DP80B51C00@0BE5D04YBE@1:DUD0BU9E0P19
DE@00`18D5<0AdmB04M?DP0404Q=D0P0Ada?00@0BDiB04]?DP1;Ce80BdmB0P1=CU00101?D580DU9B
059BDP1?Cdl204Y9A`<0DU5F0`1AD5D6051?E004055AD`1?Ce40CdmA04e=C`<0Bd]=0P1<C4h0101?
Ce40D51B051@DP1>CE<204m>E004051?E@1AD5H0DE1F055@EP<0D4mE0P1?CU@0101<Be40BdY@04]:
D01;BU0204a;D@0404i=D`1?CU@0CdiD04i=D`80CDaB301<BUD=04Y?D`80Be1D1P1<DED00`1;D5@0
BTmC04Y?D`0304e>D`@0C4eB1@1;C54<04U:C`@0BT]=1@19BT`304Q9B``0B4U>3017B4d404Y;D0<0
BdaA1@1<CE8H04i?E0`0CDiC0`1<CE8304]<D@<0BT]@3`19BTl<04Q9CQP0AdQ=0`14ADX304E6B`<0
ATM<0`17B4d00`18BDh0AdQ=04M8C@0604I7C00404M8C@18BDh0B4U>04U:C080B4U;00@0AdQ:04I7
B@16AdT0ADI80`16AdT204M8BP`0AdQ=00D0A4E704E6B015ATP0ATM904M8BP010?ooo`006015C5H`
04I=E``0B4mI3017CUP404Q<E`D0BDeH1P1:CUT404U=F0D0B4aG0`16CEL01017CUP0B4mI04Q?F@18
CeT204M>F0<0ATeG0`17CUP404Q?F@80AdiH00<0ADaF04A;E@14BeD03014BeD<04I?E00404E<DP16
CE<0ATeC04M>E0<0B4mE0P17CU@01016CE<0ADaB04E<DP1;Ce8204a@D`H0CE5D00<0C51C04]?DP1;
Ce803016CED01019D5X0BdmH04]?F01;CeP204]>E@0304Y=E01=CU<0CDiC00<0CTmD0`1;Be<404a<
E0D0CDeE00@0Bda@04a=D@1<CE40CDiB0P1>Ce<0101?D5@0CTmC04i?D`1>Ce<204e>DP0404i>F01=
CEL0CDeG04a<EP80CDeG0`1?CeT00`1<C5H0BDUC04U9D`0304a<EP0304];E@1<C5H0C4aF00@0C4aF
0P1;BeD<04a<F0`0BDiB00<0BTmC04]@E01;D5@01P1<DED00`1;D5@0BTmC04Y?D`0304e>D`@0C4eB
1@1;C54<04U:C`@0BT]=1@19BT`304Q9BaP0B4U>101:Be0504]<D@<0C4eB301=CU<<04i?E0`0CDiC
0`1<CE8304]<D@<0BT]@3`19BTlH04Q9CP`0AdQ=0`14ADX304E6B`<0ATM<1017B4d304I7C080ADI;
00@0ATM<04M8C@17B4d0B4U>0P19BTl01019BT`0B4U;04Q9B`17B4X204I7B@0404E6B016AdT0ATM9
04I7B@80AdQ:3017B4d01@14ADL0ADI804E6B016AdT0AdQ:0040oooo000H04A;EB@0ADaF3016CEL<
04Q?F@`0AdiH1018C5L504U=F0H0BTiI1019CEP504Q<E`<0ATeG00@0AdiH04Q?F@18CeT0B4mI0P17
CUP304I=E`<0AdiH1018CeT204M>F00304E<EP14BeD0A4]E00`0A4]E3016Ce@01018CE<0BDiD04U>
E019CU@304Y?E@<0BDiD0P18CE<00`1;Ce80C51C04a@D`0604eAE00404a@D`1;Ce80BdmB04U?F`<0
B4iJ0`17CET204I<F0<0AD]G00@0ATiK04I>F@16CUT0B4iJ0P18CeT01019CEP0BDaE04U<E@1:CEH2
04a?EP0404I8E017BED0AdUE04M9E@80B4YF0`19BeL304Y<F00404Q;D019C540BDaA04a?E0<0CE1E
0P1;CU<01019C540B4]@04Q;D01;CUL204U<E@0304M:D`19C5D0BDaE00<0CE1I00<0BTeF04I9DP16
BE800`1:CEH00`1;CUL0BTeF04Y=EP0604U<E@`0BTaH3019CU800`1:Ce<0Be1D04]@E00604aAE@03
04]@E01:Ce<0BTmC00<0CDiC101<CE8504]<D@`0BDY?101:Bdd504U:C0<0B4U;6018BDh304U:C`@0
BT]@1@1;C54<04e>D``0CTmD301=CU<304a=DP<0BdaA0`1:Be0?04U:CaP0B4U>3017B4d304A5BP<0
ADI;0`16Ad`304M8C@<0ATM<0`15AT/00`16Ad`0AdQ=04M8C@0304U:C`0404U:C018BD/0B4U;04M8
BP80ATM900@0ADI804I7B@16AdT0ATM90P17B4X<04M8C@0504A5A`15ATP0ADI804I7B@17B4X00@3o
ool001P0A4]E9015C5H<04I=E``0B4mI3017CUP404Q<E`D0BDeH1P1:CUT404U=F0D0B4aG0`16CEL0
1017CUP0B4mI04Q?F@18CeT204M>F0<0ATeG0`17CUP404Q?F@80AdiH00<0ADaF04A;E@14BeD03014
BeD<04I?E00404Q=D`19CU@0BDiD04U>E0<0BTmE0`19CU@204Q=D`0304]?DP1<D5<0C51C00H0CE5D
00@0C51C04]?DP1;Ce80BDmK0`18CUX304M=F@80ATaH0`15BeL01016CU/0ATiI04I>F@18CUX204Q?
F@0404U=F019C5D0BDaE04Y=EP80C4mF00@0ATQD04M9E@17BED0AdUE0P18BUH304U;E`<0BTaH00@0
B4]@04U<D@19C540C4mD0`1=D5D204]>D`0404U<D@18Be00B4]@04]>E`80BDaE00<0AdYC04U<E@19
C5D00`1=D5T00`1:CEH0ATUB04I9DP0304Y=EP0304]>E`1:CEH0BTeF00H0BDaE301:C5P<04U>DP03
04Y?D`1;D5@0Be1D00H0C55E00<0Be1D04Y?D`1:Ce<00`1=CU<404a=DPD0BdaA3019BTl404Y;C@D0
BDY<0`18BD/H04Q9CP<0BDY?101:Be0504]<D@`0CDiC301>Ce@<04e>D`<0C4eB0`1;C54304Y;D0l0
BDY?6018BDh<04M8C@<0A4E:0`15AT/304I7C0<0AdQ=0`16Ad`304E6B`0304I7C017B4d0AdQ=00<0
BDY?00@0BDY<04Q9B`18BD/0AdQ:0P16AdT01015ATP0ATM904I7B@16AdT204M8BP`0AdQ=00D0A4E7
04E6B015ATP0ATM904M8BP010?ooo`006014BeDT04E<EP`0ATeG6017CUP404Q<E`D0BDeH1P1:CUT4
04U=F0D0B4aG0`16CEL01017CUP0B4mI04Q?F@18CeT204M>F0<0ATeG0`17CUP404Q?F@80AdiH00<0
ADaF04A;E@14BeD03017BeH<04I?EP0404Q=D@19CU80BDiB04Y?D`<0Be1D0P1:Ce<01019CU80B4eA
04Q=D@1:CE8204]>D`0404a?E01=D5D0CE1E04e@E@80C4mD00@0BdiC04Y=DP1:CE80BE5N0P18D5d0
1017Ce`0ATiK04I>F`15CEX204A<F@<0@d]H00@0@deI049<F012C5P0ADeH0P18D5/01019D5X0ADaD
04E<E016BU<204Q<E@@0BdiG201:CEH0101<CeP0BdiG04]>E`1:CEH304Q;E080BDaE00@0BTeF04]>
E`1;CUL0C4mH0`1:CEH204a?F0<0Ce9K00@0CE1I04Y=EP1:CEH0BDaE0P1:CEH304]>E`0404U<E@18
Be@0B4]D04U<E@h0BTeF3@1:Ce<204]@E0H0C55E00<0Be1D04Y?D`1:Ce<00`1=CU<404a=DPD0BdaA
3019BTl404Y;C@D0BDY<0`18BD/<04Q9CQ00BDY?0`1:Be0504]<D@`0C4eB301>Ce@<04e>D`<0C4eB
0`1;C54304Y;D0l0BDY?3018BDhH04M8C@<0A4E:0`15AT/304I7C0<0AdQ=00<0ATM<04E6B`15AT/0
0`15AT/01016Ad`0B4U>04Q9CP19BTl204Y;D00404U:C018BD/0B4U;04M8BP80ATM900@0ADI804I7
B@16AdT0ATM90P17B4X<04M8C@0504A5A`15ATP0ADI804I7B@17B4X00@3oool001P0@dYD9014BeDH
04I=E``0AdiH1018C5L504U=F0H0BTiI1019CEP504Q<E`<0ATeG00@0AdiH04Q?F@18CeT0B4mI0P17
CUP304I=E`<0AdiH1018CeT204M>F00304E<EP14BeD0A4]E00`0B4aG3017D5L304M?DP0404Q@D`19
DE@0BE5D04UAE080B51C0`17Ce801016CED0AdiF04M>EP18CeL304U@F080B4mG00<0AdiF04I=E@16
CED00`17CeX01016CUT0ADeH04E=F015CEP204A<E`<0@d]F00@0A55J041=EP10CEH0@deG0P18DU/0
101:DeX0ATmF04I?EP15C58204I=D`L0BTmE1@19CU@0101<D5T0BdmH04]?F018C5D304M;E080BDeF
00<0C51I04eAFP1=DEX0101;D5@204aAE@0404eBEP1<DED0C55E04Y?D`80BDiB00<0Ada@04U>DP19
CU800`1:Ce<01018CE40Ada@04M<D019CU8204]@E1T0BTmC0P1;D5@604aAE@0304]@E01:Ce<0BTmC
00<0CDiC101<CE8504]<D@`0BDY?101:Bdd504U:C0<0B4U;3018BDh<04U:C`<0B4U>0`19BTl404Y;
D080BdaA301<CE8<04i?E0`0CDiC0`1<CE8304]<D@<0BT]@3`19BTl<04Q9CP`0ATM<3017B4d304A5
BP<0ADI;0`16Ad`304M8C@H0ADI;00@0ATM<04Q9CP18BDh0BDY?0P1;C5401019BT`0B4U;04Q9B`17
B4X204I7B@0404E6B016AdT0ATM904I7B@80AdQ:3017B4d01@14ADL0ADI804E6B016AdT0AdQ:0040
oooo000H04=:E2@0A4]E6016CEL<04M>F0@0B4aG1@19CEP604Y>F@@0BDeH1@18C5L304I=E`0404M>
F018CeT0B4mI04Q?F@80AdiH0`16CEL304M>F0@0B4mI0P17CUP00`15C5H0A4]E04A;E@0<04Q<E``0
Ae1G0`17Ce801018D5<0BE5D04UAE019DE@204Q@D`<0AdmB00@0ATeE04M>EP17CUH0B4mG0`19D5P2
04Q?E`0304M>EP16CED0ATeE00<0AdmJ00@0ATiI04E=F015CEP0ADeH0P14C5L304=;EP0404AAFP10
CEH0@4eF04==E`80B59K00@0BU=J04I?EP16CeH0ADaB0P16CE<704Y?E@D0BDiD00@0C51I04]?F01;
CeP0B4aE0`17Be@204U=EP0304a@F@1=DEX0CE5J00@0Be1D0P1<DED0101=DUH0C55E04aAE@1:Ce<2
04U>DP0304M<D019CU80BDiB00<0BTmC00@0B4eA04M<D017C500BDiB0P1;D5@I04Y?D`80Be1D1P1<
DED00`1;D5@0BTmC04Y?D`0304e>D`@0C4eB1@1;C54<04U:C`@0BT]=1@19BT`304Q9B``0B4U>3019
BTl304Q9CP<0BDY?101:Be0204]<D@`0C4eB301>Ce@<04e>D`<0C4eB0`1;C54304Y;D0l0BDY?3018
BDh<04I7C0`0AdQ=0`14ADX304E6B`<0ATM<0`17B4d604E6B`0404I7C018BDh0B4U>04U:C`80BdaA
00@0BDY<04Q9B`18BD/0AdQ:0P16AdT01015ATP0ATM904I7B@16AdT204M8BP`0AdQ=00D0A4E704E6
B015ATP0ATM904M8BP010?ooo`006013BU@T04A;EAP0ATeG3017CUP404Q<E`D0BDeH1P1:CUT404U=
F0D0B4aG0`16CEL01017CUP0B4mI04Q?F@18CeT204M>F0<0ATeG0`17CUP404Q?F@80AdiH00<0ADaF
04A;E@14BeD03019CEP<04M@F@<0AU1B00@0Ae5C04QBE018DU@0B59D0P17DE<304I@DP<0ADiG00@0
ATmH04M@F@17D5T0Ae1I0P16CeP304E>E`H0A4iH1P13CEL01018EEh0@TmG049?E`10CED204QCF@04
04YEF`17D5D0Ae1E04E>D`80B4mE0`15BU001016Be40AdaB04M<DP18CE<304U>E080BTmE0`19CEP0
101:CUT0BdmJ04]?FP1<D5/504eAG0<0BTmB00@0Be1C04Y?DP1:Ce80BDiA0P17C4l01016Bdh0ADY=
04E:C@16Bdh204Q=D0<0BTmB00@0B4e@04M<C`17C4l0BDiA0P1<DE@<04Y?DP`0Be1D00<0BTmC04]@
E01;D5@01P1<DED00`1;D5@0BTmC04Y?D`0304e>D`@0C4eB1@1;C54<04U:C`@0BT]=1@19BT`304Q9
B``0B4U>3019BTl304Q9CP<0BDY?1P1:Be0<04a=DP`0CTmD301=CU<304a=DP<0BdaA0`1:Be0?04U:
C``0B4U>3016Ad`<04M8C@<0A4E:0`15AT/304I7C0<0AdQ=1P15AT/01016Ad`0B4U>04Q9CP1:Be02
04]<D@0404U:C018BD/0B4U;04M8BP80ATM900@0ADI804I7B@16AdT0ATM90P17B4X<04M8C@0504A5
A`15ATP0ADI804I7B@17B4X00@3oool00001\
\>"],
  ImageRangeCache->{{{0, 521}, {494, 0}} -> {-0.00173502, 0.990273, 0.999052, \
0.999052}}]
}, Closed]],

Cell["Finally,", "Text"],

Cell[BoxData[
    \(\(Export["\<jjk.eps\>", me];\)\)], "Input"],

Cell["\<\
converts a JPEG file to encapsulated postscript; again, verify \
using a suitable viewer.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell[CellGroupData[{

Cell["World population history and projections", "ExerciseTitle"],

Cell[TextData[{
  "The file ",
  StyleBox["worldhis.txt",
    FontSlant->"Italic"],
  " contains estimates of the world population between about 10000 BCE and \
1950 CE while ",
  StyleBox["worldpop.txt",
    FontSlant->"Italic"],
  " contains data for 1950-1990 and projections through 2050.  These files \
were downloaded from the web site of the U.S. Census Bureau.  Extract data \
from the files and plot the history of world population.  For the historical \
data it is sufficient to use the column that gives a summary of lower bounds, \
for which the data are most complete.  Use ",
  StyleBox["MultipleListPlot", "InlineInput",
    FontWeight->"Bold"],
  " to plot the two data with different symbols and/or colors."
}], "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
ScreenStyleEnvironment->"Working",
WindowToolbars->"EditBar",
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrintingCopies->1,
PrintingPageRange->{7, 7},
PrintingOptions->{"PrintingMargins"->{{54, 54}, {72, 72}},
"PrintCellBrackets"->False,
"PrintRegistrationMarks"->False,
"PrintMultipleHorizontalPages"->False},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, -1}},
ShowSelection->True,
ShowCellLabel->False,
ShowCellTags->False,
DefaultDuplicateCellStyle->"Text",
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of \
all cells in a given style.  Make modifications to any definition using \
commands in the Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the \
Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output \
to the kernel.  Be careful when modifying, renaming, or removing these \
styles, because the front end associates special meanings with these style \
names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, \
etc.  They automatically give the cell a FrameLabel containing a reference to \
a particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext \
ButtonBoxes.  The \"Hyperlink\" style is for links within the same Notebook, \
or between Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line \
help system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Info3350915520-3336292"->{
    Cell[46076, 1292, 564, 8, 185, "Print",
      CellTags->"Info3350915520-3336292"]},
  "Info3350830501-3442347"->{
    Cell[229442, 5976, 67, 1, 34, "Print",
      CellTags->"Info3350830501-3442347"],
    Cell[229512, 5979, 573, 16, 37, "Print",
      CellTags->"Info3350830501-3442347"]},
  "Info3350830503-7920564"->{
    Cell[230165, 6003, 64, 1, 34, "Print",
      CellTags->"Info3350830503-7920564"],
    Cell[230232, 6006, 455, 13, 35, "Print",
      CellTags->"Info3350830503-7920564"]},
  "Info3350830504-7344348"->{
    Cell[230766, 6027, 63, 1, 34, "Print",
      CellTags->"Info3350830504-7344348"],
    Cell[230832, 6030, 470, 13, 35, "Print",
      CellTags->"Info3350830504-7344348"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Info3350915520-3336292", 2107552, 32597},
  {"Info3350830501-3442347", 2107670, 32600},
  {"Info3350830503-7920564", 2107875, 32605},
  {"Info3350830504-7344348", 2108080, 32610}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 125, 4, 81, "Title",
  Evaluatable->False],
Cell[1904, 59, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2384, 78, 33, 0, 83, "Section"],
Cell[2420, 80, 117, 2, 99, "Input"],
Cell[2540, 84, 55, 1, 73, "Input"],
Cell[2598, 87, 200, 5, 44, "Text"],
Cell[2801, 94, 255, 4, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[3093, 103, 82, 2, 51, "Section",
  Evaluatable->False],
Cell[3178, 107, 430, 9, 67, "Text"],

Cell[CellGroupData[{
Cell[3633, 120, 33, 0, 68, "Subsection"],
Cell[3669, 122, 153, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[3847, 129, 55, 1, 72, "Input"],
Cell[3905, 132, 11761, 295, 198, "Print"]
}, Closed]],
Cell[15681, 430, 232, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[15938, 439, 44, 1, 73, "Input"],
Cell[15985, 442, 89, 1, 59, "Output"]
}, Closed]],
Cell[16089, 446, 542, 11, 145, "Text"],

Cell[CellGroupData[{
Cell[16656, 461, 67, 1, 73, "Input"],
Cell[16726, 464, 116, 2, 59, "Output"]
}, Closed]],
Cell[16857, 469, 220, 4, 94, "Caution"],
Cell[17080, 475, 414, 10, 68, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[17531, 490, 27, 0, 68, "Subsection"],
Cell[17561, 492, 142, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[17728, 499, 50, 1, 72, "Input"],
Cell[17781, 502, 10364, 256, 196, "Print"]
}, Closed]],
Cell[28160, 761, 221, 7, 41, "Text",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[28406, 772, 44, 1, 73, "Input"],
Cell[28453, 775, 214, 3, 84, "Output"]
}, Closed]],
Cell[28682, 781, 182, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[28889, 789, 54, 1, 73, "Input"],
Cell[28946, 792, 67, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[29050, 798, 52, 1, 73, "Input"],
Cell[29105, 801, 71, 1, 59, "Output"]
}, Closed]],
Cell[29191, 805, 545, 15, 68, "Text"],

Cell[CellGroupData[{
Cell[29761, 824, 132, 2, 73, "Input"],
Cell[29896, 828, 2817, 51, 678, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[32750, 884, 65, 1, 73, "Input"],
Cell[32818, 887, 67, 1, 59, "Output"]
}, Closed]],
Cell[32900, 891, 68, 0, 58, "Exercise"],
Cell[32971, 893, 514, 14, 68, "ExerciseText"],
Cell[33488, 909, 62, 1, 73, "Input"],
Cell[33553, 912, 159, 4, 41, "ExerciseText"],

Cell[CellGroupData[{
Cell[33737, 920, 54, 1, 73, "Input"],
Cell[33794, 923, 96, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[33927, 930, 44, 1, 73, "Input"],
Cell[33974, 933, 36, 1, 59, "Output"]
}, Closed]],
Cell[34025, 937, 644, 22, 42, "ExerciseText"],

Cell[CellGroupData[{
Cell[34694, 963, 119, 3, 73, "Input"],
Cell[34816, 968, 43, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[34896, 974, 44, 1, 73, "Input"],
Cell[34943, 977, 45, 1, 59, "Output"]
}, Closed]],
Cell[35003, 981, 474, 15, 42, "ExerciseText"],

Cell[CellGroupData[{
Cell[35502, 1000, 75, 1, 73, "Input"],
Cell[35580, 1003, 43, 1, 59, "Output"]
}, Closed]],
Cell[35638, 1007, 353, 10, 42, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[36028, 1022, 27, 0, 68, "Subsection"],
Cell[36058, 1024, 537, 11, 94, "Text"],

Cell[CellGroupData[{
Cell[36620, 1039, 38, 1, 73, "Input"],
Cell[36661, 1042, 1405, 30, 409, "Output"]
}, Closed]],
Cell[38081, 1075, 375, 7, 68, "Text"],

Cell[CellGroupData[{
Cell[38481, 1086, 139, 2, 73, "Input"],
Cell[38623, 1090, 108, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[38768, 1097, 57, 1, 73, "Input"],
Cell[38828, 1100, 1476, 32, 434, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[40341, 1137, 45, 1, 73, "Input"],
Cell[40389, 1140, 667, 10, 159, "Output"]
}, Closed]],
Cell[41071, 1153, 477, 14, 69, "Text"],
Cell[41551, 1169, 55, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[41631, 1174, 45, 1, 73, "Input"],
Cell[41679, 1177, 680, 10, 184, "Output"]
}, Closed]],
Cell[42374, 1190, 160, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[42559, 1198, 48, 1, 73, "Input"],
Cell[42610, 1201, 3463, 89, 106, "Print"],
Cell[46076, 1292, 564, 8, 185, "Print",
  CellTags->"Info3350915520-3336292"]
}, Closed]],

Cell[CellGroupData[{
Cell[46677, 1305, 199, 3, 105, "Input"],
Cell[46879, 1310, 77938, 2987, 387, 26131, 2343, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[124832, 4300, 66, 0, 58, "Exercise"],
Cell[124901, 4302, 96, 3, 41, "ExerciseText"],
Cell[125000, 4307, 442, 12, 69, "ExerciseText"],
Cell[125445, 4321, 78, 0, 41, "ExerciseText"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[125572, 4327, 164, 7, 51, "Section",
  Evaluatable->False],

Cell[CellGroupData[{
Cell[125761, 4338, 34, 0, 68, "Subsection"],
Cell[125798, 4340, 1216, 38, 96, "Text"],
Cell[127017, 4380, 151, 2, 71, "Caution"],
Cell[127171, 4384, 552, 17, 69, "Text"],
Cell[127726, 4403, 57, 0, 58, "Exercise"],
Cell[127786, 4405, 86, 1, 41, "ExerciseText"],
Cell[127875, 4408, 148, 3, 74, "Input"],
Cell[128026, 4413, 253, 6, 177, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[128316, 4424, 34, 0, 68, "Subsection"],
Cell[128353, 4426, 862, 26, 159, "Text"],
Cell[129218, 4454, 61, 0, 58, "Exercise"],
Cell[129282, 4456, 711, 20, 138, "ExerciseText"],

Cell[CellGroupData[{
Cell[130018, 4480, 52, 1, 73, "Input"],
Cell[130073, 4483, 84188, 1107, 279, 10869, 175, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[214310, 5596, 89, 2, 68, "Subsection",
  Evaluatable->False],
Cell[214402, 5600, 2059, 62, 276, "Text"],
Cell[216464, 5664, 65, 0, 58, "Exercise"],
Cell[216532, 5666, 300, 5, 67, "ExerciseText"],
Cell[216835, 5673, 381, 5, 67, "ExerciseText"],
Cell[217219, 5680, 134, 3, 73, "Input"],
Cell[217356, 5685, 253, 4, 151, "Input"],
Cell[217612, 5691, 364, 8, 67, "ExerciseText"],

Cell[CellGroupData[{
Cell[218001, 5703, 137, 3, 96, "Input"],
Cell[218141, 5708, 439, 7, 37, "Message"],
Cell[218583, 5717, 439, 7, 37, "Message"],
Cell[219025, 5726, 439, 7, 37, "Message"],
Cell[219467, 5735, 329, 5, 35, "Message"]
}, Closed]],

Cell[CellGroupData[{
Cell[219833, 5745, 135, 3, 96, "Input"],
Cell[219971, 5750, 439, 7, 37, "Message"],
Cell[220413, 5759, 439, 7, 37, "Message"],
Cell[220855, 5768, 439, 7, 37, "Message"],
Cell[221297, 5777, 329, 5, 35, "Message"]
}, Closed]],

Cell[CellGroupData[{
Cell[221663, 5787, 120, 2, 96, "Input"],
Cell[221786, 5791, 439, 7, 37, "Message"],
Cell[222228, 5800, 439, 7, 37, "Message"],
Cell[222670, 5809, 439, 7, 37, "Message"],
Cell[223112, 5818, 329, 5, 35, "Message"]
}, Closed]],
Cell[223456, 5826, 113, 2, 73, "Input"],
Cell[223572, 5830, 71, 0, 41, "ExerciseText"],
Cell[223646, 5832, 131, 2, 73, "Input"],
Cell[223780, 5836, 1012, 26, 122, "ExerciseText"],
Cell[224795, 5864, 59, 1, 73, "Input"],
Cell[224857, 5867, 111, 2, 73, "Input"],
Cell[224971, 5871, 83, 1, 41, "ExerciseText"],

Cell[CellGroupData[{
Cell[225079, 5876, 44, 1, 73, "Input"],
Cell[225126, 5879, 3534, 63, 1376, "Print"]
}, Closed]],
Cell[228675, 5945, 130, 3, 41, "ExerciseText"],
Cell[228808, 5950, 67, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[228900, 5955, 42, 1, 73, "Input"],
Cell[228945, 5958, 298, 5, 35, "Message"]
}, Closed]],
Cell[229258, 5966, 59, 0, 41, "ExerciseText"],
Cell[229320, 5968, 51, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[229396, 5973, 43, 1, 73, "Input"],
Cell[229442, 5976, 67, 1, 34, "Print",
  CellTags->"Info3350830501-3442347"],
Cell[229512, 5979, 573, 16, 37, "Print",
  CellTags->"Info3350830501-3442347"]
}, Closed]],

Cell[CellGroupData[{
Cell[230122, 6000, 40, 1, 73, "Input"],
Cell[230165, 6003, 64, 1, 34, "Print",
  CellTags->"Info3350830503-7920564"],
Cell[230232, 6006, 455, 13, 35, "Print",
  CellTags->"Info3350830503-7920564"]
}, Closed]],

Cell[CellGroupData[{
Cell[230724, 6024, 39, 1, 73, "Input"],
Cell[230766, 6027, 63, 1, 34, "Print",
  CellTags->"Info3350830504-7344348"],
Cell[230832, 6030, 470, 13, 35, "Print",
  CellTags->"Info3350830504-7344348"]
}, Closed]],
Cell[231317, 6046, 178, 4, 41, "ExerciseText"],

Cell[CellGroupData[{
Cell[231520, 6054, 399, 7, 209, "Input"],
Cell[231922, 6063, 49755, 1225, 279, 12945, 766, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[281692, 7291, 107, 3, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[281848, 7300, 82, 2, 51, "Section",
  Evaluatable->False],
Cell[281933, 7304, 1205, 26, 171, "Text"],

Cell[CellGroupData[{
Cell[283163, 7334, 41, 0, 68, "Subsection"],
Cell[283207, 7336, 509, 13, 107, "Text"],
Cell[283719, 7351, 156, 5, 41, "Text",
  Evaluatable->False],
Cell[283878, 7358, 148, 2, 99, "Input"],
Cell[284029, 7362, 54, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[284108, 7366, 64, 1, 73, "Input"],
Cell[284175, 7369, 771, 24, 518, "Print"]
}, Closed]],
Cell[284961, 7396, 1132, 29, 122, "Text"],

Cell[CellGroupData[{
Cell[286118, 7429, 51, 1, 73, "Input"],
Cell[286172, 7432, 357, 6, 84, "Output"]
}, Closed]],
Cell[286544, 7441, 624, 20, 42, "Text"],
Cell[287171, 7463, 412, 8, 83, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[287620, 7476, 41, 0, 68, "Subsection"],
Cell[287664, 7478, 911, 24, 96, "Text"],

Cell[CellGroupData[{
Cell[288600, 7506, 110, 2, 73, "Input"],
Cell[288713, 7510, 694, 9, 184, "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[289468, 7526, 41, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[289534, 7530, 44, 0, 68, "Subsection"],
Cell[289581, 7532, 1587, 48, 200, "Text"],
Cell[291171, 7582, 2783, 78, 283, "Text"],
Cell[293957, 7662, 234, 6, 41, "Text"],
Cell[294194, 7670, 72, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[294291, 7675, 83, 1, 73, "Input"],
Cell[294377, 7678, 141, 7, 144, "Print"]
}, Closed]],
Cell[294533, 7688, 52, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[294610, 7692, 105, 2, 73, "Input"],
Cell[294718, 7696, 93, 1, 59, "Output"]
}, Closed]],
Cell[294826, 7700, 580, 11, 95, "Text"],

Cell[CellGroupData[{
Cell[295431, 7715, 84, 1, 73, "Input"],
Cell[295518, 7718, 75, 1, 59, "Output"]
}, Closed]],
Cell[295608, 7722, 108, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[295741, 7729, 66, 1, 73, "Input"],
Cell[295810, 7732, 216, 6, 59, "Output"]
}, Closed]],
Cell[296041, 7741, 675, 12, 120, "Text"],

Cell[CellGroupData[{
Cell[296741, 7757, 76, 1, 73, "Input"],
Cell[296820, 7760, 209, 6, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[297066, 7771, 66, 1, 73, "Input"],
Cell[297135, 7774, 198, 6, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[297370, 7785, 66, 1, 73, "Input"],
Cell[297439, 7788, 188, 6, 59, "Output"]
}, Closed]],
Cell[297642, 7797, 110, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[297777, 7805, 58, 1, 73, "Input"],
Cell[297838, 7808, 35, 1, 59, "Output"]
}, Closed]],
Cell[297888, 7812, 51, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[297964, 7816, 64, 1, 73, "Input"],
Cell[298031, 7819, 43, 1, 64, "Output"]
}, Closed]],
Cell[298089, 7823, 857, 26, 69, "Text"],

Cell[CellGroupData[{
Cell[298971, 7853, 119, 2, 99, "Input"],
Cell[299093, 7857, 207, 6, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[299337, 7868, 120, 2, 99, "Input"],
Cell[299460, 7872, 198, 6, 59, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[299707, 7884, 49, 0, 68, "Subsection"],
Cell[299759, 7886, 336, 6, 67, "Text"],
Cell[300098, 7894, 44, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[300167, 7898, 85, 1, 73, "Input"],
Cell[300255, 7901, 84, 1, 59, "Output"]
}, Closed]],
Cell[300354, 7905, 66, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[300445, 7909, 93, 1, 73, "Input"],
Cell[300541, 7912, 9018, 115, 2759, "Output"]
}, Closed]],
Cell[309574, 8030, 357, 6, 67, "Text"],

Cell[CellGroupData[{
Cell[309956, 8040, 77, 1, 73, "Input"],
Cell[310036, 8043, 8078, 103, 2484, "Output"]
}, Closed]],
Cell[318129, 8149, 105, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[318259, 8156, 92, 1, 73, "Input"],
Cell[318354, 8159, 264, 7, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[318655, 8171, 102, 2, 73, "Input"],
Cell[318760, 8175, 267, 8, 59, "Output"]
}, Closed]],
Cell[319042, 8186, 360, 10, 69, "Text"],

Cell[CellGroupData[{
Cell[319427, 8200, 146, 2, 73, "Input"],
Cell[319576, 8204, 7394, 99, 2484, "Output"]
}, Closed]],
Cell[326985, 8306, 340, 9, 68, "Text"],

Cell[CellGroupData[{
Cell[327350, 8319, 165, 3, 73, "Input"],
Cell[327518, 8324, 403, 6, 60, "Message"],
Cell[327924, 8332, 5740, 94, 1259, "Output"]
}, Closed]],
Cell[333679, 8429, 111, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[333815, 8436, 84, 1, 73, "Input"],
Cell[333902, 8439, 2169, 33, 459, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[336108, 8477, 52, 1, 73, "Input"],
Cell[336163, 8480, 31458, 648, 280, 5185, 319, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[367636, 9131, 326, 6, 67, "Text"],
Cell[367965, 9139, 570, 11, 151, "Input"],

Cell[CellGroupData[{
Cell[368560, 9154, 105, 2, 73, "Input"],
Cell[368668, 9158, 403, 6, 60, "Message"],
Cell[369074, 9166, 2169, 33, 459, "Output"]
}, Closed]],
Cell[371258, 9202, 114, 3, 58, "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[371421, 9211, 36, 0, 51, "Section"],
Cell[371460, 9213, 545, 12, 94, "Text"],

Cell[CellGroupData[{
Cell[372030, 9229, 47, 1, 73, "Input"],
Cell[372080, 9232, 845, 12, 184, "Output"]
}, Closed]],
Cell[372940, 9247, 247, 5, 67, "Text"],
Cell[373190, 9254, 117, 2, 99, "Input"],
Cell[373310, 9258, 147, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[373482, 9265, 51, 1, 73, "Input"],
Cell[373536, 9268, 1675, 99, 2168, "Print"]
}, Closed]],
Cell[375226, 9370, 328, 6, 67, "Text"],
Cell[375557, 9378, 83, 1, 73, "Input"],
Cell[375643, 9381, 181, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[375849, 9389, 167, 4, 99, "Input"],
Cell[376019, 9395, 1696027, 21928, 507, 699443, 9621, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[2072061, 31326, 24, 0, 41, "Text"],
Cell[2072088, 31328, 63, 1, 73, "Input"],
Cell[2072154, 31331, 113, 3, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[2072304, 31339, 33, 0, 54, "ExerciseMain"],

Cell[CellGroupData[{
Cell[2072362, 31343, 65, 0, 58, "ExerciseTitle"],
Cell[2072430, 31345, 744, 16, 120, "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

