(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    478179,      11324]*)
(*NotebookOutlinePosition[    515217,      12632]*)
(*  CellTagsIndexPosition[    515173,      12628]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[BoxData[
    \(Home\ Run\)], "Title",
  TextAlignment->Center,
  TextJustification->0,
  FontFamily->"Myriad Roman"],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "Air resistance has a very important effect upon the trajectory of a batted \
baseball, but the equations of motion cannot be solved analytically when \
damping is included (except for very special cases).  Therefore, we use this \
relatively simple physical problem to illustrate the use of ",
  StyleBox["NDSolve", "InlineInput",
    FontWeight->"Bold"],
  " for systems of coupled nonlinear differential equations and some useful \
programming techniques for ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ". "
}], "Text"],

Cell[TextData[{
  "We consider two types of drag forces.  When a small spherical object moves \
through a viscous fluid at relatively low speeds, such that the fluid \
exhibits a steady streamline flow around the object, Stoke's theory suggests \
that the resistive force is ",
  Cell[BoxData[
      \(TraditionalForm\`F\&\[RightVector]\  = \ \(-\[Alpha]\)\ v\&\
\[RightVector]\)]],
  " with ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha] = 6  \[Pi]\ \[Eta]\ r\)]],
  ", where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\)]],
  " is the coefficient of viscosity and ",
  Cell[BoxData[
      \(TraditionalForm\`r\)]],
  " is the radius of the object.  For objects with other shapes the drag \
coefficient will have a different form and may depend upon the orientation of \
the object relative to its velocity.  For larger objects or larger speeds, \
air drag is better approximated by a quadratic expression of the form ",
  Cell[BoxData[
      \(TraditionalForm\`F\&\[RightVector]\  = \ \(-\[Beta]\)\ v\ v\&\
\[RightVector]\)]],
  ".  The coefficient can be estimated by assuming that the fluid within the \
cylindrical volume swept by the object must be given enough kinetic energy to \
reach the same velocity as the object, from which we expect ",
  Cell[BoxData[
      \(TraditionalForm\`\[Beta] = \(\[Pi]\/2\) \[Rho]\ r\^2\ C\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Rho]\)]],
  " is the density of the fluid and ",
  Cell[BoxData[
      \(TraditionalForm\`C \[LessEqual] 1\)]],
  " is a correction factor for the fraction of the fluid that flows around \
the object without attaining the same velocity.  The parameters ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`C\)]],
  " depend upon the nature of the fluid, the shape and texture of the object, \
and other effects which are difficult to predict accurately."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(\(Off[General::spell, General::spell1];\)\), "\n", 
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", 
          FontSize \[Rule] 12};\)\)}], "Input"],

Cell[BoxData[{
    \(\(Needs["\<Utilities`Notation`\>"];\)\), "\n", 
    \(\(Needs["\<Miscellaneous`PhysicalConstants`\>"];\)\), "\n", 
    \(Needs["\<Miscellaneous`Units`\>"]\)}], "Input"],

Cell[BoxData[
    \(\(Off[Symbolize::"\<boxSymbolExists\>"];\)\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(x\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "   ", 
      RowBox[{
        RowBox[{"Symbolize", "[", 
          TagBox[\(y\_0\),
            NotationBoxTag,
            Editable->True], "]"}], 
        TagBox["",
          NotationBoxTag,
          Editable->True]}], ";", "\n", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(v\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "   ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Theta]\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "\n", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(v\_x0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(v\_y0\),
          NotationBoxTag,
          Editable->True], "]"}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Equations of motion", "Section"],

Cell[CellGroupData[{

Cell["Ideal trajectory", "Subsection"],

Cell[TextData[{
  "The ballistic trajectory in the absence of air resistance can be written \
immediately in closed form.  In addition to the initial conditions, we \
include an argument ",
  StyleBox["parameters", "InlineInput",
    FontWeight->"Bold"],
  " which is intended to be a list of replacement rules specifying physical \
parameters ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`g\)]],
  ".  Although ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " is irrelevant to the ideal trajectory, it will be needed for resisted \
motion."
}], "Text"],

Cell[BoxData[
    \(\(freeTrajectory[\(x\_0\) : _, \(y\_0\) : _, \(v\_0\) : _, \
\(\[Theta]\_0\) : _, \ 
          parameters_]\  := {x\_0 + v\_0\ Cos[\[Theta]\_0]\ t, 
            y\_0 + v\_0\ Sin[\[Theta]\_0]\ t - \(g\ t\^2\)\/2}\  /. 
          parameters;\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Baseball with air resistance", "Subsection"],

Cell["\<\
The equations for the motion of a projectile in a uniform gravitational field \
(flat earth approximation) subject to air resistance, including both linear \
and quadratic velocity dependencies are:\
\>", "Text"],

Cell[BoxData[
    RowBox[{
      RowBox[{"equationsOfMotion", "=", 
        RowBox[{
          RowBox[{"{", 
            RowBox[{
              RowBox[{
                RowBox[{"m", " ", 
                  RowBox[{
                    SuperscriptBox["x", "\[DoublePrime]",
                      MultilineFunction->None], "[", "t", "]"}]}], "\[Equal]",
                 
                RowBox[{\(-\((\[Alpha] + \[Beta]\ v[t])\)\), " ", 
                  RowBox[{
                    SuperscriptBox["x", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}]}]}], ",", 
              RowBox[{
                RowBox[{"m", " ", 
                  RowBox[{
                    SuperscriptBox["y", "\[DoublePrime]",
                      MultilineFunction->None], "[", "t", "]"}]}], "\[Equal]",
                 
                RowBox[{\(\(-m\)\ g\), "-", 
                  RowBox[{\((\[Alpha] + \[Beta]\ v[t])\), " ", 
                    RowBox[{
                      SuperscriptBox["y", "\[Prime]",
                        MultilineFunction->None], "[", "t", "]"}]}]}]}]}], 
            "}"}], "/.", 
          RowBox[{\(v[t]\), "\[Rule]", 
            SqrtBox[
              RowBox[{
                SuperscriptBox[
                  RowBox[{
                    SuperscriptBox["x", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}], "2"], "+", 
                SuperscriptBox[
                  RowBox[{
                    SuperscriptBox["y", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}], 
                  "2"]}]]}]}]}], ";"}]], "Input"],

Cell[TextData[{
  "It is convenient to express the initial velocity in terms of speed and \
angle.  The following functions translates these initial conditions into the \
Cartesian components of velocity and includes the initial position also in \
the form of a set of equations that can be given to ",
  StyleBox["NDSolve", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(ic[\(x\_0\) : _, \ \(y\_0\) : _, \(v\_0\) : _, \
\(\[Theta]\_0\) : _]\), "=", 
        RowBox[{"{", 
          RowBox[{\(x[0] \[Equal] x\_0\), ",", \(y[0] \[Equal] y\_0\), ",", 
            RowBox[{
              RowBox[{
                SuperscriptBox["x", "\[Prime]",
                  MultilineFunction->None], "[", "0", "]"}], 
              "\[Equal]", \(v\_0\ Cos[\[Theta]\_0]\)}], ",", 
            RowBox[{
              RowBox[{
                SuperscriptBox["y", "\[Prime]",
                  MultilineFunction->None], "[", "0", "]"}], 
              "\[Equal]", \(v\_0\ Sin[\[Theta]\_0]\)}]}], "}"}]}], 
      ";"}]], "Input"],

Cell["\<\
The complete set of equations is obtained by joining the equations of motion \
with the initial conditions and applying the replacement rules specifying the \
physical parameters.\
\>", "Text"],

Cell[BoxData[
    \(\(eq1[\(x\_0\) : _, \ \(y\_0\) : _, \(v\_0\) : _, \(\[Theta]\_0\) : _, \
\ parameters_]\  := \ 
        Join[equationsOfMotion, ic[x\_0, \ y\_0, v\_0, \[Theta]\_0]]\  /. 
          parameters;\)\)], "Input"],

Cell["\<\
The trajectory is the two-dimensional vector solution of these differential \
equations and is expressed as a function of the initial conditions and \
physical parameters.\
\>", "Text"],

Cell[BoxData[
    \(\(resistedTrajectory[\(x\_0\) : _, \(y\_0\) : _, \(v\_0\) : _, \(\
\[Theta]\_0\) : _, 
          parameters_] := {x[t], y[t]}\  /. 
            NDSolve[eq1[x\_0, \ y\_0, v\_0, \[Theta]\_0, \ parameters], {x, 
                y}, {t, 0, 20}]\  // Flatten;\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Baseball parameters", "Subsection"],

Cell["\<\
Choose parameters that might be suitable for a baseball game at sea level.  \
SI units are employed.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(baseballParameters = {g \[Rule] 9.8, 
          m \[Rule] 0.145, \[Alpha] \[Rule] 
            6  \[Pi]\ \[Eta]\ r, \[Beta] \[Rule] \(\[Pi]\/2\) \[Rho]\ r\^2\ \
C} /. {r \[Rule] 0.0366, \[Eta] \[Rule] 1.8\ 10\^\(-5\), \[Rho] \[Rule] 
            1.2, \ C \[Rule] 0.3}\)], "Input"],

Cell[BoxData[
    \({g \[Rule] 9.8`, 
      m \[Rule] 0.145`, \[Alpha] \[Rule] 
        0.000012418087441109733`, \[Beta] \[Rule] 
        0.0007575033339076938`}\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Going, going, gone!", "Section"],

Cell[CellGroupData[{

Cell["Compute and plot a trajectory for a home run", "Subsection"],

Cell["\<\
The length of typical major-league home run is about 400 feet, or\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Convert[400\ Feet, \ Meter] // N\)], "Input"],

Cell[BoxData[
    \(121.92`\ Meter\)], "Output"]
}, Closed]],

Cell["\<\
A little bit of trial and error, or some knowledge of the game, suggests that \
an initial speed of about 120 mph is needed.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Convert[120\ Mile/Hour, \ Meter/Second] // N\)], "Input"],

Cell[BoxData[
    \(\(53.6448`\ Meter\)\/Second\)], "Output"]
}, Closed]],

Cell[TextData[{
  "It is useful to define a function which, given initial conditions and \
parameters, produces a plot that compares the trajectories with and without \
air resistance.  Note that we set ",
  StyleBox["AspectRatio\[Rule]1", "InlineInput",
    FontWeight->"Bold"],
  " so that the trajectories will look correct; the default setting would \
distort the appearance of this plot."
}], "Text"],

Cell[BoxData[
    \(\(plotResistedVsFreeTrajectories[\(x\_0\) : _, \ \(y\_0\) : _, \(v\_0\) \
: _, \(\[Theta]\_0\) : _, parameters_] := 
        ParametricPlot[
          Evaluate[{resistedTrajectory[x\_0, y\_0, v\_0, \[Theta]\_0, 
                parameters], 
              freeTrajectory[x\_0, y\_0, v\_0, \[Theta]\_0, 
                parameters]}], {t, 0, 10}, 
          PlotRange \[Rule] {{0, 250}, {0, 50}}, Frame \[Rule] True, 
          GridLines \[Rule] Automatic, AspectRatio \[Rule] Automatic, 
          FrameLabel \[Rule] {"\<x (m)\>", "\<y (m)\>", "\<Baseball with Air \
Resistance\>", "\< \>"}];\)\)], "Input"],

Cell["\<\
Thus, for typical initial conditions and baseball parameters, we find that \
air resistance reduces the range from near 250 m to about 130 m.  Without air \
resistance, the ballpark would need to be about twice as long!\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotResistedVsFreeTrajectories[0, \ 0, 50, 35\ Degree, 
        baseballParameters];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .2 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-1.04083e-017 0.004 3.46945e-018 0.004 [
[.2 -0.0125 -8 -15.6875 ]
[.2 -0.0125 8 0 ]
[.4 -0.0125 -11 -15.6875 ]
[.4 -0.0125 11 0 ]
[.6 -0.0125 -11 -15.6875 ]
[.6 -0.0125 11 0 ]
[.8 -0.0125 -11 -15.6875 ]
[.8 -0.0125 11 0 ]
[1 -0.0125 -11 -15.6875 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -15.0625 -31.375 ]
[.5 -0.0125 15.0625 -15.6875 ]
[-0.0125 .04 -16 -7.84375 ]
[-0.0125 .04 0 7.84375 ]
[-0.0125 .08 -16 -7.84375 ]
[-0.0125 .08 0 7.84375 ]
[-0.0125 .12 -16 -7.84375 ]
[-0.0125 .12 0 7.84375 ]
[-0.0125 .16 -16 -7.84375 ]
[-0.0125 .16 0 7.84375 ]
[-0.0125 .2 -16 -7.84375 ]
[-0.0125 .2 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .1 -31.6875 -14.8125 ]
[-0.01875 .1 -16 14.8125 ]
[ 0 .2 .125 0 ]
[.5 .2125 -71.0625 0 ]
[.5 .2125 71.0625 15.6875 ]
[ 1 0 .125 0 ]
[1.0125 .1 0 -3.5 ]
[1.0125 .1 15.6875 3.5 ]
[ 0 0 0 0 ]
[ 1 .2 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.2 0 m
.2 .2 L
s
.4 0 m
.4 .2 L
s
.6 0 m
.6 .2 L
s
.8 0 m
.8 .2 L
s
0 .04 m
1 .04 L
s
0 .08 m
1 .08 L
s
0 .12 m
1 .12 L
s
0 .16 m
1 .16 L
s
0 g
.2 0 m
.2 .00625 L
s
gsave
.2 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.4 0 m
.4 .00625 L
s
gsave
.4 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.6 0 m
.6 .00625 L
s
gsave
.6 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.8 0 m
.8 .00625 L
s
gsave
.8 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(200) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(250) show
1.000 setlinewidth
grestore
.125 Mabswid
.04 0 m
.04 .00375 L
s
.08 0 m
.08 .00375 L
s
.12 0 m
.12 .00375 L
s
.16 0 m
.16 .00375 L
s
.24 0 m
.24 .00375 L
s
.28 0 m
.28 .00375 L
s
.32 0 m
.32 .00375 L
s
.36 0 m
.36 .00375 L
s
.44 0 m
.44 .00375 L
s
.48 0 m
.48 .00375 L
s
.52 0 m
.52 .00375 L
s
.56 0 m
.56 .00375 L
s
.64 0 m
.64 .00375 L
s
.68 0 m
.68 .00375 L
s
.72 0 m
.72 .00375 L
s
.76 0 m
.76 .00375 L
s
.84 0 m
.84 .00375 L
s
.88 0 m
.88 .00375 L
s
.92 0 m
.92 .00375 L
s
.96 0 m
.96 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -76.0625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
72.125 15.813 moveto
(H) show
75.938 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.313 15.813 moveto
(L) show
89.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .04 m
.00625 .04 L
s
gsave
-0.0125 .04 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .08 m
.00625 .08 L
s
gsave
-0.0125 .08 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .12 m
.00625 .12 L
s
gsave
-0.0125 .12 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
0 .16 m
.00625 .16 L
s
gsave
-0.0125 .16 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .2 m
.00625 .2 L
s
gsave
-0.0125 .2 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .008 m
.00375 .008 L
s
0 .016 m
.00375 .016 L
s
0 .024 m
.00375 .024 L
s
0 .032 m
.00375 .032 L
s
0 .048 m
.00375 .048 L
s
0 .056 m
.00375 .056 L
s
0 .064 m
.00375 .064 L
s
0 .072 m
.00375 .072 L
s
0 .088 m
.00375 .088 L
s
0 .096 m
.00375 .096 L
s
0 .104 m
.00375 .104 L
s
0 .112 m
.00375 .112 L
s
0 .128 m
.00375 .128 L
s
0 .136 m
.00375 .136 L
s
0 .144 m
.00375 .144 L
s
0 .152 m
.00375 .152 L
s
0 .168 m
.00375 .168 L
s
0 .176 m
.00375 .176 L
s
0 .184 m
.00375 .184 L
s
0 .192 m
.00375 .192 L
s
.25 Mabswid
0 0 m
0 .2 L
s
gsave
-0.01875 .1 -99.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
75.8125 11.8438 translate
90 rotate
-75.8125 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
71.625 15.813 moveto
(H) show
75.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
84.813 15.813 moveto
(L) show
88.625 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .19375 m
0 .2 L
s
.2 .19375 m
.2 .2 L
s
.4 .19375 m
.4 .2 L
s
.6 .19375 m
.6 .2 L
s
.8 .19375 m
.8 .2 L
s
.125 Mabswid
.04 .19625 m
.04 .2 L
s
.08 .19625 m
.08 .2 L
s
.12 .19625 m
.12 .2 L
s
.16 .19625 m
.16 .2 L
s
.24 .19625 m
.24 .2 L
s
.28 .19625 m
.28 .2 L
s
.32 .19625 m
.32 .2 L
s
.36 .19625 m
.36 .2 L
s
.44 .19625 m
.44 .2 L
s
.48 .19625 m
.48 .2 L
s
.52 .19625 m
.52 .2 L
s
.56 .19625 m
.56 .2 L
s
.64 .19625 m
.64 .2 L
s
.68 .19625 m
.68 .2 L
s
.72 .19625 m
.72 .2 L
s
.76 .19625 m
.76 .2 L
s
.84 .19625 m
.84 .2 L
s
.88 .19625 m
.88 .2 L
s
.92 .19625 m
.92 .2 L
s
.96 .19625 m
.96 .2 L
s
.25 Mabswid
0 .2 m
1 .2 L
s
gsave
.5 .2125 -132.063 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Baseball) show
107.000 15.813 moveto
(with) show
131.250 15.813 moveto
(Air) show
150.125 15.813 moveto
(Resistance) show
201.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 0 m
1 0 L
s
.99375 .04 m
1 .04 L
s
.99375 .08 m
1 .08 L
s
.99375 .12 m
1 .12 L
s
.99375 .16 m
1 .16 L
s
.125 Mabswid
.99625 .008 m
1 .008 L
s
.99625 .016 m
1 .016 L
s
.99625 .024 m
1 .024 L
s
.99625 .032 m
1 .032 L
s
.99625 .048 m
1 .048 L
s
.99625 .056 m
1 .056 L
s
.99625 .064 m
1 .064 L
s
.99625 .072 m
1 .072 L
s
.99625 .088 m
1 .088 L
s
.99625 .096 m
1 .096 L
s
.99625 .104 m
1 .104 L
s
.99625 .112 m
1 .112 L
s
.99625 .128 m
1 .128 L
s
.99625 .136 m
1 .136 L
s
.99625 .144 m
1 .144 L
s
.99625 .152 m
1 .152 L
s
.99625 .168 m
1 .168 L
s
.99625 .176 m
1 .176 L
s
.99625 .184 m
1 .184 L
s
.99625 .192 m
1 .192 L
s
.25 Mabswid
1 0 m
1 .2 L
s
gsave
1.0125 .1 -56.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
64.5 11.8438 translate
90 rotate
-64.5 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
66.000 15.813 moveto
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .2 L
0 .2 L
closepath
clip
newpath
.5 Mabswid
0 0 m
.06322 .04114 L
.12586 .07489 L
.15437 .08755 L
.17993 .0973 L
.20522 .10534 L
.22836 .1112 L
.24032 .11364 L
.2531 .11578 L
.26452 .11727 L
.27492 .11827 L
.2803 .11865 L
.28325 .11882 L
.28604 .11896 L
.28862 .11906 L
.28997 .1191 L
.29142 .11914 L
.29275 .11918 L
.29396 .1192 L
.29517 .11922 L
.29631 .11923 L
.29765 .11924 L
.29913 .11924 L
.29987 .11924 L
.30067 .11923 L
.3021 .11922 L
.30352 .1192 L
.30487 .11918 L
.30739 .11912 L
.31034 .11901 L
.31306 .11889 L
.31919 .11852 L
.32424 .11812 L
.32964 .11758 L
.33933 .11634 L
.35068 .11443 L
.36094 .11227 L
.38026 .10704 L
.40083 .09973 L
.43945 .08074 L
.47462 .05692 L
.50987 .02603 L
s
.50987 .02603 m
.53376 0 L
s
0 0 m
.06646 .04331 L
.13894 .08319 L
.20702 .11366 L
.27248 .13658 L
.30897 .14663 L
.34216 .15409 L
.37665 .16014 L
.40924 .16425 L
.42592 .16576 L
.43505 .16641 L
.4436 .16691 L
.45148 .16728 L
.45869 .16754 L
.4628 .16765 L
.4667 .16773 L
.4687 .16777 L
.47082 .1678 L
.47205 .16781 L
.47316 .16782 L
.47532 .16784 L
.47739 .16785 L
.47932 .16785 L
.48144 .16785 L
.48371 .16784 L
.48485 .16783 L
.48609 .16782 L
.48834 .16779 L
.49256 .16773 L
.49656 .16764 L
.50081 .16752 L
.50843 .16724 L
.51721 .16681 L
.52545 .16631 L
.54424 .16478 L
.56116 .16298 L
.57896 .16062 L
.6108 .15525 L
.64475 .14789 L
.67635 .13954 L
.74352 .11693 L
.8149 .08567 L
.88368 .04852 L
.94985 .00626 L
s
.94985 .00626 m
.95832 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{576.25, 115.25},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000=P0000[2000`400?l00000o`00003oo`3ooooo
0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?oook/0oooo00<0oooO09a8B02L
gol03@3oool00`3ogi`0B4RL0=ooo`3o0?oooi@0oooo003o0?oook/0oooo00<0_gA809cOo`3oool0
3P3oool00`3ogi`0B7Bo0?ooo`3o0?oooi<0oooo003o0?oookX0oooo00<0gia804RLg`3oool0403o
ool00`3OW4P0B9cO0?ooo`3o0?oooi80oooo003o0?oookT0oooo00<0oono07@0M02oool04@3oool0
0`3ookl0M01809cOo`3o0?oooi80oooo003o0?ooojX0oooo00D0_g@000000000M;l0oooo0=nLB002
000000030018W03Oool0oooo00D0oooo00D0gia80018W03Oool0oooo0=nLB002000000030000B02L
gml0W4P00080000000<0001d0;oog`2LB0000P0000001P00M;l0oooo0?ooo`3ooml0W4P004RLgol0
ooooTP3oool00?l0ooooZP3oool0203ooml0W4Qd0;ooo`3oool0oooO09a80000B9`0gooo1`3oool0
0`2oM00007Bo0?ooo`020?ooo`030=nLB00007@0_ooo0080oooo00L0_g@00018W03Oool0oooo0?oo
g`2LB00007Bo00@0oooo00<0gia80018W03Oool0o`3ooonA0?ooo`00o`3ooon[0?ooo`060?oo_`1d
B9`0gooo0;md000007@0_ooo1`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0080
oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo00@0oooo00<0gia80000M02oool0o`3ooonA
0?ooo`00o`3ooon/0?ooo`040?oOW01dB4P0000007Boo`P0oooo00<0oono07@0001d_ol00`3oool0
0`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`040?ooo`030?no
M00007@0_ooo0?l0ooooT@3oool00?l0oooo[03oool00`3ookl0M00004RLg`090?ooo`030?oo_`1d
0000M;oo00<0oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000
M;l0103oool00`3o_g@0001809cOo`3o0?oooi40oooo003o0?oooj/0oooo00D0oooO09a80000B9`0
W7A809cOo`080?ooo`030?oo_`1d0000M;oo00<0oooo00<0gia80000B02Lgol00P3oool01`2oM000
001d0;ooo`3oool0oooO09a80000M;l0103oool00`3o_g@0001d0;ooo`3o0?oooi40oooo003o0?oo
oj/0oooo00H0_g@00000B02Lgol0oooO09a8M02oool70?ooo`0B0?oo_`1d0000B9cO0?ooo`3ooml0
W4P00000000004P0B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO103oool00`3O
W4P0001d0;ooo`3o0?oooi40oooo003o0?ooojT0oooo00T0oooO09a8000000000000001d_`3o_g@0
00000000M02oool01P3oool00`3ooml0W4P004RLg`030?ooo`0<0;md000007@0_ooO09a800000000
001809cOo`3ookl0M0000000000007@0_ooo103oool00`3OW4P004RL0=ooo`3o0?oooi40oooo003o
0?oookT0oooo00<0gia80018W03Oool04@3oool00`3ooml0W4P004RLg`3o0?oooi80oooo003o0?oo
okT0oooo00<0oono07@0B02Lgol04@3oool00`3ookl0M01809cOo`3o0?oooi80oooo003o0?oookX0
oooo00<0gia804RLg`3oool0403oool00`3OW4P0B9cO0?ooo`3o0?oooi80oooo003o0?oook/0oooo
00<0_g@004RLg`3oool03P3oool00`3o_g@007Bo0?ooo`3o0?oooi<0oooo003o0?oook`0oooo00<0
giad09cOo`3oool0303oool00`3ookl0M7Bo0?ooo`3o0?oooi@0oooo003o0?ooool0ooooo`3ooomS
0?ooo`00o`3ooooo0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?ooool0oooo
o`3ooomS0?ooo`00o`3ooooo0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?oo
ool0ooooo`3ooomS0?ooo`00o`3ooooo0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo
003o0?ooool0ooooo`3ooomS0?ooo`00o`3ooooo0?ooool0ooooH`3oool00?l0oooo;P3oool00`3o
oml0W4P0000000020000000307Boo`3oool0oooo00<0oooo00D0omnL04P000000000001d0;ooo`17
0?ooo`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo00@0
oooo00D0omnL04P000000000001d0;ooo`130?ooo`090;md00000000000000000000B9`0gooo0?oo
o`3ooml0W4P000<0000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo0440
oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3oool0omnL04P000000000001d0;oo
o`040?ooo`050?oOW018000000000000M02oool0@@3oool00`2oM0000000000000040000000404RL
g`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo
0;L0oooo003o0?ooobh0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`170?ooo`030?oOW0180000M;oo00@0oooo00L0oono
07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
04<0oooo00<0omnL04P0001d_ol00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo
0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0440oooo00<0gia80000000000000`00
00002P00B9`0gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000M02oool20?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`110?ooo`030=nLB0000000000000<0000000D004RL0=ooo`3ooml0
W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo0;H0oooo003o0?oooc<0oooo00<0oooO09a8M02oool00P3oool00`2oM00004RL0=ooo`020?oo
o`030;md0000M;l0oooo04H0oooo00<0omnL04P0001d_ol0103oool00`2oM00004RL0=ooo`020?oo
o`070;md0000M;l0oooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo0480oooo00<0
omnL04P0001d_ol0203oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool0@@3oool00`3OW4P0B9cO0?ooo`020?ooo`060=nLM02Lgol0oooo0;md0000B9`0
gooo0P3oool01`2oM00007Bo0?ooo`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`11
0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0103oool00`3ooml0W4Qd0;ooo`02
0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0]@3oool00?l0oooo=03oool01P2o
M000M;oo0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0AP3oool00`3ogi`0B00007Bo
o`030?ooo`030?oo_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`03
0=nLB00007@0_ooo0480oooo00<0omnL04P0001d_ol02@3oool01P2oM000M;oo0?ooo`3ookl0M000
04RLg`<0oooo00<0gia80000M02oool0@P3oool00`3OW4P0M;oo0?ooo`030?ooo`030?oo_`1d0000
B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo0480oooo
00<0gia807Boo`3oool02@3oool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000
M02oool0]@3oool00?l0oooo<`3oool01`3ooml0W4P004RLg`3oool0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`160?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool0
1P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0@P3oool00`3ogi`0B000
07Boo`080?ooo`070?oog`2LB000B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
04<0oooo00<0_gA809cOo`3oool00P3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo
0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`130?ooo`030;mdB02Lgol0oooo00L0oooo00L0
oooO09a80018W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0]@3oool00?l0oooo
<`3oool01`3OW4P0000007Boo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`160?oo
o`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0
B00007Boo`<0oooo00<0okmd0000B02Lgol0@P3oool00`3ogi`0B00007Boo`080?ooo`070=nLB000
0000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo04<0oooo00<0oooO09a8M02o
ool00P3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool0
0`3o_g@0001809cOo`130?ooo`030?oog`2LB7@0_ooo00L0oooo00L0gia80000001d_ol0oooo0?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol0]@3oool00?l0oooo;`3oool02`3Ogml0oooo0?oo
_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0AP3oool0
0`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0
001d_ol30?ooo`030?noM00004P0W=oo0480oooo00<0omnL04P0001d_ol0103oool02`3Ogml0oooo
0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0A03o
ool01`3o_g@007Bo0?ooo`3oool0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B000
07Boo`<0oooo00<0okmd0000B02Lgol0A03oool00`3o_g@007Bo0?ooo`020?ooo`0;0=oOg`3oool0
oono07@000000000001d0;ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`2e0?oo
o`00o`3oool_0?ooo`060=nLB000000000000000000004P0W=oo0P3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo04H0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`03
0?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`120?ooo`030?oO
W0180000M;oo00@0oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol0A03oool01`3ookl0M01809cOo`3oool0omnL04P0001d_ol00`3o
ool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0A03oool00`3ookl0
M01809cOo`020?ooo`060=nLB000000000000000000004P0W=oo0P3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo0;D0oooo003o0?ooobl0oooo00@0oono07@000000000B9cO103oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo04H0oooo00<0omnL04P0001d_ol00`3oool0
0`3ookl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d
0;ooo`120?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@000000000B9cO103oool00`3ookl0
M00004RLg`030?ooo`030=nLB00007@0_ooo04@0oooo00L0oono07@00018W=l0oooo0?oo_`1d0000
B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo04@0oooo
00<0oono07@00018W=l00P3oool0103ookl0M00000000018W=l40?ooo`030?oo_`1d0000B9cO00<0
oooo00<0gia80000M02oool0]@3oool00?l0oooo<03oool00`2oM000M;oo0?ooo`050?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool0AP3oool00`3ogi`0B00007Boo`040?ooo`030=nL
B000B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0=nLB000B9`0gooo0080oooo00<0_g@0001d
_`3oool0@P3oool00`3ogi`0B00007Boo`050?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018
W03Oool00P3oool00`2oM00007Bo0?ooo`0o0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P0
0018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`3OW4P004RL0=oo
o`020?ooo`030;md0000M;l0oooo03l0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RL
g`030?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?oo
o`2e0?ooo`00o`3oool`0?ooo`060?oOW018000000000000000004P0W=oo0P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0A@3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono
07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
0440oooo00D0oooO09a800000000000007Boo`050?ooo`060?oOW018000000000000000004P0W=oo
0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0@03oool0203ooml0W4Q809cOo`3oool0
oono07@0000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool0@03oool0203ooml0W4Q809cOo`3oool0oono07@00000
07@0_ooo0`3oool01P3ogi`0B000000000000000001809cOo`80oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo0;H0oooo003o0?oooc40oooo00D0_g@0000000000000000004RLg`030?ooo`05
0?oo_`1d000000000000M02oool0B03oool00`3o_g@0000007Boo`050?ooo`050?oo_`1d00000000
0000M02oool0103oool01@3ookl0M0000000000007@0_ooo04@0oooo00<0okmd0000001d_ol01P3o
ool01@2oM0000000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`120?ooo`06
0?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0
oono07@000000000001d0;ooo`120?ooo`060?oog`2LB00000000000000007@0_ooo1@3oool01@2o
M0000000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`2g0?ooo`00o`3ooooo
0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?ooool0ooooo`3ooomS0?ooo`00
o`3ooooo0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?ooool0ooooo`3ooomS
0?ooo`00o`3ooooo0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?ooool0oooo
o`3ooomS0?ooo`00o`3ooooo0?ooool0ooooH`3oool00=`0ooooo`00003400000<40oooo003L0?oo
o`<000003`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0000000A0?ooo`030000003oool0oooo0;l0oooo003L0?ooo`04
0000003oool0oooo000000h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000000=0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00/0oooo0P0000020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0_`3oool0
0=`0oooo00@000000?ooo`3oool0oooo0P00001D0?ooo`030000003oool0oooo05L0oooo00<00000
0?ooo`3oool0=P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo05L0oooo00<00000
0?ooo`3oool0@03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0;l0oooo003L0?oo
o`800000103oool00`000000oooo0?ooo`1A0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?oo
o`3oool0=@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?oo
o`3oool0?P3oool2000001D0oooo0`0000310?ooo`00g03oool00`000000oooo0?ooo`040?ooo`80
0000D@3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3o
ool0803oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo03d0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`2o0?ooo`00g03oool00`000000oooo0?ooo`050?ooo`800000
D03oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo03<0oooo00<000000?ooo`3oool0
8@3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo03/0oooo0P00000J0?ooo`030000
003oool0oooo0;l0oooo003L0?ooo`030000003oool0oooo00L0oooo0P00001>0?ooo`030000O`3o
ool0oooo05L0oooo00<0001o0?ooo`3oool0<`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3o
ool0oooo05L0oooo00<0001o0?ooo`3oool0>@3oool2000001`0oooo00<000000?ooo`3oool0_`3o
ool00=`0oooo0P0000090?ooo`800000C@3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0
oooo0380oooo00<000000?ooo`3oool08P3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0
oooo03P0oooo00<000000?ooo`3oool06P3oool300000<40oooo003L0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0BP3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo
0340oooo00<000000?ooo`3oool08`3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo
03H0oooo0P00000O0?ooo`030000003oool0oooo0;l0oooo003L0?ooo`030000003oool0oooo00/0
oooo0P00001:0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0<03oool00`000000
oooo0?ooo`0T0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0=@3oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo0;l0oooo003L0?ooo`8000003P3oool00`000000oooo
0?ooo`170?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo
0?ooo`0U0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0<`3oool200000200oooo
0`0000310?ooo`00g03oool00`000000oooo0?ooo`0>0?ooo`800000A`3oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo02h0oooo00<000000?ooo`3oool09P3oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`2o0?ooo`00``3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL04P000000000
001d0;ooo`0;0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0A03oool00`0007l0
oooo0?ooo`1G0?ooo`030000O`3oool0oooo02d0oooo00<000000?ooo`3oool09`3oool00`0007l0
oooo0?ooo`1G0?ooo`030000O`3oool0oooo0300oooo0P00000U0?ooo`030000003oool0oooo0;l0
oooo00340?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
00X0oooo00<000000?ooo`3oool04@3oool2000004@0oooo00<0001o0?ooo`3oool0E`3oool00`00
07l0oooo0?ooo`0/0?ooo`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool0E`3oool00`00
07l0oooo0?ooo`0^0?ooo`8000009`3oool00`000000oooo0?ooo`2o0?ooo`00a03oool00`3ogi`0
B00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02@3oool2000001@0
oooo00<000000?ooo`3oool0@@3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo02/0
oooo00<000000?ooo`3oool0:@3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo02d0
oooo00<000000?ooo`3oool09@3oool300000<40oooo00340?ooo`030?oOW0180000M;oo00<0oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`090?ooo`030000003oool0oooo01@0oooo
0P0000110?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0:@3oool2000002`0oooo
00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0[0?ooo`800000:P3oool00`000000oooo
0?ooo`2o0?ooo`00a03oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol02@3oool00`000000oooo0?ooo`0E0?ooo`800000@03oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo02P0oooo00<000000?ooo`3oool0;03oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo02X0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?oo
o`2o0?ooo`00a03oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol02@3oool3000001L0001o0P0000320000O`030000000007l0001o0;80001o0P00000Z
0000O`@00000`@3oool00<@0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00T0oooo00<000000?ooo`3oool0603oool2000003d0oooo00<0001o0?oo
o`3oool0E`3oool00`0007l0oooo0?ooo`0V0?ooo`030000003oool0oooo02h0oooo00<0001o0?oo
o`3oool0E`3oool00`0007l0oooo0?ooo`0V0?ooo`800000;`3oool00`000000oooo0?ooo`2o0?oo
o`00a03oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol02@3oool00`000000oooo0?ooo`0J0?ooo`800000>`3oool00`0007l0oooo0?ooo`1G0?ooo`03
0000O`3oool0oooo02D0oooo00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`1G0?ooo`03
0000O`3oool0oooo02D0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`2o0?ooo`00
a03oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
2@3oool00`000000oooo0?ooo`0K0?ooo`<00000>@3oool00`0007l0oooo0?ooo`1G0?ooo`030000
O`3oool0oooo02@0oooo00<000000?ooo`3oool0<03oool00`0007l0oooo0?ooo`1G0?ooo`030000
O`3oool0oooo02<0oooo0P00000b0?ooo`030000003oool0oooo0;l0oooo00340?ooo`030?oOW018
0000M;oo00@0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`090?ooo`8000007P3o
ool3000003L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0S0?ooo`030000003o
ool0oooo0340oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0R0?ooo`030000003o
ool0oooo0300oooo0`0000310?ooo`00`P3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo00X0oooo00<000000?ooo`3oool07`3oool3000003D0
oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0Q0?ooo`800000=03oool00`0007l0
oooo0?ooo`1G0?ooo`030000O`3oool0oooo0200oooo0P00000e0?ooo`030000003oool0oooo0;l0
oooo00340?ooo`030?noM0000000M;oo00D0oooo00D0oono07@000000000001d0;ooo`0;0?ooo`03
0000003oool0oooo0240oooo0`00000c0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3o
ool0803oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3o
ool07P3oool2000003L0oooo00<000000?ooo`3oool0_`3oool00=`0oooo00<000000?ooo`3oool0
8`3oool300000340oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0O0?ooo`030000
003oool0oooo03D0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0L0?ooo`800000
>@3oool00`000000oooo0?ooo`2o0?ooo`00g03oool2000002H0oooo0P00000`0?ooo`030000O`3o
ool0oooo05L0oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0f0?ooo`030000O`3o
ool0oooo05L0oooo00<0001o0?ooo`3oool06P3oool2000003T0oooo0`0000310?ooo`00[@3oool0
0`2oM4P0W=oo0?ooo`0/0?ooo`030000003oool0oooo02H0oooo0`00000^0?ooo`030000O`3oool0
oooo05L0oooo00<0001o0?ooo`3oool0703oool2000003T0oooo00<0001o0?ooo`3oool0E`3oool0
0`0007l0oooo0?ooo`0H0?ooo`800000?@3oool00`000000oooo0?ooo`2o0?ooo`00[@3oool00`2o
M000000007Boo`060?ooo`030?oog`2LB00007Bo02<0oooo00<000000?ooo`3oool0:03oool30000
02`0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0K0?ooo`030000003oool0oooo
03T0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0G0?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0_`3oool00:d0oooo00D0_g@0000000000000000007Boo`050?oo
o`030;md0000B9`0gooo0280oooo0P00000[0?ooo`<00000:P3oool00`0007l0oooo0?ooo`1G0?oo
o`030000O`3oool0oooo01T0oooo0P00000l0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?oo
o`3oool05@3oool2000003h0oooo0`0000310?ooo`00[@3oool01`2oW9`0gooo0?oog`2LB0000000
0000B02Lgol00P3oool00`3o_g@0001809cOo`0S0?ooo`030000003oool0oooo02`0oooo0`00000X
0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`0l
0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool04`3oool200000480oooo00<00000
0?ooo`3oool0_`3oool00;40oooo00<0oono07@0000000000P0000000`00M;l0oooo0?ooo`080?oo
o`030;md00000000000000@0000000T0B9cO0?ooo`3oool0oooo0?oOW018000000000000M02oool0
2`3oool00`000000oooo0?ooo`0^0?ooo`<000009P3oool00`0007l0oooo0?ooo`1G0?ooo`030000
O`3oool0oooo01L0oooo00<000000?ooo`3oool0?@3oool00`0007l0oooo0?ooo`1G0?ooo`030000
O`3oool0oooo0140oooo0P0000140?ooo`030000003oool0oooo0;l0oooo002a0?ooo`040?oog`2L
B000000007Boo`d0oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3ookl0M00007Bo
o`3oool0okmd0000M02oool:0?ooo`030000003oool0oooo0300oooo0`00000T0?ooo`030000O`3o
ool0oooo05L0oooo00<0001o0?ooo`3oool05@3oool200000400oooo00<0001o0?ooo`3oool0E`3o
ool00`0007l0oooo0?ooo`0?0?ooo`800000AP3oool00`000000oooo0?ooo`2o0?ooo`00[@3oool0
1P2oW9`0gooo0?oOW0180000001809cOoa00oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo
0?ooo`2oM00004RL0=ooo`80oooo00<0_g@0001d_`3oool02@3oool200000380oooo1000000R0?oo
o`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0503oool00`000000oooo0?ooo`100?oo
o`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool03P3oool00`000000oooo0?ooo`140?oo
o`<00000`@3oool00:d0oooo00@0_g@000000000B9`0gooo4`3oool00`3OW4P0M;oo0?ooo`030?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02@3oool00`000000oooo0?ooo`0c0?oo
o`@00000803oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo0180oooo0P0000130?oo
o`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0303oool2000004T0oooo00<000000?oo
o`3oool0_`3oool00:d0oooo00<0_gAd0;ooo`3oool05@3oool00`2oM4P0W=oo0?ooo`020?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?ooo`0e0?ooo`@0
00007P3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo0140oooo00<000000?ooo`3o
ool0@`3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo00X0oooo0P00001;0?ooo`03
0000003oool0oooo0;l0oooo00350?ooo`030?oog`2LB7@0_ooo0080oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`090?ooo`<00000=`0007l4000008P0001o0P00002[0000O`800000
BP0007l400000<40oooo00360?ooo`070?noM000M;l0oooo0?ooo`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00T0oooo00<000000?ooo`3oool0>@3oool4000001X0oooo00<0001o0?ooo`3o
ool0E`3oool00`0007l0oooo0?ooo`0=0?ooo`800000B03oool00`0007l0oooo0?ooo`1G0?ooo`03
0000O`3oool0oooo00H0oooo0P00001?0?ooo`030000003oool0oooo0;l0oooo00360?ooo`070?oo
_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<000000?oo
o`3oool0>`3oool4000001P0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0;0?oo
o`800000BP3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo00@0oooo0P00001A0?oo
o`030000003oool0oooo0;l0oooo00360?ooo`070?oo_`1d0000B9cO0?ooo`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00T0oooo00<000000?ooo`3oool0?@3oool4000001H0oooo00<0001o
0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`090?ooo`800000C03oool00`0007l0oooo0?ooo`1G
0?ooo`030000O`3oool0oooo0080oooo0P00001C0?ooo`030000003oool0oooo0;l0oooo00310?oo
o`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00T0oooo0P0000100?ooo`D000004`3oool00`0007l0oooo0?ooo`1G0?oo
o`030000O`3oool0oooo00L0oooo0P00001>0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?oo
o`3oool00P00001C0?ooo`<00000`@3oool00:/0oooo00<0oooO09a8000000001P0000000`00M;l0
oooo0?ooo`0:0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`030000003oool0oooo0440oooo1P00000@0?ooo`03
0000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool01@3oool200000500oooo00<0001o0?ooo`3o
ool0E`3oool00`0007l000000000001G0?ooo`030000003oool0oooo0;l0oooo002Y0?ooo`030?oo
g`2LB000000000X0000000<007Bo0?ooo`3oool02@3oool01P3ooml0W4P0000000000000001d0;oo
o`@0oooo00D0oono07@000000000001d0;ooo`0;0?ooo`030000003oool0oooo04@0oooo0P000000
103oool0000000000000000=0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool00`3o
ool200000580oooo00<0001o0?ooo`3oool0EP3oool2000005T0oooo00<000000?ooo`3oool0_`3o
ool00:P0oooo00@0okmd00000000B9`0gooo2@3oool0103ookl0M0000000B02LgolS0?ooo`030000
003oool0oooo04H0oooo0`000000103oool0000000000000000:0?ooo`030000O`3oool0oooo05L0
oooo00@0001o0?ooo`3oool0oooo0P00001D0?ooo`030000O`3oool0oooo05<0oooo0`0000000`3o
ool0001o0?ooo`1H0?ooo`030000003oool0oooo0;l0oooo002W0?ooo`030?oo_`1dM;l0oooo00d0
oooo00<0oooO09adW03Oool08P3oool2000004X0oooo0P0000020?ooo`<000001`3oool00`0007l0
oooo0?ooo`1G0?ooo`040000O`3oool00000000005H0oooo00<0001o0?ooo`3oool0D@3oool20000
00@0oooo00<0001o0?ooo`3oool0E@3oool300000<40oooo003L0?ooo`030000003oool0oooo04/0
oooo0P0000030?ooo`<00000103oool00`0007l0oooo0?ooo`1F0?ooo`<00000F03oool00`0007l0
oooo0?ooo`1?0?ooo`8000001P3oool00`0007l0oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0
oooo002]0?ooo`030?oo_`1dM;l0oooo00@0oooo00<0gibL0;ooo`3oool09@3oool00`000000oooo
0?ooo`1=0?ooo`<000000`3oool200000080oooo00<0001o0?ooo`3oool0D`3oool3000000030?oo
o`0007l0oooo05P0oooo00<0001o0?ooo`3oool0C03oool3000000P0oooo00<0001o0?ooo`3oool0
E`3oool00`000000oooo0?ooo`2o0?ooo`00[@3oool00`3o_g@0000000000004000000030000M02o
ool0oooo02D0oooo0P00001A0?ooo`8000000`3oool300000580oooo0`0000040?ooo`030000O`3o
ool0oooo05L0oooo00<0001o0?ooo`3oool0BP3oool2000000/0oooo00<0001o0?ooo`3oool0E@3o
ool300000<40oooo002]0?ooo`030;md00000000000000@0000000<0001d0;ooo`3oool09@3oool0
0`000000oooo0?ooo`1B0?ooo`<000000P3oool00`0007l0000000000002000004/0oooo0`000007
0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0B03oool2000000d0oooo00<0001o
0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o0?ooo`00[P3oool00`2oM7@0_ooo0?ooo`030?oo
o`030=nLW02oool0oooo00X0oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?ooo`030?oo
o`050?oOW018000000000000M02oool02`3oool00`000000oooo0?ooo`1E0?ooo`80000000<0001o
0?ooo`3oool00P3oool5000004<0oooo0`00000:0?ooo`030000O`3oool0oooo05L0oooo00<0001o
0?ooo`3oool0AP3oool2000000l0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o
0?ooo`00[@3oool00`3ogi`0W;oo0?ooo`0A0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RL
g`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02P3oool00`000000oooo0?oo
o`1G0?ooo`800000203oool5000003/0oooo0`00000=0?ooo`030000O`3oool0oooo05L0oooo00<0
001o0?ooo`3oool0@`3oool300000140oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?oo
o`2o0?ooo`00[@3oool00`2oM00007Bo0?ooo`040?ooo`030=nLW02oool0oooo00l0oooo00<0oono
07@0B02Lgol00P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo0P00001H
0?ooo`030000O`3oool00000008000002P3oool600000340oooo1000000@0?ooo`030000O`3oool0
oooo05L0oooo00<0001o0?ooo`3oool0@@3oool2000001@0oooo00<0001o0?ooo`3oool0E@3oool3
00000<40oooo002]0?ooo`030=nLB0000000000000@0000000<0001d0;ooo`3oool03`3oool01`3o
okl0M00007Boo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`090?ooo`030000003o
ool0oooo05L0oooo00<0001o0?ooo`3oool00P3oool2000000h0oooo1@00000W0?ooo`D00000503o
ool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo03l0oooo0P00000F0?ooo`030000O`3o
ool0oooo05L0oooo00<000000?ooo`3oool0_`3oool00:d0oooo00<0oooO09a80000000010000000
0`0007@0_ooo0?ooo`0?0?ooo`070?noM0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00T0oooo00<000000?ooo`3oool0E`3oool00`0007l0oooo0?ooo`040?ooo`<00000
403oool5000001P0oooo2P00000I0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0
?03oool3000001P0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o0?ooo`00[P3o
ool00`3OW9`0_ooo0?ooo`030?ooo`030=nLW02oool0oooo00h0oooo00P0omnL04P0000004P0W=oo
0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol02@3oool300000640001o0`00000B0000
OaP00000^@0007l300000780001o100000310?ooo`00[@3oool00`3o_g@0B9cO0?ooo`0C0?ooo`05
0?noM000000000000000B02Lgol00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00T0oooo00<000000?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0:0?ooo`800000B`3oool00`00
07l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo03H0oooo0`00000N0?ooo`030000O`3oool0oooo
05L0oooo00<000000?ooo`3oool0_`3oool00:d0oooo00<0_g@0001d_`3oool0103oool00`3OW9`0
_ooo0?ooo`0>0?ooo`030;md0000M;l0oooo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`090?ooo`030000003oool0oooo05L0oooo00<0001o0?ooo`3oool0303oool3000004P0
oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0c0?ooo`<000008@3oool00`0007l0
oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo002]0?ooo`030=nLB0000000000000@00000
00<0001d0;ooo`3oool03`3oool01`2oM000B9cO0?ooo`3oool0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`090?ooo`030000003oool0oooo05L0oooo00<0001o0?ooo`3oool03`3oool30000
04D0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0`0?ooo`<00000903oool00`00
07l0oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo002]0?ooo`030?oog`2LB000000000@0
000000<0001d0;ooo`3oool02P3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0080
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`090?ooo`800000F03oool00`0007l0
oooo0?ooo`0B0?ooo`800000@`3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo02d0
oooo0`00000W0?ooo`030000O`3oool0oooo05D0oooo0`0000310?ooo`00]03oool00`3OW9`0_ooo
0?ooo`0;0?ooo`060?noM018W=l0oooo0?oo_`1d000007Bo0`3oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool02P3oool00`000000oooo0?ooo`1G0?ooo`030000O`3oool0oooo01@0oooo0`00
00100?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0:P3oool3000002X0oooo00<0
001o0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o0?ooo`00``3oool01@3OW4P0000000000000
07@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`0;0?ooo`030000003oool0oooo05L0oooo
00<0001o0?ooo`3oool05`3oool3000003d0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo
0?ooo`0W0?ooo`<00000;@3oool00`0007l0oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo
002W0?ooo`030?oo_`1dB9`0gooo00d0oooo00<0omnL04Q8W03Oool08P3oool00`000000oooo0?oo
o`1G0?ooo`030000O`3oool0oooo01X0oooo0P00000k0?ooo`030000O`3oool0oooo05L0oooo00<0
001o0?ooo`3oool0903oool300000300oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?oo
o`2o0?ooo`00Z03oool0103ogi`0B0000018W03Oool90?ooo`040?oo_`1d000004RL0=ooob<0oooo
0P00001H0?ooo`030000O`3oool0oooo01`0oooo1000000g0?ooo`030000O`3oool0oooo05L0oooo
00<0001o0?ooo`3oool08@3oool3000003<0oooo00<0001o0?ooo`3oool0E@3oool300000<40oooo
002Y0?ooo`030?oog`2LB000000000X0000000<007Bo0?ooo`3oool08`3oool00`000000oooo0?oo
o`1G0?ooo`030000O`3oool0oooo0200oooo1000000c0?ooo`030000O`3oool0oooo05L0oooo00<0
001o0?ooo`3oool07P3oool3000003H0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?oo
o`2o0?ooo`00Z`3oool00`3ooml0W4P00000000600000003001d_`3oool0oooo02D0oooo00<00000
0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0T0?ooo`@00000;`3oool00`0007l0oooo0?ooo`1G
0?ooo`030000O`3oool0oooo01X0oooo1000000i0?ooo`030000O`3oool0oooo05L0oooo00<00000
0?ooo`3oool0_`3oool00=`0oooo0P00001H0?ooo`030000O`3oool0oooo02P0oooo1000000[0?oo
o`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool05`3oool3000003d0oooo00<0001o0?oo
o`3oool0E@3oool300000<40oooo003L0?ooo`030000003oool0oooo05L0oooo00<0001o0?ooo`3o
ool0;03oool4000002L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0C0?ooo`@0
0000@03oool00`0007l0oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo00350?ooo`040?oo
g`2LB00004RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`0;0?ooo`030000003oool0oooo
05L0oooo00<0001o0?ooo`3oool0<03oool4000002<0oooo00<0001o0?ooo`3oool0E`3oool00`00
07l0oooo0?ooo`0?0?ooo`@00000A03oool00`0007l0oooo0?ooo`1G0?ooo`030000003oool0oooo
0;l0oooo00350?ooo`040?oog`2LB00004RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo00X0oooo00<000000?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0d0?ooo`@000007`3o
ool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo00X0oooo1@0000180?ooo`030000O`3o
ool0oooo05L0oooo00<000000?ooo`3oool0_`3oool00<D0oooo00@0oooO09a80000B9`0gooo0P3o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo0P00001H0?ooo`030000O`3o
ool0oooo03P0oooo1000000K0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0103o
ool6000004d0oooo00<0001o0?ooo`3oool0E@3oool300000<40oooo00310?ooo`030;md00000000
000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`090?ooo`030000003o
ool0oooo05L0oooo00<0001o0?ooo`3oool0?03oool5000001H0oooo00<0001o0?ooo`3oool0E`3o
ool00`0007l0oooo00000004000005<0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?oo
o`2o0?ooo`00`@3oool0303ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B000
07Boo`<0oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?ooo`1G0?ooo`030000O`3oool0
oooo0440oooo2000000>0?ooo`030000O`3oool0oooo05D0oooo1000001H0?ooo`030000O`3oool0
oooo05L0oooo00<000000?ooo`3oool0_`3oool00<80oooo00/0gia807Boo`3oool0oooO09a80000
B9`0gooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo0`00002S0000O`L0
0000FP0007l500000;<0001o100000310?ooo`00`P3oool02`3ooml0W4Qd0;ooo`3ooml0W4P00018
W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?oo
o`1G0?ooo`030000O`3oool0oooo0500oooo2000001<0?ooo`H000001`3oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo05L0oooo00<000000?ooo`3oool0_`3oool00<<0oooo00X0omnL
04Qd_`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`090?oo
o`030000003oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l00000000000050000
03D0oooo4000000=0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`00
0000oooo0?ooo`2o0?ooo`00a03oool02@3OW4P0M;oO09a80000B9`0gooo0?ooo`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00T0oooo00<000000?ooo`3oool0E`3oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo00D0oooo3@00000N0?ooo`X000007@3oool00`0007l0oooo0?oo
o`1G0?ooo`030000O`3oool0oooo05L0oooo00<000000?ooo`3oool0_`3oool00<@0oooo00D0oooO
09a8B018B00004RL0=ooo`020?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool02@3o
ool2000005P0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`0B0?oooah000009`3o
ool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo05D0oooo0`0000310?ooo`00a@3oool0
103ogi`0B0000018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`03
0000003oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`03
0000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o0?ooo`00
aP3oool00`3OW4P004RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool02`3oool00`000000
oooo0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0
oooo0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<000000?ooo`3oool0_`3oool00=`0oooo
00<000000?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo
00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo
003L0?ooo`800000F03oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<0
001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1E0?ooo`<00000`@3oool00=`0oooo00<00000
0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<0001o
0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo003L0?oo
o`030000003oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?oo
o`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o0?oo
o`00g03oool2000005P0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`03
0000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E@3oool300000<40oooo003L0?ooo`030000
003oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo0?ooo`1G0?ooo`030000
O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo0?ooo`2o0?ooo`00`@3o
ool00`3ooml0W4P0000000020000000307Boo`3oool0oooo00<0oooo00D0omnL04P000000000001d
0;ooo`0;0?ooo`030000003oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo
0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`000000oooo
0?ooo`2o0?ooo`00`@3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0080oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo00X0oooo00<000000?ooo`3oool0E`3oool00`0007l0
oooo0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0
oooo0?ooo`1G0?ooo`030000003oool0oooo0;l0oooo00360?ooo`030?oog`2LB7@0_ooo0080oooo
00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`090?ooo`800000F03oool00`0007l0oooo
0?ooo`1G0?ooo`030000O`3oool0oooo05L0oooo00<0001o0?ooo`3oool0E`3oool00`0007l0oooo
0?ooo`1E0?ooo`<00000`@3oool00<L0oooo00H0_g@007Boo`3oool0oono07@00018W=l30?ooo`03
0=nLB00007@0_ooo00T0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1G0?ooo`03
0000003oool0oooo05L0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1G0?ooo`03
0000003oool0oooo0;l0oooo00360?ooo`070?oog`2LB000B9cO0?ooo`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00T0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`2o0?oo
o`00aP3oool01`3OW4P0000007Boo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`09
0?ooool00000a00000310?ooo`00`P3oool02`3Ogml0oooo0?oo_`1d000000000000M02oool0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0o`3ooooo0?ooohl0oooo00320?ooo`060=nL
B000000000000000000004P0W=oo0P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
0?l0ooooo`3ooon?0?ooo`00`P3oool0103ookl0M00000000018W=l40?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0o`3ooooo0?ooohl0oooo00330?ooo`030;md001d_ol0oooo00D0
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`3o0?ooool0ooooS`3oool00<<0oooo
00H0omnL04P00000000000000000B02Lgol20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`3o0?ooool0ooooT03oool00<@0oooo00D0_g@0000000000000000004RLg`030?ooo`050?oo_`1d
000000000000M02oool0o`3ooooo0?oooi40oooo003o0?ooool0ooooo`3ooomS0?ooo`00o`3ooooo
0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?ooool0ooooo`3ooomS0?ooo`00
o`3ooooo0?ooool0ooooH`3oool00?l0ooooo`3ooooo0?ooof<0oooo003o0?oooeD0oooo00<0oooO
09a8000000001P0000001P0007@0_ooo0?ooo`3oool0oooO09a80080000000L004RL09ad00000000
07Bo0?oo_`2o_kl0M0000080000000D0M;oo0?ooo`3oool0oooo0;md0002000000060000M02oool0
oooo0?ooo`3ooml0W4P00P0000001P0004P0W=oo0?ooo`3oool0oooO09a80080000000D004RL09ad
0000000007Bo0=nLB002000000060000M02LW4P000000000000007@0_ooo203oool00`3ogi`0M9cO
0?ooo`020?ooo`030=nLM02Lgol0oooo0080oooo00X0gia8000000000000001d0;ooo`3ogi`0B000
00000018W=l0gia80P0000000`0004P0W=oO09a8000200000003001d_`3oool0oooo0080oooo00D0
oono07@000000000000004RLg`030?ooo`060=nLB0000000000000000000B9`0_ia80P0000001P00
07@0M7@0000000000000001809cOo`L0oooo00<0oooO09a8000000000P0000001@0004P0W=oo0?oo
o`3oool0okmd0080000000@0001d0;ooo`3oool0_g@00P0000001P0007@0_ooo0?ooo`3ookl0_kno
07@00080000000<0M;oo0?ooo`3OW4P00P000000100007@0_ono0;no_`1d00020000000907Boo`3o
ool0oooo0?oOW0180000000004RLg`3ooml0W4P00080000000D004RL09ad0000000007Bo0;md0002
000000030000M02ooml0W4P00080000000@004RL0=ooo`3oool0_g@00P0000000`0004P0W=oo0?oo
o`020?ooo`050;md0000000000000000M02oool0o`3oooln0?ooo`00o`3ooomG0?ooo`030?oOW018
0000B9cO00<0oooo03D0omnL04P000000000M;oo0?ooo`3o_g@0000007Boo`3oool0omnL04P00000
M;l0okno0;oo_`1d04P0W=oo0?ooo`3ookl0M01809cOo`3oool0_g@000000000M;l0oooo0?oOW018
07@0_ooo0?ooo`2oM00004RL0=ooo`3oool0omnL04P0001d_ol0oooo0?noM0000000M;oo0?ooo`3o
gi`0B000001d_`3o_kl0_ooo0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`090?ooo`070;md0000
07@0_ooo0?ooo`3ookl0M000001d_`040?ooo`0@0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`3O
_kl0gooo0=nLB00007@0_ooo0?ooo`3ooml0W4P0001d_`H0oooo00<0oono07@0B02Lgol01@3oool0
0`3OW4P0000004RLg`020?ooo`070=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`0:0?ooo`030?oO
W0180000B9cO00<0oooo00@0oono07@0000004P0W=oo0P3oool0;P2oM0000000001d_`3oool0omnL
04P0M02oool0oono07@0B02Lgol0oooo0?oo_`1d04P0W=oo0?ooo`3OW4P0001d0;ooo`3ookl0M018
09cOo`3oool0oono07@0B02Lgol0oooo0=nLB00007@0_ooo0=no_`3O_g@0000007Boo`3oool0omnL
04P00000M;l0okno0;ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l20?ooo`0?0;md00000000
04RL0=ooo`3ooml0W4Qd0;ooo`3oool0_g@000000000M;l0oooo0?oOW01807@0_ooo0?l0oooo?@3o
ool00?l0ooooE`3oool00`3ookl0M00004RLg`040?ooo`0>0?noM000000004RL0=ooo`3o_g@00018
09cOo`3oool0oooO09a80000M;l0oooo0?oo_`2o_ol20?ooo`060?noM0000000M;oo0?oOW0180000
B9cO0`3oool01P3ogi`0M9cO0?ooo`2oM00004RL0=ooo`80oooo00X0okmd0000M02oool0okmd0000
B02Lgol0oooo0?oog`2LB00007Bo0P3oool01`3OW4P0001d0;ooo`3oool0gia80000M02oool0203o
ool02@3ogi`0B00004QdM01dW=l0oooo0?noM00004P0B4Q809cOo`030?ooo`070=nLB00007@0_ooo
0?ooo`3OW4P004RL0=ooo`020?ooo`070=nLB00007@0_ooo0?ooo`3ooml0W4P0001d_`070?ooo`03
0?noM000M;l0oooo00<0oooo00@0oooO09a8000007@0_ooo0P3oool01`3OW4P0001d0;ooo`3oool0
gia80000M02oool02P3oool00`3ookl0M00004RLg`020?ooo`040?oog`2LB00004RL0=ooo`80oooo
00<0omnL04P00018W=l00`3oool0103ogi`0M9cO0?oo_`2o_ol20?ooo`090?noM0000000M;oo0?oo
o`3OW4P0001d0;ooo`3ookl0_koo0080oooo00l0okmd0000001d_ol0oooo0=nLB000B9`0gooo0?oo
o`3o_g@0001809cOo`3oool0oooO09a80000M;l00P3oool02`3OW4P004RL0=ooo`3oool0oooO09a8
0000M;l0oooo0?oOW018000007Bo00<0oooo00D0omnL09boo`3ogi`0B00004RLg`030?ooo`030?oO
W01dW=l0oooo0?l0oooo?03oool00?l0ooooE`3oool00`3ookl0M00004RLg`040?ooo`0;0?oOW018
0000001d0;ooo`3oool0okmd00000000M;l0oooO09a80000M;l00`3oool0203ookl0M00000000000
07@0_ooo0?noM00004P0W=oo1P3oool00`2oM00004RL0=ooo`020?ooo`0:0?oOW0180000M;oo0?oo
o`3o_g@00000001d_`3ooml0W4P0001d_`80oooo00L0gia80000M02oool0oooo0=nLB00007@0_ooo
00P0oooo00T0gia80000M02oool0_gAd0;oog`2LB00007Bo0?oOW01dW=l00`3oool01`3OW4P0001d
0;ooo`3oool0gia80018W03Oool00P3oool01`3OW4P0001d0;ooo`3oool0oooO09a80000M;l01`3o
ool00`3ookl0M01d0;ooo`030?ooo`030?oOW0180000B9cO00<0oooo00L0gia80000M02oool0oooo
0=nLB00007@0_ooo00X0oooo00<0oono07@00018W=l00P3oool00`3OW4P0001809cOo`030?ooo`03
0?noM00004P0W=oo00L0oooo00T0oono07@000000000001d0;ooo`3oool0gia80000M02oool00P3o
ool02@3ookl0M0000000000007@0_ooo0?ooo`3OW4P004RL0=ooo`020?ooo`060?noM000000007Bo
0?oog`2LB00007Bo0P3oool02`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00004P0
W=oo00D0oooo00<0okmd0000B02Lgol0o`3ooom20?ooo`00o`3ooomG0?ooo`030?oo_`1d0000B9cO
00@0oooo00<0okmd00000000M;l0103oool0103o_g@0000000000000M;l20?ooo`040;md00000000
000004RLg`80oooo00<0okmd0000M02oool01P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW018
0000M;oo00<0oooo00@0okmd00000000000007Bo0P3oool01`3OW4P0001d0;ooo`3oool0gia80000
M02oool01`3oool02`3ookl0M00004RLg`3oool0omnL07AdM01804P0W=oo0?ooo`2oM7@0_ooo0080
oooo00L0gia80000M02oool0oooo0=nLB000B9`0gooo0080oooo00L0gia80000M02oool0oooo0?oo
g`2LB00007Bo00P0oooo00<0_g@00000000000000`0000000`0004P0W=oo0?ooo`020?ooo`070=nL
B00007@0_ooo0?ooo`3OW4P0001d0;ooo`0:0?ooo`070?oo_`1d0000B9cO0?ooo`3o_g@0000007Bo
o`040?ooo`030?noM00007@0_ooo00H0oooo00@0_g@0000000000000B9cO0`3oool01@3OW4P0001d
0;ooo`3oool0_g@00080000000L0B9cO0?ooo`3oool0oooo0=nLB000B9`0gooo00@0oooo00@0okmd
00000000000007Bo0P3oool02`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00007@0
_ooo00D0oooo00<0okmd0000M02oool0o`3ooom20?ooo`00o`3ooomG0?ooo`030?oo_`1d0000B9cO
00<0oooo0100gia80000000004P0W=oo0?ooo`3ookl0M00007Boo`3oool0oooO09a80000M;l0oooo
0?oOW0180000B9cO0P3oool0102oW9`0gooo0?oOW01800050000000504RLg`3oool0_g@00018W03O
ool00P3oool02P3o_g@0001809cOo`3ookl0M00007Boo`3oool0oooO09a80000M;l20?ooo`070=nL
B00007@0_ooo0?ooo`3OW4P0001d0;ooo`070?ooo`070=nLB00004P0W=oo0?ooo`3ooml0W4P0001d
_`020?ooo`0;0?noM018W=l0oooo0?ooo`3OW4P0001d0;ooo`3oool0gia80018W03Oool00P3oool0
1`3OW4P0001d0;ooo`3oool0oooO09a80000M;l0203oool0203ogi`0B4RL0=ooo`3oool0oooO09a8
0000B9`0gooo0`3oool01`3OW4P0001d0;ooo`3oool0gia80000B02Lgol02P3oool00`3ookl0M000
00000002000000030000M02oool0oooo00<0oooo00<0omnL04P000000000100000001018W=l0omnL
04P00018W=l20?ooo`090;nLW03Oool0oooo0=nLB00007@0_ooo0?oOW0180000B9cO0080oooo00h0
_ibL0=ooo`3oool0gia80018W03Oool0oooo0?oo_`1d0000M;oo0?ooo`3ooml0W4P0001d_`80oooo
00/0_g@00000M02oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0B01809cOo`020?ooo`050?noM000
04P0W=oo0?oOW01800001@0000000`18W=l0oooo0?ooo`3o0?oooc/0oooo003o0?oooeL0oooo00<0
oono07@000000000100000005P1d_ol0oooo0?ooo`3oool0oono07@00018W=l0oooo0?oo_`1d0000
07Bo0?ooo`3ookl0M01d0;ooo`3oool0okmd0018W03Oool0oooo0;md0018W=l20?ooo`0C0=nLB000
04P0W=oo0?ooo`2oM00000000018W03Oool0oono07@00000B9`0gooo0?oo_`1d0000B9cO0?ooo`3o
okl0M000001d_`020?ooo`070=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`060?ooo`030?oo_`1d
000007Bo0080oooo00<0omnL04P0001d_ol00P3oool02`3ooml0W4Qd0;oog`2LB00000000000M02o
ool0oooo0=nLB000B9`0gooo0080oooo00L0gia80000000004P0W=oo0?oo_`1d0000B9cO00P0oooo
00L0oooO09a8001d_ol0oooo0?oOW0180000B9cO0080oooo00`0oooO09a800000000001d0;oog`2L
B00000000000B018M9`0gkmd0000M02oool70?ooo`030?oo_`1d0000B9cO0080oooo00<0_g@00000
0000M;l0103oool0;P2oM000B9cO0?ooo`3oool0gia80000B02Lgol0oono07@0M02oool0oooo0?no
M000B9`0gooO09a800000000001d0;ooo`3ookl0M01d0;ooo`3oool0okmd0018W03Oool0oooo0=nL
B000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0oono07@00000M;l0oooO09a80000000004Q804P0
B02Lgol0oono07@00018W=l20?ooo`0:0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3oool0_g@0
04RLg`80oooo00<0gia80000B02Lgol0o`3ooolm0?ooo`00o`3ooomG0?ooo`030?oo_`1d0000B9cO
00<0oooo00@0okmd00000000B9`0gooo0P3oool01P3ookl0M000000000000000001d0;ooo`80oooo
00H0okmd00000000000007Bo0;nLW03Oool20?ooo`050=nLB000000000000000B02Lgol00P3oool0
1`2oM00004RL0=oOW018000000000018W03Oool00P3oool01P3ookl0M000000000000000001d0;oo
o`80oooo00L0gia80000M02oool0oooo0=nLB00007@0_ooo00D0oooo00H0oooO09a8000000000000
04RLg`3o_g@30000000:04RLg`3oool0_g@000000000B9`0gooo0;md000007@0_ooo0;md00<00000
00T004RL0=ooo`3OW4P0001d0;oog`2LB00000000000M02oool02@3oool01P3o_g@004RL0=ooo`3O
W4P0001809cOo`@0oooo00X0_g@00000M02oool0oooo0;md000007@0_mnL04P0000007@0_ooo1`3o
ool00`3ookl0M00004RLg`030?ooo`030;md00000000M;oo00@0oooo00D0gia80000000000000018
09cOo`020?ooo`0<0?noM00000000000001d_`2oW9`0gooo0?ooo`2oM000001d0;ooo`3oool0okmd
0P0000001000M;l0_ibL0=ooo`2oM003000000040018W03Oool0oono07@00080000000/0001d0;oo
o`3oool0oooo0;md0000B9`0gooO09a800000000001d0;ooo`030?ooo`050=nLB000000000000000
B02Lgol00`3oool01@3OW4P000000000000004P0W=oo0?l0oooo?P3oool00?l0ooooE`3oool00`3o
okl0M00004RLg`040?ooo`030;md00000000M;oo01T0oooo00<0_g@00018W03Oool03P3oool01`3O
W4P0001d0;ooo`3oool0gia80000M02oool0603oool00`2oM00004RL0=ooo`020?ooo`030=nLB000
07@0_ooo00d0oooo00H0oono07@0B02Lgml0W4P00018W03OoolE0?ooo`030?oo_`1d0000B9cO00<0
oooo00<0okmd00000000M;l07@3oool00`2oM00004RL0=ooo`3o0?oooel0oooo003o0?oooeL0oooo
00<0oono07@00018W=l0103oool00`2oM000000007Boo`0I0?ooo`030;md0000B9`0gooo00h0oooo
00L0gia80000M02oool0oooo0=nLB00007@0_ooo01P0oooo00<0oono07A8W03Oool00P3oool00`3O
W4P0001d0;ooo`0>0?ooo`040=nLB000B4P0B00004RLgaH0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0000004RLg`0M0?ooo`030?oo_`1dB9`0gooo0?l0ooooG`3oool00?l0ooooE`3oool00`3o
gi`0B00004RLg`030?ooo`040=nLB000000004RL0=oooaT0oooo00<0_g@00018W03Oool03P3oool0
1`3OW4P0001d0;ooo`3oool0gia80000M02oool0603oool00`3ooml0W7BL0=ooo`020?ooo`030=nL
B00007@0_ooo00h0oooo00@0omnL04P0000004P0W=oo5P3oool00`3ogi`0B00004RLg`020?ooo`04
0?noM0000000001d0;oooad0oooo00<0oooO09adW03Oool0o`3ooomO0?ooo`00o`3ooomE0?ooo`03
0?oog`2LB000000000H0000000<0B9cO0?ooo`3oool05`3oool01@3ookl0M00000000000B9`0gooo
00`0oooo00T0oooO09a800000000001d0;oog`2LB00000000000M02oool0503oool00`3OW4P00018
09cOo`040?ooo`050?oog`2LB00000000000M02oool03`3oool00`2oM00004RL0=ooo`050?ooo`03
0=nLB00004P0W=oo00`0oooo00<0oooO09a8000000001@0000000`0007@0_ooo0?ooo`0B0?ooo`03
0=nLB00004P0W=oo0?l0ooooJP3oool00?l0ooooN@3oool0103ooml0W4P00018W03Oool>0?ooo`07
0;md000007@0_ooo0?ooo`2oM000001d0;ooo`0D0?ooo`030=nLB00004P0W=oo00H0oooo00<0_g@0
0000M02oool05`3oool00`3OW4P0001809cOo`0Y0?ooo`030=nLB00004P0W=oo0?l0ooooJP3oool0
0?l0ooooo`3ooooo0?ooof<0oooo003o0?ooool0ooooo`3ooomS0?ooo`00o`3ooooo0?ooool0oooo
H`3oool00?l0ooooo`3ooooo0?ooof<0oooo0000\
\>"],
  ImageRangeCache->{{{0, 863.375}, {171.875, 0}} -> {-122.201, -29.2633, \
0.833085, 0.833085}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Effect of resistance on velocity", "Subsection"],

Cell["\<\
Similarly, it is useful to define a function which, given initial conditions \
and parameters, graphically compares velocity components with and without air \
resistance.  The velocities are obtained simply by differentiating positions \
with respect to time.  For the purposes of illustration, we continue these \
trajectories for large times beyond the point where the ball returns to the \
ground.  We might imagine that the edge of the field is a cliff and that Wile \
E. Coyote is attempting to catch a well-struck ball.\
\>", "Text"],

Cell[BoxData[
    \(\(plotResistedVsFreeVelocities[\(x\_0\) : _, \ \(y\_0\) : _, \(v\_0\) : \
_, \(\[Theta]\_0\) : _, parameters_] := \n
        Plot[Evaluate[{D[
                resistedTrajectory[x\_0, y\_0, v\_0, \[Theta]\_0, 
                  parameters], t], 
              D[freeTrajectory[x\_0, y\_0, v\_0, \[Theta]\_0, parameters], 
                t]\ }], {t, 0, 20}, 
          PlotRange \[Rule] {{0, 20}, {\(-100\), 50}}, 
          AspectRatio \[Rule] 1. , Frame \[Rule] True, 
          GridLines \[Rule] Automatic, 
          FrameLabel \[Rule] {"\<t (s)\>", "\<v (m/s)\>", "\<Velocity \
Components\>", "\< \>"}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotResistedVsFreeVelocities[0, \ 0, 50, 35\ Degree, 
        baseballParameters];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-2.77556e-017 0.05 0.666667 0.00666667 [
[.125 -0.0125 -9.5 -15.6875 ]
[.125 -0.0125 9.5 0 ]
[.25 -0.0125 -5 -15.6875 ]
[.25 -0.0125 5 0 ]
[.375 -0.0125 -9.5 -15.6875 ]
[.375 -0.0125 9.5 0 ]
[.5 -0.0125 -8 -15.6875 ]
[.5 -0.0125 8 0 ]
[.625 -0.0125 -12.5 -15.6875 ]
[.625 -0.0125 12.5 0 ]
[.75 -0.0125 -8 -15.6875 ]
[.75 -0.0125 8 0 ]
[.875 -0.0125 -12.5 -15.6875 ]
[.875 -0.0125 12.5 0 ]
[1 -0.0125 -8 -15.6875 ]
[1 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -11.3125 -31.375 ]
[.5 -0.0125 11.3125 -15.6875 ]
[-0.0125 .13333 -24.3125 -7.84375 ]
[-0.0125 .13333 0 7.84375 ]
[-0.0125 .26667 -24.3125 -7.84375 ]
[-0.0125 .26667 0 7.84375 ]
[-0.0125 .4 -24.3125 -7.84375 ]
[-0.0125 .4 0 7.84375 ]
[-0.0125 .53333 -24.3125 -7.84375 ]
[-0.0125 .53333 0 7.84375 ]
[-0.0125 .66667 -10 -7.84375 ]
[-0.0125 .66667 0 7.84375 ]
[-0.0125 .8 -16 -7.84375 ]
[-0.0125 .8 0 7.84375 ]
[-0.0125 .93333 -16 -7.84375 ]
[-0.0125 .93333 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .5 -40 -19.25 ]
[-0.01875 .5 -24.3125 19.25 ]
[ 0 1 .125 0 ]
[.5 1.0125 -54.1563 0 ]
[.5 1.0125 54.1563 15.6875 ]
[ 1 0 .125 0 ]
[1.0125 .5 0 -3.5 ]
[1.0125 .5 15.6875 3.5 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.125 0 m
.125 1 L
s
.25 0 m
.25 1 L
s
.375 0 m
.375 1 L
s
.5 0 m
.5 1 L
s
.625 0 m
.625 1 L
s
.75 0 m
.75 1 L
s
.875 0 m
.875 1 L
s
0 .13333 m
1 .13333 L
s
0 .26667 m
1 .26667 L
s
0 .4 m
1 .4 L
s
0 .53333 m
1 .53333 L
s
0 .66667 m
1 .66667 L
s
0 .8 m
1 .8 L
s
0 .93333 m
1 .93333 L
s
0 g
.125 0 m
.125 .00625 L
s
gsave
.125 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2.5) show
1.000 setlinewidth
grestore
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.375 0 m
.375 .00625 L
s
gsave
.375 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(7.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.625 0 m
.625 .00625 L
s
gsave
.625 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(12.5) show
1.000 setlinewidth
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.875 0 m
.875 .00625 L
s
gsave
.875 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(17.5) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.125 Mabswid
.025 0 m
.025 .00375 L
s
.05 0 m
.05 .00375 L
s
.075 0 m
.075 .00375 L
s
.1 0 m
.1 .00375 L
s
.15 0 m
.15 .00375 L
s
.175 0 m
.175 .00375 L
s
.2 0 m
.2 .00375 L
s
.225 0 m
.225 .00375 L
s
.275 0 m
.275 .00375 L
s
.3 0 m
.3 .00375 L
s
.325 0 m
.325 .00375 L
s
.35 0 m
.35 .00375 L
s
.4 0 m
.4 .00375 L
s
.425 0 m
.425 .00375 L
s
.45 0 m
.45 .00375 L
s
.475 0 m
.475 .00375 L
s
.525 0 m
.525 .00375 L
s
.55 0 m
.55 .00375 L
s
.575 0 m
.575 .00375 L
s
.6 0 m
.6 .00375 L
s
.65 0 m
.65 .00375 L
s
.675 0 m
.675 .00375 L
s
.7 0 m
.7 .00375 L
s
.725 0 m
.725 .00375 L
s
.775 0 m
.775 .00375 L
s
.8 0 m
.8 .00375 L
s
.825 0 m
.825 .00375 L
s
.85 0 m
.85 .00375 L
s
.9 0 m
.9 .00375 L
s
.925 0 m
.925 .00375 L
s
.95 0 m
.95 .00375 L
s
.975 0 m
.975 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -72.3125 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(t) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
69.313 15.813 moveto
(H) show
73.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(s) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
77.813 15.813 moveto
(L) show
81.625 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .13333 m
.00625 .13333 L
s
gsave
-0.0125 .13333 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
0 .26667 m
.00625 .26667 L
s
gsave
-0.0125 .26667 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
0 .4 m
.00625 .4 L
s
gsave
-0.0125 .4 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .53333 m
.00625 .53333 L
s
gsave
-0.0125 .53333 -85.3125 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .66667 m
.00625 .66667 L
s
gsave
-0.0125 .66667 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .8 m
.00625 .8 L
s
gsave
-0.0125 .8 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .93333 m
.00625 .93333 L
s
gsave
-0.0125 .93333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .03333 m
.00375 .03333 L
s
0 .06667 m
.00375 .06667 L
s
0 .1 m
.00375 .1 L
s
0 .16667 m
.00375 .16667 L
s
0 .2 m
.00375 .2 L
s
0 .23333 m
.00375 .23333 L
s
0 .3 m
.00375 .3 L
s
0 .33333 m
.00375 .33333 L
s
0 .36667 m
.00375 .36667 L
s
0 .43333 m
.00375 .43333 L
s
0 .46667 m
.00375 .46667 L
s
0 .5 m
.00375 .5 L
s
0 .56667 m
.00375 .56667 L
s
0 .6 m
.00375 .6 L
s
0 .63333 m
.00375 .63333 L
s
0 .7 m
.00375 .7 L
s
0 .73333 m
.00375 .73333 L
s
0 .76667 m
.00375 .76667 L
s
0 .83333 m
.00375 .83333 L
s
0 .86667 m
.00375 .86667 L
s
0 .9 m
.00375 .9 L
s
0 .96667 m
.00375 .96667 L
s
.25 Mabswid
0 0 m
0 1 L
s
gsave
-0.01875 .5 -112.406 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
80.25 11.8438 translate
90 rotate
-80.25 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(v) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
72.000 15.813 moveto
(H) show
75.813 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.188 15.813 moveto
(\\220) show
89.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(s) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
93.688 15.813 moveto
(L) show
97.500 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .99375 m
0 1 L
s
.125 .99375 m
.125 1 L
s
.25 .99375 m
.25 1 L
s
.375 .99375 m
.375 1 L
s
.5 .99375 m
.5 1 L
s
.625 .99375 m
.625 1 L
s
.75 .99375 m
.75 1 L
s
.875 .99375 m
.875 1 L
s
.125 Mabswid
.025 .99625 m
.025 1 L
s
.05 .99625 m
.05 1 L
s
.075 .99625 m
.075 1 L
s
.1 .99625 m
.1 1 L
s
.15 .99625 m
.15 1 L
s
.175 .99625 m
.175 1 L
s
.2 .99625 m
.2 1 L
s
.225 .99625 m
.225 1 L
s
.275 .99625 m
.275 1 L
s
.3 .99625 m
.3 1 L
s
.325 .99625 m
.325 1 L
s
.35 .99625 m
.35 1 L
s
.4 .99625 m
.4 1 L
s
.425 .99625 m
.425 1 L
s
.45 .99625 m
.45 1 L
s
.475 .99625 m
.475 1 L
s
.525 .99625 m
.525 1 L
s
.55 .99625 m
.55 1 L
s
.575 .99625 m
.575 1 L
s
.6 .99625 m
.6 1 L
s
.65 .99625 m
.65 1 L
s
.675 .99625 m
.675 1 L
s
.7 .99625 m
.7 1 L
s
.725 .99625 m
.725 1 L
s
.775 .99625 m
.775 1 L
s
.8 .99625 m
.8 1 L
s
.825 .99625 m
.825 1 L
s
.85 .99625 m
.85 1 L
s
.9 .99625 m
.9 1 L
s
.925 .99625 m
.925 1 L
s
.95 .99625 m
.95 1 L
s
.975 .99625 m
.975 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
gsave
.5 1.0125 -115.156 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Velocity) show
106.688 15.813 moveto
(Components) show
167.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 0 m
1 0 L
s
.99375 .13333 m
1 .13333 L
s
.99375 .26667 m
1 .26667 L
s
.99375 .4 m
1 .4 L
s
.99375 .53333 m
1 .53333 L
s
.99375 .66667 m
1 .66667 L
s
.99375 .8 m
1 .8 L
s
.99375 .93333 m
1 .93333 L
s
.125 Mabswid
.99625 .03333 m
1 .03333 L
s
.99625 .06667 m
1 .06667 L
s
.99625 .1 m
1 .1 L
s
.99625 .16667 m
1 .16667 L
s
.99625 .2 m
1 .2 L
s
.99625 .23333 m
1 .23333 L
s
.99625 .3 m
1 .3 L
s
.99625 .33333 m
1 .33333 L
s
.99625 .36667 m
1 .36667 L
s
.99625 .43333 m
1 .43333 L
s
.99625 .46667 m
1 .46667 L
s
.99625 .5 m
1 .5 L
s
.99625 .56667 m
1 .56667 L
s
.99625 .6 m
1 .6 L
s
.99625 .63333 m
1 .63333 L
s
.99625 .7 m
1 .7 L
s
.99625 .73333 m
1 .73333 L
s
.99625 .76667 m
1 .76667 L
s
.99625 .83333 m
1 .83333 L
s
.99625 .86667 m
1 .86667 L
s
.99625 .9 m
1 .9 L
s
.99625 .96667 m
1 .96667 L
s
.25 Mabswid
1 0 m
1 1 L
s
gsave
1.0125 .5 -56.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
64.5 11.8438 translate
90 rotate
-64.5 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
66.000 15.813 moveto
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .66667 m
1 .66667 L
s
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
0 .93972 m
.01975 .91469 L
.04057 .89375 L
.08481 .86119 L
.12477 .83994 L
.1673 .82146 L
.20824 .80538 L
.25176 .78902 L
.29369 .77382 L
.33402 .75995 L
.37693 .74631 L
.41825 .73448 L
.45798 .7244 L
.50028 .71505 L
.54099 .70733 L
.58428 .70036 L
.62597 .69472 L
.66607 .69015 L
.70875 .68608 L
.74984 .68282 L
.7935 .67994 L
.83557 .67765 L
.87605 .67582 L
.9191 .6742 L
.96056 .67292 L
1 .6719 L
s
0 .85786 m
.04057 .77706 L
.08481 .7075 L
.12636 .65165 L
.16632 .6036 L
.20885 .55789 L
.2498 .51957 L
.29331 .48536 L
.33524 .45874 L
.37557 .43852 L
.41848 .42198 L
.4598 .41004 L
.49953 .40147 L
.54183 .39475 L
.58254 .39005 L
.62583 .38646 L
.66752 .384 L
.70762 .3823 L
.7503 .38101 L
.79139 .38013 L
.81236 .37978 L
.83505 .37947 L
.87712 .37902 L
.89805 .37886 L
.92017 .37871 L
.96004 .3785 L
1 .37835 L
s
0 .93972 m
.04057 .93972 L
.08481 .93972 L
.12636 .93972 L
.16632 .93972 L
.20885 .93972 L
.2498 .93972 L
.29331 .93972 L
.33524 .93972 L
.37557 .93972 L
.41848 .93972 L
.4598 .93972 L
.49953 .93972 L
.54183 .93972 L
.58254 .93972 L
.62583 .93972 L
.66752 .93972 L
.70762 .93972 L
.7503 .93972 L
.79139 .93972 L
.83505 .93972 L
.87712 .93972 L
.9176 .93972 L
.96065 .93972 L
1 .93972 L
s
0 .85786 m
.04057 .80485 L
.08481 .74704 L
.12636 .69275 L
.16632 .64054 L
.20885 .58496 L
.2498 .53146 L
.29331 .4746 L
.33524 .41981 L
.37557 .36711 L
.41848 .31104 L
.4598 .25705 L
.49953 .20514 L
.54183 .14987 L
.58254 .09667 L
.62583 .04011 L
s
.62583 .04011 m
.65652 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00j@3oool00`3ooml0W4Q809cOo`070?ooo`030?oOW018B9`0gooo0;X0oooo003Y0?ooo`03
0;mdB02Lgol0oooo00P0oooo00<0omnL04Qd_`3oool0^@3oool00>P0oooo00<0gia804RLg`3oool0
2P3oool00`3OW4P0B9cO0?ooo`2h0?ooo`00i`3oool00`3ookl0M01d0;ooo`0;0?ooo`030?oo_`1d
04P0W=oo0;P0oooo003M0?ooo`040?oOW0180000000004RLg`H0oooo00<0gia80018W03Oool00P3o
ool00`3ookl0_kno07@000020000000707Boo`3oool0oooo0?ooo`3ooml0W4P004RLg`2h0?ooo`00
g@3oool01@3OW4P0001d0;ooo`3O_kl0gooo00D0oooo00<0_g@0001d_`3oool00P3oool01`3ookl0
M01809cOo`3oool0oono07@0B02Lgol00`3oool00`3OW4P004RL0=ooo`2g0?ooo`00g@3oool00`3O
W4P004RL0=ooo`060?ooo`030?oo_`1d0000B9cO00<0oooo00L0oono0;noo`3oool0oooo0?noM000
0000M;oo00<0oooo00<0gia80000M02oool0]`3oool00=d0oooo00<0gia80018W03Oool01P3oool0
0`3ookl0M00007Boo`050?ooo`050?oo_`1d000000000000M02oool00`3oool00`3o_g@0001d0;oo
o`2g0?ooo`00g@3oool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d0000M;oo00@0oooo00@0_g@0
000000000000B9cO1@3oool00`3o_g@0001809cOo`2g0?ooo`00g@3oool00`3OW4P004RL0=ooo`06
0?ooo`030?oo_`1d0000M;oo00<0oooo00<0omnL04P00018W=l00P3oool00`2oW9`0gooo0?ooo`02
0?ooo`030?noM00007@0_ooo0;L0oooo003M0?ooo`030=nLB000B9`0gooo00H0oooo00<0oono07@0
0018W=l00`3oool01`3ookl0M01d0;ooo`3oool0okmd0018W03Oool00`3oool00`3OW4P0001d0;oo
o`2g0?ooo`00g03oool01P2oM000000000000000000004RL0=ooo`@0oooo00<0oooO09a80018W=l0
103oool01P3o_g@0000000000000M;l0_ibL0=ooo`<0oooo00<0gia80018W03Oool0]`3oool00=d0
oooo00<0_g@00018W03Oool01`3oool00`3OW4P004RL0=ooo`0;0?ooo`030?oog`2LB000B9cO0;P0
oooo003M0?ooo`030?oo_`1dB9`0gooo00L0oooo00<0oono07@0B02Lgol02`3oool00`3ookl0M018
09cOo`2h0?ooo`00g@3oool00`3ooml0W7BL0=ooo`080?ooo`030=nLB018W=l0oooo00X0oooo00<0
gia804RLg`3oool0^03oool00>T0oooo00<0_g@004RLg`3oool0203oool00`3o_g@007Bo0?ooo`2i
0?ooo`00jP3oool00`3OW7@0W=oo0?ooo`060?ooo`030?oo_`1dM;l0oooo0;X0oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo001Q0?ooo`03
0;md00000000000000@0000000T0B9cO0?ooo`3oool0okmd0000001d_ol0oooo0?oog`2LB0000`00
00000`1d_ol0oooo0?ooo`0K0?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool07P3o
ool00`3OW4P0B9cO0?ooo`040?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0
oooo0?ooo`0H0?ooo`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B00000000000
07@0_ooo01H0oooo00P0_g@000000000000000000018W03Oool0oooo0;md00H0000000T0B9cO0?oo
o`3oool0okmd0000001d_ol0oooo0?oog`2LB0000`0000000`1d_ol0oooo0?ooo`0D0?ooo`090;md
00000000000000000000B9`0gooo0?ooo`3ooml0W4P000<0000000<0M;oo0?ooo`3oool0503oool0
1P2oM000000000000000000004RL0=ooo`@0oooo00<0gia804RLg`3oool0103oool01P3o_g@00000
07Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool04P3oool00`2oM0000000000000040000
000904RLg`3oool0oooo0?ooo`3ogi`0B0000000000007@0_ooo01H0oooo001R0?ooo`030=nLB000
0000000000<0000000X004RL0=ooo`3oool0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0P3oool0
0`2oM000M;oo0?ooo`0K0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool07P3oool0
0`3ogi`0B4RL0=ooo`040?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;ooo`020?ooo`030;md
001d_ol0oooo01T0oooo00<0omnL04P0001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool05P3oool00`3ogi`0B00007Boo`040?ooo`030=nLB0000000000000<0000000X004RL0=oo
o`3oool0okmd0000001d_ol0oooo0?oog`2LB7@0_ooo0P3oool00`2oM000M;oo0?ooo`0E0?ooo`03
0?oOW0180000M;oo00<0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0E0?ooo`03
0?oOW0180000M;oo00H0oooo00<0omnL04Q8W03Oool0103oool01`3o_g@0000007Boo`3oool0oooO
09a8M02oool00P3oool00`2oM000M;oo0?ooo`0C0?ooo`030=nLB0000000000000<0000000X004RL
0=ooo`3oool0oono07@0001d_ol0oooo0?noM00007@0_ooo5@3oool006<0oooo00<0gia804RLg`3o
ool00P3oool00`3OW7@0W=oo0?ooo`090?ooo`030?oog`2LB7@0_ooo0200oooo00<0oooO09a8M02o
ool07P3oool00`3ooml0W4Q809cOo`0=0?ooo`030?oog`2LB7@0_ooo01T0oooo00<0omnL04P0001d
_ol0103oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01D0oooo00<0omnL04P0001d
_ol01@3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00T0oooo00<0oooO09a8M02o
ool05@3oool00`3ogi`0B00007Boo`080?ooo`030?oog`2LB7@0_ooo01D0oooo00<0omnL04P0001d
_ol01P3oool00`3ooml0W4Q809cOo`0=0?ooo`030?oog`2LB7@0_ooo01@0oooo00<0gia804RLg`3o
ool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo00<0_g@0001d_`3oool0503oool0
06@0oooo00<0gia807Boo`3oool03P3oool00`2oM000M;oo0?ooo`0P0?ooo`030;md001d_ol0oooo
01h0oooo00<0gia804RLg`3oool03@3oool00`2oM000M;oo0?ooo`0H0?ooo`030?oOW0180000M;oo
00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0E0?ooo`030?oOW0180000M;oo
00H0oooo00<0gia807Boo`3oool03P3oool00`2oM000M;oo0?ooo`0D0?ooo`030?oOW0180000M;oo
00T0oooo00<0_g@007Boo`3oool0503oool00`3ogi`0B00007Boo`070?ooo`030=nLB018W=l0oooo
00d0oooo00<0_g@007Boo`3oool0503oool00`3OW4P0M;oo0?ooo`030?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool0503oool006D0oooo00<0_gA809cOo`3oool0303oool00`3ooml0
W4P004RLg`0P0?ooo`030?oog`2LB000B9cO01l0oooo00<0omnL04Q8W03Oool0303oool00`3ooml0
W4P004RLg`0I0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`0E0?ooo`030?oOW0180000M;oo00L0oooo00<0_gA809cOo`3oool0303oool00`3ooml0
W4P004RLg`0E0?ooo`030?oOW0180000M;oo00P0oooo00<0oooO09a80018W=l05@3oool00`3ogi`0
B00007Boo`070?ooo`030?oOW018B9`0gooo00`0oooo00<0oooO09a80018W=l05P3oool00`2oM4P0
W=oo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0503oool006D0oooo
00<0oooO09a8M02oool0303oool00`3OW4P0000007Boo`0P0?ooo`030=nLB0000000M;oo01l0oooo
00<0oooO09a8B02Lgol0303oool00`3OW4P0000007Boo`0I0?ooo`030?oOW0180000M;oo00<0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0E0?ooo`030?oOW0180000M;oo00L0oooo
00<0oooO09a8M02oool0303oool00`3OW4P0000007Boo`0E0?ooo`030?oOW0180000M;oo00P0oooo
00<0gia80000001d_ol05@3oool00`3ogi`0B00007Boo`070?ooo`030?oog`2LB4P0W=oo00`0oooo
00<0gia80000001d_ol05P3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol0503oool006H0oooo00<0okmd001d_`3oool01`3oool01`3Ogml0oooo0?oo
_`1d000000000000M02oool0703oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0803oool0
0`3OW4P007Bo0?ooo`070?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0I0?ooo`030?oO
W0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0E0?ooo`030?oO
W0180000M;oo00P0oooo00<0okmd001d_`3oool01`3oool01`3Ogml0oooo0?oo_`1d000000000000
M02oool05@3oool00`3ogi`0B00007Boo`040?ooo`070=oOg`3oool0oono07@000000000001d0;oo
o`0E0?ooo`030?oOW0180000M;oo00P0oooo00<0gia8001d_`3oool01`3oool01`3Ogml0oooo0?oo
_`1d000000000000M02oool05`3oool01`3o_g@007Bo0?ooo`3oool0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`0D0?ooo`00IP3oool00`3ookl0M01809cOo`070?ooo`060=nLB00000000000
0000000004P0W=oo7@3oool01P3OW4P00000000000000000001809cOob40oooo00<0oono07@0M02o
ool01`3oool01P3OW4P00000000000000000001809cOoaX0oooo00<0omnL04P0001d_ol00`3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo01D0oooo00<0omnL04P0001d_ol0203oool0
0`3ookl0M01809cOo`070?ooo`060=nLB000000000000000000004P0W=oo5P3oool00`3ogi`0B000
07Boo`040?ooo`060=nLB000000000000000000004P0W=oo5P3oool00`3ogi`0B00007Boo`080?oo
o`030?oo_`1d07@0_ooo00L0oooo00H0gia800000000000000000000B02LgolH0?ooo`070?oo_`1d
04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo01@0oooo001V0?ooo`030?oo
_`1d0000B9cO00L0oooo00@0oono07@000000000B9cO7`3oool0103ookl0M00000000018W=lN0?oo
o`030?oOg`3Oool0oooo00<0oooo00<0_g@007Boo`3oool01P3oool0103ookl0M00000000018W=lL
0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0E
0?ooo`030?oOW0180000M;oo00P0oooo00<0oono07@00018W=l01`3oool0103ookl0M00000000018
W=lH0?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@000000000B9cO603oool00`3ogi`0B000
07Boo`030?ooo`030?oOg`3Oool0oooo00<0oooo00<0_g@007Boo`3oool01P3oool0103ookl0M000
00000018W=lJ0?ooo`070?oo_`1d0000B9cO0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo01@0oooo001Q0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l0203oool00`2o
M000M;oo0?ooo`0P0?ooo`030;md001d_ol0oooo01h0oooo00<0oono07A8W03Oool00`3oool00`3o
_g@007Bo0?ooo`070?ooo`030;md001d_ol0oooo01`0oooo00<0omnL04P0001d_ol0103oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo01D0oooo00<0omnL04P0001d_ol00`3oool00`3o
_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0_g@007Boo`3oool0603oool00`3o
gi`0B00007Boo`050?ooo`030;md001d_ol0oooo01P0oooo00<0omnL04P0001d_ol00`3oool00`3o
okl0M4RL0=ooo`030?ooo`030?noM000M;l0oooo00L0oooo00<0_g@007Boo`3oool05@3oool00`3o
_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0080oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`0D0?ooo`00H@3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo203o
ool01P3ogi`0B000000000000000001809cOoad0oooo00H0omnL04P00000000000000000B02LgolL
0?ooo`030;md00000000000000<0000000<0001d0;ooo`3oool01P3oool01P3ogi`0B00000000000
0000001809cOoaL0oooo00D0oooO09a800000000000007Boo`040?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`0D0?ooo`050?oog`2LB00000000000001d_ol00`3oool0203ooml0W4Q809cO
o`3oool0oono07@0000007@0_ooo203oool01P3ogi`0B000000000000000001809cOoa<0oooo00D0
oooO09a800000000000007Boo`050?ooo`060?oOW018000000000000000004P0W=oo4`3oool01@3o
oml0W4P0000000000000M;oo00@0oooo00<0_g@00000000000000`0000000`0007@0_ooo0?ooo`06
0?ooo`060?oOW018000000000000000004P0W=oo4P3oool0203ooml0W4Q809cOo`3oool0oono07@0
000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool05@3oool00680oooo00H0
oooO09a80000000000000000M02oool:0?ooo`050;md00000000000000000018W=l07P3oool01@2o
M0000000000000000000B9cO01`0oooo00<0okmd000000000000100000000`1d_ol0oooo0?ooo`06
0?ooo`050;md00000000000000000018W=l06@3oool00`3o_g@0000007Boo`050?ooo`050?oo_`1d
000000000000M02oool05`3oool00`3o_g@0000007Boo`040?ooo`060?oog`2LB000000000000000
07@0_ooo2P3oool01@2oM0000000000000000000B9cO01D0oooo00<0okmd0000001d_ol01P3oool0
1@2oM0000000000000000000B9cO01D0oooo00<0okmd0000001d_ol0103oool00`3o_g@000000000
00040000000307Boo`3oool0oooo00H0oooo00D0_g@0000000000000000004RLg`0C0?ooo`060?oo
g`2LB00000000000000007@0_ooo103oool01@3ookl0M0000000000007@0_ooo01H0oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo00120?ooool00000D`00000L
0?ooo`00@P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo0P0000060?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`060?ooo`8000001P3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`060?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`050?ooo`030000003o
ool0oooo0200oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003o
ool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`040?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0oooo
0240oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo
01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`030?ooo`030000003oool0oooo0280oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo02<0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo
00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`020?ooo`030000003oool0oooo02<0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool01@0007l0oooo0?ooo`3oool0000002H0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool0100007l0oooo0?ooo`00000W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool2
000002P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0000000X0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09@3oool3000001`0oooo00120?ooo`030000003oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0000000X0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0000003oool0:03oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo
00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo02P0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo
00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09P3oool00`000000001o0?oo
o`0Y0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?oo
o`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02H0oooo00<00000
0000O`3oool0:@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<00000
0?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0U0?oo
o`030000003oool0001o02X0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool0903oool010000000oooo0?ooo`0007lZ0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02<0oooo00D000000?ooo`3oool0oooo0000O`0Z0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool00P3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000
oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo0280oooo
00<000000?ooo`3oool00P3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool2000002P0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool08@3oool00`000000oooo0?ooo`030?ooo`030000O`3oool0oooo02P0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0U0?ooo`<00000703oool00480oooo00<00000
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0P0?ooo`030000003oool0oooo00@0oooo00<0001o
0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<00000
0?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0O0?oo
o`030000003oool0oooo00D0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0O0?ooo`030000003oool0oooo00D0oooo00<0001o0?ooo`3o
ool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3o
ool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0N0?ooo`030000
003oool0oooo00H0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0M0?ooo`030000003oool0oooo00L0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0L0?ooo`030000003oool0
oooo00P0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0K0?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3oool0:03oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480
oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0K0?ooo`030000003oool0oooo00T0
oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0J0?ooo`030000003oool0oooo00X0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<0
00000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0I0?ooo`030000003oool0oooo00/0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
00000?ooo`3oool06P3oool00480oooo0P00000X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01P0oooo00<00000
0?ooo`3oool0303oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09@3oool3000001`0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool05`3oool00`000000oooo0?ooo`0=0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`03
0000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool05`3oool00`000000oooo0?ooo`0=0?ooo`03
0000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
5P3oool00`000000oooo0?ooo`0>0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool05@3oool00`000000oooo0?ooo`0?0?ooo`030000O`3o
ool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003o
ool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0503oool0
0`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000[0?ooo`050=nLB00000000000
00000018W=l0103oool01@3ogi`0B0000000000007@0_ooo00T0oooo00<000000?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0D0?ooo`030000003oool0oooo0100oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool002X0oooo00L0gia80000001d_ol0oooo0?oog`2LB000B9cO0080oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0C0?oo
o`030000003oool0oooo0140oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool002T0oooo00<0oooO09a80000
M;l00`3oool01`3o_g@004RL0=ooo`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`07
0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool04P3oool00`000000oooo0?ooo`0B
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo000Z0?ooo`030;md0000M;l0oooo0080oooo00H0_g@00018
W03Oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0A0?ooo`030000003oool0oooo01<0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool001`0oooo00<0oooO09a8000000001`0000003`18W=l0oooo0?ooo`3oool0omnL04P0M02oool0
oooo0;md00000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<0
00000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0oooo01@0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
00000?ooo`3oool06P3oool002/0oooo00D0oono07A8B0180000000004RLg`020?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool300000;P0001o00<000000000O`0007l0T@00
07l3000001`0oooo000[0?ooo`040?oOW01800000000001d_`<0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool03`3o
ool00`000000oooo0?ooo`0E0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000Z0?ooo`060?oOW0180000
001d0;ooo`3o_g@007Bo0P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo
00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0>0?ooo`030000003oool0oooo01H0oooo
00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<000000?ooo`3oool06P3oool002X0oooo00<0_g@00000M02oool00P3oool01P2oM000B9cO0?oo
o`3ookl0M00004RLg`<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo00d0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0X0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00
:P3oool00`2oM00004RL0=ooo`020?ooo`070=nLB000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0=0?oo
o`030000003oool0oooo01L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool002X0oooo00L0okmd0000001d
_ol0oooo0?oo_`1d0000M;oo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo
00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo01P0oooo
00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<000000?ooo`3oool06P3oool002/0oooo00D0okmd000000000000000004RLg`040?ooo`050?oo
_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00/0
oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo00X0oooo00<000000?ooo`3oool06P3oool00`0007l0oooo
0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00T0oooo00<0
00000?ooo`3oool06`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool06`3oool00`0007l0oooo0?ooo`0X
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J
0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00P0oooo00<000000?oo
o`3oool0703oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool2000002P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool01`3oool00`000000oooo0?ooo`0M0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0U0?ooo`<00000703oool00480oooo00<000000?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`060?ooo`030000003oool0oooo01h0oooo00<0001o0?ooo`3oool0
:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0
6P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`060?ooo`030000003o
ool0oooo01h0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`050?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0:03oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool0
0480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0oooo
0200oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`030?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0:03oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo
00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo0280oooo
00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`020?ooo`030000003oool0oooo0280oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<00000
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool01@0007l0oooo0?ooo`3oool0000002D0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool0100007l0oooo0?ooo`00000V0?ooo`030000O`3o
ool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003o
ool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`0000009`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool00000
02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool00480oooo0P00000X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`000000oooo02L0
oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02D0
oooo0`00000L0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo02L0oooo
00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09P3oool00`000000001o0?oo
o`0X0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02H0oooo00<00000
0000O`3oool0:03oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0U0?oo
o`030000003oool0001o02T0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool0903oool010000000oooo0?ooo`0007lY0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`03
0000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02<0oooo00D000000?ooo`3oool0oooo0000O`0Y0?ooo`030000O`3oool0oooo
02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo
01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02<0oooo00D000000?ooo`3oool0oooo0000O`0Y0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3o
ool00P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000
003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo0240oooo00<000000?ooo`3oool00`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo0200oooo00<000000?ooo`3oool0103o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`800000:03oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0803oool0
0`000000oooo0?ooo`040?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02D0oooo0`00000L0?ooo`00@P3oool00`00
0000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0O0?ooo`030000003oool0oooo00D0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo00H0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0N0?ooo`030000003oool0oooo00H0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0M0?ooo`030000003oool0oooo00L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0L
0?ooo`030000003oool0oooo00P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J
0?ooo`00:`3oool01@3ogi`0B000000000000000M;oo00@0oooo00D0omnL04P000000000001d0;oo
o`090?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo01/0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo000Z0?ooo`070?oOW0180000B9cO0?ooo`3ookl0M000
07Boo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
01/0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo
01X0oooo000Y0?ooo`030?oog`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo0;md0000B9`0
gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0J0?ooo`030000003oool0
oooo00X0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00:@3oool00`3o
gi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;oo
o`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo01T0oooo00<000000?ooo`3oool02`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo000L0?ooo`030?oog`2LB000000000L0000000H0B9cO
0?ooo`3oool0omnL04P0001d_ol30?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o
_g@0001809cOo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0303oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000Y0?ooo`030?oOW0180000M;oo00<0
oooo00H0_g@00000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00L0oooo0`00002F
0000O`030000000007l0001o0;<0001o0`00000L0?ooo`00:@3oool0303ookl0M00000000018W=l0
oooo0?noM000000007Bo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01`3oool00`00
0000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0G0?ooo`030000003oool0oooo00d0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0J0?ooo`00:P3oool00`2oM00004RL0;nLB002000000060018W03Oool0oooo0?oO
W0180000M;oo0`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01H0oooo00<00000
0?ooo`3oool03P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000Z0?oo
o`030?oOW0180000M;oo00D0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`070?oo
o`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo01D0oooo00<000000?ooo`3oool03`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo000[0?ooo`030=nLB00004P0W=oo00D0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01D0oooo00<000000?ooo`3o
ool03`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000/0?ooo`030?no
M000000007Bo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?oo
o`3oool06P3oool002d0oooo00D0oooO09a800000000001d0;ooo`020?ooo`050?oo_`1d00000000
0000M02oool02@3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0C0?ooo`030000003oool0oooo0140oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0B0?oo
o`030000003oool0oooo0180oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?oo
o`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0A0?ooo`030000003oool0oooo01<0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0A0?ooo`030000
003oool0oooo01<0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3o
ool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0oooo01@0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00@P3oool2000002P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00l0oooo00<000000?ooo`3oool05@3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0U0?ooo`<00000703oool00480oooo00<000000?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool03P3oool00`00
0000oooo0?ooo`0F0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480
oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool03P3oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo
0?ooo`0G0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<0
00000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool0303oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
00000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool02`3oool00`000000oooo0?ooo`0I
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool02P3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?oo
o`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0J0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
2@3oool00`000000oooo0?ooo`0K0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0
6P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool0203oool00`000000oooo0?ooo`0L0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool01`3oool0
0`000000oooo0?ooo`0M0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool0
0480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0N0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool00480oooo0P00000X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`060?ooo`030000003oool0oooo01h0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09@3oool3000001`0oooo00120?ooo`030000003oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00D0oooo
00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo
00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo00@0oooo00<000000?ooo`3oool0803oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00<0oooo00<00000
0?ooo`3oool08@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?oo
o`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo00<0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3o
ool08P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000
003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`050000
O`3oool0oooo0?ooo`0000009@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0
oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`040000O`3oool0oooo000002H0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000
oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0000000W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`0000009`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`000000oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool2000002P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09@3oool3000001`0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0V0?ooo`030000000007l0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09P3oool00`000000001o0?ooo`0X0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`0007l0:@3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0T0?ooo`040000003oool0
oooo0000ObT0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3o
ool0oooo00@0ooooCP00000L0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool08`3oool01@000000oooo0?ooo`3oool0001o02T0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0C0?oooa`000008`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo000]0?ooo`040?oog`2LB00004RL0=ooo`<0oooo00D0
omnL04P000000000001d0;ooo`090?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0S0?ooo`050000003oool0oooo0?ooo`0007l0:@3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo00H0oooo3@00000E0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00;@3oool0103ooml0W4P0
0018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0
oooo0080oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0Q0?ooo`l000008P3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool0
02d0oooo00@0oooO09a80000B9`0gooo0P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo00L0oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo0240oooo00<000000?ooo`3oool00`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo01<0oooo3P0000060?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000Y0?ooo`030;md000000000000
00D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`070?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0P0?ooo`030000003oool0
oooo00@0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`090?ooo`X00000503oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`00703oool00`3ooml0W4P0000000070000000504RLg`3oool0oooo
0?oOW01dW=l00P3oool0203ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0
001809cOo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0P0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`050?ooo`@000007P3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00:P3oool02`3OW4P0M;oo
0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3o
ool3000007<0001o00<000000000O`0007l0<P0007l5000009l0001o0`00000L0?ooo`00:P3oool0
2`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool07P3oool00`000000oooo0?ooo`060?ooo`030000O`3oool0oooo02@0oooo1P00000W0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo000[0?ooo`0:0?oOW018M;l0oooO09a80000B9`0gooo0?ooo`3o
gi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`060?ooo`030000
O`3oool0oooo01h0oooo1P0000030?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0
6P3oool002`0oooo00T0gia807Bog`2LB00004RL0=ooo`3oool0oono07@00018W=l00`3oool00`3O
W4P0001d0;ooo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0M0?ooo`030000003oool0oooo00L0oooo00<0001o0?ooo`3oool06@3oool50000
00T0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00;03oool01@3ooml0
W4Q804Q80000B9`0gooo0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0L0?oo
o`030000003oool0oooo00P0oooo00<0001o0?ooo`3oool05@3oool4000000h0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00;@3oool0103ogi`0B0000018W03Oool20?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0K0?ooo`030000003oool0oooo00T0oooo00<0
001o0?ooo`3oool04P3oool300000180oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?oo
o`0J0?ooo`00;P3oool00`3OW4P004RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool02@3o
ool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool06`3o
ool00`000000oooo0?ooo`090?ooo`030000O`3oool0oooo00h0oooo1000000E0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool02P3oool0
0`0007l0oooo0?ooo`0;0?ooo`<000006@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0
oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0I0?ooo`030000003oool0oooo00/0oooo00<0001o0?ooo`3oool02@3oool20000
01`0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool00`2oM7@0
_ooo0?ooo`0f0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool01P3oool3000001h0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool00`2oM000B9cO
0?ooo`0f0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool0103oool200000240oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool01P2oM00000000000
00000000001d0;oooc<0oooo0P00000X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool05`3oool00`000000oooo0?ooo`0=0?ooo`040000O`3oool0oooo0?ooo`<000008`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0U0?ooo`<00000703oool000T0oooo00@0_ibL0=ooo`3oool0okmd0`0000000`00
07@0_ooo0?ooo`0_0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0F0?ooo`030000003oool0oooo00h0oooo00@0001o0?ooo`00000000009P3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000>0?ooo`040?oOW018000004RL
0=oooc00oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo01D0oooo00<000000?ooo`3oool03`3oool2000002P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`00303oool0103ogi`0B0000018W03Ooolb0?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0E0?ooo`030000003oool0
oooo00d0oooo0P0000000`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<00000
0?ooo`3oool06P3oool000T0oooo00D0_ibL0=oOW018000004RL0=ooo`0d0?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0D0?ooo`030000003oool0
oooo00`0oooo0P0000020?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`002@3oool00`2oM00004RL0=ooo`0f0?ooo`030000003oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0C0?ooo`030000003oool0oooo
00/0oooo0P0000040?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0J0?ooo`002@3oool00`2oW9`0gooo0?ooo`0f0?ooo`030000003oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0B0?ooo`030000003oool0oooo00X0
oooo0P0000060?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000
oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool04@3oool00`000000oooo0?ooo`090?ooo`800000203oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<0
00000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo0140oooo00<0
00000?ooo`3oool01`3oool2000000X0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool02P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<00000
0?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo00l0oooo00<000000?ooo`3oool01P3oool2000000d0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`03
0000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0>0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool000L0oooo00<0oooO09a800000000
1P0000000`00M;l0oooo0?ooo`0_0?ooo`800000:03oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo00h0oooo00<000000?ooo`3oool0103oool200000100oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0U0?ooo`<00000703oool000D0oooo00<0oooO09a8000000002P00
00000`00M;l0oooo0?ooo`0]0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0403o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool000@0oooo00@0okmd00000000B9`0gooo2@3oool0103ookl0M0000000B02Lgol]0?ooo`030000
003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo00<0oooo0P00000C0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`000000oooo0?ooo`0J0?ooo`000`3oool00`3ookl0M7Bo0?ooo`0=0?ooo`030?oog`2L
M9`0gooo02`0oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo00/0oooo00<000000?ooo`3oool00P3oool2000001D0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool05@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<000000?ooo`3oool06P3oool000T0oooo00<0oono07Ad_`3oool0103oool00`3O
W9`0_ooo0?ooo`0_0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0:0?ooo`030000003oool0oooo00800000603oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool000T0oooo00<0okmd00000000
0000100000000`0007@0_ooo0?ooo`0_0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`090?ooo`040000003oool0oooo000001X0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00090?ooo`03
0;md00000000000000@0000000<0001d0;ooo`3oool0;`3oool00`000000oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0203oool010000000oooo00000000000K0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?oo
o`002P3oool00`2oM7@0_ooo0?ooo`030?ooo`030=nLW02oool0oooo02l0oooo00<000000?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00L0oooo00<000000?ooo`00
00007@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3o
ool06P3oool000T0oooo00<0omnL09boo`3oool0=P3oool00`000000oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0000000N0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3o
ool00`2oM00007Bo0?ooo`040?ooo`030=nLW02oool0oooo02l0oooo00<000000?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00H0oooo0P00000O0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool0
0`3OW4P0000000000004000000030000M02oool0oooo02l0oooo0P00000X0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool01@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09@3oool3000001`0oooo00090?ooo`030?oog`2LB0000000
00@0000000<0001d0;ooo`3oool0;`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool0103oool00`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002P3oool00`3OW9`0
_ooo0?ooo`030?ooo`030=nLW02oool0oooo02l0oooo00<000000?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo0080oooo0P00000S0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool00`3o_g@0B9cO
0?ooo`0f0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool01`0007l0oooo
0?ooo`3oool000000?ooo`0000008`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool000T0oooo00<0_g@0001d_`3oool0103oool00`3OW9`0
_ooo0?ooo`0_0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool01P0007l0
oooo0?ooo`000000oooo000002@0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo00090?ooo`030=nLB0000000000000@0000000<0001d0;oo
o`3oool0;`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00D0001o0000
00000000oooo0000000U0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`002@3oool00`3ooml0W4P000000004000000030000M02oool0oooo
01H0oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3oool0omnL04P000000000001d
0;ooo`090?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool010000000oooo
0?ooo`00000V0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000
oooo0?ooo`0J0?ooo`00403oool00`3OW9`0_ooo0?ooo`0G0?ooo`030=nLB0000000000000<00000
00X004RL0=ooo`3oool0oono07@0001d_ol0oooo0?noM00007@0_ooo203oool00`000000oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02H0oooo00@000000000O`3oool000009`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00180oooo00@0
oooO09a8000007@0_ooo5@3oool00`3OW4P0B9cO0?ooo`020?ooo`060=nLM02Lgol0oooo0;md0000
B9`0gooo0P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?oo
o`3oool09@3oool01@000000oooo0000O`3oool0000002L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000?0?ooo`050?noM00000000000
0000B02Lgol0603oool00`3OW4P0M;oo0?ooo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02@0oooo00D00000
0?ooo`3oool0001o0000000X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00303oool0102oM000000000000000M;l<0?ooo`030?oog`2L
B000000000L0000000<0B9cO0?ooo`3oool0103oool00`2oM4P0W=oo0?ooo`020?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02<0oooo00D000000?ooo`3oool0oooo0000000Y0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00203oool01@3ogi`0B0000000
00000000M;oo0200oooo00<0oooO09a8M02oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00L0oooo0`00001<0000O`050000000007l0001o0000O`000000n`0007l3000001`0
oooo00050?ooo`050=nLB000000000000018W03Oool0903oool01`3o_g@007Bo0?ooo`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool08@3oool01`000000oooo0?ooo`3oool000000?ooo`0007l0:@3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool000<0oooo00<0
omnL04P00018W=l0:03oool01`3ookl0M01809cOo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`070?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool07`3oool2000000@0
oooo00<000000?ooo`0007l0:@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<000000?ooo`3oool06P3oool002h0oooo00L0oono07@00018W=l0oooo0?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo
01h0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0007lY0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool02@3o
oml0W4P00000000007@0_ooO09a800000000001d0;ooo`0G0?ooo`030?noW02Lgol0oooo0080oooo
00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo
00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0M0?ooo`030000003oool0oooo00<0oooo
00D000000?ooo`3oool0oooo0000O`0Y0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`002@3oool0203OW7@0W=oo0?oOW018000007Bo0?oo
o`3o_g@0B9cO603oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0
M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0
oooo01`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0
oooo01X0oooo00090?ooo`090=nLW02oool0oooo0;md000004P0W=oo0?ooo`2oM7@0_ooo01P0oooo
00H0oooO09a80000000000000000M02oool40?ooo`050?oo_`1d000000000000M02oool02@3oool0
0`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo01/0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00090?ooo`090?oOW018M;l0oooo
0?oOW018000007Bo0?ooo`2oW9`0gooo0300oooo00<000000?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0J0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`002@3oool01@2oM00000000000B02Lgol0gia80080000000<0M;oo0?ooo`3oool0
;`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo01T0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`040?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003o
ool0oooo02L0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00D0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0G0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01P3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`000`3oool00`3ookl0M4RL0=ooo`0=0?ooo`030?oOW018B9`0gooo
02`0oooo0P00000X0?ooo`030000O`3oool0oooo01H0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`060?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0U0?ooo`<00000703oool000@0oooo00@0omnL04P00000B9`0gooo2@3oool0103o
okl0M0000018W03Oool]0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool05@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool0
00D0oooo00<0oooO09a8000000002P0000000`00M;l0oooo0?ooo`0]0?ooo`030000003oool0oooo
02L0oooo00<0001o0?ooo`3oool05@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool000L0oooo00<0oooO09a8000000001P0000000`00M;l0
oooo0?ooo`0_0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0503oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo
00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0C0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?oo
o`0W0?ooo`030000O`3oool0oooo0180oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0:0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool04@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<00000
0?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0@0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?oo
o`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo00l0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000
003oool0oooo02L0oooo00<0001o0?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00d0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03P3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00@P3oool2000002P0oooo00<0001o0?ooo`3oool03@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00l0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02D0oooo0`00000L0?ooo`00@P3oool00`00
0000oooo0?ooo`0W0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0?0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0
oooo00<0001o0?ooo`3oool02`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0:0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04@3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo00T0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0B0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo0012
0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0203oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo0180oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?oo
o`3oool09`3oool00`0007l0oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool04`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0D0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0
1P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0
6P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`050?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool05@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool0
0`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo00@0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo
02L0oooo00<0001o0?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
01L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool00480oooo0P00000X0?ooo`030000O`3oool0oooo0080
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0U0?ooo`<00000703oool00480oooo
00<000000?ooo`3oool09`3oool01@0007l0oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3o
ool0603oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`040000
O`3oool0oooo000000X0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00
@P3oool00`000000oooo0?ooo`0W0?ooo`040000O`3oool0oooo000000T0oooo00<000000?ooo`3o
ool06P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000
O`3oool0000000T0oooo00<000000?ooo`3oool06`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3o
ool00`000000oooo0?ooo`0W0?ooo`030000O`000000oooo00T0oooo00<000000?ooo`3oool06`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00=03oool01@3ogi`0B0000000000007@0_ooo00T0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01`0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
00000?ooo`3oool06P3oool003<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo
00<000000?ooo`3oool09P3oool00`000000001o0?ooo`080?ooo`030000003oool0oooo01d0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<000000?ooo`3oool06P3oool003<0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?oo
o`070?ooo`030000003oool0oooo02H0oooo00<000000000O`3oool01`3oool00`000000oooo0?oo
o`0N0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo000b0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0001o00P0oooo00<00000
0?ooo`3oool07P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00<P3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0903oool010000000oooo0?ooo`0007l7
0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00380oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`070?ooool00000D`00000L0?ooo`00<P3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool08`3oool01@000000
oooo0?ooo`3oool0001o00D0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?oo
o`00<P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3o
ool08P3oool00`000000oooo0?ooo`020?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3o
ool08P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool01@3ooolS000001`0oooo000b0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
1`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3oool0
0P3oool00`000000oooo0?ooo`0R0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool0
8@3oool>00000280oooo00<000000?ooo`3oool06P3oool003<0oooo00<0gia80018W03Oool00P3o
ool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0103o
ool01@0007l0oooo0?ooo`3oool0000002D0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`040?oooad000001P3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo
000c0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0103oool0100007l0oooo0?ooo`00000V0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
0240oooo3P00000S0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool003@0
oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo01l0oooo00<000000?oo
o`3oool01@3oool00`0007l0oooo0000000W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01L0oooo2P0000070?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?oo
o`00@P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00H0oooo00<0001o0?ooo`00
00009`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0?0?ooo`P000004@3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`070?ooo`030000O`000000oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
1@3oool:000001T0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
003oool0oooo01X0oooo00120?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0U0?ooo`X000008`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`070?ooo`030000000007l0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo01h0oooo1`0000020?ooo`030000O`3oool0
oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0
oooo01X0oooo00120?ooo`800000703oool00`000000oooo0?ooo`070?ooo`030000003oool0001o
02T0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
01P0oooo1P0000090?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0U0?ooo`<00000703oool00480oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`080?ooo`030000003oool0001o02T0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo01<0oooo1@00000?0?ooo`030000O`3oool0oooo02P0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0
oooo00120?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool01`3oool010000000oooo
0?ooo`0007lY0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0>0?ooo`D00000503oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00L0oooo00D000000?ooo`3oool0oooo0000O`0Y0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`080?ooo`H0
00006@3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0100007l0oooo0?ooo`3oool7000001l0
oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo0080oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02D0oooo1P00000V0?ooo`030000O`3oool0oooo02P0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo
00120?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`030?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0Q0?ooo`@000000P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo0012
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`04
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0L
0?ooo`D000001P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0H0?oo
o`@000002`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0E0?ooo`<0
00003`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0A0?ooo`@00000
4P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`8000005@3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo00h0oooo0`00000F0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02D0oooo0`00000L0?ooo`00@P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool0203oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool02`3oool3000001T0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool01`3oool4000001`0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool0103oool300000200oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000
oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool02P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool01000000S0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0
6P3oool00480oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00/0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`<000009`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool08`3oool4000000030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo01l0oooo100000040?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo01/0oooo100000080?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool03P3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo01P0oooo0`00000<0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool03`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo01H0oooo0P00000?0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool0403oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo01<0oooo0`00000A0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`00@P3oool2000000d0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0A0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0403oool30000
01@0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0U0?ooo`<00000703oool00480
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0140
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0>0?ooo`8000005`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0180oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0;0?ooo`<000006@3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01<0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`080?ooo`<00000703oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01@0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`060?ooo`8000007`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01@0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`030?ooo`<000008@3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool002T0oooo00<0_g@000000000
0000100000002@18W=l0oooo0?ooo`3oool0omnL04P000000000001d0;ooo`090?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0E0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool00`00000T0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00:P3oool00`3OW4P0000000000003
0000000:0018W03Oool0oooo0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`P0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01H0oooo00<0001o0?oo
o`3oool09`3oool3000002L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo000[0?ooo`030=nLB018W=l0oooo0080oooo00H0giad09cOo`3o
ool0_g@00018W03Oool20?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01L0oooo00<0001o0?ooo`3oool09@3oool2
000000030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo000/0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0oooo0280oooo0`0000020?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo000]0?ooo`030;mdB02Lgol0oooo0080oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0oooo0200oooo0P0000050?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000003oool0oooo01X0oooo000]0?ooo`030?oog`2LB7@0_ooo0080oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`070?ooo`<000001P0007l00`000000001o0000O`020000O`030000
000007l0001o03T0001o0`00003o0000O`<0001o0`00000L0?ooo`00;P3oool01`3o_g@007Bo0?oo
o`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo00H0
oooo00D000000?ooo`3oool0oooo0000000L0?ooo`030000O`3oool0oooo01/0oooo0P00000:0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000003oool0oooo01X0oooo000^0?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool01@3oool01@000000oooo0?ooo`3o
ool0000001d0oooo00<0001o0?ooo`3oool0603oool3000000`0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool002h0oooo00L0oono07@00018W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
1`3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`0000007@3oool00`0007l0
oooo0?ooo`0E0?ooo`<000003`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00:@3oool00`3o_i`0W=oo
0?ooo`020?ooo`030?oOW0180000B9cO0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`070?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo0000000N0?ooo`03
0000O`3oool0oooo01<0oooo0P00000B0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000Y0?ooo`080?oo
g`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`080?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000007`3oool00`0007l0
oooo0?ooo`0@0?ooo`<00000503oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00:P3oool01P3ooml0W4P0
000000000000001d0;ooo`@0oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0
oooo00<0oooo00@000000?ooo`3oool00000803oool00`0007l0oooo0?ooo`0=0?ooo`<000005`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo00000200oooo00<0001o0?ooo`3oool02P3oool3000001X0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool00480oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`00000Q0?ooo`030000O`3o
ool0oooo00P0oooo0P00000M0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0
oooo0080oooo00<000000?ooo`0000008P3oool00`0007l0oooo0?ooo`060?ooo`8000007`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0J0?ooo`00@P3oool01`000000oooo0?ooo`3oool000000?ooo`0000008`3o
ool00`0007l0oooo0?ooo`030?ooo`<000008@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool2
00000080oooo00<000000?ooo`0000008`3oool0100007l0oooo0?ooo`3oool2000002@0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02D0oooo0`00
000L0?ooo`00@P3oool01P000000oooo0?ooo`000000oooo000002@0oooo00@0001o0?ooo`000000
00009P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`02000002D0oooo
0P00000X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`050000003oool000000?ooo`000000
8`3oool2000000030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00@0
00000?ooo`00000000008P3oool200000080oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`00@P3oool300000240oooo0P0000040?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
00000?ooo`3oool06P3oool00480oooo0P00000P0?ooo`8000001P3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo00120?ooo`8000007P3oool2000000P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0K0?ooo`80
00002P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000
003oool0oooo01T0oooo0P00000<0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
9`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0
:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0
6P3oool00480oooo00<000000?ooo`3oool05`3oool2000000h0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`000000oooo0?ooo`0J0?ooo`00@P3oool2000001L0oooo00<000000?ooo`3oool03P3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool0
0`0007l0oooo0?ooo`0U0?ooo`<00000703oool00480oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0?0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480
oooo00<000000?ooo`3oool04`3oool200000180oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000
oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0180oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo01<0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?oo
o`0J0?ooo`00@P3oool00`000000oooo0?ooo`0?0?ooo`8000005P3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool05P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo00/0oooo0P00000J0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03o
ool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3o
ool00480oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0K0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0L0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480
oooo0P0000080?ooo`030000003oool0oooo01d0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0
oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09@3oool3000001`0
oooo00120?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool07P3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo0012
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0803oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool08@3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool08P3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000003oool0oooo01X0oooo000]0?ooo`040?oog`2LB00004RL0=ooo`<0oooo00D0omnL04P00000
0000001d0;ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool08P3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo000]0?ooo`040?oog`2LB00004RL0=oo
o`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00D000000?ooo`3oool0oooo
0000000U0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool002d0oooo00@0
oooO09a80000B9`0gooo0P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo
00@000000?ooo`3oool000009P3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0
oooo000Y0?ooo`030;md00000000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0
001d0;ooo`070?ooo`040000003oool0oooo000002H0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0J0?ooo`00:@3oool0303ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?oo
o`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0000000W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool002X0oooo00/0gia807Boo`3oool0
oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo0`00
003o0000Odd0001o0`00000L0?ooo`00:P3oool02`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0
oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool2000002P0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00:`3oool02P3ogi`0B7Bo0?oog`2LB00004RL
0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00L0ooooo`00001C000001`0oooo
000/0?ooo`090=nLB01d_ml0W4P00018W03Oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?oo
o`3oool06P3oool002`0oooo00D0oooO09a8B018B00004RL0=ooo`020?ooo`030=nLB000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo
02L0oooo00<000000?ooo`3oool06P3oool002d0oooo00@0omnL04P00000B9`0gooo0P3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<000000?ooo`3oool06P3oool002h0oooo00<0gia80018W03Oool00`3oool0
1@3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o
0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?oo
o`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3o
ool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000
003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000
O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3o
ool2000002P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3o
ool0oooo02L0oooo00<0001o0?ooo`3oool09@3oool3000001`0oooo00120?ooo`030000003oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0
oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`00
0000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`00
07l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480
oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02P0
oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo01X0
oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo
0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0
00000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000O`3oool0oooo02P0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W
0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`0007l0oooo0?ooo`0X0?ooo`030000O`3oool0oooo02L0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`00@P3oool00`000000oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool09`3oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<0001o0?ooo`3oool0:03oool00`0007l0oooo0?ooo`0W0?ooo`03
0000O`3oool0oooo02L0oooo00<000000?ooo`3oool06P3oool00480oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01X0oooo00120?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool06P3oool00480ooooo`00001C000001`0oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003>0?ooo`040=nLB0000000000007Boobh0oooo00H0oooO09a80000000000000000M02o
oonZ0?ooo`00d@3oool00`2oM7@0_ooo0?ooo`0^0?ooo`030;md0000B9`0gooo0:/0oooo003A0?oo
o`030?oOW01dW=l0oooo02h0oooo00<0_g@00018W03Oool0Z`3oool00=40oooo00<0oooO09a8M02o
ool0;P3oool00`2oM00004RL0=ooo`2[0?ooo`00X@3oool00`3ogi`0M9cO0?ooo`060?ooo`070;md
0000000000000000M02oool0oooo0=nLB002000000040000M02oool0oooo0=nLB0<0000000<0M;oo
0?ooo`3oool00P3oool01`2oM00000000000000004P0W=oo0?ooo`3OW4P00P0000001P0007@0_ooo
0?oOW0180000000004RLg`@0oooo00<0_g@007Boo`3oool0303oool00`3ogi`0B000000000020000
00030000M02oool0oooo00<0oooo00D0gia8000000000000000007Boo`020?ooo`060=nLB0000000
00000000B02Lgml0W4P00P0000000`0007@0_ooO09a8000200000005001d_`3oool0_g@00018B018
00000P0000000`18W=l0oooo0?ooo`020?ooo`050=nLB000000000000000001d_ol00P3oool01P2o
M00000000000000007@0_ooO09a80080000000@004RL0=ooo`3oool0_g@00P000000100007@0_ooo
0?ooo`2oM002000000030000M02ooml0W4P00080000000T004RL0=ooo`3ogi`0B00000000018W=l0
oono0;no_`1d00000P0000000`1d_ol0oooo0?ooo`1f0?ooo`00X@3oool00`2oM000001d0;ooo`05
0?ooo`0C0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80000M02oool0oooo0=nLB000
0000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`0A0;md0000000004RL0=ooo`3ooml0W4Qd0;ooo`3o
ool0gia80000M02oool0oooo0=nLB00007@0_ooo0=no_`3Oool00P3oool0103ogi`0B0000018W03O
ool;0?ooo`040=nLB000000004RL0=ooo`80oooo00<0oono07@0B02Lgol00P3oool01`3OW4P00000
04RLg`3oool0oooO09a80018W=l00P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo
0?ooo`3ooml0W4P0001d_`020?ooo`070;md00000000M;oo0?ooo`3ookl0M00004RLg`020?ooo`07
0=nLB0000000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`070=nLB000B9`0gooo0?ooo`3ooml0W4P0
001d_`020?ooo`0?0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80018W03Oool0oooo
0?oog`2LB00007Bo0080oooo00/0gia80000M02oool0gkno0=oo_`1d04P0W=oo0?ooo`3ookl0M018
09cOo`1g0?ooo`00X03oool0103ookl0M0000000001d_ol40?ooo`030?oOW0180000B9cO00<0oooo
00T0omnL07BLg`3oool0gia80000M02oool0omnL04P00000M;l00`3oool01P3o_g@0001d0;ooo`3o
gi`0B000001d_`<0oooo00X0omnL09boo`3oool0gia80000M02oool0oooo0=nLB000B9`0gooo103o
ool0102oM00004RL09ad001d_ol:0?ooo`030?noM000000007Bo00D0oooo00H0oooO09a8M02oool0
omnL04P00000M;l30?ooo`070?noM00007@0_ooo0?ooo`3OW4P0001d0;ooo`020?ooo`070;md0000
B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`030;md0000B9`0gooo0080oooo00H0okmd0000M02o
ool0omnL04P00000M;l30?ooo`0?0?noM00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0oooO09a8
0000M;l0oooo0?oOW0180000B9cO00<0oooo00X0omnL07BLg`3oool0gia80018W03Oool0oooo0?oo
g`2LB00007Bo0P3oool01P3OW4P004RL0=ooo`3oool0oono0;noo`80oooo00<0okmd0000001d_ol0
M`3oool00:00oooo00D0gia80000B02Lgkl0M4RL0=ooo`030?ooo`030?noM00004P0W=oo00H0oooo
00H0gia80000M02oool0okmd0000001d_ol30?ooo`060?noM0000000M;oo0?noM00004P0W=oo1P3o
ool01`3OW4P0001d0;ooo`3oool0gia80018W03Oool00`3oool01@3o_g@0000007Boo`3ogi`0B7Bo
00T0oooo00@0oooO09a8000007@0_ooo203oool00`3o_g@0000007Boo`030?ooo`070?noM0000000
M;oo0?ooo`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?oo
o`030;md0000B9`0gooo0080oooo00H0omnL04P0001d_ol0okmd0000001d_ol30?ooo`0?0?noM000
0000M;oo0?ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00004P0W=oo00H0oooo
00L0gia80018W03Oool0oooo0?oog`2LB00007Bo0080oooo00<0gia80018W03Oool00`3oool01@3o
okl0M0000000000007@0_ooo07L0oooo002O0?ooo`060?oog`2LB00004RL0=ooo`2oM4P0W=oo0`3o
ool00`3o_g@0001d0;ooo`060?ooo`060=nLB00007@0_ooo0?noM0000000M;oo0`3oool01P3o_g@0
000007Boo`3o_g@0001d0;ooo`H0oooo00L0gia80000M02oool0oooo0=nLB000B9`0gooo0080oooo
00L0oooO09a80000B9`0gooo0?ooo`2oM4P0W=oo00P0oooo00@0omnL04P00000B9`0gooo203oool0
0`3o_g@0000007Boo`030?ooo`070?noM0000000M;oo0?ooo`3OW4P0001d0;ooo`020?ooo`070;md
0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`030;md0000B9`0gooo0080oooo00H0omnL04P0
001d_ol0okmd0000001d_ol30?ooo`0?0?noM0000000M;oo0?ooo`3OW4P004RL0=ooo`3oool0oooO
09a80000M;l0oooo0?noM00007@0_ooo00H0oooo00L0gia80018W03Oool0oooo0?oog`2LB00007Bo
0080oooo00<0gia80018W03Oool00P3oool0102oM000000000000018W=mi0?ooo`00W`3oool01P3o
_g@0000007Boo`3oool0omnL07BLg`<0oooo00<0omnL04P000000000100000002018W=l0oooo0=nL
B00007@0_ooo0?oo_`1d0000M;oo0`3oool01P2oM000001809cOo`3ogi`0B01809cOo`80oooo00/0
okmd0000B02Lgol0oooo0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`020?ooo`030?noM00004P0
W=oo0080oooo00<0okmd04RLg`3oool01`3oool00`3o_g@00000001d_`090?ooo`030?oo_`1d0000
M;oo00<0oooo00L0_g@00000B02Lgol0oooo0=nLB00004P0W=oo0080oooo00L0_g@00000M02oool0
oooo0?oog`2LB00007Bo0080oooo00<0_g@00000M02oool00P3oool01P3OW4P0001809cOo`3ookl0
M00007Boo`<0oooo00h0_g@00000B02Lgol0oooo0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3o
ool0omnL04P000D0000000T0B9cO0?ooo`2oM000001d0;ooo`3oool0oooO09a80000M;l00P3oool0
1`3OW4P004RL0=ooo`3oool0omnL04P00018W=l00P3oool00`2oW9`0gooo0?ooo`1f0?ooo`00WP3o
ool0203ooml0W4P00018W03Oool0oooo0?oog`2LB7@0_ooo0`3oool04`2oM000B9cO0?ooo`3oool0
gia80000B02Lgol0oooo0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0okmd00000000M;l0
0P3oool0502oM00004RL0=ooo`3ooml0W4P00000M02ooml0W4P00000000007@0_ooo0?ooo`3OW4P0
04RL0=ooo`3oool0oono07@00000B9`0gooo0P3oool00`3ookl0M01d0;ooo`070?ooo`030?noM000
000007Bo00X0oooo0300gia80018W03Oool0oooo0?noM000000007Bo0?oog`2LB00000000000B018
B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l0oono07@00000000004Q804Q8W03O
ool0omnL04P00000B9`0gooo0?ooo`3OW4P004RL0=ooo`3oool0okmd00000000M;l0oooO09a80000
000004Q804P0B02Lgol0oono07@00018W=l20?ooo`0?0;md0018W=l0oooo0?ooo`3OW4P0001809cO
g`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0080oooo00/0gia80018W03Oool0oooo0?oo
_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`1g0?ooo`00WP3oool00`3ogi`0B00004RLg`030?ooo`03
0?noM01d_ol0oooo00<0oooo00D0gia8000000000000001809cOo`020?ooo`030=nLB00007@0_ooo
0080oooo00D0okmd000000000000000004RLg`040?ooo`050=nLB000000000000000B02Lgol00P3o
ool0102oM000001d0;ooo`2oM0030000000;0018M01dB000000000000000B9`0gooo0?ooo`2oM000
00000018W03Oool01P3oool0103ogi`0B0000018W03Oool:0?ooo`050?noM0000000000000000018
W=l00`3oool05@2oM000001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`3o
ool0oooO09a80000B9`0gmnL04P000000000001d0;ooo`030?ooo`050?noM0000000000000000018
W=l00`3oool01`2oM00004RL0=oog`2LB00000000000M02oool00`3oool01@3OW4P0000000000000
04P0W=oo0080oooo00T0_g@00018W03Ooml0W4P00000000007@0_ooo0?ooo`2oM0000`0000000`00
B9`0gooo0?noM00200000003001d_`2oW9`0gooo07L0oooo002N0?ooo`030;md000007@0_ooo00<0
oooo00<0oono07A8W03Oool02P3oool00`3OW4P0001d0;ooo`0F0?ooo`030;md0000B9`0gooo0140
oooo00@0oooO09a8000004P0W=oo1@3oool00`3ooml0W7BL0=ooo`140?ooo`030;md0000B9`0gooo
07l0oooo002M0?ooo`030?oo_`1d000007Bo00D0oooo00<0giad09cOo`3oool02@3oool00`3OW4P0
001d0;ooo`0F0?ooo`030?oo_`1dB9`0gooo0180oooo00<0omnL04P00000M;l01@3oool00`3o_g@0
04RL0=ooo`140?ooo`030?oo_`1dB9`0gooo07l0oooo002M0?ooo`030;md00000000B9cO00D0oooo
00<0gia80018W03Oool02@3oool00`3OW4P0001d0;ooo`0F0?ooo`030?oog`2LM9`0gooo01<0oooo
00<0okmd00000000M;l00`3oool00`3OW4P00000001d_`150?ooo`030?oog`2LM9`0gooo07l0oooo
002K0?ooo`030?oog`2LB00000000080000000D007Bo0?ooo`3oool0oooo0=nLB002000000030018
W03Oool0oooo00D0oooo00D0oooO09a800000000001d0;ooo`0B0?ooo`030=nLB00004P0W=oo01P0
oooo00<0oooO09a8000000000P0000000`18W=l0oono07Ad_`370?ooo`00/@3oool00`2oM000001d
0;ooo`0B0?ooo`030=nLB00004P0W=oo0>L0oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-3.93661, -126.176, 0.0890377, \
0.667782}}]
}, Closed]],

Cell[TextData[{
  "In the absence of resistance, ",
  Cell[BoxData[
      \(TraditionalForm\`v\_x\)]],
  " is constant while ",
  Cell[BoxData[
      \(TraditionalForm\`v\_y\)]],
  " has a linear time dependence produced by constant acceleration.  However, \
air resistance eventually extinguishes the horizontal velocity and the ball \
falls vertically downward.  Furthermore, the vertical velocity eventually \
approaches a constant called the ",
  StyleBox["terminal velocity",
    FontSlant->"Italic"],
  ".  Evidently, the terminal velocity for a baseball is approximately 45 \
m/s.  It is interesting to observe that the terminal velocity is quite \
similar to the initial velocity needed for home run, although this is an \
accidental coincidence."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Dependence on initial angle", "Subsection"],

Cell["\<\
Everyone knows the optimal angle is 45\[Degree] in the absence of air \
resistance.  The effect of air resistance upon the optimal angle can be \
easily investigated graphically.  We plot a table of trajectories and supply \
the iterator as an argument to the function.\
\>", "Text"],

Cell[BoxData[
    \(\(plotAngleDependence[\(x\_0\) : _, \ \(y\_0\) : _, \(v\_0\) : _, 
          angles_, parameters_] := \n\t
        ParametricPlot[
          Evaluate[
            Table[resistedTrajectory[x\_0, y\_0, v\_0, \[Theta]\_0, 
                parameters], \ angles]\ ], {t, 0, 10}, 
          PlotRange \[Rule] {{0, 150}, {0, 50}}, Frame \[Rule] True, 
          GridLines \[Rule] Automatic, AspectRatio \[Rule] Automatic, 
          FrameLabel \[Rule] {"\<x (m)\>", "\<y (m)\>", "\<Dependence on \
Initial Angle\>", "\< \>"}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotAngleDependence[0, 0, 
        50, {\[Theta]\_0, \ 30  Degree, \ 50  Degree, \ 5\ Degree}, \ 
        baseballParameters];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .33333 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
3.29597e-017 0.00666667 1.73472e-018 0.00666667 [
[.13333 -0.0125 -8 -15.6875 ]
[.13333 -0.0125 8 0 ]
[.26667 -0.0125 -8 -15.6875 ]
[.26667 -0.0125 8 0 ]
[.4 -0.0125 -8 -15.6875 ]
[.4 -0.0125 8 0 ]
[.53333 -0.0125 -8 -15.6875 ]
[.53333 -0.0125 8 0 ]
[.66667 -0.0125 -11 -15.6875 ]
[.66667 -0.0125 11 0 ]
[.8 -0.0125 -11 -15.6875 ]
[.8 -0.0125 11 0 ]
[.93333 -0.0125 -11 -15.6875 ]
[.93333 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -15.0625 -31.375 ]
[.5 -0.0125 15.0625 -15.6875 ]
[-0.0125 .06667 -16 -7.84375 ]
[-0.0125 .06667 0 7.84375 ]
[-0.0125 .13333 -16 -7.84375 ]
[-0.0125 .13333 0 7.84375 ]
[-0.0125 .2 -16 -7.84375 ]
[-0.0125 .2 0 7.84375 ]
[-0.0125 .26667 -16 -7.84375 ]
[-0.0125 .26667 0 7.84375 ]
[-0.0125 .33333 -16 -7.84375 ]
[-0.0125 .33333 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .16667 -31.6875 -14.8125 ]
[-0.01875 .16667 -16 14.8125 ]
[ 0 .33333 .125 0 ]
[.5 .34583 -70.5625 0 ]
[.5 .34583 70.5625 15.6875 ]
[ 1 0 .125 0 ]
[1.0125 .16667 0 -3.5 ]
[1.0125 .16667 15.6875 3.5 ]
[ 0 0 0 0 ]
[ 1 .33333 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.13333 0 m
.13333 .33333 L
s
.26667 0 m
.26667 .33333 L
s
.4 0 m
.4 .33333 L
s
.53333 0 m
.53333 .33333 L
s
.66667 0 m
.66667 .33333 L
s
.8 0 m
.8 .33333 L
s
.93333 0 m
.93333 .33333 L
s
0 .06667 m
1 .06667 L
s
0 .13333 m
1 .13333 L
s
0 .2 m
1 .2 L
s
0 .26667 m
1 .26667 L
s
0 g
.13333 0 m
.13333 .00625 L
s
gsave
.13333 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.26667 0 m
.26667 .00625 L
s
gsave
.26667 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.4 0 m
.4 .00625 L
s
gsave
.4 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.53333 0 m
.53333 .00625 L
s
gsave
.53333 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.66667 0 m
.66667 .00625 L
s
gsave
.66667 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.8 0 m
.8 .00625 L
s
gsave
.8 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(120) show
1.000 setlinewidth
grestore
.93333 0 m
.93333 .00625 L
s
gsave
.93333 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(140) show
1.000 setlinewidth
grestore
.125 Mabswid
.03333 0 m
.03333 .00375 L
s
.06667 0 m
.06667 .00375 L
s
.1 0 m
.1 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.2 0 m
.2 .00375 L
s
.23333 0 m
.23333 .00375 L
s
.3 0 m
.3 .00375 L
s
.33333 0 m
.33333 .00375 L
s
.36667 0 m
.36667 .00375 L
s
.43333 0 m
.43333 .00375 L
s
.46667 0 m
.46667 .00375 L
s
.5 0 m
.5 .00375 L
s
.56667 0 m
.56667 .00375 L
s
.6 0 m
.6 .00375 L
s
.63333 0 m
.63333 .00375 L
s
.7 0 m
.7 .00375 L
s
.73333 0 m
.73333 .00375 L
s
.76667 0 m
.76667 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.86667 0 m
.86667 .00375 L
s
.9 0 m
.9 .00375 L
s
.96667 0 m
.96667 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -76.0625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
72.125 15.813 moveto
(H) show
75.938 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.313 15.813 moveto
(L) show
89.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .06667 m
.00625 .06667 L
s
gsave
-0.0125 .06667 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .13333 m
.00625 .13333 L
s
gsave
-0.0125 .13333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .2 m
.00625 .2 L
s
gsave
-0.0125 .2 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
0 .26667 m
.00625 .26667 L
s
gsave
-0.0125 .26667 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .33333 m
.00625 .33333 L
s
gsave
-0.0125 .33333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .01333 m
.00375 .01333 L
s
0 .02667 m
.00375 .02667 L
s
0 .04 m
.00375 .04 L
s
0 .05333 m
.00375 .05333 L
s
0 .08 m
.00375 .08 L
s
0 .09333 m
.00375 .09333 L
s
0 .10667 m
.00375 .10667 L
s
0 .12 m
.00375 .12 L
s
0 .14667 m
.00375 .14667 L
s
0 .16 m
.00375 .16 L
s
0 .17333 m
.00375 .17333 L
s
0 .18667 m
.00375 .18667 L
s
0 .21333 m
.00375 .21333 L
s
0 .22667 m
.00375 .22667 L
s
0 .24 m
.00375 .24 L
s
0 .25333 m
.00375 .25333 L
s
0 .28 m
.00375 .28 L
s
0 .29333 m
.00375 .29333 L
s
0 .30667 m
.00375 .30667 L
s
0 .32 m
.00375 .32 L
s
.25 Mabswid
0 0 m
0 .33333 L
s
gsave
-0.01875 .16667 -99.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
75.8125 11.8438 translate
90 rotate
-75.8125 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
71.625 15.813 moveto
(H) show
75.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
84.813 15.813 moveto
(L) show
88.625 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .32708 m
0 .33333 L
s
.13333 .32708 m
.13333 .33333 L
s
.26667 .32708 m
.26667 .33333 L
s
.4 .32708 m
.4 .33333 L
s
.53333 .32708 m
.53333 .33333 L
s
.66667 .32708 m
.66667 .33333 L
s
.8 .32708 m
.8 .33333 L
s
.93333 .32708 m
.93333 .33333 L
s
.125 Mabswid
.03333 .32958 m
.03333 .33333 L
s
.06667 .32958 m
.06667 .33333 L
s
.1 .32958 m
.1 .33333 L
s
.16667 .32958 m
.16667 .33333 L
s
.2 .32958 m
.2 .33333 L
s
.23333 .32958 m
.23333 .33333 L
s
.3 .32958 m
.3 .33333 L
s
.33333 .32958 m
.33333 .33333 L
s
.36667 .32958 m
.36667 .33333 L
s
.43333 .32958 m
.43333 .33333 L
s
.46667 .32958 m
.46667 .33333 L
s
.5 .32958 m
.5 .33333 L
s
.56667 .32958 m
.56667 .33333 L
s
.6 .32958 m
.6 .33333 L
s
.63333 .32958 m
.63333 .33333 L
s
.7 .32958 m
.7 .33333 L
s
.73333 .32958 m
.73333 .33333 L
s
.76667 .32958 m
.76667 .33333 L
s
.83333 .32958 m
.83333 .33333 L
s
.86667 .32958 m
.86667 .33333 L
s
.9 .32958 m
.9 .33333 L
s
.96667 .32958 m
.96667 .33333 L
s
.25 Mabswid
0 .33333 m
1 .33333 L
s
gsave
.5 .34583 -131.563 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Dependence) show
124.875 15.813 moveto
(on) show
139.875 15.813 moveto
(Initial) show
171.063 15.813 moveto
(Angle) show
200.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 0 m
1 0 L
s
.99375 .06667 m
1 .06667 L
s
.99375 .13333 m
1 .13333 L
s
.99375 .2 m
1 .2 L
s
.99375 .26667 m
1 .26667 L
s
.125 Mabswid
.99625 .01333 m
1 .01333 L
s
.99625 .02667 m
1 .02667 L
s
.99625 .04 m
1 .04 L
s
.99625 .05333 m
1 .05333 L
s
.99625 .08 m
1 .08 L
s
.99625 .09333 m
1 .09333 L
s
.99625 .10667 m
1 .10667 L
s
.99625 .12 m
1 .12 L
s
.99625 .14667 m
1 .14667 L
s
.99625 .16 m
1 .16 L
s
.99625 .17333 m
1 .17333 L
s
.99625 .18667 m
1 .18667 L
s
.99625 .21333 m
1 .21333 L
s
.99625 .22667 m
1 .22667 L
s
.99625 .24 m
1 .24 L
s
.99625 .25333 m
1 .25333 L
s
.99625 .28 m
1 .28 L
s
.99625 .29333 m
1 .29333 L
s
.99625 .30667 m
1 .30667 L
s
.99625 .32 m
1 .32 L
s
.25 Mabswid
1 0 m
1 .33333 L
s
gsave
1.0125 .16667 -56.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
64.5 11.8438 translate
90 rotate
-64.5 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
66.000 15.813 moveto
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .33333 L
0 .33333 L
closepath
clip
newpath
.5 Mabswid
0 0 m
.11138 .0591 L
.22168 .10593 L
.27183 .12273 L
.31676 .13511 L
.35856 .14419 L
.38122 .14807 L
.40178 .15093 L
.42366 .15325 L
.43504 .15416 L
.44081 .15454 L
.44699 .15489 L
.45237 .15514 L
.45727 .15532 L
.46217 .15547 L
.46682 .15557 L
.46942 .15561 L
.47087 .15563 L
.47218 .15564 L
.47473 .15565 L
.47714 .15566 L
.4785 .15566 L
.47994 .15565 L
.48251 .15563 L
.48394 .15562 L
.48528 .1556 L
.4883 .15555 L
.49085 .15549 L
.49354 .15542 L
.49837 .15526 L
.50354 .15504 L
.50922 .15474 L
.51932 .15407 L
.52873 .15327 L
.54607 .15136 L
.5645 .1487 L
.60203 .14115 L
.64034 .13037 L
.67513 .11772 L
.7066 .10378 L
.77073 .06751 L
.83491 .01953 L
s
.83491 .01953 m
.85561 0 L
s
0 0 m
.10536 .06857 L
.20977 .12481 L
.25728 .14592 L
.29988 .16217 L
.34204 .17557 L
.3806 .18534 L
.40054 .1894 L
.42183 .19296 L
.44087 .19545 L
.4582 .19712 L
.46716 .19776 L
.47208 .19804 L
.47673 .19826 L
.48103 .19843 L
.48329 .19851 L
.4857 .19857 L
.48792 .19863 L
.48993 .19866 L
.49195 .19869 L
.49385 .19871 L
.49609 .19873 L
.49855 .19873 L
.49979 .19873 L
.50111 .19872 L
.5035 .1987 L
.50587 .19867 L
.50812 .19863 L
.51231 .19853 L
.51724 .19836 L
.52177 .19816 L
.53199 .19754 L
.5404 .19686 L
.54941 .19596 L
.56554 .1939 L
.58447 .19072 L
.60157 .18712 L
.63377 .1784 L
.66804 .16621 L
.73242 .13457 L
.79103 .09487 L
.84978 .04339 L
s
.84978 .04339 m
.88959 0 L
s
0 0 m
.09854 .07748 L
.19625 .14259 L
.28067 .18766 L
.35643 .2178 L
.3936 .22861 L
.43217 .23681 L
.45057 .23959 L
.46016 .24074 L
.47024 .24172 L
.47899 .24238 L
.4836 .24265 L
.48854 .24289 L
.49306 .24306 L
.49718 .24317 L
.49915 .24321 L
.5013 .24324 L
.50333 .24326 L
.50521 .24326 L
.50741 .24326 L
.50852 .24326 L
.50973 .24325 L
.51188 .24322 L
.51387 .24319 L
.51623 .24314 L
.51837 .24308 L
.52318 .2429 L
.52731 .2427 L
.53178 .24244 L
.53988 .24183 L
.55589 .24013 L
.57052 .23797 L
.60052 .23172 L
.61728 .22711 L
.63281 .22211 L
.69118 .19666 L
.72028 .1798 L
.75009 .15945 L
.80181 .11619 L
.85385 .06149 L
.90105 .00103 L
s
.90105 .00103 m
.90172 0 L
s
0 0 m
.09096 .08576 L
.18121 .15913 L
.25928 .21139 L
.32945 .24805 L
.36393 .26209 L
.39976 .27366 L
.41686 .27805 L
.43517 .2819 L
.45221 .28468 L
.46775 .28652 L
.47516 .28716 L
.47924 .28744 L
.4831 .28766 L
.48666 .28783 L
.49054 .28798 L
.49224 .28802 L
.49405 .28807 L
.49576 .2881 L
.49731 .28812 L
.49932 .28813 L
.50033 .28814 L
.50145 .28814 L
.50341 .28813 L
.50524 .28811 L
.50635 .2881 L
.5074 .28808 L
.50937 .28803 L
.51148 .28797 L
.51377 .28789 L
.51765 .28772 L
.52175 .28749 L
.52909 .28695 L
.53651 .28623 L
.54327 .28542 L
.55838 .28309 L
.57155 .28045 L
.58527 .27709 L
.61367 .26809 L
.63949 .2574 L
.69253 .22754 L
.74612 .18564 L
.79504 .13597 L
.83957 .08019 L
.8842 .01294 L
s
.8842 .01294 m
.89147 0 L
s
0 0 m
.0827 .09335 L
.16479 .17429 L
.23587 .23317 L
.29985 .27585 L
.33395 .29416 L
.36411 .30757 L
.39464 .31835 L
.42282 .32566 L
.43699 .32835 L
.44468 .32952 L
.45185 .33043 L
.45841 .3311 L
.46439 .33159 L
.46756 .33179 L
.471 .33197 L
.47288 .33205 L
.47463 .33212 L
.47627 .33217 L
.47807 .33221 L
.47988 .33224 L
.48086 .33225 L
.48178 .33226 L
.48343 .33227 L
.48519 .33227 L
.48618 .33226 L
.48726 .33225 L
.48827 .33223 L
.48921 .33222 L
.49098 .33218 L
.49285 .33212 L
.49625 .33198 L
.49942 .33182 L
.50663 .3313 L
.51304 .33068 L
.51889 .32997 L
.53198 .3279 L
.54513 .32514 L
.55725 .32198 L
.5843 .31272 L
.61045 .30077 L
.63369 .28756 L
.68202 .25181 L
.72635 .20843 L
.77111 .15323 L
.81183 .09193 L
.84876 .0261 L
s
.84876 .0261 m
.86131 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{437.625, 145.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000:@0000fR000`400?l00000o`00003oo`3ooooo
0?oooi80oooo003o0?ooool0ooooTP3oool00?l0ooooD`3oool00`3ooml0W4Q809cOo`0=0?ooo`03
0?oOW018B9`0gooo0?l0oooo;03oool00?l0ooooD`3oool00`2oM4P0W=oo0?ooo`0>0?ooo`030?oO
W018M;l0oooo0?l0oooo:`3oool00?l0ooooDP3oool00`3OW4P0B9cO0?ooo`0@0?ooo`030=nLB018
W=l0oooo0?l0oooo:P3oool00?l0ooooD@3oool00`3ookl0M01d0;ooo`0A0?ooo`030?oo_`1d04P0
W=oo0?l0oooo:P3oool00?l0oooo@P3oool01@2oM0000000001d_`3oool0gia80080000000<004RL
0=ooo`3oool01@3oool01@3OW4P004RL0=ooo`3oool0gia80080000000<0001809cOg`2LB0000P00
00000`0007@0_ooO09a8000200000006001d_`3oool0oooo0?oog`2LB000B9cOo`3ooolZ0?ooo`00
o`3ooom20?ooo`080?oog`2LB7@0_ooo0?ooo`3ooml0W4P00018W03Oool70?ooo`030;md0000M;l0
oooo0080oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l0
103oool00`3OW4P004RL0=ooo`3o0?ooobT0oooo003o0?oood<0oooo00H0oono07A8W03Oool0_g@0
0000M02oool70?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool00P3oool01`2oM000
04RL0=ooo`3oool0oooO09a80000M;l0103oool00`3OW4P0001d0;ooo`3o0?ooobT0oooo003o0?oo
od@0oooo00@0omnL07A8B0000000M;oo203oool00`3ookl0M00007Boo`030?ooo`030=nLB00007@0
_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo00@0oooo00<0okmd0000M02oool0
o`3ooolY0?ooo`00o`3ooom40?ooo`030?oo_`1d0000B9cO00T0oooo00<0oono07@0001d_ol00`3o
ool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`040?ooo`03
0?noM00004P0W=oo0?l0oooo:@3oool00?l0oooo@`3oool01@3ooml0W4P00018W02LM4P0W=oo00P0
oooo00<0oono07@0001d_ol00`3oool00`3OW4P0001809cOo`020?ooo`070;md000007@0_ooo0?oo
o`3ooml0W4P0001d_`040?ooo`030?noM00007@0_ooo0?l0oooo:@3oool00?l0oooo@`3oool01P2o
M000001809cOo`3ooml0W4Qd0;ooo`L0oooo0180oono07@00018W=l0oooo0?oog`2LB00000000000
B018B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l40?ooo`030=nLB00007@0_ooo
0?l0oooo:@3oool00?l0oooo@@3oool02@3ooml0W4P000000000000007Bo0?noM0000000001d0;oo
o`060?ooo`030?oog`2LB000B9cO00<0oooo00`0_g@00000M02ooml0W4P00000000004P0W=oo0?oo
_`1d000000000000M02oool40?ooo`030=nLB000B9`0gooo0?l0oooo:@3oool00?l0ooooD@3oool0
0`3OW4P004RL0=ooo`0A0?ooo`030?oog`2LB000B9cO0?l0oooo:P3oool00?l0ooooD@3oool00`3o
okl0M01809cOo`0A0?ooo`030?oo_`1d04P0W=oo0?l0oooo:P3oool00?l0ooooDP3oool00`3OW4P0
B9cO0?ooo`0@0?ooo`030=nLB018W=l0oooo0?l0oooo:P3oool00?l0ooooD`3oool00`2oM000B9cO
0?ooo`0>0?ooo`030?noM000M;l0oooo0?l0oooo:`3oool00?l0ooooE03oool00`3OW7@0W=oo0?oo
o`0<0?ooo`030?oo_`1dM;l0oooo0?l0oooo;03oool00?l0ooooo`3ooonB0?ooo`00o`3ooooo0?oo
oi80oooo003o0?ooool0ooooTP3oool00?l0ooooo`3ooonB0?ooo`00o`3ooooo0?oooi80oooo003o
0?ooool0ooooTP3oool00?l0ooooo`3ooonB0?ooo`00o`3ooooo0?oooi80oooo003o0?ooool0oooo
TP3oool00?l0ooooo`3ooonB0?ooo`00o`3ooooo0?oooi80oooo002e0?ooo`030;md000000000000
00@0000000T0B9cO0?ooo`3oool0oooo0?oOW018000000000000M02oool0:`3oool0103ooml0W4P0
0018W03Oool30?ooo`050?oOW018000000000000M02oool0:03oool01@3ogi`0B000000000000000
M;oo00@0oooo00D0omnL04P000000000001d0;ooo`0Y0?ooo`050=nLB0000000000000000018W=l0
103oool01@3ogi`0B0000000000007@0_ooo02@0oooo00H0_g@000000000000000000018W03Oool3
0?ooo`050?oOW018000000000000M02oool0103oool01@3ogi`0B0000000000007@0_ooo0200oooo
00P0_g@000000000000000000018W03Oool0oooo0;md00H0000000T0B9cO0?ooo`3oool0oooo0?oO
W018000000000000M02oool0803oool01P2oM000000000000000000004RL0=ooo`D0oooo00@0oooO
09a80000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo07h0oooo002f0?ooo`030=nLB000
0000000000<0000000X004RL0=ooo`3oool0oono07@0001d_ol0oooo0?noM00007@0_ooo:P3oool0
103ooml0W4P00018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0V0?ooo`07
0?oOW0180000B9cO0?ooo`3ookl0M00007Boo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`0W0?ooo`070=nLB0000000M;oo0?ooo`3ooml0W4P004RLg`020?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`0T0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0200oooo00<0omnL
04P0001d_ol0103oool00`3OW4P00000000000030000000:0018W03Oool0oooo0?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooob00oooo00<0omnL04P0001d_ol01`3oool0103ooml0W4P00018W03Oool2
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`1m0?ooo`00]`3oool00`3OW4P0B9cO0?oo
o`020?ooo`060=nLM02Lgol0oooo0;md0000B9`0gooo0P3oool00`2oM00007Bo0?ooo`0Y0?ooo`04
0?oog`2LB00004RL0=ooo`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0T0?oo
o`030?oog`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo0;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool09@3oool00`3ooml0W4P0001d_`030?ooo`070?noM000B9`0gooo0?ooo`2oM000
04RL0=ooo`020?ooo`030;md0000M;l0oooo02<0oooo00<0omnL04P0001d_ol0103oool00`2oM000
04RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo01l0oooo00<0omnL04P0001d_ol01@3oool00`3OW4P0B9cO0?ooo`020?ooo`060=nLM02Lgol0
oooo0;md0000B9`0gooo0P3oool00`2oM00007Bo0?ooo`0O0?ooo`030?oOW0180000M;oo00L0oooo
00@0oooO09a80000B9`0gooo0P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo07`0
oooo002h0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`0U0?ooo`030;md00000000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0
001d0;ooo`0T0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0oono07@00018W=l3
0?ooo`030=nLB00007@0_ooo02H0oooo00<0_g@0001d_`3oool00P3oool01P2oM00004RL0=ooo`3o
okl0M00004RLg`<0oooo00<0gia80000M02oool08`3oool00`3ogi`0B00007Boo`030?ooo`030?oo
_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo
01l0oooo00<0omnL04P0001d_ol01P3oool00`3OW4P0M;oo0?ooo`030?ooo`030?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool07`3oool00`3ogi`0B00007Boo`030?ooo`030;md000000000000
00D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`1l0?ooo`00^@3oool00`2o
M4P0W=oo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol09@3oool0303o
gi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000
B02Lgol0903oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`0V0?ooo`0;0?oOW01807@0_ooo0?ooo`2oM000000007Boo`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0S0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL
04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0
7`3oool00`3ogi`0B00007Boo`070?ooo`030;mdB02Lgol0oooo0080oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0O0?ooo`030?oOW0180000M;oo00<0oooo00`0omnL07BLg`3oool0
oooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo07`0oooo
002i0?ooo`030?oog`2LB7@0_ooo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0V0?ooo`0;0=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`0T0?ooo`030?oOW0180000M;oo00<0oooo00H0_g@00000B02Lgol0omnL04P0
001d_ol30?ooo`030?noM00004P0W=oo02L0oooo00D0oono07A8B0180000000004RLg`020?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08`3oool00`3ogi`0B00007Boo`030?ooo`03
0?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0
W=oo01l0oooo00<0omnL04P0001d_ol01`3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol07`3oool00`3ogi`0B00007Boo`040?ooo`0;0=nLB01d_ol0
oooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`1l0?oo
o`00^P3oool01`3o_g@007Bo0?ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0V
0?ooo`0;0?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0T0?ooo`0<0?oo_`1d0000000004RLg`3oool0okmd00000000M;l0oooo0?oOW018
0000M;oo0`3oool00`3o_g@0001809cOo`0W0?ooo`040?oOW01800000000001d_`<0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0S0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL
04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0
7`3oool00`3ogi`0B00007Boo`080?ooo`070?noM000M;l0oooo0?ooo`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo01l0oooo00<0omnL04P0001d_ol0103oool02`3ooml0W4Qd0;ooo`3ooml0
W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0O03oool00;X0oooo
00L0oono07@0B02Lgol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol09`3oool02P3o
gi`0B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo02D0
oooo00<0_g@00018W02oW4P00P0000001P00B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd
0000B02Lgol09P3oool01P3ogi`0B0000000M02oool0okmd001d_`80oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0S0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol0
0`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol07`3oool00`3o
gi`0B00007Boo`080?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo01l0oooo00<0omnL04P0001d_ol01@3oool02P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`3o
ool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo07`0oooo002j0?ooo`070?oo_`1d0000B9cO
0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo02P0oooo00T0gia807Bog`2LB00004RL
0=ooo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0U0?ooo`030?oOW0180000M;oo
00D0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0V0?ooo`030;md000007@0_ooo
0080oooo00H0_g@004RLg`3oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo02<0oooo00<0
omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000
B9cO0`3oool00`3OW4P0001d0;ooo`0O0?ooo`030?oOW0180000M;oo00P0oooo00L0oono07@00018
W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool07`3oool00`3ogi`0B00007Boo`06
0?ooo`090=nLB01d_ml0W4P00018W03Oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02o
ool0O03oool00;D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`020?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool0:03oool01@3ooml0W4Q804Q80000B9`0gooo0080
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0V0?ooo`030=nLB00004P0W=oo00D0
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0V0?ooo`030;md0000B9`0gooo0080
oooo00L0gia8001d_`3oool0oooo0=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool08`3oool0
0`3ogi`0B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool07`3oool00`3ogi`0B00007Boo`030?ooo`030?no
W02Lgol0oooo0080oooo00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md
0000M;l0oooo01l0oooo00<0omnL04P0001d_ol01P3oool01@3ooml0W4Q804Q80000B9`0gooo0080
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`1l0?ooo`00]@3oool0203ooml0W4Q8
09cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
:P3oool0103ogi`0B0000018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0X
0?ooo`030?noM000000007Bo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo02L0oooo
00L0okmd0000001d_ol0oooo0?oo_`1d0000M;oo0080oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo0280oooo00D0oooO09a800000000000007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0N0?ooo`050?oog`2L
B00000000000001d_ol00`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool0
1`3ookl0M00007Boo`3oool0okmd0000M02oool07P3oool01@3ooml0W4P0000000000000M;oo00L0
oooo00@0omnL04P00000B9`0gooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0O@3o
ool00;H0oooo00H0oooO09a80000000000000000M02oool40?ooo`050?oo_`1d000000000000M02o
ool0;03oool00`3OW4P004RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool0:P3oool01@3o
oml0W4P00000000007@0_ooo0080oooo00D0oono07@000000000001d0;ooo`0Y0?ooo`050?noM000
0000000000000018W=l0103oool01@3ookl0M0000000000007@0_ooo02D0oooo00<0okmd0000001d
_ol01@3oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`0Q
0?ooo`030?noM0000000M;oo00@0oooo00H0oooO09a80000000000000000M02oool40?ooo`050?oo
_`1d000000000000M02oool08@3oool00`3o_g@0000007Boo`080?ooo`030=nLB000B9`0gooo00<0
oooo00D0oono07@000000000001d0;ooo`1n0?ooo`00o`3ooooo0?oooi80oooo003o0?ooool0oooo
TP3oool00?l0ooooo`3ooonB0?ooo`00o`3ooooo0?oooi80oooo003o0?ooool0ooooTP3oool00?l0
ooooo`3ooonB0?ooo`00o`3ooooo0?oooi80oooo003o0?ooool0ooooTP3oool00?l0ooooo`3ooonB
0?ooo`00o`3ooooo0?oooi80oooo00270?ooool00000W@00001]0?ooo`00Q`3oool3000000/0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool01P3oool01P000000oooo0?ooo`000000oooo000000P0oooo00@000000?oo
o`3oool0oooo0P00000=0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0000000200000380oooo00<000000?oo
o`3oool0=03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0=03oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0180oooo00@000000?oo
o`3oool000002P3oool200000080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo06/0oooo00270?ooo`030000003oool0oooo00@00000<03oool00`00
07l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0=03oool00`00
07l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool04@3oool01@00
0000oooo0?ooo`3oool0000000T0oooo0P0000030?ooo`030000003oool0oooo00d0oooo00<0001o
0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool010000000oooo0?ooo`3oool5
000002h0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo00l0oooo0P0000030?ooo`030000003oool0oooo00L0oooo00<000000?ooo`0000000P3oool0
0`000000oooo0?ooo`0>0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool0
08L0oooo0`0000000`3oool0000000000005000002`0oooo00<0001o0?ooo`3oool0=03oool00`00
07l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`00
07l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`060?ooo`030000003oool000000080oooo00<000000?ooo`3oool03`3oool00`00
07l0oooo0?ooo`0F0?ooo`<00000K@3oool008L0oooo00<000000?ooo`3oool00P3oool200000004
0?ooo`00000000000?ooo`800000:P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo
01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool00`3oool200000005
0?ooo`00000000000000003oool00P00000X0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool0<`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool01@3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0100oooo00<0001o
0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`040?oo
o`80000000<0oooo0000003oool00P0000000`3oool000000000000V0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0<`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`0000020?ooo`030000003oool0
oooo0140oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`00
0000oooo0?ooo`050?ooo`D0000000@0oooo0000003oool0oooo0P00000T0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0<`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo0180oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool0
0`000000oooo0?ooo`060?ooo`060000003oool000000?ooo`000000oooo0P0000020?ooo`800000
8P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0
203oool2000000L0oooo00<000000?ooo`3oool01@3oool01P000000oooo0?ooo`000000oooo0000
01D0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool3000000L0
oooo00D000000?ooo`000000oooo000000020?ooo`040000003oool0oooo0?ooo`800000803oool0
0`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool01`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000000P3o
ool00`000000oooo0?ooo`0C0?ooo`030000O`3oool0oooo01H0oooo0`00001]0?ooo`00Q`3oool0
0`000000oooo0?ooo`080?ooo`040000003oool000000?ooo`80000000<0oooo0000000000000`3o
ool00`000000oooo0?ooo`0M0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0<`3o
ool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03o
ool00`0007l0oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103o
ool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo01@0oooo00<0001o0?ooo`3oool0
603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool000000080oooo00@000000?ooo`3oool000000`3oool2000001d0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo00@0oooo0P00000:0?ooo`030000
003oool0oooo00<0oooo00L000000?ooo`3oool0oooo0000003oool0000001L0oooo00<0001o0?oo
o`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool000000080oooo00<000000?ooo`3oool00P0000030?ooo`8000006`3oool00`0007l0
oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool00`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo00L000000?ooo`3oool0oooo0000003oool0
000001P0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`00
0000oooo0?ooo`0:0?ooo`040000003oool0oooo00000080oooo00D000000?ooo`3oool0oooo0000
00040?ooo`8000006@3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0000000H0?ooo`030000O`3oool0oooo01P0oooo00<0
00000?ooo`3oool0J`3oool008L0oooo0`00000;0?ooo`040000003oool0oooo00000080oooo0P00
00020?ooo`800000103oool2000001L0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0d0?ooo`030000O`3oool0oooo00800000303oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`0000006@3oool00`0007l0oooo0?ooo`0F0?ooo`<00000K@3o
ool008L0oooo00<000000?ooo`3oool0303oool010000000oooo0?ooo`0000030?ooo`050000003o
ool0oooo0?ooo`0000001@3oool2000001D0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0000000h0oooo00D000000?ooo`3oool0oooo000000040?ooo`03
0000003oool0000001X0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00
Q`3oool00`000000oooo0?ooo`0=0?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3o
ool0oooo0P0000050?ooo`8000004`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0000003oool03@3oool01@000000oooo0?ooo`3oool0000000D0oooo0P00000K
0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?oo
o`3oool03P3oool010000000oooo0?ooo`0000030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool200000140oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0c
0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo00d0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0000
000K0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<00000
0?ooo`3oool03`3oool010000000oooo0?ooo`0000030?ooo`8000000`3oool2000000H0oooo0P00
000?0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0<`3oool00`0007l0oooo0?oo
o`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0<P3oool2000000030000O`3o
ool0oooo00`0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0000000L0?ooo`030000
O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0403o
ool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0080oooo0P0000060?ooo`800000
3@3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo0340oooo00@000000?ooo`3oool0
001o3@3oool010000000oooo0?ooo`0000060?ooo`8000007@3oool00`0007l0oooo0?ooo`0H0?oo
o`030000003oool0oooo06/0oooo001_0?ooo`060;md00000000000000000000B9`0gooo0`3oool0
1@3ogi`0B0000000000007@0_ooo00X0oooo0`00000A0?ooo`040000003oool0oooo000000@0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`8000002`3oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo0300oooo00D000000?ooo`3oool0oooo0000O`0=0?ooo`<0
00001P3oool2000001h0oooo00<0001o0?ooo`3oool05P3oool3000006d0oooo001`0?ooo`030?oO
W0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00T0oooo00<000000?oo
o`3oool04P3oool010000000oooo0?ooo`0000040?ooo`800000103oool2000000L0oooo0P000009
0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0<`3oool00`0007l0oooo0?ooo`0d
0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0;P3oool2000000@0oooo00<0001o
0?ooo`3oool02P3oool2000000L0oooo00<000000?ooo`0000007P3oool00`0007l0oooo0?ooo`0H
0?ooo`030000003oool0oooo06/0oooo001`0?ooo`030?oOW0180000M;oo00@0oooo00<0_g@00018
W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo0180oooo00D000000?oo
o`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3oool2
000000L0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000003oool0
oooo00@0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`8000007`3oool0
0`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo001`0?ooo`030?oOW0180000M;oo
00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030000003oool0oooo
01<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00<0oooo0P000008
0?ooo`8000001@3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00<0001o
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo02`0oooo00<00000
0?ooo`3oool01@3oool00`0007l0oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo0P00000P
0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool00700oooo00<0omnL04P0
001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?oo
o`3oool0503oool010000000oooo0?ooo`3oool2000000@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`070?ooo`8000000`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo02X0oooo0P0000080?ooo`030000O`3oool0oooo00P0oooo0P0000070?ooo`8000008@3oool0
0`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo001`0?ooo`030?oOW0180000M;oo
00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`@00000500007l00`00
0000001o0000O`020000O`030000000007l0001o0080001o0P0000050000O`8000002@0007l30000
0?l0001o200007l00`000000001o0000O`0B0000O`800000200007l00`000000001o0000O`0h0000
O`@00000K@3oool00700oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`090?ooo`80
0000=@3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0X0?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3o
ool01P3oool00`000000oooo000000070?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3o
ool0603oool00`000000oooo0?ooo`1[0?ooo`00L03oool00`3ogi`0B00007Boo`030?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`800000
2@3oool00`0007l0oooo0000000200000380oooo00<0001o0?ooo`3oool0<`3oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool09P3oool2000000`0oooo
00<0001o0?ooo`3oool0103oool2000000030?ooo`000000oooo00H0oooo0P00000S0?ooo`030000
O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool00700oooo00<0omnL04P0001d_ol00`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo0P0000060?ooo`8000001`3o
ool00`0007l0oooo0?ooo`020?ooo`800000<03oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3o
ool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0T0?ooo`8000003P3o
ool00`0007l0oooo0?ooo`030?ooo`040000003oool0oooo000000L0oooo0P00000T0?ooo`030000
O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool00700oooo00<0omnL04P0001d_ol0103o
ool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01@3o
ool2000000D0oooo00<0001o0?ooo`3oool0103oool2000002h0oooo00<0001o0?ooo`3oool0<`3o
ool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool08P3o
ool200000100oooo00<0001o0?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000L0oooo
00<000000?ooo`3oool08`3oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo
001^0?ooo`050?oog`2LB00000000000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool02@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`060?ooo`8000000`3oool00`0007l0oooo0?ooo`060?oo
o`<00000:`3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0P0?ooo`8000004P3oool01@0007l0oooo0?ooo`3oool0
000000<0oooo00<000000?ooo`3oool01@3oool2000002D0oooo00<0001o0?ooo`3oool0603oool0
0`000000oooo0?ooo`1[0?ooo`00L03oool00`3o_g@0000007Boo`050?ooo`050?oo_`1d00000000
0000M02oool02P3oool3000001/0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`800000203oool2000000030?ooo`0007l0oooo00X0oooo0P00000Y0?ooo`030000O`3oool0
oooo03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo01l0oooo00<000000?ooo`3oool04P3oool0100007l0oooo0?ooo`0000040?ooo`030000003o
ool0oooo00@0oooo0P00000V0?ooo`030000O`3oool0oooo01H0oooo0`00001]0?ooo`00Q`3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`070?ooo`8000003@3oool2000002L0oooo00<0001o0?ooo`3oool0<`3oool0
0`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool07@3oool2
000001D0oooo00<0001o0?ooo`000000103oool00`000000oooo0?ooo`040?ooo`8000009`3oool0
0`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo00270?ooo`030000003oool0oooo
01`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`800000203oool00`00
07l000000000000=0?ooo`<00000903oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0K0?ooo`8000005`3oool00`00
07l000000?ooo`040?ooo`030000003oool0oooo00<0oooo0P00000X0?ooo`030000O`3oool0oooo
01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool0100007l0
oooo0?ooo`00000?0?ooo`8000008P3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0I0?ooo`800000603oool20000
00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0000000X0?ooo`030000O`3oool0oooo
01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`030?ooo`8000001`3oool2000000D0oooo00@0001o0?ooo`3oool0oooo0P00000?0?oo
o`800000803oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0G0?ooo`8000006@3oool00`000000oooo0000O`050?oo
o`030000003oool0oooo0080oooo00<000000?ooo`000000:@3oool00`0007l0oooo0?ooo`0H0?oo
o`030000003oool0oooo06/0oooo00270?ooo`<000007`3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`0007l0oooo0?ooo`030?ooo`80
00003`3oool3000001d0oooo00<0001o0?ooo`3oool0<`3oool00`0007l0oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool05P3oool00`000000oooo0?ooo`0H0?ooo`04
0000003oool0oooo0000O`@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0000000Z0?oo
o`030000O`3oool0oooo01H0oooo0`00001]0?ooo`00Q`3oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool01P3oool200000080oooo00<0001o0?ooo`3o
ool01@3oool200000100oooo0P00000K0?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01@0oooo0P00000J0?ooo`05
0000003oool0oooo0?ooo`0007l00`3oool00`000000oooo0?ooo`020?ooo`030000003oool00000
02/0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01`3oool200000003
0000O`3oool0oooo00L0oooo0P00000@0?ooo`<00000603oool00`0007l0oooo0?ooo`0c0?ooo`03
0000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0B0?ooo`80
00006P3oool2000000@0oooo00D0001o0?ooo`3oool0oooo000000030?ooo`040000003oool0oooo
000002/0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`00
0000oooo0?ooo`0R0?ooo`030000003oool0oooo00@0oooo0P0000090?ooo`030000003oool0oooo
00T0oooo0P00000A0?ooo`8000005P3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0@0?ooo`8000006`3oool00`00
0000oooo0?ooo`040?ooo`040000O`3oool0oooo000000<0oooo00@000000?ooo`3oool00000;03o
ool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo00270?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000O`000000
000000/0oooo0P00000A0?ooo`800000503oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0>0?ooo`8000006`3oool2
000000L0oooo00<0001o0?ooo`0000000`3oool010000000oooo0?ooo`00000]0?ooo`030000O`3o
ool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0903oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00@0001o0?ooo`3oool00000303o
ool00`000000oooo0?ooo`0@0?ooo`<000004@3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3o
ool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0<0?ooo`800000703o
ool00`000000oooo0?ooo`070?ooo`030000O`3oool0000000<0oooo00<000000?ooo`000000;P3o
ool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo00270?ooo`<000009@3oool0
0`000000oooo0?ooo`050?ooo`8000001@3oool0100007l0oooo0?ooo`3oool2000000/0oooo0P00
000C0?ooo`<000003P3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0:0?ooo`800000703oool2000000X0oooo00<0
001o0000003oool00P3oool00`000000oooo0000000_0?ooo`030000O`3oool0oooo01H0oooo0`00
001]0?ooo`00Q`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool00P3oool00`0007l0oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo0P00000D
0?ooo`<000002`3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`070?ooo`<000007@3oool00`000000oooo0?ooo`0:
0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0007l01P3oool2000000`0oooo0P00
000E0?ooo`<00000203oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`040?ooo`<000007P3oool2000000d0oooo00@0
00000?ooo`3oool000000P3oool00`000000oooo0?ooo`0^0?ooo`030000O`3oool0oooo01P0oooo
00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`060?ooo`040000003oool0oooo0000O`P0oooo00<000000?ooo`3oool02`3oool2000001H0oooo
0`0000050?ooo`030000O`3oool0oooo03<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo0080oooo0P00000P0?ooo`030000003oool0oooo00`0oooo
00@000000000O`3oool000000P3oool00`000000oooo0?ooo`0_0?ooo`030000O`3oool0oooo01P0
oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`060?ooo`80000000<0001o0?ooo`3oool01`3oool2000000d0oooo0P00000G0?ooo`<00000
0P3oool00`0007l0oooo0?ooo`0c0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0000000200000200oooo0P00000>0?ooo`040000003oool0001o00000080
oooo00<000000?ooo`3oool0<03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0
oooo00270?ooo`<00000:P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0303oool2000001P0oooo0`00000e0?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool0<`3oool300000280oooo00<000000?ooo`3oool03P3oool00`000000oooo
000000020?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool05P3oool3000006d0oooo
001J0?ooo`030;mdB02Lgol0oooo0100oooo00<0_g@0000000000000100000002@18W=l0oooo0?oo
o`3oool0omnL04P000000000001d0;ooo`0:0?ooo`030000003oool0oooo02/0oooo00<000000?oo
o`3oool01P3oool00`0007l000000?ooo`0:0?ooo`8000003P3oool2000001P0oooo00@0001o0000
000000000000<P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo0340oooo0P000000
0`3oool0001o0?ooo`0P0?ooo`800000403oool3000000040000O`3oool0oooo000003<0oooo00<0
001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00FP3oool00`2oM000000007Boo`06
0?ooo`030?oog`2LB00007Bo00P0oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3o
okl0M00007Boo`3oool0okmd0000M02oool90?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool01@3oool00`0007l0oooo0000000<0?ooo`030000003oool0oooo00d0oooo0P00000F0?oo
o`040000O`3oool0oooo0?ooo`@00000;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo02h0oooo0`0000030?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool03`3oool2
00000080oooo00<0001o0?ooo`000000=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0
oooo06/0oooo001J0?ooo`050;md0000000000000000001d_ol01@3oool00`2oM00004RL0=ooo`08
0?ooo`030=nLB018W=l0oooo0080oooo00H0giad09cOo`3oool0_g@00018W03Oool20?ooo`030;md
0000M;l0oooo00P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`040?ooo`030000
O`3oool0oooo008000002`3oool2000000l0oooo0P00000D0?ooo`030000O`3oool0oooo00D0oooo
0`00000[0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0:`3oool3000000H0oooo
00<0001o0?ooo`3oool0703oool200000140oooo0P0000030?ooo`030000O`000000oooo03@0oooo
00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00FP3oool01`2oW9`0gooo0?oo
g`2LB00000000000B02Lgol00P3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB01d_ol0oooo00<0
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030000003oool0oooo02h0
oooo00<000000?ooo`3oool00`3oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo00X0
oooo0P00000?0?ooo`8000004P3oool00`0007l0oooo0?ooo`080?ooo`@000009`3oool00`0007l0
oooo0?ooo`0d0?ooo`030000O`3oool0oooo02P0oooo0`0000090?ooo`030000O`3oool0oooo01/0
oooo00<000000?ooo`3oool0403oool2000000@0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo001N0?ooo`030?oo_`1d0000000000800000
00<007Bo0?ooo`3oool02`3oool00`2oM4P0W=oo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0080oooo
00<0001o0?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00h0oooo
0P00000@0?ooo`030000O`3oool0oooo00`0oooo1000000S0?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool09@3oool3000000`0oooo00<0001o0?ooo`3oool06@3oool200000180oooo
0P0000050?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo
0?ooo`1[0?ooo`00GP3oool0103ooml0W4P00000001d_ol?0?ooo`030?oog`2LB7@0_ooo0080oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`@00000;`0007l2000000T0001o
00<000000000O`0007l02`0007l200000100001o0P00000Q0000O`@00000N`0007l3000002X0001o
00<000000000O`0007l04@0007l2000000D0001o00<000000000O`0007l0C@0007l4000006d0oooo
001J0?ooo`060;nLW03Oool0omnL04P0000004P0W=oo4P3oool01`3o_g@007Bo0?ooo`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo0380oooo00<00000
0?ooo`0007l01`3oool2000000d0oooo00<000000?ooo`3oool03`3oool3000000/0oooo00<0001o
0?ooo`3oool0503oool4000001/0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0N
0?ooo`@000004P3oool00`0007l0oooo0?ooo`0F0?ooo`8000004`3oool00`000000oooo00000004
0?ooo`030000003oool0001o03H0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[
0?ooo`00FP3oool0102oM00000000018W03OoolD0?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0<`3oool00`000000001o
0?ooo`080?ooo`030000003oool0oooo00/0oooo0P00000B0?ooo`8000002@3oool00`0007l0oooo
0?ooo`0H0?ooo`@000005`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01/0oooo
0`00000F0?ooo`030000O`3oool0oooo01@0oooo0P00000C0?ooo`80000000<0oooo0000003oool0
0`3oool010000000oooo0?ooo`0007lf0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3o
ool0J`3oool005X0oooo00<0_gAd0;ooo`3oool05@3oool01`3ookl0M00004RLg`3oool0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0140oooo0P0000070?oo
o`030000O`3oool0oooo01`0oooo1000000C0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool05`3oool4000001T0oooo00<0001o0?ooo`3oool04P3oool2000001@0oooo00@000000?oo
o`3oool00000103oool01@000000oooo0?ooo`3oool0001o03H0oooo00<0001o0?ooo`3oool0603o
ool00`000000oooo0?ooo`1[0?ooo`00K@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000
B9cO0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0
oooo03@0oooo00<0001o0000003oool02@3oool2000000d0oooo0P00000C0?ooo`8000001@3oool0
0`0007l0oooo0?ooo`0P0?ooo`D000003P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo0180oooo1@00000M0?ooo`030000O`3oool0oooo0100oooo0P00000E0?ooo`050000003oool0
oooo0?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00K@3oool0203ooml0W4Q809cOo`3o
ool0oono07@0000007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool02@3oool3
000003@0oooo00<0001o0?ooo`0000002`3oool00`000000oooo0?ooo`0<0?ooo`8000004`3oool3
00000080oooo00<0001o0?ooo`3oool09@3oool5000000T0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0=0?ooo`D000008P3oool00`0007l0oooo0?ooo`0>0?ooo`8000005@3oool2
000000<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool05P3oool3000006d0oooo001^0?ooo`060?oog`2LB00000000000000007@0_ooo
103oool01@3ookl0M0000000000007@0_ooo00X0oooo00<000000?ooo`3oool0=03oool0100007l0
oooo0?ooo`00000;0?ooo`030000003oool0oooo00d0oooo0P00000D0?ooo`80000000<0001o0?oo
o`3oool0:P3oool6000000<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`070?oo
o`H000009`3oool00`0007l0oooo0?ooo`0<0?ooo`8000005P3oool00`000000oooo0?ooo`020?oo
o`050000003oool0oooo0?ooo`0000001P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0=03oool01@00
07l0oooo0?ooo`3oool0000000/0oooo00<000000?ooo`3oool03P3oool2000001@0oooo0`00000`
0?ooo`T00000<@3oool00`0007l0oooo00000007000002d0oooo00<0001o0?ooo`3oool02P3oool2
000001L0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000070?ooo`030000O`3o
ool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3oool02@3oool2
00000100oooo00<000000?ooo`3oool04@3oool00`0007l0oooo0?ooo`0200000340oooo00<0001o
0?ooo`3oool00`3oool=000001X0oooo3000000e0?ooo`030000O`3oool0oooo00P0oooo0P00000G
0?ooo`8000001@3oool010000000oooo0?ooo`0000080?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00F03oool00`3ooml0W4P000000006
00000003001d_`3oool0oooo02<0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`03
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool200000140oooo00<0001o
0?ooo`3oool00P3oool2000002l0oooo00<0001o0?ooo`3oool0403ooolJ000000X0oooo00<0001o
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`060?ooo`800000603oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo000000T0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0H0?ooo`030000003oool0oooo06/0oooo001F0?ooo`030?oog`2LB000000000X0000000<007Bo
0?ooo`3oool08@3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo00@0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?ooo`8000003`3oool00`0007l0oooo0?ooo`04
0?ooo`<00000;03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o
0?ooo`3oool0103oool2000001T0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00
000:0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?oo
o`1[0?ooo`00E@3oool0103o_g@000000018W03Oool90?ooo`040?oo_`1d0000001809cOob40oooo
0`00000d0?ooo`030000O`3oool0oooo00D0oooo00<000000?ooo`3oool02P3oool200000140oooo
0P00000=0?ooo`030000O`3oool0oooo00L0oooo0P00000Z0?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`020?ooo`8000006P3oool00`000000oooo
0?ooo`050?ooo`030000003oool0000000/0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0F0?ooo`<00000K@3oool005@0oooo00<0oono07Ad_`3oool03@3oool00`3ooml0W7BL0=oo
o`0P0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02P3oool00`0007l0oooo0?oo
o`090?ooo`<000009`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0
001o0?ooo`3oool00P00000J0?ooo`8000001`3oool00`000000oooo0000000<0?ooo`030000O`3o
ool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo00L0oooo00<000000?ooo`3oool02`3oool2
00000140oooo0P00000:0?ooo`030000O`3oool0oooo00`0oooo1000000S0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool3000001/0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0000000=0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool0
0`000000oooo0?ooo`1[0?ooo`00FP3oool00`3ookl0M7Bo0?ooo`040?ooo`030=nLW02oool0oooo
02<0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`080?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0403oool2000000P0oooo00<0001o0?ooo`3oool0403oool30000
0200oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0a0?ooo`<0000000<0001o0?oo
o`3oool06P3oool00`000000oooo0?ooo`060?ooo`030000003oool0000000h0oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo001J0?ooo`03
0?noM0000000000000@0000000<0001d0;ooo`3oool08`3oool00`000000oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo00T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0A0?ooo`80
00001P3oool00`0007l0oooo0?ooo`0C0?ooo`@00000703oool00`0007l0oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo02h0oooo0`0000030?ooo`030000O`3oool0oooo01P0oooo0P0000080?ooo`03
0000003oool0000000l0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`03
0000003oool0oooo06/0oooo001J0?ooo`030;md00000000000000@0000000<0001d0;ooo`3oool0
8`3oool3000003@0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0<0?ooo`800000
4`3oool2000000@0oooo00<0001o0?ooo`3oool05`3oool4000001P0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0[0?ooo`<000001P3oool00`0007l0oooo0?ooo`0G0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`000000403oool00`0007l0oooo0?ooo`0d0?ooo`030000
O`3oool0oooo01H0oooo0`00001]0?ooo`00F`3oool00`2oM7@0_ooo0?ooo`030?ooo`030=nLW02o
ool0oooo02<0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0;0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool04P3oool01@000000oooo0?ooo`3oool0001o01d0oooo
0`00000E0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0:03oool3000000T0oooo
00<0001o0?ooo`3oool05P3oool00`000000oooo0?ooo`080?ooo`8000004@3oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool005X0oooo00<0
omnL09boo`3oool0:P3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo00`0oooo00<0
00000?ooo`3oool03@3oool2000001<0oooo0P0000000`3oool0001o0?ooo`0O0?ooo`@000004@3o
ool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo02D0oooo0`00000<0?ooo`030000O`3o
ool0oooo01@0oooo0P00000:0?ooo`8000004P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3o
ool0oooo01P0oooo00<000000?ooo`3oool0J`3oool005X0oooo00<0_g@0001d_`3oool0103oool0
0`3OW9`0_ooo0?ooo`0S0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool03@3oool2
000000l0oooo00<000000?ooo`3oool04P3oool2000002@0oooo1000000=0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool08P3oool3000000l0oooo00<0001o0?ooo`3oool04`3oool0
0`000000oooo0?ooo`090?ooo`8000004`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo01P0oooo00<000000?ooo`3oool0J`3oool005X0oooo00<0gia8000000000000100000000`00
07@0_ooo0?ooo`0S0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool03`3oool00`00
0000oooo0?ooo`0=0?ooo`8000004`3oool00`0007l000000000000V0?ooo`@000002@3oool00`00
07l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01h0oooo1000000B0?ooo`030000O`3oool0oooo
0140oooo0P00000;0?ooo`800000503oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
01P0oooo00<000000?ooo`3oool0J`3oool005X0oooo00<0oooO09a800000000100000000`0007@0
_ooo0?ooo`0S0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0403oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0100oooo00<0001o0?ooo`3oool00P00000X0?ooo`@0
00001@3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01/0oooo0`00000F0?ooo`03
0000O`3oool0oooo00l0oooo0P00000<0?ooo`8000005@3oool00`0007l0oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool005/0oooo00<0gibL0;ooo`3oool0
0`3oool00`3OW9`0_ooo0?ooo`090?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0
0`3oool01@3ogi`0B0000000000007@0_ooo00X0oooo0`00000d0?ooo`030000O`3oool0oooo0140
oooo00<000000?ooo`3oool03P3oool200000100oooo00<0001o0?ooo`3oool00P3oool2000002X0
oooo100000000`3oool0001o0?ooo`0e0?ooo`030000O`3oool0oooo01H0oooo1@00000I0?ooo`03
0000O`3oool0oooo00h0oooo00<000000?ooo`3oool02`3oool2000001H0oooo00<0001o0?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0F0?ooo`<00000K@3oool005X0oooo00<0okmd04RLg`3oool0
403oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo0080oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00T0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0B0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`040?oo
o`800000;03oool600000380oooo00<0001o0?ooo`3oool04@3oool5000001h0oooo00<0001o0?oo
o`3oool0303oool2000000d0oooo0P00000G0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00FP3oool00`2oM00007Bo0?ooo`040?ooo`03
0=nLW02oool0oooo00h0oooo00<0oono07@0B02Lgol00P3oool00`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0C0?ooo`03
0000003oool0oooo00l0oooo0P00000=0?ooo`030000O`3oool0oooo00H0oooo0`00000Z0?ooo`03
0000O`3oool0oooo0080oooo2000000Z0?ooo`030000O`3oool0oooo00d0oooo1000000S0?ooo`03
0000O`3oool0oooo00X0oooo0P00000>0?ooo`800000603oool00`0007l0oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool005X0oooo00<0gia8000000000000
100000000`0007@0_ooo0?ooo`0>0?ooo`070?oo_`1d0000M;oo0?ooo`3ookl0M00004RLg`030?oo
o`030=nLB00007@0_ooo00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0D0?oo
o`8000004@3oool00`000000oooo0?ooo`0:0?ooo`030000O`3oool0oooo00T0oooo0`00000W0?oo
o`030000O`3oool0oooo00X0oooo2000000R0?ooo`030000O`3oool0oooo00<0oooo2P00000W0?oo
o`030000O`3oool0oooo00P0oooo0P00000?0?ooo`8000006@3oool00`0007l0oooo0?ooo`0d0?oo
o`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool005X0oooo00<0oooO09a80000
0000100000000`0007@0_ooo0?ooo`0>0?ooo`070?noM0000000M;oo0?ooo`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0F
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0<
0?ooo`8000009@3oool00`0007l0oooo0?ooo`0B0?oooa4000001`3oool@00000340oooo00<0001o
0?ooo`3oool01P3oool2000000l0oooo0`00000J0?ooo`030000O`3oool0oooo03@0oooo00<0001o
0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00H@3oool00`3OW9`0_ooo0?ooo`0=0?oo
o`080?oOW0180000001809cOo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00P0oooo
1000001=0000O`030000000007l0001o00l0001o0P00000J0000O`<00000A`0007l8000004P0001o
0P00000@0000O`030000000007l0000006X0001o1000001]0?ooo`00K`3oool01@3o_g@000000000
000004P0W=oo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`030000
003oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo00H0oooo00<0001o0?ooo`3oool04@3oool3000001l0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo0080oooo0P00000A0?ooo`030000
003oool0000001`0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000
003oool0oooo06/0oooo001D0?ooo`030?oo_`1dB9`0gooo00d0oooo00<0omnL04Q8W03Oool02P3o
ool00`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203o
ool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo01T0oooo00<000000?ooo`3oool0403o
ool2000000H0oooo00<0001o0?ooo`3oool0503oool2000001d0oooo00<0001o0?ooo`3oool0=03o
ool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool00000008000004P3oool010000000oooo0?oo
o`00000L0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo
0?ooo`1[0?ooo`00E@3oool0103ogi`0B0000018W03Oool90?ooo`040?oo_`1d000004RL0=ooo``0
oooo00L0_g@004RLg`3oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`030?ooo`030000O`3oool0oooo01H0oooo0`00000J0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool2000001<0oooo0P0000020?ooo`030000003oool0
oooo01/0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0
oooo06/0oooo001F0?ooo`030?oog`2LB000000000X0000000<007Bo0?ooo`3oool01`3oool00`3o
oml0W9bo0?ooo`020?ooo`030=nLB00007@0_ooo0080oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`080?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool06`3oool00`00
0000oooo0?ooo`0A0?ooo`8000000`3oool00`0007l0oooo0?ooo`0I0?ooo`<000005`3oool00`00
07l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo0340oooo0`0000000`0007l0oooo0?ooo`0A0?oo
o`050000003oool0oooo0?ooo`0000007P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo01P0oooo00<000000?ooo`3oool0J`3oool005P0oooo00<0oooO09a8000000001P0000000`00
M;l0oooo0?ooo`0:0?ooo`060?noM018W=l0oooo0?oo_`1d000007Bo0`3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool02@3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo01`0
oooo0P00000C0?ooo`040000003oool0oooo0000Oah0oooo0P00000E0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0;`3oool2000000<0oooo00<0001o0?ooo`3oool0403oool01@00
0000oooo0?ooo`3oool0000001l0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H
0?ooo`030000003oool0oooo06/0oooo001_0?ooo`050=nLB000000000000000M02oool0103oool0
1@3ookl0M0000000000007@0_ooo00X0oooo0`00000d0?ooo`030000O`3oool0oooo01h0oooo00<0
00000?ooo`3oool04@3oool2000000030000O`3oool0oooo01h0oooo0`00000B0?ooo`030000O`3o
ool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool2000000D0oooo00<0001o0?ooo`3oool03P3o
ool2000000<0oooo00<000000?ooo`3oool07P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3o
ool0oooo01H0oooo0`00001]0?ooo`00Q`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0
oooo01l0oooo00<000000?ooo`3oool04P3oool200000280oooo0`00000?0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0:P3oool3000000L0oooo00<0001o0?ooo`3oool03@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo00270?ooo`030000003oool0oooo
03@0oooo00<0001o0?ooo`3oool0803oool00`000000oooo0?ooo`0A0?ooo`040000O`3oool00000
000002<0oooo0`00000<0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0:03oool2
000000X0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo0200oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0
oooo06/0oooo00270?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool08@3oool00`00
0000oooo0?ooo`0@0?ooo`040000O`3oool0oooo0?ooo`800000903oool3000000T0oooo00<0001o
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0U0?ooo`<00000303oool00`0007l0oooo0?ooo`0:
0?ooo`800000103oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03@0oooo00<0001o
0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`0d0?oo
o`030000O`3oool0oooo0280oooo00<000000?ooo`3oool03`3oool00`0007l0oooo0?ooo`030?oo
o`8000009@3oool3000000H0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0R0?oo
o`<000003`3oool00`0007l0oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool08P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?oo
o`3oool0J`3oool008L0oooo0`00000d0?ooo`030000O`3oool0oooo02<0oooo0P00000?0?ooo`03
0000O`3oool0oooo00D0oooo0P00000V0?ooo`<000000`3oool00`0007l0oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo01l0oooo0`00000B0?ooo`030000O`3oool0oooo00P0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0S0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3o
ool05P3oool3000006d0oooo00270?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0
9@3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0oooo00L0oooo00<000000?ooo`3oool0
9P3oool4000003H0oooo00<0001o0?ooo`3oool06`3oool4000001D0oooo00<0001o0?ooo`3oool0
1P3oool2000000D0oooo00<000000?ooo`3oool0903oool00`0007l0oooo0?ooo`0d0?ooo`030000
O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0=03o
ool00`0007l0oooo0?ooo`0V0?ooo`030000003oool0oooo00/0oooo00<0001o0?ooo`3oool0203o
ool2000002T0oooo00<0001o0000000000000`00000a0?ooo`030000O`3oool0oooo01L0oooo1000
000I0?ooo`030000O`3oool0oooo00@0oooo0P0000050?ooo`8000009`3oool00`0007l0oooo0?oo
o`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<00000
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo00X0oooo00<0001o
0?ooo`3oool02P3oool2000002L0oooo00<0001o0?ooo`3oool00`3oool6000002/0oooo00<0001o
0?ooo`3oool04`3oool4000001d0oooo00<0001o0?ooo`3oool00`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H
0?ooo`030000003oool0oooo06/0oooo00270?ooo`030000003oool0oooo03@0oooo00<0001o0?oo
o`3oool0:03oool2000000X0oooo00<0001o0?ooo`3oool0303oool2000002D0oooo00<0001o0?oo
o`3oool02@3oool6000002D0oooo00<0001o0?ooo`3oool03`3oool400000240oooo00@0001o0?oo
o`3oool0oooo0P0000060?ooo`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo00270?ooo`<00000=03oool00`00
07l0oooo0?ooo`0Z0?ooo`030000003oool0oooo00L0oooo00<0001o0?ooo`3oool03P3oool20000
02<0oooo00<0001o0?ooo`3oool03`3oool500000200oooo00<0001o0?ooo`3oool02P3oool50000
02D0oooo00@0001o0?ooo`00000000001`3oool00`000000oooo0?ooo`0Y0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool05P3oool3000006d0oooo00270?ooo`030000003oool0oooo
03@0oooo00<0001o0?ooo`3oool0:`3oool2000000L0oooo00<0001o0?ooo`3oool0403oool20000
0240oooo00<0001o0?ooo`3oool0503oool;000001D0oooo3@00000Z0?ooo`030000O`000000oooo
00L0oooo00<000000?ooo`3oool0:P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool04P3oool2000001l0
oooo00<0001o0?ooo`3oool07`3ooolE000000030000O`3oool0oooo03<0oooo0P0000080?ooo`03
0000003oool0oooo02/0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`03
0000003oool0oooo06/0oooo00270?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0
;P3oool2000000@0oooo00<0001o0?ooo`3oool0503oool2000001d0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0a0?ooo`80000000<0oooo0000O`3oool01P3oool00`000000oooo
0?ooo`0/0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo
0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo0300oooo00D0
00000?ooo`3oool0oooo0000O`0H0?ooo`8000006`3oool00`0007l0oooo0?ooo`0d0?ooo`030000
O`3oool0oooo02l0oooo0P0000030?ooo`030000O`3oool0oooo00@0oooo00<000000?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0
J`3oool008L0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0a0?ooo`80000000<0
oooo0000O`3oool06@3oool2000001T0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0^0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3oool00`3oool00`000000oooo0?oo
o`0^0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?oo
o`1[0?ooo`00L@3oool0103ooml0W4P00018W03Oool30?ooo`050?oOW018000000000000M02oool0
2P3oool3000003@0oooo00<0001o0?ooo`3oool0<`3oool00`000000001o0?ooo`0K0?ooo`<00000
5P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo02`0oooo0P0000060?ooo`030000
O`3oool0oooo0080oooo00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`0d0?ooo`030000
O`3oool0oooo01H0oooo0`00001]0?ooo`00L@3oool0103ooml0W4P00018W03Oool20?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030000003oool0oooo03@0oooo00<0001o0?oo
o`3oool0=03oool2000001h0oooo0P00000D0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?oo
o`3oool0:P3oool2000000P0oooo00<0001o0?ooo`3oool00P00000b0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00L@3oool0103ooml0
W4P00018W03Oool20?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool00`00
0000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00@0001o0?ooo`00000000007P3oool3
00000140oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0X0?ooo`8000002P3oool0
0`0007l0oooo0000000d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool0
0`000000oooo0?ooo`1[0?ooo`00K@3oool00`2oM0000000000000050000000407Boo`3ookl0M000
04RLg`<0oooo00<0gia80000M02oool0203oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0
oooo03@0oooo00D0001o0?ooo`3oool0oooo0000000P0?ooo`8000003`3oool00`0007l0oooo0?oo
o`0d0?ooo`030000O`3oool0oooo02D0oooo0`00000<0?ooo`030000O`000000oooo03@0oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo001]
0?ooo`0<0?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool0
0`3o_g@0001809cOo`080?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`020?ooo`800000803oool3000000`0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0S0?ooo`8000003P3oool2000003H0oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo001^0?ooo`0;0=nLB01d_ol0oooo
0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`@0
0000L@0007l00`000000001o0000O`0P0000O`@00000HP0007l3000000l0001o00<000000000O`00
07l0Q@0007l4000006d0oooo001^0?ooo`0;0?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`3oool0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo03@0oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`050?ooo`800000903oool3000000D0oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0M0?ooo`<000004@3oool010000000oooo0?oo
o`0007lf0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo
0?ooo`1[0?ooo`00K`3oool02P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol3
0?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d
0?ooo`030000O`3oool0oooo00L0oooo00<000000?ooo`3oool0903oool4000000030?ooo`0007l0
oooo03D0oooo00<0001o0?ooo`3oool06P3oool300000180oooo0P0000030?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0
oooo06/0oooo001`0?ooo`090=nLB01d_ml0W4P00018W03Oool0oooo0?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool0203oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool0203oool2000002P0oooo0`00000e0?ooo`030000O`3oool0oooo01P0oooo
0P00000D0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool00700oooo00D0
oooO09a8B018B00004RL0=ooo`020?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0
203oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0
2P3oool00`000000oooo0?ooo`0V0?ooo`030000O`3oool0000000<00000<@3oool00`0007l0oooo
0?ooo`0E0?ooo`<000005@3oool00`000000oooo0?ooo`040?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo
001a0?ooo`040?oOW018000004RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo
00T0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
00/0oooo0P00000V0?ooo`030000O`3oool0oooo00<0oooo0`00000^0?ooo`030000O`3oool0oooo
0140oooo1000000F0?ooo`8000001`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo
03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00LP3oool00`3OW4P0
04RL0=ooo`030?ooo`050?oo_`1d000000000000M02oool02P3oool3000003@0oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0=0?ooo`800000903oool00`0007l0oooo0?ooo`060?oo
o`@00000:P3oool00`0007l0oooo0?ooo`0;0?ooo`H000006@3oool00`000000oooo0?ooo`070?oo
o`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0F0?oo
o`<00000K@3oool008L0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo00l0oooo0P00000R0?ooo`030000O`3oool0oooo00X0oooo1P00000T0?ooo`03
0000O`3oool0oooo00D0oooo1P00000N0?ooo`030000003oool0oooo00P0oooo00<0001o0?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3o
ool0J`3oool008L0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000
O`3oool0oooo0140oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0@0?ooo`H00000
7P3oool00`0007l000000000000500000280oooo0P00000;0?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo
00270?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0B0?ooo`8000007`3oool00`0007l0oooo0?ooo`0F0?oooal00000:03oool00`000000oooo0?oo
o`0;0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0H0?ooo`030000003oool0oooo06/0oooo00270?ooo`030000003oool0oooo03@0oooo00<0001o
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0D0?ooo`8000007@3oool00`0007l0oooo0?ooo`0d
0?ooo`030000O`3oool0oooo02@0oooo0P00000>0?ooo`030000O`3oool0oooo03@0oooo00<0001o
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo06/0oooo00270?oo
o`<00000=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01H0oooo0P00000K0?oo
o`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool08P3oool200000100oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01H0oooo0`00001]0?oo
o`00Q`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3o
ool0603oool2000001T0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0P0?ooo`80
00004P3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3o
ool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`0d0?ooo`030000
O`3oool0oooo03@0oooo00<0001o0?ooo`3oool06P3oool2000001L0oooo00<0001o0?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0N0?ooo`800000503oool00`0007l0oooo0?ooo`0d0?ooo`030000
O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3o
ool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0703o
ool3000001@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0L0?ooo`8000005P3o
ool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603o
ool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool07`3oool300000140oooo00<0001o0?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0J0?ooo`800000603oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0
oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`1[0?ooo`00Q`3oool00`00
0000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool08P3oool20000
00l0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0H0?ooo`8000006P3oool00`00
07l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`00
0000oooo0?ooo`1[0?ooo`00Q`3oool3000003@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0T0?ooo`<00000303oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01H0
oooo0P00000L0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0F0?ooo`<00000K@3oool008L0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo02L0oooo0P00000:0?ooo`030000O`3oool0oooo03@0oooo
00<0001o0?ooo`3oool0503oool2000001h0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo
0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<0
00000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo02T0oooo0P00
00080?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool04@3oool300000200oooo00<0
001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<0
00000?ooo`3oool0J`3oool008L0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d
0?ooo`030000O`3oool0oooo02/0oooo0`0000050?ooo`030000O`3oool0oooo03@0oooo00<0001o
0?ooo`3oool03P3oool3000002<0oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d
0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0J`3oool008L0oooo00<000000?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo02h0oooo0P0000030?oo
o`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0303oool2000002H0oooo00<0001o0?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01P0oooo00<000000?oo
o`3oool0J`3oool008L0oooo0`00000d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3o
ool0<03oool3000000030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool02@3oool3000002P0
oooo00<0001o0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo01H0
oooo0`00001]0?ooo`00K@3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo00<0oooo
00D0omnL04P000000000001d0;ooo`0:0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0c0?ooo`@00000<`3oool00`0007l0oooo0?ooo`050?ooo`@0
0000:`3oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3o
ool0603oool00`000000oooo0?ooo`1[0?ooo`00K@3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md
001d_ol0oooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00T0oooo00<000000?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0d0?ooo`030000O`3oool0oooo03<0oooo00@0001o0?oo
o`3oool0oooo1000000_0?ooo`040000O`3oool0oooo0?ooo`@00000;`3oool00`0007l0oooo0?oo
o`0d0?ooo`030000O`3oool0oooo03@0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?oo
o`1[0?ooo`00LP3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0_g@0
001d_`3oool0203oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo03@0oooo00<00000
0?ooo`3oool0<`3oool00`000000oooo0?ooo`050?ooo`D00000:P3oool010000000oooo00000000
000c0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo06/0oooo001c0?ooo`060;md001d_ol0oooo0?oo_`1d0000B9cO
0`3oool00`3OW4P0001d0;ooo`080?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo200000070?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool0103oool8000000`0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`1[0?ooo`00LP3oool01`3ooml0W4P004RLg`3oool0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0103ooolL000000H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo06/0oooo001b0?ooo`070=nLB0000000M;oo0?ooo`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00P0ooooo`00002M000006d0oooo001^0?ooo`0;0=oOg`3oool0oono07@0
00000000001d0;ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`3o0?ooool0oooo
4`3oool006h0oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol0o`3ooooo0?oooa<0oooo001^0?ooo`040?oo_`1d0000000004RLg`@0
oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`3o0?ooool0oooo4`3oool006l0oooo
00<0_g@007Boo`3oool01@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0?l0oooo
o`3ooolC0?ooo`00K`3oool01P3ogi`0B000000000000000001809cOo`80oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo0?l0ooooo`3ooolD0?ooo`00L03oool01@2oM0000000000000000000
B9cO00<0oooo00D0oono07@000000000001d0;ooo`3o0?ooool0oooo5@3oool00?l0ooooo`3ooonB
0?ooo`00o`3oool30?ooo`060?oog`2LB00000000000000007@0_oooW@3oool00`3ooml0W4P00000
00030000000304RLg`3oool0oooo0>80oooo003o0?ooo`D0oooo00<0_g@00018W03Oool0WP3oool0
0`2oM00004RL0=ooo`030?ooo`030=nLW02oool0oooo0>80oooo003o0?ooo`D0oooo00<0_g@00018
W03Oool0WP3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0>80oooo003o0?ooo`D0
oooo00<0_g@00018W03Oool0W`3oool00`2oM0000000000000030000000304RLg`3oool0oooo0>40
oooo003]0?ooo`030?oog`2LB000000000H0000000<007Bo0?ooo`3oool0103oool01@2oM0000000
0000000007@0_ooo0080oooo00<0_g@00018B01800000P0000000`18W=l0oooo0?ooo`020?ooo`07
0;md0000000000000000M02oool0oooo0;md0002000000030000M02ooml0W4P00080000000@004RL
0=ooo`3ooml0W4P00P000000100004P0B4P00018W03Oool20?ooo`070;md0000000000000000M02o
ool0oooo0;md0002000000030000M02ooml0W4P00080000000@004RL0=ooo`3oool0_g@00P000000
0`0004P0W=oo0?ooo`020?ooo`050;md0000000000000000M02oool01P3oool01@3OW4P000000000
00000000M;oo0080oooo00H0_g@0000000000000001d0;oog`2LB002000000030018W03Oool0oooo
00<0oooo00<0gia80000000000000P0000001`1d_ol0_g@0000000000000001d0;oog`2LB0000P00
00002`00B9`0_ia8000000000000001d0;ooo`3ogi`0B00000000018W=l0gia80080000000<0001d
0;oog`2LB0000P0000001@00B9`0W7@000000000M;l0gia80080000000<0001d0;ooo`3oool0103o
ool01@3ookl0M000000000000000B9cO00<0oooo00H0gia800000000000000000018W02LM0020000
00030000M02ooml0W4P00080000000<004RL0=oo_`1d0000100000000`0007@0_ooo0=nLB0020000
00040000M02oool0oooo0;md0080000000<0001d0;ooo`3oool0e`3oool00>l0oooo00<0omnL04P0
0018W=l00`3oool0103o_g@0000000000018W=l30?ooo`0?0;md0000000007Bo0?ooo`3ogi`0B01d
0;ooo`3oool0_g@00000001d_ol0oooo0?oo_`1d0000B9cO0080oooo02T0_g@000000000M;l0oooo
0?oOW01807@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?oog`2LB0000000
07Boo`3oool0oono07@000000000001d0;ooo`2oM0000000001d_`3oool0omnL04P0M02oool0oooo
0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`0?0;md0000000004RL0=ooo`3ooml0W4Qd
0;ooo`3oool0_g@000000000M;l0oooo0?oOW01807@0_ooo00@0oooo00L0gia800000018W=l0oooo
0?oog`2LB000B9cO0080oooo00L0gia80018W03Oool0oooo0?oog`2LB00007Bo00L0oooo00<0oooO
09a80000M;l00`3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool0603OW4P0001d
0;ooo`3oool0gia80000M02oool0gkno0=ooo`3OW4P0001d0;ooo`3o_g@0000007Boo`3oool0omnL
04P00000M;l0okno0;ooo`3OW4P0001d0;ooo`L0oooo00<0oono07@0B02Lgol01@3oool00`3OW4P0
000004RLg`020?ooo`070=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`030;mdB02Lgol0
oooo00D0oooo00/0gia80000M02oool0oooo0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3G0?oo
o`00k`3oool00`3ookl0M00004RLg`040?ooo`080?oo_`1d0000000007Boo`3oool0omnL04P00018
W=l30?ooo`060?oOW01dW=l0oooo0;md0000B9`0gooo0P3oool01P3o_g@0001d0;ooo`3ogi`0B000
04RLg`<0oooo00h0omnL07BLg`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0
B00004RLg`80oooo00L0oooO09a80000M;l0oooo0?oOW0180000B9cO00<0oooo00h0omnL07BLg`3o
ool0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0B000001d_`<0oooo00D0omnL09bo
o`3ogi`0B00004RLg`030?ooo`030?oOW01dW=l0oooo0080oooo00<0omnL04P00000M;l00`3oool0
2`3o_g@0001d0;ooo`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo00L0oooo00<0oooO09a8
0000M;l00`3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool01`3OW4P0001d0;oo
o`3oool0gia80018W03Oool00P3oool02P3OW4P0001d0;ooo`3o_g@0001809cOo`3oool0oooO09a8
0000M;l20?ooo`030=nLB00007@0_ooo00P0oooo00<0okmd001d_`3oool00`3oool0103ooml0W4P0
0000M02oool20?ooo`070=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`050?oo_`1d0000
000000000000M;l00`3oool01P3OW4P0001d0;ooo`3ogi`0B00004RLg`<0oooo00<0omnL07BLg`3o
ool0eP3oool00>l0oooo00<0oono07@00018W=l01@3oool01`3ogi`0B0000018W03Oool0okmd0000
B02Lgol01P3oool00`2oM00004RL0=ooo`020?ooo`060?oOW0180000M;oo0?noM00004P0W=oo1P3o
ool02`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo0080oooo00L0oooO
09a80000M;l0oooo0?noM00004P0W=oo00H0oooo00/0gia80018W03Oool0oooo0?oog`2LB00007Bo
0?ooo`3o_g@0001809cOo`050?ooo`030?noM00004P0W=oo00P0oooo00<0okmd0000001d_ol00`3o
ool02`3o_g@0000007Boo`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo00L0oooo00<0oooO
09a80000M;l00`3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool01`3OW4P0001d
0;ooo`3oool0gia80018W03Oool00P3oool02P3OW4P0001d0;ooo`3oool0okmd00000000M;l0oooO
09a80000M;l20?ooo`030=nLB00007@0_ooo00P0oooo00<0oono07@0M02oool00`3oool00`3ogi`0
B00004RLg`030?ooo`070=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`060=nLB00007@0
_ooo0?oo_`1d0000B9cO0P3oool01P3OW4P0001d0;ooo`3o_g@0001809cOom`0oooo003_0?ooo`03
0?oo_`1d0000B9cO00D0oooo00L0oooO09a8000004P0W=oo0?noM00007@0_ooo00H0oooo00<0_g@0
0018W03Oool00P3oool01P3ogi`0B00007Boo`3o_g@0001d0;ooo`H0oooo00/0gia80018W03Oool0
oooo0?oog`2LB00007Bo0?ooo`3o_g@0000007Boo`020?ooo`070?oog`2LB00007Bo0?ooo`3o_g@0
001d0;ooo`060?ooo`0;0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000M02oool0
1@3oool00`3o_g@0001d0;ooo`080?ooo`030?noM0000000M;oo00<0oooo00/0okmd0000001d_ol0
oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`070?ooo`030?oog`2LB00007Bo00<0oooo00L0
gia80018W03Oool0oooo0?oog`2LB00007Bo0080oooo00L0gia80000M02oool0oooo0=nLB000B9`0
gooo0080oooo00<0gia80000M02oool00`3oool0103o_g@0000000000000M;l20?ooo`030=nLB000
07@0_ooo00T0oooo00<0_g@00000000000000`0000000`0004P0W=oo0?ooo`020?ooo`0?0=nLB000
B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0oono07@00018W=l0oooo0?oog`2LB00007Bo0080oooo
00H0gia80000M02oool0okmd0000M02ooooL0?ooo`00k`3oool00`3ookl0M00004RLg`060?ooo`05
0;md00000000M;oo0?oOW01800001@0000001@18W=l0oooo0;md000007@0_ooo0080oooo00D0gia8
0000B02Lgol0omnL04P000050000000=04RLg`3oool0_g@00000M02oool0oooo0?oog`2LB00007Bo
0?ooo`3ookl0M00004RLg`020?ooo`060?oog`2LB00007Bo0?ooo`3ogi`0B0001@0000003@18W=l0
oooo0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P0B02Lgol00P3oool01@3o_g@0
001809cOo`3ogi`0B00000D0000000L0B9cO0?ooo`3oool0oooo0?oo_`1d0000M;oo00<0oooo00/0
_g@00000B02Lgol0oooo0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`070?ooo`030?oog`2LB000
07Bo00<0oooo00L0_g@00000M02oool0oooo0?oog`2LB00007Bo0080oooo00L0gia80000M02oool0
oooo0=nLB000B9`0gooo0080oooo00X0gia80000M02oool0oono07@0001d_ol0oooo0?oog`2LB000
07Bo0P3oool00`3OW4P0001d0;ooo`090?ooo`080?oOW018B9`0gooo0?ooo`3ooml0W4P00018W03O
ool30?ooo`0?0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0oono07@00018W=l0oooo0?oo
g`2LB000B9cO0080oooo00D0gia80000M02oool0omnL04P000050000000304RLg`3oool0oooo0=D0
oooo003_0?ooo`030?oo_`1d0000B9cO00H0oooo00H0_g@00000001d_ol0oooo0;md0018W=l20?oo
o`0?0=nLB00004P0W=no07@00000000004Q804Q8W03Oool0omnL04P00000B9`0gooo0?ooo`2oM000
B9cO0080oooo00/0gia80000B02Lgml0W4P000000000B4P0B01809cOo`3ookl0M00004RLg`020?oo
o`070?noM00004P0W=oo0?ooo`3o_g@00000001d_`020?ooo`0?0;md0018W=l0oooo0?ooo`3OW4P0
001809cOg`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0080oooo00X0_g@00018W03Oool0
oooO09a8000007@0_ooo0?ooo`2oM000B9cO0P3oool00`3OW4P0001809cOo`040?ooo`0@0=nLB000
B9`0gooo0?ooo`3o_g@00000001d_`3ooml0W4P000000000B4P0B01809cOo`3ookl0M00004RLg`L0
oooo0280oooO09a80000M;l0oooo0?oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0?oo
g`2LB00000000000M02oool0oooo0=nLB000B9`0gooo0?oog`2LB00000000000M02oool0oono07@0
0018W=l0oooo0?oo_`1d000007Bo0P3oool00`3OW4P0001d0;ooo`090?ooo`070?oog`2LB000M;oo
0?ooo`3ogi`0B00004RLg`020?ooo`090?oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO
0080oooo00@0_g@0001d_`3oool0okmd0P0000002000B9`0gooo0=nLB00007@0_ooo0?ooo`2oM000
B9cO0P3oool00`3OW4P0001809cOo`3G0?ooo`00k`3oool00`3ookl0M00004RLg`050?ooo`040?oo
g`2LB000001809cOo`80oooo00h0gia8000000000000001809cOo`3oool0oooO09a80000B9`0gmnL
04P000000000001d0;ooo`<0oooo00D0gia8000000000000001809cOo`020?ooo`070;md0000B9`0
gooO09a800000000001d0;ooo`030?ooo`030?noM000000000000080000000D007Bo0?ooo`3oool0
oooo0=nLB0020000000;0000B02Lgol0oooo0?ooo`2oM00004RL0=oog`2LB00000000000M02oool0
0`3oool01@3OW4P000000000000004P0W=oo00<0oooo00D0gia8000000000000001809cOo`060?oo
o`050?noM0000000000000000018W=l00`3oool01`2oM00004RL0=oog`2LB00000000000M02oool0
1`3oool00`3ooml0W4P0001d_`030?ooo`070;md0000B9`0gooO09a800000000001d0;ooo`020?oo
o`040;md000007@0_ooo0;md00<0000000P004RL0=ooo`2oM000001d0;ooo`3oool0oono07@00080
000000L0001d0;ooo`3oool0oooo0=nLB00007@0_ooo00X0oooo00H0okmd0018W03Oool0gia80000
B02Lgol40?ooo`070;md0000B9`0gooO09a800000000001d0;ooo`020?ooo`050?oog`2LB0000000
0000001d_ol00`3oool00`3OW4P0001d0;ooo`020?ooo`050=nLB000000000000000B02Lgol0f03o
ool00>l0oooo00<0oono07@00018W=l01@3oool0103ogi`0B0000018W03OoolW0?ooo`030?oog`2L
B00007Bo03d0oooo00<0oooO09a80000M;l0403oool00`2oM00004RL0=ooo`0>0?ooo`030=nLB000
07@0_ooo00X0oooo00H0oono07@0B02Lgml0W4P00018W03OoolE0?ooo`030=nLB00007@0_ooo0=l0
oooo003_0?ooo`030?oo_`1d0000B9cO00@0oooo00@0oooO09a800000000M;oo:03oool00`3ooml0
W4P0001d_`0m0?ooo`030?oog`2LB00007Bo0100oooo00<0oono07A8W03Oool03P3oool00`3OW4P0
001d0;ooo`0;0?ooo`040=nLB000B4P0B00004RLgaH0oooo00<0gia80000M02oool0g`3oool00>l0
oooo00<0omnL04P00018W=l00`3oool0103o_g@0000000000018W=lY0?ooo`030?oog`2LB00007Bo
03d0oooo00<0oooO09a80000M;l0403oool00`3ooml0W7BL0=ooo`0>0?ooo`030=nLB00007@0_ooo
00/0oooo00@0omnL04P0000004P0W=oo5P3oool00`3OW4P0001d0;ooo`3O0?ooo`00k@3oool00`3o
oml0W4P000000006000000030000M02oool0oooo02P0oooo00@0oono07@00000000007Bo?03oool0
0`3OW4P00000000000020000000307Boo`3oool0oooo00P0oooo00<0gia80000B02Lgol01P3oool0
0`3OW4P0001809cOo`070?ooo`050?oog`2LB00000000000M02oool0303oool00`2oM00004RL0=oo
o`0D0?ooo`050?oog`2LB00000000000M02oool0g`3oool00?l0oooo8`3oool00`3ooml0W4P0001d
_`1<0?ooo`030=nLB00004P0W=oo00H0oooo00<0gia80000B02Lgol02@3oool00`2oM000001d0;oo
o`0U0?ooo`030;md000007@0_ooo0=l0oooo003o0?ooool0ooooTP3oool00?l0ooooo`3ooonB0?oo
o`00o`3ooooo0?oooi80oooo003o0?ooool0ooooTP3oool00001\
\>"],
  ImageRangeCache->{{{0, 655.438}, {217.813, 0}} -> {-49.5011, -19.0579, \
0.547655, 0.547655}}]
}, Closed]],

Cell[TextData[
"The ideal angle is apparently near 40\[Degree]; the angular shift is small \
for baseball but is larger for golf.  We leave more precise determination of \
this angle as an exercise."], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Dependence of trajectories upon damping coefficients", "Section"],

Cell[CellGroupData[{

Cell["\<\
Dependence on linear coefficient, with quadratic term neglected.\
\>", "Subsection"],

Cell[BoxData[{
    \(\(\[Alpha]Parameters = {g \[Rule] 9.8, 
          m \[Rule] 0.145, \[Beta] \[Rule] 0};\)\), "\n", 
    \(\(\[Beta]Parameters = {g \[Rule] 9.8, 
          m \[Rule] 0.145, \[Alpha] \[Rule] 0};\)\)}], "Input"],

Cell[BoxData[
    \(\(plotDamping[\(x\_0\) : _, \ \(y\_0\) : _, \(v\_0\) : _, \(\[Theta]\_0\
\) : _, parameters_, damping_, title_] := \n\t
        ParametricPlot[
          Evaluate[
            Table[resistedTrajectory[x\_0, y\_0, v\_0, \[Theta]\_0, 
                parameters], \ damping]\ ], {t, 0, 10}, 
          PlotRange \[Rule] {{0, 300}, {0, 60}}, Frame \[Rule] True, 
          GridLines \[Rule] Automatic, AspectRatio \[Rule] Automatic, 
          FrameLabel \[Rule] {"\<x (m)\>", "\<y (m)\>", 
              title, "\< \>"}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotDamping[0, 0, 50, \ 
        40  Degree, \ \[Alpha]Parameters, \ {\[Alpha], 0, 0.1, 
          0.02}, "\<Linear Damping\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .2 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-3.20924e-017 0.00333333 -8.67362e-019 0.00333333 [
[.16667 -0.0125 -8 -15.6875 ]
[.16667 -0.0125 8 0 ]
[.33333 -0.0125 -11 -15.6875 ]
[.33333 -0.0125 11 0 ]
[.5 -0.0125 -11 -15.6875 ]
[.5 -0.0125 11 0 ]
[.66667 -0.0125 -11 -15.6875 ]
[.66667 -0.0125 11 0 ]
[.83333 -0.0125 -11 -15.6875 ]
[.83333 -0.0125 11 0 ]
[1 -0.0125 -11 -15.6875 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -15.0625 -31.375 ]
[.5 -0.0125 15.0625 -15.6875 ]
[-0.0125 .03333 -16 -7.84375 ]
[-0.0125 .03333 0 7.84375 ]
[-0.0125 .06667 -16 -7.84375 ]
[-0.0125 .06667 0 7.84375 ]
[-0.0125 .1 -16 -7.84375 ]
[-0.0125 .1 0 7.84375 ]
[-0.0125 .13333 -16 -7.84375 ]
[-0.0125 .13333 0 7.84375 ]
[-0.0125 .16667 -16 -7.84375 ]
[-0.0125 .16667 0 7.84375 ]
[-0.0125 .2 -16 -7.84375 ]
[-0.0125 .2 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .1 -31.6875 -14.8125 ]
[-0.01875 .1 -16 14.8125 ]
[ 0 .2 .125 0 ]
[.5 .2125 -41.125 0 ]
[.5 .2125 41.125 15.6875 ]
[ 1 0 .125 0 ]
[1.0125 .1 0 -3.5 ]
[1.0125 .1 15.6875 3.5 ]
[ 0 0 0 0 ]
[ 1 .2 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.16667 0 m
.16667 .2 L
s
.33333 0 m
.33333 .2 L
s
.5 0 m
.5 .2 L
s
.66667 0 m
.66667 .2 L
s
.83333 0 m
.83333 .2 L
s
0 .03333 m
1 .03333 L
s
0 .06667 m
1 .06667 L
s
0 .1 m
1 .1 L
s
0 .13333 m
1 .13333 L
s
0 .16667 m
1 .16667 L
s
0 g
.16667 0 m
.16667 .00625 L
s
gsave
.16667 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.33333 0 m
.33333 .00625 L
s
gsave
.33333 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.66667 0 m
.66667 .00625 L
s
gsave
.66667 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(200) show
1.000 setlinewidth
grestore
.83333 0 m
.83333 .00625 L
s
gsave
.83333 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(250) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(300) show
1.000 setlinewidth
grestore
.125 Mabswid
.03333 0 m
.03333 .00375 L
s
.06667 0 m
.06667 .00375 L
s
.1 0 m
.1 .00375 L
s
.13333 0 m
.13333 .00375 L
s
.2 0 m
.2 .00375 L
s
.23333 0 m
.23333 .00375 L
s
.26667 0 m
.26667 .00375 L
s
.3 0 m
.3 .00375 L
s
.36667 0 m
.36667 .00375 L
s
.4 0 m
.4 .00375 L
s
.43333 0 m
.43333 .00375 L
s
.46667 0 m
.46667 .00375 L
s
.53333 0 m
.53333 .00375 L
s
.56667 0 m
.56667 .00375 L
s
.6 0 m
.6 .00375 L
s
.63333 0 m
.63333 .00375 L
s
.7 0 m
.7 .00375 L
s
.73333 0 m
.73333 .00375 L
s
.76667 0 m
.76667 .00375 L
s
.8 0 m
.8 .00375 L
s
.86667 0 m
.86667 .00375 L
s
.9 0 m
.9 .00375 L
s
.93333 0 m
.93333 .00375 L
s
.96667 0 m
.96667 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -76.0625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
72.125 15.813 moveto
(H) show
75.938 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.313 15.813 moveto
(L) show
89.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .03333 m
.00625 .03333 L
s
gsave
-0.0125 .03333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .06667 m
.00625 .06667 L
s
gsave
-0.0125 .06667 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .1 m
.00625 .1 L
s
gsave
-0.0125 .1 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
0 .13333 m
.00625 .13333 L
s
gsave
-0.0125 .13333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .16667 m
.00625 .16667 L
s
gsave
-0.0125 .16667 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
0 .2 m
.00625 .2 L
s
gsave
-0.0125 .2 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .00667 m
.00375 .00667 L
s
0 .01333 m
.00375 .01333 L
s
0 .02 m
.00375 .02 L
s
0 .02667 m
.00375 .02667 L
s
0 .04 m
.00375 .04 L
s
0 .04667 m
.00375 .04667 L
s
0 .05333 m
.00375 .05333 L
s
0 .06 m
.00375 .06 L
s
0 .07333 m
.00375 .07333 L
s
0 .08 m
.00375 .08 L
s
0 .08667 m
.00375 .08667 L
s
0 .09333 m
.00375 .09333 L
s
0 .10667 m
.00375 .10667 L
s
0 .11333 m
.00375 .11333 L
s
0 .12 m
.00375 .12 L
s
0 .12667 m
.00375 .12667 L
s
0 .14 m
.00375 .14 L
s
0 .14667 m
.00375 .14667 L
s
0 .15333 m
.00375 .15333 L
s
0 .16 m
.00375 .16 L
s
0 .17333 m
.00375 .17333 L
s
0 .18 m
.00375 .18 L
s
0 .18667 m
.00375 .18667 L
s
0 .19333 m
.00375 .19333 L
s
.25 Mabswid
0 0 m
0 .2 L
s
gsave
-0.01875 .1 -99.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
75.8125 11.8438 translate
90 rotate
-75.8125 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
71.625 15.813 moveto
(H) show
75.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
84.813 15.813 moveto
(L) show
88.625 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .19375 m
0 .2 L
s
.16667 .19375 m
.16667 .2 L
s
.33333 .19375 m
.33333 .2 L
s
.5 .19375 m
.5 .2 L
s
.66667 .19375 m
.66667 .2 L
s
.83333 .19375 m
.83333 .2 L
s
.125 Mabswid
.03333 .19625 m
.03333 .2 L
s
.06667 .19625 m
.06667 .2 L
s
.1 .19625 m
.1 .2 L
s
.13333 .19625 m
.13333 .2 L
s
.2 .19625 m
.2 .2 L
s
.23333 .19625 m
.23333 .2 L
s
.26667 .19625 m
.26667 .2 L
s
.3 .19625 m
.3 .2 L
s
.36667 .19625 m
.36667 .2 L
s
.4 .19625 m
.4 .2 L
s
.43333 .19625 m
.43333 .2 L
s
.46667 .19625 m
.46667 .2 L
s
.53333 .19625 m
.53333 .2 L
s
.56667 .19625 m
.56667 .2 L
s
.6 .19625 m
.6 .2 L
s
.63333 .19625 m
.63333 .2 L
s
.7 .19625 m
.7 .2 L
s
.73333 .19625 m
.73333 .2 L
s
.76667 .19625 m
.76667 .2 L
s
.8 .19625 m
.8 .2 L
s
.86667 .19625 m
.86667 .2 L
s
.9 .19625 m
.9 .2 L
s
.93333 .19625 m
.93333 .2 L
s
.96667 .19625 m
.96667 .2 L
s
.25 Mabswid
0 .2 m
1 .2 L
s
gsave
.5 .2125 -102.125 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Linear) show
96.813 15.813 moveto
(Damping) show
141.250 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 0 m
1 0 L
s
.99375 .03333 m
1 .03333 L
s
.99375 .06667 m
1 .06667 L
s
.99375 .1 m
1 .1 L
s
.99375 .13333 m
1 .13333 L
s
.99375 .16667 m
1 .16667 L
s
.125 Mabswid
.99625 .00667 m
1 .00667 L
s
.99625 .01333 m
1 .01333 L
s
.99625 .02 m
1 .02 L
s
.99625 .02667 m
1 .02667 L
s
.99625 .04 m
1 .04 L
s
.99625 .04667 m
1 .04667 L
s
.99625 .05333 m
1 .05333 L
s
.99625 .06 m
1 .06 L
s
.99625 .07333 m
1 .07333 L
s
.99625 .08 m
1 .08 L
s
.99625 .08667 m
1 .08667 L
s
.99625 .09333 m
1 .09333 L
s
.99625 .10667 m
1 .10667 L
s
.99625 .11333 m
1 .11333 L
s
.99625 .12 m
1 .12 L
s
.99625 .12667 m
1 .12667 L
s
.99625 .14 m
1 .14 L
s
.99625 .14667 m
1 .14667 L
s
.99625 .15333 m
1 .15333 L
s
.99625 .16 m
1 .16 L
s
.99625 .17333 m
1 .17333 L
s
.99625 .18 m
1 .18 L
s
.99625 .18667 m
1 .18667 L
s
.99625 .19333 m
1 .19333 L
s
.25 Mabswid
1 0 m
1 .2 L
s
gsave
1.0125 .1 -56.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
64.5 11.8438 translate
90 rotate
-64.5 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
66.000 15.813 moveto
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .2 L
0 .2 L
closepath
clip
newpath
.5 Mabswid
0 0 m
.05179 .04077 L
.10828 .07911 L
.16133 .10929 L
.21235 .133 L
.24078 .14395 L
.26665 .1525 L
.29383 .16004 L
.31892 .16569 L
.34459 .17017 L
.35812 .17199 L
.37245 .17353 L
.38507 .17454 L
.39216 .17496 L
.39864 .17527 L
.40465 .17547 L
.40806 .17556 L
.40968 .17559 L
.41122 .17561 L
.41268 .17563 L
.41425 .17565 L
.41569 .17566 L
.417 .17567 L
.41843 .17567 L
.42 .17567 L
.42164 .17566 L
.42317 .17565 L
.42486 .17563 L
.42666 .17561 L
.42988 .17554 L
.43335 .17546 L
.43718 .17533 L
.44403 .17503 L
.45046 .17466 L
.46243 .17375 L
.47533 .17246 L
.5022 .16869 L
.53048 .16315 L
.58157 .14909 L
.63595 .12838 L
.68829 .10285 L
.74392 .0697 L
.79752 .03189 L
s
.79752 .03189 m
.83685 0 L
s
0 0 m
.05037 .03963 L
.10219 .07444 L
.14805 .0996 L
.18975 .1173 L
.21032 .12404 L
.23168 .12953 L
.24187 .13159 L
.25277 .13336 L
.2577 .13401 L
.26289 .1346 L
.26772 .13506 L
.27211 .13539 L
.27429 .13552 L
.27666 .13565 L
.27892 .13575 L
.28096 .13582 L
.28332 .13588 L
.28462 .1359 L
.28584 .13592 L
.28697 .13593 L
.28801 .13593 L
.28914 .13593 L
.29034 .13592 L
.29139 .13591 L
.29234 .1359 L
.29344 .13588 L
.29448 .13585 L
.29632 .1358 L
.29832 .13572 L
.30282 .13548 L
.30696 .13518 L
.31552 .13432 L
.32346 .13322 L
.33204 .13169 L
.34116 .12968 L
.35788 .12489 L
.37497 .11845 L
.40453 .10335 L
.43426 .08258 L
.46128 .05826 L
.48836 .02804 L
s
.48836 .02804 m
.50918 0 L
s
0 0 m
.049 .03853 L
.09655 .07013 L
.13621 .091 L
.15453 .09858 L
.1703 .1039 L
.18642 .10805 L
.19491 .10965 L
.19864 .11022 L
.20269 .11075 L
.20618 .11112 L
.20989 .11142 L
.21191 .11155 L
.21375 .11164 L
.21463 .11167 L
.21558 .1117 L
.21641 .11173 L
.21731 .11175 L
.2182 .11176 L
.21903 .11177 L
.21984 .11177 L
.22058 .11177 L
.22138 .11176 L
.22226 .11175 L
.22317 .11173 L
.22402 .1117 L
.22556 .11165 L
.22696 .11158 L
.23008 .11138 L
.2333 .11108 L
.23667 .11069 L
.2426 .10976 L
.2488 .10845 L
.25545 .10665 L
.26675 .10256 L
.27824 .09693 L
.28865 .0904 L
.29898 .08241 L
.3161 .06528 L
.33236 .04357 L
.34631 .01967 L
s
.34631 .01967 m
.35549 0 L
s
0 0 m
.04768 .03746 L
.09132 .06614 L
.10982 .07612 L
.12563 .08335 L
.14054 .08886 L
.15351 .09243 L
.15997 .09372 L
.16326 .09425 L
.16667 .09469 L
.16948 .09497 L
.17094 .09508 L
.1725 .09518 L
.17323 .09522 L
.17391 .09525 L
.17458 .09527 L
.1752 .09529 L
.17577 .0953 L
.17639 .09531 L
.17706 .09532 L
.17767 .09532 L
.17835 .09532 L
.17899 .09531 L
.17968 .0953 L
.18041 .09528 L
.18118 .09525 L
.1819 .09522 L
.18325 .09514 L
.18563 .09495 L
.18817 .09467 L
.19088 .09427 L
.1937 .09376 L
.19858 .09259 L
.20297 .09121 L
.2075 .08945 L
.21522 .08554 L
.223 .08027 L
.23049 .0737 L
.2424 .05947 L
.25309 .04142 L
.26174 .02167 L
s
.26174 .02167 m
.26897 0 L
s
0 0 m
.04641 .03644 L
.08648 .06245 L
.10269 .07085 L
.11617 .07653 L
.12781 .08027 L
.13381 .08169 L
.13897 .08258 L
.14152 .0829 L
.14289 .08304 L
.14415 .08314 L
.14521 .08321 L
.14633 .08327 L
.14697 .08329 L
.14756 .08331 L
.1481 .08332 L
.14867 .08332 L
.149 .08332 L
.14935 .08332 L
.14967 .08332 L
.14997 .08332 L
.15054 .0833 L
.15114 .08329 L
.15177 .08326 L
.15236 .08323 L
.15368 .08315 L
.15481 .08305 L
.15583 .08294 L
.15806 .08263 L
.16007 .08226 L
.16215 .0818 L
.16574 .08077 L
.16939 .07938 L
.17318 .07755 L
.17935 .07357 L
.18528 .06832 L
.19033 .06245 L
.19506 .05549 L
.20226 .04118 L
.20539 .03307 L
.20836 .02392 L
.21329 .00454 L
s
.21329 .00454 m
.21414 0 L
s
0 0 m
.04518 .03546 L
.06414 .0484 L
.08198 .05903 L
.09635 .06612 L
.10768 .07045 L
.11312 .07203 L
.11791 .0731 L
.12041 .07352 L
.12263 .07381 L
.12376 .07393 L
.12498 .07402 L
.12558 .07406 L
.12622 .07409 L
.12653 .07411 L
.12682 .07412 L
.12737 .07413 L
.12789 .07414 L
.12818 .07414 L
.12844 .07414 L
.12893 .07414 L
.12943 .07413 L
.12989 .07412 L
.1303 .0741 L
.13124 .07405 L
.13176 .07401 L
.13224 .07397 L
.13313 .07388 L
.13509 .07362 L
.13672 .07332 L
.13818 .07298 L
.14132 .07202 L
.14427 .0708 L
.14686 .06944 L
.14954 .06769 L
.15217 .06561 L
.15665 .061 L
.16022 .05612 L
.16354 .05031 L
.16623 .04443 L
.17088 .03065 L
.17292 .0225 L
.17458 .01448 L
s
.17458 .01448 m
.1769 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{490.625, 98.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000;O0000Tb000`400?l00000o`00003oo`3ooooo
0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ0?ooo`00o`3ooomj0?ooo`030?oo
g`2LB4P0W=oo00h0oooo00<0omnL04Q8W03Oool0o`3ooomC0?ooo`00o`3ooomj0?ooo`030;mdB02L
gol0oooo00l0oooo00<0omnL04Qd_`3oool0o`3ooomB0?ooo`00o`3ooomi0?ooo`030=nLB018W=l0
oooo0140oooo00<0gia804RLg`3oool0o`3ooomA0?ooo`00o`3ooomh0?ooo`030?oo_`1d07@0_ooo
0180oooo00<0oono07@0B02Lgol0o`3ooomA0?ooo`00o`3ooomY0?ooo`050;md0000000007Bo0?oo
o`3OW4P00P0000000`00B9`0gooo0?ooo`050?ooo`030=nLB000B9`0gooo0080oooo00H0gia80000
00000000001809cOg`2LB002000000030000M02ooml0W4P00080000000H007Bo0?ooo`3oool0oooO
09a80018W=oo0?oooe40oooo003o0?ooofT0oooo00P0oooO09a8M02oool0oooo0?oog`2LB00004RL
0=ooo`L0oooo00<0_g@0001d_`3oool00`3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0
gooo0?ooo`3ooml0W4P0001d_`040?ooo`030=nLB000B9`0gooo0?l0ooooD03oool00?l0ooooJP3o
ool01P3ookl0M4RL0=ooo`2oM000001d0;ooo`L0oooo00<0oono07@00018W=l0103oool00`3OW4P0
001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`040?ooo`030=nLB00007@0
_ooo0?l0ooooD03oool00?l0ooooJ`3oool0103ogi`0M4Q80000001d_ol80?ooo`030?oo_`1d0000
M;oo00@0oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l0
103oool00`3o_g@0001d0;ooo`3o0?oooe00oooo003o0?ooof/0oooo00<0oono07@00018W=l02@3o
ool00`3ookl0M00007Boo`040?ooo`030=nLB00007@0_ooo0080oooo00L0_g@00018W03Oool0oooo
0?oog`2LB00007Bo00@0oooo00<0okmd0000B02Lgol0o`3ooom@0?ooo`00o`3ooomZ0?ooo`050?oo
g`2LB00004RL09adB02Lgol0203oool00`3ookl0M00007Boo`040?ooo`030=nLB00004P0W=oo0080
oooo00L0_g@00000M02oool0oooo0?oog`2LB00007Bo00@0oooo00<0okmd0000M02oool0o`3ooom@
0?ooo`00o`3ooomZ0?ooo`060;md000004P0W=oo0?oog`2LB7@0_ooo1`3oool00`3ookl0M00004RL
g`020?ooo`0>0?oog`2LB00000000000B018B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono07@0
0018W=l40?ooo`030=nLB00007@0_ooo0?l0ooooD03oool00?l0ooooJ03oool02@3ooml0W4P00000
0000000007Bo0?noM0000000001d0;ooo`060?ooo`030?oog`2LB000B9cO00@0oooo00`0_g@00000
M02ooml0W4P00000000004P0W=oo0?oo_`1d000000000000M02oool40?ooo`030=nLB000B9`0gooo
0?l0ooooD03oool00?l0ooooN03oool00`3OW4P004RL0=ooo`0B0?ooo`030?oog`2LB000B9cO0?l0
ooooD@3oool00?l0ooooN03oool00`3ookl0M01809cOo`0B0?ooo`030?oo_`1d04P0W=oo0?l0oooo
D@3oool00?l0ooooN@3oool00`3OW4P0B9cO0?ooo`0A0?ooo`030=nLB018W=l0oooo0?l0ooooD@3o
ool00?l0ooooNP3oool00`2oM000B9cO0?ooo`0?0?ooo`030?noM000M;l0oooo0?l0ooooDP3oool0
0?l0ooooN`3oool00`3OW7@0W=oo0?ooo`0=0?ooo`030?oo_`1dM;l0oooo0?l0ooooD`3oool00?l0
ooooo`3ooooQ0?ooo`00o`3ooooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ
0?ooo`00o`3ooooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ0?ooo`00o`3o
oooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ0?ooo`00o`3ooooo0?ooon40
oooo003o0?ooo`@0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`030?ooo`050?oO
W018000000000000M02oool09P3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0omnL
04P000000000001d0;ooo`040?ooo`050?oOW018000000000000M02oool08@3oool02@2oM0000000
00000000000004RL0=ooo`3oool0oooO09a800030000000307Boo`3oool0oooo00<0oooo00D0omnL
04P000000000001d0;ooo`0O0?ooo`030;md00000000000000@0000000T0B9cO0?ooo`3oool0oooo
0?oOW018000000000000M02oool0103oool01@3ogi`0B0000000000007@0_ooo01l0oooo00<0_g@0
000000000000100000001018W=l0oooo0?oog`2LB0030000000307Boo`3oool0oooo00<0oooo00D0
omnL04P000000000001d0;ooo`0O0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0
0`3oool01@3ogi`0B0000000000007@0_ooo00@0oooo00D0omnL04P000000000001d0;ooo`2_0?oo
o`00o`3oool40?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0
M00007Boo`3oool0okmd0000M02oool09P3oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0Q0?oo
o`030?oOW0180000M;oo00<0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`020?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0O0?ooo`030=nLB0000000000000<0000000X0
04RL0=ooo`3oool0oono07@0001d_ol0oooo0?noM00007@0_ooo0P3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool07`3oool00`3OW4P0000000000003000000050018W03Oool0oooO09a8M02o
ool00P3oool00`2oM000M;oo0?ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0N
0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RLg`3oool00P3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0[P3oool00?l0
oooo2@3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3o
ool09@3oool00`3ogi`0B00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00L0_g@0001d_`3o
ool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0803oool00`3ogi`0B00007Boo`08
0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0O
0?ooo`030=nLB018W=l0oooo0080oooo00H0giad09cOo`3oool0_g@00018W03Oool20?ooo`070;md
0000M;l0oooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01l0oooo00<0gia804RL
g`3oool00P3oool00`3OW7@0W=oo0?ooo`040?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@00018
W03Oool00P3oool00`2oM00007Bo0?ooo`0R0?ooo`030?oo_`1d04P0W=oo0080oooo00<0_g@00018
W03Oool00P3oool01`2oM00007Bo0?ooo`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?oo
o`2]0?ooo`00o`3oool:0?ooo`060;md001d_ol0oooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d
0;ooo`0U0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@00018W=l00`3oool01P3OW4P0001d
0;ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0803oool00`3ogi`0B00007Boo`090?oo
o`060;md001d_ol0oooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0P0?ooo`030=nLB01d
_ol0oooo00<0oooo00<0oono07@00018W=l00`3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`<0
oooo00<0gia80000M02oool0803oool00`3OW4P0M;oo0?ooo`090?ooo`060;md001d_ol0oooo0?oo
_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0R0?ooo`070?oo_`1d0000M;oo0?ooo`3ookl0M000
04RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`2]0?oo
o`00o`3oool90?ooo`070?oog`2LB000B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo02D0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0
W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0P0?ooo`030?oOW0180000M;oo00P0oooo
00L0oooO09a80018W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08@3oool00`2o
M4P0W=oo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d
_ol30?ooo`030?noM00004P0W=oo0240oooo00<0_gA809cOo`3oool01`3oool01`3ooml0W4P004RL
g`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0R0?ooo`070?noM0000000M;oo0?oo
o`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@00018
09cOo`2]0?ooo`00o`3oool90?ooo`070=nLB0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo02D0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`06
0?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0P0?ooo`030?oOW0180000
M;oo00P0oooo00L0gia80000001d_ol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
8@3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0
omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0240oooo00<0oooO09a8M02oool01`3oool01`3O
W4P0000007Boo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0Q0?ooo`080?oOW018
0000001809cOo`3oool0omnL04P0001d_ol30?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`2]0?ooo`00o`3oool50?ooo`0;0=oOg`3oool0oono07@000000000001d
0;ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0U0?ooo`030?oOW0180000M;oo
00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0
okmd0000B02Lgol0803oool00`3ogi`0B00007Boo`040?ooo`0;0=oOg`3oool0oono07@000000000
001d0;ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0R0?ooo`070?noM000M;l0
oooo0?ooo`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o
_g@0001809cOo`0R0?ooo`030?noM000M;l0oooo0080oooo00/0gmoO0?ooo`3ookl0M00000000000
07@0_ooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo01l0oooo00D0okmd00000000
0000001809cOo`020?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d
_ol30?ooo`030?noM00004P0W=oo0:d0oooo003o0?ooo`D0oooo00H0gia800000000000000000000
B02Lgol20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol09@3oool00`3ogi`0B000
07Boo`030?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?oo
o`030?noM00004P0W=oo0200oooo00<0omnL04P0001d_ol0103oool01P3OW4P00000000000000000
001809cOo`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0R0?ooo`070?oo_`1d
04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool0
0`3o_g@0001809cOo`0R0?ooo`030?oo_`1d04P0W=oo0080oooo00H0gia800000000000000000000
B02Lgol20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08@3oool00`2oM00007Bo
0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?oo
o`030?noM00004P0W=oo0:d0oooo003o0?ooo`D0oooo00@0oono07@000000000B9cO103oool00`3o
okl0M00004RLg`030?ooo`030=nLB00007@0_ooo02D0oooo00<0omnL04P0001d_ol00`3oool00`3o
okl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;oo
o`0P0?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@000000000B9cO103oool00`3ookl0M000
04RLg`030?ooo`030=nLB00007@0_ooo0280oooo00L0oono07@00018W=l0oooo0?oo_`1d0000B9cO
00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo0280oooo00<0
oono07@00018W=l00P3oool0103ookl0M00000000018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool08P3oool01`2oM000B9cO0?ooo`3oool0oono07@00018W=l00`3oool01P3O
W4P0001d0;ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0[@3oool00?l0oooo1P3oool0
0`2oM000M;oo0?ooo`050?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool09@3oool0
0`3ogi`0B00007Boo`040?ooo`030=nLB000B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool0803oool00`3ogi`0B00007Boo`050?ooo`030;md
001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0M0?ooo`030?no
W02Lgol0oooo0080oooo00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`070;md
0000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo01d0oooo00<0oknL09cO
o`3oool00P3oool00`3ogi`0B00004RLg`030?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018
W03Oool00P3oool00`2oM00007Bo0?ooo`0M0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia80000
M02oool00P3oool00`3OW4P004RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`3OW4P004RL0=oo
o`020?ooo`030;md0000M;l0oooo0:d0oooo003o0?ooo`H0oooo00H0omnL04P00000000000000000
B02Lgol20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0T0?ooo`050?oog`2LB0000000
0000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool07`3oool01@3ooml0W4P0000000000000M;oo00D0oooo00H0omnL
04P00000000000000000B02Lgol20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0N0?oo
o`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0N0?ooo`080?oog`2L
B4P0W=oo0?ooo`3ookl0M0000000M02oool30?ooo`060?oOW018000000000000000004P0W=oo0P3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool07`3oool01P3o_g@0B9cO0?ooo`3ookl0M000
001d_`<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo0:h0oooo003o0?ooo`L0oooo00D0_g@0000000000000000004RLg`030?oo
o`050?oo_`1d000000000000M02oool09`3oool00`3o_g@0000007Boo`050?ooo`050?oo_`1d0000
00000000M02oool0103oool01@3ookl0M0000000000007@0_ooo0280oooo00<0okmd0000001d_ol0
1P3oool01@2oM0000000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`0P0?oo
o`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0M0000000000007@0_ooo00@0oooo
00D0oono07@000000000001d0;ooo`0P0?ooo`060?oog`2LB00000000000000007@0_ooo1@3oool0
1@2oM0000000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`0Q0?ooo`050=nL
B000000000000000M02oool0103oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0oono07@0
00000000001d0;ooo`2_0?ooo`00o`3ooooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0oooo
o`3ooooQ0?ooo`00o`3ooooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ0?oo
o`00o`3ooooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ0?ooo`00e03ooooo
000005800000^P3oool00=@0oooo0`0000080?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00@000000?oo
o`3oool000001`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`000000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool0^03oool00=@0oooo1000000d0?ooo`040000003oool0oooo000000`0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0;`3oool2000003H0oooo00<000000?ooo`3oool0<`3oool3000003H0oooo0P00002j
0?ooo`00e03oool01@000000oooo0?ooo`3oool0000003<0oooo00@0001o0?ooo`3oool000002`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01@0oooo00<0001o0?ooo`3oool00P3o
ool00`000000oooo0?ooo`0`0?ooo`030000O`000000oooo03D0oooo00<0001o0?ooo`3oool0<P3o
ool010000000oooo0?ooo`0007lg0?ooo`030000003oool0oooo0;P0oooo003D0?ooo`8000000`3o
ool200000340oooo00@0001o0?ooo`3oool000002`3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo01@0oooo00<0001o0?ooo`3oool00P3oool00`000000oooo0?ooo`0`0?ooo`030000
003oool0oooo03D0oooo00<0001o0?ooo`3oool0<@3oool01@000000oooo0?ooo`3oool0001o03H0
oooo0P00002j0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`800000<03oool0100007l0oooo
0?ooo`00000;0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0503oool01@0007l0
oooo0?ooo`3oool000000380oooo00<000000000O`3oool0=P3oool00`0007l0oooo0?ooo`0_0?oo
o`800000103oool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo002m0?ooo`06
0;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo00T0oooo0P00
00050?ooo`800000;`3oool00`0007l0oooo0000000<0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool05@3oool01@0007l0oooo0?ooo`3oool000000340oooo00<000000?ooo`0007l0
=`3oool00`0007l0oooo0?ooo`0^0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool0
=03oool200000;X0oooo002n0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool01@3oool3000002d0oooo00<0001o0?ooo`00
00002`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01D0oooo00@0001o0?ooo`3o
ool00000<P3oool00`000000oooo0000O`0g0?ooo`030000O`3oool0oooo02d0oooo00<000000?oo
o`3oool01@3oool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo002n0?ooo`03
0?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`03
0000003oool0oooo00L0oooo0P00000/0?ooo`030000O`3oool0000000/0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0F0?ooo`040000O`3oool0oooo00000340oooo00@000000?oo
o`3oool0001o=`3oool00`0007l0oooo0?ooo`0[0?ooo`800000203oool00`0007l0oooo0?ooo`0e
0?ooo`030000003oool0oooo0;P0oooo002n0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@0
0018W=l00`3oool00`3OW4P0001d0;ooo`070?ooo`8000002@3oool2000002/0oooo00<0001o0000
003oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01H0oooo00<0001o0?oo
o`000000<@3oool01@000000oooo0?ooo`3oool0001o03L0oooo00<0001o0?ooo`3oool0:P3oool0
0`000000oooo0?ooo`080?ooo`030000O`3oool0oooo03@0oooo0P00002j0?ooo`00_P3oool00`3o
gi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`00
0000oooo0?ooo`090?ooo`<00000:@3oool00`0007l000000?ooo`0:0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0000000`0?ooo`030000003oool0oooo
0080oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0Y0?ooo`030000003oool0oooo
00T0oooo00<0001o0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00_P3oool00`3ogi`0
B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool3000000X0
001o0`00000Y0000O`030000000007l0001o00T0001o00<000000000O`0007l03P0007l00`000000
001o0000O`0H0000O`030000000007l0001o02h0001o00<000000000O`0007l0I@0007l200000480
001o0`00002j0?ooo`00_P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0<0?ooo`<000009P3oool00`0007l00000
0?ooo`090?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0603oool00`0007l00000
0?ooo`0^0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo
0?ooo`0V0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool0=@3oool00`000000oooo
0?ooo`2h0?ooo`00_P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01`3oool2000000h0oooo0`00000U0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo02d0oooo00<0
00000?ooo`3oool01@3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo02D0oooo00<0
00000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0d0?ooo`800000^P3oool00;h0oooo00<0omnL
04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<00000
0?ooo`3oool03P3oool3000002@0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`06
0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool08`3oool200000100oooo00<0001o
0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00Z03oool00`2oM4P0W=oo0?ooo`0C0?oo
o`030?oOW0180000M;oo00@0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`8000004@3oool300000280oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo01P0oooo00<000000000O`3oool0;@3oool00`000000oooo0?ooo`060?oo
o`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0@0?oo
o`030000O`3oool0oooo03@0oooo0P00002j0?ooo`00Z03oool00`2oM000000007Boo`060?ooo`03
0?oog`2LB00007Bo00P0oooo00D0oooO09a800000000000007Boo`040?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo0140oooo0`00000P0?ooo`0300000000
07l0oooo00T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0001o02d0oooo00<000000?ooo`3oool01`3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3o
ool0oooo0240oooo00<000000?ooo`3oool04@3oool00`0007l0oooo0?ooo`0e0?ooo`030000003o
ool0oooo0;P0oooo002X0?ooo`050;md0000000000000000001d_ol01@3oool00`2oM00004RL0=oo
o`060?ooo`030;md00000000000000@0000000T0B9cO0?ooo`3oool0oooo0?oOM00P000000000000
=@2?ool02@3oool2000001@0oooo0`00000M0?ooo`030000003oool0001o00T0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0001o02`0oooo00<000000?oo
o`3oool0203oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo01l0oooo0P00000D0?oo
o`030000O`3oool0oooo03@0oooo0P00002j0?ooo`00Z03oool01`2oW9`0gooo0?oog`2LB0000000
0000B02Lgol00P3oool00`3o_g@0001809cOo`080?ooo`030=nLB0000000000000<0000000X004RL
0=ooo`3oool0oono07@0001d_ol0oooo0?noM00007@0_ooo203oool00`000000oooo0?ooo`0E0?oo
o`<000006`3oool00`000000oooo0000O`090?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0603oool010000000oooo0?ooo`0007l[0?ooo`030000003oool0oooo00T0oooo00<0001o
0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo01@0oooo00<0001o
0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00[03oool00`3ookl0M000000000020000
0003001d_`3oool0oooo00T0oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM000
04RL0=ooo`80oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0F0?ooo`@00000603o
ool010000000oooo0?ooo`0007l80?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
5`3oool01@000000oooo0?ooo`3oool0001o02X0oooo00<000000?ooo`3oool02P3oool00`0007l0
oooo0?ooo`0e0?ooo`030000O`3oool0oooo01`0oooo0P00000G0?ooo`030000O`3oool0oooo03D0
oooo00<000000?ooo`3oool0^03oool00:`0oooo00@0oooO09a800000000M;oo3P3oool00`3OW4P0
M;oo0?ooo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool2000001T0
oooo0`00000G0?ooo`040000003oool0oooo0000O`P0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo0080oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0;0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool06`3oool00`00
0000oooo0?ooo`0G0?ooo`030000O`3oool0oooo03@0oooo0P00002j0?ooo`00Z03oool01P2oW9`0
gooo0?oOW0180000001809cOoa40oooo00<0_gA809cOo`3oool00P3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool06@3oool4000001@0oooo00D00000
0?ooo`3oool0oooo0000O`070?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`020?ooo`030000O`3oool0oooo02H0oooo00<000000?ooo`3oool0303o
ool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool0603o
ool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo002X0?ooo`040;md00000000
04RL0=oooa<0oooo00<0oooO09a8M02oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00L0oooo0`00000K0000O`@00000400007l2000000/0001o00<000000000O`0007l03@00
07l00`000000001o0000O`0H0000O`030000000007l0001o02/0001o00<000000000O`0007l0H000
07l200000540001o0`00002j0?ooo`00Z03oool00`2oM7@0_ooo0?ooo`0E0?ooo`070?noM000M;l0
oooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0
703oool6000000`0oooo00<000000?ooo`3oool0103oool00`0007l0oooo0?ooo`030?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`040?ooo`030000
O`3oool0oooo02<0oooo0P00000@0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool0
5`3oool00`000000oooo0?ooo`0K0?ooo`030000O`3oool0oooo03D0oooo00<000000?ooo`3oool0
^03oool00<00oooo00L0oono07@0B02Lgol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01`3oool2000001h0oooo4@0000070?ooo`040000O`3oool0oooo0?ooo`8000003`3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo00D0oooo00<0001o0?ooo`3oool08P3oool00`00
0000oooo0?ooo`0@0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool05@3oool20000
01h0oooo00<0001o0?ooo`3oool0=03oool200000;X0oooo00300?ooo`070?oo_`1d0000B9cO0?oo
o`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool07`3oool6
00000100oooo00@0001o0?ooo`3oool00000403oool00`000000oooo0?ooo`0G0?ooo`800000203o
ool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo0140oooo00<0001o0?ooo`3oool0=@3o
ool00`0007l0oooo0?ooo`0C0?ooo`800000803oool00`0007l0oooo0?ooo`0e0?ooo`030000003o
ool0oooo0;P0oooo002k0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l00P3oool0
0`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo0P00000Q0?ooo`L000003P3oool0
0`0007l000000000000A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0203oool0
0`0007l0oooo0?ooo`0P0?ooo`030000003oool0oooo0180oooo00<0001o0?ooo`3oool0=@3oool0
0`0007l0oooo0?ooo`0A0?ooo`8000008P3oool00`0007l0oooo0?ooo`0d0?ooo`800000^P3oool0
0;/0oooo00P0oooO09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`80oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool08P3oool<000000<0oooo1@00000A0?oo
o`800000603oool00`000000oooo0?ooo`090?ooo`030000O`3oool0oooo01l0oooo00<000000?oo
o`3oool04`3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo0100oooo00<000000?oo
o`3oool08P3oool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo002k0?ooo`03
0?oo_`1d000000000080000000<0001d0;ooo`3oool00`3oool01@3ogg@080000000000003D0Sooo
00T0oooo00<000000?ooo`3oool0903oool6000000<0oooo1@0000030?ooo`030000O`3oool0oooo
00d0oooo0P00000I0?ooo`030000003oool0oooo00X0oooo00<0001o0?ooo`3oool07@3oool20000
01H0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0>0?ooo`8000009@3oool00`00
07l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo002V0?ooo`030?oog`2LB000000000H0
000000<007Bo0?ooo`3oool02@3oool00`3ogi`0B00004RLg`020?ooo`030;md0018W=l0oooo0080
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo0P00000W0?ooo`P000001`3oool0
0`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool02`3oool0
0`0007l0oooo0?ooo`0L0?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3oool0=@3oool0
0`0007l0oooo0?ooo`0<0?ooo`8000009`3oool00`0007l0oooo0?ooo`0d0?ooo`800000^P3oool0
0:@0oooo00<0oooO09a8000000002P0000000`00M;l0oooo0?ooo`0<0?ooo`030?oo_`1d04P0W=oo
0080oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo
02P0oooo100000020?ooo`8000001@3oool00`0007l0oooo0?ooo`090?ooo`<000006P3oool00`00
0000oooo0?ooo`0<0?ooo`030000O`3oool0oooo01/0oooo00<000000?ooo`3oool05`3oool00`00
07l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo00/0oooo00<000000?ooo`3oool09`3oool00`00
07l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo002S0?ooo`040?noM000000004RL0=oo
o`T0oooo00@0oono07@0000004P0W=oo303oool01`3ookl0M00007Boo`3oool0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`070?ooo`800000:`3oool400000080oooo1`0000040?ooo`H00000
6`3oool2000000l0oooo00<0001o0?ooo`3oool06@3oool2000001X0oooo00<0001o0?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`090?ooo`800000:P3oool00`0007l0oooo0?ooo`0d0?ooo`800000
^P3oool00:80oooo00<0oono07Ad_`3oool03@3oool00`3ooml0W7BL0=ooo`0;0?ooo`070?noM000
0000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3o
ool0:`3oool5000000D0oooo00<0001o0000000000001000000O0?ooo`030000003oool0oooo00l0
oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo03D0
oooo00<0001o0?ooo`3oool01`3oool2000002`0oooo00<0001o0?ooo`3oool0=@3oool00`000000
oooo0?ooo`2h0?ooo`00_`3oool0203ogi`0B0000000B02Lgol0oooo0?oOW0180000M;oo0`3oool0
0`3o_g@0001809cOo`070?ooo`<00000;@0007l7000002D0001o0P00000[0000O`800000G@0007l2
000006@0001o0`00002j0?ooo`00Z03oool00`3ookl0M7Bo0?ooo`040?ooo`030=nLW02oool0oooo
00/0oooo00D0okmd000000000000001809cOo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol01`3oool00`000000oooo0?ooo`0_0?ooo`@0000000<0oooo0000000000000P00000N
0?ooo`<00000503oool00`0007l0oooo0?ooo`0E0?ooo`030000003oool0oooo01d0oooo00<0001o
0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0oooo02h0oooo00<0001o
0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00Z03oool00`3o_g@00000000000040000
00030000M02oool0oooo00d0oooo00<0_g@0001d_`3oool00P3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00L0oooo0P00000b0?ooo`@0000000<0001o0?ooo`3oool00`00000H0?oo
o`<000005`3oool00`0007l0oooo0?ooo`0C0?ooo`800000803oool00`0007l0oooo0?ooo`0e0?oo
o`030000O`3oool0oooo0080oooo0P00000a0?ooo`030000O`3oool0oooo03@0oooo0P00002j0?oo
o`00Z03oool00`2oM000000000000004000000030000M02oool0oooo00h0oooo00L0_g@004RLg`3o
ool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`0c
0?ooo`D000000`3oool5000000l0oooo1000000J0?ooo`030000O`3oool0oooo0140oooo0P00000R
0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool00P00000c0?ooo`030000O`3oool0
oooo03D0oooo00<000000?ooo`3oool0^03oool00:T0oooo00<0_gAd0;ooo`3oool00`3oool00`3O
W9`0_ooo0?ooo`090?ooo`030?oog`2LW;l0oooo0080oooo00<0gia80000M02oool00P3oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0=@3oool20000
00030?ooo`000000000000H0oooo3`00000N0?ooo`030000O`3oool0oooo00l0oooo0P00000T0?oo
o`030000O`3oool0oooo03D0oooo00<0001o000000000000=@3oool00`0007l0oooo0?ooo`0e0?oo
o`030000003oool0oooo0;P0oooo002X0?ooo`030?oOW02L_ol0oooo0140oooo00H0okmd04RLg`3o
ool0oono07@00000M;l30?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`800000
=P3oool01@0007l0oooo000000000000oooo00800000<@3oool00`0007l0oooo0?ooo`0=0?ooo`80
00009P3oool00`0007l0oooo0?ooo`0d0?ooo`800000=`3oool00`0007l0oooo0?ooo`0d0?ooo`80
0000^P3oool00:P0oooo00<0_g@0001d_`3oool0103oool00`3OW9`0_ooo0?ooo`0;0?ooo`060=nL
B0000000000000000000B9`0gooo0`3oool01@3ogg@080000000000003D0Sooo00T0oooo00<00000
0?ooo`3oool0=@3oool0100007l0oooo0?ooo`3oool2000000040?ooo`0000000000000002h0oooo
00<0001o0?ooo`3oool02`3oool2000002P0oooo00<0001o0?ooo`3oool0<P3oool2000000030?oo
o`0007l0oooo03H0oooo00<0001o0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00Z03o
ool00`3OW4P0000000000004000000030000M02oool0oooo00d0oooo00@0oooO09a80000B9`0gooo
0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool2000003H0oooo00<0001o0?oo
o`3oool00`3oool200000080oooo1000000Z0?ooo`030000O`3oool0oooo00T0oooo0P00000Z0?oo
o`030000O`3oool0oooo0300oooo0P0000030?ooo`030000O`3oool0oooo03D0oooo00<0001o0?oo
o`3oool0=03oool200000;X0oooo002X0?ooo`030?oog`2LB000000000@0000000<0001d0;ooo`3o
ool03@3oool0103ooml0W4P00018W03Oool20?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool01`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo00D0oooo0P0000040?oo
o`<000009`3oool00`0007l0oooo0?ooo`070?ooo`800000;03oool00`0007l0oooo0?ooo`0^0?oo
o`8000001@3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo03D0oooo00<000000?oo
o`3oool0^03oool00:T0oooo00<0gibL0;ooo`3oool00`3oool00`3OW9`0_ooo0?ooo`090?ooo`03
0;md00000000000000D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`070?oo
o`800000=P3oool00`0007l0oooo0?ooo`070?ooo`8000001@3oool4000002<0oooo00<0001o0?oo
o`3oool00`3oool4000002h0oooo00<0001o0?ooo`3oool0;03oool2000000L0oooo00<0001o0?oo
o`3oool0=@3oool00`0007l0oooo0?ooo`0d0?ooo`800000^P3oool00:P0oooo00<0okmd04RLg`3o
ool0403oool0303ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`<0
oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo00T0
oooo0P0000070?ooo`<00000803oool00`0007l0oooo0000000300000380oooo00<0001o0?ooo`3o
ool0:P3oool2000000T0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`03
0000003oool0oooo0;P0oooo002X0?ooo`030;md0000M;l0oooo00@0oooo00<0gibL0;ooo`3oool0
2P3oool02`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol01`3oool3000004<0001o0`0000070000O`H000005P0007l600000600001o0`00
001i0000O`<00000^P3oool00:P0oooo00<0gia8000000000000100000000`0007@0_ooo0?ooo`0:
0?ooo`0;0?oog`2LB7@0_ooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`070?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool03P3oool20000
00/0oooo5P0000040?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool09@3oool20000
00h0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0oooo
0;P0oooo002X0?ooo`030?oog`2LB000000000@0000000<0001d0;ooo`3oool02`3oool02P3ogi`0
B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00L0oooo
00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0@0?ooo`<000008P3oool00`0007l0oooo
0?ooo`0e0?ooo`030000O`3oool0oooo0280oooo0`00000@0?ooo`030000O`3oool0oooo03D0oooo
00<0001o0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00[`3oool00`3OW9`0_ooo0?oo
o`0<0?ooo`090=nLB01d_ml0W4P00018W03Oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool01`3oool2000003H0oooo00<0001o0?ooo`3oool04`3oool200000200oooo00<0001o0?oo
o`3oool0=@3oool00`0007l0oooo0?ooo`0O0?ooo`<000004`3oool00`0007l0oooo0?ooo`0e0?oo
o`030000O`3oool0oooo03@0oooo0P00002j0?ooo`00_P3oool01@3ooml0W4Q804Q80000B9`0gooo
0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool05@3oool3000001d0oooo00<0001o0?ooo`3oool0=@3oool00`00
07l0oooo0?ooo`0M0?ooo`8000005P3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo
03D0oooo00<000000?ooo`3oool0^03oool00:80oooo00<0oono07A8W03Oool03@3oool00`3ogi`0
B4RL0=ooo`0:0?ooo`040?oOW018000004RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo00P0oooo0P00000f0?ooo`030000O`3oool0oooo01P0oooo0`00000J0?ooo`030000O`3o
ool0oooo03D0oooo00<0001o0?ooo`3oool06P3oool3000001P0oooo00<0001o0?ooo`3oool0=@3o
ool00`0007l0oooo0?ooo`0d0?ooo`800000^P3oool00:<0oooo00@0omnL04P00000B9`0gooo2@3o
ool0103ookl0M0000018W03Oool<0?ooo`030=nLB000B9`0gooo00<0oooo00D0oono07@000000000
001d0;ooo`090?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool06`3oool3000001L0
oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0F0?ooo`@000006`3oool00`0007l0
oooo0?ooo`0e0?ooo`030000O`3oool0oooo03D0oooo00<000000?ooo`3oool0^03oool00:@0oooo
00<0oooO09a8000000002P0000000`00M;l0oooo0?ooo`070?ooo`030?oog`2LB000000000800000
00<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo00T0oooo0P00000f0?ooo`03
0000O`3oool0oooo01h0oooo1000000C0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3o
ool04P3oool4000001l0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0d0?ooo`80
0000^P3oool00:H0oooo00<0oooO09a8000000001P0000000`00M;l0oooo0?ooo`090?ooo`030?oo
g`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0203oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo0280oooo1000000?0?oo
o`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool03P3oool4000002<0oooo00<0001o0?oo
o`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0oooo0;P0oooo00300?ooo`03
0?oog`2LB7@0_ooo0080oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`80
0000=P3oool00`0007l0oooo0?ooo`0V0?ooo`@000002`3oool00`0007l0oooo0?ooo`0e0?ooo`03
0000O`3oool0oooo00X0oooo1000000W0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3o
ool0=03oool200000;X0oooo00310?ooo`060;md001d_ol0oooo0?oo_`1d0000B9cO0`3oool00`3O
W4P0001d0;ooo`070?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0:P3oool40000
00L0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`050?ooo`D00000:`3oool00`00
07l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo03D0oooo00<000000?ooo`3oool0^03oool00<00
oooo00L0oooO09a80018W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool0
0`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02h0oooo1@0000020?ooo`030000O`3oool0
oooo03D0oooo00@0001o0?ooo`3oool0oooo1000000`0?ooo`030000O`3oool0oooo03D0oooo00<0
001o0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00`03oool01`3OW4P0000007Boo`3o
ool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`<00000J`0007l4000003D0001o
1@00002R0000O`<00000^P3oool00;`0oooo00/0gmoO0?ooo`3ookl0M0000000000007@0_ooo0?oo
o`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0=@3oool0
0`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0000000P00000:03oool4000000030?ooo`0007l0
oooo03H0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000003oool0
oooo0;P0oooo002l0?ooo`060=nLB000000000000000000004P0W=oo0P3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00L0oooo0P00000f0?ooo`030000O`3oool0oooo03D0oooo00<0
001o0?ooo`3oool0203oool<000000`0oooo400000050?ooo`030000O`3oool0oooo03D0oooo00<0
001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0d0?ooo`800000^P3oool00;`0oooo00@0oono
07@000000000B9cO103oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<0
00000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo01@0oooo3000
000E0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?oo
o`0e0?ooo`030000003oool0oooo0;P0oooo002m0?ooo`030;md001d_ol0oooo00D0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`800000=P3oool00`0007l0oooo0?ooo`0e
0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e
0?ooo`030000O`3oool0oooo03@0oooo0P00002j0?ooo`00_@3oool01P3ogi`0B000000000000000
001809cOo`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3o
ool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3o
ool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0oooo03D0oooo00<000000?ooo`3o
ool0^03oool00;d0oooo00<0omnL03D0000000000P0000002@18W=l0oooo0?ooo`3oool0ommd0200
00000000000e08ooo`090?ooo`800000=P3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0
oooo03D0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`030000O`3oool0
oooo03@0oooo0P00002j0?ooo`00_03oool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol00P3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`0e0?ooo`03
0000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?ooo`03
0000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool0=@3oool00`000000oooo0?ooo`2h0?ooo`00
^`3oool00`3ooml0W4P0001d_`030?ooo`070?noM000B9`0gooo0?ooo`2oM00004RL0=ooo`020?oo
o`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?oo
o`030000O`3oool0oooo03D0oooo00<0001o0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0e0?oo
o`030000O`3oool0oooo03D0oooo00<000000?ooo`3oool0^03oool00;/0oooo00<0omnL04P0001d
_ol00`3oool01P3o_g@0001809cOo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool01`3oool2
000003H0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0
oooo03D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0d0?ooo`800000^P3oool0
0;/0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0
okmd0000B02Lgol01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`2h0?ooo`00^`3oool00`3ogi`0B00007Boo`03
0?ooo`060;md000004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`070?ooool00000
DP00002j0?ooo`00^`3oool0303ookl0M00000000018W=l0oooo0?noM000000007Bo0?ooo`3ogi`0
B00007Boo`<0oooo00<0okmd0000B02Lgol0o`3ooooo0?oooa@0oooo002l0?ooo`030;md0000B9`0
_ia80080000000H004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0?l0oooo
o`3ooolD0?ooo`00_03oool00`3ogi`0B00007Boo`050?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool0o`3ooooo0?oooa@0oooo002m0?ooo`030=nLB00004P0W=oo00D0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`3o0?ooool0oooo503oool00;h0oooo00<0okmd0000
0000M;l0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0o`3ooooo0?oooaD0oooo002o
0?ooo`050?oog`2LB00000000000M02oool00P3oool01@3ookl0M0000000000007@0_ooo0?l0oooo
o`3ooolF0?ooo`00o`3ooonG0?ooo`060?oog`2LB00000000000000007@0_ooo4@3oool00`3ooml0
W4P0000000030000000304RLg`3oool0oooo0?l0oooo:P3oool00?l0ooooV@3oool00`2oM00004RL
0=ooo`0B0?ooo`030;md0000B9`0gooo00<0oooo00<0gibL0;ooo`3oool0o`3ooolZ0?ooo`00o`3o
oonI0?ooo`030;md0000B9`0gooo0180oooo00<0oooO09adW03Oool00`3oool00`3ookl0W9cO0?oo
o`3o0?ooobX0oooo003o0?oooiT0oooo00<0_g@00018W03Oool04`3oool00`2oM000000000000003
0000000304RLg`3oool0oooo0?l0oooo:@3oool00?l0oooo@@3oool00`3ooml0W4P0000000070000
000307Boo`3oool0gia80080000000L0001d07Ad0000000000000000M02ooml0W4P00080000000@0
04RL0=ooo`3oool0_g@00P0000001@0007@0_ooo0?ooo`3ooml0W4P00080000000D004RL09ad0000
000007Bo0;md0002000000030000B02Lgol0oooo00@0oooo00<0oooO09a8000000001P0000000`00
M;l0oooo0?ooo`020?ooo`090?oog`2LB000000000000000B9`0W7@000000000M;l0gia800800000
00<0001809cOg`2LB0000P0000000`0007@0_ooO09a8000200000005001d_`3oool0_g@00018B018
00000P0000001018W=l0oooo0?ooo`3OW4P2000000070000M01dM00000000000000007@0_ooO09a8
0002000000030018W03Ookl0M00000@0000000<0001d0;ooo`3oool0o`3ooolZ0?ooo`00o`3ooom3
0?ooo`030?oOW0180000B9cO00@0oooo00l0_g@0001d_`3oool0oooo0=nLB00007@0_ooo0?ooo`3O
W4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool04`2oM0000000001d_`3oool0omnL04P0M02o
ool0okmd0000001d_ol0oooo0?oOW018000007Bo0?no_`2oool0gia80000M02oool0203oool00`3o
gi`0B00004RLg`030?ooo`040?noM0000000000004RLg`80oooo00`0okmd0000001d_ol0oooo0?oO
W018000007Bo0?no_`2oool0gia80000M02oool20?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0
001d_`020?ooo`070;md00000000M;oo0?ooo`3ookl0M00004RLg`020?ooo`0;0=nLB00007@0_ooo
0?ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool00`2oM4P0W=oo0?ooo`3o0?ooobl0
oooo003o0?oood<0oooo00<0oono07@00018W=l0103oool04`3ookl0M4RL0=ooo`3oool0gia80000
M02oool0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool0
2@3ogi`0M9cO0?noM00004P0W=oo0?ooo`3ooml0W4P0001d_`020?ooo`030=nLB00007@0_ooo00P0
oooo00<0oono07@00018W=l0103oool0303ookl0M0000000001d_ol0oooo0?noM00004P0W=oo0?oo
o`3ooml0W4P0001d_`80oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO
09a80000M;l00P3oool00`2oM00004RL0=ooo`020?ooo`0?0?noM00007@0_ooo0?ooo`3OW4P0001d
0;ooo`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo0080oooo00D0oono07@0000000000000
001d_`3o0?ooobd0oooo003o0?oood<0oooo00<0oono07@00018W=l01@3oool04P3OW7@0W=oo0?oo
o`3OW4P0001d0;ooo`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0001809cO
o`H0oooo00H0okmd00000000M;l0oooO09a80000M;l20?ooo`030=nLB00007@0_ooo00P0oooo00<0
oono07@00018W=l01@3oool02`3ogi`0B0000018W03Oool0oooo0?noM000000007Bo0?oog`2LB000
07Bo0080oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l0
0P3oool00`2oM00004RL0=ooo`020?ooo`0?0?oOW0180000M;oo0?ooo`3OW4P0001d0;ooo`3oool0
gia80018W03Oool0oooo0?oog`2LB00007Bo0080oooo00H0gia80000M02oool0oono07@00018W=oo
0?ooob`0oooo003o0?oood<0oooo00<0oono07@00018W=l0203oool03`3OW4P0001d0;ooo`3oool0
gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0001d0;ooo`080?ooo`040?noM0000000
0000001d_`80oooo00<0gia80000M02oool0203oool00`3ookl0M00004RLg`050?ooo`040?oog`2L
B000001809cOo`<0oooo00@0okmd00000000000007Bo0P3oool00`3OW4P0001d0;ooo`020?ooo`07
0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`030;md0000B9`0gooo0080oooo01L0omnL
04P0001d_ol0oooo0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo
0?oo_`1d0000B9cO0?ooo`3ooml0W4P0001d_`3o0?ooob`0oooo003o0?oood<0oooo00<0oono07@0
0018W=l0203oool03P3OW4P0001d0;ooo`3oool0_g@00000M02oool0oooo0?oog`2LB00007Bo0?oo
o`3ogi`0B0001@0000002018W=l0oono07@0001d_ol0oooo0?oog`2LB00007Bo0P3oool00`3OW4P0
001809cOo`080?ooo`030?oo_`1d0000B9cO00H0oooo00X0_g@00000001d_ol0oono07@0001d_ol0
oooo0?oog`2LB00007Bo0P3oool00`3OW4P0001809cOo`020?ooo`070;md000007@0_ooo0?ooo`3o
oml0W4P0001d_`020?ooo`030;md000007@0_ooo0080oooo01L0gia80000B02Lgol0oooo0=nLB000
07@0_ooo0?ooo`2oM000001d0;ooo`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0?ooo`3o
oml0W4P004RLg`3o0?ooob`0oooo003o0?oood<0oooo00<0oono07@00018W=l01P3oool03@3ooml0
W4P00000000007@0_ooO09a80000000004Q804P0B02Lgol0oono07@00018W=l00P3oool05P2oM000
B9cO0?ooo`3oool0gia80000B02Lgol0oono07@00018W=l0oooo0?oo_`1d000007Bo0?oog`2LB000
00000000B018M9`0gkmd0000M02oool50?ooo`030?oo_`1d0000B9cO00H0oooo02h0_g@00000001d
_ol0oono07@00018W=l0oooo0?oo_`1d000007Bo0?oog`2LB00000000000B018B4P0W=oo0?oo_`1d
000007Ad07@0M02oool0oono07@00018W=l0oono07@00000000004Q804Q8W03Oool0omnL04P00000
B9`0gooO09a800000000001d0;oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0P3oool0
102oM00007Bo0?ooo`3o_g@2000000030018W03Oool0oooo0?l0oooo:@3oool00?l0oooo@`3oool0
0`3ookl0M00004RLg`080?ooo`0;0;md000007@0_ooo0?ooo`2oM00004RL0=oog`2LB00000000000
M02oool00`3oool01@3OW4P000000000000004P0W=oo0080oooo00H0oono07@00000000000000000
M02oool20?ooo`060;md000007@0_mnL04P0000007@0_ooo1@3oool00`3ookl0M00004RLg`050?oo
o`070?oog`2LB000001809cOo`3oool0oono07@000020000000I0000M02oool0oooo0?ooo`2oM000
001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`3oool0oooO09a80000B9`0
gmnL04P000000000001d0;ooo`020?ooo`0;0;md000007@0_ooo0?ooo`2oM00004RL0=oog`2LB000
00000000M02oool00P3oool01@3ooml0W4P0000000000000M;oo0?l0oooo;@3oool00?l0oooo@`3o
ool00`3ookl0M00004RLg`0`0?ooo`030?oo_`1d0000B9cO00D0oooo00@0omnL04P00000B9`0gooo
o`3ooomO0?ooo`00o`3ooom30?ooo`030?oo_`1d0000B9cO0300oooo00<0oono07@00018W=l0103o
ool0103ooml0W4P00000001d_ooo0?ooof00oooo003o0?oood<0oooo00<0omnL04P00018W=l0<03o
ool00`3ogi`0B00004RLg`030?ooo`040?noM0000000000004RLgol0ooooH@3oool00?l0oooo@@3o
ool00`3ooml0W4P0000000030000000304RLg`3oool0oooo00@0oooo00<0gia80000B02Lgol08`3o
ool00`3ooml0W4P000000006000000030000M02oool0oooo0280oooo00<0gia80000B02Lgol0o`3o
ooll0?ooo`00o`3ooom>0?ooo`030=nLB00004P0W=oo0540oooo00<0gia80000B02Lgol0o`3oooll
0?ooo`00o`3ooooo0?ooon40oooo003o0?ooool0ooooh@3oool00?l0ooooo`3ooooQ0?ooo`00o`3o
oooo0?ooon40oooo0000\
\>"],
  ImageRangeCache->{{{0, 734.938}, {146.188, 0}} -> {-189.78, -45.7744, \
1.33941, 1.33941}}]
}, Closed]],

Cell[TextData[{
  "The value of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha]\)]],
  " needed to reduce the range to typical home run length is much larger than \
expected for a baseball.  Apparently quadratic damping dominates."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Dependence on quadratic coefficient, with linear term neglected.\
\>", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotDamping[0, 0, 50, \ 
        40  Degree, \ \[Beta]Parameters, \ {\[Beta], 0, 0.002, 
          0.0005}, "\<Quadratic Damping\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .2 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-3.20924e-017 0.00333333 -8.67362e-019 0.00333333 [
[.16667 -0.0125 -8 -15.6875 ]
[.16667 -0.0125 8 0 ]
[.33333 -0.0125 -11 -15.6875 ]
[.33333 -0.0125 11 0 ]
[.5 -0.0125 -11 -15.6875 ]
[.5 -0.0125 11 0 ]
[.66667 -0.0125 -11 -15.6875 ]
[.66667 -0.0125 11 0 ]
[.83333 -0.0125 -11 -15.6875 ]
[.83333 -0.0125 11 0 ]
[1 -0.0125 -11 -15.6875 ]
[1 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -15.0625 -31.375 ]
[.5 -0.0125 15.0625 -15.6875 ]
[-0.0125 .03333 -16 -7.84375 ]
[-0.0125 .03333 0 7.84375 ]
[-0.0125 .06667 -16 -7.84375 ]
[-0.0125 .06667 0 7.84375 ]
[-0.0125 .1 -16 -7.84375 ]
[-0.0125 .1 0 7.84375 ]
[-0.0125 .13333 -16 -7.84375 ]
[-0.0125 .13333 0 7.84375 ]
[-0.0125 .16667 -16 -7.84375 ]
[-0.0125 .16667 0 7.84375 ]
[-0.0125 .2 -16 -7.84375 ]
[-0.0125 .2 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .1 -31.6875 -14.8125 ]
[-0.01875 .1 -16 14.8125 ]
[ 0 .2 .125 0 ]
[.5 .2125 -49.1875 0 ]
[.5 .2125 49.1875 15.6875 ]
[ 1 0 .125 0 ]
[1.0125 .1 0 -3.5 ]
[1.0125 .1 15.6875 3.5 ]
[ 0 0 0 0 ]
[ 1 .2 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.16667 0 m
.16667 .2 L
s
.33333 0 m
.33333 .2 L
s
.5 0 m
.5 .2 L
s
.66667 0 m
.66667 .2 L
s
.83333 0 m
.83333 .2 L
s
0 .03333 m
1 .03333 L
s
0 .06667 m
1 .06667 L
s
0 .1 m
1 .1 L
s
0 .13333 m
1 .13333 L
s
0 .16667 m
1 .16667 L
s
0 g
.16667 0 m
.16667 .00625 L
s
gsave
.16667 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.33333 0 m
.33333 .00625 L
s
gsave
.33333 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.66667 0 m
.66667 .00625 L
s
gsave
.66667 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(200) show
1.000 setlinewidth
grestore
.83333 0 m
.83333 .00625 L
s
gsave
.83333 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(250) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(300) show
1.000 setlinewidth
grestore
.125 Mabswid
.03333 0 m
.03333 .00375 L
s
.06667 0 m
.06667 .00375 L
s
.1 0 m
.1 .00375 L
s
.13333 0 m
.13333 .00375 L
s
.2 0 m
.2 .00375 L
s
.23333 0 m
.23333 .00375 L
s
.26667 0 m
.26667 .00375 L
s
.3 0 m
.3 .00375 L
s
.36667 0 m
.36667 .00375 L
s
.4 0 m
.4 .00375 L
s
.43333 0 m
.43333 .00375 L
s
.46667 0 m
.46667 .00375 L
s
.53333 0 m
.53333 .00375 L
s
.56667 0 m
.56667 .00375 L
s
.6 0 m
.6 .00375 L
s
.63333 0 m
.63333 .00375 L
s
.7 0 m
.7 .00375 L
s
.73333 0 m
.73333 .00375 L
s
.76667 0 m
.76667 .00375 L
s
.8 0 m
.8 .00375 L
s
.86667 0 m
.86667 .00375 L
s
.9 0 m
.9 .00375 L
s
.93333 0 m
.93333 .00375 L
s
.96667 0 m
.96667 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -76.0625 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
72.125 15.813 moveto
(H) show
75.938 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.313 15.813 moveto
(L) show
89.125 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .03333 m
.00625 .03333 L
s
gsave
-0.0125 .03333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .06667 m
.00625 .06667 L
s
gsave
-0.0125 .06667 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .1 m
.00625 .1 L
s
gsave
-0.0125 .1 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
0 .13333 m
.00625 .13333 L
s
gsave
-0.0125 .13333 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .16667 m
.00625 .16667 L
s
gsave
-0.0125 .16667 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
0 .2 m
.00625 .2 L
s
gsave
-0.0125 .2 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .00667 m
.00375 .00667 L
s
0 .01333 m
.00375 .01333 L
s
0 .02 m
.00375 .02 L
s
0 .02667 m
.00375 .02667 L
s
0 .04 m
.00375 .04 L
s
0 .04667 m
.00375 .04667 L
s
0 .05333 m
.00375 .05333 L
s
0 .06 m
.00375 .06 L
s
0 .07333 m
.00375 .07333 L
s
0 .08 m
.00375 .08 L
s
0 .08667 m
.00375 .08667 L
s
0 .09333 m
.00375 .09333 L
s
0 .10667 m
.00375 .10667 L
s
0 .11333 m
.00375 .11333 L
s
0 .12 m
.00375 .12 L
s
0 .12667 m
.00375 .12667 L
s
0 .14 m
.00375 .14 L
s
0 .14667 m
.00375 .14667 L
s
0 .15333 m
.00375 .15333 L
s
0 .16 m
.00375 .16 L
s
0 .17333 m
.00375 .17333 L
s
0 .18 m
.00375 .18 L
s
0 .18667 m
.00375 .18667 L
s
0 .19333 m
.00375 .19333 L
s
.25 Mabswid
0 0 m
0 .2 L
s
gsave
-0.01875 .1 -99.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
75.8125 11.8438 translate
90 rotate
-75.8125 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
71.625 15.813 moveto
(H) show
75.438 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(m) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
84.813 15.813 moveto
(L) show
88.625 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .19375 m
0 .2 L
s
.16667 .19375 m
.16667 .2 L
s
.33333 .19375 m
.33333 .2 L
s
.5 .19375 m
.5 .2 L
s
.66667 .19375 m
.66667 .2 L
s
.83333 .19375 m
.83333 .2 L
s
.125 Mabswid
.03333 .19625 m
.03333 .2 L
s
.06667 .19625 m
.06667 .2 L
s
.1 .19625 m
.1 .2 L
s
.13333 .19625 m
.13333 .2 L
s
.2 .19625 m
.2 .2 L
s
.23333 .19625 m
.23333 .2 L
s
.26667 .19625 m
.26667 .2 L
s
.3 .19625 m
.3 .2 L
s
.36667 .19625 m
.36667 .2 L
s
.4 .19625 m
.4 .2 L
s
.43333 .19625 m
.43333 .2 L
s
.46667 .19625 m
.46667 .2 L
s
.53333 .19625 m
.53333 .2 L
s
.56667 .19625 m
.56667 .2 L
s
.6 .19625 m
.6 .2 L
s
.63333 .19625 m
.63333 .2 L
s
.7 .19625 m
.7 .2 L
s
.73333 .19625 m
.73333 .2 L
s
.76667 .19625 m
.76667 .2 L
s
.8 .19625 m
.8 .2 L
s
.86667 .19625 m
.86667 .2 L
s
.9 .19625 m
.9 .2 L
s
.93333 .19625 m
.93333 .2 L
s
.96667 .19625 m
.96667 .2 L
s
.25 Mabswid
0 .2 m
1 .2 L
s
gsave
.5 .2125 -110.188 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Quadratic) show
112.938 15.813 moveto
(Damping) show
157.375 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.99375 0 m
1 0 L
s
.99375 .03333 m
1 .03333 L
s
.99375 .06667 m
1 .06667 L
s
.99375 .1 m
1 .1 L
s
.99375 .13333 m
1 .13333 L
s
.99375 .16667 m
1 .16667 L
s
.125 Mabswid
.99625 .00667 m
1 .00667 L
s
.99625 .01333 m
1 .01333 L
s
.99625 .02 m
1 .02 L
s
.99625 .02667 m
1 .02667 L
s
.99625 .04 m
1 .04 L
s
.99625 .04667 m
1 .04667 L
s
.99625 .05333 m
1 .05333 L
s
.99625 .06 m
1 .06 L
s
.99625 .07333 m
1 .07333 L
s
.99625 .08 m
1 .08 L
s
.99625 .08667 m
1 .08667 L
s
.99625 .09333 m
1 .09333 L
s
.99625 .10667 m
1 .10667 L
s
.99625 .11333 m
1 .11333 L
s
.99625 .12 m
1 .12 L
s
.99625 .12667 m
1 .12667 L
s
.99625 .14 m
1 .14 L
s
.99625 .14667 m
1 .14667 L
s
.99625 .15333 m
1 .15333 L
s
.99625 .16 m
1 .16 L
s
.99625 .17333 m
1 .17333 L
s
.99625 .18 m
1 .18 L
s
.99625 .18667 m
1 .18667 L
s
.99625 .19333 m
1 .19333 L
s
.25 Mabswid
1 0 m
1 .2 L
s
gsave
1.0125 .1 -56.6563 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
64.5 11.8438 translate
90 rotate
-64.5 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
66.000 15.813 moveto
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .2 L
0 .2 L
closepath
clip
newpath
.5 Mabswid
0 0 m
.05179 .04077 L
.10828 .07911 L
.16133 .10929 L
.21235 .133 L
.24078 .14395 L
.26665 .1525 L
.29383 .16004 L
.31892 .16569 L
.34459 .17017 L
.35812 .17199 L
.37245 .17353 L
.38507 .17454 L
.39216 .17496 L
.39864 .17527 L
.40465 .17547 L
.40806 .17556 L
.40968 .17559 L
.41122 .17561 L
.41268 .17563 L
.41425 .17565 L
.41569 .17566 L
.417 .17567 L
.41843 .17567 L
.42 .17567 L
.42164 .17566 L
.42317 .17565 L
.42486 .17563 L
.42666 .17561 L
.42988 .17554 L
.43335 .17546 L
.43718 .17533 L
.44403 .17503 L
.45046 .17466 L
.46243 .17375 L
.47533 .17246 L
.5022 .16869 L
.53048 .16315 L
.58157 .14909 L
.63595 .12838 L
.68829 .10285 L
.74392 .0697 L
.79752 .03189 L
s
.79752 .03189 m
.83685 0 L
s
0 0 m
.05009 .0394 L
.10129 .07373 L
.14664 .09848 L
.18813 .11593 L
.20874 .1226 L
.2303 .12806 L
.24065 .13011 L
.25175 .13189 L
.2568 .13256 L
.26212 .13315 L
.26708 .13361 L
.27159 .13395 L
.27384 .13409 L
.27628 .13422 L
.27861 .13433 L
.28072 .1344 L
.28317 .13447 L
.28452 .13449 L
.28578 .13451 L
.28695 .13452 L
.28803 .13453 L
.28921 .13453 L
.29045 .13453 L
.29154 .13452 L
.29253 .1345 L
.29367 .13449 L
.29476 .13446 L
.29667 .13441 L
.29876 .13434 L
.30346 .13411 L
.30779 .13382 L
.31679 .13297 L
.32516 .13188 L
.33426 .13036 L
.34398 .12834 L
.36193 .1235 L
.38045 .11697 L
.41297 .1015 L
.44625 .08003 L
.477 .05471 L
.50828 .02306 L
s
.50828 .02306 m
.52734 0 L
s
0 0 m
.04853 .03814 L
.09538 .06918 L
.13503 .0899 L
.15371 .09758 L
.17006 .10309 L
.18708 .10752 L
.19619 .10931 L
.20464 .11059 L
.20874 .11108 L
.21256 .11145 L
.2165 .11174 L
.21858 .11187 L
.22081 .11197 L
.22269 .11204 L
.22374 .11206 L
.22473 .11208 L
.22567 .1121 L
.22652 .11211 L
.2275 .11211 L
.22842 .11211 L
.22926 .11211 L
.23014 .1121 L
.2309 .11208 L
.23172 .11207 L
.23356 .11202 L
.23527 .11195 L
.23728 .11185 L
.23913 .11173 L
.24328 .1114 L
.24692 .11102 L
.25075 .11053 L
.2573 .10946 L
.26443 .10799 L
.27829 .10411 L
.29278 .09858 L
.30626 .09197 L
.32001 .0837 L
.3437 .06549 L
.36731 .04178 L
.38851 .01509 L
s
.38851 .01509 m
.39833 0 L
s
0 0 m
.04709 .03698 L
.0903 .06526 L
.10908 .07535 L
.12555 .08287 L
.14153 .08884 L
.15591 .09298 L
.16326 .0946 L
.17105 .09593 L
.17443 .09638 L
.17797 .09677 L
.18096 .09703 L
.18422 .09723 L
.18595 .09731 L
.18692 .09734 L
.18782 .09736 L
.18868 .09738 L
.18946 .09739 L
.19035 .0974 L
.19119 .0974 L
.19196 .0974 L
.19277 .09739 L
.19346 .09738 L
.19421 .09736 L
.19588 .0973 L
.19744 .09723 L
.19924 .09712 L
.20094 .097 L
.20409 .0967 L
.20745 .0963 L
.21107 .09575 L
.21702 .0946 L
.22347 .09298 L
.23611 .08865 L
.2473 .08343 L
.25889 .07655 L
.28036 .05937 L
.29932 .03868 L
.31602 .01535 L
s
.31602 .01535 m
.32487 0 L
s
0 0 m
.04576 .03591 L
.08586 .06183 L
.10288 .07064 L
.11762 .07696 L
.13097 .08148 L
.13807 .08337 L
.14444 .08473 L
.15115 .08581 L
.15461 .08622 L
.15636 .08638 L
.15822 .08653 L
.15984 .08663 L
.16131 .0867 L
.16202 .08673 L
.16278 .08675 L
.16351 .08677 L
.16417 .08678 L
.16458 .08678 L
.16502 .08679 L
.16581 .08679 L
.16658 .08679 L
.16729 .08678 L
.16805 .08676 L
.16889 .08674 L
.16975 .08671 L
.17056 .08667 L
.17211 .08658 L
.17353 .08648 L
.17673 .08618 L
.1795 .08582 L
.18242 .08536 L
.18738 .08436 L
.19274 .08295 L
.20303 .07926 L
.21362 .07398 L
.22331 .06771 L
.23302 .0599 L
.2493 .04291 L
.26493 .02115 L
s
.26493 .02115 m
.27677 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{476.25, 95.25},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000;:0000SR000`400?l00000o`00003oo`3ooooo
0?oool`0oooo003o0?ooool0ooooc03oool00?l0ooooK`3oool00`3ooml0W4Q809cOo`0>0?ooo`03
0?oOW018B9`0gooo0?l0ooooB@3oool00?l0ooooK`3oool00`2oM4P0W=oo0?ooo`0?0?ooo`030?oO
W018M;l0oooo0?l0ooooB03oool00?l0ooooKP3oool00`3OW4P0B9cO0?ooo`0A0?ooo`030=nLB018
W=l0oooo0?l0ooooA`3oool00?l0ooooK@3oool00`3ookl0M01d0;ooo`0B0?ooo`030?oo_`1d04P0
W=oo0?l0ooooA`3oool00?l0ooooG`3oool01@2oM0000000001d_`3oool0gia80080000000<004RL
0=ooo`3oool0103oool00`3OW4P004RL0=ooo`020?ooo`060=nLB000000000000000B02Lgml0W4P0
0P0000000`0007@0_ooO09a8000200000006001d_`3oool0oooo0?oog`2LB000B9cOo`3ooom70?oo
o`00o`3ooomO0?ooo`080?oog`2LB7@0_ooo0?ooo`3ooml0W4P00018W03Oool60?ooo`030;md0000
M;l0oooo00<0oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000
M;l0103oool00`3OW4P004RL0=ooo`3o0?ooodH0oooo003o0?ooof00oooo00H0oono07A8W03Oool0
_g@00000M02oool60?ooo`030?oo_`1d0000B9cO00@0oooo00<0gia80000M02oool00P3oool01`2o
M00004RL0=ooo`3oool0oooO09a80000M;l0103oool00`3OW4P0001d0;ooo`3o0?ooodH0oooo003o
0?ooof40oooo00@0omnL07A8B0000000M;oo1`3oool00`3ookl0M00007Boo`040?ooo`030=nLB000
07@0_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo00@0oooo00<0okmd0000M02o
ool0o`3ooom60?ooo`00o`3ooomQ0?ooo`030?oo_`1d0000B9cO00P0oooo00<0oono07@0001d_ol0
103oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`040?oo
o`030?noM00004P0W=oo0?l0ooooAP3oool00?l0ooooH03oool01@3ooml0W4P00018W02LM4P0W=oo
00L0oooo00<0oono07@0001d_ol0103oool00`3OW4P0001809cOo`020?ooo`070;md000007@0_ooo
0?ooo`3ooml0W4P0001d_`040?ooo`030?noM00007@0_ooo0?l0ooooAP3oool00?l0ooooH03oool0
1P2oM000001809cOo`3ooml0W4Qd0;ooo`H0oooo00<0oono07@00018W=l00P3oool03P3ooml0W4P0
0000000004P0B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO103oool00`3OW4P0
001d0;ooo`3o0?ooodH0oooo003o0?oooeh0oooo00T0oooO09a8000000000000001d_`3o_g@00000
0000M02oool01@3oool00`3ooml0W4P004RLg`040?ooo`0<0;md000007@0_ooO09a8000000000018
09cOo`3ookl0M0000000000007@0_ooo103oool00`3OW4P004RL0=ooo`3o0?ooodH0oooo003o0?oo
ofd0oooo00<0gia80018W03Oool04P3oool00`3ooml0W4P004RLg`3o0?ooodL0oooo003o0?ooofd0
oooo00<0oono07@0B02Lgol04P3oool00`3ookl0M01809cOo`3o0?ooodL0oooo003o0?ooofh0oooo
00<0gia804RLg`3oool04@3oool00`3OW4P0B9cO0?ooo`3o0?ooodL0oooo003o0?ooofl0oooo00<0
_g@004RLg`3oool03`3oool00`3o_g@007Bo0?ooo`3o0?ooodP0oooo003o0?ooog00oooo00<0giad
09cOo`3oool03@3oool00`3ookl0M7Bo0?ooo`3o0?ooodT0oooo003o0?ooool0ooooc03oool00?l0
ooooo`3oooo<0?ooo`00o`3ooooo0?oool`0oooo003o0?ooool0ooooc03oool00?l0ooooo`3oooo<
0?ooo`00o`3ooooo0?oool`0oooo003o0?ooool0ooooc03oool00?l0ooooo`3oooo<0?ooo`00o`3o
oooo0?oool`0oooo003o0?ooool0ooooc03oool00?l0ooooo`3oooo<0?ooo`00o`3oool00`3ooml0
W4P0000000020000000307Boo`3oool0oooo00<0oooo00D0omnL04P000000000001d0;ooo`0R0?oo
o`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000000007@0_ooo00@0oooo
00D0omnL04P000000000001d0;ooo`0N0?ooo`090;md00000000000000000000B9`0gooo0?ooo`3o
oml0W4P000<0000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo01/0oooo
00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3oool0omnL04P000000000001d0;ooo`04
0?ooo`050?oOW018000000000000M02oool0703oool00`2oM0000000000000040000000404RLg`3o
ool0oooO09a800<0000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo01`0
oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?ooo`030?ooo`050?oOW018000000000000
M02oool0103oool01@3ogi`0B0000000000007@0_ooo0:l0oooo003o0?ooo`030?oog`2LB7@0_ooo
0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool08P3o
ool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0N0?ooo`030?oOW0180000M;oo00<0oooo00<0oooO
09a8M02oool00P3oool00`2oM000M;oo0?ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`0K0?ooo`030=nLB0000000000000<0000000X004RL0=ooo`3oool0oono07@0001d_ol0oooo
0?noM00007@0_ooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0703oool00`3OW4P0
000000000003000000050018W03Oool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`020?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0K0?ooo`030?oOW0180000B9cO0080oooo00<0
_g@004RLg`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3ookl0
M00007Boo`3oool0okmd0000M02oool0[P3oool00?l0oooo1@3oool00`3ooml0W4Qd0;ooo`020?oo
o`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool08@3oool00`3ogi`0B00007Boo`040?oo
o`030;md0000B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool07@3oool00`3ogi`0B00007Boo`080?ooo`030?oog`2LB7@0_ooo0080oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0K0?ooo`030=nLB018W=l0oooo0080oooo00H0
giad09cOo`3oool0_g@00018W03Oool20?ooo`070;md0000M;l0oooo0?ooo`2oM00004RL0=ooo`02
0?ooo`030;md0000M;l0oooo01`0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`04
0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0O
0?ooo`030?oo_`1d04P0W=oo0080oooo00<0_g@00018W03Oool00P3oool01`2oM00007Bo0?ooo`3o
ool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`2]0?ooo`00o`3oool60?ooo`060;md001d
_ol0oooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0Q0?ooo`030?oOW0180000M;oo00<0
oooo00<0oono07@00018W=l00`3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`<0oooo00<0gia8
0000M02oool07@3oool00`3ogi`0B00007Boo`090?ooo`060;md001d_ol0oooo0?oo_`1d0000B9cO
0`3oool00`3OW4P0001d0;ooo`0L0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l0
0`3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool07@3oool00`3O
W4P0M;oo0?ooo`090?ooo`060;md001d_ol0oooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;oo
o`0O0?ooo`070?oo_`1d0000M;oo0?ooo`3ookl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo
_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`2]0?ooo`00o`3oool50?ooo`070?oog`2LB000B9cO
0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0240oooo00<0omnL04P0001d_ol00`3o
ool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0
001809cOo`0M0?ooo`030?oOW0180000M;oo00P0oooo00L0oooO09a80018W=l0oooo0?oOW0180000
M;oo00<0oooo00<0okmd0000B02Lgol07@3oool00`2oM4P0W=oo0?ooo`020?ooo`030?oOW0180000
M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo01h0oooo
00<0_gA809cOo`3oool01`3oool01`3ooml0W4P004RLg`3oool0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0O0?ooo`070?noM0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`060?noM000
04P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`2]0?ooo`00o`3oool50?ooo`070=nL
B0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0240oooo00<0omnL04P0
001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3o
ool00`3o_g@0001809cOo`0M0?ooo`030?oOW0180000M;oo00P0oooo00L0gia80000001d_ol0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol07@3oool00`3ooml0W4Qd0;ooo`020?ooo`03
0?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0
W=oo01h0oooo00<0oooO09a8M02oool01`3oool01`3OW4P0000007Boo`3oool0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0N0?ooo`080?oOW0180000001809cOo`3oool0omnL04P0001d_ol3
0?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`2]0?ooo`00o`3o
ool10?ooo`0;0=oOg`3oool0oono07@000000000001d0;ooo`3oool0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`0Q0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool0
1P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol07@3oool00`3ogi`0B000
07Boo`040?ooo`0;0=oOg`3oool0oono07@000000000001d0;ooo`3oool0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`0N0?ooo`070?noM000M;l0oooo0?ooo`3ogi`0B00007Boo`030?ooo`06
0?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0O0?ooo`030?noM000M;l0
oooo0080oooo00/0gmoO0?ooo`3ookl0M0000000000007@0_ooo0?ooo`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo01`0oooo00D0okmd000000000000001809cOo`020?ooo`030?oOW0180000
M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0:d0oooo
003o0?ooo`40oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol08@3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0
oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo01d0oooo00<0omnL
04P0001d_ol0103oool01P3OW4P00000000000000000001809cOo`80oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0N0?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`030?oo
o`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0O0?ooo`030?oo_`1d
04P0W=oo0080oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol07P3oool00`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo00<0
oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0:d0oooo003o0?oo
o`40oooo00@0oono07@000000000B9cO103oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo0240oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`060=nLB00007@0
_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0M0?ooo`030?oOW0180000M;oo00@0oooo
00@0oono07@000000000B9cO103oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo01h0
oooo00L0oono07@00018W=l0oooo0?oo_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@0
0018W=l30?ooo`030=nLB00007@0_ooo01l0oooo00<0oono07@00018W=l00P3oool0103ookl0M000
00000018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool07`3oool01`2oM000
B9cO0?ooo`3oool0oono07@00018W=l00`3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`<0oooo
00<0gia80000M02oool0[@3oool00?l0oooo0P3oool00`2oM000M;oo0?ooo`050?ooo`030=nLB000
B9`0gooo0080oooo00<0_g@0001d_`3oool08@3oool00`3ogi`0B00007Boo`040?ooo`030=nLB000
B9`0gooo0080oooo00L0_g@0001d_`3oool0oooo0=nLB000B9`0gooo0080oooo00<0_g@0001d_`3o
ool07@3oool00`3ogi`0B00007Boo`050?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`0I0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018
W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo01X0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`03
0?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0J
0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia80000M02oool00P3oool00`3OW4P004RL0=ooo`02
0?ooo`070;md0000M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0:d0oooo
003o0?ooo`80oooo00H0omnL04P00000000000000000B02Lgol20?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`0P0?ooo`050?oog`2LB00000000000001d_ol0103oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0703oool0
1@3ooml0W4P0000000000000M;oo00D0oooo00H0omnL04P00000000000000000B02Lgol20?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0J0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M000
0000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`0K0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool3
0?ooo`060?oOW018000000000000000004P0W=oo0P3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0703oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_`<0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0:h0oooo003o0?oo
o`<0oooo00D0_g@0000000000000000004RLg`030?ooo`050?oo_`1d000000000000M02oool08`3o
ool00`3o_g@0000007Boo`050?ooo`050?oo_`1d000000000000M02oool0103oool01@3ookl0M000
0000000007@0_ooo01l0oooo00<0okmd0000001d_ol01P3oool01@2oM0000000000000000000B9cO
00<0oooo00D0oono07@000000000001d0;ooo`0L0?ooo`060?oog`2LB00000000000000007@0_ooo
103oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`0M0?oo
o`060?oog`2LB00000000000000007@0_ooo1@3oool01@2oM0000000000000000000B9cO00<0oooo
00D0oono07@000000000001d0;ooo`0N0?ooo`050=nLB000000000000000M02oool0103oool01@3o
okl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`2_0?ooo`00o`3ooooo
0?oool`0oooo003o0?ooool0ooooc03oool00?l0ooooo`3oooo<0?ooo`00o`3ooooo0?oool`0oooo
003o0?ooool0ooooc03oool00?l0ooooo`3oooo<0?ooo`00o`3ooooo0?oool`0oooo003o0?ooool0
ooooc03oool00?l0ooooo`3oooo<0?ooo`00d`3ooooo000003l00000^@3oool00=<0oooo0`000007
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`000000203oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo000000X0oooo00<000000?ooo`3oool01P3oool2000000X0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00D0oooo00@0
00000?ooo`3oool000002@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0;L0oooo
003C0?ooo`@00000<@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00d0oooo00@0
00000?ooo`3oool000004P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0`0?ooo`<00000<P3oool300000;T0oooo
003C0?ooo`050000003oool0oooo0?ooo`000000<03oool00`0007l0oooo0?ooo`0M0?ooo`030000
003oool0oooo00d0oooo00D000000?ooo`3oool0oooo0000O`0B0?ooo`030000003oool0oooo01l0
oooo00<0001o0?ooo`3oool00`3oool00`000000oooo0?ooo`0/0?ooo`030000O`3oool0oooo02l0
oooo00@000000?ooo`3oool0001o=03oool00`000000oooo0?ooo`2g0?ooo`00d`3oool2000000<0
oooo0P00000^0?ooo`030000O`3oool0oooo01d0oooo00<000000?ooo`3oool03@3oool01@000000
oooo0?ooo`3oool0001o0140oooo00<000000?ooo`3oool0803oool00`0007l0oooo0?ooo`030?oo
o`030000003oool0oooo02`0oooo00<0001o0?ooo`3oool0;P3oool01@000000oooo0?ooo`3oool0
001o0380oooo0`00002i0?ooo`00_03oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0
omnL04P000000000001d0;ooo`090?ooo`030000003oool0oooo00<0oooo0P00000]0?ooo`030000
O`3oool0oooo01`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`020?ooo`030000
O`3oool0oooo00h0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`020?ooo`030000
003oool0oooo02d0oooo00<0001o0?ooo`3oool0;03oool2000000@0oooo00<0001o0?ooo`3oool0
<P3oool00`000000oooo0?ooo`2g0?ooo`00_@3oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`8000001@3oool2000002`0oooo00<0001o0?ooo`3o
ool0703oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3o
ool03@3oool00`000000oooo0?ooo`0R0?ooo`050000O`3oool0oooo0?ooo`000000<03oool00`00
07l0oooo0?ooo`0[0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool0<03oool30000
0;T0oooo002m0?ooo`030?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`2oM000
07Bo0?ooo`070?ooo`030000003oool0oooo00D0oooo0`00000Z0?ooo`030000O`3oool0oooo01/0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?ooo`030000O`3oool0oooo00d0
oooo00<000000?ooo`3oool08P3oool0100007l0oooo0?ooo`00000a0?ooo`030000O`3oool0oooo
02X0oooo00<000000?ooo`3oool01@3oool00`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo
0;L0oooo002m0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0
001d0;ooo`070?ooo`800000203oool2000002T0oooo00<0001o0?ooo`3oool06`3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool0303oool00`000000
oooo0?ooo`0S0?ooo`030000O`3oool000000380oooo00<0001o0?ooo`3oool0:03oool2000000P0
oooo00<0001o0?ooo`3oool0<03oool300000;T0oooo002m0?ooo`030?oOW0180000M;oo00<0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo00P0oooo
0P00000X0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`050?ooo`030000O`3oool0oooo00/0oooo00<000000?ooo`3oool0903oool00`0007l00000
0?ooo`0b0?ooo`030000O`3oool0oooo02L0oooo00<000000?ooo`3oool0203oool00`0007l0oooo
0?ooo`0b0?ooo`030000003oool0oooo0;L0oooo002m0?ooo`030?oOW0180000M;oo00<0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`<000002@0007l300000480001o00<0
00000000O`0007l0300007l00`000000001o0000O`0C0000O`030000000007l0001o02D0001o00<0
00000000O`0007l0F`0007l00`000000001o0000O`0l0000O`<00000^@3oool00;d0oooo00<0omnL
04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<00000
0?ooo`3oool02P3oool3000002D0oooo00<0001o0?ooo`3oool0603oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00H0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0T
0?ooo`030000000007l0oooo03<0oooo00<0001o0?ooo`3oool0903oool2000000`0oooo00<0001o
0?ooo`3oool0<P3oool00`000000oooo0?ooo`2g0?ooo`00_@3oool00`3ogi`0B00007Boo`030?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool2000000d0oooo0`00000S0?oo
o`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`070?oo
o`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool0903oool00`000000oooo0000O`0d0?oo
o`030000O`3oool0oooo02<0oooo00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`0`0?oo
o`<00000^@3oool00:L0oooo00<0_gA809cOo`3oool04`3oool00`3ogi`0B00007Boo`030?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`0=0?ooo`<0
00008P3oool00`0007l0oooo0?ooo`0G0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool0203oool00`0007l0oooo0?ooo`080?ooo`030000003oool0oooo02@0oooo00@000000?ooo`3o
ool0001o=03oool00`0007l0oooo0?ooo`0R0?ooo`030000003oool0oooo00d0oooo00<0001o0?oo
o`3oool0<P3oool00`000000oooo0?ooo`2g0?ooo`00Y`3oool00`2oM000000007Boo`060?ooo`03
0?oog`2LB00007Bo00X0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00L0oooo0P00000?0?ooo`@00000803oool00`0007l0oooo0?ooo`0F0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`070?ooo`03
0000003oool0oooo02@0oooo00D000000?ooo`3oool0oooo0000O`0d0?ooo`030000O`3oool0oooo
0200oooo0P00000@0?ooo`030000O`3oool0oooo0300oooo0`00002i0?ooo`00Y`3oool01@2oM000
0000000000000000M;oo00D0oooo00<0_g@00018W03Oool01`3oool01@3ooml0W4P0000000000000
M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0
403oool3000001l0oooo00<0001o0?ooo`3oool05@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00X0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0T0?ooo`030000
003oool0oooo0080oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0O0?ooo`030000
003oool0oooo0100oooo00<0001o0?ooo`3oool0<P3oool00`000000oooo0?ooo`2g0?ooo`00Y`3o
ool01`2oW9`0gooo0?oog`2LB00000000000B02Lgol00P3oool00`3o_g@0001809cOo`070?ooo`03
0;md00000000000000@0000000T0B9cO0?ooo`3oool0oooo0?oOM00P000000000000=@2?ool02@3o
ool00`000000oooo0?ooo`0A0?ooo`<000007P3oool00`0007l0oooo0?ooo`0D0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`0007l0oooo0?ooo`050?ooo`030000003o
ool0oooo02@0oooo00<000000?ooo`3oool00`3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3o
ool0oooo01d0oooo0P00000C0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3oool0]`3o
ool00:/0oooo00<0oono07@0000000000P0000000`00M;l0oooo0?ooo`080?ooo`030=nLB0000000
000000<0000000X004RL0=ooo`3oool0oono07@0001d_ol0oooo0?noM00007@0_ooo203oool20000
01@0oooo0`00000L0?ooo`030000O`3oool0oooo01@0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0<0?ooo`030000O`3oool0oooo00@0oooo00<000000?ooo`3oool0903oool00`00
0000oooo0?ooo`040?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0703oool00`00
0000oooo0?ooo`0C0?ooo`030000O`3oool0oooo0300oooo0`00002i0?ooo`00Z`3oool0103ooml0
W4P00000001d_ol=0?ooo`030=nLB018W=l0oooo0080oooo00H0giad09cOo`3oool0_g@00018W03O
ool20?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0503oool4000001X0oooo00<0
001o0?ooo`3oool04`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00d0oooo00<0
001o0?ooo`3oool00`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo00D0oooo00<0
001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0J0?ooo`8000005P3oool00`0007l0oooo0?oo
o`0b0?ooo`030000003oool0oooo0;L0oooo002W0?ooo`060;nLW03Oool0omnL04P0000004P0W=oo
403oool00`3OW4P0M;oo0?ooo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0
1`3oool2000001L0oooo0`00000I0?ooo`030000O`3oool0oooo0180oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0>0?ooo`030000O`3oool0oooo0080oooo00<000000?ooo`3oool0
8`3oool2000000P0oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0I0?ooo`030000
003oool0oooo01H0oooo00<0001o0?ooo`3oool0<03oool300000;T0oooo002W0?ooo`040;md0000
000004RL0=oooa<0oooo00<0_gA809cOo`3oool00P3oool00`3ogi`0B00007Boo`030?ooo`030?no
M00004P0W=oo00L0oooo00<000000?ooo`3oool05`3oool4000001L0oooo00<0001o0?ooo`3oool0
403oool2000000d0oooo00<000000?ooo`3oool03`3oool01@0007l0oooo0?ooo`3oool0000002D0
oooo00<000000?ooo`3oool0203oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo01P0
oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0
oooo002W0?ooo`030;mdM02oool0oooo01@0oooo00<0oooO09a8M02oool00P3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00L0oooo0`00000I0000O`@000009`0007l00`000000001o
0000O`0;0000O`8000005@0007l00`000000001o0000O`0S0000O`030000000007l0001o05L0001o
0P00001=0000O`<00000^@3oool00;l0oooo00L0okmd001d_`3oool0oooo0?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0J0?ooo`D000004`3oool00`0007l0
oooo0?ooo`0=0?ooo`8000003@3oool00`000000oooo0?ooo`0B0?ooo`030000O`000000000002D0
oooo00<000000?ooo`3oool02P3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo01D0
oooo00<000000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0
oooo002o0?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00L0oooo0P00000M0?ooo`D000004@3oool00`0007l0oooo0?ooo`0;0?ooo`8000003P3oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo02<0oooo0P00000=0?ooo`030000O`3oool0oooo
0380oooo00<0001o0?ooo`3oool04`3oool2000001d0oooo00<0001o0?ooo`3oool0<03oool30000
0;T0oooo002o0?ooo`070?oo_`1d0000B9cO0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00L0oooo00<000000?ooo`3oool07@3oool7000000h0oooo00<0001o0?ooo`3oool02@3oool2
000000l0oooo00<000000?ooo`3oool04`3oool00`000000001o0?ooo`0S0?ooo`030000003oool0
oooo00d0oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0A0?ooo`8000007`3oool0
0`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0oooo002j0?ooo`030?noW02Lgol0oooo
0080oooo00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo
00L0oooo0P00000P0?ooo`T000002P3oool00`0007l0oooo0?ooo`070?ooo`8000003`3oool20000
01@0oooo0P0000000`3oool0001o0?ooo`0R0?ooo`030000003oool0oooo00h0oooo00<0001o0?oo
o`3oool0<P3oool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0oooo01l0oooo00<0001o0?oo
o`3oool0<03oool300000;T0oooo002j0?ooo`030?ooY`18000000000080000000<0=@000000M02o
ool00P3oool01`3ookl0M000020000000000000e0000M02oool0203oool00`000000oooo0?ooo`0P
0?ooo`L0000000<0oooo0000000000000P0000060?ooo`030000O`3oool0oooo00<0oooo1000000?
0?ooo`8000005@3oool01@000000oooo0?ooo`3oool0001o0240oooo0P00000A0?ooo`030000O`3o
ool0oooo0380oooo00<0001o0?ooo`3oool03P3oool200000280oooo00<0001o0?ooo`3oool0<P3o
ool00`000000oooo0?ooo`2g0?ooo`00Y@3oool00`3ooml0W4P00000000600000003001d_`3oool0
oooo00T0oooo00<0omnL04P0000[;0000`0000002`0eW=l0oooo0?ooo`3oool0oono07@0000e0000
00000000=@0007@0_ooo00P0oooo0P00000S0?ooo`L000000`3oool<00000140oooo0P00000F0?oo
o`030000003oool0oooo0080oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0A0?oo
o`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0303oool2000002@0oooo00<0001o0?oo
o`3oool0<03oool300000;T0oooo002S0?ooo`030?oog`2LB000000000X0000000<007Bo0?ooo`3o
ool0303oool00`3ookl0M01809cOo`020?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3o
ool01`3oool00`000000oooo0?ooo`0T0?ooo`P000001P3oool00`0007l0oooo0?ooo`0B0?ooo`80
00005P3oool2000000D0oooo00<0001o0?ooo`3oool07@3oool00`000000oooo0?ooo`0B0?ooo`03
0000O`3oool0oooo0380oooo00<0001o0?ooo`3oool02P3oool2000002H0oooo00<0001o0?ooo`3o
ool0<P3oool00`000000oooo0?ooo`2g0?ooo`00XP3oool0103o_g@000000018W03Oool90?ooo`04
0?oo_`1d0000001809cOo``0oooo00L0oono07@0001d_ol0oooo0?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool01`3oool2000002L0oooo1@000000103oool000000000000000030?ooo`030000
O`3oool0oooo00l0oooo0`00000G0?ooo`030000003oool0oooo00D0oooo00<0001o0?ooo`3oool0
6`3oool2000001D0oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`090?ooo`030000
003oool0oooo02H0oooo00<0001o0?ooo`3oool0<03oool300000;T0oooo002Q0?ooo`030?oo_`1d
M;l0oooo00d0oooo00<0oooO09adW03Oool02`3oool01`3o_g@0000007Boo`3oool0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo02P0oooo1@0000020?oooaD0
0000603oool2000000P0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`0E0?ooo`03
0000O`3oool0oooo0380oooo00<0001o0?ooo`3oool01`3oool2000002T0oooo00<0001o0?ooo`3o
ool0<P3oool00`000000oooo0?ooo`2g0?ooo`00_P3oool0203ogi`0B0000000B02Lgol0oooo0?oO
W0180000M;oo0`3oool00`3o_g@0001809cOo`070?ooo`<00000:P0007l6000000T0001o0P00000N
0000O`<000009@0007l2000005D0001o0P00001N0000O`<00000^@3oool00:L0oooo00<0oono07Ad
_`3oool0103oool00`3OW9`0_ooo0?ooo`0;0?ooo`050?noM000000000000000B02Lgol00P3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0;03oool3
000000030?ooo`000000000000<000008P3oool2000000d0oooo00<0001o0?ooo`3oool05P3oool2
000001X0oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`030?ooo`800000;@3oool0
0`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0oooo002W0?ooo`030?noM00000000000
00@0000000<0001d0;ooo`3oool03@3oool00`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01`3oool2000002l0oooo100000000`0007l0oooo0?ooo`020000
01d0oooo0`00000?0?ooo`030000O`3oool0oooo01@0oooo0P00000L0?ooo`030000O`3oool0oooo
0380oooo00@0001o0?ooo`3oool0oooo0P00000_0?ooo`030000O`3oool0oooo0300oooo0`00002i
0?ooo`00Y`3oool00`2oM000000000000004000000030000M02oool0oooo00h0oooo00L0_g@004RL
g`3oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?oo
o`0`0?ooo`030000003oool00000008000000P3oool600000140oooo1P00000B0?ooo`030000O`3o
ool0oooo0180oooo0P00000N0?ooo`030000O`3oool0oooo0380oooo00@0001o0?ooo`0000000000
<@3oool00`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0oooo002X0?ooo`030;mdM02o
ool0oooo00<0oooo00<0gibL0;ooo`3oool02@3oool00`3ooml0W9bo0?ooo`020?ooo`030=nLB000
07@0_ooo0080oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003o
ool0oooo0340oooo0P0000020?ooo`8000001P3ooolA000001P0oooo00<0001o0?ooo`3oool0403o
ool200000200oooo00<0001o0?ooo`3oool0<P3oool2000003<0oooo00<0001o0?ooo`3oool0<P3o
ool00`000000oooo0?ooo`2g0?ooo`00Y`3oool00`3ogi`0W;oo0?ooo`0A0?ooo`060?noM018W=l0
oooo0?oo_`1d000007Bo0`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool20000
03<0oooo00<0001o0000000000000P3oool3000002`0oooo00<0001o0?ooo`3oool03P3oool20000
0280oooo00<0001o0?ooo`3oool0<03oool2000000030000O`3oool0oooo0380oooo00<0001o0?oo
o`3oool0<03oool300000;T0oooo002W0?ooo`030;md0000M;l0oooo00@0oooo00<0gibL0;ooo`3o
ool02`3oool01P3OW4P0000000000000000004RL0=ooo`<0oooo00D0ommd020000000000000e08oo
o`090?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3oool00P0000030?ooo`@00000:03o
ool00`0007l0oooo0?ooo`0;0?ooo`<00000903oool00`0007l0oooo0?ooo`0^0?ooo`8000000P3o
ool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3oool0]`3o
ool00:L0oooo00<0gia8000000000000100000000`0007@0_ooo0?ooo`0=0?ooo`040?oog`2LB000
04RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo0P00000c0?ooo`03
0000O`3oool0oooo0080oooo0P0000050?ooo`<000009@3oool00`0007l0oooo0?ooo`080?ooo`<0
00009`3oool00`0007l0oooo0?ooo`0/0?ooo`800000103oool00`0007l0oooo0?ooo`0b0?ooo`03
0000O`3oool0oooo0300oooo0`00002i0?ooo`00Y`3oool00`3ooml0W4P000000004000000030000
M02oool0oooo00d0oooo00@0oooO09a80000B9`0gooo0P3oool00`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo0?ooo`040?ooo`80
00001P3oool400000240oooo00<0001o0?ooo`3oool0103oool4000002X0oooo00<0001o0?ooo`3o
ool0:P3oool2000000H0oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0b0?ooo`03
0000003oool0oooo0;L0oooo002X0?ooo`030=nLW02oool0oooo00<0oooo00<0gibL0;ooo`3oool0
2@3oool00`2oM0000000000000050000000407Boo`3ookl0M00004RLg`<0oooo00<0gia80000M02o
ool01`3oool2000003<0oooo00<0001o0?ooo`3oool01P3oool2000000P0oooo0`00000N0?ooo`03
0000O`3oool0oooo00@00000;P3oool00`0007l0oooo0?ooo`0W0?ooo`<00000203oool00`0007l0
oooo0?ooo`0b0?ooo`030000O`3oool0oooo0300oooo0`00002i0?ooo`00Y`3oool00`3o_g@0B9cO
0?ooo`0@0?ooo`0<0?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo
0`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3oool0
203oool2000000T0oooo1P00000E0?ooo`H00000<P3oool00`0007l0oooo0?ooo`0U0?ooo`800000
2`3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3oool0
]`3oool00:L0oooo00<0_g@0001d_`3oool0103oool00`3OW9`0_ooo0?ooo`0:0?ooo`0;0=nLB01d
_ol0oooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`07
0?ooo`<00000?`0007l3000000`0001o5@00001M0000O`<00000M@0007l300000;T0oooo002W0?oo
o`030=nLB0000000000000@0000000<0001d0;ooo`3oool02P3oool02`3ooml0W4Qd0;ooo`3ooml0
W4P00018W03Oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000
oooo0?ooo`0b0?ooo`030000O`3oool0oooo00d0oooo0`00000Q0?ooo`030000O`3oool0oooo0380
oooo00<0001o0?ooo`3oool0803oool200000100oooo00<0001o0?ooo`3oool0<P3oool00`0007l0
oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0oooo002W0?ooo`030?oog`2LB000000000@00000
00<0001d0;ooo`3oool02`3oool02P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d
_ol30?ooo`030?noM00004P0W=oo00L0oooo0P00000c0?ooo`030000O`3oool0oooo0100oooo0`00
000N0?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool07P3oool200000180oooo00<0
001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0`0?ooo`<00000^@3oool00:h0oooo00<0gibL
0;ooo`3oool0303oool02@3OW4P0M;oO09a80000B9`0gooo0?ooo`3ookl0M00004RLg`030?ooo`03
0=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0C0?ooo`<0
00006`3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo01/0oooo0`00000D0?ooo`03
0000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3oool00`000000oooo0?ooo`2g0?ooo`00
_@3oool01@3ooml0W4Q804Q80000B9`0gooo0080oooo00<0gia80018W03Oool00P3oool00`2oM000
07Bo0?ooo`070?ooo`800000<`3oool00`0007l0oooo0?ooo`0F0?ooo`8000006@3oool00`0007l0
oooo0?ooo`0b0?ooo`030000O`3oool0oooo01T0oooo0P00000G0?ooo`030000O`3oool0oooo0380
oooo00<0001o0?ooo`3oool0<03oool300000;T0oooo002Q0?ooo`030?oo_`1dB9`0gooo00d0oooo
00<0omnL04Q8W03Oool01P3oool00`3ooml0W4P0000000020000000302000000B9`0gooo0080oooo
00L0oono07@0000P000000000000=@0007@0_ooo00P0oooo00<000000?ooo`3oool0<P3oool00`00
07l0oooo0?ooo`0H0?ooo`<000005P3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo
01H0oooo0`00000I0?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3oool00`00
0000oooo0?ooo`2g0?ooo`00XP3oool0103ogi`0B0000018W03Oool90?ooo`040?oo_`1d000004RL
0=ooo`L0oooo00<0oooO09a8M02oool00P3oool00`2WB00003JL0=ooo`020?ooo`070?oo_`1d0000
=@000000000003D0001d0;ooo`080?ooo`800000<`3oool00`0007l0oooo0?ooo`0K0?ooo`@00000
4P3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0180oooo1000000L0?ooo`030000
O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<03oool300000;T0oooo002S0?ooo`030?oog`2L
B000000000X0000000<007Bo0?ooo`3oool0303oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0b0?ooo`030000O`3o
ool0oooo01l0oooo1000000>0?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool03P3o
ool400000200oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0b0?ooo`030000003o
ool0oooo0;L0oooo002U0?ooo`030?oog`2LB000000000H0000000<007Bo0?ooo`3oool03`3oool0
1P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool01`3oool2000003<0oooo
00<0001o0?ooo`3oool08`3oool4000000X0oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo
0?ooo`0:0?ooo`@00000903oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0300oooo
0`00002i0?ooo`00_`3oool01`3ooml0W4P004RLg`3oool0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`070?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3oool09`3oool4000000H0
oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`050?ooo`D00000:03oool00`0007l0
oooo0?ooo`0b0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3oool0]`3oool00;l0oooo
00L0gia80000001d_ol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool30000
0600001o1000000k0000O`@00000U@0007l300000;T0oooo002k0?ooo`0;0=oOg`3oool0oono07@0
00000000001d0;ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000
003oool0oooo0380oooo00<0001o0?ooo`3oool0;`3oool400000300oooo1`00000a0?ooo`030000
O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3oool00`000000oooo0?ooo`2g0?ooo`00^`3o
ool01P3OW4P00000000000000000001809cOo`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`070?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3oool0<@3oool00`0007l0
oooo0000000800000240oooo1P0000030?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3o
ool0<P3oool00`0007l0oooo0?ooo`0b0?ooo`030000003oool0oooo0;L0oooo002k0?ooo`040?oo
_`1d0000000004RLg`@0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`070?ooo`80
0000<`3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo00P0oooo8@0000090?ooo`03
0000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0`0?ooo`<0
0000^@3oool00;`0oooo00<0_g@007Boo`3oool01@3oool00`3OW4P004RL0=ooo`020?ooo`030;md
0000M;l0oooo00L0oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0a0?ooo`030000
O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0b0?ooo`030000
O`3oool0oooo0380oooo00<000000?ooo`3oool0]`3oool00;`0oooo00H0omnL04P0000000000000
0000B02Lgol20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`800000<`3oool0
0`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3oool0
0`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0300oooo0`00002i0?ooo`00_03oool00`3o
gi`0=@00000000020000000904RLg`3oool0oooo0?ooo`3ogg@080000000000003D0Sooo00T0oooo
00<000000?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0a0?ooo`030000O`3oool0oooo0380oooo
00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0380oooo
00<000000?ooo`3oool0]`3oool00;/0oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000M;oo0080
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo0P00000c0?ooo`030000O`3oool0
oooo0340oooo00<0001o0?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0
oooo0380oooo00<0001o0?ooo`3oool0<03oool300000;T0oooo002j0?ooo`030?oog`2LB00007Bo
00<0oooo00L0okmd0018W03Oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3o
ool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo0340oooo00<0001o0?ooo`3oool0<P3o
ool00`0007l0oooo0?ooo`0b0?ooo`030000O`3oool0oooo0380oooo00<0001o0?ooo`3oool0<P3o
ool00`000000oooo0?ooo`2g0?ooo`00^P3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0
W=oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`070?ooo`800000<`3oool00`000000oooo
0?ooo`0a0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0<P3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo0300oooo0`00002i0?ooo`00^P3oool00`3ogi`0B00007Bo
o`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`070?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo0;L0oooo002j0?ooo`030?oOW0180000M;oo00<0oooo00H0_g@00000B02Lgol0
omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00L0ooooo`00000o00000;T0oooo002j0?ooo`0<
0?oo_`1d0000000004RLg`3oool0okmd00000000M;l0oooo0?oOW0180000M;oo0`3oool00`3o_g@0
001809cOo`3o0?ooool0oooo002k0?ooo`030;md0000B9`0_ia80080000000H004RL0=ooo`3oool0
omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0?l0ooooo`3oool00;/0oooo00<0omnL04P0001d
_ol01@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0?l0ooooo`3oool00;`0oooo
00<0gia80000B02Lgol01@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0?l0oooo
o`3oool00;d0oooo00<0okmd00000000M;l0103oool01`3ookl0M00007Boo`3oool0okmd0000M02o
ool0o`3ooooo0?ooo`40oooo002n0?ooo`050?oog`2LB00000000000M02oool00P3oool01@3ookl0
M0000000000007@0_ooo0?l0ooooo`3oool20?ooo`00o`3ooonH0?ooo`060?oog`2LB00000000000
000007@0_ooo4@3oool00`3ooml0W4P0000000030000000304RLg`3oool0oooo0?l0oooo503oool0
0?l0oooo=03oool0102oM00000000000B02LgomR0?ooo`030;md0000B9`0gooo0180oooo00<0_g@0
0018W03Oool00`3oool00`3OW9`0_ooo0?ooo`3o0?oooa@0oooo003o0?oooc40oooo00D0oooO09a8
00000000001d0;ooo`1T0?ooo`030;md0000B9`0gooo0180oooo00<0oooO09adW03Oool00`3oool0
0`3ookl0W9cO0?ooo`3o0?oooa@0oooo003o0?oooc00oooo00@0gia8000000000000B9cOIP3oool0
0`2oM00004RL0=ooo`0C0?ooo`030;md00000000000000<0000000<0B9cO0?ooo`3oool0o`3ooolC
0?ooo`00o`3oool^0?ooo`030?oOW018000000000080000000<0001809cOo`3oool00`3oool02P3o
okl0M000000000000000B9bo09a80000B9`0gooo0?oog`2LB002000000070018W02LM0000000001d
_`3oool0oooO09a80002000000050000B018B00004RL0=ooo`2oM0000P0000001@0004P0W=oo0?oo
o`3ooml0W4P00080000000X004RL09ad0000000007Bo0?ooo`3ogi`0B00000000018W=l0gia80P00
0000100007@0_ooo0?ooo`2oM002000000030000B02Lgol0oooo00@0oooo00<0oooO09a800000000
1P0000000`00M;l0oooo0?ooo`020?ooo`090?oog`2LB000000000000000B9`0W7@000000000M;l0
gia80080000000<0001809cOg`2LB0000P0000000`0007@0_ooO09a8000200000005001d_`3oool0
_g@00018B01800000P0000001018W=l0oooo0?ooo`3OW4P2000000070000M01dM000000000000000
07@0_ooO09a80002000000030018W03Ookl0M00000@0000000<0001d0;ooo`3oool0o`3ooolD0?oo
o`00o`3oool]0?ooo`030=nLB000000007Bo00<0oooo00<0gia800000018W=l00`3oool01@3OW4P0
001d0;ooo`3oool0gia80080000001H0001d0;noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0
_ooO09a800000000M;oo0?ooo`3ookl0M0000000000007@0_ooo0=nLB00007@0_ooo0P3oool06@3o
_g@0000007Boo`3oool0omnL04P00000M;l0okno0;ooo`3OW4P0001d0;ooo`3O_kl0gooo0=nLB000
07@0_ooo0?ooo`2oM00000000018W03Oool0oooO09a8M02oool01P3oool00`3ogi`0B00004RLg`03
0?ooo`040?noM0000000000004RLg`80oooo00`0okmd0000001d_ol0oooo0?oOW018000007Bo0?no
_`2oool0gia80000M02oool20?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`07
0;md00000000M;oo0?ooo`3ookl0M00004RLg`020?ooo`0;0=nLB00007@0_ooo0?ooo`3OW4P004RL
0=ooo`3oool0oooO09a80000M;l00P3oool00`2oM4P0W=oo0?ooo`3o0?oooaT0oooo003o0?ooob`0
oooo00<0okmd00000000M;l01@3oool00`3OW4P0000007Boo`020?ooo`0C0;md0000B9`0gooo0?oo
o`3ooml0W4P0001d_`3oool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0?ooo`3ogi`0B00004RL
g`020?ooo`030?oog`2LB00007Bo0080oooo00<0gia80000M02oool00P3oool01`3o_g@0001809cO
o`3oool0oooO09a80000M;l00P3oool00`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0_ooo0?oO
W018000007Bo0`3oool00`3ogi`0W;oo0?ooo`050?ooo`030?oo_`1d0000B9cO00@0oooo00`0oono
07@000000000M;oo0?ooo`3o_g@0001809cOo`3oool0oooO09a80000M;l20?ooo`030=nLB00007@0
_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo00<0_g@00018W03Oool0
0P3oool03`3o_g@0001d0;ooo`3oool0gia80000M02oool0oooo0=nLB000B9`0gooo0?ooo`3ooml0
W4P0001d_`020?ooo`050?oo_`1d0000000000000000M;l0o`3ooolG0?ooo`00o`3oool[0?ooo`04
0?oog`2LB000001809cOo`D0oooo00`0oono07@00000B9`0gooo0?ooo`2oM00004RL0=ooo`3oool0
oooO09a80000M;l20?ooo`0:0?noM000000007Bo0?oog`2LB00007Bo0?ooo`3o_g@0000007Boo`80
oooo00<0oooO09a80000M;l00P3oool00`3OW4P0001d0;ooo`030?ooo`060?noM000000007Bo0?oo
g`2LB00007Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0_ooo0?noM00004P0W=oo
2`3oool00`3ookl0M00004RLg`050?ooo`0;0?oOW018000004RL0=ooo`3oool0okmd00000000M;l0
oooO09a80000M;l00P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0
W4P0001d_`020?ooo`030;md0000B9`0gooo0080oooo00l0omnL04P0001d_ol0oooo0=nLB00007@0
_ooo0?ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool01P3OW4P0001d0;ooo`3ookl0
M00004RLgol0oooo5P3oool00?l0oooo:`3oool0103ogi`0B0000018W03Oool50?ooo`0<0?oog`2L
B000001809cOo`3oool0_g@00018W03Oool0oooo0?oog`2LB00007Bo103oool0203o_g@000000000
0000M;l0oooo0?noM0000000M;oo0P3oool00`3ooml0W4P0001d_`020?ooo`030=nLB00007@0_ooo
00D0oooo00@0okmd00000000000007Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0
_ooo0?noM00007@0_ooo2`3oool00`3ookl0M00004RLg`050?ooo`040?oog`2LB000001809cOo`<0
oooo00@0okmd00000000000007Bo0P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo
0?ooo`3ooml0W4P0001d_`020?ooo`030;md0000B9`0gooo0080oooo01L0omnL04P0001d_ol0oooo
0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO
0?ooo`3ooml0W4P0001d_`3o0?oooaH0oooo003o0?ooob/0oooo00<0okmd00000000M;l01`3oool0
5`2oM000000007Boo`3oool0_g@00018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ookl0M00007Bo
o`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0080oooo00<0oooO09a80000M;l00P3oool0
0`3OW4P0001809cOo`020?ooo`070?oo_`1d0000M;oo0?ooo`3ooml0W4P0001d_`020?ooo`030=nL
B000B9`0gooo0080oooo00H0gia80000M02oool0omnL04P0B02Lgol20?ooo`030?noM00004P0W=oo
00H0oooo00<0oono07@00018W=l01P3oool02P2oM000000007Boo`3ookl0M00007Boo`3oool0oooO
09a80000M;l20?ooo`030=nLB00004P0W=oo0080oooo00L0_g@00000M02oool0oooo0?oog`2LB000
07Bo0080oooo00<0_g@00000M02oool00P3oool05`3OW4P0001809cOo`3oool0gia80000M02oool0
oooo0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0oono07@00018W=l0oooo0?oog`2LB000
B9cO0?l0oooo5P3oool00?l0oooo:`3oool00`3o_g@00000001d_`070?ooo`0C0;md00000000M;oo
0?ooo`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0?ooo`3ookl0M000
001d_`020?ooo`0E0?noM00004P0W=oo0?ooo`3o_g@00000001d_`3ooml0W4P00000000004P0B7BL
0=noM00007@0_ono07@00018W=l0oooo0?oo_`1d000007Bo0080oooo0100gia80018W03Oool0oooO
09a800000000001d0;ooo`3oool0_g@00018W03Oool0oooO09a8000007@0_ooo1P3oool00`3ookl0
M00004RLg`060?ooo`0^0;md00000000M;oo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`3ooml0
W4P00000000004P0B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO0?oo_`1d0000
00000018B018B9`0gooo0?oOW018000004RL0=oog`2LB00000000000M02ooml0W4P000000000B4P0
B01809cOo`3ookl0M00004RLg`80oooo00@0_g@0001d_`3oool0okmd0P0000000`00B9`0gooo0?oo
o`3o0?oooa<0oooo003o0?ooob/0oooo00@0omnL04P00000B9`0gooo1@3oool0303ooml0W4P00000
B02Lgkl0M00000000000B9`0gooo0?oOW01800000000001d_`80oooo00H0oono07@0000000000000
0000M02oool30?ooo`030?noM000000000000080000000/007Bo0?ooo`3oool0_g@00000M02ogi`0
B0000000M02oool0oono07@00002000000040000M02oool0oooo0;md00<0000000D004RL0=ooo`2o
M000001d0;ooo`020?ooo`050=nLB000000000000000B02Lgol01`3oool00`3ookl0M00004RLg`05
0?ooo`070?oog`2LB000001809cOo`3oool0oono07@000020000000I0000M02oool0oooo0?ooo`2o
M000001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`3oool0oooO09a80000
B9`0gmnL04P000000000001d0;ooo`020?ooo`0;0;md000007@0_ooo0?ooo`2oM00004RL0=oog`2L
B00000000000M02oool00P3oool01@3ooml0W4P0000000000000M;oo0?l0oooo5`3oool00?l0oooo
:`3oool0103ooml0W4P00000M02oool50?ooo`040?oOW018000004RL0=oooaH0oooo00<0oooO09a8
0000M;l0403oool00`2oM00004RL0=ooo`0C0?ooo`030?oo_`1d0000B9cO00D0oooo00@0omnL04P0
0000B9`0goooo`3ooom90?ooo`00o`3oool/0?ooo`030?oOW0180000B9cO00D0oooo00<0_g@00000
001d_ol05`3oool00`3ooml0W4P0001d_`0@0?ooo`030?oo_`1dB9`0gooo01<0oooo00<0oono07@0
0018W=l0103oool0103ooml0W4P00000001d_ooo0?ooodX0oooo003o0?ooobd0oooo00<0okmd0000
0018W=l00`3oool00`2oM000000007Boo`0H0?ooo`030?oog`2LB00007Bo0100oooo00<0oooO09ad
W03Oool04`3oool00`3ogi`0B00004RLg`030?ooo`040?noM0000000000004RLgol0ooooB`3oool0
0?l0oooo;P3oool00`3ooml0W4P000000002000000030018W03Oool0oooo01L0oooo00@0oono07@0
0000000007Bo5@3oool00`3OW4P0001809cOo`0<0?ooo`030?oog`2LB000000000H0000000<0001d
0;ooo`3oool08P3oool00`3OW4P0001809cOo`3o0?ooobH0oooo003o0?ooodh0oooo00<0oooO09a8
0000M;l05@3oool00`3OW4P0001809cOo`0j0?ooo`030=nLB00004P0W=oo0?l0oooo9P3oool00?l0
ooooo`3oooo<0?ooo`00o`3ooooo0?oool`0oooo003o0?ooool0ooooc03oool00?l0ooooo`3oooo<
0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 713.375}, {141.875, 0}} -> {-200.071, -48.3168, \
1.42045, 1.42045}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell["\<\
How large would one need to build a baseball field on the moon?\
\>", "Problem"],

Cell["\<\
Write a function which evaluates the range given the initial velocity and \
physical parameters.  Plot the range as a function of initial angle for \
typical baseball speeds.\
\>", "Problem"],

Cell["\<\
Evaluate the dependence of home run hitting on altitude assuming the dominant \
effect is due to air density.  Compare Denver with a sea-level city.  \
Similarly, evaluate the dependence on temperature.\
\>", "Problem"]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, -1}},
ShowCellLabel->False,
ShowCellTags->False,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"]},
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Notation Package Styles", "Section",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell["\<\
The cells below define certain styles needed by the Notation package. These \
styles serve to make visible otherwise invisible tagboxes.\
\>", "Text",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationTemplateStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 1, 0.850004],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationPatternWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 0.900008, 0.979995],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationMadeBoxesWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[0.900008, 0.889998, 1],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"]
  }, Closed]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 121, 4, 85, "Title"],
Cell[1900, 59, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2380, 78, 31, 0, 83, "Section"],
Cell[2414, 80, 553, 12, 94, "Text"],
Cell[2970, 94, 1926, 44, 202, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[4933, 143, 33, 0, 51, "Section"],
Cell[4969, 145, 221, 4, 125, "Input"],
Cell[5193, 151, 189, 3, 125, "Input"],
Cell[5385, 156, 75, 1, 73, "Input"],
Cell[5463, 159, 919, 29, 126, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[6419, 193, 38, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[6482, 197, 38, 0, 68, "Subsection"],
Cell[6523, 199, 616, 18, 68, "Text"],
Cell[7142, 219, 274, 5, 100, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[7453, 229, 50, 0, 68, "Subsection"],
Cell[7506, 231, 222, 4, 67, "Text"],
Cell[7731, 237, 1623, 39, 85, "Input"],
Cell[9357, 278, 382, 8, 68, "Text"],
Cell[9742, 288, 681, 16, 73, "Input"],
Cell[10426, 306, 203, 4, 41, "Text"],
Cell[10632, 312, 227, 4, 73, "Input"],
Cell[10862, 318, 195, 4, 41, "Text"],
Cell[11060, 324, 289, 5, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[11386, 334, 41, 0, 68, "Subsection"],
Cell[11430, 336, 124, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[11579, 343, 300, 5, 92, "Input"],
Cell[11882, 350, 176, 4, 59, "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[12119, 361, 38, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[12182, 365, 66, 0, 68, "Subsection"],
Cell[12251, 367, 89, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[12365, 373, 65, 1, 73, "Input"],
Cell[12433, 376, 48, 1, 59, "Output"]
}, Closed]],
Cell[12496, 380, 148, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[12669, 387, 77, 1, 73, "Input"],
Cell[12749, 390, 61, 1, 75, "Output"]
}, Closed]],
Cell[12825, 394, 405, 8, 68, "Text"],
Cell[13233, 404, 627, 11, 151, "Input"],
Cell[13863, 417, 243, 4, 67, "Text"],

Cell[CellGroupData[{
Cell[14131, 425, 121, 2, 73, "Input"],
Cell[14255, 429, 67042, 1656, 185, 21154, 1085, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[81346, 2091, 54, 0, 68, "Subsection"],
Cell[81403, 2093, 549, 8, 93, "Text"],
Cell[81955, 2103, 643, 12, 177, "Input"],

Cell[CellGroupData[{
Cell[82623, 2119, 119, 2, 73, "Input"],
Cell[82745, 2123, 150050, 2952, 444, 28566, 1448, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[232810, 5078, 767, 17, 119, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[233614, 5100, 49, 0, 68, "Subsection"],
Cell[233666, 5102, 293, 5, 67, "Text"],
Cell[233962, 5109, 555, 10, 151, "Input"],

Cell[CellGroupData[{
Cell[234542, 5123, 163, 3, 73, "Input"],
Cell[234708, 5128, 103125, 2290, 231, 26200, 1336, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[337848, 7421, 209, 3, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[338106, 7430, 71, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[338202, 7434, 94, 2, 68, "Subsection"],
Cell[338299, 7438, 228, 4, 99, "Input"],
Cell[338530, 7444, 554, 10, 151, "Input"],

Cell[CellGroupData[{
Cell[339109, 7458, 165, 3, 73, "Input"],
Cell[339277, 7463, 69214, 1930, 160, 26750, 1401, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[408506, 9396, 248, 6, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[408791, 9407, 94, 2, 68, "Subsection"],

Cell[CellGroupData[{
Cell[408910, 9413, 170, 3, 73, "Input"],
Cell[409083, 9418, 68445, 1879, 155, 25994, 1350, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[477589, 11304, 33, 0, 54, "ExerciseMain"],
Cell[477625, 11306, 90, 2, 57, "Problem"],
Cell[477718, 11310, 201, 4, 57, "Problem"],
Cell[477922, 11316, 229, 4, 81, "Problem"]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

