(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    918934,      47229]*)
(*NotebookOutlinePosition[    955829,      48534]*)
(*  CellTagsIndexPosition[    955785,      48530]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Koch Fractals", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "Koch curves are particularly simple fractals which enclose a finite area \
within a continuous, nonintersecting curve of infinite length.  A Koch curve \
is generated by the following self-similarity transformation.  Begin with a \
regular polygon composed of sides of equal length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0\)]],
  ".  Centered on each segment of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0\)]],
  " build an equilateral triangle with sides of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_1 = \[ScriptL]\_0/3\)]],
  " using the middle third of the original segment as a base.  Thus, after \
one iteration a simple line segment becomes"
}], "Text"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28868 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00687322 0.952381 [
[ 0 0 0 0 ]
[ 1 .28868 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28868 L
0 .28868 L
closepath
clip
newpath
0 g
.5 Mabswid
.02381 .00687 m
.34127 .00687 L
s
.34127 .00687 m
.5 .2818 L
s
.5 .2818 m
.65873 .00687 L
s
.65873 .00687 m
.97619 .00687 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 83.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000O2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002`3ooon9000008L0ooooR@00000<0?ooo`00U03o
ool00`000000oooo0?ooo`230?ooo`030000003oool0oooo09<0oooo002D0?ooo`030000003oool0
oooo08<0oooo00<000000?ooo`3oool0T`3oool009D0oooo00<000000?ooo`3oool0P@3oool00`00
0000oooo0?ooo`2D0?ooo`00U@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo09@0
oooo002F0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0U@3oool009H0oooo00<0
00000?ooo`3oool0O`3oool00`000000oooo0?ooo`2E0?ooo`00U`3oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo09H0oooo002H0?ooo`030000003oool0oooo07/0oooo00<000000?oo
o`3oool0U`3oool009P0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`2G0?ooo`00
V@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo09P0oooo002I0?ooo`030000003o
ool0oooo07T0oooo00<000000?ooo`3oool0V03oool009X0oooo00<000000?ooo`3oool0M`3oool0
0`000000oooo0?ooo`2I0?ooo`00VP3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo
09T0oooo002K0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0VP3oool009`0oooo
00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`2K0?ooo`00W03oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo09/0oooo002M0?ooo`030000003oool0oooo0740oooo00<00000
0?ooo`3oool0W03oool009d0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`2L0?oo
o`00WP3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo09d0oooo002O0?ooo`030000
003oool0oooo06d0oooo00<000000?ooo`3oool0WP3oool009l0oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`2N0?ooo`00X03oool00`000000oooo0?ooo`1[0?ooo`030000003oool0
oooo09l0oooo002P0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0W`3oool00:40
oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`2P0?ooo`00X@3oool00`000000oooo
0?ooo`1Y0?ooo`030000003oool0oooo0:00oooo002R0?ooo`030000003oool0oooo06L0oooo00<0
00000?ooo`3oool0X@3oool00:<0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`2R
0?ooo`00X`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0:80oooo002T0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool0X`3oool00:@0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`2S0?ooo`00Y@3oool00`000000oooo0?ooo`1Q0?ooo`030000003o
ool0oooo0:@0oooo002U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0Y03oool0
0:H0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`2U0?ooo`00Y`3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo0:H0oooo002W0?ooo`030000003oool0oooo05d0oooo
00<000000?ooo`3oool0YP3oool00:P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?oo
o`2W0?ooo`00Z03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0:L0oooo002Y0?oo
o`030000003oool0oooo05T0oooo00<000000?ooo`3oool0Z03oool00:T0oooo00<000000?ooo`3o
ool0F@3oool00`000000oooo0?ooo`2X0?ooo`00ZP3oool00`000000oooo0?ooo`1G0?ooo`030000
003oool0oooo0:T0oooo002[0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0ZP3o
ool00:/0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`2Z0?ooo`00[03oool00`00
0000oooo0?ooo`1C0?ooo`030000003oool0oooo0:/0oooo002/0?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool0Z`3oool00:d0oooo00<000000?ooo`3oool0D@3oool00`000000oooo
0?ooo`2/0?ooo`00[P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo0:d0oooo002^
0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0[@3oool00:l0oooo00<000000?oo
o`3oool0C@3oool00`000000oooo0?ooo`2^0?ooo`00[`3oool00`000000oooo0?ooo`1=0?ooo`03
0000003oool0oooo0:h0oooo002`0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0
[`3oool00;00oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`2_0?ooo`00/@3oool0
0`000000oooo0?ooo`190?ooo`030000003oool0oooo0;00oooo002b0?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool0/@3oool00;80oooo00<000000?ooo`3oool0A`3oool00`000000
oooo0?ooo`2a0?ooo`00/`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0;80oooo
002c0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0/P3oool00;@0oooo00<00000
0?ooo`3oool0@`3oool00`000000oooo0?ooo`2c0?ooo`00]03oool00`000000oooo0?ooo`130?oo
o`030000003oool0oooo0;<0oooo002e0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3o
ool0]03oool00;H0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`2e0?ooo`00]P3o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0;D0oooo002g0?ooo`030000003oool0
oooo03d0oooo00<000000?ooo`3oool0]P3oool00;L0oooo00<000000?ooo`3oool0?@3oool00`00
0000oooo0?ooo`2f0?ooo`00^03oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0;L0
oooo002h0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0]`3oool00;T0oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0?ooo`2h0?ooo`00^P3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo0;T0oooo002j0?ooo`030000003oool0oooo03L0oooo00<000000?oo
o`3oool0^@3oool00;/0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`2j0?ooo`00
^`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0;X0oooo002l0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0^`3oool00;`0oooo00<000000?ooo`3oool0<`3oool0
0`000000oooo0?ooo`2k0?ooo`00_@3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo
0;`0oooo002n0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0_@3oool00;h0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`2m0?ooo`00_`3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo0;h0oooo002o0?ooo`030000003oool0oooo02d0oooo00<00000
0?ooo`3oool0_P3oool00<00oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`2o0?oo
o`00`@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo00310?ooo`030000
003oool0oooo02T0oooo00<000000?ooo`3oool0`03oool00<80oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`310?ooo`00`P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0
oooo0<40oooo00330?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0`P3oool00<<0
oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`320?ooo`00a03oool00`000000oooo
0?ooo`0S0?ooo`030000003oool0oooo0<<0oooo00350?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool0a03oool00<D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`34
0?ooo`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00360?ooo`03
0000003oool0oooo01l0oooo00<000000?ooo`3oool0a@3oool00<L0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`360?ooo`00a`3oool00`000000oooo0?ooo`0M0?ooo`030000003o
ool0oooo0<H0oooo00380?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0a`3oool0
0<T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`380?ooo`00b@3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo0<P0oooo003:0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0b@3oool00<X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`390?ooo`00b`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0<X0oooo003;0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool0bP3oool00<`0oooo00<000000?ooo`3o
ool04`3oool00`000000oooo0?ooo`3;0?ooo`00c@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0<`0oooo003=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0c03o
ool00<h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3=0?ooo`00cP3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo0<d0oooo003?0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool0cP3oool00=00oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`3?0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo003A
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=40oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00dP3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo0=40oooo003B0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
d@3oool00=<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3B0?ooo`00e03oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003D0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?oo
o`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003F0?ooo`030000003oool000000=L0
oooo003F0?ooo`030000003oool000000=L0oooo003G0?ooo`030000003oool0oooo0=H0oooo003o
0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {123.688, 0}} -> {-0.028119, -0.0072184, \
0.003676, 0.003676}}],

Cell[TextData[{
  "Then build triangles with sides of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_2 = \[ScriptL]\_1/3\)]],
  " upon the middle third of each segment of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_1\)]],
  ", such that the preceding figure becomes"
}], "Text"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28868 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00687322 0.952381 [
[ 0 0 0 0 ]
[ 1 .28868 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28868 L
0 .28868 L
closepath
clip
newpath
0 g
.5 Mabswid
.02381 .00687 m
.12963 .00687 L
s
.12963 .00687 m
.18254 .09852 L
s
.18254 .09852 m
.23545 .00687 L
s
.23545 .00687 m
.34127 .00687 L
s
.34127 .00687 m
.39418 .09852 L
s
.39418 .09852 m
.34127 .19016 L
s
.34127 .19016 m
.44709 .19016 L
s
.44709 .19016 m
.5 .2818 L
s
.5 .2818 m
.55291 .19016 L
s
.55291 .19016 m
.65873 .19016 L
s
.65873 .19016 m
.60582 .09852 L
s
.60582 .09852 m
.65873 .00687 L
s
.65873 .00687 m
.76455 .00687 L
s
.76455 .00687 m
.81746 .09852 L
s
.81746 .09852 m
.87037 .00687 L
s
.87037 .00687 m
.97619 .00687 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 83.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000O2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002`3oool^000002d0oooo;P0000270?ooobh00000
;@3oool^000000`0oooo000i0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0;03o
ool00`000000oooo0?ooo`230?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0:@3o
ool00`000000oooo0?ooo`0h0?ooo`00>@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0
oooo02`0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0
oooo02T0oooo00<000000?ooo`3oool0>03oool003X0oooo00<000000?ooo`3oool09`3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo03T0oooo000j0?ooo`030000003oool0oooo02L0
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo02h0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0i0?ooo`00>`3oool00`000000oooo
0?ooo`0U0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0O`3oool00`000000oooo
0?ooo`0`0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0>P3oool003`0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo07d0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo03/0oooo000l
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0k
0?ooo`00?@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0N`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool0?03oool003d0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo07/0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo03`0oooo000n0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
=P3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0m0?ooo`00?P3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo03H0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0f0?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool0?@3oool003l0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`0M0?ooo`030000003oool0oooo03h0oooo00100?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo
03X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0o0?ooo`00@03oool00`000000
oooo0?ooo`0K0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0M@3oool00`000000
oooo0?ooo`0j0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0?`3oool00440oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo07<0oooo
00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0400oooo
00110?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`100?ooo`00@P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo03h0oooo00<00000
0?ooo`3oool0L@3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0@@3oool004<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`100?oo
o`030000003oool0oooo06l0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo0480oooo00130?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`120?ooo`00A03oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo0480oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`120?ooo`030000
003oool0oooo01<0oooo00<000000?ooo`3oool0@`3oool004@0oooo00<000000?ooo`3oool04`3o
ool00`000000oooo0?ooo`120?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0@P3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo04<0oooo00150?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`1[0?ooo`030000003oool0
oooo04@0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`140?ooo`00AP3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0J@3oool00`00
0000oooo0?ooo`160?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0A@3oool004H0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo06T0
oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo04D0
oooo00170?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0B03oool00`000000oooo
0?ooo`1W0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`160?ooo`00A`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04P0oooo00<0
00000?ooo`3oool0I`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0AP3oool004P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo04L0oooo00190?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool0C03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04`0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`180?ooo`00B@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo04`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1<0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0B03oool004X0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0
CP3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04T0oooo001:0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1Q0?ooo`030000003o
ool0oooo04h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`190?ooo`00B`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0G`3oool0
0`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0BP3oool0
04/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
05l0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
04X0oooo001<0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`1;0?ooo`00C@3oool01@000000oooo0?ooo`3oool0000005H0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`1D0?ooo`050000003oool0oooo0?ooo`000000CP3oool004d0
oooo00D000000?ooo`3oool0oooo0000001F0?ooo`030000003oool0oooo05/0oooo00<000000?oo
o`3oool0E03oool01@000000oooo0?ooo`3oool0000004h0oooo001>0?ooo`030000003oool00000
05P0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`1F0?ooo`030000003oool00000
04l0oooo001>0?ooo`030000003oool0000005P0oooo00<000000?ooo`3oool0F@3oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0000004l0oooo001?0?ooo`030000003oool0oooo05P0oooo
00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo04h0oooo
002Y0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0Z03oool00:T0oooo00<00000
0?ooo`3oool0F@3oool00`000000oooo0?ooo`2X0?ooo`00Z03oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo0:L0oooo002X0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3o
ool0Y`3oool00:L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`2V0?ooo`00YP3o
ool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0:D0oooo002V0?ooo`030000003oool0
oooo05l0oooo00<000000?ooo`3oool0Y@3oool00:D0oooo00<000000?ooo`3oool0H@3oool00`00
0000oooo0?ooo`2T0?ooo`00Y@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo0:@0
oooo002T0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0X`3oool00:@0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`2S0?ooo`00X`3oool00`000000oooo0?ooo`1U
0?ooo`030000003oool0oooo0:80oooo002R0?ooo`030000003oool0oooo06L0oooo00<000000?oo
o`3oool0X@3oool00:80oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`2Q0?ooo`00
X@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo0:00oooo002Q0?ooo`030000003o
ool0oooo06T0oooo00<000000?ooo`3oool0X03oool00:00oooo00<000000?ooo`3oool0J`3oool0
0`000000oooo0?ooo`2O0?ooo`00W`3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo
09h0oooo002O0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0WP3oool009h0oooo
00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`2M0?ooo`00WP3oool00`000000oooo0?oo
o`1_0?ooo`030000003oool0oooo09d0oooo002M0?ooo`030000003oool0oooo0740oooo00<00000
0?ooo`3oool0W03oool009`0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`2K0?oo
o`00W03oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo09/0oooo002K0?ooo`030000
003oool0oooo07D0oooo00<000000?ooo`3oool0VP3oool009/0oooo00<000000?ooo`3oool0M@3o
ool00`000000oooo0?ooo`2J0?ooo`00VP3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0
oooo09T0oooo002I0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0V03oool009T0
oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`2H0?ooo`00V03oool00`000000oooo
0?ooo`1k0?ooo`030000003oool0oooo09L0oooo002H0?ooo`030000003oool0oooo07/0oooo00<0
00000?ooo`3oool0U`3oool009L0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`2F
0?ooo`00U`3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo09H0oooo002F0?ooo`03
0000003oool0oooo07l0oooo00<000000?ooo`3oool0U@3oool009D0oooo00<000000?ooo`3oool0
P@3oool00`000000oooo0?ooo`2D0?ooo`00U@3oool00`000000oooo0?ooo`210?ooo`030000003o
ool0oooo09@0oooo002D0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0T`3oool0
09@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`2C0?ooo`00T`3oool^000002d0
oooo;P00002D0?ooo`00`@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo
00310?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0`03oool00<80oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`310?ooo`00`P3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo0<40oooo00330?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3o
ool0`P3oool00<<0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`320?ooo`00a03o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0<<0oooo00350?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0a03oool00<D0oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`340?ooo`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0
oooo00360?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0a@3oool00<L0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`360?ooo`00a`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo0<H0oooo00380?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool0a`3oool00<T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`380?ooo`00
b@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0<P0oooo003:0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0b@3oool00<X0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`390?ooo`00b`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
0<X0oooo003;0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0b`3oool00<`0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`3;0?ooo`00c@3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0<`0oooo003=0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool0c03oool00<h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3=0?oo
o`00cP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0<d0oooo003?0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0cP3oool00=00oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`3?0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo0<l0oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=40
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00dP3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo0=40oooo003B0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0d@3oool00=<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3B
0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003D0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0
oooo0000003F0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003F0?ooo`030000
003oool000000=L0oooo003F0?ooo`030000003oool000000=L0oooo003G0?ooo`030000003oool0
oooo0=H0oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {123.688, 0}} -> {-0.028119, -0.0072184, \
0.003676, 0.003676}}],

Cell["\<\
This transformation is repeated indefinitely and the Koch curve is the border \
of the resulting figure.\
\>", "Text"],

Cell[TextData[{
  "For example, the Koch snowflake based upon an equilateral triangle is \
shown below, where here the number of iterations is ",
  Cell[BoxData[
      \(TraditionalForm\`n = 4\)]],
  ".  The figure on the left shows the Koch curve, which is only the outer \
border, while the figure on the right shows the intermediate steps to \
illustrate the construction algorithm.  In addition, we also show the circle \
that circumscribes the original triangle to illustrate the finiteness of the \
area.  However, one can show that in the limit ",
  Cell[BoxData[
      \(TraditionalForm\`n \[Rule] \[Infinity]\)]],
  " the length of the Koch curve is infinite despite the fact that it \
encloses a finite area.  "
}], "Text"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .54986 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0130918 0.47619 [
[ 0 0 0 0 ]
[ 1 .54986 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .54986 L
0 .54986 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0130918 0.477324 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.302422 0.952381 [
[ 0 0 0 0 ]
[ 1 1.1547 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.1547 L
0 1.1547 L
closepath
clip
newpath
0 g
.5 Mabswid
.02381 .30242 m
.02969 .3126 L
s
.02969 .3126 m
.02381 .32279 L
s
.02381 .32279 m
.03557 .32279 L
s
.03557 .32279 m
.04145 .33297 L
s
.04145 .33297 m
.03557 .34315 L
s
.03557 .34315 m
.02381 .34315 L
s
.02381 .34315 m
.02969 .35333 L
s
.02969 .35333 m
.02381 .36352 L
s
.02381 .36352 m
.03557 .36352 L
s
.03557 .36352 m
.04145 .3737 L
s
.04145 .3737 m
.04733 .36352 L
s
.04733 .36352 m
.05908 .36352 L
s
.05908 .36352 m
.06496 .3737 L
s
.06496 .3737 m
.05908 .38388 L
s
.05908 .38388 m
.07084 .38388 L
s
.07084 .38388 m
.07672 .39406 L
s
.07672 .39406 m
.07084 .40425 L
s
.07084 .40425 m
.05908 .40425 L
s
.05908 .40425 m
.06496 .41443 L
s
.06496 .41443 m
.05908 .42461 L
s
.05908 .42461 m
.04733 .42461 L
s
.04733 .42461 m
.04145 .41443 L
s
.04145 .41443 m
.03557 .42461 L
s
.03557 .42461 m
.02381 .42461 L
s
.02381 .42461 m
.02969 .43479 L
s
.02969 .43479 m
.02381 .44498 L
s
.02381 .44498 m
.03557 .44498 L
s
.03557 .44498 m
.04145 .45516 L
s
.04145 .45516 m
.03557 .46534 L
s
.03557 .46534 m
.02381 .46534 L
s
.02381 .46534 m
.02969 .47552 L
s
.02969 .47552 m
.02381 .48571 L
s
.02381 .48571 m
.03557 .48571 L
s
.03557 .48571 m
.04145 .49589 L
s
.04145 .49589 m
.04733 .48571 L
s
.04733 .48571 m
.05908 .48571 L
s
.05908 .48571 m
.06496 .49589 L
s
.06496 .49589 m
.05908 .50607 L
s
.05908 .50607 m
.07084 .50607 L
s
.07084 .50607 m
.07672 .51626 L
s
.07672 .51626 m
.0826 .50607 L
s
.0826 .50607 m
.09436 .50607 L
s
.09436 .50607 m
.08848 .49589 L
s
.08848 .49589 m
.09436 .48571 L
s
.09436 .48571 m
.10611 .48571 L
s
.10611 .48571 m
.11199 .49589 L
s
.11199 .49589 m
.11787 .48571 L
s
.11787 .48571 m
.12963 .48571 L
s
.12963 .48571 m
.13551 .49589 L
s
.13551 .49589 m
.12963 .50607 L
s
.12963 .50607 m
.14139 .50607 L
s
.14139 .50607 m
.14727 .51626 L
s
.14727 .51626 m
.14139 .52644 L
s
.14139 .52644 m
.12963 .52644 L
s
.12963 .52644 m
.13551 .53662 L
s
.13551 .53662 m
.12963 .5468 L
s
.12963 .5468 m
.14139 .5468 L
s
.14139 .5468 m
.14727 .55699 L
s
.14727 .55699 m
.15315 .5468 L
s
.15315 .5468 m
.1649 .5468 L
s
.1649 .5468 m
.17078 .55699 L
s
.17078 .55699 m
.1649 .56717 L
s
.1649 .56717 m
.17666 .56717 L
s
.17666 .56717 m
.18254 .57735 L
s
.18254 .57735 m
.17666 .58753 L
s
.17666 .58753 m
.1649 .58753 L
s
.1649 .58753 m
.17078 .59772 L
s
.17078 .59772 m
.1649 .6079 L
s
.1649 .6079 m
.15315 .6079 L
s
.15315 .6079 m
.14727 .59772 L
s
.14727 .59772 m
.14139 .6079 L
s
.14139 .6079 m
.12963 .6079 L
s
.12963 .6079 m
.13551 .61808 L
s
.13551 .61808 m
.12963 .62826 L
s
.12963 .62826 m
.14139 .62826 L
s
.14139 .62826 m
.14727 .63845 L
s
.14727 .63845 m
.14139 .64863 L
s
.14139 .64863 m
.12963 .64863 L
s
.12963 .64863 m
.13551 .65881 L
s
.13551 .65881 m
.12963 .66899 L
s
.12963 .66899 m
.11787 .66899 L
s
.11787 .66899 m
.11199 .65881 L
s
.11199 .65881 m
.10611 .66899 L
s
.10611 .66899 m
.09436 .66899 L
s
.09436 .66899 m
.08848 .65881 L
s
.08848 .65881 m
.09436 .64863 L
s
.09436 .64863 m
.0826 .64863 L
s
.0826 .64863 m
.07672 .63845 L
s
.07672 .63845 m
.07084 .64863 L
s
.07084 .64863 m
.05908 .64863 L
s
.05908 .64863 m
.06496 .65881 L
s
.06496 .65881 m
.05908 .66899 L
s
.05908 .66899 m
.04733 .66899 L
s
.04733 .66899 m
.04145 .65881 L
s
.04145 .65881 m
.03557 .66899 L
s
.03557 .66899 m
.02381 .66899 L
s
.02381 .66899 m
.02969 .67918 L
s
.02969 .67918 m
.02381 .68936 L
s
.02381 .68936 m
.03557 .68936 L
s
.03557 .68936 m
.04145 .69954 L
s
.04145 .69954 m
.03557 .70972 L
s
.03557 .70972 m
.02381 .70972 L
s
.02381 .70972 m
.02969 .71991 L
s
.02969 .71991 m
.02381 .73009 L
s
.02381 .73009 m
.03557 .73009 L
s
.03557 .73009 m
.04145 .74027 L
s
.04145 .74027 m
.04733 .73009 L
s
.04733 .73009 m
.05908 .73009 L
s
.05908 .73009 m
.06496 .74027 L
s
.06496 .74027 m
.05908 .75045 L
s
.05908 .75045 m
.07084 .75045 L
s
.07084 .75045 m
.07672 .76064 L
s
.07672 .76064 m
.07084 .77082 L
s
.07084 .77082 m
.05908 .77082 L
s
.05908 .77082 m
.06496 .781 L
s
.06496 .781 m
.05908 .79118 L
s
.05908 .79118 m
.04733 .79118 L
s
.04733 .79118 m
.04145 .781 L
s
.04145 .781 m
.03557 .79118 L
s
.03557 .79118 m
.02381 .79118 L
s
.02381 .79118 m
.02969 .80137 L
s
.02969 .80137 m
.02381 .81155 L
s
.02381 .81155 m
.03557 .81155 L
s
.03557 .81155 m
.04145 .82173 L
s
.04145 .82173 m
.03557 .83191 L
s
.03557 .83191 m
.02381 .83191 L
s
.02381 .83191 m
.02969 .8421 L
s
.02969 .8421 m
.02381 .85228 L
s
.02381 .85228 m
.03557 .85228 L
s
.03557 .85228 m
.04145 .86246 L
s
.04145 .86246 m
.04733 .85228 L
s
.04733 .85228 m
.05908 .85228 L
s
.05908 .85228 m
.06496 .86246 L
s
.06496 .86246 m
.05908 .87264 L
s
.05908 .87264 m
.07084 .87264 L
s
.07084 .87264 m
.07672 .88283 L
s
.07672 .88283 m
.0826 .87264 L
s
.0826 .87264 m
.09436 .87264 L
s
.09436 .87264 m
.08848 .86246 L
s
.08848 .86246 m
.09436 .85228 L
s
.09436 .85228 m
.10611 .85228 L
s
.10611 .85228 m
.11199 .86246 L
s
.11199 .86246 m
.11787 .85228 L
s
.11787 .85228 m
.12963 .85228 L
s
.12963 .85228 m
.13551 .86246 L
s
.13551 .86246 m
.12963 .87264 L
s
.12963 .87264 m
.14139 .87264 L
s
.14139 .87264 m
.14727 .88283 L
s
.14727 .88283 m
.14139 .89301 L
s
.14139 .89301 m
.12963 .89301 L
s
.12963 .89301 m
.13551 .90319 L
s
.13551 .90319 m
.12963 .91337 L
s
.12963 .91337 m
.14139 .91337 L
s
.14139 .91337 m
.14727 .92356 L
s
.14727 .92356 m
.15315 .91337 L
s
.15315 .91337 m
.1649 .91337 L
s
.1649 .91337 m
.17078 .92356 L
s
.17078 .92356 m
.1649 .93374 L
s
.1649 .93374 m
.17666 .93374 L
s
.17666 .93374 m
.18254 .94392 L
s
.18254 .94392 m
.18842 .93374 L
s
.18842 .93374 m
.20018 .93374 L
s
.20018 .93374 m
.1943 .92356 L
s
.1943 .92356 m
.20018 .91337 L
s
.20018 .91337 m
.21193 .91337 L
s
.21193 .91337 m
.21781 .92356 L
s
.21781 .92356 m
.22369 .91337 L
s
.22369 .91337 m
.23545 .91337 L
s
.23545 .91337 m
.22957 .90319 L
s
.22957 .90319 m
.23545 .89301 L
s
.23545 .89301 m
.22369 .89301 L
s
.22369 .89301 m
.21781 .88283 L
s
.21781 .88283 m
.22369 .87264 L
s
.22369 .87264 m
.23545 .87264 L
s
.23545 .87264 m
.22957 .86246 L
s
.22957 .86246 m
.23545 .85228 L
s
.23545 .85228 m
.24721 .85228 L
s
.24721 .85228 m
.25309 .86246 L
s
.25309 .86246 m
.25897 .85228 L
s
.25897 .85228 m
.27072 .85228 L
s
.27072 .85228 m
.2766 .86246 L
s
.2766 .86246 m
.27072 .87264 L
s
.27072 .87264 m
.28248 .87264 L
s
.28248 .87264 m
.28836 .88283 L
s
.28836 .88283 m
.29424 .87264 L
s
.29424 .87264 m
.306 .87264 L
s
.306 .87264 m
.30012 .86246 L
s
.30012 .86246 m
.306 .85228 L
s
.306 .85228 m
.31775 .85228 L
s
.31775 .85228 m
.32363 .86246 L
s
.32363 .86246 m
.32951 .85228 L
s
.32951 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.34715 .86246 L
s
.34715 .86246 m
.34127 .87264 L
s
.34127 .87264 m
.35303 .87264 L
s
.35303 .87264 m
.35891 .88283 L
s
.35891 .88283 m
.35303 .89301 L
s
.35303 .89301 m
.34127 .89301 L
s
.34127 .89301 m
.34715 .90319 L
s
.34715 .90319 m
.34127 .91337 L
s
.34127 .91337 m
.35303 .91337 L
s
.35303 .91337 m
.35891 .92356 L
s
.35891 .92356 m
.36479 .91337 L
s
.36479 .91337 m
.37654 .91337 L
s
.37654 .91337 m
.38242 .92356 L
s
.38242 .92356 m
.37654 .93374 L
s
.37654 .93374 m
.3883 .93374 L
s
.3883 .93374 m
.39418 .94392 L
s
.39418 .94392 m
.3883 .9541 L
s
.3883 .9541 m
.37654 .9541 L
s
.37654 .9541 m
.38242 .96429 L
s
.38242 .96429 m
.37654 .97447 L
s
.37654 .97447 m
.36479 .97447 L
s
.36479 .97447 m
.35891 .96429 L
s
.35891 .96429 m
.35303 .97447 L
s
.35303 .97447 m
.34127 .97447 L
s
.34127 .97447 m
.34715 .98465 L
s
.34715 .98465 m
.34127 .99483 L
s
.34127 .99483 m
.35303 .99483 L
s
.35303 .99483 m
.35891 1.00502 L
s
.35891 1.00502 m
.35303 1.0152 L
s
.35303 1.0152 m
.34127 1.0152 L
s
.34127 1.0152 m
.34715 1.02538 L
s
.34715 1.02538 m
.34127 1.03556 L
s
.34127 1.03556 m
.35303 1.03556 L
s
.35303 1.03556 m
.35891 1.04575 L
s
.35891 1.04575 m
.36479 1.03556 L
s
.36479 1.03556 m
.37654 1.03556 L
s
.37654 1.03556 m
.38242 1.04575 L
s
.38242 1.04575 m
.37654 1.05593 L
s
.37654 1.05593 m
.3883 1.05593 L
s
.3883 1.05593 m
.39418 1.06611 L
s
.39418 1.06611 m
.40006 1.05593 L
s
.40006 1.05593 m
.41182 1.05593 L
s
.41182 1.05593 m
.40594 1.04575 L
s
.40594 1.04575 m
.41182 1.03556 L
s
.41182 1.03556 m
.42357 1.03556 L
s
.42357 1.03556 m
.42945 1.04575 L
s
.42945 1.04575 m
.43533 1.03556 L
s
.43533 1.03556 m
.44709 1.03556 L
s
.44709 1.03556 m
.45297 1.04575 L
s
.45297 1.04575 m
.44709 1.05593 L
s
.44709 1.05593 m
.45885 1.05593 L
s
.45885 1.05593 m
.46473 1.06611 L
s
.46473 1.06611 m
.45885 1.07629 L
s
.45885 1.07629 m
.44709 1.07629 L
s
.44709 1.07629 m
.45297 1.08648 L
s
.45297 1.08648 m
.44709 1.09666 L
s
.44709 1.09666 m
.45885 1.09666 L
s
.45885 1.09666 m
.46473 1.10684 L
s
.46473 1.10684 m
.47061 1.09666 L
s
.47061 1.09666 m
.48236 1.09666 L
s
.48236 1.09666 m
.48824 1.10684 L
s
.48824 1.10684 m
.48236 1.11703 L
s
.48236 1.11703 m
.49412 1.11703 L
s
.49412 1.11703 m
.5 1.12721 L
s
.5 1.12721 m
.50588 1.11703 L
s
.50588 1.11703 m
.51764 1.11703 L
s
.51764 1.11703 m
.51176 1.10684 L
s
.51176 1.10684 m
.51764 1.09666 L
s
.51764 1.09666 m
.52939 1.09666 L
s
.52939 1.09666 m
.53527 1.10684 L
s
.53527 1.10684 m
.54115 1.09666 L
s
.54115 1.09666 m
.55291 1.09666 L
s
.55291 1.09666 m
.54703 1.08648 L
s
.54703 1.08648 m
.55291 1.07629 L
s
.55291 1.07629 m
.54115 1.07629 L
s
.54115 1.07629 m
.53527 1.06611 L
s
.53527 1.06611 m
.54115 1.05593 L
s
.54115 1.05593 m
.55291 1.05593 L
s
.55291 1.05593 m
.54703 1.04575 L
s
.54703 1.04575 m
.55291 1.03556 L
s
.55291 1.03556 m
.56467 1.03556 L
s
.56467 1.03556 m
.57055 1.04575 L
s
.57055 1.04575 m
.57643 1.03556 L
s
.57643 1.03556 m
.58818 1.03556 L
s
.58818 1.03556 m
.59406 1.04575 L
s
.59406 1.04575 m
.58818 1.05593 L
s
.58818 1.05593 m
.59994 1.05593 L
s
.59994 1.05593 m
.60582 1.06611 L
s
.60582 1.06611 m
.6117 1.05593 L
s
.6117 1.05593 m
.62346 1.05593 L
s
.62346 1.05593 m
.61758 1.04575 L
s
.61758 1.04575 m
.62346 1.03556 L
s
.62346 1.03556 m
.63521 1.03556 L
s
.63521 1.03556 m
.64109 1.04575 L
s
.64109 1.04575 m
.64697 1.03556 L
s
.64697 1.03556 m
.65873 1.03556 L
s
.65873 1.03556 m
.65285 1.02538 L
s
.65285 1.02538 m
.65873 1.0152 L
s
.65873 1.0152 m
.64697 1.0152 L
s
.64697 1.0152 m
.64109 1.00502 L
s
.64109 1.00502 m
.64697 .99483 L
s
.64697 .99483 m
.65873 .99483 L
s
.65873 .99483 m
.65285 .98465 L
s
.65285 .98465 m
.65873 .97447 L
s
.65873 .97447 m
.64697 .97447 L
s
.64697 .97447 m
.64109 .96429 L
s
.64109 .96429 m
.63521 .97447 L
s
.63521 .97447 m
.62346 .97447 L
s
.62346 .97447 m
.61758 .96429 L
s
.61758 .96429 m
.62346 .9541 L
s
.62346 .9541 m
.6117 .9541 L
s
.6117 .9541 m
.60582 .94392 L
s
.60582 .94392 m
.6117 .93374 L
s
.6117 .93374 m
.62346 .93374 L
s
.62346 .93374 m
.61758 .92356 L
s
.61758 .92356 m
.62346 .91337 L
s
.62346 .91337 m
.63521 .91337 L
s
.63521 .91337 m
.64109 .92356 L
s
.64109 .92356 m
.64697 .91337 L
s
.64697 .91337 m
.65873 .91337 L
s
.65873 .91337 m
.65285 .90319 L
s
.65285 .90319 m
.65873 .89301 L
s
.65873 .89301 m
.64697 .89301 L
s
.64697 .89301 m
.64109 .88283 L
s
.64109 .88283 m
.64697 .87264 L
s
.64697 .87264 m
.65873 .87264 L
s
.65873 .87264 m
.65285 .86246 L
s
.65285 .86246 m
.65873 .85228 L
s
.65873 .85228 m
.67049 .85228 L
s
.67049 .85228 m
.67637 .86246 L
s
.67637 .86246 m
.68225 .85228 L
s
.68225 .85228 m
.694 .85228 L
s
.694 .85228 m
.69988 .86246 L
s
.69988 .86246 m
.694 .87264 L
s
.694 .87264 m
.70576 .87264 L
s
.70576 .87264 m
.71164 .88283 L
s
.71164 .88283 m
.71752 .87264 L
s
.71752 .87264 m
.72928 .87264 L
s
.72928 .87264 m
.7234 .86246 L
s
.7234 .86246 m
.72928 .85228 L
s
.72928 .85228 m
.74103 .85228 L
s
.74103 .85228 m
.74691 .86246 L
s
.74691 .86246 m
.75279 .85228 L
s
.75279 .85228 m
.76455 .85228 L
s
.76455 .85228 m
.77043 .86246 L
s
.77043 .86246 m
.76455 .87264 L
s
.76455 .87264 m
.77631 .87264 L
s
.77631 .87264 m
.78219 .88283 L
s
.78219 .88283 m
.77631 .89301 L
s
.77631 .89301 m
.76455 .89301 L
s
.76455 .89301 m
.77043 .90319 L
s
.77043 .90319 m
.76455 .91337 L
s
.76455 .91337 m
.77631 .91337 L
s
.77631 .91337 m
.78219 .92356 L
s
.78219 .92356 m
.78807 .91337 L
s
.78807 .91337 m
.79982 .91337 L
s
.79982 .91337 m
.8057 .92356 L
s
.8057 .92356 m
.79982 .93374 L
s
.79982 .93374 m
.81158 .93374 L
s
.81158 .93374 m
.81746 .94392 L
s
.81746 .94392 m
.82334 .93374 L
s
.82334 .93374 m
.8351 .93374 L
s
.8351 .93374 m
.82922 .92356 L
s
.82922 .92356 m
.8351 .91337 L
s
.8351 .91337 m
.84685 .91337 L
s
.84685 .91337 m
.85273 .92356 L
s
.85273 .92356 m
.85861 .91337 L
s
.85861 .91337 m
.87037 .91337 L
s
.87037 .91337 m
.86449 .90319 L
s
.86449 .90319 m
.87037 .89301 L
s
.87037 .89301 m
.85861 .89301 L
s
.85861 .89301 m
.85273 .88283 L
s
.85273 .88283 m
.85861 .87264 L
s
.85861 .87264 m
.87037 .87264 L
s
.87037 .87264 m
.86449 .86246 L
s
.86449 .86246 m
.87037 .85228 L
s
.87037 .85228 m
.88213 .85228 L
s
.88213 .85228 m
.88801 .86246 L
s
.88801 .86246 m
.89389 .85228 L
s
.89389 .85228 m
.90564 .85228 L
s
.90564 .85228 m
.91152 .86246 L
s
.91152 .86246 m
.90564 .87264 L
s
.90564 .87264 m
.9174 .87264 L
s
.9174 .87264 m
.92328 .88283 L
s
.92328 .88283 m
.92916 .87264 L
s
.92916 .87264 m
.94092 .87264 L
s
.94092 .87264 m
.93504 .86246 L
s
.93504 .86246 m
.94092 .85228 L
s
.94092 .85228 m
.95267 .85228 L
s
.95267 .85228 m
.95855 .86246 L
s
.95855 .86246 m
.96443 .85228 L
s
.96443 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.97031 .8421 L
s
.97031 .8421 m
.97619 .83191 L
s
.97619 .83191 m
.96443 .83191 L
s
.96443 .83191 m
.95855 .82173 L
s
.95855 .82173 m
.96443 .81155 L
s
.96443 .81155 m
.97619 .81155 L
s
.97619 .81155 m
.97031 .80137 L
s
.97031 .80137 m
.97619 .79118 L
s
.97619 .79118 m
.96443 .79118 L
s
.96443 .79118 m
.95855 .781 L
s
.95855 .781 m
.95267 .79118 L
s
.95267 .79118 m
.94092 .79118 L
s
.94092 .79118 m
.93504 .781 L
s
.93504 .781 m
.94092 .77082 L
s
.94092 .77082 m
.92916 .77082 L
s
.92916 .77082 m
.92328 .76064 L
s
.92328 .76064 m
.92916 .75045 L
s
.92916 .75045 m
.94092 .75045 L
s
.94092 .75045 m
.93504 .74027 L
s
.93504 .74027 m
.94092 .73009 L
s
.94092 .73009 m
.95267 .73009 L
s
.95267 .73009 m
.95855 .74027 L
s
.95855 .74027 m
.96443 .73009 L
s
.96443 .73009 m
.97619 .73009 L
s
.97619 .73009 m
.97031 .71991 L
s
.97031 .71991 m
.97619 .70972 L
s
.97619 .70972 m
.96443 .70972 L
s
.96443 .70972 m
.95855 .69954 L
s
.95855 .69954 m
.96443 .68936 L
s
.96443 .68936 m
.97619 .68936 L
s
.97619 .68936 m
.97031 .67918 L
s
.97031 .67918 m
.97619 .66899 L
s
.97619 .66899 m
.96443 .66899 L
s
.96443 .66899 m
.95855 .65881 L
s
.95855 .65881 m
.95267 .66899 L
s
.95267 .66899 m
.94092 .66899 L
s
.94092 .66899 m
.93504 .65881 L
s
.93504 .65881 m
.94092 .64863 L
s
.94092 .64863 m
.92916 .64863 L
s
.92916 .64863 m
.92328 .63845 L
s
.92328 .63845 m
.9174 .64863 L
s
.9174 .64863 m
.90564 .64863 L
s
.90564 .64863 m
.91152 .65881 L
s
.91152 .65881 m
.90564 .66899 L
s
.90564 .66899 m
.89389 .66899 L
s
.89389 .66899 m
.88801 .65881 L
s
.88801 .65881 m
.88213 .66899 L
s
.88213 .66899 m
.87037 .66899 L
s
.87037 .66899 m
.86449 .65881 L
s
.86449 .65881 m
.87037 .64863 L
s
.87037 .64863 m
.85861 .64863 L
s
.85861 .64863 m
.85273 .63845 L
s
.85273 .63845 m
.85861 .62826 L
s
.85861 .62826 m
.87037 .62826 L
s
.87037 .62826 m
.86449 .61808 L
s
.86449 .61808 m
.87037 .6079 L
s
.87037 .6079 m
.85861 .6079 L
s
.85861 .6079 m
.85273 .59772 L
s
.85273 .59772 m
.84685 .6079 L
s
.84685 .6079 m
.8351 .6079 L
s
.8351 .6079 m
.82922 .59772 L
s
.82922 .59772 m
.8351 .58753 L
s
.8351 .58753 m
.82334 .58753 L
s
.82334 .58753 m
.81746 .57735 L
s
.81746 .57735 m
.82334 .56717 L
s
.82334 .56717 m
.8351 .56717 L
s
.8351 .56717 m
.82922 .55699 L
s
.82922 .55699 m
.8351 .5468 L
s
.8351 .5468 m
.84685 .5468 L
s
.84685 .5468 m
.85273 .55699 L
s
.85273 .55699 m
.85861 .5468 L
s
.85861 .5468 m
.87037 .5468 L
s
.87037 .5468 m
.86449 .53662 L
s
.86449 .53662 m
.87037 .52644 L
s
.87037 .52644 m
.85861 .52644 L
s
.85861 .52644 m
.85273 .51626 L
s
.85273 .51626 m
.85861 .50607 L
s
.85861 .50607 m
.87037 .50607 L
s
.87037 .50607 m
.86449 .49589 L
s
.86449 .49589 m
.87037 .48571 L
s
.87037 .48571 m
.88213 .48571 L
s
.88213 .48571 m
.88801 .49589 L
s
.88801 .49589 m
.89389 .48571 L
s
.89389 .48571 m
.90564 .48571 L
s
.90564 .48571 m
.91152 .49589 L
s
.91152 .49589 m
.90564 .50607 L
s
.90564 .50607 m
.9174 .50607 L
s
.9174 .50607 m
.92328 .51626 L
s
.92328 .51626 m
.92916 .50607 L
s
.92916 .50607 m
.94092 .50607 L
s
.94092 .50607 m
.93504 .49589 L
s
.93504 .49589 m
.94092 .48571 L
s
.94092 .48571 m
.95267 .48571 L
s
.95267 .48571 m
.95855 .49589 L
s
.95855 .49589 m
.96443 .48571 L
s
.96443 .48571 m
.97619 .48571 L
s
.97619 .48571 m
.97031 .47552 L
s
.97031 .47552 m
.97619 .46534 L
s
.97619 .46534 m
.96443 .46534 L
s
.96443 .46534 m
.95855 .45516 L
s
.95855 .45516 m
.96443 .44498 L
s
.96443 .44498 m
.97619 .44498 L
s
.97619 .44498 m
.97031 .43479 L
s
.97031 .43479 m
.97619 .42461 L
s
.97619 .42461 m
.96443 .42461 L
s
.96443 .42461 m
.95855 .41443 L
s
.95855 .41443 m
.95267 .42461 L
s
.95267 .42461 m
.94092 .42461 L
s
.94092 .42461 m
.93504 .41443 L
s
.93504 .41443 m
.94092 .40425 L
s
.94092 .40425 m
.92916 .40425 L
s
.92916 .40425 m
.92328 .39406 L
s
.92328 .39406 m
.92916 .38388 L
s
.92916 .38388 m
.94092 .38388 L
s
.94092 .38388 m
.93504 .3737 L
s
.93504 .3737 m
.94092 .36352 L
s
.94092 .36352 m
.95267 .36352 L
s
.95267 .36352 m
.95855 .3737 L
s
.95855 .3737 m
.96443 .36352 L
s
.96443 .36352 m
.97619 .36352 L
s
.97619 .36352 m
.97031 .35333 L
s
.97031 .35333 m
.97619 .34315 L
s
.97619 .34315 m
.96443 .34315 L
s
.96443 .34315 m
.95855 .33297 L
s
.95855 .33297 m
.96443 .32279 L
s
.96443 .32279 m
.97619 .32279 L
s
.97619 .32279 m
.97031 .3126 L
s
.97031 .3126 m
.97619 .30242 L
s
.97619 .30242 m
.96443 .30242 L
s
.96443 .30242 m
.95855 .29224 L
s
.95855 .29224 m
.95267 .30242 L
s
.95267 .30242 m
.94092 .30242 L
s
.94092 .30242 m
.93504 .29224 L
s
.93504 .29224 m
.94092 .28206 L
s
.94092 .28206 m
.92916 .28206 L
s
.92916 .28206 m
.92328 .27187 L
s
.92328 .27187 m
.9174 .28206 L
s
.9174 .28206 m
.90564 .28206 L
s
.90564 .28206 m
.91152 .29224 L
s
.91152 .29224 m
.90564 .30242 L
s
.90564 .30242 m
.89389 .30242 L
s
.89389 .30242 m
.88801 .29224 L
s
.88801 .29224 m
.88213 .30242 L
s
.88213 .30242 m
.87037 .30242 L
s
.87037 .30242 m
.86449 .29224 L
s
.86449 .29224 m
.87037 .28206 L
s
.87037 .28206 m
.85861 .28206 L
s
.85861 .28206 m
.85273 .27187 L
s
.85273 .27187 m
.85861 .26169 L
s
.85861 .26169 m
.87037 .26169 L
s
.87037 .26169 m
.86449 .25151 L
s
.86449 .25151 m
.87037 .24133 L
s
.87037 .24133 m
.85861 .24133 L
s
.85861 .24133 m
.85273 .23114 L
s
.85273 .23114 m
.84685 .24133 L
s
.84685 .24133 m
.8351 .24133 L
s
.8351 .24133 m
.82922 .23114 L
s
.82922 .23114 m
.8351 .22096 L
s
.8351 .22096 m
.82334 .22096 L
s
.82334 .22096 m
.81746 .21078 L
s
.81746 .21078 m
.81158 .22096 L
s
.81158 .22096 m
.79982 .22096 L
s
.79982 .22096 m
.8057 .23114 L
s
.8057 .23114 m
.79982 .24133 L
s
.79982 .24133 m
.78807 .24133 L
s
.78807 .24133 m
.78219 .23114 L
s
.78219 .23114 m
.77631 .24133 L
s
.77631 .24133 m
.76455 .24133 L
s
.76455 .24133 m
.77043 .25151 L
s
.77043 .25151 m
.76455 .26169 L
s
.76455 .26169 m
.77631 .26169 L
s
.77631 .26169 m
.78219 .27187 L
s
.78219 .27187 m
.77631 .28206 L
s
.77631 .28206 m
.76455 .28206 L
s
.76455 .28206 m
.77043 .29224 L
s
.77043 .29224 m
.76455 .30242 L
s
.76455 .30242 m
.75279 .30242 L
s
.75279 .30242 m
.74691 .29224 L
s
.74691 .29224 m
.74103 .30242 L
s
.74103 .30242 m
.72928 .30242 L
s
.72928 .30242 m
.7234 .29224 L
s
.7234 .29224 m
.72928 .28206 L
s
.72928 .28206 m
.71752 .28206 L
s
.71752 .28206 m
.71164 .27187 L
s
.71164 .27187 m
.70576 .28206 L
s
.70576 .28206 m
.694 .28206 L
s
.694 .28206 m
.69988 .29224 L
s
.69988 .29224 m
.694 .30242 L
s
.694 .30242 m
.68225 .30242 L
s
.68225 .30242 m
.67637 .29224 L
s
.67637 .29224 m
.67049 .30242 L
s
.67049 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.65285 .29224 L
s
.65285 .29224 m
.65873 .28206 L
s
.65873 .28206 m
.64697 .28206 L
s
.64697 .28206 m
.64109 .27187 L
s
.64109 .27187 m
.64697 .26169 L
s
.64697 .26169 m
.65873 .26169 L
s
.65873 .26169 m
.65285 .25151 L
s
.65285 .25151 m
.65873 .24133 L
s
.65873 .24133 m
.64697 .24133 L
s
.64697 .24133 m
.64109 .23114 L
s
.64109 .23114 m
.63521 .24133 L
s
.63521 .24133 m
.62346 .24133 L
s
.62346 .24133 m
.61758 .23114 L
s
.61758 .23114 m
.62346 .22096 L
s
.62346 .22096 m
.6117 .22096 L
s
.6117 .22096 m
.60582 .21078 L
s
.60582 .21078 m
.6117 .2006 L
s
.6117 .2006 m
.62346 .2006 L
s
.62346 .2006 m
.61758 .19041 L
s
.61758 .19041 m
.62346 .18023 L
s
.62346 .18023 m
.63521 .18023 L
s
.63521 .18023 m
.64109 .19041 L
s
.64109 .19041 m
.64697 .18023 L
s
.64697 .18023 m
.65873 .18023 L
s
.65873 .18023 m
.65285 .17005 L
s
.65285 .17005 m
.65873 .15987 L
s
.65873 .15987 m
.64697 .15987 L
s
.64697 .15987 m
.64109 .14968 L
s
.64109 .14968 m
.64697 .1395 L
s
.64697 .1395 m
.65873 .1395 L
s
.65873 .1395 m
.65285 .12932 L
s
.65285 .12932 m
.65873 .11914 L
s
.65873 .11914 m
.64697 .11914 L
s
.64697 .11914 m
.64109 .10895 L
s
.64109 .10895 m
.63521 .11914 L
s
.63521 .11914 m
.62346 .11914 L
s
.62346 .11914 m
.61758 .10895 L
s
.61758 .10895 m
.62346 .09877 L
s
.62346 .09877 m
.6117 .09877 L
s
.6117 .09877 m
.60582 .08859 L
s
.60582 .08859 m
.59994 .09877 L
s
.59994 .09877 m
.58818 .09877 L
s
.58818 .09877 m
.59406 .10895 L
s
.59406 .10895 m
.58818 .11914 L
s
.58818 .11914 m
.57643 .11914 L
s
.57643 .11914 m
.57055 .10895 L
s
.57055 .10895 m
.56467 .11914 L
s
.56467 .11914 m
.55291 .11914 L
s
.55291 .11914 m
.54703 .10895 L
s
.54703 .10895 m
.55291 .09877 L
s
.55291 .09877 m
.54115 .09877 L
s
.54115 .09877 m
.53527 .08859 L
s
.53527 .08859 m
.54115 .07841 L
s
.54115 .07841 m
.55291 .07841 L
s
.55291 .07841 m
.54703 .06822 L
s
.54703 .06822 m
.55291 .05804 L
s
.55291 .05804 m
.54115 .05804 L
s
.54115 .05804 m
.53527 .04786 L
s
.53527 .04786 m
.52939 .05804 L
s
.52939 .05804 m
.51764 .05804 L
s
.51764 .05804 m
.51176 .04786 L
s
.51176 .04786 m
.51764 .03768 L
s
.51764 .03768 m
.50588 .03768 L
s
.50588 .03768 m
.5 .02749 L
s
.5 .02749 m
.49412 .03768 L
s
.49412 .03768 m
.48236 .03768 L
s
.48236 .03768 m
.48824 .04786 L
s
.48824 .04786 m
.48236 .05804 L
s
.48236 .05804 m
.47061 .05804 L
s
.47061 .05804 m
.46473 .04786 L
s
.46473 .04786 m
.45885 .05804 L
s
.45885 .05804 m
.44709 .05804 L
s
.44709 .05804 m
.45297 .06822 L
s
.45297 .06822 m
.44709 .07841 L
s
.44709 .07841 m
.45885 .07841 L
s
.45885 .07841 m
.46473 .08859 L
s
.46473 .08859 m
.45885 .09877 L
s
.45885 .09877 m
.44709 .09877 L
s
.44709 .09877 m
.45297 .10895 L
s
.45297 .10895 m
.44709 .11914 L
s
.44709 .11914 m
.43533 .11914 L
s
.43533 .11914 m
.42945 .10895 L
s
.42945 .10895 m
.42357 .11914 L
s
.42357 .11914 m
.41182 .11914 L
s
.41182 .11914 m
.40594 .10895 L
s
.40594 .10895 m
.41182 .09877 L
s
.41182 .09877 m
.40006 .09877 L
s
.40006 .09877 m
.39418 .08859 L
s
.39418 .08859 m
.3883 .09877 L
s
.3883 .09877 m
.37654 .09877 L
s
.37654 .09877 m
.38242 .10895 L
s
.38242 .10895 m
.37654 .11914 L
s
.37654 .11914 m
.36479 .11914 L
s
.36479 .11914 m
.35891 .10895 L
s
.35891 .10895 m
.35303 .11914 L
s
.35303 .11914 m
.34127 .11914 L
s
.34127 .11914 m
.34715 .12932 L
s
.34715 .12932 m
.34127 .1395 L
s
.34127 .1395 m
.35303 .1395 L
s
.35303 .1395 m
.35891 .14968 L
s
.35891 .14968 m
.35303 .15987 L
s
.35303 .15987 m
.34127 .15987 L
s
.34127 .15987 m
.34715 .17005 L
s
.34715 .17005 m
.34127 .18023 L
s
.34127 .18023 m
.35303 .18023 L
s
.35303 .18023 m
.35891 .19041 L
s
.35891 .19041 m
.36479 .18023 L
s
.36479 .18023 m
.37654 .18023 L
s
.37654 .18023 m
.38242 .19041 L
s
.38242 .19041 m
.37654 .2006 L
s
.37654 .2006 m
.3883 .2006 L
s
.3883 .2006 m
.39418 .21078 L
s
.39418 .21078 m
.3883 .22096 L
s
.3883 .22096 m
.37654 .22096 L
s
.37654 .22096 m
.38242 .23114 L
s
.38242 .23114 m
.37654 .24133 L
s
.37654 .24133 m
.36479 .24133 L
s
.36479 .24133 m
.35891 .23114 L
s
.35891 .23114 m
.35303 .24133 L
s
.35303 .24133 m
.34127 .24133 L
s
.34127 .24133 m
.34715 .25151 L
s
.34715 .25151 m
.34127 .26169 L
s
.34127 .26169 m
.35303 .26169 L
s
.35303 .26169 m
.35891 .27187 L
s
.35891 .27187 m
.35303 .28206 L
s
.35303 .28206 m
.34127 .28206 L
s
.34127 .28206 m
.34715 .29224 L
s
.34715 .29224 m
.34127 .30242 L
s
.34127 .30242 m
.32951 .30242 L
s
.32951 .30242 m
.32363 .29224 L
s
.32363 .29224 m
.31775 .30242 L
s
.31775 .30242 m
.306 .30242 L
s
.306 .30242 m
.30012 .29224 L
s
.30012 .29224 m
.306 .28206 L
s
.306 .28206 m
.29424 .28206 L
s
.29424 .28206 m
.28836 .27187 L
s
.28836 .27187 m
.28248 .28206 L
s
.28248 .28206 m
.27072 .28206 L
s
.27072 .28206 m
.2766 .29224 L
s
.2766 .29224 m
.27072 .30242 L
s
.27072 .30242 m
.25897 .30242 L
s
.25897 .30242 m
.25309 .29224 L
s
.25309 .29224 m
.24721 .30242 L
s
.24721 .30242 m
.23545 .30242 L
s
.23545 .30242 m
.22957 .29224 L
s
.22957 .29224 m
.23545 .28206 L
s
.23545 .28206 m
.22369 .28206 L
s
.22369 .28206 m
.21781 .27187 L
s
.21781 .27187 m
.22369 .26169 L
s
.22369 .26169 m
.23545 .26169 L
s
.23545 .26169 m
.22957 .25151 L
s
.22957 .25151 m
.23545 .24133 L
s
.23545 .24133 m
.22369 .24133 L
s
.22369 .24133 m
.21781 .23114 L
s
.21781 .23114 m
.21193 .24133 L
s
.21193 .24133 m
.20018 .24133 L
s
.20018 .24133 m
.1943 .23114 L
s
.1943 .23114 m
.20018 .22096 L
s
.20018 .22096 m
.18842 .22096 L
s
.18842 .22096 m
.18254 .21078 L
s
.18254 .21078 m
.17666 .22096 L
s
.17666 .22096 m
.1649 .22096 L
s
.1649 .22096 m
.17078 .23114 L
s
.17078 .23114 m
.1649 .24133 L
s
.1649 .24133 m
.15315 .24133 L
s
.15315 .24133 m
.14727 .23114 L
s
.14727 .23114 m
.14139 .24133 L
s
.14139 .24133 m
.12963 .24133 L
s
.12963 .24133 m
.13551 .25151 L
s
.13551 .25151 m
.12963 .26169 L
s
.12963 .26169 m
.14139 .26169 L
s
.14139 .26169 m
.14727 .27187 L
s
.14727 .27187 m
.14139 .28206 L
s
.14139 .28206 m
.12963 .28206 L
s
.12963 .28206 m
.13551 .29224 L
s
.13551 .29224 m
.12963 .30242 L
s
.12963 .30242 m
.11787 .30242 L
s
.11787 .30242 m
.11199 .29224 L
s
.11199 .29224 m
.10611 .30242 L
s
.10611 .30242 m
.09436 .30242 L
s
.09436 .30242 m
.08848 .29224 L
s
.08848 .29224 m
.09436 .28206 L
s
.09436 .28206 m
.0826 .28206 L
s
.0826 .28206 m
.07672 .27187 L
s
.07672 .27187 m
.07084 .28206 L
s
.07084 .28206 m
.05908 .28206 L
s
.05908 .28206 m
.06496 .29224 L
s
.06496 .29224 m
.05908 .30242 L
s
.05908 .30242 m
.04733 .30242 L
s
.04733 .30242 m
.04145 .29224 L
s
.04145 .29224 m
.03557 .30242 L
s
.03557 .30242 m
.02381 .30242 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0130918 0.97619 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.087607 0.824786 0.261905 0.824786 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
.08761 .2619 m
.5 .97619 L
s
.5 .97619 m
.91239 .2619 L
s
.91239 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.22507 .5 L
s
.22507 .5 m
.08761 .7381 L
s
.08761 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.5 .97619 L
s
.5 .97619 m
.63746 .7381 L
s
.63746 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.77493 .5 L
s
.77493 .5 m
.91239 .2619 L
s
.91239 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.5 .02381 L
s
.5 .02381 m
.36254 .2619 L
s
.36254 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.13343 .34127 L
s
.13343 .34127 m
.08761 .42063 L
s
.08761 .42063 m
.17925 .42063 L
s
.17925 .42063 m
.22507 .5 L
s
.22507 .5 m
.17925 .57937 L
s
.17925 .57937 m
.08761 .57937 L
s
.08761 .57937 m
.13343 .65873 L
s
.13343 .65873 m
.08761 .7381 L
s
.08761 .7381 m
.17925 .7381 L
s
.17925 .7381 m
.22507 .81746 L
s
.22507 .81746 m
.27089 .7381 L
s
.27089 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.40836 .81746 L
s
.40836 .81746 m
.36254 .89683 L
s
.36254 .89683 m
.45418 .89683 L
s
.45418 .89683 m
.5 .97619 L
s
.5 .97619 m
.54582 .89683 L
s
.54582 .89683 m
.63746 .89683 L
s
.63746 .89683 m
.59164 .81746 L
s
.59164 .81746 m
.63746 .7381 L
s
.63746 .7381 m
.72911 .7381 L
s
.72911 .7381 m
.77493 .81746 L
s
.77493 .81746 m
.82075 .7381 L
s
.82075 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.86657 .65873 L
s
.86657 .65873 m
.91239 .57937 L
s
.91239 .57937 m
.82075 .57937 L
s
.82075 .57937 m
.77493 .5 L
s
.77493 .5 m
.82075 .42063 L
s
.82075 .42063 m
.91239 .42063 L
s
.91239 .42063 m
.86657 .34127 L
s
.86657 .34127 m
.91239 .2619 L
s
.91239 .2619 m
.82075 .2619 L
s
.82075 .2619 m
.77493 .18254 L
s
.77493 .18254 m
.72911 .2619 L
s
.72911 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.59164 .18254 L
s
.59164 .18254 m
.63746 .10317 L
s
.63746 .10317 m
.54582 .10317 L
s
.54582 .10317 m
.5 .02381 L
s
.5 .02381 m
.45418 .10317 L
s
.45418 .10317 m
.36254 .10317 L
s
.36254 .10317 m
.40836 .18254 L
s
.40836 .18254 m
.36254 .2619 L
s
.36254 .2619 m
.27089 .2619 L
s
.27089 .2619 m
.22507 .18254 L
s
.22507 .18254 m
.17925 .2619 L
s
.17925 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.10288 .28836 L
s
.10288 .28836 m
.08761 .31481 L
s
.08761 .31481 m
.11815 .31481 L
s
.11815 .31481 m
.13343 .34127 L
s
.13343 .34127 m
.11815 .36772 L
s
.11815 .36772 m
.08761 .36772 L
s
.08761 .36772 m
.10288 .39418 L
s
.10288 .39418 m
.08761 .42063 L
s
.08761 .42063 m
.11815 .42063 L
s
.11815 .42063 m
.13343 .44709 L
s
.13343 .44709 m
.1487 .42063 L
s
.1487 .42063 m
.17925 .42063 L
s
.17925 .42063 m
.19452 .44709 L
s
.19452 .44709 m
.17925 .47354 L
s
.17925 .47354 m
.2098 .47354 L
s
.2098 .47354 m
.22507 .5 L
s
.22507 .5 m
.2098 .52646 L
s
.2098 .52646 m
.17925 .52646 L
s
.17925 .52646 m
.19452 .55291 L
s
.19452 .55291 m
.17925 .57937 L
s
.17925 .57937 m
.1487 .57937 L
s
.1487 .57937 m
.13343 .55291 L
s
.13343 .55291 m
.11815 .57937 L
s
.11815 .57937 m
.08761 .57937 L
s
.08761 .57937 m
.10288 .60582 L
s
.10288 .60582 m
.08761 .63228 L
s
.08761 .63228 m
.11815 .63228 L
s
.11815 .63228 m
.13343 .65873 L
s
.13343 .65873 m
.11815 .68519 L
s
.11815 .68519 m
.08761 .68519 L
s
.08761 .68519 m
.10288 .71164 L
s
.10288 .71164 m
.08761 .7381 L
s
.08761 .7381 m
.11815 .7381 L
s
.11815 .7381 m
.13343 .76455 L
s
.13343 .76455 m
.1487 .7381 L
s
.1487 .7381 m
.17925 .7381 L
s
.17925 .7381 m
.19452 .76455 L
s
.19452 .76455 m
.17925 .79101 L
s
.17925 .79101 m
.2098 .79101 L
s
.2098 .79101 m
.22507 .81746 L
s
.22507 .81746 m
.24035 .79101 L
s
.24035 .79101 m
.27089 .79101 L
s
.27089 .79101 m
.25562 .76455 L
s
.25562 .76455 m
.27089 .7381 L
s
.27089 .7381 m
.30144 .7381 L
s
.30144 .7381 m
.31671 .76455 L
s
.31671 .76455 m
.33199 .7381 L
s
.33199 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.37781 .76455 L
s
.37781 .76455 m
.36254 .79101 L
s
.36254 .79101 m
.39308 .79101 L
s
.39308 .79101 m
.40836 .81746 L
s
.40836 .81746 m
.39308 .84392 L
s
.39308 .84392 m
.36254 .84392 L
s
.36254 .84392 m
.37781 .87037 L
s
.37781 .87037 m
.36254 .89683 L
s
.36254 .89683 m
.39308 .89683 L
s
.39308 .89683 m
.40836 .92328 L
s
.40836 .92328 m
.42363 .89683 L
s
.42363 .89683 m
.45418 .89683 L
s
.45418 .89683 m
.46945 .92328 L
s
.46945 .92328 m
.45418 .94974 L
s
.45418 .94974 m
.48473 .94974 L
s
.48473 .94974 m
.5 .97619 L
s
.5 .97619 m
.51527 .94974 L
s
.51527 .94974 m
.54582 .94974 L
s
.54582 .94974 m
.53055 .92328 L
s
.53055 .92328 m
.54582 .89683 L
s
.54582 .89683 m
.57637 .89683 L
s
.57637 .89683 m
.59164 .92328 L
s
.59164 .92328 m
.60692 .89683 L
s
.60692 .89683 m
.63746 .89683 L
s
.63746 .89683 m
.62219 .87037 L
s
.62219 .87037 m
.63746 .84392 L
s
.63746 .84392 m
.60692 .84392 L
s
.60692 .84392 m
.59164 .81746 L
s
.59164 .81746 m
.60692 .79101 L
s
.60692 .79101 m
.63746 .79101 L
s
.63746 .79101 m
.62219 .76455 L
s
.62219 .76455 m
.63746 .7381 L
s
.63746 .7381 m
.66801 .7381 L
s
.66801 .7381 m
.68329 .76455 L
s
.68329 .76455 m
.69856 .7381 L
s
.69856 .7381 m
.72911 .7381 L
s
.72911 .7381 m
.74438 .76455 L
s
.74438 .76455 m
.72911 .79101 L
s
.72911 .79101 m
.75965 .79101 L
s
.75965 .79101 m
.77493 .81746 L
s
.77493 .81746 m
.7902 .79101 L
s
.7902 .79101 m
.82075 .79101 L
s
.82075 .79101 m
.80548 .76455 L
s
.80548 .76455 m
.82075 .7381 L
s
.82075 .7381 m
.8513 .7381 L
s
.8513 .7381 m
.86657 .76455 L
s
.86657 .76455 m
.88185 .7381 L
s
.88185 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.89712 .71164 L
s
.89712 .71164 m
.91239 .68519 L
s
.91239 .68519 m
.88185 .68519 L
s
.88185 .68519 m
.86657 .65873 L
s
.86657 .65873 m
.88185 .63228 L
s
.88185 .63228 m
.91239 .63228 L
s
.91239 .63228 m
.89712 .60582 L
s
.89712 .60582 m
.91239 .57937 L
s
.91239 .57937 m
.88185 .57937 L
s
.88185 .57937 m
.86657 .55291 L
s
.86657 .55291 m
.8513 .57937 L
s
.8513 .57937 m
.82075 .57937 L
s
.82075 .57937 m
.80548 .55291 L
s
.80548 .55291 m
.82075 .52646 L
s
.82075 .52646 m
.7902 .52646 L
s
.7902 .52646 m
.77493 .5 L
s
.77493 .5 m
.7902 .47354 L
s
.7902 .47354 m
.82075 .47354 L
s
.82075 .47354 m
.80548 .44709 L
s
.80548 .44709 m
.82075 .42063 L
s
.82075 .42063 m
.8513 .42063 L
s
.8513 .42063 m
.86657 .44709 L
s
.86657 .44709 m
.88185 .42063 L
s
.88185 .42063 m
.91239 .42063 L
s
.91239 .42063 m
.89712 .39418 L
s
.89712 .39418 m
.91239 .36772 L
s
.91239 .36772 m
.88185 .36772 L
s
.88185 .36772 m
.86657 .34127 L
s
.86657 .34127 m
.88185 .31481 L
s
.88185 .31481 m
.91239 .31481 L
s
.91239 .31481 m
.89712 .28836 L
s
.89712 .28836 m
.91239 .2619 L
s
.91239 .2619 m
.88185 .2619 L
s
.88185 .2619 m
.86657 .23545 L
s
.86657 .23545 m
.8513 .2619 L
s
.8513 .2619 m
.82075 .2619 L
s
.82075 .2619 m
.80548 .23545 L
s
.80548 .23545 m
.82075 .20899 L
s
.82075 .20899 m
.7902 .20899 L
s
.7902 .20899 m
.77493 .18254 L
s
.77493 .18254 m
.75965 .20899 L
s
.75965 .20899 m
.72911 .20899 L
s
.72911 .20899 m
.74438 .23545 L
s
.74438 .23545 m
.72911 .2619 L
s
.72911 .2619 m
.69856 .2619 L
s
.69856 .2619 m
.68329 .23545 L
s
.68329 .23545 m
.66801 .2619 L
s
.66801 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.62219 .23545 L
s
.62219 .23545 m
.63746 .20899 L
s
.63746 .20899 m
.60692 .20899 L
s
.60692 .20899 m
.59164 .18254 L
s
.59164 .18254 m
.60692 .15608 L
s
.60692 .15608 m
.63746 .15608 L
s
.63746 .15608 m
.62219 .12963 L
s
.62219 .12963 m
.63746 .10317 L
s
.63746 .10317 m
.60692 .10317 L
s
.60692 .10317 m
.59164 .07672 L
s
.59164 .07672 m
.57637 .10317 L
s
.57637 .10317 m
.54582 .10317 L
s
.54582 .10317 m
.53055 .07672 L
s
.53055 .07672 m
.54582 .05026 L
s
.54582 .05026 m
.51527 .05026 L
s
.51527 .05026 m
.5 .02381 L
s
.5 .02381 m
.48473 .05026 L
s
.48473 .05026 m
.45418 .05026 L
s
.45418 .05026 m
.46945 .07672 L
s
.46945 .07672 m
.45418 .10317 L
s
.45418 .10317 m
.42363 .10317 L
s
.42363 .10317 m
.40836 .07672 L
s
.40836 .07672 m
.39308 .10317 L
s
.39308 .10317 m
.36254 .10317 L
s
.36254 .10317 m
.37781 .12963 L
s
.37781 .12963 m
.36254 .15608 L
s
.36254 .15608 m
.39308 .15608 L
s
.39308 .15608 m
.40836 .18254 L
s
.40836 .18254 m
.39308 .20899 L
s
.39308 .20899 m
.36254 .20899 L
s
.36254 .20899 m
.37781 .23545 L
s
.37781 .23545 m
.36254 .2619 L
s
.36254 .2619 m
.33199 .2619 L
s
.33199 .2619 m
.31671 .23545 L
s
.31671 .23545 m
.30144 .2619 L
s
.30144 .2619 m
.27089 .2619 L
s
.27089 .2619 m
.25562 .23545 L
s
.25562 .23545 m
.27089 .20899 L
s
.27089 .20899 m
.24035 .20899 L
s
.24035 .20899 m
.22507 .18254 L
s
.22507 .18254 m
.2098 .20899 L
s
.2098 .20899 m
.17925 .20899 L
s
.17925 .20899 m
.19452 .23545 L
s
.19452 .23545 m
.17925 .2619 L
s
.17925 .2619 m
.1487 .2619 L
s
.1487 .2619 m
.13343 .23545 L
s
.13343 .23545 m
.11815 .2619 L
s
.11815 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.0927 .27072 L
s
.0927 .27072 m
.08761 .27954 L
s
.08761 .27954 m
.09779 .27954 L
s
.09779 .27954 m
.10288 .28836 L
s
.10288 .28836 m
.09779 .29718 L
s
.09779 .29718 m
.08761 .29718 L
s
.08761 .29718 m
.0927 .306 L
s
.0927 .306 m
.08761 .31481 L
s
.08761 .31481 m
.09779 .31481 L
s
.09779 .31481 m
.10288 .32363 L
s
.10288 .32363 m
.10797 .31481 L
s
.10797 .31481 m
.11815 .31481 L
s
.11815 .31481 m
.12325 .32363 L
s
.12325 .32363 m
.11815 .33245 L
s
.11815 .33245 m
.12834 .33245 L
s
.12834 .33245 m
.13343 .34127 L
s
.13343 .34127 m
.12834 .35009 L
s
.12834 .35009 m
.11815 .35009 L
s
.11815 .35009 m
.12325 .35891 L
s
.12325 .35891 m
.11815 .36772 L
s
.11815 .36772 m
.10797 .36772 L
s
.10797 .36772 m
.10288 .35891 L
s
.10288 .35891 m
.09779 .36772 L
s
.09779 .36772 m
.08761 .36772 L
s
.08761 .36772 m
.0927 .37654 L
s
.0927 .37654 m
.08761 .38536 L
s
.08761 .38536 m
.09779 .38536 L
s
.09779 .38536 m
.10288 .39418 L
s
.10288 .39418 m
.09779 .403 L
s
.09779 .403 m
.08761 .403 L
s
.08761 .403 m
.0927 .41182 L
s
.0927 .41182 m
.08761 .42063 L
s
.08761 .42063 m
.09779 .42063 L
s
.09779 .42063 m
.10288 .42945 L
s
.10288 .42945 m
.10797 .42063 L
s
.10797 .42063 m
.11815 .42063 L
s
.11815 .42063 m
.12325 .42945 L
s
.12325 .42945 m
.11815 .43827 L
s
.11815 .43827 m
.12834 .43827 L
s
.12834 .43827 m
.13343 .44709 L
s
.13343 .44709 m
.13852 .43827 L
s
.13852 .43827 m
.1487 .43827 L
s
.1487 .43827 m
.14361 .42945 L
s
.14361 .42945 m
.1487 .42063 L
s
.1487 .42063 m
.15888 .42063 L
s
.15888 .42063 m
.16398 .42945 L
s
.16398 .42945 m
.16907 .42063 L
s
.16907 .42063 m
.17925 .42063 L
s
.17925 .42063 m
.18434 .42945 L
s
.18434 .42945 m
.17925 .43827 L
s
.17925 .43827 m
.18943 .43827 L
s
.18943 .43827 m
.19452 .44709 L
s
.19452 .44709 m
.18943 .45591 L
s
.18943 .45591 m
.17925 .45591 L
s
.17925 .45591 m
.18434 .46473 L
s
.18434 .46473 m
.17925 .47354 L
s
.17925 .47354 m
.18943 .47354 L
s
.18943 .47354 m
.19452 .48236 L
s
.19452 .48236 m
.19961 .47354 L
s
.19961 .47354 m
.2098 .47354 L
s
.2098 .47354 m
.21489 .48236 L
s
.21489 .48236 m
.2098 .49118 L
s
.2098 .49118 m
.21998 .49118 L
s
.21998 .49118 m
.22507 .5 L
s
.22507 .5 m
.21998 .50882 L
s
.21998 .50882 m
.2098 .50882 L
s
.2098 .50882 m
.21489 .51764 L
s
.21489 .51764 m
.2098 .52646 L
s
.2098 .52646 m
.19961 .52646 L
s
.19961 .52646 m
.19452 .51764 L
s
.19452 .51764 m
.18943 .52646 L
s
.18943 .52646 m
.17925 .52646 L
s
.17925 .52646 m
.18434 .53527 L
s
.18434 .53527 m
.17925 .54409 L
s
.17925 .54409 m
.18943 .54409 L
s
.18943 .54409 m
.19452 .55291 L
s
.19452 .55291 m
.18943 .56173 L
s
.18943 .56173 m
.17925 .56173 L
s
.17925 .56173 m
.18434 .57055 L
s
.18434 .57055 m
.17925 .57937 L
s
.17925 .57937 m
.16907 .57937 L
s
.16907 .57937 m
.16398 .57055 L
s
.16398 .57055 m
.15888 .57937 L
s
.15888 .57937 m
.1487 .57937 L
s
.1487 .57937 m
.14361 .57055 L
s
.14361 .57055 m
.1487 .56173 L
s
.1487 .56173 m
.13852 .56173 L
s
.13852 .56173 m
.13343 .55291 L
s
.13343 .55291 m
.12834 .56173 L
s
.12834 .56173 m
.11815 .56173 L
s
.11815 .56173 m
.12325 .57055 L
s
.12325 .57055 m
.11815 .57937 L
s
.11815 .57937 m
.10797 .57937 L
s
.10797 .57937 m
.10288 .57055 L
s
.10288 .57055 m
.09779 .57937 L
s
.09779 .57937 m
.08761 .57937 L
s
.08761 .57937 m
.0927 .58818 L
s
.0927 .58818 m
.08761 .597 L
s
.08761 .597 m
.09779 .597 L
s
.09779 .597 m
.10288 .60582 L
s
.10288 .60582 m
.09779 .61464 L
s
.09779 .61464 m
.08761 .61464 L
s
.08761 .61464 m
.0927 .62346 L
s
.0927 .62346 m
.08761 .63228 L
s
.08761 .63228 m
.09779 .63228 L
s
.09779 .63228 m
.10288 .64109 L
s
.10288 .64109 m
.10797 .63228 L
s
.10797 .63228 m
.11815 .63228 L
s
.11815 .63228 m
.12325 .64109 L
s
.12325 .64109 m
.11815 .64991 L
s
.11815 .64991 m
.12834 .64991 L
s
.12834 .64991 m
.13343 .65873 L
s
.13343 .65873 m
.12834 .66755 L
s
.12834 .66755 m
.11815 .66755 L
s
.11815 .66755 m
.12325 .67637 L
s
.12325 .67637 m
.11815 .68519 L
s
.11815 .68519 m
.10797 .68519 L
s
.10797 .68519 m
.10288 .67637 L
s
.10288 .67637 m
.09779 .68519 L
s
.09779 .68519 m
.08761 .68519 L
s
.08761 .68519 m
.0927 .694 L
s
.0927 .694 m
.08761 .70282 L
s
.08761 .70282 m
.09779 .70282 L
s
.09779 .70282 m
.10288 .71164 L
s
.10288 .71164 m
.09779 .72046 L
s
.09779 .72046 m
.08761 .72046 L
s
.08761 .72046 m
.0927 .72928 L
s
.0927 .72928 m
.08761 .7381 L
s
.08761 .7381 m
.09779 .7381 L
s
.09779 .7381 m
.10288 .74691 L
s
.10288 .74691 m
.10797 .7381 L
s
.10797 .7381 m
.11815 .7381 L
s
.11815 .7381 m
.12325 .74691 L
s
.12325 .74691 m
.11815 .75573 L
s
.11815 .75573 m
.12834 .75573 L
s
.12834 .75573 m
.13343 .76455 L
s
.13343 .76455 m
.13852 .75573 L
s
.13852 .75573 m
.1487 .75573 L
s
.1487 .75573 m
.14361 .74691 L
s
.14361 .74691 m
.1487 .7381 L
s
.1487 .7381 m
.15888 .7381 L
s
.15888 .7381 m
.16398 .74691 L
s
.16398 .74691 m
.16907 .7381 L
s
.16907 .7381 m
.17925 .7381 L
s
.17925 .7381 m
.18434 .74691 L
s
.18434 .74691 m
.17925 .75573 L
s
.17925 .75573 m
.18943 .75573 L
s
.18943 .75573 m
.19452 .76455 L
s
.19452 .76455 m
.18943 .77337 L
s
.18943 .77337 m
.17925 .77337 L
s
.17925 .77337 m
.18434 .78219 L
s
.18434 .78219 m
.17925 .79101 L
s
.17925 .79101 m
.18943 .79101 L
s
.18943 .79101 m
.19452 .79982 L
s
.19452 .79982 m
.19961 .79101 L
s
.19961 .79101 m
.2098 .79101 L
s
.2098 .79101 m
.21489 .79982 L
s
.21489 .79982 m
.2098 .80864 L
s
.2098 .80864 m
.21998 .80864 L
s
.21998 .80864 m
.22507 .81746 L
s
.22507 .81746 m
.23016 .80864 L
s
.23016 .80864 m
.24035 .80864 L
s
.24035 .80864 m
.23525 .79982 L
s
.23525 .79982 m
.24035 .79101 L
s
.24035 .79101 m
.25053 .79101 L
s
.25053 .79101 m
.25562 .79982 L
s
.25562 .79982 m
.26071 .79101 L
s
.26071 .79101 m
.27089 .79101 L
s
.27089 .79101 m
.2658 .78219 L
s
.2658 .78219 m
.27089 .77337 L
s
.27089 .77337 m
.26071 .77337 L
s
.26071 .77337 m
.25562 .76455 L
s
.25562 .76455 m
.26071 .75573 L
s
.26071 .75573 m
.27089 .75573 L
s
.27089 .75573 m
.2658 .74691 L
s
.2658 .74691 m
.27089 .7381 L
s
.27089 .7381 m
.28108 .7381 L
s
.28108 .7381 m
.28617 .74691 L
s
.28617 .74691 m
.29126 .7381 L
s
.29126 .7381 m
.30144 .7381 L
s
.30144 .7381 m
.30653 .74691 L
s
.30653 .74691 m
.30144 .75573 L
s
.30144 .75573 m
.31162 .75573 L
s
.31162 .75573 m
.31671 .76455 L
s
.31671 .76455 m
.32181 .75573 L
s
.32181 .75573 m
.33199 .75573 L
s
.33199 .75573 m
.3269 .74691 L
s
.3269 .74691 m
.33199 .7381 L
s
.33199 .7381 m
.34217 .7381 L
s
.34217 .7381 m
.34726 .74691 L
s
.34726 .74691 m
.35235 .7381 L
s
.35235 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.36763 .74691 L
s
.36763 .74691 m
.36254 .75573 L
s
.36254 .75573 m
.37272 .75573 L
s
.37272 .75573 m
.37781 .76455 L
s
.37781 .76455 m
.37272 .77337 L
s
.37272 .77337 m
.36254 .77337 L
s
.36254 .77337 m
.36763 .78219 L
s
.36763 .78219 m
.36254 .79101 L
s
.36254 .79101 m
.37272 .79101 L
s
.37272 .79101 m
.37781 .79982 L
s
.37781 .79982 m
.3829 .79101 L
s
.3829 .79101 m
.39308 .79101 L
s
.39308 .79101 m
.39817 .79982 L
s
.39817 .79982 m
.39308 .80864 L
s
.39308 .80864 m
.40327 .80864 L
s
.40327 .80864 m
.40836 .81746 L
s
.40836 .81746 m
.40327 .82628 L
s
.40327 .82628 m
.39308 .82628 L
s
.39308 .82628 m
.39817 .8351 L
s
.39817 .8351 m
.39308 .84392 L
s
.39308 .84392 m
.3829 .84392 L
s
.3829 .84392 m
.37781 .8351 L
s
.37781 .8351 m
.37272 .84392 L
s
.37272 .84392 m
.36254 .84392 L
s
.36254 .84392 m
.36763 .85273 L
s
.36763 .85273 m
.36254 .86155 L
s
.36254 .86155 m
.37272 .86155 L
s
.37272 .86155 m
.37781 .87037 L
s
.37781 .87037 m
.37272 .87919 L
s
.37272 .87919 m
.36254 .87919 L
s
.36254 .87919 m
.36763 .88801 L
s
.36763 .88801 m
.36254 .89683 L
s
.36254 .89683 m
.37272 .89683 L
s
.37272 .89683 m
.37781 .90564 L
s
.37781 .90564 m
.3829 .89683 L
s
.3829 .89683 m
.39308 .89683 L
s
.39308 .89683 m
.39817 .90564 L
s
.39817 .90564 m
.39308 .91446 L
s
.39308 .91446 m
.40327 .91446 L
s
.40327 .91446 m
.40836 .92328 L
s
.40836 .92328 m
.41345 .91446 L
s
.41345 .91446 m
.42363 .91446 L
s
.42363 .91446 m
.41854 .90564 L
s
.41854 .90564 m
.42363 .89683 L
s
.42363 .89683 m
.43381 .89683 L
s
.43381 .89683 m
.4389 .90564 L
s
.4389 .90564 m
.444 .89683 L
s
.444 .89683 m
.45418 .89683 L
s
.45418 .89683 m
.45927 .90564 L
s
.45927 .90564 m
.45418 .91446 L
s
.45418 .91446 m
.46436 .91446 L
s
.46436 .91446 m
.46945 .92328 L
s
.46945 .92328 m
.46436 .9321 L
s
.46436 .9321 m
.45418 .9321 L
s
.45418 .9321 m
.45927 .94092 L
s
.45927 .94092 m
.45418 .94974 L
s
.45418 .94974 m
.46436 .94974 L
s
.46436 .94974 m
.46945 .95855 L
s
.46945 .95855 m
.47454 .94974 L
s
.47454 .94974 m
.48473 .94974 L
s
.48473 .94974 m
.48982 .95855 L
s
.48982 .95855 m
.48473 .96737 L
s
.48473 .96737 m
.49491 .96737 L
s
.49491 .96737 m
.5 .97619 L
s
.5 .97619 m
.50509 .96737 L
s
.50509 .96737 m
.51527 .96737 L
s
.51527 .96737 m
.51018 .95855 L
s
.51018 .95855 m
.51527 .94974 L
s
.51527 .94974 m
.52546 .94974 L
s
.52546 .94974 m
.53055 .95855 L
s
.53055 .95855 m
.53564 .94974 L
s
.53564 .94974 m
.54582 .94974 L
s
.54582 .94974 m
.54073 .94092 L
s
.54073 .94092 m
.54582 .9321 L
s
.54582 .9321 m
.53564 .9321 L
s
.53564 .9321 m
.53055 .92328 L
s
.53055 .92328 m
.53564 .91446 L
s
.53564 .91446 m
.54582 .91446 L
s
.54582 .91446 m
.54073 .90564 L
s
.54073 .90564 m
.54582 .89683 L
s
.54582 .89683 m
.556 .89683 L
s
.556 .89683 m
.5611 .90564 L
s
.5611 .90564 m
.56619 .89683 L
s
.56619 .89683 m
.57637 .89683 L
s
.57637 .89683 m
.58146 .90564 L
s
.58146 .90564 m
.57637 .91446 L
s
.57637 .91446 m
.58655 .91446 L
s
.58655 .91446 m
.59164 .92328 L
s
.59164 .92328 m
.59673 .91446 L
s
.59673 .91446 m
.60692 .91446 L
s
.60692 .91446 m
.60183 .90564 L
s
.60183 .90564 m
.60692 .89683 L
s
.60692 .89683 m
.6171 .89683 L
s
.6171 .89683 m
.62219 .90564 L
s
.62219 .90564 m
.62728 .89683 L
s
.62728 .89683 m
.63746 .89683 L
s
.63746 .89683 m
.63237 .88801 L
s
.63237 .88801 m
.63746 .87919 L
s
.63746 .87919 m
.62728 .87919 L
s
.62728 .87919 m
.62219 .87037 L
s
.62219 .87037 m
.62728 .86155 L
s
.62728 .86155 m
.63746 .86155 L
s
.63746 .86155 m
.63237 .85273 L
s
.63237 .85273 m
.63746 .84392 L
s
.63746 .84392 m
.62728 .84392 L
s
.62728 .84392 m
.62219 .8351 L
s
.62219 .8351 m
.6171 .84392 L
s
.6171 .84392 m
.60692 .84392 L
s
.60692 .84392 m
.60183 .8351 L
s
.60183 .8351 m
.60692 .82628 L
s
.60692 .82628 m
.59673 .82628 L
s
.59673 .82628 m
.59164 .81746 L
s
.59164 .81746 m
.59673 .80864 L
s
.59673 .80864 m
.60692 .80864 L
s
.60692 .80864 m
.60183 .79982 L
s
.60183 .79982 m
.60692 .79101 L
s
.60692 .79101 m
.6171 .79101 L
s
.6171 .79101 m
.62219 .79982 L
s
.62219 .79982 m
.62728 .79101 L
s
.62728 .79101 m
.63746 .79101 L
s
.63746 .79101 m
.63237 .78219 L
s
.63237 .78219 m
.63746 .77337 L
s
.63746 .77337 m
.62728 .77337 L
s
.62728 .77337 m
.62219 .76455 L
s
.62219 .76455 m
.62728 .75573 L
s
.62728 .75573 m
.63746 .75573 L
s
.63746 .75573 m
.63237 .74691 L
s
.63237 .74691 m
.63746 .7381 L
s
.63746 .7381 m
.64765 .7381 L
s
.64765 .7381 m
.65274 .74691 L
s
.65274 .74691 m
.65783 .7381 L
s
.65783 .7381 m
.66801 .7381 L
s
.66801 .7381 m
.6731 .74691 L
s
.6731 .74691 m
.66801 .75573 L
s
.66801 .75573 m
.67819 .75573 L
s
.67819 .75573 m
.68329 .76455 L
s
.68329 .76455 m
.68838 .75573 L
s
.68838 .75573 m
.69856 .75573 L
s
.69856 .75573 m
.69347 .74691 L
s
.69347 .74691 m
.69856 .7381 L
s
.69856 .7381 m
.70874 .7381 L
s
.70874 .7381 m
.71383 .74691 L
s
.71383 .74691 m
.71892 .7381 L
s
.71892 .7381 m
.72911 .7381 L
s
.72911 .7381 m
.7342 .74691 L
s
.7342 .74691 m
.72911 .75573 L
s
.72911 .75573 m
.73929 .75573 L
s
.73929 .75573 m
.74438 .76455 L
s
.74438 .76455 m
.73929 .77337 L
s
.73929 .77337 m
.72911 .77337 L
s
.72911 .77337 m
.7342 .78219 L
s
.7342 .78219 m
.72911 .79101 L
s
.72911 .79101 m
.73929 .79101 L
s
.73929 .79101 m
.74438 .79982 L
s
.74438 .79982 m
.74947 .79101 L
s
.74947 .79101 m
.75965 .79101 L
s
.75965 .79101 m
.76475 .79982 L
s
.76475 .79982 m
.75965 .80864 L
s
.75965 .80864 m
.76984 .80864 L
s
.76984 .80864 m
.77493 .81746 L
s
.77493 .81746 m
.78002 .80864 L
s
.78002 .80864 m
.7902 .80864 L
s
.7902 .80864 m
.78511 .79982 L
s
.78511 .79982 m
.7902 .79101 L
s
.7902 .79101 m
.80039 .79101 L
s
.80039 .79101 m
.80548 .79982 L
s
.80548 .79982 m
.81057 .79101 L
s
.81057 .79101 m
.82075 .79101 L
s
.82075 .79101 m
.81566 .78219 L
s
.81566 .78219 m
.82075 .77337 L
s
.82075 .77337 m
.81057 .77337 L
s
.81057 .77337 m
.80548 .76455 L
s
.80548 .76455 m
.81057 .75573 L
s
.81057 .75573 m
.82075 .75573 L
s
.82075 .75573 m
.81566 .74691 L
s
.81566 .74691 m
.82075 .7381 L
s
.82075 .7381 m
.83093 .7381 L
s
.83093 .7381 m
.83602 .74691 L
s
.83602 .74691 m
.84112 .7381 L
s
.84112 .7381 m
.8513 .7381 L
s
.8513 .7381 m
.85639 .74691 L
s
.85639 .74691 m
.8513 .75573 L
s
.8513 .75573 m
.86148 .75573 L
s
.86148 .75573 m
.86657 .76455 L
s
.86657 .76455 m
.87166 .75573 L
s
.87166 .75573 m
.88185 .75573 L
s
.88185 .75573 m
.87675 .74691 L
s
.87675 .74691 m
.88185 .7381 L
s
.88185 .7381 m
.89203 .7381 L
s
.89203 .7381 m
.89712 .74691 L
s
.89712 .74691 m
.90221 .7381 L
s
.90221 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.9073 .72928 L
s
.9073 .72928 m
.91239 .72046 L
s
.91239 .72046 m
.90221 .72046 L
s
.90221 .72046 m
.89712 .71164 L
s
.89712 .71164 m
.90221 .70282 L
s
.90221 .70282 m
.91239 .70282 L
s
.91239 .70282 m
.9073 .694 L
s
.9073 .694 m
.91239 .68519 L
s
.91239 .68519 m
.90221 .68519 L
s
.90221 .68519 m
.89712 .67637 L
s
.89712 .67637 m
.89203 .68519 L
s
.89203 .68519 m
.88185 .68519 L
s
.88185 .68519 m
.87675 .67637 L
s
.87675 .67637 m
.88185 .66755 L
s
.88185 .66755 m
.87166 .66755 L
s
.87166 .66755 m
.86657 .65873 L
s
.86657 .65873 m
.87166 .64991 L
s
.87166 .64991 m
.88185 .64991 L
s
.88185 .64991 m
.87675 .64109 L
s
.87675 .64109 m
.88185 .63228 L
s
.88185 .63228 m
.89203 .63228 L
s
.89203 .63228 m
.89712 .64109 L
s
.89712 .64109 m
.90221 .63228 L
s
.90221 .63228 m
.91239 .63228 L
s
.91239 .63228 m
.9073 .62346 L
s
.9073 .62346 m
.91239 .61464 L
s
.91239 .61464 m
.90221 .61464 L
s
.90221 .61464 m
.89712 .60582 L
s
.89712 .60582 m
.90221 .597 L
s
.90221 .597 m
.91239 .597 L
s
.91239 .597 m
.9073 .58818 L
s
.9073 .58818 m
.91239 .57937 L
s
.91239 .57937 m
.90221 .57937 L
s
.90221 .57937 m
.89712 .57055 L
s
.89712 .57055 m
.89203 .57937 L
s
.89203 .57937 m
.88185 .57937 L
s
.88185 .57937 m
.87675 .57055 L
s
.87675 .57055 m
.88185 .56173 L
s
.88185 .56173 m
.87166 .56173 L
s
.87166 .56173 m
.86657 .55291 L
s
.86657 .55291 m
.86148 .56173 L
s
.86148 .56173 m
.8513 .56173 L
s
.8513 .56173 m
.85639 .57055 L
s
.85639 .57055 m
.8513 .57937 L
s
.8513 .57937 m
.84112 .57937 L
s
.84112 .57937 m
.83602 .57055 L
s
.83602 .57055 m
.83093 .57937 L
s
.83093 .57937 m
.82075 .57937 L
s
.82075 .57937 m
.81566 .57055 L
s
.81566 .57055 m
.82075 .56173 L
s
.82075 .56173 m
.81057 .56173 L
s
.81057 .56173 m
.80548 .55291 L
s
.80548 .55291 m
.81057 .54409 L
s
.81057 .54409 m
.82075 .54409 L
s
.82075 .54409 m
.81566 .53527 L
s
.81566 .53527 m
.82075 .52646 L
s
.82075 .52646 m
.81057 .52646 L
s
.81057 .52646 m
.80548 .51764 L
s
.80548 .51764 m
.80039 .52646 L
s
.80039 .52646 m
.7902 .52646 L
s
.7902 .52646 m
.78511 .51764 L
s
.78511 .51764 m
.7902 .50882 L
s
.7902 .50882 m
.78002 .50882 L
s
.78002 .50882 m
.77493 .5 L
s
.77493 .5 m
.78002 .49118 L
s
.78002 .49118 m
.7902 .49118 L
s
.7902 .49118 m
.78511 .48236 L
s
.78511 .48236 m
.7902 .47354 L
s
.7902 .47354 m
.80039 .47354 L
s
.80039 .47354 m
.80548 .48236 L
s
.80548 .48236 m
.81057 .47354 L
s
.81057 .47354 m
.82075 .47354 L
s
.82075 .47354 m
.81566 .46473 L
s
.81566 .46473 m
.82075 .45591 L
s
.82075 .45591 m
.81057 .45591 L
s
.81057 .45591 m
.80548 .44709 L
s
.80548 .44709 m
.81057 .43827 L
s
.81057 .43827 m
.82075 .43827 L
s
.82075 .43827 m
.81566 .42945 L
s
.81566 .42945 m
.82075 .42063 L
s
.82075 .42063 m
.83093 .42063 L
s
.83093 .42063 m
.83602 .42945 L
s
.83602 .42945 m
.84112 .42063 L
s
.84112 .42063 m
.8513 .42063 L
s
.8513 .42063 m
.85639 .42945 L
s
.85639 .42945 m
.8513 .43827 L
s
.8513 .43827 m
.86148 .43827 L
s
.86148 .43827 m
.86657 .44709 L
s
.86657 .44709 m
.87166 .43827 L
s
.87166 .43827 m
.88185 .43827 L
s
.88185 .43827 m
.87675 .42945 L
s
.87675 .42945 m
.88185 .42063 L
s
.88185 .42063 m
.89203 .42063 L
s
.89203 .42063 m
.89712 .42945 L
s
.89712 .42945 m
.90221 .42063 L
s
.90221 .42063 m
.91239 .42063 L
s
.91239 .42063 m
.9073 .41182 L
s
.9073 .41182 m
.91239 .403 L
s
.91239 .403 m
.90221 .403 L
s
.90221 .403 m
.89712 .39418 L
s
.89712 .39418 m
.90221 .38536 L
s
.90221 .38536 m
.91239 .38536 L
s
.91239 .38536 m
.9073 .37654 L
s
.9073 .37654 m
.91239 .36772 L
s
.91239 .36772 m
.90221 .36772 L
s
.90221 .36772 m
.89712 .35891 L
s
.89712 .35891 m
.89203 .36772 L
s
.89203 .36772 m
.88185 .36772 L
s
.88185 .36772 m
.87675 .35891 L
s
.87675 .35891 m
.88185 .35009 L
s
.88185 .35009 m
.87166 .35009 L
s
.87166 .35009 m
.86657 .34127 L
s
.86657 .34127 m
.87166 .33245 L
s
.87166 .33245 m
.88185 .33245 L
s
.88185 .33245 m
.87675 .32363 L
s
.87675 .32363 m
.88185 .31481 L
s
.88185 .31481 m
.89203 .31481 L
s
.89203 .31481 m
.89712 .32363 L
s
.89712 .32363 m
.90221 .31481 L
s
.90221 .31481 m
.91239 .31481 L
s
.91239 .31481 m
.9073 .306 L
s
.9073 .306 m
.91239 .29718 L
s
.91239 .29718 m
.90221 .29718 L
s
.90221 .29718 m
.89712 .28836 L
s
.89712 .28836 m
.90221 .27954 L
s
.90221 .27954 m
.91239 .27954 L
s
.91239 .27954 m
.9073 .27072 L
s
.9073 .27072 m
.91239 .2619 L
s
.91239 .2619 m
.90221 .2619 L
s
.90221 .2619 m
.89712 .25309 L
s
.89712 .25309 m
.89203 .2619 L
s
.89203 .2619 m
.88185 .2619 L
s
.88185 .2619 m
.87675 .25309 L
s
.87675 .25309 m
.88185 .24427 L
s
.88185 .24427 m
.87166 .24427 L
s
.87166 .24427 m
.86657 .23545 L
s
.86657 .23545 m
.86148 .24427 L
s
.86148 .24427 m
.8513 .24427 L
s
.8513 .24427 m
.85639 .25309 L
s
.85639 .25309 m
.8513 .2619 L
s
.8513 .2619 m
.84112 .2619 L
s
.84112 .2619 m
.83602 .25309 L
s
.83602 .25309 m
.83093 .2619 L
s
.83093 .2619 m
.82075 .2619 L
s
.82075 .2619 m
.81566 .25309 L
s
.81566 .25309 m
.82075 .24427 L
s
.82075 .24427 m
.81057 .24427 L
s
.81057 .24427 m
.80548 .23545 L
s
.80548 .23545 m
.81057 .22663 L
s
.81057 .22663 m
.82075 .22663 L
s
.82075 .22663 m
.81566 .21781 L
s
.81566 .21781 m
.82075 .20899 L
s
.82075 .20899 m
.81057 .20899 L
s
.81057 .20899 m
.80548 .20018 L
s
.80548 .20018 m
.80039 .20899 L
s
.80039 .20899 m
.7902 .20899 L
s
.7902 .20899 m
.78511 .20018 L
s
.78511 .20018 m
.7902 .19136 L
s
.7902 .19136 m
.78002 .19136 L
s
.78002 .19136 m
.77493 .18254 L
s
.77493 .18254 m
.76984 .19136 L
s
.76984 .19136 m
.75965 .19136 L
s
.75965 .19136 m
.76475 .20018 L
s
.76475 .20018 m
.75965 .20899 L
s
.75965 .20899 m
.74947 .20899 L
s
.74947 .20899 m
.74438 .20018 L
s
.74438 .20018 m
.73929 .20899 L
s
.73929 .20899 m
.72911 .20899 L
s
.72911 .20899 m
.7342 .21781 L
s
.7342 .21781 m
.72911 .22663 L
s
.72911 .22663 m
.73929 .22663 L
s
.73929 .22663 m
.74438 .23545 L
s
.74438 .23545 m
.73929 .24427 L
s
.73929 .24427 m
.72911 .24427 L
s
.72911 .24427 m
.7342 .25309 L
s
.7342 .25309 m
.72911 .2619 L
s
.72911 .2619 m
.71892 .2619 L
s
.71892 .2619 m
.71383 .25309 L
s
.71383 .25309 m
.70874 .2619 L
s
.70874 .2619 m
.69856 .2619 L
s
.69856 .2619 m
.69347 .25309 L
s
.69347 .25309 m
.69856 .24427 L
s
.69856 .24427 m
.68838 .24427 L
s
.68838 .24427 m
.68329 .23545 L
s
.68329 .23545 m
.67819 .24427 L
s
.67819 .24427 m
.66801 .24427 L
s
.66801 .24427 m
.6731 .25309 L
s
.6731 .25309 m
.66801 .2619 L
s
.66801 .2619 m
.65783 .2619 L
s
.65783 .2619 m
.65274 .25309 L
s
.65274 .25309 m
.64765 .2619 L
s
.64765 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.63237 .25309 L
s
.63237 .25309 m
.63746 .24427 L
s
.63746 .24427 m
.62728 .24427 L
s
.62728 .24427 m
.62219 .23545 L
s
.62219 .23545 m
.62728 .22663 L
s
.62728 .22663 m
.63746 .22663 L
s
.63746 .22663 m
.63237 .21781 L
s
.63237 .21781 m
.63746 .20899 L
s
.63746 .20899 m
.62728 .20899 L
s
.62728 .20899 m
.62219 .20018 L
s
.62219 .20018 m
.6171 .20899 L
s
.6171 .20899 m
.60692 .20899 L
s
.60692 .20899 m
.60183 .20018 L
s
.60183 .20018 m
.60692 .19136 L
s
.60692 .19136 m
.59673 .19136 L
s
.59673 .19136 m
.59164 .18254 L
s
.59164 .18254 m
.59673 .17372 L
s
.59673 .17372 m
.60692 .17372 L
s
.60692 .17372 m
.60183 .1649 L
s
.60183 .1649 m
.60692 .15608 L
s
.60692 .15608 m
.6171 .15608 L
s
.6171 .15608 m
.62219 .1649 L
s
.62219 .1649 m
.62728 .15608 L
s
.62728 .15608 m
.63746 .15608 L
s
.63746 .15608 m
.63237 .14727 L
s
.63237 .14727 m
.63746 .13845 L
s
.63746 .13845 m
.62728 .13845 L
s
.62728 .13845 m
.62219 .12963 L
s
.62219 .12963 m
.62728 .12081 L
s
.62728 .12081 m
.63746 .12081 L
s
.63746 .12081 m
.63237 .11199 L
s
.63237 .11199 m
.63746 .10317 L
s
.63746 .10317 m
.62728 .10317 L
s
.62728 .10317 m
.62219 .09436 L
s
.62219 .09436 m
.6171 .10317 L
s
.6171 .10317 m
.60692 .10317 L
s
.60692 .10317 m
.60183 .09436 L
s
.60183 .09436 m
.60692 .08554 L
s
.60692 .08554 m
.59673 .08554 L
s
.59673 .08554 m
.59164 .07672 L
s
.59164 .07672 m
.58655 .08554 L
s
.58655 .08554 m
.57637 .08554 L
s
.57637 .08554 m
.58146 .09436 L
s
.58146 .09436 m
.57637 .10317 L
s
.57637 .10317 m
.56619 .10317 L
s
.56619 .10317 m
.5611 .09436 L
s
.5611 .09436 m
.556 .10317 L
s
.556 .10317 m
.54582 .10317 L
s
.54582 .10317 m
.54073 .09436 L
s
.54073 .09436 m
.54582 .08554 L
s
.54582 .08554 m
.53564 .08554 L
s
.53564 .08554 m
.53055 .07672 L
s
.53055 .07672 m
.53564 .0679 L
s
.53564 .0679 m
.54582 .0679 L
s
.54582 .0679 m
.54073 .05908 L
s
.54073 .05908 m
.54582 .05026 L
s
.54582 .05026 m
.53564 .05026 L
s
.53564 .05026 m
.53055 .04145 L
s
.53055 .04145 m
.52546 .05026 L
s
.52546 .05026 m
.51527 .05026 L
s
.51527 .05026 m
.51018 .04145 L
s
.51018 .04145 m
.51527 .03263 L
s
.51527 .03263 m
.50509 .03263 L
s
.50509 .03263 m
.5 .02381 L
s
.5 .02381 m
.49491 .03263 L
s
.49491 .03263 m
.48473 .03263 L
s
.48473 .03263 m
.48982 .04145 L
s
.48982 .04145 m
.48473 .05026 L
s
.48473 .05026 m
.47454 .05026 L
s
.47454 .05026 m
.46945 .04145 L
s
.46945 .04145 m
.46436 .05026 L
s
.46436 .05026 m
.45418 .05026 L
s
.45418 .05026 m
.45927 .05908 L
s
.45927 .05908 m
.45418 .0679 L
s
.45418 .0679 m
.46436 .0679 L
s
.46436 .0679 m
.46945 .07672 L
s
.46945 .07672 m
.46436 .08554 L
s
.46436 .08554 m
.45418 .08554 L
s
.45418 .08554 m
.45927 .09436 L
s
.45927 .09436 m
.45418 .10317 L
s
.45418 .10317 m
.444 .10317 L
s
.444 .10317 m
.4389 .09436 L
s
.4389 .09436 m
.43381 .10317 L
s
.43381 .10317 m
.42363 .10317 L
s
.42363 .10317 m
.41854 .09436 L
s
.41854 .09436 m
.42363 .08554 L
s
.42363 .08554 m
.41345 .08554 L
s
.41345 .08554 m
.40836 .07672 L
s
.40836 .07672 m
.40327 .08554 L
s
.40327 .08554 m
.39308 .08554 L
s
.39308 .08554 m
.39817 .09436 L
s
.39817 .09436 m
.39308 .10317 L
s
.39308 .10317 m
.3829 .10317 L
s
.3829 .10317 m
.37781 .09436 L
s
.37781 .09436 m
.37272 .10317 L
s
.37272 .10317 m
.36254 .10317 L
s
.36254 .10317 m
.36763 .11199 L
s
.36763 .11199 m
.36254 .12081 L
s
.36254 .12081 m
.37272 .12081 L
s
.37272 .12081 m
.37781 .12963 L
s
.37781 .12963 m
.37272 .13845 L
s
.37272 .13845 m
.36254 .13845 L
s
.36254 .13845 m
.36763 .14727 L
s
.36763 .14727 m
.36254 .15608 L
s
.36254 .15608 m
.37272 .15608 L
s
.37272 .15608 m
.37781 .1649 L
s
.37781 .1649 m
.3829 .15608 L
s
.3829 .15608 m
.39308 .15608 L
s
.39308 .15608 m
.39817 .1649 L
s
.39817 .1649 m
.39308 .17372 L
s
.39308 .17372 m
.40327 .17372 L
s
.40327 .17372 m
.40836 .18254 L
s
.40836 .18254 m
.40327 .19136 L
s
.40327 .19136 m
.39308 .19136 L
s
.39308 .19136 m
.39817 .20018 L
s
.39817 .20018 m
.39308 .20899 L
s
.39308 .20899 m
.3829 .20899 L
s
.3829 .20899 m
.37781 .20018 L
s
.37781 .20018 m
.37272 .20899 L
s
.37272 .20899 m
.36254 .20899 L
s
.36254 .20899 m
.36763 .21781 L
s
.36763 .21781 m
.36254 .22663 L
s
.36254 .22663 m
.37272 .22663 L
s
.37272 .22663 m
.37781 .23545 L
s
.37781 .23545 m
.37272 .24427 L
s
.37272 .24427 m
.36254 .24427 L
s
.36254 .24427 m
.36763 .25309 L
s
.36763 .25309 m
.36254 .2619 L
s
.36254 .2619 m
.35235 .2619 L
s
.35235 .2619 m
.34726 .25309 L
s
.34726 .25309 m
.34217 .2619 L
s
.34217 .2619 m
.33199 .2619 L
s
.33199 .2619 m
.3269 .25309 L
s
.3269 .25309 m
.33199 .24427 L
s
.33199 .24427 m
.32181 .24427 L
s
.32181 .24427 m
.31671 .23545 L
s
.31671 .23545 m
.31162 .24427 L
s
.31162 .24427 m
.30144 .24427 L
s
.30144 .24427 m
.30653 .25309 L
s
.30653 .25309 m
.30144 .2619 L
s
.30144 .2619 m
.29126 .2619 L
s
.29126 .2619 m
.28617 .25309 L
s
.28617 .25309 m
.28108 .2619 L
s
.28108 .2619 m
.27089 .2619 L
s
.27089 .2619 m
.2658 .25309 L
s
.2658 .25309 m
.27089 .24427 L
s
.27089 .24427 m
.26071 .24427 L
s
.26071 .24427 m
.25562 .23545 L
s
.25562 .23545 m
.26071 .22663 L
s
.26071 .22663 m
.27089 .22663 L
s
.27089 .22663 m
.2658 .21781 L
s
.2658 .21781 m
.27089 .20899 L
s
.27089 .20899 m
.26071 .20899 L
s
.26071 .20899 m
.25562 .20018 L
s
.25562 .20018 m
.25053 .20899 L
s
.25053 .20899 m
.24035 .20899 L
s
.24035 .20899 m
.23525 .20018 L
s
.23525 .20018 m
.24035 .19136 L
s
.24035 .19136 m
.23016 .19136 L
s
.23016 .19136 m
.22507 .18254 L
s
.22507 .18254 m
.21998 .19136 L
s
.21998 .19136 m
.2098 .19136 L
s
.2098 .19136 m
.21489 .20018 L
s
.21489 .20018 m
.2098 .20899 L
s
.2098 .20899 m
.19961 .20899 L
s
.19961 .20899 m
.19452 .20018 L
s
.19452 .20018 m
.18943 .20899 L
s
.18943 .20899 m
.17925 .20899 L
s
.17925 .20899 m
.18434 .21781 L
s
.18434 .21781 m
.17925 .22663 L
s
.17925 .22663 m
.18943 .22663 L
s
.18943 .22663 m
.19452 .23545 L
s
.19452 .23545 m
.18943 .24427 L
s
.18943 .24427 m
.17925 .24427 L
s
.17925 .24427 m
.18434 .25309 L
s
.18434 .25309 m
.17925 .2619 L
s
.17925 .2619 m
.16907 .2619 L
s
.16907 .2619 m
.16398 .25309 L
s
.16398 .25309 m
.15888 .2619 L
s
.15888 .2619 m
.1487 .2619 L
s
.1487 .2619 m
.14361 .25309 L
s
.14361 .25309 m
.1487 .24427 L
s
.1487 .24427 m
.13852 .24427 L
s
.13852 .24427 m
.13343 .23545 L
s
.13343 .23545 m
.12834 .24427 L
s
.12834 .24427 m
.11815 .24427 L
s
.11815 .24427 m
.12325 .25309 L
s
.12325 .25309 m
.11815 .2619 L
s
.11815 .2619 m
.10797 .2619 L
s
.10797 .2619 m
.10288 .25309 L
s
.10288 .25309 m
.09779 .2619 L
s
.09779 .2619 m
.08761 .2619 L
s
newpath
.5 .5 .47619 0 365.73 arc
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{361.313, 198.625},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008M0001:B000`400?l00000o`00003oo`3ooooo
0?oooal0oooo003o0?ooool0oooo7`3oool00?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo
003o0?ooool0oooo7`3oool00?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo003o0?ooool0
oooo7`3oool00?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo003o0?ooool0oooo7`3oool0
0?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo00270?ooo`030000003oool0oooo0?l0oooo
U03oool008L0oooo0P00003o0?oooiD0oooo00260?ooo`030000003oool000000?l0ooooU@3oool0
08<0oooo100000020?ooo`@00000o`3ooonA0?ooo`00P`3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0?l0ooooT03oool007l0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3ooon;0?ooo`00OP3o
ool2000000@0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`3oool200000?l0oooo
S@3oool007h0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`050?ooo`070000003o
ool0oooo0?ooo`000000oooo0000003o0?oooh`0oooo001j0?ooo`@000000P3oool4000000P0oooo
0`0000020?ooo`@00000o`3ooon90?ooo`00N`3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0?l0ooooR03oool007`0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o
0?ooohP0oooo001k0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0o`3ooon80?oo
o`00N`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0?l0ooooQ`3oool007X0oooo
1000000C0?ooo`@00000o`3ooon90?ooo`00OP3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0?l0ooooR`3oool006d0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0o`3ooomj0?ooo`00K@3oool20000
00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`800000o`3ooomk0?oo
o`00K03oool00`000000oooo0000000?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool0303oool00`000000oooo0000003o0?ooog/0oooo001Y0?ooo`@000000P3oool4000000L0oooo
1000000C0?ooo`@00000203oool400000080oooo0`00003[0?ooo`P00000Q03oool006T0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0=h0oooo3000
00030?ooo`<000000P3oool<000007P0oooo001U0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo1@00000?
0?ooo`030000003oool0000000h0oooo1P00001b0?ooo`00I03oool2000000@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`020?ooo`800000103oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo0P00003?0?ooo`H000004@3oool300000080oooo0`00000B0?oo
o`D00000K@3oool006@0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`050?ooo`07
0000003oool0oooo0?ooo`000000oooo000000030?ooo`030000003oool0oooo01H0oooo00L00000
0?ooo`3oool0oooo0000003oool000000080oooo00<000000?ooo`3oool01@3oool01`000000oooo
0?ooo`3oool000000?ooo`000000b`3oool3000001L0oooo00<000000?ooo`0000000`3oool00`00
0000oooo0?ooo`0F0?ooo`<00000JP3oool00600oooo100000020?ooo`@00000203oool400000003
0?ooo`0000000000008000006@3oool400000080oooo100000070?ooo`@000000P3oool400000<D0
oooo0`00000F0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool2000000<0
oooo00<000000?ooo`3oool0503oool3000006L0oooo001Q0?ooo`030000003oool0oooo04T0oooo
00<000000?ooo`3oool0`P3oool2000001P0oooo0P0000030?ooo`800000103oool200000080oooo
0P00000I0?ooo`<00000I03oool00680oooo00<000000?ooo`3oool0B03oool00`000000oooo0?oo
o`2m0?ooo`D000005`3oool9000000H0oooo2@00000H0?ooo`D00000G`3oool00640oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`2k0?ooo`8000007@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0L
0?ooo`030000003oool0oooo05`0oooo001Q0?ooo`030000003oool0oooo04X0oooo00<000000?oo
o`3oool0^03oool2000001l0oooo0P0000040?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0M0?ooo`800000G03oool00600oooo100000170?ooo`@0
0000^03oool200000240oooo0P0000030?ooo`030000003oool0oooo00P0oooo0P0000020?ooo`80
00008@3oool3000005T0oooo001T0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0
]`3oool300000280oooo100000000`3oool000000000000;0?ooo`030000003oool0000000<00000
8`3oool2000005L0oooo001U0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0]03o
ool3000002P0oooo00<000000?ooo`0000003@3oool2000002P0oooo0`00001D0?ooo`00I03oool0
0`000000oooo0?ooo`130?ooo`030000003oool0oooo0;80oooo0P00000M0?ooo`030000003oool0
oooo00`0oooo0P00000=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06`3oool2
00000580oooo001T0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0/03oool00`00
0000oooo0?ooo`0L0?ooo`<00000303oool2000000h0oooo0P00000<0?ooo`<000007P3oool00`00
0000oooo0?ooo`1?0?ooo`00H03oool4000004L0oooo1000002]0?ooo`8000007`3oool00`000000
oooo0000000<0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0000000O0?ooo`800000C`3oool00640oooo00<000000?ooo`3oool0B@3oool00`000000oooo
0?ooo`2Z0?ooo`8000007P3oool4000000030?ooo`0000000000008000001P3oool400000100oooo
100000060?ooo`@0000000<0oooo0000000000000P00000N0?ooo`800000C@3oool00680oooo00<0
00000?ooo`3oool0B03oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo01l0oooo0P00
00030?ooo`800000203oool200000180oooo0P0000080?ooo`8000000`3oool200000240oooo00<0
00000?ooo`3oool0BP3oool00640oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`2W
0?ooo`8000007P3oool010000000oooo0?ooo`3oool2000000@0oooo00<000000?ooo`3oool00P3o
ool010000000oooo0?ooo`3oool2000001<0oooo00<000000?ooo`3oool00P3oool010000000oooo
0?ooo`3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`80
0000BP3oool00640oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`2T0?ooo`800000
7`3oool2000000<0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool2000000<0
oooo00<000000?ooo`3oool04P3oool200000080oooo0P0000030?ooo`030000003oool0oooo00<0
oooo00@000000?ooo`3oool0oooo0P00000O0?ooo`800000B03oool00600oooo100000020?ooo`@0
0000>`3oool400000080oooo1000002U0?ooo`030000003oool0oooo01`0oooo5`00000F0?oooaL0
00007P3oool00`000000oooo0?ooo`150?ooo`00I03oool2000000<0oooo00<000000?ooo`3oool0
>03oool00`000000oooo0?ooo`020?ooo`800000Z03oool00`000000oooo0?ooo`0M0?ooo`800000
503oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
7P3oool2000004D0oooo001U0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0=`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0:@0oooo0P00000Q0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo0200oooo0P0000130?ooo`00JP3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo0:P0oooo0P00000R0?ooo`030000003oool000000180oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0?0?ooo`800000903oool00`000000oooo0?ooo`100?ooo`00J@3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool08P3oool40000
0100oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0?0?ooo`<00000903oool00`00
0000oooo0?ooo`0o0?ooo`00J@3oool4000003H0oooo0`00002W0?ooo`030000003oool0oooo02H0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0n0?ooo`00K03oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool09`3oool2000000h0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool0?@3oool003X0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?oo
o`0a0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?oo
o`0X0?ooo`8000003@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00/0oooo0P00
000Z0?ooo`800000?@3oool003T0oooo0P00000b0?ooo`030000003oool0oooo0380oooo00<00000
0?ooo`3oool0<03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo02T0oooo00<00000
0?ooo`000000303oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00T0oooo00@00000
0?ooo`3oool00000:`3oool00`000000oooo0?ooo`0j0?ooo`00>@3oool00`000000oooo0000000`
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0000001a
0?ooo`800000:@3oool300000080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0N
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`00000Z0?ooo`030000003oool0
oooo03T0oooo000e0?ooo`@000000P3oool4000002X0oooo1000000f0?ooo`<00000:`3oool40000
00030?ooo`000000000000800000K@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool00000
00<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo
03P0oooo000f0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`0h0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`1[0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0=`3oool00340oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03L0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06@0oooo
00<000000?ooo`3oool0:`3oool2000000D0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool2000002d0oooo
00<000000?ooo`3oool0=P3oool00300oooo0P0000040?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool200000200oooo0P0000040?ooo`030000003oool0oooo03L0oooo00<0
00000?ooo`3oool00P3oool200000240oooo0P0000040?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool2000006D0oooo00<000000?ooo`3oool0;03oool9000000H0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`040?ooo`T00000;P3oool00`000000oooo0?oo
o`0e0?ooo`00<03oool00`000000oooo000000030?ooo`030000003oool0oooo00@0oooo00L00000
0?ooo`3oool0oooo0000003oool000000200oooo00<000000?ooo`0000000P3oool00`000000oooo
0?ooo`0i0?ooo`070000003oool0oooo0?ooo`000000oooo0000000O0?ooo`030000003oool00000
00<0oooo00<000000?ooo`3oool01@3oool01`000000oooo0?ooo`3oool000000?ooo`000000H`3o
ool00`000000oooo0?ooo`0`0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo02@0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool200000380oooo
00<000000?ooo`3oool0=03oool002d0oooo0`0000020?ooo`@00000203oool400000080oooo0`00
000I0?ooo`@000000P3oool4000003/0oooo100000020?ooo`@000006@3oool400000080oooo1000
00070?ooo`@000000P3oool4000005l0oooo00<000000?ooo`3oool0<P3oool010000000oooo0?oo
o`3oool2000000@0oooo00<000000?ooo`3oool09@3oool010000000oooo0?ooo`3oool2000000@0
oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0c0?ooo`00;@3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0B@3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`0g0?ooo`8000000`3oool00`000000oooo0?ooo`0V0?ooo`040000003oool0oooo0?ooo`80
0000>@3oool00`000000oooo0?ooo`0b0?ooo`00;P3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0g0?ooo`@00000
0P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0000000<00000>@3oool00`000000oooo
0?ooo`0a0?ooo`00;P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool0B@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool0F`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0000002X0oooo0`00
000m0?ooo`030000003oool0oooo0300oooo000]0?ooo`030000003oool0oooo01H0oooo00<00000
0?ooo`3oool0603oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0:P3oool2000002/0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo02l0oooo000]0?ooo`<00000503oool3000001T0oooo100000170?oo
o`@000006@3oool4000001<0oooo1000001J0?ooo`030000003oool0oooo0100oooo0`00000Z0?oo
o`800000;03oool2000002X0oooo0`00000B0?ooo`030000003oool0oooo02h0oooo000`0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`130?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1K0?ooo`03
0000003oool0oooo00l0oooo0`0000000`3oool0000000000002000002@0oooo1000000^0?ooo`@0
0000903oool4000000030?ooo`000000000000800000403oool00`000000oooo0?ooo`0]0?ooo`00
803oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0100oooo0`0000020?ooo`030000
003oool0000002D0oooo0`00000^0?ooo`80000000<0oooo0000003oool08`3oool00`000000oooo
000000030?ooo`8000004P3oool00`000000oooo0?ooo`0/0?ooo`007`3oool2000000l0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0=0?ooo`8000003`3oool00`000000oooo0?oo
o`130?ooo`030000003oool0oooo00h0oooo0P00000?0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool03@3oool2000004X0oooo00<000000?ooo`3oool04P3oool010000000oooo0?oo
o`3oool2000002H0oooo0P00000`0?ooo`8000009P3oool2000000<0oooo0P00000C0?ooo`030000
003oool0oooo02/0oooo000O0?ooo`030000003oool0000000h0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0000000l0oooo00<000000?ooo`3oool0A03o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0000000h0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0000004P0oooo00<000000?ooo`3oool03P3o
ool010000000oooo0?ooo`3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0K0?ooo`040000003oool0oooo0?ooo`800000<@3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`80
00000`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02X0oooo000K0?ooo`@00000
0P3oool4000000P0oooo0`00000D0?ooo`<00000203oool400000080oooo100000070?ooo`@00000
A`3oool4000000P0oooo100000020?ooo`<00000203oool4000001<0oooo100000080?ooo`<00000
0P3oool4000004D0oooo00<000000?ooo`3oool03@3oool2000000<0oooo00<000000?ooo`3oool0
0`3oool200000080oooo0P00000L0?ooo`8000000`3oool00`000000oooo0?ooo`0`0?ooo`040000
003oool0oooo0?ooo`8000006`3oool00`000000oooo000000020?ooo`8000001@3oool01P000000
oooo0?ooo`000000oooo00000100oooo00<000000?ooo`3oool0:@3oool001`0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo00/0oooo2@0000060?oo
o`T000005@3oool8000003@0oooo2000000F0?ooo`P000001`3oool8000000d0oooo00<000000?oo
o`3oool0:@3oool001L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`180?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0;0?ooo`80
00001@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0503oool2000000D0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0X0?ooo`00
5`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool2000000@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`020?ooo`800000
103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo0P0000030?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool00P3oool2000000@0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`800000103oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo0080oooo0P0000040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool2000003h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool2000001L0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo02P0oooo000F0?ooo`030000003o
ool0000000<0oooo00<000000?ooo`3oool0103oool01`000000oooo0?ooo`3oool000000?ooo`00
00000`3oool00`000000oooo0?ooo`0F0?ooo`070000003oool0oooo0?ooo`000000oooo00000003
0?ooo`030000003oool0oooo00@0oooo00L000000?ooo`3oool0oooo0000003oool0000000<0oooo
00<000000?ooo`3oool0BP3oool01P000000oooo0?ooo`000000oooo000000<0oooo00<000000?oo
o`3oool01@3oool01`000000oooo0?ooo`3oool000000?ooo`0000000P3oool00`000000oooo0?oo
o`0G0?ooo`060000003oool0oooo0000003oool000000`3oool00`000000oooo0?ooo`050?ooo`07
0000003oool0oooo0?ooo`000000oooo0000000l0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`0000000P3oool00`000000oooo0?ooo`090?ooo`040000003oool0oooo0?ooo`8000005P3o
ool00`000000oooo000000020?ooo`800000=P3oool010000000oooo0?ooo`3oool2000001L0oooo
0P0000030?ooo`030000003oool0oooo00P0oooo00@000000?ooo`3oool0oooo0P00000@0?ooo`03
0000003oool0oooo02L0oooo000C0?ooo`@0000000<0oooo0000000000000P0000080?ooo`@00000
0P3oool4000001P0oooo100000020?ooo`@00000203oool400000080oooo0`00001=0?ooo`@00000
0P3oool4000000L0oooo100000020?ooo`@000006@3oool400000080oooo100000070?ooo`@00000
0P3oool4000003P0oooo00<000000?ooo`3oool03P3oool4000000030?ooo`000000oooo00/0oooo
00<000000?ooo`0000000`00000E0?ooo`@0000000<0oooo0000003oool0=`3oool00`000000oooo
0?ooo`03000001H0oooo0`0000020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`000000
0`00000@0?ooo`030000003oool0oooo02H0oooo000C0?ooo`030000003oool0oooo0>D0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`0B0?ooo`8000003@3oool3000001/0oooo0`00
000i0?ooo`8000006`3oool00`000000oooo0000000=0?ooo`800000503oool00`000000oooo0?oo
o`0U0?ooo`00503oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo03P0oooo0P00000>
0?ooo`030000003oool0oooo00/0oooo0P00000=0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0T0?ooo`00503oool00`000000oooo0?ooo`3T0?oo
o`030000003oool0oooo03D0oooo00<000000?ooo`3oool0103oool3000000`0oooo00<000000?oo
o`3oool03@3oool2000000`0oooo0P00000=0?ooo`030000003oool0oooo03X0oooo00<000000?oo
o`3oool02`3oool2000000`0oooo0P00000?0?ooo`030000003oool0oooo00X0oooo0`0000060?oo
o`030000003oool0oooo02@0oooo000C0?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3o
ool0=@3oool010000000oooo0?ooo`3oool4000000030?ooo`0000000000008000001P3oool40000
0100oooo100000060?ooo`@0000000@0oooo00000000000000001`3oool4000003`0oooo10000007
0?ooo`@0000000@0oooo00000000000000001`3oool4000000l0oooo100000070?ooo`<0000000<0
oooo0000000000000P0000040?ooo`030000003oool0oooo02<0oooo000C0?ooo`@00000hP3oool4
000003D0oooo00<000000?ooo`3oool00P3oool00`000000oooo000000030?ooo`8000001`3oool0
0`000000oooo0000000B0?ooo`8000001`3oool00`000000oooo000000020?ooo`030000003oool0
000000L0oooo00<000000?ooo`000000?P3oool00`000000oooo000000070?ooo`030000003oool0
00000080oooo00<000000?ooo`0000001`3oool00`000000oooo0000000A0?ooo`030000003oool0
000000L0oooo00<000000?ooo`0000000P3oool00`000000oooo000000040?ooo`030000003oool0
oooo02<0oooo000F0?ooo`030000003oool0oooo0=l0oooo00<000000?ooo`3oool0=P3oool00`00
0000oooo0?ooo`040?ooo`8000000`3oool2000000P0oooo0P00000B0?ooo`800000203oool20000
00<0oooo00<000000?ooo`3oool01`3oool2000003h0oooo0P0000090?ooo`8000000`3oool00`00
0000oooo0?ooo`070?ooo`8000004@3oool2000000T0oooo00@000000?ooo`3oool0oooo0P000006
0?ooo`030000003oool0oooo0280oooo000G0?ooo`030000003oool0oooo0=h0oooo00<000000?oo
o`3oool0=P3oool00`000000oooo000000040?ooo`030000003oool0oooo0080oooo0P0000030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool04@3oool2000000<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00@000000?oo
o`3oool000008`3oool001L0oooo00<000000?ooo`3oool0gP3oool00`000000oooo0?ooo`0e0?oo
o`050000003oool000000?ooo`0000000P3oool2000000D0oooo00H000000?ooo`3oool000000?oo
o`0000020?ooo`800000503oool01P000000oooo0?ooo`000000oooo00000080oooo0P0000040?oo
o`8000000P3oool2000000<0oooo0P0000100?ooo`8000000P3oool2000000<0oooo0P0000040?oo
o`8000000P3oool2000000<0oooo0P00000C0?ooo`040000003oool0oooo0?ooo`8000000`3oool0
0`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0?oo
o`0Q0?ooo`005P3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo03@0oooob`00000R
0?ooo`004`3oool400000>80oooo1000000b0?ooo`<00000a`3oool200000280oooo000C0?ooo`03
0000003oool0oooo0>D0oooo00<000000?ooo`3oool0<03oool010000000oooo0?ooo`0000360?oo
o`80000000<0oooo0000003oool0803oool001@0oooo00<000000?ooo`3oool0h`3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0000000800000a@3oool2000000030?ooo`000000oooo0200
oooo000D0?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0300000<@0oooo100000000`3oool000000?ooo`0O0?ooo`004`3oool00`000000oooo0?oo
o`3U0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool00P3oool200000<80oooo0P00
00040?ooo`030000003oool0oooo01h0oooo000C0?ooo`@0000000<0oooo0000000000000P00003G
0?ooo`@000000P3oool4000002l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`30
0?ooo`8000001@3oool00`000000oooo0?ooo`0M0?ooo`005P3oool2000000@0oooo00<000000?oo
o`3oool0d`3oool00`000000oooo0?ooo`020?ooo`800000<`3oool00`000000oooo0?ooo`030?oo
o`<00000`@3oool2000000H0oooo00<000000?ooo`3oool0703oool001L0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0<03oool00`000000oooo0?ooo`020?ooo`<000000P3oool00`000000oooo0?ooo`2m0?ooo`03
0000003oool0000000<000000`3oool00`000000oooo0?ooo`0L0?ooo`00703oool00`000000oooo
0?ooo`3C0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool00P3oool00`000000oooo
000000020?ooo`030000003oool0oooo0;`0oooo0P0000020?ooo`8000001@3oool00`000000oooo
0?ooo`0K0?ooo`00703oool00`000000oooo0?ooo`3D0?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0;/0oooo00@0
00000?ooo`3oool0oooo0P0000050?ooo`030000003oool0oooo01/0oooo000K0?ooo`@00000d@3o
ool4000003H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0;X0oooo0P0000030?ooo`8000001@3oool00`000000oooo0?ooo`0K0?ooo`007`3oool0
0`000000oooo0?ooo`3=0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0;P0oooo0P0000050?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool06P3oool00200oooo00<000000?ooo`3oool0c03oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo00@0oooo2@00002i0?ooo`T000001@3oool00`00
0000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo03H0
oooo00<000000?ooo`3oool01`3oool00`000000oooo000000020?ooo`800000^03oool200000080
oooo0P0000090?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0<h0oooo
00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo
0P00002g0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0I0?ooo`006`3oool400000=40oooo1000000d0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`000000]P3oool2000000l0oooo00<000000?ooo`3oool06@3oool001`0oooo00<00000
0?ooo`3oool0d`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00d0oooo1000002d
0?ooo`@000003`3oool00`000000oooo0?ooo`0H0?ooo`005`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0=80oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool0/P3oool200000140oooo00<000000?oo
o`3oool0603oool001L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3C0?ooo`03
0000003oool0oooo0080oooo0P00000^0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool0/03oool200000180oooo00<000000?ooo`3oool0603oool001H0oooo00<000000?ooo`000000
0`3oool00`000000oooo0?ooo`3D0?ooo`070000003oool0oooo0?ooo`000000oooo0000000/0?oo
o`030000003oool0oooo0140oooo0`00002a0?ooo`8000004`3oool00`000000oooo0?ooo`0G0?oo
o`004`3oool4000000030?ooo`000000000000800000e`3oool400000080oooo1000000Y0?ooo`03
0000003oool0oooo00h0oooo100000000`3oool000000?ooo`2`0?ooo`D00000403oool00`000000
oooo0?ooo`0G0?ooo`004`3oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo02P0oooo
00<000000?ooo`3oool03P3oool00`000000oooo000000030?ooo`030000003oool0oooo0:d0oooo
00H000000?ooo`3oool000000?ooo`00000@0?ooo`030000003oool0oooo01L0oooo000D0?ooo`03
0000003oool0oooo0><0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0@0?ooo`80
00000`3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo008000004P3oool00`000000
oooo0?ooo`0F0?ooo`00503oool00`000000oooo0?ooo`3T0?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0F0?ooo`004`3oool00`000000oooo0?oo
o`3U0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0000
00020?ooo`8000001@3oool00`000000oooo0?ooo`2Z0?ooo`800000103oool200000080oooo0P00
000=0?ooo`030000003oool0oooo01H0oooo000C0?ooo`@00000hP3oool4000002L0oooo00<00000
0?ooo`3oool02@3oool8000000L0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`04
0?ooo`T000002P3oool00`000000oooo0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`3O0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo00D0oooo0P00002Y0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01D0oooo000G0?ooo`03
0000003oool0oooo0=h0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`2W0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool2000000/0oooo00<000000?ooo`3o
ool05@3oool001L0oooo00<000000?ooo`3oool0gP3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo00T0oooo0P0000030?ooo`030000003oool0oooo00P0oooo0P00002W0?ooo`030000
003oool0oooo00P0oooo00@000000?ooo`3oool0oooo0P00000<0?ooo`030000003oool0oooo01@0
oooo000F0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0:03oool00`000000oooo
0?ooo`090?ooo`<000000P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0:@0oooo
0P00000;0?ooo`030000003oool0000000<000002`3oool00`000000oooo0?ooo`0D0?ooo`004`3o
ool400000>80oooo1000000V0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`0000003@3o
ool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo00X0oooo0`00000>0?ooo`030000003o
ool0oooo01@0oooo000C0?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool09@3oool0
0`000000oooo0?ooo`0=0?ooo`8000003@3oool00`000000oooo0?ooo`2R0?ooo`8000003@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01@0oooo000D0?ooo`030000003oool0oooo
0><0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0<0?ooo`8000003`3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo00`0oooo0P00000>0?ooo`030000003oool0oooo
01@0oooo000D0?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`0;0?ooo`@000003`3oool200000:40oooo00<000000?ooo`3oool03P3oool4000000`0
oooo00<000000?ooo`3oool04`3oool001<0oooo00<000000?ooo`3oool0i@3oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`0000004@3oool00`000000oooo
0?ooo`2N0?ooo`8000004@3oool2000000d0oooo00<000000?ooo`3oool04`3oool001<0oooo1000
00000`3oool0000000000002000000P0oooo100000020?ooo`@00000]03oool400000080oooo1000
00070?ooo`@000000P3oool4000002D0oooo00<000000?ooo`3oool0303oool200000140oooo0P00
002O0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0C0?ooo`005P3oool2000000@0oooo00<000000?ooo`3oool0103oool01`000000oooo0?ooo`3o
ool000000?ooo`0000000`3oool00`000000oooo0?ooo`2a0?ooo`070000003oool0oooo0?ooo`00
0000oooo000000030?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool20000
02T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
09`0oooo0P00000C0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool04`3oool001L0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0;0?ooo`800000503oool00`000000
oooo0?ooo`2K0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0C0?ooo`00703oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0;0?oooaP00000V`3ooolH000000`0oooo
00<000000?ooo`3oool04`3oool001`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0000
00020?ooo`8000001@3oool01P000000oooo0?ooo`000000oooo00000080oooo0P00002J0?ooo`80
00000P3oool2000000<0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool20000
0140oooo00<000000?ooo`3oool04P3oool001/0oooo100000020?ooo`@00000203oool300000:l0
oooo100000080?ooo`<000000P3oool4000002d0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0080oooo0P0000030?ooo`030000003oool0oooo0080
oooo0P00002I0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0B0?ooo`007`3oool2000000l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo
0?ooo`0=0?ooo`800000<@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0000000<0oooo
0P0000070?ooo`030000003oool0000009P0oooo0P0000080?ooo`8000000`3oool2000001H0oooo
00<000000?ooo`3oool04P3oool00200oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?oo
o`0C0?ooo`@0000000<0oooo0000000000000P0000060?ooo`@00000UP3oool4000000L0oooo0`00
00000`3oool0000000000002000001D0oooo00<000000?ooo`3oool04P3oool00300oooo00<00000
0?ooo`3oool0Z`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo01H0oooo0`00000<
0?ooo`030000003oool0oooo09@0oooo0P00000;0?ooo`030000003oool0000001P0oooo00<00000
0?ooo`3oool04P3oool00300oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0o0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool02`3oool2000009@0oooo0P00000=0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool04P3oool002d0oooo0`00002_0?ooo`@0
0000?P3oool00`000000oooo0?ooo`0U0?ooo`<00000T`3oool2000002P0oooo00<000000?ooo`3o
ool04@3oool002d0oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`0m0?ooo`030000
003oool0oooo0280oooo100000000`3oool000000?ooo`2A0?ooo`H000009@3oool00`000000oooo
0?ooo`0A0?ooo`00;P3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo03`0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo000000020?ooo`800000T@3oool00`000000oooo0?oo
o`02000002H0oooo00<000000?ooo`3oool04@3oool002h0oooo00<000000?ooo`3oool0/03oool0
0`000000oooo0?ooo`0l0?ooo`030000003oool0oooo02@0oooo0P0000030?ooo`030000003oool0
oooo08h0oooo0P0000020?ooo`8000009P3oool00`000000oooo0?ooo`0A0?ooo`00;@3oool00`00
0000oooo0?ooo`2b0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0903oool00`00
0000oooo0?ooo`020?ooo`800000S`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
02@0oooo00<000000?ooo`3oool04@3oool002d0oooo0`0000020?ooo`@00000X`3oool400000080
oooo1000000m0?ooo`030000003oool0oooo02<0oooo0P0000050?ooo`030000003oool0oooo08`0
oooo0P0000040?ooo`8000009@3oool00`000000oooo0?ooo`0A0?ooo`00<03oool2000000@0oooo
00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`020?ooo`800000@@3oool00`000000oooo
0?ooo`0S0?ooo`T00000S03oool9000002D0oooo00<000000?ooo`3oool04@3oool00340oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0?`3oool00`000000oooo0?ooo`0V0?ooo`8000000`3oool2000008/0oooo00H0
00000?ooo`3oool000000?ooo`00000X0?ooo`030000003oool0oooo0140oooo000f0?ooo`030000
003oool0oooo09l0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0W0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0R03oool2000000<0oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`0A0?ooo`00=P3oool00`000000oooo0?ooo`2P0?ooo`030000003o
ool0oooo04@0oooo00<000000?ooo`3oool0:`3oool3000008T0oooo0P00000]0?ooo`030000003o
ool0oooo0140oooo000e0?ooo`@00000W@3oool4000004H0oooo00<000000?ooo`3oool0:`3oool3
000008P0oooo1000000/0?ooo`030000003oool0oooo0140oooo000i0?ooo`030000003oool0oooo
09T0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0]0?ooo`800000QP3oool20000
02l0oooo00<000000?ooo`3oool04@3oool003X0oooo00<000000?ooo`3oool0V03oool00`000000
oooo0?ooo`180?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0Q03oool2000002l0
oooo00<000000?ooo`3oool04@3oool003T0oooo00<000000?ooo`3oool0V@3oool00`000000oooo
0?ooo`180?ooo`030000003oool0oooo02d0oooo0`0000250?ooo`800000;`3oool00`000000oooo
0?ooo`0A0?ooo`00>@3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo04L0oooo00<0
00000?ooo`3oool0:`3oool300000080oooo00<000000?ooo`3oool0P@3oool00`000000oooo0000
0003000002`0oooo00<000000?ooo`3oool04@3oool003D0oooo1000002M0?ooo`@00000AP3oool0
0`000000oooo0?ooo`0[0?ooo`<000000P3oool00`000000oooo0?ooo`210?ooo`030000003oool0
oooo00800000;@3oool00`000000oooo0?ooo`0A0?ooo`00=P3oool00`000000oooo0?ooo`2O0?oo
o`030000003oool0oooo04D0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0O`3oool01@000000oooo0?ooo`3oool0
000000@0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0A0?ooo`00<@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo02L0oooo00@000000?ooo`3oool0oooo0P000004
0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3o
ool2000002P0oooo00<000000?ooo`3oool04@3oool00300oooo0P0000040?ooo`030000003oool0
oooo09l0oooo00<000000?ooo`3oool00P3oool200000440oooo00<000000?ooo`3oool09P3oool0
0`000000oooo000000020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0O@3oool0
0`000000oooo0?ooo`030?ooo`80000000@0oooo0000003oool00000:03oool00`000000oooo0?oo
o`0A0?ooo`00<03oool00`000000oooo000000030?ooo`030000003oool0oooo0:00oooo00L00000
0?ooo`3oool0oooo0000003oool000000400oooo00<000000?ooo`3oool08`3oool9000000L0oooo
00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`050?ooo`P000009@3oool00`000000oooo
0?ooo`0A0?ooo`00;@3oool300000080oooo1000002S0?ooo`@000000P3oool4000003d0oooo00<0
00000?ooo`3oool08`3oool2000000@0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`1k0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0T0?ooo`030000003oool0oooo0140oooo000]0?ooo`030000003oool0oooo0;40oooo00<00000
0?ooo`3oool0?03oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0140oooo000^0?oo
o`030000003oool0oooo0;00oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0S0?oo
o`030000003oool000000080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1i0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P00000V0?ooo`030000003oool0oooo
0140oooo000^0?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool0?@3oool00`000000
oooo0?ooo`0R0?ooo`@0000000<0oooo0000003oool02`3oool00`000000oooo0?ooo`1g0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`0000000`00000U0?ooo`030000003oool0oooo0140
oooo000]0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0?03oool00`000000oooo
0?ooo`0U0?ooo`8000003@3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo00X0oooo
0P00000X0?ooo`030000003oool0oooo0140oooo000]0?ooo`<00000[`3oool4000003h0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0;0?ooo`8000003P3oool00`000000oooo0?oo
o`1e0?ooo`030000003oool0oooo00/0oooo0P00000=0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool04@3oool00300oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`10
0?ooo`030000003oool0oooo01H0oooo0`00000<0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0L`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`000000603oool00`000000oooo0?ooo`0B0?ooo`00803oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo0:X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo01<0oooo100000000`3oool0000000000002000000H0oooo1000000@0?oo
o`030000003oool0oooo07<0oooo00<000000?ooo`3oool03@3oool4000000L0oooo0`0000000`3o
ool0000000000002000001D0oooo00<000000?ooo`3oool04P3oool001l0oooo0P00000?0?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool03@3oool200000340oooo00<000000?ooo`3o
ool04`3oool00`000000oooo000000030?ooo`8000001`3oool00`000000oooo0000000B0?ooo`03
0000003oool0oooo0740oooo00<000000?ooo`3oool03`3oool2000000P0oooo0P0000030?ooo`80
00005P3oool00`000000oooo0?ooo`0B0?ooo`007`3oool00`000000oooo0000000>0?ooo`030000
003oool0oooo0:`0oooo00<000000?ooo`3oool0303oool00`000000oooo0000000`0?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`800000
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
L@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`0B0?ooo`006`3oool400000080oooo100000080?ooo`<00000[`3o
ool4000000P0oooo0`0000020?ooo`@00000;@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool000000080oooo0P0000050?ooo`060000003oool0oooo0000003oool000000P3oool2000001@0
oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0A0?ooo`8000000P3oool2000000<0
oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool200000140oooo00<000000?oo
o`3oool04P3oool001`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo0;40oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo02d0oooo00<000000?ooo`3oool02`3ooolG000001D0oooo00<000000?ooo`3o
ool0K`3oool00`000000oooo0?ooo`0B0?oooaP00000303oool00`000000oooo0?ooo`0C0?ooo`00
5`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00/0oooo0P00000D0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`005`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool2000000@0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`020?ooo`800000103o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P00000Y0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo06/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool04`3oool001H0oooo00<000000?ooo`0000000`3oool0
0`000000oooo0?ooo`040?ooo`070000003oool0oooo0?ooo`000000oooo000000030?ooo`030000
003oool0oooo0;80oooo00H000000?ooo`3oool000000?ooo`0000030?ooo`030000003oool0oooo
00D0oooo00L000000?ooo`3oool0oooo0000003oool0000002P0oooo00<000000?ooo`3oool0303o
ool200000140oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1[0?ooo`030000003o
ool0oooo01H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo01<0oooo000C0?ooo`@0000000<0oooo0000000000000P0000080?ooo`@000000P3oool4
00000;@0oooo100000020?ooo`@000001`3oool400000080oooo1000000U0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`0000004@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo06T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0@0?ooo`8000003@3oool0
0`000000oooo0?ooo`0C0?ooo`004`3oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo
02@0oooo00<000000?ooo`3oool02`3oool4000000l0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`1Y0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03P3oool40000
00`0oooo00<000000?ooo`3oool04`3oool001@0oooo00<000000?ooo`3oool0h`3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo00`0oooo0P00000?0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0
oooo0P00000>0?ooo`030000003oool0oooo01@0oooo000C0?ooo`030000003oool0oooo0>D0oooo
00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0=0?ooo`8000003@3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0503oool001<0oooo1000
003R0?ooo`@000009P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0000000d0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo01`0oooo00<0
00000?ooo`3oool02P3oool3000000h0oooo00<000000?ooo`3oool0503oool001H0oooo00<00000
0?ooo`3oool0h03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00T0oooo0`000002
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool02P3oool00`000000oooo00000003
000000/0oooo00<000000?ooo`3oool0503oool001L0oooo00<000000?ooo`3oool0gP3oool00`00
0000oooo0?ooo`0Y0?ooo`030000003oool0oooo00T0oooo0P0000030?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`3oool2000000`0oooo00<00000
0?ooo`3oool0503oool001L0oooo00<000000?ooo`3oool0gP3oool00`000000oooo0?ooo`0Y0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0200oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0P0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool2000000`0oooo00<000000?oo
o`3oool0503oool001H0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`0Y0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo01D0oooo000C0?ooo`@00000hP3oool4000002L0oooo00<000000?ooo`3oool0
2@3oool8000000L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1O0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0103oool9000000X0oooo00<000000?ooo`3oool0
5@3oool001<0oooo00<000000?ooo`3oool0i@3oool00`000000oooo0?ooo`0V0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`0000000P3oool2000000D0oooo00<000000?ooo`3oool0903o
ool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool00`3o
ool200000080oooo00<000000?ooo`000000303oool00`000000oooo0?ooo`0F0?ooo`00503oool0
0`000000oooo0?ooo`3T0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool02`3oool2
000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0
oooo05d0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0F0?ooo`00503oool00`00
0000oooo0?ooo`3S0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`020?ooo`8000000`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
05/0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`02000000@0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0F0?ooo`004`3oool00`000000oooo0?ooo`3U0?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool03P3oool00`000000oooo000000030?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0V0?ooo`03
0000003oool0oooo008000004@3oool00`000000oooo0?ooo`0G0?ooo`004`3oool4000000030?oo
o`000000000000800000e`3oool400000080oooo1000000Y0?ooo`030000003oool0oooo00h0oooo
100000000`3oool000000?ooo`0Y0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0
:03oool500000100oooo00<000000?ooo`3oool05`3oool001H0oooo00<000000?ooo`0000000`3o
ool00`000000oooo0?ooo`3D0?ooo`070000003oool0oooo0?ooo`000000oooo0000000/0?ooo`03
0000003oool0oooo0140oooo0P00000[0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3o
ool0:03oool2000001<0oooo00<000000?ooo`3oool05`3oool001L0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo0080oooo0P00000^0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`1G0?ooo`030000
003oool0oooo02T0oooo0P00000B0?ooo`030000003oool0oooo01P0oooo000G0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0dP3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo05D0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo01P0oooo000L0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`0=0?ooo`@00000;P3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0
oooo02/0oooo1000000?0?ooo`030000003oool0oooo01P0oooo000K0?ooo`@00000d@3oool40000
03@0oooo00<000000?ooo`3oool0303oool00`000000oooo0000000`0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool0;@3oool2000000l0oooo00<000000?ooo`3oool06@3oool001l0
oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool00P3oool200000300oooo00<000000?ooo`3oool0D`3oool00`000000
oooo0?ooo`0^0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo03H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo000000020?ooo`800000<P3oool00`000000oooo
0?ooo`1A0?ooo`030000003oool0oooo02l0oooo0P0000020?ooo`8000002@3oool00`000000oooo
0?ooo`0J0?ooo`00803oool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo03L0oooo00<0
00000?ooo`3oool0103oool900000380oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`0`0?ooo`T000001@3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`3=
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0a
0?ooo`8000001@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01X0oooo000K0?oo
o`@00000d@3oool4000003H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo03@0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0c0?oo
o`800000103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo000L0?ooo`03
0000003oool0oooo0=@0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`1=0?ooo`03
0000003oool0oooo03@0oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo
01/0oooo000L0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`020?ooo`030000003oool000000080oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`1;0?ooo`030000003oool0oooo03D0oooo0P0000020?ooo`8000001@3oool00`000000
oooo0?ooo`0K0?ooo`005`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=80oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0080oooo
0`0000020?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0B`3oool00`000000oooo
0?ooo`0f0?ooo`030000003oool0000000<000000`3oool00`000000oooo0?ooo`0L0?ooo`005P3o
ool2000000@0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`020?ooo`800000<`3o
ool00`000000oooo0?ooo`030?ooo`<00000>P3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo03P0oooo0P0000060?ooo`030000003oool0oooo01`0oooo000C0?ooo`@0000000<0oooo
0000000000000P00003G0?ooo`@000000P3oool4000002l0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0i0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo01d0oooo000C0?ooo`030000003oool0oooo
0>D0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`020?ooo`800000?03oool00`00
0000oooo0?ooo`170?ooo`030000003oool0oooo03T0oooo0P0000040?ooo`030000003oool0oooo
01h0oooo000D0?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0;`3oool00`000000
oooo0?ooo`03000003h0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0k0?ooo`@0
000000<0oooo0000003oool07`3oool001@0oooo00<000000?ooo`3oool0h`3oool00`000000oooo
0?ooo`0a0?ooo`030000003oool0000000800000?P3oool00`000000oooo0?ooo`150?ooo`030000
003oool0oooo03`0oooo0P0000000`3oool000000?ooo`0P0?ooo`004`3oool00`000000oooo0?oo
o`3U0?ooo`030000003oool0oooo0300oooo00@000000?ooo`3oool00000@03oool00`000000oooo
0?ooo`130?ooo`030000003oool0oooo03d0oooo0P0000000`3oool000000?ooo`0P0?ooo`004`3o
ool400000>80oooo1000000b0?ooo`<00000@03oool00`000000oooo0?ooo`130?ooo`030000003o
ool0oooo03h0oooo0P00000R0?ooo`005P3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0
oooo03@0ooooA00000120?ooodD000008P3oool001L0oooo00<000000?ooo`3oool0gP3oool00`00
0000oooo0?ooo`0e0?ooo`050000003oool000000?ooo`0000000P3oool2000000D0oooo00H00000
0?ooo`3oool000000?ooo`0000020?ooo`800000503oool01P000000oooo0?ooo`000000oooo0000
0080oooo0P0000050?ooo`060000003oool0oooo0000003oool000000P3oool200000400oooo0P00
00020?ooo`030000003oool000000080oooo0P0000050?ooo`060000003oool0oooo0000003oool0
00000P3oool2000001<0oooo0P0000020?ooo`030000003oool000000080oooo00<000000?ooo`3o
ool00`3oool200000080oooo00D000000?ooo`000000oooo0000000S0?ooo`005`3oool00`000000
oooo0?ooo`3N0?ooo`030000003oool0oooo03H0oooo00@000000?ooo`00000000000`3oool00`00
0000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`0B
0?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0n0?ooo`8000000`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool010000000oooo0?ooo`00000S0?ooo`005P3oool00`000000oooo0?ooo`3O0?ooo`030000003o
ool0oooo03H0oooo00<000000?ooo`000000103oool2000000<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`8000004P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo0P0000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool2000003h0oooo0P0000040?ooo`030000003oool0oooo0080oooo0P0000030?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00P3oool200000140oooo0P0000040?ooo`030000003o
ool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000040?ooo`030000003oool0000002@0oooo
000C0?ooo`@00000hP3oool4000003D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0000
00030?ooo`8000001`3oool00`000000oooo0000000B0?ooo`8000001`3oool00`000000oooo0000
00020?ooo`800000203oool00`000000oooo0000000n0?ooo`800000203oool00`000000oooo0000
00020?ooo`800000203oool00`000000oooo0000000A0?ooo`800000203oool00`000000oooo0000
00020?ooo`8000001@3oool00`000000oooo0?ooo`0S0?ooo`004`3oool00`000000oooo0?ooo`3U
0?ooo`030000003oool0oooo03D0oooo00@000000?ooo`3oool0oooo100000000`3oool000000000
0002000000H0oooo1000000@0?ooo`@000001P3oool4000000040?ooo`0000000000000000L0oooo
1000000l0?ooo`@000001`3oool4000000040?ooo`0000000000000000L0oooo1000000?0?ooo`@0
00001`3oool3000000030?ooo`000000000000800000103oool00`000000oooo0?ooo`0S0?ooo`00
503oool00`000000oooo0?ooo`3T0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0
103oool3000000`0oooo00<000000?ooo`3oool03@3oool2000000`0oooo0P00000=0?ooo`030000
003oool0oooo03X0oooo0P00000<0?ooo`800000303oool2000000l0oooo00<000000?ooo`3oool0
2P3oool3000000H0oooo00<000000?ooo`3oool0903oool001@0oooo00<000000?ooo`3oool0h`3o
ool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0h0?ooo`8000003P3oool00`000000oooo0?ooo`0;0?ooo`8000003@3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0903o
ool001<0oooo00<000000?ooo`3oool0i@3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0
oooo0180oooo0P00000=0?ooo`<000006`3oool3000003T0oooo0P00000K0?ooo`030000003oool0
000000d0oooo0P00000D0?ooo`030000003oool0oooo02D0oooo000C0?ooo`@0000000<0oooo0000
000000000P0000080?ooo`@000000P3oool4000001P0oooo100000020?ooo`@00000203oool40000
0080oooo0`00001=0?ooo`@000000P3oool4000000L0oooo100000020?ooo`@000006@3oool40000
0080oooo100000070?ooo`@000000P3oool4000003P0oooo00<000000?ooo`3oool03P3oool40000
00030?ooo`000000oooo00/0oooo00<000000?ooo`0000000`00000E0?ooo`@0000000<0oooo0000
003oool0=`3oool2000000040?ooo`0000000000000001H0oooo0`0000020?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`0000000`00000@0?ooo`030000003oool0oooo02H0oooo000F0?oo
o`030000003oool0000000<0oooo00<000000?ooo`3oool0103oool01`000000oooo0?ooo`3oool0
00000?ooo`0000000`3oool00`000000oooo0?ooo`0F0?ooo`070000003oool0oooo0?ooo`000000
oooo000000030?ooo`030000003oool0oooo00@0oooo00L000000?ooo`3oool0oooo0000003oool0
000000<0oooo00<000000?ooo`3oool0BP3oool01P000000oooo0?ooo`000000oooo000000<0oooo
00<000000?ooo`3oool01@3oool01`000000oooo0?ooo`3oool000000?ooo`0000000P3oool00`00
0000oooo0?ooo`0G0?ooo`060000003oool0oooo0000003oool000000`3oool00`000000oooo0?oo
o`050?ooo`070000003oool0oooo0?ooo`000000oooo0000000l0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`090?ooo`040000003oool0oooo0?oo
o`8000005P3oool00`000000oooo000000020?ooo`800000=P3oool200000080oooo0P00000G0?oo
o`8000000`3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo0?ooo`800000403oool0
0`000000oooo0?ooo`0W0?ooo`005`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool2000000@0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`020?ooo`800000103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo0P0000030?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool00P3oool20000
00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`800000103oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0080oooo0P0000040?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool2000003h0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool20000
01L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0d0?ooo`8000000`3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0:03oool001L0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0;0?ooo`8000001@3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0503oool2000000D0
oooo00<000000?ooo`3oool0<P3oool2000000@0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo02P0oooo000L0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04T0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0;0?ooo`T000001P3oool9000001D0oooo
2000000d0?ooo`P000005P3oool8000000L0oooo2000000=0?ooo`030000003oool0oooo02T0oooo
000K0?ooo`@000000P3oool4000000P0oooo0`00000D0?ooo`<00000203oool400000080oooo1000
00070?ooo`@00000A`3oool4000000P0oooo100000020?ooo`<00000203oool4000001<0oooo1000
00080?ooo`<000000P3oool4000004D0oooo00<000000?ooo`3oool03@3oool2000000<0oooo00<0
00000?ooo`3oool00`3oool200000080oooo0P00000L0?ooo`8000000`3oool00`000000oooo0?oo
o`0`0?ooo`8000000P3oool2000001/0oooo00<000000?ooo`0000000P3oool2000000D0oooo00H0
00000?ooo`3oool000000?ooo`00000@0?ooo`030000003oool0oooo02T0oooo000O0?ooo`030000
003oool0000000h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0000000l0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0000000h0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0000004P0oooo00<000000?ooo`3oool03P3oool010000000oooo0?ooo`3oool2000000@0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0K0?ooo`040000003oool0oooo0?oo
o`800000<@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`8000000`3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo02X0oooo000O0?ooo`8000003`3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo00d0oooo0P00000?0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3o
ool03P3oool2000000l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0=0?ooo`80
0000BP3oool00`000000oooo0?ooo`0B0?ooo`040000003oool0oooo0?ooo`8000009P3oool20000
0300oooo0P00000V0?ooo`8000000`3oool2000001<0oooo00<000000?ooo`3oool0:`3oool00200
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0@0?ooo`<000000P3oool00`000000
oooo0000000U0?ooo`<00000;P3oool2000000030?ooo`000000oooo02<0oooo00<000000?ooo`00
00000`3oool200000180oooo00<000000?ooo`3oool0;03oool00300oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0
7P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0
3`3oool3000000030?ooo`000000000000800000903oool4000002h0oooo1000000T0?ooo`@00000
00<0oooo0000000000000P00000@0?ooo`030000003oool0oooo02d0oooo000]0?ooo`<00000503o
ool3000001T0oooo100000170?ooo`@000006@3oool4000001<0oooo1000001J0?ooo`030000003o
ool0oooo0100oooo0`00000Z0?ooo`800000;03oool2000002X0oooo0`00000B0?ooo`030000003o
ool0oooo02h0oooo000]0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`1:0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`1I0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0:P3oool2
000002/0oooo0P00000[0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0;`3oool0
02h0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
04T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
05/0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0000000Z0?ooo`<00000?@3oool00`00
0000oooo0?ooo`0`0?ooo`00;P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0g0?ooo`@000000P3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0000000<00000>@3oool00`000000oooo0?ooo`0a0?ooo`00
;@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
B@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
G@3oool00`000000oooo0?ooo`0g0?ooo`8000000`3oool00`000000oooo0?ooo`0V0?ooo`800000
0P3oool2000003T0oooo00<000000?ooo`3oool0<P3oool002d0oooo0`0000020?ooo`@00000203o
ool400000080oooo0`00000I0?ooo`@000000P3oool4000003/0oooo100000020?ooo`@000006@3o
ool400000080oooo100000070?ooo`@000000P3oool4000005l0oooo00<000000?ooo`3oool0<P3o
ool010000000oooo0?ooo`3oool2000000@0oooo00<000000?ooo`3oool09@3oool010000000oooo
0?ooo`3oool2000000@0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0c0?ooo`00
<03oool00`000000oooo000000030?ooo`030000003oool0oooo00@0oooo00L000000?ooo`3oool0
oooo0000003oool000000200oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`0i0?oo
o`070000003oool0oooo0?ooo`000000oooo0000000O0?ooo`030000003oool0000000<0oooo00<0
00000?ooo`3oool01@3oool01`000000oooo0?ooo`3oool000000?ooo`000000HP3oool00`000000
oooo0?ooo`0a0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02@0
oooo0P0000040?ooo`040000003oool0oooo0?ooo`800000<`3oool00`000000oooo0?ooo`0c0?oo
o`00<03oool2000000@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`80
0000803oool2000000@0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`020?ooo`80
00008@3oool2000000@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`80
0000I03oool00`000000oooo0?ooo`0]0?ooo`T000001P3oool00`000000oooo0?ooo`0R0?ooo`80
00001@3oool9000002l0oooo00<000000?ooo`3oool0=03oool00340oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0
;03oool2000000D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool2000002h0oooo00<000000?ooo`3oool0
=@3oool003H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0Y0?ooo`030000003o
ool0oooo03P0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo06X0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0803oool2000000P0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo03H0oooo000e0?ooo`@000000P3oool4
000002X0oooo1000000f0?ooo`<00000:`3oool4000000030?ooo`000000000000800000K03oool2
000002`0oooo0P0000030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool07`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0:P3oool2
000003T0oooo000i0?ooo`030000003oool000000300oooo00<000000?ooo`3oool0<`3oool00`00
0000oooo0?ooo`0_0?ooo`030000003oool000000740oooo0P00000Y0?ooo`030000003oool00000
0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0N0?ooo`8000002P3oool01P00
0000oooo0?ooo`000000oooo000002X0oooo00<000000?ooo`3oool0>@3oool003T0oooo0P00000b
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1b
0?ooo`030000003oool0oooo02H0oooo100000000`3oool000000?ooo`0;0?ooo`030000003oool0
oooo01d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`03000002T0oooo00<00000
0?ooo`3oool0>P3oool003X0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0a0?oo
o`030000003oool0oooo0340oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0X0?oo
o`8000003@3oool00`000000oooo0?ooo`0L0?ooo`800000303oool2000002X0oooo0P00000m0?oo
o`00K03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3o
ool09`3oool2000000h0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool0?@3oool006T0oooo1000000f0?ooo`<00000
Y`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0
?P3oool006T0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`2V0?ooo`030000003o
ool0oooo0280oooo1000000@0?ooo`030000003oool0oooo01P0oooo0P00000@0?ooo`<00000903o
ool00`000000oooo0?ooo`0o0?ooo`00JP3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo0:P0oooo0P00000R0?ooo`030000003oool000000180oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0?0?ooo`800000903oool00`000000oooo0?ooo`100?ooo`00I@3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`2T0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0F0?ooo`8000004P3oool00`000000oooo0?ooo`0P0?ooo`800000@`3oool006@0
oooo0P0000030?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool00P3oool200000:L0
oooo0P00000O0?ooo`800000503oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool07P3oool2000004D0oooo001P0?ooo`@000000P3oool4000003/0oooo
100000020?ooo`@00000Y@3oool00`000000oooo0?ooo`0L0?oooaL000005P3ooolG000001h0oooo
00<000000?ooo`3oool0A@3oool00640oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?oo
o`2T0?ooo`030000003oool0oooo01h0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00@0
00000?ooo`3oool0oooo0P0000030?ooo`030000003oool0oooo0180oooo0P0000020?ooo`800000
0`3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`800000803oool00`000000
oooo0?ooo`160?ooo`00H@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0:H0oooo
0P00000O0?ooo`040000003oool0oooo0?ooo`800000103oool00`000000oooo0?ooo`020?ooo`04
0000003oool0oooo0?ooo`8000004`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo
0?ooo`800000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01`0oooo0P000019
0?ooo`00HP3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0:P0oooo00<000000?oo
o`3oool0803oool2000000<0oooo0P0000080?ooo`8000004P3oool2000000P0oooo0P0000030?oo
o`8000008P3oool00`000000oooo0?ooo`190?ooo`00H@3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo0:T0oooo0`00000O0?ooo`8000000`3oool00`000000oooo000000070?ooo`<0
00004@3oool00`000000oooo000000070?ooo`8000000`3oool00`000000oooo0000000N0?ooo`<0
0000C03oool00600oooo100000170?ooo`@00000[@3oool2000001`0oooo100000000`3oool00000
00000002000000H0oooo1000000@0?ooo`@000001P3oool4000000030?ooo`000000000000800000
703oool2000004l0oooo001T0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0/03o
ool00`000000oooo0?ooo`0L0?ooo`<00000303oool2000000h0oooo0P00000<0?ooo`<000007P3o
ool00`000000oooo0?ooo`1?0?ooo`00I03oool00`000000oooo0?ooo`130?ooo`030000003oool0
oooo0;80oooo0P00000M0?ooo`030000003oool0oooo00`0oooo0P00000=0?ooo`8000003@3oool0
0`000000oooo0?ooo`0K0?ooo`800000DP3oool006D0oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`2d0?ooo`<00000:03oool00`000000oooo0000000=0?ooo`800000:@3oool20000
05@0oooo001T0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0]`3oool2000002<0
oooo100000000`3oool000000000000;0?ooo`030000003oool0000000<00000903oool2000005H0
oooo001P0?ooo`@00000A`3oool400000;L0oooo00<000000?ooo`3oool08@3oool2000000<0oooo
00<000000?ooo`3oool0203oool200000080oooo0P00000S0?ooo`800000F03oool00640oooo00<0
00000?ooo`3oool0BP3oool00`000000oooo0?ooo`2f0?ooo`8000008@3oool2000000@0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01l0oooo0P00
001J0?ooo`00H@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0;T0oooo0`00000N
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3oool2000000@0oooo00<00000
0?ooo`3oool07@3oool2000005`0oooo001R0?ooo`030000003oool0oooo04P0oooo00<000000?oo
o`3oool0_03oool2000001/0oooo2@0000060?ooo`T000006`3oool3000005h0oooo001Q0?ooo`03
0000003oool0oooo04T0oooo00<000000?ooo`3oool0_P3oool6000001P0oooo0P0000030?ooo`80
0000103oool200000080oooo0P00000I0?ooo`H00000H@3oool00600oooo100000020?ooo`@00000
203oool4000000030?ooo`0000000000008000006@3oool400000080oooo100000070?ooo`@00000
0P3oool400000<D0oooo0`00000F0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0P3oool2000000<0oooo00<000000?ooo`3oool0503oool3000006L0oooo001T0?ooo`030000003o
ool000000080oooo00<000000?ooo`3oool01@3oool01`000000oooo0?ooo`3oool000000?ooo`00
00000`3oool00`000000oooo0?ooo`0F0?ooo`070000003oool0oooo0?ooo`000000oooo00000002
0?ooo`030000003oool0oooo00D0oooo00L000000?ooo`3oool0oooo0000003oool000000</0oooo
0`00000H0?ooo`8000000`3oool00`000000oooo0?ooo`0G0?ooo`800000JP3oool006@0oooo0P00
00040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool00P3oool2000000@0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`800000c`3oool500000180oooo00<0
00000?ooo`0000000P3oool300000180oooo1P00001/0?ooo`00I@3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3B
0?ooo`H00000303oool4000000040?ooo`0000000000000000`0oooo1P00001b0?ooo`00J@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0gP3oool9
000000H0oooo0`0000050?ooo`T00000N03oool006T0oooo100000020?ooo`@000001`3oool40000
01<0oooo100000080?ooo`@000000P3oool300000>P0oooo3P0000210?ooo`00K03oool00`000000
oooo0000000?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0303oool00`000000
oooo0000003o0?ooog/0oooo001]0?ooo`8000003`3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo00h0oooo0P00003o0?ooog/0oooo001]0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0?l0ooooNP3o
ool007h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3o0?oooh/0oooo001j0?oo
o`@000004`3oool400000?l0ooooR@3oool007/0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`3o0?ooohL0oooo001k0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
o`3ooon80?ooo`00O03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0?l0ooooR03o
ool007/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3o0?ooohP0oooo001j0?oo
o`@000000P3oool4000000P0oooo0`0000020?ooo`@00000o`3ooon90?ooo`00OP3oool00`000000
oooo000000020?ooo`030000003oool0oooo00D0oooo00L000000?ooo`3oool0oooo0000003oool0
00000?l0ooooS03oool007h0oooo0P0000040?ooo`030000003oool0oooo00@0oooo00@000000?oo
o`3oool0oooo0P00003o0?ooohd0oooo001o0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0ooooR`3oool008<0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o0?oooi00oooo00230?ooo`@00000
0P3oool400000?l0ooooT@3oool008H0oooo00<000000?ooo`000000o`3ooonE0?ooo`00Q`3oool2
00000?l0ooooU@3oool008L0oooo00<000000?ooo`3oool0o`3ooonD0?ooo`00o`3ooooo0?oooal0
oooo003o0?ooool0oooo7`3oool00?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo003o0?oo
ool0oooo7`3oool00?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo003o0?ooool0oooo7`3o
ool00?l0ooooo`3ooolO0?ooo`00o`3ooooo0?oooal0oooo003o0?ooool0oooo7`3oool00?l0oooo
o`3ooolO0?ooo`00o`3ooooo0?oooal0oooo0000\
\>"],
  ImageRangeCache->{{{0, 540.938}, {296.938, 0}} -> {-0.051787, -0.0274986, \
0.00583314, 0.00583314}, {{13.3125, 258.188}, {289.813, 7.0625}} -> \
{-0.0821032, -0.348102, 0.00643208, 0.00643208}, {{282.688, 527.563}, \
{289.813, 7.0625}} -> {-1.50589, -0.446592, 0.00742694, 0.00742694}}]
}, Closed]],

Cell[CellGroupData[{

Cell["Implementation", "Section"],

Cell[CellGroupData[{

Cell["Initialization", "Subsection"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Off[General::spell, \ General::spell1]\)}], "Input"],

Cell[BoxData[
    \(Needs["\<Utilities`Notation`\>"]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(n\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[ScriptL]\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(L\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(A\_0\),
          NotationBoxTag,
          Editable->True], "]"}]}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Basic algorithm", "Subsection"],

Cell[CellGroupData[{

Cell["The vertices of the triangle built on a line segment", "Subsubsection"],

Cell[TextData[{
  "To construct an equilateral triangle upon a line segment, we must \
calculate the base points and the vertex point.  The simplest way to do this \
is with vectors.  Let ",
  Cell[BoxData[
      \(TraditionalForm\`r\&\[RightVector] = {x, y}\)]],
  " represent a two-dimensional vector and let ",
  Cell[BoxData[
      FormBox[
        RowBox[{" ", 
          FormBox[\(r\&\[RightVector]\_1 = {x, y}\),
            "TraditionalForm"]}], TraditionalForm]]],
  " and",
  Cell[BoxData[
      FormBox[
        RowBox[{" ", 
          FormBox[
            RowBox[{" ", 
              FormBox[\(r\&\[RightVector]\_2 = {x, y}\),
                "TraditionalForm"]}],
            "TraditionalForm"]}], TraditionalForm]]],
  " represent the end points of a line segment.  The base points of the \
triangle are then ",
  Cell[BoxData[
      \(TraditionalForm
      \`r\&\[RightVector]\_a = 
        r\&\[RightVector]\_1 + 
          \(1\/3\) \((r\&\[RightVector]\_2 - r\&\[RightVector]\_1)\)\)]],
  " and ",
  Cell[BoxData[
      FormBox[
        RowBox[{" ", 
          
          FormBox[\(r\&\[RightVector]\_b = 
              r\&\[RightVector]\_1 + 
                \(2\/3\) \((r\&\[RightVector]\_2 - r\&\[RightVector]\_1)\)\),
            "TraditionalForm"]}], TraditionalForm]]],
  ".  To compute the vertex, we must move along a perpendicular to the line \
segment a distance equal to ",
  Cell[BoxData[
      \(TraditionalForm\`1\/\@12\)]],
  " its length.  Recognizing that the vector ",
  Cell[BoxData[
      \(TraditionalForm\`{y, \(-x\)}\)]],
  " is perpendicular to ",
  Cell[BoxData[
      \(TraditionalForm\`{x, y}\)]],
  ", we can compute the vertex point as"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(r\& \[RightVector] \)\_c = \(1\/2\) \((\(r\& \
\[RightVector] \)\_1 + \(r\& \[RightVector] \)\_2)\) + {y\_2 - y\_1, x\_1 - x\
\_2}\/\@12\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell["These points are evaluated using the following functions.", "Text"],

Cell[BoxData[{
    \(\(Clear[perp];\)\), "\n", 
    \(\(perp[
          vector : {x_, y_}] := {y, \(-x\)}\/\@\(x\^2 + y\^2\);\)\), "\n", 
    \(perp[pt1 : {x1_, y1_}, pt2 : {x2_, y2_}] := perp[pt1 - pt2]\)}], "Input"],

Cell[BoxData[
    \(\(newpt1[pt1 : {x1_, y1_}, pt2 : {x2_, y2_}] := 
        pt1 + \(1\/3\) \((pt2 - pt1)\);\)\)], "Input"],

Cell[BoxData[
    \(\(newpt2[pt1 : {x1_, y1_}, 
          pt2 : {x2_, y2_}] := \(1\/2\) \((pt1 + pt2)\) + \(1\/\@12\) 
            perp[pt1, 
              pt2]\ \@\(\((pt2 - pt1)\) . \((pt2 - pt1)\)\);\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(newpt2[{x1, y1}, {x2, y2}] // Simplify\)], "Input"],

Cell[BoxData[
    \({1\/6\ \((3\ x1 + 3\ x2 + \@3\ \((y1 - y2)\))\), 
      1\/6\ \((\(-\@3\)\ x1 + \@3\ x2 + 3\ \((y1 + y2)\))\)}\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(\(newpt3[pt1 : {x1_, y1_}, pt2 : {x2_, y2_}] := 
        pt1 + \(2\/3\) \((pt2 - pt1)\);\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Replacement rule for line segments", "Subsubsection"],

Cell["\<\
The following rule replaces a line segment by the four segments required to \
place a triangle atop the middle third of the original segment.\
\>", "Text"],

Cell[BoxData[{
    \(\(Clear[KochRule];\)\), "\n", 
    \(\(KochRule = 
        Line[{pt1 : {x1_, y1_}, 
              pt2 : {x2_, y2_}}] \[RuleDelayed] {Line[{pt1, 
                newpt1[pt1, pt2]}], 
            Line[{newpt1[pt1, pt2], newpt2[pt1, pt2]}], 
            Line[{newpt2[pt1, pt2], newpt3[pt1, pt2]}], 
            Line[{newpt3[pt1, pt2], pt2}]};\)\)}], "Input"],

Cell[BoxData[{
    \(\(Clear[KochSplit];\)\), "\n", 
    \(KochSplit[x_] := x /. KochRule\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Basic function which displays Koch transformation of polygon", \
"Subsubsection"],

Cell[BoxData[{
    \(\(Clear[BasicDisplayKoch];\)\), "\n", 
    \(BasicDisplayKoch[polygon : {{_, _} .. }, n_, opts___Rule] := 
      Show[Graphics[Nest[KochSplit, Line /@ Partition[polygon, 2, 1], n]], 
        opts, AspectRatio \[Rule] Automatic]\)}], "Input"],

Cell["\<\
The basic algorithm is illustrated by the following iteration sequence \
applied to a simple line segment.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Table[
        BasicDisplayKoch[{{0. , 0. }, {1. , 0. }}, n], {n, 1, 
          4}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28868 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00687322 0.952381 [
[ 0 0 0 0 ]
[ 1 .28868 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28868 L
0 .28868 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .00687 m
.34127 .00687 L
s
.34127 .00687 m
.5 .2818 L
s
.5 .2818 m
.65873 .00687 L
s
.65873 .00687 m
.97619 .00687 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 83.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000O2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002`3ooon9000008L0ooooR@00000<0?ooo`00U03o
ool00`000000oooo0?ooo`230?ooo`030000003oool0oooo09<0oooo002D0?ooo`030000003oool0
oooo08<0oooo00<000000?ooo`3oool0T`3oool009D0oooo00<000000?ooo`3oool0P@3oool00`00
0000oooo0?ooo`2D0?ooo`00U@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo09@0
oooo002F0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0U@3oool009H0oooo00<0
00000?ooo`3oool0O`3oool00`000000oooo0?ooo`2E0?ooo`00U`3oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo09H0oooo002H0?ooo`030000003oool0oooo07/0oooo00<000000?oo
o`3oool0U`3oool009P0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`2G0?ooo`00
V@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo09P0oooo002I0?ooo`030000003o
ool0oooo07T0oooo00<000000?ooo`3oool0V03oool009X0oooo00<000000?ooo`3oool0M`3oool0
0`000000oooo0?ooo`2I0?ooo`00VP3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo
09T0oooo002K0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0VP3oool009`0oooo
00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`2K0?ooo`00W03oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo09/0oooo002M0?ooo`030000003oool0oooo0740oooo00<00000
0?ooo`3oool0W03oool009d0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`2L0?oo
o`00WP3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo09d0oooo002O0?ooo`030000
003oool0oooo06d0oooo00<000000?ooo`3oool0WP3oool009l0oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`2N0?ooo`00X03oool00`000000oooo0?ooo`1[0?ooo`030000003oool0
oooo09l0oooo002P0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0W`3oool00:40
oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`2P0?ooo`00X@3oool00`000000oooo
0?ooo`1Y0?ooo`030000003oool0oooo0:00oooo002R0?ooo`030000003oool0oooo06L0oooo00<0
00000?ooo`3oool0X@3oool00:<0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`2R
0?ooo`00X`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0:80oooo002T0?ooo`03
0000003oool0oooo06<0oooo00<000000?ooo`3oool0X`3oool00:@0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`2S0?ooo`00Y@3oool00`000000oooo0?ooo`1Q0?ooo`030000003o
ool0oooo0:@0oooo002U0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0Y03oool0
0:H0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`2U0?ooo`00Y`3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo0:H0oooo002W0?ooo`030000003oool0oooo05d0oooo
00<000000?ooo`3oool0YP3oool00:P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?oo
o`2W0?ooo`00Z03oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0:L0oooo002Y0?oo
o`030000003oool0oooo05T0oooo00<000000?ooo`3oool0Z03oool00:T0oooo00<000000?ooo`3o
ool0F@3oool00`000000oooo0?ooo`2X0?ooo`00ZP3oool00`000000oooo0?ooo`1G0?ooo`030000
003oool0oooo0:T0oooo002[0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0ZP3o
ool00:/0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`2Z0?ooo`00[03oool00`00
0000oooo0?ooo`1C0?ooo`030000003oool0oooo0:/0oooo002/0?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool0Z`3oool00:d0oooo00<000000?ooo`3oool0D@3oool00`000000oooo
0?ooo`2/0?ooo`00[P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo0:d0oooo002^
0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0[@3oool00:l0oooo00<000000?oo
o`3oool0C@3oool00`000000oooo0?ooo`2^0?ooo`00[`3oool00`000000oooo0?ooo`1=0?ooo`03
0000003oool0oooo0:h0oooo002`0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0
[`3oool00;00oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`2_0?ooo`00/@3oool0
0`000000oooo0?ooo`190?ooo`030000003oool0oooo0;00oooo002b0?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool0/@3oool00;80oooo00<000000?ooo`3oool0A`3oool00`000000
oooo0?ooo`2a0?ooo`00/`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0;80oooo
002c0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0/P3oool00;@0oooo00<00000
0?ooo`3oool0@`3oool00`000000oooo0?ooo`2c0?ooo`00]03oool00`000000oooo0?ooo`130?oo
o`030000003oool0oooo0;<0oooo002e0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3o
ool0]03oool00;H0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`2e0?ooo`00]P3o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0;D0oooo002g0?ooo`030000003oool0
oooo03d0oooo00<000000?ooo`3oool0]P3oool00;L0oooo00<000000?ooo`3oool0?@3oool00`00
0000oooo0?ooo`2f0?ooo`00^03oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0;L0
oooo002h0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0]`3oool00;T0oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0?ooo`2h0?ooo`00^P3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo0;T0oooo002j0?ooo`030000003oool0oooo03L0oooo00<000000?oo
o`3oool0^@3oool00;/0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`2j0?ooo`00
^`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0;X0oooo002l0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0^`3oool00;`0oooo00<000000?ooo`3oool0<`3oool0
0`000000oooo0?ooo`2k0?ooo`00_@3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo
0;`0oooo002n0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0_@3oool00;h0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`2m0?ooo`00_`3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo0;h0oooo002o0?ooo`030000003oool0oooo02d0oooo00<00000
0?ooo`3oool0_P3oool00<00oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`2o0?oo
o`00`@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo00310?ooo`030000
003oool0oooo02T0oooo00<000000?ooo`3oool0`03oool00<80oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`310?ooo`00`P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0
oooo0<40oooo00330?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0`P3oool00<<0
oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`320?ooo`00a03oool00`000000oooo
0?ooo`0S0?ooo`030000003oool0oooo0<<0oooo00350?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool0a03oool00<D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`34
0?ooo`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00360?ooo`03
0000003oool0oooo01l0oooo00<000000?ooo`3oool0a@3oool00<L0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`360?ooo`00a`3oool00`000000oooo0?ooo`0M0?ooo`030000003o
ool0oooo0<H0oooo00380?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0a`3oool0
0<T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`380?ooo`00b@3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo0<P0oooo003:0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0b@3oool00<X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`390?ooo`00b`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0<X0oooo003;0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool0bP3oool00<`0oooo00<000000?ooo`3o
ool04`3oool00`000000oooo0?ooo`3;0?ooo`00c@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0<`0oooo003=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0c03o
ool00<h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3=0?ooo`00cP3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo0<d0oooo003?0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool0cP3oool00=00oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`3?0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo003A
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=40oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00dP3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo0=40oooo003B0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
d@3oool00=<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3B0?ooo`00e03oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003D0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?oo
o`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003F0?ooo`030000003oool000000=L0
oooo003F0?ooo`030000003oool000000=L0oooo003G0?ooo`030000003oool0oooo0=H0oooo003o
0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {123.688, 0}} -> {-0.028119, -0.0072184, \
0.003676, 0.003676}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28868 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00687322 0.952381 [
[ 0 0 0 0 ]
[ 1 .28868 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28868 L
0 .28868 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .00687 m
.12963 .00687 L
s
.12963 .00687 m
.18254 .09852 L
s
.18254 .09852 m
.23545 .00687 L
s
.23545 .00687 m
.34127 .00687 L
s
.34127 .00687 m
.39418 .09852 L
s
.39418 .09852 m
.34127 .19016 L
s
.34127 .19016 m
.44709 .19016 L
s
.44709 .19016 m
.5 .2818 L
s
.5 .2818 m
.55291 .19016 L
s
.55291 .19016 m
.65873 .19016 L
s
.65873 .19016 m
.60582 .09852 L
s
.60582 .09852 m
.65873 .00687 L
s
.65873 .00687 m
.76455 .00687 L
s
.76455 .00687 m
.81746 .09852 L
s
.81746 .09852 m
.87037 .00687 L
s
.87037 .00687 m
.97619 .00687 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 83.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000O2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002`3oool^000002d0oooo;P0000270?ooobh00000
;@3oool^000000`0oooo000i0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0;03o
ool00`000000oooo0?ooo`230?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0:@3o
ool00`000000oooo0?ooo`0h0?ooo`00>@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0
oooo02`0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0
oooo02T0oooo00<000000?ooo`3oool0>03oool003X0oooo00<000000?ooo`3oool09`3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo03T0oooo000j0?ooo`030000003oool0oooo02L0
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo02h0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0i0?ooo`00>`3oool00`000000oooo
0?ooo`0U0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0O`3oool00`000000oooo
0?ooo`0`0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0>P3oool003`0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo07d0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo03/0oooo000l
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0k
0?ooo`00?@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0N`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool0?03oool003d0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo07/0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo03`0oooo000n0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
=P3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0m0?ooo`00?P3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo03H0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0f0?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool0?@3oool003l0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`0M0?ooo`030000003oool0oooo03h0oooo00100?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo
03X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0o0?ooo`00@03oool00`000000
oooo0?ooo`0K0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0M@3oool00`000000
oooo0?ooo`0j0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0?`3oool00440oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo07<0oooo
00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0400oooo
00110?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`100?ooo`00@P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo03h0oooo00<00000
0?ooo`3oool0L@3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0@@3oool004<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`100?oo
o`030000003oool0oooo06l0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo0480oooo00130?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`120?ooo`00A03oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo0480oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`120?ooo`030000
003oool0oooo01<0oooo00<000000?ooo`3oool0@`3oool004@0oooo00<000000?ooo`3oool04`3o
ool00`000000oooo0?ooo`120?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0@P3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo04<0oooo00150?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`1[0?ooo`030000003oool0
oooo04@0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`140?ooo`00AP3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0J@3oool00`00
0000oooo0?ooo`160?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0A@3oool004H0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo06T0
oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo04D0
oooo00170?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0B03oool00`000000oooo
0?ooo`1W0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`160?ooo`00A`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo04P0oooo00<0
00000?ooo`3oool0I`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0AP3oool004P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo04L0oooo00190?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool0C03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04`0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`180?ooo`00B@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo04`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1<0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0B03oool004X0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0
CP3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04T0oooo001:0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1Q0?ooo`030000003o
ool0oooo04h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`190?ooo`00B`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0G`3oool0
0`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0BP3oool0
04/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
05l0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
04X0oooo001<0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0DP3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`1;0?ooo`00C@3oool01@000000oooo0?ooo`3oool0000005H0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`1D0?ooo`050000003oool0oooo0?ooo`000000CP3oool004d0
oooo00D000000?ooo`3oool0oooo0000001F0?ooo`030000003oool0oooo05/0oooo00<000000?oo
o`3oool0E03oool01@000000oooo0?ooo`3oool0000004h0oooo001>0?ooo`030000003oool00000
05P0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`1F0?ooo`030000003oool00000
04l0oooo001>0?ooo`030000003oool0000005P0oooo00<000000?ooo`3oool0F@3oool00`000000
oooo0?ooo`1F0?ooo`030000003oool0000004l0oooo001?0?ooo`030000003oool0oooo05P0oooo
00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo04h0oooo
002Y0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0Z03oool00:T0oooo00<00000
0?ooo`3oool0F@3oool00`000000oooo0?ooo`2X0?ooo`00Z03oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo0:L0oooo002X0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3o
ool0Y`3oool00:L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`2V0?ooo`00YP3o
ool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0:D0oooo002V0?ooo`030000003oool0
oooo05l0oooo00<000000?ooo`3oool0Y@3oool00:D0oooo00<000000?ooo`3oool0H@3oool00`00
0000oooo0?ooo`2T0?ooo`00Y@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo0:@0
oooo002T0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0X`3oool00:@0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`2S0?ooo`00X`3oool00`000000oooo0?ooo`1U
0?ooo`030000003oool0oooo0:80oooo002R0?ooo`030000003oool0oooo06L0oooo00<000000?oo
o`3oool0X@3oool00:80oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`2Q0?ooo`00
X@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo0:00oooo002Q0?ooo`030000003o
ool0oooo06T0oooo00<000000?ooo`3oool0X03oool00:00oooo00<000000?ooo`3oool0J`3oool0
0`000000oooo0?ooo`2O0?ooo`00W`3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo
09h0oooo002O0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0WP3oool009h0oooo
00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`2M0?ooo`00WP3oool00`000000oooo0?oo
o`1_0?ooo`030000003oool0oooo09d0oooo002M0?ooo`030000003oool0oooo0740oooo00<00000
0?ooo`3oool0W03oool009`0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`2K0?oo
o`00W03oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo09/0oooo002K0?ooo`030000
003oool0oooo07D0oooo00<000000?ooo`3oool0VP3oool009/0oooo00<000000?ooo`3oool0M@3o
ool00`000000oooo0?ooo`2J0?ooo`00VP3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0
oooo09T0oooo002I0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0V03oool009T0
oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`2H0?ooo`00V03oool00`000000oooo
0?ooo`1k0?ooo`030000003oool0oooo09L0oooo002H0?ooo`030000003oool0oooo07/0oooo00<0
00000?ooo`3oool0U`3oool009L0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`2F
0?ooo`00U`3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo09H0oooo002F0?ooo`03
0000003oool0oooo07l0oooo00<000000?ooo`3oool0U@3oool009D0oooo00<000000?ooo`3oool0
P@3oool00`000000oooo0?ooo`2D0?ooo`00U@3oool00`000000oooo0?ooo`210?ooo`030000003o
ool0oooo09@0oooo002D0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0T`3oool0
09@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`2C0?ooo`00T`3oool^000002d0
oooo;P00002D0?ooo`00`@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo
00310?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0`03oool00<80oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`310?ooo`00`P3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo0<40oooo00330?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3o
ool0`P3oool00<<0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`320?ooo`00a03o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0<<0oooo00350?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0a03oool00<D0oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`340?ooo`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0
oooo00360?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0a@3oool00<L0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`360?ooo`00a`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo0<H0oooo00380?ooo`030000003oool0oooo01/0oooo00<000000?oo
o`3oool0a`3oool00<T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`380?ooo`00
b@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0<P0oooo003:0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0b@3oool00<X0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`390?ooo`00b`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
0<X0oooo003;0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0b`3oool00<`0oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`3;0?ooo`00c@3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0<`0oooo003=0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool0c03oool00<h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3=0?oo
o`00cP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0<d0oooo003?0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0cP3oool00=00oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`3?0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo0<l0oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=40
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00dP3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo0=40oooo003B0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0d@3oool00=<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3B
0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003D0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0
oooo0000003F0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003F0?ooo`030000
003oool000000=L0oooo003F0?ooo`030000003oool000000=L0oooo003G0?ooo`030000003oool0
oooo0=H0oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {123.688, 0}} -> {-0.028119, -0.0072184, \
0.003676, 0.003676}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28868 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00687322 0.952381 [
[ 0 0 0 0 ]
[ 1 .28868 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28868 L
0 .28868 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .00687 m
.05908 .00687 L
s
.05908 .00687 m
.07672 .03742 L
s
.07672 .03742 m
.09436 .00687 L
s
.09436 .00687 m
.12963 .00687 L
s
.12963 .00687 m
.14727 .03742 L
s
.14727 .03742 m
.12963 .06797 L
s
.12963 .06797 m
.1649 .06797 L
s
.1649 .06797 m
.18254 .09852 L
s
.18254 .09852 m
.20018 .06797 L
s
.20018 .06797 m
.23545 .06797 L
s
.23545 .06797 m
.21781 .03742 L
s
.21781 .03742 m
.23545 .00687 L
s
.23545 .00687 m
.27072 .00687 L
s
.27072 .00687 m
.28836 .03742 L
s
.28836 .03742 m
.306 .00687 L
s
.306 .00687 m
.34127 .00687 L
s
.34127 .00687 m
.35891 .03742 L
s
.35891 .03742 m
.34127 .06797 L
s
.34127 .06797 m
.37654 .06797 L
s
.37654 .06797 m
.39418 .09852 L
s
.39418 .09852 m
.37654 .12906 L
s
.37654 .12906 m
.34127 .12906 L
s
.34127 .12906 m
.35891 .15961 L
s
.35891 .15961 m
.34127 .19016 L
s
.34127 .19016 m
.37654 .19016 L
s
.37654 .19016 m
.39418 .22071 L
s
.39418 .22071 m
.41182 .19016 L
s
.41182 .19016 m
.44709 .19016 L
s
.44709 .19016 m
.46473 .22071 L
s
.46473 .22071 m
.44709 .25125 L
s
.44709 .25125 m
.48236 .25125 L
s
.48236 .25125 m
.5 .2818 L
s
.5 .2818 m
.51764 .25125 L
s
.51764 .25125 m
.55291 .25125 L
s
.55291 .25125 m
.53527 .22071 L
s
.53527 .22071 m
.55291 .19016 L
s
.55291 .19016 m
.58818 .19016 L
s
.58818 .19016 m
.60582 .22071 L
s
.60582 .22071 m
.62346 .19016 L
s
.62346 .19016 m
.65873 .19016 L
s
.65873 .19016 m
.64109 .15961 L
s
.64109 .15961 m
.65873 .12906 L
s
.65873 .12906 m
.62346 .12906 L
s
.62346 .12906 m
.60582 .09852 L
s
.60582 .09852 m
.62346 .06797 L
s
.62346 .06797 m
.65873 .06797 L
s
.65873 .06797 m
.64109 .03742 L
s
.64109 .03742 m
.65873 .00687 L
s
.65873 .00687 m
.694 .00687 L
s
.694 .00687 m
.71164 .03742 L
s
.71164 .03742 m
.72928 .00687 L
s
.72928 .00687 m
.76455 .00687 L
s
.76455 .00687 m
.78219 .03742 L
s
.78219 .03742 m
.76455 .06797 L
s
.76455 .06797 m
.79982 .06797 L
s
.79982 .06797 m
.81746 .09852 L
s
.81746 .09852 m
.8351 .06797 L
s
.8351 .06797 m
.87037 .06797 L
s
.87037 .06797 m
.85273 .03742 L
s
.85273 .03742 m
.87037 .00687 L
s
.87037 .00687 m
.90564 .00687 L
s
.90564 .00687 m
.92328 .03742 L
s
.92328 .03742 m
.94092 .00687 L
s
.94092 .00687 m
.97619 .00687 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 83.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000O2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002`3oool@000000h0oooo4000000]0?oooa000000
3P3oool@000008L0oooo4000000>0?oooa000000;@3oool@000000h0oooo4000000<0?ooo`006`3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:@3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`230?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01X0oooo000K0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo08<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool06P3oool001`0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0P@3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0K0?ooo`00703oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02L0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01/0
oooo000M0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`0U0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool09@3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0703oool001h0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool0O@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?ooo`007P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01d0oooo000O0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo07/0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool07P3oool00200oooo00D000000?ooo`3oool0oooo0000000I0?ooo`030000003oool0oooo
01l0oooo00<000000?ooo`3oool05`3oool01@000000oooo0?ooo`3oool0000001X0oooo00<00000
0?ooo`3oool0N@3oool00`000000oooo0?ooo`0H0?ooo`050000003oool0oooo0?ooo`0000006@3o
ool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01L0oooo00D000000?ooo`3oool0oooo
0000000Q0?ooo`00803oool01@000000oooo0?ooo`3oool0000001T0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0G0?ooo`050000003oool0oooo0?ooo`0000006P3oool00`000000
oooo0?ooo`1i0?ooo`030000003oool0oooo01P0oooo00D000000?ooo`3oool0oooo0000000I0?oo
o`030000003oool0oooo01l0oooo00<000000?ooo`3oool05`3oool01@000000oooo0?ooo`3oool0
00000240oooo000Q0?ooo`030000003oool0000001/0oooo00<000000?ooo`3oool07@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool0M`3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0000001/0oooo00<000000?ooo`3oool07@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool000000280oooo000Q0?ooo`030000003oool0000001/0
oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0000001`0
oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0000001/0
oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool000000280
oooo000R0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0M@3oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`0K0?ooo`030000003oool0oooo0240oooo000o0?ooo`030000003oool0oooo01d0oooo00<0
00000?ooo`3oool0>03oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo03P0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`0n0?ooo`00?`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0h
0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0?P3oool003h0oooo00<000000?oo
o`3oool07`3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo07T0oooo00<000000?oo
o`3oool0=P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo03d0oooo000n0?ooo`03
0000003oool0oooo01l0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`1i0?ooo`03
0000003oool0oooo03H0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0m0?ooo`00
?@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0
N`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
?03oool003`0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo07d0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo03/0oooo000l0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`1m0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool08`3oool0
0`000000oooo0?ooo`0k0?ooo`00>`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo
02D0oooo00<000000?ooo`3oool0>P3oool003X0oooo00<000000?ooo`3oool09`3oool00`000000
oooo0?ooo`0^0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo03T0oooo000j0?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0i0?ooo`00>@3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?oo
o`0/0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0>03oool003T0oooo00<00000
0?ooo`3oool0:@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo08<0oooo00<00000
0?ooo`3oool0;03oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo03P0oooo000h0?oo
oa0000003`3oool@000002`0oooo4000001Y0?oooa000000;03oool@000000l0oooo4000000i0?oo
o`00B03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3o
ool0I@3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0A`3oool004P0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo06D0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo04L0oooo00190?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0C03o
ool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`180?ooo`00B@3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo04`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0B03oool004X0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`1>0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0CP3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo04T0oooo001;0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0500
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1:0?ooo`00B`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0G`3oool00`000000oooo
0?ooo`1@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0BP3oool004`0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo05d0oooo00<0
00000?ooo`3oool0DP3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo04/0oooo001=
0?ooo`050000003oool0oooo0?ooo`000000EP3oool00`000000oooo0?ooo`1K0?ooo`030000003o
ool0oooo05@0oooo00D000000?ooo`3oool0oooo0000001>0?ooo`00C@3oool01@000000oooo0?oo
o`3oool0000005H0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`1D0?ooo`050000
003oool0oooo0?ooo`000000CP3oool004h0oooo00<000000?ooo`000000F03oool00`000000oooo
0?ooo`1I0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`000000C`3oool004h0oooo00<0
00000?ooo`000000F03oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo05H0oooo00<0
00000?ooo`000000C`3oool004l0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0CP3oool00:T0oooo00<000000?oo
o`3oool0F@3oool00`000000oooo0?ooo`2X0?ooo`00Z@3oool00`000000oooo0?ooo`1I0?ooo`03
0000003oool0oooo0:P0oooo002X0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0
Y`3oool00:P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`2W0?ooo`00Y`3oool0
0`000000oooo0?ooo`1M0?ooo`030000003oool0oooo0:H0oooo002V0?ooo`030000003oool0oooo
05l0oooo00<000000?ooo`3oool0Y@3oool00:H0oooo00<000000?ooo`3oool0G`3oool00`000000
oooo0?ooo`2U0?ooo`00Y@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo0:@0oooo
002T0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0X`3oool00:@0oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`2S0?ooo`00X`3oool00`000000oooo0?ooo`1U0?oo
o`030000003oool0oooo0:80oooo002S0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3o
ool0XP3oool009<0oooo4000001Y0?oooa000000U03oool009@0oooo00<000000?ooo`3oool0P`3o
ool00`000000oooo0?ooo`2C0?ooo`00U03oool00`000000oooo0?ooo`230?ooo`030000003oool0
oooo09<0oooo002E0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0U03oool009D0
oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2D0?ooo`00UP3oool00`000000oooo
0?ooo`1o0?ooo`030000003oool0oooo09D0oooo002G0?ooo`030000003oool0oooo07d0oooo00<0
00000?ooo`3oool0UP3oool009L0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`2F
0?ooo`00V03oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo09L0oooo002I0?ooo`03
0000003oool0oooo07T0oooo00<000000?ooo`3oool0V03oool009T0oooo00<000000?ooo`3oool0
N@3oool00`000000oooo0?ooo`2H0?ooo`00VP3oool00`000000oooo0?ooo`1g0?ooo`030000003o
ool0oooo09T0oooo002J0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0V@3oool0
09/0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`2J0?ooo`00VP3oool00`000000
oooo0?ooo`1g0?ooo`030000003oool0oooo09T0oooo002J0?ooo`030000003oool0oooo07L0oooo
00<000000?ooo`3oool0V@3oool009T0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?oo
o`2H0?ooo`00V@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo09P0oooo002H0?oo
o`030000003oool0oooo07/0oooo00<000000?ooo`3oool0U`3oool009L0oooo00<000000?ooo`3o
ool0O@3oool00`000000oooo0?ooo`2F0?ooo`00U`3oool00`000000oooo0?ooo`1m0?ooo`030000
003oool0oooo09H0oooo002F0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0U@3o
ool009D0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2D0?ooo`00U@3oool00`00
0000oooo0?ooo`210?ooo`030000003oool0oooo09@0oooo002D0?ooo`030000003oool0oooo08<0
oooo00<000000?ooo`3oool0T`3oool009@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo
0?ooo`2C0?ooo`00T`3oool@000000h0oooo4000000]0?oooa0000003P3oool@000009@0oooo002S
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Y
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`2R
0?ooo`00X`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool0XP3oool00:@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo0:<0oooo002T0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`2S0?ooo`00Y@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool0Y03oool00:H0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0:D0oooo002V0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2U0?ooo`00Y`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0YP3oool00:P0oooo
00D000000?ooo`3oool0oooo0000000I0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3o
ool05`3oool01@000000oooo0?ooo`3oool000000:T0oooo002X0?ooo`050000003oool0oooo0?oo
o`0000006@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01L0oooo00D000000?oo
o`3oool0oooo0000002Y0?ooo`00Z@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo
01d0oooo00<000000?ooo`3oool06@3oool00`000000oooo0000002Z0?ooo`00Z@3oool00`000000
oooo0000000K0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0000002Z0?ooo`00ZP3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2Y0?ooo`00a`3oool00`000000oooo0?oo
o`0M0?ooo`030000003oool0oooo0<H0oooo00370?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool0aP3oool00<H0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`350?oo
o`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00350?ooo`030000
003oool0oooo0240oooo00<000000?ooo`3oool0a03oool00<@0oooo00<000000?ooo`3oool08`3o
ool00`000000oooo0?ooo`330?ooo`00a03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0
oooo0<<0oooo00330?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0`P3oool00<80
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`310?ooo`00`P3oool00`000000oooo
0?ooo`0W0?ooo`030000003oool0oooo0<40oooo00310?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool0`03oool00<40oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`30
0?ooo`00`03oool@000000l0oooo400000310?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo0<l0oooo003@0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
c`3oool00=40oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00d@3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo0=00oooo003B0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0d@3oool00=80oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`3A0?ooo`00d`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0=80oooo
003D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0d`3oool00=@0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`3C0?ooo`00e@3oool01@000000oooo0?ooo`3oool0
00000=H0oooo003E0?ooo`050000003oool0oooo0?ooo`000000eP3oool00=H0oooo00<000000?oo
o`000000e`3oool00=H0oooo00<000000?ooo`000000e`3oool00=L0oooo00<000000?ooo`3oool0
eP3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {123.688, 0}} -> {-0.028119, -0.0072184, \
0.003676, 0.003676}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .28868 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00687322 0.952381 [
[ 0 0 0 0 ]
[ 1 .28868 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .28868 L
0 .28868 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .00687 m
.03557 .00687 L
s
.03557 .00687 m
.04145 .01706 L
s
.04145 .01706 m
.04733 .00687 L
s
.04733 .00687 m
.05908 .00687 L
s
.05908 .00687 m
.06496 .01706 L
s
.06496 .01706 m
.05908 .02724 L
s
.05908 .02724 m
.07084 .02724 L
s
.07084 .02724 m
.07672 .03742 L
s
.07672 .03742 m
.0826 .02724 L
s
.0826 .02724 m
.09436 .02724 L
s
.09436 .02724 m
.08848 .01706 L
s
.08848 .01706 m
.09436 .00687 L
s
.09436 .00687 m
.10611 .00687 L
s
.10611 .00687 m
.11199 .01706 L
s
.11199 .01706 m
.11787 .00687 L
s
.11787 .00687 m
.12963 .00687 L
s
.12963 .00687 m
.13551 .01706 L
s
.13551 .01706 m
.12963 .02724 L
s
.12963 .02724 m
.14139 .02724 L
s
.14139 .02724 m
.14727 .03742 L
s
.14727 .03742 m
.14139 .0476 L
s
.14139 .0476 m
.12963 .0476 L
s
.12963 .0476 m
.13551 .05779 L
s
.13551 .05779 m
.12963 .06797 L
s
.12963 .06797 m
.14139 .06797 L
s
.14139 .06797 m
.14727 .07815 L
s
.14727 .07815 m
.15315 .06797 L
s
.15315 .06797 m
.1649 .06797 L
s
.1649 .06797 m
.17078 .07815 L
s
.17078 .07815 m
.1649 .08833 L
s
.1649 .08833 m
.17666 .08833 L
s
.17666 .08833 m
.18254 .09852 L
s
.18254 .09852 m
.18842 .08833 L
s
.18842 .08833 m
.20018 .08833 L
s
.20018 .08833 m
.1943 .07815 L
s
.1943 .07815 m
.20018 .06797 L
s
.20018 .06797 m
.21193 .06797 L
s
.21193 .06797 m
.21781 .07815 L
s
.21781 .07815 m
.22369 .06797 L
s
.22369 .06797 m
.23545 .06797 L
s
.23545 .06797 m
.22957 .05779 L
s
.22957 .05779 m
.23545 .0476 L
s
.23545 .0476 m
.22369 .0476 L
s
.22369 .0476 m
.21781 .03742 L
s
.21781 .03742 m
.22369 .02724 L
s
.22369 .02724 m
.23545 .02724 L
s
.23545 .02724 m
.22957 .01706 L
s
.22957 .01706 m
.23545 .00687 L
s
.23545 .00687 m
.24721 .00687 L
s
.24721 .00687 m
.25309 .01706 L
s
.25309 .01706 m
.25897 .00687 L
s
.25897 .00687 m
.27072 .00687 L
s
.27072 .00687 m
.2766 .01706 L
s
.2766 .01706 m
.27072 .02724 L
s
.27072 .02724 m
.28248 .02724 L
s
.28248 .02724 m
.28836 .03742 L
s
.28836 .03742 m
.29424 .02724 L
s
.29424 .02724 m
.306 .02724 L
s
.306 .02724 m
.30012 .01706 L
s
.30012 .01706 m
.306 .00687 L
s
.306 .00687 m
.31775 .00687 L
s
.31775 .00687 m
.32363 .01706 L
s
.32363 .01706 m
.32951 .00687 L
s
.32951 .00687 m
.34127 .00687 L
s
.34127 .00687 m
.34715 .01706 L
s
.34715 .01706 m
.34127 .02724 L
s
.34127 .02724 m
.35303 .02724 L
s
.35303 .02724 m
.35891 .03742 L
s
.35891 .03742 m
.35303 .0476 L
s
.35303 .0476 m
.34127 .0476 L
s
.34127 .0476 m
.34715 .05779 L
s
.34715 .05779 m
.34127 .06797 L
s
.34127 .06797 m
.35303 .06797 L
s
.35303 .06797 m
.35891 .07815 L
s
.35891 .07815 m
.36479 .06797 L
s
.36479 .06797 m
.37654 .06797 L
s
.37654 .06797 m
.38242 .07815 L
s
.38242 .07815 m
.37654 .08833 L
s
.37654 .08833 m
.3883 .08833 L
s
.3883 .08833 m
.39418 .09852 L
s
.39418 .09852 m
.3883 .1087 L
s
.3883 .1087 m
.37654 .1087 L
s
.37654 .1087 m
.38242 .11888 L
s
.38242 .11888 m
.37654 .12906 L
s
.37654 .12906 m
.36479 .12906 L
s
.36479 .12906 m
.35891 .11888 L
s
.35891 .11888 m
.35303 .12906 L
s
.35303 .12906 m
.34127 .12906 L
s
.34127 .12906 m
.34715 .13925 L
s
.34715 .13925 m
.34127 .14943 L
s
.34127 .14943 m
.35303 .14943 L
s
.35303 .14943 m
.35891 .15961 L
s
.35891 .15961 m
.35303 .16979 L
s
.35303 .16979 m
.34127 .16979 L
s
.34127 .16979 m
.34715 .17998 L
s
.34715 .17998 m
.34127 .19016 L
s
.34127 .19016 m
.35303 .19016 L
s
.35303 .19016 m
.35891 .20034 L
s
.35891 .20034 m
.36479 .19016 L
s
.36479 .19016 m
.37654 .19016 L
s
.37654 .19016 m
.38242 .20034 L
s
.38242 .20034 m
.37654 .21052 L
s
.37654 .21052 m
.3883 .21052 L
s
.3883 .21052 m
.39418 .22071 L
s
.39418 .22071 m
.40006 .21052 L
s
.40006 .21052 m
.41182 .21052 L
s
.41182 .21052 m
.40594 .20034 L
s
.40594 .20034 m
.41182 .19016 L
s
.41182 .19016 m
.42357 .19016 L
s
.42357 .19016 m
.42945 .20034 L
s
.42945 .20034 m
.43533 .19016 L
s
.43533 .19016 m
.44709 .19016 L
s
.44709 .19016 m
.45297 .20034 L
s
.45297 .20034 m
.44709 .21052 L
s
.44709 .21052 m
.45885 .21052 L
s
.45885 .21052 m
.46473 .22071 L
s
.46473 .22071 m
.45885 .23089 L
s
.45885 .23089 m
.44709 .23089 L
s
.44709 .23089 m
.45297 .24107 L
s
.45297 .24107 m
.44709 .25125 L
s
.44709 .25125 m
.45885 .25125 L
s
.45885 .25125 m
.46473 .26144 L
s
.46473 .26144 m
.47061 .25125 L
s
.47061 .25125 m
.48236 .25125 L
s
.48236 .25125 m
.48824 .26144 L
s
.48824 .26144 m
.48236 .27162 L
s
.48236 .27162 m
.49412 .27162 L
s
.49412 .27162 m
.5 .2818 L
s
.5 .2818 m
.50588 .27162 L
s
.50588 .27162 m
.51764 .27162 L
s
.51764 .27162 m
.51176 .26144 L
s
.51176 .26144 m
.51764 .25125 L
s
.51764 .25125 m
.52939 .25125 L
s
.52939 .25125 m
.53527 .26144 L
s
.53527 .26144 m
.54115 .25125 L
s
.54115 .25125 m
.55291 .25125 L
s
.55291 .25125 m
.54703 .24107 L
s
.54703 .24107 m
.55291 .23089 L
s
.55291 .23089 m
.54115 .23089 L
s
.54115 .23089 m
.53527 .22071 L
s
.53527 .22071 m
.54115 .21052 L
s
.54115 .21052 m
.55291 .21052 L
s
.55291 .21052 m
.54703 .20034 L
s
.54703 .20034 m
.55291 .19016 L
s
.55291 .19016 m
.56467 .19016 L
s
.56467 .19016 m
.57055 .20034 L
s
.57055 .20034 m
.57643 .19016 L
s
.57643 .19016 m
.58818 .19016 L
s
.58818 .19016 m
.59406 .20034 L
s
.59406 .20034 m
.58818 .21052 L
s
.58818 .21052 m
.59994 .21052 L
s
.59994 .21052 m
.60582 .22071 L
s
.60582 .22071 m
.6117 .21052 L
s
.6117 .21052 m
.62346 .21052 L
s
.62346 .21052 m
.61758 .20034 L
s
.61758 .20034 m
.62346 .19016 L
s
.62346 .19016 m
.63521 .19016 L
s
.63521 .19016 m
.64109 .20034 L
s
.64109 .20034 m
.64697 .19016 L
s
.64697 .19016 m
.65873 .19016 L
s
.65873 .19016 m
.65285 .17998 L
s
.65285 .17998 m
.65873 .16979 L
s
.65873 .16979 m
.64697 .16979 L
s
.64697 .16979 m
.64109 .15961 L
s
.64109 .15961 m
.64697 .14943 L
s
.64697 .14943 m
.65873 .14943 L
s
.65873 .14943 m
.65285 .13925 L
s
.65285 .13925 m
.65873 .12906 L
s
.65873 .12906 m
.64697 .12906 L
s
.64697 .12906 m
.64109 .11888 L
s
.64109 .11888 m
.63521 .12906 L
s
.63521 .12906 m
.62346 .12906 L
s
.62346 .12906 m
.61758 .11888 L
s
.61758 .11888 m
.62346 .1087 L
s
.62346 .1087 m
.6117 .1087 L
s
.6117 .1087 m
.60582 .09852 L
s
.60582 .09852 m
.6117 .08833 L
s
.6117 .08833 m
.62346 .08833 L
s
.62346 .08833 m
.61758 .07815 L
s
.61758 .07815 m
.62346 .06797 L
s
.62346 .06797 m
.63521 .06797 L
s
.63521 .06797 m
.64109 .07815 L
s
.64109 .07815 m
.64697 .06797 L
s
.64697 .06797 m
.65873 .06797 L
s
.65873 .06797 m
.65285 .05779 L
s
.65285 .05779 m
.65873 .0476 L
s
.65873 .0476 m
.64697 .0476 L
s
.64697 .0476 m
.64109 .03742 L
s
.64109 .03742 m
.64697 .02724 L
s
.64697 .02724 m
.65873 .02724 L
s
.65873 .02724 m
.65285 .01706 L
s
.65285 .01706 m
.65873 .00687 L
s
.65873 .00687 m
.67049 .00687 L
s
.67049 .00687 m
.67637 .01706 L
s
.67637 .01706 m
.68225 .00687 L
s
.68225 .00687 m
.694 .00687 L
s
.694 .00687 m
.69988 .01706 L
s
.69988 .01706 m
.694 .02724 L
s
.694 .02724 m
.70576 .02724 L
s
.70576 .02724 m
.71164 .03742 L
s
.71164 .03742 m
.71752 .02724 L
s
.71752 .02724 m
.72928 .02724 L
s
.72928 .02724 m
.7234 .01706 L
s
.7234 .01706 m
.72928 .00687 L
s
.72928 .00687 m
.74103 .00687 L
s
.74103 .00687 m
.74691 .01706 L
s
.74691 .01706 m
.75279 .00687 L
s
.75279 .00687 m
.76455 .00687 L
s
.76455 .00687 m
.77043 .01706 L
s
.77043 .01706 m
.76455 .02724 L
s
.76455 .02724 m
.77631 .02724 L
s
.77631 .02724 m
.78219 .03742 L
s
.78219 .03742 m
.77631 .0476 L
s
.77631 .0476 m
.76455 .0476 L
s
.76455 .0476 m
.77043 .05779 L
s
.77043 .05779 m
.76455 .06797 L
s
.76455 .06797 m
.77631 .06797 L
s
.77631 .06797 m
.78219 .07815 L
s
.78219 .07815 m
.78807 .06797 L
s
.78807 .06797 m
.79982 .06797 L
s
.79982 .06797 m
.8057 .07815 L
s
.8057 .07815 m
.79982 .08833 L
s
.79982 .08833 m
.81158 .08833 L
s
.81158 .08833 m
.81746 .09852 L
s
.81746 .09852 m
.82334 .08833 L
s
.82334 .08833 m
.8351 .08833 L
s
.8351 .08833 m
.82922 .07815 L
s
.82922 .07815 m
.8351 .06797 L
s
.8351 .06797 m
.84685 .06797 L
s
.84685 .06797 m
.85273 .07815 L
s
.85273 .07815 m
.85861 .06797 L
s
.85861 .06797 m
.87037 .06797 L
s
.87037 .06797 m
.86449 .05779 L
s
.86449 .05779 m
.87037 .0476 L
s
.87037 .0476 m
.85861 .0476 L
s
.85861 .0476 m
.85273 .03742 L
s
.85273 .03742 m
.85861 .02724 L
s
.85861 .02724 m
.87037 .02724 L
s
.87037 .02724 m
.86449 .01706 L
s
.86449 .01706 m
.87037 .00687 L
s
.87037 .00687 m
.88213 .00687 L
s
.88213 .00687 m
.88801 .01706 L
s
.88801 .01706 m
.89389 .00687 L
s
.89389 .00687 m
.90564 .00687 L
s
.90564 .00687 m
.91152 .01706 L
s
.91152 .01706 m
.90564 .02724 L
s
.90564 .02724 m
.9174 .02724 L
s
.9174 .02724 m
.92328 .03742 L
s
.92328 .03742 m
.92916 .02724 L
s
.92916 .02724 m
.94092 .02724 L
s
.94092 .02724 m
.93504 .01706 L
s
.93504 .01706 m
.94092 .00687 L
s
.94092 .00687 m
.95267 .00687 L
s
.95267 .00687 m
.95855 .01706 L
s
.95855 .01706 m
.96443 .00687 L
s
.96443 .00687 m
.97619 .00687 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 83.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000O2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002`3oool6000000@0oooo1P00000>0?ooo`H00000
103oool6000002d0oooo1P0000040?ooo`H000003P3oool6000000@0oooo1P0000270?ooo`H00000
103oool6000000h0oooo1P0000040?ooo`H00000;@3oool6000000@0oooo1P00000>0?ooo`H00000
103oool6000000`0oooo000A0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3o
ool0:@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo00000006
0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00
00060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool010000000oooo0?oo
o`0000060?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool0000000D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`05
0000003oool0oooo0?ooo`0000004P3oool00140oooo00@000000?ooo`3oool000001P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`220?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3oool000001P3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000
1P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000
1`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000
4`3oool00180oooo0P0000080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3o
ool2000000P0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0000000P0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`060?ooo`800000203o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000080?ooo`030000003o
ool0oooo02L0oooo00<000000?ooo`3oool01@3oool00`000000oooo000000070?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000C0?ooo`004`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0B0?ooo`00703oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
01/0oooo000K0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0X0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:03oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool06`3oool001/0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02T0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0J0?ooo`006P3oool6000000@0oooo1P00
000>0?ooo`H000008`3oool6000000h0oooo1P0000040?ooo`H000003P3oool6000007d0oooo1P00
000>0?ooo`H00000103oool6000000h0oooo1P00000S0?ooo`H000003P3oool6000000@0oooo1P00
000K0?ooo`00803oool01@000000oooo0?ooo`3oool0000001T0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0G0?ooo`050000003oool0oooo0?ooo`0000006P3oool00`000000oooo
0?ooo`1i0?ooo`030000003oool0oooo01P0oooo00D000000?ooo`3oool0oooo0000000I0?ooo`03
0000003oool0oooo01l0oooo00<000000?ooo`3oool05`3oool01@000000oooo0?ooo`3oool00000
0240oooo000P0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool07`3oool00`00
0000oooo0?ooo`0H0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool0N@3oool0
0`000000oooo0?ooo`0H0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0H0?ooo`040000003oool0oooo00000240oooo000Q0?ooo`030000003o
ool0000001/0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0000001`0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0000001/0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool000000280oooo000Q0?ooo`800000703oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo01X0oooo0P00000L0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool06P3oool2
000001`0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0J0?ooo`8000008P3oool0
0280oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo
01`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool08@3oool003l0oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`0h0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0>03oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo03h0oooo000n0?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo03L0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0n0?ooo`00?P3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?oo
o`0f0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0?@3oool003P0oooo1P00000S
0?ooo`H00000;03oool6000007d0oooo1P00000/0?ooo`H000008`3oool6000003T0oooo000i0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`230?oo
o`030000003oool0oooo02`0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0h0?oo
o`00>@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3o
ool0P`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3o
ool0>03oool003X0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0^0?ooo`030000
003oool0oooo0840oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0W0?ooo`030000
003oool0oooo03T0oooo000j0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0;P3o
ool00`000000oooo0?ooo`210?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`0i0?ooo`00>`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0
oooo02D0oooo00<000000?ooo`3oool0>P3oool003X0oooo00<000000?ooo`3oool09`3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo03T0oooo000i0?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo02d0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0i0?ooo`00>@3oool00`000000oooo
0?ooo`0Y0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0P`3oool00`000000oooo
0?ooo`0/0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0>03oool003P0oooo1P00
00040?ooo`H000003`3oool6000000@0oooo1P00000/0?ooo`H00000103oool6000006T0oooo1P00
00040?ooo`H00000;03oool6000000@0oooo1P00000?0?ooo`H00000103oool6000003T0oooo000n
0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`000000>03oool010000000oooo0?ooo`0000060?ooo`03
0000003oool0oooo06D0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool00000
03P0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00<0oooo00D000000?ooo`3oool0oooo0000000o0?ooo`00?P3oool010000000oooo0?ooo`00
00060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000
000i0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool0I03oool00`000000oooo
0?ooo`050?ooo`030000003oool0000003T0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000@03oool003l0oooo
0P0000080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0000000j0?ooo`800000203oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`000000>P3oool2000000P0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`050?ooo`030000003oool000000400oooo00100?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03X0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`0o0?ooo`00B@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`1<
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0B03oool004T0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo06<0oooo00<000000?oo
o`3oool0C03oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo04P0oooo00180?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`1U0?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`170?ooo`00
B03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
I@3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
A`3oool004L0oooo1P0000050?ooo`H00000BP3oool6000005l0oooo1P00001:0?ooo`H000001@3o
ool6000004P0oooo001=0?ooo`050000003oool0oooo0?ooo`000000EP3oool00`000000oooo0?oo
o`1K0?ooo`030000003oool0oooo05@0oooo00D000000?ooo`3oool0oooo0000001>0?ooo`00C@3o
ool010000000oooo0?ooo`00001G0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0
E@3oool010000000oooo0?ooo`00001?0?ooo`00CP3oool00`000000oooo0000001H0?ooo`030000
003oool0oooo05T0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0000001?0?ooo`00C`3o
ool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0F03o
ool00`000000oooo0?ooo`1>0?ooo`00Z@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0
oooo0:P0oooo002X0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0Z03oool00:P0
oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`2W0?ooo`00XP3oool6000005l0oooo
1P00002S0?ooo`00X`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0:80oooo002S
0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0XP3oool00:@0oooo00<000000?oo
o`3oool0H`3oool00`000000oooo0?ooo`2S0?ooo`00Y03oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0:<0oooo002K0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
H@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo09X0oooo002J0?ooo`800000203o
ool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000V`3o
ool009T0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`1T0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`000000V`3oool009T0oooo00@000000?ooo`3oool000001P3o
ool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo
0000002J0?ooo`00T`3oool6000000@0oooo1P00001Y0?ooo`H00000103oool6000009@0oooo002D
0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0T`3oool009@0oooo00<000000?oo
o`3oool0PP3oool00`000000oooo0?ooo`2D0?ooo`00U@3oool00`000000oooo0?ooo`210?ooo`03
0000003oool0oooo09@0oooo002F0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0
U@3oool009D0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2D0?ooo`00U@3oool0
0`000000oooo0?ooo`210?ooo`030000003oool0oooo09@0oooo002D0?ooo`030000003oool0oooo
08<0oooo00<000000?ooo`3oool0T`3oool009@0oooo00<000000?ooo`3oool0P`3oool00`000000
oooo0?ooo`2C0?ooo`00T`3oool6000007d0oooo1P00002D0?ooo`00V@3oool00`000000oooo0?oo
o`1i0?ooo`030000003oool0oooo09P0oooo002I0?ooo`030000003oool0oooo07P0oooo00<00000
0?ooo`3oool0V@3oool009X0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`2I0?oo
o`00V`3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo09X0oooo002J0?ooo`030000
003oool0oooo07L0oooo00<000000?ooo`3oool0V@3oool009X0oooo00<000000?ooo`3oool0M`3o
ool00`000000oooo0?ooo`2I0?ooo`00V@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0
oooo09P0oooo002I0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0V03oool009<0
oooo1P00001m0?ooo`H00000U03oool009@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo
0?ooo`2C0?ooo`00U03oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo09@0oooo002E
0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0U03oool009H0oooo00<000000?oo
o`3oool0O`3oool00`000000oooo0?ooo`2E0?ooo`00U@3oool00`000000oooo0?ooo`210?ooo`03
0000003oool0oooo09@0oooo002D0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0
U03oool009@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`2C0?ooo`00T`3oool6
000000@0oooo1P00000>0?ooo`H00000103oool6000002d0oooo1P0000040?ooo`H000003P3oool6
000000@0oooo1P00002D0?ooo`00V@3oool01@000000oooo0?ooo`3oool0000000D0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo
000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool0000009X0oooo002I0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000009X0oooo002J
0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0000009/0oooo002J0?ooo`800000203oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00H0oooo0P0000080?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool01P3oool2000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?oo
o`800000V`3oool009/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0U0?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool0VP3oool00:@0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0:<0oooo002S0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0X0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`2S0?ooo`00X`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0XP3oool0
0:80oooo1P0000040?ooo`H000003P3oool6000002<0oooo1P00000>0?ooo`H00000103oool60000
0:<0oooo002X0?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool07`3oool00`00
0000oooo0?ooo`0G0?ooo`050000003oool0oooo0?ooo`000000Z@3oool00:P0oooo00@000000?oo
o`3oool000006P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`000000ZP3oool00:T0oooo0P00000L0?ooo`030000003oool0oooo01d0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0000002Z0?ooo`00ZP3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2Y0?ooo`00
a`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0<H0oooo00370?ooo`030000003o
ool0oooo01d0oooo00<000000?ooo`3oool0aP3oool00<H0oooo00<000000?ooo`3oool07`3oool0
0`000000oooo0?ooo`350?ooo`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
0<D0oooo00300?ooo`H000008`3oool600000<40oooo00310?ooo`030000003oool0oooo02T0oooo
00<000000?ooo`3oool0`03oool00<40oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`310?ooo`00`P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0<40oooo00330?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool0`P3oool00<80oooo00<000000?ooo`3o
ool09`3oool00`000000oooo0?ooo`310?ooo`00`@3oool00`000000oooo0?ooo`0X0?ooo`030000
003oool0oooo0<40oooo00310?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0`03o
ool00<00oooo1P0000040?ooo`H000003`3oool6000000@0oooo1P0000310?ooo`00aP3oool01@00
0000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`03
0?ooo`050000003oool0oooo0?ooo`000000a`3oool00<H0oooo00@000000?ooo`3oool000001P3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000
a`3oool00<L0oooo00<000000?ooo`0000001`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`000000b03oool00<L0oooo0P0000080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool01P3oool200000<P0oooo00380?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
0<L0oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=00oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`3@0?ooo`00d03oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo0<l0oooo003?0?ooo`H000001@3oool600000=00oooo003E0?oo
o`050000003oool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F
0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00
e`3oool00`000000oooo0?ooo`3F0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {123.688, 0}} -> {-0.028119, -0.0072184, \
0.003676, 0.003676}}]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["More elaborate display function", "Subsection"],

Cell[TextData[{
  "Now that we have an algorithm which produces and displays Koch curves, we \
can write a more elaborate function which permits several variations to be \
chosen.  We use ",
  StyleBox["Options", "InlineInput",
    FontWeight->"Bold"],
  " to establish default options for our function and the ",
  StyleBox["FilterOptions", "InlineInput",
    FontWeight->"Bold"],
  " package to distinguish between our options and those applicable to ",
  StyleBox["Graphics", "InlineInput",
    FontWeight->"Bold"],
  ".  We also provide usage messages for the function and its options."
}], "Text"],

Cell[BoxData[
    \(\(Needs["\<Utilities`FilterOptions`\>"];\)\)], "Input"],

Cell[BoxData[{
    \(\(Clear[DisplayKoch];\)\), "\n", 
    \(\(DisplayKoch::usage = "\<DisplayKoch[polygon,n,opts] produces a Koch \
curve of order n based upon the list of points in polygon. By default, the \
triangle rule is applied in the outward direction assuming that the polygon \
is defined counterclockwise.
\n  Options: 
\n    KochType\[Rule]\"sieve\" applies the triangle-building rule in reverse.
\n    ShowIterations\[Rule]True can be used to display intermediate \
iterations.
\n    CloseFigure\[Rule]True connects the last point to the first to close \
the figure.\>";\)\n\), "\n", 
    \(\(ShowIterations::usage = "\<a True|False option for DisplayKoch.\>";\)\
\[IndentingNewLine]\), "\n", 
    \(\(KochType::usage = "\<an option for DisplayKoch: 
\n  \"snowflake\" produces outward triangles 
\n  \"sieve\" produces inward triangles\>";\)\n\), "\[IndentingNewLine]", 
    \(\(CloseFigure::usage = "\<a True|False option for DisplayKoch that can \
be used when the list of points is not a closed polygon.\>";\)\
\[IndentingNewLine]\), "\n", 
    \(\(Options[DisplayKoch] = {ShowIterations \[Rule] False, 
          KochType \[Rule] "\<snowflake\>", 
          CloseFigure \[Rule] False};\)\n\), "\n", 
    \(DisplayKoch[polygon : {{_, _} .. }, n_, opts___Rule] := 
      Module[{curve, type, showIterations, close, NestFunction, optKoch, 
          optPlot}, \n\t\n (*\ 
          establish\ options\ *) \[IndentingNewLine]\n{type, showIterations, 
            close} = \({KochType, ShowIterations, CloseFigure} /. {opts}\) /. 
            Options[DisplayKoch]; \n\t\t\n (*\ 
          establish\ initial\ line\ *) \[IndentingNewLine]\n
        curve = \ If[type \[Equal] "\<sieve\>", Reverse[polygon], polygon]; \n
        If[close, AppendTo[curve, First[curve]]]; \n
        curve = Line /@ Partition[curve, 2, 1]; \n\t\n (*\ 
          apply\ Koch\ algorithm\ and\ produce\ figure\ \
*) \[IndentingNewLine]\nNestFunction = If[showIterations, NestList, Nest]; \n
        optPlot = FilterOptions[Graphics, opts]; \n
        Show[Graphics[NestFunction[KochSplit, curve, n]], optPlot, 
          AspectRatio \[Rule] Automatic]]\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayKoch[{{0. , 0. }, {0.5, \@3\/2. }, {1. , 0. }}, 4, 
        ShowIterations \[Rule] False, KochType \[Rule] "\<sieve\>", 
        CloseFigure \[Rule] True];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .86603 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0206197 0.952381 [
[ 0 0 0 0 ]
[ 1 .86603 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .86603 L
0 .86603 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.97619 .02062 m
.97031 .0308 L
s
.97031 .0308 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .04098 L
s
.96443 .04098 m
.95855 .05117 L
s
.95855 .05117 m
.9468 .05117 L
s
.9468 .05117 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .05117 L
s
.93504 .05117 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .06135 L
s
.92916 .06135 m
.92328 .07153 L
s
.92328 .07153 m
.93504 .07153 L
s
.93504 .07153 m
.94092 .08171 L
s
.94092 .08171 m
.93504 .0919 L
s
.93504 .0919 m
.92328 .0919 L
s
.92328 .0919 m
.92916 .10208 L
s
.92916 .10208 m
.92328 .11226 L
s
.92328 .11226 m
.91152 .11226 L
s
.91152 .11226 m
.90564 .10208 L
s
.90564 .10208 m
.89976 .11226 L
s
.89976 .11226 m
.88801 .11226 L
s
.88801 .11226 m
.88213 .10208 L
s
.88213 .10208 m
.88801 .0919 L
s
.88801 .0919 m
.87625 .0919 L
s
.87625 .0919 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .0919 L
s
.86449 .0919 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .10208 L
s
.85861 .10208 m
.85273 .11226 L
s
.85273 .11226 m
.84098 .11226 L
s
.84098 .11226 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .11226 L
s
.82922 .11226 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .12245 L
s
.82334 .12245 m
.81746 .13263 L
s
.81746 .13263 m
.82922 .13263 L
s
.82922 .13263 m
.8351 .14281 L
s
.8351 .14281 m
.82922 .15299 L
s
.82922 .15299 m
.81746 .15299 L
s
.81746 .15299 m
.82334 .16318 L
s
.82334 .16318 m
.81746 .17336 L
s
.81746 .17336 m
.82922 .17336 L
s
.82922 .17336 m
.8351 .18354 L
s
.8351 .18354 m
.84098 .17336 L
s
.84098 .17336 m
.85273 .17336 L
s
.85273 .17336 m
.85861 .18354 L
s
.85861 .18354 m
.85273 .19372 L
s
.85273 .19372 m
.86449 .19372 L
s
.86449 .19372 m
.87037 .20391 L
s
.87037 .20391 m
.86449 .21409 L
s
.86449 .21409 m
.85273 .21409 L
s
.85273 .21409 m
.85861 .22427 L
s
.85861 .22427 m
.85273 .23445 L
s
.85273 .23445 m
.84098 .23445 L
s
.84098 .23445 m
.8351 .22427 L
s
.8351 .22427 m
.82922 .23445 L
s
.82922 .23445 m
.81746 .23445 L
s
.81746 .23445 m
.82334 .24464 L
s
.82334 .24464 m
.81746 .25482 L
s
.81746 .25482 m
.82922 .25482 L
s
.82922 .25482 m
.8351 .265 L
s
.8351 .265 m
.82922 .27518 L
s
.82922 .27518 m
.81746 .27518 L
s
.81746 .27518 m
.82334 .28537 L
s
.82334 .28537 m
.81746 .29555 L
s
.81746 .29555 m
.8057 .29555 L
s
.8057 .29555 m
.79982 .28537 L
s
.79982 .28537 m
.79394 .29555 L
s
.79394 .29555 m
.78219 .29555 L
s
.78219 .29555 m
.77631 .28537 L
s
.77631 .28537 m
.78219 .27518 L
s
.78219 .27518 m
.77043 .27518 L
s
.77043 .27518 m
.76455 .265 L
s
.76455 .265 m
.75867 .27518 L
s
.75867 .27518 m
.74691 .27518 L
s
.74691 .27518 m
.75279 .28537 L
s
.75279 .28537 m
.74691 .29555 L
s
.74691 .29555 m
.73516 .29555 L
s
.73516 .29555 m
.72928 .28537 L
s
.72928 .28537 m
.7234 .29555 L
s
.7234 .29555 m
.71164 .29555 L
s
.71164 .29555 m
.70576 .28537 L
s
.70576 .28537 m
.71164 .27518 L
s
.71164 .27518 m
.69988 .27518 L
s
.69988 .27518 m
.694 .265 L
s
.694 .265 m
.69988 .25482 L
s
.69988 .25482 m
.71164 .25482 L
s
.71164 .25482 m
.70576 .24464 L
s
.70576 .24464 m
.71164 .23445 L
s
.71164 .23445 m
.69988 .23445 L
s
.69988 .23445 m
.694 .22427 L
s
.694 .22427 m
.68812 .23445 L
s
.68812 .23445 m
.67637 .23445 L
s
.67637 .23445 m
.67049 .22427 L
s
.67049 .22427 m
.67637 .21409 L
s
.67637 .21409 m
.66461 .21409 L
s
.66461 .21409 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .21409 L
s
.65285 .21409 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .22427 L
s
.64697 .22427 m
.64109 .23445 L
s
.64109 .23445 m
.62934 .23445 L
s
.62934 .23445 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .23445 L
s
.61758 .23445 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .24464 L
s
.6117 .24464 m
.60582 .25482 L
s
.60582 .25482 m
.61758 .25482 L
s
.61758 .25482 m
.62346 .265 L
s
.62346 .265 m
.61758 .27518 L
s
.61758 .27518 m
.60582 .27518 L
s
.60582 .27518 m
.6117 .28537 L
s
.6117 .28537 m
.60582 .29555 L
s
.60582 .29555 m
.59406 .29555 L
s
.59406 .29555 m
.58818 .28537 L
s
.58818 .28537 m
.5823 .29555 L
s
.5823 .29555 m
.57055 .29555 L
s
.57055 .29555 m
.56467 .28537 L
s
.56467 .28537 m
.57055 .27518 L
s
.57055 .27518 m
.55879 .27518 L
s
.55879 .27518 m
.55291 .265 L
s
.55291 .265 m
.54703 .27518 L
s
.54703 .27518 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .28537 L
s
.54115 .28537 m
.53527 .29555 L
s
.53527 .29555 m
.52352 .29555 L
s
.52352 .29555 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .29555 L
s
.51176 .29555 m
.5 .29555 L
s
.5 .29555 m
.50588 .30573 L
s
.50588 .30573 m
.5 .31591 L
s
.5 .31591 m
.51176 .31591 L
s
.51176 .31591 m
.51764 .3261 L
s
.51764 .3261 m
.51176 .33628 L
s
.51176 .33628 m
.5 .33628 L
s
.5 .33628 m
.50588 .34646 L
s
.50588 .34646 m
.5 .35664 L
s
.5 .35664 m
.51176 .35664 L
s
.51176 .35664 m
.51764 .36683 L
s
.51764 .36683 m
.52352 .35664 L
s
.52352 .35664 m
.53527 .35664 L
s
.53527 .35664 m
.54115 .36683 L
s
.54115 .36683 m
.53527 .37701 L
s
.53527 .37701 m
.54703 .37701 L
s
.54703 .37701 m
.55291 .38719 L
s
.55291 .38719 m
.54703 .39737 L
s
.54703 .39737 m
.53527 .39737 L
s
.53527 .39737 m
.54115 .40756 L
s
.54115 .40756 m
.53527 .41774 L
s
.53527 .41774 m
.52352 .41774 L
s
.52352 .41774 m
.51764 .40756 L
s
.51764 .40756 m
.51176 .41774 L
s
.51176 .41774 m
.5 .41774 L
s
.5 .41774 m
.50588 .42792 L
s
.50588 .42792 m
.5 .4381 L
s
.5 .4381 m
.51176 .4381 L
s
.51176 .4381 m
.51764 .44829 L
s
.51764 .44829 m
.51176 .45847 L
s
.51176 .45847 m
.5 .45847 L
s
.5 .45847 m
.50588 .46865 L
s
.50588 .46865 m
.5 .47883 L
s
.5 .47883 m
.51176 .47883 L
s
.51176 .47883 m
.51764 .48902 L
s
.51764 .48902 m
.52352 .47883 L
s
.52352 .47883 m
.53527 .47883 L
s
.53527 .47883 m
.54115 .48902 L
s
.54115 .48902 m
.53527 .4992 L
s
.53527 .4992 m
.54703 .4992 L
s
.54703 .4992 m
.55291 .50938 L
s
.55291 .50938 m
.55879 .4992 L
s
.55879 .4992 m
.57055 .4992 L
s
.57055 .4992 m
.56467 .48902 L
s
.56467 .48902 m
.57055 .47883 L
s
.57055 .47883 m
.5823 .47883 L
s
.5823 .47883 m
.58818 .48902 L
s
.58818 .48902 m
.59406 .47883 L
s
.59406 .47883 m
.60582 .47883 L
s
.60582 .47883 m
.6117 .48902 L
s
.6117 .48902 m
.60582 .4992 L
s
.60582 .4992 m
.61758 .4992 L
s
.61758 .4992 m
.62346 .50938 L
s
.62346 .50938 m
.61758 .51956 L
s
.61758 .51956 m
.60582 .51956 L
s
.60582 .51956 m
.6117 .52975 L
s
.6117 .52975 m
.60582 .53993 L
s
.60582 .53993 m
.61758 .53993 L
s
.61758 .53993 m
.62346 .55011 L
s
.62346 .55011 m
.62934 .53993 L
s
.62934 .53993 m
.64109 .53993 L
s
.64109 .53993 m
.64697 .55011 L
s
.64697 .55011 m
.64109 .56029 L
s
.64109 .56029 m
.65285 .56029 L
s
.65285 .56029 m
.65873 .57048 L
s
.65873 .57048 m
.65285 .58066 L
s
.65285 .58066 m
.64109 .58066 L
s
.64109 .58066 m
.64697 .59084 L
s
.64697 .59084 m
.64109 .60102 L
s
.64109 .60102 m
.62934 .60102 L
s
.62934 .60102 m
.62346 .59084 L
s
.62346 .59084 m
.61758 .60102 L
s
.61758 .60102 m
.60582 .60102 L
s
.60582 .60102 m
.6117 .61121 L
s
.6117 .61121 m
.60582 .62139 L
s
.60582 .62139 m
.61758 .62139 L
s
.61758 .62139 m
.62346 .63157 L
s
.62346 .63157 m
.61758 .64175 L
s
.61758 .64175 m
.60582 .64175 L
s
.60582 .64175 m
.6117 .65194 L
s
.6117 .65194 m
.60582 .66212 L
s
.60582 .66212 m
.59406 .66212 L
s
.59406 .66212 m
.58818 .65194 L
s
.58818 .65194 m
.5823 .66212 L
s
.5823 .66212 m
.57055 .66212 L
s
.57055 .66212 m
.56467 .65194 L
s
.56467 .65194 m
.57055 .64175 L
s
.57055 .64175 m
.55879 .64175 L
s
.55879 .64175 m
.55291 .63157 L
s
.55291 .63157 m
.54703 .64175 L
s
.54703 .64175 m
.53527 .64175 L
s
.53527 .64175 m
.54115 .65194 L
s
.54115 .65194 m
.53527 .66212 L
s
.53527 .66212 m
.52352 .66212 L
s
.52352 .66212 m
.51764 .65194 L
s
.51764 .65194 m
.51176 .66212 L
s
.51176 .66212 m
.5 .66212 L
s
.5 .66212 m
.50588 .6723 L
s
.50588 .6723 m
.5 .68249 L
s
.5 .68249 m
.51176 .68249 L
s
.51176 .68249 m
.51764 .69267 L
s
.51764 .69267 m
.51176 .70285 L
s
.51176 .70285 m
.5 .70285 L
s
.5 .70285 m
.50588 .71303 L
s
.50588 .71303 m
.5 .72322 L
s
.5 .72322 m
.51176 .72322 L
s
.51176 .72322 m
.51764 .7334 L
s
.51764 .7334 m
.52352 .72322 L
s
.52352 .72322 m
.53527 .72322 L
s
.53527 .72322 m
.54115 .7334 L
s
.54115 .7334 m
.53527 .74358 L
s
.53527 .74358 m
.54703 .74358 L
s
.54703 .74358 m
.55291 .75376 L
s
.55291 .75376 m
.54703 .76395 L
s
.54703 .76395 m
.53527 .76395 L
s
.53527 .76395 m
.54115 .77413 L
s
.54115 .77413 m
.53527 .78431 L
s
.53527 .78431 m
.52352 .78431 L
s
.52352 .78431 m
.51764 .77413 L
s
.51764 .77413 m
.51176 .78431 L
s
.51176 .78431 m
.5 .78431 L
s
.5 .78431 m
.50588 .79449 L
s
.50588 .79449 m
.5 .80468 L
s
.5 .80468 m
.51176 .80468 L
s
.51176 .80468 m
.51764 .81486 L
s
.51764 .81486 m
.51176 .82504 L
s
.51176 .82504 m
.5 .82504 L
s
.5 .82504 m
.50588 .83522 L
s
.50588 .83522 m
.5 .84541 L
s
.5 .84541 m
.49412 .83522 L
s
.49412 .83522 m
.5 .82504 L
s
.5 .82504 m
.48824 .82504 L
s
.48824 .82504 m
.48236 .81486 L
s
.48236 .81486 m
.48824 .80468 L
s
.48824 .80468 m
.5 .80468 L
s
.5 .80468 m
.49412 .79449 L
s
.49412 .79449 m
.5 .78431 L
s
.5 .78431 m
.48824 .78431 L
s
.48824 .78431 m
.48236 .77413 L
s
.48236 .77413 m
.47648 .78431 L
s
.47648 .78431 m
.46473 .78431 L
s
.46473 .78431 m
.45885 .77413 L
s
.45885 .77413 m
.46473 .76395 L
s
.46473 .76395 m
.45297 .76395 L
s
.45297 .76395 m
.44709 .75376 L
s
.44709 .75376 m
.45297 .74358 L
s
.45297 .74358 m
.46473 .74358 L
s
.46473 .74358 m
.45885 .7334 L
s
.45885 .7334 m
.46473 .72322 L
s
.46473 .72322 m
.47648 .72322 L
s
.47648 .72322 m
.48236 .7334 L
s
.48236 .7334 m
.48824 .72322 L
s
.48824 .72322 m
.5 .72322 L
s
.5 .72322 m
.49412 .71303 L
s
.49412 .71303 m
.5 .70285 L
s
.5 .70285 m
.48824 .70285 L
s
.48824 .70285 m
.48236 .69267 L
s
.48236 .69267 m
.48824 .68249 L
s
.48824 .68249 m
.5 .68249 L
s
.5 .68249 m
.49412 .6723 L
s
.49412 .6723 m
.5 .66212 L
s
.5 .66212 m
.48824 .66212 L
s
.48824 .66212 m
.48236 .65194 L
s
.48236 .65194 m
.47648 .66212 L
s
.47648 .66212 m
.46473 .66212 L
s
.46473 .66212 m
.45885 .65194 L
s
.45885 .65194 m
.46473 .64175 L
s
.46473 .64175 m
.45297 .64175 L
s
.45297 .64175 m
.44709 .63157 L
s
.44709 .63157 m
.44121 .64175 L
s
.44121 .64175 m
.42945 .64175 L
s
.42945 .64175 m
.43533 .65194 L
s
.43533 .65194 m
.42945 .66212 L
s
.42945 .66212 m
.4177 .66212 L
s
.4177 .66212 m
.41182 .65194 L
s
.41182 .65194 m
.40594 .66212 L
s
.40594 .66212 m
.39418 .66212 L
s
.39418 .66212 m
.3883 .65194 L
s
.3883 .65194 m
.39418 .64175 L
s
.39418 .64175 m
.38242 .64175 L
s
.38242 .64175 m
.37654 .63157 L
s
.37654 .63157 m
.38242 .62139 L
s
.38242 .62139 m
.39418 .62139 L
s
.39418 .62139 m
.3883 .61121 L
s
.3883 .61121 m
.39418 .60102 L
s
.39418 .60102 m
.38242 .60102 L
s
.38242 .60102 m
.37654 .59084 L
s
.37654 .59084 m
.37066 .60102 L
s
.37066 .60102 m
.35891 .60102 L
s
.35891 .60102 m
.35303 .59084 L
s
.35303 .59084 m
.35891 .58066 L
s
.35891 .58066 m
.34715 .58066 L
s
.34715 .58066 m
.34127 .57048 L
s
.34127 .57048 m
.34715 .56029 L
s
.34715 .56029 m
.35891 .56029 L
s
.35891 .56029 m
.35303 .55011 L
s
.35303 .55011 m
.35891 .53993 L
s
.35891 .53993 m
.37066 .53993 L
s
.37066 .53993 m
.37654 .55011 L
s
.37654 .55011 m
.38242 .53993 L
s
.38242 .53993 m
.39418 .53993 L
s
.39418 .53993 m
.3883 .52975 L
s
.3883 .52975 m
.39418 .51956 L
s
.39418 .51956 m
.38242 .51956 L
s
.38242 .51956 m
.37654 .50938 L
s
.37654 .50938 m
.38242 .4992 L
s
.38242 .4992 m
.39418 .4992 L
s
.39418 .4992 m
.3883 .48902 L
s
.3883 .48902 m
.39418 .47883 L
s
.39418 .47883 m
.40594 .47883 L
s
.40594 .47883 m
.41182 .48902 L
s
.41182 .48902 m
.4177 .47883 L
s
.4177 .47883 m
.42945 .47883 L
s
.42945 .47883 m
.43533 .48902 L
s
.43533 .48902 m
.42945 .4992 L
s
.42945 .4992 m
.44121 .4992 L
s
.44121 .4992 m
.44709 .50938 L
s
.44709 .50938 m
.45297 .4992 L
s
.45297 .4992 m
.46473 .4992 L
s
.46473 .4992 m
.45885 .48902 L
s
.45885 .48902 m
.46473 .47883 L
s
.46473 .47883 m
.47648 .47883 L
s
.47648 .47883 m
.48236 .48902 L
s
.48236 .48902 m
.48824 .47883 L
s
.48824 .47883 m
.5 .47883 L
s
.5 .47883 m
.49412 .46865 L
s
.49412 .46865 m
.5 .45847 L
s
.5 .45847 m
.48824 .45847 L
s
.48824 .45847 m
.48236 .44829 L
s
.48236 .44829 m
.48824 .4381 L
s
.48824 .4381 m
.5 .4381 L
s
.5 .4381 m
.49412 .42792 L
s
.49412 .42792 m
.5 .41774 L
s
.5 .41774 m
.48824 .41774 L
s
.48824 .41774 m
.48236 .40756 L
s
.48236 .40756 m
.47648 .41774 L
s
.47648 .41774 m
.46473 .41774 L
s
.46473 .41774 m
.45885 .40756 L
s
.45885 .40756 m
.46473 .39737 L
s
.46473 .39737 m
.45297 .39737 L
s
.45297 .39737 m
.44709 .38719 L
s
.44709 .38719 m
.45297 .37701 L
s
.45297 .37701 m
.46473 .37701 L
s
.46473 .37701 m
.45885 .36683 L
s
.45885 .36683 m
.46473 .35664 L
s
.46473 .35664 m
.47648 .35664 L
s
.47648 .35664 m
.48236 .36683 L
s
.48236 .36683 m
.48824 .35664 L
s
.48824 .35664 m
.5 .35664 L
s
.5 .35664 m
.49412 .34646 L
s
.49412 .34646 m
.5 .33628 L
s
.5 .33628 m
.48824 .33628 L
s
.48824 .33628 m
.48236 .3261 L
s
.48236 .3261 m
.48824 .31591 L
s
.48824 .31591 m
.5 .31591 L
s
.5 .31591 m
.49412 .30573 L
s
.49412 .30573 m
.5 .29555 L
s
.5 .29555 m
.48824 .29555 L
s
.48824 .29555 m
.48236 .28537 L
s
.48236 .28537 m
.47648 .29555 L
s
.47648 .29555 m
.46473 .29555 L
s
.46473 .29555 m
.45885 .28537 L
s
.45885 .28537 m
.46473 .27518 L
s
.46473 .27518 m
.45297 .27518 L
s
.45297 .27518 m
.44709 .265 L
s
.44709 .265 m
.44121 .27518 L
s
.44121 .27518 m
.42945 .27518 L
s
.42945 .27518 m
.43533 .28537 L
s
.43533 .28537 m
.42945 .29555 L
s
.42945 .29555 m
.4177 .29555 L
s
.4177 .29555 m
.41182 .28537 L
s
.41182 .28537 m
.40594 .29555 L
s
.40594 .29555 m
.39418 .29555 L
s
.39418 .29555 m
.3883 .28537 L
s
.3883 .28537 m
.39418 .27518 L
s
.39418 .27518 m
.38242 .27518 L
s
.38242 .27518 m
.37654 .265 L
s
.37654 .265 m
.38242 .25482 L
s
.38242 .25482 m
.39418 .25482 L
s
.39418 .25482 m
.3883 .24464 L
s
.3883 .24464 m
.39418 .23445 L
s
.39418 .23445 m
.38242 .23445 L
s
.38242 .23445 m
.37654 .22427 L
s
.37654 .22427 m
.37066 .23445 L
s
.37066 .23445 m
.35891 .23445 L
s
.35891 .23445 m
.35303 .22427 L
s
.35303 .22427 m
.35891 .21409 L
s
.35891 .21409 m
.34715 .21409 L
s
.34715 .21409 m
.34127 .20391 L
s
.34127 .20391 m
.33539 .21409 L
s
.33539 .21409 m
.32363 .21409 L
s
.32363 .21409 m
.32951 .22427 L
s
.32951 .22427 m
.32363 .23445 L
s
.32363 .23445 m
.31188 .23445 L
s
.31188 .23445 m
.306 .22427 L
s
.306 .22427 m
.30012 .23445 L
s
.30012 .23445 m
.28836 .23445 L
s
.28836 .23445 m
.29424 .24464 L
s
.29424 .24464 m
.28836 .25482 L
s
.28836 .25482 m
.30012 .25482 L
s
.30012 .25482 m
.306 .265 L
s
.306 .265 m
.30012 .27518 L
s
.30012 .27518 m
.28836 .27518 L
s
.28836 .27518 m
.29424 .28537 L
s
.29424 .28537 m
.28836 .29555 L
s
.28836 .29555 m
.2766 .29555 L
s
.2766 .29555 m
.27072 .28537 L
s
.27072 .28537 m
.26484 .29555 L
s
.26484 .29555 m
.25309 .29555 L
s
.25309 .29555 m
.24721 .28537 L
s
.24721 .28537 m
.25309 .27518 L
s
.25309 .27518 m
.24133 .27518 L
s
.24133 .27518 m
.23545 .265 L
s
.23545 .265 m
.22957 .27518 L
s
.22957 .27518 m
.21781 .27518 L
s
.21781 .27518 m
.22369 .28537 L
s
.22369 .28537 m
.21781 .29555 L
s
.21781 .29555 m
.20606 .29555 L
s
.20606 .29555 m
.20018 .28537 L
s
.20018 .28537 m
.1943 .29555 L
s
.1943 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.17666 .28537 L
s
.17666 .28537 m
.18254 .27518 L
s
.18254 .27518 m
.17078 .27518 L
s
.17078 .27518 m
.1649 .265 L
s
.1649 .265 m
.17078 .25482 L
s
.17078 .25482 m
.18254 .25482 L
s
.18254 .25482 m
.17666 .24464 L
s
.17666 .24464 m
.18254 .23445 L
s
.18254 .23445 m
.17078 .23445 L
s
.17078 .23445 m
.1649 .22427 L
s
.1649 .22427 m
.15902 .23445 L
s
.15902 .23445 m
.14727 .23445 L
s
.14727 .23445 m
.14139 .22427 L
s
.14139 .22427 m
.14727 .21409 L
s
.14727 .21409 m
.13551 .21409 L
s
.13551 .21409 m
.12963 .20391 L
s
.12963 .20391 m
.13551 .19372 L
s
.13551 .19372 m
.14727 .19372 L
s
.14727 .19372 m
.14139 .18354 L
s
.14139 .18354 m
.14727 .17336 L
s
.14727 .17336 m
.15902 .17336 L
s
.15902 .17336 m
.1649 .18354 L
s
.1649 .18354 m
.17078 .17336 L
s
.17078 .17336 m
.18254 .17336 L
s
.18254 .17336 m
.17666 .16318 L
s
.17666 .16318 m
.18254 .15299 L
s
.18254 .15299 m
.17078 .15299 L
s
.17078 .15299 m
.1649 .14281 L
s
.1649 .14281 m
.17078 .13263 L
s
.17078 .13263 m
.18254 .13263 L
s
.18254 .13263 m
.17666 .12245 L
s
.17666 .12245 m
.18254 .11226 L
s
.18254 .11226 m
.17078 .11226 L
s
.17078 .11226 m
.1649 .10208 L
s
.1649 .10208 m
.15902 .11226 L
s
.15902 .11226 m
.14727 .11226 L
s
.14727 .11226 m
.14139 .10208 L
s
.14139 .10208 m
.14727 .0919 L
s
.14727 .0919 m
.13551 .0919 L
s
.13551 .0919 m
.12963 .08171 L
s
.12963 .08171 m
.12375 .0919 L
s
.12375 .0919 m
.11199 .0919 L
s
.11199 .0919 m
.11787 .10208 L
s
.11787 .10208 m
.11199 .11226 L
s
.11199 .11226 m
.10024 .11226 L
s
.10024 .11226 m
.09436 .10208 L
s
.09436 .10208 m
.08848 .11226 L
s
.08848 .11226 m
.07672 .11226 L
s
.07672 .11226 m
.07084 .10208 L
s
.07084 .10208 m
.07672 .0919 L
s
.07672 .0919 m
.06496 .0919 L
s
.06496 .0919 m
.05908 .08171 L
s
.05908 .08171 m
.06496 .07153 L
s
.06496 .07153 m
.07672 .07153 L
s
.07672 .07153 m
.07084 .06135 L
s
.07084 .06135 m
.07672 .05117 L
s
.07672 .05117 m
.06496 .05117 L
s
.06496 .05117 m
.05908 .04098 L
s
.05908 .04098 m
.0532 .05117 L
s
.0532 .05117 m
.04145 .05117 L
s
.04145 .05117 m
.03557 .04098 L
s
.03557 .04098 m
.04145 .0308 L
s
.04145 .0308 m
.02969 .0308 L
s
.02969 .0308 m
.02381 .02062 L
s
.02381 .02062 m
.03557 .02062 L
s
.03557 .02062 m
.04145 .0308 L
s
.04145 .0308 m
.04733 .02062 L
s
.04733 .02062 m
.05908 .02062 L
s
.05908 .02062 m
.06496 .0308 L
s
.06496 .0308 m
.05908 .04098 L
s
.05908 .04098 m
.07084 .04098 L
s
.07084 .04098 m
.07672 .05117 L
s
.07672 .05117 m
.0826 .04098 L
s
.0826 .04098 m
.09436 .04098 L
s
.09436 .04098 m
.08848 .0308 L
s
.08848 .0308 m
.09436 .02062 L
s
.09436 .02062 m
.10611 .02062 L
s
.10611 .02062 m
.11199 .0308 L
s
.11199 .0308 m
.11787 .02062 L
s
.11787 .02062 m
.12963 .02062 L
s
.12963 .02062 m
.13551 .0308 L
s
.13551 .0308 m
.12963 .04098 L
s
.12963 .04098 m
.14139 .04098 L
s
.14139 .04098 m
.14727 .05117 L
s
.14727 .05117 m
.14139 .06135 L
s
.14139 .06135 m
.12963 .06135 L
s
.12963 .06135 m
.13551 .07153 L
s
.13551 .07153 m
.12963 .08171 L
s
.12963 .08171 m
.14139 .08171 L
s
.14139 .08171 m
.14727 .0919 L
s
.14727 .0919 m
.15315 .08171 L
s
.15315 .08171 m
.1649 .08171 L
s
.1649 .08171 m
.17078 .0919 L
s
.17078 .0919 m
.1649 .10208 L
s
.1649 .10208 m
.17666 .10208 L
s
.17666 .10208 m
.18254 .11226 L
s
.18254 .11226 m
.18842 .10208 L
s
.18842 .10208 m
.20018 .10208 L
s
.20018 .10208 m
.1943 .0919 L
s
.1943 .0919 m
.20018 .08171 L
s
.20018 .08171 m
.21193 .08171 L
s
.21193 .08171 m
.21781 .0919 L
s
.21781 .0919 m
.22369 .08171 L
s
.22369 .08171 m
.23545 .08171 L
s
.23545 .08171 m
.22957 .07153 L
s
.22957 .07153 m
.23545 .06135 L
s
.23545 .06135 m
.22369 .06135 L
s
.22369 .06135 m
.21781 .05117 L
s
.21781 .05117 m
.22369 .04098 L
s
.22369 .04098 m
.23545 .04098 L
s
.23545 .04098 m
.22957 .0308 L
s
.22957 .0308 m
.23545 .02062 L
s
.23545 .02062 m
.24721 .02062 L
s
.24721 .02062 m
.25309 .0308 L
s
.25309 .0308 m
.25897 .02062 L
s
.25897 .02062 m
.27072 .02062 L
s
.27072 .02062 m
.2766 .0308 L
s
.2766 .0308 m
.27072 .04098 L
s
.27072 .04098 m
.28248 .04098 L
s
.28248 .04098 m
.28836 .05117 L
s
.28836 .05117 m
.29424 .04098 L
s
.29424 .04098 m
.306 .04098 L
s
.306 .04098 m
.30012 .0308 L
s
.30012 .0308 m
.306 .02062 L
s
.306 .02062 m
.31775 .02062 L
s
.31775 .02062 m
.32363 .0308 L
s
.32363 .0308 m
.32951 .02062 L
s
.32951 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.34715 .0308 L
s
.34715 .0308 m
.34127 .04098 L
s
.34127 .04098 m
.35303 .04098 L
s
.35303 .04098 m
.35891 .05117 L
s
.35891 .05117 m
.35303 .06135 L
s
.35303 .06135 m
.34127 .06135 L
s
.34127 .06135 m
.34715 .07153 L
s
.34715 .07153 m
.34127 .08171 L
s
.34127 .08171 m
.35303 .08171 L
s
.35303 .08171 m
.35891 .0919 L
s
.35891 .0919 m
.36479 .08171 L
s
.36479 .08171 m
.37654 .08171 L
s
.37654 .08171 m
.38242 .0919 L
s
.38242 .0919 m
.37654 .10208 L
s
.37654 .10208 m
.3883 .10208 L
s
.3883 .10208 m
.39418 .11226 L
s
.39418 .11226 m
.3883 .12245 L
s
.3883 .12245 m
.37654 .12245 L
s
.37654 .12245 m
.38242 .13263 L
s
.38242 .13263 m
.37654 .14281 L
s
.37654 .14281 m
.36479 .14281 L
s
.36479 .14281 m
.35891 .13263 L
s
.35891 .13263 m
.35303 .14281 L
s
.35303 .14281 m
.34127 .14281 L
s
.34127 .14281 m
.34715 .15299 L
s
.34715 .15299 m
.34127 .16318 L
s
.34127 .16318 m
.35303 .16318 L
s
.35303 .16318 m
.35891 .17336 L
s
.35891 .17336 m
.35303 .18354 L
s
.35303 .18354 m
.34127 .18354 L
s
.34127 .18354 m
.34715 .19372 L
s
.34715 .19372 m
.34127 .20391 L
s
.34127 .20391 m
.35303 .20391 L
s
.35303 .20391 m
.35891 .21409 L
s
.35891 .21409 m
.36479 .20391 L
s
.36479 .20391 m
.37654 .20391 L
s
.37654 .20391 m
.38242 .21409 L
s
.38242 .21409 m
.37654 .22427 L
s
.37654 .22427 m
.3883 .22427 L
s
.3883 .22427 m
.39418 .23445 L
s
.39418 .23445 m
.40006 .22427 L
s
.40006 .22427 m
.41182 .22427 L
s
.41182 .22427 m
.40594 .21409 L
s
.40594 .21409 m
.41182 .20391 L
s
.41182 .20391 m
.42357 .20391 L
s
.42357 .20391 m
.42945 .21409 L
s
.42945 .21409 m
.43533 .20391 L
s
.43533 .20391 m
.44709 .20391 L
s
.44709 .20391 m
.45297 .21409 L
s
.45297 .21409 m
.44709 .22427 L
s
.44709 .22427 m
.45885 .22427 L
s
.45885 .22427 m
.46473 .23445 L
s
.46473 .23445 m
.45885 .24464 L
s
.45885 .24464 m
.44709 .24464 L
s
.44709 .24464 m
.45297 .25482 L
s
.45297 .25482 m
.44709 .265 L
s
.44709 .265 m
.45885 .265 L
s
.45885 .265 m
.46473 .27518 L
s
.46473 .27518 m
.47061 .265 L
s
.47061 .265 m
.48236 .265 L
s
.48236 .265 m
.48824 .27518 L
s
.48824 .27518 m
.48236 .28537 L
s
.48236 .28537 m
.49412 .28537 L
s
.49412 .28537 m
.5 .29555 L
s
.5 .29555 m
.50588 .28537 L
s
.50588 .28537 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .27518 L
s
.51176 .27518 m
.51764 .265 L
s
.51764 .265 m
.52939 .265 L
s
.52939 .265 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .265 L
s
.54115 .265 m
.55291 .265 L
s
.55291 .265 m
.54703 .25482 L
s
.54703 .25482 m
.55291 .24464 L
s
.55291 .24464 m
.54115 .24464 L
s
.54115 .24464 m
.53527 .23445 L
s
.53527 .23445 m
.54115 .22427 L
s
.54115 .22427 m
.55291 .22427 L
s
.55291 .22427 m
.54703 .21409 L
s
.54703 .21409 m
.55291 .20391 L
s
.55291 .20391 m
.56467 .20391 L
s
.56467 .20391 m
.57055 .21409 L
s
.57055 .21409 m
.57643 .20391 L
s
.57643 .20391 m
.58818 .20391 L
s
.58818 .20391 m
.59406 .21409 L
s
.59406 .21409 m
.58818 .22427 L
s
.58818 .22427 m
.59994 .22427 L
s
.59994 .22427 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .22427 L
s
.6117 .22427 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .21409 L
s
.61758 .21409 m
.62346 .20391 L
s
.62346 .20391 m
.63521 .20391 L
s
.63521 .20391 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .20391 L
s
.64697 .20391 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .19372 L
s
.65285 .19372 m
.65873 .18354 L
s
.65873 .18354 m
.64697 .18354 L
s
.64697 .18354 m
.64109 .17336 L
s
.64109 .17336 m
.64697 .16318 L
s
.64697 .16318 m
.65873 .16318 L
s
.65873 .16318 m
.65285 .15299 L
s
.65285 .15299 m
.65873 .14281 L
s
.65873 .14281 m
.64697 .14281 L
s
.64697 .14281 m
.64109 .13263 L
s
.64109 .13263 m
.63521 .14281 L
s
.63521 .14281 m
.62346 .14281 L
s
.62346 .14281 m
.61758 .13263 L
s
.61758 .13263 m
.62346 .12245 L
s
.62346 .12245 m
.6117 .12245 L
s
.6117 .12245 m
.60582 .11226 L
s
.60582 .11226 m
.6117 .10208 L
s
.6117 .10208 m
.62346 .10208 L
s
.62346 .10208 m
.61758 .0919 L
s
.61758 .0919 m
.62346 .08171 L
s
.62346 .08171 m
.63521 .08171 L
s
.63521 .08171 m
.64109 .0919 L
s
.64109 .0919 m
.64697 .08171 L
s
.64697 .08171 m
.65873 .08171 L
s
.65873 .08171 m
.65285 .07153 L
s
.65285 .07153 m
.65873 .06135 L
s
.65873 .06135 m
.64697 .06135 L
s
.64697 .06135 m
.64109 .05117 L
s
.64109 .05117 m
.64697 .04098 L
s
.64697 .04098 m
.65873 .04098 L
s
.65873 .04098 m
.65285 .0308 L
s
.65285 .0308 m
.65873 .02062 L
s
.65873 .02062 m
.67049 .02062 L
s
.67049 .02062 m
.67637 .0308 L
s
.67637 .0308 m
.68225 .02062 L
s
.68225 .02062 m
.694 .02062 L
s
.694 .02062 m
.69988 .0308 L
s
.69988 .0308 m
.694 .04098 L
s
.694 .04098 m
.70576 .04098 L
s
.70576 .04098 m
.71164 .05117 L
s
.71164 .05117 m
.71752 .04098 L
s
.71752 .04098 m
.72928 .04098 L
s
.72928 .04098 m
.7234 .0308 L
s
.7234 .0308 m
.72928 .02062 L
s
.72928 .02062 m
.74103 .02062 L
s
.74103 .02062 m
.74691 .0308 L
s
.74691 .0308 m
.75279 .02062 L
s
.75279 .02062 m
.76455 .02062 L
s
.76455 .02062 m
.77043 .0308 L
s
.77043 .0308 m
.76455 .04098 L
s
.76455 .04098 m
.77631 .04098 L
s
.77631 .04098 m
.78219 .05117 L
s
.78219 .05117 m
.77631 .06135 L
s
.77631 .06135 m
.76455 .06135 L
s
.76455 .06135 m
.77043 .07153 L
s
.77043 .07153 m
.76455 .08171 L
s
.76455 .08171 m
.77631 .08171 L
s
.77631 .08171 m
.78219 .0919 L
s
.78219 .0919 m
.78807 .08171 L
s
.78807 .08171 m
.79982 .08171 L
s
.79982 .08171 m
.8057 .0919 L
s
.8057 .0919 m
.79982 .10208 L
s
.79982 .10208 m
.81158 .10208 L
s
.81158 .10208 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .10208 L
s
.82334 .10208 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .0919 L
s
.82922 .0919 m
.8351 .08171 L
s
.8351 .08171 m
.84685 .08171 L
s
.84685 .08171 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .08171 L
s
.85861 .08171 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .07153 L
s
.86449 .07153 m
.87037 .06135 L
s
.87037 .06135 m
.85861 .06135 L
s
.85861 .06135 m
.85273 .05117 L
s
.85273 .05117 m
.85861 .04098 L
s
.85861 .04098 m
.87037 .04098 L
s
.87037 .04098 m
.86449 .0308 L
s
.86449 .0308 m
.87037 .02062 L
s
.87037 .02062 m
.88213 .02062 L
s
.88213 .02062 m
.88801 .0308 L
s
.88801 .0308 m
.89389 .02062 L
s
.89389 .02062 m
.90564 .02062 L
s
.90564 .02062 m
.91152 .0308 L
s
.91152 .0308 m
.90564 .04098 L
s
.90564 .04098 m
.9174 .04098 L
s
.9174 .04098 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .04098 L
s
.92916 .04098 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .0308 L
s
.93504 .0308 m
.94092 .02062 L
s
.94092 .02062 m
.95267 .02062 L
s
.95267 .02062 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .02062 L
s
.96443 .02062 m
.97619 .02062 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 249.375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001MR000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`002P3oool6000000@0oooo1P00
000>0?ooo`H00000103oool6000002d0oooo1P0000040?ooo`H000003P3oool6000000D0oooo1P00
00270?ooo`H000001@3oool6000000h0oooo1P0000040?ooo`H00000;@3oool6000000@0oooo1P00
000>0?ooo`H00000103oool6000000/0oooo000:0?ooo`030000003oool0oooo0080oooo00D00000
0?ooo`3oool0oooo000000060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3o
ool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0:P3oool00`000000oooo
0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo
08D0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3o
ool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo
000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool01@000000oooo
0?ooo`3oool0000000@0oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool0
0P3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool0103oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo02T0oooo00<000000?oo
o`3oool0103oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3oool0P`3o
ool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000103oool0
0`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`020?ooo`030000003oool00000
00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`050?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`030000003oool00000
00@0oooo00<000000?ooo`3oool02P3oool000`0oooo00<000000?ooo`3oool00P3oool2000000P0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`800000203oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo00H0oooo0P0000080?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01P3oool00`000000oooo000000080?ooo`030000003oool0oooo0840
oooo00<000000?ooo`3oool01P3oool00`000000oooo000000080?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01P3oool2000000P0oooo00<000000?ooo`3oool09`3oool00`000000
oooo0?ooo`060?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0
oooo0P0000040?ooo`030000003oool0oooo00/0oooo000<0?ooo`H000002P3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`210?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00P0oooo1P00000=0?ooo`00403oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0W0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0@0?ooo`00403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo08<0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0?0?ooo`003`3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Y0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo08@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0?0?oo
o`003`3oool00`000000oooo0?ooo`070?ooo`H00000103oool6000000h0oooo1`00000R0?ooo`H0
00003P3oool6000000@0oooo1P00000?0?ooo`H00000O@3oool6000000l0oooo1P0000040?ooo`H0
00003P3oool600000280oooo1`00000>0?ooo`H00000103oool6000000T0oooo00<000000?ooo`3o
ool03P3oool000l0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000040?ooo`040000
003oool0oooo000001X0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0H0?ooo`04
0000003oool0oooo000001X0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0H0?oo
o`050000003oool0oooo0?ooo`0000006P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo01P0oooo00D000000?ooo`3oool0oooo000000040?ooo`800000203oool00`000000oooo0?oo
o`0?0?ooo`00403oool00`000000oooo0?ooo`050?ooo`030000003oool0000000@0oooo00@00000
0?ooo`3oool000006`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01T0oooo00@0
00000?ooo`3oool000006`3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`0000006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`000000103oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo00l0
oooo000@0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000040?ooo`80
0000703oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01X0oooo0P00000L0?ooo`03
0000003oool0oooo07P0oooo00<000000?ooo`3oool06P3oool00`000000oooo0000000L0?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool06P3oool00`000000oooo000000040?ooo`04
0000003oool0oooo000000H0oooo00<000000?ooo`3oool0403oool00140oooo1`0000040?ooo`H0
00007@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01/0oooo1P0000040?ooo`L0
00004P3oool00200oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo03T0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0i0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0O0?ooo`00803o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0>03o
ool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01l0oooo000O0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo07T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool07P3oool001l0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0N`3oool00`00
0000oooo0?ooo`0f0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0N0?ooo`007P3oool00`000000oooo0?ooo`0F0?ooo`L000008P3oool6000002d0
oooo1P00001m0?ooo`H00000;@3oool600000280oooo1`00000H0?ooo`030000003oool0oooo01d0
oooo000O0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`0/0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool07P3oool001l0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02d0oooo00<0
00000?ooo`3oool0P`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02P0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0O0?ooo`00803oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`22
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01l0oooo000L0?ooo`H00000603oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`3oool05P3oool6000001d0oooo000K0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo0840oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0W0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool06P3oool001/0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0
P`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`0J0?ooo`006P3oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`230?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo01T0oooo000J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`0[0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06@3oool0
01T0oooo00<000000?ooo`3oool06`3oool7000000@0oooo1P00000>0?ooo`H00000103oool60000
02d0oooo1P0000040?ooo`H00000J@3oool6000000@0oooo1P00000]0?ooo`H00000103oool60000
00h0oooo1P0000040?ooo`L000007@3oool00`000000oooo0?ooo`0H0?ooo`006P3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0000000@0oooo00D000000?ooo`3oool0oooo000000050?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0
000003P0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo06H0oooo00<0
00000?ooo`3oool0103oool010000000oooo0?ooo`00000i0?ooo`040000003oool0oooo000000H0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo0000
00D0oooo0P00000L0?ooo`030000003oool0oooo01T0oooo000J0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`0000001@3oool00`000000oooo000000070?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000j0?ooo`030000003oool0000000L0
oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo0000
03T0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00@0oooo00@000000?ooo`3oool00000103oool010000000oooo0?ooo`00000J0?ooo`030000
003oool0oooo01X0oooo000K0?ooo`030000003oool0oooo01L0oooo00D000000?ooo`3oool0oooo
000000040?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0000003X0oooo00<000000?ooo`0000001`3oool00`000000oooo
0?ooo`1T0?ooo`030000003oool0oooo00H0oooo0P00000k0?ooo`800000203oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000050?ooo`040000003oool0oooo000001X0
oooo00<000000?ooo`3oool06P3oool001`0oooo1P00000>0?ooo`H00000103oool6000000T0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03X0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool6000000@0oooo1P00000>0?ooo`H000007@3oool00200oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo06@0oooo00<0
00000?ooo`3oool0C@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01l0oooo000O
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1U
0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0O
0?ooo`007`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04X0oooo00<000000?oo
o`3oool0IP3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool07P3oool001h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool6000000@0oooo1P00001;0?ooo`H0
0000G`3oool6000004/0oooo1P0000040?ooo`H000002@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0M0?ooo`00
7`3oool00`000000oooo0?ooo`060?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00H0oooo0P0000040?ooo`050000003oool0oooo0?ooo`000000EP3oool00`000000
oooo0?ooo`1M0?ooo`030000003oool0oooo05@0oooo00D000000?ooo`3oool0oooo000000040?oo
o`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000080?oo
o`030000003oool0oooo01h0oooo000O0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`00
00001`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`00
0000103oool010000000oooo0?ooo`00001G0?ooo`030000003oool0oooo05/0oooo00<000000?oo
o`3oool0E@3oool010000000oooo0?ooo`0000040?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool07P3oool00200oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`000006
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00
00040?ooo`030000003oool0000005P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?oo
o`1F0?ooo`030000003oool0000000@0oooo00@000000?ooo`3oool000001P3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0O0?ooo`00803oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`00
00001@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3o
ool0oooo000000040?ooo`800000F@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo
05L0oooo0P0000040?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003o
ool0oooo01l0oooo000Q0?ooo`H00000103oool6000000h0oooo1P0000040?ooo`H00000FP3oool0
0`000000oooo0?ooo`1I0?ooo`030000003oool0oooo05P0oooo1P0000040?ooo`H000003P3oool6
000000@0oooo1P00000R0?ooo`00C@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo
05X0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1=0?ooo`00C@3oool00`000000
oooo0?ooo`1H0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0F03oool00`000000
oooo0?ooo`1<0?ooo`00C03oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo05`0oooo
00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo0?oo
o`1C0?ooo`H00000G`3oool6000005D0oooo00<000000?ooo`3oool0B`3oool004`0oooo00<00000
0?ooo`3oool0D`3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0B`3oool004d0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1U0?oo
o`030000003oool0oooo05<0oooo00<000000?ooo`3oool0C03oool004d0oooo00<000000?ooo`3o
ool0D`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3o
ool0C03oool004h0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo05<0oooo00<000000?ooo`3oool0C@3oool004T0oooo1P00001;0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo04T0oooo1P00001:0?ooo`00B03oool00`000000oooo0?ooo`1>0?ooo`030000003oool0
000000L0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`060?ooo`800000D03oool0
0`000000oooo0?ooo`180?ooo`00B03oool00`000000oooo0?ooo`1>0?ooo`030000003oool00000
00L0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo
00000500oooo00<000000?ooo`3oool0A`3oool004L0oooo00<000000?ooo`3oool0CP3oool01@00
0000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo00000500oooo00<000000?ooo`3oool0A`3oool004L0oooo00<00000
0?ooo`3oool0B@3oool6000000@0oooo1P00001Y0?ooo`H00000103oool6000004/0oooo00<00000
0?ooo`3oool0AP3oool004L0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`240?oo
o`030000003oool0oooo04T0oooo00<000000?ooo`3oool0A`3oool004P0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0A`3oool004P0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`220?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool0B03oool004T0oooo1P0000160?ooo`030000003o
ool0oooo0840oooo00<000000?ooo`3oool0A03oool6000004X0oooo001>0?ooo`030000003oool0
oooo04@0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`140?ooo`030000003oool0
oooo04d0oooo001=0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0P`3oool00`00
0000oooo0?ooo`140?ooo`030000003oool0oooo04`0oooo001=0?ooo`030000003oool0oooo04@0
oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo04`0
oooo001<0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo04/0oooo001<0?ooo`030000003oool0oooo04@0oooo1P00
001m0?ooo`H00000AP3oool00`000000oooo0?ooo`1;0?ooo`00C03oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`1<
0?ooo`00C@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo07T0oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`1<0?ooo`00C@3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo07P0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`1=0?ooo`00
?`3oool6000000@0oooo1P00001;0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0
B@3oool6000000@0oooo1P0000100?ooo`00?`3oool00`000000oooo0?ooo`030?ooo`050000003o
ool0oooo0?ooo`000000D03oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo04h0oooo
00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo03h0oooo000n0?ooo`030000
003oool0oooo00@0oooo00@000000?ooo`3oool00000D03oool00`000000oooo0?ooo`1i0?ooo`03
0000003oool0oooo04h0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0m0?oo
o`00?P3oool00`000000oooo0?ooo`050?ooo`030000003oool000000500oooo00<000000?ooo`3o
ool0N@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0000000L0oooo00<000000?ooo`3o
ool0?@3oool003d0oooo00<000000?ooo`3oool01P3oool200000500oooo00<000000?ooo`3oool0
N`3oool00`000000oooo0?ooo`1>0?ooo`800000203oool00`000000oooo0?ooo`0l0?ooo`00?@3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04T0oooo1P00001m0?ooo`H00000B`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03`0oooo000m0?ooo`030000003oool0
oooo05<0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`1C0?ooo`030000003oool0
oooo03d0oooo000n0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0P`3oool00`00
0000oooo0?ooo`1C0?ooo`030000003oool0oooo03d0oooo000n0?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03h0
oooo000j0?ooo`H00000E@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo05<0oooo
1P00000k0?ooo`00>@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0880oooo00<0
00000?ooo`3oool0F@3oool00`000000oooo0?ooo`0h0?ooo`00>03oool00`000000oooo0?ooo`1I
0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0h
0?ooo`00>03oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo08@0oooo00<000000?oo
o`3oool0F@3oool00`000000oooo0?ooo`0g0?ooo`00=`3oool00`000000oooo0?ooo`1I0?ooo`H0
0000103oool6000000h0oooo1P0000040?ooo`H00000;@3oool6000000@0oooo1P00000>0?ooo`H0
0000103oool6000005/0oooo00<000000?ooo`3oool0=P3oool003P0oooo00<000000?ooo`3oool0
E`3oool2000000@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001P3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000
00040?ooo`800000F@3oool00`000000oooo0?ooo`0g0?ooo`00>03oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0000000@0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000oooo
0?ooo`0Y0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001P3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000103oool00`00
0000oooo0000001I0?ooo`030000003oool0oooo03L0oooo000i0?ooo`030000003oool0oooo05D0
oooo00@000000?ooo`3oool00000103oool00`000000oooo000000070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000070?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool01@3oool00`000000oooo000000080?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool01@3oool00`000000oooo000000040?ooo`040000003oool0oooo
000005L0oooo00<000000?ooo`3oool0>03oool003T0oooo00<000000?ooo`3oool0E@3oool01@00
0000oooo0?ooo`3oool0000000@0oooo0P0000080?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool01P3oool2000000P0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`06
0?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P000004
0?ooo`050000003oool0oooo0?ooo`000000E`3oool00`000000oooo0?ooo`0h0?ooo`00>P3oool6
000004/0oooo1P0000040?ooo`H000002@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00L0oooo1P0000040?ooo`H00000B`3oool6000003/0oooo000n0?ooo`030000003oool0oooo
04/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0n0?ooo`00?P3oool00`000000
oooo0?ooo`1;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:03oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0?@3oool003d0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02T0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo03d0oooo
000m0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool6000000@0oooo1P00
000>0?ooo`H000008`3oool6000000h0oooo1P0000040?ooo`H000002@3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo03`0oooo000m0?ooo`030000003oool0oooo00H0oooo0P00000j
0?ooo`030000003oool0000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`06
0?ooo`800000103oool01@000000oooo0?ooo`3oool0000001X0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0H0?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`000000
1`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo0P00000j0?ooo`030000
003oool0000000L0oooo00<000000?ooo`3oool0?@3oool003h0oooo00<000000?ooo`3oool0103o
ool010000000oooo0?ooo`00000i0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`00
00006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01T0oooo00@000000?ooo`3o
ool00000103oool00`000000oooo000000070?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool01@3oool010000000oooo0?ooo`00000i0?ooo`030000003oool0000000L0oooo00<00000
0?ooo`3oool0?@3oool003h0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00000h
0?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00@0oooo00@000000?ooo`3oool00000103oool00`000000oooo0000000L0?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool06P3oool2000000@0oooo00D000000?ooo`3oool0
oooo000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool010000000
oooo0?ooo`00000h0?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0n
0?ooo`00?`3oool6000000@0oooo1P00000]0?ooo`H00000103oool6000000h0oooo1P0000040?oo
o`H000007P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01`0oooo1P0000040?oo
o`H000003P3oool6000000@0oooo1P00000]0?ooo`H00000103oool600000400oooo001>0?ooo`03
0000003oool0oooo02/0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo01d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0Z0?ooo`03
0000003oool0oooo02/0oooo00<000000?ooo`3oool0C@3oool004d0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`1<0?ooo`00C@3oool00`000000oooo0?ooo`0]0?ooo`030000003o
ool0oooo02P0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0]0?ooo`030000003o
ool0oooo04`0oooo001<0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0B`3oool0
04`0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
01H0oooo1P00000S0?ooo`H00000603oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
02l0oooo00<000000?ooo`3oool0B`3oool004`0oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`0F0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`1<0?ooo`00C@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02P0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo04`0oooo
001=0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0C@3oool004T0oooo1P00000]
0?ooo`H000008P3oool6000001T0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0G
0?ooo`H000008P3oool6000002d0oooo1P00001:0?ooo`00B@3oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo0200oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo04P0oooo00180?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3o
ool07P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3o
ool0A`3oool004P0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0K0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`170?ooo`00A`3o
ool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo04H0oooo00170?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0
oooo01/0oooo1P0000040?ooo`H000003`3oool6000000@0oooo1P00000M0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo04H0oooo00170?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000006`3oool00`00
0000oooo0?ooo`0M0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`000000103oool01000
0000oooo0?ooo`0000060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool0
1@000000oooo0?ooo`3oool0000000@0oooo0P00000L0?ooo`030000003oool0oooo01d0oooo00<0
00000?ooo`3oool06P3oool2000001`0oooo00<000000?ooo`3oool0A`3oool004P0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo0000000K0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool06@3oool00`000000oooo000000040?ooo`040000003oool0oooo000000H0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000@0oooo00@0
00000?ooo`3oool000006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01P0oooo
00@000000?ooo`3oool000006`3oool00`000000oooo0?ooo`170?ooo`00B03oool00`000000oooo
0?ooo`0H0?ooo`050000003oool0oooo0?ooo`0000006@3oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo01P0oooo00D000000?ooo`3oool0oooo000000040?ooo`800000203oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000103oool01000
0000oooo0?ooo`00000J0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0603oool0
10000000oooo0?ooo`00000J0?ooo`030000003oool0oooo04P0oooo00190?ooo`H000003P3oool7
000000@0oooo1P00000>0?ooo`H000008P3oool6000000l0oooo1P0000040?ooo`H000002@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo1P0000040?ooo`H000003`3oool6
00000280oooo1P00000>0?ooo`H00000103oool7000000h0oooo1P00001:0?ooo`00C@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0Y0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`1=0?ooo`00C@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`0W0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`1<0?ooo`00C03oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`H000001@3oool6000000T0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0B`3oool004`0oooo00<000000?ooo`3oool01P3oool2000000P0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`800000203oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`000000203oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00H0oooo0P0000040?ooo`050000003oool0oooo0?ooo`000000103o
ool2000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`030000003o
ool0000000P0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`060?ooo`800000203o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo0P0000080?ooo`030000003o
ool0oooo04/0oooo001=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000103oool0
1@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`0X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`0000001`3oool00`000000oooo0?oo
o`1<0?ooo`00C@3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`00
00001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3o
ool00000103oool00`000000oooo000000040?ooo`040000003oool0oooo000000H0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001P3o
ool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool00000
1P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool0
00001P3oool00`000000oooo0?ooo`1<0?ooo`00CP3oool00`000000oooo0?ooo`030?ooo`050000
003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0
oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo02X0oooo00<000000?oo
o`3oool00`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0000
00040?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo02X0
oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000001@3oool0
0`000000oooo0?ooo`1=0?ooo`00CP3oool6000000@0oooo1P00000?0?ooo`H00000103oool60000
02`0oooo1P0000050?ooo`H000003P3oool6000000@0oooo2`0000040?ooo`H000003P3oool60000
00D0oooo1P00000/0?ooo`H00000103oool6000000l0oooo1P0000040?ooo`H00000C`3oool00=H0
oooo00<000000?ooo`000000e`3oool00=D0oooo00@000000?ooo`3oool00000e`3oool00=D0oooo
00D000000?ooo`3oool0oooo0000003F0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0=<0oooo003E0?ooo`050000003oool0oooo0?ooo`000000eP3oool00=D0oooo00D0
00000?ooo`3oool0oooo0000003F0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool0
0`000000oooo0000003G0?ooo`00dP3oool;00000=<0oooo003A0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0d03oool00=00oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`3@0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo003?
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0cP3oool00=00oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`3?0?ooo`00d03oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo0=00oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
d03oool00=80oooo2`00003C0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`00
0000oooo0000003G0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`05
0000003oool0oooo0?ooo`000000eP3oool00=@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`3C0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`040000
003oool0oooo00000=L0oooo003F0?ooo`030000003oool000000=L0oooo00380?ooo`H00000103o
ool;000000@0oooo1P0000390?ooo`00a`3oool00`000000oooo0?ooo`030?ooo`050000003oool0
oooo0?ooo`0000002`3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0<H0oooo
00360?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000303oool010000000oooo0?oo
o`0000060?ooo`030000003oool0oooo0<H0oooo00360?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`0000003@3oool2000000P0oooo00<000000?ooo`3oool0a@3oool00<D0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0a03oool00<H0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`350?oo
o`00aP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00370?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool0aP3oool00<L0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`360?ooo`00``3oool6000001d0oooo1P0000340?ooo`00`P3oool00`00
0000oooo0?ooo`0W0?ooo`030000003oool0oooo0<40oooo00310?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool0`@3oool00<40oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`300?ooo`00`03oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0;l0oooo0031
0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0`03oool00<40oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`300?ooo`00`P3oool00`000000oooo0?ooo`0W0?ooo`03
0000003oool0oooo0<40oooo00320?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
`@3oool00<<0oooo1P00000M0?ooo`H00000a03oool00<L0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`360?ooo`00aP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
0<H0oooo00360?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0a@3oool00<D0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0a03oool00<H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000
000=0?ooo`800000203oool00`000000oooo0?ooo`350?ooo`00aP3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0000000`0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?oo
o`360?ooo`00a`3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000002`3o
ool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0<H0oooo00380?ooo`H00000103o
ool;000000@0oooo1P0000390?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`00
0000oooo0000003G0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`05
0000003oool0oooo0?ooo`000000eP3oool00=@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`3C0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003E0?ooo`040000
003oool0oooo00000=L0oooo003F0?ooo`030000003oool000000=L0oooo003B0?ooo`/00000d`3o
ool00=40oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00d@3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo0=00oooo003@0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0c`3oool00=00oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`3?0?ooo`00c`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0<h0oooo003@
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0c`3oool00=00oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`3@0?ooo`00d@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0=00oooo003B0?ooo`/00000d`3oool00=H0oooo00<000000?ooo`000000e`3o
ool00=D0oooo00@000000?ooo`3oool00000e`3oool00=D0oooo00D000000?ooo`3oool0oooo0000
003F0?ooo`00e03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003E0?oo
o`050000003oool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F
0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00
Z@3oool6000000D0oooo1P00000>0?ooo`H00000103oool;000000@0oooo1P00000>0?ooo`H00000
1@3oool600000:X0oooo002X0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo
000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo
0?ooo`3oool0000000/0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0
oooo0:P0oooo002X0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000001`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000303oool0
10000000oooo0?ooo`0000060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103o
ool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo0:L0oooo002W0?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`0000001`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`0000003@3oool2000000P0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`060?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0Y`3o
ool00:L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0YP3oool00:L0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo0:L0oooo002X0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`2W0?ooo`00Z03oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool07@3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0Z03oool00:@0oooo1P00
000?0?ooo`H00000103oool6000001d0oooo1P0000040?ooo`H000003`3oool600000:D0oooo002T
0?ooo`030000003oool0oooo01L0oooo00D000000?ooo`3oool0oooo0000000Y0?ooo`050000003o
ool0oooo0?ooo`0000006@3oool00`000000oooo0?ooo`2S0?ooo`00X`3oool00`000000oooo0?oo
o`0I0?ooo`040000003oool0oooo000002T0oooo00@000000?ooo`3oool000006`3oool00`000000
oooo0?ooo`2R0?ooo`00X`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0000002/0oooo
00<000000?ooo`0000006`3oool00`000000oooo0?ooo`2R0?ooo`00XP3oool00`000000oooo0?oo
o`0K0?ooo`800000:`3oool2000001d0oooo00<000000?ooo`3oool0X@3oool00:80oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool0X@3oool00:80oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`2R0?oo
o`00X`3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo0:80oooo002S0?ooo`030000
003oool0oooo06@0oooo00<000000?ooo`3oool0X`3oool00:@0oooo1P00001K0?ooo`H00000Y@3o
ool00:T0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`2X0?ooo`00Z03oool00`00
0000oooo0?ooo`1K0?ooo`030000003oool0oooo0:L0oooo002X0?ooo`030000003oool0oooo05/0
oooo00<000000?ooo`3oool0Y`3oool00:L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`2V0?ooo`00Y`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo0:H0oooo002W
0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0Y`3oool00:P0oooo00<000000?oo
o`3oool0F`3oool00`000000oooo0?ooo`2W0?ooo`00Z03oool00`000000oooo0?ooo`1J0?ooo`03
0000003oool0oooo0:P0oooo002J0?ooo`H00000103oool6000005/0oooo1P0000040?ooo`H00000
V`3oool009T0oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`00001V0?ooo`050000
003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`2I0?ooo`00V@3oool00`000000oooo
0?ooo`040?ooo`040000003oool0oooo000006L0oooo00<000000?ooo`0000001`3oool00`000000
oooo0?ooo`2H0?ooo`00V03oool00`000000oooo0?ooo`060?ooo`800000J03oool00`000000oooo
000000070?ooo`030000003oool0oooo09P0oooo002H0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0I`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo09L0oooo002H
0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0U`3oool009T0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`2H0?ooo`00V@3oool00`000000oooo0?ooo`1i0?ooo`03
0000003oool0oooo09P0oooo002J0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0
V@3oool009D0oooo1P00001i0?ooo`H00000UP3oool009@0oooo00<000000?ooo`3oool0PP3oool0
0`000000oooo0?ooo`2D0?ooo`00U03oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo
09<0oooo002C0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0T`3oool009<0oooo
00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`2B0?ooo`00T`3oool00`000000oooo0?oo
o`250?ooo`030000003oool0oooo0980oooo002D0?ooo`030000003oool0oooo08<0oooo00<00000
0?ooo`3oool0T`3oool009@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`2C0?oo
o`00U@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo09@0oooo002E0?ooo`H00000
N@3oool6000009H0oooo002I0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0V@3o
ool009T0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`2H0?ooo`00V03oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo09P0oooo002H0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo09L0
oooo002H0?ooo`030000003oool0oooo00H0oooo0P00001X0?ooo`030000003oool0000000L0oooo
00<000000?ooo`3oool0V03oool009T0oooo00<000000?ooo`3oool0103oool010000000oooo0?oo
o`00001W0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0V03oool009T0oooo00<0
00000?ooo`3oool0103oool010000000oooo0?ooo`00001V0?ooo`050000003oool0oooo0?ooo`00
00001@3oool00`000000oooo0?ooo`2I0?ooo`00VP3oool6000000@0oooo1P00001K0?ooo`H00000
103oool6000009/0oooo002Y0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0Z03o
ool00:P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`2W0?ooo`00Z03oool00`00
0000oooo0?ooo`1K0?ooo`030000003oool0oooo0:L0oooo002W0?ooo`030000003oool0oooo05d0
oooo00<000000?ooo`3oool0YP3oool00:L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo
0?ooo`2V0?ooo`00Y`3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0:L0oooo002X
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0Y`3oool00:P0oooo00<000000?oo
o`3oool0FP3oool00`000000oooo0?ooo`2X0?ooo`00Y03oool6000005/0oooo1P00002U0?ooo`00
X`3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0:<0oooo002S0?ooo`030000003o
ool0oooo06D0oooo00<000000?ooo`3oool0XP3oool00:80oooo00<000000?ooo`3oool0IP3oool0
0`000000oooo0?ooo`2R0?ooo`00XP3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2Q0?ooo`00XP3oool00`000000
oooo0?ooo`0K0?ooo`800000:`3oool2000001d0oooo00<000000?ooo`3oool0X@3oool00:<0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0000000[0?ooo`030000003oool0000001/0oooo
00<000000?ooo`3oool0XP3oool00:<0oooo00<000000?ooo`3oool06@3oool010000000oooo0?oo
o`00000Y0?ooo`040000003oool0oooo000001/0oooo00<000000?ooo`3oool0XP3oool00:@0oooo
00<000000?ooo`3oool05`3oool01@000000oooo0?ooo`3oool0000002T0oooo00D000000?ooo`3o
ool0oooo0000000I0?ooo`030000003oool0oooo0:<0oooo002T0?ooo`H000003`3oool6000000@0
oooo1P00000M0?ooo`H00000103oool6000000l0oooo1P00002U0?ooo`00Z03oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool07@3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0Z03oool00:P0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01h0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0:L0oooo002W
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0O
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`2W
0?ooo`00Y`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`2V0?ooo`00Y`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`80
00003@3oool2000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`03
0000003oool0000000P0oooo00<000000?ooo`3oool0YP3oool00:P0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo000000070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0000000=0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`050?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
Y`3oool00:P0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000H0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000/0
oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00@0oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo0:L0oooo002Y0?oo
o`030000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000000/0oooo00D0
00000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
0`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0Z03oool00:T0oooo
1P0000050?ooo`H000003P3oool6000000@0oooo2`0000040?ooo`H000003P3oool6000000D0oooo
1P00002Z0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00e@3oool010000000oooo0?ooo`00
003G0?ooo`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003D0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F
0?ooo`00e@3oool010000000oooo0?ooo`00003G0?ooo`00eP3oool00`000000oooo0000003G0?oo
o`00dP3oool;00000=<0oooo003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
d03oool00=40oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00d03oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo003@0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0c`3oool00<l0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`3>0?ooo`00d03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0<l0oooo
003@0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0d03oool00=40oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`3@0?ooo`00dP3oool;00000=<0oooo003F0?ooo`03
0000003oool000000=L0oooo003F0?ooo`030000003oool000000=L0oooo003E0?ooo`050000003o
ool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?ooo`00e03o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003E0?ooo`050000003oool0
oooo0?ooo`000000eP3oool00=D0oooo00@000000?ooo`3oool00000e`3oool00=H0oooo00<00000
0?ooo`000000e`3oool00<P0oooo1P0000040?ooo`/00000103oool600000<T0oooo00370?ooo`03
0000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`040000003oool0oooo
000000H0oooo00<000000?ooo`3oool0aP3oool00<H0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0000000<0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool0aP3oool0
0<H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000=0?ooo`800000203oool00`00
0000oooo0?ooo`350?ooo`00a@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`340?ooo`00aP3oool00`000000oooo
0?ooo`0O0?ooo`030000003oool0oooo0<D0oooo00360?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool0a@3oool00<L0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`36
0?ooo`00a`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0<H0oooo00330?ooo`H0
00007@3oool600000<@0oooo00320?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
`@3oool00<40oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`310?ooo`00`@3oool0
0`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo00300?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool0_`3oool00<40oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`300?ooo`00`@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0<00oooo
00320?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0`@3oool00<80oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`310?ooo`00``3oool6000001d0oooo1P0000340?oo
o`00a`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0<H0oooo00360?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0aP3oool00<H0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`350?ooo`00a@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`340?ooo`00aP3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0000000d0oooo0P0000080?ooo`030000003oool0oooo
0<D0oooo00360?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000303oool010000000
oooo0?ooo`0000060?ooo`030000003oool0oooo0<H0oooo00370?ooo`030000003oool0oooo00<0
oooo00D000000?ooo`3oool0oooo0000000;0?ooo`040000003oool0oooo000000H0oooo00<00000
0?ooo`3oool0aP3oool00<P0oooo1P0000040?ooo`/00000103oool600000<T0oooo003F0?ooo`03
0000003oool000000=L0oooo003F0?ooo`030000003oool000000=L0oooo003E0?ooo`050000003o
ool0oooo0?ooo`000000eP3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?ooo`00e03o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=<0oooo003E0?ooo`050000003oool0
oooo0?ooo`000000eP3oool00=D0oooo00@000000?ooo`3oool00000e`3oool00=H0oooo00<00000
0?ooo`000000e`3oool00=80oooo2`00003C0?ooo`00d@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0=00oooo003@0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
d03oool00=00oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3?0?ooo`00c`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0<h0oooo003@0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0c`3oool00=00oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`3?0?ooo`00d@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0=00oooo
003A0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0d03oool00=80oooo2`00003C
0?ooo`00eP3oool00`000000oooo0000003G0?ooo`00e@3oool010000000oooo0?ooo`00003G0?oo
o`00e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003D0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0d`3oool00=D0oooo00D000000?ooo`3oool0oooo0000003F0?ooo`00
e@3oool01@000000oooo0?ooo`3oool000000=H0oooo003F0?ooo`030000003oool000000=L0oooo
003F0?ooo`030000003oool000000=L0oooo003G0?ooo`030000003oool0oooo0=H0oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {373.063, 0}} -> {-0.0252817, -0.0216552, \
0.00365625, 0.00365625}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Typical Koch curves", "Section"],

Cell[CellGroupData[{

Cell["Simple figures based upon equilateral and square", "Subsection"],

Cell["\<\
The following specify an equilateral triangle and a square in a \
counterclockwise direction of traversal.\
\>", "Text"],

Cell[BoxData[{
    \(\(triangle = {{0. , 0. }, {0.5, \@3\/2. }, {1. , 0. }, {0. , 
            0. }};\)\), "\n", 
    \(\(square = {{0. , 0. }, {1. , 0. }, {1. , 1. }, {0. , 1. }, {0. , 
            0. }};\)\)}], "Input"],

Cell[TextData[{
  "The effect of outward versus inward triangulation are compared below for a \
",
  Cell[BoxData[
      \(TraditionalForm\`4\^th\)]],
  " order Koch curve based upon the equilateral triangle.  The first pair of \
figures shows the intermediate iterations while the second pair shows the \
final result."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
        GraphicsArray[{DisplayKoch[triangle, 4, ShowIterations \[Rule] True, 
              DisplayFunction \[Rule] Identity], 
            DisplayKoch[triangle, 4, ShowIterations \[Rule] True, 
              KochType \[Rule] "\<sieve\>", 
              DisplayFunction \[Rule] Identity]}], 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .54986 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0130918 0.47619 [
[ 0 0 0 0 ]
[ 1 .54986 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .54986 L
0 .54986 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0130918 0.477324 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.302422 0.952381 [
[ 0 0 0 0 ]
[ 1 1.1547 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.1547 L
0 1.1547 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .30242 m
.5 1.12721 L
s
.5 1.12721 m
.97619 .30242 L
s
.97619 .30242 m
.02381 .30242 L
s
.02381 .30242 m
.18254 .57735 L
s
.18254 .57735 m
.02381 .85228 L
s
.02381 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.5 1.12721 L
s
.5 1.12721 m
.65873 .85228 L
s
.65873 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.81746 .57735 L
s
.81746 .57735 m
.97619 .30242 L
s
.97619 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.5 .02749 L
s
.5 .02749 m
.34127 .30242 L
s
.34127 .30242 m
.02381 .30242 L
s
.02381 .30242 m
.07672 .39406 L
s
.07672 .39406 m
.02381 .48571 L
s
.02381 .48571 m
.12963 .48571 L
s
.12963 .48571 m
.18254 .57735 L
s
.18254 .57735 m
.12963 .66899 L
s
.12963 .66899 m
.02381 .66899 L
s
.02381 .66899 m
.07672 .76064 L
s
.07672 .76064 m
.02381 .85228 L
s
.02381 .85228 m
.12963 .85228 L
s
.12963 .85228 m
.18254 .94392 L
s
.18254 .94392 m
.23545 .85228 L
s
.23545 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.39418 .94392 L
s
.39418 .94392 m
.34127 1.03556 L
s
.34127 1.03556 m
.44709 1.03556 L
s
.44709 1.03556 m
.5 1.12721 L
s
.5 1.12721 m
.55291 1.03556 L
s
.55291 1.03556 m
.65873 1.03556 L
s
.65873 1.03556 m
.60582 .94392 L
s
.60582 .94392 m
.65873 .85228 L
s
.65873 .85228 m
.76455 .85228 L
s
.76455 .85228 m
.81746 .94392 L
s
.81746 .94392 m
.87037 .85228 L
s
.87037 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.92328 .76064 L
s
.92328 .76064 m
.97619 .66899 L
s
.97619 .66899 m
.87037 .66899 L
s
.87037 .66899 m
.81746 .57735 L
s
.81746 .57735 m
.87037 .48571 L
s
.87037 .48571 m
.97619 .48571 L
s
.97619 .48571 m
.92328 .39406 L
s
.92328 .39406 m
.97619 .30242 L
s
.97619 .30242 m
.87037 .30242 L
s
.87037 .30242 m
.81746 .21078 L
s
.81746 .21078 m
.76455 .30242 L
s
.76455 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.60582 .21078 L
s
.60582 .21078 m
.65873 .11914 L
s
.65873 .11914 m
.55291 .11914 L
s
.55291 .11914 m
.5 .02749 L
s
.5 .02749 m
.44709 .11914 L
s
.44709 .11914 m
.34127 .11914 L
s
.34127 .11914 m
.39418 .21078 L
s
.39418 .21078 m
.34127 .30242 L
s
.34127 .30242 m
.23545 .30242 L
s
.23545 .30242 m
.18254 .21078 L
s
.18254 .21078 m
.12963 .30242 L
s
.12963 .30242 m
.02381 .30242 L
s
.02381 .30242 m
.04145 .33297 L
s
.04145 .33297 m
.02381 .36352 L
s
.02381 .36352 m
.05908 .36352 L
s
.05908 .36352 m
.07672 .39406 L
s
.07672 .39406 m
.05908 .42461 L
s
.05908 .42461 m
.02381 .42461 L
s
.02381 .42461 m
.04145 .45516 L
s
.04145 .45516 m
.02381 .48571 L
s
.02381 .48571 m
.05908 .48571 L
s
.05908 .48571 m
.07672 .51626 L
s
.07672 .51626 m
.09436 .48571 L
s
.09436 .48571 m
.12963 .48571 L
s
.12963 .48571 m
.14727 .51626 L
s
.14727 .51626 m
.12963 .5468 L
s
.12963 .5468 m
.1649 .5468 L
s
.1649 .5468 m
.18254 .57735 L
s
.18254 .57735 m
.1649 .6079 L
s
.1649 .6079 m
.12963 .6079 L
s
.12963 .6079 m
.14727 .63845 L
s
.14727 .63845 m
.12963 .66899 L
s
.12963 .66899 m
.09436 .66899 L
s
.09436 .66899 m
.07672 .63845 L
s
.07672 .63845 m
.05908 .66899 L
s
.05908 .66899 m
.02381 .66899 L
s
.02381 .66899 m
.04145 .69954 L
s
.04145 .69954 m
.02381 .73009 L
s
.02381 .73009 m
.05908 .73009 L
s
.05908 .73009 m
.07672 .76064 L
s
.07672 .76064 m
.05908 .79118 L
s
.05908 .79118 m
.02381 .79118 L
s
.02381 .79118 m
.04145 .82173 L
s
.04145 .82173 m
.02381 .85228 L
s
.02381 .85228 m
.05908 .85228 L
s
.05908 .85228 m
.07672 .88283 L
s
.07672 .88283 m
.09436 .85228 L
s
.09436 .85228 m
.12963 .85228 L
s
.12963 .85228 m
.14727 .88283 L
s
.14727 .88283 m
.12963 .91337 L
s
.12963 .91337 m
.1649 .91337 L
s
.1649 .91337 m
.18254 .94392 L
s
.18254 .94392 m
.20018 .91337 L
s
.20018 .91337 m
.23545 .91337 L
s
.23545 .91337 m
.21781 .88283 L
s
.21781 .88283 m
.23545 .85228 L
s
.23545 .85228 m
.27072 .85228 L
s
.27072 .85228 m
.28836 .88283 L
s
.28836 .88283 m
.306 .85228 L
s
.306 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.35891 .88283 L
s
.35891 .88283 m
.34127 .91337 L
s
.34127 .91337 m
.37654 .91337 L
s
.37654 .91337 m
.39418 .94392 L
s
.39418 .94392 m
.37654 .97447 L
s
.37654 .97447 m
.34127 .97447 L
s
.34127 .97447 m
.35891 1.00502 L
s
.35891 1.00502 m
.34127 1.03556 L
s
.34127 1.03556 m
.37654 1.03556 L
s
.37654 1.03556 m
.39418 1.06611 L
s
.39418 1.06611 m
.41182 1.03556 L
s
.41182 1.03556 m
.44709 1.03556 L
s
.44709 1.03556 m
.46473 1.06611 L
s
.46473 1.06611 m
.44709 1.09666 L
s
.44709 1.09666 m
.48236 1.09666 L
s
.48236 1.09666 m
.5 1.12721 L
s
.5 1.12721 m
.51764 1.09666 L
s
.51764 1.09666 m
.55291 1.09666 L
s
.55291 1.09666 m
.53527 1.06611 L
s
.53527 1.06611 m
.55291 1.03556 L
s
.55291 1.03556 m
.58818 1.03556 L
s
.58818 1.03556 m
.60582 1.06611 L
s
.60582 1.06611 m
.62346 1.03556 L
s
.62346 1.03556 m
.65873 1.03556 L
s
.65873 1.03556 m
.64109 1.00502 L
s
.64109 1.00502 m
.65873 .97447 L
s
.65873 .97447 m
.62346 .97447 L
s
.62346 .97447 m
.60582 .94392 L
s
.60582 .94392 m
.62346 .91337 L
s
.62346 .91337 m
.65873 .91337 L
s
.65873 .91337 m
.64109 .88283 L
s
.64109 .88283 m
.65873 .85228 L
s
.65873 .85228 m
.694 .85228 L
s
.694 .85228 m
.71164 .88283 L
s
.71164 .88283 m
.72928 .85228 L
s
.72928 .85228 m
.76455 .85228 L
s
.76455 .85228 m
.78219 .88283 L
s
.78219 .88283 m
.76455 .91337 L
s
.76455 .91337 m
.79982 .91337 L
s
.79982 .91337 m
.81746 .94392 L
s
.81746 .94392 m
.8351 .91337 L
s
.8351 .91337 m
.87037 .91337 L
s
.87037 .91337 m
.85273 .88283 L
s
.85273 .88283 m
.87037 .85228 L
s
.87037 .85228 m
.90564 .85228 L
s
.90564 .85228 m
.92328 .88283 L
s
.92328 .88283 m
.94092 .85228 L
s
.94092 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.95855 .82173 L
s
.95855 .82173 m
.97619 .79118 L
s
.97619 .79118 m
.94092 .79118 L
s
.94092 .79118 m
.92328 .76064 L
s
.92328 .76064 m
.94092 .73009 L
s
.94092 .73009 m
.97619 .73009 L
s
.97619 .73009 m
.95855 .69954 L
s
.95855 .69954 m
.97619 .66899 L
s
.97619 .66899 m
.94092 .66899 L
s
.94092 .66899 m
.92328 .63845 L
s
.92328 .63845 m
.90564 .66899 L
s
.90564 .66899 m
.87037 .66899 L
s
.87037 .66899 m
.85273 .63845 L
s
.85273 .63845 m
.87037 .6079 L
s
.87037 .6079 m
.8351 .6079 L
s
.8351 .6079 m
.81746 .57735 L
s
.81746 .57735 m
.8351 .5468 L
s
.8351 .5468 m
.87037 .5468 L
s
.87037 .5468 m
.85273 .51626 L
s
.85273 .51626 m
.87037 .48571 L
s
.87037 .48571 m
.90564 .48571 L
s
.90564 .48571 m
.92328 .51626 L
s
.92328 .51626 m
.94092 .48571 L
s
.94092 .48571 m
.97619 .48571 L
s
.97619 .48571 m
.95855 .45516 L
s
.95855 .45516 m
.97619 .42461 L
s
.97619 .42461 m
.94092 .42461 L
s
.94092 .42461 m
.92328 .39406 L
s
.92328 .39406 m
.94092 .36352 L
s
.94092 .36352 m
.97619 .36352 L
s
.97619 .36352 m
.95855 .33297 L
s
.95855 .33297 m
.97619 .30242 L
s
.97619 .30242 m
.94092 .30242 L
s
.94092 .30242 m
.92328 .27187 L
s
.92328 .27187 m
.90564 .30242 L
s
.90564 .30242 m
.87037 .30242 L
s
.87037 .30242 m
.85273 .27187 L
s
.85273 .27187 m
.87037 .24133 L
s
.87037 .24133 m
.8351 .24133 L
s
.8351 .24133 m
.81746 .21078 L
s
.81746 .21078 m
.79982 .24133 L
s
.79982 .24133 m
.76455 .24133 L
s
.76455 .24133 m
.78219 .27187 L
s
.78219 .27187 m
.76455 .30242 L
s
.76455 .30242 m
.72928 .30242 L
s
.72928 .30242 m
.71164 .27187 L
s
.71164 .27187 m
.694 .30242 L
s
.694 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.64109 .27187 L
s
.64109 .27187 m
.65873 .24133 L
s
.65873 .24133 m
.62346 .24133 L
s
.62346 .24133 m
.60582 .21078 L
s
.60582 .21078 m
.62346 .18023 L
s
.62346 .18023 m
.65873 .18023 L
s
.65873 .18023 m
.64109 .14968 L
s
.64109 .14968 m
.65873 .11914 L
s
.65873 .11914 m
.62346 .11914 L
s
.62346 .11914 m
.60582 .08859 L
s
.60582 .08859 m
.58818 .11914 L
s
.58818 .11914 m
.55291 .11914 L
s
.55291 .11914 m
.53527 .08859 L
s
.53527 .08859 m
.55291 .05804 L
s
.55291 .05804 m
.51764 .05804 L
s
.51764 .05804 m
.5 .02749 L
s
.5 .02749 m
.48236 .05804 L
s
.48236 .05804 m
.44709 .05804 L
s
.44709 .05804 m
.46473 .08859 L
s
.46473 .08859 m
.44709 .11914 L
s
.44709 .11914 m
.41182 .11914 L
s
.41182 .11914 m
.39418 .08859 L
s
.39418 .08859 m
.37654 .11914 L
s
.37654 .11914 m
.34127 .11914 L
s
.34127 .11914 m
.35891 .14968 L
s
.35891 .14968 m
.34127 .18023 L
s
.34127 .18023 m
.37654 .18023 L
s
.37654 .18023 m
.39418 .21078 L
s
.39418 .21078 m
.37654 .24133 L
s
.37654 .24133 m
.34127 .24133 L
s
.34127 .24133 m
.35891 .27187 L
s
.35891 .27187 m
.34127 .30242 L
s
.34127 .30242 m
.306 .30242 L
s
.306 .30242 m
.28836 .27187 L
s
.28836 .27187 m
.27072 .30242 L
s
.27072 .30242 m
.23545 .30242 L
s
.23545 .30242 m
.21781 .27187 L
s
.21781 .27187 m
.23545 .24133 L
s
.23545 .24133 m
.20018 .24133 L
s
.20018 .24133 m
.18254 .21078 L
s
.18254 .21078 m
.1649 .24133 L
s
.1649 .24133 m
.12963 .24133 L
s
.12963 .24133 m
.14727 .27187 L
s
.14727 .27187 m
.12963 .30242 L
s
.12963 .30242 m
.09436 .30242 L
s
.09436 .30242 m
.07672 .27187 L
s
.07672 .27187 m
.05908 .30242 L
s
.05908 .30242 m
.02381 .30242 L
s
.02381 .30242 m
.02969 .3126 L
s
.02969 .3126 m
.02381 .32279 L
s
.02381 .32279 m
.03557 .32279 L
s
.03557 .32279 m
.04145 .33297 L
s
.04145 .33297 m
.03557 .34315 L
s
.03557 .34315 m
.02381 .34315 L
s
.02381 .34315 m
.02969 .35333 L
s
.02969 .35333 m
.02381 .36352 L
s
.02381 .36352 m
.03557 .36352 L
s
.03557 .36352 m
.04145 .3737 L
s
.04145 .3737 m
.04733 .36352 L
s
.04733 .36352 m
.05908 .36352 L
s
.05908 .36352 m
.06496 .3737 L
s
.06496 .3737 m
.05908 .38388 L
s
.05908 .38388 m
.07084 .38388 L
s
.07084 .38388 m
.07672 .39406 L
s
.07672 .39406 m
.07084 .40425 L
s
.07084 .40425 m
.05908 .40425 L
s
.05908 .40425 m
.06496 .41443 L
s
.06496 .41443 m
.05908 .42461 L
s
.05908 .42461 m
.04733 .42461 L
s
.04733 .42461 m
.04145 .41443 L
s
.04145 .41443 m
.03557 .42461 L
s
.03557 .42461 m
.02381 .42461 L
s
.02381 .42461 m
.02969 .43479 L
s
.02969 .43479 m
.02381 .44498 L
s
.02381 .44498 m
.03557 .44498 L
s
.03557 .44498 m
.04145 .45516 L
s
.04145 .45516 m
.03557 .46534 L
s
.03557 .46534 m
.02381 .46534 L
s
.02381 .46534 m
.02969 .47552 L
s
.02969 .47552 m
.02381 .48571 L
s
.02381 .48571 m
.03557 .48571 L
s
.03557 .48571 m
.04145 .49589 L
s
.04145 .49589 m
.04733 .48571 L
s
.04733 .48571 m
.05908 .48571 L
s
.05908 .48571 m
.06496 .49589 L
s
.06496 .49589 m
.05908 .50607 L
s
.05908 .50607 m
.07084 .50607 L
s
.07084 .50607 m
.07672 .51626 L
s
.07672 .51626 m
.0826 .50607 L
s
.0826 .50607 m
.09436 .50607 L
s
.09436 .50607 m
.08848 .49589 L
s
.08848 .49589 m
.09436 .48571 L
s
.09436 .48571 m
.10611 .48571 L
s
.10611 .48571 m
.11199 .49589 L
s
.11199 .49589 m
.11787 .48571 L
s
.11787 .48571 m
.12963 .48571 L
s
.12963 .48571 m
.13551 .49589 L
s
.13551 .49589 m
.12963 .50607 L
s
.12963 .50607 m
.14139 .50607 L
s
.14139 .50607 m
.14727 .51626 L
s
.14727 .51626 m
.14139 .52644 L
s
.14139 .52644 m
.12963 .52644 L
s
.12963 .52644 m
.13551 .53662 L
s
.13551 .53662 m
.12963 .5468 L
s
.12963 .5468 m
.14139 .5468 L
s
.14139 .5468 m
.14727 .55699 L
s
.14727 .55699 m
.15315 .5468 L
s
.15315 .5468 m
.1649 .5468 L
s
.1649 .5468 m
.17078 .55699 L
s
.17078 .55699 m
.1649 .56717 L
s
.1649 .56717 m
.17666 .56717 L
s
.17666 .56717 m
.18254 .57735 L
s
.18254 .57735 m
.17666 .58753 L
s
.17666 .58753 m
.1649 .58753 L
s
.1649 .58753 m
.17078 .59772 L
s
.17078 .59772 m
.1649 .6079 L
s
.1649 .6079 m
.15315 .6079 L
s
.15315 .6079 m
.14727 .59772 L
s
.14727 .59772 m
.14139 .6079 L
s
.14139 .6079 m
.12963 .6079 L
s
.12963 .6079 m
.13551 .61808 L
s
.13551 .61808 m
.12963 .62826 L
s
.12963 .62826 m
.14139 .62826 L
s
.14139 .62826 m
.14727 .63845 L
s
.14727 .63845 m
.14139 .64863 L
s
.14139 .64863 m
.12963 .64863 L
s
.12963 .64863 m
.13551 .65881 L
s
.13551 .65881 m
.12963 .66899 L
s
.12963 .66899 m
.11787 .66899 L
s
.11787 .66899 m
.11199 .65881 L
s
.11199 .65881 m
.10611 .66899 L
s
.10611 .66899 m
.09436 .66899 L
s
.09436 .66899 m
.08848 .65881 L
s
.08848 .65881 m
.09436 .64863 L
s
.09436 .64863 m
.0826 .64863 L
s
.0826 .64863 m
.07672 .63845 L
s
.07672 .63845 m
.07084 .64863 L
s
.07084 .64863 m
.05908 .64863 L
s
.05908 .64863 m
.06496 .65881 L
s
.06496 .65881 m
.05908 .66899 L
s
.05908 .66899 m
.04733 .66899 L
s
.04733 .66899 m
.04145 .65881 L
s
.04145 .65881 m
.03557 .66899 L
s
.03557 .66899 m
.02381 .66899 L
s
.02381 .66899 m
.02969 .67918 L
s
.02969 .67918 m
.02381 .68936 L
s
.02381 .68936 m
.03557 .68936 L
s
.03557 .68936 m
.04145 .69954 L
s
.04145 .69954 m
.03557 .70972 L
s
.03557 .70972 m
.02381 .70972 L
s
.02381 .70972 m
.02969 .71991 L
s
.02969 .71991 m
.02381 .73009 L
s
.02381 .73009 m
.03557 .73009 L
s
.03557 .73009 m
.04145 .74027 L
s
.04145 .74027 m
.04733 .73009 L
s
.04733 .73009 m
.05908 .73009 L
s
.05908 .73009 m
.06496 .74027 L
s
.06496 .74027 m
.05908 .75045 L
s
.05908 .75045 m
.07084 .75045 L
s
.07084 .75045 m
.07672 .76064 L
s
.07672 .76064 m
.07084 .77082 L
s
.07084 .77082 m
.05908 .77082 L
s
.05908 .77082 m
.06496 .781 L
s
.06496 .781 m
.05908 .79118 L
s
.05908 .79118 m
.04733 .79118 L
s
.04733 .79118 m
.04145 .781 L
s
.04145 .781 m
.03557 .79118 L
s
.03557 .79118 m
.02381 .79118 L
s
.02381 .79118 m
.02969 .80137 L
s
.02969 .80137 m
.02381 .81155 L
s
.02381 .81155 m
.03557 .81155 L
s
.03557 .81155 m
.04145 .82173 L
s
.04145 .82173 m
.03557 .83191 L
s
.03557 .83191 m
.02381 .83191 L
s
.02381 .83191 m
.02969 .8421 L
s
.02969 .8421 m
.02381 .85228 L
s
.02381 .85228 m
.03557 .85228 L
s
.03557 .85228 m
.04145 .86246 L
s
.04145 .86246 m
.04733 .85228 L
s
.04733 .85228 m
.05908 .85228 L
s
.05908 .85228 m
.06496 .86246 L
s
.06496 .86246 m
.05908 .87264 L
s
.05908 .87264 m
.07084 .87264 L
s
.07084 .87264 m
.07672 .88283 L
s
.07672 .88283 m
.0826 .87264 L
s
.0826 .87264 m
.09436 .87264 L
s
.09436 .87264 m
.08848 .86246 L
s
.08848 .86246 m
.09436 .85228 L
s
.09436 .85228 m
.10611 .85228 L
s
.10611 .85228 m
.11199 .86246 L
s
.11199 .86246 m
.11787 .85228 L
s
.11787 .85228 m
.12963 .85228 L
s
.12963 .85228 m
.13551 .86246 L
s
.13551 .86246 m
.12963 .87264 L
s
.12963 .87264 m
.14139 .87264 L
s
.14139 .87264 m
.14727 .88283 L
s
.14727 .88283 m
.14139 .89301 L
s
.14139 .89301 m
.12963 .89301 L
s
.12963 .89301 m
.13551 .90319 L
s
.13551 .90319 m
.12963 .91337 L
s
.12963 .91337 m
.14139 .91337 L
s
.14139 .91337 m
.14727 .92356 L
s
.14727 .92356 m
.15315 .91337 L
s
.15315 .91337 m
.1649 .91337 L
s
.1649 .91337 m
.17078 .92356 L
s
.17078 .92356 m
.1649 .93374 L
s
.1649 .93374 m
.17666 .93374 L
s
.17666 .93374 m
.18254 .94392 L
s
.18254 .94392 m
.18842 .93374 L
s
.18842 .93374 m
.20018 .93374 L
s
.20018 .93374 m
.1943 .92356 L
s
.1943 .92356 m
.20018 .91337 L
s
.20018 .91337 m
.21193 .91337 L
s
.21193 .91337 m
.21781 .92356 L
s
.21781 .92356 m
.22369 .91337 L
s
.22369 .91337 m
.23545 .91337 L
s
.23545 .91337 m
.22957 .90319 L
s
.22957 .90319 m
.23545 .89301 L
s
.23545 .89301 m
.22369 .89301 L
s
.22369 .89301 m
.21781 .88283 L
s
.21781 .88283 m
.22369 .87264 L
s
.22369 .87264 m
.23545 .87264 L
s
.23545 .87264 m
.22957 .86246 L
s
.22957 .86246 m
.23545 .85228 L
s
.23545 .85228 m
.24721 .85228 L
s
.24721 .85228 m
.25309 .86246 L
s
.25309 .86246 m
.25897 .85228 L
s
.25897 .85228 m
.27072 .85228 L
s
.27072 .85228 m
.2766 .86246 L
s
.2766 .86246 m
.27072 .87264 L
s
.27072 .87264 m
.28248 .87264 L
s
.28248 .87264 m
.28836 .88283 L
s
.28836 .88283 m
.29424 .87264 L
s
.29424 .87264 m
.306 .87264 L
s
.306 .87264 m
.30012 .86246 L
s
.30012 .86246 m
.306 .85228 L
s
.306 .85228 m
.31775 .85228 L
s
.31775 .85228 m
.32363 .86246 L
s
.32363 .86246 m
.32951 .85228 L
s
.32951 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.34715 .86246 L
s
.34715 .86246 m
.34127 .87264 L
s
.34127 .87264 m
.35303 .87264 L
s
.35303 .87264 m
.35891 .88283 L
s
.35891 .88283 m
.35303 .89301 L
s
.35303 .89301 m
.34127 .89301 L
s
.34127 .89301 m
.34715 .90319 L
s
.34715 .90319 m
.34127 .91337 L
s
.34127 .91337 m
.35303 .91337 L
s
.35303 .91337 m
.35891 .92356 L
s
.35891 .92356 m
.36479 .91337 L
s
.36479 .91337 m
.37654 .91337 L
s
.37654 .91337 m
.38242 .92356 L
s
.38242 .92356 m
.37654 .93374 L
s
.37654 .93374 m
.3883 .93374 L
s
.3883 .93374 m
.39418 .94392 L
s
.39418 .94392 m
.3883 .9541 L
s
.3883 .9541 m
.37654 .9541 L
s
.37654 .9541 m
.38242 .96429 L
s
.38242 .96429 m
.37654 .97447 L
s
.37654 .97447 m
.36479 .97447 L
s
.36479 .97447 m
.35891 .96429 L
s
.35891 .96429 m
.35303 .97447 L
s
.35303 .97447 m
.34127 .97447 L
s
.34127 .97447 m
.34715 .98465 L
s
.34715 .98465 m
.34127 .99483 L
s
.34127 .99483 m
.35303 .99483 L
s
.35303 .99483 m
.35891 1.00502 L
s
.35891 1.00502 m
.35303 1.0152 L
s
.35303 1.0152 m
.34127 1.0152 L
s
.34127 1.0152 m
.34715 1.02538 L
s
.34715 1.02538 m
.34127 1.03556 L
s
.34127 1.03556 m
.35303 1.03556 L
s
.35303 1.03556 m
.35891 1.04575 L
s
.35891 1.04575 m
.36479 1.03556 L
s
.36479 1.03556 m
.37654 1.03556 L
s
.37654 1.03556 m
.38242 1.04575 L
s
.38242 1.04575 m
.37654 1.05593 L
s
.37654 1.05593 m
.3883 1.05593 L
s
.3883 1.05593 m
.39418 1.06611 L
s
.39418 1.06611 m
.40006 1.05593 L
s
.40006 1.05593 m
.41182 1.05593 L
s
.41182 1.05593 m
.40594 1.04575 L
s
.40594 1.04575 m
.41182 1.03556 L
s
.41182 1.03556 m
.42357 1.03556 L
s
.42357 1.03556 m
.42945 1.04575 L
s
.42945 1.04575 m
.43533 1.03556 L
s
.43533 1.03556 m
.44709 1.03556 L
s
.44709 1.03556 m
.45297 1.04575 L
s
.45297 1.04575 m
.44709 1.05593 L
s
.44709 1.05593 m
.45885 1.05593 L
s
.45885 1.05593 m
.46473 1.06611 L
s
.46473 1.06611 m
.45885 1.07629 L
s
.45885 1.07629 m
.44709 1.07629 L
s
.44709 1.07629 m
.45297 1.08648 L
s
.45297 1.08648 m
.44709 1.09666 L
s
.44709 1.09666 m
.45885 1.09666 L
s
.45885 1.09666 m
.46473 1.10684 L
s
.46473 1.10684 m
.47061 1.09666 L
s
.47061 1.09666 m
.48236 1.09666 L
s
.48236 1.09666 m
.48824 1.10684 L
s
.48824 1.10684 m
.48236 1.11703 L
s
.48236 1.11703 m
.49412 1.11703 L
s
.49412 1.11703 m
.5 1.12721 L
s
.5 1.12721 m
.50588 1.11703 L
s
.50588 1.11703 m
.51764 1.11703 L
s
.51764 1.11703 m
.51176 1.10684 L
s
.51176 1.10684 m
.51764 1.09666 L
s
.51764 1.09666 m
.52939 1.09666 L
s
.52939 1.09666 m
.53527 1.10684 L
s
.53527 1.10684 m
.54115 1.09666 L
s
.54115 1.09666 m
.55291 1.09666 L
s
.55291 1.09666 m
.54703 1.08648 L
s
.54703 1.08648 m
.55291 1.07629 L
s
.55291 1.07629 m
.54115 1.07629 L
s
.54115 1.07629 m
.53527 1.06611 L
s
.53527 1.06611 m
.54115 1.05593 L
s
.54115 1.05593 m
.55291 1.05593 L
s
.55291 1.05593 m
.54703 1.04575 L
s
.54703 1.04575 m
.55291 1.03556 L
s
.55291 1.03556 m
.56467 1.03556 L
s
.56467 1.03556 m
.57055 1.04575 L
s
.57055 1.04575 m
.57643 1.03556 L
s
.57643 1.03556 m
.58818 1.03556 L
s
.58818 1.03556 m
.59406 1.04575 L
s
.59406 1.04575 m
.58818 1.05593 L
s
.58818 1.05593 m
.59994 1.05593 L
s
.59994 1.05593 m
.60582 1.06611 L
s
.60582 1.06611 m
.6117 1.05593 L
s
.6117 1.05593 m
.62346 1.05593 L
s
.62346 1.05593 m
.61758 1.04575 L
s
.61758 1.04575 m
.62346 1.03556 L
s
.62346 1.03556 m
.63521 1.03556 L
s
.63521 1.03556 m
.64109 1.04575 L
s
.64109 1.04575 m
.64697 1.03556 L
s
.64697 1.03556 m
.65873 1.03556 L
s
.65873 1.03556 m
.65285 1.02538 L
s
.65285 1.02538 m
.65873 1.0152 L
s
.65873 1.0152 m
.64697 1.0152 L
s
.64697 1.0152 m
.64109 1.00502 L
s
.64109 1.00502 m
.64697 .99483 L
s
.64697 .99483 m
.65873 .99483 L
s
.65873 .99483 m
.65285 .98465 L
s
.65285 .98465 m
.65873 .97447 L
s
.65873 .97447 m
.64697 .97447 L
s
.64697 .97447 m
.64109 .96429 L
s
.64109 .96429 m
.63521 .97447 L
s
.63521 .97447 m
.62346 .97447 L
s
.62346 .97447 m
.61758 .96429 L
s
.61758 .96429 m
.62346 .9541 L
s
.62346 .9541 m
.6117 .9541 L
s
.6117 .9541 m
.60582 .94392 L
s
.60582 .94392 m
.6117 .93374 L
s
.6117 .93374 m
.62346 .93374 L
s
.62346 .93374 m
.61758 .92356 L
s
.61758 .92356 m
.62346 .91337 L
s
.62346 .91337 m
.63521 .91337 L
s
.63521 .91337 m
.64109 .92356 L
s
.64109 .92356 m
.64697 .91337 L
s
.64697 .91337 m
.65873 .91337 L
s
.65873 .91337 m
.65285 .90319 L
s
.65285 .90319 m
.65873 .89301 L
s
.65873 .89301 m
.64697 .89301 L
s
.64697 .89301 m
.64109 .88283 L
s
.64109 .88283 m
.64697 .87264 L
s
.64697 .87264 m
.65873 .87264 L
s
.65873 .87264 m
.65285 .86246 L
s
.65285 .86246 m
.65873 .85228 L
s
.65873 .85228 m
.67049 .85228 L
s
.67049 .85228 m
.67637 .86246 L
s
.67637 .86246 m
.68225 .85228 L
s
.68225 .85228 m
.694 .85228 L
s
.694 .85228 m
.69988 .86246 L
s
.69988 .86246 m
.694 .87264 L
s
.694 .87264 m
.70576 .87264 L
s
.70576 .87264 m
.71164 .88283 L
s
.71164 .88283 m
.71752 .87264 L
s
.71752 .87264 m
.72928 .87264 L
s
.72928 .87264 m
.7234 .86246 L
s
.7234 .86246 m
.72928 .85228 L
s
.72928 .85228 m
.74103 .85228 L
s
.74103 .85228 m
.74691 .86246 L
s
.74691 .86246 m
.75279 .85228 L
s
.75279 .85228 m
.76455 .85228 L
s
.76455 .85228 m
.77043 .86246 L
s
.77043 .86246 m
.76455 .87264 L
s
.76455 .87264 m
.77631 .87264 L
s
.77631 .87264 m
.78219 .88283 L
s
.78219 .88283 m
.77631 .89301 L
s
.77631 .89301 m
.76455 .89301 L
s
.76455 .89301 m
.77043 .90319 L
s
.77043 .90319 m
.76455 .91337 L
s
.76455 .91337 m
.77631 .91337 L
s
.77631 .91337 m
.78219 .92356 L
s
.78219 .92356 m
.78807 .91337 L
s
.78807 .91337 m
.79982 .91337 L
s
.79982 .91337 m
.8057 .92356 L
s
.8057 .92356 m
.79982 .93374 L
s
.79982 .93374 m
.81158 .93374 L
s
.81158 .93374 m
.81746 .94392 L
s
.81746 .94392 m
.82334 .93374 L
s
.82334 .93374 m
.8351 .93374 L
s
.8351 .93374 m
.82922 .92356 L
s
.82922 .92356 m
.8351 .91337 L
s
.8351 .91337 m
.84685 .91337 L
s
.84685 .91337 m
.85273 .92356 L
s
.85273 .92356 m
.85861 .91337 L
s
.85861 .91337 m
.87037 .91337 L
s
.87037 .91337 m
.86449 .90319 L
s
.86449 .90319 m
.87037 .89301 L
s
.87037 .89301 m
.85861 .89301 L
s
.85861 .89301 m
.85273 .88283 L
s
.85273 .88283 m
.85861 .87264 L
s
.85861 .87264 m
.87037 .87264 L
s
.87037 .87264 m
.86449 .86246 L
s
.86449 .86246 m
.87037 .85228 L
s
.87037 .85228 m
.88213 .85228 L
s
.88213 .85228 m
.88801 .86246 L
s
.88801 .86246 m
.89389 .85228 L
s
.89389 .85228 m
.90564 .85228 L
s
.90564 .85228 m
.91152 .86246 L
s
.91152 .86246 m
.90564 .87264 L
s
.90564 .87264 m
.9174 .87264 L
s
.9174 .87264 m
.92328 .88283 L
s
.92328 .88283 m
.92916 .87264 L
s
.92916 .87264 m
.94092 .87264 L
s
.94092 .87264 m
.93504 .86246 L
s
.93504 .86246 m
.94092 .85228 L
s
.94092 .85228 m
.95267 .85228 L
s
.95267 .85228 m
.95855 .86246 L
s
.95855 .86246 m
.96443 .85228 L
s
.96443 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.97031 .8421 L
s
.97031 .8421 m
.97619 .83191 L
s
.97619 .83191 m
.96443 .83191 L
s
.96443 .83191 m
.95855 .82173 L
s
.95855 .82173 m
.96443 .81155 L
s
.96443 .81155 m
.97619 .81155 L
s
.97619 .81155 m
.97031 .80137 L
s
.97031 .80137 m
.97619 .79118 L
s
.97619 .79118 m
.96443 .79118 L
s
.96443 .79118 m
.95855 .781 L
s
.95855 .781 m
.95267 .79118 L
s
.95267 .79118 m
.94092 .79118 L
s
.94092 .79118 m
.93504 .781 L
s
.93504 .781 m
.94092 .77082 L
s
.94092 .77082 m
.92916 .77082 L
s
.92916 .77082 m
.92328 .76064 L
s
.92328 .76064 m
.92916 .75045 L
s
.92916 .75045 m
.94092 .75045 L
s
.94092 .75045 m
.93504 .74027 L
s
.93504 .74027 m
.94092 .73009 L
s
.94092 .73009 m
.95267 .73009 L
s
.95267 .73009 m
.95855 .74027 L
s
.95855 .74027 m
.96443 .73009 L
s
.96443 .73009 m
.97619 .73009 L
s
.97619 .73009 m
.97031 .71991 L
s
.97031 .71991 m
.97619 .70972 L
s
.97619 .70972 m
.96443 .70972 L
s
.96443 .70972 m
.95855 .69954 L
s
.95855 .69954 m
.96443 .68936 L
s
.96443 .68936 m
.97619 .68936 L
s
.97619 .68936 m
.97031 .67918 L
s
.97031 .67918 m
.97619 .66899 L
s
.97619 .66899 m
.96443 .66899 L
s
.96443 .66899 m
.95855 .65881 L
s
.95855 .65881 m
.95267 .66899 L
s
.95267 .66899 m
.94092 .66899 L
s
.94092 .66899 m
.93504 .65881 L
s
.93504 .65881 m
.94092 .64863 L
s
.94092 .64863 m
.92916 .64863 L
s
.92916 .64863 m
.92328 .63845 L
s
.92328 .63845 m
.9174 .64863 L
s
.9174 .64863 m
.90564 .64863 L
s
.90564 .64863 m
.91152 .65881 L
s
.91152 .65881 m
.90564 .66899 L
s
.90564 .66899 m
.89389 .66899 L
s
.89389 .66899 m
.88801 .65881 L
s
.88801 .65881 m
.88213 .66899 L
s
.88213 .66899 m
.87037 .66899 L
s
.87037 .66899 m
.86449 .65881 L
s
.86449 .65881 m
.87037 .64863 L
s
.87037 .64863 m
.85861 .64863 L
s
.85861 .64863 m
.85273 .63845 L
s
.85273 .63845 m
.85861 .62826 L
s
.85861 .62826 m
.87037 .62826 L
s
.87037 .62826 m
.86449 .61808 L
s
.86449 .61808 m
.87037 .6079 L
s
.87037 .6079 m
.85861 .6079 L
s
.85861 .6079 m
.85273 .59772 L
s
.85273 .59772 m
.84685 .6079 L
s
.84685 .6079 m
.8351 .6079 L
s
.8351 .6079 m
.82922 .59772 L
s
.82922 .59772 m
.8351 .58753 L
s
.8351 .58753 m
.82334 .58753 L
s
.82334 .58753 m
.81746 .57735 L
s
.81746 .57735 m
.82334 .56717 L
s
.82334 .56717 m
.8351 .56717 L
s
.8351 .56717 m
.82922 .55699 L
s
.82922 .55699 m
.8351 .5468 L
s
.8351 .5468 m
.84685 .5468 L
s
.84685 .5468 m
.85273 .55699 L
s
.85273 .55699 m
.85861 .5468 L
s
.85861 .5468 m
.87037 .5468 L
s
.87037 .5468 m
.86449 .53662 L
s
.86449 .53662 m
.87037 .52644 L
s
.87037 .52644 m
.85861 .52644 L
s
.85861 .52644 m
.85273 .51626 L
s
.85273 .51626 m
.85861 .50607 L
s
.85861 .50607 m
.87037 .50607 L
s
.87037 .50607 m
.86449 .49589 L
s
.86449 .49589 m
.87037 .48571 L
s
.87037 .48571 m
.88213 .48571 L
s
.88213 .48571 m
.88801 .49589 L
s
.88801 .49589 m
.89389 .48571 L
s
.89389 .48571 m
.90564 .48571 L
s
.90564 .48571 m
.91152 .49589 L
s
.91152 .49589 m
.90564 .50607 L
s
.90564 .50607 m
.9174 .50607 L
s
.9174 .50607 m
.92328 .51626 L
s
.92328 .51626 m
.92916 .50607 L
s
.92916 .50607 m
.94092 .50607 L
s
.94092 .50607 m
.93504 .49589 L
s
.93504 .49589 m
.94092 .48571 L
s
.94092 .48571 m
.95267 .48571 L
s
.95267 .48571 m
.95855 .49589 L
s
.95855 .49589 m
.96443 .48571 L
s
.96443 .48571 m
.97619 .48571 L
s
.97619 .48571 m
.97031 .47552 L
s
.97031 .47552 m
.97619 .46534 L
s
.97619 .46534 m
.96443 .46534 L
s
.96443 .46534 m
.95855 .45516 L
s
.95855 .45516 m
.96443 .44498 L
s
.96443 .44498 m
.97619 .44498 L
s
.97619 .44498 m
.97031 .43479 L
s
.97031 .43479 m
.97619 .42461 L
s
.97619 .42461 m
.96443 .42461 L
s
.96443 .42461 m
.95855 .41443 L
s
.95855 .41443 m
.95267 .42461 L
s
.95267 .42461 m
.94092 .42461 L
s
.94092 .42461 m
.93504 .41443 L
s
.93504 .41443 m
.94092 .40425 L
s
.94092 .40425 m
.92916 .40425 L
s
.92916 .40425 m
.92328 .39406 L
s
.92328 .39406 m
.92916 .38388 L
s
.92916 .38388 m
.94092 .38388 L
s
.94092 .38388 m
.93504 .3737 L
s
.93504 .3737 m
.94092 .36352 L
s
.94092 .36352 m
.95267 .36352 L
s
.95267 .36352 m
.95855 .3737 L
s
.95855 .3737 m
.96443 .36352 L
s
.96443 .36352 m
.97619 .36352 L
s
.97619 .36352 m
.97031 .35333 L
s
.97031 .35333 m
.97619 .34315 L
s
.97619 .34315 m
.96443 .34315 L
s
.96443 .34315 m
.95855 .33297 L
s
.95855 .33297 m
.96443 .32279 L
s
.96443 .32279 m
.97619 .32279 L
s
.97619 .32279 m
.97031 .3126 L
s
.97031 .3126 m
.97619 .30242 L
s
.97619 .30242 m
.96443 .30242 L
s
.96443 .30242 m
.95855 .29224 L
s
.95855 .29224 m
.95267 .30242 L
s
.95267 .30242 m
.94092 .30242 L
s
.94092 .30242 m
.93504 .29224 L
s
.93504 .29224 m
.94092 .28206 L
s
.94092 .28206 m
.92916 .28206 L
s
.92916 .28206 m
.92328 .27187 L
s
.92328 .27187 m
.9174 .28206 L
s
.9174 .28206 m
.90564 .28206 L
s
.90564 .28206 m
.91152 .29224 L
s
.91152 .29224 m
.90564 .30242 L
s
.90564 .30242 m
.89389 .30242 L
s
.89389 .30242 m
.88801 .29224 L
s
.88801 .29224 m
.88213 .30242 L
s
.88213 .30242 m
.87037 .30242 L
s
.87037 .30242 m
.86449 .29224 L
s
.86449 .29224 m
.87037 .28206 L
s
.87037 .28206 m
.85861 .28206 L
s
.85861 .28206 m
.85273 .27187 L
s
.85273 .27187 m
.85861 .26169 L
s
.85861 .26169 m
.87037 .26169 L
s
.87037 .26169 m
.86449 .25151 L
s
.86449 .25151 m
.87037 .24133 L
s
.87037 .24133 m
.85861 .24133 L
s
.85861 .24133 m
.85273 .23114 L
s
.85273 .23114 m
.84685 .24133 L
s
.84685 .24133 m
.8351 .24133 L
s
.8351 .24133 m
.82922 .23114 L
s
.82922 .23114 m
.8351 .22096 L
s
.8351 .22096 m
.82334 .22096 L
s
.82334 .22096 m
.81746 .21078 L
s
.81746 .21078 m
.81158 .22096 L
s
.81158 .22096 m
.79982 .22096 L
s
.79982 .22096 m
.8057 .23114 L
s
.8057 .23114 m
.79982 .24133 L
s
.79982 .24133 m
.78807 .24133 L
s
.78807 .24133 m
.78219 .23114 L
s
.78219 .23114 m
.77631 .24133 L
s
.77631 .24133 m
.76455 .24133 L
s
.76455 .24133 m
.77043 .25151 L
s
.77043 .25151 m
.76455 .26169 L
s
.76455 .26169 m
.77631 .26169 L
s
.77631 .26169 m
.78219 .27187 L
s
.78219 .27187 m
.77631 .28206 L
s
.77631 .28206 m
.76455 .28206 L
s
.76455 .28206 m
.77043 .29224 L
s
.77043 .29224 m
.76455 .30242 L
s
.76455 .30242 m
.75279 .30242 L
s
.75279 .30242 m
.74691 .29224 L
s
.74691 .29224 m
.74103 .30242 L
s
.74103 .30242 m
.72928 .30242 L
s
.72928 .30242 m
.7234 .29224 L
s
.7234 .29224 m
.72928 .28206 L
s
.72928 .28206 m
.71752 .28206 L
s
.71752 .28206 m
.71164 .27187 L
s
.71164 .27187 m
.70576 .28206 L
s
.70576 .28206 m
.694 .28206 L
s
.694 .28206 m
.69988 .29224 L
s
.69988 .29224 m
.694 .30242 L
s
.694 .30242 m
.68225 .30242 L
s
.68225 .30242 m
.67637 .29224 L
s
.67637 .29224 m
.67049 .30242 L
s
.67049 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.65285 .29224 L
s
.65285 .29224 m
.65873 .28206 L
s
.65873 .28206 m
.64697 .28206 L
s
.64697 .28206 m
.64109 .27187 L
s
.64109 .27187 m
.64697 .26169 L
s
.64697 .26169 m
.65873 .26169 L
s
.65873 .26169 m
.65285 .25151 L
s
.65285 .25151 m
.65873 .24133 L
s
.65873 .24133 m
.64697 .24133 L
s
.64697 .24133 m
.64109 .23114 L
s
.64109 .23114 m
.63521 .24133 L
s
.63521 .24133 m
.62346 .24133 L
s
.62346 .24133 m
.61758 .23114 L
s
.61758 .23114 m
.62346 .22096 L
s
.62346 .22096 m
.6117 .22096 L
s
.6117 .22096 m
.60582 .21078 L
s
.60582 .21078 m
.6117 .2006 L
s
.6117 .2006 m
.62346 .2006 L
s
.62346 .2006 m
.61758 .19041 L
s
.61758 .19041 m
.62346 .18023 L
s
.62346 .18023 m
.63521 .18023 L
s
.63521 .18023 m
.64109 .19041 L
s
.64109 .19041 m
.64697 .18023 L
s
.64697 .18023 m
.65873 .18023 L
s
.65873 .18023 m
.65285 .17005 L
s
.65285 .17005 m
.65873 .15987 L
s
.65873 .15987 m
.64697 .15987 L
s
.64697 .15987 m
.64109 .14968 L
s
.64109 .14968 m
.64697 .1395 L
s
.64697 .1395 m
.65873 .1395 L
s
.65873 .1395 m
.65285 .12932 L
s
.65285 .12932 m
.65873 .11914 L
s
.65873 .11914 m
.64697 .11914 L
s
.64697 .11914 m
.64109 .10895 L
s
.64109 .10895 m
.63521 .11914 L
s
.63521 .11914 m
.62346 .11914 L
s
.62346 .11914 m
.61758 .10895 L
s
.61758 .10895 m
.62346 .09877 L
s
.62346 .09877 m
.6117 .09877 L
s
.6117 .09877 m
.60582 .08859 L
s
.60582 .08859 m
.59994 .09877 L
s
.59994 .09877 m
.58818 .09877 L
s
.58818 .09877 m
.59406 .10895 L
s
.59406 .10895 m
.58818 .11914 L
s
.58818 .11914 m
.57643 .11914 L
s
.57643 .11914 m
.57055 .10895 L
s
.57055 .10895 m
.56467 .11914 L
s
.56467 .11914 m
.55291 .11914 L
s
.55291 .11914 m
.54703 .10895 L
s
.54703 .10895 m
.55291 .09877 L
s
.55291 .09877 m
.54115 .09877 L
s
.54115 .09877 m
.53527 .08859 L
s
.53527 .08859 m
.54115 .07841 L
s
.54115 .07841 m
.55291 .07841 L
s
.55291 .07841 m
.54703 .06822 L
s
.54703 .06822 m
.55291 .05804 L
s
.55291 .05804 m
.54115 .05804 L
s
.54115 .05804 m
.53527 .04786 L
s
.53527 .04786 m
.52939 .05804 L
s
.52939 .05804 m
.51764 .05804 L
s
.51764 .05804 m
.51176 .04786 L
s
.51176 .04786 m
.51764 .03768 L
s
.51764 .03768 m
.50588 .03768 L
s
.50588 .03768 m
.5 .02749 L
s
.5 .02749 m
.49412 .03768 L
s
.49412 .03768 m
.48236 .03768 L
s
.48236 .03768 m
.48824 .04786 L
s
.48824 .04786 m
.48236 .05804 L
s
.48236 .05804 m
.47061 .05804 L
s
.47061 .05804 m
.46473 .04786 L
s
.46473 .04786 m
.45885 .05804 L
s
.45885 .05804 m
.44709 .05804 L
s
.44709 .05804 m
.45297 .06822 L
s
.45297 .06822 m
.44709 .07841 L
s
.44709 .07841 m
.45885 .07841 L
s
.45885 .07841 m
.46473 .08859 L
s
.46473 .08859 m
.45885 .09877 L
s
.45885 .09877 m
.44709 .09877 L
s
.44709 .09877 m
.45297 .10895 L
s
.45297 .10895 m
.44709 .11914 L
s
.44709 .11914 m
.43533 .11914 L
s
.43533 .11914 m
.42945 .10895 L
s
.42945 .10895 m
.42357 .11914 L
s
.42357 .11914 m
.41182 .11914 L
s
.41182 .11914 m
.40594 .10895 L
s
.40594 .10895 m
.41182 .09877 L
s
.41182 .09877 m
.40006 .09877 L
s
.40006 .09877 m
.39418 .08859 L
s
.39418 .08859 m
.3883 .09877 L
s
.3883 .09877 m
.37654 .09877 L
s
.37654 .09877 m
.38242 .10895 L
s
.38242 .10895 m
.37654 .11914 L
s
.37654 .11914 m
.36479 .11914 L
s
.36479 .11914 m
.35891 .10895 L
s
.35891 .10895 m
.35303 .11914 L
s
.35303 .11914 m
.34127 .11914 L
s
.34127 .11914 m
.34715 .12932 L
s
.34715 .12932 m
.34127 .1395 L
s
.34127 .1395 m
.35303 .1395 L
s
.35303 .1395 m
.35891 .14968 L
s
.35891 .14968 m
.35303 .15987 L
s
.35303 .15987 m
.34127 .15987 L
s
.34127 .15987 m
.34715 .17005 L
s
.34715 .17005 m
.34127 .18023 L
s
.34127 .18023 m
.35303 .18023 L
s
.35303 .18023 m
.35891 .19041 L
s
.35891 .19041 m
.36479 .18023 L
s
.36479 .18023 m
.37654 .18023 L
s
.37654 .18023 m
.38242 .19041 L
s
.38242 .19041 m
.37654 .2006 L
s
.37654 .2006 m
.3883 .2006 L
s
.3883 .2006 m
.39418 .21078 L
s
.39418 .21078 m
.3883 .22096 L
s
.3883 .22096 m
.37654 .22096 L
s
.37654 .22096 m
.38242 .23114 L
s
.38242 .23114 m
.37654 .24133 L
s
.37654 .24133 m
.36479 .24133 L
s
.36479 .24133 m
.35891 .23114 L
s
.35891 .23114 m
.35303 .24133 L
s
.35303 .24133 m
.34127 .24133 L
s
.34127 .24133 m
.34715 .25151 L
s
.34715 .25151 m
.34127 .26169 L
s
.34127 .26169 m
.35303 .26169 L
s
.35303 .26169 m
.35891 .27187 L
s
.35891 .27187 m
.35303 .28206 L
s
.35303 .28206 m
.34127 .28206 L
s
.34127 .28206 m
.34715 .29224 L
s
.34715 .29224 m
.34127 .30242 L
s
.34127 .30242 m
.32951 .30242 L
s
.32951 .30242 m
.32363 .29224 L
s
.32363 .29224 m
.31775 .30242 L
s
.31775 .30242 m
.306 .30242 L
s
.306 .30242 m
.30012 .29224 L
s
.30012 .29224 m
.306 .28206 L
s
.306 .28206 m
.29424 .28206 L
s
.29424 .28206 m
.28836 .27187 L
s
.28836 .27187 m
.28248 .28206 L
s
.28248 .28206 m
.27072 .28206 L
s
.27072 .28206 m
.2766 .29224 L
s
.2766 .29224 m
.27072 .30242 L
s
.27072 .30242 m
.25897 .30242 L
s
.25897 .30242 m
.25309 .29224 L
s
.25309 .29224 m
.24721 .30242 L
s
.24721 .30242 m
.23545 .30242 L
s
.23545 .30242 m
.22957 .29224 L
s
.22957 .29224 m
.23545 .28206 L
s
.23545 .28206 m
.22369 .28206 L
s
.22369 .28206 m
.21781 .27187 L
s
.21781 .27187 m
.22369 .26169 L
s
.22369 .26169 m
.23545 .26169 L
s
.23545 .26169 m
.22957 .25151 L
s
.22957 .25151 m
.23545 .24133 L
s
.23545 .24133 m
.22369 .24133 L
s
.22369 .24133 m
.21781 .23114 L
s
.21781 .23114 m
.21193 .24133 L
s
.21193 .24133 m
.20018 .24133 L
s
.20018 .24133 m
.1943 .23114 L
s
.1943 .23114 m
.20018 .22096 L
s
.20018 .22096 m
.18842 .22096 L
s
.18842 .22096 m
.18254 .21078 L
s
.18254 .21078 m
.17666 .22096 L
s
.17666 .22096 m
.1649 .22096 L
s
.1649 .22096 m
.17078 .23114 L
s
.17078 .23114 m
.1649 .24133 L
s
.1649 .24133 m
.15315 .24133 L
s
.15315 .24133 m
.14727 .23114 L
s
.14727 .23114 m
.14139 .24133 L
s
.14139 .24133 m
.12963 .24133 L
s
.12963 .24133 m
.13551 .25151 L
s
.13551 .25151 m
.12963 .26169 L
s
.12963 .26169 m
.14139 .26169 L
s
.14139 .26169 m
.14727 .27187 L
s
.14727 .27187 m
.14139 .28206 L
s
.14139 .28206 m
.12963 .28206 L
s
.12963 .28206 m
.13551 .29224 L
s
.13551 .29224 m
.12963 .30242 L
s
.12963 .30242 m
.11787 .30242 L
s
.11787 .30242 m
.11199 .29224 L
s
.11199 .29224 m
.10611 .30242 L
s
.10611 .30242 m
.09436 .30242 L
s
.09436 .30242 m
.08848 .29224 L
s
.08848 .29224 m
.09436 .28206 L
s
.09436 .28206 m
.0826 .28206 L
s
.0826 .28206 m
.07672 .27187 L
s
.07672 .27187 m
.07084 .28206 L
s
.07084 .28206 m
.05908 .28206 L
s
.05908 .28206 m
.06496 .29224 L
s
.06496 .29224 m
.05908 .30242 L
s
.05908 .30242 m
.04733 .30242 L
s
.04733 .30242 m
.04145 .29224 L
s
.04145 .29224 m
.03557 .30242 L
s
.03557 .30242 m
.02381 .30242 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0130918 0.97619 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0206197 0.952381 [
[ 0 0 0 0 ]
[ 1 .86603 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .86603 L
0 .86603 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .02062 m
.97619 .02062 L
s
.97619 .02062 m
.5 .84541 L
s
.5 .84541 m
.02381 .02062 L
s
.02381 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.5 .29555 L
s
.5 .29555 m
.65873 .02062 L
s
.65873 .02062 m
.97619 .02062 L
s
.97619 .02062 m
.81746 .29555 L
s
.81746 .29555 m
.5 .29555 L
s
.5 .29555 m
.65873 .57048 L
s
.65873 .57048 m
.5 .84541 L
s
.5 .84541 m
.34127 .57048 L
s
.34127 .57048 m
.5 .29555 L
s
.5 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.02381 .02062 L
s
.02381 .02062 m
.12963 .02062 L
s
.12963 .02062 m
.18254 .11226 L
s
.18254 .11226 m
.23545 .02062 L
s
.23545 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.39418 .11226 L
s
.39418 .11226 m
.34127 .20391 L
s
.34127 .20391 m
.44709 .20391 L
s
.44709 .20391 m
.5 .29555 L
s
.5 .29555 m
.55291 .20391 L
s
.55291 .20391 m
.65873 .20391 L
s
.65873 .20391 m
.60582 .11226 L
s
.60582 .11226 m
.65873 .02062 L
s
.65873 .02062 m
.76455 .02062 L
s
.76455 .02062 m
.81746 .11226 L
s
.81746 .11226 m
.87037 .02062 L
s
.87037 .02062 m
.97619 .02062 L
s
.97619 .02062 m
.92328 .11226 L
s
.92328 .11226 m
.81746 .11226 L
s
.81746 .11226 m
.87037 .20391 L
s
.87037 .20391 m
.81746 .29555 L
s
.81746 .29555 m
.71164 .29555 L
s
.71164 .29555 m
.65873 .20391 L
s
.65873 .20391 m
.60582 .29555 L
s
.60582 .29555 m
.5 .29555 L
s
.5 .29555 m
.55291 .38719 L
s
.55291 .38719 m
.5 .47883 L
s
.5 .47883 m
.60582 .47883 L
s
.60582 .47883 m
.65873 .57048 L
s
.65873 .57048 m
.60582 .66212 L
s
.60582 .66212 m
.5 .66212 L
s
.5 .66212 m
.55291 .75376 L
s
.55291 .75376 m
.5 .84541 L
s
.5 .84541 m
.44709 .75376 L
s
.44709 .75376 m
.5 .66212 L
s
.5 .66212 m
.39418 .66212 L
s
.39418 .66212 m
.34127 .57048 L
s
.34127 .57048 m
.39418 .47883 L
s
.39418 .47883 m
.5 .47883 L
s
.5 .47883 m
.44709 .38719 L
s
.44709 .38719 m
.5 .29555 L
s
.5 .29555 m
.39418 .29555 L
s
.39418 .29555 m
.34127 .20391 L
s
.34127 .20391 m
.28836 .29555 L
s
.28836 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.12963 .20391 L
s
.12963 .20391 m
.18254 .11226 L
s
.18254 .11226 m
.07672 .11226 L
s
.07672 .11226 m
.02381 .02062 L
s
.02381 .02062 m
.05908 .02062 L
s
.05908 .02062 m
.07672 .05117 L
s
.07672 .05117 m
.09436 .02062 L
s
.09436 .02062 m
.12963 .02062 L
s
.12963 .02062 m
.14727 .05117 L
s
.14727 .05117 m
.12963 .08171 L
s
.12963 .08171 m
.1649 .08171 L
s
.1649 .08171 m
.18254 .11226 L
s
.18254 .11226 m
.20018 .08171 L
s
.20018 .08171 m
.23545 .08171 L
s
.23545 .08171 m
.21781 .05117 L
s
.21781 .05117 m
.23545 .02062 L
s
.23545 .02062 m
.27072 .02062 L
s
.27072 .02062 m
.28836 .05117 L
s
.28836 .05117 m
.306 .02062 L
s
.306 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.35891 .05117 L
s
.35891 .05117 m
.34127 .08171 L
s
.34127 .08171 m
.37654 .08171 L
s
.37654 .08171 m
.39418 .11226 L
s
.39418 .11226 m
.37654 .14281 L
s
.37654 .14281 m
.34127 .14281 L
s
.34127 .14281 m
.35891 .17336 L
s
.35891 .17336 m
.34127 .20391 L
s
.34127 .20391 m
.37654 .20391 L
s
.37654 .20391 m
.39418 .23445 L
s
.39418 .23445 m
.41182 .20391 L
s
.41182 .20391 m
.44709 .20391 L
s
.44709 .20391 m
.46473 .23445 L
s
.46473 .23445 m
.44709 .265 L
s
.44709 .265 m
.48236 .265 L
s
.48236 .265 m
.5 .29555 L
s
.5 .29555 m
.51764 .265 L
s
.51764 .265 m
.55291 .265 L
s
.55291 .265 m
.53527 .23445 L
s
.53527 .23445 m
.55291 .20391 L
s
.55291 .20391 m
.58818 .20391 L
s
.58818 .20391 m
.60582 .23445 L
s
.60582 .23445 m
.62346 .20391 L
s
.62346 .20391 m
.65873 .20391 L
s
.65873 .20391 m
.64109 .17336 L
s
.64109 .17336 m
.65873 .14281 L
s
.65873 .14281 m
.62346 .14281 L
s
.62346 .14281 m
.60582 .11226 L
s
.60582 .11226 m
.62346 .08171 L
s
.62346 .08171 m
.65873 .08171 L
s
.65873 .08171 m
.64109 .05117 L
s
.64109 .05117 m
.65873 .02062 L
s
.65873 .02062 m
.694 .02062 L
s
.694 .02062 m
.71164 .05117 L
s
.71164 .05117 m
.72928 .02062 L
s
.72928 .02062 m
.76455 .02062 L
s
.76455 .02062 m
.78219 .05117 L
s
.78219 .05117 m
.76455 .08171 L
s
.76455 .08171 m
.79982 .08171 L
s
.79982 .08171 m
.81746 .11226 L
s
.81746 .11226 m
.8351 .08171 L
s
.8351 .08171 m
.87037 .08171 L
s
.87037 .08171 m
.85273 .05117 L
s
.85273 .05117 m
.87037 .02062 L
s
.87037 .02062 m
.90564 .02062 L
s
.90564 .02062 m
.92328 .05117 L
s
.92328 .05117 m
.94092 .02062 L
s
.94092 .02062 m
.97619 .02062 L
s
.97619 .02062 m
.95855 .05117 L
s
.95855 .05117 m
.92328 .05117 L
s
.92328 .05117 m
.94092 .08171 L
s
.94092 .08171 m
.92328 .11226 L
s
.92328 .11226 m
.88801 .11226 L
s
.88801 .11226 m
.87037 .08171 L
s
.87037 .08171 m
.85273 .11226 L
s
.85273 .11226 m
.81746 .11226 L
s
.81746 .11226 m
.8351 .14281 L
s
.8351 .14281 m
.81746 .17336 L
s
.81746 .17336 m
.85273 .17336 L
s
.85273 .17336 m
.87037 .20391 L
s
.87037 .20391 m
.85273 .23445 L
s
.85273 .23445 m
.81746 .23445 L
s
.81746 .23445 m
.8351 .265 L
s
.8351 .265 m
.81746 .29555 L
s
.81746 .29555 m
.78219 .29555 L
s
.78219 .29555 m
.76455 .265 L
s
.76455 .265 m
.74691 .29555 L
s
.74691 .29555 m
.71164 .29555 L
s
.71164 .29555 m
.694 .265 L
s
.694 .265 m
.71164 .23445 L
s
.71164 .23445 m
.67637 .23445 L
s
.67637 .23445 m
.65873 .20391 L
s
.65873 .20391 m
.64109 .23445 L
s
.64109 .23445 m
.60582 .23445 L
s
.60582 .23445 m
.62346 .265 L
s
.62346 .265 m
.60582 .29555 L
s
.60582 .29555 m
.57055 .29555 L
s
.57055 .29555 m
.55291 .265 L
s
.55291 .265 m
.53527 .29555 L
s
.53527 .29555 m
.5 .29555 L
s
.5 .29555 m
.51764 .3261 L
s
.51764 .3261 m
.5 .35664 L
s
.5 .35664 m
.53527 .35664 L
s
.53527 .35664 m
.55291 .38719 L
s
.55291 .38719 m
.53527 .41774 L
s
.53527 .41774 m
.5 .41774 L
s
.5 .41774 m
.51764 .44829 L
s
.51764 .44829 m
.5 .47883 L
s
.5 .47883 m
.53527 .47883 L
s
.53527 .47883 m
.55291 .50938 L
s
.55291 .50938 m
.57055 .47883 L
s
.57055 .47883 m
.60582 .47883 L
s
.60582 .47883 m
.62346 .50938 L
s
.62346 .50938 m
.60582 .53993 L
s
.60582 .53993 m
.64109 .53993 L
s
.64109 .53993 m
.65873 .57048 L
s
.65873 .57048 m
.64109 .60102 L
s
.64109 .60102 m
.60582 .60102 L
s
.60582 .60102 m
.62346 .63157 L
s
.62346 .63157 m
.60582 .66212 L
s
.60582 .66212 m
.57055 .66212 L
s
.57055 .66212 m
.55291 .63157 L
s
.55291 .63157 m
.53527 .66212 L
s
.53527 .66212 m
.5 .66212 L
s
.5 .66212 m
.51764 .69267 L
s
.51764 .69267 m
.5 .72322 L
s
.5 .72322 m
.53527 .72322 L
s
.53527 .72322 m
.55291 .75376 L
s
.55291 .75376 m
.53527 .78431 L
s
.53527 .78431 m
.5 .78431 L
s
.5 .78431 m
.51764 .81486 L
s
.51764 .81486 m
.5 .84541 L
s
.5 .84541 m
.48236 .81486 L
s
.48236 .81486 m
.5 .78431 L
s
.5 .78431 m
.46473 .78431 L
s
.46473 .78431 m
.44709 .75376 L
s
.44709 .75376 m
.46473 .72322 L
s
.46473 .72322 m
.5 .72322 L
s
.5 .72322 m
.48236 .69267 L
s
.48236 .69267 m
.5 .66212 L
s
.5 .66212 m
.46473 .66212 L
s
.46473 .66212 m
.44709 .63157 L
s
.44709 .63157 m
.42945 .66212 L
s
.42945 .66212 m
.39418 .66212 L
s
.39418 .66212 m
.37654 .63157 L
s
.37654 .63157 m
.39418 .60102 L
s
.39418 .60102 m
.35891 .60102 L
s
.35891 .60102 m
.34127 .57048 L
s
.34127 .57048 m
.35891 .53993 L
s
.35891 .53993 m
.39418 .53993 L
s
.39418 .53993 m
.37654 .50938 L
s
.37654 .50938 m
.39418 .47883 L
s
.39418 .47883 m
.42945 .47883 L
s
.42945 .47883 m
.44709 .50938 L
s
.44709 .50938 m
.46473 .47883 L
s
.46473 .47883 m
.5 .47883 L
s
.5 .47883 m
.48236 .44829 L
s
.48236 .44829 m
.5 .41774 L
s
.5 .41774 m
.46473 .41774 L
s
.46473 .41774 m
.44709 .38719 L
s
.44709 .38719 m
.46473 .35664 L
s
.46473 .35664 m
.5 .35664 L
s
.5 .35664 m
.48236 .3261 L
s
.48236 .3261 m
.5 .29555 L
s
.5 .29555 m
.46473 .29555 L
s
.46473 .29555 m
.44709 .265 L
s
.44709 .265 m
.42945 .29555 L
s
.42945 .29555 m
.39418 .29555 L
s
.39418 .29555 m
.37654 .265 L
s
.37654 .265 m
.39418 .23445 L
s
.39418 .23445 m
.35891 .23445 L
s
.35891 .23445 m
.34127 .20391 L
s
.34127 .20391 m
.32363 .23445 L
s
.32363 .23445 m
.28836 .23445 L
s
.28836 .23445 m
.306 .265 L
s
.306 .265 m
.28836 .29555 L
s
.28836 .29555 m
.25309 .29555 L
s
.25309 .29555 m
.23545 .265 L
s
.23545 .265 m
.21781 .29555 L
s
.21781 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.1649 .265 L
s
.1649 .265 m
.18254 .23445 L
s
.18254 .23445 m
.14727 .23445 L
s
.14727 .23445 m
.12963 .20391 L
s
.12963 .20391 m
.14727 .17336 L
s
.14727 .17336 m
.18254 .17336 L
s
.18254 .17336 m
.1649 .14281 L
s
.1649 .14281 m
.18254 .11226 L
s
.18254 .11226 m
.14727 .11226 L
s
.14727 .11226 m
.12963 .08171 L
s
.12963 .08171 m
.11199 .11226 L
s
.11199 .11226 m
.07672 .11226 L
s
.07672 .11226 m
.05908 .08171 L
s
.05908 .08171 m
.07672 .05117 L
s
.07672 .05117 m
.04145 .05117 L
s
.04145 .05117 m
.02381 .02062 L
s
.02381 .02062 m
.03557 .02062 L
s
.03557 .02062 m
.04145 .0308 L
s
.04145 .0308 m
.04733 .02062 L
s
.04733 .02062 m
.05908 .02062 L
s
.05908 .02062 m
.06496 .0308 L
s
.06496 .0308 m
.05908 .04098 L
s
.05908 .04098 m
.07084 .04098 L
s
.07084 .04098 m
.07672 .05117 L
s
.07672 .05117 m
.0826 .04098 L
s
.0826 .04098 m
.09436 .04098 L
s
.09436 .04098 m
.08848 .0308 L
s
.08848 .0308 m
.09436 .02062 L
s
.09436 .02062 m
.10611 .02062 L
s
.10611 .02062 m
.11199 .0308 L
s
.11199 .0308 m
.11787 .02062 L
s
.11787 .02062 m
.12963 .02062 L
s
.12963 .02062 m
.13551 .0308 L
s
.13551 .0308 m
.12963 .04098 L
s
.12963 .04098 m
.14139 .04098 L
s
.14139 .04098 m
.14727 .05117 L
s
.14727 .05117 m
.14139 .06135 L
s
.14139 .06135 m
.12963 .06135 L
s
.12963 .06135 m
.13551 .07153 L
s
.13551 .07153 m
.12963 .08171 L
s
.12963 .08171 m
.14139 .08171 L
s
.14139 .08171 m
.14727 .0919 L
s
.14727 .0919 m
.15315 .08171 L
s
.15315 .08171 m
.1649 .08171 L
s
.1649 .08171 m
.17078 .0919 L
s
.17078 .0919 m
.1649 .10208 L
s
.1649 .10208 m
.17666 .10208 L
s
.17666 .10208 m
.18254 .11226 L
s
.18254 .11226 m
.18842 .10208 L
s
.18842 .10208 m
.20018 .10208 L
s
.20018 .10208 m
.1943 .0919 L
s
.1943 .0919 m
.20018 .08171 L
s
.20018 .08171 m
.21193 .08171 L
s
.21193 .08171 m
.21781 .0919 L
s
.21781 .0919 m
.22369 .08171 L
s
.22369 .08171 m
.23545 .08171 L
s
.23545 .08171 m
.22957 .07153 L
s
.22957 .07153 m
.23545 .06135 L
s
.23545 .06135 m
.22369 .06135 L
s
.22369 .06135 m
.21781 .05117 L
s
.21781 .05117 m
.22369 .04098 L
s
.22369 .04098 m
.23545 .04098 L
s
.23545 .04098 m
.22957 .0308 L
s
.22957 .0308 m
.23545 .02062 L
s
.23545 .02062 m
.24721 .02062 L
s
.24721 .02062 m
.25309 .0308 L
s
.25309 .0308 m
.25897 .02062 L
s
.25897 .02062 m
.27072 .02062 L
s
.27072 .02062 m
.2766 .0308 L
s
.2766 .0308 m
.27072 .04098 L
s
.27072 .04098 m
.28248 .04098 L
s
.28248 .04098 m
.28836 .05117 L
s
.28836 .05117 m
.29424 .04098 L
s
.29424 .04098 m
.306 .04098 L
s
.306 .04098 m
.30012 .0308 L
s
.30012 .0308 m
.306 .02062 L
s
.306 .02062 m
.31775 .02062 L
s
.31775 .02062 m
.32363 .0308 L
s
.32363 .0308 m
.32951 .02062 L
s
.32951 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.34715 .0308 L
s
.34715 .0308 m
.34127 .04098 L
s
.34127 .04098 m
.35303 .04098 L
s
.35303 .04098 m
.35891 .05117 L
s
.35891 .05117 m
.35303 .06135 L
s
.35303 .06135 m
.34127 .06135 L
s
.34127 .06135 m
.34715 .07153 L
s
.34715 .07153 m
.34127 .08171 L
s
.34127 .08171 m
.35303 .08171 L
s
.35303 .08171 m
.35891 .0919 L
s
.35891 .0919 m
.36479 .08171 L
s
.36479 .08171 m
.37654 .08171 L
s
.37654 .08171 m
.38242 .0919 L
s
.38242 .0919 m
.37654 .10208 L
s
.37654 .10208 m
.3883 .10208 L
s
.3883 .10208 m
.39418 .11226 L
s
.39418 .11226 m
.3883 .12245 L
s
.3883 .12245 m
.37654 .12245 L
s
.37654 .12245 m
.38242 .13263 L
s
.38242 .13263 m
.37654 .14281 L
s
.37654 .14281 m
.36479 .14281 L
s
.36479 .14281 m
.35891 .13263 L
s
.35891 .13263 m
.35303 .14281 L
s
.35303 .14281 m
.34127 .14281 L
s
.34127 .14281 m
.34715 .15299 L
s
.34715 .15299 m
.34127 .16318 L
s
.34127 .16318 m
.35303 .16318 L
s
.35303 .16318 m
.35891 .17336 L
s
.35891 .17336 m
.35303 .18354 L
s
.35303 .18354 m
.34127 .18354 L
s
.34127 .18354 m
.34715 .19372 L
s
.34715 .19372 m
.34127 .20391 L
s
.34127 .20391 m
.35303 .20391 L
s
.35303 .20391 m
.35891 .21409 L
s
.35891 .21409 m
.36479 .20391 L
s
.36479 .20391 m
.37654 .20391 L
s
.37654 .20391 m
.38242 .21409 L
s
.38242 .21409 m
.37654 .22427 L
s
.37654 .22427 m
.3883 .22427 L
s
.3883 .22427 m
.39418 .23445 L
s
.39418 .23445 m
.40006 .22427 L
s
.40006 .22427 m
.41182 .22427 L
s
.41182 .22427 m
.40594 .21409 L
s
.40594 .21409 m
.41182 .20391 L
s
.41182 .20391 m
.42357 .20391 L
s
.42357 .20391 m
.42945 .21409 L
s
.42945 .21409 m
.43533 .20391 L
s
.43533 .20391 m
.44709 .20391 L
s
.44709 .20391 m
.45297 .21409 L
s
.45297 .21409 m
.44709 .22427 L
s
.44709 .22427 m
.45885 .22427 L
s
.45885 .22427 m
.46473 .23445 L
s
.46473 .23445 m
.45885 .24464 L
s
.45885 .24464 m
.44709 .24464 L
s
.44709 .24464 m
.45297 .25482 L
s
.45297 .25482 m
.44709 .265 L
s
.44709 .265 m
.45885 .265 L
s
.45885 .265 m
.46473 .27518 L
s
.46473 .27518 m
.47061 .265 L
s
.47061 .265 m
.48236 .265 L
s
.48236 .265 m
.48824 .27518 L
s
.48824 .27518 m
.48236 .28537 L
s
.48236 .28537 m
.49412 .28537 L
s
.49412 .28537 m
.5 .29555 L
s
.5 .29555 m
.50588 .28537 L
s
.50588 .28537 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .27518 L
s
.51176 .27518 m
.51764 .265 L
s
.51764 .265 m
.52939 .265 L
s
.52939 .265 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .265 L
s
.54115 .265 m
.55291 .265 L
s
.55291 .265 m
.54703 .25482 L
s
.54703 .25482 m
.55291 .24464 L
s
.55291 .24464 m
.54115 .24464 L
s
.54115 .24464 m
.53527 .23445 L
s
.53527 .23445 m
.54115 .22427 L
s
.54115 .22427 m
.55291 .22427 L
s
.55291 .22427 m
.54703 .21409 L
s
.54703 .21409 m
.55291 .20391 L
s
.55291 .20391 m
.56467 .20391 L
s
.56467 .20391 m
.57055 .21409 L
s
.57055 .21409 m
.57643 .20391 L
s
.57643 .20391 m
.58818 .20391 L
s
.58818 .20391 m
.59406 .21409 L
s
.59406 .21409 m
.58818 .22427 L
s
.58818 .22427 m
.59994 .22427 L
s
.59994 .22427 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .22427 L
s
.6117 .22427 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .21409 L
s
.61758 .21409 m
.62346 .20391 L
s
.62346 .20391 m
.63521 .20391 L
s
.63521 .20391 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .20391 L
s
.64697 .20391 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .19372 L
s
.65285 .19372 m
.65873 .18354 L
s
.65873 .18354 m
.64697 .18354 L
s
.64697 .18354 m
.64109 .17336 L
s
.64109 .17336 m
.64697 .16318 L
s
.64697 .16318 m
.65873 .16318 L
s
.65873 .16318 m
.65285 .15299 L
s
.65285 .15299 m
.65873 .14281 L
s
.65873 .14281 m
.64697 .14281 L
s
.64697 .14281 m
.64109 .13263 L
s
.64109 .13263 m
.63521 .14281 L
s
.63521 .14281 m
.62346 .14281 L
s
.62346 .14281 m
.61758 .13263 L
s
.61758 .13263 m
.62346 .12245 L
s
.62346 .12245 m
.6117 .12245 L
s
.6117 .12245 m
.60582 .11226 L
s
.60582 .11226 m
.6117 .10208 L
s
.6117 .10208 m
.62346 .10208 L
s
.62346 .10208 m
.61758 .0919 L
s
.61758 .0919 m
.62346 .08171 L
s
.62346 .08171 m
.63521 .08171 L
s
.63521 .08171 m
.64109 .0919 L
s
.64109 .0919 m
.64697 .08171 L
s
.64697 .08171 m
.65873 .08171 L
s
.65873 .08171 m
.65285 .07153 L
s
.65285 .07153 m
.65873 .06135 L
s
.65873 .06135 m
.64697 .06135 L
s
.64697 .06135 m
.64109 .05117 L
s
.64109 .05117 m
.64697 .04098 L
s
.64697 .04098 m
.65873 .04098 L
s
.65873 .04098 m
.65285 .0308 L
s
.65285 .0308 m
.65873 .02062 L
s
.65873 .02062 m
.67049 .02062 L
s
.67049 .02062 m
.67637 .0308 L
s
.67637 .0308 m
.68225 .02062 L
s
.68225 .02062 m
.694 .02062 L
s
.694 .02062 m
.69988 .0308 L
s
.69988 .0308 m
.694 .04098 L
s
.694 .04098 m
.70576 .04098 L
s
.70576 .04098 m
.71164 .05117 L
s
.71164 .05117 m
.71752 .04098 L
s
.71752 .04098 m
.72928 .04098 L
s
.72928 .04098 m
.7234 .0308 L
s
.7234 .0308 m
.72928 .02062 L
s
.72928 .02062 m
.74103 .02062 L
s
.74103 .02062 m
.74691 .0308 L
s
.74691 .0308 m
.75279 .02062 L
s
.75279 .02062 m
.76455 .02062 L
s
.76455 .02062 m
.77043 .0308 L
s
.77043 .0308 m
.76455 .04098 L
s
.76455 .04098 m
.77631 .04098 L
s
.77631 .04098 m
.78219 .05117 L
s
.78219 .05117 m
.77631 .06135 L
s
.77631 .06135 m
.76455 .06135 L
s
.76455 .06135 m
.77043 .07153 L
s
.77043 .07153 m
.76455 .08171 L
s
.76455 .08171 m
.77631 .08171 L
s
.77631 .08171 m
.78219 .0919 L
s
.78219 .0919 m
.78807 .08171 L
s
.78807 .08171 m
.79982 .08171 L
s
.79982 .08171 m
.8057 .0919 L
s
.8057 .0919 m
.79982 .10208 L
s
.79982 .10208 m
.81158 .10208 L
s
.81158 .10208 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .10208 L
s
.82334 .10208 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .0919 L
s
.82922 .0919 m
.8351 .08171 L
s
.8351 .08171 m
.84685 .08171 L
s
.84685 .08171 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .08171 L
s
.85861 .08171 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .07153 L
s
.86449 .07153 m
.87037 .06135 L
s
.87037 .06135 m
.85861 .06135 L
s
.85861 .06135 m
.85273 .05117 L
s
.85273 .05117 m
.85861 .04098 L
s
.85861 .04098 m
.87037 .04098 L
s
.87037 .04098 m
.86449 .0308 L
s
.86449 .0308 m
.87037 .02062 L
s
.87037 .02062 m
.88213 .02062 L
s
.88213 .02062 m
.88801 .0308 L
s
.88801 .0308 m
.89389 .02062 L
s
.89389 .02062 m
.90564 .02062 L
s
.90564 .02062 m
.91152 .0308 L
s
.91152 .0308 m
.90564 .04098 L
s
.90564 .04098 m
.9174 .04098 L
s
.9174 .04098 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .04098 L
s
.92916 .04098 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .0308 L
s
.93504 .0308 m
.94092 .02062 L
s
.94092 .02062 m
.95267 .02062 L
s
.95267 .02062 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .02062 L
s
.96443 .02062 m
.97619 .02062 L
s
.97619 .02062 m
.97031 .0308 L
s
.97031 .0308 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .04098 L
s
.96443 .04098 m
.95855 .05117 L
s
.95855 .05117 m
.9468 .05117 L
s
.9468 .05117 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .05117 L
s
.93504 .05117 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .06135 L
s
.92916 .06135 m
.92328 .07153 L
s
.92328 .07153 m
.93504 .07153 L
s
.93504 .07153 m
.94092 .08171 L
s
.94092 .08171 m
.93504 .0919 L
s
.93504 .0919 m
.92328 .0919 L
s
.92328 .0919 m
.92916 .10208 L
s
.92916 .10208 m
.92328 .11226 L
s
.92328 .11226 m
.91152 .11226 L
s
.91152 .11226 m
.90564 .10208 L
s
.90564 .10208 m
.89976 .11226 L
s
.89976 .11226 m
.88801 .11226 L
s
.88801 .11226 m
.88213 .10208 L
s
.88213 .10208 m
.88801 .0919 L
s
.88801 .0919 m
.87625 .0919 L
s
.87625 .0919 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .0919 L
s
.86449 .0919 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .10208 L
s
.85861 .10208 m
.85273 .11226 L
s
.85273 .11226 m
.84098 .11226 L
s
.84098 .11226 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .11226 L
s
.82922 .11226 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .12245 L
s
.82334 .12245 m
.81746 .13263 L
s
.81746 .13263 m
.82922 .13263 L
s
.82922 .13263 m
.8351 .14281 L
s
.8351 .14281 m
.82922 .15299 L
s
.82922 .15299 m
.81746 .15299 L
s
.81746 .15299 m
.82334 .16318 L
s
.82334 .16318 m
.81746 .17336 L
s
.81746 .17336 m
.82922 .17336 L
s
.82922 .17336 m
.8351 .18354 L
s
.8351 .18354 m
.84098 .17336 L
s
.84098 .17336 m
.85273 .17336 L
s
.85273 .17336 m
.85861 .18354 L
s
.85861 .18354 m
.85273 .19372 L
s
.85273 .19372 m
.86449 .19372 L
s
.86449 .19372 m
.87037 .20391 L
s
.87037 .20391 m
.86449 .21409 L
s
.86449 .21409 m
.85273 .21409 L
s
.85273 .21409 m
.85861 .22427 L
s
.85861 .22427 m
.85273 .23445 L
s
.85273 .23445 m
.84098 .23445 L
s
.84098 .23445 m
.8351 .22427 L
s
.8351 .22427 m
.82922 .23445 L
s
.82922 .23445 m
.81746 .23445 L
s
.81746 .23445 m
.82334 .24464 L
s
.82334 .24464 m
.81746 .25482 L
s
.81746 .25482 m
.82922 .25482 L
s
.82922 .25482 m
.8351 .265 L
s
.8351 .265 m
.82922 .27518 L
s
.82922 .27518 m
.81746 .27518 L
s
.81746 .27518 m
.82334 .28537 L
s
.82334 .28537 m
.81746 .29555 L
s
.81746 .29555 m
.8057 .29555 L
s
.8057 .29555 m
.79982 .28537 L
s
.79982 .28537 m
.79394 .29555 L
s
.79394 .29555 m
.78219 .29555 L
s
.78219 .29555 m
.77631 .28537 L
s
.77631 .28537 m
.78219 .27518 L
s
.78219 .27518 m
.77043 .27518 L
s
.77043 .27518 m
.76455 .265 L
s
.76455 .265 m
.75867 .27518 L
s
.75867 .27518 m
.74691 .27518 L
s
.74691 .27518 m
.75279 .28537 L
s
.75279 .28537 m
.74691 .29555 L
s
.74691 .29555 m
.73516 .29555 L
s
.73516 .29555 m
.72928 .28537 L
s
.72928 .28537 m
.7234 .29555 L
s
.7234 .29555 m
.71164 .29555 L
s
.71164 .29555 m
.70576 .28537 L
s
.70576 .28537 m
.71164 .27518 L
s
.71164 .27518 m
.69988 .27518 L
s
.69988 .27518 m
.694 .265 L
s
.694 .265 m
.69988 .25482 L
s
.69988 .25482 m
.71164 .25482 L
s
.71164 .25482 m
.70576 .24464 L
s
.70576 .24464 m
.71164 .23445 L
s
.71164 .23445 m
.69988 .23445 L
s
.69988 .23445 m
.694 .22427 L
s
.694 .22427 m
.68812 .23445 L
s
.68812 .23445 m
.67637 .23445 L
s
.67637 .23445 m
.67049 .22427 L
s
.67049 .22427 m
.67637 .21409 L
s
.67637 .21409 m
.66461 .21409 L
s
.66461 .21409 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .21409 L
s
.65285 .21409 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .22427 L
s
.64697 .22427 m
.64109 .23445 L
s
.64109 .23445 m
.62934 .23445 L
s
.62934 .23445 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .23445 L
s
.61758 .23445 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .24464 L
s
.6117 .24464 m
.60582 .25482 L
s
.60582 .25482 m
.61758 .25482 L
s
.61758 .25482 m
.62346 .265 L
s
.62346 .265 m
.61758 .27518 L
s
.61758 .27518 m
.60582 .27518 L
s
.60582 .27518 m
.6117 .28537 L
s
.6117 .28537 m
.60582 .29555 L
s
.60582 .29555 m
.59406 .29555 L
s
.59406 .29555 m
.58818 .28537 L
s
.58818 .28537 m
.5823 .29555 L
s
.5823 .29555 m
.57055 .29555 L
s
.57055 .29555 m
.56467 .28537 L
s
.56467 .28537 m
.57055 .27518 L
s
.57055 .27518 m
.55879 .27518 L
s
.55879 .27518 m
.55291 .265 L
s
.55291 .265 m
.54703 .27518 L
s
.54703 .27518 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .28537 L
s
.54115 .28537 m
.53527 .29555 L
s
.53527 .29555 m
.52352 .29555 L
s
.52352 .29555 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .29555 L
s
.51176 .29555 m
.5 .29555 L
s
.5 .29555 m
.50588 .30573 L
s
.50588 .30573 m
.5 .31591 L
s
.5 .31591 m
.51176 .31591 L
s
.51176 .31591 m
.51764 .3261 L
s
.51764 .3261 m
.51176 .33628 L
s
.51176 .33628 m
.5 .33628 L
s
.5 .33628 m
.50588 .34646 L
s
.50588 .34646 m
.5 .35664 L
s
.5 .35664 m
.51176 .35664 L
s
.51176 .35664 m
.51764 .36683 L
s
.51764 .36683 m
.52352 .35664 L
s
.52352 .35664 m
.53527 .35664 L
s
.53527 .35664 m
.54115 .36683 L
s
.54115 .36683 m
.53527 .37701 L
s
.53527 .37701 m
.54703 .37701 L
s
.54703 .37701 m
.55291 .38719 L
s
.55291 .38719 m
.54703 .39737 L
s
.54703 .39737 m
.53527 .39737 L
s
.53527 .39737 m
.54115 .40756 L
s
.54115 .40756 m
.53527 .41774 L
s
.53527 .41774 m
.52352 .41774 L
s
.52352 .41774 m
.51764 .40756 L
s
.51764 .40756 m
.51176 .41774 L
s
.51176 .41774 m
.5 .41774 L
s
.5 .41774 m
.50588 .42792 L
s
.50588 .42792 m
.5 .4381 L
s
.5 .4381 m
.51176 .4381 L
s
.51176 .4381 m
.51764 .44829 L
s
.51764 .44829 m
.51176 .45847 L
s
.51176 .45847 m
.5 .45847 L
s
.5 .45847 m
.50588 .46865 L
s
.50588 .46865 m
.5 .47883 L
s
.5 .47883 m
.51176 .47883 L
s
.51176 .47883 m
.51764 .48902 L
s
.51764 .48902 m
.52352 .47883 L
s
.52352 .47883 m
.53527 .47883 L
s
.53527 .47883 m
.54115 .48902 L
s
.54115 .48902 m
.53527 .4992 L
s
.53527 .4992 m
.54703 .4992 L
s
.54703 .4992 m
.55291 .50938 L
s
.55291 .50938 m
.55879 .4992 L
s
.55879 .4992 m
.57055 .4992 L
s
.57055 .4992 m
.56467 .48902 L
s
.56467 .48902 m
.57055 .47883 L
s
.57055 .47883 m
.5823 .47883 L
s
.5823 .47883 m
.58818 .48902 L
s
.58818 .48902 m
.59406 .47883 L
s
.59406 .47883 m
.60582 .47883 L
s
.60582 .47883 m
.6117 .48902 L
s
.6117 .48902 m
.60582 .4992 L
s
.60582 .4992 m
.61758 .4992 L
s
.61758 .4992 m
.62346 .50938 L
s
.62346 .50938 m
.61758 .51956 L
s
.61758 .51956 m
.60582 .51956 L
s
.60582 .51956 m
.6117 .52975 L
s
.6117 .52975 m
.60582 .53993 L
s
.60582 .53993 m
.61758 .53993 L
s
.61758 .53993 m
.62346 .55011 L
s
.62346 .55011 m
.62934 .53993 L
s
.62934 .53993 m
.64109 .53993 L
s
.64109 .53993 m
.64697 .55011 L
s
.64697 .55011 m
.64109 .56029 L
s
.64109 .56029 m
.65285 .56029 L
s
.65285 .56029 m
.65873 .57048 L
s
.65873 .57048 m
.65285 .58066 L
s
.65285 .58066 m
.64109 .58066 L
s
.64109 .58066 m
.64697 .59084 L
s
.64697 .59084 m
.64109 .60102 L
s
.64109 .60102 m
.62934 .60102 L
s
.62934 .60102 m
.62346 .59084 L
s
.62346 .59084 m
.61758 .60102 L
s
.61758 .60102 m
.60582 .60102 L
s
.60582 .60102 m
.6117 .61121 L
s
.6117 .61121 m
.60582 .62139 L
s
.60582 .62139 m
.61758 .62139 L
s
.61758 .62139 m
.62346 .63157 L
s
.62346 .63157 m
.61758 .64175 L
s
.61758 .64175 m
.60582 .64175 L
s
.60582 .64175 m
.6117 .65194 L
s
.6117 .65194 m
.60582 .66212 L
s
.60582 .66212 m
.59406 .66212 L
s
.59406 .66212 m
.58818 .65194 L
s
.58818 .65194 m
.5823 .66212 L
s
.5823 .66212 m
.57055 .66212 L
s
.57055 .66212 m
.56467 .65194 L
s
.56467 .65194 m
.57055 .64175 L
s
.57055 .64175 m
.55879 .64175 L
s
.55879 .64175 m
.55291 .63157 L
s
.55291 .63157 m
.54703 .64175 L
s
.54703 .64175 m
.53527 .64175 L
s
.53527 .64175 m
.54115 .65194 L
s
.54115 .65194 m
.53527 .66212 L
s
.53527 .66212 m
.52352 .66212 L
s
.52352 .66212 m
.51764 .65194 L
s
.51764 .65194 m
.51176 .66212 L
s
.51176 .66212 m
.5 .66212 L
s
.5 .66212 m
.50588 .6723 L
s
.50588 .6723 m
.5 .68249 L
s
.5 .68249 m
.51176 .68249 L
s
.51176 .68249 m
.51764 .69267 L
s
.51764 .69267 m
.51176 .70285 L
s
.51176 .70285 m
.5 .70285 L
s
.5 .70285 m
.50588 .71303 L
s
.50588 .71303 m
.5 .72322 L
s
.5 .72322 m
.51176 .72322 L
s
.51176 .72322 m
.51764 .7334 L
s
.51764 .7334 m
.52352 .72322 L
s
.52352 .72322 m
.53527 .72322 L
s
.53527 .72322 m
.54115 .7334 L
s
.54115 .7334 m
.53527 .74358 L
s
.53527 .74358 m
.54703 .74358 L
s
.54703 .74358 m
.55291 .75376 L
s
.55291 .75376 m
.54703 .76395 L
s
.54703 .76395 m
.53527 .76395 L
s
.53527 .76395 m
.54115 .77413 L
s
.54115 .77413 m
.53527 .78431 L
s
.53527 .78431 m
.52352 .78431 L
s
.52352 .78431 m
.51764 .77413 L
s
.51764 .77413 m
.51176 .78431 L
s
.51176 .78431 m
.5 .78431 L
s
.5 .78431 m
.50588 .79449 L
s
.50588 .79449 m
.5 .80468 L
s
.5 .80468 m
.51176 .80468 L
s
.51176 .80468 m
.51764 .81486 L
s
.51764 .81486 m
.51176 .82504 L
s
.51176 .82504 m
.5 .82504 L
s
.5 .82504 m
.50588 .83522 L
s
.50588 .83522 m
.5 .84541 L
s
.5 .84541 m
.49412 .83522 L
s
.49412 .83522 m
.5 .82504 L
s
.5 .82504 m
.48824 .82504 L
s
.48824 .82504 m
.48236 .81486 L
s
.48236 .81486 m
.48824 .80468 L
s
.48824 .80468 m
.5 .80468 L
s
.5 .80468 m
.49412 .79449 L
s
.49412 .79449 m
.5 .78431 L
s
.5 .78431 m
.48824 .78431 L
s
.48824 .78431 m
.48236 .77413 L
s
.48236 .77413 m
.47648 .78431 L
s
.47648 .78431 m
.46473 .78431 L
s
.46473 .78431 m
.45885 .77413 L
s
.45885 .77413 m
.46473 .76395 L
s
.46473 .76395 m
.45297 .76395 L
s
.45297 .76395 m
.44709 .75376 L
s
.44709 .75376 m
.45297 .74358 L
s
.45297 .74358 m
.46473 .74358 L
s
.46473 .74358 m
.45885 .7334 L
s
.45885 .7334 m
.46473 .72322 L
s
.46473 .72322 m
.47648 .72322 L
s
.47648 .72322 m
.48236 .7334 L
s
.48236 .7334 m
.48824 .72322 L
s
.48824 .72322 m
.5 .72322 L
s
.5 .72322 m
.49412 .71303 L
s
.49412 .71303 m
.5 .70285 L
s
.5 .70285 m
.48824 .70285 L
s
.48824 .70285 m
.48236 .69267 L
s
.48236 .69267 m
.48824 .68249 L
s
.48824 .68249 m
.5 .68249 L
s
.5 .68249 m
.49412 .6723 L
s
.49412 .6723 m
.5 .66212 L
s
.5 .66212 m
.48824 .66212 L
s
.48824 .66212 m
.48236 .65194 L
s
.48236 .65194 m
.47648 .66212 L
s
.47648 .66212 m
.46473 .66212 L
s
.46473 .66212 m
.45885 .65194 L
s
.45885 .65194 m
.46473 .64175 L
s
.46473 .64175 m
.45297 .64175 L
s
.45297 .64175 m
.44709 .63157 L
s
.44709 .63157 m
.44121 .64175 L
s
.44121 .64175 m
.42945 .64175 L
s
.42945 .64175 m
.43533 .65194 L
s
.43533 .65194 m
.42945 .66212 L
s
.42945 .66212 m
.4177 .66212 L
s
.4177 .66212 m
.41182 .65194 L
s
.41182 .65194 m
.40594 .66212 L
s
.40594 .66212 m
.39418 .66212 L
s
.39418 .66212 m
.3883 .65194 L
s
.3883 .65194 m
.39418 .64175 L
s
.39418 .64175 m
.38242 .64175 L
s
.38242 .64175 m
.37654 .63157 L
s
.37654 .63157 m
.38242 .62139 L
s
.38242 .62139 m
.39418 .62139 L
s
.39418 .62139 m
.3883 .61121 L
s
.3883 .61121 m
.39418 .60102 L
s
.39418 .60102 m
.38242 .60102 L
s
.38242 .60102 m
.37654 .59084 L
s
.37654 .59084 m
.37066 .60102 L
s
.37066 .60102 m
.35891 .60102 L
s
.35891 .60102 m
.35303 .59084 L
s
.35303 .59084 m
.35891 .58066 L
s
.35891 .58066 m
.34715 .58066 L
s
.34715 .58066 m
.34127 .57048 L
s
.34127 .57048 m
.34715 .56029 L
s
.34715 .56029 m
.35891 .56029 L
s
.35891 .56029 m
.35303 .55011 L
s
.35303 .55011 m
.35891 .53993 L
s
.35891 .53993 m
.37066 .53993 L
s
.37066 .53993 m
.37654 .55011 L
s
.37654 .55011 m
.38242 .53993 L
s
.38242 .53993 m
.39418 .53993 L
s
.39418 .53993 m
.3883 .52975 L
s
.3883 .52975 m
.39418 .51956 L
s
.39418 .51956 m
.38242 .51956 L
s
.38242 .51956 m
.37654 .50938 L
s
.37654 .50938 m
.38242 .4992 L
s
.38242 .4992 m
.39418 .4992 L
s
.39418 .4992 m
.3883 .48902 L
s
.3883 .48902 m
.39418 .47883 L
s
.39418 .47883 m
.40594 .47883 L
s
.40594 .47883 m
.41182 .48902 L
s
.41182 .48902 m
.4177 .47883 L
s
.4177 .47883 m
.42945 .47883 L
s
.42945 .47883 m
.43533 .48902 L
s
.43533 .48902 m
.42945 .4992 L
s
.42945 .4992 m
.44121 .4992 L
s
.44121 .4992 m
.44709 .50938 L
s
.44709 .50938 m
.45297 .4992 L
s
.45297 .4992 m
.46473 .4992 L
s
.46473 .4992 m
.45885 .48902 L
s
.45885 .48902 m
.46473 .47883 L
s
.46473 .47883 m
.47648 .47883 L
s
.47648 .47883 m
.48236 .48902 L
s
.48236 .48902 m
.48824 .47883 L
s
.48824 .47883 m
.5 .47883 L
s
.5 .47883 m
.49412 .46865 L
s
.49412 .46865 m
.5 .45847 L
s
.5 .45847 m
.48824 .45847 L
s
.48824 .45847 m
.48236 .44829 L
s
.48236 .44829 m
.48824 .4381 L
s
.48824 .4381 m
.5 .4381 L
s
.5 .4381 m
.49412 .42792 L
s
.49412 .42792 m
.5 .41774 L
s
.5 .41774 m
.48824 .41774 L
s
.48824 .41774 m
.48236 .40756 L
s
.48236 .40756 m
.47648 .41774 L
s
.47648 .41774 m
.46473 .41774 L
s
.46473 .41774 m
.45885 .40756 L
s
.45885 .40756 m
.46473 .39737 L
s
.46473 .39737 m
.45297 .39737 L
s
.45297 .39737 m
.44709 .38719 L
s
.44709 .38719 m
.45297 .37701 L
s
.45297 .37701 m
.46473 .37701 L
s
.46473 .37701 m
.45885 .36683 L
s
.45885 .36683 m
.46473 .35664 L
s
.46473 .35664 m
.47648 .35664 L
s
.47648 .35664 m
.48236 .36683 L
s
.48236 .36683 m
.48824 .35664 L
s
.48824 .35664 m
.5 .35664 L
s
.5 .35664 m
.49412 .34646 L
s
.49412 .34646 m
.5 .33628 L
s
.5 .33628 m
.48824 .33628 L
s
.48824 .33628 m
.48236 .3261 L
s
.48236 .3261 m
.48824 .31591 L
s
.48824 .31591 m
.5 .31591 L
s
.5 .31591 m
.49412 .30573 L
s
.49412 .30573 m
.5 .29555 L
s
.5 .29555 m
.48824 .29555 L
s
.48824 .29555 m
.48236 .28537 L
s
.48236 .28537 m
.47648 .29555 L
s
.47648 .29555 m
.46473 .29555 L
s
.46473 .29555 m
.45885 .28537 L
s
.45885 .28537 m
.46473 .27518 L
s
.46473 .27518 m
.45297 .27518 L
s
.45297 .27518 m
.44709 .265 L
s
.44709 .265 m
.44121 .27518 L
s
.44121 .27518 m
.42945 .27518 L
s
.42945 .27518 m
.43533 .28537 L
s
.43533 .28537 m
.42945 .29555 L
s
.42945 .29555 m
.4177 .29555 L
s
.4177 .29555 m
.41182 .28537 L
s
.41182 .28537 m
.40594 .29555 L
s
.40594 .29555 m
.39418 .29555 L
s
.39418 .29555 m
.3883 .28537 L
s
.3883 .28537 m
.39418 .27518 L
s
.39418 .27518 m
.38242 .27518 L
s
.38242 .27518 m
.37654 .265 L
s
.37654 .265 m
.38242 .25482 L
s
.38242 .25482 m
.39418 .25482 L
s
.39418 .25482 m
.3883 .24464 L
s
.3883 .24464 m
.39418 .23445 L
s
.39418 .23445 m
.38242 .23445 L
s
.38242 .23445 m
.37654 .22427 L
s
.37654 .22427 m
.37066 .23445 L
s
.37066 .23445 m
.35891 .23445 L
s
.35891 .23445 m
.35303 .22427 L
s
.35303 .22427 m
.35891 .21409 L
s
.35891 .21409 m
.34715 .21409 L
s
.34715 .21409 m
.34127 .20391 L
s
.34127 .20391 m
.33539 .21409 L
s
.33539 .21409 m
.32363 .21409 L
s
.32363 .21409 m
.32951 .22427 L
s
.32951 .22427 m
.32363 .23445 L
s
.32363 .23445 m
.31188 .23445 L
s
.31188 .23445 m
.306 .22427 L
s
.306 .22427 m
.30012 .23445 L
s
.30012 .23445 m
.28836 .23445 L
s
.28836 .23445 m
.29424 .24464 L
s
.29424 .24464 m
.28836 .25482 L
s
.28836 .25482 m
.30012 .25482 L
s
.30012 .25482 m
.306 .265 L
s
.306 .265 m
.30012 .27518 L
s
.30012 .27518 m
.28836 .27518 L
s
.28836 .27518 m
.29424 .28537 L
s
.29424 .28537 m
.28836 .29555 L
s
.28836 .29555 m
.2766 .29555 L
s
.2766 .29555 m
.27072 .28537 L
s
.27072 .28537 m
.26484 .29555 L
s
.26484 .29555 m
.25309 .29555 L
s
.25309 .29555 m
.24721 .28537 L
s
.24721 .28537 m
.25309 .27518 L
s
.25309 .27518 m
.24133 .27518 L
s
.24133 .27518 m
.23545 .265 L
s
.23545 .265 m
.22957 .27518 L
s
.22957 .27518 m
.21781 .27518 L
s
.21781 .27518 m
.22369 .28537 L
s
.22369 .28537 m
.21781 .29555 L
s
.21781 .29555 m
.20606 .29555 L
s
.20606 .29555 m
.20018 .28537 L
s
.20018 .28537 m
.1943 .29555 L
s
.1943 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.17666 .28537 L
s
.17666 .28537 m
.18254 .27518 L
s
.18254 .27518 m
.17078 .27518 L
s
.17078 .27518 m
.1649 .265 L
s
.1649 .265 m
.17078 .25482 L
s
.17078 .25482 m
.18254 .25482 L
s
.18254 .25482 m
.17666 .24464 L
s
.17666 .24464 m
.18254 .23445 L
s
.18254 .23445 m
.17078 .23445 L
s
.17078 .23445 m
.1649 .22427 L
s
.1649 .22427 m
.15902 .23445 L
s
.15902 .23445 m
.14727 .23445 L
s
.14727 .23445 m
.14139 .22427 L
s
.14139 .22427 m
.14727 .21409 L
s
.14727 .21409 m
.13551 .21409 L
s
.13551 .21409 m
.12963 .20391 L
s
.12963 .20391 m
.13551 .19372 L
s
.13551 .19372 m
.14727 .19372 L
s
.14727 .19372 m
.14139 .18354 L
s
.14139 .18354 m
.14727 .17336 L
s
.14727 .17336 m
.15902 .17336 L
s
.15902 .17336 m
.1649 .18354 L
s
.1649 .18354 m
.17078 .17336 L
s
.17078 .17336 m
.18254 .17336 L
s
.18254 .17336 m
.17666 .16318 L
s
.17666 .16318 m
.18254 .15299 L
s
.18254 .15299 m
.17078 .15299 L
s
.17078 .15299 m
.1649 .14281 L
s
.1649 .14281 m
.17078 .13263 L
s
.17078 .13263 m
.18254 .13263 L
s
.18254 .13263 m
.17666 .12245 L
s
.17666 .12245 m
.18254 .11226 L
s
.18254 .11226 m
.17078 .11226 L
s
.17078 .11226 m
.1649 .10208 L
s
.1649 .10208 m
.15902 .11226 L
s
.15902 .11226 m
.14727 .11226 L
s
.14727 .11226 m
.14139 .10208 L
s
.14139 .10208 m
.14727 .0919 L
s
.14727 .0919 m
.13551 .0919 L
s
.13551 .0919 m
.12963 .08171 L
s
.12963 .08171 m
.12375 .0919 L
s
.12375 .0919 m
.11199 .0919 L
s
.11199 .0919 m
.11787 .10208 L
s
.11787 .10208 m
.11199 .11226 L
s
.11199 .11226 m
.10024 .11226 L
s
.10024 .11226 m
.09436 .10208 L
s
.09436 .10208 m
.08848 .11226 L
s
.08848 .11226 m
.07672 .11226 L
s
.07672 .11226 m
.07084 .10208 L
s
.07084 .10208 m
.07672 .0919 L
s
.07672 .0919 m
.06496 .0919 L
s
.06496 .0919 m
.05908 .08171 L
s
.05908 .08171 m
.06496 .07153 L
s
.06496 .07153 m
.07672 .07153 L
s
.07672 .07153 m
.07084 .06135 L
s
.07084 .06135 m
.07672 .05117 L
s
.07672 .05117 m
.06496 .05117 L
s
.06496 .05117 m
.05908 .04098 L
s
.05908 .04098 m
.0532 .05117 L
s
.0532 .05117 m
.04145 .05117 L
s
.04145 .05117 m
.03557 .04098 L
s
.03557 .04098 m
.04145 .0308 L
s
.04145 .0308 m
.02969 .0308 L
s
.02969 .0308 m
.02381 .02062 L
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 158.313},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000kB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K03oool00`000000oooo0?ooo`3o0?oood80oooo001[0?ooo`<00000o`3ooom30?ooo`00
J03oool4000000040?ooo`000000000000000?l0oooo@@3oool006P0oooo00<000000?ooo`000000
0P3oool200000?l0oooo@P3oool006D0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0o`3ooolk0?ooo`00I03oool200000080oooo0P00
00040?ooo`8000000P3oool200000?l0oooo?@3oool00640oooo2@0000050?ooo`P00000o`3ooolj
0?ooo`00HP3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`3o0?ooocT0oooo001R0?ooo`8000000`3oool00`000000
oooo0?ooo`060?ooo`8000000`3oool00`000000oooo0?ooo`3o0?ooocT0oooo001R0?ooo`800000
0P3oool2000000T0oooo00<000000?ooo`3oool00P00003o0?oooc/0oooo001Q0?ooo`@0000000<0
oooo0000003oool02P3oool00`000000oooo0000000200000?l0oooo>P3oool006@0oooo0`00000;
0?ooo`800000o`3ooolm0?ooo`00E`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`3o0?ooobh0oooo001F0?ooo`<0
00002`3oool2000000d0oooo00<000000?ooo`3oool02@3oool300000?l0oooo;`3oool005@0oooo
0`0000000`3oool0000000000002000000D0oooo1000000?0?ooo`<000001P3oool3000000030?oo
o`000000000000800000o`3oool/0?ooo`00E03oool2000000<0oooo0P0000070?ooo`800000403o
ool2000000L0oooo0P0000030?ooo`800000o`3oool]0?ooo`00D03oool00`000000oooo0?ooo`02
0?ooo`040000003oool0oooo0?ooo`8000000`3oool010000000oooo0?ooo`3oool200000140oooo
00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00D000000?ooo`3o
ool0oooo0000003o0?ooobT0oooo001?0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo008000000`3oool00`000000oooo0?ooo`0@0?ooo`80000000@0oooo0000003o
ool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00800000o`3ooolY0?oo
o`00C@3ooolF000001<0oooo5P00003o0?ooobD0oooo001=0?ooo`8000004P3oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3ooolT0?ooo`00CP3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`3o0?ooob@0oooo001=0?ooo`<00000403oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo00h0oooo0P00003o0?ooobH0oooo001=0?ooo`<000003`3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo00`0oooo1000003o0?ooobD0oooo001?0?ooo`80
00003P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00/0oooo0P00003o0?ooobP0
oooo001@0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo0?l0oooo9`3oool004l0oooo00<000000?ooo`000000303o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00X0oooo0P00003o0?ooobP0oooo001=
0?ooo`<0000000<0oooo0000003oool02P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`0000000`00003o0?ooobD0oooo001=0?ooo`8000000`3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool01`3oool01000
0000oooo0?ooo`3oool2000005X0oooo^`00000@0?ooo`00CP3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0F@3oool010000000oooo000000000002
0?ooo`800000103oool010000000oooo0?ooo`3oool200000080oooo0P00000B0?ooo`030000003o
ool0oooo008000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`800000
0P3oool2000003/0oooo00@000000?ooo`3oool0oooo0P0000020?ooo`800000103oool01P000000
oooo0?ooo`000000oooo00000080oooo0P00000A0?ooo`040000003oool0oooo0?ooo`8000000P3o
ool2000000@0oooo00@000000?ooo`3oool0oooo0P0000000`3oool000000?ooo`0@0?ooo`00C@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0F@3o
ool4000000<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`050000003o
ool0oooo0?ooo`0000004P3oool01@000000oooo0?ooo`3oool0000000@0oooo00D000000?ooo`3o
ool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000>`3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0000000@0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`00
0000103oool00`000000oooo0?ooo`020?ooo`<000004@3oool004d0oooo200000060?ooo`030000
003oool0oooo01l0oooo00<000000?ooo`3oool0103oool8000005/0oooo0P0000030?ooo`030000
003oool000000080oooo0P0000070?ooo`<00000403oool2000000L0oooo0P0000020?ooo`800000
1`3oool00`000000oooo0000000i0?ooo`8000001`3oool00`000000oooo000000020?ooo`800000
1`3oool3000000l0oooo0P0000070?ooo`030000003oool000000080oooo0P0000040?ooo`800000
4P3oool004l0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool08@3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00800000G`3oool2000000<0oooo1000
0000103oool000000000000000060?ooo`<000003`3oool4000000H0oooo0`000000103oool00000
0000000000060?ooo`@00000>03oool3000000H0oooo10000000103oool000000000000000060?oo
o`<000003`3oool3000000H0oooo10000000103oool000000000000000040?ooo`030000003oool0
oooo0100oooo001@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0Q0?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`1N
0?ooo`030000003oool0oooo008000000P3oool2000000/0oooo0P00000=0?ooo`8000002`3oool2
000000`0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0:0?ooo`8000002`3oool2
000000d0oooo00<000000?ooo`3oool02P3oool2000000030?ooo`000000000000<0oooo00<00000
0?ooo`3oool04@3oool005@0oooo0P0000030?ooo`030000003oool0oooo02<0oooo00<000000?oo
o`3oool00P00001T0?ooo`T00000303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool02P3oool2000003H0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool80000
01<0oooo001D0?ooo`<0000000<0oooo0000003oool09P3oool5000006@0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0000000`0oooo0P00000I0?ooo`80
0000=@3oool2000001T0oooo00<000000?ooo`0000002`3oool2000000d0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0B0?ooo`00EP3oool00`000000oooo0000000W0?ooo`800000
J03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo0`0000000`3oool00000
0?ooo`0:0?ooo`030000003oool0000000<000004`3oool4000000030?ooo`000000oooo0380oooo
0P000000103oool0000000000000000D0?ooo`<0000000<0oooo0000003oool02P3oool00`000000
oooo00000002000000/0oooo00@000000?ooo`3oool0oooo0P00000D0?ooo`00;P3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo02H0oooo0P00000X0?ooo`030000003oool0oooo03d0
oooo00@000000?ooo`3oool0oooo0P00000:0?ooo`8000000`3oool00`000000oooo0?ooo`070?oo
o`040000003oool0oooo0?ooo`800000503oool00`000000oooo000000020?ooo`030000003oool0
oooo0340oooo00<000000?ooo`3oool00P00000E0?ooo`8000000`3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo008000002`3oool2000000<0oooo00<000000?ooo`3oool04`3oool0
02d0oooo0`00000V0?ooo`800000:03oool2000002L0oooo0`00000o0?ooo`030000003oool00000
00<000002P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool02@3oool3000000030?ooo`000000oooo01D0oooo000Z0?ooo`@0000000@0oooo00000000
000000008P3oool3000002X0oooo1000000Q0?ooo`@0000000@0oooo0000000000000000?@3oool3
000000`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool04P3oool2000000@0oooo0P00000`0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
000001H0oooo000[0?ooo`8000000P3oool2000002<0oooo0`00000[0?ooo`8000008P3oool00`00
0000oooo000000020?ooo`800000?`3oool00`000000oooo0?ooo`0;0?ooo`P000001P3oool80000
01<0oooo2000000_0?ooo`P00000503oool8000000D0oooo2000000=0?ooo`030000003oool0oooo
01D0oooo000W0?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0Z0?oo
o`8000000`3oool00`000000oooo0?ooo`0I0?ooo`040000003oool0oooo0?ooo`8000000`3oool0
1@000000oooo0?ooo`3oool0000003/0oooo0P00000;0?ooo`80000000@0oooo0000003oool00000
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000006P3oool200000080oooo
0P00000]0?ooo`8000000P3oool2000001T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P000000103oool000000?ooo`00000;0?ooo`8000005`3oool002H0oooo
0P0000030?ooo`030000003oool0oooo0080oooo0P0000020?ooo`8000006@3oool2000000<0oooo
00<000000?ooo`3oool0;03oool2000000030?ooo`000000000001X0oooo0P0000020?ooo`800000
103oool2000000040?ooo`000000oooo000003/0oooo100000060?ooo`<0000000@0oooo00000000
00000000103oool010000000oooo0?ooo`3oool2000000<0oooo00<000000?ooo`3oool06@3oool0
10000000oooo0?ooo`3oool2000002`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000030?ooo`<00000
00<0oooo0000000000000P0000060?ooo`<00000603oool002@0oooo200000050?ooo`P00000503o
ool8000002l0oooo2000000C0?ooo`P000001P3oool8000003T0oooo0P0000070?ooo`8000000P3o
ool2000000@0oooo0P0000030?ooo`8000008P3oool00`000000oooo0000000[0?ooo`8000008`3o
ool2000000<0oooo0P0000030?ooo`8000000`3oool2000000L0oooo0`00000H0?ooo`00903oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0;P3oool2
000000@0oooo00<000000?ooo`3oool04P3oool2000000D0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo0P0000030?ooo`030000003oool0
oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`<0000000<0oooo0000000000000P00
000Q0?ooo`@00000:P3oool300000280oooo0`0000000`3oool0000000000002000000<0oooo00<0
00000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0
5`3oool002D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo02l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo03T0oooo00<000000?ooo`3oool00P0000030?ooo`030000003oool0oooo0080oooo00@0
00000?ooo`3oool0oooo0P000000103oool0000000000000000W0?ooo`030000003oool0oooo02L0
oooo0P00000V0?ooo`<0000000<0oooo0000000000000P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo008000000`3oool2000001T0oooo000T0?ooo`8000000`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo008000005@3oool2000000<0oooo00<000000?ooo`3oool0
<03oool200000080oooo0P00000D0?ooo`030000003oool000000080oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`02000003/0oooo5P00000X0?ooo`800000:03oool00`000000oooo
0?ooo`0V0?oooaH000006P3oool002@0oooo0`0000000`3oool000000?ooo`0:0?ooo`030000003o
ool0000000800000503oool3000000030?ooo`000000000003<0oooo00<000000?ooo`0000000`00
000C0?ooo`@0000000<0oooo0000003oool02P3oool5000003/0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0W0?ooo`8000009`3oool3000002L0oooo0P00000B0?ooo`030000003o
ool0oooo01T0oooo000V0?ooo`030000003oool0000000/0oooo0P00000I0?ooo`030000003oool0
000003D0oooo0P00000I0?ooo`800000303oool2000003h0oooo0P00000A0?ooo`030000003oool0
oooo02@0oooo1@00000W0?ooo`030000003oool00000008000009P3oool00`000000oooo0?ooo`0?
0?ooo`8000006`3oool001T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0d0?oo
o`800000303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo0P00000<0?oo
o`030000003oool0oooo0300oooo00<000000?ooo`3oool03P3oool2000002H0oooo00<000000?oo
o`0000000P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo008000009P3oool00`00
0000oooo0000000@0?ooo`030000003oool0oooo01X0oooo000H0?ooo`<000002`3oool2000000d0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0000000;0?ooo`800000=P3oool2000000/0
oooo0P00000<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool300000380
oooo00<000000?ooo`3oool03@3oool300000280oooo00@000000?ooo`3oool0oooo0P0000030?oo
o`030000003oool0oooo0240oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo01l0oooo1000000>0?ooo`030000003oool0oooo01/0oooo000F0?ooo`<0
000000<0oooo0000000000000P0000060?ooo`<000003`3oool3000000H0oooo0`0000000`3oool0
000000000002000000H0oooo0`00000h0?ooo`@000001P3oool3000000040?ooo`00000000000000
00H0oooo1000000>0?ooo`@000001P3oool3000000030?ooo`000000000000800000;`3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo0P0000020?ooo`800000103oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo0080oooo0P0000020?ooo`800000903oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo01/0oooo000F0?ooo`8000000`3oool2000000L0
oooo0`00000?0?ooo`8000001`3oool2000000<0oooo0P0000070?ooo`<00000>@3oool2000000L0
oooo0P0000020?ooo`8000001`3oool00`000000oooo0000000@0?ooo`8000001`3oool2000000<0
oooo0P00000a0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool07`3oool8000000H0
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`040?ooo`P000008@3oool2000000`0
oooo00<000000?ooo`3oool0703oool00180oooo00<000000?ooo`3oool00P3oool010000000oooo
0?ooo`3oool2000000<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo
0?ooo`000000103oool00`000000oooo0?ooo`0i0?ooo`050000003oool0oooo0?ooo`000000103o
ool01@000000oooo0?ooo`3oool0000000@0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo
0?ooo`000000;@3oool00`000000oooo0?ooo`0:0?ooo`8000008@3oool2000000D0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0P0?ooo`800000303oool00`000000oooo0?ooo`0L
0?ooo`004@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020000
00<0oooo0P00000A0?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo008000000`3oool2000003/0oooo00<000000?ooo`3oool00P000003
0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000020?ooo`8000004P3o
ool01P000000oooo0?ooo`000000oooo00000080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`02000002h0oooo00<000000?ooo`3oool02@3oool00`000000oooo00000002000001l0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01d0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01d0
oooo1@00000;0?ooo`030000003oool0oooo01d0oooo000?0?oook/00000:P3oool2000000T0oooo
00<000000?ooo`3oool00P00000O0?ooo`030000003oool000000080oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
0P00000O0?ooo`030000003oool000000080oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0N0?ooo`003`3oool200000;L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool07@3oool4000000030?ooo`00
0000oooo00T0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`090?ooo`030000003o
ool00000008000007`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool07P3oool00100oooo00<000000?ooo`3oool0]03oool2000002d0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0200oooo0P00000<
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02P3oool200000240oooo0P000004
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool07`3oool000l0oooo0`00002e0?oo
o`800000;@3oool00`000000oooo0?ooo`040?ooo`P000008P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0O0?oo
o`P000001P3oool00`000000oooo0?ooo`0O0?ooo`003`3oool300000;@0oooo1000000]0?ooo`03
0000003oool0oooo0080oooo0P0000020?ooo`800000903oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`80
00000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0200oooo000A0?ooo`800000
/P3oool200000300oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo01l0oooo1000000>0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
3@3oool300000280oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0200
oooo000B0?ooo`800000/@3oool200000340oooo00<000000?ooo`3oool00P00000V0?ooo`030000
003oool000000100oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000
9P3oool2000000<0oooo00<000000?ooo`3oool08@3oool00140oooo00<000000?ooo`000000/@3o
ool200000380oooo00<000000?ooo`0000000P00000V0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02@0oooo0`00
00000`3oool000000?ooo`0S0?ooo`003`3oool3000000030?ooo`00000000000:l0oooo00<00000
0?ooo`0000000`00000_0?ooo`<000009`3oool200000180oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`000000903oool0
00l0oooo0P0000030?ooo`030000003oool0oooo0:`0oooo0P0000020?ooo`800000<@3oool00`00
0000oooo0?ooo`0V0?oooaH000004`3ooolF000002P0oooo00<000000?ooo`3oool08`3oool00100
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2[0?ooo`040000003oool0oooo0?oo
o`800000<@3oool2000002H0oooo0`0000000`3oool00000000000020?ooo`800000103oool01P00
0000oooo0?ooo`000000oooo00000080oooo0P00000B0?ooo`030000003oool0oooo008000000`3o
ool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`80000000@0oooo000000000000
00009P3oool2000002D0oooo000?0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
ZP3oool2000000@0oooo00<000000?ooo`3oool0<03oool300000280oooo0`0000000`3oool00000
00000002000000<0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0
oooo00<000000?ooo`3oool03`3oool2000000<0oooo00@000000?ooo`3oool0oooo0P0000030?oo
o`030000003oool0oooo0080oooo0`0000000`3oool000000000000200000240oooo0`00000V0?oo
o`003`3oool800000:/0oooo2000000a0?ooo`8000008`3oool2000000<0oooo0P0000030?ooo`03
0000003oool000000080oooo0P0000070?ooo`<00000403oool2000000L0oooo0P0000020?ooo`80
0000103oool2000000<0oooo0P00000R0?ooo`<000009P3oool00140oooo0P0000030?ooo`800000
Z@3oool00`000000oooo0?ooo`02000003H0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000030?ooo`@0
000000@0oooo00000000000000001P3oool3000000l0oooo100000050?ooo`@0000000@0oooo0000
000000000000103oool010000000oooo0?ooo`3oool2000000<0oooo00<000000?ooo`3oool06@3o
ool01@000000oooo0?ooo`3oool0000002L0oooo000B0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0YP3oool2000000<0oooo00<000000?ooo`3oool0=03oool200000080oooo0P00
000I0?ooo`030000003oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0200000080oooo0P00000;0?ooo`8000003@3oool2000000/0oooo0P0000000`3oool000000000
00030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P00000J0?ooo`8000000P3o
ool00`000000oooo0?ooo`0V0?ooo`005P3oool300000:L0oooo0P00000k0?ooo`P00000503oool8
000000H0oooo2000000<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool8
000000H0oooo2000000C0?ooo`P00000:03oool001H0oooo0`00002V0?ooo`@00000>`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0000000;0?ooo`<00000303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0B0?ooo`800000103oool00`00
0000oooo0?ooo`0W0?ooo`00603oool200000:@0oooo0P00000n0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0
oooo00@000000?ooo`3oool0oooo0P00000:0?ooo`<0000000<0oooo0000003oool02P3oool00`00
0000oooo00000003000000X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`060?oo
o`040000003oool0oooo0?ooo`8000005@3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo02L0oooo000I0?ooo`800000X`3oool2000003l0oooo00<000000?ooo`3oool00P00000E0?oo
o`8000000`3oool00`000000oooo0?ooo`070?ooo`040000003oool0oooo0?ooo`8000002P3oool2
000000<0oooo00<000000?ooo`3oool01`3oool010000000oooo0?ooo`3oool2000000X0oooo0P00
00030?ooo`030000003oool0oooo00L0oooo00@000000?ooo`3oool0oooo0P00000D0?ooo`030000
003oool000000080oooo00<000000?ooo`3oool0:03oool001P0oooo00<000000?ooo`000000X`3o
ool2000003l0oooo0P000000103oool0000000000000000D0?ooo`<0000000<0oooo0000003oool0
2P3oool00`000000oooo00000003000000X0oooo00@000000?ooo`3oool0oooo0P0000080?ooo`04
0000003oool0oooo0?ooo`8000002P3oool3000000030?ooo`000000oooo00X0oooo00<000000?oo
o`0000000`00000C0?ooo`H00000:P3oool001H0oooo0`0000020?ooo`030000003oool0oooo09l0
oooo00<000000?ooo`0000000`00000m0?ooo`8000006@3oool00`000000oooo0000000<0?ooo`80
0000303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0000000`0oooo0P00000I0?ooo`80
0000:`3oool001H0oooo0P0000030?ooo`030000003oool0oooo09l0oooo00<000000?ooo`3oool0
0P00000o0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo200000050?ooo`T00000303oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool2000002/0oooo
000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`2M0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000000k0?ooo`030000
003oool0oooo00X0oooo0P00000;0?ooo`8000003@3oool2000000/0oooo0P000000103oool00000
0?ooo`0000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P0000020?ooo`80
00002`3oool2000000d0oooo0P00000;0?ooo`800000303oool00`000000oooo0?ooo`0Z0?ooo`00
4@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2M0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00P00000l0?ooo`<000001P3oool4000000040?oo
o`0000000000000000H0oooo0`00000?0?ooo`@000001@3oool4000000040?ooo`00000000000000
00@0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`@0000000@0oooo00000000000000001P3o
ool3000000l0oooo100000050?ooo`@0000000@0oooo00000000000000001P3oool4000002`0oooo
000?0?ooo`P000001P3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo00@0oooo2000
000h0?ooo`8000001`3oool00`000000oooo000000020?ooo`8000001`3oool300000100oooo0P00
00070?ooo`8000000P3oool2000000@0oooo0P0000030?ooo`8000000`3oool00`000000oooo0000
00020?ooo`8000001`3oool300000100oooo0P0000070?ooo`8000000P3oool2000000L0oooo00<0
00000?ooo`000000;@3oool000l0oooo0P0000040?ooo`030000003oool0oooo00D0oooo0P00002K
0?ooo`8000001`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool00P3o
ool01@000000oooo0?ooo`3oool000000140oooo0P0000030?ooo`040000003oool0oooo0?ooo`80
00000`3oool00`000000oooo0?ooo`020?ooo`<0000000<0oooo0000000000000P0000030?ooo`03
0000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo
00l0oooo0P0000030?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`02
0?ooo`050000003oool0oooo0?ooo`000000;P3oool00100oooo00D000000?ooo`3oool0oooo0000
00090?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3oool01P3oool01@000000oooo0?oo
o`3oool0000003/0oooo0P0000020?ooo`8000000P3oool2000000@0oooo00@000000?ooo`3oool0
oooo0P0000020?ooo`8000004P3oool00`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3o
ool00P3oool010000000oooo0?ooo`3oool2000000050?ooo`00000000000000003oool00P000002
0?ooo`800000103oool01P000000oooo0?ooo`000000oooo00000080oooo0P00000B0?ooo`030000
003oool0oooo008000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`80
00000P3oool2000002h0oooo000?0?ooo`030000003oool000000080oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
0P00000k0?ooogd00000;`3oool000l0oooo0`0000000`3oool000000?ooo`0:0?ooo`030000003o
ool0oooo09L0oooo00<000000?ooo`3oool0203oool00`000000oooo00000003000003/0oooo00<0
00000?ooo`3oool0>@3oool00`000000oooo0000000k0?ooo`030000003oool0oooo02h0oooo000B
0?ooo`800000303oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool0?@3oool00`000000oooo0?ooo`0i0?ooo`@00000>P3oool00`000000oooo0?ooo`0^
0?ooo`004@3oool2000000d0oooo0P00002E0?ooo`800000303oool2000003l0oooo00<000000?oo
o`3oool0=`3oool6000003P0oooo00<000000?ooo`3oool0;`3oool000l0oooo0`00000?0?ooo`03
0000003oool0oooo09<0oooo00<000000?ooo`3oool0303oool4000003d0oooo00<000000?ooo`3o
ool0=@3oool2000000<0oooo0P00000g0?ooo`030000003oool0oooo0300oooo000?0?ooo`<00000
3`3oool2000009<0oooo0P00000?0?ooo`800000?P3oool00`000000oooo0?ooo`0e0?ooo`030000
003oool0oooo00<0oooo0P00000f0?ooo`030000003oool0oooo0300oooo000@0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`0>0?ooo`800000?`3o
ool00`000000oooo0?ooo`0c0?ooo`8000001@3oool2000003D0oooo00<000000?ooo`3oool0<@3o
ool000l0oooo0P00000B0?ooo`030000003oool0oooo08l0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`0m0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`0000001`00000d
0?ooo`030000003oool0oooo0340oooo000?0?oooaH00000S`3ooolF000003l0oooo00<000000?oo
o`3oool0<@3oool00`000000oooo0?ooo`03000000050?ooo`000000oooo0?ooo`000000<`3oool0
0`000000oooo0?ooo`0b0?ooo`004@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`02000000<0oooo0P00002=0?ooo`80000000@0oooo0000003oool000000P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00800000@`3oool00`000000oooo0?ooo`0a
0?ooo`070000003oool0oooo0?ooo`000000oooo000000030?ooo`800000<P3oool00`000000oooo
0?ooo`0b0?ooo`004P3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`800000
0`3oool010000000oooo0?ooo`3oool2000008d0oooo0P0000020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000004@0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`020?ooo`<00000103oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo03<0oooo000F0?ooo`8000000`3oool2000000L0oooo0`00002;0?ooo`80
00001`3oool2000000<0oooo0P0000190?ooo`030000003oool0oooo02d0oooo3`00000_0?ooo`03
0000003oool0oooo03@0oooo000F0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<00000
RP3oool4000000H0oooo0`0000000`3oool0000000000002000004P0oooo00<000000?ooo`3oool0
;03oool200000080oooo0P0000050?ooo`8000000P3oool2000002h0oooo00<000000?ooo`3oool0
=03oool001P0oooo0`00000;0?ooo`800000R@3oool2000000X0oooo0`00001<0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`050000003o
ool0oooo0?ooo`000000;@3oool00`000000oooo0?ooo`0e0?ooo`006@3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo08H0oooo0P00000<0?ooo`030000003oool0oooo04/0oooo00<0
00000?ooo`3oool0:P3oool2000000h0oooo0`00000/0?ooo`030000003oool0oooo03D0oooo000V
0?ooo`<00000Q`3oool2000005/0oooo00<000000?ooo`3oool0:@3oool3000000d0oooo1000000Z
0?ooo`030000003oool0oooo03H0oooo000T0?ooo`<0000000<0oooo0000003oool0QP3oool50000
05P0oooo00<000000?ooo`3oool0:03oool200000140oooo0P00000Z0?ooo`030000003oool0oooo
03H0oooo000T0?ooo`8000000P3oool2000008D0oooo0P0000000`3oool000000000001J0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool04@3oool2000002P0oooo00<000000?ooo`3o
ool0=`3oool002D0oooo00@000000?ooo`3oool0oooo0P0000230?ooo`8000000`3oool00`000000
oooo0?ooo`1I0?ooo`030000003oool0oooo02D0oooo0`00000B0?ooo`8000009`3oool00`000000
oooo0?ooo`0h0?ooo`00903oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0840oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo02D0oooo
00<000000?ooo`0000000P00000=0?ooo`D000009`3oool00`000000oooo0?ooo`0h0?ooo`00903o
ool800000840oooo2000001K0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool00P00
000>0?ooo`030000003oool000000080oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`0i0?ooo`009P3oool2000000<0oooo00<000000?ooo`3oool0OP3oool2000000040?ooo`000000
oooo000005h0oooo00<000000?ooo`3oool08`3oool01@000000oooo0?ooo`3oool0000000@0oooo
00<000000?ooo`3oool0103oool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3o
ool08`3oool00`000000oooo0?ooo`0i0?ooo`009`3oool010000000oooo0?ooo`3oool2000007l0
oooo0P0000020?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`020?ooo`8000000P3oool2000000D0oooo0P0000020?ooo`800000103oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo03X0oooo000[0?ooo`800000OP3oool2000006@0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`030?ooo`l000001@3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo03X0oooo000Z0?ooo`@00000O@3oool3000006@0oooo00<0
00000?ooo`3oool07`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo0`00
00050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`0k0?ooo`00;@3oool00`000000oooo0?ooo`1j0?ooo`800000I`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo00000004
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0l
0?ooo`00;P3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo06H0oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0000000D0oooo0000
003oool0oooo000000090?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0?03oool0
02d0oooo0P00001k0?ooo`800000J03oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`0000001@0000000`3oool000000?ooo`090?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0?@3oool002X0oooo100000000`3oool000000?ooo`1h0?ooo`03
0000003oool0000000800000IP3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00T0
oooo0P0000050?ooo`8000002`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03d0
oooo000[0?ooo`8000000P3oool00`000000oooo0?ooo`1g0?ooo`040000003oool00000000006P0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03h0
oooo000W0?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`1e0?ooo`05
0000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0o0?ooo`009P3oool2000000<0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0080
oooo0P000000103oool000000?ooo`00001T0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool03@3oool6000000h0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0o0?oo
o`00903oool8000000D0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`030?ooo`P0
0000HP3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0`0000000`3oool0
00000?ooo`0?0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0@03oool002@0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0740oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0640oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool000000140oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`100?ooo`009@3oool010000000oooo0?oo
o`3oool2000000L0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`050?ooo`800000
0`3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
403oool300000180oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`110?ooo`00903o
ool2000000<0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`1_0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool00P00001T0?ooo`030000003oool0oooo01<0oooo:P00
000D0?ooo`030000003oool0oooo0440oooo000T0?ooo`<0000000<0oooo0000003oool02@3oool0
0`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00P0oooo1@00001T0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool00P0000030?ooo`030000003oool0oooo0080oooo00@00000
0?ooo`3oool0oooo0P0000050?ooo`8000000P3oool2000000@0oooo00H000000?ooo`3oool00000
0?ooo`0000020?ooo`8000004P3oool00`000000oooo0?ooo`120?ooo`009P3oool00`000000oooo
0000000;0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool02@3oool2000006P0oooo
00<000000?ooo`3oool03`3oool2000000<0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool00000
00@0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0@`3oool001T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo06/0oooo00<000000?ooo`3oool02P3oool2000000`0oooo00<000000?oo
o`3oool0F@3oool00`000000oooo0?ooo`0?0?ooo`8000001`3oool200000080oooo0P00000>0?oo
o`030000003oool000000080oooo0P0000070?ooo`<00000403oool00`000000oooo0?ooo`130?oo
o`00603oool3000000/0oooo0P00000=0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`090?ooo`<00000F`3oool00`000000oooo0?ooo`0=0?ooo`@0
00001@3oool4000000040?ooo`0000000000000000d0oooo10000000103oool00000000000000006
0?ooo`<000003`3oool00`000000oooo0?ooo`140?ooo`005P3oool3000000030?ooo`0000000000
008000001P3oool3000000l0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0<0?oo
o`@000001P3oool3000000030?ooo`000000000000800000F03oool00`000000oooo0?ooo`0<0?oo
o`8000002`3oool2000001<0oooo0P00000;0?ooo`8000003P3oool00`000000oooo0?ooo`140?oo
o`005P3oool2000000<0oooo0P0000070?ooo`<000003`3oool00`000000oooo0?ooo`1Y0?ooo`03
0000003oool0oooo00d0oooo0P0000070?ooo`8000000`3oool2000005X0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04D0oooo000B0?ooo`030000
003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo00l0
oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo0000001F0?ooo`030000003oool0oooo00X0oooo0`00000]0?ooo`030000003oool0
000000`0oooo00<000000?ooo`3oool0A@3oool00140oooo0P0000030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P0000030?ooo`8000004@3oool00`000000oooo0?ooo`1W0?oo
o`030000003oool0oooo00l0oooo0P000000103oool000000?ooo`0000020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P00001G0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`0000000`00000X0?ooo`<0000000<0oooo0000003oool02P3oool00`000000oooo0?ooo`16
0?ooo`003`3ooolF000001<0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0A0?oo
oaH00000E03oool00`000000oooo0?ooo`070?ooo`040000003oool0oooo0?ooo`800000:@3oool2
000000<0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`170?ooo`003`3oool20000
0180oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo
00L0oooo00@000000?ooo`3oool0oooo0P00000Z0?ooo`040000003oool0oooo0?ooo`800000203o
ool00`000000oooo0?ooo`170?ooo`00403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo00l0oooo0P00001F0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`180?ooo`003`3oool300000100oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`1Q0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool20000
05H0oooo00<000000?ooo`3oool01@3oool8000002P0oooo200000060?ooo`030000003oool0oooo
04P0oooo000?0?ooo`<000003`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0640
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0<0?ooo`@00000EP3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00800000;P3oool2000000<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`190?ooo`004@3oool2000000h0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
2`3oool2000005T0oooo00<000000?ooo`3oool00P3oool2000000<0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
B@3oool00180oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo05l0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo05T0oooo00@000000?ooo`3oool0oooo0P00000g0?ooo`8000000`3oool00`000000oooo
0?ooo`1:0?ooo`004@3oool00`000000oooo0000000<0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool0G@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00X0oooo0P00
001K0?ooo`030000003oool0000000<00000=P3oool3000000030?ooo`000000oooo04`0oooo000?
0?ooo`<0000000<0oooo0000003oool02P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo05/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
000000<00000F03oool3000003/0oooo00<000000?ooo`000000C@3oool000l0oooo0P0000030?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`1K0?oo
o`030000003oool0oooo01/0oooo00<000000?ooo`3oool01`3oool010000000oooo0?ooo`3oool2
000005X0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1<0?ooo`00403oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool07@3oool00`00
0000oooo0?ooo`1I0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool01P3oool01000
0000oooo0?ooo`3oool2000005X0oooo0P00000k0?ooo`800000CP3oool000l0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool0F@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`1I0?ooo`@00000=P3oool3000004l0oooo000?0?ooo`P0
00001P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3o
ool07`3oool00`000000oooo0?ooo`040?ooo`P00000F`3oool2000003L0oooo0P00001@0?ooo`00
4@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo05D0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00800000G`3oool2000000<0oooo00<000000?ooo`3oool0;@3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo04h0oooo000B0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo05D0oooo00<0
00000?ooo`3oool08@3oool010000000oooo0?ooo`3oool2000000<0oooo00<000000?ooo`3oool0
GP3oool00`000000oooo0?ooo`02000002h0oooo0P0000030?ooo`030000003oool0oooo04l0oooo
000F0?ooo`8000000`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo05<0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`02000006@0oooo2@00000X0?ooo`P00000D@3o
ool001H0oooo0`0000000`3oool000000?ooo`0U0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0903oool5000006@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0W
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0D03oool001P0oooo00<000000?oo
o`0000009`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo02D0oooo0P00001W0?oo
o`8000000`3oool2000002X0oooo00@000000?ooo`3oool0oooo0P00001B0?ooo`006@3oool00`00
0000oooo0?ooo`0V0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool09@3oool20000
06P0oooo00@000000?ooo`3oool0oooo0P00000Y0?ooo`8000000`3oool00`000000oooo0?ooo`1A
0?ooo`00603oool2000002T0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0V0?oo
o`800000J@3oool00`000000oooo00000003000002P0oooo0`0000000`3oool000000?ooo`1C0?oo
o`005P3oool3000002/0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0X0?ooo`@0
0000IP3oool3000002d0oooo00<000000?ooo`000000E03oool001H0oooo0`00000[0?ooo`030000
003oool0oooo04d0oooo00<000000?ooo`3oool0:@3oool2000006P0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
D`3oool00180oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo04/0oooo00<000000?ooo`3oool0:P3oool2000000<0oooo00<000000?ooo`3oool0HP3o
ool2000000/0oooo0P00000C0?ooo`8000002`3oool2000005D0oooo000A0?ooo`8000000`3oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0:`3oool2
000000030?ooo`000000000006D0oooo100000050?ooo`@0000000@0oooo00000000000000003@3o
ool4000000040?ooo`0000000000000000H0oooo0`00001F0?ooo`003`3oool8000002l0oooo00<0
00000?ooo`3oool0B@3oool00`000000oooo0?ooo`0]0?ooo`P00000H@3oool3000000L0oooo0P00
00020?ooo`8000003P3oool00`000000oooo000000020?ooo`8000001`3oool3000005H0oooo000?
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo02d0oooo0P0000040?ooo`030000003oool0oooo0640oooo0P000003
0?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00@0oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`00
0000103oool00`000000oooo0?ooo`1E0?ooo`00403oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0_0?ooo`040000
003oool0oooo0?ooo`800000I03oool00`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3o
ool00P3oool010000000oooo0?ooo`3oool2000000D0oooo0P0000020?ooo`800000103oool01P00
0000oooo0?ooo`000000oooo00000080oooo0P00001G0?ooo`003`3oool2000000<0oooo00<00000
0?ooo`3oool0<@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0300oooo0P000002
0?ooo`800000I03oool[000005P0oooo000?0?ooo`<0000000<0oooo000000000000<`3oool00`00
0000oooo0?ooo`150?ooo`030000003oool0oooo0340oooo00<000000?ooo`0000000`00001T0?oo
o`030000003oool0oooo0100oooo0`00000B0?ooo`030000003oool0oooo05L0oooo000A0?ooo`03
0000003oool0000003D0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0c0?ooo`80
0000I`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool000000180oooo00<000000?ooo`3o
ool0E`3oool00180oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`130?ooo`030000
003oool0oooo03<0oooo0P00001X0?ooo`030000003oool0oooo00h0oooo0`0000000`3oool00000
0?ooo`0?0?ooo`030000003oool0oooo05P0oooo000A0?ooo`800000=`3oool00`000000oooo0?oo
o`110?ooo`030000003oool0oooo03@0oooo0P00001Y0?ooo`030000003oool0oooo00d0oooo1P00
000>0?ooo`030000003oool0oooo05T0oooo000?0?ooo`<00000>03oool00`000000oooo0?ooo`11
0?ooo`030000003oool0oooo03D0oooo1000001V0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05T0oooo000?0?oo
o`<00000>@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo03L0oooo0P00001X0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo05X0oooo000@0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3o
ool0?@3oool00`000000oooo0?ooo`0h0?ooo`800000J03oool00`000000oooo0?ooo`0:0?ooo`80
00001@3oool2000000`0oooo00<000000?ooo`3oool0FP3oool000l0oooo0P00000k0?ooo`030000
003oool0oooo03d0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`1W0?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`0000001@0000000`3oool000000?ooo`090?ooo`030000
003oool0oooo05/0oooo000?0?ooocl00000?@3ooolo000006P0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`03000000050?ooo`000000oooo0?ooo`0000002@3oool2000005d0oooo
000A0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3o
ool200000140oooo00@000000?ooo`3oool0oooo0P0000020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P0000030?ooo`800000>`3oool2000000030?ooo`000000000000<0oooo
00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool200000080oooo0P00000B0?ooo`06
0000003oool0oooo0000003oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00800000K@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`000000103oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo05`0oooo000B0?oo
o`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`050000003oool0
oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0
>@3oool01@000000oooo0?ooo`3oool0000000@0oooo00D000000?ooo`3oool0oooo000000040?oo
o`050000003oool0oooo0?ooo`0000004P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<0
00000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000006h0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`020?ooo`<000001@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo05d0oooo000F0?ooo`8000000`3oool2000000L0oooo0`00000?0?ooo`8000001`3o
ool2000000<0oooo0P0000070?ooo`<00000>@3oool2000000L0oooo0P0000020?ooo`8000001`3o
ool00`000000oooo0000000@0?ooo`8000001`3oool2000000<0oooo0P00001b0?ooo`030000003o
ool0oooo00@0oooo3`0000060?ooo`030000003oool0oooo05d0oooo000F0?ooo`<0000000<0oooo
0000000000000P0000060?ooo`<000003`3oool3000000H0oooo0`0000000`3oool0000000000002
000000H0oooo0`00000h0?ooo`@000001P3oool3000000040?ooo`0000000000000000H0oooo1000
000>0?ooo`@000001P3oool3000000030?ooo`000000000000800000LP3oool00`000000oooo0?oo
o`020?ooo`8000000P3oool2000000D0oooo0P0000020?ooo`800000103oool00`000000oooo0?oo
o`1N0?ooo`00603oool3000000/0oooo0P00000=0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`0000002`3oool2000003H0oooo0P00000;0?ooo`800000303oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00T0oooo0`00001e0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo0?ooo`8000000`3o
ool00`000000oooo0?ooo`1N0?ooo`006@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo03@0oooo0P00000<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool2
000000`0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`02000000h0oooo00<00000
0?ooo`0000000P3oool00`000000oooo0?ooo`1O0?ooo`009P3oool00`000000oooo0000000;0?oo
o`8000006@3oool00`000000oooo0000000e0?ooo`8000006@3oool2000000`0oooo0P0000240?oo
o`80000000@0oooo00000000000000003@3oool600000640oooo000T0?ooo`<0000000<0oooo0000
003oool02P3oool00`000000oooo00000002000001@0oooo0`0000000`3oool000000000000c0?oo
o`H000004`3oool4000000030?ooo`000000oooo00X0oooo1@0000220?ooo`<000004P3oool20000
0680oooo000T0?ooo`8000000`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080
00005@3oool2000000<0oooo00<000000?ooo`3oool0<03oool200000080oooo0P00000D0?ooo`03
0000003oool000000080oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`02000008@0
oooo00<000000?ooo`3oool04@3oool200000680oooo000U0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`220?ooo`8000004P3oool00`000000oooo
0?ooo`1Q0?ooo`00903oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0;P3oool2000000@0oooo00<000000?ooo`3oool04P3oool2000000D0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo08<0oooo0`00
000=0?ooo`@00000H`3oool002@0oooo200000050?ooo`P00000503oool8000002l0oooo2000000C
0?ooo`P000001P3oool8000008@0oooo0P00000>0?ooo`030000003oool0000006@0oooo000V0?oo
o`8000000`3oool00`000000oooo0?ooo`020?ooo`8000000P3oool2000001T0oooo0P0000030?oo
o`030000003oool0oooo02`0oooo0P0000000`3oool000000000000J0?ooo`8000000P3oool20000
00@0oooo0P000000103oool000000?ooo`0000280?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool010000000oooo0?ooo`3oool2000006@0oooo000W0?ooo`040000003oool0
oooo0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0Z0?ooo`8000000`3oool00`000000oooo0?oo
o`0I0?ooo`040000003oool0oooo0?ooo`8000000`3oool01@000000oooo0?ooo`3oool0000008T0
oooo0P0000020?ooo`8000001@3oool200000080oooo0P00001U0?ooo`00:`3oool200000080oooo
0P00000S0?ooo`<00000:`3oool200000280oooo00<000000?ooo`0000000P3oool2000008h0oooo
3`00001V0?ooo`00:P3oool4000000040?ooo`000000000000000280oooo0`00000Z0?ooo`@00000
8@3oool4000000040?ooo`0000000000000008h0oooo00<000000?ooo`3oool00P3oool3000000@0
oooo0P00001V0?ooo`00;@3oool3000002H0oooo0P00000X0?ooo`8000009`3oool300000900oooo
00<000000?ooo`3oool00P3oool00`000000oooo000000040?ooo`030000003oool0oooo06D0oooo
000^0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool09P3oool2000002P0oooo00<0
00000?ooo`3oool0T03oool00`000000oooo0?ooo`03000000050?ooo`000000oooo0?ooo`000000
J03oool005H0oooo00<000000?ooo`0000009`3oool200000;/0oooo0P0000000`3oool000000000
0004000000030?ooo`000000oooo06L0oooo001D0?ooo`<0000000<0oooo0000003oool09P3oool5
00000;T0oooo0P0000050?ooo`800000J@3oool005@0oooo0P0000030?ooo`030000003oool0oooo
02<0oooo00<000000?ooo`3oool00P00002k0?ooo`030000003oool0oooo00<0oooo0P00001Y0?oo
o`00D03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool08@3oool010000000oooo0?ooo`3oool2000000<0oooo00<000000?ooo`3oool0]@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo06P0oooo001?0?ooo`8000000`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0200000;P0oooo1P00001Z0?ooo`00C@3oool8000000H0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`040?ooo`P00000]03oool3000000030?ooo`000000oooo06X0oooo
001=0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`2d0?ooo`030000003oool0000006`0oooo001>0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`2d0?ooo`<00000K03oool004d0oooo0P0000030?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0703oool2000000P0oooo00@000000?oo
o`3oool0oooo0P00002g0?ooo`030000003oool0oooo06/0oooo001=0?ooo`<0000000<0oooo0000
003oool02P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`0000000`00003o0?ooobD0oooo001?0?ooo`030000003oool0000000`0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0:0?ooo`800000o`3ooolX0?ooo`00D03oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`3o0?ooobL0oooo001?0?ooo`8000003P3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00/0oooo0P00003o0?ooobP0oooo001=0?ooo`<000003`3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo00`0oooo1000003o0?ooobD0oooo001=0?ooo`<00000403oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00003o0?ooobH0oooo001>0?ooo`03
0000003oool0oooo00l0oooo0P00000E0?ooo`8000004@3oool00`000000oooo0?ooo`3o0?ooob@0
oooo001=0?ooo`8000004P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0o`3ooolT0?ooo`00C@3ooolF000001<0oooo5P00003o0?ooobD0oooo001?
0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3oool0
0`000000oooo0?ooo`0@0?ooo`80000000@0oooo0000003oool000000P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00800000o`3ooolY0?ooo`00D03oool00`000000oooo0?ooo`02
0?ooo`040000003oool0oooo0?ooo`8000000`3oool010000000oooo0?ooo`3oool200000140oooo
00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00D000000?ooo`3o
ool0oooo0000003o0?ooobT0oooo001D0?ooo`8000000`3oool2000000L0oooo0P00000@0?ooo`80
00001`3oool2000000<0oooo0P00003o0?ooobd0oooo001D0?ooo`<0000000<0oooo000000000000
0P0000050?ooo`@000003`3oool3000000H0oooo0`0000000`3oool000000000000200000?l0oooo
;03oool005H0oooo0`00000;0?ooo`8000003@3oool2000000X0oooo0`00003o0?ooobl0oooo001G
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo0?l0oooo;P3oool006@0oooo0`00000;0?ooo`800000o`3ooolm0?oo
o`00H@3oool4000000030?ooo`000000oooo00X0oooo00<000000?ooo`0000000P00003o0?ooocX0
oooo001R0?ooo`8000000P3oool2000000T0oooo00<000000?ooo`3oool00P00003o0?oooc/0oooo
001R0?ooo`8000000`3oool00`000000oooo0?ooo`060?ooo`8000000`3oool00`000000oooo0?oo
o`3o0?ooocT0oooo001R0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0?l0oooo>@3oool00640oooo2@0000050?oo
o`P00000o`3ooolj0?ooo`00I03oool200000080oooo0P0000040?ooo`8000000P3oool200000?l0
oooo?@3oool006D0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0o`3ooolk0?ooo`00J03oool00`000000oooo000000020?ooo`800000
o`3ooom20?ooo`00J03oool4000000040?ooo`000000000000000?l0oooo@@3oool006/0oooo0`00
003o0?oood<0oooo001/0?ooo`030000003oool0oooo0?l0oooo@P3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {236.438, 0}} -> {-0.0524626, -0.0274986, \
0.00732573, 0.00732573}, {{10.6875, 205.75}, {230.75, 5.625}} -> {-0.0828296, \
-0.34818, 0.00807849, 0.00807849}, {{225.25, 420.25}, {230.75, 5.625}} -> \
{-1.2379, -0.203722, 0.008077, 0.008077}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
        GraphicsArray[{DisplayKoch[triangle, 4, 
              DisplayFunction \[Rule] Identity], 
            DisplayKoch[triangle, 4, KochType \[Rule] "\<sieve\>", 
              DisplayFunction \[Rule] Identity]}], 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .54986 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0130918 0.47619 [
[ 0 0 0 0 ]
[ 1 .54986 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .54986 L
0 .54986 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0130918 0.477324 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.302422 0.952381 [
[ 0 0 0 0 ]
[ 1 1.1547 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.1547 L
0 1.1547 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .30242 m
.02969 .3126 L
s
.02969 .3126 m
.02381 .32279 L
s
.02381 .32279 m
.03557 .32279 L
s
.03557 .32279 m
.04145 .33297 L
s
.04145 .33297 m
.03557 .34315 L
s
.03557 .34315 m
.02381 .34315 L
s
.02381 .34315 m
.02969 .35333 L
s
.02969 .35333 m
.02381 .36352 L
s
.02381 .36352 m
.03557 .36352 L
s
.03557 .36352 m
.04145 .3737 L
s
.04145 .3737 m
.04733 .36352 L
s
.04733 .36352 m
.05908 .36352 L
s
.05908 .36352 m
.06496 .3737 L
s
.06496 .3737 m
.05908 .38388 L
s
.05908 .38388 m
.07084 .38388 L
s
.07084 .38388 m
.07672 .39406 L
s
.07672 .39406 m
.07084 .40425 L
s
.07084 .40425 m
.05908 .40425 L
s
.05908 .40425 m
.06496 .41443 L
s
.06496 .41443 m
.05908 .42461 L
s
.05908 .42461 m
.04733 .42461 L
s
.04733 .42461 m
.04145 .41443 L
s
.04145 .41443 m
.03557 .42461 L
s
.03557 .42461 m
.02381 .42461 L
s
.02381 .42461 m
.02969 .43479 L
s
.02969 .43479 m
.02381 .44498 L
s
.02381 .44498 m
.03557 .44498 L
s
.03557 .44498 m
.04145 .45516 L
s
.04145 .45516 m
.03557 .46534 L
s
.03557 .46534 m
.02381 .46534 L
s
.02381 .46534 m
.02969 .47552 L
s
.02969 .47552 m
.02381 .48571 L
s
.02381 .48571 m
.03557 .48571 L
s
.03557 .48571 m
.04145 .49589 L
s
.04145 .49589 m
.04733 .48571 L
s
.04733 .48571 m
.05908 .48571 L
s
.05908 .48571 m
.06496 .49589 L
s
.06496 .49589 m
.05908 .50607 L
s
.05908 .50607 m
.07084 .50607 L
s
.07084 .50607 m
.07672 .51626 L
s
.07672 .51626 m
.0826 .50607 L
s
.0826 .50607 m
.09436 .50607 L
s
.09436 .50607 m
.08848 .49589 L
s
.08848 .49589 m
.09436 .48571 L
s
.09436 .48571 m
.10611 .48571 L
s
.10611 .48571 m
.11199 .49589 L
s
.11199 .49589 m
.11787 .48571 L
s
.11787 .48571 m
.12963 .48571 L
s
.12963 .48571 m
.13551 .49589 L
s
.13551 .49589 m
.12963 .50607 L
s
.12963 .50607 m
.14139 .50607 L
s
.14139 .50607 m
.14727 .51626 L
s
.14727 .51626 m
.14139 .52644 L
s
.14139 .52644 m
.12963 .52644 L
s
.12963 .52644 m
.13551 .53662 L
s
.13551 .53662 m
.12963 .5468 L
s
.12963 .5468 m
.14139 .5468 L
s
.14139 .5468 m
.14727 .55699 L
s
.14727 .55699 m
.15315 .5468 L
s
.15315 .5468 m
.1649 .5468 L
s
.1649 .5468 m
.17078 .55699 L
s
.17078 .55699 m
.1649 .56717 L
s
.1649 .56717 m
.17666 .56717 L
s
.17666 .56717 m
.18254 .57735 L
s
.18254 .57735 m
.17666 .58753 L
s
.17666 .58753 m
.1649 .58753 L
s
.1649 .58753 m
.17078 .59772 L
s
.17078 .59772 m
.1649 .6079 L
s
.1649 .6079 m
.15315 .6079 L
s
.15315 .6079 m
.14727 .59772 L
s
.14727 .59772 m
.14139 .6079 L
s
.14139 .6079 m
.12963 .6079 L
s
.12963 .6079 m
.13551 .61808 L
s
.13551 .61808 m
.12963 .62826 L
s
.12963 .62826 m
.14139 .62826 L
s
.14139 .62826 m
.14727 .63845 L
s
.14727 .63845 m
.14139 .64863 L
s
.14139 .64863 m
.12963 .64863 L
s
.12963 .64863 m
.13551 .65881 L
s
.13551 .65881 m
.12963 .66899 L
s
.12963 .66899 m
.11787 .66899 L
s
.11787 .66899 m
.11199 .65881 L
s
.11199 .65881 m
.10611 .66899 L
s
.10611 .66899 m
.09436 .66899 L
s
.09436 .66899 m
.08848 .65881 L
s
.08848 .65881 m
.09436 .64863 L
s
.09436 .64863 m
.0826 .64863 L
s
.0826 .64863 m
.07672 .63845 L
s
.07672 .63845 m
.07084 .64863 L
s
.07084 .64863 m
.05908 .64863 L
s
.05908 .64863 m
.06496 .65881 L
s
.06496 .65881 m
.05908 .66899 L
s
.05908 .66899 m
.04733 .66899 L
s
.04733 .66899 m
.04145 .65881 L
s
.04145 .65881 m
.03557 .66899 L
s
.03557 .66899 m
.02381 .66899 L
s
.02381 .66899 m
.02969 .67918 L
s
.02969 .67918 m
.02381 .68936 L
s
.02381 .68936 m
.03557 .68936 L
s
.03557 .68936 m
.04145 .69954 L
s
.04145 .69954 m
.03557 .70972 L
s
.03557 .70972 m
.02381 .70972 L
s
.02381 .70972 m
.02969 .71991 L
s
.02969 .71991 m
.02381 .73009 L
s
.02381 .73009 m
.03557 .73009 L
s
.03557 .73009 m
.04145 .74027 L
s
.04145 .74027 m
.04733 .73009 L
s
.04733 .73009 m
.05908 .73009 L
s
.05908 .73009 m
.06496 .74027 L
s
.06496 .74027 m
.05908 .75045 L
s
.05908 .75045 m
.07084 .75045 L
s
.07084 .75045 m
.07672 .76064 L
s
.07672 .76064 m
.07084 .77082 L
s
.07084 .77082 m
.05908 .77082 L
s
.05908 .77082 m
.06496 .781 L
s
.06496 .781 m
.05908 .79118 L
s
.05908 .79118 m
.04733 .79118 L
s
.04733 .79118 m
.04145 .781 L
s
.04145 .781 m
.03557 .79118 L
s
.03557 .79118 m
.02381 .79118 L
s
.02381 .79118 m
.02969 .80137 L
s
.02969 .80137 m
.02381 .81155 L
s
.02381 .81155 m
.03557 .81155 L
s
.03557 .81155 m
.04145 .82173 L
s
.04145 .82173 m
.03557 .83191 L
s
.03557 .83191 m
.02381 .83191 L
s
.02381 .83191 m
.02969 .8421 L
s
.02969 .8421 m
.02381 .85228 L
s
.02381 .85228 m
.03557 .85228 L
s
.03557 .85228 m
.04145 .86246 L
s
.04145 .86246 m
.04733 .85228 L
s
.04733 .85228 m
.05908 .85228 L
s
.05908 .85228 m
.06496 .86246 L
s
.06496 .86246 m
.05908 .87264 L
s
.05908 .87264 m
.07084 .87264 L
s
.07084 .87264 m
.07672 .88283 L
s
.07672 .88283 m
.0826 .87264 L
s
.0826 .87264 m
.09436 .87264 L
s
.09436 .87264 m
.08848 .86246 L
s
.08848 .86246 m
.09436 .85228 L
s
.09436 .85228 m
.10611 .85228 L
s
.10611 .85228 m
.11199 .86246 L
s
.11199 .86246 m
.11787 .85228 L
s
.11787 .85228 m
.12963 .85228 L
s
.12963 .85228 m
.13551 .86246 L
s
.13551 .86246 m
.12963 .87264 L
s
.12963 .87264 m
.14139 .87264 L
s
.14139 .87264 m
.14727 .88283 L
s
.14727 .88283 m
.14139 .89301 L
s
.14139 .89301 m
.12963 .89301 L
s
.12963 .89301 m
.13551 .90319 L
s
.13551 .90319 m
.12963 .91337 L
s
.12963 .91337 m
.14139 .91337 L
s
.14139 .91337 m
.14727 .92356 L
s
.14727 .92356 m
.15315 .91337 L
s
.15315 .91337 m
.1649 .91337 L
s
.1649 .91337 m
.17078 .92356 L
s
.17078 .92356 m
.1649 .93374 L
s
.1649 .93374 m
.17666 .93374 L
s
.17666 .93374 m
.18254 .94392 L
s
.18254 .94392 m
.18842 .93374 L
s
.18842 .93374 m
.20018 .93374 L
s
.20018 .93374 m
.1943 .92356 L
s
.1943 .92356 m
.20018 .91337 L
s
.20018 .91337 m
.21193 .91337 L
s
.21193 .91337 m
.21781 .92356 L
s
.21781 .92356 m
.22369 .91337 L
s
.22369 .91337 m
.23545 .91337 L
s
.23545 .91337 m
.22957 .90319 L
s
.22957 .90319 m
.23545 .89301 L
s
.23545 .89301 m
.22369 .89301 L
s
.22369 .89301 m
.21781 .88283 L
s
.21781 .88283 m
.22369 .87264 L
s
.22369 .87264 m
.23545 .87264 L
s
.23545 .87264 m
.22957 .86246 L
s
.22957 .86246 m
.23545 .85228 L
s
.23545 .85228 m
.24721 .85228 L
s
.24721 .85228 m
.25309 .86246 L
s
.25309 .86246 m
.25897 .85228 L
s
.25897 .85228 m
.27072 .85228 L
s
.27072 .85228 m
.2766 .86246 L
s
.2766 .86246 m
.27072 .87264 L
s
.27072 .87264 m
.28248 .87264 L
s
.28248 .87264 m
.28836 .88283 L
s
.28836 .88283 m
.29424 .87264 L
s
.29424 .87264 m
.306 .87264 L
s
.306 .87264 m
.30012 .86246 L
s
.30012 .86246 m
.306 .85228 L
s
.306 .85228 m
.31775 .85228 L
s
.31775 .85228 m
.32363 .86246 L
s
.32363 .86246 m
.32951 .85228 L
s
.32951 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.34715 .86246 L
s
.34715 .86246 m
.34127 .87264 L
s
.34127 .87264 m
.35303 .87264 L
s
.35303 .87264 m
.35891 .88283 L
s
.35891 .88283 m
.35303 .89301 L
s
.35303 .89301 m
.34127 .89301 L
s
.34127 .89301 m
.34715 .90319 L
s
.34715 .90319 m
.34127 .91337 L
s
.34127 .91337 m
.35303 .91337 L
s
.35303 .91337 m
.35891 .92356 L
s
.35891 .92356 m
.36479 .91337 L
s
.36479 .91337 m
.37654 .91337 L
s
.37654 .91337 m
.38242 .92356 L
s
.38242 .92356 m
.37654 .93374 L
s
.37654 .93374 m
.3883 .93374 L
s
.3883 .93374 m
.39418 .94392 L
s
.39418 .94392 m
.3883 .9541 L
s
.3883 .9541 m
.37654 .9541 L
s
.37654 .9541 m
.38242 .96429 L
s
.38242 .96429 m
.37654 .97447 L
s
.37654 .97447 m
.36479 .97447 L
s
.36479 .97447 m
.35891 .96429 L
s
.35891 .96429 m
.35303 .97447 L
s
.35303 .97447 m
.34127 .97447 L
s
.34127 .97447 m
.34715 .98465 L
s
.34715 .98465 m
.34127 .99483 L
s
.34127 .99483 m
.35303 .99483 L
s
.35303 .99483 m
.35891 1.00502 L
s
.35891 1.00502 m
.35303 1.0152 L
s
.35303 1.0152 m
.34127 1.0152 L
s
.34127 1.0152 m
.34715 1.02538 L
s
.34715 1.02538 m
.34127 1.03556 L
s
.34127 1.03556 m
.35303 1.03556 L
s
.35303 1.03556 m
.35891 1.04575 L
s
.35891 1.04575 m
.36479 1.03556 L
s
.36479 1.03556 m
.37654 1.03556 L
s
.37654 1.03556 m
.38242 1.04575 L
s
.38242 1.04575 m
.37654 1.05593 L
s
.37654 1.05593 m
.3883 1.05593 L
s
.3883 1.05593 m
.39418 1.06611 L
s
.39418 1.06611 m
.40006 1.05593 L
s
.40006 1.05593 m
.41182 1.05593 L
s
.41182 1.05593 m
.40594 1.04575 L
s
.40594 1.04575 m
.41182 1.03556 L
s
.41182 1.03556 m
.42357 1.03556 L
s
.42357 1.03556 m
.42945 1.04575 L
s
.42945 1.04575 m
.43533 1.03556 L
s
.43533 1.03556 m
.44709 1.03556 L
s
.44709 1.03556 m
.45297 1.04575 L
s
.45297 1.04575 m
.44709 1.05593 L
s
.44709 1.05593 m
.45885 1.05593 L
s
.45885 1.05593 m
.46473 1.06611 L
s
.46473 1.06611 m
.45885 1.07629 L
s
.45885 1.07629 m
.44709 1.07629 L
s
.44709 1.07629 m
.45297 1.08648 L
s
.45297 1.08648 m
.44709 1.09666 L
s
.44709 1.09666 m
.45885 1.09666 L
s
.45885 1.09666 m
.46473 1.10684 L
s
.46473 1.10684 m
.47061 1.09666 L
s
.47061 1.09666 m
.48236 1.09666 L
s
.48236 1.09666 m
.48824 1.10684 L
s
.48824 1.10684 m
.48236 1.11703 L
s
.48236 1.11703 m
.49412 1.11703 L
s
.49412 1.11703 m
.5 1.12721 L
s
.5 1.12721 m
.50588 1.11703 L
s
.50588 1.11703 m
.51764 1.11703 L
s
.51764 1.11703 m
.51176 1.10684 L
s
.51176 1.10684 m
.51764 1.09666 L
s
.51764 1.09666 m
.52939 1.09666 L
s
.52939 1.09666 m
.53527 1.10684 L
s
.53527 1.10684 m
.54115 1.09666 L
s
.54115 1.09666 m
.55291 1.09666 L
s
.55291 1.09666 m
.54703 1.08648 L
s
.54703 1.08648 m
.55291 1.07629 L
s
.55291 1.07629 m
.54115 1.07629 L
s
.54115 1.07629 m
.53527 1.06611 L
s
.53527 1.06611 m
.54115 1.05593 L
s
.54115 1.05593 m
.55291 1.05593 L
s
.55291 1.05593 m
.54703 1.04575 L
s
.54703 1.04575 m
.55291 1.03556 L
s
.55291 1.03556 m
.56467 1.03556 L
s
.56467 1.03556 m
.57055 1.04575 L
s
.57055 1.04575 m
.57643 1.03556 L
s
.57643 1.03556 m
.58818 1.03556 L
s
.58818 1.03556 m
.59406 1.04575 L
s
.59406 1.04575 m
.58818 1.05593 L
s
.58818 1.05593 m
.59994 1.05593 L
s
.59994 1.05593 m
.60582 1.06611 L
s
.60582 1.06611 m
.6117 1.05593 L
s
.6117 1.05593 m
.62346 1.05593 L
s
.62346 1.05593 m
.61758 1.04575 L
s
.61758 1.04575 m
.62346 1.03556 L
s
.62346 1.03556 m
.63521 1.03556 L
s
.63521 1.03556 m
.64109 1.04575 L
s
.64109 1.04575 m
.64697 1.03556 L
s
.64697 1.03556 m
.65873 1.03556 L
s
.65873 1.03556 m
.65285 1.02538 L
s
.65285 1.02538 m
.65873 1.0152 L
s
.65873 1.0152 m
.64697 1.0152 L
s
.64697 1.0152 m
.64109 1.00502 L
s
.64109 1.00502 m
.64697 .99483 L
s
.64697 .99483 m
.65873 .99483 L
s
.65873 .99483 m
.65285 .98465 L
s
.65285 .98465 m
.65873 .97447 L
s
.65873 .97447 m
.64697 .97447 L
s
.64697 .97447 m
.64109 .96429 L
s
.64109 .96429 m
.63521 .97447 L
s
.63521 .97447 m
.62346 .97447 L
s
.62346 .97447 m
.61758 .96429 L
s
.61758 .96429 m
.62346 .9541 L
s
.62346 .9541 m
.6117 .9541 L
s
.6117 .9541 m
.60582 .94392 L
s
.60582 .94392 m
.6117 .93374 L
s
.6117 .93374 m
.62346 .93374 L
s
.62346 .93374 m
.61758 .92356 L
s
.61758 .92356 m
.62346 .91337 L
s
.62346 .91337 m
.63521 .91337 L
s
.63521 .91337 m
.64109 .92356 L
s
.64109 .92356 m
.64697 .91337 L
s
.64697 .91337 m
.65873 .91337 L
s
.65873 .91337 m
.65285 .90319 L
s
.65285 .90319 m
.65873 .89301 L
s
.65873 .89301 m
.64697 .89301 L
s
.64697 .89301 m
.64109 .88283 L
s
.64109 .88283 m
.64697 .87264 L
s
.64697 .87264 m
.65873 .87264 L
s
.65873 .87264 m
.65285 .86246 L
s
.65285 .86246 m
.65873 .85228 L
s
.65873 .85228 m
.67049 .85228 L
s
.67049 .85228 m
.67637 .86246 L
s
.67637 .86246 m
.68225 .85228 L
s
.68225 .85228 m
.694 .85228 L
s
.694 .85228 m
.69988 .86246 L
s
.69988 .86246 m
.694 .87264 L
s
.694 .87264 m
.70576 .87264 L
s
.70576 .87264 m
.71164 .88283 L
s
.71164 .88283 m
.71752 .87264 L
s
.71752 .87264 m
.72928 .87264 L
s
.72928 .87264 m
.7234 .86246 L
s
.7234 .86246 m
.72928 .85228 L
s
.72928 .85228 m
.74103 .85228 L
s
.74103 .85228 m
.74691 .86246 L
s
.74691 .86246 m
.75279 .85228 L
s
.75279 .85228 m
.76455 .85228 L
s
.76455 .85228 m
.77043 .86246 L
s
.77043 .86246 m
.76455 .87264 L
s
.76455 .87264 m
.77631 .87264 L
s
.77631 .87264 m
.78219 .88283 L
s
.78219 .88283 m
.77631 .89301 L
s
.77631 .89301 m
.76455 .89301 L
s
.76455 .89301 m
.77043 .90319 L
s
.77043 .90319 m
.76455 .91337 L
s
.76455 .91337 m
.77631 .91337 L
s
.77631 .91337 m
.78219 .92356 L
s
.78219 .92356 m
.78807 .91337 L
s
.78807 .91337 m
.79982 .91337 L
s
.79982 .91337 m
.8057 .92356 L
s
.8057 .92356 m
.79982 .93374 L
s
.79982 .93374 m
.81158 .93374 L
s
.81158 .93374 m
.81746 .94392 L
s
.81746 .94392 m
.82334 .93374 L
s
.82334 .93374 m
.8351 .93374 L
s
.8351 .93374 m
.82922 .92356 L
s
.82922 .92356 m
.8351 .91337 L
s
.8351 .91337 m
.84685 .91337 L
s
.84685 .91337 m
.85273 .92356 L
s
.85273 .92356 m
.85861 .91337 L
s
.85861 .91337 m
.87037 .91337 L
s
.87037 .91337 m
.86449 .90319 L
s
.86449 .90319 m
.87037 .89301 L
s
.87037 .89301 m
.85861 .89301 L
s
.85861 .89301 m
.85273 .88283 L
s
.85273 .88283 m
.85861 .87264 L
s
.85861 .87264 m
.87037 .87264 L
s
.87037 .87264 m
.86449 .86246 L
s
.86449 .86246 m
.87037 .85228 L
s
.87037 .85228 m
.88213 .85228 L
s
.88213 .85228 m
.88801 .86246 L
s
.88801 .86246 m
.89389 .85228 L
s
.89389 .85228 m
.90564 .85228 L
s
.90564 .85228 m
.91152 .86246 L
s
.91152 .86246 m
.90564 .87264 L
s
.90564 .87264 m
.9174 .87264 L
s
.9174 .87264 m
.92328 .88283 L
s
.92328 .88283 m
.92916 .87264 L
s
.92916 .87264 m
.94092 .87264 L
s
.94092 .87264 m
.93504 .86246 L
s
.93504 .86246 m
.94092 .85228 L
s
.94092 .85228 m
.95267 .85228 L
s
.95267 .85228 m
.95855 .86246 L
s
.95855 .86246 m
.96443 .85228 L
s
.96443 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.97031 .8421 L
s
.97031 .8421 m
.97619 .83191 L
s
.97619 .83191 m
.96443 .83191 L
s
.96443 .83191 m
.95855 .82173 L
s
.95855 .82173 m
.96443 .81155 L
s
.96443 .81155 m
.97619 .81155 L
s
.97619 .81155 m
.97031 .80137 L
s
.97031 .80137 m
.97619 .79118 L
s
.97619 .79118 m
.96443 .79118 L
s
.96443 .79118 m
.95855 .781 L
s
.95855 .781 m
.95267 .79118 L
s
.95267 .79118 m
.94092 .79118 L
s
.94092 .79118 m
.93504 .781 L
s
.93504 .781 m
.94092 .77082 L
s
.94092 .77082 m
.92916 .77082 L
s
.92916 .77082 m
.92328 .76064 L
s
.92328 .76064 m
.92916 .75045 L
s
.92916 .75045 m
.94092 .75045 L
s
.94092 .75045 m
.93504 .74027 L
s
.93504 .74027 m
.94092 .73009 L
s
.94092 .73009 m
.95267 .73009 L
s
.95267 .73009 m
.95855 .74027 L
s
.95855 .74027 m
.96443 .73009 L
s
.96443 .73009 m
.97619 .73009 L
s
.97619 .73009 m
.97031 .71991 L
s
.97031 .71991 m
.97619 .70972 L
s
.97619 .70972 m
.96443 .70972 L
s
.96443 .70972 m
.95855 .69954 L
s
.95855 .69954 m
.96443 .68936 L
s
.96443 .68936 m
.97619 .68936 L
s
.97619 .68936 m
.97031 .67918 L
s
.97031 .67918 m
.97619 .66899 L
s
.97619 .66899 m
.96443 .66899 L
s
.96443 .66899 m
.95855 .65881 L
s
.95855 .65881 m
.95267 .66899 L
s
.95267 .66899 m
.94092 .66899 L
s
.94092 .66899 m
.93504 .65881 L
s
.93504 .65881 m
.94092 .64863 L
s
.94092 .64863 m
.92916 .64863 L
s
.92916 .64863 m
.92328 .63845 L
s
.92328 .63845 m
.9174 .64863 L
s
.9174 .64863 m
.90564 .64863 L
s
.90564 .64863 m
.91152 .65881 L
s
.91152 .65881 m
.90564 .66899 L
s
.90564 .66899 m
.89389 .66899 L
s
.89389 .66899 m
.88801 .65881 L
s
.88801 .65881 m
.88213 .66899 L
s
.88213 .66899 m
.87037 .66899 L
s
.87037 .66899 m
.86449 .65881 L
s
.86449 .65881 m
.87037 .64863 L
s
.87037 .64863 m
.85861 .64863 L
s
.85861 .64863 m
.85273 .63845 L
s
.85273 .63845 m
.85861 .62826 L
s
.85861 .62826 m
.87037 .62826 L
s
.87037 .62826 m
.86449 .61808 L
s
.86449 .61808 m
.87037 .6079 L
s
.87037 .6079 m
.85861 .6079 L
s
.85861 .6079 m
.85273 .59772 L
s
.85273 .59772 m
.84685 .6079 L
s
.84685 .6079 m
.8351 .6079 L
s
.8351 .6079 m
.82922 .59772 L
s
.82922 .59772 m
.8351 .58753 L
s
.8351 .58753 m
.82334 .58753 L
s
.82334 .58753 m
.81746 .57735 L
s
.81746 .57735 m
.82334 .56717 L
s
.82334 .56717 m
.8351 .56717 L
s
.8351 .56717 m
.82922 .55699 L
s
.82922 .55699 m
.8351 .5468 L
s
.8351 .5468 m
.84685 .5468 L
s
.84685 .5468 m
.85273 .55699 L
s
.85273 .55699 m
.85861 .5468 L
s
.85861 .5468 m
.87037 .5468 L
s
.87037 .5468 m
.86449 .53662 L
s
.86449 .53662 m
.87037 .52644 L
s
.87037 .52644 m
.85861 .52644 L
s
.85861 .52644 m
.85273 .51626 L
s
.85273 .51626 m
.85861 .50607 L
s
.85861 .50607 m
.87037 .50607 L
s
.87037 .50607 m
.86449 .49589 L
s
.86449 .49589 m
.87037 .48571 L
s
.87037 .48571 m
.88213 .48571 L
s
.88213 .48571 m
.88801 .49589 L
s
.88801 .49589 m
.89389 .48571 L
s
.89389 .48571 m
.90564 .48571 L
s
.90564 .48571 m
.91152 .49589 L
s
.91152 .49589 m
.90564 .50607 L
s
.90564 .50607 m
.9174 .50607 L
s
.9174 .50607 m
.92328 .51626 L
s
.92328 .51626 m
.92916 .50607 L
s
.92916 .50607 m
.94092 .50607 L
s
.94092 .50607 m
.93504 .49589 L
s
.93504 .49589 m
.94092 .48571 L
s
.94092 .48571 m
.95267 .48571 L
s
.95267 .48571 m
.95855 .49589 L
s
.95855 .49589 m
.96443 .48571 L
s
.96443 .48571 m
.97619 .48571 L
s
.97619 .48571 m
.97031 .47552 L
s
.97031 .47552 m
.97619 .46534 L
s
.97619 .46534 m
.96443 .46534 L
s
.96443 .46534 m
.95855 .45516 L
s
.95855 .45516 m
.96443 .44498 L
s
.96443 .44498 m
.97619 .44498 L
s
.97619 .44498 m
.97031 .43479 L
s
.97031 .43479 m
.97619 .42461 L
s
.97619 .42461 m
.96443 .42461 L
s
.96443 .42461 m
.95855 .41443 L
s
.95855 .41443 m
.95267 .42461 L
s
.95267 .42461 m
.94092 .42461 L
s
.94092 .42461 m
.93504 .41443 L
s
.93504 .41443 m
.94092 .40425 L
s
.94092 .40425 m
.92916 .40425 L
s
.92916 .40425 m
.92328 .39406 L
s
.92328 .39406 m
.92916 .38388 L
s
.92916 .38388 m
.94092 .38388 L
s
.94092 .38388 m
.93504 .3737 L
s
.93504 .3737 m
.94092 .36352 L
s
.94092 .36352 m
.95267 .36352 L
s
.95267 .36352 m
.95855 .3737 L
s
.95855 .3737 m
.96443 .36352 L
s
.96443 .36352 m
.97619 .36352 L
s
.97619 .36352 m
.97031 .35333 L
s
.97031 .35333 m
.97619 .34315 L
s
.97619 .34315 m
.96443 .34315 L
s
.96443 .34315 m
.95855 .33297 L
s
.95855 .33297 m
.96443 .32279 L
s
.96443 .32279 m
.97619 .32279 L
s
.97619 .32279 m
.97031 .3126 L
s
.97031 .3126 m
.97619 .30242 L
s
.97619 .30242 m
.96443 .30242 L
s
.96443 .30242 m
.95855 .29224 L
s
.95855 .29224 m
.95267 .30242 L
s
.95267 .30242 m
.94092 .30242 L
s
.94092 .30242 m
.93504 .29224 L
s
.93504 .29224 m
.94092 .28206 L
s
.94092 .28206 m
.92916 .28206 L
s
.92916 .28206 m
.92328 .27187 L
s
.92328 .27187 m
.9174 .28206 L
s
.9174 .28206 m
.90564 .28206 L
s
.90564 .28206 m
.91152 .29224 L
s
.91152 .29224 m
.90564 .30242 L
s
.90564 .30242 m
.89389 .30242 L
s
.89389 .30242 m
.88801 .29224 L
s
.88801 .29224 m
.88213 .30242 L
s
.88213 .30242 m
.87037 .30242 L
s
.87037 .30242 m
.86449 .29224 L
s
.86449 .29224 m
.87037 .28206 L
s
.87037 .28206 m
.85861 .28206 L
s
.85861 .28206 m
.85273 .27187 L
s
.85273 .27187 m
.85861 .26169 L
s
.85861 .26169 m
.87037 .26169 L
s
.87037 .26169 m
.86449 .25151 L
s
.86449 .25151 m
.87037 .24133 L
s
.87037 .24133 m
.85861 .24133 L
s
.85861 .24133 m
.85273 .23114 L
s
.85273 .23114 m
.84685 .24133 L
s
.84685 .24133 m
.8351 .24133 L
s
.8351 .24133 m
.82922 .23114 L
s
.82922 .23114 m
.8351 .22096 L
s
.8351 .22096 m
.82334 .22096 L
s
.82334 .22096 m
.81746 .21078 L
s
.81746 .21078 m
.81158 .22096 L
s
.81158 .22096 m
.79982 .22096 L
s
.79982 .22096 m
.8057 .23114 L
s
.8057 .23114 m
.79982 .24133 L
s
.79982 .24133 m
.78807 .24133 L
s
.78807 .24133 m
.78219 .23114 L
s
.78219 .23114 m
.77631 .24133 L
s
.77631 .24133 m
.76455 .24133 L
s
.76455 .24133 m
.77043 .25151 L
s
.77043 .25151 m
.76455 .26169 L
s
.76455 .26169 m
.77631 .26169 L
s
.77631 .26169 m
.78219 .27187 L
s
.78219 .27187 m
.77631 .28206 L
s
.77631 .28206 m
.76455 .28206 L
s
.76455 .28206 m
.77043 .29224 L
s
.77043 .29224 m
.76455 .30242 L
s
.76455 .30242 m
.75279 .30242 L
s
.75279 .30242 m
.74691 .29224 L
s
.74691 .29224 m
.74103 .30242 L
s
.74103 .30242 m
.72928 .30242 L
s
.72928 .30242 m
.7234 .29224 L
s
.7234 .29224 m
.72928 .28206 L
s
.72928 .28206 m
.71752 .28206 L
s
.71752 .28206 m
.71164 .27187 L
s
.71164 .27187 m
.70576 .28206 L
s
.70576 .28206 m
.694 .28206 L
s
.694 .28206 m
.69988 .29224 L
s
.69988 .29224 m
.694 .30242 L
s
.694 .30242 m
.68225 .30242 L
s
.68225 .30242 m
.67637 .29224 L
s
.67637 .29224 m
.67049 .30242 L
s
.67049 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.65285 .29224 L
s
.65285 .29224 m
.65873 .28206 L
s
.65873 .28206 m
.64697 .28206 L
s
.64697 .28206 m
.64109 .27187 L
s
.64109 .27187 m
.64697 .26169 L
s
.64697 .26169 m
.65873 .26169 L
s
.65873 .26169 m
.65285 .25151 L
s
.65285 .25151 m
.65873 .24133 L
s
.65873 .24133 m
.64697 .24133 L
s
.64697 .24133 m
.64109 .23114 L
s
.64109 .23114 m
.63521 .24133 L
s
.63521 .24133 m
.62346 .24133 L
s
.62346 .24133 m
.61758 .23114 L
s
.61758 .23114 m
.62346 .22096 L
s
.62346 .22096 m
.6117 .22096 L
s
.6117 .22096 m
.60582 .21078 L
s
.60582 .21078 m
.6117 .2006 L
s
.6117 .2006 m
.62346 .2006 L
s
.62346 .2006 m
.61758 .19041 L
s
.61758 .19041 m
.62346 .18023 L
s
.62346 .18023 m
.63521 .18023 L
s
.63521 .18023 m
.64109 .19041 L
s
.64109 .19041 m
.64697 .18023 L
s
.64697 .18023 m
.65873 .18023 L
s
.65873 .18023 m
.65285 .17005 L
s
.65285 .17005 m
.65873 .15987 L
s
.65873 .15987 m
.64697 .15987 L
s
.64697 .15987 m
.64109 .14968 L
s
.64109 .14968 m
.64697 .1395 L
s
.64697 .1395 m
.65873 .1395 L
s
.65873 .1395 m
.65285 .12932 L
s
.65285 .12932 m
.65873 .11914 L
s
.65873 .11914 m
.64697 .11914 L
s
.64697 .11914 m
.64109 .10895 L
s
.64109 .10895 m
.63521 .11914 L
s
.63521 .11914 m
.62346 .11914 L
s
.62346 .11914 m
.61758 .10895 L
s
.61758 .10895 m
.62346 .09877 L
s
.62346 .09877 m
.6117 .09877 L
s
.6117 .09877 m
.60582 .08859 L
s
.60582 .08859 m
.59994 .09877 L
s
.59994 .09877 m
.58818 .09877 L
s
.58818 .09877 m
.59406 .10895 L
s
.59406 .10895 m
.58818 .11914 L
s
.58818 .11914 m
.57643 .11914 L
s
.57643 .11914 m
.57055 .10895 L
s
.57055 .10895 m
.56467 .11914 L
s
.56467 .11914 m
.55291 .11914 L
s
.55291 .11914 m
.54703 .10895 L
s
.54703 .10895 m
.55291 .09877 L
s
.55291 .09877 m
.54115 .09877 L
s
.54115 .09877 m
.53527 .08859 L
s
.53527 .08859 m
.54115 .07841 L
s
.54115 .07841 m
.55291 .07841 L
s
.55291 .07841 m
.54703 .06822 L
s
.54703 .06822 m
.55291 .05804 L
s
.55291 .05804 m
.54115 .05804 L
s
.54115 .05804 m
.53527 .04786 L
s
.53527 .04786 m
.52939 .05804 L
s
.52939 .05804 m
.51764 .05804 L
s
.51764 .05804 m
.51176 .04786 L
s
.51176 .04786 m
.51764 .03768 L
s
.51764 .03768 m
.50588 .03768 L
s
.50588 .03768 m
.5 .02749 L
s
.5 .02749 m
.49412 .03768 L
s
.49412 .03768 m
.48236 .03768 L
s
.48236 .03768 m
.48824 .04786 L
s
.48824 .04786 m
.48236 .05804 L
s
.48236 .05804 m
.47061 .05804 L
s
.47061 .05804 m
.46473 .04786 L
s
.46473 .04786 m
.45885 .05804 L
s
.45885 .05804 m
.44709 .05804 L
s
.44709 .05804 m
.45297 .06822 L
s
.45297 .06822 m
.44709 .07841 L
s
.44709 .07841 m
.45885 .07841 L
s
.45885 .07841 m
.46473 .08859 L
s
.46473 .08859 m
.45885 .09877 L
s
.45885 .09877 m
.44709 .09877 L
s
.44709 .09877 m
.45297 .10895 L
s
.45297 .10895 m
.44709 .11914 L
s
.44709 .11914 m
.43533 .11914 L
s
.43533 .11914 m
.42945 .10895 L
s
.42945 .10895 m
.42357 .11914 L
s
.42357 .11914 m
.41182 .11914 L
s
.41182 .11914 m
.40594 .10895 L
s
.40594 .10895 m
.41182 .09877 L
s
.41182 .09877 m
.40006 .09877 L
s
.40006 .09877 m
.39418 .08859 L
s
.39418 .08859 m
.3883 .09877 L
s
.3883 .09877 m
.37654 .09877 L
s
.37654 .09877 m
.38242 .10895 L
s
.38242 .10895 m
.37654 .11914 L
s
.37654 .11914 m
.36479 .11914 L
s
.36479 .11914 m
.35891 .10895 L
s
.35891 .10895 m
.35303 .11914 L
s
.35303 .11914 m
.34127 .11914 L
s
.34127 .11914 m
.34715 .12932 L
s
.34715 .12932 m
.34127 .1395 L
s
.34127 .1395 m
.35303 .1395 L
s
.35303 .1395 m
.35891 .14968 L
s
.35891 .14968 m
.35303 .15987 L
s
.35303 .15987 m
.34127 .15987 L
s
.34127 .15987 m
.34715 .17005 L
s
.34715 .17005 m
.34127 .18023 L
s
.34127 .18023 m
.35303 .18023 L
s
.35303 .18023 m
.35891 .19041 L
s
.35891 .19041 m
.36479 .18023 L
s
.36479 .18023 m
.37654 .18023 L
s
.37654 .18023 m
.38242 .19041 L
s
.38242 .19041 m
.37654 .2006 L
s
.37654 .2006 m
.3883 .2006 L
s
.3883 .2006 m
.39418 .21078 L
s
.39418 .21078 m
.3883 .22096 L
s
.3883 .22096 m
.37654 .22096 L
s
.37654 .22096 m
.38242 .23114 L
s
.38242 .23114 m
.37654 .24133 L
s
.37654 .24133 m
.36479 .24133 L
s
.36479 .24133 m
.35891 .23114 L
s
.35891 .23114 m
.35303 .24133 L
s
.35303 .24133 m
.34127 .24133 L
s
.34127 .24133 m
.34715 .25151 L
s
.34715 .25151 m
.34127 .26169 L
s
.34127 .26169 m
.35303 .26169 L
s
.35303 .26169 m
.35891 .27187 L
s
.35891 .27187 m
.35303 .28206 L
s
.35303 .28206 m
.34127 .28206 L
s
.34127 .28206 m
.34715 .29224 L
s
.34715 .29224 m
.34127 .30242 L
s
.34127 .30242 m
.32951 .30242 L
s
.32951 .30242 m
.32363 .29224 L
s
.32363 .29224 m
.31775 .30242 L
s
.31775 .30242 m
.306 .30242 L
s
.306 .30242 m
.30012 .29224 L
s
.30012 .29224 m
.306 .28206 L
s
.306 .28206 m
.29424 .28206 L
s
.29424 .28206 m
.28836 .27187 L
s
.28836 .27187 m
.28248 .28206 L
s
.28248 .28206 m
.27072 .28206 L
s
.27072 .28206 m
.2766 .29224 L
s
.2766 .29224 m
.27072 .30242 L
s
.27072 .30242 m
.25897 .30242 L
s
.25897 .30242 m
.25309 .29224 L
s
.25309 .29224 m
.24721 .30242 L
s
.24721 .30242 m
.23545 .30242 L
s
.23545 .30242 m
.22957 .29224 L
s
.22957 .29224 m
.23545 .28206 L
s
.23545 .28206 m
.22369 .28206 L
s
.22369 .28206 m
.21781 .27187 L
s
.21781 .27187 m
.22369 .26169 L
s
.22369 .26169 m
.23545 .26169 L
s
.23545 .26169 m
.22957 .25151 L
s
.22957 .25151 m
.23545 .24133 L
s
.23545 .24133 m
.22369 .24133 L
s
.22369 .24133 m
.21781 .23114 L
s
.21781 .23114 m
.21193 .24133 L
s
.21193 .24133 m
.20018 .24133 L
s
.20018 .24133 m
.1943 .23114 L
s
.1943 .23114 m
.20018 .22096 L
s
.20018 .22096 m
.18842 .22096 L
s
.18842 .22096 m
.18254 .21078 L
s
.18254 .21078 m
.17666 .22096 L
s
.17666 .22096 m
.1649 .22096 L
s
.1649 .22096 m
.17078 .23114 L
s
.17078 .23114 m
.1649 .24133 L
s
.1649 .24133 m
.15315 .24133 L
s
.15315 .24133 m
.14727 .23114 L
s
.14727 .23114 m
.14139 .24133 L
s
.14139 .24133 m
.12963 .24133 L
s
.12963 .24133 m
.13551 .25151 L
s
.13551 .25151 m
.12963 .26169 L
s
.12963 .26169 m
.14139 .26169 L
s
.14139 .26169 m
.14727 .27187 L
s
.14727 .27187 m
.14139 .28206 L
s
.14139 .28206 m
.12963 .28206 L
s
.12963 .28206 m
.13551 .29224 L
s
.13551 .29224 m
.12963 .30242 L
s
.12963 .30242 m
.11787 .30242 L
s
.11787 .30242 m
.11199 .29224 L
s
.11199 .29224 m
.10611 .30242 L
s
.10611 .30242 m
.09436 .30242 L
s
.09436 .30242 m
.08848 .29224 L
s
.08848 .29224 m
.09436 .28206 L
s
.09436 .28206 m
.0826 .28206 L
s
.0826 .28206 m
.07672 .27187 L
s
.07672 .27187 m
.07084 .28206 L
s
.07084 .28206 m
.05908 .28206 L
s
.05908 .28206 m
.06496 .29224 L
s
.06496 .29224 m
.05908 .30242 L
s
.05908 .30242 m
.04733 .30242 L
s
.04733 .30242 m
.04145 .29224 L
s
.04145 .29224 m
.03557 .30242 L
s
.03557 .30242 m
.02381 .30242 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0130918 0.97619 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0206197 0.952381 [
[ 0 0 0 0 ]
[ 1 .86603 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .86603 L
0 .86603 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .02062 m
.03557 .02062 L
s
.03557 .02062 m
.04145 .0308 L
s
.04145 .0308 m
.04733 .02062 L
s
.04733 .02062 m
.05908 .02062 L
s
.05908 .02062 m
.06496 .0308 L
s
.06496 .0308 m
.05908 .04098 L
s
.05908 .04098 m
.07084 .04098 L
s
.07084 .04098 m
.07672 .05117 L
s
.07672 .05117 m
.0826 .04098 L
s
.0826 .04098 m
.09436 .04098 L
s
.09436 .04098 m
.08848 .0308 L
s
.08848 .0308 m
.09436 .02062 L
s
.09436 .02062 m
.10611 .02062 L
s
.10611 .02062 m
.11199 .0308 L
s
.11199 .0308 m
.11787 .02062 L
s
.11787 .02062 m
.12963 .02062 L
s
.12963 .02062 m
.13551 .0308 L
s
.13551 .0308 m
.12963 .04098 L
s
.12963 .04098 m
.14139 .04098 L
s
.14139 .04098 m
.14727 .05117 L
s
.14727 .05117 m
.14139 .06135 L
s
.14139 .06135 m
.12963 .06135 L
s
.12963 .06135 m
.13551 .07153 L
s
.13551 .07153 m
.12963 .08171 L
s
.12963 .08171 m
.14139 .08171 L
s
.14139 .08171 m
.14727 .0919 L
s
.14727 .0919 m
.15315 .08171 L
s
.15315 .08171 m
.1649 .08171 L
s
.1649 .08171 m
.17078 .0919 L
s
.17078 .0919 m
.1649 .10208 L
s
.1649 .10208 m
.17666 .10208 L
s
.17666 .10208 m
.18254 .11226 L
s
.18254 .11226 m
.18842 .10208 L
s
.18842 .10208 m
.20018 .10208 L
s
.20018 .10208 m
.1943 .0919 L
s
.1943 .0919 m
.20018 .08171 L
s
.20018 .08171 m
.21193 .08171 L
s
.21193 .08171 m
.21781 .0919 L
s
.21781 .0919 m
.22369 .08171 L
s
.22369 .08171 m
.23545 .08171 L
s
.23545 .08171 m
.22957 .07153 L
s
.22957 .07153 m
.23545 .06135 L
s
.23545 .06135 m
.22369 .06135 L
s
.22369 .06135 m
.21781 .05117 L
s
.21781 .05117 m
.22369 .04098 L
s
.22369 .04098 m
.23545 .04098 L
s
.23545 .04098 m
.22957 .0308 L
s
.22957 .0308 m
.23545 .02062 L
s
.23545 .02062 m
.24721 .02062 L
s
.24721 .02062 m
.25309 .0308 L
s
.25309 .0308 m
.25897 .02062 L
s
.25897 .02062 m
.27072 .02062 L
s
.27072 .02062 m
.2766 .0308 L
s
.2766 .0308 m
.27072 .04098 L
s
.27072 .04098 m
.28248 .04098 L
s
.28248 .04098 m
.28836 .05117 L
s
.28836 .05117 m
.29424 .04098 L
s
.29424 .04098 m
.306 .04098 L
s
.306 .04098 m
.30012 .0308 L
s
.30012 .0308 m
.306 .02062 L
s
.306 .02062 m
.31775 .02062 L
s
.31775 .02062 m
.32363 .0308 L
s
.32363 .0308 m
.32951 .02062 L
s
.32951 .02062 m
.34127 .02062 L
s
.34127 .02062 m
.34715 .0308 L
s
.34715 .0308 m
.34127 .04098 L
s
.34127 .04098 m
.35303 .04098 L
s
.35303 .04098 m
.35891 .05117 L
s
.35891 .05117 m
.35303 .06135 L
s
.35303 .06135 m
.34127 .06135 L
s
.34127 .06135 m
.34715 .07153 L
s
.34715 .07153 m
.34127 .08171 L
s
.34127 .08171 m
.35303 .08171 L
s
.35303 .08171 m
.35891 .0919 L
s
.35891 .0919 m
.36479 .08171 L
s
.36479 .08171 m
.37654 .08171 L
s
.37654 .08171 m
.38242 .0919 L
s
.38242 .0919 m
.37654 .10208 L
s
.37654 .10208 m
.3883 .10208 L
s
.3883 .10208 m
.39418 .11226 L
s
.39418 .11226 m
.3883 .12245 L
s
.3883 .12245 m
.37654 .12245 L
s
.37654 .12245 m
.38242 .13263 L
s
.38242 .13263 m
.37654 .14281 L
s
.37654 .14281 m
.36479 .14281 L
s
.36479 .14281 m
.35891 .13263 L
s
.35891 .13263 m
.35303 .14281 L
s
.35303 .14281 m
.34127 .14281 L
s
.34127 .14281 m
.34715 .15299 L
s
.34715 .15299 m
.34127 .16318 L
s
.34127 .16318 m
.35303 .16318 L
s
.35303 .16318 m
.35891 .17336 L
s
.35891 .17336 m
.35303 .18354 L
s
.35303 .18354 m
.34127 .18354 L
s
.34127 .18354 m
.34715 .19372 L
s
.34715 .19372 m
.34127 .20391 L
s
.34127 .20391 m
.35303 .20391 L
s
.35303 .20391 m
.35891 .21409 L
s
.35891 .21409 m
.36479 .20391 L
s
.36479 .20391 m
.37654 .20391 L
s
.37654 .20391 m
.38242 .21409 L
s
.38242 .21409 m
.37654 .22427 L
s
.37654 .22427 m
.3883 .22427 L
s
.3883 .22427 m
.39418 .23445 L
s
.39418 .23445 m
.40006 .22427 L
s
.40006 .22427 m
.41182 .22427 L
s
.41182 .22427 m
.40594 .21409 L
s
.40594 .21409 m
.41182 .20391 L
s
.41182 .20391 m
.42357 .20391 L
s
.42357 .20391 m
.42945 .21409 L
s
.42945 .21409 m
.43533 .20391 L
s
.43533 .20391 m
.44709 .20391 L
s
.44709 .20391 m
.45297 .21409 L
s
.45297 .21409 m
.44709 .22427 L
s
.44709 .22427 m
.45885 .22427 L
s
.45885 .22427 m
.46473 .23445 L
s
.46473 .23445 m
.45885 .24464 L
s
.45885 .24464 m
.44709 .24464 L
s
.44709 .24464 m
.45297 .25482 L
s
.45297 .25482 m
.44709 .265 L
s
.44709 .265 m
.45885 .265 L
s
.45885 .265 m
.46473 .27518 L
s
.46473 .27518 m
.47061 .265 L
s
.47061 .265 m
.48236 .265 L
s
.48236 .265 m
.48824 .27518 L
s
.48824 .27518 m
.48236 .28537 L
s
.48236 .28537 m
.49412 .28537 L
s
.49412 .28537 m
.5 .29555 L
s
.5 .29555 m
.50588 .28537 L
s
.50588 .28537 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .27518 L
s
.51176 .27518 m
.51764 .265 L
s
.51764 .265 m
.52939 .265 L
s
.52939 .265 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .265 L
s
.54115 .265 m
.55291 .265 L
s
.55291 .265 m
.54703 .25482 L
s
.54703 .25482 m
.55291 .24464 L
s
.55291 .24464 m
.54115 .24464 L
s
.54115 .24464 m
.53527 .23445 L
s
.53527 .23445 m
.54115 .22427 L
s
.54115 .22427 m
.55291 .22427 L
s
.55291 .22427 m
.54703 .21409 L
s
.54703 .21409 m
.55291 .20391 L
s
.55291 .20391 m
.56467 .20391 L
s
.56467 .20391 m
.57055 .21409 L
s
.57055 .21409 m
.57643 .20391 L
s
.57643 .20391 m
.58818 .20391 L
s
.58818 .20391 m
.59406 .21409 L
s
.59406 .21409 m
.58818 .22427 L
s
.58818 .22427 m
.59994 .22427 L
s
.59994 .22427 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .22427 L
s
.6117 .22427 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .21409 L
s
.61758 .21409 m
.62346 .20391 L
s
.62346 .20391 m
.63521 .20391 L
s
.63521 .20391 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .20391 L
s
.64697 .20391 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .19372 L
s
.65285 .19372 m
.65873 .18354 L
s
.65873 .18354 m
.64697 .18354 L
s
.64697 .18354 m
.64109 .17336 L
s
.64109 .17336 m
.64697 .16318 L
s
.64697 .16318 m
.65873 .16318 L
s
.65873 .16318 m
.65285 .15299 L
s
.65285 .15299 m
.65873 .14281 L
s
.65873 .14281 m
.64697 .14281 L
s
.64697 .14281 m
.64109 .13263 L
s
.64109 .13263 m
.63521 .14281 L
s
.63521 .14281 m
.62346 .14281 L
s
.62346 .14281 m
.61758 .13263 L
s
.61758 .13263 m
.62346 .12245 L
s
.62346 .12245 m
.6117 .12245 L
s
.6117 .12245 m
.60582 .11226 L
s
.60582 .11226 m
.6117 .10208 L
s
.6117 .10208 m
.62346 .10208 L
s
.62346 .10208 m
.61758 .0919 L
s
.61758 .0919 m
.62346 .08171 L
s
.62346 .08171 m
.63521 .08171 L
s
.63521 .08171 m
.64109 .0919 L
s
.64109 .0919 m
.64697 .08171 L
s
.64697 .08171 m
.65873 .08171 L
s
.65873 .08171 m
.65285 .07153 L
s
.65285 .07153 m
.65873 .06135 L
s
.65873 .06135 m
.64697 .06135 L
s
.64697 .06135 m
.64109 .05117 L
s
.64109 .05117 m
.64697 .04098 L
s
.64697 .04098 m
.65873 .04098 L
s
.65873 .04098 m
.65285 .0308 L
s
.65285 .0308 m
.65873 .02062 L
s
.65873 .02062 m
.67049 .02062 L
s
.67049 .02062 m
.67637 .0308 L
s
.67637 .0308 m
.68225 .02062 L
s
.68225 .02062 m
.694 .02062 L
s
.694 .02062 m
.69988 .0308 L
s
.69988 .0308 m
.694 .04098 L
s
.694 .04098 m
.70576 .04098 L
s
.70576 .04098 m
.71164 .05117 L
s
.71164 .05117 m
.71752 .04098 L
s
.71752 .04098 m
.72928 .04098 L
s
.72928 .04098 m
.7234 .0308 L
s
.7234 .0308 m
.72928 .02062 L
s
.72928 .02062 m
.74103 .02062 L
s
.74103 .02062 m
.74691 .0308 L
s
.74691 .0308 m
.75279 .02062 L
s
.75279 .02062 m
.76455 .02062 L
s
.76455 .02062 m
.77043 .0308 L
s
.77043 .0308 m
.76455 .04098 L
s
.76455 .04098 m
.77631 .04098 L
s
.77631 .04098 m
.78219 .05117 L
s
.78219 .05117 m
.77631 .06135 L
s
.77631 .06135 m
.76455 .06135 L
s
.76455 .06135 m
.77043 .07153 L
s
.77043 .07153 m
.76455 .08171 L
s
.76455 .08171 m
.77631 .08171 L
s
.77631 .08171 m
.78219 .0919 L
s
.78219 .0919 m
.78807 .08171 L
s
.78807 .08171 m
.79982 .08171 L
s
.79982 .08171 m
.8057 .0919 L
s
.8057 .0919 m
.79982 .10208 L
s
.79982 .10208 m
.81158 .10208 L
s
.81158 .10208 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .10208 L
s
.82334 .10208 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .0919 L
s
.82922 .0919 m
.8351 .08171 L
s
.8351 .08171 m
.84685 .08171 L
s
.84685 .08171 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .08171 L
s
.85861 .08171 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .07153 L
s
.86449 .07153 m
.87037 .06135 L
s
.87037 .06135 m
.85861 .06135 L
s
.85861 .06135 m
.85273 .05117 L
s
.85273 .05117 m
.85861 .04098 L
s
.85861 .04098 m
.87037 .04098 L
s
.87037 .04098 m
.86449 .0308 L
s
.86449 .0308 m
.87037 .02062 L
s
.87037 .02062 m
.88213 .02062 L
s
.88213 .02062 m
.88801 .0308 L
s
.88801 .0308 m
.89389 .02062 L
s
.89389 .02062 m
.90564 .02062 L
s
.90564 .02062 m
.91152 .0308 L
s
.91152 .0308 m
.90564 .04098 L
s
.90564 .04098 m
.9174 .04098 L
s
.9174 .04098 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .04098 L
s
.92916 .04098 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .0308 L
s
.93504 .0308 m
.94092 .02062 L
s
.94092 .02062 m
.95267 .02062 L
s
.95267 .02062 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .02062 L
s
.96443 .02062 m
.97619 .02062 L
s
.97619 .02062 m
.97031 .0308 L
s
.97031 .0308 m
.95855 .0308 L
s
.95855 .0308 m
.96443 .04098 L
s
.96443 .04098 m
.95855 .05117 L
s
.95855 .05117 m
.9468 .05117 L
s
.9468 .05117 m
.94092 .04098 L
s
.94092 .04098 m
.93504 .05117 L
s
.93504 .05117 m
.92328 .05117 L
s
.92328 .05117 m
.92916 .06135 L
s
.92916 .06135 m
.92328 .07153 L
s
.92328 .07153 m
.93504 .07153 L
s
.93504 .07153 m
.94092 .08171 L
s
.94092 .08171 m
.93504 .0919 L
s
.93504 .0919 m
.92328 .0919 L
s
.92328 .0919 m
.92916 .10208 L
s
.92916 .10208 m
.92328 .11226 L
s
.92328 .11226 m
.91152 .11226 L
s
.91152 .11226 m
.90564 .10208 L
s
.90564 .10208 m
.89976 .11226 L
s
.89976 .11226 m
.88801 .11226 L
s
.88801 .11226 m
.88213 .10208 L
s
.88213 .10208 m
.88801 .0919 L
s
.88801 .0919 m
.87625 .0919 L
s
.87625 .0919 m
.87037 .08171 L
s
.87037 .08171 m
.86449 .0919 L
s
.86449 .0919 m
.85273 .0919 L
s
.85273 .0919 m
.85861 .10208 L
s
.85861 .10208 m
.85273 .11226 L
s
.85273 .11226 m
.84098 .11226 L
s
.84098 .11226 m
.8351 .10208 L
s
.8351 .10208 m
.82922 .11226 L
s
.82922 .11226 m
.81746 .11226 L
s
.81746 .11226 m
.82334 .12245 L
s
.82334 .12245 m
.81746 .13263 L
s
.81746 .13263 m
.82922 .13263 L
s
.82922 .13263 m
.8351 .14281 L
s
.8351 .14281 m
.82922 .15299 L
s
.82922 .15299 m
.81746 .15299 L
s
.81746 .15299 m
.82334 .16318 L
s
.82334 .16318 m
.81746 .17336 L
s
.81746 .17336 m
.82922 .17336 L
s
.82922 .17336 m
.8351 .18354 L
s
.8351 .18354 m
.84098 .17336 L
s
.84098 .17336 m
.85273 .17336 L
s
.85273 .17336 m
.85861 .18354 L
s
.85861 .18354 m
.85273 .19372 L
s
.85273 .19372 m
.86449 .19372 L
s
.86449 .19372 m
.87037 .20391 L
s
.87037 .20391 m
.86449 .21409 L
s
.86449 .21409 m
.85273 .21409 L
s
.85273 .21409 m
.85861 .22427 L
s
.85861 .22427 m
.85273 .23445 L
s
.85273 .23445 m
.84098 .23445 L
s
.84098 .23445 m
.8351 .22427 L
s
.8351 .22427 m
.82922 .23445 L
s
.82922 .23445 m
.81746 .23445 L
s
.81746 .23445 m
.82334 .24464 L
s
.82334 .24464 m
.81746 .25482 L
s
.81746 .25482 m
.82922 .25482 L
s
.82922 .25482 m
.8351 .265 L
s
.8351 .265 m
.82922 .27518 L
s
.82922 .27518 m
.81746 .27518 L
s
.81746 .27518 m
.82334 .28537 L
s
.82334 .28537 m
.81746 .29555 L
s
.81746 .29555 m
.8057 .29555 L
s
.8057 .29555 m
.79982 .28537 L
s
.79982 .28537 m
.79394 .29555 L
s
.79394 .29555 m
.78219 .29555 L
s
.78219 .29555 m
.77631 .28537 L
s
.77631 .28537 m
.78219 .27518 L
s
.78219 .27518 m
.77043 .27518 L
s
.77043 .27518 m
.76455 .265 L
s
.76455 .265 m
.75867 .27518 L
s
.75867 .27518 m
.74691 .27518 L
s
.74691 .27518 m
.75279 .28537 L
s
.75279 .28537 m
.74691 .29555 L
s
.74691 .29555 m
.73516 .29555 L
s
.73516 .29555 m
.72928 .28537 L
s
.72928 .28537 m
.7234 .29555 L
s
.7234 .29555 m
.71164 .29555 L
s
.71164 .29555 m
.70576 .28537 L
s
.70576 .28537 m
.71164 .27518 L
s
.71164 .27518 m
.69988 .27518 L
s
.69988 .27518 m
.694 .265 L
s
.694 .265 m
.69988 .25482 L
s
.69988 .25482 m
.71164 .25482 L
s
.71164 .25482 m
.70576 .24464 L
s
.70576 .24464 m
.71164 .23445 L
s
.71164 .23445 m
.69988 .23445 L
s
.69988 .23445 m
.694 .22427 L
s
.694 .22427 m
.68812 .23445 L
s
.68812 .23445 m
.67637 .23445 L
s
.67637 .23445 m
.67049 .22427 L
s
.67049 .22427 m
.67637 .21409 L
s
.67637 .21409 m
.66461 .21409 L
s
.66461 .21409 m
.65873 .20391 L
s
.65873 .20391 m
.65285 .21409 L
s
.65285 .21409 m
.64109 .21409 L
s
.64109 .21409 m
.64697 .22427 L
s
.64697 .22427 m
.64109 .23445 L
s
.64109 .23445 m
.62934 .23445 L
s
.62934 .23445 m
.62346 .22427 L
s
.62346 .22427 m
.61758 .23445 L
s
.61758 .23445 m
.60582 .23445 L
s
.60582 .23445 m
.6117 .24464 L
s
.6117 .24464 m
.60582 .25482 L
s
.60582 .25482 m
.61758 .25482 L
s
.61758 .25482 m
.62346 .265 L
s
.62346 .265 m
.61758 .27518 L
s
.61758 .27518 m
.60582 .27518 L
s
.60582 .27518 m
.6117 .28537 L
s
.6117 .28537 m
.60582 .29555 L
s
.60582 .29555 m
.59406 .29555 L
s
.59406 .29555 m
.58818 .28537 L
s
.58818 .28537 m
.5823 .29555 L
s
.5823 .29555 m
.57055 .29555 L
s
.57055 .29555 m
.56467 .28537 L
s
.56467 .28537 m
.57055 .27518 L
s
.57055 .27518 m
.55879 .27518 L
s
.55879 .27518 m
.55291 .265 L
s
.55291 .265 m
.54703 .27518 L
s
.54703 .27518 m
.53527 .27518 L
s
.53527 .27518 m
.54115 .28537 L
s
.54115 .28537 m
.53527 .29555 L
s
.53527 .29555 m
.52352 .29555 L
s
.52352 .29555 m
.51764 .28537 L
s
.51764 .28537 m
.51176 .29555 L
s
.51176 .29555 m
.5 .29555 L
s
.5 .29555 m
.50588 .30573 L
s
.50588 .30573 m
.5 .31591 L
s
.5 .31591 m
.51176 .31591 L
s
.51176 .31591 m
.51764 .3261 L
s
.51764 .3261 m
.51176 .33628 L
s
.51176 .33628 m
.5 .33628 L
s
.5 .33628 m
.50588 .34646 L
s
.50588 .34646 m
.5 .35664 L
s
.5 .35664 m
.51176 .35664 L
s
.51176 .35664 m
.51764 .36683 L
s
.51764 .36683 m
.52352 .35664 L
s
.52352 .35664 m
.53527 .35664 L
s
.53527 .35664 m
.54115 .36683 L
s
.54115 .36683 m
.53527 .37701 L
s
.53527 .37701 m
.54703 .37701 L
s
.54703 .37701 m
.55291 .38719 L
s
.55291 .38719 m
.54703 .39737 L
s
.54703 .39737 m
.53527 .39737 L
s
.53527 .39737 m
.54115 .40756 L
s
.54115 .40756 m
.53527 .41774 L
s
.53527 .41774 m
.52352 .41774 L
s
.52352 .41774 m
.51764 .40756 L
s
.51764 .40756 m
.51176 .41774 L
s
.51176 .41774 m
.5 .41774 L
s
.5 .41774 m
.50588 .42792 L
s
.50588 .42792 m
.5 .4381 L
s
.5 .4381 m
.51176 .4381 L
s
.51176 .4381 m
.51764 .44829 L
s
.51764 .44829 m
.51176 .45847 L
s
.51176 .45847 m
.5 .45847 L
s
.5 .45847 m
.50588 .46865 L
s
.50588 .46865 m
.5 .47883 L
s
.5 .47883 m
.51176 .47883 L
s
.51176 .47883 m
.51764 .48902 L
s
.51764 .48902 m
.52352 .47883 L
s
.52352 .47883 m
.53527 .47883 L
s
.53527 .47883 m
.54115 .48902 L
s
.54115 .48902 m
.53527 .4992 L
s
.53527 .4992 m
.54703 .4992 L
s
.54703 .4992 m
.55291 .50938 L
s
.55291 .50938 m
.55879 .4992 L
s
.55879 .4992 m
.57055 .4992 L
s
.57055 .4992 m
.56467 .48902 L
s
.56467 .48902 m
.57055 .47883 L
s
.57055 .47883 m
.5823 .47883 L
s
.5823 .47883 m
.58818 .48902 L
s
.58818 .48902 m
.59406 .47883 L
s
.59406 .47883 m
.60582 .47883 L
s
.60582 .47883 m
.6117 .48902 L
s
.6117 .48902 m
.60582 .4992 L
s
.60582 .4992 m
.61758 .4992 L
s
.61758 .4992 m
.62346 .50938 L
s
.62346 .50938 m
.61758 .51956 L
s
.61758 .51956 m
.60582 .51956 L
s
.60582 .51956 m
.6117 .52975 L
s
.6117 .52975 m
.60582 .53993 L
s
.60582 .53993 m
.61758 .53993 L
s
.61758 .53993 m
.62346 .55011 L
s
.62346 .55011 m
.62934 .53993 L
s
.62934 .53993 m
.64109 .53993 L
s
.64109 .53993 m
.64697 .55011 L
s
.64697 .55011 m
.64109 .56029 L
s
.64109 .56029 m
.65285 .56029 L
s
.65285 .56029 m
.65873 .57048 L
s
.65873 .57048 m
.65285 .58066 L
s
.65285 .58066 m
.64109 .58066 L
s
.64109 .58066 m
.64697 .59084 L
s
.64697 .59084 m
.64109 .60102 L
s
.64109 .60102 m
.62934 .60102 L
s
.62934 .60102 m
.62346 .59084 L
s
.62346 .59084 m
.61758 .60102 L
s
.61758 .60102 m
.60582 .60102 L
s
.60582 .60102 m
.6117 .61121 L
s
.6117 .61121 m
.60582 .62139 L
s
.60582 .62139 m
.61758 .62139 L
s
.61758 .62139 m
.62346 .63157 L
s
.62346 .63157 m
.61758 .64175 L
s
.61758 .64175 m
.60582 .64175 L
s
.60582 .64175 m
.6117 .65194 L
s
.6117 .65194 m
.60582 .66212 L
s
.60582 .66212 m
.59406 .66212 L
s
.59406 .66212 m
.58818 .65194 L
s
.58818 .65194 m
.5823 .66212 L
s
.5823 .66212 m
.57055 .66212 L
s
.57055 .66212 m
.56467 .65194 L
s
.56467 .65194 m
.57055 .64175 L
s
.57055 .64175 m
.55879 .64175 L
s
.55879 .64175 m
.55291 .63157 L
s
.55291 .63157 m
.54703 .64175 L
s
.54703 .64175 m
.53527 .64175 L
s
.53527 .64175 m
.54115 .65194 L
s
.54115 .65194 m
.53527 .66212 L
s
.53527 .66212 m
.52352 .66212 L
s
.52352 .66212 m
.51764 .65194 L
s
.51764 .65194 m
.51176 .66212 L
s
.51176 .66212 m
.5 .66212 L
s
.5 .66212 m
.50588 .6723 L
s
.50588 .6723 m
.5 .68249 L
s
.5 .68249 m
.51176 .68249 L
s
.51176 .68249 m
.51764 .69267 L
s
.51764 .69267 m
.51176 .70285 L
s
.51176 .70285 m
.5 .70285 L
s
.5 .70285 m
.50588 .71303 L
s
.50588 .71303 m
.5 .72322 L
s
.5 .72322 m
.51176 .72322 L
s
.51176 .72322 m
.51764 .7334 L
s
.51764 .7334 m
.52352 .72322 L
s
.52352 .72322 m
.53527 .72322 L
s
.53527 .72322 m
.54115 .7334 L
s
.54115 .7334 m
.53527 .74358 L
s
.53527 .74358 m
.54703 .74358 L
s
.54703 .74358 m
.55291 .75376 L
s
.55291 .75376 m
.54703 .76395 L
s
.54703 .76395 m
.53527 .76395 L
s
.53527 .76395 m
.54115 .77413 L
s
.54115 .77413 m
.53527 .78431 L
s
.53527 .78431 m
.52352 .78431 L
s
.52352 .78431 m
.51764 .77413 L
s
.51764 .77413 m
.51176 .78431 L
s
.51176 .78431 m
.5 .78431 L
s
.5 .78431 m
.50588 .79449 L
s
.50588 .79449 m
.5 .80468 L
s
.5 .80468 m
.51176 .80468 L
s
.51176 .80468 m
.51764 .81486 L
s
.51764 .81486 m
.51176 .82504 L
s
.51176 .82504 m
.5 .82504 L
s
.5 .82504 m
.50588 .83522 L
s
.50588 .83522 m
.5 .84541 L
s
.5 .84541 m
.49412 .83522 L
s
.49412 .83522 m
.5 .82504 L
s
.5 .82504 m
.48824 .82504 L
s
.48824 .82504 m
.48236 .81486 L
s
.48236 .81486 m
.48824 .80468 L
s
.48824 .80468 m
.5 .80468 L
s
.5 .80468 m
.49412 .79449 L
s
.49412 .79449 m
.5 .78431 L
s
.5 .78431 m
.48824 .78431 L
s
.48824 .78431 m
.48236 .77413 L
s
.48236 .77413 m
.47648 .78431 L
s
.47648 .78431 m
.46473 .78431 L
s
.46473 .78431 m
.45885 .77413 L
s
.45885 .77413 m
.46473 .76395 L
s
.46473 .76395 m
.45297 .76395 L
s
.45297 .76395 m
.44709 .75376 L
s
.44709 .75376 m
.45297 .74358 L
s
.45297 .74358 m
.46473 .74358 L
s
.46473 .74358 m
.45885 .7334 L
s
.45885 .7334 m
.46473 .72322 L
s
.46473 .72322 m
.47648 .72322 L
s
.47648 .72322 m
.48236 .7334 L
s
.48236 .7334 m
.48824 .72322 L
s
.48824 .72322 m
.5 .72322 L
s
.5 .72322 m
.49412 .71303 L
s
.49412 .71303 m
.5 .70285 L
s
.5 .70285 m
.48824 .70285 L
s
.48824 .70285 m
.48236 .69267 L
s
.48236 .69267 m
.48824 .68249 L
s
.48824 .68249 m
.5 .68249 L
s
.5 .68249 m
.49412 .6723 L
s
.49412 .6723 m
.5 .66212 L
s
.5 .66212 m
.48824 .66212 L
s
.48824 .66212 m
.48236 .65194 L
s
.48236 .65194 m
.47648 .66212 L
s
.47648 .66212 m
.46473 .66212 L
s
.46473 .66212 m
.45885 .65194 L
s
.45885 .65194 m
.46473 .64175 L
s
.46473 .64175 m
.45297 .64175 L
s
.45297 .64175 m
.44709 .63157 L
s
.44709 .63157 m
.44121 .64175 L
s
.44121 .64175 m
.42945 .64175 L
s
.42945 .64175 m
.43533 .65194 L
s
.43533 .65194 m
.42945 .66212 L
s
.42945 .66212 m
.4177 .66212 L
s
.4177 .66212 m
.41182 .65194 L
s
.41182 .65194 m
.40594 .66212 L
s
.40594 .66212 m
.39418 .66212 L
s
.39418 .66212 m
.3883 .65194 L
s
.3883 .65194 m
.39418 .64175 L
s
.39418 .64175 m
.38242 .64175 L
s
.38242 .64175 m
.37654 .63157 L
s
.37654 .63157 m
.38242 .62139 L
s
.38242 .62139 m
.39418 .62139 L
s
.39418 .62139 m
.3883 .61121 L
s
.3883 .61121 m
.39418 .60102 L
s
.39418 .60102 m
.38242 .60102 L
s
.38242 .60102 m
.37654 .59084 L
s
.37654 .59084 m
.37066 .60102 L
s
.37066 .60102 m
.35891 .60102 L
s
.35891 .60102 m
.35303 .59084 L
s
.35303 .59084 m
.35891 .58066 L
s
.35891 .58066 m
.34715 .58066 L
s
.34715 .58066 m
.34127 .57048 L
s
.34127 .57048 m
.34715 .56029 L
s
.34715 .56029 m
.35891 .56029 L
s
.35891 .56029 m
.35303 .55011 L
s
.35303 .55011 m
.35891 .53993 L
s
.35891 .53993 m
.37066 .53993 L
s
.37066 .53993 m
.37654 .55011 L
s
.37654 .55011 m
.38242 .53993 L
s
.38242 .53993 m
.39418 .53993 L
s
.39418 .53993 m
.3883 .52975 L
s
.3883 .52975 m
.39418 .51956 L
s
.39418 .51956 m
.38242 .51956 L
s
.38242 .51956 m
.37654 .50938 L
s
.37654 .50938 m
.38242 .4992 L
s
.38242 .4992 m
.39418 .4992 L
s
.39418 .4992 m
.3883 .48902 L
s
.3883 .48902 m
.39418 .47883 L
s
.39418 .47883 m
.40594 .47883 L
s
.40594 .47883 m
.41182 .48902 L
s
.41182 .48902 m
.4177 .47883 L
s
.4177 .47883 m
.42945 .47883 L
s
.42945 .47883 m
.43533 .48902 L
s
.43533 .48902 m
.42945 .4992 L
s
.42945 .4992 m
.44121 .4992 L
s
.44121 .4992 m
.44709 .50938 L
s
.44709 .50938 m
.45297 .4992 L
s
.45297 .4992 m
.46473 .4992 L
s
.46473 .4992 m
.45885 .48902 L
s
.45885 .48902 m
.46473 .47883 L
s
.46473 .47883 m
.47648 .47883 L
s
.47648 .47883 m
.48236 .48902 L
s
.48236 .48902 m
.48824 .47883 L
s
.48824 .47883 m
.5 .47883 L
s
.5 .47883 m
.49412 .46865 L
s
.49412 .46865 m
.5 .45847 L
s
.5 .45847 m
.48824 .45847 L
s
.48824 .45847 m
.48236 .44829 L
s
.48236 .44829 m
.48824 .4381 L
s
.48824 .4381 m
.5 .4381 L
s
.5 .4381 m
.49412 .42792 L
s
.49412 .42792 m
.5 .41774 L
s
.5 .41774 m
.48824 .41774 L
s
.48824 .41774 m
.48236 .40756 L
s
.48236 .40756 m
.47648 .41774 L
s
.47648 .41774 m
.46473 .41774 L
s
.46473 .41774 m
.45885 .40756 L
s
.45885 .40756 m
.46473 .39737 L
s
.46473 .39737 m
.45297 .39737 L
s
.45297 .39737 m
.44709 .38719 L
s
.44709 .38719 m
.45297 .37701 L
s
.45297 .37701 m
.46473 .37701 L
s
.46473 .37701 m
.45885 .36683 L
s
.45885 .36683 m
.46473 .35664 L
s
.46473 .35664 m
.47648 .35664 L
s
.47648 .35664 m
.48236 .36683 L
s
.48236 .36683 m
.48824 .35664 L
s
.48824 .35664 m
.5 .35664 L
s
.5 .35664 m
.49412 .34646 L
s
.49412 .34646 m
.5 .33628 L
s
.5 .33628 m
.48824 .33628 L
s
.48824 .33628 m
.48236 .3261 L
s
.48236 .3261 m
.48824 .31591 L
s
.48824 .31591 m
.5 .31591 L
s
.5 .31591 m
.49412 .30573 L
s
.49412 .30573 m
.5 .29555 L
s
.5 .29555 m
.48824 .29555 L
s
.48824 .29555 m
.48236 .28537 L
s
.48236 .28537 m
.47648 .29555 L
s
.47648 .29555 m
.46473 .29555 L
s
.46473 .29555 m
.45885 .28537 L
s
.45885 .28537 m
.46473 .27518 L
s
.46473 .27518 m
.45297 .27518 L
s
.45297 .27518 m
.44709 .265 L
s
.44709 .265 m
.44121 .27518 L
s
.44121 .27518 m
.42945 .27518 L
s
.42945 .27518 m
.43533 .28537 L
s
.43533 .28537 m
.42945 .29555 L
s
.42945 .29555 m
.4177 .29555 L
s
.4177 .29555 m
.41182 .28537 L
s
.41182 .28537 m
.40594 .29555 L
s
.40594 .29555 m
.39418 .29555 L
s
.39418 .29555 m
.3883 .28537 L
s
.3883 .28537 m
.39418 .27518 L
s
.39418 .27518 m
.38242 .27518 L
s
.38242 .27518 m
.37654 .265 L
s
.37654 .265 m
.38242 .25482 L
s
.38242 .25482 m
.39418 .25482 L
s
.39418 .25482 m
.3883 .24464 L
s
.3883 .24464 m
.39418 .23445 L
s
.39418 .23445 m
.38242 .23445 L
s
.38242 .23445 m
.37654 .22427 L
s
.37654 .22427 m
.37066 .23445 L
s
.37066 .23445 m
.35891 .23445 L
s
.35891 .23445 m
.35303 .22427 L
s
.35303 .22427 m
.35891 .21409 L
s
.35891 .21409 m
.34715 .21409 L
s
.34715 .21409 m
.34127 .20391 L
s
.34127 .20391 m
.33539 .21409 L
s
.33539 .21409 m
.32363 .21409 L
s
.32363 .21409 m
.32951 .22427 L
s
.32951 .22427 m
.32363 .23445 L
s
.32363 .23445 m
.31188 .23445 L
s
.31188 .23445 m
.306 .22427 L
s
.306 .22427 m
.30012 .23445 L
s
.30012 .23445 m
.28836 .23445 L
s
.28836 .23445 m
.29424 .24464 L
s
.29424 .24464 m
.28836 .25482 L
s
.28836 .25482 m
.30012 .25482 L
s
.30012 .25482 m
.306 .265 L
s
.306 .265 m
.30012 .27518 L
s
.30012 .27518 m
.28836 .27518 L
s
.28836 .27518 m
.29424 .28537 L
s
.29424 .28537 m
.28836 .29555 L
s
.28836 .29555 m
.2766 .29555 L
s
.2766 .29555 m
.27072 .28537 L
s
.27072 .28537 m
.26484 .29555 L
s
.26484 .29555 m
.25309 .29555 L
s
.25309 .29555 m
.24721 .28537 L
s
.24721 .28537 m
.25309 .27518 L
s
.25309 .27518 m
.24133 .27518 L
s
.24133 .27518 m
.23545 .265 L
s
.23545 .265 m
.22957 .27518 L
s
.22957 .27518 m
.21781 .27518 L
s
.21781 .27518 m
.22369 .28537 L
s
.22369 .28537 m
.21781 .29555 L
s
.21781 .29555 m
.20606 .29555 L
s
.20606 .29555 m
.20018 .28537 L
s
.20018 .28537 m
.1943 .29555 L
s
.1943 .29555 m
.18254 .29555 L
s
.18254 .29555 m
.17666 .28537 L
s
.17666 .28537 m
.18254 .27518 L
s
.18254 .27518 m
.17078 .27518 L
s
.17078 .27518 m
.1649 .265 L
s
.1649 .265 m
.17078 .25482 L
s
.17078 .25482 m
.18254 .25482 L
s
.18254 .25482 m
.17666 .24464 L
s
.17666 .24464 m
.18254 .23445 L
s
.18254 .23445 m
.17078 .23445 L
s
.17078 .23445 m
.1649 .22427 L
s
.1649 .22427 m
.15902 .23445 L
s
.15902 .23445 m
.14727 .23445 L
s
.14727 .23445 m
.14139 .22427 L
s
.14139 .22427 m
.14727 .21409 L
s
.14727 .21409 m
.13551 .21409 L
s
.13551 .21409 m
.12963 .20391 L
s
.12963 .20391 m
.13551 .19372 L
s
.13551 .19372 m
.14727 .19372 L
s
.14727 .19372 m
.14139 .18354 L
s
.14139 .18354 m
.14727 .17336 L
s
.14727 .17336 m
.15902 .17336 L
s
.15902 .17336 m
.1649 .18354 L
s
.1649 .18354 m
.17078 .17336 L
s
.17078 .17336 m
.18254 .17336 L
s
.18254 .17336 m
.17666 .16318 L
s
.17666 .16318 m
.18254 .15299 L
s
.18254 .15299 m
.17078 .15299 L
s
.17078 .15299 m
.1649 .14281 L
s
.1649 .14281 m
.17078 .13263 L
s
.17078 .13263 m
.18254 .13263 L
s
.18254 .13263 m
.17666 .12245 L
s
.17666 .12245 m
.18254 .11226 L
s
.18254 .11226 m
.17078 .11226 L
s
.17078 .11226 m
.1649 .10208 L
s
.1649 .10208 m
.15902 .11226 L
s
.15902 .11226 m
.14727 .11226 L
s
.14727 .11226 m
.14139 .10208 L
s
.14139 .10208 m
.14727 .0919 L
s
.14727 .0919 m
.13551 .0919 L
s
.13551 .0919 m
.12963 .08171 L
s
.12963 .08171 m
.12375 .0919 L
s
.12375 .0919 m
.11199 .0919 L
s
.11199 .0919 m
.11787 .10208 L
s
.11787 .10208 m
.11199 .11226 L
s
.11199 .11226 m
.10024 .11226 L
s
.10024 .11226 m
.09436 .10208 L
s
.09436 .10208 m
.08848 .11226 L
s
.08848 .11226 m
.07672 .11226 L
s
.07672 .11226 m
.07084 .10208 L
s
.07084 .10208 m
.07672 .0919 L
s
.07672 .0919 m
.06496 .0919 L
s
.06496 .0919 m
.05908 .08171 L
s
.05908 .08171 m
.06496 .07153 L
s
.06496 .07153 m
.07672 .07153 L
s
.07672 .07153 m
.07084 .06135 L
s
.07084 .06135 m
.07672 .05117 L
s
.07672 .05117 m
.06496 .05117 L
s
.06496 .05117 m
.05908 .04098 L
s
.05908 .04098 m
.0532 .05117 L
s
.0532 .05117 m
.04145 .05117 L
s
.04145 .05117 m
.03557 .04098 L
s
.03557 .04098 m
.04145 .0308 L
s
.04145 .0308 m
.02969 .0308 L
s
.02969 .0308 m
.02381 .02062 L
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 158.313},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000kB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K03oool00`000000oooo0?ooo`3o0?oood80oooo001[0?ooo`800000o`3ooom40?ooo`00
J03oool4000000040?ooo`000000000000000?l0oooo@@3oool006P0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`3o0?oood00oooo001U0?ooo`050000003oool0oooo0?ooo`000000
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?l0oooo>`3oool006@0oooo0P00
00020?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0oooo0P00003o0?ooocd0oooo
001Q0?ooo`@0000000@0oooo00000000000000001P3oool300000080oooo0`00003o0?ooocX0oooo
001R0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3oooli0?ooo`00H`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0oooo>@3oool00680oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`3o0?ooocT0oooo001Q0?ooo`@000003`3oool300000?l0
oooo>P3oool006@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`3o0?oooc/0oooo
001G0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo0?l0oooo;P3oool005H0oooo0P00000<0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool02@3oool200000?l0oooo<03oool005@0oooo0`0000000`3o
ool0000000000002000000D0oooo1000000?0?ooo`<000001P3oool3000000030?ooo`0000000000
00800000o`3oool/0?ooo`00E03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0o`3oool[0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`050000
003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0
oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo0000003o0?ooobT0oooo001?0?ooo`8000000`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo008000000`3oool00`000000oooo0?ooo`0@0?ooo`060000003oool0oooo
0000003oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00800000o`3o
oolY0?ooo`00C@3oool3000000030?ooo`0000000000008000001P3oool3000000040?ooo`000000
0000000001@0oooo0`0000020?ooo`<000001P3oool3000000030?ooo`000000000000800000o`3o
oolU0?ooo`00C@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0?l0oooo903oool0
04h0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`3o0?ooob@0oooo001=0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool0o`3ooolT0?ooo`00C@3oool3000003P0oooo
1000003o0?ooobD0oooo001?0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0o`3o
oolW0?ooo`00D03oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0?l0oooo9`3oool0
04l0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`3o0?ooobL0oooo001=0?ooo`<0
0000>03oool400000?l0oooo9@3oool004d0oooo00<000000?ooo`3oool0>P3oool00`000000oooo
0?ooo`1H0?ooo`@0000000@0oooo00000000000000001P3oool300000080oooo0`00000D0?ooo`<0
000000<0oooo0000000000000P0000050?ooo`@0000000@0oooo0000000000000000?@3oool30000
0080oooo0`0000060?ooo`<000000P3oool3000001<0oooo10000000103oool00000000000000006
0?ooo`@0000000@0oooo0000000000000000403oool004h0oooo00<000000?ooo`3oool0>@3oool0
0`000000oooo0?ooo`1I0?ooo`040000003oool0000000000080oooo00<000000?ooo`3oool00`3o
ool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`02000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool200000080
oooo00<000000?ooo`3oool0>P3oool010000000oooo0?ooo`3oool200000080oooo00<000000?oo
o`3oool00`3oool01P000000oooo0?ooo`000000oooo00000080oooo00<000000?ooo`3oool0403o
ool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3oool00`3oool010000000oooo
0?ooo`3oool2000000030?ooo`000000oooo0100oooo001=0?ooo`030000003oool0oooo03X0oooo
00<000000?ooo`3oool0FP3oool3000000<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`020?ooo`050000003oool0oooo0?ooo`0000004P3oool01@000000oooo0?ooo`3oool00000
00@0oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000>`3o
ool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103oool01@000000oooo
0?ooo`3oool0000000@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`<000004@3oool004d0
oooo0`0000000`3oool0000000000002000002l0oooo0`0000000`3oool0000000000002000005`0
oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool04@3oool004l0oooo0P0000030?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3o
ool00P00001P0?ooo`040000003oool0oooo0?ooo`@0000000@0oooo00000000000000001P3oool3
000000l0oooo100000060?ooo`<0000000@0oooo00000000000000001P3oool4000003P0oooo0`00
00060?ooo`<000000P3oool3000000H0oooo0`00000?0?ooo`<000001P3oool4000000040?ooo`00
00000000000000@0oooo00<000000?ooo`3oool0403oool00500oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0[0?ooo`050000003oool0oooo0?ooo`000000H03oool00`000000oooo
0?ooo`0200000080oooo0P00000;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
2P3oool2000000`0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0:0?ooo`800000
2`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00X0oooo0P0000000`3oool00000
000000030?ooo`030000003oool0oooo0140oooo001D0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0H`3oool3000000030?ooo`000000000000800000303oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0d0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo0`0000000`3oool0000000000002000001<0oooo001D
0?ooo`<00000:P3oool4000006X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0H
0?ooo`00EP3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo06`0oooo00<000000?oo
o`3oool0203oool3000000l0oooo1000000C0?ooo`@00000>03oool3000001@0oooo0`00000?0?oo
o`<000002`3oool00`000000oooo0?ooo`0G0?ooo`00;P3oool00`000000oooo0?ooo`0V0?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`120?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0603oool002d0oooo0P00000W0?ooo`030000
003oool0oooo02L0oooo00<000000?ooo`3oool09P3oool200000480oooo1000000:0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0i0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`090?ooo`<00000
603oool002X0oooo10000000103oool0000000000000000R0?ooo`<00000:P3oool400000240oooo
0`0000020?ooo`<00000?P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo01<0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01H0oooo
000[0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool08@3oool00`000000oooo0?oo
o`0/0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0m0?ooo`030000003oool0oooo00/0oooo0`0000020?ooo`<000001P3oool3000000030?ooo`00
00000000008000004`3oool4000000040?ooo`0000000000000002l0oooo10000000103oool00000
00000000000D0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<0000000@0oooo00000000
000000003@3oool00`000000oooo0?ooo`0E0?ooo`009`3oool00`000000oooo0?ooo`020?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0:`3oool01@000000oooo0?ooo`3oool0000001/0oooo00D00000
0?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000>`3oool00`000000oooo
0?ooo`0:0?ooo`80000000@0oooo0000003oool000000P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo008000006P3oool200000080oooo00<000000?ooo`3oool0;03oool010000000
oooo0?ooo`3oool2000001T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P000000103oool000000?ooo`00000;0?ooo`030000003oool0oooo01H0oooo000V0?oo
o`8000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`8000006@3oool2
000000<0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`02000001X0oooo0P000002
0?ooo`030000003oool0oooo00<0oooo00H000000?ooo`3oool000000?ooo`00000k0?ooo`@00000
1P3oool3000000040?ooo`0000000000000000@0oooo00<000000?ooo`3oool00P3oool01@000000
oooo0?ooo`3oool0000001/0oooo00D000000?ooo`3oool0oooo0000000]0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`020?ooo`050000003oool0
oooo0?ooo`000000103oool300000080oooo0`0000060?ooo`<00000603oool002@0oooo0`000000
103oool000000000000000060?ooo`@0000000@0oooo0000000000000000503oool3000000030?oo
o`000000000000800000;`3oool3000000030?ooo`0000000000008000004`3oool4000000040?oo
o`0000000000000000H0oooo0`0000020?ooo`<00000>P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0200oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0603oool002@0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool04P3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo03T0oooo00D000000?ooo`3oool0oooo000000040?oo
o`050000003oool0oooo0?ooo`000000103oool3000000030?ooo`0000000000008000008@3oool3
000002/0oooo0`00000R0?ooo`<0000000<0oooo0000000000000P0000030?ooo`030000003oool0
oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo01L0oooo000U
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0i
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0i
0?ooo`030000003oool0oooo008000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo0?ooo`80000000<0oooo000000000000:03oool00`000000oooo0?ooo`0W0?ooo`030000003o
ool0oooo02D0oooo0P0000020?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo008000000`3oool00`000000oooo0?ooo`0H0?ooo`00903oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0>P3oool3000000030?ooo`000000
0000008000001@3oool4000000040?ooo`0000000000000002T0oooo00<000000?ooo`3oool09P3o
ool00`000000oooo0?ooo`0V0?ooo`<000000P3oool3000000H0oooo0`0000000`3oool000000000
0002000001X0oooo000T0?ooo`<000003`3oool3000001@0oooo0`00000h0?ooo`@000004`3oool4
000000l0oooo0`00001>0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02d0oooo000V0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
02@0oooo0`00000[0?ooo`<000009P3oool00`000000oooo0?ooo`0/0?ooo`006@3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0T0?ooo`030000003oool0oooo02d0oooo000H0?ooo`800000303oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`0000002`3oool00`000000oooo
0?ooo`0e0?ooo`030000003oool0oooo00X0oooo0P00000<0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool02@3oool2000004<0oooo0`00000R0?ooo`050000003oool0oooo0?ooo`00
0000;@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01l0oooo0`00000]0?ooo`00
5P3oool3000000030?ooo`0000000000008000001P3oool3000000l0oooo0`0000060?ooo`<00000
0P3oool3000000H0oooo0`00000h0?ooo`@000001P3oool3000000040?ooo`0000000000000000H0
oooo1000000?0?ooo`<000001P3oool3000000030?ooo`000000000000800000?P3oool00`000000
oooo0?ooo`0R0?ooo`8000000P3oool00`000000oooo0?ooo`0/0?ooo`040000003oool0oooo0?oo
o`800000903oool00`000000oooo0?ooo`0Z0?ooo`005P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo03d0oooo00<000000?ooo`3oool07`3oool4000000040?ooo`00000000000000
02l0oooo10000000103oool0000000000000000R0?ooo`030000003oool0oooo02T0oooo000B0?oo
o`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`050000003oool0
oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0
>@3oool01@000000oooo0?ooo`3oool0000000@0oooo00D000000?ooo`3oool0oooo000000040?oo
o`050000003oool0oooo0?ooo`0000004P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<0
00000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000003/0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
:P3oool00140oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P00
00030?ooo`030000003oool0oooo0100oooo00@000000?ooo`3oool0oooo0P0000020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P0000030?ooo`030000003oool0oooo03X0oooo
00<000000?ooo`3oool00P0000030?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0
oooo0P0000020?ooo`030000003oool0oooo0140oooo00H000000?ooo`3oool000000?ooo`000002
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P00000l0?ooo`<000007`3oool0
0`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01d0oooo0`00000]0?ooo`003`3oool30000
00030?ooo`0000000000008000001P3oool3000000030?ooo`0000000000008000004`3oool30000
0080oooo0`0000060?ooo`<0000000<0oooo0000000000000P00000m0?ooo`<0000000<0oooo0000
000000000P0000050?ooo`@0000000@0oooo0000000000000000503oool300000080oooo0`000006
0?ooo`<0000000<0oooo0000000000000P00000i0?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool0>P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo02d0oooo000?0?oo
o`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0M0?oo
o`@00000>03oool3000001l0oooo00<000000?ooo`3oool0;03oool00100oooo00<000000?ooo`3o
ool0]03oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3o
ool0=@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo02d0oooo000?0?ooo`030000
003oool0oooo0;H0oooo00<000000?ooo`3oool0<P3oool4000000040?ooo`000000000000000280
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0O0?ooo`<000000P3oool3000002P0
oooo000?0?ooo`<00000]03oool400000380oooo00@000000?ooo`3oool0oooo0P00000T0?ooo`03
0000003oool0oooo03D0oooo00<000000?ooo`3oool08P3oool200000080oooo00<000000?ooo`3o
ool09P3oool00140oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`0d0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool07`3oool4000003P0oooo0`00000R0?ooo`050000
003oool0oooo0?ooo`0000009`3oool00180oooo00<000000?ooo`3oool0/03oool00`000000oooo
0?ooo`0d0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0>P3oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo02H0oooo000A0?ooo`030000003oool0oooo0;40oooo00<0
00000?ooo`3oool0<`3oool3000002H0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?oo
o`0T0?ooo`<000009P3oool000l0oooo0`00002d0?ooo`@00000<03oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0T
0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool09P3oool4000000040?ooo`0000000000000000H0oooo0`0000020?ooo`<00000503oool3
000000040?ooo`0000000000000000H0oooo10000000103oool0000000000000000X0?ooo`030000
003oool0oooo02<0oooo000@0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0<03o
ool00`000000oooo0?ooo`0U0?ooo`8000000P3oool200000080oooo00<000000?ooo`3oool00`3o
ool01P000000oooo0?ooo`000000oooo00000080oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`02000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool20000
00030?ooo`000000000002L0oooo00<000000?ooo`3oool0903oool000l0oooo00<000000?ooo`3o
ool0]P3oool00`000000oooo0?ooo`0`0?ooo`<000008P3oool3000000030?ooo`00000000000080
00000`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool01@00
0000oooo0?ooo`3oool0000000@0oooo0`0000000`3oool000000000000200000240oooo0`00000V
0?ooo`003`3oool3000000030?ooo`000000000000800000Z`3oool3000000030?ooo`0000000000
00800000<P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00<0oooo00D000000?oo
o`3oool0oooo000000050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool09P3oool0
0140oooo0P0000030?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool00P00000f0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0`0000020?ooo`<000001P3oool3
000000l0oooo100000050?ooo`@0000000@0oooo0000000000000000103oool01@000000oooo0?oo
o`3oool0000000@0oooo00<000000?ooo`3oool06@3oool01@000000oooo0?ooo`3oool0000002L0
oooo000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0Y`3oool01@000000oooo
0?ooo`3oool0000003H0oooo00@000000?ooo`3oool0oooo0P00000I0?ooo`030000003oool00000
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0200000080oooo0P00000;0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02P3oool2000000030?ooo`0000000000
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02000001X0oooo0P0000020?oo
o`030000003oool0oooo02H0oooo000F0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0>@3oool4000000040?ooo`0000000000000001@0oooo0`0000020?ooo`<000001P3oool30000
00030?ooo`000000000000800000303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00/0oooo0`0000000`3oool0000000000002000000H0oooo0`0000000`3oool00000000000020000
01<0oooo0`0000020?ooo`<00000:03oool001H0oooo0`00002V0?ooo`@00000@03oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo02d0oooo000H0?ooo`030000003oool0oooo0:<0oooo
00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool02@3oool3000000l0oooo1000000:0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0/0?ooo`006@3oool00`000000oooo0?oo
o`2R0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool04`3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0]0?ooo`00603oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo0440oooo0`00000D
0?ooo`<000003`3oool4000000X0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`09
0?ooo`<000003`3oool4000001<0oooo0`00000]0?ooo`005P3oool300000:H0oooo1000000n0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0:P3oool001H0oooo00<000000?ooo`3o
ool0Z03oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo0`0000020?ooo`<0
00001P3oool3000000030?ooo`000000000000800000303oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0Y0?ooo`00
4P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0:L0oooo00D000000?ooo`3oool0
oooo0000000k0?ooo`030000003oool0oooo00X0oooo0P00000;0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool02P3oool2000000040?ooo`000000oooo00000080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0200000080oooo0P00000;0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool02P3oool2000000`0oooo00<000000?ooo`3oool0:P3oool00140
oooo0P0000030?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool00P00000l0?ooo`<0
00001P3oool4000000040?ooo`0000000000000000H0oooo0`00000?0?ooo`@000001@3oool40000
00040?ooo`0000000000000000@0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3o
ool300000080oooo0`0000060?ooo`<000003`3oool4000000D0oooo10000000103oool000000000
000000060?ooo`<00000;@3oool000l0oooo0`0000000`3oool000000000000200000:/0oooo0`00
00000`3oool0000000000002000003T0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`030?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool0;@3oool000l0oooo00<000000?ooo`3oool0]P3o
ool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
000000040?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000000A0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool00000
00@0oooo0`0000000`3oool0000000000002000000<0oooo00<000000?ooo`3oool00P3oool01@00
0000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003o
ool0oooo0?ooo`000000;P3oool00100oooo00<000000?ooo`3oool0]03oool00`000000oooo0?oo
o`0i0?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`040000
003oool0oooo0?ooo`8000000P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0080
00000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`80000000<0oooo0000
000000000P3oool200000080oooo00<000000?ooo`3oool00`3oool01P000000oooo0?ooo`000000
oooo00000080oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`02000000<0oooo00<0
00000?ooo`3oool00P3oool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3oool0
;@3oool000l0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0i0?ooo`@0000000@0
oooo00000000000000001P3oool300000080oooo0`00000D0?ooo`<0000000@0oooo000000000000
00001P3oool4000000030?ooo`000000000000@0000000@0oooo00000000000000001P3oool30000
0080oooo0`00000D0?ooo`<0000000@0oooo00000000000000001P3oool4000000040?ooo`000000
0000000002l0oooo000?0?ooo`<00000]03oool4000007L0oooo00<000000?ooo`000000K03oool0
0180oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`1i0?ooo`800000K@3oool00140
oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`1h0?ooo`H00000JP3oool000l0oooo
0`00002d0?ooo`@00000M@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06P0oooo
000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo06L0oooo000@0?ooo`030000003oool0oooo0;@0oooo00<00000
0?ooo`3oool0M@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06P0oooo000?0?oo
o`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0M@3oool6000006X0oooo000?0?ooo`<0
000000<0oooo0000000000000P0000060?ooo`<0000000<0oooo0000000000000P00002?0?ooo`<0
00000P3oool3000000H0oooo0`0000000`3oool0000000000002000007L0oooo0P00001]0?ooo`00
4@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02000000<0oooo
00<000000?ooo`3oool0S03oool01P000000oooo0?ooo`000000oooo00000080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`02000007/0oooo00<000000?ooo`000000K03oool00180
oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`2;0?ooo`050000003oool0oooo0?ooo`000000103oool0
0`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000N`3oool2000006d0oooo000F
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`2<
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1g
0?ooo`@0000000<0oooo00000000000010000000103oool0000000000000001V0?ooo`005P3oool3
000000030?ooo`0000000000008000001P3oool3000008/0oooo0`0000060?ooo`<0000000<0oooo
0000000000000P00001g0?ooo`040000003oool0oooo0?ooo`8000001@3oool200000080oooo00<0
00000?ooo`3oool0I03oool001P0oooo0P00000<0?ooo`030000003oool0oooo08P0oooo00<00000
0?ooo`3oool02@3oool2000007/0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`04
0?ooo`050000003oool0oooo0?ooo`000000I@3oool001T0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`260?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0N@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo06@0oooo000V0?ooo`030000003oool0oooo
08P0oooo00<000000?ooo`3oool0Q@3oool3000000d0oooo0`00001T0?ooo`00903oool3000008/0
oooo0`0000230?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0H@3oool002@0oooo
00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0H03oool002D0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?oo
o`220?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0H@3oool002@0oooo00<00000
0?ooo`3oool0S03oool00`000000oooo0?ooo`230?ooo`<000003@3oool3000006@0oooo000T0?oo
o`<0000000@0oooo0000000000000000PP3oool300000080oooo0`0000250?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0I03oool002H0oooo0P0000030?ooo`030000003oool0oooo
07h0oooo00H000000?ooo`3oool000000?ooo`0000280?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000006D0oooo000W0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0O@3oool01@000000oooo0?ooo`3oool0000008T0oooo
00@000000?ooo`3oool0oooo0P0000050?ooo`8000000P3oool00`000000oooo0?ooo`1T0?ooo`00
:`3oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo08`0oooo100000000`3oool00000
00000004000000040?ooo`0000000000000006H0oooo000Z0?ooo`@00000O@3oool3000009<0oooo
0P00001]0?ooo`00;@3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo09@0oooo00<0
00000?ooo`000000K03oool002h0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`2D
0?ooo`800000K@3oool002d0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`2C0?oo
o`H00000JP3oool002X0oooo1000001m0?ooo`<00000T@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo06P0oooo000[0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0
T03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo06L0oooo000W0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0O@3oool01@000000oooo0?ooo`3oool0000008h0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1X0?ooo`009P3oool2000000<0oooo00<0
00000?ooo`3oool0OP3oool01P000000oooo0?ooo`000000oooo000008h0oooo1P00001Z0?ooo`00
903oool3000000040?ooo`000000000000000880oooo0`0000020?ooo`<00000S03oool2000006d0
oooo000T0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0R`3oool00`000000oooo
0000001/0?ooo`009@3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo08/0oooo0P00
001]0?ooo`00903oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo07P0oooo0`000000
103oool000000000000000060?ooo`@0000000<0oooo00000000000010000000103oool000000000
000000060?ooo`<000000P3oool3000005P0oooo000T0?ooo`<00000R`3oool3000007P0oooo00<0
00000?ooo`3oool00P0000030?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo
0P0000050?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`060000003oool0oooo0000003o
ool000000P3oool00`000000oooo0?ooo`1F0?ooo`009P3oool00`000000oooo0?ooo`280?ooo`03
0000003oool0oooo07P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo
0?ooo`000000103oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0E@3o
ool001T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`260?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0EP3oool001P0oooo0P00000<0?ooo`030000003oool0oooo
08P0oooo00<000000?ooo`3oool02@3oool2000006`0oooo100000050?ooo`@0000000@0oooo0000
0000000000003@3oool300000080oooo0`0000060?ooo`<00000EP3oool001H0oooo0`0000000`3o
ool0000000000002000000H0oooo0`00002;0?ooo`<000001P3oool3000000030?ooo`0000000000
00800000I`3oool00`000000oooo0?ooo`0:0?ooo`8000004`3oool2000000/0oooo00<000000?oo
o`3oool0E03oool001H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo08`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo06H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0D`3oool00180oooo00<000000?ooo`3oool0
0P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`2;0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`020?oo
o`050000003oool0oooo0?ooo`000000I03oool00`000000oooo0?ooo`0/0?ooo`030000003oool0
oooo05@0oooo000A0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
008000000`3oool00`000000oooo0?ooo`2<0?ooo`060000003oool0oooo0000003oool000000P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00800000I@3oool4000002P0oooo0`00
001F0?ooo`003`3oool3000000030?ooo`0000000000008000001P3oool3000000030?ooo`000000
000000800000S`3oool300000080oooo0`0000060?ooo`<0000000<0oooo0000000000000P00001S
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0EP3oool000l0oooo00<000000?oo
o`3oool0]P3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo02T0oooo00<000000?oo
o`3oool0E@3oool00100oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool0EP3oool000l0oooo00<000000?ooo`3oool0
]P3oool00`000000oooo0?ooo`1L0?ooo`<0000000<0oooo0000000000000P00000X0?ooo`<00000
00<0oooo0000000000000P00001A0?ooo`003`3oool300000;@0oooo1000001L0?ooo`030000003o
ool0oooo00800000;P3oool2000000<0oooo00<000000?ooo`3oool0C`3oool00140oooo00<00000
0?ooo`3oool0/@3oool00`000000oooo0?ooo`1N0?ooo`050000003oool0oooo0?ooo`000000;`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04h0oooo000B0?ooo`030000003oool0
oooo0;00oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0
oooo04l0oooo000A0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0G03oool40000
03H0oooo0`00001?0?ooo`003`3oool300000;@0oooo1000001I0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool0C@3oool000l0oooo00<000000?ooo`3oool0]P3oool00`000000oooo
0?ooo`1H0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0C03oool00100oooo00<0
00000?ooo`3oool0]03oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo03X0oooo00<0
00000?ooo`3oool0C@3oool000l0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`1I
0?ooo`@00000=P3oool3000004l0oooo000?0?ooo`<0000000<0oooo0000000000000P00002[0?oo
o`<0000000<0oooo0000000000000P00001L0?ooo`030000003oool0oooo03D0oooo00<000000?oo
o`3oool0C`3oool00140oooo0P0000030?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool00P00001P0?ooo`050000003oool0oooo0?ooo`000000;`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo04h0oooo000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0Y`3oool01@000000oooo0?ooo`3oool000000600oooo00<000000?ooo`3oool00P00000^0?oo
o`8000000`3oool00`000000oooo0?ooo`1?0?ooo`005P3oool00`000000oooo0?ooo`2X0?ooo`03
0000003oool0oooo06<0oooo0`0000000`3oool0000000000002000002P0oooo0`0000000`3oool0
00000000000200000540oooo000F0?ooo`<00000YP3oool4000006X0oooo00<000000?ooo`3oool0
9`3oool00`000000oooo0?ooo`1F0?ooo`00603oool00`000000oooo0?ooo`2S0?ooo`030000003o
ool0oooo06/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1E0?ooo`006@3oool0
0`000000oooo0?ooo`2R0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`1F0?ooo`00603oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo
06X0oooo1000000X0?ooo`<00000EP3oool001H0oooo0`00002V0?ooo`@00000I`3oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo05@0oooo000F0?ooo`030000003oool0oooo0:P0oooo
00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1C0?ooo`004P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo0:L0oooo00D000000?ooo`3oool0oooo0000001T0?ooo`030000
003oool0oooo00X0oooo0P00000C0?ooo`8000002`3oool00`000000oooo0?ooo`1D0?ooo`004@3o
ool2000000<0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`02000006D0oooo1000
00050?ooo`@0000000@0oooo00000000000000003@3oool300000080oooo0`0000060?ooo`<00000
EP3oool000l0oooo0`0000000`3oool000000000000200000:/0oooo0`0000000`3oool000000000
0002000006<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo05H0oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0H@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000
00040?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000040?ooo`050000
003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`1E0?ooo`00403oool00`000000oooo
0?ooo`2d0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool00P0000030?ooo`030000
003oool0oooo0080oooo00@000000?ooo`3oool0oooo0P0000050?ooo`8000000P3oool00`000000
oooo0?ooo`030?ooo`060000003oool0oooo0000003oool000000P3oool00`000000oooo0?ooo`1F
0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo06<0oooo0`000000103o
ool000000000000000060?ooo`@0000000<0oooo00000000000010000000103oool0000000000000
00060?ooo`<000000P3oool3000005P0oooo000?0?ooo`<00000]03oool4000007L0oooo0P00001]
0?ooo`004@3oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo07T0oooo00<000000?oo
o`000000K03oool00180oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`1i0?ooo`80
0000K@3oool00140oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`1h0?ooo`H00000
JP3oool000l0oooo0`00002d0?ooo`@00000M@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo06P0oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0M03oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo06L0oooo000@0?ooo`030000003oool0oooo
0;@0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
06P0oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0M@3oool6000006X0
oooo000?0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<0000000<0oooo000000000000
0P00000C0?ooo`<000000P3oool3000000H0oooo0`0000000`3oool0000000000002000003d0oooo
0`0000000`3oool0000000000002000000D0oooo10000000103oool0000000000000000D0?ooo`<0
00000P3oool3000000H0oooo0`0000000`3oool0000000000002000007L0oooo0P00001]0?ooo`00
4@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02000000<0oooo
00<000000?ooo`3oool0403oool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3oool0>P3oool00`000000
oooo0?ooo`02000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool20000
0080oooo00<000000?ooo`3oool04@3oool01P000000oooo0?ooo`000000oooo00000080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`02000007/0oooo00<000000?ooo`000000K03o
ool00180oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00D0
00000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000103oool00`000000
oooo0?ooo`0i0?ooo`050000003oool0oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0
000000@0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000
103oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000N`3oool2000006d0
oooo000F0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0M`3oool4000000030?oo
o`000000000000@0000000@0oooo0000000000000000IP3oool001H0oooo0`0000000`3oool00000
00000002000000H0oooo0`00000?0?ooo`<000001P3oool300000080oooo0`0000060?ooo`<00000
>03oool4000000H0oooo0`000000103oool000000000000000060?ooo`@000003`3oool3000000H0
oooo0`0000000`3oool0000000000002000007L0oooo00@000000?ooo`3oool0oooo0P0000050?oo
o`8000000P3oool00`000000oooo0?ooo`1T0?ooo`00603oool2000000`0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`090?ooo`030000003oool0000000/0oooo00<000000?ooo`3o
ool0=@3oool00`000000oooo0?ooo`0:0?ooo`800000303oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo00T0oooo0P00001k0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0103oool01@000000oooo0?ooo`3oool0000006D0oooo000I0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo07T0
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`1T0?ooo`009P3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0=@3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0Q@3oool3000000d0oooo
0`00001T0?ooo`00903oool3000000l0oooo0`00000D0?ooo`<00000>03oool4000001<0oooo1000
000?0?ooo`<00000P`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0640oooo000T
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0j
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`22
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0H03oool002D0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo03T0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0880oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`1Q0?ooo`00903oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool0P`3oool3000000d0oooo0`00001T0?ooo`00
903oool3000000040?ooo`0000000000000000H0oooo10000000103oool0000000000000000D0?oo
o`<0000000<0oooo0000000000000P00000_0?ooo`<0000000<0oooo0000000000000P00000C0?oo
o`@0000000@0oooo00000000000000001P3oool300000080oooo0`0000250?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0I03oool002H0oooo0P0000030?ooo`030000003oool0oooo
0080oooo00@000000?ooo`3oool0oooo0P00000I0?ooo`8000000`3oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo008000006P3oool200000080oooo00<000000?ooo`3oool00`3oool0
1P000000oooo0?ooo`000000oooo000008P0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`040?ooo`050000003oool0oooo0?ooo`000000I@3oool002L0oooo00<000000?ooo`3oool0
0P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo02/0oooo00D000000?ooo`3oool0oooo0000000K0?oo
o`050000003oool0oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0000008T0oooo00@0
00000?ooo`3oool0oooo0P0000050?ooo`8000000P3oool00`000000oooo0?ooo`1T0?ooo`00:`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0;03o
ool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0S03o
ool4000000030?ooo`000000000000@0000000@0oooo0000000000000000IP3oool002X0oooo1000
0000103oool0000000000000000R0?ooo`<00000:P3oool400000240oooo0`0000020?ooo`<00000
T`3oool2000006d0oooo000]0?ooo`8000009`3oool00`000000oooo0?ooo`0W0?ooo`030000003o
ool0oooo02H0oooo0P00002F0?ooo`030000003oool0000006`0oooo000^0?ooo`030000003oool0
oooo02H0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0
oooo09@0oooo0P00001]0?ooo`00EP3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo
0;d0oooo1P00001Z0?ooo`00E03oool3000002X0oooo1000002j0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0J03oool005@0oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`2i0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0I`3oool00500oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0[0?ooo`050000003oool0oooo0?ooo`000000
]`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06P0oooo001?0?ooo`8000000`3o
ool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00800000^03oool6000006X0oooo001=
0?ooo`<0000000<0oooo0000000000000P00000_0?ooo`<0000000<0oooo0000000000000P00002e
0?ooo`800000K@3oool004d0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`2d0?oo
o`030000003oool0000006`0oooo001>0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3o
ool0]03oool2000006d0oooo001=0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
]@3oool00`000000oooo0?ooo`1[0?ooo`00C@3oool3000003P0oooo1000003o0?ooobD0oooo001?
0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0o`3ooolW0?ooo`00D03oool00`00
0000oooo0?ooo`0d0?ooo`030000003oool0oooo0?l0oooo9`3oool004l0oooo00<000000?ooo`3o
ool0=@3oool00`000000oooo0?ooo`3o0?ooobL0oooo001=0?ooo`<00000>03oool400000?l0oooo
9@3oool004d0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`3o0?ooob@0oooo001>
0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0o`3ooolT0?ooo`00C@3oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo0?l0oooo903oool004d0oooo0`0000000`3oool0
000000000002000000H0oooo0`000000103oool0000000000000000D0?ooo`<000000P3oool30000
00H0oooo0`0000000`3oool000000000000200000?l0oooo9@3oool004l0oooo0P0000030?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P0000030?ooo`030000003oool0oooo0100
oooo00H000000?ooo`3oool000000?ooo`0000020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P00003o0?ooobT0oooo001@0?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool0
1@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?oo
o`3oool000000?l0oooo:@3oool005@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0?l0oooo:`3oool005@0oooo0`0000000`3oool0000000000002
000000D0oooo1000000?0?ooo`<000001P3oool3000000030?ooo`000000000000800000o`3oool/
0?ooo`00EP3oool2000000`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`090?oo
o`800000o`3oool`0?ooo`00E`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`3o0?ooobh0oooo001T0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool0o`3ooolk0?ooo`00H@3oool4000000l0oooo0`00
003o0?ooocX0oooo001R0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3oooli
0?ooo`00H`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0oooo>@3oool00680
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`3o0?ooocT0oooo001Q0?ooo`@00000
00@0oooo00000000000000001P3oool300000080oooo0`00003o0?ooocX0oooo001T0?ooo`800000
0P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`800000o`3ooolm0?ooo`00
I@3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`3o0?oooc/0oooo001X0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
o`3ooom00?ooo`00J03oool4000000040?ooo`000000000000000?l0oooo@@3oool006/0oooo0P00
003o0?oood@0oooo001/0?ooo`030000003oool0oooo0?l0oooo@P3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {236.438, 0}} -> {-0.0524626, -0.0274986, \
0.00732573, 0.00732573}, {{10.6875, 205.75}, {230.75, 5.625}} -> {-0.0828296, \
-0.34818, 0.00807849, 0.00807849}, {{225.25, 420.25}, {230.75, 5.625}} -> \
{-1.2379, -0.203722, 0.008077, 0.008077}}]
}, Closed]],

Cell["Similar figures based upon a square are shown below.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
        GraphicsArray[{DisplayKoch[square, 4, 
              DisplayFunction \[Rule] Identity], 
            DisplayKoch[square, 4, KochType \[Rule] "\<sieve\>", 
              DisplayFunction \[Rule] Identity]}], 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .47619 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0113379 0.47619 [
[ 0 0 0 0 ]
[ 1 .47619 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .47619 L
0 .47619 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0113379 0.477324 0.464853 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0238095 0.952381 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .02381 m
.03557 .02381 L
s
.03557 .02381 m
.04145 .03399 L
s
.04145 .03399 m
.04733 .02381 L
s
.04733 .02381 m
.05908 .02381 L
s
.05908 .02381 m
.06496 .03399 L
s
.06496 .03399 m
.05908 .04417 L
s
.05908 .04417 m
.07084 .04417 L
s
.07084 .04417 m
.07672 .05436 L
s
.07672 .05436 m
.0826 .04417 L
s
.0826 .04417 m
.09436 .04417 L
s
.09436 .04417 m
.08848 .03399 L
s
.08848 .03399 m
.09436 .02381 L
s
.09436 .02381 m
.10611 .02381 L
s
.10611 .02381 m
.11199 .03399 L
s
.11199 .03399 m
.11787 .02381 L
s
.11787 .02381 m
.12963 .02381 L
s
.12963 .02381 m
.13551 .03399 L
s
.13551 .03399 m
.12963 .04417 L
s
.12963 .04417 m
.14139 .04417 L
s
.14139 .04417 m
.14727 .05436 L
s
.14727 .05436 m
.14139 .06454 L
s
.14139 .06454 m
.12963 .06454 L
s
.12963 .06454 m
.13551 .07472 L
s
.13551 .07472 m
.12963 .0849 L
s
.12963 .0849 m
.14139 .0849 L
s
.14139 .0849 m
.14727 .09509 L
s
.14727 .09509 m
.15315 .0849 L
s
.15315 .0849 m
.1649 .0849 L
s
.1649 .0849 m
.17078 .09509 L
s
.17078 .09509 m
.1649 .10527 L
s
.1649 .10527 m
.17666 .10527 L
s
.17666 .10527 m
.18254 .11545 L
s
.18254 .11545 m
.18842 .10527 L
s
.18842 .10527 m
.20018 .10527 L
s
.20018 .10527 m
.1943 .09509 L
s
.1943 .09509 m
.20018 .0849 L
s
.20018 .0849 m
.21193 .0849 L
s
.21193 .0849 m
.21781 .09509 L
s
.21781 .09509 m
.22369 .0849 L
s
.22369 .0849 m
.23545 .0849 L
s
.23545 .0849 m
.22957 .07472 L
s
.22957 .07472 m
.23545 .06454 L
s
.23545 .06454 m
.22369 .06454 L
s
.22369 .06454 m
.21781 .05436 L
s
.21781 .05436 m
.22369 .04417 L
s
.22369 .04417 m
.23545 .04417 L
s
.23545 .04417 m
.22957 .03399 L
s
.22957 .03399 m
.23545 .02381 L
s
.23545 .02381 m
.24721 .02381 L
s
.24721 .02381 m
.25309 .03399 L
s
.25309 .03399 m
.25897 .02381 L
s
.25897 .02381 m
.27072 .02381 L
s
.27072 .02381 m
.2766 .03399 L
s
.2766 .03399 m
.27072 .04417 L
s
.27072 .04417 m
.28248 .04417 L
s
.28248 .04417 m
.28836 .05436 L
s
.28836 .05436 m
.29424 .04417 L
s
.29424 .04417 m
.306 .04417 L
s
.306 .04417 m
.30012 .03399 L
s
.30012 .03399 m
.306 .02381 L
s
.306 .02381 m
.31775 .02381 L
s
.31775 .02381 m
.32363 .03399 L
s
.32363 .03399 m
.32951 .02381 L
s
.32951 .02381 m
.34127 .02381 L
s
.34127 .02381 m
.34715 .03399 L
s
.34715 .03399 m
.34127 .04417 L
s
.34127 .04417 m
.35303 .04417 L
s
.35303 .04417 m
.35891 .05436 L
s
.35891 .05436 m
.35303 .06454 L
s
.35303 .06454 m
.34127 .06454 L
s
.34127 .06454 m
.34715 .07472 L
s
.34715 .07472 m
.34127 .0849 L
s
.34127 .0849 m
.35303 .0849 L
s
.35303 .0849 m
.35891 .09509 L
s
.35891 .09509 m
.36479 .0849 L
s
.36479 .0849 m
.37654 .0849 L
s
.37654 .0849 m
.38242 .09509 L
s
.38242 .09509 m
.37654 .10527 L
s
.37654 .10527 m
.3883 .10527 L
s
.3883 .10527 m
.39418 .11545 L
s
.39418 .11545 m
.3883 .12563 L
s
.3883 .12563 m
.37654 .12563 L
s
.37654 .12563 m
.38242 .13582 L
s
.38242 .13582 m
.37654 .146 L
s
.37654 .146 m
.36479 .146 L
s
.36479 .146 m
.35891 .13582 L
s
.35891 .13582 m
.35303 .146 L
s
.35303 .146 m
.34127 .146 L
s
.34127 .146 m
.34715 .15618 L
s
.34715 .15618 m
.34127 .16637 L
s
.34127 .16637 m
.35303 .16637 L
s
.35303 .16637 m
.35891 .17655 L
s
.35891 .17655 m
.35303 .18673 L
s
.35303 .18673 m
.34127 .18673 L
s
.34127 .18673 m
.34715 .19691 L
s
.34715 .19691 m
.34127 .2071 L
s
.34127 .2071 m
.35303 .2071 L
s
.35303 .2071 m
.35891 .21728 L
s
.35891 .21728 m
.36479 .2071 L
s
.36479 .2071 m
.37654 .2071 L
s
.37654 .2071 m
.38242 .21728 L
s
.38242 .21728 m
.37654 .22746 L
s
.37654 .22746 m
.3883 .22746 L
s
.3883 .22746 m
.39418 .23764 L
s
.39418 .23764 m
.40006 .22746 L
s
.40006 .22746 m
.41182 .22746 L
s
.41182 .22746 m
.40594 .21728 L
s
.40594 .21728 m
.41182 .2071 L
s
.41182 .2071 m
.42357 .2071 L
s
.42357 .2071 m
.42945 .21728 L
s
.42945 .21728 m
.43533 .2071 L
s
.43533 .2071 m
.44709 .2071 L
s
.44709 .2071 m
.45297 .21728 L
s
.45297 .21728 m
.44709 .22746 L
s
.44709 .22746 m
.45885 .22746 L
s
.45885 .22746 m
.46473 .23764 L
s
.46473 .23764 m
.45885 .24783 L
s
.45885 .24783 m
.44709 .24783 L
s
.44709 .24783 m
.45297 .25801 L
s
.45297 .25801 m
.44709 .26819 L
s
.44709 .26819 m
.45885 .26819 L
s
.45885 .26819 m
.46473 .27837 L
s
.46473 .27837 m
.47061 .26819 L
s
.47061 .26819 m
.48236 .26819 L
s
.48236 .26819 m
.48824 .27837 L
s
.48824 .27837 m
.48236 .28856 L
s
.48236 .28856 m
.49412 .28856 L
s
.49412 .28856 m
.5 .29874 L
s
.5 .29874 m
.50588 .28856 L
s
.50588 .28856 m
.51764 .28856 L
s
.51764 .28856 m
.51176 .27837 L
s
.51176 .27837 m
.51764 .26819 L
s
.51764 .26819 m
.52939 .26819 L
s
.52939 .26819 m
.53527 .27837 L
s
.53527 .27837 m
.54115 .26819 L
s
.54115 .26819 m
.55291 .26819 L
s
.55291 .26819 m
.54703 .25801 L
s
.54703 .25801 m
.55291 .24783 L
s
.55291 .24783 m
.54115 .24783 L
s
.54115 .24783 m
.53527 .23764 L
s
.53527 .23764 m
.54115 .22746 L
s
.54115 .22746 m
.55291 .22746 L
s
.55291 .22746 m
.54703 .21728 L
s
.54703 .21728 m
.55291 .2071 L
s
.55291 .2071 m
.56467 .2071 L
s
.56467 .2071 m
.57055 .21728 L
s
.57055 .21728 m
.57643 .2071 L
s
.57643 .2071 m
.58818 .2071 L
s
.58818 .2071 m
.59406 .21728 L
s
.59406 .21728 m
.58818 .22746 L
s
.58818 .22746 m
.59994 .22746 L
s
.59994 .22746 m
.60582 .23764 L
s
.60582 .23764 m
.6117 .22746 L
s
.6117 .22746 m
.62346 .22746 L
s
.62346 .22746 m
.61758 .21728 L
s
.61758 .21728 m
.62346 .2071 L
s
.62346 .2071 m
.63521 .2071 L
s
.63521 .2071 m
.64109 .21728 L
s
.64109 .21728 m
.64697 .2071 L
s
.64697 .2071 m
.65873 .2071 L
s
.65873 .2071 m
.65285 .19691 L
s
.65285 .19691 m
.65873 .18673 L
s
.65873 .18673 m
.64697 .18673 L
s
.64697 .18673 m
.64109 .17655 L
s
.64109 .17655 m
.64697 .16637 L
s
.64697 .16637 m
.65873 .16637 L
s
.65873 .16637 m
.65285 .15618 L
s
.65285 .15618 m
.65873 .146 L
s
.65873 .146 m
.64697 .146 L
s
.64697 .146 m
.64109 .13582 L
s
.64109 .13582 m
.63521 .146 L
s
.63521 .146 m
.62346 .146 L
s
.62346 .146 m
.61758 .13582 L
s
.61758 .13582 m
.62346 .12563 L
s
.62346 .12563 m
.6117 .12563 L
s
.6117 .12563 m
.60582 .11545 L
s
.60582 .11545 m
.6117 .10527 L
s
.6117 .10527 m
.62346 .10527 L
s
.62346 .10527 m
.61758 .09509 L
s
.61758 .09509 m
.62346 .0849 L
s
.62346 .0849 m
.63521 .0849 L
s
.63521 .0849 m
.64109 .09509 L
s
.64109 .09509 m
.64697 .0849 L
s
.64697 .0849 m
.65873 .0849 L
s
.65873 .0849 m
.65285 .07472 L
s
.65285 .07472 m
.65873 .06454 L
s
.65873 .06454 m
.64697 .06454 L
s
.64697 .06454 m
.64109 .05436 L
s
.64109 .05436 m
.64697 .04417 L
s
.64697 .04417 m
.65873 .04417 L
s
.65873 .04417 m
.65285 .03399 L
s
.65285 .03399 m
.65873 .02381 L
s
.65873 .02381 m
.67049 .02381 L
s
.67049 .02381 m
.67637 .03399 L
s
.67637 .03399 m
.68225 .02381 L
s
.68225 .02381 m
.694 .02381 L
s
.694 .02381 m
.69988 .03399 L
s
.69988 .03399 m
.694 .04417 L
s
.694 .04417 m
.70576 .04417 L
s
.70576 .04417 m
.71164 .05436 L
s
.71164 .05436 m
.71752 .04417 L
s
.71752 .04417 m
.72928 .04417 L
s
.72928 .04417 m
.7234 .03399 L
s
.7234 .03399 m
.72928 .02381 L
s
.72928 .02381 m
.74103 .02381 L
s
.74103 .02381 m
.74691 .03399 L
s
.74691 .03399 m
.75279 .02381 L
s
.75279 .02381 m
.76455 .02381 L
s
.76455 .02381 m
.77043 .03399 L
s
.77043 .03399 m
.76455 .04417 L
s
.76455 .04417 m
.77631 .04417 L
s
.77631 .04417 m
.78219 .05436 L
s
.78219 .05436 m
.77631 .06454 L
s
.77631 .06454 m
.76455 .06454 L
s
.76455 .06454 m
.77043 .07472 L
s
.77043 .07472 m
.76455 .0849 L
s
.76455 .0849 m
.77631 .0849 L
s
.77631 .0849 m
.78219 .09509 L
s
.78219 .09509 m
.78807 .0849 L
s
.78807 .0849 m
.79982 .0849 L
s
.79982 .0849 m
.8057 .09509 L
s
.8057 .09509 m
.79982 .10527 L
s
.79982 .10527 m
.81158 .10527 L
s
.81158 .10527 m
.81746 .11545 L
s
.81746 .11545 m
.82334 .10527 L
s
.82334 .10527 m
.8351 .10527 L
s
.8351 .10527 m
.82922 .09509 L
s
.82922 .09509 m
.8351 .0849 L
s
.8351 .0849 m
.84685 .0849 L
s
.84685 .0849 m
.85273 .09509 L
s
.85273 .09509 m
.85861 .0849 L
s
.85861 .0849 m
.87037 .0849 L
s
.87037 .0849 m
.86449 .07472 L
s
.86449 .07472 m
.87037 .06454 L
s
.87037 .06454 m
.85861 .06454 L
s
.85861 .06454 m
.85273 .05436 L
s
.85273 .05436 m
.85861 .04417 L
s
.85861 .04417 m
.87037 .04417 L
s
.87037 .04417 m
.86449 .03399 L
s
.86449 .03399 m
.87037 .02381 L
s
.87037 .02381 m
.88213 .02381 L
s
.88213 .02381 m
.88801 .03399 L
s
.88801 .03399 m
.89389 .02381 L
s
.89389 .02381 m
.90564 .02381 L
s
.90564 .02381 m
.91152 .03399 L
s
.91152 .03399 m
.90564 .04417 L
s
.90564 .04417 m
.9174 .04417 L
s
.9174 .04417 m
.92328 .05436 L
s
.92328 .05436 m
.92916 .04417 L
s
.92916 .04417 m
.94092 .04417 L
s
.94092 .04417 m
.93504 .03399 L
s
.93504 .03399 m
.94092 .02381 L
s
.94092 .02381 m
.95267 .02381 L
s
.95267 .02381 m
.95855 .03399 L
s
.95855 .03399 m
.96443 .02381 L
s
.96443 .02381 m
.97619 .02381 L
s
.97619 .02381 m
.97619 .03557 L
s
.97619 .03557 m
.96601 .04145 L
s
.96601 .04145 m
.97619 .04733 L
s
.97619 .04733 m
.97619 .05908 L
s
.97619 .05908 m
.96601 .06496 L
s
.96601 .06496 m
.95583 .05908 L
s
.95583 .05908 m
.95583 .07084 L
s
.95583 .07084 m
.94564 .07672 L
s
.94564 .07672 m
.95583 .0826 L
s
.95583 .0826 m
.95583 .09436 L
s
.95583 .09436 m
.96601 .08848 L
s
.96601 .08848 m
.97619 .09436 L
s
.97619 .09436 m
.97619 .10611 L
s
.97619 .10611 m
.96601 .11199 L
s
.96601 .11199 m
.97619 .11787 L
s
.97619 .11787 m
.97619 .12963 L
s
.97619 .12963 m
.96601 .13551 L
s
.96601 .13551 m
.95583 .12963 L
s
.95583 .12963 m
.95583 .14139 L
s
.95583 .14139 m
.94564 .14727 L
s
.94564 .14727 m
.93546 .14139 L
s
.93546 .14139 m
.93546 .12963 L
s
.93546 .12963 m
.92528 .13551 L
s
.92528 .13551 m
.9151 .12963 L
s
.9151 .12963 m
.9151 .14139 L
s
.9151 .14139 m
.90491 .14727 L
s
.90491 .14727 m
.9151 .15315 L
s
.9151 .15315 m
.9151 .1649 L
s
.9151 .1649 m
.90491 .17078 L
s
.90491 .17078 m
.89473 .1649 L
s
.89473 .1649 m
.89473 .17666 L
s
.89473 .17666 m
.88455 .18254 L
s
.88455 .18254 m
.89473 .18842 L
s
.89473 .18842 m
.89473 .20018 L
s
.89473 .20018 m
.90491 .1943 L
s
.90491 .1943 m
.9151 .20018 L
s
.9151 .20018 m
.9151 .21193 L
s
.9151 .21193 m
.90491 .21781 L
s
.90491 .21781 m
.9151 .22369 L
s
.9151 .22369 m
.9151 .23545 L
s
.9151 .23545 m
.92528 .22957 L
s
.92528 .22957 m
.93546 .23545 L
s
.93546 .23545 m
.93546 .22369 L
s
.93546 .22369 m
.94564 .21781 L
s
.94564 .21781 m
.95583 .22369 L
s
.95583 .22369 m
.95583 .23545 L
s
.95583 .23545 m
.96601 .22957 L
s
.96601 .22957 m
.97619 .23545 L
s
.97619 .23545 m
.97619 .24721 L
s
.97619 .24721 m
.96601 .25309 L
s
.96601 .25309 m
.97619 .25897 L
s
.97619 .25897 m
.97619 .27072 L
s
.97619 .27072 m
.96601 .2766 L
s
.96601 .2766 m
.95583 .27072 L
s
.95583 .27072 m
.95583 .28248 L
s
.95583 .28248 m
.94564 .28836 L
s
.94564 .28836 m
.95583 .29424 L
s
.95583 .29424 m
.95583 .306 L
s
.95583 .306 m
.96601 .30012 L
s
.96601 .30012 m
.97619 .306 L
s
.97619 .306 m
.97619 .31775 L
s
.97619 .31775 m
.96601 .32363 L
s
.96601 .32363 m
.97619 .32951 L
s
.97619 .32951 m
.97619 .34127 L
s
.97619 .34127 m
.96601 .34715 L
s
.96601 .34715 m
.95583 .34127 L
s
.95583 .34127 m
.95583 .35303 L
s
.95583 .35303 m
.94564 .35891 L
s
.94564 .35891 m
.93546 .35303 L
s
.93546 .35303 m
.93546 .34127 L
s
.93546 .34127 m
.92528 .34715 L
s
.92528 .34715 m
.9151 .34127 L
s
.9151 .34127 m
.9151 .35303 L
s
.9151 .35303 m
.90491 .35891 L
s
.90491 .35891 m
.9151 .36479 L
s
.9151 .36479 m
.9151 .37654 L
s
.9151 .37654 m
.90491 .38242 L
s
.90491 .38242 m
.89473 .37654 L
s
.89473 .37654 m
.89473 .3883 L
s
.89473 .3883 m
.88455 .39418 L
s
.88455 .39418 m
.87437 .3883 L
s
.87437 .3883 m
.87437 .37654 L
s
.87437 .37654 m
.86418 .38242 L
s
.86418 .38242 m
.854 .37654 L
s
.854 .37654 m
.854 .36479 L
s
.854 .36479 m
.86418 .35891 L
s
.86418 .35891 m
.854 .35303 L
s
.854 .35303 m
.854 .34127 L
s
.854 .34127 m
.84382 .34715 L
s
.84382 .34715 m
.83363 .34127 L
s
.83363 .34127 m
.83363 .35303 L
s
.83363 .35303 m
.82345 .35891 L
s
.82345 .35891 m
.81327 .35303 L
s
.81327 .35303 m
.81327 .34127 L
s
.81327 .34127 m
.80309 .34715 L
s
.80309 .34715 m
.7929 .34127 L
s
.7929 .34127 m
.7929 .35303 L
s
.7929 .35303 m
.78272 .35891 L
s
.78272 .35891 m
.7929 .36479 L
s
.7929 .36479 m
.7929 .37654 L
s
.7929 .37654 m
.78272 .38242 L
s
.78272 .38242 m
.77254 .37654 L
s
.77254 .37654 m
.77254 .3883 L
s
.77254 .3883 m
.76236 .39418 L
s
.76236 .39418 m
.77254 .40006 L
s
.77254 .40006 m
.77254 .41182 L
s
.77254 .41182 m
.78272 .40594 L
s
.78272 .40594 m
.7929 .41182 L
s
.7929 .41182 m
.7929 .42357 L
s
.7929 .42357 m
.78272 .42945 L
s
.78272 .42945 m
.7929 .43533 L
s
.7929 .43533 m
.7929 .44709 L
s
.7929 .44709 m
.78272 .45297 L
s
.78272 .45297 m
.77254 .44709 L
s
.77254 .44709 m
.77254 .45885 L
s
.77254 .45885 m
.76236 .46473 L
s
.76236 .46473 m
.75217 .45885 L
s
.75217 .45885 m
.75217 .44709 L
s
.75217 .44709 m
.74199 .45297 L
s
.74199 .45297 m
.73181 .44709 L
s
.73181 .44709 m
.73181 .45885 L
s
.73181 .45885 m
.72163 .46473 L
s
.72163 .46473 m
.73181 .47061 L
s
.73181 .47061 m
.73181 .48236 L
s
.73181 .48236 m
.72163 .48824 L
s
.72163 .48824 m
.71144 .48236 L
s
.71144 .48236 m
.71144 .49412 L
s
.71144 .49412 m
.70126 .5 L
s
.70126 .5 m
.71144 .50588 L
s
.71144 .50588 m
.71144 .51764 L
s
.71144 .51764 m
.72163 .51176 L
s
.72163 .51176 m
.73181 .51764 L
s
.73181 .51764 m
.73181 .52939 L
s
.73181 .52939 m
.72163 .53527 L
s
.72163 .53527 m
.73181 .54115 L
s
.73181 .54115 m
.73181 .55291 L
s
.73181 .55291 m
.74199 .54703 L
s
.74199 .54703 m
.75217 .55291 L
s
.75217 .55291 m
.75217 .54115 L
s
.75217 .54115 m
.76236 .53527 L
s
.76236 .53527 m
.77254 .54115 L
s
.77254 .54115 m
.77254 .55291 L
s
.77254 .55291 m
.78272 .54703 L
s
.78272 .54703 m
.7929 .55291 L
s
.7929 .55291 m
.7929 .56467 L
s
.7929 .56467 m
.78272 .57055 L
s
.78272 .57055 m
.7929 .57643 L
s
.7929 .57643 m
.7929 .58818 L
s
.7929 .58818 m
.78272 .59406 L
s
.78272 .59406 m
.77254 .58818 L
s
.77254 .58818 m
.77254 .59994 L
s
.77254 .59994 m
.76236 .60582 L
s
.76236 .60582 m
.77254 .6117 L
s
.77254 .6117 m
.77254 .62346 L
s
.77254 .62346 m
.78272 .61758 L
s
.78272 .61758 m
.7929 .62346 L
s
.7929 .62346 m
.7929 .63521 L
s
.7929 .63521 m
.78272 .64109 L
s
.78272 .64109 m
.7929 .64697 L
s
.7929 .64697 m
.7929 .65873 L
s
.7929 .65873 m
.80309 .65285 L
s
.80309 .65285 m
.81327 .65873 L
s
.81327 .65873 m
.81327 .64697 L
s
.81327 .64697 m
.82345 .64109 L
s
.82345 .64109 m
.83363 .64697 L
s
.83363 .64697 m
.83363 .65873 L
s
.83363 .65873 m
.84382 .65285 L
s
.84382 .65285 m
.854 .65873 L
s
.854 .65873 m
.854 .64697 L
s
.854 .64697 m
.86418 .64109 L
s
.86418 .64109 m
.854 .63521 L
s
.854 .63521 m
.854 .62346 L
s
.854 .62346 m
.86418 .61758 L
s
.86418 .61758 m
.87437 .62346 L
s
.87437 .62346 m
.87437 .6117 L
s
.87437 .6117 m
.88455 .60582 L
s
.88455 .60582 m
.89473 .6117 L
s
.89473 .6117 m
.89473 .62346 L
s
.89473 .62346 m
.90491 .61758 L
s
.90491 .61758 m
.9151 .62346 L
s
.9151 .62346 m
.9151 .63521 L
s
.9151 .63521 m
.90491 .64109 L
s
.90491 .64109 m
.9151 .64697 L
s
.9151 .64697 m
.9151 .65873 L
s
.9151 .65873 m
.92528 .65285 L
s
.92528 .65285 m
.93546 .65873 L
s
.93546 .65873 m
.93546 .64697 L
s
.93546 .64697 m
.94564 .64109 L
s
.94564 .64109 m
.95583 .64697 L
s
.95583 .64697 m
.95583 .65873 L
s
.95583 .65873 m
.96601 .65285 L
s
.96601 .65285 m
.97619 .65873 L
s
.97619 .65873 m
.97619 .67049 L
s
.97619 .67049 m
.96601 .67637 L
s
.96601 .67637 m
.97619 .68225 L
s
.97619 .68225 m
.97619 .694 L
s
.97619 .694 m
.96601 .69988 L
s
.96601 .69988 m
.95583 .694 L
s
.95583 .694 m
.95583 .70576 L
s
.95583 .70576 m
.94564 .71164 L
s
.94564 .71164 m
.95583 .71752 L
s
.95583 .71752 m
.95583 .72928 L
s
.95583 .72928 m
.96601 .7234 L
s
.96601 .7234 m
.97619 .72928 L
s
.97619 .72928 m
.97619 .74103 L
s
.97619 .74103 m
.96601 .74691 L
s
.96601 .74691 m
.97619 .75279 L
s
.97619 .75279 m
.97619 .76455 L
s
.97619 .76455 m
.96601 .77043 L
s
.96601 .77043 m
.95583 .76455 L
s
.95583 .76455 m
.95583 .77631 L
s
.95583 .77631 m
.94564 .78219 L
s
.94564 .78219 m
.93546 .77631 L
s
.93546 .77631 m
.93546 .76455 L
s
.93546 .76455 m
.92528 .77043 L
s
.92528 .77043 m
.9151 .76455 L
s
.9151 .76455 m
.9151 .77631 L
s
.9151 .77631 m
.90491 .78219 L
s
.90491 .78219 m
.9151 .78807 L
s
.9151 .78807 m
.9151 .79982 L
s
.9151 .79982 m
.90491 .8057 L
s
.90491 .8057 m
.89473 .79982 L
s
.89473 .79982 m
.89473 .81158 L
s
.89473 .81158 m
.88455 .81746 L
s
.88455 .81746 m
.89473 .82334 L
s
.89473 .82334 m
.89473 .8351 L
s
.89473 .8351 m
.90491 .82922 L
s
.90491 .82922 m
.9151 .8351 L
s
.9151 .8351 m
.9151 .84685 L
s
.9151 .84685 m
.90491 .85273 L
s
.90491 .85273 m
.9151 .85861 L
s
.9151 .85861 m
.9151 .87037 L
s
.9151 .87037 m
.92528 .86449 L
s
.92528 .86449 m
.93546 .87037 L
s
.93546 .87037 m
.93546 .85861 L
s
.93546 .85861 m
.94564 .85273 L
s
.94564 .85273 m
.95583 .85861 L
s
.95583 .85861 m
.95583 .87037 L
s
.95583 .87037 m
.96601 .86449 L
s
.96601 .86449 m
.97619 .87037 L
s
.97619 .87037 m
.97619 .88213 L
s
.97619 .88213 m
.96601 .88801 L
s
.96601 .88801 m
.97619 .89389 L
s
.97619 .89389 m
.97619 .90564 L
s
.97619 .90564 m
.96601 .91152 L
s
.96601 .91152 m
.95583 .90564 L
s
.95583 .90564 m
.95583 .9174 L
s
.95583 .9174 m
.94564 .92328 L
s
.94564 .92328 m
.95583 .92916 L
s
.95583 .92916 m
.95583 .94092 L
s
.95583 .94092 m
.96601 .93504 L
s
.96601 .93504 m
.97619 .94092 L
s
.97619 .94092 m
.97619 .95267 L
s
.97619 .95267 m
.96601 .95855 L
s
.96601 .95855 m
.97619 .96443 L
s
.97619 .96443 m
.97619 .97619 L
s
.97619 .97619 m
.96443 .97619 L
s
.96443 .97619 m
.95855 .96601 L
s
.95855 .96601 m
.95267 .97619 L
s
.95267 .97619 m
.94092 .97619 L
s
.94092 .97619 m
.93504 .96601 L
s
.93504 .96601 m
.94092 .95583 L
s
.94092 .95583 m
.92916 .95583 L
s
.92916 .95583 m
.92328 .94564 L
s
.92328 .94564 m
.9174 .95583 L
s
.9174 .95583 m
.90564 .95583 L
s
.90564 .95583 m
.91152 .96601 L
s
.91152 .96601 m
.90564 .97619 L
s
.90564 .97619 m
.89389 .97619 L
s
.89389 .97619 m
.88801 .96601 L
s
.88801 .96601 m
.88213 .97619 L
s
.88213 .97619 m
.87037 .97619 L
s
.87037 .97619 m
.86449 .96601 L
s
.86449 .96601 m
.87037 .95583 L
s
.87037 .95583 m
.85861 .95583 L
s
.85861 .95583 m
.85273 .94564 L
s
.85273 .94564 m
.85861 .93546 L
s
.85861 .93546 m
.87037 .93546 L
s
.87037 .93546 m
.86449 .92528 L
s
.86449 .92528 m
.87037 .9151 L
s
.87037 .9151 m
.85861 .9151 L
s
.85861 .9151 m
.85273 .90491 L
s
.85273 .90491 m
.84685 .9151 L
s
.84685 .9151 m
.8351 .9151 L
s
.8351 .9151 m
.82922 .90491 L
s
.82922 .90491 m
.8351 .89473 L
s
.8351 .89473 m
.82334 .89473 L
s
.82334 .89473 m
.81746 .88455 L
s
.81746 .88455 m
.81158 .89473 L
s
.81158 .89473 m
.79982 .89473 L
s
.79982 .89473 m
.8057 .90491 L
s
.8057 .90491 m
.79982 .9151 L
s
.79982 .9151 m
.78807 .9151 L
s
.78807 .9151 m
.78219 .90491 L
s
.78219 .90491 m
.77631 .9151 L
s
.77631 .9151 m
.76455 .9151 L
s
.76455 .9151 m
.77043 .92528 L
s
.77043 .92528 m
.76455 .93546 L
s
.76455 .93546 m
.77631 .93546 L
s
.77631 .93546 m
.78219 .94564 L
s
.78219 .94564 m
.77631 .95583 L
s
.77631 .95583 m
.76455 .95583 L
s
.76455 .95583 m
.77043 .96601 L
s
.77043 .96601 m
.76455 .97619 L
s
.76455 .97619 m
.75279 .97619 L
s
.75279 .97619 m
.74691 .96601 L
s
.74691 .96601 m
.74103 .97619 L
s
.74103 .97619 m
.72928 .97619 L
s
.72928 .97619 m
.7234 .96601 L
s
.7234 .96601 m
.72928 .95583 L
s
.72928 .95583 m
.71752 .95583 L
s
.71752 .95583 m
.71164 .94564 L
s
.71164 .94564 m
.70576 .95583 L
s
.70576 .95583 m
.694 .95583 L
s
.694 .95583 m
.69988 .96601 L
s
.69988 .96601 m
.694 .97619 L
s
.694 .97619 m
.68225 .97619 L
s
.68225 .97619 m
.67637 .96601 L
s
.67637 .96601 m
.67049 .97619 L
s
.67049 .97619 m
.65873 .97619 L
s
.65873 .97619 m
.65285 .96601 L
s
.65285 .96601 m
.65873 .95583 L
s
.65873 .95583 m
.64697 .95583 L
s
.64697 .95583 m
.64109 .94564 L
s
.64109 .94564 m
.64697 .93546 L
s
.64697 .93546 m
.65873 .93546 L
s
.65873 .93546 m
.65285 .92528 L
s
.65285 .92528 m
.65873 .9151 L
s
.65873 .9151 m
.64697 .9151 L
s
.64697 .9151 m
.64109 .90491 L
s
.64109 .90491 m
.63521 .9151 L
s
.63521 .9151 m
.62346 .9151 L
s
.62346 .9151 m
.61758 .90491 L
s
.61758 .90491 m
.62346 .89473 L
s
.62346 .89473 m
.6117 .89473 L
s
.6117 .89473 m
.60582 .88455 L
s
.60582 .88455 m
.6117 .87437 L
s
.6117 .87437 m
.62346 .87437 L
s
.62346 .87437 m
.61758 .86418 L
s
.61758 .86418 m
.62346 .854 L
s
.62346 .854 m
.63521 .854 L
s
.63521 .854 m
.64109 .86418 L
s
.64109 .86418 m
.64697 .854 L
s
.64697 .854 m
.65873 .854 L
s
.65873 .854 m
.65285 .84382 L
s
.65285 .84382 m
.65873 .83363 L
s
.65873 .83363 m
.64697 .83363 L
s
.64697 .83363 m
.64109 .82345 L
s
.64109 .82345 m
.64697 .81327 L
s
.64697 .81327 m
.65873 .81327 L
s
.65873 .81327 m
.65285 .80309 L
s
.65285 .80309 m
.65873 .7929 L
s
.65873 .7929 m
.64697 .7929 L
s
.64697 .7929 m
.64109 .78272 L
s
.64109 .78272 m
.63521 .7929 L
s
.63521 .7929 m
.62346 .7929 L
s
.62346 .7929 m
.61758 .78272 L
s
.61758 .78272 m
.62346 .77254 L
s
.62346 .77254 m
.6117 .77254 L
s
.6117 .77254 m
.60582 .76236 L
s
.60582 .76236 m
.59994 .77254 L
s
.59994 .77254 m
.58818 .77254 L
s
.58818 .77254 m
.59406 .78272 L
s
.59406 .78272 m
.58818 .7929 L
s
.58818 .7929 m
.57643 .7929 L
s
.57643 .7929 m
.57055 .78272 L
s
.57055 .78272 m
.56467 .7929 L
s
.56467 .7929 m
.55291 .7929 L
s
.55291 .7929 m
.54703 .78272 L
s
.54703 .78272 m
.55291 .77254 L
s
.55291 .77254 m
.54115 .77254 L
s
.54115 .77254 m
.53527 .76236 L
s
.53527 .76236 m
.54115 .75217 L
s
.54115 .75217 m
.55291 .75217 L
s
.55291 .75217 m
.54703 .74199 L
s
.54703 .74199 m
.55291 .73181 L
s
.55291 .73181 m
.54115 .73181 L
s
.54115 .73181 m
.53527 .72163 L
s
.53527 .72163 m
.52939 .73181 L
s
.52939 .73181 m
.51764 .73181 L
s
.51764 .73181 m
.51176 .72163 L
s
.51176 .72163 m
.51764 .71144 L
s
.51764 .71144 m
.50588 .71144 L
s
.50588 .71144 m
.5 .70126 L
s
.5 .70126 m
.49412 .71144 L
s
.49412 .71144 m
.48236 .71144 L
s
.48236 .71144 m
.48824 .72163 L
s
.48824 .72163 m
.48236 .73181 L
s
.48236 .73181 m
.47061 .73181 L
s
.47061 .73181 m
.46473 .72163 L
s
.46473 .72163 m
.45885 .73181 L
s
.45885 .73181 m
.44709 .73181 L
s
.44709 .73181 m
.45297 .74199 L
s
.45297 .74199 m
.44709 .75217 L
s
.44709 .75217 m
.45885 .75217 L
s
.45885 .75217 m
.46473 .76236 L
s
.46473 .76236 m
.45885 .77254 L
s
.45885 .77254 m
.44709 .77254 L
s
.44709 .77254 m
.45297 .78272 L
s
.45297 .78272 m
.44709 .7929 L
s
.44709 .7929 m
.43533 .7929 L
s
.43533 .7929 m
.42945 .78272 L
s
.42945 .78272 m
.42357 .7929 L
s
.42357 .7929 m
.41182 .7929 L
s
.41182 .7929 m
.40594 .78272 L
s
.40594 .78272 m
.41182 .77254 L
s
.41182 .77254 m
.40006 .77254 L
s
.40006 .77254 m
.39418 .76236 L
s
.39418 .76236 m
.3883 .77254 L
s
.3883 .77254 m
.37654 .77254 L
s
.37654 .77254 m
.38242 .78272 L
s
.38242 .78272 m
.37654 .7929 L
s
.37654 .7929 m
.36479 .7929 L
s
.36479 .7929 m
.35891 .78272 L
s
.35891 .78272 m
.35303 .7929 L
s
.35303 .7929 m
.34127 .7929 L
s
.34127 .7929 m
.34715 .80309 L
s
.34715 .80309 m
.34127 .81327 L
s
.34127 .81327 m
.35303 .81327 L
s
.35303 .81327 m
.35891 .82345 L
s
.35891 .82345 m
.35303 .83363 L
s
.35303 .83363 m
.34127 .83363 L
s
.34127 .83363 m
.34715 .84382 L
s
.34715 .84382 m
.34127 .854 L
s
.34127 .854 m
.35303 .854 L
s
.35303 .854 m
.35891 .86418 L
s
.35891 .86418 m
.36479 .854 L
s
.36479 .854 m
.37654 .854 L
s
.37654 .854 m
.38242 .86418 L
s
.38242 .86418 m
.37654 .87437 L
s
.37654 .87437 m
.3883 .87437 L
s
.3883 .87437 m
.39418 .88455 L
s
.39418 .88455 m
.3883 .89473 L
s
.3883 .89473 m
.37654 .89473 L
s
.37654 .89473 m
.38242 .90491 L
s
.38242 .90491 m
.37654 .9151 L
s
.37654 .9151 m
.36479 .9151 L
s
.36479 .9151 m
.35891 .90491 L
s
.35891 .90491 m
.35303 .9151 L
s
.35303 .9151 m
.34127 .9151 L
s
.34127 .9151 m
.34715 .92528 L
s
.34715 .92528 m
.34127 .93546 L
s
.34127 .93546 m
.35303 .93546 L
s
.35303 .93546 m
.35891 .94564 L
s
.35891 .94564 m
.35303 .95583 L
s
.35303 .95583 m
.34127 .95583 L
s
.34127 .95583 m
.34715 .96601 L
s
.34715 .96601 m
.34127 .97619 L
s
.34127 .97619 m
.32951 .97619 L
s
.32951 .97619 m
.32363 .96601 L
s
.32363 .96601 m
.31775 .97619 L
s
.31775 .97619 m
.306 .97619 L
s
.306 .97619 m
.30012 .96601 L
s
.30012 .96601 m
.306 .95583 L
s
.306 .95583 m
.29424 .95583 L
s
.29424 .95583 m
.28836 .94564 L
s
.28836 .94564 m
.28248 .95583 L
s
.28248 .95583 m
.27072 .95583 L
s
.27072 .95583 m
.2766 .96601 L
s
.2766 .96601 m
.27072 .97619 L
s
.27072 .97619 m
.25897 .97619 L
s
.25897 .97619 m
.25309 .96601 L
s
.25309 .96601 m
.24721 .97619 L
s
.24721 .97619 m
.23545 .97619 L
s
.23545 .97619 m
.22957 .96601 L
s
.22957 .96601 m
.23545 .95583 L
s
.23545 .95583 m
.22369 .95583 L
s
.22369 .95583 m
.21781 .94564 L
s
.21781 .94564 m
.22369 .93546 L
s
.22369 .93546 m
.23545 .93546 L
s
.23545 .93546 m
.22957 .92528 L
s
.22957 .92528 m
.23545 .9151 L
s
.23545 .9151 m
.22369 .9151 L
s
.22369 .9151 m
.21781 .90491 L
s
.21781 .90491 m
.21193 .9151 L
s
.21193 .9151 m
.20018 .9151 L
s
.20018 .9151 m
.1943 .90491 L
s
.1943 .90491 m
.20018 .89473 L
s
.20018 .89473 m
.18842 .89473 L
s
.18842 .89473 m
.18254 .88455 L
s
.18254 .88455 m
.17666 .89473 L
s
.17666 .89473 m
.1649 .89473 L
s
.1649 .89473 m
.17078 .90491 L
s
.17078 .90491 m
.1649 .9151 L
s
.1649 .9151 m
.15315 .9151 L
s
.15315 .9151 m
.14727 .90491 L
s
.14727 .90491 m
.14139 .9151 L
s
.14139 .9151 m
.12963 .9151 L
s
.12963 .9151 m
.13551 .92528 L
s
.13551 .92528 m
.12963 .93546 L
s
.12963 .93546 m
.14139 .93546 L
s
.14139 .93546 m
.14727 .94564 L
s
.14727 .94564 m
.14139 .95583 L
s
.14139 .95583 m
.12963 .95583 L
s
.12963 .95583 m
.13551 .96601 L
s
.13551 .96601 m
.12963 .97619 L
s
.12963 .97619 m
.11787 .97619 L
s
.11787 .97619 m
.11199 .96601 L
s
.11199 .96601 m
.10611 .97619 L
s
.10611 .97619 m
.09436 .97619 L
s
.09436 .97619 m
.08848 .96601 L
s
.08848 .96601 m
.09436 .95583 L
s
.09436 .95583 m
.0826 .95583 L
s
.0826 .95583 m
.07672 .94564 L
s
.07672 .94564 m
.07084 .95583 L
s
.07084 .95583 m
.05908 .95583 L
s
.05908 .95583 m
.06496 .96601 L
s
.06496 .96601 m
.05908 .97619 L
s
.05908 .97619 m
.04733 .97619 L
s
.04733 .97619 m
.04145 .96601 L
s
.04145 .96601 m
.03557 .97619 L
s
.03557 .97619 m
.02381 .97619 L
s
.02381 .97619 m
.02381 .96443 L
s
.02381 .96443 m
.03399 .95855 L
s
.03399 .95855 m
.02381 .95267 L
s
.02381 .95267 m
.02381 .94092 L
s
.02381 .94092 m
.03399 .93504 L
s
.03399 .93504 m
.04417 .94092 L
s
.04417 .94092 m
.04417 .92916 L
s
.04417 .92916 m
.05436 .92328 L
s
.05436 .92328 m
.04417 .9174 L
s
.04417 .9174 m
.04417 .90564 L
s
.04417 .90564 m
.03399 .91152 L
s
.03399 .91152 m
.02381 .90564 L
s
.02381 .90564 m
.02381 .89389 L
s
.02381 .89389 m
.03399 .88801 L
s
.03399 .88801 m
.02381 .88213 L
s
.02381 .88213 m
.02381 .87037 L
s
.02381 .87037 m
.03399 .86449 L
s
.03399 .86449 m
.04417 .87037 L
s
.04417 .87037 m
.04417 .85861 L
s
.04417 .85861 m
.05436 .85273 L
s
.05436 .85273 m
.06454 .85861 L
s
.06454 .85861 m
.06454 .87037 L
s
.06454 .87037 m
.07472 .86449 L
s
.07472 .86449 m
.0849 .87037 L
s
.0849 .87037 m
.0849 .85861 L
s
.0849 .85861 m
.09509 .85273 L
s
.09509 .85273 m
.0849 .84685 L
s
.0849 .84685 m
.0849 .8351 L
s
.0849 .8351 m
.09509 .82922 L
s
.09509 .82922 m
.10527 .8351 L
s
.10527 .8351 m
.10527 .82334 L
s
.10527 .82334 m
.11545 .81746 L
s
.11545 .81746 m
.10527 .81158 L
s
.10527 .81158 m
.10527 .79982 L
s
.10527 .79982 m
.09509 .8057 L
s
.09509 .8057 m
.0849 .79982 L
s
.0849 .79982 m
.0849 .78807 L
s
.0849 .78807 m
.09509 .78219 L
s
.09509 .78219 m
.0849 .77631 L
s
.0849 .77631 m
.0849 .76455 L
s
.0849 .76455 m
.07472 .77043 L
s
.07472 .77043 m
.06454 .76455 L
s
.06454 .76455 m
.06454 .77631 L
s
.06454 .77631 m
.05436 .78219 L
s
.05436 .78219 m
.04417 .77631 L
s
.04417 .77631 m
.04417 .76455 L
s
.04417 .76455 m
.03399 .77043 L
s
.03399 .77043 m
.02381 .76455 L
s
.02381 .76455 m
.02381 .75279 L
s
.02381 .75279 m
.03399 .74691 L
s
.03399 .74691 m
.02381 .74103 L
s
.02381 .74103 m
.02381 .72928 L
s
.02381 .72928 m
.03399 .7234 L
s
.03399 .7234 m
.04417 .72928 L
s
.04417 .72928 m
.04417 .71752 L
s
.04417 .71752 m
.05436 .71164 L
s
.05436 .71164 m
.04417 .70576 L
s
.04417 .70576 m
.04417 .694 L
s
.04417 .694 m
.03399 .69988 L
s
.03399 .69988 m
.02381 .694 L
s
.02381 .694 m
.02381 .68225 L
s
.02381 .68225 m
.03399 .67637 L
s
.03399 .67637 m
.02381 .67049 L
s
.02381 .67049 m
.02381 .65873 L
s
.02381 .65873 m
.03399 .65285 L
s
.03399 .65285 m
.04417 .65873 L
s
.04417 .65873 m
.04417 .64697 L
s
.04417 .64697 m
.05436 .64109 L
s
.05436 .64109 m
.06454 .64697 L
s
.06454 .64697 m
.06454 .65873 L
s
.06454 .65873 m
.07472 .65285 L
s
.07472 .65285 m
.0849 .65873 L
s
.0849 .65873 m
.0849 .64697 L
s
.0849 .64697 m
.09509 .64109 L
s
.09509 .64109 m
.0849 .63521 L
s
.0849 .63521 m
.0849 .62346 L
s
.0849 .62346 m
.09509 .61758 L
s
.09509 .61758 m
.10527 .62346 L
s
.10527 .62346 m
.10527 .6117 L
s
.10527 .6117 m
.11545 .60582 L
s
.11545 .60582 m
.12563 .6117 L
s
.12563 .6117 m
.12563 .62346 L
s
.12563 .62346 m
.13582 .61758 L
s
.13582 .61758 m
.146 .62346 L
s
.146 .62346 m
.146 .63521 L
s
.146 .63521 m
.13582 .64109 L
s
.13582 .64109 m
.146 .64697 L
s
.146 .64697 m
.146 .65873 L
s
.146 .65873 m
.15618 .65285 L
s
.15618 .65285 m
.16637 .65873 L
s
.16637 .65873 m
.16637 .64697 L
s
.16637 .64697 m
.17655 .64109 L
s
.17655 .64109 m
.18673 .64697 L
s
.18673 .64697 m
.18673 .65873 L
s
.18673 .65873 m
.19691 .65285 L
s
.19691 .65285 m
.2071 .65873 L
s
.2071 .65873 m
.2071 .64697 L
s
.2071 .64697 m
.21728 .64109 L
s
.21728 .64109 m
.2071 .63521 L
s
.2071 .63521 m
.2071 .62346 L
s
.2071 .62346 m
.21728 .61758 L
s
.21728 .61758 m
.22746 .62346 L
s
.22746 .62346 m
.22746 .6117 L
s
.22746 .6117 m
.23764 .60582 L
s
.23764 .60582 m
.22746 .59994 L
s
.22746 .59994 m
.22746 .58818 L
s
.22746 .58818 m
.21728 .59406 L
s
.21728 .59406 m
.2071 .58818 L
s
.2071 .58818 m
.2071 .57643 L
s
.2071 .57643 m
.21728 .57055 L
s
.21728 .57055 m
.2071 .56467 L
s
.2071 .56467 m
.2071 .55291 L
s
.2071 .55291 m
.21728 .54703 L
s
.21728 .54703 m
.22746 .55291 L
s
.22746 .55291 m
.22746 .54115 L
s
.22746 .54115 m
.23764 .53527 L
s
.23764 .53527 m
.24783 .54115 L
s
.24783 .54115 m
.24783 .55291 L
s
.24783 .55291 m
.25801 .54703 L
s
.25801 .54703 m
.26819 .55291 L
s
.26819 .55291 m
.26819 .54115 L
s
.26819 .54115 m
.27837 .53527 L
s
.27837 .53527 m
.26819 .52939 L
s
.26819 .52939 m
.26819 .51764 L
s
.26819 .51764 m
.27837 .51176 L
s
.27837 .51176 m
.28856 .51764 L
s
.28856 .51764 m
.28856 .50588 L
s
.28856 .50588 m
.29874 .5 L
s
.29874 .5 m
.28856 .49412 L
s
.28856 .49412 m
.28856 .48236 L
s
.28856 .48236 m
.27837 .48824 L
s
.27837 .48824 m
.26819 .48236 L
s
.26819 .48236 m
.26819 .47061 L
s
.26819 .47061 m
.27837 .46473 L
s
.27837 .46473 m
.26819 .45885 L
s
.26819 .45885 m
.26819 .44709 L
s
.26819 .44709 m
.25801 .45297 L
s
.25801 .45297 m
.24783 .44709 L
s
.24783 .44709 m
.24783 .45885 L
s
.24783 .45885 m
.23764 .46473 L
s
.23764 .46473 m
.22746 .45885 L
s
.22746 .45885 m
.22746 .44709 L
s
.22746 .44709 m
.21728 .45297 L
s
.21728 .45297 m
.2071 .44709 L
s
.2071 .44709 m
.2071 .43533 L
s
.2071 .43533 m
.21728 .42945 L
s
.21728 .42945 m
.2071 .42357 L
s
.2071 .42357 m
.2071 .41182 L
s
.2071 .41182 m
.21728 .40594 L
s
.21728 .40594 m
.22746 .41182 L
s
.22746 .41182 m
.22746 .40006 L
s
.22746 .40006 m
.23764 .39418 L
s
.23764 .39418 m
.22746 .3883 L
s
.22746 .3883 m
.22746 .37654 L
s
.22746 .37654 m
.21728 .38242 L
s
.21728 .38242 m
.2071 .37654 L
s
.2071 .37654 m
.2071 .36479 L
s
.2071 .36479 m
.21728 .35891 L
s
.21728 .35891 m
.2071 .35303 L
s
.2071 .35303 m
.2071 .34127 L
s
.2071 .34127 m
.19691 .34715 L
s
.19691 .34715 m
.18673 .34127 L
s
.18673 .34127 m
.18673 .35303 L
s
.18673 .35303 m
.17655 .35891 L
s
.17655 .35891 m
.16637 .35303 L
s
.16637 .35303 m
.16637 .34127 L
s
.16637 .34127 m
.15618 .34715 L
s
.15618 .34715 m
.146 .34127 L
s
.146 .34127 m
.146 .35303 L
s
.146 .35303 m
.13582 .35891 L
s
.13582 .35891 m
.146 .36479 L
s
.146 .36479 m
.146 .37654 L
s
.146 .37654 m
.13582 .38242 L
s
.13582 .38242 m
.12563 .37654 L
s
.12563 .37654 m
.12563 .3883 L
s
.12563 .3883 m
.11545 .39418 L
s
.11545 .39418 m
.10527 .3883 L
s
.10527 .3883 m
.10527 .37654 L
s
.10527 .37654 m
.09509 .38242 L
s
.09509 .38242 m
.0849 .37654 L
s
.0849 .37654 m
.0849 .36479 L
s
.0849 .36479 m
.09509 .35891 L
s
.09509 .35891 m
.0849 .35303 L
s
.0849 .35303 m
.0849 .34127 L
s
.0849 .34127 m
.07472 .34715 L
s
.07472 .34715 m
.06454 .34127 L
s
.06454 .34127 m
.06454 .35303 L
s
.06454 .35303 m
.05436 .35891 L
s
.05436 .35891 m
.04417 .35303 L
s
.04417 .35303 m
.04417 .34127 L
s
.04417 .34127 m
.03399 .34715 L
s
.03399 .34715 m
.02381 .34127 L
s
.02381 .34127 m
.02381 .32951 L
s
.02381 .32951 m
.03399 .32363 L
s
.03399 .32363 m
.02381 .31775 L
s
.02381 .31775 m
.02381 .306 L
s
.02381 .306 m
.03399 .30012 L
s
.03399 .30012 m
.04417 .306 L
s
.04417 .306 m
.04417 .29424 L
s
.04417 .29424 m
.05436 .28836 L
s
.05436 .28836 m
.04417 .28248 L
s
.04417 .28248 m
.04417 .27072 L
s
.04417 .27072 m
.03399 .2766 L
s
.03399 .2766 m
.02381 .27072 L
s
.02381 .27072 m
.02381 .25897 L
s
.02381 .25897 m
.03399 .25309 L
s
.03399 .25309 m
.02381 .24721 L
s
.02381 .24721 m
.02381 .23545 L
s
.02381 .23545 m
.03399 .22957 L
s
.03399 .22957 m
.04417 .23545 L
s
.04417 .23545 m
.04417 .22369 L
s
.04417 .22369 m
.05436 .21781 L
s
.05436 .21781 m
.06454 .22369 L
s
.06454 .22369 m
.06454 .23545 L
s
.06454 .23545 m
.07472 .22957 L
s
.07472 .22957 m
.0849 .23545 L
s
.0849 .23545 m
.0849 .22369 L
s
.0849 .22369 m
.09509 .21781 L
s
.09509 .21781 m
.0849 .21193 L
s
.0849 .21193 m
.0849 .20018 L
s
.0849 .20018 m
.09509 .1943 L
s
.09509 .1943 m
.10527 .20018 L
s
.10527 .20018 m
.10527 .18842 L
s
.10527 .18842 m
.11545 .18254 L
s
.11545 .18254 m
.10527 .17666 L
s
.10527 .17666 m
.10527 .1649 L
s
.10527 .1649 m
.09509 .17078 L
s
.09509 .17078 m
.0849 .1649 L
s
.0849 .1649 m
.0849 .15315 L
s
.0849 .15315 m
.09509 .14727 L
s
.09509 .14727 m
.0849 .14139 L
s
.0849 .14139 m
.0849 .12963 L
s
.0849 .12963 m
.07472 .13551 L
s
.07472 .13551 m
.06454 .12963 L
s
.06454 .12963 m
.06454 .14139 L
s
.06454 .14139 m
.05436 .14727 L
s
.05436 .14727 m
.04417 .14139 L
s
.04417 .14139 m
.04417 .12963 L
s
.04417 .12963 m
.03399 .13551 L
s
.03399 .13551 m
.02381 .12963 L
s
.02381 .12963 m
.02381 .11787 L
s
.02381 .11787 m
.03399 .11199 L
s
.03399 .11199 m
.02381 .10611 L
s
.02381 .10611 m
.02381 .09436 L
s
.02381 .09436 m
.03399 .08848 L
s
.03399 .08848 m
.04417 .09436 L
s
.04417 .09436 m
.04417 .0826 L
s
.04417 .0826 m
.05436 .07672 L
s
.05436 .07672 m
.04417 .07084 L
s
.04417 .07084 m
.04417 .05908 L
s
.04417 .05908 m
.03399 .06496 L
s
.03399 .06496 m
.02381 .05908 L
s
.02381 .05908 m
.02381 .04733 L
s
.02381 .04733 m
.03399 .04145 L
s
.03399 .04145 m
.02381 .03557 L
s
.02381 .03557 m
.02381 .02381 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0113379 0.97619 0.464853 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.198107 0.603785 0.198107 0.603785 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.19811 .19811 m
.19811 .20556 L
s
.19811 .20556 m
.19165 .20929 L
s
.19165 .20929 m
.19811 .21302 L
s
.19811 .21302 m
.19811 .22047 L
s
.19811 .22047 m
.19165 .2242 L
s
.19165 .2242 m
.1852 .22047 L
s
.1852 .22047 m
.1852 .22792 L
s
.1852 .22792 m
.17874 .23165 L
s
.17874 .23165 m
.1852 .23538 L
s
.1852 .23538 m
.1852 .24283 L
s
.1852 .24283 m
.19165 .23911 L
s
.19165 .23911 m
.19811 .24283 L
s
.19811 .24283 m
.19811 .25029 L
s
.19811 .25029 m
.19165 .25401 L
s
.19165 .25401 m
.19811 .25774 L
s
.19811 .25774 m
.19811 .26519 L
s
.19811 .26519 m
.19165 .26892 L
s
.19165 .26892 m
.1852 .26519 L
s
.1852 .26519 m
.1852 .27265 L
s
.1852 .27265 m
.17874 .27638 L
s
.17874 .27638 m
.17229 .27265 L
s
.17229 .27265 m
.17229 .26519 L
s
.17229 .26519 m
.16583 .26892 L
s
.16583 .26892 m
.15937 .26519 L
s
.15937 .26519 m
.15937 .27265 L
s
.15937 .27265 m
.15292 .27638 L
s
.15292 .27638 m
.15937 .2801 L
s
.15937 .2801 m
.15937 .28756 L
s
.15937 .28756 m
.15292 .29128 L
s
.15292 .29128 m
.14646 .28756 L
s
.14646 .28756 m
.14646 .29501 L
s
.14646 .29501 m
.14001 .29874 L
s
.14001 .29874 m
.14646 .30247 L
s
.14646 .30247 m
.14646 .30992 L
s
.14646 .30992 m
.15292 .30619 L
s
.15292 .30619 m
.15937 .30992 L
s
.15937 .30992 m
.15937 .31737 L
s
.15937 .31737 m
.15292 .3211 L
s
.15292 .3211 m
.15937 .32483 L
s
.15937 .32483 m
.15937 .33228 L
s
.15937 .33228 m
.16583 .32855 L
s
.16583 .32855 m
.17229 .33228 L
s
.17229 .33228 m
.17229 .32483 L
s
.17229 .32483 m
.17874 .3211 L
s
.17874 .3211 m
.1852 .32483 L
s
.1852 .32483 m
.1852 .33228 L
s
.1852 .33228 m
.19165 .32855 L
s
.19165 .32855 m
.19811 .33228 L
s
.19811 .33228 m
.19811 .33974 L
s
.19811 .33974 m
.19165 .34346 L
s
.19165 .34346 m
.19811 .34719 L
s
.19811 .34719 m
.19811 .35464 L
s
.19811 .35464 m
.19165 .35837 L
s
.19165 .35837 m
.1852 .35464 L
s
.1852 .35464 m
.1852 .3621 L
s
.1852 .3621 m
.17874 .36583 L
s
.17874 .36583 m
.1852 .36955 L
s
.1852 .36955 m
.1852 .37701 L
s
.1852 .37701 m
.19165 .37328 L
s
.19165 .37328 m
.19811 .37701 L
s
.19811 .37701 m
.19811 .38446 L
s
.19811 .38446 m
.19165 .38819 L
s
.19165 .38819 m
.19811 .39191 L
s
.19811 .39191 m
.19811 .39937 L
s
.19811 .39937 m
.19165 .4031 L
s
.19165 .4031 m
.1852 .39937 L
s
.1852 .39937 m
.1852 .40682 L
s
.1852 .40682 m
.17874 .41055 L
s
.17874 .41055 m
.17229 .40682 L
s
.17229 .40682 m
.17229 .39937 L
s
.17229 .39937 m
.16583 .4031 L
s
.16583 .4031 m
.15937 .39937 L
s
.15937 .39937 m
.15937 .40682 L
s
.15937 .40682 m
.15292 .41055 L
s
.15292 .41055 m
.15937 .41428 L
s
.15937 .41428 m
.15937 .42173 L
s
.15937 .42173 m
.15292 .42546 L
s
.15292 .42546 m
.14646 .42173 L
s
.14646 .42173 m
.14646 .42919 L
s
.14646 .42919 m
.14001 .43291 L
s
.14001 .43291 m
.13355 .42919 L
s
.13355 .42919 m
.13355 .42173 L
s
.13355 .42173 m
.1271 .42546 L
s
.1271 .42546 m
.12064 .42173 L
s
.12064 .42173 m
.12064 .41428 L
s
.12064 .41428 m
.1271 .41055 L
s
.1271 .41055 m
.12064 .40682 L
s
.12064 .40682 m
.12064 .39937 L
s
.12064 .39937 m
.11419 .4031 L
s
.11419 .4031 m
.10773 .39937 L
s
.10773 .39937 m
.10773 .40682 L
s
.10773 .40682 m
.10128 .41055 L
s
.10128 .41055 m
.09482 .40682 L
s
.09482 .40682 m
.09482 .39937 L
s
.09482 .39937 m
.08836 .4031 L
s
.08836 .4031 m
.08191 .39937 L
s
.08191 .39937 m
.08191 .40682 L
s
.08191 .40682 m
.07545 .41055 L
s
.07545 .41055 m
.08191 .41428 L
s
.08191 .41428 m
.08191 .42173 L
s
.08191 .42173 m
.07545 .42546 L
s
.07545 .42546 m
.069 .42173 L
s
.069 .42173 m
.069 .42919 L
s
.069 .42919 m
.06254 .43291 L
s
.06254 .43291 m
.069 .43664 L
s
.069 .43664 m
.069 .44409 L
s
.069 .44409 m
.07545 .44037 L
s
.07545 .44037 m
.08191 .44409 L
s
.08191 .44409 m
.08191 .45155 L
s
.08191 .45155 m
.07545 .45528 L
s
.07545 .45528 m
.08191 .459 L
s
.08191 .459 m
.08191 .46646 L
s
.08191 .46646 m
.07545 .47018 L
s
.07545 .47018 m
.069 .46646 L
s
.069 .46646 m
.069 .47391 L
s
.069 .47391 m
.06254 .47764 L
s
.06254 .47764 m
.05609 .47391 L
s
.05609 .47391 m
.05609 .46646 L
s
.05609 .46646 m
.04963 .47018 L
s
.04963 .47018 m
.04318 .46646 L
s
.04318 .46646 m
.04318 .47391 L
s
.04318 .47391 m
.03672 .47764 L
s
.03672 .47764 m
.04318 .48136 L
s
.04318 .48136 m
.04318 .48882 L
s
.04318 .48882 m
.03672 .49255 L
s
.03672 .49255 m
.03026 .48882 L
s
.03026 .48882 m
.03026 .49627 L
s
.03026 .49627 m
.02381 .5 L
s
.02381 .5 m
.03026 .50373 L
s
.03026 .50373 m
.03026 .51118 L
s
.03026 .51118 m
.03672 .50745 L
s
.03672 .50745 m
.04318 .51118 L
s
.04318 .51118 m
.04318 .51864 L
s
.04318 .51864 m
.03672 .52236 L
s
.03672 .52236 m
.04318 .52609 L
s
.04318 .52609 m
.04318 .53354 L
s
.04318 .53354 m
.04963 .52982 L
s
.04963 .52982 m
.05609 .53354 L
s
.05609 .53354 m
.05609 .52609 L
s
.05609 .52609 m
.06254 .52236 L
s
.06254 .52236 m
.069 .52609 L
s
.069 .52609 m
.069 .53354 L
s
.069 .53354 m
.07545 .52982 L
s
.07545 .52982 m
.08191 .53354 L
s
.08191 .53354 m
.08191 .541 L
s
.08191 .541 m
.07545 .54472 L
s
.07545 .54472 m
.08191 .54845 L
s
.08191 .54845 m
.08191 .55591 L
s
.08191 .55591 m
.07545 .55963 L
s
.07545 .55963 m
.069 .55591 L
s
.069 .55591 m
.069 .56336 L
s
.069 .56336 m
.06254 .56709 L
s
.06254 .56709 m
.069 .57081 L
s
.069 .57081 m
.069 .57827 L
s
.069 .57827 m
.07545 .57454 L
s
.07545 .57454 m
.08191 .57827 L
s
.08191 .57827 m
.08191 .58572 L
s
.08191 .58572 m
.07545 .58945 L
s
.07545 .58945 m
.08191 .59318 L
s
.08191 .59318 m
.08191 .60063 L
s
.08191 .60063 m
.08836 .5969 L
s
.08836 .5969 m
.09482 .60063 L
s
.09482 .60063 m
.09482 .59318 L
s
.09482 .59318 m
.10128 .58945 L
s
.10128 .58945 m
.10773 .59318 L
s
.10773 .59318 m
.10773 .60063 L
s
.10773 .60063 m
.11419 .5969 L
s
.11419 .5969 m
.12064 .60063 L
s
.12064 .60063 m
.12064 .59318 L
s
.12064 .59318 m
.1271 .58945 L
s
.1271 .58945 m
.12064 .58572 L
s
.12064 .58572 m
.12064 .57827 L
s
.12064 .57827 m
.1271 .57454 L
s
.1271 .57454 m
.13355 .57827 L
s
.13355 .57827 m
.13355 .57081 L
s
.13355 .57081 m
.14001 .56709 L
s
.14001 .56709 m
.14646 .57081 L
s
.14646 .57081 m
.14646 .57827 L
s
.14646 .57827 m
.15292 .57454 L
s
.15292 .57454 m
.15937 .57827 L
s
.15937 .57827 m
.15937 .58572 L
s
.15937 .58572 m
.15292 .58945 L
s
.15292 .58945 m
.15937 .59318 L
s
.15937 .59318 m
.15937 .60063 L
s
.15937 .60063 m
.16583 .5969 L
s
.16583 .5969 m
.17229 .60063 L
s
.17229 .60063 m
.17229 .59318 L
s
.17229 .59318 m
.17874 .58945 L
s
.17874 .58945 m
.1852 .59318 L
s
.1852 .59318 m
.1852 .60063 L
s
.1852 .60063 m
.19165 .5969 L
s
.19165 .5969 m
.19811 .60063 L
s
.19811 .60063 m
.19811 .60809 L
s
.19811 .60809 m
.19165 .61181 L
s
.19165 .61181 m
.19811 .61554 L
s
.19811 .61554 m
.19811 .62299 L
s
.19811 .62299 m
.19165 .62672 L
s
.19165 .62672 m
.1852 .62299 L
s
.1852 .62299 m
.1852 .63045 L
s
.1852 .63045 m
.17874 .63417 L
s
.17874 .63417 m
.1852 .6379 L
s
.1852 .6379 m
.1852 .64536 L
s
.1852 .64536 m
.19165 .64163 L
s
.19165 .64163 m
.19811 .64536 L
s
.19811 .64536 m
.19811 .65281 L
s
.19811 .65281 m
.19165 .65654 L
s
.19165 .65654 m
.19811 .66026 L
s
.19811 .66026 m
.19811 .66772 L
s
.19811 .66772 m
.19165 .67145 L
s
.19165 .67145 m
.1852 .66772 L
s
.1852 .66772 m
.1852 .67517 L
s
.1852 .67517 m
.17874 .6789 L
s
.17874 .6789 m
.17229 .67517 L
s
.17229 .67517 m
.17229 .66772 L
s
.17229 .66772 m
.16583 .67145 L
s
.16583 .67145 m
.15937 .66772 L
s
.15937 .66772 m
.15937 .67517 L
s
.15937 .67517 m
.15292 .6789 L
s
.15292 .6789 m
.15937 .68263 L
s
.15937 .68263 m
.15937 .69008 L
s
.15937 .69008 m
.15292 .69381 L
s
.15292 .69381 m
.14646 .69008 L
s
.14646 .69008 m
.14646 .69753 L
s
.14646 .69753 m
.14001 .70126 L
s
.14001 .70126 m
.14646 .70499 L
s
.14646 .70499 m
.14646 .71244 L
s
.14646 .71244 m
.15292 .70872 L
s
.15292 .70872 m
.15937 .71244 L
s
.15937 .71244 m
.15937 .7199 L
s
.15937 .7199 m
.15292 .72362 L
s
.15292 .72362 m
.15937 .72735 L
s
.15937 .72735 m
.15937 .73481 L
s
.15937 .73481 m
.16583 .73108 L
s
.16583 .73108 m
.17229 .73481 L
s
.17229 .73481 m
.17229 .72735 L
s
.17229 .72735 m
.17874 .72362 L
s
.17874 .72362 m
.1852 .72735 L
s
.1852 .72735 m
.1852 .73481 L
s
.1852 .73481 m
.19165 .73108 L
s
.19165 .73108 m
.19811 .73481 L
s
.19811 .73481 m
.19811 .74226 L
s
.19811 .74226 m
.19165 .74599 L
s
.19165 .74599 m
.19811 .74971 L
s
.19811 .74971 m
.19811 .75717 L
s
.19811 .75717 m
.19165 .76089 L
s
.19165 .76089 m
.1852 .75717 L
s
.1852 .75717 m
.1852 .76462 L
s
.1852 .76462 m
.17874 .76835 L
s
.17874 .76835 m
.1852 .77208 L
s
.1852 .77208 m
.1852 .77953 L
s
.1852 .77953 m
.19165 .7758 L
s
.19165 .7758 m
.19811 .77953 L
s
.19811 .77953 m
.19811 .78698 L
s
.19811 .78698 m
.19165 .79071 L
s
.19165 .79071 m
.19811 .79444 L
s
.19811 .79444 m
.19811 .80189 L
s
.19811 .80189 m
.20556 .80189 L
s
.20556 .80189 m
.20929 .80835 L
s
.20929 .80835 m
.21302 .80189 L
s
.21302 .80189 m
.22047 .80189 L
s
.22047 .80189 m
.2242 .80835 L
s
.2242 .80835 m
.22047 .8148 L
s
.22047 .8148 m
.22792 .8148 L
s
.22792 .8148 m
.23165 .82126 L
s
.23165 .82126 m
.23538 .8148 L
s
.23538 .8148 m
.24283 .8148 L
s
.24283 .8148 m
.23911 .80835 L
s
.23911 .80835 m
.24283 .80189 L
s
.24283 .80189 m
.25029 .80189 L
s
.25029 .80189 m
.25401 .80835 L
s
.25401 .80835 m
.25774 .80189 L
s
.25774 .80189 m
.26519 .80189 L
s
.26519 .80189 m
.26892 .80835 L
s
.26892 .80835 m
.26519 .8148 L
s
.26519 .8148 m
.27265 .8148 L
s
.27265 .8148 m
.27638 .82126 L
s
.27638 .82126 m
.27265 .82771 L
s
.27265 .82771 m
.26519 .82771 L
s
.26519 .82771 m
.26892 .83417 L
s
.26892 .83417 m
.26519 .84063 L
s
.26519 .84063 m
.27265 .84063 L
s
.27265 .84063 m
.27638 .84708 L
s
.27638 .84708 m
.2801 .84063 L
s
.2801 .84063 m
.28756 .84063 L
s
.28756 .84063 m
.29128 .84708 L
s
.29128 .84708 m
.28756 .85354 L
s
.28756 .85354 m
.29501 .85354 L
s
.29501 .85354 m
.29874 .85999 L
s
.29874 .85999 m
.30247 .85354 L
s
.30247 .85354 m
.30992 .85354 L
s
.30992 .85354 m
.30619 .84708 L
s
.30619 .84708 m
.30992 .84063 L
s
.30992 .84063 m
.31737 .84063 L
s
.31737 .84063 m
.3211 .84708 L
s
.3211 .84708 m
.32483 .84063 L
s
.32483 .84063 m
.33228 .84063 L
s
.33228 .84063 m
.32855 .83417 L
s
.32855 .83417 m
.33228 .82771 L
s
.33228 .82771 m
.32483 .82771 L
s
.32483 .82771 m
.3211 .82126 L
s
.3211 .82126 m
.32483 .8148 L
s
.32483 .8148 m
.33228 .8148 L
s
.33228 .8148 m
.32855 .80835 L
s
.32855 .80835 m
.33228 .80189 L
s
.33228 .80189 m
.33974 .80189 L
s
.33974 .80189 m
.34346 .80835 L
s
.34346 .80835 m
.34719 .80189 L
s
.34719 .80189 m
.35464 .80189 L
s
.35464 .80189 m
.35837 .80835 L
s
.35837 .80835 m
.35464 .8148 L
s
.35464 .8148 m
.3621 .8148 L
s
.3621 .8148 m
.36583 .82126 L
s
.36583 .82126 m
.36955 .8148 L
s
.36955 .8148 m
.37701 .8148 L
s
.37701 .8148 m
.37328 .80835 L
s
.37328 .80835 m
.37701 .80189 L
s
.37701 .80189 m
.38446 .80189 L
s
.38446 .80189 m
.38819 .80835 L
s
.38819 .80835 m
.39191 .80189 L
s
.39191 .80189 m
.39937 .80189 L
s
.39937 .80189 m
.4031 .80835 L
s
.4031 .80835 m
.39937 .8148 L
s
.39937 .8148 m
.40682 .8148 L
s
.40682 .8148 m
.41055 .82126 L
s
.41055 .82126 m
.40682 .82771 L
s
.40682 .82771 m
.39937 .82771 L
s
.39937 .82771 m
.4031 .83417 L
s
.4031 .83417 m
.39937 .84063 L
s
.39937 .84063 m
.40682 .84063 L
s
.40682 .84063 m
.41055 .84708 L
s
.41055 .84708 m
.41428 .84063 L
s
.41428 .84063 m
.42173 .84063 L
s
.42173 .84063 m
.42546 .84708 L
s
.42546 .84708 m
.42173 .85354 L
s
.42173 .85354 m
.42919 .85354 L
s
.42919 .85354 m
.43291 .85999 L
s
.43291 .85999 m
.42919 .86645 L
s
.42919 .86645 m
.42173 .86645 L
s
.42173 .86645 m
.42546 .8729 L
s
.42546 .8729 m
.42173 .87936 L
s
.42173 .87936 m
.41428 .87936 L
s
.41428 .87936 m
.41055 .8729 L
s
.41055 .8729 m
.40682 .87936 L
s
.40682 .87936 m
.39937 .87936 L
s
.39937 .87936 m
.4031 .88581 L
s
.4031 .88581 m
.39937 .89227 L
s
.39937 .89227 m
.40682 .89227 L
s
.40682 .89227 m
.41055 .89872 L
s
.41055 .89872 m
.40682 .90518 L
s
.40682 .90518 m
.39937 .90518 L
s
.39937 .90518 m
.4031 .91164 L
s
.4031 .91164 m
.39937 .91809 L
s
.39937 .91809 m
.40682 .91809 L
s
.40682 .91809 m
.41055 .92455 L
s
.41055 .92455 m
.41428 .91809 L
s
.41428 .91809 m
.42173 .91809 L
s
.42173 .91809 m
.42546 .92455 L
s
.42546 .92455 m
.42173 .931 L
s
.42173 .931 m
.42919 .931 L
s
.42919 .931 m
.43291 .93746 L
s
.43291 .93746 m
.43664 .931 L
s
.43664 .931 m
.44409 .931 L
s
.44409 .931 m
.44037 .92455 L
s
.44037 .92455 m
.44409 .91809 L
s
.44409 .91809 m
.45155 .91809 L
s
.45155 .91809 m
.45528 .92455 L
s
.45528 .92455 m
.459 .91809 L
s
.459 .91809 m
.46646 .91809 L
s
.46646 .91809 m
.47018 .92455 L
s
.47018 .92455 m
.46646 .931 L
s
.46646 .931 m
.47391 .931 L
s
.47391 .931 m
.47764 .93746 L
s
.47764 .93746 m
.47391 .94391 L
s
.47391 .94391 m
.46646 .94391 L
s
.46646 .94391 m
.47018 .95037 L
s
.47018 .95037 m
.46646 .95682 L
s
.46646 .95682 m
.47391 .95682 L
s
.47391 .95682 m
.47764 .96328 L
s
.47764 .96328 m
.48136 .95682 L
s
.48136 .95682 m
.48882 .95682 L
s
.48882 .95682 m
.49255 .96328 L
s
.49255 .96328 m
.48882 .96974 L
s
.48882 .96974 m
.49627 .96974 L
s
.49627 .96974 m
.5 .97619 L
s
.5 .97619 m
.50373 .96974 L
s
.50373 .96974 m
.51118 .96974 L
s
.51118 .96974 m
.50745 .96328 L
s
.50745 .96328 m
.51118 .95682 L
s
.51118 .95682 m
.51864 .95682 L
s
.51864 .95682 m
.52236 .96328 L
s
.52236 .96328 m
.52609 .95682 L
s
.52609 .95682 m
.53354 .95682 L
s
.53354 .95682 m
.52982 .95037 L
s
.52982 .95037 m
.53354 .94391 L
s
.53354 .94391 m
.52609 .94391 L
s
.52609 .94391 m
.52236 .93746 L
s
.52236 .93746 m
.52609 .931 L
s
.52609 .931 m
.53354 .931 L
s
.53354 .931 m
.52982 .92455 L
s
.52982 .92455 m
.53354 .91809 L
s
.53354 .91809 m
.541 .91809 L
s
.541 .91809 m
.54472 .92455 L
s
.54472 .92455 m
.54845 .91809 L
s
.54845 .91809 m
.55591 .91809 L
s
.55591 .91809 m
.55963 .92455 L
s
.55963 .92455 m
.55591 .931 L
s
.55591 .931 m
.56336 .931 L
s
.56336 .931 m
.56709 .93746 L
s
.56709 .93746 m
.57081 .931 L
s
.57081 .931 m
.57827 .931 L
s
.57827 .931 m
.57454 .92455 L
s
.57454 .92455 m
.57827 .91809 L
s
.57827 .91809 m
.58572 .91809 L
s
.58572 .91809 m
.58945 .92455 L
s
.58945 .92455 m
.59318 .91809 L
s
.59318 .91809 m
.60063 .91809 L
s
.60063 .91809 m
.5969 .91164 L
s
.5969 .91164 m
.60063 .90518 L
s
.60063 .90518 m
.59318 .90518 L
s
.59318 .90518 m
.58945 .89872 L
s
.58945 .89872 m
.59318 .89227 L
s
.59318 .89227 m
.60063 .89227 L
s
.60063 .89227 m
.5969 .88581 L
s
.5969 .88581 m
.60063 .87936 L
s
.60063 .87936 m
.59318 .87936 L
s
.59318 .87936 m
.58945 .8729 L
s
.58945 .8729 m
.58572 .87936 L
s
.58572 .87936 m
.57827 .87936 L
s
.57827 .87936 m
.57454 .8729 L
s
.57454 .8729 m
.57827 .86645 L
s
.57827 .86645 m
.57081 .86645 L
s
.57081 .86645 m
.56709 .85999 L
s
.56709 .85999 m
.57081 .85354 L
s
.57081 .85354 m
.57827 .85354 L
s
.57827 .85354 m
.57454 .84708 L
s
.57454 .84708 m
.57827 .84063 L
s
.57827 .84063 m
.58572 .84063 L
s
.58572 .84063 m
.58945 .84708 L
s
.58945 .84708 m
.59318 .84063 L
s
.59318 .84063 m
.60063 .84063 L
s
.60063 .84063 m
.5969 .83417 L
s
.5969 .83417 m
.60063 .82771 L
s
.60063 .82771 m
.59318 .82771 L
s
.59318 .82771 m
.58945 .82126 L
s
.58945 .82126 m
.59318 .8148 L
s
.59318 .8148 m
.60063 .8148 L
s
.60063 .8148 m
.5969 .80835 L
s
.5969 .80835 m
.60063 .80189 L
s
.60063 .80189 m
.60809 .80189 L
s
.60809 .80189 m
.61181 .80835 L
s
.61181 .80835 m
.61554 .80189 L
s
.61554 .80189 m
.62299 .80189 L
s
.62299 .80189 m
.62672 .80835 L
s
.62672 .80835 m
.62299 .8148 L
s
.62299 .8148 m
.63045 .8148 L
s
.63045 .8148 m
.63417 .82126 L
s
.63417 .82126 m
.6379 .8148 L
s
.6379 .8148 m
.64536 .8148 L
s
.64536 .8148 m
.64163 .80835 L
s
.64163 .80835 m
.64536 .80189 L
s
.64536 .80189 m
.65281 .80189 L
s
.65281 .80189 m
.65654 .80835 L
s
.65654 .80835 m
.66026 .80189 L
s
.66026 .80189 m
.66772 .80189 L
s
.66772 .80189 m
.67145 .80835 L
s
.67145 .80835 m
.66772 .8148 L
s
.66772 .8148 m
.67517 .8148 L
s
.67517 .8148 m
.6789 .82126 L
s
.6789 .82126 m
.67517 .82771 L
s
.67517 .82771 m
.66772 .82771 L
s
.66772 .82771 m
.67145 .83417 L
s
.67145 .83417 m
.66772 .84063 L
s
.66772 .84063 m
.67517 .84063 L
s
.67517 .84063 m
.6789 .84708 L
s
.6789 .84708 m
.68263 .84063 L
s
.68263 .84063 m
.69008 .84063 L
s
.69008 .84063 m
.69381 .84708 L
s
.69381 .84708 m
.69008 .85354 L
s
.69008 .85354 m
.69753 .85354 L
s
.69753 .85354 m
.70126 .85999 L
s
.70126 .85999 m
.70499 .85354 L
s
.70499 .85354 m
.71244 .85354 L
s
.71244 .85354 m
.70872 .84708 L
s
.70872 .84708 m
.71244 .84063 L
s
.71244 .84063 m
.7199 .84063 L
s
.7199 .84063 m
.72362 .84708 L
s
.72362 .84708 m
.72735 .84063 L
s
.72735 .84063 m
.73481 .84063 L
s
.73481 .84063 m
.73108 .83417 L
s
.73108 .83417 m
.73481 .82771 L
s
.73481 .82771 m
.72735 .82771 L
s
.72735 .82771 m
.72362 .82126 L
s
.72362 .82126 m
.72735 .8148 L
s
.72735 .8148 m
.73481 .8148 L
s
.73481 .8148 m
.73108 .80835 L
s
.73108 .80835 m
.73481 .80189 L
s
.73481 .80189 m
.74226 .80189 L
s
.74226 .80189 m
.74599 .80835 L
s
.74599 .80835 m
.74971 .80189 L
s
.74971 .80189 m
.75717 .80189 L
s
.75717 .80189 m
.76089 .80835 L
s
.76089 .80835 m
.75717 .8148 L
s
.75717 .8148 m
.76462 .8148 L
s
.76462 .8148 m
.76835 .82126 L
s
.76835 .82126 m
.77208 .8148 L
s
.77208 .8148 m
.77953 .8148 L
s
.77953 .8148 m
.7758 .80835 L
s
.7758 .80835 m
.77953 .80189 L
s
.77953 .80189 m
.78698 .80189 L
s
.78698 .80189 m
.79071 .80835 L
s
.79071 .80835 m
.79444 .80189 L
s
.79444 .80189 m
.80189 .80189 L
s
.80189 .80189 m
.80189 .79444 L
s
.80189 .79444 m
.80835 .79071 L
s
.80835 .79071 m
.80189 .78698 L
s
.80189 .78698 m
.80189 .77953 L
s
.80189 .77953 m
.80835 .7758 L
s
.80835 .7758 m
.8148 .77953 L
s
.8148 .77953 m
.8148 .77208 L
s
.8148 .77208 m
.82126 .76835 L
s
.82126 .76835 m
.8148 .76462 L
s
.8148 .76462 m
.8148 .75717 L
s
.8148 .75717 m
.80835 .76089 L
s
.80835 .76089 m
.80189 .75717 L
s
.80189 .75717 m
.80189 .74971 L
s
.80189 .74971 m
.80835 .74599 L
s
.80835 .74599 m
.80189 .74226 L
s
.80189 .74226 m
.80189 .73481 L
s
.80189 .73481 m
.80835 .73108 L
s
.80835 .73108 m
.8148 .73481 L
s
.8148 .73481 m
.8148 .72735 L
s
.8148 .72735 m
.82126 .72362 L
s
.82126 .72362 m
.82771 .72735 L
s
.82771 .72735 m
.82771 .73481 L
s
.82771 .73481 m
.83417 .73108 L
s
.83417 .73108 m
.84063 .73481 L
s
.84063 .73481 m
.84063 .72735 L
s
.84063 .72735 m
.84708 .72362 L
s
.84708 .72362 m
.84063 .7199 L
s
.84063 .7199 m
.84063 .71244 L
s
.84063 .71244 m
.84708 .70872 L
s
.84708 .70872 m
.85354 .71244 L
s
.85354 .71244 m
.85354 .70499 L
s
.85354 .70499 m
.85999 .70126 L
s
.85999 .70126 m
.85354 .69753 L
s
.85354 .69753 m
.85354 .69008 L
s
.85354 .69008 m
.84708 .69381 L
s
.84708 .69381 m
.84063 .69008 L
s
.84063 .69008 m
.84063 .68263 L
s
.84063 .68263 m
.84708 .6789 L
s
.84708 .6789 m
.84063 .67517 L
s
.84063 .67517 m
.84063 .66772 L
s
.84063 .66772 m
.83417 .67145 L
s
.83417 .67145 m
.82771 .66772 L
s
.82771 .66772 m
.82771 .67517 L
s
.82771 .67517 m
.82126 .6789 L
s
.82126 .6789 m
.8148 .67517 L
s
.8148 .67517 m
.8148 .66772 L
s
.8148 .66772 m
.80835 .67145 L
s
.80835 .67145 m
.80189 .66772 L
s
.80189 .66772 m
.80189 .66026 L
s
.80189 .66026 m
.80835 .65654 L
s
.80835 .65654 m
.80189 .65281 L
s
.80189 .65281 m
.80189 .64536 L
s
.80189 .64536 m
.80835 .64163 L
s
.80835 .64163 m
.8148 .64536 L
s
.8148 .64536 m
.8148 .6379 L
s
.8148 .6379 m
.82126 .63417 L
s
.82126 .63417 m
.8148 .63045 L
s
.8148 .63045 m
.8148 .62299 L
s
.8148 .62299 m
.80835 .62672 L
s
.80835 .62672 m
.80189 .62299 L
s
.80189 .62299 m
.80189 .61554 L
s
.80189 .61554 m
.80835 .61181 L
s
.80835 .61181 m
.80189 .60809 L
s
.80189 .60809 m
.80189 .60063 L
s
.80189 .60063 m
.80835 .5969 L
s
.80835 .5969 m
.8148 .60063 L
s
.8148 .60063 m
.8148 .59318 L
s
.8148 .59318 m
.82126 .58945 L
s
.82126 .58945 m
.82771 .59318 L
s
.82771 .59318 m
.82771 .60063 L
s
.82771 .60063 m
.83417 .5969 L
s
.83417 .5969 m
.84063 .60063 L
s
.84063 .60063 m
.84063 .59318 L
s
.84063 .59318 m
.84708 .58945 L
s
.84708 .58945 m
.84063 .58572 L
s
.84063 .58572 m
.84063 .57827 L
s
.84063 .57827 m
.84708 .57454 L
s
.84708 .57454 m
.85354 .57827 L
s
.85354 .57827 m
.85354 .57081 L
s
.85354 .57081 m
.85999 .56709 L
s
.85999 .56709 m
.86645 .57081 L
s
.86645 .57081 m
.86645 .57827 L
s
.86645 .57827 m
.8729 .57454 L
s
.8729 .57454 m
.87936 .57827 L
s
.87936 .57827 m
.87936 .58572 L
s
.87936 .58572 m
.8729 .58945 L
s
.8729 .58945 m
.87936 .59318 L
s
.87936 .59318 m
.87936 .60063 L
s
.87936 .60063 m
.88581 .5969 L
s
.88581 .5969 m
.89227 .60063 L
s
.89227 .60063 m
.89227 .59318 L
s
.89227 .59318 m
.89872 .58945 L
s
.89872 .58945 m
.90518 .59318 L
s
.90518 .59318 m
.90518 .60063 L
s
.90518 .60063 m
.91164 .5969 L
s
.91164 .5969 m
.91809 .60063 L
s
.91809 .60063 m
.91809 .59318 L
s
.91809 .59318 m
.92455 .58945 L
s
.92455 .58945 m
.91809 .58572 L
s
.91809 .58572 m
.91809 .57827 L
s
.91809 .57827 m
.92455 .57454 L
s
.92455 .57454 m
.931 .57827 L
s
.931 .57827 m
.931 .57081 L
s
.931 .57081 m
.93746 .56709 L
s
.93746 .56709 m
.931 .56336 L
s
.931 .56336 m
.931 .55591 L
s
.931 .55591 m
.92455 .55963 L
s
.92455 .55963 m
.91809 .55591 L
s
.91809 .55591 m
.91809 .54845 L
s
.91809 .54845 m
.92455 .54472 L
s
.92455 .54472 m
.91809 .541 L
s
.91809 .541 m
.91809 .53354 L
s
.91809 .53354 m
.92455 .52982 L
s
.92455 .52982 m
.931 .53354 L
s
.931 .53354 m
.931 .52609 L
s
.931 .52609 m
.93746 .52236 L
s
.93746 .52236 m
.94391 .52609 L
s
.94391 .52609 m
.94391 .53354 L
s
.94391 .53354 m
.95037 .52982 L
s
.95037 .52982 m
.95682 .53354 L
s
.95682 .53354 m
.95682 .52609 L
s
.95682 .52609 m
.96328 .52236 L
s
.96328 .52236 m
.95682 .51864 L
s
.95682 .51864 m
.95682 .51118 L
s
.95682 .51118 m
.96328 .50745 L
s
.96328 .50745 m
.96974 .51118 L
s
.96974 .51118 m
.96974 .50373 L
s
.96974 .50373 m
.97619 .5 L
s
.97619 .5 m
.96974 .49627 L
s
.96974 .49627 m
.96974 .48882 L
s
.96974 .48882 m
.96328 .49255 L
s
.96328 .49255 m
.95682 .48882 L
s
.95682 .48882 m
.95682 .48136 L
s
.95682 .48136 m
.96328 .47764 L
s
.96328 .47764 m
.95682 .47391 L
s
.95682 .47391 m
.95682 .46646 L
s
.95682 .46646 m
.95037 .47018 L
s
.95037 .47018 m
.94391 .46646 L
s
.94391 .46646 m
.94391 .47391 L
s
.94391 .47391 m
.93746 .47764 L
s
.93746 .47764 m
.931 .47391 L
s
.931 .47391 m
.931 .46646 L
s
.931 .46646 m
.92455 .47018 L
s
.92455 .47018 m
.91809 .46646 L
s
.91809 .46646 m
.91809 .459 L
s
.91809 .459 m
.92455 .45528 L
s
.92455 .45528 m
.91809 .45155 L
s
.91809 .45155 m
.91809 .44409 L
s
.91809 .44409 m
.92455 .44037 L
s
.92455 .44037 m
.931 .44409 L
s
.931 .44409 m
.931 .43664 L
s
.931 .43664 m
.93746 .43291 L
s
.93746 .43291 m
.931 .42919 L
s
.931 .42919 m
.931 .42173 L
s
.931 .42173 m
.92455 .42546 L
s
.92455 .42546 m
.91809 .42173 L
s
.91809 .42173 m
.91809 .41428 L
s
.91809 .41428 m
.92455 .41055 L
s
.92455 .41055 m
.91809 .40682 L
s
.91809 .40682 m
.91809 .39937 L
s
.91809 .39937 m
.91164 .4031 L
s
.91164 .4031 m
.90518 .39937 L
s
.90518 .39937 m
.90518 .40682 L
s
.90518 .40682 m
.89872 .41055 L
s
.89872 .41055 m
.89227 .40682 L
s
.89227 .40682 m
.89227 .39937 L
s
.89227 .39937 m
.88581 .4031 L
s
.88581 .4031 m
.87936 .39937 L
s
.87936 .39937 m
.87936 .40682 L
s
.87936 .40682 m
.8729 .41055 L
s
.8729 .41055 m
.87936 .41428 L
s
.87936 .41428 m
.87936 .42173 L
s
.87936 .42173 m
.8729 .42546 L
s
.8729 .42546 m
.86645 .42173 L
s
.86645 .42173 m
.86645 .42919 L
s
.86645 .42919 m
.85999 .43291 L
s
.85999 .43291 m
.85354 .42919 L
s
.85354 .42919 m
.85354 .42173 L
s
.85354 .42173 m
.84708 .42546 L
s
.84708 .42546 m
.84063 .42173 L
s
.84063 .42173 m
.84063 .41428 L
s
.84063 .41428 m
.84708 .41055 L
s
.84708 .41055 m
.84063 .40682 L
s
.84063 .40682 m
.84063 .39937 L
s
.84063 .39937 m
.83417 .4031 L
s
.83417 .4031 m
.82771 .39937 L
s
.82771 .39937 m
.82771 .40682 L
s
.82771 .40682 m
.82126 .41055 L
s
.82126 .41055 m
.8148 .40682 L
s
.8148 .40682 m
.8148 .39937 L
s
.8148 .39937 m
.80835 .4031 L
s
.80835 .4031 m
.80189 .39937 L
s
.80189 .39937 m
.80189 .39191 L
s
.80189 .39191 m
.80835 .38819 L
s
.80835 .38819 m
.80189 .38446 L
s
.80189 .38446 m
.80189 .37701 L
s
.80189 .37701 m
.80835 .37328 L
s
.80835 .37328 m
.8148 .37701 L
s
.8148 .37701 m
.8148 .36955 L
s
.8148 .36955 m
.82126 .36583 L
s
.82126 .36583 m
.8148 .3621 L
s
.8148 .3621 m
.8148 .35464 L
s
.8148 .35464 m
.80835 .35837 L
s
.80835 .35837 m
.80189 .35464 L
s
.80189 .35464 m
.80189 .34719 L
s
.80189 .34719 m
.80835 .34346 L
s
.80835 .34346 m
.80189 .33974 L
s
.80189 .33974 m
.80189 .33228 L
s
.80189 .33228 m
.80835 .32855 L
s
.80835 .32855 m
.8148 .33228 L
s
.8148 .33228 m
.8148 .32483 L
s
.8148 .32483 m
.82126 .3211 L
s
.82126 .3211 m
.82771 .32483 L
s
.82771 .32483 m
.82771 .33228 L
s
.82771 .33228 m
.83417 .32855 L
s
.83417 .32855 m
.84063 .33228 L
s
.84063 .33228 m
.84063 .32483 L
s
.84063 .32483 m
.84708 .3211 L
s
.84708 .3211 m
.84063 .31737 L
s
.84063 .31737 m
.84063 .30992 L
s
.84063 .30992 m
.84708 .30619 L
s
.84708 .30619 m
.85354 .30992 L
s
.85354 .30992 m
.85354 .30247 L
s
.85354 .30247 m
.85999 .29874 L
s
.85999 .29874 m
.85354 .29501 L
s
.85354 .29501 m
.85354 .28756 L
s
.85354 .28756 m
.84708 .29128 L
s
.84708 .29128 m
.84063 .28756 L
s
.84063 .28756 m
.84063 .2801 L
s
.84063 .2801 m
.84708 .27638 L
s
.84708 .27638 m
.84063 .27265 L
s
.84063 .27265 m
.84063 .26519 L
s
.84063 .26519 m
.83417 .26892 L
s
.83417 .26892 m
.82771 .26519 L
s
.82771 .26519 m
.82771 .27265 L
s
.82771 .27265 m
.82126 .27638 L
s
.82126 .27638 m
.8148 .27265 L
s
.8148 .27265 m
.8148 .26519 L
s
.8148 .26519 m
.80835 .26892 L
s
.80835 .26892 m
.80189 .26519 L
s
.80189 .26519 m
.80189 .25774 L
s
.80189 .25774 m
.80835 .25401 L
s
.80835 .25401 m
.80189 .25029 L
s
.80189 .25029 m
.80189 .24283 L
s
.80189 .24283 m
.80835 .23911 L
s
.80835 .23911 m
.8148 .24283 L
s
.8148 .24283 m
.8148 .23538 L
s
.8148 .23538 m
.82126 .23165 L
s
.82126 .23165 m
.8148 .22792 L
s
.8148 .22792 m
.8148 .22047 L
s
.8148 .22047 m
.80835 .2242 L
s
.80835 .2242 m
.80189 .22047 L
s
.80189 .22047 m
.80189 .21302 L
s
.80189 .21302 m
.80835 .20929 L
s
.80835 .20929 m
.80189 .20556 L
s
.80189 .20556 m
.80189 .19811 L
s
.80189 .19811 m
.79444 .19811 L
s
.79444 .19811 m
.79071 .19165 L
s
.79071 .19165 m
.78698 .19811 L
s
.78698 .19811 m
.77953 .19811 L
s
.77953 .19811 m
.7758 .19165 L
s
.7758 .19165 m
.77953 .1852 L
s
.77953 .1852 m
.77208 .1852 L
s
.77208 .1852 m
.76835 .17874 L
s
.76835 .17874 m
.76462 .1852 L
s
.76462 .1852 m
.75717 .1852 L
s
.75717 .1852 m
.76089 .19165 L
s
.76089 .19165 m
.75717 .19811 L
s
.75717 .19811 m
.74971 .19811 L
s
.74971 .19811 m
.74599 .19165 L
s
.74599 .19165 m
.74226 .19811 L
s
.74226 .19811 m
.73481 .19811 L
s
.73481 .19811 m
.73108 .19165 L
s
.73108 .19165 m
.73481 .1852 L
s
.73481 .1852 m
.72735 .1852 L
s
.72735 .1852 m
.72362 .17874 L
s
.72362 .17874 m
.72735 .17229 L
s
.72735 .17229 m
.73481 .17229 L
s
.73481 .17229 m
.73108 .16583 L
s
.73108 .16583 m
.73481 .15937 L
s
.73481 .15937 m
.72735 .15937 L
s
.72735 .15937 m
.72362 .15292 L
s
.72362 .15292 m
.7199 .15937 L
s
.7199 .15937 m
.71244 .15937 L
s
.71244 .15937 m
.70872 .15292 L
s
.70872 .15292 m
.71244 .14646 L
s
.71244 .14646 m
.70499 .14646 L
s
.70499 .14646 m
.70126 .14001 L
s
.70126 .14001 m
.69753 .14646 L
s
.69753 .14646 m
.69008 .14646 L
s
.69008 .14646 m
.69381 .15292 L
s
.69381 .15292 m
.69008 .15937 L
s
.69008 .15937 m
.68263 .15937 L
s
.68263 .15937 m
.6789 .15292 L
s
.6789 .15292 m
.67517 .15937 L
s
.67517 .15937 m
.66772 .15937 L
s
.66772 .15937 m
.67145 .16583 L
s
.67145 .16583 m
.66772 .17229 L
s
.66772 .17229 m
.67517 .17229 L
s
.67517 .17229 m
.6789 .17874 L
s
.6789 .17874 m
.67517 .1852 L
s
.67517 .1852 m
.66772 .1852 L
s
.66772 .1852 m
.67145 .19165 L
s
.67145 .19165 m
.66772 .19811 L
s
.66772 .19811 m
.66026 .19811 L
s
.66026 .19811 m
.65654 .19165 L
s
.65654 .19165 m
.65281 .19811 L
s
.65281 .19811 m
.64536 .19811 L
s
.64536 .19811 m
.64163 .19165 L
s
.64163 .19165 m
.64536 .1852 L
s
.64536 .1852 m
.6379 .1852 L
s
.6379 .1852 m
.63417 .17874 L
s
.63417 .17874 m
.63045 .1852 L
s
.63045 .1852 m
.62299 .1852 L
s
.62299 .1852 m
.62672 .19165 L
s
.62672 .19165 m
.62299 .19811 L
s
.62299 .19811 m
.61554 .19811 L
s
.61554 .19811 m
.61181 .19165 L
s
.61181 .19165 m
.60809 .19811 L
s
.60809 .19811 m
.60063 .19811 L
s
.60063 .19811 m
.5969 .19165 L
s
.5969 .19165 m
.60063 .1852 L
s
.60063 .1852 m
.59318 .1852 L
s
.59318 .1852 m
.58945 .17874 L
s
.58945 .17874 m
.59318 .17229 L
s
.59318 .17229 m
.60063 .17229 L
s
.60063 .17229 m
.5969 .16583 L
s
.5969 .16583 m
.60063 .15937 L
s
.60063 .15937 m
.59318 .15937 L
s
.59318 .15937 m
.58945 .15292 L
s
.58945 .15292 m
.58572 .15937 L
s
.58572 .15937 m
.57827 .15937 L
s
.57827 .15937 m
.57454 .15292 L
s
.57454 .15292 m
.57827 .14646 L
s
.57827 .14646 m
.57081 .14646 L
s
.57081 .14646 m
.56709 .14001 L
s
.56709 .14001 m
.57081 .13355 L
s
.57081 .13355 m
.57827 .13355 L
s
.57827 .13355 m
.57454 .1271 L
s
.57454 .1271 m
.57827 .12064 L
s
.57827 .12064 m
.58572 .12064 L
s
.58572 .12064 m
.58945 .1271 L
s
.58945 .1271 m
.59318 .12064 L
s
.59318 .12064 m
.60063 .12064 L
s
.60063 .12064 m
.5969 .11419 L
s
.5969 .11419 m
.60063 .10773 L
s
.60063 .10773 m
.59318 .10773 L
s
.59318 .10773 m
.58945 .10128 L
s
.58945 .10128 m
.59318 .09482 L
s
.59318 .09482 m
.60063 .09482 L
s
.60063 .09482 m
.5969 .08836 L
s
.5969 .08836 m
.60063 .08191 L
s
.60063 .08191 m
.59318 .08191 L
s
.59318 .08191 m
.58945 .07545 L
s
.58945 .07545 m
.58572 .08191 L
s
.58572 .08191 m
.57827 .08191 L
s
.57827 .08191 m
.57454 .07545 L
s
.57454 .07545 m
.57827 .069 L
s
.57827 .069 m
.57081 .069 L
s
.57081 .069 m
.56709 .06254 L
s
.56709 .06254 m
.56336 .069 L
s
.56336 .069 m
.55591 .069 L
s
.55591 .069 m
.55963 .07545 L
s
.55963 .07545 m
.55591 .08191 L
s
.55591 .08191 m
.54845 .08191 L
s
.54845 .08191 m
.54472 .07545 L
s
.54472 .07545 m
.541 .08191 L
s
.541 .08191 m
.53354 .08191 L
s
.53354 .08191 m
.52982 .07545 L
s
.52982 .07545 m
.53354 .069 L
s
.53354 .069 m
.52609 .069 L
s
.52609 .069 m
.52236 .06254 L
s
.52236 .06254 m
.52609 .05609 L
s
.52609 .05609 m
.53354 .05609 L
s
.53354 .05609 m
.52982 .04963 L
s
.52982 .04963 m
.53354 .04318 L
s
.53354 .04318 m
.52609 .04318 L
s
.52609 .04318 m
.52236 .03672 L
s
.52236 .03672 m
.51864 .04318 L
s
.51864 .04318 m
.51118 .04318 L
s
.51118 .04318 m
.50745 .03672 L
s
.50745 .03672 m
.51118 .03026 L
s
.51118 .03026 m
.50373 .03026 L
s
.50373 .03026 m
.5 .02381 L
s
.5 .02381 m
.49627 .03026 L
s
.49627 .03026 m
.48882 .03026 L
s
.48882 .03026 m
.49255 .03672 L
s
.49255 .03672 m
.48882 .04318 L
s
.48882 .04318 m
.48136 .04318 L
s
.48136 .04318 m
.47764 .03672 L
s
.47764 .03672 m
.47391 .04318 L
s
.47391 .04318 m
.46646 .04318 L
s
.46646 .04318 m
.47018 .04963 L
s
.47018 .04963 m
.46646 .05609 L
s
.46646 .05609 m
.47391 .05609 L
s
.47391 .05609 m
.47764 .06254 L
s
.47764 .06254 m
.47391 .069 L
s
.47391 .069 m
.46646 .069 L
s
.46646 .069 m
.47018 .07545 L
s
.47018 .07545 m
.46646 .08191 L
s
.46646 .08191 m
.459 .08191 L
s
.459 .08191 m
.45528 .07545 L
s
.45528 .07545 m
.45155 .08191 L
s
.45155 .08191 m
.44409 .08191 L
s
.44409 .08191 m
.44037 .07545 L
s
.44037 .07545 m
.44409 .069 L
s
.44409 .069 m
.43664 .069 L
s
.43664 .069 m
.43291 .06254 L
s
.43291 .06254 m
.42919 .069 L
s
.42919 .069 m
.42173 .069 L
s
.42173 .069 m
.42546 .07545 L
s
.42546 .07545 m
.42173 .08191 L
s
.42173 .08191 m
.41428 .08191 L
s
.41428 .08191 m
.41055 .07545 L
s
.41055 .07545 m
.40682 .08191 L
s
.40682 .08191 m
.39937 .08191 L
s
.39937 .08191 m
.4031 .08836 L
s
.4031 .08836 m
.39937 .09482 L
s
.39937 .09482 m
.40682 .09482 L
s
.40682 .09482 m
.41055 .10128 L
s
.41055 .10128 m
.40682 .10773 L
s
.40682 .10773 m
.39937 .10773 L
s
.39937 .10773 m
.4031 .11419 L
s
.4031 .11419 m
.39937 .12064 L
s
.39937 .12064 m
.40682 .12064 L
s
.40682 .12064 m
.41055 .1271 L
s
.41055 .1271 m
.41428 .12064 L
s
.41428 .12064 m
.42173 .12064 L
s
.42173 .12064 m
.42546 .1271 L
s
.42546 .1271 m
.42173 .13355 L
s
.42173 .13355 m
.42919 .13355 L
s
.42919 .13355 m
.43291 .14001 L
s
.43291 .14001 m
.42919 .14646 L
s
.42919 .14646 m
.42173 .14646 L
s
.42173 .14646 m
.42546 .15292 L
s
.42546 .15292 m
.42173 .15937 L
s
.42173 .15937 m
.41428 .15937 L
s
.41428 .15937 m
.41055 .15292 L
s
.41055 .15292 m
.40682 .15937 L
s
.40682 .15937 m
.39937 .15937 L
s
.39937 .15937 m
.4031 .16583 L
s
.4031 .16583 m
.39937 .17229 L
s
.39937 .17229 m
.40682 .17229 L
s
.40682 .17229 m
.41055 .17874 L
s
.41055 .17874 m
.40682 .1852 L
s
.40682 .1852 m
.39937 .1852 L
s
.39937 .1852 m
.4031 .19165 L
s
.4031 .19165 m
.39937 .19811 L
s
.39937 .19811 m
.39191 .19811 L
s
.39191 .19811 m
.38819 .19165 L
s
.38819 .19165 m
.38446 .19811 L
s
.38446 .19811 m
.37701 .19811 L
s
.37701 .19811 m
.37328 .19165 L
s
.37328 .19165 m
.37701 .1852 L
s
.37701 .1852 m
.36955 .1852 L
s
.36955 .1852 m
.36583 .17874 L
s
.36583 .17874 m
.3621 .1852 L
s
.3621 .1852 m
.35464 .1852 L
s
.35464 .1852 m
.35837 .19165 L
s
.35837 .19165 m
.35464 .19811 L
s
.35464 .19811 m
.34719 .19811 L
s
.34719 .19811 m
.34346 .19165 L
s
.34346 .19165 m
.33974 .19811 L
s
.33974 .19811 m
.33228 .19811 L
s
.33228 .19811 m
.32855 .19165 L
s
.32855 .19165 m
.33228 .1852 L
s
.33228 .1852 m
.32483 .1852 L
s
.32483 .1852 m
.3211 .17874 L
s
.3211 .17874 m
.32483 .17229 L
s
.32483 .17229 m
.33228 .17229 L
s
.33228 .17229 m
.32855 .16583 L
s
.32855 .16583 m
.33228 .15937 L
s
.33228 .15937 m
.32483 .15937 L
s
.32483 .15937 m
.3211 .15292 L
s
.3211 .15292 m
.31737 .15937 L
s
.31737 .15937 m
.30992 .15937 L
s
.30992 .15937 m
.30619 .15292 L
s
.30619 .15292 m
.30992 .14646 L
s
.30992 .14646 m
.30247 .14646 L
s
.30247 .14646 m
.29874 .14001 L
s
.29874 .14001 m
.29501 .14646 L
s
.29501 .14646 m
.28756 .14646 L
s
.28756 .14646 m
.29128 .15292 L
s
.29128 .15292 m
.28756 .15937 L
s
.28756 .15937 m
.2801 .15937 L
s
.2801 .15937 m
.27638 .15292 L
s
.27638 .15292 m
.27265 .15937 L
s
.27265 .15937 m
.26519 .15937 L
s
.26519 .15937 m
.26892 .16583 L
s
.26892 .16583 m
.26519 .17229 L
s
.26519 .17229 m
.27265 .17229 L
s
.27265 .17229 m
.27638 .17874 L
s
.27638 .17874 m
.27265 .1852 L
s
.27265 .1852 m
.26519 .1852 L
s
.26519 .1852 m
.26892 .19165 L
s
.26892 .19165 m
.26519 .19811 L
s
.26519 .19811 m
.25774 .19811 L
s
.25774 .19811 m
.25401 .19165 L
s
.25401 .19165 m
.25029 .19811 L
s
.25029 .19811 m
.24283 .19811 L
s
.24283 .19811 m
.23911 .19165 L
s
.23911 .19165 m
.24283 .1852 L
s
.24283 .1852 m
.23538 .1852 L
s
.23538 .1852 m
.23165 .17874 L
s
.23165 .17874 m
.22792 .1852 L
s
.22792 .1852 m
.22047 .1852 L
s
.22047 .1852 m
.2242 .19165 L
s
.2242 .19165 m
.22047 .19811 L
s
.22047 .19811 m
.21302 .19811 L
s
.21302 .19811 m
.20929 .19165 L
s
.20929 .19165 m
.20556 .19811 L
s
.20556 .19811 m
.19811 .19811 L
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 137.125},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000cB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`003`3oool300000080oooo0`00
00060?ooo`<0000000<0oooo0000000000000P00000C0?ooo`@0000000@0oooo0000000000000000
1P3oool300000080oooo0`00000m0?ooo`<0000000<0oooo0000000000000P0000050?ooo`@00000
00@0oooo0000000000000000503oool300000080oooo0`0000060?ooo`<0000000<0oooo00000000
00000P00001h0?ooo`030000003oool0oooo06/0oooo000?0?ooo`050000003oool000000?ooo`00
00000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3oool00`000000
oooo0?ooo`0@0?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0?ooo`030?oo
o`060000003oool0oooo0000003oool000000P3oool00`000000oooo0?ooo`0j0?ooo`030000003o
ool0oooo008000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`800000
0P3oool00`000000oooo0?ooo`0A0?ooo`060000003oool0oooo0000003oool000000P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3oool00`000000oooo0?ooo`1e0?oo
o`800000K@3oool000l0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`020?oo
o`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000
103oool00`000000oooo0?ooo`0i0?ooo`050000003oool0oooo0?ooo`000000103oool01@000000
oooo0?ooo`3oool0000000@0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`050000003oool0
oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000
0`3oool00`000000oooo0?ooo`1d0?ooo`80000000<0oooo000000000000J`3oool000l0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0LP3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000006P0
oooo000@0?ooo`800000103oool3000000030?ooo`0000000000008000001P3oool3000000l0oooo
0`0000060?ooo`<000000P3oool3000000H0oooo0`00000h0?ooo`@000001P3oool3000000040?oo
o`0000000000000000H0oooo1000000?0?ooo`<000001P3oool3000000030?ooo`00000000000080
0000103oool2000007<0oooo0P0000000`3oool00000000000030?ooo`H00000IP3oool000l0oooo
0P0000070?ooo`800000303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00X0oooo
0P00000;0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool02P3oool2000000`0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`090?ooo`8000002@3oool200000780oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1U0?ooo`003`3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo06D0oooo000?0?ooo`050000003oool0oooo0?ooo`0000004P3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0=@3o
ool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool04@3o
ool01@000000oooo0?ooo`3oool000000780oooo0P0000090?ooo`<00000IP3oool00100oooo1000
000@0?ooo`<000003`3oool3000001@0oooo0`00000h0?ooo`@000004`3oool4000000l0oooo0`00
000@0?ooo`@00000K03oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`1N0?ooo`004`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0j
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo06/0oooo0P0000000`3oool00000000000040?ooo`8000002@3oool3
000000<0oooo1@00001N0?ooo`00503oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
06X0oooo00@000000?ooo`3oool000000P3oool00`000000oooo000000020?ooo`030000003oool0
oooo00T0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000
003oool0oooo05T0oooo000E0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool04P3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0JP3o
ool2000000030?ooo`000000oooo0080oooo00<000000?ooo`0000000P3oool00`000000oooo0?oo
o`090?ooo`060000003oool000000000003oool000000`3oool010000000oooo00000000001K0?oo
o`004`3oool2000000l0oooo0`000000103oool000000000000000060?ooo`@0000000@0oooo0000
000000000000503oool3000000030?ooo`000000000000800000;`3oool3000000030?ooo`000000
0000008000004`3oool4000000040?ooo`0000000000000000H0oooo0`0000020?ooo`<000003`3o
ool2000006T0oooo1@0000030?ooo`H00000303oool2000000030?ooo`000000000000<0oooo1P00
001I0?ooo`004@3oool00`000000oooo0000000B0?ooo`8000000`3oool00`000000oooo0?ooo`02
0?ooo`040000003oool0oooo0?ooo`8000006@3oool2000000<0oooo00<000000?ooo`3oool0;03o
ool00`000000oooo0?ooo`02000001X0oooo0P0000020?ooo`030000003oool0oooo00<0oooo00@0
00000?ooo`3oool0oooo0P00000C0?ooo`030000003oool0000006L0oooo00<000000?ooo`3oool0
8`3oool00`000000oooo0?ooo`1H0?ooo`003`3oool2000000030?ooo`000000000001<0oooo00<0
00000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo02/0oooo00D000000?ooo`3oool0
oooo0000000K0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0140oooo0P0000000`3oool000000000001U0?ooo`8000008`3oool30000
05T0oooo000?0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0Q0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0803oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0I@3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo05T0oooo000?0?ooo`030000003oool0oooo01P0oooo
10000000103oool0000000000000000R0?ooo`<00000:`3oool300000240oooo10000000103oool0
000000000000000K0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`1I0?ooo`00403oool00`000000oooo0?ooo`0J0?ooo`8000009`3oool00`000000
oooo0?ooo`0X0?ooo`030000003oool0oooo02D0oooo0P00000M0?ooo`030000003oool0oooo06@0
oooo0P00000S0?ooo`<00000F@3oool00140oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0V0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`1H0?ooo`003`3oool2000004D0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`140?ooo`800000I@3oool5000001d0oooo1P00001I0?ooo`003`3oool00`000000oooo0?ooo`12
0?ooo`<00000:`3oool3000004@0oooo00<000000?ooo`3oool0I@3oool010000000oooo0?ooo`00
000L0?ooo`040000003oool0oooo000005/0oooo000?0?ooo`050000003oool0oooo0?ooo`000000
0`3oool01@000000oooo0?ooo`3oool0000003P0oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0g0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0
000006T0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1L0?ooo`00403oool40000
00<0oooo1@00000d0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0:`3oool01@00
0000oooo0?ooo`3oool0000003D0oooo1@0000030?ooo`@00000JP3oool2000001/0oooo0P00001N
0?ooo`004`3oool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3oool0<@3oool2
000000<0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`02000003D0oooo00D00000
0?ooo`3oool0oooo000000030?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0A03o
ool001@0oooo0`0000050?ooo`800000;`3oool3000000030?ooo`000000000000800000;`3oool3
000000030?ooo`000000000000800000;`3oool2000000D0oooo0`00001D0?ooo`80000000<0oooo
0000000000005@3oool2000001/0oooo0P00000E0?ooo`H00000@`3oool001`0oooo00<000000?oo
o`3oool0;P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool0F03oool00`000000oooo000000020?ooo`040000003oool0oooo00000140oooo00@00000
0?ooo`3oool00000703oool010000000oooo0?ooo`00000@0?ooo`040000003oool0oooo00000080
oooo00@000000?ooo`3oool00000@@3oool001/0oooo00<000000?ooo`3oool0<03oool00`000000
oooo0?ooo`0i0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0EP3oool200000005
0?ooo`000000oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0>0?ooo`80000000<0oooo0000
003oool0703oool010000000oooo00000000000?0?ooo`80000000<0oooo0000003oool00P3oool0
10000000oooo0000000000110?ooo`006`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo03X0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1D0?ooo`H000000`3oool5
000000d0oooo1@00000M0?ooo`H00000303oool2000000030?ooo`000000000000@0oooo1@00000o
0?ooo`006`3oool01@000000oooo0?ooo`3oool0000002d0oooo0`00000h0?ooo`@00000;@3oool0
1@000000oooo0?ooo`3oool0000005L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo03h0oooo000L0?ooo`@00000;`3oool00`000000oooo0?ooo`0e
0?ooo`030000003oool0oooo02l0oooo1000001G0?ooo`<000002@3oool2000000d0oooo0P00000S
0?ooo`<00000303oool2000000T0oooo0`00000o0?ooo`007`3oool00`000000oooo0?ooo`0^0?oo
o`030000003oool0oooo03@0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`1J0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0o0?oo
o`00803oool2000002d0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0]0?ooo`80
0000E@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0200oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0g0?ooo`007`3oool2000002`0oooo0`00000h0?ooo`@00000
;03oool200000580oooo0P0000000`3oool00000000000030?ooo`<000002@3oool2000000@0oooo
0P0000000`3oool00000000000040?ooo`8000008`3oool3000000<0oooo1@0000040?ooo`800000
2@3oool3000000<0oooo1P00000f0?ooo`007`3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1=0?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`090?ooo`04
0000003oool0oooo00000080oooo00<000000?ooo`0000000P3oool010000000oooo0?ooo`00000U
0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo
0?ooo`00000:0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`0b0?ooo`007@3oool00`000000oooo0000000^0?ooo`030000003oool0oooo
03T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0000001;0?ooo`H000000`3oool20000
00030?ooo`000000000000`0oooo1P0000030?ooo`H000009P3oool2000000030?ooo`0000000000
00<0oooo1P00000<0?ooo`80000000<0oooo000000000000103oool2000000030?ooo`0000000000
0380oooo000K0?ooo`80000000<0oooo000000000000;@3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo02`0oooo0P0000000`3oool00000000000190?ooo`030000003oool0oooo07<0
oooo00<000000?ooo`3oool0<03oool001/0oooo00<000000?ooo`3oool0;`3oool3000000030?oo
o`0000000000008000001P3oool3000000040?ooo`0000000000000001@0oooo0`0000020?ooo`<0
00001P3oool3000000030?ooo`000000000000800000<@3oool00`000000oooo0?ooo`160?ooo`80
0000M@3oool200000340oooo000K0?ooo`030000003oool0oooo0340oooo0P0000030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P0000030?ooo`030000003oool0oooo0100oooo
00@000000?ooo`3oool0oooo0P0000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P00000e0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0L`3oool00`000000
oooo0?ooo`0`0?ooo`005@3oool00`000000oooo0?ooo`040?ooo`800000<P3oool00`000000oooo
0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool00000
00@0oooo00D000000?ooo`3oool0oooo0000000c0?ooo`8000001P3oool00`000000oooo0?ooo`11
0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0<03oool001<0oooo0P0000000`3o
ool00000000000030?ooo`800000=`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0=P3oool2000000<0oooo0P0000000`3oool000000000000o0?oo
o`<00000M@3oool4000002l0oooo000A0?ooo`030000003oool0000000<0oooo00D000000?ooo`00
0000oooo0000000h0?ooo`<0000000<0oooo0000000000000P0000050?ooo`@000003`3oool30000
00H0oooo0`0000020?ooo`<00000>03oool01@000000oooo0000003oool0000000<0oooo00<00000
0?ooo`000000?@3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo02d0oooo000@0?oo
o`040000003oool00000000000<0oooo0P0000000`3oool000000000000j0?ooo`800000303oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`000000>`3oool2
000000030?ooo`000000000000<0oooo0P0000000`3oool000000?ooo`0j0?ooo`030000003oool0
oooo07X0oooo00<000000?ooo`3oool0;03oool000l0oooo00D000000?ooo`3oool0oooo00000003
0?ooo`050000003oool0oooo0?ooo`000000>`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0j0?ooo`050000003o
ool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0000003/0oooo0P00001g0?ooo`03
0000003oool0000002l0oooo000?0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`000000
M@3oool010000000oooo00000000000_0?ooo`003`3oool00`000000oooo0?ooo`1?0?ooo`@00000
3`3oool300000580oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1c0?ooo`030000
003oool0oooo0300oooo000@0?ooo`800000D03oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo04l0oooo0P00000m0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0;`3o
ool000l0oooo0P00001B0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0D03oool2
000003d0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0`0?ooo`003`3oool00`00
0000oooo0?ooo`1@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0D@3oool00`00
0000oooo0?ooo`0d0?ooo`030000003oool000000080oooo00<000000?ooo`000000M@3oool01P00
0000oooo0?ooo`000000oooo00000080oooo00<000000?ooo`3oool0:03oool000l0oooo00D00000
0?ooo`3oool0oooo0000001=0?ooo`@0000000@0oooo00000000000000001P3oool4000000040?oo
o`0000000000000004h0oooo00D000000?ooo`3oool0oooo0000000f0?ooo`<000000P3oool20000
07L0oooo0`0000000`3oool0000000000002000002X0oooo000@0?ooo`040000003oool000000000
0500oooo0P0000020?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0oooo0P00001A
0?ooo`80000000<0oooo0000003oool0=P3oool00`000000oooo000000020?ooo`030000003oool0
oooo07P0oooo00<000000?ooo`0000000P3oool00`000000oooo0?ooo`0X0?ooo`004@3oool00`00
0000oooo0000001A0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`000000=P3oool300000080oooo0P00001l
0?ooo`040000003oool0oooo0?ooo`800000:@3oool001<0oooo00<000000?ooo`3oool0DP3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0>03oool0
0`000000oooo0?ooo`220?ooo`030000003oool0oooo02P0oooo000D0?ooo`800000DP3oool40000
00040?ooo`0000000000000005<0oooo0P00000h0?ooo`@00000Q03oool3000002P0oooo000C0?oo
o`800000EP3oool2000005L0oooo0P00000g0?ooo`030000003oool0oooo08L0oooo00<000000?oo
o`3oool09P3oool00140oooo00<000000?ooo`000000F03oool00`000000oooo0?ooo`1F0?ooo`03
0000003oool0000003@0oooo00<000000?ooo`3oool0R@3oool00`000000oooo0?ooo`0U0?ooo`00
3`3oool2000000030?ooo`00000000000;40oooo0P0000000`3oool000000000000c0?ooo`800000
Q`3oool2000002P0oooo000?0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0<@3o
ool010000000oooo0000000000240?ooo`030000003oool0000002P0oooo000?0?ooo`030000003o
ool0oooo0;L0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`220?ooo`030000003o
ool0oooo02P0oooo000?0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0<P3oool3
000000<0oooo00<000000?ooo`3oool0NP3oool010000000oooo0?ooo`3oool2000002T0oooo000@
0?ooo`800000]@3oool2000003H0oooo0P0000001@3oool000000000003oool0000007X0oooo00<0
00000?ooo`0000000P3oool00`000000oooo0?ooo`0X0?ooo`003`3oool200000;L0oooo0P00000f
0?ooo`<000000P3oool3000007D0oooo100000000`3oool0000000000002000002X0oooo000?0?oo
o`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1c0?oo
o`030000003oool0oooo0300oooo000?0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@00
0000oooo0?ooo`3oool0000000/0oooo00D000000?ooo`3oool0oooo000000030?ooo`050000003o
ool0oooo0?ooo`000000L@3oool01@000000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3o
ool0oooo0000000;0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3o
ool0000003d0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0`0?ooo`00403oool4
000000<0oooo1@00000;0?ooo`D000000`3oool500000740oooo1@0000030?ooo`D000002`3oool5
000000<0oooo1000000m0?ooo`800000M@3oool200000340oooo000C0?ooo`050000003oool0oooo
0?ooo`0000000`3oool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`0000000`3o
ool01@000000oooo0?ooo`3oool000000740oooo00D000000?ooo`3oool0oooo000000030?ooo`05
0000003oool0oooo0?ooo`0000002`3oool01@000000oooo0?ooo`3oool0000000<0oooo00<00000
0?ooo`3oool0?`3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0300oooo000C0?oo
o`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`090?ooo`050000003oool0
oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool000000740oooo00D000000?ooo`3oool0
oooo000000030?ooo`050000003oool0oooo0?ooo`0000002`3oool01@000000oooo0?ooo`3oool0
000000<0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0000001e0?ooo`040000003oool0
oooo000002l0oooo000D0?ooo`<000001@3oool2000000L0oooo0P0000050?ooo`<000001@3oool2
000006d0oooo0P0000050?ooo`<000001@3oool2000000L0oooo0P0000050?ooo`<00000@03oool2
000007L0oooo0`00000_0?ooo`006`3oool2000000T0oooo0P00000;0?ooo`800000K`3oool20000
00/0oooo0P0000090?ooo`800000AP3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo
02`0oooo000K0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`1_0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`150?ooo`800000M`3oool00`000000oooo0000000_0?ooo`006`3oool01@000000oooo
0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`050000003oool0oooo
0?ooo`000000J@3oool01@000000oooo0?ooo`3oool0000000/0oooo00D000000?ooo`3oool0oooo
000000030?ooo`050000003oool0oooo0?ooo`000000A`3oool00`000000oooo0000001e0?ooo`04
0000003oool00000000002l0oooo000L0?ooo`@000000`3oool4000000d0oooo1000001Y0?ooo`@0
00003@3oool4000000<0oooo1000001:0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3o
ool0<03oool001l0oooo00D000000?ooo`3oool0oooo0000000C0?ooo`030000003oool0oooo06L0
oooo00<000000?ooo`3oool04@3oool01@000000oooo0?ooo`3oool0000004`0oooo0P00001e0?oo
o`800000<@3oool00200oooo00<000000?ooo`0000005@3oool00`000000oooo0?ooo`1U0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`000000CP3oool00`000000oooo0?ooo`1c0?ooo`03
0000003oool0oooo0300oooo000Q0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
H`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0
L`3oool00`000000oooo0?ooo`0`0?ooo`00=`3oool2000006L0oooo0P00001?0?ooo`<000000P3o
ool3000000L0oooo0`0000020?ooo`<00000M@3oool4000000030?ooo`0000000000008000001P3o
ool4000000030?ooo`0000000000008000006`3oool003D0oooo00<000000?ooo`000000J@3oool0
0`000000oooo0000001=0?ooo`030000003oool000000080oooo00<000000?ooo`0000001`3oool0
0`000000oooo000000020?ooo`030000003oool0oooo07P0oooo00<000000?ooo`0000000P3oool0
0`000000oooo0?ooo`040?ooo`060000003oool0oooo0000003oool000000P3oool00`000000oooo
0?ooo`0I0?ooo`00<`3oool2000000030?ooo`000000000006T0oooo0P0000000`3oool000000000
00190?ooo`<000000P3oool200000080oooo0P0000040?ooo`<000000P3oool2000007`0oooo00@0
00000?ooo`3oool0oooo0P0000040?ooo`8000000`3oool010000000oooo0?ooo`3oool2000001X0
oooo000c0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0B@3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0PP3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool06@3oool003<0oooo00<0
00000?ooo`3oool0K`3oool00`000000oooo0?ooo`160?ooo`040000003oool0oooo000000H0oooo
00H000000?ooo`3oool000000?ooo`0000020?ooo`030000003oool0oooo0880oooo00<000000?oo
o`0000000P3oool00`000000oooo000000070?ooo`030000003oool0000001T0oooo000d0?ooo`80
0000K@3oool2000004T0oooo0`0000080?ooo`<0000000@0oooo0000000000000000QP3oool20000
0080oooo0P0000090?ooo`8000006@3oool003@0oooo00<000000?ooo`3oool0K@3oool00`000000
oooo0?ooo`160?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0R@3oool2000000d0
oooo00<000000?ooo`3oool05P3oool003<0oooo00<000000?ooo`3oool0K`3oool00`000000oooo
0?ooo`160?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool05`3oool003<0oooo00<0
00000?ooo`3oool0K`3oool00`000000oooo0?ooo`160?ooo`@00000XP3oool3000001T0oooo000c
0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0000005T0oooo
00D000000?ooo`3oool0oooo000000030?ooo`050000003oool0oooo0?ooo`000000B`3oool00`00
0000oooo0?ooo`2P0?ooo`030000003oool0oooo01T0oooo000d0?ooo`@000000`3oool5000005T0
oooo1@0000030?ooo`@00000BP3oool300000:80oooo0P00000J0?ooo`00=`3oool01@000000oooo
0?ooo`3oool0000000<0oooo00<000000?ooo`3oool0E`3oool01@000000oooo0?ooo`3oool00000
00<0oooo00<000000?ooo`3oool0C03oool200000:80oooo00<000000?ooo`3oool06@3oool003P0
oooo0`0000050?ooo`800000E@3oool2000000D0oooo0`00001@0?ooo`030000003oool0oooo0:00
oooo00<000000?ooo`3oool06@3oool003l0oooo0P00001G0?ooo`800000C`3oool4000000030?oo
o`000000000000800000XP3oool300000080oooo0`00000D0?ooo`00?`3oool00`000000oooo0?oo
o`1G0?ooo`030000003oool0oooo04d0oooo00H000000?ooo`3oool000000?ooo`00002W0?ooo`06
0000003oool0oooo0000003oool00000503oool003l0oooo00<000000?ooo`3oool0E`3oool00`00
0000oooo0?ooo`1<0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0Y`3oool20000
00<0oooo00<000000?ooo`3oool04@3oool003l0oooo00D000000?ooo`3oool0oooo0000001A0?oo
o`050000003oool0oooo0?ooo`000000C`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0
oooo0180oooo00100?ooo`@00000D@3oool4000004h0oooo00<000000?ooo`000000/@3oool01000
0000oooo0?ooo`00000A0?ooo`00@`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
04l0oooo0P00002c0?ooo`<000004@3oool004@0oooo0P00001=0?ooo`800000D03oool200000;P0
oooo00<000000?ooo`3oool03P3oool004<0oooo0P00001?0?ooo`800000D03oool200000;H0oooo
00<000000?ooo`3oool03`3oool00440oooo00<000000?ooo`000000D@3oool00`000000oooo0000
001?0?ooo`<00000/@3oool400000140oooo00100?ooo`040000003oool0000000000540oooo0P00
00000`3oool000000?ooo`1?0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool04P3o
ool003l0oooo00D000000?ooo`3oool0oooo0000001A0?ooo`050000003oool0oooo0?ooo`000000
C`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo0180oooo000o0?ooo`030000003o
ool0oooo05L0oooo00<000000?ooo`3oool0C03oool200000080oooo0`00002W0?ooo`@0000000<0
oooo0000000000004`3oool003l0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1=
0?ooo`80000000<0oooo0000003oool00P00002U0?ooo`8000000P3oool3000001@0oooo000i0?oo
o`030000003oool0oooo00@0oooo0P00001E0?ooo`8000001P3oool00`000000oooo0?ooo`170?oo
o`070000003oool000000000003oool000000?ooo`0200000:80oooo00<000000?ooo`0000000P3o
ool00`000000oooo0000000D0?ooo`00=`3oool2000000030?ooo`000000000000<0oooo0P00001G
0?ooo`8000000`3oool2000000030?ooo`000000000004l0oooo00<000000?ooo`3oool0X03oool0
0`000000oooo0?ooo`0I0?ooo`00=@3oool00`000000oooo000000030?ooo`050000003oool00000
0?ooo`000000F@3oool01@000000oooo0000003oool0000000<0oooo00<000000?ooo`000000B`3o
ool200000:@0oooo00<000000?ooo`3oool0603oool003<0oooo0P0000000`3oool0000000000003
0?ooo`80000000<0oooo000000000000F@3oool2000000030?ooo`000000000000<0oooo0P000000
0`3oool000000000001:0?ooo`800000XP3oool00`000000oooo0?ooo`0I0?ooo`00<`3oool00`00
0000oooo0?ooo`1_0?ooo`030000003oool0oooo04H0oooo00@000000?ooo`3oool00000XP3oool0
0`000000oooo0000000I0?ooo`00<`3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo
04H0oooo0`00002T0?ooo`8000006@3oool003<0oooo00<000000?ooo`3oool0K`3oool00`000000
oooo0?ooo`150?ooo`8000003@3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo00/0
oooo0P00000H0?ooo`00=03oool2000006d0oooo0P0000190?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`000000R@3oool010000000oooo00000000000;0?ooo`030000003oool0oooo01L0
oooo000c0?ooo`800000K`3oool2000004P0oooo100000060?ooo`@0000000<0oooo000000000000
0P0000240?ooo`<000000P3oool3000000L0oooo0`00000I0?ooo`00<`3oool00`000000oooo0?oo
o`1_0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo01T0oooo000c0?ooo`050000003oool0oooo0?ooo`000000J@3o
ool01@000000oooo0?ooo`3oool0000004/0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo01T0oooo000d0?ooo`@00000J@3oool4000004X0oooo0`00
00000`3oool0000000000002000000030?ooo`000000000000@0oooo0`0000000`3oool000000000
0002000007X0oooo0`0000020?ooo`800000103oool200000080oooo0`0000020?ooo`8000006P3o
ool003L0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`1=0?ooo`<000000P3oool3
000000L0oooo0`0000020?ooo`800000M`3oool010000000oooo0000003oool2000000030?ooo`00
0000oooo00D0oooo0P0000000`3oool000000?ooo`02000000030?ooo`000000oooo01X0oooo000g
0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0C@3oool00`000000oooo00000002
0?ooo`030000003oool0000000L0oooo00<000000?ooo`0000000P3oool00`000000oooo0000001e
0?ooo`070000003oool000000000003oool000000?ooo`02000000H0oooo00L000000?ooo`000000
00000?ooo`000000oooo008000006`3oool00240oooo00<000000?ooo`3oool0503oool2000006D0
oooo0P00000F0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0L`3oool00`000000
oooo0?ooo`0`0?ooo`007`3oool2000000030?ooo`000000000001<0oooo0P00001W0?ooo`800000
4`3oool2000000030?ooo`000000000004`0oooo00<000000?ooo`3oool0M@3oool00`000000oooo
0?ooo`0_0?ooo`007@3oool00`000000oooo000000030?ooo`030000003oool0000000l0oooo00<0
00000?ooo`000000J@3oool00`000000oooo0000000?0?ooo`030000003oool0000000<0oooo00<0
00000?ooo`000000B`3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0300oooo000K
0?ooo`80000000<0oooo0000000000000`3oool2000000030?ooo`000000000000/0oooo0P000000
0`3oool000000000001Y0?ooo`80000000<0oooo0000000000002`3oool2000000030?ooo`000000
000000<0oooo0P0000000`3oool00000000000170?ooo`030000003oool0000007D0oooo00@00000
0?ooo`3oool00000;`3oool001/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo04D0oooo0P00001g0?ooo`<00000;`3oool001/0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo06l0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo04@0oooo0P00001j0?oo
o`800000;P3oool001/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo06l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo04D0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`0]0?ooo`00
5@3oool00`000000oooo0?ooo`040?ooo`8000001`3oool2000000H0oooo00<000000?ooo`3oool0
103oool2000006d0oooo0P0000060?ooo`030000003oool0oooo00@0oooo0P0000070?ooo`800000
1P3oool00`000000oooo0?ooo`0o0?ooo`<00000M@3oool4000002l0oooo000C0?ooo`80000000<0
oooo0000000000000`3oool2000000T0oooo0P0000030?ooo`80000000<0oooo0000000000000`3o
ool2000006l0oooo0P0000030?ooo`80000000<0oooo0000000000000`3oool2000000T0oooo0P00
00030?ooo`80000000<0oooo000000000000@@3oool00`000000oooo0?ooo`1c0?ooo`030000003o
ool0oooo0300oooo000A0?ooo`030000003oool0000000<0oooo00D000000?ooo`000000oooo0000
000;0?ooo`050000003oool000000?ooo`0000000`3oool01@000000oooo0000003oool000000740
oooo00D000000?ooo`000000oooo000000030?ooo`050000003oool000000?ooo`0000002`3oool0
1@000000oooo0000003oool0000000<0oooo00<000000?ooo`000000?`3oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo0300oooo000?0?ooo`80000000<0oooo0000000000000`3oool2
000000030?ooo`000000000000/0oooo0P0000000`3oool00000000000030?ooo`80000000<0oooo
000000000000L@3oool2000000030?ooo`000000000000<0oooo0P0000000`3oool000000000000;
0?ooo`80000000<0oooo0000000000000`3oool2000000030?ooo`000000000003`0oooo0P00001e
0?ooo`800000<@3oool000l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0k0?oo
o`030000003oool0oooo07<0oooo00<000000?ooo`3oool0<03oool000l0oooo00<000000?ooo`3o
ool0]`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool000000080oooo00<000000?ooo`00
0000M@3oool01P000000oooo0?ooo`000000oooo00000080oooo00<000000?ooo`3oool0:03oool0
0100oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0e0?ooo`<000000P3oool20000
07L0oooo0`0000000`3oool0000000000002000002X0oooo000A0?ooo`030000003oool0oooo0;<0
oooo00<000000?ooo`3oool0=03oool2000000<0oooo0P00001l0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool09`3oool000l0oooo0P00002g0?ooo`800000=@3oool2000008@0oooo
00<000000?ooo`3oool0:03oool000l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?oo
o`0a0?ooo`040000003oool0oooo000008@0oooo00<000000?ooo`000000:03oool000l0oooo00D0
00000?ooo`3oool0oooo0000002a0?ooo`050000003oool0oooo0?ooo`000000<`3oool3000008H0
oooo0P00000X0?ooo`00403oool4000005P0oooo00<000000?ooo`3oool0EP3oool4000003<0oooo
0P0000290?ooo`8000009`3oool001<0oooo00<000000?ooo`3oool0E@3oool2000005P0oooo00<0
00000?ooo`3oool0=@3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo02H0oooo000D
0?ooo`800000DP3oool4000000040?ooo`0000000000000005<0oooo0P00000h0?ooo`800000Q`3o
ool2000002P0oooo000D0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03H0oooo00@000000?ooo`0000000000Q03o
ool00`000000oooo0000000X0?ooo`004`3oool00`000000oooo0?ooo`1?0?ooo`050000003oool0
oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04l0oooo00<0
00000?ooo`3oool0>03oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo02P0oooo000A
0?ooo`030000003oool000000500oooo0P0000020?ooo`030000003oool0oooo00<0oooo00@00000
0?ooo`3oool0oooo0P00001A0?ooo`030000003oool0000003H0oooo0`0000000`3oool000000000
0002000007X0oooo0`0000020?ooo`800000:@3oool000l0oooo0P0000000`3oool000000000001=
0?ooo`@0000000@0oooo00000000000000001P3oool4000000040?ooo`0000000000000004h0oooo
0P0000000`3oool000000000000f0?ooo`<000000P3oool2000007L0oooo00@000000?ooo`000000
oooo0P0000000`3oool000000?ooo`0Y0?ooo`003`3oool00`000000oooo0?ooo`1@0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0d0?ooo`030000
003oool000000080oooo00<000000?ooo`000000M@3oool01`000000oooo000000000000oooo0000
003oool00P00000Z0?ooo`003`3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo07<0
oooo00<000000?ooo`3oool0<03oool00100oooo0P00001@0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0C`3oool2000003d0oooo00<000000?ooo`3oool0M@3oool00`000000oooo
0?ooo`0_0?ooo`003`3oool200000500oooo1000000?0?ooo`<00000D@3oool2000003d0oooo00<0
00000?ooo`3oool0L`3oool00`000000oooo0?ooo`0`0?ooo`003`3oool00`000000oooo0?ooo`1B
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0i
0?ooo`030000003oool0000007D0oooo00@000000?ooo`3oool00000;`3oool000l0oooo00<00000
0?ooo`3oool0A@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo03T0oooo0P00001g
0?ooo`<00000;`3oool000l0oooo00D000000?ooo`3oool0oooo000000030?ooo`050000003oool0
oooo0?ooo`000000>P3oool2000000`0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`090?ooo`030000003oool0000003/0oooo00D000000?ooo`3oool0oooo000000030?ooo`050000
003oool0oooo0?ooo`000000>P3oool2000007X0oooo0P00000^0?ooo`00403oool4000000<0oooo
1@00000h0?ooo`<0000000<0oooo0000000000000P0000050?ooo`@000003`3oool3000000H0oooo
0`0000020?ooo`<00000>03oool5000000<0oooo1000000l0?ooo`030000003oool0oooo07P0oooo
00<000000?ooo`3oool0;@3oool001<0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000
003oool0oooo03H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo03L0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo03d0
oooo0P00001g0?ooo`030000003oool0000002l0oooo000D0?ooo`<000001@3oool200000380oooo
00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3o
ool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000<`3oool2000000D0oooo0`000010
0?ooo`030000003oool0000007D0oooo00@000000?ooo`0000000000;`3oool001/0oooo0P00000b
0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3oool0
0`000000oooo0?ooo`0@0?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00800000=03oool2000004T0oooo00<000000?ooo`3oool0L`3o
ool00`000000oooo0?ooo`0`0?ooo`006`3oool00`000000oooo0?ooo`0_0?ooo`<0000000<0oooo
0000000000000P0000060?ooo`<0000000@0oooo0000000000000000503oool300000080oooo0`00
00060?ooo`<0000000<0oooo0000000000000P00000a0?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool0M@3oool00`000000oooo0?ooo`0_0?ooo`006`3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`17
0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0<03oool001/0oooo00D000000?oo
o`3oool0oooo0000000^0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0;03oool0
1@000000oooo0?ooo`3oool0000004T0oooo1P0000030?ooo`80000000<0oooo000000000000303o
ool6000000<0oooo1P00000V0?ooo`80000000<0oooo0000000000000`3oool6000000`0oooo0P00
00000`3oool00000000000040?ooo`80000000<0oooo000000000000<P3oool001`0oooo1000000]
0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0;03oool4000004`0oooo00@00000
0?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00T0oooo
00@000000?ooo`3oool000000P3oool00`000000oooo000000020?ooo`040000003oool0oooo0000
02D0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003o
ool0oooo000000X0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?oo
o`030000003oool0oooo0380oooo000O0?ooo`030000003oool0oooo02/0oooo0`00000h0?ooo`@0
0000;@3oool00`000000oooo0?ooo`1@0?ooo`040000003oool0oooo000000@0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo000000D0oooo00<00000
0?ooo`3oool08`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
0000000g0?ooo`00803oool2000002d0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?oo
o`0]0?ooo`800000D`3oool2000000030?ooo`000000000000<0oooo0`0000090?ooo`800000103o
ool2000000030?ooo`000000000000@0oooo0P00000S0?ooo`<000000`3oool5000000@0oooo0P00
00090?ooo`<000000`3oool6000003H0oooo000O0?ooo`800000;`3oool00`000000oooo0?ooo`0d
0?ooo`030000003oool0oooo02h0oooo0P00001D0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0803oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03L0oooo000M0?oo
o`030000003oool0000002l0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0000005P0oooo0`0000090?ooo`8000003@3oool2000002<0oooo0`00000<0?oo
o`8000002@3oool3000003l0oooo000K0?ooo`80000000<0oooo000000000000;@3oool3000003P0
oooo1000000]0?ooo`80000000<0oooo000000000000E`3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0?P3oool001/0oooo00<000000?ooo`3oool0
;`3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0
E03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
8`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
?P3oool001/0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0i0?ooo`030000003o
ool0oooo0300oooo00<000000?ooo`3oool0E03oool6000000<0oooo1@00000=0?ooo`D000007@3o
ool6000000`0oooo0P0000000`3oool00000000000040?ooo`D00000?`3oool001/0oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0300oooo00<00000
0?ooo`3oool0E`3oool00`000000oooo000000020?ooo`040000003oool0oooo00000140oooo00@0
00000?ooo`3oool00000703oool010000000oooo0?ooo`00000@0?ooo`040000003oool0oooo0000
0080oooo00@000000?ooo`3oool00000@@3oool001D0oooo00<000000?ooo`3oool0103oool20000
02l0oooo0`0000000`3oool0000000000002000002l0oooo0`0000000`3oool00000000000020000
02l0oooo0P0000060?ooo`030000003oool0oooo05<0oooo0P0000000`3oool000000000000E0?oo
o`8000006`3oool2000001D0oooo1P0000130?ooo`004`3oool2000000030?ooo`000000000000<0
oooo0P00000b0?ooo`8000000`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0080
0000=03oool2000000<0oooo0P0000000`3oool000000000001E0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo04@0
oooo000A0?ooo`030000003oool0000000<0oooo00D000000?ooo`000000oooo0000000d0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0:`3oool01@000000oooo0?ooo`3oool00000
03D0oooo00D000000?ooo`000000oooo000000030?ooo`030000003oool0000006`0oooo00<00000
0?ooo`3oool0603oool00`000000oooo0?ooo`1N0?ooo`003`3oool2000000030?ooo`0000000000
00<0oooo0P0000000`3oool000000000000h0?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool0=`3oool2000000030?ooo`000000000000<0oooo0P0000000`3oool000000000001Y0?oo
o`8000006`3oool2000005h0oooo000?0?ooo`030000003oool0oooo0480oooo0`00000[0?ooo`<0
0000A03oool00`000000oooo0?ooo`1U0?ooo`040000003oool0oooo000001`0oooo00@000000?oo
o`3oool00000F`3oool000l0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0X0?oo
o`030000003oool0oooo04D0oooo00<000000?ooo`3oool0H`3oool5000001d0oooo1P00001I0?oo
o`003`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3o
ool09P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0H`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo05P0oooo000@0?ooo`800000
6`3oool2000002L0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0U0?ooo`800000
703oool2000006H0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1H0?ooo`003`3o
ool2000001T0oooo10000000103oool0000000000000000R0?ooo`<00000:`3oool300000240oooo
10000000103oool0000000000000000J0?ooo`800000I@3oool2000002<0oooo0`00001I0?ooo`00
3`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`1I0?ooo`003`3oool01@000000oooo0?ooo`3oool0000001<0oooo
00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo02/0oooo00D000000?ooo`3o
ool0oooo0000000K0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0140oooo00D000000?ooo`3oool0oooo0000001U0?ooo`8000008`3o
ool3000005T0oooo000@0?ooo`@000004P3oool2000000<0oooo00<000000?ooo`3oool00P3oool0
10000000oooo0?ooo`3oool2000001T0oooo0P0000030?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool00P00000J0?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`040000003o
ool0oooo0?ooo`8000004`3oool4000006H0oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`1H0?ooo`004`3oool00`000000oooo0?ooo`0>0?ooo`<0000000@0oooo0000000000000000
1P3oool4000000040?ooo`0000000000000001@0oooo0`0000000`3oool0000000000002000002l0
oooo0`0000000`3oool0000000000002000001<0oooo10000000103oool000000000000000060?oo
o`<000000P3oool300000100oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0S0?oo
o`030000003oool0oooo05P0oooo000C0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool0I`3oool5000000<0oooo1P00000<0?ooo`80000000<0oooo0000000000000`3oool6000005T0
oooo000D0?ooo`8000003`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool03@3oool2000006`0oooo00@000000?ooo`3oool000000P3oool00`000000
oooo000000020?ooo`030000003oool0oooo00T0oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo05T0oooo000C0?ooo`8000003`3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0>P3oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool03P3oool20000
06d0oooo0P0000000`3oool00000000000040?ooo`8000002@3oool3000000<0oooo1@00001N0?oo
o`004@3oool00`000000oooo0000000@0?ooo`<000003`3oool3000001@0oooo0`00000h0?ooo`@0
00004`3oool4000000l0oooo0`00000@0?ooo`030000003oool0000006d0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0GP3oool000l0oooo0P0000000`3oool000000000000B0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0A0?ooo`80000000<0oooo000000000000
L`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo06H0oooo000?0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0d0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`1`0?ooo`8000002@3o
ool3000006H0oooo000?0?ooo`030000003oool0oooo00H0oooo0P00000<0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool02P3oool2000000/0oooo00<000000?ooo`3oool0=@3oool0
0`000000oooo0?ooo`0:0?ooo`800000303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo00T0oooo0P00000:0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`1U0?ooo`00403oool00`000000oooo0?ooo`030?ooo`<0000000<0oooo0000
000000000P0000060?ooo`<000003`3oool3000000H0oooo0`0000020?ooo`<000001P3oool30000
03P0oooo100000060?ooo`<0000000@0oooo00000000000000001P3oool4000000l0oooo0`000006
0?ooo`<0000000<0oooo0000000000000P0000050?ooo`030000003oool0oooo0740oooo0P000000
0`3oool00000000000030?ooo`H00000IP3oool00140oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0M03oool01000
0000oooo0?ooo`0000020?ooo`040000003oool0oooo000006P0oooo000?0?ooo`80000000<0oooo
0000003oool00`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo
000000040?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0i0?ooo`05
0000003oool0oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0000000@0oooo00D00000
0?ooo`3oool0oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo
0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000P3oool2000007H0oooo00D000000?ooo`3o
ool0oooo0000001[0?ooo`003`3oool01@000000oooo0000003oool000000080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3oool0403oool01000
0000oooo0?ooo`3oool200000080oooo00<000000?ooo`3oool00`3oool01P000000oooo0?ooo`00
0000oooo00000080oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`02000000<0oooo
00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3o
ool04@3oool01P000000oooo0?ooo`000000oooo00000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3oool0M03oool2000000030?ooo`000000
000006/0oooo000?0?ooo`<000000P3oool3000000H0oooo0`0000000`3oool00000000000020000
01<0oooo10000000103oool000000000000000060?ooo`<000000P3oool3000003d0oooo0`000000
0`3oool0000000000002000000D0oooo10000000103oool0000000000000000D0?ooo`<000000P3o
ool3000000H0oooo0`0000000`3oool0000000000002000007P0oooo00<000000?ooo`3oool0J`3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {204.688, 0}} -> {-0.0528339, -0.0238146, \
0.00732832, 0.00732832}, {{10.8125, 205.75}, {199.813, 4.875}} -> \
{-0.0832456, -0.0512639, 0.00807959, 0.00807959}, {{225.188, 420.125}, \
{199.813, 4.875}} -> {-2.24136, -0.369536, 0.0127444, 0.0127444}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Figures based upon higher-order polygons", "Subsection"],

Cell[BoxData[
    \(Needs["\<Geometry`Polytopes`\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayKoch[Vertices[Dodecagon], 4, ShowIterations \[Rule] False, 
        KochType \[Rule] "\<snowflake\>", 
        CloseFigure \[Rule] True];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.47619 0.5 0.47619 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.91239 .7381 m
.91024 .74025 L
s
.91024 .74025 m
.9073 .73946 L
s
.9073 .73946 m
.90809 .7424 L
s
.90809 .7424 m
.90594 .74455 L
s
.90594 .74455 m
.903 .74376 L
s
.903 .74376 m
.90221 .74082 L
s
.90221 .74082 m
.90006 .74298 L
s
.90006 .74298 m
.89712 .74219 L
s
.89712 .74219 m
.89791 .74513 L
s
.89791 .74513 m
.89576 .74728 L
s
.89576 .74728 m
.89869 .74807 L
s
.89869 .74807 m
.89948 .75101 L
s
.89948 .75101 m
.89733 .75316 L
s
.89733 .75316 m
.89439 .75237 L
s
.89439 .75237 m
.89518 .75531 L
s
.89518 .75531 m
.89303 .75746 L
s
.89303 .75746 m
.89009 .75667 L
s
.89009 .75667 m
.8893 .75373 L
s
.8893 .75373 m
.88715 .75589 L
s
.88715 .75589 m
.88421 .7551 L
s
.88421 .7551 m
.88342 .75216 L
s
.88342 .75216 m
.88557 .75001 L
s
.88557 .75001 m
.88263 .74922 L
s
.88263 .74922 m
.88185 .74628 L
s
.88185 .74628 m
.87969 .74843 L
s
.87969 .74843 m
.87675 .74764 L
s
.87675 .74764 m
.87754 .75058 L
s
.87754 .75058 m
.87539 .75274 L
s
.87539 .75274 m
.87245 .75195 L
s
.87245 .75195 m
.87166 .74901 L
s
.87166 .74901 m
.86951 .75116 L
s
.86951 .75116 m
.86657 .75037 L
s
.86657 .75037 m
.86736 .75331 L
s
.86736 .75331 m
.86521 .75546 L
s
.86521 .75546 m
.86815 .75625 L
s
.86815 .75625 m
.86893 .75919 L
s
.86893 .75919 m
.86678 .76134 L
s
.86678 .76134 m
.86384 .76056 L
s
.86384 .76056 m
.86463 .7635 L
s
.86463 .7635 m
.86248 .76565 L
s
.86248 .76565 m
.86542 .76643 L
s
.86542 .76643 m
.86621 .76937 L
s
.86621 .76937 m
.86836 .76722 L
s
.86836 .76722 m
.8713 .76801 L
s
.8713 .76801 m
.87208 .77095 L
s
.87208 .77095 m
.86993 .7731 L
s
.86993 .7731 m
.87287 .77389 L
s
.87287 .77389 m
.87366 .77683 L
s
.87366 .77683 m
.87151 .77898 L
s
.87151 .77898 m
.86857 .77819 L
s
.86857 .77819 m
.86936 .78113 L
s
.86936 .78113 m
.8672 .78328 L
s
.8672 .78328 m
.86427 .7825 L
s
.86427 .7825 m
.86348 .77956 L
s
.86348 .77956 m
.86133 .78171 L
s
.86133 .78171 m
.85839 .78092 L
s
.85839 .78092 m
.85917 .78386 L
s
.85917 .78386 m
.85702 .78601 L
s
.85702 .78601 m
.85996 .7868 L
s
.85996 .7868 m
.86075 .78974 L
s
.86075 .78974 m
.8586 .79189 L
s
.8586 .79189 m
.85566 .7911 L
s
.85566 .7911 m
.85645 .79404 L
s
.85645 .79404 m
.85429 .79619 L
s
.85429 .79619 m
.85135 .79541 L
s
.85135 .79541 m
.85057 .79247 L
s
.85057 .79247 m
.84841 .79462 L
s
.84841 .79462 m
.84548 .79383 L
s
.84548 .79383 m
.84469 .79089 L
s
.84469 .79089 m
.84684 .78874 L
s
.84684 .78874 m
.8439 .78795 L
s
.8439 .78795 m
.84311 .78501 L
s
.84311 .78501 m
.84096 .78717 L
s
.84096 .78717 m
.83802 .78638 L
s
.83802 .78638 m
.83881 .78932 L
s
.83881 .78932 m
.83666 .79147 L
s
.83666 .79147 m
.83372 .79068 L
s
.83372 .79068 m
.83293 .78774 L
s
.83293 .78774 m
.83078 .78989 L
s
.83078 .78989 m
.82784 .78911 L
s
.82784 .78911 m
.82705 .78617 L
s
.82705 .78617 m
.8292 .78401 L
s
.8292 .78401 m
.82626 .78323 L
s
.82626 .78323 m
.82548 .78029 L
s
.82548 .78029 m
.82763 .77814 L
s
.82763 .77814 m
.83057 .77892 L
s
.83057 .77892 m
.82978 .77598 L
s
.82978 .77598 m
.83193 .77383 L
s
.83193 .77383 m
.82899 .77304 L
s
.82899 .77304 m
.8282 .77011 L
s
.8282 .77011 m
.82605 .77226 L
s
.82605 .77226 m
.82311 .77147 L
s
.82311 .77147 m
.82233 .76853 L
s
.82233 .76853 m
.82448 .76638 L
s
.82448 .76638 m
.82154 .76559 L
s
.82154 .76559 m
.82075 .76265 L
s
.82075 .76265 m
.8186 .7648 L
s
.8186 .7648 m
.81566 .76402 L
s
.81566 .76402 m
.81645 .76695 L
s
.81645 .76695 m
.81429 .76911 L
s
.81429 .76911 m
.81136 .76832 L
s
.81136 .76832 m
.81057 .76538 L
s
.81057 .76538 m
.80842 .76753 L
s
.80842 .76753 m
.80548 .76674 L
s
.80548 .76674 m
.80626 .76968 L
s
.80626 .76968 m
.80411 .77183 L
s
.80411 .77183 m
.80705 .77262 L
s
.80705 .77262 m
.80784 .77556 L
s
.80784 .77556 m
.80569 .77771 L
s
.80569 .77771 m
.80275 .77693 L
s
.80275 .77693 m
.80354 .77987 L
s
.80354 .77987 m
.80138 .78202 L
s
.80138 .78202 m
.79844 .78123 L
s
.79844 .78123 m
.79766 .77829 L
s
.79766 .77829 m
.7955 .78044 L
s
.7955 .78044 m
.79257 .77965 L
s
.79257 .77965 m
.79178 .77671 L
s
.79178 .77671 m
.79393 .77456 L
s
.79393 .77456 m
.79099 .77378 L
s
.79099 .77378 m
.7902 .77084 L
s
.7902 .77084 m
.78805 .77299 L
s
.78805 .77299 m
.78511 .7722 L
s
.78511 .7722 m
.7859 .77514 L
s
.7859 .77514 m
.78375 .77729 L
s
.78375 .77729 m
.78081 .7765 L
s
.78081 .7765 m
.78002 .77356 L
s
.78002 .77356 m
.77787 .77572 L
s
.77787 .77572 m
.77493 .77493 L
s
.77493 .77493 m
.77572 .77787 L
s
.77572 .77787 m
.77356 .78002 L
s
.77356 .78002 m
.7765 .78081 L
s
.7765 .78081 m
.77729 .78375 L
s
.77729 .78375 m
.77514 .7859 L
s
.77514 .7859 m
.7722 .78511 L
s
.7722 .78511 m
.77299 .78805 L
s
.77299 .78805 m
.77084 .7902 L
s
.77084 .7902 m
.77378 .79099 L
s
.77378 .79099 m
.77456 .79393 L
s
.77456 .79393 m
.77671 .79178 L
s
.77671 .79178 m
.77965 .79257 L
s
.77965 .79257 m
.78044 .7955 L
s
.78044 .7955 m
.77829 .79766 L
s
.77829 .79766 m
.78123 .79844 L
s
.78123 .79844 m
.78202 .80138 L
s
.78202 .80138 m
.77987 .80354 L
s
.77987 .80354 m
.77693 .80275 L
s
.77693 .80275 m
.77771 .80569 L
s
.77771 .80569 m
.77556 .80784 L
s
.77556 .80784 m
.77262 .80705 L
s
.77262 .80705 m
.77183 .80411 L
s
.77183 .80411 m
.76968 .80626 L
s
.76968 .80626 m
.76674 .80548 L
s
.76674 .80548 m
.76753 .80842 L
s
.76753 .80842 m
.76538 .81057 L
s
.76538 .81057 m
.76832 .81136 L
s
.76832 .81136 m
.76911 .81429 L
s
.76911 .81429 m
.76695 .81645 L
s
.76695 .81645 m
.76402 .81566 L
s
.76402 .81566 m
.7648 .8186 L
s
.7648 .8186 m
.76265 .82075 L
s
.76265 .82075 m
.76559 .82154 L
s
.76559 .82154 m
.76638 .82448 L
s
.76638 .82448 m
.76853 .82233 L
s
.76853 .82233 m
.77147 .82311 L
s
.77147 .82311 m
.77226 .82605 L
s
.77226 .82605 m
.77011 .8282 L
s
.77011 .8282 m
.77304 .82899 L
s
.77304 .82899 m
.77383 .83193 L
s
.77383 .83193 m
.77598 .82978 L
s
.77598 .82978 m
.77892 .83057 L
s
.77892 .83057 m
.77814 .82763 L
s
.77814 .82763 m
.78029 .82548 L
s
.78029 .82548 m
.78323 .82626 L
s
.78323 .82626 m
.78401 .8292 L
s
.78401 .8292 m
.78617 .82705 L
s
.78617 .82705 m
.78911 .82784 L
s
.78911 .82784 m
.78989 .83078 L
s
.78989 .83078 m
.78774 .83293 L
s
.78774 .83293 m
.79068 .83372 L
s
.79068 .83372 m
.79147 .83666 L
s
.79147 .83666 m
.78932 .83881 L
s
.78932 .83881 m
.78638 .83802 L
s
.78638 .83802 m
.78717 .84096 L
s
.78717 .84096 m
.78501 .84311 L
s
.78501 .84311 m
.78795 .8439 L
s
.78795 .8439 m
.78874 .84684 L
s
.78874 .84684 m
.79089 .84469 L
s
.79089 .84469 m
.79383 .84548 L
s
.79383 .84548 m
.79462 .84841 L
s
.79462 .84841 m
.79247 .85057 L
s
.79247 .85057 m
.79541 .85135 L
s
.79541 .85135 m
.79619 .85429 L
s
.79619 .85429 m
.79404 .85645 L
s
.79404 .85645 m
.7911 .85566 L
s
.7911 .85566 m
.79189 .8586 L
s
.79189 .8586 m
.78974 .86075 L
s
.78974 .86075 m
.7868 .85996 L
s
.7868 .85996 m
.78601 .85702 L
s
.78601 .85702 m
.78386 .85917 L
s
.78386 .85917 m
.78092 .85839 L
s
.78092 .85839 m
.78171 .86133 L
s
.78171 .86133 m
.77956 .86348 L
s
.77956 .86348 m
.7825 .86427 L
s
.7825 .86427 m
.78328 .8672 L
s
.78328 .8672 m
.78113 .86936 L
s
.78113 .86936 m
.77819 .86857 L
s
.77819 .86857 m
.77898 .87151 L
s
.77898 .87151 m
.77683 .87366 L
s
.77683 .87366 m
.77389 .87287 L
s
.77389 .87287 m
.7731 .86993 L
s
.7731 .86993 m
.77095 .87208 L
s
.77095 .87208 m
.76801 .8713 L
s
.76801 .8713 m
.76722 .86836 L
s
.76722 .86836 m
.76937 .86621 L
s
.76937 .86621 m
.76643 .86542 L
s
.76643 .86542 m
.76565 .86248 L
s
.76565 .86248 m
.7635 .86463 L
s
.7635 .86463 m
.76056 .86384 L
s
.76056 .86384 m
.76134 .86678 L
s
.76134 .86678 m
.75919 .86893 L
s
.75919 .86893 m
.75625 .86815 L
s
.75625 .86815 m
.75546 .86521 L
s
.75546 .86521 m
.75331 .86736 L
s
.75331 .86736 m
.75037 .86657 L
s
.75037 .86657 m
.75116 .86951 L
s
.75116 .86951 m
.74901 .87166 L
s
.74901 .87166 m
.75195 .87245 L
s
.75195 .87245 m
.75274 .87539 L
s
.75274 .87539 m
.75058 .87754 L
s
.75058 .87754 m
.74764 .87675 L
s
.74764 .87675 m
.74843 .87969 L
s
.74843 .87969 m
.74628 .88185 L
s
.74628 .88185 m
.74922 .88263 L
s
.74922 .88263 m
.75001 .88557 L
s
.75001 .88557 m
.75216 .88342 L
s
.75216 .88342 m
.7551 .88421 L
s
.7551 .88421 m
.75589 .88715 L
s
.75589 .88715 m
.75373 .8893 L
s
.75373 .8893 m
.75667 .89009 L
s
.75667 .89009 m
.75746 .89303 L
s
.75746 .89303 m
.75531 .89518 L
s
.75531 .89518 m
.75237 .89439 L
s
.75237 .89439 m
.75316 .89733 L
s
.75316 .89733 m
.75101 .89948 L
s
.75101 .89948 m
.74807 .89869 L
s
.74807 .89869 m
.74728 .89576 L
s
.74728 .89576 m
.74513 .89791 L
s
.74513 .89791 m
.74219 .89712 L
s
.74219 .89712 m
.74298 .90006 L
s
.74298 .90006 m
.74082 .90221 L
s
.74082 .90221 m
.74376 .903 L
s
.74376 .903 m
.74455 .90594 L
s
.74455 .90594 m
.7424 .90809 L
s
.7424 .90809 m
.73946 .9073 L
s
.73946 .9073 m
.74025 .91024 L
s
.74025 .91024 m
.7381 .91239 L
s
.7381 .91239 m
.73516 .91318 L
s
.73516 .91318 m
.733 .91103 L
s
.733 .91103 m
.73222 .91397 L
s
.73222 .91397 m
.72928 .91476 L
s
.72928 .91476 m
.72713 .9126 L
s
.72713 .9126 m
.72791 .90966 L
s
.72791 .90966 m
.72497 .91045 L
s
.72497 .91045 m
.72282 .9083 L
s
.72282 .9083 m
.72203 .91124 L
s
.72203 .91124 m
.71909 .91203 L
s
.71909 .91203 m
.72125 .91418 L
s
.72125 .91418 m
.72046 .91712 L
s
.72046 .91712 m
.71752 .91791 L
s
.71752 .91791 m
.71537 .91575 L
s
.71537 .91575 m
.71458 .91869 L
s
.71458 .91869 m
.71164 .91948 L
s
.71164 .91948 m
.70949 .91733 L
s
.70949 .91733 m
.71028 .91439 L
s
.71028 .91439 m
.70734 .91518 L
s
.70734 .91518 m
.70518 .91303 L
s
.70518 .91303 m
.70597 .91009 L
s
.70597 .91009 m
.70891 .9093 L
s
.70891 .9093 m
.70676 .90715 L
s
.70676 .90715 m
.70755 .90421 L
s
.70755 .90421 m
.70461 .905 L
s
.70461 .905 m
.70246 .90284 L
s
.70246 .90284 m
.70167 .90578 L
s
.70167 .90578 m
.69873 .90657 L
s
.69873 .90657 m
.69658 .90442 L
s
.69658 .90442 m
.69737 .90148 L
s
.69737 .90148 m
.69443 .90227 L
s
.69443 .90227 m
.69227 .90012 L
s
.69227 .90012 m
.69149 .90305 L
s
.69149 .90305 m
.68855 .90384 L
s
.68855 .90384 m
.6907 .90599 L
s
.6907 .90599 m
.68991 .90893 L
s
.68991 .90893 m
.68697 .90972 L
s
.68697 .90972 m
.68482 .90757 L
s
.68482 .90757 m
.68403 .91051 L
s
.68403 .91051 m
.68109 .9113 L
s
.68109 .9113 m
.68324 .91345 L
s
.68324 .91345 m
.68246 .91639 L
s
.68246 .91639 m
.6854 .9156 L
s
.6854 .9156 m
.68755 .91775 L
s
.68755 .91775 m
.68676 .92069 L
s
.68676 .92069 m
.68382 .92148 L
s
.68382 .92148 m
.68597 .92363 L
s
.68597 .92363 m
.68519 .92657 L
s
.68519 .92657 m
.68225 .92736 L
s
.68225 .92736 m
.68009 .92521 L
s
.68009 .92521 m
.67931 .92815 L
s
.67931 .92815 m
.67637 .92893 L
s
.67637 .92893 m
.67422 .92678 L
s
.67422 .92678 m
.675 .92384 L
s
.675 .92384 m
.67206 .92463 L
s
.67206 .92463 m
.66991 .92248 L
s
.66991 .92248 m
.66912 .92542 L
s
.66912 .92542 m
.66618 .9262 L
s
.66618 .9262 m
.66834 .92836 L
s
.66834 .92836 m
.66755 .9313 L
s
.66755 .9313 m
.66461 .93208 L
s
.66461 .93208 m
.66246 .92993 L
s
.66246 .92993 m
.66167 .93287 L
s
.66167 .93287 m
.65873 .93366 L
s
.65873 .93366 m
.65658 .93151 L
s
.65658 .93151 m
.65737 .92857 L
s
.65737 .92857 m
.65443 .92936 L
s
.65443 .92936 m
.65227 .9272 L
s
.65227 .9272 m
.65306 .92426 L
s
.65306 .92426 m
.656 .92348 L
s
.656 .92348 m
.65385 .92132 L
s
.65385 .92132 m
.65464 .91839 L
s
.65464 .91839 m
.6517 .91917 L
s
.6517 .91917 m
.64955 .91702 L
s
.64955 .91702 m
.64876 .91996 L
s
.64876 .91996 m
.64582 .92075 L
s
.64582 .92075 m
.64367 .9186 L
s
.64367 .9186 m
.64446 .91566 L
s
.64446 .91566 m
.64152 .91644 L
s
.64152 .91644 m
.63936 .91429 L
s
.63936 .91429 m
.64015 .91135 L
s
.64015 .91135 m
.64309 .91057 L
s
.64309 .91057 m
.64094 .90841 L
s
.64094 .90841 m
.64173 .90547 L
s
.64173 .90547 m
.64467 .90469 L
s
.64467 .90469 m
.64682 .90684 L
s
.64682 .90684 m
.64761 .9039 L
s
.64761 .9039 m
.65054 .90311 L
s
.65054 .90311 m
.64839 .90096 L
s
.64839 .90096 m
.64918 .89802 L
s
.64918 .89802 m
.64624 .89881 L
s
.64624 .89881 m
.64409 .89666 L
s
.64409 .89666 m
.64488 .89372 L
s
.64488 .89372 m
.64782 .89293 L
s
.64782 .89293 m
.64566 .89078 L
s
.64566 .89078 m
.64645 .88784 L
s
.64645 .88784 m
.64351 .88863 L
s
.64351 .88863 m
.64136 .88647 L
s
.64136 .88647 m
.64057 .88941 L
s
.64057 .88941 m
.63763 .8902 L
s
.63763 .8902 m
.63548 .88805 L
s
.63548 .88805 m
.63627 .88511 L
s
.63627 .88511 m
.63333 .8859 L
s
.63333 .8859 m
.63118 .88374 L
s
.63118 .88374 m
.63039 .88668 L
s
.63039 .88668 m
.62745 .88747 L
s
.62745 .88747 m
.6296 .88962 L
s
.6296 .88962 m
.62882 .89256 L
s
.62882 .89256 m
.62588 .89335 L
s
.62588 .89335 m
.62372 .8912 L
s
.62372 .8912 m
.62294 .89414 L
s
.62294 .89414 m
.62 .89493 L
s
.62 .89493 m
.61785 .89277 L
s
.61785 .89277 m
.61863 .88983 L
s
.61863 .88983 m
.61569 .89062 L
s
.61569 .89062 m
.61354 .88847 L
s
.61354 .88847 m
.61433 .88553 L
s
.61433 .88553 m
.61727 .88474 L
s
.61727 .88474 m
.61512 .88259 L
s
.61512 .88259 m
.6159 .87965 L
s
.6159 .87965 m
.61297 .88044 L
s
.61297 .88044 m
.61081 .87829 L
s
.61081 .87829 m
.61003 .88123 L
s
.61003 .88123 m
.60709 .88202 L
s
.60709 .88202 m
.60493 .87986 L
s
.60493 .87986 m
.60572 .87692 L
s
.60572 .87692 m
.60278 .87771 L
s
.60278 .87771 m
.60063 .87556 L
s
.60063 .87556 m
.59984 .8785 L
s
.59984 .8785 m
.5969 .87929 L
s
.5969 .87929 m
.59906 .88144 L
s
.59906 .88144 m
.59827 .88438 L
s
.59827 .88438 m
.59533 .88517 L
s
.59533 .88517 m
.59318 .88301 L
s
.59318 .88301 m
.59239 .88595 L
s
.59239 .88595 m
.58945 .88674 L
s
.58945 .88674 m
.5916 .88889 L
s
.5916 .88889 m
.59081 .89183 L
s
.59081 .89183 m
.59375 .89104 L
s
.59375 .89104 m
.59591 .8932 L
s
.59591 .8932 m
.59512 .89614 L
s
.59512 .89614 m
.59218 .89692 L
s
.59218 .89692 m
.59433 .89908 L
s
.59433 .89908 m
.59354 .90201 L
s
.59354 .90201 m
.5906 .9028 L
s
.5906 .9028 m
.58845 .90065 L
s
.58845 .90065 m
.58766 .90359 L
s
.58766 .90359 m
.58472 .90438 L
s
.58472 .90438 m
.58257 .90223 L
s
.58257 .90223 m
.58336 .89929 L
s
.58336 .89929 m
.58042 .90007 L
s
.58042 .90007 m
.57827 .89792 L
s
.57827 .89792 m
.57748 .90086 L
s
.57748 .90086 m
.57454 .90165 L
s
.57454 .90165 m
.57669 .9038 L
s
.57669 .9038 m
.57591 .90674 L
s
.57591 .90674 m
.57297 .90753 L
s
.57297 .90753 m
.57081 .90538 L
s
.57081 .90538 m
.57003 .90832 L
s
.57003 .90832 m
.56709 .9091 L
s
.56709 .9091 m
.56924 .91126 L
s
.56924 .91126 m
.56845 .91419 L
s
.56845 .91419 m
.57139 .91341 L
s
.57139 .91341 m
.57354 .91556 L
s
.57354 .91556 m
.57276 .9185 L
s
.57276 .9185 m
.56982 .91929 L
s
.56982 .91929 m
.57197 .92144 L
s
.57197 .92144 m
.57118 .92438 L
s
.57118 .92438 m
.57412 .92359 L
s
.57412 .92359 m
.57627 .92574 L
s
.57627 .92574 m
.57706 .9228 L
s
.57706 .9228 m
.58 .92201 L
s
.58 .92201 m
.58215 .92417 L
s
.58215 .92417 m
.58136 .92711 L
s
.58136 .92711 m
.5843 .92632 L
s
.5843 .92632 m
.58645 .92847 L
s
.58645 .92847 m
.58567 .93141 L
s
.58567 .93141 m
.58273 .9322 L
s
.58273 .9322 m
.58488 .93435 L
s
.58488 .93435 m
.58409 .93729 L
s
.58409 .93729 m
.58115 .93808 L
s
.58115 .93808 m
.579 .93592 L
s
.579 .93592 m
.57821 .93886 L
s
.57821 .93886 m
.57527 .93965 L
s
.57527 .93965 m
.57742 .9418 L
s
.57742 .9418 m
.57664 .94474 L
s
.57664 .94474 m
.57958 .94395 L
s
.57958 .94395 m
.58173 .94611 L
s
.58173 .94611 m
.58094 .94905 L
s
.58094 .94905 m
.578 .94983 L
s
.578 .94983 m
.58015 .95199 L
s
.58015 .95199 m
.57937 .95492 L
s
.57937 .95492 m
.57643 .95571 L
s
.57643 .95571 m
.57427 .95356 L
s
.57427 .95356 m
.57349 .9565 L
s
.57349 .9565 m
.57055 .95729 L
s
.57055 .95729 m
.56839 .95514 L
s
.56839 .95514 m
.56918 .9522 L
s
.56918 .9522 m
.56624 .95298 L
s
.56624 .95298 m
.56409 .95083 L
s
.56409 .95083 m
.5633 .95377 L
s
.5633 .95377 m
.56036 .95456 L
s
.56036 .95456 m
.56252 .95671 L
s
.56252 .95671 m
.56173 .95965 L
s
.56173 .95965 m
.55879 .96044 L
s
.55879 .96044 m
.55664 .95829 L
s
.55664 .95829 m
.55585 .96123 L
s
.55585 .96123 m
.55291 .96201 L
s
.55291 .96201 m
.55076 .95986 L
s
.55076 .95986 m
.55155 .95692 L
s
.55155 .95692 m
.54861 .95771 L
s
.54861 .95771 m
.54645 .95556 L
s
.54645 .95556 m
.54724 .95262 L
s
.54724 .95262 m
.55018 .95183 L
s
.55018 .95183 m
.54803 .94968 L
s
.54803 .94968 m
.54882 .94674 L
s
.54882 .94674 m
.54588 .94753 L
s
.54588 .94753 m
.54373 .94538 L
s
.54373 .94538 m
.54294 .94831 L
s
.54294 .94831 m
.54 .9491 L
s
.54 .9491 m
.53785 .94695 L
s
.53785 .94695 m
.53863 .94401 L
s
.53863 .94401 m
.5357 .9448 L
s
.5357 .9448 m
.53354 .94265 L
s
.53354 .94265 m
.53276 .94559 L
s
.53276 .94559 m
.52982 .94637 L
s
.52982 .94637 m
.53197 .94853 L
s
.53197 .94853 m
.53118 .95147 L
s
.53118 .95147 m
.52824 .95225 L
s
.52824 .95225 m
.52609 .9501 L
s
.52609 .9501 m
.5253 .95304 L
s
.5253 .95304 m
.52236 .95383 L
s
.52236 .95383 m
.52451 .95598 L
s
.52451 .95598 m
.52373 .95892 L
s
.52373 .95892 m
.52667 .95813 L
s
.52667 .95813 m
.52882 .96028 L
s
.52882 .96028 m
.52803 .96322 L
s
.52803 .96322 m
.52509 .96401 L
s
.52509 .96401 m
.52724 .96616 L
s
.52724 .96616 m
.52646 .9691 L
s
.52646 .9691 m
.52352 .96989 L
s
.52352 .96989 m
.52136 .96774 L
s
.52136 .96774 m
.52058 .97068 L
s
.52058 .97068 m
.51764 .97146 L
s
.51764 .97146 m
.51548 .96931 L
s
.51548 .96931 m
.51627 .96637 L
s
.51627 .96637 m
.51333 .96716 L
s
.51333 .96716 m
.51118 .96501 L
s
.51118 .96501 m
.51039 .96795 L
s
.51039 .96795 m
.50745 .96874 L
s
.50745 .96874 m
.50961 .97089 L
s
.50961 .97089 m
.50882 .97383 L
s
.50882 .97383 m
.50588 .97462 L
s
.50588 .97462 m
.50373 .97246 L
s
.50373 .97246 m
.50294 .9754 L
s
.50294 .9754 m
.5 .97619 L
s
.5 .97619 m
.49706 .9754 L
s
.49706 .9754 m
.49627 .97246 L
s
.49627 .97246 m
.49412 .97462 L
s
.49412 .97462 m
.49118 .97383 L
s
.49118 .97383 m
.49039 .97089 L
s
.49039 .97089 m
.49255 .96874 L
s
.49255 .96874 m
.48961 .96795 L
s
.48961 .96795 m
.48882 .96501 L
s
.48882 .96501 m
.48667 .96716 L
s
.48667 .96716 m
.48373 .96637 L
s
.48373 .96637 m
.48452 .96931 L
s
.48452 .96931 m
.48236 .97146 L
s
.48236 .97146 m
.47942 .97068 L
s
.47942 .97068 m
.47864 .96774 L
s
.47864 .96774 m
.47648 .96989 L
s
.47648 .96989 m
.47354 .9691 L
s
.47354 .9691 m
.47276 .96616 L
s
.47276 .96616 m
.47491 .96401 L
s
.47491 .96401 m
.47197 .96322 L
s
.47197 .96322 m
.47118 .96028 L
s
.47118 .96028 m
.47333 .95813 L
s
.47333 .95813 m
.47627 .95892 L
s
.47627 .95892 m
.47549 .95598 L
s
.47549 .95598 m
.47764 .95383 L
s
.47764 .95383 m
.4747 .95304 L
s
.4747 .95304 m
.47391 .9501 L
s
.47391 .9501 m
.47176 .95225 L
s
.47176 .95225 m
.46882 .95147 L
s
.46882 .95147 m
.46803 .94853 L
s
.46803 .94853 m
.47018 .94637 L
s
.47018 .94637 m
.46724 .94559 L
s
.46724 .94559 m
.46646 .94265 L
s
.46646 .94265 m
.4643 .9448 L
s
.4643 .9448 m
.46137 .94401 L
s
.46137 .94401 m
.46215 .94695 L
s
.46215 .94695 m
.46 .9491 L
s
.46 .9491 m
.45706 .94831 L
s
.45706 .94831 m
.45627 .94538 L
s
.45627 .94538 m
.45412 .94753 L
s
.45412 .94753 m
.45118 .94674 L
s
.45118 .94674 m
.45197 .94968 L
s
.45197 .94968 m
.44982 .95183 L
s
.44982 .95183 m
.45276 .95262 L
s
.45276 .95262 m
.45355 .95556 L
s
.45355 .95556 m
.45139 .95771 L
s
.45139 .95771 m
.44845 .95692 L
s
.44845 .95692 m
.44924 .95986 L
s
.44924 .95986 m
.44709 .96201 L
s
.44709 .96201 m
.44415 .96123 L
s
.44415 .96123 m
.44336 .95829 L
s
.44336 .95829 m
.44121 .96044 L
s
.44121 .96044 m
.43827 .95965 L
s
.43827 .95965 m
.43748 .95671 L
s
.43748 .95671 m
.43964 .95456 L
s
.43964 .95456 m
.4367 .95377 L
s
.4367 .95377 m
.43591 .95083 L
s
.43591 .95083 m
.43376 .95298 L
s
.43376 .95298 m
.43082 .9522 L
s
.43082 .9522 m
.43161 .95514 L
s
.43161 .95514 m
.42945 .95729 L
s
.42945 .95729 m
.42651 .9565 L
s
.42651 .9565 m
.42573 .95356 L
s
.42573 .95356 m
.42357 .95571 L
s
.42357 .95571 m
.42063 .95492 L
s
.42063 .95492 m
.41985 .95199 L
s
.41985 .95199 m
.422 .94983 L
s
.422 .94983 m
.41906 .94905 L
s
.41906 .94905 m
.41827 .94611 L
s
.41827 .94611 m
.42042 .94395 L
s
.42042 .94395 m
.42336 .94474 L
s
.42336 .94474 m
.42258 .9418 L
s
.42258 .9418 m
.42473 .93965 L
s
.42473 .93965 m
.42179 .93886 L
s
.42179 .93886 m
.421 .93592 L
s
.421 .93592 m
.41885 .93808 L
s
.41885 .93808 m
.41591 .93729 L
s
.41591 .93729 m
.41512 .93435 L
s
.41512 .93435 m
.41727 .9322 L
s
.41727 .9322 m
.41433 .93141 L
s
.41433 .93141 m
.41355 .92847 L
s
.41355 .92847 m
.4157 .92632 L
s
.4157 .92632 m
.41864 .92711 L
s
.41864 .92711 m
.41785 .92417 L
s
.41785 .92417 m
.42 .92201 L
s
.42 .92201 m
.42294 .9228 L
s
.42294 .9228 m
.42373 .92574 L
s
.42373 .92574 m
.42588 .92359 L
s
.42588 .92359 m
.42882 .92438 L
s
.42882 .92438 m
.42803 .92144 L
s
.42803 .92144 m
.43018 .91929 L
s
.43018 .91929 m
.42724 .9185 L
s
.42724 .9185 m
.42646 .91556 L
s
.42646 .91556 m
.42861 .91341 L
s
.42861 .91341 m
.43155 .91419 L
s
.43155 .91419 m
.43076 .91126 L
s
.43076 .91126 m
.43291 .9091 L
s
.43291 .9091 m
.42997 .90832 L
s
.42997 .90832 m
.42919 .90538 L
s
.42919 .90538 m
.42703 .90753 L
s
.42703 .90753 m
.42409 .90674 L
s
.42409 .90674 m
.42331 .9038 L
s
.42331 .9038 m
.42546 .90165 L
s
.42546 .90165 m
.42252 .90086 L
s
.42252 .90086 m
.42173 .89792 L
s
.42173 .89792 m
.41958 .90007 L
s
.41958 .90007 m
.41664 .89929 L
s
.41664 .89929 m
.41743 .90223 L
s
.41743 .90223 m
.41528 .90438 L
s
.41528 .90438 m
.41234 .90359 L
s
.41234 .90359 m
.41155 .90065 L
s
.41155 .90065 m
.4094 .9028 L
s
.4094 .9028 m
.40646 .90201 L
s
.40646 .90201 m
.40567 .89908 L
s
.40567 .89908 m
.40782 .89692 L
s
.40782 .89692 m
.40488 .89614 L
s
.40488 .89614 m
.40409 .8932 L
s
.40409 .8932 m
.40625 .89104 L
s
.40625 .89104 m
.40919 .89183 L
s
.40919 .89183 m
.4084 .88889 L
s
.4084 .88889 m
.41055 .88674 L
s
.41055 .88674 m
.40761 .88595 L
s
.40761 .88595 m
.40682 .88301 L
s
.40682 .88301 m
.40467 .88517 L
s
.40467 .88517 m
.40173 .88438 L
s
.40173 .88438 m
.40094 .88144 L
s
.40094 .88144 m
.4031 .87929 L
s
.4031 .87929 m
.40016 .8785 L
s
.40016 .8785 m
.39937 .87556 L
s
.39937 .87556 m
.39722 .87771 L
s
.39722 .87771 m
.39428 .87692 L
s
.39428 .87692 m
.39507 .87986 L
s
.39507 .87986 m
.39291 .88202 L
s
.39291 .88202 m
.38997 .88123 L
s
.38997 .88123 m
.38919 .87829 L
s
.38919 .87829 m
.38703 .88044 L
s
.38703 .88044 m
.3841 .87965 L
s
.3841 .87965 m
.38488 .88259 L
s
.38488 .88259 m
.38273 .88474 L
s
.38273 .88474 m
.38567 .88553 L
s
.38567 .88553 m
.38646 .88847 L
s
.38646 .88847 m
.38431 .89062 L
s
.38431 .89062 m
.38137 .88983 L
s
.38137 .88983 m
.38215 .89277 L
s
.38215 .89277 m
.38 .89493 L
s
.38 .89493 m
.37706 .89414 L
s
.37706 .89414 m
.37628 .8912 L
s
.37628 .8912 m
.37412 .89335 L
s
.37412 .89335 m
.37118 .89256 L
s
.37118 .89256 m
.3704 .88962 L
s
.3704 .88962 m
.37255 .88747 L
s
.37255 .88747 m
.36961 .88668 L
s
.36961 .88668 m
.36882 .88374 L
s
.36882 .88374 m
.36667 .8859 L
s
.36667 .8859 m
.36373 .88511 L
s
.36373 .88511 m
.36452 .88805 L
s
.36452 .88805 m
.36237 .8902 L
s
.36237 .8902 m
.35943 .88941 L
s
.35943 .88941 m
.35864 .88647 L
s
.35864 .88647 m
.35649 .88863 L
s
.35649 .88863 m
.35355 .88784 L
s
.35355 .88784 m
.35434 .89078 L
s
.35434 .89078 m
.35218 .89293 L
s
.35218 .89293 m
.35512 .89372 L
s
.35512 .89372 m
.35591 .89666 L
s
.35591 .89666 m
.35376 .89881 L
s
.35376 .89881 m
.35082 .89802 L
s
.35082 .89802 m
.35161 .90096 L
s
.35161 .90096 m
.34946 .90311 L
s
.34946 .90311 m
.35239 .9039 L
s
.35239 .9039 m
.35318 .90684 L
s
.35318 .90684 m
.35533 .90469 L
s
.35533 .90469 m
.35827 .90547 L
s
.35827 .90547 m
.35906 .90841 L
s
.35906 .90841 m
.35691 .91057 L
s
.35691 .91057 m
.35985 .91135 L
s
.35985 .91135 m
.36064 .91429 L
s
.36064 .91429 m
.35848 .91644 L
s
.35848 .91644 m
.35554 .91566 L
s
.35554 .91566 m
.35633 .9186 L
s
.35633 .9186 m
.35418 .92075 L
s
.35418 .92075 m
.35124 .91996 L
s
.35124 .91996 m
.35045 .91702 L
s
.35045 .91702 m
.3483 .91917 L
s
.3483 .91917 m
.34536 .91839 L
s
.34536 .91839 m
.34615 .92132 L
s
.34615 .92132 m
.344 .92348 L
s
.344 .92348 m
.34694 .92426 L
s
.34694 .92426 m
.34773 .9272 L
s
.34773 .9272 m
.34557 .92936 L
s
.34557 .92936 m
.34263 .92857 L
s
.34263 .92857 m
.34342 .93151 L
s
.34342 .93151 m
.34127 .93366 L
s
.34127 .93366 m
.33833 .93287 L
s
.33833 .93287 m
.33754 .92993 L
s
.33754 .92993 m
.33539 .93208 L
s
.33539 .93208 m
.33245 .9313 L
s
.33245 .9313 m
.33166 .92836 L
s
.33166 .92836 m
.33382 .9262 L
s
.33382 .9262 m
.33088 .92542 L
s
.33088 .92542 m
.33009 .92248 L
s
.33009 .92248 m
.32794 .92463 L
s
.32794 .92463 m
.325 .92384 L
s
.325 .92384 m
.32578 .92678 L
s
.32578 .92678 m
.32363 .92893 L
s
.32363 .92893 m
.32069 .92815 L
s
.32069 .92815 m
.31991 .92521 L
s
.31991 .92521 m
.31775 .92736 L
s
.31775 .92736 m
.31481 .92657 L
s
.31481 .92657 m
.31403 .92363 L
s
.31403 .92363 m
.31618 .92148 L
s
.31618 .92148 m
.31324 .92069 L
s
.31324 .92069 m
.31245 .91775 L
s
.31245 .91775 m
.3146 .9156 L
s
.3146 .9156 m
.31754 .91639 L
s
.31754 .91639 m
.31676 .91345 L
s
.31676 .91345 m
.31891 .9113 L
s
.31891 .9113 m
.31597 .91051 L
s
.31597 .91051 m
.31518 .90757 L
s
.31518 .90757 m
.31303 .90972 L
s
.31303 .90972 m
.31009 .90893 L
s
.31009 .90893 m
.3093 .90599 L
s
.3093 .90599 m
.31145 .90384 L
s
.31145 .90384 m
.30851 .90305 L
s
.30851 .90305 m
.30773 .90012 L
s
.30773 .90012 m
.30557 .90227 L
s
.30557 .90227 m
.30263 .90148 L
s
.30263 .90148 m
.30342 .90442 L
s
.30342 .90442 m
.30127 .90657 L
s
.30127 .90657 m
.29833 .90578 L
s
.29833 .90578 m
.29754 .90284 L
s
.29754 .90284 m
.29539 .905 L
s
.29539 .905 m
.29245 .90421 L
s
.29245 .90421 m
.29324 .90715 L
s
.29324 .90715 m
.29109 .9093 L
s
.29109 .9093 m
.29403 .91009 L
s
.29403 .91009 m
.29482 .91303 L
s
.29482 .91303 m
.29266 .91518 L
s
.29266 .91518 m
.28972 .91439 L
s
.28972 .91439 m
.29051 .91733 L
s
.29051 .91733 m
.28836 .91948 L
s
.28836 .91948 m
.28542 .91869 L
s
.28542 .91869 m
.28463 .91575 L
s
.28463 .91575 m
.28248 .91791 L
s
.28248 .91791 m
.27954 .91712 L
s
.27954 .91712 m
.27875 .91418 L
s
.27875 .91418 m
.28091 .91203 L
s
.28091 .91203 m
.27797 .91124 L
s
.27797 .91124 m
.27718 .9083 L
s
.27718 .9083 m
.27503 .91045 L
s
.27503 .91045 m
.27209 .90966 L
s
.27209 .90966 m
.27287 .9126 L
s
.27287 .9126 m
.27072 .91476 L
s
.27072 .91476 m
.26778 .91397 L
s
.26778 .91397 m
.267 .91103 L
s
.267 .91103 m
.26484 .91318 L
s
.26484 .91318 m
.2619 .91239 L
s
.2619 .91239 m
.25975 .91024 L
s
.25975 .91024 m
.26054 .9073 L
s
.26054 .9073 m
.2576 .90809 L
s
.2576 .90809 m
.25545 .90594 L
s
.25545 .90594 m
.25624 .903 L
s
.25624 .903 m
.25918 .90221 L
s
.25918 .90221 m
.25702 .90006 L
s
.25702 .90006 m
.25781 .89712 L
s
.25781 .89712 m
.25487 .89791 L
s
.25487 .89791 m
.25272 .89576 L
s
.25272 .89576 m
.25193 .89869 L
s
.25193 .89869 m
.24899 .89948 L
s
.24899 .89948 m
.24684 .89733 L
s
.24684 .89733 m
.24763 .89439 L
s
.24763 .89439 m
.24469 .89518 L
s
.24469 .89518 m
.24254 .89303 L
s
.24254 .89303 m
.24333 .89009 L
s
.24333 .89009 m
.24627 .8893 L
s
.24627 .8893 m
.24411 .88715 L
s
.24411 .88715 m
.2449 .88421 L
s
.2449 .88421 m
.24784 .88342 L
s
.24784 .88342 m
.24999 .88557 L
s
.24999 .88557 m
.25078 .88263 L
s
.25078 .88263 m
.25372 .88185 L
s
.25372 .88185 m
.25157 .87969 L
s
.25157 .87969 m
.25236 .87675 L
s
.25236 .87675 m
.24942 .87754 L
s
.24942 .87754 m
.24726 .87539 L
s
.24726 .87539 m
.24805 .87245 L
s
.24805 .87245 m
.25099 .87166 L
s
.25099 .87166 m
.24884 .86951 L
s
.24884 .86951 m
.24963 .86657 L
s
.24963 .86657 m
.24669 .86736 L
s
.24669 .86736 m
.24454 .86521 L
s
.24454 .86521 m
.24375 .86815 L
s
.24375 .86815 m
.24081 .86893 L
s
.24081 .86893 m
.23866 .86678 L
s
.23866 .86678 m
.23944 .86384 L
s
.23944 .86384 m
.2365 .86463 L
s
.2365 .86463 m
.23435 .86248 L
s
.23435 .86248 m
.23357 .86542 L
s
.23357 .86542 m
.23063 .86621 L
s
.23063 .86621 m
.23278 .86836 L
s
.23278 .86836 m
.23199 .8713 L
s
.23199 .8713 m
.22905 .87208 L
s
.22905 .87208 m
.2269 .86993 L
s
.2269 .86993 m
.22611 .87287 L
s
.22611 .87287 m
.22317 .87366 L
s
.22317 .87366 m
.22102 .87151 L
s
.22102 .87151 m
.22181 .86857 L
s
.22181 .86857 m
.21887 .86936 L
s
.21887 .86936 m
.21672 .8672 L
s
.21672 .8672 m
.2175 .86427 L
s
.2175 .86427 m
.22044 .86348 L
s
.22044 .86348 m
.21829 .86133 L
s
.21829 .86133 m
.21908 .85839 L
s
.21908 .85839 m
.21614 .85917 L
s
.21614 .85917 m
.21399 .85702 L
s
.21399 .85702 m
.2132 .85996 L
s
.2132 .85996 m
.21026 .86075 L
s
.21026 .86075 m
.20811 .8586 L
s
.20811 .8586 m
.2089 .85566 L
s
.2089 .85566 m
.20596 .85645 L
s
.20596 .85645 m
.20381 .85429 L
s
.20381 .85429 m
.20459 .85135 L
s
.20459 .85135 m
.20753 .85057 L
s
.20753 .85057 m
.20538 .84841 L
s
.20538 .84841 m
.20617 .84548 L
s
.20617 .84548 m
.20911 .84469 L
s
.20911 .84469 m
.21126 .84684 L
s
.21126 .84684 m
.21205 .8439 L
s
.21205 .8439 m
.21499 .84311 L
s
.21499 .84311 m
.21283 .84096 L
s
.21283 .84096 m
.21362 .83802 L
s
.21362 .83802 m
.21068 .83881 L
s
.21068 .83881 m
.20853 .83666 L
s
.20853 .83666 m
.20932 .83372 L
s
.20932 .83372 m
.21226 .83293 L
s
.21226 .83293 m
.21011 .83078 L
s
.21011 .83078 m
.21089 .82784 L
s
.21089 .82784 m
.21383 .82705 L
s
.21383 .82705 m
.21599 .8292 L
s
.21599 .8292 m
.21677 .82626 L
s
.21677 .82626 m
.21971 .82548 L
s
.21971 .82548 m
.22186 .82763 L
s
.22186 .82763 m
.22108 .83057 L
s
.22108 .83057 m
.22402 .82978 L
s
.22402 .82978 m
.22617 .83193 L
s
.22617 .83193 m
.22696 .82899 L
s
.22696 .82899 m
.22989 .8282 L
s
.22989 .8282 m
.22774 .82605 L
s
.22774 .82605 m
.22853 .82311 L
s
.22853 .82311 m
.23147 .82233 L
s
.23147 .82233 m
.23362 .82448 L
s
.23362 .82448 m
.23441 .82154 L
s
.23441 .82154 m
.23735 .82075 L
s
.23735 .82075 m
.2352 .8186 L
s
.2352 .8186 m
.23598 .81566 L
s
.23598 .81566 m
.23305 .81645 L
s
.23305 .81645 m
.23089 .81429 L
s
.23089 .81429 m
.23168 .81136 L
s
.23168 .81136 m
.23462 .81057 L
s
.23462 .81057 m
.23247 .80842 L
s
.23247 .80842 m
.23326 .80548 L
s
.23326 .80548 m
.23032 .80626 L
s
.23032 .80626 m
.22817 .80411 L
s
.22817 .80411 m
.22738 .80705 L
s
.22738 .80705 m
.22444 .80784 L
s
.22444 .80784 m
.22229 .80569 L
s
.22229 .80569 m
.22307 .80275 L
s
.22307 .80275 m
.22013 .80354 L
s
.22013 .80354 m
.21798 .80138 L
s
.21798 .80138 m
.21877 .79844 L
s
.21877 .79844 m
.22171 .79766 L
s
.22171 .79766 m
.21956 .7955 L
s
.21956 .7955 m
.22035 .79257 L
s
.22035 .79257 m
.22329 .79178 L
s
.22329 .79178 m
.22544 .79393 L
s
.22544 .79393 m
.22622 .79099 L
s
.22622 .79099 m
.22916 .7902 L
s
.22916 .7902 m
.22701 .78805 L
s
.22701 .78805 m
.2278 .78511 L
s
.2278 .78511 m
.22486 .7859 L
s
.22486 .7859 m
.22271 .78375 L
s
.22271 .78375 m
.2235 .78081 L
s
.2235 .78081 m
.22644 .78002 L
s
.22644 .78002 m
.22428 .77787 L
s
.22428 .77787 m
.22507 .77493 L
s
.22507 .77493 m
.22213 .77572 L
s
.22213 .77572 m
.21998 .77356 L
s
.21998 .77356 m
.21919 .7765 L
s
.21919 .7765 m
.21625 .77729 L
s
.21625 .77729 m
.2141 .77514 L
s
.2141 .77514 m
.21489 .7722 L
s
.21489 .7722 m
.21195 .77299 L
s
.21195 .77299 m
.2098 .77084 L
s
.2098 .77084 m
.20901 .77378 L
s
.20901 .77378 m
.20607 .77456 L
s
.20607 .77456 m
.20822 .77671 L
s
.20822 .77671 m
.20743 .77965 L
s
.20743 .77965 m
.2045 .78044 L
s
.2045 .78044 m
.20234 .77829 L
s
.20234 .77829 m
.20156 .78123 L
s
.20156 .78123 m
.19862 .78202 L
s
.19862 .78202 m
.19646 .77987 L
s
.19646 .77987 m
.19725 .77693 L
s
.19725 .77693 m
.19431 .77771 L
s
.19431 .77771 m
.19216 .77556 L
s
.19216 .77556 m
.19295 .77262 L
s
.19295 .77262 m
.19589 .77183 L
s
.19589 .77183 m
.19374 .76968 L
s
.19374 .76968 m
.19452 .76674 L
s
.19452 .76674 m
.19158 .76753 L
s
.19158 .76753 m
.18943 .76538 L
s
.18943 .76538 m
.18864 .76832 L
s
.18864 .76832 m
.18571 .76911 L
s
.18571 .76911 m
.18355 .76695 L
s
.18355 .76695 m
.18434 .76402 L
s
.18434 .76402 m
.1814 .7648 L
s
.1814 .7648 m
.17925 .76265 L
s
.17925 .76265 m
.17846 .76559 L
s
.17846 .76559 m
.17552 .76638 L
s
.17552 .76638 m
.17767 .76853 L
s
.17767 .76853 m
.17689 .77147 L
s
.17689 .77147 m
.17395 .77226 L
s
.17395 .77226 m
.1718 .77011 L
s
.1718 .77011 m
.17101 .77304 L
s
.17101 .77304 m
.16807 .77383 L
s
.16807 .77383 m
.17022 .77598 L
s
.17022 .77598 m
.16943 .77892 L
s
.16943 .77892 m
.17237 .77814 L
s
.17237 .77814 m
.17452 .78029 L
s
.17452 .78029 m
.17374 .78323 L
s
.17374 .78323 m
.1708 .78401 L
s
.1708 .78401 m
.17295 .78617 L
s
.17295 .78617 m
.17216 .78911 L
s
.17216 .78911 m
.16922 .78989 L
s
.16922 .78989 m
.16707 .78774 L
s
.16707 .78774 m
.16628 .79068 L
s
.16628 .79068 m
.16334 .79147 L
s
.16334 .79147 m
.16119 .78932 L
s
.16119 .78932 m
.16198 .78638 L
s
.16198 .78638 m
.15904 .78717 L
s
.15904 .78717 m
.15689 .78501 L
s
.15689 .78501 m
.1561 .78795 L
s
.1561 .78795 m
.15316 .78874 L
s
.15316 .78874 m
.15531 .79089 L
s
.15531 .79089 m
.15452 .79383 L
s
.15452 .79383 m
.15159 .79462 L
s
.15159 .79462 m
.14943 .79247 L
s
.14943 .79247 m
.14865 .79541 L
s
.14865 .79541 m
.14571 .79619 L
s
.14571 .79619 m
.14355 .79404 L
s
.14355 .79404 m
.14434 .7911 L
s
.14434 .7911 m
.1414 .79189 L
s
.1414 .79189 m
.13925 .78974 L
s
.13925 .78974 m
.14004 .7868 L
s
.14004 .7868 m
.14298 .78601 L
s
.14298 .78601 m
.14083 .78386 L
s
.14083 .78386 m
.14161 .78092 L
s
.14161 .78092 m
.13867 .78171 L
s
.13867 .78171 m
.13652 .77956 L
s
.13652 .77956 m
.13573 .7825 L
s
.13573 .7825 m
.1328 .78328 L
s
.1328 .78328 m
.13064 .78113 L
s
.13064 .78113 m
.13143 .77819 L
s
.13143 .77819 m
.12849 .77898 L
s
.12849 .77898 m
.12634 .77683 L
s
.12634 .77683 m
.12713 .77389 L
s
.12713 .77389 m
.13007 .7731 L
s
.13007 .7731 m
.12792 .77095 L
s
.12792 .77095 m
.1287 .76801 L
s
.1287 .76801 m
.13164 .76722 L
s
.13164 .76722 m
.13379 .76937 L
s
.13379 .76937 m
.13458 .76643 L
s
.13458 .76643 m
.13752 .76565 L
s
.13752 .76565 m
.13537 .7635 L
s
.13537 .7635 m
.13616 .76056 L
s
.13616 .76056 m
.13322 .76134 L
s
.13322 .76134 m
.13107 .75919 L
s
.13107 .75919 m
.13185 .75625 L
s
.13185 .75625 m
.13479 .75546 L
s
.13479 .75546 m
.13264 .75331 L
s
.13264 .75331 m
.13343 .75037 L
s
.13343 .75037 m
.13049 .75116 L
s
.13049 .75116 m
.12834 .74901 L
s
.12834 .74901 m
.12755 .75195 L
s
.12755 .75195 m
.12461 .75274 L
s
.12461 .75274 m
.12246 .75058 L
s
.12246 .75058 m
.12325 .74764 L
s
.12325 .74764 m
.12031 .74843 L
s
.12031 .74843 m
.11815 .74628 L
s
.11815 .74628 m
.11737 .74922 L
s
.11737 .74922 m
.11443 .75001 L
s
.11443 .75001 m
.11658 .75216 L
s
.11658 .75216 m
.11579 .7551 L
s
.11579 .7551 m
.11285 .75589 L
s
.11285 .75589 m
.1107 .75373 L
s
.1107 .75373 m
.10991 .75667 L
s
.10991 .75667 m
.10697 .75746 L
s
.10697 .75746 m
.10482 .75531 L
s
.10482 .75531 m
.10561 .75237 L
s
.10561 .75237 m
.10267 .75316 L
s
.10267 .75316 m
.10052 .75101 L
s
.10052 .75101 m
.10131 .74807 L
s
.10131 .74807 m
.10424 .74728 L
s
.10424 .74728 m
.10209 .74513 L
s
.10209 .74513 m
.10288 .74219 L
s
.10288 .74219 m
.09994 .74298 L
s
.09994 .74298 m
.09779 .74082 L
s
.09779 .74082 m
.097 .74376 L
s
.097 .74376 m
.09406 .74455 L
s
.09406 .74455 m
.09191 .7424 L
s
.09191 .7424 m
.0927 .73946 L
s
.0927 .73946 m
.08976 .74025 L
s
.08976 .74025 m
.08761 .7381 L
s
.08761 .7381 m
.08682 .73516 L
s
.08682 .73516 m
.08897 .733 L
s
.08897 .733 m
.08603 .73222 L
s
.08603 .73222 m
.08524 .72928 L
s
.08524 .72928 m
.0874 .72713 L
s
.0874 .72713 m
.09034 .72791 L
s
.09034 .72791 m
.08955 .72497 L
s
.08955 .72497 m
.0917 .72282 L
s
.0917 .72282 m
.08876 .72203 L
s
.08876 .72203 m
.08797 .71909 L
s
.08797 .71909 m
.08582 .72125 L
s
.08582 .72125 m
.08288 .72046 L
s
.08288 .72046 m
.08209 .71752 L
s
.08209 .71752 m
.08425 .71537 L
s
.08425 .71537 m
.08131 .71458 L
s
.08131 .71458 m
.08052 .71164 L
s
.08052 .71164 m
.08267 .70949 L
s
.08267 .70949 m
.08561 .71028 L
s
.08561 .71028 m
.08482 .70734 L
s
.08482 .70734 m
.08697 .70518 L
s
.08697 .70518 m
.08991 .70597 L
s
.08991 .70597 m
.0907 .70891 L
s
.0907 .70891 m
.09285 .70676 L
s
.09285 .70676 m
.09579 .70755 L
s
.09579 .70755 m
.095 .70461 L
s
.095 .70461 m
.09716 .70246 L
s
.09716 .70246 m
.09422 .70167 L
s
.09422 .70167 m
.09343 .69873 L
s
.09343 .69873 m
.09558 .69658 L
s
.09558 .69658 m
.09852 .69737 L
s
.09852 .69737 m
.09773 .69443 L
s
.09773 .69443 m
.09988 .69227 L
s
.09988 .69227 m
.09695 .69149 L
s
.09695 .69149 m
.09616 .68855 L
s
.09616 .68855 m
.09401 .6907 L
s
.09401 .6907 m
.09107 .68991 L
s
.09107 .68991 m
.09028 .68697 L
s
.09028 .68697 m
.09243 .68482 L
s
.09243 .68482 m
.08949 .68403 L
s
.08949 .68403 m
.0887 .68109 L
s
.0887 .68109 m
.08655 .68324 L
s
.08655 .68324 m
.08361 .68246 L
s
.08361 .68246 m
.0844 .6854 L
s
.0844 .6854 m
.08225 .68755 L
s
.08225 .68755 m
.07931 .68676 L
s
.07931 .68676 m
.07852 .68382 L
s
.07852 .68382 m
.07637 .68597 L
s
.07637 .68597 m
.07343 .68519 L
s
.07343 .68519 m
.07264 .68225 L
s
.07264 .68225 m
.07479 .68009 L
s
.07479 .68009 m
.07185 .67931 L
s
.07185 .67931 m
.07107 .67637 L
s
.07107 .67637 m
.07322 .67422 L
s
.07322 .67422 m
.07616 .675 L
s
.07616 .675 m
.07537 .67206 L
s
.07537 .67206 m
.07752 .66991 L
s
.07752 .66991 m
.07458 .66912 L
s
.07458 .66912 m
.0738 .66618 L
s
.0738 .66618 m
.07164 .66834 L
s
.07164 .66834 m
.0687 .66755 L
s
.0687 .66755 m
.06792 .66461 L
s
.06792 .66461 m
.07007 .66246 L
s
.07007 .66246 m
.06713 .66167 L
s
.06713 .66167 m
.06634 .65873 L
s
.06634 .65873 m
.06849 .65658 L
s
.06849 .65658 m
.07143 .65737 L
s
.07143 .65737 m
.07064 .65443 L
s
.07064 .65443 m
.0728 .65227 L
s
.0728 .65227 m
.07574 .65306 L
s
.07574 .65306 m
.07652 .656 L
s
.07652 .656 m
.07868 .65385 L
s
.07868 .65385 m
.08161 .65464 L
s
.08161 .65464 m
.08083 .6517 L
s
.08083 .6517 m
.08298 .64955 L
s
.08298 .64955 m
.08004 .64876 L
s
.08004 .64876 m
.07925 .64582 L
s
.07925 .64582 m
.0814 .64367 L
s
.0814 .64367 m
.08434 .64446 L
s
.08434 .64446 m
.08356 .64152 L
s
.08356 .64152 m
.08571 .63936 L
s
.08571 .63936 m
.08865 .64015 L
s
.08865 .64015 m
.08943 .64309 L
s
.08943 .64309 m
.09159 .64094 L
s
.09159 .64094 m
.09453 .64173 L
s
.09453 .64173 m
.09531 .64467 L
s
.09531 .64467 m
.09316 .64682 L
s
.09316 .64682 m
.0961 .64761 L
s
.0961 .64761 m
.09689 .65054 L
s
.09689 .65054 m
.09904 .64839 L
s
.09904 .64839 m
.10198 .64918 L
s
.10198 .64918 m
.10119 .64624 L
s
.10119 .64624 m
.10334 .64409 L
s
.10334 .64409 m
.10628 .64488 L
s
.10628 .64488 m
.10707 .64782 L
s
.10707 .64782 m
.10922 .64566 L
s
.10922 .64566 m
.11216 .64645 L
s
.11216 .64645 m
.11137 .64351 L
s
.11137 .64351 m
.11353 .64136 L
s
.11353 .64136 m
.11059 .64057 L
s
.11059 .64057 m
.1098 .63763 L
s
.1098 .63763 m
.11195 .63548 L
s
.11195 .63548 m
.11489 .63627 L
s
.11489 .63627 m
.1141 .63333 L
s
.1141 .63333 m
.11626 .63118 L
s
.11626 .63118 m
.11332 .63039 L
s
.11332 .63039 m
.11253 .62745 L
s
.11253 .62745 m
.11038 .6296 L
s
.11038 .6296 m
.10744 .62882 L
s
.10744 .62882 m
.10665 .62588 L
s
.10665 .62588 m
.1088 .62372 L
s
.1088 .62372 m
.10586 .62294 L
s
.10586 .62294 m
.10507 .62 L
s
.10507 .62 m
.10723 .61785 L
s
.10723 .61785 m
.11017 .61863 L
s
.11017 .61863 m
.10938 .61569 L
s
.10938 .61569 m
.11153 .61354 L
s
.11153 .61354 m
.11447 .61433 L
s
.11447 .61433 m
.11526 .61727 L
s
.11526 .61727 m
.11741 .61512 L
s
.11741 .61512 m
.12035 .6159 L
s
.12035 .6159 m
.11956 .61297 L
s
.11956 .61297 m
.12171 .61081 L
s
.12171 .61081 m
.11877 .61003 L
s
.11877 .61003 m
.11798 .60709 L
s
.11798 .60709 m
.12014 .60493 L
s
.12014 .60493 m
.12308 .60572 L
s
.12308 .60572 m
.12229 .60278 L
s
.12229 .60278 m
.12444 .60063 L
s
.12444 .60063 m
.1215 .59984 L
s
.1215 .59984 m
.12071 .5969 L
s
.12071 .5969 m
.11856 .59906 L
s
.11856 .59906 m
.11562 .59827 L
s
.11562 .59827 m
.11483 .59533 L
s
.11483 .59533 m
.11699 .59318 L
s
.11699 .59318 m
.11405 .59239 L
s
.11405 .59239 m
.11326 .58945 L
s
.11326 .58945 m
.11111 .5916 L
s
.11111 .5916 m
.10817 .59081 L
s
.10817 .59081 m
.10896 .59375 L
s
.10896 .59375 m
.1068 .59591 L
s
.1068 .59591 m
.10386 .59512 L
s
.10386 .59512 m
.10308 .59218 L
s
.10308 .59218 m
.10092 .59433 L
s
.10092 .59433 m
.09799 .59354 L
s
.09799 .59354 m
.0972 .5906 L
s
.0972 .5906 m
.09935 .58845 L
s
.09935 .58845 m
.09641 .58766 L
s
.09641 .58766 m
.09562 .58472 L
s
.09562 .58472 m
.09777 .58257 L
s
.09777 .58257 m
.10071 .58336 L
s
.10071 .58336 m
.09993 .58042 L
s
.09993 .58042 m
.10208 .57827 L
s
.10208 .57827 m
.09914 .57748 L
s
.09914 .57748 m
.09835 .57454 L
s
.09835 .57454 m
.0962 .57669 L
s
.0962 .57669 m
.09326 .57591 L
s
.09326 .57591 m
.09247 .57297 L
s
.09247 .57297 m
.09462 .57081 L
s
.09462 .57081 m
.09168 .57003 L
s
.09168 .57003 m
.0909 .56709 L
s
.0909 .56709 m
.08874 .56924 L
s
.08874 .56924 m
.08581 .56845 L
s
.08581 .56845 m
.08659 .57139 L
s
.08659 .57139 m
.08444 .57354 L
s
.08444 .57354 m
.0815 .57276 L
s
.0815 .57276 m
.08071 .56982 L
s
.08071 .56982 m
.07856 .57197 L
s
.07856 .57197 m
.07562 .57118 L
s
.07562 .57118 m
.07641 .57412 L
s
.07641 .57412 m
.07426 .57627 L
s
.07426 .57627 m
.0772 .57706 L
s
.0772 .57706 m
.07799 .58 L
s
.07799 .58 m
.07583 .58215 L
s
.07583 .58215 m
.07289 .58136 L
s
.07289 .58136 m
.07368 .5843 L
s
.07368 .5843 m
.07153 .58645 L
s
.07153 .58645 m
.06859 .58567 L
s
.06859 .58567 m
.0678 .58273 L
s
.0678 .58273 m
.06565 .58488 L
s
.06565 .58488 m
.06271 .58409 L
s
.06271 .58409 m
.06192 .58115 L
s
.06192 .58115 m
.06408 .579 L
s
.06408 .579 m
.06114 .57821 L
s
.06114 .57821 m
.06035 .57527 L
s
.06035 .57527 m
.0582 .57742 L
s
.0582 .57742 m
.05526 .57664 L
s
.05526 .57664 m
.05605 .57958 L
s
.05605 .57958 m
.05389 .58173 L
s
.05389 .58173 m
.05095 .58094 L
s
.05095 .58094 m
.05017 .578 L
s
.05017 .578 m
.04801 .58015 L
s
.04801 .58015 m
.04508 .57937 L
s
.04508 .57937 m
.04429 .57643 L
s
.04429 .57643 m
.04644 .57427 L
s
.04644 .57427 m
.0435 .57349 L
s
.0435 .57349 m
.04271 .57055 L
s
.04271 .57055 m
.04486 .56839 L
s
.04486 .56839 m
.0478 .56918 L
s
.0478 .56918 m
.04702 .56624 L
s
.04702 .56624 m
.04917 .56409 L
s
.04917 .56409 m
.04623 .5633 L
s
.04623 .5633 m
.04544 .56036 L
s
.04544 .56036 m
.04329 .56252 L
s
.04329 .56252 m
.04035 .56173 L
s
.04035 .56173 m
.03956 .55879 L
s
.03956 .55879 m
.04171 .55664 L
s
.04171 .55664 m
.03877 .55585 L
s
.03877 .55585 m
.03799 .55291 L
s
.03799 .55291 m
.04014 .55076 L
s
.04014 .55076 m
.04308 .55155 L
s
.04308 .55155 m
.04229 .54861 L
s
.04229 .54861 m
.04444 .54645 L
s
.04444 .54645 m
.04738 .54724 L
s
.04738 .54724 m
.04817 .55018 L
s
.04817 .55018 m
.05032 .54803 L
s
.05032 .54803 m
.05326 .54882 L
s
.05326 .54882 m
.05247 .54588 L
s
.05247 .54588 m
.05462 .54373 L
s
.05462 .54373 m
.05169 .54294 L
s
.05169 .54294 m
.0509 .54 L
s
.0509 .54 m
.05305 .53785 L
s
.05305 .53785 m
.05599 .53863 L
s
.05599 .53863 m
.0552 .5357 L
s
.0552 .5357 m
.05735 .53354 L
s
.05735 .53354 m
.05441 .53276 L
s
.05441 .53276 m
.05363 .52982 L
s
.05363 .52982 m
.05147 .53197 L
s
.05147 .53197 m
.04853 .53118 L
s
.04853 .53118 m
.04775 .52824 L
s
.04775 .52824 m
.0499 .52609 L
s
.0499 .52609 m
.04696 .5253 L
s
.04696 .5253 m
.04617 .52236 L
s
.04617 .52236 m
.04402 .52451 L
s
.04402 .52451 m
.04108 .52373 L
s
.04108 .52373 m
.04187 .52667 L
s
.04187 .52667 m
.03972 .52882 L
s
.03972 .52882 m
.03678 .52803 L
s
.03678 .52803 m
.03599 .52509 L
s
.03599 .52509 m
.03384 .52724 L
s
.03384 .52724 m
.0309 .52646 L
s
.0309 .52646 m
.03011 .52352 L
s
.03011 .52352 m
.03226 .52136 L
s
.03226 .52136 m
.02932 .52058 L
s
.02932 .52058 m
.02854 .51764 L
s
.02854 .51764 m
.03069 .51548 L
s
.03069 .51548 m
.03363 .51627 L
s
.03363 .51627 m
.03284 .51333 L
s
.03284 .51333 m
.03499 .51118 L
s
.03499 .51118 m
.03205 .51039 L
s
.03205 .51039 m
.03126 .50745 L
s
.03126 .50745 m
.02911 .50961 L
s
.02911 .50961 m
.02617 .50882 L
s
.02617 .50882 m
.02538 .50588 L
s
.02538 .50588 m
.02754 .50373 L
s
.02754 .50373 m
.0246 .50294 L
s
.0246 .50294 m
.02381 .5 L
s
.02381 .5 m
.0246 .49706 L
s
.0246 .49706 m
.02754 .49627 L
s
.02754 .49627 m
.02538 .49412 L
s
.02538 .49412 m
.02617 .49118 L
s
.02617 .49118 m
.02911 .49039 L
s
.02911 .49039 m
.03126 .49255 L
s
.03126 .49255 m
.03205 .48961 L
s
.03205 .48961 m
.03499 .48882 L
s
.03499 .48882 m
.03284 .48667 L
s
.03284 .48667 m
.03363 .48373 L
s
.03363 .48373 m
.03069 .48452 L
s
.03069 .48452 m
.02854 .48236 L
s
.02854 .48236 m
.02932 .47942 L
s
.02932 .47942 m
.03226 .47864 L
s
.03226 .47864 m
.03011 .47648 L
s
.03011 .47648 m
.0309 .47354 L
s
.0309 .47354 m
.03384 .47276 L
s
.03384 .47276 m
.03599 .47491 L
s
.03599 .47491 m
.03678 .47197 L
s
.03678 .47197 m
.03972 .47118 L
s
.03972 .47118 m
.04187 .47333 L
s
.04187 .47333 m
.04108 .47627 L
s
.04108 .47627 m
.04402 .47549 L
s
.04402 .47549 m
.04617 .47764 L
s
.04617 .47764 m
.04696 .4747 L
s
.04696 .4747 m
.0499 .47391 L
s
.0499 .47391 m
.04775 .47176 L
s
.04775 .47176 m
.04853 .46882 L
s
.04853 .46882 m
.05147 .46803 L
s
.05147 .46803 m
.05363 .47018 L
s
.05363 .47018 m
.05441 .46724 L
s
.05441 .46724 m
.05735 .46646 L
s
.05735 .46646 m
.0552 .4643 L
s
.0552 .4643 m
.05599 .46137 L
s
.05599 .46137 m
.05305 .46215 L
s
.05305 .46215 m
.0509 .46 L
s
.0509 .46 m
.05169 .45706 L
s
.05169 .45706 m
.05462 .45627 L
s
.05462 .45627 m
.05247 .45412 L
s
.05247 .45412 m
.05326 .45118 L
s
.05326 .45118 m
.05032 .45197 L
s
.05032 .45197 m
.04817 .44982 L
s
.04817 .44982 m
.04738 .45276 L
s
.04738 .45276 m
.04444 .45355 L
s
.04444 .45355 m
.04229 .45139 L
s
.04229 .45139 m
.04308 .44845 L
s
.04308 .44845 m
.04014 .44924 L
s
.04014 .44924 m
.03799 .44709 L
s
.03799 .44709 m
.03877 .44415 L
s
.03877 .44415 m
.04171 .44336 L
s
.04171 .44336 m
.03956 .44121 L
s
.03956 .44121 m
.04035 .43827 L
s
.04035 .43827 m
.04329 .43748 L
s
.04329 .43748 m
.04544 .43964 L
s
.04544 .43964 m
.04623 .4367 L
s
.04623 .4367 m
.04917 .43591 L
s
.04917 .43591 m
.04702 .43376 L
s
.04702 .43376 m
.0478 .43082 L
s
.0478 .43082 m
.04486 .43161 L
s
.04486 .43161 m
.04271 .42945 L
s
.04271 .42945 m
.0435 .42651 L
s
.0435 .42651 m
.04644 .42573 L
s
.04644 .42573 m
.04429 .42357 L
s
.04429 .42357 m
.04508 .42063 L
s
.04508 .42063 m
.04801 .41985 L
s
.04801 .41985 m
.05017 .422 L
s
.05017 .422 m
.05095 .41906 L
s
.05095 .41906 m
.05389 .41827 L
s
.05389 .41827 m
.05605 .42042 L
s
.05605 .42042 m
.05526 .42336 L
s
.05526 .42336 m
.0582 .42258 L
s
.0582 .42258 m
.06035 .42473 L
s
.06035 .42473 m
.06114 .42179 L
s
.06114 .42179 m
.06408 .421 L
s
.06408 .421 m
.06192 .41885 L
s
.06192 .41885 m
.06271 .41591 L
s
.06271 .41591 m
.06565 .41512 L
s
.06565 .41512 m
.0678 .41727 L
s
.0678 .41727 m
.06859 .41433 L
s
.06859 .41433 m
.07153 .41355 L
s
.07153 .41355 m
.07368 .4157 L
s
.07368 .4157 m
.07289 .41864 L
s
.07289 .41864 m
.07583 .41785 L
s
.07583 .41785 m
.07799 .42 L
s
.07799 .42 m
.0772 .42294 L
s
.0772 .42294 m
.07426 .42373 L
s
.07426 .42373 m
.07641 .42588 L
s
.07641 .42588 m
.07562 .42882 L
s
.07562 .42882 m
.07856 .42803 L
s
.07856 .42803 m
.08071 .43018 L
s
.08071 .43018 m
.0815 .42724 L
s
.0815 .42724 m
.08444 .42646 L
s
.08444 .42646 m
.08659 .42861 L
s
.08659 .42861 m
.08581 .43155 L
s
.08581 .43155 m
.08874 .43076 L
s
.08874 .43076 m
.0909 .43291 L
s
.0909 .43291 m
.09168 .42997 L
s
.09168 .42997 m
.09462 .42919 L
s
.09462 .42919 m
.09247 .42703 L
s
.09247 .42703 m
.09326 .42409 L
s
.09326 .42409 m
.0962 .42331 L
s
.0962 .42331 m
.09835 .42546 L
s
.09835 .42546 m
.09914 .42252 L
s
.09914 .42252 m
.10208 .42173 L
s
.10208 .42173 m
.09993 .41958 L
s
.09993 .41958 m
.10071 .41664 L
s
.10071 .41664 m
.09777 .41743 L
s
.09777 .41743 m
.09562 .41528 L
s
.09562 .41528 m
.09641 .41234 L
s
.09641 .41234 m
.09935 .41155 L
s
.09935 .41155 m
.0972 .4094 L
s
.0972 .4094 m
.09799 .40646 L
s
.09799 .40646 m
.10092 .40567 L
s
.10092 .40567 m
.10308 .40782 L
s
.10308 .40782 m
.10386 .40488 L
s
.10386 .40488 m
.1068 .40409 L
s
.1068 .40409 m
.10896 .40625 L
s
.10896 .40625 m
.10817 .40919 L
s
.10817 .40919 m
.11111 .4084 L
s
.11111 .4084 m
.11326 .41055 L
s
.11326 .41055 m
.11405 .40761 L
s
.11405 .40761 m
.11699 .40682 L
s
.11699 .40682 m
.11483 .40467 L
s
.11483 .40467 m
.11562 .40173 L
s
.11562 .40173 m
.11856 .40094 L
s
.11856 .40094 m
.12071 .4031 L
s
.12071 .4031 m
.1215 .40016 L
s
.1215 .40016 m
.12444 .39937 L
s
.12444 .39937 m
.12229 .39722 L
s
.12229 .39722 m
.12308 .39428 L
s
.12308 .39428 m
.12014 .39507 L
s
.12014 .39507 m
.11798 .39291 L
s
.11798 .39291 m
.11877 .38997 L
s
.11877 .38997 m
.12171 .38919 L
s
.12171 .38919 m
.11956 .38703 L
s
.11956 .38703 m
.12035 .3841 L
s
.12035 .3841 m
.11741 .38488 L
s
.11741 .38488 m
.11526 .38273 L
s
.11526 .38273 m
.11447 .38567 L
s
.11447 .38567 m
.11153 .38646 L
s
.11153 .38646 m
.10938 .38431 L
s
.10938 .38431 m
.11017 .38137 L
s
.11017 .38137 m
.10723 .38215 L
s
.10723 .38215 m
.10507 .38 L
s
.10507 .38 m
.10586 .37706 L
s
.10586 .37706 m
.1088 .37628 L
s
.1088 .37628 m
.10665 .37412 L
s
.10665 .37412 m
.10744 .37118 L
s
.10744 .37118 m
.11038 .3704 L
s
.11038 .3704 m
.11253 .37255 L
s
.11253 .37255 m
.11332 .36961 L
s
.11332 .36961 m
.11626 .36882 L
s
.11626 .36882 m
.1141 .36667 L
s
.1141 .36667 m
.11489 .36373 L
s
.11489 .36373 m
.11195 .36452 L
s
.11195 .36452 m
.1098 .36237 L
s
.1098 .36237 m
.11059 .35943 L
s
.11059 .35943 m
.11353 .35864 L
s
.11353 .35864 m
.11137 .35649 L
s
.11137 .35649 m
.11216 .35355 L
s
.11216 .35355 m
.10922 .35434 L
s
.10922 .35434 m
.10707 .35218 L
s
.10707 .35218 m
.10628 .35512 L
s
.10628 .35512 m
.10334 .35591 L
s
.10334 .35591 m
.10119 .35376 L
s
.10119 .35376 m
.10198 .35082 L
s
.10198 .35082 m
.09904 .35161 L
s
.09904 .35161 m
.09689 .34946 L
s
.09689 .34946 m
.0961 .35239 L
s
.0961 .35239 m
.09316 .35318 L
s
.09316 .35318 m
.09531 .35533 L
s
.09531 .35533 m
.09453 .35827 L
s
.09453 .35827 m
.09159 .35906 L
s
.09159 .35906 m
.08943 .35691 L
s
.08943 .35691 m
.08865 .35985 L
s
.08865 .35985 m
.08571 .36064 L
s
.08571 .36064 m
.08356 .35848 L
s
.08356 .35848 m
.08434 .35554 L
s
.08434 .35554 m
.0814 .35633 L
s
.0814 .35633 m
.07925 .35418 L
s
.07925 .35418 m
.08004 .35124 L
s
.08004 .35124 m
.08298 .35045 L
s
.08298 .35045 m
.08083 .3483 L
s
.08083 .3483 m
.08161 .34536 L
s
.08161 .34536 m
.07868 .34615 L
s
.07868 .34615 m
.07652 .344 L
s
.07652 .344 m
.07574 .34694 L
s
.07574 .34694 m
.0728 .34773 L
s
.0728 .34773 m
.07064 .34557 L
s
.07064 .34557 m
.07143 .34263 L
s
.07143 .34263 m
.06849 .34342 L
s
.06849 .34342 m
.06634 .34127 L
s
.06634 .34127 m
.06713 .33833 L
s
.06713 .33833 m
.07007 .33754 L
s
.07007 .33754 m
.06792 .33539 L
s
.06792 .33539 m
.0687 .33245 L
s
.0687 .33245 m
.07164 .33166 L
s
.07164 .33166 m
.0738 .33382 L
s
.0738 .33382 m
.07458 .33088 L
s
.07458 .33088 m
.07752 .33009 L
s
.07752 .33009 m
.07537 .32794 L
s
.07537 .32794 m
.07616 .325 L
s
.07616 .325 m
.07322 .32578 L
s
.07322 .32578 m
.07107 .32363 L
s
.07107 .32363 m
.07185 .32069 L
s
.07185 .32069 m
.07479 .31991 L
s
.07479 .31991 m
.07264 .31775 L
s
.07264 .31775 m
.07343 .31481 L
s
.07343 .31481 m
.07637 .31403 L
s
.07637 .31403 m
.07852 .31618 L
s
.07852 .31618 m
.07931 .31324 L
s
.07931 .31324 m
.08225 .31245 L
s
.08225 .31245 m
.0844 .3146 L
s
.0844 .3146 m
.08361 .31754 L
s
.08361 .31754 m
.08655 .31676 L
s
.08655 .31676 m
.0887 .31891 L
s
.0887 .31891 m
.08949 .31597 L
s
.08949 .31597 m
.09243 .31518 L
s
.09243 .31518 m
.09028 .31303 L
s
.09028 .31303 m
.09107 .31009 L
s
.09107 .31009 m
.09401 .3093 L
s
.09401 .3093 m
.09616 .31145 L
s
.09616 .31145 m
.09695 .30851 L
s
.09695 .30851 m
.09988 .30773 L
s
.09988 .30773 m
.09773 .30557 L
s
.09773 .30557 m
.09852 .30263 L
s
.09852 .30263 m
.09558 .30342 L
s
.09558 .30342 m
.09343 .30127 L
s
.09343 .30127 m
.09422 .29833 L
s
.09422 .29833 m
.09716 .29754 L
s
.09716 .29754 m
.095 .29539 L
s
.095 .29539 m
.09579 .29245 L
s
.09579 .29245 m
.09285 .29324 L
s
.09285 .29324 m
.0907 .29109 L
s
.0907 .29109 m
.08991 .29403 L
s
.08991 .29403 m
.08697 .29482 L
s
.08697 .29482 m
.08482 .29266 L
s
.08482 .29266 m
.08561 .28972 L
s
.08561 .28972 m
.08267 .29051 L
s
.08267 .29051 m
.08052 .28836 L
s
.08052 .28836 m
.08131 .28542 L
s
.08131 .28542 m
.08425 .28463 L
s
.08425 .28463 m
.08209 .28248 L
s
.08209 .28248 m
.08288 .27954 L
s
.08288 .27954 m
.08582 .27875 L
s
.08582 .27875 m
.08797 .28091 L
s
.08797 .28091 m
.08876 .27797 L
s
.08876 .27797 m
.0917 .27718 L
s
.0917 .27718 m
.08955 .27503 L
s
.08955 .27503 m
.09034 .27209 L
s
.09034 .27209 m
.0874 .27287 L
s
.0874 .27287 m
.08524 .27072 L
s
.08524 .27072 m
.08603 .26778 L
s
.08603 .26778 m
.08897 .267 L
s
.08897 .267 m
.08682 .26484 L
s
.08682 .26484 m
.08761 .2619 L
s
.08761 .2619 m
.08976 .25975 L
s
.08976 .25975 m
.0927 .26054 L
s
.0927 .26054 m
.09191 .2576 L
s
.09191 .2576 m
.09406 .25545 L
s
.09406 .25545 m
.097 .25624 L
s
.097 .25624 m
.09779 .25918 L
s
.09779 .25918 m
.09994 .25702 L
s
.09994 .25702 m
.10288 .25781 L
s
.10288 .25781 m
.10209 .25487 L
s
.10209 .25487 m
.10424 .25272 L
s
.10424 .25272 m
.10131 .25193 L
s
.10131 .25193 m
.10052 .24899 L
s
.10052 .24899 m
.10267 .24684 L
s
.10267 .24684 m
.10561 .24763 L
s
.10561 .24763 m
.10482 .24469 L
s
.10482 .24469 m
.10697 .24254 L
s
.10697 .24254 m
.10991 .24333 L
s
.10991 .24333 m
.1107 .24627 L
s
.1107 .24627 m
.11285 .24411 L
s
.11285 .24411 m
.11579 .2449 L
s
.11579 .2449 m
.11658 .24784 L
s
.11658 .24784 m
.11443 .24999 L
s
.11443 .24999 m
.11737 .25078 L
s
.11737 .25078 m
.11815 .25372 L
s
.11815 .25372 m
.12031 .25157 L
s
.12031 .25157 m
.12325 .25236 L
s
.12325 .25236 m
.12246 .24942 L
s
.12246 .24942 m
.12461 .24726 L
s
.12461 .24726 m
.12755 .24805 L
s
.12755 .24805 m
.12834 .25099 L
s
.12834 .25099 m
.13049 .24884 L
s
.13049 .24884 m
.13343 .24963 L
s
.13343 .24963 m
.13264 .24669 L
s
.13264 .24669 m
.13479 .24454 L
s
.13479 .24454 m
.13185 .24375 L
s
.13185 .24375 m
.13107 .24081 L
s
.13107 .24081 m
.13322 .23866 L
s
.13322 .23866 m
.13616 .23944 L
s
.13616 .23944 m
.13537 .2365 L
s
.13537 .2365 m
.13752 .23435 L
s
.13752 .23435 m
.13458 .23357 L
s
.13458 .23357 m
.13379 .23063 L
s
.13379 .23063 m
.13164 .23278 L
s
.13164 .23278 m
.1287 .23199 L
s
.1287 .23199 m
.12792 .22905 L
s
.12792 .22905 m
.13007 .2269 L
s
.13007 .2269 m
.12713 .22611 L
s
.12713 .22611 m
.12634 .22317 L
s
.12634 .22317 m
.12849 .22102 L
s
.12849 .22102 m
.13143 .22181 L
s
.13143 .22181 m
.13064 .21887 L
s
.13064 .21887 m
.1328 .21672 L
s
.1328 .21672 m
.13573 .2175 L
s
.13573 .2175 m
.13652 .22044 L
s
.13652 .22044 m
.13867 .21829 L
s
.13867 .21829 m
.14161 .21908 L
s
.14161 .21908 m
.14083 .21614 L
s
.14083 .21614 m
.14298 .21399 L
s
.14298 .21399 m
.14004 .2132 L
s
.14004 .2132 m
.13925 .21026 L
s
.13925 .21026 m
.1414 .20811 L
s
.1414 .20811 m
.14434 .2089 L
s
.14434 .2089 m
.14355 .20596 L
s
.14355 .20596 m
.14571 .20381 L
s
.14571 .20381 m
.14865 .20459 L
s
.14865 .20459 m
.14943 .20753 L
s
.14943 .20753 m
.15159 .20538 L
s
.15159 .20538 m
.15452 .20617 L
s
.15452 .20617 m
.15531 .20911 L
s
.15531 .20911 m
.15316 .21126 L
s
.15316 .21126 m
.1561 .21205 L
s
.1561 .21205 m
.15689 .21499 L
s
.15689 .21499 m
.15904 .21283 L
s
.15904 .21283 m
.16198 .21362 L
s
.16198 .21362 m
.16119 .21068 L
s
.16119 .21068 m
.16334 .20853 L
s
.16334 .20853 m
.16628 .20932 L
s
.16628 .20932 m
.16707 .21226 L
s
.16707 .21226 m
.16922 .21011 L
s
.16922 .21011 m
.17216 .21089 L
s
.17216 .21089 m
.17295 .21383 L
s
.17295 .21383 m
.1708 .21599 L
s
.1708 .21599 m
.17374 .21677 L
s
.17374 .21677 m
.17452 .21971 L
s
.17452 .21971 m
.17237 .22186 L
s
.17237 .22186 m
.16943 .22108 L
s
.16943 .22108 m
.17022 .22402 L
s
.17022 .22402 m
.16807 .22617 L
s
.16807 .22617 m
.17101 .22696 L
s
.17101 .22696 m
.1718 .22989 L
s
.1718 .22989 m
.17395 .22774 L
s
.17395 .22774 m
.17689 .22853 L
s
.17689 .22853 m
.17767 .23147 L
s
.17767 .23147 m
.17552 .23362 L
s
.17552 .23362 m
.17846 .23441 L
s
.17846 .23441 m
.17925 .23735 L
s
.17925 .23735 m
.1814 .2352 L
s
.1814 .2352 m
.18434 .23598 L
s
.18434 .23598 m
.18355 .23305 L
s
.18355 .23305 m
.18571 .23089 L
s
.18571 .23089 m
.18864 .23168 L
s
.18864 .23168 m
.18943 .23462 L
s
.18943 .23462 m
.19158 .23247 L
s
.19158 .23247 m
.19452 .23326 L
s
.19452 .23326 m
.19374 .23032 L
s
.19374 .23032 m
.19589 .22817 L
s
.19589 .22817 m
.19295 .22738 L
s
.19295 .22738 m
.19216 .22444 L
s
.19216 .22444 m
.19431 .22229 L
s
.19431 .22229 m
.19725 .22307 L
s
.19725 .22307 m
.19646 .22013 L
s
.19646 .22013 m
.19862 .21798 L
s
.19862 .21798 m
.20156 .21877 L
s
.20156 .21877 m
.20234 .22171 L
s
.20234 .22171 m
.2045 .21956 L
s
.2045 .21956 m
.20743 .22035 L
s
.20743 .22035 m
.20822 .22329 L
s
.20822 .22329 m
.20607 .22544 L
s
.20607 .22544 m
.20901 .22622 L
s
.20901 .22622 m
.2098 .22916 L
s
.2098 .22916 m
.21195 .22701 L
s
.21195 .22701 m
.21489 .2278 L
s
.21489 .2278 m
.2141 .22486 L
s
.2141 .22486 m
.21625 .22271 L
s
.21625 .22271 m
.21919 .2235 L
s
.21919 .2235 m
.21998 .22644 L
s
.21998 .22644 m
.22213 .22428 L
s
.22213 .22428 m
.22507 .22507 L
s
.22507 .22507 m
.22428 .22213 L
s
.22428 .22213 m
.22644 .21998 L
s
.22644 .21998 m
.2235 .21919 L
s
.2235 .21919 m
.22271 .21625 L
s
.22271 .21625 m
.22486 .2141 L
s
.22486 .2141 m
.2278 .21489 L
s
.2278 .21489 m
.22701 .21195 L
s
.22701 .21195 m
.22916 .2098 L
s
.22916 .2098 m
.22622 .20901 L
s
.22622 .20901 m
.22544 .20607 L
s
.22544 .20607 m
.22329 .20822 L
s
.22329 .20822 m
.22035 .20743 L
s
.22035 .20743 m
.21956 .2045 L
s
.21956 .2045 m
.22171 .20234 L
s
.22171 .20234 m
.21877 .20156 L
s
.21877 .20156 m
.21798 .19862 L
s
.21798 .19862 m
.22013 .19646 L
s
.22013 .19646 m
.22307 .19725 L
s
.22307 .19725 m
.22229 .19431 L
s
.22229 .19431 m
.22444 .19216 L
s
.22444 .19216 m
.22738 .19295 L
s
.22738 .19295 m
.22817 .19589 L
s
.22817 .19589 m
.23032 .19374 L
s
.23032 .19374 m
.23326 .19452 L
s
.23326 .19452 m
.23247 .19158 L
s
.23247 .19158 m
.23462 .18943 L
s
.23462 .18943 m
.23168 .18864 L
s
.23168 .18864 m
.23089 .18571 L
s
.23089 .18571 m
.23305 .18355 L
s
.23305 .18355 m
.23598 .18434 L
s
.23598 .18434 m
.2352 .1814 L
s
.2352 .1814 m
.23735 .17925 L
s
.23735 .17925 m
.23441 .17846 L
s
.23441 .17846 m
.23362 .17552 L
s
.23362 .17552 m
.23147 .17767 L
s
.23147 .17767 m
.22853 .17689 L
s
.22853 .17689 m
.22774 .17395 L
s
.22774 .17395 m
.22989 .1718 L
s
.22989 .1718 m
.22696 .17101 L
s
.22696 .17101 m
.22617 .16807 L
s
.22617 .16807 m
.22402 .17022 L
s
.22402 .17022 m
.22108 .16943 L
s
.22108 .16943 m
.22186 .17237 L
s
.22186 .17237 m
.21971 .17452 L
s
.21971 .17452 m
.21677 .17374 L
s
.21677 .17374 m
.21599 .1708 L
s
.21599 .1708 m
.21383 .17295 L
s
.21383 .17295 m
.21089 .17216 L
s
.21089 .17216 m
.21011 .16922 L
s
.21011 .16922 m
.21226 .16707 L
s
.21226 .16707 m
.20932 .16628 L
s
.20932 .16628 m
.20853 .16334 L
s
.20853 .16334 m
.21068 .16119 L
s
.21068 .16119 m
.21362 .16198 L
s
.21362 .16198 m
.21283 .15904 L
s
.21283 .15904 m
.21499 .15689 L
s
.21499 .15689 m
.21205 .1561 L
s
.21205 .1561 m
.21126 .15316 L
s
.21126 .15316 m
.20911 .15531 L
s
.20911 .15531 m
.20617 .15452 L
s
.20617 .15452 m
.20538 .15159 L
s
.20538 .15159 m
.20753 .14943 L
s
.20753 .14943 m
.20459 .14865 L
s
.20459 .14865 m
.20381 .14571 L
s
.20381 .14571 m
.20596 .14355 L
s
.20596 .14355 m
.2089 .14434 L
s
.2089 .14434 m
.20811 .1414 L
s
.20811 .1414 m
.21026 .13925 L
s
.21026 .13925 m
.2132 .14004 L
s
.2132 .14004 m
.21399 .14298 L
s
.21399 .14298 m
.21614 .14083 L
s
.21614 .14083 m
.21908 .14161 L
s
.21908 .14161 m
.21829 .13867 L
s
.21829 .13867 m
.22044 .13652 L
s
.22044 .13652 m
.2175 .13573 L
s
.2175 .13573 m
.21672 .1328 L
s
.21672 .1328 m
.21887 .13064 L
s
.21887 .13064 m
.22181 .13143 L
s
.22181 .13143 m
.22102 .12849 L
s
.22102 .12849 m
.22317 .12634 L
s
.22317 .12634 m
.22611 .12713 L
s
.22611 .12713 m
.2269 .13007 L
s
.2269 .13007 m
.22905 .12792 L
s
.22905 .12792 m
.23199 .1287 L
s
.23199 .1287 m
.23278 .13164 L
s
.23278 .13164 m
.23063 .13379 L
s
.23063 .13379 m
.23357 .13458 L
s
.23357 .13458 m
.23435 .13752 L
s
.23435 .13752 m
.2365 .13537 L
s
.2365 .13537 m
.23944 .13616 L
s
.23944 .13616 m
.23866 .13322 L
s
.23866 .13322 m
.24081 .13107 L
s
.24081 .13107 m
.24375 .13185 L
s
.24375 .13185 m
.24454 .13479 L
s
.24454 .13479 m
.24669 .13264 L
s
.24669 .13264 m
.24963 .13343 L
s
.24963 .13343 m
.24884 .13049 L
s
.24884 .13049 m
.25099 .12834 L
s
.25099 .12834 m
.24805 .12755 L
s
.24805 .12755 m
.24726 .12461 L
s
.24726 .12461 m
.24942 .12246 L
s
.24942 .12246 m
.25236 .12325 L
s
.25236 .12325 m
.25157 .12031 L
s
.25157 .12031 m
.25372 .11815 L
s
.25372 .11815 m
.25078 .11737 L
s
.25078 .11737 m
.24999 .11443 L
s
.24999 .11443 m
.24784 .11658 L
s
.24784 .11658 m
.2449 .11579 L
s
.2449 .11579 m
.24411 .11285 L
s
.24411 .11285 m
.24627 .1107 L
s
.24627 .1107 m
.24333 .10991 L
s
.24333 .10991 m
.24254 .10697 L
s
.24254 .10697 m
.24469 .10482 L
s
.24469 .10482 m
.24763 .10561 L
s
.24763 .10561 m
.24684 .10267 L
s
.24684 .10267 m
.24899 .10052 L
s
.24899 .10052 m
.25193 .10131 L
s
.25193 .10131 m
.25272 .10424 L
s
.25272 .10424 m
.25487 .10209 L
s
.25487 .10209 m
.25781 .10288 L
s
.25781 .10288 m
.25702 .09994 L
s
.25702 .09994 m
.25918 .09779 L
s
.25918 .09779 m
.25624 .097 L
s
.25624 .097 m
.25545 .09406 L
s
.25545 .09406 m
.2576 .09191 L
s
.2576 .09191 m
.26054 .0927 L
s
.26054 .0927 m
.25975 .08976 L
s
.25975 .08976 m
.2619 .08761 L
s
.2619 .08761 m
.26484 .08682 L
s
.26484 .08682 m
.267 .08897 L
s
.267 .08897 m
.26778 .08603 L
s
.26778 .08603 m
.27072 .08524 L
s
.27072 .08524 m
.27287 .0874 L
s
.27287 .0874 m
.27209 .09034 L
s
.27209 .09034 m
.27503 .08955 L
s
.27503 .08955 m
.27718 .0917 L
s
.27718 .0917 m
.27797 .08876 L
s
.27797 .08876 m
.28091 .08797 L
s
.28091 .08797 m
.27875 .08582 L
s
.27875 .08582 m
.27954 .08288 L
s
.27954 .08288 m
.28248 .08209 L
s
.28248 .08209 m
.28463 .08425 L
s
.28463 .08425 m
.28542 .08131 L
s
.28542 .08131 m
.28836 .08052 L
s
.28836 .08052 m
.29051 .08267 L
s
.29051 .08267 m
.28972 .08561 L
s
.28972 .08561 m
.29266 .08482 L
s
.29266 .08482 m
.29482 .08697 L
s
.29482 .08697 m
.29403 .08991 L
s
.29403 .08991 m
.29109 .0907 L
s
.29109 .0907 m
.29324 .09285 L
s
.29324 .09285 m
.29245 .09579 L
s
.29245 .09579 m
.29539 .095 L
s
.29539 .095 m
.29754 .09716 L
s
.29754 .09716 m
.29833 .09422 L
s
.29833 .09422 m
.30127 .09343 L
s
.30127 .09343 m
.30342 .09558 L
s
.30342 .09558 m
.30263 .09852 L
s
.30263 .09852 m
.30557 .09773 L
s
.30557 .09773 m
.30773 .09988 L
s
.30773 .09988 m
.30851 .09695 L
s
.30851 .09695 m
.31145 .09616 L
s
.31145 .09616 m
.3093 .09401 L
s
.3093 .09401 m
.31009 .09107 L
s
.31009 .09107 m
.31303 .09028 L
s
.31303 .09028 m
.31518 .09243 L
s
.31518 .09243 m
.31597 .08949 L
s
.31597 .08949 m
.31891 .0887 L
s
.31891 .0887 m
.31676 .08655 L
s
.31676 .08655 m
.31754 .08361 L
s
.31754 .08361 m
.3146 .0844 L
s
.3146 .0844 m
.31245 .08225 L
s
.31245 .08225 m
.31324 .07931 L
s
.31324 .07931 m
.31618 .07852 L
s
.31618 .07852 m
.31403 .07637 L
s
.31403 .07637 m
.31481 .07343 L
s
.31481 .07343 m
.31775 .07264 L
s
.31775 .07264 m
.31991 .07479 L
s
.31991 .07479 m
.32069 .07185 L
s
.32069 .07185 m
.32363 .07107 L
s
.32363 .07107 m
.32578 .07322 L
s
.32578 .07322 m
.325 .07616 L
s
.325 .07616 m
.32794 .07537 L
s
.32794 .07537 m
.33009 .07752 L
s
.33009 .07752 m
.33088 .07458 L
s
.33088 .07458 m
.33382 .0738 L
s
.33382 .0738 m
.33166 .07164 L
s
.33166 .07164 m
.33245 .0687 L
s
.33245 .0687 m
.33539 .06792 L
s
.33539 .06792 m
.33754 .07007 L
s
.33754 .07007 m
.33833 .06713 L
s
.33833 .06713 m
.34127 .06634 L
s
.34127 .06634 m
.34342 .06849 L
s
.34342 .06849 m
.34263 .07143 L
s
.34263 .07143 m
.34557 .07064 L
s
.34557 .07064 m
.34773 .0728 L
s
.34773 .0728 m
.34694 .07574 L
s
.34694 .07574 m
.344 .07652 L
s
.344 .07652 m
.34615 .07868 L
s
.34615 .07868 m
.34536 .08161 L
s
.34536 .08161 m
.3483 .08083 L
s
.3483 .08083 m
.35045 .08298 L
s
.35045 .08298 m
.35124 .08004 L
s
.35124 .08004 m
.35418 .07925 L
s
.35418 .07925 m
.35633 .0814 L
s
.35633 .0814 m
.35554 .08434 L
s
.35554 .08434 m
.35848 .08356 L
s
.35848 .08356 m
.36064 .08571 L
s
.36064 .08571 m
.35985 .08865 L
s
.35985 .08865 m
.35691 .08943 L
s
.35691 .08943 m
.35906 .09159 L
s
.35906 .09159 m
.35827 .09453 L
s
.35827 .09453 m
.35533 .09531 L
s
.35533 .09531 m
.35318 .09316 L
s
.35318 .09316 m
.35239 .0961 L
s
.35239 .0961 m
.34946 .09689 L
s
.34946 .09689 m
.35161 .09904 L
s
.35161 .09904 m
.35082 .10198 L
s
.35082 .10198 m
.35376 .10119 L
s
.35376 .10119 m
.35591 .10334 L
s
.35591 .10334 m
.35512 .10628 L
s
.35512 .10628 m
.35218 .10707 L
s
.35218 .10707 m
.35434 .10922 L
s
.35434 .10922 m
.35355 .11216 L
s
.35355 .11216 m
.35649 .11137 L
s
.35649 .11137 m
.35864 .11353 L
s
.35864 .11353 m
.35943 .11059 L
s
.35943 .11059 m
.36237 .1098 L
s
.36237 .1098 m
.36452 .11195 L
s
.36452 .11195 m
.36373 .11489 L
s
.36373 .11489 m
.36667 .1141 L
s
.36667 .1141 m
.36882 .11626 L
s
.36882 .11626 m
.36961 .11332 L
s
.36961 .11332 m
.37255 .11253 L
s
.37255 .11253 m
.3704 .11038 L
s
.3704 .11038 m
.37118 .10744 L
s
.37118 .10744 m
.37412 .10665 L
s
.37412 .10665 m
.37628 .1088 L
s
.37628 .1088 m
.37706 .10586 L
s
.37706 .10586 m
.38 .10507 L
s
.38 .10507 m
.38215 .10723 L
s
.38215 .10723 m
.38137 .11017 L
s
.38137 .11017 m
.38431 .10938 L
s
.38431 .10938 m
.38646 .11153 L
s
.38646 .11153 m
.38567 .11447 L
s
.38567 .11447 m
.38273 .11526 L
s
.38273 .11526 m
.38488 .11741 L
s
.38488 .11741 m
.3841 .12035 L
s
.3841 .12035 m
.38703 .11956 L
s
.38703 .11956 m
.38919 .12171 L
s
.38919 .12171 m
.38997 .11877 L
s
.38997 .11877 m
.39291 .11798 L
s
.39291 .11798 m
.39507 .12014 L
s
.39507 .12014 m
.39428 .12308 L
s
.39428 .12308 m
.39722 .12229 L
s
.39722 .12229 m
.39937 .12444 L
s
.39937 .12444 m
.40016 .1215 L
s
.40016 .1215 m
.4031 .12071 L
s
.4031 .12071 m
.40094 .11856 L
s
.40094 .11856 m
.40173 .11562 L
s
.40173 .11562 m
.40467 .11483 L
s
.40467 .11483 m
.40682 .11699 L
s
.40682 .11699 m
.40761 .11405 L
s
.40761 .11405 m
.41055 .11326 L
s
.41055 .11326 m
.4084 .11111 L
s
.4084 .11111 m
.40919 .10817 L
s
.40919 .10817 m
.40625 .10896 L
s
.40625 .10896 m
.40409 .1068 L
s
.40409 .1068 m
.40488 .10386 L
s
.40488 .10386 m
.40782 .10308 L
s
.40782 .10308 m
.40567 .10092 L
s
.40567 .10092 m
.40646 .09799 L
s
.40646 .09799 m
.4094 .0972 L
s
.4094 .0972 m
.41155 .09935 L
s
.41155 .09935 m
.41234 .09641 L
s
.41234 .09641 m
.41528 .09562 L
s
.41528 .09562 m
.41743 .09777 L
s
.41743 .09777 m
.41664 .10071 L
s
.41664 .10071 m
.41958 .09993 L
s
.41958 .09993 m
.42173 .10208 L
s
.42173 .10208 m
.42252 .09914 L
s
.42252 .09914 m
.42546 .09835 L
s
.42546 .09835 m
.42331 .0962 L
s
.42331 .0962 m
.42409 .09326 L
s
.42409 .09326 m
.42703 .09247 L
s
.42703 .09247 m
.42919 .09462 L
s
.42919 .09462 m
.42997 .09168 L
s
.42997 .09168 m
.43291 .0909 L
s
.43291 .0909 m
.43076 .08874 L
s
.43076 .08874 m
.43155 .08581 L
s
.43155 .08581 m
.42861 .08659 L
s
.42861 .08659 m
.42646 .08444 L
s
.42646 .08444 m
.42724 .0815 L
s
.42724 .0815 m
.43018 .08071 L
s
.43018 .08071 m
.42803 .07856 L
s
.42803 .07856 m
.42882 .07562 L
s
.42882 .07562 m
.42588 .07641 L
s
.42588 .07641 m
.42373 .07426 L
s
.42373 .07426 m
.42294 .0772 L
s
.42294 .0772 m
.42 .07799 L
s
.42 .07799 m
.41785 .07583 L
s
.41785 .07583 m
.41864 .07289 L
s
.41864 .07289 m
.4157 .07368 L
s
.4157 .07368 m
.41355 .07153 L
s
.41355 .07153 m
.41433 .06859 L
s
.41433 .06859 m
.41727 .0678 L
s
.41727 .0678 m
.41512 .06565 L
s
.41512 .06565 m
.41591 .06271 L
s
.41591 .06271 m
.41885 .06192 L
s
.41885 .06192 m
.421 .06408 L
s
.421 .06408 m
.42179 .06114 L
s
.42179 .06114 m
.42473 .06035 L
s
.42473 .06035 m
.42258 .0582 L
s
.42258 .0582 m
.42336 .05526 L
s
.42336 .05526 m
.42042 .05605 L
s
.42042 .05605 m
.41827 .05389 L
s
.41827 .05389 m
.41906 .05095 L
s
.41906 .05095 m
.422 .05017 L
s
.422 .05017 m
.41985 .04801 L
s
.41985 .04801 m
.42063 .04508 L
s
.42063 .04508 m
.42357 .04429 L
s
.42357 .04429 m
.42573 .04644 L
s
.42573 .04644 m
.42651 .0435 L
s
.42651 .0435 m
.42945 .04271 L
s
.42945 .04271 m
.43161 .04486 L
s
.43161 .04486 m
.43082 .0478 L
s
.43082 .0478 m
.43376 .04702 L
s
.43376 .04702 m
.43591 .04917 L
s
.43591 .04917 m
.4367 .04623 L
s
.4367 .04623 m
.43964 .04544 L
s
.43964 .04544 m
.43748 .04329 L
s
.43748 .04329 m
.43827 .04035 L
s
.43827 .04035 m
.44121 .03956 L
s
.44121 .03956 m
.44336 .04171 L
s
.44336 .04171 m
.44415 .03877 L
s
.44415 .03877 m
.44709 .03799 L
s
.44709 .03799 m
.44924 .04014 L
s
.44924 .04014 m
.44845 .04308 L
s
.44845 .04308 m
.45139 .04229 L
s
.45139 .04229 m
.45355 .04444 L
s
.45355 .04444 m
.45276 .04738 L
s
.45276 .04738 m
.44982 .04817 L
s
.44982 .04817 m
.45197 .05032 L
s
.45197 .05032 m
.45118 .05326 L
s
.45118 .05326 m
.45412 .05247 L
s
.45412 .05247 m
.45627 .05462 L
s
.45627 .05462 m
.45706 .05169 L
s
.45706 .05169 m
.46 .0509 L
s
.46 .0509 m
.46215 .05305 L
s
.46215 .05305 m
.46137 .05599 L
s
.46137 .05599 m
.4643 .0552 L
s
.4643 .0552 m
.46646 .05735 L
s
.46646 .05735 m
.46724 .05441 L
s
.46724 .05441 m
.47018 .05363 L
s
.47018 .05363 m
.46803 .05147 L
s
.46803 .05147 m
.46882 .04853 L
s
.46882 .04853 m
.47176 .04775 L
s
.47176 .04775 m
.47391 .0499 L
s
.47391 .0499 m
.4747 .04696 L
s
.4747 .04696 m
.47764 .04617 L
s
.47764 .04617 m
.47549 .04402 L
s
.47549 .04402 m
.47627 .04108 L
s
.47627 .04108 m
.47333 .04187 L
s
.47333 .04187 m
.47118 .03972 L
s
.47118 .03972 m
.47197 .03678 L
s
.47197 .03678 m
.47491 .03599 L
s
.47491 .03599 m
.47276 .03384 L
s
.47276 .03384 m
.47354 .0309 L
s
.47354 .0309 m
.47648 .03011 L
s
.47648 .03011 m
.47864 .03226 L
s
.47864 .03226 m
.47942 .02932 L
s
.47942 .02932 m
.48236 .02854 L
s
.48236 .02854 m
.48452 .03069 L
s
.48452 .03069 m
.48373 .03363 L
s
.48373 .03363 m
.48667 .03284 L
s
.48667 .03284 m
.48882 .03499 L
s
.48882 .03499 m
.48961 .03205 L
s
.48961 .03205 m
.49255 .03126 L
s
.49255 .03126 m
.49039 .02911 L
s
.49039 .02911 m
.49118 .02617 L
s
.49118 .02617 m
.49412 .02538 L
s
.49412 .02538 m
.49627 .02754 L
s
.49627 .02754 m
.49706 .0246 L
s
.49706 .0246 m
.5 .02381 L
s
.5 .02381 m
.50294 .0246 L
s
.50294 .0246 m
.50373 .02754 L
s
.50373 .02754 m
.50588 .02538 L
s
.50588 .02538 m
.50882 .02617 L
s
.50882 .02617 m
.50961 .02911 L
s
.50961 .02911 m
.50745 .03126 L
s
.50745 .03126 m
.51039 .03205 L
s
.51039 .03205 m
.51118 .03499 L
s
.51118 .03499 m
.51333 .03284 L
s
.51333 .03284 m
.51627 .03363 L
s
.51627 .03363 m
.51548 .03069 L
s
.51548 .03069 m
.51764 .02854 L
s
.51764 .02854 m
.52058 .02932 L
s
.52058 .02932 m
.52136 .03226 L
s
.52136 .03226 m
.52352 .03011 L
s
.52352 .03011 m
.52646 .0309 L
s
.52646 .0309 m
.52724 .03384 L
s
.52724 .03384 m
.52509 .03599 L
s
.52509 .03599 m
.52803 .03678 L
s
.52803 .03678 m
.52882 .03972 L
s
.52882 .03972 m
.52667 .04187 L
s
.52667 .04187 m
.52373 .04108 L
s
.52373 .04108 m
.52451 .04402 L
s
.52451 .04402 m
.52236 .04617 L
s
.52236 .04617 m
.5253 .04696 L
s
.5253 .04696 m
.52609 .0499 L
s
.52609 .0499 m
.52824 .04775 L
s
.52824 .04775 m
.53118 .04853 L
s
.53118 .04853 m
.53197 .05147 L
s
.53197 .05147 m
.52982 .05363 L
s
.52982 .05363 m
.53276 .05441 L
s
.53276 .05441 m
.53354 .05735 L
s
.53354 .05735 m
.5357 .0552 L
s
.5357 .0552 m
.53863 .05599 L
s
.53863 .05599 m
.53785 .05305 L
s
.53785 .05305 m
.54 .0509 L
s
.54 .0509 m
.54294 .05169 L
s
.54294 .05169 m
.54373 .05462 L
s
.54373 .05462 m
.54588 .05247 L
s
.54588 .05247 m
.54882 .05326 L
s
.54882 .05326 m
.54803 .05032 L
s
.54803 .05032 m
.55018 .04817 L
s
.55018 .04817 m
.54724 .04738 L
s
.54724 .04738 m
.54645 .04444 L
s
.54645 .04444 m
.54861 .04229 L
s
.54861 .04229 m
.55155 .04308 L
s
.55155 .04308 m
.55076 .04014 L
s
.55076 .04014 m
.55291 .03799 L
s
.55291 .03799 m
.55585 .03877 L
s
.55585 .03877 m
.55664 .04171 L
s
.55664 .04171 m
.55879 .03956 L
s
.55879 .03956 m
.56173 .04035 L
s
.56173 .04035 m
.56252 .04329 L
s
.56252 .04329 m
.56036 .04544 L
s
.56036 .04544 m
.5633 .04623 L
s
.5633 .04623 m
.56409 .04917 L
s
.56409 .04917 m
.56624 .04702 L
s
.56624 .04702 m
.56918 .0478 L
s
.56918 .0478 m
.56839 .04486 L
s
.56839 .04486 m
.57055 .04271 L
s
.57055 .04271 m
.57349 .0435 L
s
.57349 .0435 m
.57427 .04644 L
s
.57427 .04644 m
.57643 .04429 L
s
.57643 .04429 m
.57937 .04508 L
s
.57937 .04508 m
.58015 .04801 L
s
.58015 .04801 m
.578 .05017 L
s
.578 .05017 m
.58094 .05095 L
s
.58094 .05095 m
.58173 .05389 L
s
.58173 .05389 m
.57958 .05605 L
s
.57958 .05605 m
.57664 .05526 L
s
.57664 .05526 m
.57742 .0582 L
s
.57742 .0582 m
.57527 .06035 L
s
.57527 .06035 m
.57821 .06114 L
s
.57821 .06114 m
.579 .06408 L
s
.579 .06408 m
.58115 .06192 L
s
.58115 .06192 m
.58409 .06271 L
s
.58409 .06271 m
.58488 .06565 L
s
.58488 .06565 m
.58273 .0678 L
s
.58273 .0678 m
.58567 .06859 L
s
.58567 .06859 m
.58645 .07153 L
s
.58645 .07153 m
.5843 .07368 L
s
.5843 .07368 m
.58136 .07289 L
s
.58136 .07289 m
.58215 .07583 L
s
.58215 .07583 m
.58 .07799 L
s
.58 .07799 m
.57706 .0772 L
s
.57706 .0772 m
.57627 .07426 L
s
.57627 .07426 m
.57412 .07641 L
s
.57412 .07641 m
.57118 .07562 L
s
.57118 .07562 m
.57197 .07856 L
s
.57197 .07856 m
.56982 .08071 L
s
.56982 .08071 m
.57276 .0815 L
s
.57276 .0815 m
.57354 .08444 L
s
.57354 .08444 m
.57139 .08659 L
s
.57139 .08659 m
.56845 .08581 L
s
.56845 .08581 m
.56924 .08874 L
s
.56924 .08874 m
.56709 .0909 L
s
.56709 .0909 m
.57003 .09168 L
s
.57003 .09168 m
.57081 .09462 L
s
.57081 .09462 m
.57297 .09247 L
s
.57297 .09247 m
.57591 .09326 L
s
.57591 .09326 m
.57669 .0962 L
s
.57669 .0962 m
.57454 .09835 L
s
.57454 .09835 m
.57748 .09914 L
s
.57748 .09914 m
.57827 .10208 L
s
.57827 .10208 m
.58042 .09993 L
s
.58042 .09993 m
.58336 .10071 L
s
.58336 .10071 m
.58257 .09777 L
s
.58257 .09777 m
.58472 .09562 L
s
.58472 .09562 m
.58766 .09641 L
s
.58766 .09641 m
.58845 .09935 L
s
.58845 .09935 m
.5906 .0972 L
s
.5906 .0972 m
.59354 .09799 L
s
.59354 .09799 m
.59433 .10092 L
s
.59433 .10092 m
.59218 .10308 L
s
.59218 .10308 m
.59512 .10386 L
s
.59512 .10386 m
.59591 .1068 L
s
.59591 .1068 m
.59375 .10896 L
s
.59375 .10896 m
.59081 .10817 L
s
.59081 .10817 m
.5916 .11111 L
s
.5916 .11111 m
.58945 .11326 L
s
.58945 .11326 m
.59239 .11405 L
s
.59239 .11405 m
.59318 .11699 L
s
.59318 .11699 m
.59533 .11483 L
s
.59533 .11483 m
.59827 .11562 L
s
.59827 .11562 m
.59906 .11856 L
s
.59906 .11856 m
.5969 .12071 L
s
.5969 .12071 m
.59984 .1215 L
s
.59984 .1215 m
.60063 .12444 L
s
.60063 .12444 m
.60278 .12229 L
s
.60278 .12229 m
.60572 .12308 L
s
.60572 .12308 m
.60493 .12014 L
s
.60493 .12014 m
.60709 .11798 L
s
.60709 .11798 m
.61003 .11877 L
s
.61003 .11877 m
.61081 .12171 L
s
.61081 .12171 m
.61297 .11956 L
s
.61297 .11956 m
.6159 .12035 L
s
.6159 .12035 m
.61512 .11741 L
s
.61512 .11741 m
.61727 .11526 L
s
.61727 .11526 m
.61433 .11447 L
s
.61433 .11447 m
.61354 .11153 L
s
.61354 .11153 m
.61569 .10938 L
s
.61569 .10938 m
.61863 .11017 L
s
.61863 .11017 m
.61785 .10723 L
s
.61785 .10723 m
.62 .10507 L
s
.62 .10507 m
.62294 .10586 L
s
.62294 .10586 m
.62372 .1088 L
s
.62372 .1088 m
.62588 .10665 L
s
.62588 .10665 m
.62882 .10744 L
s
.62882 .10744 m
.6296 .11038 L
s
.6296 .11038 m
.62745 .11253 L
s
.62745 .11253 m
.63039 .11332 L
s
.63039 .11332 m
.63118 .11626 L
s
.63118 .11626 m
.63333 .1141 L
s
.63333 .1141 m
.63627 .11489 L
s
.63627 .11489 m
.63548 .11195 L
s
.63548 .11195 m
.63763 .1098 L
s
.63763 .1098 m
.64057 .11059 L
s
.64057 .11059 m
.64136 .11353 L
s
.64136 .11353 m
.64351 .11137 L
s
.64351 .11137 m
.64645 .11216 L
s
.64645 .11216 m
.64566 .10922 L
s
.64566 .10922 m
.64782 .10707 L
s
.64782 .10707 m
.64488 .10628 L
s
.64488 .10628 m
.64409 .10334 L
s
.64409 .10334 m
.64624 .10119 L
s
.64624 .10119 m
.64918 .10198 L
s
.64918 .10198 m
.64839 .09904 L
s
.64839 .09904 m
.65054 .09689 L
s
.65054 .09689 m
.64761 .0961 L
s
.64761 .0961 m
.64682 .09316 L
s
.64682 .09316 m
.64467 .09531 L
s
.64467 .09531 m
.64173 .09453 L
s
.64173 .09453 m
.64094 .09159 L
s
.64094 .09159 m
.64309 .08943 L
s
.64309 .08943 m
.64015 .08865 L
s
.64015 .08865 m
.63936 .08571 L
s
.63936 .08571 m
.64152 .08356 L
s
.64152 .08356 m
.64446 .08434 L
s
.64446 .08434 m
.64367 .0814 L
s
.64367 .0814 m
.64582 .07925 L
s
.64582 .07925 m
.64876 .08004 L
s
.64876 .08004 m
.64955 .08298 L
s
.64955 .08298 m
.6517 .08083 L
s
.6517 .08083 m
.65464 .08161 L
s
.65464 .08161 m
.65385 .07868 L
s
.65385 .07868 m
.656 .07652 L
s
.656 .07652 m
.65306 .07574 L
s
.65306 .07574 m
.65227 .0728 L
s
.65227 .0728 m
.65443 .07064 L
s
.65443 .07064 m
.65737 .07143 L
s
.65737 .07143 m
.65658 .06849 L
s
.65658 .06849 m
.65873 .06634 L
s
.65873 .06634 m
.66167 .06713 L
s
.66167 .06713 m
.66246 .07007 L
s
.66246 .07007 m
.66461 .06792 L
s
.66461 .06792 m
.66755 .0687 L
s
.66755 .0687 m
.66834 .07164 L
s
.66834 .07164 m
.66618 .0738 L
s
.66618 .0738 m
.66912 .07458 L
s
.66912 .07458 m
.66991 .07752 L
s
.66991 .07752 m
.67206 .07537 L
s
.67206 .07537 m
.675 .07616 L
s
.675 .07616 m
.67422 .07322 L
s
.67422 .07322 m
.67637 .07107 L
s
.67637 .07107 m
.67931 .07185 L
s
.67931 .07185 m
.68009 .07479 L
s
.68009 .07479 m
.68225 .07264 L
s
.68225 .07264 m
.68519 .07343 L
s
.68519 .07343 m
.68597 .07637 L
s
.68597 .07637 m
.68382 .07852 L
s
.68382 .07852 m
.68676 .07931 L
s
.68676 .07931 m
.68755 .08225 L
s
.68755 .08225 m
.6854 .0844 L
s
.6854 .0844 m
.68246 .08361 L
s
.68246 .08361 m
.68324 .08655 L
s
.68324 .08655 m
.68109 .0887 L
s
.68109 .0887 m
.68403 .08949 L
s
.68403 .08949 m
.68482 .09243 L
s
.68482 .09243 m
.68697 .09028 L
s
.68697 .09028 m
.68991 .09107 L
s
.68991 .09107 m
.6907 .09401 L
s
.6907 .09401 m
.68855 .09616 L
s
.68855 .09616 m
.69149 .09695 L
s
.69149 .09695 m
.69227 .09988 L
s
.69227 .09988 m
.69443 .09773 L
s
.69443 .09773 m
.69737 .09852 L
s
.69737 .09852 m
.69658 .09558 L
s
.69658 .09558 m
.69873 .09343 L
s
.69873 .09343 m
.70167 .09422 L
s
.70167 .09422 m
.70246 .09716 L
s
.70246 .09716 m
.70461 .095 L
s
.70461 .095 m
.70755 .09579 L
s
.70755 .09579 m
.70676 .09285 L
s
.70676 .09285 m
.70891 .0907 L
s
.70891 .0907 m
.70597 .08991 L
s
.70597 .08991 m
.70518 .08697 L
s
.70518 .08697 m
.70734 .08482 L
s
.70734 .08482 m
.71028 .08561 L
s
.71028 .08561 m
.70949 .08267 L
s
.70949 .08267 m
.71164 .08052 L
s
.71164 .08052 m
.71458 .08131 L
s
.71458 .08131 m
.71537 .08425 L
s
.71537 .08425 m
.71752 .08209 L
s
.71752 .08209 m
.72046 .08288 L
s
.72046 .08288 m
.72125 .08582 L
s
.72125 .08582 m
.71909 .08797 L
s
.71909 .08797 m
.72203 .08876 L
s
.72203 .08876 m
.72282 .0917 L
s
.72282 .0917 m
.72497 .08955 L
s
.72497 .08955 m
.72791 .09034 L
s
.72791 .09034 m
.72713 .0874 L
s
.72713 .0874 m
.72928 .08524 L
s
.72928 .08524 m
.73222 .08603 L
s
.73222 .08603 m
.733 .08897 L
s
.733 .08897 m
.73516 .08682 L
s
.73516 .08682 m
.7381 .08761 L
s
.7381 .08761 m
.74025 .08976 L
s
.74025 .08976 m
.73946 .0927 L
s
.73946 .0927 m
.7424 .09191 L
s
.7424 .09191 m
.74455 .09406 L
s
.74455 .09406 m
.74376 .097 L
s
.74376 .097 m
.74082 .09779 L
s
.74082 .09779 m
.74298 .09994 L
s
.74298 .09994 m
.74219 .10288 L
s
.74219 .10288 m
.74513 .10209 L
s
.74513 .10209 m
.74728 .10424 L
s
.74728 .10424 m
.74807 .10131 L
s
.74807 .10131 m
.75101 .10052 L
s
.75101 .10052 m
.75316 .10267 L
s
.75316 .10267 m
.75237 .10561 L
s
.75237 .10561 m
.75531 .10482 L
s
.75531 .10482 m
.75746 .10697 L
s
.75746 .10697 m
.75667 .10991 L
s
.75667 .10991 m
.75373 .1107 L
s
.75373 .1107 m
.75589 .11285 L
s
.75589 .11285 m
.7551 .11579 L
s
.7551 .11579 m
.75216 .11658 L
s
.75216 .11658 m
.75001 .11443 L
s
.75001 .11443 m
.74922 .11737 L
s
.74922 .11737 m
.74628 .11815 L
s
.74628 .11815 m
.74843 .12031 L
s
.74843 .12031 m
.74764 .12325 L
s
.74764 .12325 m
.75058 .12246 L
s
.75058 .12246 m
.75274 .12461 L
s
.75274 .12461 m
.75195 .12755 L
s
.75195 .12755 m
.74901 .12834 L
s
.74901 .12834 m
.75116 .13049 L
s
.75116 .13049 m
.75037 .13343 L
s
.75037 .13343 m
.75331 .13264 L
s
.75331 .13264 m
.75546 .13479 L
s
.75546 .13479 m
.75625 .13185 L
s
.75625 .13185 m
.75919 .13107 L
s
.75919 .13107 m
.76134 .13322 L
s
.76134 .13322 m
.76056 .13616 L
s
.76056 .13616 m
.7635 .13537 L
s
.7635 .13537 m
.76565 .13752 L
s
.76565 .13752 m
.76643 .13458 L
s
.76643 .13458 m
.76937 .13379 L
s
.76937 .13379 m
.76722 .13164 L
s
.76722 .13164 m
.76801 .1287 L
s
.76801 .1287 m
.77095 .12792 L
s
.77095 .12792 m
.7731 .13007 L
s
.7731 .13007 m
.77389 .12713 L
s
.77389 .12713 m
.77683 .12634 L
s
.77683 .12634 m
.77898 .12849 L
s
.77898 .12849 m
.77819 .13143 L
s
.77819 .13143 m
.78113 .13064 L
s
.78113 .13064 m
.78328 .1328 L
s
.78328 .1328 m
.7825 .13573 L
s
.7825 .13573 m
.77956 .13652 L
s
.77956 .13652 m
.78171 .13867 L
s
.78171 .13867 m
.78092 .14161 L
s
.78092 .14161 m
.78386 .14083 L
s
.78386 .14083 m
.78601 .14298 L
s
.78601 .14298 m
.7868 .14004 L
s
.7868 .14004 m
.78974 .13925 L
s
.78974 .13925 m
.79189 .1414 L
s
.79189 .1414 m
.7911 .14434 L
s
.7911 .14434 m
.79404 .14355 L
s
.79404 .14355 m
.79619 .14571 L
s
.79619 .14571 m
.79541 .14865 L
s
.79541 .14865 m
.79247 .14943 L
s
.79247 .14943 m
.79462 .15159 L
s
.79462 .15159 m
.79383 .15452 L
s
.79383 .15452 m
.79089 .15531 L
s
.79089 .15531 m
.78874 .15316 L
s
.78874 .15316 m
.78795 .1561 L
s
.78795 .1561 m
.78501 .15689 L
s
.78501 .15689 m
.78717 .15904 L
s
.78717 .15904 m
.78638 .16198 L
s
.78638 .16198 m
.78932 .16119 L
s
.78932 .16119 m
.79147 .16334 L
s
.79147 .16334 m
.79068 .16628 L
s
.79068 .16628 m
.78774 .16707 L
s
.78774 .16707 m
.78989 .16922 L
s
.78989 .16922 m
.78911 .17216 L
s
.78911 .17216 m
.78617 .17295 L
s
.78617 .17295 m
.78401 .1708 L
s
.78401 .1708 m
.78323 .17374 L
s
.78323 .17374 m
.78029 .17452 L
s
.78029 .17452 m
.77814 .17237 L
s
.77814 .17237 m
.77892 .16943 L
s
.77892 .16943 m
.77598 .17022 L
s
.77598 .17022 m
.77383 .16807 L
s
.77383 .16807 m
.77304 .17101 L
s
.77304 .17101 m
.77011 .1718 L
s
.77011 .1718 m
.77226 .17395 L
s
.77226 .17395 m
.77147 .17689 L
s
.77147 .17689 m
.76853 .17767 L
s
.76853 .17767 m
.76638 .17552 L
s
.76638 .17552 m
.76559 .17846 L
s
.76559 .17846 m
.76265 .17925 L
s
.76265 .17925 m
.7648 .1814 L
s
.7648 .1814 m
.76402 .18434 L
s
.76402 .18434 m
.76695 .18355 L
s
.76695 .18355 m
.76911 .18571 L
s
.76911 .18571 m
.76832 .18864 L
s
.76832 .18864 m
.76538 .18943 L
s
.76538 .18943 m
.76753 .19158 L
s
.76753 .19158 m
.76674 .19452 L
s
.76674 .19452 m
.76968 .19374 L
s
.76968 .19374 m
.77183 .19589 L
s
.77183 .19589 m
.77262 .19295 L
s
.77262 .19295 m
.77556 .19216 L
s
.77556 .19216 m
.77771 .19431 L
s
.77771 .19431 m
.77693 .19725 L
s
.77693 .19725 m
.77987 .19646 L
s
.77987 .19646 m
.78202 .19862 L
s
.78202 .19862 m
.78123 .20156 L
s
.78123 .20156 m
.77829 .20234 L
s
.77829 .20234 m
.78044 .2045 L
s
.78044 .2045 m
.77965 .20743 L
s
.77965 .20743 m
.77671 .20822 L
s
.77671 .20822 m
.77456 .20607 L
s
.77456 .20607 m
.77378 .20901 L
s
.77378 .20901 m
.77084 .2098 L
s
.77084 .2098 m
.77299 .21195 L
s
.77299 .21195 m
.7722 .21489 L
s
.7722 .21489 m
.77514 .2141 L
s
.77514 .2141 m
.77729 .21625 L
s
.77729 .21625 m
.7765 .21919 L
s
.7765 .21919 m
.77356 .21998 L
s
.77356 .21998 m
.77572 .22213 L
s
.77572 .22213 m
.77493 .22507 L
s
.77493 .22507 m
.77787 .22428 L
s
.77787 .22428 m
.78002 .22644 L
s
.78002 .22644 m
.78081 .2235 L
s
.78081 .2235 m
.78375 .22271 L
s
.78375 .22271 m
.7859 .22486 L
s
.7859 .22486 m
.78511 .2278 L
s
.78511 .2278 m
.78805 .22701 L
s
.78805 .22701 m
.7902 .22916 L
s
.7902 .22916 m
.79099 .22622 L
s
.79099 .22622 m
.79393 .22544 L
s
.79393 .22544 m
.79178 .22329 L
s
.79178 .22329 m
.79257 .22035 L
s
.79257 .22035 m
.7955 .21956 L
s
.7955 .21956 m
.79766 .22171 L
s
.79766 .22171 m
.79844 .21877 L
s
.79844 .21877 m
.80138 .21798 L
s
.80138 .21798 m
.80354 .22013 L
s
.80354 .22013 m
.80275 .22307 L
s
.80275 .22307 m
.80569 .22229 L
s
.80569 .22229 m
.80784 .22444 L
s
.80784 .22444 m
.80705 .22738 L
s
.80705 .22738 m
.80411 .22817 L
s
.80411 .22817 m
.80626 .23032 L
s
.80626 .23032 m
.80548 .23326 L
s
.80548 .23326 m
.80842 .23247 L
s
.80842 .23247 m
.81057 .23462 L
s
.81057 .23462 m
.81136 .23168 L
s
.81136 .23168 m
.81429 .23089 L
s
.81429 .23089 m
.81645 .23305 L
s
.81645 .23305 m
.81566 .23598 L
s
.81566 .23598 m
.8186 .2352 L
s
.8186 .2352 m
.82075 .23735 L
s
.82075 .23735 m
.82154 .23441 L
s
.82154 .23441 m
.82448 .23362 L
s
.82448 .23362 m
.82233 .23147 L
s
.82233 .23147 m
.82311 .22853 L
s
.82311 .22853 m
.82605 .22774 L
s
.82605 .22774 m
.8282 .22989 L
s
.8282 .22989 m
.82899 .22696 L
s
.82899 .22696 m
.83193 .22617 L
s
.83193 .22617 m
.82978 .22402 L
s
.82978 .22402 m
.83057 .22108 L
s
.83057 .22108 m
.82763 .22186 L
s
.82763 .22186 m
.82548 .21971 L
s
.82548 .21971 m
.82626 .21677 L
s
.82626 .21677 m
.8292 .21599 L
s
.8292 .21599 m
.82705 .21383 L
s
.82705 .21383 m
.82784 .21089 L
s
.82784 .21089 m
.83078 .21011 L
s
.83078 .21011 m
.83293 .21226 L
s
.83293 .21226 m
.83372 .20932 L
s
.83372 .20932 m
.83666 .20853 L
s
.83666 .20853 m
.83881 .21068 L
s
.83881 .21068 m
.83802 .21362 L
s
.83802 .21362 m
.84096 .21283 L
s
.84096 .21283 m
.84311 .21499 L
s
.84311 .21499 m
.8439 .21205 L
s
.8439 .21205 m
.84684 .21126 L
s
.84684 .21126 m
.84469 .20911 L
s
.84469 .20911 m
.84548 .20617 L
s
.84548 .20617 m
.84841 .20538 L
s
.84841 .20538 m
.85057 .20753 L
s
.85057 .20753 m
.85135 .20459 L
s
.85135 .20459 m
.85429 .20381 L
s
.85429 .20381 m
.85645 .20596 L
s
.85645 .20596 m
.85566 .2089 L
s
.85566 .2089 m
.8586 .20811 L
s
.8586 .20811 m
.86075 .21026 L
s
.86075 .21026 m
.85996 .2132 L
s
.85996 .2132 m
.85702 .21399 L
s
.85702 .21399 m
.85917 .21614 L
s
.85917 .21614 m
.85839 .21908 L
s
.85839 .21908 m
.86133 .21829 L
s
.86133 .21829 m
.86348 .22044 L
s
.86348 .22044 m
.86427 .2175 L
s
.86427 .2175 m
.8672 .21672 L
s
.8672 .21672 m
.86936 .21887 L
s
.86936 .21887 m
.86857 .22181 L
s
.86857 .22181 m
.87151 .22102 L
s
.87151 .22102 m
.87366 .22317 L
s
.87366 .22317 m
.87287 .22611 L
s
.87287 .22611 m
.86993 .2269 L
s
.86993 .2269 m
.87208 .22905 L
s
.87208 .22905 m
.8713 .23199 L
s
.8713 .23199 m
.86836 .23278 L
s
.86836 .23278 m
.86621 .23063 L
s
.86621 .23063 m
.86542 .23357 L
s
.86542 .23357 m
.86248 .23435 L
s
.86248 .23435 m
.86463 .2365 L
s
.86463 .2365 m
.86384 .23944 L
s
.86384 .23944 m
.86678 .23866 L
s
.86678 .23866 m
.86893 .24081 L
s
.86893 .24081 m
.86815 .24375 L
s
.86815 .24375 m
.86521 .24454 L
s
.86521 .24454 m
.86736 .24669 L
s
.86736 .24669 m
.86657 .24963 L
s
.86657 .24963 m
.86951 .24884 L
s
.86951 .24884 m
.87166 .25099 L
s
.87166 .25099 m
.87245 .24805 L
s
.87245 .24805 m
.87539 .24726 L
s
.87539 .24726 m
.87754 .24942 L
s
.87754 .24942 m
.87675 .25236 L
s
.87675 .25236 m
.87969 .25157 L
s
.87969 .25157 m
.88185 .25372 L
s
.88185 .25372 m
.88263 .25078 L
s
.88263 .25078 m
.88557 .24999 L
s
.88557 .24999 m
.88342 .24784 L
s
.88342 .24784 m
.88421 .2449 L
s
.88421 .2449 m
.88715 .24411 L
s
.88715 .24411 m
.8893 .24627 L
s
.8893 .24627 m
.89009 .24333 L
s
.89009 .24333 m
.89303 .24254 L
s
.89303 .24254 m
.89518 .24469 L
s
.89518 .24469 m
.89439 .24763 L
s
.89439 .24763 m
.89733 .24684 L
s
.89733 .24684 m
.89948 .24899 L
s
.89948 .24899 m
.89869 .25193 L
s
.89869 .25193 m
.89576 .25272 L
s
.89576 .25272 m
.89791 .25487 L
s
.89791 .25487 m
.89712 .25781 L
s
.89712 .25781 m
.90006 .25702 L
s
.90006 .25702 m
.90221 .25918 L
s
.90221 .25918 m
.903 .25624 L
s
.903 .25624 m
.90594 .25545 L
s
.90594 .25545 m
.90809 .2576 L
s
.90809 .2576 m
.9073 .26054 L
s
.9073 .26054 m
.91024 .25975 L
s
.91024 .25975 m
.91239 .2619 L
s
.91239 .2619 m
.91318 .26484 L
s
.91318 .26484 m
.91103 .267 L
s
.91103 .267 m
.91397 .26778 L
s
.91397 .26778 m
.91476 .27072 L
s
.91476 .27072 m
.9126 .27287 L
s
.9126 .27287 m
.90966 .27209 L
s
.90966 .27209 m
.91045 .27503 L
s
.91045 .27503 m
.9083 .27718 L
s
.9083 .27718 m
.91124 .27797 L
s
.91124 .27797 m
.91203 .28091 L
s
.91203 .28091 m
.91418 .27875 L
s
.91418 .27875 m
.91712 .27954 L
s
.91712 .27954 m
.91791 .28248 L
s
.91791 .28248 m
.91575 .28463 L
s
.91575 .28463 m
.91869 .28542 L
s
.91869 .28542 m
.91948 .28836 L
s
.91948 .28836 m
.91733 .29051 L
s
.91733 .29051 m
.91439 .28972 L
s
.91439 .28972 m
.91518 .29266 L
s
.91518 .29266 m
.91303 .29482 L
s
.91303 .29482 m
.91009 .29403 L
s
.91009 .29403 m
.9093 .29109 L
s
.9093 .29109 m
.90715 .29324 L
s
.90715 .29324 m
.90421 .29245 L
s
.90421 .29245 m
.905 .29539 L
s
.905 .29539 m
.90284 .29754 L
s
.90284 .29754 m
.90578 .29833 L
s
.90578 .29833 m
.90657 .30127 L
s
.90657 .30127 m
.90442 .30342 L
s
.90442 .30342 m
.90148 .30263 L
s
.90148 .30263 m
.90227 .30557 L
s
.90227 .30557 m
.90012 .30773 L
s
.90012 .30773 m
.90305 .30851 L
s
.90305 .30851 m
.90384 .31145 L
s
.90384 .31145 m
.90599 .3093 L
s
.90599 .3093 m
.90893 .31009 L
s
.90893 .31009 m
.90972 .31303 L
s
.90972 .31303 m
.90757 .31518 L
s
.90757 .31518 m
.91051 .31597 L
s
.91051 .31597 m
.9113 .31891 L
s
.9113 .31891 m
.91345 .31676 L
s
.91345 .31676 m
.91639 .31754 L
s
.91639 .31754 m
.9156 .3146 L
s
.9156 .3146 m
.91775 .31245 L
s
.91775 .31245 m
.92069 .31324 L
s
.92069 .31324 m
.92148 .31618 L
s
.92148 .31618 m
.92363 .31403 L
s
.92363 .31403 m
.92657 .31481 L
s
.92657 .31481 m
.92736 .31775 L
s
.92736 .31775 m
.92521 .31991 L
s
.92521 .31991 m
.92815 .32069 L
s
.92815 .32069 m
.92893 .32363 L
s
.92893 .32363 m
.92678 .32578 L
s
.92678 .32578 m
.92384 .325 L
s
.92384 .325 m
.92463 .32794 L
s
.92463 .32794 m
.92248 .33009 L
s
.92248 .33009 m
.92542 .33088 L
s
.92542 .33088 m
.9262 .33382 L
s
.9262 .33382 m
.92836 .33166 L
s
.92836 .33166 m
.9313 .33245 L
s
.9313 .33245 m
.93208 .33539 L
s
.93208 .33539 m
.92993 .33754 L
s
.92993 .33754 m
.93287 .33833 L
s
.93287 .33833 m
.93366 .34127 L
s
.93366 .34127 m
.93151 .34342 L
s
.93151 .34342 m
.92857 .34263 L
s
.92857 .34263 m
.92936 .34557 L
s
.92936 .34557 m
.9272 .34773 L
s
.9272 .34773 m
.92426 .34694 L
s
.92426 .34694 m
.92348 .344 L
s
.92348 .344 m
.92132 .34615 L
s
.92132 .34615 m
.91839 .34536 L
s
.91839 .34536 m
.91917 .3483 L
s
.91917 .3483 m
.91702 .35045 L
s
.91702 .35045 m
.91996 .35124 L
s
.91996 .35124 m
.92075 .35418 L
s
.92075 .35418 m
.9186 .35633 L
s
.9186 .35633 m
.91566 .35554 L
s
.91566 .35554 m
.91644 .35848 L
s
.91644 .35848 m
.91429 .36064 L
s
.91429 .36064 m
.91135 .35985 L
s
.91135 .35985 m
.91057 .35691 L
s
.91057 .35691 m
.90841 .35906 L
s
.90841 .35906 m
.90547 .35827 L
s
.90547 .35827 m
.90469 .35533 L
s
.90469 .35533 m
.90684 .35318 L
s
.90684 .35318 m
.9039 .35239 L
s
.9039 .35239 m
.90311 .34946 L
s
.90311 .34946 m
.90096 .35161 L
s
.90096 .35161 m
.89802 .35082 L
s
.89802 .35082 m
.89881 .35376 L
s
.89881 .35376 m
.89666 .35591 L
s
.89666 .35591 m
.89372 .35512 L
s
.89372 .35512 m
.89293 .35218 L
s
.89293 .35218 m
.89078 .35434 L
s
.89078 .35434 m
.88784 .35355 L
s
.88784 .35355 m
.88863 .35649 L
s
.88863 .35649 m
.88647 .35864 L
s
.88647 .35864 m
.88941 .35943 L
s
.88941 .35943 m
.8902 .36237 L
s
.8902 .36237 m
.88805 .36452 L
s
.88805 .36452 m
.88511 .36373 L
s
.88511 .36373 m
.8859 .36667 L
s
.8859 .36667 m
.88374 .36882 L
s
.88374 .36882 m
.88668 .36961 L
s
.88668 .36961 m
.88747 .37255 L
s
.88747 .37255 m
.88962 .3704 L
s
.88962 .3704 m
.89256 .37118 L
s
.89256 .37118 m
.89335 .37412 L
s
.89335 .37412 m
.8912 .37628 L
s
.8912 .37628 m
.89414 .37706 L
s
.89414 .37706 m
.89493 .38 L
s
.89493 .38 m
.89277 .38215 L
s
.89277 .38215 m
.88983 .38137 L
s
.88983 .38137 m
.89062 .38431 L
s
.89062 .38431 m
.88847 .38646 L
s
.88847 .38646 m
.88553 .38567 L
s
.88553 .38567 m
.88474 .38273 L
s
.88474 .38273 m
.88259 .38488 L
s
.88259 .38488 m
.87965 .3841 L
s
.87965 .3841 m
.88044 .38703 L
s
.88044 .38703 m
.87829 .38919 L
s
.87829 .38919 m
.88123 .38997 L
s
.88123 .38997 m
.88202 .39291 L
s
.88202 .39291 m
.87986 .39507 L
s
.87986 .39507 m
.87692 .39428 L
s
.87692 .39428 m
.87771 .39722 L
s
.87771 .39722 m
.87556 .39937 L
s
.87556 .39937 m
.8785 .40016 L
s
.8785 .40016 m
.87929 .4031 L
s
.87929 .4031 m
.88144 .40094 L
s
.88144 .40094 m
.88438 .40173 L
s
.88438 .40173 m
.88517 .40467 L
s
.88517 .40467 m
.88301 .40682 L
s
.88301 .40682 m
.88595 .40761 L
s
.88595 .40761 m
.88674 .41055 L
s
.88674 .41055 m
.88889 .4084 L
s
.88889 .4084 m
.89183 .40919 L
s
.89183 .40919 m
.89104 .40625 L
s
.89104 .40625 m
.8932 .40409 L
s
.8932 .40409 m
.89614 .40488 L
s
.89614 .40488 m
.89692 .40782 L
s
.89692 .40782 m
.89908 .40567 L
s
.89908 .40567 m
.90201 .40646 L
s
.90201 .40646 m
.9028 .4094 L
s
.9028 .4094 m
.90065 .41155 L
s
.90065 .41155 m
.90359 .41234 L
s
.90359 .41234 m
.90438 .41528 L
s
.90438 .41528 m
.90223 .41743 L
s
.90223 .41743 m
.89929 .41664 L
s
.89929 .41664 m
.90007 .41958 L
s
.90007 .41958 m
.89792 .42173 L
s
.89792 .42173 m
.90086 .42252 L
s
.90086 .42252 m
.90165 .42546 L
s
.90165 .42546 m
.9038 .42331 L
s
.9038 .42331 m
.90674 .42409 L
s
.90674 .42409 m
.90753 .42703 L
s
.90753 .42703 m
.90538 .42919 L
s
.90538 .42919 m
.90832 .42997 L
s
.90832 .42997 m
.9091 .43291 L
s
.9091 .43291 m
.91126 .43076 L
s
.91126 .43076 m
.91419 .43155 L
s
.91419 .43155 m
.91341 .42861 L
s
.91341 .42861 m
.91556 .42646 L
s
.91556 .42646 m
.9185 .42724 L
s
.9185 .42724 m
.91929 .43018 L
s
.91929 .43018 m
.92144 .42803 L
s
.92144 .42803 m
.92438 .42882 L
s
.92438 .42882 m
.92359 .42588 L
s
.92359 .42588 m
.92574 .42373 L
s
.92574 .42373 m
.9228 .42294 L
s
.9228 .42294 m
.92201 .42 L
s
.92201 .42 m
.92417 .41785 L
s
.92417 .41785 m
.92711 .41864 L
s
.92711 .41864 m
.92632 .4157 L
s
.92632 .4157 m
.92847 .41355 L
s
.92847 .41355 m
.93141 .41433 L
s
.93141 .41433 m
.9322 .41727 L
s
.9322 .41727 m
.93435 .41512 L
s
.93435 .41512 m
.93729 .41591 L
s
.93729 .41591 m
.93808 .41885 L
s
.93808 .41885 m
.93592 .421 L
s
.93592 .421 m
.93886 .42179 L
s
.93886 .42179 m
.93965 .42473 L
s
.93965 .42473 m
.9418 .42258 L
s
.9418 .42258 m
.94474 .42336 L
s
.94474 .42336 m
.94395 .42042 L
s
.94395 .42042 m
.94611 .41827 L
s
.94611 .41827 m
.94905 .41906 L
s
.94905 .41906 m
.94983 .422 L
s
.94983 .422 m
.95199 .41985 L
s
.95199 .41985 m
.95492 .42063 L
s
.95492 .42063 m
.95571 .42357 L
s
.95571 .42357 m
.95356 .42573 L
s
.95356 .42573 m
.9565 .42651 L
s
.9565 .42651 m
.95729 .42945 L
s
.95729 .42945 m
.95514 .43161 L
s
.95514 .43161 m
.9522 .43082 L
s
.9522 .43082 m
.95298 .43376 L
s
.95298 .43376 m
.95083 .43591 L
s
.95083 .43591 m
.95377 .4367 L
s
.95377 .4367 m
.95456 .43964 L
s
.95456 .43964 m
.95671 .43748 L
s
.95671 .43748 m
.95965 .43827 L
s
.95965 .43827 m
.96044 .44121 L
s
.96044 .44121 m
.95829 .44336 L
s
.95829 .44336 m
.96123 .44415 L
s
.96123 .44415 m
.96201 .44709 L
s
.96201 .44709 m
.95986 .44924 L
s
.95986 .44924 m
.95692 .44845 L
s
.95692 .44845 m
.95771 .45139 L
s
.95771 .45139 m
.95556 .45355 L
s
.95556 .45355 m
.95262 .45276 L
s
.95262 .45276 m
.95183 .44982 L
s
.95183 .44982 m
.94968 .45197 L
s
.94968 .45197 m
.94674 .45118 L
s
.94674 .45118 m
.94753 .45412 L
s
.94753 .45412 m
.94538 .45627 L
s
.94538 .45627 m
.94831 .45706 L
s
.94831 .45706 m
.9491 .46 L
s
.9491 .46 m
.94695 .46215 L
s
.94695 .46215 m
.94401 .46137 L
s
.94401 .46137 m
.9448 .4643 L
s
.9448 .4643 m
.94265 .46646 L
s
.94265 .46646 m
.94559 .46724 L
s
.94559 .46724 m
.94637 .47018 L
s
.94637 .47018 m
.94853 .46803 L
s
.94853 .46803 m
.95147 .46882 L
s
.95147 .46882 m
.95225 .47176 L
s
.95225 .47176 m
.9501 .47391 L
s
.9501 .47391 m
.95304 .4747 L
s
.95304 .4747 m
.95383 .47764 L
s
.95383 .47764 m
.95598 .47549 L
s
.95598 .47549 m
.95892 .47627 L
s
.95892 .47627 m
.95813 .47333 L
s
.95813 .47333 m
.96028 .47118 L
s
.96028 .47118 m
.96322 .47197 L
s
.96322 .47197 m
.96401 .47491 L
s
.96401 .47491 m
.96616 .47276 L
s
.96616 .47276 m
.9691 .47354 L
s
.9691 .47354 m
.96989 .47648 L
s
.96989 .47648 m
.96774 .47864 L
s
.96774 .47864 m
.97068 .47942 L
s
.97068 .47942 m
.97146 .48236 L
s
.97146 .48236 m
.96931 .48452 L
s
.96931 .48452 m
.96637 .48373 L
s
.96637 .48373 m
.96716 .48667 L
s
.96716 .48667 m
.96501 .48882 L
s
.96501 .48882 m
.96795 .48961 L
s
.96795 .48961 m
.96874 .49255 L
s
.96874 .49255 m
.97089 .49039 L
s
.97089 .49039 m
.97383 .49118 L
s
.97383 .49118 m
.97462 .49412 L
s
.97462 .49412 m
.97246 .49627 L
s
.97246 .49627 m
.9754 .49706 L
s
.9754 .49706 m
.97619 .5 L
s
.97619 .5 m
.9754 .50294 L
s
.9754 .50294 m
.97246 .50373 L
s
.97246 .50373 m
.97462 .50588 L
s
.97462 .50588 m
.97383 .50882 L
s
.97383 .50882 m
.97089 .50961 L
s
.97089 .50961 m
.96874 .50745 L
s
.96874 .50745 m
.96795 .51039 L
s
.96795 .51039 m
.96501 .51118 L
s
.96501 .51118 m
.96716 .51333 L
s
.96716 .51333 m
.96637 .51627 L
s
.96637 .51627 m
.96931 .51548 L
s
.96931 .51548 m
.97146 .51764 L
s
.97146 .51764 m
.97068 .52058 L
s
.97068 .52058 m
.96774 .52136 L
s
.96774 .52136 m
.96989 .52352 L
s
.96989 .52352 m
.9691 .52646 L
s
.9691 .52646 m
.96616 .52724 L
s
.96616 .52724 m
.96401 .52509 L
s
.96401 .52509 m
.96322 .52803 L
s
.96322 .52803 m
.96028 .52882 L
s
.96028 .52882 m
.95813 .52667 L
s
.95813 .52667 m
.95892 .52373 L
s
.95892 .52373 m
.95598 .52451 L
s
.95598 .52451 m
.95383 .52236 L
s
.95383 .52236 m
.95304 .5253 L
s
.95304 .5253 m
.9501 .52609 L
s
.9501 .52609 m
.95225 .52824 L
s
.95225 .52824 m
.95147 .53118 L
s
.95147 .53118 m
.94853 .53197 L
s
.94853 .53197 m
.94637 .52982 L
s
.94637 .52982 m
.94559 .53276 L
s
.94559 .53276 m
.94265 .53354 L
s
.94265 .53354 m
.9448 .5357 L
s
.9448 .5357 m
.94401 .53863 L
s
.94401 .53863 m
.94695 .53785 L
s
.94695 .53785 m
.9491 .54 L
s
.9491 .54 m
.94831 .54294 L
s
.94831 .54294 m
.94538 .54373 L
s
.94538 .54373 m
.94753 .54588 L
s
.94753 .54588 m
.94674 .54882 L
s
.94674 .54882 m
.94968 .54803 L
s
.94968 .54803 m
.95183 .55018 L
s
.95183 .55018 m
.95262 .54724 L
s
.95262 .54724 m
.95556 .54645 L
s
.95556 .54645 m
.95771 .54861 L
s
.95771 .54861 m
.95692 .55155 L
s
.95692 .55155 m
.95986 .55076 L
s
.95986 .55076 m
.96201 .55291 L
s
.96201 .55291 m
.96123 .55585 L
s
.96123 .55585 m
.95829 .55664 L
s
.95829 .55664 m
.96044 .55879 L
s
.96044 .55879 m
.95965 .56173 L
s
.95965 .56173 m
.95671 .56252 L
s
.95671 .56252 m
.95456 .56036 L
s
.95456 .56036 m
.95377 .5633 L
s
.95377 .5633 m
.95083 .56409 L
s
.95083 .56409 m
.95298 .56624 L
s
.95298 .56624 m
.9522 .56918 L
s
.9522 .56918 m
.95514 .56839 L
s
.95514 .56839 m
.95729 .57055 L
s
.95729 .57055 m
.9565 .57349 L
s
.9565 .57349 m
.95356 .57427 L
s
.95356 .57427 m
.95571 .57643 L
s
.95571 .57643 m
.95492 .57937 L
s
.95492 .57937 m
.95199 .58015 L
s
.95199 .58015 m
.94983 .578 L
s
.94983 .578 m
.94905 .58094 L
s
.94905 .58094 m
.94611 .58173 L
s
.94611 .58173 m
.94395 .57958 L
s
.94395 .57958 m
.94474 .57664 L
s
.94474 .57664 m
.9418 .57742 L
s
.9418 .57742 m
.93965 .57527 L
s
.93965 .57527 m
.93886 .57821 L
s
.93886 .57821 m
.93592 .579 L
s
.93592 .579 m
.93808 .58115 L
s
.93808 .58115 m
.93729 .58409 L
s
.93729 .58409 m
.93435 .58488 L
s
.93435 .58488 m
.9322 .58273 L
s
.9322 .58273 m
.93141 .58567 L
s
.93141 .58567 m
.92847 .58645 L
s
.92847 .58645 m
.92632 .5843 L
s
.92632 .5843 m
.92711 .58136 L
s
.92711 .58136 m
.92417 .58215 L
s
.92417 .58215 m
.92201 .58 L
s
.92201 .58 m
.9228 .57706 L
s
.9228 .57706 m
.92574 .57627 L
s
.92574 .57627 m
.92359 .57412 L
s
.92359 .57412 m
.92438 .57118 L
s
.92438 .57118 m
.92144 .57197 L
s
.92144 .57197 m
.91929 .56982 L
s
.91929 .56982 m
.9185 .57276 L
s
.9185 .57276 m
.91556 .57354 L
s
.91556 .57354 m
.91341 .57139 L
s
.91341 .57139 m
.91419 .56845 L
s
.91419 .56845 m
.91126 .56924 L
s
.91126 .56924 m
.9091 .56709 L
s
.9091 .56709 m
.90832 .57003 L
s
.90832 .57003 m
.90538 .57081 L
s
.90538 .57081 m
.90753 .57297 L
s
.90753 .57297 m
.90674 .57591 L
s
.90674 .57591 m
.9038 .57669 L
s
.9038 .57669 m
.90165 .57454 L
s
.90165 .57454 m
.90086 .57748 L
s
.90086 .57748 m
.89792 .57827 L
s
.89792 .57827 m
.90007 .58042 L
s
.90007 .58042 m
.89929 .58336 L
s
.89929 .58336 m
.90223 .58257 L
s
.90223 .58257 m
.90438 .58472 L
s
.90438 .58472 m
.90359 .58766 L
s
.90359 .58766 m
.90065 .58845 L
s
.90065 .58845 m
.9028 .5906 L
s
.9028 .5906 m
.90201 .59354 L
s
.90201 .59354 m
.89908 .59433 L
s
.89908 .59433 m
.89692 .59218 L
s
.89692 .59218 m
.89614 .59512 L
s
.89614 .59512 m
.8932 .59591 L
s
.8932 .59591 m
.89104 .59375 L
s
.89104 .59375 m
.89183 .59081 L
s
.89183 .59081 m
.88889 .5916 L
s
.88889 .5916 m
.88674 .58945 L
s
.88674 .58945 m
.88595 .59239 L
s
.88595 .59239 m
.88301 .59318 L
s
.88301 .59318 m
.88517 .59533 L
s
.88517 .59533 m
.88438 .59827 L
s
.88438 .59827 m
.88144 .59906 L
s
.88144 .59906 m
.87929 .5969 L
s
.87929 .5969 m
.8785 .59984 L
s
.8785 .59984 m
.87556 .60063 L
s
.87556 .60063 m
.87771 .60278 L
s
.87771 .60278 m
.87692 .60572 L
s
.87692 .60572 m
.87986 .60493 L
s
.87986 .60493 m
.88202 .60709 L
s
.88202 .60709 m
.88123 .61003 L
s
.88123 .61003 m
.87829 .61081 L
s
.87829 .61081 m
.88044 .61297 L
s
.88044 .61297 m
.87965 .6159 L
s
.87965 .6159 m
.88259 .61512 L
s
.88259 .61512 m
.88474 .61727 L
s
.88474 .61727 m
.88553 .61433 L
s
.88553 .61433 m
.88847 .61354 L
s
.88847 .61354 m
.89062 .61569 L
s
.89062 .61569 m
.88983 .61863 L
s
.88983 .61863 m
.89277 .61785 L
s
.89277 .61785 m
.89493 .62 L
s
.89493 .62 m
.89414 .62294 L
s
.89414 .62294 m
.8912 .62372 L
s
.8912 .62372 m
.89335 .62588 L
s
.89335 .62588 m
.89256 .62882 L
s
.89256 .62882 m
.88962 .6296 L
s
.88962 .6296 m
.88747 .62745 L
s
.88747 .62745 m
.88668 .63039 L
s
.88668 .63039 m
.88374 .63118 L
s
.88374 .63118 m
.8859 .63333 L
s
.8859 .63333 m
.88511 .63627 L
s
.88511 .63627 m
.88805 .63548 L
s
.88805 .63548 m
.8902 .63763 L
s
.8902 .63763 m
.88941 .64057 L
s
.88941 .64057 m
.88647 .64136 L
s
.88647 .64136 m
.88863 .64351 L
s
.88863 .64351 m
.88784 .64645 L
s
.88784 .64645 m
.89078 .64566 L
s
.89078 .64566 m
.89293 .64782 L
s
.89293 .64782 m
.89372 .64488 L
s
.89372 .64488 m
.89666 .64409 L
s
.89666 .64409 m
.89881 .64624 L
s
.89881 .64624 m
.89802 .64918 L
s
.89802 .64918 m
.90096 .64839 L
s
.90096 .64839 m
.90311 .65054 L
s
.90311 .65054 m
.9039 .64761 L
s
.9039 .64761 m
.90684 .64682 L
s
.90684 .64682 m
.90469 .64467 L
s
.90469 .64467 m
.90547 .64173 L
s
.90547 .64173 m
.90841 .64094 L
s
.90841 .64094 m
.91057 .64309 L
s
.91057 .64309 m
.91135 .64015 L
s
.91135 .64015 m
.91429 .63936 L
s
.91429 .63936 m
.91644 .64152 L
s
.91644 .64152 m
.91566 .64446 L
s
.91566 .64446 m
.9186 .64367 L
s
.9186 .64367 m
.92075 .64582 L
s
.92075 .64582 m
.91996 .64876 L
s
.91996 .64876 m
.91702 .64955 L
s
.91702 .64955 m
.91917 .6517 L
s
.91917 .6517 m
.91839 .65464 L
s
.91839 .65464 m
.92132 .65385 L
s
.92132 .65385 m
.92348 .656 L
s
.92348 .656 m
.92426 .65306 L
s
.92426 .65306 m
.9272 .65227 L
s
.9272 .65227 m
.92936 .65443 L
s
.92936 .65443 m
.92857 .65737 L
s
.92857 .65737 m
.93151 .65658 L
s
.93151 .65658 m
.93366 .65873 L
s
.93366 .65873 m
.93287 .66167 L
s
.93287 .66167 m
.92993 .66246 L
s
.92993 .66246 m
.93208 .66461 L
s
.93208 .66461 m
.9313 .66755 L
s
.9313 .66755 m
.92836 .66834 L
s
.92836 .66834 m
.9262 .66618 L
s
.9262 .66618 m
.92542 .66912 L
s
.92542 .66912 m
.92248 .66991 L
s
.92248 .66991 m
.92463 .67206 L
s
.92463 .67206 m
.92384 .675 L
s
.92384 .675 m
.92678 .67422 L
s
.92678 .67422 m
.92893 .67637 L
s
.92893 .67637 m
.92815 .67931 L
s
.92815 .67931 m
.92521 .68009 L
s
.92521 .68009 m
.92736 .68225 L
s
.92736 .68225 m
.92657 .68519 L
s
.92657 .68519 m
.92363 .68597 L
s
.92363 .68597 m
.92148 .68382 L
s
.92148 .68382 m
.92069 .68676 L
s
.92069 .68676 m
.91775 .68755 L
s
.91775 .68755 m
.9156 .6854 L
s
.9156 .6854 m
.91639 .68246 L
s
.91639 .68246 m
.91345 .68324 L
s
.91345 .68324 m
.9113 .68109 L
s
.9113 .68109 m
.91051 .68403 L
s
.91051 .68403 m
.90757 .68482 L
s
.90757 .68482 m
.90972 .68697 L
s
.90972 .68697 m
.90893 .68991 L
s
.90893 .68991 m
.90599 .6907 L
s
.90599 .6907 m
.90384 .68855 L
s
.90384 .68855 m
.90305 .69149 L
s
.90305 .69149 m
.90012 .69227 L
s
.90012 .69227 m
.90227 .69443 L
s
.90227 .69443 m
.90148 .69737 L
s
.90148 .69737 m
.90442 .69658 L
s
.90442 .69658 m
.90657 .69873 L
s
.90657 .69873 m
.90578 .70167 L
s
.90578 .70167 m
.90284 .70246 L
s
.90284 .70246 m
.905 .70461 L
s
.905 .70461 m
.90421 .70755 L
s
.90421 .70755 m
.90715 .70676 L
s
.90715 .70676 m
.9093 .70891 L
s
.9093 .70891 m
.91009 .70597 L
s
.91009 .70597 m
.91303 .70518 L
s
.91303 .70518 m
.91518 .70734 L
s
.91518 .70734 m
.91439 .71028 L
s
.91439 .71028 m
.91733 .70949 L
s
.91733 .70949 m
.91948 .71164 L
s
.91948 .71164 m
.91869 .71458 L
s
.91869 .71458 m
.91575 .71537 L
s
.91575 .71537 m
.91791 .71752 L
s
.91791 .71752 m
.91712 .72046 L
s
.91712 .72046 m
.91418 .72125 L
s
.91418 .72125 m
.91203 .71909 L
s
.91203 .71909 m
.91124 .72203 L
s
.91124 .72203 m
.9083 .72282 L
s
.9083 .72282 m
.91045 .72497 L
s
.91045 .72497 m
.90966 .72791 L
s
.90966 .72791 m
.9126 .72713 L
s
.9126 .72713 m
.91476 .72928 L
s
.91476 .72928 m
.91397 .73222 L
s
.91397 .73222 m
.91103 .733 L
s
.91103 .733 m
.91318 .73516 L
s
.91318 .73516 m
.91239 .7381 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00e03oool00`000000oooo00000002000000030?ooo`000000oooo0=@0oooo003C0?ooo`03
0000003oool0000000<0oooo00<000000?ooo`000000e03oool00<d0oooo0`0000030?ooo`030000
003oool0oooo00D0oooo00@000000?ooo`3oool0oooo0`00003>0?ooo`00c03oool01@000000oooo
0000003oool000000080oooo0P0000050?ooo`8000000P3oool01@000000oooo0000003oool00000
0<d0oooo003;0?ooo`030000003oool0oooo0080oooo0`0000080?ooo`040000003oool000000000
00@0oooo00<000000?ooo`3oool0bP3oool00</0oooo0P0000050?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool00`3oool200000<`0oooo002o0?ooo`8000002P3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00P0oooo0P0000300?ooo`00_03oool4000000030?ooo`00
0000oooo00P0oooo00<000000?ooo`0000004`3oool00`000000oooo000000090?ooo`030000003o
ool0000000<00000_@3oool00;L0oooo0`0000020?ooo`030000003oool0oooo0080oooo0P000009
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool01`3oool2000000@0oooo00<00000
0?ooo`3oool00`00002h0?ooo`00]@3oool300000080oooo00@000000?ooo`00000000001@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`030?ooo`80000000@0oooo0000003oool0oooo0`00002f0?ooo`00]03oool0
1@000000oooo0?ooo`000000oooo00<000001@3oool3000000H0oooo0`00000E0?ooo`<000001P3o
ool3000000@0oooo00H000000?ooo`00000000000?ooo`0000020?ooo`030000003oool0oooo0;<0
oooo002d0?ooo`8000001@3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`00
00000P3oool010000000oooo0?ooo`00000E0?ooo`030000003oool0000000<0oooo00D000000?oo
o`3oool0oooo000000060?ooo`030000003oool0oooo00<0oooo0P00002e0?ooo`00]03oool00`00
0000oooo0?ooo`0;0?ooo`@0000000@0oooo0000003oool000006`3oool010000000oooo0000003o
ool4000000d0oooo00<000000?ooo`3oool0/`3oool00;@0oooo00<000000?ooo`0000003@3oool0
10000000oooo0000003oool3000001T0oooo100000020?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`000000]@3oool00;D0oooo0P00000?0?ooo`80000000<0oooo0000003oool06P3oool0
10000000oooo00000000000?0?ooo`800000]P3oool00;H0oooo00<000000?ooo`3oool0?`3oool0
0`000000oooo0?ooo`2e0?ooo`00]03oool4000003l0oooo1000002e0?ooo`00/`3oool00`000000
oooo000000130?ooo`030000003oool000000;@0oooo002A0?ooo`<000007P3oool00`000000oooo
0?ooo`170?ooo`030000003oool0oooo01`0oooo0`00002B0?ooo`00S`3oool300000080oooo00<0
00000?ooo`3oool06`3oool2000004L0oooo0P00000M0?ooo`030000003oool0oooo00<00000T03o
ool008X0oooo0P0000020?ooo`050000003oool0oooo0000003oool00P00000M0?ooo`030000003o
ool0oooo04L0oooo00<000000?ooo`3oool06`3oool2000000050?ooo`000000oooo0?ooo`000000
0P3oool2000008/0oooo00270?ooo`<000000P3oool010000000oooo0?ooo`0000050?ooo`030000
003oool0oooo01/0oooo0P0000150?ooo`8000007@3oool00`000000oooo0?ooo`030?ooo`070000
003oool0oooo0000003oool000000?ooo`02000008P0oooo00270?ooo`030000003oool000000080
oooo0`0000050?ooo`8000007P3oool00`000000oooo00000002000003d0oooo0`0000000`3oool0
00000?ooo`0M0?ooo`8000001@3oool300000080oooo00<000000?ooo`000000R03oool008P0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01d0oooo
0P0000000`3oool000000?ooo`0l0?ooo`040000003oool00000000001l0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo08L0oooo001l0?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool02P3oool00`000000oooo0000000200000200oooo
00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0N0?ooo`@000003@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo07/0oooo001h0?ooo`80000000@0oooo0000003oool00000
203oool00`000000oooo0?ooo`0;0?ooo`80000000@0oooo0000003oool000007P3oool00`000000
oooo0?ooo`0k0?ooo`030000003oool0oooo01`0oooo00@000000?ooo`000000oooo0P00000=0?oo
o`030000003oool0oooo00H0oooo00@000000?ooo`000000oooo0P00001i0?ooo`00M03oool02000
0000oooo0?ooo`3oool000000?ooo`000000oooo0`0000080?ooo`800000403oool3000001/0oooo
00@000000?ooo`3oool00000>P3oool00`000000oooo0000000K0?ooo`<00000403oool2000000P0
oooo0`000000103oool000000?ooo`0000030?ooo`030000003oool0oooo07<0oooo001`0?ooo`@0
000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0J0?ooo`800000>`3oool2000001`0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0
oooo00H000000?ooo`3oool0oooo0000003oool400000740oooo001_0?ooo`060000003oool0oooo
0?ooo`000000oooo100000050?ooo`030000003oool0oooo00D0oooo00@000000?ooo`0000000000
3`3oool3000001d0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0K0?ooo`<00000
3`3oool2000000030?ooo`000000oooo00H0oooo00<000000?ooo`3oool00`3oool4000000030?oo
o`000000oooo0080oooo00<000000?ooo`3oool0KP3oool006l0oooo0P0000060?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0000000B0?ooo`030000003o
ool0oooo01/0oooo0`00000i0?ooo`<000007@3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0000000L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`800000L03o
ool006h0oooo00<000000?ooo`3oool03@3oool4000000<0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0000000K0?ooo`<00000?@3oool3000001/0oooo00<000000?ooo`0000004P3oool0
10000000oooo0?ooo`3oool4000000l0oooo00<000000?ooo`3oool0K@3oool006h0oooo00<00000
0?ooo`3oool03@3oool01P000000oooo0000003oool000000?ooo`<000003`3oool2000000030?oo
o`000000oooo01D0oooo0`0000030?ooo`030000003oool0oooo03l0oooo00@000000?ooo`3oool0
oooo0`00000F0?ooo`040000003oool00000000000l0oooo0`0000001P3oool000000?ooo`000000
oooo000000l0oooo00<000000?ooo`3oool0K@3oool006l0oooo00<000000?ooo`3oool0403oool3
00000180oooo00<000000?ooo`3oool05@3oool01@000000oooo0000003oool000000080oooo0P00
000o0?ooo`8000000P3oool01@000000oooo0000003oool0000001L0oooo00<000000?ooo`3oool0
403oool300000180oooo00<000000?ooo`3oool0KP3oool006/0oooo0P0000000`3oool000000?oo
o`0D0?ooo`030000003oool0oooo0100oooo0P00000E0?ooo`030000003oool0oooo0080oooo0`00
00130?ooo`<00000103oool00`000000oooo0?ooo`0C0?ooo`8000004P3oool00`000000oooo0?oo
o`0C0?ooo`040000003oool00000000006`0oooo001Z0?ooo`030000003oool0000000<00000:@3o
ool00`000000oooo0?ooo`0B0?ooo`800000C`3oool2000001@0oooo00<000000?ooo`3oool09`3o
ool4000000030?ooo`000000oooo06X0oooo001Y0?ooo`800000;P3oool00`000000oooo0?ooo`05
0?ooo`<000002P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00P0oooo0`000007
0?ooo`030000003oool0oooo02`0oooo0P00001Z0?ooo`00J03oool00`000000oooo0?ooo`0/0?oo
o`8000001P3oool2000000040?ooo`000000oooo000000T0oooo0`00001=0?ooo`<000002@3oool0
10000000oooo0000003oool2000000H0oooo0P00000^0?ooo`030000003oool0oooo06L0oooo001X
0?ooo`<00000;@3oool00`000000oooo0000000200000080oooo00<000000?ooo`3oool00P3oool2
000000X0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`080?ooo`800000103oool0
0`000000oooo0?ooo`03000000030?ooo`000000oooo02`0oooo0`00001Y0?ooo`00J@3oool00`00
0000oooo0?ooo`0/0?ooo`<000000P3oool010000000oooo0000000000050?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00<0oooo0P000000103oool000000?ooo`3oool3000002h0oooo00<000000?ooo`3oool0J03o
ool006T0oooo00<000000?ooo`000000<03oool3000000D0oooo0`0000060?ooo`<00000C`3oool3
000000H0oooo0`0000040?ooo`@00000<03oool00`000000oooo0000001Z0?ooo`00JP3oool01000
0000oooo00000000000`0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo0000
00020?ooo`040000003oool0oooo000004l0oooo00<000000?ooo`0000000`3oool01@000000oooo
0?ooo`3oool0000000H0oooo00<000000?ooo`3oool0;P3oool2000000030?ooo`000000oooo06X0
oooo001/0?ooo`030000003oool0oooo03H0oooo0P0000001P3oool000000?ooo`000000oooo0000
05D0oooo00H000000?ooo`000000oooo0000003oool2000003P0oooo00<000000?ooo`3oool0J`3o
ool006/0oooo0P00000j0?ooo`040000003oool000000?ooo`<00000D`3oool400000080oooo00<0
00000?ooo`3oool0>03oool2000006`0oooo001Z0?ooo`030000003oool0000003`0oooo0P000000
0`3oool000000?ooo`1D0?ooo`040000003oool00000000003`0oooo00<000000?ooo`000000J`3o
ool00600oooo0P0000080?ooo`030000003oool0oooo0=L0oooo00<000000?ooo`3oool01P3oool2
00000640oooo001O0?ooo`030000003oool00000008000001`3oool200000=D0oooo0P0000070?oo
o`<0000000<0oooo0000003oool0G`3oool005h0oooo0P0000001@3oool000000?ooo`3oool00000
0080oooo0`0000000`3oool000000?ooo`3F0?ooo`030000003oool00000008000000P3oool01@00
0000oooo0?ooo`000000oooo00800000G`3oool005d0oooo00<000000?ooo`3oool00`3oool01000
0000oooo0?ooo`0000020?ooo`<00000e`3oool300000080oooo00@000000?ooo`3oool000001@3o
ool00`000000oooo0?ooo`1L0?ooo`00G@3oool3000000@0oooo100000000`3oool000000?ooo`3J
0?ooo`030000003oool0000000<00000103oool3000005h0oooo001N0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
05d0oooo001I0?ooo`<0000000<0oooo000000000000l@3oool2000000040?ooo`00000000000000
05X0oooo001H0?ooo`8000000P3oool00`000000oooo0000003a0?ooo`070000003oool000000?oo
o`000000oooo0000001I0?ooo`00E`3oool010000000oooo00000000003i0?ooo`80000000<0oooo
0000003oool0E`3oool005L0oooo00<000000?ooo`3oool0o03oool00`000000oooo0?ooo`1G0?oo
o`00F03oool200000?/0oooo0P00001I0?ooo`00F03oool00`000000oooo0?ooo`3k0?ooo`030000
003oool0oooo05L0oooo001H0?ooo`@00000m`3oool4000005T0oooo001K0?ooo`800000m@3oool2
000005`0oooo001K0?ooo`030000003oool0oooo0?D0oooo00<000000?ooo`3oool0FP3oool005X0
oooo0`00003e0?ooo`<00000F`3oool005T0oooo00<000000?ooo`3oool0n@3oool00`000000oooo
0?ooo`1H0?ooo`00FP3oool00`000000oooo0?ooo`3g0?ooo`030000003oool0oooo05T0oooo001J
0?ooo`8000001@3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo00<0oooo0P00001K
0?ooo`00FP3oool01@000000oooo0?ooo`000000oooo00<00000j`3oool3000000050?ooo`000000
oooo0?ooo`000000F`3oool005X0oooo10000000103oool000000?ooo`3oool200000>L0oooo0P00
00020?ooo`030000003oool0000000<00000F`3oool005h0oooo00L000000?ooo`3oool0oooo0000
003oool000000>D0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`1M0?ooo`00HP3o
ool2000000030?ooo`000000oooo0>80oooo00@000000?ooo`0000000000H`3oool006H0oooo00<0
00000?ooo`3oool0g`3oool00`000000oooo0?ooo`1U0?ooo`00I@3oool00`000000oooo0?ooo`3Q
0?ooo`030000003oool0oooo06@0oooo001T0?ooo`800000h`3oool2000006D0oooo001S0?ooo`03
0000003oool0oooo0>D0oooo00<000000?ooo`3oool0HP3oool006<0oooo0`00003S0?ooo`<00000
I03oool00600oooo00D000000?ooo`3oool0oooo0000003U0?ooo`050000003oool0oooo0?ooo`00
0000H@3oool005l0oooo00<000000?ooo`0000000`00003U0?ooo`@0000000<0oooo0000003oool0
G`3oool005h0oooo0P0000020?ooo`030000003oool0oooo0>L0oooo00D000000?ooo`000000oooo
0000001O0?ooo`00G@3oool010000000oooo00000000003]0?ooo`80000000<0oooo0000003oool0
G@3oool005h0oooo00<000000?ooo`3oool0k`3oool00`000000oooo0?ooo`1M0?ooo`00?P3oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0>d0oooo00<000000?ooo`3oool07P3oool0
0`000000oooo0?ooo`0m0?ooo`00?@3oool00`000000oooo00000003000001/0oooo00@000000?oo
o`3oool00000j`3oool010000000oooo0?ooo`00000K0?ooo`@0000000<0oooo0000003oool0?@3o
ool003`0oooo0P0000020?ooo`030000003oool0000000<0oooo00<000000?ooo`3oool05@3oool4
00000>/0oooo1000000G0?ooo`0;0000003oool0oooo0?ooo`000000oooo0000003oool000000?oo
o`000000?@3oool003`0oooo0`0000030?ooo`050000003oool0oooo0000003oool01000000F0?oo
o`800000j@3oool2000001H0oooo100000001@3oool000000?ooo`3oool0000000<0oooo0`00000m
0?ooo`00?03oool00`000000oooo0?ooo`030?ooo`@000000P3oool00`000000oooo0000000F0?oo
o`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0503oool00`000000oooo000000020?oo
o`@000001@3oool00`000000oooo0?ooo`0k0?ooo`00?@3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0000000@0oooo00<000000?ooo`3oool04`3oool300000>T0oooo0`00000E0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`0000001@3oool00`000000oooo0?ooo`0l0?ooo`00
>@3oool2000000030?ooo`000000oooo00/0oooo0P00000:0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0k@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo
0P00000<0?ooo`040000003oool00000000003X0oooo000h0?ooo`030000003oool0000000<00000
3@3oool00`000000oooo0?ooo`060?ooo`040000003oool000000?ooo`8000001P3oool200000>/0
oooo0P0000060?ooo`80000000@0oooo0000003oool00000203oool00`000000oooo0?ooo`0;0?oo
o`@0000000<0oooo0000003oool0>03oool003L0oooo0P0000000`3oool000000?ooo`0=0?ooo`80
0000203oool200000080oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00003]0?oo
o`040000003oool0oooo000000<0oooo00L000000?ooo`000000oooo0000003oool0000000P0oooo
0P00000>0?ooo`040000003oool00000000003P0oooo000f0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool01P3oool010000000oooo0000000000030?ooo`050000003oool0oooo0000
003oool00`00003/0?ooo`060000003oool000000000003oool000000P3oool010000000oooo0?oo
o`3oool2000000030?ooo`000000oooo00L0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`0e0?ooo`00=P3oool3000000l0oooo0P0000080?ooo`030000003oool0oooo00<0oooo1@00
00000`3oool000000?ooo`0200000>/0oooo0P0000000`3oool000000?ooo`05000000@0oooo00<0
00000?ooo`3oool01`3oool2000000l0oooo0`00000g0?ooo`00=`3oool00`000000oooo0?ooo`0?
0?ooo`@000001P3oool2000000D0oooo00<000000?ooo`000000m@3oool00`000000oooo00000005
0?ooo`8000001P3oool300000180oooo00<000000?ooo`3oool0=P3oool003L0oooo00<000000?oo
o`000000403oool00`000000oooo000000030?ooo`80000000<0oooo0000003oool0o`3oool70?oo
o`040000003oool00000000000<0oooo00<000000?ooo`000000403oool00`000000oooo0000000h
0?ooo`00>03oool00`000000oooo0000000@0?ooo`060000003oool0oooo0?ooo`000000oooo0`00
003o0?ooo`P0oooo0`0000000`3oool000000?ooo`020?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`000000>@3oool003X0oooo0P00000@0?ooo`060000003oool000000000003oool00000
o`3oool<0?ooo`060000003oool000000000003oool00000403oool2000003/0oooo000i0?ooo`80
00004P3oool00`000000oooo0?ooo`3o0?oooa80oooo00<000000?ooo`3oool0403oool2000003X0
oooo000h0?ooo`030000003oool000000?l0oooo>P3oool00`000000oooo0000000i0?ooo`00;P3o
ool2000000P0oooo00<000000?ooo`3oool0o`3oooll0?ooo`030000003oool0oooo00H0oooo0P00
000_0?ooo`00;@3oool00`000000oooo00000002000000H0oooo0`00003o0?ooocX0oooo0`000006
0?ooo`<0000000<0oooo0000003oool0;@3oool002`0oooo0P0000001@3oool000000?ooo`3oool0
00000080oooo0P0000020?ooo`030000003oool0oooo0?l0oooo>P3oool00`000000oooo0?ooo`02
00000080oooo00D000000?ooo`3oool000000?ooo`02000002d0oooo000[0?ooo`030000003oool0
oooo00<0oooo00L000000?ooo`3oool000000?ooo`000000oooo00800000o`3oooll0?ooo`<00000
0P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo02X0oooo000[0?ooo`800000
1@3oool4000000030?ooo`000000oooo0?l0oooo?`3oool00`000000oooo00000003000000D0oooo
0P00000/0?ooo`00;03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0ooooB03o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02/0oooo000X0?ooo`80000000<0oooo
000000000000o`3ooomF0?ooo`80000000<0oooo000000000000:@3oool002H0oooo0P0000020?oo
o`030000003oool000000?l0ooooEP3oool00`000000oooo000000020?ooo`8000009`3oool002D0
oooo00<000000?ooo`000000o`3ooomP0?ooo`030000003oool0000002H0oooo000U0?ooo`030000
003oool0oooo0?l0ooooHP3oool00`000000oooo0?ooo`0T0?ooo`009@3oool00`000000oooo0?oo
o`3o0?ooof80oooo00<000000?ooo`3oool0903oool002D0oooo0P00003o0?ooof80oooo0P00000V
0?ooo`00903oool00`000000oooo0?ooo`3o0?ooof@0oooo00<000000?ooo`3oool08`3oool002D0
oooo0P00003o0?ooof80oooo0P00000V0?ooo`009P3oool00`000000oooo0?ooo`3o0?ooof00oooo
00<000000?ooo`3oool09@3oool002H0oooo0P00003o0?ooof00oooo0P00000W0?ooo`008`3oool3
00000?l0ooooH`3oool010000000oooo00000000000T0?ooo`008`3oool00`000000oooo0000003o
0?ooof@0oooo00<000000?ooo`000000903oool002@0oooo00<000000?ooo`3oool0o`3ooomT0?oo
o`030000003oool0oooo02<0oooo000S0?ooo`030000003oool0oooo0?l0ooooIP3oool00`000000
oooo0?ooo`0R0?ooo`008P3oool00`000000oooo0000003o0?ooofH0oooo00<000000?ooo`000000
8`3oool002<0oooo0P0000020?ooo`030000003oool0oooo0?l0ooooGP3oool00`000000oooo0?oo
o`02000002@0oooo000T0?ooo`040000003oool000000?ooo`800000o`3ooomL0?ooo`80000000@0
oooo0000003oool000009@3oool002D0oooo00@000000?ooo`3oool00000o`3ooomN0?ooo`040000
003oool0oooo000002H0oooo000X0?ooo`800000o`3ooomL0?ooo`800000:@3oool002P0oooo00<0
00000?ooo`3oool0o`3ooomL0?ooo`030000003oool0oooo02L0oooo000Y0?ooo`800000o`3ooomJ
0?ooo`800000:P3oool002X0oooo00<000000?ooo`3oool0o`3ooomH0?ooo`030000003oool0oooo
02T0oooo000Y0?ooo`<00000o`3ooomH0?ooo`<00000:P3oool002L0oooo0`00003o0?oooe`0oooo
0`00000X0?ooo`008`3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0?l0oooo
FP3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0280oooo000O0?ooo`800000
00@0oooo0000003oool000000P3oool00`000000oooo0?ooo`3o0?oooeh0oooo00L000000?ooo`3o
ool000000?ooo`000000oooo00800000803oool001l0oooo00<000000?ooo`0000000P3oool30000
0?l0ooooHP3oool300000080oooo00<000000?ooo`000000803oool00200oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`0O0?ooo`007P3oool200000?l0ooooL03oool2000001l0oooo000N0?ooo`030000003o
ool0oooo0?l0ooooL03oool00`000000oooo0?ooo`0M0?ooo`007`3oool200000?l0ooooKP3oool2
00000200oooo000P0?ooo`030000003oool0oooo0?l0ooooK03oool00`000000oooo0?ooo`0O0?oo
o`007P3oool010000000oooo00000000003o0?ooof`0oooo0P0000000`3oool000000?ooo`0N0?oo
o`007@3oool00`000000oooo0000003o0?ooog00oooo00<000000?ooo`0000007P3oool001d0oooo
00<000000?ooo`3oool0o`3ooomb0?ooo`030000003oool0oooo01`0oooo000L0?ooo`<00000o`3o
oomb0?ooo`<000007@3oool001`0oooo00<000000?ooo`3oool0o`3ooomd0?ooo`030000003oool0
oooo01/0oooo000L0?ooo`030000003oool000000?l0ooooLP3oool00`000000oooo0000000M0?oo
o`007@3oool2000000050?ooo`000000oooo0?ooo`000000o`3ooomX0?ooo`050000003oool0oooo
0000003oool00P00000N0?ooo`007`3oool400000?l0ooooJ@3oool500000200oooo000R0?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool0o`3ooomJ0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool08@3oool00280oooo0P0000050?ooo`<000000P3oool00`000000oooo
0?ooo`3o0?oooe00oooo00<000000?ooo`3oool00`0000050?ooo`8000008`3oool00280oooo00<0
00000?ooo`3oool00`3oool01`000000oooo0?ooo`000000oooo0000003oool00P00003o0?ooodh0
oooo0`0000020?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool08@3oool002<0
oooo0P0000001@3oool000000?ooo`3oool000000080oooo0P0000020?ooo`030000003oool0oooo
0?l0ooooC03oool00`000000oooo0?ooo`0200000080oooo00D000000?ooo`3oool000000?ooo`02
000002@0oooo000T0?ooo`030000003oool00000008000001P3oool200000?l0ooooCP3oool20000
00H0oooo0`0000000`3oool000000?ooo`0T0?ooo`00903oool3000000P0oooo00<000000?ooo`3o
ool0o`3ooom>0?ooo`030000003oool0oooo00H0oooo0`00000U0?ooo`00;`3oool00`000000oooo
0000003o0?oood`0oooo00<000000?ooo`000000<03oool00300oooo0P00003o0?oood`0oooo0P00
000a0?ooo`00<@3oool200000?l0ooooBP3oool200000380oooo000^0?ooo`@00000o`3ooom<0?oo
o`@00000;`3oool002d0oooo00@000000?ooo`3oool00000o`3ooom>0?ooo`030000003oool00000
02l0oooo000]0?ooo`030000003oool0oooo0?l0ooooDP3oool00`000000oooo0?ooo`0/0?ooo`00
;@3oool200000?l0ooooDP3oool2000002h0oooo000]0?ooo`030000003oool0oooo0?l0ooooDP3o
ool00`000000oooo0?ooo`0/0?ooo`00;P3oool2000000030?ooo`000000oooo0?l0ooooB`3oool0
10000000oooo00000000000_0?ooo`00;`3oool00`000000oooo0000000200000?l0ooooB03oool3
000000030?ooo`000000oooo02l0oooo000`0?ooo`80000000<0oooo0000003oool0o`3ooom70?oo
o`040000003oool0000000000340oooo000d0?ooo`030000003oool0oooo0?l0ooooA03oool00`00
0000oooo0?ooo`0c0?ooo`00<`3oool00`000000oooo0?ooo`3o0?ooodH0oooo00<000000?ooo`3o
ool0<P3oool00380oooo00@000000?ooo`3oool00000o`3ooom40?ooo`040000003oool0oooo0000
03<0oooo000c0?ooo`800000o`3ooom50?ooo`<00000=03oool003@0oooo00<000000?ooo`3oool0
o`3ooom40?ooo`030000003oool0oooo03<0oooo000c0?ooo`<00000o`3ooom40?ooo`<00000=03o
ool00340oooo0P0000000`3oool000000?ooo`3o0?ooodD0oooo00@000000?ooo`0000000000<P3o
ool002/0oooo100000020?ooo`030000003oool0oooo0?l0ooooBP3oool00`000000oooo0?ooo`04
000002`0oooo000Z0?ooo`050000003oool000000?ooo`0000000P3oool200000?l0ooooBP3oool2
00000080oooo00D000000?ooo`000000oooo0000000[0?ooo`00:@3oool00`000000oooo0?ooo`02
0?ooo`<00000o`3ooom>0?ooo`<00000103oool00`000000oooo0?ooo`0X0?ooo`00:@3oool20000
0?l0ooooFP3oool2000002X0oooo000L0?ooo`<000002P3oool00`000000oooo0?ooo`3o0?oooeX0
oooo00<000000?ooo`3oool0203oool3000001d0oooo000K0?ooo`050000003oool000000?ooo`00
00002P3oool200000?l0ooooF03oool2000000X0oooo00D000000?ooo`000000oooo0000000L0?oo
o`00503oool400000080oooo00<000000?ooo`3oool00P3oool2000000X0oooo00<000000?ooo`3o
ool0o`3ooomF0?ooo`030000003oool0oooo00P0oooo0P0000040?ooo`030000003oool0oooo00@0
00005@3oool001<0oooo00<000000?ooo`0000000P3oool010000000oooo0000000000050?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooomE0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool00`3oool2000000070?ooo`000000oooo0?ooo`000000oooo0000000D
0?ooo`004`3oool010000000oooo0?ooo`3oool4000000D0oooo0P0000060?ooo`@00000o`3ooomI
0?ooo`<000001P3oool2000000D0oooo100000030?ooo`030000003oool0oooo0180oooo000B0?oo
o`<000001@3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`0000000P3oool0
10000000oooo0?ooo`00003o0?oooeX0oooo00<000000?ooo`0000000`3oool01@000000oooo0?oo
o`3oool0000000D0oooo00<000000?ooo`3oool00`3oool3000001<0oooo000B0?ooo`030000003o
ool0oooo00/0oooo0P0000001@3oool000000?ooo`000000oooo00800000o`3ooomN0?ooo`800000
00D0oooo0000003oool000000?ooo`02000000d0oooo00<000000?ooo`3oool04@3oool00180oooo
00<000000?ooo`0000003@3oool00`000000oooo0000000300000?l0ooooH03oool4000000030?oo
o`000000oooo00`0oooo00<000000?ooo`0000004`3oool001<0oooo0P00000B0?ooo`030000003o
ool0oooo0?l0ooooGP3oool00`000000oooo0?ooo`0@0?ooo`800000503oool001D0oooo00<00000
0?ooo`3oool0o`3ooon20?ooo`030000003oool0oooo01@0oooo000A0?ooo`@00000o`3ooon60?oo
o`@000004P3oool00140oooo00<000000?ooo`000000o`3ooon80?ooo`030000003oool000000180
oooo000A0?ooo`030000003oool0oooo0?l0ooooRP3oool00`000000oooo0?ooo`0@0?ooo`00403o
ool200000?l0ooooS03oool200000140oooo000@0?ooo`030000003oool0oooo0?l0ooooS03oool0
0`000000oooo0?ooo`0?0?ooo`004@3oool2000000030?ooo`000000oooo0?l0ooooQ@3oool01000
0000oooo00000000000B0?ooo`004P3oool00`000000oooo0000000200000?l0ooooPP3oool30000
00030?ooo`000000oooo0180oooo000C0?ooo`80000000<0oooo0000003oool0o`3ooon10?ooo`04
0000003oool00000000001@0oooo000F0?ooo`800000o`3ooon00?ooo`8000005`3oool001D0oooo
00<000000?ooo`3oool0o`3ooon10?ooo`030000003oool0oooo01D0oooo000E0?ooo`040000003o
ool0oooo00000?l0ooooOP3oool010000000oooo0?ooo`00000F0?ooo`005P3oool200000?l0oooo
O`3oool3000001L0oooo000G0?ooo`030000003oool0oooo0?l0ooooOP3oool00`000000oooo0?oo
o`0F0?ooo`005P3oool300000?l0ooooOP3oool3000001L0oooo000D0?ooo`80000000<0oooo0000
003oool0o`3ooomo0?ooo`040000003oool00000000001D0oooo000>0?ooo`@000000P3oool00`00
0000oooo0?ooo`3o0?oooh@0oooo00<000000?ooo`3oool01000000?0?ooo`003@3oool00`000000
oooo000000020?ooo`040000003oool0000000000?l0ooooQ03oool2000000070?ooo`000000oooo
0?ooo`000000oooo0000000>0?ooo`00303oool00`000000oooo0?ooo`020?ooo`030000003oool0
00000?l0ooooR03oool00`000000oooo000000040?ooo`030000003oool0oooo00/0oooo000<0?oo
o`800000o`3ooonD0?ooo`8000003@3oool000`0oooo00<000000?ooo`3oool0o`3ooonD0?ooo`03
0000003oool0oooo00/0oooo000=0?ooo`800000o`3ooonB0?ooo`8000003P3oool000h0oooo00<0
00000?ooo`3oool0o`3ooon@0?ooo`030000003oool0oooo00d0oooo000?0?ooo`030000003oool0
oooo0?l0ooooSP3oool00`000000oooo0?ooo`0>0?ooo`002`3oool400000?l0ooooTP3oool40000
00`0oooo000:0?ooo`040000003oool0oooo00000?l0ooooU03oool010000000oooo0?ooo`00000;
0?ooo`002`3oool00`000000oooo0?ooo`3o0?oooiH0oooo00<000000?ooo`3oool02P3oool000X0
oooo00<000000?ooo`3oool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000
003oool0oooo0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?oo
o`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool0o`3ooonF
0?ooo`030000003oool0oooo00X0oooo000:0?ooo`040000003oool0oooo00000?l0ooooU03oool0
10000000oooo0?ooo`00000;0?ooo`002`3oool300000?l0ooooU03oool3000000`0oooo000>0?oo
o`800000o`3ooon@0?ooo`8000003`3oool000h0oooo00<000000?ooo`3oool0o`3ooon@0?ooo`03
0000003oool0oooo00d0oooo000=0?ooo`800000o`3ooonB0?ooo`8000003P3oool000`0oooo00<0
00000?ooo`3oool0o`3ooonD0?ooo`030000003oool0oooo00/0oooo000<0?ooo`800000o`3ooonD
0?ooo`8000003@3oool000`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0000003o0?oo
ohP0oooo00<000000?ooo`000000103oool00`000000oooo0?ooo`0;0?ooo`003@3oool00`000000
oooo000000020?ooo`040000003oool0000000000?l0ooooQ03oool2000000070?ooo`000000oooo
0?ooo`000000oooo0000000>0?ooo`003P3oool400000080oooo00<000000?ooo`3oool0o`3ooon4
0?ooo`030000003oool0oooo00@000003`3oool001@0oooo00@000000?ooo`3oool00000o`3ooon0
0?ooo`040000003oool0oooo000001D0oooo000E0?ooo`@00000o`3ooomn0?ooo`@000005P3oool0
01L0oooo00<000000?ooo`3oool0o`3ooomn0?ooo`030000003oool0oooo01H0oooo000F0?ooo`80
0000o`3ooomo0?ooo`030000003oool0000001L0oooo000E0?ooo`040000003oool0000000000?l0
ooooOP3oool2000000030?ooo`000000oooo01D0oooo000E0?ooo`030000003oool0oooo0?l0oooo
P@3oool00`000000oooo0?ooo`0E0?ooo`005P3oool200000?l0ooooP03oool2000001L0oooo000C
0?ooo`80000000<0oooo0000003oool0o`3ooon10?ooo`040000003oool00000000001@0oooo000B
0?ooo`030000003oool0000000800000o`3ooon20?ooo`<0000000<0oooo0000003oool04P3oool0
0140oooo0P0000000`3oool000000?ooo`3o0?ooohD0oooo00@000000?ooo`00000000004P3oool0
0100oooo00<000000?ooo`3oool0o`3ooon<0?ooo`030000003oool0oooo00l0oooo000@0?ooo`80
0000o`3ooon<0?ooo`8000004@3oool00140oooo00<000000?ooo`3oool0o`3ooon:0?ooo`030000
003oool0oooo0100oooo000A0?ooo`030000003oool000000?l0ooooR03oool00`000000oooo0000
000B0?ooo`004@3oool300000?l0ooooR03oool300000180oooo000D0?ooo`800000o`3ooon40?oo
o`8000005@3oool001<0oooo0P00000?0?ooo`040000003oool0oooo00000?l0ooooH03oool01000
0000oooo0?ooo`00000?0?ooo`800000503oool00180oooo00<000000?ooo`0000003@3oool00`00
0000oooo0000000300000?l0ooooH03oool300000080oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0000000C0?ooo`004P3oool00`000000oooo0?ooo`0;0?ooo`80000000D0oooo0000003o
ool000000?ooo`0200000?l0ooooGP3oool2000000050?ooo`000000oooo0000003oool00P00000=
0?ooo`030000003oool0oooo0140oooo000B0?ooo`<000001@3oool00`000000oooo0?ooo`030?oo
o`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`00003o0?oooeX0oooo00<0
00000?ooo`0000000`3oool01@000000oooo0?ooo`3oool0000000D0oooo00<000000?ooo`3oool0
0`3oool3000001<0oooo000C0?ooo`040000003oool0oooo0?ooo`@000001@3oool2000000H0oooo
0`00003o0?oooeX0oooo0`0000060?ooo`8000001@3oool4000000<0oooo00<000000?ooo`3oool0
4P3oool001<0oooo00<000000?ooo`0000000P3oool010000000oooo0000000000050?ooo`030000
003oool0oooo00L0oooo0P00003o0?oooeL0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`030?ooo`80000000L0oooo0000003oool0oooo0000003oool0000001@0oooo000D0?ooo`@0
00000P3oool00`000000oooo0?ooo`020?ooo`8000002P3oool00`000000oooo0?ooo`3o0?oooeH0
oooo00<000000?ooo`3oool0203oool2000000@0oooo00<000000?ooo`3oool01000000E0?ooo`00
6`3oool01@000000oooo0000003oool0000000X0oooo0P00003o0?oooeP0oooo0P00000:0?ooo`05
0000003oool000000?ooo`000000703oool001`0oooo0`00000:0?ooo`030000003oool0oooo0?l0
ooooFP3oool00`000000oooo0?ooo`080?ooo`<000007@3oool002T0oooo0P00003o0?oooeX0oooo
0P00000Z0?ooo`00:@3oool00`000000oooo0?ooo`020?ooo`<00000o`3ooom>0?ooo`<00000103o
ool00`000000oooo0?ooo`0X0?ooo`00:P3oool01@000000oooo0000003oool000000080oooo0P00
003o0?ooodX0oooo0P0000020?ooo`050000003oool000000?ooo`000000:`3oool002/0oooo1000
00020?ooo`030000003oool0oooo0?l0ooooBP3oool00`000000oooo0?ooo`04000002`0oooo000a
0?ooo`040000003oool0oooo00000?l0ooooAP3oool010000000oooo0?ooo`00000b0?ooo`00<P3o
ool400000?l0ooooA03oool4000003<0oooo000d0?ooo`030000003oool0oooo0?l0ooooA03oool0
0`000000oooo0?ooo`0c0?ooo`00<`3oool200000?l0ooooA@3oool00`000000oooo0000000d0?oo
o`00<P3oool010000000oooo00000000003o0?oood@0oooo0P0000000`3oool000000?ooo`0b0?oo
o`00<`3oool00`000000oooo0?ooo`3o0?ooodH0oooo00<000000?ooo`3oool0<P3oool003@0oooo
00<000000?ooo`3oool0o`3ooom40?ooo`030000003oool0oooo03<0oooo000`0?ooo`80000000<0
oooo0000003oool0o`3ooom70?ooo`040000003oool0000000000340oooo000_0?ooo`030000003o
ool0000000800000o`3ooom80?ooo`<0000000<0oooo0000003oool0;`3oool002h0oooo0P000000
0`3oool000000?ooo`3o0?oood/0oooo00@000000?ooo`0000000000;`3oool002d0oooo00<00000
0?ooo`3oool0o`3ooomB0?ooo`030000003oool0oooo02`0oooo000]0?ooo`800000o`3ooomB0?oo
o`800000;P3oool002d0oooo00<000000?ooo`3oool0o`3ooomB0?ooo`030000003oool0oooo02`0
oooo000]0?ooo`040000003oool0oooo00000?l0ooooCP3oool00`000000oooo0000000_0?ooo`00
;P3oool300000?l0ooooCP3oool3000002l0oooo000a0?ooo`800000o`3ooom:0?ooo`800000<P3o
ool00300oooo0P00003o0?oood`0oooo0P00000a0?ooo`00;`3oool00`000000oooo0000003o0?oo
od`0oooo00<000000?ooo`000000<03oool002@0oooo0`0000080?ooo`030000003oool0oooo0?l0
ooooCP3oool00`000000oooo0?ooo`060?ooo`<000009@3oool002@0oooo00<000000?ooo`000000
0P0000060?ooo`800000o`3ooom>0?ooo`8000001P3oool3000000030?ooo`000000oooo02@0oooo
000S0?ooo`80000000D0oooo0000003oool0oooo000000020?ooo`8000000P3oool00`000000oooo
0?ooo`3o0?oood`0oooo00<000000?ooo`3oool00P0000020?ooo`050000003oool0oooo0000003o
ool00P00000T0?ooo`008P3oool00`000000oooo0?ooo`030?ooo`070000003oool0oooo0000003o
ool000000?ooo`0200000?l0ooooCP3oool300000080oooo00@000000?ooo`3oool000001@3oool0
0`000000oooo0?ooo`0Q0?ooo`008P3oool2000000D0oooo0`0000020?ooo`030000003oool0oooo
0?l0ooooD03oool00`000000oooo0?ooo`03000000D0oooo0P00000S0?ooo`008P3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0?l0ooooFP3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0240oooo000O0?ooo`@00000o`3ooomY0?ooo`030000003oool0000000800000
803oool001d0oooo0P0000000`3oool000000?ooo`0200000?l0ooooJ03oool2000000030?ooo`00
0000oooo008000007P3oool001`0oooo00<000000?ooo`000000o`3ooomb0?ooo`030000003oool0
000001d0oooo000L0?ooo`030000003oool0oooo0?l0ooooM03oool00`000000oooo0?ooo`0K0?oo
o`00703oool300000?l0ooooLP3oool3000001d0oooo000M0?ooo`030000003oool0oooo0?l0oooo
LP3oool00`000000oooo0?ooo`0L0?ooo`007@3oool00`000000oooo0000003o0?ooog00oooo00<0
00000?ooo`0000007P3oool001h0oooo00@000000?ooo`0000000000o`3ooom/0?ooo`80000000<0
oooo0000003oool07P3oool00200oooo00<000000?ooo`3oool0o`3ooom/0?ooo`030000003oool0
oooo01l0oooo000O0?ooo`800000o`3ooom^0?ooo`800000803oool001h0oooo00<000000?ooo`3o
ool0o`3ooom`0?ooo`030000003oool0oooo01d0oooo000N0?ooo`030000003oool0oooo0?l0oooo
L03oool00`000000oooo0?ooo`0M0?ooo`007`3oool2000000D0oooo00<000000?ooo`3oool0o`3o
oomP0?ooo`030000003oool0oooo00<0oooo0P00000P0?ooo`007`3oool00`000000oooo00000002
0?ooo`<00000o`3ooomR0?ooo`<000000P3oool00`000000oooo0000000P0?ooo`007`3oool20000
00040?ooo`000000oooo00000080oooo00<000000?ooo`3oool0o`3ooomN0?ooo`070000003oool0
oooo0000003oool000000?ooo`0200000200oooo000S0?ooo`040000003oool0oooo00000080oooo
00<000000?ooo`3oool0o`3ooomJ0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3o
ool08P3oool002L0oooo0`00003o0?oooe`0oooo0`00000X0?ooo`00:@3oool300000?l0ooooF03o
ool3000002X0oooo000Z0?ooo`030000003oool0oooo0?l0ooooF03oool00`000000oooo0?ooo`0Y
0?ooo`00:@3oool200000?l0ooooFP3oool2000002X0oooo000X0?ooo`030000003oool0oooo0?l0
ooooG03oool00`000000oooo0?ooo`0W0?ooo`00:03oool200000?l0ooooG03oool2000002T0oooo
000U0?ooo`040000003oool0oooo00000?l0ooooGP3oool010000000oooo0?ooo`00000V0?ooo`00
903oool010000000oooo0000003oool200000?l0ooooG03oool2000000040?ooo`000000oooo0000
02D0oooo000S0?ooo`8000000P3oool00`000000oooo0?ooo`3o0?oooeh0oooo00<000000?ooo`3o
ool00P00000T0?ooo`008P3oool00`000000oooo0000003o0?ooofH0oooo00<000000?ooo`000000
8`3oool002<0oooo00<000000?ooo`3oool0o`3ooomV0?ooo`030000003oool0oooo0280oooo000T
0?ooo`030000003oool0oooo0?l0ooooI03oool00`000000oooo0?ooo`0S0?ooo`008`3oool00`00
0000oooo0000003o0?ooof@0oooo00<000000?ooo`000000903oool002<0oooo0`00003o0?ooof<0
oooo00@000000?ooo`0000000000903oool002H0oooo0P00003o0?ooof00oooo0P00000W0?ooo`00
9P3oool00`000000oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool09@3oool002D0oooo0P00
003o0?ooof80oooo0P00000V0?ooo`00903oool00`000000oooo0?ooo`3o0?ooof@0oooo00<00000
0?ooo`3oool08`3oool002D0oooo0P00003o0?ooof80oooo0P00000V0?ooo`009@3oool00`000000
oooo0?ooo`3o0?ooof80oooo00<000000?ooo`3oool0903oool002D0oooo00<000000?ooo`3oool0
o`3ooomR0?ooo`030000003oool0oooo02@0oooo000U0?ooo`030000003oool000000?l0ooooH03o
ool00`000000oooo0000000V0?ooo`009P3oool200000080oooo00<000000?ooo`000000o`3ooomF
0?ooo`030000003oool000000080oooo0P00000W0?ooo`00:03oool2000000030?ooo`0000000000
0?l0ooooEP3oool2000000030?ooo`000000000002T0oooo000/0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0o`3ooom80?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0:`3oool002/0oooo0P0000050?ooo`@0000000<0oooo0000003oool0o`3ooolo0?ooo`030000
003oool0000000<000001@3oool2000002`0oooo000[0?ooo`030000003oool0oooo00<0oooo00L0
00000?ooo`3oool000000?ooo`000000oooo00800000o`3oooll0?ooo`<000000P3oool010000000
oooo0?ooo`0000050?ooo`030000003oool0oooo02X0oooo000/0?ooo`80000000D0oooo0000003o
ool0oooo000000020?ooo`8000000P3oool00`000000oooo0?ooo`3o0?ooocX0oooo00<000000?oo
o`3oool00P0000020?ooo`050000003oool0oooo0000003oool00P00000]0?ooo`00;@3oool00`00
0000oooo00000002000000H0oooo0`00003o0?ooocX0oooo0`0000060?ooo`<0000000<0oooo0000
003oool0;@3oool002h0oooo0P0000080?ooo`030000003oool0oooo0?l0oooo?03oool00`000000
oooo0?ooo`060?ooo`800000;`3oool003P0oooo00<000000?ooo`000000o`3ooolj0?ooo`030000
003oool0000003T0oooo000i0?ooo`8000004P3oool00`000000oooo0?ooo`3o0?oooa80oooo00<0
00000?ooo`3oool0403oool2000003X0oooo000j0?ooo`800000403oool01P000000oooo00000000
0000oooo00000?l0oooo303oool01P000000oooo000000000000oooo00000100oooo0P00000k0?oo
o`00>03oool00`000000oooo0000000@0?ooo`060000003oool0oooo0?ooo`000000oooo0`00003o
0?ooo`P0oooo0`0000000`3oool000000?ooo`020?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`000000>@3oool003L0oooo00<000000?ooo`000000403oool00`000000oooo000000030?oo
o`80000000<0oooo0000003oool0o`3oool70?ooo`040000003oool00000000000<0oooo00<00000
0?ooo`000000403oool00`000000oooo0000000h0?ooo`00=`3oool00`000000oooo0?ooo`0?0?oo
o`@000001P3oool2000000D0oooo00<000000?ooo`000000m@3oool00`000000oooo000000050?oo
o`8000001P3oool300000180oooo00<000000?ooo`3oool0=P3oool003H0oooo0`00000?0?ooo`80
0000203oool00`000000oooo0?ooo`030?ooo`D0000000D0oooo0000003oool0oooo0000003[0?oo
o`050000003oool0oooo0000003oool01@0000040?ooo`030000003oool0oooo00L0oooo0P00000?
0?ooo`<00000=`3oool003H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`060?oo
o`040000003oool0oooo000000<0oooo00D000000?ooo`3oool000000?ooo`0300000>`0oooo1000
00001@3oool000000?ooo`3oool0000000<0oooo00@000000?ooo`3oool00000203oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo03D0oooo000g0?ooo`80000000<0oooo0000003oool0
3@3oool2000000P0oooo0P0000020?ooo`030000003oool0000000<0oooo00@000000?ooo`3oool0
0000k@3oool010000000oooo0?ooo`0000030?ooo`040000003oool000000?ooo`<00000203oool2
000000h0oooo00@000000?ooo`0000000000>03oool003P0oooo00<000000?ooo`0000000`00000=
0?ooo`030000003oool0oooo00H0oooo00@000000?ooo`000000oooo0P0000060?ooo`800000j`3o
ool2000000H0oooo0P000000103oool000000?ooo`0000080?ooo`030000003oool0oooo00/0oooo
100000000`3oool000000?ooo`0h0?ooo`00>@3oool2000000030?ooo`000000oooo00/0oooo0P00
000:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0k@3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo00P0oooo0P00000<0?ooo`040000003oool00000000003X0oooo
000m0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`000000103oool00`000000oooo0?oo
o`0C0?ooo`<00000j@3oool3000001D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0000
00050?ooo`030000003oool0oooo03`0oooo000l0?ooo`030000003oool0oooo00<0oooo10000002
0?ooo`030000003oool0000001H0oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool000000080oooo100000050?ooo`030000003oool0oooo03/0oooo000l0?oo
o`030000003oool0000000<0oooo00D000000?ooo`3oool000000?ooo`04000001H0oooo0P00003Y
0?ooo`8000005P3oool4000000050?ooo`000000oooo0?ooo`0000000`3oool00`000000oooo0000
000m0?ooo`00?03oool200000080oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`0E
0?ooo`@00000j`3oool4000001L0oooo00P000000?ooo`3oool0oooo0000003oool000000?ooo`<0
0000?@3oool003d0oooo00<000000?ooo`3oool00`00000K0?ooo`040000003oool0oooo00000>/0
oooo00@000000?ooo`3oool000006`3oool300000080oooo00<000000?ooo`3oool0?03oool003h0
oooo0P00000O0?ooo`030000003oool0oooo0>d0oooo00<000000?ooo`3oool07@3oool2000003l0
oooo001N0?ooo`030000003oool0oooo0>l0oooo00<000000?ooo`3oool0G@3oool005d0oooo00@0
00000?ooo`3oool00000k@3oool010000000oooo0?ooo`00001N0?ooo`00GP3oool200000080oooo
00<000000?ooo`3oool0i`3oool00`000000oooo00000002000005l0oooo001O0?ooo`030000003o
ool0oooo00<00000i@3oool300000080oooo00<000000?ooo`3oool0GP3oool00600oooo0P000002
0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool00P00001Q0?ooo`00H`3oool30000
0><0oooo0`00001T0?ooo`00H`3oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo0680
oooo001T0?ooo`800000h`3oool2000006D0oooo001U0?ooo`030000003oool0oooo0>40oooo00<0
00000?ooo`3oool0I03oool006H0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`1U
0?ooo`00HP3oool2000000030?ooo`000000oooo0>80oooo00@000000?ooo`0000000000H`3oool0
05h0oooo00L000000?ooo`3oool0oooo0000003oool000000>D0oooo00<000000?ooo`0000000`3o
ool00`000000oooo0?ooo`1M0?ooo`00FP3oool4000000040?ooo`000000oooo0?ooo`800000i`3o
ool200000080oooo00<000000?ooo`0000000`00001K0?ooo`00FP3oool01@000000oooo0?ooo`00
0000oooo00@00000j@3oool4000000050?ooo`000000oooo0?ooo`000000F`3oool005X0oooo0P00
00050?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool00`3oool2000005/0oooo001J
0?ooo`030000003oool0oooo0?L0oooo00<000000?ooo`3oool0F@3oool005T0oooo00<000000?oo
o`3oool0n@3oool00`000000oooo0?ooo`1H0?ooo`00FP3oool300000?D0oooo0`00001K0?ooo`00
F`3oool00`000000oooo0?ooo`3e0?ooo`030000003oool0oooo05X0oooo001K0?ooo`800000m@3o
ool2000005`0oooo001H0?ooo`@00000m`3oool4000005T0oooo001H0?ooo`030000003oool0oooo
0?/0oooo00<000000?ooo`3oool0E`3oool005P0oooo0P00003k0?ooo`800000F@3oool005L0oooo
00<000000?ooo`3oool0o03oool00`000000oooo0?ooo`1G0?ooo`00E`3oool010000000oooo0?oo
o`00003i0?ooo`040000003oool0oooo000005P0oooo001H0?ooo`8000000P3oool00`000000oooo
0000003a0?ooo`040000003oool000000?ooo`<00000F@3oool005T0oooo0`0000000`3oool00000
0000003a0?ooo`80000000@0oooo0000000000000000FP3oool005h0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
G@3oool005d0oooo0`0000040?ooo`@0000000<0oooo0000003oool0fP3oool00`000000oooo0000
0003000000@0oooo0`00001N0?ooo`00G@3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo00000080oooo0`00003G0?ooo`<000000P3oool010000000oooo0?ooo`0000050?ooo`030000
003oool0oooo05`0oooo001N0?ooo`80000000D0oooo0000003oool0oooo000000020?ooo`<00000
00<0oooo0000003oool0eP3oool00`000000oooo0000000200000080oooo00D000000?ooo`3oool0
00000?ooo`02000005l0oooo001O0?ooo`030000003oool0000000800000203oool00`000000oooo
0?ooo`3C0?ooo`030000003oool0oooo00H0oooo0`0000000`3oool000000?ooo`1O0?ooo`00H03o
ool2000000P0oooo0P00003G0?ooo`800000203oool200000640oooo001Z0?ooo`030000003oool0
000003`0oooo00@000000?ooo`3oool00000E@3oool010000000oooo0?ooo`00000l0?ooo`030000
003oool0000006/0oooo001[0?ooo`800000>P3oool00`000000oooo00000004000005<0oooo1000
00020?ooo`030000003oool0oooo03P0oooo0P00001/0?ooo`00K03oool00`000000oooo0?ooo`0f
0?ooo`80000000H0oooo0000003oool000000?ooo`00001E0?ooo`060000003oool000000?ooo`00
0000oooo0P00000h0?ooo`030000003oool0oooo06/0oooo001Z0?ooo`040000003oool000000000
0300oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool000000080oooo00@00000
0?ooo`3oool00000C`3oool00`000000oooo000000030?ooo`050000003oool0oooo0?ooo`000000
1P3oool00`000000oooo0?ooo`0^0?ooo`80000000<0oooo0000003oool0JP3oool006T0oooo00<0
00000?ooo`000000<03oool3000000D0oooo0`0000060?ooo`<00000C`3oool3000000H0oooo0`00
00040?ooo`@00000<03oool00`000000oooo0000001Z0?ooo`00J@3oool00`000000oooo0?ooo`0/
0?ooo`<000000P3oool010000000oooo0000000000050?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0B`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo0P00
0000103oool000000?ooo`3oool3000002h0oooo00<000000?ooo`3oool0J03oool006P0oooo0`00
000]0?ooo`030000003oool00000008000000P3oool00`000000oooo0?ooo`020?ooo`8000002P3o
ool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00P0oooo0P0000040?ooo`030000003o
ool0oooo00<0000000<0oooo0000003oool0;03oool3000006T0oooo001X0?ooo`030000003oool0
oooo02`0oooo00<000000?ooo`3oool01@3oool01P000000oooo0?ooo`000000oooo000000T0oooo
0`00001=0?ooo`<000002@3oool00`000000oooo000000020?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1W0?ooo`00J@3oool2000002d0oooo0P00
00060?ooo`@000002P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00P0oooo1000
00060?ooo`800000;@3oool2000006X0oooo001Z0?ooo`030000003oool0000000<00000:@3oool0
0`000000oooo0?ooo`0B0?ooo`8000001@3oool00`000000oooo0?ooo`110?ooo`030000003oool0
oooo00<0oooo0P00000D0?ooo`030000003oool0oooo02L0oooo100000000`3oool000000?ooo`1Z
0?ooo`00J`3oool2000000030?ooo`000000oooo01@0oooo00<000000?ooo`3oool0403oool20000
01D0oooo00<000000?ooo`3oool00P3oool300000480oooo00@000000?ooo`0000000000103oool0
0`000000oooo0?ooo`0C0?ooo`8000004P3oool00`000000oooo0?ooo`0C0?ooo`040000003oool0
0000000006`0oooo001_0?ooo`030000003oool0oooo0100oooo0`00000B0?ooo`030000003oool0
oooo01D0oooo00D000000?ooo`000000oooo000000020?ooo`800000?`3oool200000080oooo00D0
00000?ooo`000000oooo0000000G0?ooo`030000003oool0oooo0100oooo0`00000B0?ooo`030000
003oool0oooo06h0oooo001^0?ooo`030000003oool0oooo00d0oooo00H000000?ooo`000000oooo
0000003oool3000000l0oooo0P0000000`3oool000000?ooo`0E0?ooo`<000000`3oool00`000000
oooo0?ooo`0o0?ooo`040000003oool0oooo0?ooo`<000005P3oool010000000oooo00000000000?
0?ooo`<0000000H0oooo0000003oool000000?ooo`00000?0?ooo`030000003oool0oooo06d0oooo
001^0?ooo`030000003oool0oooo00d0oooo100000030?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`0000006`3oool00`000000oooo0000000m0?ooo`030000003oool0000001/0oooo00<0
00000?ooo`0000004P3oool010000000oooo0?ooo`3oool4000000l0oooo00<000000?ooo`3oool0
K@3oool006l0oooo0P0000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0000000B0?ooo`030000003oool0oooo01X0oooo1000000i0?ooo`@00000703o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`040?ooo`800000L03oool006l0oooo00H000000?ooo`3oool0oooo0000
003oool3000000H0oooo00<000000?ooo`3oool01@3oool010000000oooo00000000000?0?ooo`<0
00007@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01/0oooo0`00000?0?ooo`80
000000<0oooo0000003oool01P3oool00`000000oooo0?ooo`040?ooo`<0000000<0oooo0000003o
ool00P3oool00`000000oooo0?ooo`1^0?ooo`00L03oool4000000040?ooo`000000oooo0?ooo`80
00001@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool06P3oool2000003/0oooo0P00000L0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`030?ooo`8000000P3oool00`000000oooo0000000300000740
oooo001d0?ooo`080000003oool0oooo0?ooo`000000oooo0000003oool3000000P0oooo0P00000@
0?ooo`<000006`3oool010000000oooo00000000000j0?ooo`030000003oool0000001/0oooo0`00
000@0?ooo`800000203oool3000000040?ooo`000000oooo000000<0oooo00<000000?ooo`3oool0
L`3oool007P0oooo0P000000103oool000000?ooo`0000080?ooo`030000003oool0oooo00/0oooo
00H000000?ooo`3oool000000?ooo`00000N0?ooo`030000003oool0oooo03/0oooo00<000000?oo
o`3oool0703oool00`000000oooo000000020?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool01P3oool010000000oooo0000003oool2000007T0oooo001l0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool02P3oool500000200oooo00<000000?ooo`3oool0>@3oool00`00
0000oooo0?ooo`0N0?ooo`@000003@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
07/0oooo00280?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0M0?ooo`80000000<0oooo0000003oool0?03oool010000000oooo00000000000O0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`270?oo
o`00Q`3oool00`000000oooo000000020?ooo`<000001@3oool2000001h0oooo00<000000?ooo`00
00000P00000m0?ooo`<0000000<0oooo0000003oool07@3oool2000000D0oooo0`0000020?ooo`03
0000003oool0000008P0oooo00270?ooo`<000000P3oool010000000oooo0?ooo`0000050?ooo`03
0000003oool0oooo01/0oooo0P0000150?ooo`8000007@3oool00`000000oooo0?ooo`030?ooo`07
0000003oool0oooo0000003oool000000?ooo`02000008P0oooo002:0?ooo`8000000P3oool01@00
0000oooo0?ooo`000000oooo008000007@3oool00`000000oooo0?ooo`170?ooo`030000003oool0
oooo01/0oooo0P0000001@3oool000000?ooo`3oool000000080oooo0P00002;0?ooo`00S`3oool3
00000080oooo00<000000?ooo`3oool06`3oool2000004L0oooo0P00000M0?ooo`030000003oool0
oooo00<00000T03oool00940oooo0`00000N0?ooo`030000003oool0oooo04L0oooo00<000000?oo
o`3oool0703oool300000980oooo002c0?ooo`030000003oool0000004<0oooo00<000000?ooo`00
0000]03oool00;@0oooo1000000o0?ooo`@00000]@3oool00;H0oooo00<000000?ooo`3oool0?`3o
ool00`000000oooo0?ooo`2e0?ooo`00]@3oool2000000l0oooo00@000000?ooo`3oool000006`3o
ool010000000oooo0?ooo`00000?0?ooo`800000]P3oool00;@0oooo00<000000?ooo`0000003@3o
ool00`000000oooo00000004000001T0oooo100000020?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`000000]@3oool00;@0oooo00<000000?ooo`3oool02`3oool300000080oooo00<00000
0?ooo`0000006`3oool00`000000oooo000000020?ooo`<000003@3oool00`000000oooo0?ooo`2c
0?ooo`00]03oool2000000D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020000
0080oooo00@000000?ooo`3oool000005@3oool00`000000oooo000000030?ooo`8000000P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo0P00002e0?ooo`00]03oool01@00
0000oooo0?ooo`000000oooo00<000001@3oool3000000H0oooo0`00000E0?ooo`<000001P3oool3
000000@0oooo00H000000?ooo`00000000000?ooo`0000020?ooo`030000003oool0oooo0;<0oooo
002e0?ooo`<000000P3oool010000000oooo0000000000050?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo
0P000000103oool000000?ooo`3oool300000;H0oooo002g0?ooo`<000000P3oool00`000000oooo
0?ooo`020?ooo`8000002@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00L0oooo
0P0000040?ooo`030000003oool0oooo00<00000^03oool00;`0oooo100000000`3oool000000?oo
o`080?ooo`030000003oool0000001<0oooo00<000000?ooo`0000002@3oool00`000000oooo0000
000300000;d0oooo002o0?ooo`8000002P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00P0oooo0P0000300?ooo`00b`3oool2000000D0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`030?ooo`800000c03oool00</0oooo00<000000?ooo`3oool00P3oool3000000P0
oooo00@000000?ooo`0000000000103oool00`000000oooo0?ooo`3:0?ooo`00c03oool01@000000
oooo0000003oool000000080oooo0P0000050?ooo`8000000P3oool01@000000oooo0000003oool0
00000<d0oooo003=0?ooo`<000000`3oool00`000000oooo0?ooo`050?ooo`040000003oool0oooo
0?ooo`<00000cP3oool00=<0oooo00<000000?ooo`0000000`3oool00`000000oooo0000003D0?oo
o`00e03oool00`000000oooo00000002000000030?ooo`000000oooo0=@0oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-1.05001, -1.05001, 0.00730867, \
0.00730867}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Fractal properties", "Section"],

Cell[CellGroupData[{

Cell["Infinite length encloses finite area", "Subsection"],

Cell[TextData[{
  "Consider a regular polygon with ",
  Cell[BoxData[
      \(TraditionalForm\`n\_0\)]],
  " sides of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0\)]],
  " enclosing area ",
  Cell[BoxData[
      \(TraditionalForm\`A\_0\)]],
  ".  Each iteration multiplies the number of line segments by 4 and reduces \
the length of each segment by a factor of ",
  Cell[BoxData[
      \(TraditionalForm\`1\/3\)]],
  ".  Thus, the length after ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " iterations becomes ",
  Cell[BoxData[
      \(TraditionalForm\`L\_n = \(\((4\/3)\)\^n\) L\_0\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`L\_0 = \(n\_0\) \[ScriptL]\_0\)]],
  " and clearly diverges as ",
  Cell[BoxData[
      \(TraditionalForm\`n \[Rule] \[Infinity]\)]],
  ".  Nevertheless, the following comparison between the circle that \
circumscribes the original triangle and a Koch triangle demonstrates that the \
enclosed area remains finite even though the length becomes infinite as  ",
  Cell[BoxData[
      \(TraditionalForm\`n \[Rule] \[Infinity]\)]],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[
        GraphicsArray[{DisplayKoch[triangle, 4, 
              DisplayFunction \[Rule] Identity], 
            Show[{DisplayKoch[triangle, 4, ShowIterations \[Rule] True, 
                  DisplayFunction \[Rule] Identity], 
                Graphics[Circle[{0.5, \@3\/2 - 1\/\@3}, 1\/\@3]]}, 
              DisplayFunction \[Rule] Identity]}]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .54986 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.47619 0.0130918 0.47619 [
[ 0 0 0 0 ]
[ 1 .54986 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .54986 L
0 .54986 L
closepath
clip
newpath
% Start of sub-graphic
p
0.0238095 0.0130918 0.477324 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.302422 0.952381 [
[ 0 0 0 0 ]
[ 1 1.1547 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.1547 L
0 1.1547 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .30242 m
.02969 .3126 L
s
.02969 .3126 m
.02381 .32279 L
s
.02381 .32279 m
.03557 .32279 L
s
.03557 .32279 m
.04145 .33297 L
s
.04145 .33297 m
.03557 .34315 L
s
.03557 .34315 m
.02381 .34315 L
s
.02381 .34315 m
.02969 .35333 L
s
.02969 .35333 m
.02381 .36352 L
s
.02381 .36352 m
.03557 .36352 L
s
.03557 .36352 m
.04145 .3737 L
s
.04145 .3737 m
.04733 .36352 L
s
.04733 .36352 m
.05908 .36352 L
s
.05908 .36352 m
.06496 .3737 L
s
.06496 .3737 m
.05908 .38388 L
s
.05908 .38388 m
.07084 .38388 L
s
.07084 .38388 m
.07672 .39406 L
s
.07672 .39406 m
.07084 .40425 L
s
.07084 .40425 m
.05908 .40425 L
s
.05908 .40425 m
.06496 .41443 L
s
.06496 .41443 m
.05908 .42461 L
s
.05908 .42461 m
.04733 .42461 L
s
.04733 .42461 m
.04145 .41443 L
s
.04145 .41443 m
.03557 .42461 L
s
.03557 .42461 m
.02381 .42461 L
s
.02381 .42461 m
.02969 .43479 L
s
.02969 .43479 m
.02381 .44498 L
s
.02381 .44498 m
.03557 .44498 L
s
.03557 .44498 m
.04145 .45516 L
s
.04145 .45516 m
.03557 .46534 L
s
.03557 .46534 m
.02381 .46534 L
s
.02381 .46534 m
.02969 .47552 L
s
.02969 .47552 m
.02381 .48571 L
s
.02381 .48571 m
.03557 .48571 L
s
.03557 .48571 m
.04145 .49589 L
s
.04145 .49589 m
.04733 .48571 L
s
.04733 .48571 m
.05908 .48571 L
s
.05908 .48571 m
.06496 .49589 L
s
.06496 .49589 m
.05908 .50607 L
s
.05908 .50607 m
.07084 .50607 L
s
.07084 .50607 m
.07672 .51626 L
s
.07672 .51626 m
.0826 .50607 L
s
.0826 .50607 m
.09436 .50607 L
s
.09436 .50607 m
.08848 .49589 L
s
.08848 .49589 m
.09436 .48571 L
s
.09436 .48571 m
.10611 .48571 L
s
.10611 .48571 m
.11199 .49589 L
s
.11199 .49589 m
.11787 .48571 L
s
.11787 .48571 m
.12963 .48571 L
s
.12963 .48571 m
.13551 .49589 L
s
.13551 .49589 m
.12963 .50607 L
s
.12963 .50607 m
.14139 .50607 L
s
.14139 .50607 m
.14727 .51626 L
s
.14727 .51626 m
.14139 .52644 L
s
.14139 .52644 m
.12963 .52644 L
s
.12963 .52644 m
.13551 .53662 L
s
.13551 .53662 m
.12963 .5468 L
s
.12963 .5468 m
.14139 .5468 L
s
.14139 .5468 m
.14727 .55699 L
s
.14727 .55699 m
.15315 .5468 L
s
.15315 .5468 m
.1649 .5468 L
s
.1649 .5468 m
.17078 .55699 L
s
.17078 .55699 m
.1649 .56717 L
s
.1649 .56717 m
.17666 .56717 L
s
.17666 .56717 m
.18254 .57735 L
s
.18254 .57735 m
.17666 .58753 L
s
.17666 .58753 m
.1649 .58753 L
s
.1649 .58753 m
.17078 .59772 L
s
.17078 .59772 m
.1649 .6079 L
s
.1649 .6079 m
.15315 .6079 L
s
.15315 .6079 m
.14727 .59772 L
s
.14727 .59772 m
.14139 .6079 L
s
.14139 .6079 m
.12963 .6079 L
s
.12963 .6079 m
.13551 .61808 L
s
.13551 .61808 m
.12963 .62826 L
s
.12963 .62826 m
.14139 .62826 L
s
.14139 .62826 m
.14727 .63845 L
s
.14727 .63845 m
.14139 .64863 L
s
.14139 .64863 m
.12963 .64863 L
s
.12963 .64863 m
.13551 .65881 L
s
.13551 .65881 m
.12963 .66899 L
s
.12963 .66899 m
.11787 .66899 L
s
.11787 .66899 m
.11199 .65881 L
s
.11199 .65881 m
.10611 .66899 L
s
.10611 .66899 m
.09436 .66899 L
s
.09436 .66899 m
.08848 .65881 L
s
.08848 .65881 m
.09436 .64863 L
s
.09436 .64863 m
.0826 .64863 L
s
.0826 .64863 m
.07672 .63845 L
s
.07672 .63845 m
.07084 .64863 L
s
.07084 .64863 m
.05908 .64863 L
s
.05908 .64863 m
.06496 .65881 L
s
.06496 .65881 m
.05908 .66899 L
s
.05908 .66899 m
.04733 .66899 L
s
.04733 .66899 m
.04145 .65881 L
s
.04145 .65881 m
.03557 .66899 L
s
.03557 .66899 m
.02381 .66899 L
s
.02381 .66899 m
.02969 .67918 L
s
.02969 .67918 m
.02381 .68936 L
s
.02381 .68936 m
.03557 .68936 L
s
.03557 .68936 m
.04145 .69954 L
s
.04145 .69954 m
.03557 .70972 L
s
.03557 .70972 m
.02381 .70972 L
s
.02381 .70972 m
.02969 .71991 L
s
.02969 .71991 m
.02381 .73009 L
s
.02381 .73009 m
.03557 .73009 L
s
.03557 .73009 m
.04145 .74027 L
s
.04145 .74027 m
.04733 .73009 L
s
.04733 .73009 m
.05908 .73009 L
s
.05908 .73009 m
.06496 .74027 L
s
.06496 .74027 m
.05908 .75045 L
s
.05908 .75045 m
.07084 .75045 L
s
.07084 .75045 m
.07672 .76064 L
s
.07672 .76064 m
.07084 .77082 L
s
.07084 .77082 m
.05908 .77082 L
s
.05908 .77082 m
.06496 .781 L
s
.06496 .781 m
.05908 .79118 L
s
.05908 .79118 m
.04733 .79118 L
s
.04733 .79118 m
.04145 .781 L
s
.04145 .781 m
.03557 .79118 L
s
.03557 .79118 m
.02381 .79118 L
s
.02381 .79118 m
.02969 .80137 L
s
.02969 .80137 m
.02381 .81155 L
s
.02381 .81155 m
.03557 .81155 L
s
.03557 .81155 m
.04145 .82173 L
s
.04145 .82173 m
.03557 .83191 L
s
.03557 .83191 m
.02381 .83191 L
s
.02381 .83191 m
.02969 .8421 L
s
.02969 .8421 m
.02381 .85228 L
s
.02381 .85228 m
.03557 .85228 L
s
.03557 .85228 m
.04145 .86246 L
s
.04145 .86246 m
.04733 .85228 L
s
.04733 .85228 m
.05908 .85228 L
s
.05908 .85228 m
.06496 .86246 L
s
.06496 .86246 m
.05908 .87264 L
s
.05908 .87264 m
.07084 .87264 L
s
.07084 .87264 m
.07672 .88283 L
s
.07672 .88283 m
.0826 .87264 L
s
.0826 .87264 m
.09436 .87264 L
s
.09436 .87264 m
.08848 .86246 L
s
.08848 .86246 m
.09436 .85228 L
s
.09436 .85228 m
.10611 .85228 L
s
.10611 .85228 m
.11199 .86246 L
s
.11199 .86246 m
.11787 .85228 L
s
.11787 .85228 m
.12963 .85228 L
s
.12963 .85228 m
.13551 .86246 L
s
.13551 .86246 m
.12963 .87264 L
s
.12963 .87264 m
.14139 .87264 L
s
.14139 .87264 m
.14727 .88283 L
s
.14727 .88283 m
.14139 .89301 L
s
.14139 .89301 m
.12963 .89301 L
s
.12963 .89301 m
.13551 .90319 L
s
.13551 .90319 m
.12963 .91337 L
s
.12963 .91337 m
.14139 .91337 L
s
.14139 .91337 m
.14727 .92356 L
s
.14727 .92356 m
.15315 .91337 L
s
.15315 .91337 m
.1649 .91337 L
s
.1649 .91337 m
.17078 .92356 L
s
.17078 .92356 m
.1649 .93374 L
s
.1649 .93374 m
.17666 .93374 L
s
.17666 .93374 m
.18254 .94392 L
s
.18254 .94392 m
.18842 .93374 L
s
.18842 .93374 m
.20018 .93374 L
s
.20018 .93374 m
.1943 .92356 L
s
.1943 .92356 m
.20018 .91337 L
s
.20018 .91337 m
.21193 .91337 L
s
.21193 .91337 m
.21781 .92356 L
s
.21781 .92356 m
.22369 .91337 L
s
.22369 .91337 m
.23545 .91337 L
s
.23545 .91337 m
.22957 .90319 L
s
.22957 .90319 m
.23545 .89301 L
s
.23545 .89301 m
.22369 .89301 L
s
.22369 .89301 m
.21781 .88283 L
s
.21781 .88283 m
.22369 .87264 L
s
.22369 .87264 m
.23545 .87264 L
s
.23545 .87264 m
.22957 .86246 L
s
.22957 .86246 m
.23545 .85228 L
s
.23545 .85228 m
.24721 .85228 L
s
.24721 .85228 m
.25309 .86246 L
s
.25309 .86246 m
.25897 .85228 L
s
.25897 .85228 m
.27072 .85228 L
s
.27072 .85228 m
.2766 .86246 L
s
.2766 .86246 m
.27072 .87264 L
s
.27072 .87264 m
.28248 .87264 L
s
.28248 .87264 m
.28836 .88283 L
s
.28836 .88283 m
.29424 .87264 L
s
.29424 .87264 m
.306 .87264 L
s
.306 .87264 m
.30012 .86246 L
s
.30012 .86246 m
.306 .85228 L
s
.306 .85228 m
.31775 .85228 L
s
.31775 .85228 m
.32363 .86246 L
s
.32363 .86246 m
.32951 .85228 L
s
.32951 .85228 m
.34127 .85228 L
s
.34127 .85228 m
.34715 .86246 L
s
.34715 .86246 m
.34127 .87264 L
s
.34127 .87264 m
.35303 .87264 L
s
.35303 .87264 m
.35891 .88283 L
s
.35891 .88283 m
.35303 .89301 L
s
.35303 .89301 m
.34127 .89301 L
s
.34127 .89301 m
.34715 .90319 L
s
.34715 .90319 m
.34127 .91337 L
s
.34127 .91337 m
.35303 .91337 L
s
.35303 .91337 m
.35891 .92356 L
s
.35891 .92356 m
.36479 .91337 L
s
.36479 .91337 m
.37654 .91337 L
s
.37654 .91337 m
.38242 .92356 L
s
.38242 .92356 m
.37654 .93374 L
s
.37654 .93374 m
.3883 .93374 L
s
.3883 .93374 m
.39418 .94392 L
s
.39418 .94392 m
.3883 .9541 L
s
.3883 .9541 m
.37654 .9541 L
s
.37654 .9541 m
.38242 .96429 L
s
.38242 .96429 m
.37654 .97447 L
s
.37654 .97447 m
.36479 .97447 L
s
.36479 .97447 m
.35891 .96429 L
s
.35891 .96429 m
.35303 .97447 L
s
.35303 .97447 m
.34127 .97447 L
s
.34127 .97447 m
.34715 .98465 L
s
.34715 .98465 m
.34127 .99483 L
s
.34127 .99483 m
.35303 .99483 L
s
.35303 .99483 m
.35891 1.00502 L
s
.35891 1.00502 m
.35303 1.0152 L
s
.35303 1.0152 m
.34127 1.0152 L
s
.34127 1.0152 m
.34715 1.02538 L
s
.34715 1.02538 m
.34127 1.03556 L
s
.34127 1.03556 m
.35303 1.03556 L
s
.35303 1.03556 m
.35891 1.04575 L
s
.35891 1.04575 m
.36479 1.03556 L
s
.36479 1.03556 m
.37654 1.03556 L
s
.37654 1.03556 m
.38242 1.04575 L
s
.38242 1.04575 m
.37654 1.05593 L
s
.37654 1.05593 m
.3883 1.05593 L
s
.3883 1.05593 m
.39418 1.06611 L
s
.39418 1.06611 m
.40006 1.05593 L
s
.40006 1.05593 m
.41182 1.05593 L
s
.41182 1.05593 m
.40594 1.04575 L
s
.40594 1.04575 m
.41182 1.03556 L
s
.41182 1.03556 m
.42357 1.03556 L
s
.42357 1.03556 m
.42945 1.04575 L
s
.42945 1.04575 m
.43533 1.03556 L
s
.43533 1.03556 m
.44709 1.03556 L
s
.44709 1.03556 m
.45297 1.04575 L
s
.45297 1.04575 m
.44709 1.05593 L
s
.44709 1.05593 m
.45885 1.05593 L
s
.45885 1.05593 m
.46473 1.06611 L
s
.46473 1.06611 m
.45885 1.07629 L
s
.45885 1.07629 m
.44709 1.07629 L
s
.44709 1.07629 m
.45297 1.08648 L
s
.45297 1.08648 m
.44709 1.09666 L
s
.44709 1.09666 m
.45885 1.09666 L
s
.45885 1.09666 m
.46473 1.10684 L
s
.46473 1.10684 m
.47061 1.09666 L
s
.47061 1.09666 m
.48236 1.09666 L
s
.48236 1.09666 m
.48824 1.10684 L
s
.48824 1.10684 m
.48236 1.11703 L
s
.48236 1.11703 m
.49412 1.11703 L
s
.49412 1.11703 m
.5 1.12721 L
s
.5 1.12721 m
.50588 1.11703 L
s
.50588 1.11703 m
.51764 1.11703 L
s
.51764 1.11703 m
.51176 1.10684 L
s
.51176 1.10684 m
.51764 1.09666 L
s
.51764 1.09666 m
.52939 1.09666 L
s
.52939 1.09666 m
.53527 1.10684 L
s
.53527 1.10684 m
.54115 1.09666 L
s
.54115 1.09666 m
.55291 1.09666 L
s
.55291 1.09666 m
.54703 1.08648 L
s
.54703 1.08648 m
.55291 1.07629 L
s
.55291 1.07629 m
.54115 1.07629 L
s
.54115 1.07629 m
.53527 1.06611 L
s
.53527 1.06611 m
.54115 1.05593 L
s
.54115 1.05593 m
.55291 1.05593 L
s
.55291 1.05593 m
.54703 1.04575 L
s
.54703 1.04575 m
.55291 1.03556 L
s
.55291 1.03556 m
.56467 1.03556 L
s
.56467 1.03556 m
.57055 1.04575 L
s
.57055 1.04575 m
.57643 1.03556 L
s
.57643 1.03556 m
.58818 1.03556 L
s
.58818 1.03556 m
.59406 1.04575 L
s
.59406 1.04575 m
.58818 1.05593 L
s
.58818 1.05593 m
.59994 1.05593 L
s
.59994 1.05593 m
.60582 1.06611 L
s
.60582 1.06611 m
.6117 1.05593 L
s
.6117 1.05593 m
.62346 1.05593 L
s
.62346 1.05593 m
.61758 1.04575 L
s
.61758 1.04575 m
.62346 1.03556 L
s
.62346 1.03556 m
.63521 1.03556 L
s
.63521 1.03556 m
.64109 1.04575 L
s
.64109 1.04575 m
.64697 1.03556 L
s
.64697 1.03556 m
.65873 1.03556 L
s
.65873 1.03556 m
.65285 1.02538 L
s
.65285 1.02538 m
.65873 1.0152 L
s
.65873 1.0152 m
.64697 1.0152 L
s
.64697 1.0152 m
.64109 1.00502 L
s
.64109 1.00502 m
.64697 .99483 L
s
.64697 .99483 m
.65873 .99483 L
s
.65873 .99483 m
.65285 .98465 L
s
.65285 .98465 m
.65873 .97447 L
s
.65873 .97447 m
.64697 .97447 L
s
.64697 .97447 m
.64109 .96429 L
s
.64109 .96429 m
.63521 .97447 L
s
.63521 .97447 m
.62346 .97447 L
s
.62346 .97447 m
.61758 .96429 L
s
.61758 .96429 m
.62346 .9541 L
s
.62346 .9541 m
.6117 .9541 L
s
.6117 .9541 m
.60582 .94392 L
s
.60582 .94392 m
.6117 .93374 L
s
.6117 .93374 m
.62346 .93374 L
s
.62346 .93374 m
.61758 .92356 L
s
.61758 .92356 m
.62346 .91337 L
s
.62346 .91337 m
.63521 .91337 L
s
.63521 .91337 m
.64109 .92356 L
s
.64109 .92356 m
.64697 .91337 L
s
.64697 .91337 m
.65873 .91337 L
s
.65873 .91337 m
.65285 .90319 L
s
.65285 .90319 m
.65873 .89301 L
s
.65873 .89301 m
.64697 .89301 L
s
.64697 .89301 m
.64109 .88283 L
s
.64109 .88283 m
.64697 .87264 L
s
.64697 .87264 m
.65873 .87264 L
s
.65873 .87264 m
.65285 .86246 L
s
.65285 .86246 m
.65873 .85228 L
s
.65873 .85228 m
.67049 .85228 L
s
.67049 .85228 m
.67637 .86246 L
s
.67637 .86246 m
.68225 .85228 L
s
.68225 .85228 m
.694 .85228 L
s
.694 .85228 m
.69988 .86246 L
s
.69988 .86246 m
.694 .87264 L
s
.694 .87264 m
.70576 .87264 L
s
.70576 .87264 m
.71164 .88283 L
s
.71164 .88283 m
.71752 .87264 L
s
.71752 .87264 m
.72928 .87264 L
s
.72928 .87264 m
.7234 .86246 L
s
.7234 .86246 m
.72928 .85228 L
s
.72928 .85228 m
.74103 .85228 L
s
.74103 .85228 m
.74691 .86246 L
s
.74691 .86246 m
.75279 .85228 L
s
.75279 .85228 m
.76455 .85228 L
s
.76455 .85228 m
.77043 .86246 L
s
.77043 .86246 m
.76455 .87264 L
s
.76455 .87264 m
.77631 .87264 L
s
.77631 .87264 m
.78219 .88283 L
s
.78219 .88283 m
.77631 .89301 L
s
.77631 .89301 m
.76455 .89301 L
s
.76455 .89301 m
.77043 .90319 L
s
.77043 .90319 m
.76455 .91337 L
s
.76455 .91337 m
.77631 .91337 L
s
.77631 .91337 m
.78219 .92356 L
s
.78219 .92356 m
.78807 .91337 L
s
.78807 .91337 m
.79982 .91337 L
s
.79982 .91337 m
.8057 .92356 L
s
.8057 .92356 m
.79982 .93374 L
s
.79982 .93374 m
.81158 .93374 L
s
.81158 .93374 m
.81746 .94392 L
s
.81746 .94392 m
.82334 .93374 L
s
.82334 .93374 m
.8351 .93374 L
s
.8351 .93374 m
.82922 .92356 L
s
.82922 .92356 m
.8351 .91337 L
s
.8351 .91337 m
.84685 .91337 L
s
.84685 .91337 m
.85273 .92356 L
s
.85273 .92356 m
.85861 .91337 L
s
.85861 .91337 m
.87037 .91337 L
s
.87037 .91337 m
.86449 .90319 L
s
.86449 .90319 m
.87037 .89301 L
s
.87037 .89301 m
.85861 .89301 L
s
.85861 .89301 m
.85273 .88283 L
s
.85273 .88283 m
.85861 .87264 L
s
.85861 .87264 m
.87037 .87264 L
s
.87037 .87264 m
.86449 .86246 L
s
.86449 .86246 m
.87037 .85228 L
s
.87037 .85228 m
.88213 .85228 L
s
.88213 .85228 m
.88801 .86246 L
s
.88801 .86246 m
.89389 .85228 L
s
.89389 .85228 m
.90564 .85228 L
s
.90564 .85228 m
.91152 .86246 L
s
.91152 .86246 m
.90564 .87264 L
s
.90564 .87264 m
.9174 .87264 L
s
.9174 .87264 m
.92328 .88283 L
s
.92328 .88283 m
.92916 .87264 L
s
.92916 .87264 m
.94092 .87264 L
s
.94092 .87264 m
.93504 .86246 L
s
.93504 .86246 m
.94092 .85228 L
s
.94092 .85228 m
.95267 .85228 L
s
.95267 .85228 m
.95855 .86246 L
s
.95855 .86246 m
.96443 .85228 L
s
.96443 .85228 m
.97619 .85228 L
s
.97619 .85228 m
.97031 .8421 L
s
.97031 .8421 m
.97619 .83191 L
s
.97619 .83191 m
.96443 .83191 L
s
.96443 .83191 m
.95855 .82173 L
s
.95855 .82173 m
.96443 .81155 L
s
.96443 .81155 m
.97619 .81155 L
s
.97619 .81155 m
.97031 .80137 L
s
.97031 .80137 m
.97619 .79118 L
s
.97619 .79118 m
.96443 .79118 L
s
.96443 .79118 m
.95855 .781 L
s
.95855 .781 m
.95267 .79118 L
s
.95267 .79118 m
.94092 .79118 L
s
.94092 .79118 m
.93504 .781 L
s
.93504 .781 m
.94092 .77082 L
s
.94092 .77082 m
.92916 .77082 L
s
.92916 .77082 m
.92328 .76064 L
s
.92328 .76064 m
.92916 .75045 L
s
.92916 .75045 m
.94092 .75045 L
s
.94092 .75045 m
.93504 .74027 L
s
.93504 .74027 m
.94092 .73009 L
s
.94092 .73009 m
.95267 .73009 L
s
.95267 .73009 m
.95855 .74027 L
s
.95855 .74027 m
.96443 .73009 L
s
.96443 .73009 m
.97619 .73009 L
s
.97619 .73009 m
.97031 .71991 L
s
.97031 .71991 m
.97619 .70972 L
s
.97619 .70972 m
.96443 .70972 L
s
.96443 .70972 m
.95855 .69954 L
s
.95855 .69954 m
.96443 .68936 L
s
.96443 .68936 m
.97619 .68936 L
s
.97619 .68936 m
.97031 .67918 L
s
.97031 .67918 m
.97619 .66899 L
s
.97619 .66899 m
.96443 .66899 L
s
.96443 .66899 m
.95855 .65881 L
s
.95855 .65881 m
.95267 .66899 L
s
.95267 .66899 m
.94092 .66899 L
s
.94092 .66899 m
.93504 .65881 L
s
.93504 .65881 m
.94092 .64863 L
s
.94092 .64863 m
.92916 .64863 L
s
.92916 .64863 m
.92328 .63845 L
s
.92328 .63845 m
.9174 .64863 L
s
.9174 .64863 m
.90564 .64863 L
s
.90564 .64863 m
.91152 .65881 L
s
.91152 .65881 m
.90564 .66899 L
s
.90564 .66899 m
.89389 .66899 L
s
.89389 .66899 m
.88801 .65881 L
s
.88801 .65881 m
.88213 .66899 L
s
.88213 .66899 m
.87037 .66899 L
s
.87037 .66899 m
.86449 .65881 L
s
.86449 .65881 m
.87037 .64863 L
s
.87037 .64863 m
.85861 .64863 L
s
.85861 .64863 m
.85273 .63845 L
s
.85273 .63845 m
.85861 .62826 L
s
.85861 .62826 m
.87037 .62826 L
s
.87037 .62826 m
.86449 .61808 L
s
.86449 .61808 m
.87037 .6079 L
s
.87037 .6079 m
.85861 .6079 L
s
.85861 .6079 m
.85273 .59772 L
s
.85273 .59772 m
.84685 .6079 L
s
.84685 .6079 m
.8351 .6079 L
s
.8351 .6079 m
.82922 .59772 L
s
.82922 .59772 m
.8351 .58753 L
s
.8351 .58753 m
.82334 .58753 L
s
.82334 .58753 m
.81746 .57735 L
s
.81746 .57735 m
.82334 .56717 L
s
.82334 .56717 m
.8351 .56717 L
s
.8351 .56717 m
.82922 .55699 L
s
.82922 .55699 m
.8351 .5468 L
s
.8351 .5468 m
.84685 .5468 L
s
.84685 .5468 m
.85273 .55699 L
s
.85273 .55699 m
.85861 .5468 L
s
.85861 .5468 m
.87037 .5468 L
s
.87037 .5468 m
.86449 .53662 L
s
.86449 .53662 m
.87037 .52644 L
s
.87037 .52644 m
.85861 .52644 L
s
.85861 .52644 m
.85273 .51626 L
s
.85273 .51626 m
.85861 .50607 L
s
.85861 .50607 m
.87037 .50607 L
s
.87037 .50607 m
.86449 .49589 L
s
.86449 .49589 m
.87037 .48571 L
s
.87037 .48571 m
.88213 .48571 L
s
.88213 .48571 m
.88801 .49589 L
s
.88801 .49589 m
.89389 .48571 L
s
.89389 .48571 m
.90564 .48571 L
s
.90564 .48571 m
.91152 .49589 L
s
.91152 .49589 m
.90564 .50607 L
s
.90564 .50607 m
.9174 .50607 L
s
.9174 .50607 m
.92328 .51626 L
s
.92328 .51626 m
.92916 .50607 L
s
.92916 .50607 m
.94092 .50607 L
s
.94092 .50607 m
.93504 .49589 L
s
.93504 .49589 m
.94092 .48571 L
s
.94092 .48571 m
.95267 .48571 L
s
.95267 .48571 m
.95855 .49589 L
s
.95855 .49589 m
.96443 .48571 L
s
.96443 .48571 m
.97619 .48571 L
s
.97619 .48571 m
.97031 .47552 L
s
.97031 .47552 m
.97619 .46534 L
s
.97619 .46534 m
.96443 .46534 L
s
.96443 .46534 m
.95855 .45516 L
s
.95855 .45516 m
.96443 .44498 L
s
.96443 .44498 m
.97619 .44498 L
s
.97619 .44498 m
.97031 .43479 L
s
.97031 .43479 m
.97619 .42461 L
s
.97619 .42461 m
.96443 .42461 L
s
.96443 .42461 m
.95855 .41443 L
s
.95855 .41443 m
.95267 .42461 L
s
.95267 .42461 m
.94092 .42461 L
s
.94092 .42461 m
.93504 .41443 L
s
.93504 .41443 m
.94092 .40425 L
s
.94092 .40425 m
.92916 .40425 L
s
.92916 .40425 m
.92328 .39406 L
s
.92328 .39406 m
.92916 .38388 L
s
.92916 .38388 m
.94092 .38388 L
s
.94092 .38388 m
.93504 .3737 L
s
.93504 .3737 m
.94092 .36352 L
s
.94092 .36352 m
.95267 .36352 L
s
.95267 .36352 m
.95855 .3737 L
s
.95855 .3737 m
.96443 .36352 L
s
.96443 .36352 m
.97619 .36352 L
s
.97619 .36352 m
.97031 .35333 L
s
.97031 .35333 m
.97619 .34315 L
s
.97619 .34315 m
.96443 .34315 L
s
.96443 .34315 m
.95855 .33297 L
s
.95855 .33297 m
.96443 .32279 L
s
.96443 .32279 m
.97619 .32279 L
s
.97619 .32279 m
.97031 .3126 L
s
.97031 .3126 m
.97619 .30242 L
s
.97619 .30242 m
.96443 .30242 L
s
.96443 .30242 m
.95855 .29224 L
s
.95855 .29224 m
.95267 .30242 L
s
.95267 .30242 m
.94092 .30242 L
s
.94092 .30242 m
.93504 .29224 L
s
.93504 .29224 m
.94092 .28206 L
s
.94092 .28206 m
.92916 .28206 L
s
.92916 .28206 m
.92328 .27187 L
s
.92328 .27187 m
.9174 .28206 L
s
.9174 .28206 m
.90564 .28206 L
s
.90564 .28206 m
.91152 .29224 L
s
.91152 .29224 m
.90564 .30242 L
s
.90564 .30242 m
.89389 .30242 L
s
.89389 .30242 m
.88801 .29224 L
s
.88801 .29224 m
.88213 .30242 L
s
.88213 .30242 m
.87037 .30242 L
s
.87037 .30242 m
.86449 .29224 L
s
.86449 .29224 m
.87037 .28206 L
s
.87037 .28206 m
.85861 .28206 L
s
.85861 .28206 m
.85273 .27187 L
s
.85273 .27187 m
.85861 .26169 L
s
.85861 .26169 m
.87037 .26169 L
s
.87037 .26169 m
.86449 .25151 L
s
.86449 .25151 m
.87037 .24133 L
s
.87037 .24133 m
.85861 .24133 L
s
.85861 .24133 m
.85273 .23114 L
s
.85273 .23114 m
.84685 .24133 L
s
.84685 .24133 m
.8351 .24133 L
s
.8351 .24133 m
.82922 .23114 L
s
.82922 .23114 m
.8351 .22096 L
s
.8351 .22096 m
.82334 .22096 L
s
.82334 .22096 m
.81746 .21078 L
s
.81746 .21078 m
.81158 .22096 L
s
.81158 .22096 m
.79982 .22096 L
s
.79982 .22096 m
.8057 .23114 L
s
.8057 .23114 m
.79982 .24133 L
s
.79982 .24133 m
.78807 .24133 L
s
.78807 .24133 m
.78219 .23114 L
s
.78219 .23114 m
.77631 .24133 L
s
.77631 .24133 m
.76455 .24133 L
s
.76455 .24133 m
.77043 .25151 L
s
.77043 .25151 m
.76455 .26169 L
s
.76455 .26169 m
.77631 .26169 L
s
.77631 .26169 m
.78219 .27187 L
s
.78219 .27187 m
.77631 .28206 L
s
.77631 .28206 m
.76455 .28206 L
s
.76455 .28206 m
.77043 .29224 L
s
.77043 .29224 m
.76455 .30242 L
s
.76455 .30242 m
.75279 .30242 L
s
.75279 .30242 m
.74691 .29224 L
s
.74691 .29224 m
.74103 .30242 L
s
.74103 .30242 m
.72928 .30242 L
s
.72928 .30242 m
.7234 .29224 L
s
.7234 .29224 m
.72928 .28206 L
s
.72928 .28206 m
.71752 .28206 L
s
.71752 .28206 m
.71164 .27187 L
s
.71164 .27187 m
.70576 .28206 L
s
.70576 .28206 m
.694 .28206 L
s
.694 .28206 m
.69988 .29224 L
s
.69988 .29224 m
.694 .30242 L
s
.694 .30242 m
.68225 .30242 L
s
.68225 .30242 m
.67637 .29224 L
s
.67637 .29224 m
.67049 .30242 L
s
.67049 .30242 m
.65873 .30242 L
s
.65873 .30242 m
.65285 .29224 L
s
.65285 .29224 m
.65873 .28206 L
s
.65873 .28206 m
.64697 .28206 L
s
.64697 .28206 m
.64109 .27187 L
s
.64109 .27187 m
.64697 .26169 L
s
.64697 .26169 m
.65873 .26169 L
s
.65873 .26169 m
.65285 .25151 L
s
.65285 .25151 m
.65873 .24133 L
s
.65873 .24133 m
.64697 .24133 L
s
.64697 .24133 m
.64109 .23114 L
s
.64109 .23114 m
.63521 .24133 L
s
.63521 .24133 m
.62346 .24133 L
s
.62346 .24133 m
.61758 .23114 L
s
.61758 .23114 m
.62346 .22096 L
s
.62346 .22096 m
.6117 .22096 L
s
.6117 .22096 m
.60582 .21078 L
s
.60582 .21078 m
.6117 .2006 L
s
.6117 .2006 m
.62346 .2006 L
s
.62346 .2006 m
.61758 .19041 L
s
.61758 .19041 m
.62346 .18023 L
s
.62346 .18023 m
.63521 .18023 L
s
.63521 .18023 m
.64109 .19041 L
s
.64109 .19041 m
.64697 .18023 L
s
.64697 .18023 m
.65873 .18023 L
s
.65873 .18023 m
.65285 .17005 L
s
.65285 .17005 m
.65873 .15987 L
s
.65873 .15987 m
.64697 .15987 L
s
.64697 .15987 m
.64109 .14968 L
s
.64109 .14968 m
.64697 .1395 L
s
.64697 .1395 m
.65873 .1395 L
s
.65873 .1395 m
.65285 .12932 L
s
.65285 .12932 m
.65873 .11914 L
s
.65873 .11914 m
.64697 .11914 L
s
.64697 .11914 m
.64109 .10895 L
s
.64109 .10895 m
.63521 .11914 L
s
.63521 .11914 m
.62346 .11914 L
s
.62346 .11914 m
.61758 .10895 L
s
.61758 .10895 m
.62346 .09877 L
s
.62346 .09877 m
.6117 .09877 L
s
.6117 .09877 m
.60582 .08859 L
s
.60582 .08859 m
.59994 .09877 L
s
.59994 .09877 m
.58818 .09877 L
s
.58818 .09877 m
.59406 .10895 L
s
.59406 .10895 m
.58818 .11914 L
s
.58818 .11914 m
.57643 .11914 L
s
.57643 .11914 m
.57055 .10895 L
s
.57055 .10895 m
.56467 .11914 L
s
.56467 .11914 m
.55291 .11914 L
s
.55291 .11914 m
.54703 .10895 L
s
.54703 .10895 m
.55291 .09877 L
s
.55291 .09877 m
.54115 .09877 L
s
.54115 .09877 m
.53527 .08859 L
s
.53527 .08859 m
.54115 .07841 L
s
.54115 .07841 m
.55291 .07841 L
s
.55291 .07841 m
.54703 .06822 L
s
.54703 .06822 m
.55291 .05804 L
s
.55291 .05804 m
.54115 .05804 L
s
.54115 .05804 m
.53527 .04786 L
s
.53527 .04786 m
.52939 .05804 L
s
.52939 .05804 m
.51764 .05804 L
s
.51764 .05804 m
.51176 .04786 L
s
.51176 .04786 m
.51764 .03768 L
s
.51764 .03768 m
.50588 .03768 L
s
.50588 .03768 m
.5 .02749 L
s
.5 .02749 m
.49412 .03768 L
s
.49412 .03768 m
.48236 .03768 L
s
.48236 .03768 m
.48824 .04786 L
s
.48824 .04786 m
.48236 .05804 L
s
.48236 .05804 m
.47061 .05804 L
s
.47061 .05804 m
.46473 .04786 L
s
.46473 .04786 m
.45885 .05804 L
s
.45885 .05804 m
.44709 .05804 L
s
.44709 .05804 m
.45297 .06822 L
s
.45297 .06822 m
.44709 .07841 L
s
.44709 .07841 m
.45885 .07841 L
s
.45885 .07841 m
.46473 .08859 L
s
.46473 .08859 m
.45885 .09877 L
s
.45885 .09877 m
.44709 .09877 L
s
.44709 .09877 m
.45297 .10895 L
s
.45297 .10895 m
.44709 .11914 L
s
.44709 .11914 m
.43533 .11914 L
s
.43533 .11914 m
.42945 .10895 L
s
.42945 .10895 m
.42357 .11914 L
s
.42357 .11914 m
.41182 .11914 L
s
.41182 .11914 m
.40594 .10895 L
s
.40594 .10895 m
.41182 .09877 L
s
.41182 .09877 m
.40006 .09877 L
s
.40006 .09877 m
.39418 .08859 L
s
.39418 .08859 m
.3883 .09877 L
s
.3883 .09877 m
.37654 .09877 L
s
.37654 .09877 m
.38242 .10895 L
s
.38242 .10895 m
.37654 .11914 L
s
.37654 .11914 m
.36479 .11914 L
s
.36479 .11914 m
.35891 .10895 L
s
.35891 .10895 m
.35303 .11914 L
s
.35303 .11914 m
.34127 .11914 L
s
.34127 .11914 m
.34715 .12932 L
s
.34715 .12932 m
.34127 .1395 L
s
.34127 .1395 m
.35303 .1395 L
s
.35303 .1395 m
.35891 .14968 L
s
.35891 .14968 m
.35303 .15987 L
s
.35303 .15987 m
.34127 .15987 L
s
.34127 .15987 m
.34715 .17005 L
s
.34715 .17005 m
.34127 .18023 L
s
.34127 .18023 m
.35303 .18023 L
s
.35303 .18023 m
.35891 .19041 L
s
.35891 .19041 m
.36479 .18023 L
s
.36479 .18023 m
.37654 .18023 L
s
.37654 .18023 m
.38242 .19041 L
s
.38242 .19041 m
.37654 .2006 L
s
.37654 .2006 m
.3883 .2006 L
s
.3883 .2006 m
.39418 .21078 L
s
.39418 .21078 m
.3883 .22096 L
s
.3883 .22096 m
.37654 .22096 L
s
.37654 .22096 m
.38242 .23114 L
s
.38242 .23114 m
.37654 .24133 L
s
.37654 .24133 m
.36479 .24133 L
s
.36479 .24133 m
.35891 .23114 L
s
.35891 .23114 m
.35303 .24133 L
s
.35303 .24133 m
.34127 .24133 L
s
.34127 .24133 m
.34715 .25151 L
s
.34715 .25151 m
.34127 .26169 L
s
.34127 .26169 m
.35303 .26169 L
s
.35303 .26169 m
.35891 .27187 L
s
.35891 .27187 m
.35303 .28206 L
s
.35303 .28206 m
.34127 .28206 L
s
.34127 .28206 m
.34715 .29224 L
s
.34715 .29224 m
.34127 .30242 L
s
.34127 .30242 m
.32951 .30242 L
s
.32951 .30242 m
.32363 .29224 L
s
.32363 .29224 m
.31775 .30242 L
s
.31775 .30242 m
.306 .30242 L
s
.306 .30242 m
.30012 .29224 L
s
.30012 .29224 m
.306 .28206 L
s
.306 .28206 m
.29424 .28206 L
s
.29424 .28206 m
.28836 .27187 L
s
.28836 .27187 m
.28248 .28206 L
s
.28248 .28206 m
.27072 .28206 L
s
.27072 .28206 m
.2766 .29224 L
s
.2766 .29224 m
.27072 .30242 L
s
.27072 .30242 m
.25897 .30242 L
s
.25897 .30242 m
.25309 .29224 L
s
.25309 .29224 m
.24721 .30242 L
s
.24721 .30242 m
.23545 .30242 L
s
.23545 .30242 m
.22957 .29224 L
s
.22957 .29224 m
.23545 .28206 L
s
.23545 .28206 m
.22369 .28206 L
s
.22369 .28206 m
.21781 .27187 L
s
.21781 .27187 m
.22369 .26169 L
s
.22369 .26169 m
.23545 .26169 L
s
.23545 .26169 m
.22957 .25151 L
s
.22957 .25151 m
.23545 .24133 L
s
.23545 .24133 m
.22369 .24133 L
s
.22369 .24133 m
.21781 .23114 L
s
.21781 .23114 m
.21193 .24133 L
s
.21193 .24133 m
.20018 .24133 L
s
.20018 .24133 m
.1943 .23114 L
s
.1943 .23114 m
.20018 .22096 L
s
.20018 .22096 m
.18842 .22096 L
s
.18842 .22096 m
.18254 .21078 L
s
.18254 .21078 m
.17666 .22096 L
s
.17666 .22096 m
.1649 .22096 L
s
.1649 .22096 m
.17078 .23114 L
s
.17078 .23114 m
.1649 .24133 L
s
.1649 .24133 m
.15315 .24133 L
s
.15315 .24133 m
.14727 .23114 L
s
.14727 .23114 m
.14139 .24133 L
s
.14139 .24133 m
.12963 .24133 L
s
.12963 .24133 m
.13551 .25151 L
s
.13551 .25151 m
.12963 .26169 L
s
.12963 .26169 m
.14139 .26169 L
s
.14139 .26169 m
.14727 .27187 L
s
.14727 .27187 m
.14139 .28206 L
s
.14139 .28206 m
.12963 .28206 L
s
.12963 .28206 m
.13551 .29224 L
s
.13551 .29224 m
.12963 .30242 L
s
.12963 .30242 m
.11787 .30242 L
s
.11787 .30242 m
.11199 .29224 L
s
.11199 .29224 m
.10611 .30242 L
s
.10611 .30242 m
.09436 .30242 L
s
.09436 .30242 m
.08848 .29224 L
s
.08848 .29224 m
.09436 .28206 L
s
.09436 .28206 m
.0826 .28206 L
s
.0826 .28206 m
.07672 .27187 L
s
.07672 .27187 m
.07084 .28206 L
s
.07084 .28206 m
.05908 .28206 L
s
.05908 .28206 m
.06496 .29224 L
s
.06496 .29224 m
.05908 .30242 L
s
.05908 .30242 m
.04733 .30242 L
s
.04733 .30242 m
.04145 .29224 L
s
.04145 .29224 m
.03557 .30242 L
s
.03557 .30242 m
.02381 .30242 L
s
MathSubEnd
P
% End of sub-graphic
% Start of sub-graphic
p
0.522676 0.0130918 0.97619 0.536766 MathSubStart
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.087607 0.824786 0.261905 0.824786 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.08761 .2619 m
.5 .97619 L
s
.5 .97619 m
.91239 .2619 L
s
.91239 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.22507 .5 L
s
.22507 .5 m
.08761 .7381 L
s
.08761 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.5 .97619 L
s
.5 .97619 m
.63746 .7381 L
s
.63746 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.77493 .5 L
s
.77493 .5 m
.91239 .2619 L
s
.91239 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.5 .02381 L
s
.5 .02381 m
.36254 .2619 L
s
.36254 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.13343 .34127 L
s
.13343 .34127 m
.08761 .42063 L
s
.08761 .42063 m
.17925 .42063 L
s
.17925 .42063 m
.22507 .5 L
s
.22507 .5 m
.17925 .57937 L
s
.17925 .57937 m
.08761 .57937 L
s
.08761 .57937 m
.13343 .65873 L
s
.13343 .65873 m
.08761 .7381 L
s
.08761 .7381 m
.17925 .7381 L
s
.17925 .7381 m
.22507 .81746 L
s
.22507 .81746 m
.27089 .7381 L
s
.27089 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.40836 .81746 L
s
.40836 .81746 m
.36254 .89683 L
s
.36254 .89683 m
.45418 .89683 L
s
.45418 .89683 m
.5 .97619 L
s
.5 .97619 m
.54582 .89683 L
s
.54582 .89683 m
.63746 .89683 L
s
.63746 .89683 m
.59164 .81746 L
s
.59164 .81746 m
.63746 .7381 L
s
.63746 .7381 m
.72911 .7381 L
s
.72911 .7381 m
.77493 .81746 L
s
.77493 .81746 m
.82075 .7381 L
s
.82075 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.86657 .65873 L
s
.86657 .65873 m
.91239 .57937 L
s
.91239 .57937 m
.82075 .57937 L
s
.82075 .57937 m
.77493 .5 L
s
.77493 .5 m
.82075 .42063 L
s
.82075 .42063 m
.91239 .42063 L
s
.91239 .42063 m
.86657 .34127 L
s
.86657 .34127 m
.91239 .2619 L
s
.91239 .2619 m
.82075 .2619 L
s
.82075 .2619 m
.77493 .18254 L
s
.77493 .18254 m
.72911 .2619 L
s
.72911 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.59164 .18254 L
s
.59164 .18254 m
.63746 .10317 L
s
.63746 .10317 m
.54582 .10317 L
s
.54582 .10317 m
.5 .02381 L
s
.5 .02381 m
.45418 .10317 L
s
.45418 .10317 m
.36254 .10317 L
s
.36254 .10317 m
.40836 .18254 L
s
.40836 .18254 m
.36254 .2619 L
s
.36254 .2619 m
.27089 .2619 L
s
.27089 .2619 m
.22507 .18254 L
s
.22507 .18254 m
.17925 .2619 L
s
.17925 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.10288 .28836 L
s
.10288 .28836 m
.08761 .31481 L
s
.08761 .31481 m
.11815 .31481 L
s
.11815 .31481 m
.13343 .34127 L
s
.13343 .34127 m
.11815 .36772 L
s
.11815 .36772 m
.08761 .36772 L
s
.08761 .36772 m
.10288 .39418 L
s
.10288 .39418 m
.08761 .42063 L
s
.08761 .42063 m
.11815 .42063 L
s
.11815 .42063 m
.13343 .44709 L
s
.13343 .44709 m
.1487 .42063 L
s
.1487 .42063 m
.17925 .42063 L
s
.17925 .42063 m
.19452 .44709 L
s
.19452 .44709 m
.17925 .47354 L
s
.17925 .47354 m
.2098 .47354 L
s
.2098 .47354 m
.22507 .5 L
s
.22507 .5 m
.2098 .52646 L
s
.2098 .52646 m
.17925 .52646 L
s
.17925 .52646 m
.19452 .55291 L
s
.19452 .55291 m
.17925 .57937 L
s
.17925 .57937 m
.1487 .57937 L
s
.1487 .57937 m
.13343 .55291 L
s
.13343 .55291 m
.11815 .57937 L
s
.11815 .57937 m
.08761 .57937 L
s
.08761 .57937 m
.10288 .60582 L
s
.10288 .60582 m
.08761 .63228 L
s
.08761 .63228 m
.11815 .63228 L
s
.11815 .63228 m
.13343 .65873 L
s
.13343 .65873 m
.11815 .68519 L
s
.11815 .68519 m
.08761 .68519 L
s
.08761 .68519 m
.10288 .71164 L
s
.10288 .71164 m
.08761 .7381 L
s
.08761 .7381 m
.11815 .7381 L
s
.11815 .7381 m
.13343 .76455 L
s
.13343 .76455 m
.1487 .7381 L
s
.1487 .7381 m
.17925 .7381 L
s
.17925 .7381 m
.19452 .76455 L
s
.19452 .76455 m
.17925 .79101 L
s
.17925 .79101 m
.2098 .79101 L
s
.2098 .79101 m
.22507 .81746 L
s
.22507 .81746 m
.24035 .79101 L
s
.24035 .79101 m
.27089 .79101 L
s
.27089 .79101 m
.25562 .76455 L
s
.25562 .76455 m
.27089 .7381 L
s
.27089 .7381 m
.30144 .7381 L
s
.30144 .7381 m
.31671 .76455 L
s
.31671 .76455 m
.33199 .7381 L
s
.33199 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.37781 .76455 L
s
.37781 .76455 m
.36254 .79101 L
s
.36254 .79101 m
.39308 .79101 L
s
.39308 .79101 m
.40836 .81746 L
s
.40836 .81746 m
.39308 .84392 L
s
.39308 .84392 m
.36254 .84392 L
s
.36254 .84392 m
.37781 .87037 L
s
.37781 .87037 m
.36254 .89683 L
s
.36254 .89683 m
.39308 .89683 L
s
.39308 .89683 m
.40836 .92328 L
s
.40836 .92328 m
.42363 .89683 L
s
.42363 .89683 m
.45418 .89683 L
s
.45418 .89683 m
.46945 .92328 L
s
.46945 .92328 m
.45418 .94974 L
s
.45418 .94974 m
.48473 .94974 L
s
.48473 .94974 m
.5 .97619 L
s
.5 .97619 m
.51527 .94974 L
s
.51527 .94974 m
.54582 .94974 L
s
.54582 .94974 m
.53055 .92328 L
s
.53055 .92328 m
.54582 .89683 L
s
.54582 .89683 m
.57637 .89683 L
s
.57637 .89683 m
.59164 .92328 L
s
.59164 .92328 m
.60692 .89683 L
s
.60692 .89683 m
.63746 .89683 L
s
.63746 .89683 m
.62219 .87037 L
s
.62219 .87037 m
.63746 .84392 L
s
.63746 .84392 m
.60692 .84392 L
s
.60692 .84392 m
.59164 .81746 L
s
.59164 .81746 m
.60692 .79101 L
s
.60692 .79101 m
.63746 .79101 L
s
.63746 .79101 m
.62219 .76455 L
s
.62219 .76455 m
.63746 .7381 L
s
.63746 .7381 m
.66801 .7381 L
s
.66801 .7381 m
.68329 .76455 L
s
.68329 .76455 m
.69856 .7381 L
s
.69856 .7381 m
.72911 .7381 L
s
.72911 .7381 m
.74438 .76455 L
s
.74438 .76455 m
.72911 .79101 L
s
.72911 .79101 m
.75965 .79101 L
s
.75965 .79101 m
.77493 .81746 L
s
.77493 .81746 m
.7902 .79101 L
s
.7902 .79101 m
.82075 .79101 L
s
.82075 .79101 m
.80548 .76455 L
s
.80548 .76455 m
.82075 .7381 L
s
.82075 .7381 m
.8513 .7381 L
s
.8513 .7381 m
.86657 .76455 L
s
.86657 .76455 m
.88185 .7381 L
s
.88185 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.89712 .71164 L
s
.89712 .71164 m
.91239 .68519 L
s
.91239 .68519 m
.88185 .68519 L
s
.88185 .68519 m
.86657 .65873 L
s
.86657 .65873 m
.88185 .63228 L
s
.88185 .63228 m
.91239 .63228 L
s
.91239 .63228 m
.89712 .60582 L
s
.89712 .60582 m
.91239 .57937 L
s
.91239 .57937 m
.88185 .57937 L
s
.88185 .57937 m
.86657 .55291 L
s
.86657 .55291 m
.8513 .57937 L
s
.8513 .57937 m
.82075 .57937 L
s
.82075 .57937 m
.80548 .55291 L
s
.80548 .55291 m
.82075 .52646 L
s
.82075 .52646 m
.7902 .52646 L
s
.7902 .52646 m
.77493 .5 L
s
.77493 .5 m
.7902 .47354 L
s
.7902 .47354 m
.82075 .47354 L
s
.82075 .47354 m
.80548 .44709 L
s
.80548 .44709 m
.82075 .42063 L
s
.82075 .42063 m
.8513 .42063 L
s
.8513 .42063 m
.86657 .44709 L
s
.86657 .44709 m
.88185 .42063 L
s
.88185 .42063 m
.91239 .42063 L
s
.91239 .42063 m
.89712 .39418 L
s
.89712 .39418 m
.91239 .36772 L
s
.91239 .36772 m
.88185 .36772 L
s
.88185 .36772 m
.86657 .34127 L
s
.86657 .34127 m
.88185 .31481 L
s
.88185 .31481 m
.91239 .31481 L
s
.91239 .31481 m
.89712 .28836 L
s
.89712 .28836 m
.91239 .2619 L
s
.91239 .2619 m
.88185 .2619 L
s
.88185 .2619 m
.86657 .23545 L
s
.86657 .23545 m
.8513 .2619 L
s
.8513 .2619 m
.82075 .2619 L
s
.82075 .2619 m
.80548 .23545 L
s
.80548 .23545 m
.82075 .20899 L
s
.82075 .20899 m
.7902 .20899 L
s
.7902 .20899 m
.77493 .18254 L
s
.77493 .18254 m
.75965 .20899 L
s
.75965 .20899 m
.72911 .20899 L
s
.72911 .20899 m
.74438 .23545 L
s
.74438 .23545 m
.72911 .2619 L
s
.72911 .2619 m
.69856 .2619 L
s
.69856 .2619 m
.68329 .23545 L
s
.68329 .23545 m
.66801 .2619 L
s
.66801 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.62219 .23545 L
s
.62219 .23545 m
.63746 .20899 L
s
.63746 .20899 m
.60692 .20899 L
s
.60692 .20899 m
.59164 .18254 L
s
.59164 .18254 m
.60692 .15608 L
s
.60692 .15608 m
.63746 .15608 L
s
.63746 .15608 m
.62219 .12963 L
s
.62219 .12963 m
.63746 .10317 L
s
.63746 .10317 m
.60692 .10317 L
s
.60692 .10317 m
.59164 .07672 L
s
.59164 .07672 m
.57637 .10317 L
s
.57637 .10317 m
.54582 .10317 L
s
.54582 .10317 m
.53055 .07672 L
s
.53055 .07672 m
.54582 .05026 L
s
.54582 .05026 m
.51527 .05026 L
s
.51527 .05026 m
.5 .02381 L
s
.5 .02381 m
.48473 .05026 L
s
.48473 .05026 m
.45418 .05026 L
s
.45418 .05026 m
.46945 .07672 L
s
.46945 .07672 m
.45418 .10317 L
s
.45418 .10317 m
.42363 .10317 L
s
.42363 .10317 m
.40836 .07672 L
s
.40836 .07672 m
.39308 .10317 L
s
.39308 .10317 m
.36254 .10317 L
s
.36254 .10317 m
.37781 .12963 L
s
.37781 .12963 m
.36254 .15608 L
s
.36254 .15608 m
.39308 .15608 L
s
.39308 .15608 m
.40836 .18254 L
s
.40836 .18254 m
.39308 .20899 L
s
.39308 .20899 m
.36254 .20899 L
s
.36254 .20899 m
.37781 .23545 L
s
.37781 .23545 m
.36254 .2619 L
s
.36254 .2619 m
.33199 .2619 L
s
.33199 .2619 m
.31671 .23545 L
s
.31671 .23545 m
.30144 .2619 L
s
.30144 .2619 m
.27089 .2619 L
s
.27089 .2619 m
.25562 .23545 L
s
.25562 .23545 m
.27089 .20899 L
s
.27089 .20899 m
.24035 .20899 L
s
.24035 .20899 m
.22507 .18254 L
s
.22507 .18254 m
.2098 .20899 L
s
.2098 .20899 m
.17925 .20899 L
s
.17925 .20899 m
.19452 .23545 L
s
.19452 .23545 m
.17925 .2619 L
s
.17925 .2619 m
.1487 .2619 L
s
.1487 .2619 m
.13343 .23545 L
s
.13343 .23545 m
.11815 .2619 L
s
.11815 .2619 m
.08761 .2619 L
s
.08761 .2619 m
.0927 .27072 L
s
.0927 .27072 m
.08761 .27954 L
s
.08761 .27954 m
.09779 .27954 L
s
.09779 .27954 m
.10288 .28836 L
s
.10288 .28836 m
.09779 .29718 L
s
.09779 .29718 m
.08761 .29718 L
s
.08761 .29718 m
.0927 .306 L
s
.0927 .306 m
.08761 .31481 L
s
.08761 .31481 m
.09779 .31481 L
s
.09779 .31481 m
.10288 .32363 L
s
.10288 .32363 m
.10797 .31481 L
s
.10797 .31481 m
.11815 .31481 L
s
.11815 .31481 m
.12325 .32363 L
s
.12325 .32363 m
.11815 .33245 L
s
.11815 .33245 m
.12834 .33245 L
s
.12834 .33245 m
.13343 .34127 L
s
.13343 .34127 m
.12834 .35009 L
s
.12834 .35009 m
.11815 .35009 L
s
.11815 .35009 m
.12325 .35891 L
s
.12325 .35891 m
.11815 .36772 L
s
.11815 .36772 m
.10797 .36772 L
s
.10797 .36772 m
.10288 .35891 L
s
.10288 .35891 m
.09779 .36772 L
s
.09779 .36772 m
.08761 .36772 L
s
.08761 .36772 m
.0927 .37654 L
s
.0927 .37654 m
.08761 .38536 L
s
.08761 .38536 m
.09779 .38536 L
s
.09779 .38536 m
.10288 .39418 L
s
.10288 .39418 m
.09779 .403 L
s
.09779 .403 m
.08761 .403 L
s
.08761 .403 m
.0927 .41182 L
s
.0927 .41182 m
.08761 .42063 L
s
.08761 .42063 m
.09779 .42063 L
s
.09779 .42063 m
.10288 .42945 L
s
.10288 .42945 m
.10797 .42063 L
s
.10797 .42063 m
.11815 .42063 L
s
.11815 .42063 m
.12325 .42945 L
s
.12325 .42945 m
.11815 .43827 L
s
.11815 .43827 m
.12834 .43827 L
s
.12834 .43827 m
.13343 .44709 L
s
.13343 .44709 m
.13852 .43827 L
s
.13852 .43827 m
.1487 .43827 L
s
.1487 .43827 m
.14361 .42945 L
s
.14361 .42945 m
.1487 .42063 L
s
.1487 .42063 m
.15888 .42063 L
s
.15888 .42063 m
.16398 .42945 L
s
.16398 .42945 m
.16907 .42063 L
s
.16907 .42063 m
.17925 .42063 L
s
.17925 .42063 m
.18434 .42945 L
s
.18434 .42945 m
.17925 .43827 L
s
.17925 .43827 m
.18943 .43827 L
s
.18943 .43827 m
.19452 .44709 L
s
.19452 .44709 m
.18943 .45591 L
s
.18943 .45591 m
.17925 .45591 L
s
.17925 .45591 m
.18434 .46473 L
s
.18434 .46473 m
.17925 .47354 L
s
.17925 .47354 m
.18943 .47354 L
s
.18943 .47354 m
.19452 .48236 L
s
.19452 .48236 m
.19961 .47354 L
s
.19961 .47354 m
.2098 .47354 L
s
.2098 .47354 m
.21489 .48236 L
s
.21489 .48236 m
.2098 .49118 L
s
.2098 .49118 m
.21998 .49118 L
s
.21998 .49118 m
.22507 .5 L
s
.22507 .5 m
.21998 .50882 L
s
.21998 .50882 m
.2098 .50882 L
s
.2098 .50882 m
.21489 .51764 L
s
.21489 .51764 m
.2098 .52646 L
s
.2098 .52646 m
.19961 .52646 L
s
.19961 .52646 m
.19452 .51764 L
s
.19452 .51764 m
.18943 .52646 L
s
.18943 .52646 m
.17925 .52646 L
s
.17925 .52646 m
.18434 .53527 L
s
.18434 .53527 m
.17925 .54409 L
s
.17925 .54409 m
.18943 .54409 L
s
.18943 .54409 m
.19452 .55291 L
s
.19452 .55291 m
.18943 .56173 L
s
.18943 .56173 m
.17925 .56173 L
s
.17925 .56173 m
.18434 .57055 L
s
.18434 .57055 m
.17925 .57937 L
s
.17925 .57937 m
.16907 .57937 L
s
.16907 .57937 m
.16398 .57055 L
s
.16398 .57055 m
.15888 .57937 L
s
.15888 .57937 m
.1487 .57937 L
s
.1487 .57937 m
.14361 .57055 L
s
.14361 .57055 m
.1487 .56173 L
s
.1487 .56173 m
.13852 .56173 L
s
.13852 .56173 m
.13343 .55291 L
s
.13343 .55291 m
.12834 .56173 L
s
.12834 .56173 m
.11815 .56173 L
s
.11815 .56173 m
.12325 .57055 L
s
.12325 .57055 m
.11815 .57937 L
s
.11815 .57937 m
.10797 .57937 L
s
.10797 .57937 m
.10288 .57055 L
s
.10288 .57055 m
.09779 .57937 L
s
.09779 .57937 m
.08761 .57937 L
s
.08761 .57937 m
.0927 .58818 L
s
.0927 .58818 m
.08761 .597 L
s
.08761 .597 m
.09779 .597 L
s
.09779 .597 m
.10288 .60582 L
s
.10288 .60582 m
.09779 .61464 L
s
.09779 .61464 m
.08761 .61464 L
s
.08761 .61464 m
.0927 .62346 L
s
.0927 .62346 m
.08761 .63228 L
s
.08761 .63228 m
.09779 .63228 L
s
.09779 .63228 m
.10288 .64109 L
s
.10288 .64109 m
.10797 .63228 L
s
.10797 .63228 m
.11815 .63228 L
s
.11815 .63228 m
.12325 .64109 L
s
.12325 .64109 m
.11815 .64991 L
s
.11815 .64991 m
.12834 .64991 L
s
.12834 .64991 m
.13343 .65873 L
s
.13343 .65873 m
.12834 .66755 L
s
.12834 .66755 m
.11815 .66755 L
s
.11815 .66755 m
.12325 .67637 L
s
.12325 .67637 m
.11815 .68519 L
s
.11815 .68519 m
.10797 .68519 L
s
.10797 .68519 m
.10288 .67637 L
s
.10288 .67637 m
.09779 .68519 L
s
.09779 .68519 m
.08761 .68519 L
s
.08761 .68519 m
.0927 .694 L
s
.0927 .694 m
.08761 .70282 L
s
.08761 .70282 m
.09779 .70282 L
s
.09779 .70282 m
.10288 .71164 L
s
.10288 .71164 m
.09779 .72046 L
s
.09779 .72046 m
.08761 .72046 L
s
.08761 .72046 m
.0927 .72928 L
s
.0927 .72928 m
.08761 .7381 L
s
.08761 .7381 m
.09779 .7381 L
s
.09779 .7381 m
.10288 .74691 L
s
.10288 .74691 m
.10797 .7381 L
s
.10797 .7381 m
.11815 .7381 L
s
.11815 .7381 m
.12325 .74691 L
s
.12325 .74691 m
.11815 .75573 L
s
.11815 .75573 m
.12834 .75573 L
s
.12834 .75573 m
.13343 .76455 L
s
.13343 .76455 m
.13852 .75573 L
s
.13852 .75573 m
.1487 .75573 L
s
.1487 .75573 m
.14361 .74691 L
s
.14361 .74691 m
.1487 .7381 L
s
.1487 .7381 m
.15888 .7381 L
s
.15888 .7381 m
.16398 .74691 L
s
.16398 .74691 m
.16907 .7381 L
s
.16907 .7381 m
.17925 .7381 L
s
.17925 .7381 m
.18434 .74691 L
s
.18434 .74691 m
.17925 .75573 L
s
.17925 .75573 m
.18943 .75573 L
s
.18943 .75573 m
.19452 .76455 L
s
.19452 .76455 m
.18943 .77337 L
s
.18943 .77337 m
.17925 .77337 L
s
.17925 .77337 m
.18434 .78219 L
s
.18434 .78219 m
.17925 .79101 L
s
.17925 .79101 m
.18943 .79101 L
s
.18943 .79101 m
.19452 .79982 L
s
.19452 .79982 m
.19961 .79101 L
s
.19961 .79101 m
.2098 .79101 L
s
.2098 .79101 m
.21489 .79982 L
s
.21489 .79982 m
.2098 .80864 L
s
.2098 .80864 m
.21998 .80864 L
s
.21998 .80864 m
.22507 .81746 L
s
.22507 .81746 m
.23016 .80864 L
s
.23016 .80864 m
.24035 .80864 L
s
.24035 .80864 m
.23525 .79982 L
s
.23525 .79982 m
.24035 .79101 L
s
.24035 .79101 m
.25053 .79101 L
s
.25053 .79101 m
.25562 .79982 L
s
.25562 .79982 m
.26071 .79101 L
s
.26071 .79101 m
.27089 .79101 L
s
.27089 .79101 m
.2658 .78219 L
s
.2658 .78219 m
.27089 .77337 L
s
.27089 .77337 m
.26071 .77337 L
s
.26071 .77337 m
.25562 .76455 L
s
.25562 .76455 m
.26071 .75573 L
s
.26071 .75573 m
.27089 .75573 L
s
.27089 .75573 m
.2658 .74691 L
s
.2658 .74691 m
.27089 .7381 L
s
.27089 .7381 m
.28108 .7381 L
s
.28108 .7381 m
.28617 .74691 L
s
.28617 .74691 m
.29126 .7381 L
s
.29126 .7381 m
.30144 .7381 L
s
.30144 .7381 m
.30653 .74691 L
s
.30653 .74691 m
.30144 .75573 L
s
.30144 .75573 m
.31162 .75573 L
s
.31162 .75573 m
.31671 .76455 L
s
.31671 .76455 m
.32181 .75573 L
s
.32181 .75573 m
.33199 .75573 L
s
.33199 .75573 m
.3269 .74691 L
s
.3269 .74691 m
.33199 .7381 L
s
.33199 .7381 m
.34217 .7381 L
s
.34217 .7381 m
.34726 .74691 L
s
.34726 .74691 m
.35235 .7381 L
s
.35235 .7381 m
.36254 .7381 L
s
.36254 .7381 m
.36763 .74691 L
s
.36763 .74691 m
.36254 .75573 L
s
.36254 .75573 m
.37272 .75573 L
s
.37272 .75573 m
.37781 .76455 L
s
.37781 .76455 m
.37272 .77337 L
s
.37272 .77337 m
.36254 .77337 L
s
.36254 .77337 m
.36763 .78219 L
s
.36763 .78219 m
.36254 .79101 L
s
.36254 .79101 m
.37272 .79101 L
s
.37272 .79101 m
.37781 .79982 L
s
.37781 .79982 m
.3829 .79101 L
s
.3829 .79101 m
.39308 .79101 L
s
.39308 .79101 m
.39817 .79982 L
s
.39817 .79982 m
.39308 .80864 L
s
.39308 .80864 m
.40327 .80864 L
s
.40327 .80864 m
.40836 .81746 L
s
.40836 .81746 m
.40327 .82628 L
s
.40327 .82628 m
.39308 .82628 L
s
.39308 .82628 m
.39817 .8351 L
s
.39817 .8351 m
.39308 .84392 L
s
.39308 .84392 m
.3829 .84392 L
s
.3829 .84392 m
.37781 .8351 L
s
.37781 .8351 m
.37272 .84392 L
s
.37272 .84392 m
.36254 .84392 L
s
.36254 .84392 m
.36763 .85273 L
s
.36763 .85273 m
.36254 .86155 L
s
.36254 .86155 m
.37272 .86155 L
s
.37272 .86155 m
.37781 .87037 L
s
.37781 .87037 m
.37272 .87919 L
s
.37272 .87919 m
.36254 .87919 L
s
.36254 .87919 m
.36763 .88801 L
s
.36763 .88801 m
.36254 .89683 L
s
.36254 .89683 m
.37272 .89683 L
s
.37272 .89683 m
.37781 .90564 L
s
.37781 .90564 m
.3829 .89683 L
s
.3829 .89683 m
.39308 .89683 L
s
.39308 .89683 m
.39817 .90564 L
s
.39817 .90564 m
.39308 .91446 L
s
.39308 .91446 m
.40327 .91446 L
s
.40327 .91446 m
.40836 .92328 L
s
.40836 .92328 m
.41345 .91446 L
s
.41345 .91446 m
.42363 .91446 L
s
.42363 .91446 m
.41854 .90564 L
s
.41854 .90564 m
.42363 .89683 L
s
.42363 .89683 m
.43381 .89683 L
s
.43381 .89683 m
.4389 .90564 L
s
.4389 .90564 m
.444 .89683 L
s
.444 .89683 m
.45418 .89683 L
s
.45418 .89683 m
.45927 .90564 L
s
.45927 .90564 m
.45418 .91446 L
s
.45418 .91446 m
.46436 .91446 L
s
.46436 .91446 m
.46945 .92328 L
s
.46945 .92328 m
.46436 .9321 L
s
.46436 .9321 m
.45418 .9321 L
s
.45418 .9321 m
.45927 .94092 L
s
.45927 .94092 m
.45418 .94974 L
s
.45418 .94974 m
.46436 .94974 L
s
.46436 .94974 m
.46945 .95855 L
s
.46945 .95855 m
.47454 .94974 L
s
.47454 .94974 m
.48473 .94974 L
s
.48473 .94974 m
.48982 .95855 L
s
.48982 .95855 m
.48473 .96737 L
s
.48473 .96737 m
.49491 .96737 L
s
.49491 .96737 m
.5 .97619 L
s
.5 .97619 m
.50509 .96737 L
s
.50509 .96737 m
.51527 .96737 L
s
.51527 .96737 m
.51018 .95855 L
s
.51018 .95855 m
.51527 .94974 L
s
.51527 .94974 m
.52546 .94974 L
s
.52546 .94974 m
.53055 .95855 L
s
.53055 .95855 m
.53564 .94974 L
s
.53564 .94974 m
.54582 .94974 L
s
.54582 .94974 m
.54073 .94092 L
s
.54073 .94092 m
.54582 .9321 L
s
.54582 .9321 m
.53564 .9321 L
s
.53564 .9321 m
.53055 .92328 L
s
.53055 .92328 m
.53564 .91446 L
s
.53564 .91446 m
.54582 .91446 L
s
.54582 .91446 m
.54073 .90564 L
s
.54073 .90564 m
.54582 .89683 L
s
.54582 .89683 m
.556 .89683 L
s
.556 .89683 m
.5611 .90564 L
s
.5611 .90564 m
.56619 .89683 L
s
.56619 .89683 m
.57637 .89683 L
s
.57637 .89683 m
.58146 .90564 L
s
.58146 .90564 m
.57637 .91446 L
s
.57637 .91446 m
.58655 .91446 L
s
.58655 .91446 m
.59164 .92328 L
s
.59164 .92328 m
.59673 .91446 L
s
.59673 .91446 m
.60692 .91446 L
s
.60692 .91446 m
.60183 .90564 L
s
.60183 .90564 m
.60692 .89683 L
s
.60692 .89683 m
.6171 .89683 L
s
.6171 .89683 m
.62219 .90564 L
s
.62219 .90564 m
.62728 .89683 L
s
.62728 .89683 m
.63746 .89683 L
s
.63746 .89683 m
.63237 .88801 L
s
.63237 .88801 m
.63746 .87919 L
s
.63746 .87919 m
.62728 .87919 L
s
.62728 .87919 m
.62219 .87037 L
s
.62219 .87037 m
.62728 .86155 L
s
.62728 .86155 m
.63746 .86155 L
s
.63746 .86155 m
.63237 .85273 L
s
.63237 .85273 m
.63746 .84392 L
s
.63746 .84392 m
.62728 .84392 L
s
.62728 .84392 m
.62219 .8351 L
s
.62219 .8351 m
.6171 .84392 L
s
.6171 .84392 m
.60692 .84392 L
s
.60692 .84392 m
.60183 .8351 L
s
.60183 .8351 m
.60692 .82628 L
s
.60692 .82628 m
.59673 .82628 L
s
.59673 .82628 m
.59164 .81746 L
s
.59164 .81746 m
.59673 .80864 L
s
.59673 .80864 m
.60692 .80864 L
s
.60692 .80864 m
.60183 .79982 L
s
.60183 .79982 m
.60692 .79101 L
s
.60692 .79101 m
.6171 .79101 L
s
.6171 .79101 m
.62219 .79982 L
s
.62219 .79982 m
.62728 .79101 L
s
.62728 .79101 m
.63746 .79101 L
s
.63746 .79101 m
.63237 .78219 L
s
.63237 .78219 m
.63746 .77337 L
s
.63746 .77337 m
.62728 .77337 L
s
.62728 .77337 m
.62219 .76455 L
s
.62219 .76455 m
.62728 .75573 L
s
.62728 .75573 m
.63746 .75573 L
s
.63746 .75573 m
.63237 .74691 L
s
.63237 .74691 m
.63746 .7381 L
s
.63746 .7381 m
.64765 .7381 L
s
.64765 .7381 m
.65274 .74691 L
s
.65274 .74691 m
.65783 .7381 L
s
.65783 .7381 m
.66801 .7381 L
s
.66801 .7381 m
.6731 .74691 L
s
.6731 .74691 m
.66801 .75573 L
s
.66801 .75573 m
.67819 .75573 L
s
.67819 .75573 m
.68329 .76455 L
s
.68329 .76455 m
.68838 .75573 L
s
.68838 .75573 m
.69856 .75573 L
s
.69856 .75573 m
.69347 .74691 L
s
.69347 .74691 m
.69856 .7381 L
s
.69856 .7381 m
.70874 .7381 L
s
.70874 .7381 m
.71383 .74691 L
s
.71383 .74691 m
.71892 .7381 L
s
.71892 .7381 m
.72911 .7381 L
s
.72911 .7381 m
.7342 .74691 L
s
.7342 .74691 m
.72911 .75573 L
s
.72911 .75573 m
.73929 .75573 L
s
.73929 .75573 m
.74438 .76455 L
s
.74438 .76455 m
.73929 .77337 L
s
.73929 .77337 m
.72911 .77337 L
s
.72911 .77337 m
.7342 .78219 L
s
.7342 .78219 m
.72911 .79101 L
s
.72911 .79101 m
.73929 .79101 L
s
.73929 .79101 m
.74438 .79982 L
s
.74438 .79982 m
.74947 .79101 L
s
.74947 .79101 m
.75965 .79101 L
s
.75965 .79101 m
.76475 .79982 L
s
.76475 .79982 m
.75965 .80864 L
s
.75965 .80864 m
.76984 .80864 L
s
.76984 .80864 m
.77493 .81746 L
s
.77493 .81746 m
.78002 .80864 L
s
.78002 .80864 m
.7902 .80864 L
s
.7902 .80864 m
.78511 .79982 L
s
.78511 .79982 m
.7902 .79101 L
s
.7902 .79101 m
.80039 .79101 L
s
.80039 .79101 m
.80548 .79982 L
s
.80548 .79982 m
.81057 .79101 L
s
.81057 .79101 m
.82075 .79101 L
s
.82075 .79101 m
.81566 .78219 L
s
.81566 .78219 m
.82075 .77337 L
s
.82075 .77337 m
.81057 .77337 L
s
.81057 .77337 m
.80548 .76455 L
s
.80548 .76455 m
.81057 .75573 L
s
.81057 .75573 m
.82075 .75573 L
s
.82075 .75573 m
.81566 .74691 L
s
.81566 .74691 m
.82075 .7381 L
s
.82075 .7381 m
.83093 .7381 L
s
.83093 .7381 m
.83602 .74691 L
s
.83602 .74691 m
.84112 .7381 L
s
.84112 .7381 m
.8513 .7381 L
s
.8513 .7381 m
.85639 .74691 L
s
.85639 .74691 m
.8513 .75573 L
s
.8513 .75573 m
.86148 .75573 L
s
.86148 .75573 m
.86657 .76455 L
s
.86657 .76455 m
.87166 .75573 L
s
.87166 .75573 m
.88185 .75573 L
s
.88185 .75573 m
.87675 .74691 L
s
.87675 .74691 m
.88185 .7381 L
s
.88185 .7381 m
.89203 .7381 L
s
.89203 .7381 m
.89712 .74691 L
s
.89712 .74691 m
.90221 .7381 L
s
.90221 .7381 m
.91239 .7381 L
s
.91239 .7381 m
.9073 .72928 L
s
.9073 .72928 m
.91239 .72046 L
s
.91239 .72046 m
.90221 .72046 L
s
.90221 .72046 m
.89712 .71164 L
s
.89712 .71164 m
.90221 .70282 L
s
.90221 .70282 m
.91239 .70282 L
s
.91239 .70282 m
.9073 .694 L
s
.9073 .694 m
.91239 .68519 L
s
.91239 .68519 m
.90221 .68519 L
s
.90221 .68519 m
.89712 .67637 L
s
.89712 .67637 m
.89203 .68519 L
s
.89203 .68519 m
.88185 .68519 L
s
.88185 .68519 m
.87675 .67637 L
s
.87675 .67637 m
.88185 .66755 L
s
.88185 .66755 m
.87166 .66755 L
s
.87166 .66755 m
.86657 .65873 L
s
.86657 .65873 m
.87166 .64991 L
s
.87166 .64991 m
.88185 .64991 L
s
.88185 .64991 m
.87675 .64109 L
s
.87675 .64109 m
.88185 .63228 L
s
.88185 .63228 m
.89203 .63228 L
s
.89203 .63228 m
.89712 .64109 L
s
.89712 .64109 m
.90221 .63228 L
s
.90221 .63228 m
.91239 .63228 L
s
.91239 .63228 m
.9073 .62346 L
s
.9073 .62346 m
.91239 .61464 L
s
.91239 .61464 m
.90221 .61464 L
s
.90221 .61464 m
.89712 .60582 L
s
.89712 .60582 m
.90221 .597 L
s
.90221 .597 m
.91239 .597 L
s
.91239 .597 m
.9073 .58818 L
s
.9073 .58818 m
.91239 .57937 L
s
.91239 .57937 m
.90221 .57937 L
s
.90221 .57937 m
.89712 .57055 L
s
.89712 .57055 m
.89203 .57937 L
s
.89203 .57937 m
.88185 .57937 L
s
.88185 .57937 m
.87675 .57055 L
s
.87675 .57055 m
.88185 .56173 L
s
.88185 .56173 m
.87166 .56173 L
s
.87166 .56173 m
.86657 .55291 L
s
.86657 .55291 m
.86148 .56173 L
s
.86148 .56173 m
.8513 .56173 L
s
.8513 .56173 m
.85639 .57055 L
s
.85639 .57055 m
.8513 .57937 L
s
.8513 .57937 m
.84112 .57937 L
s
.84112 .57937 m
.83602 .57055 L
s
.83602 .57055 m
.83093 .57937 L
s
.83093 .57937 m
.82075 .57937 L
s
.82075 .57937 m
.81566 .57055 L
s
.81566 .57055 m
.82075 .56173 L
s
.82075 .56173 m
.81057 .56173 L
s
.81057 .56173 m
.80548 .55291 L
s
.80548 .55291 m
.81057 .54409 L
s
.81057 .54409 m
.82075 .54409 L
s
.82075 .54409 m
.81566 .53527 L
s
.81566 .53527 m
.82075 .52646 L
s
.82075 .52646 m
.81057 .52646 L
s
.81057 .52646 m
.80548 .51764 L
s
.80548 .51764 m
.80039 .52646 L
s
.80039 .52646 m
.7902 .52646 L
s
.7902 .52646 m
.78511 .51764 L
s
.78511 .51764 m
.7902 .50882 L
s
.7902 .50882 m
.78002 .50882 L
s
.78002 .50882 m
.77493 .5 L
s
.77493 .5 m
.78002 .49118 L
s
.78002 .49118 m
.7902 .49118 L
s
.7902 .49118 m
.78511 .48236 L
s
.78511 .48236 m
.7902 .47354 L
s
.7902 .47354 m
.80039 .47354 L
s
.80039 .47354 m
.80548 .48236 L
s
.80548 .48236 m
.81057 .47354 L
s
.81057 .47354 m
.82075 .47354 L
s
.82075 .47354 m
.81566 .46473 L
s
.81566 .46473 m
.82075 .45591 L
s
.82075 .45591 m
.81057 .45591 L
s
.81057 .45591 m
.80548 .44709 L
s
.80548 .44709 m
.81057 .43827 L
s
.81057 .43827 m
.82075 .43827 L
s
.82075 .43827 m
.81566 .42945 L
s
.81566 .42945 m
.82075 .42063 L
s
.82075 .42063 m
.83093 .42063 L
s
.83093 .42063 m
.83602 .42945 L
s
.83602 .42945 m
.84112 .42063 L
s
.84112 .42063 m
.8513 .42063 L
s
.8513 .42063 m
.85639 .42945 L
s
.85639 .42945 m
.8513 .43827 L
s
.8513 .43827 m
.86148 .43827 L
s
.86148 .43827 m
.86657 .44709 L
s
.86657 .44709 m
.87166 .43827 L
s
.87166 .43827 m
.88185 .43827 L
s
.88185 .43827 m
.87675 .42945 L
s
.87675 .42945 m
.88185 .42063 L
s
.88185 .42063 m
.89203 .42063 L
s
.89203 .42063 m
.89712 .42945 L
s
.89712 .42945 m
.90221 .42063 L
s
.90221 .42063 m
.91239 .42063 L
s
.91239 .42063 m
.9073 .41182 L
s
.9073 .41182 m
.91239 .403 L
s
.91239 .403 m
.90221 .403 L
s
.90221 .403 m
.89712 .39418 L
s
.89712 .39418 m
.90221 .38536 L
s
.90221 .38536 m
.91239 .38536 L
s
.91239 .38536 m
.9073 .37654 L
s
.9073 .37654 m
.91239 .36772 L
s
.91239 .36772 m
.90221 .36772 L
s
.90221 .36772 m
.89712 .35891 L
s
.89712 .35891 m
.89203 .36772 L
s
.89203 .36772 m
.88185 .36772 L
s
.88185 .36772 m
.87675 .35891 L
s
.87675 .35891 m
.88185 .35009 L
s
.88185 .35009 m
.87166 .35009 L
s
.87166 .35009 m
.86657 .34127 L
s
.86657 .34127 m
.87166 .33245 L
s
.87166 .33245 m
.88185 .33245 L
s
.88185 .33245 m
.87675 .32363 L
s
.87675 .32363 m
.88185 .31481 L
s
.88185 .31481 m
.89203 .31481 L
s
.89203 .31481 m
.89712 .32363 L
s
.89712 .32363 m
.90221 .31481 L
s
.90221 .31481 m
.91239 .31481 L
s
.91239 .31481 m
.9073 .306 L
s
.9073 .306 m
.91239 .29718 L
s
.91239 .29718 m
.90221 .29718 L
s
.90221 .29718 m
.89712 .28836 L
s
.89712 .28836 m
.90221 .27954 L
s
.90221 .27954 m
.91239 .27954 L
s
.91239 .27954 m
.9073 .27072 L
s
.9073 .27072 m
.91239 .2619 L
s
.91239 .2619 m
.90221 .2619 L
s
.90221 .2619 m
.89712 .25309 L
s
.89712 .25309 m
.89203 .2619 L
s
.89203 .2619 m
.88185 .2619 L
s
.88185 .2619 m
.87675 .25309 L
s
.87675 .25309 m
.88185 .24427 L
s
.88185 .24427 m
.87166 .24427 L
s
.87166 .24427 m
.86657 .23545 L
s
.86657 .23545 m
.86148 .24427 L
s
.86148 .24427 m
.8513 .24427 L
s
.8513 .24427 m
.85639 .25309 L
s
.85639 .25309 m
.8513 .2619 L
s
.8513 .2619 m
.84112 .2619 L
s
.84112 .2619 m
.83602 .25309 L
s
.83602 .25309 m
.83093 .2619 L
s
.83093 .2619 m
.82075 .2619 L
s
.82075 .2619 m
.81566 .25309 L
s
.81566 .25309 m
.82075 .24427 L
s
.82075 .24427 m
.81057 .24427 L
s
.81057 .24427 m
.80548 .23545 L
s
.80548 .23545 m
.81057 .22663 L
s
.81057 .22663 m
.82075 .22663 L
s
.82075 .22663 m
.81566 .21781 L
s
.81566 .21781 m
.82075 .20899 L
s
.82075 .20899 m
.81057 .20899 L
s
.81057 .20899 m
.80548 .20018 L
s
.80548 .20018 m
.80039 .20899 L
s
.80039 .20899 m
.7902 .20899 L
s
.7902 .20899 m
.78511 .20018 L
s
.78511 .20018 m
.7902 .19136 L
s
.7902 .19136 m
.78002 .19136 L
s
.78002 .19136 m
.77493 .18254 L
s
.77493 .18254 m
.76984 .19136 L
s
.76984 .19136 m
.75965 .19136 L
s
.75965 .19136 m
.76475 .20018 L
s
.76475 .20018 m
.75965 .20899 L
s
.75965 .20899 m
.74947 .20899 L
s
.74947 .20899 m
.74438 .20018 L
s
.74438 .20018 m
.73929 .20899 L
s
.73929 .20899 m
.72911 .20899 L
s
.72911 .20899 m
.7342 .21781 L
s
.7342 .21781 m
.72911 .22663 L
s
.72911 .22663 m
.73929 .22663 L
s
.73929 .22663 m
.74438 .23545 L
s
.74438 .23545 m
.73929 .24427 L
s
.73929 .24427 m
.72911 .24427 L
s
.72911 .24427 m
.7342 .25309 L
s
.7342 .25309 m
.72911 .2619 L
s
.72911 .2619 m
.71892 .2619 L
s
.71892 .2619 m
.71383 .25309 L
s
.71383 .25309 m
.70874 .2619 L
s
.70874 .2619 m
.69856 .2619 L
s
.69856 .2619 m
.69347 .25309 L
s
.69347 .25309 m
.69856 .24427 L
s
.69856 .24427 m
.68838 .24427 L
s
.68838 .24427 m
.68329 .23545 L
s
.68329 .23545 m
.67819 .24427 L
s
.67819 .24427 m
.66801 .24427 L
s
.66801 .24427 m
.6731 .25309 L
s
.6731 .25309 m
.66801 .2619 L
s
.66801 .2619 m
.65783 .2619 L
s
.65783 .2619 m
.65274 .25309 L
s
.65274 .25309 m
.64765 .2619 L
s
.64765 .2619 m
.63746 .2619 L
s
.63746 .2619 m
.63237 .25309 L
s
.63237 .25309 m
.63746 .24427 L
s
.63746 .24427 m
.62728 .24427 L
s
.62728 .24427 m
.62219 .23545 L
s
.62219 .23545 m
.62728 .22663 L
s
.62728 .22663 m
.63746 .22663 L
s
.63746 .22663 m
.63237 .21781 L
s
.63237 .21781 m
.63746 .20899 L
s
.63746 .20899 m
.62728 .20899 L
s
.62728 .20899 m
.62219 .20018 L
s
.62219 .20018 m
.6171 .20899 L
s
.6171 .20899 m
.60692 .20899 L
s
.60692 .20899 m
.60183 .20018 L
s
.60183 .20018 m
.60692 .19136 L
s
.60692 .19136 m
.59673 .19136 L
s
.59673 .19136 m
.59164 .18254 L
s
.59164 .18254 m
.59673 .17372 L
s
.59673 .17372 m
.60692 .17372 L
s
.60692 .17372 m
.60183 .1649 L
s
.60183 .1649 m
.60692 .15608 L
s
.60692 .15608 m
.6171 .15608 L
s
.6171 .15608 m
.62219 .1649 L
s
.62219 .1649 m
.62728 .15608 L
s
.62728 .15608 m
.63746 .15608 L
s
.63746 .15608 m
.63237 .14727 L
s
.63237 .14727 m
.63746 .13845 L
s
.63746 .13845 m
.62728 .13845 L
s
.62728 .13845 m
.62219 .12963 L
s
.62219 .12963 m
.62728 .12081 L
s
.62728 .12081 m
.63746 .12081 L
s
.63746 .12081 m
.63237 .11199 L
s
.63237 .11199 m
.63746 .10317 L
s
.63746 .10317 m
.62728 .10317 L
s
.62728 .10317 m
.62219 .09436 L
s
.62219 .09436 m
.6171 .10317 L
s
.6171 .10317 m
.60692 .10317 L
s
.60692 .10317 m
.60183 .09436 L
s
.60183 .09436 m
.60692 .08554 L
s
.60692 .08554 m
.59673 .08554 L
s
.59673 .08554 m
.59164 .07672 L
s
.59164 .07672 m
.58655 .08554 L
s
.58655 .08554 m
.57637 .08554 L
s
.57637 .08554 m
.58146 .09436 L
s
.58146 .09436 m
.57637 .10317 L
s
.57637 .10317 m
.56619 .10317 L
s
.56619 .10317 m
.5611 .09436 L
s
.5611 .09436 m
.556 .10317 L
s
.556 .10317 m
.54582 .10317 L
s
.54582 .10317 m
.54073 .09436 L
s
.54073 .09436 m
.54582 .08554 L
s
.54582 .08554 m
.53564 .08554 L
s
.53564 .08554 m
.53055 .07672 L
s
.53055 .07672 m
.53564 .0679 L
s
.53564 .0679 m
.54582 .0679 L
s
.54582 .0679 m
.54073 .05908 L
s
.54073 .05908 m
.54582 .05026 L
s
.54582 .05026 m
.53564 .05026 L
s
.53564 .05026 m
.53055 .04145 L
s
.53055 .04145 m
.52546 .05026 L
s
.52546 .05026 m
.51527 .05026 L
s
.51527 .05026 m
.51018 .04145 L
s
.51018 .04145 m
.51527 .03263 L
s
.51527 .03263 m
.50509 .03263 L
s
.50509 .03263 m
.5 .02381 L
s
.5 .02381 m
.49491 .03263 L
s
.49491 .03263 m
.48473 .03263 L
s
.48473 .03263 m
.48982 .04145 L
s
.48982 .04145 m
.48473 .05026 L
s
.48473 .05026 m
.47454 .05026 L
s
.47454 .05026 m
.46945 .04145 L
s
.46945 .04145 m
.46436 .05026 L
s
.46436 .05026 m
.45418 .05026 L
s
.45418 .05026 m
.45927 .05908 L
s
.45927 .05908 m
.45418 .0679 L
s
.45418 .0679 m
.46436 .0679 L
s
.46436 .0679 m
.46945 .07672 L
s
.46945 .07672 m
.46436 .08554 L
s
.46436 .08554 m
.45418 .08554 L
s
.45418 .08554 m
.45927 .09436 L
s
.45927 .09436 m
.45418 .10317 L
s
.45418 .10317 m
.444 .10317 L
s
.444 .10317 m
.4389 .09436 L
s
.4389 .09436 m
.43381 .10317 L
s
.43381 .10317 m
.42363 .10317 L
s
.42363 .10317 m
.41854 .09436 L
s
.41854 .09436 m
.42363 .08554 L
s
.42363 .08554 m
.41345 .08554 L
s
.41345 .08554 m
.40836 .07672 L
s
.40836 .07672 m
.40327 .08554 L
s
.40327 .08554 m
.39308 .08554 L
s
.39308 .08554 m
.39817 .09436 L
s
.39817 .09436 m
.39308 .10317 L
s
.39308 .10317 m
.3829 .10317 L
s
.3829 .10317 m
.37781 .09436 L
s
.37781 .09436 m
.37272 .10317 L
s
.37272 .10317 m
.36254 .10317 L
s
.36254 .10317 m
.36763 .11199 L
s
.36763 .11199 m
.36254 .12081 L
s
.36254 .12081 m
.37272 .12081 L
s
.37272 .12081 m
.37781 .12963 L
s
.37781 .12963 m
.37272 .13845 L
s
.37272 .13845 m
.36254 .13845 L
s
.36254 .13845 m
.36763 .14727 L
s
.36763 .14727 m
.36254 .15608 L
s
.36254 .15608 m
.37272 .15608 L
s
.37272 .15608 m
.37781 .1649 L
s
.37781 .1649 m
.3829 .15608 L
s
.3829 .15608 m
.39308 .15608 L
s
.39308 .15608 m
.39817 .1649 L
s
.39817 .1649 m
.39308 .17372 L
s
.39308 .17372 m
.40327 .17372 L
s
.40327 .17372 m
.40836 .18254 L
s
.40836 .18254 m
.40327 .19136 L
s
.40327 .19136 m
.39308 .19136 L
s
.39308 .19136 m
.39817 .20018 L
s
.39817 .20018 m
.39308 .20899 L
s
.39308 .20899 m
.3829 .20899 L
s
.3829 .20899 m
.37781 .20018 L
s
.37781 .20018 m
.37272 .20899 L
s
.37272 .20899 m
.36254 .20899 L
s
.36254 .20899 m
.36763 .21781 L
s
.36763 .21781 m
.36254 .22663 L
s
.36254 .22663 m
.37272 .22663 L
s
.37272 .22663 m
.37781 .23545 L
s
.37781 .23545 m
.37272 .24427 L
s
.37272 .24427 m
.36254 .24427 L
s
.36254 .24427 m
.36763 .25309 L
s
.36763 .25309 m
.36254 .2619 L
s
.36254 .2619 m
.35235 .2619 L
s
.35235 .2619 m
.34726 .25309 L
s
.34726 .25309 m
.34217 .2619 L
s
.34217 .2619 m
.33199 .2619 L
s
.33199 .2619 m
.3269 .25309 L
s
.3269 .25309 m
.33199 .24427 L
s
.33199 .24427 m
.32181 .24427 L
s
.32181 .24427 m
.31671 .23545 L
s
.31671 .23545 m
.31162 .24427 L
s
.31162 .24427 m
.30144 .24427 L
s
.30144 .24427 m
.30653 .25309 L
s
.30653 .25309 m
.30144 .2619 L
s
.30144 .2619 m
.29126 .2619 L
s
.29126 .2619 m
.28617 .25309 L
s
.28617 .25309 m
.28108 .2619 L
s
.28108 .2619 m
.27089 .2619 L
s
.27089 .2619 m
.2658 .25309 L
s
.2658 .25309 m
.27089 .24427 L
s
.27089 .24427 m
.26071 .24427 L
s
.26071 .24427 m
.25562 .23545 L
s
.25562 .23545 m
.26071 .22663 L
s
.26071 .22663 m
.27089 .22663 L
s
.27089 .22663 m
.2658 .21781 L
s
.2658 .21781 m
.27089 .20899 L
s
.27089 .20899 m
.26071 .20899 L
s
.26071 .20899 m
.25562 .20018 L
s
.25562 .20018 m
.25053 .20899 L
s
.25053 .20899 m
.24035 .20899 L
s
.24035 .20899 m
.23525 .20018 L
s
.23525 .20018 m
.24035 .19136 L
s
.24035 .19136 m
.23016 .19136 L
s
.23016 .19136 m
.22507 .18254 L
s
.22507 .18254 m
.21998 .19136 L
s
.21998 .19136 m
.2098 .19136 L
s
.2098 .19136 m
.21489 .20018 L
s
.21489 .20018 m
.2098 .20899 L
s
.2098 .20899 m
.19961 .20899 L
s
.19961 .20899 m
.19452 .20018 L
s
.19452 .20018 m
.18943 .20899 L
s
.18943 .20899 m
.17925 .20899 L
s
.17925 .20899 m
.18434 .21781 L
s
.18434 .21781 m
.17925 .22663 L
s
.17925 .22663 m
.18943 .22663 L
s
.18943 .22663 m
.19452 .23545 L
s
.19452 .23545 m
.18943 .24427 L
s
.18943 .24427 m
.17925 .24427 L
s
.17925 .24427 m
.18434 .25309 L
s
.18434 .25309 m
.17925 .2619 L
s
.17925 .2619 m
.16907 .2619 L
s
.16907 .2619 m
.16398 .25309 L
s
.16398 .25309 m
.15888 .2619 L
s
.15888 .2619 m
.1487 .2619 L
s
.1487 .2619 m
.14361 .25309 L
s
.14361 .25309 m
.1487 .24427 L
s
.1487 .24427 m
.13852 .24427 L
s
.13852 .24427 m
.13343 .23545 L
s
.13343 .23545 m
.12834 .24427 L
s
.12834 .24427 m
.11815 .24427 L
s
.11815 .24427 m
.12325 .25309 L
s
.12325 .25309 m
.11815 .2619 L
s
.11815 .2619 m
.10797 .2619 L
s
.10797 .2619 m
.10288 .25309 L
s
.10288 .25309 m
.09779 .2619 L
s
.09779 .2619 m
.08761 .2619 L
s
newpath
.5 .5 .47619 0 365.73 arc
s
MathSubEnd
P
% End of sub-graphic
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 158.313},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000kB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K03oool00`000000oooo0?ooo`3o0?oood80oooo001[0?ooo`800000o`3ooom40?ooo`00
J03oool4000000040?ooo`000000000000000?l0oooo@@3oool006P0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`3o0?oood00oooo001U0?ooo`050000003oool0oooo0?ooo`000000
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?l0oooo>`3oool006@0oooo0P00
00020?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0oooo0P00003o0?ooocd0oooo
001Q0?ooo`@0000000@0oooo00000000000000001P3oool300000080oooo0`00003o0?ooocX0oooo
001R0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3oooli0?ooo`00H`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0oooo>@3oool00680oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`3o0?ooocT0oooo001Q0?ooo`@000003`3oool300000?l0
oooo>P3oool006@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`3o0?oooc/0oooo
001G0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo0?l0oooo;P3oool005H0oooo0P00000<0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool02@3oool200000?l0oooo<03oool005@0oooo0`0000000`3o
ool0000000000002000000D0oooo1000000?0?ooo`<000001P3oool3000000030?ooo`0000000000
00800000^`3oool6000006X0oooo001D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`2^0?ooo``0000000@0oooo00000000000000000P3oool;0000
05l0oooo001@0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool03`3oool01@000000oooo0?ooo`3oool0
000000@0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool000000:P0oooo1000
000;0?ooo`<0000000@0oooo00000000000000002`3oool5000005X0oooo001?0?ooo`8000000`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3oool00`000000oooo0?oo
o`0@0?ooo`060000003oool0oooo0000003oool000000P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00800000X`3oool5000000l0oooo0P0000020?ooo`8000004@3oool4000005H0
oooo001=0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<0000000@0oooo000000000000
0000503oool300000080oooo0`0000060?ooo`<0000000<0oooo0000000000000P00002M0?ooo`80
00004@3oool01@000000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo0000000A
0?ooo`<00000D`3oool004d0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`2H0?oo
o`@000004P3oool200000080oooo00<000000?ooo`3oool00P3oool2000000030?ooo`0000000000
01@0oooo1000001?0?ooo`00CP3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo09D0
oooo0`00000D0?ooo`L000001@3oool7000001D0oooo0P00001=0?ooo`00C@3oool00`000000oooo
0?ooo`0j0?ooo`030000003oool0oooo09<0oooo0P00000H0?ooo`050000003oool0oooo0?ooo`00
00001`3oool01@000000oooo0?ooo`3oool0000001P0oooo0P00001;0?ooo`00C@3oool3000003P0
oooo1000002B0?ooo`8000006@3oool200000080oooo00<000000?ooo`3oool01P3oool200000080
oooo00<000000?ooo`3oool0603oool2000004T0oooo001?0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool0TP3oool2000001/0oooo0`0000000`3oool000000?ooo`080?ooo`030000
003oool00000008000006`3oool3000004H0oooo001@0?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool0T03oool2000001l0oooo0P00000;0?ooo`800000803oool2000004@0oooo001?
0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0SP3oool2000001H0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool05@3oool200000480oooo001=0?ooo`<00000>03oool4000008X0oooo0P00000E0?oo
o`<0000000@0oooo00000000000000001@3oool3000000d0oooo0`0000050?ooo`<0000000@0oooo
00000000000000005P3oool200000400oooo001=0?ooo`030000003oool0oooo03X0oooo00<00000
0?ooo`3oool0Q`3oool2000001L0oooo00@000000?ooo`000000oooo0P0000060?ooo`030000003o
ool0000000d0oooo0P0000060?ooo`040000003oool000000?ooo`8000006@3oool2000003h0oooo
001>0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0Q@3oool2000001H0oooo00D0
00000?ooo`3oool0oooo000000030?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000
oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0
000000<0oooo00<000000?ooo`3oool05@3oool2000003`0oooo001=0?ooo`030000003oool0oooo
03X0oooo00<000000?ooo`3oool0P`3oool2000001L0oooo0P0000020?ooo`8000000`3oool00`00
0000oooo0?ooo`0200000080oooo0P00000?0?ooo`80000000<0oooo0000000000000P3oool20000
00<0oooo00<000000?ooo`3oool00P00000I0?ooo`800000>P3oool004d0oooo0`0000000`3oool0
000000000002000002l0oooo0`0000000`3oool0000000000002000008<0oooo00<000000?ooo`3o
ool05@3ooolC00000140oooo4`00000H0?ooo`030000003oool0oooo03L0oooo001?0?ooo`800000
0`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00800000QP3oool00`000000oooo
0?ooo`0F0?ooo`8000003`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0f0?ooo`00D03oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo02/0oooo00D000000?ooo`3oool0oooo000000240?ooo`800000
6P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo03D0oooo001D0?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool0Q03oool2000001/0oooo0`00000=0?ooo`030000003o
ool0oooo01<0oooo00<000000?ooo`3oool02`3oool3000001/0oooo0P00000e0?ooo`00E03oool3
000002X0oooo100000240?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00X0oooo0P00000O0?ooo`030000003oool0
oooo0380oooo001F0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0Q@3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0<@3oool002h0
oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02L0
oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0000000T0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`070?ooo`<000008@3oool200000340
oooo000]0?ooo`8000009`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02H0oooo
0P00001K0?ooo`030000003oool0oooo01h0oooo0`0000000`3oool000000?ooo`080?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0000000200000240oooo
00<000000?ooo`3oool0;P3oool002X0oooo10000000103oool0000000000000000R0?ooo`<00000
:P3oool400000240oooo0`0000020?ooo`<00000E`3oool00`000000oooo0?ooo`0O0?ooo`800000
0`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
1@3oool01@000000oooo0?ooo`3oool0000002<0oooo00<000000?ooo`3oool0;@3oool002/0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo05@0oooo
0P00000S0?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo00D0oooo00D000000?ooo`3oool0oooo0000000T0?ooo`030000003oool0oooo
02`0oooo000W0?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000040?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0[0?oo
o`050000003oool0oooo0?ooo`0000006`3oool01@000000oooo0?ooo`3oool0000000@0oooo00D0
00000?ooo`3oool0oooo0000001A0?ooo`030000003oool0oooo0280oooo1`0000050?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool00`3oool7000002@0oooo00<000000?ooo`3oool0
:`3oool002H0oooo0P0000030?ooo`030000003oool0oooo0080oooo00@000000?ooo`3oool0oooo
0P00000I0?ooo`8000000`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00800000
6P3oool200000080oooo00<000000?ooo`3oool00`3oool01P000000oooo0?ooo`000000oooo0000
04l0oooo00<000000?ooo`3oool09@3oool200000080oooo0P0000030?ooo`030000003oool0oooo
01d0oooo00D000000?ooo`3oool0oooo000000020?ooo`800000:03oool00`000000oooo0?ooo`0Z
0?ooo`00903oool3000000040?ooo`0000000000000000H0oooo10000000103oool0000000000000
000D0?ooo`<0000000<0oooo0000000000000P00000_0?ooo`<0000000<0oooo0000000000000P00
000C0?ooo`@0000000@0oooo00000000000000001P3oool300000080oooo0`00001;0?ooo`030000
003oool0oooo02L0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo01d0
oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo02L0oooo0P00000Z0?oo
o`00903oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0[0?ooo`050000003oool000000?ooo`0000008@3oool01000
0000oooo00000000000^0?ooo`030000003oool0oooo02P0oooo000U0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo
02`0oooo0`0000000`3oool000000?ooo`0P0?ooo`030000003oool0000000800000;P3oool00`00
0000oooo0?ooo`0W0?ooo`00903oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool09P3oool002@0oooo0`00000?0?ooo`<00000503oool3000003P0oooo
1000000C0?ooo`@000003`3oool3000004P0oooo00<000000?ooo`3oool0303oool300000200oooo
0P00000S0?ooo`8000008@3oool3000000h0oooo00<000000?ooo`3oool09@3oool002H0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo03D0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo04P0oooo00<0
00000?ooo`3oool02`3oool3000000040?ooo`0000000000000001`0oooo0`00000U0?ooo`<00000
7@3oool3000000040?ooo`0000000000000000`0oooo00<000000?ooo`3oool09@3oool001T0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0<0?ooo`8000000P3oool2000001d0oooo
0P00000V0?ooo`8000007P3oool200000080oooo0P00000>0?ooo`030000003oool0oooo02@0oooo
000H0?ooo`800000303oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`0000002`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00X0oooo0P00
000<0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool2000003/0oooo00<0
00000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0
oooo0000000F0?ooo`050000003oool0oooo0?ooo`0000009`3oool01@000000oooo0?ooo`3oool0
000001L0oooo00D000000?ooo`3oool0oooo000000030?ooo`050000003oool0oooo0?ooo`000000
2`3oool2000002@0oooo000F0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<000003`3o
ool3000000H0oooo0`0000020?ooo`<000001P3oool3000003P0oooo100000060?ooo`<0000000@0
oooo00000000000000001P3oool4000000l0oooo0`0000060?ooo`<0000000<0oooo000000000000
0P00000f0?ooo`030000003oool0oooo00X0oooo0P0000020?ooo`030000003oool0oooo0080oooo
0P0000000`3oool000000000000E0?ooo`8000000P3oool00`000000oooo0?ooo`0V0?ooo`800000
00<0oooo0000000000005P3oool200000080oooo00<000000?ooo`3oool00P3oool2000000030?oo
o`000000000000`0oooo00<000000?ooo`3oool08P3oool001H0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0>P3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo00T0oooo1`0000050?ooo`L00000403o
ool7000002T0oooo1`00000A0?ooo`L000001@3oool7000000X0oooo00<000000?ooo`3oool08@3o
ool00180oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000000@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00D0
00000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo0?ooo`000000103oool00`000000
oooo0?ooo`0i0?ooo`050000003oool0oooo0?ooo`000000103oool01@000000oooo0?ooo`3oool0
000000@0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`050000003oool0oooo0?ooo`000000
103oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000<@3oool00`000000
oooo0?ooo`0;0?ooo`050000003oool0oooo0?ooo`0000001`3oool01@000000oooo0?ooo`3oool0
00000180oooo00D000000?ooo`3oool0oooo0000000[0?ooo`050000003oool0oooo0?ooo`000000
4`3oool01@000000oooo0?ooo`3oool0000000L0oooo00D000000?ooo`3oool0oooo0000000<0?oo
o`030000003oool0oooo0200oooo000A0?ooo`8000000`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo008000000`3oool00`000000oooo0?ooo`0@0?ooo`040000003oool0oooo0?oo
o`8000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo008000000`3oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo008000000`3oool00`000000oooo0?ooo`020?oo
o`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo0?ooo`0A0?ooo`060000003oool0
oooo0000003oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00800000
<@3oool00`000000oooo0?ooo`0:0?ooo`8000000P3oool00`000000oooo0?ooo`060?ooo`050000
003oool0oooo0?ooo`0000004@3oool200000080oooo0P00000[0?ooo`050000003oool0oooo0?oo
o`0000004P3oool200000080oooo00<000000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool0
000000d0oooo00<000000?ooo`3oool07`3oool000l0oooo0`0000000`3oool00000000000020000
00H0oooo0`0000000`3oool0000000000002000001<0oooo0`0000020?ooo`<000001P3oool30000
00030?ooo`000000000000800000?@3oool3000000030?ooo`0000000000008000001@3oool40000
00040?ooo`0000000000000001@0oooo0`0000020?ooo`<000001P3oool3000000030?ooo`000000
000000800000;03oool00`000000oooo0?ooo`0;0?ooo`<0000000<0oooo0000003oool0203oool0
0`000000oooo0000000200000100oooo0`0000000`3oool000000?ooo`0/0?ooo`030000003oool0
0000008000004@3oool3000000030?ooo`000000oooo00P0oooo00<000000?ooo`0000000P00000<
0?ooo`030000003oool0oooo01l0oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?oo
o`3oool0:`3oool00`000000oooo0?ooo`0=0?ooo`8000002`3oool2000001@0oooo0P00000_0?oo
o`8000005@3oool2000000/0oooo0P00000?0?ooo`030000003oool0oooo01h0oooo000@0?ooo`03
0000003oool0oooo0;@0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool07P3oool000l0oooo00<000000?ooo`3oool0
]P3oool00`000000oooo0?ooo`0Y0?ooo`040000003oool0oooo0?ooo`<0000000@0oooo00000000
000000001@3oool3000000d0oooo0`0000050?ooo`H000001@3oool300000340oooo0`0000050?oo
o`<0000000@0oooo00000000000000001@3oool3000000d0oooo0`0000050?ooo`<0000000@0oooo
00000000000000000P3oool00`000000oooo0?ooo`0M0?ooo`003`3oool300000;@0oooo1000000Y
0?ooo`030000003oool0oooo0080oooo00@000000?ooo`000000oooo0P0000060?ooo`030000003o
ool0000000d0oooo0P0000060?ooo`050000003oool000000?ooo`0000001P3oool00`000000oooo
0000000a0?ooo`8000001P3oool010000000oooo0000003oool2000000H0oooo00<000000?ooo`00
00003@3oool2000000H0oooo00@000000?ooo`000000oooo0P0000030?ooo`030000003oool0oooo
01d0oooo000A0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0:`3oool00`000000
oooo000000030?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0
000000l0oooo00D000000?ooo`3oool0oooo000000030?ooo`040000003oool0oooo000000<0oooo
00D000000?ooo`3oool0oooo0000000c0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@00
0000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`050000003o
ool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0000000<0oooo0P00000N0?ooo`00
4P3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo02X0oooo00@000000?ooo`000000
00000P3oool2000000<0oooo00<000000?ooo`3oool00P0000020?ooo`8000003`3oool00`000000
oooo0?ooo`0200000080oooo0P0000020?ooo`80000000<0oooo0000000000000P3oool2000003<0
oooo00<000000?ooo`3oool00P0000020?ooo`8000000`3oool00`000000oooo0?ooo`0200000080
oooo0P00000?0?ooo`030000003oool0oooo008000000P3oool2000000<0oooo00<000000?ooo`3o
ool00P0000000`3oool000000?ooo`0L0?ooo`004@3oool00`000000oooo0?ooo`2a0?ooo`030000
003oool0oooo02X0ooooXP00000L0?ooo`003`3oool300000;@0oooo1000000W0?ooo`<00000WP3o
ool2000001`0oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0000002N0?ooo`030000003oool0000001/0oooo000@0?ooo`030000003oool0oooo
0;@0oooo00<000000?ooo`3oool09P3oool00`000000oooo00000002000009`0oooo1000000K0?oo
o`003`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo02D0oooo00D000000?ooo`3o
ool0oooo0000002K0?ooo`8000000`3oool00`000000oooo0?ooo`0H0?ooo`003`3oool300000003
0?ooo`000000000000800000Z`3oool3000000030?ooo`0000000000008000009@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0H0?ooo`004@3oool2000000<0oooo00<000000?ooo`3oool0Z03oool00`000000oooo
0?ooo`02000002T0oooo00<000000?ooo`3oool00P3oool3000009P0oooo0`0000040?ooo`030000
003oool0oooo01L0oooo000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0Y`3o
ool01@000000oooo0?ooo`3oool0000002P0oooo00@000000?ooo`3oool0oooo0`0000000`3oool0
00000?ooo`2G0?ooo`030000003oool00000008000000P3oool00`000000oooo0?ooo`0G0?ooo`00
5P3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0
0P3oool01@000000oooo0?ooo`3oool0000009H0oooo00D000000?ooo`3oool0oooo000000040?oo
o`030000003oool0oooo01H0oooo000F0?ooo`<00000YP3oool4000002X0oooo00<000000?ooo`3o
ool00P3oool2000000<0oooo00<000000?ooo`3oool0U03oool01@000000oooo0?ooo`3oool00000
00@0oooo00<000000?ooo`3oool05P3oool001P0oooo00<000000?ooo`3oool0X`3oool00`000000
oooo0?ooo`0/0?ooo`030000003oool0oooo0080oooo1`00002D0?ooo`L00000103oool00`000000
oooo0?ooo`0E0?ooo`006@3oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool01@3oool200000080oooo0P00002B0?ooo`80000000<0oooo000000000000
1`3oool00`000000oooo0?ooo`0E0?ooo`00603oool00`000000oooo0?ooo`2S0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool000000980oooo
00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0oooo01@0oooo000F0?ooo`<00000
YP3oool4000002T0oooo00<000000?ooo`3oool02@3oool300000900oooo0P00000<0?ooo`030000
003oool0oooo01@0oooo000F0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`0:0?ooo`<00000T03oool3000000`0oooo00<000000?ooo`3oool04`3o
ool00180oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2W0?ooo`050000003oool0
oooo0?ooo`0000009@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo08`0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`004@3oool2000000<0oooo00<00000
0?ooo`3oool0Z03oool00`000000oooo0?ooo`02000002@0oooo00<000000?ooo`3oool03@3oool2
000008h0oooo0P00000>0?ooo`030000003oool0oooo01<0oooo000?0?ooo`<0000000<0oooo0000
000000000P00002[0?ooo`<0000000<0oooo0000000000000P00000P0?ooo`030000003oool0oooo
00/0oooo0`0000000`3oool000000?ooo`2;0?ooo`030000003oool0000000800000303oool00`00
0000oooo0?ooo`0C0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01l0
oooo00<000000?ooo`3oool02`3oool200000080oooo0P00002:0?ooo`80000000<0oooo00000000
00003@3oool00`000000oooo0?ooo`0C0?ooo`00403oool00`000000oooo0?ooo`2d0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool0203oool01@000000oooo0?ooo`3oool0000000<0
oooo00<000000?ooo`3oool0R03oool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`0B0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool0203oool200000080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo0080oooo0P0000000`3oool0
00000000000:0?ooo`030000003oool0oooo0180oooo000?0?ooo`<00000]03oool4000001l0oooo
00<000000?ooo`3oool01P3oool7000000D0oooo00<000000?ooo`3oool0QP3oool00`000000oooo
0?ooo`030?ooo`L00000203oool00`000000oooo0?ooo`0A0?ooo`004P3oool00`000000oooo0?oo
o`2`0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0203oool01@000000oooo0?oo
o`3oool0000000L0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`050?ooo`050000
003oool0oooo0?ooo`0000002@3oool00`000000oooo0?ooo`0A0?ooo`004@3oool00`000000oooo
0?ooo`2a0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool01`3oool200000080oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`230?ooo`8000001`3oool01@000000oooo
0?ooo`3oool0000000T0oooo00<000000?ooo`3oool04@3oool000l0oooo0`00002d0?ooo`@00000
7P3oool00`000000oooo0?ooo`070?ooo`<0000000<0oooo0000003oool0203oool00`000000oooo
0?ooo`220?ooo`030000003oool0oooo00L0oooo00<000000?ooo`0000000P0000080?ooo`030000
003oool0oooo0140oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`090?ooo`8000002P3oool200000880oooo00<000000?ooo`3oool02@3o
ool2000000/0oooo00<000000?ooo`3oool0403oool00100oooo00<000000?ooo`3oool0]03oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`200?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0@0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo
01`0oooo00<000000?ooo`3oool0203oool3000000d0oooo00<000000?ooo`3oool0OP3oool00`00
0000oooo0?ooo`0;0?ooo`<000002@3oool00`000000oooo0?ooo`0@0?ooo`003`3oool300000003
0?ooo`0000000000008000001P3oool3000000030?ooo`000000000000800000S`3oool300000080
oooo0`0000060?ooo`<0000000<0oooo0000000000000P00000M0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`0000003@3oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo00/0
oooo0P00000:0?ooo`030000003oool0oooo0100oooo000A0?ooo`8000000`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo008000000`3oool00`000000oooo0?ooo`2<0?ooo`060000
003oool0oooo0000003oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
008000008@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0O03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0403oool00180oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool00000
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2;0?ooo`050000003oool0oooo
0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000008@3o
ool00`000000oooo0?ooo`080?ooo`8000003`3oool2000007`0oooo0P00000?0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool03`3oool001H0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool02@3ooolC
000007`0oooo4`00000:0?ooo`030000003oool0oooo00l0oooo000F0?ooo`<0000000<0oooo0000
000000000P0000060?ooo`<00000R`3oool3000000H0oooo0`0000000`3oool00000000000020000
02<0oooo00<000000?ooo`3oool02`3oool200000080oooo0P0000030?ooo`030000003oool0oooo
008000000P3oool2000007X0oooo00<000000?ooo`3oool00P0000020?ooo`8000000`3oool00`00
0000oooo0?ooo`02000000d0oooo00<000000?ooo`3oool03`3oool001P0oooo0P00000<0?ooo`03
0000003oool0oooo08P0oooo00<000000?ooo`3oool02@3oool2000002L0oooo00<000000?ooo`3o
ool0303oool01@000000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo00000003
0?ooo`030000003oool0oooo07P0oooo00D000000?ooo`3oool0oooo000000030?ooo`050000003o
ool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00l0oooo
000I0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool03`3oool3000000040?ooo`00
00000000000000D0oooo0`00001h0?ooo`<000001@3oool3000000040?ooo`000000000000000100
oooo00<000000?ooo`3oool03`3oool002H0oooo00<000000?ooo`3oool0R03oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo0140oooo0`0000090?ooo`800000MP3oool2000000T0oooo
0`00000B0?ooo`030000003oool0oooo00l0oooo000T0?ooo`<00000R`3oool300000340oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo07@0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00l0oooo000T
0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0M
0?ooo`<00000M03oool3000001l0oooo00<000000?ooo`3oool03P3oool002D0oooo00<000000?oo
o`3oool0R`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01/0oooo0`0000000`3o
ool000000?ooo`1c0?ooo`030000003oool00000008000007@3oool00`000000oooo0?ooo`0>0?oo
o`00903oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3o
ool0703oool00`000000oooo0?ooo`0200000780oooo00D000000?ooo`3oool0oooo0000000N0?oo
o`030000003oool0oooo00h0oooo000T0?ooo`<0000000@0oooo0000000000000000PP3oool30000
0080oooo0`00000a0?ooo`030000003oool0oooo01/0oooo0P0000030?ooo`030000003oool0oooo
0700oooo00D000000?ooo`3oool0oooo0000000N0?ooo`030000003oool0oooo00h0oooo000V0?oo
o`8000000`3oool00`000000oooo0?ooo`1n0?ooo`060000003oool0oooo0000003oool00000=03o
ool00`000000oooo0?ooo`0K0?ooo`L00000L03oool7000001d0oooo00<000000?ooo`3oool03P3o
ool002L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1m0?ooo`050000003oool0
oooo0?ooo`000000=@3oool00`000000oooo0?ooo`0M0?ooo`8000000P3oool2000006h0oooo0P00
00000`3oool000000000000P0?ooo`030000003oool0oooo00h0oooo000[0?ooo`030000003oool0
oooo07h0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0O0?ooo`050000003oool0
oooo0?ooo`000000KP3oool01@000000oooo0?ooo`3oool000000200oooo00<000000?ooo`3oool0
3P3oool002X0oooo1000001m0?ooo`<00000=`3oool00`000000oooo0?ooo`0R0?ooo`<00000K03o
ool2000002@0oooo00<000000?ooo`3oool03P3oool002d0oooo00<000000?ooo`3oool0NP3oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo0280oooo0`00001/0?ooo`<000008`3oool0
0`000000oooo0?ooo`0>0?ooo`00;P3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool0903oool2000006X0oooo00<000000?ooo`3oool0903oool00`00
0000oooo0?ooo`0>0?ooo`00;@3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo03P0
oooo00<000000?ooo`3oool0903oool2000006X0oooo0P00000U0?ooo`030000003oool0oooo00h0
oooo000Z0?ooo`@00000O@3oool3000003L0oooo00<000000?ooo`3oool08P3oool3000000030?oo
o`000000oooo06L0oooo00<000000?ooo`0000000P00000S0?ooo`030000003oool0oooo00h0oooo
000[0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?oo
o`0R0?ooo`8000000P3oool00`000000oooo0?ooo`1U0?ooo`050000003oool0oooo0?ooo`000000
903oool00`000000oooo0?ooo`0>0?ooo`009`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo07d0oooo00D000000?ooo`3oool0oooo0000000d0?ooo`030000003oool0oooo01l0oooo
00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo06@0oooo00D000000?ooo`3o
ool0oooo000000030?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03P3oool002H0
oooo0P0000030?ooo`030000003oool0oooo07h0oooo00H000000?ooo`3oool000000?ooo`00000d
0?ooo`030000003oool0oooo01/0oooo1`0000040?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool00`3oool7000001d0oooo00<000000?ooo`3oool03P3oool002@0oooo0`000000103o
ool000000000000000220?ooo`<000000P3oool300000340oooo00<000000?ooo`3oool06`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0HP3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`0>0?ooo`00903oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`3oool0703oool01@000000oooo0?ooo`3oool0000000L0oooo00<00000
0?ooo`3oool0H03oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`0000007P3o
ool00`000000oooo0?ooo`0>0?ooo`009@3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool06`3oool200000080oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`1P0?ooo`030000003oool0oooo00D0oooo0P0000000`3oool000000000000N
0?ooo`030000003oool0oooo00h0oooo000T0?ooo`030000003oool0oooo08`0oooo00<000000?oo
o`3oool0<03oool00`000000oooo0?ooo`0K0?ooo`<0000000<0oooo0000003oool0203oool00`00
0000oooo0?ooo`1N0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`0000000P00000M0?oo
o`030000003oool0oooo00h0oooo000T0?ooo`<00000R`3oool300000340oooo00<000000?ooo`3o
ool07@3oool2000000X0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo01h0oooo00<000000?ooo`3oool03P3oool002H0oooo00<000000?ooo`3oool0
R03oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
3`3oool001T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`260?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0?0?ooo`<0000000@0
oooo00000000000000001@3oool3000000`0oooo00<000000?ooo`3oool0F`3oool00`000000oooo
0?ooo`0;0?ooo`<000001@3oool3000000040?ooo`000000000000000100oooo00<000000?ooo`3o
ool03`3oool001P0oooo0P00000<0?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool0
2@3oool2000002L0oooo00<000000?ooo`3oool03`3oool010000000oooo0000003oool2000000H0
oooo00<000000?ooo`0000003@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo00/0
oooo0P0000060?ooo`040000003oool000000?ooo`8000004@3oool00`000000oooo0?ooo`0?0?oo
o`005P3oool3000000030?ooo`0000000000008000001P3oool3000008/0oooo0`0000060?ooo`<0
000000<0oooo0000000000000P00000S0?ooo`030000003oool0oooo00`0oooo00D000000?ooo`3o
ool0oooo000000030?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool03@3oool01@000000oooo0?ooo`3o
ool0000000<0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool03`3oool001H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool02`3oool200000080oooo0P00
00030?ooo`030000003oool0oooo008000000P3oool2000000l0oooo00<000000?ooo`3oool0E`3o
ool00`000000oooo0?ooo`0>0?ooo`80000000<0oooo0000000000000P3oool2000000<0oooo00<0
00000?ooo`3oool00P00000=0?ooo`030000003oool0oooo00l0oooo000B0?ooo`030000003oool0
oooo0080oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0R`3oool01@000000oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0
0P3oool01@000000oooo0?ooo`3oool000000240oooo00<000000?ooo`3oool0203ooolC00000140
oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0?0?oooa<000002P3oool00`000000
oooo0?ooo`0?0?ooo`004@3oool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`02000000<0oooo00<000000?ooo`3oool0S03oool01P000000oooo0?ooo`000000oooo0000
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0200000240oooo00<000000?oo
o`3oool0203oool2000000l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`1E0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo00l0oooo000?0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<0
000000<0oooo0000000000000P00002?0?ooo`<000000P3oool3000000H0oooo0`0000000`3oool0
000000000002000001d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0@0?ooo`00
3`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0
203oool3000000d0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1C0?ooo`030000
003oool0oooo01<0oooo00<000000?ooo`3oool02`3oool3000000T0oooo00<000000?ooo`3oool0
403oool00100oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0M0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo0580oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0:0?ooo`8000002`3o
ool00`000000oooo0?ooo`0@0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0
oooo01d0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0@0?ooo`003`3oool30000
0;@0oooo1000000N0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`0000002@3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`070?ooo`<000002P3oool00`000000oooo0?ooo`0A0?ooo`004@3oool00`000000
oooo0?ooo`2a0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool01`3oool300000003
0?ooo`000000oooo00P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`1>0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool01`3oool00`000000oooo00000002000000P0
oooo00<000000?ooo`3oool04@3oool00180oooo00<000000?ooo`3oool0/03oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo00P0oooo00D000000?ooo`3oool0oooo000000070?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000090?ooo`030000003oool0oooo
0140oooo000A0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`060?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00D0
oooo00D000000?ooo`3oool0oooo000000090?ooo`030000003oool0oooo0140oooo000?0?ooo`<0
0000]03oool4000001l0oooo00<000000?ooo`3oool01P3oool7000000D0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool00`3oool7000000P0oooo00<000000?ooo`3oool04@3oool000l0oooo00<000000?ooo`3oool0
]P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00L0oooo0P0000020?ooo`800000
0`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0
7@3oool01@000000oooo0?ooo`3oool000000080oooo0P00000:0?ooo`030000003oool0oooo0180
oooo000@0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`080?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0M0?ooo`03
0000003oool0oooo04P0oooo00<000000?ooo`3oool07@3oool01@000000oooo0?ooo`3oool00000
00<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0B0?ooo`003`3oool00`000000
oooo0?ooo`2f0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool02`3oool01@000000
oooo0000003oool000000200oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0O0?oo
o`040000003oool00000000000d0oooo00<000000?ooo`3oool04`3oool000l0oooo0`0000000`3o
ool000000000000200000:/0oooo0`0000000`3oool000000000000200000200oooo00<000000?oo
o`3oool02`3oool3000000030?ooo`000000oooo0200oooo00<000000?ooo`3oool0AP3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0000000800000303oool00`000000oooo0?ooo`0C0?oo
o`004@3oool2000000<0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`02000002@0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo04D0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01<0
oooo000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0Y`3oool01@000000oooo
0?ooo`3oool0000002D0oooo00<000000?ooo`3oool0303oool2000002<0oooo00<000000?ooo`3o
ool0A03oool00`000000oooo0?ooo`0Q0?ooo`8000003P3oool00`000000oooo0?ooo`0C0?ooo`00
5P3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
2P3oool3000002@0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0S0?ooo`<00000
2`3oool00`000000oooo0?ooo`0D0?ooo`005P3oool300000:H0oooo1000000Y0?ooo`030000003o
ool0oooo00T0oooo0P00000V0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool09@3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01@0oooo000H0?ooo`030000003oool0
oooo0:<0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`060?ooo`050000003oool0
oooo0?ooo`0000009`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo02D0oooo00D0
00000?ooo`3oool0oooo000000070?ooo`030000003oool0oooo01D0oooo000I0?ooo`030000003o
ool0oooo0:80oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`020?ooo`L00000:03o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo02L0oooo1`0000040?ooo`030000003o
ool0oooo01D0oooo000H0?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool0;03oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`0n0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo01D0oooo000F0?ooo`<00000YP3oool40000
02/0oooo00<000000?ooo`3oool00P3oool01@000000oooo0?ooo`3oool0000002X0oooo00<00000
0?ooo`3oool0?@3oool00`000000oooo0?ooo`0Y0?ooo`050000003oool0oooo0?ooo`000000103o
ool00`000000oooo0?ooo`0F0?ooo`005P3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo02X0oooo00@000000?ooo`3oool0oooo0P0000020?ooo`800000:`3oool00`000000oooo0?oo
o`0k0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool00P0000040?ooo`030000003o
ool0oooo01H0oooo000B0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0Y`3oool0
1@000000oooo0?ooo`3oool0000002T0oooo00<000000?ooo`3oool00`0000000`3oool000000?oo
o`0[0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0000
000200000080oooo00<000000?ooo`3oool05`3oool00140oooo0P0000030?ooo`030000003oool0
oooo0:P0oooo00<000000?ooo`3oool00P00000Y0?ooo`030000003oool0oooo0080oooo0P00000^
0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo01L0oooo000?0?ooo`<0000000<0oooo0000000000000P00002[0?oo
o`<0000000<0oooo0000000000000P00000U0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0;@3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo02d0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`0H0?ooo`003`3oool00`000000oooo0?ooo`2f0?ooo`03
0000003oool0oooo02D0oooo00<000000?ooo`0000000P00000`0?ooo`030000003oool0oooo03L0
oooo00<000000?ooo`3oool0;`3oool3000000030?ooo`000000oooo01T0oooo000@0?ooo`030000
003oool0oooo0;@0oooo00<000000?ooo`3oool09P3oool01@000000oooo0000003oool000000340
oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0_0?ooo`80000000<0oooo0000003o
ool06P3oool000l0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool000000380oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0000001/0oooo000?0?ooo`<00000]03oool4000002L0oooo0`00000c0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0<P3oool2000001`0oooo000A0?ooo`030000003oool0
oooo0;40oooo00<000000?ooo`3oool0:P3ooolg000003@0oooo=`00000L0?ooo`004P3oool00`00
0000oooo0?ooo`2`0?ooo`030000003oool0oooo02X0oooo00@000000?ooo`3oool000000`3oool0
1@000000oooo0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`050000
003oool0oooo0?ooo`0000000`3oool00`000000oooo0?ooo`0200000080oooo00D000000?ooo`3o
ool0oooo0000000c0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3o
ool0000000<0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`050000003oool0oooo0?ooo`00
00000`3oool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`0000007@3oool00140
oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0080
oooo0P0000030?ooo`030000003oool0oooo00@0oooo0P00000?0?ooo`030000003oool0oooo00@0
oooo0P0000020?ooo`030000003oool0oooo00@0oooo0`00000b0?ooo`030000003oool0oooo00@0
oooo0P0000030?ooo`030000003oool0oooo00@0oooo0P00000?0?ooo`030000003oool0oooo00@0
oooo0P0000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0703oool000l0oooo
0`00002d0?ooo`@00000:@3oool00`000000oooo0?ooo`020?ooo`<0000000@0oooo000000000000
00001@3oool3000000d0oooo0`0000050?ooo`<0000000<0oooo0000000000001@3oool300000340
oooo0`0000050?ooo`<0000000@0oooo00000000000000001@3oool3000000d0oooo0`0000050?oo
o`<0000000@0oooo00000000000000000P3oool00`000000oooo0?ooo`0M0?ooo`003`3oool00`00
0000oooo0?ooo`2f0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool00`3oool30000
00T0oooo00<000000?ooo`3oool02P3oool2000000T0oooo0P0000090?ooo`800000;`3oool20000
00T0oooo0`0000090?ooo`030000003oool0oooo00X0oooo0P0000090?ooo`<00000103oool00`00
0000oooo0?ooo`0M0?ooo`00403oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo02/0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo02d0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01h0
oooo000?0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0:`3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0000000T0oooo0`00000D0?ooo`030000003oool0000002d0oooo
0`00000E0?ooo`030000003oool0000000T0oooo0`00000?0?ooo`030000003oool0oooo01h0oooo
000?0?ooo`<0000000<0oooo0000000000000P0000060?ooo`<0000000<0oooo0000000000000P00
000C0?ooo`<000000P3oool3000000H0oooo0`0000000`3oool0000000000002000003d0oooo0`00
00000`3oool0000000000002000000D0oooo10000000103oool0000000000000000D0?ooo`<00000
0P3oool3000000H0oooo0`0000000`3oool0000000000002000002`0oooo00<000000?ooo`3oool0
2`3oool3000000030?ooo`000000oooo00P0oooo00<000000?ooo`0000000P00000@0?ooo`<00000
00<0oooo000000000000;03oool00`000000oooo0000000200000140oooo0`0000000`3oool00000
0?ooo`080?ooo`030000003oool0000000800000303oool00`000000oooo0?ooo`0O0?ooo`004@3o
ool2000000<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02000000<0oooo00<0
00000?ooo`3oool0403oool010000000oooo0?ooo`3oool200000080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3oool0>P3oool00`000000oooo
0?ooo`02000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool200000080
oooo00<000000?ooo`3oool04@3oool01P000000oooo0?ooo`000000oooo00000080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0200000340oooo00<000000?ooo`3oool02`3oool0
1@000000oooo0?ooo`3oool0000000L0oooo00D000000?ooo`3oool0oooo0000000B0?ooo`050000
003oool0oooo0?ooo`000000:`3oool01@000000oooo0?ooo`3oool0000001<0oooo00D000000?oo
o`3oool0oooo000000070?ooo`050000003oool0oooo0?ooo`0000003@3oool00`000000oooo0?oo
o`0O0?ooo`004P3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3o
ool01@000000oooo0?ooo`3oool0000000@0oooo00D000000?ooo`3oool0oooo000000040?ooo`03
0000003oool0oooo03T0oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0oooo
0?ooo`000000103oool01@000000oooo0?ooo`3oool000000180oooo00D000000?ooo`3oool0oooo
000000040?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo0000000b0?ooo`03
0000003oool0oooo00T0oooo0P0000030?ooo`030000003oool0oooo00D0oooo00D000000?ooo`3o
ool0oooo0000000A0?ooo`8000000`3oool2000002X0oooo00D000000?ooo`3oool0oooo0000000B
0?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`0000002`3o
ool200000280oooo000F0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=@3oool0
0`000000oooo0?ooo`080?ooo`L000001@3oool700000100oooo1`00000Y0?ooo`L000004@3oool7
000000D0oooo1`0000090?ooo`030000003oool0oooo0280oooo000F0?ooo`<0000000<0oooo0000
000000000P0000060?ooo`<000003`3oool3000000H0oooo0`0000020?ooo`<000001P3oool30000
03P0oooo100000060?ooo`<0000000@0oooo00000000000000001P3oool4000000l0oooo0`000006
0?ooo`<0000000<0oooo0000000000000P00000g0?ooo`030000003oool0oooo00T0oooo0P000002
0?ooo`8000000`3oool00`000000oooo0?ooo`02000001D0oooo0P0000020?ooo`8000009`3oool2
000000030?ooo`000000000001H0oooo0P0000020?ooo`8000000`3oool00`000000oooo0?ooo`02
000000`0oooo00<000000?ooo`3oool08P3oool001P0oooo0P00000<0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool02@3oool00`000000oooo0000000;0?ooo`030000003oool0oooo
03D0oooo00<000000?ooo`3oool02P3oool2000000`0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`090?ooo`800000>`3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo
0?ooo`0000000`3oool01@000000oooo0?ooo`3oool0000001H0oooo00D000000?ooo`3oool0oooo
0000000W0?ooo`050000003oool0oooo0?ooo`0000005`3oool01@000000oooo0?ooo`3oool00000
00<0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`030000003oool0oooo02<0oooo000I0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo03X0oooo00<000000?ooo`3oool0303oool3000000040?ooo`0000000000
000001`0oooo0`00000U0?ooo`<000007@3oool3000000040?ooo`0000000000000000d0oooo00<0
00000?ooo`3oool0903oool002H0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool03@3oool300000200oooo0P00000S
0?ooo`8000008@3oool3000000h0oooo00<000000?ooo`3oool09@3oool002@0oooo0`00000?0?oo
o`<00000503oool3000003P0oooo1000000C0?ooo`@000003`3oool3000004P0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo02D0oooo000T0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`180?ooo`03
0000003oool0oooo02h0oooo0`00000Q0?ooo`<00000<03oool2000002P0oooo000U0?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0i0?ooo`030000
003oool0oooo01<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo02/0oooo0`0000000`3oool000000?ooo`0P0?ooo`030000003oool0000000800000
;@3oool00`000000oooo0?ooo`0X0?ooo`00903oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo01<0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0Z0?ooo`8000000`3o
ool00`000000oooo0?ooo`0M0?ooo`050000003oool0oooo0?ooo`000000;P3oool00`000000oooo
0?ooo`0X0?ooo`00903oool3000000040?ooo`0000000000000000H0oooo10000000103oool00000
00000000000D0?ooo`<0000000<0oooo0000000000000P00000_0?ooo`<0000000<0oooo00000000
00000P00000C0?ooo`@0000000@0oooo00000000000000001P3oool300000080oooo0`00001;0?oo
o`030000003oool0oooo02L0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0
oooo01d0oooo00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo02L0oooo00<0
00000?ooo`3oool0:@3oool002H0oooo0P0000030?ooo`030000003oool0oooo0080oooo00@00000
0?ooo`3oool0oooo0P00000I0?ooo`8000000`3oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo008000006P3oool200000080oooo00<000000?ooo`3oool00`3oool01P000000oooo0?oo
o`000000oooo000004l0oooo00<000000?ooo`3oool08`3oool7000000D0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`030?ooo`L000009@3oool00`000000oooo0?ooo`0Z0?ooo`00
9`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000103oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0:`3oool01@000000
oooo0?ooo`3oool0000001/0oooo00D000000?ooo`3oool0oooo000000040?ooo`050000003oool0
oooo0?ooo`000000D@3oool2000002<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`800000703oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool08P3oool2000002d0oooo000[0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0200oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0240oooo00D00000
0?ooo`3oool0oooo000000070?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool01@3o
ool01@000000oooo0?ooo`3oool0000002<0oooo00<000000?ooo`3oool0;@3oool002X0oooo1000
0000103oool0000000000000000R0?ooo`<00000:P3oool400000240oooo0`0000020?ooo`<00000
E`3oool00`000000oooo0?ooo`0O0?ooo`8000000P3oool00`000000oooo0?ooo`060?ooo`800000
6@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo008000008`3oool00`000000oooo
0?ooo`0]0?ooo`00;@3oool2000002L0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`0V0?ooo`800000F`3oool2000001l0oooo0`0000000`3oool000000?ooo`080?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0000000200000200oooo0P00
000`0?ooo`00;P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02H0oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo01l0oooo0P00000;
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0P
0?ooo`030000003oool0oooo0300oooo001F0?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool0Q@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01l0oooo00<000000?oo
o`3oool0<@3oool005@0oooo0`00000Z0?ooo`@00000Q03oool2000001`0oooo0`00000=0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool02`3oool3000001`0oooo0P00000d0?ooo`00
E03oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0000000=0?ooo`8000004`3oool2000000d0oooo0P00000L0?ooo`030000
003oool0oooo03@0oooo001@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0:`3o
ool01@000000oooo0?ooo`3oool0000008@0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo03D0oooo001?0?ooo`8000000`3oool00`000000oooo0?oo
o`0/0?ooo`030000003oool0oooo00800000Q@3oool2000001P0oooo0P00000@0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo03H0oooo001=0?ooo`<0000000<0oooo0000000000000P00000_0?ooo`<0000000<0oooo
0000000000000P0000230?ooo`8000005P3ooolC00000140oooo4`00000G0?ooo`800000>@3oool0
04d0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo
01H0oooo00D000000?ooo`3oool0oooo000000030?ooo`050000003oool0oooo0?ooo`0000000`3o
ool00`000000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`0000000`3oool01@000000oooo
0?ooo`3oool0000000<0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0i0?ooo`00
CP3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0
603oool2000000<0oooo00<000000?ooo`3oool0103oool3000000h0oooo00<000000?ooo`3oool0
103oool2000000<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0j0?ooo`00C@3o
ool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo08H0oooo0P00000H0?ooo`<0000000@0
oooo00000000000000001@3oool3000000d0oooo0`0000050?ooo`<0000000@0oooo000000000000
00006@3oool2000003d0oooo001=0?ooo`<00000>03oool4000008T0oooo0P00000H0?ooo`<00000
2@3oool2000000/0oooo0P0000090?ooo`<000006@3oool2000003l0oooo001?0?ooo`030000003o
ool0oooo03D0oooo00<000000?ooo`3oool0S@3oool2000001L0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool05P3o
ool200000440oooo001@0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0S`3oool2
00000200oooo00<000000?ooo`0000002@3oool300000200oooo0`0000130?ooo`00C`3oool00`00
0000oooo0?ooo`0e0?ooo`030000003oool0oooo0940oooo0P00000L0?ooo`<0000000<0oooo0000
003oool0203oool00`000000oooo00000002000001`0oooo0P0000160?ooo`00C@3oool3000003P0
oooo1000002A0?ooo`8000006`3oool01@000000oooo0?ooo`3oool0000000L0oooo00D000000?oo
o`3oool0oooo0000000K0?ooo`800000B03oool004d0oooo00<000000?ooo`3oool0>P3oool00`00
0000oooo0?ooo`2B0?ooo`800000603oool2000000<0oooo0P0000050?ooo`8000000`3oool00`00
0000oooo0?ooo`0G0?ooo`800000BP3oool004h0oooo00<000000?ooo`3oool0>@3oool00`000000
oooo0?ooo`2D0?ooo`D000004`3oool7000000D0oooo1`00000D0?ooo`@00000C03oool004d0oooo
00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`2I0?ooo`8000004`3oool200000080oooo
0P0000030?ooo`030000003oool0oooo00800000503oool300000500oooo001=0?ooo`<0000000<0
oooo0000000000000P0000060?ooo`<0000000@0oooo0000000000000000503oool300000080oooo
0`0000060?ooo`<0000000<0oooo0000000000000P00002L0?ooo`D000003`3oool01@000000oooo
0?ooo`3oool0000000<0oooo00D000000?ooo`3oool0oooo0000000@0?ooo`@00000D`3oool004l0
oooo0P0000030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P0000030?ooo`03
0000003oool0oooo0100oooo00H000000?ooo`3oool000000?ooo`0000020?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P00002U0?ooo`@000003P3oool3000000040?ooo`000000
0000000000h0oooo1@00001G0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`050000003oool0
oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00l0oooo00D0
00000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0
oooo0000002Y0?ooo`L000002@3oool3000000T0oooo1`00001L0?ooo`00E03oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0/P3ooolE000006<0oooo
001D0?ooo`<0000000<0oooo0000000000000P0000050?ooo`@000003`3oool3000000H0oooo0`00
00000`3oool000000000000200000?l0oooo;03oool005H0oooo0P00000<0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool02@3oool200000?l0oooo<03oool005L0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool0o`3oool^0?ooo`00I03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0
oooo>`3oool00640oooo1000000?0?ooo`<00000o`3ooolj0?ooo`00HP3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0?l0oooo>@3oool006<0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`3o0?ooocT0oooo001R0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0o`3oooli0?ooo`00H@3oool4000000040?ooo`0000000000000000H0oooo0`0000020?oo
o`<00000o`3ooolj0?ooo`00I03oool200000080oooo00<000000?ooo`3oool00`3oool010000000
oooo0?ooo`3oool200000?l0oooo?@3oool006D0oooo00D000000?ooo`3oool0oooo000000040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0o`3ooolk0?ooo`00J03oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0?l0oooo@03oool006P0oooo10000000103oool00000
00000000003o0?oood40oooo001[0?ooo`800000o`3ooom40?ooo`00K03oool00`000000oooo0?oo
o`3o0?oood80oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {236.438, 0}} -> {-0.0524626, -0.0274986, \
0.00732573, 0.00732573}, {{10.6875, 205.75}, {230.75, 5.625}} -> {-0.0828296, \
-0.34818, 0.00807849, 0.00807849}, {{225.25, 420.25}, {230.75, 5.625}} -> \
{-1.50676, -0.446566, 0.00932652, 0.00932652}}]
}, Closed]],

Cell[TextData[{
  "The area is computed by adding or subtracting the area of each new \
triangle to or from the original area, such that ",
  Cell[BoxData[
      \(TraditionalForm\`A = A\_0 \[PlusMinus] \[Delta]\[VeryThinSpace]A\)]],
  " where the sign is positive for outward or negative for inward \
triangulation.  The area of an equilateral triangle with sides of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_n\)]],
  " is simply ",
  Cell[BoxData[
      \(TraditionalForm\`\(\@3\/4\) \[ScriptL]\_n\^2\)]],
  " so that change in area for iteration becomes ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]A\_n = 
        n\_0\ \(\@3\/4\) \(\[Sum]\+\(n = 1\)\%\[Infinity]\( 
                  4\^\(n - 1\)\) \[ScriptL]\_n\^2\)\)]],
  ".  Recognizing that ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_\(\(n\)\(\ \)\) = \(\((1\/3)\)\^n\) \
\[ScriptL]\_0\)]],
  ", we find that ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[VeryThinSpace]A\_n = \(n\_0\) \(\[ScriptL]\
\_0\^2\) \(\@3\/16\) \((4\/9)\)\^n\)]],
  " is a simple geometric series."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Delta]A = \(n\_0\) \(\[ScriptL]\_0\^2\) \(\@3\/16\) \(\[Sum]\+\(n = 1\
\)\%\[Infinity]\((4\/9)\)\^n\)\)], "Input"],

Cell[BoxData[
    \(1\/20\ \@3\ n\_0\ \[ScriptL]\_0\%2\)], "Output"]
}, Closed]],

Cell["\<\
For example, the area enclosed by a Koch snowflake obtained by outward \
triangulation of an equilateral triangle is larger by a factor\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((A\_0 + \[Delta]A)\)/
        A\_0 /. {n\_0 \[Rule] 3, \[ScriptL]\_0 \[Rule] 1, 
        A\_0 \[Rule] \@3\/4}\)], "Input"],

Cell[BoxData[
    \(8\/5\)], "Output"]
}, Closed]],

Cell[TextData[{
  "compared with its original area.  It is instructive to compare this area \
with that of the circumscribed circle of radius ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0/\@3\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((A\_0 + \[Delta]A)\)/\((\[Pi]\ r\^2)\) /. {n\_0 \[Rule] 
            3, \[ScriptL]\_0 \[Rule] 1, A\_0 \[Rule] \@3\/4, 
          r \[Rule] 1/\@3}\  // N\)], "Input"],

Cell[BoxData[
    \(0.6615946745061505`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Hence, the snowflake occupies 66% of the circumscribed circle.  It is easy \
to show that as ",
  Cell[BoxData[
      \(TraditionalForm\`n\_0\)]],
  " increases, the area enclosed by the outward Koch curve approaches that of \
the circumscribed circle as ",
  Cell[BoxData[
      \(TraditionalForm\`n \[Rule] \[Infinity]\)]],
  "."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Fractal dimension", "Subsection"],

Cell[TextData[{
  "The size of an object can be measured in terms of an elementary volume ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\^d\)]],
  ", where ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\)]],
  " is a length scale and ",
  StyleBox["d",
    FontSlant->"Italic"],
  " is its dimensionality.  The ",
  StyleBox["d",
    FontSlant->"Italic"],
  "-dimensional volume is then the number of elementary volumes, ",
  Cell[BoxData[
      \(TraditionalForm\`N[\[CurlyEpsilon]]\)]],
  ", contained within the object.  Ordinarily one expects that upon reduction \
of the length scale, ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] \[Rule] \[Lambda]\ \
\[CurlyEpsilon]\)]],
  ", the number of elementary volumes scales according to ",
  Cell[BoxData[
      \(TraditionalForm\`N \[Proportional] \[CurlyEpsilon]\^\(-d\) \
\[Proportional] \[Lambda]\^\(-d\)\)]],
  ".  For example, as the number of sides increases a regular polygon becomes \
an increasingly accurate approximation to a circle.  If we decrease the \
length of each side by a rational factor \[Lambda] and correspondingly \
increase the number of sides by the factor ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\^\(-1\)\)]],
  ", the length of the perimeter remains constant as the area approaches that \
of the circumscribed circle.  Thus, a circle is a one-dimensional object of \
finite length enclosing finite area.  By contrast, the Koch curve also \
encloses a finite area but has infinite length and, hence, cannot be \
described as a one-dimensional object.  Therefore, it is necessary to \
generalize our notions of dimension to include the possibility of nonintegral \
or ",
  StyleBox["fractal",
    FontSlant->"Italic"],
  " dimension.  The procedure proposed by Hausdorff and Besicovitch applies \
essentially the same scaling argument to objects generated by iterated \
transformations."
}], "Text"],

Cell[TextData[{
  "The number of line segments in a Koch curve scales as ",
  Cell[BoxData[
      \(TraditionalForm\`N \[Proportional] 4\^n\)]],
  " while the elementary length scales as ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda] = \((1\/3)\)\^n\)]],
  ".  Thus, equating these scaling relationships and solving for the \
dimensionality ",
  Cell[BoxData[
      \(TraditionalForm\`d\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[4\^n \[Equal] \((1\/3)\)\^\(\(-n\)\ d\), d] // 
      PowerExpand\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"ifun"\), \(\(:\)\(\ \)\), "\<\"Inverse functions are \
being used by \\!\\(Solve\\), so some solutions may not be found; use Reduce \
for complete solution information. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({{d \[Rule] Log[4]\/Log[3]}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // N\)], "Input"],

Cell[BoxData[
    \({{d \[Rule] 1.2618595071429146`}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "we find that the Koch curve is characterized by a non-integer dimension, \
",
  Cell[BoxData[
      \(TraditionalForm\`d \[TildeTilde] 1.26\)]],
  ".  This fractal dimension occurs because the number of segments increases \
more rapidly than the length of each segment decreases.  Other \
transformations which share this behavior will also yield fractal dimensions \
",
  Cell[BoxData[
      \(TraditionalForm\`1 < d < 2\)]],
  " for which a finite area is enclosed by an infinite length.  Thus, it is \
useful to distinguish between ",
  StyleBox["topological dimension",
    FontSlant->"Italic"],
  ", which is unity for a curve constructed by applying a linear \
self-similarity transformation to a line, and ",
  StyleBox["Hausdorff",
    FontSlant->"Italic"],
  " dimension.  These concepts can be readily generalized to figures \
constructed by applying self-similarity transformations to simple objects of \
larger, but integral, topological dimension.  When the Hausdorff dimension ",
  StyleBox["d",
    FontSlant->"Italic"],
  " of a closed figure is greater than its topological dimension ",
  Cell[BoxData[
      \(TraditionalForm\`d\_0\)]],
  ", an infinite ",
  Cell[BoxData[
      \(TraditionalForm\`d\_0\)]],
  "-dimensional length encloses a finite ",
  Cell[BoxData[
      \(TraditionalForm\`\((d\_0 + 1)\)\)]],
  "-dimensional area.  Conversely, when ",
  Cell[BoxData[
      \(TraditionalForm\`d < d\_0\)]],
  ", a finite ",
  Cell[BoxData[
      \(TraditionalForm\`d\_0\)]],
  "-dimensional length is reduced to a ",
  Cell[BoxData[
      \(TraditionalForm\`\((d\_0 - 1)\)\)]],
  "-fractal dust (see ",
  StyleBox["Cantor set",
    FontSlant->"Italic"],
  " problem below). "
}], "Text"],

Cell["\<\
This property of fractal curves is sometimes described by analogy with \
coastlines.  Using a coarse length scale, such as kilometers, the length of a \
coast has one apparent value.  However, upon closer examination, use of a \
meter stick would result in a greater length estimate as irregular features \
of smaller scale are probed.  Decreasing the length scale to centimeter or \
millimeter scales produces ever greater estimates of the length as finer \
structures become apparent.  As this process is repeated, the coast line \
appears to have almost infinite length by the time we reach atomic or \
subatomic dimensions.  Evidently, coast lines are fractal by nature.\
\>", "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Additional exercises", "ExerciseMain"],

Cell[CellGroupData[{

Cell["Cantor set", "ExerciseTitle"],

Cell[TextData[{
  "Consider a line segment of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0\)]],
  ".  In the first iteration remove the middle third of that segment to \
produce two segments each of length ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0/3\)]],
  ".  In the next iteration remove the middle third of each of those, and so \
forth.  The Cantor set consists of the endpoints of every segment as this \
process is iterated an infinite number of times.  By analyzing the \
relationship between the number of segments and their total length, deduce \
the Hausdorff dimension of the Cantor set."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Devil's staircase", "ExerciseTitle"],

Cell[TextData[{
  "The ",
  StyleBox["devil's staircase",
    FontSlant->"Italic"],
  " provides an amusing display of a Cantor subset based upon ",
  StyleBox["n",
    FontSlant->"Italic"],
  " iterations.  Let ",
  Cell[BoxData[
      \(TraditionalForm\`\[ScriptL]\_0 = 1\)]],
  " and let the total height of the staircase also be one unit.  The \
staircase consists of a series of line segments, alternating between rising \
and horizontal, that begins at {0,0} and ends at {1,1} with vertices at each \
member of the Cantor set.  Make the vertical rise simply proportional to the \
horizontal distance and write a function which displays the staircase for \
specified ",
  StyleBox["n",
    FontSlant->"Italic"],
  "."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["A variation of Koch curve", "ExerciseTitle"],

Cell["\<\
Develop a Koch-like curve in which a square is erected on the middle third of \
each line segment.  Determine the fractal dimension of this object.\
\>", "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->"EditBar",
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
ShowSelection->True,
ShowCellLabel->False,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"]},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Notation Package Styles", "Section",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell["\<\
The cells below define certain styles needed by the Notation package. These \
styles serve to make visible otherwise invisible tagboxes.\
\>", "Text",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationTemplateStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 1, 0.850004],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationPatternWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 0.900008, 0.979995],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationMadeBoxesWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[0.900008, 0.889998, 1],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"]
  }, Closed]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 79, 2, 81, "Title"],
Cell[1858, 57, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2338, 76, 31, 0, 83, "Section"],
Cell[2372, 78, 737, 15, 93, "Text"],
Cell[3112, 95, 8546, 151, 137, 740, 50, "GraphicsData", "PostScript", \
"Graphics"],
Cell[11661, 248, 313, 8, 41, "Text"],
Cell[11977, 258, 12939, 236, 137, 1148, 86, "GraphicsData", "PostScript", \
"Graphics"],
Cell[24919, 496, 128, 3, 41, "Text"],
Cell[25050, 501, 733, 14, 119, "Text"],
Cell[25786, 517, 131555, 6335, 310, 61658, 5468, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[157378, 6857, 33, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[157436, 6861, 36, 0, 68, "Subsection"],
Cell[157475, 6863, 119, 2, 99, "Input"],
Cell[157597, 6867, 65, 1, 73, "Input"],
Cell[157665, 6870, 547, 17, 73, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[158249, 6892, 37, 0, 42, "Subsection"],

Cell[CellGroupData[{
Cell[158311, 6896, 77, 0, 55, "Subsubsection"],
Cell[158391, 6898, 1694, 48, 130, "Text"],
Cell[160088, 6948, 266, 6, 66, "DisplayFormula",
  Evaluatable->False],
Cell[160357, 6956, 73, 0, 41, "Text"],
Cell[160433, 6958, 217, 4, 166, "Input"],
Cell[160653, 6964, 123, 2, 96, "Input"],
Cell[160779, 6968, 217, 4, 105, "Input"],

Cell[CellGroupData[{
Cell[161021, 6976, 71, 1, 73, "Input"],
Cell[161095, 6979, 144, 2, 75, "Output"]
}, Closed]],
Cell[161254, 6984, 123, 2, 96, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[161414, 6991, 59, 0, 55, "Subsubsection"],
Cell[161476, 6993, 165, 3, 41, "Text"],
Cell[161644, 6998, 376, 8, 151, "Input"],
Cell[162023, 7008, 103, 2, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[162163, 7015, 87, 1, 55, "Subsubsection"],
Cell[162253, 7018, 262, 4, 125, "Input"],
Cell[162518, 7024, 130, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[162673, 7031, 121, 3, 73, "Input"],

Cell[CellGroupData[{
Cell[162819, 7038, 8560, 152, 137, 754, 51, "GraphicsData", "PostScript", \
"Graphics"],
Cell[171382, 7192, 12953, 237, 137, 1162, 87, "GraphicsData", "PostScript", \
"Graphics"],
Cell[184338, 7431, 18052, 424, 137, 2782, 231, "GraphicsData", "PostScript", \
"Graphics"],
Cell[202393, 7857, 26804, 1028, 137, 9242, 807, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[229270, 8893, 53, 0, 68, "Subsection"],
Cell[229326, 8895, 602, 13, 95, "Text"],
Cell[229931, 8910, 75, 1, 73, "Input"],
Cell[230009, 8913, 2156, 37, 1035, "Input"],

Cell[CellGroupData[{
Cell[232190, 8954, 199, 3, 105, "Input"],
Cell[232392, 8959, 77938, 2987, 387, 26131, 2343, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[310391, 11953, 38, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[310454, 11957, 70, 0, 68, "Subsection"],
Cell[310527, 11959, 130, 3, 41, "Text"],
Cell[310660, 11964, 221, 4, 135, "Input"],
Cell[310884, 11970, 332, 8, 67, "Text"],

Cell[CellGroupData[{
Cell[311241, 11982, 385, 7, 125, "Input"],
Cell[311629, 11991, 121288, 6869, 250, 70017, 6232, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[432954, 18865, 312, 6, 99, "Input"],
Cell[433269, 18873, 95815, 5234, 250, 53049, 4702, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[529099, 24110, 68, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[529192, 24114, 308, 6, 99, "Input"],
Cell[529503, 24122, 113709, 6784, 218, 69792, 6238, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[643261, 30912, 62, 0, 42, "Subsection"],
Cell[643326, 30914, 65, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[643416, 30919, 181, 3, 73, "Input"],
Cell[643600, 30924, 149199, 9822, 444, 103617, 9255, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[792860, 40753, 37, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[792922, 40757, 58, 0, 68, "Subsection"],
Cell[792983, 40759, 1118, 32, 98, "Text"],

Cell[CellGroupData[{
Cell[794126, 40795, 388, 7, 176, "Input"],
Cell[794517, 40804, 114295, 6124, 250, 61686, 5470, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[908827, 46931, 1094, 26, 109, "Text"],

Cell[CellGroupData[{
Cell[909946, 46961, 136, 2, 119, "Input"],
Cell[910085, 46965, 68, 1, 75, "Output"]
}, Closed]],
Cell[910168, 46969, 159, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[910352, 46976, 144, 3, 105, "Input"],
Cell[910499, 46981, 38, 1, 75, "Output"]
}, Closed]],
Cell[910552, 46985, 219, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[910796, 46994, 187, 3, 105, "Input"],
Cell[910986, 46999, 53, 1, 59, "Output"]
}, Closed]],
Cell[911054, 47003, 362, 10, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[911453, 47018, 39, 0, 68, "Subsection"],
Cell[911495, 47020, 1928, 43, 197, "Text"],
Cell[913426, 47065, 409, 11, 72, "Text"],

Cell[CellGroupData[{
Cell[913860, 47080, 104, 2, 98, "Input"],
Cell[913967, 47084, 368, 6, 60, "Message"],
Cell[914338, 47092, 62, 1, 78, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[914437, 47098, 39, 1, 73, "Input"],
Cell[914479, 47101, 67, 1, 59, "Output"]
}, Closed]],
Cell[914561, 47105, 1728, 46, 171, "Text"],
Cell[916292, 47153, 698, 10, 119, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[917039, 47169, 44, 0, 54, "ExerciseMain"],

Cell[CellGroupData[{
Cell[917108, 47173, 35, 0, 58, "ExerciseTitle"],
Cell[917146, 47175, 655, 13, 93, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[917838, 47193, 42, 0, 58, "ExerciseTitle"],
Cell[917883, 47195, 742, 19, 93, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[918662, 47219, 50, 0, 58, "ExerciseTitle"],
Cell[918715, 47221, 179, 3, 41, "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

