(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   1670972,      32435]*)
(*NotebookOutlinePosition[   1705046,      33662]*)
(*  CellTagsIndexPosition[   1705002,      33658]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Binary Collisions", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell["\<\
\tIn this notebook we study some of the properties of nonrelativistic binary \
collisions.  Inelasticity and mass transfer are included.  Tools are \
developed to display the vector relationships and to produce both momentum \
and velocity diagrams.  Kinematic relationships are also plotted.\
\>", "Text"],

Cell[TextData[{
  "\tThe reaction is described by the mass relationship ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1 + m\_2 \[Equal] m\_3 + m\_4\)]],
  ", where ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  " is designated the projectile, ",
  Cell[BoxData[
      \(TraditionalForm\`m\_2\)]],
  " the target, ",
  Cell[BoxData[
      \(TraditionalForm\`m\_3\)]],
  " the ejectile, and ",
  Cell[BoxData[
      \(TraditionalForm\`m\_4\)]],
  " the recoil particle.  If no mass is transferred between the projectile \
and the target, then ",
  Cell[BoxData[
      \(TraditionalForm\`m\_3 \[Rule] m\_1\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`m\_4 \[Rule] m\_2\)]],
  ".   The particle velocities are designated ",
  Cell[BoxData[
      \(TraditionalForm\`{\(v\& \[RightVector] \)\_1, \(v\& \[RightVector] \)\
\_2, \(v\& \[RightVector] \)\_3, \(v\& \[RightVector] \)\_4}\)]],
  ".  For convenience we define the lab frame to be the rest frame of the \
target, such that ",
  Cell[BoxData[
      \(TraditionalForm\`v\_2 \[Rule] 0\)]],
  " in the lab frame.  Other frames can be developed easily using Galilean \
transformations.  Conservation of momentum is then represented by the \
equation ",
  Cell[BoxData[
      \(TraditionalForm\`\(p\& \[RightVector] \)\_1 + \(p\& \[RightVector] \)\
\_2 \[Equal] \(p\& \[RightVector] \)\_3 + \(p\& \[RightVector] \)\_4\)]],
  " and conservation of energy is represented by "
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(\(p\& \[RightVector] \)\_1\[CenterDot]\(p\& \
\[RightVector] \)\_1\)\/\(2  m\_1\)\  + \ \(\(p\& \[RightVector] \)\_2\
\[CenterDot]\(p\& \[RightVector] \)\_2\)\/\(2  m\_2\)\  + \ 
        Q\  = \ \(\(p\& \[RightVector] \)\_3\[CenterDot]\(p\& \[RightVector] \
\)\_3\)\/\(2  m\_3\)\  + \ \(\(p\& \[RightVector] \)\_4\[CenterDot]\(p\& \
\[RightVector] \)\_4\)\/\(2  m\_4\)\)], "DisplayFormula"],

Cell[TextData[{
  "where the Q-value, ",
  Cell[BoxData[
      \(TraditionalForm\`Q\)]],
  ", represents the internal energy liberated by the reaction.  Thus, the \
Q-value is positive for an exothermic reaction in which the final kinetic \
energy is greater than the initial kinetic energy , whereas  the Q-value is \
negative for an endothermic reaction in which some of the initial kinetic \
energy is converted into internal energy.  Elastic scattering requires the \
Q-value to vanish and for there to be no mass transfer."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[CellGroupData[{

Cell["Options and packages", "Subsection"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(\(Off[General::spell, \ General::spell1];\)\), "\n", 
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", 
          FontSize \[Rule] 12};\)\)}], "Input"],

Cell[BoxData[{
    \(\(Needs["\<Graphics`PlotField`\>"];\)\), "\n", 
    \(Needs["\<Graphics`Arrow`\>"]\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Definitions", "Subsection"],

Cell["\<\
Establish symbolic names for the masses, momenta, and velocities that \
describe the  reaction.  These names can be used in replacement rules but \
should not be set to specific values.\
\>", "Text"],

Cell[BoxData[{
    \(\(masses = {m1, m2, m3, m4};\)\), "\n", 
    \(\(momenta = {p1, p2, p3, p4};\)\), "\n", 
    \(\(velocities = {v1, v2, v3, v4};\)\)}], "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Implementation", "Section"],

Cell[CellGroupData[{

Cell["Procedures for drawing vectors", "Subsection"],

Cell[BoxData[
    \(describeVector[vector_?VectorQ, position_: {0, 0}, 
        style_: {HeadScaling \[Rule] Absolute, HeadLength \[Rule] 5.0, 
            HeadWidth \[Rule] 0.5}] := {vector, position, style}\)], "Input"],

Cell[BoxData[
    \(drawVector[VectorStuff_] := 
      Module[{vector, position, style}, 
        vector = VectorStuff\[LeftDoubleBracket]1\[RightDoubleBracket]; 
        position = VectorStuff\[LeftDoubleBracket]2\[RightDoubleBracket]; 
        style = VectorStuff\[LeftDoubleBracket]3\[RightDoubleBracket] /. 
            List \[Rule] Sequence; 
        Graphics[Arrow[position, position + vector, style]]]\)], "Input"],

Cell[BoxData[
    \(drawVectors[VectorList_] := \((drawVector[#1] &)\) /@ 
        VectorList\)], "Input"],

Cell[BoxData[
    \(\(drawVectorSum[v1_?VectorQ, v2_?VectorQ] := 
        Module[{position1, position2, position3, style1, style2, style3, v3, 
            vector1, vector2, vector3}, position1 = \(position3 = {0, 0}\); 
          position2 = v1; v3 = v1 + v2; 
          style1 = \(style2 = \(style3 = {HeadScaling \[Rule] Absolute, 
                  HeadLength \[Rule] 5.0, HeadWidth \[Rule] 0.5}\)\); 
          vector1 = describeVector[v1, position1, style1]; 
          vector2 = describeVector[v2, position2, style2]; 
          vector3 = describeVector[v3, position3, style3]; 
          Show[{drawVectors[{vector1, vector2}], 
              Graphics[AbsoluteThickness[1]], drawVector[vector3]}, 
            AspectRatio \[Rule] Automatic]];\)\)], "Input"],

Cell[BoxData[
    \(\(drawVectorDifference[v1_?VectorQ, v2_?VectorQ] := 
        Module[{position1, position2, position3, style1, style2, style3, v3, 
            vector1, vector2, vector3}, position1 = \(position2 = {0, 0}\); 
          position3 = v1; v3 = v2 - v1; 
          style1 = \(style2 = \(style3 = {HeadScaling \[Rule] Absolute, 
                  HeadLength \[Rule] 5.0, HeadWidth \[Rule] 0.5}\)\); 
          vector1 = describeVector[v1, position1, style1]; 
          vector2 = describeVector[v2, position2, style2]; 
          vector3 = describeVector[v3, position3, style3]; 
          Show[{drawVectors[{vector1, vector2}], 
              Graphics[AbsoluteThickness[1]], drawVector[vector3]}, 
            AspectRatio \[Rule] Automatic]];\)\)], "Input"],

Cell[BoxData[
    \(\(drawCollision[p1i_?VectorQ, p2i_?VectorQ, p1f_?VectorQ, 
          p2f_?VectorQ] := 
        Module[{vector1i, vector2i, vector1f, vector2f}, 
          vector1i = describeVector[p1i, \(-p1i\)]; 
          vector2i = describeVector[p2i, \(-p2i\)]; 
          vector1f = describeVector[p1f]; vector2f = describeVector[p2f]; 
          Show[drawVectors[{vector1i, vector2i, vector1f, vector2f}], 
            AspectRatio \[Rule] Automatic]];\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Procedures for calculating and drawing collisions", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
Establish equations expressing conservation of mass, momentum and energy\
\>", "Subsubsection"],

Cell[BoxData[
    \(\(collisionEqs = {m1 + m2 \[Equal] m3 + m4, p1 + p2 \[Equal] p3 + p4, 
          p1 . p1\/\(2\ m1\) + p2 . p2\/\(2\ m2\) + qvalue \[Equal] 
            p3 . p3\/\(2\ m3\) + p4 . p4\/\(2\ m4\)};\)\)], "Input"],

Cell[BoxData[
    \(\(momentumComponentRules = {p1 \[Rule] {p1x, p1y}, 
          p2 \[Rule] {p2x, p2y}, p3 \[Rule] {p3x, p3y}, 
          p4 \[Rule] {p4x, p4y}, p1x \[Rule] p1r\ Cos[theta1], 
          p1y \[Rule] p1r\ Sin[theta1], p2x \[Rule] p2r\ Cos[theta2], 
          p2y \[Rule] p2r\ Sin[theta2], p3x \[Rule] p3r\ Cos[theta3], 
          p3y \[Rule] p3r\ Sin[theta3]};\)\)], "Input"],

Cell[BoxData[{
    \(\(magnitude[p_?VectorQ] := \@\(p . p\);\)\), "\[IndentingNewLine]", 
    \(angle[p_?VectorQ] := 
      ArcTan[p\[LeftDoubleBracket]2\[RightDoubleBracket], 
        p\[LeftDoubleBracket]1\[RightDoubleBracket]]\)}], "Input"],

Cell[BoxData[
    \(cmVelocity[masses_, 
        momenta_] := \(momenta\[LeftDoubleBracket]1\[RightDoubleBracket] + \
momenta\[LeftDoubleBracket]2\[RightDoubleBracket]\)\/\(masses\
\[LeftDoubleBracket]1\[RightDoubleBracket] + masses\[LeftDoubleBracket]2\
\[RightDoubleBracket]\)\)], "Input"],

Cell[BoxData[
    \(\(LabToCM[masses_, momenta_] := 
        Module[{Vcm, velocities}, Vcm = cmVelocity[masses, momenta]; 
          velocities = momenta\/masses - Vcm; 
          Simplify[{momenta \[Rule] masses\ velocities}]];\)\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Define general solution and procedures for plotting momenta and angles\
\>", "Subsubsection"],

Cell[BoxData[
    \(\(generalSolution[p3r, p4x, p4y] = 
        Solve[collisionEqs //. momentumComponentRules, {m4, p3r, p4x, 
            p4y}];\)\)], "Input"],

Cell[BoxData[{
    \(\(\(finalMomenta::usage\  = \ "\<finalMomenta[angle,sol,values] \
calculates the ejectile and recoil momenta given the scattering angle (in \
degrees) and the solution and initialization rules, sol and values.\>";\)\(\n\
\)
    \)\), "\n", 
    \(finalMomenta[angle_, sol_, 
        values_] := \(\(\({magnitude[p3], magnitude[p4]} //. 
              momentumComponentRules\) /. sol\) //. 
          values\) /. {theta3 \[Rule] angle\ Degree}\)}], "Input"],

Cell[BoxData[{
    \(\(\(plotFinalMomenta::usage\  = \ "\<plotFinalMomenta[sol,values,thmax] \
plots the ejectile and recoil momenta for angles up to thmax (in degrees) \
given solution and initialization rules, sol and values.\>";\)\(\n\)
    \)\), "\n", 
    \(\(plotFinalMomenta[sol_, values_, thmax_] := 
        Plot[Evaluate[finalMomenta[theta3, sol, values]], {theta3, 0, thmax}, 
          Frame \[Rule] True, GridLines \[Rule] Automatic, 
          FrameLabel \[Rule] {\*"\"\<\!\(\[Theta]\_3\) [deg]\>\"", \
\*"\"\<\!\(p\_3\)/\!\(p\_1\), \!\(p\_4\)/\!\(p\_1\)\>\""}, 
          PlotLabel \[Rule] "\<scattered and recoil momenta\>"];\)\)}], \
"Input"],

Cell[BoxData[{
    \(\(\(recoilAngle::usage\  = \ "\<recoilAngle[angle,sol,values] \
calculates the recoil angle (in degrees) given a scattering angle (in \
degrees) and the replacement rules that define a solution and the collision \
parameters.\>";\)\(\n\)
    \)\), "\n", 
    \(recoilAngle[angle_, sol_, 
        values_] := \(\(ArcTan[p4x, \(-p4y\)]\/Degree /. sol\) //. 
          values\) /. {theta3 \[Rule] angle\ Degree}\)}], "Input"],

Cell[BoxData[{
    \(\(\(plotRecoilAngle::usage\  = \ "\<plotRecoilAngle[sol,values,thmax]  \
plots the recoil angle (in degrees) given solution and initialization rules, \
sol and values.  The maximum scattering angle plotted is thmax (in \
degrees).\>";\)\(\n\)
    \)\), "\n", 
    \(\(plotRecoilAngle[sol_, values_, thmax_] := 
        Plot[Evaluate[recoilAngle[theta3, sol, values]], {theta3, 0.0001, 
            thmax}, Frame \[Rule] True, GridLines \[Rule] Automatic, 
          FrameLabel \[Rule] {\*"\"\<\!\(\[Theta]\_3\) [deg]\>\"", \
\*"\"\<\!\(\[Theta]\_4\) [deg]\>\""}, 
          PlotLabel \[Rule] "\<recoil angle\>"];\)\)}], "Input"],

Cell[BoxData[{
    \(\(\(calcCollision::usage\  = \ "\<calcCollision[sol,values,LabAngle]  \
evaluates many quantities describing a collision given the solution \
replacement rules, sol, the list of rules defining the masses and initial \
conditions, values, the laboratory angle for particle 3, LabAngle.  The \
following quantities are calculated using this information: \n LabMomenta = \
list of laboratory momenta \n LabVelocities = list of laboratory velocities \n\
 cmMomenta = list of cm momenta \n cmVelocities = list of cm velocities \n \
mui,muf = initial, final reduced masses \n vCM, pCM = velocity, momentum of \
CM \n vi,vf = initial, final relative velocities \n pi,pf = initial, final \
relative momenta in lab \n ki,kf = initial, final relative momenta in cm \>";\
\)\(\n\)
    \)\), "\n", 
    \(\(calcCollision[sol_, values_, LabAngle_] := 
        Module[{}, 
          LabMomenta = \(\(\(momenta //. momentumComponentRules\) /. sol\) //. 
                values\) /. theta3 \[Rule] LabAngle; 
          LabVelocities = \(LabMomenta\/masses /. sol\) //. values; 
          cmMomenta = \(\(\(\(momenta /. LabToCM[masses, momenta]\) //. 
                    momentumComponentRules\) /. sol\) //. values\) /. 
              theta3 \[Rule] LabAngle; 
          cmVelocities = \(cmMomenta\/masses /. sol\) //. values; {mui, muf, 
              vi, vf, pi, pf, vCM, 
              pCM} = \({\(masses\[LeftDoubleBracket]1\[RightDoubleBracket]\ \
masses\[LeftDoubleBracket]2\[RightDoubleBracket]\)\/\(masses\
\[LeftDoubleBracket]1\[RightDoubleBracket] + masses\[LeftDoubleBracket]2\
\[RightDoubleBracket]\), \(masses\[LeftDoubleBracket]3\[RightDoubleBracket]\ \
masses\[LeftDoubleBracket]4\[RightDoubleBracket]\)\/\(masses\
\[LeftDoubleBracket]3\[RightDoubleBracket] + masses\[LeftDoubleBracket]4\
\[RightDoubleBracket]\), 
                  LabVelocities\[LeftDoubleBracket]1\[RightDoubleBracket] - 
                    LabVelocities\[LeftDoubleBracket]2\[RightDoubleBracket], 
                  LabVelocities\[LeftDoubleBracket]3\[RightDoubleBracket] - 
                    LabVelocities\[LeftDoubleBracket]4\[RightDoubleBracket], 
                  LabMomenta\[LeftDoubleBracket]1\[RightDoubleBracket] - 
                    LabMomenta\[LeftDoubleBracket]2\[RightDoubleBracket], 
                  LabMomenta\[LeftDoubleBracket]3\[RightDoubleBracket] - 
                    LabMomenta\[LeftDoubleBracket]4\[RightDoubleBracket], \
\(LabMomenta\[LeftDoubleBracket]1\[RightDoubleBracket] + LabMomenta\
\[LeftDoubleBracket]2\[RightDoubleBracket]\)\/\(masses\[LeftDoubleBracket]1\
\[RightDoubleBracket] + masses\[LeftDoubleBracket]2\[RightDoubleBracket]\), 
                  LabMomenta\[LeftDoubleBracket]1\[RightDoubleBracket] + 
                    LabMomenta\[LeftDoubleBracket]2\[RightDoubleBracket]} /. 
                sol\) //. values; ki = mui\ vi; kf = muf\ vf;];\)\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Procedures for drawing vector relationships", "Subsubsection"],

Cell[BoxData[{
    \(\(\(PlotCMmomenta::usage\  = \ "\<PlotCMmomenta[cmMomenta]  produces a \
plot which displays the initial and final momenta in the cm frame together \
with circles of constant momentum.\>";\)\(\n\)
    \)\), "\n", 
    \(\(PlotCMmomenta[cmMomenta_] := 
        Module[{pi, pf, vectorList}, 
          pi = 1\/2\ \((cmMomenta\[LeftDoubleBracket]1\[RightDoubleBracket] - 
                  cmMomenta\[LeftDoubleBracket]2\[RightDoubleBracket])\); 
          pf = 1\/2\ \((cmMomenta\[LeftDoubleBracket]3\[RightDoubleBracket] - 
                  cmMomenta\[LeftDoubleBracket]4\[RightDoubleBracket])\); 
          vectorList = \((describeVector[#1] &)\) /@ cmMomenta; {drawVectors[
              vectorList], 
            Graphics[{Circle[{0, 0}, magnitude[pi]], 
                Circle[{0, 0}, magnitude[pf]], AbsolutePointSize[3], 
                Point[{0, 0}]}]}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(\(VelocityPlot2::usage\  = \ \
"\<VelocityPlot2[LabVelocities,vCM,cmVelocities] produces two plots, named \
plot1 and plot2, which display velocity relationships in the initial and \
final states, respectively.  Each plot contains two lab velocity vectors and \
the cm velocity vector based at the origin, and the two velocity vectors in \
the cm frame based at the cm velocity.\>";\)\(\n\)
    \)\), "\n", 
    \(\(VelocityPlot2[LabVelocities_, vCM_, cmVelocities_] := 
        Module[{initialVelocities, finalVelocities}, 
          initialVelocities = 
            Join[\((describeVector[#1] &)\) /@ {LabVelocities\
\[LeftDoubleBracket]1\[RightDoubleBracket], 
                  LabVelocities\[LeftDoubleBracket]2\[RightDoubleBracket], 
                  vCM}, {describeVector[
                  cmVelocities\[LeftDoubleBracket]1\[RightDoubleBracket], 
                  vCM], describeVector[
                  cmVelocities\[LeftDoubleBracket]2\[RightDoubleBracket], 
                  vCM]}]; 
          finalVelocities = 
            Join[\((describeVector[#1] &)\) /@ {LabVelocities\
\[LeftDoubleBracket]3\[RightDoubleBracket], 
                  LabVelocities\[LeftDoubleBracket]4\[RightDoubleBracket], 
                  vCM}, {describeVector[
                  cmVelocities\[LeftDoubleBracket]3\[RightDoubleBracket], 
                  vCM], describeVector[
                  cmVelocities\[LeftDoubleBracket]4\[RightDoubleBracket], 
                  vCM]}]; plot1 = drawVectors[initialVelocities]; 
          plot2 = drawVectors[finalVelocities];];\)\)}], "Input"],

Cell[BoxData[{
    \(\(\(MomentumPlot2::usage\  = \ "\<MomentumPlot2[LabMomenta] produces \
two figures: plot1 shows the sum of initial momentum vectors and plot2 shows \
the sum of final momentum vectors.\>";\)\(\n\)
    \)\), "\n", 
    \(\(MomentumPlot2[LabMomenta_] := 
        Module[{pCM}, 
          pCM = LabMomenta\[LeftDoubleBracket]1\[RightDoubleBracket] + 
              LabMomenta\[LeftDoubleBracket]2\[RightDoubleBracket]; 
          initialMomenta = {describeVector[
                LabMomenta\[LeftDoubleBracket]1\[RightDoubleBracket]], 
              describeVector[
                LabMomenta\[LeftDoubleBracket]2\[RightDoubleBracket], 
                LabMomenta\[LeftDoubleBracket]1\[RightDoubleBracket]], 
              describeVector[pCM], 
              describeVector[
                cmMomenta\[LeftDoubleBracket]1\[RightDoubleBracket], pCM], 
              describeVector[
                cmMomenta\[LeftDoubleBracket]2\[RightDoubleBracket], pCM]}; 
          finalMomenta = {describeVector[
                LabMomenta\[LeftDoubleBracket]3\[RightDoubleBracket]], 
              describeVector[
                LabMomenta\[LeftDoubleBracket]4\[RightDoubleBracket], 
                LabMomenta\[LeftDoubleBracket]3\[RightDoubleBracket]], 
              describeVector[pCM], 
              describeVector[
                cmMomenta\[LeftDoubleBracket]3\[RightDoubleBracket], pCM], 
              describeVector[
                cmMomenta\[LeftDoubleBracket]4\[RightDoubleBracket], pCM]}; 
          plot1 = drawVectors[initialMomenta]; 
          plot2 = drawVectors[finalMomenta];];\)\)}], "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Simplification rules", "Subsection"],

Cell[BoxData[{
    \(\(MyAssumptions \[Rule] {m1 > 0, m2 > 0, m3 > 0, m4 > 0, mtotal > 0, 
          mui > 0, muf > 0, qvalue \[Element] Reals, vi \[Element] Reals, 
          vf \[Element] Reals};\)\), "\[IndentingNewLine]", 
    \(\(MySimplify = 
        Simplify[#, MyAssumptions] &;\)\), "\[IndentingNewLine]", 
    \(\(MyFullSimplify = FullSimplify[#, MyAssumptions] &;\)\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Rules for lab to cm transformations", "Subsection"],

Cell[CellGroupData[{

Cell["Define reduced masses, relative and cm velocities", "Subsubsection"],

Cell[BoxData[
    \(\(Clear[m1, m2, m3, m4, v1, v2, v2, v4, mui, muf, vi, vf, vCM, 
        pCM];\)\)], "Input"],

Cell[BoxData[
    \(\(reducedMassRules = {mui \[Rule] \(m1\ m2\)\/\(m1 + m2\), 
          muf \[Rule] \(m3\ m4\)\/\(m3 + m4\), 
          mtotal \[Rule] m1 + m2};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(vcmRules = {v1cm \[Rule] v1 - \(m1\ v1 + m2\ v2\)\/\(m1 + m2\), 
            v2cm \[Rule] v2 - \(m1\ v1 + m2\ v2\)\/\(m1 + m2\), 
            v3cm \[Rule] v3 - \(m3\ v3 + m4\ v4\)\/\(m3 + m4\), 
            v4cm \[Rule] 
              v4 - \(m3\ v3 + m4\ v4\)\/\(m3 + m4\)} /. {v1 \[Rule] vi + v2, 
            v3 \[Rule] vf + v4}\  // Simplify\)], "Input"],

Cell[BoxData[
    \({v1cm \[Rule] \(m2\ vi\)\/\(m1 + m2\), 
      v2cm \[Rule] \(-\(\(m1\ vi\)\/\(m1 + m2\)\)\), 
      v3cm \[Rule] \(m4\ vf\)\/\(m3 + m4\), 
      v4cm \[Rule] \(-\(\(m3\ vf\)\/\(m3 + m4\)\)\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(vLabToCM = 
      Join[Simplify[\(Solve[{vi \[Equal] v1 - v2, 
                vCM \[Equal] \(m1\ v1 + m2\ v2\)\/\(m1 + m2\)}, {v1, 
                v2}]\)\[LeftDoubleBracket]1\[RightDoubleBracket]], 
        Simplify[\(Solve[{vf \[Equal] v3 - v4, 
                vCM \[Equal] \(m3\ v3 + m4\ v4\)\/\(m3 + m4\)}, {v3, 
                v4}]\)\[LeftDoubleBracket]1\[RightDoubleBracket]]]\)], "Input"],

Cell[BoxData[
    \({v1 \[Rule] \(m1\ vCM + m2\ \((vCM + vi)\)\)\/\(m1 + m2\), 
      v2 \[Rule] \(m1\ vCM + m2\ vCM - m1\ vi\)\/\(m1 + m2\), 
      v3 \[Rule] \(m3\ vCM + m4\ \((vCM + vf)\)\)\/\(m3 + m4\), 
      v4 \[Rule] \(m3\ vCM + m4\ vCM - m3\ vf\)\/\(m3 + m4\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(masses\ velocities /. vLabToCM\ \  // 
        Collect[#, {vi, vf, vcm}] &\)\  // Simplify\)], "Input"],

Cell[BoxData[
    \({\(m1\ \((m1\ vCM + m2\ \((vCM + vi)\))\)\)\/\(m1 + m2\), \(m2\ \((m2\ \
vCM + m1\ \((vCM - vi)\))\)\)\/\(m1 + m2\), \(m3\ \((m3\ vCM + m4\ \((vCM + \
vf)\))\)\)\/\(m3 + m4\), \(m4\ \((m4\ vCM + m3\ \((vCM - vf)\))\)\)\/\(m3 + \
m4\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(massRules = \(Solve[{mtotal \[Equal] m1 + m2 \[Equal] m3 + m4, 
            mui \[Equal] \(m1\ m2\)\/\(m1 + m2\), 
            muf \[Equal] \(m3\ m4\)\/\(m3 + m4\)}, {m1, m2, m3, 
            m4}]\)\[LeftDoubleBracket]1\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \({m1 \[Rule] 1\/2\ \((mtotal - \@mtotal\ \@\(mtotal - 4\ mui\))\), 
      m2 \[Rule] 1\/2\ \((mtotal + \@mtotal\ \@\(mtotal - 4\ mui\))\), 
      m3 \[Rule] 1\/2\ \((mtotal - \@mtotal\ \@\(mtotal - 4\ muf\))\), 
      m4 \[Rule] 
        1\/2\ \((mtotal + \@mtotal\ \@\(mtotal - 4\ muf\))\)}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Express cm momenta in terms of relative velocities and reduced masses\
\>", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(stuff = 
      Eliminate[{p1cm \[Equal] m1\ v1cm, p2cm \[Equal] m2\ v2cm, 
            p3cm \[Equal] m3\ v3cm, p4cm \[Equal] m4\ v4cm, 
            mtotal \[Equal] m1 + m2 \[Equal] m3 + m4, 
            mui \[Equal] \(m1\ m2\)\/mtotal, 
            muf \[Equal] \(m3\ m4\)\/mtotal} /. vcmRules, {m1, m2, m3, m4, 
          v1, v2, v3, v4}]\)], "Input"],

Cell[BoxData[
    \(p1cm \[Equal] \(-p2cm\) && p3cm \[Equal] \(-p4cm\) && 
      muf\ vf \[Equal] \(-p4cm\) && mui\ vi \[Equal] \(-p2cm\) && 
      mtotal \[NotEqual] 0\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(pcmRules = \(Solve[
          stuff, {p1cm, p2cm, p3cm, 
            p4cm}]\)\[LeftDoubleBracket]1\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \({p1cm \[Rule] mui\ vi, p2cm \[Rule] \(-mui\)\ vi, p3cm \[Rule] muf\ vf, 
      p4cm \[Rule] \(-muf\)\ vf}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Separate kinetic energy into internal and cm contributions", \
"Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Solve[
        Eliminate[{Simplify[
              KiLab \[Equal] \((Collect[#1, {vi, 
                          vCM}] &)\)[\(m1\ v1\^2\)\/2 + \(m2\ v2\^2\)\/2 /. 
                    vLabToCM]], mtotal \[Equal] m1 + m2, 
            mui \[Equal] \(m1\ m2\)\/mtotal}, {m1, m2}], 
        KiLab]\)\[LeftDoubleBracket]1\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \({KiLab \[Rule] 1\/2\ \((mtotal\ vCM\^2 + mui\ vi\^2)\)}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The first term is the kinetic energy associated with the motion of the \
center of mass,  whereas the second is the kinetic energy due to the internal \
motions relative to the center of mass.  The corresponding relationship for \
the final state is obviously obtained by the replacements ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(mui \[Rule] muf\),
          "InlineInput"], TraditionalForm]]],
  ", ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(vi \[Rule] vf\),
          "InlineInput"], TraditionalForm]]],
  ".  Therefore, we can define the following rules."
}], "Text"],

Cell[BoxData[
    \(\(kineticEnergyRules = {KiLab \[Rule] KCM + KiCM, 
          KfLab \[Rule] KCM + KfCM, KCM \[Rule] \(mtotal\ vCM\^2\)\/2, 
          KiCM \[Rule] \(mui\ vi\^2\)\/2, 
          KfCM \[Rule] \(muf\ vf\^2\)\/2};\)\)], "Input"],

Cell[TextData[{
  "Conservation of energy can now be analyzed using the relative velocities \
and the ",
  StyleBox["qvalue", "InlineInput"],
  ".  For example, we can solve for the relative velocity in the final state \
in terms of the initial relative velocity, the qvalue, and the reduced \
masses.  Furthermore, because the momenta of the outgoing particles must be \
equal and opposite in the center of mass or zero momentum frame, the cm \
velocities of each particle are inversely proportional to their masses.  Note \
that since these results are independent of scattering angle, we can develop \
simple relationships for the magnitudes of the cm vectors as follows."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(vfRule = \(Solve[
            KfLab - KiLab - qvalue \[Equal] 0 //. kineticEnergyRules, 
            vf]\)\[LeftDoubleBracket]2\[RightDoubleBracket] // 
        Simplify\)], "Input"],

Cell[BoxData[
    \({vf \[Rule] \@\(2\ qvalue + mui\ vi\^2\)\/\@muf}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\({p3cm, p4cm} /. pcmRules\) /. vfRule\)], "Input"],

Cell[BoxData[
    \({\@muf\ \@\(2\ qvalue + mui\ vi\^2\), \(-\@muf\)\ \@\(2\ qvalue + mui\ \
vi\^2\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[\(\({v3cm, v4cm}\ \((m3 + m4)\)\)\/mtotal /. vcmRules\) /. 
        vfRule]\)], "Input"],

Cell[BoxData[
    \({\(m4\ \@\(2\ qvalue + mui\ vi\^2\)\)\/\(mtotal\ \@muf\), \(-\(\(m3\ \@\
\(2\ qvalue + mui\ vi\^2\)\)\/\(mtotal\ \@muf\)\)\)}\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples", "Section"],

Cell[CellGroupData[{

Cell["Drawing vectors", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[AbsoluteThickness[1]], 
        drawVector[describeVector[{1, 3}]]]\ ;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.196201 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
1 Mabswid
[ ] 0 setdash
.02381 .01472 m
.97619 .60332 L
s
% Start of user PostScript
0. 0. 1. 3. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`002P3oool00`000000oooo0?ooo`3o0?oooj@0oooo000;0?ooo`800000o`3ooonT0?ooo`00
3@3oool200000?l0ooooXP3oool000l0oooo00<000000?ooo`3oool0o`3ooonO0?ooo`00403oool2
00000?l0ooooW`3oool00180oooo00<000000?ooo`3oool0o`3ooonL0?ooo`004`3oool200000?l0
ooooW03oool001D0oooo0P00003o0?oooiX0oooo000G0?ooo`030000003oool0oooo0?l0ooooU`3o
ool001P0oooo0P00003o0?oooiL0oooo000J0?ooo`800000o`3ooonE0?ooo`00703oool00`000000
oooo0?ooo`3o0?oooi80oooo000M0?ooo`800000o`3ooonB0?ooo`007`3oool00`000000oooo0?oo
o`3o0?ooohl0oooo000P0?ooo`800000o`3ooon?0?ooo`008P3oool200000?l0ooooS@3oool002@0
oooo00<000000?ooo`3oool0o`3ooon:0?ooo`009@3oool200000?l0ooooRP3oool002L0oooo00<0
00000?ooo`3oool0o`3ooon70?ooo`00:03oool200000?l0ooooQ`3oool002X0oooo0P00003o0?oo
ohD0oooo000/0?ooo`030000003oool0oooo0?l0ooooPP3oool002d0oooo0P00003o0?oooh80oooo
000_0?ooo`800000o`3ooon00?ooo`00<@3oool00`000000oooo0?ooo`3o0?ooogd0oooo000b0?oo
o`800000o`3ooomm0?ooo`00=03oool00`000000oooo0?ooo`3o0?ooogX0oooo000e0?ooo`800000
o`3ooomj0?ooo`00=`3oool200000?l0ooooN03oool003T0oooo00<000000?ooo`3oool0o`3ooome
0?ooo`00>P3oool200000?l0ooooM@3oool003`0oooo0P00003o0?ooog<0oooo000n0?ooo`030000
003oool0oooo0?l0ooooL03oool003l0oooo0P00003o0?ooog00oooo00110?ooo`030000003oool0
oooo0?l0ooooK@3oool00480oooo0P00003o0?ooofd0oooo00140?ooo`800000o`3ooom[0?ooo`00
AP3oool00`000000oooo0?ooo`3o0?ooofP0oooo00170?ooo`800000o`3ooomX0?ooo`00B@3oool2
00000?l0ooooIP3oool004/0oooo00<000000?ooo`3oool0o`3ooomS0?ooo`00C03oool200000?l0
ooooH`3oool004h0oooo00<000000?ooo`3oool0o`3ooomP0?ooo`00C`3oool200000?l0ooooH03o
ool00540oooo0P00003o0?oooeh0oooo001C0?ooo`030000003oool0oooo0?l0ooooF`3oool005@0
oooo0P00003o0?oooe/0oooo001F0?ooo`030000003oool0oooo0?l0ooooF03oool005L0oooo0P00
003o0?oooeP0oooo001I0?ooo`800000o`3ooomF0?ooo`00F`3oool00`000000oooo0?ooo`3o0?oo
oe<0oooo001L0?ooo`800000o`3ooomC0?ooo`00GP3oool200000?l0ooooD@3oool00600oooo00<0
00000?ooo`3oool0o`3ooom>0?ooo`00H@3oool200000?l0ooooCP3oool006<0oooo00<000000?oo
o`3oool0o`3ooom;0?ooo`00I03oool200000?l0ooooB`3oool006H0oooo0P00003o0?ooodT0oooo
001X0?ooo`030000003oool0oooo0?l0ooooAP3oool006T0oooo0P00003o0?ooodH0oooo001[0?oo
o`800000o`3ooom40?ooo`00K@3oool00`000000oooo0?ooo`3o0?oood40oooo001^0?ooo`800000
o`3ooom10?ooo`00L03oool00`000000oooo0?ooo`3o0?oooch0oooo001a0?ooo`800000o`3oooln
0?ooo`00L`3oool200000?l0oooo?03oool007D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`00
MP3oool200000?l0oooo>@3oool007P0oooo0P00003o0?ooocL0oooo001j0?ooo`030000003oool0
oooo0?l0oooo=03oool007/0oooo0P00003o0?oooc@0oooo001m0?ooo`030000003oool0oooo0?l0
oooo<@3oool007h0oooo0P00003o0?oooc40oooo00200?ooo`800000o`3oool_0?ooo`00PP3oool0
0`000000oooo0?ooo`3o0?ooob`0oooo00230?ooo`800000o`3oool/0?ooo`00Q@3oool00`000000
oooo0?ooo`3o0?ooobT0oooo00260?ooo`800000o`3ooolY0?ooo`00R03oool200000?l0oooo9`3o
ool008X0oooo00<000000?ooo`3oool0o`3ooolT0?ooo`00R`3oool200000?l0oooo903oool008d0
oooo0P00003o0?ooob80oooo002?0?ooo`030000003oool0oooo0?l0oooo7`3oool00900oooo0P00
003o0?oooal0oooo002B0?ooo`030000003oool0oooo0?l0oooo703oool009<0oooo0P00003o0?oo
oa`0oooo002E0?ooo`800000o`3ooolJ0?ooo`00U`3oool00`000000oooo0?ooo`3o0?oooaL0oooo
002H0?ooo`800000o`3ooolG0?ooo`00VP3oool200000?l0oooo5@3oool009`0oooo00<000000?oo
o`3oool0o`3ooolB0?ooo`00W@3oool200000?l0oooo4P3oool009l0oooo00<000000?ooo`3oool0
o`3oool?0?ooo`00X03oool200000?l0oooo3`3oool00:80oooo0P00003o0?ooo`d0oooo002T0?oo
o`030000003oool0oooo0?l0oooo2P3oool00:D0oooo0P00003o0?ooo`X0oooo002W0?ooo`800000
o`3oool80?ooo`00Z@3oool00`000000oooo0?ooo`3o0?ooo`D0oooo002Z0?ooo`800000o`3oool5
0?ooo`00[03oool00`000000oooo0?ooo`3o0?ooo`80oooo002]0?ooo`800000o`3oool20?ooo`00
[`3oool200000?l0oooo002a0?ooo`030000003oool0oooo0?`0oooo002b0?ooo`800000o03oool0
0;@0oooo00<000000?ooo`3oool0n@3oool00;D0oooo0P00003i0?ooo`00]`3oool200000?L0oooo
002i0?ooo`030000003oool0oooo0?@0oooo002j0?ooo`800000m03oool00;`0oooo0P00003b0?oo
o`00_P3oool00`000000oooo0?ooo`3_0?ooo`00_`3oool200000>l0oooo00310?ooo`030000003o
ool0oooo0>`0oooo00320?ooo`800000k03oool00<@0oooo0P00003Z0?ooo`00aP3oool00`000000
oooo0?ooo`3W0?ooo`00a`3oool200000>L0oooo00390?ooo`800000i@3oool00</0oooo00<00000
0?ooo`3oool0hP3oool00<`0oooo0P00003R0?ooo`00cP3oool00`000000oooo0?ooo`3O0?ooo`00
c`3oool200000=l0oooo003A0?ooo`800000g@3oool00=<0oooo00<000000?ooo`3oool0fP3oool0
0=@0oooo0P00003J0?ooo`00eP3oool00`000000oooo0?ooo`3G0?ooo`00e`3oool200000=L0oooo
003I0?ooo`800000e@3oool00=/0oooo00<000000?ooo`3oool0dP3oool00=`0oooo0P00003B0?oo
o`00gP3oool200000=00oooo003P0?ooo`030000003oool0oooo0<d0oooo003Q0?ooo`800000c@3o
ool00><0oooo00<000000?ooo`3oool0bP3oool00>@0oooo0P00003:0?ooo`00iP3oool200000<P0
oooo003X0?ooo`030000003oool0oooo0<D0oooo003Y0?ooo`800000a@3oool00>/0oooo0P000033
0?ooo`00k@3oool00`000000oooo0?ooo`300?ooo`00kP3oool200000<00oooo003`0?ooo`030000
003oool0oooo0;d0oooo003a0?ooo`800000_@3oool00?<0oooo0P00002k0?ooo`00m@3oool00`00
0000oooo0?ooo`2h0?ooo`00mP3oool200000;P0oooo003h0?ooo`800000]P3oool00?X0oooo00<0
00000?ooo`3oool0/`3oool00?/0oooo0P00002c0?ooo`00o@3oool00`000000oooo0?ooo`2`0?oo
o`00oP3oool200000;00oooo003o0?ooo`40oooo0P00002^0?ooo`00o`3oool30?ooo`030000003o
ool0oooo0:/0oooo003o0?ooo`@0oooo0P00002[0?ooo`00o`3oool60?ooo`030000003oool0oooo
0:P0oooo003o0?ooo`L0oooo0P00002X0?ooo`00o`3oool90?ooo`800000YP3oool00?l0oooo2`3o
ool00`000000oooo0?ooo`2S0?ooo`00o`3oool<0?ooo`800000X`3oool00?l0oooo3P3oool20000
0:40oooo003o0?oooa00oooo00<000000?ooo`3oool0WP3oool00?l0oooo4@3oool2000009h0oooo
003o0?oooa<0oooo00<000000?ooo`3oool0V`3oool00?l0oooo503oool2000009/0oooo003o0?oo
oaH0oooo0P00002I0?ooo`00o`3ooolH0?ooo`030000003oool0oooo09H0oooo003o0?oooaT0oooo
0P00002F0?ooo`00o`3ooolK0?ooo`800000U03oool00?l0oooo7@3oool00`000000oooo0?ooo`2A
0?ooo`00o`3ooolN0?ooo`800000T@3oool00?l0oooo803oool00`000000oooo0?ooo`2>0?ooo`00
o`3ooolQ0?ooo`800000SP3oool00?l0oooo8`3oool2000008`0oooo003o0?ooobD0oooo00<00000
0?ooo`3oool0R@3oool00?l0oooo9P3oool2000008T0oooo003o0?ooobP0oooo0P0000270?ooo`00
o`3ooolZ0?ooo`030000003oool0oooo08@0oooo003o0?ooob/0oooo0P0000240?ooo`00o`3oool]
0?ooo`030000003oool0oooo0840oooo003o0?ooobh0oooo0P0000210?ooo`00o`3oool`0?ooo`80
0000O`3oool00?l0oooo<P3oool00`000000oooo0?ooo`1l0?ooo`00o`3ooolc0?ooo`800000O03o
ool00?l0oooo=@3oool00`000000oooo0?ooo`1i0?ooo`00o`3ooolf0?ooo`800000N@3oool00?l0
oooo>03oool2000007L0oooo003o0?ooocX0oooo00<000000?ooo`3oool0M03oool00?l0oooo>`3o
ool2000007@0oooo003o0?ooocd0oooo0P00001b0?ooo`00o`3ooolo0?ooo`030000003oool0oooo
06l0oooo003o0?oood00oooo0P00001_0?ooo`00o`3ooom20?ooo`030000003oool0oooo06`0oooo
003o0?oood<0oooo0P00001/0?ooo`00o`3ooom50?ooo`800000JP3oool00?l0ooooA`3oool00`00
0000oooo0?ooo`1W0?ooo`00o`3ooom80?ooo`800000I`3oool00?l0ooooBP3oool2000006D0oooo
003o0?oood`0oooo00<000000?ooo`3oool0HP3oool00?l0ooooC@3oool200000680oooo003o0?oo
odl0oooo00<000000?ooo`3oool0G`3oool00?l0ooooD03oool2000005l0oooo003o0?oooe80oooo
0P00001M0?ooo`00o`3ooomD0?ooo`030000003oool0oooo05X0oooo003o0?oooeD0oooo0P00001J
0?ooo`00o`3ooomG0?ooo`800000F03oool00?l0ooooF@3oool00`000000oooo0?ooo`1E0?ooo`00
o`3ooomJ0?ooo`800000E@3oool00?l0ooooG03oool00`000000oooo0?ooo`1B0?ooo`00o`3ooomM
0?ooo`800000DP3oool00?l0ooooG`3oool200000500oooo003o0?ooof40oooo00<000000?ooo`3o
ool0C@3oool00?l0ooooHP3oool2000004d0oooo003o0?ooof@0oooo00<000000?ooo`3oool0BP3o
ool00?l0ooooI@3oool2000004X0oooo003o0?ooofL0oooo0P0000180?ooo`00o`3ooomY0?ooo`03
0000003oool0oooo04D0oooo003o0?ooofX0oooo0P0000150?ooo`00o`3ooom/0?ooo`800000@`3o
ool00?l0ooooKP3oool00`000000oooo0?ooo`100?ooo`00o`3ooom_0?ooo`800000@03oool00?l0
ooooL@3oool00`000000oooo0?ooo`0m0?ooo`00o`3ooomb0?ooo`800000?@3oool00?l0ooooM03o
ool2000003/0oooo003o0?ooogH0oooo00<000000?ooo`3oool0>03oool00?l0ooooM`3oool20000
03P0oooo003o0?ooogT0oooo0P00000f0?ooo`00o`3ooomk0?ooo`030000003oool0oooo03<0oooo
003o0?ooog`0oooo0P00000c0?ooo`00o`3ooomn0?ooo`030000003oool0oooo0300oooo003o0?oo
ogl0oooo0P00000`0?ooo`00o`3ooon10?ooo`800000;P3oool00?l0ooooP`3oool00`000000oooo
0?ooo`0[0?ooo`00o`3ooon40?ooo`800000:`3oool00?l0ooooQP3oool2000002T0oooo003o0?oo
ohP0oooo00<000000?ooo`3oool09P3oool00?l0ooooR@3oool2000002H0oooo003o0?oooh/0oooo
00<000000?ooo`3oool08`3oool00?l0ooooS03oool2000002<0oooo003o0?ooohh0oooo0P00000Q
0?ooo`00o`3ooon@0?ooo`030000003oool0oooo01h0oooo003o0?oooi40oooo0P00000N0?ooo`00
o`3ooonC0?ooo`030000003oool0oooo01/0oooo003o0?oooi@0oooo0P00000K0?ooo`00o`3ooonF
0?ooo`8000006@3oool00?l0ooooV03oool00`000000oooo0?ooo`0F0?ooo`00o`3ooonI0?ooo`80
00005P3oool00?l0ooooV`3oool2000001@0oooo003o0?oooid0oooo00<000000?ooo`0000004@3o
ool00?l0ooooWP3oool300000100oooo003o0?oooih0oooo1000000?0?ooo`00o`3ooonM0?ooo`L0
00003@3oool00?l0ooooW`3oool6000000`0oooo003o0?oooj<0oooo0`00000;0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.0259845, -0.0750159, \
0.00366114, 0.0177716}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawVectorSum[{1, 2}, {3, \(-5\)}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.25 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.744048 0.238095 [
[ 0 0 0 0 ]
[ 1 1.25 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.25 L
0 1.25 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .74405 m
.2619 1.22024 L
s
% Start of user PostScript
0. 0. 1. 2. MAarrowhead1
% End of user PostScript
.2619 1.22024 m
.97619 .02976 L
s
% Start of user PostScript
1. 2. 4. -3. MAarrowhead1
% End of user PostScript
1 Mabswid
.02381 .74405 m
.97619 .02976 L
s
% Start of user PostScript
0. 0. 4. -3. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{230.313, 287.875},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005I0001[b000`400?l00000o`00003oo`3ooomJ
0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ
0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooom@
0?ooo`800000203oool00?l0ooooCP3oool4000000P0oooo003o0?oood`0oooo1@0000090?ooo`00
o`3ooom:0?ooo`L000002@3oool00?l0ooooB`3oool5000000X0oooo003o0?ooodX0oooo1P00000:
0?ooo`00o`3ooom90?ooo`030000003oool0000000@000002P3oool00?l0ooooB03oool010000000
oooo0?ooo`3oool3000000/0oooo003o0?ooodH0oooo0P0000040?ooo`030000003oool0000000/0
oooo003o0?ooodD0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?ooo`00o`3o
oom40?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0303oool00?l0oooo@P3oool2
000000H0oooo00<000000?ooo`3oool03@3oool00?l0oooo@@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00d0oooo003o0?oood00oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`0>0?ooo`00o`3oooln0?ooo`8000002@3oool00`000000oooo0?ooo`0>0?ooo`00o`3o
oolm0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03`3oool00?l0oooo?03oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo0100oooo003o0?ooocX0oooo0P00000;0?oo
o`030000003oool0oooo0100oooo003o0?ooocT0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0A0?ooo`00o`3ooolh0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
4@3oool00?l0oooo=P3oool2000000d0oooo00<000000?ooo`3oool04P3oool00?l0oooo=@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01<0oooo003o0?oooc@0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`00o`3ooolb0?ooo`8000003`3oool00`000000
oooo0?ooo`0D0?ooo`00o`3ooola0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
503oool00?l0oooo<03oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01D0oooo003o
0?ooobh0oooo0P00000A0?ooo`030000003oool0oooo01H0oooo003o0?ooobd0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0F0?ooo`00o`3oool/0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool05`3oool00?l0oooo:P3oool2000001@0oooo00<000000?ooo`3oool0
5`3oool00?l0oooo:@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01P0oooo003o
0?ooobP0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0I0?ooo`00o`3ooolV0?oo
o`8000005P3oool00`000000oooo0?ooo`0I0?ooo`00o`3ooolU0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool06P3oool00?l0oooo903oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo01X0oooo003o0?ooob80oooo0P00000H0?ooo`030000003oool0oooo01/0oooo003o
0?ooob40oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0L0?ooo`00o`3ooolP0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0703oool00?l0oooo7P3oool2000001X0
oooo00<000000?ooo`3oool07@3oool00?l0oooo7@3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo01d0oooo003o0?oooa`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0N0?ooo`00o`3ooolJ0?ooo`800000703oool00`000000oooo0?ooo`0O0?ooo`00o`3ooolI0?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool07`3oool00?l0oooo603oool00`000000
oooo0?ooo`0L0?ooo`030000003oool0oooo0200oooo003o0?oooaH0oooo0P00000O0?ooo`030000
003oool0oooo0200oooo003o0?oooaD0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0Q0?ooo`00o`3ooolD0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool08P3oool0
0?l0oooo4P3oool200000240oooo00<000000?ooo`3oool08P3oool00?l0oooo4@3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo02<0oooo003o0?oooa00oooo00<000000?ooo`3oool0
8@3oool00`000000oooo0?ooo`0S0?ooo`00o`3oool>0?ooo`8000008`3oool00`000000oooo0?oo
o`0T0?ooo`00o`3oool=0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool09@3oool0
0?l0oooo303oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02D0oooo003o0?ooo`X0
oooo0P00000U0?ooo`030000003oool0oooo02H0oooo003o0?ooo`T0oooo00<000000?ooo`3oool0
9@3oool00`000000oooo0?ooo`0V0?ooo`00o`3oool80?ooo`030000003oool0oooo02D0oooo00<0
00000?ooo`3oool09`3oool00?l0oooo1P3oool2000002L0oooo00<000000?ooo`3oool0:03oool0
0?l0oooo1@3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02P0oooo003o0?ooo`@0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0Y0?ooo`00o`3oool20?ooo`800000
:P3oool00`000000oooo0?ooo`0Y0?ooo`00o`3oool10?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool0:P3oool00?l0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0[
0?ooo`00o@3oool2000002`0oooo00<000000?ooo`3oool0:`3oool00?`0oooo00<000000?ooo`3o
ool0:`3oool00`000000oooo0?ooo`0/0?ooo`00n`3oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo02`0oooo003i0?ooo`800000;P3oool00`000000oooo0?ooo`0]0?ooo`00n03oool0
0`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02h0oooo003g0?ooo`030000003oool0oooo
02h0oooo00<000000?ooo`3oool0;P3oool00?D0oooo0P00000`0?ooo`030000003oool0oooo02l0
oooo003d0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0;`3oool00?<0oooo00<0
00000?ooo`3oool0<03oool00`000000oooo0?ooo`0`0?ooo`00l@3oool200000380oooo00<00000
0?ooo`3oool0<@3oool00?00oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0a0?oo
o`00k`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0380oooo003]0?ooo`800000
=@3oool00`000000oooo0?ooo`0b0?ooo`00k03oool00`000000oooo0?ooo`0d0?ooo`030000003o
ool0oooo03<0oooo003[0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0=03oool0
0>T0oooo0P00000g0?ooo`030000003oool0oooo03@0oooo003X0?ooo`030000003oool0oooo03H0
oooo00<000000?ooo`3oool0=@3oool00>L0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`0e0?ooo`00i@3oool2000003T0oooo00<000000?ooo`3oool0=P3oool00>@0oooo00<00000
0?ooo`3oool0>03oool00`000000oooo0?ooo`0g0?ooo`00h`3oool00`000000oooo0?ooo`0i0?oo
o`030000003oool0oooo03L0oooo003Q0?ooo`800000>`3oool00`000000oooo0?ooo`0h0?ooo`00
h03oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo03P0oooo003O0?ooo`030000003o
ool0oooo03/0oooo00<000000?ooo`3oool0>@3oool00=d0oooo0P00000m0?ooo`030000003oool0
oooo03X0oooo003L0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0>P3oool00=/0
oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0k0?ooo`00f@3oool200000400oooo
00<000000?ooo`3oool0>`3oool00=P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?oo
o`0l0?ooo`00e`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo03d0oooo003E0?oo
o`800000@P3oool00`000000oooo0?ooo`0m0?ooo`00e03oool00`000000oooo0?ooo`110?ooo`03
0000003oool0oooo03h0oooo003C0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0
?P3oool00=40oooo0P0000140?ooo`030000003oool0oooo03l0oooo003@0?ooo`030000003oool0
oooo04<0oooo00<000000?ooo`3oool0@03oool00<l0oooo00<000000?ooo`3oool0A03oool00`00
0000oooo0?ooo`100?ooo`00c@3oool2000004H0oooo00<000000?ooo`3oool0@@3oool00<`0oooo
00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`110?ooo`00b`3oool00`000000oooo0?oo
o`160?ooo`030000003oool0oooo0480oooo00390?ooo`800000B03oool00`000000oooo0?ooo`13
0?ooo`00b03oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo04<0oooo00370?ooo`03
0000003oool0oooo04P0oooo00<000000?ooo`3oool0A03oool00<D0oooo0P00001;0?ooo`030000
003oool0oooo04@0oooo00340?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0A@3o
ool00<<0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`160?ooo`00`@3oool20000
04d0oooo00<000000?ooo`3oool0AP3oool00<00oooo00<000000?ooo`3oool0C03oool00`000000
oooo0?ooo`170?ooo`00_`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo04L0oooo
002m0?ooo`800000C`3oool00`000000oooo0?ooo`180?ooo`00_03oool00`000000oooo0?ooo`1>
0?ooo`030000003oool0oooo04T0oooo002k0?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0B@3oool00;T0oooo0P00001A0?ooo`030000003oool0oooo04X0oooo002h0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool0BP3oool00;L0oooo00<000000?ooo`3oool0D@3o
ool00`000000oooo0?ooo`1;0?ooo`00]@3oool2000005<0oooo00<000000?ooo`3oool0C03oool0
0;@0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1<0?ooo`00/`3oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo04d0oooo002a0?ooo`800000EP3oool00`000000oooo
0?ooo`1=0?ooo`00/03oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo04h0oooo002_
0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0C`3oool00:d0oooo0P00001H0?oo
o`030000003oool0oooo04l0oooo002/0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3o
ool0D03oool00:/0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1@0?ooo`00Z@3o
ool2000005X0oooo00<000000?ooo`3oool0D@3oool00:P0oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`1B0?ooo`00Y`3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo
0580oooo002U0?ooo`800000G03oool00`000000oooo0?ooo`1C0?ooo`00Y03oool00`000000oooo
0?ooo`1L0?ooo`030000003oool0oooo05<0oooo002S0?ooo`030000003oool0oooo05`0oooo00<0
00000?ooo`3oool0E03oool00:40oooo0P00001N0?ooo`030000003oool0oooo05D0oooo002P0?oo
o`030000003oool0oooo05h0oooo00<000000?ooo`3oool0E@3oool009l0oooo00<000000?ooo`3o
ool0GP3oool00`000000oooo0?ooo`1F0?ooo`00W@3oool200000640oooo00<000000?ooo`3oool0
EP3oool009`0oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`1G0?ooo`00V`3oool0
0`000000oooo0?ooo`1P0?ooo`030000003oool0oooo05P0oooo002I0?ooo`800000H`3oool00`00
0000oooo0?ooo`1H0?ooo`00V03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo05T0
oooo002G0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0F@3oool009D0oooo0P00
001U0?ooo`030000003oool0oooo05X0oooo002D0?ooo`030000003oool0oooo06@0oooo00<00000
0?ooo`3oool0F`3oool009<0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`1K0?oo
o`00T@3oool2000006L0oooo00<000000?ooo`3oool0G03oool00900oooo00<000000?ooo`3oool0
I`3oool00`000000oooo0?ooo`1L0?ooo`00S`3oool00`000000oooo0?ooo`1W0?ooo`030000003o
ool0oooo05d0oooo002=0?ooo`800000J@3oool00`000000oooo0?ooo`1N0?ooo`00S03oool00`00
0000oooo0?ooo`1Y0?ooo`030000003oool0oooo05h0oooo002;0?ooo`030000003oool0oooo06T0
oooo00<000000?ooo`3oool0G`3oool008T0oooo0P00001/0?ooo`030000003oool0oooo05l0oooo
00280?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0H03oool008L0oooo00<00000
0?ooo`3oool0J`3oool00`000000oooo0?ooo`1Q0?ooo`00Q@3oool2000006h0oooo00<000000?oo
o`3oool0H@3oool008@0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`1R0?ooo`00
P`3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo0680oooo00210?ooo`800000L03o
ool00`000000oooo0?ooo`1S0?ooo`00P03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0
oooo06@0oooo001o0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0I03oool007d0
oooo0P00001b0?ooo`030000003oool0oooo06D0oooo001l0?ooo`030000003oool0oooo0780oooo
00<000000?ooo`3oool0I@3oool007/0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?oo
o`1V0?ooo`00N@3oool2000007@0oooo00<000000?ooo`3oool0I`3oool007P0oooo00<000000?oo
o`3oool0M03oool00`000000oooo0?ooo`1W0?ooo`00M`3oool00`000000oooo0?ooo`1d0?ooo`03
0000003oool0oooo06P0oooo001e0?ooo`800000M`3oool00`000000oooo0?ooo`1X0?ooo`00M03o
ool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo06T0oooo001c0?ooo`030000003oool0
oooo07H0oooo00<000000?ooo`3oool0JP3oool00740oooo0P00001i0?ooo`030000003oool0oooo
06X0oooo001`0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0J`3oool006l0oooo
00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1[0?ooo`00K@3oool2000007/0oooo00<0
00000?ooo`3oool0K03oool006`0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`1]
0?ooo`00J`3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo06d0oooo001Y0?ooo`80
0000O@3oool00`000000oooo0?ooo`1^0?ooo`00J03oool00`000000oooo0?ooo`1m0?ooo`030000
003oool0oooo06h0oooo001W0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0K`3o
ool006D0oooo0P00001o0?ooo`030000003oool0oooo0700oooo001T0?ooo`030000003oool0oooo
07l0oooo00<000000?ooo`3oool0L03oool006<0oooo00<000000?ooo`3oool0O`3oool00`000000
oooo0?ooo`1a0?ooo`00H@3oool200000880oooo00<000000?ooo`3oool0L@3oool00600oooo00<0
00000?ooo`3oool0P@3oool00`000000oooo0?ooo`1b0?ooo`00G`3oool00`000000oooo0?ooo`21
0?ooo`030000003oool0oooo07<0oooo001M0?ooo`800000Q03oool00`000000oooo0?ooo`1c0?oo
o`00G03oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo07@0oooo001K0?ooo`030000
003oool0oooo08@0oooo00<000000?ooo`3oool0M03oool005T0oooo0P0000260?ooo`030000003o
ool0oooo07D0oooo001H0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0MP3oool0
05L0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`1f0?ooo`00E@3oool2000008P0
oooo00<000000?ooo`3oool0M`3oool005@0oooo00<000000?ooo`3oool0R03oool00`000000oooo
0?ooo`1g0?ooo`00D`3oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo07P0oooo001A
0?ooo`800000RP3oool00`000000oooo0?ooo`1i0?ooo`00D03oool00`000000oooo0?ooo`2:0?oo
o`030000003oool0oooo07T0oooo001?0?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3o
ool0NP3oool004d0oooo0P00002=0?ooo`030000003oool0oooo07X0oooo001<0?ooo`030000003o
ool0oooo08`0oooo00<000000?ooo`3oool0N`3oool004/0oooo00<000000?ooo`3oool0S03oool0
0`000000oooo0?ooo`1l0?ooo`00B@3oool2000008l0oooo00<000000?ooo`3oool0O03oool004P0
oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`1m0?ooo`00A`3oool00`000000oooo
0?ooo`2?0?ooo`030000003oool0oooo07d0oooo00150?ooo`800000T@3oool00`000000oooo0?oo
o`1n0?ooo`00A03oool00`000000oooo0?ooo`2@0?ooo`030000003oool0oooo07l0oooo00130?oo
o`030000003oool0oooo0940oooo00<000000?ooo`3oool0O`3oool00440oooo0P00002C0?ooo`03
0000003oool0oooo0800oooo00100?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool0
P03oool003l0oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`210?ooo`00?@3oool2
000009D0oooo00<000000?ooo`3oool0PP3oool003`0oooo00<000000?ooo`3oool0U@3oool00`00
0000oooo0?ooo`220?ooo`00>`3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo08<0
oooo000i0?ooo`800000V03oool00`000000oooo0?ooo`230?ooo`00>03oool00`000000oooo0?oo
o`2G0?ooo`030000003oool0oooo08@0oooo000g0?ooo`030000003oool0oooo09L0oooo00<00000
0?ooo`3oool0Q@3oool003D0oooo0P00002J0?ooo`030000003oool0oooo08D0oooo000d0?ooo`03
0000003oool0oooo09T0oooo00<000000?ooo`3oool0QP3oool003<0oooo00<000000?ooo`3oool0
VP3oool00`000000oooo0?ooo`260?ooo`00<@3oool2000009`0oooo00<000000?ooo`3oool0Q`3o
ool00300oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`280?ooo`00;`3oool00`00
0000oooo0?ooo`2L0?ooo`030000003oool0oooo08P0oooo000]0?ooo`800000WP3oool00`000000
oooo0?ooo`290?ooo`00;03oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo08T0oooo
000[0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool0RP3oool002T0oooo0P00002P
0?ooo`030000003oool0oooo08/0oooo000X0?ooo`030000003oool0oooo0:00oooo00<000000?oo
o`3oool0R`3oool002L0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`2<0?ooo`00
9@3oool200000:<0oooo00<000000?ooo`3oool0S03oool002@0oooo00<000000?ooo`3oool0XP3o
ool00`000000oooo0?ooo`2=0?ooo`008`3oool00`000000oooo0?ooo`2R0?ooo`030000003oool0
oooo08h0oooo000Q0?ooo`800000Y@3oool00`000000oooo0?ooo`2>0?ooo`00803oool00`000000
oooo0?ooo`2T0?ooo`030000003oool0oooo08l0oooo000O0?ooo`030000003oool0oooo0:D0oooo
00<000000?ooo`3oool0S`3oool001d0oooo0P00002W0?ooo`030000003oool0oooo0900oooo000L
0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool0T@3oool001/0oooo00<000000?oo
o`3oool0Y`3oool00`000000oooo0?ooo`2A0?ooo`006@3oool200000:T0oooo00<000000?ooo`3o
ool0TP3oool001P0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`2B0?ooo`005`3o
ool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo09<0oooo000E0?ooo`800000Z`3oool0
0`000000oooo0?ooo`2D0?ooo`00503oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo
09@0oooo000C0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0U@3oool00140oooo
0P00002^0?ooo`030000003oool0oooo09D0oooo000@0?ooo`030000003oool0oooo0:d0oooo00<0
00000?ooo`3oool0UP3oool000l0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`2G
0?ooo`003@3oool200000;00oooo00<000000?ooo`3oool0U`3oool000`0oooo00<000000?ooo`3o
ool0[`3oool00`000000oooo0?ooo`2H0?ooo`002`3oool00`000000oooo0?ooo`2`0?ooo`030000
003oool0oooo09P0oooo00090?ooo`800000/P3oool00`000000oooo0?ooo`2I0?ooo`00203oool0
0`000000oooo0?ooo`2a0?ooo`030000003oool0oooo09X0oooo00080?ooo`030000003oool0oooo
0;40oooo00<000000?ooo`3oool0VP3oool000T0oooo00<000000?ooo`3oool0[`3oool00`000000
oooo0?ooo`2K0?ooo`002@3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo09/0oooo
000:0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool0W03oool000X0oooo00<00000
0?ooo`3oool0[03oool00`000000oooo0?ooo`2M0?ooo`002`3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo09d0oooo000;0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3o
ool0WP3oool000`0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`2N0?ooo`00303o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo09l0oooo000=0?ooo`030000003oool0
oooo0:H0oooo00<000000?ooo`3oool0X03oool000d0oooo00<000000?ooo`3oool0YP3oool00`00
0000oooo0?ooo`2P0?ooo`003P3oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo0:40
oooo000>0?ooo`030000003oool0oooo0:@0oooo00<000000?ooo`3oool0X@3oool000l0oooo00<0
00000?ooo`3oool0XP3oool00`000000oooo0?ooo`2R0?ooo`003`3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo0:<0oooo000@0?ooo`030000003oool0oooo0:00oooo00<000000?oo
o`3oool0X`3oool00100oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`2T0?ooo`00
4@3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo0:@0oooo000A0?ooo`030000003o
ool0oooo09d0oooo00<000000?ooo`3oool0Y@3oool00180oooo00<000000?ooo`3oool0V`3oool0
0`000000oooo0?ooo`2V0?ooo`004P3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo
0:H0oooo000C0?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3oool0Y`3oool001<0oooo
00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`2W0?ooo`00503oool00`000000oooo0?oo
o`2G0?ooo`030000003oool0oooo0:P0oooo000D0?ooo`030000003oool0oooo09H0oooo00<00000
0?ooo`3oool0Z@3oool001D0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`2Y0?oo
o`005@3oool00`000000oooo0?ooo`2D0?ooo`030000003oool0oooo0:X0oooo000F0?ooo`030000
003oool0oooo09<0oooo00<000000?ooo`3oool0ZP3oool001H0oooo00<000000?ooo`3oool0TP3o
ool00`000000oooo0?ooo`2[0?ooo`005`3oool00`000000oooo0?ooo`2@0?ooo`030000003oool0
oooo0:`0oooo000G0?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0[03oool001P0
oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`2]0?ooo`00603oool00`000000oooo
0?ooo`2>0?ooo`030000003oool0oooo0:d0oooo000I0?ooo`030000003oool0oooo08`0oooo00<0
00000?ooo`3oool0[P3oool001T0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`2_
0?ooo`006P3oool00`000000oooo0?ooo`2:0?ooo`030000003oool0oooo0:l0oooo000J0?ooo`03
0000003oool0oooo08T0oooo00<000000?ooo`3oool0/03oool001/0oooo00<000000?ooo`3oool0
R03oool00`000000oooo0?ooo`2`0?ooo`006`3oool00`000000oooo0?ooo`270?ooo`030000003o
ool0oooo0;40oooo000L0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0/P3oool0
01`0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`2b0?ooo`007@3oool00`000000
oooo0?ooo`230?ooo`030000003oool0oooo0;<0oooo000M0?ooo`030000003oool0oooo08<0oooo
00<000000?ooo`3oool0/`3oool001h0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?oo
o`2d0?ooo`007P3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo0;D0oooo000O0?oo
o`030000003oool0oooo07l0oooo00<000000?ooo`3oool0]@3oool001l0oooo00<000000?ooo`3o
ool0OP3oool00`000000oooo0?ooo`2f0?ooo`00803oool00`000000oooo0?ooo`1m0?ooo`030000
003oool0oooo0;H0oooo000P0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0]`3o
ool00240oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`2h0?ooo`008@3oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo0;P0oooo000R0?ooo`030000003oool0oooo07P0
oooo00<000000?ooo`3oool0^@3oool00280oooo00<000000?ooo`3oool0N03oool00`000000oooo
0?ooo`2i0?ooo`008`3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo0;X0oooo000S
0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0^`3oool002@0oooo00<000000?oo
o`3oool0M03oool00`000000oooo0?ooo`2k0?ooo`00903oool00`000000oooo0?ooo`1c0?ooo`03
0000003oool0oooo0;`0oooo000U0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0
_03oool002D0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`2m0?ooo`009P3oool0
0`000000oooo0?ooo`1_0?ooo`030000003oool0oooo0;h0oooo000V0?ooo`030000003oool0oooo
06l0oooo00<000000?ooo`3oool0_P3oool002L0oooo00<000000?ooo`3oool0K@3oool00`000000
oooo0?ooo`2o0?ooo`009`3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo0;l0oooo
000X0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0`03oool002P0oooo00<00000
0?ooo`3oool0JP3oool00`000000oooo0?ooo`310?ooo`00:@3oool00`000000oooo0?ooo`1Y0?oo
o`030000003oool0oooo0<40oooo000Y0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3o
ool0`P3oool002X0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`320?ooo`00:P3o
ool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo0<<0oooo000[0?ooo`030000003oool0
oooo06@0oooo00<000000?ooo`3oool0a03oool002/0oooo00<000000?ooo`3oool0I03oool00`00
0000oooo0?ooo`340?ooo`00;03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0<D0
oooo000/0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0a@3oool002d0oooo00<0
00000?ooo`3oool0H03oool00`000000oooo0?ooo`360?ooo`00;@3oool00`000000oooo0?ooo`1O
0?ooo`030000003oool0oooo0<L0oooo000^0?ooo`030000003oool0oooo05h0oooo00<000000?oo
o`3oool0a`3oool002h0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`380?ooo`00
;`3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0<P0oooo000_0?ooo`030000003o
ool0oooo05/0oooo00<000000?ooo`3oool0b@3oool00300oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`3:0?ooo`00<03oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo
0<X0oooo000a0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0b`3oool00340oooo
00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`3;0?ooo`00<P3oool00`000000oooo0?oo
o`1E0?ooo`030000003oool0oooo0<`0oooo000b0?ooo`030000003oool0oooo05@0oooo00<00000
0?ooo`3oool0c@3oool003<0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`3=0?oo
o`00<`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo0<h0oooo000d0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool0cP3oool003@0oooo00<000000?ooo`3oool0D03o
ool00`000000oooo0?ooo`3?0?ooo`00=@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0
oooo0=00oooo000e0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0d03oool003H0
oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`3A0?ooo`00=P3oool00`000000oooo
0?ooo`1<0?ooo`030000003oool0oooo0=40oooo000g0?ooo`030000003oool0oooo04X0oooo00<0
00000?ooo`3oool0dP3oool003L0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`3C
0?ooo`00>03oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0=<0oooo000h0?ooo`03
0000003oool0oooo04L0oooo00<000000?ooo`3oool0e03oool003T0oooo00<000000?ooo`3oool0
AP3oool00`000000oooo0?ooo`3D0?ooo`00>@3oool00`000000oooo0?ooo`150?ooo`030000003o
ool0oooo0=D0oooo000j0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0eP3oool0
03X0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`3F0?ooo`00>`3oool00`000000
oooo0?ooo`110?ooo`030000003oool0oooo0=L0oooo000k0?ooo`030000003oool0oooo0440oooo
00<000000?ooo`3oool0e`3oool003`0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?oo
o`3H0?ooo`00?03oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo0=T0oooo000m0?oo
o`030000003oool0oooo03d0oooo00<000000?ooo`3oool0f@3oool003d0oooo00<000000?ooo`3o
ool0?03oool00`000000oooo0?ooo`3J0?ooo`00?P3oool00`000000oooo0?ooo`0k0?ooo`030000
003oool0oooo0=X0oooo000n0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0f`3o
ool003l0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`3L0?ooo`00?`3oool00`00
0000oooo0?ooo`0h0?ooo`030000003oool0oooo0=`0oooo00100?ooo`030000003oool0oooo03H0
oooo00<000000?ooo`3oool0g@3oool00400oooo00<000000?ooo`3oool0=P3oool00`000000oooo
0?ooo`3M0?ooo`00@@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0=h0oooo0011
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0g`3oool00480oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`3O0?ooo`00@P3oool00`000000oooo0?ooo`0a0?ooo`03
0000003oool0oooo0>00oooo00130?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0
h03oool004<0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`3Q0?ooo`00A03oool0
0`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0>80oooo00140?ooo`030000003oool0oooo
02d0oooo00<000000?ooo`3oool0hP3oool004D0oooo00<000000?ooo`3oool0:`3oool00`000000
oooo0?ooo`3S0?ooo`00A@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0><0oooo
00160?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0i03oool004H0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`3U0?ooo`00A`3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo0>D0oooo00170?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3o
ool0iP3oool004P0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`3V0?ooo`00B03o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0>L0oooo00190?ooo`030000003oool0
oooo0280oooo00<000000?ooo`3oool0j03oool004T0oooo00<000000?ooo`3oool08P3oool00`00
0000oooo0?ooo`3X0?ooo`00BP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0>T0
oooo001:0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0j@3oool004/0oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`3Z0?ooo`00B`3oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo0>/0oooo001<0?ooo`030000003oool0oooo01`0oooo00<000000?oo
o`3oool0j`3oool004`0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`3/0?ooo`00
C@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0>`0oooo001=0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0k@3oool004h0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`3^0?ooo`00CP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
0>h0oooo001?0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0k`3oool004l0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3_0?ooo`00D03oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0?00oooo001@0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0l@3oool00540oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`3a0?oo
o`00D@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0?80oooo001B0?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool0lP3oool00580oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`3c0?ooo`00D`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo0?@0oooo001C0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0m03oool005@0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`3e0?ooo`00E03oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo0?D0oooo001E0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0mP3oool005D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`3g
0?ooo`00EP3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0?L0oooo001F0?ooo`<0
00001@3oool00`000000oooo0?ooo`3h0?ooo`00E@3oool4000000D0oooo00<000000?ooo`3oool0
n03oool005H0oooo100000030?ooo`030000003oool0oooo0?T0oooo001G0?ooo`<000000P3oool0
0`000000oooo0?ooo`3j0?ooo`00E`3oool300000080oooo00<000000?ooo`3oool0nP3oool005P0
oooo0P0000000`3oool000000?ooo`3l0?ooo`00F@3oool300000?d0oooo001J0?ooo`030000003o
ool0oooo0?`0oooo003o0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo
003o0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo
003o0?oooeX0oooo0000\
\>"],
  ImageRangeCache->{{{0, 344.438}, {430.813, 0}} -> {-0.100021, -3.12665, \
0.0182909, 0.0182909}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawVectorDifference[{\(-10\), 2}, {3, 5}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .38462 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.75641 0.0732601 0.00915751 0.0732601 [
[ 0 0 0 0 ]
[ 1 .38462 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .38462 L
0 .38462 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.75641 .00916 m
.02381 .15568 L
s
% Start of user PostScript
0. 0. -10. 2. MAarrowhead1
% End of user PostScript
.75641 .00916 m
.97619 .37546 L
s
% Start of user PostScript
0. 0. 3. 5. MAarrowhead1
% End of user PostScript
1 Mabswid
.02381 .15568 m
.97619 .37546 L
s
% Start of user PostScript
-10. 2. 3. 5. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 110.75},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000YR000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooom40?ooo`<00000JP3o
ool00?l0oooo?`3oool5000000<0oooo00<000000?ooo`3oool0I`3oool00?l0oooo>P3oool50000
00P0oooo00<000000?ooo`3oool0I`3oool00?l0oooo=@3oool5000000h0oooo00<000000?ooo`3o
ool0IP3oool00?l0oooo<03oool5000001<0oooo00<000000?ooo`3oool0IP3oool00?l0oooo:`3o
ool5000001T0oooo00<000000?ooo`3oool0I@3oool00?l0oooo9P3oool5000001l0oooo00<00000
0?ooo`3oool0I03oool00?l0oooo8@3oool5000002@0oooo00<000000?ooo`3oool0I03oool00?l0
oooo703oool5000002X0oooo00<000000?ooo`3oool0H`3oool00?l0oooo5`3oool5000002l0oooo
00<000000?ooo`3oool0H`3oool00?l0oooo4P3oool5000003D0oooo00<000000?ooo`3oool0HP3o
ool00?l0oooo3@3oool5000003/0oooo00<000000?ooo`3oool0H@3oool00?l0oooo203oool50000
0400oooo00<000000?ooo`3oool0H@3oool00?l0oooo0`3oool5000004H0oooo00<000000?ooo`3o
ool0H03oool00?d0oooo1@00001;0?ooo`030000003oool0oooo0600oooo003h0?ooo`D00000D@3o
ool00`000000oooo0?ooo`1O0?ooo`00l`3oool5000005L0oooo00<000000?ooo`3oool0GP3oool0
0>h0oooo1@00001L0?ooo`030000003oool0oooo05h0oooo003Y0?ooo`D00000HP3oool00`000000
oooo0?ooo`1M0?ooo`00i03oool5000006L0oooo00<000000?ooo`3oool0G@3oool00=l0oooo1@00
001]0?ooo`030000003oool0oooo05`0oooo003J0?ooo`D00000LP3oool00`000000oooo0?ooo`1L
0?ooo`00e@3oool5000007P0oooo00<000000?ooo`3oool0F`3oool00=00oooo1@00001n0?ooo`03
0000003oool0oooo05X0oooo003;0?ooo`D00000P`3oool00`000000oooo0?ooo`1J0?ooo`00aP3o
ool5000008T0oooo00<000000?ooo`3oool0F@3oool00<40oooo1@00002>0?ooo`030000003oool0
oooo05T0oooo002l0?ooo`D00000U03oool00`000000oooo0?ooo`1H0?ooo`00]`3oool5000009X0
oooo00<000000?ooo`3oool0E`3oool00;80oooo1@00002O0?ooo`030000003oool0oooo05L0oooo
002]0?ooo`D00000Y@3oool00`000000oooo0?ooo`1F0?ooo`00Z@3oool400000:X0oooo00<00000
0?ooo`3oool0EP3oool00:@0oooo1@00002_0?ooo`030000003oool0oooo05D0oooo002O0?ooo`D0
0000]@3oool00`000000oooo0?ooo`1D0?ooo`00VP3oool500000;X0oooo00<000000?ooo`3oool0
E03oool009D0oooo1@0000300?ooo`030000003oool0oooo05<0oooo002@0?ooo`D00000a@3oool0
0`000000oooo0?ooo`1C0?ooo`00R`3oool500000</0oooo00<000000?ooo`3oool0DP3oool008H0
oooo1@00003A0?ooo`030000003oool0oooo0540oooo00210?ooo`D00000eP3oool00`000000oooo
0?ooo`1A0?ooo`00O03oool500000=`0oooo00<000000?ooo`3oool0D03oool007L0oooo1@00003Q
0?ooo`030000003oool0oooo0500oooo001b0?ooo`D00000i`3oool00`000000oooo0?ooo`1?0?oo
o`00K@3oool500000>d0oooo00<000000?ooo`3oool0CP3oool006P0oooo1@00003b0?ooo`030000
003oool0oooo04h0oooo001S0?ooo`D00000n03oool00`000000oooo0?ooo`1=0?ooo`00GP3oool5
00000?d0oooo00<000000?ooo`3oool0C@3oool005T0oooo1@00003o0?ooo`@0oooo00<000000?oo
o`3oool0C03oool005@0oooo1@00003o0?ooo`X0oooo00<000000?ooo`3oool0B`3oool004l0oooo
1@00003o0?ooo`l0oooo00<000000?ooo`3oool0B`3oool004X0oooo1@00003o0?oooaD0oooo00<0
00000?ooo`3oool0BP3oool004D0oooo1@00003o0?oooaX0oooo00<000000?ooo`3oool0BP3oool0
0400oooo1@00003o0?ooob00oooo00<000000?ooo`3oool0B@3oool003/0oooo1@00003o0?ooobH0
oooo00<000000?ooo`3oool0B03oool003H0oooo1@00003o0?ooob/0oooo00<000000?ooo`3oool0
B03oool00340oooo1@00003o0?oooc40oooo00<000000?ooo`3oool0A`3oool002`0oooo1@00003o
0?ooocH0oooo00<000000?ooo`3oool0A`3oool002L0oooo1@00003o0?oooc`0oooo00<000000?oo
o`3oool0AP3oool00280oooo1@00003o0?oood80oooo00<000000?ooo`3oool0A@3oool001d0oooo
1@00003o0?ooodL0oooo00<000000?ooo`3oool0A@3oool00140oooo0P0000050?ooo`D00000o`3o
oom=0?ooo`030000003oool0oooo04@0oooo000?0?ooo`T00000o`3ooomB0?ooo`030000003oool0
oooo04@0oooo000=0?ooo`H00000o`3ooomH0?ooo`030000003oool0oooo04<0oooo000;0?ooo`P0
0000o`3ooomH0?ooo`030000003oool0oooo04<0oooo000>0?ooo`D00000o`3ooomI0?ooo`030000
003oool0oooo0480oooo000B0?ooo`@00000o`3ooomG0?ooo`030000003oool0oooo0440oooo000F
0?ooo`D00000o`3ooomB0?ooo`030000003oool0oooo0440oooo000K0?ooo`@00000o`3ooom?0?oo
o`030000003oool0oooo0400oooo000O0?ooo`@00000o`3ooom;0?ooo`030000003oool0oooo0400
oooo000S0?ooo`@00000o`3ooom80?ooo`030000003oool0oooo03l0oooo000W0?ooo`D00000o`3o
oom40?ooo`030000003oool0oooo03h0oooo000/0?ooo`@00000o`3ooom00?ooo`030000003oool0
oooo03h0oooo000`0?ooo`@00000o`3ooolm0?ooo`030000003oool0oooo03d0oooo000d0?ooo`D0
0000o`3ooolh0?ooo`030000003oool0oooo03d0oooo000i0?ooo`@00000o`3ooole0?ooo`030000
003oool0oooo03`0oooo000m0?ooo`@00000o`3ooolb0?ooo`030000003oool0oooo03/0oooo0011
0?ooo`@00000o`3oool^0?ooo`030000003oool0oooo03/0oooo00150?ooo`D00000o`3ooolZ0?oo
o`030000003oool0oooo03X0oooo001:0?ooo`@00000o`3ooolV0?ooo`030000003oool0oooo03X0
oooo001>0?ooo`@00000o`3ooolS0?ooo`030000003oool0oooo03T0oooo001B0?ooo`D00000o`3o
oolO0?ooo`030000003oool0oooo03P0oooo001G0?ooo`@00000o`3ooolK0?ooo`030000003oool0
oooo03P0oooo001K0?ooo`@00000o`3ooolH0?ooo`030000003oool0oooo03L0oooo001O0?ooo`D0
0000o`3ooolC0?ooo`030000003oool0oooo03L0oooo001T0?ooo`@00000o`3oool@0?ooo`030000
003oool0oooo03H0oooo001X0?ooo`@00000o`3oool=0?ooo`030000003oool0oooo03D0oooo001/
0?ooo`@00000o`3oool90?ooo`030000003oool0oooo03D0oooo001`0?ooo`D00000o`3oool50?oo
o`030000003oool0oooo03@0oooo001e0?ooo`@00000o`3oool10?ooo`030000003oool0oooo03@0
oooo001i0?ooo`@00000o@3oool00`000000oooo0?ooo`0c0?ooo`00O@3oool500000?T0oooo00<0
00000?ooo`3oool0<P3oool00880oooo1000003e0?ooo`030000003oool0oooo0380oooo00260?oo
o`@00000lP3oool00`000000oooo0?ooo`0a0?ooo`00RP3oool400000>h0oooo00<000000?ooo`3o
ool0<@3oool008h0oooo1@00003Z0?ooo`030000003oool0oooo0300oooo002C0?ooo`@00000i`3o
ool00`000000oooo0?ooo`0_0?ooo`00U`3oool400000><0oooo00<000000?ooo`3oool0;`3oool0
09/0oooo1@00003O0?ooo`030000003oool0oooo02h0oooo002P0?ooo`@00000f`3oool00`000000
oooo0?ooo`0^0?ooo`00Y03oool400000=P0oooo00<000000?ooo`3oool0;@3oool00:P0oooo1@00
003C0?ooo`030000003oool0oooo02d0oooo002]0?ooo`@00000d03oool00`000000oooo0?ooo`0/
0?ooo`00/@3oool400000<d0oooo00<000000?ooo`3oool0:`3oool00;D0oooo100000390?ooo`03
0000003oool0oooo02/0oooo002i0?ooo`D00000a@3oool00`000000oooo0?ooo`0Z0?ooo`00_P3o
ool400000<40oooo00<000000?ooo`3oool0:P3oool00<80oooo1000002n0?ooo`030000003oool0
oooo02T0oooo00360?ooo`D00000^P3oool00`000000oooo0?ooo`0X0?ooo`00b`3oool400000;H0
oooo00<000000?ooo`3oool0:03oool00<l0oooo1000002c0?ooo`030000003oool0oooo02L0oooo
003C0?ooo`@00000[`3oool00`000000oooo0?ooo`0W0?ooo`00e`3oool500000:/0oooo00<00000
0?ooo`3oool09P3oool00=`0oooo1000002X0?ooo`030000003oool0oooo02D0oooo003P0?ooo`@0
0000Y03oool00`000000oooo0?ooo`0U0?ooo`00i03oool500000:00oooo00<000000?ooo`3oool0
903oool00>T0oooo1000002L0?ooo`030000003oool0oooo02@0oooo003]0?ooo`@00000V@3oool0
0`000000oooo0?ooo`0S0?ooo`00l@3oool5000009D0oooo00<000000?ooo`3oool08P3oool00?H0
oooo1000002A0?ooo`030000003oool0oooo0280oooo003j0?ooo`@00000SP3oool00`000000oooo
0?ooo`0Q0?ooo`00oP3oool4000008X0oooo00<000000?ooo`3oool08@3oool00?l0oooo0`3oool5
000008H0oooo00<000000?ooo`3oool0803oool00?l0oooo203oool4000008<0oooo00<000000?oo
o`3oool07`3oool00?l0oooo303oool4000007l0oooo00<000000?ooo`3oool07`3oool00?l0oooo
403oool5000007/0oooo00<000000?ooo`3oool07P3oool00?l0oooo5@3oool4000007L0oooo00<0
00000?ooo`3oool07P3oool00?l0oooo6@3oool4000007@0oooo00<000000?ooo`3oool07@3oool0
0?l0oooo7@3oool500000700oooo00<000000?ooo`3oool0703oool00?l0oooo8P3oool4000006`0
oooo00<000000?ooo`3oool0703oool00?l0oooo9P3oool4000006T0oooo00<000000?ooo`3oool0
6`3oool00?l0oooo:P3oool4000006D0oooo00<000000?ooo`3oool06`3oool00?l0oooo;P3oool5
00000640oooo00<000000?ooo`3oool06P3oool00?l0oooo<`3oool4000005h0oooo00<000000?oo
o`3oool06@3oool00?l0oooo=`3oool4000005X0oooo00<000000?ooo`3oool06@3oool00?l0oooo
>`3oool5000005H0oooo00<000000?ooo`3oool0603oool00?l0oooo@03oool400000580oooo00<0
00000?ooo`3oool0603oool00?l0ooooA03oool4000004l0oooo00<000000?ooo`3oool05`3oool0
0?l0ooooB03oool4000004`0oooo00<000000?ooo`3oool05P3oool00?l0ooooC03oool5000004L0
oooo00<000000?ooo`3oool05P3oool00?l0ooooD@3oool4000004@0oooo00<000000?ooo`3oool0
5@3oool00?l0ooooE@3oool400000400oooo00<000000?ooo`3oool05@3oool00?l0ooooF@3oool5
000003`0oooo00<000000?ooo`3oool0503oool00?l0ooooGP3oool4000003P0oooo00<000000?oo
o`3oool0503oool00?l0ooooHP3oool4000003D0oooo00<000000?ooo`3oool04`3oool00?l0oooo
IP3oool500000340oooo00<000000?ooo`3oool04P3oool00?l0ooooJ`3oool4000002d0oooo00<0
00000?ooo`3oool04P3oool00?l0ooooK`3oool4000002X0oooo00<000000?ooo`3oool04@3oool0
0?l0ooooL`3oool4000002H0oooo00<000000?ooo`3oool04@3oool00?l0ooooM`3oool500000280
oooo00<000000?ooo`3oool0403oool00?l0ooooO03oool4000001l0oooo00<000000?ooo`3oool0
3`3oool00?l0ooooP03oool4000001/0oooo00@000000?ooo`3oool000003P3oool00?l0ooooQ03o
ool5000001L0oooo0`00000>0?ooo`00o`3ooon90?ooo`@000004P3oool4000000h0oooo003o0?oo
ohd0oooo1000000?0?ooo`@000003@3oool00?l0ooooT@3oool4000000T0oooo00<000000?ooo`3o
ool00`00000=0?ooo`00o`3ooonE0?ooo`D00000103oool6000000d0oooo003o0?oooiX0oooo2P00
000=0?ooo`00o`3ooonM0?ooo`P00000303oool00?l0ooooW@3oool8000000`0oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {165.125, 0}} -> {-10.3517, -0.125, 0.0476917, \
0.0476917}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawCollision[{2, 0}, {\(-1\), 0}, {1, 
          1}, {\(-2\), \(-1\)}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .66667 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.65873 0.31746 0.333333 0.31746 [
[ 0 0 0 0 ]
[ 1 .66667 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .66667 L
0 .66667 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .33333 m
.65873 .33333 L
s
% Start of user PostScript
-2. 0. 0. 0. MAarrowhead1
% End of user PostScript
.97619 .33333 m
.65873 .33333 L
s
% Start of user PostScript
1. 0. 0. 0. MAarrowhead1
% End of user PostScript
.65873 .33333 m
.97619 .65079 L
s
% Start of user PostScript
0. 0. 1. 1. MAarrowhead1
% End of user PostScript
.65873 .33333 m
.02381 .01587 L
s
% Start of user PostScript
0. 0. -2. -1. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 192},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`000182000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`002P3oool200000?l0ooooY@3oool000/0oooo1@00003o0?oooj40oooo
000<0?ooo`L00000o`3ooonN0?ooo`003@3oool500000?l0ooooW`3oool000l0oooo1000003o0?oo
oih0oooo000@0?ooo`030000003oool0oooo00800000o`3ooonL0?ooo`005@3oool200000?l0oooo
VP3oool001L0oooo0P00003o0?oooiP0oooo000I0?ooo`800000o`3ooonF0?ooo`006`3oool20000
0?l0ooooU03oool001d0oooo0P00003o0?oooi80oooo000O0?ooo`800000o`3ooon@0?ooo`008@3o
ool200000?l0ooooSP3oool002<0oooo0P00003o0?oooh`0oooo000U0?ooo`800000o`3ooon:0?oo
o`009`3oool200000?l0ooooR03oool002T0oooo0P00003o0?ooohH0oooo000[0?ooo`800000o`3o
oon40?ooo`00;@3oool200000?l0ooooPP3oool002l0oooo0P00003o0?oooh00oooo000a0?ooo`80
0000o`3ooomn0?ooo`00<`3oool200000?l0ooooO03oool003D0oooo0P00003o0?ooogX0oooo000g
0?ooo`800000o`3ooomh0?ooo`00>@3oool200000?l0ooooMP3oool003/0oooo0P00003o0?ooog@0
oooo000m0?ooo`800000o`3ooomb0?ooo`00?`3oool200000?l0ooooL03oool00440oooo0P00003o
0?ooofh0oooo00130?ooo`800000o`3ooom/0?ooo`00A@3oool200000?l0ooooJP3oool004L0oooo
0P00003o0?ooofP0oooo00190?ooo`800000o`3ooomV0?ooo`00B`3oool200000?l0ooooI03oool0
04d0oooo0P00003o0?ooof80oooo001?0?ooo`800000o`3ooomP0?ooo`00D@3oool200000?l0oooo
GP3oool005<0oooo0P00003o0?oooe`0oooo001E0?ooo`800000o`3ooomJ0?ooo`00E`3oool20000
0?l0ooooF03oool005T0oooo0P00003o0?oooeH0oooo001K0?ooo`800000o`3ooomD0?ooo`00G@3o
ool200000?l0ooooDP3oool005l0oooo0P00003o0?oooe00oooo001Q0?ooo`800000o`3ooom>0?oo
o`00H`3oool200000?l0ooooC03oool006D0oooo0P00003o0?ooodX0oooo001W0?ooo`800000o`3o
oom80?ooo`00J@3oool200000?l0ooooAP3oool006/0oooo0P00003o0?oood@0oooo001]0?ooo`80
0000o`3ooom20?ooo`00K`3oool200000?l0oooo@03oool00740oooo0P00003o0?oooch0oooo001c
0?ooo`800000o`3oooll0?ooo`00M@3oool200000?l0oooo>P3oool007L0oooo0P00003o0?ooocP0
oooo001i0?ooo`800000o`3ooolf0?ooo`00N`3oool200000?l0oooo=03oool007d0oooo0P00003o
0?oooc80oooo001o0?ooo`800000o`3oool`0?ooo`00P@3oool200000?l0oooo;P3oool008<0oooo
0P00003o0?ooob`0oooo00250?ooo`800000o`3ooolZ0?ooo`00Q`3oool200000?l0oooo:03oool0
08T0oooo0P00003o0?ooobH0oooo002;0?ooo`800000o`3ooolT0?ooo`00S@3oool200000?l0oooo
8P3oool008l0oooo0P00003o0?ooob00oooo002A0?ooo`800000o`3ooolN0?ooo`00T`3oool20000
0?l0oooo703oool009D0oooo0P00003o0?oooaX0oooo002G0?ooo`800000o`3ooolH0?ooo`00V@3o
ool200000?l0oooo5P3oool009/0oooo0P00003o0?oooa@0oooo002M0?ooo`800000o`3ooolB0?oo
o`00W`3oool200000?l0oooo403oool00:40oooo0P00003o0?ooo`h0oooo002S0?ooo`800000o`3o
ool<0?ooo`00Y@3oool200000?l0oooo2P3oool00:L0oooo0P00003o0?ooo`P0oooo002Y0?ooo`80
0000o`3oool60?ooo`00Z`3oool200000?l0oooo103oool00:d0oooo0P00003o0?ooo`80oooo002_
0?ooo`800000o`3oool00;40oooo0P00003m0?ooo`00/`3oool200000?/0oooo002e0?ooo`800000
n@3oool00;L0oooo0P00003g0?ooo`00^@3oool200000?D0oooo002k0?ooo`800000l`3oool00;d0
oooo0P00003a0?ooo`00_`3oool200000>l0oooo00310?ooo`800000k@3oool00<<0oooo0P00003[
0?ooo`00a@3oool200000>T0oooo00370?ooo`800000i`3oool00<T0oooo0P00003U0?ooo`00b`3o
ool200000><0oooo003=0?ooo`800000h@3oool00<l0oooo0P00003O0?ooo`00d@3oool200000=d0
oooo003C0?ooo`800000f`3oool00=D0oooo0P00003I0?ooo`00e`3oool200000=L0oooo003I0?oo
o`800000e@3oool00=/0oooo0P00003C0?ooo`00g@3oool200000=40oooo003O0?ooo`800000c`3o
ool00>40oooo0P00003=0?ooo`00h`3oool200000</0oooo003U0?ooo`800000b@3oool00>L0oooo
0P0000370?ooo`00j@3oool200000<D0oooo003[0?ooo`800000``3oool00>d0oooo0P0000310?oo
o`00k`3oool200000;l0oooo003a0?ooo`800000_@3oool00?<0oooo0P00002k0?ooo`00m@3oool2
00000;T0oooo003g0?ooo`800000]`3oool00?T0oooo0P00002e0?ooo`00n`3oool200000;<0oooo
003m0?ooo`800000/@3oool00?l0oooo0P00002_0?ooo`00o`3oool20?ooo`800000[@3oool00?l0
oooo103oool200000:/0oooo003o0?ooo`H0oooo0P00002Y0?ooo`00o`3oool80?ooo`800000Y`3o
ool00?l0oooo2P3oool200000:D0oooo003o0?ooo``0oooo0P00002S0?ooo`00o`3oool>0?ooo`80
0000X@3oool00?l0oooo403oool2000009l0oooo003o0?oooa80oooo0P00002M0?ooo`00o`3ooolD
0?ooo`800000V`3oool00?l0oooo5P3oool2000009T0oooo003o0?oooaD0oooo0P0000000`3oool0
0000000000090?ooo`800000S03oool00?l0oooo5@3oool7000000<0oooo1P00002<0?ooo`002P3o
oooo000009`000002`3oool00?l0oooo5@3oool600000080oooo2000002<0?ooo`00o`3ooolE0?oo
o`8000001`3oool010000000oooo0?ooo`3oool3000008`0oooo003o0?oooal0oooo00<000000?oo
o`3oool0S`3oool00?l0oooo803oool00`000000oooo0?ooo`2>0?ooo`00o`3ooolQ0?ooo`030000
003oool0oooo08d0oooo003o0?ooob80oooo00<000000?ooo`3oool0S03oool00?l0oooo8`3oool0
0`000000oooo0?ooo`2;0?ooo`00o`3ooolT0?ooo`030000003oool0oooo08X0oooo003o0?ooobD0
oooo00<000000?ooo`3oool0R@3oool00?l0oooo9P3oool00`000000oooo0?ooo`280?ooo`00o`3o
oolW0?ooo`030000003oool0oooo08L0oooo003o0?ooobP0oooo00<000000?ooo`3oool0QP3oool0
0?l0oooo:@3oool00`000000oooo0?ooo`250?ooo`00o`3ooolZ0?ooo`030000003oool0oooo08@0
oooo003o0?ooob/0oooo00<000000?ooo`3oool0P`3oool00?l0oooo;03oool00`000000oooo0?oo
o`220?ooo`00o`3oool]0?ooo`030000003oool0oooo0840oooo003o0?ooobh0oooo00<000000?oo
o`3oool0P03oool00?l0oooo;`3oool00`000000oooo0?ooo`1o0?ooo`00o`3oool`0?ooo`030000
003oool0oooo07h0oooo003o0?oooc40oooo00<000000?ooo`3oool0O@3oool00?l0oooo<P3oool0
0`000000oooo0?ooo`1l0?ooo`00o`3ooolc0?ooo`030000003oool0oooo07/0oooo003o0?oooc@0
oooo00<000000?ooo`3oool0NP3oool00?l0oooo=@3oool00`000000oooo0?ooo`1i0?ooo`00o`3o
oolf0?ooo`030000003oool0oooo07P0oooo003o0?ooocL0oooo00<000000?ooo`3oool0M`3oool0
0?l0oooo>03oool00`000000oooo0?ooo`1f0?ooo`00o`3oooli0?ooo`030000003oool0oooo07D0
oooo003o0?ooocX0oooo00<000000?ooo`3oool0M03oool00?l0oooo>`3oool00`000000oooo0?oo
o`1c0?ooo`00o`3oooll0?ooo`030000003oool0oooo0780oooo003o0?ooocd0oooo00<000000?oo
o`3oool0L@3oool00?l0oooo?P3oool00`000000oooo0?ooo`1`0?ooo`00o`3ooolo0?ooo`030000
003oool0oooo06l0oooo003o0?oood00oooo00<000000?ooo`3oool0KP3oool00?l0oooo@@3oool0
0`000000oooo0?ooo`1]0?ooo`00o`3ooom20?ooo`030000003oool0oooo06`0oooo003o0?oood<0
oooo00<000000?ooo`3oool0J`3oool00?l0ooooA03oool00`000000oooo0?ooo`1Z0?ooo`00o`3o
oom50?ooo`030000003oool0oooo06T0oooo003o0?ooodH0oooo00<000000?ooo`3oool0J03oool0
0?l0ooooA`3oool00`000000oooo0?ooo`1W0?ooo`00o`3ooom80?ooo`030000003oool0oooo06H0
oooo003o0?ooodT0oooo00<000000?ooo`3oool0I@3oool00?l0ooooBP3oool00`000000oooo0?oo
o`1T0?ooo`00o`3ooom;0?ooo`030000003oool0oooo06<0oooo003o0?oood`0oooo00<000000?oo
o`3oool0HP3oool00?l0ooooC@3oool00`000000oooo0?ooo`1Q0?ooo`00o`3ooom>0?ooo`030000
003oool0oooo0600oooo003o0?ooodl0oooo00<000000?ooo`3oool0G`3oool00?l0ooooD03oool0
0`000000oooo0?ooo`1N0?ooo`00o`3ooomA0?ooo`030000003oool0oooo05d0oooo003o0?oooe80
oooo00<000000?ooo`3oool0G03oool00?l0ooooD`3oool00`000000oooo0?ooo`1K0?ooo`00o`3o
oomD0?ooo`030000003oool0oooo05X0oooo003o0?oooeD0oooo00<000000?ooo`3oool0F@3oool0
0?l0ooooEP3oool00`000000oooo0?ooo`1H0?ooo`00o`3ooomG0?ooo`030000003oool0oooo05L0
oooo003o0?oooeP0oooo00<000000?ooo`3oool0EP3oool00?l0ooooF@3oool00`000000oooo0?oo
o`1E0?ooo`00o`3ooomJ0?ooo`030000003oool0oooo05@0oooo003o0?oooe/0oooo00<000000?oo
o`3oool0D`3oool00?l0ooooG03oool00`000000oooo0?ooo`1B0?ooo`00o`3ooomM0?ooo`030000
003oool0oooo0540oooo003o0?oooeh0oooo00<000000?ooo`3oool0D03oool00?l0ooooG`3oool0
0`000000oooo0?ooo`1?0?ooo`00o`3ooomP0?ooo`030000003oool0oooo04h0oooo003o0?ooof40
oooo00<000000?ooo`3oool0C@3oool00?l0ooooHP3oool00`000000oooo0?ooo`1<0?ooo`00o`3o
oomS0?ooo`030000003oool0oooo04/0oooo003o0?ooof@0oooo00<000000?ooo`3oool0BP3oool0
0?l0ooooI@3oool00`000000oooo0?ooo`190?ooo`00o`3ooomV0?ooo`030000003oool0oooo04P0
oooo003o0?ooofL0oooo00<000000?ooo`3oool0A`3oool00?l0ooooJ03oool00`000000oooo0?oo
o`160?ooo`00o`3ooomY0?ooo`030000003oool0oooo04D0oooo003o0?ooofX0oooo00<000000?oo
o`3oool0A03oool00?l0ooooJ`3oool00`000000oooo0?ooo`130?ooo`00o`3ooom/0?ooo`030000
003oool0oooo0480oooo003o0?ooofd0oooo00<000000?ooo`3oool0@@3oool00?l0ooooKP3oool0
0`000000oooo0?ooo`100?ooo`00o`3ooom_0?ooo`030000003oool0oooo03l0oooo003o0?ooog00
oooo00<000000?ooo`3oool0?P3oool00?l0ooooL@3oool00`000000oooo0?ooo`0m0?ooo`00o`3o
oomb0?ooo`030000003oool0oooo03`0oooo003o0?ooog<0oooo00<000000?ooo`3oool0>`3oool0
0?l0ooooM03oool00`000000oooo0?ooo`0j0?ooo`00o`3ooome0?ooo`030000003oool0oooo03T0
oooo003o0?ooogH0oooo00<000000?ooo`3oool0>03oool00?l0ooooM`3oool00`000000oooo0?oo
o`0g0?ooo`00o`3ooomh0?ooo`030000003oool0oooo03H0oooo003o0?ooogT0oooo00<000000?oo
o`3oool0=@3oool00?l0ooooNP3oool00`000000oooo0?ooo`0d0?ooo`00o`3ooomk0?ooo`030000
003oool0oooo03<0oooo003o0?ooog`0oooo00<000000?ooo`3oool0<P3oool00?l0ooooO@3oool0
0`000000oooo0?ooo`0a0?ooo`00o`3ooomn0?ooo`030000003oool0oooo0300oooo003o0?ooogl0
oooo00<000000?ooo`3oool0;`3oool00?l0ooooP03oool00`000000oooo0?ooo`0^0?ooo`00o`3o
oon10?ooo`030000003oool0oooo02d0oooo003o0?oooh80oooo00<000000?ooo`3oool0;03oool0
0?l0ooooP`3oool00`000000oooo0?ooo`0[0?ooo`00o`3ooon40?ooo`030000003oool0oooo02X0
oooo003o0?ooohD0oooo00<000000?ooo`3oool0:@3oool00?l0ooooQP3oool00`000000oooo0?oo
o`0X0?ooo`00o`3ooon70?ooo`030000003oool0oooo02L0oooo003o0?ooohP0oooo00<000000?oo
o`3oool09P3oool00?l0ooooR@3oool00`000000oooo0?ooo`0U0?ooo`00o`3ooon:0?ooo`030000
003oool0oooo02@0oooo003o0?oooh/0oooo00<000000?ooo`3oool08`3oool00?l0ooooS03oool0
0`000000oooo0?ooo`0R0?ooo`00o`3ooon=0?ooo`030000003oool0oooo0240oooo003o0?ooohh0
oooo00<000000?ooo`3oool0803oool00?l0ooooS`3oool00`000000oooo0?ooo`0O0?ooo`00o`3o
oon@0?ooo`030000003oool0oooo01h0oooo003o0?oooi40oooo00<000000?ooo`3oool07@3oool0
0?l0ooooTP3oool00`000000oooo0?ooo`0L0?ooo`00o`3ooonC0?ooo`030000003oool0oooo01/0
oooo003o0?oooi@0oooo00<000000?ooo`3oool06P3oool00?l0ooooU@3oool00`000000oooo0?oo
o`0I0?ooo`00o`3ooonF0?ooo`030000003oool0oooo01P0oooo003o0?oooiL0oooo00<000000?oo
o`3oool05`3oool00?l0ooooV03oool00`000000oooo0?ooo`0F0?ooo`00o`3ooonI0?ooo`030000
003oool0oooo01D0oooo003o0?oooiX0oooo00<000000?ooo`3oool0503oool00?l0ooooV`3oool0
0`000000oooo0?ooo`0C0?ooo`00o`3ooonL0?ooo`030000003oool0oooo0180oooo003o0?oooid0
oooo00<000000?ooo`3oool04@3oool00?l0ooooWP3oool010000000oooo0?ooo`00000?0?ooo`00
o`3ooonO0?ooo`@000003P3oool00?l0ooooW`3oool4000000h0oooo003o0?oooih0oooo1P00000=
0?ooo`00o`3ooonO0?ooo`D000003@3oool00?l0ooooX@3oool4000000`0oooo003o0?oooj<0oooo
0P00000<0?ooo`00o`3ooonU0?ooo`030000003oool0oooo00T0oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {287, 0}} -> {-2.07684, -1.05, 0.0109757, \
0.0109757}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(vectorList = 
      Join[\((describeVector[#1] &)\) /@ {{1, 1}, {\(-1\), 
              2}}, {describeVector[{2, 3}, {\(-2\), \(-5\)}]}]\)], "Input"],

Cell[BoxData[
    \({{{1, 1}, {0, 0}, {HeadScaling \[Rule] Absolute, 
          HeadLength \[Rule] 5.`, HeadWidth \[Rule] 0.5`}}, {{\(-1\), 2}, {0, 
          0}, {HeadScaling \[Rule] Absolute, HeadLength \[Rule] 5.`, 
          HeadWidth \[Rule] 0.5`}}, {{2, 
          3}, {\(-2\), \(-5\)}, {HeadScaling \[Rule] Absolute, 
          HeadLength \[Rule] 5.`, HeadWidth \[Rule] 0.5`}}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[drawVectors[vectorList]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.65873 0.31746 0.435146 0.0840863 [
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.65873 .43515 m
.97619 .51923 L
s
% Start of user PostScript
0. 0. 1. 1. MAarrowhead1
% End of user PostScript
.65873 .43515 m
.34127 .60332 L
s
% Start of user PostScript
0. 0. -1. 2. MAarrowhead1
% End of user PostScript
.02381 .01472 m
.65873 .26697 L
s
% Start of user PostScript
-2. -5. 0. -2. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`002P3oool200000?l0ooooY@3oool000`0oooo0P00003o0?oooj<0oooo000>0?ooo`<00000
o`3ooonP0?ooo`004@3oool200000?l0ooooWP3oool001<0oooo0`00003o0?oooi/0oooo000F0?oo
o`800000o`3ooonI0?ooo`00603oool300000?l0ooooUP3oool001/0oooo0P00003o0?oooi@0oooo
000M0?ooo`<00000o`3ooonA0?ooo`00803oool300000?l0ooooSP3oool002<0oooo0P00003o0?oo
oh`0oooo000U0?ooo`<00000o`3ooon90?ooo`00:03oool200000?l0ooooQ`3oool002X0oooo0`00
003o0?oooh@0oooo000]0?ooo`800000o`3ooon20?ooo`00;`3oool300000?l0ooooO`3oool00380
oooo0P00003o0?ooogd0oooo000d0?ooo`<00000o`3ooomj0?ooo`00=`3oool200000?l0ooooN03o
ool003T0oooo0`00003o0?ooogD0oooo000l0?ooo`800000o`3ooomc0?ooo`00?P3oool300000?l0
ooooL03oool00440oooo0P00003o0?ooofh0oooo00130?ooo`<00000o`3ooom[0?ooo`00AP3oool2
00000?l0ooooJ@3oool004P0oooo0`00003o0?ooofH0oooo001;0?ooo`800000o`3ooomT0?ooo`00
C@3oool300000?l0ooooH@3oool00500oooo0`00003o0?oooeh0oooo001C0?ooo`800000o`3ooomL
0?ooo`00E@3oool300000?l0ooooF@3oool005P0oooo0P00003o0?oooeL0oooo001J0?ooo`<00000
o`3ooomD0?ooo`00G@3oool200000?l0ooooDP3oool005l0oooo0`00003o0?ooodl0oooo001R0?oo
o`800000o`3ooom=0?ooo`00I03oool300000?l0ooooBP3oool006L0oooo0P00003o0?ooodP0oooo
001Y0?ooo`<00000o`3ooom50?ooo`00K03oool200000?l0oooo@`3oool006h0oooo0`00003o0?oo
od00oooo001a0?ooo`800000o`3oooln0?ooo`00L`3oool300000?l0oooo>`3oool007H0oooo0P00
003o0?ooocT0oooo001h0?ooo`<00000o`3ooolf0?ooo`00N`3oool300000?l0oooo<`3oool007h0
oooo0P00003o0?oooc40oooo00200?ooo`<00000o`3oool^0?ooo`00P`3oool200000?l0oooo;03o
ool008D0oooo0`00003o0?ooobT0oooo00280?ooo`800000o`3ooolW0?ooo`00RP3oool300000?l0
oooo903oool008d0oooo0P00003o0?ooob80oooo002?0?ooo`<00000o`3ooolO0?ooo`00TP3oool2
00000?l0oooo7@3oool009@0oooo0`00003o0?oooaX0oooo002G0?ooo`800000o`3ooolH0?ooo`00
V@3oool300000?l0oooo5@3oool009`0oooo0P00003o0?oooa<0oooo002N0?ooo`<00000o`3oool@
0?ooo`00X@3oool200000?l0oooo3P3oool00:<0oooo0`00003o0?ooo`/0oooo002V0?ooo`800000
o`3oool90?ooo`00Z03oool300000?l0oooo1P3oool00:/0oooo0`00003o0?ooo`<0oooo002^0?oo
o`800000o`3oool10?ooo`00/03oool300000?d0oooo002c0?ooo`800000n`3oool00;D0oooo0`00
003h0?ooo`00^03oool200000?H0oooo002j0?ooo`<00000l`3oool00;d0oooo0P00003a0?ooo`00
_`3oool300000>h0oooo00320?ooo`800000k03oool00<@0oooo0`00003Y0?ooo`00a`3oool20000
0>L0oooo00390?ooo`<00000i03oool00<`0oooo0P00003R0?ooo`00cP3oool300000=l0oooo003A
0?ooo`800000g@3oool00=<0oooo0`00003J0?ooo`00eP3oool300000=L0oooo003I0?ooo`800000
e@3oool00=/0oooo0`00003B0?ooo`00gP3oool200000=00oooo003P0?ooo`<00000c@3oool00><0
oooo0P00003;0?ooo`00i@3oool300000<P0oooo003X0?ooo`800000aP3oool00>X0oooo0`000033
0?ooo`00k@3oool200000<40oooo003_0?ooo`<00000_P3oool00?80oooo0P00002l0?ooo`00m03o
ool300000;T0oooo003g0?ooo`800000]`3oool00?T0oooo0`00002d0?ooo`00o03oool200000;80
oooo003n0?ooo`<00000[`3oool00?l0oooo0P3oool200000:d0oooo003o0?ooo`@0oooo0`00002Z
0?ooo`00o`3oool70?ooo`<00000Y`3oool00?l0oooo2P3oool200000:D0oooo003o0?ooo``0oooo
0`00002R0?ooo`00o`3oool?0?ooo`800000X03oool00?l0oooo4@3oool300000080oooo00<00000
0?ooo`3oool0V03oool00?l0oooo503oool5000009P0oooo003o0?oooaD0oooo1@00002G0?ooo`00
o`3ooolE0?ooo`L00000U@3oool00?l0oooo6@3oool4000009@0oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oooa`0oooo0P00002C0?ooo`00o`3ooolJ0?ooo`8000000P3oool4000008l0oooo
003o0?oooaP0oooo0P0000080?ooo`@00000R`3oool00?l0oooo5P3oool2000000h0oooo10000027
0?ooo`00o`3ooolD0?ooo`800000503oool4000008<0oooo003o0?oooa80oooo0P00000J0?ooo`<0
0000P03oool00?l0oooo403oool2000001l0oooo1000001l0?ooo`00o`3oool>0?ooo`8000009@3o
ool4000007P0oooo003o0?ooo``0oooo0P00000[0?ooo`@00000M03oool00?l0oooo2`3oool00`00
0000oooo0?ooo`0_0?ooo`@00000L03oool00?l0oooo2@3oool2000003H0oooo0`00001]0?ooo`00
o`3oool70?ooo`800000>`3oool4000006T0oooo003o0?ooo`D0oooo0P0000110?ooo`@00000I@3o
ool00?l0oooo0`3oool2000004L0oooo1000001Q0?ooo`00o`3oool10?ooo`800000C@3oool40000
05d0oooo003n0?ooo`800000D`3oool3000005X0oooo003l0?ooo`800000F03oool4000005H0oooo
003j0?ooo`800000GP3oool400000580oooo003i0?ooo`030000003oool0oooo0680oooo1000001>
0?ooo`00m`3oool2000006T0oooo1000001:0?ooo`00m@3oool2000006l0oooo100000160?ooo`00
l`3oool2000007D0oooo0`0000130?ooo`00l@3oool2000007X0oooo1000000o0?ooo`00k`3oool2
00000800oooo1000000k0?ooo`00k@3oool2000008H0oooo1000000g0?ooo`00j`3oool2000008`0
oooo1000000c0?ooo`00j@3oool200000980oooo0`00000`0?ooo`00j03oool00`000000oooo0?oo
o`2E0?ooo`@00000;03oool00>H0oooo0P00002L0?ooo`@00000:03oool00>@0oooo0P00002R0?oo
o`@00000903oool00>80oooo0P00002X0?ooo`@00000803oool00>00oooo0P00002^0?ooo`<00000
7@3oool00=h0oooo0P00002c0?ooo`@000001P3oool00`000000oooo0?ooo`0@0?ooo`00g03oool2
00000;T0oooo100000020?ooo`<00000403oool00=X0oooo0P00002o0?ooo`L000003P3oool00=P0
oooo0P0000320?ooo`P00000303oool00=L0oooo00<000000?ooo`3oool0`P3oool9000000/0oooo
003E0?ooo`800000f@3oool00=<0oooo0P00003K0?ooo`00d@3oool200000=d0oooo003?0?ooo`80
0000g`3oool00<d0oooo0P00003Q0?ooo`00b`3oool200000><0oooo00390?ooo`800000i@3oool0
0<L0oooo0P00003W0?ooo`00aP3oool00`000000oooo0?ooo`3W0?ooo`00a03oool200000>X0oooo
00320?ooo`800000k03oool00<00oooo0P00003^0?ooo`00_P3oool200000?00oooo002l0?ooo`80
0000lP3oool00;X0oooo0P00003d0?ooo`00^03oool200000?H0oooo002f0?ooo`800000n03oool0
0;D0oooo00<000000?ooo`3oool0n03oool00;<0oooo0P00003k0?ooo`00/@3oool200000?d0oooo
002_0?ooo`800000o`3oool00:d0oooo0P00003o0?ooo`80oooo002[0?ooo`800000o`3oool40?oo
o`00Z@3oool200000?l0oooo1P3oool00:L0oooo0P00003o0?ooo`P0oooo002U0?ooo`800000o`3o
ool:0?ooo`00Y03oool00`000000oooo0?ooo`3o0?ooo`X0oooo002R0?ooo`800000o`3oool=0?oo
o`00X03oool200000?l0oooo3`3oool009h0oooo0P00003o0?oooa40oooo002H0?ooo`040000003o
ool0oooo0?ooo`800000o`3ooolC0?ooo`00U`3oool500000?l0oooo5@3oool009H0oooo1000003o
0?oooaL0oooo002E0?ooo`H00000o`3ooolF0?ooo`00U03oool500000?l0oooo603oool009<0oooo
0P00003o0?oooa`0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.07796, -5.17503, \
0.0109834, 0.041467}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["One dimensional collisions", "Subsection"],

Cell[CellGroupData[{

Cell["\<\
General one-dimensional inelastic collision with mass transfer\
\>", "Subsubsection"],

Cell[TextData[{
  "It is easier to use scalar equations for one-dimensional collisions than \
to decide whether the scattering angle should be 0 or ",
  Cell[BoxData[
      \(TraditionalForm\`\[Pi]\)]],
  ". "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(collisionEqs1D = collisionEqs /. {Dot \[Rule] Times}\)], "Input"],

Cell[BoxData[
    \({m1 + m2 \[Equal] m3 + m4, p1 + p2 \[Equal] p3 + p4, 
      p1\^2\/\(2\ m1\) + p2\^2\/\(2\ m2\) + qvalue \[Equal] 
        p3\^2\/\(2\ m3\) + p4\^2\/\(2\ m4\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol = \ \(\({p1, p2, p3, p4} /. 
          Solve[collisionEqs1D, {p3, p4, 
              m4}]\)\(//\)\(MySimplify\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \({{p1, 
        p2, \(m1\ m2\ m3\ p1 + m1\ m2\ m3\ p2 + \@\(m1\ m2\ \((m1 + m2 - \
m3)\)\ m3\ \((m1\^2\ p2\^2 + 2\ m1\ m2\ \((\(-p1\)\ p2 + m1\ qvalue)\) + \
m2\^2\ \((p1\^2 + 2\ m1\ qvalue)\))\)\)\)\/\(m1\ m2\ \((m1 + m2)\)\), \(\(1\/\
\(m1\ m2\ \((m1 + m2)\)\)\)\((m1\^2\ m2\ p1 + m1\ m2\^2\ p1 - 
            m1\ m2\ m3\ p1 + m1\^2\ m2\ p2 + m1\ m2\^2\ p2 - 
            m1\ m2\ m3\ p2 - \@\(m1\ m2\ \((m1 + m2 - m3)\)\ m3\ \((m1\^2\ p2\
\^2 + 2\ m1\ m2\ \((\(-p1\)\ p2 + m1\ qvalue)\) + m2\^2\ \((p1\^2 + 2\ m1\ \
qvalue)\))\)\))\)\)}, {p1, 
        p2, \(m1\ m2\ m3\ p1 + m1\ m2\ m3\ p2 - \@\(m1\ m2\ \((m1 + m2 - \
m3)\)\ m3\ \((m1\^2\ p2\^2 + 2\ m1\ m2\ \((\(-p1\)\ p2 + m1\ qvalue)\) + \
m2\^2\ \((p1\^2 + 2\ m1\ qvalue)\))\)\)\)\/\(m1\ m2\ \((m1 + m2)\)\), \(\(1\/\
\(m1\ m2\ \((m1 + m2)\)\)\)\((m1\^2\ m2\ p1 + m1\ m2\^2\ p1 - 
            m1\ m2\ m3\ p1 + m1\^2\ m2\ p2 + m1\ m2\^2\ p2 - 
            m1\ m2\ m3\ p2 + \@\(m1\ m2\ \((m1 + m2 - m3)\)\ m3\ \((m1\^2\ p2\
\^2 + 2\ m1\ m2\ \((\(-p1\)\ p2 + m1\ qvalue)\) + m2\^2\ \((p1\^2 + 2\ m1\ \
qvalue)\))\)\))\)\)}}\)], "Output"]
}, Closed]],

Cell["\<\
To select the appropriate solution, consider an elastic collision between \
equal masses. \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(sol /. {m3 \[Rule] m1, m4 \[Rule] m2, p2 \[Rule] 0, 
              qvalue \[Rule] 0}\  // Simplify\) // PowerExpand\) // 
      Simplify\)], "Input"],

Cell[BoxData[
    \({{p1, 0, p1, 0}, {p1, 
        0, \(\((m1 - m2)\)\ p1\)\/\(m1 + m2\), \(2\ m2\ p1\)\/\(m1 + \
m2\)}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Recognizing that the first solution corresponds to no scattering (",
  Cell[BoxData[
      FormBox[
        StyleBox[\(p3 \[Rule] p1\),
          "InlineInput"], TraditionalForm]]],
  ") in this case, we choose the second solution."
}], "Text"],

Cell[BoxData[
    \(\(momenta1D = 
        sol\[LeftDoubleBracket]2\[RightDoubleBracket];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(velocities1D\  = \(\(momenta1D\/masses /. {p1 \[Rule] m1\ v1, 
                p2 \[Rule] m2\ v2} // Simplify\) // PowerExpand\) // 
        Simplify\)], "Input"],

Cell[BoxData[
    \({v1, 
      v2, \(m1\ v1 - \(\@\(m1 + m2 - m3\)\ \@\(2\ m2\ qvalue + m1\ \((2\ \
qvalue + m2\ \((v1 - v2)\)\^2)\)\)\)\/\@m3 + m2\ v2\)\/\(m1 + m2\), \(m1\^2\ \
v1 + \@\(m1 + m2 - m3\)\ \@m3\ \@\(2\ m2\ qvalue + m1\ \((2\ qvalue + m2\ \
\((v1 - v2)\)\^2)\)\) + m2\^2\ v2 - m2\ m3\ v2 + m1\ \((\(-m3\)\ v1 + m2\ \
\((v1 + v2)\))\)\)\/\(\((m1 + m2)\)\ m4\)}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["One-dimensional elastic collisions", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(momenta1D /. {m3 \[Rule] m1, m4 \[Rule] m2, qvalue \[Rule] 0} // 
          Simplify\) // PowerExpand\) // Simplify\)], "Input"],

Cell[BoxData[
    \({p1, 
      p2, \(m1\ p1 - m2\ p1 + 2\ m1\ p2\)\/\(m1 + m2\), \(2\ m2\ p1 - m1\ p2 \
+ m2\ p2\)\/\(m1 + m2\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabVelocities = \(\(\(velocities1D /. {m3 \[Rule] m1, m4 \[Rule] m2, 
                  qvalue \[Rule] 0} // Simplify\) // PowerExpand\) // 
          Simplify\) // Collect[#, {v1, v2}] &\)], "Input"],

Cell[BoxData[
    \({v1, 
      v2, \(\((m1 - m2)\)\ v1\)\/\(m1 + m2\) + \(2\ m2\ v2\)\/\(m1 + m2\), \
\(2\ m1\ v1\)\/\(m1 + m2\) + \(\((\(-m1\) + m2)\)\ v2\)\/\(m1 + m2\)}\)], \
"Output"]
}, Closed]],

Cell["It is useful to examine a few limiting cases.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabVelocities /. {m2 \[Rule] m1}\)], "Input"],

Cell[BoxData[
    \({v1, v2, v2, v1}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Thus, we find that when the masses are equal a one-dimensional elastic \
collision simply exchanges the velocities of the two particles.  For example, \
if ",
  Cell[BoxData[
      \(TraditionalForm\`m\_2\)]],
  " is initially at rest, after the collision it acquires the original \
velocity of ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  " is brought to rest."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Limit[LabVelocities, m2 \[Rule] \[Infinity]]\)], "Input"],

Cell[BoxData[
    \({v1, v2, \(-v1\) + 2\ v2, v2}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "On the other hand, if the target is extremely massive its velocity is \
unaffected by the collision whereas the velocity of the projectile is \
reversed and is increased by twice the target velocity.  For example, suppose \
that ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  " represents a baseball and that ",
  Cell[BoxData[
      \(TraditionalForm\`m\_2\)]],
  " represents the combined mass of the bat and batter and is much larger \
than ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  ".  The speed of a line drive is then the speed of the pitch plus twice the \
bat speed."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Limit[LabVelocities, m1 \[Rule] \[Infinity]]\)], "Input"],

Cell[BoxData[
    \({v1, v2, v1, 2\ v1 - v2}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Alternatively, if the projectile is much more massive than the target, the \
recoiling target acquires twice the projectile velocity minus its original \
velocity.  For example, if ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1\)]],
  " is a large truck which strikes a small car, ",
  Cell[BoxData[
      \(TraditionalForm\`m\_2\)]],
  ", originally at rest at a stop light, the car recoils with twice the \
truck's velocity while the truck continues with its original velocity \
(assuming that the collision is approximately elastic)."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["One-dimensional inelastic collision without mass transfer", 
  "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabMomenta = \(\(momenta1D /. {m3 \[Rule] m1, m4 \[Rule] m2} // 
            Simplify\) // PowerExpand\) // Simplify\)], "Input"],

Cell[BoxData[
    \({p1, 
      p2, \(m1\ \((p1 + p2)\) - \@\(m1\^2\ p2\^2 + 2\ m1\ m2\ \((\(-p1\)\ p2 \
+ m1\ qvalue)\) + m2\^2\ \((p1\^2 + 2\ m1\ qvalue)\)\)\)\/\(m1 + m2\), \(m2\ \
\((p1 + p2)\) + \@\(m1\^2\ p2\^2 + 2\ m1\ m2\ \((\(-p1\)\ p2 + m1\ qvalue)\) \
+ m2\^2\ \((p1\^2 + 2\ m1\ qvalue)\)\)\)\/\(m1 + m2\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabVelocities = \(\(velocities1D /. {m3 \[Rule] m1, m4 \[Rule] m2} // 
            Simplify\) // PowerExpand\) // Simplify\)], "Input"],

Cell[BoxData[
    \({v1, 
      v2, \(m1\ v1 - \(\@m2\ \@\(2\ m2\ qvalue + m1\ \((2\ qvalue + m2\ \((v1 \
- v2)\)\^2)\)\)\)\/\@m1 + m2\ v2\)\/\(m1 + m2\), \(m1\ v1 + \(\@m1\ \@\(2\ m2\
\ qvalue + m1\ \((2\ qvalue + m2\ \((v1 - v2)\)\^2)\)\)\)\/\@m2 + m2\ \
v2\)\/\(m1 + m2\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabVelocities /. {v2 \[Rule] 0}\  // Simplify\)], "Input"],

Cell[BoxData[
    \({v1, 
      0, \(m1\ v1 - \(\@m2\ \@\(2\ m1\ qvalue + 2\ m2\ qvalue + m1\ m2\ v1\^2\
\)\)\/\@m1\)\/\(m1 + m2\), \(m1\ v1 + \(\@m1\ \@\(2\ m1\ qvalue + 2\ m2\ \
qvalue + m1\ m2\ v1\^2\)\)\/\@m2\)\/\(m1 + m2\)}\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Elastic colinear collision for equal masses", "Subsection"],

Cell[BoxData[
    \(\(values = {m1 \[Rule] 1, m2 \[Rule] m1, m3 \[Rule] m1, 
          qvalue \[Rule] 0, p1r \[Rule] 1, theta1 \[Rule] 0, p2r \[Rule] 0, 
          theta2 \[Rule] \[Pi]};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol = 
      Simplify[PowerExpand[\(generalSolution[p3r, p4x, 
                p4y]\)\[LeftDoubleBracket]1\[RightDoubleBracket] //. 
            values]]\)], "Input"],

Cell[BoxData[
    \({p4x \[Rule] Sin[theta3]\^2, p4y \[Rule] \(-Cos[theta3]\)\ Sin[theta3], 
      p3r \[Rule] Cos[theta3], m4 \[Rule] 1}\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(calcCollision[sol, values, 60\ Degree]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabMomenta\)], "Input"],

Cell[BoxData[
    \({{1, 0}, {0, 0}, {1\/4, \@3\/4}, {3\/4, \(-\(\@3\/4\)\)}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(LabMomenta\[LeftDoubleBracket]3\[RightDoubleBracket] . 
      LabMomenta\[LeftDoubleBracket]4\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawCollision @@ LabMomenta;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .49487 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.568027 0.544218 0.247436 0.544218 [
[ 0 0 0 0 ]
[ 1 .49487 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .49487 L
0 .49487 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .24744 m
.56803 .24744 L
s
% Start of user PostScript
-1. 0. 0. 0. MAarrowhead1
% End of user PostScript
.56803 .24744 m
.56803 .24744 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.56803 .24744 m
.70408 .48309 L
s
% Start of user PostScript
0. 0. 0.25 0.4330127 MAarrowhead1
% End of user PostScript
.56803 .24744 m
.97619 .01178 L
s
% Start of user PostScript
0. 0. 0.75 -0.4330127 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 142.5},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000eB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooonT
0?ooo`8000002`3oool00?l0ooooX03oool5000000`0oooo003o0?oooid0oooo1`00000=0?ooo`00
o`3ooonN0?ooo`D000003P3oool00?l0ooooWP3oool300000100oooo003o0?oooi`0oooo0P000000
0`3oool000000?ooo`0@0?ooo`00o`3ooonJ0?ooo`8000005@3oool00?l0ooooV03oool2000001L0
oooo003o0?oooiL0oooo00<000000?ooo`3oool05`3oool00?l0ooooU@3oool2000001X0oooo003o
0?oooi<0oooo0P00000L0?ooo`00o`3ooonA0?ooo`8000007P3oool00?l0ooooT03oool00`000000
oooo0?ooo`0N0?ooo`00o`3ooon>0?ooo`8000008@3oool00?l0ooooS03oool2000002<0oooo003o
0?ooohX0oooo0P00000U0?ooo`00o`3ooon90?ooo`030000003oool0oooo02D0oooo003o0?ooohL0
oooo0P00000X0?ooo`00o`3ooon50?ooo`800000:P3oool00?l0ooooQ03oool00`000000oooo0?oo
o`0Z0?ooo`00o`3ooon20?ooo`800000;@3oool00?l0ooooP03oool2000002l0oooo003o0?ooogh0
oooo0P00000a0?ooo`00o`3ooomm0?ooo`030000003oool0oooo0340oooo003o0?ooog/0oooo0P00
000d0?ooo`00o`3ooomi0?ooo`800000=P3oool00?l0ooooM`3oool2000003P0oooo003o0?ooogH0
oooo00<000000?ooo`3oool0>03oool00?l0ooooM03oool2000003/0oooo003o0?ooog80oooo0P00
000m0?ooo`00o`3ooom`0?ooo`800000?`3oool00?l0ooooK`3oool00`000000oooo0?ooo`0o0?oo
o`00o`3ooom]0?ooo`800000@P3oool00?l0ooooJ`3oool2000004@0oooo003o0?ooofT0oooo0P00
00160?ooo`00o`3ooomX0?ooo`030000003oool0oooo04H0oooo003o0?ooofH0oooo0P0000190?oo
o`00o`3ooomT0?ooo`800000B`3oool00?l0ooooHP3oool2000004d0oooo003o0?ooof40oooo00<0
00000?ooo`3oool0C@3oool00?l0ooooG`3oool200000500oooo003o0?oooed0oooo0P00001B0?oo
o`00o`3ooomK0?ooo`800000E03oool00?l0ooooFP3oool00`000000oooo0?ooo`1D0?ooo`00o`3o
oomH0?ooo`800000E`3oool00?l0ooooEP3oool2000005T0oooo003o0?oooe@0oooo0P00001K0?oo
o`00o`3ooomC0?ooo`030000003oool0oooo05/0oooo003o0?oooe40oooo0P00001N0?ooo`00o`3o
oom?0?ooo`800000H03oool00?l0ooooCP3oool00`000000oooo0?ooo`1P0?ooo`00o`3ooom<0?oo
o`800000H`3oool00?l0ooooBP3oool2000006D0oooo003o0?ooodP0oooo0P00001W0?ooo`00o`3o
oom70?ooo`030000003oool0oooo06L0oooo003o0?ooodD0oooo0P00001Z0?ooo`00o`3ooom30?oo
o`800000K03oool00?l0oooo@@3oool2000006h0oooo003o0?oood00oooo00<000000?ooo`3oool0
KP3oool00?l0oooo?P3oool200000740oooo003o0?oooc`0oooo0P00001c0?ooo`00o`3ooolj0?oo
o`800000M@3oool00?l0oooo>@3oool00`000000oooo0?ooo`1e0?ooo`00o`3ooolg0?ooo`800000
N03oool00?l0oooo=@3oool2000007X0oooo003o0?oooc<0oooo0P00001l0?ooo`00o`3ooolb0?oo
o`030000003oool0oooo07`0oooo003o0?oooc00oooo0P00001o0?ooo`00o`3oool^0?ooo`800000
P@3oool00?l0oooo;03oool2000008<0oooo003o0?ooob/0oooo00<000000?ooo`3oool0P`3oool0
0?l0oooo:@3oool2000008H0oooo003o0?ooobL0oooo0P0000280?ooo`00o`3ooolU0?ooo`800000
RP3oool00?l0oooo903oool00`000000oooo0?ooo`2:0?ooo`00o`3ooolR0?ooo`800000S@3oool0
0?l0oooo803oool2000008l0oooo003o0?oooah0oooo0P00002A0?ooo`00o`3ooolM0?ooo`030000
003oool0oooo0940oooo003o0?oooa/0oooo0P00002D0?ooo`00o`3ooolI0?ooo`800000UP3oool0
0?l0oooo5`3oool2000009P0oooo003o0?oooaH0oooo00<000000?ooo`3oool0V03oool00?l0oooo
503oool2000009/0oooo003o0?oooa80oooo0P00002M0?ooo`00o`3ooolA0?ooo`030000003oool0
oooo09d0oooo003o0?ooo`l0oooo0P00002P0?ooo`00o`3oool=0?ooo`800000XP3oool00?l0oooo
2`3oool200000:@0oooo003o0?ooo`X0oooo00<000000?ooo`3oool0Y03oool00?l0oooo203oool2
00000:L0oooo003o0?ooo`H0oooo0P00002Y0?ooo`00o`3oool40?ooo`800000Z`3oool00?l0oooo
0`3oool00`000000oooo0?ooo`2[0?ooo`00o`3oool10?ooo`800000[P3oool00?h0oooo0P00002`
0?ooo`00o03oool200000;80oooo003k0?ooo`030000003oool0oooo0;80oooo003i0?ooo`800000
]@3oool00>d0oooo0P0000080?ooo`800000]`3oool00>d0oooo1P0000020?ooo`800000^@3oool0
00X0ooooj`00002k0?ooo`00k@3oool600000080oooo00<000000?ooo`3oool0^03oool00>d0oooo
0P0000060?ooo`030000003oool0oooo0;P0oooo003f0?ooo`030000003oool0oooo0;L0oooo003f
0?ooo`030000003oool0oooo0;L0oooo003g0?ooo`030000003oool0oooo0;H0oooo003h0?ooo`03
0000003oool0oooo0;D0oooo003h0?ooo`030000003oool0oooo0;D0oooo003i0?ooo`030000003o
ool0oooo0;@0oooo003i0?ooo`030000003oool0oooo0;@0oooo003j0?ooo`030000003oool0oooo
0;<0oooo003j0?ooo`030000003oool0oooo0;<0oooo003k0?ooo`030000003oool0oooo0;80oooo
003l0?ooo`030000003oool0oooo0;40oooo003l0?ooo`030000003oool0oooo0;40oooo003m0?oo
o`030000003oool0oooo0;00oooo003m0?ooo`030000003oool0oooo0;00oooo003n0?ooo`030000
003oool0oooo0:l0oooo003o0?ooo`030000003oool0oooo0:h0oooo003o0?ooo`030000003oool0
oooo0:h0oooo003o0?ooo`40oooo00<000000?ooo`3oool0[@3oool00?l0oooo0@3oool00`000000
oooo0?ooo`2]0?ooo`00o`3oool20?ooo`030000003oool0oooo0:`0oooo003o0?ooo`80oooo00<0
00000?ooo`3oool0[03oool00?l0oooo0`3oool00`000000oooo0?ooo`2[0?ooo`00o`3oool40?oo
o`030000003oool0oooo0:X0oooo003o0?ooo`@0oooo00<000000?ooo`3oool0ZP3oool00?l0oooo
1@3oool00`000000oooo0?ooo`2Y0?ooo`00o`3oool50?ooo`030000003oool0oooo0:T0oooo003o
0?ooo`H0oooo00<000000?ooo`3oool0Z03oool00?l0oooo1`3oool00`000000oooo0?ooo`2W0?oo
o`00o`3oool70?ooo`030000003oool0oooo0:L0oooo003o0?ooo`P0oooo00<000000?ooo`3oool0
YP3oool00?l0oooo203oool00`000000oooo0?ooo`2V0?ooo`00o`3oool90?ooo`030000003oool0
oooo0:D0oooo003o0?ooo`T0oooo00<000000?ooo`3oool0Y@3oool00?l0oooo2P3oool00`000000
oooo0?ooo`2T0?ooo`00o`3oool;0?ooo`030000003oool0oooo0:<0oooo003o0?ooo`/0oooo00<0
00000?ooo`3oool0X`3oool00?l0oooo303oool00`000000oooo0?ooo`2R0?ooo`00o`3oool<0?oo
o`030000003oool0oooo0:80oooo003o0?ooo`d0oooo00<000000?ooo`3oool0X@3oool00?l0oooo
3P3oool00`000000oooo0?ooo`2P0?ooo`00o`3oool>0?ooo`030000003oool0oooo0:00oooo003o
0?ooo`l0oooo00<000000?ooo`3oool0W`3oool00?l0oooo3`3oool00`000000oooo0?ooo`2O0?oo
o`00o`3oool@0?ooo`030000003oool0oooo09h0oooo003o0?oooa00oooo00<000000?ooo`3oool0
WP3oool00?l0oooo4@3oool00`000000oooo0?ooo`2M0?ooo`00o`3ooolB0?ooo`030000003oool0
oooo09`0oooo003o0?oooa80oooo00<000000?ooo`3oool0W03oool00?l0oooo4`3oool00`000000
oooo0?ooo`2K0?ooo`00o`3ooolC0?ooo`030000003oool0oooo09/0oooo003o0?oooa@0oooo00<0
00000?ooo`3oool0VP3oool00?l0oooo5@3oool00`000000oooo0?ooo`2I0?ooo`00o`3ooolE0?oo
o`030000003oool0oooo09T0oooo003o0?oooaH0oooo00<000000?ooo`3oool0V03oool00?l0oooo
5P3oool00`000000oooo0?ooo`2H0?ooo`00o`3ooolG0?ooo`030000003oool0oooo09L0oooo003o
0?oooaL0oooo00<000000?ooo`3oool0U`3oool00?l0oooo603oool00`000000oooo0?ooo`2F0?oo
o`00o`3ooolI0?ooo`030000003oool0oooo09D0oooo003o0?oooaT0oooo00<000000?ooo`3oool0
U@3oool00?l0oooo6P3oool00`000000oooo0?ooo`2D0?ooo`00o`3ooolJ0?ooo`030000003oool0
oooo09@0oooo003o0?oooa/0oooo00<000000?ooo`3oool0T`3oool00?l0oooo703oool00`000000
oooo0?ooo`2B0?ooo`00o`3ooolL0?ooo`030000003oool0oooo0980oooo003o0?oooad0oooo00<0
00000?ooo`3oool0T@3oool00?l0oooo7@3oool00`000000oooo0?ooo`2A0?ooo`00o`3ooolN0?oo
o`030000003oool0oooo0900oooo003o0?oooah0oooo00<000000?ooo`3oool0T03oool00?l0oooo
7`3oool00`000000oooo0?ooo`2?0?ooo`00o`3ooolP0?ooo`030000003oool0oooo08h0oooo003o
0?ooob00oooo00<000000?ooo`3oool0SP3oool00?l0oooo8@3oool00`000000oooo0?ooo`2=0?oo
o`00o`3ooolQ0?ooo`030000003oool0oooo08d0oooo003o0?ooob80oooo00<000000?ooo`3oool0
S03oool00?l0oooo8`3oool00`000000oooo0?ooo`2;0?ooo`00o`3ooolS0?ooo`030000003oool0
oooo08/0oooo003o0?ooob@0oooo00<000000?ooo`3oool0RP3oool00?l0oooo903oool00`000000
oooo0?ooo`2:0?ooo`00o`3ooolU0?ooo`030000003oool0oooo08T0oooo003o0?ooobD0oooo00<0
00000?ooo`3oool0R@3oool00?l0oooo9P3oool00`000000oooo0?ooo`280?ooo`00o`3ooolW0?oo
o`030000003oool0oooo08L0oooo003o0?ooobL0oooo00<000000?ooo`3oool0Q`3oool00?l0oooo
:03oool00`000000oooo0?ooo`260?ooo`00o`3ooolX0?ooo`030000003oool0oooo08H0oooo003o
0?ooobT0oooo00<000000?ooo`3oool0Q@3oool00?l0oooo:P3oool00`000000oooo0?ooo`240?oo
o`00o`3ooolZ0?ooo`030000003oool0oooo08@0oooo003o0?ooob/0oooo00<000000?ooo`3oool0
P`3oool00?l0oooo:`3oool00`000000oooo0?ooo`230?ooo`00o`3oool/0?ooo`030000003oool0
00000880oooo003o0?ooob`0oooo0`0000220?ooo`00o`3ooolZ0?ooo`H00000P@3oool00?l0oooo
:`3oool500000840oooo003o0?ooob`0oooo100000210?ooo`00o`3oool]0?ooo`<00000P@3oool0
0?l0oooo;`3oool200000800oooo003o0?oooc00oooo00<000000?ooo`3oool0OP3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {212.75, 0}} -> {-1.04609, -0.454668, \
0.00641127, 0.00641127}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawVectorSum[LabMomenta\[LeftDoubleBracket]3\[RightDoubleBracket], 
        LabMomenta\[LeftDoubleBracket]4\[RightDoubleBracket]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .43301 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0103098 0.952381 [
[ 0 0 0 0 ]
[ 1 .43301 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .43301 L
0 .43301 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .01031 m
.2619 .4227 L
s
% Start of user PostScript
0. 0. 0.25 0.4330127 MAarrowhead1
% End of user PostScript
.2619 .4227 m
.97619 .01031 L
s
% Start of user PostScript
0.25 0.4330127 1. 0. MAarrowhead1
% End of user PostScript
1 Mabswid
.02381 .01031 m
.97619 .01031 L
s
% Start of user PostScript
0. 0. 1. 0. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 124.688},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000^b000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooonM0?ooo`8000004P3oool00?l0ooooW@3o
ool6000000h0oooo000:0?ooool00000W000000;0?ooo`002`3oool00`000000oooo0?ooo`3o0?oo
ohl0oooo2000000<0?ooo`002`3oool00`000000oooo0?ooo`3o0?ooohl0oooo1`00000=0?ooo`00
303oool00`000000oooo0?ooo`3o0?ooohl0oooo1@00000>0?ooo`00303oool00`000000oooo0?oo
o`3o0?ooohl0oooo0`00000@0?ooo`003@3oool00`000000oooo0?ooo`3o0?oooh`0oooo0P000000
0`3oool000000?ooo`0@0?ooo`003@3oool00`000000oooo0?ooo`3o0?ooohX0oooo0P00000E0?oo
o`003P3oool00`000000oooo0?ooo`3o0?ooohL0oooo0P00000G0?ooo`003`3oool00`000000oooo
0?ooo`3o0?ooohD0oooo00<000000?ooo`3oool05`3oool000l0oooo00<000000?ooo`3oool0o`3o
oon30?ooo`8000006P3oool00100oooo00<000000?ooo`3oool0o`3ooon00?ooo`800000703oool0
0100oooo00<000000?ooo`3oool0o`3ooomo0?ooo`030000003oool0oooo01`0oooo000A0?ooo`03
0000003oool0oooo0?l0ooooO03oool2000001l0oooo000B0?ooo`030000003oool0oooo0?l0oooo
N@3oool200000240oooo000B0?ooo`030000003oool0oooo0?l0ooooM`3oool2000002<0oooo000C
0?ooo`030000003oool0oooo0?l0ooooM@3oool00`000000oooo0?ooo`0S0?ooo`004`3oool00`00
0000oooo0?ooo`3o0?ooog<0oooo0P00000V0?ooo`00503oool00`000000oooo0?ooo`3o0?ooog00
oooo0P00000X0?ooo`00503oool00`000000oooo0?ooo`3o0?ooofh0oooo0P00000Z0?ooo`005@3o
ool00`000000oooo0?ooo`3o0?ooof`0oooo00<000000?ooo`3oool0:P3oool001H0oooo00<00000
0?ooo`3oool0o`3ooomY0?ooo`800000;@3oool001H0oooo00<000000?ooo`3oool0o`3ooomW0?oo
o`800000;`3oool001L0oooo00<000000?ooo`3oool0o`3ooomT0?ooo`800000<@3oool001L0oooo
00<000000?ooo`3oool0o`3ooomS0?ooo`030000003oool0oooo0340oooo000H0?ooo`030000003o
ool0oooo0?l0ooooH03oool2000003@0oooo000I0?ooo`030000003oool0oooo0?l0ooooG@3oool2
000003H0oooo000I0?ooo`030000003oool0oooo0?l0ooooG03oool00`000000oooo0?ooo`0f0?oo
o`006P3oool00`000000oooo0?ooo`3o0?oooeT0oooo0P00000i0?ooo`006P3oool00`000000oooo
0?ooo`3o0?oooeL0oooo0P00000k0?ooo`006`3oool00`000000oooo0?ooo`3o0?oooe@0oooo0P00
000m0?ooo`006`3oool00`000000oooo0?ooo`3o0?oooe<0oooo00<000000?ooo`3oool0?@3oool0
01`0oooo00<000000?ooo`3oool0o`3ooom@0?ooo`800000@03oool001d0oooo00<000000?ooo`3o
ool0o`3ooom=0?ooo`800000@P3oool001d0oooo00<000000?ooo`3oool0o`3ooom;0?ooo`800000
A03oool001h0oooo00<000000?ooo`3oool0o`3ooom90?ooo`030000003oool0oooo04@0oooo000N
0?ooo`030000003oool0oooo0?l0ooooA`3oool2000004L0oooo000O0?ooo`030000003oool0oooo
0?l0ooooA03oool2000004T0oooo000P0?ooo`030000003oool0oooo0?l0oooo@@3oool2000004/0
oooo000P0?ooo`030000003oool0oooo0?l0oooo@03oool00`000000oooo0?ooo`1;0?ooo`008@3o
ool00`000000oooo0?ooo`3o0?ooocd0oooo0P00001>0?ooo`008@3oool00`000000oooo0?ooo`3o
0?oooc/0oooo0P00001@0?ooo`008P3oool00`000000oooo0?ooo`3o0?ooocT0oooo00<000000?oo
o`3oool0D03oool00280oooo00<000000?ooo`3oool0o`3ooolg0?ooo`800000D`3oool002<0oooo
00<000000?ooo`3oool0o`3ooold0?ooo`800000E@3oool002@0oooo00<000000?ooo`3oool0o`3o
oola0?ooo`800000E`3oool002@0oooo00<000000?ooo`3oool0o`3oool`0?ooo`030000003oool0
oooo05L0oooo000U0?ooo`030000003oool0oooo0?l0oooo;@3oool2000005X0oooo000U0?ooo`03
0000003oool0oooo0?l0oooo:`3oool2000005`0oooo000V0?ooo`030000003oool0oooo0?l0oooo
:03oool2000005h0oooo000W0?ooo`030000003oool0oooo0?l0oooo9P3oool00`000000oooo0?oo
o`1N0?ooo`009`3oool00`000000oooo0?ooo`3o0?ooob@0oooo0P00001Q0?ooo`00:03oool00`00
0000oooo0?ooo`3o0?ooob40oooo0P00001S0?ooo`00:03oool00`000000oooo0?ooo`3o0?oooal0
oooo0P00001U0?ooo`00:@3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0
I@3oool002T0oooo00<000000?ooo`3oool0o`3ooolK0?ooo`800000J03oool002X0oooo00<00000
0?ooo`3oool0o`3ooolH0?ooo`800000JP3oool002/0oooo00<000000?ooo`3oool0o`3ooolF0?oo
o`030000003oool0oooo06X0oooo000[0?ooo`030000003oool0oooo0?l0oooo503oool2000006d0
oooo000/0?ooo`030000003oool0oooo0?l0oooo4@3oool2000006l0oooo000/0?ooo`030000003o
ool0oooo0?l0oooo3`3oool200000740oooo000]0?ooo`030000003oool0oooo0?l0oooo3@3oool0
0`000000oooo0?ooo`1a0?ooo`00;@3oool00`000000oooo0?ooo`3o0?ooo`/0oooo0P00001d0?oo
o`00;P3oool00`000000oooo0?ooo`3o0?ooo`P0oooo0P00001f0?ooo`00;`3oool00`000000oooo
0?ooo`3o0?ooo`D0oooo0P00001h0?ooo`00;`3oool00`000000oooo0?ooo`3o0?ooo`@0oooo00<0
00000?ooo`3oool0N03oool00300oooo00<000000?ooo`3oool0o`3oool10?ooo`800000N`3oool0
0300oooo00<000000?ooo`3oool0oP3oool2000007d0oooo000a0?ooo`030000003oool0oooo0?/0
oooo0P00001o0?ooo`00<P3oool00`000000oooo0?ooo`3i0?ooo`030000003oool0oooo07l0oooo
000b0?ooo`030000003oool0oooo0?L0oooo0P0000220?ooo`00<`3oool00`000000oooo0?ooo`3d
0?ooo`800000Q03oool003<0oooo00<000000?ooo`3oool0lP3oool2000008H0oooo000d0?ooo`03
0000003oool0oooo0?00oooo00<000000?ooo`3oool0QP3oool003@0oooo00<000000?ooo`3oool0
kP3oool2000008T0oooo000e0?ooo`030000003oool0oooo0>/0oooo0P00002;0?ooo`00=P3oool0
0`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo08/0oooo000f0?ooo`030000003oool0oooo
0>L0oooo0P00002>0?ooo`00=`3oool00`000000oooo0?ooo`3T0?ooo`800000T03oool003L0oooo
00<000000?ooo`3oool0hP3oool200000980oooo000h0?ooo`030000003oool0oooo0>00oooo00<0
00000?ooo`3oool0TP3oool003T0oooo00<000000?ooo`3oool0g@3oool2000009D0oooo000i0?oo
o`030000003oool0oooo0=/0oooo0P00002G0?ooo`00>P3oool00`000000oooo0?ooo`3H0?ooo`80
0000V@3oool003X0oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`2I0?ooo`00>`3o
ool00`000000oooo0?ooo`3D0?ooo`800000W03oool003/0oooo00<000000?ooo`3oool0dP3oool2
000009h0oooo000l0?ooo`030000003oool0oooo0<l0oooo0P00002P0?ooo`00?@3oool00`000000
oooo0?ooo`3=0?ooo`030000003oool0oooo0:00oooo000m0?ooo`030000003oool0oooo0</0oooo
0P00002S0?ooo`00?P3oool00`000000oooo0?ooo`380?ooo`800000Y@3oool003h0oooo00<00000
0?ooo`3oool0a`3oool00`000000oooo0?ooo`2U0?ooo`00?`3oool00`000000oooo0?ooo`340?oo
o`800000Z03oool00400oooo00<000000?ooo`3oool0`@3oool200000:X0oooo00100?ooo`030000
003oool0oooo0;l0oooo0P00002/0?ooo`00@@3oool00`000000oooo0?ooo`2m0?ooo`030000003o
ool0oooo0:`0oooo00110?ooo`030000003oool0oooo0;/0oooo0P00002_0?ooo`00@P3oool00`00
0000oooo0?ooo`2h0?ooo`800000/@3oool00480oooo00<000000?ooo`3oool0]P3oool200000;<0
oooo00130?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0/`3oool004@0oooo00<0
00000?ooo`3oool0/@3oool200000;H0oooo00140?ooo`030000003oool0oooo0:l0oooo0P00002h
0?ooo`00A@3oool00`000000oooo0?ooo`2/0?ooo`800000^P3oool004D0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`2j0?ooo`00AP3oool00`000000oooo0?ooo`2X0?ooo`800000
_@3oool004L0oooo00<000000?ooo`3oool0Y@3oool200000;l0oooo00170?ooo`030000003oool0
oooo0:@0oooo00<000000?ooo`3oool0_`3oool004P0oooo00<000000?ooo`3oool0X@3oool20000
0<80oooo00180?ooo`030000003oool0oooo09l0oooo0P0000340?ooo`00B@3oool00`000000oooo
0?ooo`2L0?ooo`800000aP3oool004T0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?oo
o`360?ooo`00BP3oool00`000000oooo0?ooo`2H0?ooo`800000b@3oool004/0oooo00<000000?oo
o`3oool0U@3oool200000</0oooo001;0?ooo`030000003oool0oooo09<0oooo0P00003=0?ooo`00
C03oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo0<d0oooo001<0?ooo`030000003o
ool0oooo08l0oooo0P00003@0?ooo`00C@3oool00`000000oooo0?ooo`2<0?ooo`800000dP3oool0
04h0oooo00<000000?ooo`3oool0R@3oool200000=@0oooo001>0?ooo`030000003oool0oooo08P0
oooo00<000000?ooo`3oool0e03oool004l0oooo00<000000?ooo`3oool0Q@3oool200000=L0oooo
001?0?ooo`030000003oool0oooo08<0oooo0P00003I0?ooo`00D03oool00`000000oooo0?ooo`20
0?ooo`800000f`3oool00500oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`3K0?oo
o`00D@3oool00`000000oooo0?ooo`1l0?ooo`800000gP3oool00580oooo00<000000?ooo`3oool0
N@3oool200000>00oooo001B0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0h03o
ool005<0oooo00<000000?ooo`3oool0M@3oool200000><0oooo001C0?ooo`030000003oool0oooo
07<0oooo0P00003U0?ooo`00E03oool00`000000oooo0?ooo`1`0?ooo`800000i`3oool005@0oooo
00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`3W0?ooo`00E@3oool00`000000oooo0?oo
o`1/0?ooo`800000jP3oool005H0oooo00<000000?ooo`3oool0J@3oool200000>`0oooo001F0?oo
o`030000003oool0oooo06L0oooo0P00003^0?ooo`00E`3oool00`000000oooo0?ooo`1U0?ooo`03
0000003oool0oooo0>h0oooo001G0?ooo`030000003oool0oooo06<0oooo0P00003a0?ooo`00F03o
ool00`000000oooo0?ooo`1P0?ooo`800000l`3oool005T0oooo00<000000?ooo`3oool0G@3oool2
00000?D0oooo001I0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0m@3oool005X0
oooo00<000000?ooo`3oool0F@3oool200000?P0oooo001J0?ooo`030000003oool0oooo05L0oooo
0P00003j0?ooo`00F`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0?X0oooo001K
0?ooo`030000003oool0oooo05<0oooo0P00003m0?ooo`00G03oool00`000000oooo0?ooo`1@0?oo
o`800000o`3oool005d0oooo00<000000?ooo`3oool0C@3oool200000?l0oooo0P3oool005d0oooo
00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`3o0?ooo`80oooo001N0?ooo`030000003o
ool0oooo04T0oooo0P00003o0?ooo`D0oooo001N0?ooo`030000003oool0oooo04L0oooo0P00003o
0?ooo`L0oooo001O0?ooo`030000003oool0oooo04@0oooo0P00003o0?ooo`T0oooo001P0?ooo`03
0000003oool0oooo0480oooo00<000000?ooo`3oool0o`3oool90?ooo`00H03oool00`000000oooo
0?ooo`100?ooo`800000o`3oool<0?ooo`00H@3oool00`000000oooo0?ooo`0m0?ooo`800000o`3o
ool>0?ooo`00H@3oool00`000000oooo0?ooo`0k0?ooo`800000o`3oool@0?ooo`00HP3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo0?l0oooo403oool00680oooo00<000000?ooo`3o
ool0=`3oool200000?l0oooo4`3oool006<0oooo00<000000?ooo`3oool0=03oool200000?l0oooo
5@3oool006@0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`3o0?oooaD0oooo001T
0?ooo`030000003oool0oooo0300oooo0P00003o0?oooaP0oooo001U0?ooo`030000003oool0oooo
02d0oooo0P00003o0?oooaX0oooo001U0?ooo`030000003oool0oooo02/0oooo0P00003o0?oooa`0
oooo001V0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0o`3ooolL0?ooo`00I`3o
ool00`000000oooo0?ooo`0V0?ooo`800000o`3ooolO0?ooo`00I`3oool00`000000oooo0?ooo`0T
0?ooo`800000o`3ooolQ0?ooo`00J03oool00`000000oooo0?ooo`0Q0?ooo`800000o`3ooolS0?oo
o`00J03oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0?l0oooo8`3oool006T0oooo
00<000000?ooo`3oool07@3oool200000?l0oooo9P3oool006T0oooo00<000000?ooo`3oool06`3o
ool200000?l0oooo:03oool006X0oooo00<000000?ooo`3oool0603oool200000?l0oooo:P3oool0
06/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?ooobX0oooo001[0?ooo`03
0000003oool0oooo01@0oooo0P00003o0?ooobd0oooo001/0?ooo`030000003oool0oooo0140oooo
0P00003o0?ooobl0oooo001/0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0o`3o
ool_0?ooo`00K@3oool00`000000oooo0000000=0?ooo`800000o`3ooolb0?ooo`00K@3oool30000
00/0oooo0P00003o0?oooc@0oooo001[0?ooo`H00000203oool200000?l0oooo=P3oool006`0oooo
1@0000070?ooo`030000003oool0oooo0?l0oooo=P3oool006d0oooo100000050?ooo`800000o`3o
ooli0?ooo`00KP3oool3000000<0oooo0P00003o0?oooc/0oooo001`0?ooo`@00000o`3ooolm0?oo
o`00L@3oool00`000000oooo0?ooo`3o0?ooocd0oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {186, 0}} -> {-0.0267759, -0.0108276, \
0.00366665, 0.00366665}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[\(\(p3 . p4 //. momentumComponentRules\) /. sol\) //. 
        values]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(\(VelocityPlot2[LabVelocities, vCM, cmVelocities];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[RGBColor[0, 1, 0]], plot1, 
        Graphics[Circle[vCM, magnitude[vi]\/2]], Graphics[RGBColor[1, 0, 0]], 
        plot2, Graphics[Circle[vCM, magnitude[vf]\/2]], 
        AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.5 0.952381 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 1 0 r
.5 Mabswid
[ ] 0 setdash
.02381 .5 m
.97619 .5 L
s
% Start of user PostScript
0. 0. 1. 0. MAarrowhead1
% End of user PostScript
.02381 .5 m
.02381 .5 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.02381 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.97619 .5 L
s
% Start of user PostScript
0.5 0. 1. 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.02381 .5 L
s
% Start of user PostScript
0.5 0. 0. 0. MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
1 0 0 r
.02381 .5 m
.2619 .91239 L
s
% Start of user PostScript
0. 0. 0.25 0.4330127 MAarrowhead1
% End of user PostScript
.02381 .5 m
.7381 .08761 L
s
% Start of user PostScript
0. 0. 0.75 -0.4330127 MAarrowhead1
% End of user PostScript
.02381 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.2619 .91239 L
s
% Start of user PostScript
0.5 0. 0.25 0.4330127 MAarrowhead1
% End of user PostScript
.5 .5 m
.7381 .08761 L
s
% Start of user PostScript
0.5 0. 0.75 -0.4330127 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ0?l00<P0oooo002m0?ooo`X0o`008@3oool:0?l00;h0oooo002e0?ooo`P0o`00
=@3oool80?l00;H0oooo002`0?ooo`D0o`00A@3oool50?l00;40oooo002[0?ooo`D0o`00C`3oool5
0?l00:`0oooo002V0?ooo`D0o`00F@3oool20?l00:X0oooo002T0?ooo`80o`00H03oool50?l00:D0
oooo002O0?ooo`D0o`00I`3oool50?l00:00oooo002M0?ooo`80o`00L@3oool30?l009d0oooo002J
0?ooo`<0o`00MP3oool20?l009/0oooo002E0?ooo`D0o`00N`3oool50?l009H0oooo002C0?ooo`80
o`00Q@3oool20?l009@0oooo002@0?ooo`<0o`00R@3oool30?l00940oooo002>0?ooo`80o`00S`3o
ool20?l008l0oooo002<0?ooo`80o`00T`3oool20?l008d0oooo00290?ooo`<0o`00U`3oool30?l0
08X0oooo00270?ooo`80o`00W@3oool20?l008P0oooo00250?ooo`80o`00X@3oool20?l008H0oooo
00220?ooo`<0o`00Y@3oool30?l008<0oooo00200?ooo`80o`00Z`3oool20?l00840oooo001n0?oo
o`80o`00[`3oool20?l007l0oooo001k0?ooo`<0o`00/`3oool30?l007`0oooo001i0?ooo`80o`00
^@3oool20?l007X0oooo001h0?ooo`030?l0003oool0oooo0;/0oooo00<0o`000?ooo`3oool0M`3o
ool007L0oooo00<0o`000?ooo`3oool0_@3oool00`3o0000oooo0?ooo`1f0?ooo`00M@3oool20?l0
0<40oooo0P3o001f0?ooo`00LP3oool30?l00<D0oooo0`3o001c0?ooo`00L03oool20?l00</0oooo
0P3o001a0?ooo`00K`3oool00`3o0000oooo0?ooo`370?ooo`L0o`00L03oool006h0oooo00<0o`00
0?ooo`3oool0a@3oool80?l00080oooo00<0o`000?ooo`3oool0K@3oool006`0oooo0P3o00380?oo
o`P0o`000`3oool20?l006d0oooo001Z0?ooo`80o`00b`3oool60?l000H0oooo0P3o001[0?ooo`00
J@3oool00`3o0000oooo0?ooo`390?ooo`P0o`00203oool00`3o0000oooo0?ooo`1X0?ooo`00J03o
ool00`3o0000oooo0?ooo`380?ooo`80o`000`3oool50?l000T0oooo0P3o001X0?ooo`00I@3oool3
0?l00<X0oooo00<0o`000?ooo`3oool0103oool30?l000`0oooo0P3o001V0?ooo`00H`3oool20?l0
0</0oooo0P3o00070?ooo`030?l0003oool0o`0000h0oooo0P3o001T0?ooo`00HP3oool00`3o0000
oooo0?ooo`390?ooo`80o`002@3oool00`3o0000oooo0?ooo`0@0?ooo`030?l0003oool0oooo0640
oooo001Q0?ooo`030?l0003oool0oooo0<P0oooo0P3o000:0?ooo`030?l0003oool0oooo0180oooo
00<0o`000?ooo`3oool0H03oool005l0oooo0P3o003:0?ooo`030?l0003oool0oooo00X0oooo00<0
o`000?ooo`3oool04`3oool20?l00600oooo001N0?ooo`030?l0003oool0oooo0<P0oooo0P3o000<
0?ooo`030?l0003oool0oooo01H0oooo00<0o`000?ooo`3oool0G@3oool005d0oooo00<0o`000?oo
o`3oool0a`3oool20?l000d0oooo00<0o`000?ooo`3oool0603oool00`3o0000oooo0?ooo`1L0?oo
o`00F`3oool20?l00<P0oooo0P3o000?0?ooo`030?l0003oool0oooo01T0oooo0P3o001L0?ooo`00
FP3oool00`3o0000oooo0?ooo`370?ooo`030?l0003oool0oooo00h0oooo00<0o`000?ooo`3oool0
703oool00`3o0000oooo0?ooo`1I0?ooo`00F@3oool00`3o0000oooo0?ooo`360?ooo`80o`004@3o
ool00`3o0000oooo0?ooo`0M0?ooo`030?l0003oool0oooo05P0oooo001H0?ooo`030?l0003oool0
oooo0<D0oooo0P3o000B0?ooo`030?l0003oool0oooo01l0oooo00<0o`000?ooo`3oool0E`3oool0
05L0oooo00<0o`000?ooo`3oool0a03oool20?l001@0oooo00<0o`000?ooo`3oool0803oool00`3o
0000oooo0?ooo`1F0?ooo`00E@3oool20?l00<H0oooo00<0o`000?ooo`3oool04`3oool00`3o0000
oooo0?ooo`0R0?ooo`80o`00EP3oool005@0oooo00<0o`000?ooo`3oool0a03oool20?l001D0oooo
00<0o`000?ooo`3oool09@3oool00`3o0000oooo0?ooo`1C0?ooo`00D`3oool00`3o0000oooo0?oo
o`330?ooo`80o`005`3oool00`3o0000oooo0?ooo`0V0?ooo`030?l0003oool0oooo0580oooo001A
0?ooo`80o`00a@3oool00`3o0000oooo0?ooo`0F0?ooo`030?l0003oool0oooo02P0oooo0P3o001B
0?ooo`00D03oool00`3o0000oooo0?ooo`330?ooo`80o`006@3oool00`3o0000oooo0?ooo`0Z0?oo
o`030?l0003oool0oooo04l0oooo001?0?ooo`030?l0003oool0oooo0<80oooo0P3o000J0?ooo`03
0?l0003oool0oooo02`0oooo00<0o`000?ooo`3oool0CP3oool004h0oooo00<0o`000?ooo`3oool0
`@3oool20?l001`0oooo00<0o`000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`1=0?ooo`00C@3o
ool00`3o0000oooo0?ooo`310?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0;`3o
ool00`3o0000oooo0?ooo`1<0?ooo`00C03oool00`3o0000oooo0?ooo`300?ooo`80o`007@3oool0
0`3o0000oooo0?ooo`0a0?ooo`030?l0003oool0oooo04/0oooo001;0?ooo`030?l0003oool0oooo
0;l0oooo0P3o000O0?ooo`030?l0003oool0oooo0380oooo00<0o`000?ooo`3oool0BP3oool004X0
oooo00<0o`000?ooo`3oool0_P3oool20?l00200oooo00<0o`000?ooo`3oool0=03oool00`3o0000
oooo0?ooo`190?ooo`00B@3oool00`3o0000oooo0?ooo`2n0?ooo`030?l0003oool0oooo0200oooo
00<0o`000?ooo`3oool0=@3oool00`3o0000oooo0?ooo`180?ooo`00B03oool00`3o0000oooo0?oo
o`2m0?ooo`80o`008P3oool00`3o0000oooo0?ooo`0g0?ooo`030?l0003oool0oooo04L0oooo0017
0?ooo`030?l0003oool0oooo0;`0oooo0P3o000S0?ooo`030?l0003oool0oooo03T0oooo00<0o`00
0?ooo`3oool0AP3oool004H0oooo00<0o`000?ooo`3oool0_03oool00`3o0000oooo0?ooo`0S0?oo
o`030?l0003oool0oooo03X0oooo00<0o`000?ooo`3oool0A@3oool004D0oooo00<0o`000?ooo`3o
ool0^`3oool20?l002D0oooo00<0o`000?ooo`3oool0?03oool00`3o0000oooo0?ooo`140?ooo`00
A03oool00`3o0000oooo0?ooo`2j0?ooo`80o`009`3oool00`3o0000oooo0?ooo`0m0?ooo`030?l0
003oool0oooo04<0oooo00130?ooo`030?l0003oool0oooo0;T0oooo0P3o000X0?ooo`030?l0003o
ool0oooo03l0oooo00<0o`000?ooo`3oool0@P3oool00480oooo00<0o`000?ooo`3oool0^@3oool0
0`3o0000oooo0?ooo`0X0?ooo`030?l0003oool0oooo0400oooo00<0o`000?ooo`3oool0@@3oool0
0440oooo00<0o`000?ooo`3oool0^03oool20?l002X0oooo00<0o`000?ooo`3oool0@P3oool00`3o
0000oooo0?ooo`100?ooo`00@03oool00`3o0000oooo0?ooo`2g0?ooo`80o`00:`3oool00`3o0000
oooo0?ooo`140?ooo`030?l0003oool0oooo03l0oooo000o0?ooo`030?l0003oool0oooo0;H0oooo
0P3o000]0?ooo`030?l0003oool0oooo04D0oooo00<0o`000?ooo`3oool0?P3oool003h0oooo00<0
o`000?ooo`3oool0]P3oool00`3o0000oooo0?ooo`0/0?ooo`030?l0003oool0oooo04L0oooo00<0
o`000?ooo`3oool0?@3oool003d0oooo00<0o`000?ooo`3oool0]@3oool20?l002l0oooo00<0o`00
0?ooo`3oool0B03oool00`3o0000oooo0?ooo`0l0?ooo`00?03oool00`3o0000oooo0?ooo`2d0?oo
o`80o`00<03oool00`3o0000oooo0?ooo`1:0?ooo`030?l0003oool0oooo03/0oooo000k0?ooo`03
0?l0003oool0oooo0;@0oooo00<0o`000?ooo`3oool0<03oool00`3o0000oooo0?ooo`1;0?ooo`03
0?l0003oool0oooo03X0oooo000j0?ooo`030?l0003oool0oooo0;<0oooo0P3o000b0?ooo`030?l0
003oool0oooo04`0oooo00<0o`000?ooo`3oool0>P3oool003X0oooo00<0o`000?ooo`3oool0/@3o
ool20?l003<0oooo00<0o`000?ooo`3oool0CP3oool00`3o0000oooo0?ooo`0i0?ooo`00>@3oool0
0`3o0000oooo0?ooo`2`0?ooo`80o`00=@3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0
oooo03P0oooo000h0?ooo`030?l0003oool0oooo0;00oooo00<0o`000?ooo`3oool0=03oool00`3o
0000oooo0?ooo`1A0?ooo`030?l0003oool0oooo03L0oooo000g0?ooo`030?l0003oool0oooo0:l0
oooo0P3o000g0?ooo`030?l0003oool0oooo0540oooo00<0o`000?ooo`3oool0=`3oool003L0oooo
00<0o`000?ooo`3oool0[@3oool20?l003P0oooo00<0o`000?ooo`3oool0D`3oool00`3o0000oooo
0?ooo`0f0?ooo`00=P3oool00`3o0000oooo0?ooo`2/0?ooo`80o`00>@3oool00`3o0000oooo0?oo
o`1E0?ooo`030?l0003oool0oooo03D0oooo000e0?ooo`030?l0003oool0oooo0:`0oooo00<0o`00
0?ooo`3oool0>@3oool00`3o0000oooo0?ooo`1F0?ooo`030?l0003oool0oooo03@0oooo000d0?oo
o`030?l0003oool0oooo0:/0oooo0P3o000k0?ooo`030?l0003oool0oooo05P0oooo00<0o`000?oo
o`3oool0<`3oool003<0oooo00<0o`000?ooo`3oool0ZP3oool20?l003d0oooo00<0o`000?ooo`3o
ool0F@3oool00`3o0000oooo0?ooo`0b0?ooo`00<P3oool00`3o0000oooo0?ooo`2Y0?ooo`80o`00
?P3oool00`3o0000oooo0?ooo`1J0?ooo`030?l0003oool0oooo0380oooo000b0?ooo`030?l0003o
ool0oooo0:P0oooo00<0o`000?ooo`3oool0?P3oool00`3o0000oooo0?ooo`1K0?ooo`030?l0003o
ool0oooo0340oooo000a0?ooo`030?l0003oool0oooo0:L0oooo0P3o00100?ooo`030?l0003oool0
oooo05d0oooo00<0o`000?ooo`3oool0<03oool00300oooo00<0o`000?ooo`3oool0YP3oool20?l0
0440oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`0_0?ooo`00;`3oool00`3o0000
oooo0?ooo`2V0?ooo`030?l0003oool0oooo0440oooo00<0o`000?ooo`3oool0G`3oool00`3o0000
oooo0?ooo`0_0?ooo`00;`3oool00`3o0000oooo0?ooo`2T0?ooo`80o`00@`3oool00`3o0000oooo
0?ooo`1Q0?ooo`030?l0003oool0oooo02h0oooo000^0?ooo`030?l0003oool0oooo0:<0oooo0P3o
00150?ooo`030?l0003oool0oooo0680oooo00<0o`000?ooo`3oool0;@3oool002d0oooo00<0o`00
0?ooo`3oool0XP3oool20?l004H0oooo00<0o`000?ooo`3oool0I03oool00`3o0000oooo0?ooo`0/
0?ooo`00;03oool00`3o0000oooo0?ooo`2R0?ooo`030?l0003oool0oooo04H0oooo00<0o`000?oo
o`3oool0I03oool00`3o0000oooo0?ooo`0/0?ooo`00;03oool00`3o0000oooo0?ooo`2P0?ooo`80
o`00B03oool00`3o0000oooo0?ooo`1V0?ooo`030?l0003oool0oooo02/0oooo000/0?ooo`030?l0
003oool0oooo09h0oooo0P3o00190?ooo`030?l0003oool0oooo06L0oooo00<0o`000?ooo`3oool0
:`3oool002/0oooo00<0o`000?ooo`3oool0W@3oool20?l004/0oooo00<0o`000?ooo`3oool0I`3o
ool00`3o0000oooo0?ooo`0[0?ooo`00:`3oool00`3o0000oooo0?ooo`2L0?ooo`030?l0003oool0
oooo04X0oooo00<0o`000?ooo`3oool0J@3oool00`3o0000oooo0?ooo`0Z0?ooo`00:P3oool00`3o
0000oooo0?ooo`2K0?ooo`80o`00C@3oool00`3o0000oooo0?ooo`1Z0?ooo`030?l0003oool0oooo
02T0oooo000Y0?ooo`030?l0003oool0oooo09X0oooo0P3o001>0?ooo`030?l0003oool0oooo06`0
oooo00<0o`000?ooo`3oool0:03oool002P0oooo00<0o`000?ooo`3oool0VP3oool00`3o0000oooo
0?ooo`1=0?ooo`030?l0003oool0oooo06d0oooo00<0o`000?ooo`3oool0:03oool002P0oooo00<0
o`000?ooo`3oool0V03oool20?l00500oooo00<0o`000?ooo`3oool0KP3oool00`3o0000oooo0?oo
o`0W0?ooo`009`3oool00`3o0000oooo0?ooo`2G0?ooo`80o`00D@3oool00`3o0000oooo0?ooo`1_
0?ooo`030?l0003oool0oooo02L0oooo000W0?ooo`030?l0003oool0oooo09D0oooo0P3o001C0?oo
o`030?l0003oool0oooo0700oooo00<0o`000?ooo`3oool09P3oool002H0oooo00<0o`000?ooo`3o
ool0U@3oool00`3o0000oooo0?ooo`1B0?ooo`030?l0003oool0oooo0780oooo00<0o`000?ooo`3o
ool09@3oool002D0oooo00<0o`000?ooo`3oool0U03oool20?l005D0oooo00<0o`000?ooo`3oool0
L`3oool00`3o0000oooo0?ooo`0T0?ooo`009@3oool00`3o0000oooo0?ooo`2B0?ooo`80o`00EP3o
ool00`3o0000oooo0?ooo`1d0?ooo`030?l0003oool0oooo02@0oooo000T0?ooo`030?l0003oool0
oooo0940oooo0P3o001G0?ooo`030?l0003oool0oooo07H0oooo00<0o`000?ooo`3oool08`3oool0
02@0oooo00<0o`000?ooo`3oool0T03oool00`3o0000oooo0?ooo`1G0?ooo`030?l0003oool0oooo
07H0oooo00<0o`000?ooo`3oool08`3oool002<0oooo00<0o`000?ooo`3oool0S`3oool20?l005T0
oooo00<0o`000?ooo`3oool0M`3oool00`3o0000oooo0?ooo`0S0?ooo`008`3oool00`3o0000oooo
0?ooo`2=0?ooo`80o`00F`3oool00`3o0000oooo0?ooo`1h0?ooo`030?l0003oool0oooo0280oooo
000R0?ooo`030?l0003oool0oooo08`0oooo0P3o001L0?ooo`030?l0003oool0oooo07T0oooo00<0
o`000?ooo`3oool08P3oool00280oooo00<0o`000?ooo`3oool0R`3oool00`3o0000oooo0?ooo`1L
0?ooo`030?l0003oool0oooo07X0oooo00<0o`000?ooo`3oool08@3oool00240oooo00<0o`000?oo
o`3oool0RP3oool20?l005h0oooo00<0o`000?ooo`3oool0O03oool00`3o0000oooo0?ooo`0P0?oo
o`00803oool00`3o0000oooo0?ooo`290?ooo`80o`00G`3oool00`3o0000oooo0?ooo`1n0?ooo`03
0?l0003oool0oooo01l0oooo000P0?ooo`030?l0003oool0oooo08P0oooo00<0o`000?ooo`3oool0
G`3oool00`3o0000oooo0?ooo`1n0?ooo`030?l0003oool0oooo01l0oooo000O0?ooo`030?l0003o
ool0oooo08L0oooo0P3o001Q0?ooo`030?l0003oool0oooo0800oooo00<0o`000?ooo`3oool07P3o
ool001l0oooo00<0o`000?ooo`3oool0Q@3oool20?l006<0oooo00<0o`000?ooo`3oool0P03oool0
0`3o0000oooo0?ooo`0N0?ooo`007P3oool00`3o0000oooo0?ooo`240?ooo`80o`00I03oool00`3o
0000oooo0?ooo`210?ooo`030?l0003oool0oooo01h0oooo000N0?ooo`030?l0003oool0oooo08<0
oooo00<0o`000?ooo`3oool0H`3oool00`3o0000oooo0?ooo`230?ooo`030?l0003oool0oooo01d0
oooo000M0?ooo`030?l0003oool0oooo0880oooo0P3o001V0?ooo`030?l0003oool0oooo08<0oooo
00<0o`000?ooo`3oool07@3oool001d0oooo00<0o`000?ooo`3oool0P03oool20?l006L0oooo00<0
o`000?ooo`3oool0Q@3oool00`3o0000oooo0?ooo`0L0?ooo`007@3oool00`3o0000oooo0?ooo`1n
0?ooo`80o`00J@3oool00`3o0000oooo0?ooo`250?ooo`030?l0003oool0oooo01`0oooo000L0?oo
o`030?l0003oool0oooo07h0oooo00<0o`000?ooo`3oool0J03oool00`3o0000oooo0?ooo`270?oo
o`030?l0003oool0oooo01/0oooo000L0?ooo`030?l0003oool0oooo07`0oooo0P3o001[0?ooo`03
0?l0003oool0oooo08L0oooo00<0o`000?ooo`3oool06`3oool001/0oooo00<0o`000?ooo`3oool0
N`3oool20?l006`0oooo00<0o`000?ooo`3oool0R03oool00`3o0000oooo0?ooo`0K0?ooo`006`3o
ool00`3o0000oooo0?ooo`1j0?ooo`030?l0003oool0oooo06/0oooo00<0o`000?ooo`3oool0RP3o
ool00`3o0000oooo0?ooo`0J0?ooo`006P3oool00`3o0000oooo0?ooo`1i0?ooo`80o`00KP3oool0
0`3o0000oooo0?ooo`2:0?ooo`030?l0003oool0oooo01X0oooo000J0?ooo`030?l0003oool0oooo
07L0oooo0P3o001_0?ooo`030?l0003oool0oooo08`0oooo00<0o`000?ooo`3oool06@3oool001X0
oooo00<0o`000?ooo`3oool0M@3oool20?l00740oooo00<0o`000?ooo`3oool0S03oool00`3o0000
oooo0?ooo`0I0?ooo`006@3oool00`3o0000oooo0?ooo`1e0?ooo`030?l0003oool0oooo0700oooo
00<0o`000?ooo`3oool0SP3oool00`3o0000oooo0?ooo`0H0?ooo`006@3oool00`3o0000oooo0?oo
o`1c0?ooo`80o`00L`3oool00`3o0000oooo0?ooo`2>0?ooo`030?l0003oool0oooo01P0oooo000H
0?ooo`030?l0003oool0oooo0780oooo0P3o001d0?ooo`030?l0003oool0oooo08l0oooo00<0o`00
0?ooo`3oool0603oool001P0oooo00<0o`000?ooo`3oool0L03oool20?l007D0oooo00<0o`000?oo
o`3oool0T@3oool00`3o0000oooo0?ooo`0G0?ooo`005`3oool00`3o0000oooo0?ooo`1`0?ooo`03
0?l0003oool0oooo07D0oooo00<0o`000?ooo`3oool0T@3oool00`3o0000oooo0?ooo`0G0?ooo`00
5`3oool00`3o0000oooo0?ooo`1^0?ooo`80o`00M`3oool00`3o0000oooo0?ooo`2C0?ooo`030?l0
003oool0oooo01H0oooo000G0?ooo`030?l0003oool0oooo06`0oooo0P3o001i0?ooo`030?l0003o
ool0oooo09<0oooo00<0o`000?ooo`3oool05P3oool001H0oooo00<0o`000?ooo`3oool0J`3oool2
0?l007X0oooo00<0o`000?ooo`3oool0U@3oool00`3o0000oooo0?ooo`0E0?ooo`005P3oool00`3o
0000oooo0?ooo`1Z0?ooo`030?l0003oool0oooo07X0oooo00<0o`000?ooo`3oool0U@3oool00`3o
0000oooo0?ooo`0E0?ooo`005@3oool00`3o0000oooo0?ooo`1Y0?ooo`80o`00O03oool00`3o0000
oooo0?ooo`2F0?ooo`030?l0003oool0oooo01D0oooo000E0?ooo`030?l0003oool0oooo06L0oooo
0P3o001m0?ooo`030?l0003oool0oooo09P0oooo00<0o`000?ooo`3oool0503oool001D0oooo00<0
o`000?ooo`3oool0IP3oool00`3o0000oooo0?ooo`1m0?ooo`030?l0003oool0oooo09P0oooo00<0
o`000?ooo`3oool0503oool001@0oooo00<0o`000?ooo`3oool0I@3oool20?l007l0oooo00<0o`00
0?ooo`3oool0VP3oool00`3o0000oooo0?ooo`0C0?ooo`00503oool00`3o0000oooo0?ooo`1S0?oo
o`80o`00P@3oool00`3o0000oooo0?ooo`2J0?ooo`030?l0003oool0oooo01<0oooo000D0?ooo`03
0?l0003oool0oooo0640oooo0P3o00220?ooo`030?l0003oool0oooo09/0oooo00<0o`000?ooo`3o
ool04`3oool001@0oooo00<0o`000?ooo`3oool0H03oool00`3o0000oooo0?ooo`210?ooo`030?l0
003oool0oooo09`0oooo00<0o`000?ooo`3oool04`3oool001<0oooo00<0o`000?ooo`3oool0G`3o
ool20?l008@0oooo00<0o`000?ooo`3oool0W03oool00`3o0000oooo0?ooo`0C0?ooo`004`3oool0
0`3o0000oooo0?ooo`1M0?ooo`80o`00Q@3oool00`3o0000oooo0?ooo`2N0?ooo`030?l0003oool0
oooo0180oooo000C0?ooo`030?l0003oool0oooo05/0oooo0P3o00270?ooo`030?l0003oool0oooo
09h0oooo00<0o`000?ooo`3oool04P3oool00180oooo00<0o`000?ooo`3oool0F`3oool00`3o0000
oooo0?ooo`260?ooo`030?l0003oool0oooo09l0oooo00<0o`000?ooo`3oool04P3oool00180oooo
00<0o`000?ooo`3oool0F@3oool20?l008T0oooo00<0o`000?ooo`3oool0X03oool00`3o0000oooo
0?ooo`0A0?ooo`004@3oool00`3o0000oooo0?ooo`1H0?ooo`80o`00RP3oool00`3o0000oooo0?oo
o`2Q0?ooo`030?l0003oool0oooo0140oooo000A0?ooo`030?l0003oool0oooo05L0oooo00<0o`00
0?ooo`3oool0R@3oool00`3o0000oooo0?ooo`2S0?ooo`030?l0003oool0oooo0100oooo000A0?oo
o`030?l0003oool0oooo05D0oooo0P3o002<0?ooo`030?l0003oool0oooo0:<0oooo00<0o`000?oo
o`3oool0403oool00140oooo00<0o`000?ooo`3oool0D`3oool20?l008d0oooo00<0o`000?ooo`3o
ool0Y03oool00`3o0000oooo0?ooo`0@0?ooo`004@3oool00`3o0000oooo0?ooo`1A0?ooo`80o`00
S`3oool00`3o0000oooo0?ooo`2T0?ooo`030?l0003oool0oooo0100oooo000@0?ooo`030?l0003o
ool0oooo0540oooo00<0o`000?ooo`3oool0SP3oool00`3o0000oooo0?ooo`2U0?ooo`030?l0003o
ool0oooo0100oooo000@0?ooo`030?l0003oool0oooo04l0oooo0P3o002A0?ooo`030?l0003oool0
oooo0:H0oooo00<0o`000?ooo`3oool03`3oool00100oooo00<0o`000?ooo`3oool0C@3oool20?l0
0980oooo00<0o`000?ooo`3oool0Y`3oool00`3o0000oooo0?ooo`0?0?ooo`00403oool00`3o0000
oooo0?ooo`1;0?ooo`80o`00T`3oool00`3o0000oooo0?ooo`2Y0?ooo`030?l0003oool0oooo00h0
oooo000@0?ooo`030?l0003oool0oooo04X0oooo00<0o`000?ooo`3oool0T`3oool00`3o0000oooo
0?ooo`2Y0?ooo`030?l0003oool0oooo00h0oooo000?0?ooo`030?l0003oool0oooo04T0oooo0P3o
002E0?ooo`030?l0003oool0oooo0:X0oooo00<0o`000?ooo`3oool03P3oool000l0oooo00<0o`00
0?ooo`3oool0A`3oool20?l009L0oooo00<0o`000?ooo`3oool0ZP3oool00`3o0000oooo0?ooo`0>
0?ooo`003`3oool00`3o0000oooo0?ooo`160?ooo`030?l0003oool0oooo09H0oooo00<0o`000?oo
o`3oool0Z`3oool00`3o0000oooo0?ooo`0>0?ooo`003P3oool00`3o0000oooo0?ooo`150?ooo`80
o`00V03oool00`3o0000oooo0?ooo`2]0?ooo`030?l0003oool0oooo00d0oooo000>0?ooo`030?l0
003oool0oooo04<0oooo0P3o002J0?ooo`030?l0003oool0oooo0:d0oooo00<0o`000?ooo`3oool0
3@3oool000h0oooo00<0o`000?ooo`3oool0@@3oool20?l009/0oooo00<0o`000?ooo`3oool0[P3o
ool00`3o0000oooo0?ooo`0=0?ooo`003P3oool00`3o0000oooo0?ooo`100?ooo`030?l0003oool0
oooo09/0oooo00<0o`000?ooo`3oool0[P3oool00`3o0000oooo0?ooo`0=0?ooo`003P3oool00`3o
0000oooo0?ooo`0n0?ooo`80o`00W@3oool00`3o0000oooo0?ooo`2_0?ooo`030?l0003oool0oooo
00d0oooo000=0?ooo`030?l0003oool0oooo03d0oooo0P3o002O0?ooo`030?l0003oool0oooo0;00
oooo00<0o`000?ooo`3oool0303oool000d0oooo00<0o`000?ooo`3oool0>`3oool20?l00:00oooo
00<0o`000?ooo`3oool0/@3oool00`3o0000oooo0?ooo`0<0?ooo`003@3oool00`3o0000oooo0?oo
o`0j0?ooo`030?l0003oool0oooo09l0oooo00<0o`000?ooo`3oool0/P3oool00`3o0000oooo0?oo
o`0<0?ooo`003@3oool00`3o0000oooo0?ooo`0h0?ooo`80o`00XP3oool00`3o0000oooo0?ooo`2b
0?ooo`030?l0003oool0oooo00`0oooo000=0?ooo`030?l0003oool0oooo03H0oooo0P3o002S0?oo
o`030?l0003oool0oooo0;<0oooo00<0o`000?ooo`3oool0303oool000`0oooo00<0o`000?ooo`3o
ool0=@3oool20?l00:D0oooo00<0o`000?ooo`3oool0]03oool00`3o0000oooo0?ooo`0;0?ooo`00
303oool00`3o0000oooo0?ooo`0d0?ooo`030?l0003oool0oooo0:@0oooo00<0o`000?ooo`3oool0
]@3oool00`3o0000oooo0?ooo`0;0?ooo`00303oool00`3o0000oooo0?ooo`0b0?ooo`80o`00Y`3o
ool00`3o0000oooo0?ooo`2e0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0
oooo0300oooo0P3o002X0?ooo`030?l0003oool0oooo0;H0oooo00<0o`000?ooo`3oool02`3oool0
00`0oooo00<0o`000?ooo`3oool0;`3oool00`3o0000oooo0?ooo`2W0?ooo`030?l0003oool0oooo
0;L0oooo00<0o`000?ooo`3oool02`3oool000`0oooo00<0o`000?ooo`3oool0;@3oool20?l00:X0
oooo00<0o`000?ooo`3oool0]`3oool00`3o0000oooo0?ooo`0;0?ooo`00303oool00`3o0000oooo
0?ooo`0[0?ooo`80o`00Z`3oool00`3o0000oooo0?ooo`2h0?ooo`030?l0003oool0oooo00/0oooo
000;0?ooo`030?l0003oool0oooo02X0oooo0P3o002]0?ooo`030?l0003oool0oooo0;P0oooo00<0
o`000?ooo`3oool02`3oool000/0oooo00<0o`000?ooo`3oool0:@3oool00`3o0000oooo0?ooo`2/
0?ooo`030?l0003oool0oooo0;X0oooo00<0o`000?ooo`3oool02P3oool000/0oooo00<0o`000?oo
o`3oool09`3oool20?l00:h0oooo00<0o`000?ooo`3oool0^`3oool00`3o0000oooo0?ooo`0:0?oo
o`002`3oool00`3o0000oooo0?ooo`0U0?ooo`80o`00/03oool00`3o0000oooo0?ooo`2k0?ooo`03
0?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo02<0oooo0P3o002a0?ooo`030?l0
003oool0oooo0;`0oooo00<0o`000?ooo`3oool02P3oool000/0oooo00<0o`000?ooo`3oool08P3o
ool00`3o0000oooo0?ooo`2a0?ooo`030?l0003oool0oooo0;`0oooo00<0o`000?ooo`3oool02P3o
ool000/0oooo00<0o`000?ooo`3oool0803oool20?l00;<0oooo00<0o`000?ooo`3oool0_@3oool0
0`3o0000oooo0?ooo`0:0?ooo`002`3oool00`3o0000oooo0?ooo`0N0?ooo`80o`00]@3oool00`3o
0000oooo0?ooo`2m0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo01d0
oooo00<0o`000?ooo`3oool0]03oool00`3o0000oooo0?ooo`2n0?ooo`030?l0003oool0oooo00X0
oooo000;0?ooo`030?l0003oool0oooo01/0oooo0P3o002f0?ooo`030?l0003oool0oooo0;l0oooo
00<0o`000?ooo`3oool02P3oool000/0oooo00<0o`000?ooo`3oool06@3oool20?l00;P0oooo00<0
o`000?ooo`3oool0_`3oool00`3o0000oooo0?ooo`0:0?ooo`002P3oool00`3o0000oooo0?ooo`0H
0?ooo`80o`00^@3oool00`3o0000oooo0?ooo`310?ooo`030?l0003oool0oooo00T0oooo000:0?oo
o`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool0^@3oool00`3o0000oooo0?ooo`310?oo
o`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo01D0oooo0P3o002k0?ooo`03
0?l0003oool0oooo0<80oooo00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3oool0
4`3oool20?l00;d0oooo00<0o`000?ooo`3oool0`P3oool00`3o0000oooo0?ooo`090?ooo`002P3o
ool00`3o0000oooo0?ooo`0A0?ooo`80o`00_P3oool00`3o0000oooo0?ooo`330?ooo`030?l0003o
ool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0100oooo00<0o`000?ooo`3oool0_@3oool0
0`3o0000oooo0?ooo`340?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo
00h0oooo0P3o00300?ooo`030?l0003oool0oooo0<@0oooo00<0o`000?ooo`3oool02@3oool000X0
oooo00<0o`000?ooo`3oool0303oool20?l00<40oooo00<0o`000?ooo`3oool0a@3oool00`3o0000
oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0:0?ooo`80o`00``3oool00`3o0000oooo
0?ooo`350?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo00T0oooo00<0
o`000?ooo`3oool0`P3oool00`3o0000oooo0?ooo`360?ooo`030?l0003oool0oooo00T0oooo000:
0?ooo`030?l0003oool0oooo00L0oooo0P3o00340?ooo`030?l0003oool0oooo0<L0oooo00<0o`00
0?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3oool01@3oool20?l00<H0oooo00<0o`000?oo
o`3oool0a`3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`040?ooo`03
0?l0003oool0oooo0<D0oooo00<0o`000?ooo`3oool0b03oool00`3o0000oooo0?ooo`090?ooo`00
2P3oool00`3o0000oooo0?ooo`020?ooo`80o`00b03oool00`3o0000oooo0?ooo`380?ooo`030?l0
003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0080o`0000<0oooo003o0000o`00_P3o
ool20?l000H0oooo00<0o`000?ooo`3oool0`P3oool2003o00D0oooo00<0o`000?ooo`3oool02@3o
ool000X0oooo0`3o0005003o0;h0oooo1P3o00020?ooo`030?l0003oool0oooo0<80oooo1P00o`00
0`3oool0o`000?ooo`0:0?ooo`002P3oooo>0?l00<`00?l000<0o`000?ooo`3oool02@3oool000X0
oooo0P3o0006003o0;h0oooo1`3o00360?ooo`H00?l000<0oooo0?l0003oool02P3oool000X0oooo
0P3o00040?ooo`800?l0_P3oool20?l000@0oooo00<0o`000?ooo`3oool0a03oool2003o00D0oooo
00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3o0000b03oool00`3o0000oooo0?oo
o`3<0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0o`000<P0oooo00<0o`00
0?ooo`3oool0c03oool00`3o0000oooo0?ooo`090?ooo`002P3oool0103o0000oooo0?ooo`3o0036
0?ooo`030?l0003oool0oooo0<d0oooo00<0o`000?ooo`3oool02@3oool000X0oooo00@0o`000?oo
o`3oool0o`00aP3oool00`3o0000oooo0?ooo`3=0?ooo`030?l0003oool0oooo00T0oooo000:0?oo
o`050?l0003oool0oooo0?ooo`3o0000a03oool00`3o0000oooo0?ooo`3>0?ooo`030?l0003oool0
oooo00T0oooo000:0?ooo`030?l0003oool0oooo0080oooo00<0o`000?ooo`3oool0`03oool00`3o
0000oooo0?ooo`3?0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0080
oooo00<0o`000?ooo`3oool0`03oool00`3o0000oooo0?ooo`3?0?ooo`030?l0003oool0oooo00T0
oooo000:0?ooo`030?l0003oool0oooo00<0oooo00<0o`000?ooo`3oool0_P3oool00`3o0000oooo
0?ooo`3@0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo00<0oooo00<0
o`000?ooo`3oool0_P3oool00`3o0000oooo0?ooo`3@0?ooo`030?l0003oool0oooo00T0oooo000:
0?ooo`030?l0003oool0oooo00@0oooo00<0o`000?ooo`3oool0_03oool00`3o0000oooo0?ooo`3A
0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo00@0oooo00<0o`000?oo
o`3oool0^`3oool00`3o0000oooo0?ooo`3B0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`03
0?l0003oool0oooo00D0oooo00<0o`000?ooo`3oool0^P3oool00`3o0000oooo0?ooo`3B0?ooo`03
0?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo00H0oooo00<0o`000?ooo`3oool0
^03oool00`3o0000oooo0?ooo`3C0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003o
ool0oooo00H0oooo00<0o`000?ooo`3oool0^03oool00`3o0000oooo0?ooo`3C0?ooo`030?l0003o
ool0oooo00T0oooo000;0?ooo`030?l0003oool0oooo00H0oooo00<0o`000?ooo`3oool0]P3oool0
0`3o0000oooo0?ooo`3C0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo
00H0oooo00<0o`000?ooo`3oool0]P3oool00`3o0000oooo0?ooo`3C0?ooo`030?l0003oool0oooo
00X0oooo000;0?ooo`030?l0003oool0oooo00L0oooo00<0o`000?ooo`3oool0]03oool00`3o0000
oooo0?ooo`3D0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo00L0oooo
00<0o`000?ooo`3oool0/`3oool00`3o0000oooo0?ooo`3E0?ooo`030?l0003oool0oooo00X0oooo
000;0?ooo`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool0/P3oool00`3o0000oooo0?oo
o`3E0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo00T0oooo00<0o`00
0?ooo`3oool0/03oool00`3o0000oooo0?ooo`3F0?ooo`030?l0003oool0oooo00X0oooo000;0?oo
o`030?l0003oool0oooo00T0oooo00<0o`000?ooo`3oool0/03oool00`3o0000oooo0?ooo`3F0?oo
o`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo00X0oooo00<0o`000?ooo`3o
ool0[P3oool00`3o0000oooo0?ooo`3G0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0
003oool0oooo00X0oooo00<0o`000?ooo`3oool0[P3oool00`3o0000oooo0?ooo`3G0?ooo`030?l0
003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo00/0oooo00<0o`000?ooo`3oool0[03o
ool00`3o0000oooo0?ooo`3H0?ooo`030?l0003oool0oooo00X0oooo000<0?ooo`030?l0003oool0
oooo00X0oooo00<0o`000?ooo`3oool0Z`3oool00`3o0000oooo0?ooo`3H0?ooo`030?l0003oool0
oooo00/0oooo000<0?ooo`030?l0003oool0oooo00/0oooo00<0o`000?ooo`3oool0ZP3oool00`3o
0000oooo0?ooo`3H0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0oooo00`0
oooo00<0o`000?ooo`3oool0Z03oool00`3o0000oooo0?ooo`3I0?ooo`030?l0003oool0oooo00/0
oooo000<0?ooo`030?l0003oool0oooo00`0oooo00<0o`000?ooo`3oool0Z03oool00`3o0000oooo
0?ooo`3I0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0oooo00d0oooo00<0
o`000?ooo`3oool0YP3oool00`3o0000oooo0?ooo`3J0?ooo`030?l0003oool0oooo00/0oooo000<
0?ooo`030?l0003oool0oooo00d0oooo00<0o`000?ooo`3oool0Y@3oool00`3o0000oooo0?ooo`3K
0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0oooo00h0oooo00<0o`000?oo
o`3oool0Y03oool00`3o0000oooo0?ooo`3K0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`03
0?l0003oool0oooo00h0oooo00<0o`000?ooo`3oool0X`3oool00`3o0000oooo0?ooo`3L0?ooo`03
0?l0003oool0oooo00/0oooo000=0?ooo`030?l0003oool0oooo00h0oooo00<0o`000?ooo`3oool0
XP3oool00`3o0000oooo0?ooo`3K0?ooo`030?l0003oool0oooo00`0oooo000=0?ooo`030?l0003o
ool0oooo00l0oooo00<0o`000?ooo`3oool0X03oool00`3o0000oooo0?ooo`3L0?ooo`030?l0003o
ool0oooo00`0oooo000=0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3oool0X03oool0
0`3o0000oooo0?ooo`3L0?ooo`030?l0003oool0oooo00`0oooo000=0?ooo`030?l0003oool0oooo
0100oooo00<0o`000?ooo`3oool0WP3oool00`3o0000oooo0?ooo`3M0?ooo`030?l0003oool0oooo
00`0oooo000=0?ooo`030?l0003oool0oooo0100oooo00<0o`000?ooo`3oool0W@3oool00`3o0000
oooo0?ooo`3N0?ooo`030?l0003oool0oooo00`0oooo000>0?ooo`030?l0003oool0oooo0100oooo
00<0o`000?ooo`3oool0W03oool00`3o0000oooo0?ooo`3M0?ooo`030?l0003oool0oooo00d0oooo
000>0?ooo`030?l0003oool0oooo0100oooo00<0o`000?ooo`3oool0V`3oool00`3o0000oooo0?oo
o`3N0?ooo`030?l0003oool0oooo00d0oooo000>0?ooo`030?l0003oool0oooo0140oooo00<0o`00
0?ooo`3oool0VP3oool00`3o0000oooo0?ooo`3N0?ooo`030?l0003oool0oooo00d0oooo000>0?oo
o`030?l0003oool0oooo0180oooo00<0o`000?ooo`3oool0V03oool00`3o0000oooo0?ooo`3O0?oo
o`030?l0003oool0oooo00d0oooo000>0?ooo`030?l0003oool0oooo0180oooo00<0o`000?ooo`3o
ool0V03oool00`3o0000oooo0?ooo`3O0?ooo`030?l0003oool0oooo00d0oooo000?0?ooo`030?l0
003oool0oooo0180oooo00<0o`000?ooo`3oool0UP3oool00`3o0000oooo0?ooo`3O0?ooo`030?l0
003oool0oooo00h0oooo000?0?ooo`030?l0003oool0oooo0180oooo00<0o`000?ooo`3oool0U@3o
ool00`3o0000oooo0?ooo`3P0?ooo`030?l0003oool0oooo00h0oooo000?0?ooo`030?l0003oool0
oooo01<0oooo00<0o`000?ooo`3oool0U03oool00`3o0000oooo0?ooo`3P0?ooo`030?l0003oool0
oooo00h0oooo000?0?ooo`030?l0003oool0oooo01@0oooo00<0o`000?ooo`3oool0TP3oool00`3o
0000oooo0?ooo`3Q0?ooo`030?l0003oool0oooo00h0oooo000?0?ooo`030?l0003oool0oooo01@0
oooo00<0o`000?ooo`3oool0TP3oool00`3o0000oooo0?ooo`3Q0?ooo`030?l0003oool0oooo00h0
oooo000@0?ooo`030?l0003oool0oooo01@0oooo00<0o`000?ooo`3oool0T03oool00`3o0000oooo
0?ooo`3Q0?ooo`030?l0003oool0oooo00l0oooo000@0?ooo`030?l0003oool0oooo01@0oooo00<0
o`000?ooo`3oool0S`3oool00`3o0000oooo0?ooo`3R0?ooo`030?l0003oool0oooo00l0oooo000@
0?ooo`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool0SP3oool00`3o0000oooo0?ooo`3Q
0?ooo`030?l0003oool0oooo0100oooo000A0?ooo`030?l0003oool0oooo01@0oooo00<0o`000?oo
o`3oool0S@3oool00`3o0000oooo0?ooo`3R0?ooo`030?l0003oool0oooo0100oooo000A0?ooo`03
0?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool0S03oool00`3o0000oooo0?ooo`3R0?ooo`03
0?l0003oool0oooo0100oooo000A0?ooo`030?l0003oool0oooo01H0oooo00<0o`000?ooo`3oool0
RP3oool00`3o0000oooo0?ooo`3S0?ooo`030?l0003oool0oooo0100oooo000A0?ooo`030?l0003o
ool0oooo01H0oooo00<0o`000?ooo`3oool0RP3oool00`3o0000oooo0?ooo`3S0?ooo`030?l0003o
ool0oooo0100oooo000A0?ooo`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool0R03oool0
0`3o0000oooo0?ooo`3S0?ooo`030?l0003oool0oooo0140oooo000B0?ooo`030?l0003oool0oooo
01H0oooo00<0o`000?ooo`3oool0Q`3oool00`3o0000oooo0?ooo`3T0?ooo`030?l0003oool0oooo
0140oooo000B0?ooo`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool0QP3oool00`3o0000
oooo0?ooo`3S0?ooo`030?l0003oool0oooo0180oooo000C0?ooo`030?l0003oool0oooo01H0oooo
00<0o`000?ooo`3oool0Q@3oool00`3o0000oooo0?ooo`3T0?ooo`030?l0003oool0oooo0180oooo
000C0?ooo`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool0Q03oool00`3o0000oooo0?oo
o`3T0?ooo`030?l0003oool0oooo0180oooo000D0?ooo`030?l0003oool0oooo01L0oooo00<0o`00
0?ooo`3oool0PP3oool00`3o0000oooo0?ooo`3T0?ooo`030?l0003oool0oooo01<0oooo000D0?oo
o`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool0PP3oool00`3o0000oooo0?ooo`3T0?oo
o`030?l0003oool0oooo01<0oooo000D0?ooo`030?l0003oool0oooo01P0oooo00<0o`000?ooo`3o
ool0P03oool00`3o0000oooo0?ooo`3U0?ooo`030?l0003oool0oooo01<0oooo000D0?ooo`030?l0
003oool0oooo01P0oooo00<0o`000?ooo`3oool0O`3oool00`3o0000oooo0?ooo`3V0?ooo`030?l0
003oool0oooo01<0oooo000D0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0OP3o
ool00`3o0000oooo0?ooo`3V0?ooo`030?l0003oool0oooo01<0oooo000E0?ooo`030?l0003oool0
oooo01P0oooo00<0o`000?ooo`3oool0O@3oool00`3o0000oooo0?ooo`3V0?ooo`030?l0003oool0
oooo01@0oooo000E0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0O03oool00`3o
0000oooo0?ooo`3V0?ooo`030?l0003oool0oooo01@0oooo000E0?ooo`030?l0003oool0oooo01X0
oooo00<0o`000?ooo`3oool0NP3oool00`3o0000oooo0?ooo`3V0?ooo`030?l0003oool0oooo01D0
oooo000F0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0N@3oool00`3o0000oooo
0?ooo`3W0?ooo`030?l0003oool0oooo01D0oooo000F0?ooo`030?l0003oool0oooo01X0oooo00<0
o`000?ooo`3oool0N03oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003oool0oooo01D0oooo000G
0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0M`3oool00`3o0000oooo0?ooo`3W
0?ooo`030?l0003oool0oooo01H0oooo000G0?ooo`030?l0003oool0oooo01X0oooo00<0o`000?oo
o`3oool0MP3oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003oool0oooo01H0oooo000G0?ooo`03
0?l0003oool0oooo01X0oooo00<0o`000?ooo`3oool0M@3oool00`3o0000oooo0?ooo`3W0?ooo`03
0?l0003oool0oooo01L0oooo000H0?ooo`030?l0003oool0oooo01X0oooo00<0o`000?ooo`3oool0
M03oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003oool0oooo01L0oooo000H0?ooo`030?l0003o
ool0oooo01/0oooo00<0o`000?ooo`3oool0LP3oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003o
ool0oooo01P0oooo000I0?ooo`030?l0003oool0oooo01X0oooo00<0o`000?ooo`3oool0L@3oool0
0`3o0000oooo0?ooo`3X0?ooo`030?l0003oool0oooo01P0oooo000I0?ooo`030?l0003oool0oooo
01/0oooo00<0o`000?ooo`3oool0L03oool00`3o0000oooo0?ooo`3X0?ooo`030?l0003oool0oooo
01P0oooo000J0?ooo`030?l0003oool0oooo01X0oooo00<0o`000?ooo`3oool0K`3oool00`3o0000
oooo0?ooo`3X0?ooo`030?l0003oool0oooo01T0oooo000J0?ooo`030?l0003oool0oooo01/0oooo
00<0o`000?ooo`3oool0KP3oool00`3o0000oooo0?ooo`3X0?ooo`030?l0003oool0oooo01T0oooo
000J0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0K@3oool00`3o0000oooo0?oo
o`3X0?ooo`030?l0003oool0oooo01X0oooo000K0?ooo`030?l0003oool0oooo01/0oooo00<0o`00
0?ooo`3oool0K03oool00`3o0000oooo0?ooo`3X0?ooo`030?l0003oool0oooo01X0oooo000K0?oo
o`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0JP3oool00`3o0000oooo0?ooo`3X0?oo
o`030?l0003oool0oooo01/0oooo000L0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3o
ool0J@3oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo01/0oooo000L0?ooo`030?l0
003oool0oooo01`0oooo00<0o`000?ooo`3oool0J03oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0
003oool0oooo01/0oooo000M0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0I`3o
ool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo01`0oooo000M0?ooo`030?l0003oool0
oooo01`0oooo00<0o`000?ooo`3oool0IP3oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0
oooo01`0oooo000M0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0I@3oool00`3o
0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo01d0oooo000N0?ooo`030?l0003oool0oooo01`0
oooo00<0o`000?ooo`3oool0H`3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo01d0
oooo000N0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0HP3oool00`3o0000oooo
0?ooo`3Y0?ooo`030?l0003oool0oooo01h0oooo000O0?ooo`030?l0003oool0oooo01`0oooo00<0
o`000?ooo`3oool0H@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo01h0oooo000O
0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0H03oool00`3o0000oooo0?ooo`3Z
0?ooo`030?l0003oool0oooo01h0oooo000P0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?oo
o`3oool0G`3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo01l0oooo000P0?ooo`03
0?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0GP3oool00`3o0000oooo0?ooo`3Z0?ooo`03
0?l0003oool0oooo01l0oooo000Q0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0
G@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo0200oooo000R0?ooo`030?l0003o
ool0oooo01`0oooo00<0o`000?ooo`3oool0F`3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003o
ool0oooo0240oooo000R0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0FP3oool0
0`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo0280oooo000S0?ooo`030?l0003oool0oooo
01`0oooo00<0o`000?ooo`3oool0F@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo
0280oooo000S0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0F03oool00`3o0000
oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02<0oooo000T0?ooo`030?l0003oool0oooo01`0oooo
00<0o`000?ooo`3oool0E`3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo02<0oooo
000T0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0EP3oool00`3o0000oooo0?oo
o`3Z0?ooo`030?l0003oool0oooo02<0oooo000U0?ooo`030?l0003oool0oooo01`0oooo00<0o`00
0?ooo`3oool0E@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo02@0oooo000U0?oo
o`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0D`3oool00`3o0000oooo0?ooo`3[0?oo
o`030?l0003oool0oooo02@0oooo000V0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3o
ool0DP3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo02D0oooo000W0?ooo`030?l0
003oool0oooo01`0oooo00<0o`000?ooo`3oool0D@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0
003oool0oooo02H0oooo000W0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0D03o
ool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02L0oooo000X0?ooo`030?l0003oool0
oooo01`0oooo00<0o`000?ooo`3oool0C`3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0
oooo02L0oooo000X0?ooo`030?l0003oool0oooo01d0oooo00<0o`000?ooo`3oool0CP3oool00`3o
0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02P0oooo000Y0?ooo`030?l0003oool0oooo01`0
oooo00<0o`000?ooo`3oool0C@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo02P0
oooo000Z0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0B`3oool00`3o0000oooo
0?ooo`3Z0?ooo`030?l0003oool0oooo02T0oooo000[0?ooo`030?l0003oool0oooo01`0oooo00<0
o`000?ooo`3oool0BP3oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02X0oooo000[
0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0B@3oool00`3o0000oooo0?ooo`3Y
0?ooo`030?l0003oool0oooo02/0oooo000/0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?oo
o`3oool0B03oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02/0oooo000/0?ooo`03
0?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0A`3oool00`3o0000oooo0?ooo`3Z0?ooo`03
0?l0003oool0oooo02/0oooo000]0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0
A@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo02`0oooo000]0?ooo`030?l0003o
ool0oooo01`0oooo00<0o`000?ooo`3oool0A@3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003o
ool0oooo02`0oooo000^0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0@`3oool0
0`3o0000oooo0?ooo`3Z0?ooo`030?l0003oool0oooo02d0oooo000_0?ooo`030?l0003oool0oooo
01`0oooo00<0o`000?ooo`3oool0@P3oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo
02h0oooo000_0?ooo`030?l0003oool0oooo01`0oooo00<0o`000?ooo`3oool0@@3oool00`3o0000
oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02l0oooo000`0?ooo`030?l0003oool0oooo01`0oooo
00<0o`000?ooo`3oool0@03oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo02l0oooo
000a0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0?`3oool00`3o0000oooo0?oo
o`3Y0?ooo`030?l0003oool0oooo0300oooo000b0?ooo`030?l0003oool0oooo01/0oooo00<0o`00
0?ooo`3oool0?@3oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo0340oooo000b0?oo
o`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0?@3oool00`3o0000oooo0?ooo`3X0?oo
o`030?l0003oool0oooo0380oooo000c0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3o
ool0>`3oool00`3o0000oooo0?ooo`3Y0?ooo`030?l0003oool0oooo0380oooo000d0?ooo`030?l0
003oool0oooo01/0oooo00<0o`000?ooo`3oool0>P3oool00`3o0000oooo0?ooo`3X0?ooo`030?l0
003oool0oooo03<0oooo000e0?ooo`030?l0003oool0oooo01X0oooo00<0o`000?ooo`3oool0>@3o
ool00`3o0000oooo0?ooo`3X0?ooo`030?l0003oool0oooo03@0oooo000f0?ooo`030?l0003oool0
oooo01X0oooo00<0o`000?ooo`3oool0>03oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003oool0
oooo03D0oooo000g0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0=`3oool00`3o
0000oooo0?ooo`3W0?ooo`030?l0003oool0oooo03H0oooo000g0?ooo`030?l0003oool0oooo01X0
oooo00<0o`000?ooo`3oool0=@3oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003oool0oooo03L0
oooo000h0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0=@3oool00`3o0000oooo
0?ooo`3W0?ooo`030?l0003oool0oooo03L0oooo000i0?ooo`030?l0003oool0oooo01T0oooo00<0
o`000?ooo`3oool0<`3oool00`3o0000oooo0?ooo`3W0?ooo`030?l0003oool0oooo03P0oooo000j
0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0<P3oool00`3o0000oooo0?ooo`3V
0?ooo`030?l0003oool0oooo03T0oooo000j0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?oo
o`3oool0<@3oool00`3o0000oooo0?ooo`3V0?ooo`030?l0003oool0oooo03X0oooo000k0?ooo`03
0?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0;`3oool00`3o0000oooo0?ooo`3W0?ooo`03
0?l0003oool0oooo03X0oooo000l0?ooo`030?l0003oool0oooo01P0oooo00<0o`000?ooo`3oool0
;`3oool00`3o0000oooo0?ooo`3V0?ooo`030?l0003oool0oooo03/0oooo000m0?ooo`030?l0003o
ool0oooo01P0oooo00<0o`000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`3V0?ooo`030?l0003o
ool0oooo03`0oooo000n0?ooo`030?l0003oool0oooo01P0oooo00<0o`000?ooo`3oool0;03oool0
0`3o0000oooo0?ooo`3U0?ooo`030?l0003oool0oooo03d0oooo000o0?ooo`030?l0003oool0oooo
01L0oooo00<0o`000?ooo`3oool0:`3oool00`3o0000oooo0?ooo`3U0?ooo`030?l0003oool0oooo
03h0oooo00100?ooo`030?l0003oool0oooo01L0oooo00<0o`000?ooo`3oool0:P3oool00`3o0000
oooo0?ooo`3T0?ooo`030?l0003oool0oooo03l0oooo00110?ooo`030?l0003oool0oooo01H0oooo
00<0o`000?ooo`3oool0:@3oool00`3o0000oooo0?ooo`3T0?ooo`030?l0003oool0oooo0400oooo
00120?ooo`030?l0003oool0oooo01H0oooo00<0o`000?ooo`3oool09`3oool00`3o0000oooo0?oo
o`3T0?ooo`030?l0003oool0oooo0440oooo00130?ooo`030?l0003oool0oooo01D0oooo00<0o`00
0?ooo`3oool09`3oool00`3o0000oooo0?ooo`3S0?ooo`030?l0003oool0oooo0480oooo00140?oo
o`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool09@3oool00`3o0000oooo0?ooo`3S0?oo
o`030?l0003oool0oooo04<0oooo00150?ooo`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3o
ool0903oool00`3o0000oooo0?ooo`3R0?ooo`030?l0003oool0oooo04@0oooo00160?ooo`030?l0
003oool0oooo01@0oooo00<0o`000?ooo`3oool08`3oool00`3o0000oooo0?ooo`3R0?ooo`030?l0
003oool0oooo04D0oooo00170?ooo`030?l0003oool0oooo01@0oooo00<0o`000?ooo`3oool08P3o
ool00`3o0000oooo0?ooo`3Q0?ooo`030?l0003oool0oooo04H0oooo00180?ooo`030?l0003oool0
oooo01<0oooo00<0o`000?ooo`3oool08@3oool00`3o0000oooo0?ooo`3Q0?ooo`030?l0003oool0
oooo04L0oooo00190?ooo`030?l0003oool0oooo01<0oooo00<0o`000?ooo`3oool07`3oool00`3o
0000oooo0?ooo`3Q0?ooo`030?l0003oool0oooo04P0oooo001:0?ooo`030?l0003oool0oooo0180
oooo00<0o`000?ooo`3oool07`3oool00`3o0000oooo0?ooo`3P0?ooo`030?l0003oool0oooo04T0
oooo001;0?ooo`030?l0003oool0oooo0180oooo00<0o`000?ooo`3oool07@3oool00`3o0000oooo
0?ooo`3P0?ooo`030?l0003oool0oooo04X0oooo001<0?ooo`030?l0003oool0oooo0180oooo00<0
o`000?ooo`3oool0703oool00`3o0000oooo0?ooo`3O0?ooo`030?l0003oool0oooo04/0oooo001=
0?ooo`030?l0003oool0oooo0140oooo00<0o`000?ooo`3oool06`3oool00`3o0000oooo0?ooo`3O
0?ooo`030?l0003oool0oooo04`0oooo001>0?ooo`030?l0003oool0oooo0140oooo00<0o`000?oo
o`3oool06@3oool00`3o0000oooo0?ooo`3O0?ooo`030?l0003oool0oooo04d0oooo001?0?ooo`03
0?l0003oool0oooo0100oooo00<0o`000?ooo`3oool06@3oool00`3o0000oooo0?ooo`3N0?ooo`03
0?l0003oool0oooo04h0oooo001@0?ooo`030?l0003oool0oooo0100oooo00<0o`000?ooo`3oool0
5`3oool00`3o0000oooo0?ooo`3N0?ooo`030?l0003oool0oooo04l0oooo001A0?ooo`030?l0003o
ool0oooo00l0oooo00<0o`000?ooo`3oool05`3oool00`3o0000oooo0?ooo`3M0?ooo`030?l0003o
ool0oooo0500oooo001B0?ooo`80o`00403oool00`3o0000oooo0?ooo`0E0?ooo`030?l0003oool0
oooo0=`0oooo0P3o001C0?ooo`00E03oool00`3o0000oooo0?ooo`0>0?ooo`030?l0003oool0oooo
01@0oooo00<0o`000?ooo`3oool0f`3oool00`3o0000oooo0?ooo`1C0?ooo`00E@3oool00`3o0000
oooo0?ooo`0=0?ooo`030?l0003oool0oooo01<0oooo00<0o`000?ooo`3oool0f`3oool00`3o0000
oooo0?ooo`1D0?ooo`00EP3oool20?l000h0oooo00<0o`000?ooo`3oool04@3oool00`3o0000oooo
0?ooo`3J0?ooo`80o`00E`3oool005P0oooo00<0o`000?ooo`3oool02`3oool00`3o0000oooo0?oo
o`0A0?ooo`030?l0003oool0oooo0=T0oooo00<0o`000?ooo`3oool0E`3oool005T0oooo00<0o`00
0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0?0?ooo`030?l0003oool0oooo0=T0oooo00<0o`00
0?ooo`3oool0F03oool005X0oooo00<0o`000?ooo`3oool02P3oool00`3o0000oooo0?ooo`0?0?oo
o`030?l0003oool0oooo0=P0oooo00<0o`000?ooo`3oool0F@3oool005/0oooo00<0o`000?ooo`3o
ool02P3oool00`3o0000oooo0?ooo`0=0?ooo`030?l0003oool0oooo0=P0oooo00<0o`000?ooo`3o
ool0FP3oool005`0oooo0P3o000;0?ooo`030?l0003oool0oooo00`0oooo00<0o`000?ooo`3oool0
eP3oool20?l005d0oooo001N0?ooo`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`3F0?ooo`030?l0003oool0oooo05d0oooo001O0?ooo`030?l0003oool0
oooo00P0oooo00<0o`000?ooo`3oool02@3oool00`3o0000oooo0?ooo`3F0?ooo`030?l0003oool0
oooo05h0oooo001P0?ooo`80o`00203oool00`3o0000oooo0?ooo`090?ooo`030?l0003oool0oooo
0=@0oooo0P3o001Q0?ooo`00HP3oool00`3o0000oooo0?ooo`060?ooo`030?l0003oool0oooo00L0
oooo00<0o`000?ooo`3oool0e03oool00`3o0000oooo0?ooo`1Q0?ooo`00H`3oool00`3o0000oooo
0?ooo`050?ooo`040?l0003oool0oooo0?l000@0oooo00<0o`000?ooo`3o0000e03oool20?l006@0
oooo001T0?ooo`<0o`001@3oool30?l000@0oooo0`3o003B0?ooo`80o`00IP3oool006L0oooo0P3o
00020?ooo`@0o`000`3oool50?l00<l0oooo0P3o001X0?ooo`00J@3oool00`3o0000oooo0?ooo`04
0?l00080oooo103o003?0?ooo`030?l0003oool0oooo06P0oooo001Z0?ooo`030?l0003oool0oooo
00<0o`0000<0oooo0?l0003o00000P3o003?0?ooo`030?l0003oool0oooo06T0oooo001[0?ooo`D0
o`0000@0oooo0?l0003o0000o`00cP3oool20?l006`0oooo001]0?ooo`<0o`0000<0oooo0?l0003o
0000c@3oool20?l006h0oooo001_0?ooo`<0o`00c@3oool00`3o0000oooo0?ooo`1^0?ooo`00L03o
ool20?l00</0oooo0P3o001a0?ooo`00LP3oool20?l00<L0oooo0P3o001c0?ooo`00M03oool20?l0
0<<0oooo0P3o001e0?ooo`00MP3oool20?l00;l0oooo0P3o001g0?ooo`00N03oool00`3o0000oooo
0?ooo`2k0?ooo`030?l0003oool0oooo07L0oooo001i0?ooo`80o`00^@3oool20?l007X0oooo001k
0?ooo`80o`00]@3oool20?l007`0oooo001m0?ooo`80o`00/@3oool20?l007h0oooo001o0?ooo`<0
o`00Z`3oool30?l00800oooo00220?ooo`80o`00Y`3oool20?l008<0oooo00240?ooo`80o`00X`3o
ool20?l008D0oooo00260?ooo`<0o`00W@3oool30?l008L0oooo00290?ooo`80o`00V@3oool20?l0
08X0oooo002;0?ooo`80o`00U@3oool20?l008`0oooo002=0?ooo`<0o`00S`3oool30?l008h0oooo
002@0?ooo`80o`00R`3oool20?l00940oooo002B0?ooo`<0o`00Q@3oool30?l009<0oooo002E0?oo
o`@0o`00O03oool50?l009H0oooo002I0?ooo`<0o`00M`3oool20?l009/0oooo002L0?ooo`80o`00
L`3oool20?l009d0oooo002N0?ooo`D0o`00J@3oool50?l009l0oooo002S0?ooo`D0o`00H@3oool3
0?l00:@0oooo002X0?ooo`<0o`00F@3oool50?l00:L0oooo002[0?ooo`D0o`00C`3oool50?l00:`0
oooo002`0?ooo`D0o`00A@3oool50?l00;40oooo002e0?ooo`L0o`00=`3oool70?l00;H0oooo002l
0?ooo`/0o`008@3oool;0?l00;d0oooo00370?ooob40o`00b03oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.0250052, -0.525005, \
0.00365433, 0.00365433}}]
}, Closed]],

Cell[TextData[{
  "It should be clear from this plot that the laboratory scattering angle is \
limited to ",
  Cell[BoxData[
      \(TraditionalForm\`90  \[Degree]\)]],
  ", where the struck particle carries the full momentum and the projectile \
ends up at rest."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotFinalMomenta[sol, values, 90. ];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.010582 0.0147151 0.588604 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.23545 -0.0125 -8 -15.6875 ]
[.23545 -0.0125 8 0 ]
[.44709 -0.0125 -8 -15.6875 ]
[.44709 -0.0125 8 0 ]
[.65873 -0.0125 -8 -15.6875 ]
[.65873 -0.0125 8 0 ]
[.87037 -0.0125 -8 -15.6875 ]
[.87037 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.0313 -31.375 ]
[.5 -0.0125 21.0313 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .13244 -19 -7.84375 ]
[-0.0125 .13244 0 7.84375 ]
[-0.0125 .25016 -19 -7.84375 ]
[-0.0125 .25016 0 7.84375 ]
[-0.0125 .36788 -19 -7.84375 ]
[-0.0125 .36788 0 7.84375 ]
[-0.0125 .4856 -19 -7.84375 ]
[-0.0125 .4856 0 7.84375 ]
[-0.0125 .60332 -10 -7.84375 ]
[-0.0125 .60332 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.6875 -30.4375 ]
[-0.01875 .30902 -19 30.4375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -72.125 0 ]
[.5 .63053 72.125 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.23545 0 m
.23545 .61803 L
s
.44709 0 m
.44709 .61803 L
s
.65873 0 m
.65873 .61803 L
s
.87037 0 m
.87037 .61803 L
s
0 .13244 m
1 .13244 L
s
0 .25016 m
1 .25016 L
s
0 .36788 m
1 .36788 L
s
0 .4856 m
1 .4856 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.23545 0 m
.23545 .00625 L
s
gsave
.23545 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.44709 0 m
.44709 .00625 L
s
gsave
.44709 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.65873 0 m
.65873 .00625 L
s
gsave
.65873 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.87037 0 m
.87037 .00625 L
s
gsave
.87037 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.125 Mabswid
.07672 0 m
.07672 .00375 L
s
.12963 0 m
.12963 .00375 L
s
.18254 0 m
.18254 .00375 L
s
.28836 0 m
.28836 .00375 L
s
.34127 0 m
.34127 .00375 L
s
.39418 0 m
.39418 .00375 L
s
.5 0 m
.5 .00375 L
s
.55291 0 m
.55291 .00375 L
s
.60582 0 m
.60582 .00375 L
s
.71164 0 m
.71164 .00375 L
s
.76455 0 m
.76455 .00375 L
s
.81746 0 m
.81746 .00375 L
s
.92328 0 m
.92328 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.0313 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .13244 m
.00625 .13244 L
s
gsave
-0.0125 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .25016 m
.00625 .25016 L
s
gsave
-0.0125 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .36788 m
.00625 .36788 L
s
gsave
-0.0125 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .4856 m
.00625 .4856 L
s
gsave
-0.0125 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .60332 m
.00625 .60332 L
s
gsave
-0.0125 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04415 m
.00375 .04415 L
s
0 .07358 m
.00375 .07358 L
s
0 .10301 m
.00375 .10301 L
s
0 .16187 m
.00375 .16187 L
s
0 .1913 m
.00375 .1913 L
s
0 .22073 m
.00375 .22073 L
s
0 .27959 m
.00375 .27959 L
s
0 .30902 m
.00375 .30902 L
s
0 .33845 m
.00375 .33845 L
s
0 .39731 m
.00375 .39731 L
s
0 .42674 m
.00375 .42674 L
s
0 .45617 m
.00375 .45617 L
s
0 .51503 m
.00375 .51503 L
s
0 .54446 m
.00375 .54446 L
s
0 .57389 m
.00375 .57389 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -118.281 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
91.4375 11.8438 translate
90 rotate
-91.4375 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
69.000 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
73.813 15.813 moveto
(\\220) show
77.625 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
83.625 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
88.438 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(,) show
94.438 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
100.438 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
105.250 15.813 moveto
(\\220) show
109.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
115.063 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
119.875 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.23545 .61178 m
.23545 .61803 L
s
.44709 .61178 m
.44709 .61803 L
s
.65873 .61178 m
.65873 .61803 L
s
.87037 .61178 m
.87037 .61803 L
s
.125 Mabswid
.07672 .61428 m
.07672 .61803 L
s
.12963 .61428 m
.12963 .61803 L
s
.18254 .61428 m
.18254 .61803 L
s
.28836 .61428 m
.28836 .61803 L
s
.34127 .61428 m
.34127 .61803 L
s
.39418 .61428 m
.39418 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.55291 .61428 m
.55291 .61803 L
s
.60582 .61428 m
.60582 .61803 L
s
.71164 .61428 m
.71164 .61803 L
s
.76455 .61428 m
.76455 .61803 L
s
.81746 .61428 m
.81746 .61803 L
s
.92328 .61428 m
.92328 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .13244 m
1 .13244 L
s
.99375 .25016 m
1 .25016 L
s
.99375 .36788 m
1 .36788 L
s
.99375 .4856 m
1 .4856 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .04415 m
1 .04415 L
s
.99625 .07358 m
1 .07358 L
s
.99625 .10301 m
1 .10301 L
s
.99625 .16187 m
1 .16187 L
s
.99625 .1913 m
1 .1913 L
s
.99625 .22073 m
1 .22073 L
s
.99625 .27959 m
1 .27959 L
s
.99625 .30902 m
1 .30902 L
s
.99625 .33845 m
1 .33845 L
s
.99625 .39731 m
1 .39731 L
s
.99625 .42674 m
1 .42674 L
s
.99625 .45617 m
1 .45617 L
s
.99625 .51503 m
1 .51503 L
s
.99625 .54446 m
1 .54446 L
s
.99625 .57389 m
1 .57389 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -133.125 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(scattered) show
108.250 15.813 moveto
(and) show
128.500 15.813 moveto
(recoil) show
158.563 15.813 moveto
(momenta) show
203.250 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60332 L
.02605 .60331 L
.02729 .60331 L
.02846 .6033 L
.03053 .60328 L
.03279 .60325 L
.03527 .60321 L
.0379 .60316 L
.04262 .60304 L
.04749 .60287 L
.05205 .60268 L
.06244 .60212 L
.07305 .60138 L
.08274 .60054 L
.10458 .5981 L
.12357 .59537 L
.14429 .59174 L
.18493 .58266 L
.22406 .5715 L
.26565 .55711 L
.30571 .54083 L
.34426 .523 L
.38527 .50178 L
.42475 .47924 L
.46273 .45571 L
.50315 .42875 L
.54206 .40107 L
.58342 .3699 L
.62326 .33831 L
.66159 .30661 L
.70238 .27159 L
.74164 .23678 L
.77939 .20243 L
.8196 .16505 L
.85828 .12846 L
.89942 .08904 L
.93905 .05075 L
.97619 .01472 L
s
.02381 .01472 m
.06244 .0522 L
.10458 .0929 L
.14415 .13078 L
.18221 .16675 L
.22272 .20437 L
.26171 .23979 L
.30316 .27641 L
.34309 .31054 L
.3815 .34217 L
.42237 .37437 L
.46172 .40383 L
.49955 .43061 L
.53984 .45735 L
.57861 .48124 L
.61984 .50455 L
.65954 .52486 L
.69774 .54233 L
.73838 .55862 L
.77751 .572 L
.81909 .58367 L
.85916 .59239 L
.87754 .59554 L
.89771 .59839 L
.91765 .60058 L
.92854 .6015 L
.93871 .60219 L
.94354 .60247 L
.94878 .60272 L
.95341 .6029 L
.95832 .60306 L
.96273 .60317 L
.96524 .60322 L
.9676 .60326 L
.96976 .60329 L
.97173 .6033 L
.97277 .60331 L
.97391 .60331 L
.97509 .60332 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00k03oool01@3OW4P00000000000000000M;oo0140oooo00<0oooO09a8
000000000`0000000`18W=l0oooo0=nLB0030000000307Boo`3oool0oooo09l0oooo003/0?ooo`03
0=nLB00007@0_ooo01<0oooo00<0_g@00018W03Oool00`3oool01`3OW9`0_ooo0?ooo`3oool0oono
07@0001d_ol0X@3oool00=l0oooo00@0gia800000000000007Bo2@3oool00`3OW4P004RL0=ooo`0C
0?ooo`030?oog`2LM9`0gooo00<0oooo00L0oono09bLg`3oool0oooo0?oo_`1d0000M;oo0:40oooo
003R0?ooo`030?noM000M;l0oooo00L0oooo00<0gia80018W03Oool0503oool00`2oM00000000000
00030000000604RLg`3oool0oooo0?oo_`1d0000M;ooX@3oool00=L0oooo00D0oono07@000000000
001809cOo`060?ooo`030?oo_`1d07@0_ooo00L0oooo00<0gia80018W03Oool00`3oool0203ooml0
W4P0000000000000001804Q80000B9`0gooo0P3oool0202oM00000000000000007@0_ooo0?ooo`3o
okl0M000100000001`0007@0_ooo0?ooo`3oool0oono07@0001d_ol0X@3oool00=L0oooo00H0_g@0
001d_`3oool0omnL04P0B02Lgol50?ooo`030=nLB00007@0_ooo00L0oooo00<0gia80018W03Oool0
0P3oool0503ooml0W4P00000001d_ol0oooo0?oo_`1d000000000000M02oool0_g@000000000M;l0
oooo0?oOW01807@0_ooo0?ooo`2oM4P0W=oo1`3oool00`3ookl0M00007Boo`2Q0?ooo`00eP3oool0
0`3ooml0W4P007Boo`020?ooo`030=nLB000B9`0gooo0080oooo00@0oono07@000000000B9cO203o
ool00`3OW4P004RL0=ooo`020?ooo`030?oOW0180000B9cO0080oooo00L0oooO09a80000M;l0oooo
0?oOW0180000B9cO00<0oooo00D0omnL07BLg`3oool0oono07@0000200000003001d_`3oool0oooo
0080oooo00<0oono07@0001d_ol0X@3oool00=H0oooo00<0oono07@0001d_ol00P3oool00`3ookl0
M00007Boo`030?ooo`030?oog`2LW;l0oooo00P0oooo00<0gia80018W03Oool00P3oool00`3o_g@0
000007Boo`020?ooo`070?oog`2LB00007Bo0?ooo`3o_g@0001809cOo`060?ooo`060=nLB00007@0
_ooo0?oo_`1d0000B9cO0`3oool00`3ookl0M00007Boo`2Q0?ooo`00eP3oool00`3ookl0M00004RL
g`030?ooo`030;md0000M;l0oooo00<0oooo00<0okmd04RLg`3oool01`3oool00`3OW4P004RL0=oo
o`020?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?noM00007@0_ooo00D0
oooo00L0oono07@00018W=l0oooo0?oog`2LB00007Bo00<0oooo00<0oono07@0001d_ol0X@3oool0
0=H0oooo00<0oooO09a80000M;l00`3oool0203OW4P0001d0;ooo`3ogi`0M9cO0?ooo`3OW4P007Bo
203oool00`3OW4P004RL0=ooo`020?ooo`030?oo_`1d0000B9cO0080oooo00H0oooO09a80000M;l0
oooo0?oOW01800050000000804RLg`3ookl0M00004RLg`3oool0oooO09a80018W=l30?ooo`030?oo
_`1d0000M;oo0:40oooo003G0?ooo`030;md00000000000000@0000000H0M;oo0?ooo`3OW4P00000
0000B02Lgol80?ooo`030=nLB000B9`0gooo00<0oooo00L0okmd0000B02Lgol0oooo0?noM0000000
07Bo0080oooo00`0_g@004RLg`3oool0oooo0=nLB00004P0W=oo0?ooo`2oM00007Bo0?ooo`3o_g@2
000000060018W03Oool0oooo0?oo_`1d0000M;ooX@3oool00=L0oooo00<0gia80000M02oool00P3o
ool00`3ookl0M00004RLg`0=0?ooo`030=nLB000B9`0gooo00@0oooo00<0okmd0000000000000P00
00001@00M;l0oooo0?ooo`3oool0gia80080000000H0001809cOo`3oool0oooo0?oog`2LB0020000
000307Boo`3oool0oooo0080oooo00<0oono07@0001d_ol0X@3oool00=L0oooo00<0omnL04P0001d
_ol00`3oool00`2oM00007Bo0?ooo`0<0?ooo`030=nLB000B9`0gooo00L0oooo00<0oooO09a80000
M;l04`3oool00`3ookl0M00007Boo`2Q0?ooo`00f03oool00`2oM00004RL0=ooo`020?ooo`030=nL
B000M;l0oooo00`0oooo00<0gia80018W03Oool01`3oool00`3ooml0W4P0001d_`0C0?ooo`030?oo
_`1d0000M;oo0:40oooo003H0?ooo`030?oo_`1d0000M;oo0080oooo00<0gia8001d_`3oool0303o
ool00`3OW4P004RL0=ooo`070?ooo`030?oog`2LB00007Bo01<0oooo00<0oono07@0001d_ol0X@3o
ool00=T0oooo00H0okmd0000B02Lgol0oooo0;md001d_ol=0?ooo`030=nLB00007@0_ooo00H0oooo
00@0oono07@00000000007Bo4`3oool00`3ookl0M00007Boo`2Q0?ooo`00fP3oool01@3ogi`0B000
00000000B9`0gooo00d0oooo00D0gia8000000000000000007Boo`050?ooo`030?oog`2LB00007Bo
0140oooo00D0gia8000000000000000007Boo`2Q0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00H@3oool01@3ogi`0
B0000000000007@0_ooo03@0oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3oool0
omnL04P000000000001d0;ooo`0e0?ooo`040?oog`2LB00004RL0=ooo`<0oooo00D0omnL04P00000
0000001d0;ooo`0b0?ooo`050?oOW018000000000000001d_ol0103oool01@3ogi`0B00000000000
07@0_ooo0380oooo00D0gia8000000000000000004RLg`040?ooo`050?oOW018000000000000M02o
ool0A@3oool00600oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo03@0oooo00<0gia80000
000000000`0000002P00B9`0gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000M02ooold0?ooo`04
0?oog`2LB00004RL0=ooo`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0300oooo00L0
omnL04P00018W=l0oooo0?oo_`1d0000M;oo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo0300oooo00L0gia80000001d_ol0oooo0?oog`2LB000B9cO0080oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo04@0oooo001P0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3o
ool0=03oool00`3OW4P0B9cO0?ooo`020?ooo`060=nLM02Lgol0oooo0;md0000B9`0gooo0P3oool0
0`2oM00007Bo0?ooo`0c0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0_g@00018W03Oool00P3o
ool00`2oM00007Bo0?ooo`0^0?ooo`030?oog`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo
0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0;P3oool00`3ooml0W4P0001d_`030?ooo`07
0?noM000B9`0gooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo04<0oooo001O0?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0=@3oool00`3OW4P0M;oo0?ooo`030?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0;`3oool00`2oM0000000000000050000
000407Boo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0;P3oool00`3ogi`0B00007Boo`03
0?ooo`060?noM00004P0W=oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0_0?ooo`030;md
0000M;l0oooo0080oooo00H0_g@00018W03Oool0oono07@00018W=l30?ooo`030=nLB00007@0_ooo
04<0oooo001O0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0=P3oool00`2oM4P0
W=oo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0;`3oool0303ogi`0
M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02L
gol0;P3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool0
0`3o_g@0001809cOo`0_0?ooo`0;0?oOW01807@0_ooo0?ooo`2oM000000007Boo`3oool0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`130?ooo`00G`3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo03H0oooo00<0oooO09a8M02oool00P3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo0300oooo00/0gia807Boo`3oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo02h0oooo00<0omnL04P0001d_ol00`3oool01P2oM0000018
09cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0<03oool01@3ookl0M4Q804P000000000
B9cO0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`130?ooo`00G`3oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo03L0oooo00L0okmd001d_`3oool0oooo0?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol0<03oool02`3ooml0W4Qd0;ooo`3ooml0W4P00018W03O
ool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0;P3oool0303ookl0M00000000018
W=l0oooo0?noM000000007Bo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0<03oool0
103ogi`0B00000000000M;l30?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0@`3o
ool005l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0g0?ooo`070?oo_`1d04P0
W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0340oooo00X0omnL04Qd_`3ooml0
W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0_0?ooo`030;md0000
B9`0_ia80080000000H004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo02l0
oooo00H0omnL04P0000007@0_ooo0?noM000M;l20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0@`3oool005l0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0g0?oo
o`070?oo_`1d0000B9cO0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0380oooo00T0
gia807Bog`2LB00004RL0=ooo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0_0?oo
o`030?oOW0180000M;oo00D0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0_0?oo
o`030;md000007@0_ooo0080oooo00H0_g@004RLg`3oool0oono07@00018W=l30?ooo`030=nLB000
07@0_ooo04<0oooo001P0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0<P3oool0
0`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0080oooo00<0gia80018W03Oool00P3oool0
0`2oM00007Bo0?ooo`0b0?ooo`050?oog`2LB4P0B4P00018W03Oool00P3oool00`3OW4P004RL0=oo
o`020?ooo`030;md0000M;l0oooo0300oooo00<0gia80000B02Lgol01@3oool00`3OW4P004RL0=oo
o`020?ooo`030;md0000M;l0oooo02l0oooo00<0_g@00018W03Oool00P3oool01`3OW4P007Bo0?oo
o`3oool0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`130?ooo`00H03oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0<`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0=03oool0103ogi`0B0000018W03Oool2
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0b0?ooo`030?noM000000007Bo00@0oooo
00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0300oooo00L0okmd0000001d_ol0oooo0?oo_`1d
0000M;oo0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo04@0oooo001Q0?ooo`050?oo
_`1d000000000000M02oool0=@3oool01P3ooml0W4P0000000000000001d0;ooo`@0oooo00D0oono
07@000000000001d0;ooo`0f0?ooo`030=nLB000B9`0gooo00<0oooo00D0oono07@000000000001d
0;ooo`0d0?ooo`050?oog`2LB00000000000M02oool00P3oool01@3ookl0M0000000000007@0_ooo
0380oooo00D0okmd000000000000000004RLg`040?ooo`050?oo_`1d000000000000M02oool0A@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool004h0oooo00D0omnL04P000000000
001d0;ooo`090?ooool00000<P00000S0?ooo`00C@3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool0203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool08@3oool004d0oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0Q0?ooo`00
C03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool0
A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001<0?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`140?ooo`030000O`3o
ool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo02D0oooo00<000000?ooo`3oool08@3oool004`0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`070?ooo`<00000103oool00`000000oooo0?ooo`0m0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo01d0oooo00<000000?ooo`3oool00`3oool3000002<0oooo001<0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
03`0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
03d0oooo00<0001o0?ooo`3oool0703oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
0240oooo001<0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo03/0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool06`3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo0240oooo001<0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gia80000M02oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03X0oooo
00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0240oooo
001=0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo03T0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?oo
o`0m0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool06@3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo0240oooo001=0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0>03oool00`0007l0oooo
0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo
0?ooo`0H0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool08@3oool004h0oooo00D0
oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo00X0oooo0P00000h0?ooo`030000
O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000
O`3oool0oooo01L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0Q0?ooo`00G03o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0?P3o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool05P3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0E0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool08@3oool005`0oooo0P00000?0?ooo`030000003oool0oooo
03<0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
03d0oooo00<0001o0?ooo`3oool0503oool00`000000oooo0?ooo`0<0?ooo`<000008`3oool005`0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0b0?ooo`030000O`3oool0oooo03h0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01<0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo0340oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo
0?ooo`0m0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool04P3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool0<03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0_
0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo0100oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0Q
0?ooo`00G03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo02h0oooo00<0001o0?oo
o`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?oo
o`3oool03`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0240oooo001L0?ooo`03
0000003oool0oooo01@0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0n0?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0>0?ooo`03
0000003oool0oooo01@0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0/0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0
?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00d0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo02/0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo0240oooo001L0?ooo`800000603oool00`000000oooo0?ooo`0Z0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo00/0oooo00<000000?ooo`3oool05@3oool3000002<0oooo001L0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0:@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo
03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`0X0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02L0oooo
00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool0203oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0240oooo
001L0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`070?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool08@3oool005`0oooo00<00000
0?ooo`3oool0703oool00`000000oooo0?ooo`0U0?ooo`030000O`3oool0oooo03h0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00H0oooo00<00000
0?ooo`3oool0703oool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo02@0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?oo
o`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3o
ool08`3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3o
ool08@3oool005`0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0R0?ooo`030000
O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000
O`3oool0oooo00<0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0Q0?ooo`00G03o
ool200000240oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`020?ooo`030000003o
ool0oooo01h0oooo0`00000S0?ooo`00G03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0200oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo03d0oooo00D0001o0?ooo`3oool0oooo0000000S0?ooo`030000003oool0oooo0240oooo001L
0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0n
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool0100007l0oooo00000000
000T0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo02<0oooo0P00000O
0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`000000oooo02D0oooo00<000000?ooo`3oool08@3oool00440oooo00D0omnL04P0
00000000001d0;ooo`030?ooo`050?noM0000000M;oo0?ooo`2oM0001P0000000`18W=l0oooo0?oo
o`050?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0703oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`000000oooo0?oo
o`0U0?ooo`030000003oool0oooo0240oooo00100?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`020?ooo`030?noM0000000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo
0?ooo`050?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool06`3oool00`0007l0oooo
0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?03oool00`000000001o
0?ooo`0V0?ooo`030000003oool0oooo0240oooo00100?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool01`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00D0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`0J0?ooo`030000O`3oool0oooo03h0oooo00<0
001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0k0?ooo`030000003oool0001o02L0oooo00<0
00000?ooo`3oool08@3oool003l0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`08
0?ooo`030=nLB01d_ol0oooo00T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0I
0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0j
0?ooo`040000003oool0oooo0000ObL0oooo00<000000?ooo`3oool08@3oool003l0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo00P0oooo00<00000
0?ooo`3oool0:@3oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0oooo03h0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0i0?ooo`050000003oool0oooo0?ooo`0007l09`3o
ool00`000000oooo0?ooo`0Q0?ooo`00?`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00T0oooo00<0oooO09a8M02oool0203oool3000002X0001o00<000000000O`0007l0d`0007l0
0`000000001o0000O`0X0000O`<000008`3oool003l0oooo00<0omnL04P0001d_ol00`3oool00`3o
_g@0001809cOo`0:0?ooo`030?noM000M;l0oooo00L0oooo00<000000?ooo`3oool0:`3oool00`00
0000oooo0?ooo`0F0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`00
07l0oooo0?ooo`0g0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3oool09@3oool00`00
0000oooo0?ooo`0Q0?ooo`00?`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0
oooo00<0oono07@0B02Lgol01`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo01D0
oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03H0
oooo00<000000?ooo`3oool0103oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240
oooo000o0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`3ookl0M000
04RLg`070?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0503oool00`0007l0oooo
0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0=@3oool00`000000oooo
0?ooo`050?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool00400oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?noW02Lgol0oooo0080oooo00<0
omnL04P00018W=l01`3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01<0oooo00<0
001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo0010
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3o
okl0M0000000M02oool70?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool04P3oool0
0`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0<`3oool0
0`000000oooo0?ooo`070?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool0
0440oooo00D0oono07@000000000001d0;ooo`080?ooo`060?oog`2LB00000000000000007@0_ooo
203oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0140oooo00<0001o0?ooo`3oool0
?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3oool0
203oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003o
ool0oooo0340oooo00<000000?ooo`3oool0403oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool0<@3oool00`000000oooo0?ooo`090?ooo`030000O`3o
ool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`0?0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool0
0`0007l0oooo0?ooo`0`0?ooo`030000003oool0oooo00X0oooo00<0001o0?ooo`3oool09@3oool0
0`000000oooo0?ooo`0Q0?ooo`00G03oool2000003@0oooo00<000000?ooo`3oool03P3oool00`00
07l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0;`3oool00`00
0000oooo0?ooo`0;0?ooo`030000O`3oool0oooo02<0oooo0`00000S0?ooo`00G03oool00`000000
oooo0?ooo`0d0?ooo`030000003oool0oooo00d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo02h0oooo00<000000?ooo`3oool0303oool00`0007l0
oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool0;@3oool00`000000oooo0?ooo`0=0?ooo`030000O`3oool0oooo02D0oooo
00<000000?ooo`3oool08@3oool002`0oooo00<0oooO09bL_`3oool0203oool00`2oW9`0gooo0?oo
o`0R0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool02`3oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0;03oool00`000000oooo0?oo
o`0>0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002`0oooo00<0omnL
04P000000000200000000`00B9`0gooo0?ooo`0R0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool02P3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool0:`3oool00`000000oooo0?ooo`0?0?ooo`030000O`3oool0oooo02D0oooo00<00000
0?ooo`3oool08@3oool002`0oooo00<0gia8000000000000200000000`00B9`0gooo0?ooo`0R0?oo
o`030000003oool0oooo03P0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0n0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0:P3oool00`000000oooo0?ooo`0@0?oo
o`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002d0oooo00<0_gAd0;ooo`3o
ool00P3oool00`3ooml0W9bo0?ooo`020?ooo`030;nLW03Oool0oooo0280oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`080?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`0Y0?ooo`030000003oool0oooo0140oooo00<0001o0?ooo`3o
ool09@3oool00`000000oooo0?ooo`0Q0?ooo`00;03oool00`3OW4P007Bo0?ooo`040?ooo`030=nL
W02oool0oooo02H0oooo00<000000?ooo`3oool0>P3oool2000000P0oooo00<0001o0?ooo`3oool0
?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02L0oooo0P00000D0?ooo`030000
O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002`0oooo00<0_g@00000001d_ol00`3o
ool00`3ogi`0B4RL0=ooo`0W0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool01@3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool09P3o
ool00`000000oooo0?ooo`0D0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3o
ool002`0oooo00<0okmd000000000000100000000`1d_ol0oooo0?ooo`0V0?ooo`800000?P3oool0
0`000000oooo0?ooo`040?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool08`3oool3
000002<0oooo000]0?ooo`060?oOW018000000000000000004P0W=oo:@3oool00`000000oooo0?oo
o`0n0?ooo`030000003oool0oooo00<0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?oo
o`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo03l0oooo00<00000
0?ooo`3oool00P3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool08`3oool00`000000oooo0?ooo`0G0?ooo`030000O`3oool0oooo02D0oooo00<00000
0?ooo`3oool08@3oool002l0oooo00<0giad09cOo`3oool00`3oool00`3ogi`0B4RL0=ooo`0T0?oo
o`030000003oool0oooo0400oooo00D000000?ooo`3oool0oooo0000O`100?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0
oooo02D0oooo00<000000?ooo`3oool08@3oool002h0oooo00<0okmd04RLg`3oool01@3oool00`2o
M7@0_ooo0?ooo`0S0?ooo`030000003oool0oooo0440oooo00@000000?ooo`3oool0001o@03oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0240oooo00<000000?ooo`3oool06@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000^0?ooo`060=nLB01d_ol0oooo
0?noM000B9`0gooo0P3oool00`3O_kl0gooo0?ooo`0S0?ooo`030000003oool0oooo0480oooo00<0
00000?ooo`0007l0@03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0200oooo00<0
00000?ooo`3oool06P3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000^
0?ooo`090=nLB000M;l0oono07AdW02LB000B9cO0?ooo`3o_g@0B9cO02D0oooo00<000000?ooo`3o
ool0@`3oool00`000000001o0?ooo`0o0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool07P3oool2000001d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00
;P3oool0203ookl0M0000000M02oool0omnL04P00000000007Bo9P3oool00`000000oooo0?ooo`14
0?ooo`030000003oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0M
0?ooo`030000003oool0oooo01d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q
0?ooo`00G03oool00`000000oooo0?ooo`140?ooo`030000O`000000oooo03h0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`0L0?ooo`030000003oool0oooo01h0oooo00<0001o0?oo
o`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool2000004D0oooo00@0001o0?ooo`00
00000000?@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01/0oooo00<000000?oo
o`3oool07`3oool00`0007l0oooo0?ooo`0S0?ooo`<000008`3oool003D0oooo00@0oooO09a80000
07@0_ooo8`3oool00`000000oooo0?ooo`140?ooo`050000O`3oool0oooo0?ooo`000000?03oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool0803oool0
0`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000b0?ooo`050?noM00000000000
0000B02Lgol09@3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo0080oooo00<00000
0?ooo`3oool0>@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01P0oooo0P00000S
0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002l0oooo00@0_g@00000
0000000007Bo3P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol01@3o
ool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0
0`3oool00`000000oooo0?ooo`0h0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
5`3oool00`000000oooo0?ooo`0S0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool0
8@3oool002/0oooo00D0omnL04P000000000000007Boo`0@0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000B9`0gooo1`3oool0
0`000000oooo0?ooo`140?ooo`030000O`3oool0oooo00@0oooo00<000000?ooo`3oool0=`3oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01H0oooo00<000000?ooo`3oool0903oool0
0`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000X0?ooo`050=nLB00000000000
0018W03Oool04`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo00@0oooO
09a80000B9`0gooo1`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo00D0oooo00<0
00000?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01D0oooo00<0
00000?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000V
0?ooo`030?oOW0180000B9cO01H0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`05
0?ooo`030;md00000000000000D0000000<0M;oo0?ooo`3oool0103oool00`000000oooo0?ooo`14
0?ooo`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo01@0oooo00<000000?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U
0?ooo`030000003oool0oooo0240oooo000/0?ooo`030?oog`2LW;l0oooo00P0oooo00<0_ibL0=oo
o`3oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0omnL07BL
g`3oool0oooo0?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo
0?ooo`070?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo
0?ooo`0C0?ooo`030000003oool0oooo02L0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo
0?ooo`0Q0?ooo`00;03oool00`3ogi`0B00000000008000000030018W03Oool0oooo00D0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=nLB01d_ol0oooo0?oog`2LB000
04RL0=ooo`070?ooo`<00000C`0007l2000008T0001o00<000000000O`0007l0CP0007l3000002<0
oooo000/0?ooo`030=nLB0000000000000P0000000<004RL0=ooo`3oool01@3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo
00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo
0340oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0A0?ooo`030000003oool0oooo
02T0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00;@3oool00`2oM7@0
_ooo0?ooo`020?ooo`030?oog`2LW;l0oooo0080oooo00<0_ibL0=ooo`3oool01@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00H0omnL04Qd_`3ooml0W4P00018W03Oool7
0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool02`3oool00`000000oooo0?ooo`0`
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool03`3oool2000002`0oooo00<0001o
0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00;03oool00`3OW4P007Bo0?ooo`040?oo
o`030=nLW02oool0oooo00T0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?oo
o`050=nLB01d_ml0W4P00018W03Oool01`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0
oooo00`0oooo00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo00h0oooo00<000000?ooo`3oool0;03oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0
oooo0240oooo000/0?ooo`030;md00000000M;oo00<0oooo00<0omnL04Q8W03Oool02`3oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00P0oooo00D0oooO09a8B018B00004RL0=ooo`07
0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`0^
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002`0oooo00<0okmd0000
00000000100000000`1d_ol0oooo0?ooo`0:0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`0:0?ooo`040?oOW018000004RL0=ooo`L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo
0?ooo`0>0?ooo`800000;P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo
000]0?ooo`060?oOW018000000000000000004P0W=oo3P3oool01@3ookl0M0000000000007@0_ooo
00`0oooo00<0gia80018W03Oool01`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo
0100oooo00<000000?ooo`3oool0:`3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
00/0oooo00<000000?ooo`3oool0;`3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo
0240oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool04@3oool00`000000
oooo0?ooo`0Z0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool02P3oool00`000000
oooo0?ooo`0`0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0B0?ooo`030000003oool0oooo02T0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`090?ooo`030000003oool0oooo0340oooo
00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00;P3oool00`3ooml0W7BL0=oo
o`050?ooo`030=nLW02oool0oooo02<0oooo0P0000150?ooo`030000O`3oool0oooo01<0oooo00<0
00000?ooo`3oool0:03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00P0oooo00<0
00000?ooo`3oool0<P3oool00`0007l0oooo0?ooo`0S0?ooo`<000008`3oool002h0oooo00<0okmd
0000000000001@0000000`0007@0_ooo0?ooo`0S0?ooo`030000003oool0oooo04@0oooo00<0001o
0?ooo`3oool0503oool00`000000oooo0?ooo`0W0?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool01`3oool00`000000oooo0?ooo`0c0?ooo`030000O`3oool0oooo02D0oooo00<00000
0?ooo`3oool08@3oool003H0oooo00<0gibL0;ooo`3oool08`3oool00`000000oooo0?ooo`140?oo
o`030000O`3oool0oooo01D0oooo00<000000?ooo`3oool09P3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0U0?oo
o`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3o
ool05P3oool00`000000oooo0?ooo`0U0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool01@3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3o
ool08@3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0G0?ooo`030000
003oool0oooo02@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`030?ooo`800000
>03oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003o
ool0oooo04@0oooo00<0001o0?ooo`3oool0603oool2000002@0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo00<0001o0?ooo`3oool09@3o
ool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`140?ooo`030000O`3oool0
oooo01X0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`0m0?ooo`050000O`3oool0
oooo0?ooo`000000>`3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000a
0?ooo`060?oog`2LB00004RL0=ooo`3ooml0W9bo9@3oool00`000000oooo0?ooo`140?ooo`030000
O`3oool0oooo01/0oooo00<000000?ooo`3oool0803oool00`0007l0oooo0?ooo`0m0?ooo`040000
O`3oool0oooo000003`0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00
<@3oool01P3ookl0M0000018W03Oool0_gA809cOobD0oooo00<000000?ooo`3oool0A03oool00`00
07l0oooo0?ooo`0L0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0?@3oool00`00
07l0oooo0000000m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool00380
oooo00@0gia8000000000000M;oo9P3oool2000004D0oooo00<0001o0?ooo`3oool07@3oool00`00
0000oooo0?ooo`0N0?ooo`030000O`3oool0oooo03d0oooo00<0001o0000003oool0?@3oool00`00
07l0oooo0?ooo`0S0?ooo`<000008`3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0
oooo0?ooo`0N0?ooo`030000003oool0oooo01d0oooo00<0001o0?ooo`3oool0?@3oool00`000000
oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0O0?ooo`030000003oool0oooo01`0oooo
00<0001o0?ooo`3oool0>`3oool2000000030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
9@3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`140?ooo`030000O`3o
ool0oooo0200oooo0P00000L0?ooo`030000O`3oool0oooo03X0oooo00@000000?ooo`3oool0001o
?`3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003o
ool0oooo04@0oooo00<0001o0?ooo`3oool08P3oool00`000000oooo0?ooo`0I0?ooo`030000O`3o
ool0oooo03T0oooo00D000000?ooo`3oool0oooo0000O`0o0?ooo`030000O`3oool0oooo02D0oooo
00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?oo
o`0S0?ooo`030000003oool0oooo01P0oooo00<0001o0?ooo`3oool0=`3oool2000000@0oooo00<0
001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000/
0?ooo`030?oog`2LW;l0oooo00P0oooo00<0_ibL0=ooo`3oool08P3oool00`000000oooo0?ooo`14
0?ooo`030000O`3oool0oooo02@0oooo0P00000H0?ooo`030000O`3oool0oooo03H0oooo00<00000
0?ooo`3oool0103oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<00000
0?ooo`3oool08@3oool002`0oooo00<0omnL04P000000000200000000`00B9`0gooo0?ooo`0R0?oo
o`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool09P3oool00`000000oooo0?ooo`0E0?oo
o`030000O`3oool0oooo03D0oooo00<000000?ooo`3oool01@3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002`0oooo00<0gia800000000
0000200000000`00B9`0gooo0?ooo`0R0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3o
ool09`3oool00`000000oooo0?ooo`0D0?ooo`030000O`3oool0oooo03<0oooo0P0000080?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00
;@3oool00`2oM7@0_ooo0?ooo`020?ooo`030?oog`2LW;l0oooo0080oooo00<0_ibL0=ooo`3oool0
8P3oool2000004D0oooo00<0001o0?ooo`3oool0:03oool2000001@0oooo00<0001o0?ooo`3oool0
<P3oool00`000000oooo0?ooo`080?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
8`3oool3000002<0oooo000/0?ooo`030=nLB000M;l0oooo00@0oooo00<0gibL0;ooo`3oool09P3o
ool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo02X0oooo00<000000?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0a0?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000/0?ooo`030;md00000000
M;oo00<0oooo00<0omnL04Q8W03Oool09`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0
oooo02/0oooo00<000000?ooo`3oool0403oool00`0007l0oooo0?ooo`0`0?ooo`030000003oool0
oooo00X0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0
oooo0240oooo000/0?ooo`030?noM0000000000000@0000000<0M;oo0?ooo`3oool09P3oool00`00
0000oooo0?ooo`140?ooo`030000O`3oool0oooo02`0oooo00<000000?ooo`3oool03`3oool00`00
07l0oooo0?ooo`0^0?ooo`8000003@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
02D0oooo00<000000?ooo`3oool08@3oool002d0oooo00H0omnL04P00000000000000000B02Lgol>
0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050?oOW018
000000000000001d_ol0203oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo02d0oooo
0P00000?0?ooo`030000O`3oool0oooo02d0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo
0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool00400oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool01`3ogi`0
B00004RLg`3oool0oono07@0001d_ol01`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0
oooo02l0oooo00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`0/0?ooo`030000003oool0
oooo00h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0
oooo0240oooo000b0?ooo`040?oOW018000004RL0=ooo`X0oooo00<0_g@00018W03Oool00P3oool0
0`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03Oool01P3oool0
0`000000oooo0?ooo`140?ooo`030000O`3oool0oooo0300oooo00<000000?ooo`3oool02`3oool0
0`0007l0oooo0?ooo`0[0?ooo`030000003oool0oooo00l0oooo00<0001o0?ooo`3oool0?@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000a0?ooo`050=nLB000B9`0gmnL
04Q8W03Oool02@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00D0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030000003oool0oooo04@0oooo00<0001o
0?ooo`3oool0<@3oool00`000000oooo0?ooo`0:0?ooo`030000O`3oool0oooo02X0oooo00<00000
0?ooo`3oool0403oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<00000
0?ooo`3oool08@3oool00300oooo00H0_g@004RLg`3oool0omnL04Q8W03Oool90?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00H0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0b0?ooo`030000003o
ool0oooo00T0oooo00<0001o0?ooo`3oool0:03oool2000001<0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000_0?ooo`070=nLB01d_ol0
oooo0?ooo`3ogi`0B4RL0=ooo`090?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
1@3oool00`3ogi`0B00007Boo`030?ooo`030;md000004P0W=oo00H0oooo0`00001j0000O`800000
<`0007l00`000000001o0000O`1i0000O`<000008`3oool002h0oooo00<0gia80000000000001@00
00000`0007@0_ooo0?ooo`060?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3o
ool0203ookl0M00000000018W=l0oooo0?noM000000007Bo1`3oool00`000000oooo0?ooo`140?oo
o`030000O`3oool0oooo03D0oooo00<000000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0V0?oo
o`030000003oool0oooo01@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000003oool0oooo0240oooo000c0?ooo`030?oOW018B9`0gooo00T0oooo00<0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`060?ooo`030;md0000B9`0_ia80080000000<004RL0=ooo`3o
ool01P3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03H0oooo00<000000?ooo`3o
ool01@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000o0?ooo`030?oo
_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`3ogi`0B00007Boo`0;0?ooo`030000
003oool0oooo04@0oooo00<0001o0?ooo`3oool0=`3oool00`000000oooo0?ooo`040?ooo`030000
O`3oool0oooo02<0oooo0P00000H0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
9@3oool00`000000oooo0?ooo`0Q0?ooo`00=@3oool0103ooml0W4P00000M02oool70?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`3OW4P0001809cOo`0:0?ooo`030000
003oool0oooo04@0oooo00<0001o0?ooo`3oool0>03oool2000000@0oooo00<0001o0?ooo`3oool0
8P3oool00`000000oooo0?ooo`0H0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
9@3oool00`000000oooo0?ooo`0Q0?ooo`00<P3oool01@3o_g@000000000000004P0W=oo00T0oooo
00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00T0oooo00<0okmd00000000M;l02@3oool00`00
0000oooo0?ooo`140?ooo`030000O`3oool0oooo03X0oooo00D000000?ooo`3oool0oooo0000O`0S
0?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000003oool0oooo0240oooo000_0?ooo`040;md000000000000001d_`h0oooo00D0oono
07@000000000001d0;ooo`0;0?ooo`050?oog`2LB00000000000M02oool01P3oool00`000000oooo
0?ooo`140?ooo`030000O`3oool0oooo03/0oooo00@000000?ooo`3oool0001o8@3oool2000001`0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240
oooo000[0?ooo`050?oOW018000000000000001d_ol0;03oool00`000000oooo0?ooo`140?ooo`03
0000O`3oool0oooo03`0oooo0P0000000`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo01`0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240
oooo000X0?ooo`050=nLB000000000000018W03Oool0;`3oool00`000000oooo0?ooo`140?ooo`03
0000O`3oool0oooo03h0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0M0?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00
9P3oool00`3ogi`0B00004RLg`0c0?ooo`800000A@3oool00`0007l0oooo0?ooo`0n0?ooo`030000
O`000000oooo01/0oooo0P00000P0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
8`3oool3000002<0oooo000/0?ooo`030?oog`2LW;l0oooo00P0oooo00<0_ibL0=ooo`3oool08P3o
ool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03h0oooo00@0001o0?ooo`0000000000
6@3oool00`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
9@3oool00`000000oooo0?ooo`0Q0?ooo`00;03oool00`3ogi`0B00000000008000000030018W03O
ool0oooo0280oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`050000O`3o
ool0oooo0?ooo`0000005`3oool00`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00;03oool00`3OW4P000000000
0008000000030018W03Oool0oooo0280oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo0080oooo0P00000C0?ooo`800000903oool00`0007l0oooo0?oo
o`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool002d0oooo00<0_gAd
0;ooo`3oool00P3oool00`3ooml0W9bo0?ooo`020?ooo`030;nLW03Oool0oooo0280oooo00<00000
0?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00@0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`0T0?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00;03oool00`3OW4P007Bo0?ooo`040?oo
o`030=nLW02oool0oooo02H0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?oo
o`030000O`3oool0oooo00D0oooo0P00000>0?ooo`030000003oool0oooo02D0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo000/0?ooo`03
0;md00000000M;oo00<0oooo00<0omnL04Q8W03Oool09`3oool00`000000oooo0?ooo`140?ooo`03
0000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool01`3oool00`000000oooo0?ooo`090?ooo`80
0000:03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3o
ool08@3oool002`0oooo00<0okmd000000000000100000000`1d_ol0oooo0?ooo`0V0?ooo`030000
003oool0oooo04@0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`080?ooo`800000
203oool00`000000oooo0?ooo`0X0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
9@3oool00`000000oooo0?ooo`0Q0?ooo`00;@3oool01P3ogi`0B000000000000000001809cOobT0
oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00X0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0Y0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool2000004D0oooo
00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0;0?ooo`8000000P3oool2000002`0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0S0?ooo`<000008`3oool005`0oooo00<0
00000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00d0oooo0P00
000^0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?oo
o`0Q0?ooo`00;P3oool00`3ooml0W7BL0=ooo`050?ooo`030=nLW02oool0oooo02<0oooo00<00000
0?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00`0oooo1000000]
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q
0?ooo`00;P3oool00`3o_g@0000000000005000000030000M02oool0oooo02<0oooo00<000000?oo
o`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00X0oooo0P0000040?oo
o`030000003oool0oooo02X0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000003oool0oooo0240oooo000f0?ooo`030=nLW02oool0oooo02<0oooo00<000000?ooo`3o
ool0A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3o
ool01@3oool2000002X0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0
?P3oool00`0007l0oooo0?ooo`070?ooo`8000002P3oool00`000000oooo0?ooo`0W0?ooo`030000
O`3oool0oooo03d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00G03o
ool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool01P3o
ool00`000000oooo0?ooo`0;0?ooo`8000009`3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0A03oool0
0`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00@0oooo0P00000@0?ooo`030000003oool0
oooo02@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0
oooo0240oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?P3oool00`00
07l0oooo0?ooo`020?ooo`8000004`3oool2000002@0oooo00<0001o0?ooo`3oool0?@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`800000A@3oool00`0007l0
oooo0?ooo`0n0?ooo`050000O`3oool0oooo0?ooo`0000005`3oool00`000000oooo0?ooo`0Q0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool08`3oool3000002<0oooo001L0?ooo`03
0000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?P3oool0100007l0oooo00000000000I0?oo
o`8000008@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?oo
o`3oool08@3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0n0?ooo`03
0000O`000000oooo01`0oooo00<000000?ooo`3oool07P3oool00`0007l0oooo0?ooo`0m0?ooo`03
0000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0
A03oool00`0007l0oooo0?ooo`0m0?ooo`8000007`3oool2000001h0oooo00<0001o0?ooo`3oool0
?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo00110?ooo`050?oOW018
000000000000M02oool00`3oool00`3o_g@0000007Boo`030?ooo`050=nLB0000000000000000018
W=l0203oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03/0oooo0P0000000`3oool0
001o0?ooo`0P0?ooo`030000003oool0oooo01/0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo00100?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0gia80000001d_ol0oooo0?oog`2L
B000B9cO00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0j0?ooo`050000003o
ool0oooo0?ooo`0007l08P3oool2000001/0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000003oool0oooo0240oooo00100?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool01@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00H0oooo00<0
00000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0h0?ooo`800000103oool00`0007l0oooo0?oo
o`0R0?ooo`030000003oool0oooo01P0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000003oool0oooo0240oooo000o0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01P3oool00`2oM00007Bo0?ooo`020?ooo`030;md0000B9`0gooo00H0oooo00<00000
0?ooo`3oool0A03oool00`0007l0oooo0?ooo`0g0?ooo`030000003oool0oooo00@0oooo00<0001o
0?ooo`3oool08`3oool2000001P0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000003oool0oooo0240oooo000o0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01P3oool01`3ogi`0B01d0;ooo`3oool0_g@00000001d_ol01`3oool00`000000oooo0?oo
o`140?ooo`030000O`3oool0oooo03D0oooo0P0000070?ooo`030000O`3oool0oooo02D0oooo00<0
00000?ooo`3oool05@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<0
00000?ooo`3oool08@3oool003l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`07
0?ooo`050?oo_`1dB4P0B00000000018W=l0203oool3000007X0001o0P00000b0000O`800000N`00
07l3000002<0oooo000o0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool0
103ogi`0B00000000000M;l90?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0<P3o
ool00`000000oooo0?ooo`090?ooo`030000O`3oool0oooo02P0oooo0P00000C0?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00?`3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0omnL04P0000007@0_ooo0?no
M000M;l80?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0<03oool2000000`0oooo
00<0001o0?ooo`3oool0:P3oool00`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00?`3oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@00000M02oool00P3oool00`2oM000B9cO0?oo
o`060?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0;P3oool2000000h0oooo00<0
001o0?ooo`3oool0:`3oool200000100oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000003oool0oooo0240oooo00100?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool01P3oool00`2oM00004RL0=ooo`020?ooo`030=nLB000M;l0oooo00H0oooo00<00000
0?ooo`3oool0A03oool00`0007l0oooo0?ooo`0/0?ooo`800000403oool00`0007l0oooo0?ooo`0]
0?ooo`8000003P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<00000
0?ooo`3oool08@3oool00400oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00L0
okmd0000001d_ol0oooo0?oo_`1d0000M;oo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0
oooo0?ooo`0Z0?ooo`8000004P3oool00`0007l0oooo0?ooo`0_0?ooo`800000303oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool00440oooo
00D0oono07@000000000001d0;ooo`090?ooo`050?noM0000000000000000018W=l0203oool00`00
0000oooo0?ooo`140?ooo`030000O`3oool0oooo02P0oooo0P00000D0?ooo`030000O`3oool0oooo
0340oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
02D0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0
oooo0?ooo`0V0?ooo`8000005P3oool00`0007l0oooo0?ooo`0b0?ooo`8000002@3oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0T0?ooo`800000603oool00`0007l0oooo
0?ooo`0d0?ooo`8000001`3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02D0oooo
00<000000?ooo`3oool08@3oool005`0oooo0P0000150?ooo`030000O`3oool0oooo0280oooo0P00
000J0?ooo`030000O`3oool0oooo03H0oooo0P0000050?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool08`3oool3000002<0oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o
0?ooo`3oool0803oool2000001`0oooo00<0001o0?ooo`3oool0>03oool2000000<0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?oo
o`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool07P3oool2000001h0oooo00<0001o0?oo
o`3oool0>P3oool2000000030?ooo`0007l0oooo03h0oooo00<0001o0?ooo`3oool09@3oool00`00
0000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo01`0
oooo0P00000P0?ooo`030000O`3oool0oooo03`0oooo0`00000n0?ooo`030000O`3oool0oooo02D0
oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo
0?ooo`0J0?ooo`8000008P3oool00`0007l0oooo0?ooo`0m0?ooo`040000O`3oool00000000003`0
oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo
0?ooo`140?ooo`030000O`3oool0oooo01P0oooo0P00000T0?ooo`030000O`3oool0oooo03d0oooo
00@0001o0?ooo`3oool0oooo0P00000j0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3o
ool08@3oool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0F0?ooo`800000
9P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00<0oooo0`00000g0?ooo`030000
O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0A03o
ool00`0007l0oooo0?ooo`0D0?ooo`800000:03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo00H0oooo0P00000e0?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3o
ool005`0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0B0?ooo`800000:P3oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00P0oooo0`00000b0?ooo`030000O`3oool0
oooo02D0oooo00<000000?ooo`3oool08@3oool005`0oooo0P0000150?ooo`030000O`3oool0oooo
0100oooo0P00000/0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool02`3oool20000
0300oooo00<0001o0?ooo`3oool08`3oool3000002<0oooo001L0?ooo`030000003oool0oooo04@0
oooo00<0001o0?ooo`3oool03P3oool2000002h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0=0?ooo`800000;P3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240
oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool02`3oool300000300oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0?0?ooo`<00000:`3oool00`0007l0oooo
0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo04@0oooo00<0
001o0?ooo`3oool02@3oool2000003<0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0B0?ooo`800000:@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L
0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool01P3oool3000003D0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0D0?ooo`<000009P3oool00`0007l0oooo0?ooo`0U
0?ooo`030000003oool0oooo0240oooo001L0?ooo`030000003oool0oooo04@0oooo00<0001o0?oo
o`3oool00`3oool3000003P0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0G0?oo
o`<000008`3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`03
0000003oool0oooo04@0oooo00<0001o0?ooo`0000000`00000k0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool06P3oool300000200oooo00<0001o0?ooo`3oool09@3oool00`000000
oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`130?ooo`<00000?`3oool00`0007l0oooo
0?ooo`0m0?ooo`030000O`3oool0oooo01d0oooo0`00000M0?ooo`030000O`3oool0oooo02D0oooo
00<000000?ooo`3oool08@3oool005`0oooo00<000000?ooo`3oool0@03oool3000000030?ooo`00
07l0oooo03l0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0P0?ooo`<000006P3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001L0?ooo`800000?@3oool4
000000@0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo02<0oooo0`00000G0?ooo`030000O`3oool0oooo02<0oooo0`00000S0?ooo`00G03oool00`00
0000oooo0?ooo`0h0?ooo`@00000203oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo
03d0oooo00<0001o0?ooo`3oool09P3oool4000001<0oooo00<0001o0?ooo`3oool09@3oool00`00
0000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo0?ooo`0d0?ooo`@00000303oool00`0007l0
oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0:P3oool300000100
oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0Q0?ooo`00G03oool00`000000oooo
0?ooo`0`0?ooo`@00000403oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool0;@3oool4000000`0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo
0?ooo`0Q0?ooo`00CP3oool01P2oM000000000000000000004RL0=ooo`P0oooo00<000000?ooo`3o
ool0;03oool4000001@0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`03
0000O`3oool0oooo0340oooo1P0000060?ooo`030000O`3oool0oooo02D0oooo00<000000?ooo`3o
ool08@3oool004l0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0V0?ooo`H00000
603oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
=`3oool6000000030000O`3oool0oooo02D0oooo00<000000?ooo`3oool08@3oool004l0oooo00<0
omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0P0?ooo`H000007P3oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool600000280oooo00<0
00000?ooo`3oool08@3oool004l0oooo00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0J
0?ooo`H00000903oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`030?ooo`H00000703oool00`000000oooo0?ooo`0Q
0?ooo`00C`3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo00l0oooo2`00000Z0?oo
o`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo00T0oooo2`00000A0?ooo`030000003oool0oooo0240oooo001?0?ooo`03
0?oOW0180000M;oo00X0oooo0`0000040?ooo`/00000=@3oool00`0007l0oooo0?ooo`0n0?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0D0?ooo`X0
00001@3oool3000002<0oooo001?0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0
A03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
?@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo0240oooo001?0?ooo`030?oOW018
0000M;oo00X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0m0?ooo`030000003o
ool0oooo03h0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0m0?ooo`030000003o
ool0oooo02D0oooo00<000000?ooo`3oool08@3oool004l0oooo00<0omnL04P0001d_ol02P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool08@3oool0
04l0oooo00<0omnL04P0001d_ol02P3ooooo0000038000008`3oool004d0oooo00D0oooO09a80000
0000000007Boo`3o0?oooel0oooo001?0?ooo`030?noM0000000M;oo0?l0ooooG`3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool008/0oooo00<0oono0;no_`1d00000P0000001@1d
_ol0oooo0?ooo`3oool0_g@00080000000D0001809cOo`3oool0oooO09a800020000000>0018W02L
M0000000001d_`3oool0omnL04P000000000B9cO0?ooo`3ogi`0B00000000018W=l20?ooo`070;md
0000000000000000M02oool0oooo0;md0002000000030000B02Lgol0oooo0080oooo00D0_g@00000
00000000001d0;ooo`020?ooo`080?oog`2LB00000000000000004P0B4P00018W03Oool30?ooo`09
0?oog`2LB000000000000000B9`0W7@000000000M;l0_g@00080000000<0001d0;oog`2LB0000P00
00001000B9`0gooo0?oog`2LB002000000040000B018B00004RL0=ooo`D0oooo00D0_g@000000000
0000001809cOo`030?ooo`050;md0000000000000000M02oool00`3oool01@2oM000000000000000
04P0W=oo00<0oooo00D0gia8000000000000000007Boo`020?ooo`050=nLB000000000000000M02L
W4P00P0000000`0007@0_ooo0?ooo`040?ooo`060=nLB000000000000000B02Lgml0W4P00P000000
0`0007@0_ooO09a8000200000004001d_`3oool0oooo0=nLB0<0000000@0M;oo0?ooo`3oool0gia8
0P0000000`0004P0W=oO09a80002000000030000M02ooml0W4P00080000000@007Bo0?ooo`3oool0
_g@00P000000100007@0_ooo0?ooo`2oM002000000030000M02ooml0W4P00080000000P004RL0=oo
o`3ogi`0B00000000018W=l0oooO09a80080000000@004RL09ad0000000007BoD`3oool008/0oooo
02d0oono07@0B02Lgol0oooo0?oo_`1d04P0W=oo0?ooo`2oM00000000018W03Oool0oooO09a8M02o
ool0okmd0000001d_ol0oooo0?oOW018000007Bo0?no_`2oool0gia80000M02oool0gkno0=ooo`3O
W4P0001d0;ooo`3O_kl0gooo0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80000M02o
ool00`3oool04@2oM0000000001d_`3oool0omnL04P0M02oool0oooO09a800000000M;oo0?ooo`3o
okl0M0000000000007@0_ooo0080oooo01/0okmd0000001d_ol0oooo0?oOW018000007Bo0?no_`2o
ool0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ooml0W4P00000001d_ol0oooo0?oo_`1d
000000000000M02oool01@3oool00`3OW4P0001d0;ooo`030?ooo`0G0;md0000000007Bo0?ooo`3o
gi`0B01d0;ooo`3oool0_g@000000000B9`0gooo0?oog`2LB7@0_ooo0?ooo`3OW4P0000004RLg`3o
ool0oooO09a80018W=l00P3oool01`3OW4P0001d0;ooo`3oool0gia80000M02oool01`3oool00`3O
W4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`070=nLB000
0000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`030=nLB00007@0_ooo0080oooo00L0_g@00018W03O
ool0oooo0?oog`2LB00007Bo0080oooo00l0_g@000000000M;l0oooo0?oOW01807@0_ooo0?ooo`3O
W4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool03@3OW4P0001d0;ooo`3O_kl0gkmd0000001d
_ol0oooo0?oOW018000007Bo0?no_`2oool0DP3oool008/0oooo00X0oono0;noo`3oool0oooo0?no
M0000000M;oo0?oOW018000007Bo0`3oool02@3ogi`0W;oo0?noM00004P0W=oo0?ooo`3ooml0W4P0
001d_`020?ooo`030=nLB000B9`0gooo0080oooo00L0gia80018W03Oool0oooo0?oOW0180000B9cO
00<0oooo00H0omnL07BLg`3oool0gia80000M02oool20?ooo`030?oOW0180000B9cO00<0oooo00D0
omnL07BLg`3ogi`0B00004RLg`020?ooo`030?oog`2LB00007Bo00@0oooo00L0okmd0000B02Lgol0
oooo0?oog`2LB00007Bo0080oooo00/0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0
B00004RLg`020?ooo`030?oog`2LB00007Bo00L0oooo00<0gia80000M02oool00P3oool00`3ogi`0
B00004RLg`030?ooo`050?oOW01dW=l0omnL04P00000M;l00`3oool01@3ogi`0W;oo0?oOW0180000
07Bo00<0oooo00/0okmd0000M02oool0oooo0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`070?oo
o`030=nLB00007@0_ooo0080oooo00/0_g@00018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3ogi`0
B000001d_`030?ooo`070?noM00007@0_ooo0?ooo`3OW4P0001d0;ooo`020?ooo`0;0;md0000B9`0
gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool02P3ogi`0M9cO0?ooo`3OW4P0
04RL0=ooo`3oool0oooO09a80000M;l20?ooo`0;0=nLB000B9`0gooo0?ooo`3o_g@0001809cOo`3o
ool0oooO09a80000M;l0E03oool008d0oooo00P0oono07@000000000001d0;ooo`3o_g@0001809cO
o`H0oooo00H0okmd00000000M;l0oooO09a80000M;l20?ooo`030=nLB000B9`0gooo0080oooo00L0
gia80018W03Oool0oooo0?noM00004P0W=oo00H0oooo00<0gia80000M02oool00P3oool00`3o_g@0
001809cOo`050?ooo`030?noM0000000M;oo0080oooo00<0oooO09a80000M;l01@3oool01P3o_g@0
0000001d_`3ooml0W4P0001d_`80oooo00/0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o
_g@0000007Boo`020?ooo`030?oog`2LB00007Bo00L0oooo00<0gia80000M02oool00P3oool00`3o
_g@0001809cOo`050?ooo`030?noM00004P0W=oo00D0oooo00<0okmd0000001d_ol00`3oool02`3o
_g@0000007Boo`3oool0gia80000M02oool0oooo0=nLB00007@0_ooo00L0oooo00<0gia80000M02o
ool00P3oool02`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo00<0oooo
00L0okmd0000001d_ol0oooo0=nLB00007@0_ooo0080oooo00/0_g@00018W03Oool0oooo0?oog`2L
B00007Bo0?ooo`3o_g@0001809cOo`060?ooo`070=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`02
0?ooo`030=nLB000B9`0gooo0080oooo00H0okmd00000000M;l0oooO09a80000M;mD0?ooo`00S03o
ool0102oM000000000000018W=l20?ooo`030?noM00007@0_ooo00P0oooo00@0okmd000000000000
07Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`070=nLB000B9`0gooo0?ooo`3o_g@0001d0;ooo`06
0?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000M02oool01@3oool00`3o_g@0000007Boo`02
0?ooo`030?oog`2LB00007Bo00L0oooo00@0okmd00000000000007Bo0P3oool02`3OW4P004RL0=oo
o`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo0080oooo00<0oooO09a80000M;l01`3oool0
0`3OW4P0001d0;ooo`020?ooo`030?noM00007@0_ooo00D0oooo00<0okmd0000M02oool01@3oool0
0`3o_g@0000007Boo`030?ooo`0;0?noM0000000M;oo0?ooo`3OW4P0001d0;ooo`3oool0gia80000
M02oool01`3oool00`3OW4P0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0001d
_`3oool0okmd0000001d_ol00`3oool01`3o_g@0000007Boo`3oool0gia80000M02oool00P3oool0
2`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00007@0_ooo00H0oooo00L0gia80018
W03Oool0oooo0?oog`2LB00007Bo0080oooo00<0gia80018W03Oool0103oool0103o_g@000000000
0000M;mD0?ooo`00R`3oool00`3ogi`0B00004RLg`020?ooo`050;nLW03Oool0omnL04P0B02Lgol0
0P3oool02P3o_g@0001809cOo`3ookl0M00007Boo`3oool0oooO09a80000M;l20?ooo`030=nLB000
B9`0gooo0080oooo00H0gia80018W03Oool0oooo0?oOW01800050000000504RLg`3oool0gia80000
B02Lgol00P3oool00`3ogi`0B000000000040000000404RLg`3ookl0M00004RLg`80oooo00<0oooO
09a80000M;l0103oool01`3ookl0M00007Boo`3oool0oooO09a80000M;l00P3oool02`2oM000001d
0;ooo`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0080oooo00<0oooO09a80000M;l01`3o
ool00`3OW4P0001809cOo`020?ooo`030?oOW0180000000000@0000000@0B9cO0?oOW01804P0W=oo
0P3oool01P3o_g@0001809cOo`3ookl0M00007Boo`<0oooo00/0_g@00000B02Lgol0oooo0=nLB000
07@0_ooo0?ooo`3OW4P0001d0;ooo`070?ooo`030=nLB00004P0W=oo0080oooo00/0_g@00000M02o
ool0oooo0?oog`2LB00007Bo0?ooo`3ookl0M00007Boo`030?ooo`070;md000004P0W=oo0?ooo`3O
W4P0001809cOo`020?ooo`0:0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P000D0
000000T0B9cO0?ooo`2oM000001d0;ooo`3oool0oooO09a80000M;l00P3oool02`3OW4P004RL0=oo
o`3oool0oono07@0001d_ol0oooo0?oog`2LB00007Bo05@0oooo002;0?ooo`0F0?oo_`1d07@0_ooo
0?ooo`3o_g@004RL0=ooo`3oool0_g@00018W03Oool0oooO09a8000007@0_ooo0?oo_`1d0000B9cO
0?ooo`3ookl0M000001d_`80oooo00<0gia80018W03Oool00P3oool00`3OW4P004RL0=ooo`020?oo
o`0@0;md0018W=l0oooo0?ooo`3OW4P0001809cOg`2LB00000000000B018M9`0gkmd0000M02oool0
_g@004RLg`80oooo00/0gia80000B02Lgol0oooo0?noM00004P0W=oo0?ooo`3o_g@00000001d_`04
0?ooo`0@0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`3ooml0W4P000000000B4P0B01809cOo`3o
okl0M00004RLg`80oooo00L0okmd0000B02Lgol0oooo0?noM000000007Bo00D0oooo00X0oooO09a8
00000000001804QdW03O_g@0001d0;ooo`2oM000B9cO0P3oool0703OW4P0001809cOo`3oool0_g@0
0018W03Oool0oooO09a8000007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0okmd00000000M;l0oooO
09a800000000001d0;ooo`3oool0gia80000M02oool50?ooo`0>0?oog`2LB00000000000B018B4P0
W=oo0?oo_`1d000007Ad07@0M02oool0oono07@00018W=l20?ooo`0E0=nLB000B9`0gooo0?ooo`3o
_g@00000001d_`3ooml0W4P00000000004P0B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d
0000B9cO0080oooo00l0_g@004RLg`3oool0oooo0=nLB00004P0W=oO09a80000000004Q804P0B02L
gol0oono07@00018W=l00P3oool02`3OW4P004RL0=ooo`3oool0oono07@00018W=l0oooo0?oo_`1d
000007Bo05@0oooo002<0?ooo`060?noM00000000000001d_`2oW9`0gooo0P3oool01@3OW4P00000
0000000004P0W=oo0080oooo00P0oono07@00000000000000000M02oool0oooo0;md00<0000000<0
04RL09ad000000000P0000001000B9`0gooo0?ooo`3OW4P20000000<0000B02Lgol0oooo0?ooo`2o
M000001d0;oOW0180000001d0;ooo`3oool0gia80P0000000`0004P0W=oo0?ooo`020?ooo`030?no
M000000000000080000000<007Bo0?ooo`3oool00`3oool01P3ookl0M000000000000000001d0;oo
o`80oooo00L0_g@00018W03Ooml0W4P00000000007@0_ooo00<0oooo00<0okmd0000000000000P00
00000`00M;l0oooo0?ooo`050?ooo`080;md000007@0_mnL04P0000007@0_ooo0?ooo`3OW4P20000
00030000B02Lgol0oooo0080oooo00D0gia8000000000000001809cOo`030?ooo`050?noM0000000
000000000018W=l00`3oool01`2oM000001d0;ooo`3oool0gia80000M02oool01`3oool0302oM000
001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`<0oooo00D0okmd00000000
0000000004RLg`030?ooo`0<0;md000007@0_ooO09a800000000001809cOo`3ookl0M00000000000
07@0_ooo0`3oool01@3OW4P000000000000004P0W=oo0080oooo00T0_g@00018W03Ooml0W4P00000
000007@0_ooo0?ooo`2oM0000`0000001000B9`0gooo0?oo_`1d0002000000030000M02oool0oooo
05<0oooo002S0?ooo`030;md0000B9`0gooo0080oooo00<0_g@00018W03Oool0703oool00`3ooml0
W4P0001d_`0J0?ooo`030?oog`2LB00007Bo02X0oooo00<0gia80000M02oool0?@3oool00`2oM000
04RL0=ooo`1L0?ooo`00X`3oool00`3ookl0M4RL0=ooo`020?ooo`030?oo_`1dB9`0gooo01`0oooo
00<0oooO09a80000M;l06P3oool00`3ooml0W4P0001d_`0Z0?ooo`030=nLB00007@0_ooo03d0oooo
00<0oono07A8W03Oool0G03oool00:<0oooo00<0oooO09adW03Oool00P3oool00`3ooml0W7BL0=oo
o`0L0?ooo`030?oog`2LB00007Bo01X0oooo00<0oooO09a80000M;l0:P3oool00`3OW4P0001d0;oo
o`0m0?ooo`030?oog`2LM9`0gooo05`0oooo00360?ooo`040?oo_`1d00000000001d_aT0oooo00@0
oono07@00000000007Bo9P3oool01`3OW4P0001809cOg`2LB00000000000M02oool0W03oool00<L0
oooo00<0oooO09a80000M;l06P3oool00`3ooml0W4P0001d_`0V0?ooo`070=nLB00004P0W=oo0?oo
o`2oM000001d0;ooo`2L0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-31.0108, -0.309528, \
0.466765, 0.00839157}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotRecoilAngle[sol, values, 90. ];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.010582 0.0147151 0.00654005 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.23545 -0.0125 -8 -15.6875 ]
[.23545 -0.0125 8 0 ]
[.44709 -0.0125 -8 -15.6875 ]
[.44709 -0.0125 8 0 ]
[.65873 -0.0125 -8 -15.6875 ]
[.65873 -0.0125 8 0 ]
[.87037 -0.0125 -8 -15.6875 ]
[.87037 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.0313 -31.375 ]
[.5 -0.0125 21.0313 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .14552 -16 -7.84375 ]
[-0.0125 .14552 0 7.84375 ]
[-0.0125 .27632 -16 -7.84375 ]
[-0.0125 .27632 0 7.84375 ]
[-0.0125 .40712 -16 -7.84375 ]
[-0.0125 .40712 0 7.84375 ]
[-0.0125 .53792 -16 -7.84375 ]
[-0.0125 .53792 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -31.6875 -21.0313 ]
[-0.01875 .30902 -16 21.0313 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -29.875 0 ]
[.5 .63053 29.875 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.23545 0 m
.23545 .61803 L
s
.44709 0 m
.44709 .61803 L
s
.65873 0 m
.65873 .61803 L
s
.87037 0 m
.87037 .61803 L
s
0 .14552 m
1 .14552 L
s
0 .27632 m
1 .27632 L
s
0 .40712 m
1 .40712 L
s
0 .53792 m
1 .53792 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.23545 0 m
.23545 .00625 L
s
gsave
.23545 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.44709 0 m
.44709 .00625 L
s
gsave
.44709 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.65873 0 m
.65873 .00625 L
s
gsave
.65873 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
.87037 0 m
.87037 .00625 L
s
gsave
.87037 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.125 Mabswid
.07672 0 m
.07672 .00375 L
s
.12963 0 m
.12963 .00375 L
s
.18254 0 m
.18254 .00375 L
s
.28836 0 m
.28836 .00375 L
s
.34127 0 m
.34127 .00375 L
s
.39418 0 m
.39418 .00375 L
s
.5 0 m
.5 .00375 L
s
.55291 0 m
.55291 .00375 L
s
.60582 0 m
.60582 .00375 L
s
.71164 0 m
.71164 .00375 L
s
.76455 0 m
.76455 .00375 L
s
.81746 0 m
.81746 .00375 L
s
.92328 0 m
.92328 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.0313 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .14552 m
.00625 .14552 L
s
gsave
-0.0125 .14552 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .27632 m
.00625 .27632 L
s
gsave
-0.0125 .27632 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .40712 m
.00625 .40712 L
s
gsave
-0.0125 .40712 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
0 .53792 m
.00625 .53792 L
s
gsave
-0.0125 .53792 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04742 m
.00375 .04742 L
s
0 .08012 m
.00375 .08012 L
s
0 .11282 m
.00375 .11282 L
s
0 .17822 m
.00375 .17822 L
s
0 .21092 m
.00375 .21092 L
s
0 .24362 m
.00375 .24362 L
s
0 .30902 m
.00375 .30902 L
s
0 .34172 m
.00375 .34172 L
s
0 .37442 m
.00375 .37442 L
s
0 .43982 m
.00375 .43982 L
s
0 .47252 m
.00375 .47252 L
s
0 .50522 m
.00375 .50522 L
s
0 .57062 m
.00375 .57062 L
s
0 .60332 m
.00375 .60332 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -105.875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
82.0313 11.8438 translate
90 rotate
-82.0313 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.23545 .61178 m
.23545 .61803 L
s
.44709 .61178 m
.44709 .61803 L
s
.65873 .61178 m
.65873 .61803 L
s
.87037 .61178 m
.87037 .61803 L
s
.125 Mabswid
.07672 .61428 m
.07672 .61803 L
s
.12963 .61428 m
.12963 .61803 L
s
.18254 .61428 m
.18254 .61803 L
s
.28836 .61428 m
.28836 .61803 L
s
.34127 .61428 m
.34127 .61803 L
s
.39418 .61428 m
.39418 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.55291 .61428 m
.55291 .61803 L
s
.60582 .61428 m
.60582 .61803 L
s
.71164 .61428 m
.71164 .61803 L
s
.76455 .61428 m
.76455 .61803 L
s
.81746 .61428 m
.81746 .61803 L
s
.92328 .61428 m
.92328 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .14552 m
1 .14552 L
s
.99375 .27632 m
1 .27632 L
s
.99375 .40712 m
1 .40712 L
s
.99375 .53792 m
1 .53792 L
s
.125 Mabswid
.99625 .04742 m
1 .04742 L
s
.99625 .08012 m
1 .08012 L
s
.99625 .11282 m
1 .11282 L
s
.99625 .17822 m
1 .17822 L
s
.99625 .21092 m
1 .21092 L
s
.99625 .24362 m
1 .24362 L
s
.99625 .30902 m
1 .30902 L
s
.99625 .34172 m
1 .34172 L
s
.99625 .37442 m
1 .37442 L
s
.99625 .43982 m
1 .43982 L
s
.99625 .47252 m
1 .47252 L
s
.99625 .50522 m
1 .50522 L
s
.99625 .57062 m
1 .57062 L
s
.99625 .60332 m
1 .60332 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -90.875 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(recoil) show
93.063 15.813 moveto
(angle) show
118.750 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.06245 .57944 L
.10458 .5534 L
.14415 .52894 L
.18221 .50542 L
.22272 .48039 L
.26171 .45629 L
.30316 .43067 L
.34309 .406 L
.3815 .38225 L
.42237 .357 L
.46172 .33268 L
.49955 .30929 L
.53984 .2844 L
.57861 .26043 L
.61984 .23495 L
.65954 .21041 L
.69774 .18681 L
.73838 .16169 L
.77751 .13751 L
.81909 .11181 L
.85916 .08704 L
.89771 .06322 L
.93871 .03788 L
.97619 .01472 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00jP3oool01@3OW4P00000000000000000M;oo0100oooo00<0oooO09a8
000000000`0000001018W=l0oooo0?ooo`3OW4P30000000307Boo`3oool0oooo0:40oooo003Z0?oo
o`030=nLB00007@0_ooo0180oooo00<0_g@00018W03Oool00`3oool00`3OW9`0_ooo0?ooo`020?oo
o`030?oo_`1d0000M;oo0:<0oooo003M0?ooo`040=nLB00000000000001d_`T0oooo00<0gia80018
W03Oool04P3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0080oooo00<0oono07@0
001d_ol0X`3oool00>00oooo00<0okmd001d_`3oool01`3oool00`3OW4P004RL0=ooo`0C0?ooo`03
0;md00000000000000<0000000L0B9cO0?ooo`3oool0oooo0?oo_`1d0000M;oo0:<0oooo003E0?oo
o`050?oo_`1d000000000000B02Lgol01P3oool00`3ookl0M01d0;ooo`070?ooo`030=nLB000B9`0
gooo0080oooo00P0oooO09a80000000000000000B018B00004RL0=ooo`80oooo00P0_g@000000000
0000001d0;ooo`3oool0oono07@000@0000000<0001d0;ooo`3oool00P3oool00`3ookl0M00007Bo
o`2S0?ooo`00e@3oool01P2oM00007Bo0?ooo`3ogi`0B01809cOo`D0oooo00<0gia80000M02oool0
1`3oool0603OW4P004RL0=ooo`3oool0oooO09a800000000M;oo0?ooo`3ookl0M0000000000007@0
_ooo0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0_gA809cOo`P0oooo00<0oono07@0001d
_ol0X`3oool00=@0oooo00<0oooO09a8001d_ol00P3oool00`3OW4P004RL0=ooo`020?ooo`040?oo
_`1d0000000004RLg`P0oooo00L0gia80018W03Oool0oooo0?oOW0180000B9cO0080oooo00L0oooO
09a80000M;l0oooo0?oOW0180000B9cO00<0oooo00D0omnL07BLg`3oool0oono07@0000200000003
001d_`3oool0oooo00<0oooo00<0oono07@0001d_ol0X`3oool00=@0oooo00<0oono07@0001d_ol0
0P3oool00`3ookl0M00007Boo`030?ooo`030?oog`2LW;l0oooo00P0oooo00L0gia80018W03Oool0
oooo0?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?noM00004P0W=oo00H0oooo00H0
gia80000M02oool0oono07@00018W=l40?ooo`030?oo_`1d0000M;oo0:<0oooo003D0?ooo`030?oo
_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool00`3oool00`3o_g@0B9cO0?ooo`070?ooo`070=nL
B000B9`0gooo0?ooo`3o_g@0000007Boo`020?ooo`070?oog`2LB00007Bo0?ooo`3o_g@0001d0;oo
o`050?ooo`070?oo_`1d0000B9cO0?ooo`3ooml0W4P0001d_`040?ooo`030?oo_`1d0000M;oo0:<0
oooo003D0?ooo`030?oog`2LB00007Bo00<0oooo00P0gia80000M02oool0omnL07BLg`3oool0gia8
001d_`P0oooo00L0gia80018W03Oool0oooo0?oo_`1d0000B9cO0080oooo00H0oooO09a80000M;l0
oooo0?oOW01800050000000804RLg`3ookl0M00004RLg`3oool0oooO09a80018W=l40?ooo`030?oo
_`1d0000M;oo0:<0oooo003E0?ooo`030;md00000000000000@0000000H0M;oo0?ooo`3OW4P00000
0000B02Lgol80?ooo`030=nLB000B9`0gooo0080oooo00L0okmd0000B02Lgol0oooo0?noM0000000
07Bo0080oooo00`0_g@004RLg`3oool0oooo0=nLB00004P0W=oo0?ooo`2oM00007Bo0?ooo`3o_g@2
000000070018W03Oool0oooo0?ooo`3ookl0M00007Boo`2S0?ooo`00e@3oool00`3OW4P0001d0;oo
o`020?ooo`030?oo_`1d0000B9cO00d0oooo00<0gia80018W03Oool00`3oool00`3o_g@000000000
000200000005001d_`3oool0oooo0?ooo`3OW4P00P0000001P0004P0W=oo0?ooo`3oool0oooO09a8
0080000000<0M;oo0?ooo`3oool00`3oool00`3ookl0M00007Boo`2S0?ooo`00e@3oool00`3ogi`0
B00007Boo`030?ooo`030;md0000M;l0oooo00`0oooo00<0gia80018W03Oool01P3oool00`3ooml0
W4P0001d_`0D0?ooo`030?oo_`1d0000M;oo0:<0oooo003F0?ooo`030;md0000B9`0gooo0080oooo
00<0gia8001d_`3oool0303oool00`3OW4P004RL0=ooo`060?ooo`030?oog`2LB00007Bo01@0oooo
00<0oono07@0001d_ol0X`3oool00=H0oooo00<0oono07@0001d_ol00P3oool00`3OW4P007Bo0?oo
o`0<0?ooo`030=nLB000B9`0gooo00H0oooo00<0oooO09a80000M;l0503oool00`3ookl0M00007Bo
o`2S0?ooo`00e`3oool01P3o_g@0001809cOo`3oool0_g@007Boo`d0oooo00<0gia80000M02oool0
1@3oool0103ookl0M00000000000M;lD0?ooo`030?oo_`1d0000M;oo0:<0oooo003H0?ooo`050?oO
W018000000000018W03Oool03@3oool01@3OW4P00000000000000000M;oo00@0oooo00<0oooO09a8
0000M;l04P3oool01@3OW4P00000000000000000M;oo0:<0oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo001N0?ooo`05
0?oOW018000000000000M02oool0=@3oool00`2oM0000000000000040000000904RLg`3oool0oooo
0?ooo`3ogi`0B0000000000007@0_ooo03@0oooo00@0oooO09a80000B9`0gooo0`3oool01@3ogi`0
B0000000000007@0_ooo03<0oooo00D0omnL04P000000000000007Boo`040?ooo`050?oOW0180000
00000000M02oool0<P3oool01@3OW4P00000000000000000B9cO00@0oooo00D0omnL04P000000000
001d0;ooo`170?ooo`00G@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0=@3oool00`3O
W4P00000000000030000000:0018W03Oool0oooo0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oooc<0
oooo00@0oooO09a80000B9`0gooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0<@3o
ool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool0<03oool01`3OW4P0000007Boo`3oool0oooO09a80018W=l00P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0AP3oool005d0oooo00<0_g@00018W03Oool00P3oool00`2oM000
07Bo0?ooo`0e0?ooo`030=nLB018W=l0oooo0080oooo00H0giad09cOo`3oool0_g@00018W03Oool2
0?ooo`030;md0000M;l0oooo0380oooo00@0oooO09a80000B9`0gooo0P3oool00`2oM00004RL0=oo
o`020?ooo`030;md0000M;l0oooo02l0oooo00<0oooO09a80000M;l00`3oool01`3o_g@004RL0=oo
o`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0^0?ooo`030?oog`2LB00007Bo00<0
oooo00L0okmd0018W03Oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0A@3oool0
05`0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0f0?ooo`030=nLB01d_ol0oooo
00<0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0^0?ooo`030;md000000000000
00D0000000@0M;oo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0_0?ooo`030?oOW0180000
M;oo00<0oooo00H0okmd0000B02Lgol0oono07@00018W=l30?ooo`030=nLB00007@0_ooo02l0oooo
00<0_g@0001d_`3oool00P3oool01P2oM00004RL0=ooo`3ookl0M00004RLg`<0oooo00<0gia80000
M02oool0A@3oool005`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0g0?ooo`03
0;mdB02Lgol0oooo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0^0?ooo`0<
0?oOW01dW=l0oooo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0
001809cOo`0_0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol3
0?ooo`030?noM00004P0W=oo02l0oooo00/0omnL04P0M02oool0oooo0;md00000000M;oo0?ooo`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo04D0oooo001L0?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol0=`3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol0;`3oool02`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0oooo0?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol0;`3oool00`3ogi`0B00007Boo`030?ooo`060;md
000004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0`0?ooo`050?oo_`1dB4P0B000
00000018W=l00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo04D0oooo001L0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0>03oool01`3o_g@007Bo0?ooo`3oool0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0_0?ooo`0;0?oog`2LB7@0_ooo0?oog`2LB000
04RL0=ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0_0?ooo`0<0?oo_`1d0000
000004RLg`3oool0okmd00000000M;l0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0`
0?ooo`040?oOW01800000000001d_`<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`150?ooo`00G03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo03P0oooo00L0oono
07@0B02Lgol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0<03oool02P3ogi`0B7Bo
0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0300oooo00<0
_g@00018W02oW4P00P0000001P00B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02L
gol0;`3oool01P3ogi`0B0000000M02oool0okmd001d_`80oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`150?ooo`00G03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
03P0oooo00L0oono07@00018W=l0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0<@3o
ool02@3OW4P0M;oO09a80000B9`0gooo0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
0300oooo00<0omnL04P0001d_ol01@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo
02l0oooo00<0_g@00000M02oool00P3oool01P2oM000B9cO0?ooo`3ookl0M00004RLg`<0oooo00<0
gia80000M02oool0A@3oool005d0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0c
0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo0340oooo00D0oooO09a8B018B00004RL0=ooo`020?ooo`030=nLB000
B9`0gooo0080oooo00<0_g@0001d_`3oool0<@3oool00`3OW4P0001809cOo`050?ooo`030=nLB000
B9`0gooo0080oooo00<0_g@0001d_`3oool0;`3oool00`2oM00004RL0=ooo`020?ooo`070=nLB000
M;l0oooo0?ooo`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo04D0oooo001M0?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0d0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000
M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0c0?ooo`040?oOW018000004RL
0=ooo`80oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo03<0oooo00<0okmd00000000M;l0
103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0<03oool01`3o_g@0000007Boo`3oool0
oono07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0AP3oool005h0oooo
00D0oono07@000000000001d0;ooo`0f0?ooo`060?oog`2LB00000000000000007@0_ooo103oool0
1@3ookl0M0000000000007@0_ooo03D0oooo00<0gia80018W03Oool00`3oool01@3ookl0M0000000
000007@0_ooo03D0oooo00D0oooO09a800000000001d0;ooo`020?ooo`050?oo_`1d000000000000
M02oool0<P3oool01@3o_g@00000000000000000B9cO00@0oooo00D0oono07@000000000001d0;oo
o`170?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00C03oool01@3ogi`0B000
0000000007@0_ooo00T0ooooo`00000a000002H0oooo001;0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0T0?ooo`00B`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02@0
oooo001:0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo
0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo
0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool004X0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo04@0oooo00<0
001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00BP3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00L0oooo0`0000140?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01d0oooo00<00000
0?ooo`3oool00P3oool3000002H0oooo001:0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol01`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?oo
o`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01/0oooo0P0000070?oo
o`030000003oool0oooo02@0oooo001:0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0T0?ooo`00BP3oool00`3ookl0M00004RLg`030?ooo`030=nL
B00007@0_ooo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000
O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0H0?ooo`800000
2P3oool00`000000oooo0?ooo`0T0?ooo`00B`3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000
M;l0oooo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0F0?ooo`800000303o
ool00`000000oooo0?ooo`0T0?ooo`00B`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
203oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo01D0oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`0T0?ooo`00C03oool01@3ookl0M0000000000007@0_ooo00T0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0C0?ooo`8000003`3oool00`000000oooo
0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0180oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`14
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo0100oooo0P00000B0?ooo`030000003oool0oooo02@0oooo001J0?oo
o`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool03P3oool2000001@0oooo00<000000?oo
o`3oool0903oool005X0oooo0P0000150?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo00d0oooo00<000000?ooo`3o
ool04`3oool2000002H0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0
?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
2`3oool2000001L0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0903o
ool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`080?ooo`8000006P3oool0
0`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo
03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
00H0oooo0P00000L0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo04@0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool01@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02@0
oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo
0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool00`3oool2000001l0oooo
00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?oo
o`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`020?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0903oool005X0oooo00<00000
0?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0200000280oooo00<000000?ooo`3oool0903oool0
05X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l000000000000T0?ooo`030000003oool0oooo
02@0oooo001J0?ooo`800000A@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0
oooo00<0001o0?ooo`3oool0?@3oool00`000000oooo0?ooo`0S0?ooo`8000009P3oool005X0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo
00<0001o0?ooo`3oool0>`3oool2000000030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0
903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo03h0oooo00<0001o0?ooo`3oool0>P3oool010000000oooo0?ooo`0007lV0?ooo`030000
003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03P0oooo0P0000030?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0=P3oool2
000000D0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`00
0000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`00
07l0oooo0?ooo`0d0?ooo`8000001`3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo
02@0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0n0?ooo`030000O`3oool0oooo0380oooo0P0000090?ooo`030000O`3oool0oooo02@0
oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo
0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0<03oool2000000/0oooo
00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?oo
o`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?oo
o`0^0?ooo`8000003@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J
0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n
0?ooo`030000O`3oool0oooo02d0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0T
0?ooo`030000003oool0oooo02@0oooo001J0?ooo`800000A@3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0:`3oool200000100oooo00<0001o0?oo
o`3oool08`3oool2000002H0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo02X0oooo00<000000?ooo`3o
ool0403oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000
003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000
O`3oool0oooo02P0oooo0P00000C0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0
903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo03h0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0C0?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool09@3oool2
000001H0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00@@3oool00`2o
M0000000000000040000000904RLg`3oool0oooo0?ooo`3ogi`0B0000000000007@0_ooo00T0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo
00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo02@0oooo
00<000000?ooo`3oool0903oool00480oooo00<0gia80000000000000`0000002P00B9`0gooo0?oo
o`3ookl0M00007Boo`3oool0okmd0000M02oool80?ooo`030000003oool0oooo04@0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo0280oooo0P00000I
0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool004<0oooo00<0gia804RL
g`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo00<0_g@0001d_`3oool01`3o
ool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3o
ool00`0007l0oooo0?ooo`0Q0?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool0903o
ool00`000000oooo0?ooo`0T0?ooo`00A03oool00`3OW4P0M;oo0?ooo`030?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0O0?ooo`800000703oool0
0`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo00150?ooo`030;mdB02Lgol0oooo
0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo
04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo
01h0oooo00<000000?ooo`3oool0703oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo
02@0oooo00150?ooo`030?oog`2LB7@0_ooo0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`070?ooo`<00000i00007l2000004@0001o0`00000V0?ooo`00AP3oool01`3o_g@007Bo
0?ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo
04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo
01/0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo
02@0oooo00160?ooo`070?oo_`1d04P0W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool06@3oool200000280oooo00<0001o0?ooo`3oool0903oool0
0`000000oooo0?ooo`0T0?ooo`00AP3oool01`3ookl0M00004RLg`3oool0oono07@00018W=l00`3o
ool00`3OW4P0001d0;ooo`070?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool08P3o
ool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo00110?ooo`030?noW02Lgol0
oooo0080oooo00<0omnL04P00018W=l00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0
oooo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool05P3oool2000002D0oooo00<0001o0?ooo`3oool0903oool0
0`000000oooo0?ooo`0T0?ooo`00@@3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`140?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0E0?oo
o`030000003oool0oooo02D0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?oo
o`00@P3oool01P3ooml0W4P0000000000000001d0;ooo`@0oooo00D0oono07@000000000001d0;oo
o`090?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo01<0oooo0P00000X0?ooo`030000O`3oool0oooo02@0oooo00<0
00000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool04@3oool2000002X0oooo00<0001o
0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0@0?oo
o`030000003oool0oooo02X0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?oo
o`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool0?P3oool00`0007l0oooo0?ooo`0>0?ooo`800000;@3oool00`0007l0oooo0?ooo`0T0?ooo`03
0000003oool0oooo02@0oooo001J0?ooo`800000A@3oool00`0007l0oooo0?ooo`0m0?ooo`030000
O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0303oool2000002l0oooo00<0001o0?ooo`3oool0
8`3oool2000002H0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00X0oooo0P00000a0?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool02@3oool0
0`000000oooo0?ooo`0a0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool0
05X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
03h0oooo00<0001o0?ooo`3oool01`3oool2000003@0oooo00<0001o0?ooo`3oool0903oool00`00
0000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`050?ooo`800000=P3oool00`0007l0
oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo04@0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo00@0oooo
00<000000?ooo`3oool0=P3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo
001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo0080oooo0P00000i0?ooo`030000O`3oool0oooo02@0oooo00<0
00000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool00P00000k0?ooo`030000O`3oool0
oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`00
07l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o000000000000?@3oool00`00
07l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo04@0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000003oool0oooo03d0
oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool2000004D0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0l0?ooo`80000000<0001o0?ooo`3oool0
?@3oool00`0007l0oooo0?ooo`0S0?ooo`8000009P3oool005X0oooo00<000000?ooo`3oool0A03o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03/0oooo00@000000?ooo`3oool0001o
?`3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000^0?ooo`030?oOW018
0000000000<0000000<007Bo0?ooo`3oool08`3oool00`000000oooo0?ooo`140?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool0>@3oool2000000<0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000/0?ooo`030=nLB0000000
000000H0000000<004RL0=ooo`3oool08P3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool0>03oool00`000000oooo0?ooo`030?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00:P3oool01P3o
okl0M0000000000007@0_ono09bLg`D0oooo00<0_gAd0;ooo`3oool08P3oool00`000000oooo0?oo
o`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0=P3oool2000000H0oooo00<0
001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000Y
0?ooo`030?oOW01804P0W=oo0080oooo00<0oono09bLg`3oool0103oool00`2oW9`0gooo0?ooo`0R
0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0e
0?ooo`030000003oool0oooo00H0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T
0?ooo`030000003oool0oooo02@0oooo000Y0?ooo`030;nLW03Oool0oooo0080oooo00<0oono09bL
g`3oool00P3oool00`3ookl0M00007Boo`0T0?ooo`030000003oool0oooo04@0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`0c0?ooo`8000002@3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool002P0oooo00<0oono07Ad_`3o
ool00`3oool00`3ookl0M000000000030000000304RLg`3oool0oooo02<0oooo00<000000?ooo`3o
ool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0380oooo00<000000?ooo`3o
ool02@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3o
ool0903oool002P0oooo00<0oooO09a8000000001`0000000`00B9`0gooo0?ooo`0U0?ooo`030000
003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0`0?ooo`800000
303oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0
903oool002T0oooo00<0oono07@000000000100000000`1d_ol0oooo0?ooo`0W0?ooo`800000A@3o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02h0oooo0P00000>0?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool08`3oool2000002H0oooo000d0?ooo`040?oOW0180000
04RL0=ooob80oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo02d0oooo00<000000?ooo`3oool03P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool0903oool003<0oooo00D0gia80018W03Ogi`0B4RL0=oo
o`0R0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0[0?ooo`8000004@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<0
00000?ooo`3oool0903oool00380oooo00H0_g@004RLg`3oool0omnL04Q8W03OoolR0?ooo`030000
003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0Z0?ooo`030000
003oool0oooo0140oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000
003oool0oooo02@0oooo000a0?ooo`070=nLB01d_ol0oooo0?ooo`3ogi`0B4RL0=ooo`0R0?ooo`03
0000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0X0?ooo`80
0000503oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3o
ool0903oool00300oooo00<0gia80000000000001@0000000`0007@0_ooo0?ooo`0:0?ooo`040?oo
g`2LB00004RL0=ooo`<0oooo00D0omnL04P000000000001d0;ooo`090?ooo`030000003oool0oooo
04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0V0?ooo`8000005P3oool00`00
07l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool003D0
oooo00<0omnL04Q8W03Oool03@3oool0103ooml0W4P00018W03Oool20?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo00150?ooo`040?oog`2LB000
04RL0=ooo`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003o
ool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0S0?ooo`8000006@3o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903o
ool00440oooo00<0_g@00000000000001@000000101d_ol0oono07@00018W=l30?ooo`030=nLB000
07@0_ooo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo0280oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool0903oool00440oooo00`0omnL07BLg`3oool0oooo0?oo
g`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00L0oooo00<00000
0?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0200oooo0P00000L
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T
0?ooo`00@P3oool02`3OW4P0M;oo0?ooo`3ooml0W4P00018W03Oool0oooo0?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01`3oool300000:D0001o0P0000230000O`<000009P3oool00480oooo
00/0oooO09a8M02oool0oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo01d0oooo00<000000?ooo`3oool07P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool0903oool004<0oooo00X0omnL04Qd_`3ooml0W4P00018
W03Oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo
04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0K0?ooo`8000008@3oool00`00
07l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool002P0
oooo00<0omnL04P0000000003P0000000`0007@0_ooo0?ooo`080?ooo`090=nLB01d_ml0W4P00018
W03Oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?oo
o`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?oo
o`0Q0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?oo
o`0T0?ooo`00:03oool00`3ogi`0M9cO0?ooo`0>0?ooo`030;mdM02oool0oooo00P0oooo00D0oooO
09a8B018B00004RL0=ooo`020?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3o
ool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0603o
ool2000002@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003o
ool0oooo02@0oooo000X0?ooo`030?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3oool02@3oool0
103ogi`0B0000018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`03
0000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0F0?ooo`80
00009P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3o
ool0903oool002P0oooo00<0omnL09boo`3oool03P3oool00`3OW9`0_ooo0?ooo`0:0?ooo`030=nL
B000B9`0gooo00<0oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo04@0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0D0?ooo`800000:03oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo
00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo0180oooo
0P00000Z0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo
0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0403oool2000002`0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?oo
o`0T0?ooo`030000003oool0oooo02@0oooo000`0?ooo`050;md00000000000000000018W=l09@3o
ool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool03P3o
ool2000002h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003o
ool0oooo02@0oooo000^0?ooo`030?oo_`1d0000000000@0000000<0B9cO0?ooo`3oool08P3oool2
000004D0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0=0?ooo`030000003oool0
oooo02h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0S0?ooo`8000009P3oool0
02h0oooo00<0gia8001d_`3oool00P3oool0103ookl0M0000000M02ooolS0?ooo`030000003oool0
oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0;0?ooo`800000<@3oool0
0`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool0
02h0oooo00<0_gAd0;ooo`3oool00`3oool00`3ookl0M01d0;ooo`0S0?ooo`030000003oool0oooo
04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo
0340oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo
02@0oooo000Y0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL09boo`3oool00`3oool00`3o_i`0
W=oo0?ooo`0S0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`090?ooo`030000003oool0oooo0380oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000Y0?ooo`030;md00000000000000T00000
00<0001d0;ooo`3oool08P3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool01`3oool2000003D0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo
0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oo_`1d0000000000X0000000<0
07Bo0?ooo`3oool08P3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00=03oool00`3ogi`0W;oo0?ooo`0S
0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`05
0?ooo`030000003oool0oooo03H0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T
0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`030?ooo`800000>@3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool002l0oooo00<0oono07@00000
00000P0000000`18W=l0oooo0?ooo`0S0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo03T0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000^0?ooo`030?oO
W0180000000000@0000000<0B9cO0?ooo`3oool08P3oool00`000000oooo0?ooo`140?ooo`030000
O`3oool0oooo03d0oooo00<0001o0?ooo`3oool00P00000l0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00;P3oool02@3OW7@0W=oo0?no
M01d_ol0oooo0?noM0000000001d0;ooo`0S0?ooo`800000A@3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`000000000003h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0S0?oo
o`8000009P3oool002h0oooo00@0_gA809cOo`3o_g@0M;oo0P3oool00`3OW4P0001d0;ooo`0S0?oo
o`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool00`000000oooo0?ooo`0n0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?oo
o`00;P3oool0103o_g@000000000001d_ol20?ooo`030;md0000M;l0oooo02<0oooo00<000000?oo
o`3oool0A03oool00`0007l0oooo0?ooo`0k0?ooo`80000000<0001o0?ooo`3oool0?P3oool00`00
07l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool002l0
oooo00H0gia80000001d_ol0okmd00000000M;lU0?ooo`030000003oool0oooo04@0oooo00<0001o
0?ooo`3oool0>@3oool200000080oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?oo
o`3oool0A03oool00`0007l0oooo0?ooo`0g0?ooo`800000103oool00`0007l0oooo0?ooo`0n0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?oo
o`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo03H0oooo00<000000?ooo`3o
ool0103oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3o
ool0903oool00`000000oooo0?ooo`0T0?ooo`00;`3oool02`3o_g@0000000000000B9`0gono07@0
0000M;l0gia80000000007@0_ooo0200oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?oo
o`0d0?ooo`8000001`3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00;P3oool00`3o_g@0000000000002
000000070000M02o_g@0000004RLg`3oool0_g@0001d_`0P0?ooo`030000003oool0oooo04@0oooo
00<0001o0?ooo`3oool0<`3oool00`000000oooo0?ooo`070?ooo`030000O`3oool0oooo03h0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo
000^0?ooo`0=0;mdM02oool0oooo0?ooo`3ooml0W4Qd0;oOW01807@0_ooo0?ooo`3o_g@004RL0=oo
o`0O0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0<@3oool2000000X0oooo00<0
001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<0
00000?ooo`3oool0903oool002h0oooo00<0_gA809cOo`3oool00P3oool0202oM7@0_mnL04P0B02L
gol0oooo0?oOW018B9`0gooo7`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo0300
oooo00<000000?ooo`3oool02P3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00;P3oool00`3ogi`0B000
0000000200000007001d_`3ogi`0B01809cOo`3oool0okmd001d_`0P0?ooo`800000A@3oool00`00
07l0oooo0?ooo`0^0?ooo`8000003@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo
03d0oooo00<0001o0?ooo`3oool08`3oool2000002H0oooo000^0?ooo`0<0?oo_`1d000000000000
0000B9`0gooo0?oOW0180000M;oo0?oog`2LB4P0W=oo803oool00`000000oooo0?ooo`140?ooo`03
0000O`3oool0oooo02d0oooo00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0n0?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00
;P3oool00`3ookl0W9cO0?ooo`040?ooo`040;md00000000001809cOob40oooo00<000000?ooo`3o
ool0A03oool00`0007l0oooo0?ooo`0[0?ooo`800000403oool00`0007l0oooo0?ooo`0n0?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00
FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo02X0oooo00<000000?ooo`3oool0
403oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3o
ool0oooo02P0oooo0P00000C0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oOW02L_ol0
oooo00h0oooo00<0gibL0;ooo`3oool01`3oool01@3ogi`0B000000000000000M;oo00@0oooo00D0
omnL04P000000000001d0;ooo`090?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0
9P3oool2000001D0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000
O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool002P0oooo00<0omnL09boo`3oool03P3o
ool00`3OW9`0_ooo0?ooo`060?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00007Boo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo04@0oooo00<0001o
0?ooo`3oool09@3oool00`000000oooo0?ooo`0E0?ooo`030000O`3oool0oooo03h0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo000X0?oo
o`030?oOW01dW=l0oooo00h0oooo00<0_gAd0;ooo`3oool01@3oool00`3ooml0W4P0001d_`030?oo
o`070?noM000B9`0gooo0?ooo`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0
00000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0S0?ooo`800000603oool00`0007l0oooo0?oo
o`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?oo
o`0T0?ooo`00:03oool00`3ogi`0B0000000000>000000030000M02oool0oooo00D0oooo00<0omnL
04P0001d_ol00`3oool01P3o_g@0001809cOo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0
1`3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo0240oooo0P00000J0?ooo`030000
O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000
003oool0oooo02@0oooo00110?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL
04P0001d_ol30?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0
oooo0?ooo`0O0?ooo`800000703oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00@@3oool00`3ogi`0B000
07Boo`030?ooo`060;md000004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`070?oo
o`<00000I@0007l00`000000001o0000O`320000O`<000009P3oool00440oooo00`0oono07@00000
0000B9cO0?ooo`3o_g@00000001d_`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00L0
oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0L0?ooo`8000007`3oool00`0007l0
oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000
oooo0?ooo`0T0?ooo`00@P3oool00`2oM00004RL0;nLB002000000060018W03Oool0oooo0?oOW018
0000M;oo0`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo04@0oooo00<0001o0?oo
o`3oool06P3oool200000240oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool00480oooo00<0omnL04P0001d
_ol01@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3o
ool0A03oool00`0007l0oooo0?ooo`0I0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3o
ool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3o
ool0903oool004<0oooo00<0gia80000B02Lgol01@3oool00`3OW4P004RL0=ooo`020?ooo`030;md
0000M;l0oooo00L0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0G0?ooo`800000
903oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
903oool00`000000oooo0?ooo`0T0?ooo`00A03oool00`3o_g@00000001d_`040?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3o
ool05@3oool2000002H0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`03
0000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool004D0oooo00D0oooO09a800000000
001d0;ooo`020?ooo`050?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`140?oo
o`030000O`3oool0oooo01<0oooo0P00000X0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`03
0000003oool0oooo04@0oooo00<0001o0?ooo`3oool04P3oool00`000000oooo0?ooo`0X0?ooo`03
0000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`03
0000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0
403oool2000002/0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000
O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03o
ool00`0007l0oooo0?ooo`0>0?ooo`800000;@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3o
ool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool2
000004D0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0S0?ooo`8000009P3oool0
05X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`0;0?ooo`800000<03oool00`00
07l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`00
0000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`030000O`3oool0oooo00X0
oooo00<000000?ooo`3oool0<03oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo
0?ooo`140?ooo`030000O`3oool0oooo00P0oooo0P00000c0?ooo`030000O`3oool0oooo03h0oooo
00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo
001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool01P3oool2000003D0oooo00<0
001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<0
00000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`05
0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?oo
o`3oool0A03oool00`0007l0oooo0?ooo`030?ooo`800000>03oool00`0007l0oooo0?ooo`0n0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?oo
o`00FP3oool00`000000oooo0?ooo`140?ooo`040000O`3oool0oooo0?ooo`800000>P3oool00`00
07l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`00
0000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`140?ooo`040000O`3oool0oooo0000
03`0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0A03oool00`0007l0
00000000000m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`800000A03oool2000003l0oooo
00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02<0oooo
0P00000V0?ooo`00FP3oool00`000000oooo0?ooo`110?ooo`80000000<0oooo0000O`3oool0?P3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903o
ool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`100?ooo`050000003oool0
oooo0?ooo`0007l0?`3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0n
0?ooo`800000103oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?oo
o`030000003oool0oooo03d0oooo00<000000?ooo`3oool0103oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?oo
o`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo03/0oooo0P0000070?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`03
0000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0
>P3oool00`000000oooo0?ooo`070?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0
903oool005X0oooo00<000000?ooo`3oool0>03oool2000000X0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903o
ool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo00X0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`00
0000oooo0?ooo`0e0?ooo`8000003@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo
03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo
02@0oooo001J0?ooo`800000=@3oool00`000000oooo0?ooo`0=0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02<0
oooo0P00000V0?ooo`00FP3oool00`000000oooo0?ooo`0b0?ooo`800000403oool00`0007l0oooo
0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo
0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo0300oooo0P00
000B0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?oo
o`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0B0?ooo`030000O`3oool0oooo03d0oooo00<0001o
0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<00000
0?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0;@3oool2000001D0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?oo
o`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00@`3oool01@3OW4P00000000000000000B9cO
00@0oooo00D0omnL04P000000000001d0;ooo`090?ooo`030000003oool0oooo02/0oooo0P00000G
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool00480oooo00L0gia80000
001d_ol0oooo0?oog`2LB000B9cO0080oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0
oooo00<000000?ooo`3oool0:@3oool2000001T0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0
oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000
oooo0?ooo`0T0?ooo`00@@3oool00`3ooml0W4P0001d_`030?ooo`070?noM000B9`0gooo0?ooo`2o
M00004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3oool0:03oool00`00
0000oooo0?ooo`0I0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`00
07l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool00480
oooo00<0_g@0001d_`3oool00P3oool01P2oM00004RL0=ooo`3ookl0M00004RLg`<0oooo00<0gia8
0000M02oool01`3oool00`000000oooo0?ooo`0V0?ooo`800000703oool00`0007l0oooo0?ooo`0m
0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T
0?ooo`030000003oool0oooo02@0oooo00120?ooo`0;0?oOW01807@0_ooo0?ooo`2oM000000007Bo
o`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo02D0
oooo00<000000?ooo`3oool0703oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0
oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0
oooo00130?ooo`050?oo_`1dB4P0B00000000018W=l00P3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00L0oooo0`00000S0000O`800000o`0007l60000O`<000009P3oool004<0oooo
00@0omnL04P00000000007Bo0`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0
oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo03d0
oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0
oooo00<000000?ooo`3oool0903oool00480oooo00H0omnL04P0000007@0_ooo0?noM000M;l20?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0P0?oo
o`8000008P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?oo
o`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo00120?ooo`03
0;md000007@0_ooo0080oooo00H0_g@004RLg`3oool0oono07@00018W=l30?ooo`030=nLB00007@0
_ooo00L0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0R0?ooo`030000O`3oool0
oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo02@0oooo00<000000?ooo`3oool0903oool00480oooo00<0_g@00018W03Oool00P3oool01`3O
W4P007Bo0?ooo`3oool0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003o
ool0oooo01d0oooo0P00000U0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903o
ool00480oooo00L0okmd0000001d_ol0oooo0?oo_`1d0000M;oo0080oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0U0?oo
o`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool004<0oooo00D0okmd00000000
0000000004RLg`040?ooo`050?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`0J
0?ooo`800000:03oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o
0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?oo
o`030000003oool0oooo01P0oooo0P00000Z0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?oo
o`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?oo
o`3oool0903oool005X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0Z0?ooo`03
0000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`03
0000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0
5@3oool2000002d0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000
O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3o
ool2000001@0oooo0P00000_0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3o
ool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02<0oooo0P00000V0?ooo`00FP3oool0
0`000000oooo0?ooo`0A0?ooo`800000<@3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0
oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0
oooo02@0oooo001J0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0<@3oool00`00
07l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`00
07l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo00h0
oooo0P00000d0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0?P3oool00`0007l0
oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0903oool005X0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo03d0oooo
00<0001o0?ooo`3oool0?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo
00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool02`3oool2000003L0oooo00<0
001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0
001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0n
0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool00`000000oooo0?ooo`0T
0?ooo`00FP3oool00`000000oooo0?ooo`080?ooo`800000>P3oool00`0007l0oooo0?ooo`0m0?oo
o`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3oool0?@3oool00`0007l0oooo0?ooo`0T0?oo
o`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0>P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo03h0oooo00<0001o0?ooo`3o
ool0?@3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000
003oool0oooo00D0oooo0P00000m0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0
?P3oool00`0007l0oooo0?ooo`0m0?ooo`030000O`3oool0oooo02@0oooo00<000000?ooo`3oool0
903oool005X0oooo0P0000050?ooo`030000003oool0oooo03d0oooo00<0001o0?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool08`3o
ool2000002H0oooo001J0?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0?@3oool0
0`0007l0oooo0?ooo`0n0?ooo`030000O`3oool0oooo03d0oooo00<0001o0?ooo`3oool0903oool0
0`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0903oool005X0oooo
o`00000a000002H0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?ooo`L0oooo00<0oooO09a8000000000`000000
0`18W=l0oooo0?ooo`2Q0?ooo`00o`3oool70?ooo`030;md0000B9`0gooo00<0oooo00<0gibL0;oo
o`3oool0X@3oool00?l0oooo1`3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0:40
oooo003o0?ooo`P0oooo00<0_g@00000000000000`0000000`18W=l0oooo0?ooo`2P0?ooo`00b03o
ool01@2oM00000000000000004P0W=oo00<0oooo00D0_g@0000000000000001d0;ooo`030?ooo`05
0;md0000000000000000B02Lgol00`3oool01@3OW4P00000000000000000M;oo0080oooo00D0gia8
000000000000001d09bLB002000000030000M02oool0oooo00@0oooo00T0oooO09a8000000000000
0018W02LM0000000001d_`2oM0000P0000000`0007@0_ooO09a80002000000030018W03Ookl0M000
00@0000000<0001d0;ooo`3OW4P00P000000100007@0_ooo0?ooo`2oM002000000030000M02oool0
oooo09H0oooo00390?ooo`030=nLB00007@0_ooo00<0oooo01L0_g@000000000M;l0oooo0?oOW018
07@0_ooo0?ooo`2oM00000000018W03Oool0oooO09a8M02oool0oooo0=nLB0000000B9cO0?ooo`3o
oml0W4P004RLg`020?ooo`070=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`060?ooo`0@0?noM000
0000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`80
oooo00<0_gA809cOo`3oool01@3oool02`3OW4P0001d0;ooo`3oool0_g@000000000M;l0oooo0?oO
W01807@0_ooo09H0oooo00390?ooo`030=nLB00007@0_ooo0080oooo00<0omnL04P00018W=l00`3o
ool01@3ogi`0M9cO0?oOW018000007Bo00<0oooo00D0omnL09boo`3ogi`0B000001d_`030?ooo`0;
0?noM00007@0_ooo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02oool01P3oool01`3o_g@00018
09cOo`3oool0oooO09a80000M;l00P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3o
ool01@3ookl0M00000000000000007Bo00<0oooo00H0gia80000M02oool0omnL04P00018W=l30?oo
o`030?oOW01dW=l0oooo09D0oooo00390?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02L
gol01@3oool00`3o_g@0001809cOo`050?ooo`030?noM0000000M;oo00<0oooo00/0okmd0000001d
_ol0oooo0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`070?ooo`060?noM000000007Bo0?oog`2L
B00007Bo0P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool01P3OW4P0001d0;oo
o`3ookl0M00004RLg`80oooo00H0gia80000M02oool0okmd0000B02LgonK0?ooo`00b@3oool00`3O
W4P0001d0;ooo`020?ooo`030?noM00007@0_ooo00D0oooo00<0okmd0000M02oool01@3oool00`3o
_g@0000007Boo`030?ooo`0;0?noM0000000M;oo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02o
ool02@3oool0103o_g@0000000000000M;l20?ooo`0?0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d
_`3oool0oono07@00018W=l0oooo0?oog`2LB00007Bo0080oooo00H0gia80000M02oool0okmd0000
M02ooonK0?ooo`00b@3oool00`3OW4P0001809cOo`020?ooo`030?oOW0180000000000@0000000@0
B9cO0?oOW01804P0W=oo0P3oool01P3o_g@0001809cOo`3ookl0M00007Boo`<0oooo00/0_g@00000
B02Lgol0oooo0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`060?ooo`070?oo_`1d0000M;oo0?oo
o`3ooml0W4P0001d_`020?ooo`0?0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0oono07@0
0018W=l0oooo0?oog`2LB000B9cO0080oooo00D0gia80000M02oool0omnL04P000050000000304RL
g`3oool0oooo09@0oooo00370?ooo`0:0?oog`2LB00000000000B018M9`0gkmd0000M02oool0_g@0
04RLg`80oooo01`0gia80000B02Lgol0oooo0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3oool0
gia80018W03Oool0oooo0?noM000000007Bo0?oog`2LB00000000000M02oool0oooo0=nLB00007@0
_ooo1P3oool0403ookl0M00004RLg`3oool0oono07@00000M;l0oooO09a80000000004Q804P0B02L
gol0oono07@00018W=l20?ooo`040;md0000M;l0oooo0?noM080000000P004RL0=ooo`3OW4P0001d
0;ooo`3oool0_g@004RLg`80oooo00<0gia80000B02Lgol0UP3oool00<T0oooo00P0_g@00000M02o
gi`0B0000000M02oool0oooo0=nLB080000000<0001809cOo`3oool00P3oool01@3OW4P000000000
000004P0W=oo00<0oooo00D0okmd000000000000000004RLg`030?ooo`070;md000007@0_ooo0?oo
o`3OW4P0001d0;ooo`070?ooo`060?oo_`1d000000000000000007@0_ooo0P3oool01`2oM00004RL
0=oog`2LB00000000000M02oool00P3oool01@3ooml0W4P0000000000000M;oo00<0oooo00<0gia8
0000M02oool00P3oool01@3OW4P000000000000004P0W=oo09L0oooo003/0?ooo`030=nLB00007@0
_ooo0200oooo00<0gia80000M02oool0WP3oool00>`0oooo00<0gia80000M02oool0803oool00`3O
W4P0001d0;ooo`2N0?ooo`00k03oool00`3OW4P0001d0;ooo`0P0?ooo`030=nLB00007@0_ooo09h0
oooo003X0?ooo`070=nLB00004P0W=oO09a800000000001d0;ooo`0N0?ooo`050?oog`2LB0000000
0000M02oool0WP3oool00>P0oooo00L0gia80000B02Lgol0oooo0;md000007@0_ooo0200oooo00<0
_g@00000M02oool0WP3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-30.3106, -27.8575, \
0.466765, 0.75524}}]
}, Closed]],

Cell[BoxData[
    \(\(PlotCMmomenta[cmMomenta];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[%, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.952381 0.5 0.952381 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.5 .5 m
.97619 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.02381 .5 L
s
% Start of user PostScript
0. 0. -0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.2619 .91239 L
s
% Start of user PostScript
0. 0. -0.25 0.4330127 MAarrowhead1
% End of user PostScript
.5 .5 m
.7381 .08761 L
s
% Start of user PostScript
0. 0. 0.25 -0.4330127 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
newpath
.5 .5 .47619 0 365.73 arc
s
3 Mabswid
.5 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ00000<P0oooo002m0?ooo`X000008@3oool:00000;h0oooo002e0?ooo`P00000
=@3oool800000;H0oooo002`0?ooo`D00000A@3oool500000;40oooo002[0?ooo`D00000C`3oool5
00000:`0oooo002V0?ooo`D00000F@3oool200000:X0oooo002T0?ooo`800000H03oool500000:D0
oooo002O0?ooo`D00000I`3oool500000:00oooo002M0?ooo`800000L@3oool3000009d0oooo002J
0?ooo`<00000MP3oool2000009/0oooo002E0?ooo`D00000N`3oool5000009H0oooo002C0?ooo`80
0000Q@3oool2000009@0oooo002@0?ooo`<00000R@3oool300000940oooo002>0?ooo`800000S`3o
ool2000008l0oooo002<0?ooo`800000T`3oool2000008d0oooo00290?ooo`<00000U`3oool30000
08X0oooo00270?ooo`800000W@3oool2000008P0oooo00250?ooo`800000X@3oool2000008H0oooo
00220?ooo`<00000Y@3oool3000008<0oooo00200?ooo`800000Z`3oool200000840oooo001n0?oo
o`800000[`3oool2000007l0oooo001k0?ooo`<00000/`3oool3000007`0oooo001i0?ooo`800000
^@3oool2000007X0oooo001h0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0M`3o
ool007L0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`1f0?ooo`00M@3oool20000
0<40oooo0P00001f0?ooo`00LP3oool300000<D0oooo0`00001c0?ooo`00L03oool200000</0oooo
0P00001a0?ooo`00K`3oool00`000000oooo0?ooo`3;0?ooo`<00000L03oool006h0oooo00<00000
0?ooo`3oool0b`3oool200000080oooo00<000000?ooo`3oool0K@3oool006`0oooo0P00003=0?oo
o`<000000`3oool2000006d0oooo001Z0?ooo`800000cP3oool3000000H0oooo0P00001[0?ooo`00
J@3oool00`000000oooo0?ooo`3=0?ooo`@00000203oool00`000000oooo0?ooo`1X0?ooo`00J03o
ool00`000000oooo0?ooo`3=0?ooo`D000002@3oool2000006P0oooo001U0?ooo`<00000d@3oool3
000000`0oooo0P00001V0?ooo`00H`3oool200000=@0oooo00<000000?ooo`0000003P3oool20000
06@0oooo001R0?ooo`030000003oool0oooo0=@0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`1Q0?ooo`00H@3oool00`000000oooo0?ooo`3D0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0H03oool005l0oooo0P00003G0?ooo`030000003oool0oooo01<0oooo0P00
001P0?ooo`00GP3oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo01H0oooo00<00000
0?ooo`3oool0G@3oool005d0oooo00<000000?ooo`3oool0eP3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo05`0oooo001K0?ooo`800000f@3oool00`000000oooo0?ooo`0I0?ooo`80
0000G03oool005X0oooo00<000000?ooo`3oool0f03oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo05T0oooo001I0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`1H0?ooo`00F03oool00`000000oooo0?ooo`3I0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool0E`3oool005L0oooo00<000000?ooo`3oool0fP3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo05H0oooo001E0?ooo`800000g03oool00`000000
oooo0?ooo`0R0?ooo`800000EP3oool005@0oooo00<000000?ooo`3oool0f`3oool00`000000oooo
0?ooo`0U0?ooo`030000003oool0oooo05<0oooo001C0?ooo`030000003oool0oooo0=`0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`1B0?ooo`00D@3oool200000=h0oooo00<00000
0?ooo`3oool0:03oool200000580oooo001@0?ooo`030000003oool0oooo0=h0oooo00<000000?oo
o`3oool0:P3oool00`000000oooo0?ooo`1?0?ooo`00C`3oool00`000000oooo0?ooo`3N0?ooo`03
0000003oool0oooo02`0oooo00<000000?ooo`3oool0CP3oool004h0oooo00<000000?ooo`3oool0
g`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo04d0oooo001=0?ooo`030000003o
ool0oooo0=l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`1<0?ooo`00C03oool0
0`000000oooo0?ooo`3O0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0B`3oool0
04/0oooo00<000000?ooo`3oool0h03oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
04X0oooo001:0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0=03oool00`000000
oooo0?ooo`190?ooo`00B@3oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool0B03oool004P0oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?oo
o`0g0?ooo`030000003oool0oooo04L0oooo00170?ooo`030000003oool0oooo0>40oooo00<00000
0?ooo`3oool0>@3oool00`000000oooo0?ooo`160?ooo`00AP3oool00`000000oooo0?ooo`3R0?oo
o`030000003oool0oooo03X0oooo00<000000?ooo`3oool0A@3oool004D0oooo00<000000?ooo`3o
ool0hP3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo04@0oooo00140?ooo`030000
003oool0oooo0><0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`130?ooo`00@`3o
ool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0@P3o
ool00480oooo00<000000?ooo`3oool0i03oool00`000000oooo0?ooo`100?ooo`030000003oool0
oooo0440oooo00110?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`100?ooo`00@03oool00`000000oooo0?ooo`3T0?ooo`030000003oool0oooo04@0
oooo00<000000?ooo`3oool0?`3oool003l0oooo00<000000?ooo`3oool0i@3oool00`000000oooo
0?ooo`150?ooo`030000003oool0oooo03h0oooo000n0?ooo`030000003oool0oooo0>D0oooo00<0
00000?ooo`3oool0A`3oool00`000000oooo0?ooo`0m0?ooo`00?@3oool00`000000oooo0?ooo`3V
0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool0?03oool003`0oooo00<000000?oo
o`3oool0iP3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo03/0oooo000k0?ooo`03
0000003oool0oooo0>L0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0j0?ooo`00
>P3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0
>P3oool003X0oooo00<000000?ooo`3oool0iP3oool00`000000oooo0?ooo`1>0?ooo`030000003o
ool0oooo03T0oooo000i0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0C`3oool0
0`000000oooo0?ooo`0h0?ooo`00>03oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo
0540oooo00<000000?ooo`3oool0=`3oool003L0oooo00<000000?ooo`3oool0j03oool00`000000
oooo0?ooo`1A0?ooo`030000003oool0oooo03L0oooo000g0?ooo`030000003oool0oooo0>L0oooo
00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0f0?ooo`00=P3oool00`000000oooo0?oo
o`3W0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0=@3oool003D0oooo00<00000
0?ooo`3oool0j03oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo03@0oooo000d0?oo
o`030000003oool0oooo0>P0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0c0?oo
o`00<`3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3o
ool0<P3oool00380oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`1J0?ooo`030000
003oool0oooo0380oooo000b0?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0F`3o
ool00`000000oooo0?ooo`0a0?ooo`00<@3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0
oooo05d0oooo00<000000?ooo`3oool0<03oool00300oooo00<000000?ooo`3oool0j@3oool00`00
0000oooo0?ooo`1O0?ooo`030000003oool0oooo02l0oooo000_0?ooo`030000003oool0oooo0>X0
oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`0_0?ooo`00;`3oool00`000000oooo
0?ooo`3Y0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0;P3oool002h0oooo00<0
00000?ooo`3oool0jP3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo02d0oooo000]
0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`0/
0?ooo`00;03oool00`000000oooo0?ooo`3[0?ooo`030000003oool0oooo06@0oooo00<000000?oo
o`3oool0;03oool002`0oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`1V0?ooo`03
0000003oool0oooo02/0oooo000/0?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0
I`3oool00`000000oooo0?ooo`0[0?ooo`00:`3oool00`000000oooo0?ooo`3Z0?ooo`030000003o
ool0oooo06L0oooo00<000000?ooo`3oool0:`3oool002/0oooo00<000000?ooo`3oool0j@3oool0
0`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo02X0oooo000Z0?ooo`030000003oool0oooo
0>X0oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0Y0?ooo`00:@3oool00`000000
oooo0?ooo`3Z0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0:03oool002P0oooo
00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo02P0oooo
000X0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?oo
o`0W0?ooo`009`3oool00`000000oooo0?ooo`3Z0?ooo`030000003oool0oooo06l0oooo00<00000
0?ooo`3oool09`3oool002L0oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`1`0?oo
o`030000003oool0oooo02H0oooo000V0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3o
ool0LP3oool00`000000oooo0?ooo`0U0?ooo`009@3oool00`000000oooo0?ooo`3[0?ooo`030000
003oool0oooo07<0oooo00<000000?ooo`3oool0903oool002D0oooo00<000000?ooo`3oool0jP3o
ool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo02@0oooo000T0?ooo`030000003oool0
oooo0>X0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0S0?ooo`00903oool00`00
0000oooo0?ooo`3Z0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool08`3oool002<0
oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo02<0
oooo000S0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool0N03oool00`000000oooo
0?ooo`0R0?ooo`008P3oool00`000000oooo0?ooo`3Z0?ooo`030000003oool0oooo07T0oooo00<0
00000?ooo`3oool08P3oool00280oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`1j
0?ooo`030000003oool0oooo0240oooo000Q0?ooo`030000003oool0oooo0>X0oooo00<000000?oo
o`3oool0O03oool00`000000oooo0?ooo`0P0?ooo`00803oool00`000000oooo0?ooo`3Z0?ooo`03
0000003oool0oooo07h0oooo00<000000?ooo`3oool07`3oool00200oooo00<000000?ooo`3oool0
jP3oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo01l0oooo000O0?ooo`030000003o
ool0oooo0>X0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0N0?ooo`007`3oool0
0`000000oooo0?ooo`3Z0?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool07P3oool0
01h0oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo
01h0oooo000N0?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0P`3oool00`000000
oooo0?ooo`0M0?ooo`007@3oool00`000000oooo0?ooo`3Z0?ooo`030000003oool0oooo08<0oooo
00<000000?ooo`3oool07@3oool001d0oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?oo
o`250?ooo`030000003oool0oooo01`0oooo000M0?ooo`030000003oool0oooo0>T0oooo00<00000
0?ooo`3oool0Q@3oool00`000000oooo0?ooo`0L0?ooo`00703oool00`000000oooo0?ooo`3Y0?oo
o`030000003oool0oooo08L0oooo00<000000?ooo`3oool06`3oool001`0oooo00<000000?ooo`3o
ool0j@3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo01/0oooo000K0?ooo`030000
003oool0oooo0>T0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0K0?ooo`006`3o
ool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3oool06P3o
ool001X0oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`2:0?ooo`030000003oool0
oooo01X0oooo000J0?ooo`030000003oool0oooo0>P0oooo00<000000?ooo`3oool0S03oool00`00
0000oooo0?ooo`0I0?ooo`006P3oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo08`0
oooo00<000000?ooo`3oool06@3oool001T0oooo00<000000?ooo`3oool0j03oool00`000000oooo
0?ooo`2>0?ooo`030000003oool0oooo01P0oooo000I0?ooo`030000003oool0oooo0>P0oooo00<0
00000?ooo`3oool0SP3oool00`000000oooo0?ooo`0H0?ooo`00603oool00`000000oooo0?ooo`3X
0?ooo`030000003oool0oooo08l0oooo00<000000?ooo`3oool0603oool001P0oooo00<000000?oo
o`3oool0i`3oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo01L0oooo000G0?ooo`03
0000003oool0oooo0>P0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`0G0?ooo`00
5`3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool0
5P3oool001L0oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`2C0?ooo`030000003o
ool0oooo01H0oooo000F0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0U@3oool0
0`000000oooo0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo
09D0oooo00<000000?ooo`3oool05@3oool001D0oooo00<000000?ooo`3oool0i`3oool00`000000
oooo0?ooo`2F0?ooo`030000003oool0oooo01D0oooo000E0?ooo`030000003oool0oooo0>H0oooo
00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`0D0?ooo`005@3oool00`000000oooo0?oo
o`3V0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0503oool001@0oooo00<00000
0?ooo`3oool0iP3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo01<0oooo000D0?oo
o`030000003oool0oooo0>H0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0C0?oo
o`00503oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo09/0oooo00<000000?ooo`3o
ool04`3oool001@0oooo00<000000?ooo`3oool0i03oool00`000000oooo0?ooo`2L0?ooo`030000
003oool0oooo01<0oooo000C0?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool0W03o
ool00`000000oooo0?ooo`0C0?ooo`004`3oool00`000000oooo0?ooo`3T0?ooo`030000003oool0
oooo09h0oooo00<000000?ooo`3oool04P3oool001<0oooo00<000000?ooo`3oool0i03oool00`00
0000oooo0?ooo`2N0?ooo`030000003oool0oooo0180oooo000B0?ooo`030000003oool0oooo0>@0
oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`0B0?ooo`004P3oool00`000000oooo
0?ooo`3T0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool04@3oool00140oooo00<0
00000?ooo`3oool0i03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0140oooo000A
0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0X`3oool00`000000oooo0?ooo`0@
0?ooo`004@3oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo0:<0oooo00<000000?oo
o`3oool0403oool00140oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`2T0?ooo`03
0000003oool0oooo0100oooo000A0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0
Y03oool00`000000oooo0?ooo`0@0?ooo`00403oool00`000000oooo0?ooo`3R0?ooo`030000003o
ool0oooo0:D0oooo00<000000?ooo`3oool0403oool00100oooo00<000000?ooo`3oool0hP3oool0
0`000000oooo0?ooo`2V0?ooo`030000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo
0>40oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000
oooo0?ooo`3P0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3oool03P3oool00100oooo
00<000000?ooo`3oool0h03oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo00h0oooo
000?0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`0>0?ooo`003`3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo0:X0oooo00<00000
0?ooo`3oool03P3oool000l0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00h0oooo000>0?ooo`030000003oool0oooo0=l0oooo00<000000?ooo`3o
ool0[@3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`3O0?ooo`030000
003oool0oooo0:d0oooo00<000000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3oool0gP3o
ool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0
oooo0=h0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`00
0000oooo0?ooo`3M0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool03@3oool000d0
oooo00<000000?ooo`3oool0gP3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo00`0
oooo000=0?ooo`030000003oool0oooo0=d0oooo00<000000?ooo`3oool0/@3oool00`000000oooo
0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`3L0?ooo`030000003oool0oooo0;80oooo00<0
00000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0g03oool00`000000oooo0?ooo`2b
0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0=/0oooo00<000000?oo
o`3oool0/`3oool00`000000oooo0?ooo`0<0?ooo`00303oool00`000000oooo0?ooo`3L0?ooo`03
0000003oool0oooo0;@0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0
f`3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003o
ool0oooo0=/0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0;0?ooo`00303oool0
0`000000oooo0?ooo`3J0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool02`3oool0
00`0oooo00<000000?ooo`3oool0f@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo
00/0oooo000<0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0]`3oool00`000000
oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`3H0?ooo`030000003oool0oooo0;P0oooo
00<000000?ooo`3oool02`3oool000/0oooo00<000000?ooo`3oool0f@3oool00`000000oooo0?oo
o`2h0?ooo`030000003oool0oooo00/0oooo000;0?ooo`030000003oool0oooo0=P0oooo00<00000
0?ooo`3oool0^P3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`3G0?oo
o`030000003oool0oooo0;/0oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3o
ool0e`3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000
003oool0oooo0=H0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`0:0?ooo`002`3o
ool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool02P3o
ool000/0oooo00<000000?ooo`3oool0e@3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0
oooo00X0oooo000;0?ooo`030000003oool0oooo0=D0oooo00<000000?ooo`3oool0_@3oool00`00
0000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`3D0?ooo`030000003oool0oooo0;h0
oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0d`3oool00`000000oooo
0?ooo`2o0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0=<0oooo00<0
00000?ooo`3oool0_`3oool00`000000oooo0?ooo`0:0?ooo`002P3oool00`000000oooo0?ooo`3C
0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?oo
o`3oool0d`3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo00T0oooo000:0?ooo`03
0000003oool0oooo0=80oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`090?ooo`00
2P3oool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool0
2@3oool000X0oooo00<000000?ooo`3oool0d@3oool00`000000oooo0?ooo`330?ooo`030000003o
ool0oooo00T0oooo000:0?ooo`030000003oool0oooo0=00oooo00<000000?ooo`3oool0a03oool0
0`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo
0<@0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0c`3oool00`000000
oooo0?ooo`350?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<l0oooo
00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?oo
o`3>0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool02@3oool000X0oooo00<00000
0?ooo`3oool0c@3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00T0oooo000:0?oo
o`030000003oool0oooo0<d0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`090?oo
o`002P3oool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3o
ool02@3oool000X0oooo00<000000?ooo`3oool0c03oool00`000000oooo0?ooo`380?ooo`030000
003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo00<0oooo0P0000360?ooo`030000003o
ool0oooo0<80oooo0P0000050?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0
000000D00000a03oool300000<@0oooo1P0000000`3oool000000?ooo`0:0?ooo`002P3ooooo0000
09`000002`3oool000X0oooo00<000000?ooo`0000001@0000330?ooo`@00000a03oool600000003
0?ooo`000000oooo00X0oooo000:0?ooo`030000003oool0oooo00<0oooo0P0000340?ooo`800000
a@3oool2000000D0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0b03o
ool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0
oooo0<P0oooo00<000000?ooo`3oool0c03oool00`000000oooo0?ooo`090?ooo`002P3oool00`00
0000oooo0?ooo`370?ooo`030000003oool0oooo0<d0oooo00<000000?ooo`3oool02@3oool000X0
oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo00T0
oooo000:0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0cP3oool00`000000oooo
0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo0<l0oooo00<0
00000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`3?
0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<@0oooo00<000000?oo
o`3oool0d03oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`340?ooo`03
0000003oool0oooo0=00oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0
``3oool00`000000oooo0?ooo`3A0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003o
ool0oooo0<80oooo00<000000?ooo`3oool0dP3oool00`000000oooo0?ooo`090?ooo`002P3oool0
0`000000oooo0?ooo`320?ooo`030000003oool0oooo0=80oooo00<000000?ooo`3oool02@3oool0
00X0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo
00T0oooo000:0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0d`3oool00`000000
oooo0?ooo`090?ooo`002`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo0=<0oooo
00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?oo
o`3C0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0;h0oooo00<00000
0?ooo`3oool0e03oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2m0?oo
o`030000003oool0oooo0=D0oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3o
ool0_@3oool00`000000oooo0?ooo`3E0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000
003oool0oooo0;`0oooo00<000000?ooo`3oool0eP3oool00`000000oooo0?ooo`0:0?ooo`002`3o
ool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3oool02P3o
ool000/0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`3G0?ooo`030000003oool0
oooo00X0oooo000;0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0e`3oool00`00
0000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo0=P0
oooo00<000000?ooo`3oool02P3oool000`0oooo00<000000?ooo`3oool0^03oool00`000000oooo
0?ooo`3H0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0;P0oooo00<0
00000?ooo`3oool0f03oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`2g
0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?oo
o`3oool0]`3oool00`000000oooo0?ooo`3I0?ooo`030000003oool0oooo00/0oooo000<0?ooo`03
0000003oool0oooo0;H0oooo00<000000?ooo`3oool0fP3oool00`000000oooo0?ooo`0;0?ooo`00
303oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool0
2`3oool000`0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`3K0?ooo`030000003o
ool0oooo00/0oooo000<0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0g03oool0
0`000000oooo0?ooo`0;0?ooo`003@3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo
0=/0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0/P3oool00`000000
oooo0?ooo`3L0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0;80oooo
00<000000?ooo`3oool0g03oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?oo
o`2a0?ooo`030000003oool0oooo0=d0oooo00<000000?ooo`3oool0303oool000d0oooo00<00000
0?ooo`3oool0/03oool00`000000oooo0?ooo`3N0?ooo`030000003oool0oooo00`0oooo000>0?oo
o`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0g@3oool00`000000oooo0?ooo`0=0?oo
o`003P3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo0=h0oooo00<000000?ooo`3o
ool03@3oool000h0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`3N0?ooo`030000
003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool0g`3o
ool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0
oooo0=l0oooo00<000000?ooo`3oool03@3oool000l0oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`3O0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0oooo0:X0
oooo00<000000?ooo`3oool0h03oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo
0?ooo`2Z0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool03P3oool000l0oooo00<0
00000?ooo`3oool0Z@3oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo00h0oooo000?
0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`0>
0?ooo`00403oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0>40oooo00<000000?oo
o`3oool03`3oool00100oooo00<000000?ooo`3oool0YP3oool00`000000oooo0?ooo`3R0?ooo`03
0000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool0
h@3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`2T0?ooo`030000003o
ool0oooo0>80oooo00<000000?ooo`3oool0403oool00140oooo00<000000?ooo`3oool0Y03oool0
0`000000oooo0?ooo`3R0?ooo`030000003oool0oooo0100oooo000A0?ooo`030000003oool0oooo
0:<0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000
oooo0?ooo`2S0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0403oool00140oooo
00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo0140oooo
000B0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0i03oool00`000000oooo0?oo
o`0A0?ooo`004P3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo0><0oooo00<00000
0?ooo`3oool04P3oool001<0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`3T0?oo
o`030000003oool0oooo0180oooo000C0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3o
ool0i03oool00`000000oooo0?ooo`0B0?ooo`00503oool00`000000oooo0?ooo`2L0?ooo`030000
003oool0oooo0>@0oooo00<000000?ooo`3oool04`3oool001@0oooo00<000000?ooo`3oool0W03o
ool00`000000oooo0?ooo`3T0?ooo`030000003oool0oooo01<0oooo000D0?ooo`030000003oool0
oooo09/0oooo00<000000?ooo`3oool0i@3oool00`000000oooo0?ooo`0C0?ooo`00503oool00`00
0000oooo0?ooo`2J0?ooo`030000003oool0oooo0>H0oooo00<000000?ooo`3oool04`3oool001@0
oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`3V0?ooo`030000003oool0oooo01<0
oooo000E0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0iP3oool00`000000oooo
0?ooo`0D0?ooo`005@3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo0>H0oooo00<0
00000?ooo`3oool0503oool001D0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`3V
0?ooo`030000003oool0oooo01D0oooo000F0?ooo`030000003oool0oooo09D0oooo00<000000?oo
o`3oool0i`3oool00`000000oooo0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`2E0?ooo`03
0000003oool0oooo0>L0oooo00<000000?ooo`3oool05@3oool001L0oooo00<000000?ooo`3oool0
T`3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo01H0oooo000G0?ooo`030000003o
ool0oooo09<0oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`0F0?ooo`005`3oool0
0`000000oooo0?ooo`2B0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool05`3oool0
01P0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo
01L0oooo000H0?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0i`3oool00`000000
oooo0?ooo`0H0?ooo`006@3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo0>P0oooo
00<000000?ooo`3oool0603oool001T0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?oo
o`3X0?ooo`030000003oool0oooo01P0oooo000J0?ooo`030000003oool0oooo08`0oooo00<00000
0?ooo`3oool0j03oool00`000000oooo0?ooo`0I0?ooo`006P3oool00`000000oooo0?ooo`2<0?oo
o`030000003oool0oooo0>P0oooo00<000000?ooo`3oool06@3oool001X0oooo00<000000?ooo`3o
ool0R`3oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo01X0oooo000K0?ooo`030000
003oool0oooo08X0oooo00<000000?ooo`3oool0j03oool00`000000oooo0?ooo`0J0?ooo`006`3o
ool00`000000oooo0?ooo`290?ooo`030000003oool0oooo0>P0oooo00<000000?ooo`3oool06`3o
ool001`0oooo00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0
oooo01/0oooo000L0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0j@3oool00`00
0000oooo0?ooo`0K0?ooo`007@3oool00`000000oooo0?ooo`250?ooo`030000003oool0oooo0>T0
oooo00<000000?ooo`3oool0703oool001d0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo
0?ooo`3Y0?ooo`030000003oool0oooo01`0oooo000M0?ooo`030000003oool0oooo08@0oooo00<0
00000?ooo`3oool0j@3oool00`000000oooo0?ooo`0M0?ooo`007P3oool00`000000oooo0?ooo`22
0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool07@3oool001h0oooo00<000000?oo
o`3oool0PP3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo01h0oooo000O0?ooo`03
0000003oool0oooo0800oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`0N0?ooo`00
7`3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool0
7P3oool00200oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`3Z0?ooo`030000003o
ool0oooo01l0oooo000P0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0jP3oool0
0`000000oooo0?ooo`0O0?ooo`008@3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo
0>X0oooo00<000000?ooo`3oool0803oool00280oooo00<000000?ooo`3oool0NP3oool00`000000
oooo0?ooo`3Z0?ooo`030000003oool0oooo0240oooo000R0?ooo`030000003oool0oooo07X0oooo
00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`0R0?ooo`008`3oool00`000000oooo0?oo
o`1h0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool08P3oool002<0oooo00<00000
0?ooo`3oool0N03oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo02<0oooo000T0?oo
o`030000003oool0oooo07H0oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`0S0?oo
o`00903oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3o
ool08`3oool002D0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`3Z0?ooo`030000
003oool0oooo02@0oooo000U0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0j`3o
ool00`000000oooo0?ooo`0T0?ooo`009P3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0
oooo0>X0oooo00<000000?ooo`3oool09@3oool002L0oooo00<000000?ooo`3oool0L03oool00`00
0000oooo0?ooo`3Z0?ooo`030000003oool0oooo02H0oooo000W0?ooo`030000003oool0oooo0700
oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`0W0?ooo`00:03oool00`000000oooo
0?ooo`1^0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool09`3oool002P0oooo00<0
00000?ooo`3oool0KP3oool00`000000oooo0?ooo`3Y0?ooo`030000003oool0oooo02P0oooo000Y
0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`0X
0?ooo`00:P3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo0>X0oooo00<000000?oo
o`3oool0:@3oool002/0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`3Y0?ooo`03
0000003oool0oooo02X0oooo000[0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0
j@3oool00`000000oooo0?ooo`0[0?ooo`00;03oool00`000000oooo0?ooo`1W0?ooo`030000003o
ool0oooo0>T0oooo00<000000?ooo`3oool0:`3oool002`0oooo00<000000?ooo`3oool0IP3oool0
0`000000oooo0?ooo`3Z0?ooo`030000003oool0oooo02/0oooo000]0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`0/0?ooo`00;@3oool00`000000
oooo0?ooo`1T0?ooo`030000003oool0oooo0>X0oooo00<000000?ooo`3oool0;03oool002h0oooo
00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`3Z0?ooo`030000003oool0oooo02d0oooo
000_0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?oo
o`0^0?ooo`00;`3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0>T0oooo00<00000
0?ooo`3oool0;`3oool00300oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`3Y0?oo
o`030000003oool0oooo02l0oooo000a0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3o
ool0j@3oool00`000000oooo0?ooo`0`0?ooo`00<P3oool00`000000oooo0?ooo`1K0?ooo`030000
003oool0oooo0>T0oooo00<000000?ooo`3oool0<@3oool00380oooo00<000000?ooo`3oool0F`3o
ool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo0380oooo000c0?ooo`030000003oool0
oooo05T0oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`0b0?ooo`00=03oool00`00
0000oooo0?ooo`1H0?ooo`030000003oool0oooo0>P0oooo00<000000?ooo`3oool0<`3oool003D0
oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`3X0?ooo`030000003oool0oooo03@0
oooo000f0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0i`3oool00`000000oooo
0?ooo`0e0?ooo`00=`3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo0>L0oooo00<0
00000?ooo`3oool0=P3oool003L0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`3W
0?ooo`030000003oool0oooo03L0oooo000h0?ooo`030000003oool0oooo0540oooo00<000000?oo
o`3oool0i`3oool00`000000oooo0?ooo`0g0?ooo`00>@3oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo0>L0oooo00<000000?ooo`3oool0>03oool003X0oooo00<000000?ooo`3oool0
CP3oool00`000000oooo0?ooo`3V0?ooo`030000003oool0oooo03T0oooo000j0?ooo`030000003o
ool0oooo04d0oooo00<000000?ooo`3oool0iP3oool00`000000oooo0?ooo`0j0?ooo`00>`3oool0
0`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0>P3oool0
03`0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`3V0?ooo`030000003oool0oooo
03/0oooo000m0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool0iP3oool00`000000
oooo0?ooo`0l0?ooo`00?P3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo0>D0oooo
00<000000?ooo`3oool0?@3oool003l0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?oo
o`3U0?ooo`030000003oool0oooo03h0oooo00100?ooo`030000003oool0oooo04@0oooo00<00000
0?ooo`3oool0i03oool00`000000oooo0?ooo`0o0?ooo`00@@3oool00`000000oooo0?ooo`120?oo
o`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0@03oool00480oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`3T0?ooo`030000003oool0oooo0440oooo00130?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`120?ooo`00A03o
ool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0@`3o
ool004D0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`3R0?ooo`030000003oool0
oooo04@0oooo00160?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0hP3oool00`00
0000oooo0?ooo`150?ooo`00A`3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0>40
oooo00<000000?ooo`3oool0AP3oool004P0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`3Q0?ooo`030000003oool0oooo04L0oooo00190?ooo`030000003oool0oooo03D0oooo00<0
00000?ooo`3oool0h@3oool00`000000oooo0?ooo`180?ooo`00BP3oool00`000000oooo0?ooo`0d
0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0B@3oool004/0oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo04X0oooo001<0?ooo`03
0000003oool0oooo0340oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`1;0?ooo`00
C@3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0=l0oooo00<000000?ooo`3oool0
C03oool004h0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`3O0?ooo`030000003o
ool0oooo04d0oooo001?0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0gP3oool0
0`000000oooo0?ooo`1>0?ooo`00D03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo
0=h0oooo00<000000?ooo`3oool0C`3oool00540oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`3M0?ooo`030000003oool0oooo0500oooo001B0?ooo`800000:03oool00`000000oooo
0?ooo`3L0?ooo`800000D`3oool005@0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?oo
o`3K0?ooo`030000003oool0oooo05<0oooo001E0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool0f`3oool00`000000oooo0?ooo`1D0?ooo`00EP3oool200000280oooo00<000000?oo
o`3oool0fP3oool2000005L0oooo001H0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3o
ool0f@3oool00`000000oooo0?ooo`1G0?ooo`00F@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo0=T0oooo00<000000?ooo`3oool0F03oool005X0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`3H0?ooo`030000003oool0oooo05T0oooo001K0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool0f03oool00`000000oooo0?ooo`1J0?ooo`00G03oool20000
01X0oooo00<000000?ooo`3oool0eP3oool2000005d0oooo001N0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool0eP3oool00`000000oooo0?ooo`1M0?ooo`00G`3oool00`000000oooo
0?ooo`0D0?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3oool0GP3oool00600oooo0P00
000D0?ooo`030000003oool0oooo0=@0oooo0P00001Q0?ooo`00HP3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo0=@0oooo00<000000?ooo`3oool0H@3oool006<0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0000003D0?ooo`800000I03oool006@0oooo0`00000<0?ooo`<0
0000dP3oool2000006H0oooo001W0?ooo`8000002@3oool500000<l0oooo0P00001X0?ooo`00J@3o
ool00`000000oooo0?ooo`060?ooo`@00000c`3oool00`000000oooo0?ooo`1X0?ooo`00JP3oool0
0`000000oooo0?ooo`040?ooo`@00000c`3oool00`000000oooo0?ooo`1Y0?ooo`00J`3oool20000
00@0oooo0`00003>0?ooo`800000K03oool006d0oooo0P0000020?ooo`800000c@3oool2000006h0
oooo001_0?ooo`<00000c@3oool00`000000oooo0?ooo`1^0?ooo`00L03oool200000</0oooo0P00
001a0?ooo`00LP3oool200000<L0oooo0P00001c0?ooo`00M03oool200000<<0oooo0P00001e0?oo
o`00MP3oool200000;l0oooo0P00001g0?ooo`00N03oool00`000000oooo0?ooo`2k0?ooo`030000
003oool0oooo07L0oooo001i0?ooo`800000^@3oool2000007X0oooo001k0?ooo`800000]@3oool2
000007`0oooo001m0?ooo`800000/@3oool2000007h0oooo001o0?ooo`<00000Z`3oool300000800
oooo00220?ooo`800000Y`3oool2000008<0oooo00240?ooo`800000X`3oool2000008D0oooo0026
0?ooo`<00000W@3oool3000008L0oooo00290?ooo`800000V@3oool2000008X0oooo002;0?ooo`80
0000U@3oool2000008`0oooo002=0?ooo`<00000S`3oool3000008h0oooo002@0?ooo`800000R`3o
ool200000940oooo002B0?ooo`<00000Q@3oool3000009<0oooo002E0?ooo`@00000O03oool50000
09H0oooo002I0?ooo`<00000M`3oool2000009/0oooo002L0?ooo`800000L`3oool2000009d0oooo
002N0?ooo`D00000J@3oool5000009l0oooo002S0?ooo`D00000H@3oool300000:@0oooo002X0?oo
o`<00000F@3oool500000:L0oooo002[0?ooo`D00000C`3oool500000:`0oooo002`0?ooo`D00000
A@3oool500000;40oooo002e0?ooo`L00000=`3oool700000;H0oooo002l0?ooo`/000008@3oool;
00000;d0oooo00370?ooob400000b03oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.525005, -0.525005, \
0.00365433, 0.00365433}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Colinear inelastic collision between equal masses", "Subsection"],

Cell["\<\
Inelastic collisions can be classified as endothermic when kinetic energy is \
transferred into internal energy or as exothermic when internal energy is \
liberated and becomes kinetic energy.  Thus, endothermic reactions have \
negative qvalue and exothermic reactions positive qvalue.  For simplicity, we \
consider inelastic collisions between equal masses and examine the \
kinematical relationships between the laboratory frame, defined as the rest \
frame of particle 2, and the center-of-mass frame. \
\>", "Text"],

Cell[CellGroupData[{

Cell["Endothermic reaction", "Subsubsection"],

Cell[BoxData[
    \(\(values = {m1 \[Rule] 1, m2 \[Rule] 1, m3 \[Rule] 1, 
          qvalue \[Rule] \(-0.2\), p1r \[Rule] 1, theta1 \[Rule] 0, 
          p2r \[Rule] 0, theta2 \[Rule] \[Pi]};\)\)], "Input"],

Cell[BoxData[
    \(\(sol = \(generalSolution[p3r, p4x, 
              p4y]\)\[LeftDoubleBracket]1\[RightDoubleBracket] //. 
          values;\)\)], "Input"],

Cell[BoxData[
    \(calcCollision[sol, values, 20\ Degree]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawCollision @@ LabMomenta;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .26631 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.627771 0.603961 0.133155 0.603961 [
[ 0 0 0 0 ]
[ 1 .26631 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .26631 L
0 .26631 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .13316 m
.62777 .13316 L
s
% Start of user PostScript
-1. 0. 0. 0. MAarrowhead1
% End of user PostScript
.62777 .13316 m
.62777 .13316 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.62777 .13316 m
.97619 .25997 L
s
% Start of user PostScript
0. 0. 0.57689064 0.20997102 MAarrowhead1
% End of user PostScript
.62777 .13316 m
.88331 .00634 L
s
% Start of user PostScript
0. 0. 0.42310936 -0.20997102 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 76.6875},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000Lb000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooomk0?ooo`800000=03oool00?l0ooooM`3o
ool5000003D0oooo003o0?ooogD0oooo1P00000f0?ooo`00o`3ooome0?ooo`D00000=`3oool00?l0
ooooM03oool4000003T0oooo003o0?ooog80oooo0P0000020?ooo`030000003oool0oooo03P0oooo
003o0?ooog00oooo0P00000o0?ooo`00o`3ooom^0?ooo`800000@@3oool00?l0ooooK03oool20000
04<0oooo003o0?ooofX0oooo0P0000150?ooo`00o`3ooomX0?ooo`800000A`3oool00?l0ooooIP3o
ool2000004T0oooo003o0?ooof@0oooo0P00001;0?ooo`00o`3ooomR0?ooo`800000C@3oool00?l0
ooooH03oool2000004l0oooo003o0?oooeh0oooo0P00001A0?ooo`00o`3ooomL0?ooo`800000D`3o
ool00?l0ooooFP3oool2000005D0oooo003o0?oooeP0oooo0P00001G0?ooo`00o`3ooomF0?ooo`80
0000F@3oool00?l0ooooE03oool2000005/0oooo003o0?oooe80oooo0P00001M0?ooo`00o`3ooom@
0?ooo`800000G`3oool00?l0ooooCP3oool200000640oooo003o0?oood`0oooo0P00001S0?ooo`00
o`3ooom:0?ooo`800000I@3oool00?l0ooooB03oool2000006L0oooo003o0?ooodH0oooo0P00001Y
0?ooo`00o`3ooom40?ooo`800000J`3oool00?l0oooo@P3oool2000006d0oooo003o0?oood00oooo
0P00001_0?ooo`00o`3oooln0?ooo`800000L@3oool00?l0oooo?03oool2000007<0oooo003o0?oo
ocX0oooo0P00001e0?ooo`00o`3ooolh0?ooo`800000M`3oool00?l0oooo=P3oool2000007T0oooo
003o0?oooc@0oooo0P00001k0?ooo`00o`3ooolb0?ooo`800000O@3oool00?l0oooo<03oool20000
07l0oooo003o0?ooobh0oooo0P0000210?ooo`00o`3oool/0?ooo`800000P`3oool00?l0oooo:P3o
ool2000008D0oooo003o0?ooobP0oooo0P0000270?ooo`00o`3ooolV0?ooo`800000R@3oool00?l0
oooo903oool2000008/0oooo003o0?ooob80oooo0P00002=0?ooo`00o`3ooolP0?ooo`800000S`3o
ool00?l0oooo7P3oool200000940oooo003o0?oooa`0oooo0P00002C0?ooo`00o`3ooolJ0?ooo`80
0000U@3oool00?l0oooo603oool2000009L0oooo003o0?oooaH0oooo0P00002I0?ooo`00o`3ooolD
0?ooo`800000V`3oool00?l0oooo1`3oool4000000L0oooo0P00002M0?ooo`00o`3oool70?ooo`/0
0000W`3oool000/0ooooo`00000600000:00oooo003o0?ooo`L0oooo0`0000070?ooo`<00000W@3o
ool00?l0oooo503oool2000009/0oooo003o0?oooaH0oooo0`00002H0?ooo`00o`3ooolI0?ooo`<0
0000U@3oool00?l0oooo703oool300000980oooo003o0?oooal0oooo0P00002@0?ooo`00o`3ooolQ
0?ooo`<00000S@3oool00?l0oooo903oool3000008X0oooo003o0?ooobL0oooo0`0000270?ooo`00
o`3ooolZ0?ooo`800000Q@3oool00?l0oooo;03oool300000880oooo003o0?ooobl0oooo0`00001o
0?ooo`00o`3ooolb0?ooo`<00000O03oool00?l0oooo=@3oool3000007T0oooo003o0?ooocP0oooo
0P00001g0?ooo`00o`3ooolj0?ooo`<00000M03oool00?l0oooo?@3oool300000740oooo003o0?oo
od00oooo0`00001^0?ooo`00o`3ooom30?ooo`800000K03oool00?l0ooooA@3oool3000006T0oooo
003o0?ooodP0oooo0`00001V0?ooo`00o`3ooom;0?ooo`<00000H`3oool00?l0ooooCP3oool20000
0640oooo003o0?oooe00oooo0`00001N0?ooo`00o`3ooomC0?ooo`<00000F`3oool00?l0ooooEP3o
ool3000005P0oooo003o0?oooeT0oooo0P00001F0?ooo`00o`3ooomK0?ooo`<00000D`3oool00?l0
ooooGP3oool300000500oooo003o0?ooof40oooo0`00001=0?ooo`00o`3ooomT0?ooo`800000B`3o
ool00?l0ooooIP3oool3000004P0oooo003o0?ooofT0oooo0`0000150?ooo`00o`3ooom/0?ooo`<0
0000@P3oool00?l0ooooK`3oool200000400oooo003o0?ooog40oooo0`00000m0?ooo`00o`3ooomd
0?ooo`<00000>P3oool00?l0ooooM`3oool3000003L0oooo003o0?ooogX0oooo0P00000e0?ooo`00
o`3oooml0?ooo`<00000<P3oool00?l0ooooO`3oool3000002l0oooo003o0?oooh80oooo0`00000/
0?ooo`00o`3ooon50?ooo`<00000:@3oool00?l0ooooR03oool2000002L0oooo003o0?ooohX0oooo
0`00000T0?ooo`00o`3ooon=0?ooo`<000008@3oool00?l0ooooT03oool3000001h0oooo003o0?oo
oi<0oooo0P00000L0?ooo`00o`3ooonE0?ooo`<000001@3oool00`000000oooo0?ooo`0A0?ooo`00
o`3ooonH0?ooo`<000000P3oool300000140oooo003o0?oooi/0oooo1P00000@0?ooo`00o`3ooonL
0?ooo`H000003`3oool00?l0ooooW03oool8000000d0oooo003o0?oooj00oooo1@00000<0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {114, 0}} -> {-1.04509, -0.220472, 0.00580189, \
0.00580189}}]
}, Closed]],

Cell[TextData[{
  "For an endothermic reaction, the lab angle between the two emerging \
particles (with equal masses) is now less than ",
  Cell[BoxData[
      \(TraditionalForm\`90  \[Degree]\)]],
  ".  (Recall that this angle is ",
  Cell[BoxData[
      \(TraditionalForm\`90  \[Degree]\)]],
  " for an elastic collision between equal masses.)"
}], "Text"],

Cell[BoxData[
    \(\(PlotCMmomenta[cmMomenta];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[%, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.952381 0.5 0.952381 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.5 .5 m
.97619 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.02381 .5 L
s
% Start of user PostScript
0. 0. -0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.57323 .69997 L
s
% Start of user PostScript
0. 0. 0.076890639 0.20997102 MAarrowhead1
% End of user PostScript
.5 .5 m
.42677 .30003 L
s
% Start of user PostScript
0. 0. -0.076890639 -0.20997102 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
newpath
.5 .5 .21296 0 365.73 arc
s
3 Mabswid
.5 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ00000<P0oooo002m0?ooo`X000008@3oool:00000;h0oooo002e0?ooo`P00000
=@3oool800000;H0oooo002`0?ooo`D00000A@3oool500000;40oooo002[0?ooo`D00000C`3oool5
00000:`0oooo002V0?ooo`D00000F@3oool200000:X0oooo002T0?ooo`800000H03oool500000:D0
oooo002O0?ooo`D00000I`3oool500000:00oooo002M0?ooo`800000L@3oool3000009d0oooo002J
0?ooo`<00000MP3oool2000009/0oooo002E0?ooo`D00000N`3oool5000009H0oooo002C0?ooo`80
0000Q@3oool2000009@0oooo002@0?ooo`<00000R@3oool300000940oooo002>0?ooo`800000S`3o
ool2000008l0oooo002<0?ooo`800000T`3oool2000008d0oooo00290?ooo`<00000U`3oool30000
08X0oooo00270?ooo`800000W@3oool2000008P0oooo00250?ooo`800000X@3oool2000008H0oooo
00220?ooo`<00000Y@3oool3000008<0oooo00200?ooo`800000Z`3oool200000840oooo001n0?oo
o`800000[`3oool2000007l0oooo001k0?ooo`<00000/`3oool3000007`0oooo001i0?ooo`800000
^@3oool2000007X0oooo001h0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0M`3o
ool007L0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`1f0?ooo`00M@3oool20000
0<40oooo0P00001f0?ooo`00LP3oool300000<D0oooo0`00001c0?ooo`00L03oool200000</0oooo
0P00001a0?ooo`00K`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo06h0oooo001^
0?ooo`030000003oool0oooo0<l0oooo00<000000?ooo`3oool0K@3oool006`0oooo0P00003C0?oo
o`800000K@3oool006X0oooo0P00003G0?ooo`800000J`3oool006T0oooo00<000000?ooo`3oool0
f@3oool00`000000oooo0?ooo`1X0?ooo`00J03oool00`000000oooo0?ooo`3K0?ooo`800000J03o
ool006D0oooo0`00003P0?ooo`800000IP3oool006<0oooo0P00003U0?ooo`800000I03oool00680
oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`1Q0?ooo`00H@3oool00`000000oooo
0?ooo`3Y0?ooo`030000003oool0oooo0600oooo001O0?ooo`800000k@3oool200000600oooo001N
0?ooo`030000003oool0oooo0>l0oooo00<000000?ooo`3oool0G@3oool005d0oooo00<000000?oo
o`3oool0l@3oool00`000000oooo0?ooo`1L0?ooo`00F`3oool200000?D0oooo0P00001L0?ooo`00
FP3oool00`000000oooo0?ooo`3g0?ooo`030000003oool0oooo05T0oooo001I0?ooo`030000003o
ool0oooo0?T0oooo00<000000?ooo`3oool0F03oool005P0oooo00<000000?ooo`3oool0n`3oool0
0`000000oooo0?ooo`1G0?ooo`00E`3oool00`000000oooo0?ooo`3m0?ooo`030000003oool0oooo
05H0oooo001E0?ooo`800000o`3oool20?ooo`800000EP3oool005@0oooo00<000000?ooo`3oool0
o`3oool40?ooo`030000003oool0oooo05<0oooo001C0?ooo`030000003oool0oooo0?l0oooo1P3o
ool00`000000oooo0?ooo`1B0?ooo`00D@3oool200000?l0oooo2P3oool200000580oooo001@0?oo
o`030000003oool0oooo0?l0oooo303oool00`000000oooo0?ooo`1?0?ooo`00C`3oool00`000000
oooo0?ooo`3o0?ooo`h0oooo00<000000?ooo`3oool0CP3oool004h0oooo00<000000?ooo`3oool0
o`3oool@0?ooo`030000003oool0oooo04d0oooo001=0?ooo`030000003oool0oooo0?l0oooo4P3o
ool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo0?ooo`3o0?oooa@0oooo00<00000
0?ooo`3oool0B`3oool004/0oooo00<000000?ooo`3oool0o`3ooolF0?ooo`030000003oool0oooo
04X0oooo001:0?ooo`030000003oool0oooo0?l0oooo603oool00`000000oooo0?ooo`190?ooo`00
B@3oool00`000000oooo0?ooo`3o0?oooaX0oooo00<000000?ooo`3oool0B03oool004P0oooo00<0
00000?ooo`3oool0o`3ooolL0?ooo`030000003oool0oooo04L0oooo00170?ooo`030000003oool0
oooo0?l0oooo7P3oool00`000000oooo0?ooo`160?ooo`00AP3oool00`000000oooo0?ooo`3o0?oo
ob00oooo00<000000?ooo`3oool0A@3oool004D0oooo00<000000?ooo`3oool0o`3ooolR0?ooo`03
0000003oool0oooo04@0oooo00140?ooo`030000003oool0oooo0?l0oooo903oool00`000000oooo
0?ooo`130?ooo`00@`3oool00`000000oooo0?ooo`3o0?ooobH0oooo00<000000?ooo`3oool0@P3o
ool00480oooo00<000000?ooo`3oool0o`3ooolX0?ooo`030000003oool0oooo0440oooo00110?oo
o`030000003oool0oooo0?l0oooo:P3oool00`000000oooo0?ooo`100?ooo`00@03oool00`000000
oooo0?ooo`3o0?ooob`0oooo00<000000?ooo`3oool0?`3oool003l0oooo00<000000?ooo`3oool0
o`3oool^0?ooo`030000003oool0oooo03h0oooo000n0?ooo`030000003oool0oooo0?l0oooo<03o
ool00`000000oooo0?ooo`0m0?ooo`00?@3oool00`000000oooo0?ooo`3o0?oooc80oooo00<00000
0?ooo`3oool0?03oool003`0oooo00<000000?ooo`3oool0o`3ooold0?ooo`030000003oool0oooo
03/0oooo000k0?ooo`030000003oool0oooo0?l0oooo=P3oool00`000000oooo0?ooo`0j0?ooo`00
>P3oool00`000000oooo0?ooo`3o0?ooocL0oooo00<000000?ooo`3oool0>P3oool003X0oooo00<0
00000?ooo`3oool0o`3ooolh0?ooo`030000003oool0oooo03T0oooo000i0?ooo`030000003oool0
oooo0?l0oooo>P3oool00`000000oooo0?ooo`0h0?ooo`00>03oool00`000000oooo0?ooo`3o0?oo
oc`0oooo00<000000?ooo`3oool0=`3oool003L0oooo00<000000?ooo`3oool0o`3ooolm0?ooo`03
0000003oool0oooo03L0oooo000g0?ooo`030000003oool0oooo0?l0oooo?P3oool00`000000oooo
0?ooo`0f0?ooo`00=P3oool00`000000oooo0?ooo`3o0?oood00oooo00<000000?ooo`3oool0=@3o
ool003D0oooo00<000000?ooo`3oool0o`3ooom20?ooo`030000003oool0oooo03@0oooo000d0?oo
o`030000003oool0oooo0?l0ooooA03oool00`000000oooo0?ooo`0c0?ooo`00<`3oool00`000000
oooo0?ooo`3o0?ooodH0oooo00<000000?ooo`3oool0<P3oool00380oooo00<000000?ooo`3oool0
o`3ooom70?ooo`030000003oool0oooo0380oooo000b0?ooo`030000003oool0oooo0?l0ooooB03o
ool00`000000oooo0?ooo`0a0?ooo`00<@3oool00`000000oooo0?ooo`3o0?ooodX0oooo00<00000
0?ooo`3oool0<03oool00300oooo00<000000?ooo`3oool0o`3ooom<0?ooo`030000003oool0oooo
02l0oooo000_0?ooo`030000003oool0oooo0?l0ooooC@3oool00`000000oooo0?ooo`0_0?ooo`00
;`3oool00`000000oooo0?ooo`3o0?ooodh0oooo00<000000?ooo`3oool0;P3oool002h0oooo00<0
00000?ooo`3oool0o`3ooom@0?ooo`030000003oool0oooo02d0oooo000]0?ooo`030000003oool0
oooo0?l0ooooDP3oool00`000000oooo0?ooo`0/0?ooo`00;03oool00`000000oooo0?ooo`3o0?oo
oe<0oooo00<000000?ooo`3oool0;03oool002`0oooo00<000000?ooo`3oool0o`3ooomD0?ooo`03
0000003oool0oooo02/0oooo000/0?ooo`030000003oool0oooo0?l0ooooE03oool00`000000oooo
0?ooo`0[0?ooo`00:`3oool00`000000oooo0?ooo`3o0?oooeD0oooo00<000000?ooo`3oool0:`3o
ool002/0oooo00<000000?ooo`3oool0o`3ooomF0?ooo`030000003oool0oooo02X0oooo000Z0?oo
o`030000003oool0oooo0?l0ooooF03oool00`000000oooo0?ooo`0Y0?ooo`00:@3oool00`000000
oooo0?ooo`3o0?oooeX0oooo00<000000?ooo`3oool0:03oool002P0oooo00<000000?ooo`3oool0
o`3ooomK0?ooo`030000003oool0oooo02P0oooo000X0?ooo`030000003oool0oooo0?l0ooooG03o
ool00`000000oooo0?ooo`0W0?ooo`009`3oool00`000000oooo0?ooo`3o0?oooed0oooo00<00000
0?ooo`3oool09`3oool002L0oooo00<000000?ooo`3oool0o`3ooomN0?ooo`030000003oool0oooo
02H0oooo000V0?ooo`030000003oool0oooo0?l0ooooH03oool00`000000oooo0?ooo`0U0?ooo`00
9@3oool00`000000oooo0?ooo`3o0?ooof80oooo00<000000?ooo`3oool0903oool002D0oooo00<0
00000?ooo`3oool0o`3ooomR0?ooo`030000003oool0oooo02@0oooo000T0?ooo`030000003oool0
oooo0?l0ooooI03oool00`000000oooo0?ooo`0S0?ooo`00903oool00`000000oooo0?ooo`3o0?oo
of@0oooo00<000000?ooo`3oool08`3oool002<0oooo00<000000?ooo`3oool0o`3ooomU0?ooo`03
0000003oool0oooo02<0oooo000S0?ooo`030000003oool0oooo0?l0ooooIP3oool00`000000oooo
0?ooo`0R0?ooo`008P3oool00`000000oooo0?ooo`3o0?ooofL0oooo00<000000?ooo`3oool08P3o
ool00280oooo00<000000?ooo`3oool0o`3ooomX0?ooo`030000003oool0oooo0240oooo000Q0?oo
o`030000003oool0oooo0?l0ooooJP3oool00`000000oooo0?ooo`0P0?ooo`00803oool00`000000
oooo0?ooo`3o0?ooof`0oooo00<000000?ooo`3oool07`3oool00200oooo00<000000?ooo`3oool0
o`3ooom/0?ooo`030000003oool0oooo01l0oooo000O0?ooo`030000003oool0oooo0:h0oooo3`00
002`0?ooo`030000003oool0oooo01h0oooo000O0?ooo`030000003oool0oooo0:L0oooo1`00000?
0?ooo`L00000Z@3oool00`000000oooo0?ooo`0N0?ooo`007P3oool00`000000oooo0?ooo`2T0?oo
o`@000007@3oool400000:D0oooo00<000000?ooo`3oool07P3oool001h0oooo00<000000?ooo`3o
ool0W`3oool5000002D0oooo1@00002Q0?ooo`030000003oool0oooo01d0oooo000M0?ooo`030000
003oool0oooo09`0oooo1000000_0?ooo`@00000W@3oool00`000000oooo0?ooo`0M0?ooo`007@3o
ool00`000000oooo0?ooo`2J0?ooo`800000=`3oool2000009`0oooo00<000000?ooo`3oool0703o
ool001d0oooo00<000000?ooo`3oool0U`3oool3000003/0oooo0`00002I0?ooo`030000003oool0
oooo01`0oooo000L0?ooo`030000003oool0oooo09@0oooo1P00000o0?ooo`@00000UP3oool00`00
0000oooo0?ooo`0K0?ooo`00703oool00`000000oooo0?ooo`2B0?ooo`800000103oool300000480
oooo0P00002D0?ooo`030000003oool0oooo01/0oooo000K0?ooo`030000003oool0oooo0940oooo
0P0000060?ooo`<00000A03oool200000980oooo00<000000?ooo`3oool06`3oool001/0oooo00<0
00000?ooo`3oool0S`3oool2000000T0oooo0`0000150?ooo`800000T@3oool00`000000oooo0?oo
o`0J0?ooo`006P3oool00`000000oooo0?ooo`2>0?ooo`8000002`3oool4000004H0oooo0P00002?
0?ooo`030000003oool0oooo01X0oooo000J0?ooo`030000003oool0oooo08d0oooo00<000000?oo
o`3oool02`3oool5000004L0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0I0?oo
o`006P3oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo00`0oooo0`00001:0?ooo`03
0000003oool0oooo08`0oooo00<000000?ooo`3oool06@3oool001T0oooo00<000000?ooo`3oool0
R`3oool200000140oooo00<000000?ooo`3oool0B@3oool2000008d0oooo00<000000?ooo`3oool0
603oool001T0oooo00<000000?ooo`3oool0R@3oool2000001<0oooo00<000000?ooo`3oool0B`3o
ool2000008/0oooo00<000000?ooo`3oool0603oool001P0oooo00<000000?ooo`3oool0R03oool2
000001H0oooo00<000000?ooo`3oool0C03oool2000008T0oooo00<000000?ooo`3oool0603oool0
01P0oooo00<000000?ooo`3oool0QP3oool2000001P0oooo00<000000?ooo`3oool0CP3oool20000
08P0oooo00<000000?ooo`3oool05`3oool001L0oooo00<000000?ooo`3oool0QP3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0QP3oool00`000000
oooo0?ooo`0G0?ooo`005`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo01H0oooo
000G0?ooo`030000003oool0oooo08@0oooo0P00000L0?ooo`030000003oool0oooo0500oooo0P00
00260?ooo`030000003oool0oooo01H0oooo000F0?ooo`030000003oool0oooo08@0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo08@0oooo00<00000
0?ooo`3oool05@3oool001H0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0N0?oo
o`030000003oool0oooo0580oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0E0?oo
o`005@3oool00`000000oooo0?ooo`220?ooo`8000008@3oool00`000000oooo0?ooo`1C0?ooo`80
0000P`3oool00`000000oooo0?ooo`0E0?ooo`005@3oool00`000000oooo0?ooo`210?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`220?ooo`030000
003oool0oooo01@0oooo000E0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0503o
ool001@0oooo00<000000?ooo`3oool0P03oool2000002D0oooo00<000000?ooo`3oool0E@3oool2
00000880oooo00<000000?ooo`3oool04`3oool001@0oooo00<000000?ooo`3oool0O`3oool00`00
0000oooo0?ooo`0V0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3oool0O`3oool00`00
0000oooo0?ooo`0C0?ooo`00503oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo02L0
oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo01<0
oooo000D0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`1F0?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool04`3oool001<0oooo00<0
00000?ooo`3oool0OP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo05L0oooo00<0
00000?ooo`3oool0O@3oool00`000000oooo0?ooo`0C0?ooo`004`3oool00`000000oooo0?ooo`1l
0?ooo`800000;03oool00`000000oooo0?ooo`1H0?ooo`800000OP3oool00`000000oooo0?ooo`0B
0?ooo`004`3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo02d0oooo00<000000?oo
o`3oool0F03oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo0180oooo000B0?ooo`03
0000003oool0oooo07`0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1I0?ooo`03
0000003oool0oooo07/0oooo00<000000?ooo`3oool04P3oool00180oooo00<000000?ooo`3oool0
N`3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0
N`3oool00`000000oooo0?ooo`0A0?ooo`004@3oool00`000000oooo0?ooo`1k0?ooo`030000003o
ool0oooo0300oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1j0?ooo`030000003o
ool0oooo0140oooo000A0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0403oool0
0140oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
05X0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000
oooo0?ooo`1h0?ooo`800000=@3oool00`000000oooo0?ooo`1K0?ooo`800000NP3oool00`000000
oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0100oooo
000@0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?oo
o`1L0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0403oool00100oooo00<00000
0?ooo`3oool0M`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo05`0oooo00<00000
0?ooo`3oool0N03oool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000oooo0?ooo`1g0?oo
o`030000003oool0oooo03L0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`1g0?oo
o`030000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3o
ool03P3oool00100oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0i0?ooo`030000
003oool0oooo05d0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0>0?ooo`003`3o
ool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0G@3o
ool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0
oooo07H0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0
oooo07D0oooo00<000000?ooo`3oool03P3oool000l0oooo00<000000?ooo`3oool0M@3oool00`00
0000oooo0?ooo`0l0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0M@3oool00`00
0000oooo0?ooo`0>0?ooo`003P3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo03h0
oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo00d0
oooo000>0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0?`3oool00`000000oooo
0?ooo`1N0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool03@3oool000h0oooo00<0
00000?ooo`3oool0M03oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo05l0oooo00<0
00000?ooo`3oool0M03oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`1c
0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`1d
0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo07<0oooo00<000000?oo
o`3oool0@@3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo07<0oooo00<000000?oo
o`3oool03@3oool000d0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`120?ooo`03
0000003oool0oooo05l0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0<0?ooo`00
3@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0
G`3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003o
ool0oooo0780oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`1O0?ooo`030000003o
ool0oooo07<0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0LP3oool0
0`000000oooo0?ooo`150?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0LP3oool0
0`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo
04D0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo
00`0oooo000<0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0A@3oool00`000000
oooo0?ooo`1O0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo05h0oooo
00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?oo
o`1b0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?oo
o`1b0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0780oooo00<00000
0?ooo`3oool0A`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0780oooo00<00000
0?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo05h0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0;0?oo
o`00303oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0G`3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1N0?ooo`030000
003oool0oooo0740oooo00<000000?ooo`3oool02`3oool000/0oooo00<000000?ooo`3oool0L@3o
ool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0L03o
ool00`000000oooo0?ooo`0;0?ooo`002`3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0
oooo00X0oooo000;0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0C03oool00`00
0000oooo0?ooo`1N0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool02P3oool000/0
oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo05h0
oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo
0?ooo`1`0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0G`3oool00`000000oooo
0?ooo`1`0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0700oooo00<0
00000?ooo`3oool0CP3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo0700oooo00<0
00000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`1>
0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`0:
0?ooo`002`3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0G@3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo00X0oooo000;0?ooo`03
0000003oool0oooo06l0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1M0?ooo`03
0000003oool0oooo0700oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0
K`3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0
K`3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`1_0?ooo`030000003o
ool0oooo0540oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`1_0?ooo`030000003o
ool0oooo00X0oooo000:0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0D@3oool0
0`000000oooo0?ooo`1M0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool02@3oool0
00X0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo
05d0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000
oooo0?ooo`1_0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0G@3oool00`000000
oooo0?ooo`1_0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?oo
o`1D0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?oo
o`090?ooo`002P3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo05@0oooo00<00000
0?ooo`3oool0G03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00T0oooo000:0?oo
o`030000003oool0oooo06l0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1L0?oo
o`030000003oool0oooo06l0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3o
ool0K`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3o
ool0K`3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`1^0?ooo`030000
003oool0oooo05H0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`1_0?ooo`030000
003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0EP3o
ool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool02@3o
ool000X0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0
oooo05/0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`090?ooo`002P3oool00`00
0000oooo0?ooo`1^0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0F`3oool00`00
0000oooo0?ooo`1^0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo06h0
oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo06h0
oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0KP3oool00`000000oooo
0?ooo`1H0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0KP3oool00`000000oooo
0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`030?ooo`800000J@3oool00`000000oooo0?oo
o`1H0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0I`3oool2000000D0oooo00<0
00000?ooo`3oool02@3oool000X0oooo00<000000?ooo`0000001@00001Y0?ooo`030000003oool0
oooo05P0oooo0P00001K0?ooo`030000003oool0oooo06L0oooo1P0000000`3oool000000?ooo`0:
0?ooo`002P3ooooo000009`000002`3oool000X0oooo00<000000?ooo`0000001@00001Y0?ooo`03
0000003oool0oooo05L0oooo1000001J0?ooo`030000003oool0oooo06L0oooo1P0000000`3oool0
00000?ooo`0:0?ooo`002P3oool00`000000oooo0?ooo`030?ooo`800000J@3oool00`000000oooo
0?ooo`1H0?ooo`<00000FP3oool00`000000oooo0?ooo`1W0?ooo`8000001@3oool00`000000oooo
0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo05X0oooo00<0
00000?ooo`3oool0F03oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo00T0oooo000:
0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1H
0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?oo
o`3oool0KP3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo05L0oooo00<000000?oo
o`3oool0KP3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`1^0?ooo`03
0000003oool0oooo05/0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`1^0?ooo`03
0000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0
G03oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0
2@3oool000X0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1L0?ooo`030000003o
ool0oooo05D0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`090?ooo`002P3oool0
0`000000oooo0?ooo`1_0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0E@3oool0
0`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo
06l0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
06l0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0K`3oool00`000000
oooo0?ooo`1L0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0K`3oool00`000000
oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo05`0oooo
00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00T0oooo
000:0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?oo
o`1C0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool02@3oool000X0oooo00<00000
0?ooo`3oool0K`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo05<0oooo00<00000
0?ooo`3oool0K`3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`1_0?oo
o`030000003oool0oooo05h0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`1`0?oo
o`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3o
ool0G@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3o
ool02@3oool000/0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1M0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0:0?ooo`002`3o
ool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0D03o
ool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0
oooo06l0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0
oooo0700oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0L03oool00`00
0000oooo0?ooo`1M0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0L03oool00`00
0000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo05h0
oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo00X0
oooo000;0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0GP3oool00`000000oooo
0?ooo`1>0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool02P3oool000/0oooo00<0
00000?ooo`3oool0L03oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo04d0oooo00<0
00000?ooo`3oool0L03oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`1`
0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1a
0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0740oooo00<000000?oo
o`3oool0GP3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo0740oooo00<000000?oo
o`3oool02P3oool000/0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1O0?ooo`03
0000003oool0oooo04/0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`0:0?ooo`00
303oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0
B`3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003o
ool0oooo0700oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1:0?ooo`030000003o
ool0oooo0740oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0L@3oool0
0`000000oooo0?ooo`1O0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0L@3oool0
0`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo
05l0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo
00/0oooo000<0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0G`3oool00`000000
oooo0?ooo`170?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo04L0oooo
00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo07<0oooo00<00000
0?ooo`3oool0G`3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo07<0oooo00<00000
0?ooo`3oool02`3oool000d0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`1O0?oo
o`030000003oool0oooo04D0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0<0?oo
o`003@3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3o
ool0A@3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000
003oool0oooo0780oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`130?ooo`030000
003oool0oooo07<0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0L`3o
ool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0L`3o
ool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0
oooo0600oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0
oooo00`0oooo000>0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0G`3oool00`00
0000oooo0?ooo`110?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool03@3oool000h0
oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0400
oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo
0?ooo`1d0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0?`3oool00`000000oooo
0?ooo`1d0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo07@0oooo00<0
00000?ooo`3oool0G`3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo07D0oooo00<0
00000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`1N
0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0=
0?ooo`003`3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo05h0oooo00<000000?oo
o`3oool0?03oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo00h0oooo000?0?ooo`03
0000003oool0oooo07H0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`0k0?ooo`03
0000003oool0oooo07H0oooo00<000000?ooo`3oool03P3oool000l0oooo00<000000?ooo`3oool0
MP3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0
M`3oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo0?ooo`1g0?ooo`030000003o
ool0oooo05d0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`1g0?ooo`030000003o
ool0oooo00h0oooo000?0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0G@3oool0
0`000000oooo0?ooo`0h0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool03P3oool0
0100oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000
oooo0?ooo`1g0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`1h0?ooo`030000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo07P0oooo
00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo07L0oooo
00<000000?ooo`3oool0403oool00140oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?oo
o`1M0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?oo
o`0@0?ooo`004@3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo05`0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0100oooo000A0?oo
o`030000003oool0oooo07T0oooo0P00001L0?ooo`030000003oool0oooo0380oooo0P00001k0?oo
o`030000003oool0oooo0100oooo000A0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3o
ool0403oool00140oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`1J0?ooo`030000
003oool0oooo0300oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0A0?ooo`004P3o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0;P3o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0140oooo000B0?ooo`030000003oool0
oooo07`0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0
oooo07/0oooo00<000000?ooo`3oool04P3oool001<0oooo00<000000?ooo`3oool0N`3oool00`00
0000oooo0?ooo`1I0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0O03oool00`00
0000oooo0?ooo`0B0?ooo`004`3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo05T0
oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo0180
oooo000D0?ooo`030000003oool0oooo07`0oooo0P00001I0?ooo`030000003oool0oooo02T0oooo
0P00001n0?ooo`030000003oool0oooo01<0oooo000D0?ooo`030000003oool0oooo07d0oooo00<0
00000?ooo`3oool0E`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo07h0oooo00<0
00000?ooo`3oool04`3oool001@0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`1G
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`0C
0?ooo`00503oool00`000000oooo0?ooo`1o0?ooo`030000003oool0oooo05H0oooo00<000000?oo
o`3oool09P3oool00`000000oooo0?ooo`1o0?ooo`030000003oool0oooo01<0oooo000D0?ooo`03
0000003oool0oooo0800oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`0T0?ooo`03
0000003oool0oooo0800oooo00<000000?ooo`3oool04`3oool001D0oooo00<000000?ooo`3oool0
P03oool2000005H0oooo00<000000?ooo`3oool08P3oool200000880oooo00<000000?ooo`3oool0
503oool001D0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`0D0?ooo`005@3oool0
0`000000oooo0?ooo`220?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`210?ooo`030000003oool0oooo01D0oooo000F0?ooo`030000003oool0oooo
0880oooo0P00001D0?ooo`030000003oool0oooo01h0oooo0P0000240?ooo`030000003oool0oooo
01D0oooo000F0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0D@3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool05@3oool001L0oooo
00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`0F0?ooo`005`3oool00`000000oooo0?oo
o`250?ooo`800000D@3oool00`000000oooo0?ooo`0I0?ooo`800000Q`3oool00`000000oooo0?oo
o`0F0?ooo`005`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo0500oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo01L0oooo000H0?oo
o`030000003oool0oooo08H0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo08H0oooo00<000000?ooo`3oool05`3oool001P0oooo00<000000?ooo`3o
ool0Q`3oool2000004l0oooo00<000000?ooo`3oool05@3oool2000008P0oooo00<000000?ooo`3o
ool0603oool001T0oooo00<000000?ooo`3oool0R03oool2000004h0oooo00<000000?ooo`3oool0
4P3oool2000008X0oooo00<000000?ooo`3oool0603oool001T0oooo00<000000?ooo`3oool0RP3o
ool2000004`0oooo00<000000?ooo`000000403oool2000008`0oooo00<000000?ooo`3oool0603o
ool001X0oooo00<000000?ooo`3oool0R`3oool2000004X0oooo0`00000>0?ooo`800000S@3oool0
0`000000oooo0?ooo`0I0?ooo`006P3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo
04D0oooo1@00000=0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool06@3oool001X0
oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`150?ooo`@00000303oool00`000000
oooo0?ooo`2=0?ooo`030000003oool0oooo01X0oooo000K0?ooo`030000003oool0oooo08h0oooo
0P0000160?ooo`<000002P3oool200000900oooo00<000000?ooo`3oool06P3oool001/0oooo00<0
00000?ooo`3oool0T03oool2000004@0oooo100000070?ooo`800000T@3oool00`000000oooo0?oo
o`0K0?ooo`00703oool00`000000oooo0?ooo`2A0?ooo`800000@`3oool3000000D0oooo0P00002C
0?ooo`030000003oool0oooo01/0oooo000L0?ooo`030000003oool0oooo09<0oooo1@00000o0?oo
o`L00000U@3oool00`000000oooo0?ooo`0K0?ooo`007@3oool00`000000oooo0?ooo`2G0?ooo`80
0000?@3oool2000009T0oooo00<000000?ooo`3oool0703oool001d0oooo00<000000?ooo`3oool0
V@3oool2000003T0oooo0P00002K0?ooo`030000003oool0oooo01`0oooo000M0?ooo`030000003o
ool0oooo09/0oooo1000000a0?ooo`@00000W03oool00`000000oooo0?ooo`0M0?ooo`007P3oool0
0`000000oooo0?ooo`2N0?ooo`D000009`3oool500000:00oooo00<000000?ooo`3oool07@3oool0
01h0oooo00<000000?ooo`3oool0X`3oool4000001l0oooo1000002T0?ooo`030000003oool0oooo
01h0oooo000O0?ooo`030000003oool0oooo0:H0oooo1`00000A0?ooo`L00000Z03oool00`000000
oooo0?ooo`0N0?ooo`007`3oool00`000000oooo0?ooo`2]0?oooa400000[`3oool00`000000oooo
0?ooo`0N0?ooo`00803oool00`000000oooo0?ooo`3o0?ooof`0oooo00<000000?ooo`3oool07`3o
ool00200oooo00<000000?ooo`3oool0o`3ooom/0?ooo`030000003oool0oooo01l0oooo000Q0?oo
o`030000003oool0oooo0?l0ooooJP3oool00`000000oooo0?ooo`0P0?ooo`008P3oool00`000000
oooo0?ooo`3o0?ooofP0oooo00<000000?ooo`3oool08@3oool00280oooo00<000000?ooo`3oool0
o`3ooomW0?ooo`030000003oool0oooo0280oooo000S0?ooo`030000003oool0oooo0?l0ooooIP3o
ool00`000000oooo0?ooo`0R0?ooo`008`3oool00`000000oooo0?ooo`3o0?ooofD0oooo00<00000
0?ooo`3oool08`3oool002@0oooo00<000000?ooo`3oool0o`3ooomT0?ooo`030000003oool0oooo
02<0oooo000T0?ooo`030000003oool0oooo0?l0ooooI03oool00`000000oooo0?ooo`0S0?ooo`00
9@3oool00`000000oooo0?ooo`3o0?ooof80oooo00<000000?ooo`3oool0903oool002D0oooo00<0
00000?ooo`3oool0o`3ooomR0?ooo`030000003oool0oooo02@0oooo000V0?ooo`030000003oool0
oooo0?l0ooooH03oool00`000000oooo0?ooo`0U0?ooo`009`3oool00`000000oooo0?ooo`3o0?oo
oeh0oooo00<000000?ooo`3oool09P3oool002L0oooo00<000000?ooo`3oool0o`3ooomM0?ooo`03
0000003oool0oooo02L0oooo000X0?ooo`030000003oool0oooo0?l0ooooG03oool00`000000oooo
0?ooo`0W0?ooo`00:03oool00`000000oooo0?ooo`3o0?oooe/0oooo00<000000?ooo`3oool0:03o
ool002T0oooo00<000000?ooo`3oool0o`3ooomJ0?ooo`030000003oool0oooo02P0oooo000Z0?oo
o`030000003oool0oooo0?l0ooooF03oool00`000000oooo0?ooo`0Y0?ooo`00:`3oool00`000000
oooo0?ooo`3o0?oooeH0oooo00<000000?ooo`3oool0:P3oool002/0oooo00<000000?ooo`3oool0
o`3ooomE0?ooo`030000003oool0oooo02/0oooo000/0?ooo`030000003oool0oooo0?l0ooooE03o
ool00`000000oooo0?ooo`0[0?ooo`00;03oool00`000000oooo0?ooo`3o0?oooe@0oooo00<00000
0?ooo`3oool0:`3oool002d0oooo00<000000?ooo`3oool0o`3ooomB0?ooo`030000003oool0oooo
02`0oooo000]0?ooo`030000003oool0oooo0?l0ooooDP3oool00`000000oooo0?ooo`0/0?ooo`00
;P3oool00`000000oooo0?ooo`3o0?oooe00oooo00<000000?ooo`3oool0;@3oool002l0oooo00<0
00000?ooo`3oool0o`3ooom>0?ooo`030000003oool0oooo02h0oooo000_0?ooo`030000003oool0
oooo0?l0ooooC@3oool00`000000oooo0?ooo`0_0?ooo`00<03oool00`000000oooo0?ooo`3o0?oo
od`0oooo00<000000?ooo`3oool0;`3oool00340oooo00<000000?ooo`3oool0o`3ooom:0?ooo`03
0000003oool0oooo0300oooo000b0?ooo`030000003oool0oooo0?l0ooooB03oool00`000000oooo
0?ooo`0a0?ooo`00<P3oool00`000000oooo0?ooo`3o0?ooodL0oooo00<000000?ooo`3oool0<P3o
ool003<0oooo00<000000?ooo`3oool0o`3ooom60?ooo`030000003oool0oooo0380oooo000d0?oo
o`030000003oool0oooo0?l0ooooA03oool00`000000oooo0?ooo`0c0?ooo`00=@3oool00`000000
oooo0?ooo`3o0?oood80oooo00<000000?ooo`3oool0=03oool003H0oooo00<000000?ooo`3oool0
o`3ooom00?ooo`030000003oool0oooo03D0oooo000g0?ooo`030000003oool0oooo0?l0oooo?P3o
ool00`000000oooo0?ooo`0f0?ooo`00=`3oool00`000000oooo0?ooo`3o0?ooocd0oooo00<00000
0?ooo`3oool0=`3oool003P0oooo00<000000?ooo`3oool0o`3oooll0?ooo`030000003oool0oooo
03L0oooo000i0?ooo`030000003oool0oooo0?l0oooo>P3oool00`000000oooo0?ooo`0h0?ooo`00
>P3oool00`000000oooo0?ooo`3o0?ooocP0oooo00<000000?ooo`3oool0>@3oool003X0oooo00<0
00000?ooo`3oool0o`3ooolg0?ooo`030000003oool0oooo03X0oooo000k0?ooo`030000003oool0
oooo0?l0oooo=P3oool00`000000oooo0?ooo`0j0?ooo`00?03oool00`000000oooo0?ooo`3o0?oo
oc@0oooo00<000000?ooo`3oool0>`3oool003d0oooo00<000000?ooo`3oool0o`3ooolb0?ooo`03
0000003oool0oooo03`0oooo000n0?ooo`030000003oool0oooo0?l0oooo<03oool00`000000oooo
0?ooo`0m0?ooo`00?`3oool00`000000oooo0?ooo`3o0?ooobh0oooo00<000000?ooo`3oool0?P3o
ool00400oooo00<000000?ooo`3oool0o`3oool/0?ooo`030000003oool0oooo03l0oooo00110?oo
o`030000003oool0oooo0?l0oooo:P3oool00`000000oooo0?ooo`100?ooo`00@P3oool00`000000
oooo0?ooo`3o0?ooobP0oooo00<000000?ooo`3oool0@@3oool004<0oooo00<000000?ooo`3oool0
o`3ooolV0?ooo`030000003oool0oooo0480oooo00140?ooo`030000003oool0oooo0?l0oooo903o
ool00`000000oooo0?ooo`130?ooo`00A@3oool00`000000oooo0?ooo`3o0?ooob80oooo00<00000
0?ooo`3oool0A03oool004H0oooo00<000000?ooo`3oool0o`3ooolP0?ooo`030000003oool0oooo
04D0oooo00170?ooo`030000003oool0oooo0?l0oooo7P3oool00`000000oooo0?ooo`160?ooo`00
B03oool00`000000oooo0?ooo`3o0?oooa`0oooo00<000000?ooo`3oool0A`3oool004T0oooo00<0
00000?ooo`3oool0o`3ooolJ0?ooo`030000003oool0oooo04P0oooo001:0?ooo`030000003oool0
oooo0?l0oooo603oool00`000000oooo0?ooo`190?ooo`00B`3oool00`000000oooo0?ooo`3o0?oo
oaH0oooo00<000000?ooo`3oool0BP3oool004`0oooo00<000000?ooo`3oool0o`3ooolD0?ooo`03
0000003oool0oooo04/0oooo001=0?ooo`030000003oool0oooo0?l0oooo4P3oool00`000000oooo
0?ooo`1<0?ooo`00CP3oool00`000000oooo0?ooo`3o0?oooa00oooo00<000000?ooo`3oool0C@3o
ool004l0oooo00<000000?ooo`3oool0o`3oool>0?ooo`030000003oool0oooo04h0oooo001@0?oo
o`030000003oool0oooo0?l0oooo303oool00`000000oooo0?ooo`1?0?ooo`00D@3oool00`000000
oooo0?ooo`3o0?ooo`X0oooo00<000000?ooo`3oool0D03oool00580oooo0P00003o0?ooo`P0oooo
0P00001C0?ooo`00E03oool00`000000oooo0?ooo`3o0?ooo`@0oooo00<000000?ooo`3oool0D`3o
ool005D0oooo00<000000?ooo`3oool0o`3oool20?ooo`030000003oool0oooo05@0oooo001F0?oo
o`800000o`3oool2000005L0oooo001H0?ooo`030000003oool0oooo0?/0oooo00<000000?ooo`3o
ool0E`3oool005T0oooo00<000000?ooo`3oool0n@3oool00`000000oooo0?ooo`1H0?ooo`00FP3o
ool00`000000oooo0?ooo`3g0?ooo`030000003oool0oooo05T0oooo001K0?ooo`030000003oool0
oooo0?D0oooo00<000000?ooo`3oool0FP3oool005`0oooo0P00003c0?ooo`800000G@3oool005h0
oooo00<000000?ooo`3oool0k`3oool00`000000oooo0?ooo`1M0?ooo`00G`3oool00`000000oooo
0?ooo`3]0?ooo`030000003oool0oooo05h0oooo001P0?ooo`800000j`3oool200000640oooo001R
0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0H@3oool006<0oooo00<000000?oo
o`3oool0i03oool2000006@0oooo001T0?ooo`<00000h@3oool2000006H0oooo001W0?ooo`800000
g@3oool2000006P0oooo001Y0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0J03o
ool006X0oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`1Y0?ooo`00J`3oool20000
0=D0oooo0P00001/0?ooo`00K@3oool200000=40oooo0P00001^0?ooo`00K`3oool00`000000oooo
0?ooo`3=0?ooo`030000003oool0oooo06h0oooo001`0?ooo`800000b`3oool200000740oooo001b
0?ooo`800000a`3oool2000007<0oooo001d0?ooo`800000``3oool2000007D0oooo001f0?ooo`80
0000_`3oool2000007L0oooo001h0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0
M`3oool007T0oooo0P00002i0?ooo`800000NP3oool007/0oooo0P00002e0?ooo`800000O03oool0
07d0oooo0P00002a0?ooo`800000OP3oool007l0oooo0`00002[0?ooo`<00000P03oool00880oooo
0P00002W0?ooo`800000P`3oool008@0oooo0P00002S0?ooo`800000Q@3oool008H0oooo0`00002M
0?ooo`<00000Q`3oool008T0oooo0P00002I0?ooo`800000RP3oool008/0oooo0P00002E0?ooo`80
0000S03oool008d0oooo0`00002?0?ooo`<00000SP3oool00900oooo0P00002;0?ooo`800000T@3o
ool00980oooo0`0000250?ooo`<00000T`3oool009D0oooo1000001l0?ooo`D00000UP3oool009T0
oooo0`00001g0?ooo`800000V`3oool009`0oooo0P00001c0?ooo`800000W@3oool009h0oooo1@00
001Y0?ooo`D00000W`3oool00:<0oooo1@00001Q0?ooo`<00000Y03oool00:P0oooo0`00001I0?oo
o`D00000Y`3oool00:/0oooo1@00001?0?ooo`D00000[03oool00;00oooo1@0000150?ooo`D00000
/@3oool00;D0oooo1`00000g0?ooo`L00000]P3oool00;`0oooo2`00000Q0?ooo`/00000_@3oool0
0<L0oooo8@0000380?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.525005, -0.525005, \
0.00365433, 0.00365433}}]
}, Closed]],

Cell[BoxData[
    \(\(VelocityPlot2[LabVelocities, vCM, cmVelocities];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[RGBColor[0, 1, 0]], plot1, 
        Graphics[Circle[vCM, magnitude[vi]\/2]], Graphics[RGBColor[1, 0, 0]], 
        plot2, Graphics[Circle[vCM, magnitude[vf]\/2]], 
        AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.5 0.952381 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 1 0 r
.5 Mabswid
[ ] 0 setdash
.02381 .5 m
.97619 .5 L
s
% Start of user PostScript
0. 0. 1. 0. MAarrowhead1
% End of user PostScript
.02381 .5 m
.02381 .5 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.02381 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.97619 .5 L
s
% Start of user PostScript
0.5 0. 1. 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.02381 .5 L
s
% Start of user PostScript
0.5 0. 0. 0. MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
1 0 0 r
.02381 .5 m
.57323 .69997 L
s
% Start of user PostScript
0. 0. 0.57689064 0.20997102 MAarrowhead1
% End of user PostScript
.02381 .5 m
.42677 .30003 L
s
% Start of user PostScript
0. 0. 0.42310936 -0.20997102 MAarrowhead1
% End of user PostScript
.02381 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.57323 .69997 L
s
% Start of user PostScript
0.5 0. 0.57689064 0.20997102 MAarrowhead1
% End of user PostScript
.5 .5 m
.42677 .30003 L
s
% Start of user PostScript
0.5 0. 0.42310936 -0.20997102 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .21296 0 365.73 arc
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ003o0<P0oooo002m0?ooo`X00?l08@3oool:003o0;h0oooo002e0?ooo`P00?l0
=@3oool8003o0;H0oooo002`0?ooo`D00?l0A@3oool5003o0;40oooo002[0?ooo`D00?l0C`3oool5
003o0:`0oooo002V0?ooo`D00?l0F@3oool2003o0:X0oooo002T0?ooo`800?l0H03oool5003o0:D0
oooo002O0?ooo`D00?l0I`3oool5003o0:00oooo002M0?ooo`800?l0L@3oool3003o09d0oooo002J
0?ooo`<00?l0MP3oool2003o09/0oooo002E0?ooo`D00?l0N`3oool5003o09H0oooo002C0?ooo`80
0?l0Q@3oool2003o09@0oooo002@0?ooo`<00?l0R@3oool3003o0940oooo002>0?ooo`800?l0S`3o
ool2003o08l0oooo002<0?ooo`800?l0T`3oool2003o08d0oooo00290?ooo`<00?l0U`3oool3003o
08X0oooo00270?ooo`800?l0W@3oool2003o08P0oooo00250?ooo`800?l0X@3oool2003o08H0oooo
00220?ooo`<00?l0Y@3oool3003o08<0oooo00200?ooo`800?l0Z`3oool2003o0840oooo001n0?oo
o`800?l0[`3oool2003o07l0oooo001k0?ooo`<00?l0/`3oool3003o07`0oooo001i0?ooo`800?l0
^@3oool2003o07X0oooo001h0?ooo`03003o003oool0oooo0;/0oooo00<00?l00?ooo`3oool0M`3o
ool007L0oooo00<00?l00?ooo`3oool0_@3oool00`00o`00oooo0?ooo`1f0?ooo`00M@3oool2003o
0<40oooo0P00o`1f0?ooo`00LP3oool3003o0<D0oooo0`00o`1c0?ooo`00L03oool2003o0</0oooo
0P00o`1a0?ooo`00K`3oool00`00o`00oooo0?ooo`3=0?ooo`03003o003oool0oooo06h0oooo001^
0?ooo`03003o003oool0oooo0<l0oooo00<00?l00?ooo`3oool0K@3oool006`0oooo0P00o`3C0?oo
o`800?l0K@3oool006X0oooo0P00o`3G0?ooo`800?l0J`3oool006T0oooo00<00?l00?ooo`3oool0
f@3oool00`00o`00oooo0?ooo`1X0?ooo`00J03oool00`00o`00oooo0?ooo`3K0?ooo`800?l0J03o
ool006D0oooo0`00o`3P0?ooo`800?l0IP3oool006<0oooo0P00o`3U0?ooo`800?l0I03oool00680
oooo00<00?l00?ooo`3oool0i`3oool00`00o`00oooo0?ooo`1Q0?ooo`00H@3oool00`00o`00oooo
0?ooo`3Y0?ooo`03003o003oool0oooo0600oooo001O0?ooo`800?l0k@3oool2003o0600oooo001N
0?ooo`03003o003oool0oooo0>l0oooo00<00?l00?ooo`3oool0G@3oool005d0oooo00<00?l00?oo
o`3oool0l@3oool00`00o`00oooo0?ooo`1L0?ooo`00F`3oool2003o0?D0oooo0P00o`1L0?ooo`00
FP3oool00`00o`00oooo0?ooo`3g0?ooo`03003o003oool0oooo05T0oooo001I0?ooo`03003o003o
ool0oooo0?T0oooo00<00?l00?ooo`3oool0F03oool005P0oooo00<00?l00?ooo`3oool0n`3oool0
0`00o`00oooo0?ooo`1G0?ooo`00E`3oool00`00o`00oooo0?ooo`3m0?ooo`03003o003oool0oooo
05H0oooo001E0?ooo`800?l0o`3oool20?ooo`800?l0EP3oool005@0oooo00<00?l00?ooo`3oool0
o`3oool40?ooo`03003o003oool0oooo05<0oooo001C0?ooo`03003o003oool0oooo0?l0oooo1P3o
ool00`00o`00oooo0?ooo`1B0?ooo`00D@3oool2003o0?l0oooo2P3oool2003o0580oooo001@0?oo
o`03003o003oool0oooo0?l0oooo303oool00`00o`00oooo0?ooo`1?0?ooo`00C`3oool00`00o`00
oooo0?ooo`3o0?ooo`h0oooo00<00?l00?ooo`3oool0CP3oool004h0oooo00<00?l00?ooo`3oool0
o`3oool@0?ooo`03003o003oool0oooo04d0oooo001=0?ooo`03003o003oool0oooo0?l0oooo4P3o
ool00`00o`00oooo0?ooo`1<0?ooo`00C03oool00`00o`00oooo0?ooo`3o0?oooa@0oooo00<00?l0
0?ooo`3oool0B`3oool004/0oooo00<00?l00?ooo`3oool0o`3ooolF0?ooo`03003o003oool0oooo
04X0oooo001:0?ooo`03003o003oool0oooo0?l0oooo603oool00`00o`00oooo0?ooo`190?ooo`00
B@3oool00`00o`00oooo0?ooo`3o0?oooaX0oooo00<00?l00?ooo`3oool0B03oool004P0oooo00<0
0?l00?ooo`3oool0o`3ooolL0?ooo`03003o003oool0oooo04L0oooo00170?ooo`03003o003oool0
oooo0?l0oooo7P3oool00`00o`00oooo0?ooo`160?ooo`00AP3oool00`00o`00oooo0?ooo`3o0?oo
ob00oooo00<00?l00?ooo`3oool0A@3oool004D0oooo00<00?l00?ooo`3oool0o`3ooolR0?ooo`03
003o003oool0oooo04@0oooo00140?ooo`03003o003oool0oooo0?l0oooo903oool00`00o`00oooo
0?ooo`130?ooo`00@`3oool00`00o`00oooo0?ooo`3o0?ooobH0oooo00<00?l00?ooo`3oool0@P3o
ool00480oooo00<00?l00?ooo`3oool0o`3ooolX0?ooo`03003o003oool0oooo0440oooo00110?oo
o`03003o003oool0oooo0?l0oooo:P3oool00`00o`00oooo0?ooo`100?ooo`00@03oool00`00o`00
oooo0?ooo`3o0?ooob`0oooo00<00?l00?ooo`3oool0?`3oool003l0oooo00<00?l00?ooo`3oool0
o`3oool^0?ooo`03003o003oool0oooo03h0oooo000n0?ooo`03003o003oool0oooo0?l0oooo<03o
ool00`00o`00oooo0?ooo`0m0?ooo`00?@3oool00`00o`00oooo0?ooo`3o0?oooc80oooo00<00?l0
0?ooo`3oool0?03oool003`0oooo00<00?l00?ooo`3oool0o`3ooold0?ooo`03003o003oool0oooo
03/0oooo000k0?ooo`03003o003oool0oooo0?l0oooo=P3oool00`00o`00oooo0?ooo`0j0?ooo`00
>P3oool00`00o`00oooo0?ooo`3o0?ooocL0oooo00<00?l00?ooo`3oool0>P3oool003X0oooo00<0
0?l00?ooo`3oool0o`3ooolh0?ooo`03003o003oool0oooo03T0oooo000i0?ooo`03003o003oool0
oooo0?l0oooo>P3oool00`00o`00oooo0?ooo`0h0?ooo`00>03oool00`00o`00oooo0?ooo`3o0?oo
oc`0oooo00<00?l00?ooo`3oool0=`3oool003L0oooo00<00?l00?ooo`3oool0o`3ooolm0?ooo`03
003o003oool0oooo03L0oooo000g0?ooo`03003o003oool0oooo0?l0oooo?P3oool00`00o`00oooo
0?ooo`0f0?ooo`00=P3oool00`00o`00oooo0?ooo`3o0?oood00oooo00<00?l00?ooo`3oool0=@3o
ool003D0oooo00<00?l00?ooo`3oool0o`3ooom20?ooo`03003o003oool0oooo03@0oooo000d0?oo
o`03003o003oool0oooo0?l0ooooA03oool00`00o`00oooo0?ooo`0c0?ooo`00<`3oool00`00o`00
oooo0?ooo`3o0?ooodH0oooo00<00?l00?ooo`3oool0<P3oool00380oooo00<00?l00?ooo`3oool0
o`3ooom70?ooo`03003o003oool0oooo0380oooo000b0?ooo`03003o003oool0oooo0?l0ooooB03o
ool00`00o`00oooo0?ooo`0a0?ooo`00<@3oool00`00o`00oooo0?ooo`3o0?ooodX0oooo00<00?l0
0?ooo`3oool0<03oool00300oooo00<00?l00?ooo`3oool0o`3ooom<0?ooo`03003o003oool0oooo
02l0oooo000_0?ooo`03003o003oool0oooo0?l0ooooC@3oool00`00o`00oooo0?ooo`0_0?ooo`00
;`3oool00`00o`00oooo0?ooo`3o0?ooodh0oooo00<00?l00?ooo`3oool0;P3oool002h0oooo00<0
0?l00?ooo`3oool0o`3ooom@0?ooo`03003o003oool0oooo02d0oooo000]0?ooo`03003o003oool0
oooo0?l0ooooDP3oool00`00o`00oooo0?ooo`0/0?ooo`00;03oool00`00o`00oooo0?ooo`3o0?oo
oe<0oooo00<00?l00?ooo`3oool0;03oool002`0oooo00<00?l00?ooo`3oool0o`3ooomD0?ooo`03
003o003oool0oooo02/0oooo000/0?ooo`03003o003oool0oooo0?l0ooooE03oool00`00o`00oooo
0?ooo`0[0?ooo`00:`3oool00`00o`00oooo0?ooo`3o0?oooeD0oooo00<00?l00?ooo`3oool0:`3o
ool002/0oooo00<00?l00?ooo`3oool0o`3ooomF0?ooo`03003o003oool0oooo02X0oooo000Z0?oo
o`03003o003oool0oooo0?l0ooooF03oool00`00o`00oooo0?ooo`0Y0?ooo`00:@3oool00`00o`00
oooo0?ooo`3o0?oooeX0oooo00<00?l00?ooo`3oool0:03oool002P0oooo00<00?l00?ooo`3oool0
o`3ooomK0?ooo`03003o003oool0oooo02P0oooo000X0?ooo`03003o003oool0oooo0?l0ooooG03o
ool00`00o`00oooo0?ooo`0W0?ooo`009`3oool00`00o`00oooo0?ooo`3o0?oooed0oooo00<00?l0
0?ooo`3oool09`3oool002L0oooo00<00?l00?ooo`3oool0o`3ooomN0?ooo`03003o003oool0oooo
02H0oooo000V0?ooo`03003o003oool0oooo0?l0ooooH03oool00`00o`00oooo0?ooo`0U0?ooo`00
9@3oool00`00o`00oooo0?ooo`3o0?ooof80oooo00<00?l00?ooo`3oool0903oool002D0oooo00<0
0?l00?ooo`3oool0o`3ooomR0?ooo`03003o003oool0oooo02@0oooo000T0?ooo`03003o003oool0
oooo0?l0ooooI03oool00`00o`00oooo0?ooo`0S0?ooo`00903oool00`00o`00oooo0?ooo`3o0?oo
of@0oooo00<00?l00?ooo`3oool08`3oool002<0oooo00<00?l00?ooo`3oool0o`3ooomU0?ooo`03
003o003oool0oooo02<0oooo000S0?ooo`03003o003oool0oooo0?l0ooooIP3oool00`00o`00oooo
0?ooo`0R0?ooo`008P3oool00`00o`00oooo0?ooo`3o0?ooofL0oooo00<00?l00?ooo`3oool08P3o
ool00280oooo00<00?l00?ooo`3oool0o`3ooomX0?ooo`03003o003oool0oooo0240oooo000Q0?oo
o`03003o003oool0oooo0?l0ooooJP3oool00`00o`00oooo0?ooo`0P0?ooo`00803oool00`00o`00
oooo0?ooo`3o0?ooof`0oooo00<00?l00?ooo`3oool07`3oool00200oooo00<00?l00?ooo`3oool0
o`3ooom/0?ooo`03003o003oool0oooo01l0oooo000O0?ooo`03003o003oool0oooo0:h0oooo3`3o
002`0?ooo`03003o003oool0oooo01h0oooo000O0?ooo`03003o003oool0oooo0:L0oooo1`3o000?
0?ooo`L0o`00Z@3oool00`00o`00oooo0?ooo`0N0?ooo`007P3oool00`00o`00oooo0?ooo`2T0?oo
o`@0o`007@3oool40?l00:D0oooo00<00?l00?ooo`3oool07P3oool001h0oooo00<00?l00?ooo`3o
ool0W`3oool50?l002D0oooo1@3o002Q0?ooo`03003o003oool0oooo01d0oooo000M0?ooo`03003o
003oool0oooo09`0oooo103o000_0?ooo`@0o`00W@3oool00`00o`00oooo0?ooo`0M0?ooo`007@3o
ool00`00o`00oooo0?ooo`2J0?ooo`80o`00=`3oool20?l009`0oooo00<00?l00?ooo`3oool0703o
ool001d0oooo00<00?l00?ooo`3oool0U03oool60?l003/0oooo0`3o002I0?ooo`03003o003oool0
oooo01`0oooo000L0?ooo`03003o003oool0oooo0940oooo2@3o000o0?ooo`@0o`00UP3oool00`00
o`00oooo0?ooo`0K0?ooo`00703oool00`00o`00oooo0?ooo`2A0?ooo`H0o`0000@0oooo0?l0003o
0000o`00@P3oool20?l009@0oooo00<00?l00?ooo`3oool06`3oool001/0oooo00<00?l00?ooo`3o
ool0T@3oool60?l00080oooo0`3o00140?ooo`80o`00TP3oool00`00o`00oooo0?ooo`0K0?ooo`00
6`3oool00`00o`00oooo0?ooo`2?0?ooo`<0o`0000@0oooo0?l0003o0000o`00103oool30?l004D0
oooo0P3o002A0?ooo`03003o003oool0oooo01X0oooo000J0?ooo`03003o003oool0oooo08h0oooo
0`3o00040?ooo`030?l0003oool0oooo00<0oooo103o00160?ooo`80o`00S`3oool00`00o`00oooo
0?ooo`0J0?ooo`006P3oool00`00o`00oooo0?ooo`2=0?ooo`80o`00303oool50?l004L0oooo00<0
o`000?ooo`3oool0S@3oool00`00o`00oooo0?ooo`0I0?ooo`006P3oool00`00o`00oooo0?ooo`2;
0?ooo`80o`003P3oool30?l004X0oooo00<0o`000?ooo`3oool0S03oool00`00o`00oooo0?ooo`0I
0?ooo`006@3oool00`00o`00oooo0?ooo`2:0?ooo`<0o`004@3oool00`3o0000oooo0?ooo`190?oo
o`80o`00S@3oool00`00o`00oooo0?ooo`0H0?ooo`006@3oool00`00o`00oooo0?ooo`280?ooo`<0
o`004`3oool00`3o0000oooo0?ooo`1;0?ooo`80o`00R`3oool00`00o`00oooo0?ooo`0H0?ooo`00
603oool00`00o`00oooo0?ooo`270?ooo`<0o`005P3oool00`3o0000oooo0?ooo`1<0?ooo`80o`00
R@3oool00`00o`00oooo0?ooo`0H0?ooo`00603oool00`00o`00oooo0?ooo`250?ooo`<0o`00603o
ool00`3o0000oooo0?ooo`1>0?ooo`80o`00R03oool00`00o`00oooo0?ooo`0G0?ooo`005`3oool0
0`00o`00oooo0?ooo`240?ooo`<0o`006P3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0
oooo08H0oooo00<00?l00?ooo`3oool05`3oool001L0oooo00<00?l00?ooo`3oool0PP3oool20?l0
0080oooo00<0o`000?ooo`3oool06@3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0oooo
08H0oooo00<00?l00?ooo`3oool05P3oool001L0oooo00<00?l00?ooo`3oool0P03oool20?l00080
oooo0P3o000L0?ooo`030?l0003oool0oooo0500oooo0P3o00260?ooo`03003o003oool0oooo01H0
oooo000F0?ooo`03003o003oool0oooo07l0oooo0P3o00030?ooo`030?l0003oool0oooo01d0oooo
00<0o`000?ooo`3oool0D@3oool00`3o0000oooo0?ooo`240?ooo`03003o003oool0oooo01D0oooo
000F0?ooo`03003o003oool0oooo07d0oooo0P3o00040?ooo`030?l0003oool0oooo01h0oooo00<0
o`000?ooo`3oool0DP3oool00`3o0000oooo0?ooo`230?ooo`03003o003oool0oooo01D0oooo000E
0?ooo`03003o003oool0oooo07`0oooo0P3o00040?ooo`80o`008@3oool00`3o0000oooo0?ooo`1C
0?ooo`80o`00P`3oool00`00o`00oooo0?ooo`0E0?ooo`005@3oool00`00o`00oooo0?ooo`1j0?oo
o`80o`001@3oool00`3o0000oooo0?ooo`0R0?ooo`030?l0003oool0oooo05<0oooo00<0o`000?oo
o`3oool0PP3oool00`00o`00oooo0?ooo`0D0?ooo`005@3oool00`00o`00oooo0?ooo`1h0?ooo`80
o`001`3oool00`3o0000oooo0?ooo`0R0?ooo`030?l0003oool0oooo05@0oooo00<0o`000?ooo`3o
ool0P@3oool00`00o`00oooo0?ooo`0D0?ooo`00503oool00`00o`00oooo0?ooo`1g0?ooo`80o`00
1`3oool20?l002D0oooo00<0o`000?ooo`3oool0E@3oool20?l00880oooo00<00?l00?ooo`3oool0
4`3oool001@0oooo00<00?l00?ooo`3oool0M@3oool20?l000P0oooo00<0o`000?ooo`3oool09P3o
ool00`3o0000oooo0?ooo`1F0?ooo`030?l0003oool0oooo07l0oooo00<00?l00?ooo`3oool04`3o
ool001@0oooo00<00?l00?ooo`3oool0L`3oool20?l000T0oooo00<0o`000?ooo`3oool09`3oool0
0`3o0000oooo0?ooo`1G0?ooo`030?l0003oool0oooo07h0oooo00<00?l00?ooo`3oool04`3oool0
01@0oooo00<00?l00?ooo`3oool0L@3oool20?l000X0oooo00<0o`000?ooo`3oool0:@3oool00`3o
0000oooo0?ooo`1F0?ooo`030?l0003oool0oooo07h0oooo00<00?l00?ooo`3oool04`3oool001<0
oooo00<00?l00?ooo`3oool0L03oool20?l000`0oooo00<0o`000?ooo`3oool0:@3oool00`3o0000
oooo0?ooo`1G0?ooo`030?l0003oool0oooo07d0oooo00<00?l00?ooo`3oool04`3oool001<0oooo
00<00?l00?ooo`3oool0KP3oool20?l000`0oooo0P3o000/0?ooo`030?l0003oool0oooo05P0oooo
0P3o001n0?ooo`03003o003oool0oooo0180oooo000C0?ooo`03003o003oool0oooo06`0oooo0P3o
000=0?ooo`030?l0003oool0oooo02d0oooo00<0o`000?ooo`3oool0F03oool00`3o0000oooo0?oo
o`1l0?ooo`03003o003oool0oooo0180oooo000B0?ooo`03003o003oool0oooo06/0oooo0P3o000?
0?ooo`030?l0003oool0oooo02d0oooo00<0o`000?ooo`3oool0F@3oool00`3o0000oooo0?ooo`1k
0?ooo`03003o003oool0oooo0180oooo000B0?ooo`03003o003oool0oooo06T0oooo0P3o000@0?oo
o`030?l0003oool0oooo02h0oooo00<0o`000?ooo`3oool0FP3oool00`3o0000oooo0?ooo`1k0?oo
o`03003o003oool0oooo0140oooo000A0?ooo`03003o003oool0oooo06P0oooo0P3o000A0?ooo`03
0?l0003oool0oooo0300oooo00<0o`000?ooo`3oool0FP3oool00`3o0000oooo0?ooo`1j0?ooo`03
003o003oool0oooo0140oooo000A0?ooo`03003o003oool0oooo06H0oooo0P3o000B0?ooo`030?l0
003oool0oooo0340oooo00<0o`000?ooo`3oool0FP3oool00`3o0000oooo0?ooo`1k0?ooo`03003o
003oool0oooo0100oooo000A0?ooo`03003o003oool0oooo06@0oooo0P3o000D0?ooo`030?l0003o
ool0oooo0380oooo00<0o`000?ooo`3oool0FP3oool00`3o0000oooo0?ooo`1j0?ooo`03003o003o
ool0oooo0100oooo000A0?ooo`03003o003oool0oooo0680oooo0P3o000D0?ooo`80o`00=@3oool0
0`3o0000oooo0?ooo`1K0?ooo`80o`00NP3oool00`00o`00oooo0?ooo`0@0?ooo`004@3oool00`00
o`00oooo0?ooo`1P0?ooo`80o`005@3oool00`3o0000oooo0?ooo`0e0?ooo`030?l0003oool0oooo
05`0oooo00<0o`000?ooo`3oool0N03oool00`00o`00oooo0?ooo`0@0?ooo`00403oool00`00o`00
oooo0?ooo`1O0?ooo`80o`005`3oool00`3o0000oooo0?ooo`0f0?ooo`030?l0003oool0oooo05`0
oooo00<0o`000?ooo`3oool0M`3oool00`00o`00oooo0?ooo`0@0?ooo`00403oool00`00o`00oooo
0?ooo`1M0?ooo`80o`00603oool00`3o0000oooo0?ooo`0g0?ooo`030?l0003oool0oooo05`0oooo
00<0o`000?ooo`3oool0N03oool00`00o`00oooo0?ooo`0?0?ooo`00403oool00`00o`00oooo0?oo
o`1K0?ooo`80o`006P3oool00`3o0000oooo0?ooo`0g0?ooo`030?l0003oool0oooo05d0oooo00<0
o`000?ooo`3oool0M`3oool00`00o`00oooo0?ooo`0?0?ooo`00403oool00`00o`00oooo0?ooo`1I
0?ooo`80o`006`3oool00`3o0000oooo0?ooo`0i0?ooo`030?l0003oool0oooo05`0oooo00<0o`00
0?ooo`3oool0N03oool00`00o`00oooo0?ooo`0>0?ooo`00403oool00`00o`00oooo0?ooo`1G0?oo
o`80o`007@3oool00`3o0000oooo0?ooo`0i0?ooo`030?l0003oool0oooo05d0oooo00<0o`000?oo
o`3oool0M`3oool00`00o`00oooo0?ooo`0>0?ooo`003`3oool00`00o`00oooo0?ooo`1F0?ooo`80
o`007P3oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo05d0oooo00<0o`000?ooo`3o
ool0MP3oool00`00o`00oooo0?ooo`0>0?ooo`003`3oool00`00o`00oooo0?ooo`1D0?ooo`80o`00
803oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0
M@3oool00`00o`00oooo0?ooo`0>0?ooo`003`3oool00`00o`00oooo0?ooo`1B0?ooo`80o`008@3o
ool00`3o0000oooo0?ooo`0l0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0M@3o
ool00`00o`00oooo0?ooo`0>0?ooo`003P3oool00`00o`00oooo0?ooo`1A0?ooo`80o`008P3oool0
0`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0M@3oool0
0`00o`00oooo0?ooo`0=0?ooo`003P3oool00`00o`00oooo0?ooo`1?0?ooo`80o`008`3oool00`3o
0000oooo0?ooo`0o0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0M@3oool00`00
o`00oooo0?ooo`0=0?ooo`003P3oool00`00o`00oooo0?ooo`1=0?ooo`80o`009@3oool00`3o0000
oooo0?ooo`0o0?ooo`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0M03oool00`00o`00
oooo0?ooo`0=0?ooo`003P3oool00`00o`00oooo0?ooo`1;0?ooo`80o`009P3oool00`3o0000oooo
0?ooo`110?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0M03oool00`00o`00oooo
0?ooo`0=0?ooo`003P3oool00`00o`00oooo0?ooo`190?ooo`80o`00:03oool00`3o0000oooo0?oo
o`110?ooo`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0L`3oool00`00o`00oooo0?oo
o`0=0?ooo`003@3oool00`00o`00oooo0?ooo`180?ooo`80o`00:@3oool00`3o0000oooo0?ooo`12
0?ooo`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0M03oool00`00o`00oooo0?ooo`0<
0?ooo`003@3oool00`00o`00oooo0?ooo`160?ooo`80o`00:`3oool00`3o0000oooo0?ooo`130?oo
o`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0L`3oool00`00o`00oooo0?ooo`0<0?oo
o`003@3oool00`00o`00oooo0?ooo`140?ooo`80o`00;03oool00`3o0000oooo0?ooo`140?ooo`03
0?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0L`3oool00`00o`00oooo0?ooo`0<0?ooo`00
3@3oool00`00o`00oooo0?ooo`120?ooo`80o`00;P3oool00`3o0000oooo0?ooo`150?ooo`030?l0
003oool0oooo05l0oooo00<0o`000?ooo`3oool0LP3oool00`00o`00oooo0?ooo`0<0?ooo`003@3o
ool00`00o`00oooo0?ooo`100?ooo`80o`00<03oool00`3o0000oooo0?ooo`150?ooo`030?l0003o
ool0oooo05l0oooo00<0o`000?ooo`3oool0LP3oool00`00o`00oooo0?ooo`0<0?ooo`00303oool0
0`00o`00oooo0?ooo`0o0?ooo`80o`00<P3oool00`3o0000oooo0?ooo`150?ooo`030?l0003oool0
oooo05l0oooo00<0o`000?ooo`3oool0L`3oool00`00o`00oooo0?ooo`0;0?ooo`00303oool00`00
o`00oooo0?ooo`0m0?ooo`80o`00=03oool00`3o0000oooo0?ooo`160?ooo`030?l0003oool0oooo
05h0oooo00<0o`000?ooo`3oool0L`3oool00`00o`00oooo0?ooo`0;0?ooo`00303oool00`00o`00
oooo0?ooo`0k0?ooo`80o`00=@3oool00`3o0000oooo0?ooo`170?ooo`030?l0003oool0oooo05l0
oooo00<0o`000?ooo`3oool0LP3oool00`00o`00oooo0?ooo`0;0?ooo`00303oool00`00o`00oooo
0?ooo`0i0?ooo`80o`00=`3oool00`3o0000oooo0?ooo`170?ooo`030?l0003oool0oooo05l0oooo
00<0o`000?ooo`3oool0LP3oool00`00o`00oooo0?ooo`0;0?ooo`00303oool00`00o`00oooo0?oo
o`0g0?ooo`80o`00>03oool00`3o0000oooo0?ooo`190?ooo`030?l0003oool0oooo05h0oooo00<0
o`000?ooo`3oool0LP3oool00`00o`00oooo0?ooo`0;0?ooo`00303oool00`00o`00oooo0?ooo`0e
0?ooo`80o`00>P3oool00`3o0000oooo0?ooo`190?ooo`030?l0003oool0oooo05l0oooo00<0o`00
0?ooo`3oool0L@3oool00`00o`00oooo0?ooo`0;0?ooo`00303oool00`00o`00oooo0?ooo`0c0?oo
o`80o`00>`3oool00`3o0000oooo0?ooo`1;0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?oo
o`3oool0L@3oool00`00o`00oooo0?ooo`0;0?ooo`002`3oool00`00o`00oooo0?ooo`0b0?ooo`80
o`00?@3oool00`3o0000oooo0?ooo`1;0?ooo`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3o
ool0L03oool00`00o`00oooo0?ooo`0;0?ooo`002`3oool00`00o`00oooo0?ooo`0`0?ooo`80o`00
?`3oool00`3o0000oooo0?ooo`1;0?ooo`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0
L@3oool00`00o`00oooo0?ooo`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0^0?ooo`80o`00@@3o
ool00`3o0000oooo0?ooo`1<0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0L@3o
ool00`00o`00oooo0?ooo`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0/0?ooo`80o`00@P3oool0
0`3o0000oooo0?ooo`1=0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0L@3oool0
0`00o`00oooo0?ooo`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0Z0?ooo`80o`00A03oool00`3o
0000oooo0?ooo`1=0?ooo`030?l0003oool0oooo05l0oooo00<0o`000?ooo`3oool0L03oool00`00
o`00oooo0?ooo`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0X0?ooo`80o`00AP3oool00`3o0000
oooo0?ooo`1>0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0L03oool00`00o`00
oooo0?ooo`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0V0?ooo`80o`00B03oool00`3o0000oooo
0?ooo`1>0?ooo`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0L03oool00`00o`00oooo
0?ooo`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0T0?ooo`80o`00BP3oool00`3o0000oooo0?oo
o`1?0?ooo`030?l0003oool0oooo05d0oooo00<0o`000?ooo`3oool0L03oool00`00o`00oooo0?oo
o`0:0?ooo`002`3oool00`00o`00oooo0?ooo`0R0?ooo`80o`00B`3oool00`3o0000oooo0?ooo`1@
0?ooo`030?l0003oool0oooo05d0oooo00<0o`000?ooo`3oool0L03oool00`00o`00oooo0?ooo`0:
0?ooo`002`3oool00`00o`00oooo0?ooo`0P0?ooo`80o`00C@3oool00`3o0000oooo0?ooo`1@0?oo
o`030?l0003oool0oooo05h0oooo00<0o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`0:0?oo
o`002`3oool00`00o`00oooo0?ooo`0N0?ooo`80o`00C`3oool00`3o0000oooo0?ooo`1A0?ooo`03
0?l0003oool0oooo05d0oooo00<0o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`0:0?ooo`00
2P3oool00`00o`00oooo0?ooo`0M0?ooo`80o`00D@3oool00`3o0000oooo0?ooo`1A0?ooo`030?l0
003oool0oooo05d0oooo00<0o`000?ooo`3oool0L03oool00`00o`00oooo0?ooo`090?ooo`002P3o
ool00`00o`00oooo0?ooo`0K0?ooo`80o`00DP3oool00`3o0000oooo0?ooo`1B0?ooo`030?l0003o
ool0oooo05d0oooo00<0o`000?ooo`3oool0L03oool00`00o`00oooo0?ooo`090?ooo`002P3oool0
0`00o`00oooo0?ooo`0I0?ooo`80o`00E03oool00`3o0000oooo0?ooo`1C0?ooo`030?l0003oool0
oooo05d0oooo00<0o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00
o`00oooo0?ooo`0G0?ooo`80o`00EP3oool00`3o0000oooo0?ooo`1C0?ooo`030?l0003oool0oooo
05d0oooo00<0o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00
oooo0?ooo`0E0?ooo`80o`00F03oool00`3o0000oooo0?ooo`1D0?ooo`030?l0003oool0oooo05`0
oooo00<0o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo
0?ooo`0C0?ooo`80o`00FP3oool00`3o0000oooo0?ooo`1D0?ooo`030?l0003oool0oooo05`0oooo
00<0o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?oo
o`0A0?ooo`80o`00G03oool00`3o0000oooo0?ooo`1D0?ooo`030?l0003oool0oooo05`0oooo00<0
o`000?ooo`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`0?
0?ooo`80o`00GP3oool00`3o0000oooo0?ooo`1E0?ooo`030?l0003oool0oooo05/0oooo00<0o`00
0?ooo`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`0=0?oo
o`80o`00G`3oool00`3o0000oooo0?ooo`1F0?ooo`030?l0003oool0oooo05/0oooo00<0o`000?oo
o`3oool0K`3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`0;0?ooo`80
o`00H@3oool00`3o0000oooo0?ooo`1F0?ooo`030?l0003oool0oooo05`0oooo00<0o`000?ooo`3o
ool0KP3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`090?ooo`80o`00
H`3oool00`3o0000oooo0?ooo`1G0?ooo`030?l0003oool0oooo05/0oooo00<0o`000?ooo`3oool0
KP3oool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`070?ooo`80o`00I@3o
ool00`3o0000oooo0?ooo`1G0?ooo`030?l0003oool0oooo05/0oooo00<0o`000?ooo`3oool0KP3o
ool00`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`050?ooo`80o`00I`3oool0
0`3o0000oooo0?ooo`1H0?ooo`030?l0003oool0oooo05X0oooo00<0o`000?ooo`3oool0KP3oool0
0`00o`00oooo0?ooo`090?ooo`002P3oool00`00o`00oooo0?ooo`030?ooo`80o`00J@3oool00`3o
0000oooo0?ooo`1H0?ooo`030?l0003oool0oooo05X0oooo00<0o`000?ooo`3oool0KP3oool00`00
o`00oooo0?ooo`090?ooo`002P3oool01000o`00oooo0?ooo`3oool20?l000800?l0J@3oool00`3o
0000oooo0?ooo`1B0?ooo`80o`00103oool00`3o0000oooo0?ooo`1J0?ooo`030?l0003oool0oooo
06L0oooo0P00o`050?ooo`03003o003oool0oooo00T0oooo000:0?ooo`04003o003oool0o`000?l0
00@00?l0J@3oool00`3o0000oooo0?ooo`1B0?ooo`H0o`0000<0oooo0?l0003oool0FP3oool00`3o
0000oooo0?ooo`1W0?ooo`H00?l000<0oooo003o003oool02P3oool000X0oooocP3o001K003o0003
0?l00000o`000?l006l00?l02`3oool000X0oooo00<00?l00?ooo`3o00000P3o0003003o06T0oooo
00<0o`000?ooo`3oool0DP3oool60?l000030?ooo`3o0000oooo05X0oooo00<0o`000?ooo`3oool0
I`3oool6003o00030?ooo`00o`00oooo00X0oooo000:0?ooo`03003o003oool0oooo0080oooo0P3o
00000`00o`00oooo0?ooo`1W0?ooo`030?l0003oool0oooo0580oooo0P3o00060?ooo`030?l0003o
ool0oooo05P0oooo00<0o`000?ooo`3oool0I`3oool2003o00D0oooo00<00?l00?ooo`3oool02@3o
ool000X0oooo00<00?l00?ooo`3oool0103oool30?l006L0oooo00<0o`000?ooo`3oool0FP3oool0
0`3o0000oooo0?ooo`1H0?ooo`030?l0003oool0oooo06h0oooo00<00?l00?ooo`3oool02@3oool0
00X0oooo00<00?l00?ooo`3oool01`3oool30?l006@0oooo00<0o`000?ooo`3oool0FP3oool00`3o
0000oooo0?ooo`1H0?ooo`030?l0003oool0oooo06h0oooo00<00?l00?ooo`3oool02@3oool000X0
oooo00<00?l00?ooo`3oool02P3oool30?l00640oooo00<0o`000?ooo`3oool0F`3oool00`3o0000
oooo0?ooo`1G0?ooo`030?l0003oool0oooo06h0oooo00<00?l00?ooo`3oool02@3oool000X0oooo
00<00?l00?ooo`3oool03@3oool20?l005l0oooo00<0o`000?ooo`3oool0F`3oool00`3o0000oooo
0?ooo`1G0?ooo`030?l0003oool0oooo06h0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<0
0?l00?ooo`3oool03`3oool30?l005`0oooo00<0o`000?ooo`3oool0G03oool00`3o0000oooo0?oo
o`1F0?ooo`030?l0003oool0oooo06h0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l0
0?ooo`3oool04P3oool30?l005T0oooo00<0o`000?ooo`3oool0G03oool00`3o0000oooo0?ooo`1E
0?ooo`030?l0003oool0oooo06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?oo
o`3oool05@3oool30?l005L0oooo00<0o`000?ooo`3oool0F`3oool00`3o0000oooo0?ooo`1E0?oo
o`030?l0003oool0oooo06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3o
ool0603oool20?l005D0oooo00<0o`000?ooo`3oool0G03oool00`3o0000oooo0?ooo`1D0?ooo`03
0?l0003oool0oooo06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3oool0
6P3oool30?l00580oooo00<0o`000?ooo`3oool0G03oool00`3o0000oooo0?ooo`1D0?ooo`030?l0
003oool0oooo06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3oool07@3o
ool30?l004l0oooo00<0o`000?ooo`3oool0G03oool00`3o0000oooo0?ooo`1D0?ooo`030?l0003o
ool0oooo06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3oool0803oool3
0?l004`0oooo00<0o`000?ooo`3oool0G@3oool00`3o0000oooo0?ooo`1C0?ooo`030?l0003oool0
oooo06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3oool08`3oool20?l0
04X0oooo00<0o`000?ooo`3oool0G@3oool00`3o0000oooo0?ooo`1C0?ooo`030?l0003oool0oooo
06l0oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3oool09@3oool30?l004L0
oooo00<0o`000?ooo`3oool0GP3oool00`3o0000oooo0?ooo`1A0?ooo`030?l0003oool0oooo0700
oooo00<00?l00?ooo`3oool02@3oool000X0oooo00<00?l00?ooo`3oool0:03oool30?l004D0oooo
00<0o`000?ooo`3oool0G@3oool00`3o0000oooo0?ooo`1A0?ooo`030?l0003oool0oooo0700oooo
00<00?l00?ooo`3oool02@3oool000/0oooo00<00?l00?ooo`3oool0:P3oool30?l00480oooo00<0
o`000?ooo`3oool0G@3oool00`3o0000oooo0?ooo`1A0?ooo`030?l0003oool0oooo06l0oooo00<0
0?l00?ooo`3oool02P3oool000/0oooo00<00?l00?ooo`3oool0;@3oool20?l00400oooo00<0o`00
0?ooo`3oool0GP3oool00`3o0000oooo0?ooo`1@0?ooo`030?l0003oool0oooo06l0oooo00<00?l0
0?ooo`3oool02P3oool000/0oooo00<00?l00?ooo`3oool0;`3oool30?l003d0oooo00<0o`000?oo
o`3oool0GP3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0oooo0700oooo00<00?l00?oo
o`3oool02P3oool000/0oooo00<00?l00?ooo`3oool0<P3oool30?l003/0oooo00<0o`000?ooo`3o
ool0G@3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0oooo0700oooo00<00?l00?ooo`3o
ool02P3oool000/0oooo00<00?l00?ooo`3oool0=@3oool30?l003P0oooo00<0o`000?ooo`3oool0
GP3oool00`3o0000oooo0?ooo`1>0?ooo`030?l0003oool0oooo0700oooo00<00?l00?ooo`3oool0
2P3oool000/0oooo00<00?l00?ooo`3oool0>03oool30?l003D0oooo00<0o`000?ooo`3oool0GP3o
ool00`3o0000oooo0?ooo`1>0?ooo`030?l0003oool0oooo0700oooo00<00?l00?ooo`3oool02P3o
ool000/0oooo00<00?l00?ooo`3oool0>`3oool20?l003<0oooo00<0o`000?ooo`3oool0G`3oool0
0`3o0000oooo0?ooo`1=0?ooo`030?l0003oool0oooo0700oooo00<00?l00?ooo`3oool02P3oool0
00/0oooo00<00?l00?ooo`3oool0?@3oool30?l00300oooo00<0o`000?ooo`3oool0G`3oool00`3o
0000oooo0?ooo`1<0?ooo`030?l0003oool0oooo0740oooo00<00?l00?ooo`3oool02P3oool000/0
oooo00<00?l00?ooo`3oool0@03oool30?l002h0oooo00<0o`000?ooo`3oool0GP3oool00`3o0000
oooo0?ooo`1<0?ooo`030?l0003oool0oooo0740oooo00<00?l00?ooo`3oool02P3oool000/0oooo
00<00?l00?ooo`3oool0@`3oool30?l002/0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo
0?ooo`1;0?ooo`030?l0003oool0oooo0740oooo00<00?l00?ooo`3oool02P3oool000`0oooo00<0
0?l00?ooo`3oool0A@3oool20?l002T0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?oo
o`1;0?ooo`030?l0003oool0oooo0700oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l0
0?ooo`3oool0A`3oool30?l002H0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`1:
0?ooo`030?l0003oool0oooo0740oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l00?oo
o`3oool0BP3oool30?l002@0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`190?oo
o`030?l0003oool0oooo0740oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l00?ooo`3o
ool0C@3oool30?l00240oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`180?ooo`03
0?l0003oool0oooo0780oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l00?ooo`3oool0
D03oool20?l00200oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`170?ooo`030?l0
003oool0oooo0780oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l00?ooo`3oool0DP3o
ool30?l001d0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`170?ooo`030?l0003o
ool0oooo0780oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l00?ooo`3oool0E@3oool3
0?l001/0oooo00<0o`000?ooo`3oool0GP3oool00`3o0000oooo0?ooo`160?ooo`030?l0003oool0
oooo07<0oooo00<00?l00?ooo`3oool02`3oool000`0oooo00<00?l00?ooo`3oool0F03oool30?l0
01P0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`150?ooo`030?l0003oool0oooo
07<0oooo00<00?l00?ooo`3oool02`3oool000d0oooo00<00?l00?ooo`3oool0FP3oool20?l001H0
oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`150?ooo`030?l0003oool0oooo0780
oooo00<00?l00?ooo`3oool0303oool000d0oooo00<00?l00?ooo`3oool0G03oool30?l001<0oooo
00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`150?ooo`030?l0003oool0oooo0780oooo
00<00?l00?ooo`3oool0303oool000d0oooo00<00?l00?ooo`3oool0G`3oool30?l00100oooo00<0
o`000?ooo`3oool0H03oool00`3o0000oooo0?ooo`130?ooo`030?l0003oool0oooo07<0oooo00<0
0?l00?ooo`3oool0303oool000d0oooo00<00?l00?ooo`3oool0HP3oool30?l000h0oooo00<0o`00
0?ooo`3oool0G`3oool00`3o0000oooo0?ooo`130?ooo`030?l0003oool0oooo07<0oooo00<00?l0
0?ooo`3oool0303oool000d0oooo00<00?l00?ooo`3oool0I@3oool20?l000`0oooo00<0o`000?oo
o`3oool0H03oool00`3o0000oooo0?ooo`110?ooo`030?l0003oool0oooo07@0oooo00<00?l00?oo
o`3oool0303oool000h0oooo00<00?l00?ooo`3oool0IP3oool30?l000X0oooo00<0o`000?ooo`3o
ool0G`3oool00`3o0000oooo0?ooo`110?ooo`030?l0003oool0oooo07<0oooo00<00?l00?ooo`3o
ool03@3oool000h0oooo00<00?l00?ooo`3oool0J@3oool30?l000L0oooo00<0o`000?ooo`3oool0
G`3oool00`3o0000oooo0?ooo`100?ooo`030?l0003oool0oooo07@0oooo00<00?l00?ooo`3oool0
3@3oool000h0oooo00<00?l00?ooo`3oool0K03oool30?l000D0oooo00<0o`000?ooo`3oool0G`3o
ool00`3o0000oooo0?ooo`0o0?ooo`030?l0003oool0oooo07@0oooo00<00?l00?ooo`3oool03@3o
ool000h0oooo00<00?l00?ooo`3oool0K`3oool20?l000<0oooo00<0o`000?ooo`3oool0G`3oool0
0`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo07D0oooo00<00?l00?ooo`3oool03@3oool0
00h0oooo00<00?l00?ooo`3oool0L@3oool30?l000030?ooo`3o0000oooo05l0oooo00<0o`000?oo
o`3oool0?P3oool00`3o0000oooo0?ooo`1e0?ooo`03003o003oool0oooo00d0oooo000?0?ooo`03
003o003oool0oooo07<0oooo0`3o001P0?ooo`030?l0003oool0oooo03`0oooo00<0o`000?ooo`3o
ool0M@3oool00`00o`00oooo0?ooo`0>0?ooo`003`3oool00`00o`00oooo0?ooo`1f0?ooo`<0o`00
G@3oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo07H0oooo00<00?l00?ooo`3oool0
3P3oool000l0oooo00<00?l00?ooo`3oool0MP3oool00`3o0000oooo0?ooo`020?l005/0oooo00<0
o`000?ooo`3oool0>P3oool00`3o0000oooo0?ooo`1g0?ooo`03003o003oool0oooo00h0oooo000?
0?ooo`03003o003oool0oooo07L0oooo00@0o`000?ooo`3oool0oooo0`3o001I0?ooo`030?l0003o
ool0oooo03T0oooo00<0o`000?ooo`3oool0M`3oool00`00o`00oooo0?ooo`0>0?ooo`003`3oool0
0`00o`00oooo0?ooo`1g0?ooo`030?l0003oool0oooo00@0oooo0`3o001F0?ooo`030?l0003oool0
oooo03P0oooo00<0o`000?ooo`3oool0N03oool00`00o`00oooo0?ooo`0>0?ooo`00403oool00`00
o`00oooo0?ooo`1g0?ooo`030?l0003oool0oooo00H0oooo0`3o001D0?ooo`030?l0003oool0oooo
03L0oooo00<0o`000?ooo`3oool0M`3oool00`00o`00oooo0?ooo`0?0?ooo`00403oool00`00o`00
oooo0?ooo`1g0?ooo`030?l0003oool0oooo00T0oooo0P3o001B0?ooo`030?l0003oool0oooo03H0
oooo00<0o`000?ooo`3oool0N03oool00`00o`00oooo0?ooo`0?0?ooo`00403oool00`00o`00oooo
0?ooo`1h0?ooo`030?l0003oool0oooo00X0oooo0`3o001?0?ooo`030?l0003oool0oooo03H0oooo
00<0o`000?ooo`3oool0M`3oool00`00o`00oooo0?ooo`0@0?ooo`004@3oool00`00o`00oooo0?oo
o`1g0?ooo`030?l0003oool0oooo00d0oooo0`3o001=0?ooo`030?l0003oool0oooo03@0oooo00<0
o`000?ooo`3oool0N03oool00`00o`00oooo0?ooo`0@0?ooo`004@3oool00`00o`00oooo0?ooo`1h
0?ooo`030?l0003oool0oooo00l0oooo0`3o001:0?ooo`030?l0003oool0oooo03@0oooo00<0o`00
0?ooo`3oool0N03oool00`00o`00oooo0?ooo`0@0?ooo`004@3oool00`00o`00oooo0?ooo`1i0?oo
o`80o`004P3oool20?l004P0oooo00<0o`000?ooo`3oool0<P3oool20?l007/0oooo00<00?l00?oo
o`3oool0403oool00140oooo00<00?l00?ooo`3oool0NP3oool00`3o0000oooo0?ooo`0B0?ooo`<0
o`00AP3oool00`3o0000oooo0?ooo`0`0?ooo`030?l0003oool0oooo07/0oooo00<00?l00?ooo`3o
ool0403oool00140oooo00<00?l00?ooo`3oool0N`3oool00`3o0000oooo0?ooo`0D0?ooo`<0o`00
@`3oool00`3o0000oooo0?ooo`0`0?ooo`030?l0003oool0oooo07X0oooo00<00?l00?ooo`3oool0
4@3oool00180oooo00<00?l00?ooo`3oool0N`3oool00`3o0000oooo0?ooo`0F0?ooo`<0o`00@@3o
ool00`3o0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo07/0oooo00<00?l00?ooo`3oool04@3o
ool00180oooo00<00?l00?ooo`3oool0O03oool00`3o0000oooo0?ooo`0H0?ooo`80o`00?`3oool0
0`3o0000oooo0?ooo`0]0?ooo`030?l0003oool0oooo07/0oooo00<00?l00?ooo`3oool04P3oool0
01<0oooo00<00?l00?ooo`3oool0N`3oool00`3o0000oooo0?ooo`0J0?ooo`<0o`00?03oool00`3o
0000oooo0?ooo`0/0?ooo`030?l0003oool0oooo07`0oooo00<00?l00?ooo`3oool04P3oool001<0
oooo00<00?l00?ooo`3oool0O03oool00`3o0000oooo0?ooo`0L0?ooo`<0o`00>P3oool00`3o0000
oooo0?ooo`0[0?ooo`030?l0003oool0oooo07`0oooo00<00?l00?ooo`3oool04P3oool001@0oooo
00<00?l00?ooo`3oool0O03oool20?l001l0oooo0`3o000g0?ooo`030?l0003oool0oooo02T0oooo
0P3o001n0?ooo`03003o003oool0oooo01<0oooo000D0?ooo`03003o003oool0oooo07d0oooo00<0
o`000?ooo`3oool0803oool20?l003D0oooo00<0o`000?ooo`3oool0:03oool00`3o0000oooo0?oo
o`1n0?ooo`03003o003oool0oooo01<0oooo000D0?ooo`03003o003oool0oooo07h0oooo00<0o`00
0?ooo`3oool08@3oool30?l003<0oooo00<0o`000?ooo`3oool09`3oool00`3o0000oooo0?ooo`1n
0?ooo`03003o003oool0oooo01<0oooo000D0?ooo`03003o003oool0oooo07l0oooo00<0o`000?oo
o`3oool08`3oool30?l00300oooo00<0o`000?ooo`3oool09P3oool00`3o0000oooo0?ooo`1o0?oo
o`03003o003oool0oooo01<0oooo000D0?ooo`03003o003oool0oooo0800oooo00<0o`000?ooo`3o
ool09@3oool30?l002h0oooo00<0o`000?ooo`3oool0903oool00`3o0000oooo0?ooo`200?ooo`03
003o003oool0oooo01<0oooo000E0?ooo`03003o003oool0oooo0800oooo0P3o000X0?ooo`<0o`00
:`3oool00`3o0000oooo0?ooo`0R0?ooo`80o`00PP3oool00`00o`00oooo0?ooo`0D0?ooo`005@3o
ool00`00o`00oooo0?ooo`210?ooo`030?l0003oool0oooo02T0oooo0P3o000Y0?ooo`030?l0003o
ool0oooo0240oooo00<0o`000?ooo`3oool0PP3oool00`00o`00oooo0?ooo`0D0?ooo`005@3oool0
0`00o`00oooo0?ooo`220?ooo`030?l0003oool0oooo02X0oooo0`3o000W0?ooo`030?l0003oool0
oooo0200oooo00<0o`000?ooo`3oool0P@3oool00`00o`00oooo0?ooo`0E0?ooo`005P3oool00`00
o`00oooo0?ooo`220?ooo`80o`00;@3oool30?l002@0oooo00<0o`000?ooo`3oool07P3oool20?l0
08@0oooo00<00?l00?ooo`3oool05@3oool001H0oooo00<00?l00?ooo`3oool0Q03oool00`3o0000
oooo0?ooo`0]0?ooo`<0o`008@3oool00`3o0000oooo0?ooo`0M0?ooo`030?l0003oool0oooo08@0
oooo00<00?l00?ooo`3oool05@3oool001L0oooo00<00?l00?ooo`3oool0Q03oool00`3o0000oooo
0?ooo`0_0?ooo`80o`00803oool00`3o0000oooo0?ooo`0K0?ooo`030?l0003oool0oooo08@0oooo
00<00?l00?ooo`3oool05P3oool001L0oooo00<00?l00?ooo`3oool0Q@3oool20?l00340oooo0`3o
000M0?ooo`030?l0003oool0oooo01T0oooo0P3o00270?ooo`03003o003oool0oooo01H0oooo000G
0?ooo`03003o003oool0oooo08H0oooo00<0o`000?ooo`3oool0<P3oool30?l001/0oooo00<0o`00
0?ooo`3oool05`3oool00`3o0000oooo0?ooo`260?ooo`03003o003oool0oooo01L0oooo000H0?oo
o`03003o003oool0oooo08H0oooo00<0o`000?ooo`3oool0=03oool30?l001P0oooo00<0o`000?oo
o`3oool05`3oool00`3o0000oooo0?ooo`260?ooo`03003o003oool0oooo01L0oooo000H0?ooo`03
003o003oool0oooo08L0oooo0P3o000g0?ooo`80o`005P3oool00`3o0000oooo0?ooo`0E0?ooo`80
o`00R03oool00`00o`00oooo0?ooo`0H0?ooo`006@3oool00`00o`00oooo0?ooo`280?ooo`80o`00
=`3oool30?l001@0oooo00<0o`000?ooo`3oool04P3oool20?l008X0oooo00<00?l00?ooo`3oool0
603oool001T0oooo00<00?l00?ooo`3oool0RP3oool20?l003P0oooo0`3o000A0?ooo`030?l0003o
ool0o`000100oooo0P3o002<0?ooo`03003o003oool0oooo01P0oooo000J0?ooo`03003o003oool0
oooo08/0oooo0P3o000i0?ooo`<0o`003P3oool30?l000h0oooo0P3o002=0?ooo`03003o003oool0
oooo01T0oooo000J0?ooo`03003o003oool0oooo08d0oooo00<0o`000?ooo`3oool0>@3oool20?l0
00X0oooo1@3o000=0?ooo`030?l0003oool0oooo08d0oooo00<00?l00?ooo`3oool06@3oool001X0
oooo00<00?l00?ooo`3oool0SP3oool00`3o0000oooo0?ooo`0j0?ooo`<0o`00203oool40?l000`0
oooo00<0o`000?ooo`3oool0S@3oool00`00o`00oooo0?ooo`0J0?ooo`006`3oool00`00o`00oooo
0?ooo`2>0?ooo`80o`00?@3oool30?l00080oooo00@0o`000?ooo`3oool0oooo0`3o000:0?ooo`80
o`00T03oool00`00o`00oooo0?ooo`0J0?ooo`006`3oool00`00o`00oooo0?ooo`2@0?ooo`80o`00
?P3oool50?l000030?ooo`3o0000o`000080o`001`3oool20?l00940oooo00<00?l00?ooo`3oool0
6`3oool001`0oooo00<00?l00?ooo`3oool0T@3oool20?l003d0oooo2@3o00050?ooo`80o`00T`3o
ool00`00o`00oooo0?ooo`0K0?ooo`00703oool00`00o`00oooo0?ooo`2C0?ooo`D0o`00>03oool>
0?l009D0oooo00<00?l00?ooo`3oool06`3oool001d0oooo00<00?l00?ooo`3oool0U`3oool20?l0
03X0oooo1@3o002I0?ooo`03003o003oool0oooo01`0oooo000M0?ooo`03003o003oool0oooo09T0
oooo0P3o000i0?ooo`80o`00V`3oool00`00o`00oooo0?ooo`0L0?ooo`007@3oool00`00o`00oooo
0?ooo`2K0?ooo`@0o`00<@3oool40?l009`0oooo00<00?l00?ooo`3oool07@3oool001h0oooo00<0
0?l00?ooo`3oool0WP3oool50?l002L0oooo1@3o002P0?ooo`03003o003oool0oooo01d0oooo000N
0?ooo`03003o003oool0oooo0:<0oooo103o000O0?ooo`@0o`00Y03oool00`00o`00oooo0?ooo`0N
0?ooo`007`3oool00`00o`00oooo0?ooo`2V0?ooo`L0o`004@3oool70?l00:P0oooo00<00?l00?oo
o`3oool07P3oool001l0oooo00<00?l00?ooo`3oool0[@3ooolA0?l00:l0oooo00<00?l00?ooo`3o
ool07P3oool00200oooo00<00?l00?ooo`3oool0o`3ooom/0?ooo`03003o003oool0oooo01l0oooo
000P0?ooo`03003o003oool0oooo0?l0ooooK03oool00`00o`00oooo0?ooo`0O0?ooo`008@3oool0
0`00o`00oooo0?ooo`3o0?ooofX0oooo00<00?l00?ooo`3oool0803oool00280oooo00<00?l00?oo
o`3oool0o`3ooomX0?ooo`03003o003oool0oooo0240oooo000R0?ooo`03003o003oool0oooo0?l0
ooooI`3oool00`00o`00oooo0?ooo`0R0?ooo`008`3oool00`00o`00oooo0?ooo`3o0?ooofH0oooo
00<00?l00?ooo`3oool08P3oool002<0oooo00<00?l00?ooo`3oool0o`3ooomU0?ooo`03003o003o
ool0oooo02<0oooo000T0?ooo`03003o003oool0oooo0?l0ooooI03oool00`00o`00oooo0?ooo`0S
0?ooo`00903oool00`00o`00oooo0?ooo`3o0?ooof@0oooo00<00?l00?ooo`3oool08`3oool002D0
oooo00<00?l00?ooo`3oool0o`3ooomR0?ooo`03003o003oool0oooo02@0oooo000U0?ooo`03003o
003oool0oooo0?l0ooooHP3oool00`00o`00oooo0?ooo`0T0?ooo`009P3oool00`00o`00oooo0?oo
o`3o0?ooof00oooo00<00?l00?ooo`3oool09@3oool002L0oooo00<00?l00?ooo`3oool0o`3ooomN
0?ooo`03003o003oool0oooo02H0oooo000W0?ooo`03003o003oool0oooo0?l0ooooG@3oool00`00
o`00oooo0?ooo`0W0?ooo`00:03oool00`00o`00oooo0?ooo`3o0?oooe`0oooo00<00?l00?ooo`3o
ool09`3oool002P0oooo00<00?l00?ooo`3oool0o`3ooomK0?ooo`03003o003oool0oooo02P0oooo
000Y0?ooo`03003o003oool0oooo0?l0ooooFP3oool00`00o`00oooo0?ooo`0X0?ooo`00:P3oool0
0`00o`00oooo0?ooo`3o0?oooeP0oooo00<00?l00?ooo`3oool0:@3oool002/0oooo00<00?l00?oo
o`3oool0o`3ooomF0?ooo`03003o003oool0oooo02X0oooo000[0?ooo`03003o003oool0oooo0?l0
ooooE@3oool00`00o`00oooo0?ooo`0[0?ooo`00;03oool00`00o`00oooo0?ooo`3o0?oooe@0oooo
00<00?l00?ooo`3oool0:`3oool002`0oooo00<00?l00?ooo`3oool0o`3ooomD0?ooo`03003o003o
ool0oooo02/0oooo000]0?ooo`03003o003oool0oooo0?l0ooooDP3oool00`00o`00oooo0?ooo`0/
0?ooo`00;@3oool00`00o`00oooo0?ooo`3o0?oooe80oooo00<00?l00?ooo`3oool0;03oool002h0
oooo00<00?l00?ooo`3oool0o`3ooom@0?ooo`03003o003oool0oooo02d0oooo000_0?ooo`03003o
003oool0oooo0?l0ooooCP3oool00`00o`00oooo0?ooo`0^0?ooo`00;`3oool00`00o`00oooo0?oo
o`3o0?ooodd0oooo00<00?l00?ooo`3oool0;`3oool00300oooo00<00?l00?ooo`3oool0o`3ooom<
0?ooo`03003o003oool0oooo02l0oooo000a0?ooo`03003o003oool0oooo0?l0ooooBP3oool00`00
o`00oooo0?ooo`0`0?ooo`00<P3oool00`00o`00oooo0?ooo`3o0?ooodP0oooo00<00?l00?ooo`3o
ool0<@3oool00380oooo00<00?l00?ooo`3oool0o`3ooom70?ooo`03003o003oool0oooo0380oooo
000c0?ooo`03003o003oool0oooo0?l0ooooAP3oool00`00o`00oooo0?ooo`0b0?ooo`00=03oool0
0`00o`00oooo0?ooo`3o0?oood@0oooo00<00?l00?ooo`3oool0<`3oool003D0oooo00<00?l00?oo
o`3oool0o`3ooom20?ooo`03003o003oool0oooo03@0oooo000f0?ooo`03003o003oool0oooo0?l0
oooo@03oool00`00o`00oooo0?ooo`0e0?ooo`00=`3oool00`00o`00oooo0?ooo`3o0?oooch0oooo
00<00?l00?ooo`3oool0=P3oool003L0oooo00<00?l00?ooo`3oool0o`3ooolm0?ooo`03003o003o
ool0oooo03L0oooo000h0?ooo`03003o003oool0oooo0?l0oooo?03oool00`00o`00oooo0?ooo`0g
0?ooo`00>@3oool00`00o`00oooo0?ooo`3o0?ooocX0oooo00<00?l00?ooo`3oool0>03oool003X0
oooo00<00?l00?ooo`3oool0o`3ooolh0?ooo`03003o003oool0oooo03T0oooo000j0?ooo`03003o
003oool0oooo0?l0oooo=`3oool00`00o`00oooo0?ooo`0j0?ooo`00>`3oool00`00o`00oooo0?oo
o`3o0?ooocH0oooo00<00?l00?ooo`3oool0>P3oool003`0oooo00<00?l00?ooo`3oool0o`3ooold
0?ooo`03003o003oool0oooo03/0oooo000m0?ooo`03003o003oool0oooo0?l0oooo<P3oool00`00
o`00oooo0?ooo`0l0?ooo`00?P3oool00`00o`00oooo0?ooo`3o0?oooc00oooo00<00?l00?ooo`3o
ool0?@3oool003l0oooo00<00?l00?ooo`3oool0o`3oool^0?ooo`03003o003oool0oooo03h0oooo
00100?ooo`03003o003oool0oooo0?l0oooo;03oool00`00o`00oooo0?ooo`0o0?ooo`00@@3oool0
0`00o`00oooo0?ooo`3o0?ooobX0oooo00<00?l00?ooo`3oool0@03oool00480oooo00<00?l00?oo
o`3oool0o`3ooolX0?ooo`03003o003oool0oooo0440oooo00130?ooo`03003o003oool0oooo0?l0
oooo9P3oool00`00o`00oooo0?ooo`120?ooo`00A03oool00`00o`00oooo0?ooo`3o0?ooob@0oooo
00<00?l00?ooo`3oool0@`3oool004D0oooo00<00?l00?ooo`3oool0o`3ooolR0?ooo`03003o003o
ool0oooo04@0oooo00160?ooo`03003o003oool0oooo0?l0oooo803oool00`00o`00oooo0?ooo`15
0?ooo`00A`3oool00`00o`00oooo0?ooo`3o0?oooah0oooo00<00?l00?ooo`3oool0AP3oool004P0
oooo00<00?l00?ooo`3oool0o`3ooolL0?ooo`03003o003oool0oooo04L0oooo00190?ooo`03003o
003oool0oooo0?l0oooo6P3oool00`00o`00oooo0?ooo`180?ooo`00BP3oool00`00o`00oooo0?oo
o`3o0?oooaP0oooo00<00?l00?ooo`3oool0B@3oool004/0oooo00<00?l00?ooo`3oool0o`3ooolF
0?ooo`03003o003oool0oooo04X0oooo001<0?ooo`03003o003oool0oooo0?l0oooo503oool00`00
o`00oooo0?ooo`1;0?ooo`00C@3oool00`00o`00oooo0?ooo`3o0?oooa80oooo00<00?l00?ooo`3o
ool0C03oool004h0oooo00<00?l00?ooo`3oool0o`3oool@0?ooo`03003o003oool0oooo04d0oooo
001?0?ooo`03003o003oool0oooo0?l0oooo3P3oool00`00o`00oooo0?ooo`1>0?ooo`00D03oool0
0`00o`00oooo0?ooo`3o0?ooo``0oooo00<00?l00?ooo`3oool0C`3oool00540oooo00<00?l00?oo
o`3oool0o`3oool:0?ooo`03003o003oool0oooo0500oooo001B0?ooo`800?l0o`3oool80?ooo`80
0?l0D`3oool005@0oooo00<00?l00?ooo`3oool0o`3oool40?ooo`03003o003oool0oooo05<0oooo
001E0?ooo`03003o003oool0oooo0?l0oooo0P3oool00`00o`00oooo0?ooo`1D0?ooo`00EP3oool2
003o0?l0oooo0P00o`1G0?ooo`00F03oool00`00o`00oooo0?ooo`3k0?ooo`03003o003oool0oooo
05L0oooo001I0?ooo`03003o003oool0oooo0?T0oooo00<00?l00?ooo`3oool0F03oool005X0oooo
00<00?l00?ooo`3oool0m`3oool00`00o`00oooo0?ooo`1I0?ooo`00F`3oool00`00o`00oooo0?oo
o`3e0?ooo`03003o003oool0oooo05X0oooo001L0?ooo`800?l0l`3oool2003o05d0oooo001N0?oo
o`03003o003oool0oooo0>l0oooo00<00?l00?ooo`3oool0G@3oool005l0oooo00<00?l00?ooo`3o
ool0k@3oool00`00o`00oooo0?ooo`1N0?ooo`00H03oool2003o0>/0oooo0P00o`1Q0?ooo`00HP3o
ool00`00o`00oooo0?ooo`3W0?ooo`03003o003oool0oooo0640oooo001S0?ooo`03003o003oool0
oooo0>@0oooo0P00o`1T0?ooo`00I03oool3003o0>40oooo0P00o`1V0?ooo`00I`3oool2003o0=d0
oooo0P00o`1X0?ooo`00J@3oool00`00o`00oooo0?ooo`3I0?ooo`03003o003oool0oooo06P0oooo
001Z0?ooo`03003o003oool0oooo0=L0oooo00<00?l00?ooo`3oool0J@3oool006/0oooo0P00o`3E
0?ooo`800?l0K03oool006d0oooo0P00o`3A0?ooo`800?l0KP3oool006l0oooo00<00?l00?ooo`3o
ool0c@3oool00`00o`00oooo0?ooo`1^0?ooo`00L03oool2003o0</0oooo0P00o`1a0?ooo`00LP3o
ool2003o0<L0oooo0P00o`1c0?ooo`00M03oool2003o0<<0oooo0P00o`1e0?ooo`00MP3oool2003o
0;l0oooo0P00o`1g0?ooo`00N03oool00`00o`00oooo0?ooo`2k0?ooo`03003o003oool0oooo07L0
oooo001i0?ooo`800?l0^@3oool2003o07X0oooo001k0?ooo`800?l0]@3oool2003o07`0oooo001m
0?ooo`800?l0/@3oool2003o07h0oooo001o0?ooo`<00?l0Z`3oool3003o0800oooo00220?ooo`80
0?l0Y`3oool2003o08<0oooo00240?ooo`800?l0X`3oool2003o08D0oooo00260?ooo`<00?l0W@3o
ool3003o08L0oooo00290?ooo`800?l0V@3oool2003o08X0oooo002;0?ooo`800?l0U@3oool2003o
08`0oooo002=0?ooo`<00?l0S`3oool3003o08h0oooo002@0?ooo`800?l0R`3oool2003o0940oooo
002B0?ooo`<00?l0Q@3oool3003o09<0oooo002E0?ooo`@00?l0O03oool5003o09H0oooo002I0?oo
o`<00?l0M`3oool2003o09/0oooo002L0?ooo`800?l0L`3oool2003o09d0oooo002N0?ooo`D00?l0
J@3oool5003o09l0oooo002S0?ooo`D00?l0H@3oool3003o0:@0oooo002X0?ooo`<00?l0F@3oool5
003o0:L0oooo002[0?ooo`D00?l0C`3oool5003o0:`0oooo002`0?ooo`D00?l0A@3oool5003o0;40
oooo002e0?ooo`L00?l0=`3oool7003o0;H0oooo002l0?ooo`/00?l08@3oool;003o0;d0oooo0037
0?ooob400?l0b03oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.0250052, -0.525005, \
0.00365433, 0.00365433}}]
}, Closed]],

Cell["\<\
Because the outgoing velocities must lie on the inner circle, the velocity \
plot  demonstrates that there is are maximum scattering and recoil angles for \
this  endothermic reaction.  The maximum scattering angle is derived in a \
later section using conservation of energy and momentum, without assuming \
equal masses.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Exothermic reaction", "Subsubsection"],

Cell[BoxData[
    \(\(values = {m1 \[Rule] 1, m2 \[Rule] 1, m3 \[Rule] 1, 
          qvalue \[Rule] \(+0.2\), p1r \[Rule] 1, theta1 \[Rule] 0, 
          p2r \[Rule] 0, theta2 \[Rule] \[Pi]};\)\)], "Input"],

Cell[BoxData[
    \(\(sol = \(generalSolution[p3r, p4x, 
              p4y]\)\[LeftDoubleBracket]1\[RightDoubleBracket] //. 
          values;\)\)], "Input"],

Cell[BoxData[
    \(calcCollision[sol, values, 20\ Degree]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawCollision @@ LabMomenta;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .37303 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.488148 0.464338 0.186515 0.464338 [
[ 0 0 0 0 ]
[ 1 .37303 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .37303 L
0 .37303 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .18651 m
.48815 .18651 L
s
% Start of user PostScript
-1. 0. 0. 0. MAarrowhead1
% End of user PostScript
.48815 .18651 m
.48815 .18651 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.48815 .18651 m
.97619 .36415 L
s
% Start of user PostScript
0. 0. 1.0510491 0.38255057 MAarrowhead1
% End of user PostScript
.48815 .18651 m
.46444 .00888 L
s
% Start of user PostScript
0. 0. -0.051049055 -0.38255057 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 107.375},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000XB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00b03oool00`000000oooo0?oo
o`3U0?ooo`00b03oool00`000000oooo0?ooo`3U0?ooo`00b03oool200000>H0oooo00380?ooo`80
0000iP3oool00<L0oooo1000003U0?ooo`00a`3oool400000>D0oooo00370?ooo`D00000i03oool0
0<L0oooo1@00003T0?ooo`00a`3oool300000>H0oooo00390?ooo`030000003oool0oooo0>@0oooo
00390?ooo`030000003oool0oooo0>@0oooo00390?ooo`030000003oool0oooo0>@0oooo003:0?oo
o`030000003oool0oooo0><0oooo003:0?ooo`030000003oool0oooo0><0oooo003:0?ooo`030000
003oool0oooo0><0oooo003:0?ooo`030000003oool0oooo0><0oooo003:0?ooo`030000003oool0
oooo0><0oooo003:0?ooo`030000003oool0oooo0><0oooo003:0?ooo`030000003oool0oooo0><0
oooo003:0?ooo`030000003oool0oooo0><0oooo003;0?ooo`030000003oool0oooo0>80oooo003;
0?ooo`030000003oool0oooo0>80oooo003;0?ooo`030000003oool0oooo0>80oooo003;0?ooo`03
0000003oool0oooo0>80oooo003;0?ooo`030000003oool0oooo0>80oooo003;0?ooo`030000003o
ool0oooo0>80oooo003;0?ooo`030000003oool0oooo0>80oooo003<0?ooo`030000003oool0oooo
0>40oooo003<0?ooo`030000003oool0oooo0>40oooo003<0?ooo`030000003oool0oooo0>40oooo
003<0?ooo`030000003oool0oooo0>40oooo003<0?ooo`030000003oool0oooo0>40oooo003<0?oo
o`030000003oool0oooo0>40oooo003<0?ooo`030000003oool0oooo0>40oooo003<0?ooo`030000
003oool0oooo0>40oooo003=0?ooo`030000003oool0oooo0>00oooo003=0?ooo`030000003oool0
oooo0>00oooo003=0?ooo`030000003oool0oooo0>00oooo003=0?ooo`030000003oool0oooo0>00
oooo003=0?ooo`030000003oool0oooo0>00oooo003=0?ooo`030000003oool0oooo0>00oooo003=
0?ooo`030000003oool0oooo0>00oooo003>0?ooo`030000003oool0oooo0=l0oooo003>0?ooo`03
0000003oool0oooo0=l0oooo003>0?ooo`030000003oool0oooo0=l0oooo003>0?ooo`030000003o
ool0oooo0=l0oooo003>0?ooo`030000003oool0oooo0=l0oooo003>0?ooo`030000003oool0oooo
0=l0oooo003>0?ooo`030000003oool0oooo0=l0oooo003>0?ooo`030000003oool0oooo0=l0oooo
003?0?ooo`030000003oool0oooo0=h0oooo003?0?ooo`030000003oool0oooo0=h0oooo003?0?oo
o`030000003oool0oooo0=h0oooo003?0?ooo`030000003oool0oooo0=h0oooo003?0?ooo`030000
003oool0oooo0=h0oooo003?0?ooo`030000003oool0oooo0=h0oooo003?0?ooo`030000003oool0
oooo0=h0oooo003?0?ooo`030000003oool0oooo0=h0oooo003@0?ooo`030000003oool0oooo0=d0
oooo003@0?ooo`030000003oool0oooo0=d0oooo003@0?ooo`030000003oool0oooo0=d0oooo003@
0?ooo`030000003oool0oooo0=d0oooo003@0?ooo`030000003oool0oooo0=d0oooo003@0?ooo`03
0000003oool0oooo0=d0oooo003@0?ooo`030000003oool0oooo0=d0oooo003A0?ooo`030000003o
ool0oooo0=`0oooo003A0?ooo`030000003oool0oooo0=`0oooo003A0?ooo`030000003oool0oooo
0=`0oooo003A0?ooo`030000003oool0oooo0=`0oooo003A0?ooo`030000003oool0oooo0=`0oooo
003A0?ooo`030000003oool0oooo0=`0oooo003A0?ooo`030000003oool0oooo0=`0oooo003A0?oo
o`030000003oool0oooo0=`0oooo003B0?ooo`030000003oool0oooo0=/0oooo003B0?ooo`030000
003oool0oooo0=/0oooo003:0?ooo`@00000103oool00`000000oooo0?ooo`3K0?ooo`002`3oooo9
00000=`0oooo003:0?ooo`L000000`3oool300000=T0oooo003:0?ooo`<000002P3oool200000=L0
oooo003I0?ooo`<00000e03oool00=`0oooo0`00003A0?ooo`00g`3oool200000<l0oooo003Q0?oo
o`<00000c03oool00>@0oooo0`0000390?ooo`00i`3oool300000<H0oooo003Z0?ooo`800000a03o
ool00>`0oooo0`0000310?ooo`00k`3oool300000;h0oooo003b0?ooo`800000_03oool00?@0oooo
0`00002i0?ooo`00m`3oool300000;H0oooo003j0?ooo`<00000/`3oool00?d0oooo0P00002a0?oo
o`00o`3oool300000:h0oooo003o0?ooo`<0oooo0`00002[0?ooo`00o`3oool60?ooo`800000Z@3o
ool00?l0oooo203oool300000:H0oooo003o0?ooo`/0oooo0`00002S0?ooo`00o`3oool>0?ooo`<0
0000X03oool00?l0oooo4@3oool2000009h0oooo003o0?oooa<0oooo0`00002K0?ooo`00o`3ooolF
0?ooo`<00000V03oool00?l0oooo6@3oool2000009H0oooo003o0?oooa/0oooo0`00002C0?ooo`00
o`3ooolN0?ooo`<00000T03oool00?l0oooo8@3oool3000008d0oooo003o0?ooob@0oooo0P00002;
0?ooo`00o`3ooolV0?ooo`<00000R03oool00?l0oooo:@3oool3000008D0oooo003o0?ooob`0oooo
0P0000230?ooo`00o`3oool^0?ooo`<00000P03oool00?l0oooo<@3oool3000007d0oooo003o0?oo
oc@0oooo0`00001j0?ooo`00o`3ooolg0?ooo`800000N03oool00?l0oooo>@3oool3000007D0oooo
003o0?oooc`0oooo0`00001b0?ooo`00o`3ooolo0?ooo`800000L03oool00?l0oooo@@3oool30000
06d0oooo003o0?oood@0oooo0`00001Z0?ooo`00o`3ooom70?ooo`<00000I`3oool00?l0ooooBP3o
ool2000006D0oooo003o0?oood`0oooo0`00001R0?ooo`00o`3ooom?0?ooo`<00000G`3oool00?l0
ooooDP3oool2000005d0oooo003o0?oooe@0oooo0`00001J0?ooo`00o`3ooomG0?ooo`<00000E`3o
ool00?l0ooooFP3oool3000005@0oooo003o0?oooed0oooo0P00001B0?ooo`00o`3ooomO0?ooo`<0
0000C`3oool00?l0ooooHP3oool3000004`0oooo003o0?ooofD0oooo0P00001:0?ooo`00o`3ooomW
0?ooo`<00000A`3oool00?l0ooooJP3oool3000004@0oooo003o0?ooofd0oooo0`0000110?ooo`00
o`3ooom`0?ooo`800000?`3oool00?l0ooooLP3oool3000003`0oooo003o0?ooogD0oooo0`00000i
0?ooo`00o`3ooomh0?ooo`800000=`3oool00?l0ooooNP3oool3000003@0oooo003o0?ooogd0oooo
0`00000a0?ooo`00o`3ooon00?ooo`<00000;P3oool00?l0ooooP`3oool2000002`0oooo003o0?oo
ohD0oooo0`00000Y0?ooo`00o`3ooon80?ooo`<000009P3oool00?l0ooooR`3oool2000002@0oooo
003o0?ooohd0oooo0`00000Q0?ooo`00o`3ooon@0?ooo`<000007P3oool00?l0ooooT`3oool30000
01/0oooo003o0?oooiH0oooo0P0000060?ooo`030000003oool0oooo0100oooo003o0?oooiP0oooo
0`0000030?ooo`8000004@3oool00?l0ooooV`3oool600000100oooo003o0?oooid0oooo1P00000>
0?ooo`00o`3ooonM0?ooo`L000003@3oool00?l0ooooX@3oool4000000`0oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {160.063, 0}} -> {-1.05609, -0.401683, \
0.00752862, 0.00752862}}]
}, Closed]],

Cell[TextData[{
  "For an exothermic reaction, the lab angle between the two emerging \
particles (with equal masses) is now greater than ",
  Cell[BoxData[
      \(TraditionalForm\`90  \[Degree]\)]],
  ".  (Recall that this angle is ",
  Cell[BoxData[
      \(TraditionalForm\`90  \[Degree]\)]],
  " for an elastic collision between equal masses.)"
}], "Text"],

Cell[BoxData[
    \(\(PlotCMmomenta[cmMomenta];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[%, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.709863 0.5 0.709863 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.5 .5 m
.85493 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.14507 .5 L
s
% Start of user PostScript
0. 0. -0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.89117 .77156 L
s
% Start of user PostScript
0. 0. 0.55104905 0.38255057 MAarrowhead1
% End of user PostScript
.5 .5 m
.10883 .22844 L
s
% Start of user PostScript
0. 0. -0.55104905 -0.38255057 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .35493 0 365.73 arc
s
newpath
.5 .5 .47619 0 365.73 arc
s
3 Mabswid
.5 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ00000<P0oooo002m0?ooo`X000008@3oool:00000;h0oooo002e0?ooo`P00000
=@3oool800000;H0oooo002`0?ooo`D00000A@3oool500000;40oooo002[0?ooo`D00000C`3oool5
00000:`0oooo002V0?ooo`D00000F@3oool200000:X0oooo002T0?ooo`800000H03oool500000:D0
oooo002O0?ooo`D00000I`3oool500000:00oooo002M0?ooo`800000L@3oool3000009d0oooo002J
0?ooo`<00000MP3oool2000009/0oooo002E0?ooo`D00000N`3oool5000009H0oooo002C0?ooo`80
0000Q@3oool2000009@0oooo002@0?ooo`<00000R@3oool300000940oooo002>0?ooo`800000S`3o
ool2000008l0oooo002<0?ooo`800000T`3oool2000008d0oooo00290?ooo`<00000U`3oool30000
08X0oooo00270?ooo`800000W@3oool2000008P0oooo00250?ooo`800000X@3oool2000008H0oooo
00220?ooo`<00000Y@3oool3000008<0oooo00200?ooo`800000Z`3oool200000840oooo001n0?oo
o`800000[`3oool2000007l0oooo001k0?ooo`<00000/`3oool3000007`0oooo001i0?ooo`800000
^@3oool2000007X0oooo001h0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool0M`3o
ool007L0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`1f0?ooo`00M@3oool20000
0<40oooo0P00001f0?ooo`00LP3oool300000<D0oooo0`00001c0?ooo`00L03oool200000</0oooo
0P00001a0?ooo`00K`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo06h0oooo001^
0?ooo`030000003oool0oooo0<l0oooo00<000000?ooo`3oool0K@3oool006`0oooo0P00003C0?oo
o`800000K@3oool006X0oooo0P00003G0?ooo`800000J`3oool006T0oooo00<000000?ooo`3oool0
f@3oool00`000000oooo0?ooo`1X0?ooo`00J03oool00`000000oooo0?ooo`3K0?ooo`800000J03o
ool006D0oooo0`00003P0?ooo`800000IP3oool006<0oooo0P00003U0?ooo`800000I03oool00680
oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`1Q0?ooo`00H@3oool00`000000oooo
0?ooo`3Y0?ooo`030000003oool0oooo0600oooo001O0?ooo`800000k@3oool200000600oooo001N
0?ooo`030000003oool0oooo0>l0oooo00<000000?ooo`3oool0G@3oool005d0oooo00<000000?oo
o`3oool0l@3oool00`000000oooo0?ooo`1L0?ooo`00F`3oool200000?D0oooo0P00001L0?ooo`00
FP3oool00`000000oooo0?ooo`3g0?ooo`030000003oool0oooo05T0oooo001I0?ooo`030000003o
ool0oooo0?T0oooo00<000000?ooo`3oool0F03oool005P0oooo00<000000?ooo`3oool0n`3oool0
0`000000oooo0?ooo`1G0?ooo`00E`3oool00`000000oooo0?ooo`3m0?ooo`030000003oool0oooo
05H0oooo001E0?ooo`800000o`3oool20?ooo`800000EP3oool005@0oooo00<000000?ooo`3oool0
o`3oool40?ooo`030000003oool0oooo05<0oooo001C0?ooo`030000003oool0oooo0?l0oooo1P3o
ool00`000000oooo0?ooo`1B0?ooo`00D@3oool200000?l0oooo2P3oool200000580oooo001@0?oo
o`030000003oool0oooo0?l0oooo303oool00`000000oooo0?ooo`1?0?ooo`00C`3oool00`000000
oooo0?ooo`3o0?ooo`h0oooo00<000000?ooo`3oool0CP3oool004h0oooo00<000000?ooo`3oool0
N`3ooolG000007d0oooo00<000000?ooo`3oool0C@3oool004d0oooo00<000000?ooo`3oool0LP3o
ool:000001L0oooo2P00001d0?ooo`030000003oool0oooo04`0oooo001<0?ooo`030000003oool0
oooo06d0oooo1P00000[0?ooo`H00000K`3oool00`000000oooo0?ooo`1;0?ooo`00B`3oool00`00
0000oooo0?ooo`1Y0?ooo`D00000=`3oool5000006/0oooo00<000000?ooo`3oool0BP3oool004X0
oooo00<000000?ooo`3oool0IP3oool400000440oooo1000001X0?ooo`030000003oool0oooo04T0
oooo00190?ooo`030000003oool0oooo06<0oooo100000190?ooo`@00000I@3oool00`000000oooo
0?ooo`180?ooo`00B03oool00`000000oooo0?ooo`1Q0?ooo`<00000D@3oool3000006<0oooo00<0
00000?ooo`3oool0A`3oool004L0oooo00<000000?ooo`3oool0GP3oool4000005L0oooo1000001P
0?ooo`030000003oool0oooo04H0oooo00160?ooo`030000003oool0oooo05d0oooo0P00001O0?oo
o`800000G`3oool00`000000oooo0?ooo`150?ooo`00A@3oool00`000000oooo0?ooo`1K0?ooo`<0
0000H`3oool3000005d0oooo00<000000?ooo`3oool0A03oool004@0oooo00<000000?ooo`3oool0
FP3oool2000006T0oooo0P00001L0?ooo`030000003oool0oooo04<0oooo00130?ooo`030000003o
ool0oooo05L0oooo1000001]0?ooo`@00000F@3oool00`000000oooo0?ooo`120?ooo`00@P3oool0
0`000000oooo0?ooo`1G0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool0E`3oool0
0`000000oooo0?ooo`110?ooo`00@@3oool00`000000oooo0?ooo`1F0?ooo`800000M`3oool20000
05P0oooo00<000000?ooo`3oool0@03oool00400oooo00<000000?ooo`3oool0D`3oool4000007/0
oooo1000001E0?ooo`030000003oool0oooo03l0oooo000o0?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03h0oooo
000n0?ooo`030000003oool0oooo0580oooo0P0000250?ooo`800000E03oool00`000000oooo0?oo
o`0m0?ooo`00?@3oool00`000000oooo0?ooo`1A0?ooo`800000R@3oool2000005<0oooo00<00000
0?ooo`3oool0?03oool003`0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`2;0?oo
o`800000DP3oool00`000000oooo0?ooo`0k0?ooo`00>`3oool00`000000oooo0?ooo`1@0?ooo`80
0000T03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo03X0oooo000j0?ooo`030000
003oool0oooo04l0oooo0P00002C0?ooo`800000D03oool00`000000oooo0?ooo`0j0?ooo`00>P3o
ool00`000000oooo0?ooo`1<0?ooo`<00000U`3oool2000004l0oooo00<000000?ooo`3oool0>@3o
ool003T0oooo00<000000?ooo`3oool0B`3oool2000009`0oooo0`00001=0?ooo`030000003oool0
oooo03P0oooo000h0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0W`3oool00`00
0000oooo0?ooo`1;0?ooo`030000003oool0oooo03L0oooo000g0?ooo`030000003oool0oooo04X0
oooo0P00002S0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0=`3oool003L0oooo
00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo04X0oooo
00<000000?ooo`3oool0=P3oool003H0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?oo
o`2V0?ooo`800000B`3oool00`000000oooo0?ooo`0e0?ooo`00=@3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0d
0?ooo`00=03oool00`000000oooo0?ooo`180?ooo`800000[@3oool2000004X0oooo00<000000?oo
o`3oool0<`3oool003<0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`2_0?ooo`03
0000003oool0oooo04P0oooo00<000000?ooo`3oool0<P3oool00380oooo00<000000?ooo`3oool0
A`3oool200000;<0oooo0P0000180?ooo`030000003oool0oooo0380oooo000b0?ooo`030000003o
ool0oooo04H0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`160?ooo`030000003o
ool0oooo0340oooo000a0?ooo`030000003oool0oooo04D0oooo0P00002i0?ooo`800000A`3oool0
0`000000oooo0?ooo`0`0?ooo`00<03oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo
0;/0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0_0?ooo`00;`3oool00`000000
oooo0?ooo`150?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0A03oool00`000000
oooo0?ooo`0_0?ooo`00;`3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo0;l0oooo
00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0^0?ooo`00;P3oool2000004D0oooo00<0
00000?ooo`3oool0`@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo02d0oooo000]
0?ooo`030000003oool0000000800000@@3oool200000<D0oooo0P0000150?ooo`030000003oool0
oooo02`0oooo000/0?ooo`040000003oool0oooo0?ooo`D00000?@3oool00`000000oooo0?ooo`37
0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0;03oool002`0oooo00@000000?oo
o`3oool0oooo1`00000j0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool0@P3oool0
0`000000oooo0?ooo`0[0?ooo`00;03oool00`000000oooo0?ooo`020?ooo`D00000>P3oool00`00
0000oooo0?ooo`3;0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0:`3oool002/0
oooo00<000000?ooo`3oool0103oool4000003T0oooo00<000000?ooo`3oool0c@3oool00`000000
oooo0?ooo`100?ooo`030000003oool0oooo02/0oooo000[0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool00P00000e0?ooo`800000d@3oool00`000000oooo0?ooo`100?ooo`030000
003oool0oooo02X0oooo000Z0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0<@3o
ool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0:@3o
ool002T0oooo00<000000?ooo`3oool03@3oool200000300oooo00<000000?ooo`3oool0e03oool0
0`000000oooo0?ooo`100?ooo`030000003oool0oooo02P0oooo000X0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`3E0?ooo`800000@03oool00`00
0000oooo0?ooo`0X0?ooo`00:03oool00`000000oooo0?ooo`0A0?ooo`800000;03oool00`000000
oooo0?ooo`3H0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool09`3oool002L0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0=X0oooo
00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0W0?ooo`009`3oool00`000000oooo0?oo
o`0E0?ooo`8000009`3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo03h0oooo00<0
00000?ooo`3oool09P3oool002H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0S
0?ooo`030000003oool0oooo0=d0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`0U
0?ooo`009@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0240oooo00<000000?oo
o`3oool0g`3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo02@0oooo000U0?ooo`03
0000003oool0oooo01/0oooo0P00000P0?ooo`030000003oool0oooo0>40oooo00<000000?ooo`3o
ool0?@3oool00`000000oooo0?ooo`0T0?ooo`00903oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`0m0?ooo`030000
003oool0oooo02<0oooo000T0?ooo`030000003oool0oooo01l0oooo0P00000K0?ooo`030000003o
ool0oooo0>D0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0S0?ooo`008`3oool0
0`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0i@3oool0
0`000000oooo0?ooo`0l0?ooo`030000003oool0oooo02<0oooo000S0?ooo`030000003oool0oooo
02<0oooo0P00000G0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0?03oool00`00
0000oooo0?ooo`0R0?ooo`008P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0j@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0280
oooo000R0?ooo`030000003oool0oooo02L0oooo0P00000B0?ooo`030000003oool0oooo0>/0oooo
00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0Q0?ooo`008@3oool00`000000oooo0?oo
o`0Z0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0k@3oool00`000000oooo0?oo
o`0k0?ooo`030000003oool0oooo0200oooo000P0?ooo`030000003oool0oooo02`0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`3_0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool07`3oool00200oooo00<000000?ooo`3oool0;@3oool2000000`0oooo00<000000?oo
o`3oool0k`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo01l0oooo000O0?ooo`03
0000003oool0oooo0300oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`3a0?ooo`03
0000003oool0oooo03/0oooo00<000000?ooo`3oool07P3oool001l0oooo00<000000?ooo`3oool0
<@3oool2000000L0oooo00<000000?ooo`3oool0l`3oool00`000000oooo0?ooo`0j0?ooo`030000
003oool0oooo01h0oooo000N0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`3d0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool07P3o
ool001h0oooo00<000000?ooo`3oool0=@3oool2000000<0oooo00<000000?ooo`3oool0m@3oool0
0`000000oooo0?ooo`0j0?ooo`030000003oool0oooo01d0oooo000M0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`000000n@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo
01d0oooo000M0?ooo`030000003oool0oooo03T0oooo0P00003i0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool0703oool001d0oooo00<000000?ooo`3oool0>@3oool00`000000oooo
0000003i0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0703oool001`0oooo00<0
00000?ooo`3oool0>@3oool01@000000oooo0?ooo`3oool000000?T0oooo00<000000?ooo`3oool0
>@3oool00`000000oooo0?ooo`0K0?ooo`00703oool00`000000oooo0?ooo`0h0?ooo`030000003o
ool0oooo00<0oooo0P00003h0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool06`3o
ool001/0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo0?H0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0K0?ooo`006`3oool00`00
0000oooo0?ooo`0h0?ooo`030000003oool0oooo00L0oooo0P00003f0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool06P3oool001X0oooo00<000000?ooo`3oool0>03oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo0?@0oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`0J0?ooo`006P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00`0oooo
0P00003e0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool06@3oool001X0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?80oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0I0?ooo`006@3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo0100oooo0P00003c0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool0603oool001T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0?00oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0H0?oo
o`00603oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool0k`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo01P0oooo000H0?ooo`030000
003oool0oooo03L0oooo00<000000?ooo`3oool05@3oool200000?00oooo00<000000?ooo`3oool0
=`3oool00`000000oooo0?ooo`0G0?ooo`005`3oool00`000000oooo0?ooo`0g0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0k@3oool00`000000oooo0?ooo`0g0?ooo`030000003o
ool0oooo01L0oooo000G0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool06@3oool2
00000>h0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0F0?ooo`005`3oool00`00
0000oooo0?ooo`0f0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0j`3oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo01H0oooo000F0?ooo`030000003oool0oooo03H0
oooo00<000000?ooo`3oool07P3oool200000>`0oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool0jP3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo01D0oooo
000E0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool08P3oool200000>X0oooo00<0
00000?ooo`3oool0=P3oool00`000000oooo0?ooo`0E0?ooo`005@3oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0j03oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo01@0oooo000E0?ooo`030000003oool0oooo03D0oooo00<000000?oo
o`3oool09P3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0oooo03H0oooo00<000000?oo
o`3oool0503oool001@0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0W0?ooo`80
0000j03oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo01<0oooo000D0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`3V0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool04`3oool001@0oooo00<000000?ooo`3oool0=@3o
ool00`000000oooo0?ooo`0[0?ooo`800000iP3oool00`000000oooo0?ooo`0e0?ooo`030000003o
ool0oooo01<0oooo000D0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0;@3oool0
0`000000oooo0?ooo`3S0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool04`3oool0
01<0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0_0?ooo`800000h`3oool00`00
0000oooo0?ooo`0e0?ooo`030000003oool0oooo01<0oooo000C0?ooo`030000003oool0oooo03D0
oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo03D0
oooo00<000000?ooo`3oool04P3oool001<0oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`0c0?ooo`800000h@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0180oooo
000B0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?oo
o`3O0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool04P3oool00180oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo0=l0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`0A0?ooo`004@3oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo03P0oooo0P00003O0?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool04@3oool00140oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo0=`0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0@0?ooo`00
4@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo03`0oooo0P00003L0?ooo`030000
003oool0oooo03D0oooo00<000000?ooo`3oool0403oool00140oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo0=X0oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo0400oooo0P00003J0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0403oool0
0100oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo
0=P0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0@0?ooo`00403oool00`000000
oooo0?ooo`0d0?ooo`030000003oool0oooo04<0oooo0P00003H0?ooo`030000003oool0oooo03@0
oooo00<000000?ooo`3oool03`3oool00100oooo00<000000?ooo`3oool0<`3oool00`000000oooo
0?ooo`160?ooo`030000003oool0oooo0=D0oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`0?0?ooo`00403oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo04L0oooo00<0
00000?ooo`3oool0e@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo00h0oooo000@
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0B@3oool200000=D0oooo00<00000
0?ooo`3oool0=03oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo0?ooo`0c0?oo
o`030000003oool0oooo04/0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`0c0?oo
o`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3o
ool0C03oool200000=<0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0>0?ooo`00
3`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0
d03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00h0oooo000>0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0D03oool200000=00oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo0580oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo00d0oooo000>0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0D`3oool20000
0<h0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000
oooo0?ooo`0c0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0b`3oool00`000000
oooo0?ooo`0c0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`3;0?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool03@3oool000d0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?oo
o`1H0?ooo`800000b`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00`0oooo000=
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`38
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`1L0?ooo`800000b03oool00`000000oooo0?ooo`0c0?oo
o`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool0GP3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool0303oool000d0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1O0?ooo`800000
aP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00`0oooo000<0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`330?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`1S0?ooo`800000``3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo00/0oooo000<0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0I@3oool00`00
0000oooo0?ooo`310?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool02`3oool000`0
oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo0<00
oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo06L0oooo0P0000300?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`1Y0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`0;0?ooo`00303oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo06X0oooo0P00002m
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool02`3oool000/0oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo0;/0oooo00<000000?oo
o`3oool0<@3oool00`000000oooo0?ooo`0;0?ooo`002`3oool00`000000oooo0?ooo`0b0?ooo`03
0000003oool0oooo06h0oooo0P00002k0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool02P3oool000/0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1`0?ooo`030000
003oool0oooo0;P0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0:0?ooo`002`3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0740oooo0P00002h0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0<P3oool0
0`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo
07D0oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
00X0oooo000;0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0MP3oool200000;D0
oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo
0?ooo`0a0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0/P3oool00`000000oooo
0?ooo`0a0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0N@3oool200000;80oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?oo
o`0:0?ooo`002`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo07/0oooo00<00000
0?ooo`3oool0[`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo00X0oooo000;0?oo
o`030000003oool0oooo0340oooo00<000000?ooo`3oool0O03oool200000:l0oooo00<000000?oo
o`3oool0<@3oool00`000000oooo0?ooo`0:0?ooo`002P3oool00`000000oooo0?ooo`0b0?ooo`03
0000003oool0oooo07h0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0b0?ooo`03
0000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0
O`3oool200000:`0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`090?ooo`002P3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0Z@3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0
oooo0380oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo0380oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0<@3oool00`00
0000oooo0?ooo`240?ooo`800000Z03oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
00T0oooo000:0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0QP3oool00`000000
oooo0?ooo`2V0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool02@3oool000X0oooo
00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`270?ooo`800000YP3oool00`000000oooo
0?ooo`0a0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0R@3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2:
0?ooo`800000X`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo00T0oooo000:0?oo
o`030000003oool0oooo0340oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`2P0?oo
o`030000003oool0oooo0340oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3o
ool0<@3oool00`000000oooo0?ooo`2=0?ooo`800000X03oool00`000000oooo0?ooo`0a0?ooo`03
0000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0
S`3oool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0
2@3oool000X0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2@0?ooo`030000003o
ool0oooo09`0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`090?ooo`002P3oool0
0`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0940oooo0P00002L0?ooo`030000003oool0
oooo0340oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0<@3oool00`00
0000oooo0?ooo`040?ooo`800000S@3oool00`000000oooo0?ooo`2A0?ooo`8000001P3oool00`00
0000oooo0?ooo`0a0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0340
oooo00<000000?ooo`3oool01P00002>0?ooo`<00000T03oool600000080oooo00<000000?ooo`3o
ool0<@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`0a0?ooool00000
=000000c0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`0000001P00002>0?ooo`D00000SP3oool7000000030?ooo`000000oooo0380oooo00<0
00000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`03
0?ooo`<00000S`3oool200000080oooo00<000000?ooo`3oool0R`3oool3000000D0oooo00<00000
0?ooo`3oool0<@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`0a0?oo
o`030000003oool0oooo09X0oooo0P00002C0?ooo`030000003oool0oooo0340oooo00<000000?oo
o`3oool02@3oool000X0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2L0?ooo`03
0000003oool0oooo0900oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`090?ooo`00
2P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0
S`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003o
ool0oooo0340oooo00<000000?ooo`3oool0WP3oool2000008l0oooo00<000000?ooo`3oool0<@3o
ool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo0:00oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo00T0oooo000:0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0X@3oool20000
08`0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool0R@3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0Y03oool2000008T0oooo00<000000?ooo`3oool0<@3oool00`000000oooo
0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0:H0oooo00<0
00000?ooo`3oool0QP3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo00T0oooo000:
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0Y`3oool2000008D0oooo00<00000
0?ooo`3oool0<P3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`0b0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`0b0?oo
o`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool0Z@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool02@3oool000X0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`2Z0?ooo`800000
P@3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`1n0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`2]0?ooo`800000OP3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo00X0oooo000;0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0[`3oool00`00
0000oooo0?ooo`1k0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool02P3oool000/0
oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2`0?ooo`800000N`3oool00`000000
oooo0?ooo`0a0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?oo
o`2c0?ooo`800000N03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo00X0oooo000;
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`1d
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo07<0oooo00<000000?oo
o`3oool0<P3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`0b0?ooo`03
0000003oool0oooo0;H0oooo0P00001c0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool02P3oool000/0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`2h0?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0:0?ooo`002`3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0;T0oooo0P00001`0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool02P3oool000`0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`2k0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
0;/0oooo0P00001/0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool02`3oool000`0
oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo06T0
oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo0;h0oooo0P00001Y0?ooo`030000003oool0oooo0380oooo
00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`300?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`0;0?ooo`00303oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0<40oooo00<00000
0?ooo`3oool0I@3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00/0oooo000<0?oo
o`030000003oool0oooo0380oooo00<000000?ooo`3oool0`P3oool2000006@0oooo00<000000?oo
o`3oool0<`3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0c0?ooo`03
0000003oool0oooo0<<0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0c0?ooo`03
0000003oool0oooo00/0oooo000=0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0
a03oool200000640oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0<0?ooo`003@3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0GP3o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0
oooo0380oooo00<000000?ooo`3oool0a`3oool2000005d0oooo00<000000?ooo`3oool0<`3oool0
0`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
0<P0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
00`0oooo000=0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0b@3oool2000005X0
oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0<0?ooo`003P3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo0</0oooo00<000000?ooo`3oool0E`3oool00`000000oooo
0?ooo`0b0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool0b`3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo03<0oooo00<0
00000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`3<
0?ooo`800000E@3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00d0oooo000>0?oo
o`030000003oool0oooo03<0oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?ooo`1B0?oo
o`030000003oool0oooo03<0oooo00<000000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3o
ool0<`3oool00`000000oooo0?ooo`3?0?ooo`800000D@3oool00`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo00d0oooo000?0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
d03oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
3P3oool000l0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`3A0?ooo`800000CP3o
ool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0
oooo03<0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0
oooo03<0oooo00<000000?ooo`3oool03P3oool000l0oooo00<000000?ooo`3oool0<`3oool00`00
0000oooo0?ooo`3D0?ooo`800000BP3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo
00h0oooo000?0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0e@3oool00`000000
oooo0?ooo`170?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool03P3oool00100oooo
00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo04D0oooo
00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000oooo0?oo
o`0d0?ooo`030000003oool0oooo0=H0oooo0P0000150?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool03`3oool00100oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`3H
0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0@
0?ooo`004@3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0=T0oooo0P0000110?oo
o`030000003oool0oooo03@0oooo00<000000?ooo`3oool0403oool00140oooo00<000000?ooo`3o
ool0=03oool00`000000oooo0?ooo`3J0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3o
ool0=03oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`0d0?ooo`030000
003oool0oooo0=/0oooo0P00000m0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0
403oool00140oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`3L0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0@0?ooo`004@3oool0
0`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0=d0oooo0P00000j0?ooo`030000003oool0
oooo03@0oooo00<000000?ooo`3oool04@3oool00180oooo00<000000?ooo`3oool0=03oool00`00
0000oooo0?ooo`3O0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0=03oool00`00
0000oooo0?ooo`0A0?ooo`004P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0=l0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0180
oooo000C0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0h03oool2000003@0oooo
00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0B0?ooo`004`3oool00`000000oooo0?oo
o`0e0?ooo`030000003oool0oooo0>40oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?oo
o`0e0?ooo`030000003oool0oooo0180oooo000D0?ooo`030000003oool0oooo03@0oooo00<00000
0?ooo`3oool0hP3oool200000300oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0C
0?ooo`00503oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0><0oooo00<000000?oo
o`3oool0;@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo01<0oooo000D0?ooo`03
0000003oool0oooo03D0oooo00<000000?ooo`3oool0i03oool2000002d0oooo00<000000?ooo`3o
ool0=@3oool00`000000oooo0?ooo`0C0?ooo`00503oool00`000000oooo0?ooo`0e0?ooo`030000
003oool0oooo0>H0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0e0?ooo`030000
003oool0oooo01<0oooo000D0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0iP3o
ool2000002T0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0C0?ooo`005@3oool0
0`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0>P0oooo00<000000?ooo`3oool09@3oool0
0`000000oooo0?ooo`0f0?ooo`030000003oool0oooo01@0oooo000E0?ooo`030000003oool0oooo
03H0oooo00<000000?ooo`3oool0j03oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo
03H0oooo00<000000?ooo`3oool0503oool001D0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`3Y0?ooo`8000008`3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo01D0
oooo000F0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0jP3oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool05@3oool001H0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`3Z0?ooo`8000007`3oool00`000000oooo0?oo
o`0g0?ooo`030000003oool0oooo01D0oooo000G0?ooo`030000003oool0oooo03H0oooo00<00000
0?ooo`3oool0k03oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool05P3oool001L0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`3/0?oo
o`8000006`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo01H0oooo000G0?ooo`03
0000003oool0oooo03L0oooo00<000000?ooo`3oool0kP3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo03L0oooo00<000000?ooo`3oool05`3oool001P0oooo00<000000?ooo`3oool0
=`3oool00`000000oooo0?ooo`3^0?ooo`8000005`3oool00`000000oooo0?ooo`0g0?ooo`030000
003oool0oooo01L0oooo000H0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0k`3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0603o
ool001T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`3`0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0H0?ooo`006@3oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo0?40oooo0P00000B0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool0603oool001X0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`3c0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`0I0?ooo`006P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo0?<0oooo
0P00000>0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool06@3oool001X0oooo00<0
00000?ooo`3oool0>03oool00`000000oooo0?ooo`3d0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0J0?ooo`006`3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo0?D0oooo0P0000090?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool06P3oool001/0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`3f0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0K0?oo
o`00703oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo0?H0oooo0P0000050?ooo`03
0000003oool0oooo03P0oooo00<000000?ooo`3oool06`3oool001`0oooo00<000000?ooo`3oool0
>@3oool00`000000oooo0?ooo`3g0?ooo`050000003oool0oooo0?ooo`000000>`3oool00`000000
oooo0?ooo`0K0?ooo`007@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0?L0oooo
00<000000?ooo`000000>`3oool00`000000oooo0?ooo`0L0?ooo`007@3oool00`000000oooo0?oo
o`0i0?ooo`030000003oool0oooo0?P0oooo0P00000k0?ooo`030000003oool0oooo01`0oooo000M
0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0m`3oool00`000000oooo0000000i
0?ooo`030000003oool0oooo01d0oooo000N0?ooo`030000003oool0oooo03X0oooo00<000000?oo
o`3oool0m@3oool010000000oooo0?ooo`3oool2000003L0oooo00<000000?ooo`3oool07@3oool0
01h0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`3d0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0N0?ooo`007`3oool00`000000
oooo0?ooo`0j0?ooo`030000003oool0oooo0?<0oooo00<000000?ooo`3oool01@3oool2000003<0
oooo00<000000?ooo`3oool07P3oool001l0oooo00<000000?ooo`3oool0>`3oool00`000000oooo
0?ooo`3a0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo
0?ooo`0N0?ooo`00803oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo0>l0oooo00<0
00000?ooo`3oool02P3oool2000002l0oooo00<000000?ooo`3oool07`3oool00200oooo00<00000
0?ooo`3oool0>`3oool00`000000oooo0?ooo`3_0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0;03oool00`000000oooo0?ooo`0O0?ooo`008@3oool00`000000oooo0?ooo`0k0?oo
o`030000003oool0oooo0>d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo0200oooo000R0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3o
ool0j`3oool00`000000oooo0?ooo`0@0?ooo`800000:@3oool00`000000oooo0?ooo`0Q0?ooo`00
8P3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo0280oooo000S0?ooo`030000003o
ool0oooo03`0oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`0E0?ooo`8000009@3o
ool00`000000oooo0?ooo`0R0?ooo`008`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo0>D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo02<0oooo000T0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0i@3oool00`00
0000oooo0?ooo`0I0?ooo`8000008@3oool00`000000oooo0?ooo`0S0?ooo`00903oool00`000000
oooo0?ooo`0m0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo000U0?ooo`030000003oool0oooo03d0oooo
00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`0N0?ooo`8000007@3oool00`000000oooo
0?ooo`0T0?ooo`009@3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo0=l0oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02@0oooo000V
0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0g@3oool00`000000oooo0?ooo`0S
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool09@3oool002L0oooo00<000000?oo
o`3oool0?P3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo02D0oooo0P00000G0?oo
o`030000003oool0oooo02H0oooo000W0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3o
ool0f`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool09`3oool002P0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`3I0?ooo`030000
003oool0oooo02T0oooo0P00000C0?ooo`030000003oool0oooo02L0oooo000X0?ooo`030000003o
ool0oooo03l0oooo0P00003H0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0X0?ooo`00:@3oool00`000000oooo0?ooo`100?ooo`030000003oool0
oooo0=@0oooo00<000000?ooo`3oool0;P3oool2000000l0oooo00<000000?ooo`3oool0:03oool0
02X0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0Y0?ooo`00:`3oool00`000000
oooo0?ooo`100?ooo`030000003oool0oooo0<l0oooo0P00000e0?ooo`80000000<0oooo0000003o
ool01`3oool00`000000oooo0?ooo`0Z0?ooo`00:`3oool00`000000oooo0?ooo`110?ooo`030000
003oool0oooo0<d0oooo00<000000?ooo`3oool0=`3oool3000000H0oooo00<000000?ooo`3oool0
:`3oool002`0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`3;0?ooo`030000003o
ool0oooo03P0oooo100000050?ooo`030000003oool0oooo02/0oooo000/0?ooo`030000003oool0
oooo0480oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`0h0?ooo`H00000103oool0
0`000000oooo0?ooo`0[0?ooo`00;@3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo
0<L0oooo00<000000?ooo`3oool0>`3oool500000080oooo00<000000?ooo`3oool0;03oool002d0
oooo00<000000?ooo`3oool0@`3oool200000<D0oooo0P0000100?ooo`@0000000<0oooo0000003o
ool0;@3oool002h0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`310?ooo`030000
003oool0oooo04<0oooo0P00000_0?ooo`00;`3oool00`000000oooo0?ooo`140?ooo`030000003o
ool0oooo0;l0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0^0?ooo`00;`3oool0
0`000000oooo0?ooo`150?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0A03oool0
0`000000oooo0?ooo`0_0?ooo`00<03oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo
0;/0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`0_0?ooo`00<@3oool00`000000
oooo0?ooo`150?ooo`800000^P3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo0300
oooo000b0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0]@3oool2000004L0oooo
00<000000?ooo`3oool0<@3oool00380oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?oo
o`2c0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0<P3oool003<0oooo00<00000
0?ooo`3oool0A`3oool200000;40oooo0P0000190?ooo`030000003oool0oooo0380oooo000d0?oo
o`030000003oool0oooo04P0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`180?oo
o`030000003oool0oooo03<0oooo000e0?ooo`030000003oool0oooo04P0oooo0P00002[0?ooo`80
0000BP3oool00`000000oooo0?ooo`0d0?ooo`00=P3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo0:L0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0e0?ooo`00=`3o
ool00`000000oooo0?ooo`190?ooo`800000YP3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo03H0oooo000g0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0XP3oool0
0`000000oooo0?ooo`190?ooo`030000003oool0oooo03L0oooo000h0?ooo`030000003oool0oooo
04/0oooo00<000000?ooo`3oool0W`3oool2000004`0oooo00<000000?ooo`3oool0=`3oool003T0
oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo04/0
oooo00<000000?ooo`3oool0>03oool003X0oooo00<000000?ooo`3oool0B`3oool4000009L0oooo
1000001=0?ooo`030000003oool0oooo03T0oooo000j0?ooo`030000003oool0oooo04l0oooo00<0
00000?ooo`3oool0T`3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo03X0oooo000k
0?ooo`030000003oool0oooo04l0oooo0P00002A0?ooo`800000D@3oool00`000000oooo0?ooo`0j
0?ooo`00?03oool00`000000oooo0?ooo`1@0?ooo`800000S@3oool200000580oooo00<000000?oo
o`3oool0>`3oool003d0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`290?ooo`03
0000003oool0oooo0540oooo00<000000?ooo`3oool0?03oool003h0oooo00<000000?ooo`3oool0
D@3oool2000008L0oooo0P00001C0?ooo`030000003oool0oooo03d0oooo000o0?ooo`030000003o
ool0oooo0580oooo0P0000230?ooo`800000E03oool00`000000oooo0?ooo`0n0?ooo`00@03oool0
0`000000oooo0?ooo`1C0?ooo`<00000O@3oool3000005D0oooo00<000000?ooo`3oool0?`3oool0
0440oooo00<000000?ooo`3oool0E@3oool2000007T0oooo0P00001G0?ooo`030000003oool0oooo
0400oooo00120?ooo`030000003oool0oooo05H0oooo0P00001e0?ooo`800000F03oool00`000000
oooo0?ooo`110?ooo`00@`3oool00`000000oooo0?ooo`1G0?ooo`<00000K`3oool3000005T0oooo
00<000000?ooo`3oool0@P3oool004@0oooo00<000000?ooo`3oool0F@3oool2000006/0oooo0P00
001K0?ooo`030000003oool0oooo04<0oooo00150?ooo`030000003oool0oooo05X0oooo0`00001U
0?ooo`<00000G03oool00`000000oooo0?ooo`140?ooo`00AP3oool00`000000oooo0?ooo`1L0?oo
o`800000H@3oool2000005h0oooo00<000000?ooo`3oool0A@3oool004L0oooo00<000000?ooo`3o
ool0G@3oool4000005T0oooo1000001O0?ooo`030000003oool0oooo04H0oooo00180?ooo`030000
003oool0oooo0600oooo1000001A0?ooo`@00000HP3oool00`000000oooo0?ooo`170?ooo`00B@3o
ool00`000000oooo0?ooo`1S0?ooo`<00000B`3oool3000006D0oooo00<000000?ooo`3oool0B03o
ool004X0oooo00<000000?ooo`3oool0I@3oool4000004<0oooo1000001W0?ooo`030000003oool0
oooo04T0oooo001;0?ooo`030000003oool0oooo06P0oooo1P00000g0?ooo`H00000JP3oool00`00
0000oooo0?ooo`1:0?ooo`00C03oool00`000000oooo0?ooo`1]0?ooo`D00000;@3oool5000006l0
oooo00<000000?ooo`3oool0B`3oool004d0oooo00<000000?ooo`3oool0L@3oool:000001T0oooo
2P00001c0?ooo`030000003oool0oooo04`0oooo001>0?ooo`030000003oool0oooo07X0oooo6@00
001l0?ooo`030000003oool0oooo04d0oooo001?0?ooo`030000003oool0oooo0?l0oooo3P3oool0
0`000000oooo0?ooo`1>0?ooo`00D03oool00`000000oooo0?ooo`3o0?ooo``0oooo00<000000?oo
o`3oool0C`3oool00540oooo00<000000?ooo`3oool0o`3oool:0?ooo`030000003oool0oooo0500
oooo001B0?ooo`800000o`3oool80?ooo`800000D`3oool005@0oooo00<000000?ooo`3oool0o`3o
ool40?ooo`030000003oool0oooo05<0oooo001E0?ooo`030000003oool0oooo0?l0oooo0P3oool0
0`000000oooo0?ooo`1D0?ooo`00EP3oool200000?l0oooo0P00001G0?ooo`00F03oool00`000000
oooo0?ooo`3k0?ooo`030000003oool0oooo05L0oooo001I0?ooo`030000003oool0oooo0?T0oooo
00<000000?ooo`3oool0F03oool005X0oooo00<000000?ooo`3oool0m`3oool00`000000oooo0?oo
o`1I0?ooo`00F`3oool00`000000oooo0?ooo`3e0?ooo`030000003oool0oooo05X0oooo001L0?oo
o`800000l`3oool2000005d0oooo001N0?ooo`030000003oool0oooo0>l0oooo00<000000?ooo`3o
ool0G@3oool005l0oooo00<000000?ooo`3oool0k@3oool00`000000oooo0?ooo`1N0?ooo`00H03o
ool200000>/0oooo0P00001Q0?ooo`00HP3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0
oooo0640oooo001S0?ooo`030000003oool0oooo0>@0oooo0P00001T0?ooo`00I03oool300000>40
oooo0P00001V0?ooo`00I`3oool200000=d0oooo0P00001X0?ooo`00J@3oool00`000000oooo0?oo
o`3I0?ooo`030000003oool0oooo06P0oooo001Z0?ooo`030000003oool0oooo0=L0oooo00<00000
0?ooo`3oool0J@3oool006/0oooo0P00003E0?ooo`800000K03oool006d0oooo0P00003A0?ooo`80
0000KP3oool006l0oooo00<000000?ooo`3oool0c@3oool00`000000oooo0?ooo`1^0?ooo`00L03o
ool200000</0oooo0P00001a0?ooo`00LP3oool200000<L0oooo0P00001c0?ooo`00M03oool20000
0<<0oooo0P00001e0?ooo`00MP3oool200000;l0oooo0P00001g0?ooo`00N03oool00`000000oooo
0?ooo`2k0?ooo`030000003oool0oooo07L0oooo001i0?ooo`800000^@3oool2000007X0oooo001k
0?ooo`800000]@3oool2000007`0oooo001m0?ooo`800000/@3oool2000007h0oooo001o0?ooo`<0
0000Z`3oool300000800oooo00220?ooo`800000Y`3oool2000008<0oooo00240?ooo`800000X`3o
ool2000008D0oooo00260?ooo`<00000W@3oool3000008L0oooo00290?ooo`800000V@3oool20000
08X0oooo002;0?ooo`800000U@3oool2000008`0oooo002=0?ooo`<00000S`3oool3000008h0oooo
002@0?ooo`800000R`3oool200000940oooo002B0?ooo`<00000Q@3oool3000009<0oooo002E0?oo
o`@00000O03oool5000009H0oooo002I0?ooo`<00000M`3oool2000009/0oooo002L0?ooo`800000
L`3oool2000009d0oooo002N0?ooo`D00000J@3oool5000009l0oooo002S0?ooo`D00000H@3oool3
00000:@0oooo002X0?ooo`<00000F@3oool500000:L0oooo002[0?ooo`D00000C`3oool500000:`0
oooo002`0?ooo`D00000A@3oool500000;40oooo002e0?ooo`L00000=`3oool700000;H0oooo002l
0?ooo`/000008@3oool;00000;d0oooo00370?ooob400000b03oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.704368, -0.704368, 0.0049028, \
0.0049028}}]
}, Closed]],

Cell[BoxData[
    \(\(VelocityPlot2[LabVelocities, vCM, cmVelocities];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[RGBColor[0, 1, 0]], plot1, 
        Graphics[Circle[vCM, magnitude[vi]\/2]], Graphics[RGBColor[1, 0, 0]], 
        plot2, Graphics[Circle[vCM, magnitude[vf]\/2]], 
        AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.145069 0.709863 0.5 0.709863 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 1 0 r
.5 Mabswid
[ ] 0 setdash
.14507 .5 m
.85493 .5 L
s
% Start of user PostScript
0. 0. 1. 0. MAarrowhead1
% End of user PostScript
.14507 .5 m
.14507 .5 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.14507 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.85493 .5 L
s
% Start of user PostScript
0.5 0. 1. 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.14507 .5 L
s
% Start of user PostScript
0.5 0. 0. 0. MAarrowhead1
% End of user PostScript
newpath
.5 .5 .35493 0 365.73 arc
s
1 0 0 r
.14507 .5 m
.89117 .77156 L
s
% Start of user PostScript
0. 0. 1.0510491 0.38255057 MAarrowhead1
% End of user PostScript
.14507 .5 m
.10883 .22844 L
s
% Start of user PostScript
0. 0. -0.051049055 -0.38255057 MAarrowhead1
% End of user PostScript
.14507 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. 0.5 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.89117 .77156 L
s
% Start of user PostScript
0.5 0. 1.0510491 0.38255057 MAarrowhead1
% End of user PostScript
.5 .5 m
.10883 .22844 L
s
% Start of user PostScript
0.5 0. -0.051049055 -0.38255057 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ0?l00<P0oooo002m0?ooo`X0o`008@3oool:0?l00;h0oooo002e0?ooo`P0o`00
=@3oool80?l00;H0oooo002`0?ooo`D0o`00A@3oool50?l00;40oooo002[0?ooo`D0o`00C`3oool5
0?l00:`0oooo002V0?ooo`D0o`00F@3oool20?l00:X0oooo002T0?ooo`80o`00H03oool50?l00:D0
oooo002O0?ooo`D0o`00I`3oool50?l00:00oooo002M0?ooo`80o`00L@3oool30?l009d0oooo002J
0?ooo`<0o`00MP3oool20?l009/0oooo002E0?ooo`D0o`00N`3oool50?l009H0oooo002C0?ooo`80
o`00Q@3oool20?l009@0oooo002@0?ooo`<0o`00R@3oool30?l00940oooo002>0?ooo`80o`00S`3o
ool20?l008l0oooo002<0?ooo`80o`00T`3oool20?l008d0oooo00290?ooo`<0o`00U`3oool30?l0
08X0oooo00270?ooo`80o`00W@3oool20?l008P0oooo00250?ooo`80o`00X@3oool20?l008H0oooo
00220?ooo`<0o`00Y@3oool30?l008<0oooo00200?ooo`80o`00Z`3oool20?l00840oooo001n0?oo
o`80o`00[`3oool20?l007l0oooo001k0?ooo`<0o`00/`3oool30?l007`0oooo001i0?ooo`80o`00
^@3oool20?l007X0oooo001h0?ooo`030?l0003oool0oooo0;/0oooo00<0o`000?ooo`3oool0M`3o
ool007L0oooo00<0o`000?ooo`3oool0_@3oool00`3o0000oooo0?ooo`1f0?ooo`00M@3oool20?l0
0<40oooo0P3o001f0?ooo`00LP3oool30?l00<D0oooo0`3o001c0?ooo`00L03oool20?l00</0oooo
0P3o001a0?ooo`00K`3oool00`3o0000oooo0?ooo`3=0?ooo`030?l0003oool0oooo06h0oooo001^
0?ooo`030?l0003oool0oooo0<l0oooo00<0o`000?ooo`3oool0K@3oool006`0oooo0P3o003C0?oo
o`80o`00K@3oool006X0oooo0P3o003G0?ooo`80o`00J`3oool006T0oooo00<0o`000?ooo`3oool0
f@3oool00`3o0000oooo0?ooo`1X0?ooo`00J03oool00`3o0000oooo0?ooo`3K0?ooo`80o`00J03o
ool006D0oooo0`3o003P0?ooo`80o`00IP3oool006<0oooo0P3o003U0?ooo`80o`00I03oool00680
oooo00<0o`000?ooo`3oool0i`3oool00`3o0000oooo0?ooo`1Q0?ooo`00H@3oool00`3o0000oooo
0?ooo`3Y0?ooo`030?l0003oool0oooo0600oooo001O0?ooo`80o`00k@3oool20?l00600oooo001N
0?ooo`030?l0003oool0oooo0>l0oooo00<0o`000?ooo`3oool0G@3oool005d0oooo00<0o`000?oo
o`3oool0l@3oool00`3o0000oooo0?ooo`1L0?ooo`00F`3oool20?l00?D0oooo0P3o001L0?ooo`00
FP3oool00`3o0000oooo0?ooo`3g0?ooo`030?l0003oool0oooo05T0oooo001I0?ooo`030?l0003o
ool0oooo0?T0oooo00<0o`000?ooo`3oool0F03oool005P0oooo00<0o`000?ooo`3oool0n`3oool0
0`3o0000oooo0?ooo`1G0?ooo`00E`3oool00`3o0000oooo0?ooo`3m0?ooo`030?l0003oool0oooo
05H0oooo001E0?ooo`80o`00o`3oool20?ooo`80o`00EP3oool005@0oooo00<0o`000?ooo`3oool0
o`3oool40?ooo`030?l0003oool0oooo05<0oooo001C0?ooo`030?l0003oool0oooo0?l0oooo1P3o
ool00`3o0000oooo0?ooo`1B0?ooo`00D@3oool20?l00?l0oooo2P3oool20?l00580oooo001@0?oo
o`030?l0003oool0oooo0?l0oooo303oool00`3o0000oooo0?ooo`1?0?ooo`00C`3oool00`3o0000
oooo0?ooo`3o0?ooo`h0oooo00<0o`000?ooo`3oool0CP3oool004h0oooo00<0o`000?ooo`3oool0
N`3ooolG003o07d0oooo00<0o`000?ooo`3oool0C@3oool004d0oooo00<0o`000?ooo`3oool0LP3o
ool:003o01L0oooo2P00o`1d0?ooo`030?l0003oool0oooo04`0oooo001<0?ooo`030?l0003oool0
oooo06d0oooo1P00o`0[0?ooo`H00?l0K`3oool00`3o0000oooo0?ooo`1;0?ooo`00B`3oool00`3o
0000oooo0?ooo`1Y0?ooo`D00?l0=`3oool5003o06/0oooo00<0o`000?ooo`3oool0BP3oool004X0
oooo00<0o`000?ooo`3oool0IP3oool4003o0440oooo1000o`1X0?ooo`030?l0003oool0oooo04T0
oooo00190?ooo`030?l0003oool0oooo06<0oooo1000o`190?ooo`@00?l0I@3oool00`3o0000oooo
0?ooo`180?ooo`00B03oool00`3o0000oooo0?ooo`1Q0?ooo`<00?l0D@3oool3003o06<0oooo00<0
o`000?ooo`3oool0A`3oool004L0oooo00<0o`000?ooo`3oool0GP3oool4003o05L0oooo1000o`1P
0?ooo`030?l0003oool0oooo04H0oooo00160?ooo`030?l0003oool0oooo05d0oooo0P00o`1O0?oo
o`800?l0G`3oool00`3o0000oooo0?ooo`150?ooo`00A@3oool00`3o0000oooo0?ooo`1K0?ooo`<0
0?l0H`3oool3003o05d0oooo00<0o`000?ooo`3oool0A03oool004@0oooo00<0o`000?ooo`3oool0
FP3oool2003o06T0oooo0P00o`1L0?ooo`030?l0003oool0oooo04<0oooo00130?ooo`030?l0003o
ool0oooo05L0oooo1000o`1]0?ooo`@00?l0F@3oool00`3o0000oooo0?ooo`120?ooo`00@P3oool0
0`3o0000oooo0?ooo`1G0?ooo`03003o003oool0oooo07<0oooo00<00?l00?ooo`3oool0E`3oool0
0`3o0000oooo0?ooo`110?ooo`00@@3oool00`3o0000oooo0?ooo`1F0?ooo`800?l0M`3oool2003o
05P0oooo00<0o`000?ooo`3oool0@03oool00400oooo00<0o`000?ooo`3oool0D`3oool4003o07/0
oooo1000o`1E0?ooo`030?l0003oool0oooo03l0oooo000o0?ooo`030?l0003oool0oooo05<0oooo
00<00?l00?ooo`3oool0P@3oool00`00o`00oooo0?ooo`1C0?ooo`030?l0003oool0oooo03h0oooo
000n0?ooo`030?l0003oool0oooo0580oooo0P00o`250?ooo`800?l0E03oool00`3o0000oooo0?oo
o`0m0?ooo`00?@3oool00`3o0000oooo0?ooo`1A0?ooo`800?l0R@3oool2003o05<0oooo00<0o`00
0?ooo`3oool0?03oool003`0oooo00<0o`000?ooo`3oool0D@3oool00`00o`00oooo0?ooo`2;0?oo
o`800?l0DP3oool00`3o0000oooo0?ooo`0k0?ooo`00>`3oool00`3o0000oooo0?ooo`1@0?ooo`80
0?l0T03oool00`00o`00oooo0?ooo`1@0?ooo`030?l0003oool0oooo03X0oooo000j0?ooo`030?l0
003oool0oooo04l0oooo0P00o`2C0?ooo`800?l0D03oool00`3o0000oooo0?ooo`0j0?ooo`00>P3o
ool00`3o0000oooo0?ooo`1<0?ooo`<00?l0U`3oool2003o04l0oooo00<0o`000?ooo`3oool0>@3o
ool003T0oooo00<0o`000?ooo`3oool0B`3oool2003o09`0oooo0`00o`1=0?ooo`030?l0003oool0
oooo03P0oooo000h0?ooo`030?l0003oool0oooo04/0oooo00<00?l00?ooo`3oool0W`3oool00`00
o`00oooo0?ooo`1;0?ooo`030?l0003oool0oooo03L0oooo000g0?ooo`030?l0003oool0oooo04X0
oooo0P00o`2S0?ooo`03003o003oool0oooo04X0oooo00<0o`000?ooo`3oool0=`3oool003L0oooo
00<0o`000?ooo`3oool0B@3oool00`00o`00oooo0?ooo`2T0?ooo`03003o003oool0oooo04X0oooo
00<0o`000?ooo`3oool0=P3oool003H0oooo00<0o`000?ooo`3oool0B@3oool00`00o`00oooo0?oo
o`2V0?ooo`800?l0B`3oool00`3o0000oooo0?ooo`0e0?ooo`00=@3oool00`3o0000oooo0?ooo`19
0?ooo`03003o003oool0oooo0:T0oooo00<00?l00?ooo`3oool0B@3oool00`3o0000oooo0?ooo`0d
0?ooo`00=03oool00`3o0000oooo0?ooo`180?ooo`800?l0[@3oool2003o04X0oooo00<0o`000?oo
o`3oool0<`3oool003<0oooo00<0o`000?ooo`3oool0B03oool00`00o`00oooo0?ooo`2_0?ooo`03
003o003oool0oooo04P0oooo00<0o`000?ooo`3oool0<P3oool00380oooo00<0o`000?ooo`3oool0
A`3oool2003o0;<0oooo0P00o`180?ooo`030?l0003oool0oooo0380oooo000b0?ooo`030?l0003o
ool0oooo04H0oooo00<00?l00?ooo`3oool0]@3oool00`00o`00oooo0?ooo`160?ooo`030?l0003o
ool0oooo0340oooo000a0?ooo`030?l0003oool0oooo04D0oooo0P00o`2i0?ooo`800?l0A`3oool0
0`3o0000oooo0?ooo`0`0?ooo`00<03oool00`3o0000oooo0?ooo`150?ooo`03003o003oool0oooo
0;/0oooo00<00?l00?ooo`3oool0A@3oool00`3o0000oooo0?ooo`0_0?ooo`00;`3oool00`3o0000
oooo0?ooo`150?ooo`03003o003oool0oooo0;d0oooo00<00?l00?ooo`3oool0A03oool00`3o0000
oooo0?ooo`0_0?ooo`00;`3oool00`3o0000oooo0?ooo`140?ooo`03003o003oool0oooo0;l0oooo
00<00?l00?ooo`3oool0A03oool00`3o0000oooo0?ooo`0^0?ooo`00;P3oool20?l004D0oooo00<0
0?l00?ooo`3oool0`@3oool00`00o`00oooo0?ooo`140?ooo`030?l0003oool0oooo02d0oooo000]
0?ooo`D0o`00@@3oool2003o0<D0oooo0P00o`150?ooo`030?l0003oool0oooo02`0oooo000/0?oo
o`030?l0003oool0o`0000H0o`00?@3oool00`00o`00oooo0?ooo`370?ooo`03003o003oool0oooo
0480oooo00<0o`000?ooo`3oool0;03oool002`0oooo00<0o`000?ooo`3o0000203o000j0?ooo`03
003o003oool0oooo0<T0oooo00<00?l00?ooo`3oool0@P3oool00`3o0000oooo0?ooo`0[0?ooo`00
;03oool00`3o0000oooo0?l000070?l003X0oooo00<00?l00?ooo`3oool0b`3oool00`00o`00oooo
0?ooo`110?ooo`030?l0003oool0oooo02/0oooo000[0?ooo`030?l0003oool0oooo00<0o`0000<0
oooo0?l0003o00000P3o000i0?ooo`03003o003oool0oooo0<d0oooo00<00?l00?ooo`3oool0@03o
ool00`3o0000oooo0?ooo`0[0?ooo`00:`3oool00`3o0000oooo0?ooo`040?l000040?ooo`3o0000
oooo0?ooo`80o`00=@3oool2003o0=40oooo00<00?l00?ooo`3oool0@03oool00`3o0000oooo0?oo
o`0Z0?ooo`00:P3oool0103o0000oooo0?ooo`3oool40?l000H0oooo00<0o`000?ooo`3oool0<@3o
ool00`00o`00oooo0?ooo`3B0?ooo`03003o003oool0oooo0400oooo00<0o`000?ooo`3oool0:@3o
ool002T0oooo00<0o`000?ooo`3oool00P3oool20?l000T0oooo0P3o000`0?ooo`03003o003oool0
oooo0=@0oooo00<00?l00?ooo`3oool0@03oool00`3o0000oooo0?ooo`0X0?ooo`00:03oool00`3o
0000oooo0?ooo`040?ooo`030?l0003oool0oooo00T0oooo00<0o`000?ooo`3oool0;@3oool00`00
o`00oooo0?ooo`3E0?ooo`800?l0@03oool00`3o0000oooo0?ooo`0X0?ooo`00:03oool00`3o0000
oooo0?ooo`040?ooo`030?l0003oool0oooo00X0oooo0P3o000/0?ooo`03003o003oool0oooo0=P0
oooo00<00?l00?ooo`3oool0?P3oool00`3o0000oooo0?ooo`0W0?ooo`009`3oool00`3o0000oooo
0?ooo`060?ooo`030?l0003oool0oooo00/0oooo00<0o`000?ooo`3oool0:03oool00`00o`00oooo
0?ooo`3J0?ooo`03003o003oool0oooo03d0oooo00<0o`000?ooo`3oool09`3oool002L0oooo00<0
o`000?ooo`3oool01P3oool00`3o0000oooo0?ooo`0<0?ooo`80o`009`3oool00`00o`00oooo0?oo
o`3K0?ooo`03003o003oool0oooo03h0oooo00<0o`000?ooo`3oool09P3oool002H0oooo00<0o`00
0?ooo`3oool01`3oool00`3o0000oooo0?ooo`0>0?ooo`030?l0003oool0oooo02<0oooo00<00?l0
0?ooo`3oool0g@3oool00`00o`00oooo0?ooo`0n0?ooo`030?l0003oool0oooo02D0oooo000U0?oo
o`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool03`3oool00`3o0000oooo0?ooo`0Q0?oo
o`03003o003oool0oooo0=l0oooo00<00?l00?ooo`3oool0?P3oool00`3o0000oooo0?ooo`0T0?oo
o`009@3oool00`3o0000oooo0?ooo`080?ooo`030?l0003oool0oooo0100oooo0P3o000P0?ooo`03
003o003oool0oooo0>40oooo00<00?l00?ooo`3oool0?@3oool00`3o0000oooo0?ooo`0T0?ooo`00
903oool00`3o0000oooo0?ooo`090?ooo`030?l0003oool0oooo0180oooo00<0o`000?ooo`3oool0
703oool00`00o`00oooo0?ooo`3S0?ooo`03003o003oool0oooo03d0oooo00<0o`000?ooo`3oool0
8`3oool002@0oooo00<0o`000?ooo`3oool02@3oool00`3o0000oooo0?ooo`0C0?ooo`80o`006`3o
ool00`00o`00oooo0?ooo`3U0?ooo`03003o003oool0oooo03`0oooo00<0o`000?ooo`3oool08`3o
ool002<0oooo00<0o`000?ooo`3oool02P3oool00`3o0000oooo0?ooo`0E0?ooo`030?l0003oool0
oooo01P0oooo00<00?l00?ooo`3oool0i@3oool00`00o`00oooo0?ooo`0l0?ooo`030?l0003oool0
oooo02<0oooo000S0?ooo`030?l0003oool0oooo00/0oooo00<0o`000?ooo`3oool05@3oool20?l0
01L0oooo00<00?l00?ooo`3oool0i`3oool00`00o`00oooo0?ooo`0l0?ooo`030?l0003oool0oooo
0280oooo000R0?ooo`030?l0003oool0oooo00`0oooo00<0o`000?ooo`3oool05`3oool00`3o0000
oooo0?ooo`0C0?ooo`03003o003oool0oooo0>T0oooo00<00?l00?ooo`3oool0>`3oool00`3o0000
oooo0?ooo`0R0?ooo`008P3oool00`3o0000oooo0?ooo`0<0?ooo`030?l0003oool0oooo01P0oooo
0P3o000B0?ooo`03003o003oool0oooo0>/0oooo00<00?l00?ooo`3oool0>`3oool00`3o0000oooo
0?ooo`0Q0?ooo`008@3oool00`3o0000oooo0?ooo`0=0?ooo`030?l0003oool0oooo01X0oooo00<0
o`000?ooo`3oool03P3oool00`00o`00oooo0?ooo`3]0?ooo`03003o003oool0oooo03/0oooo00<0
o`000?ooo`3oool0803oool00200oooo00<0o`000?ooo`3oool03P3oool00`3o0000oooo0?ooo`0K
0?ooo`030?l0003oool0oooo00`0oooo00<00?l00?ooo`3oool0k`3oool00`00o`00oooo0?ooo`0k
0?ooo`030?l0003oool0oooo01l0oooo000P0?ooo`030?l0003oool0oooo00h0oooo00<0o`000?oo
o`3oool0703oool20?l000`0oooo00<00?l00?ooo`3oool0k`3oool00`00o`00oooo0?ooo`0k0?oo
o`030?l0003oool0oooo01l0oooo000O0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3o
ool07P3oool00`3o0000oooo0?ooo`080?ooo`03003o003oool0oooo0?40oooo00<00?l00?ooo`3o
ool0>`3oool00`3o0000oooo0?ooo`0N0?ooo`007`3oool00`3o0000oooo0?ooo`0@0?ooo`030?l0
003oool0oooo01h0oooo0P3o00070?ooo`03003o003oool0oooo0?<0oooo00<00?l00?ooo`3oool0
>P3oool00`3o0000oooo0?ooo`0N0?ooo`007P3oool00`3o0000oooo0?ooo`0A0?ooo`030?l0003o
ool0oooo0200oooo00<0o`000?ooo`3oool00`3oool00`00o`00oooo0?ooo`3d0?ooo`03003o003o
ool0oooo03X0oooo00<0o`000?ooo`3oool07P3oool001h0oooo00<0o`000?ooo`3oool04@3oool0
0`3o0000oooo0?ooo`0Q0?ooo`80o`000`3oool00`00o`00oooo0?ooo`3e0?ooo`03003o003oool0
oooo03X0oooo00<0o`000?ooo`3oool07@3oool001d0oooo00<0o`000?ooo`3oool04P3oool00`3o
0000oooo0?ooo`0S0?ooo`030?l0003oool00?l00?T0oooo00<00?l00?ooo`3oool0>@3oool00`3o
0000oooo0?ooo`0M0?ooo`007@3oool00`3o0000oooo0?ooo`0B0?ooo`030?l0003oool0oooo02@0
oooo0P3o003i0?ooo`03003o003oool0oooo03X0oooo00<0o`000?ooo`3oool0703oool001d0oooo
00<0o`000?ooo`3oool04P3oool00`3o0000oooo0?ooo`0T0?ooo`03003o003oool0o`000?T0oooo
00<00?l00?ooo`3oool0>@3oool00`3o0000oooo0?ooo`0L0?ooo`00703oool00`3o0000oooo0?oo
o`0C0?ooo`030?l0003oool0oooo02<0oooo00D00?l00?ooo`3oool0oooo0?l0003i0?ooo`03003o
003oool0oooo03T0oooo00<0o`000?ooo`3oool06`3oool001`0oooo00<0o`000?ooo`3oool0503o
ool00`3o0000oooo0?ooo`0Q0?ooo`03003o003oool0oooo00<0oooo0P3o003h0?ooo`03003o003o
ool0oooo03P0oooo00<0o`000?ooo`3oool06`3oool001/0oooo00<0o`000?ooo`3oool05@3oool0
0`3o0000oooo0?ooo`0Q0?ooo`03003o003oool0oooo00D0oooo00<0o`000?ooo`3oool0mP3oool0
0`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01/0oooo000K0?ooo`030?l0003oool0oooo
01D0oooo00<0o`000?ooo`3oool0803oool00`00o`00oooo0?ooo`070?ooo`80o`00mP3oool00`00
o`00oooo0?ooo`0h0?ooo`030?l0003oool0oooo01X0oooo000J0?ooo`030?l0003oool0oooo01H0
oooo00<0o`000?ooo`3oool07`3oool00`00o`00oooo0?ooo`0:0?ooo`030?l0003oool0oooo0?@0
oooo00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?ooo`0J0?ooo`006P3oool00`3o0000oooo
0?ooo`0F0?ooo`030?l0003oool0oooo01h0oooo00<00?l00?ooo`3oool0303oool20?l00?D0oooo
00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?ooo`0I0?ooo`006P3oool00`3o0000oooo0?oo
o`0F0?ooo`030?l0003oool0oooo01d0oooo00<00?l00?ooo`3oool03`3oool00`3o0000oooo0?oo
o`3b0?ooo`03003o003oool0oooo03L0oooo00<0o`000?ooo`3oool06@3oool001T0oooo00<0o`00
0?ooo`3oool05`3oool00`3o0000oooo0?ooo`0M0?ooo`03003o003oool0oooo0100oooo0P3o003c
0?ooo`03003o003oool0oooo03L0oooo00<0o`000?ooo`3oool0603oool001T0oooo00<0o`000?oo
o`3oool05`3oool00`3o0000oooo0?ooo`0M0?ooo`03003o003oool0oooo0180oooo00<0o`000?oo
o`3oool0l03oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01P0oooo000H0?ooo`03
0?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo0?ooo`0D0?ooo`03
0?l0003oool0oooo0>l0oooo00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?ooo`0H0?ooo`00
603oool00`3o0000oooo0?ooo`0I0?ooo`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3oool0
5@3oool20?l00?00oooo00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?ooo`0G0?ooo`005`3o
ool00`3o0000oooo0?ooo`0J0?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool0603o
ool00`3o0000oooo0?ooo`3]0?ooo`03003o003oool0oooo03L0oooo00<0o`000?ooo`3oool05`3o
ool001L0oooo00<0o`000?ooo`3oool06P3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0
oooo01T0oooo0P3o003^0?ooo`03003o003oool0oooo03L0oooo00<0o`000?ooo`3oool05P3oool0
01L0oooo00<0o`000?ooo`3oool06P3oool00`3o0000oooo0?ooo`0I0?ooo`03003o003oool0oooo
01`0oooo00<0o`000?ooo`3oool0j`3oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo
01H0oooo000F0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0603oool00`00o`00
oooo0?ooo`0N0?ooo`80o`00k03oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01D0
oooo000F0?ooo`030?l0003oool0oooo01/0oooo00<0o`000?ooo`3oool0603oool00`00o`00oooo
0?ooo`0P0?ooo`030?l0003oool0oooo0>X0oooo00<00?l00?ooo`3oool0=P3oool00`3o0000oooo
0?ooo`0E0?ooo`005@3oool00`3o0000oooo0?ooo`0M0?ooo`030?l0003oool0oooo01H0oooo00<0
0?l00?ooo`3oool08P3oool20?l00>X0oooo00<00?l00?ooo`3oool0=P3oool00`3o0000oooo0?oo
o`0E0?ooo`005@3oool00`3o0000oooo0?ooo`0M0?ooo`030?l0003oool0oooo01H0oooo00<00?l0
0?ooo`3oool0903oool00`3o0000oooo0?ooo`3X0?ooo`03003o003oool0oooo03H0oooo00<0o`00
0?ooo`3oool0503oool001D0oooo00<0o`000?ooo`3oool07@3oool00`3o0000oooo0?ooo`0E0?oo
o`03003o003oool0oooo02H0oooo00<0o`000?ooo`3oool0i`3oool00`00o`00oooo0?ooo`0f0?oo
o`030?l0003oool0oooo01@0oooo000D0?ooo`030?l0003oool0oooo01h0oooo00<0o`000?ooo`3o
ool05@3oool00`00o`00oooo0?ooo`0W0?ooo`80o`00j03oool00`00o`00oooo0?ooo`0f0?ooo`03
0?l0003oool0oooo01<0oooo000D0?ooo`030?l0003oool0oooo01h0oooo00<0o`000?ooo`3oool0
503oool00`00o`00oooo0?ooo`0Z0?ooo`030?l0003oool0oooo0>H0oooo00<00?l00?ooo`3oool0
=@3oool00`3o0000oooo0?ooo`0C0?ooo`00503oool00`3o0000oooo0?ooo`0N0?ooo`030?l0003o
ool0oooo01@0oooo00<00?l00?ooo`3oool0:`3oool20?l00>H0oooo00<00?l00?ooo`3oool0=@3o
ool00`3o0000oooo0?ooo`0C0?ooo`00503oool00`3o0000oooo0?ooo`0N0?ooo`030?l0003oool0
oooo01@0oooo00<00?l00?ooo`3oool0;@3oool00`3o0000oooo0?ooo`3S0?ooo`03003o003oool0
oooo03D0oooo00<0o`000?ooo`3oool04`3oool001<0oooo00<0o`000?ooo`3oool0803oool00`3o
0000oooo0?ooo`0B0?ooo`03003o003oool0oooo02l0oooo0P3o003S0?ooo`03003o003oool0oooo
03D0oooo00<0o`000?ooo`3oool04`3oool001<0oooo00<0o`000?ooo`3oool0803oool00`3o0000
oooo0?ooo`0B0?ooo`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool0h@3oool00`00o`00
oooo0?ooo`0e0?ooo`030?l0003oool0oooo0180oooo000C0?ooo`030?l0003oool0oooo0200oooo
00<0o`000?ooo`3oool04@3oool00`00o`00oooo0?ooo`0c0?ooo`80o`00h@3oool00`00o`00oooo
0?ooo`0e0?ooo`030?l0003oool0oooo0180oooo000B0?ooo`030?l0003oool0oooo0240oooo00<0
o`000?ooo`3oool04@3oool00`00o`00oooo0?ooo`0e0?ooo`030?l0003oool0oooo0=l0oooo00<0
0?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0B0?ooo`004P3oool00`3o0000oooo0?ooo`0Q
0?ooo`030?l0003oool0oooo0100oooo00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?ooo`3O
0?ooo`03003o003oool0oooo03@0oooo00<0o`000?ooo`3oool04@3oool00140oooo00<0o`000?oo
o`3oool08P3oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo03P0oooo0P3o003O0?oo
o`03003o003oool0oooo03@0oooo00<0o`000?ooo`3oool04@3oool00140oooo00<0o`000?ooo`3o
ool08P3oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo03X0oooo00<0o`000?ooo`3o
ool0g03oool00`00o`00oooo0?ooo`0e0?ooo`030?l0003oool0oooo0100oooo000A0?ooo`030?l0
003oool0oooo0280oooo00<0o`000?ooo`3oool03`3oool00`00o`00oooo0?ooo`0l0?ooo`80o`00
g03oool00`00o`00oooo0?ooo`0e0?ooo`030?l0003oool0oooo0100oooo000A0?ooo`030?l0003o
ool0oooo02<0oooo00<0o`000?ooo`3oool03P3oool00`00o`00oooo0?ooo`0n0?ooo`030?l0003o
ool0oooo0=X0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0@0?ooo`004@3oool0
0`3o0000oooo0?ooo`0S0?ooo`030?l0003oool0oooo00d0oooo00<00?l00?ooo`3oool0@03oool2
0?l00=X0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0@0?ooo`00403oool00`3o
0000oooo0?ooo`0T0?ooo`030?l0003oool0oooo00d0oooo00<00?l00?ooo`3oool0@P3oool00`3o
0000oooo0?ooo`3H0?ooo`03003o003oool0oooo03<0oooo00<0o`000?ooo`3oool0403oool00100
oooo00<0o`000?ooo`3oool0903oool00`3o0000oooo0?ooo`0=0?ooo`03003o003oool0oooo04<0
oooo0P3o003H0?ooo`03003o003oool0oooo03@0oooo00<0o`000?ooo`3oool03`3oool00100oooo
00<0o`000?ooo`3oool0903oool00`3o0000oooo0?ooo`0<0?ooo`03003o003oool0oooo04H0oooo
00<0o`000?ooo`3oool0e@3oool00`00o`00oooo0?ooo`0d0?ooo`030?l0003oool0oooo00l0oooo
000@0?ooo`030?l0003oool0oooo02@0oooo00<0o`000?ooo`3oool0303oool00`00o`00oooo0?oo
o`170?ooo`030?l0003oool0oooo0=D0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?oo
o`0>0?ooo`00403oool00`3o0000oooo0?ooo`0T0?ooo`030?l0003oool0oooo00/0oooo00<00?l0
0?ooo`3oool0B@3oool20?l00=D0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0>
0?ooo`003`3oool00`3o0000oooo0?ooo`0V0?ooo`030?l0003oool0oooo00X0oooo00<00?l00?oo
o`3oool0B`3oool00`3o0000oooo0?ooo`3C0?ooo`03003o003oool0oooo03<0oooo00<0o`000?oo
o`3oool03P3oool000l0oooo00<0o`000?ooo`3oool09P3oool00`3o0000oooo0?ooo`0:0?ooo`03
003o003oool0oooo04`0oooo0P3o003C0?ooo`03003o003oool0oooo03<0oooo00<0o`000?ooo`3o
ool03P3oool000l0oooo00<0o`000?ooo`3oool09P3oool00`3o0000oooo0?ooo`0:0?ooo`03003o
003oool0oooo04h0oooo00<0o`000?ooo`3oool0d03oool00`00o`00oooo0?ooo`0c0?ooo`030?l0
003oool0oooo00h0oooo000>0?ooo`030?l0003oool0oooo02L0oooo00<0o`000?ooo`3oool02@3o
ool00`00o`00oooo0?ooo`1@0?ooo`80o`00d03oool00`00o`00oooo0?ooo`0d0?ooo`030?l0003o
ool0oooo00d0oooo000>0?ooo`030?l0003oool0oooo02L0oooo00<0o`000?ooo`3oool02@3oool0
0`00o`00oooo0?ooo`1B0?ooo`030?l0003oool0oooo0<h0oooo00<00?l00?ooo`3oool0<`3oool0
0`3o0000oooo0?ooo`0=0?ooo`003P3oool00`3o0000oooo0?ooo`0W0?ooo`030?l0003oool0oooo
00T0oooo00<00?l00?ooo`3oool0D`3oool20?l00<h0oooo00<00?l00?ooo`3oool0<`3oool00`3o
0000oooo0?ooo`0=0?ooo`003P3oool00`3o0000oooo0?ooo`0W0?ooo`030?l0003oool0oooo00T0
oooo00<00?l00?ooo`3oool0E@3oool00`3o0000oooo0?ooo`3;0?ooo`03003o003oool0oooo03<0
oooo00<0o`000?ooo`3oool03@3oool000h0oooo00<0o`000?ooo`3oool0:03oool00`3o0000oooo
0?ooo`070?ooo`03003o003oool0oooo05L0oooo00<0o`000?ooo`3oool0b`3oool00`00o`00oooo
0?ooo`0b0?ooo`030?l0003oool0oooo00d0oooo000=0?ooo`030?l0003oool0oooo02T0oooo00<0
o`000?ooo`3oool01`3oool00`00o`00oooo0?ooo`1H0?ooo`80o`00b`3oool00`00o`00oooo0?oo
o`0c0?ooo`030?l0003oool0oooo00`0oooo000=0?ooo`030?l0003oool0oooo02T0oooo00<0o`00
0?ooo`3oool01`3oool00`00o`00oooo0?ooo`1J0?ooo`030?l0003oool0oooo0<P0oooo00<00?l0
0?ooo`3oool0<`3oool00`3o0000oooo0?ooo`0<0?ooo`003@3oool00`3o0000oooo0?ooo`0Y0?oo
o`030?l0003oool0oooo00H0oooo00<00?l00?ooo`3oool0G03oool20?l00<P0oooo00<00?l00?oo
o`3oool0<`3oool00`3o0000oooo0?ooo`0<0?ooo`003@3oool00`3o0000oooo0?ooo`0Y0?ooo`03
0?l0003oool0oooo00H0oooo00<00?l00?ooo`3oool0GP3oool00`3o0000oooo0?ooo`360?ooo`03
003o003oool0oooo0380oooo00<0o`000?ooo`3oool0303oool000d0oooo00<0o`000?ooo`3oool0
:@3oool00`3o0000oooo0?ooo`060?ooo`03003o003oool0oooo05l0oooo0P3o00360?ooo`03003o
003oool0oooo0380oooo00<0o`000?ooo`3oool0303oool000`0oooo00<0o`000?ooo`3oool0:P3o
ool00`3o0000oooo0?ooo`060?ooo`03003o003oool0oooo0640oooo00<0o`000?ooo`3oool0``3o
ool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0
oooo02X0oooo00<0o`000?ooo`3oool01@3oool00`00o`00oooo0?ooo`1S0?ooo`80o`00``3oool0
0`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0oooo
02/0oooo00<0o`000?ooo`3oool0103oool00`00o`00oooo0?ooo`1U0?ooo`030?l0003oool0oooo
0<40oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`0;0?ooo`00303oool00`3o0000
oooo0?ooo`0[0?ooo`030?l0003oool0oooo00@0oooo00<00?l00?ooo`3oool0IP3oool00`3o0000
oooo0?ooo`300?ooo`03003o003oool0oooo0380oooo00<0o`000?ooo`3oool02`3oool000`0oooo
00<0o`000?ooo`3oool0:`3oool00`3o0000oooo0?ooo`040?ooo`03003o003oool0oooo06L0oooo
0P3o00300?ooo`03003o003oool0oooo0380oooo00<0o`000?ooo`3oool02`3oool000`0oooo00<0
o`000?ooo`3oool0:`3oool00`3o0000oooo0?ooo`040?ooo`03003o003oool0oooo06T0oooo00<0
o`000?ooo`3oool0_@3oool00`00o`00oooo0?ooo`0b0?ooo`030?l0003oool0oooo00/0oooo000<
0?ooo`030?l0003oool0oooo02/0oooo00<0o`000?ooo`3oool0103oool00`00o`00oooo0?ooo`1Z
0?ooo`80o`00_@3oool00`00o`00oooo0?ooo`0b0?ooo`030?l0003oool0oooo00/0oooo000;0?oo
o`030?l0003oool0oooo02`0oooo00<0o`000?ooo`3oool00`3oool00`00o`00oooo0?ooo`1]0?oo
o`030?l0003oool0oooo0;/0oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`0;0?oo
o`002`3oool00`3o0000oooo0?ooo`0/0?ooo`030?l0003oool0oooo00<0oooo00<00?l00?ooo`3o
ool0KP3oool20?l00;/0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`0:0?ooo`00
2`3oool00`3o0000oooo0?ooo`0]0?ooo`030?l0003oool0oooo0080oooo00<00?l00?ooo`3oool0
L03oool00`3o0000oooo0?ooo`2h0?ooo`03003o003oool0oooo0380oooo00<0o`000?ooo`3oool0
2P3oool000/0oooo00<0o`000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`020?ooo`03003o003o
ool0oooo0740oooo0P3o002h0?ooo`03003o003oool0oooo0380oooo00<0o`000?ooo`3oool02P3o
ool000/0oooo00<0o`000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`020?ooo`03003o003oool0
oooo07<0oooo00<0o`000?ooo`3oool0]@3oool00`00o`00oooo0?ooo`0b0?ooo`030?l0003oool0
oooo00X0oooo000;0?ooo`030?l0003oool0oooo02d0oooo00D0o`000?ooo`3oool0oooo003o001g
0?ooo`030?l0003oool0oooo0;@0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`0:
0?ooo`002`3oool00`3o0000oooo0?ooo`0]0?ooo`050?l0003oool0oooo0?ooo`00o`00N03oool2
0?l00;D0oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`0:0?ooo`002`3oool00`3o
0000oooo0?ooo`0]0?ooo`050?l0003oool0oooo0?ooo`00o`00NP3oool00`3o0000oooo0?ooo`2b
0?ooo`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool02P3oool000/0oooo00<0o`000?oo
o`3oool0;@3oool01@3o0000oooo0?ooo`3oool00?l007/0oooo0P3o002b0?ooo`03003o003oool0
oooo0340oooo00<0o`000?ooo`3oool02P3oool000/0oooo00<0o`000?ooo`3oool0;P3oool0103o
0000oooo0?ooo`00o`1m0?ooo`030?l0003oool0oooo0:l0oooo00<00?l00?ooo`3oool0<@3oool0
0`3o0000oooo0?ooo`0:0?ooo`002`3oool00`3o0000oooo0?ooo`0^0?ooo`040?l0003oool0oooo
003o07h0oooo0P3o002_0?ooo`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool02P3oool0
00X0oooo00<0o`000?ooo`3oool0;`3oool0103o0000oooo0?ooo`00o`200?ooo`030?l0003oool0
oooo0:`0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o
0000oooo0?ooo`0_0?ooo`040?l0003oool0oooo003o0840oooo0P3o002/0?ooo`03003o003oool0
oooo0380oooo00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3oool0;`3oool0103o
0000oooo0?ooo`00o`230?ooo`030?l0003oool0oooo0:T0oooo00<00?l00?ooo`3oool0<P3oool0
0`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0_0?ooo`040?l0003oool0oooo
003o08@0oooo00<0o`000?ooo`3oool0Z03oool00`00o`00oooo0?ooo`0b0?ooo`030?l0003oool0
oooo00T0oooo000:0?ooo`030?l0003oool0oooo02l0oooo00<0o`000?ooo`00o`00QP3oool20?l0
0:P0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000
oooo0?ooo`0_0?ooo`030?l0003oool00?l008P0oooo00<0o`000?ooo`3oool0YP3oool00`00o`00
oooo0?ooo`0a0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0300oooo
00<0o`00003o003oool0R03oool20?l00:H0oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo
0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0`0?ooo`030?l00000o`00oooo08X0oooo00<0
o`000?ooo`3oool0X`3oool00`00o`00oooo0?ooo`0a0?ooo`030?l0003oool0oooo00T0oooo000:
0?ooo`030?l0003oool0oooo0300oooo00<0o`00003o003oool0R`3oool20?l00:<0oooo00<00?l0
0?ooo`3oool0<@3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0`0?oo
o`030?l00000o`00oooo08d0oooo00<0o`000?ooo`3oool0X03oool00`00o`00oooo0?ooo`0a0?oo
o`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0300oooo00<0o`00003o003o
ool0SP3oool20?l00:00oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`090?ooo`00
2P3oool00`3o0000oooo0?ooo`0`0?ooo`030?l00000o`00oooo0900oooo00<0o`000?ooo`3oool0
W@3oool00`00o`00oooo0?ooo`0a0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003o
ool0oooo0300oooo00<0o`00003o003oool0T@3oool00`3o0000oooo0?ooo`2L0?ooo`03003o003o
ool0oooo0340oooo00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3oool0<@3oool0
0`3o0000oooo0?ooo`2A0?ooo`80o`00W03oool00`00o`00oooo0?ooo`0a0?ooo`030?l0003oool0
oooo00T0oooo000:0?ooo`030?l0003oool0oooo0340oooo00<0o`000?ooo`3oool0103oool2003o
08T0oooo0P3o00020?ooo`030?l0003oool0oooo0940oooo0P00o`060?ooo`03003o003oool0oooo
0340oooo00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3oool0<@3oool00`3o0000
oooo0?ooo`06003o08T0oooo1`3o002A0?ooo`H00?l00P3oool00`00o`00oooo0?ooo`0a0?ooo`03
0?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0340ooooVP3o002I003o03<0oooo
00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?ooo`3oool0<@3oool00`00o`00oooo0?l0
00020?l000@00?l0R@3oool60?l00080oooo0P3o002>0?ooo`L00?l000<0oooo003o003oool0<P3o
ool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0a0?ooo`03003o003oool0
oooo0080oooo0P3o0002003o08T0oooo0P3o00080?ooo`030?l0003oool0oooo08/0oooo0`00o`05
0?ooo`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool02@3oool000X0oooo00<0o`000?oo
o`3oool0<@3oool00`00o`00oooo0?ooo`040?ooo`<0o`00T`3oool20?l009<0oooo00<00?l00?oo
o`3oool0<@3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0a0?ooo`03
003o003oool0oooo00L0oooo0`3o002B0?ooo`030?l0003oool0oooo0900oooo00<00?l00?ooo`3o
ool0<@3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0a0?ooo`03003o
003oool0oooo00X0oooo0`3o002@0?ooo`030?l0003oool0oooo08l0oooo00<00?l00?ooo`3oool0
<@3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0a0?ooo`03003o003o
ool0oooo00d0oooo0P3o002?0?ooo`80o`00S`3oool00`00o`00oooo0?ooo`0a0?ooo`030?l0003o
ool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0340oooo00<00?l00?ooo`3oool03`3oool3
0?l008h0oooo00<0o`000?ooo`3oool0S03oool00`00o`00oooo0?ooo`0a0?ooo`030?l0003oool0
oooo00T0oooo000:0?ooo`030?l0003oool0oooo0340oooo00<00?l00?ooo`3oool04P3oool30?l0
08`0oooo0P3o002<0?ooo`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool02@3oool000X0
oooo00<0o`000?ooo`3oool0<@3oool00`00o`00oooo0?ooo`0E0?ooo`<0o`00R`3oool00`3o0000
oooo0?ooo`290?ooo`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool02@3oool000X0oooo
00<0o`000?ooo`3oool0<@3oool00`00o`00oooo0?ooo`0H0?ooo`80o`00RP3oool20?l008T0oooo
00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?oo
o`0a0?ooo`03003o003oool0oooo01X0oooo0`3o00290?ooo`030?l0003oool0oooo08H0oooo00<0
0?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`090?ooo`002P3oool00`3o0000oooo0?ooo`0a
0?ooo`03003o003oool0oooo01d0oooo0`3o00270?ooo`80o`00Q@3oool00`00o`00oooo0?ooo`0b
0?ooo`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0380oooo00<00?l00?oo
o`3oool07`3oool30?l008H0oooo00<0o`000?ooo`3oool0PP3oool00`00o`00oooo0?ooo`0b0?oo
o`030?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0380oooo00<00?l00?ooo`3o
ool08P3oool20?l008D0oooo00<0o`000?ooo`3oool0P@3oool00`00o`00oooo0?ooo`0b0?ooo`03
0?l0003oool0oooo00T0oooo000:0?ooo`030?l0003oool0oooo0380oooo00<00?l00?ooo`3oool0
903oool30?l008<0oooo0P3o00210?ooo`03003o003oool0oooo0380oooo00<0o`000?ooo`3oool0
2@3oool000X0oooo00<0o`000?ooo`3oool0<P3oool00`00o`00oooo0?ooo`0W0?ooo`<0o`00PP3o
ool00`3o0000oooo0?ooo`1n0?ooo`03003o003oool0oooo0380oooo00<0o`000?ooo`3oool02@3o
ool000/0oooo00<0o`000?ooo`3oool0<@3oool00`00o`00oooo0?ooo`0Z0?ooo`<0o`00P03oool2
0?l007h0oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`0:0?ooo`002`3oool00`3o
0000oooo0?ooo`0a0?ooo`03003o003oool0oooo02d0oooo0P3o00200?ooo`030?l0003oool0oooo
07/0oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`0:0?ooo`002`3oool00`3o0000
oooo0?ooo`0a0?ooo`03003o003oool0oooo02l0oooo0`3o001n0?ooo`80o`00N`3oool00`00o`00
oooo0?ooo`0a0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo0340oooo
00<00?l00?ooo`3oool0<P3oool30?l007d0oooo00<0o`000?ooo`3oool0N03oool00`00o`00oooo
0?ooo`0a0?ooo`030?l0003oool0oooo00X0oooo000;0?ooo`030?l0003oool0oooo0340oooo00<0
0?l00?ooo`3oool0=@3oool30?l007/0oooo0P3o001h0?ooo`03003o003oool0oooo0340oooo00<0
o`000?ooo`3oool02P3oool000/0oooo00<0o`000?ooo`3oool0<@3oool00`00o`00oooo0?ooo`0h
0?ooo`80o`00N`3oool00`3o0000oooo0?ooo`1d0?ooo`03003o003oool0oooo0380oooo00<0o`00
0?ooo`3oool02P3oool000/0oooo00<0o`000?ooo`3oool0<P3oool00`00o`00oooo0?ooo`0i0?oo
o`<0o`00N@3oool00`3o0000oooo0?ooo`1c0?ooo`03003o003oool0oooo0380oooo00<0o`000?oo
o`3oool02P3oool000/0oooo00<0o`000?ooo`3oool0<P3oool00`00o`00oooo0?ooo`0l0?ooo`<0
o`00M`3oool20?l007<0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`0:0?ooo`00
2`3oool00`3o0000oooo0?ooo`0b0?ooo`03003o003oool0oooo03l0oooo0`3o001f0?ooo`030?l0
003oool0oooo0700oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`0:0?ooo`002`3o
ool00`3o0000oooo0?ooo`0b0?ooo`03003o003oool0oooo0480oooo0P3o001e0?ooo`80o`00L03o
ool00`00o`00oooo0?ooo`0b0?ooo`030?l0003oool0oooo00X0oooo000<0?ooo`030?l0003oool0
oooo0340oooo00<00?l00?ooo`3oool0A03oool30?l007@0oooo00<0o`000?ooo`3oool0K@3oool0
0`00o`00oooo0?ooo`0a0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`030?l0003oool0oooo
0380oooo00<00?l00?ooo`3oool0AP3oool30?l00780oooo0P3o001/0?ooo`03003o003oool0oooo
0380oooo00<0o`000?ooo`3oool02`3oool000`0oooo00<0o`000?ooo`3oool0<P3oool00`00o`00
oooo0?ooo`190?ooo`<0o`00L@3oool00`3o0000oooo0?ooo`1Y0?ooo`03003o003oool0oooo0380
oooo00<0o`000?ooo`3oool02`3oool000`0oooo00<0o`000?ooo`3oool0<P3oool00`00o`00oooo
0?ooo`1<0?ooo`80o`00L03oool20?l006T0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo
0?ooo`0;0?ooo`00303oool00`3o0000oooo0?ooo`0b0?ooo`03003o003oool0oooo04h0oooo0`3o
001_0?ooo`030?l0003oool0oooo06H0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?oo
o`0;0?ooo`00303oool00`3o0000oooo0?ooo`0b0?ooo`03003o003oool0oooo0540oooo0`3o001]
0?ooo`030?l0003oool0oooo06D0oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`0;
0?ooo`00303oool00`3o0000oooo0?ooo`0b0?ooo`03003o003oool0oooo05@0oooo0`3o001[0?oo
o`80o`00I03oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo00/0oooo000<0?ooo`03
0?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0EP3oool20?l006/0oooo00<0o`000?ooo`3o
ool0H@3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo00/0oooo000=0?ooo`030?l0
003oool0oooo0380oooo00<00?l00?ooo`3oool0F03oool30?l006T0oooo0P3o001Q0?ooo`03003o
003oool0oooo0380oooo00<0o`000?ooo`3oool0303oool000d0oooo00<0o`000?ooo`3oool0<P3o
ool00`00o`00oooo0?ooo`1K0?ooo`<0o`00J03oool00`3o0000oooo0?ooo`1N0?ooo`03003o003o
ool0oooo0380oooo00<0o`000?ooo`3oool0303oool000d0oooo00<0o`000?ooo`3oool0<P3oool0
0`00o`00oooo0?ooo`1N0?ooo`<0o`00IP3oool20?l005d0oooo00<00?l00?ooo`3oool0<`3oool0
0`3o0000oooo0?ooo`0<0?ooo`003@3oool00`3o0000oooo0?ooo`0c0?ooo`03003o003oool0oooo
0600oooo0P3o001V0?ooo`030?l0003oool0oooo05X0oooo00<00?l00?ooo`3oool0<`3oool00`3o
0000oooo0?ooo`0<0?ooo`003@3oool00`3o0000oooo0?ooo`0c0?ooo`03003o003oool0oooo0680
oooo0`3o001T0?ooo`80o`00FP3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo00`0
oooo000>0?ooo`030?l0003oool0oooo0380oooo00<00?l00?ooo`3oool0I@3oool30?l006<0oooo
00<0o`000?ooo`3oool0E`3oool00`00o`00oooo0?ooo`0b0?ooo`030?l0003oool0oooo00d0oooo
000>0?ooo`030?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0I`3oool30?l00640oooo00<0
o`000?ooo`3oool0E@3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo00d0oooo000>
0?ooo`030?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0JP3oool20?l00600oooo0P3o001E
0?ooo`03003o003oool0oooo03<0oooo00<0o`000?ooo`3oool03@3oool000h0oooo00<0o`000?oo
o`3oool0<`3oool00`00o`00oooo0?ooo`1/0?ooo`<0o`00G`3oool00`3o0000oooo0?ooo`1B0?oo
o`03003o003oool0oooo03<0oooo00<0o`000?ooo`3oool03@3oool000h0oooo00<0o`000?ooo`3o
ool0<`3oool00`00o`00oooo0?ooo`1_0?ooo`<0o`00G@3oool20?l00540oooo00<00?l00?ooo`3o
ool0=03oool00`3o0000oooo0?ooo`0=0?ooo`003`3oool00`3o0000oooo0?ooo`0c0?ooo`03003o
003oool0oooo0740oooo0`3o001L0?ooo`030?l0003oool0oooo04h0oooo00<00?l00?ooo`3oool0
<`3oool00`3o0000oooo0?ooo`0>0?ooo`003`3oool00`3o0000oooo0?ooo`0c0?ooo`03003o003o
ool0oooo07@0oooo0P3o001K0?ooo`80o`00CP3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003o
ool0oooo00h0oooo000?0?ooo`030?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0MP3oool3
0?l005X0oooo00<0o`000?ooo`3oool0B`3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0
oooo00h0oooo000?0?ooo`030?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0N@3oool30?l0
05P0oooo0P3o001:0?ooo`03003o003oool0oooo03@0oooo00<0o`000?ooo`3oool03P3oool000l0
oooo00<0o`000?ooo`3oool0=03oool00`00o`00oooo0?ooo`1k0?ooo`<0o`00E`3oool00`3o0000
oooo0?ooo`170?ooo`03003o003oool0oooo03@0oooo00<0o`000?ooo`3oool03P3oool00100oooo
00<0o`000?ooo`3oool0<`3oool00`00o`00oooo0?ooo`1n0?ooo`80o`00EP3oool00`3o0000oooo
0?ooo`150?ooo`03003o003oool0oooo03@0oooo00<0o`000?ooo`3oool03`3oool00100oooo00<0
o`000?ooo`3oool0=03oool00`00o`00oooo0?ooo`1o0?ooo`<0o`00E03oool20?l004D0oooo00<0
0?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0?0?ooo`00403oool00`3o0000oooo0?ooo`0d
0?ooo`03003o003oool0oooo0880oooo0`3o001C0?ooo`030?l0003oool0oooo0480oooo00<00?l0
0?ooo`3oool0<`3oool00`3o0000oooo0?ooo`0@0?ooo`004@3oool00`3o0000oooo0?ooo`0c0?oo
o`03003o003oool0oooo08D0oooo0`3o001A0?ooo`80o`00@@3oool00`00o`00oooo0?ooo`0d0?oo
o`030?l0003oool0oooo0100oooo000A0?ooo`030?l0003oool0oooo03@0oooo00<00?l00?ooo`3o
ool0Q`3oool20?l00540oooo00<0o`000?ooo`3oool0?P3oool00`00o`00oooo0?ooo`0d0?ooo`03
0?l0003oool0oooo0100oooo000A0?ooo`030?l0003oool0oooo03@0oooo00<00?l00?ooo`3oool0
R@3oool30?l004l0oooo0P3o000m0?ooo`03003o003oool0oooo03D0oooo00<0o`000?ooo`3oool0
403oool00140oooo00<0o`000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`2;0?ooo`<0o`00CP3o
ool00`3o0000oooo0?ooo`0j0?ooo`03003o003oool0oooo03D0oooo00<0o`000?ooo`3oool0403o
ool00140oooo00<0o`000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`2>0?ooo`<0o`00C03oool2
0?l003X0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0A0?ooo`004P3oool00`3o
0000oooo0?ooo`0d0?ooo`03003o003oool0oooo0940oooo0P3o001<0?ooo`030?l0003oool0oooo
03L0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0A0?ooo`004P3oool00`3o0000
oooo0?ooo`0e0?ooo`03003o003oool0oooo0980oooo0`3o001:0?ooo`030?l0003oool0oooo03D0
oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`0B0?ooo`004`3oool00`3o0000oooo
0?ooo`0d0?ooo`03003o003oool0oooo09D0oooo0`3o00180?ooo`80o`00=03oool00`00o`00oooo
0?ooo`0e0?ooo`030?l0003oool0oooo0180oooo000C0?ooo`030?l0003oool0oooo03D0oooo00<0
0?l00?ooo`3oool0U`3oool30?l004L0oooo00<0o`000?ooo`3oool0<@3oool00`00o`00oooo0?oo
o`0e0?ooo`030?l0003oool0oooo0180oooo000D0?ooo`030?l0003oool0oooo03@0oooo00<00?l0
0?ooo`3oool0VP3oool30?l004D0oooo0P3o000`0?ooo`03003o003oool0oooo03D0oooo00<0o`00
0?ooo`3oool04`3oool001@0oooo00<0o`000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`2L0?oo
o`80o`00A@3oool00`3o0000oooo0?ooo`0]0?ooo`03003o003oool0oooo03D0oooo00<0o`000?oo
o`3oool04`3oool001@0oooo00<0o`000?ooo`3oool0=@3oool00`00o`00oooo0?ooo`2N0?ooo`<0
o`00@`3oool20?l002d0oooo00<00?l00?ooo`3oool0=@3oool00`3o0000oooo0?ooo`0C0?ooo`00
503oool00`3o0000oooo0?ooo`0e0?ooo`03003o003oool0oooo0:40oooo0`3o00120?ooo`030?l0
003oool0oooo02X0oooo00<00?l00?ooo`3oool0=@3oool00`3o0000oooo0?ooo`0C0?ooo`00503o
ool00`3o0000oooo0?ooo`0f0?ooo`03003o003oool0oooo0:<0oooo0`3o00100?ooo`80o`00:@3o
ool00`00o`00oooo0?ooo`0f0?ooo`030?l0003oool0oooo01<0oooo000E0?ooo`030?l0003oool0
oooo03D0oooo00<00?l00?ooo`3oool0YP3oool20?l00400oooo00<0o`000?ooo`3oool09@3oool0
0`00o`00oooo0?ooo`0f0?ooo`030?l0003oool0oooo01@0oooo000E0?ooo`030?l0003oool0oooo
03H0oooo00<00?l00?ooo`3oool0Y`3oool30?l003h0oooo00<0o`000?ooo`3oool0903oool00`00
o`00oooo0?ooo`0f0?ooo`030?l0003oool0oooo01@0oooo000E0?ooo`030?l0003oool0oooo03H0
oooo00<00?l00?ooo`3oool0ZP3oool30?l003`0oooo0P3o000S0?ooo`03003o003oool0oooo03H0
oooo00<0o`000?ooo`3oool05@3oool001H0oooo00<0o`000?ooo`3oool0=P3oool00`00o`00oooo
0?ooo`2/0?ooo`<0o`00>`3oool00`3o0000oooo0?ooo`0P0?ooo`03003o003oool0oooo03H0oooo
00<0o`000?ooo`3oool05@3oool001H0oooo00<0o`000?ooo`3oool0=`3oool00`00o`00oooo0?oo
o`2^0?ooo`80o`00>P3oool20?l001l0oooo00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?oo
o`0E0?ooo`005`3oool00`3o0000oooo0?ooo`0f0?ooo`03003o003oool0oooo0;00oooo0`3o000i
0?ooo`030?l0003oool0oooo01/0oooo00<00?l00?ooo`3oool0=`3oool00`3o0000oooo0?ooo`0F
0?ooo`005`3oool00`3o0000oooo0?ooo`0g0?ooo`03003o003oool0oooo0;80oooo0`3o000g0?oo
o`80o`006`3oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01H0oooo000G0?ooo`03
0?l0003oool0oooo03L0oooo00<00?l00?ooo`3oool0]@3oool30?l003H0oooo00<0o`000?ooo`3o
ool05`3oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01L0oooo000H0?ooo`030?l0
003oool0oooo03L0oooo00<00?l00?ooo`3oool0]`3oool20?l003D0oooo0P3o000G0?ooo`03003o
003oool0oooo03L0oooo00<0o`000?ooo`3oool05`3oool001P0oooo00<0o`000?ooo`3oool0>03o
ool00`00o`00oooo0?ooo`2h0?ooo`<0o`00=03oool00`3o0000oooo0?ooo`0C0?ooo`03003o003o
ool0oooo03L0oooo00<0o`000?ooo`3oool0603oool001T0oooo00<0o`000?ooo`3oool0=`3oool0
0`00o`00oooo0?ooo`2k0?ooo`<0o`00<P3oool00`3o0000oooo0?ooo`0B0?ooo`03003o003oool0
oooo03L0oooo00<0o`000?ooo`3oool0603oool001T0oooo00<0o`000?ooo`3oool0=`3oool00`00
o`00oooo0?ooo`2n0?ooo`<0o`00<03oool20?l00180oooo00<00?l00?ooo`3oool0=`3oool00`3o
0000oooo0?ooo`0H0?ooo`006P3oool00`3o0000oooo0?ooo`0f0?ooo`03003o003oool0oooo0<40
oooo0P3o000`0?ooo`030?l0003oool0oooo00h0oooo00<00?l00?ooo`3oool0=`3oool00`3o0000
oooo0?ooo`0I0?ooo`006P3oool00`3o0000oooo0?ooo`0g0?ooo`03003o003oool0oooo0<80oooo
0`3o000^0?ooo`80o`003P3oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01T0oooo
000J0?ooo`030?l0003oool0oooo03P0oooo00<00?l00?ooo`3oool0a03oool30?l002d0oooo00<0
o`000?ooo`3oool02P3oool00`00o`00oooo0?ooo`0g0?ooo`030?l0003oool0oooo01X0oooo000K
0?ooo`030?l0003oool0oooo03L0oooo00<00?l00?ooo`3oool0a`3oool30?l002/0oooo0P3o0009
0?ooo`03003o003oool0oooo03P0oooo00<0o`000?ooo`3oool06P3oool001/0oooo00<0o`000?oo
o`3oool0>03oool00`00o`00oooo0?ooo`390?ooo`80o`00:`3oool00`3o0000oooo0?ooo`060?oo
o`03003o003oool0oooo03L0oooo00<0o`000?ooo`3oool06`3oool001`0oooo00<0o`000?ooo`3o
ool0>03oool00`00o`00oooo0?ooo`3:0?ooo`<0o`00:@3oool20?l000D0oooo00<00?l00?ooo`3o
ool0>03oool00`3o0000oooo0?ooo`0K0?ooo`00703oool00`3o0000oooo0?ooo`0i0?ooo`03003o
003oool0oooo0<`0oooo0`3o000X0?ooo`050?l0003oool0oooo0?ooo`00o`00>`3oool00`3o0000
oooo0?ooo`0K0?ooo`007@3oool00`3o0000oooo0?ooo`0i0?ooo`03003o003oool0oooo0<h0oooo
0`3o000V0?ooo`030?l0003oool00?l003/0oooo00<0o`000?ooo`3oool0703oool001d0oooo00<0
o`000?ooo`3oool0>@3oool00`00o`00oooo0?ooo`3A0?ooo`80o`009@3oool20?l003/0oooo00<0
o`000?ooo`3oool0703oool001d0oooo00<0o`000?ooo`3oool0>P3oool00`00o`00oooo0?ooo`3B
0?ooo`<0o`008P3oool00`00o`00oooo0?l0000i0?ooo`030?l0003oool0oooo01d0oooo000N0?oo
o`030?l0003oool0oooo03X0oooo00<00?l00?ooo`3oool0e03oool30?l001h0oooo00@00?l00?oo
o`3oool0oooo0P3o000g0?ooo`030?l0003oool0oooo01d0oooo000N0?ooo`030?l0003oool0oooo
03X0oooo00<00?l00?ooo`3oool0e`3oool30?l001X0oooo00<00?l00?ooo`3oool0103oool00`3o
0000oooo0?ooo`0c0?ooo`030?l0003oool0oooo01h0oooo000O0?ooo`030?l0003oool0oooo03X0
oooo00<00?l00?ooo`3oool0f@3oool20?l001P0oooo00<00?l00?ooo`3oool01@3oool20?l003<0
oooo00<0o`000?ooo`3oool07P3oool001l0oooo00<0o`000?ooo`3oool0>`3oool00`00o`00oooo
0?ooo`3J0?ooo`<0o`00503oool00`00o`00oooo0?ooo`080?ooo`030?l0003oool0oooo0300oooo
00<0o`000?ooo`3oool07P3oool00200oooo00<0o`000?ooo`3oool0>`3oool00`00o`00oooo0?oo
o`3L0?ooo`<0o`00403oool00`00o`00oooo0?ooo`0:0?ooo`80o`00;`3oool00`3o0000oooo0?oo
o`0O0?ooo`00803oool00`3o0000oooo0?ooo`0k0?ooo`03003o003oool0oooo0=l0oooo0`3o000=
0?ooo`03003o003oool0oooo00`0oooo00<0o`000?ooo`3oool0;03oool00`3o0000oooo0?ooo`0O
0?ooo`008@3oool00`3o0000oooo0?ooo`0k0?ooo`03003o003oool0oooo0>40oooo0P3o000:0?oo
o`03003o003oool0oooo00h0oooo00<0o`000?ooo`3oool0:P3oool00`3o0000oooo0?ooo`0P0?oo
o`008P3oool00`3o0000oooo0?ooo`0k0?ooo`03003o003oool0oooo0>80oooo0`3o00060?ooo`03
003o003oool0oooo0100oooo0P3o000Y0?ooo`030?l0003oool0oooo0240oooo000R0?ooo`030?l0
003oool0oooo03`0oooo00<00?l00?ooo`3oool0i03oool30?l00080oooo00<00?l00?ooo`3oool0
4`3oool00`3o0000oooo0?ooo`0U0?ooo`030?l0003oool0oooo0280oooo000S0?ooo`030?l0003o
ool0oooo03`0oooo00<00?l00?ooo`3oool0iP3oool30?l001H0oooo0P3o000U0?ooo`030?l0003o
ool0oooo0280oooo000S0?ooo`030?l0003oool0oooo03d0oooo00<00?l00?ooo`3oool0i@3oool0
0`00o`00oooo0?ooo`020?l001H0oooo00<0o`000?ooo`3oool08@3oool00`3o0000oooo0?ooo`0S
0?ooo`00903oool00`3o0000oooo0?ooo`0l0?ooo`03003o003oool0oooo0>D0oooo00<00?l00?oo
o`3oool00P3oool30?l001@0oooo0P3o000Q0?ooo`030?l0003oool0oooo02<0oooo000T0?ooo`03
0?l0003oool0oooo03d0oooo00<00?l00?ooo`3oool0h`3oool00`00o`00oooo0?ooo`060?ooo`<0
o`004`3oool00`3o0000oooo0?ooo`0N0?ooo`030?l0003oool0oooo02<0oooo000U0?ooo`030?l0
003oool0oooo03d0oooo00<00?l00?ooo`3oool0h@3oool00`00o`00oooo0?ooo`0:0?ooo`<0o`00
4@3oool20?l001d0oooo00<0o`000?ooo`3oool0903oool002D0oooo00<0o`000?ooo`3oool0?P3o
ool00`00o`00oooo0?ooo`3O0?ooo`03003o003oool0oooo00h0oooo0P3o000A0?ooo`030?l0003o
ool0oooo01X0oooo00<0o`000?ooo`3oool0903oool002H0oooo00<0o`000?ooo`3oool0?P3oool0
0`00o`00oooo0?ooo`3M0?ooo`03003o003oool0oooo0140oooo0`3o000?0?ooo`030?l0003oool0
oooo01P0oooo00<0o`000?ooo`3oool09@3oool002L0oooo00<0o`000?ooo`3oool0?P3oool00`00
o`00oooo0?ooo`3K0?ooo`03003o003oool0oooo01D0oooo0`3o000=0?ooo`80o`005`3oool00`3o
0000oooo0?ooo`0V0?ooo`009`3oool00`3o0000oooo0?ooo`0n0?ooo`03003o003oool0oooo0=/0
oooo00<00?l00?ooo`3oool0603oool30?l000`0oooo00<0o`000?ooo`3oool04`3oool00`3o0000
oooo0?ooo`0W0?ooo`00:03oool00`3o0000oooo0?ooo`0n0?ooo`03003o003oool0oooo0=T0oooo
00<00?l00?ooo`3oool0703oool20?l000/0oooo0P3o000C0?ooo`030?l0003oool0oooo02L0oooo
000X0?ooo`030?l0003oool0oooo03l0oooo0P00o`3H0?ooo`03003o003oool0oooo01l0oooo0`3o
000:0?ooo`030?l0003oool0oooo00l0oooo00<0o`000?ooo`3oool0:03oool002T0oooo00<0o`00
0?ooo`3oool0@03oool00`00o`00oooo0?ooo`3D0?ooo`03003o003oool0oooo02<0oooo0`3o0008
0?ooo`80o`003`3oool00`3o0000oooo0?ooo`0X0?ooo`00:P3oool00`3o0000oooo0?ooo`100?oo
o`03003o003oool0oooo0=80oooo00<00?l00?ooo`3oool09`3oool30?l000L0oooo00<0o`000?oo
o`3oool02`3oool00`3o0000oooo0?ooo`0Y0?ooo`00:`3oool00`3o0000oooo0?ooo`100?ooo`03
003o003oool0oooo0<l0oooo0P00o`0]0?ooo`80o`001P3oool20?l000030?ooo`3o0000oooo00L0
oooo00<0o`000?ooo`3oool0:P3oool002/0oooo00<0o`000?ooo`3oool0@@3oool00`00o`00oooo
0?ooo`3=0?ooo`03003o003oool0oooo02l0oooo0`3o00050?ooo`<0o`001P3oool00`3o0000oooo
0?ooo`0[0?ooo`00;03oool00`3o0000oooo0?ooo`110?ooo`03003o003oool0oooo0</0oooo00<0
0?l00?ooo`3oool0<`3oool30?l00080oooo103o00050?ooo`030?l0003oool0oooo02/0oooo000/
0?ooo`030?l0003oool0oooo0480oooo00<00?l00?ooo`3oool0b@3oool00`00o`00oooo0?ooo`0g
0?ooo`L0o`00103oool00`3o0000oooo0?ooo`0[0?ooo`00;@3oool00`3o0000oooo0?ooo`120?oo
o`03003o003oool0oooo0<L0oooo00<00?l00?ooo`3oool0>@3oool70?l00080oooo00<0o`000?oo
o`3oool0;03oool002d0oooo00<0o`000?ooo`3oool0@`3oool2003o0<D0oooo0P00o`0l0?ooo`P0
o`0000<0oooo0?l0003oool0;@3oool002h0oooo00<0o`000?ooo`3oool0A03oool00`00o`00oooo
0?ooo`310?ooo`03003o003oool0oooo0400oooo1@3o000_0?ooo`00;`3oool00`3o0000oooo0?oo
o`140?ooo`03003o003oool0oooo0;l0oooo00<00?l00?ooo`3oool0A03oool00`3o0000oooo0?oo
o`0^0?ooo`00;`3oool00`3o0000oooo0?ooo`150?ooo`03003o003oool0oooo0;d0oooo00<00?l0
0?ooo`3oool0A03oool00`3o0000oooo0?ooo`0_0?ooo`00<03oool00`3o0000oooo0?ooo`150?oo
o`03003o003oool0oooo0;/0oooo00<00?l00?ooo`3oool0A@3oool00`3o0000oooo0?ooo`0_0?oo
o`00<@3oool00`3o0000oooo0?ooo`150?ooo`800?l0^P3oool00`00o`00oooo0?ooo`150?ooo`03
0?l0003oool0oooo0300oooo000b0?ooo`030?l0003oool0oooo04H0oooo00<00?l00?ooo`3oool0
]@3oool2003o04L0oooo00<0o`000?ooo`3oool0<@3oool00380oooo00<0o`000?ooo`3oool0A`3o
ool00`00o`00oooo0?ooo`2c0?ooo`03003o003oool0oooo04H0oooo00<0o`000?ooo`3oool0<P3o
ool003<0oooo00<0o`000?ooo`3oool0A`3oool2003o0;40oooo0P00o`190?ooo`030?l0003oool0
oooo0380oooo000d0?ooo`030?l0003oool0oooo04P0oooo00<00?l00?ooo`3oool0[@3oool00`00
o`00oooo0?ooo`180?ooo`030?l0003oool0oooo03<0oooo000e0?ooo`030?l0003oool0oooo04P0
oooo0P00o`2[0?ooo`800?l0BP3oool00`3o0000oooo0?ooo`0d0?ooo`00=P3oool00`3o0000oooo
0?ooo`190?ooo`03003o003oool0oooo0:L0oooo00<00?l00?ooo`3oool0B@3oool00`3o0000oooo
0?ooo`0e0?ooo`00=`3oool00`3o0000oooo0?ooo`190?ooo`800?l0YP3oool00`00o`00oooo0?oo
o`190?ooo`030?l0003oool0oooo03H0oooo000g0?ooo`030?l0003oool0oooo04/0oooo00<00?l0
0?ooo`3oool0XP3oool00`00o`00oooo0?ooo`190?ooo`030?l0003oool0oooo03L0oooo000h0?oo
o`030?l0003oool0oooo04/0oooo00<00?l00?ooo`3oool0W`3oool2003o04`0oooo00<0o`000?oo
o`3oool0=`3oool003T0oooo00<0o`000?ooo`3oool0B`3oool00`00o`00oooo0?ooo`2M0?ooo`03
003o003oool0oooo04/0oooo00<0o`000?ooo`3oool0>03oool003X0oooo00<0o`000?ooo`3oool0
B`3oool4003o09L0oooo1000o`1=0?ooo`030?l0003oool0oooo03T0oooo000j0?ooo`030?l0003o
ool0oooo04l0oooo00<00?l00?ooo`3oool0T`3oool00`00o`00oooo0?ooo`1>0?ooo`030?l0003o
ool0oooo03X0oooo000k0?ooo`030?l0003oool0oooo04l0oooo0P00o`2A0?ooo`800?l0D@3oool0
0`3o0000oooo0?ooo`0j0?ooo`00?03oool00`3o0000oooo0?ooo`1@0?ooo`800?l0S@3oool2003o
0580oooo00<0o`000?ooo`3oool0>`3oool003d0oooo00<0o`000?ooo`3oool0D@3oool00`00o`00
oooo0?ooo`290?ooo`03003o003oool0oooo0540oooo00<0o`000?ooo`3oool0?03oool003h0oooo
00<0o`000?ooo`3oool0D@3oool2003o08L0oooo0P00o`1C0?ooo`030?l0003oool0oooo03d0oooo
000o0?ooo`030?l0003oool0oooo0580oooo0P00o`230?ooo`800?l0E03oool00`3o0000oooo0?oo
o`0n0?ooo`00@03oool00`3o0000oooo0?ooo`1C0?ooo`<00?l0O@3oool3003o05D0oooo00<0o`00
0?ooo`3oool0?`3oool00440oooo00<0o`000?ooo`3oool0E@3oool2003o07T0oooo0P00o`1G0?oo
o`030?l0003oool0oooo0400oooo00120?ooo`030?l0003oool0oooo05H0oooo0P00o`1e0?ooo`80
0?l0F03oool00`3o0000oooo0?ooo`110?ooo`00@`3oool00`3o0000oooo0?ooo`1G0?ooo`<00?l0
K`3oool3003o05T0oooo00<0o`000?ooo`3oool0@P3oool004@0oooo00<0o`000?ooo`3oool0F@3o
ool2003o06/0oooo0P00o`1K0?ooo`030?l0003oool0oooo04<0oooo00150?ooo`030?l0003oool0
oooo05X0oooo0`00o`1U0?ooo`<00?l0G03oool00`3o0000oooo0?ooo`140?ooo`00AP3oool00`3o
0000oooo0?ooo`1L0?ooo`800?l0H@3oool2003o05h0oooo00<0o`000?ooo`3oool0A@3oool004L0
oooo00<0o`000?ooo`3oool0G@3oool4003o05T0oooo1000o`1O0?ooo`030?l0003oool0oooo04H0
oooo00180?ooo`030?l0003oool0oooo0600oooo1000o`1A0?ooo`@00?l0HP3oool00`3o0000oooo
0?ooo`170?ooo`00B@3oool00`3o0000oooo0?ooo`1S0?ooo`<00?l0B`3oool3003o06D0oooo00<0
o`000?ooo`3oool0B03oool004X0oooo00<0o`000?ooo`3oool0I@3oool4003o04<0oooo1000o`1W
0?ooo`030?l0003oool0oooo04T0oooo001;0?ooo`030?l0003oool0oooo06P0oooo1P00o`0g0?oo
o`H00?l0JP3oool00`3o0000oooo0?ooo`1:0?ooo`00C03oool00`3o0000oooo0?ooo`1]0?ooo`D0
0?l0;@3oool5003o06l0oooo00<0o`000?ooo`3oool0B`3oool004d0oooo00<0o`000?ooo`3oool0
L@3oool:003o01T0oooo2P00o`1c0?ooo`030?l0003oool0oooo04`0oooo001>0?ooo`030?l0003o
ool0oooo07X0oooo6@00o`1l0?ooo`030?l0003oool0oooo04d0oooo001?0?ooo`030?l0003oool0
oooo0?l0oooo3P3oool00`3o0000oooo0?ooo`1>0?ooo`00D03oool00`3o0000oooo0?ooo`3o0?oo
o``0oooo00<0o`000?ooo`3oool0C`3oool00540oooo00<0o`000?ooo`3oool0o`3oool:0?ooo`03
0?l0003oool0oooo0500oooo001B0?ooo`80o`00o`3oool80?ooo`80o`00D`3oool005@0oooo00<0
o`000?ooo`3oool0o`3oool40?ooo`030?l0003oool0oooo05<0oooo001E0?ooo`030?l0003oool0
oooo0?l0oooo0P3oool00`3o0000oooo0?ooo`1D0?ooo`00EP3oool20?l00?l0oooo0P3o001G0?oo
o`00F03oool00`3o0000oooo0?ooo`3k0?ooo`030?l0003oool0oooo05L0oooo001I0?ooo`030?l0
003oool0oooo0?T0oooo00<0o`000?ooo`3oool0F03oool005X0oooo00<0o`000?ooo`3oool0m`3o
ool00`3o0000oooo0?ooo`1I0?ooo`00F`3oool00`3o0000oooo0?ooo`3e0?ooo`030?l0003oool0
oooo05X0oooo001L0?ooo`80o`00l`3oool20?l005d0oooo001N0?ooo`030?l0003oool0oooo0>l0
oooo00<0o`000?ooo`3oool0G@3oool005l0oooo00<0o`000?ooo`3oool0k@3oool00`3o0000oooo
0?ooo`1N0?ooo`00H03oool20?l00>/0oooo0P3o001Q0?ooo`00HP3oool00`3o0000oooo0?ooo`3W
0?ooo`030?l0003oool0oooo0640oooo001S0?ooo`030?l0003oool0oooo0>@0oooo0P3o001T0?oo
o`00I03oool30?l00>40oooo0P3o001V0?ooo`00I`3oool20?l00=d0oooo0P3o001X0?ooo`00J@3o
ool00`3o0000oooo0?ooo`3I0?ooo`030?l0003oool0oooo06P0oooo001Z0?ooo`030?l0003oool0
oooo0=L0oooo00<0o`000?ooo`3oool0J@3oool006/0oooo0P3o003E0?ooo`80o`00K03oool006d0
oooo0P3o003A0?ooo`80o`00KP3oool006l0oooo00<0o`000?ooo`3oool0c@3oool00`3o0000oooo
0?ooo`1^0?ooo`00L03oool20?l00</0oooo0P3o001a0?ooo`00LP3oool20?l00<L0oooo0P3o001c
0?ooo`00M03oool20?l00<<0oooo0P3o001e0?ooo`00MP3oool20?l00;l0oooo0P3o001g0?ooo`00
N03oool00`3o0000oooo0?ooo`2k0?ooo`030?l0003oool0oooo07L0oooo001i0?ooo`80o`00^@3o
ool20?l007X0oooo001k0?ooo`80o`00]@3oool20?l007`0oooo001m0?ooo`80o`00/@3oool20?l0
07h0oooo001o0?ooo`<0o`00Z`3oool30?l00800oooo00220?ooo`80o`00Y`3oool20?l008<0oooo
00240?ooo`80o`00X`3oool20?l008D0oooo00260?ooo`<0o`00W@3oool30?l008L0oooo00290?oo
o`80o`00V@3oool20?l008X0oooo002;0?ooo`80o`00U@3oool20?l008`0oooo002=0?ooo`<0o`00
S`3oool30?l008h0oooo002@0?ooo`80o`00R`3oool20?l00940oooo002B0?ooo`<0o`00Q@3oool3
0?l009<0oooo002E0?ooo`@0o`00O03oool50?l009H0oooo002I0?ooo`<0o`00M`3oool20?l009/0
oooo002L0?ooo`80o`00L`3oool20?l009d0oooo002N0?ooo`D0o`00J@3oool50?l009l0oooo002S
0?ooo`D0o`00H@3oool30?l00:@0oooo002X0?ooo`<0o`00F@3oool50?l00:L0oooo002[0?ooo`D0
o`00C`3oool50?l00:`0oooo002`0?ooo`D0o`00A@3oool50?l00;40oooo002e0?ooo`L0o`00=`3o
ool70?l00;H0oooo002l0?ooo`/0o`008@3oool;0?l00;d0oooo00370?ooob40o`00b03oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001
\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.204369, -0.704368, 0.0049028, \
0.0049028}}]
}, Closed]],

Cell["\<\
Since the outgoing velocities may appear anywhere on the outer circle, the \
scattering and recoil angles for this exothermic reaction are not restricted. \
\
\>", "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Elastic colinear collision for unequal masses", "Subsection"],

Cell["\<\
We now consider an elastic collision between unequal masses, such that m2/m1 \
= mratio.  For simplicity we assume that the target is at rest in the lab.  \
\
\>", "Text"],

Cell[BoxData[
    \(\(values = {m1 \[Rule] 1, m2 \[Rule] mratio\ m1, m3 \[Rule] m1, 
          qvalue \[Rule] 0, p1r \[Rule] 1, theta1 \[Rule] 0, p2r \[Rule] 0, 
          theta2 \[Rule] \[Pi]};\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol\  = \ \(\(\(generalSolution[p3r, p4x, 
                  p4y]\)\[LeftDoubleBracket]1\[RightDoubleBracket] //. 
              values\  // Simplify\) // PowerExpand\) // Simplify\)], "Input"],

Cell[BoxData[
    \({p4x \[Rule] \(-\(\(\(-1\) - 2\ mratio + 
                Cos[2\ theta3] + \@2\ Cos[
                    theta3]\ \@\(\(-1\) + 2\ mratio\^2 + Cos[2\ theta3]\)\)\/\
\(2\ \((1 + mratio)\)\)\)\), 
      p4y \[Rule] \(-\(\(\((2\ Cos[
                        theta3] + \@2\ \@\(\(-1\) + 2\ mratio\^2 + Cos[2\ \
theta3]\))\)\ Sin[theta3]\)\/\(2\ \((1 + mratio)\)\)\)\), 
      p3r \[Rule] \(2\ Cos[theta3] + \@2\ \@\(\(-1\) + 2\ mratio\^2 + Cos[2\ \
theta3]\)\)\/\(2 + 2\ mratio\), m4 \[Rule] mratio}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell["Projectile heavier than target", "Subsubsection"],

Cell[TextData[{
  "Suppose that the projectile is heavier than the target, such that ",
  StyleBox["mratio<1", "InlineInput"],
  ".  We then find that there is a maximum scattering angle, which is derived \
below.  Plots are made for the specific case ",
  StyleBox["mratio\[Rule]0.5", "InlineInput"],
  "."
}], "Text"],

Cell[BoxData[
    \(calcCollision[sol, Join[values, {mratio \[Rule] 0.5}], 
      20\ Degree]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawCollision @@ LabMomenta;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .32735 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.547912 0.524102 0.163675 0.524102 [
[ 0 0 0 0 ]
[ 1 .32735 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .32735 L
0 .32735 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .16367 m
.54791 .16367 L
s
% Start of user PostScript
-1. 0. 0. 0. MAarrowhead1
% End of user PostScript
.54791 .16367 m
.54791 .16367 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.54791 .16367 m
.97619 .31956 L
s
% Start of user PostScript
0. 0. 0.81716593 0.29742408 MAarrowhead1
% End of user PostScript
.54791 .16367 m
.64374 .00779 L
s
% Start of user PostScript
0. 0. 0.18283407 -0.29742408 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{287.813, 94.1875},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006_0000SB000`400?l00000o`00003oo`3ooon`
0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00o`3ooon`0?ooo`00o`3ooolE0?ooo`030000003o
ool0oooo09P0oooo003o0?oooa@0oooo0P00002J0?ooo`00o`3ooolC0?ooo`800000V`3oool00?l0
oooo4P3oool3000009/0oooo003o0?oooa40oooo1000002K0?ooo`00o`3oool@0?ooo`D00000V`3o
ool00?l0oooo4@3oool3000009`0oooo003o0?oooa40oooo00<000000?ooo`000000W03oool00?l0
oooo403oool00`000000oooo0?ooo`2M0?ooo`00o`3oool?0?ooo`030000003oool0oooo09h0oooo
003o0?ooo`l0oooo00<000000?ooo`3oool0WP3oool00?l0oooo3P3oool00`000000oooo0?ooo`2O
0?ooo`00o`3oool>0?ooo`030000003oool0oooo09l0oooo003o0?ooo`d0oooo00<000000?ooo`3o
ool0X03oool00?l0oooo303oool00`000000oooo0?ooo`2Q0?ooo`00o`3oool<0?ooo`030000003o
ool0oooo0:40oooo003o0?ooo`/0oooo00<000000?ooo`3oool0XP3oool00?l0oooo2P3oool00`00
0000oooo0?ooo`2S0?ooo`00o`3oool:0?ooo`030000003oool0oooo0:<0oooo003o0?ooo`T0oooo
00<000000?ooo`3oool0Y03oool00?l0oooo2@3oool00`000000oooo0?ooo`2T0?ooo`00o`3oool8
0?ooo`030000003oool0oooo0:D0oooo003o0?ooo`L0oooo00<000000?ooo`3oool0YP3oool00?l0
oooo1`3oool00`000000oooo0?ooo`2V0?ooo`00o`3oool60?ooo`030000003oool0oooo0:L0oooo
003o0?ooo`D0oooo00<000000?ooo`3oool0Z03oool00?l0oooo1@3oool00`000000oooo0?ooo`2X
0?ooo`00o`3oool40?ooo`030000003oool0oooo0:T0oooo003o0?ooo`@0oooo00<000000?ooo`3o
ool0Z@3oool00?l0oooo0`3oool00`000000oooo0?ooo`2Z0?ooo`00o`3oool20?ooo`030000003o
ool0oooo0:/0oooo003o0?ooo`80oooo00<000000?ooo`3oool0Z`3oool00?l0oooo0@3oool00`00
0000oooo0?ooo`2/0?ooo`00o`3oool00`000000oooo0?ooo`2]0?ooo`00o`3oool00`000000oooo
0?ooo`2]0?ooo`00oP3oool00`000000oooo0?ooo`2^0?ooo`00oP3oool00`000000oooo0?ooo`2^
0?ooo`00o@3oool00`000000oooo0?ooo`2_0?ooo`00o03oool00`000000oooo0?ooo`2`0?ooo`00
o03oool00`000000oooo0?ooo`2`0?ooo`00n`3oool00`000000oooo0?ooo`2a0?ooo`00n`3oool0
0`000000oooo0?ooo`2a0?ooo`00nP3oool00`000000oooo0?ooo`2b0?ooo`00n@3oool00`000000
oooo0?ooo`2c0?ooo`00n@3oool00`000000oooo0?ooo`2c0?ooo`00n03oool00`000000oooo0?oo
o`2d0?ooo`00m`3oool00`000000oooo0?ooo`2e0?ooo`00m`3oool00`000000oooo0?ooo`2e0?oo
o`00mP3oool00`000000oooo0?ooo`2f0?ooo`00mP3oool00`000000oooo0?ooo`2f0?ooo`00m@3o
ool00`000000oooo0?ooo`2g0?ooo`00m03oool00`000000oooo0?ooo`2h0?ooo`00m03oool00`00
0000oooo0?ooo`2h0?ooo`00l`3oool00`000000oooo0?ooo`2i0?ooo`00lP3oool00`000000oooo
0?ooo`2j0?ooo`00lP3oool00`000000oooo0?ooo`2j0?ooo`00l@3oool00`000000oooo0?ooo`2k
0?ooo`00l@3oool00`000000oooo0?ooo`2k0?ooo`00l03oool00`000000oooo0?ooo`2l0?ooo`00
k`3oool00`000000oooo0?ooo`2m0?ooo`00k`3oool00`000000oooo0?ooo`2m0?ooo`00kP3oool0
0`000000oooo0?ooo`2n0?ooo`00k@3oool00`000000oooo0?ooo`2o0?ooo`00k@3oool00`000000
oooo0?ooo`2o0?ooo`00i03oool2000000H0oooo00<000000?ooo`3oool0`03oool00>@0oooo1P00
00020?ooo`030000003oool0oooo0<00oooo000;0?ooon800000`P3oool00>@0oooo1P0000030?oo
o`<00000_`3oool00>@0oooo0P00000:0?ooo`800000_@3oool00?80oooo0`00002j0?ooo`00m@3o
ool300000;L0oooo003h0?ooo`<00000]03oool00?/0oooo0P00002b0?ooo`00o@3oool300000:l0
oooo003o0?ooo`40oooo0`00002/0?ooo`00o`3oool40?ooo`<00000Z@3oool00?l0oooo1`3oool2
00000:L0oooo003o0?ooo`T0oooo0`00002T0?ooo`00o`3oool<0?ooo`<00000X@3oool00?l0oooo
3`3oool3000009h0oooo003o0?oooa80oooo0P00002L0?ooo`00o`3ooolD0?ooo`<00000V@3oool0
0?l0oooo5`3oool3000009H0oooo003o0?oooaX0oooo0`00002C0?ooo`00o`3ooolM0?ooo`800000
T@3oool00?l0oooo7`3oool3000008h0oooo003o0?ooob80oooo0`00002;0?ooo`00o`3ooolU0?oo
o`<00000R03oool00?l0oooo:03oool2000008H0oooo003o0?ooobX0oooo0`0000230?ooo`00o`3o
ool]0?ooo`<00000P03oool00?l0oooo<03oool3000007d0oooo003o0?oooc<0oooo0P00001k0?oo
o`00o`3ooole0?ooo`<00000N03oool00?l0oooo>03oool3000007D0oooo003o0?oooc/0oooo0`00
001b0?ooo`00o`3oooln0?ooo`800000L03oool00?l0oooo@03oool3000006d0oooo003o0?oood<0
oooo0`00001Z0?ooo`00o`3ooom60?ooo`800000J03oool00?l0ooooB03oool3000006D0oooo003o
0?oood/0oooo0`00001R0?ooo`00o`3ooom>0?ooo`<00000G`3oool00?l0ooooD@3oool2000005d0
oooo003o0?oooe<0oooo0`00001J0?ooo`00o`3ooomF0?ooo`<00000E`3oool00?l0ooooF@3oool3
000005@0oooo003o0?oooe`0oooo0P00001B0?ooo`00o`3ooomN0?ooo`<00000C`3oool00?l0oooo
H@3oool3000004`0oooo003o0?ooof@0oooo0`0000190?ooo`00o`3ooomW0?ooo`800000A`3oool0
0?l0ooooJ@3oool3000004@0oooo003o0?ooof`0oooo0`0000110?ooo`00o`3ooom_0?ooo`<00000
?P3oool00?l0ooooLP3oool2000003`0oooo003o0?ooog@0oooo0`00000i0?ooo`00o`3ooomg0?oo
o`<00000=P3oool00?l0ooooNP3oool3000003<0oooo003o0?ooogd0oooo0P00000a0?ooo`00o`3o
oomo0?ooo`<00000;P3oool00?l0ooooPP3oool3000002/0oooo003o0?ooohD0oooo0`00000X0?oo
o`00o`3ooon80?ooo`8000009P3oool00?l0ooooRP3oool3000002<0oooo003o0?ooohd0oooo0`00
000P0?ooo`00o`3ooon@0?ooo`<000007@3oool00?l0ooooT`3oool2000001/0oooo003o0?oooiD0
oooo0`00000H0?ooo`00o`3ooonH0?ooo`<000000P3oool00`000000oooo0?ooo`0@0?ooo`00o`3o
oonK0?ooo`D00000403oool00?l0ooooW03oool6000000h0oooo003o0?oooi`0oooo2000000<0?oo
o`00o`3ooonP0?ooo`D000002`3oool00?l0oooo/03oool00?l0oooo/03oool00?l0oooo/03oool0
0001\
\>"],
  ImageRangeCache->{{{0, 430.688}, {140.25, 0}} -> {-1.05044, -0.312299, \
0.0066802, 0.0066802}}]
}, Closed]],

Cell[BoxData[
    \(\(VelocityPlot2[LabVelocities, vCM, cmVelocities];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[RGBColor[0, 1, 0]], plot1, Graphics[RGBColor[1, 0, 0]], 
        plot2, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .89227 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.587767 0.952381 [
[ 0 0 0 0 ]
[ 1 .89227 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .89227 L
0 .89227 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
0 0 Mdot
grestore
} def
% End of user PostScript
0 1 0 r
.5 Mabswid
[ ] 0 setdash
.02381 .58777 m
.97619 .58777 L
s
% Start of user PostScript
0. 0. 1. 0. MAarrowhead1
% End of user PostScript
.02381 .58777 m
.02381 .58777 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.02381 .58777 m
.65873 .58777 L
s
% Start of user PostScript
0. 0. 0.66666667 0. MAarrowhead1
% End of user PostScript
.65873 .58777 m
.97619 .58777 L
s
% Start of user PostScript
0.66666667 0. 1. 0. MAarrowhead1
% End of user PostScript
.65873 .58777 m
.02381 .58777 L
s
% Start of user PostScript
0.66666667 0. 0. 0. MAarrowhead1
% End of user PostScript
1 0 0 r
.02381 .58777 m
.80206 .87103 L
s
% Start of user PostScript
0. 0. 0.81716593 0.29742408 MAarrowhead1
% End of user PostScript
.02381 .58777 m
.37206 .02124 L
s
% Start of user PostScript
0. 0. 0.36566813 -0.59484815 MAarrowhead1
% End of user PostScript
.02381 .58777 m
.65873 .58777 L
s
% Start of user PostScript
0. 0. 0.66666667 0. MAarrowhead1
% End of user PostScript
.65873 .58777 m
.80206 .87103 L
s
% Start of user PostScript
0.66666667 0. 0.81716593 0.29742408 MAarrowhead1
% End of user PostScript
.65873 .58777 m
.37206 .02124 L
s
% Start of user PostScript
0.66666667 0. 0.36566813 -0.59484815 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 256.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001PB000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00X03oool00`3o0000oooo0?oo
o`3o0?ooo`h0oooo002O0?ooo`<0o`00o`3oool?0?ooo`00WP3oool50?l00?l0oooo3P3oool009d0
oooo0`3o00000`3oool0o`000?l0003o0?ooo`h0oooo002L0?ooo`@0o`0000@0oooo0?l0003o0000
o`00o`3oool=0?ooo`00V`3oool50?l000030?ooo`3o0000o`000080o`00o`3oool<0?ooo`00VP3o
ool50?l00080oooo1@3o003o0?ooo`/0oooo002K0?ooo`@0o`000`3oool30?l00?l0oooo303oool0
09/0oooo00@0o`000?ooo`3o0000o`000`3oool00`3o0000oooo0?l0003o0?ooo``0oooo002J0?oo
o`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool0o`3oool90?ooo`00VP3oool00`3o0000
oooo0?ooo`080?ooo`030?l0003oool0oooo0?l0oooo2@3oool009T0oooo00<0o`000?ooo`3oool0
2P3oool00`3o0000oooo0?ooo`3o0?ooo`P0oooo002I0?ooo`030?l0003oool0oooo00X0oooo00<0
o`000?ooo`3oool0o`3oool80?ooo`00V03oool00`3o0000oooo0?ooo`0<0?ooo`030?l0003oool0
oooo0?l0oooo1`3oool009L0oooo00<0o`000?ooo`3oool03@3oool00`3o0000oooo0?ooo`3o0?oo
o`L0oooo002G0?ooo`030?l0003oool0oooo00h0oooo00<0o`000?ooo`3oool0o`3oool60?ooo`00
UP3oool00`3o0000oooo0?ooo`0?0?ooo`030?l0003oool0oooo0?l0oooo1P3oool009H0oooo00<0
o`000?ooo`3oool0403oool00`3o0000oooo0?ooo`3o0?ooo`D0oooo002E0?ooo`030?l0003oool0
oooo0140oooo00<0o`000?ooo`3oool0o`3oool50?ooo`00U03oool00`3o0000oooo0?ooo`0C0?oo
o`030?l0003oool0oooo0?l0oooo103oool009@0oooo00<0o`000?ooo`3oool04`3oool00`3o0000
oooo0?ooo`3o0?ooo`@0oooo002C0?ooo`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool0
o`3oool30?ooo`00TP3oool00`3o0000oooo0?ooo`0F0?ooo`030?l0003oool0oooo0?l0oooo0`3o
ool00980oooo00<0o`000?ooo`3oool05`3oool00`3o0000oooo0?ooo`3o0?ooo`80oooo002A0?oo
o`030?l0003oool0oooo01P0oooo00<0o`000?ooo`3oool0o`3oool20?ooo`00T@3oool00`3o0000
oooo0?ooo`0I0?ooo`030?l0003oool0oooo0?l0oooo0@3oool00900oooo00<0o`000?ooo`3oool0
6P3oool00`3o0000oooo0?ooo`3o0?ooo`40oooo002?0?ooo`030?l0003oool0oooo01`0oooo00<0
o`000?ooo`3oool0o`3oool008l0oooo00<0o`000?ooo`3oool0703oool00`3o0000oooo0?ooo`3o
0?ooo`00SP3oool00`3o0000oooo0?ooo`0N0?ooo`030?l0003oool0oooo0?h0oooo002>0?ooo`03
0?l0003oool0oooo01h0oooo00<0o`000?ooo`3oool0oP3oool008d0oooo00<0o`000?ooo`3oool0
803oool00`3o0000oooo0?ooo`3m0?ooo`00S03oool00`3o0000oooo0?ooo`0Q0?ooo`030?l0003o
ool0oooo0?d0oooo002<0?ooo`030?l0003oool0oooo0280oooo00<0o`000?ooo`3oool0o03oool0
08/0oooo00<0o`000?ooo`3oool08`3oool00`3o0000oooo0?ooo`3l0?ooo`00RP3oool00`3o0000
oooo0?ooo`0U0?ooo`030?l0003oool0oooo0?/0oooo002:0?ooo`030?l0003oool0oooo02D0oooo
00<0o`000?ooo`3oool0n`3oool008T0oooo00<0o`000?ooo`3oool09`3oool00`3o0000oooo0?oo
o`3j0?ooo`00R@3oool00`3o0000oooo0?ooo`0W0?ooo`030?l0003oool0oooo0?X0oooo00280?oo
o`030?l0003oool0oooo02T0oooo00<0o`000?ooo`3oool0n@3oool008L0oooo00<0o`000?ooo`3o
ool0:P3oool00`3o0000oooo0?ooo`3i0?ooo`00Q`3oool00`3o0000oooo0?ooo`0[0?ooo`030?l0
003oool0oooo0?P0oooo00260?ooo`030?l0003oool0oooo02`0oooo00<0o`000?ooo`3oool0n03o
ool008H0oooo00<0o`000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`3g0?ooo`00Q@3oool00`3o
0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo0?L0oooo00240?ooo`030?l0003oool0oooo0300
oooo00<0o`000?ooo`3oool0mP3oool008@0oooo00<0o`000?ooo`3oool0<03oool00`3o0000oooo
0?ooo`3f0?ooo`00P`3oool00`3o0000oooo0?ooo`0b0?ooo`030?l0003oool0oooo0?D0oooo0022
0?ooo`030?l0003oool0oooo03<0oooo00<0o`000?ooo`3oool0m@3oool00880oooo00<0o`000?oo
o`3oool0=03oool00`3o0000oooo0?ooo`3d0?ooo`00P@3oool00`3o0000oooo0?ooo`0e0?ooo`03
0?l0003oool0oooo0?@0oooo00210?ooo`030?l0003oool0oooo03H0oooo00<0o`000?ooo`3oool0
l`3oool00800oooo00<0o`000?ooo`3oool0=`3oool00`3o0000oooo0?ooo`3c0?ooo`00O`3oool0
0`3o0000oooo0?ooo`0i0?ooo`030?l0003oool0oooo0?80oooo001o0?ooo`030?l0003oool0oooo
03T0oooo00<0o`000?ooo`3oool0lP3oool007h0oooo00<0o`000?ooo`3oool0>`3oool00`3o0000
oooo0?ooo`3a0?ooo`00OP3oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo0?40oooo
001m0?ooo`030?l0003oool0oooo03d0oooo00<0o`000?ooo`3oool0l03oool007`0oooo00<0o`00
0?ooo`3oool0?P3oool00`3o0000oooo0?ooo`3`0?ooo`00O03oool00`3o0000oooo0?ooo`0o0?oo
o`030?l0003oool0oooo0>l0oooo001k0?ooo`030?l0003oool0oooo0400oooo00<0o`000?ooo`3o
ool0k`3oool007X0oooo00<0o`000?ooo`3oool0@P3oool00`3o0000oooo0?ooo`3^0?ooo`00NP3o
ool00`3o0000oooo0?ooo`120?ooo`030?l0003oool0oooo0>h0oooo001i0?ooo`030?l0003oool0
oooo04@0oooo00<0o`000?ooo`3oool0k@3oool007T0oooo00<0o`000?ooo`3oool0A03oool00`3o
0000oooo0?ooo`3]0?ooo`00N03oool00`3o0000oooo0?ooo`160?ooo`030?l0003oool0oooo0>`0
oooo001g0?ooo`030?l0003oool0oooo04L0oooo00<0o`000?ooo`3oool0k03oool007L0oooo00<0
o`000?ooo`3oool0B03oool00`3o0000oooo0?ooo`3[0?ooo`00MP3oool00`3o0000oooo0?ooo`19
0?ooo`030?l0003oool0oooo0>/0oooo001f0?ooo`030?l0003oool0oooo04X0oooo00<0o`000?oo
o`3oool0jP3oool007D0oooo00<0o`000?ooo`3oool0B`3oool00`3o0000oooo0?ooo`3Z0?ooo`00
M03oool00`3o0000oooo0?ooo`1=0?ooo`030?l0003oool0oooo0>T0oooo001d0?ooo`030?l0003o
ool0oooo04d0oooo00<0o`000?ooo`3oool0j@3oool007<0oooo00<0o`000?ooo`3oool0C`3oool0
0`3o0000oooo0?ooo`3X0?ooo`00L`3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0oooo
0>P0oooo001b0?ooo`030?l0003oool0oooo0540oooo00<0o`000?ooo`3oool0i`3oool00740oooo
00<0o`000?ooo`3oool0DP3oool00`3o0000oooo0?ooo`3W0?ooo`00L@3oool00`3o0000oooo0?oo
o`1C0?ooo`030?l0003oool0oooo0>H0oooo001`0?ooo`030?l0003oool0oooo05@0oooo00<0o`00
0?ooo`3oool0iP3oool006l0oooo00<0o`000?ooo`3oool0EP3oool00`3o0000oooo0?ooo`3U0?oo
o`00K`3oool00`3o0000oooo0?ooo`1F0?ooo`030?l0003oool0oooo0>D0oooo001^0?ooo`030?l0
003oool0oooo05P0oooo00<0o`000?ooo`3oool0i03oool006h0oooo00<0o`000?ooo`3oool0F03o
ool00`3o0000oooo0?ooo`3T0?ooo`00K@3oool00`3o0000oooo0?ooo`1J0?ooo`030?l0003oool0
oooo0><0oooo001/0?ooo`030?l0003oool0oooo05/0oooo00<0o`000?ooo`3oool0h`3oool006`0
oooo00<0o`000?ooo`3oool0G03oool00`3o0000oooo0?ooo`3R0?ooo`00J`3oool00`3o0000oooo
0?ooo`1M0?ooo`030?l0003oool0oooo0>80oooo001[0?ooo`030?l0003oool0oooo05h0oooo00<0
o`000?ooo`3oool0h@3oool006X0oooo00<0o`000?ooo`3oool0G`3oool00`3o0000oooo0?ooo`3Q
0?ooo`00J@3oool00`3o0000oooo0?ooo`1Q0?ooo`030?l0003oool0oooo0>00oooo001Y0?ooo`03
0?l0003oool0oooo0640oooo00<0o`000?ooo`3oool0h03oool006P0oooo00<0o`000?ooo`3oool0
H`3oool00`3o0000oooo0?ooo`3O0?ooo`00I`3oool00`3o0000oooo0?ooo`1T0?ooo`030?l0003o
ool0oooo0=l0oooo001W0?ooo`030?l0003oool0oooo06D0oooo00<0o`000?ooo`3oool0gP3oool0
06H0oooo00<0o`000?ooo`3oool0IP3oool00`3o0000oooo0?ooo`3N0?ooo`00IP3oool00`3o0000
oooo0?ooo`1W0?ooo`030?l0003oool0oooo0=d0oooo001U0?ooo`030?l0003oool0oooo06P0oooo
00<0o`000?ooo`3oool0g@3oool006@0oooo00<0o`000?ooo`3oool0JP3oool00`3o0000oooo0?oo
o`3L0?ooo`00I03oool00`3o0000oooo0?ooo`1Z0?ooo`030?l0003oool0oooo0=`0oooo001S0?oo
o`030?l0003oool0oooo06`0oooo00<0o`000?ooo`3oool0f`3oool006<0oooo00<0o`000?ooo`3o
ool0K03oool00`3o0000oooo0?ooo`3K0?ooo`00HP3oool00`3o0000oooo0?ooo`1^0?ooo`030?l0
003oool0oooo0=X0oooo001Q0?ooo`030?l0003oool0oooo06l0oooo00<0o`000?ooo`3oool0fP3o
ool00640oooo00<0o`000?ooo`3oool0L03oool00`3o0000oooo0?ooo`3I0?ooo`00H03oool00`3o
0000oooo0?ooo`1a0?ooo`030?l0003oool0oooo0=T0oooo001O0?ooo`030?l0003oool0oooo07<0
oooo00<0o`000?ooo`3oool0f03oool005l0oooo00<0o`000?ooo`3oool0L`3oool00`3o0000oooo
0?ooo`3H0?ooo`00GP3oool00`3o0000oooo0?ooo`1e0?ooo`030?l0003oool0oooo0=L0oooo001N
0?ooo`030?l0003oool0oooo07D0oooo00<0o`000?ooo`3oool0e`3oool005d0oooo00<0o`000?oo
o`3oool0M`3oool00`3o0000oooo0?ooo`3F0?ooo`00G03oool00`3o0000oooo0?ooo`1h0?ooo`03
0?l0003oool0oooo0=H0oooo001L0?ooo`030?l0003oool0oooo07T0oooo00<0o`000?ooo`3oool0
e@3oool005/0oooo00<0o`000?ooo`3oool0NP3oool00`3o0000oooo0?ooo`3E0?ooo`00F`3oool0
0`3o0000oooo0?ooo`1k0?ooo`030?l0003oool0oooo0=@0oooo001J0?ooo`030?l0003oool0oooo
07`0oooo00<0o`000?ooo`3oool0e03oool005T0oooo00<0o`000?ooo`3oool0OP3oool00`3o0000
oooo0?ooo`3C0?ooo`00F@3oool00`3o0000oooo0?ooo`1n0?ooo`030?l0003oool0oooo0=<0oooo
001H0?ooo`030?l0003oool0oooo0800oooo00<0o`000?ooo`3oool0dP3oool005L0oooo00<0o`00
0?ooo`3oool0P@3oool00`3o0000oooo0?ooo`3B0?ooo`00E`3oool00`3o0000oooo0?ooo`220?oo
o`030?l0003oool0oooo0=40oooo001F0?ooo`030?l0003oool0oooo08<0oooo00<0o`000?ooo`3o
ool0d@3oool005H0oooo00<0o`000?ooo`3oool0Q03oool00`3o0000oooo0?ooo`3@0?ooo`00E@3o
ool00`3o0000oooo0?ooo`250?ooo`030?l0003oool0oooo0=00oooo001D0?ooo`030?l0003oool0
oooo08L0oooo00<0o`000?ooo`3oool0c`3oool005@0oooo00<0o`000?ooo`3oool0R03oool00`3o
0000oooo0?ooo`3>0?ooo`00D`3oool00`3o0000oooo0?ooo`290?ooo`030?l0003oool0oooo0<h0
oooo001C0?ooo`030?l0003oool0oooo08X0oooo00<0o`000?ooo`3oool0c@3oool00580oooo00<0
o`000?ooo`3oool0R`3oool00`3o0000oooo0?ooo`3=0?ooo`00D@3oool00`3o0000oooo0?ooo`2=
0?ooo`030?l0003oool0oooo0<`0oooo001A0?ooo`030?l0003oool0oooo08d0oooo00<0o`000?oo
o`3oool0c03oool00500oooo00<0o`000?ooo`3oool0S`3oool00`3o0000oooo0?ooo`3;0?ooo`00
C`3oool00`3o0000oooo0?ooo`2@0?ooo`030?l0003oool0oooo0</0oooo001?0?ooo`030?l0003o
ool0oooo0940oooo00<0o`000?ooo`3oool0bP3oool004h0oooo00<0o`000?ooo`3oool0TP3oool0
0`3o0000oooo0?ooo`3:0?ooo`00CP3oool00`3o0000oooo0?ooo`2C0?ooo`030?l0003oool0oooo
0<T0oooo001=0?ooo`030?l0003oool0oooo09@0oooo00<0o`000?ooo`3oool0b@3oool004`0oooo
00<0o`000?ooo`3oool0UP3oool00`3o0000oooo0?ooo`380?ooo`00C03oool00`3o0000oooo0?oo
o`2F0?ooo`030?l0003oool0oooo0<P0oooo001;0?ooo`030?l0003oool0oooo09P0oooo00<0o`00
0?ooo`3oool0a`3oool004/0oooo00<0o`000?ooo`3oool0V03oool00`3o0000oooo0?ooo`370?oo
o`00BP3oool00`3o0000oooo0?ooo`2J0?ooo`030?l0003oool0oooo0<H0oooo00190?ooo`030?l0
003oool0oooo09/0oooo00<0o`000?ooo`3oool0aP3oool004T0oooo00<0o`000?ooo`3oool0W03o
ool00`3o0000oooo0?ooo`350?ooo`00B03oool00`3o0000oooo0?ooo`2M0?ooo`030?l0003oool0
oooo0<D0oooo00170?ooo`030?l0003oool0oooo09l0oooo00<0o`000?ooo`3oool0a03oool004L0
oooo00<0o`000?ooo`3oool0W`3oool00`3o0000oooo0?ooo`340?ooo`00AP3oool00`3o0000oooo
0?ooo`2Q0?ooo`030?l0003oool0oooo0<<0oooo00160?ooo`030?l0003oool0oooo0:40oooo00<0
o`000?ooo`3oool0``3oool004D0oooo00<0o`000?ooo`3oool0X`3oool00`3o0000oooo0?ooo`32
0?ooo`00A03oool00`3o0000oooo0?ooo`2T0?ooo`030?l0003oool0oooo0<80oooo00140?ooo`03
0?l0003oool0oooo0:D0oooo00<0o`000?ooo`3oool0`@3oool004<0oooo00<0o`000?ooo`3oool0
YP3oool00`3o0000oooo0?ooo`310?ooo`00@`3oool00`3o0000oooo0?ooo`2W0?ooo`030?l0003o
ool0oooo0<00oooo00120?ooo`030?l0003oool0oooo0:P0oooo00<0o`000?ooo`3oool0`03oool0
0440oooo00<0o`000?ooo`3oool0ZP3oool00`3o0000oooo0?ooo`2o0?ooo`00@@3oool00`3o0000
oooo0?ooo`2Z0?ooo`030?l0003oool0oooo0;l0oooo00100?ooo`030?l0003oool0oooo0:`0oooo
00<0o`000?ooo`3oool0_P3oool003l0oooo00<0o`000?ooo`3oool0[@3oool00`3o0000oooo0?oo
o`2n0?ooo`00?`3oool00`3o0000oooo0?ooo`2^0?ooo`030?l0003oool0oooo0;d0oooo000n0?oo
o`030?l0003oool0oooo0:l0oooo00<0o`000?ooo`3oool0_@3oool003h0oooo00<0o`000?ooo`3o
ool0/03oool00`3o0000oooo0?ooo`2l0?ooo`00?@3oool00`3o0000oooo0?ooo`2a0?ooo`030?l0
003oool0oooo0;`0oooo000l0?ooo`030?l0003oool0oooo0;<0oooo00<0o`000?ooo`3oool0^`3o
ool003`0oooo00<0o`000?ooo`3oool0/`3oool00`3o0000oooo0?ooo`2k0?ooo`00>`3oool00`3o
0000oooo0?ooo`2e0?ooo`030?l0003oool0oooo0;X0oooo000k0?ooo`030?l0003oool0oooo0;D0
oooo00<0o`000?ooo`3oool0^P3oool003X0oooo00<0o`000?ooo`3oool0]`3oool00`3o0000oooo
0?ooo`2i0?ooo`00>@3oool00`3o0000oooo0?ooo`2h0?ooo`030?l0003oool0oooo0;T0oooo000i
0?ooo`030?l0003oool0oooo0;T0oooo00<0o`000?ooo`3oool0^03oool003P0oooo00<0o`000?oo
o`3oool0^P3oool00`3o0000oooo0?ooo`2h0?ooo`00=`3oool00`3o0000oooo0?ooo`2l0?ooo`03
0?l0003oool0oooo0;L0oooo000g0?ooo`030?l0003oool0oooo0;`0oooo00<0o`000?ooo`3oool0
]`3oool003H0oooo00<0o`000?ooo`3oool0_P3oool00`3o0000oooo0?ooo`2f0?ooo`00=P3oool0
0`3o0000oooo0?ooo`2n0?ooo`030?l0003oool0oooo0;H0oooo000e0?ooo`030?l0003oool0oooo
0<00oooo00<0o`000?ooo`3oool0]@3oool003@0oooo00<0o`000?ooo`3oool0`@3oool00`3o0000
oooo0?ooo`2e0?ooo`00=03oool00`3o0000oooo0?ooo`320?ooo`030?l0003oool0oooo0;@0oooo
000c0?ooo`030?l0003oool0oooo0<<0oooo00<0o`000?ooo`3oool0]03oool003<0oooo00<0o`00
0?ooo`3oool0a03oool00`3o0000oooo0?ooo`2c0?ooo`00<P3oool00`3o0000oooo0?ooo`350?oo
o`030?l0003oool0oooo0;<0oooo000a0?ooo`030?l0003oool0oooo0<L0oooo00<0o`000?ooo`3o
ool0/P3oool00340oooo00<0o`000?ooo`3oool0a`3oool00`3o0000oooo0?ooo`2b0?ooo`00<03o
ool00`3o0000oooo0?ooo`390?ooo`030?l0003oool0oooo0;40oooo000_0?ooo`030?l0003oool0
oooo0<X0oooo00<0o`000?ooo`3oool0/@3oool002l0oooo00<0o`000?ooo`3oool0b`3oool00`3o
0000oooo0?ooo`2`0?ooo`00;P3oool00`3o0000oooo0?ooo`3<0?ooo`030?l0003oool0oooo0;00
oooo000^0?ooo`030?l0003oool0oooo0<d0oooo00<0o`000?ooo`3oool0[`3oool002d0oooo00<0
o`000?ooo`3oool0cP3oool00`3o0000oooo0?ooo`2_0?ooo`00;03oool00`3o0000oooo0?ooo`3@
0?ooo`030?l0003oool0oooo0:h0oooo000/0?ooo`030?l0003oool0oooo0=00oooo00<0o`000?oo
o`3oool0[P3oool002/0oooo00<0o`000?ooo`3oool0dP3oool00`3o0000oooo0?ooo`2]0?ooo`00
:`3oool00`3o0000oooo0?ooo`3B0?ooo`030?l0003oool0oooo0:d0oooo000Z0?ooo`030?l0003o
ool0oooo0=@0oooo00<0o`000?ooo`3oool0[03oool002T0oooo00<0o`000?ooo`3oool0e@3oool0
0`3o0000oooo0?ooo`2/0?ooo`00:@3oool00`3o0000oooo0?ooo`3F0?ooo`030?l0003oool0oooo
0:/0oooo000X0?ooo`030?l0003oool0oooo0=L0oooo00<0o`000?ooo`3oool0Z`3oool002P0oooo
00<0o`000?ooo`3oool0f03oool00`3o0000oooo0?ooo`2Z0?ooo`009`3oool00`3o0000oooo0?oo
o`3I0?ooo`030?l0003oool0oooo0:X0oooo000V0?ooo`030?l0003oool0oooo0=/0oooo00<0o`00
0?ooo`3oool0Z@3oool002H0oooo00<0o`000?ooo`3oool0f`3oool00`3o0000oooo0?ooo`2Y0?oo
o`009@3oool00`3o0000oooo0?ooo`3M0?ooo`030?l0003oool0oooo0:P0oooo000T0?ooo`030?l0
003oool0oooo0=h0oooo00<0o`000?ooo`3oool0Z03oool002@0oooo00<0o`000?ooo`3oool0g`3o
ool00`3o0000oooo0?ooo`2W0?ooo`008`3oool00`3o0000oooo0?ooo`3P0?ooo`030?l0003oool0
oooo0:L0oooo000S0?ooo`030?l0003oool0oooo0>40oooo00<0o`000?ooo`3oool0YP3oool00280
oooo00<0o`000?ooo`3oool0hP3oool00`3o0000oooo0?ooo`2V0?ooo`008@3oool00`3o0000oooo
0?ooo`3T0?ooo`030?l0003oool0oooo0:D0oooo000Q0?ooo`030?l0003oool0oooo0>@0oooo00<0
o`000?ooo`3oool0Y@3oool00200oooo00<0o`000?ooo`3oool0iP3oool00`3o0000oooo0?ooo`2T
0?ooo`00803oool00`3o0000oooo0?ooo`3V0?ooo`030?l0003oool0oooo0:@0oooo000O0?ooo`03
0?l0003oool0oooo0>P0oooo00<0o`000?ooo`3oool0X`3oool001h0oooo00<0o`000?ooo`3oool0
j@3oool00`3o0000oooo0?ooo`2S0?ooo`007P3oool00`3o0000oooo0?ooo`3Z0?ooo`030?l0003o
ool0oooo0:80oooo000M0?ooo`030?l0003oool0oooo0>/0oooo00<0o`000?ooo`3oool0XP3oool0
01`0oooo00<0o`000?ooo`3oool0k@3oool00`3o0000oooo0?ooo`2Q0?ooo`00703oool00`3o0000
oooo0?ooo`3]0?ooo`030?l0003oool0oooo0:40oooo000K0?ooo`030?l0003oool0oooo0>l0oooo
00<0o`000?ooo`3oool0X03oool001/0oooo00<0o`000?ooo`3oool0k`3oool00`3o0000oooo0?oo
o`2P0?ooo`006P3oool00`3o0000oooo0?ooo`3a0?ooo`030?l0003oool0oooo09l0oooo000I0?oo
o`030?l0003oool0oooo0?80oooo00<0o`000?ooo`3oool0W`3oool001T0oooo00<0o`000?ooo`3o
ool0l`3oool00`3o0000oooo0?ooo`2N0?ooo`00603oool00`3o0000oooo0?ooo`3d0?ooo`030?l0
003oool0oooo09h0oooo000H0?ooo`030?l0003oool0oooo0?D0oooo00<0o`000?ooo`3oool0W@3o
ool001L0oooo00<0o`000?ooo`3oool0mP3oool00`3o0000oooo0?ooo`2M0?ooo`005P3oool00`3o
0000oooo0?ooo`3h0?ooo`030?l0003oool0oooo09`0oooo000F0?ooo`030?l0003oool0oooo0?P0
oooo00<0o`000?ooo`3oool0W03oool001D0oooo00<0o`000?ooo`3oool0nP3oool00`3o0000oooo
0?ooo`2K0?ooo`00503oool00`3o0000oooo0?ooo`3k0?ooo`030?l0003oool0oooo09/0oooo000D
0?ooo`030?l0003oool0oooo0?`0oooo00<0o`000?ooo`3oool0VP3oool001<0oooo00<0o`000?oo
o`3oool0o@3oool00`3o0000oooo0?ooo`2J0?ooo`004`3oool00`3o0000oooo0?ooo`3n0?ooo`03
0?l0003oool0oooo09T0oooo000B0?ooo`030?l0003oool0oooo0?l0oooo00<0o`000?ooo`3oool0
V@3oool00140oooo00<0o`000?ooo`3oool0o`3oool20?ooo`030?l0003oool0oooo09P0oooo000A
0?ooo`030?l0003oool0oooo0?l0oooo0P3oool00`3o0000oooo0?ooo`2H0?ooo`00403oool00`3o
0000oooo0?ooo`3o0?ooo`@0oooo00<0o`000?ooo`3oool0U`3oool00100oooo00<0o`000?ooo`3o
ool0o`3oool40?ooo`030?l0003oool0oooo09L0oooo000?0?ooo`030?l0003oool0oooo0?l0oooo
1P3oool00`3o0000oooo0?ooo`2F0?ooo`003P3oool00`3o0000oooo0?ooo`3o0?ooo`L0oooo00<0
o`000?ooo`3oool0UP3oool000h0oooo00<0o`000?ooo`3oool0o`3oool80?ooo`030?l0003oool0
oooo09D0oooo000=0?ooo`030?l0003oool0oooo0?l0oooo2@3oool00`3o0000oooo0?ooo`2E0?oo
o`00303oool00`3o0000oooo0?ooo`3o0?ooo`/0oooo00<0o`000?ooo`3oool0U03oool000`0oooo
00<0o`000?ooo`3oool0o`3oool;0?ooo`030?l0003oool0oooo09@0oooo000;0?ooo`030?l0003o
ool0oooo0080oooo0P00o`3o0?ooo`<0oooo0P3o00040?ooo`030?l0003oool0oooo0800oooo0P00
o`0A0?ooo`002`3oool00`3o00000?l0003o0004003o0?l0oooo0`3oool70?l00880oooo1P00o`0=
0?ooo`002P3ooooo0?l001<0o`00R@00o`0;0?ooo`00303oool30?l000<00?l0o`3oool30?ooo`H0
o`000P3oool00`3o0000oooo0?ooo`1n0?ooo`H00?l03@3oool000l0oooo0P3o00000`00o`00oooo
0?ooo`3o0?ooo`40oooo0P3o00060?ooo`030?l0003oool0oooo07h0oooo0P00o`0A0?ooo`004@3o
ool30?l00?l0oooo2P3oool00`3o0000oooo0?ooo`2@0?ooo`00503oool30?l00?l0oooo1`3oool0
0`3o0000oooo0?ooo`2@0?ooo`005`3oool30?l00?l0oooo1@3oool00`3o0000oooo0?ooo`2?0?oo
o`006P3oool20?l00?l0oooo0`3oool00`3o0000oooo0?ooo`2?0?ooo`00703oool30?l00?l0oooo
0@3oool00`3o0000oooo0?ooo`2>0?ooo`007`3oool30?l00?d0oooo00<0o`000?ooo`3oool0SP3o
ool00280oooo0`3o003k0?ooo`030?l0003oool0oooo08d0oooo000U0?ooo`80o`00n@3oool00`3o
0000oooo0?ooo`2=0?ooo`009`3oool30?l00?L0oooo00<0o`000?ooo`3oool0S03oool002X0oooo
0`3o003d0?ooo`030?l0003oool0oooo08`0oooo000]0?ooo`<0o`00lP3oool00`3o0000oooo0?oo
o`2;0?ooo`00<03oool20?l00?00oooo00<0o`000?ooo`3oool0R`3oool00380oooo0`3o003^0?oo
o`030?l0003oool0oooo08X0oooo000e0?ooo`<0o`00j`3oool00`3o0000oooo0?ooo`2:0?ooo`00
>03oool30?l00>T0oooo00<0o`000?ooo`3oool0R@3oool003/0oooo0P3o003W0?ooo`030?l0003o
ool0oooo08T0oooo000m0?ooo`<0o`00i@3oool00`3o0000oooo0?ooo`280?ooo`00@03oool30?l0
0>80oooo00<0o`000?ooo`3oool0R03oool004<0oooo0`3o003P0?ooo`030?l0003oool0oooo08L0
oooo00160?ooo`80o`00gP3oool00`3o0000oooo0?ooo`270?ooo`00B03oool30?l00=`0oooo00<0
o`000?ooo`3oool0QP3oool004/0oooo0`3o003I0?ooo`030?l0003oool0oooo08H0oooo001>0?oo
o`<0o`00e`3oool00`3o0000oooo0?ooo`250?ooo`00D@3oool20?l00=D0oooo00<0o`000?ooo`3o
ool0Q@3oool005<0oooo0`3o003C0?ooo`030?l0003oool0oooo08@0oooo001F0?ooo`<0o`00d03o
ool00`3o0000oooo0?ooo`240?ooo`00F@3oool30?l00<h0oooo00<0o`000?ooo`3oool0P`3oool0
05`0oooo0P3o003<0?ooo`030?l0003oool0oooo08<0oooo001N0?ooo`<0o`00bP3oool00`3o0000
oooo0?ooo`220?ooo`00H@3oool30?l00<L0oooo00<0o`000?ooo`3oool0PP3oool006@0oooo0P3o
00360?ooo`030?l0003oool0oooo0840oooo001V0?ooo`<0o`00``3oool00`3o0000oooo0?ooo`21
0?ooo`00J@3oool30?l00<40oooo00<0o`000?ooo`3oool0P03oool006`0oooo0`3o002n0?ooo`03
0?l0003oool0oooo0800oooo001_0?ooo`80o`00_@3oool00`3o0000oooo0?ooo`1o0?ooo`00L@3o
ool30?l00;X0oooo00<0o`000?ooo`3oool0O`3oool007@0oooo0`3o002h0?ooo`030?l0003oool0
oooo07h0oooo001g0?ooo`<0o`00]@3oool00`3o0000oooo0?ooo`1n0?ooo`00NP3oool20?l00;@0
oooo00<0o`000?ooo`3oool0O@3oool007`0oooo0`3o002a0?ooo`030?l0003oool0oooo07d0oooo
001o0?ooo`<0o`00[`3oool00`3o0000oooo0?ooo`1l0?ooo`00PP3oool30?l00:`0oooo00<0o`00
0?ooo`3oool0O03oool008D0oooo0P3o002[0?ooo`030?l0003oool0oooo07/0oooo00270?ooo`<0
o`00Z03oool00`3o0000oooo0?ooo`1k0?ooo`00RP3oool30?l00:H0oooo00<0o`000?ooo`3oool0
NP3oool008d0oooo0`3o002S0?ooo`030?l0003oool0oooo07X0oooo002@0?ooo`80o`00XP3oool0
0`3o0000oooo0?ooo`1i0?ooo`00TP3oool30?l009l0oooo00<0o`000?ooo`3oool0N@3oool009D0
oooo0`3o002M0?ooo`030?l0003oool0oooo07P0oooo002H0?ooo`<0o`00VP3oool00`3o0000oooo
0?ooo`1h0?ooo`00V`3oool20?l009T0oooo00<0o`000?ooo`3oool0M`3oool009d0oooo0`3o002F
0?ooo`030?l0003oool0oooo07L0oooo002P0?ooo`<0o`00U03oool00`3o0000oooo0?ooo`1f0?oo
o`00X`3oool30?l00940oooo00<0o`000?ooo`3oool0MP3oool00:H0oooo0P3o002@0?ooo`030?l0
003oool0oooo07D0oooo002X0?ooo`<0o`00S@3oool00`3o0000oooo0?ooo`1e0?ooo`00Z`3oool3
0?l008/0oooo00<0o`000?ooo`3oool0M03oool00:h0oooo0`3o00280?ooo`030?l0003oool0oooo
07@0oooo002a0?ooo`80o`00Q`3oool00`3o0000oooo0?ooo`1c0?ooo`00/`3oool30?l008D0oooo
00<0o`000?ooo`3oool0LP3oool00;H0oooo0`3o00220?ooo`030?l0003oool0oooo0780oooo002i
0?ooo`<0o`00P03oool00`3o0000oooo0?ooo`1a0?ooo`00_03oool20?l007h0oooo00<0o`000?oo
o`3oool0L@3oool00;h0oooo0`3o001l0?ooo`030?l0003oool0oooo0700oooo00310?ooo`<0o`00
N@3oool00`3o0000oooo0?ooo`1`0?ooo`00a03oool30?l007L0oooo00<0o`000?ooo`3oool0K`3o
ool00<L0oooo0P3o001e0?ooo`030?l0003oool0oooo06l0oooo00390?ooo`<0o`00L`3oool00`3o
0000oooo0?ooo`1^0?ooo`00c03oool30?l00700oooo00<0o`000?ooo`3oool0KP3oool00<l0oooo
0`3o001^0?ooo`030?l0003oool0oooo06d0oooo003B0?ooo`80o`00K03oool00`3o0000oooo0?oo
o`1]0?ooo`00e03oool30?l006X0oooo00<0o`000?ooo`3oool0K03oool00=L0oooo0`3o001W0?oo
o`030?l0003oool0oooo06`0oooo003J0?ooo`<0o`00I@3oool00`3o0000oooo0?ooo`1[0?ooo`00
g@3oool20?l006<0oooo00<0o`000?ooo`3oool0J`3oool00=l0oooo0`3o001Q0?ooo`030?l0003o
ool0oooo06X0oooo003R0?ooo`<0o`00GP3oool00`3o0000oooo0?ooo`1Z0?ooo`00i@3oool30?l0
05`0oooo00<0o`000?ooo`3oool0J@3oool00>P0oooo0P3o001J0?ooo`030?l0003oool0oooo06T0
oooo003Z0?ooo`<0o`00F03oool00`3o0000oooo0?ooo`1X0?ooo`00k@3oool30?l005D0oooo00<0
o`000?ooo`3oool0J03oool00?00oooo0`3o001C0?ooo`030?l0003oool0oooo06L0oooo003c0?oo
o`80o`00D@3oool00`3o0000oooo0?ooo`1W0?ooo`00m@3oool30?l004l0oooo00<0o`000?ooo`3o
ool0IP3oool00?P0oooo0`3o001<0?ooo`030?l0003oool0oooo06H0oooo003k0?ooo`<0o`00BP3o
ool00`3o0000oooo0?ooo`1U0?ooo`00oP3oool20?l004P0oooo00<0o`000?ooo`3oool0I@3oool0
0?l0oooo0@3oool30?l004H0oooo00<0o`000?ooo`3oool0I03oool00?l0oooo103oool30?l004<0
oooo00<0o`000?ooo`3oool0I03oool00?l0oooo1`3oool20?l00480oooo00<0o`000?ooo`3oool0
H`3oool00?l0oooo2@3oool30?l003l0oooo00<0o`000?ooo`3oool0H`3oool00?l0oooo303oool3
0?l003d0oooo00<0o`000?ooo`3oool0HP3oool00?l0oooo3`3oool30?l003X0oooo00<0o`000?oo
o`3oool0HP3oool00?l0oooo4P3oool20?l003T0oooo00<0o`000?ooo`3oool0H@3oool00?l0oooo
503oool30?l003H0oooo00<0o`000?ooo`3oool0H@3oool00?l0oooo5`3oool30?l003@0oooo00<0
o`000?ooo`3oool0H03oool00?l0oooo6P3oool30?l00340oooo00<0o`000?ooo`3oool0H03oool0
0?l0oooo7@3oool20?l00300oooo00<0o`000?ooo`3oool0G`3oool00?l0oooo7`3oool30?l002d0
oooo00<0o`000?ooo`3oool0G`3oool00?l0oooo8P3oool30?l002/0oooo00<0o`000?ooo`3oool0
GP3oool00?l0oooo9@3oool30?l002P0oooo00<0o`000?ooo`3oool0GP3oool00?l0oooo:03oool2
0?l002L0oooo00<0o`000?ooo`3oool0G@3oool00?l0oooo:P3oool30?l002@0oooo00<0o`000?oo
o`3oool0G@3oool00?l0oooo;@3oool30?l00280oooo00<0o`000?ooo`3oool0G03oool00?l0oooo
<03oool30?l001l0oooo00<0o`000?ooo`3oool0G03oool00?l0oooo<`3oool20?l001h0oooo00<0
o`000?ooo`3oool0F`3oool00?l0oooo=@3oool30?l001/0oooo00<0o`000?ooo`3oool0F`3oool0
0?l0oooo>03oool30?l001T0oooo00<0o`000?ooo`3oool0FP3oool00?l0oooo>`3oool30?l001H0
oooo00<0o`000?ooo`3oool0FP3oool00?l0oooo?P3oool20?l001D0oooo00<0o`000?ooo`3oool0
F@3oool00?l0oooo@03oool30?l00180oooo00<0o`000?ooo`3oool0F@3oool00?l0oooo@`3oool3
0?l00100oooo00<0o`000?ooo`3oool0F03oool00?l0ooooAP3oool30?l000d0oooo00<0o`000?oo
o`3o0000F03oool00?l0ooooB@3oool20?l000/0oooo0`3o001H0?ooo`00o`3ooom;0?ooo`<0o`00
1`3oool50?l005L0oooo003o0?ooodh0oooo0`3o00030?ooo`030?l0003oool0o`0000<0o`00E`3o
ool00?l0ooooD@3oool90?l005L0oooo003o0?oooe<0oooo1`3o001G0?ooo`00o`3ooomB0?ooo`T0
o`00EP3oool00?l0ooooDP3oool90?l005H0oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {384.375, 0}} -> {-0.0252693, -0.61716, \
0.00365617, 0.00365617}}]
}, Closed]],

Cell[BoxData[
    \(\(PlotCMmomenta[cmMomenta];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[%, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 1.42857 0.5 1.42857 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.5 .5 m
.97619 .5 L
s
% Start of user PostScript
0. 0. 0.33333333 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.02381 .5 L
s
% Start of user PostScript
0. 0. -0.33333333 0. MAarrowhead1
% End of user PostScript
.5 .5 m
.715 .92489 L
s
% Start of user PostScript
0. 0. 0.15049927 0.29742408 MAarrowhead1
% End of user PostScript
.5 .5 m
.285 .07511 L
s
% Start of user PostScript
0. 0. -0.15049927 -0.29742408 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
newpath
.5 .5 .47619 0 365.73 arc
s
3 Mabswid
.5 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ00000<P0oooo002m0?ooo`X000008@3oool:00000;h0oooo002e0?ooo`P00000
=@3oool800000;H0oooo002`0?ooo`D00000A@3oool500000;40oooo002[0?ooo`D00000C`3oool5
00000:`0oooo002V0?ooo`D00000F@3oool200000:X0oooo002T0?ooo`800000H03oool500000:D0
oooo002O0?ooo`D00000I`3oool500000:00oooo002M0?ooo`800000L@3oool3000009d0oooo002J
0?ooo`<00000MP3oool2000009/0oooo002E0?ooo`D00000N`3oool5000009H0oooo002C0?ooo`80
0000Q@3oool2000009@0oooo002@0?ooo`<00000R@3oool300000940oooo002>0?ooo`800000S`3o
ool2000008l0oooo002<0?ooo`800000T`3oool2000008d0oooo00290?ooo`<00000U`3oool30000
08X0oooo00270?ooo`800000W@3oool2000008P0oooo00250?ooo`800000X@3oool2000008H0oooo
00220?ooo`<00000Y@3oool3000008<0oooo00200?ooo`800000Z`3oool200000840oooo001n0?oo
o`800000[`3oool2000007l0oooo001k0?ooo`<00000/`3oool3000007`0oooo001i0?ooo`800000
^@3oool2000007X0oooo001h0?ooo`040000003oool0000000000;X0oooo00<000000?ooo`3oool0
M`3oool007L0oooo00@000000?ooo`3oool0oooo0P00002j0?ooo`030000003oool0oooo07H0oooo
001e0?ooo`800000103oool200000;/0oooo0P00001f0?ooo`00LP3oool3000000H0oooo0`00002l
0?ooo`<00000L`3oool00700oooo0P0000090?ooo`@00000_P3oool200000740oooo001_0?ooo`03
0000003oool0oooo00X0oooo1000002o0?ooo`030000003oool0oooo06h0oooo001^0?ooo`030000
003oool0oooo00/0oooo0`0000310?ooo`030000003oool0oooo06d0oooo001/0?ooo`800000403o
ool00`000000oooo0?ooo`300?ooo`800000K@3oool006X0oooo0P00000C0?ooo`030000003oool0
oooo0<40oooo0P00001[0?ooo`00J@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
0<<0oooo00<000000?ooo`3oool0J03oool006P0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`330?ooo`800000J03oool006D0oooo0`00000H0?ooo`030000003oool0oooo0<D0oooo
0P00001V0?ooo`00H`3oool2000001`0oooo00<000000?ooo`3oool0aP3oool2000006@0oooo001R
0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`1Q
0?ooo`00H@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0<P0oooo00<000000?oo
o`3oool0H03oool005l0oooo0P00000Q0?ooo`030000003oool0oooo0<T0oooo0P00001P0?ooo`00
GP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3oool0
G@3oool005d0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3;0?ooo`030000003o
ool0oooo05`0oooo001K0?ooo`8000009`3oool00`000000oooo0?ooo`3;0?ooo`800000G03oool0
05X0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo
05T0oooo001I0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0c@3oool00`000000
oooo0?ooo`1H0?ooo`00F03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0<h0oooo
00<000000?ooo`3oool0E`3oool005L0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?oo
o`3>0?ooo`030000003oool0oooo05H0oooo001E0?ooo`800000;`3oool00`000000oooo0?ooo`3?
0?ooo`800000EP3oool005@0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`3@0?oo
o`030000003oool0oooo05<0oooo001C0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3o
ool0d@3oool00`000000oooo0?ooo`1B0?ooo`00D@3oool2000003D0oooo00<000000?ooo`3oool0
d@3oool200000580oooo001@0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0d`3o
ool00`000000oooo0?ooo`1?0?ooo`00C`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo0=<0oooo00<000000?ooo`3oool0CP3oool004h0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`3D0?ooo`030000003oool0oooo04d0oooo001=0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool0e03oool00`000000oooo0?ooo`1<0?ooo`00C03oool00`000000oooo
0?ooo`0k0?ooo`030000003oool0oooo0=D0oooo00<000000?ooo`3oool0B`3oool004/0oooo00<0
00000?ooo`3oool0?@3oool00`000000oooo0?ooo`3E0?ooo`030000003oool0oooo04X0oooo001:
0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0eP3oool00`000000oooo0?ooo`19
0?ooo`00B@3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo0=H0oooo00<000000?oo
o`3oool0B03oool004P0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`3G0?ooo`03
0000003oool0oooo04L0oooo00170?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0
e`3oool00`000000oooo0?ooo`160?ooo`00AP3oool00`000000oooo0?ooo`140?ooo`030000003o
ool0oooo0=P0oooo00<000000?ooo`3oool0A@3oool004D0oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`3H0?ooo`030000003oool0oooo04@0oooo00140?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool0f@3oool00`000000oooo0?ooo`130?ooo`00@`3oool00`000000
oooo0?ooo`190?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0@P3oool00480oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`3J0?ooo`030000003oool0oooo0440oooo
00110?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0fP3oool00`000000oooo0?oo
o`100?ooo`00@03oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo0=/0oooo00<00000
0?ooo`3oool0?`3oool003l0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`3K0?oo
o`030000003oool0oooo03h0oooo000n0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3o
ool0g03oool00`000000oooo0?ooo`0m0?ooo`00?@3oool00`000000oooo0?ooo`1B0?ooo`030000
003oool0oooo0=`0oooo00<000000?ooo`3oool0?03oool003`0oooo00<000000?ooo`3oool0D`3o
ool00`000000oooo0?ooo`3M0?ooo`030000003oool0oooo03/0oooo000k0?ooo`030000003oool0
oooo05D0oooo00<000000?ooo`3oool0g@3oool00`000000oooo0?ooo`0j0?ooo`00>P3oool00`00
0000oooo0?ooo`1F0?ooo`030000003oool0oooo0=d0oooo00<000000?ooo`3oool0>P3oool003X0
oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`3M0?ooo`030000003oool0oooo03T0
oooo000i0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0gP3oool00`000000oooo
0?ooo`0h0?ooo`00>03oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0=h0oooo00<0
00000?ooo`3oool0=`3oool003L0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`3N
0?ooo`030000003oool0oooo03L0oooo000g0?ooo`030000003oool0oooo05`0oooo00<000000?oo
o`3oool0gP3oool00`000000oooo0?ooo`0f0?ooo`00=P3oool00`000000oooo0?ooo`1M0?ooo`03
0000003oool0oooo0=l0oooo00<000000?ooo`3oool0=@3oool003D0oooo00<000000?ooo`3oool0
G`3oool00`000000oooo0?ooo`3O0?ooo`030000003oool0oooo03@0oooo000d0?ooo`030000003o
ool0oooo0600oooo00<000000?ooo`3oool0h03oool00`000000oooo0?ooo`0c0?ooo`00<`3oool0
0`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0<P3oool0
0380oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo
0380oooo000b0?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0h03oool00`000000
oooo0?ooo`0a0?ooo`00<@3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo0>00oooo
00<000000?ooo`3oool0<03oool00300oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?oo
o`3Q0?ooo`030000003oool0oooo02l0oooo000_0?ooo`030000003oool0oooo06T0oooo00<00000
0?ooo`3oool0h03oool00`000000oooo0?ooo`0_0?ooo`00;`3oool00`000000oooo0?ooo`1Y0?oo
o`030000003oool0oooo0>40oooo00<000000?ooo`3oool0;P3oool002h0oooo00<000000?ooo`3o
ool0J`3oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo02d0oooo000]0?ooo`030000
003oool0oooo06`0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`0/0?ooo`00;03o
ool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo0>40oooo00<000000?ooo`3oool0;03o
ool002`0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0
oooo02/0oooo000/0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0h@3oool00`00
0000oooo0?ooo`0[0?ooo`00:`3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo0>40
oooo00<000000?ooo`3oool0:`3oool002/0oooo00<000000?ooo`3oool0L@3oool00`000000oooo
0?ooo`3Q0?ooo`030000003oool0oooo02X0oooo000Z0?ooo`030000003oool0oooo0780oooo00<0
00000?ooo`3oool0hP3oool00`000000oooo0?ooo`0Y0?ooo`00:@3oool00`000000oooo0?ooo`1d
0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0:03oool002P0oooo00<000000?oo
o`3oool0M@3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo02P0oooo000X0?ooo`03
0000003oool0oooo07H0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`0W0?ooo`00
9`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0
9`3oool002L0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`3R0?ooo`030000003o
ool0oooo02H0oooo000V0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0h`3oool0
0`000000oooo0?ooo`0U0?ooo`009@3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo
0><0oooo00<000000?ooo`3oool0903oool002D0oooo00<000000?ooo`3oool0N`3oool00`000000
oooo0?ooo`3S0?ooo`030000003oool0oooo02@0oooo000T0?ooo`030000003oool0oooo07d0oooo
00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`0S0?ooo`00903oool00`000000oooo0?oo
o`1m0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool08`3oool002<0oooo00<00000
0?ooo`3oool0O`3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo02<0oooo000S0?oo
o`030000003oool0oooo07l0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`0R0?oo
o`008P3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3o
ool08P3oool00280oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`3S0?ooo`030000
003oool0oooo0240oooo000Q0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0h`3o
ool00`000000oooo0?ooo`0P0?ooo`00803oool00`000000oooo0?ooo`240?ooo`030000003oool0
oooo0>@0oooo00<000000?ooo`3oool07`3oool00200oooo00<000000?ooo`3oool0Q@3oool00`00
0000oooo0?ooo`3S0?ooo`030000003oool0oooo01l0oooo000O0?ooo`030000003oool0oooo08H0
oooo00<000000?ooo`3oool0i03oool00`000000oooo0?ooo`0N0?ooo`007`3oool00`000000oooo
0?ooo`270?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool07P3oool001h0oooo00<0
00000?ooo`3oool0R03oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo01h0oooo000N
0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`0M
0?ooo`007@3oool00`000000oooo0?ooo`2:0?ooo`030000003oool0oooo0><0oooo00<000000?oo
o`3oool07@3oool001d0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`3S0?ooo`03
0000003oool0oooo01`0oooo000M0?ooo`030000003oool0oooo08/0oooo00<000000?ooo`3oool0
h`3oool00`000000oooo0?ooo`0L0?ooo`00703oool00`000000oooo0?ooo`2=0?ooo`030000003o
ool0oooo0><0oooo00<000000?ooo`3oool06`3oool001`0oooo00<000000?ooo`3oool0S@3oool0
0`000000oooo0?ooo`3S0?ooo`030000003oool0oooo01/0oooo000K0?ooo`030000003oool0oooo
08l0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`0K0?ooo`006`3oool00`000000
oooo0?ooo`2?0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool06P3oool001X0oooo
00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo01X0oooo
000J0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?oo
o`0I0?ooo`006P3oool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo0>80oooo00<00000
0?ooo`3oool06@3oool001T0oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`3S0?oo
o`030000003oool0oooo01P0oooo000I0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3o
ool0hP3oool00`000000oooo0?ooo`0H0?ooo`00603oool00`000000oooo0?ooo`2E0?ooo`030000
003oool0oooo0>80oooo00<000000?ooo`3oool0603oool001P0oooo00<000000?ooo`3oool0UP3o
ool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo01L0oooo000G0?ooo`030000003oool0
oooo09L0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`0G0?ooo`005`3oool00`00
0000oooo0?ooo`2H0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool05P3oool001L0
oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo01H0
oooo000F0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool0hP3oool00`000000oooo
0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo0>80oooo00<0
00000?ooo`3oool05@3oool001D0oooo00<000000?ooo`3oool0W03oool00`000000oooo0?ooo`3Q
0?ooo`030000003oool0oooo01D0oooo000E0?ooo`030000003oool0oooo09`0oooo00<000000?oo
o`3oool0hP3oool00`000000oooo0?ooo`0D0?ooo`005@3oool00`000000oooo0?ooo`2M0?ooo`03
0000003oool0oooo0>40oooo00<000000?ooo`3oool0503oool001@0oooo00<000000?ooo`3oool0
WP3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo01<0oooo000D0?ooo`030000003o
ool0oooo09l0oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`0C0?ooo`00503oool0
0`000000oooo0?ooo`2O0?ooo`030000003oool0oooo0>40oooo00<000000?ooo`3oool04`3oool0
01@0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo
01<0oooo000C0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0h03oool00`000000
oooo0?ooo`0C0?ooo`004`3oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo0>00oooo
00<000000?ooo`3oool04P3oool001<0oooo00<000000?ooo`3oool0X`3oool00`000000oooo0?oo
o`3O0?ooo`030000003oool0oooo0180oooo000B0?ooo`030000003oool0oooo0:@0oooo00<00000
0?ooo`3oool0g`3oool00`000000oooo0?ooo`0B0?ooo`004P3oool00`000000oooo0?ooo`2U0?oo
o`030000003oool0oooo0=l0oooo00<000000?ooo`3oool04@3oool00140oooo00<000000?ooo`3o
ool0YP3oool00`000000oooo0?ooo`3O0?ooo`030000003oool0oooo0140oooo000A0?ooo`030000
003oool0oooo0:L0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`0@0?ooo`004@3o
ool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0=l0oooo00<000000?ooo`3oool0403o
ool00140oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`3N0?ooo`030000003oool0
oooo0100oooo000A0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0gP3oool00`00
0000oooo0?ooo`0@0?ooo`00403oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0=d0
oooo00<000000?ooo`3oool0403oool00100oooo00<000000?ooo`3oool0ZP3oool00`000000oooo
0?ooo`3N0?ooo`030000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo0:/0oooo00<0
00000?ooo`3oool0g@3oool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo0=h0oooo00<000000?ooo`3oool03P3oool00100oooo00<000000?oo
o`3oool0[03oool00`000000oooo0?ooo`3M0?ooo`030000003oool0oooo00h0oooo000?0?ooo`03
0000003oool0oooo0:d0oooo00<000000?ooo`3oool0g@3oool00`000000oooo0?ooo`0>0?ooo`00
3`3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo0=`0oooo00<000000?ooo`3oool0
3P3oool000l0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?ooo`3L0?ooo`030000003o
ool0oooo00h0oooo000>0?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool0g03oool0
0`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo
0=`0oooo00<000000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3oool0/@3oool00`000000
oooo0?ooo`3K0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo0;40oooo
00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?oo
o`2b0?ooo`030000003oool0oooo0=X0oooo00<000000?ooo`3oool03@3oool000d0oooo00<00000
0?ooo`3oool0/`3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo00`0oooo000=0?oo
o`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0fP3oool00`000000oooo0?ooo`0<0?oo
o`003@3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0=X0oooo00<000000?ooo`3o
ool0303oool000d0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`3I0?ooo`030000
003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0f@3o
ool00`000000oooo0?ooo`0<0?ooo`00303oool00`000000oooo0?ooo`2g0?ooo`030000003oool0
oooo0=T0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0]`3oool00`00
0000oooo0?ooo`3I0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0;P0
oooo00<000000?ooo`3oool0f03oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo
0?ooo`2h0?ooo`030000003oool0oooo0=P0oooo00<000000?ooo`3oool02`3oool000`0oooo00<0
00000?ooo`3oool0^@3oool00`000000oooo0?ooo`3G0?ooo`030000003oool0oooo00/0oooo000<
0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`0;
0?ooo`00303oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo0=H0oooo00<000000?oo
o`3oool02`3oool000/0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`3F0?ooo`03
0000003oool0oooo00/0oooo000;0?ooo`030000003oool0oooo0;`0oooo00<000000?ooo`3oool0
eP3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2l0?ooo`030000003o
ool0oooo0=H0oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0_@3oool0
0`000000oooo0?ooo`3E0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo
0;d0oooo00<000000?ooo`3oool0e@3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000
oooo0?ooo`2n0?ooo`030000003oool0oooo0=@0oooo00<000000?ooo`3oool02P3oool000/0oooo
00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`3D0?ooo`030000003oool0oooo00X0oooo
000;0?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool0d`3oool00`000000oooo0?oo
o`0:0?ooo`002`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo0=<0oooo00<00000
0?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`3B0?oo
o`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3o
ool0dP3oool00`000000oooo0?ooo`0:0?ooo`002P3oool00`000000oooo0?ooo`320?ooo`030000
003oool0oooo0=80oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0`P3o
ool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0
oooo0<<0oooo00<000000?ooo`3oool0d@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`00
0000oooo0?ooo`330?ooo`030000003oool0oooo0=40oooo00<000000?ooo`3oool02@3oool000X0
oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`3@0?ooo`030000003oool0oooo00T0
oooo000:0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool0d03oool00`000000oooo
0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo0<l0oooo00<0
00000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`3?
0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<H0oooo00<000000?oo
o`3oool0cP3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`360?ooo`03
0000003oool0oooo0<h0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0
a`3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003o
ool0oooo0<L0oooo00<000000?ooo`3oool0c@3oool00`000000oooo0?ooo`090?ooo`002P3oool0
0`000000oooo0?ooo`380?ooo`030000003oool0oooo0<`0oooo00<000000?ooo`3oool02@3oool0
00X0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo
00T0oooo000:0?ooo`030000003oool0oooo00<0oooo0P0000340?ooo`030000003oool0oooo0<@0
oooo0P0000050?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0000000D00000
a03oool200000<D0oooo1P0000000`3oool000000?ooo`0:0?ooo`002P3ooooo000009`000002`3o
ool000X0oooo00<000000?ooo`0000001@0000330?ooo`@00000a03oool6000000030?ooo`000000
oooo00X0oooo000:0?ooo`030000003oool0oooo00<0oooo0P0000340?ooo`<00000a03oool20000
00D0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0c03oool00`000000
oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<`0oooo
00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?oo
o`3=0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02@3oool000X0oooo00<00000
0?ooo`3oool0c@3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00T0oooo000:0?oo
o`030000003oool0oooo0<h0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`090?oo
o`002P3oool00`000000oooo0?ooo`3>0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3o
ool02@3oool000X0oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`350?ooo`030000
003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<l0oooo00<000000?ooo`3oool0a@3o
ool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`3@0?ooo`030000003oool0
oooo0<@0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0d03oool00`00
0000oooo0?ooo`340?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0=40
oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo
0?ooo`3A0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool02@3oool000X0oooo00<0
00000?ooo`3oool0dP3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo00T0oooo000:
0?ooo`030000003oool0oooo0=80oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`09
0?ooo`002`3oool00`000000oooo0?ooo`3B0?ooo`030000003oool0oooo0<00oooo00<000000?oo
o`3oool02P3oool000/0oooo00<000000?ooo`3oool0dP3oool00`000000oooo0?ooo`300?ooo`03
0000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0
_`3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`3C0?ooo`030000003o
ool0oooo0;l0oooo00<000000?ooo`3oool02P3oool000/0oooo00<000000?ooo`3oool0e03oool0
0`000000oooo0?ooo`2n0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo
0=@0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000
oooo0?ooo`3E0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool02P3oool000/0oooo
00<000000?ooo`3oool0e@3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo00X0oooo
000;0?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?oo
o`0:0?ooo`002`3oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo0;`0oooo00<00000
0?ooo`3oool02P3oool000`0oooo00<000000?ooo`3oool0eP3oool00`000000oooo0?ooo`2j0?oo
o`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3o
ool0^P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`3G0?ooo`030000
003oool0oooo0;T0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0e`3o
ool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0
oooo0=P0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`3H0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool02`3oool000`0
oooo00<000000?ooo`3oool0f@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo00/0
oooo000<0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`0;0?ooo`003@3oool00`000000oooo0?ooo`3I0?ooo`030000003oool0oooo0;D0oooo00<0
00000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0f@3oool00`000000oooo0?ooo`2e
0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0=X0oooo00<000000?oo
o`3oool0]03oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`3J0?ooo`03
0000003oool0oooo0;@0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0
f`3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo00`0oooo000>0?ooo`030000003o
ool0oooo0=X0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0=0?ooo`003P3oool0
0`000000oooo0?ooo`3K0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool03@3oool0
00h0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo
00d0oooo000>0?ooo`030000003oool0oooo0=`0oooo00<000000?ooo`3oool0/03oool00`000000
oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`3L0?ooo`030000003oool0oooo0;00oooo
00<000000?ooo`3oool03@3oool000l0oooo00<000000?ooo`3oool0g03oool00`000000oooo0?oo
o`2^0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0oooo0=`0oooo00<00000
0?ooo`3oool0[P3oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo0?ooo`3M0?oo
o`030000003oool0oooo0:d0oooo00<000000?ooo`3oool03P3oool000l0oooo00<000000?ooo`3o
ool0g@3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000
003oool0oooo0=h0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0>0?ooo`00403o
ool00`000000oooo0?ooo`3M0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool03`3o
ool00100oooo00<000000?ooo`3oool0gP3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo00l0oooo000@0?ooo`030000003oool0oooo0=h0oooo00<000000?ooo`3oool0Z@3oool00`00
0000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`3N0?ooo`030000003oool0oooo0:P0
oooo00<000000?ooo`3oool0403oool00140oooo00<000000?ooo`3oool0gP3oool00`000000oooo
0?ooo`2X0?ooo`030000003oool0oooo0100oooo000A0?ooo`030000003oool0oooo0=l0oooo00<0
00000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`3O
0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0403oool00140oooo00<000000?oo
o`3oool0h03oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo0140oooo000B0?ooo`03
0000003oool0oooo0=l0oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`0A0?ooo`00
4P3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool0
4P3oool001<0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`2S0?ooo`030000003o
ool0oooo0180oooo000C0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0XP3oool0
0`000000oooo0?ooo`0B0?ooo`00503oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo
0:00oooo00<000000?ooo`3oool04`3oool001@0oooo00<000000?ooo`3oool0h03oool00`000000
oooo0?ooo`2P0?ooo`030000003oool0oooo01<0oooo000D0?ooo`030000003oool0oooo0>40oooo
00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`0C0?ooo`00503oool00`000000oooo0?oo
o`3Q0?ooo`030000003oool0oooo09l0oooo00<000000?ooo`3oool04`3oool001@0oooo00<00000
0?ooo`3oool0hP3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo01<0oooo000E0?oo
o`030000003oool0oooo0>40oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`0D0?oo
o`005@3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo09`0oooo00<000000?ooo`3o
ool0503oool001D0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?ooo`2K0?ooo`030000
003oool0oooo01D0oooo000F0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0VP3o
ool00`000000oooo0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0
oooo09X0oooo00<000000?ooo`3oool05@3oool001L0oooo00<000000?ooo`3oool0hP3oool00`00
0000oooo0?ooo`2H0?ooo`030000003oool0oooo01H0oooo000G0?ooo`030000003oool0oooo0>80
oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`0F0?ooo`005`3oool00`000000oooo
0?ooo`3S0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool05`3oool001P0oooo00<0
00000?ooo`3oool0hP3oool00`000000oooo0?ooo`2F0?ooo`030000003oool0oooo01L0oooo000H
0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`0H
0?ooo`006@3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo09@0oooo00<000000?oo
o`3oool0603oool001T0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`2C0?ooo`03
0000003oool0oooo01P0oooo000J0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0
TP3oool00`000000oooo0?ooo`0I0?ooo`006P3oool00`000000oooo0?ooo`3S0?ooo`030000003o
ool0oooo0940oooo00<000000?ooo`3oool06@3oool001X0oooo00<000000?ooo`3oool0h`3oool0
0`000000oooo0?ooo`2@0?ooo`030000003oool0oooo01X0oooo000K0?ooo`030000003oool0oooo
0><0oooo00<000000?ooo`3oool0S`3oool00`000000oooo0?ooo`0J0?ooo`006`3oool00`000000
oooo0?ooo`3S0?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3oool06`3oool001`0oooo
00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo01/0oooo
000L0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?oo
o`0K0?ooo`007@3oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo08/0oooo00<00000
0?ooo`3oool0703oool001d0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`2;0?oo
o`030000003oool0oooo01`0oooo000M0?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3o
ool0R@3oool00`000000oooo0?ooo`0M0?ooo`007P3oool00`000000oooo0?ooo`3S0?ooo`030000
003oool0oooo08T0oooo00<000000?ooo`3oool07@3oool001h0oooo00<000000?ooo`3oool0i03o
ool00`000000oooo0?ooo`270?ooo`030000003oool0oooo01h0oooo000O0?ooo`030000003oool0
oooo0><0oooo00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`0N0?ooo`007`3oool00`00
0000oooo0?ooo`3T0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool07P3oool00200
oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`250?ooo`030000003oool0oooo01l0
oooo000P0?ooo`030000003oool0oooo0>@0oooo00<000000?ooo`3oool0Q03oool00`000000oooo
0?ooo`0O0?ooo`008@3oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo08<0oooo00<0
00000?ooo`3oool0803oool00280oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`21
0?ooo`030000003oool0oooo0240oooo000R0?ooo`030000003oool0oooo0><0oooo00<000000?oo
o`3oool0P03oool00`000000oooo0?ooo`0R0?ooo`008`3oool00`000000oooo0?ooo`3S0?ooo`03
0000003oool0oooo07l0oooo00<000000?ooo`3oool08P3oool002<0oooo00<000000?ooo`3oool0
h`3oool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo02<0oooo000T0?ooo`030000003o
ool0oooo0><0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`0S0?ooo`00903oool0
0`000000oooo0?ooo`3S0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool08`3oool0
02D0oooo00<000000?ooo`3oool0h`3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo
02@0oooo000U0?ooo`030000003oool0oooo0><0oooo00<000000?ooo`3oool0N`3oool00`000000
oooo0?ooo`0T0?ooo`009P3oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo07T0oooo
00<000000?ooo`3oool09@3oool002L0oooo00<000000?ooo`3oool0hP3oool00`000000oooo0?oo
o`1h0?ooo`030000003oool0oooo02H0oooo000W0?ooo`030000003oool0oooo0><0oooo00<00000
0?ooo`3oool0MP3oool00`000000oooo0?ooo`0W0?ooo`00:03oool00`000000oooo0?ooo`3R0?oo
o`030000003oool0oooo07H0oooo00<000000?ooo`3oool09`3oool002P0oooo00<000000?ooo`3o
ool0h`3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo02P0oooo000Y0?ooo`030000
003oool0oooo0>80oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0X0?ooo`00:P3o
ool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0:@3o
ool002/0oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0
oooo02X0oooo000[0?ooo`030000003oool0oooo0>80oooo00<000000?ooo`3oool0K`3oool00`00
0000oooo0?ooo`0[0?ooo`00;03oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo06l0
oooo00<000000?ooo`3oool0:`3oool002`0oooo00<000000?ooo`3oool0hP3oool00`000000oooo
0?ooo`1^0?ooo`030000003oool0oooo02/0oooo000]0?ooo`030000003oool0oooo0>40oooo00<0
00000?ooo`3oool0K@3oool00`000000oooo0?ooo`0/0?ooo`00;@3oool00`000000oooo0?ooo`3R
0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0;03oool002h0oooo00<000000?oo
o`3oool0h@3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo02d0oooo000_0?ooo`03
0000003oool0oooo0>40oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0^0?ooo`00
;`3oool00`000000oooo0?ooo`3Q0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0
;`3oool00300oooo00<000000?ooo`3oool0h@3oool00`000000oooo0?ooo`1W0?ooo`030000003o
ool0oooo02l0oooo000a0?ooo`030000003oool0oooo0>00oooo00<000000?ooo`3oool0IP3oool0
0`000000oooo0?ooo`0`0?ooo`00<P3oool00`000000oooo0?ooo`3P0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool0<@3oool00380oooo00<000000?ooo`3oool0h@3oool00`000000
oooo0?ooo`1R0?ooo`030000003oool0oooo0380oooo000c0?ooo`030000003oool0oooo0>00oooo
00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`0b0?ooo`00=03oool00`000000oooo0?oo
o`3P0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0<`3oool003D0oooo00<00000
0?ooo`3oool0g`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo03@0oooo000f0?oo
o`030000003oool0oooo0=l0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`0e0?oo
o`00=`3oool00`000000oooo0?ooo`3N0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3o
ool0=P3oool003L0oooo00<000000?ooo`3oool0g`3oool00`000000oooo0?ooo`1J0?ooo`030000
003oool0oooo03L0oooo000h0?ooo`030000003oool0oooo0=h0oooo00<000000?ooo`3oool0FP3o
ool00`000000oooo0?ooo`0g0?ooo`00>@3oool00`000000oooo0?ooo`3N0?ooo`030000003oool0
oooo05P0oooo00<000000?ooo`3oool0>03oool003X0oooo00<000000?ooo`3oool0g@3oool00`00
0000oooo0?ooo`1G0?ooo`030000003oool0oooo03T0oooo000j0?ooo`030000003oool0oooo0=h0
oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0j0?ooo`00>`3oool00`000000oooo
0?ooo`3M0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0>P3oool003`0oooo00<0
00000?ooo`3oool0g@3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03/0oooo000m
0?ooo`030000003oool0oooo0=`0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0l
0?ooo`00?P3oool00`000000oooo0?ooo`3L0?ooo`030000003oool0oooo0500oooo00<000000?oo
o`3oool0?@3oool003l0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo03h0oooo00100?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool0
C@3oool00`000000oooo0?ooo`0o0?ooo`00@@3oool00`000000oooo0?ooo`3J0?ooo`030000003o
ool0oooo04`0oooo00<000000?ooo`3oool0@03oool00480oooo00<000000?ooo`3oool0fP3oool0
0`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0440oooo00130?ooo`030000003oool0oooo
0=T0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`120?ooo`00A03oool00`000000
oooo0?ooo`3I0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0@`3oool004D0oooo
00<000000?ooo`3oool0f03oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04@0oooo
00160?ooo`030000003oool0oooo0=P0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?oo
o`150?ooo`00A`3oool00`000000oooo0?ooo`3G0?ooo`030000003oool0oooo04<0oooo00<00000
0?ooo`3oool0AP3oool004P0oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`110?oo
o`030000003oool0oooo04L0oooo00190?ooo`030000003oool0oooo0=H0oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`180?ooo`00BP3oool00`000000oooo0?ooo`3F0?ooo`030000
003oool0oooo03h0oooo00<000000?ooo`3oool0B@3oool004/0oooo00<000000?ooo`3oool0e@3o
ool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo04X0oooo001<0?ooo`030000003oool0
oooo0=D0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1;0?ooo`00C@3oool00`00
0000oooo0?ooo`3D0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0C03oool004h0
oooo00<000000?ooo`3oool0e03oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo04d0
oooo001?0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`1>0?ooo`00D03oool00`000000oooo0?ooo`3C0?ooo`030000003oool0oooo03D0oooo00<0
00000?ooo`3oool0C`3oool00540oooo00<000000?ooo`3oool0dP3oool00`000000oooo0?ooo`0d
0?ooo`030000003oool0oooo0500oooo001B0?ooo`800000d`3oool00`000000oooo0?ooo`0a0?oo
o`800000D`3oool005@0oooo00<000000?ooo`3oool0d03oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo05<0oooo001E0?ooo`030000003oool0oooo0=00oooo00<000000?ooo`3oool0
;P3oool00`000000oooo0?ooo`1D0?ooo`00EP3oool200000=00oooo00<000000?ooo`3oool0;03o
ool2000005L0oooo001H0?ooo`030000003oool0oooo0<h0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`1G0?ooo`00F@3oool00`000000oooo0?ooo`3=0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool0F03oool005X0oooo00<000000?ooo`3oool0c@3oool00`000000
oooo0?ooo`0W0?ooo`030000003oool0oooo05T0oooo001K0?ooo`030000003oool0oooo0<`0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`1J0?ooo`00G03oool200000<d0oooo00<0
00000?ooo`3oool08`3oool2000005d0oooo001N0?ooo`030000003oool0oooo0<X0oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`1M0?ooo`00G`3oool00`000000oooo0?ooo`3:0?oo
o`030000003oool0oooo0200oooo00<000000?ooo`3oool0GP3oool00600oooo0P00003:0?ooo`03
0000003oool0oooo01h0oooo0P00001Q0?ooo`00HP3oool00`000000oooo0?ooo`380?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool0H@3oool006<0oooo00<000000?ooo`3oool0a`3o
ool00`000000oooo0?ooo`0J0?ooo`800000I03oool006@0oooo0`0000370?ooo`030000003oool0
oooo01L0oooo0P00001V0?ooo`00I`3oool200000<D0oooo00<000000?ooo`3oool05@3oool20000
06P0oooo001Y0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`1X0?ooo`00JP3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool0J@3oool006/0oooo0P0000330?ooo`030000003oool0oooo00l0oooo0P00
001/0?ooo`00K@3oool200000<40oooo0`00000=0?ooo`800000KP3oool006l0oooo00<000000?oo
o`3oool0_@3oool4000000`0oooo00<000000?ooo`3oool0KP3oool00700oooo0P00002n0?ooo`@0
00002@3oool200000740oooo001b0?ooo`800000_@3oool3000000L0oooo0P00001c0?ooo`00M03o
ool200000;/0oooo0`0000050?ooo`800000M@3oool007H0oooo0P00002j0?ooo`8000000`3oool2
000007L0oooo001h0?ooo`030000003oool0oooo0;P0oooo0P0000000`3oool000000?ooo`1h0?oo
o`00N@3oool200000;T0oooo0P00001j0?ooo`00N`3oool200000;D0oooo0P00001l0?ooo`00O@3o
ool200000;40oooo0P00001n0?ooo`00O`3oool300000:/0oooo0`0000200?ooo`00PP3oool20000
0:L0oooo0P0000230?ooo`00Q03oool200000:<0oooo0P0000250?ooo`00QP3oool3000009d0oooo
0`0000270?ooo`00R@3oool2000009T0oooo0P00002:0?ooo`00R`3oool2000009D0oooo0P00002<
0?ooo`00S@3oool3000008l0oooo0`00002>0?ooo`00T03oool2000008/0oooo0P00002A0?ooo`00
TP3oool3000008D0oooo0`00002C0?ooo`00U@3oool4000007`0oooo1@00002F0?ooo`00V@3oool3
000007L0oooo0P00002K0?ooo`00W03oool2000007<0oooo0P00002M0?ooo`00WP3oool5000006T0
oooo1@00002O0?ooo`00X`3oool500000640oooo0`00002T0?ooo`00Z03oool3000005T0oooo1@00
002W0?ooo`00Z`3oool5000004l0oooo1@00002/0?ooo`00/03oool5000004D0oooo1@00002a0?oo
o`00]@3oool7000003L0oooo1`00002f0?ooo`00_03oool;00000240oooo2`00002m0?ooo`00a`3o
oolQ00000<P0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo
003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.350004, -0.350004, \
0.00243622, 0.00243622}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[\(\(p3r //. momentumComponentRules\) /. sol\) //. 
        values]\)], "Input"],

Cell[BoxData[
    \(\(2\ Cos[theta3] + \@2\ \@\(\(-1\) + 2\ mratio\^2 + Cos[2\ \
theta3]\)\)\/\(2 + 2\ mratio\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Notice that if ",
  StyleBox["mratio<1", "InlineInput"],
  ", the range of ",
  StyleBox["theta3", "InlineInput"],
  " must be restricted."
}], "Text"],

Cell[BoxData[
    \(maxTheta[mratio_] := 
      theta3\/Degree /. \(Solve[\(-1\) + 2\ mratio\^2 + Cos[2\ theta3] == 0, 
            theta3]\)\[LeftDoubleBracket]2\[RightDoubleBracket]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(thmax = N[maxTheta[0.5]]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"ifun"\), \(\(:\)\(\ \)\), "\<\"Inverse functions are \
being used by \\!\\(Solve\\), so some solutions may not be found; use Reduce \
for complete solution information. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \(30.000000000000004`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotFinalMomenta[sol, Join[values, {mratio \[Rule] 0.5}], 
        0.9999\ thmax];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0317492 0.0147151 0.588604 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.18256 -0.0125 -5 -15.6875 ]
[.18256 -0.0125 5 0 ]
[.3413 -0.0125 -8 -15.6875 ]
[.3413 -0.0125 8 0 ]
[.50005 -0.0125 -8 -15.6875 ]
[.50005 -0.0125 8 0 ]
[.65879 -0.0125 -8 -15.6875 ]
[.65879 -0.0125 8 0 ]
[.81754 -0.0125 -8 -15.6875 ]
[.81754 -0.0125 8 0 ]
[.97629 -0.0125 -8 -15.6875 ]
[.97629 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.0313 -31.375 ]
[.5 -0.0125 21.0313 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .13244 -19 -7.84375 ]
[-0.0125 .13244 0 7.84375 ]
[-0.0125 .25016 -19 -7.84375 ]
[-0.0125 .25016 0 7.84375 ]
[-0.0125 .36788 -19 -7.84375 ]
[-0.0125 .36788 0 7.84375 ]
[-0.0125 .4856 -19 -7.84375 ]
[-0.0125 .4856 0 7.84375 ]
[-0.0125 .60332 -10 -7.84375 ]
[-0.0125 .60332 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.6875 -30.4375 ]
[-0.01875 .30902 -19 30.4375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -72.125 0 ]
[.5 .63053 72.125 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.18256 0 m
.18256 .61803 L
s
.3413 0 m
.3413 .61803 L
s
.50005 0 m
.50005 .61803 L
s
.65879 0 m
.65879 .61803 L
s
.81754 0 m
.81754 .61803 L
s
0 .13244 m
1 .13244 L
s
0 .25016 m
1 .25016 L
s
0 .36788 m
1 .36788 L
s
0 .4856 m
1 .4856 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.18256 0 m
.18256 .00625 L
s
gsave
.18256 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.3413 0 m
.3413 .00625 L
s
gsave
.3413 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.50005 0 m
.50005 .00625 L
s
gsave
.50005 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.65879 0 m
.65879 .00625 L
s
gsave
.65879 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.81754 0 m
.81754 .00625 L
s
gsave
.81754 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(25) show
1.000 setlinewidth
grestore
.97629 0 m
.97629 .00625 L
s
gsave
.97629 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 0 m
.05556 .00375 L
s
.08731 0 m
.08731 .00375 L
s
.11906 0 m
.11906 .00375 L
s
.15081 0 m
.15081 .00375 L
s
.2143 0 m
.2143 .00375 L
s
.24605 0 m
.24605 .00375 L
s
.2778 0 m
.2778 .00375 L
s
.30955 0 m
.30955 .00375 L
s
.37305 0 m
.37305 .00375 L
s
.4048 0 m
.4048 .00375 L
s
.43655 0 m
.43655 .00375 L
s
.4683 0 m
.4683 .00375 L
s
.5318 0 m
.5318 .00375 L
s
.56355 0 m
.56355 .00375 L
s
.5953 0 m
.5953 .00375 L
s
.62704 0 m
.62704 .00375 L
s
.69054 0 m
.69054 .00375 L
s
.72229 0 m
.72229 .00375 L
s
.75404 0 m
.75404 .00375 L
s
.78579 0 m
.78579 .00375 L
s
.84929 0 m
.84929 .00375 L
s
.88104 0 m
.88104 .00375 L
s
.91279 0 m
.91279 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.0313 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .13244 m
.00625 .13244 L
s
gsave
-0.0125 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .25016 m
.00625 .25016 L
s
gsave
-0.0125 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .36788 m
.00625 .36788 L
s
gsave
-0.0125 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .4856 m
.00625 .4856 L
s
gsave
-0.0125 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .60332 m
.00625 .60332 L
s
gsave
-0.0125 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04415 m
.00375 .04415 L
s
0 .07358 m
.00375 .07358 L
s
0 .10301 m
.00375 .10301 L
s
0 .16187 m
.00375 .16187 L
s
0 .1913 m
.00375 .1913 L
s
0 .22073 m
.00375 .22073 L
s
0 .27959 m
.00375 .27959 L
s
0 .30902 m
.00375 .30902 L
s
0 .33845 m
.00375 .33845 L
s
0 .39731 m
.00375 .39731 L
s
0 .42674 m
.00375 .42674 L
s
0 .45617 m
.00375 .45617 L
s
0 .51503 m
.00375 .51503 L
s
0 .54446 m
.00375 .54446 L
s
0 .57389 m
.00375 .57389 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -118.281 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
91.4375 11.8438 translate
90 rotate
-91.4375 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
69.000 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
73.813 15.813 moveto
(\\220) show
77.625 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
83.625 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
88.438 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(,) show
94.438 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
100.438 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
105.250 15.813 moveto
(\\220) show
109.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
115.063 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
119.875 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.18256 .61178 m
.18256 .61803 L
s
.3413 .61178 m
.3413 .61803 L
s
.50005 .61178 m
.50005 .61803 L
s
.65879 .61178 m
.65879 .61803 L
s
.81754 .61178 m
.81754 .61803 L
s
.97629 .61178 m
.97629 .61803 L
s
.125 Mabswid
.05556 .61428 m
.05556 .61803 L
s
.08731 .61428 m
.08731 .61803 L
s
.11906 .61428 m
.11906 .61803 L
s
.15081 .61428 m
.15081 .61803 L
s
.2143 .61428 m
.2143 .61803 L
s
.24605 .61428 m
.24605 .61803 L
s
.2778 .61428 m
.2778 .61803 L
s
.30955 .61428 m
.30955 .61803 L
s
.37305 .61428 m
.37305 .61803 L
s
.4048 .61428 m
.4048 .61803 L
s
.43655 .61428 m
.43655 .61803 L
s
.4683 .61428 m
.4683 .61803 L
s
.5318 .61428 m
.5318 .61803 L
s
.56355 .61428 m
.56355 .61803 L
s
.5953 .61428 m
.5953 .61803 L
s
.62704 .61428 m
.62704 .61803 L
s
.69054 .61428 m
.69054 .61803 L
s
.72229 .61428 m
.72229 .61803 L
s
.75404 .61428 m
.75404 .61803 L
s
.78579 .61428 m
.78579 .61803 L
s
.84929 .61428 m
.84929 .61803 L
s
.88104 .61428 m
.88104 .61803 L
s
.91279 .61428 m
.91279 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .13244 m
1 .13244 L
s
.99375 .25016 m
1 .25016 L
s
.99375 .36788 m
1 .36788 L
s
.99375 .4856 m
1 .4856 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .04415 m
1 .04415 L
s
.99625 .07358 m
1 .07358 L
s
.99625 .10301 m
1 .10301 L
s
.99625 .16187 m
1 .16187 L
s
.99625 .1913 m
1 .1913 L
s
.99625 .22073 m
1 .22073 L
s
.99625 .27959 m
1 .27959 L
s
.99625 .30902 m
1 .30902 L
s
.99625 .33845 m
1 .33845 L
s
.99625 .39731 m
1 .39731 L
s
.99625 .42674 m
1 .42674 L
s
.99625 .45617 m
1 .45617 L
s
.99625 .51503 m
1 .51503 L
s
.99625 .54446 m
1 .54446 L
s
.99625 .57389 m
1 .57389 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -133.125 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(scattered) show
108.250 15.813 moveto
(and) show
128.500 15.813 moveto
(recoil) show
158.563 15.813 moveto
(momenta) show
203.250 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.02499 .60332 L
.02605 .60332 L
.02729 .60332 L
.02846 .60332 L
.03053 .60331 L
.03279 .6033 L
.03527 .6033 L
.0379 .60328 L
.04262 .60326 L
.04749 .60322 L
.05205 .60318 L
.06244 .60305 L
.07305 .60289 L
.08274 .6027 L
.10458 .60216 L
.12357 .60155 L
.14429 .60073 L
.18493 .59869 L
.22406 .59615 L
.26565 .59284 L
.30571 .58903 L
.34426 .5848 L
.38527 .57966 L
.42475 .57408 L
.46273 .5681 L
.50315 .56105 L
.54206 .55357 L
.58342 .54483 L
.62326 .53557 L
.66159 .52583 L
.70238 .51446 L
.74164 .50241 L
.77939 .48961 L
.8196 .47437 L
.85828 .45767 L
.89942 .43663 L
.91995 .42412 L
.93905 .41042 L
.95701 .39408 L
.96218 .38815 L
.96706 .38151 L
.97144 .3739 L
.97264 .37128 L
.97391 .36799 L
.975 .3644 L
.97619 .3572 L
s
.02381 .01472 m
.06244 .02722 L
.10458 .04085 L
.14415 .05365 L
.18221 .06595 L
.22272 .07905 L
.26171 .09164 L
.30316 .10503 L
.34309 .11792 L
.3815 .13031 L
.42237 .1435 L
.46172 .1562 L
.49955 .16843 L
.53984 .18146 L
.57861 .19403 L
.61984 .20746 L
.65954 .22046 L
.69774 .23307 L
.73838 .24665 L
.77751 .25997 L
.81909 .2745 L
.85916 .28913 L
.89771 .30422 L
.91726 .31255 L
.93871 .32266 L
.94807 .32763 L
.9581 .33364 L
.96296 .337 L
.96756 .34066 L
.97169 .34473 L
.97283 .3461 L
.97403 .3478 L
.97506 .34963 L
.97619 .35321 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00j@3oool01@3OW4P00000000000000000M;oo0140oooo00<0oooO09a8
000000000`0000001018W=l0oooo0?ooo`3OW4P30000000307Boo`3oool0oooo0:40oooo003Y0?oo
o`030=nLB00007@0_ooo01<0oooo00<0_g@00018W03Oool00`3oool00`3OW9`0_ooo0?ooo`020?oo
o`030?oo_`1d0000M;oo0:<0oooo003M0?ooo`040=nLB00000000000001d_`P0oooo00<0gia80018
W03Oool04`3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0080oooo00<0oono07@0
001d_ol0X`3oool00>00oooo00<0okmd001d_`3oool01P3oool00`3OW4P004RL0=ooo`0D0?ooo`03
0;md00000000000000<0000000L0B9cO0?ooo`3oool0oooo0?oo_`1d0000M;oo0:<0oooo003E0?oo
o`050?oo_`1d000000000000B02Lgol01P3oool00`3ookl0M01d0;ooo`060?ooo`030=nLB000B9`0
gooo00<0oooo00P0oooO09a80000000000000000B018B00004RL0=ooo`80oooo00P0_g@000000000
0000001d0;ooo`3oool0oono07@000@0000000<0001d0;ooo`3oool00P3oool00`3ookl0M00007Bo
o`2S0?ooo`00e@3oool01P2oM00007Bo0?ooo`3ogi`0B01809cOo`D0oooo00<0gia80000M02oool0
1P3oool00`3OW4P004RL0=ooo`020?ooo`0D0?oog`2LB000000007Boo`3oool0oono07@000000000
001d0;ooo`2oM0000000001d_`3oool0omnL04P0M02oool0oooo0;mdB02Lgol80?ooo`030?oo_`1d
0000M;oo0:<0oooo003D0?ooo`030?oog`2LB000M;oo0080oooo00<0gia80018W03Oool00P3oool0
103ookl0M00000000018W=l70?ooo`030=nLB000B9`0gooo0080oooo00<0omnL04P00018W=l00P3o
ool01`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool01@3ogi`0M9cO0?ooo`3ookl0M000
0080000000<007Bo0?ooo`3oool00`3oool00`3ookl0M00007Boo`2S0?ooo`00e03oool00`3ookl0
M00007Boo`020?ooo`030?oo_`1d0000M;oo00<0oooo00<0oooO09bL_`3oool01`3oool00`3OW4P0
04RL0=ooo`020?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?noM00004P0
W=oo00H0oooo00H0gia80000M02oool0oono07@00018W=l40?ooo`030?oo_`1d0000M;oo0:<0oooo
003D0?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool00`3oool00`3o_g@0B9cO0?oo
o`060?ooo`030=nLB000B9`0gooo0080oooo00<0okmd0000001d_ol00P3oool01`3ooml0W4P0001d
_`3oool0okmd0000M02oool01@3oool01`3ookl0M00004RLg`3oool0oooO09a80000M;l0103oool0
0`3ookl0M00007Boo`2S0?ooo`00e03oool00`3ooml0W4P0001d_`030?ooo`080=nLB00007@0_ooo
0?oOW01dW=l0oooo0=nLB000M;l70?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@00018W=l0
0P3oool01P3ooml0W4P0001d_`3oool0omnL04P000D0000000P0B9cO0?oo_`1d0000B9cO0?ooo`3o
oml0W4P004RLg`@0oooo00<0oono07@0001d_ol0X`3oool00=D0oooo00<0_g@00000000000001000
00001P1d_ol0oooo0=nLB0000000001809cOo`L0oooo00<0gia80018W03Oool00`3oool01`3o_g@0
001809cOo`3oool0okmd00000000M;l00P3oool0302oM000B9cO0?ooo`3oool0gia80000B02Lgol0
oooo0;md0000M;l0oooo0?noM080000000L004RL0=ooo`3oool0oooo0?oo_`1d0000M;oo0:<0oooo
003E0?ooo`030=nLB00007@0_ooo0080oooo00<0oono07@00018W=l0303oool00`3OW4P004RL0=oo
o`040?ooo`030?noM000000000000080000000D007Bo0?ooo`3oool0oooo0=nLB002000000060000
B02Lgol0oooo0?ooo`3ooml0W4P00P0000000`1d_ol0oooo0?ooo`030?ooo`030?oo_`1d0000M;oo
0:<0oooo003E0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@0001d_`3oool02`3oool00`3OW4P0
04RL0=ooo`070?ooo`030?oog`2LB00007Bo01@0oooo00<0oono07@0001d_ol0X`3oool00=H0oooo
00<0_g@00018W03Oool00P3oool00`3OW4P007Bo0?ooo`0;0?ooo`030=nLB000B9`0gooo00L0oooo
00<0oooO09a80000M;l0503oool00`3ookl0M00007Boo`2S0?ooo`00eP3oool00`3ookl0M00007Bo
o`020?ooo`030=nLB000M;l0oooo00/0oooo00<0gia80018W03Oool01`3oool00`3ooml0W4P0001d
_`0D0?ooo`030?oo_`1d0000M;oo0:<0oooo003G0?ooo`060?noM00004P0W=oo0?ooo`2oM000M;oo
303oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d00000000001d_a@0oooo00<0oono07@0001d
_ol0X`3oool00=P0oooo00D0omnL04P00000000004RL0=ooo`0<0?ooo`050=nLB000000000000000
001d_ol01@3oool00`3ooml0W4P0001d_`0B0?ooo`050=nLB000000000000000001d_ol0X`3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool005h0oooo00D0omnL04P000000000001d0;ooo`0Z0?ooo`030?oog`2LB0000000
0080000000<0M;oo0?ooo`3oool09@3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0
omnL04P000000000001d0;ooo`0R0?ooo`090;md00000000000000000000B9`0gooo0?ooo`3ooml0
W4P000<0000000<0M;oo0?ooo`3oool0803oool00`2oM0000000000000040000000904RLg`3oool0
oooo0?ooo`3ogi`0B0000000000007@0_ooo0200oooo00<0_g@0000000000000100000001018W=l0
oooo0?oog`2LB0030000000307Boo`3oool0oooo01l0oooo00<0oono07@0000000000P0000000`1d
_ol0oooo0?ooo`030?ooo`050?oOW018000000000000M02oool09`3oool005d0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo02T0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?oo
o`0V0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0280
oooo00<0omnL04P0001d_ol00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0240
oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000
M02ooolP0?ooo`030=nLB0000000000000<0000000D004RL0=ooo`3ooml0W4Qd0;ooo`020?ooo`03
0;md001d_ol0oooo01l0oooo00<0omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0V0?ooo`00G@3oool00`2oM00004RL0=ooo`020?oo
o`030;md0000M;l0oooo02d0oooo00<0oooO09a8M02oool09P3oool00`3ogi`0B00007Boo`040?oo
o`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool08@3oool00`3ogi`0B00007Boo`080?oo
o`030?oog`2LB7@0_ooo0280oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM000
04RL0=ooo`80oooo00<0_g@0001d_`3oool0803oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02L
gol0oooo00@0oooo00<0oooO09a8M02oool0903oool00`3ookl0M01809cOo`020?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool09@3oool005`0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`0^0?ooo`030;md001d_ol0oooo02D0oooo00<0omnL04P0001d_ol00`3oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0240oooo00<0omnL04P0001d_ol02@3oool0
0`2oM000M;oo0?ooo`0R0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`0Q0?ooo`030=nLB01d_ol0oooo00T0oooo00<0_g@007Boo`3oool08`3oool0
1`3ookl0M00007Boo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0U0?ooo`00G03o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02d0oooo00<0oooO09a80018W=l09P3o
ool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08@3o
ool00`3ogi`0B00007Boo`080?ooo`030?oog`2LB000B9cO02@0oooo00<0_gA809cOo`3oool00P3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0280oooo00<0_gA809cOo`3oool01`3o
ool00`3ooml0W4P004RLg`0T0?ooo`070?noM0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo02D0oooo001L0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
;@3oool00`3OW4P0000007Boo`0V0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0Q0?ooo`030?oOW0180000M;oo00P0oooo00<0gia80000001d_ol0
903oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
8P3oool00`3ooml0W4Qd0;ooo`070?ooo`030=nLB0000000M;oo02<0oooo00P0omnL04P0000004P0
W=oo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol09@3oool005`0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`0Y0?ooo`070=oOg`3oool0oono07@000000000001d0;oo
o`0V0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0Q0?ooo`030?oOW0180000M;oo00@0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo02D0
oooo00L0okmd001d_`3oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08`3oool0
0`3o_g@007Bo0?ooo`020?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0Q0?ooo`050?no
M000000000000000B02Lgol00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02D0
oooo001L0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0:@3oool01P3OW4P00000
000000000000001809cOobL0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo0240oooo00<0omnL04P0001d_ol0103oool01P3OW4P00000000000000000
001809cOobH0oooo00L0oono07@0B02Lgol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol08`3oool00`3ookl0M01809cOo`020?ooo`060=nLB000000000000000000004P0W=oo903oool0
0`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol09@3oool0
05`0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0Y0?ooo`040?oo_`1d00000000
04RLgbT0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo0240oooo00<0omnL04P0001d_ol0103oool0103ookl0M00000000018W=lX0?ooo`070?oo_`1d
0000B9cO0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo02<0oooo00<0oono07@00018
W=l00P3oool0103ookl0M00000000018W=lW0?ooo`070;md0018W=l0oooo0?ooo`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo02D0oooo001M0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool0:P3oool00`2oM000M;oo0?ooo`0Y0?ooo`030?oOW0180000M;oo00@0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`0Q0?ooo`030?oOW0180000M;oo00D0oooo00<0_g@0
07Boo`3oool08`3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0080oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`0N0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL
04P00018W=l00`3oool00`2oM000M;oo0?ooo`0R0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia8
0000M02oool00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo02D0oooo001M0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0[0?ooo`060?oOW018000000000000000004P0
W=oo903oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo0200oooo00D0oooO09a800000000000007Boo`050?ooo`060?oOW0180000000000000000
04P0W=oo803oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool07`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
0`3oool01P3ogi`0B000000000000000001809cOob00oooo00H0okmd04RLg`3oool0oono07@00000
M;l30?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0V0?ooo`00GP3oool01@3ookl0M000
0000000007@0_ooo02d0oooo00D0_g@0000000000000000004RLg`0V0?ooo`030?noM0000000M;oo
00D0oooo00D0oono07@000000000001d0;ooo`0S0?ooo`030?noM0000000M;oo00H0oooo00D0_g@0
000000000000000004RLg`0Q0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0
M0000000000007@0_ooo0240oooo00H0oooO09a80000000000000000M02oool50?ooo`050;md0000
0000000000000018W=l08@3oool01@3OW4P000000000000007@0_ooo00@0oooo00D0oono07@00000
0000001d0;ooo`0W0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00C03oool0
1@3ogi`0B0000000000007@0_ooo00T0ooooo`00000a000002H0oooo001;0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0T0?ooo`00B`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo02@0oooo001:0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool0
0`000000oooo0?ooo`0T0?ooo`00BP3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00L0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
03@0oooo00<000000?ooo`3oool0903oool004X0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`070?ooo`<00000103oool2000002h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0b0?ooo`<000009P3oool004X0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`070?ooo`030000003oool0oooo00H0oooo0`00000[0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00BP3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool02@3oool3000002P0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001:
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`0<
0?ooo`<000009@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<00000
0?ooo`3oool0903oool004/0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?oo
o`030000003oool0oooo00l0oooo0`00000R0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00B`3oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool0203oool00`000000oooo0?ooo`0B0?ooo`<000007`3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool004`0oooo00D0oono07@0
00000000001d0;ooo`090?ooo`030000003oool0oooo01D0oooo0`00000L0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`00
0000oooo0?ooo`0H0?ooo`<000006@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
03@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool06`3oool3000001H0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0
oooo001J0?ooo`8000007`3oool3000001<0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0c0?ooo`8000009P3oool005X0oooo00<000000?ooo`3oool08@3oool300000100oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J
0?ooo`030000003oool0oooo02@0oooo0`00000=0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0W0?oo
o`<000002P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?oo
o`3oool0903oool005X0oooo00<000000?ooo`3oool0:P3oool3000000L0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000
003oool0oooo02d0oooo0`0000040?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0`0?ooo`<0000000<0
oooo0000O`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J
0?ooo`030000003oool0oooo03<0oooo0`00000^0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<00000
0?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo000000030000
02X0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool2000003D0
oooo00<0001o0?ooo`3oool00`3oool4000002H0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0<`3oool2000002H0
oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool01`3oool400000280oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?oo
o`0d0?ooo`030000O`3oool0oooo00/0oooo0`00000O0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0>
0?ooo`<00000703oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J0?oo
o`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool04@3oool3000001T0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo01@0oooo0`00000F0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3o
ool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0G0?ooo`<00000
4`3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003o
ool0oooo03@0oooo00<0001o0?ooo`3oool06P3oool4000000l0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03o
ool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0
oooo01h0oooo0`00000<0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool0
05X0oooo0P00000e0?ooo`030000O`3oool0oooo0240oooo0`0000090?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
03<0oooo0P00000V0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02@0
oooo0`0000060?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool005X0oooo
00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0W0?ooo`<000000`3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0
001o0?ooo`3oool0:P3oool3000000030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903o
ool003h0oooo00D0omnL04P000000000001d0;ooo`030?ooo`050?noM0000000M;oo0?ooo`2oM000
1P0000000`18W=l0oooo0?ooo`060?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0
;@3oool3000002d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool003d0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool00`3OW4P000000000000300000003
0018W03Oool0oooo00H0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo00<00000:P3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00?@3oool00`2oM00004RL
0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo
0?ooo`060?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`030?ooo`800000:03oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00?03oool00`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool02P3oool00`000000oooo0?oo
o`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool01@3oool3000002D0oooo00<0
001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool0903oool003`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`09
0?ooo`030;mdB02Lgol0oooo00T0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo00P0oooo0`00000R0?ooo`030000O`3oool0oooo02h0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000l0?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`3ooml0W4Qd0;ooo`090?oo
o`<00000LP0007l300000;D0001o0`00000V0?ooo`00?03oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00X0oooo00<0okmd001d_`3oool0203oool00`000000oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool03P3oool4000001/0oooo00<0001o0?ooo`3o
ool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3o
ool0903oool003`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?oo
_`1d04P0W=oo00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo0180oooo0`00000H0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000l0?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`3ookl0M00004RLg`080?ooo`030000003o
ool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0E0?ooo`<000005@3o
ool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03o
ool00`000000oooo0?ooo`0T0?ooo`00?@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0
oooo00D0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`080?ooo`030000003oool0
oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0H0?ooo`<000004P3oool0
0`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool0
0`000000oooo0?ooo`0T0?ooo`00?@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01P3o
ool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo203oool00`000000oooo0?ooo`0d0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool06`3oool3000000l0oooo00<0001o0?oo
o`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?oo
o`3oool0903oool003h0oooo00D0oono07@000000000001d0;ooo`080?ooo`060?oog`2LB0000000
0000000007@0_ooo2@3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool07P3oool3000000`0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool005X0oooo00<00000
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0240oooo0`000009
0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0T0?ooo`<000001P3oool00`0007l0oooo0?ooo`0^0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?oo
o`00FP3oool2000003D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0W0?ooo`80
0000103oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0<`3oool2000002H0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0Y0?ooo`<0000000<0oooo0000O`3oool0;`3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool005X0oooo00<0
00000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02`0oooo0`00
000_0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?oo
o`0T0?ooo`00:P3oool00`3ooml0W9bo0?ooo`080?ooo`030;nLW03Oool0oooo0280oooo00<00000
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`0000000P00000/0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool0
0`000000oooo0?ooo`0T0?ooo`00:P3oool00`3ogi`0B00000000008000000030018W03Oool0oooo
0280oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool00P3oool4000002P0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000Z0?ooo`030=nLB0000000000000P0
000000<004RL0=ooo`3oool08P3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`060?ooo`<000009@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool002/0oooo
00<0_gAd0;ooo`3oool00P3oool00`3ooml0W9bo0?ooo`020?ooo`030;nLW03Oool0oooo0280oooo
00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool02@3oool300000280oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000Z0?ooo`030=nLB000M;l0oooo00@0oooo00<0
gibL0;ooo`3oool09P3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0<0?ooo`@000007P3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool002X0oooo00<0_g@0
0000001d_ol00`3oool00`3ogi`0B4RL0=ooo`0W0?ooo`030000003oool0oooo03@0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0100oooo1000000J
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T
0?ooo`00:P3oool00`3o_g@00000000000040000000307Boo`3oool0oooo02H0oooo0P00000e0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0D0?oo
o`<000005`3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03<0oooo0P00000V0?oo
o`00:`3oool01P3ogi`0B000000000000000001809cOobT0oooo00<000000?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool05`3oool3
000001@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo01X0oooo1000000@0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00;03oool00`3OW7@0
W=oo0?ooo`030?ooo`030?oOW018B9`0gooo02D0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool07P3oool3000000d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0
oooo000[0?ooo`030?noM018W=l0oooo00D0oooo00<0_gAd0;ooo`3oool0903oool00`000000oooo
0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0Q0?ooo`<000002P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo
00<000000?ooo`3oool0903oool002/0oooo00H0gia807Boo`3oool0okmd0018W03Oool20?ooo`03
0=no_`3Oool0oooo02@0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0903oool3000000L0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000[0?ooo`090=nL
B000M;l0oono07AdW02LB000B9cO0?ooo`3o_g@0B9cO02H0oooo00<000000?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool09`3oool3
000000@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02@0oooo000[0?ooo`080?oo_`1d0000001d0;ooo`3ogi`0B00000000000M;lW0?ooo`030000
003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02X0oooo0`0000000`3oool0001o0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo
00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool3000002h0oooo00<0
001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool2000003D0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o
0?ooo`0000000P00000[0?ooo`030000O`3oool0oooo03<0oooo0P00000V0?ooo`00<`3oool0103o
oml0W4P00000M02ooolS0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool00P3oool3
000002P0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00<03oool01@3o
_g@000000000000004P0W=oo02D0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`05
0?ooo`<000009@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000]0?oo
o`040;md000000000000001d_`d0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM000
0000M;oo00D0oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`0d0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000
O`3oool0oooo00P0oooo0`00000R0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0
903oool002T0oooo00D0omnL04P000000000000007Boo`0?0?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0oooO09a80000B9`0gooo203oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo00/0oooo0`00000O0?ooo`030000O`3oool0
oooo03@0oooo00<000000?ooo`3oool0903oool002H0oooo00D0gia800000000000004RL0=ooo`0B
0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02@3oool0103ooml0W4P00018W03O
ool80?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool03P3oool3000001`0oooo00<0
001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00903oool00`3ogi`0B00004RLg`0E
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`2oM000000000000005
0000000307Boo`3oool0oooo00D0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0A
0?ooo`<000006@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000Z0?oo
o`030?oog`2LW;l0oooo00P0oooo00<0_ibL0=ooo`3oool0103oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00D0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`P0oooo
00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0D0?ooo`<000005P3oool00`0007l0oooo
0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000Z0?ooo`030?oOW0180000000000P0000000<0
04RL0=ooo`3oool0103oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0
gia807Boo`3oool0oooO09a80000B9`0gooo00P0oooo0`00003O0000O`<00000B00007l3000002H0
oooo000Z0?ooo`030=nLB0000000000000P0000000<004RL0=ooo`3oool0103oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo
00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0J0?ooo`<00000403oool00`00
07l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000[0?ooo`030;mdM02oool0oooo0080
oooo00<0oooO09bL_`3oool00P3oool00`2oW9`0gooo0?ooo`040?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01`3oool01P3ogi`0B7Bo0?oog`2LB00004RL0=ooo`P0oooo00<00000
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0M0?ooo`<000003@3oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo02@0oooo000Z0?ooo`030=nLB000M;l0oooo00@0oooo00<0gibL0;oo
o`3oool0203oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00P0oooo00D0gia807Bo
g`2LB00004RL0=ooo`080?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0803oool3
000000X0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00:P3oool00`2o
M000000007Boo`030?ooo`030?oOW018B9`0gooo00X0oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`080?ooo`050?oog`2LB4P0B4P00018W03Oool0203oool00`000000oooo0?ooo`0d
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^
0?ooo`030000O`3oool0oooo02<0oooo0P0000080?ooo`030000O`3oool0oooo03@0oooo00<00000
0?ooo`3oool0903oool002X0oooo00<0okmd000000000000100000000`1d_ol0oooo0?ooo`090?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`040?oOW018000004RL0=ooo`P0oooo
00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0U0?ooo`8000001P3oool00`0007l0oooo
0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000[0?ooo`060?oOW018000000000000000004P0
W=oo3@3oool01@3ookl0M0000000000007@0_ooo00`0oooo00<0gia80018W03Oool0203oool00`00
0000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02L0oooo0`0000030?ooo`030000O`3oool0oooo
03@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0Z0?ooo`80000000<0oooo0000O`3oool0=@3oool00`000000oooo0?ooo`0T0?ooo`00
FP3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02`0oooo0`00000e0?ooo`030000
003oool0oooo02@0oooo000[0?ooo`030?oog`2LM9`0gooo00D0oooo00<0gibL0;ooo`3oool0903o
ool2000003D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0000000200000340oooo0P00
000V0?ooo`00:`3oool00`3o_g@0000000000005000000030000M02oool0oooo02@0oooo00<00000
0?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0080oooo1000000^
0?ooo`030000003oool0oooo02@0oooo000c0?ooo`030=nLW02oool0oooo02@0oooo00<000000?oo
o`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo00H0oooo0`00000[0?oo
o`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`090?ooo`<00000:03oool00`000000oooo0?ooo`0T0?ooo`00
FP3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
303oool2000002H0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo00h0oooo0P00000T0?ooo`030000003o
ool0oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0@0?ooo`8000008P3oool00`000000oooo0?ooo`0T0?ooo`00;`3oool01P3o
oml0W4P00018W03Oool0oooO09bL_bD0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo0180oooo0P00000P0?ooo`030000003oool0oooo02@0oooo000_
0?ooo`060?oo_`1d000004RL0=ooo`2oM4P0W=oo9@3oool00`000000oooo0?ooo`0d0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0503oool3000001d0oooo00<000000?ooo`3oool0
903oool00300oooo00@0gia8000000000000M;oo9P3oool2000003D0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0G0?ooo`<000006@3oool2000002H0oooo001J0?ooo`030000003o
ool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0J0?ooo`<000005`3o
ool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool07@3oool2000001D0oooo00<000000?ooo`3oool0903oool0
05X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
01l0oooo0P00000C0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo03@0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0Q0?ooo`8000004@3oool00`000000
oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool08`3oool2000000l0oooo00<000000?ooo`3oool0903oool002X0oooo00<0
oooO09bL_`3oool0203oool00`2oW9`0gooo0?ooo`0R0?ooo`030000003oool0oooo03@0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool0903oool002X0oooo00<0omnL04P000000000200000000`00B9`0gooo0?ooo`0R
0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0V
0?ooo`800000303oool00`000000oooo0?ooo`0T0?ooo`00:P3oool00`3OW4P00000000000080000
00030018W03Oool0oooo0280oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0T0?oo
o`00:`3oool00`2oM7@0_ooo0?ooo`020?ooo`030?oog`2LW;l0oooo0080oooo00<0_ibL0=ooo`3o
ool08P3oool2000003D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0Y0?ooo`03
0000003oool0oooo00L0oooo0P00000V0?ooo`00:P3oool00`3OW4P007Bo0?ooo`040?ooo`030=nL
W02oool0oooo02H0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0T0?ooo`00:P3o
ool00`2oM000000007Boo`030?ooo`030?oOW018B9`0gooo02L0oooo00<000000?ooo`3oool0=03o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02/0oooo0P0000070?ooo`030000003o
ool0oooo02@0oooo000Z0?ooo`030?noM0000000000000@0000000<0M;oo0?ooo`3oool09P3oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;03oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo02@0oooo000[0?ooo`060?oOW01800000000
0000000004P0W=oo3@3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol0
0`3oool01@3ogi`0B000000000000000M;oo00T0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0T0?ooo`00?@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o
_g@0000007Boo`020?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00007Boo`080?ooo`030000003o
ool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0903oool002l0oooo00@0omnL04P00000B9`0gooo2P3o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0oooO09a80000M;l00`3o
ool00`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0903o
ool002h0oooo00D0gia80018W03Ogi`0B4RL0=ooo`090?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<0
00000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0T0?ooo`00;@3oool01P2oM000B9cO0?ooo`3o
gi`0B4RL0=ooo`T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0d0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo02@0oooo000/0?ooo`070=nLB01d_ol0oooo0?ooo`3ogi`0B4RL0=ooo`090?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool00`3ogi`0B00007Boo`030?ooo`03
0;md000004P0W=oo00L0oooo0`00003o0000ObD0001o00<000000000O`0007l00`0007l3000002H0
oooo000[0?ooo`030=nLB0000000000000D0000000<0001d0;ooo`3oool01P3oool00`3ogi`0B000
07Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o_g@00000
001d_`P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0T0?ooo`00<03oool00`3o
gi`0B4RL0=ooo`090?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`2o
M00004RL0;nLB002000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0=03oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02`0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0T0?ooo`00?03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0
oooo00<0omnL04P0001d_ol0303oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02@0
oooo000c0?ooo`040?oog`2LB000001d0;ooo`H0oooo00<0gia80018W03Oool00P3oool00`2oM000
07Bo0?ooo`070?ooo`030=nLB00004P0W=oo00/0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0T0?ooo`00<03oool01@3o_g@000000000000004P0W=oo00P0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00T0oooo00<0okmd00000000M;l02P3oool00`000000oooo0?ooo`0d
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0:@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo02@0oooo000]0?ooo`040;md000000000000001d_`d0oooo00D0oono
07@000000000001d0;ooo`0;0?ooo`050?oog`2LB00000000000M02oool01`3oool00`000000oooo
0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0:@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo02@0oooo000Y0?ooo`050?oOW018000000000000001d_ol0
;03oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
:03oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02@0oooo000V0?ooo`050=nLB000
000000000018W03Oool0;`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo02@0oooo
000T0?ooo`030?oOW0180000B9cO03<0oooo0P00000e0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool09P3oool00`000000oooo0?ooo`0:0?ooo`8000009P3oool002X0oooo00<0oooO
09bL_`3oool0203oool00`2oW9`0gooo0?ooo`0R0?ooo`030000003oool0oooo03@0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool0903oool002X0oooo00<0omnL04P000000000200000000`00B9`0gooo0?ooo`0R0?oo
o`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0T0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0903oool002X0oooo00<0gia800000000
0000200000000`00B9`0gooo0?ooo`0R0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0S0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0903oool002/0oooo00<0_gAd0;ooo`3oool00P3oool00`3ooml0W9bo0?ooo`020?ooo`030;nL
W03Oool0oooo0280oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo0280oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0T0?ooo`00:P3o
ool00`3OW4P007Bo0?ooo`040?ooo`030=nLW02oool0oooo02H0oooo00<000000?ooo`3oool0=03o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0200oooo0P00000B0?ooo`030000003o
ool0oooo02@0oooo000Z0?ooo`030;md00000000M;oo00<0oooo00<0omnL04Q8W03Oool09`3oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool07`3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo02@0oooo000Z0?ooo`030?noM00000000000
00@0000000<0M;oo0?ooo`3oool09P3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool07@3oool2000001D0oooo00<000000?ooo`3oool0903oool002/0
oooo00H0omnL04P00000000000000000B02LgolY0?ooo`030000003oool0oooo03@0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0L0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool0903oool005X0oooo0P00000e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool06P3oool2000001L0oooo0P00000V0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo02@0oooo000[0?ooo`030?oog`2LM9`0gooo00D0oooo00<0gibL0;ooo`3oool0
903oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
5`3oool2000001/0oooo00<000000?ooo`3oool0903oool002/0oooo00<0okmd0000000000001@00
00000`0007@0_ooo0?ooo`0T0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0E0?ooo`8000007@3oool00`000000oooo0?ooo`0T0?ooo`00<`3oool0
0`3OW9`0_ooo0?ooo`0T0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0C0?ooo`8000007`3oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`00
0000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool04@3oool20000
0240oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo00l0oooo0P00000S0?ooo`030000003oool0oooo02@0
oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0=0?ooo`8000009@3oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?oo
o`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool02`3oool2000002L0oooo00<0
00000?ooo`3oool0903oool005X0oooo0P00000e0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool02@3oool2000002P0oooo0P00000V0?ooo`00FP3oool00`000000oooo0?ooo`0d0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool01`3oool2000002/0oooo00<000000?oo
o`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo00D0oooo0P00000]0?ooo`030000003oool0oooo02@0oooo001J0?ooo`030000
003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`030?ooo`800000
;`3oool00`000000oooo0?ooo`0T0?ooo`00?P3oool01@3ogi`0B0000000000007@0_ooo00<0oooo
00<0okmd0000001d_ol00`3oool01@3OW4P00000000000000000B9cO00T0oooo00<000000?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`040000O`3oool0oooo0?ooo`800000<@3oool00`00
0000oooo0?ooo`0T0?ooo`00?@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool0
0`3o_g@0000007Boo`020?ooo`070=nLB0000000M;oo0?ooo`3ooml0W4P004RLg`080?ooo`030000
003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool0100007l0oooo00000000000c0?ooo`03
0000003oool0oooo02@0oooo000m0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0
1@3oool00`3ooml0W4P0001d_`030?ooo`030?noM000B9`0gooo00L0oooo00<000000?ooo`3oool0
=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;P3oool00`0007l0oooo0?ooo`0]0?ooo`800000=@3oool00`000000oooo0?ooo`0T0?ooo`00?03o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@0001d_`3oool00P3o
ool00`2oM00004RL0=ooo`070?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0:P3o
ool3000000030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool003`0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070?oOW01807@0_ooo0?ooo`2oM0000000
07Boo`080?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool09P3oool4000000<0oooo
00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00?03oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00L0oooo00D0oono07A8B0180000000004RLg`090?ooo`<00000
j`0007l3000003`0001o0`00000V0?ooo`00?03oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00L0oooo00@0omnL04P00000000007Bo2P3oool00`000000oooo0?ooo`0d0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000
O`3oool0oooo0200oooo0`00000:0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0
903oool003`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060?oOW018
0000001d0;ooo`3o_g@007Bo2@3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo01d0
oooo0`00000=0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool003`0oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030;md000007@0_ooo0080oooo
00<0_g@004RLg`3oool01`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo01/0oooo
0P00000@0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool003d0oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0gooo0080oooo00<0
gia8001d_`3oool01`3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo01P0oooo0`00
000B0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool003d0oooo00L0oono
07@0001d_ol0oooo0?noM00007@0_ooo00L0oooo00L0okmd0000001d_ol0oooo0?oo_`1d0000M;oo
00P0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0E0?ooo`<000005@3oool00`00
07l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo000n0?ooo`050?oo_`1d000000000000
M02oool02@3oool01@3o_g@00000000000000000B9cO00T0oooo00<000000?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool0
0`0007l0oooo0?ooo`0B0?ooo`<00000603oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool03`3oool30000
01/0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000
oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0^0?ooo`030000O`3oool0oooo00`0oooo0`00000N0?ooo`030000O`3oool0oooo03@0
oooo00<000000?ooo`3oool0903oool005X0oooo0P00000e0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo00T0oooo
0`00000Q0?ooo`030000O`3oool0oooo03<0oooo0P00000V0?ooo`00FP3oool00`000000oooo0?oo
o`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0^0?ooo`030000O`3oool0oooo00D0oooo1000000T0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool0100007l0oooo0?ooo`3o
ool4000002P0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool0
0`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0^0?ooo`@00000;03oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02X0oooo100000000`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool09P3oool4000000@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`03
0000003oool0oooo02@0oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0280oooo100000080?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3o
ool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0N0?ooo`@00000303oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo03@0oooo00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool06P3oool4
00000100oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02@0oooo001J0?ooo`800000=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool05@3oool5000001@0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0c0?ooo`8000009P3oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool04@3oool4000001T0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0
oooo001J0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo00d0oooo1000000M0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?oo
o`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`080?ooo`D000008@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool00P3oool6000002H0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J0?oo
o`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0/0?oo
o`H00000;03oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?oo
o`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02H0oooo1P0000000`3oool0001o0?ooo`0_0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo
0?ooo`0d0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0803oool6000000L0oooo
00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo
00<000000?ooo`3oool0903oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo01X0oooo1P00000=0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J
0?ooo`800000=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo01D0oooo1@00000C
0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0c
0?ooo`8000009P3oool005X0oooo00<000000?ooo`3oool0=03oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo00l0oooo1P00000H0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001J0?ooo`03
0000003oool0oooo03@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`050?ooo`X0
00007P3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0=03oool00`000000oooo0?ooo`0T0?ooo`00FP3oool00`000000oooo0?ooo`0d0?ooo`030000
O`3oool0oooo02T0oooo3000000X0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001<0?ooo`060;md0000
0000000000000000B9`0gooo203oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo0200
oooo2@0000040?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool004d0oooo
00<0omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo01X0oooo
1P00000=0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool004d0oooo00<0
omnL04P0001d_ol02P3oool00`000000oooo0?ooo`0d0?ooo`030000O`3oool0oooo0140oooo2@00
000C0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0903oool004d0oooo00<0omnL
04P0001d_ol02P3oool00`000000oooo0?ooo`0_0?oooaT00000703oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo02@0oooo001=0?ooo`030?oOW0180000M;oo00X0oooo00<000000?oo
o`3oool06P3ooolE000000D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0d0?oo
o`030000003oool0oooo02@0oooo001=0?ooo`030?oOW0180000M;oo00X0oooo0`0000040?oooaH0
00006P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0;P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0380oooo0`00000V0?ooo`00
C@3oool00`3ogi`0B00007Boo`0:0?ooo`030000003oool0oooo03@0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02@0oooo001=0?ooo`030?oOW018
0000M;oo00X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0]0?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0^0?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo02@0oooo001=0?ooo`030?oOW0180000M;oo00X0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo02@0oooo001=0?ooo`030?oOW0180000M;oo
00X0ooooo`00000a000002H0oooo001;0?ooo`050?oog`2LB00000000000001d_ol0o`3ooomQ0?oo
o`00C@3oool00`3o_g@0000007Boo`3o0?ooof40oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o
0?oook40oooo00280?ooo`030?oo_`2o_kl0M0000080000000D0M;oo0?ooo`3oool0oooo0;md0002
000000050000B02Lgol0oooo0?oog`2LB0000P0000003P00B9`0W7@000000000M;l0oooo0?oOW018
0000000004RLg`3oool0omnL04P000000000B9cO0P3oool01`2oM00000000000000007@0_ooo0?oo
o`2oM0000P0000000`0004P0W=oo0?ooo`020?ooo`050;md0000000000000000M02oool00P3oool0
203ooml0W4P0000000000000001804Q80000B9`0gooo103oool02@3ooml0W4P000000000000004RL
09ad0000000007Bo0;md0002000000030000M02ooml0W4P00080000000@004RL0=ooo`3ooml0W4P0
0P000000100004P0B4P00018W03Oool50?ooo`050;md0000000000000000B02Lgol00`3oool01@2o
M00000000000000007@0_ooo00<0oooo00D0_g@0000000000000001809cOo`030?ooo`050=nLB000
000000000000001d_ol00P3oool01@3OW4P000000000000007@0W9a80080000000<0001d0;ooo`3o
ool0103oool01P3OW4P000000000000004P0W=oO09a80080000000<0001d0;oog`2LB0000P000000
1000M;l0oooo0?ooo`3OW4P30000000407Boo`3oool0oooo0=nLB080000000<0001809cOg`2LB000
0P0000000`0007@0_ooO09a8000200000004001d_`3oool0oooo0;md0080000000@0001d0;ooo`3o
ool0_g@00P0000000`0007@0_ooO09a80002000000080018W03Oool0omnL04P000000000B9cO0?oo
g`2LB002000000040018W02LM0000000001d_eD0oooo00280?ooo`0]0?oo_`1d04P0W=oo0?ooo`3o
okl0M01809cOo`3oool0_g@000000000B9`0gooo0?oog`2LB7@0_ooo0?noM0000000M;oo0?ooo`3o
gi`0B000001d_`3o_kl0_ooo0=nLB00007@0_ooo0=no_`3Oool0gia80000M02oool0gkno0=ooo`2o
M0000000001d_`3oool0omnL04P0M02oool0oooo0=nLB00007@0_ooo00<0oooo0140_g@000000000
M;l0oooo0?oOW01807@0_ooo0?oog`2LB000000007Boo`3oool0oono07@000000000001d0;ooo`03
0?ooo`0K0?noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB000B9`0gooo0?ooo`3o
oml0W4P0001d_`3oool0oooO09a800000000M;oo0?ooo`3ookl0M0000000000007@0_ooo00D0oooo
00<0gia80000M02oool00`3oool05`2oM0000000001d_`3oool0omnL04P0M02oool0oooo0;md0000
000004RL0=ooo`3ooml0W4Qd0;ooo`3oool0gia800000018W=l0oooo0?oog`2LB000B9cO0080oooo
00L0gia80000M02oool0oooo0=nLB00007@0_ooo00L0oooo00<0gia80000M02oool00P3oool01`2o
M00004RL0=ooo`3oool0oooO09a80000M;l00P3oool01`3OW4P0000004RLg`3oool0oooO09a80018
W=l00P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`02
0?ooo`0?0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80018W03Oool0oooo0?oog`2L
B00007Bo0080oooo00d0gia80000M02oool0gkno0=noM0000000M;oo0?ooo`3ogi`0B000001d_`3o
_kl0_ooo05@0oooo00280?ooo`0:0?oo_`2o_ol0oooo0?ooo`3o_g@0000007Boo`3ogi`0B000001d
_`<0oooo00T0omnL09boo`3o_g@0001809cOo`3oool0oooO09a80000M;l00P3oool00`3OW4P004RL
0=ooo`020?ooo`070=nLB000B9`0gooo0?ooo`3ogi`0B00004RLg`030?ooo`060?oOW01dW=l0oooo
0=nLB00007@0_ooo0P3oool00`3ogi`0B00004RLg`030?ooo`050?oOW01dW=l0omnL04P00018W=l0
0P3oool00`3ooml0W4P0001d_`050?ooo`070?noM00004P0W=oo0?ooo`3ooml0W4P0001d_`020?oo
o`0;0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l00P3oool00`3ooml0
W4P0001d_`070?ooo`030=nLB00007@0_ooo0080oooo00<0omnL04P00018W=l00`3oool01@3ogi`0
M9cO0?oOW018000007Bo00<0oooo00D0omnL09boo`3ogi`0B000001d_`030?ooo`0;0?noM00007@0
_ooo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02oool01`3oool00`3OW4P0001d0;ooo`020?oo
o`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00000M;l00`3oool01`3o_g@0
001d0;ooo`3oool0gia80000M02oool00P3oool02`2oM00004RL0=ooo`3oool0oooO09a80000M;l0
oooo0?oOW0180000B9cO00<0oooo00X0omnL07BLg`3oool0gia80018W03Oool0oooo0?oog`2LB000
07Bo0P3oool02`3OW4P004RL0=ooo`3oool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo05H0oooo
002:0?ooo`080?oo_`1d000000000000M02oool0okmd0000B02Lgol60?ooo`060?noM000000007Bo
0?oog`2LB00007Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`070=nLB000B9`0gooo0?ooo`3o_g@0
001809cOo`060?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02Lgol01@3oool00`3o_g@0
000007Boo`020?ooo`030?oog`2LB00007Bo00H0oooo00H0okmd00000000M;l0oooO09a80000M;l2
0?ooo`0;0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000001d_ol00P3oool00`3o
oml0W4P0001d_`070?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02Lgol01@3oool00`3o
_g@0001809cOo`050?ooo`030?noM0000000M;oo00<0oooo00/0okmd0000001d_ol0oooo0=nLB000
07@0_ooo0?ooo`3OW4P0001d0;ooo`070?ooo`030=nLB00007@0_ooo0080oooo00/0_g@00018W03O
ool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0000007Boo`030?ooo`070?noM0000000M;oo0?ooo`3O
W4P0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000B02L
gol01P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool00`3OW4P004RL0=ooo`02
0?ooo`060?noM000000007Bo0?oog`2LB00007BoEP3oool008T0oooo00@0_g@0000000000000B9cO
0P3oool00`3o_g@0001d0;ooo`080?ooo`040?noM00000000000001d_`80oooo00<0gia80018W03O
ool00P3oool01`3OW4P004RL0=ooo`3oool0okmd0000M02oool01P3oool00`3OW4P0001d0;ooo`02
0?ooo`030?noM00007@0_ooo00D0oooo00<0okmd0000001d_ol00P3oool00`3ooml0W4P0001d_`08
0?ooo`040?noM00000000000001d_`80oooo00/0gia80018W03Oool0oooo0?oog`2LB00007Bo0?oo
o`3o_g@0000007Boo`020?ooo`030?oog`2LB00007Bo00L0oooo00<0gia80000M02oool00P3oool0
0`3o_g@0001d0;ooo`050?ooo`030?noM00007@0_ooo00D0oooo00<0okmd0000001d_ol00`3oool0
2`3o_g@0000007Boo`3oool0gia80000M02oool0oooo0=nLB00007@0_ooo00L0oooo00<0gia80000
M02oool00P3oool02`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM0000000M;oo00<0
oooo00L0okmd0000001d_ol0oooo0=nLB00007@0_ooo0080oooo00/0_g@00018W03Oool0oooo0?oo
g`2LB00007Bo0?ooo`3o_g@0001d0;ooo`060?ooo`070=nLB000B9`0gooo0?ooo`3ooml0W4P0001d
_`020?ooo`030=nLB000B9`0gooo00@0oooo00@0okmd00000000000007BoEP3oool008P0oooo00<0
omnL04P00018W=l00P3oool01@2oW9`0gooo0?oOW01804P0W=oo0080oooo00X0okmd0000B02Lgol0
oono07@0001d_ol0oooo0?oog`2LB00007Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`060=nLB000
B9`0gooo0?ooo`3ogi`0B0001@0000001@18W=l0oooo0=nLB00004P0W=oo0080oooo00<0omnL04P0
00000000100000001018W=l0oono07@00018W=l20?ooo`030?oog`2LB00007Bo00D0oooo00L0oono
07@0001d_ol0oooo0?oog`2LB00007Bo0080oooo00/0_g@00000M02oool0oooo0?oog`2LB00007Bo
0?ooo`3ookl0M00004RLg`020?ooo`030?oog`2LB00007Bo00L0oooo00<0gia80000B02Lgol00P3o
ool00`3ogi`0B000000000040000000404RLg`3ogi`0B01809cOo`80oooo00H0okmd0000B02Lgol0
oono07@0001d_ol30?ooo`0;0;md000004P0W=oo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02o
ool01`3oool00`3OW4P0001809cOo`020?ooo`0;0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3o
ool0oono07@0001d_ol00`3oool01`2oM000001809cOo`3oool0gia80000B02Lgol00P3oool02P2o
M000001d0;ooo`3oool0oooO09a80000M;l0oooo0?oOW01800050000000904RLg`3oool0_g@00000
M02oool0oooo0?oog`2LB00007Bo0080oooo00/0gia80018W03Oool0oooo0?oo_`1d0000M;oo0?oo
o`3ooml0W4P0001d_`1F0?ooo`00R03oool05P3ookl0M01d0;ooo`3oool0okmd0018W03Oool0oooo
0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3ookl0M00004RLg`3oool0oono07@00000M;l20?oo
o`030=nLB000B9`0gooo0080oooo00<0gia80018W03Oool00P3oool0402oM000B9cO0?ooo`3oool0
gia80000B02Lgml0W4P00000000004P0B7BL0=noM00007@0_ooo0;md0018W=l20?ooo`0;0=nLB000
04P0W=oo0?ooo`3o_g@0001809cOo`3oool0okmd00000000M;l01@3oool0403ookl0M00004RLg`3o
ool0oono07@00000M;l0oooO09a80000000004Q804P0B02Lgol0oono07@00018W=l20?ooo`070?no
M00004P0W=oo0?ooo`3o_g@00000001d_`050?ooo`0:0?oog`2LB00000000000B018M9`0gkmd0000
M02oool0_g@004RLg`80oooo01`0gia80000B02Lgol0oooo0;md0000B9`0gooo0?oog`2LB000001d
0;ooo`3oool0gia80018W03Oool0oooo0?noM000000007Bo0?oog`2LB00000000000M02oool0oooo
0=nLB00007@0_ooo1@3oool03P3ooml0W4P00000000004P0B4Q809cOo`3ookl0M000001dM01d07@0
_ooo0?oo_`1d0000B9cO0P3oool05@3OW4P004RL0=ooo`3oool0okmd00000000M;l0oooO09a80000
0000001804Q8B02Lgol0oono07@00000M7@0M01d0;ooo`3ookl0M00004RLg`020?ooo`0?0;md0018
W=l0oooo0?ooo`3OW4P0001809cOg`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0080oooo
00/0gia80018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`1F0?ooo`00R@3oool0
1P3o_g@0000000000000M;l0_ibL0=ooo`80oooo00D0gia8000000000000001809cOo`020?ooo`08
0?oo_`1d000000000000000007@0_ooo0?ooo`2oM003000000030018W02LM00000000080000000@0
04RL0=ooo`3oool0gia80P000000300004P0W=oo0?ooo`3oool0_g@00000M02ogi`0B0000000M02o
ool0oooo0=nLB080000000<0001809cOo`3oool00P3oool00`3o_g@000000000000200000003001d
_`3oool0oooo00@0oooo00H0oono07@00000000000000000M02oool20?ooo`070;md0000B9`0gooO
09a800000000001d0;ooo`030?ooo`030?noM000000000000080000000<007Bo0?ooo`3oool01@3o
ool0202oM000001d0;oOW0180000001d0;ooo`3oool0gia80P0000000`0004P0W=oo0?ooo`020?oo
o`050=nLB000000000000000B02Lgol00`3oool01@3o_g@00000000000000000B9cO00<0oooo00L0
_g@00000M02oool0oooo0=nLB00007@0_ooo00L0oooo00`0_g@00000M02ooml0W4P00000000004P0
W=oo0?oo_`1d000000000000M02oool30?ooo`050?noM0000000000000000018W=l00`3oool0302o
M000001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`<0oooo00D0gia80000
00000000001809cOo`020?ooo`090;md0000B9`0gooO09a800000000001d0;ooo`3oool0_g@000<0
000000@004RL0=ooo`3ookl0M0000P0000000`0007@0_ooo0?ooo`1E0?ooo`00X03oool00`2oM000
04RL0=ooo`020?ooo`030;md0000B9`0gooo01`0oooo00<0oooO09a80000M;l06`3oool00`3ooml0
W4P0001d_`0Z0?ooo`030=nLB00007@0_ooo03d0oooo00<0_g@00018W03Oool0GP3oool00:00oooo
00<0oono07A8W03Oool00P3oool00`3ookl0M4RL0=ooo`0L0?ooo`030?oog`2LB00007Bo01/0oooo
00<0oooO09a80000M;l0:P3oool00`3OW4P0001d0;ooo`0m0?ooo`030?oo_`1dB9`0gooo05h0oooo
002P0?ooo`030?oog`2LM9`0gooo0080oooo00<0oooO09adW03Oool0703oool00`3ooml0W4P0001d
_`0K0?ooo`030?oog`2LB00007Bo02X0oooo00<0gia80000M02oool0?@3oool00`3ooml0W7BL0=oo
o`1N0?ooo`00``3oool0103ookl0M00000000000M;lJ0?ooo`040?oo_`1d00000000001d_bH0oooo
00L0gia80000B02Lgml0W4P00000000007@0_ooo09h0oooo00340?ooo`030?oog`2LB00007Bo01/0
oooo00<0oooO09a80000M;l09P3oool01`3OW4P0001809cOo`3oool0_g@00000M02oool0WP3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-10.0967, -0.309528, \
0.155573, 0.00839157}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotRecoilAngle[sol, Join[values, {mratio \[Rule] 0.5}], 
        0.9999\ thmax];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.0317492 -0.285325 0.00987384 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.18256 -0.0125 -5 -15.6875 ]
[.18256 -0.0125 5 0 ]
[.3413 -0.0125 -8 -15.6875 ]
[.3413 -0.0125 8 0 ]
[.50005 -0.0125 -8 -15.6875 ]
[.50005 -0.0125 8 0 ]
[.65879 -0.0125 -8 -15.6875 ]
[.65879 -0.0125 8 0 ]
[.81754 -0.0125 -8 -15.6875 ]
[.81754 -0.0125 8 0 ]
[.97629 -0.0125 -8 -15.6875 ]
[.97629 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.0313 -31.375 ]
[.5 -0.0125 21.0313 -15.6875 ]
[-0.0125 .01089 -16 -7.84375 ]
[-0.0125 .01089 0 7.84375 ]
[-0.0125 .10963 -16 -7.84375 ]
[-0.0125 .10963 0 7.84375 ]
[-0.0125 .20837 -16 -7.84375 ]
[-0.0125 .20837 0 7.84375 ]
[-0.0125 .30711 -16 -7.84375 ]
[-0.0125 .30711 0 7.84375 ]
[-0.0125 .40584 -16 -7.84375 ]
[-0.0125 .40584 0 7.84375 ]
[-0.0125 .50458 -16 -7.84375 ]
[-0.0125 .50458 0 7.84375 ]
[-0.0125 .60332 -16 -7.84375 ]
[-0.0125 .60332 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -31.6875 -21.0313 ]
[-0.01875 .30902 -16 21.0313 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -29.875 0 ]
[.5 .63053 29.875 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.18256 0 m
.18256 .61803 L
s
.3413 0 m
.3413 .61803 L
s
.50005 0 m
.50005 .61803 L
s
.65879 0 m
.65879 .61803 L
s
.81754 0 m
.81754 .61803 L
s
0 .10963 m
1 .10963 L
s
0 .20837 m
1 .20837 L
s
0 .30711 m
1 .30711 L
s
0 .40584 m
1 .40584 L
s
0 .50458 m
1 .50458 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.18256 0 m
.18256 .00625 L
s
gsave
.18256 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.3413 0 m
.3413 .00625 L
s
gsave
.3413 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.50005 0 m
.50005 .00625 L
s
gsave
.50005 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.65879 0 m
.65879 .00625 L
s
gsave
.65879 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.81754 0 m
.81754 .00625 L
s
gsave
.81754 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(25) show
1.000 setlinewidth
grestore
.97629 0 m
.97629 .00625 L
s
gsave
.97629 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 0 m
.05556 .00375 L
s
.08731 0 m
.08731 .00375 L
s
.11906 0 m
.11906 .00375 L
s
.15081 0 m
.15081 .00375 L
s
.2143 0 m
.2143 .00375 L
s
.24605 0 m
.24605 .00375 L
s
.2778 0 m
.2778 .00375 L
s
.30955 0 m
.30955 .00375 L
s
.37305 0 m
.37305 .00375 L
s
.4048 0 m
.4048 .00375 L
s
.43655 0 m
.43655 .00375 L
s
.4683 0 m
.4683 .00375 L
s
.5318 0 m
.5318 .00375 L
s
.56355 0 m
.56355 .00375 L
s
.5953 0 m
.5953 .00375 L
s
.62704 0 m
.62704 .00375 L
s
.69054 0 m
.69054 .00375 L
s
.72229 0 m
.72229 .00375 L
s
.75404 0 m
.75404 .00375 L
s
.78579 0 m
.78579 .00375 L
s
.84929 0 m
.84929 .00375 L
s
.88104 0 m
.88104 .00375 L
s
.91279 0 m
.91279 .00375 L
s
.94454 0 m
.94454 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.0313 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01089 m
.00625 .01089 L
s
gsave
-0.0125 .01089 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
0 .10963 m
.00625 .10963 L
s
gsave
-0.0125 .10963 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .20837 m
.00625 .20837 L
s
gsave
-0.0125 .20837 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
0 .30711 m
.00625 .30711 L
s
gsave
-0.0125 .30711 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
0 .40584 m
.00625 .40584 L
s
gsave
-0.0125 .40584 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(70) show
1.000 setlinewidth
grestore
0 .50458 m
.00625 .50458 L
s
gsave
-0.0125 .50458 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
0 .60332 m
.00625 .60332 L
s
gsave
-0.0125 .60332 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(90) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .03064 m
.00375 .03064 L
s
0 .05039 m
.00375 .05039 L
s
0 .07013 m
.00375 .07013 L
s
0 .08988 m
.00375 .08988 L
s
0 .12938 m
.00375 .12938 L
s
0 .14912 m
.00375 .14912 L
s
0 .16887 m
.00375 .16887 L
s
0 .18862 m
.00375 .18862 L
s
0 .22811 m
.00375 .22811 L
s
0 .24786 m
.00375 .24786 L
s
0 .26761 m
.00375 .26761 L
s
0 .28736 m
.00375 .28736 L
s
0 .32685 m
.00375 .32685 L
s
0 .3466 m
.00375 .3466 L
s
0 .36635 m
.00375 .36635 L
s
0 .3861 m
.00375 .3861 L
s
0 .42559 m
.00375 .42559 L
s
0 .44534 m
.00375 .44534 L
s
0 .46509 m
.00375 .46509 L
s
0 .48483 m
.00375 .48483 L
s
0 .52433 m
.00375 .52433 L
s
0 .54408 m
.00375 .54408 L
s
0 .56383 m
.00375 .56383 L
s
0 .58357 m
.00375 .58357 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -105.875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
82.0313 11.8438 translate
90 rotate
-82.0313 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.18256 .61178 m
.18256 .61803 L
s
.3413 .61178 m
.3413 .61803 L
s
.50005 .61178 m
.50005 .61803 L
s
.65879 .61178 m
.65879 .61803 L
s
.81754 .61178 m
.81754 .61803 L
s
.97629 .61178 m
.97629 .61803 L
s
.125 Mabswid
.05556 .61428 m
.05556 .61803 L
s
.08731 .61428 m
.08731 .61803 L
s
.11906 .61428 m
.11906 .61803 L
s
.15081 .61428 m
.15081 .61803 L
s
.2143 .61428 m
.2143 .61803 L
s
.24605 .61428 m
.24605 .61803 L
s
.2778 .61428 m
.2778 .61803 L
s
.30955 .61428 m
.30955 .61803 L
s
.37305 .61428 m
.37305 .61803 L
s
.4048 .61428 m
.4048 .61803 L
s
.43655 .61428 m
.43655 .61803 L
s
.4683 .61428 m
.4683 .61803 L
s
.5318 .61428 m
.5318 .61803 L
s
.56355 .61428 m
.56355 .61803 L
s
.5953 .61428 m
.5953 .61803 L
s
.62704 .61428 m
.62704 .61803 L
s
.69054 .61428 m
.69054 .61803 L
s
.72229 .61428 m
.72229 .61803 L
s
.75404 .61428 m
.75404 .61803 L
s
.78579 .61428 m
.78579 .61803 L
s
.84929 .61428 m
.84929 .61803 L
s
.88104 .61428 m
.88104 .61803 L
s
.91279 .61428 m
.91279 .61803 L
s
.94454 .61428 m
.94454 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01089 m
1 .01089 L
s
.99375 .10963 m
1 .10963 L
s
.99375 .20837 m
1 .20837 L
s
.99375 .30711 m
1 .30711 L
s
.99375 .40584 m
1 .40584 L
s
.99375 .50458 m
1 .50458 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .03064 m
1 .03064 L
s
.99625 .05039 m
1 .05039 L
s
.99625 .07013 m
1 .07013 L
s
.99625 .08988 m
1 .08988 L
s
.99625 .12938 m
1 .12938 L
s
.99625 .14912 m
1 .14912 L
s
.99625 .16887 m
1 .16887 L
s
.99625 .18862 m
1 .18862 L
s
.99625 .22811 m
1 .22811 L
s
.99625 .24786 m
1 .24786 L
s
.99625 .26761 m
1 .26761 L
s
.99625 .28736 m
1 .28736 L
s
.99625 .32685 m
1 .32685 L
s
.99625 .3466 m
1 .3466 L
s
.99625 .36635 m
1 .36635 L
s
.99625 .3861 m
1 .3861 L
s
.99625 .42559 m
1 .42559 L
s
.99625 .44534 m
1 .44534 L
s
.99625 .46509 m
1 .46509 L
s
.99625 .48483 m
1 .48483 L
s
.99625 .52433 m
1 .52433 L
s
.99625 .54408 m
1 .54408 L
s
.99625 .56383 m
1 .56383 L
s
.99625 .58357 m
1 .58357 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -90.875 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(recoil) show
93.063 15.813 moveto
(angle) show
118.750 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.06245 .58529 L
.10458 .56562 L
.14415 .5471 L
.18221 .52924 L
.22272 .51015 L
.26171 .49169 L
.30316 .47194 L
.34309 .45276 L
.3815 .43415 L
.42237 .41414 L
.46172 .39463 L
.49955 .37561 L
.53984 .35503 L
.57861 .33483 L
.61984 .31285 L
.65955 .29109 L
.69774 .26948 L
.73838 .24556 L
.77751 .22138 L
.81909 .19399 L
.85916 .16519 L
.89771 .13389 L
.91726 .11579 L
.93871 .09285 L
.94807 .08114 L
.9581 .06652 L
.96296 .05812 L
.96756 .04876 L
.97169 .03807 L
.97283 .0344 L
.97403 .02979 L
.97506 .02477 L
.97619 .01472 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00i`3oool01@3OW4P00000000000000000M;oo0140oooo00<0oooO09a8
000000000`0000001018W=l0oooo0?ooo`3OW4P30000000307Boo`3oool0oooo0:<0oooo003W0?oo
o`030=nLB00007@0_ooo01<0oooo00<0_g@00018W03Oool00`3oool00`3OW9`0_ooo0?ooo`020?oo
o`030?oo_`1d0000M;oo0:D0oooo003K0?ooo`040=nLB00000000000001d_`P0oooo00<0gia80018
W03Oool04`3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0080oooo00<0oono07@0
001d_ol0Y@3oool00=h0oooo00<0okmd001d_`3oool01P3oool00`3OW4P004RL0=ooo`0D0?ooo`03
0;md00000000000000<0000000L0B9cO0?ooo`3oool0oooo0?oo_`1d0000M;oo0:D0oooo003C0?oo
o`050?oo_`1d000000000000B02Lgol01P3oool00`3ookl0M01d0;ooo`060?ooo`030=nLB000B9`0
gooo00<0oooo00P0oooO09a80000000000000000B018B00004RL0=ooo`80oooo00P0_g@000000000
0000001d0;ooo`3oool0oono07@000@0000000<0001d0;ooo`3oool00P3oool00`3ookl0M00007Bo
o`2U0?ooo`00d`3oool01P2oM00007Bo0?ooo`3ogi`0B01809cOo`D0oooo00<0gia80000M02oool0
1P3oool00`3OW4P004RL0=ooo`020?ooo`0D0?oog`2LB000000007Boo`3oool0oono07@000000000
001d0;ooo`2oM0000000001d_`3oool0omnL04P0M02oool0oooo0;mdB02Lgol80?ooo`030?oo_`1d
0000M;oo0:D0oooo003B0?ooo`030?oog`2LB000M;oo0080oooo00<0gia80018W03Oool00P3oool0
103ookl0M00000000018W=l70?ooo`030=nLB000B9`0gooo0080oooo00<0omnL04P00018W=l00P3o
ool01`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool01@3ogi`0M9cO0?ooo`3ookl0M000
0080000000<007Bo0?ooo`3oool00`3oool00`3ookl0M00007Boo`2U0?ooo`00dP3oool00`3ookl0
M00007Boo`020?ooo`030?oo_`1d0000M;oo00<0oooo00<0oooO09bL_`3oool01`3oool00`3OW4P0
04RL0=ooo`020?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?noM00004P0
W=oo00H0oooo00H0gia80000M02oool0oono07@00018W=l40?ooo`030?oo_`1d0000M;oo0:D0oooo
003B0?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool00`3oool00`3o_g@0B9cO0?oo
o`060?ooo`030=nLB000B9`0gooo0080oooo00<0okmd0000001d_ol00P3oool01`3ooml0W4P0001d
_`3oool0okmd0000M02oool01@3oool01`3ookl0M00004RLg`3oool0oooO09a80000M;l0103oool0
0`3ookl0M00007Boo`2U0?ooo`00dP3oool00`3ooml0W4P0001d_`030?ooo`080=nLB00007@0_ooo
0?oOW01dW=l0oooo0=nLB000M;l70?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@00018W=l0
0P3oool01P3ooml0W4P0001d_`3oool0omnL04P000D0000000P0B9cO0?oo_`1d0000B9cO0?ooo`3o
oml0W4P004RLg`@0oooo00<0oono07@0001d_ol0Y@3oool00=<0oooo00<0_g@00000000000001000
00001P1d_ol0oooo0=nLB0000000001809cOo`L0oooo00<0gia80018W03Oool00`3oool01`3o_g@0
001809cOo`3oool0okmd00000000M;l00P3oool0302oM000B9cO0?ooo`3oool0gia80000B02Lgol0
oooo0;md0000M;l0oooo0?noM080000000L004RL0=ooo`3oool0oooo0?oo_`1d0000M;oo0:D0oooo
003C0?ooo`030=nLB00007@0_ooo0080oooo00<0oono07@00018W=l0303oool00`3OW4P004RL0=oo
o`040?ooo`030?noM000000000000080000000D007Bo0?ooo`3oool0oooo0=nLB002000000060000
B02Lgol0oooo0?ooo`3ooml0W4P00P0000000`1d_ol0oooo0?ooo`030?ooo`030?oo_`1d0000M;oo
0:D0oooo003C0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@0001d_`3oool02`3oool00`3OW4P0
04RL0=ooo`070?ooo`030?oog`2LB00007Bo01@0oooo00<0oono07@0001d_ol0Y@3oool00=@0oooo
00<0_g@00018W03Oool00P3oool00`3OW4P007Bo0?ooo`0;0?ooo`030=nLB000B9`0gooo00L0oooo
00<0oooO09a80000M;l0503oool00`3ookl0M00007Boo`2U0?ooo`00e03oool00`3ookl0M00007Bo
o`020?ooo`030=nLB000M;l0oooo00/0oooo00<0gia80018W03Oool01`3oool00`3ooml0W4P0001d
_`0D0?ooo`030?oo_`1d0000M;oo0:D0oooo003E0?ooo`060?noM00004P0W=oo0?ooo`2oM000M;oo
303oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d00000000001d_a@0oooo00<0oono07@0001d
_ol0Y@3oool00=H0oooo00D0omnL04P00000000004RL0=ooo`0<0?ooo`050=nLB000000000000000
001d_ol01@3oool00`3ooml0W4P0001d_`0B0?ooo`050=nLB000000000000000001d_ol0Y@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool005`0oooo00D0omnL04P000000000001d0;ooo`0Y0?ooo`030?oog`2LB0000000
0080000000<0M;oo0?ooo`3oool09P3oool01P2oM000000000000000000004RL0=ooo`<0oooo00D0
omnL04P000000000001d0;ooo`0R0?ooo`090;md00000000000000000000B9`0gooo0?ooo`3ooml0
W4P000<0000000<0M;oo0?ooo`3oool07`3oool00`2oM0000000000000040000000904RLg`3oool0
oooo0?ooo`3ogi`0B0000000000007@0_ooo0240oooo00<0_g@0000000000000100000001018W=l0
oooo0?oog`2LB0030000000307Boo`3oool0oooo01l0oooo00<0oono07@0000000000P0000000`1d
_ol0oooo0?ooo`030?ooo`050?oOW018000000000000M02oool0:@3oool005/0oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo02P0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?oo
o`0W0?ooo`030?oOW0180000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0280
oooo00<0omnL04P0001d_ol00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0200
oooo00<0gia80000000000000`0000002P00B9`0gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000
M02ooolQ0?ooo`030=nLB0000000000000<0000000D004RL0=ooo`3ooml0W4Qd0;ooo`020?ooo`03
0;md001d_ol0oooo01l0oooo00<0omnL04P00018W=l00P3oool00`2oM000B9cO0?ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0X0?ooo`00F`3oool00`2oM00004RL0=ooo`020?oo
o`030;md0000M;l0oooo02`0oooo00<0oooO09a8M02oool09`3oool00`3ogi`0B00007Boo`040?oo
o`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool08@3oool00`3ogi`0B00007Boo`080?oo
o`030?oog`2LB7@0_ooo0240oooo00<0gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM000
04RL0=ooo`80oooo00<0_g@0001d_`3oool08@3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02L
gol0oooo00@0oooo00<0oooO09a8M02oool0903oool00`3ookl0M01809cOo`020?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool09`3oool005X0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`0]0?ooo`030;md001d_ol0oooo02H0oooo00<0omnL04P0001d_ol00`3oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0240oooo00<0omnL04P0001d_ol02@3oool0
0`2oM000M;oo0?ooo`0Q0?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l00`3oool0
0`3OW4P0001d0;ooo`0R0?ooo`030=nLB01d_ol0oooo00T0oooo00<0_g@007Boo`3oool08`3oool0
1`3ookl0M00007Boo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0W0?ooo`00FP3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02`0oooo00<0oooO09a80018W=l09`3o
ool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol08@3o
ool00`3ogi`0B00007Boo`080?ooo`030?oog`2LB000B9cO02<0oooo00<0_gA809cOo`3oool00P3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02<0oooo00<0_gA809cOo`3oool01`3o
ool00`3ooml0W4P004RLg`0T0?ooo`070?noM0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo02L0oooo001J0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
;03oool00`3OW4P0000007Boo`0W0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`0Q0?ooo`030?oOW0180000M;oo00P0oooo00<0gia80000001d_ol0
8`3oool00`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
8`3oool00`3ooml0W4Qd0;ooo`070?ooo`030=nLB0000000M;oo02<0oooo00P0omnL04P0000004P0
W=oo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol09`3oool005X0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`0X0?ooo`070=oOg`3oool0oono07@000000000001d0;oo
o`0W0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0Q0?ooo`030?oOW0180000M;oo00@0oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo02@0
oooo00L0okmd001d_`3oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0903oool0
0`3o_g@007Bo0?ooo`020?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0Q0?ooo`050?no
M000000000000000B02Lgol00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo02L0
oooo001J0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0:03oool01P3OW4P00000
000000000000001809cOobP0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo0240oooo00<0omnL04P0001d_ol0103oool01P3OW4P00000000000000000
001809cOobD0oooo00L0oono07@0B02Lgol0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol0903oool00`3ookl0M01809cOo`020?ooo`060=nLB000000000000000000004P0W=oo903oool0
0`2oM00007Bo0?ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol09`3oool0
05X0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0X0?ooo`040?oo_`1d00000000
04RLgbX0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo0240oooo00<0omnL04P0001d_ol0103oool0103ookl0M00000000018W=lW0?ooo`070?oo_`1d
0000B9cO0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo02@0oooo00<0oono07@00018
W=l00P3oool0103ookl0M00000000018W=lW0?ooo`070;md0018W=l0oooo0?ooo`3ookl0M00004RL
g`030?ooo`030=nLB00007@0_ooo02L0oooo001K0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0
001d_`3oool0:@3oool00`2oM000M;oo0?ooo`0Z0?ooo`030?oOW0180000M;oo00@0oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`0Q0?ooo`030?oOW0180000M;oo00D0oooo00<0_g@0
07Boo`3oool08P3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO0080oooo00<0gia8
0018W03Oool00P3oool00`2oM00007Bo0?ooo`0O0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL
04P00018W=l00`3oool00`2oM000M;oo0?ooo`0R0?ooo`030?oog`2LW;l0oooo0080oooo00<0gia8
0000M02oool00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo02L0oooo001K0?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0Z0?ooo`060?oOW018000000000000000004P0
W=oo9@3oool01@3ooml0W4P0000000000000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM000
07@0_ooo0200oooo00D0oooO09a800000000000007Boo`050?ooo`060?oOW0180000000000000000
04P0W=oo7`3oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo0P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0803oool0203ooml0W4Q809cOo`3oool0oono07@0000007@0_ooo
0`3oool01P3ogi`0B000000000000000001809cOob00oooo00H0okmd04RLg`3oool0oono07@00000
M;l30?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0X0?ooo`00G03oool01@3ookl0M000
0000000007@0_ooo02`0oooo00D0_g@0000000000000000004RLg`0W0?ooo`030?noM0000000M;oo
00D0oooo00D0oono07@000000000001d0;ooo`0S0?ooo`030?noM0000000M;oo00H0oooo00D0_g@0
000000000000000004RLg`0P0?ooo`060?oog`2LB00000000000000007@0_ooo103oool01@3ookl0
M0000000000007@0_ooo0280oooo00H0oooO09a80000000000000000M02oool50?ooo`050;md0000
0000000000000018W=l08@3oool01@3OW4P000000000000007@0_ooo00@0oooo00D0oono07@00000
0000001d0;ooo`0Y0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00?`3oool00`3ookl0M0000000
00020000000307Boo`3oool0oooo00<0oooo00D0omnL04P000000000001d0;ooo`3o0?ooof80oooo
000o0?ooo`030?oOW0180000B9cO0080oooo00<0_g@004RLg`3oool00P3oool01`3ookl0M00007Bo
o`3oool0okmd0000M02oool01`3ooooo000003800000:03oool004@0oooo00<0oono07@0B02Lgol0
0P3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00H0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02H0oooo00140?ooo`070?oo_`1d
0000M;oo0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02H0oooo00140?ooo`070?no
M0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00<000000?oo
o`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?oo
o`3oool09P3oool004<0oooo00P0omnL04P0000004P0W=oo0?ooo`3ogi`0B00007Boo`<0oooo00<0
okmd0000B02Lgol01P3oool3000003D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0b0?ooo`<00000:03oool00440oooo00D0okmd000000000000001809cOo`020?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02h0oooo00<0001o0?ooo`3oool0;@3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo02H0oooo00130?ooo`030;md0000M;l0oooo0080oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`060?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool09P3oool0
04@0oooo00L0_g@004RLg`3oool0oooo0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3o
ool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0;03o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02H0oooo000o0?ooo`030?oog`2LW;l0
oooo0080oooo00<0gia80000M02oool00P3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0
oooo00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo02`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0V0?ooo`00@03oool01P3o
_g@0B9cO0?ooo`3ookl0M000001d_`<0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00L0
oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02`0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0V0?ooo`00@@3oool01@3OW4P00000
0000000007@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`080?ooo`<00000=@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02`0oooo00<000000?ooo`3oool0103oool20000
02P0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0/0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool09P3oool005L0oooo
00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02/0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?oo
o`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0Z0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?oo
o`030000O`3oool0oooo02X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0V0?oo
o`00E`3oool3000003D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0Z0?ooo`03
0000003oool0oooo00H0oooo0P00000X0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0:@3oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0Y0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool09P3o
ool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo02P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0V0?ooo`00E`3oool00`00
0000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0:03oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0W0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool09P3oool005L0oooo0`00000e0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo
00<0001o0?ooo`3oool09P3oool00`000000oooo0?ooo`0:0?ooo`800000:03oool005L0oooo00<0
00000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo02D0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?ooo`0e
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool09@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;P3oool00`0007l0oooo0?ooo`0T0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`03
0000O`3oool0oooo02<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0V0?ooo`00
E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0
8P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo02H0oooo001G0?ooo`<00000=@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo0280oooo00<000000?ooo`3oool03P3o
ool2000002P0oooo00130?ooo`040?oog`2LB00004RL0=ooo`<0oooo00D0omnL04P000000000001d
0;ooo`080?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo
0?ooo`0Q0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool09P3oool004<0oooo00@0
oooO09a80000B9`0gooo0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool00`00
0000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0803oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo00130?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0O0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool09P3oool003l0oooo00<0_g@00000000000001@000000101d_ol0
oono07@00018W=l30?ooo`030=nLB00007@0_ooo00H0oooo00<000000?ooo`3oool0=@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo01h0oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0V0?ooo`00?`3oool0303ogi`0M9cO0?ooo`3oool0oooO09a80000B9`0gooo0?oo
o`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool07P3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo02H0oooo00100?ooo`0;0=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`3oool0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`<00000o`0007lG0000O`0300000000
07l0001o0180001o0`00000X0?ooo`00@03oool02`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0
oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0703oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo02H0oooo00110?ooo`0:0?oOW018M;l0oooO09a80000B9`0gooo0?ooo`3o
gi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool06`3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo02H0oooo00120?ooo`090=nLB01d_ml0W4P00018W03Oool0oooo0?oo_`1d0000B9cO
00<0oooo00<0gia80000M02oool01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02h0oooo00<0001o0?ooo`3oool06P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
02H0oooo00120?ooo`050?oog`2LB4P0B4P00018W03Oool00P3oool00`3OW4P004RL0=ooo`020?oo
o`030;md0000M;l0oooo00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?oo
o`030000O`3oool0oooo01T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0V0?oo
o`00@`3oool0103ogi`0B0000018W03Oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`070?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0G0?ooo`8000006`3oool00`000000oooo0?ooo`0V0?ooo`00A03oool00`3OW4P004RL0=ooo`03
0?ooo`050?oo_`1d000000000000M02oool0203oool3000003D0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0F0?ooo`030000003oool0oooo01X0oooo0P00000X0?ooo`00E`3oool0
0`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool05@3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0D0?ooo`030000003oool0oooo
01d0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0^0?ooo`030000O`3oool0oooo01<0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo
00<0001o0?ooo`3oool04@3oool200000240oooo00<000000?ooo`3oool09P3oool005L0oooo00<0
00000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo0100oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`0V0?ooo`00E`3oool3000003D0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0?0?ooo`030000003oool0oooo0240oooo0P00000X
0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?oo
o`3oool03P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02H0oooo001G0?ooo`03
0000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0=0?ooo`03
0000003oool0oooo02@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo00/0oooo0P00000W0?ooo`030000
003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0:0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool09P3o
ool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo00T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0V0?ooo`00E`3oool30000
03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`070?ooo`800000:P3oool20000
02P0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`060?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool09P3oool005L0oooo
00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo00@0oooo
0P00000^0?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`030?ooo`030000003oool0oooo02h0oooo00<0
00000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^
0?ooo`040000O`3oool0oooo0?ooo`800000<@3oool00`000000oooo0?ooo`0V0?ooo`00E`3oool0
0`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00@0001o0?ooo`3oool00000<`3o
ool00`000000oooo0?ooo`0V0?ooo`00E`3oool3000003D0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool0
0`0007l000000000000c0?ooo`800000:03oool003l0oooo00<0oooO09a8000000000P0000000`1d
_ol0oooo0?ooo`030?ooo`050?oOW018000000000000M02oool0203oool00`000000oooo0?ooo`0e
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02h0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0V
0?ooo`00?`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0080oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0/0?ooo`80000000<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00A03oool0
0`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01P3oool0
0`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02/0oooo00@000000?ooo`3oool0001o=P3o
ool00`000000oooo0?ooo`0V0?ooo`00A@3oool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo
00<0gia80000M02oool01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02T0oooo
0P0000030?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool004@0oooo00L0
oooO09a80018W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`000000
oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02P0oooo00<000000?ooo`3oool00`3oool00`0007l0
oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo00140?ooo`070=nLB0000000M;oo0?ooo`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo0`00003_0000O`030000000007l0001o
03T0001o0`00000X0?ooo`00@03oool02`3Ogml0oooo0?oo_`1d000000000000M02oool0oooo0?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02D0oooo0P0000070?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0
9P3oool00400oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02@0
oooo00<000000?ooo`3oool01`3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0
oooo000/0?ooo`030?oOW0180000000000<0000000<007Bo0?ooo`3oool02`3oool0103ookl0M000
00000018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool00`000000
oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo0280oooo0P00000:0?ooo`030000O`3oool0oooo03@0
oooo00<000000?ooo`3oool09P3oool002X0oooo00<0gia80000000000001P0000000`00B9`0gooo
0?ooo`0;0?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo
0?ooo`060?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool08@3oool00`000000oooo
0?ooo`0:0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool002P0oooo00H0
oono07@000000000001d0;oo_`2LW=l50?ooo`030;mdM02oool0oooo00/0oooo00H0omnL04P00000
000000000000B02Lgol20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`030000
003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool07`3oool2000000d0oooo00<0001o0?ooo`3oool0
=03oool00`000000oooo0?ooo`0V0?ooo`009`3oool00`3ogi`0B01809cOo`020?ooo`030?oo_`2L
W=l0oooo00@0oooo00<0_ibL0=ooo`3oool0303oool01@2oM0000000000000000000B9cO00<0oooo
00D0oono07@000000000001d0;ooo`080?ooo`<00000=@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0N0?ooo`03
0000003oool0oooo00d0oooo00<0001o0?ooo`3oool0<`3oool2000002P0oooo000W0?ooo`030;nL
W03Oool0oooo0080oooo00<0oono09bLg`3oool00P3oool00`3ookl0M00007Boo`0S0?ooo`030000
003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0703oool200000100oooo00<0001o0?ooo`3oool0
=03oool00`000000oooo0?ooo`0V0?ooo`009P3oool00`3ookl0M7Bo0?ooo`030?ooo`030?oo_`1d
0000000000<0000000<0B9cO0?ooo`3oool08P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo01X0oooo0P00000B0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3o
ool002H0oooo00<0oooO09a8000000001`0000000`00B9`0gooo0?ooo`0T0?ooo`030000003oool0
oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool06@3oool00`000000oooo0?ooo`0B0?ooo`030000O`3oool0
oooo03@0oooo00<000000?ooo`3oool09P3oool002L0oooo00<0oono07@000000000100000000`1d
_ol0oooo0?ooo`0V0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool05`3oool20000
01D0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00<P3oool0103ogi`0
B0000018W03OoolQ0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool05P3oool00`00
0000oooo0?ooo`0E0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool00340
oooo00D0gia80018W03Ogi`0B4RL0=ooo`0Q0?ooo`<00000=@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0D0?oo
o`800000603oool00`0007l0oooo0?ooo`0c0?ooo`800000:03oool00300oooo00H0_g@004RLg`3o
ool0omnL04Q8W03OoolQ0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool04P3oool2
000001X0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00;`3oool01`3O
W4P0M;oo0?ooo`3oool0omnL04Q8W03Oool08@3oool00`000000oooo0?ooo`0e0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo0100oooo0P00000L0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3o
ool002h0oooo00<0gia80000000000001@0000000`0007@0_ooo0?ooo`0N0?ooo`030000003oool0
oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool03P3oool2000001h0oooo00<0001o0?ooo`3oool0=03oool0
0`000000oooo0?ooo`0V0?ooo`00<`3oool00`3ogi`0B4RL0=ooo`0Q0?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0303oool200000200oooo00<0001o0?ooo`3oool0=03oool00`00
0000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo00X0
oooo0P00000R0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo
0`00000e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool08P3oool00`0007l0oooo
0?ooo`0c0?ooo`800000:03oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`070?ooo`8000009@3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo001G
0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0U
0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?oo
o`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`040?ooo`800000:03oool00`0007l0oooo0?ooo`0d0?oo
o`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool00P3oool2000002X0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00
9P3oool00`3ogi`0B0000000000>000000030000M02oool0oooo01d0oooo00<000000?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool01@0007l0oooo0?ooo`3oool0000002`0oooo00<0001o0?ooo`3oool0=03oool00`000000
oooo0?ooo`0V0?ooo`009P3oool00`3ogi`0M9cO0?ooo`0>0?ooo`030;mdM02oool0oooo01d0oooo
0`00000e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`040000O`3oool00000000002d0oooo00<0001o0?ooo`3oool0<`3oool2000002P0
oooo000V0?ooo`030?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3oool01`3oool01@3ogi`0B000
000000000000M;oo00@0oooo00D0omnL04P000000000001d0;ooo`080?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0000003oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo
02H0oooo000V0?ooo`030?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3oool01P3oool01`3ogi`0
B00004RLg`3oool0oono07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
1`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0/0?ooo`800000<03oool00`0007l0oooo0?ooo`0d0?ooo`030000
003oool0oooo02H0oooo000o0?ooo`030?oog`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo
0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool01P3oool00`000000oooo0?ooo`0e0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0Z0?ooo`80
000000<0oooo0000O`3oool0;`3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0
oooo000o0?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0oono07@00018W=l30?oo
o`030=nLB00007@0_ooo00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0:03oool2000000<0oooo00<0001o0?oo
o`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo000^0?ooo`05
0;md00000000000000000018W=l0303oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo
0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`060?ooo`030000003oool0oooo03D0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02H0oooo0P00
00050?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?oo
o`0V0?ooo`00;03oool00`3ookl0M000000000040000000304RLg`3oool0oooo00T0oooo00<0omnL
04P0001d_ol00`3oool01P2oM000001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol0
1P3oool300000;`0001o0P00001]0000O`<00000:03oool002`0oooo00<0gia8001d_`3oool00P3o
ool0103ookl0M0000000M02oool:0?ooo`0<0?oo_`1d0000000004RLg`3oool0okmd00000000M;l0
oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`060?ooo`030000003oool0oooo03D0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo0280oooo
0P0000090?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo
0?ooo`0V0?ooo`00;03oool00`2oM7@0_ooo0?ooo`030?ooo`030?oo_`1d07@0_ooo00/0oooo00<0
_g@00018W02oW4P00P0000001P00B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02L
gol01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0P0?ooo`8000002`3oool00`0007l0oooo0?ooo`0^0?ooo`03
0000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool002L0oooo00<0oknL09cOo`3oool0
0P3oool00`3ogi`0W;oo0?ooo`030?ooo`030?noW02Lgol0oooo00/0oooo00<0omnL04P0001d_ol0
1@3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<000000?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0
7P3oool2000000d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000
003oool0oooo02H0oooo000W0?ooo`030;md00000000000000T0000000<0001d0;ooo`3oool02`3o
ool00`3OW4P0001809cOo`050?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01P3o
ool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0L0?ooo`8000003`3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3o
ool0oooo03@0oooo00<000000?ooo`3oool09P3oool002H0oooo00<0oono07@0000000002P000000
0`00M;l0oooo0?ooo`0<0?ooo`030?noM000000007Bo00@0oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo00L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool06P3oool200000140oooo00<0001o0?ooo`3oool0
;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo000b0?ooo`030?oOW02L
_ol0oooo00h0oooo00D0oooO09a800000000001d0;ooo`020?ooo`050?oo_`1d000000000000M02o
ool0203oool3000003D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo01P0oooo0P00000C0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3o
ool0<`3oool2000002P0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo01H0oooo0P00000E0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00;@3o
ool00`3ookl0M000000000020000000304RLg`3oool0oooo0280oooo00<000000?ooo`3oool0=@3o
ool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool05@3o
ool00`000000oooo0?ooo`0E0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03o
ool00`000000oooo0?ooo`0V0?ooo`00;03oool00`3ogi`0B000000000040000000304RLg`3oool0
oooo0240oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool04`3oool2000001P0oooo00<0001o0?ooo`3oool0;P3oool0
0`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo000/0?ooo`090=nLM02Lgol0okmd
07Boo`3oool0okmd0000000007@0_ooo0280oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool04@3oool2000001X0oooo
00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo
000/0?ooo`040;mdB02Lgol0okmd07Boo`80oooo00<0gia80000M02oool08P3oool00`000000oooo
0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0?0?ooo`800000703oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo
00<000000?ooo`3oool09P3oool002`0oooo00@0okmd000000000000M;oo0P3oool00`2oM00007Bo
0?ooo`0R0?ooo`<00000=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool03@3oool2000001h0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo
0?ooo`0c0?ooo`800000:03oool002d0oooo00H0gia80000001d_ol0okmd00000000M;lT0?ooo`03
0000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo00/0oooo0P00000P0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3o
ool0=03oool00`000000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`090?ooo`800000
8P3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0
9P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool01`3oool2000002@0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo000]0?ooo`0;0?noM0000000
00000018W03Ookl0M000001d_`3OW4P000000000M02oool07`3oool00`000000oooo0?ooo`0e0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`050?oo
o`8000009P3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?oo
o`3oool09P3oool002`0oooo00<0okmd0000000000000P0000001`0007@0_kmd00000018W=l0oooo
0;md0000M;l07`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`040?ooo`030000003oool0oooo02H0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo000/0?oo
o`0=0;mdM02oool0oooo0?ooo`3ooml0W4Qd0;oOW01807@0_ooo0?ooo`3o_g@004RL0=ooo`0N0?oo
o`<00000=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?oo
o`3oool00P3oool2000002T0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0c0?oo
o`800000:03oool002`0oooo00<0_gA809cOo`3oool00P3oool0202oM7@0_mnL04P0B02Lgol0oooo
0?oOW018B9`0gooo7P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`02000002/0oooo00<0001o0?ooo`3oool0;P3o
ool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo000/0?ooo`030?oOW0180000
00000080000000L007Bo0?oOW01804P0W=oo0?ooo`3o_g@007Bo01l0oooo00<000000?ooo`3oool0
=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o000000000000
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0
9P3oool002`0oooo00`0oono07@00000000000000018W03Oool0omnL04P0001d_ol0oooO09a8B02L
golO0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0/0?ooo`800000;`3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool09P3oool002`0oooo00<0oono09bLg`3oool0103oool0102oM00000000000B02L
golP0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0Z0?ooo`80000000<0oooo0000O`3oool0;P3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3o
ool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02P0oooo0P0000030?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02H0oooo001G0?ooo`<00000=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02H0oooo0P0000050?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`00
07l0oooo0?ooo`0c0?ooo`800000:03oool002H0oooo00<0omnL09boo`3oool03P3oool00`3OW9`0
_ooo0?ooo`080?ooo`030=nLB018W=l0oooo00D0oooo00D0omnL04P000000000001d0;ooo`080?oo
o`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0T0?oo
o`8000001`3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?oo
o`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`009P3oool00`3ogi`0W;oo0?ooo`0>0?ooo`03
0=nLW02oool0oooo00P0oooo00<0omnL04Q8W03Oool0103oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool01`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool08P3oool2000000T0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^
0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool002H0oooo00<0omnL07BL
g`3oool03P3oool00`2oM7@0_ooo0?ooo`080?ooo`030?oog`2LB4P0W=oo00@0oooo00<0_g@00018
W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030000003oool0oooo03D0oooo00<0001o0?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0P0?ooo`8000002`3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?oo
o`009P3oool00`3ogi`0B0000000000>000000030000M02oool0oooo00T0oooo00<0gia804RLg`3o
ool00P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<000000?ooo`3o
ool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo01h0oooo0P00000=0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`03
0000003oool0oooo02H0oooo00130?ooo`030?oOW018B9`0gooo0080oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@0001809cOo`060?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0L0?ooo`8000003`3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00@`3o
ool00`3ooml0W4Q809cOo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3o
ool300000880001o0P00002W0000O`<00000:03oool004@0oooo00L0gia8001d_`3oool0oooo0?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0603oool2000001<0oooo00<0001o0?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool0
9P3oool004@0oooo00L0oono07@0M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool05@3oool3000001D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`03
0000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool003l0oooo00<0omoO0=ooo`3oool0
0`3oool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool01P3oool00`00
0000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool04`3oool20000
01P0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo
03@0oooo00<000000?ooo`3oool09P3oool003l0oooo00<0oono07A8W03Oool00`3oool01`3o_g@0
07Bo0?ooo`3oool0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030000003oool0
oooo03D0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0A0?ooo`8000006P3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool0
0`000000oooo0?ooo`0V0?ooo`00@03oool00`2oM0000000000000030000000:0000M02oool0oooo
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`L0oooo00<000000?ooo`3oool0=@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo00l0oooo0P00000L0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0
oooo00100?ooo`030?noM0000000000000@0000000P0M;oo0?ooo`3oool0oono07@000000000001d
0;ooo`P0oooo0`00000e0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool03@3oool2
000001h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo03<0oooo0P00000X0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool02`3oool200000200oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0
oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo00T0
oooo0P00000R0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`070?ooo`800000903oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo
0?ooo`0V0?ooo`00E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool01@3oool2000002H0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<00000
0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo00<0oooo0P00000X
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo02H0oooo001G0?ooo`<00000=@3oool00`0007l0oooo0?ooo`0]0?oo
o`040000O`3oool0oooo0?ooo`800000:P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02h0oooo00<0001o0?ooo`3oool0<`3oool2000002P0oooo001G0?ooo`030000003oool0oooo
03D0oooo00<0001o0?ooo`3oool0;@3oool0100007l0oooo00000000000/0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0
oooo02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool0;@3oool20000
02h0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo
03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0
oooo0?ooo`0[0?ooo`80000000<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00
E`3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo02T0oooo0P0000020?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000
O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3o
ool00`0007l0oooo0?ooo`0W0?ooo`800000103oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003o
ool0oooo02H0oooo001G0?ooo`<00000=@3oool00`0007l0oooo0?ooo`0U0?ooo`8000001P3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool0
0`0007l0oooo0?ooo`0c0?ooo`800000:03oool005L0oooo00<000000?ooo`3oool0=@3oool00`00
07l0oooo0?ooo`0S0?ooo`800000203oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo
02H0oooo001G0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool08@3oool2000000X0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0
oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo
0?ooo`0e0?ooo`030000O`3oool0oooo01l0oooo0P00000<0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo
00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?oo
o`0M0?ooo`8000003P3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo001G
0?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool06`3oool200000100oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o
0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00E`3oool3000003D0oooo00<0001o0?oo
o`3oool06@3oool200000180oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0<`3oool2000002P0oooo00110?ooo`05
0=nLB0000000000000000018W=l0103oool01@3ogi`0B0000000000007@0_ooo00P0oooo00<00000
0?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0G0?ooo`800000503oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo02H0oooo00100?ooo`070=nLB0000000M;oo0?ooo`3ooml0W4P004RL
g`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`030000003oool0oooo03D0
oooo00<0001o0?ooo`3oool0503oool3000001H0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000
oooo0?ooo`0V0?ooo`00?`3oool00`3ooml0W4P0001d_`030?ooo`070?noM000B9`0gooo0?ooo`2o
M00004RL0=ooo`020?ooo`030;md0000M;l0oooo00H0oooo00<000000?ooo`3oool0=@3oool00`00
07l0oooo0?ooo`0B0?ooo`8000006@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo
02H0oooo00100?ooo`030;md0000M;l0oooo0080oooo00H0_g@00018W03Oool0oono07@00018W=l3
0?ooo`030=nLB00007@0_ooo00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0@
0?ooo`8000006`3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo00100?oo
o`0;0?oOW01807@0_ooo0?ooo`2oM000000007Boo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool03@3oool3000001d0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0
oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00@@3oool01@3ookl0M4Q8
04P000000000B9cO0080oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`<0
0000@`0007l200000>H0001o0`00000X0?ooo`00@@3oool0103ogi`0B00000000000M;l30?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`03
0000O`3oool0oooo00T0oooo0P00000R0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3o
ool09P3oool00400oooo00H0omnL04P0000007@0_ooo0?noM000M;l20?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo
00L0oooo0P00000T0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool00400
oooo00<0_g@00000M02oool00P3oool01P2oM000B9cO0?ooo`3ookl0M00004RLg`<0oooo00<0gia8
0000M02oool01P3oool00`000000oooo0?ooo`0e0?ooo`030000O`3oool0oooo00D0oooo0P00000V
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^
0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool00400oooo00<0_g@00018
W03Oool00P3oool01`3OW4P007Bo0?ooo`3oool0gia80018W03Oool00P3oool00`2oM00007Bo0?oo
o`060?ooo`030000003oool0oooo03D0oooo00<0001o0?ooo`3oool00`3oool2000002P0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00@03oool01`3o_g@0000007Boo`3o
ool0oono07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool00`00
0000oooo0?ooo`0e0?ooo`040000O`3oool0oooo0?ooo`800000:P3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d
0?ooo`030000003oool0oooo02H0oooo00110?ooo`050?noM0000000000000000018W=l0103oool0
1@3ookl0M0000000000007@0_ooo00P0oooo0`00000e0?ooo`040000O`3oool00000000002`0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo
00<0001o0?ooo`3oool0<`3oool2000002P0oooo001G0?ooo`030000003oool0oooo03D0oooo0P00
000^0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<00000
0?ooo`3oool0<`3oool2000000030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool0
0`000000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?ooo`0a0?ooo`8000000P3oool00`00
07l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0
oooo00<000000?ooo`3oool0;`3oool2000000@0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo02d0oooo
0P0000060?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo
0?ooo`0V0?ooo`00E`3oool3000002/0oooo0P0000080?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo00<0
001o0?ooo`3oool0<`3oool2000002P0oooo001G0?ooo`030000003oool0oooo02P0oooo0`00000:
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V
0?ooo`00E`3oool00`000000oooo0?ooo`0V0?ooo`8000003@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?oo
o`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3o
ool0903oool2000000l0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`03
0000003oool0oooo02H0oooo001G0?ooo`030000003oool0oooo0240oooo0`00000A0?ooo`030000
O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00E`3o
ool00`000000oooo0?ooo`0O0?ooo`800000503oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3o
ool0oooo03@0oooo00<000000?ooo`3oool09P3oool005L0oooo0`00000L0?ooo`<000005P3oool0
0`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03<0oooo0P00000X0?ooo`00E`3oool00`00
0000oooo0?ooo`0J0?ooo`8000006@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo
03@0oooo00<000000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool0603oool2000001/0
oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0
oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0
oooo001G0?ooo`030000003oool0oooo01H0oooo0P00000M0?ooo`030000O`3oool0oooo02d0oooo
00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02h0oooo
00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0V0?ooo`00E`3oool00`000000oooo0?oo
o`0D0?ooo`8000007`3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool09P3oool005L0oooo00<000000?ooo`3oool04P3oool200000240oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo001G0?oo
o`<00000403oool2000002<0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0c0?oo
o`800000:03oool003l0oooo00D0oono07@00000000004RL0=ooo`060?ooo`050?oOW01800000000
0000M02oool0203oool00`000000oooo0?ooo`0>0?ooo`8000009@3oool00`0007l0oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^
0?ooo`030000O`3oool0oooo03@0oooo00<000000?ooo`3oool09P3oool00480oooo00<0gia80000
001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool00`000000oooo0?oo
o`0<0?ooo`8000009`3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0
001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0oooo03@0oooo00<0
00000?ooo`3oool09P3oool004<0oooo00<0oono07@0001d_ol00`3oool00`2oM00004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00H0oooo00<000000?ooo`3oool02P3oool2000002T0oooo00<0001o
0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o
0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo02H0oooo00140?oo
o`070?oOW01804P0W=oo0?ooo`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0
00000?ooo`3oool0203oool2000002/0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?oo
o`0d0?ooo`030000003oool0oooo02H0oooo00100?ooo`030?oog`2LB000000000<0000000D007Bo
0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00<000000?ooo`3oool01P3o
ool2000002d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02d0oooo00<0001o0?ooo`3oool0;P3oool00`0007l0oooo0?ooo`0d0?ooo`030000003o
ool0oooo02H0oooo00100?ooo`0;0;md00000000M;oo0?ooo`3ooml0W4P00000M02oool0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`<000001@3oool00`000000oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02h0oooo00<0001o0?ooo`3oool0<P3oool3000002P0oooo000o0?ooo`03
0?oOW0180000B9cO00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0
W=oo00H0oooo00<000000?ooo`3oool0=@3oool00`0007l0oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02d0oooo00<0001o0?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0^0?ooo`030000O`3oool0
oooo03@0oooo00<000000?ooo`3oool09P3oool003l0oooo00<0omnL04P0001d_ol00`3oool01P3o
_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool09P3oool003l0oooo00<0omnL
04P0001d_ol00`3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool0
1P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
9P3oool003l0oooo00<0oooO09a8001d_ol00P3oool0203ooml0W4P00018W03Oool0oooo0=nLB000
B9`0gooo0P3oool00`2oM00007Bo0?ooo`060?ooool00000<P00000X0?ooo`00@03oool01`3o_g@0
001d0;ooo`3oool0gia80000B02Lgol00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
o`3ooomQ0?ooo`00@@3oool01@3o_g@000000000000004P0W=oo00@0oooo00D0oono07@000000000
001d0;ooo`3o0?ooof80oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?ooo`@0oooo00<0oooO09a8000000000`0000000`18W=l0oooo0?ooo`2T0?ooo`00
o`3oool40?ooo`030;md0000B9`0gooo00<0oooo00<0gibL0;ooo`3oool0Y03oool00?l0oooo103o
ool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0:@0oooo003o0?ooo`D0oooo00<0_g@0
0000000000000`0000000`18W=l0oooo0?ooo`2S0?ooo`00a@3oool01@2oM00000000000000004P0
W=oo00<0oooo00D0_g@0000000000000001d0;ooo`030?ooo`050;md0000000000000000B02Lgol0
0`3oool01@3OW4P00000000000000000M;oo0080oooo00D0gia8000000000000001d09bLB0020000
00030000M02oool0oooo00@0oooo00T0oooO09a80000000000000018W02LM0000000001d_`2oM000
0P0000000`0007@0_ooO09a80002000000030018W03Ookl0M00000@0000000<0001d0;ooo`3OW4P0
0P000000100007@0_ooo0?ooo`2oM002000000030000M02oool0oooo09T0oooo00360?ooo`030=nL
B00007@0_ooo00<0oooo01L0_g@000000000M;l0oooo0?oOW01807@0_ooo0?ooo`2oM00000000018
W03Oool0oooO09a8M02oool0oooo0=nLB0000000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`070=nL
B00007@0_ooo0?ooo`3OW4P0001d0;ooo`060?ooo`0@0?noM0000000M;oo0?ooo`3ogi`0B000001d
_`3o_kl0_ooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`80oooo00<0_gA809cOo`3oool01@3o
ool02`3OW4P0001d0;ooo`3oool0_g@000000000M;l0oooo0?oOW01807@0_ooo09T0oooo00360?oo
o`030=nLB00007@0_ooo0080oooo00<0omnL04P00018W=l00`3oool01@3ogi`0M9cO0?oOW0180000
07Bo00<0oooo00D0omnL09boo`3ogi`0B000001d_`030?ooo`0;0?noM00007@0_ooo0?ooo`3OW4P0
001d0;ooo`3oool0gia80000M02oool01P3oool01`3o_g@0001809cOo`3oool0oooO09a80000M;l0
0P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool01@3ookl0M000000000000000
07Bo00<0oooo00H0gia80000M02oool0omnL04P00018W=l30?ooo`030?oOW01dW=l0oooo09P0oooo
00360?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02Lgol01@3oool00`3o_g@0001809cO
o`050?ooo`030?noM0000000M;oo00<0oooo00/0okmd0000001d_ol0oooo0=nLB00007@0_ooo0?oo
o`3OW4P0001d0;ooo`070?ooo`060?noM000000007Bo0?oog`2LB00007Bo0P3oool01`3OW4P004RL
0=ooo`3oool0oooO09a80000M;l00P3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`80oooo00H0
gia80000M02oool0okmd0000B02LgonN0?ooo`00aP3oool00`3OW4P0001d0;ooo`020?ooo`030?no
M00007@0_ooo00D0oooo00<0okmd0000M02oool01@3oool00`3o_g@0000007Boo`030?ooo`0;0?no
M0000000M;oo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02oool02@3oool0103o_g@000000000
0000M;l20?ooo`0?0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0oono07@00018W=l0oooo
0?oog`2LB00007Bo0080oooo00H0gia80000M02oool0okmd0000M02ooonN0?ooo`00aP3oool00`3O
W4P0001809cOo`020?ooo`030?oOW0180000000000@0000000@0B9cO0?oOW01804P0W=oo0P3oool0
1P3o_g@0001809cOo`3ookl0M00007Boo`<0oooo00/0_g@00000B02Lgol0oooo0=nLB00007@0_ooo
0?ooo`3OW4P0001d0;ooo`060?ooo`070?oo_`1d0000M;oo0?ooo`3ooml0W4P0001d_`020?ooo`0?
0;md000007@0_ooo0?ooo`3ooml0W4P0001d_`3oool0oono07@00018W=l0oooo0?oog`2LB000B9cO
0080oooo00D0gia80000M02oool0omnL04P000050000000304RLg`3oool0oooo09L0oooo00340?oo
o`0:0?oog`2LB00000000000B018M9`0gkmd0000M02oool0_g@004RLg`80oooo01`0gia80000B02L
gol0oooo0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3oool0gia80018W03Oool0oooo0?noM000
000007Bo0?oog`2LB00000000000M02oool0oooo0=nLB00007@0_ooo1P3oool0403ookl0M00004RL
g`3oool0oono07@00000M;l0oooO09a80000000004Q804P0B02Lgol0oono07@00018W=l20?ooo`04
0;md0000M;l0oooo0?noM080000000P004RL0=ooo`3OW4P0001d0;ooo`3oool0_g@004RLg`80oooo
00<0gia80000B02Lgol0V@3oool00<H0oooo00P0_g@00000M02ogi`0B0000000M02oool0oooo0=nL
B080000000<0001809cOo`3oool00P3oool01@3OW4P000000000000004P0W=oo00<0oooo00D0okmd
000000000000000004RLg`030?ooo`070;md000007@0_ooo0?ooo`3OW4P0001d0;ooo`070?ooo`06
0?oo_`1d000000000000000007@0_ooo0P3oool01`2oM00004RL0=oog`2LB00000000000M02oool0
0P3oool01@3ooml0W4P0000000000000M;oo00<0oooo00<0gia80000M02oool00P3oool01@3OW4P0
00000000000004P0W=oo09X0oooo003Y0?ooo`030=nLB00007@0_ooo0200oooo00<0gia80000M02o
ool0X@3oool00>T0oooo00<0gia80000M02oool0803oool00`3OW4P0001d0;ooo`2Q0?ooo`00j@3o
ool00`3OW4P0001d0;ooo`0P0?ooo`030=nLB00007@0_ooo0:40oooo003U0?ooo`070=nLB00004P0
W=oO09a800000000001d0;ooo`0N0?ooo`050?oog`2LB00000000000M02oool0X@3oool00>D0oooo
00L0gia80000B02Lgol0oooo0;md000007@0_ooo0200oooo00<0_g@00000M02oool0X@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-9.86336, 11.9357, 0.155573, \
0.500242}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Target heavier than projectile", "Subsubsection"],

Cell[TextData[{
  "Suppose that the projectile is heavier than the target, such that ",
  StyleBox["mratio>1", "InlineInput"],
  ".  The scattering angle is then unrestricted.  We examine below the \
specific case of ",
  StyleBox["mratio\[Rule]2", "InlineInput"],
  ".  There is a maximum scattering angle, which is derived below. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotFinalMomenta[sol, Join[values, {mratio \[Rule] 2.0}], 
        180. ];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00529101 0.0147151 0.441453 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.15608 -0.0125 -8 -15.6875 ]
[.15608 -0.0125 8 0 ]
[.28836 -0.0125 -8 -15.6875 ]
[.28836 -0.0125 8 0 ]
[.42063 -0.0125 -8 -15.6875 ]
[.42063 -0.0125 8 0 ]
[.55291 -0.0125 -11 -15.6875 ]
[.55291 -0.0125 11 0 ]
[.68519 -0.0125 -11 -15.6875 ]
[.68519 -0.0125 11 0 ]
[.81746 -0.0125 -11 -15.6875 ]
[.81746 -0.0125 11 0 ]
[.94974 -0.0125 -11 -15.6875 ]
[.94974 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.0313 -31.375 ]
[.5 -0.0125 21.0313 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .10301 -19 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .1913 -19 -7.84375 ]
[-0.0125 .1913 0 7.84375 ]
[-0.0125 .27959 -19 -7.84375 ]
[-0.0125 .27959 0 7.84375 ]
[-0.0125 .36788 -19 -7.84375 ]
[-0.0125 .36788 0 7.84375 ]
[-0.0125 .45617 -10 -7.84375 ]
[-0.0125 .45617 0 7.84375 ]
[-0.0125 .54446 -19 -7.84375 ]
[-0.0125 .54446 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.6875 -30.4375 ]
[-0.01875 .30902 -19 30.4375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -72.125 0 ]
[.5 .63053 72.125 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.15608 0 m
.15608 .61803 L
s
.28836 0 m
.28836 .61803 L
s
.42063 0 m
.42063 .61803 L
s
.55291 0 m
.55291 .61803 L
s
.68519 0 m
.68519 .61803 L
s
.81746 0 m
.81746 .61803 L
s
.94974 0 m
.94974 .61803 L
s
0 .10301 m
1 .10301 L
s
0 .1913 m
1 .1913 L
s
0 .27959 m
1 .27959 L
s
0 .36788 m
1 .36788 L
s
0 .45617 m
1 .45617 L
s
0 .54446 m
1 .54446 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.15608 0 m
.15608 .00625 L
s
gsave
.15608 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(25) show
1.000 setlinewidth
grestore
.28836 0 m
.28836 .00625 L
s
gsave
.28836 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.42063 0 m
.42063 .00625 L
s
gsave
.42063 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(75) show
1.000 setlinewidth
grestore
.55291 0 m
.55291 .00625 L
s
gsave
.55291 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.68519 0 m
.68519 .00625 L
s
gsave
.68519 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(125) show
1.000 setlinewidth
grestore
.81746 0 m
.81746 .00625 L
s
gsave
.81746 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.94974 0 m
.94974 .00625 L
s
gsave
.94974 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(175) show
1.000 setlinewidth
grestore
.125 Mabswid
.05026 0 m
.05026 .00375 L
s
.07672 0 m
.07672 .00375 L
s
.10317 0 m
.10317 .00375 L
s
.12963 0 m
.12963 .00375 L
s
.18254 0 m
.18254 .00375 L
s
.20899 0 m
.20899 .00375 L
s
.23545 0 m
.23545 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.31481 0 m
.31481 .00375 L
s
.34127 0 m
.34127 .00375 L
s
.36772 0 m
.36772 .00375 L
s
.39418 0 m
.39418 .00375 L
s
.44709 0 m
.44709 .00375 L
s
.47354 0 m
.47354 .00375 L
s
.5 0 m
.5 .00375 L
s
.52646 0 m
.52646 .00375 L
s
.57937 0 m
.57937 .00375 L
s
.60582 0 m
.60582 .00375 L
s
.63228 0 m
.63228 .00375 L
s
.65873 0 m
.65873 .00375 L
s
.71164 0 m
.71164 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.76455 0 m
.76455 .00375 L
s
.79101 0 m
.79101 .00375 L
s
.84392 0 m
.84392 .00375 L
s
.87037 0 m
.87037 .00375 L
s
.89683 0 m
.89683 .00375 L
s
.92328 0 m
.92328 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.0313 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .1913 m
.00625 .1913 L
s
gsave
-0.0125 .1913 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .27959 m
.00625 .27959 L
s
gsave
-0.0125 .27959 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .36788 m
.00625 .36788 L
s
gsave
-0.0125 .36788 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .45617 m
.00625 .45617 L
s
gsave
-0.0125 .45617 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .54446 m
.00625 .54446 L
s
gsave
-0.0125 .54446 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.2) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .03679 m
.00375 .03679 L
s
0 .05886 m
.00375 .05886 L
s
0 .08093 m
.00375 .08093 L
s
0 .12508 m
.00375 .12508 L
s
0 .14715 m
.00375 .14715 L
s
0 .16922 m
.00375 .16922 L
s
0 .21337 m
.00375 .21337 L
s
0 .23544 m
.00375 .23544 L
s
0 .25751 m
.00375 .25751 L
s
0 .30166 m
.00375 .30166 L
s
0 .32373 m
.00375 .32373 L
s
0 .3458 m
.00375 .3458 L
s
0 .38995 m
.00375 .38995 L
s
0 .41202 m
.00375 .41202 L
s
0 .4341 m
.00375 .4341 L
s
0 .47824 m
.00375 .47824 L
s
0 .50031 m
.00375 .50031 L
s
0 .52239 m
.00375 .52239 L
s
0 .56653 m
.00375 .56653 L
s
0 .5886 m
.00375 .5886 L
s
0 .61068 m
.00375 .61068 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -118.281 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
91.4375 11.8438 translate
90 rotate
-91.4375 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
69.000 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
73.813 15.813 moveto
(\\220) show
77.625 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
83.625 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
88.438 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(,) show
94.438 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
100.438 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
105.250 15.813 moveto
(\\220) show
109.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(p) show
115.063 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
119.875 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.15608 .61178 m
.15608 .61803 L
s
.28836 .61178 m
.28836 .61803 L
s
.42063 .61178 m
.42063 .61803 L
s
.55291 .61178 m
.55291 .61803 L
s
.68519 .61178 m
.68519 .61803 L
s
.81746 .61178 m
.81746 .61803 L
s
.94974 .61178 m
.94974 .61803 L
s
.125 Mabswid
.05026 .61428 m
.05026 .61803 L
s
.07672 .61428 m
.07672 .61803 L
s
.10317 .61428 m
.10317 .61803 L
s
.12963 .61428 m
.12963 .61803 L
s
.18254 .61428 m
.18254 .61803 L
s
.20899 .61428 m
.20899 .61803 L
s
.23545 .61428 m
.23545 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.31481 .61428 m
.31481 .61803 L
s
.34127 .61428 m
.34127 .61803 L
s
.36772 .61428 m
.36772 .61803 L
s
.39418 .61428 m
.39418 .61803 L
s
.44709 .61428 m
.44709 .61803 L
s
.47354 .61428 m
.47354 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.52646 .61428 m
.52646 .61803 L
s
.57937 .61428 m
.57937 .61803 L
s
.60582 .61428 m
.60582 .61803 L
s
.63228 .61428 m
.63228 .61803 L
s
.65873 .61428 m
.65873 .61803 L
s
.71164 .61428 m
.71164 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.76455 .61428 m
.76455 .61803 L
s
.79101 .61428 m
.79101 .61803 L
s
.84392 .61428 m
.84392 .61803 L
s
.87037 .61428 m
.87037 .61803 L
s
.89683 .61428 m
.89683 .61803 L
s
.92328 .61428 m
.92328 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .1913 m
1 .1913 L
s
.99375 .27959 m
1 .27959 L
s
.99375 .36788 m
1 .36788 L
s
.99375 .45617 m
1 .45617 L
s
.99375 .54446 m
1 .54446 L
s
.125 Mabswid
.99625 .03679 m
1 .03679 L
s
.99625 .05886 m
1 .05886 L
s
.99625 .08093 m
1 .08093 L
s
.99625 .12508 m
1 .12508 L
s
.99625 .14715 m
1 .14715 L
s
.99625 .16922 m
1 .16922 L
s
.99625 .21337 m
1 .21337 L
s
.99625 .23544 m
1 .23544 L
s
.99625 .25751 m
1 .25751 L
s
.99625 .30166 m
1 .30166 L
s
.99625 .32373 m
1 .32373 L
s
.99625 .3458 m
1 .3458 L
s
.99625 .38995 m
1 .38995 L
s
.99625 .41202 m
1 .41202 L
s
.99625 .4341 m
1 .4341 L
s
.99625 .47824 m
1 .47824 L
s
.99625 .50031 m
1 .50031 L
s
.99625 .52239 m
1 .52239 L
s
.99625 .56653 m
1 .56653 L
s
.99625 .5886 m
1 .5886 L
s
.99625 .61068 m
1 .61068 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -133.125 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(scattered) show
108.250 15.813 moveto
(and) show
128.500 15.813 moveto
(recoil) show
158.563 15.813 moveto
(momenta) show
203.250 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .45617 m
.02499 .45617 L
.02605 .45616 L
.02729 .45615 L
.02846 .45614 L
.03053 .45611 L
.03279 .45607 L
.03527 .45601 L
.0379 .45593 L
.04262 .45574 L
.04749 .4555 L
.05205 .45521 L
.06244 .45438 L
.07305 .45327 L
.08274 .45202 L
.10458 .44841 L
.14429 .43914 L
.18248 .42713 L
.22313 .41142 L
.26226 .39389 L
.30384 .3733 L
.34391 .35215 L
.38246 .33123 L
.42346 .30898 L
.46295 .28816 L
.50092 .26914 L
.54134 .25037 L
.58025 .23398 L
.62161 .21852 L
.66146 .2056 L
.69979 .19494 L
.74057 .18543 L
.77984 .17796 L
.81759 .17222 L
.83856 .16961 L
.85779 .16757 L
.87783 .16578 L
.89648 .16443 L
.9154 .16335 L
.92605 .16288 L
.9361 .16251 L
.94592 .16223 L
.95109 .16212 L
.95656 .16202 L
.96116 .16196 L
.96361 .16193 L
.96623 .16191 L
.96864 .16189 L
.97084 .16188 L
.97189 .16187 L
Mistroke
.97305 .16187 L
.97414 .16187 L
.97515 .16187 L
.97619 .16187 L
Mfstroke
.02381 .01472 m
.06244 .07085 L
.10458 .13121 L
.14415 .18627 L
.18221 .23705 L
.22272 .28811 L
.26171 .3338 L
.30316 .37817 L
.34309 .4165 L
.3815 .44916 L
.42237 .47934 L
.46172 .50413 L
.49955 .52424 L
.53984 .54204 L
.57861 .55604 L
.61984 .56805 L
.65954 .57725 L
.69774 .58429 L
.73838 .59018 L
.77751 .59455 L
.81909 .59804 L
.83996 .59942 L
.85916 .60048 L
.8791 .60138 L
.89771 .60207 L
.91654 .6026 L
.92715 .60283 L
.9372 .60301 L
.94695 .60315 L
.95246 .60321 L
.95755 .60325 L
.96183 .60328 L
.96655 .6033 L
.96892 .60331 L
.97017 .60331 L
.97148 .60331 L
.97263 .60332 L
.97389 .60332 L
.97504 .60332 L
.97611 .60332 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00j`3oool01@3OW4P00000000000000000M;oo0140oooo00<0oooO09a8
000000000`0000001018W=l0oooo0?ooo`3OW4P30000000307Boo`3oool0oooo09l0oooo003[0?oo
o`030=nLB00007@0_ooo01<0oooo00<0_g@00018W03Oool00`3oool00`3OW9`0_ooo0?ooo`020?oo
o`030?oo_`1d0000M;oo0:40oooo003O0?ooo`040=nLB00000000000001d_`P0oooo00<0gia80018
W03Oool04`3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0080oooo00<0oono07@0
001d_ol0X@3oool00>80oooo00<0okmd001d_`3oool01P3oool00`3OW4P004RL0=ooo`0D0?ooo`03
0;md00000000000000<0000000L0B9cO0?ooo`3oool0oooo0?oo_`1d0000M;oo0:40oooo003G0?oo
o`050?oo_`1d000000000000B02Lgol01P3oool00`3ookl0M01d0;ooo`060?ooo`030=nLB000B9`0
gooo00<0oooo00P0oooO09a80000000000000000B018B00004RL0=ooo`80oooo00P0_g@000000000
0000001d0;ooo`3oool0oono07@000@0000000<0001d0;ooo`3oool00P3oool00`3ookl0M00007Bo
o`2Q0?ooo`00e`3oool01P2oM00007Bo0?ooo`3ogi`0B01809cOo`D0oooo00<0gia80000M02oool0
1P3oool00`3OW4P004RL0=ooo`020?ooo`0D0?oog`2LB000000007Boo`3oool0oono07@000000000
001d0;ooo`2oM0000000001d_`3oool0omnL04P0M02oool0oooo0;mdB02Lgol80?ooo`030?oo_`1d
0000M;oo0:40oooo003F0?ooo`030?oog`2LB000M;oo0080oooo00<0gia80018W03Oool00P3oool0
103ookl0M00000000018W=l70?ooo`030=nLB000B9`0gooo0080oooo00<0omnL04P00018W=l00P3o
ool01`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool01@3ogi`0M9cO0?ooo`3ookl0M000
0080000000<007Bo0?ooo`3oool00`3oool00`3ookl0M00007Boo`2Q0?ooo`00eP3oool00`3ookl0
M00007Boo`020?ooo`030?oo_`1d0000M;oo00<0oooo00<0oooO09bL_`3oool01`3oool00`3OW4P0
04RL0=ooo`020?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?noM00004P0
W=oo00H0oooo00H0gia80000M02oool0oono07@00018W=l40?ooo`030?oo_`1d0000M;oo0:40oooo
003F0?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool00`3oool00`3o_g@0B9cO0?oo
o`060?ooo`030=nLB000B9`0gooo0080oooo00<0okmd0000001d_ol00P3oool01`3ooml0W4P0001d
_`3oool0okmd0000M02oool01@3oool01`3ookl0M00004RLg`3oool0oooO09a80000M;l0103oool0
0`3ookl0M00007Boo`2Q0?ooo`00eP3oool00`3ooml0W4P0001d_`030?ooo`080=nLB00007@0_ooo
0?oOW01dW=l0oooo0=nLB000M;l70?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@00018W=l0
0P3oool01P3ooml0W4P0001d_`3oool0omnL04P000D0000000P0B9cO0?oo_`1d0000B9cO0?ooo`3o
oml0W4P004RLg`@0oooo00<0oono07@0001d_ol0X@3oool00=L0oooo00<0_g@00000000000001000
00001P1d_ol0oooo0=nLB0000000001809cOo`L0oooo00<0gia80018W03Oool00`3oool01`3o_g@0
001809cOo`3oool0okmd00000000M;l00P3oool0302oM000B9cO0?ooo`3oool0gia80000B02Lgol0
oooo0;md0000M;l0oooo0?noM080000000L004RL0=ooo`3oool0oooo0?oo_`1d0000M;oo0:40oooo
003G0?ooo`030=nLB00007@0_ooo0080oooo00<0oono07@00018W=l0303oool00`3OW4P004RL0=oo
o`040?ooo`030?noM000000000000080000000D007Bo0?ooo`3oool0oooo0=nLB002000000060000
B02Lgol0oooo0?ooo`3ooml0W4P00P0000000`1d_ol0oooo0?ooo`030?ooo`030?oo_`1d0000M;oo
0:40oooo003G0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@0001d_`3oool02`3oool00`3OW4P0
04RL0=ooo`070?ooo`030?oog`2LB00007Bo01@0oooo00<0oono07@0001d_ol0X@3oool00=P0oooo
00<0_g@00018W03Oool00P3oool00`3OW4P007Bo0?ooo`0;0?ooo`030=nLB000B9`0gooo00L0oooo
00<0oooO09a80000M;l0503oool00`3ookl0M00007Boo`2Q0?ooo`00f03oool00`3ookl0M00007Bo
o`020?ooo`030=nLB000M;l0oooo00/0oooo00<0gia80018W03Oool01`3oool00`3ooml0W4P0001d
_`0D0?ooo`030?oo_`1d0000M;oo0:40oooo003I0?ooo`060?noM00004P0W=oo0?ooo`2oM000M;oo
303oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d00000000001d_a@0oooo00<0oono07@0001d
_ol0X@3oool00=X0oooo00D0omnL04P00000000004RL0=ooo`0<0?ooo`050=nLB000000000000000
001d_ol01@3oool00`3ooml0W4P0001d_`0B0?ooo`050=nLB000000000000000001d_ol0X@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00600oooo00D0omnL04P000000000001d0;ooo`0M0?ooo`030;md000000000000
00@0000000@0B9cO0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`0G0?ooo`030?oog`2LB000
00000080000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo01/0oooo00<0
gia804RLg`3oool00`3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo01D0oooo00H0
_g@000000000000000000018W03Oool30?ooo`050?oOW018000000000000M02oool0103oool01@3o
gi`0B0000000000007@0_ooo0140oooo00P0_g@000000000000000000018W03Oool0oooo0;md00H0
000000@0B9cO0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`0@0?ooo`090;md000000000000
00000000B9`0gooo0?ooo`3ooml0W4P000<0000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B000
0000000007@0_ooo0140oooo00H0_g@000000000000000000018W03Oool40?ooo`030=nLB018W=l0
oooo00<0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`0X0?ooo`00G`3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool07@3oool00`3OW4P0000000000003000000050018W03O
ool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0G0?ooo`030?oog`2LB7@0_ooo0080oooo
00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool06P3oool00`3o
gi`0B4RL0=ooo`030?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool05P3oool00`3o
gi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`0A0?ooo`030?oOW0180000M;oo00@0oooo00<0gia800000000
00000`0000001@00B9`0gooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool04@3oool00`3o
gi`0B00007Boo`030?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool00P3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool04@3oool00`3ogi`0B00007Boo`060?ooo`030?oOW018
B9`0gooo00<0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0X0?ooo`00G`3oool0
0`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01d0oooo00<0gia804RLg`3oool00P3oool0
0`3OW7@0W=oo0?ooo`040?ooo`030?oog`2LB7@0_ooo01`0oooo00<0oooO09a8M02oool00P3oool0
0`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01T0oooo00<0oooO09a8B02Lgol0203oool0
0`3ooml0W4Qd0;ooo`0F0?ooo`030?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool0
1`2oM00007Bo0?ooo`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0@0?ooo`030?oO
W0180000M;oo00D0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`040?ooo`030?oo
g`2LB7@0_ooo0140oooo00<0omnL04P0001d_ol0203oool00`3ooml0W4Qd0;ooo`020?ooo`030;md
0000B9`0gooo0080oooo00<0_g@0001d_`3oool0403oool00`3ogi`0B00007Boo`060?ooo`030?oo
g`2LB4P0W=oo00P0oooo00<0oooO09a8M02oool0:03oool005h0oooo00<0oono07@00018W=l00`3o
ool00`3OW4P0001d0;ooo`0N0?ooo`030=nLB01d_ol0oooo00T0oooo00<0_g@007Boo`3oool0703o
ool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool06P3oool00`3OW4P0
B9cO0?ooo`080?ooo`030;md001d_ol0oooo01D0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0
M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0@
0?ooo`030?oOW0180000M;oo00H0oooo00<0gia807Boo`3oool02@3oool00`2oM000M;oo0?ooo`0@
0?ooo`030?oOW0180000M;oo00T0oooo00H0_g@007Boo`3oool0oono07@00018W=l30?ooo`030=nL
B00007@0_ooo0100oooo00<0omnL04P0001d_ol01`3oool00`3OW4P0B9cO0?ooo`080?ooo`030;md
001d_ol0oooo02L0oooo001N0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol07`3o
ool00`2oM4P0W=oo0?ooo`070?ooo`030?oog`2LB000B9cO01`0oooo00L0oooO09a80018W=l0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol06P3oool00`3ogi`0B4RL0=ooo`070?ooo`03
0?oog`2LB000B9cO01H0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`06
0?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0@0?ooo`030?oOW0180000
M;oo00L0oooo00<0_gA809cOo`3oool01`3oool00`3ooml0W4P004RLg`0A0?ooo`030?oOW0180000
M;oo00P0oooo00L0oooO09a80018W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
403oool00`3ogi`0B00007Boo`070?ooo`030?oOW018B9`0gooo00L0oooo00<0oooO09a80018W=l0
:03oool005h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0O0?ooo`030?oog`2L
B7@0_ooo00L0oooo00<0gia80000001d_ol0703oool01`3OW4P0000007Boo`3oool0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`0J0?ooo`030?oog`2LB4P0W=oo00L0oooo00<0gia80000001d
_ol05P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02L
gol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0100oooo00<0omnL04P0001d_ol01`3oool0
0`3ooml0W4Qd0;ooo`070?ooo`030=nLB0000000M;oo0140oooo00<0omnL04P0001d_ol0203oool0
1`3OW4P0000007Boo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0@0?ooo`030?oO
W0180000M;oo00L0oooo00<0oooO09a8B02Lgol01`3oool00`3OW4P0000007Boo`0X0?ooo`00GP3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0200oooo00<0okmd001d_`3oool00P3o
ool01`3Ogml0oooo0?oo_`1d000000000000M02oool0603oool02`3Ogml0oooo0?oo_`1d00000000
0000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol06`3oool00`3OW4P007Bo
0?ooo`020?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0F0?ooo`030?oOW0180000M;oo
00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0
okmd0000B02Lgol0403oool00`3ogi`0B00007Boo`080?ooo`030?noM000M;l0oooo0080oooo00L0
gmoO0?ooo`3ookl0M0000000000007@0_ooo0140oooo00<0omnL04P0001d_ol0103oool02`3Ogml0
oooo0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
403oool00`3ogi`0B00007Boo`080?ooo`030=nLB000M;l0oooo0080oooo00L0gmoO0?ooo`3ookl0
M0000000000007@0_ooo02P0oooo001N0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol0803oool00`3ookl0M01809cOo`020?ooo`060=nLB000000000000000000004P0W=oo6@3oool0
1P3OW4P00000000000000000001809cOo`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`0K0?ooo`030?oo_`1d07@0_ooo0080oooo00H0gia800000000000000000000B02LgolG0?oo
o`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0
B00007Boo`<0oooo00<0okmd0000B02Lgol0403oool00`3ogi`0B00007Boo`080?ooo`030?oo_`1d
04P0W=oo0080oooo00H0gia800000000000000000000B02LgolB0?ooo`030?oOW0180000M;oo00@0
oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0403oool00`3ogi`0B00007Boo`080?ooo`030?oo_`1d07@0_ooo0080oooo00H0gia8
00000000000000000000B02LgolY0?ooo`00GP3oool00`3ookl0M00004RLg`030?ooo`030=nLB000
07@0_ooo0200oooo00<0oono07@00018W=l00P3oool0103ookl0M00000000018W=lK0?ooo`040?oo
_`1d0000000004RLg`@0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0F0?ooo`03
0?oOg`3Oool0oooo00<0oooo00P0_g@007Boo`3oool0oooo0?oo_`1d0000000004RLgaT0oooo00<0
omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000
B9cO0`3oool00`3OW4P0001d0;ooo`0@0?ooo`030?oOW0180000M;oo00P0oooo00<0oono07@00018
W=l00P3oool0103ookl0M00000000018W=lD0?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@0
00000000B9cO103oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0100oooo00<0omnL
04P0001d_ol00`3oool00`3ogml0gooo0?ooo`030?ooo`080;md001d_ol0oooo0?ooo`3ookl0M000
00000018W=l[0?ooo`00G`3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo01/0oooo
00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`030?ooo`030;md001d_ol0oooo01`0oooo
00<0_g@007Boo`3oool01@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo01H0oooo
00<0oono07A8W03Oool00`3oool00`3o_g@007Bo0?ooo`020?ooo`030;md001d_ol0oooo01T0oooo
00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0100oooo00<0omnL04P0001d_ol00`3oool00`3o
_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00<0oooo00<0_g@007Boo`3oool0503oool00`3o
gi`0B00007Boo`050?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`0@0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07A8W03Oool00`3oool00`3o
_g@007Bo0?ooo`020?ooo`030;md001d_ol0oooo02/0oooo001O0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`0L0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool30?ooo`06
0?oOW018000000000000000004P0W=oo6@3oool01P3ogi`0B000000000000000001809cOo`80oooo
00L0oono07@0001d_ol0oooo0?noM00007@0_ooo01P0oooo00<0_g@00000000000000`0000001P00
07@0_ooo0?ooo`3oool0omnL04P00080000000<0001809cOo`3oool04`3oool01@3ooml0W4P00000
00000000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00l0oooo00D0oooO09a800000000000007Boo`030?ooo`080?oo
g`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool30?ooo`060?oOW018000000000000000004P0W=oo
3`3oool01@3ooml0W4P0000000000000M;oo00D0oooo00H0omnL04P00000000000000000B02Lgol2
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0?0?ooo`050?oog`2LB00000000000001d
_ol0103oool00`2oM000000000000003000000060000M02oool0oooo0?ooo`3ogi`0B0000P000000
0`0004P0W=oo0?ooo`0W0?ooo`00H03oool01@3ookl0M0000000000007@0_ooo01h0oooo00H0oooO
09a80000000000000000M02oool50?ooo`050;md00000000000000000018W=l06P3oool01@2oM000
0000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`0I0?ooo`030?noM0000000
000000@0000000D0M;oo0?ooo`3oool0oooo0;md00030000000304RLg`3oool0oooo01@0oooo00<0
okmd0000001d_ol01@3oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000
001d0;ooo`0B0?ooo`030?noM0000000M;oo00@0oooo00H0oooO09a80000000000000000M02oool5
0?ooo`050;md00000000000000000018W=l04@3oool00`3o_g@0000007Boo`060?ooo`050;md0000
0000000000000018W=l00`3oool01@3ookl0M0000000000007@0_ooo0180oooo00<0okmd0000001d
_ol0103oool00`3o_g@00000000000040000000507Boo`3oool0oooo0?ooo`2oM0000`0000000`18
W=l0oooo0?ooo`0V0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00CP3oool0
1@3ogi`0B0000000000007@0_ooo00P0ooooo`00000b000002@0oooo001=0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0R0?ooo`00C@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0280oooo001<0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
0`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool0
04`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030000003oool0oooo
02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00C03oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo0`0000050?ooo`030000003oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool02P3oool3000002@0oooo001<0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo02@0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00C03oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`0T0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool004`0oooo00<0oono
07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool08`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001=0?oo
o`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01P3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo0280oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?oo
o`00C@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo0280oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R
0?ooo`00CP3oool01@3ookl0M0000000000007@0_ooo00P0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`0Q0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool0
05/0oooo0`00000:0?ooo`030000003oool0oooo0200oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02`3oool20000
02@0oooo001K0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0803oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo
001K0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool07P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool07@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?oo
o`030000003oool0oooo0280oooo001K0?ooo`<000003@3oool00`000000oooo0?ooo`0M0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo00/0oooo0P00000T0?ooo`00F`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo01`0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01/0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303o
ool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo01/0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`00
0000oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01T0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0R0?ooo`00F`3oool300000180oooo00<000000?ooo`3oool0603oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0;0?ooo`800000903oool005/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0G
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool00400oooo00D0omnL04P0
00000000001d0;ooo`030?ooo`050?noM0000000M;oo0?ooo`2oM0001P0000000`18W=l0oooo0?oo
o`050?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0<0?ooo`030000003oool0oooo0280oooo000o0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d
0;ooo`020?ooo`030?noM0000000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo
0?ooo`050?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool05P3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo0280oooo000o0?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool01`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00D0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<0
00000?ooo`3oool08P3oool003h0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`08
0?ooo`030=nLB01d_ol0oooo00T0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0D
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool003h0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo00P0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0D0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?oo
o`3oool08P3oool003h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`03
0?oog`2LB7@0_ooo00P0oooo0`00000G0000O`030000000007l0001o0?l0001o4P0007l3000002@0
oooo000n0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3o_g@007Bo
0?ooo`070?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool04P3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo0280oooo000n0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol02P3oool00`3ookl0M01809cOo`070?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool04@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000n
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`3ookl0M00004RLg`07
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool04@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo0280oooo000o0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d
_`3oool01@3oool00`3o_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00L0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?oo
o`3oool08P3oool003l0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0oooO
09a8B02Lgol0oooo0?oo_`1d0000001d0;ooo`L0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`0?0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool00400oooo
00D0oono07@000000000001d0;ooo`080?ooo`060?oog`2LB00000000000000007@0_ooo203oool3
000001`0oooo00<000000?ooo`3oool03P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0;0?ooo`800000903oool0
05/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0>0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`0=0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0<0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`0;0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<00000
0?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0:0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo0`00000Q0?ooo`03
0000003oool0oooo00T0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02`3oool2000002@0oooo001K0?ooo`030000
003oool0oooo0240oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo0280oooo001K0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0203o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool01`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo0280oooo001K0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool01`3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0
oooo00<0001o0?ooo`3oool04`3ooolK000000H0oooo00<000000?ooo`3oool08P3oool005/0oooo
00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`050?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`070?ooo``000004P3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`<000009@3oool00`000000oooo0?oo
o`050?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool0100007l0oooo0?oo
o`3oool6000001h0oooo00<0001o0?ooo`3oool02`3oool2000002@0oooo000[0?ooo`030?oog`2L
W;l0oooo00P0oooo00<0_ibL0=ooo`3oool08P3oool00`000000oooo0?ooo`0V0?ooo`030000003o
ool0oooo00@0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0T0?ooo`H00000903o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000[0?ooo`030?oOW0180000
000000P0000000<004RL0=ooo`3oool01`3oool01@3ogi`0B0000000000007@0_ooo00<0oooo00<0
okmd0000001d_ol01@3oool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo02L0oooo
00<000000?ooo`3oool00`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01h0oooo
1P0000020?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0R0?ooo`00:`3oool00`3OW4P0000000000008000000030018W03Oool0oooo00H0oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol01@3oool0103ooml0
W4P00018W03Oool70?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool00P3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01P0oooo1P0000080?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00;03oool00`2oM7@0
_ooo0?ooo`020?ooo`030?oog`2LW;l0oooo0080oooo00<0_ibL0=ooo`3oool01P3oool00`2oM000
04RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo00@0oooO09a80000B9`0gooo1`3oool00`00
0000oooo0?ooo`0X0?ooo`030000003oool0oooo0080oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0B0?ooo`H000003P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool002/0oooo00<0gia8001d_`3oool0103oool00`3OW9`0
_ooo0?ooo`090?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01@3oool00`2oM000
0000000000050000000307Boo`3oool0oooo00@0oooo00<000000?ooo`3oool0:@3oool01@000000
oooo0?ooo`3oool0001o02L0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool03@3oool5000001@0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000[0?ooo`03
0;md00000000M;oo00<0oooo00<0omnL04Q8W03Oool02P3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00D0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`L0oooo00<0
00000?ooo`3oool0:P3oool010000000oooo0?ooo`0007lW0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00T0oooo
1000000I0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0R0?ooo`00:`3oool00`3o_g@00000000000040000000307Boo`3oool0oooo00T0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=nLB01d_ol0oooo0?oog`2LB000
04RL0=ooo`070?ooo`<00000:`0007l00`000000001o0000O`2W0000O`@00000DP0007l3000002@0
oooo000/0?ooo`060?oOW018000000000000000004P0W=oo303oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00H0oooo00L0oooO09a8M02oool0oooO09a80000B9`0gooo00L0oooo00<0
00000?ooo`3oool0:`3oool00`000000oooo0000O`0W0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`040000O`3oool0oooo0?ooo`@00000
8@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0
8P3oool003h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`060?oOW018
M;l0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`0/0?ooo`030000000007l0oooo02H0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo1000000U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000
oooo0?ooo`0R0?ooo`00;P3oool00`3OW7@0W=oo0?ooo`030?ooo`030?oOW018B9`0gooo00L0oooo
00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?ooo`050=nLB01d_ml0W4P00018W03O
ool01`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0240oooo100000000`0007l0
oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000
oooo0?ooo`0R0?ooo`00;@3oool00`3o_g@0B9cO0?ooo`050?ooo`030;mdM02oool0oooo00L0oooo
00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`080?ooo`050?oog`2LB4P0B4P00018W03O
ool01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`000000oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01d0oooo100000040?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo0280oooo000]0?ooo`060=nLB01d_ol0oooo0?noM000B9`0gooo0P3oool00`3O
_kl0gooo0?ooo`070?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`040?oOW018
000004RL0=ooo`L0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0000000U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0J0?ooo`<00000
203oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0R0?ooo`00;@3oool02@3OW4P007Bo0?oo_`1dM9`0W4P004RLg`3o
ool0okmd04RLg`0:0?ooo`050?oo_`1d000000000000M02oool0303oool00`3OW4P004RL0=ooo`07
0?ooo`<00000;@3oool00`0007l0oooo0000000U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0G0?ooo`<000002`3oool00`0007l0oooo0?ooo`0V
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02`3oool2000002@0oooo000]0?oo
o`080?oo_`1d0000001d0;ooo`3ogi`0B00000000000M;lV0?ooo`030000003oool0oooo02d0oooo
00@0001o0?ooo`3oool00000903oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool0503oool3000000h0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo
00<000000?ooo`3oool0;@3oool01@0007l0oooo0?ooo`3oool0000002<0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0140oooo0`00000A0?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
0P3oool00`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0?0?ooo`800000503oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00=03o
ool0103ooml0W4P00000M02ooolS0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
0`3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0=0?ooo`8000005P3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00<@3o
ool01@3o_g@000000000000004P0W=oo02D0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`030?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo00X0oooo0`00000H0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo
000^0?ooo`040;md000000000000001d_bT0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`040?ooo`030000003oool0oooo01h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo00P0oooo0P00000K0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo
000Z0?ooo`050?oOW018000000000000001d_ol0;03oool3000002d0oooo00<0001o0?ooo`3oool0
1@3oool00`000000oooo0?ooo`0M0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`050?ooo`<000007@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02`3oool2000002@0oooo000W0?ooo`050=nLB000
000000000018W03Oool0;`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo00H0oooo
00<000000?ooo`3oool0703oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool00`3oool200000200oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool002D0oooo00<0
omnL04P00018W=l0<`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo00H0oooo00<0
00000?ooo`3oool0703oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00@0
001o0?ooo`3oool0oooo0P00000R0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000[0?ooo`030?oog`2L
W;l0oooo00P0oooo00<0_ibL0=ooo`3oool08P3oool00`000000oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo00L0oooo00<000000?ooo`3oool06`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00@0001o0000000000000000903oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00:`3o
ool00`3ogi`0B00000000008000000030018W03Oool0oooo0280oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`080?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0T0?ooo`8000009`3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00:`3oool0
0`3OW4P0000000000008000000030018W03Oool0oooo0280oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`090?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0Q0?ooo`<0000000<0oooo0000O`3oool09P3oool00`0007l0oooo0?ooo`0V
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R
0?ooo`00;03oool00`2oM7@0_ooo0?ooo`020?ooo`030?oog`2LW;l0oooo0080oooo00<0_ibL0=oo
o`3oool08P3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo00T0oooo00<000000?oo
o`3oool06@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01l0oooo0P0000040?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool002/0oooo00<0gia8001d_`3o
ool0103oool00`3OW9`0_ooo0?ooo`0V0?ooo`<00000;@3oool00`0007l0oooo0?ooo`0:0?ooo`03
0000003oool0oooo01P0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0M0?ooo`80
00001P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0;0?ooo`800000903oool002/0oooo00<0_g@00000001d_ol0
0`3oool00`3ogi`0B4RL0=ooo`0<0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0
000007Boo`030?ooo`050?oOW018000000000000001d_ol0203oool00`000000oooo0?ooo`0]0?oo
o`030000O`3oool0oooo00/0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo01X0oooo0`0000080?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?oo
o`3oool08P3oool002/0oooo00<0okmd000000000000100000000`1d_ol0oooo0?ooo`0:0?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo0080oooo00L0omnL
04P00018W=l0oooo0?oo_`1d0000M;oo00L0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0H0?ooo`8000002`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo
000/0?ooo`060?oOW018000000000000000004P0W=oo3@3oool00`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00D0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`060?ooo`03
0000003oool0oooo02d0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0F0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool05P3oool2000000d0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool0303oool00`000000oooo0?ooo`0R0?ooo`00?P3oool00`3ookl0M00004RLg`030?ooo`030=nL
B00007@0_ooo00D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030000
003oool0oooo02d0oooo00<0001o0?ooo`3oool03@3oool00`000000oooo0?ooo`0E0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0503oool2000000l0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0R0?ooo`00?P3oool00`3ogi`0B00007Boo`030?ooo`030?noM000
04P0W=oo00D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`030000003o
ool0oooo02d0oooo00<0001o0?ooo`3oool03P3oool00`000000oooo0?ooo`0D0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool04P3oool200000140oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303o
ool00`000000oooo0?ooo`0R0?ooo`00;@3oool00`3ooml0W7BL0=ooo`050?ooo`030=nLW02oool0
oooo00H0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`030?oOW0180000
M;oo00<0oooo00<0_g@00000B02Lgol01P3oool3000003l0001o00<000000000O`0007l0CP0007l2
000009T0001o0`00000T0?ooo`00;@3oool00`3o_g@0000000000005000000030000M02oool0oooo
00H0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oo_`1d00000000
04RLg`3oool0okmd00000000M;l70?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
3`3oool00`000000oooo0?ooo`0C0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
3P3oool2000001D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00=@3o
ool00`3OW9`0_ooo0?ooo`060?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3o
ool00`2oM00004RL0;nLB002000000030018W03Oool0oooo00H0oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0oooo0180oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0<0?ooo`8000005`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo0280oooo000n0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
0`3ogi`0B00007Boo`0;0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool04@3oool0
0`000000oooo0?ooo`0A0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02P3oool2
000001T0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00?`3oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol02P3oool00`00
0000oooo0?ooo`0]0?ooo`030000O`3oool0oooo0180oooo00<000000?ooo`3oool0403oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00P0oooo0P00000K0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool003l0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo00T0oooo00<0okmd00000000M;l02@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0
oooo01<0oooo00<000000?ooo`3oool03`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo00H0oooo0P00000M0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool0
0400oooo00D0oono07@000000000001d0;ooo`0;0?ooo`050?oog`2LB00000000000M02oool01P3o
ool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01@0oooo00<000000?ooo`3oool03P3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00@0oooo0P00000O0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool08P3oool00300oooo00H0oooO09a80000B9`0gooo0?oo
g`2LW;lU0?ooo`<00000;@3oool00`0007l0oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`020?ooo`8000008@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0;0?ooo`800000903oool00300oooo00H0oono07@00000B9`0gooo0;mdB02LgolU0?ooo`03
0000003oool0oooo02d0oooo00<0001o0?ooo`3oool05@3oool00`000000oooo0?ooo`0=0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool00P00000S0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0
oooo00<000000?ooo`3oool08P3oool00340oooo00@0gia8000000000000M;oo9P3oool00`000000
oooo0?ooo`0]0?ooo`030000O`3oool0oooo01H0oooo00<000000?ooo`3oool0303oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`000000000002D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000
oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01L0oooo
00<000000?ooo`3oool02`3oool00`0007l0oooo0?ooo`0T0?ooo`8000009`3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo02d0oooo00<0
001o0?ooo`3oool0603oool00`000000oooo0?ooo`0:0?ooo`030000O`3oool0oooo0280oooo0P00
00000`3oool0001o0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3o
ool005/0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0I0?ooo`030000003oool0
oooo00T0oooo00<0001o0?ooo`3oool08@3oool01@000000oooo0?ooo`3oool0001o02L0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo0?ooo`0]
0?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool0203oool00`0007l0oooo0?ooo`0O
0?ooo`800000103oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<00000
0?ooo`3oool08P3oool005/0oooo0`00000]0?ooo`030000O`3oool0oooo01X0oooo00<000000?oo
o`3oool0203oool00`0007l0oooo0?ooo`0M0?ooo`8000001P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo00/0oooo0P00000T0?ooo`00:`3oool00`3ooml0W9bo0?ooo`080?ooo`03
0;nLW03Oool0oooo0280oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0K0?ooo`03
0000003oool0oooo00L0oooo00<0001o0?ooo`3oool06`3oool2000000P0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000[0?ooo`030?oO
W0180000000000P0000000<004RL0=ooo`3oool08P3oool00`000000oooo0?ooo`0]0?ooo`030000
O`3oool0oooo01`0oooo00<000000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0J0?ooo`030000
003oool0oooo00P0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo0280oooo000[0?ooo`030=nLB0000000000000P0000000<004RL0=ooo`3oool08P3o
ool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01d0oooo00<000000?ooo`3oool01@3o
ool00`0007l0oooo0?ooo`0H0?ooo`8000002`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool08P3oool002`0oooo00<0_gAd0;ooo`3oool00P3oool0
0`3ooml0W9bo0?ooo`020?ooo`030;nLW03Oool0oooo0280oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0N0?ooo`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool05P3oool2
000000d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo0280oooo000[0?ooo`030=nLB000M;l0oooo00@0oooo00<0gibL0;ooo`3oool09P3oool00`00
0000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01l0oooo00<000000?ooo`3oool00`3oool00`00
07l0oooo0?ooo`0D0?ooo`8000003`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool002/0oooo00<0_g@00000001d_ol00`3oool00`3ogi`0
B4RL0=ooo`0W0?ooo`<00000;@3oool00`0007l0oooo0?ooo`0P0?ooo`030000003oool0oooo0080
oooo00<0001o0?ooo`3oool04P3oool200000140oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0;0?ooo`800000903oool002/0oooo00<0okmd000000000000100000000`1d_ol0oooo
0?ooo`0V0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool08@3oool01@000000oooo
0?ooo`3oool0001o0180oooo0P00000C0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool0303oool00`000000oooo0?ooo`0R0?ooo`00;03oool01P3ogi`0B000000000000000001809cO
o`h0oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0gia8
000000000000000004RLg`080?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool08P3o
ool010000000oooo0?ooo`0007l@0?ooo`8000005@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool003l0oooo00L0oono07@0001d_ol0oooo
0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool01`3OW4P0000007Boo`3oool0oooO
09a80018W=l01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02<0oooo00<00000
0?ooo`0007l03P3oool2000001L0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo0280oooo000a0?ooo`040?oOW018000004RL0=ooo`X0oooo00<0_g@0
0018W03Oool00P3oool00`2oM00007Bo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0okmd
0018W03Oool01P3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02<0oooo00<00000
0?ooo`0007l0303oool2000001T0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo0280oooo000`0?ooo`050=nLB000B9`0gmnL04Q8W03Oool02@3oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@0001d_`3oool00P3oool0
0`2oM00004RL0=ooo`060?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0903oool0
0`000000001o0?ooo`0:0?ooo`030000003oool0oooo01T0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000_0?ooo`060;md0018W=l0oooo
0?oOW018B9`0gooo2@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0
omnL04P0M02oool0oooo0;md00000000M;oo00L0oooo00<000000?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000003oool0oooo00L0oooo0P00000L0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00;P3oool01`3OW4P0M;oo
0?ooo`3oool0omnL04Q8W03Oool02@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo
00L0oooo00D0oono07A8B0180000000004RLg`080?ooo`<00000EP0007l00`000000001o0000O`04
0000O`800000c00007l3000002@0oooo000]0?ooo`030=nLB0000000000000D0000000<0001d0;oo
o`3oool01P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00@0omnL04P0
0000000007Bo2@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`0000000`3oool200000200oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo0280oooo000b0?ooo`030?oOW018B9`0gooo00T0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`060?oOW0180000001d0;ooo`3o_g@007Bo203o
ool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool00`00
000R0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?oo
o`0R0?ooo`00?P3oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00H0oooo00<0_g@0
0000M02oool00P3oool00`2oM000B9cO0?ooo`060?ooo`030000003oool0oooo02d0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo00000002000002<0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000d0?ooo`040?oog`2LB000001d
0;ooo`L0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md0000B9`0
gooo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02D0oooo0P0000030?ooo`030000003oool0oooo0200oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000a0?ooo`050?noM00000000000
0000B02Lgol02@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool01`3oool01`3o_g@00000
07Boo`3oool0oono07@0001d_ol01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02<0oooo0P0000000`0007l0oooo0?ooo`030?ooo`030000003oool0oooo01l0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo000^0?ooo`04
0;md000000000000001d_`h0oooo00D0oono07@000000000001d0;ooo`090?ooo`050?noM0000000
000000000018W=l0203oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo0240oooo0P00
00020?ooo`030000O`3oool0oooo00@0oooo00<000000?ooo`3oool07P3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool002X0oooo00D0omnL
04P000000000000007Boo`0/0?ooo`<00000;@3oool00`0007l0oooo0?ooo`0O0?ooo`800000103o
ool00`0007l0oooo0?ooo`050?ooo`030000003oool0oooo01d0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0;0?ooo`800000903oool002L0oooo00D0gia800000000000004RL0=oo
o`0_0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0703oool3000000H0oooo00<0
001o0?ooo`3oool01P3oool00`000000oooo0?ooo`0L0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`009@3oool00`3ogi`0B00004RLg`0c
0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool06P3oool2000000T0oooo00<0001o
0?ooo`3oool01`3oool2000001`0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo0280oooo000[0?ooo`030?oog`2LW;l0oooo00P0oooo00<0_ibL0=oo
o`3oool08P3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01P0oooo0P00000;0?oo
o`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool002/0oooo00<0omnL04P00000
0000200000000`00B9`0gooo0?ooo`0R0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3o
ool05@3oool3000000d0oooo00<0001o0?ooo`3oool02P3oool00`000000oooo0?ooo`0H0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00
:`3oool00`3OW4P0000000000008000000030018W03Oool0oooo0280oooo00<000000?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0C0?ooo`800000403oool00`0007l0oooo0?ooo`0;0?ooo`030000
003oool0oooo01L0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo0280oooo000/0?ooo`030;mdM02oool0oooo0080oooo00<0oooO09bL_`3oool00P3o
ool00`2oW9`0gooo0?ooo`0R0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0403o
ool300000180oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0F0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00:`3oool0
0`3OW4P007Bo0?ooo`040?ooo`030=nLW02oool0oooo02H0oooo0`00000]0?ooo`030000O`3oool0
oooo00h0oooo0P00000E0?ooo`030000O`3oool0oooo00d0oooo00<000000?ooo`3oool05@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00/0oooo0P00000T0?ooo`00:`3oool00`2o
M000000007Boo`030?ooo`030?oOW018B9`0gooo02L0oooo00<000000?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0<0?ooo`8000005`3oool00`0007l0oooo0?ooo`0>0?ooo`030000003oool0oooo
01@0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo
0280oooo000[0?ooo`030?noM0000000000000@0000000<0M;oo0?ooo`3oool09P3oool00`000000
oooo0?ooo`0]0?ooo`030000O`3oool0oooo00T0oooo0`00000I0?ooo`030000O`3oool0oooo00l0
oooo00<000000?ooo`3oool04`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0
oooo00<000000?ooo`3oool08P3oool002`0oooo00H0omnL04P00000000000000000B02LgolY0?oo
o`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool01`3oool2000001`0oooo00<0001o0?oo
o`3oool0403oool00`000000oooo0?ooo`0B0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo00@0oooo0`00000N0?ooo`030000O`3oool0oooo0140oooo00<000000?ooo`3o
ool04@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3o
ool08P3oool005/0oooo00<000000?ooo`3oool0;@3oool0100007l0oooo0?ooo`3oool300000240
oooo00<0001o0?ooo`3oool04P3oool00`000000oooo0?ooo`0@0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00;@3oool00`3ooml0W7BL
0=ooo`050?ooo`030=nLW02oool0oooo02<0oooo0`00000]0?ooo`040000O`0000000000000002@0
oooo00<0001o0?ooo`3oool04`3oool200000100oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0;0?ooo`800000903oool002d0oooo00<0okmd0000000000001@0000000`0007@0_ooo
0?ooo`0S0?ooo`030000003oool0oooo02/0oooo0`00000W0?ooo`030000O`3oool0oooo01D0oooo
00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool08P3oool003D0oooo00<0gibL0;ooo`3oool05P3oool01P2oM00000000000
0000000004RL0=ooo`L0oooo00<000000?ooo`3oool0:03oool300000080oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0F0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001?0?ooo`030?oO
W0180000M;oo00T0oooo00<000000?ooo`3oool0903oool4000000D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0G0?ooo`030000003oool0oooo00/0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001?0?ooo`030?oOW018
0000M;oo00T0oooo00<000000?ooo`3oool0803oool4000000T0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0H0?ooo`030000003oool0oooo00X0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001?0?ooo`030?oOW0180000
M;oo00T0oooo00<000000?ooo`3oool06`3oool5000000d0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0I0?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001?0?ooo`030?oOW0180000M;oo
00T0oooo00<000000?ooo`3oool03`3oool<00000180oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0J0?ooo`8000002@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool004l0oooo00<0omnL04P0001d_ol02@3oool3000000D0
001o2P00001U0000O`030000000007l0001o0;@0001o0`00000T0?ooo`00C`3oool00`3ogi`0B000
07Boo`090?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0M0?ooo`8000001P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool08P3oool004l0oooo00<0omnL04P0001d_ol02@3oool00`000000oooo0?oo
o`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool07`3oool00`000000oooo0?oo
o`030?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?oo
o`0R0?ooo`00C`3oool00`3ogi`0B00007Boo`090?ooo`030000003oool0oooo02d0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0P0?ooo`030000003oool0oooo0080oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001?0?oo
o`030?oOW0180000M;oo00T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo0240oooo0P0000020?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00C@3oool01@3ooml0W4P0000000000000M;oo
00T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02<0oooo00<000000?ooo`0007l09`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo
0280oooo001?0?ooo`030?noM0000000M;oo00T0oooo00<000000?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02@0oooo0P00000W0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0
oooo00<000000?ooo`3oool08P3oool005/0oooo0`00000]0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l000000?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00/0oooo
0P00000T0?ooo`00F`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool0100007l0oooo00000000000T0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00D0001o0?ooo`3oool0oooo0000000S0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool00P3o
ool200000240oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool0
0`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`040?ooo`030000003oool0oooo01h0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0R0?ooo`00F`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`050?ooo`8000007P3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`<00000;@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool01`3oool00`000000
oooo0?ooo`0K0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo00/0oooo0P00000T0?ooo`00F`3oool00`000000oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`080?ooo`8000006`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo
001K0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo00X0oooo00<000000?ooo`3oool0603oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo02d0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00/0oooo0P00000H
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool03@3oool2000001H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?oo
o`00F`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0?0?ooo`800000503oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo0280oooo001K0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0140oooo0P00000B0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo0`00000]0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0C0?ooo`800000403o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0;0?ooo`800000903oool005/0oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool05@3oool2
000000h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00@03oool01P2o
M000000000000000000004RL0=ooo`80oooo00D0okmd0000001d_ol0oooo0;md00060000000304RL
g`3oool0oooo00D0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool05`3oool2000000`0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0R0?ooo`00@@3oool00`3ogi`0B00007Boo`040?ooo`030?noM000
0000M;oo0080oooo00<0gia80000000000000`0000000`00B9`0gooo0?ooo`050?ooo`030000003o
ool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo01T0oooo0P00000:0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3o
ool00440oooo00<0omnL04P0001d_ol02P3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0
oooo00D0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool06`3oool3000000L0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0R0?ooo`00@@3oool00`3ogi`0B00007Boo`0;0?ooo`030=nLB01d_ol0oooo
00T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool07P3oool2000000D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`00
0000oooo0?ooo`0R0?ooo`00@@3oool00`3ogi`0B00007Boo`0<0?ooo`030;mdB02Lgol0oooo00P0
oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool0803oool300000080oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000
oooo0?ooo`0R0?ooo`00@@3oool00`3ogi`0B00007Boo`0<0?ooo`030?oog`2LB7@0_ooo00P0oooo
0`00002S0000O`<00000Q@0007l3000002@0oooo00110?ooo`030?oOW0180000M;oo00d0oooo00<0
okmd001d_`3oool01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`040000O`0000000000000002@0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool08P3oool00440oooo00<0omnL04P0001d_ol03@3oool00`3ookl0M01809cO
o`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00@0001o0?ooo`3oool0oooo0`00000Q0?ooo`030000
O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo0280oooo00110?ooo`030?oOW0180000M;oo00d0oooo00<0oono07@00018W=l01`3o
ool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00@0oooo0`00000N0?ooo`030000O`3o
ool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo0280oooo00110?ooo`030?oOW0180000M;oo00P0oooo00<0oknL09cOo`3oool00P3oool0
0`3ogi`0B00004RLg`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool01`3oool3
000001/0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo00`0oooo00<000000?ooo`3oool08P3oool003l0oooo00D0oooO09a800000000000007Boo`08
0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool70?ooo`030000003oool0oooo02d0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool02P3oool4000001L0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool00440oooo
00<0okmd0000001d_ol02@3oool01P3ooml0W4P0000000000000001d0;ooo`P0oooo00<000000?oo
o`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0>0?ooo`<00000503oool00`0007l0oooo0?ooo`0V0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?oo
o`00F`3oool3000002d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool04@3oool300000140oooo00<0001o0?ooo`3o
ool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00/0oooo0P00000T0?ooo`00
F`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01@0oooo1000000=0?ooo`030000
O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo0280oooo001K0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0603o
ool4000000T0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0L0?ooo`D00000103oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`00
0000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0240oooo1P00000W0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00F`3oool00`000000
oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`0000001@00000Q0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0R0?ooo`00
F`3oool3000002d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`050?ooo`H00000
6`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00/0oooo0P00000T0?ooo`00F`3o
ool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02`3o
ool900000180oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo0280oooo001K0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0D0?ooo``000001P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0;@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0200oooo2@00000U0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool07@0000080?ooo`03
0000O`3oool0oooo00`0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0
7@3ooolA000000H0oooo00<000000?ooo`3oool08P3oool005/0oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo0280oooo001K0?ooo`<000001@3oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool09@3oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00/0oooo0P00000T0?ooo`00F`3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool08P3oool005/0
ooooo`00000b000002@0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo002:0?ooo`030?oo_`2o_kl0M0000080000000D0M;oo0?ooo`3oool0
oooo0;md0002000000050000B02Lgol0oooo0?oog`2LB0000P0000003P00B9`0W7@000000000M;l0
oooo0?oOW0180000000004RLg`3oool0omnL04P000000000B9cO0P3oool01`2oM000000000000000
07@0_ooo0?ooo`2oM0000P0000000`0004P0W=oo0?ooo`020?ooo`050;md0000000000000000M02o
ool00P3oool0203ooml0W4P0000000000000001804Q80000B9`0gooo103oool02@3ooml0W4P00000
0000000004RL09ad0000000007Bo0;md0002000000030000M02ooml0W4P00080000000@004RL0=oo
o`3ooml0W4P00P000000100004P0B4P00018W03Oool40?ooo`050;md0000000000000000B02Lgol0
0`3oool01@2oM00000000000000007@0_ooo00<0oooo00D0_g@0000000000000001809cOo`030?oo
o`050=nLB000000000000000001d_ol00P3oool01@3OW4P000000000000007@0W9a80080000000<0
001d0;ooo`3oool0103oool01P3OW4P000000000000004P0W=oO09a80080000000<0001d0;oog`2L
B0000P0000001000M;l0oooo0?ooo`3OW4P30000000407Boo`3oool0oooo0=nLB080000000<00018
09cOg`2LB0000P0000000`0007@0_ooO09a8000200000004001d_`3oool0oooo0;md0080000000@0
001d0;ooo`3oool0_g@00P0000000`0007@0_ooO09a80002000000080018W03Oool0omnL04P00000
0000B9cO0?oog`2LB002000000040018W02LM0000000001d_e@0oooo002:0?ooo`0]0?oo_`1d04P0
W=oo0?ooo`3ookl0M01809cOo`3oool0_g@000000000B9`0gooo0?oog`2LB7@0_ooo0?noM0000000
M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB00007@0_ooo0=no_`3Oool0gia80000M02oool0
gkno0=ooo`2oM0000000001d_`3oool0omnL04P0M02oool0oooo0=nLB00007@0_ooo00<0oooo0140
_g@000000000M;l0oooo0?oOW01807@0_ooo0?oog`2LB000000007Boo`3oool0oono07@000000000
001d0;ooo`030?ooo`0K0?noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB000B9`0
gooo0?ooo`3ooml0W4P0001d_`3oool0oooO09a800000000M;oo0?ooo`3ookl0M0000000000007@0
_ooo00@0oooo00<0gia80000M02oool00`3oool05`2oM0000000001d_`3oool0omnL04P0M02oool0
oooo0;md0000000004RL0=ooo`3ooml0W4Qd0;ooo`3oool0gia800000018W=l0oooo0?oog`2LB000
B9cO0080oooo00L0gia80000M02oool0oooo0=nLB00007@0_ooo00L0oooo00<0gia80000M02oool0
0P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool01`3OW4P0000004RLg`3oool0
oooO09a80018W=l00P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0
W4P0001d_`020?ooo`0?0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80018W03Oool0
oooo0?oog`2LB00007Bo0080oooo00d0gia80000M02oool0gkno0=noM0000000M;oo0?ooo`3ogi`0
B000001d_`3o_kl0_ooo05<0oooo002:0?ooo`0:0?oo_`2o_ol0oooo0?ooo`3o_g@0000007Boo`3o
gi`0B000001d_`<0oooo00T0omnL09boo`3o_g@0001809cOo`3oool0oooO09a80000M;l00P3oool0
0`3OW4P004RL0=ooo`020?ooo`070=nLB000B9`0gooo0?ooo`3ogi`0B00004RLg`030?ooo`060?oO
W01dW=l0oooo0=nLB00007@0_ooo0P3oool00`3ogi`0B00004RLg`030?ooo`050?oOW01dW=l0omnL
04P00018W=l00P3oool00`3ooml0W4P0001d_`050?ooo`070?noM00004P0W=oo0?ooo`3ooml0W4P0
001d_`020?ooo`0;0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l00P3o
ool00`3ooml0W4P0001d_`060?ooo`030=nLB00007@0_ooo0080oooo00<0omnL04P00018W=l00`3o
ool01@3ogi`0M9cO0?oOW018000007Bo00<0oooo00D0omnL09boo`3ogi`0B000001d_`030?ooo`0;
0?noM00007@0_ooo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02oool01`3oool00`3OW4P0001d
0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00000M;l00`3o
ool01`3o_g@0001d0;ooo`3oool0gia80000M02oool00P3oool02`2oM00004RL0=ooo`3oool0oooO
09a80000M;l0oooo0?oOW0180000B9cO00<0oooo00X0omnL07BLg`3oool0gia80018W03Oool0oooo
0?oog`2LB00007Bo0P3oool02`3OW4P004RL0=ooo`3oool0okmd0000B02Lgol0oooo0?oog`2LB000
07Bo05D0oooo002<0?ooo`080?oo_`1d000000000000M02oool0okmd0000B02Lgol60?ooo`060?no
M000000007Bo0?oog`2LB00007Bo0P3oool00`3OW4P004RL0=ooo`020?ooo`070=nLB000B9`0gooo
0?ooo`3o_g@0001809cOo`060?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02Lgol01@3o
ool00`3o_g@0000007Boo`020?ooo`030?oog`2LB00007Bo00H0oooo00H0okmd00000000M;l0oooO
09a80000M;l20?ooo`0;0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0okmd0000001d_ol0
0P3oool00`3ooml0W4P0001d_`060?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02Lgol0
1@3oool00`3o_g@0001809cOo`050?ooo`030?noM0000000M;oo00<0oooo00/0okmd0000001d_ol0
oooo0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`070?ooo`030=nLB00007@0_ooo0080oooo00/0
_g@00018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0000007Boo`030?ooo`070?noM0000000
M;oo0?ooo`3OW4P0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0
okmd0000B02Lgol01P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool00`3OW4P0
04RL0=ooo`020?ooo`060?noM000000007Bo0?oog`2LB00007BoE@3oool008/0oooo00@0_g@00000
00000000B9cO0P3oool00`3o_g@0001d0;ooo`080?ooo`040?noM00000000000001d_`80oooo00<0
gia80018W03Oool00P3oool01`3OW4P004RL0=ooo`3oool0okmd0000M02oool01P3oool00`3OW4P0
001d0;ooo`020?ooo`030?noM00007@0_ooo00D0oooo00<0okmd0000001d_ol00P3oool00`3ooml0
W4P0001d_`080?ooo`040?noM00000000000001d_`80oooo00/0gia80018W03Oool0oooo0?oog`2L
B00007Bo0?ooo`3o_g@0000007Boo`020?ooo`030?oog`2LB00007Bo00H0oooo00<0gia80000M02o
ool00P3oool00`3o_g@0001d0;ooo`050?ooo`030?noM00007@0_ooo00D0oooo00<0okmd0000001d
_ol00`3oool02`3o_g@0000007Boo`3oool0gia80000M02oool0oooo0=nLB00007@0_ooo00L0oooo
00<0gia80000M02oool00P3oool02`2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM000
0000M;oo00<0oooo00L0okmd0000001d_ol0oooo0=nLB00007@0_ooo0080oooo00/0_g@00018W03O
ool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0001d0;ooo`060?ooo`070=nLB000B9`0gooo0?ooo`3o
oml0W4P0001d_`020?ooo`030=nLB000B9`0gooo00@0oooo00@0okmd00000000000007BoE@3oool0
08X0oooo00<0omnL04P00018W=l00P3oool01@2oW9`0gooo0?oOW01804P0W=oo0080oooo00X0okmd
0000B02Lgol0oono07@0001d_ol0oooo0?oog`2LB00007Bo0P3oool00`3OW4P004RL0=ooo`020?oo
o`060=nLB000B9`0gooo0?ooo`3ogi`0B0001@0000001@18W=l0oooo0=nLB00004P0W=oo0080oooo
00<0omnL04P000000000100000001018W=l0oono07@00018W=l20?ooo`030?oog`2LB00007Bo00D0
oooo00L0oono07@0001d_ol0oooo0?oog`2LB00007Bo0080oooo00/0_g@00000M02oool0oooo0?oo
g`2LB00007Bo0?ooo`3ookl0M00004RLg`020?ooo`030?oog`2LB00007Bo00H0oooo00<0gia80000
B02Lgol00P3oool00`3ogi`0B000000000040000000404RLg`3ogi`0B01809cOo`80oooo00H0okmd
0000B02Lgol0oono07@0001d_ol30?ooo`0;0;md000004P0W=oo0?ooo`3OW4P0001d0;ooo`3oool0
gia80000M02oool01`3oool00`3OW4P0001809cOo`020?ooo`0;0;md000007@0_ooo0?ooo`3ooml0
W4P0001d_`3oool0oono07@0001d_ol00`3oool01`2oM000001809cOo`3oool0gia80000B02Lgol0
0P3oool02P2oM000001d0;ooo`3oool0oooO09a80000M;l0oooo0?oOW01800050000000904RLg`3o
ool0_g@00000M02oool0oooo0?oog`2LB00007Bo0080oooo00/0gia80018W03Oool0oooo0?oo_`1d
0000M;oo0?ooo`3ooml0W4P0001d_`1E0?ooo`00RP3oool05P3ookl0M01d0;ooo`3oool0okmd0018
W03Oool0oooo0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3ookl0M00004RLg`3oool0oono07@0
0000M;l20?ooo`030=nLB000B9`0gooo0080oooo00<0gia80018W03Oool00P3oool0402oM000B9cO
0?ooo`3oool0gia80000B02Lgml0W4P00000000004P0B7BL0=noM00007@0_ooo0;md0018W=l20?oo
o`0;0=nLB00004P0W=oo0?ooo`3o_g@0001809cOo`3oool0okmd00000000M;l01@3oool0403ookl0
M00004RLg`3oool0oono07@00000M;l0oooO09a80000000004Q804P0B02Lgol0oono07@00018W=l2
0?ooo`070?noM00004P0W=oo0?ooo`3o_g@00000001d_`040?ooo`0:0?oog`2LB00000000000B018
M9`0gkmd0000M02oool0_g@004RLg`80oooo01`0gia80000B02Lgol0oooo0;md0000B9`0gooo0?oo
g`2LB000001d0;ooo`3oool0gia80018W03Oool0oooo0?noM000000007Bo0?oog`2LB00000000000
M02oool0oooo0=nLB00007@0_ooo1@3oool03P3ooml0W4P00000000004P0B4Q809cOo`3ookl0M000
001dM01d07@0_ooo0?oo_`1d0000B9cO0P3oool05@3OW4P004RL0=ooo`3oool0okmd00000000M;l0
oooO09a800000000001804Q8B02Lgol0oono07@00000M7@0M01d0;ooo`3ookl0M00004RLg`020?oo
o`0?0;md0018W=l0oooo0?ooo`3OW4P0001809cOg`2LB00000000018B01804P0W=oo0?oo_`1d0000
B9cO0080oooo00/0gia80018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`1E0?oo
o`00R`3oool01P3o_g@0000000000000M;l0_ibL0=ooo`80oooo00D0gia8000000000000001809cO
o`020?ooo`080?oo_`1d000000000000000007@0_ooo0?ooo`2oM003000000030018W02LM0000000
0080000000@004RL0=ooo`3oool0gia80P000000300004P0W=oo0?ooo`3oool0_g@00000M02ogi`0
B0000000M02oool0oooo0=nLB080000000<0001809cOo`3oool00P3oool00`3o_g@0000000000002
00000003001d_`3oool0oooo00@0oooo00H0oono07@00000000000000000M02oool20?ooo`070;md
0000B9`0gooO09a800000000001d0;ooo`030?ooo`030?noM000000000000080000000<007Bo0?oo
o`3oool0103oool0202oM000001d0;oOW0180000001d0;ooo`3oool0gia80P0000000`0004P0W=oo
0?ooo`020?ooo`050=nLB000000000000000B02Lgol00`3oool01@3o_g@00000000000000000B9cO
00<0oooo00L0_g@00000M02oool0oooo0=nLB00007@0_ooo00L0oooo00`0_g@00000M02ooml0W4P0
0000000004P0W=oo0?oo_`1d000000000000M02oool30?ooo`050?noM0000000000000000018W=l0
0`3oool0302oM000001d0;oog`2LB00000000000B02Lgol0oono07@000000000001d0;ooo`<0oooo
00D0gia8000000000000001809cOo`020?ooo`090;md0000B9`0gooO09a800000000001d0;ooo`3o
ool0_g@000<0000000@004RL0=ooo`3ookl0M0000P0000000`0007@0_ooo0?ooo`1D0?ooo`00XP3o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000B9`0gooo01`0oooo00<0oooO09a80000M;l06`3o
ool00`3ooml0W4P0001d_`0Y0?ooo`030=nLB00007@0_ooo03d0oooo00<0_g@00018W03Oool0G@3o
ool00:80oooo00<0oono07A8W03Oool00P3oool00`3ookl0M4RL0=ooo`0L0?ooo`030?oog`2LB000
07Bo01/0oooo00<0oooO09a80000M;l0:@3oool00`3OW4P0001d0;ooo`0m0?ooo`030?oo_`1dB9`0
gooo05d0oooo002R0?ooo`030?oog`2LM9`0gooo0080oooo00<0oooO09adW03Oool0703oool00`3o
oml0W4P0001d_`0K0?ooo`030?oog`2LB00007Bo02T0oooo00<0gia80000M02oool0?@3oool00`3o
oml0W7BL0=ooo`1M0?ooo`00a@3oool0103ookl0M00000000000M;lJ0?ooo`040?oo_`1d00000000
001d_bD0oooo00L0gia80000B02Lgml0W4P00000000007@0_ooo09d0oooo00360?ooo`030?oog`2L
B00007Bo01/0oooo00<0oooO09a80000M;l09@3oool01`3OW4P0001809cOo`3oool0_g@00000M02o
ool0W@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001
\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-61.695, -0.412703, \
0.933529, 0.0111888}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotRecoilAngle[sol, Join[values, {mratio \[Rule] 2.0}], 
        180. ];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00529101 0.0147151 0.00654005 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.15608 -0.0125 -8 -15.6875 ]
[.15608 -0.0125 8 0 ]
[.28836 -0.0125 -8 -15.6875 ]
[.28836 -0.0125 8 0 ]
[.42063 -0.0125 -8 -15.6875 ]
[.42063 -0.0125 8 0 ]
[.55291 -0.0125 -11 -15.6875 ]
[.55291 -0.0125 11 0 ]
[.68519 -0.0125 -11 -15.6875 ]
[.68519 -0.0125 11 0 ]
[.81746 -0.0125 -11 -15.6875 ]
[.81746 -0.0125 11 0 ]
[.94974 -0.0125 -11 -15.6875 ]
[.94974 -0.0125 11 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.0313 -31.375 ]
[.5 -0.0125 21.0313 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .14552 -16 -7.84375 ]
[-0.0125 .14552 0 7.84375 ]
[-0.0125 .27632 -16 -7.84375 ]
[-0.0125 .27632 0 7.84375 ]
[-0.0125 .40712 -16 -7.84375 ]
[-0.0125 .40712 0 7.84375 ]
[-0.0125 .53792 -16 -7.84375 ]
[-0.0125 .53792 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -31.6875 -21.0313 ]
[-0.01875 .30902 -16 21.0313 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -29.875 0 ]
[.5 .63053 29.875 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.15608 0 m
.15608 .61803 L
s
.28836 0 m
.28836 .61803 L
s
.42063 0 m
.42063 .61803 L
s
.55291 0 m
.55291 .61803 L
s
.68519 0 m
.68519 .61803 L
s
.81746 0 m
.81746 .61803 L
s
.94974 0 m
.94974 .61803 L
s
0 .14552 m
1 .14552 L
s
0 .27632 m
1 .27632 L
s
0 .40712 m
1 .40712 L
s
0 .53792 m
1 .53792 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.15608 0 m
.15608 .00625 L
s
gsave
.15608 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(25) show
1.000 setlinewidth
grestore
.28836 0 m
.28836 .00625 L
s
gsave
.28836 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.42063 0 m
.42063 .00625 L
s
gsave
.42063 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(75) show
1.000 setlinewidth
grestore
.55291 0 m
.55291 .00625 L
s
gsave
.55291 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.68519 0 m
.68519 .00625 L
s
gsave
.68519 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(125) show
1.000 setlinewidth
grestore
.81746 0 m
.81746 .00625 L
s
gsave
.81746 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.94974 0 m
.94974 .00625 L
s
gsave
.94974 -0.0125 -72 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(175) show
1.000 setlinewidth
grestore
.125 Mabswid
.05026 0 m
.05026 .00375 L
s
.07672 0 m
.07672 .00375 L
s
.10317 0 m
.10317 .00375 L
s
.12963 0 m
.12963 .00375 L
s
.18254 0 m
.18254 .00375 L
s
.20899 0 m
.20899 .00375 L
s
.23545 0 m
.23545 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.31481 0 m
.31481 .00375 L
s
.34127 0 m
.34127 .00375 L
s
.36772 0 m
.36772 .00375 L
s
.39418 0 m
.39418 .00375 L
s
.44709 0 m
.44709 .00375 L
s
.47354 0 m
.47354 .00375 L
s
.5 0 m
.5 .00375 L
s
.52646 0 m
.52646 .00375 L
s
.57937 0 m
.57937 .00375 L
s
.60582 0 m
.60582 .00375 L
s
.63228 0 m
.63228 .00375 L
s
.65873 0 m
.65873 .00375 L
s
.71164 0 m
.71164 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.76455 0 m
.76455 .00375 L
s
.79101 0 m
.79101 .00375 L
s
.84392 0 m
.84392 .00375 L
s
.87037 0 m
.87037 .00375 L
s
.89683 0 m
.89683 .00375 L
s
.92328 0 m
.92328 .00375 L
s
.97619 0 m
.97619 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.0313 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .14552 m
.00625 .14552 L
s
gsave
-0.0125 .14552 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
0 .27632 m
.00625 .27632 L
s
gsave
-0.0125 .27632 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
0 .40712 m
.00625 .40712 L
s
gsave
-0.0125 .40712 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(60) show
1.000 setlinewidth
grestore
0 .53792 m
.00625 .53792 L
s
gsave
-0.0125 .53792 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(80) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04742 m
.00375 .04742 L
s
0 .08012 m
.00375 .08012 L
s
0 .11282 m
.00375 .11282 L
s
0 .17822 m
.00375 .17822 L
s
0 .21092 m
.00375 .21092 L
s
0 .24362 m
.00375 .24362 L
s
0 .30902 m
.00375 .30902 L
s
0 .34172 m
.00375 .34172 L
s
0 .37442 m
.00375 .37442 L
s
0 .43982 m
.00375 .43982 L
s
0 .47252 m
.00375 .47252 L
s
0 .50522 m
.00375 .50522 L
s
0 .57062 m
.00375 .57062 L
s
0 .60332 m
.00375 .60332 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -105.875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
82.0313 11.8438 translate
90 rotate
-82.0313 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
68.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
76.063 15.813 moveto
(@) show
80.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
97.125 15.813 moveto
(D) show
101.063 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.15608 .61178 m
.15608 .61803 L
s
.28836 .61178 m
.28836 .61803 L
s
.42063 .61178 m
.42063 .61803 L
s
.55291 .61178 m
.55291 .61803 L
s
.68519 .61178 m
.68519 .61803 L
s
.81746 .61178 m
.81746 .61803 L
s
.94974 .61178 m
.94974 .61803 L
s
.125 Mabswid
.05026 .61428 m
.05026 .61803 L
s
.07672 .61428 m
.07672 .61803 L
s
.10317 .61428 m
.10317 .61803 L
s
.12963 .61428 m
.12963 .61803 L
s
.18254 .61428 m
.18254 .61803 L
s
.20899 .61428 m
.20899 .61803 L
s
.23545 .61428 m
.23545 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.31481 .61428 m
.31481 .61803 L
s
.34127 .61428 m
.34127 .61803 L
s
.36772 .61428 m
.36772 .61803 L
s
.39418 .61428 m
.39418 .61803 L
s
.44709 .61428 m
.44709 .61803 L
s
.47354 .61428 m
.47354 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.52646 .61428 m
.52646 .61803 L
s
.57937 .61428 m
.57937 .61803 L
s
.60582 .61428 m
.60582 .61803 L
s
.63228 .61428 m
.63228 .61803 L
s
.65873 .61428 m
.65873 .61803 L
s
.71164 .61428 m
.71164 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.76455 .61428 m
.76455 .61803 L
s
.79101 .61428 m
.79101 .61803 L
s
.84392 .61428 m
.84392 .61803 L
s
.87037 .61428 m
.87037 .61803 L
s
.89683 .61428 m
.89683 .61803 L
s
.92328 .61428 m
.92328 .61803 L
s
.97619 .61428 m
.97619 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .14552 m
1 .14552 L
s
.99375 .27632 m
1 .27632 L
s
.99375 .40712 m
1 .40712 L
s
.99375 .53792 m
1 .53792 L
s
.125 Mabswid
.99625 .04742 m
1 .04742 L
s
.99625 .08012 m
1 .08012 L
s
.99625 .11282 m
1 .11282 L
s
.99625 .17822 m
1 .17822 L
s
.99625 .21092 m
1 .21092 L
s
.99625 .24362 m
1 .24362 L
s
.99625 .30902 m
1 .30902 L
s
.99625 .34172 m
1 .34172 L
s
.99625 .37442 m
1 .37442 L
s
.99625 .43982 m
1 .43982 L
s
.99625 .47252 m
1 .47252 L
s
.99625 .50522 m
1 .50522 L
s
.99625 .57062 m
1 .57062 L
s
.99625 .60332 m
1 .60332 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -90.875 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(recoil) show
93.063 15.813 moveto
(angle) show
118.750 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .60332 m
.06245 .56753 L
.10458 .52866 L
.14415 .4925 L
.18221 .45817 L
.22272 .42231 L
.26171 .38863 L
.30316 .35393 L
.34309 .32178 L
.3815 .29222 L
.42237 .26241 L
.46172 .23544 L
.49955 .21119 L
.53984 .18723 L
.57861 .16593 L
.61984 .14509 L
.65954 .12665 L
.69774 .11025 L
.73838 .09405 L
.77751 .07949 L
.81909 .06492 L
.85916 .05156 L
.89771 .03917 L
.93871 .02632 L
.97619 .01472 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00j@3oool01@3OW4P00000000000000000M;oo0140oooo00<0oooO09a8
000000000`0000001018W=l0oooo0?ooo`3OW4P30000000307Boo`3oool0oooo0:40oooo003Y0?oo
o`030=nLB00007@0_ooo01<0oooo00<0_g@00018W03Oool00`3oool00`3OW9`0_ooo0?ooo`020?oo
o`030?oo_`1d0000M;oo0:<0oooo003L0?ooo`040=nLB00000000000001d_`T0oooo00<0gia80018
W03Oool04`3oool00`3ooml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo0080oooo00<0oono07@0
001d_ol0X`3oool00=l0oooo00<0okmd001d_`3oool01`3oool00`3OW4P004RL0=ooo`0D0?ooo`03
0;md00000000000000<0000000L0B9cO0?ooo`3oool0oooo0?oo_`1d0000M;oo0:<0oooo003D0?oo
o`050?oo_`1d000000000000B02Lgol01P3oool00`3ookl0M01d0;ooo`070?ooo`030=nLB000B9`0
gooo00<0oooo00P0oooO09a80000000000000000B018B00004RL0=ooo`80oooo00P0_g@000000000
0000001d0;ooo`3oool0oono07@000@0000000<0001d0;ooo`3oool00P3oool00`3ookl0M00007Bo
o`2S0?ooo`00e03oool01P2oM00007Bo0?ooo`3ogi`0B01809cOo`D0oooo00<0gia80000M02oool0
1`3oool00`3OW4P004RL0=ooo`020?ooo`0D0?oog`2LB000000007Boo`3oool0oono07@000000000
001d0;ooo`2oM0000000001d_`3oool0omnL04P0M02oool0oooo0;mdB02Lgol80?ooo`030?oo_`1d
0000M;oo0:<0oooo003C0?ooo`030?oog`2LB000M;oo0080oooo00<0gia80018W03Oool00P3oool0
103ookl0M00000000018W=l80?ooo`030=nLB000B9`0gooo0080oooo00<0omnL04P00018W=l00P3o
ool01`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool01@3ogi`0M9cO0?ooo`3ookl0M000
0080000000<007Bo0?ooo`3oool00`3oool00`3ookl0M00007Boo`2S0?ooo`00d`3oool00`3ookl0
M00007Boo`020?ooo`030?oo_`1d0000M;oo00<0oooo00<0oooO09bL_`3oool0203oool00`3OW4P0
04RL0=ooo`020?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?noM00004P0
W=oo00H0oooo00H0gia80000M02oool0oono07@00018W=l40?ooo`030?oo_`1d0000M;oo0:<0oooo
003C0?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool00`3oool00`3o_g@0B9cO0?oo
o`070?ooo`030=nLB000B9`0gooo0080oooo00<0okmd0000001d_ol00P3oool01`3ooml0W4P0001d
_`3oool0okmd0000M02oool01@3oool01`3ookl0M00004RLg`3oool0oooO09a80000M;l0103oool0
0`3ookl0M00007Boo`2S0?ooo`00d`3oool00`3ooml0W4P0001d_`030?ooo`080=nLB00007@0_ooo
0?oOW01dW=l0oooo0=nLB000M;l80?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@00018W=l0
0P3oool01P3ooml0W4P0001d_`3oool0omnL04P000D0000000P0B9cO0?oo_`1d0000B9cO0?ooo`3o
oml0W4P004RLg`@0oooo00<0oono07@0001d_ol0X`3oool00=@0oooo00<0_g@00000000000001000
00001P1d_ol0oooo0=nLB0000000001809cOo`P0oooo00<0gia80018W03Oool00`3oool01`3o_g@0
001809cOo`3oool0okmd00000000M;l00P3oool0302oM000B9cO0?ooo`3oool0gia80000B02Lgol0
oooo0;md0000M;l0oooo0?noM080000000L004RL0=ooo`3oool0oooo0?oo_`1d0000M;oo0:<0oooo
003D0?ooo`030=nLB00007@0_ooo0080oooo00<0oono07@00018W=l03@3oool00`3OW4P004RL0=oo
o`040?ooo`030?noM000000000000080000000D007Bo0?ooo`3oool0oooo0=nLB002000000060000
B02Lgol0oooo0?ooo`3ooml0W4P00P0000000`1d_ol0oooo0?ooo`030?ooo`030?oo_`1d0000M;oo
0:<0oooo003D0?ooo`030?oOW0180000M;oo00<0oooo00<0_g@0001d_`3oool0303oool00`3OW4P0
04RL0=ooo`070?ooo`030?oog`2LB00007Bo01@0oooo00<0oono07@0001d_ol0X`3oool00=D0oooo
00<0_g@00018W03Oool00P3oool00`3OW4P007Bo0?ooo`0<0?ooo`030=nLB000B9`0gooo00L0oooo
00<0oooO09a80000M;l0503oool00`3ookl0M00007Boo`2S0?ooo`00e@3oool00`3ookl0M00007Bo
o`020?ooo`030=nLB000M;l0oooo00`0oooo00<0gia80018W03Oool01`3oool00`3ooml0W4P0001d
_`0D0?ooo`030?oo_`1d0000M;oo0:<0oooo003F0?ooo`060?noM00004P0W=oo0?ooo`2oM000M;oo
3@3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d00000000001d_a@0oooo00<0oono07@0001d
_ol0X`3oool00=L0oooo00D0omnL04P00000000004RL0=ooo`0=0?ooo`050=nLB000000000000000
001d_ol01@3oool00`3ooml0W4P0001d_`0B0?ooo`050=nLB000000000000000001d_ol0X`3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool005h0oooo00D0omnL04P000000000001d0;ooo`0M0?ooo`030;md000000000000
00@0000000@0B9cO0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`0G0?ooo`030?oog`2LB000
00000080000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B0000000000007@0_ooo01/0oooo00<0
gia804RLg`3oool00`3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo01D0oooo00H0
_g@000000000000000000018W03Oool30?ooo`050?oOW018000000000000M02oool0103oool01@3o
gi`0B0000000000007@0_ooo0140oooo00P0_g@000000000000000000018W03Oool0oooo0;md00H0
000000@0B9cO0?ooo`3ooml0W4P00`0000000`1d_ol0oooo0?ooo`0@0?ooo`090;md000000000000
00000000B9`0gooo0?ooo`3ooml0W4P000<0000000<0M;oo0?ooo`3oool00`3oool01@3ogi`0B000
0000000007@0_ooo0140oooo00H0_g@000000000000000000018W03Oool40?ooo`030=nLB018W=l0
oooo00<0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?ooo`0Z0?ooo`00G@3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool07@3oool00`3OW4P0000000000003000000050018W03O
ool0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0G0?ooo`030?oog`2LB7@0_ooo0080oooo
00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool06P3oool00`3o
gi`0B4RL0=ooo`030?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool05P3oool00`3o
gi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`0A0?ooo`030?oOW0180000M;oo00@0oooo00<0gia800000000
00000`0000001@00B9`0gooo0?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool04@3oool00`3o
gi`0B00007Boo`030?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool00P3oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool04@3oool00`3ogi`0B00007Boo`060?ooo`030?oOW018
B9`0gooo00<0oooo00<0oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`0Z0?ooo`00G@3oool0
0`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01d0oooo00<0gia804RLg`3oool00P3oool0
0`3OW7@0W=oo0?ooo`040?ooo`030?oog`2LB7@0_ooo01`0oooo00<0oooO09a8M02oool00P3oool0
0`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo01T0oooo00<0oooO09a8B02Lgol0203oool0
0`3ooml0W4Qd0;ooo`0F0?ooo`030?oOW0180000M;oo00@0oooo00<0_g@00018W03Oool00P3oool0
1`2oM00007Bo0?ooo`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`0@0?ooo`030?oO
W0180000M;oo00D0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`040?ooo`030?oo
g`2LB7@0_ooo0140oooo00<0omnL04P0001d_ol0203oool00`3ooml0W4Qd0;ooo`020?ooo`030;md
0000B9`0gooo0080oooo00<0_g@0001d_`3oool0403oool00`3ogi`0B00007Boo`060?ooo`030?oo
g`2LB4P0W=oo00P0oooo00<0oooO09a8M02oool0:P3oool005`0oooo00<0oono07@00018W=l00`3o
ool00`3OW4P0001d0;ooo`0N0?ooo`030=nLB01d_ol0oooo00T0oooo00<0_g@007Boo`3oool0703o
ool01P2oM000M;oo0?ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02oool06P3oool00`3OW4P0
B9cO0?ooo`080?ooo`030;md001d_ol0oooo01D0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0
M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`0@
0?ooo`030?oOW0180000M;oo00H0oooo00<0gia807Boo`3oool02@3oool00`2oM000M;oo0?ooo`0@
0?ooo`030?oOW0180000M;oo00T0oooo00H0_g@007Boo`3oool0oono07@00018W=l30?ooo`030=nL
B00007@0_ooo0100oooo00<0omnL04P0001d_ol01`3oool00`3OW4P0B9cO0?ooo`080?ooo`030;md
001d_ol0oooo02T0oooo001L0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol07`3o
ool00`2oM4P0W=oo0?ooo`070?ooo`030?oog`2LB000B9cO01`0oooo00L0oooO09a80018W=l0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol06P3oool00`3ogi`0B4RL0=ooo`070?ooo`03
0?oog`2LB000B9cO01H0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`06
0?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`0@0?ooo`030?oOW0180000
M;oo00L0oooo00<0_gA809cOo`3oool01`3oool00`3ooml0W4P004RLg`0A0?ooo`030?oOW0180000
M;oo00P0oooo00L0oooO09a80018W=l0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
403oool00`3ogi`0B00007Boo`070?ooo`030?oOW018B9`0gooo00L0oooo00<0oooO09a80018W=l0
:P3oool005`0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0O0?ooo`030?oog`2L
B7@0_ooo00L0oooo00<0gia80000001d_ol0703oool01`3OW4P0000007Boo`3oool0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`0J0?ooo`030?oog`2LB4P0W=oo00L0oooo00<0gia80000001d
_ol05P3oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02L
gol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo0100oooo00<0omnL04P0001d_ol01`3oool0
0`3ooml0W4Qd0;ooo`070?ooo`030=nLB0000000M;oo0140oooo00<0omnL04P0001d_ol0203oool0
1`3OW4P0000007Boo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0@0?ooo`030?oO
W0180000M;oo00L0oooo00<0oooO09a8B02Lgol01`3oool00`3OW4P0000007Boo`0Z0?ooo`00G03o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo0200oooo00<0okmd001d_`3oool00P3o
ool01`3Ogml0oooo0?oo_`1d000000000000M02oool0603oool02`3Ogml0oooo0?oo_`1d00000000
0000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol06`3oool00`3OW4P007Bo
0?ooo`020?ooo`070=oOg`3oool0oono07@000000000001d0;ooo`0F0?ooo`030?oOW0180000M;oo
00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0
okmd0000B02Lgol0403oool00`3ogi`0B00007Boo`080?ooo`030?noM000M;l0oooo0080oooo00L0
gmoO0?ooo`3ookl0M0000000000007@0_ooo0140oooo00<0omnL04P0001d_ol0103oool02`3Ogml0
oooo0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0
403oool00`3ogi`0B00007Boo`080?ooo`030=nLB000M;l0oooo0080oooo00L0gmoO0?ooo`3ookl0
M0000000000007@0_ooo02X0oooo001L0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol0803oool00`3ookl0M01809cOo`020?ooo`060=nLB000000000000000000004P0W=oo6@3oool0
1P3OW4P00000000000000000001809cOo`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`0K0?ooo`030?oo_`1d07@0_ooo0080oooo00H0gia800000000000000000000B02LgolG0?oo
o`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d_ol00`3oool01P3o_g@0001809cOo`3ogi`0
B00007Boo`<0oooo00<0okmd0000B02Lgol0403oool00`3ogi`0B00007Boo`080?ooo`030?oo_`1d
04P0W=oo0080oooo00H0gia800000000000000000000B02LgolB0?ooo`030?oOW0180000M;oo00@0
oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol0403oool00`3ogi`0B00007Boo`080?ooo`030?oo_`1d07@0_ooo0080oooo00H0gia8
00000000000000000000B02Lgol[0?ooo`00G03oool00`3ookl0M00004RLg`030?ooo`030=nLB000
07@0_ooo0200oooo00<0oono07@00018W=l00P3oool0103ookl0M00000000018W=lK0?ooo`040?oo
_`1d0000000004RLg`@0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0F0?ooo`03
0?oOg`3Oool0oooo00<0oooo00P0_g@007Boo`3oool0oooo0?oo_`1d0000000004RLgaT0oooo00<0
omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000
B9cO0`3oool00`3OW4P0001d0;ooo`0@0?ooo`030?oOW0180000M;oo00P0oooo00<0oono07@00018
W=l00P3oool0103ookl0M00000000018W=lD0?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@0
00000000B9cO103oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo0100oooo00<0omnL
04P0001d_ol00`3oool00`3ogml0gooo0?ooo`030?ooo`080;md001d_ol0oooo0?ooo`3ookl0M000
00000018W=l]0?ooo`00G@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo01/0oooo
00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`030?ooo`030;md001d_ol0oooo01`0oooo
00<0_g@007Boo`3oool01@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo01H0oooo
00<0oono07A8W03Oool00`3oool00`3o_g@007Bo0?ooo`020?ooo`030;md001d_ol0oooo01T0oooo
00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`3O
W4P004RL0=ooo`020?ooo`030;md0000M;l0oooo0100oooo00<0omnL04P0001d_ol00`3oool00`3o
_i`0W=oo0?ooo`020?ooo`030?oOW0180000B9cO00<0oooo00<0_g@007Boo`3oool0503oool00`3o
gi`0B00007Boo`050?ooo`030;md001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`0@0?ooo`030?oOW0180000M;oo00<0oooo00<0oono07A8W03Oool00`3oool00`3o
_g@007Bo0?ooo`020?ooo`030;md001d_ol0oooo02d0oooo001M0?ooo`070?oo_`1d0000M;oo0?oo
o`3o_g@0001d0;ooo`0L0?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool30?ooo`06
0?oOW018000000000000000004P0W=oo6@3oool01P3ogi`0B000000000000000001809cOo`80oooo
00L0oono07@0001d_ol0oooo0?noM00007@0_ooo01P0oooo00<0_g@00000000000000`0000001P00
07@0_ooo0?ooo`3oool0omnL04P00080000000<0001809cOo`3oool04`3oool01@3ooml0W4P00000
00000000M;oo00@0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0oono07@0
001d_ol0oooo0?noM00007@0_ooo00l0oooo00D0oooO09a800000000000007Boo`030?ooo`080?oo
g`2LB4P0W=oo0?ooo`3ookl0M0000000M02oool30?ooo`060?oOW018000000000000000004P0W=oo
3`3oool01@3ooml0W4P0000000000000M;oo00D0oooo00H0omnL04P00000000000000000B02Lgol2
0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0?0?ooo`050?oog`2LB00000000000001d
_ol0103oool00`2oM000000000000003000000060000M02oool0oooo0?ooo`3ogi`0B0000P000000
0`0004P0W=oo0?ooo`0Y0?ooo`00GP3oool01@3ookl0M0000000000007@0_ooo01h0oooo00H0oooO
09a80000000000000000M02oool50?ooo`050;md00000000000000000018W=l06P3oool01@2oM000
0000000000000000B9cO00<0oooo00D0oono07@000000000001d0;ooo`0I0?ooo`030?noM0000000
000000@0000000D0M;oo0?ooo`3oool0oooo0;md00030000000304RLg`3oool0oooo01@0oooo00<0
okmd0000001d_ol01@3oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000
001d0;ooo`0B0?ooo`030?noM0000000M;oo00@0oooo00H0oooO09a80000000000000000M02oool5
0?ooo`050;md00000000000000000018W=l04@3oool00`3o_g@0000007Boo`060?ooo`050;md0000
0000000000000018W=l00`3oool01@3ookl0M0000000000007@0_ooo0180oooo00<0okmd0000001d
_ol0103oool00`3o_g@00000000000040000000507Boo`3oool0oooo0?ooo`2oM0000`0000000`18
W=l0oooo0?ooo`0X0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00B`3oool0
1@3ogi`0B0000000000007@0_ooo00T0ooooo`00000b000002H0oooo001:0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0T0?ooo`00BP3oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo02@0oooo00190?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool0
0`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool0
04T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo
02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00B@3oool00`3ogi`0
B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo0`00000]0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00@0
oooo0P0000040?ooo`<000009P3oool004T0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0100000080?ooo`030000
003oool0oooo02@0oooo00190?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3o
ool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`@00000303oool00`000000oooo0?ooo`0T0?ooo`00B@3oool0
0`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool08@3oool4
000000030?ooo`0007l0oooo00d0oooo00<000000?ooo`3oool0903oool004X0oooo00<0gia80018
W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01h0oooo0`0000050?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool004X0oooo00L0oono07@0001d
_ol0oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool06`3oool3000000P0oooo00<0001o
0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00B`3oool01@3ookl0M0000000000007@0
_ooo00T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool0603oool3000000/0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0E0?ooo`<000003P3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0180
oooo0`00000A0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo
00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool03`3oool3000001@0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0T0?ooo`00F@3oool2000002h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo00`0oooo0`00000G0?ooo`030000O`3oool0oooo00/0oooo0P00
000V0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`090?ooo`<000006P3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00H0oooo0`00000M0?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool00`3oool300000200oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00
F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`03000002<0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool3000002H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?oo
o`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool08`3oool2000000030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool0803oool300000080oooo00<0001o0?ooo`3oool09P3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0M0?ooo`<000001@3oool00`0007l0
oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo
00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01/0oooo
0P0000080?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0T0?ooo`00F@3oool2000002h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0H0?ooo`<000002P3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00/0oooo0P00
000V0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool05@3oool3000000d0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0B0?ooo`<00000403oool00`0007l0oooo0?ooo`0V0?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?ooo`3o
ool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00l0oooo0`00000C0?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00
F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
303oool3000001H0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0:0?ooo`8000006@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00P0oooo0P00000K0?ooo`030000O`3oool0
oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`00
0000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool01@3oool30000
01d0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo
02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`030?ooo`800000803oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0
oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo00<000008P3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo0P00000^0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0000000000009@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo00/0oooo0P00000V0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02@0oooo0P00000W0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
0240oooo0`0000000`3oool0001o0?ooo`0V0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?oo
o`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo01l0oooo0P0000040?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000
003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool07@3oool2000000H0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0
903oool00400oooo00<0_g@0000000000000100000002@18W=l0oooo0?ooo`3oool0omnL04P00000
0000001d0;ooo`090?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool06`3oool20000
00P0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool0903oool00440oooo00<0gia80000000000000`0000002P00B9`0
gooo0?ooo`3ookl0M00007Boo`3oool0okmd0000M02oool80?ooo`030000003oool0oooo02d0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool06@3oool2000000X0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool00480oooo00<0
gia804RLg`3oool00P3oool01P3OW7@0W=oo0?ooo`2oM00004RL0=ooo`80oooo00<0_g@0001d_`3o
ool01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01L0oooo0P00000<0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo02@0oooo00130?ooo`030=nLB01d_ol0oooo00<0oooo00<0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
5@3oool2000000h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool004@0oooo00<0_gA809cOo`3oool00P3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0C0?ooo`800000403oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00A03oool0
0`3ooml0W4Qd0;ooo`020?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool3
00000;T0001o0P00001`0000O`<000009P3oool004D0oooo00L0okmd001d_`3oool0oooo0?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo00l0oooo0P00000D0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo00150?ooo`070?oo_`1d04P0
W=oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0=0?ooo`8000005P3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00A@3o
ool01`3ookl0M00004RLg`3oool0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`070?ooo`03
0000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool02`3oool2000001P0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3o
ool0903oool00400oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`020?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo00T0oooo0P00000J0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo00100?ooo`080?oog`2L
B4P0W=oo0?ooo`3ookl0M0000000M02oool20?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`080?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool01`3oool2000001`0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<0
00000?ooo`3oool0903oool00440oooo00H0oooO09a80000000000000000M02oool40?ooo`050?oo
_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00D0
oooo0P00000N0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool00`3oool200000200oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<0
00000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool0100007l0oooo0?ooo`3oool200000280oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0
903oool005T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0100007l0oooo00000000000T0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo02@0oooo001I0?ooo`800000;P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool2000002H0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00/0oooo0P00000V0?ooo`00F@3oool0
0`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0S0?ooo`80000000<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T
0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0Q0?ooo`8000000P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?oo
o`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01l0oooo0P0000040?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`03
0000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?ooo`3oool0
;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
7P3oool00`000000oooo0?ooo`040?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0
903oool005T0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool0703oool2000000L0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303o
ool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0K0?ooo`030000003oool0
oooo00L0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`00
0000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0I0?ooo`8000002P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo
02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool02P3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01H0oooo
0P00000=0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo0P00
000^0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0E0?ooo`030000003oool0oooo00d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool02`3oool2000002H0oooo001I
0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo01<0oooo0P00000@0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<00000
0?ooo`3oool0903oool002h0oooo00<0omnL04P0000000000`0000000`00M;l0oooo0?ooo`0R0?oo
o`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo0140oooo0P00000B0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?oo
o`3oool0903oool002`0oooo00<0gia80000000000001P0000000`00B9`0gooo0?ooo`0Q0?ooo`03
0000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo0100oooo00<000000?ooo`3oool04P3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo02@0oooo000Z0?ooo`060?oo_`1d000000000000M02ookl0W9cO1@3oool00`2o
M7@0_ooo0?ooo`0Q0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00h0oooo0P00000E0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool0903oool002T0oooo00<0omnL04P0B02Lgol00P3oool00`3ookl0
W9cO0?ooo`040?ooo`030;nLW03Oool0oooo0240oooo00<000000?ooo`3oool0;@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0303oool2000001L0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0
oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00:@3oool00`2oW9`0gooo
0?ooo`020?ooo`030?oo_`2LW=l0oooo0080oooo00<0oono07@0001d_ol08`3oool00`000000oooo
0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0:0?ooo`8000006@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo
000X0?ooo`030?oo_`1dM;l0oooo00<0oooo00<0oono07@0000000000`0000000`18W=l0oooo0?oo
o`0R0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0<0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oog`2LB000000000L0000000<004RL
0=ooo`3oool0903oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`070?ooo`800000703oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo02@0oooo000Y0?ooo`030?oo_`1d0000000000@0000000<0M;oo0?oo
o`3oool09P3oool2000002h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool0703oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0;0?oo
o`8000009P3oool003@0oooo00@0omnL04P00000B9`0gooo8@3oool00`000000oooo0?ooo`0]0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`040?oo
o`8000007`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo000c0?ooo`05
0=nLB000B9`0gmnL04Q8W03Oool08@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`030?ooo`030000003oool0oooo
01l0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00<P3oool01P2oM000
B9cO0?ooo`3ogi`0B4RL0=ooob40oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00@0001o0?ooo`3oool0oooo0P00000R0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool0903oool00340oooo00L0gia807Boo`3oool0oooo0?oO
W018B9`0gooo0240oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00@0001o0?ooo`3oool00000903oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo02@0oooo000`0?ooo`030=nLB0000000000000D0000000<0001d0;ooo`3oool0
2@3oool0103ooml0W4P00018W03Oool30?ooo`050?oOW018000000000000M02oool02@3oool00`00
0000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l000000000000U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool003D0
oooo00<0omnL04Q8W03Oool0303oool0103ooml0W4P00018W03Oool20?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000003oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303o
ool00`000000oooo0?ooo`0T0?ooo`00A03oool0103ooml0W4P00018W03Oool20?ooo`030;md0000
B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool0903oool00`000000001o0?ooo`0V0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool0903oool00400oooo00<0_g@00000000000001@000000
101d_ol0oono07@00018W=l30?ooo`030=nLB00007@0_ooo00L0oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0280oooo0P0000000`3oool0001o0?oo
o`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool00400oooo00`0omnL
07BLg`3oool0oooo0?oog`2LB00004RL0=ooo`3oool0omnL04P0001d_ol30?ooo`030?noM00004P0
W=oo00L0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo0240oooo00D000000?ooo`3oool0oooo0000O`0W0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<0
00000?ooo`3oool0903oool00440oooo00/0gia807Boo`3oool0oooO09a80000B9`0gooo0?ooo`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo0`00001h0000O`030000000007l0001o
0;00001o0`00000V0?ooo`00@@3oool02`3ooml0W4Qd0;ooo`3ooml0W4P00018W03Oool0oooo0?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo0?ooo`0]0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool07P3oool2000000D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo00120?ooo`0:0?oOW018
M;l0oooO09a80000B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01`3oool0
0`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool07@3oool0
0`000000oooo0?ooo`050?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0T0?ooo`00:03oool00`3ogi`0B0000000000>000000030000M02oool0oooo
00L0oooo00T0gia807Bog`2LB00004RL0=ooo`3oool0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0L0?ooo`030000003oool0oooo00H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oOW01dW=l0oooo00h0oooo00<0
_gAd0;ooo`3oool01`3oool01@3ooml0W4Q804Q80000B9`0gooo0080oooo00<0gia80018W03Oool0
0P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0J0?ooo`8000002@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool002P0oooo00<0omnL09boo`3oool03P3o
ool00`3OW9`0_ooo0?ooo`080?ooo`040?oOW018000004RL0=ooo`80oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo00P0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo01T0oooo00<000000?ooo`3oool02@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?oo
o`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool002P0oooo00<0omnL09boo`3o
ool03P3oool00`3OW9`0_ooo0?ooo`090?ooo`030=nLB000B9`0gooo00<0oooo00D0oono07@00000
0000001d0;ooo`090?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0H0?ooo`030000003oool0oooo00X0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0F0?ooo`8000003@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo
00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo01D0oooo
00<000000?ooo`3oool03@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo
00<000000?ooo`3oool0903oool00300oooo00D0_g@0000000000000000004RLg`0T0?ooo`030000
003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0D0?ooo`030000
003oool0oooo00h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo02@0oooo000^0?ooo`030?oo_`1d0000000000@0000000<0B9cO0?ooo`3oool08@3o
ool2000002h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0B0?ooo`8000004@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00/0oooo0P00000V0?ooo`00;P3oool0
0`3OW4P007Bo0?ooo`020?ooo`040?oo_`1d0000001d0;ooob80oooo00<000000?ooo`3oool0;@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo0140oooo00<000000?ooo`3oool04@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903o
ool002h0oooo00<0_gAd0;ooo`3oool00`3oool00`3ookl0M01d0;ooo`0R0?ooo`030000003oool0
oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0@0?ooo`030000003oool0
oooo0180oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo02@0oooo000Y0?ooo`030?noW02Lgol0oooo0080oooo00<0omnL09boo`3oool00`3oool00`3o
_i`0W=oo0?ooo`0R0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0>0?ooo`8000005@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo
00`0oooo00<000000?ooo`3oool0903oool002T0oooo00<0_g@00000000000002@0000000`0007@0
_ooo0?ooo`0Q0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0=0?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0
oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0
oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oo_`1d0000000000X00000
00<007Bo0?ooo`3oool08@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00=03oool00`3ogi`0W;oo0?oo
o`0R0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0;0?ooo`030000003oool0oooo01L0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`090?ooo`8000006P3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V
0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool002l0oooo00<0oono07@0
000000000P0000000`18W=l0oooo0?ooo`0R0?ooo`030000003oool0oooo02d0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`080?ooo`030000003oool0oooo01X0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo000^0?ooo`03
0?oOW0180000000000@0000000<0B9cO0?ooo`3oool08@3oool00`000000oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool01`3oool00`000000oooo0?ooo`0K0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00
;P3oool02@3OW7@0W=oo0?noM01d_ol0oooo0?noM0000000001d0;ooo`0R0?ooo`800000;P3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00H0oooo00<000000?ooo`3oool0703oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00/0oooo0P00000V0?ooo`00;P3oool0102o
M4P0W=oo0?noM01d_ol20?ooo`030=nLB00007@0_ooo0280oooo00<000000?ooo`3oool0;@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo00D0oooo00<000000?ooo`3oool07@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool0
02h0oooo00@0okmd000000000000M;oo0P3oool00`2oM00007Bo0?ooo`0R0?ooo`030000003oool0
oooo02d0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`030?ooo`800000803oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool0
02l0oooo00H0gia80000001d_ol0okmd00000000M;lT0?ooo`030000003oool0oooo02d0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`020?ooo`030000003oool0oooo0200oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I
0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool01@0007l0oooo0?ooo`3o
ool0000002<0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool09@3oool0
100007l0oooo0?ooo`00000T0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303o
ool00`000000oooo0?ooo`0T0?ooo`00;`3oool02`3o_g@0000000000000B9`0gono07@00000M;l0
gia80000000007@0_ooo01l0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0000002D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?oo
o`030000003oool0oooo02@0oooo000^0?ooo`030?noM000000000000080000000L0001d0;noM000
0000B9cO0?ooo`2oM00007Bo01l0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`000000oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo02@0oooo000^0?ooo`0=0;mdM02oool0oooo0?ooo`3ooml0W4Qd0;oO
W01807@0_ooo0?ooo`3o_g@004RL0=ooo`0N0?ooo`030000003oool0oooo02d0oooo00<0001o0?oo
o`3oool09@3oool00`000000oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?oo
o`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00;P3oool00`2oM4P0W=oo0?ooo`020?ooo`08
0;mdM02ogi`0B01809cOo`3oool0omnL04Q8W03OoolN0?ooo`030000003oool0oooo02d0oooo00<0
001o0?ooo`3oool0903oool00`000000001o0?ooo`0V0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0
001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00;P3oool00`3ogi`0B00000000002
00000007001d_`3ogi`0B01809cOo`3oool0okmd001d_`0O0?ooo`800000;P3oool00`0007l0oooo
0?ooo`0R0?ooo`80000000<0oooo0000O`3oool09P3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00/0oooo0P00000V0?ooo`00;P3oool0303ookl0M00000000000000004RL0=ooo`3o
gi`0B00007Boo`3ooml0W4Q809cOoal0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?oo
o`0Q0?ooo`050000003oool0oooo0?ooo`0007l09`3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool002h0oooo00<0oono09bLg`3oool0103o
ool0102oM00000000000B02LgolP0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool0
803oool00`000000oooo0?ooo`020?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003o
ool0oooo02d0oooo00<0001o0?ooo`3oool07`3oool00`000000oooo0?ooo`030?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool07P3oool0
0`000000oooo0?ooo`040?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool0
0`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oOW02L_ol0oooo
00h0oooo00<0gibL0;ooo`3oool01P3oool01@3ogi`0B000000000000000M;oo00@0oooo00D0omnL
04P000000000001d0;ooo`090?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool07@3o
ool00`000000oooo0?ooo`050?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oOW02L_ol0
oooo00h0oooo00<0gibL0;ooo`3oool01@3oool01`3ogi`0B00004RLg`3oool0oono07@0001d_ol0
0P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`0]0?oo
o`030000O`3oool0oooo01`0oooo00<000000?ooo`3oool01P3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?oo
o`00:03oool00`3ogi`0M9cO0?ooo`0>0?ooo`030;mdM02oool0oooo00@0oooo00<0oooO09a80000
M;l00`3oool01`3o_g@004RL0=ooo`3oool0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`07
0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool06`3oool00`000000oooo0?ooo`07
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo02@0oooo000X0?ooo`030?oOW0180000000000h0000000<0001d0;oo
o`3oool0103oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo0?oo_`1d0000B9cO0`3o
ool00`3OW4P0001d0;ooo`070?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool06@3o
ool2000000X0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3o
ool0oooo00`0oooo00<000000?ooo`3oool0903oool00400oooo00<0omnL04P0001d_ol00`3oool0
1P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol01`3oool00`000000oooo
0?ooo`0]0?ooo`030000O`3oool0oooo01P0oooo00<000000?ooo`3oool02P3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0T0?ooo`00@03oool00`3ogi`0B00007Boo`030?ooo`060;md000004P0W=oo0?oOW0180000
M;oo0`3oool00`3o_g@0001809cOo`070?ooo`<00000A`0007l00`000000001o0000O`3Q0000O`<0
00009P3oool00400oooo00`0oono07@000000000B9cO0?ooo`3o_g@00000001d_`3oool0omnL04P0
001d_ol30?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`0F0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool00440oooo00<0
_g@00018W02oW4P00P0000001P00B9`0gooo0?ooo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02L
gol01`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo01D0oooo00<000000?ooo`3o
ool03@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3o
ool0303oool00`000000oooo0?ooo`0T0?ooo`00@@3oool00`3ogi`0B00007Boo`050?ooo`030?oo
_`1d0000B9cO00<0oooo00<0gia80000M02oool01`3oool00`000000oooo0?ooo`0]0?ooo`030000
O`3oool0oooo01<0oooo0P00000@0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo00120?ooo`030=nLB000
04P0W=oo00D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003o
ool0oooo02d0oooo00<0001o0?ooo`3oool04P3oool00`000000oooo0?ooo`0@0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003o
ool0oooo02@0oooo00130?ooo`030?noM000000007Bo00@0oooo00L0oono07@0001d_ol0oooo0?no
M00007@0_ooo00P0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`0A0?ooo`030000
003oool0oooo0140oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool004@0oooo00D0oooO09a800000000001d
0;ooo`020?ooo`050?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo0100oooo00<000000?ooo`3oool04P3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00
F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo00l0oooo00<000000?ooo`3oool0
4`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3o
ool0oooo00d0oooo0P00000F0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0
oooo02d0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0F0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo02@0oooo001I0?ooo`800000;P3oool00`0007l0oooo0?ooo`0;0?ooo`030000003oool0oooo
01L0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo
00/0oooo0P00000V0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo00X0
oooo00<000000?ooo`3oool0603oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0
oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo
0?ooo`0]0?ooo`030000O`3oool0oooo00T0oooo00<000000?ooo`3oool06@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo
0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo00L0oooo0P00
000L0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?oo
o`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o
0?ooo`3oool01P3oool00`000000oooo0?ooo`0L0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?oo
o`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool01@3oool00`000000oooo0?ooo`0M0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?oo
o`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3o
ool0103oool00`000000oooo0?ooo`0N0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000
003oool0oooo02d0oooo00<0001o0?ooo`3oool00`3oool00`000000oooo0?ooo`0O0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`3oool00P3o
ool00`000000oooo0?ooo`0P0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0
oooo02d0oooo00D0001o0?ooo`3oool0oooo0000000S0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I
0?ooo`800000;P3oool0100007l0oooo0?ooo`00000T0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0
001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0;0?ooo`8000009P3oool005T0oooo00<00000
0?ooo`3oool0;@3oool00`0007l000000000000U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?oo
o`030000003oool0oooo02d0oooo00<000000?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?oo
o`00F@3oool00`000000oooo0?ooo`0/0?ooo`030000000007l0oooo02H0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3o
ool0903oool005T0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0000O`0W0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000
003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo02X0oooo00@000000?ooo`3oool0001o
9`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0Y0?ooo`050000003o
ool0oooo0?ooo`0007l09`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo
00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo
00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?oo
o`0X0?ooo`030000003oool0oooo0080oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?oo
o`0T0?ooo`00F@3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00<0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o
0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o
0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo00@0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?oo
o`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?oo
o`00F@3oool2000002H0oooo00<000000?ooo`3oool01@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0;0?ooo`80
00009P3oool005T0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`060?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000
O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V0?ooo`030000
O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?ooo`3oool08P3o
ool2000000T0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3o
ool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00T0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool0
0`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool0
0`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo
00X0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00@P3oool01@3OW4P0
0000000000000000B9cO00@0oooo00D0omnL04P000000000001d0;ooo`090?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool02`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo02@0oooo00110?ooo`070=nLB0000000M;oo0?ooo`3ooml0W4P004RLg`020?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`080?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3o
ool0303oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo00100?ooo`030?oo
g`2LB00007Bo00<0oooo00L0okmd0018W03Oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool01`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00d0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?oo
o`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00@@3oool00`2oM00007Bo0?ooo`020?ooo`06
0;md0000B9`0gooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`070?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool03P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo02@0oooo00110?ooo`0;0?oOW01807@0_ooo0?ooo`2oM000000007Boo`3oool0omnL04P0001d
_ol00`3oool00`3o_g@0001809cOo`070?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool03`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3o
ool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo00120?ooo`050?oo
_`1dB4P0B00000000018W=l00P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0
oooo0`00000J0000O`030000000007l0001o0?l0001o3`0007l3000002H0oooo00120?ooo`040?oO
W01800000000001d_`<0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool04@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`03
0000003oool0oooo02@0oooo00110?ooo`060?oOW0180000001d0;ooo`3o_g@007Bo0P3oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`0B0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0V0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool00440
oooo00<0_g@00000M02oool00P3oool01P2oM000B9cO0?ooo`3ookl0M00004RLg`<0oooo00<0gia8
0000M02oool01`3oool00`000000oooo0?ooo`0F0?ooo`8000005@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<
0?ooo`030000003oool0oooo02@0oooo00110?ooo`030;md0000B9`0gooo0080oooo00L0gia8001d
_`3oool0oooo0=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?oo
o`0T0?ooo`00@@3oool01`3o_g@0000007Boo`3oool0oono07@0001d_ol00P3oool01`3ookl0M000
07Boo`3oool0okmd0000M02oool0203oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
01H0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo
02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00@P3oool01@3o_g@0
0000000000000000B9cO00@0oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool05`3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo02@0oooo001I0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0603oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0
oooo001I0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool06P3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`800000403oool00`000000oooo0?oo
o`0K0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0V0?ooo`030000O`3oool0oooo00/0oooo0P00000V0?ooo`00F@3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo01`0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T
0?ooo`00F@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01d0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?oo
o`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?oo
o`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo01h0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`03
0000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0303oool00`000000oooo0?ooo`0T0?ooo`00
F@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool0
9@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02H0oooo00<0001o0?ooo`3oool0
303oool00`000000oooo0?ooo`0T0?ooo`00F@3oool00`000000oooo0?ooo`090?ooo`8000008P3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3o
ool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3o
ool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool08P3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0
oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0
oooo02@0oooo001I0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool08`3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`00
07l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`00
07l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0903oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U0?ooo`030000O`3oool0oooo02D0
oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo0?ooo`0<0?ooo`030000003oool0oooo02@0
oooo001I0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo
0?ooo`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09P3oool00`0007l0oooo
0?ooo`0<0?ooo`030000003oool0oooo02@0oooo001I0?ooo`8000001@3oool00`000000oooo0?oo
o`0V0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0U0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?oo
o`0V0?ooo`030000O`3oool0oooo00/0oooo0P00000V0?ooo`00F@3oool00`000000oooo0?ooo`0]
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0U
0?ooo`030000O`3oool0oooo02D0oooo00<0001o0?ooo`3oool09@3oool00`0007l0oooo0?ooo`0V
0?ooo`030000O`3oool0oooo00`0oooo00<000000?ooo`3oool0903oool005T0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool09@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool09@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0903oool005T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo02@0oooo001I0?ooool00000<P00000V0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3oool6
0?ooo`030?oog`2LB000000000<0000000<0B9cO0?ooo`3oool0XP3oool00?l0oooo1P3oool00`2o
M00004RL0=ooo`030?ooo`030=nLW02oool0oooo0:80oooo003o0?ooo`H0oooo00<0oooO09adW03O
ool00`3oool00`3ookl0W9cO0?ooo`2R0?ooo`00o`3oool70?ooo`030;md00000000000000<00000
00<0B9cO0?ooo`3oool0X@3oool00<L0oooo00D0_g@0000000000000001809cOo`030?ooo`050;md
0000000000000000M02oool00`3oool01@2oM00000000000000004P0W=oo00<0oooo00D0gia80000
00000000000007Boo`020?ooo`050=nLB000000000000000M02LW4P00P0000000`0007@0_ooo0?oo
o`040?ooo`090?oog`2LB000000000000000B9`0W7@000000000M;l0_g@00080000000<0001d0;oo
g`2LB0000P0000000`00B9`0gono07@00004000000030000M02oool0gia80080000000@0001d0;oo
o`3oool0_g@00P0000000`0007@0_ooo0?ooo`2G0?ooo`00b03oool00`3OW4P0001d0;ooo`030?oo
o`0G0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0_g@000000000B9`0gooo0?oog`2LB7@0
_ooo0?ooo`3OW4P0000004RLg`3oool0oooO09a80018W=l00P3oool01`3OW4P0001d0;ooo`3oool0
gia80000M02oool01P3oool0403o_g@0000007Boo`3oool0omnL04P00000M;l0okno0;ooo`3OW4P0
04RL0=ooo`3oool0oooO09a80000M;l20?ooo`030;mdB02Lgol0oooo00D0oooo00/0gia80000M02o
ool0oooo0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`2G0?ooo`00b03oool00`3OW4P0001d0;oo
o`020?ooo`030?oOW0180000B9cO00<0oooo00D0omnL07BLg`3ogi`0B000001d_`030?ooo`050?oO
W02L_ol0omnL04P00000M;l00`3oool02`3o_g@0001d0;ooo`3oool0gia80000M02oool0oooo0=nL
B00007@0_ooo00H0oooo00L0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00L0gia80018
W03Oool0oooo0?oog`2LB00007Bo0080oooo00D0oono07@0000000000000001d_`030?ooo`060=nL
B00007@0_ooo0?oOW0180000B9cO0`3oool00`3ogi`0M9cO0?ooo`2F0?ooo`00b03oool00`3OW4P0
001d0;ooo`020?ooo`030?noM00004P0W=oo00D0oooo00<0okmd0000B02Lgol01@3oool00`3o_g@0
000007Boo`030?ooo`0;0?noM0000000M;oo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02oool0
1`3oool01P3o_g@00000001d_`3ooml0W4P0001d_`80oooo00L0gia80018W03Oool0oooo0?oog`2L
B00007Bo0080oooo00H0gia80000M02oool0oono07@00018W=l20?ooo`060=nLB00007@0_ooo0?no
M00004P0W=ooW03oool00<P0oooo00<0gia80000M02oool00P3oool00`3o_g@0001d0;ooo`050?oo
o`030?noM00007@0_ooo00D0oooo00<0okmd0000001d_ol00`3oool02`3o_g@0000007Boo`3oool0
gia80000M02oool0oooo0=nLB00007@0_ooo00T0oooo00@0okmd00000000000007Bo0P3oool03`3O
W4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0?ooo`3ooml0W4P0001d_`02
0?ooo`060=nLB00007@0_ooo0?noM00007@0_oooW03oool00<P0oooo00<0gia80000B02Lgol00P3o
ool00`3ogi`0B000000000040000000404RLg`3ogi`0B01809cOo`80oooo00H0okmd0000B02Lgol0
oono07@0001d_ol30?ooo`0;0;md000004P0W=oo0?ooo`3OW4P0001d0;ooo`3oool0gia80000M02o
ool01P3oool01`3ookl0M00007Boo`3oool0oooO09a80000M;l00P3oool03`2oM000001d0;ooo`3o
ool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0?ooo`3ooml0W4P004RLg`020?ooo`050=nLB000
07@0_ooo0?oOW01800001@0000000`18W=l0oooo0?ooo`2E0?ooo`00aP3oool02P3ooml0W4P00000
000004P0B7BL0=noM00007@0_ooo0;md0018W=l20?ooo`0L0=nLB00004P0W=oo0?ooo`2oM00004RL
0=ooo`3ooml0W4P00000M02oool0oooo0=nLB000B9`0gooo0?ooo`3o_g@00000001d_`3ooml0W4P0
0000000007@0_ooo0?ooo`3OW4P0001d0;ooo`H0oooo0100oono07@00018W=l0oooo0?oo_`1d0000
07Bo0?oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0P3oool0102oM00007Bo0?ooo`3o
_g@2000000080018W03Oool0gia80000M02oool0oooo0;md0018W=l20?ooo`030=nLB00004P0W=oo
09L0oooo00380?ooo`080;md000007@0_mnL04P0000007@0_ooo0?ooo`3OW4P2000000030000B02L
gol0oooo0080oooo00D0gia8000000000000001809cOo`030?ooo`050?noM0000000000000000018
W=l00`3oool01`2oM000001d0;ooo`3oool0gia80000M02oool01`3oool01P3ookl0M00000000000
0000001d0;ooo`80oooo00L0_g@00018W03Ooml0W4P00000000007@0_ooo0080oooo00D0oooO09a8
00000000000007Boo`030?ooo`030=nLB00007@0_ooo0080oooo00D0gia8000000000000001809cO
o`2H0?ooo`00j`3oool00`3OW4P0001d0;ooo`0P0?ooo`030=nLB00007@0_ooo09l0oooo003[0?oo
o`030=nLB00007@0_ooo0200oooo00<0gia80000M02oool0W`3oool00>/0oooo00<0gia80000M02o
ool0803oool00`3OW4P0001d0;ooo`2O0?ooo`00i`3oool01`3OW4P0001809cOg`2LB00000000000
M02oool07P3oool01@3ooml0W4P00000000007@0_ooo09l0oooo003W0?ooo`070=nLB00004P0W=oo
0?ooo`2oM000001d0;ooo`0P0?ooo`030;md000007@0_ooo09l0oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-60.2947, -27.8575, \
0.933529, 0.75524}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Maximum scattering angle", "Subsection"],

Cell["\<\
We have encountered several situations in which there is a maximum possible \
scattering angle.  For simplicity we assume that the target is at rest in the \
lab.  Then for elastic collisions there is a maximum scattering angle when \
the projectile is heavier than the target.  Alternatively, for endothermic \
reactions there is a maximum scattering angle even when the projectile mass \
is less than or equal to the target mass.   To obtain a general symbolic \
relationship for the maximum scattering angle, we apply conservation of \
momentum and energy to the triangle formed by the incident momentum and the \
two exit momenta, as  shown below.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawVectorSum[LabMomenta\[LeftDoubleBracket]3\[RightDoubleBracket], 
        LabMomenta\[LeftDoubleBracket]4\[RightDoubleBracket]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .29742 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.00708153 0.952381 [
[ 0 0 0 0 ]
[ 1 .29742 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 .29742 L
0 .29742 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.02381 .00708 m
.80206 .29034 L
s
% Start of user PostScript
0. 0. 0.81716593 0.29742408 MAarrowhead1
% End of user PostScript
.80206 .29034 m
.97619 .00708 L
s
% Start of user PostScript
0.81716593 0.29742408 1. 0. MAarrowhead1
% End of user PostScript
1 Mabswid
.02381 .00708 m
.97619 .00708 L
s
% Start of user PostScript
0. 0. 1. 0. MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 85.625},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0000P2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooonM0?ooo`8000004P3oool00?l0ooooW@3oool6000000h0oooo000;0?ooool00000
VP00000<0?ooo`003@3oool300000?l0ooooS@3oool8000000`0oooo000@0?ooo`800000o`3ooon;
0?ooo`8000000`3oool2000000d0oooo000B0?ooo`<00000o`3ooon<0?ooo`<000003@3oool001D0
oooo0`00003o0?ooohP0oooo1000000=0?ooo`00603oool300000?l0ooooQ03oool5000000d0oooo
000K0?ooo`800000o`3ooon10?ooo`D000003P3oool001d0oooo0`00003o0?oooh00oooo0`00000>
0?ooo`00803oool300000?l0ooooO03oool00`000000oooo0?ooo`0?0?ooo`008`3oool300000?l0
ooooN03oool00`000000oooo0?ooo`0@0?ooo`009P3oool200000?l0ooooMP3oool00`000000oooo
0?ooo`0@0?ooo`00:03oool300000?l0ooooLP3oool00`000000oooo0?ooo`0A0?ooo`00:`3oool3
00000?l0ooooK`3oool00`000000oooo0?ooo`0A0?ooo`00;P3oool300000?l0ooooJ`3oool00`00
0000oooo0?ooo`0B0?ooo`00<@3oool200000?l0ooooJ03oool00`000000oooo0?ooo`0C0?ooo`00
<`3oool300000?l0ooooI@3oool00`000000oooo0?ooo`0C0?ooo`00=P3oool300000?l0ooooH@3o
ool00`000000oooo0?ooo`0D0?ooo`00>@3oool300000?l0ooooG@3oool00`000000oooo0?ooo`0E
0?ooo`00?03oool200000?l0ooooF`3oool00`000000oooo0?ooo`0E0?ooo`00?P3oool300000?l0
ooooE`3oool00`000000oooo0?ooo`0F0?ooo`00@@3oool300000?l0ooooE03oool00`000000oooo
0?ooo`0F0?ooo`00A03oool300000?l0ooooD03oool00`000000oooo0?ooo`0G0?ooo`00A`3oool2
00000?l0ooooC@3oool00`000000oooo0?ooo`0H0?ooo`00B@3oool300000?l0ooooBP3oool00`00
0000oooo0?ooo`0H0?ooo`00C03oool300000?l0ooooAP3oool00`000000oooo0?ooo`0I0?ooo`00
C`3oool300000?l0oooo@`3oool00`000000oooo0?ooo`0I0?ooo`00DP3oool200000?l0oooo@03o
ool00`000000oooo0?ooo`0J0?ooo`00E03oool300000?l0oooo?03oool00`000000oooo0?ooo`0K
0?ooo`00E`3oool300000?l0oooo>@3oool00`000000oooo0?ooo`0K0?ooo`00FP3oool300000?l0
oooo=@3oool00`000000oooo0?ooo`0L0?ooo`00G@3oool200000?l0oooo<P3oool00`000000oooo
0?ooo`0M0?ooo`00G`3oool300000?l0oooo;`3oool00`000000oooo0?ooo`0M0?ooo`00HP3oool3
00000?l0oooo:`3oool00`000000oooo0?ooo`0N0?ooo`00I@3oool300000?l0oooo:03oool00`00
0000oooo0?ooo`0N0?ooo`00J03oool200000?l0oooo9@3oool00`000000oooo0?ooo`0O0?ooo`00
JP3oool300000?l0oooo8@3oool00`000000oooo0?ooo`0P0?ooo`00K@3oool300000?l0oooo7P3o
ool00`000000oooo0?ooo`0P0?ooo`00L03oool300000?l0oooo6P3oool00`000000oooo0?ooo`0Q
0?ooo`00L`3oool200000?l0oooo5`3oool00`000000oooo0?ooo`0R0?ooo`00M@3oool300000?l0
oooo503oool00`000000oooo0?ooo`0R0?ooo`00N03oool300000?l0oooo403oool00`000000oooo
0?ooo`0S0?ooo`00N`3oool300000?l0oooo3@3oool00`000000oooo0?ooo`0S0?ooo`00OP3oool2
00000?l0oooo2P3oool00`000000oooo0?ooo`0T0?ooo`00P03oool300000?l0oooo1P3oool00`00
0000oooo0?ooo`0U0?ooo`00P`3oool300000?l0oooo0`3oool00`000000oooo0?ooo`0U0?ooo`00
QP3oool300000?h0oooo00<000000?ooo`3oool09P3oool008T0oooo0P00003k0?ooo`030000003o
ool0oooo02L0oooo002;0?ooo`<00000n03oool00`000000oooo0?ooo`0W0?ooo`00SP3oool30000
0?@0oooo00<000000?ooo`3oool0:03oool00940oooo0`00003a0?ooo`030000003oool0oooo02P0
oooo002D0?ooo`800000kP3oool00`000000oooo0?ooo`0Y0?ooo`00UP3oool300000>X0oooo00<0
00000?ooo`3oool0:P3oool009T0oooo0`00003W0?ooo`030000003oool0oooo02X0oooo002L0?oo
o`<00000h`3oool00`000000oooo0?ooo`0[0?ooo`00W`3oool200000>40oooo00<000000?ooo`3o
ool0:`3oool00:40oooo0`00003M0?ooo`030000003oool0oooo02`0oooo002T0?ooo`<00000f@3o
ool00`000000oooo0?ooo`0]0?ooo`00Y`3oool300000=H0oooo00<000000?ooo`3oool0;@3oool0
0:X0oooo0P00003C0?ooo`030000003oool0oooo02h0oooo002/0?ooo`<00000c`3oool00`000000
oooo0?ooo`0_0?ooo`00[`3oool300000<`0oooo00<000000?ooo`3oool0;`3oool00;80oooo0`00
00380?ooo`030000003oool0oooo0300oooo002e0?ooo`<00000a@3oool00`000000oooo0?ooo`0`
0?ooo`00^03oool200000<80oooo00<000000?ooo`3oool0<@3oool00;X0oooo0`00002n0?ooo`03
0000003oool0oooo0380oooo002m0?ooo`<00000^`3oool00`000000oooo0?ooo`0b0?ooo`00`03o
ool300000;L0oooo00<000000?ooo`3oool0<`3oool00<<0oooo0P00002d0?ooo`030000003oool0
oooo03@0oooo00350?ooo`<00000/@3oool00`000000oooo0?ooo`0d0?ooo`00b03oool300000:d0
oooo00<000000?ooo`3oool0=@3oool00</0oooo0`00002Z0?ooo`030000003oool0oooo03D0oooo
003>0?ooo`800000Y`3oool00`000000oooo0?ooo`0f0?ooo`00d03oool300000:<0oooo00<00000
0?ooo`3oool0=`3oool00=<0oooo0`00002P0?ooo`030000003oool0oooo03L0oooo003F0?ooo`<0
0000W03oool00`000000oooo0?ooo`0h0?ooo`00f@3oool2000009X0oooo00<000000?ooo`3oool0
>03oool00=/0oooo0`00002F0?ooo`030000003oool0oooo03T0oooo003N0?ooo`<00000TP3oool0
0`000000oooo0?ooo`0j0?ooo`00h@3oool3000008l0oooo00<000000?ooo`3oool0>P3oool00>@0
oooo0P00002<0?ooo`030000003oool0oooo03/0oooo003V0?ooo`<00000R03oool00`000000oooo
0?ooo`0l0?ooo`00j@3oool3000008D0oooo00<000000?ooo`3oool0?03oool00>`0oooo0`000021
0?ooo`030000003oool0oooo03d0oooo003_0?ooo`800000O`3oool00`000000oooo0?ooo`0m0?oo
o`00l@3oool3000007/0oooo00<000000?ooo`3oool0?P3oool00?@0oooo0`00001g0?ooo`030000
003oool0oooo03l0oooo003g0?ooo`<00000M03oool00`000000oooo0?ooo`0o0?ooo`00nP3oool2
00000740oooo00<000000?ooo`3oool0@03oool00?`0oooo0`00001]0?ooo`030000003oool0oooo
0440oooo003o0?ooo`<00000JP3oool00`000000oooo0?ooo`110?ooo`00o`3oool30?ooo`<00000
IP3oool00`000000oooo0?ooo`120?ooo`00o`3oool60?ooo`800000I03oool00`000000oooo0?oo
o`120?ooo`00o`3oool80?ooo`<00000H03oool00`000000oooo0?ooo`130?ooo`00o`3oool;0?oo
o`<00000G03oool00`000000oooo0?ooo`140?ooo`00o`3oool>0?ooo`<00000F@3oool00`000000
oooo0?ooo`140?ooo`00o`3ooolA0?ooo`800000EP3oool00`000000oooo0?ooo`150?ooo`00o`3o
oolC0?ooo`<00000DP3oool00`000000oooo0?ooo`160?ooo`00o`3ooolF0?ooo`<00000C`3oool0
0`000000oooo0?ooo`160?ooo`00o`3ooolI0?ooo`<00000B`3oool00`000000oooo0?ooo`170?oo
o`00o`3ooolL0?ooo`800000B@3oool00`000000oooo0?ooo`170?ooo`00o`3ooolN0?ooo`<00000
A@3oool00`000000oooo0?ooo`180?ooo`00o`3ooolQ0?ooo`<00000@@3oool00`000000oooo0?oo
o`190?ooo`00o`3ooolT0?ooo`<00000?P3oool00`000000oooo0?ooo`190?ooo`00o`3ooolW0?oo
o`800000>`3oool00`000000oooo0?ooo`1:0?ooo`00o`3ooolY0?ooo`<00000>03oool00`000000
oooo0?ooo`1:0?ooo`00o`3oool/0?ooo`<00000=03oool00`000000oooo0?ooo`1;0?ooo`00o`3o
ool_0?ooo`<00000<03oool00`000000oooo0?ooo`1<0?ooo`00o`3ooolb0?ooo`800000;P3oool0
0`000000oooo0?ooo`1<0?ooo`00o`3ooold0?ooo`<00000:P3oool00`000000oooo0?ooo`1=0?oo
o`00o`3ooolg0?ooo`<000009P3oool00`000000oooo0?ooo`1>0?ooo`00o`3ooolj0?ooo`<00000
8`3oool00`000000oooo0?ooo`1>0?ooo`00o`3ooolm0?ooo`800000803oool00`000000oooo0?oo
o`1?0?ooo`00o`3ooolo0?ooo`<000007@3oool00`000000oooo0?ooo`1?0?ooo`00o`3ooom20?oo
o`<000006@3oool00`000000oooo0?ooo`1@0?ooo`00o`3ooom50?ooo`<000005@3oool00`000000
oooo0?ooo`1A0?ooo`00o`3ooom80?ooo`8000004`3oool00`000000oooo0?ooo`1A0?ooo`00o`3o
oom:0?ooo`<000003`3oool00`000000oooo0?ooo`1B0?ooo`00o`3ooom=0?ooo`<000000`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo05<0oooo003o0?oooe00oooo1P0000050?oo
o`030000003oool0oooo05<0oooo003o0?oooe80oooo1@0000030?ooo`030000003oool0oooo05@0
oooo003o0?oooe80oooo1`0000000`3oool000000?ooo`1E0?ooo`00o`3ooomB0?ooo`P00000E`3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {127.438, 0}} -> {-0.0280919, -0.00743561, \
0.00367581, 0.00367581}}]
}, Closed]],

Cell[TextData[{
  "Use conservation of energy and apply the law of cosines, where ",
  StyleBox["x = Cos[theta3]", "InlineInput"],
  ".  We eliminate ",
  StyleBox["p4", "InlineInput"],
  " and express ",
  StyleBox["p3", "InlineInput"],
  " in units of ",
  StyleBox["p1", "InlineInput"],
  " and the ",
  StyleBox["qvalue", "InlineInput"],
  " as a fraction of the initial kinetic energy, such that ",
  StyleBox["qvalue\[Rule]-inelasticity ", "InlineInput"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`p1\^2\)], "InlineInput"], "InlineInput"],
  StyleBox["/(2 m1)", "InlineInput"],
  ".  Similarly, we define ",
  StyleBox["mratio\[Rule]m2/m1", "InlineInput"],
  " as the ratio of masses."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq = 
      Simplify[\(\(Eliminate[{p1\^2\/\(2\ m1\) + qvalue \[Equal] 
                    p3\^2\/\(2\ m1\) + p4\^2\/\(2\ m2\), 
                  p4\^2 \[Equal] p1\^2 + p3\^2 - 2\ p1\ p3\ x}, 
                p4]\)\[LeftDoubleBracket]1\[RightDoubleBracket] /. {qvalue \
\[Rule] \(-\(\(inelasticity\ p1\^2\)\/\(2\ m1\)\)\), 
              m2 \[Rule] m1\ mratio}\) /. {p1 \[Rule] 1, 
            m1 \[Rule] 1}]\)], "Input"],

Cell[BoxData[
    \(\(1 + p3\^2 + mratio\ \((\(-1\) + inelasticity + p3\^2)\) - 2\ p3\ \
x\)\/mratio \[Equal] 0\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(xsol = 
      x /. \(Solve[eq, 
            x]\)\[LeftDoubleBracket]1\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \(\(1 - mratio + inelasticity\ mratio + p3\^2 + mratio\ p3\^2\)\/\(2\ \
p3\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Now we maximize ",
  StyleBox["theta3", "InlineInput"],
  " by minimizing ",
  StyleBox["x", "InlineInput"],
  " wrt ",
  StyleBox["p3", "InlineInput"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(p3sol = 
      Simplify[\(Solve[\[PartialD]\_p3 xsol \[Equal] 0, 
            p3]\)\[LeftDoubleBracket]1\[RightDoubleBracket]]\)], "Input"],

Cell[BoxData[
    \({p3 \[Rule] \(-\(\@\(1 + \((\(-1\) + inelasticity)\)\ mratio\)\/\@\(1 + \
mratio\)\)\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(cos2 = \(xsol\^2 /. p3sol\  // Simplify\)\  // 
        Collect[#, inelasticity] &\)], "Input"],

Cell[BoxData[
    \(\((1 - mratio)\)\ \((1 + mratio)\) + 
      inelasticity\ mratio\ \((1 + mratio)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Finally, recognizing that the expression above gives ",
  Cell[BoxData[
      \(TraditionalForm\`Cos[\[Theta]\_3]\^2\)]],
  ", we construct a maximum angle function that recognizes that there are \
situations in which the scattering angle is not restricted.  This function \
depends upon the mass ratio and the relative inelasticity (negative is \
endothermic, positive is exothermic) and is displayed in two forms below."
}], "Text"],

Cell[BoxData[{
    \(\(Clear[thetaMax];\)\), "\n", 
    \(thetaMax[m_, x_] := 
      Module[{temp}, 
        temp = cos2 /. {mratio \[Rule] m, inelasticity \[Rule] x}; 
        If[temp > 0 && temp < 1, ArcCos[\@temp], \[Pi]]]\)}], "Input"],

Cell[BoxData[
    \(\(plotThetaMax3D = 
        Plot3D[Evaluate[thetaMax[mratio, inelasticity]\/Degree], {mratio, 0, 
            1}, {inelasticity, 0, 0.5}, PlotPoints \[Rule] 25, 
          PlotRange \[Rule] {Automatic, Automatic, {0, 180}}, 
          AxesLabel \[Rule] {"\<mass ratio\>", "\<relative inelasticity\>", \
\*"\"\<Max[\!\(\[Theta]\_3\)] (deg)\>\""}, 
          PlotLabel \[Rule] "\<Maximum Scattering Angle\>", 
          DisplayFunction \[Rule] Identity];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[plotThetaMax3D, ViewPoint \[Rule] {3, 2, 1}, 
        BoxRatios \[Rule] {1, 1, 1}, PlotLabel \[Rule] {}, 
        DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .94581 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% SurfaceGraphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.0219431 1.08092 1.98788e-016 1.08092 [
[.5 .95831 -5.9375 0 ]
[.5 .95831 5.9375 15.6875 ]
[.43414 .95932 -5.04311 0 ]
[.43414 .95932 4.95689 15.6875 ]
[.36072 .94703 -9.65336 0 ]
[.36072 .94703 9.34664 15.6875 ]
[.28135 .93375 -9.73082 0 ]
[.28135 .93375 9.26918 15.6875 ]
[.19527 .91935 -9.81506 0 ]
[.19527 .91935 9.18494 15.6875 ]
[.1016 .90367 -9.90701 0 ]
[.1016 .90367 9.09299 15.6875 ]
[-0.00072 .88655 -5.26726 0 ]
[-0.00072 .88655 4.73274 15.6875 ]
[.23727 .99423 -27.1624 0 ]
[.23727 .99423 25.6501 15.6875 ]
[.03776 .16666 -4.73274 -15.6875 ]
[.03776 .16666 5.26726 0 ]
[.1424 .13486 -9.10132 -15.6875 ]
[.1424 .13486 9.89868 0 ]
[.25338 .10114 -9.21789 -15.6875 ]
[.25338 .10114 9.78211 0 ]
[.37129 .06531 -9.34272 -15.6875 ]
[.37129 .06531 9.65728 0 ]
[.49681 .02717 -9.47671 -15.6875 ]
[.49681 .02717 9.52329 0 ]
[.63069 -0.01351 -9.6209 -15.6875 ]
[.63069 -0.01351 9.3791 0 ]
[.31299 .01596 -46.1824 -15.6875 ]
[.31299 .01596 48.3801 0 ]
[.02483 .17436 -10 -11.564 ]
[.02483 .17436 0 4.1235 ]
[.01471 .35756 -16 -10.9233 ]
[.01471 .35756 0 4.76424 ]
[.00423 .54839 -22 -10.2618 ]
[.00423 .54839 0 5.42569 ]
[-0.00663 .74734 -22 -9.57864 ]
[-0.00663 .74734 0 6.10886 ]
[-0.05789 .48869 -71.75 -10.3958 ]
[-0.05789 .48869 0 5.2917 ]
[ 0 0 0 0 ]
[ 1 .94581 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
gsave
.5 .95831 -66.9375 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
63.000 15.813 moveto
(8) show
66.938 15.813 moveto
(<) show
1.000 setlinewidth
grestore
.25 Mabswid
[ ] 0 setdash
.43426 .94581 m
0 .87305 L
s
.43426 .94581 m
.43432 .93905 L
s
gsave
.43414 .95932 -66.0431 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.36094 .93352 m
.36105 .92677 L
s
gsave
.36072 .94703 -70.6534 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.28168 .92025 m
.28185 .91349 L
s
gsave
.28135 .93375 -70.7308 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.19572 .90584 m
.19595 .89909 L
s
gsave
.19527 .91935 -70.8151 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.10218 .89017 m
.10247 .88342 L
s
gsave
.1016 .90367 -70.907 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .87305 m
.00036 .86631 L
s
gsave
-0.00072 .88655 -66.2673 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.41645 .94282 m
.41649 .93877 L
s
.3983 .93978 m
.39835 .93573 L
s
.3798 .93668 m
.37986 .93263 L
s
.34171 .9303 m
.34179 .92625 L
s
.3221 .92702 m
.32218 .92296 L
s
.30209 .92367 m
.30218 .91961 L
s
.26085 .91676 m
.26096 .9127 L
s
.23959 .91319 m
.23971 .90914 L
s
.21789 .90956 m
.21801 .90551 L
s
.17309 .90205 m
.17323 .898 L
s
.14996 .89818 m
.15011 .89413 L
s
.12633 .89422 m
.12649 .89017 L
s
.07749 .88604 m
.07767 .88199 L
s
.05224 .88181 m
.05244 .87776 L
s
.02642 .87748 m
.02662 .87343 L
s
gsave
.23727 .99423 -88.1624 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(mass) show
90.000 15.813 moveto
(ratio) show
111.813 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.03704 .18015 m
.63086 0 L
s
.03704 .18015 m
.03668 .1869 L
s
gsave
.03776 .16666 -65.7327 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.14183 .14836 m
.14155 .15511 L
s
gsave
.1424 .13486 -70.1013 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
.25297 .11464 m
.25277 .1214 L
s
gsave
.25338 .10114 -70.2179 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.37106 .07882 m
.37095 .08557 L
s
gsave
.37129 .06531 -70.3427 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.3) show
1.000 setlinewidth
grestore
.49677 .04068 m
.49676 .04744 L
s
gsave
.49681 .02717 -70.4767 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.63086 0 m
.63095 .00676 L
s
gsave
.63069 -0.01351 -70.6209 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.05751 .17394 m
.05731 .17799 L
s
.07823 .16766 m
.07803 .17171 L
s
.09918 .1613 m
.09899 .16535 L
s
.12038 .15487 m
.1202 .15892 L
s
.16353 .14178 m
.16337 .14583 L
s
.18549 .13512 m
.18534 .13917 L
s
.20772 .12837 m
.20758 .13242 L
s
.23021 .12155 m
.23008 .1256 L
s
.27602 .10765 m
.27591 .1117 L
s
.29934 .10058 m
.29924 .10463 L
s
.32295 .09341 m
.32286 .09747 L
s
.34686 .08616 m
.34678 .09021 L
s
.39557 .07138 m
.39552 .07543 L
s
.42039 .06385 m
.42035 .06791 L
s
.44553 .05623 m
.4455 .06028 L
s
.47099 .0485 m
.47097 .05256 L
s
.5229 .03276 m
.5229 .03681 L
s
.54936 .02473 m
.54937 .02878 L
s
.57617 .01659 m
.5762 .02065 L
s
.60334 .00835 m
.60337 .0124 L
s
gsave
.31299 .01596 -107.182 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(relative) show
101.625 15.813 moveto
(inelasticity) show
153.563 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.03704 .18015 m
0 .87305 L
s
.03704 .18015 m
.04314 .18305 L
s
gsave
.02483 .17436 -71 -15.564 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.02729 .3625 m
.03358 .36496 L
s
gsave
.01471 .35756 -77 -14.9233 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.01714 .55237 m
.0236 .55436 L
s
gsave
.00423 .54839 -83 -14.2618 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.00656 .75026 m
.01316 .75172 L
s
gsave
-0.00663 .74734 -83 -13.5786 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.125 Mabswid
.03512 .21604 m
.0388 .21773 L
s
.03319 .25222 m
.03689 .25385 L
s
.03124 .28868 m
.03497 .29027 L
s
.02927 .32544 m
.03302 .32697 L
s
.02529 .39985 m
.02909 .40128 L
s
.02328 .43752 m
.0271 .43889 L
s
.02125 .47549 m
.02509 .4768 L
s
.0192 .51377 m
.02306 .51503 L
s
.01506 .59129 m
.01895 .59243 L
s
.01296 .63054 m
.01687 .63161 L
s
.01085 .67011 m
.01477 .67112 L
s
.00871 .71002 m
.01266 .71096 L
s
.00439 .79084 m
.00837 .79165 L
s
.00221 .83177 m
.00619 .83251 L
s
gsave
-0.05789 .48869 -132.75 -14.3958 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Max) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.063 15.813 moveto
(@) show
89.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
94.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
99.063 15.813 moveto
(D) show
106.000 15.813 moveto
(H) show
109.813 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
126.938 15.813 moveto
(L) show
130.750 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.25 Mabswid
.43922 .37091 m
.43426 .94581 L
s
.43426 .94581 m
1 .90077 L
s
1 .90077 m
.9679 .25222 L
s
.9679 .25222 m
.43922 .37091 L
s
.03704 .18015 m
.63086 0 L
s
.63086 0 m
.64108 .8025 L
s
.64108 .8025 m
0 .87305 L
s
0 .87305 m
.03704 .18015 L
s
.43922 .37091 m
.43426 .94581 L
s
.43426 .94581 m
0 .87305 L
s
0 .87305 m
.03704 .18015 L
s
.03704 .18015 m
.43922 .37091 L
s
.9679 .25222 m
.63086 0 L
s
.63086 0 m
.64108 .8025 L
s
.64108 .8025 m
1 .90077 L
s
1 .90077 m
.9679 .25222 L
s
0 0 m
1 0 L
1 .94581 L
0 .94581 L
closepath
clip
newpath
.5 Mabswid
.588 .45 .662 r
.43426 .94581 .45499 .94416 .44488 .36489 .42524 .37154 Metetra
.512 .898 .951 r
.42524 .37154 .44488 .36489 .43077 .36561 .41103 .3722 Metetra
.504 .896 .949 r
.41103 .3722 .43077 .36561 .41642 .36622 .39659 .3729 Metetra
.504 .896 .948 r
.39659 .3729 .41642 .36622 .40183 .36684 .3819 .37365 Metetra
.506 .897 .947 r
.3819 .37365 .40183 .36684 .38699 .36752 .36696 .37447 Metetra
.508 .899 .946 r
.36696 .37447 .38699 .36752 .3719 .36827 .35177 .37537 Metetra
.51 .9 .945 r
.35177 .37537 .3719 .36827 .35654 .36911 .33631 .37637 Metetra
.512 .902 .943 r
.33631 .37637 .35654 .36911 .34091 .37007 .32058 .3775 Metetra
.513 .903 .941 r
.32058 .3775 .34091 .37007 .325 .37115 .30456 .37878 Metetra
.514 .905 .938 r
.30456 .37878 .325 .37115 .3088 .3724 .28826 .38023 Metetra
.515 .906 .935 r
.28826 .38023 .3088 .3724 .2923 .37384 .27166 .3819 Metetra
.516 .908 .931 r
.27166 .3819 .2923 .37384 .2755 .37551 .25475 .38381 Metetra
.516 .909 .925 r
.25475 .38381 .2755 .37551 .25838 .37745 .23752 .38601 Metetra
.515 .91 .919 r
.23752 .38601 .25838 .37745 .24094 .3797 .21996 .38856 Metetra
.514 .911 .91 r
.21996 .38856 .24094 .3797 .22315 .38234 .20205 .39154 Metetra
.512 .912 .9 r
.20205 .39154 .22315 .38234 .205 .38545 .18379 .39503 Metetra
.507 .911 .886 r
.18379 .39503 .205 .38545 .18648 .38912 .16514 .39915 Metetra
.501 .909 .866 r
.16514 .39915 .18648 .38912 .16757 .39351 .1461 .40407 Metetra
.49 .903 .839 r
.1461 .40407 .16757 .39351 .14825 .39879 .12664 .41002 Metetra
.473 .891 .799 r
.12664 .41002 .14825 .39879 .12848 .40526 .10672 .41735 Metetra
.441 .864 .734 r
.10672 .41735 .12848 .40526 .10824 .41339 .08629 .42668 Metetra
.378 .803 .619 r
.08629 .42668 .10824 .41339 .08745 .42397 .06528 .43916 Metetra
.06528 .43916 .08745 .42397 .066 .43879 .04349 .45779 Metetra
.297 .039 .343 r
.04349 .45779 .066 .43879 .04354 .46382 .0192 .51377 Metetra
.588 .45 .662 r
.45499 .94416 .47595 .94249 .4613 .93995 .44488 .36489 Metetra
.894 .687 .616 r
.44488 .36489 .4613 .93995 .4507 .35849 .43077 .36561 Metetra
.532 .906 .956 r
.43077 .36561 .4507 .35849 .43645 .35919 .41642 .36622 Metetra
.523 .903 .954 r
.41642 .36622 .43645 .35919 .42196 .35977 .40183 .36684 Metetra
.521 .902 .952 r
.40183 .36684 .42196 .35977 .40722 .36034 .38699 .36752 Metetra
.521 .903 .951 r
.38699 .36752 .40722 .36034 .39222 .36096 .3719 .36827 Metetra
.522 .904 .95 r
.3719 .36827 .39222 .36096 .37697 .36165 .35654 .36911 Metetra
.523 .905 .949 r
.35654 .36911 .37697 .36165 .36144 .36244 .34091 .37007 Metetra
.525 .907 .947 r
.34091 .37007 .36144 .36244 .34564 .36334 .325 .37115 Metetra
.526 .908 .944 r
.325 .37115 .34564 .36334 .32955 .36439 .3088 .3724 Metetra
.527 .91 .941 r
.3088 .3724 .32955 .36439 .31316 .36562 .2923 .37384 Metetra
.527 .911 .937 r
.2923 .37384 .31316 .36562 .29647 .36704 .2755 .37551 Metetra
.528 .913 .933 r
.2755 .37551 .29647 .36704 .27946 .36872 .25838 .37745 Metetra
.528 .914 .927 r
.25838 .37745 .27946 .36872 .26213 .37069 .24094 .3797 Metetra
.527 .916 .92 r
.24094 .3797 .26213 .37069 .24446 .373 .22315 .38234 Metetra
.526 .917 .911 r
.22315 .38234 .24446 .373 .22643 .37573 .205 .38545 Metetra
.524 .917 .9 r
.205 .38545 .22643 .37573 .20804 .37898 .18648 .38912 Metetra
.519 .917 .884 r
.18648 .38912 .20804 .37898 .18926 .38285 .16757 .39351 Metetra
.512 .914 .863 r
.16757 .39351 .18926 .38285 .17008 .38752 .14825 .39879 Metetra
.501 .907 .832 r
.14825 .39879 .17008 .38752 .15047 .39322 .12848 .40526 Metetra
.48 .892 .784 r
.12848 .40526 .15047 .39322 .13039 .40029 .10824 .41339 Metetra
.44 .857 .704 r
.10824 .41339 .13039 .40029 .1098 .40934 .08745 .42397 Metetra
.08745 .42397 .1098 .40934 .08862 .4215 .066 .43879 Metetra
.066 .43879 .08862 .4215 .06668 .43964 .04354 .46382 Metetra
.326 .778 .953 r
.47595 .94249 .49712 .9408 .48257 .93824 .4613 .93995 Metetra
.586 .449 .661 r
.4613 .93995 .48257 .93824 .47082 .35032 .4507 .35849 Metetra
.577 .922 .96 r
.4507 .35849 .47082 .35032 .45667 .3517 .43645 .35919 Metetra
.545 .91 .958 r
.43645 .35919 .45667 .3517 .44228 .35236 .42196 .35977 Metetra
.538 .908 .957 r
.42196 .35977 .44228 .35236 .42765 .35289 .40722 .36034 Metetra
.535 .907 .956 r
.40722 .36034 .42765 .35289 .41276 .35341 .39222 .36096 Metetra
.535 .908 .954 r
.39222 .36096 .41276 .35341 .3976 .35397 .37697 .36165 Metetra
.535 .909 .953 r
.37697 .36165 .3976 .35397 .38218 .3546 .36144 .36244 Metetra
.536 .91 .951 r
.36144 .36244 .38218 .3546 .36649 .35533 .34564 .36334 Metetra
.537 .911 .949 r
.34564 .36334 .36649 .35533 .35051 .35619 .32955 .36439 Metetra
.538 .913 .947 r
.32955 .36439 .35051 .35619 .33423 .3572 .31316 .36562 Metetra
.539 .914 .943 r
.31316 .36562 .33423 .3572 .31765 .3584 .29647 .36704 Metetra
.54 .916 .94 r
.29647 .36704 .31765 .3584 .30076 .35982 .27946 .36872 Metetra
.54 .918 .935 r
.27946 .36872 .30076 .35982 .28354 .3615 .26213 .37069 Metetra
.54 .92 .929 r
.26213 .37069 .28354 .3615 .26599 .3635 .24446 .373 Metetra
.54 .921 .921 r
.24446 .373 .26599 .3635 .24809 .36587 .22643 .37573 Metetra
.538 .923 .912 r
.22643 .37573 .24809 .36587 .22982 .3687 .20804 .37898 Metetra
.536 .923 .899 r
.20804 .37898 .22982 .3687 .21118 .3721 .18926 .38285 Metetra
.532 .923 .882 r
.18926 .38285 .21118 .3721 .19214 .37619 .17008 .38752 Metetra
.524 .92 .859 r
.17008 .38752 .19214 .37619 .17267 .38116 .15047 .39322 Metetra
.511 .912 .823 r
.15047 .39322 .17267 .38116 .15276 .38731 .13039 .40029 Metetra
.486 .892 .767 r
.13039 .40029 .15276 .38731 .13236 .39507 .1098 .40934 Metetra
.434 .844 .665 r
.1098 .40934 .13236 .39507 .11141 .40522 .08862 .4215 Metetra
.08862 .4215 .11141 .40522 .08982 .4194 .06668 .43964 Metetra
.326 .778 .953 r
.49712 .9408 .51852 .9391 .50406 .93651 .48257 .93824 Metetra
.586 .449 .661 r
.48257 .93824 .50406 .93651 .48937 .93388 .47082 .35032 Metetra
.898 .689 .612 r
.47082 .35032 .48937 .93388 .47709 .34341 .45667 .3517 Metetra
.575 .922 .96 r
.45667 .3517 .47709 .34341 .46281 .34449 .44228 .35236 Metetra
.557 .914 .96 r
.44228 .35236 .46281 .34449 .44828 .34508 .42765 .35289 Metetra
.551 .912 .959 r
.42765 .35289 .44828 .34508 .43349 .34556 .41276 .35341 Metetra
.548 .912 .958 r
.41276 .35341 .43349 .34556 .41845 .34602 .3976 .35397 Metetra
.548 .912 .957 r
.3976 .35397 .41845 .34602 .40314 .34652 .38218 .3546 Metetra
.548 .913 .955 r
.38218 .3546 .40314 .34652 .38755 .3471 .36649 .35533 Metetra
.549 .914 .954 r
.36649 .35533 .38755 .3471 .37168 .34777 .35051 .35619 Metetra
.549 .916 .951 r
.35051 .35619 .37168 .34777 .35552 .34858 .33423 .3572 Metetra
.55 .917 .949 r
.33423 .3572 .35552 .34858 .33905 .34956 .31765 .3584 Metetra
.551 .919 .945 r
.31765 .3584 .33905 .34956 .32228 .35073 .30076 .35982 Metetra
.552 .921 .941 r
.30076 .35982 .32228 .35073 .30518 .35213 .28354 .3615 Metetra
.552 .923 .936 r
.28354 .3615 .30518 .35213 .28775 .35382 .26599 .3635 Metetra
.553 .925 .93 r
.26599 .3635 .28775 .35382 .26998 .35584 .24809 .36587 Metetra
.552 .927 .922 r
.24809 .36587 .26998 .35584 .25184 .35828 .22982 .3687 Metetra
.551 .928 .912 r
.22982 .3687 .25184 .35828 .23333 .36121 .21118 .3721 Metetra
.549 .929 .899 r
.21118 .3721 .23333 .36121 .21443 .36476 .19214 .37619 Metetra
.545 .929 .88 r
.19214 .37619 .21443 .36476 .19511 .36908 .17267 .38116 Metetra
.537 .925 .854 r
.17267 .38116 .19511 .36908 .17536 .37439 .15276 .38731 Metetra
.522 .916 .813 r
.15276 .38731 .17536 .37439 .15514 .38104 .13236 .39507 Metetra
.491 .891 .745 r
.13236 .39507 .15514 .38104 .1344 .38957 .11141 .40522 Metetra
.42 .822 .612 r
.11141 .40522 .1344 .38957 .11308 .40103 .08982 .4194 Metetra
.326 .778 .953 r
.51852 .9391 .54015 .93738 .52579 .93476 .50406 .93651 Metetra
.326 .778 .953 r
.50406 .93651 .52579 .93476 .51119 .9321 .48937 .93388 Metetra
.585 .447 .66 r
.48937 .93388 .51119 .9321 .49771 .33305 .47709 .34341 Metetra
.644 .944 .957 r
.47709 .34341 .49771 .33305 .48353 .33592 .46281 .34449 Metetra
.581 .923 .961 r
.46281 .34449 .48353 .33592 .46911 .33683 .44828 .34508 Metetra
.568 .918 .961 r
.44828 .34508 .46911 .33683 .45443 .33736 .43349 .34556 Metetra
.563 .916 .961 r
.43349 .34556 .45443 .33736 .4395 .33778 .41845 .34602 Metetra
.561 .916 .96 r
.41845 .34602 .4395 .33778 .4243 .33818 .40314 .34652 Metetra
.56 .916 .959 r
.40314 .34652 .4243 .33818 .40883 .33861 .38755 .3471 Metetra
.56 .917 .957 r
.38755 .3471 .40883 .33861 .39307 .33912 .37168 .34777 Metetra
.561 .918 .956 r
.37168 .34777 .39307 .33912 .37703 .33974 .35552 .34858 Metetra
.561 .92 .953 r
.35552 .34858 .37703 .33974 .36068 .3405 .33905 .34956 Metetra
.562 .921 .95 r
.33905 .34956 .36068 .3405 .34402 .34142 .32228 .35073 Metetra
.563 .923 .947 r
.32228 .35073 .34402 .34142 .32705 .34256 .30518 .35213 Metetra
.564 .925 .943 r
.30518 .35213 .32705 .34256 .30974 .34395 .28775 .35382 Metetra
.565 .928 .938 r
.28775 .35382 .30974 .34395 .29209 .34564 .26998 .35584 Metetra
.565 .93 .931 r
.26998 .35584 .29209 .34564 .27409 .34769 .25184 .35828 Metetra
.565 .932 .923 r
.25184 .35828 .27409 .34769 .25571 .35019 .23333 .36121 Metetra
.565 .934 .912 r
.23333 .36121 .25571 .35019 .23695 .35322 .21443 .36476 Metetra
.563 .935 .898 r
.21443 .36476 .23695 .35322 .21778 .35692 .19511 .36908 Metetra
.558 .935 .878 r
.19511 .36908 .21778 .35692 .19819 .36148 .17536 .37439 Metetra
.55 .931 .849 r
.17536 .37439 .19819 .36148 .17814 .36715 .15514 .38104 Metetra
.533 .919 .802 r
.15514 .38104 .17814 .36715 .1576 .37434 .1344 .38957 Metetra
.495 .886 .718 r
.1344 .38957 .1576 .37434 .13651 .38375 .11308 .40103 Metetra
.326 .778 .953 r
.54015 .93738 .56201 .93564 .54775 .93299 .52579 .93476 Metetra
.326 .778 .953 r
.52579 .93476 .54775 .93299 .53325 .9303 .51119 .9321 Metetra
.585 .447 .66 r
.51119 .9321 .53325 .9303 .51852 .92757 .49771 .33305 Metetra
.902 .691 .608 r
.49771 .33305 .51852 .92757 .50446 .32601 .48353 .33592 Metetra
.623 .938 .958 r
.48353 .33592 .50446 .32601 .49015 .32792 .46911 .33683 Metetra
.589 .926 .962 r
.46911 .33683 .49015 .32792 .47559 .3287 .45443 .33736 Metetra
.579 .921 .962 r
.45443 .33736 .47559 .3287 .46077 .32916 .4395 .33778 Metetra
.575 .92 .962 r
.4395 .33778 .46077 .32916 .44568 .3295 .4243 .33818 Metetra
.573 .92 .962 r
.4243 .33818 .44568 .3295 .43032 .32983 .40883 .33861 Metetra
.572 .92 .961 r
.40883 .33861 .43032 .32983 .41468 .3302 .39307 .33912 Metetra
.572 .921 .959 r
.39307 .33912 .41468 .3302 .39876 .33064 .37703 .33974 Metetra
.573 .922 .957 r
.37703 .33974 .39876 .33064 .38253 .3312 .36068 .3405 Metetra
.573 .924 .955 r
.36068 .3405 .38253 .3312 .366 .33189 .34402 .34142 Metetra
.574 .926 .952 r
.34402 .34142 .366 .33189 .34915 .33277 .32705 .34256 Metetra
.575 .928 .948 r
.32705 .34256 .34915 .33277 .33197 .33387 .30974 .34395 Metetra
.577 .93 .944 r
.30974 .34395 .33197 .33387 .31445 .33523 .29209 .34564 Metetra
.578 .932 .939 r
.29209 .34564 .31445 .33523 .29658 .33692 .27409 .34769 Metetra
.578 .935 .932 r
.27409 .34769 .29658 .33692 .27834 .33899 .25571 .35019 Metetra
.579 .937 .924 r
.25571 .35019 .27834 .33899 .25972 .34154 .23695 .35322 Metetra
.578 .94 .912 r
.23695 .35322 .25972 .34154 .2407 .34467 .21778 .35692 Metetra
.577 .941 .897 r
.21778 .35692 .2407 .34467 .22126 .34853 .19819 .36148 Metetra
.573 .941 .876 r
.19819 .36148 .22126 .34853 .20137 .35333 .17814 .36715 Metetra
.564 .937 .843 r
.17814 .36715 .20137 .35333 .18102 .35937 .1576 .37434 Metetra
.544 .922 .789 r
.1576 .37434 .18102 .35937 .16015 .36715 .13651 .38375 Metetra
.326 .778 .953 r
.56201 .93564 .5841 .93388 .56995 .9312 .54775 .93299 Metetra
.326 .778 .953 r
.54775 .93299 .56995 .9312 .55556 .92849 .53325 .9303 Metetra
.326 .778 .953 r
.53325 .9303 .55556 .92849 .54093 .92572 .51852 .92757 Metetra
.583 .446 .66 r
.51852 .92757 .54093 .92572 .52605 .92291 .50446 .32601 Metetra
.905 .692 .605 r
.50446 .32601 .52605 .92291 .51139 .31789 .49015 .32792 Metetra
.622 .938 .959 r
.49015 .32792 .51139 .31789 .49695 .31939 .47559 .3287 Metetra
.599 .928 .962 r
.47559 .3287 .49695 .31939 .48225 .32006 .46077 .32916 Metetra
.59 .925 .964 r
.46077 .32916 .48225 .32006 .46728 .32044 .44568 .3295 Metetra
.587 .923 .964 r
.44568 .3295 .46728 .32044 .45204 .32072 .43032 .32983 Metetra
.585 .923 .963 r
.43032 .32983 .45204 .32072 .43652 .32097 .41468 .3302 Metetra
.584 .924 .962 r
.41468 .3302 .43652 .32097 .42071 .32126 .39876 .33064 Metetra
.584 .925 .96 r
.39876 .33064 .42071 .32126 .40461 .32163 .38253 .3312 Metetra
.585 .926 .958 r
.38253 .3312 .40461 .32163 .3882 .32211 .366 .33189 Metetra
.586 .928 .956 r
.366 .33189 .3882 .32211 .37148 .32274 .34915 .33277 Metetra
.587 .93 .953 r
.34915 .33277 .37148 .32274 .35443 .32356 .33197 .33387 Metetra
.588 .932 .95 r
.33197 .33387 .35443 .32356 .33704 .32461 .31445 .33523 Metetra
.589 .934 .945 r
.31445 .33523 .33704 .32461 .3193 .32594 .29658 .33692 Metetra
.591 .937 .94 r
.29658 .33692 .3193 .32594 .3012 .32761 .27834 .33899 Metetra
.592 .94 .933 r
.27834 .33899 .3012 .32761 .28272 .3297 .25972 .34154 Metetra
.593 .943 .924 r
.25972 .34154 .28272 .3297 .26385 .33228 .2407 .34467 Metetra
.593 .945 .912 r
.2407 .34467 .26385 .33228 .24457 .3355 .22126 .34853 Metetra
.592 .947 .896 r
.22126 .34853 .24457 .3355 .22485 .33951 .20137 .35333 Metetra
.588 .947 .873 r
.20137 .35333 .22485 .33951 .20468 .34455 .18102 .35937 Metetra
.579 .942 .837 r
.18102 .35937 .20468 .34455 .18401 .35098 .16015 .36715 Metetra
.326 .778 .953 r
.5841 .93388 .60644 .9321 .59239 .9294 .56995 .9312 Metetra
.326 .778 .953 r
.56995 .9312 .59239 .9294 .57811 .92665 .55556 .92849 Metetra
.326 .778 .953 r
.55556 .92849 .57811 .92665 .56358 .92385 .54093 .92572 Metetra
.326 .778 .953 r
.54093 .92572 .56358 .92385 .54881 .92101 .52605 .92291 Metetra
.581 .444 .659 r
.52605 .92291 .54881 .92101 .53283 .30475 .51139 .31789 Metetra
.702 .96 .945 r
.51139 .31789 .53283 .30475 .51852 .30906 .49695 .31939 Metetra
.627 .938 .96 r
.49695 .31939 .51852 .30906 .50394 .31032 .48225 .32006 Metetra
.609 .931 .963 r
.48225 .32006 .50394 .31032 .48909 .31089 .46728 .32044 Metetra
.602 .928 .964 r
.46728 .32044 .48909 .31089 .47398 .31118 .45204 .32072 Metetra
.598 .927 .964 r
.45204 .32072 .47398 .31118 .45858 .31138 .43652 .32097 Metetra
.597 .926 .964 r
.43652 .32097 .45858 .31138 .4429 .31155 .42071 .32126 Metetra
.596 .927 .963 r
.42071 .32126 .4429 .31155 .42692 .31175 .40461 .32163 Metetra
.596 .928 .961 r
.40461 .32163 .42692 .31175 .41064 .31204 .3882 .32211 Metetra
.597 .93 .959 r
.3882 .32211 .41064 .31204 .39405 .31244 .37148 .32274 Metetra
.598 .931 .957 r
.37148 .32274 .39405 .31244 .37713 .31299 .35443 .32356 Metetra
.599 .933 .954 r
.35443 .32356 .37713 .31299 .35987 .31374 .33704 .32461 Metetra
.601 .936 .951 r
.33704 .32461 .35987 .31374 .34228 .31473 .3193 .32594 Metetra
.602 .939 .946 r
.3193 .32594 .34228 .31473 .32432 .31602 .3012 .32761 Metetra
.604 .941 .941 r
.3012 .32761 .32432 .31602 .30598 .31766 .28272 .3297 Metetra
.605 .945 .934 r
.28272 .3297 .30598 .31766 .28726 .31974 .26385 .33228 Metetra
.607 .948 .925 r
.26385 .33228 .28726 .31974 .26813 .32236 .24457 .3355 Metetra
.608 .951 .912 r
.24457 .3355 .26813 .32236 .24858 .32565 .22485 .33951 Metetra
.607 .953 .895 r
.22485 .33951 .24858 .32565 .22858 .3298 .20468 .34455 Metetra
.604 .953 .87 r
.20468 .34455 .22858 .3298 .20811 .33507 .18401 .35098 Metetra
.326 .778 .953 r
.60644 .9321 .62902 .9303 .61508 .92757 .59239 .9294 Metetra
.326 .778 .953 r
.59239 .9294 .61508 .92757 .60091 .92479 .57811 .92665 Metetra
.326 .778 .953 r
.57811 .92665 .60091 .92479 .58649 .92196 .56358 .92385 Metetra
.326 .778 .953 r
.56358 .92385 .58649 .92196 .57184 .91909 .54881 .92101 Metetra
.581 .444 .659 r
.54881 .92101 .57184 .91909 .55693 .91617 .53283 .30475 Metetra
.908 .694 .6 r
.53283 .30475 .55693 .91617 .54029 .29648 .51852 .30906 Metetra
.683 .956 .949 r
.51852 .30906 .54029 .29648 .52584 .29958 .50394 .31032 Metetra
.635 .94 .96 r
.50394 .31032 .52584 .29958 .51113 .30066 .48909 .31089 Metetra
.62 .934 .964 r
.48909 .31089 .51113 .30066 .49614 .30113 .47398 .31118 Metetra
.614 .931 .965 r
.47398 .31118 .49614 .30113 .48087 .30134 .45858 .31138 Metetra
.61 .93 .965 r
.45858 .31138 .48087 .30134 .46531 .30144 .4429 .31155 Metetra
.608 .93 .965 r
.4429 .31155 .46531 .30144 .44947 .30152 .42692 .31175 Metetra
.608 .93 .964 r
.42692 .31175 .44947 .30152 .43332 .30163 .41064 .31204 Metetra
.608 .931 .962 r
.41064 .31204 .43332 .30163 .41685 .30182 .39405 .31244 Metetra
.609 .933 .96 r
.39405 .31244 .41685 .30182 .40007 .30213 .37713 .31299 Metetra
.61 .935 .958 r
.37713 .31299 .40007 .30213 .38296 .3026 .35987 .31374 Metetra
.612 .937 .955 r
.35987 .31374 .38296 .3026 .3655 .30327 .34228 .31473 Metetra
.613 .94 .951 r
.34228 .31473 .3655 .30327 .34768 .30419 .32432 .31602 Metetra
.615 .943 .947 r
.32432 .31602 .34768 .30419 .32949 .30541 .30598 .31766 Metetra
.617 .946 .941 r
.30598 .31766 .32949 .30541 .31092 .30701 .28726 .31974 Metetra
.62 .949 .934 r
.28726 .31974 .31092 .30701 .29195 .30907 .26813 .32236 Metetra
.622 .953 .925 r
.26813 .32236 .29195 .30907 .27257 .3117 .24858 .32565 Metetra
.623 .956 .912 r
.24858 .32565 .27257 .3117 .25274 .31505 .22858 .3298 Metetra
.624 .959 .895 r
.22858 .3298 .25274 .31505 .23245 .31931 .20811 .33507 Metetra
.326 .778 .953 r
.62902 .9303 .65185 .92849 .63802 .92572 .61508 .92757 Metetra
.326 .778 .953 r
.61508 .92757 .63802 .92572 .62396 .92291 .60091 .92479 Metetra
.326 .778 .953 r
.60091 .92479 .62396 .92291 .60966 .92005 .58649 .92196 Metetra
.326 .778 .953 r
.58649 .92196 .60966 .92005 .59512 .91715 .57184 .91909 Metetra
.326 .778 .953 r
.57184 .91909 .59512 .91715 .58032 .91419 .55693 .91617 Metetra
.58 .442 .658 r
.55693 .91617 .58032 .91419 .56528 .91118 .54029 .29648 Metetra
.911 .695 .597 r
.54029 .29648 .56528 .91118 .54796 .28687 .52584 .29958 Metetra
.681 .955 .95 r
.52584 .29958 .54796 .28687 .53338 .28942 .51113 .30066 Metetra
.645 .942 .96 r
.51113 .30066 .53338 .28942 .51852 .29037 .49614 .30113 Metetra
.632 .936 .964 r
.49614 .30113 .51852 .29037 .50338 .29074 .48087 .30134 Metetra
.625 .934 .965 r
.48087 .30134 .50338 .29074 .48796 .29086 .46531 .30144 Metetra
.622 .933 .965 r
.46531 .30144 .48796 .29086 .47225 .29087 .44947 .30152 Metetra
.621 .933 .965 r
.44947 .30152 .47225 .29087 .45623 .29085 .43332 .30163 Metetra
.62 .933 .964 r
.43332 .30163 .45623 .29085 .43991 .29086 .41685 .30182 Metetra
.62 .934 .963 r
.41685 .30182 .43991 .29086 .42326 .29094 .40007 .30213 Metetra
.621 .936 .961 r
.40007 .30213 .42326 .29094 .40628 .29115 .38296 .3026 Metetra
.623 .938 .958 r
.38296 .3026 .40628 .29115 .38897 .29151 .3655 .30327 Metetra
.624 .94 .956 r
.3655 .30327 .38897 .29151 .3713 .29209 .34768 .30419 Metetra
.626 .943 .952 r
.34768 .30419 .3713 .29209 .35326 .29291 .32949 .30541 Metetra
.629 .946 .947 r
.32949 .30541 .35326 .29291 .33484 .29406 .31092 .30701 Metetra
.631 .95 .942 r
.31092 .30701 .33484 .29406 .31603 .2956 .29195 .30907 Metetra
.634 .953 .935 r
.29195 .30907 .31603 .2956 .29681 .29761 .27257 .3117 Metetra
.637 .957 .925 r
.27257 .3117 .29681 .29761 .27715 .30023 .25274 .31505 Metetra
.639 .961 .912 r
.25274 .31505 .27715 .30023 .25705 .3036 .23245 .31931 Metetra
.326 .778 .953 r
.65185 .92849 .67493 .92665 .66122 .92385 .63802 .92572 Metetra
.326 .778 .953 r
.63802 .92572 .66122 .92385 .64727 .92101 .62396 .92291 Metetra
.326 .778 .953 r
.62396 .92291 .64727 .92101 .63309 .91812 .60966 .92005 Metetra
.326 .778 .953 r
.60966 .92005 .63309 .91812 .61867 .91518 .59512 .91715 Metetra
.326 .778 .953 r
.59512 .91715 .61867 .91518 .60399 .91219 .58032 .91419 Metetra
.326 .778 .953 r
.58032 .91419 .60399 .91219 .58906 .90915 .56528 .91118 Metetra
.578 .441 .658 r
.56528 .91118 .58906 .90915 .57387 .90605 .54796 .28687 Metetra
.914 .697 .594 r
.54796 .28687 .57387 .90605 .55584 .2763 .53338 .28942 Metetra
.687 .955 .95 r
.53338 .28942 .55584 .2763 .54112 .27856 .51852 .29037 Metetra
.656 .944 .96 r
.51852 .29037 .54112 .27856 .52612 .27939 .50338 .29074 Metetra
.644 .939 .963 r
.50338 .29074 .52612 .27939 .51084 .27968 .48796 .29086 Metetra
.638 .936 .965 r
.48796 .29086 .51084 .27968 .49526 .2797 .47225 .29087 Metetra
.634 .935 .965 r
.47225 .29087 .49526 .2797 .47939 .27961 .45623 .29085 Metetra
.633 .935 .965 r
.45623 .29085 .47939 .27961 .4632 .27948 .43991 .29086 Metetra
.632 .936 .964 r
.43991 .29086 .4632 .27948 .4467 .27937 .42326 .29094 Metetra
.633 .937 .963 r
.42326 .29094 .4467 .27937 .42987 .27934 .40628 .29115 Metetra
.634 .939 .961 r
.40628 .29115 .42987 .27934 .4127 .27942 .38897 .29151 Metetra
.635 .941 .959 r
.38897 .29151 .4127 .27942 .39517 .27967 .3713 .29209 Metetra
.637 .944 .956 r
.3713 .29209 .39517 .27967 .37729 .28013 .35326 .29291 Metetra
.639 .946 .952 r
.35326 .29291 .37729 .28013 .35903 .28085 .33484 .29406 Metetra
.642 .95 .948 r
.33484 .29406 .35903 .28085 .34038 .28189 .31603 .2956 Metetra
.645 .953 .942 r
.31603 .2956 .34038 .28189 .32132 .28334 .29681 .29761 Metetra
.648 .957 .935 r
.29681 .29761 .32132 .28334 .30184 .28528 .27715 .30023 Metetra
.652 .962 .925 r
.27715 .30023 .30184 .28528 .28191 .28785 .25705 .3036 Metetra
.326 .778 .953 r
.67493 .92665 .69827 .92479 .68468 .92196 .66122 .92385 Metetra
.326 .778 .953 r
.66122 .92385 .68468 .92196 .67085 .91909 .64727 .92101 Metetra
.326 .778 .953 r
.64727 .92101 .67085 .91909 .65679 .91617 .63309 .91812 Metetra
.326 .778 .953 r
.63309 .91812 .65679 .91617 .64249 .91319 .61867 .91518 Metetra
.326 .778 .953 r
.61867 .91518 .64249 .91319 .62793 .91017 .60399 .91219 Metetra
.326 .778 .953 r
.60399 .91219 .62793 .91017 .61313 .90709 .58906 .90915 Metetra
.326 .778 .953 r
.58906 .90915 .61313 .90709 .59806 .90396 .57387 .90605 Metetra
.576 .439 .657 r
.57387 .90605 .59806 .90396 .58272 .90077 .55584 .2763 Metetra
.917 .698 .591 r
.55584 .2763 .58272 .90077 .56394 .26484 .54112 .27856 Metetra
.696 .956 .95 r
.54112 .27856 .56394 .26484 .54909 .26693 .52612 .27939 Metetra
.668 .946 .959 r
.52612 .27939 .54909 .26693 .53395 .26768 .51084 .27968 Metetra
.656 .941 .962 r
.51084 .27968 .53395 .26768 .51852 .26788 .49526 .2797 Metetra
.65 .939 .964 r
.49526 .2797 .51852 .26788 .50278 .2678 .47939 .27961 Metetra
.647 .938 .965 r
.47939 .27961 .50278 .2678 .48674 .2676 .4632 .27948 Metetra
.645 .938 .964 r
.4632 .27948 .48674 .2676 .47039 .26735 .4467 .27937 Metetra
.645 .939 .964 r
.4467 .27937 .47039 .26735 .4537 .26711 .42987 .27934 Metetra
.645 .94 .962 r
.42987 .27934 .4537 .26711 .43668 .26695 .4127 .27942 Metetra
.646 .942 .961 r
.4127 .27942 .43668 .26695 .41931 .2669 .39517 .27967 Metetra
.648 .944 .959 r
.39517 .27967 .41931 .2669 .40158 .26701 .37729 .28013 Metetra
.65 .946 .956 r
.37729 .28013 .40158 .26701 .38348 .26733 .35903 .28085 Metetra
.653 .949 .952 r
.35903 .28085 .38348 .26733 .36499 .26791 .34038 .28189 Metetra
.656 .953 .948 r
.34038 .28189 .36499 .26791 .3461 .26883 .32132 .28334 Metetra
.659 .957 .942 r
.32132 .28334 .3461 .26883 .32679 .27016 .30184 .28528 Metetra
.663 .961 .935 r
.30184 .28528 .32679 .27016 .30704 .272 .28191 .28785 Metetra
.326 .778 .953 r
.69827 .92479 .72187 .92291 .7084 .92005 .68468 .92196 Metetra
.326 .778 .953 r
.68468 .92196 .7084 .92005 .6947 .91715 .67085 .91909 Metetra
.326 .778 .953 r
.67085 .91909 .6947 .91715 .68076 .91419 .65679 .91617 Metetra
.326 .778 .953 r
.65679 .91617 .68076 .91419 .66658 .91118 .64249 .91319 Metetra
.326 .778 .953 r
.64249 .91319 .66658 .91118 .65215 .90812 .62793 .91017 Metetra
.326 .778 .953 r
.62793 .91017 .65215 .90812 .63747 .90501 .61313 .90709 Metetra
.326 .778 .953 r
.61313 .90709 .63747 .90501 .62253 .90184 .59806 .90396 Metetra
.326 .778 .953 r
.59806 .90396 .62253 .90184 .60732 .89861 .58272 .90077 Metetra
.575 .437 .656 r
.58272 .90077 .60732 .89861 .58692 .24411 .56394 .26484 Metetra
.809 .967 .906 r
.56394 .26484 .58692 .24411 .57228 .25244 .54909 .26693 Metetra
.708 .958 .948 r
.54909 .26693 .57228 .25244 .55729 .25446 .53395 .26768 Metetra
.682 .949 .957 r
.53395 .26768 .55729 .25446 .54201 .25516 .51852 .26788 Metetra
.669 .944 .961 r
.51852 .26788 .54201 .25516 .52643 .25528 .50278 .2678 Metetra
.663 .941 .963 r
.50278 .2678 .52643 .25528 .51053 .2551 .48674 .2676 Metetra
.659 .94 .964 r
.48674 .2676 .51053 .2551 .49433 .25477 .47039 .26735 Metetra
.658 .94 .964 r
.47039 .26735 .49433 .25477 .47779 .25439 .4537 .26711 Metetra
.657 .941 .963 r
.4537 .26711 .47779 .25439 .46093 .25402 .43668 .26695 Metetra
.657 .942 .962 r
.43668 .26695 .46093 .25402 .44371 .2537 .41931 .2669 Metetra
.659 .944 .96 r
.41931 .2669 .44371 .2537 .42614 .2535 .40158 .26701 Metetra
.66 .946 .958 r
.40158 .26701 .42614 .2535 .4082 .25345 .38348 .26733 Metetra
.663 .949 .955 r
.38348 .26733 .4082 .25345 .38988 .25361 .36499 .26791 Metetra
.666 .952 .952 r
.36499 .26791 .38988 .25361 .37116 .25404 .3461 .26883 Metetra
.669 .955 .948 r
.3461 .26883 .37116 .25404 .35202 .25479 .32679 .27016 Metetra
.673 .959 .942 r
.32679 .27016 .35202 .25479 .33246 .25597 .30704 .272 Metetra
.326 .778 .953 r
.72187 .92291 .74573 .92101 .73239 .91812 .7084 .92005 Metetra
.326 .778 .953 r
.7084 .92005 .73239 .91812 .71881 .91518 .6947 .91715 Metetra
.326 .778 .953 r
.6947 .91715 .71881 .91518 .70501 .91219 .68076 .91419 Metetra
.326 .778 .953 r
.68076 .91419 .70501 .91219 .69096 .90915 .66658 .91118 Metetra
.326 .778 .953 r
.66658 .91118 .69096 .90915 .67666 .90605 .65215 .90812 Metetra
.326 .778 .953 r
.65215 .90812 .67666 .90605 .66211 .9029 .63747 .90501 Metetra
.326 .778 .953 r
.63747 .90501 .66211 .9029 .6473 .89969 .62253 .90184 Metetra
.326 .778 .953 r
.62253 .90184 .6473 .89969 .63222 .89643 .60732 .89861 Metetra
.575 .437 .656 r
.60732 .89861 .63222 .89643 .61686 .8931 .58692 .24411 Metetra
.919 .698 .587 r
.58692 .24411 .61686 .8931 .59562 .23091 .57228 .25244 Metetra
.814 .967 .903 r
.57228 .25244 .59562 .23091 .58085 .23902 .55729 .25446 Metetra
.722 .96 .944 r
.55729 .25446 .58085 .23902 .56573 .24108 .54201 .25516 Metetra
.696 .951 .955 r
.54201 .25516 .56573 .24108 .55031 .24175 .52643 .25528 Metetra
.683 .946 .959 r
.52643 .25528 .55031 .24175 .53457 .24179 .51053 .2551 Metetra
.676 .944 .962 r
.51053 .2551 .53457 .24179 .51852 .24151 .49433 .25477 Metetra
.672 .943 .962 r
.49433 .25477 .51852 .24151 .50214 .24107 .47779 .25439 Metetra
.67 .942 .963 r
.47779 .25439 .50214 .24107 .48543 .24054 .46093 .25402 Metetra
.67 .943 .962 r
.46093 .25402 .48543 .24054 .46838 .24001 .44371 .2537 Metetra
.67 .944 .961 r
.44371 .2537 .46838 .24001 .45097 .23953 .42614 .2535 Metetra
.671 .946 .959 r
.42614 .2535 .45097 .23953 .4332 .23915 .4082 .25345 Metetra
.673 .948 .957 r
.4082 .25345 .4332 .23915 .41504 .23892 .38988 .25361 Metetra
.676 .951 .955 r
.38988 .25361 .41504 .23892 .39649 .23889 .37116 .25404 Metetra
.679 .954 .951 r
.37116 .25404 .39649 .23889 .37753 .23913 .35202 .25479 Metetra
.683 .958 .947 r
.35202 .25479 .37753 .23913 .35815 .23969 .33246 .25597 Metetra
.326 .778 .953 r
.74573 .92101 .76987 .91909 .75665 .91617 .73239 .91812 Metetra
.326 .778 .953 r
.73239 .91812 .75665 .91617 .74321 .91319 .71881 .91518 Metetra
.326 .778 .953 r
.71881 .91518 .74321 .91319 .72953 .91017 .70501 .91219 Metetra
.326 .778 .953 r
.70501 .91219 .72953 .91017 .71562 .90709 .69096 .90915 Metetra
.326 .778 .953 r
.69096 .90915 .71562 .90709 .70145 .90396 .67666 .90605 Metetra
.326 .778 .953 r
.67666 .90605 .70145 .90396 .68704 .90077 .66211 .9029 Metetra
.326 .778 .953 r
.66211 .9029 .68704 .90077 .67236 .89752 .6473 .89969 Metetra
.326 .778 .953 r
.6473 .89969 .67236 .89752 .65742 .89422 .63222 .89643 Metetra
.326 .778 .953 r
.63222 .89643 .65742 .89422 .6422 .89085 .61686 .8931 Metetra
.573 .435 .655 r
.61686 .8931 .6422 .89085 .6267 .88742 .59562 .23091 Metetra
.922 .7 .584 r
.59562 .23091 .6267 .88742 .60455 .21527 .58085 .23902 Metetra
.833 .964 .891 r
.58085 .23902 .60455 .21527 .58967 .22446 .56573 .24108 Metetra
.739 .962 .94 r
.56573 .24108 .58967 .22446 .57442 .22667 .55031 .24175 Metetra
.711 .953 .951 r
.55031 .24175 .57442 .22667 .55885 .22735 .53457 .24179 Metetra
.697 .948 .957 r
.53457 .24179 .55885 .22735 .54296 .22734 .51852 .24151 Metetra
.69 .946 .959 r
.51852 .24151 .54296 .22734 .52675 .22697 .50214 .24107 Metetra
.685 .945 .961 r
.50214 .24107 .52675 .22697 .5102 .22639 .48543 .24054 Metetra
.683 .944 .961 r
.48543 .24054 .5102 .22639 .49331 .22571 .46838 .24001 Metetra
.683 .945 .96 r
.46838 .24001 .49331 .22571 .47607 .22501 .45097 .23953 Metetra
.683 .946 .96 r
.45097 .23953 .47607 .22501 .45847 .22435 .4332 .23915 Metetra
.684 .948 .958 r
.4332 .23915 .45847 .22435 .44048 .22377 .41504 .23892 Metetra
.686 .95 .956 r
.41504 .23892 .44048 .22377 .42211 .22333 .39649 .23889 Metetra
.689 .952 .954 r
.39649 .23889 .42211 .22333 .40333 .22309 .37753 .23913 Metetra
.692 .956 .95 r
.37753 .23913 .40333 .22309 .38414 .2231 .35815 .23969 Metetra
.326 .778 .953 r
.76987 .91909 .79428 .91715 .78119 .91419 .75665 .91617 Metetra
.326 .778 .953 r
.75665 .91617 .78119 .91419 .76789 .91118 .74321 .91319 Metetra
.326 .778 .953 r
.74321 .91319 .76789 .91118 .75435 .90812 .72953 .91017 Metetra
.326 .778 .953 r
.72953 .91017 .75435 .90812 .74057 .90501 .71562 .90709 Metetra
.326 .778 .953 r
.71562 .90709 .74057 .90501 .72654 .90184 .70145 .90396 Metetra
.326 .778 .953 r
.70145 .90396 .72654 .90184 .71226 .89861 .68704 .90077 Metetra
.326 .778 .953 r
.68704 .90077 .71226 .89861 .69773 .89533 .67236 .89752 Metetra
.326 .778 .953 r
.67236 .89752 .69773 .89533 .68293 .89198 .65742 .89422 Metetra
.326 .778 .953 r
.65742 .89422 .68293 .89198 .66785 .88857 .6422 .89085 Metetra
.326 .778 .953 r
.6422 .89085 .66785 .88857 .6525 .8851 .6267 .88742 Metetra
.571 .433 .654 r
.6267 .88742 .6525 .8851 .63685 .88157 .60455 .21527 Metetra
.925 .701 .58 r
.60455 .21527 .63685 .88157 .62091 .87797 .58967 .22446 Metetra
.928 .703 .577 r
.58967 .22446 .62091 .87797 .59875 .20855 .57442 .22667 Metetra
.758 .963 .933 r
.57442 .22667 .59875 .20855 .58337 .2111 .55885 .22735 Metetra
.727 .955 .947 r
.55885 .22735 .58337 .2111 .56766 .21186 .54296 .22734 Metetra
.712 .951 .953 r
.54296 .22734 .56766 .21186 .55161 .21182 .52675 .22697 Metetra
.704 .948 .956 r
.52675 .22697 .55161 .21182 .53524 .21135 .5102 .22639 Metetra
.699 .946 .958 r
.5102 .22639 .53524 .21135 .51852 .21065 .49331 .22571 Metetra
.697 .946 .959 r
.49331 .22571 .51852 .21065 .50145 .20981 .47607 .22501 Metetra
.696 .946 .958 r
.47607 .22501 .50145 .20981 .48402 .20893 .45847 .22435 Metetra
.696 .947 .958 r
.45847 .22435 .48402 .20893 .46621 .20806 .44048 .22377 Metetra
.697 .949 .956 r
.44048 .22377 .46621 .20806 .44802 .20726 .42211 .22333 Metetra
.699 .951 .955 r
.42211 .22333 .44802 .20726 .42942 .20659 .40333 .22309 Metetra
.702 .954 .952 r
.40333 .22309 .42942 .20659 .41041 .20609 .38414 .2231 Metetra
.326 .778 .953 r
.79428 .91715 .81896 .91518 .80602 .91219 .78119 .91419 Metetra
.326 .778 .953 r
.78119 .91419 .80602 .91219 .79285 .90915 .76789 .91118 Metetra
.326 .778 .953 r
.76789 .91118 .79285 .90915 .77945 .90605 .75435 .90812 Metetra
.326 .778 .953 r
.75435 .90812 .77945 .90605 .76581 .9029 .74057 .90501 Metetra
.326 .778 .953 r
.74057 .90501 .76581 .9029 .75193 .89969 .72654 .90184 Metetra
.326 .778 .953 r
.72654 .90184 .75193 .89969 .7378 .89643 .71226 .89861 Metetra
.326 .778 .953 r
.71226 .89861 .7378 .89643 .7234 .8931 .69773 .89533 Metetra
.326 .778 .953 r
.69773 .89533 .7234 .8931 .70875 .88972 .68293 .89198 Metetra
.326 .778 .953 r
.68293 .89198 .70875 .88972 .69382 .88627 .66785 .88857 Metetra
.326 .778 .953 r
.66785 .88857 .69382 .88627 .67861 .88275 .6525 .8851 Metetra
.326 .778 .953 r
.6525 .8851 .67861 .88275 .66311 .87917 .63685 .88157 Metetra
.326 .778 .953 r
.63685 .88157 .66311 .87917 .64732 .87552 .62091 .87797 Metetra
.567 .429 .653 r
.62091 .87797 .64732 .87552 .63123 .87181 .59875 .20855 Metetra
.931 .704 .574 r
.59875 .20855 .63123 .87181 .60809 .19099 .58337 .2111 Metetra
.781 .964 .923 r
.58337 .2111 .60809 .19099 .59258 .19419 .56766 .21186 Metetra
.746 .957 .941 r
.56766 .21186 .59258 .19419 .57672 .19513 .55161 .21182 Metetra
.728 .952 .949 r
.55161 .21182 .57672 .19513 .56053 .1951 .53524 .21135 Metetra
.719 .949 .953 r
.53524 .21135 .56053 .1951 .54399 .19456 .51852 .21065 Metetra
.713 .948 .955 r
.51852 .21065 .54399 .19456 .5271 .19373 .50145 .20981 Metetra
.71 .947 .956 r
.50145 .20981 .5271 .19373 .50985 .19273 .48402 .20893 Metetra
.709 .948 .956 r
.48402 .20893 .50985 .19273 .49222 .19165 .46621 .20806 Metetra
.709 .948 .955 r
.46621 .20806 .49222 .19165 .47421 .19056 .44802 .20726 Metetra
.71 .95 .954 r
.44802 .20726 .47421 .19056 .45581 .18952 .42942 .20659 Metetra
.712 .952 .952 r
.42942 .20659 .45581 .18952 .43699 .18858 .41041 .20609 Metetra
.326 .778 .953 r
.81896 .91518 .84393 .91319 .83113 .91017 .80602 .91219 Metetra
.326 .778 .953 r
.80602 .91219 .83113 .91017 .81811 .90709 .79285 .90915 Metetra
.326 .778 .953 r
.79285 .90915 .81811 .90709 .80485 .90396 .77945 .90605 Metetra
.326 .778 .953 r
.77945 .90605 .80485 .90396 .79136 .90077 .76581 .9029 Metetra
.326 .778 .953 r
.76581 .9029 .79136 .90077 .77762 .89752 .75193 .89969 Metetra
.326 .778 .953 r
.75193 .89969 .77762 .89752 .76364 .89422 .7378 .89643 Metetra
.326 .778 .953 r
.7378 .89643 .76364 .89422 .74939 .89085 .7234 .8931 Metetra
.326 .778 .953 r
.7234 .8931 .74939 .89085 .73489 .88742 .70875 .88972 Metetra
.326 .778 .953 r
.70875 .88972 .73489 .88742 .72011 .88393 .69382 .88627 Metetra
.326 .778 .953 r
.69382 .88627 .72011 .88393 .70505 .88037 .67861 .88275 Metetra
.326 .778 .953 r
.67861 .88275 .70505 .88037 .68971 .87675 .66311 .87917 Metetra
.326 .778 .953 r
.66311 .87917 .68971 .87675 .67407 .87305 .64732 .87552 Metetra
.326 .778 .953 r
.64732 .87552 .67407 .87305 .65813 .86929 .63123 .87181 Metetra
.565 .427 .652 r
.63123 .87181 .65813 .86929 .64188 .86545 .60809 .19099 Metetra
.933 .705 .57 r
.60809 .19099 .64188 .86545 .61769 .17121 .59258 .19419 Metetra
.81 .964 .907 r
.59258 .19419 .61769 .17121 .60207 .17567 .57672 .19513 Metetra
.766 .959 .932 r
.57672 .19513 .60207 .17567 .58607 .17696 .56053 .1951 Metetra
.746 .954 .942 r
.56053 .1951 .58607 .17696 .56972 .17702 .54399 .19456 Metetra
.735 .951 .948 r
.54399 .19456 .56972 .17702 .55302 .17645 .5271 .19373 Metetra
.728 .949 .95 r
.5271 .19373 .55302 .17645 .53595 .1755 .50985 .19273 Metetra
.724 .948 .952 r
.50985 .19273 .53595 .1755 .51852 .17434 .49222 .19165 Metetra
.722 .948 .952 r
.49222 .19165 .51852 .17434 .5007 .17305 .47421 .19056 Metetra
.722 .949 .952 r
.47421 .19056 .5007 .17305 .48249 .17172 .45581 .18952 Metetra
.723 .95 .951 r
.45581 .18952 .48249 .17172 .46386 .17042 .43699 .18858 Metetra
.326 .778 .953 r
.84393 .91319 .86919 .91118 .85654 .90812 .83113 .91017 Metetra
.326 .778 .953 r
.83113 .91017 .85654 .90812 .84366 .90501 .81811 .90709 Metetra
.326 .778 .953 r
.81811 .90709 .84366 .90501 .83055 .90184 .80485 .90396 Metetra
.326 .778 .953 r
.80485 .90396 .83055 .90184 .81721 .89861 .79136 .90077 Metetra
.326 .778 .953 r
.79136 .90077 .81721 .89861 .80363 .89533 .77762 .89752 Metetra
.326 .778 .953 r
.77762 .89752 .80363 .89533 .78979 .89198 .76364 .89422 Metetra
.326 .778 .953 r
.76364 .89422 .78979 .89198 .7757 .88857 .74939 .89085 Metetra
.326 .778 .953 r
.74939 .89085 .7757 .88857 .76135 .8851 .73489 .88742 Metetra
.326 .778 .953 r
.73489 .88742 .76135 .8851 .74673 .88157 .72011 .88393 Metetra
.326 .778 .953 r
.72011 .88393 .74673 .88157 .73183 .87797 .70505 .88037 Metetra
.326 .778 .953 r
.70505 .88037 .73183 .87797 .71665 .87429 .68971 .87675 Metetra
.326 .778 .953 r
.68971 .87675 .71665 .87429 .70117 .87055 .67407 .87305 Metetra
.326 .778 .953 r
.67407 .87305 .70117 .87055 .68539 .86673 .65813 .86929 Metetra
.326 .778 .953 r
.65813 .86929 .68539 .86673 .6693 .86284 .64188 .86545 Metetra
.563 .425 .651 r
.64188 .86545 .6693 .86284 .65289 .85887 .61769 .17121 Metetra
.935 .706 .566 r
.61769 .17121 .65289 .85887 .62755 .14766 .60207 .17567 Metetra
.85 .958 .879 r
.60207 .17567 .62755 .14766 .61183 .15512 .58607 .17696 Metetra
.791 .96 .92 r
.58607 .17696 .61183 .15512 .5957 .15709 .56972 .17702 Metetra
.766 .956 .934 r
.56972 .17702 .5957 .15709 .5792 .15739 .55302 .17645 Metetra
.752 .952 .941 r
.55302 .17645 .5792 .15739 .56234 .15684 .53595 .1755 Metetra
.744 .95 .945 r
.53595 .1755 .56234 .15684 .5451 .15581 .51852 .17434 Metetra
.739 .949 .947 r
.51852 .17434 .5451 .15581 .52748 .15449 .5007 .17305 Metetra
.737 .949 .948 r
.5007 .17305 .52748 .15449 .50946 .153 .48249 .17172 Metetra
.735 .949 .948 r
.48249 .17172 .50946 .153 .49104 .15142 .46386 .17042 Metetra
.326 .778 .953 r
.86919 .91118 .89475 .90915 .88224 .90605 .85654 .90812 Metetra
.326 .778 .953 r
.85654 .90812 .88224 .90605 .86952 .9029 .84366 .90501 Metetra
.326 .778 .953 r
.84366 .90501 .86952 .9029 .85656 .89969 .83055 .90184 Metetra
.326 .778 .953 r
.83055 .90184 .85656 .89969 .84337 .89643 .81721 .89861 Metetra
.326 .778 .953 r
.81721 .89861 .84337 .89643 .82994 .8931 .80363 .89533 Metetra
.326 .778 .953 r
.80363 .89533 .82994 .8931 .81627 .88972 .78979 .89198 Metetra
.326 .778 .953 r
.78979 .89198 .81627 .88972 .80234 .88627 .7757 .88857 Metetra
.326 .778 .953 r
.7757 .88857 .80234 .88627 .78815 .88275 .76135 .8851 Metetra
.326 .778 .953 r
.76135 .8851 .78815 .88275 .77369 .87917 .74673 .88157 Metetra
.326 .778 .953 r
.74673 .88157 .77369 .87917 .75895 .87552 .73183 .87797 Metetra
.326 .778 .953 r
.73183 .87797 .75895 .87552 .74393 .87181 .71665 .87429 Metetra
.326 .778 .953 r
.71665 .87429 .74393 .87181 .72862 .86801 .70117 .87055 Metetra
.326 .778 .953 r
.70117 .87055 .72862 .86801 .71301 .86415 .68539 .86673 Metetra
.326 .778 .953 r
.68539 .86673 .71301 .86415 .69708 .8602 .6693 .86284 Metetra
.326 .778 .953 r
.6693 .86284 .69708 .8602 .68084 .85618 .65289 .85887 Metetra
.56 .422 .65 r
.65289 .85887 .68084 .85618 .66426 .85208 .62755 .14766 Metetra
.937 .706 .563 r
.62755 .14766 .66426 .85208 .64735 .84789 .61183 .15512 Metetra
.939 .708 .561 r
.61183 .15512 .64735 .84789 .62188 .13174 .5957 .15709 Metetra
.822 .959 .901 r
.5957 .15709 .62188 .13174 .60562 .13513 .5792 .15739 Metetra
.789 .957 .923 r
.5792 .15739 .60562 .13513 .58898 .1359 .56234 .15684 Metetra
.771 .953 .933 r
.56234 .15684 .58898 .1359 .57195 .1355 .5451 .15581 Metetra
.761 .951 .938 r
.5451 .15581 .57195 .1355 .55454 .13445 .52748 .15449 Metetra
.755 .95 .941 r
.52748 .15449 .55454 .13445 .53673 .13301 .50946 .153 Metetra
.751 .949 .943 r
.50946 .153 .53673 .13301 .51852 .13132 .49104 .15142 Metetra
.326 .778 .953 r
.89475 .90915 .9206 .90709 .90825 .90396 .88224 .90605 Metetra
.326 .778 .953 r
.88224 .90605 .90825 .90396 .89568 .90077 .86952 .9029 Metetra
.326 .778 .953 r
.86952 .9029 .89568 .90077 .88288 .89752 .85656 .89969 Metetra
.326 .778 .953 r
.85656 .89969 .88288 .89752 .86985 .89422 .84337 .89643 Metetra
.326 .778 .953 r
.84337 .89643 .86985 .89422 .85659 .89085 .82994 .8931 Metetra
.326 .778 .953 r
.82994 .8931 .85659 .89085 .84307 .88742 .81627 .88972 Metetra
.326 .778 .953 r
.81627 .88972 .84307 .88742 .82931 .88393 .80234 .88627 Metetra
.326 .778 .953 r
.80234 .88627 .82931 .88393 .81528 .88037 .78815 .88275 Metetra
.326 .778 .953 r
.78815 .88275 .81528 .88037 .80099 .87675 .77369 .87917 Metetra
.326 .778 .953 r
.77369 .87917 .80099 .87675 .78642 .87305 .75895 .87552 Metetra
.326 .778 .953 r
.75895 .87552 .78642 .87305 .77157 .86929 .74393 .87181 Metetra
.326 .778 .953 r
.74393 .87181 .77157 .86929 .75643 .86545 .72862 .86801 Metetra
.326 .778 .953 r
.72862 .86801 .75643 .86545 .74099 .86153 .71301 .86415 Metetra
.326 .778 .953 r
.71301 .86415 .74099 .86153 .72523 .85753 .69708 .8602 Metetra
.326 .778 .953 r
.69708 .8602 .72523 .85753 .70917 .85346 .68084 .85618 Metetra
.326 .778 .953 r
.68084 .85618 .70917 .85346 .69277 .8493 .66426 .85208 Metetra
.326 .778 .953 r
.66426 .85208 .69277 .8493 .67603 .84505 .64735 .84789 Metetra
.556 .418 .648 r
.64735 .84789 .67603 .84505 .65895 .84072 .62188 .13174 Metetra
.942 .709 .557 r
.62188 .13174 .65895 .84072 .63218 .10309 .60562 .13513 Metetra
.867 .951 .865 r
.60562 .13513 .63218 .10309 .61584 .11035 .58898 .1359 Metetra
.817 .956 .905 r
.58898 .1359 .61584 .11035 .59905 .11214 .57195 .1355 Metetra
.794 .954 .921 r
.57195 .1355 .59905 .11214 .58187 .11211 .55454 .13445 Metetra
.78 .952 .929 r
.55454 .13445 .58187 .11211 .56428 .11116 .53673 .13301 Metetra
.772 .95 .934 r
.53673 .13301 .56428 .11116 .54629 .10966 .51852 .13132 Metetra
.326 .778 .953 r
.9206 .90709 .94675 .90501 .93456 .90184 .90825 .90396 Metetra
.326 .778 .953 r
.90825 .90396 .93456 .90184 .92216 .89861 .89568 .90077 Metetra
.326 .778 .953 r
.89568 .90077 .92216 .89861 .90952 .89533 .88288 .89752 Metetra
.326 .778 .953 r
.88288 .89752 .90952 .89533 .89666 .89198 .86985 .89422 Metetra
.326 .778 .953 r
.86985 .89422 .89666 .89198 .88356 .88857 .85659 .89085 Metetra
.326 .778 .953 r
.85659 .89085 .88356 .88857 .87021 .8851 .84307 .88742 Metetra
.326 .778 .953 r
.84307 .88742 .87021 .8851 .85661 .88157 .82931 .88393 Metetra
.326 .778 .953 r
.82931 .88393 .85661 .88157 .84276 .87797 .81528 .88037 Metetra
.326 .778 .953 r
.81528 .88037 .84276 .87797 .82864 .87429 .80099 .87675 Metetra
.326 .778 .953 r
.80099 .87675 .82864 .87429 .81424 .87055 .78642 .87305 Metetra
.326 .778 .953 r
.78642 .87305 .81424 .87055 .79957 .86673 .77157 .86929 Metetra
.326 .778 .953 r
.77157 .86929 .79957 .86673 .7846 .86284 .75643 .86545 Metetra
.326 .778 .953 r
.75643 .86545 .7846 .86284 .76934 .85887 .74099 .86153 Metetra
.326 .778 .953 r
.74099 .86153 .76934 .85887 .75377 .85482 .72523 .85753 Metetra
.326 .778 .953 r
.72523 .85753 .75377 .85482 .73788 .85069 .70917 .85346 Metetra
.326 .778 .953 r
.70917 .85346 .73788 .85069 .72166 .84648 .69277 .8493 Metetra
.326 .778 .953 r
.69277 .8493 .72166 .84648 .70511 .84217 .67603 .84505 Metetra
.326 .778 .953 r
.67603 .84505 .70511 .84217 .68822 .83778 .65895 .84072 Metetra
.553 .415 .647 r
.65895 .84072 .68822 .83778 .67096 .83329 .63218 .10309 Metetra
.943 .709 .554 r
.63218 .10309 .67096 .83329 .65333 .82871 .61584 .11035 Metetra
.945 .71 .551 r
.61584 .11035 .65333 .82871 .62633 .08096 .59905 .11214 Metetra
.855 .951 .876 r
.59905 .11214 .62633 .08096 .60944 .08532 .58187 .11211 Metetra
.821 .953 .904 r
.58187 .11211 .60944 .08532 .5921 .08618 .56428 .11116 Metetra
.802 .951 .917 r
.56428 .11116 .5921 .08618 .57435 .08556 .54629 .10966 Metetra
.326 .778 .953 r
.94675 .90501 .97322 .9029 .96119 .89969 .93456 .90184 Metetra
.326 .778 .953 r
.93456 .90184 .96119 .89969 .94895 .89643 .92216 .89861 Metetra
.326 .778 .953 r
.92216 .89861 .94895 .89643 .93649 .8931 .90952 .89533 Metetra
.326 .778 .953 r
.90952 .89533 .93649 .8931 .92379 .88972 .89666 .89198 Metetra
.326 .778 .953 r
.89666 .89198 .92379 .88972 .91086 .88627 .88356 .88857 Metetra
.326 .778 .953 r
.88356 .88857 .91086 .88627 .89769 .88275 .87021 .8851 Metetra
.326 .778 .953 r
.87021 .8851 .89769 .88275 .88426 .87917 .85661 .88157 Metetra
.326 .778 .953 r
.85661 .88157 .88426 .87917 .87058 .87552 .84276 .87797 Metetra
.326 .778 .953 r
.84276 .87797 .87058 .87552 .85664 .87181 .82864 .87429 Metetra
.326 .778 .953 r
.82864 .87429 .85664 .87181 .84242 .86801 .81424 .87055 Metetra
.326 .778 .953 r
.81424 .87055 .84242 .86801 .82793 .86415 .79957 .86673 Metetra
.326 .778 .953 r
.79957 .86673 .82793 .86415 .81315 .8602 .7846 .86284 Metetra
.326 .778 .953 r
.7846 .86284 .81315 .8602 .79807 .85618 .76934 .85887 Metetra
.326 .778 .953 r
.76934 .85887 .79807 .85618 .78268 .85208 .75377 .85482 Metetra
.326 .778 .953 r
.75377 .85482 .78268 .85208 .76698 .84789 .73788 .85069 Metetra
.326 .778 .953 r
.73788 .85069 .76698 .84789 .75096 .84362 .72166 .84648 Metetra
.326 .778 .953 r
.72166 .84648 .75096 .84362 .7346 .83925 .70511 .84217 Metetra
.326 .778 .953 r
.70511 .84217 .7346 .83925 .71789 .8348 .68822 .83778 Metetra
.326 .778 .953 r
.68822 .83778 .71789 .8348 .70083 .83025 .67096 .83329 Metetra
.326 .778 .953 r
.67096 .83329 .70083 .83025 .6834 .8256 .65333 .82871 Metetra
.548 .41 .645 r
.65333 .82871 .6834 .8256 .66559 .82085 .62633 .08096 Metetra
.947 .711 .547 r
.62633 .08096 .66559 .82085 .64739 .81599 .60944 .08532 Metetra
.949 .711 .545 r
.60944 .08532 .64739 .81599 .62011 .05365 .5921 .08618 Metetra
.857 .948 .874 r
.5921 .08618 .62011 .05365 .60265 .05681 .57435 .08556 Metetra
.326 .778 .953 r
.97322 .9029 1 .90077 .98814 .89752 .96119 .89969 Metetra
.326 .778 .953 r
.96119 .89969 .98814 .89752 .97607 .89422 .94895 .89643 Metetra
.326 .778 .953 r
.94895 .89643 .97607 .89422 .96378 .89085 .93649 .8931 Metetra
.326 .778 .953 r
.93649 .8931 .96378 .89085 .95126 .88742 .92379 .88972 Metetra
.326 .778 .953 r
.92379 .88972 .95126 .88742 .9385 .88393 .91086 .88627 Metetra
.326 .778 .953 r
.91086 .88627 .9385 .88393 .92551 .88037 .89769 .88275 Metetra
.326 .778 .953 r
.89769 .88275 .92551 .88037 .91226 .87675 .88426 .87917 Metetra
.326 .778 .953 r
.88426 .87917 .91226 .87675 .89877 .87305 .87058 .87552 Metetra
.326 .778 .953 r
.87058 .87552 .89877 .87305 .88501 .86929 .85664 .87181 Metetra
.326 .778 .953 r
.85664 .87181 .88501 .86929 .87098 .86545 .84242 .86801 Metetra
.326 .778 .953 r
.84242 .86801 .87098 .86545 .85667 .86153 .82793 .86415 Metetra
.326 .778 .953 r
.82793 .86415 .85667 .86153 .84207 .85753 .81315 .8602 Metetra
.326 .778 .953 r
.81315 .8602 .84207 .85753 .82719 .85346 .79807 .85618 Metetra
.326 .778 .953 r
.79807 .85618 .82719 .85346 .81199 .8493 .78268 .85208 Metetra
.326 .778 .953 r
.78268 .85208 .81199 .8493 .79649 .84505 .76698 .84789 Metetra
.326 .778 .953 r
.76698 .84789 .79649 .84505 .78066 .84072 .75096 .84362 Metetra
.326 .778 .953 r
.75096 .84362 .78066 .84072 .7645 .83629 .7346 .83925 Metetra
.326 .778 .953 r
.7346 .83925 .7645 .83629 .74799 .83177 .71789 .8348 Metetra
.326 .778 .953 r
.71789 .8348 .74799 .83177 .73113 .82716 .70083 .83025 Metetra
.326 .778 .953 r
.70083 .83025 .73113 .82716 .7139 .82244 .6834 .8256 Metetra
.326 .778 .953 r
.6834 .8256 .7139 .82244 .6963 .81762 .66559 .82085 Metetra
.326 .778 .953 r
.66559 .82085 .6963 .81762 .6783 .81269 .64739 .81599 Metetra
.543 .404 .642 r
.64739 .81599 .6783 .81269 .6599 .80766 .62011 .05365 Metetra
.951 .712 .541 r
.62011 .05365 .6599 .80766 .64108 .8025 .60265 .05681 Metetra
0 g
.25 Mabswid
.03704 .18015 m
.63086 0 L
s
.63086 0 m
.64108 .8025 L
s
.64108 .8025 m
0 .87305 L
s
0 .87305 m
.03704 .18015 L
s
.9679 .25222 m
.63086 0 L
s
.63086 0 m
.64108 .8025 L
s
.64108 .8025 m
1 .90077 L
s
1 .90077 m
.9679 .25222 L
s
.03704 .18015 m
0 .87305 L
s
.03704 .18015 m
.04314 .18305 L
s
gsave
.02483 .17436 -71 -15.564 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.02729 .3625 m
.03358 .36496 L
s
gsave
.01471 .35756 -77 -14.9233 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
.01714 .55237 m
.0236 .55436 L
s
gsave
.00423 .54839 -83 -14.2618 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
.00656 .75026 m
.01316 .75172 L
s
gsave
-0.00663 .74734 -83 -13.5786 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.125 Mabswid
.03512 .21604 m
.0388 .21773 L
s
.03319 .25222 m
.03689 .25385 L
s
.03124 .28868 m
.03497 .29027 L
s
.02927 .32544 m
.03302 .32697 L
s
.02529 .39985 m
.02909 .40128 L
s
.02328 .43752 m
.0271 .43889 L
s
.02125 .47549 m
.02509 .4768 L
s
.0192 .51377 m
.02306 .51503 L
s
.01506 .59129 m
.01895 .59243 L
s
.01296 .63054 m
.01687 .63161 L
s
.01085 .67011 m
.01477 .67112 L
s
.00871 .71002 m
.01266 .71096 L
s
.00439 .79084 m
.00837 .79165 L
s
.00221 .83177 m
.00619 .83251 L
s
gsave
-0.05789 .48869 -132.75 -14.3958 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Max) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.063 15.813 moveto
(@) show
89.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
94.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
99.063 15.813 moveto
(D) show
106.000 15.813 moveto
(H) show
109.813 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
126.938 15.813 moveto
(L) show
130.750 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{385.875, 364.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40009200028b000`400?l00000o`00003oo`3ooooo
0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo
003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0
ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool0
0?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3o
oom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00
o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oo
od@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o
0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0oooo
A03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0
ooooo`3ooom40?ooo`00o`3ooonE0?ooo`050?oOW018000000000000M02oool00`3oool01P3o_g@0
000007Boo`3oool0oooO09a800<0000000<0M;oo0?ooo`3oool0VP3oool00?l0ooooU03oool01`3o
okl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a8M02o
ool00P3oool00`2oM000M;oo0?ooo`2J0?ooo`00o`3ooonD0?ooo`030;md0000B9`0gooo0080oooo
00<0_g@0001d_`3oool02P3oool00`3ooml0W4Qd0;ooo`2J0?ooo`00o`3ooonC0?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`2I0?ooo`00o`3ooonC
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RLg`2J
0?ooo`00o`3ooonC0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3O
W4P0000007Boo`2J0?ooo`00o`3ooonC0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02L
gol01P3oool01`3Ogml0oooo0?oo_`1d000000000000M02oool0VP3oool00?l0ooooT`3oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia800000000000000000000B02L
gonK0?ooo`00o`3ooonC0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool01P3oool0
103ookl0M00000000018W=nM0?ooo`00o`3ooom;0?ooo`040=nLB0000000000007BoodD0oooo00<0
gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo09d0oooo003o
0?ooodh0oooo00<0_gAd0;ooo`3oool0@`3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
203oool01P3ogi`0B000000000000000001809cOoiX0oooo003o0?ooodh0oooo00<0omnL07BLg`3o
ool0A03oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@0000000000000000004RLg`2J
0?ooo`00o`3ooom>0?ooo`030?oog`2LB7@0_ooo0?80oooo003>0?ooo`050;md0000000000000000
B02Lgol00`3oool01`2oM00000000000000007@0_ooo0?ooo`3OW4P00P0000000`0007@0_ooO09a8
00020000000:0018W02LM0000000001d_`3oool0omnL04P000000000B9cO0=nLB080000000L0001d
0;ooo`3oool0oooo0?oog`2LB7@0_ooo00D0oooo00D0_g@0000000000000001d0;ooo`070?ooo`05
0=nLB000000000000000M01dM0000P0000000`0007@0_ooO09a80002000000040018W03Oool0oooo
0;md0080000000@0001d0;ooo`3oool0gia80P0000000`0007@0_ooO09a80002000000070018W02L
M0000000001d_`3ookl0_kno07@000020000000807Boo`3oool0oooo0?oOW0180000000004RLg`3O
W4P2000000040000M02oool0oooo0;md0080000000@0001809cOo`3oool0gia80P0000001P0007@0
_ooo0?oOW0180000000004RLg`@0oooo00<0_g@007Boo`3oool0l@3oool00<l0oooo00<0gia80000
M02oool00`3oool0702oM0000000001d_`3oool0omnL04P0M02oool0oooo0=nLB00007@0_ooo0?no
M0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB00007@0_ooo0=no_`3Oool0gia80000
M02oool30?ooo`030?noM0000000B9cO00@0oooo00L0_g@000000000M;l0oooo0?oOW01807@0_ooo
00L0oooo00/0gia80000M02oool0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`0e
0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80000M02oool0okmd0000001d_ol0oooo
0?oOW018000007Bo0?no_`2ookl0M01809cOo`3oool0oono07@0B02Lgol0oooo0=nLB00007@0_ooo
0=no_`3Oool0gia80000M02oool0oooo0;md0000000004RL0=ooo`3ooml0W4Qd0;ooo`3oool0gia8
0000M02oool0oooo0=nLB00007@0_ooo0=no_`3Oool00P3oool0103ogi`0B0000018W03Ooooa0?oo
o`00c`3oool00`3OW4P0001d0;ooo`020?ooo`030?oOW0180000B9cO00<0oooo00d0omnL07BLg`3o
ool0gia80000M02oool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00<0gia80018W03O
ool00P3oool00`3OW4P0001d0;ooo`020?ooo`050?oog`2LB00004RL09adM02oool00P3oool00`3o
gi`0B00004RLg`030?ooo`030?oOW01dW=l0oooo00H0oooo00l0gia80000M02oool0oooo0=nLB000
B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool0403ogi`0M9cO0?ooo`3O
W4P0001d0;ooo`3o_g@0001809cOo`3oool0oooO09a80000M;l0oooo0?oo_`2o_ol20?ooo`070?no
M0000000M;oo0?ooo`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0_ooo0?oOW018000007Bo0`3o
ool02P3ogi`0W;oo0?ooo`3OW4P0001d0;ooo`3oool0gia80018W03Oool40?ooo`040;md0000B9`0
W7@007Booo40oooo003?0?ooo`030=nLB00007@0_ooo0080oooo00<0okmd0000B02Lgol01P3oool0
2P3OW4P0001d0;ooo`3oool0okmd00000000M;l0oooO09a80000M;l20?ooo`030=nLB000B9`0gooo
0080oooo00<0gia80000M02oool00P3oool01@3ogi`0B00007Boo`3o_g@0B9cO0080oooo00<0okmd
0000B02Lgol0303oool03`3OW4P0001d0;ooo`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo
0?ooo`3o_g@0001809cOo`060?ooo`0:0=nLB00007@0_ooo0?ooo`3o_g@00000001d_`3ooml0W4P0
001d_`<0oooo00T0oono07@000000000001d0;ooo`3oool0gia80018W03Oool00P3oool01P3OW4P0
001d0;ooo`3o_g@0001809cOo`H0oooo00L0gia80000M02oool0oooo0=nLB000B9`0gooo00<0oooo
00D0okmd0000001d_ol0omnL04Qd_`3a0?ooo`00c`3oool00`3OW4P0001d0;ooo`020?ooo`030?no
M00007@0_ooo00H0oooo00<0gia80000M02oool00`3oool0103o_g@0000000000000M;l20?ooo`03
0=nLB000B9`0gooo0080oooo00<0gia80000M02oool00P3oool02P2oM00004RL0=ooo`3ooml0W4Qd
0;ooo`3oool0okmd0000M02oool<0?ooo`0?0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0
oooO09a80000M;l0oooo0?noM00007@0_ooo00H0oooo00<0gia80000M02oool00`3oool0103o_g@0
000000000000M;l20?ooo`040;md00000000000004RLg`<0oooo00<0gia80018W03Oool00P3oool0
1P3OW4P0001d0;ooo`3o_g@0001d0;ooo`H0oooo00L0gia80000M02oool0oooo0=nLB000B9`0gooo
0080oooo00L0oooO09a80000B9`0gooo0?ooo`2oM4P0W=oo0?00oooo003?0?ooo`030=nLB00004P0
W=oo0080oooo00<0omnL04P000000000100000003018W=l0oooo0=nLB00007@0_ooo0?oo_`1d0000
M;oo0?ooo`3ooml0W4P0001d_`80oooo00<0gia80018W03Oool00P3oool01`3OW4P0001d0;ooo`3o
ool0omnL04P0001d_ol00P3oool01@3o_g@0B9cO0?ooo`3ogi`0B00000D0000000<0B9cO0?ooo`3o
ool01@3oool03P3OW4P0001d0;ooo`3oool0_g@00000M02oool0oooo0?oog`2LB00007Bo0?ooo`3o
gi`0B0001@0000004018W=l0oooo0=nLB00007@0_ooo0?oo_`1d0000M;oo0?ooo`3ooml0W4P0001d
_`3oool0omnL04P00018W=l20?ooo`060;nLW03Oool0oooo0=nLB000B9`0gooo0P3oool01P3OW4P0
001d0;ooo`3ogi`0B01809cOo`80oooo00/0okmd0000B02Lgol0oooo0=nLB00007@0_ooo0?ooo`3O
W4P004RL0=ooo`020?ooo`030?noM00004P0W=oo0080oooo00<0okmd04RLg`3oool01@3oool01@3o
gi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol01@3oool0103ooml0W4P00018W03O
oooF0?ooo`00c@3oool02P3ooml0W4P00000000004P0B7BL0=noM00007@0_ooo0;md0018W=l20?oo
o`0>0=nLB00004P0W=oo0?ooo`3OW4P0001d0;ooo`3ookl0M00004RLg`3oool0oono07@00000M;l2
0?ooo`0<0=nLB000B9`0gooo0?oog`2LB00000000000M02oool0oooO09a8000007@0_ooo0P3oool0
1P3ooml0W4Qd0;ooo`3oool0_g@004RLg`80oooo00<0gia80000B02Lgol01@3oool03@3ooml0W4P0
0000000007@0_ooO09a80000000004Q804P0B02Lgol0oono07@00018W=l00P3oool0>02oM000B9cO
0?ooo`3oool0gia80000B02Lgol0oooo0=nLB00007@0_ooo0?oo_`1d0000B9cO0?ooo`3ookl0M000
001d_`3oool0oono07@0M02oool0oooo0?noM000B9`0gooo0?ooo`3OW4P004RL0=ooo`3ooml0W4P0
0000000007@0_ooo0?ooo`2oM00004RL0=ooo`3ooml0W4P00000M02ooml0W4P00000000007@0_ooo
0?ooo`3OW4P004RL0=ooo`3oool0oono07@00000B9`0gooo0P3oool00`3ookl0M01d0;ooo`040?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0
oooO09a80000B9`0goooeP3oool00<l0oooo00P0_g@00000M02ogi`0B0000000M02oool0oooo0=nL
B080000000X0001809cOo`3oool0oooo0=nLB00007@0_ooo0?ooo`3ookl0M0000P000000100007@0
_ooo0?ooo`2oM003000000060018W03Oool0_g@00000M02ooml0W4P00P0000002@0007@0_ooo0?oo
g`2LB00000000018W03Oool0oooo0=nLB002000000030000B02Lgol0oooo00L0oooo00/0_g@00000
M02oool0oooo0;md0000B9`0gooO09a800000000001d0;ooo`030?ooo`050=nLB000000000000000
B02Lgol00P3oool01P3OW4P0001d0;ooo`3oool0oono07@00080000000D0001d0;ooo`3oool0oooo
0?noM00200000004001d_`2oW9`0gooo0;md00<0000000D004RL0=ooo`2oM000001d0;ooo`020?oo
o`050=nLB000000000000000B02Lgol00P3oool0102oM000001d0;ooo`2oM0030000000;0018M01d
B000000000000000B9`0gooo0?ooo`2oM00000000018W03Oool00`3oool00`2oM00004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00T0oooo00@0oooO09a80000B9`0goooeP3oool00=d0oooo00<0gia8
0000M02oool02@3oool00`2oM00004RL0=ooo`0b0?ooo`030=nLB00007@0_ooo0100oooo00<0_g@0
0018W03Oool04P3oool00`2oM00004RL0=ooo`0=0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool01@3oool00`2oM0000000000000050000000307Boo`3oool0oooo0=<0oooo003M0?oo
o`030=nLB00007@0_ooo00T0oooo00<0oono07A8W03Oool0<P3oool00`3OW4P0001d0;ooo`0@0?oo
o`030?oo_`1dB9`0gooo0180oooo00<0oono07A8W03Oool03@3oool00`3ogi`0B00007Boo`030?oo
o`030?noM00004P0W=oo00D0oooo00P0omnL07BLg`3oool0oooo0?oog`2LB00004RL0=oooc00oooo
0P00002T0?ooo`00g@3oool00`3OW4P0001d0;ooo`090?ooo`030?oog`2LM9`0gooo0380oooo00<0
gia80000M02oool0403oool00`3ooml0W7BL0=ooo`0B0?ooo`030?oog`2LM9`0gooo00d0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?ooo`070=nLB01d_ol0oooo0?oog`2LB000
04RL0=ooo`0]0?ooo`<0000000@0oooo0000000000000000XP3oool00=/0oooo00D0oooO09a80000
0000001d0;ooo`0>0?ooo`030=nLB00004P0W=oo01P0oooo00<0gia80000B02Lgol0403oool01@3o
oml0W4P00000000007@0_ooo01D0oooo00<0gia80000B02Lgol02@3oool00`3OW4P0001809cOo`0A
0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3ooml0W4Qd0;ooo`3o
oml0W4P00018W03Oool0:@3oool4000000@0oooo00@000000?ooo`3oool00000X@3oool00=d0oooo
00<0_g@00000M02oool03P3oool00`3OW4P0001809cOo`0H0?ooo`030=nLB00004P0W=oo0180oooo
00<0_g@00000M02oool05@3oool00`3OW4P0001809cOo`090?ooo`030=nLB00004P0W=oo0140oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`060?oOW018M;l0oooO09a80000
B9`0gooo9P3oool3000000P0oooo00D000000?ooo`3oool0oooo0000002P0?ooo`00o`3ooomG0?oo
o`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool01@3OW4P0M;oO09a80000B9`0
gooo02<0oooo0`00000;0?ooo`030000003oool0oooo0080oooo0P00002N0?ooo`00o`3ooomH0?oo
o`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool0203oool01@3ooml0W4Q804Q80000B9`0
gooo01l0oooo100000020?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`2K0?ooo`00o`3ooomH0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`0:0?ooo`040?oOW018000004RL0=oooa`0oooo0`0000060?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`2J0?ooo`00o`3ooomI0?ooo`050?oo_`1d
000000000000M02oool0303oool00`3OW4P004RL0=ooo`0I0?ooo`<000005@3oool00`000000oooo
0?ooo`060?ooo`800000VP3oool00?l0ooooPP3oool5000001L0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`2G0?ooo`00o`3ooomo0?ooo`<00000103oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0UP3oool00?l0ooooO03oool30000
00L0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0:0?ooo`800000UP3oool00?l0
ooooN@3oool300000280oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`2C0?ooo`00
o`3ooome0?ooo`@0000000<0oooo0000003oool08P3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo0980oooo003o0?ooob80oooo00D0omnL04P000000000001d0;ooo`030?ooo`060?no
M0000000M;oo0?ooo`3ookl0M0000`0000000`1d_ol0oooo0?ooo`0l0?ooo`<000001@3oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo00h0oooo0P00002B0?ooo`00o`3ooolQ0?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ogi`0B000
04RLg`020?ooo`030;md0018W=l0oooo03T0oooo0`00000/0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0S`3oool00?l0oooo8@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000
M;l0oooo00X0oooo00<0oono07@0B02Lgol0=@3oool4000002l0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`2>0?ooo`00o`3ooolP0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia8
0000M02oool02P3oool00`3ookl0M00007Boo`0b0?ooo`<000000`3oool00`000000oooo0?ooo`0]
0?ooo`030000003oool0oooo0180oooo0P00002>0?ooo`00o`3ooolP0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol02P3oool00`3o_g@0000007Boo`0_0?ooo`<000001P3oool00`00
0000oooo0?ooo`0]0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0R`3oool00?l0
oooo803oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00@0omnL04P00000
04P0W=oo:`3oool4000003T0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`2:0?oo
o`00o`3ooolP0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01`3oool01@3o_g@0
00000000000004P0W=oo02T0oooo0`0000020?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool05P3oool2000008X0oooo003o0?ooob00oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`090?ooo`030;md0000M;l0oooo02H0oooo0`0000050?ooo`030000003oool0oooo03P0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`270?ooo`00o`3ooolP0?ooo`030?oo
_`1d0000B9cO00<0oooo00<0gia80000M02oool02P3oool00`2oM000B9cO0?ooo`0R0?ooo`<00000
203oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
QP3oool00?l0oooo8@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0
oooO09bL_`3oool00P3oool00`3OW4P0001d0;ooo`0N0?ooo`@00000AP3oool00`000000oooo0?oo
o`0J0?ooo`800000QP3oool00?l0oooo8@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0
1`3oool01P3o_g@0B9cO0?ooo`3ookl0M000001d_a`0oooo0`0000030?ooo`030000003oool0oooo
04@0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`230?ooo`00o`3ooolR0?ooo`05
0?oo_`1d000000000000M02oool02@3oool01@3OW4P000000000000007@0_ooo01T0oooo0`000006
0?ooo`030000003oool0oooo03h0oooo0P0000040?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool0PP3oool00?l0ooooBP3oool4000004H0oooo100000000`3c]XX000000?ooo`030?oo
o`030000003oool0oooo01h0oooo0P0000220?ooo`00o`3ooom70?ooo`<000000P3oool00`000000
oooo0?ooo`130?ooo`8000001@3c]XX00`000000oooo0?ooo`020?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0O`3oool00?l0ooooA03oool3000000D0oooo00<000000?ooo`3oool0
@P3oool2000000H0lkJ:00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo07h0oooo003o0?oood00oooo1000001<0?ooo`030000003Kl]l0000000H0lkJ:00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0R0?ooo`800000OP3oool00>d0oooo00D0omnL
04P000000000001d0;ooo`030?ooo`050?noM0000000M;oo0?ooo`2oM0001P0000000`18W=l0oooo
0?ooo`0i0?ooo`<0000000<0oooo0000003oool0C03oool010000000fo;O0=_bg`0000060?>fRP03
0000003oool0oooo0080oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`1k0?ooo`00
k03oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?oo
o`030=nLB0000000000000<0000000<004RL0=ooo`3oool0=P3oool3000000@0oooo00<000000?oo
o`3oool0BP3oool01@000000fo;O0=_bg`3Kl]l0000000H0lkJ:00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo07X0oooo003/0?ooo`030;md0000B9`0gooo
0080oooo00<0_g@0001d_`3oool01`3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo
03<0oooo0`0000070?ooo`030000003oool0oooo04T0oooo00<000000=_bg`3Kl]l00P3Kl]l00`00
0000lkJ:0?>fRP040?>fRP030000003oool0oooo0080oooo00<000000?ooo`3oool09P3oool20000
07X0oooo003[0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`3OW4P0
M;oo0?ooo`0c0?ooo`@00000E03oool2000000D0fo;O00<000000?>fRP3c]XX0103c]XX00`000000
oooo0?ooo`020?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0M`3oool00>/0oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030;mdB02Lgol0oooo02l0oooo
0`0000030?ooo`030000003oool0oooo0540oooo00<000000=_bg`3Kl]l01@3Kl]l00`000000lkJ:
0?>fRP040?>fRP030000003oool0oooo0080oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`1f0?ooo`00j`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0
oooO09a8M02oool0;03oool3000000H0oooo00<000000?ooo`3oool0D03oool00`000000fo;O0=_b
g`050=_bg`8000001`3c]XX01@000000oooo0?ooo`3oool0000002`0oooo0P00001f0?ooo`00j`3o
ool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0okmd001d_`3oool09`3o
ool4000005/0oooo00<000000=_bg`3Kl]l01@3Kl]l00`000000l[J;000000070?>fRP050000003o
ool0oooo0?ooo`000000;P3oool00`000000oooo0?ooo`1c0?ooo`00j`3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00X0oooo00<0oono07@0B02Lgol0903oool300000080oooo00<0
00000?ooo`3oool0F@3oool00`000000fo;O0=_bg`050=_bg`050000003b]X/0l[J;0?:fR`000000
1P3c]XX300000080oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1b0?ooo`00j`3o
ool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00X0oooo00<0oono07@00018W=l08@3o
ool3000000D0oooo00<000000?ooo`3oool0F03oool00`000000fo;O0=_bg`060=_bg`050000003b
]X/0l[J;0?:fR`0000001P3c]XX00`000000l[J<000000020?ooo`030000003oool0oooo02d0oooo
0P00001b0?ooo`00k03oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0
oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`0M0?ooo`@00000203oool00`000000oooo0?oo
o`1G0?ooo`@000001@3Kl]l010000000l[J;0?:fR`3b]X/2000000H0lkJ:00<000000?:fS0000000
0P3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo06l0oooo003/0?ooo`070?oo_`1d
0000M;oo0?ooo`3o_g@0001d0;ooo`060?ooo`080?oog`2LB4P0W=oo0?ooo`3ookl0M0000000M02o
oolJ0?ooo`<0000000<0oooo0000003oool0HP3oool00`000000cO?Z0<gcjP020<gcjPT0000000<0
d_?W0000003c]XX01@3c]XX00`000000l[J<000000020?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool0KP3oool00>d0oooo00D0oono07@000000000001d0;ooo`080?ooo`060?oog`2L
B00000000000000007@0_ooo603oool3000000@0oooo00<000000?ooo`3oool0H03oool00`000000
cO?Z0<gcjP060<gcjP030000003BlnL0d_?W00@0d_?W00<000000?>fRP3c]XX0103c]XX00`000000
l[J<000000020?ooo`030000003oool0oooo0340oooo0P00001^0?ooo`00o`3ooolF0?ooo`<00000
J@3oool00`000000cO?Z0<gcjP060<gcjP030000003BlnL0d_?W00@0d_?W0P0000060?>fRP030000
003b]X`000000080oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`1[0?ooo`00o`3o
oolB0?ooo`@00000J`3oool00`000000cO?Z0<gcjP060<gcjP030000003BlnL0d_?W00@0d_?W00<0
00000=[ch00000001P3c]XX00`000000l[J<000000020?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool0JP3oool00?l0oooo3`3oool3000000<0oooo00<000000?ooo`3oool0J03oool0
0`000000cO?Z0<gcjP060<gcjP030000003BlnL0d_?W00@0d_?W00@000000=[ch03Jln0000001P3c
]XX00`000000l[J<000000020?ooo`030000003oool0oooo03D0oooo0P00001Z0?ooo`00^`3oool0
1@3ogi`0B0000000000007@0_ooo00<0oooo00<0okmd0000001d_ol00`3oool01P2oM00000000000
0000000004RL0=oooc`0oooo0`0000060?ooo`030000003oool0oooo06H0oooo0P0000080<gcjP03
0000003BlnL0d_?W00<0d_?W0P0000030=[ch0030000003c]XX0lkJ:00@0lkJ:00<000000?:fS000
00000P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo06L0oooo002j0?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00@0oooo00<0omnL04P0
001d_ol0>P3oool400000740oooo00<000000<gcjP3=lnX01P3=lnX2000000D0d_?W00<000000=[c
h03Jln000`3Jln000`000000lkJ:0?>fRP040?>fRP030000003b]X`000000080oooo00<000000?oo
o`3oool0>03oool00`000000oooo0?ooo`1V0?ooo`00^P3oool00`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00P0oooo00<0omnL04P0001d_ol0=`3oool3000000<0oooo00<000000?ooo`3o
ool0KP3oool00`000000cO?Z0<gcjP060<gcjP030000003BlnL0d_?W00@0d_?W00<000000=[ch03J
ln00103Jln000`000000lkJ:0?>fRP040?>fRP030000003b]X`0l[J<00<00000>`3oool2000006H0
oooo002i0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`3ogi`0B000
07Boo`0d0?ooo`<000001P3oool00`000000oooo0?ooo`1]0?ooo`030000003=lnX0cO?Z00H0cO?Z
00<000000=;ci`3BlnL0103BlnL00`000000f_?P0=[ch0050=[ch0030000003c]XX0lkJ:00@0lkJ:
00H000000?:fS03b]X`000000?6eS@00000m0?ooo`030000003oool0oooo06<0oooo002i0?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`3ogi`0B00007Boo`0`0?ooo`@0
0000N03oool00`000000cO?Z0<gcjP060<gcjP030000003BlnL0d_?W00@0d_?W00<000000=[ch03J
ln001@3Jln02000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00000n0?ooo`030000003oool0
oooo0680oooo002i0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol0203oool00`3o
gi`0B00007Boo`0]0?ooo`<000000P3oool00`000000oooo0?ooo`1f0?ooo`@000001@3=lnX00`00
0000d_?W0=;ci`040=;ci`030000003Jln00f_?P00D0f_?P00<000000?6eS@0000001P3c]XX01P00
0000l[J<0?:fS0000000lKF=000003l0oooo0P00001R0?ooo`00^@3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM00004P0W=oo00P0oooo00<0omnL04P0001d_ol0:P3oool3000000D0oooo00<00000
0?ooo`3oool0M@3oool00`000000aO?_0<Gck`020<GckaL000001P3c]XX01P000000l[J<0?:fS000
0000lKF=00000440oooo00<000000?ooo`3oool0G`3oool00;T0oooo00<0omnL04P0001d_ol00`3o
ool00`3o_g@0001809cOo`080?ooo`030?oOW0180000M;oo02L0oooo0`00001n0?ooo`8000001`35
lnl2000000H0ao?]00<000000<_dj`3;m>/01@3;m>/01@000000dOCW0=7di`3Am>L0000000H0lkJ:
00H000000?:fS03b]X`000000?6eS@0000120?ooo`030000003oool0oooo05h0oooo002i0?ooo`03
0?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool0203oool00`3ogi`0B00007Boo`0S0?ooo`@0
000000<0oooo0000003oool0O@3oool00`000000aO?_0<Gck`060<Gck`0300000037lnd0ao?]00D0
ao?]00<000000<_dj`3;m>/01@3;m>/00`000000dOCW0=7di`020=7di`030000003c]XX0lkJ:00D0
lkJ:00D000000?:fS0000000lKF=000000130?ooo`800000GP3oool00;X0oooo00<0gia80018W03O
ool00P3oool00`2oM00007Bo0?ooo`080?ooo`030?oOW0180000M;oo0200oooo0`0000050?ooo`03
0000003oool0oooo07/0oooo00<000000<Gck`35lnl01@35lnl2000000H0ao?]0P0000070<_dj`03
0000003Am>L0dOCW00<0dOCW00<000000?>fRP3c]XX01@3c]XX01@000000l[J<0000003a]Hd00000
04D0oooo00<000000?ooo`3oool0F`3oool00;X0oooo00L0oono07@0001d_ol0oooo0?noM00007@0
_ooo00L0oooo00D0oooO09a800000000000007Boo`0M0?ooo`<00000203oool00`000000oooo0?oo
o`1i0?ooo`8000001`35lnl00`000000ao?]0<Ock@050<Ock@030000003;m>/0boC[00H0boC[00<0
00000=7di`3Am>L0103Am>L00`000000lkJ:0?>fRP050?>fRP050000003b]X`000000?6eS@000000
AP3oool00`000000oooo0?ooo`1J0?ooo`00^`3oool01@3ookl0M0000000000007@0_ooo00X0oooo
00<0okmd0000001d_ol06@3oool4000000030?ooo`000000oooo08<0oooo00<000000<Gck`35lnl0
1P35lnl00`000000ao?]0<Ock@050<Ock@030000003;m>/0boC[00H0boC[00<000000=7di`3Am>L0
103Am>L2000000L0lkJ:00D000000?:fS0000000lKF=000000170?ooo`800000FP3oool00><0oooo
0`0000050?ooo`030000003oool0oooo0840oooo00<000000<Gck`35lnl01@35lnl2000000L0ao?]
00<000000<_dj`3;m>/01@3;m>/2000000H0dOCW00@000000=gcg@3Mlmd000001P3c]XX010000000
l[J<0000003a]Hd2000004P0oooo00<000000?ooo`3oool0E`3oool00>00oooo0`00002:0?ooo`80
00001`35lnl00`000000ao?]0<Ock@060<Ock@030000003;m>/0boC[00D0boC[00<000000=7di`3A
m>L01@3Am>L01@000000gO?M0=gcg@3Mlmd0000000H0lkJ:00@000000?:fS0000000lKF=0P000019
0?ooo`800000E`3oool009@0oooo00D0omnL04P000000000001d0;ooo`130?ooo`@0000000<0oooo
0000003oool0R@3oool00`000000aO?_0<Gck`050<Gck`8000001`37lnd2000000L0boC[00<00000
0=7di`3Am>L01@3Am>L00`000000gO?M0=gcg@020=gcg@030000003c]XX0lkJ:00@0lkJ:00@00000
0?:fS0000000lKF=0P00001;0?ooo`030000003oool0oooo05@0oooo002C0?ooo`070?oo_`1d0000
M;oo0?ooo`3o_g@0001d0;ooo`0o0?ooo`<000001@3oool00`000000oooo0?ooo`270?ooo``00000
1@37lnd00`000000boC[0<_dj`060<_dj`030000003Am>L0dOCW00D0dOCW00<000000=gcg@3Mlmd0
0`3Mlmd00`000000lkJ:0?>fRP040?>fRP040000003b]X`000000?6eS@800000C03oool00`000000
oooo0?ooo`1C0?ooo`00T`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo03/0oooo
0`00002A0?ooo`0300000030l_40`?;a00D0`?;a00<000000<7cl031lo000P31lo08000000D0boC[
00<000000=7di`3Am>L01@3Am>L00`000000gO?M0=gcg@05000000H0lkJ:00@000000?:fS0000000
lKF=0P00001=0?ooo`800000D`3oool00980oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d
0;ooo`0h0?ooo`<0000000<0oooo0000003oool0S`3oool2000000H0`?;a0P0000080<7cl0030000
0032lo00`_?`0080`_?`400000020<Wdk0030000003BmNH0000000H0lkJ:00@000000?:fS0000000
lKF=0P00001?0?ooo`030000003oool0oooo0500oooo002B0?ooo`030?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol0=03oool4000000@0oooo00<000000?ooo`3oool0S@3oool00`000000`?;a
0<3bl@050<3bl@0300000031lo00`O?`00H0`O?`0P0000080<;cl00300000035lnh0aO?^00D0aO?^
00<000000<Wdk039m>`00`39m>`010000000d_GV0=;eiP0000060?>fRP040000003b]X`000000?6e
S@800000D03oool00`000000oooo0?ooo`1?0?ooo`00TP3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo0340oooo0`00002G0?ooo`0300000030l_40`?;a00D0`?;a00<000000<7cl031
lo001P31lo000`000000`_?`0<;cl0060<;cl08000001P35lnh2000000@0bOC/0P0000030=;eiP03
0000003c]XX0lkJ:00@0lkJ:00@000000?:fS0000000lKF=0P00001A0?ooo`800000C`3oool00980
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0^0?ooo`<0000000<0oooo0000003o
ool0U@3oool2000000H0`?;a0P0000080<7cl00300000032lo00`_?`00H0`_?`00<000000<GckP35
lnh01@35lnh00`000000bOC/0<Wdk0030<Wdk0030000003BmNH0d_GV00<0d_GV00<000000?>fRP3c
]XX0103c]XX00`000000l[J<0?:fS003000005<0oooo00<000000?ooo`3oool0C03oool00980oooo
00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0Z0?ooo`@00000103oool00`000000oooo
0?ooo`2C0?ooo`0300000030l_40`?;a00D0`?;a00<000000<7cl031lo001P31lo02000000P0`_?`
00<000000<GckP35lnh01@35lnh00`000000bOC/0<Wdk0030<Wdk0030000003BmNH0d_GV00@0d_GV
00<000000?>fRP3c]XX0103c]XX00`000000l[J<0?:fS003000005@0oooo00<000000?ooo`3oool0
B`3oool00980oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`0W0?ooo`<00000W@3o
ool00`000000`?;a0<3bl@050<3bl@0300000031lo00`O?`00H0`O?`00<000000<;cl032lo001P32
lo02000000H0aO?^0P0000050<Wdk0030000003BmNH0d_GV00@0d_GV0P0000060?>fRP030000003b
]X`0l[J<00<00000E@3oool2000004/0oooo00260?ooo`050?o4G`0D000000000000=@2?ool0203o
ool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo02@0oooo0`0000020?ooo`030000003o
ool0oooo09T0oooo0P0000060<3bl@8000002031lo000`000000`_?`0<;cl0060<;cl00300000035
lnh0aO?^00D0aO?^00<000000<Wdk039m>`01039m>`00`000000d_GV0=;eiP040=;eiP030000003`
]Hl0000000H0lkJ:00<000000?:fS03b]X`00`00001G0?ooo`030000003oool0oooo04P0oooo0025
0?ooo`070?ooS`0e0000=Hoo0?ooo`3oScD0000e08ooo`070?ooo`070?oo_`1d0000M;oo0?ooo`3o
_g@0001d0;ooo`0Q0?ooo`@000001@3oool00`000000oooo0?ooo`2H0?ooo``000001@31lo020000
00P0`_?`00<000000<GckP35lnh01@35lnh00`000000bOC/0<Wdk0030<Wdk08000001P3BmNH01000
0000l;F?0?2eS`0000060?>fRP030000003b]X`0l[J<00<00000F03oool00`000000oooo0?ooo`17
0?ooo`00Q@3oool00`2?=@0001AO0<Coo`020?ooo`0308le0000=Hl0oooo00L0oooo00D0oono07@0
00000000001d0;ooo`0O0?ooo`<000002@3oool00`000000oooo0?ooo`2F0?ooo`8000001`2ll_80
0`000000_?;b0;cblP020;cblPP000001@32lo02000000H0aO?^0P0000050<Wdk0030000003BmNH0
d_GV00D0d_GV00<000000?2eS`3`]Hl00P0000060?>fRP030000003b]X`0l[J<00<00000F@3oool2
000004L0oooo00240?ooo`030?oo_`0e000055o400<0oooo00<0a5lD0000=@2?ool0:03oool30000
0080oooo00<000000?ooo`3oool0W`3oool00`000000_?;b0;cblP050;cblP8000001`2ll_820000
00@0_O;b7`0000000`3:mN/000000?>fRP050?>fRP030000003b]X`0l[J<00<00000F`3oool00`00
0000oooo0?ooo`140?ooo`00Q03oool00`3ogi`0500003F?o`030?ooo`030?n?=@0001@0GlCo02D0
oooo0`0000050?ooo`030000003oool0oooo09d0oooo0P0000070;cblP030000002ll_80_?;b00H0
_?;b00<000000;gblP2ml_801P2ml_82000000H0__?a0P0000060<3cl00300000034m>l0a?C_00D0
a?C_00<000000<[ej`3:mN/00`3:mN/00`000000lkJ:0?>fRP050?>fRP030000003b]X`000000080
0000G03oool00`000000oooo0?ooo`130?ooo`00Q03oool00`3ogi`0500003F?o`030?ooo`030?n?
=@0001@0GlCo0240oooo1000002W0?ooo`030000002ll_80_?;b00H0_?;b00<000000;cblP2ll_80
1@2ll_82000000P0_O;b00<000000;kcl@2nlo401@2nlo400`000000`??`0<3cl0040<3cl0800000
1P34m>l2000000@0b_G[0`0000070?>fRP030000003b]X`0000000800000G@3oool2000004<0oooo
00240?ooo`030?oOW00D0000=Hoo00<0oooo00<0ohle0000501Oa?l07P3oool3000000<0oooo00<0
00000?ooo`3oool0X`3oool2000000L0_?;b0P0000070;cblP030000002ml_80_O;b00H0_O;b0P00
00060;kcl@8000001P30lo000`000000a?C_0<Cdk`050<Cdk`030000003:mN/0b_G[00<0b_G[00@0
00000=Weh@3ImN4000001`3c]XX00`000000l[J<00000002000005l0oooo00<000000?ooo`3oool0
@03oool008@0oooo00<0omnL01@0000eSol00`3oool00`3oScD0000D05o4o`0K0?ooo`<000001P3o
ool00`000000oooo0?ooo`2R0?ooo`030000002ll_80_?;b00H0_?;b00<000000;cblP2ll_801@2l
l_82000000P0_O;b00<000000;kcl@2nlo401@2nlo400`000000`??`0<3cl0050<3cl00300000034
m>l0a?C_00D0a?C_00<000000<[ej`3:mN/00`3:mN/01@000000fOGQ0=Weh@3ImN40000000L0lkJ:
00<000000?:fS00000000P00001P0?ooo`030000003oool0oooo03l0oooo00240?ooo`030?oo_`0e
000055o400<0oooo00<0a5lD0000=@2?ool05`3oool500000:/0oooo0P0000080;cblP030000002l
l_80_?;b00D0_?;b00<000000;gblP2ml_801P2ml_82000000H0__?a0P0000070<3cl00300000034
m>l0a?C_00@0a?C_0P0000050<[ej`030000003ImN40fOGQ0080fOGQ0P0000060?>fRP060000003b
]X`000000000003a]Hd00000H03oool2000003l0oooo00250?ooo`030<AO500055l0a?oo0080oooo
00<0ScD0000eS`3oool0503oool3000000@0oooo00<000000?ooo`3oool0Z03oool4000000D0_?;b
0P0000060;cblP800000202ml_800`000000__?a0;kcl@050;kcl@0300000030lo00`??`00H0`??`
00<000000<Cdk`34m>l01034m>l00`000000b_G[0<[ej`040<[ej`030000003ImN40fOGQ0080fOGQ
00<000000>ndT00000001P3c]XX01P000000l[J<000000000000lKF=00000680oooo00<000000?oo
o`3oool0?03oool008D0oooo00L0oon?03D0000eSol0oooo0?n?=@0003D0Sooo0180oooo0`000007
0?ooo`030000003oool0oooo0:H0oooo0P0000040;Wcla0000001@2ml_82000000H0__?a0P000007
0<3cl08000001P34m>l00`000000b_G[0<[ej`030<[ej`800000103ImN44000000H0lkJ:00H00000
0?:fS000000000000?6eS@00001S0?ooo`030000003oool0oooo03/0oooo00260?ooo`050?ooS`0e
000000000000=@2?ool03`3oool500000;40oooo00<000000;Wcl`2ilo<01P2ilo<2000000H0^?;c
0P0000040;Sbla0000001@30lo000`000000a?C_0<Cdk`040<Cdk`8000001@3:mN/00`000000fOGQ
0=Weh@0500000080coKX00<000000?>fRP3c]XX0103c]XX01P000000l[J<000000000000lKF=0000
06@0oooo0P00000k0?ooo`00U`3oool3000000@0oooo00<000000?ooo`3oool0[@3oool2000000P0
^O?c00<000000;Sbl`2hl_<01@2hl_<00`000000^?;c0;Sbl`060;Sbl`8000001P2il_<2000000@0
^_;c5P0000020<CekP030000003?m^P0coKX0080coKX00<000000?>fRP3c]XX0103c]XX01P000000
l[J<000000000000lKF=000006H0oooo00<000000?ooo`3oool0>03oool009D0oooo0P00002f0?oo
o`030000002ilo<0^O?c00H0^O?c0P0000070;Sbl`030000002hl_<0^?;c00H0^?;c00<000000;Wb
l`2il_<01@2il_<00`000000^_;c0;[bl`060;[bl`8000001@2llo82000000D0__Ca0P0000040<Ce
kP800000103?m^P2000000H0lkJ:00H000000?:fS000000000000?6eS@00001W0?ooo`030000003o
ool0oooo03L0oooo002E0?ooo`@00000/P3oool2000000P0^O?c00<000000;Sbl`2hl_<01@2hl_<2
000000P0^?;c00<000000;Wbl`2il_<0102il_<2000000L0^_;c0P0000060;cclP030000002nm?40
__Ca00@0__Ca00<000000<CekP34mNh00`34mNh00`000000coKX0<ofj0030<ofj0030000003^]940
000000H0lkJ:00H000000?:fS000000000000?6eS@00001X0?ooo`800000=`3oool009D0oooo00<0
00000?ooo`3oool00`00002_0?ooo`030000002ilo<0^O?c00H0^O?c0P0000070;Sbl`030000002h
l_<0^?;c00H0^?;c0P0000060;Wbl`030000002jl_<0^_;c00H0^_;c00<000000;cclP2llo801@2l
lo800`000000__Ca0;kdl@040;kdl@0300000034mNh0a?G^00<0a?G^00<000000<ofj03?m^P00`3?
m^P010000000k[BA0>jdT@0000060?>fRP060000003b]X`0l[J<0000003a]Hd00000JP3oool00`00
0000oooo0?ooo`0d0?ooo`00U@3oool00`000000oooo0?ooo`030?ooo`800000Z`3oool2000000P0
^O?c00<000000;Sbl`2hl_<01P2hl_<00`000000^?;c0;Sbl`060;Sbl`030000002il_<0^O;c00@0
^O;c0P0000070;[bl`8000001P2llo82000000D0__Ca0P0000050<CekP030000003?m^P0coKX00<0
coKX00D000000>jdT@3^]940k[BA000000060?>fRP060000003b]X`0l[J<0000003a]Hd00000J`3o
ool00`000000oooo0?ooo`0c0?ooo`00U@3oool00`000000oooo0?ooo`050?ooo`800000Z03oool4
000000D0^O?c0P0000070;Sbl`800000202hl_<00`000000^O;c0;Wbl`040;Wbl`030000002jl_<0
^_;c00D0^_;c0P0000070;cclP030000002nm?40__Ca00@0__Ca00<000000<CekP34mNh01034mNh0
0`000000coKX0<ofj0030<ofj0030000003^]940k[BA0080k[BA00<000000?>fRP3c]XX0103c]XX0
1P000000l[J<0?:fS0000000lKF=000006`0oooo0P00000c0?ooo`00U@3oool00`000000oooo0?oo
o`070?ooo`800000Y03oool2000000@0]_?c4@0000050;Sbl`8000001@2il_<2000000H0^_;c0P00
00080;cclP030000002nm?40__Ca00@0__Ca00<000000<CekP34mNh00`34mNh2000000D0coKX00@0
00000>jdT@3^]940k[BA0`0000060?>fRP060000003b]X`0l[J<0000003a]Hd00000KP3oool00`00
0000oooo0?ooo`0`0?ooo`00U@3oool00`000000oooo0?ooo`090?ooo`800000X@3oool00`000000
]_?c0;Kcl`060;Kcl`8000001`2elo@2000000@0]O;d5`0000050;cclP8000001@2nm?42000000D0
a?G^00<000000<ofj03?m^P0103?m^P8000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00001_
0?ooo`030000003oool0oooo02l0oooo002E0?ooo`030000003oool0oooo00/0oooo0P00002M0?oo
o`800000202flo<00`000000]O?d0;Gcm0050;Gcm0800000202el_@2000000D0]O;d0P0000060;Gb
m0800000102fl_@K00000080aoK/0P0000000`3^]9800000000000060?>fRP060000003b]X`0l[J<
0000003a]Hd00000L03oool2000002l0oooo002E0?ooo`030000003oool0oooo00d0oooo0P00002J
0?ooo`030000002flo<0]_?c00H0]_?c0P0000070;Gcm0030000002el_@0]O;d00H0]O;d0P000006
0;Gbm0030000002el_@0]O;d00D0]O;d00<000000;Kbm02fl_@01P2fl_@2000000D0^?;c0P000005
0;[clP8000001@2nm?02000000<0aoK/00@000000>jdTP3^]980k[BB0P0000060?>fRP060000003b
]X`0l[J<0000003a]Hd00000LP3oool00`000000oooo0?ooo`0/0?ooo`00U@3oool00`000000oooo
0?ooo`0?0?ooo`<00000U@3oool2000000L0]_?c0P0000070;Gcm0800000202el_@00`000000]O;d
0;Gbm0040;Gbm08000001`2el_@00`000000]_;d0;Kbm0050;Kbm08000001P2hl_<00`000000^_?b
0;[clP040;[clP030000002nm?00__C`00@0__C`00<000000<Ofk037m^`00P37m^`00`000000k[BB
0>jdTP020>jdTP8000001P3c]XX01P000000l[J<0?:fS0000000lKF=000007<0oooo00<000000?oo
o`3oool0:`3oool009@0oooo00<000000?ooo`3oool04`3oool200000980oooo00<000000;Kcl`2f
lo<01P2flo<00`000000]O?d0;Gcm0050;Gcm0800000202el_@2000000D0]O;d0P0000070;Gbm080
00001`2fl_@00`000000^?;c0;Sbl`040;Sbl`8000001P2jlo800`000000__C`0;kdl0040;kdl003
00000037m^`0aoK/0080aoK/00<000000>jdTP3^]9800`3^]982000000L0lkJ:00D000000?:fS000
0000lKF=0000001d0?ooo`800000:`3oool009@0oooo00<000000?ooo`3oool05@3oool2000008h0
oooo0P0000070;Kcl`8000001P2elo@2000000P0]O;d0P0000060;Gbm0030000002el_@0]O;d00H0
]O;d00<000000;Kbm02fl_@01@2fl_@2000000D0^?;c0P0000060;[clP8000001@2nm?02000000@0
aoK/00<000000>jdTP3^]980103^]982000000L0lkJ:00D000000?:fS0000000lKF=0000001f0?oo
o`030000003oool0oooo02P0oooo002D0?ooo`030000003oool0oooo01L0oooo0P00002;0?oooaD0
00001@2el_@2000000H0]O;d0P0000080;Gbm0030000002fl_@0]_;d00@0]_;d0P0000060;Sbl`03
0000002jlo80^_?b00D0^_?b00<000000;kdl02nm?00102nm?000`000000aoK/0<Ofk0020<Ofk080
00001`3^]982000000L0lkJ:00D000000?:fS0000000lKF=0000001g0?ooo`030000003oool0oooo
02L0oooo002D0?ooo`030000003oool0oooo01T0oooo0P0000270?ooo`8000001`2cm?<2000000H0
/_?d0P0000040;;bm10000001@2el_@2000000D0]_;d0P0000060;Sbl`8000001`2jlo800`000000
__C`0;kdl0040;kdl00300000037m^`0aoK/0080aoK/00<000000>jdTP3^]980100000030>jdTP80
00001`3c]XX01@000000l[J<0000003a]Hd0000007P0oooo0P00000W0?ooo`00U03oool00`000000
oooo0?ooo`0K0?ooo`800000Q03oool00`000000/oCc0;?dl`060;?dl`030000002blo@0/_?d00D0
/_?d00<000000;;bm02bl_@01P2bl_@2000000H0/_;e0P0000040;;bmAL00000102jlo82000000D0
__C`0P0000000`37m^`0000000000005000000040<;fkP0000000000000000<0k[BB0P0000070?>f
RP030000003b]X`000000080lKF=00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0T0?oo
o`00U03oool00`000000oooo0?ooo`0M0?ooo`800000P@3oool00`000000/oCc0;?dl`050;?dl`80
00001P2blo@2000000P0/_;d00<000000;;bm@2bl_D01@2bl_D00`000000/_;e0;;bm@060;;bm@80
00001@2bl_D2000000H0/_;e0P0000040;Cbm0h0000000<0^_Cb0000000000001032m^h010000000
kK>C0>fcT`0000030>jdTP030000003^]940000000H0lkJ:00<000000?:fS00000000P3a]Hd00`00
0000oooo0?ooo`1h0?ooo`030000003oool0oooo02<0oooo002D0?ooo`030000003oool0oooo01l0
oooo0P00001m0?ooo`8000001`2cm?<00`000000/_?d0;;cm0050;;cm0030000002bl_@0/_;d00H0
/_;d0P0000060;;bm@8000001`2bl_D2000000H0/_;e00<000000;;bm@2bl_D01@2bl_D00`000000
]?;d0;Cbm0050;Cbm08000001@2flo<2000000@0^_Cb00<000000<;fkP32m^h00`32m^h01@000000
kK>C0>fcT`3]/i<0000000<0k[BB0`0000060?>fRP030000003b]X`000000080lKF=00<000000?oo
o`3oool0N@3oool2000002<0oooo002D0?ooo`<000008@3oool2000007X0oooo00<000000;?dl`2c
m?<01@2cm?<2000000H0/_?d0P0000070;;bm08000001P2bl_D2000000L0/_;e0P0000060;;bm@80
00001P2bl_D2000000H0]?;d0P0000050;Kcl`800000102jm?82000000@0`_K^0P0000040>fcT`04
0000003^]980k[BB0>jdTP<000001P3c]XX00`000000l[J<000000020?6eS@030000003oool0oooo
07/0oooo00<000000?ooo`3oool0803oool009@0oooo00<000000?ooo`3oool08`3oool2000007L0
oooo00<000000;?dl`2cm?<01@2cm?<00`000000/_?d0;;cm0050;;cm0030000002bl_@0/_;d00H0
/_;d00<000000;;bm@2bl_D01@2bl_D00`000000/_;e0;;bm@050;;bm@8000001P2bl_D2000000H0
/_;e0P0000060;Cbm08000001@2flo<2000000@0^_Cb0P0000050<;fkP030000003]/i<0kK>C00@0
kK>C00@000000>jdTP3^]980k[BB0`0000060?>fRP030000003b]X`000000080lKF=00<000000?oo
o`3oool0O03oool00`000000oooo0?ooo`0O0?ooo`00U03oool00`000000oooo0?ooo`0U0?ooo`<0
0000LP3oool2000000H0/oCc0P0000060;;cm08000001`2bl_@2000000H0/_;e0P0000060;;bm@80
00001`2bl_D00`000000/_;e0;;bm@050;;bm@030000002dl_@0]?;d00@0]?;d0P0000050;Kcl`80
00001@2jm?800`000000`_K^0<;fkP040<;fkP030000003]/i<0kK>C00D0kK>C00@000000>jdTP3^
]980k[BB0`0000060?>fRP030000003b]X`000000080lKF=00<000000?ooo`3oool0O@3oool20000
01l0oooo002D0?ooo`030000003oool0oooo02P0oooo0P00001_0?ooob`000001@2bl_D2000000H0
/_;e0P0000050;Cbm08000001@2flo<2000000D0^_Cb0P0000050<;fkP80000000<0kK>C00000000
00000P0000030>fcT`040000003^]980k[BB0>jdTP<000001P3c]XX01P000000l[J<000000000000
lKF=00000840oooo00<000000?ooo`3oool0703oool009@0oooo00<000000?ooo`3oool0:P3oool2
000006/0oooo0P0000060;7dl`8000001P2`lo@2000000L0[o?d0P0000060:obm@8000001@2^l_D2
000000@0[_7e;0000000102mm_0000000000000000030>fcT`030000003^]980k[BB0080k[BB0P00
00060?>fRP060000003b]X`000000000003a]Hd00000PP3oool00`000000oooo0?ooo`0K0?ooo`00
U03oool00`000000oooo0?ooo`0/0?ooo`800000I`3oool2000000L0/OCc00<000000;3cm02`lo@0
1@2`lo@00`000000[o?d0:ocm0050:ocm08000001P2_l_D2000000H0[_;e00<000000:kam@2^lOD0
1P2^lOD2000000H0[_7f0P0000050:oamP8000001@2`l_D2000000D0/_;d0P0000050;Kcl`800000
102mm_0010000000eOKT0=Gfi00000030>fcT`030000003^]980k[BB0080k[BB0P0000060?>fRP06
0000003b]X`000000000003a]Hd00000P`3oool2000001/0oooo002D0?ooo`030000003oool0oooo
02h0oooo0P00001T0?ooo`030000002am?<0/OCc00D0/OCc0P0000060;3cm08000001`2_lo@00`00
0000[o;e0:obm@050:obm@030000002^l_D0[_;e00@0[_;e0P0000070:kam@8000001P2^lOH20000
00H0[o7f00<000000;3bm@2`l_D0102`l_D00`000000/_;d0;;bm0040;;bm0030000002flo<0]_?c
00@0]_?c00<000000;gfl02mm_000`2mm_001@000000eOKT0=Gfi03Em^@0000000<0kK>C00<00000
0>jdTP3^]9800P3^]982000000H0lkJ:00H000000?:fS000000000000?6eS@0000250?ooo`030000
003oool0oooo01P0oooo002D0?ooo`030000003oool0oooo0300oooo0P00001P0?ooo`8000001P2a
m?<2000000H0/??d0P0000070:ocm08000001P2_l_D2000000D0[_;e0P0000070:kam@8000001P2^
lOH2000000H0[o7f0P0000050;3bm@8000001@2bl_@2000000D0]_?c0P0000040;gfl08000000`3E
m^@2000000<0kK>C0P0000030>jdTP8000001P3c]XX01P000000l[J<000000000000lKF=000008H0
oooo00<000000?ooo`3oool05`3oool009@0oooo00<000000?ooo`3oool0<P3oool2000005`0oooo
0P0000070;7dl`030000002`lo@0/??d00D0/??d00<000000:ocm02_lo@01@2_lo@2000000H0[o;e
0P0000060:kbm@030000002^lOD0[_7e00D0[_7e0P0000060:kamP8000001P2_lOH2000000D0/?;e
0P0000050;;bm08000001@2flo<2000000D0_OK`00<000000=Gfi03Em^@00P3Em^@3000000<0kK>C
0P0000030>jdTP8000001P3c]XX01P000000l[J<000000000000lKF=000008L0oooo0P00000G0?oo
o`00U03oool00`000000oooo0?ooo`0d0?ooo`800000F03oool2000000L0/OCc0P0000060;3cm080
00001P2_lo@2000000H0[o;e0P0000060:kbm@8000001P2^lOD2000000H0[_7f0P0000070:oamP03
0000002`l_D0/?;e00@0/?;e00<000000;;bm02bl_@0102bl_@00`000000]_?c0;Kcl`040;Kcl`03
0000002mm_00_OK`00@0_OK`00<000000=Gfi03Em^@00P3Em^@2000000030=3gi`000000kK>C0080
kK>C0P0000030>jdTP8000001P3c]XX01P000000l[J<000000000000lKF=000008T0oooo00<00000
0?ooo`3oool0503oool009@0oooo00<000000?ooo`3oool0=P3oool2000005D0oooo?@0000040:oa
mP8000001@2`l_D2000000D0/_;d0P0000050;Kcl`8000001@2mm_02000000030=Gfi00000000000
008000000P3@mnL2000000<0kK>C0P0000030>jdTP8000001P3c]XX01P000000l[J<000000000000
lKF=000008X0oooo00<000000?ooo`3oool04`3oool009@0oooo00<000000?ooo`3oool0>03oool3
00000500oooo0P0000070:kelP8000001P2]m?<2000000H0[O?d0P0000060:cbm08000001P2[l_D2
000000H0Zo7f0P0000060:_amP800000102[lOHS000000040;Sel@00000000000=3gi`80000000<0
jk:F0000003]/i<00`3]/i<010000000k[BB0>jdTP3^]982000000L0lkJ:00D000000?:fS0000000
lKF=0000002;0?ooo`8000004`3oool009<0oooo00<000000?ooo`3oool0?03oool2000004d0oooo
00<000000:kelP2^mO801@2^mO82000000H0[OCc0P0000060:gcm08000001P2/l_@2000000H0Zo;e
0P0000060:_amP8000001P2[lOH2000000H0Zo7f0P0000060:camP8000001@2]lOH2000000D0[_;e
0P0000050;;cm08000000`2hmO4200000080d?OW00<000000>^bUP3[/YH00P0000040>fcT`040000
003^]980k[BB0>jdTP8000001`3c]XX01@000000l[J<0000003a]Hd0000008d0oooo00<000000?oo
o`3oool0403oool009<0oooo00<000000?ooo`3oool0?P3oool2000004T0oooo0P0000070:kelP03
0000002]m?<0[OCc00@0[OCc0P0000060:gcm08000001P2/l_@2000000H0Zo;e0P0000060:_amP80
00001P2[lOH2000000H0Zo7f0P0000060:camP8000001@2]lOH2000000D0[_;e0P0000050;;cm080
0000102hmO401P000000d?OW0=3gi`3@mnL000000>^bUP@00000103]/i<010000000k[BB0>jdTP3^
]982000000L0lkJ:00D000000?:fS0000000lKF=0000002>0?ooo`030000003oool0oooo00l0oooo
002C0?ooo`030000003oool0oooo0400oooo0P0000150?ooo`8000001`2^mO82000000D0[OCc0P00
00060:gcm08000001P2/l_@2000000H0Zo;e0P0000060:_amP8000001P2[lOH2000000H0Zo7f0P00
00070:camP030000002]lOH0[O7f00@0[O7f00<000000:kbm@2^l_D0102^l_D00`000000/_?d0;;c
m0040;;cm0030000002hmO40^?Ga0080^?Ga0P0000030=3gi`@0000000@0TfnW0000003[/YH00000
0`3]/i<010000000k[BB0>jdTP3^]982000000L0lkJ:00D000000?:fS0000000lKF=0000002?0?oo
o`8000003`3oool009<0oooo0P0000130?ooo`800000@P3oool00`000000[_Gb0:kelP050:kelP80
00001@2]m?<2000000H0[O?d0P0000060:cbm08000001P2[l_D2000000H0Zo7f0P0000060:_amP80
00001P2[lOH2000000L0[?7f0P0000050:gamP8000001@2^l_D2000000D0/_?d0P0000030;Sel@80
0000103@mnL2000000070<ogi`000000TfnW09=_Y`000000jk:F000000030>fcT`040000003^]980
k[BB0>jdTP8000001`3c]XX01@000000l[J<0000003a]Hd0000008l0oooo0`00000>0?ooo`00T`3o
ool00`000000oooo000000140?ooo`800000?P3oool2000000H0[_Gb0P0000020:gdlcL000001@2/
lOH2000000D0[O7f0P0000050:kbm@8000001@2blo@2000000030;Sel@000000000000P000000P3?
mnL01`000000TfnW09=_Y`2CKjL000000>^bUP0000000`3]/i<010000000k[BB0>jdTP3^]9820000
00L0lkJ:00<000000?:fS00000000P3a]Hd00`000000oooo0?ooo`280?ooo`@000000P3oool00`00
0000oooo0?ooo`0<0?ooo`00T`3oool00`000000oooo0?ooo`160?ooo`800000>`3oool=00000080
ZoCb0P0000060:[cl`8000001P2Yl_@2000000H0Z?;e0P0000060:Sam@8000001P2XlOH2000000D0
Z?3f0P0000040:S`mQd0000000<0[_;d000000000000102emO82000000030<ogi`000000000000@0
TfnW00<000000>^bUP0000000`3]/i<01`000000k[BB0>jdTP3^]98000000>jdT@0000001P3c]XX0
0`000000l[J<000000020?6eS@030000003oool0oooo08<0oooo1@0000060?ooo`030000003oool0
oooo00`0oooo002C0?ooo`030000003oool0oooo04P0oooo0P00000g0?ooo`8000001P2/mO420000
00H0ZoCb00<000000:[cl`2Zlo<0102Zlo<2000000H0ZO;d0P0000060:Sbm@8000001P2XlOD20000
00H0Z?7f0P0000050:S`mP8000001`2Xl?H2000000D0Z?3f0P0000050:W`mP8000001@2[lOH20000
00@0[_;d0P0000040;GelP8000000P3?mnL00`000000TfnW09=_Y`0409=_Y`030000003[/YH00000
00<0kK>C00L000000>jdTP3^]980k[BB0000003^]940000000H0lkJ:00<000000?:fS00000000P3a
]Hd00`000000oooo0?ooo`1o0?ooo`@000002`3oool00`000000oooo0?ooo`0<0?ooo`00T`3oool0
0`000000oooo0?ooo`1:0?ooo`800000<`3oool2000000H0[?Ga0P0000060:_dlP8000001@2Zlo<2
000000H0ZO;d0P0000060:Sbm@8000001P2XlOD2000000H0Z?7f0P0000060:S`mP030000002Xl?H0
Z?3f00D0Z?3f0P0000050:S`mP8000001@2Yl?H2000000D0Zo7f0P0000040:kbm08000001@2emO80
1@000000coOW0<ogi`3?mnL0000000L0TfnW00<000000>^bUP0000000`3]/i<010000000k[BB0>jd
TP3^]983000000H0lkJ:00<000000?:fS00000000P3a]Hd00`000000oooo0?ooo`1k0?ooo`@00000
3`3oool00`000000oooo0?ooo`0<0?ooo`00T`3oool00`000000oooo0?ooo`1<0?ooo`800000<03o
ool00`000000[?Ga0:cel@040:cel@8000001P2[m?82000000D0Z_?c0P0000060:Wbm08000001P2X
l_D2000000H0Z?7e0P0000060:SamP8000001P2Xl?H2000000H0Z?3f0P0000050:S`mP8000001@2Y
l?H2000000D0Zo7f0P0000040:kbm08000001@2emO82000000<0coOW00<0000009=_Y`2CKjL01P2C
KjL00`000000jk:F000000030>fcT`040000003^]980k[BB0>jdTP<000001P3c]XX00`000000l[J<
000000020?6eS@030000003oool0oooo07H0oooo1@00000C0?ooo`030000003oool0oooo00`0oooo
002C0?ooo`030000003oool0oooo04h0oooo0`00000[0?ooo`8000001@2/mO42000000L0ZoCb00<0
00000:[cl`2Zlo<00`2Zlo<2000000H0ZO;d0P0000060:Sbm@8000001P2XlOD2000000H0Z?7f0P00
00060:S`mP8000001P2Xl?H2000000D0Z?3f0P0000050:W`mP8000001@2[lOH2000000@0[_;d0P00
00050;GelP80000000<0coOW0000000000000P00000909=_Y`030000003[/YH0000000<0kK>C00@0
00000>jdTP3^]980k[BB0`0000060?>fRP030000003b]X`000000080lKF=00<000000?ooo`3oool0
LP3oool4000001P0oooo00<000000?ooo`3oool0303oool009<0oooo00<000000?ooo`3oool0D@3o
ool2000002L0oooo0P0000050:cel@8000001`2[m?8200000080Z_?cF@000000102bm?<000000000
0000000909=_Y`030000003[/YH0000000<0kK>C00@000000>jdTP3^]980k[BB0`0000060?>fRP06
0000003b]X`000000000003a]Hd00000K`3oool5000001`0oooo00<000000?ooo`3oool0303oool0
09<0oooo00<000000?ooo`3oool0D`3oool2000002<0oooo0P0000020:celA80000000<0Yo?b0000
000000001P2Wl_<2000000H0Y_;d0P0000060:Gam@8000001P2UlOH2000000H0YO3f0P0000050:G`
mP8000001@2Ul?H2000000D0YO3g0P0000050:K`mP8000001@2Wl?H2000000D0Zo;e0P0000030;;d
l`8000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`030000003[/YH0000000<0kK>C00@000000>jd
TP3^]980k[BB0`0000060?>fRP060000003b]X`000000000003a]Hd00000J`3oool400000240oooo
00<000000?ooo`3oool0303oool009<0oooo00<000000?ooo`3oool0E@3oool200000200oooo1@00
00020:Wfk`8000001`2Xm?42000000D0Yo?b00<000000:Obl`2Wl_<0102Wl_<2000000H0Y_;d0P00
00060:Gam@8000001P2UlOH2000000H0YO3f0P0000050:G`mP8000001@2Ul?H2000000D0YO3g0P00
00050:K`mP8000001@2Wl?H2000000D0Zo;e0P0000030;;dl`800000103Z/YL00`000000TfnW09=_
Y`0709=_Y`<000000`3]/i<010000000k[BB0>jdTP3^]983000000H0lkJ:00H000000?:fS0000000
00000?6eS@00001V0?ooo`D000009@3oool00`000000oooo0?ooo`0<0?ooo`00T`3oool00`000000
oooo0?ooo`1G0?ooo`800000703oool2000000H0ZOK_00<000000:Sdl@2Xm?401@2Xm?42000000D0
Yo?b0P0000050:Obl`8000001`2Vl_@00`000000YO7e0:Gam@040:Gam@8000001P2UlOH2000000H0
YO3f0P0000050:G`mP8000001P2Ul?H00`000000YO3g0:G`m`040:G`m`030000002Vl?H0Y_3f00@0
Y_3f00<000000:O`mP2Wl?H0102Wl?H00`000000Zo;e0:_bm@040:_bm@040000002bm?<0/_Cc0;;d
l`8000001P3Z/YL00`000000TfnW09=_Y`0709=_Y`<000000`3]/i<010000000k[BB0>jdTP3^]983
000000H0lkJ:00H000000?:fS000000000000?6eS@00001R0?ooo`@00000:P3oool00`000000oooo
0?ooo`0<0?ooo`00T`3oool00`000000oooo0?ooo`1I0?ooo`8000006@3oool00`000000ZOK_0:Wf
k`040:Wfk`8000001P2Xm?42000000D0Yo?b0P0000050:Obl`8000001`2Vl_@2000000D0YO7e0P00
00060:GamP8000001P2Ul?H2000000D0YO3f0P0000060:G`mP8000001@2Ul?L2000000D0Y_3f0P00
00050:O`mP8000001@2[l_D200000080/_Cc0P0000000`3Z/YL0000000000002000000<0j[:G00<0
000009=_Y`2CKjL01`2CKjL3000000<0kK>C00@000000>jdTP3^]980k[BB0`0000060?>fRP060000
003b]X`000000000003a]Hd00000GP3oool4000002l0oooo00<000000?ooo`3oool02`3oool009<0
oooo00<000000?ooo`3oool0F`3oool2000001D0oooo0P0000050:Wfk`8000001P2Xm?42000000H0
Yo?b00<000000:Obl`2Wl_<00`2Wl_<2000000L0Y_;d0P0000050:Gam@8000001P2UlOH2000000H0
YO3f0P0000050:G`mP8000001P2Ul?H2000000D0YO3g0P0000050:K`mP8000001@2Wl?H2000000D0
Zo;e0P0000000`2bm?<0000000000004000000040:odl`0000000000000000<0j[:G00<0000009=_
Y`2CKjL01`2CKjL3000000<0kK>C00@000000>jdTP3^]980k[BB0`0000060?>fRP060000003b]X`0
00000000003a]Hd00000F@3oool5000003<0oooo00<000000?ooo`3oool02`3oool009<0oooo00<0
00000?ooo`3oool0G@3oool200000140oooo0P0000060:Wfk`030000002Xm?40Z?Ca00@0Z?Ca0P00
00060:OclP8000000P2Wl_=D000000<0[oCc0P0000020>VbV0050000003Z/YL0j[:G0>ZbU`000000
2@2CKjL3000000<0kK>C0P0000020>jdTP<000001P3c]XX01P000000l[J<000000000000lKF=0000
05D0oooo1000000h0?ooo`030000003oool0oooo00/0oooo002C0?ooo`030000003oool0oooo05l0
oooo0P00000>0?ooo`030000002Ym^l0ZOK_00@0ZOK_0P0000020:SdlA40000000<0Y??b00000000
00001`2Sl_<2000000D0Xo7d0P0000060:;`m@8000001@2Rl?H2000000H0X^of0P0000050:7_mP80
00001P2RkoL2000000D0X^og0P0000050:?_m`8000001@2Tl?H200000080Yo7e0P0000040:odl`03
0000003Y/YP0jK:H0080jK:H00D000000>ZbU`3Z/YL0j[:G0000000:09=_Y`8000000`3]/i<20000
0080k[BB0`0000060?>fRP060000003b]X`000000000003a]Hd00000D03oool5000003`0oooo00<0
00000?ooo`3oool02`3oool009<0oooo00<000000?ooo`3oool0H@3oool3000000T0oooo0P000002
0:Wfk`T000000P2Um>l2000000H0YO?a0P0000040:CclP8000001`2Sl_<2000000D0Xo7d0P000006
0:;`m@8000001@2Rl?H2000000H0X^of0P0000050:7_mP8000001P2RkoL2000000D0X^og0P000005
0:?_m`8000001@2Tl?H200000080Yo7e0P0000040:odl`8000001@3Y/YP01@000000j[:G0>ZbU`3Z
/YL0000000X0TfnW0P0000030>fcT`8000000P3^]983000000L0lkJ:0`0000000`3a]Hd000000?oo
o`1;0?ooo`@00000@@3oool00`000000oooo0?ooo`0;0?ooo`00T`3oool00`000000oooo0?ooo`1T
0?ooo`8000001P3oool500000080Y_K/0P0000060:Gdk`030000002Ulo40YO?a00@0YO?a0P000005
0:CclP030000002Sl_<0Xo;c00D0Xo;c0P0000050:?am08000001P2Rl?D2000000D0X_3f0P000006
0:;_mP8000001@2QkoH2000000H0X^og0P0000050:;_m`8000001@2SkoL2000000D0Y?3f0P000002
0:Oam@800000102_m?<2000000040>VbV00000000000000000<0jK:H00D000000>ZbU`3Z/YL0j[:G
0000000:09=_Y`8000000`3]/i<200000080k[BB0`0000070?>fRP<0000000<0lKF=0000003oool0
AP3oool5000004D0oooo00<000000?ooo`3oool02`3oool00980oooo0`00001W0?ooo`8000000P3o
ool2000000H0Y_K/00<000000:Gdk`2Um>l0102Um>l2000000D0YO?a0P0000050:CclP8000001P2S
l_<2000000D0Xo7d0P0000060:;`m@8000001@2Rl?H2000000H0X^of0P0000050:7_mP8000001P2R
koL2000000D0X^og0P0000050:?_m`8000001@2Tl?HB000000<0jK:H00D000000>ZbU`3Z/YL0j[:G
0000000:09=_Y`8000000`3]/i<200000080k[BB0`0000070?>fRP<0000000<0lKF=0000003oool0
@P3oool4000004X0oooo00<000000?ooo`3oool02`3oool00980oooo00<000000?ooo`3oool0J@3o
ool2000000H0Y_K/0P0000050:Gdk`8000001@2Ulo42000000D0Y??b0P0000060:?bl`8000000P2S
lOA4000000D0YO7e0P0000030:kdl`040000003Y/IP0jK6H000000<0jK:H00D000000>ZbU`3Z/YL0
j[:G0000000:09=_Y`8000000`3]/i<200000080k[BB0`0000070?>fRP<0000000<0lKF=0000003o
ool0?@3oool5000004h0oooo00<000000?ooo`3oool02`3oool00980oooo00<000000?ooo`3oool0
J03oool2000000D0Y_K/0P0000060:Gdk`030000002Ulo40YO?a00<0YO?a0P0000020:CclQ400000
00<0X?7c0000000000001@2Ol?@2000000H0Wo3e0P00000609k_mP8000001@2NkoH2000000@0W^kf
0`00000509k^m`8000001@2Ok_L2000000D0Wnog0P0000030:7_mP8000001@2UlOD2000000<0[_Cc
0P0000030>VaV0050000003Y/YP0jK:H0>VbV00000000`3Z/YL00`000000TfnW09=_Y`0809=_Y`80
00000`3]/i<2000000<0k[BB0P0000070?>fRP<0000000<0lKF=0000003oool0>@3oool4000005<0
oooo00<000000?ooo`3oool02`3oool00980oooo00<000000?ooo`3oool0IP3oool2000000H0Y_K/
00<000000:Gdk`2Um>l0102Um>l200000080YO?a200000020:7cl@8000001@2Ql_82000000D0X?7c
0P00000509o`m08000001P2Ol?D2000000H0W^of0P00000509k_mP800000102Nk_H2000000D0W^kg
0`00000409o^m`<00000102OkoL3000000<0XNof0P0000050:Gam@800000102^m?<010000000jK6H
0000000000020>VaV0050000003Y/YP0jK:H0>VbV00000000`3Z/YL00`000000TfnW09=_Y`0809=_
Y`030000003[/YH000000080kK>C0P0000030>jdTP8000001`3c]XX3000000030?6eS@000000oooo
03D0oooo1000001G0?ooo`030000003oool0oooo00/0oooo002B0?ooo`030000003oool0oooo06D0
oooo00<000000:Kfk02Vm^`0102Vm^`200000080YOC_2@0000000`2Rlnl00000000000050:7cl@80
00001@2Ql_82000000D0X?7c0P00000509o`m08000001P2Ol?D2000000H0W^of0P00000509k_mP80
00000`2Nk_H3000000D0W^kg0P00000509o^m`8000001@2OkoL2000000@0XNof0P0000050:Gam@80
0000102^m?<600000080jK6H00D000000>VbV03Y/YP0jK:H000000030>ZbU`030000002CKjL0TfnW
00P0TfnW00<000000>^bUP0000000P3]/i<2000000<0k[BB0P0000070?>fRP<000000P3a]Hd00`00
0000oooo0?ooo`0^0?ooo`D00000F`3oool00`000000oooo0?ooo`0;0?ooo`00TP3oool00`000000
oooo0?ooo`1S0?ooo`8000000P2Vm^`900000080XoC/0P0000040:;ck`8000001@2Qlo42000000D0
XO;b0P0000050:3al`8000001@2Ol?@2000000H0Wo3e0P00000609k_mP8000001@2NkoHc00000080
j;6I00@000000>VaV03Y/IP000000`3Y/YP01@000000j[:G0>ZbU`3Z/YL0000000X0TfnW00<00000
0>^bUP0000000P3]/i<2000000<0k[BB00<000000>jdT@0000001P3c]XX300000080lKF=00<00000
0?ooo`3oool0:P3oool400000600oooo00<000000?ooo`3oool02`3oool00980oooo00<000000?oo
o`3oool0HP3oool500000080XoKY0P0000050:?dk08000001@2Rlnl00`000000XO?a0:7cl@030:7c
l@8000001@2Ql_82000000D0X?7c0P00000209o`m1h0000000<0Vnkf0000000000001P2Kk_H20000
00D0Vnkg0P00000509c^m`800000102Mk_L2000000D0W^of0P0000040:;`m@8000000P2^m?800`00
0000j;6I0>RaV@020>RaV@040000003Y/IP0jK6H000000<0jK:H0P0000020>ZbU`030000002CKjL0
TfnW00P0TfnW00<000000>^bUP0000000P3]/i<2000000<0k[BB00<000000>jdT@0000001P3c]XX3
00000080lKF=00<000000?ooo`3oool09@3oool5000006@0oooo00<000000?ooo`3oool02`3oool0
0980oooo00<000000?ooo`3oool0H03oool2000000D0XoKY0P0000050:?dk08000001@2Rlnl20000
00@0XO?a0P0000020:7blQ00000000<0W?3c0000000000001P2Lko@2000000@0W>oe0`00000509_^
m@800000102Kk_H2000000H0Vnkf0P00000509_^m`8000001@2Lk_L2000000@0WNkg0P00000509k_
mP8000000`2Rl?D300000080[_Cb0P0000000`3X/IT00000000000020>RaV@040000003Y/IP0jK6H
000000<0jK:H0P0000030>ZbU`030000002CKjL0TfnW00L0TfnW00<000000>^bUP0000000P3]/i<2
000000<0k[BB0`0000060?>fRP<000000P3a]Hd00`000000oooo0?ooo`0Q0?ooo`@00000J@3oool0
0`000000oooo0?ooo`0;0?ooo`00TP3oool00`000000oooo0?ooo`1O0?ooo`030000002Sm^T0XoKY
00<0XoKY0P0000050:?dk08000001@2Rlnl200000080XO?a200000000`2Ml_0000000000000609ga
lP800000102Ll?<2000000H0W>od0P00000409c_m@8000001@2Kk_D3000000@0Vnkf0P00000609_^
mP8000001@2Kk_L2000000D0W>kg0P00000409g^m`8000001@2NkoH2000000<0X_3e0P0000030:kd
lPL000000P3X/IT010000000jK6H0>VaV00000030>VbV08000000`3Z/YL00`000000TfnW09=_Y`07
09=_Y`030000003[/YH000000080kK>C0P0000030>jdTP<000001P3c]XX2000000040?:fS0000000
lKF=000001h0oooo1@00001]0?ooo`030000003oool0oooo00/0oooo002B0?ooo`030000003oool0
oooo05d0oooo0P0000040:?fj@8000001@2Sm>`200000080X_?_2@0000000`2Nl^l0000000000004
09gbl08000001P2MlO82000000@0W?3c0P00000609c_m08000000`2LkoD3000000D0Vnke0P000005
09_^mP8000001P2Kk_H2000000D0Vnkg0P00000509c^m`800000102Mk_L2000000D0W^of3@000000
0`2cmO4000000000000209AaZ0040000003X/IT0j;6I00000080jK6H00@000000>VbV03Y/YP0jK:H
0P0000030>ZbU`030000002CKjL0TfnW00L0TfnW00<000000>^bUP0000000`3]/i<010000000k[BB
0>jdTP3^]983000000H0lkJ:0P000000103b]X`000000?6eS@00000J0?ooo`@00000LP3oool00`00
0000oooo0?ooo`0;0?ooo`00TP3oool00`000000oooo0?ooo`1K0?ooo`800000102Sm^T200000080
XoC/2@00000209ock0800000102Nl^l2000000D0WO;`00<0000009galP2MlO80102MlO82000000@0
W?3c0P00000609c_m3/0000000<0Vnkf0000000000000`2Pl?D2000000<0/oGa00<0000009AaZ02D
LJP00P2DLJP010000000j;6I0>RaV@0000020>VaV0040000003Y/YP0jK:H0>VbV08000000`3Z/YL0
0`000000TfnW09=_Y`0709=_Y`030000003[/YH0000000<0kK>C00@000000>jdTP3^]980k[BB0`00
00060?>fRP80000000@0l[J<0000003a]Hd000005P3oool4000007H0oooo00<000000?ooo`3oool0
2`3oool00980oooo00<000000?ooo`3oool0FP3oool010000000XoKY0:?fj@2Sm^T700000080WoCY
0P00000509ock0800000102Nl^l2000000D0WO;`0P00000509galP80000000<0W?3c000000000000
3@0000000`2Ik_@000000000000609W^m08000001@2Hk_D2000000D0V>gf0P00000509S]mP800000
1@2HkOH2000000D0VNgf0P00000409[]mP800000102Kk_H2000000<0X?3e0P0000030;?el@800000
1@2DLJP010000000j;6I0>RaV@0000020>VaV08000000P3Y/YP2000000<0j[:G00<0000009=_Y`2C
KjL01`2CKjL00`000000jk:F000000030>fcT`040000003^]980k[BB0>jdTP<000001P3c]XX20000
00040?:fS0000000lKF=00000140oooo1@00001j0?ooo`030000003oool0oooo00/0oooo002B0?oo
o`030000003oool0oooo05P0oooo2000000509odj@8000001@2Oln`2000000@0W_;_0P00000509gb
l08000000P2MlO880000000309[`lP000000000000H0VNoc0P00000309W^m08000001P2Ik_@20000
00D0V>ke0P00000409S]mP<00000102HkOH3000000@0V>gf0`00000409W]mP<00000102JkOH20000
00@0Vnkf0P0000030:3`m@8000000`2cmO42000000L0U76X00@000000>RaV@3X/IT000000P3Y/IP2
000000<0jK:H00D000000>ZbU`3Z/YL0j[:G0000000909=_Y`030000003[/YH0000000<0kK>C00@0
00000>jdTP3^]980k[BB0`0000060?>fRP80000000@0l[J<0000003a]Hd000003@3oool4000007l0
oooo00<000000?ooo`3oool02`3oool00980oooo00<000000?ooo`3oool0E`3oool3000000<0WoGU
00<0000009odj@2Om>T00`2Om>T2000000D0Wo?/0P00000209kbk`l0000000<0V_3`000000000000
102Jl?82000000H0VNoc0P00000309W^m08000001P2Ik_@2000000D0V>ke0P00000409S]mP800000
1@2HkOH2000000D0V>gf0P00000509W]mP8000001@2JkOH2000000@0Vnkf3000000909AaZ0040000
003X/IT0j;6I00000080jK6H0P0000030>VbV0050000003Z/YL0j[:G0>ZbU`0000002@2CKjL00`00
0000jk:F000000030>fcT`040000003^]980k[BB0>jdTP<000001P3c]XX2000000040?:fS0000000
lKF=000000P0oooo1@0000230?ooo`030000003oool0oooo00/0oooo002B0?ooo`030000003oool0
oooo05D0oooo0P00000409oei@800000102Om>T2000000D0Wo?/2000000609[ak`800000102Jl?02
000000@0V_3b0P00000609W_l`8000000`2Ik_@2000000H0VNkd0P00000509S^m@80000000<0V>gf
000000000000:@00000409o`m@80000000<0ik6J0000002DLJP0202DLJP010000000j;6I0>RaV@00
00020>VaV08000000`3Y/YP01@000000j[:G0>ZbU`3Z/YL0000000T0TfnW0`0000030>fcT`040000
003^]980k[BB0>jdTP<000001P3c]XX2000000040?:fS0000000lKF=000000@0oooo100000290?oo
o`030000003oool0oooo00X0oooo002B0?ooo`030000003oool0oooo05@0oooo00<0000009oei@2O
mND00P2OmND2000000@0WoCY0P00000209ock0L000000P2Kl^`2000000H0V_7_0P00000409[`l080
0000102Jl?82000000H0VNoc6@0000000`2EkOD000000000000409G/m@8000001@2Ek?H2000000D0
UNcf0P00000509K/mP800000102Gk?H200000080VNgf0`00000409o`m@8000000`3W/IX00`000000
U76X09AaZ00709AaZ0040000003X/IT0j;6I00000080jK6H0P0000030>VbV0050000003Z/YL0j[:G
0>ZbU`0000002@2CKjL3000000<0kK>C00@000000>jdTP3^]980k[BB0`0000060?>fRP80000000<0
l[J<0000003a]Hd01@00002=0?ooo`030000003oool0oooo00X0oooo002B0?ooo`030000003oool0
oooo0580oooo0P00000409oei@040000002Om>T0WoCY09odj@L000000P2KlnT2000000<0Vo;/0P00
000609[ak`800000102Jl?020000000309[`lP000000000000d0000000<0U^kc0000000000001@2F
k_<2000000D0U^gd0P00000509G]m@8000000`2Ek?D3000000@0UNcf0`00000409G/mP<00000102F
k?H3000000@0Uncf0P00000209W]mP8000001@2Ol?D500000080ik6J00<0000009AaZ02DLJP01`2D
LJP00`000000j;6I0>RaV@02000000030>VaV0000000000000<0jK:H00D000000>ZbU`3Z/YL0j[:G
0000000909=_Y`<000000`3]/i<010000000k[BB0>jdTP3^]983000000L0lkJ:00D000000?:fS000
0000lKF=0000002A0?ooo`030000003oool0oooo00X0oooo002B0?ooo`<00000D03oool2000000@0
WoGU2000000509_cj@8000000`2Kl^`2000000H0V_7_0P0000000`2Jl?0000000000000600000003
09O_l0000000000000D0U^oa0P00000409K^l`800000102Fk_<3000000D0U^gd0P00000509G]m@80
00000`2Ek?D2000000D0UNcf0P00000509G/mP8000001@2Fk?H2000000D0Uncf4P0000020>NaVP03
0000002DLJP0U76X00L0U76X00<000000>RaV@3X/IT00P0000020>VaV0050000003Y/YP0jK:H0>Vb
V00000000`3Z/YL00`000000TfnW09=_Y`0709=_Y`<000000`3]/i<200000080k[BB0`0000070?>f
RP050000003b]X`000000?6eS@000000T@3oool00`000000oooo0?ooo`0:0?ooo`00TP3oool00`00
0000oooo0?ooo`1?0?ooo`040000002OmND0WoGU09oei@D000000P2KlnD2000000D0Vo?Y0P000003
09_bk08000000P2JlNl90000000309O`kP000000000000@0Uno`0P00000409K_l@<00000102Fk_<2
000000@0U^kc0P00000609K]m08000001@2EkODV000000D0U^gf0P00000309o`m@8000000P3V/9/0
10000000ik6J0>NaVP00000909AaZ0030000003X/IT0j;6I008000000P3Y/IP01@000000jK:H0>Vb
V03Y/YP0000000<0j[:G00<0000009=_Y`2CKjL01`2CKjL3000000<0kK>C0P0000020>jdTP<00000
1`3c]XX01@000000l[J<0000003a]Hd000000940oooo00<000000?ooo`3oool02P3oool00940oooo
00<000000?ooo`3oool0CP3oool8000000@0Vo?U00<0000009_cj@2KlnT00`2KlnT20000000309_b
k0000000000000H0000000<0Uo7/0000000000001@2Gl>h2000000@0Uno`0P00000409K_l@800000
1@2Fk_<20000000309K^l`000000000001H000001@2BjoD2000000D0T^_f0P00000509;[mP800000
102CjoH300000080U>_f0P00000409K]mP<000000`2Ol?D400000080i[2K00@000000>NaVP3W/IX0
00002@2DLJP00`000000j;6I0>RaV@0200000080jK6H00D000000>VbV03Y/YP0jK:H000000030>Zb
U`030000002CKjL0TfnW00L0TfnW0`0000030>fcT`8000000P3^]983000000L0lkJ:00D000000?:f
S0000000lKF=0000002A0?ooo`030000003oool0oooo00X0oooo002A0?ooo`030000003oool0oooo
04d0oooo0`00000309[cgP030000002KlnD0Vo?U0080Vo?U0P00000209_cj@L0000000<0Uo7Y0000
00000000102GlN`2000000D0Uo3^0P00000409O_l080000000<0U^oa0000000000003@0000000`2C
kO8000000000000509;]l`8000001@2Bk?@200000080T^cd0`00000509;[m@8000001@2BjoH20000
00@0T^_f0`00000209?[mP@000000`2DjoH2000000@0U^gf0P00000409o`m@H000000P3V/9/01000
0000ik6J0>NaVP00000909AaZ0030000003X/IT0j;6I008000000P3Y/IP01@000000jK:H0>VbV03Y
/YP0000000<0j[:G00<0000009=_Y`2CKjL01`2CKjL3000000<0kK>C0P0000020>jdTP<000001`3c
]XX01@000000l[J<0000003a]Hd000000940oooo00<000000?ooo`3oool02P3oool007H0oooo00<0
oooO09a8000000000P0000000`1d_ol0oooo0?ooo`030?ooo`050?oOW018000000000000M02?ool0
2`3oool00`000000oooo0?ooo`1;0?ooo`8000001@2Jlmh010000000Vo?U09_ci@2KlnD8000000@0
Uo7Y0P00000409Oak08000001@2Gl>h20000000309O_l0000000000000H0000000<0Tnk`00000000
0000102Ck_42000000D0Tngb0P00000409;]l`<00000102Bk?@300000080T^cd0P00000609;[m@80
00001@2BjoH2000000@0T^_f7@00000209EbZ0040000003V/9/0i[2K00000080ik6J00<0000009Aa
Z02DLJP01`2DLJP2000000030>RaV@00000000000080jK6H00D000000>VbV03Y/YP0jK:H00000003
0>ZbU`030000002CKjL0TfnW00L0TfnW0`0000030>fcT`8000000P3^]983000000L0lkJ:00D00000
0?:fS0000000lKF=0000002A0?ooo`030000003oool0oooo00X0oooo001f0?ooo`030?oog`2LB7@0
_ooo0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3oool0okmd0000=@2?ool0
2P3oool00`000000oooo0?ooo`1:0?ooo`030000002Jlmh0V_?N00<0V_?N0P00000209_ci@@00000
0P2Gl^D2000000D0Uo7Y00<0000009Oak02GlN`00P2GlN`20000000309O`kP000000000000L00000
00<0Tno^000000000000102Ck_02000000@0Tnka0P00000509?]lP800000102BkO<2000000D0T^cd
80000005093ZmP800000102AjoH2000000@0U>cf0`0000000`2TlO@000000000000409EbZ0040000
003V/9/0i[2K00000080ik6J00<0000009AaZ02DLJP01`2DLJP2000000030>RaV@00000000000080
jK6H00D000000>VbV03Y/YP0jK:H000000030>ZbU`030000002CKjL0TfnW00L0TfnW0`0000030>fc
T`8000000P3^]983000000L0lkJ:00D000000?:fS0000000lKF=0000002A0?ooo`030000003oool0
oooo00X0oooo001k0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@00018W03Oool00P3oool00`2?
=@0003F?0?ooo`090?ooo`030000003oool0oooo04T0oooo00<0000009[cgP2Jlmh00`2Jlmh00`00
0000Vo?U00000003000000<0Uo;U0P00000509Oaj@8000000P2GlN`600000080U>o/0P00000409?_
kP800000102Ck_02000000@0Tnka0P00000509?]lP80000000<0T^gc0000000000003P00000608o[
m08000001@2?j_@2000000<0Sn[e0`00000308oZm@8000001@2@j_H2000000@0TN_f0P00000309C/
mP<000000P2TlO@2000000H0UG:X00@000000>J`V`3V/9/000000P3W/IX00`000000U76X09AaZ007
09AaZ08000000P3X/IT010000000jK6H0>VaV00000030>VbV0050000003Z/YL0j[:G0>ZbU`000000
2@2CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]983000000L0lkJ:00<000000?:fS000
00000P3a]Hd00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo00X0oooo001l0?ooo`060;md
001d_ol0oooo0?oo_`1d0000B9cO0`3oool00`34Ga@0000e08ooo`090?ooo`030000003oool0oooo
04L0oooo0P00000409[cgPH000000`2Gl^D2000000D0Uo7Y1`00000509C_k0800000102Cknh20000
00@0Tnk`0P0000000`2Ck_4000000000000?000000D0Sn_c0P00000308o[l`8000001@2?jo@30000
00@0Sn[d0`00000208oZm@<00000102?j_D2000000D0T>[f0P000004097[mP`00000202ELZP01000
0000i[2K0>J`V`0000020>NaVP030000002DLJP0U76X00L0U76X0P0000020>RaV@040000003Y/IP0
jK6H000000<0jK:H00D000000>ZbU`3Z/YL0j[:G0000000909=_Y`80000000@0jk:F0000003]/i<0
kK>C0P0000020>jdTP80000000<0k[BA0000003c]XX01@3c]XX00`000000l[J<000000020?6eS@03
0000003oool0oooo08h0oooo00<000000?ooo`3oool02P3oool007/0oooo00L0oooO09a80018W=l0
oooo0?oOW0180000M;oo00<0oooo00<0ohle0000501Oa?l02@3oool00`000000oooo0?ooo`160?oo
o`040000002Jlmh0V_?N09[cgP@0000000<0U_;O0000000000000`2Gl^D200000080Uo7Y1`000000
0`2Cl>T000000000000509C_k0800000102Cknh20000000309?^l0000000000000H0000000<0T>g`
000000000000102@k?42000000<0T>cb0`00000508o[l`8000000`2?jo<2000000D0Sn_d0P000005
08oZm1l00000102Ck?D2000000030>F`W0000000UG:X00L0UG:X00@000000>J`V`3V/9/000000P3W
/IX00`000000U76X09AaZ00709AaZ08000000P3X/IT010000000jK6H0>VaV00000030>VbV0050000
003Z/YL0j[:G0>ZbU`0000002@2CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]9820000
00030>jdT@000000lkJ:00D0lkJ:00<000000?:fS00000000P3a]Hd00`000000oooo0?ooo`2>0?oo
o`030000003oool0oooo00X0oooo001k0?ooo`070=nLB0000000M;oo0?ooo`3ogi`0B00007Boo`03
0?ooo`030?n?=@0001@0GlCo00T0oooo00<000000?ooo`3oool0A03oool80000000309Kbg`000000
00000080Uo;U1`00000209?`i@8000000P2Cl>T2000000D0U>o/0P0000000`2Cknh0000000000006
00000003093^kP000000000000@0T>g`0P000003093/l@<000000`2@k?82000000H0Sn_c6@000005
08cYm0800000102<jOD3000000<0SNWe0`00000208kYm@<00000102Ck?D2000000<0iK2L00<00000
09EbZ02ELZP01P2ELZP010000000i[2K0>J`V`0000020>NaVP030000002DLJP0U76X00L0U76X0P00
00020>RaV@040000003Y/IP0jK6H000000<0jK:H00D000000>ZbU`3Z/YL0j[:G0000000909=_Y`80
000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP80000000<0k[BA0000003c]XX01@3c]XX01P00
0000l[J<0?:fS0000000lKF=00000900oooo00<000000?ooo`3oool02P3oool007L0oooo00/0gmoO
0?ooo`3ookl0M0000000000007@0_ooo0?ooo`3ogi`0B00007Boo`030?ooo`030?n?=@0001@0GlCo
00T0oooo00<000000?ooo`3oool0@`3oool300000080U?7F0P0000000`2Fl]l00000000000060000
00D0To3U0P00000209?`j@80000000<0U>o/0000000000001`0000000`2@k^`0000000000004093^
kP800000102@kO02000000<0T>ca0P0000000`2@k?8000000000000?000000D0S>[c0P00000408cY
m0<000000`2<jO@2000000D0S>Wd0P00000208cYm@@000000`2=jOD3000000<0S^We0P00000509?/
m@D000000P3U/9`00`000000UG:X09EbZ00609EbZ0040000003V/9/0i[2K00000080ik6J00<00000
09AaZ02DLJP0202DLJP010000000j;6I0>RaV@0000020>VaV0050000003Y/YP0jK:H0>VbV0000000
0`3Z/YL00`000000TfnW09=_Y`0709=_Y`80000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP04
0000002@KJH00000000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00002@0?ooo`030000003o
ool0oooo00X0oooo001g0?ooo`060=nLB000000000000000000004P0W=oo0P3oool00`3ogi`0B000
07Boo`030?ooo`030?n?=@0001@0GlCo00T0oooo00<000000?ooo`3oool0@@3oool2000000<0U?7F
0P0000000`2Fl]l000000000000200000080T_3P0P00000509?`i@80000000<0To3Y000000000000
10000002093_j@800000102@k^`2000000@0T>k^0P000004093]l0`000001P2=jo4200000080S>[b
0`00000508cZl`8000000`2<jO@3000000@0S>Wd0P00000508cYm1d0000000D0UW:Y0000003U/9`0
iK2L0000000809EbZ0040000003V/9/0i[2K00000080ik6J00<0000009AaZ02DLJP0202DLJP01000
0000j;6I0>RaV@0000020>VaV0050000003Y/YP0jK:H0>VbV00000000`3Z/YL00`000000TfnW09=_
Y`0709=_Y`80000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP040000002@KJH00000000000H0
lkJ:00H000000?:fS03b]X`000000?6eS@00002@0?ooo`030000003oool0oooo00X0oooo001g0?oo
o`040?oo_`1d0000000004RLg`@0oooo00<0oono07@00018W=l00`3oool00`34Ga@0000e08ooo`09
0?ooo`030000003oool0oooo0400oooo00@0000009CaeP2DlMH0U?7F1P00000409;`h0030000002C
l>D0To3U00<0To3U1P000005093_j@800000102@k^`2000000@0T>k^0P0000000`2@kO0000000000
0007000000@0SN_`0P00000508g[l@<000000P2<j_82000000H0S>[c6000000508SXm08000000`29
j?@3000000@0RnSe0`00000209?/m@8000000`2FLZT010000000iK2L0>F`W000000809EbZ0040000
003V/9/0i[2K00000080ik6J00<0000009AaZ02DLJP0202DLJP010000000j;6I0>RaV@0000020>Va
V0040000003Y/YP0jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`030000003[/YH0
00000080kK>C0P0000020>jdTP040000002@KJH00000000000H0lkJ:00H000000?:fS03b]X`00000
0?6eS@00002@0?ooo`030000003oool0oooo00X0oooo001h0?ooo`030;md001d_ol0oooo00D0oooo
00<0gia80018W03Oool00P3oool00`2?=@0003F?0?ooo`090?ooo`030000003oool0oooo03l0oooo
00<0000009CaeP2DlMH0100000000`2@kmL0000009;`h00309;`h08000000P2Cl>D500000003093_
i@000000000000D0T>oY0P000004093^k080000000<0T>k^0000000000002000000308g/k`800000
102=jo02000000D0SN_a4`00000408WXl`<000001028j?<300000080R>Sd0P00000508SXm0800000
0P29j?@3000000@0RnSe0`00000309?/m@8000001@2FLZT010000000iK2L0>F`W000000809EbZ003
0000003V/9/0i[2K0080000000<0ik6J0000002DLJP02@2DLJP010000000j;6I0>RaV@0000020>Va
V0040000003Y/YP0jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`030000003[/YH0
00000080kK>C0P0000020>jdTP040000002@KJH00000000000H0lkJ:00H000000?:fS03b]X`00000
0?6eS@00002@0?ooo`030000003oool0oooo00X0oooo001h0?ooo`060?oOW0180000000000000000
04P0W=oo0P3oool01`3ookl0M00007Boo`3oool0okmd0000=@2?ool02P3oool00`000000oooo0?oo
o`0m0?ooo`80000000@0U?7F00000000000000000`2@kmL010000000T_3P09;`h02Bl>0800000080
T>oU0P000005093_j@80000000<0T>k/0000000000002000000308g/kP<000000`2=k>l200000003
08g[l0000000000000h000001@29jO8300000080RNWb0`00000308WXl`<000000`28j?<4000000<0
R>Sd0P00000508SXm1<000001`2FLZT010000000iK2L0>F`W000000809EbZ0030000003V/9/0i[2K
0080000000<0ik6J0000002DLJP02@2DLJP010000000j;6I0>RaV@0000020>VaV0040000003Y/YP0
jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`030000003[/YH000000080kK>C0P00
00020>jdTP040000002@KJH00000000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00002@0?oo
o`030000003oool0oooo00X0oooo001i0?ooo`050;md00000000000000000018W=l00`3oool01@3o
okl0M0000000000007@0Sooo00/0oooo00<000000?ooo`3oool0?03oool00`000000U?7F00000002
000000<0T>oG0P00000209;`h0@000000P2>kn0200000080T>oU0P0000000`2@knT0000000000009
000000<0SNg/0`00000308g/kP80000000<0SNc_0000000000002000000508[Zl08000000`29jO42
000000<0RNWb1000000308WYlP80000000<0RNSc0000000000006P00000408GWm0<000000`28io@3
000000030>B_W@000000UW:Y00L0UW:Y00@000000>F`W03U/9`00000202ELZP00`000000i[2K0>J`
V`02000000030>NaVP000000U76X00T0U76X00@000000>RaV@3X/IT000000P3Y/IP010000000jK:H
0>VbV03Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL00`000000jk:F000000020>fcT`80
00000`3^]983000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00002@0?ooo`030000003oool0
oooo00X0oooo002A0?ooo`<00000>`3oool4000000@0T>oG00<0000009;`h00000000`00000308k_
h080000000<0T>oU0000000000001000000208c]iP8000000P2=kNT3000000<0SNg/0P0000000`2=
k>h0000000000008000000<0R^__0`00000508[Zl08000000`29jO4C000000<0QNOb0`00000408GW
l`<000001025io<6000000@0QNOd0`00000308SWm0<00000103T[id00`000000UW:Y09IbZ@0609Ib
Z@040000003U/9`0iK2L000000P0UG:X00<000000>J`V`3V/9/00P0000020>NaVP030000002DLJP0
U76X00L0U76X00@000000>RaV@3X/IT000000P3Y/IP010000000jK:H0>VbV03Y/YP200000080j[:G
00<0000009=_Y`2CKjL0202CKjL00`000000jk:F000000020>fcT`8000000`3^]983000000H0lkJ:
00H000000?:fS03b]X`000000?6eS@00002@0?ooo`030000003oool0oooo00X0oooo002A0?ooo`04
0000003oool0oooo000003P0oooo0`0000000`2;k<T00000093_e`02093_e`H000000`2>kn060000
00D0S>gV0P00000208g]j@80000000<0SNg/0000000000002000000308[[k@<000000`2:jnl20000
00H0R^[`2`00000408KXl@<000000`26j?8300000080QNOb0`00000408GWl`<000001025io<I0000
00030>B_W@000000UW:Y00L0UW:Y00@000000>F`W03U/9`00000202ELZP2000000030>J`V`000000
00000080ik6J00<0000009AaZ02DLJP01`2DLJP010000000j;6I0>RaV@0000020>VaV08000000P3Y
/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL00`000000jk:F000000020>fcT`8000000`3^
]983000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00002@0?ooo`030000003oool0oooo00X0
oooo002A0?ooo`030000003oool0oooo03P0oooo0P0000000`2;k<T0000000000002093_e`<00000
0P2<k]T00`000000S^oP08k_h0040000000308_]h@000000000000D0S>gV0P0000000`2=kNT00000
00000006000000@0R^_[0P00000308[[k@80000000<0R^__0000000000003`00000308KXl0800000
0`26j?43000000@0Q^Sb60000004087UlP<000001021iO82000000@0PnGc0`000000102FLjT00000
0>B_W@00000809IbZ@040000003U/9`0iK2L000000P0UG:X0P0000000`3V/9/00000000000020>Na
VP030000002DLJP0U76X00L0U76X00@000000>RaV@3X/IT000000P3Y/IP200000080jK:H0P000002
0>ZbU`030000002CKjL0TfnW00P0TfnW00<000000>^bUP0000000P3]/i<2000000<0k[BB0`000006
0?>fRP80000000@0l[J<0000003a]Hd00000T@3oool00`000000oooo0?ooo`090?ooo`00T@3oool0
0`000000oooo0?ooo`0f0?ooo`8000000P2;k<T00`000000T>oG0000000300000080S>kI1P000003
08_]h@030000002<kNH0S>gV00<0S>gV1P00000208W/j@<00000102:jn/<000000D0QnW^0`000003
08KYk`8000000`26j?0B000000@0P^Ka0`00000408;VlP<0000000@0PNGb00000000000000000`21
iO83000000@0PNGb0`00000308?Ul`<00000102FLjT00`000000i:nM0000000809IbZ@040000003U
/9`0iK2L000000P0UG:X0P0000000`3V/9/00000000000020>NaVP030000002DLJP0U76X00L0U76X
00@000000>RaV@3X/IT000000P3Y/IP200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00P0
TfnW0`0000020>fcT`8000000`3^]983000000H0lkJ:0P000000103b]X`000000?6eS@00002A0?oo
o`030000003oool0oooo00T0oooo002A0?ooo`030000003oool0oooo03D0oooo00<0000008_/b@2;
k<T01P00000208c^f@H000000`2;kN4200000080S>gV1@0000000`29k>H000000000000208W/j@80
000000<0R^_[0000000000002000000408OYk@8000000`27jNh4000000@0Q^W_2`00000408?Wl0<0
00000`23i_42000000<0P^Ka0`00000408;VlQX000001`2FLjT00`000000i:nM0000000809IbZ@04
0000003U/9`0iK2L000000P0UG:X0P0000000`3V/9/00000000000020>NaVP030000002DLJP0U76X
00L0U76X00@000000>RaV@3X/IT000000P3Y/IP200000080jK:H0P0000020>ZbU`030000002CKjL0
TfnW00P0TfnW0`0000020>fcT`8000000`3^]983000000H0lkJ:0P000000103b]X`000000?6eS@00
002A0?ooo`030000003oool0oooo00T0oooo002@0?ooo`030000003oool0oooo03D0oooo00<00000
08_/b@0000000P00000208S[c@030000002<k]T0S>kI00<000000P29k=X010000000RngQ08_]h@2;
kN4800000080RNcV0P0000000`29k>T00000000000040000000308KZj@000000000000<0Q^[[0P00
000408OYkA<000000`23io02000000<0PnO`0`00000408?VlA0000005P3oool00`000000UW>Y09Ic
Z@0609IcZ@030000003T[id0000000P0UW:Y00@000000>F`W03U/9`00000202ELZP2000000030>J`
V`00000000000080ik6J00<0000009AaZ02DLJP01`2DLJP010000000j;6I0>RaV@0000020>VaV080
00000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL300000080kK>C00<00000095]Y`00
00000P3^]983000000H0lkJ:0P000000103b]X`000000?6eS@00002A0?ooo`030000003oool0oooo
00T0oooo002@0?ooo`030000003oool0oooo03<0oooo1@00000208S[c@H000000P29k=X200000080
RngQ1000000208S/h@8000000P29k>H6000000<0Q^[Y0`00000308KZj`80000000<0QnW]00000000
00001`00000508CXkP<0000000@0PnO_00000000000000001023io0@000002@0oooo00<0000009Ic
Z@2FLjT01P2FLjT00`000000i:nM0000000809IbZ@040000003U/9`0iK2L000000P0UG:X0P000002
0>J`V`040000003W/IX0ik6J000000T0U76X00@000000>RaV@3X/IT000000P3Y/IP200000080jK:H
0P0000020>ZbU`030000002CKjL0TfnW00P0TfnW0`0000030>fcT`8000000P3^]983000000H0lkJ:
0P000000103b]X`000000?6eS@00002A0?ooo`030000003oool0oooo00T0oooo002@0?ooo`030000
003oool0oooo0380oooo1000000208S[c@H000000P29k=X7000000<0R>cQ0P0000000`29k>H00000
000000020000000308KZiP000000000000<0Q^[Y0P0000000`26j^/0000000000008000000<0Q>S/
0P00000408CXkQ000000<P3oool00`000000UW>Y09IcZ@0609IcZ@030000003T[id0000000P0UW:Y
00@000000>F`W03U/9`00000202ELZP200000080i[2K00@000000>NaVP3W/IX000002@2DLJP01000
0000j;6I0>RaV@0000020>VaV08000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL3
000000<0kK>C0P0000020>jdTP<000001P3c]XX2000000040?:fS0000000lKF=00000940oooo00<0
00000?ooo`3oool02@3oool00900oooo00<000000?ooo`3oool0<03oool30000000307kR]`000000
R>_=00<000000P25j]000`000000RNcJ08W/fP040000000308K[f`000000000000<0R>cQ1@000003
08KZiP8000000P26j^T8000000<0PnS[0`00000308CXk080000000<0Q>S^0000000000001@000010
0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT010000000iK2L0>F`W000
000809EbZ08000000P3V/9/010000000ik6J0>NaVP00000909AaZ0030000003X/IT0j;6I00800000
00<0jK6H0000000000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL3000000<0kK>C
0P0000020>jdTP<000001P3c]XX2000000040?:fS0000000lKF=00000940oooo00<000000?ooo`3o
ool02@3oool00900oooo00<000000?ooo`3oool0;`3oool20000000307kR]`000000000000<00000
0P25j]0600000080Q^_K0P0000000`28k>40000000000005000000<0Q^[V1P00000408?Yj0800000
0`23j>/20000000308CXk0000000000000H00000A`3oool00`000000UW>Y09IcZ@0609IcZ@030000
003T[id0000000P0UW:Y00@000000>F`W03U/9`000002@2ELZP010000000i[2K0>J`V`0000020>Na
VP030000002DLJP0U76X00L0U76X00<000000>RaV@3X/IT00P0000000`3Y/IP00000000000020>Vb
V08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`<000000`3]/i<200000080k[BB0`0000060?>f
RP80000000D0l[J<0000003a]Hd0lKF=0000002@0?ooo`030000003oool0oooo00T0oooo002@0?oo
o`030000003oool0oooo02d0oooo0P00000207kR]`@000000P25j]0600000080Q^_K1P00000208CZ
hP80000000<0Q^[V0000000000001P00000408?Yj0X00000C`3oool00`000000UW>Y09IcZ@0609Ic
Z@030000003T[id0000000P0UW:Y00@000000>F`W03U/9`000002@2ELZP010000000i[2K0>J`V`00
00020>NaVP030000002DLJP0U76X00L0U76X00<000000>RaV@3X/IT00P0000000`3Y/IP000000000
00020>VbV08000000`3Z/YL00`000000TfnW09=_Y`0709=_Y`<000000`3]/i<200000080k[BB0`00
00060?>fRP80000000D0l[J<0000003a]Hd0lKF=0000002@0?ooo`030000003oool0oooo00T0oooo
002@0?ooo`030000003oool0oooo02`0oooo00<0000007kR]`1nh[L01000000208GZd0<0000000@0
P^WB000000000000Q^_K4000000208?YiP80000000<0PnWX0000000000001@00001G0?ooo`030000
002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT2000000030>F`W0000000UG:X00P0UG:X
00@000000>J`V`3V/9/000000P3W/IX00`000000U76X09AaZ00709AaZ0030000003X/IT0j;6I0080
000000<0jK6H0000000000000P3Y/YP2000000<0j[:G00<0000009=_Y`2CKjL01`2CKjL200000004
0>^bUP000000kK>C0>fcT`8000000P3^]982000000030>jdT@000000lkJ:00@0lkJ:0P0000001@3b
]X`000000?6eS@3a]Hd000000900oooo00<000000?ooo`3oool02@3oool00900oooo00<000000?oo
o`3oool0:`3oool00`000000O^:g07kR]`080000000308;YdP000000000000@0000000<0PnWL0000
000000001@000003087YhP80000000<0PnWV0000000000001000001N0?ooo`030000002FLjT0UW>Y
00H0UW>Y00<000000>B_W@000000202FLZT2000000030>F`W0000000UG:X00P0UG:X00@000000>J`
V`3V/9/000000P3W/IX00`000000U76X09AaZ00709AaZ0030000003X/IT0j;6I0080000000<0jK6H
0000000000000P3Y/YP2000000<0j[:G00<0000009=_Y`2CKjL01`2CKjL2000000040>^bUP000000
kK>C0>fcT`8000000P3^]982000000030>jdT@000000lkJ:00@0lkJ:0P0000020?:fS0030000003a
]Hd000000900oooo00<000000?ooo`3oool02@3oool00900oooo00<000000?ooo`3oool0:@3oool2
0000000307kR]`000000000000X0000000<0P>SD00000000000010000002083Xg@8000000`21jN85
000006@0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000i:nM0000000809IbZ@80000000<0iK2L
0000002ELZP0202ELZP010000000i[2K0>J`V`0000020>NaVP030000002DLJP0U76X00L0U76X00<0
00000>RaV@3X/IT00P0000000`3Y/IP00000000000020>VbV08000000`3Z/YL00`000000TfnW09=_
Y`0709=_Y`80000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP80000000<0k[BA0000003c]XX0
103c]XX200000080l[J<00<000000?6eS@000000T03oool00`000000oooo0?ooo`090?ooo`00T03o
ool00`000000oooo0?ooo`0X0?ooo`030000001nh[L0000000@0000000<0O>C40000000000000`00
0002083Xe0D000000`20j=d200000003087YhP000000000000<00000J03oool00`000000UW>Y09Ic
Z@0609IcZ@030000003T[id0000000P0UW:Y0P0000000`3U/9`0000009EbZ00809EbZ0040000003V
/9/0i[2K00000080ik6J00<0000009AaZ02DLJP01`2DLJP00`000000j;6I0>RaV@02000000030>Va
V0000000000000<0jK:H00D000000>ZbU`3Z/YL0j[:G0000000909=_Y`80000000@0jk:F0000003]
/i<0kK>C0P0000020>jdTP040000002@KJH00000000000D0lkJ:0P0000020?:fS0030000003a]Hd0
00000900oooo00<000000?ooo`3oool02@3oool00900oooo00<000000?ooo`3oool09`3oool60000
000307cTa0000000000000<000000P20j=@5000000<0P>SM1@00001]0?ooo`030000002FLjT0UW>Y
00H0UW>Y00<000000>B_W@000000202FLZT2000000030>F`W0000000UG:X00P0UG:X00@000000>J`
V`3V/9/000000P3W/IX00`000000U76X09AaZ00709AaZ080000000@0j;6I000000000000jK6H0P00
00030>VbV0050000003Z/YL0j[:G0>ZbU`0000002@2CKjL2000000040>^bUP000000kK>C0>fcT`80
00000P3^]98010000000T6fV0000000000050?>fRP8000000P3b]X`00`000000lKF=0000002@0?oo
o`030000003oool0oooo00T0oooo002@0?ooo`030000003oool0oooo02D0oooo2@000000101ji<P0
000000000020j=@600000080P>SM1000001a0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_
W@000000202FLZT2000000030>F`W0000000UG:X00P0UG:X00@000000>J`V`3V/9/000000P3W/IX0
0`000000U76X09AaZ00709AaZ080000000@0j;6I000000000000jK6H0P0000030>VbV0050000003Z
/YL0j[:G0>ZbU`0000002@2CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000
T6fV0000000000050?>fRP8000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo
00T0oooo002@0?ooo`<00000903oool30000000306oHZP00000000000080000000<0N^C800000000
00002`00001e0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT200000003
0>F`W0000000UG:X00P0UG:X00@000000>J`V`3V/9/000000P3W/IX00`000000U76X09AaZ00709Aa
Z080000000@0j;6I000000000000jK6H0P0000030>VbV0050000003Z/YL0j[:G0>ZbU`0000002@2C
KjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV0000000000060?>fRP06
0000003b]X`0l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`090?ooo`00T03oool00`00
0000oooo0?ooo`0S0?ooo`80000000<0KmRZ0000000000001P00000207WTc0H00000N@3oool00`00
0000UW>Y09IcZ@0609IcZ@030000003T[id0000000P0UW:Y0P0000020>F`W0030000002ELZP0UG:X
00H0UG:X00@000000>J`V`3V/9/000000P3W/IX00`000000U76X09AaZ00709AaZ080000000<0j;6I
0000000000000P3Y/IP01@000000jK:H0>VbV03Y/YP0000000<0j[:G00<0000009=_Y`2CKjL01`2C
KjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV0000000000060?>fRP06
0000003b]X`0l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`090?ooo`00T03oool00`00
0000oooo0?ooo`0Q0?ooo`80000000<0KmRZ0000000000001@00000307WTc0@00000O@3oool00`00
0000UW>Y09IcZ@0609IcZ@030000003T[id0000000P0UW:Y0P0000020>F`W0030000002ELZP0UG:X
00H0UG:X00@000000>J`V`3V/9/000000P3W/IX00`000000U76X09AaZ00709AaZ080000000<0j;6I
0000000000000P3Y/IP010000000jK:H0>VbV03Y/YP200000080j[:G00<0000009=_Y`2CKjL01`2C
KjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV0000000000060?>fRP06
0000003b]X`0l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`090?ooo`00T03oool00`00
0000oooo0?ooo`0P0?ooo`80000000<0KmRZ0000000000001000000307WTc0<00000P03oool00`00
0000UW>Y09IcZ@0609IcZ@<00000202FLZT200000080iK2L00<0000009EbZ02ELZP01P2ELZP01000
0000i[2K0>J`V`0000020>NaVP030000002DLJP0U76X00L0U76X0P0000000`3X/IT0000000000002
0>VaV0040000003Y/YP0jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`80000000@0
jk:F0000003]/i<0kK>C0P0000020>jdTP040000002@KJH00000000000H0lkJ:00H000000?:fS03b
]X`000000?6eS@00002@0?ooo`030000003oool0oooo00T0oooo002@0?ooo`030000003oool0oooo
01l0oooo00<0000006oHZP0000001@00000207WTc0<00000P`3oool00`000000UW>Y09IcZ@0609Ic
Z@<000002@2FLZT010000000iK2L0>F`W000000809EbZ0040000003V/9/0i[2K00000080ik6J00<0
000009AaZ02DLJP01`2DLJP2000000030>RaV@00000000000080jK6H00@000000>VbV03Y/YP0jK:H
0P0000020>ZbU`030000002CKjL0TfnW00L0TfnW0P000000103[/YH000000>fcT`3]/i<200000080
k[BB00@00000091]YP00000000001P3c]XX2000000040?:fS0000000lKF=00000900oooo00<00000
0?ooo`3oool02@3oool00900oooo00<000000?ooo`3oool07P3oool300000003073K]00000000000
00D00000QP3oool00`000000UW>Y09IcZ@0609IcZ@<000002@2FLZT010000000iK2L0>F`W0000008
09EbZ0040000003V/9/0i[2K00000080ik6J00<0000009AaZ02DLJP01`2DLJP2000000030>RaV@00
000000000080jK6H00@000000>VbV03Y/YP0jK:H0P0000020>ZbU`030000002CKjL0TfnW00L0TfnW
0P000000103[/YH000000>fcT`3]/i<200000080k[BB00@00000091]YP00000000001P3c]XX20000
00040?:fS0000000lKF=00000900oooo00<000000?ooo`3oool02@3oool00900oooo00<000000?oo
o`3oool0703oool:000008T0oooo00<0000009IcZ@2FLjT01P2FLjT3000000T0UW:Y00@000000>F`
W03U/9`00000202ELZP00`000000i[2K0>J`V`02000000030>NaVP000000U76X00P0U76X0P000000
0`3X/IT00000000000020>VaV0040000003Y/YP0jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_
Y`0709=_Y`80000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP040000002@KJH00000000000H0
lkJ:0P000000103b]X`000000?6eS@00002A0?ooo`030000003oool0oooo00P0oooo002?0?ooo`03
0000003oool0oooo01`0oooo2000002<0?ooo`030000002FLjT0UW>Y00H0UW>Y0P0000000`3T[id0
000009IbZ@0709IbZ@040000003U/9`0iK2L000000P0UG:X00<000000>J`V`3V/9/00P0000000`3W
/IX0000009AaZ00809AaZ080000000<0j;6I0000000000000P3Y/IP200000080jK:H0P0000020>Zb
U`030000002CKjL0TfnW00L0TfnW0P000000103[/YH000000>fcT`3]/i<200000080k[BB00@00000
091]YP00000000001P3c]XX2000000040?:fS0000000lKF=00000940oooo00<000000?ooo`3oool0
203oool008l0oooo00<000000?ooo`3oool06`3oool6000008l0oooo00<0000009IcZ@2FLjT01P2F
LjT2000000030>B_W@000000UW:Y00L0UW:Y00@000000>F`W03U/9`00000202ELZP00`000000i[2K
0>J`V`02000000030>NaVP000000U76X00P0U76X0P0000000`3X/IT00000000000020>VaV0800000
0P3Y/YP200000080j[:G00<0000009=_Y`2CKjL01`2CKjL010000000TfnW0000000000020>fcT`80
00000P3^]98010000000T6fV0000000000060?>fRP80000000@0l[J<0000003a]Hd00000T@3oool0
0`000000oooo0?ooo`080?ooo`00S`3oool00`000000oooo0?ooo`0J0?ooo`D00000T@3oool00`00
0000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT010000000iK2L0>F`W0000008
09EbZ0030000003V/9/0i[2K0080000000<0ik6J0000002DLJP0202DLJP2000000030>RaV@000000
00000080jK6H0P0000020>VbV08000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`040000002CKjL0
000000000080kK>C00H00000095]Y`000000k[BB0000002@KJH2000000H0lkJ:0P000000103b]X`0
00000?6eS@00002A0?ooo`030000003oool0oooo00P0oooo002?0?ooo`030000003oool0oooo01P0
oooo1@00002C0?ooo`030000002FLjT0UW>Y00H0UW>Y0P0000000`3T[id0000009IbZ@0709IbZ@04
0000003U/9`0iK2L000000P0UG:X00<000000>J`V`3V/9/00P0000000`3W/IX0000009AaZ00809Aa
Z080000000<0j;6I0000000000000P3Y/IP200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW
00L0TfnW00@0000009=_Y`00000000000P3]/i<01P000000TFfW0000003^]9800000091]YP800000
1P3c]XX2000000040?:fS0000000lKF=00000940oooo00<000000?ooo`3oool0203oool008l0oooo
00<000000?ooo`3oool05`3oool4000009D0oooo00<0000009IcZ@2FLjT01P2FLjT2000000030>B_
W@000000UW:Y00L0UW:Y00@000000>F`W03U/9`00000202ELZP00`000000i[2K0>J`V`0200000003
0>NaVP000000U76X00P0U76X0P0000000`3X/IT00000000000020>VaV08000000P3Y/YP200000080
j[:G00<0000009=_Y`2CKjL0202CKjL300000080kK>C00H00000095]Y`000000k[BB0000002@KJH2
000000H0lkJ:0P0000001@3b]X`000000?6eS@3a]Hd000000900oooo00<000000?ooo`3oool0203o
ool008l0oooo00<000000?ooo`3oool05P3oool3000009L0oooo00<0000009IcZ@2FLjT01P2FLjT2
000000030>B_W@000000UW:Y00L0UW:Y00@000000>F`W03U/9`00000202ELZP00`000000i[2K0>J`
V`02000000030>NaVP000000U76X00P0U76X0P0000000`3X/IT00000000000020>VaV08000000P3Y
/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL300000080kK>C00H00000095]Y`000000k[BB
0000002@KJH2000000H0lkJ:0P0000001@3b]X`000000?6eS@3a]Hd000000900oooo00<000000?oo
o`3oool0203oool008l0oooo00<000000?ooo`3oool0503oool3000009T0oooo00<0000009IcZ@2F
LjT01P2FLjT2000000030>B_W@000000UW:Y00L0UW:Y00@000000>F`W03U/9`00000202ELZP00`00
0000i[2K0>J`V`02000000030>NaVP000000U76X00P0U76X0P0000020>RaV@030000003Y/IP0jK6H
008000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL300000080kK>C00<00000095]
Y`0000000P3^]982000000030>jdT@000000lkJ:00@0lkJ:0P0000020?:fS0030000003a]Hd00000
0900oooo00<000000?ooo`3oool0203oool008l0oooo00<000000?ooo`3oool04`3oool3000009X0
oooo00<0000009IcZ@2FLjT01P2FLjT2000000030>B_W@000000UW:Y00L0UW:Y00@000000>F`W03U
/9`00000202ELZP2000000050>J`V`00000000000>NaVP0000002@2DLJP200000080j;6I00<00000
0>VaV03Y/IP00P0000020>VbV08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`<000000P3]/i<0
0`000000TFfW000000020>jdTP80000000<0k[BA0000003c]XX0103c]XX200000080l[J<00<00000
0?6eS@000000T03oool00`000000oooo0?ooo`080?ooo`00S`3oool00`000000oooo0?ooo`0A0?oo
o`<00000W03oool00`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT01000
0000iK2L0>F`W000000809EbZ080000000D0i[2K000000000000ik6J0000000909AaZ08000000P3X
/IT00`000000jK6H0>VaV00200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00P0TfnW0P00
00001P3[/YH000000>fcT`000000TFfW00000080k[BB0P0000000`3^]94000000?>fRP040?>fRP80
00000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00P0oooo002?0?ooo`<00000
403oool3000009d0oooo00<0000009IcZ@2FLjT01P2FLjT2000000030>B_W@000000UW:Y00L0UW:Y
00@000000>F`W03U/9`00000202ELZP2000000050>J`V`00000000000>NaVP0000002P2DLJP00`00
0000j;6I0>RaV@02000000030>VaV000000000000080jK:H0P0000020>ZbU`030000002CKjL0TfnW
00P0TfnW0P0000001P3[/YH000000>fcT`000000TFfW00000080k[BB0P0000000`3^]94000000?>f
RP040?>fRP8000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00P0oooo002?
0?ooo`030000003oool0oooo00l0oooo0P00002O0?ooo`030000002FLjT0UW>Y00L0UW>Y00<00000
0>B_W@000000202FLZT010000000iK2L0>F`W000000809EbZ080000000D0i[2K000000000000ik6J
0000000:09AaZ0030000003X/IT0j;6I0080000000<0jK6H0000000000000P3Y/YP200000080j[:G
00<0000009=_Y`2CKjL0202CKjL2000000060>^bUP000000kK>C0000002AKJL000000P3^]9801000
0000T6fV0000000000050?>fRP8000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0
oooo00P0oooo002?0?ooo`030000003oool0oooo00h0oooo0P00002P0?ooo`030000002FLjT0UW>Y
00L0UW>Y00<000000>B_W@000000202FLZT010000000iK2L0>F`W000000809EbZ080000000D0i[2K
000000000000ik6J0000000:09AaZ0030000003X/IT0j;6I0080000000<0jK6H0000000000000P3Y
/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL2000000060>^bUP000000kK>C0000002AKJL0
00000P3^]98010000000T6fV0000000000050?>fRP8000000P3b]X`00`000000lKF=0000002@0?oo
o`030000003oool0oooo00P0oooo002?0?ooo`030000003oool0oooo00d0oooo0P00002R0?ooo`03
0000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT010000000iK2L0>F`W000000809Eb
Z080000000D0i[2K000000000000ik6J0000000:09AaZ0030000003X/IT0j;6I0080000000<0jK6H
0000000000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL2000000040>^bUP000000
kK>C0>fcT`8000000P3^]98010000000T6fV0000000000050?>fRP8000000P3b]X`00`000000lKF=
0000002@0?ooo`030000003oool0oooo00P0oooo002?0?ooo`030000003oool0oooo00`0oooo0P00
002S0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT010000000iK2L0>F`
W000000809EbZ080000000D0i[2K000000000000ik6J0000000:09AaZ0030000003X/IT0j;6I0080
000000<0jK6H0000000000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0202CKjL200000004
0>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV0000000000050?>fRP8000000P3b]X`0
0`000000lKF=0000002@0?ooo`030000003oool0oooo00P0oooo002?0?ooo`030000003oool0oooo
00X0oooo0`00002T0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT01000
0000iK2L0>F`W000000809EbZ080000000<0i[2K0000000000000P3W/IX00`000000U76X09AaZ007
09AaZ0030000003X/IT0j;6I0080000000<0jK6H0000000000000P3Y/YP200000080j[:G00<00000
09=_Y`2CKjL0202CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV0000
000000050?>fRP8000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00P0oooo
002?0?ooo`030000003oool0oooo00T0oooo00<0000004`9E`000000Y@3oool00`000000UW>Y09Ic
Z@0609IcZ@030000003T[id0000000P0UW:Y00@000000>F`W03U/9`00000202ELZP2000000030>J`
V`00000000000080ik6J00<0000009AaZ02DLJP01`2DLJP00`000000j;6I0>RaV@02000000030>Va
V000000000000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00P0TfnW0P000000103[/YH00000
0>fcT`3]/i<200000080k[BB00@00000091]YP00000000001@3c]XX01`000000lkJ:0000003b]X`0
00000?6eS@000000T03oool00`000000oooo0?ooo`080?ooo`00S`3oool00`000000oooo0?ooo`08
0?ooo`040000001<2EL0000000000:D0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000i:nM0000
000809IbZ@040000003U/9`0iK2L000000P0UG:X0P0000000`3V/9/00000000000020>NaVP030000
002DLJP0U76X00L0U76X00<000000>RaV@3X/IT00P0000000`3Y/IP00000000000020>VbV0800000
0P3Z/YL00`000000TfnW09=_Y`0809=_Y`80000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP04
0000002@KJH00000000000D0lkJ:00L000000?>fRP000000l[J<0000003a]Hd000000900oooo00<0
00000?ooo`3oool0203oool008l0oooo00<000000?ooo`3oool01`3oool010000000C0UG00000000
002V0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT010000000iK2L0>F`
W000000809EbZ080000000<0i[2K0000000000000P3W/IX00`000000U76X09AaZ00709AaZ0030000
003X/IT0j;6I0080000000<0jK6H0000000000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL0
202CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV0000000000050?>f
RP070000003c]XX000000?:fS0000000lKF=0000002@0?ooo`030000003oool0oooo00P0oooo002?
0?ooo`030000003oool0oooo00H0oooo00@0000004`9E`0000000000Y`3oool00`000000UW>Y09Ic
Z@0609IcZ@030000003T[id0000000P0UW:Y00@000000>F`W03U/9`00000202ELZP2000000030>J`
V`00000000000080ik6J00<0000009AaZ02DLJP01`2DLJP00`000000j;6I0>RaV@02000000030>Va
V000000000000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00P0TfnW0P000000103[/YH00000
0>fcT`3]/i<200000080k[BB00@00000091]YP00000000001@3c]XX01`000000lkJ:0000003b]X`0
00000?6eS@000000T03oool00`000000oooo0?ooo`080?ooo`009@3oool01@3OW4P0000000000000
0000M;oo00l0oooo00D0gia8000000000000000007Boo`080?ooo`030?oog`2LB4P0W=oo0140oooo
00<0oooO09a8000000000`0000001@18W=l0oooo0?oOW018B9`0gooo02X0oooo00<000000?ooo`3o
ool01P3oool00`000000C0UG0000002X0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@00
0000202FLZT2000000030>F`W0000000UG:X00L0UG:X0P0000000`3V/9/00000000000020>NaVP03
0000002DLJP0U76X00L0U76X0P000000103X/IT000000000003Y/IP200000080jK:H00D0000009=`
Z0000000j[:G0000000:09=_Y`80000000@0jk:F0000003]/i<0kK>C0P0000020>jdTP040000002@
KJH00000000000D0lkJ:00L000000?>fRP000000l[J<0000003a]Hd000000940oooo00<000000?oo
o`3oool01`3oool002D0oooo00<0gia80000M02oool02@3oool0103OW4P0000000000000M;l60?oo
o`030?oo_`1d0000M;oo00P0oooo00<0_gA809cOo`3oool04@3oool00`2oM00004RL0=ooo`030?oo
o`050=nLW02oool0oooo0?oOW018M;l0:@3oool00`000000oooo0?ooo`060?ooo`030000001<2EL0
00000:T0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000i:nM0000000809IbZ@80000000<0iK2L
0000002ELZP01`2ELZP2000000030>J`V`00000000000080ik6J00<0000009AaZ02DLJP01`2DLJP2
000000040>RaV@00000000000>VaV08000000P3Y/YP01@000000Tg2X0000003Z/YL0000000X0TfnW
0P000000103[/YH000000>fcT`3]/i<200000080k[BB00@00000091]YP00000000001@3c]XX01`00
0000lkJ:0000003b]X`000000?6eS@000000T@3oool00`000000oooo0?ooo`070?ooo`009@3oool0
0`3OW4P004RL0=ooo`0<0?ooo`030?noM000M;l0oooo00@0oooo00<0oono07@0001d_ol01`3oool0
0`3OW4P0B9cO0?ooo`0B0?ooo`030?oog`2LM9`0gooo00<0oooo00H0oono09bLg`3oool0oooo0=nL
B018W=lX0?ooo`030000003oool0oooo00H0oooo0P00002Z0?ooo`030000002FLjT0UW>Y00H0UW>Y
00<000000>B_W@000000202FLZT2000000030>F`W0000000UG:X00L0UG:X0P0000000`3V/9/00000
000000020>NaVP030000002DLJP0U76X00L0U76X0P000000103X/IT000000000003Y/IP200000080
jK:H00<0000009=`Z00000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`80000000@0jk:F0000003]
/i<0kK>C0P0000020>jdTP040000002@KJH00000000000D0lkJ:00L000000?>fRP000000l[J<0000
003a]Hd000000940oooo00<000000?ooo`3oool01`3oool002D0oooo00<0gia80018W03Oool0303o
ool00`3ookl0M01d0;ooo`040?ooo`030?oo_`1d0000M;oo00H0oooo00<0oono07@0M02oool0503o
ool00`2oM0000000000000030000000604RLg`3oool0oooo0?oo_`1d04P0W=oo9`3oool00`000000
oooo0?ooo`050?ooo`800000Z`3oool00`000000UW>Y09IcZ@0609IcZ@030000003T[id0000000P0
UW:Y0P0000000`3U/9`0000009EbZ00709EbZ080000000<0i[2K0000000000000P3W/IX00`000000
U76X09AaZ00709AaZ080000000@0j;6I000000000000jK6H0P0000020>VbV0030000002CL:P00000
0080j[:G00<0000009=_Y`2CKjL01`2CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]980
10000000T6fV0000000000050?>fRP070000003c]XX000000?:fS0000000lKF=0000002A0?ooo`03
0000003oool0oooo00L0oooo00020?ooo`030?oog`2LB00000000080000000P0B9cO0?ooo`3oool0
_gA809cOo`3oool0omnL04P000<0000000<004RL0=oog`2LB0000P0000002@00B9`0W7@000000000
M;l0_g@000000000M;l0oooo0=nLB002000000060018W03Oool0oooo0=nLB000B9`0gooo0P3oool0
1@3ookl0M0000000000004P0W=oo00D0oooo00<0gia80000M02oool0103oool00`3ookl0M00007Bo
o`060?ooo`030=nLB000B9`0gooo00<0oooo00P0oooO09a80000000000000000B018B00004RL0=oo
o`80oooo00P0_g@0000000000000001d0;ooo`3oool0oono07@000@0000000L0001d0;ooo`3oool0
oooo0?oog`2LB000B9cO02L0oooo00<000000?ooo`3oool01@3oool200000:/0oooo00<0000009Ic
Z@2FLjT01P2FLjT00`000000i:nM0000000809IbZ@80000000<0iK2L0000002ELZP01`2ELZP20000
00030>J`V`00000000000080ik6J00<0000009AaZ02DLJP01`2DLJP20000000;0>RaV@0000000000
0>VaV0000000Tg2X0000003Y/YP0000009=`Z00000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`04
0000002CKjL0000000000080kK>C00H00000095]Y`000000k[BB0000002@KJH2000000D0lkJ:00D0
00000?>fRP000000l[J<000000020?6eS@030000003oool0oooo08h0oooo00<000000?ooo`3oool0
1`3oool000@0oooo00<0omnL04Q8W03Oool00P3oool00`3ogi`0B000001d_`030?ooo`0D0=nLB000
04P0W=oo0?ooo`3o_g@0000007Boo`3oool0omnL04P00000M;l0okno0;oog`2LB7@0_ooo0?ooo`3o
oml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo0080oooo00H0_g@0001d_`3oool0omnL04P0
B02Lgol20?ooo`040?oo_`1d0000000004RLg`D0oooo00<0oono07@0001d_ol01P3oool00`2oM000
07Bo0?ooo`020?ooo`0D0?oog`2LB000000007Boo`3oool0oono07@000000000001d0;ooo`2oM000
0000001d_`3oool0omnL04P0M02oool0oooo0;mdB02Lgol80?ooo`030=nLB000B9`0gooo02H0oooo
00<000000?ooo`3oool0103oool200000:`0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000i:nM
0000000809IbZ@80000000<0iK2L0000002ELZP01`2ELZP2000000030>J`V`00000000000080ik6J
00<0000009AaZ02DLJP01`2DLJP20000000;0>RaV@00000000000>VaV0000000Tg2X0000003Y/YP0
000009=`Z00000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`040000002CKjL0000000000080kK>C
00H00000095]Y`000000k[BB0000002@KJH2000000D0lkJ:00<000000?>fRP0000000P3b]X`00`00
0000lKF=0000002@0?ooo`030000003oool0oooo00L0oooo00040?ooo`030?oo_`1dB9`0gooo0080
oooo00@0_g@00000000004P0W=oo0P3oool02`3OW4P0001d0;ooo`3oool0okmd0000B02Lgol0oooo
0?oog`2LB00007Bo0080oooo00H0oono07A8W03Oool0_g@00000M02oool30?ooo`070=nLB000B9`0
gooo0?ooo`3ooml0W4P007Boo`020?ooo`030=nLB000B9`0gooo0080oooo00<0oooO09bL_`3oool0
1@3oool00`3ookl0M00007Boo`050?ooo`030?oo_`1d0000B9cO00<0oooo00<0omnL04P00018W=l0
0P3oool01`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3oool01@3ogi`0M9cO0?ooo`3ookl0
M0000080000000<007Bo0?ooo`3oool00`3oool00`3OW4P0001d0;ooo`0V0?ooo`030000003oool0
oooo00@0oooo0P00002/0?ooo`030000002FLjT0UW>Y00H0UW>Y00<000000>B_W@000000202FLZT2
000000030>F`W0000000UG:X00L0UG:X0P0000000`3V/9/00000000000020>NaVP030000002DLJP0
U76X00L0U76X0P0000001`3X/IT000000000003Y/IP0000009=`Z00000000P3Y/YP200000080j[:G
00<0000009=_Y`2CKjL01`2CKjL010000000TfnW0000000000020>fcT`090000002AKJL000000>jd
TP000000T6fV0000003^]940000000@0lkJ:00<000000?>fRP0000000P3b]X`00`000000lKF=0000
002@0?ooo`030000003oool0oooo00L0oooo00040?ooo`090?oo_`1dB9`0gooo0?ooo`3ogi`0B000
001d_`3o_g@0B9cO0080oooo00<0gia80000M02oool00P3oool01P3o_g@00000001d_`3ooml0W4P0
001d_`<0oooo00@0omnL07A8B0000000M;oo103oool01`3OW4P004RL0=ooo`3oool0oono07@0001d
_ol00P3oool00`3ookl0M00007Boo`030?ooo`030?noM018W=l0oooo00@0oooo00<0oono07@0001d
_ol01@3oool00`3ookl0M00007Boo`030?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000
M;l0oooo0?noM00004P0W=oo00H0oooo00H0gia80000M02oool0oono07@00018W=l40?ooo`030?no
M00007@0_ooo02H0oooo0`0000030?ooo`800000[@3oool00`000000UW>Y09IcZ@0609IcZ@030000
003T[id0000000P0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ080000000H0i[2K0000002ELZP0
00000>NaVP00000909AaZ080000000L0j;6I000000000000jK6H0000002CL:P000000080jK:H0P00
00020>ZbU`030000002CKjL0TfnW00L0TfnW00@0000009=_Y`00000000000P3]/i<02@000000TFfW
0000003^]9800000091]YP000000k[BA000000040?>fRP030000003c]XX000000080l[J<00<00000
0?6eS@000000T03oool00`000000oooo0?ooo`070?ooo`00103oool03P3ookl0M4RL0=ooo`3oool0
_g@00000B02Lgol0oono07@0M02oool0oooo0=nLB00007@0_ooo103oool0103o_g@0000000000000
M;l30?ooo`030?oo_`1d0000B9cO00D0oooo00L0gia80018W03Oool0oooo0?oo_`1d0000B9cO00<0
oooo00L0_g@0001d_`3ogi`0M9cO0?ooo`3OW4P007Bo00D0oooo00<0oono07@0001d_ol01@3oool0
0`3ookl0M00007Boo`030?ooo`030?noM0000000M;oo0080oooo00L0oooO09a80000M;l0oooo0?no
M00007@0_ooo00D0oooo00L0oono07@00018W=l0oooo0?oog`2LB00007Bo00@0oooo00<0okmd0000
B02Lgol09P3oool00`000000oooo0?ooo`030?ooo`800000[@3oool00`000000UW>Y09IcZ@0609Ic
Z@030000003T[id0000000P0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ08000000P3V/9/20000
00030>NaVP000000U76X00P0U76X0P0000001`3X/IT000000000003Y/IP0000009=`Z00000000P3Y
/YP200000080j[:G00<0000009=_Y`2CKjL01`2CKjL010000000TfnW0000000000020>fcT`090000
002AKJL000000>jdTP000000T6fV0000003^]940000000@0lkJ:00<000000?>fRP0000000P3b]X`0
0`000000lKF=0000002@0?ooo`030000003oool0oooo00L0oooo00040?ooo`0F0?oo_`1dB9`0gooo
0?oOW018000004RL0=ooo`3oool0gia804RLg`3oool0gia80000M02oool0oooo0?oo_`1d0000M;oo
0?ooo`3ooml0W4P0001d_`80oooo00D0oooO09a80000B9`0W7A809cOo`040?ooo`070=nLB000B9`0
gooo0?ooo`3ooml0W4P0001d_`030?ooo`070=nLB00007@0_ooo0=nLB0000000001809cOo`050?oo
o`030?oo_`1d0000M;oo00D0oooo00<0oono07@0001d_ol00`3oool00`3ookl0M00004RLg`020?oo
o`060?oog`2LB00007Bo0?ooo`3ogi`0B0001@0000002018W=l0oono07@00018W=l0oooo0?oog`2L
B000B9cO103oool00`3o_g@0001d0;ooo`0V0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0[03oool00`000000UW>Y09IcZ@0609IcZ@030000003T[id0000000P0UW:Y0P0000000`3U
/9`0000009EbZ00709EbZ08000000P3V/9/2000000030>NaVP000000U76X00P0U76X0P0000001`3X
/IT000000000003Y/IP0000009=`Z00000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL01`2C
KjL010000000TfnW0000000000020>fcT`090000002AKJL000000>jdTP000000T6fV0000003^]940
000000@0lkJ:00<000000?>fRP0000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0
oooo00L0oooo00040?ooo`060?oo_`1dB9`0gooo0=nLB0000000M;oo0P3oool03P3ookl0M01d0;oo
o`3OW4P0001d0;ooo`3oool0oono07@00018W=l0oooo0?oo_`1d000007Bo0P3oool01P2oM0000018
09cOo`3ooml0W4Qd0;ooo`<0oooo00<0gia80018W03Oool00P3oool00`2oM0000000000000040000
000307Boo`3oool0oooo00L0oooo00<0oono07@0001d_ol01@3oool00`3ookl0M00004RLg`040?oo
o`070?noM00004P0W=oo0?ooo`3o_g@00000001d_`020?ooo`0<0;md0018W=l0oooo0?ooo`3OW4P0
001809cOo`3oool0_g@0001d_`3oool0okmd0P0000001`00B9`0gooo0?ooo`3oool0gia80000M02o
ool09P3oool00`000000oooo0?ooo`020?ooo`800000[P3oool00`000000UW>Y09IcZ@0609IcZ@03
0000003T[id0000000P0UW:Y0P0000000`3U/9`0000009EbZ00809EbZ0030000003V/9/0i[2K0080
000000<0ik6J0000002DLJP0202DLJP2000000030>RaV@00000000000080jK6H0P0000020>VbV080
00000P3Z/YL00`000000TfnW09=_Y`0709=_Y`0?0000002CKjL000000>^bUP000000kK>C0000002A
KJL000000>jdTP000000T6fV0000003^]940000000@0lkJ:00<000000?>fRP0000000P3b]X`00`00
0000lKF=0000002@0?ooo`030000003oool0oooo00L0oooo00040?ooo`060?oo_`1dB9`0gono07@0
0000B9`0gooo0`3oool01@2oM000M;oo0=nLB00007@0_ooo0080oooo00P0oono07@0000000000000
0000M02oool0oooO09a80080000000D007Bo0?noM0000000001d0;ooo`020?ooo`030=nLB000B9`0
gooo0080oooo00<0gia80000M02oool00P3oool00`3ookl0M00004RLg`090?ooo`030?oo_`1d0000
M;oo00D0oooo00<0oooO09a80018W=l01@3oool00`3o_g@000000000000200000005001d_`3oool0
oooo0?ooo`3OW4P00P0000001P0004P0W=oo0?ooo`3oool0oooO09a80080000000<0M;oo0?ooo`3o
ool00`3oool00`3OW4P004RL0=ooo`0V0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0[@3oool00`000000UW>Y09IcZ@0609IcZ@030000003T[id0000000P0UW:Y0P0000000`3U/9`0
000009EbZ00809EbZ0030000003V/9/0i[2K0080000000<0ik6J0000002DLJP0202DLJP200000003
0>RaV@00000000000080jK6H0P0000020>VbV08000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`0;
0000002CKjL000000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000@0lkJ:
00<000000?>fRP0000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00L0oooo
00040?ooo`050?oo_`1dB9`0_ia80000001d_ol0103oool01@3ogi`0B4RL0;nLB00007@0_ooo01<0
oooo00<0gia80018W03Oool00P3oool00`3ogi`0B00007Boo`030?ooo`030;md0000M;l0oooo00P0
oooo00<0oono07@0001d_ol01P3oool00`3OW4P004RL0=ooo`070?ooo`030?oog`2LB00007Bo01<0
oooo00<0oooO09a80018W=l09`3oool010000000oooo0?ooo`3oool200000:l0oooo00<0000009Ic
Z@2FLjT01P2FLjT00`000000i:nM0000000809IbZ@80000000<0iK2L0000002ELZP0202ELZP00`00
0000i[2K0>J`V`02000000030>NaVP000000U76X00P0U76X0P0000000`3X/IT00000000000020>Va
V08000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL01`2CKjL02`000000TfnW0000003[/YH0
00000>fcT`000000TFfW0000003^]98000000080T6fV0P0000040?>fRP030000003c]XX000000080
l[J<00<000000?6eS@000000T03oool00`000000oooo0?ooo`070?ooo`00103oool01@3ookl0M4Q8
04P00000B9`0gooo00D0oooo00@0_g@004Q8B00007@0_ooo4`3oool00`3OW4P004RL0=ooo`030?oo
o`030;md0000B9`0gooo0080oooo00<0gia8001d_`3oool0203oool00`3ookl0M00007Boo`060?oo
o`030?oo_`1d04P0W=oo00L0oooo00<0oooO09a80000M;l04`3oool00`3ookl0M01809cOo`0W0?oo
o`050000003oool0oooo0?ooo`000000/03oool00`000000UW>Y09IcZ@0609IcZ@<00000202FLZT2
000000030>F`W0000000UG:X00P0UG:X00<000000>J`V`3V/9/00P0000000`3W/IX0000009AaZ008
09AaZ080000000<0j;6I0000000000000P3Y/IP200000080jK:H0P0000020>ZbU`030000002CKjL0
TfnW00L0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB00000002091]
YP800000103c]XX00`000000lkJ:000000020?:fS0030000003a]Hd000000900oooo00<000000?oo
o`3oool01`3oool000@0oooo00@0omnL04P000000000M;oo1P3oool0103o_g@000000000B02LgolC
0?ooo`030=nLB000B9`0gooo00<0oooo00<0oono07@0001d_ol00P3oool00`3OW4P007Bo0?ooo`08
0?ooo`030?oo_`1d0000M;oo00L0oooo00<0gia804RLg`3oool01P3oool00`3ooml0W4P0001d_`0C
0?ooo`030=nLB018W=l0oooo02L0oooo00<000000?ooo`3oool00P00002`0?ooo`030000002FLjT0
UW>Y00H0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP00`000000i[2K0>J`V`02
000000030>NaVP000000U76X00P0U76X0P0000001@3X/IT0000009AaZ0000000jK6H008000000P3Y
/YP200000080j[:G00<0000009=_Y`2CKjL01`2CKjL02`000000TfnW0000003[/YH000000>fcT`00
0000TFfW0000003^]98000000080T6fV0P0000040?>fRP030000003c]XX000000080l[J<00<00000
0?6eS@000000T03oool00`000000oooo0?ooo`070?ooo`000P3oool01P3ooml0W4P0000000000000
001d0;ooo`H0oooo00H0oooO09a80000000000000018W03OoolA0?ooo`030=nLB00007@0_ooo00@0
oooo00H0okmd0000B02Lgol0oooo0;md001d_ol90?ooo`030?oo_`1d0000M;oo00P0oooo00<0_g@0
04RLg`3oool0103oool0103ookl0M00000000000M;lB0?ooo`030?noM000M;l0oooo02P0oooo00@0
00000?ooo`3oool00000/@3oool00`000000UW>Y09IcZ@0609IcZ@<00000202FLZT2000000030>F`
W0000000UG:X00P0UG:X00<000000>J`V`3V/9/00P0000000`3W/IX0000009AaZ00809AaZ0800000
00D0j;6I0000002DLJP000000>VaV00200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00L0
TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB00000002091]YP800000
103c]XX020000000lkJ:0?>fRP000000l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`07
0?ooo`009@3oool01@3OW4P00000000000000000M;oo00<0oooo00D0omnL04P00000000004RL0=oo
o`070?ooo`050=nLB000000000000000001d_ol02@3oool00`3OW7@0W=oo0?ooo`040?ooo`030?oo
g`2LB00007Bo0140oooo00<0oono07Ad_`3oool0:@3oool010000000oooo0?ooo`00002a0?ooo`03
0000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP00`000000
i[2K0>J`V`02000000030>NaVP000000U76X00P0U76X0P0000001@3X/IT0000009AaZ0000000jK6H
008000000P3Y/YP200000080j[:G00<0000009=_Y`2CKjL01`2CKjL02`000000TfnW0000003[/YH0
00000>fcT`000000TFfW0000003^]98000000080T6fV0P0000040?>fRP080000003c]XX0lkJ:0000
003b]X`000000?6eS@00002@0?ooo`030000003oool0oooo00L0oooo002>0?ooo`030000003oool0
00000;80oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y0P0000000`3U/9`0000009EbZ008
09EbZ080000000D0i[2K000000000000ik6J0000000909AaZ080000000D0j;6I0000002DLJP00000
0>VaV00200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00L0TfnW00T0000009=_Y`000000
jk:F0000003]/i<00000095]Y`0000000P3^]98010000000T6fV0000000000040?>fRP080000003c
]XX0lkJ:0000003b]X`000000?6eS@00002@0?ooo`030000003oool0oooo00L0oooo002>0?ooo`03
0000003oool000000;80oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y0P0000000`3U/9`0
000009EbZ00809EbZ080000000D0i[2K000000000000ik6J0000000909AaZ080000000D0j;6I0000
002DLJP000000>VaV00200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00P0TfnW0P000000
1P3[/YH000000>fcT`000000TFfW00000080k[BB00@00000091]YP0000000000103c]XX020000000
lkJ:0?>fRP000000l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`070?ooo`00SP3oool2
00000;<0oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y0P0000000`3U/9`0000009EbZ008
09EbZ080000000D0i[2K000000000000ik6J0000000909AaZ080000000D0j;6I0000002DLJP00000
0>VaV00200000080jK:H0P0000020>ZbU`030000002CKjL0TfnW00P0TfnW0P0000001P3[/YH00000
0>fcT`000000TFfW00000080k[BB00@00000091]YP0000000000103c]XX020000000lkJ:0?>fRP00
0000l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`070?ooo`00S@3oool00`000000oooo
0000002c0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0
202ELZP2000000050>J`V`00000000000>NaVP0000002@2DLJP2000000050>RaV@000000U76X0000
003Y/IP00P0000020>VbV08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`80000000H0jk:F0000
003]/i<00000095]Y`0000020>jdTP040000002@KJH00000000000@0lkJ:00P000000?>fRP3c]XX0
00000?:fS0000000lKF=00000900oooo00<000000?ooo`3oool01`3oool008d0oooo0P00002d0?oo
o`030000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP20000
00050>J`V`00000000000>NaVP0000002@2DLJP2000000050>RaV@000000U76X0000003Y/IP00P00
00020>VbV08000000P3Z/YL00`000000TfnW09=_Y`0809=_Y`80000000H0jk:F0000003]/i<00000
095]Y`0000020>jdTP040000002@KJH00000000000@0lkJ:00P000000?>fRP3c]XX000000?:fS000
0000lKF=00000940oooo00<000000?ooo`3oool01P3oool008d0oooo0P00002d0?ooo`030000002F
LjT0UW>Y00H0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP2000000050>J`V`00
000000000>NaVP0000002@2DLJP2000000050>RaV@000000U76X0000003Y/IP00P0000020>VbV080
00000P3Z/YL00`000000TfnW09=_Y`0809=_Y`80000000H0jk:F0000003]/i<00000095]Y`000002
0>jdTP040000002@KJH00000000000@0lkJ:00P000000?>fRP3c]XX000000?:fS0000000lKF=0000
0940oooo00<000000?ooo`3oool01P3oool008d0oooo00<000000?ooo`3oool0/`3oool00`000000
UW>Y09IcZ@0609IcZ@<00000202FLZT2000000030>F`W0000000UG:X00P0UG:X0P0000001@3V/9/0
00000000003W/IX0000000T0U76X0P0000001@3X/IT0000009AaZ0000000jK6H008000000P3Y/YP0
1@000000Tg2X0000003Z/YL0000000X0TfnW0P000000103[/YH000000>fcT`3]/i<200000080k[BB
00@00000091]YP0000000000103c]XX010000000lkJ:0?>fRP0000020?:fS0030000003a]Hd00000
0900oooo00<000000?ooo`3oool01P3oool008d0oooo0`00002c0?ooo`030000002FLjT0UW>Y00H0
UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP2000000050>J`V`00000000000>Na
VP0000002@2DLJP2000000050>RaV@000000U76X0000003Y/IP00P0000020>VbV0050000002CL:P0
00000>ZbU`0000002P2CKjL2000000040>^bUP000000kK>C0>fcT`8000000P3^]98010000000T6fV
0000000000040?>fRP040000003c]XX0lkJ:00000080l[J<00<000000?6eS@000000T03oool00`00
0000oooo0?ooo`060?ooo`00S@3oool00`000000oooo0?ooo`2c0?ooo`030000002FLjT0UW>Y00H0
UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP2000000050>J`V`00000000000>Na
VP0000002@2DLJP2000000050>RaV@000000U76X0000003Y/IP00P0000020>VbV0050000002CL:P0
00000>ZbU`0000002P2CKjL20000000:0>^bUP000000kK>C0>fcT`000000TFfW0000003^]9800000
091]YP800000103c]XX010000000lkJ:0?>fRP0000020?:fS0030000003a]Hd000000900oooo00<0
00000?ooo`3oool01P3oool008d0oooo00<000000?ooo`3oool0/`3oool00`000000UW>Y09IcZ@06
09IcZ@<00000202FLZT2000000030>F`W0000000UG:X00P0UG:X0P0000001@3V/9/000000000003W
/IX0000000T0U76X0P0000020>RaV@80000000<0jK6H0000000000000P3Y/YP01@000000Tg2X0000
003Z/YL0000000X0TfnW00@0000009=_Y`00000000000P3]/i<02@000000TFfW0000003^]9800000
091]YP000000k[BA000000030?>fRP040000003c]XX0lkJ:00000080l[J<00<000000?6eS@000000
T03oool00`000000oooo0?ooo`060?ooo`00S@3oool00`000000oooo0?ooo`2c0?ooo`030000002F
LjT0UW>Y00H0UW>Y0`00000809IbZ@8000000P3U/9`00`000000UG:X09EbZ00609EbZ080000000D0
i[2K000000000000ik6J0000000909AaZ0040000002DLJP000000>RaV@80000000X0jK6H0000002C
L:P000000>VbV0000000Tg2X0000003Z/YL000002P2CKjL010000000TfnW0000000000020>fcT`09
0000002AKJL000000>jdTP000000T6fV0000003^]940000000<0lkJ:00@000000?>fRP3c]XX00000
0P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00H0oooo002=0?ooo`030000003o
ool0oooo0;<0oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y0P0000020>F`W0030000002E
LZP0UG:X00H0UG:X0P0000001@3V/9/000000000003W/IX0000000T0U76X00@0000009AaZ0000000
j;6I0P0000002P3Y/IP0000009=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_Y`040000
002CKjL0000000000080kK>C00T00000095]Y`000000k[BB0000002@KJH000000>jdT@0000000`3c
]XX010000000lkJ:0?>fRP0000020?:fS0030000003a]Hd000000900oooo00<000000?ooo`3oool0
1P3oool008d0oooo00<000000?ooo`3oool0/`3oool00`000000UW>Y09IcZ@0609IcZ@<00000202F
LZT200000080iK2L00<0000009EbZ02ELZP01P2ELZP2000000050>J`V`00000000000>NaVP000000
2P2DLJP20000000=0>RaV@00000000000>VaV0000000Tg2X0000003Y/YP0000009=`Z0000000j[:G
0000000:09=_Y`040000002CKjL0000000000080kK>C00T00000095]Y`000000k[BB0000002@KJH0
00000>jdT@0000000`3c]XX010000000lkJ:0?>fRP0000020?:fS0030000003a]Hd000000900oooo
00<000000?ooo`3oool01P3oool008d0oooo00<000000?ooo`3oool0/`3oool00`000000UW>Y09Ic
Z@0609IcZ@<000002@2FLZT010000000iK2L0>F`W000000809EbZ080000000D0i[2K000000000000
ik6J0000000:09AaZ080000000d0j;6I000000000000jK6H0000002CL:P000000>VbV0000000Tg2X
0000003Z/YL0000000X0TfnW00@0000009=_Y`00000000000P3]/i<02@000000TFfW0000003^]980
0000091]YP000000k[BA000000030?>fRP040000003c]XX0lkJ:00000080l[J<00<000000?6eS@00
0000T03oool00`000000oooo0?ooo`060?ooo`00J`3oool01P2oM000000000000000000004RL0=oo
o`<0oooo00D0omnL04P000000000001d0;ooo`040?ooo`050?oOW018000000000000M02oool02`3o
ool00`000000oooo0?ooo`2c0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000909IbZ@040000003U
/9`0iK2L000000P0UG:X0P0000001@3V/9/000000000003W/IX0000000X0U76X0P0000003@3X/IT0
00000000003Y/IP0000009=`Z0000000jK:H0000002CL:P000000>ZbU`0000002P2CKjL02`000000
TfnW0000003[/YH000000>fcT`000000TFfW0000003^]98000000080T6fV0P0000030?>fRP040000
003c]XX0lkJ:00000080l[J<00<000000?6eS@000000T03oool00`000000oooo0?ooo`060?ooo`00
K03oool00`3ogi`0B00007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?oo
o`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`0:0?ooo`030000003oool0oooo0;<0oooo00<0
000009IcZ@2FLjT01P2FLjT3000000T0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ080000000D0
i[2K000000000000ik6J0000000:09AaZ080000000d0j;6I000000000000jK6H0000002CL:P00000
0>VbV0000000Tg2X0000003Z/YL0000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000
095]Y`000000k[BB00000002091]YP8000000`3c]XX010000000lkJ:0?>fRP0000020?:fS0030000
003a]Hd000000900oooo00<000000?ooo`3oool01P3oool006`0oooo00<0omnL04P0001d_ol0103o
ool00`2oM00004RL0=ooo`020?ooo`070;md0000M;l0oooo0?ooo`2oM00004RL0=ooo`020?ooo`03
0;md0000M;l0oooo00T0oooo00<000000?ooo`3oool0/`3oool00`000000UW>Y09IcZ@0609IcZ@80
000000<0i:nM0000002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X0P0000001@3V/9/0
00000000003W/IX0000000X0U76X0P0000003@3X/IT000000000003Y/IP0000009=`Z0000000jK:H
0000002CL:P000000>ZbU`0000002P2CKjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW
0000003^]98000000080T6fV0P0000030?>fRP040000003c]XX0lkJ:00000080l[J<00<000000?6e
S@000000T03oool00`000000oooo0?ooo`060?ooo`00K03oool00`3ogi`0B00007Boo`030?ooo`03
0?oo_`1d0000B9cO00<0oooo00H0gia80000M02oool0oono07@00018W=l30?ooo`030=nLB00007@0
_ooo00T0oooo00<000000?ooo`3oool0/`3oool00`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM
0000002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X0P0000001P3V/9/0000009EbZ000
0000ik6J000000T0U76X0P0000003@3X/IT000000000003Y/IP0000009=`Z0000000jK:H0000002C
L:P000000>ZbU`0000002P2CKjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000003^
]98000000080T6fV0P0000030?>fRP090000003c]XX0lkJ:0?>fRP000000l[J<0000003a]Hd00000
0900oooo00<000000?ooo`3oool01P3oool006`0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0
B00007Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`09
0?ooo`030000003oool0oooo0;<0oooo00<0000009IcZ@2FLjT01P2FLjT2000000030>B_W@000000
UW:Y00L0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ080000000H0i[2K0000002ELZP000000>Na
VP00000909AaZ080000000d0j;6I000000000000jK6H0000002CL:P000000>VbV0000000Tg2X0000
003Z/YL0000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB0000
0002091]YP8000000`3c]XX02@000000lkJ:0?>fRP3c]XX000000?:fS0000000lKF=0000002@0?oo
o`030000003oool0oooo00H0oooo001/0?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d
_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol02@3oool0
0`000000oooo0?ooo`2c0?ooo`030000002FLjT0UW>Y00H0UW>Y0P0000000`3T[id0000009IbZ@07
09IbZ@80000000<0iK2L0000002ELZP01`2ELZP2000000060>J`V`000000UG:X0000003W/IX00000
2@2DLJP20000000=0>RaV@00000000000>VaV0000000Tg2X0000003Y/YP0000009=`Z0000000j[:G
0000000:09=_Y`0;0000002CKjL000000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@
KJH2000000<0lkJ:00T000000?>fRP3c]XX0lkJ:0000003b]X`000000?6eS@000000T03oool00`00
0000oooo0?ooo`060?ooo`00K03oool00`3ogi`0B00007Boo`030?ooo`030?oOW0180000M;oo00<0
oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?ooo`030?noM00004P0W=oo00T0oooo00<00000
0?ooo`3oool0/`3oool00`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT2
000000030>F`W0000000UG:X00L0UG:X0P0000001P3V/9/0000009EbZ0000000ik6J000000T0U76X
0P0000002`3X/IT000000000003Y/IP0000009=`Z0000000jK:H0000002CL:P000000080j[:G00<0
000009=_Y`2CKjL01`2CKjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000003^]980
00000080T6fV0P0000030?>fRP090000003c]XX0lkJ:0?>fRP000000l[J<0000003a]Hd000000900
oooo00<000000?ooo`3oool01P3oool006`0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B000
07Boo`030?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`090?oo
o`030000003oool0oooo0;<0oooo00<0000009IcZ@2FLjT01P2FLjT2000000030>B_W@000000UW:Y
00L0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ080000000H0i[2K0000002ELZP000000>NaVP00
000909AaZ080000000/0j;6I000000000000jK6H0000002CL:P000000>VbV0000000Tg2X00000002
0>ZbU`030000002CKjL0TfnW00L0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`00
0000k[BB00000002091]YP8000000`3c]XX02@000000lkJ:0?>fRP3c]XX000000?:fS0000000lKF=
0000002@0?ooo`030000003oool0oooo00H0oooo001/0?ooo`030?oOW0180000M;oo00<0oooo00<0
oono07@00018W=l00`3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`<0oooo00<0gia80000M02o
ool0203oool00`000000oooo0?ooo`2d0?ooo`030000002FLjT0UW>Y00H0UW>Y0P0000000`3T[id0
000009IbZ@0709IbZ@80000000<0iK2L0000002ELZP01`2ELZP2000000060>J`V`000000UG:X0000
003W/IX000002@2DLJP2000000030>RaV@000000U76X0080000000H0Tg2X0000003Y/YP0000009=`
Z00000020>ZbU`030000002CKjL0TfnW00L0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000
095]Y`000000k[BB00000002091]YP8000000`3c]XX02@000000lkJ:0?>fRP3c]XX000000?:fS000
0000lKF=0000002@0?ooo`030000003oool0oooo00H0oooo001/0?ooo`030?oOW0180000M;oo00@0
oooo00<0gia80018W03Oool00P3oool01`2oM00007Bo0?ooo`3oool0gia80018W03Oool00P3oool0
0`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo0;@0oooo00<0000009IcZ@2FLjT01P2FLjT2
000000030>B_W@000000UW:Y00L0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ080000000H0i[2K
0000002ELZP000000>NaVP00000909AaZ080000000<0j;6I0000002DLJP00P0000001P2CL:P00000
0>VbV0000000Tg2X00000080j[:G00<0000009=_Y`2CKjL01`2CKjL02`000000TfnW0000003[/YH0
00000>fcT`000000TFfW0000003^]98000000080T6fV0P0000030?>fRP090000003c]XX0lkJ:0?>f
RP000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool01P3oool006X0oooo00D0oooO
09a800000000000007Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`800000]@3oool00`000000UW>Y09IcZ@06
09IcZ@80000000<0i:nM0000002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X0P000000
1P3V/9/0000009EbZ0000000ik6J000000T0U76X0P0000000`3X/IT0000009AaZ0020000000409=`
Z0000000jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`0;0000002CKjL000000>^b
UP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000<0lkJ:00T000000?>fRP3c]XX0
lkJ:0000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`060?ooo`00K03oool00`3o
_g@0000007Boo`050?ooo`050?oo_`1d000000000000M02oool0103oool01@3ookl0M00000000000
07@0_ooo00X0oooo00@000000?ooo`0000000000/`3oool00`000000UW>Y09IcZ@0609IcZ@800000
00<0i:nM0000002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X0P0000001P3V/9/00000
09EbZ0000000ik6J000000T0U76X0P0000000`3X/IT0000009AaZ0020000000409=`Z0000000jK:H
0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`0;0000002CKjL000000>^bUP000000kK>C
0000002AKJL000000>jdTP0000000P2@KJH2000000<0lkJ:00D000000?>fRP3c]XX0lkJ:00000002
0?:fS0030000003a]Hd000000900oooo00<000000?ooo`3oool01@3oool008`0oooo00<000000?oo
o`3oool0]03oool00`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT20000
00030>F`W0000000UG:X00L0UG:X0P0000001P3V/9/0000009EbZ0000000ik6J000000T0U76X0P00
00000`3X/IT0000009AaZ0020000000409=`Z0000000jK:H0>VbV08000000P3Z/YL00`000000TfnW
09=_Y`0709=_Y`0;0000002CKjL000000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@
KJH2000000<0lkJ:00D000000?>fRP3c]XX0lkJ:000000020?:fS0030000003a]Hd000000900oooo
00<000000?ooo`3oool01@3oool008`0oooo00<000000?ooo`3oool0]03oool00`000000UW>Y09Ic
Z@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X0P00
00001P3V/9/0000009EbZ0000000ik6J000000T0U76X0P0000000`3X/IT0000009AaZ00200000004
09=`Z0000000jK:H0>VbV08000000P3Z/YL00`000000TfnW09=_Y`0709=_Y`0;0000002CKjL00000
0>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000<0lkJ:00D000000?>fRP3c
]XX0lkJ:000000020?:fS0030000003a]Hd000000900oooo00<000000?ooo`3oool01@3oool008`0
oooo00<000000?ooo`3oool0]03oool00`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000002F
LZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X0P0000001P3V/9/0000009EbZ0000000ik6J
000000T0U76X0P0000001@3X/IT0000009AaZ0000000jK6H008000000P3Y/YP200000080j[:G00<0
000009=_Y`2CKjL01`2CKjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000003^]980
00000080T6fV00<000000>jdT@0000000`3c]XX010000000lkJ:0?>fRP0000020?:fS0030000003a
]Hd000000900oooo00<000000?ooo`3oool01@3oool008`0oooo00<000000?ooo`3oool0]03oool0
0`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT2000000030>F`W0000000
UG:X00L0UG:X0P0000001P3V/9/0000009EbZ0000000ik6J000000T0U76X0P0000001@3X/IT00000
09AaZ0000000jK6H008000000P3Y/YP01@000000Tg2X0000003Z/YL0000000T0TfnW00/0000009=_
Y`000000jk:F0000003]/i<00000095]Y`000000k[BB00000002091]YP030000003^]940000000<0
lkJ:00@000000?>fRP3c]XX000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo
00D0oooo002<0?ooo`030000003oool0oooo0;@0oooo00<0000009IcZ@2FLjT01P2FLjT200000003
0>B_W@000000UW:Y00L0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ0080000002ELZP000000000
002ELZP000000>NaVP00000909AaZ080000000D0j;6I0000002DLJP000000>VaV00200000080jK:H
00D0000009=`Z0000000j[:G0000000909=_Y`070000002CKjL000000>^bUP000000kK>C00000002
095]Y`070000003^]9800000091]YP000000k[BA000000030?>fRP040000003c]XX0lkJ:00000080
l[J<00<000000?6eS@000000T03oool00`000000oooo0?ooo`050?ooo`00S03oool00`000000oooo
0?ooo`2d0?ooo`030000002FLjT0UW>Y00H0UW>Y0P0000000`3T[id0000009IbZ@0709IbZ@800000
00<0iK2L0000002ELZP01`2ELZP020000000UG:X000000000000UG:X0000003W/IX000002@2DLJP2
000000050>RaV@000000U76X0000003Y/IP00P0000020>VbV0050000002CL:P000000>ZbU`000000
2@2CKjL01`000000TfnW0000003[/YH000000>fcT`0000000P2AKJL01`000000k[BB0000002@KJH0
00000>jdT@0000000`3c]XX010000000lkJ:0?>fRP0000020?:fS0030000003a]Hd000000900oooo
00<000000?ooo`3oool01@3oool008`0oooo00<000000?ooo`3oool0]03oool00`000000UW>Y09Ic
Z@0609IcZ@80000000<0i:nM0000002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X00P0
000009EbZ0000000000009EbZ0000000ik6J000000T0U76X0P0000001@3X/IT0000009AaZ0000000
jK6H008000000P3Y/YP01@000000Tg2X0000003Z/YL0000000T0TfnW00<0000009=_Y`2CKjL00P00
00002`3]/i<00000095]Y`2AKJL000000>jdTP000000T6fV0000003^]940000000<0lkJ:00@00000
0?>fRP3c]XX000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00D0oooo002<
0?ooo`030000003oool0oooo0;@0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000
000809IbZ@80000000<0iK2L0000002ELZP01`2ELZP020000000UG:X000000000000UG:X0000003W
/IX000002@2DLJP20000000>0>RaV@000000U76X0000003Y/IP0000009=`Z0000000jK:H0000002C
L:P000000>ZbU`00000:09=_Y`060000002CKjL000000000003]/i<000000P2AKJL01`000000k[BB
0000002@KJH000000>jdT@0000000`3c]XX010000000lkJ:0?>fRP0000020?:fS0030000003a]Hd0
00000900oooo00<000000?ooo`3oool01@3oool008`0oooo00<000000?ooo`3oool0]03oool00`00
0000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y0P0000000`3U/9`0000009EbZ007
09EbZ0080000002ELZP000000000002ELZP000000>NaVP00000909AaZ080000000h0j;6I0000002D
LJP000000>VaV0000000Tg2X0000003Y/YP0000009=`Z0000000j[:G000000X0TfnW00@0000009=_
Y`000000jk:F0P000002095]Y`030000003^]98000000080T6fV0P0000030?>fRP040000003c]XX0
lkJ:00000080l[J<00<000000?6eS@000000T03oool00`000000oooo0?ooo`050?ooo`00S03oool0
0`000000oooo0?ooo`2d0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202F
LZT2000000030>F`W0000000UG:X00L0UG:X00@0000009EbZ0000000i[2K0P0000000`3W/IX00000
09AaZ00809AaZ080000000h0j;6I0000002DLJP000000>VaV0000000Tg2X0000003Y/YP0000009=`
Z0000000j[:G000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB
00000002091]YP8000000`3c]XX02@000000lkJ:0?>fRP3c]XX000000?:fS0000000lKF=0000002@
0?ooo`030000003oool0oooo00D0oooo002<0?ooo`030000003oool0oooo0;@0oooo00<0000009Ic
Z@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@80000000<0iK2L0000002ELZP01`2ELZP0
10000000UG:X0000003V/9/2000000030>NaVP000000U76X00P0U76X0P0000003P3X/IT0000009Aa
Z0000000jK6H0000002CL:P000000>VbV0000000Tg2X0000003Z/YL000002P2CKjL02`000000TfnW
0000003[/YH000000>fcT`000000TFfW0000003^]98000000080T6fV0P0000030?>fRP090000003c
]XX0lkJ:0?>fRP000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool01@3oool008`0
oooo00<000000?ooo`3oool0]03oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT000000000
00P0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ0040000002ELZP000000>J`V`80000000<0ik6J
0000002DLJP0202DLJP040000000U76X000000000000U76X0000003Y/IP0000009=`Z0000000jK:H
0000002CL:P000000>ZbU`00000:09=_Y`0;0000002CKjL000000>^bUP000000kK>C0000002AKJL0
00000>jdTP0000000P2@KJH2000000<0lkJ:00T000000?>fRP3c]XX0lkJ:0000003b]X`000000?6e
S@000000T03oool00`000000oooo0?ooo`050?ooo`00S03oool00`000000oooo0?ooo`2d0?ooo`03
0000002FLjT0UW>Y00L0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP200000005
0>J`V`00000000000>NaVP0000002@2DLJP040000000U76X000000000000U76X0000003Y/IP00000
09=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_Y`0;0000002CKjL000000>^bUP000000
kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000<0lkJ:00T000000?>fRP3c]XX0lkJ:0000
003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`050?ooo`00S03oool00`000000oooo
0?ooo`2d0?ooo`030000002FLjT0UW>Y00L0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0
202ELZP2000000050>J`V`00000000000>NaVP0000002@2DLJP040000000U76X000000000000U76X
0000003Y/IP0000009=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_Y`0;0000002CKjL0
00000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000<0lkJ:00T000000?>f
RP3c]XX0lkJ:0000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`050?ooo`00S03o
ool00`000000oooo0?ooo`2e0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@80000000<0
iK2L0000002ELZP0202ELZP2000000050>J`V`00000000000>NaVP0000002@2DLJP040000000U76X
000000000000U76X0000003Y/IP0000009=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_
Y`0;0000002CKjL000000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000<0
lkJ:00T000000?>fRP3c]XX0lkJ:0000003b]X`000000?6eS@000000T03oool00`000000oooo0?oo
o`050?ooo`00R`3oool200000;L0oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y0P000000
0`3U/9`0000009EbZ00809EbZ080000000D0i[2K000000000000ik6J0000000909AaZ00@0000002D
LJP000000000002DLJP000000>VaV0000000Tg2X0000003Y/YP0000009=`Z0000000j[:G000000X0
TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB00000002091]YP800000
0`3c]XX02@000000lkJ:0?>fRP3c]XX000000?:fS0000000lKF=0000002@0?ooo`030000003oool0
oooo00D0oooo002;0?ooo`030000003oool000000;H0oooo00<0000009IcZ@2FLjT01P2FLjT30000
00P0UW:Y0P0000000`3U/9`0000009EbZ00809EbZ080000000D0i[2K000000000000ik6J00000009
09AaZ00@0000002DLJP000000000002DLJP000000>VaV0000000Tg2X0000003Y/YP0000009=`Z000
0000j[:G000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB0000
0002091]YP8000000`3c]XX02@000000lkJ:0?>fRP3c]XX000000?:fS0000000lKF=0000002@0?oo
o`030000003oool0oooo00D0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2F
LjT01P2FLjT3000000P0UW:Y0P0000000`3U/9`0000009EbZ00809EbZ080000000<0i[2K0000002E
LZP00P00000909AaZ00@0000002DLJP000000000002DLJP000000>VaV0000000Tg2X0000003Y/YP0
000009=`Z0000000j[:G000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`00
0000k[BB00000002091]YP8000000`3c]XX01@000000lkJ:0?>fRP3c]XX000000080l[J<0P00002@
0?ooo`030000003oool0oooo00D0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009Ic
Z@2FLjT01P2FLjT3000000P0UW:Y0P0000000`3U/9`0000009EbZ00809EbZ080000000<0i[2K0000
002ELZP00P00000909AaZ00@0000002DLJP000000000002DLJP000000>VaV0000000Tg2X0000003Y
/YP0000009=`Z0000000j[:G000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]
Y`000000k[BB00000002091]YP8000000`3c]XX01@000000lkJ:0?>fRP3c]XX000000080l[J<00<0
00000?6eS@000000S`3oool00`000000oooo0?ooo`050?ooo`00R`3oool00`000000oooo0?ooo`2f
0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@80000000<0iK2L0000002ELZP0202ELZP2
000000030>J`V`000000UG:X008000002@2DLJP040000000U76X000000000000U76X0000003Y/IP0
000009=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_Y`0;0000002CKjL000000>^bUP00
0000kK>C0000002AKJL000000>jdTP0000000P2@KJH2000000<0lkJ:00D000000?>fRP3c]XX0lkJ:
000000020?:fS0030000003a]Hd000000900oooo00<000000?ooo`3oool0103oool008/0oooo00<0
00000?ooo`3oool0]P3oool00`000000UW>Y09IcZ@0609IcZ@<00000202FLZT2000000030>F`W000
0000UG:X00P0UG:X0P0000000`3V/9/0000009EbZ002000000T0U76X0100000009AaZ00000000000
09AaZ0000000jK6H0000002CL:P000000>VbV0000000Tg2X0000003Z/YL000002P2CKjL02`000000
TfnW0000003[/YH000000>fcT`000000TFfW0000003^]98000000080T6fV0P0000030?>fRP050000
003c]XX0lkJ:0?>fRP0000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0oooo00@0
oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y
0P0000000`3U/9`0000009EbZ00809EbZ080000000<0i[2K0000002ELZP00P00000909AaZ0040000
002DLJP0000000000080U76X0P000000202CL:P000000>VbV0000000Tg2X0000003Z/YL000002P2C
KjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000003^]98000000080T6fV0P000003
0?>fRP050000003c]XX0lkJ:0?>fRP0000000P3b]X`00`000000lKF=0000002@0?ooo`030000003o
ool0oooo00@0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2FLjT01P2FLjT3
000000P0UW:Y00@0000009IbZ@00000000002@2ELZP2000000030>J`V`000000UG:X008000002@2D
LJP010000000U76X00000000000209AaZ080000000P0Tg2X0000003Y/YP0000009=`Z0000000j[:G
000000X0TfnW00/0000009=_Y`000000jk:F0000003]/i<00000095]Y`000000k[BB00000002091]
YP030000003^]94000000080lkJ:00D000000?>fRP3c]XX0lkJ:000000020?:fS0030000003a]Hd0
00000900oooo00<000000?ooo`3oool0103oool008/0oooo00<000000?ooo`3oool0]P3oool00`00
0000UW>Y09IcZ@0609IcZ@<00000202FLZT010000000UW:Y00000000000909EbZ080000000<0i[2K
0000002ELZP00P00000909AaZ0060000002DLJP000000>RaV@000000U76X0P000000202CL:P00000
0>VbV0000000Tg2X0000003Z/YL000002P2CKjL02`000000TfnW0000003[/YH000000>fcT`000000
TFfW0000003^]98000000080T6fV00<000000>jdT@0000000P3c]XX01@000000lkJ:0?>fRP3c]XX0
00000080l[J<00<000000?6eS@000000T03oool00`000000oooo0?ooo`040?ooo`00R`3oool00`00
0000oooo0?ooo`2f0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@040000002FLZT00000
000000T0UG:X0P0000000`3V/9/0000009EbZ002000000T0U76X00H0000009AaZ0000000j;6I0000
002DLJP20000000809=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_Y`0;0000002CKjL0
00000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH00`000000k[BA000000020?>f
RP050000003c]XX0lkJ:0?>fRP0000000P3b]X`00`000000lKF=0000002@0?ooo`030000003oool0
oooo00@0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2FLjT01P2FLjT30000
00P0UW:Y00@0000009IbZ@00000000002@2ELZP2000000030>J`V`000000UG:X008000002@2DLJP0
1P000000U76X0000003X/IT0000009AaZ080000000P0Tg2X0000003Y/YP0000009=`Z0000000j[:G
000000X0TfnW00L0000009=_Y`000000jk:F0000003]/i<000000080TFfW0P000002091]YP030000
003^]94000000080lkJ:00D000000?>fRP3c]XX0lkJ:000000020?:fS0030000003a]Hd000000900
oooo00<000000?ooo`3oool0103oool008/0oooo00<000000?ooo`3oool0]P3oool00`000000UW>Y
09IcZ@0609IcZ@<00000202FLZT010000000UW:Y00000000000909EbZ080000000<0i[2K0000002E
LZP00P00000:09AaZ080000000<0j;6I0000002DLJP00P000000102CL:P000000>VbV000000209=`
Z08000002P2CKjL01`000000TfnW0000003[/YH000000>fcT`0000000P2AKJL200000080T6fV00<0
00000>jdT@0000000P3c]XX01@000000lkJ:0?>fRP3c]XX000000080l[J<00<000000?6eS@000000
T03oool00`000000oooo0?ooo`040?ooo`00R`3oool00`000000oooo0?ooo`2f0?ooo`030000002F
LjT0UW>Y00H0UW>Y0`00000809IbZ@040000002FLZT00000000000T0UG:X0P0000000`3V/9/00000
09EbZ002000000X0U76X0P0000000`3X/IT0000009AaZ0020000000409=`Z0000000jK:H00000080
Tg2X0P00000:09=_Y`070000002CKjL000000>^bUP000000kK>C00000002095]Y`8000000P2@KJH0
0`000000k[BA000000020?>fRP050000003c]XX0lkJ:0?>fRP0000000P3b]X`00`000000lKF=0000
002@0?ooo`030000003oool0oooo00@0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<00000
09IcZ@2FLjT01P2FLjT3000000P0UW:Y00@0000009IbZ@00000000002@2ELZP2000000030>J`V`00
0000UG:X008000002P2DLJP2000000030>RaV@000000U76X0080000000@0Tg2X0000003Y/YP00000
0P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL00P000000103]/i<00000095]Y`2AKJL2000000<0
T6fV0P0000020?>fRP030000003c]XX0lkJ:0080lkJ:00D000000?:fS0000000lKF=0000002@0?oo
o`030000003oool0oooo00@0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2F
LjT01P2FLjT3000000P0UW:Y00@0000009IbZ@00000000002@2ELZP2000000060>J`V`000000UG:X
0000003W/IX000002@2DLJP2000000030>RaV@000000U76X008000000P2CL:P200000080Tg2X0P00
000:09=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV0P0000020?>fRP030000003c
]XX0lkJ:0080lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003oool0oooo00@0oooo
002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y00@0
000009IbZ@00000000002@2ELZP2000000060>J`V`000000UG:X0000003W/IX000002@2DLJP20000
00030>RaV@000000U76X008000000P2CL:P200000080Tg2X00<000000>ZbU`0000002@2CKjL00`00
0000TfnW09=_Y`0400000080TFfW0P000003091]YP8000000P3c]XX00`000000lkJ:0?>fRP020?>f
RP050000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`040?ooo`00R`3oool00`00
0000oooo0?ooo`2f0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000809IbZ@040000002FLZT00000
000000T0UG:X0P0000001P3V/9/0000009EbZ0000000ik6J000000T0U76X0P0000000`3X/IT00000
09AaZ00200000080Tg2X0P00000209=`Z0030000003Z/YL0000000T0TfnW00<0000009=_Y`2CKjL0
10000002095]Y`8000000`2@KJH200000080lkJ:00<000000?>fRP3c]XX00P3c]XX01@000000l[J<
0000003a]Hd000000900oooo00<000000?ooo`3oool0103oool008/0oooo00<000000?ooo`3oool0
]P3oool00`000000UW>Y09IcZ@0609IcZ@<00000202FLZT010000000UW:Y00000000000909EbZ080
000000H0i[2K0000002ELZP000000>NaVP00000909AaZ080000000<0j;6I0000002DLJP00P000002
09=`Z08000000P2CL:P00`000000j[:G0000000909=_Y`030000002CKjL0TfnW00@000000P2AKJL2
000000<0T6fV0P0000020?>fRP030000003c]XX0lkJ:0080lkJ:00D000000?:fS0000000lKF=0000
002@0?ooo`030000003oool0oooo00@0oooo002;0?ooo`800000]`3oool00`000000UW>Y09IcZ@06
09IcZ@<00000202FLZT010000000UW:Y00000000000909EbZ0080000002ELZP000000000002ELZP0
00000>NaVP00000909AaZ080000000<0j;6I0000002DLJP00P00000209=`Z0070000003Y/YP00000
09=`Z0000000j[:G0000000909=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV0P00
00020?>fRP030000003c]XX0lkJ:0080lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000
003oool0oooo00@0oooo002;0?ooo`030000003oool0oooo0;H0oooo00<0000009IcZ@2FLjT01P2F
LjT3000000P0UW:Y00@0000009IbZ@00000000002@2ELZP020000000UG:X000000000000UG:X0000
003W/IX000002@2DLJP2000000030>RaV@000000U76X008000000P2CL:P01`000000jK:H0000002C
L:P000000>ZbU`0000002@2CKjL00`000000TfnW09=_Y`0400000080TFfW00<000000>jdTP000000
0P2@KJH200000080lkJ:00<000000?>fRP3c]XX00P3c]XX01@000000l[J<0000003a]Hd000000900
oooo00<000000?ooo`3oool0103oool008X0oooo00<000000?ooo`3oool0]`3oool00`000000UW>Y
09IcZ@0609IcZ@<00000202FLZT010000000UW:Y00000000000909EbZ0080000002ELZP000000000
002ELZP000000>NaVP00000909AaZ080000000<0j;6I0000002DLJP00P00000209=`Z0070000003Y
/YP0000009=`Z0000000j[:G0000000909=_Y`050000002CKjL0TfnW0000003[/YH00P000002095]
Y`030000003^]98000000080T6fV0P0000020?>fRP030000003c]XX0lkJ:0080lkJ:00<000000?:f
S03b]X`00P00002@0?ooo`030000003oool0oooo00@0oooo002:0?ooo`030000003oool0oooo0;L0
oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y00@0000009IbZ@00000000002@2ELZP02000
0000UG:X000000000000UG:X0000003W/IX000002@2DLJP20000000>0>RaV@000000U76X0000003Y
/IP0000009=`Z0000000jK:H0000002CL:P000000>ZbU`00000909=_Y`050000002CKjL0TfnW0000
003[/YH00P000002095]Y`030000003^]98000000080T6fV0P0000020?>fRP030000003c]XX0lkJ:
0080lkJ:00<000000?:fS03b]X`00P00002@0?ooo`030000003oool0oooo00@0oooo002:0?ooo`03
0000003oool0oooo0;L0oooo00<0000009IcZ@2FLjT01P2FLjT3000000P0UW:Y00D0000009IbZ@00
0000iK2L0000000809EbZ0080000002ELZP000000000002ELZP000000>NaVP00000909AaZ0800000
00h0j;6I0000002DLJP000000>VaV0000000Tg2X0000003Y/YP0000009=`Z0000000j[:G000000T0
TfnW00D0000009=_Y`2CKjL000000>^bUP0200000080TFfW00<000000>jdTP0000000P2@KJH20000
0080lkJ:00<000000?>fRP3c]XX00P3c]XX01P000000l[J<0?:fS0000000lKF=000008l0oooo00<0
00000?ooo`3oool0103oool008X0oooo00<000000?ooo`3oool0]`3oool00`000000UW>Y09IcZ@06
09IcZ@<00000202FLZT01@000000UW:Y0000003U/9`0000000P0UG:X00P0000009EbZ00000000000
09EbZ0000000ik6J000000T0U76X0100000009AaZ0000000000009AaZ0000000jK6H0000002CL:P0
00000>VbV0000000Tg2X0000003Z/YL000002@2CKjL01@000000TfnW09=_Y`000000jk:F00800000
0P2AKJL00`000000k[BB00000002091]YP8000000P3c]XX00`000000lkJ:0?>fRP020?>fRP060000
003b]X`0l[J<0000003a]Hd00000S`3oool00`000000oooo0?ooo`040?ooo`00RP3oool00`000000
oooo0?ooo`2g0?ooo`030000002FLjT0UW>Y00H0UW>Y0`00000909IbZ@80000000<0iK2L0000002E
LZP01`2ELZP020000000UG:X000000000000UG:X0000003W/IX000002@2DLJP040000000U76X0000
00000000U76X0000003Y/IP0000009=`Z0000000jK:H0000002CL:P000000>ZbU`00000:09=_Y`0;
0000002CKjL000000>^bUP000000kK>C0000002AKJL000000>jdTP0000000P2@KJH200000080lkJ:
00<000000?>fRP3c]XX00P3c]XX01P000000l[J<0?:fS0000000lKF=000008l0oooo00<000000?oo
o`3oool0103oool008X0oooo00<000000?ooo`3oool0]`3oool00`000000UW>Y09IcZ@0609IcZ@<0
00002@2FLZT2000000030>F`W0000000UG:X00L0UG:X00P0000009EbZ0000000000009EbZ0000000
ik6J000000T0U76X0100000009AaZ0000000000009AaZ0000000jK6H0000002CL:P000000>VbV000
0000Tg2X0000003Z/YL000002P2CKjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000
003^]98000000080T6fV0P0000020?>fRP030000003c]XX0lkJ:0080lkJ:00H000000?:fS03b]X`0
00000?6eS@00002@0?ooo`030000003oool0oooo00<0oooo002:0?ooo`030000003oool0oooo0;L0
oooo00<0000009IcZ@2FLjT01P2FLjT3000000T0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ008
0000002ELZP000000000002ELZP000000>NaVP00000909AaZ00@0000002DLJP000000000002DLJP0
00000>VaV0000000Tg2X0000003Y/YP0000009=`Z0000000j[:G000000X0TfnW00/0000009=_Y`00
0000jk:F0000003]/i<00000095]Y`000000k[BB00000002091]YP050000003^]94000000?>fRP00
0000103c]XX01P000000l[J<0?:fS0000000lKF=00000900oooo00<000000?ooo`3oool00`3oool0
08X0oooo00<000000?ooo`3oool0]`3oool00`000000UW>Y09IcZ@0609IcZ@80000000<0i:nM0000
002FLZT01`2FLZT2000000030>F`W0000000UG:X00L0UG:X00P0000009EbZ0000000000009EbZ000
0000ik6J000000T0U76X0100000009AaZ0000000000009AaZ0000000jK6H0000002CL:P000000>Vb
V0000000Tg2X0000003Z/YL000002P2CKjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW
0000003^]98000000080T6fV00D000000>jdT@000000lkJ:000000040?>fRP060000003b]X`0l[J<
0000003a]Hd00000T03oool00`000000oooo0?ooo`030?ooo`00RP3oool00`000000oooo0?ooo`2g
0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT2000000030>F`W000
0000UG:X00L0UG:X00P0000009EbZ0000000000009EbZ0000000ik6J000000T0U76X0100000009Aa
Z0000000000009AaZ0000000jK6H0000002CL:P000000>VbV0000000Tg2X0000003Z/YL000002P2C
KjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000003^]98000000080T6fV00D00000
0>jdT@000000lkJ:000000040?>fRP060000003b]X`0l[J<0000003a]Hd00000T03oool00`000000
oooo0?ooo`030?ooo`00RP3oool00`000000oooo0?ooo`2g0?ooo`030000002FLjT0UW>Y00H0UW>Y
00@0000009IcZ@0000000000202FLZT2000000030>F`W0000000UG:X00L0UG:X00@0000009EbZ000
000000000P2ELZP2000000T0U76X0100000009AaZ0000000000009AaZ0000000jK6H0000002CL:P0
00000>VbV0000000Tg2X0000003Z/YL000002P2CKjL02`000000TfnW0000003[/YH000000>fcT`00
0000TFfW0000003^]98000000080T6fV00D000000>jdT@000000lkJ:000000040?>fRP060000003b
]X`0l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`030?ooo`00RP3oool00`000000oooo
0?ooo`2g0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT200000003
0>F`W0000000UG:X00L0UG:X00@0000009EbZ000000000000P2ELZP2000000T0U76X0100000009Aa
Z0000000000009AaZ0000000jK6H0000002CL:P000000>VbV0000000Tg2X0000003Z/YL000002P2C
KjL02`000000TfnW0000003[/YH000000>fcT`000000TFfW0000003^]98000000080T6fV00D00000
0>jdT@000000lkJ:000000040?>fRP060000003b]X`0l[J<0000003a]Hd00000T03oool00`000000
oooo0?ooo`030?ooo`00RP3oool00`000000oooo0?ooo`2g0?ooo`030000002FLjT0UW>Y00H0UW>Y
00@0000009IcZ@0000000000202FLZT2000000030>F`W0000000UG:X00L0UG:X00@0000009EbZ000
000000000P2ELZP2000000T0U76X00@0000009AaZ000000000000P2DLJP20000000809=`Z0000000
jK:H0000002CL:P000000>ZbU`00000:09=_Y`070000002CKjL000000>^bUP000000kK>C00000002
095]Y`8000000P2@KJH01@000000k[BA0000003c]XX0000000D0lkJ:00D000000?:fS0000000lKF=
0000002@0?ooo`030000003oool0oooo00<0oooo002:0?ooo`030000003oool0oooo0;L0oooo00<0
000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@80000000<0iK2L0000002ELZP0
1`2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000U76X00000000000209Aa
Z080000000P0Tg2X0000003Y/YP0000009=`Z0000000j[:G000000X0TfnW00L0000009=_Y`000000
jk:F0000003]/i<000000080TFfW0P000002091]YP050000003^]94000000?>fRP0000001@3c]XX0
1@000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool00`3oool008X0oooo00<00000
0?ooo`3oool0]`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y0P00
00000`3U/9`0000009EbZ00709EbZ0040000002ELZP0000000000080UG:X0P00000909AaZ0040000
002DLJP0000000000080U76X0P000000202CL:P000000>VbV0000000Tg2X0000003Z/YL000002P2C
KjL01`000000TfnW0000003[/YH000000>fcT`0000000P2AKJL2000000<0T6fV0P0000000`3c]XX0
00000?>fRP040?>fRP050000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`030?oo
o`00RP3oool00`000000oooo0?ooo`2g0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@00
00000000202FLZT2000000030>F`W0000000UG:X00L0UG:X00@0000009EbZ000000000000P2ELZP2
000000T0U76X00@0000009AaZ000000000000P2DLJP20000000409=`Z0000000jK:H00000080Tg2X
0P00000:09=_Y`070000002CKjL000000>^bUP000000kK>C00000002095]Y`8000000`2@KJH20000
00030?>fRP000000lkJ:00@0lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003oool0
oooo00<0oooo002:0?ooo`030000003oool0oooo0;L0oooo00<0000009IcZ@2FLjT01P2FLjT01000
0000UW>Y00000000000809IbZ@80000000<0iK2L0000002ELZP01`2ELZP010000000UG:X00000000
000209EbZ08000002@2DLJP010000000U76X00000000000209AaZ080000000@0Tg2X0000003Y/YP0
00000P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL010000002095]Y`8000000`2@KJH200000003
0?>fRP000000lkJ:00@0lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003oool0oooo
00<0oooo002:0?ooo`800000^03oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT000000000
00P0UW:Y0P0000000`3U/9`0000009EbZ00709EbZ0040000002ELZP0000000000080UG:X0P000009
09AaZ0040000002DLJP0000000000080U76X0P000000102CL:P000000>VbV000000209=`Z0800000
2P2CKjL00`000000TfnW09=_Y`0400000080TFfW0P000003091]YP80000000<0lkJ:0000003c]XX0
103c]XX01@000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool00`3oool008X0oooo
00<000000?ooo`3oool0]`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0
UW:Y0P0000000`3U/9`0000009EbZ00709EbZ0040000002ELZP0000000000080UG:X0P00000909Aa
Z0040000002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00<00000
09=_Y`2CKjL010000002095]Y`8000000`2@KJH2000000030?>fRP000000lkJ:00@0lkJ:00D00000
0?:fS0000000lKF=0000002@0?ooo`030000003oool0oooo00<0oooo002:0?ooo`030000003oool0
oooo0;L0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@80000000<0
iK2L0000002ELZP01`2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000U76X
00000000000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@0
00000P2AKJL2000000<0T6fV0P0000000`3c]XX000000?>fRP040?>fRP050000003b]X`000000?6e
S@000000T03oool00`000000oooo0?ooo`030?ooo`00R@3oool00`000000oooo0?ooo`2h0?ooo`03
0000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT010000000UW:Y000000000008
09EbZ0060000002ELZP000000>J`V`000000UG:X0P00000909AaZ0040000002DLJP0000000000080
U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL010000002095]Y`80
00000`2@KJH2000000030?>fRP000000lkJ:00@0lkJ:00<000000?:fS03b]X`00P00002@0?ooo`03
0000003oool0oooo00<0oooo00290?ooo`030000003oool0oooo0;P0oooo00<0000009IcZ@2FLjT0
1P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000P0UG:X00H0000009Eb
Z0000000i[2K0000002ELZP2000000T0U76X00@0000009AaZ000000000000P2DLJP200000080Tg2X
0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`0400000080TFfW0P000003091]YP800000
00<0lkJ:0000003c]XX0103c]XX00`000000l[J<0?:fS00200000900oooo00<000000?ooo`3oool0
0`3oool008T0oooo00<000000?ooo`3oool0^03oool00`000000UW>Y09IcZ@0609IcZ@040000002F
LjT00000000000P0UW:Y00@0000009IbZ@00000000002@2ELZP2000000030>J`V`000000UG:X0080
00002@2DLJP010000000U76X00000000000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_
Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV0P0000000`3c]XX000000?>fRP040?>f
RP030000003b]X`0l[J<00800000T03oool00`000000oooo0?ooo`030?ooo`00R@3oool00`000000
oooo0?ooo`2h0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT01000
0000UW:Y00000000000909EbZ080000000<0i[2K0000002ELZP00P00000909AaZ0040000002DLJP0
000000000080U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL01000
0002095]Y`8000000`2@KJH2000000030?>fRP000000lkJ:00@0lkJ:00H000000?:fS03b]X`00000
0?6eS@00002?0?ooo`030000003oool0oooo00<0oooo00290?ooo`030000003oool0oooo0;P0oooo
00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT000000000
00T0UG:X0P0000000`3V/9/0000009EbZ002000000T0U76X00@0000009AaZ000000000000P2DLJP2
00000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`0400000080TFfW0P000003
091]YP80000000<0lkJ:0000003c]XX0103c]XX01P000000l[J<0?:fS0000000lKF=000008l0oooo
00<000000?ooo`3oool00`3oool008T0oooo00<000000?ooo`3oool0^03oool00`000000UW>Y09Ic
Z@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@00000000002@2ELZP200000003
0>J`V`000000UG:X008000002@2DLJP010000000U76X00000000000209AaZ08000000P2CL:P20000
0080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV0P0000000`3c
]XX000000?>fRP040?>fRP060000003b]X`0l[J<0000003a]Hd00000S`3oool00`000000oooo0?oo
o`030?ooo`00R@3oool00`000000oooo0?ooo`2h0?ooo`030000002FLjT0UW>Y00H0UW>Y00@00000
09IcZ@0000000000202FLZT010000000UW:Y00000000000909EbZ080000000<0i[2K0000002ELZP0
0P00000909AaZ0040000002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P2000000X0
TfnW00<0000009=_Y`2CKjL010000002095]Y`8000000`2@KJH2000000030?>fRP000000lkJ:00@0
lkJ:00H000000?:fS03b]X`000000?6eS@00002?0?ooo`030000003oool0oooo00<0oooo00290?oo
o`030000003oool0oooo0;P0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y000000000008
09IbZ@040000002FLZT00000000000T0UG:X0P0000000`3V/9/0000009EbZ002000000T0U76X00@0
000009AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW
09=_Y`0400000080TFfW0P000003091]YP@000001@3c]XX01P000000l[J<0?:fS0000000lKF=0000
0900oooo00<000000?ooo`3oool00P3oool008T0oooo00<000000?ooo`3oool0^03oool00`000000
UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@00000000002@2ELZP0
1@000000UG:X000000000000UG:X008000002@2DLJP01P000000U76X0000003X/IT0000009AaZ080
00000P2CL:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0
T6fV100000050?>fRP060000003b]X`0l[J<0000003a]Hd00000T03oool00`000000oooo0?ooo`02
0?ooo`00R@3oool00`000000oooo0?ooo`2h0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009Ic
Z@0000000000202FLZT010000000UW:Y00000000000909EbZ0050000002ELZP000000000002ELZP0
0P00000909AaZ0060000002DLJP000000>RaV@000000U76X0P00000209=`Z08000000P2CL:P20000
00X0TfnW00<0000009=_Y`2CKjL010000002095]Y`8000000`2@KJH4000000D0lkJ:00H000000?:f
S03b]X`000000?6eS@00002@0?ooo`030000003oool0oooo0080oooo00290?ooo`030000003oool0
oooo0;P0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002F
LZT00000000000T0UG:X00D0000009EbZ0000000000009EbZ002000000T0U76X00H0000009AaZ000
0000j;6I0000002DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`04
00000080TFfW0P000003091]YP@000001@3c]XX01P000000l[J<0?:fS0000000lKF=00000900oooo
00<000000?ooo`3oool00P3oool008T0oooo00<000000?ooo`3oool0^03oool00`000000UW>Y09Ic
Z@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@00000000002@2ELZP01@000000
UG:X000000000000UG:X008000002P2DLJP01@000000U76X000000000000U76X008000000P2CL:P2
00000080Tg2X00<000000>ZbU`0000002@2CKjL00`000000TfnW09=_Y`0400000080TFfW00<00000
0>jdTP0000000P2@KJH4000000H0lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003o
ool0oooo0080oooo00290?ooo`030000003oool0oooo0;P0oooo00<0000009IcZ@2FLjT01P2FLjT0
10000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00D0000009EbZ0000000
000009EbZ002000000X0U76X00D0000009AaZ0000000000009AaZ00200000080Tg2X0P00000209=`
Z0030000003Z/YL0000000T0TfnW00<0000009=_Y`2CKjL010000002095]Y`030000003^]9800000
0080T6fV100000060?>fRP050000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`02
0?ooo`00R@3oool00`000000oooo0?ooo`2h0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009Ic
Z@0000000000202FLZT010000000UW:Y00000000000909EbZ0080000002ELZP000000000002ELZP0
00000>NaVP00000909AaZ0050000002DLJP000000000002DLJP00P00000209=`Z08000000P2CL:P0
0`000000j[:G0000000909=_Y`030000002CKjL0TfnW00@000000`2AKJL200000080T6fV00@00000
0>jdT@00000000001P3c]XX01@000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool0
0P3oool008T0oooo0P00002i0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000
202FLZT010000000UW:Y00000000000909EbZ0080000002ELZP000000000002ELZP000000>NaVP00
000909AaZ0050000002DLJP000000000002DLJP00P00000209=`Z0070000003Y/YP0000009=`Z000
0000j[:G0000000909=_Y`050000002CKjL0TfnW0000003[/YH00P000003095]Y`8000000`2@KJH3
000000H0lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003oool0oooo0080oooo0029
0?ooo`030000003oool0oooo0;P0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000
000809IbZ@040000002FLZT00000000000T0UG:X00P0000009EbZ0000000000009EbZ0000000ik6J
000000T0U76X00D0000009AaZ0000000000009AaZ00200000080Tg2X00L000000>VbV0000000Tg2X
0000003Z/YL0000000T0TfnW00P0000009=_Y`2CKjL000000>^bUP000000kK>C00000080TFfW0P00
0003091]YP<000001P3c]XX01@000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool0
0P3oool008T0oooo00<000000?ooo`3oool0^03oool00`000000UW>Y09IcZ@0609IcZ@040000002F
LjT00000000000P0UW:Y00@0000009IbZ@00000000002@2ELZP010000000UG:X00000000000209Eb
Z08000002@2DLJP01@000000U76X000000000000U76X008000000P2CL:P01`000000jK:H0000002C
L:P000000>ZbU`0000002@2CKjL020000000TfnW09=_Y`000000jk:F0000003]/i<000000P2AKJL2
000000<0T6fV0`0000060?>fRP050000003b]X`000000?6eS@000000T03oool00`000000oooo0?oo
o`020?ooo`00R@3oool00`000000oooo0?ooo`2h0?ooo`030000002FLjT0UW>Y00H0UW>Y00@00000
09IcZ@0000000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP0000000000080
UG:X0P00000909AaZ0040000002DLJP0000000000080U76X0P000000202CL:P000000>VbV0000000
Tg2X0000003Z/YL000002@2CKjL01@000000TfnW09=_Y`000000jk:F00<000000P2AKJL2000000<0
T6fV0`0000060?>fRP050000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`020?oo
o`00R@3oool00`000000oooo0?ooo`2h0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@00
00000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP0000000000080UG:X0P00
000909AaZ0040000002DLJP0000000000080U76X0P000000102CL:P000000>VbV000000209=`Z080
00002P2CKjL00`000000TfnW09=_Y`0400000080TFfW0P000003091]YP<000001P3c]XX00`000000
l[J<0?:fS00200000900oooo00<000000?ooo`3oool00P3oool008P0oooo00<000000?ooo`3oool0
^@3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@00
000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000U76X00000000
000209AaZ080000000@0Tg2X0000003Y/YP000000P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL0
10000002095]Y`8000000`2@KJH3000000H0lkJ:00<000000?:fS03b]X`00P00002@0?ooo`030000
003oool0oooo0080oooo00280?ooo`030000003oool0oooo0;T0oooo00<0000009IcZ@2FLjT01P2F
LjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00@0000009EbZ000
000000000P2ELZP2000000T0U76X00@0000009AaZ000000000000P2DLJP20000000409=`Z0000000
jK:H00000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV0P00
00000`3c]XX000000?>fRP040?>fRP030000003b]X`0l[J<00800000T03oool00`000000oooo0?oo
o`020?ooo`00R03oool00`000000oooo0?ooo`2i0?ooo`030000002FLjT0UW>Y00H0UW>Y00@00000
09IcZ@0000000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP0000000000080
UG:X0P00000909AaZ0040000002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P20000
00X0TfnW00<0000009=_Y`2CKjL010000002095]Y`8000000`2@KJH2000000030?>fRP000000lkJ:
00@0lkJ:00<000000?:fS03b]X`00P00002@0?ooo`030000003oool0oooo0080oooo00280?ooo`03
0000003oool0oooo0;T0oooo00<0000009IcZ@2FLjT01`2FLjT3000000P0UW:Y00@0000009IbZ@00
000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000U76X00000000
000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@000000P2A
KJL2000000<0T6fV0P0000000`3c]XX000000?>fRP040?>fRP060000003b]X`0l[J<0000003a]Hd0
0000S`3oool00`000000oooo0?ooo`020?ooo`00IP3oool02@2oM000000000000000000004RL0=oo
o`3oool0oooO09a800030000000307Boo`3oool0oooo00<0oooo00D0omnL04P000000000001d0;oo
o`0;0?ooo`030000003oool0oooo0;T0oooo00<0000009IcZ@2FLjT01`2FLjT3000000P0UW:Y00@0
000009IbZ@00000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000
U76X00000000000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW
00@000000P2AKJL2000000<0T6fV0P0000000`3c]XX000000?>fRP040?>fRP060000003b]X`0l[J<
0000003a]Hd00000S`3oool00`000000oooo0?ooo`020?ooo`00I`3oool00`3ogi`0B00007Boo`03
0?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool00P3oool01`3ookl0M00007Boo`3o
ool0okmd0000M02oool02P3oool00`000000oooo0?ooo`2j0?ooo`030000002FLjT0UW>Y00H0UW>Y
0`00000809IbZ@040000002FLZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP20000
00T0U76X00@0000009AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL0
0`000000TfnW09=_Y`0400000080TFfW0P000003091]YP80000000<0lkJ:0000003c]XX0103c]XX0
1P000000l[J<0?:fS0000000lKF=000008l0oooo00<000000?ooo`3oool00P3oool006L0oooo00<0
omnL04P0001d_ol0203oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0
_g@0001d_`3oool02@3oool00`000000oooo0?ooo`2j0?ooo`030000002FLjT0UW>Y00H0UW>Y00<0
000009IcZ@000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP0000000000080
UG:X0P00000909AaZ0040000002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P20000
00X0TfnW00<0000009=_Y`2CKjL010000002095]Y`8000000`2@KJH2000000030?>fRP000000lkJ:
00@0lkJ:00H000000?:fS03b]X`000000?6eS@00002?0?ooo`030000003oool0oooo0080oooo001W
0?ooo`030?oOW0180000M;oo00T0oooo00H0_g@007Boo`3oool0oono07@00018W=l30?ooo`030=nL
B00007@0_ooo00T0oooo00<000000?ooo`3oool0^P3oool00`000000UW>Y09IcZ@0609IcZ@030000
002FLjT0000000P0UW:Y00@0000009IbZ@00000000002@2ELZP010000000UG:X00000000000209Eb
Z08000002@2DLJP010000000U76X00000000000209AaZ08000000P2CL:P200000080Tg2X0P00000:
09=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV100000050?>fRP060000003b]X`0
l[J<0000003a]Hd00000S`3oool00`000000oooo0?ooo`020?ooo`00I`3oool00`3ogi`0B00007Bo
o`080?ooo`070?oog`2LB000B9cO0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0
oooo00<000000?ooo`3oool0^P3oool00`000000UW>Y09IcZ@0609IcZ@030000002FLjT0000000P0
UW:Y00@0000009IbZ@00000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP0
10000000U76X00000000000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`030000002C
KjL0TfnW00@000000P2AKJL2000000<0T6fV100000050?>fRP060000003b]X`0l[J<0000003a]Hd0
0000T03oool00`000000oooo0?ooo`010?ooo`00I`3oool00`3ogi`0B00007Boo`080?ooo`070=nL
B0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<000000?oo
o`3oool0^P3oool00`000000UW>Y09IcZ@0609IcZ@030000002FLjT0000000P0UW:Y00@0000009Ib
Z@00000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000U76X0000
0000000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@00000
0P2AKJL2000000<0T6fV100000050?>fRP060000003b]X`0l[J<0000003a]Hd00000T03oool00`00
0000oooo0?ooo`010?ooo`00I`3oool00`3ogi`0B00007Boo`040?ooo`0;0=oOg`3oool0oono07@0
00000000001d0;ooo`3oool0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030000
003oool0oooo0;X0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000UW>Y0000000809IbZ@040000
002FLZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP2000000T0U76X00@0000009Aa
Z000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`04
00000080TFfW0P000003091]YP@000001@3c]XX01P000000l[J<0?:fS0000000lKF=00000900oooo
00<000000?ooo`3oool00@3oool006L0oooo00<0omnL04P0001d_ol0103oool01P3OW4P000000000
00000000001809cOo`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`03
0000003oool0oooo0;X0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000UW>Y0000000809IbZ@04
0000002FLZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP2000000T0U76X00@00000
09AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_
Y`0400000080TFfW0P000003091]YP@000001P3c]XX01@000000l[J<0000003a]Hd000000900oooo
00<000000?ooo`3oool00@3oool006L0oooo00<0omnL04P0001d_ol0103oool0103ookl0M0000000
0018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000M02oool02@3oool00`000000oooo
0?ooo`2j0?ooo`030000002FLjT0UW>Y00H0UW>Y00<0000009IcZ@000000202FLZT010000000UW:Y
00000000000909EbZ0040000002ELZP0000000000080UG:X0P00000909AaZ0040000002DLJP00000
00000080U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL010000002
095]Y`8000000`2@KJH4000000H0lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003o
ool0oooo0040oooo001W0?ooo`030?oOW0180000M;oo00D0oooo00<0_g@007Boo`3oool01@3oool0
0`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo1000002i0?ooo`030000002FLjT0
UW>Y00H0UW>Y00<0000009IcZ@000000202FLZT010000000UW:Y00000000000909EbZ0040000002E
LZP0000000000080UG:X0P00000909AaZ0040000002DLJP0000000000080U76X0P00000209=`Z080
00000P2CL:P2000000X0TfnW00<0000009=_Y`2CKjL010000002095]Y`8000000`2@KJH4000000H0
lkJ:00D000000?:fS0000000lKF=0000002@0?ooo`030000003oool0oooo0040oooo001U0?ooo`05
0?oog`2LB00000000000001d_ol01@3oool01P3ogi`0B000000000000000001809cOo`80oooo00L0
oono07@0001d_ol0oooo0?noM00007@0_ooo00X0oooo00<000000?ooo`3oool0^P3oool00`000000
UW>Y09IcZ@0609IcZ@030000002FLjT0000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP01000
0000UG:X00000000000209EbZ08000002@2DLJP010000000U76X00000000000209AaZ08000000P2C
L:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00@000000P2AKJL2000000<0T6fV1000
00060?>fRP050000003b]X`000000?6eS@000000T03oool00`000000oooo0?ooo`010?ooo`00I`3o
ool00`3o_g@0000007Boo`060?ooo`050;md00000000000000000018W=l00`3oool01@3ookl0M000
0000000007@0_ooo00/0oooo00<000000?ooo`3oool0^P3oool00`000000UW>Y09IcZ@0609IcZ@03
0000002FLjT0000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000UG:X000000000002
09EbZ08000002@2DLJP010000000U76X00000000000209AaZ08000000P2CL:P200000080Tg2X0P00
000:09=_Y`030000002CKjL0TfnW00@000000`2AKJL00`000000T6fV091]YP02091]YP<000001P3c
]XX01@000000l[J<0000003a]Hd000000900oooo00<000000?ooo`3oool00@3oool008P0oooo00<0
00000?ooo`3oool0^P3oool00`000000UW>Y09IcZ@0609IcZ@030000002FLjT0000000P0UW:Y00<0
000009IbZ@2FLZT00P00000809EbZ0040000002ELZP0000000000080UG:X0P00000909AaZ0040000
002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00<0000009=_Y`2C
KjL010000003095]Y`030000002@KJH0T6fV0080T6fV0`0000060?>fRP050000003b]X`000000?6e
S@000000T03oool00`000000oooo0?ooo`010?ooo`00R03oool00`000000oooo0?ooo`2j0?ooo`03
0000002FLjT0UW>Y00H0UW>Y00<0000009IcZ@000000202FLZT00`000000UW:Y09IbZ@02000000P0
UG:X00@0000009EbZ000000000000P2ELZP2000000T0U76X00@0000009AaZ000000000000P2DLJP2
00000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`04000000<0TFfW00<00000
091]YP2@KJH00P2@KJH3000000H0lkJ:00<000000?:fS03b]X`00P00002@0?ooo`030000003oool0
oooo0040oooo00280?ooo`030000003oool0oooo0;X0oooo00<0000009IcZ@2FLjT01P2FLjT00`00
0000UW>Y0000000909IbZ@040000002FLZT00000000000P0UG:X00@0000009EbZ000000000000P2E
LZP2000000T0U76X00@0000009AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z0800000
2P2CKjL00`000000TfnW09=_Y`04000000<0TFfW00<00000091]YP2@KJH00P2@KJH3000000H0lkJ:
00<000000?:fS03b]X`00P00002@0?ooo`030000003oool0oooo0040oooo00270?ooo`030000003o
ool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000UW>Y0000000909IbZ@040000002F
LZT00000000000P0UG:X00@0000009EbZ02ELZP000000P2ELZP2000000T0U76X00@0000009AaZ000
000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`020000
0003099_Y`000000TFfW0080TFfW00<00000091]YP2@KJH00P2@KJH3000000H0lkJ:00<000000?:f
S03b]X`00P00002@0?ooo`030000003oool0oooo0040oooo00270?ooo`030000003oool0oooo0;/0
oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000
000000P0UG:X00@0000009EbZ02ELZP000000P2ELZP2000000T0U76X00@0000009AaZ02DLJP00000
0P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`0200000003099_
Y`000000TFfW0080TFfW00<00000091]YP2@KJH00P2@KJH3000000H0lkJ:00<000000?:fS03b]X`0
0P00002@0?ooo`030000003oool0oooo0040oooo00270?ooo`030000003oool0oooo0;/0oooo00<0
000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000P0
UG:X00@0000009EbZ02ELZP000000P2ELZP2000000T0U76X00@0000009AaZ02DLJP000000P2DLJP2
00000080Tg2X0P00000209=`Z08000002P2CKjL01`000000TfnW09=_Y`2CKjL00000099_Y`000000
0`2AKJL00`000000T6fV091]YP02091]YP<000001P3c]XX00`000000l[J<0?:fS00200000900oooo
00<000000?ooo`3oool00@3oool008L0oooo00<000000?ooo`3oool0^`3oool00`000000UW>Y09Ic
Z@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@0000000000202ELZP010000000
UG:X09EbZ000000209EbZ08000002@2DLJP010000000U76X09AaZ000000309AaZ0030000002CL:P0
Tg2X008000000P2CL:P2000000X0TfnW00L0000009=_Y`2CKjL0TfnW0000002BKjL0000000<0TFfW
0P000003091]YP<000001P3c]XX01P000000l[J<0?:fS0000000lKF=000008l0oooo00<000000?oo
o`3oool00@3oool008L0oooo00<000000?ooo`3oool0^`3oool00`000000UW>Y09IcZ@0609IcZ@04
0000002FLjT00000000000P0UW:Y00@0000009IbZ@0000000000202ELZP010000000UG:X09EbZ000
000209EbZ08000002@2DLJP010000000U76X09AaZ000000309AaZ0030000002CL:P0Tg2X00800000
0`2CL:P00`000000TfnW09=_Y`0809=_Y`070000002CKjL0TfnW09=_Y`000000TVnW00000003095]
Y`8000000`2@KJH3000000H0lkJ:00H000000?:fS03b]X`000000?6eS@00002?0?ooo`030000003o
ool0oooo0040oooo00270?ooo`030000003oool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2FLjT0
10000000UW>Y00000000000809IbZ@040000002FLZT00000000000P0UG:X00@0000009EbZ02ELZP0
00000`2ELZP00`000000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X00<0000009=`
Z02CL:P00P00000309=`Z0030000002CKjL0TfnW00P0TfnW00L0000009=_Y`2CKjL0TfnW0000002B
KjL0000000<0TFfW0P000003091]YP<000001P3c]XX01P000000l[J<0?:fS0000000lKF=000008l0
oooo00<000000?ooo`3oool00@3oool008L0oooo00<000000?ooo`3oool0^`3oool00`000000UW>Y
09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@0000000000202ELZP01000
0000UG:X09EbZ000000309EbZ0030000002DLJP0U76X00L0U76X00@0000009AaZ02DLJP000000`2D
LJP01@000000Tg2X09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL01`000000TfnW
09=_Y`2CKjL00000099_Y`0000000`2AKJL2000000<0T6fV0`0000060?>fRP060000003b]X`0l[J<
0000003a]Hd00000S`3oool00`000000oooo0?ooo`010?ooo`00Q`3oool00`000000oooo0?ooo`2k
0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT010000000UW:Y0000
0000000809EbZ0030000002ELZP0UG:X008000000P2ELZP00`000000U76X09AaZ00709AaZ0040000
002DLJP0U76X000000<0U76X00D0000009=`Z02CL:P0Tg2X0000000309=`Z0030000002CKjL0TfnW
00P0TfnW00H0000009=_Y`2CKjL0TfnW0000002BKjL200000080TFfW0P000003091]YP<000001P3c
]XX01P000000l[J<0?:fS0000000lKF=000008l0oooo00<000000?ooo`3oool00@3oool008L0oooo
00<000000?ooo`3oool0^`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0
UW:Y00@0000009IbZ@0000000000202ELZP00`000000UG:X09EbZ00200000080UG:X00<0000009Aa
Z02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X09=`Z00000000`2C
L:P2000000T0TfnW00H0000009=_Y`2CKjL0TfnW0000002BKjL200000080TFfW0P000003091]YP<0
00001P3c]XX01P000000l[J<0?:fS0000000lKF=000008l0oooo00<000000?ooo`3oool00@3oool0
08L0oooo00<000000?ooo`3oool0^`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000
000000P0UW:Y00@0000009IbZ@00000000002@2ELZP010000000UG:X00000000000209EbZ0030000
002DLJP0U76X00L0U76X00<0000009AaZ02DLJP00P00000209AaZ0050000002CL:P0Tg2X09=`Z000
00000`2CL:P2000000T0TfnW00@0000009=_Y`2CKjL0TfnW10000002095]Y`8000000`2@KJH30000
00H0lkJ:00H000000?:fS03b]X`000000?6eS@00002?0?ooo`030000003oool0oooo0040oooo0027
0?ooo`030000003oool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000
000809IbZ@040000002FLZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP00`000000
U76X09AaZ00709AaZ0030000002DLJP0U76X008000000P2DLJP01@000000Tg2X09=`Z02CL:P00000
00<0Tg2X0P00000909=_Y`040000002CKjL0TfnW09=_Y`@000000P2AKJL2000000<0T6fV0`000007
0?>fRP050000003b]X`000000?6eS@000000T03oool100000040oooo0@3oool008L0oooo00<00000
0?ooo`3oool0^`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@0
000009IbZ@00000000002@2ELZP010000000UG:X00000000000209EbZ0030000002DLJP0U76X00L0
U76X00<0000009AaZ02DLJP00P00000209AaZ0040000002CL:P0Tg2X09=`Z08000000P2CL:P20000
00T0TfnW00@0000009=_Y`2CKjL0TfnW10000002095]Y`8000000`2@KJH3000000L0lkJ:00D00000
0?:fS0000000lKF=0000002@0?ooo`4000000@3oool10?ooo`00Q`3oool300000;/0oooo00<00000
09IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X
00@0000009EbZ000000000000P2ELZP00`000000U76X09AaZ00809AaZ0040000002DLJP000000000
0080U76X00@0000009=`Z02CL:P0Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`04
00000080TFfW0P000003091]YP<000001`3c]XX01@000000l[J<0000003a]Hd000000900oooo0@00
00010?ooo`40oooo00270?ooo`030000003oool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2FLjT0
10000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00@0000009EbZ0000000
00000P2ELZP00`000000U76X09AaZ00809AaZ0040000002DLJP0000000000080U76X00@0000009=`
Z02CL:P0Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`0400000080TFfW0P000003
091]YP<000001`3c]XX01@000000l[J<0000003a]Hd000000900oooo0@0000010?ooo`40oooo0027
0?ooo`030000003oool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000
000809IbZ@040000002FLZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP2000000T0
U76X00@0000009AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`00
0000TfnW09=_Y`0400000080TFfW0P000003091]YP<000001`3c]XX01@000000l[J<0000003a]Hd0
00000900oooo0@0000010?ooo`40oooo00270?ooo`030000003oool0oooo0;/0oooo00<0000009Ic
Z@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00@0
000009EbZ000000000000P2ELZP2000000T0U76X00@0000009AaZ000000000000P2DLJP200000080
Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW09=_Y`0400000080TFfW0P000003091]YP<0
00001`3c]XX01@000000l[J<0000003a]Hd000000900oooo0@0000010?ooo`40oooo00270?ooo`03
0000003oool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809Ib
Z@040000002FLZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP2000000T0U76X00@0
000009AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000TfnW
09=_Y`0400000080TFfW0P000004091]YP8000001`3c]XX00`000000l[J<0?:fS00200000900oooo
0@0000010?ooo`40oooo00270?ooo`030000003oool0oooo0;/0oooo00<0000009IcZ@2FLjT01P2F
LjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00@0000009EbZ000
000000000P2ELZP2000000T0U76X00@0000009AaZ000000000000P2DLJP200000080Tg2X0P000002
09=`Z08000002P2CKjL00`000000TfnW09=_Y`04000000<0TFfW00<00000091]YP0000001@000006
0?>fRP030000003b]X`0l[J<00800000T03oool100000040oooo0@3oool008L0oooo00<000000?oo
o`3oool0^`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@00000
09IbZ@00000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000U76X
00000000000209AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW00T0
00000`2@KJH00`000000T6bV091/YP0400000080lkJ:00<000000?:fS03b]X`00P00002@0?ooo`40
00000@3oool10?ooo`00Q`3oool00`000000oooo0?ooo`2k0?ooo`030000002FLjT0UW>Y00H0UW>Y
00@0000009IcZ@0000000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP00000
00000080UG:X0P00000909AaZ0040000002DLJP0000000000080U76X0P00000209=`Z08000000P2C
L:P2000000@0TfnW3@000003095]Y`030000002@KJH0T6fV0080T6fV00<00000091/YP2@K:H00P00
000208]WY0@0000000<0l[J<000000000000T03oool100000040oooo0@3oool008H0oooo00<00000
0?ooo`3oool0_03oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@0
000009IbZ@00000000002@2ELZP010000000UG:X00000000000209EbZ08000002@2DLJP010000000
U76X00000000000209AaZ08000000P2CL:P:000000H0TfnW00<0000009=_Y`2CKjL00P0000000`2B
KjL00000095]Y`02095]Y`030000002@KJH0T6fV0080T6fV00<00000091/YP2@K:H00P00000608]W
Y0<00000T03oool100000040oooo0@3oool008H0oooo00<000000?ooo`3oool0_03oool00`000000
UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0UW:Y00@0000009IbZ@00000000002@2ELZP0
10000000UG:X00000000000209EbZ08000002@2DLJP<00000080Tg2X0P00000:09=_Y`030000002C
KjL0TfnW0080000000<0TVnW0000002AKJL00P2AKJL00`000000T6fV091]YP02091]YP030000002@
K:H0T6bV00800000202;Ij@5000008`0oooo0@0000010?ooo`40oooo00260?ooo`030000003oool0
oooo0;`0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002F
LZT00000000000T0UG:X00@0000009EbZ000000000000P2ELZP20000000309AaZ0000000000000L0
000000<0U76X0000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2CKjL00`000000
TfnW09=_Y`0200000003099_Y`000000TFfW0080TFfW00<00000091]YP2@KJH00P2@KJH00`000000
T6bV091/YP02000000H0RfNT2P0000290?ooo`4000000@3oool10?ooo`00QP3oool00`000000oooo
0?ooo`2l0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT010000000
UW:Y00000000000909EbZ0T00000202DLJP010000000U76X00000000000209AaZ08000000P2CL:P2
00000080Tg2X0P00000:09=_Y`030000002CKjL0TfnW0080000000<0TVnW0000002AKJL00P2AKJL0
0`000000T6fV091]YP02091]YP030000002@K:H0T6bV00P00000201Cao<6000008D0oooo0@000001
0?ooo`40oooo00260?ooo`030000003oool0oooo0;`0oooo00<0000009IcZ@2FLjT01P2FLjT01000
0000UW>Y09IcZ@00000809IbZ@030000002FLZT0000000/0000000<0UG:X0000000000000P2ELZP2
000000T0U76X00@0000009AaZ000000000000P2DLJP200000080Tg2X0P00000209=`Z08000002P2C
KjL01`000000TfnW09=_Y`2CKjL00000099_Y`0000000`2AKJL00`000000T6fV091]YP02091]YP03
0000002@K:H0T6bV0080000000<0SV^U0000000000000P00000905?7l`X00000P@3oool100000040
oooo0@3oool008H0oooo00<000000?ooo`3oool0_03oool00`000000UW>Y09IcZ@0609IcZ@040000
002FLjT0UW>Y00000080UW:Y2P00000909EbZ0040000002ELZP0000000000080UG:X0P00000909Aa
Z0040000002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00L00000
09=_Y`2CKjL0TfnW0000002BKjL0000000<0TFfW00<00000091]YP2@KJH00P2@KJH00`000000T6bV
091/YP020000000308i[Y@000000S6RU0080S6RU2@00000805?7l`D00000OP3oool100000040oooo
0@3oool008H0oooo00<000000?ooo`3oool0_03oool00`000000UW>Y09IcZ@0309IcZ@T000001P2F
LZT010000000UW:Y00000000000909EbZ0040000002ELZP0000000000080UG:X0P00000909AaZ004
0000002DLJP0000000000080U76X0P00000209=`Z08000000P2CL:P2000000X0TfnW00L0000009=_
Y`2CKjL0TfnW0000002BKjL0000000<0TFfW00<00000091]YP2@KJH00P2@KJH00`000000T6bV091/
YP020000000308i[Y@000000S6RU00H0S6RU1000000905?7l`T00000NP3oool100000040oooo0@3o
ool008H0oooo00<000000?ooo`3oool0^@3oool9000000<0UW>Y00@0000009IcZ@2FLjT00000202F
LZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0000000000080UG:X0P00000909AaZ004
0000002DLJP0000000000080U76X0P00000209=`Z08000000`2CL:P00`000000TfnW09=_Y`0809=_
Y`070000002CKjL0TfnW09=_Y`000000TVnW00000003095]Y`030000002@KJH0T6fV0080T6fV00<0
0000091/YP2@K:H00P0000000`2>JjD0000008aXY@0908aXY@X00000201Cao<4000007L0oooo0@00
00010?ooo`40oooo00260?ooo`030000003oool0oooo0;00oooo2@0000030?ooo`030000002FLjT0
UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000
002ELZP0000000000080UG:X0P00000909AaZ0040000002DLJP0U76X000000<0U76X00<0000009=`
Z02CL:P00P00000309=`Z0030000002CKjL0TfnW00P0TfnW00L0000009=_Y`2CKjL0TfnW0000002B
KjL0000000<0TFfW00<00000091]YP2@KJH00P2@KJH01`000000T6bV091/YP2@K:H0000008i[Y@00
00000`2<J:D7000000D0DlOc1000000905?7l`P00000L`3oool100000040oooo0@3oool008H0oooo
00<000000?ooo`3oool0Y`3oool9000000`0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y
09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0000009EbZ02ELZP000000P2ELZP2
000000T0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000Tg2X09=`Z02CL:P0000000<0Tg2X
00<0000009=_Y`2CKjL0202CKjL01`000000TfnW09=_Y`2CKjL00000099_Y`0000000`2AKJL00`00
0000T6fV091]YP02091]YP060000002@K:H0T6bV091/YP000000SV^U1000000>05?7l`/00000201C
ao<4000006l0oooo0@0000010?ooo`40oooo00260?ooo`<00000W@3oool:000001D0oooo00<00000
09IcZ@2FLjT01P2FLjT010000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X
00@0000009EbZ02ELZP000000P2ELZP2000000T0U76X00@0000009AaZ02DLJP000000`2DLJP01@00
0000Tg2X09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL01`000000TfnW09=_Y`2C
KjL00000099_Y`0000000`2AKJL00`000000T6fV091]YP02091]YP050000002@K:H0T6bV091/YP00
00000`2>JjD4000000H0DlOc1P00000505?7l`@000002@1Cao<8000006`0oooo0@0000010?ooo`40
oooo00260?ooo`030000003oool0oooo09@0oooo2@00000O0?ooo`030000002FLjT0UW>Y00H0UW>Y
00@0000009IcZ@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X
00000080UG:X0P00000909AaZ0040000002DLJP0U76X000000<0U76X00D0000009=`Z02CL:P0Tg2X
0000000309=`Z0030000002CKjL0TfnW00P0TfnW00L0000009=_Y`2CKjL0TfnW0000002BKjL00000
00<0TFfW00<00000091]YP2@KJH00P2@KJH200000080T6bV00<0000008i[Y@2>JjD00P2>JjD00`00
0000SFZU08eZY@06000000h0DlOc2P00000805?7l`@00000J@3oool100000040oooo00260?ooo`03
0000003oool0oooo08/0oooo2@00000X0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2F
LjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0
000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X09=`Z000
00000`2CL:P00`000000TfnW09=_Y`0809=_Y`070000002CKjL0TfnW09=_Y`000000TVnW00000003
095]Y`030000002@KJH0T6fV0080T6fV0P000002091/YP030000002>JjD0SV^U0080SV^U00<00000
08eZY@2=JZD0102=JZD4000000H0DlOc1P00000505?7l`@000002@1Cao<8000006D0oooo0@000001
0?ooo`00QP3oool00`000000oooo0?ooo`220?ooo`T00000<@3oool00`000000UW>Y09IcZ@0609Ic
Z@040000002FLjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000UG:X09Eb
Z000000309EbZ0030000002DLJP0U76X00L0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000
Tg2X09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL01`000000TfnW09=_Y`2CKjL0
0000099_Y`0000000`2AKJL2000000<0T6fV0P000002091/YP030000002>JjD0SV^U0080SV^U00<0
000008eZY@2=JZD0102=JZD:000000d0DlOc2`00000805?7l`@00000H@3oool100000040oooo0026
0?ooo`030000003oool0oooo07T0oooo2@00000j0?ooo`030000002FLjT0UW>Y00H0UW>Y00@00000
09IcZ@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0
UG:X00<0000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X
09=`Z00000000`2CL:P00`000000TfnW09=_Y`0809=_Y`070000002CKjL0TfnW09=_Y`000000TVnW
00000003095]Y`8000000`2@KJH200000080T6bV00<0000008i[Y@2>JjD00P2>JjD7000000P0DlOc
1000000505?7l`H000001@1Cao<4000000T0DlOc2000001N0?ooo`4000000@3oool008H0oooo00<0
00000?ooo`3oool0K`3oool:000004<0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y09Ic
Z@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0000009EbZ02ELZP000000`2ELZP00`00
0000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X00D0000009=`Z02CL:P0Tg2X0000
000309=`Z0030000002CKjL0TfnW00P0TfnW00L0000009=_Y`2CKjL0TfnW0000002BKjL0000000<0
TFfW0P000003091]YP8000000P2@K:H00`000000SV^U08i[Y@0408i[Y@@00000201Cao<:000000h0
DlOc2P00000805?7l`@00000FP3oool100000040oooo00260?ooo`030000003oool0oooo06H0oooo
2@00001=0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000
UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2DLJP0
10000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X09=`Z00000000`2CL:P00`000000TfnW
09=_Y`0809=_Y`070000002CKjL0TfnW09=_Y`000000TVnW00000003095]Y`800000102@KJH01000
0000T6bV091/YP00000608i[Y@`00000201Cao<4000000H0DlOc1P00000505?7l`@00000201Cao<8
000005L0oooo0@0000010?ooo`00QP3oool00`000000oooo0?ooo`1M0?ooo`T00000EP3oool00`00
0000UW>Y09IcZ@0609IcZ@040000002FLjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT000002@2E
LZP010000000UG:X09EbZ000000309EbZ0030000002DLJP0U76X00L0U76X00@0000009AaZ02DLJP0
00000`2DLJP01@000000Tg2X09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL01P00
0000TfnW09=_Y`2CKjL00000099_Y`8000000P2AKJL2000000@0T6fV00<00000091/YP2@K:H01`00
000705?7l`@000002@1Cao<:000000X0DlOc3@00000805?7l`@00000D`3oool100000040oooo0026
0?ooo`030000003oool0oooo05@0oooo2@00001O0?ooo`030000002FLjT0UW>Y00H0UW>Y00@00000
09IcZ@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0
UG:X00<0000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X
09=`Z00000000`2CL:P00`000000TfnW09=_Y`0809=_Y`060000002CKjL0TfnW09=_Y`000000TVnW
0P000003095]Y`030000002@KJH0T6fV0080T6fV00<00000091/YP2@K:H00P2@K:H4000000P0DlOc
3@00000:05?7l`X00000201Cao<4000000P0DlOc2000001@0?ooo`4000000@3oool008D0oooo00<0
00000?ooo`3oool0C03oool9000006P0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y09Ic
Z@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0000009EbZ02ELZP000000`2ELZP00`00
0000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X00D0000009=`Z02CL:P0Tg2X0000
000309=`Z0030000002CKjL0TfnW00P0TfnW00H0000009=_Y`2CKjL0TfnW0000002BKjL2000000<0
TFfW00<00000091]YP2@KJH00P2@KJH00`000000T6bV091/YP02091/YP040000002?K:H0SfbV08m/
YPP00000201Cao<4000000D0DlOc1P00000505?7l`@00000201Cao<=000000P0DlOc1000001<0?oo
o`4000000@3oool008D0oooo00<000000?ooo`3oool0@P3oool:00000740oooo00<0000009IcZ@2F
LjT01P2FLjT010000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@00000
09EbZ02ELZP000000`2ELZP00`000000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X
00D0000009=`Z02CL:P0Tg2X0000000309=`Z0030000002CKjL0TfnW00P0TfnW00H0000009=_Y`2C
KjL0TfnW0000002BKjL2000000<0TFfW00<00000091]YP2@KJH00P2@KJH00`000000T6bV091/YP02
091/YP030000002?K:H0SfbV00@0SfbV1000000805?7l`X000003@1Cao<:000000P0DlOc10000008
05?7l`P00000B@3oool100000040oooo00250?ooo`030000003oool0oooo03T0oooo2@00001k0?oo
o`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000UW:Y09IbZ@00
000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2DLJP010000000U76X
09AaZ000000309AaZ0050000002CL:P0Tg2X09=`Z00000000`2CL:P00`000000TfnW09=_Y`0809=_
Y`040000002CKjL0TfnW09=_Y`80000000<0TVnW0000002AKJL00P2AKJL00`000000T6fV091]YP02
091]YP030000002@K:H0T6bV0080T6bV00<0000008m/YP2?K:H0102?K:H<000000P0DlOc10000005
05?7l`H000001@1Cao<4000000P0DlOc3@00000805?7l`@00000A@3oool100000040oooo00250?oo
o`030000003oool0oooo0300oooo2@0000240?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009Ic
Z@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X
00<0000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X09=`
Z00000000`2CL:P00`000000TfnW09=_Y`0809=_Y`040000002CKjL0TfnW09=_Y`80000000<0TVnW
0000002AKJL00P2AKJL00`000000T6fV091]YP02091]YP030000002@K:H0T6bV0080T6bV1`000007
05?7l`@000002@1Cao<9000000X0DlOc3@00000805?7l`<00000201Cao<:00000440oooo0@000001
0?ooo`00Q@3oool00`000000oooo0?ooo`0W0?ooo`T00000S@3oool00`000000UW>Y09IcZ@0609Ic
Z@040000002FLjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000UG:X09Eb
Z000000309EbZ0030000002DLJP0U76X00L0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000
Tg2X09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL010000000TfnW09=_Y`2CKjL2
00000003099_Y`000000TFfW0080TFfW00<00000091]YP2@KJH00P2@KJH00`000000T6bV091/YP04
091/YP@00000201Cao<=000000L0DlOc3000000805?7l`@00000201Cao<<000000L0DlOc2000000l
0?ooo`4000000@3oool008D0oooo00<000000?ooo`3oool07P3oool9000009H0oooo00<0000009Ic
Z@2FLjT01P2FLjT010000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0
000009EbZ02ELZP000000`2ELZP00`000000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0
U76X00D0000009=`Z02CL:P0Tg2X0000000309=`Z08000002P2CKjL00`000000TfnW09=_Y`020000
0003099_Y`000000TFfW0080TFfW00<00000091]YP2@KJH00P2@KJH00`000000T6bV091/YP04091/
YP`000002P1Cao<:000000L0DlOc1000000805?7l`d000002@1Cao<:000000H0DlOc1@00000i0?oo
o`4000000@3oool008D0oooo00<000000?ooo`3oool0503oool:000009l0oooo00<0000009IcZ@2F
LjT01P2FLjT010000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@00000
09EbZ02ELZP000000`2ELZP00`000000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X
00D0000009=`Z02CL:P0Tg2X0000000309=`Z08000002P2CKjL00`000000TfnW09=_Y`0200000003
099_Y`000000TFfW0080TFfW00<00000091]YP2@KJH00P2@KJH7000000L0DlOc1000000505?7l`H0
00001@1Cao<4000000P0DlOc3000000805?7l`<000001@1Cao<6000000D0DlOc0`00000805?7l`P0
0000=P3oool100000040oooo00250?ooo`030000003oool0oooo00/0oooo2@00002Y0?ooo`030000
002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909Eb
Z0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000
000309AaZ0040000002CL:P0Tg2X09=`Z08000000P2CL:P2000000X0TfnW00L0000009=_Y`2CKjL0
TfnW0000002BKjL0000000<0TFfW00<00000091]YP2@KJH0102@KJH4000000P0DlOc2@00000:05?7
l`d000001`1Cao<4000000P0DlOc2@00000:05?7l``000001`1Cao<4000003<0oooo0@0000010?oo
o`00Q@3oool00`000000oooo0?ooo`020?ooo`T00000/P3oool00`000000UW>Y09IcZ@0609IcZ@04
0000002FLjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000UG:X09EbZ000
000309EbZ0030000002DLJP0U76X00L0U76X00@0000009AaZ02DLJP000000`2DLJP010000000Tg2X
09=`Z02CL:P200000080Tg2X0P00000:09=_Y`070000002CKjL0TfnW09=_Y`000000TVnW00000003
095]Y`030000002@KJH0T6fV00@0T6fV3000000705?7l``00000201Cao<4000000P0DlOc30000006
05?7l`d000001`1Cao<4000000P0DlOc2000000_0?ooo`4000000@3oool008D0oooo1@00002k0?oo
o`030000002FLjT0UW>Y00L0UW>Y00<0000009IcZ@000000202FLZT010000000UW:Y09IbZ@000009
09EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2DLJP010000000U76X09Aa
Z000000309AaZ08000000P2CL:P200000080Tg2X0P00000:09=_Y`070000002CKjL0TfnW09=_Y`00
0000TVnW00000003095]Y`L000001`1Cao<<000000L0DlOc1000000805?7l`d000001P1Cao<<0000
00P0DlOc0`00000805?7l`d000001P1Cao<7000002X0oooo0@0000010?ooo`00R03oool6000000<0
oooo00<000000?ooo`3oool0/@3oool00`000000UW>Y09IcZ@0709IcZ@030000002FLjT0000000P0
UW:Y00<0000009IbZ@2FLZT00P00000809EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009Aa
Z02DLJP01`2DLJP00`000000U76X09AaZ00200000080U76X0P00000209=`Z08000000P2CL:P20000
00X0TfnW00L0000009=_Y`2CKjL0TfnW0000002BKjL0000000L0TFfW2P00000705?7l`@00000201C
ao<<000000T0DlOc2P00000705?7l`@00000201Cao<<000000T0DlOc2P00000505?7l`D000009`3o
ool100000040oooo002>0?ooo`H00000203oool00`000000oooo0?ooo`2W0?ooo`030000002FLjT0
UW>Y00H0UW>Y00<0000009IcZ@000000202FLZT00`000000UW:Y09IbZ@02000000P0UG:X00@00000
09EbZ02ELZP000000`2ELZP00`000000U76X09AaZ00709AaZ0030000002DLJP0U76X008000000P2D
LJP200000080Tg2X0P00000309=`Z0030000002CKjL0TfnW00P0TfnW00L0000009=_Y`2CKjL0TfnW
0000002BKjL0000000<0TFfW1P00000405?7l`@000001`1Cao<<000000L0DlOc1000000505?7l`H0
0000101Cao<4000000P0DlOc3000000705?7l`<000001@1Cao<6000000D0DlOc0`00000805?7l`L0
0000903oool100000040oooo002D0?ooo`H000000P3oool00`000000oooo0?ooo`2W0?ooo`030000
002FLjT0UW>Y00H0UW>Y00<0000009IcZ@0000002@2FLZT010000000UW:Y00000000000809EbZ003
0000002ELZP0UG:X008000000P2ELZP00`000000U76X09AaZ00809AaZ0040000002DLJP000000000
0080U76X0P00000209=`Z08000000`2CL:P00`000000TfnW09=_Y`0809=_Y`060000002CKjL0TfnW
09=_Y`000000TVnW1000000:05?7l`/000001`1Cao<4000000P0DlOc2@00000:05?7l``000001`1C
ao<3000000P0DlOc2@00000:05?7l``000001`1Cao<300000280oooo0@00000009X0oooo1P000007
0?ooo`030000003oool0oooo09`0oooo00<0000009IcZ@2FLjT01P2FLjT00`000000UW>Y00000009
09IbZ@040000002FLZT00000000000P0UG:X00<0000009EbZ02ELZP00P00000209EbZ08000002@2D
LJP010000000U76X00000000000209AaZ08000000`2CL:P01@000000Tg2X09=`Z02CL:P0000000X0
TfnW00D0000009=_Y`2CKjL0TfnW00000003099_Y`030000002BKZL0000000T000001`1Cao<30000
00P0DlOc3000000705?7l``000001`1Cao<3000000P0DlOc3000000605?7l`d000001`1Cao<30000
00P0DlOc2000000N0?ooo`400000002P0?ooo`H0000000<0oooo0000003oool02@3oool00`000000
oooo0?ooo`2A0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@0000000000202FLZT01000
0000UW:Y00000000000909EbZ0040000002ELZP0000000000080UG:X0P00000909AaZ0040000002D
LJP00000000000<0U76X00D0000009=`Z02CL:P0Tg2X0000000309=`Z0030000002CKjL0TfnW00P0
TfnW00D0000009=_Y`2CKjL0TfnW00000003099_Y`030000002BKZL0TVjW00@0TVjW1000000705?7
l``000001`1Cao<<000000L0DlOc0`00000805?7l``000001P1Cao<<000000P0DlOc0`00000805?7
l``000001P1Cao<7000001T0oooo0@0000000:H0oooo1P0000060?ooo`030000003oool0oooo0940
oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000
000000T0UG:X00@0000009EbZ000000000000P2ELZP2000000T0U76X00@0000009AaZ00000000000
0`2DLJP01@000000Tg2X09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL01@000000
TfnW09=_Y`2CKjL0000000<0TVnW00<00000099^Y`2BKZL0102BKZL;000000L0DlOc3000000705?7
l`<00000201Cao<<000000H0DlOc3000000705?7l`@00000201Cao<<000000H0DlOc3000000505?7
l`D000005P3oool100000000[03oool5000000030?ooo`000000oooo0980oooo00<0000009IcZ@2F
LjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00@00000
09EbZ000000000000P2ELZP2000000T0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000Tg2X
09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL01@000000TfnW09=_Y`2CKjL00000
00<0TVnW1`00000705?7l`/000001`1Cao<4000000L0DlOc3000000705?7l`/000001`1Cao<40000
00P0DlOc3000000605?7l``000001`1Cao<2000000P0DlOc2@00000B0?ooo`400000002a0?ooo`H0
00001P3oool00`000000oooo0?ooo`260?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@00
00000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP0000000000080UG:X0P00
000909AaZ0040000002DLJP0U76X000000<0U76X00D0000009=`Z02CL:P0Tg2X0000000309=`Z003
0000002CKjL0TfnW00P0TfnW00D0000009=_Y`2CKjL0TfnW00000005099_Y``000001`1Cao<30000
00P0DlOc2`00000705?7l``000001`1Cao<3000000P0DlOc3000000605?7l``000001`1Cao<20000
00P0DlOc3000000605?7l`L000003P3oool100000000]`3oool7000008P0oooo00<0000009IcZ@2F
LjT01P2FLjT010000000UW>Y00000000000809IbZ@040000002FLZT00000000000T0UG:X00@00000
09EbZ000000000000P2ELZP2000000T0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000Tg2X
09=`Z02CL:P0000000<0Tg2X00<0000009=_Y`2CKjL0202CKjL010000000TfnW09=_Y`2CKjL60000
00L0DlOc1000000705?7l``000001P1Cao<<000000L0DlOc0`00000805?7l``000001P1Cao<<0000
00L0DlOc0P00000805?7l``000001P1Cao<<000000D0DlOc1@00000;0?ooo`400000002m0?ooo`H0
00001@3oool00`000000oooo0?ooo`1k0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@00
00000000202FLZT010000000UW:Y00000000000909EbZ0040000002ELZP00000000000<0UG:X00<0
000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0050000002CL:P0Tg2X09=`Z000
00000`2CL:P00`000000TfnW09=_Y`0809=_Y`030000002CKjL0TfnW00<0TfnW1000000705?7l`/0
00001`1Cao<;000000L0DlOc1000000705?7l``000001P1Cao<<000000L0DlOc0`00000705?7l``0
00001P1Cao<<000000L0DlOc0`00000805?7l`P000001`3oool100000000``3oool6000000T0oooo
00<000000?ooo`3oool0L@3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT00000000000P0
UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000UG:X09EbZ000000309EbZ0030000002DLJP0
U76X00L0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000Tg2X09=`Z02CL:P0000000<0Tg2X
00<0000009=_Y`2CKjL0202CKjL00`000000TfnW09=_Y`0309=_Y`/000001`1Cao<;000000L0DlOc
0`00000805?7l`/000001P1Cao<<000000L0DlOc0`00000805?7l`/000001P1Cao<<000000L0DlOc
0`00000805?7l``000001@1Cao<7000000<0oooo0@0000000<P0oooo1`0000030?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0I`3oool00`000000UW>Y09IcZ@0609IcZ@040000002F
LjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000UG:X09EbZ000000309Eb
Z0030000002DLJP0U76X00L0U76X00@0000009AaZ02DLJP000000`2DLJP01@000000Tg2X09=`Z02C
L:P0000000<0Tg2X00<0000009=_Y`2CKjL00`2CKjL;000000L0DlOc2`00000705?7l`<000001`1C
ao<<000000H0DlOc2`00000705?7l`@000001`1Cao<<000000H0DlOc2`00000705?7l`<00000201C
ao<<000000D0DlOc3000000505?7l`@000000@3oool100000000PP3oool01P2oM000000000000000
000004RL0=ooodL0oooo1@0000080?ooo`030000003oool0oooo06L0oooo00<0000009IcZ@2FLjT0
1P2FLjT010000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0000009Eb
Z02ELZP000000`2ELZP00`000000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X00D0
000009=`Z02CL:P0Tg2X0000000309=`Z0H000001`1Cao<;000000L0DlOc0`00000705?7l`/00000
1`1Cao<;000000L0DlOc0`00000805?7l`/000001P1Cao<<000000L0DlOc0P00000805?7l``00000
1@1Cao<<000000L0DlOc0P00000805?7l`H0000000230?ooo`030?oOW0180000M;oo04h0oooo1P00
00020?ooo`030000003oool0oooo06L0oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y09Ic
Z@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0000009EbZ02ELZP000000`2ELZP00`00
0000U76X09AaZ00709AaZ0040000002DLJP0U76X000000<0U76X00D0000009=`Z02CL:P0Tg2X0000
000509=`Z0/000001`1Cao<3000000L0DlOc2`00000705?7l`/000001`1Cao<3000000L0DlOc3000
000605?7l`/000001`1Cao<3000000L0DlOc3000000605?7l`/000001`1Cao<3000000L0DlOc3000
00060?ooo`00P`3oool00`3ogi`0B00007Boo`1D0?ooo`H000001P3oool00`000000oooo0?ooo`1M
0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000UW:Y09Ib
Z@00000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2DLJP010000000
U76X09AaZ000000309AaZ0040000002CL:P0Tg2X09=`Z0H000001`1Cao<3000000L0DlOc2`000007
05?7l`/000001`1Cao<3000000L0DlOc2`00000605?7l``000001`1Cao<2000000P0DlOc2`000006
05?7l``000001`1Cao<2000000P0DlOc2`00000B0?ooo`00P`3oool00`3ogi`0B00007Boo`1J0?oo
o`L00000203oool00`000000oooo0?ooo`1D0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009Ic
Z@2FLjT00000202FLZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X
00<0000009AaZ02DLJP01`2DLJP010000000U76X09AaZ000000309AaZ0030000002CL:P0Tg2X00<0
Tg2X5@00000705?7l`/000001P1Cao<F000000H0DlOc2`00000605?7laL000001P1Cao<;000000H0
DlOc5`00000M0?ooo`00P`3oool00`3ogi`0B00007Boo`1P0?ooo`H000000`3oool00`000000oooo
0?ooo`1D0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000
UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2DLJP0
10000000U76X09AaZ000000309AaZ0H000001`1Cao<3000000L0DlOc2`00000605?7l`/000001`1C
ao<3000000L0DlOc2`00000605?7l`/000001`1Cao<3000000L0DlOc3000000505?7la@00000;03o
ool008<0oooo00<0omnL04P0001d_ol0IP3oool6000000L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`110?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202F
LZT010000000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02D
LJP01`2DLJP010000000U76X09AaZ000000509AaZ0@000001`1Cao<:000000L0DlOc2P00000705?7
l`<000001`1Cao<;000000H0DlOc2`00000705?7l`<000001`1Cao<;000000H0DlOc4P00000i0?oo
o`00P`3oool00`3ogi`0B00007Boo`1/0?ooo`H0000000<0oooo0000003oool01`3oool00`000000
oooo0?ooo`110?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT01000
0000UW:Y09IbZ@00000909EbZ0040000002ELZP0UG:X000000<0UG:X00<0000009AaZ02DLJP01`2D
LJP010000000U76X09AaZ000000509AaZ0/000001P1Cao<;000000H0DlOc1000000705?7l`X00000
1`1Cao<:000000L0DlOc0`00000705?7l`/000001P1Cao<@000004H0oooo00230?ooo`030?oOW018
0000M;oo0780oooo1@0000050?ooo`030000003oool0oooo0440oooo00<0000009IcZ@2FLjT01P2F
LjT010000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00@0000009EbZ02E
LZP000000`2ELZP00`000000U76X09AaZ00709AaZ0030000002DLJP0U76X00H000001`1Cao<:0000
00L0DlOc5@00000605?7l`/000001P1Cao<E000000H0DlOc3`00001B0?ooo`00P`3oool00`3ogi`0
B00007Boo`0R0?ooo`050?oOW018000000000000M02oool00`3oool00`3o_g@0000007Boo`030?oo
o`050=nLB0000000000000000018W=l0@P3oool6000000P0oooo00<000000?ooo`3oool0>03oool0
0`000000UW>Y09IcZ@0609IcZ@040000002FLjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT00000
2@2ELZP010000000UG:X09EbZ000000309EbZ0030000002DLJP0U76X00L0U76X00<0000009AaZ02D
LJP00`2DLJP:000000L0DlOc2P00000605?7l`@000001`1Cao<:000000H0DlOc2`00000605?7l`<0
00001`1Cao<=000005l0oooo00230?ooo`030?oOW0180000M;oo0240oooo00L0oono07@0001d_ol0
oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool01`3OW4P0000007Boo`3oool0
oooO09a80018W=l0A`3oool600000080oooo00<000000?ooo`3oool0>03oool00`000000UW>Y09Ic
Z@0609IcZ@040000002FLjT0UW>Y000000P0UW:Y00@0000009IbZ@2FLZT000002@2ELZP010000000
UG:X09EbZ000000309EbZ0030000002DLJP0U76X00<0U76X2P00000605?7l`/000001P1Cao<30000
00L0DlOc2`00000605?7l`X000001`1Cao<3000000L0DlOc2`00001[0?ooo`00P@3oool01@3ooml0
W4P0000000000000M;oo0240oooo00<0_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`050?oo
o`030?oog`2LB00007Bo00<0oooo00<0okmd0018W03Oool0C03oool6000000D0oooo00<000000?oo
o`3oool0;`3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT0UW>Y000000P0UW:Y00@00000
09IbZ@2FLZT000002@2ELZP010000000UG:X09EbZ000000309EbZ0H000001P1Cao<:000000L0DlOc
5000000605?7l`/000001P1Cao<E000007H0oooo00230?ooo`030?noM0000000M;oo0200oooo00<0
oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`030;md0000M;l0oooo0080oooo00<0
_g@00018W03Oool0DP3oool6000000P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0N0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT00000202FLZT010000000UW:Y
09IbZ@00000909EbZ0040000002ELZP0UG:X000000D0UG:X2P00000705?7l`X000001P1Cao<30000
00L0DlOc2P00000705?7la400000Q03oool00:H0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
001809cOo`060?ooo`070?oOW01807@0_ooo0?ooo`2oM000000007Boo`1I0?ooo`H000000P3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo01h0oooo00<0000009IcZ@2FLjT01P2FLjT0
10000000UW>Y09IcZ@00000809IbZ@040000002FLZT0UW:Y000000T0UG:X00<0000009EbZ02ELZP0
1P00000705?7l`X000001P1Cao<D000000H0DlOc3`00002A0?ooo`00YP3oool00`3ogi`0B00007Bo
o`030?ooo`030?noM00004P0W=oo00L0oooo00D0oono07A8B0180000000004RLg`1P0?ooo`D00000
1@3oool00`000000oooo0?ooo`0N0?ooo`030000002FLjT0UW>Y00H0UW>Y00@0000009IcZ@2FLjT0
0000202FLZT010000000UW:Y09IbZ@00000909EbZ0030000002ELZP0UG:X00<0UG:X2P00000605?7
l`X000001P1Cao<3000000L0DlOc3@00002M0?ooo`00YP3oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00L0oooo00@0omnL04P00000000007BoIP3oool6000000P0oooo00<000000?oo
o`3oool05@3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT0UW>Y000000P0UW:Y00<00000
09IbZ@2FLZT04000000605?7l`X000001P1Cao<D00000:X0oooo002V0?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd0000B02Lgol01P3oool01P3ogi`0B0000000M02oool0okmd001d_aL0oooo00D0
omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00<0oooo00D0omnL04P000000000
000007Boo`110?ooo`H000000P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00d0
oooo00<0000009IcZ@2FLjT01P2FLjT010000000UW>Y09IcZ@00000809IbZ@030000002FLZT0UW:Y
0080UW:Y1000000705?7l`T000001`1Cao<A00000;H0oooo002V0?ooo`030?oo_`1d0000B9cO00<0
oooo00<0gia80000M02oool01P3oool00`2oM000001d0;ooo`020?ooo`030;md0018W=l0oooo01@0
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00<0okmd0000001d_ol00P3oool0
1`3ogi`0B00004RLg`3oool0oono07@0001d_ol0AP3oool6000000@0oooo00<000000?ooo`3oool0
3@3oool00`000000UW>Y09IcZ@0609IcZ@040000002FLjT0UW>Y000000@0UW:Y5000000605?7l`h0
0000``3oool00:L0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`060?ooo`030;md
0000B9`0gooo0080oooo00<0gia8001d_`3oool0503oool00`2oM00004RL0=ooo`020?ooo`030;md
0000M;l0oooo00D0oooo00<0oooO09a80000M;l00`3oool00`3o_g@004RL0=ooo`1;0?ooo`H00000
1P3oool00`000000oooo0?ooo`050?ooo`030000002FLjT0UW>Y00H0UW>Y00<0000009IcZ@2FLjT0
1@00000605?7l`@000001P1Cao<;00000=00oooo002W0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0
001d0;ooo`070?ooo`070?noM0000000M;oo0?ooo`3ookl0M00007Boo`0D0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool01@3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0
W=oo0540oooo1`0000070?ooo`030000002FLjT0UW>Y00H0UW>Y00<0000009IcZ@2FLjT00P2FLjTC
00000=/0oooo002X0?ooo`050?oo_`1d000000000000M02oool02@3oool01@3o_g@0000000000000
0000B9cO01D0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd0000B02Lgol0E`3oool600000080oooo00<0000009IcZ@2FLjT00P2F
LjT>00000>T0oooo003@0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01@3oool0
0`3ogi`0B00007Boo`030?ooo`030;md000004P0W=oo05d0oooo2@00003e0?ooo`00d03oool00`3o
gi`0B00007Boo`030?ooo`030?noM00004P0W=oo00D0oooo00P0oono07@000000000B9cO0?ooo`3o
_g@00000001d_a80oooo00D0omnL04P000000000001d0;ooo`030?ooo`030?noM0000000M;oo00D0
oooo00@0oooO09a80000B9`0goooo`3ooolg0?ooo`00d03oool00`3ogi`0B00007Boo`030?ooo`03
0?noM00004P0W=oo00H0oooo00<0_g@00018W02oW4P00P0000000`00B9`0gooo0?ooo`0@0?ooo`07
0?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?noM0000000M;oo00D0oooo00@0oooO
09a80000B9`0goooo`3ooolg0?ooo`00d03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00H0oooo00<0omnL04P0001d_ol05@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo00T0oooo00@0oooO09a80000B9`0goooo`3ooolg0?ooo`00d@3oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000M;l0oooo00L0oooo00<0gia80000B02Lgol04`3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00D0oooo00<0_g@00000000000001@0000000`1d_ol0oooo0?ooo`3o
0?oooc@0oooo003A0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`090?ooo`030?noM000
000007Bo0180oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`050?ooo`080?oOW01d
W=l0oooo0?ooo`3ooml0W4P00018W03Ooooo0?ooocL0oooo003B0?ooo`050?oo_`1d000000000000
M02oool02`3oool01@3ooml0W4P00000000007@0_ooo00l0oooo00<0omnL04P0001d_ol00`3oool0
0`3o_g@0001809cOo`060?ooo`070=nLB01d_ol0oooo0?oog`2LB00004RL0=ooo`3o0?ooocL0oooo
003f0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol01P3oool01`3ooml0W4Qd0;oo
o`3ooml0W4P00018W03Oool0403oool01@3ogi`0B0000000000007@0_ooo00<0oooo00D0okmd0000
001d_ol0oooo0;md00060000000304RLg`3oool0oooo0?l0oooo4@3oool00?H0oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@0001809cOo`070?ooo`060?oOW018M;l0oooO09a80000B9`0gooo3`3o
ool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool00`3o_g@0000007Boo`020?ooo`03
0=nLB0000000000000<0000000<004RL0=ooo`3oool0;P3oool01`2oM000001809cOo`3ogi`0B000
0018W03Oool0f`3oool00?H0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`080?oo
o`050=nLB01d_ml0W4P00018W03Oool03`3oool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0
oooo00L0oooo00<0gia804RLg`3oool00P3oool00`3OW7@0W=oo0?ooo`0]0?ooo`030=nLB00007@0
_ooo0080oooo00<0oono07@00000M;l0f`3oool00?L0oooo00<0gia80018W03Oool00P3oool00`2o
M00007Bo0?ooo`080?ooo`050?oog`2LB4P0B4P00018W03Oool03P3oool00`3ookl0M00004RLg`03
0?ooo`030=nLB00007@0_ooo00P0oooo00<0gia807Boo`3oool0<03oool00`3ooml0W4P004RLg`04
0?ooo`030=nLB000B9`0gooo0=X0oooo003g0?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;oo
o`0:0?ooo`040?oOW018000004RL0=ooo`h0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@00018
09cOo`090?ooo`030;mdB02Lgol0oooo02l0oooo00<0oooO09a80018W=l0103oool00`3OW4P004RL
0=ooo`3J0?ooo`00n03oool01@3ookl0M0000000000007@0_ooo00`0oooo00<0gia80018W03Oool0
3P3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0oooo00<0oooO09a8M02oool0
;`3oool00`3ooml0W4P004RLg`040?ooo`030=nLB000B9`0gooo0=X0oooo003o0?oooa/0oooo00<0
omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030?noM000M;l0oooo01<0oooo00D0
omnL04P000000000001d0;ooo`0F0?ooo`030?oog`2LB000B9cO00@0oooo00<0gia80018W03Oool0
fP3oool00?l0oooo6`3oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00X0oooo00<0
oono07@0B02Lgol04P3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool05@3oool00`3ooml0
W4P004RLg`040?ooo`030=nLB000B9`0gooo0=X0oooo003o0?oooa/0oooo00<0oono07@00018W=l0
0`3oool00`3OW4P0001d0;ooo`0:0?ooo`030?oo_`1d0000B9cO0180oooo00<0_g@00018W03Oool0
0P3oool00`2oM00007Bo0?ooo`0D0?ooo`030?oo_`1d0000M;oo00@0oooo00<0okmd0000M02oool0
fP3oool00?l0oooo703oool00`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00D0oooo00<0
oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`0A0?ooo`030?oo_`1d0000B9cO00<0oooo00<0
gia80000M02oool04`3oool00`3ogi`0B01809cOo`060?ooo`030?oOW01804P0W=oo0=T0oooo003o
0?oooa`0oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00H0oooo00P0oooO09a8B02Lgol0
oooo0?oo_`1d0000001d0;oooa40oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`0D
0?ooo`030?oo_`1d0000M;oo00@0oooo00<0okmd0000M02oool0fP3oool00?l0oooo7@3oool01@3o
okl0M0000000000007@0_ooo00P0oooo00H0oooO09a80000000000000000M02ooolB0?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd0000B02Lgol0503oool00`3ooml0W4P004RLg`040?ooo`030=nL
B000B9`0gooo0=X0oooo003o0?oood80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`0D0?ooo`030?oog`2LB000B9cO00@0oooo00<0gia80018W03Oool0fP3oool00?l0oooo@P3oool0
0`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo01@0oooo00<0oooO09a80018W=l0103oool0
0`3OW4P004RL0=ooo`3J0?ooo`00o`3ooom20?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool0503oool00`3ooml0W4P004RLg`040?ooo`030=nLB000B9`0gooo0=X0oooo003o0?oood<0
oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`0D0?ooo`030?oog`2LB000B9cO00@0
oooo00<0gia80018W03Oool0fP3oool00?l0oooo@`3oool01`3ookl0M00007Boo`3oool0okmd0000
M02oool05P3oool00`3OW4P0001d0;ooo`020?ooo`030?oo_`1d000007Bo0=/0oooo003o0?oood@0
oooo00D0oono07@000000000001d0;ooo`0G0?ooo`080?oog`2LB000001809cOo`3ogi`0B0000018
W03OoooK0?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?oo
o`00b@3oool01P3OW4P000000000000004P0W=oO09a80080000000<0001d0;oog`2LB0000P000000
0`00M;l0oooO09a80002000000070018W02LM0000000001d_`3ookl0_kno07@000020000000507Bo
o`3oool0oono0;no_`1d00000P0000000`1d_ol0oooo0?ooo`030?ooo`080;md0000000000000000
B02Lgol0oooo0?oog`2LB0020000000:0018W02LM0000000001d_`3oool0omnL04P000000000B9cO
0=nLB080000000@0001d0;ooo`3oool0gia80`0000000`1d_ol0oooo0?ooo`3o0?oooc80oooo003:
0?ooo`030=nLB00007@0_ooo0080oooo01h0_g@00018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o
_g@0000007Boo`3oool0omnL04P00000M;l0okno0;oo_`1d04P0W=oo0?ooo`3ookl0M01809cOo`3o
okl0M01809cOo`3oool0oono07@0B02Lgol50?ooo`030=nLB00007@0_ooo0080oooo01T0okmd0000
001d_ol0oooo0?oOW018000007Bo0?no_`2oool0gia80000M02oool0gkno0=ooo`3OW4P0001d0;oo
o`3oool0gia800000018W=l0oooo0?oog`2LB000B9cO0?l0oooo<`3oool00<X0oooo00<0gia80000
M02oool00P3oool04P2oM00004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00004P0W=oo0?oo
o`3ooml0W4P0001d_`3oool0oono0;noo`80oooo00D0okmd0000001d_ol0oono0;noo`020?ooo`03
0?noM0000000M;oo00D0oooo00<0gia80000M02oool00P3oool01`3o_g@0001809cOo`3oool0oooO
09a80000M;l00P3oool00`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0_ooo0?oOW018000007Bo
0`3oool00`3o_g@0001d0;ooo`3o0?oooc80oooo003:0?ooo`030=nLB00007@0_ooo0080oooo00L0
_g@00018W03Oool0oooo0?oog`2LB00007Bo0080oooo00H0okmd00000000M;l0oooO09a80000M;l3
0?ooo`050?oo_`1d000000000000M02oool00P3oool01@3ookl0M0000000000007@0_ooo00D0oooo
00<0gia80000M02oool00`3oool01P3o_g@00000001d_`3ooml0W4P0001d_`80oooo00<0gia80018
W03Oool00P3oool01P3OW4P0001d0;ooo`3o_g@0000007Boo`<0oooo00<0okmd0000001d_ol0o`3o
oolb0?ooo`00bP3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0
001d_`040?ooo`040?noM00000000000001d_`80oooo00@0_g@0000000000000B9cO0`3oool0102o
M000000000000018W=l70?ooo`030=nLB00007@0_ooo00D0oooo00@0okmd00000000000007Bo0P3o
ool00`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0_ooo0?noM0000000M;oo0`3oool00`3o_g@0
000007Boo`3o0?oooc80oooo003:0?ooo`030=nLB00004P0W=oo0080oooo01<0_g@00000M02oool0
oooo0?oog`2LB00007Bo0?ooo`3ookl0M00007Boo`3oool0oooO09a80000M;l0oooo0?oOW0180000
B9cO0080oooo00D0_ibL0=ooo`3ogi`0B00004RLg`020?ooo`030;nLW03Oool0oooo00@0oooo00<0
gia80000B02Lgol00P3oool01`3ookl0M00007Boo`3oool0oooO09a80000M;l00P3oool00`3OW4P0
04RL0=ooo`020?ooo`060=nLB00007@0_ooo0?oo_`1d0000M;oo0`3oool00`2oM000001809cOo`3o
0?oooc80oooo00380?ooo`0U0?oog`2LB00000000000B018B4P0W=oo0?oo_`1d000007Ad07@0M02o
ool0oono07@00018W=l0oooo0?oo_`1d0000B9cO0?ooo`3ookl0M000001d_`3oool0oono07@0M02o
ool0oooo0?noM000B9`0gooo0?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`030?ooo`0>0?oog`2L
B00000000000B018M9`0gkmd0000M02ookl0M00004RLg`3oool0oono07@00000M;l20?ooo`0@0=nL
B000B9`0gooo0?oog`2LB00000000000M02oool0oooo0=nLB000B9`0gooo0?ooo`3o_g@00000001d
_ol0oooo<`3oool00<X0oooo00`0_g@00000M02ooml0W4P00000000004P0W=oo0?oo_`1d00000000
0000M02oool20?ooo`060?oo_`1d000000000000000007@0_ooo0P3oool0203o_g@0000000000000
M;l0_ibL0=ooo`3oool0okmd0P0000000`00M;l0_ibL0=ooo`050?ooo`080;md000007@0_mnL04P0
000007@0_ooo0?oo_`1d0002000000040000M02oool0oooo0;md00<0000000D004RL0=ooo`2oM000
001d0;ooo`020?ooo`050?noM0000000000000000018W=l0o`3ooold0?ooo`00o`3oool10?ooo`03
0;md0000B9`0gooo0?l0oooo@03oool00?l0oooo0@3oool00`3ookl0M4RL0=ooo`3o0?oood00oooo
003o0?ooo`40oooo00<0oooO09adW03Oool0o`3ooom00?ooo`00o`3oool60?ooo`030=nLB00004P0
W=oo0?l0oooo>`3oool00?l0oooo1P3oool00`3OW4P0001809cOo`3o0?oooc/0oooo003o0?ooool0
ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool0
0?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3o
oom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00
o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oo
od@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o
0?ooool0ooooA03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0oooo
A03oool00?l0ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0
ooooo`3ooom40?ooo`00o`3ooooo0?oood@0oooo003o0?ooool0ooooA03oool00?l0ooooo`3ooom4
0?ooo`00o`3ooooo0?oood@0oooo0000\
\>"],
  ImageRangeCache->{{{0, 577.813}, {546.375, 0}} -> {-0.257284, -0.114994, \
0.00312228, 0.00312228}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(plotThetaMax = 
        Plot[Evaluate[
            Table[thetaMax[mratio, inelasticity]\/Degree, {mratio, 0, 2, 
                0.2}]], {inelasticity, \(-1\), 1}, 
          PlotRange \[Rule] {Automatic, {0, 180}}, PlotPoints \[Rule] 50, 
          Frame \[Rule] True, GridLines \[Rule] Automatic, 
          FrameLabel \[Rule] {"\<relative inelasticity\>", \*"\"\<Max[\!\(\
\[Theta]\_3\)] (deg)\>\""}, 
          PlotLabel \[Rule] "\<Maximum Scattering Angle\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.47619 -1.30104e-017 0.00343352 [
[.02381 -0.0125 -9.15625 -15.6875 ]
[.02381 -0.0125 9.15625 0 ]
[.2619 -0.0125 -13.6563 -15.6875 ]
[.2619 -0.0125 13.6563 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -47.2813 -31.375 ]
[.5 -0.0125 47.2813 -15.6875 ]
[-0.0125 .08584 -16 -7.84375 ]
[-0.0125 .08584 0 7.84375 ]
[-0.0125 .17168 -16 -7.84375 ]
[-0.0125 .17168 0 7.84375 ]
[-0.0125 .25751 -16 -7.84375 ]
[-0.0125 .25751 0 7.84375 ]
[-0.0125 .34335 -22 -7.84375 ]
[-0.0125 .34335 0 7.84375 ]
[-0.0125 .42919 -22 -7.84375 ]
[-0.0125 .42919 0 7.84375 ]
[-0.0125 .51503 -22 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[-0.0125 .60087 -22 -7.84375 ]
[-0.0125 .60087 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -37.6875 -35.875 ]
[-0.01875 .30902 -22 35.875 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -68.6875 0 ]
[.5 .63053 68.6875 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.2619 0 m
.2619 .61803 L
s
.5 0 m
.5 .61803 L
s
.7381 0 m
.7381 .61803 L
s
0 .08584 m
1 .08584 L
s
0 .17168 m
1 .17168 L
s
0 .25751 m
1 .25751 L
s
0 .34335 m
1 .34335 L
s
0 .42919 m
1 .42919 L
s
0 .51503 m
1 .51503 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.1563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.6563 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.313 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -108.281 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(relative) show
101.625 15.813 moveto
(inelasticity) show
153.563 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .08584 m
.00625 .08584 L
s
gsave
-0.0125 .08584 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(25) show
1.000 setlinewidth
grestore
0 .17168 m
.00625 .17168 L
s
gsave
-0.0125 .17168 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(50) show
1.000 setlinewidth
grestore
0 .25751 m
.00625 .25751 L
s
gsave
-0.0125 .25751 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(75) show
1.000 setlinewidth
grestore
0 .34335 m
.00625 .34335 L
s
gsave
-0.0125 .34335 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
0 .42919 m
.00625 .42919 L
s
gsave
-0.0125 .42919 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(125) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
0 .60087 m
.00625 .60087 L
s
gsave
-0.0125 .60087 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(175) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .01717 m
.00375 .01717 L
s
0 .03434 m
.00375 .03434 L
s
0 .0515 m
.00375 .0515 L
s
0 .06867 m
.00375 .06867 L
s
0 .10301 m
.00375 .10301 L
s
0 .12017 m
.00375 .12017 L
s
0 .13734 m
.00375 .13734 L
s
0 .15451 m
.00375 .15451 L
s
0 .18884 m
.00375 .18884 L
s
0 .20601 m
.00375 .20601 L
s
0 .22318 m
.00375 .22318 L
s
0 .24035 m
.00375 .24035 L
s
0 .27468 m
.00375 .27468 L
s
0 .29185 m
.00375 .29185 L
s
0 .30902 m
.00375 .30902 L
s
0 .32618 m
.00375 .32618 L
s
0 .36052 m
.00375 .36052 L
s
0 .37769 m
.00375 .37769 L
s
0 .39486 m
.00375 .39486 L
s
0 .41202 m
.00375 .41202 L
s
0 .44636 m
.00375 .44636 L
s
0 .46353 m
.00375 .46353 L
s
0 .48069 m
.00375 .48069 L
s
0 .49786 m
.00375 .49786 L
s
0 .5322 m
.00375 .5322 L
s
0 .54936 m
.00375 .54936 L
s
0 .56653 m
.00375 .56653 L
s
0 .5837 m
.00375 .5837 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -126.719 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
96.875 11.8438 translate
90 rotate
-96.875 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
%%BeginResource: font Times-Roman-MISO
%%BeginFont: Times-Roman-MISO
/Times-Roman /Times-Roman-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Max) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
85.063 15.813 moveto
(@) show
89.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(q) show
94.250 17.688 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
99.063 15.813 moveto
(D) show
106.000 15.813 moveto
(H) show
109.813 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(deg) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
126.938 15.813 moveto
(L) show
130.750 15.813 moveto
%%IncludeResource: font Times-Roman-MISO
%%IncludeFont: Times-Roman-MISO
/Times-Roman-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .08584 m
1 .08584 L
s
.99375 .17168 m
1 .17168 L
s
.99375 .25751 m
1 .25751 L
s
.99375 .34335 m
1 .34335 L
s
.99375 .42919 m
1 .42919 L
s
.99375 .51503 m
1 .51503 L
s
.99375 .60087 m
1 .60087 L
s
.125 Mabswid
.99625 .01717 m
1 .01717 L
s
.99625 .03434 m
1 .03434 L
s
.99625 .0515 m
1 .0515 L
s
.99625 .06867 m
1 .06867 L
s
.99625 .10301 m
1 .10301 L
s
.99625 .12017 m
1 .12017 L
s
.99625 .13734 m
1 .13734 L
s
.99625 .15451 m
1 .15451 L
s
.99625 .18884 m
1 .18884 L
s
.99625 .20601 m
1 .20601 L
s
.99625 .22318 m
1 .22318 L
s
.99625 .24035 m
1 .24035 L
s
.99625 .27468 m
1 .27468 L
s
.99625 .29185 m
1 .29185 L
s
.99625 .30902 m
1 .30902 L
s
.99625 .32618 m
1 .32618 L
s
.99625 .36052 m
1 .36052 L
s
.99625 .37769 m
1 .37769 L
s
.99625 .39486 m
1 .39486 L
s
.99625 .41202 m
1 .41202 L
s
.99625 .44636 m
1 .44636 L
s
.99625 .46353 m
1 .46353 L
s
.99625 .48069 m
1 .48069 L
s
.99625 .49786 m
1 .49786 L
s
.99625 .5322 m
1 .5322 L
s
.99625 .54936 m
1 .54936 L
s
.99625 .56653 m
1 .56653 L
s
.99625 .5837 m
1 .5837 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -129.688 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Maximum) show
116.125 15.813 moveto
(Scattering) show
167.313 15.813 moveto
(Angle) show
196.375 15.813 moveto
%%IncludeResource: font Times-Roman
%%IncludeFont: Times-Roman
/Times-Roman findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.49127 .61803 L
.50987 .61803 L
.52967 .61803 L
.54872 .61803 L
.56898 .61803 L
.5885 .61803 L
.60727 .61803 L
.62725 .61803 L
.64648 .61803 L
.66497 .61803 L
.68466 .61803 L
.70361 .61803 L
.72376 .61803 L
.74317 .61803 L
.76183 .61803 L
.7817 .61803 L
.80082 .61803 L
.82115 .61803 L
.84073 .61803 L
.85957 .61803 L
.87961 .61803 L
.89891 .61803 L
.91747 .61803 L
.93723 .61803 L
.95624 .61803 L
.97619 .61803 L
Mistroke
Mfstroke
.02381 .10969 m
.04273 .10759 L
.06337 .10528 L
.08275 .10307 L
.10139 .10093 L
.12123 .09861 L
.14033 .09634 L
.16063 .0939 L
.18019 .0915 L
.199 .08915 L
.21902 .08661 L
.23829 .0841 L
.25683 .08164 L
.27656 .07896 L
.29555 .0763 L
.31574 .0734 L
.33519 .07051 L
.3539 .06764 L
.3738 .06447 L
.39297 .06128 L
.41334 .05773 L
.43296 .05412 L
.45184 .05042 L
.47193 .0462 L
.49127 .04177 L
.50987 .03704 L
.52967 .03127 L
.54872 .02451 L
.55913 .0199 L
.5639 .01739 L
.56898 .01424 L
.57362 .01059 L
.57489 .00935 L
.57626 .00779 L
.57688 .00696 L
.57755 .00596 L
.57817 .00483 L
.57873 .00351 L
.57929 .00123 L
.57981 .61803 L
.58036 .61803 L
.58096 .61803 L
.58304 .61803 L
.58775 .61803 L
.6078 .61803 L
.6271 .61803 L
.64565 .61803 L
.66541 .61803 L
.68443 .61803 L
.70464 .61803 L
Mistroke
.72412 .61803 L
.74285 .61803 L
.76278 .61803 L
.78198 .61803 L
.80237 .61803 L
.82202 .61803 L
.84093 .61803 L
.86103 .61803 L
.8804 .61803 L
.89902 .61803 L
.91885 .61803 L
.93793 .61803 L
.95822 .61803 L
.97619 .61803 L
Mfstroke
.02381 .19932 m
.04273 .1945 L
.06337 .18936 L
.08275 .18462 L
.10139 .18013 L
.12123 .17541 L
.14033 .17091 L
.16063 .16616 L
.18019 .16162 L
.199 .15726 L
.21902 .15263 L
.23829 .14816 L
.25683 .14386 L
.27656 .13925 L
.29555 .13479 L
.31574 .13 L
.33519 .12532 L
.3539 .12076 L
.3738 .11583 L
.39297 .11098 L
.41334 .10569 L
.43296 .10045 L
.45184 .09524 L
.47193 .08948 L
.49127 .08368 L
.50987 .0778 L
.52967 .07112 L
.54872 .06418 L
.56898 .056 L
.5885 .04697 L
.60727 .0364 L
.61696 .02959 L
.62191 .02545 L
.62725 .02006 L
.62967 .01707 L
.63222 .01322 L
.63327 .01126 L
.6344 .00868 L
.63505 .00678 L
.63563 .00438 L
.63622 .61803 L
.63678 .61803 L
.64574 .61803 L
.66458 .61803 L
.68462 .61803 L
.70392 .61803 L
.72247 .61803 L
.74223 .61803 L
.76125 .61803 L
.78147 .61803 L
Mistroke
.80094 .61803 L
.81967 .61803 L
.83961 .61803 L
.8588 .61803 L
.87919 .61803 L
.89884 .61803 L
.91775 .61803 L
.93786 .61803 L
.95722 .61803 L
.97585 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.1708 .61803 L
.17527 .61803 L
.17764 .61803 L
.17895 .61803 L
.18019 .61803 L
.18137 .61803 L
.18169 .61803 L
.18203 .61803 L
.18265 .30606 L
.18326 .30152 L
.18383 .29897 L
.1849 .29542 L
.18703 .29028 L
.18932 .28596 L
.199 .27297 L
.20877 .26337 L
.21909 .25494 L
.23769 .24214 L
.25749 .23048 L
.27654 .22041 L
.2968 .21054 L
.31632 .20159 L
.33509 .19337 L
.35507 .18493 L
.3743 .17702 L
.39279 .16955 L
.41248 .16168 L
.43143 .15416 L
.45158 .14616 L
.47099 .1384 L
.48965 .13084 L
.50952 .12264 L
.52864 .11451 L
.54897 .10555 L
.56855 .09647 L
.58739 .08717 L
.60743 .07641 L
.62673 .06476 L
.64529 .05155 L
.65561 .04265 L
.66505 .03263 L
.66972 .02636 L
Mistroke
.67218 .02238 L
.67353 .01986 L
.6748 .01717 L
.67602 .01413 L
.67665 .01224 L
.67733 .00986 L
.67789 .00728 L
.67852 .00209 L
.67908 .61803 L
.67961 .61803 L
.68406 .61803 L
.70315 .61803 L
.72343 .61803 L
.74297 .61803 L
.76177 .61803 L
.78177 .61803 L
.80103 .61803 L
.81955 .61803 L
.83926 .61803 L
.85824 .61803 L
.87841 .61803 L
.89785 .61803 L
.91654 .61803 L
.93643 .61803 L
.95558 .61803 L
.97593 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.36434 .61803 L
.3738 .61803 L
.37611 .61803 L
.37855 .61803 L
.37982 .61803 L
.38017 .61803 L
.38054 .61803 L
.38089 .61803 L
.38122 .30346 L
.38182 .29893 L
.38216 .29711 L
.38248 .29565 L
.38367 .29115 L
.38582 .2851 L
.38818 .27982 L
.39297 .27128 L
.40245 .2583 L
.41266 .24708 L
.43281 .22892 L
.45222 .21404 L
.47088 .20109 L
.49075 .18821 L
.50988 .17635 L
.5302 .1641 L
.54978 .15243 L
.56862 .14118 L
.58867 .12902 L
.60797 .11691 L
.62652 .10468 L
.64628 .09063 L
.6653 .07548 L
.67497 .06679 L
Mistroke
.68551 .05605 L
.69573 .0435 L
.70018 .03683 L
.70499 .028 L
.70758 .02184 L
.70872 .01852 L
.70932 .0165 L
.70996 .01402 L
.71058 .01114 L
.71114 .00764 L
.71171 .61803 L
.71224 .61803 L
.71464 .61803 L
.72492 .61803 L
.74372 .61803 L
.76372 .61803 L
.78298 .61803 L
.8015 .61803 L
.82121 .61803 L
.84019 .61803 L
.86036 .61803 L
.8798 .61803 L
.89849 .61803 L
.91838 .61803 L
.93753 .61803 L
.95788 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.48193 .61803 L
.48681 .61803 L
.49127 .61803 L
.49347 .61803 L
.49588 .61803 L
.49698 .61803 L
.49814 .61803 L
.49914 .61803 L
.49971 .61803 L
.50024 .30273 L
.50081 .29757 L
.50142 .29381 L
.50266 .28819 L
.50487 .28077 L
.50987 .26869 L
.51947 .25196 L
.52973 .23796 L
.55006 .21531 L
.56964 .1966 L
.58848 .18005 L
.60852 .16322 L
.62782 .14725 L
.64638 .13172 L
.66614 .11451 L
.68515 .0966 L
.70537 .07472 L
Mistroke
.71548 .06164 L
.72039 .05433 L
.72485 .04685 L
.72921 .03825 L
.7315 .0329 L
.73392 .02613 L
.73516 .0219 L
.73581 .0193 L
.73651 .01606 L
.73716 .01235 L
.73775 .00754 L
.73834 .61803 L
.7389 .61803 L
.74358 .61803 L
.76238 .61803 L
.78238 .61803 L
.80164 .61803 L
.82015 .61803 L
.83987 .61803 L
.85884 .61803 L
.87902 .61803 L
.89845 .61803 L
.91714 .61803 L
.93703 .61803 L
.95618 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.49127 .61803 L
.50987 .61803 L
.52967 .61803 L
.54872 .61803 L
.55843 .61803 L
.56898 .61803 L
.5739 .61803 L
.57659 .61803 L
.57791 .61803 L
.57849 .61803 L
.57912 .61803 L
.57966 .30107 L
.58016 .29592 L
.5807 .2921 L
.58126 .2888 L
.58358 .27884 L
.58615 .27061 L
.5885 .26437 L
.59789 .24483 L
.60808 .22828 L
.62812 .20146 L
.64742 .1789 L
.66598 .15841 L
.68574 .13676 L
.70475 .11507 L
.72497 .08942 L
Mistroke
.73507 .07452 L
.74444 .05813 L
.74912 .04822 L
.75352 .03675 L
.75595 .0286 L
.75733 .02281 L
.75798 .01948 L
.75859 .01575 L
.7591 .01172 L
.75965 .00435 L
.76026 .61803 L
.76082 .61803 L
.76318 .61803 L
.78272 .61803 L
.80152 .61803 L
.82152 .61803 L
.84078 .61803 L
.85929 .61803 L
.87901 .61803 L
.89798 .61803 L
.91816 .61803 L
.93759 .61803 L
.95628 .61803 L
.97618 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.49127 .61803 L
.50987 .61803 L
.52967 .61803 L
.54872 .61803 L
.56898 .61803 L
.5885 .61803 L
.60727 .61803 L
.61677 .61803 L
.62181 .61803 L
.62725 .61803 L
.62977 .61803 L
.63242 .61803 L
.63469 .61803 L
.63531 .61803 L
.63565 .61803 L
.63597 .61803 L
.6366 .29684 L
.63717 .29158 L
.63829 .28422 L
.63949 .27826 L
.64163 .26972 L
.64648 .25499 L
.65607 .23323 L
.66504 .21671 L
.68405 .18683 L
.70427 .15819 L
Mistroke
.72374 .13093 L
.74248 .1028 L
.75294 .08498 L
.76241 .06602 L
.76716 .05454 L
.76966 .04754 L
.7723 .03894 L
.77461 .02952 L
.77524 .02639 L
.77592 .02256 L
.77651 .01865 L
.77715 .01315 L
.77777 .00165 L
.77834 .61803 L
.77889 .61803 L
.77948 .61803 L
.7816 .61803 L
.80012 .61803 L
.81983 .61803 L
.83881 .61803 L
.85898 .61803 L
.87842 .61803 L
.89711 .61803 L
.917 .61803 L
.93615 .61803 L
.9565 .61803 L
.97611 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.49127 .61803 L
.50987 .61803 L
.52967 .61803 L
.54872 .61803 L
.56898 .61803 L
.5885 .61803 L
.60727 .61803 L
.62725 .61803 L
.64648 .61803 L
.66497 .61803 L
.66958 .61803 L
.67444 .61803 L
.67699 .61803 L
.67769 .61803 L
.67807 .61803 L
.67843 .61803 L
.67877 .30084 L
.67913 .29521 L
.67977 .28882 L
.68036 .28435 L
.68101 .28021 L
.68232 .27324 L
.68466 .26324 L
.69418 .2346 L
.70442 .2116 L
.72463 .17383 L
Mistroke
.74411 .14019 L
.76284 .10612 L
.7733 .08424 L
.78278 .05983 L
.78753 .04352 L
.78883 .03796 L
.7902 .03109 L
.79138 .02377 L
.79204 .01839 L
.79266 .01129 L
.79324 .61803 L
.79387 .61803 L
.79502 .61803 L
.79716 .61803 L
.80197 .61803 L
.82197 .61803 L
.84123 .61803 L
.85974 .61803 L
.87946 .61803 L
.89843 .61803 L
.91861 .61803 L
.93804 .61803 L
.95673 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.49127 .61803 L
.50987 .61803 L
.52967 .61803 L
.54872 .61803 L
.56898 .61803 L
.5885 .61803 L
.60727 .61803 L
.62725 .61803 L
.64648 .61803 L
.66497 .61803 L
.68466 .61803 L
.69371 .61803 L
.69884 .61803 L
.70361 .61803 L
.70835 .61803 L
.70972 .61803 L
.71039 .61803 L
.71101 .61803 L
.71157 .61803 L
.71217 .29429 L
.71277 .28749 L
.71342 .2819 L
.716 .26643 L
.71876 .2543 L
.72376 .23696 L
.74398 .18608 L
Mistroke
.76345 .14498 L
.78218 .10377 L
.79184 .07857 L
.79719 .06147 L
.79954 .05254 L
.80212 .0408 L
.8033 .03415 L
.80457 .02525 L
.80493 .0221 L
.80528 .01865 L
.80592 .0092 L
.80652 .61803 L
.80685 .61803 L
.80717 .61803 L
.81262 .61803 L
.82251 .61803 L
.84177 .61803 L
.86028 .61803 L
.88 .61803 L
.89897 .61803 L
.91915 .61803 L
.93858 .61803 L
.95727 .61803 L
.97619 .61803 L
Mfstroke
.02381 .61803 m
.04273 .61803 L
.06337 .61803 L
.08275 .61803 L
.10139 .61803 L
.12123 .61803 L
.14033 .61803 L
.16063 .61803 L
.18019 .61803 L
.199 .61803 L
.21902 .61803 L
.23829 .61803 L
.25683 .61803 L
.27656 .61803 L
.29555 .61803 L
.31574 .61803 L
.33519 .61803 L
.3539 .61803 L
.3738 .61803 L
.39297 .61803 L
.41334 .61803 L
.43296 .61803 L
.45184 .61803 L
.47193 .61803 L
.49127 .61803 L
.50987 .61803 L
.52967 .61803 L
.54872 .61803 L
.56898 .61803 L
.5885 .61803 L
.60727 .61803 L
.62725 .61803 L
.64648 .61803 L
.66497 .61803 L
.68466 .61803 L
.70361 .61803 L
.71323 .61803 L
.72376 .61803 L
.72862 .61803 L
.73129 .61803 L
.73381 .61803 L
.73616 .61803 L
.73681 .61803 L
.73749 .61803 L
.73807 .61803 L
.73871 .29163 L
.73929 .28479 L
.73982 .27989 L
.74101 .27106 L
.74317 .25874 L
Mistroke
.748 .23798 L
.75248 .22249 L
.76264 .193 L
.78183 .14444 L
.79163 .11942 L
.80222 .08922 L
.80721 .07233 L
.80993 .06161 L
.81247 .04989 L
.81463 .03736 L
.81526 .03293 L
.81585 .02814 L
.81637 .02307 L
.81695 .01573 L
.81755 .61803 L
.81818 .61803 L
.81931 .61803 L
.82187 .61803 L
.84159 .61803 L
.86057 .61803 L
.88074 .61803 L
.90017 .61803 L
.91887 .61803 L
.93876 .61803 L
.95791 .61803 L
.97619 .61803 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{369.375, 228.313},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008Z0001ER000`400?l00000o`00003oo`3ooooo
0?ooob`0oooo003o0?ooool0oooo;03oool00?l0ooooo`3oool/0?ooo`00o`3ooom]0?ooo`040=nL
B0000000000007BookX0oooo003o0?ooog00oooo00<0_gAd0;ooo`3oool0^03oool00?l0ooooL03o
ool00`3ogi`0M9cO0?ooo`2h0?ooo`00o`3ooom`0?ooo`030?oog`2LB7@0_ooo0;P0oooo003`0?oo
o`050;md0000000000000000B02Lgol00`3oool01`2oM00000000000000007@0_ooo0?ooo`3OW4P0
0P0000000`0007@0_ooO09a800020000000:0018W02LM0000000001d_`3oool0omnL04P000000000
B9cO0=nLB080000000L0001d0;ooo`3oool0oooo0?oog`2LB7@0_ooo00D0oooo00D0_g@000000000
0000001d0;ooo`070?ooo`050=nLB000000000000000M01dM0000P0000000`0007@0_ooO09a80002
000000040018W03Oool0oooo0;md0080000000@0001d0;ooo`3oool0gia80P0000000`0007@0_ooO
09a80002000000070018W02LM0000000001d_`3ookl0_kno07@000020000000807Boo`3oool0oooo
0?oOW0180000000004RLg`3OW4P2000000040000M02oool0oooo0;md0080000000@0001809cOo`3o
ool0gia80P0000001P0007@0_ooo0?oOW0180000000004RLg`@0oooo00<0_g@007Boo`3oool0]`3o
ool00?40oooo00<0gia80000M02oool00`3oool0702oM0000000001d_`3oool0omnL04P0M02oool0
oooo0=nLB00007@0_ooo0?noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0_ooo0=nLB00007@0
_ooo0=no_`3Oool0gia80000M02oool30?ooo`030?noM0000000B9cO00@0oooo00L0_g@000000000
M;l0oooo0?oOW01807@0_ooo00L0oooo00/0gia80000M02oool0oooo0=nLB000B9`0gooo0?ooo`3o
oml0W4P0001d_`020?ooo`0e0;md0000000007Bo0?ooo`3ogi`0B01d0;ooo`3oool0gia80000M02o
ool0okmd0000001d_ol0oooo0?oOW018000007Bo0?no_`2ookl0M01809cOo`3oool0oono07@0B02L
gol0oooo0=nLB00007@0_ooo0=no_`3Oool0gia80000M02oool0oooo0;md0000000004RL0=ooo`3o
oml0W4Qd0;ooo`3oool0gia80000M02oool0oooo0=nLB00007@0_ooo0=no_`3Oool00P3oool0103o
gi`0B0000018W03Ooong0?ooo`00l@3oool00`3OW4P0001d0;ooo`020?ooo`030?oOW0180000B9cO
00<0oooo00d0omnL07BLg`3oool0gia80000M02oool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo
0080oooo00<0gia80018W03Oool00P3oool00`3OW4P0001d0;ooo`020?ooo`050?oog`2LB00004RL
09adM02oool00P3oool00`3ogi`0B00004RLg`030?ooo`030?oOW01dW=l0oooo00H0oooo00l0gia8
0000M02oool0oooo0=nLB000B9`0gooo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l00`3o
ool0403ogi`0M9cO0?ooo`3OW4P0001d0;ooo`3o_g@0001809cOo`3oool0oooO09a80000M;l0oooo
0?oo_`2o_ol20?ooo`070?noM0000000M;oo0?ooo`3OW4P004RL0=ooo`020?ooo`060=nLB00007@0
_ooo0?oOW018000007Bo0`3oool02P3ogi`0W;oo0?ooo`3OW4P0001d0;ooo`3oool0gia80018W03O
ool40?ooo`040;md0000B9`0W7@007BookL0oooo003a0?ooo`030=nLB00007@0_ooo0080oooo00<0
okmd0000B02Lgol01P3oool02P3OW4P0001d0;ooo`3oool0okmd00000000M;l0oooO09a80000M;l2
0?ooo`030=nLB000B9`0gooo0080oooo00<0gia80000M02oool00P3oool01@3ogi`0B00007Boo`3o
_g@0B9cO0080oooo00<0okmd0000B02Lgol0303oool03`3OW4P0001d0;ooo`3oool0gia80018W03O
ool0oooo0?oog`2LB00007Bo0?ooo`3o_g@0001809cOo`060?ooo`0:0=nLB00007@0_ooo0?ooo`3o
_g@00000001d_`3ooml0W4P0001d_`<0oooo00T0oono07@000000000001d0;ooo`3oool0gia80018
W03Oool00P3oool01P3OW4P0001d0;ooo`3o_g@0001809cOo`H0oooo00L0gia80000M02oool0oooo
0=nLB000B9`0gooo00<0oooo00D0okmd0000001d_ol0omnL04Qd_`2g0?ooo`00l@3oool00`3OW4P0
001d0;ooo`020?ooo`030?noM00007@0_ooo00H0oooo00<0gia80000M02oool00`3oool0103o_g@0
000000000000M;l20?ooo`030=nLB000B9`0gooo0080oooo00<0gia80000M02oool00P3oool02P2o
M00004RL0=ooo`3ooml0W4Qd0;ooo`3oool0okmd0000M02oool<0?ooo`0?0=nLB00007@0_ooo0?oo
o`3OW4P004RL0=ooo`3oool0oooO09a80000M;l0oooo0?noM00007@0_ooo00H0oooo00<0gia80000
M02oool00`3oool0103o_g@0000000000000M;l20?ooo`040;md00000000000004RLg`<0oooo00<0
gia80018W03Oool00P3oool01P3OW4P0001d0;ooo`3o_g@0001d0;ooo`H0oooo00L0gia80000M02o
ool0oooo0=nLB000B9`0gooo0080oooo00L0oooO09a80000B9`0gooo0?ooo`2oM4P0W=oo0;H0oooo
003a0?ooo`030=nLB00004P0W=oo0080oooo00<0omnL04P000000000100000003018W=l0oooo0=nL
B00007@0_ooo0?oo_`1d0000M;oo0?ooo`3ooml0W4P0001d_`80oooo00<0gia80018W03Oool00P3o
ool01`3OW4P0001d0;ooo`3oool0omnL04P0001d_ol00P3oool01@3o_g@0B9cO0?ooo`3ogi`0B000
00D0000000<0B9cO0?ooo`3oool01@3oool03P3OW4P0001d0;ooo`3oool0_g@00000M02oool0oooo
0?oog`2LB00007Bo0?ooo`3ogi`0B0001@0000004018W=l0oooo0=nLB00007@0_ooo0?oo_`1d0000
M;oo0?ooo`3ooml0W4P0001d_`3oool0omnL04P00018W=l20?ooo`060;nLW03Oool0oooo0=nLB000
B9`0gooo0P3oool01P3OW4P0001d0;ooo`3ogi`0B01809cOo`80oooo00/0okmd0000B02Lgol0oooo
0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`020?ooo`030?noM00004P0W=oo0080oooo00<0okmd
04RLg`3oool0]@3oool00>l0oooo00X0oooO09a800000000001804QdW03O_g@0001d0;ooo`2oM000
B9cO0P3oool03P3OW4P0001809cOo`3oool0gia80000M02oool0oono07@00018W=l0oooo0?oo_`1d
000007Bo0P3oool0303OW4P004RL0=ooo`3ooml0W4P00000000007@0_ooo0?oog`2LB000001d0;oo
o`80oooo00H0oooO09a8M02oool0oooo0;md0018W=l20?ooo`030=nLB00004P0W=oo00D0oooo00d0
oooO09a800000000001d0;oog`2LB00000000018B01804P0W=oo0?oo_`1d0000B9cO0080oooo03P0
_g@004RLg`3oool0oooo0=nLB00004P0W=oo0?ooo`3OW4P0001d0;ooo`3ookl0M00004RLg`3oool0
oono07@00000M;l0oooo0?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`3oool0gia80018W03Oool0
oooO09a800000000001d0;ooo`3oool0_g@00018W03Oool0oooO09a8000007@0_ooO09a800000000
001d0;ooo`3oool0gia80018W03Oool0oooo0?oo_`1d000004RL0=ooo`80oooo00<0oono07@0M02o
ool0]@3oool00?40oooo00P0_g@00000M02ogi`0B0000000M02oool0oooo0=nLB080000000X00018
09cOo`3oool0oooo0=nLB00007@0_ooo0?ooo`3ookl0M0000P000000100007@0_ooo0?ooo`2oM003
000000060018W03Oool0_g@00000M02ooml0W4P00P0000002@0007@0_ooo0?oog`2LB00000000018
W03Oool0oooo0=nLB002000000030000B02Lgol0oooo00L0oooo00/0_g@00000M02oool0oooo0;md
0000B9`0gooO09a800000000001d0;ooo`030?ooo`050=nLB000000000000000B02Lgol00P3oool0
1P3OW4P0001d0;ooo`3oool0oono07@00080000000D0001d0;ooo`3oool0oooo0?noM00200000004
001d_`2oW9`0gooo0;md00<0000000D004RL0=ooo`2oM000001d0;ooo`020?ooo`050=nLB0000000
00000000B02Lgol00P3oool0102oM000001d0;ooo`2oM0030000000;0018M01dB000000000000000
B9`0gooo0?ooo`2oM00000000018W03Oool0]03oool00?l0oooo00<0gia80000M02oool02@3oool0
0`2oM00004RL0=ooo`0b0?ooo`030=nLB00007@0_ooo0100oooo00<0_g@00018W03Oool04P3oool0
0`2oM00004RL0=ooo`2o0?ooo`00o`3oool00`3OW4P0001d0;ooo`090?ooo`030?oo_`1dB9`0gooo
0380oooo00<0gia80000M02oool0403oool00`3ookl0M4RL0=ooo`0B0?ooo`030?oo_`1dB9`0gooo
0;l0oooo003o0?ooo`030=nLB00007@0_ooo00T0oooo00<0oooO09adW03Oool0<P3oool00`3OW4P0
001d0;ooo`0@0?ooo`030?oog`2LM9`0gooo0180oooo00<0oooO09adW03Oool0_`3oool00?d0oooo
00D0oooO09a800000000001d0;ooo`0>0?ooo`030=nLB00004P0W=oo01P0oooo00<0gia80000B02L
gol0403oool01@3ooml0W4P00000000007@0_ooo01D0oooo00<0gia80000B02Lgol02@3oool00`3O
W4P0001809cOo`330?ooo`00o`3oool00`2oM000001d0;ooo`0>0?ooo`030=nLB00004P0W=oo01P0
oooo00<0gia80000B02Lgol04P3oool00`2oM000001d0;ooo`0E0?ooo`030=nLB00004P0W=oo00T0
oooo00<0gia80000B02Lgol0``3oool00?l0ooooo`3oool/0?ooo`00o`3ooooo0?ooob`0oooo003o
0?ooool0oooo;03oool00?l0ooooo`3oool/0?ooo`00o`3ooooo0?ooob`0oooo003o0?ooool0oooo
;03oool00?l0ooooo`3oool/0?ooo`00o`3ooooo0?ooob`0oooo003o0?ooool0oooo;03oool00?l0
ooooo`3oool/0?ooo`00K`3oool01P2oM000000000000000000004RL0=oooeP0oooo00D0omnL04P0
00000000001d0;ooo`030?ooo`060?noM0000000M;oo0?ooo`3ooml0W4P00`0000000`1d_ol0oooo
0?ooo`1A0?ooo`050?oOW018000000000000M02oool0F03oool01@3ogi`0B0000000000007@0_ooo
00<0oooo00H0okmd0000001d_ol0oooo0?oog`2LB0030000000307Boo`3oool0oooo05L0oooo00H0
_g@000000000000000000018W03OoolZ0?ooo`00L03oool00`3ogi`0B00007Boo`1I0?ooo`070?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?noM0000000M;oo0?ooo`3ooml0W4Qd0;oo
o`020?ooo`030;md001d_ol0oooo0500oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo05H0
oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo0080oooo00L0okmd0000001d_ol0oooo0?oo
g`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool0F03oool00`3ogi`0B00007Boo`0/0?ooo`00L03o
ool00`3ogi`0B00007Boo`1I0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool02P3o
ool00`3ooml0W4Qd0;ooo`1@0?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0E@3o
ool00`2oM00004RL0=ooo`020?ooo`030;md0000M;l0oooo00X0oooo00<0oooO09a8M02oool0F03o
ool00`3ogi`0B00007Boo`0/0?ooo`00L03oool00`3ogi`0B00007Boo`1H0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool02`3oool00`2oM000M;oo0?ooo`1>0?ooo`030?oo_`1d0000
B9cO00<0oooo00<0gia80000M02oool0E03oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0
_ooo00/0oooo00<0_g@007Boo`3oool0E`3oool00`3ogi`0B00007Boo`0/0?ooo`00H@3oool00`3o
oml0W4P0000000070000000304RLg`3oool0oooo0080oooo00<0omnL04P0001d_ol0B`3oool00`3o
oml0W4P0000000070000000604RLg`3oool0oooo0?oOW0180000M;oo0`3oool00`3o_g@0001809cO
o`0:0?ooo`030?oog`2LB000B9cO04l0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cO
o`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02P3oool00`3ooml0W4P004RL
g`1H0?ooo`030?oOW0180000M;oo02`0oooo001`0?ooo`030?oOW0180000M;oo05P0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`0:0?ooo`030=nLB0000000M;oo04l0oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`1D0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd
0000B02Lgol02P3oool00`3OW4P0000007Boo`1H0?ooo`030?oOW0180000M;oo02`0oooo001`0?oo
o`030?oOW0180000M;oo05P0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0001809cOo`060?oo
o`070=oOg`3oool0oono07@000000000001d0;ooo`1?0?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol0E03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00L0
gmoO0?ooo`3ookl0M0000000000007@0_ooo05P0oooo00<0omnL04P0001d_ol0;03oool00700oooo
00<0omnL04P0001d_ol0F03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo
00H0gia800000000000000000000B02Lgom@0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol0E03oool00`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00H0oooo00H0gia80000
0000000000000000B02LgomI0?ooo`030?oOW0180000M;oo02`0oooo001`0?ooo`030?oOW0180000
M;oo05P0oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`060?ooo`040?oo_`1d0000
000004RLge80oooo00<0oono07@00018W=l00`3oool00`3OW4P0001d0;ooo`1D0?ooo`030?oo_`1d
0000B9cO00<0oooo00<0gia80000M02oool01P3oool0103ookl0M00000000018W=mK0?ooo`030?oO
W0180000M;oo02`0oooo001`0?ooo`030?oOW0180000M;oo05T0oooo00<0gia80018W03Oool00P3o
ool00`2oM00007Bo0?ooo`070?ooo`030;md001d_ol0oooo05<0oooo00<0gia80018W03Oool00P3o
ool00`2oM00007Bo0?ooo`1E0?ooo`030=nLB000B9`0gooo0080oooo00<0_g@0001d_`3oool01`3o
ool00`2oM000M;oo0?ooo`1K0?ooo`030?oOW0180000M;oo02`0oooo001^0?ooo`050?oog`2LB000
00000000001d_ol0F@3oool01`3ookl0M00007Boo`3oool0okmd0000M02oool0203oool01P3ogi`0
B000000000000000001809cOoe00oooo00L0oono07@0001d_ol0oooo0?noM00007@0_ooo05H0oooo
00L0oono07@0001d_ol0oooo0?noM00007@0_ooo00P0oooo00H0omnL04P00000000000000000B02L
gomF0?ooo`050?oog`2LB00000000000001d_ol0;03oool00700oooo00<0okmd0000001d_ol0FP3o
ool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@0000000000000000004RLg`1A0?ooo`05
0?oo_`1d000000000000M02oool0F03oool01@3ookl0M0000000000007@0_ooo00X0oooo00D0_g@0
000000000000000004RLg`1H0?ooo`030?noM0000000M;oo02`0oooo003o0?ooool0oooo;03oool0
0?l0ooooo`3oool/0?ooo`00o`3ooooo0?ooob`0oooo003o0?ooool0oooo;03oool00?l0ooooo`3o
ool/0?ooo`00o`3ooooo0?ooob`0oooo003o0?ooool0oooo;03oool00?l0ooooo`3oool/0?ooo`00
o`3ooooo0?ooob`0oooo003o0?ooool0oooo;03oool00680ooooo`00002W00000280oooo001R0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00h0oooo00@000000?ooo`3oool00000503oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00X0oooo0P0000050?oo
o`030000003oool0oooo00h0oooo00@000000?ooo`3oool000003P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo000001@0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0P0?ooo`00HP3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0640oooo00<00000
0?ooo`3oool0H@3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0P
0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo0640oooo00<000000?oo
o`3oool07@3oool00`000000oooo0000000F0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000060?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0P0?oo
o`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo0640oooo00<000000?ooo`3o
ool0703oool010000000oooo0?ooo`00000E0?ooo`8000004@3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool2000000H0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0803oool00680oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo0?ooo`1Q0?ooo`03
0000003oool0oooo01X0oooo0P0000030?ooo`030000003oool0oooo01<0oooo0P00000@0?ooo`80
00003@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo0P0000060?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo06/0oooo00<0001o0?ooo`3oool0
H@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
4P3oool00`000000oooo0000000@0?ooo`800000303oool2000000X0oooo00<000000?ooo`3oool0
1P3oool2000000H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3o
ool2000006`0oooo00<0001o0?ooo`3oool0H@3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool04P3oool00`000000oooo0000000@0?ooo`800000303o
ool2000000T0oooo0P0000080?ooo`8000001P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`180?ooo`<000008P3o
ool00680oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo0?ooo`1Q0?ooo`030000003oool0
oooo01H0oooo0P0000070?ooo`030000003oool0oooo0140oooo00@000000?ooo`3oool000003`3o
ool00`000000oooo0000000<0?ooo`8000002@3oool2000000P0oooo0P0000050?ooo`8000001@3o
ool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`800000C03oool00`000000oooo
0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo0640oooo00<0
00000?ooo`3oool0503oool2000000T0oooo00<000000?ooo`3oool0403oool01@000000oooo0?oo
o`3oool0000000l0oooo00<000000?ooo`000000303oool2000000T0oooo0P0000080?ooo`800000
1@3oool2000000D0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool2000004`0
oooo00<000000?ooo`3oool0803oool00680oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo
0?ooo`1Q0?ooo`030000003oool0oooo0140oooo0`00000;0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0<0?ooo`040000003oool0oooo000000/0
oooo00<000000?ooo`0000002@3oool2000000L0oooo00<000000?ooo`0000001@3oool2000000D0
oooo00<000000?ooo`3oool00P3oool2000000<0oooo0P00001<0?ooo`030000003oool0oooo0200
oooo001R0?ooo`030000003oool0oooo06/0oooo00<0001o0?ooo`3oool0H@3oool00`000000oooo
0?ooo`0?0?ooo`8000003P3oool00`000000oooo0?ooo`0=0?ooo`8000001@3oool00`000000oooo
0?ooo`0;0?ooo`050000003oool0oooo0?ooo`0000002P3oool010000000oooo0?ooo`0000090?oo
o`8000001`3oool00`000000oooo000000050?ooo`800000103oool2000000@0oooo0P0000030?oo
o`800000C03oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`03
0000O`3oool0oooo0640oooo00<000000?ooo`3oool0303oool300000100oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00D000000?ooo`3o
ool0oooo000000090?ooo`050000003oool0oooo0?ooo`000000203oool00`000000oooo00000007
0?ooo`030000003oool0000000D0oooo0P0000040?ooo`800000103oool2000000<0oooo0P00001<
0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo06/0oooo00<0001o0?oo
o`3oool0H@3oool00`000000oooo0?ooo`080?ooo`@000004`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00L0oooo00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0
000000H0oooo00@000000?ooo`3oool000001@3oool2000000@0oooo0P0000040?ooo`8000000`3o
ool2000004`0oooo00<000000?ooo`3oool0803oool00680oooo0P00001/0?ooo`030000O`3oool0
oooo0640oooo00<000000?ooo`3oool0103oool4000001L0oooo00<000000?ooo`3oool02@3oool2
000000T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`060?ooo`030000003oool0
000000H0oooo00@000000?ooo`3oool00000103oool00`000000oooo000000040?ooo`800000103o
ool2000000<0oooo0P00001:0?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0J`3oool0
0`0007l0oooo0?ooo`1Q0?ooo`030000003oool0oooo00@000006`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`05
0?ooo`040000003oool0oooo000000H0oooo00@000000?ooo`3oool00000103oool00`000000oooo
000000040?ooo`800000103oool2000000<0oooo0P00001<0?ooo`030000003oool0oooo0200oooo
001R0?ooo`030000003oool0oooo06/0oooo00<0001o0?ooo`3oool0H03oool4000001l0oooo00<0
00000?ooo`3oool01P3oool2000000`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`050?ooo`040000003oool0oooo000000H0oooo00@000000?ooo`3oool00000103oool00`000000
oooo000000040?ooo`8000000`3oool00`000000oooo000000030?ooo`800000C03oool00`000000
oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo05`0oooo
100000000`3oool000000?ooo`0P0?ooo`030000003oool0oooo00@0oooo0P00000>0?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0
oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0000000<0oooo00<000000?oo
o`0000000`3oool00`000000oooo000000030?ooo`800000C03oool00`000000oooo0?ooo`0P0?oo
o`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo05P0oooo100000050?ooo`03
0000003oool0oooo01l0oooo00<000000?ooo`3oool00P3oool200000100oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001@3oool01@00
0000oooo0?ooo`3oool0000000<0oooo00@000000?ooo`3oool000000`3oool00`000000oooo0000
00030?ooo`030000003oool000000080oooo00<000000?ooo`000000C03oool00`000000oooo0?oo
o`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo05@0oooo10000009
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool00P00000B0?ooo`030000003oool0
oooo00<0oooo0P00000:0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000P3oool01`000000oooo0?oo
o`3oool000000?ooo`0000000`3oool00`000000oooo000000020?ooo`030000003oool0000004`0
oooo00<000000?ooo`3oool0803oool00680oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo
0?ooo`1@0?ooo`@000003@3oool00`000000oooo0?ooo`0O0?ooo`<00000503oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000020?ooo`04
0000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool01P000000oooo0?ooo`00
0000oooo000004`0oooo00<000000?ooo`3oool0803oool00680oooo0P00001/0?ooo`030000O`3o
ool0oooo04X0oooo1P00000A0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`050000003o
ool0oooo0?ooo`0000003@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0?ooo`00
00001@3oool010000000oooo0?ooo`0000030?ooo`040000003oool0oooo00000080oooo00@00000
0?ooo`3oool000000P3oool01P000000oooo0?ooo`000000oooo000004X0oooo0`00000R0?ooo`00
HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo04@0oooo1P00000G0?ooo`030000
003oool0oooo01`0oooo0P0000000`3oool000000?ooo`0E0?ooo`040000003oool0oooo000000h0
oooo00D000000?ooo`3oool0oooo000000090?ooo`050000003oool0oooo0?ooo`0000001P3oool0
1@000000oooo0?ooo`3oool0000000D0oooo00@000000?ooo`3oool000000`3oool010000000oooo
0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00H000000?ooo`3oool000000?ooo`00
001<0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo06/0oooo00<0001o
0?ooo`3oool0@03oool4000001d0oooo00<000000?ooo`3oool06@3oool3000000<0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0000000?0?ooo`040000003oool0oooo000000X0oooo00@0
00000?ooo`3oool000001`3oool010000000oooo0?ooo`0000060?ooo`040000003oool0oooo0000
00<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003o
ool000000080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool0
0`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo03T0oooo1`00000Q0?ooo`030000003oool0
oooo01L0oooo0P0000060?ooo`030000003oool0oooo01@0oooo0P00000@0?ooo`030000003oool0
000000/0oooo00@000000?ooo`3oool000001`3oool010000000oooo0?ooo`0000060?ooo`030000
003oool0000000@0oooo00<000000?ooo`0000000`3oool00`000000oooo000000030?ooo`030000
003oool000000080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3o
ool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo03<0oooo1P00000X0?ooo`030000003o
ool0oooo01@0oooo0`0000080?ooo`030000003oool0oooo01<0oooo0P00000A0?ooo`800000303o
ool00`000000oooo000000080?ooo`030000003oool0000000L0oooo00<000000?ooo`000000103o
ool00`000000oooo000000030?ooo`030000003oool0000000<0oooo00<000000?ooo`0000000P3o
ool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0
oooo06/0oooo00<0001o0?ooo`3oool0;`3oool4000002h0oooo00<000000?ooo`3oool04@3oool3
000000/0oooo00<000000?ooo`3oool04P3oool00`000000oooo0000000A0?ooo`800000303oool2
000000T0oooo00<000000?ooo`0000001`3oool2000000D0oooo00<000000?ooo`0000000`3oool0
0`000000oooo000000030?ooo`030000003oool000000080oooo00<000000?ooo`3oool0BP3oool0
0`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo
02T0oooo1P00000b0?ooo`030000003oool0oooo00l0oooo0P00000>0?ooo`030000003oool0oooo
0140oooo00@000000?ooo`3oool000004@3oool00`000000oooo0?ooo`0;0?ooo`8000002@3oool2
000000P0oooo0P0000050?ooo`800000103oool2000000@0oooo00<000000?ooo`0000000P3oool0
0`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`800000K03oool00`00
07l0oooo0?ooo`0Q0?ooo`P00000>03oool00`000000oooo0?ooo`0<0?ooo`<00000403oool00`00
0000oooo0?ooo`0?0?ooo`8000000`3oool00`000000oooo0?ooo`0>0?ooo`8000003@3oool00`00
0000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`040?ooo`800000103oool20000
00@0oooo0P0000030?ooo`030000003oool0oooo04P0oooo0`00000R0?ooo`00HP3oool00`000000
oooo0?ooo`1[0?ooo`030000O`3oool0oooo01T0oooo200000100?ooo`030000003oool0oooo00T0
oooo0`00000C0?ooo`030000003oool0oooo00d0oooo0P0000050?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`000000303oool2000000X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo0P0000040?ooo`8000000`3oool00`000000
oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo06/0oooo
00<0001o0?ooo`3oool04P3oool7000004P0oooo00<000000?ooo`3oool01P3oool3000001H0oooo
00<000000?ooo`3oool02`3oool2000000L0oooo00<000000?ooo`3oool02`3oool200000080oooo
00<000000?ooo`3oool02P3oool2000000X0oooo00<000000?ooo`3oool01P3oool2000000H0oooo
00<000000?ooo`3oool00`3oool2000000@0oooo0P0000030?ooo`030000003oool0oooo04X0oooo
00<000000?ooo`3oool0803oool004L0oooo00<0_g@0000000000000100000001018W=l0oooo0?oo
g`2LB0030000000307Boo`3oool0oooo00X0oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo
0?ooo`0>0?ooo`@00000C`3oool00`000000oooo0?ooo`030?ooo`<000006@3oool00`000000oooo
0?ooo`090?ooo`8000002@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool02@3oool00`000000oooo000000090?ooo`800000203oool2000000H0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`8000000`3oool00`000000oooo
0?ooo`1:0?ooo`030000003oool0oooo0200oooo00180?ooo`030=nLB0000000000000<0000000D0
04RL0=ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00X0oooo00<000000?ooo`3oool0
J`3oool00`0007l0oooo0?ooo`080?ooo`H00000D`3oool00`000000oooo0?ooo`03000001`0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000000P0
oooo00<000000?ooo`0000001`3oool00`000000oooo000000050?ooo`8000001@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`0P0?ooo`00B@3oool00`3OW4P0B9cO0?ooo`020?ooo`030=nLM02Lgol0oooo00@0oooo
00<0oooO09a8M02oool02P3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo00P00000
F@3oool3000001l0oooo00<000000?ooo`3oool01P3oool2000000`0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00L0oooo00D000000?ooo`3oool0
oooo000000080?ooo`030000003oool0000000L0oooo00<000000?ooo`0000001@3oool2000000D0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0
oooo00<000000?ooo`3oool0803oool004X0oooo00<0gia807Boo`3oool02@3oool00`2oM000M;oo
0?ooo`090?ooo`030000003oool0oooo06H0oooo2000001O0?ooo`<000008@3oool00`000000oooo
0?ooo`040?ooo`8000003P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo
00@000000?ooo`3oool000001P3oool010000000oooo0?ooo`0000040?ooo`030000003oool00000
00@0oooo0P0000040?ooo`800000103oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo
0200oooo001;0?ooo`030;mdB02Lgol0oooo00L0oooo00<0oooO09a80018W=l02P3oool00`000000
oooo0?ooo`1N0?ooo`P000001@3oool00`0007l0oooo0?ooo`1L0?ooo`<000000P3oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo0080oooo0P00000@0?ooo`030000003oool0oooo00D0
oooo0P0000080?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000001P3oool010000000oooo0?ooo`000004
0?ooo`030000003oool0000000@0oooo0P0000040?ooo`800000103oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo0200oooo001;0?ooo`030?oog`2LB7@0_ooo00L0oooo00<0gia80000
001d_ol02P3oool3000005D0001o2@00001X0000O`@000001@0007l00`000000001o0000O`0O0000
O`030000000007l0001o008000004P0007l00`000000001o0000O`040000O`030000000007l0001o
00P0001o00<000000000O`0007l0100007l00`000000001o0000O`040000O`030000000007l0001o
00@0001o00D000000000O`0007l0001o000000050000O`050000000007l0001o0000O`0000001000
07l00`000000001o000000040000O`8000000`0007l00`000000001o000000040000O`0300000000
07l0001o04L0001o1000000R0?ooo`00C03oool00`3o_g@007Bo0?ooo`020?ooo`070=oOg`3oool0
oono07@000000000001d0;ooo`0:0?ooo`030000003oool0oooo04d0oooo2000000F0?ooo`030000
O`3oool0oooo05@0oooo100000090?ooo`030000003oool0oooo01l0oooo0`00000D0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00D000000?ooo`3oool0oooo000000030?ooo`040000003oool0oooo0000
00<0oooo00<000000?ooo`0000000`3oool00`000000oooo000000040?ooo`030000003oool0oooo
04X0oooo00<000000?ooo`3oool0803oool004`0oooo00<0oono07@0B02Lgol00P3oool01P3OW4P0
0000000000000000001809cOo`/0oooo00<000000?ooo`3oool0A@3oool8000001h0oooo00<0001o
0?ooo`3oool0DP3oool2000000d0oooo00<000000?ooo`3oool07P3oool2000001H0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000P3o
ool01`000000oooo0?ooo`3oool000000?ooo`0000000`3oool00`000000oooo000000040?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool004`0oooo00<0oono07@00018W=l0
0P3oool0103ookl0M00000000018W=l=0?ooo`030000003oool0oooo03d0oooo2000000V0?ooo`03
0000O`3oool0oooo04l0oooo0`00000?0?ooo`030000003oool0oooo01`0oooo0P0000000`3oool0
00000?ooo`0E0?ooo`030000003oool0oooo008000003@3oool01@000000oooo0?ooo`3oool00000
00T0oooo00D000000?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000001@3o
ool010000000oooo0?ooo`0000030?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?oo
o`3oool0803oool004L0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`030?ooo`03
0;md001d_ol0oooo00d0oooo00<000000?ooo`3oool0=@3oool8000002h0oooo00<0001o0?ooo`3o
ool0C03oool300000180oooo00<000000?ooo`3oool06P3oool2000000<0oooo00<000000?ooo`3o
ool0503oool3000000l0oooo00@000000?ooo`3oool000002P3oool01@000000oooo0?ooo`3oool0
000000H0oooo00D000000?ooo`3oool0oooo000000050?ooo`040000003oool0oooo000000<0oooo
00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00A`3oool0203ooml0
W4Q809cOo`3oool0oono07@0000007@0_ooo0`3oool01P3ogi`0B000000000000000001809cOo`X0
oooo00<000000?ooo`3oool0;03oool9000003H0oooo00<0001o0?ooo`3oool0B03oool4000001D0
oooo00<000000?ooo`3oool0603oool2000000D0oooo00<000000?ooo`3oool04`3oool200000140
oooo00<000000?ooo`0000002`3oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo
000000H0oooo00<000000?ooo`000000103oool010000000oooo0?ooo`0000020?ooo`040000003o
ool0oooo00000080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0200oooo00180?ooo`060?oog`2LB00000000000000007@0_ooo1@3oool01@2oM000
0000000000000000B9cO00X0oooo00<000000?ooo`3oool0903oool8000003l0oooo00<0001o0?oo
o`3oool0A03oool4000001T0oooo00<000000?ooo`3oool05P3oool2000000L0oooo00<000000?oo
o`3oool04@3oool2000000030?ooo`000000oooo0100oooo0P00000<0?ooo`030000003oool00000
00P0oooo00@000000?ooo`3oool000001P3oool00`000000oooo000000040?ooo`030000003oool0
000000<0oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo0P00000M0?ooo`P00000A`3oool00`00
07l0oooo0?ooo`100?ooo`@000007@3oool00`000000oooo0?ooo`0D0?ooo`8000002@3oool00`00
0000oooo0?ooo`0@0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0;
0?ooo`8000002@3oool00`000000oooo000000070?ooo`8000001@3oool00`000000oooo00000003
0?ooo`030000003oool0000000<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`18
0?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0503oool8000004l0oooo00<0001o0?oo
o`3oool0>`3oool500000240oooo00<000000?ooo`3oool04@3oool3000000/0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00h0oooo0P00000=0?oo
o`030000003oool0oooo00P0oooo0P0000080?ooo`8000001@3oool00`000000oooo000000030?oo
o`030000003oool0000000<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo00`0oooo2000001G0?ooo`03
0000O`3oool0oooo03L0oooo1000000V0?ooo`030000003oool0oooo00l0oooo0P00000>0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0000000`0oooo0P00000:0?ooo`800000203oool00`000000oooo0?ooo`040?ooo`80
0000103oool00`000000oooo000000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`070?ooo`D00000
G`3oool00`0007l0oooo0?ooo`0c0?ooo`@00000:P3oool00`000000oooo0?ooo`0=0?ooo`800000
403oool00`000000oooo0?ooo`0<0?ooo`8000001P3oool00`000000oooo0?ooo`0<0?ooo`040000
003oool0oooo000000`0oooo0P00000:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0103oool2000000@0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000
O`3oool0oooo02l0oooo1000000^0?ooo`030000003oool0oooo00/0oooo0P00000B0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0;0?ooo`050000
003oool0oooo0?ooo`0000002`3oool00`000000oooo000000090?ooo`800000203oool2000000H0
oooo00<000000?ooo`3oool00`3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo06/0oooo
00<0001o0?ooo`3oool0:`3oool400000380oooo00<000000?ooo`3oool0203oool3000001@0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000000T0
oooo0P0000080?ooo`8000001P3oool00`000000oooo0?ooo`030?ooo`800000103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo
00<000000?ooo`3oool0J`3oool00`0007l0oooo0?ooo`0W0?ooo`@00000=P3oool00`000000oooo
0?ooo`060?ooo`8000005`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00L0oooo
00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0000000L0oooo00<000000?ooo`00
00001@3oool2000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo0P00001/0?ooo`030000
O`3oool0oooo02<0oooo1000000j0?ooo`030000003oool0oooo00<0oooo0`00000I0?ooo`030000
003oool0oooo00L0oooo0P00000;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
1@3oool010000000oooo0?ooo`0000070?ooo`030000003oool0000000D0oooo0P0000050?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`180?ooo`<0
00008P3oool00680oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo0?ooo`0O0?ooo`@00000
?P3oool00`000000oooo0?ooo`03000001`0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool010000000oooo
0?ooo`0000060?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`0000001@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`00
0000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo01/0
oooo100000120?ooo`<000007`3oool00`000000oooo0?ooo`040?ooo`8000003P3oool00`000000
oooo0?ooo`050?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00D000000?oo
o`3oool0oooo000000040?ooo`030000003oool0000000@0oooo0P0000050?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`00
0000oooo0?ooo`1[0?ooo`030000O`3oool0oooo01L0oooo100000140?ooo`<000008@3oool00`00
0000oooo0?ooo`020?ooo`800000403oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo
0?ooo`0000000`3oool010000000oooo0?ooo`0000040?ooo`8000001@3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo00<00000
0?ooo`3oool0J`3oool00`0007l0oooo0?ooo`0B0?ooo`D00000AP3oool200000080oooo00<00000
0?ooo`3oool07`3oool01@000000oooo0?ooo`3oool000000180oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000P3oool01`000000oooo
0?ooo`3oool000000?ooo`0000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
04X0oooo00<000000?ooo`3oool0803oool00680oooo00<000000?ooo`3oool0J`3oool00`0007l0
oooo0?ooo`0>0?ooo`@00000B@3oool2000000@0oooo00<000000?ooo`3oool07`3oool010000000
oooo00000000000C0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool01000
0000oooo0?ooo`0000030?ooo`070000003oool0oooo0?ooo`000000oooo000000050?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3o
ool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo00X0oooo1000001;0?ooo`8000001P3o
ool00`000000oooo0?ooo`0O0?ooo`8000005@3oool01@000000oooo0?ooo`3oool0000000d0oooo
00D000000?ooo`3oool0oooo000000090?ooo`050000003oool0oooo0?ooo`0000001P3oool01@00
0000oooo0?ooo`3oool0000000D0oooo00@000000?ooo`3oool000000`3oool01`000000oooo0?oo
o`3oool000000?ooo`0000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0
oooo00<000000?ooo`3oool0803oool00680oooo0P00001/0?ooo`030000O`3oool0oooo00H0oooo
1000001=0?ooo`800000203oool00`000000oooo0?ooo`0M0?ooo`<000005P3oool010000000oooo
00000000000>0?ooo`040000003oool0oooo000000X0oooo00D000000?ooo`3oool0oooo00000006
0?ooo`050000003oool0oooo0?ooo`0000001@3oool00`000000oooo000000040?ooo`040000003o
ool0oooo00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo04P0oooo0`00000R0?ooo`00HP3oool00`000000oooo0?ooo`1[0?ooo`030000O`3o
ool0oooo0080oooo1000001>0?ooo`<000002P3oool00`000000oooo0?ooo`0K0?ooo`8000000P3o
ool00`000000oooo0?ooo`0D0?ooo`800000403oool00`000000oooo0000000;0?ooo`040000003o
ool0oooo000000L0oooo00@000000?ooo`3oool000001P3oool00`000000oooo000000040?ooo`04
0000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo00<000000?ooo`3o
ool0J`3oool00`0007l000000000000200000500oooo0P00000=0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00l0
oooo0P00000<0?ooo`030000003oool0000000P0oooo00<000000?ooo`0000001`3oool2000000D0
oooo00<000000?ooo`0000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo
0?ooo`1X0?ooo`@00000D@3oool3000000l0oooo00<000000?ooo`3oool0603oool2000000D0oooo
00<000000?ooo`3oool04P3oool300000140oooo00<000000?ooo`3oool02`3oool2000000T0oooo
00<000000?ooo`0000001`3oool2000000D0oooo00<000000?ooo`0000000`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo
0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`1T0?ooo`@000000`3oool00`0007l0oooo0?oo
o`1<0?ooo`<000004P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool04@3oool010000000oooo0?ooo`00000@0?ooo`8000003@3oool2000000T0oooo
0P0000080?ooo`8000001@3oool00`000000oooo000000030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
001R0?ooo`030000003oool0oooo0600oooo100000070?ooo`030000O`3oool0oooo04T0oooo0`00
000E0?ooo`030000003oool0oooo01D0oooo0P0000080?ooo`030000003oool0oooo0100oooo00D0
00000?ooo`3oool0oooo0000000@0?ooo`8000003@3oool00`000000oooo0?ooo`080?ooo`800000
203oool00`000000oooo0?ooo`040?ooo`800000103oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3o
ool00`000000oooo0?ooo`1L0?ooo`@000002`3oool00`0007l0oooo0?ooo`160?ooo`<00000603o
ool00`000000oooo0?ooo`0C0?ooo`8000002P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0000000<0?ooo`8000002P3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo0P0000040?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003o
ool0oooo0200oooo001R0?ooo`800000F03oool5000000l0oooo00<0001o0?ooo`3oool0@`3oool3
000001/0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`040000003oool0
oooo000000/0oooo00<000000?ooo`0000002P3oool00`000000oooo0?ooo`060?ooo`8000001P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0D`3oool4
000001@0oooo00<0001o0?ooo`3oool0@@3oool2000001h0oooo00<000000?ooo`3oool0403oool2
000000d0oooo00<000000?ooo`3oool0303oool2000000H0oooo00<000000?ooo`3oool02`3oool0
1@000000oooo0?ooo`3oool0000000/0oooo00<000000?ooo`0000002@3oool2000000P0oooo0P00
00060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo00<00000
0?ooo`3oool0C@3oool6000001P0oooo00<0001o0?ooo`3oool0?P3oool300000200oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0203oool010000000oooo0?ooo`0000090?ooo`8000001`3oool00`000000oooo0000
00060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool004L0oooo00<0oooO
09a8000000000P0000000`1d_ol0oooo0?ooo`030?ooo`050?oOW018000000000000M02oool02`3o
ool00`000000oooo0?ooo`170?ooo`H000007P3oool00`0007l0oooo0?ooo`0l0?ooo`8000008`3o
ool00`000000oooo0?ooo`0=0?ooo`800000403oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00L0oooo00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0000000L0oooo
00<000000?ooo`0000001@3oool2000000D0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool004L0oooo00<0
oooO09a8M02oool00P3oool00`2oM000M;oo0?ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0
001d0;ooo`0:0?ooo`030000003oool0oooo04<0oooo1000000T0?ooo`030000O`3oool0oooo03X0
oooo0P00000U0?ooo`030000003oool0oooo00/0oooo0P00000B0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool01`3oool01@000000oooo0?ooo`3oool0000000P0oooo00<000000?oo
o`0000001P3oool010000000oooo0?ooo`0000050?ooo`8000001@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P
0?ooo`00C03oool00`3ooml0W4Qd0;ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0_g@0001d
_`3oool02@3oool00`000000oooo0?ooo`0o0?ooo`@00000:03oool00`0007l0oooo0?ooo`0h0?oo
o`8000009`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool01@3oool010000000oooo0?ooo`0000060?ooo`040000003oool0oooo000000D0oooo0P00
00050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo0200oooo001=0?ooo`060;md001d_ol0oooo0?oo_`1d0000B9cO
0`3oool00`3OW4P0001d0;ooo`090?ooo`030000003oool0oooo03/0oooo1000000/0?ooo`030000
O`3oool0oooo03H0oooo0P00000Y0?ooo`030000003oool0oooo00P0oooo0P00000E0?ooo`030000
003oool0oooo00H0oooo0P00000<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
1@3oool010000000oooo0?ooo`0000060?ooo`040000003oool0oooo000000@0oooo00<000000?oo
o`0000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00C03oool01`3ooml0W4P004RLg`3oool0omnL
04P0001d_ol00`3oool00`3o_g@0001809cOo`090?ooo`030000003oool0oooo03L0oooo1000000`
0?ooo`030000O`3oool0oooo03<0oooo0`00000[0?ooo`030000003oool0oooo00H0oooo0P00000G
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo000000050?ooo`050000003o
ool0oooo0?ooo`000000103oool00`000000oooo000000050?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
001<0?ooo`070=nLB0000000M;oo0?ooo`3ogi`0B00007Boo`030?ooo`030?noM00004P0W=oo00T0
oooo0`00000c0000O`@00000J00007l2000002h0001o00<000000000O`0007l0100007l2000001T0
001o00<000000000O`0007l0100007l00`000000001o0000O`0=0000O`030000000007l0001o00D0
001o00<000000000O`0007l01`0007l00`000000001o0000O`040000O`030000000007l0001o00@0
001o00<000000000O`0007l0100007l01@000000001o0000O`0007l0000000D0001o00D000000000
O`0007l0001o000000030000O`040000000007l0001o000000D0001o00<000000000O`0007l00`00
07l00`000000001o0000O`020000O`030000000007l0001o04L0001o1000000R0?ooo`00B03oool0
2`3Ogml0oooo0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0oooo00<0okmd0000
B02Lgol02@3oool00`000000oooo0?ooo`0^0?ooo`D00000>03oool00`0007l0oooo0?ooo`0^0?oo
o`<00000<03oool00`000000oooo0?ooo`020?ooo`8000006`3oool00`000000oooo0?ooo`020?oo
o`800000403oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00D000000?oo
o`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool0
04P0oooo00H0gia800000000000000000000B02Lgol20?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd0000B02Lgol02@3oool00`000000oooo0?ooo`0Z0?ooo`@00000?@3oool00`0007l0oooo0?oo
o`0[0?ooo`<00000<`3oool01@000000oooo0?ooo`3oool0000001d0oooo00D000000?ooo`3oool0
oooo0000000B0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool01@000000
oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?oo
o`00B03oool0103ookl0M00000000018W=l40?ooo`030?oo_`1d0000B9cO00<0oooo00<0gia80000
M02oool02@3oool00`000000oooo0?ooo`0V0?ooo`@00000@@3oool00`0007l0oooo0?ooo`0Y0?oo
o`800000=P3oool010000000oooo00000000000N0?ooo`040000003oool0oooo000001<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00D0
00000?ooo`3oool0oooo000000050?ooo`040000003oool0oooo000000<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00B@3oool00`2oM000M;oo0?ooo`050?ooo`030=nL
B000B9`0gooo0080oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`0R0?ooo`@00000
A@3oool00`0007l0oooo0?ooo`0V0?ooo`<00000>03oool200000200oooo00<000000?ooo`000000
503oool01@000000oooo0?ooo`3oool0000000d0oooo00D000000?ooo`3oool0oooo000000090?oo
o`050000003oool0oooo0?ooo`0000001P3oool01@000000oooo0?ooo`3oool0000000D0oooo00@0
00000?ooo`3oool000000`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
00190?ooo`060?oOW018000000000000000004P0W=oo0P3oool01`3ookl0M00007Boo`3oool0okmd
0000M02oool02P3oool00`000000oooo0?ooo`0N0?ooo`@00000B@3oool00`0007l0oooo0?ooo`0T
0?ooo`800000>P3oool200000240oooo0P00000E0?ooo`040000003oool0oooo000000h0oooo00D0
00000?ooo`3oool0oooo000000090?ooo`050000003oool0oooo0?ooo`0000001P3oool010000000
oooo0?ooo`0000060?ooo`040000003oool0oooo000000<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool0
0`000000oooo0?ooo`0P0?ooo`00BP3oool01@2oM0000000000000000000B9cO00<0oooo00D0oono
07@000000000001d0;ooo`0;0?ooo`030000003oool0oooo01X0oooo1000001=0?ooo`030000O`3o
ool0oooo0280oooo0P00000j0?ooo`80000000<0oooo0000003oool0803oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0000000l0oooo00@000000?ooo`3oool000002P3oool010000000oooo
0?ooo`0000070?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`000000103oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`8000005`3oool400000540
oooo00<0001o0?ooo`3oool0803oool2000003/0oooo00D000000?ooo`3oool0oooo0000000O0?oo
o`<000005P3oool200000100oooo00<000000?ooo`0000002`3oool010000000oooo0?ooo`000007
0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`000000103oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`180?ooo`<000008P3oool002H0oooo00<0gibL0;ooo`3oool02@3oool00`3OW9`0_ooo0?ooo`0]
0?ooo`030000003oool0oooo0180oooo1000001E0?ooo`030000O`3oool0oooo01h0oooo0P00000k
0?ooo`800000103oool00`000000oooo0?ooo`0L0?ooo`040000003oool0oooo000001H0oooo0P00
000@0?ooo`030000003oool0000000/0oooo00<000000?ooo`000000203oool00`000000oooo0000
00070?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000V
0?ooo`030=nLB01d_ol0oooo00T0oooo00<0_gAd0;ooo`3oool0;@3oool00`000000oooo0?ooo`0>
0?ooo`@00000F@3oool00`0007l0oooo0?ooo`0K0?ooo`<00000?03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo01/0oooo00D000000?ooo`3oool0oooo0000000F0?ooo`030000003o
ool0oooo00l0oooo0P00000<0?ooo`8000002@3oool00`000000oooo000000070?ooo`8000001@3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000V0?ooo`030=nLB0000000
000000T0000000<0001d0;ooo`3oool0;@3oool00`000000oooo0?ooo`0:0?ooo`@00000G@3oool0
0`0007l0oooo0?ooo`0I0?ooo`800000?@3oool2000000L0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo01<0oooo0P00000A0?ooo`800000303oool2
000000T0oooo0P0000080?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo0200oooo000V0?ooo`050=nLB000000000000018W03Oool01`3oool00`2oM7@0_ooo0?ooo`0]
0?ooo`030000003oool0oooo00L0oooo0`00001Q0?ooo`030000O`3oool0oooo01H0oooo0`00000m
0?ooo`8000002@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0000000A0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo0200oooo000W0?ooo`030?oOW018000000000080000000<0B9cO0?oo
o`3oool00`3oool00`3OW9`0_ooo0?ooo`0]0?ooo`030000003oool0oooo06/0oooo00<0001o0?oo
o`3oool0503oool2000003l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool010000000oooo0?ooo`00000@
0?ooo`8000003@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000Z0?oo
o`060=nLB000000000000000000007@0_ooo<P3oool00`000000oooo0?ooo`1[0?ooo`030000O`3o
ool0oooo0180oooo0P0000100?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0100oooo00D000000?ooo`3oool0oooo
0000000@0?ooo`800000303oool2000000X0oooo00<000000?ooo`3oool01P3oool2000000H0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00;03oool00`3ookl0M0000000
000200000003001d_`3oool0oooo02h0oooo0P00001/0?ooo`030000O`3oool0oooo0100oooo0P00
00100?ooo`8000003@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0403oool01@000000oooo0?ooo`3oool0000000l0oooo00<000000?ooo`000000
303oool2000000X0oooo00<000000?ooo`3oool01P3oool2000000H0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
B03oool300000280oooo000_0?ooo`050?noM000000000000000M02oool0;P3oool00`000000oooo
0?ooo`1[0?ooo`030000O`3oool0oooo00h0oooo0P0000110?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0<0?ooo`040000003oool0oooo000000/0
oooo00<000000?ooo`0000002@3oool2000000L0oooo00<000000?ooo`0000001P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000^0?ooo`040?oo_`1d00000000001d_c00
oooo00<000000?ooo`3oool0J`3oool00`0007l0oooo0?ooo`0;0?ooo`<00000@P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0303oool010000000
oooo0?ooo`00000;0?ooo`030000003oool0000000T0oooo0P0000070?ooo`030000003oool00000
00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00;03oool0103OW4P0
00000000M02ooolb0?ooo`030000003oool0oooo06/0oooo00<0001o0?ooo`3oool02@3oool20000
04<0oooo0P00000A0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02`3oool01@00
0000oooo0?ooo`3oool0000000X0oooo00@000000?ooo`3oool00000203oool00`000000oooo0000
00070?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?oo
o`0P0?ooo`00:@3oool0103ogi`0B00000000018W=l50?ooo`030=nLW02oool0oooo02d0oooo00<0
00000?ooo`3oool0J`3oool00`0007l0oooo0?ooo`070?ooo`800000A03oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool01@000000oooo0?oo
o`3oool0000000X0oooo00@000000?ooo`3oool00000203oool00`000000oooo000000060?ooo`04
0000003oool0oooo000000H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?oo
o`009`3oool0102oM00000000018W03Oool70?ooo`030;mdM02oool0oooo02d0oooo00<000000?oo
o`3oool0J`3oool00`0007l0oooo0?ooo`050?ooo`800000A@3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00L0oooo00D000000?ooo`3oool0oooo000000070?ooo`040000003oool0
oooo000000H0oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0200oooo000V0?ooo`030=nLB0000000000000T0000000<0001d0;ooo`3oool0;@3o
ool00`000000oooo0?ooo`1[0?ooo`030000O`3oool0oooo00<0oooo0P0000150?ooo`8000005@3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`0000001`3oool010000000oooo
0?ooo`0000050?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0200oooo000V0?ooo`030=nLB0000000000000T0000000<0001d0;ooo`3o
ool0;@3oool2000006`0oooo00@0001o0?ooo`3oool0oooo0P0000160?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00D0oooo00@000000?ooo`3oool000001@3oool01@000000oooo0?ooo`3oool0000000H0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0B03oool300000280oooo000V0?ooo`030=nLB01d_ol0oooo00T0oooo00<0_gAd
0;ooo`3oool0;@3oool00`000000oooo0?ooo`1[0?ooo`040000O`3oool00000000004L0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00D000000?ooo`3oool0
oooo000000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002H0oooo
00<0gibL0;ooo`3oool02@3oool00`3OW9`0_ooo0?ooo`0]0?ooo`030000003oool0oooo06/0oooo
0P0000170?ooo`8000006@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00D000000?ooo`3o
ool0oooo000000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680
oooo00<000000?ooo`3oool0J@3oool2000000030000O`3oool0oooo04D0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3o
ool0oooo000000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002`0
oooo00L0_g@00000B02Lgol0oono07@000000000B9cO02l0oooo00<000000?ooo`3oool0I`3oool2
00000080oooo00<0001o0?ooo`3oool0A03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00:`3oool01@3o
_g@000000000B02Lgol0_g@00080000000<0001d0;ooo`3oool0;@3oool00`000000oooo0?ooo`1V
0?ooo`030000003oool0oooo0080oooo00<0001o0?ooo`3oool0@P3oool2000001d0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000[0?oo
o`090;mdM02oool0oooo0?oo_`1dB9`0gooo0?oOW018B9`0gooo02h0oooo00<000000?ooo`3oool0
I03oool2000000D0oooo00<0001o0?ooo`3oool0@@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`070?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00:`3o
ool0202oM4P0W=oo0?ooo`3OW7@0W=oo0?ooo`3o_i`0W=oo;`3oool00`000000oooo0?ooo`1R0?oo
o`8000001`3oool00`0007l0oooo0?ooo`100?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000[0?ooo`03
0=nLB0000000000000@0000000<004RL0=ooo`3oool0;@3oool200000680oooo00<000000?ooo`3o
ool01`3oool00`0007l0oooo0?ooo`0n0?ooo`8000008@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
04P0oooo0`00000R0?ooo`00:`3oool00`3ookl0M00000000004000000030018W03Oool0oooo02d0
oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`080?ooo`030000O`3oool0oooo03d0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0
oooo00D000000?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000001@3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool0
0340oooo00<0_gA809cOo`3oool05P3oool00`3OW4P0B9cO0?ooo`030?ooo`030?oog`2LB0000000
0080000000<0M;oo0?ooo`3oool02P3oool00`000000oooo0?ooo`1N0?ooo`8000002`3oool00`00
07l0oooo0?ooo`0l0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0000001P3oool01@000000oooo0?ooo`3o
ool0000000D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003o
ool0oooo0200oooo000[0?ooo`030;mdM02oool0oooo00@0oooo00<0_gAd0;ooo`3oool05@3oool0
0`3ogi`0B4RL0=ooo`030?ooo`030?oog`2LB7@0_ooo0080oooo00<0_g@007Boo`3oool02P3oool0
0`000000oooo0?ooo`1M0?ooo`030000003oool0oooo00/0oooo00<0001o0?ooo`3oool0>`3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01P3oool0
1@000000oooo0?ooo`3oool0000000H0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00:`3o
ool00`2oM000B9cO0?ooo`030?ooo`030=nLB00007@0_ooo01H0oooo00<0oooO09a8B02Lgol0203o
ool00`3ooml0W4Qd0;ooo`0:0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0303o
ool00`0007l0oooo0?ooo`0j0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`060?ooo`040000003oool0oooo000000L0oooo00D000000?ooo`3oool0
oooo000000050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`0P0?ooo`00:`3oool02@2oM00000000000M02oool0oooO09a80018W=l0gibL0;ooo`0G
0?ooo`030=nLB018W=l0oooo00P0oooo00<0_g@007Boo`3oool02@3oool00`000000oooo0?ooo`1J
0?ooo`8000003`3oool00`0007l0oooo0?ooo`0j0?ooo`030000003oool0oooo02@0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00X0oooo00D00000
0?ooo`3oool0oooo000000090?ooo`040000003oool0oooo000000L0oooo00@000000?ooo`3oool0
00001P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3o
ool0803oool002/0oooo00H0_ibL0=oOW018000000000000B02LgolJ0?ooo`030?oOW018B9`0gooo
00L0oooo00<0oooO09a80018W=l02P3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo
00l0oooo00<0001o0?ooo`3oool0>@3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`050000003oool0oooo
0?ooo`0000002@3oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo000000H0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
000[0?ooo`030;mdM02ogi`0B0000080000000@0001809cOo`3OW9`0_ooo5`3oool00`3ooml0W4Q8
09cOo`070?ooo`030=nLB0000000M;oo00X0oooo0`00001G0000O`800000C@0007l00`000000001o
0000O`0V0000O`030000000007l0001o00@0001o00<000000000O`0007l0600007l01@000000001o
0000O`0007l000000180001o00<000000000O`0007l00P0007l00`000000001o0000O`0:0000O`05
0000000007l0001o0000O`0000002@0007l010000000001o0000O`0000070000O`040000000007l0
001o000000H0001o00<000000000O`0007l0100007l00`000000001o0000O`030000O`0300000000
07l0001o00<0001o00<000000000O`0007l00P0007l00`000000001o0000O`170000O`@000008P3o
ool002/0oooo00<0_g@00018W03Oool00P3oool0102oM00000000000M02ooolH0?ooo`030=nLB000
M;l0oooo0080oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00X0oooo00<000000?ooo`3o
ool0E@3oool2000001@0oooo00<0001o0?ooo`3oool0=`3oool00`000000oooo0?ooo`0W0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool06@3oool01@000000oooo0?ooo`3oool00000
0180oooo00D000000?ooo`3oool0oooo0000000=0?ooo`040000003oool0oooo000000X0oooo00<0
00000?ooo`000000203oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00:`3oool00`2oM4P0W=oo0?oo
o`030?ooo`030?noM00007@0_ooo01P0oooo00<0oono07@0M02oool00P3oool01P3OW4P000000000
00000000001809cOo`/0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0D0?ooo`03
0000O`3oool0oooo03L0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo01T0oooo00@000000?ooo`3oool000004`3oool01@000000oooo0?ooo`3oool0
000000d0oooo00@000000?ooo`3oool000002P3oool00`000000oooo000000080?ooo`040000003o
ool0oooo000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0200oooo000[0?ooo`030;nLW03Oool0oooo00@0oooo00<0gibL0;ooo`3oool04P3o
ool00`3ogml0gooo0?ooo`030?ooo`080;md001d_ol0oooo0?ooo`3ookl0M00000000018W=l=0?oo
o`030000003oool0oooo05<0oooo00<000000?ooo`3oool05@3oool00`0007l0oooo0?ooo`0f0?oo
o`030000003oool0oooo02P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0J0?oo
o`040000003oool0oooo000001<0oooo00D000000?ooo`3oool0oooo0000000=0?ooo`040000003o
ool0oooo000000X0oooo00<000000?ooo`000000203oool00`000000oooo000000070?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00A`3o
ool00`3ookl0M4RL0=ooo`030?ooo`030?noM000M;l0oooo0080oooo00<0_g@007Boo`3oool03@3o
ool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo01H0oooo00<0001o0?ooo`3oool0=@3o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0000000D0?ooo`040000003oool0oooo000000h0oooo00<000000?ooo`000000
2`3oool00`000000oooo000000080?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo00180?ooo`030;md0000
0000000000<0000000H0001d0;ooo`3oool0oooo0?oOW0180002000000030000B02Lgol0oooo00T0
oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0G0?ooo`030000O`3oool0oooo03D0
oooo00<000000?ooo`3oool0:@3oool01@000000oooo0?ooo`3oool0000001d0oooo00<000000?oo
o`000000503oool010000000oooo0?ooo`00000>0?ooo`030000003oool0000000/0oooo00<00000
0?ooo`000000203oool00`000000oooo000000070?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`009@3oool00`3ogi`0B0000000000>0000
00030000M02oool0oooo00l0oooo00<0okmd000000000000100000001@1d_ol0oooo0?ooo`3oool0
_g@000<0000000<0B9cO0?ooo`3oool0203oool00`000000oooo0?ooo`1@0?ooo`030000003oool0
oooo01P0oooo00<0001o0?ooo`3oool0=03oool00`000000oooo0?ooo`0Z0?ooo`050000003oool0
oooo0?ooo`0000007@3oool00`000000oooo0000000D0?ooo`040000003oool0oooo000000h0oooo
00<000000?ooo`0000002`3oool00`000000oooo000000080?ooo`030000003oool0000000L0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
000U0?ooo`030?oOW01dW=l0oooo00h0oooo00<0_gAd0;ooo`3oool0:@3oool200000500oooo00<0
00000?ooo`3oool06@3oool00`0007l0oooo0?ooo`0c0?ooo`030000003oool0oooo02/0oooo00@0
00000?ooo`3oool000007P3oool2000001D0oooo00<000000?ooo`0000003`3oool00`000000oooo
0000000;0?ooo`8000002@3oool00`000000oooo000000070?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0B03oool300000280oooo000U0?ooo`030?oOW02L_ol0oooo00h0oooo00<0
gibL0;ooo`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo01T0oooo00<0
001o0?ooo`3oool0<`3oool00`000000oooo0?ooo`0[0?ooo`040000003oool0oooo000001h0oooo
0P00000E0?ooo`030000003oool0000000l0oooo0P00000<0?ooo`8000002@3oool2000000P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
000U0?ooo`030?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3oool0:@3oool00`000000oooo0?oo
o`1>0?ooo`030000003oool0oooo01X0oooo00<0001o0?ooo`3oool0<P3oool00`000000oooo0?oo
o`0/0?ooo`030000003oool0000001l0oooo0P00000E0?ooo`030000003oool0000000l0oooo0P00
000<0?ooo`8000002@3oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo04d0oooo00<00000
0?ooo`3oool06`3oool00`0007l0oooo0?ooo`0a0?ooo`030000003oool0oooo02d0oooo00<00000
0?ooo`0000007`3oool2000001D0oooo0P00000@0?ooo`800000303oool2000000T0oooo0P000008
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P
0?ooo`00:`3oool00`3ogi`0B0000000000300000003001d_`3oool0oooo02h0oooo00<000000?oo
o`3oool0C@3oool00`000000oooo0?ooo`0K0?ooo`030000O`3oool0oooo0340oooo00<000000?oo
o`3oool0;@3oool00`000000oooo0000000O0?ooo`030000003oool0oooo01@0oooo0P00000@0?oo
o`800000303oool2000000T0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00:@3oool00`3OW4P000000000000600000003
0018W03Oool0oooo02d0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0L0?ooo`03
0000O`3oool0oooo0340oooo00<000000?ooo`3oool0;@3oool200000200oooo00<000000?ooo`3o
ool0503oool200000100oooo0P00000<0?ooo`8000002@3oool2000000P0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000W0?ooo`060?oo
_`1d000000000000M02ookl0W9cO1@3oool00`2oM7@0_ooo0?ooo`0]0?ooo`030000003oool0oooo
04/0oooo00<000000?ooo`3oool07@3oool00`0007l0oooo0?ooo`0`0?ooo`030000003oool0oooo
02h0oooo0P00000P0?ooo`030000003oool0oooo01@0oooo0P00000@0?ooo`800000303oool20000
00T0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`00
0000oooo0?ooo`0P0?ooo`009P3oool00`3ogi`0B01809cOo`020?ooo`030?oo_`2LW=l0oooo00@0
oooo00<0_ibL0=ooo`3oool0;@3oool2000004`0oooo00<000000?ooo`3oool07@3oool00`0007l0
oooo0?ooo`0`0?ooo`030000003oool0oooo02h0oooo0P00000P0?ooo`030000003oool0oooo01@0
oooo0P00000@0?ooo`800000303oool2000000T0oooo0P0000080?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0B03oool300000280oooo000V0?ooo`030;nLW03Oool0oooo0080oooo
00<0oono09bLg`3oool00P3oool00`3ookl0M00007Boo`0_0?ooo`030000003oool0oooo04X0oooo
00<000000?ooo`3oool07P3oool00`0007l0oooo0?ooo`0`0?ooo`030000003oool0oooo02h0oooo
0P00000P0?ooo`030000003oool0oooo01@0oooo0P00000@0?ooo`800000303oool2000000T0oooo
0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo
0?ooo`0P0?ooo`009@3oool00`3ookl0M7Bo0?ooo`030?ooo`030?oo_`1d0000000000<0000000<0
B9cO0?ooo`3oool0;P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo01h0oooo00<0
001o0?ooo`3oool0<03oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool0503oool200000100oooo0P00000<0?ooo`8000002@3oool2000000P0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000U
0?ooo`030?oog`2LB000000000L0000000<004RL0=ooo`3oool0<03oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo01h0oooo00<0001o0?ooo`3oool0;`3oool200000300oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`0D0?ooo`800000403oool00`000000oooo0?ooo`0;
0?ooo`8000002@3oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:
0?ooo`030000003oool0oooo0200oooo000V0?ooo`030?oo_`1d0000000000@0000000<0M;oo0?oo
o`3oool0<P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo01h0oooo00<0001o0?oo
o`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<000000?oo
o`3oool0503oool200000100oooo00<000000?ooo`3oool02`3oool2000000T0oooo0P0000080?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?oo
o`00;P3oool00`3OW7@0W=oo0?ooo`030?ooo`030?oOW018B9`0gooo02/0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01@0oooo0P00000@0?ooo`03
0000003oool0oooo00/0oooo0P0000090?ooo`800000203oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002d0oooo00<0okmd04RLg`3oool0
1@3oool00`2oM7@0_ooo0?ooo`0Z0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0
7`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0D0?ooo`800000403oool00`000000oooo0?ooo`0;0?ooo`800000
2@3oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0200oooo000]0?ooo`060=nLB01d_ol0oooo0?noM000B9`0gooo0P3oool00`3O_kl0
gooo0?ooo`0Z0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0
oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`0D0?ooo`800000403oool00`000000oooo0?ooo`0;0?ooo`8000002@3oool2000000P0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200
oooo000]0?ooo`090=nLB000M;l0oono07AdW02LB000B9cO0?ooo`3o_g@0B9cO02`0oooo0P00001:
0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0503oool200000100oooo00<00000
0?ooo`3oool02`3oool2000000T0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0B03oool300000280oooo000]0?ooo`080?oo_`1d0000001d0;ooo`3ogi`0B0000000
0000M;l]0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo
0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool2000000P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?oo
o`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l000000?oo
o`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?oo
o`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o
0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0803oool00680oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?oo
o`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?oo
o`030000O`000000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0200oooo000U0?ooo`030?oOW02L_ol0oooo00h0oooo00<0gibL0;ooo`3o
ool0:@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o0000003o
ool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3o
ool0803oool002D0oooo00<0omnL09boo`3oool03P3oool00`3OW9`0_ooo0?ooo`0Y0?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l000000?ooo`070?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`009@3o
ool00`3ogi`0M9cO0?ooo`0>0?ooo`030;mdM02oool0oooo02T0oooo0P00001:0?ooo`030000003o
ool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303o
ool2000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo04P0oooo0`00000R0?ooo`009@3oool00`3ogi`0B0000000000>000000030000M02oool0
oooo02T0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0
oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000
oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo0200oooo00100?ooo`060;md00000000000000000000B9`0gooo0`3oool01@3ogi`0B0000000
000007@0_ooo00@0oooo00D0omnL04P000000000001d0;ooo`0;0?ooo`030000003oool0oooo04T0
oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0
oooo0P00000<0?ooo`8000002@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B000
07Boo`040?ooo`070?oo_`1d0000M;oo0?ooo`3o_g@0001d0;ooo`020?ooo`070?oo_`1d0000M;oo
0?ooo`3o_g@0001d0;ooo`0:0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3o
ool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3o
ool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3o
ool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Boo`040?ooo`030;md0000B9`0
gooo0080oooo00L0_g@0001d_`3oool0oooo0;md0000B9`0gooo0080oooo00<0_g@0001d_`3oool0
2@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0
;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o0000003oool0
1`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
803oool00440oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M00004RLg`030?ooo`060=nLB000
07@0_ooo0?oo_`1d0000B9cO0`3oool00`3OW4P0001d0;ooo`090?ooo`030000003oool0oooo04T0
oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0
oooo0P00000<0?ooo`8000002@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B000
07Boo`030?ooo`030?oOW0180000M;oo00<0oooo00H0okmd0000B02Lgol0omnL04P0001d_ol30?oo
o`030?noM00004P0W=oo00T0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?oo
o`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?oo
o`030000O`000000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0200oooo00110?ooo`030?oOW0180000M;oo00<0oooo00<0omnL04P0001d
_ol00`3oool01P3o_g@0001809cOo`3ogi`0B00007Boo`<0oooo00<0okmd0000B02Lgol02@3oool3
000004T0001o00<000000000O`0007l0D@0007l00`000000001o0000O`0_0000O`030000000007l0
001o01l0001o00<000000000O`0007l05@0007l00`000000001o0000O`0>0000O`800000300007l2
000000X0001o00<000000000O`0007l01P0007l00`000000001o0000O`040000O`030000000007l0
001o00@0001o00<000000000O`0007l00P0007l00`000000001o0000O`020000O`030000000007l0
001o04L0001o1000000R0?ooo`00:@3oool00`3ooml0W4P00000000600000003001d_`3oool0oooo
00`0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`030?ooo`060?noM00004P0W=oo
0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`090?ooo`030000003oool0oooo04T0oooo00<0
00000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<0
00000?ooo`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00
000<0?ooo`8000002@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`009`3oool00`3ooml0W4P00000000:
00000003001d_`3oool0oooo00X0oooo00<0omnL04P0001d_ol00`3oool00`3ogi`0B00007Boo`03
0?ooo`060?noM00004P0W=oo0?oOW0180000M;oo0`3oool00`3o_g@0001809cOo`090?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l000000?ooo`070?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`009P3o
ool0103o_g@000000018W03Oool90?ooo`040?oo_`1d0000001809cOo`X0oooo00<0omnL04P0001d
_ol00`3oool00`3ookl0M00004RLg`030?ooo`060=nLB00007@0_ooo0?oo_`1d0000B9cO0`3oool0
0`3OW4P0001d0;ooo`090?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool0
0`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool0
0`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool0
0`000000oooo0?ooo`0P0?ooo`009@3oool00`3ookl0M7Bo0?ooo`0=0?ooo`030?oog`2LM9`0gooo
00T0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`070;md0000M;l0oooo
0?ooo`3OW4P004RL0=ooo`020?ooo`030;md0000M;l0oooo00T0oooo00<000000?ooo`3oool0B@3o
ool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3o
ool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000o0?ooo`050?oog`2LB000
00000000001d_ol0103oool01`3ookl0M00007Boo`3oool0okmd0000M02oool00P3oool01`3ookl0
M00007Boo`3oool0okmd0000M02oool02P3oool00`000000oooo0?ooo`190?ooo`030000003oool0
oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2
000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo04X0oooo00<000000?ooo`3oool0803oool00440oooo00<0okmd0000001d_ol01@3oool01@3o
okl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000001d0;ooo`0;0?ooo`030000003o
ool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003o
ool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l000000?ooo`070?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00;@3oool0
1@2oM0000000000000000000B9cO0300oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0
001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0
001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04P0oooo0`00
000R0?ooo`00:`3oool00`3ookl0M000000000040000000304RLg`3oool0oooo02d0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000[0?oo
o`030=nLB000M;l0oooo0080oooo00@0oono07@0000007@0_ooo;P3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>
0?ooo`800000303oool2000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002/0oooo00<0_gAd0;oo
o`3oool00`3oool00`3ookl0M01d0;ooo`0^0?ooo`030000003oool0oooo04T0oooo00<000000?oo
o`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?oo
o`8000002@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`009P3oool00`3o_i`0W=oo0?ooo`020?ooo`03
0?oOW02L_ol0oooo00<0oooo00<0oknL09cOo`3oool0;P3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`80
0000303oool2000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002H0oooo00<0_g@0000000000000
2@0000000`0007@0_ooo0?ooo`0]0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0
7`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`800000
2@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
BP3oool00`000000oooo0?ooo`0P0?ooo`009@3oool00`3ookl0M0000000000:00000003001d_`3o
ool0oooo02d0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3o
ool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`00
0000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003o
ool0oooo0200oooo000a0?ooo`030?oOW02L_ol0oooo02h0oooo0P00001:0?ooo`030000003oool0
oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2
000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo04P0oooo0`00000R0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo
01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool20000
00T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
04X0oooo00<000000?ooo`3oool0803oool002`0oooo00<0oono07@0000000000P0000000`18W=l0
oooo0?ooo`0^0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0
oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l0
00000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000
oooo0?ooo`0P0?ooo`00:`3oool00`3ogi`0B000000000040000000304RLg`3oool0oooo02d0oooo
00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
000[0?ooo`090=nLM02Lgol0okmd07Boo`3oool0okmd0000000007@0_ooo02h0oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000[0?ooo`04
0;mdB02Lgol0okmd07Boo`80oooo00<0gia80000M02oool0;P3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?oo
o`800000303oool2000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002/0oooo00@0okmd00000000
0000M;oo0P3oool00`2oM00007Bo0?ooo`0^0?ooo`030000003oool0oooo04T0oooo00<000000?oo
o`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?oo
o`8000002@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00;03oool01P3OW4P0000007Boo`3o_g@00000
001d_c00oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0
oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000
oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo0200oooo001R0?ooo`800000BP3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo
02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo
00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`180?ooo`<000008P3oool00680
oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0
oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200
oooo000/0?ooo`0;0?noM000000000000018W03Ookl0M000001d_`3OW4P000000000M02oool0:`3o
ool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool05@3o
ool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o0000003oool01`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803o
ool002/0oooo00<0okmd0000000000000P0000001`0007@0_kmd00000018W=l0oooo0;md0000M;l0
:`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0
;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o0000003oool0
1`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
803oool002/0oooo00d0_gAd0;ooo`3oool0oooo0?oog`2LB7@0_mnL04P0M02oool0oooo0?noM000
B9`0gooo02X0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3o
ool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`00
0000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003o
ool0oooo0200oooo000[0?ooo`030;mdB02Lgol0oooo0080oooo00P0_gAd0;oOW01804P0W=oo0?oo
o`3ogi`0B4RL0=ooobX0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`03
0000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`03
0000O`000000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo0200oooo000[0?ooo`030?oOW018000000000080000000L007Bo0?oOW01804P0
W=oo0?ooo`3o_g@007Bo02/0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?oo
o`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?oo
o`030000O`000000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0200oooo000[0?ooo`0<0?oo_`1d0000000000000000B9`0gooo0?oOW018
0000M;oo0?oog`2LB4P0W=oo:`3oool2000004X0oooo00<000000?ooo`3oool07`3oool00`0007l0
oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l0
00000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0B03oool300000280
oooo000[0?ooo`030?oo_`2LW=l0oooo00@0oooo00@0_g@00000000004P0W=oo;03oool00`000000
oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000
oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o0000003oool01`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00400oooo
00P0_g@000000000000000000018W03Oool0oooo0;md00H0000000@0B9cO0?ooo`3ooml0W4P00`00
00000`1d_ol0oooo0?ooo`0:0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3o
ool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3o
ool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3o
ool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Boo`040?ooo`030=nLB0000000
000000<0000000D004RL0=ooo`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00X0oooo00<0
00000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<0
00000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000U
0?ooo`030?oo_`1dB9`0gooo00d0oooo00<0omnL04Q8W03Oool02@3oool00`3ogi`0B00007Boo`05
0?ooo`030=nLB018W=l0oooo0080oooo00<0giad09cOo`3oool0103oool00`3ooml0W4Qd0;ooo`0:
0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_
0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`8000002@3oool00`0007l000000?ooo`07
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P
0?ooo`009P3oool0103ogi`0B0000018W03Oool90?ooo`040?oo_`1d000004RL0=ooo`X0oooo00<0
omnL04P0001d_ol01P3oool00`3OW4P0M;oo0?ooo`090?ooo`030;md001d_ol0oooo00T0oooo00<0
00000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<0
00000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo000W
0?ooo`030?oog`2LB000000000X0000000<007Bo0?ooo`3oool02P3oool00`3ogi`0B00007Boo`07
0?ooo`030;mdB02Lgol0oooo00L0oooo00<0oooO09a80018W=l02P3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>
0?ooo`800000303oool2000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool002T0oooo00<0oooO09a8
000000001P0000000`00M;l0oooo0?ooo`0<0?ooo`030?oOW0180000M;oo00L0oooo00<0oooO09a8
M02oool01`3oool00`3OW4P0000007Boo`0:0?ooo`<00000B@0007l00`000000001o0000O`1A0000
O`030000000007l0001o02l0001o00<000000000O`0007l07`0007l00`000000001o0000O`0E0000
O`030000000007l0001o00h0001o0P00000<0000O`8000002P0007l00`000000001o0000O`060000
O`030000000007l0001o00@0001o00<000000000O`0007l0100007l00`000000001o0000O`020000
O`030000000007l0001o0080001o00<000000000O`0007l0A`0007l400000280oooo00110?ooo`03
0?oOW0180000M;oo00P0oooo00<0okmd001d_`3oool00P3oool01`3Ogml0oooo0?oo_`1d00000000
0000M02oool02P3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o
0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0803oool00440oooo00<0omnL04P0001d_ol0203oool00`3ookl0M01809cOo`020?oo
o`060=nLB000000000000000000004P0W=oo2`3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303o
ool2000000T0oooo00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo04X0oooo00<000000?ooo`3oool0803oool00440oooo00<0omnL04P0001d_ol0203oool0
0`3ookl0M00004RLg`020?ooo`040?oo_`1d0000000004RLg`d0oooo00<000000?ooo`3oool0B@3o
ool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3o
ool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo00110?ooo`030?oOW0180000
M;oo00<0oooo00<0oknL09cOo`3oool00P3oool00`3ogi`0B00004RLg`030?ooo`030;md001d_ol0
oooo00d0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0
oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000
oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo0200oooo000o0?ooo`050?oog`2LB00000000000001d_ol00`3oool0203ooml0W4Q809cOo`3o
ool0oono07@0000007@0_ooo0`3oool01P3ogi`0B000000000000000001809cOo`X0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool03P3oool2000000`0oooo0P0000090?ooo`030000O`000000oooo00L0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo00110?oo
o`030?noM0000000M;oo00@0oooo00H0oooO09a80000000000000000M02oool50?ooo`050;md0000
0000000000000018W=l02P3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo
00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo
00<0001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo
00<000000?ooo`3oool0803oool00680oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0
001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0
001o0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04P0oooo0`00
000R0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o
0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000303oool2000000T0oooo00<0001o
0000003oool01`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0803oool00680oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?oo
o`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000`0oooo0P0000090?oo
o`030000O`000000oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3o
ool07`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo0P00000<0?ooo`80
00002@3oool00`0007l000000?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000
003oool0oooo01l0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`800000
303oool2000000T0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003o
ool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool0503oool200000100oooo0P00000<0?ooo`030000003o
ool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3o
ool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0
oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2
000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`00
07l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool20000
01H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680
oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000
oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0
oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0B@3oool00`000000oooo
0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`030000003oool0oooo00P0oooo
0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo
0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0
001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00
000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R
0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_
0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00000@
0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?oo
o`030000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3o
ool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0
7`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0
7P3oool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000
203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
803oool00680oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403o
ool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0B@3oool0
0`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool0
0`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`030000003oool0
oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool0
0`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo
01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool20000
00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo
0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0
oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0
oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo
00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo
0P00000<0?ooo`030000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<0
00000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<00000
0?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`08
0?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0803oool00680oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0001o0?oo
o`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?oo
o`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`03
0000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@03oool02@2oM000000000000000000004RL0=oo
o`3oool0oooO09a800030000000307Boo`3oool0oooo00<0oooo00D0omnL04P000000000001d0;oo
o`0;0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?oo
o`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00
000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00440oooo00<0omnL
04P0001d_ol00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo0080oooo00L0oono
07@0001d_ol0oooo0?noM00007@0_ooo00X0oooo00<000000?ooo`3oool0B@3oool00`000000oooo
0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`030000003oool0oooo00P0oooo
0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo
0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Boo`080?ooo`030?oog`2LB7@0_ooo0080oooo00<0
_g@00018W03Oool00P3oool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo04T0oooo00<0
00000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<0
00000?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?oo
o`080?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<0
00000?ooo`3oool0803oool00440oooo00<0omnL04P0001d_ol02@3oool01P2oM000M;oo0?ooo`3o
okl0M00004RLg`<0oooo00<0gia80000M02oool02@3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000
003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0
203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0200oooo00110?ooo`030?oOW0180000M;oo00P0oooo00L0oooO09a80018W=l0oooo
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3o
ool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo0200oooo00110?ooo`030?oOW0180000M;oo00P0oooo00L0gia80000001d_ol0
oooo0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool3000004T0001o00<000000000
O`0007l0D@0007l00`000000001o0000O`0_0000O`030000000007l0001o01h0001o0P00000F0000
O`800000400007l2000000`0001o00<000000000O`0007l0200007l2000000P0001o00<000000000
O`0007l0100007l00`000000001o0000O`040000O`030000000007l0001o0080001o00<000000000
O`0007l00P0007l00`000000001o0000O`170000O`@000008P3oool00440oooo00<0omnL04P0001d
_ol0103oool02`3Ogml0oooo0?oo_`1d000000000000M02oool0oooo0?oOW0180000M;oo00<0oooo
00<0okmd0000B02Lgol02@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo
00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo
0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
00110?ooo`030?oOW0180000M;oo00@0oooo00H0gia800000000000000000000B02Lgol20?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3o
ool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo0200oooo00110?ooo`030?oOW0180000M;oo00@0oooo00@0oono07@000000000
B9cO103oool00`3ookl0M00004RLg`030?ooo`030=nLB00007@0_ooo00T0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`03
0000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Boo`050?ooo`030;md
001d_ol0oooo00D0oooo00<0gia80018W03Oool00P3oool00`2oM00007Bo0?ooo`090?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000
303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo04X0oooo00<000000?ooo`3oool0803oool003l0oooo00D0oooO09a8000000000000
07Boo`050?ooo`060?oOW018000000000000000004P0W=oo0P3oool01`3ookl0M00007Boo`3oool0
okmd0000M02oool02P3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0
001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00
000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo0011
0?ooo`030?noM0000000M;oo00H0oooo00D0_g@0000000000000000004RLg`030?ooo`050?oo_`1d
000000000000M02oool02`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo
00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo
0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo
001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?oo
o`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00
000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo0P00001:
0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_
0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<00000
0?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`18
0?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?oo
o`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0N0?oo
o`8000005P3oool200000100oooo0P00000<0?ooo`030000003oool0oooo00P0oooo0P0000080?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?oo
o`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`80
0000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000
303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo00<000000?ooo`3oool0B@3o
ool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`030000003o
ool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3o
ool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0
oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2
000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`00
07l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool20000
01H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680
oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000
oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0
oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0B@3oool00`000000oooo
0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`030000003oool0oooo00P0oooo
0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool00`000000oooo
0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0
001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00
000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo001R
0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_
0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00000@
0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?oo
o`030000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`03
0000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3o
ool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`03
0000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0
7`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0
7P3oool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000
203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
803oool00680oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403o
ool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3oool0B@3oool0
0`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0;`3oool0
0`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`030000003oool0
oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0BP3oool0
0`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo
01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool20000
00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo
0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0
oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0
oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00680oooo
00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo
0P00000<0?ooo`030000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<0
00000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo0200oooo001R0?ooo`030000003oool0oooo04T0oooo00<00000
0?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`08
0?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0803oool00680oooo0P00001:0?ooo`030000003oool0oooo01l0oooo00<0001o0?oo
o`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?oo
o`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`180?ooo`<000008P3oool00680oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`03
0000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00HP3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000
003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<000000?ooo`3oool0
203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0200oooo00100?ooo`060;md00000000000000000000B9`0gooo103oool00`3OW4P0
B9cO0?ooo`030?ooo`030?oog`2LB00000000080000000<0M;oo0?ooo`3oool02P3oool00`000000
oooo0?ooo`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000
oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0
oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`1:0?ooo`030000003oool0oooo0200oooo00110?ooo`030?oOW0180000M;oo00H0oooo
00<0omnL04Q8W03Oool00`3oool00`3ooml0W4Qd0;ooo`020?ooo`030;md001d_ol0oooo00X0oooo
00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo
00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo
0P00000<0?ooo`030000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Bo
o`060?ooo`030?oog`2LB4P0W=oo00P0oooo00<0oooO09a8M02oool02P3oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo01l0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?oo
o`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?ooo`800000403oool2000000`0oooo00<0
00000?ooo`3oool0203oool2000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo0200oooo00110?ooo`030?oOW0180000M;oo00L0oooo00<0gia8
04RLg`3oool0203oool00`2oM000M;oo0?ooo`090?ooo`030000003oool0oooo04T0oooo00<00000
0?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`08
0?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04X0oooo00<00000
0?ooo`3oool0803oool00440oooo00<0omnL04P0001d_ol01`3oool00`3ogi`0B4RL0=ooo`070?oo
o`030?oog`2LB000B9cO00X0oooo0`0000190?ooo`030000003oool0oooo01l0oooo00<0001o0?oo
o`3oool0;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo0P00000F0?oo
o`800000403oool2000000`0oooo00<000000?ooo`3oool0203oool2000000P0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`170?ooo`@000008P3oool00440oooo00<0omnL04P0001d
_ol01`3oool00`3ooml0W4Q809cOo`070?ooo`030=nLB0000000M;oo00X0oooo00<000000?ooo`3o
ool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3o
ool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?ooo`03
0000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Boo`080?ooo`030=nL
B000M;l0oooo0080oooo00L0gmoO0?ooo`3ookl0M0000000000007@0_ooo00X0oooo00<000000?oo
o`3oool0B@3oool00`000000oooo0?ooo`0O0?ooo`030000O`3oool0oooo02l0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`0N0?ooo`8000005P3oool200000100oooo0P00000<0?oo
o`030000003oool0oooo00P0oooo0P0000080?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0BP3oool00`000000oooo0?ooo`0P0?ooo`00@@3oool00`3ogi`0B00007Boo`080?ooo`03
0?oo_`1d07@0_ooo0080oooo00H0gia800000000000000000000B02Lgol;0?ooo`030000003oool0
oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000303oool0
0`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo04X0oooo00<000000?ooo`3oool0803oool00440oooo00<0omnL04P0001d_ol00`3oool00`3o
gml0gooo0?ooo`030?ooo`080;md001d_ol0oooo0?ooo`3ookl0M00000000018W=l=0?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool07`3oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool07P3oool2000001H0oooo0P00000@0?ooo`800000
303oool00`000000oooo0?ooo`080?ooo`800000203oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo04X0oooo00<000000?ooo`3oool0803oool00440oooo00<0omnL04P0001d_ol00`3o
ool00`3ookl0M4RL0=ooo`030?ooo`030?noM000M;l0oooo0080oooo00<0_g@007Boo`3oool03@3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool07P3o
ool2000001H0oooo0P00000@0?ooo`800000303oool00`000000oooo0?ooo`080?ooo`800000203o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`0P0?ooo`00?`3oool01@3ooml0W4P0000000000000M;oo00@0oooo00<0
_g@00000000000000`0000001P0007@0_ooo0?ooo`3oool0omnL04P00080000000<0001809cOo`3o
ool02@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool02P3oool2000000D0oooo00<000000?ooo`3oool03P3oool2000000030?ooo`000000oooo00d0
oooo0P0000030?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool2000000P0
oooo00<000000?ooo`3oool0103oool010000000oooo0?ooo`0000030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0803oool00440oooo00<0okmd0000001d_ol0103oool00`3o_g@0
0000000000040000000507Boo`3oool0oooo0?ooo`2oM0000`0000000`18W=l0oooo0?ooo`080?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0:0?oo
o`8000001@3oool00`000000oooo0?ooo`0>0?ooo`80000000<0oooo0000003oool03@3oool20000
00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`080?ooo`800000203oool00`00
0000oooo0?ooo`040?ooo`040000003oool0oooo000000<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0P0?ooo`00HP3ooooo00000:L000008P3oool00?l0ooooo`3oool/0?ooo`00
o`3ooooo0?ooob`0oooo003o0?ooool0oooo;03oool00?l0ooooo`3oool/0?ooo`00o`3ooooo0?oo
ob`0oooo003o0?ooool0oooo;03oool00?l0ooooo`3oool/0?ooo`00o`3ooomP0?ooo`030?oog`2L
B000000000<0000000<0B9cO0?ooo`3oool06P3oool00`3ooml0W4P0000000030000000304RLg`3o
ool0oooo09l0oooo003o0?ooof00oooo00<0_g@00018W03Oool00`3oool00`3OW9`0_ooo0?ooo`0J
0?ooo`030;md0000B9`0gooo00<0oooo00<0gibL0;ooo`3oool0W`3oool00?l0ooooH03oool00`3o
oml0W7BL0=ooo`030?ooo`030?oo_`2LW=l0oooo01X0oooo00<0oooO09adW03Oool00`3oool00`3o
okl0W9cO0?ooo`2O0?ooo`00o`3ooomQ0?ooo`030;md00000000000000<0000000<0B9cO0?ooo`3o
ool06P3oool00`2oM0000000000000030000000304RLg`3oool0oooo09h0oooo003?0?ooo`030?oo
g`2LB00000000080000000P0B9cO0?ooo`3oool0_gA809cOo`3oool0omnL04P000<0000000<004RL
0=oog`2LB0000P0000002@00B9`0W7@000000000M;l0_g@000000000M;l0oooo0=nLB00200000006
0018W02oW4P000000000000007@0W9a80P0000000`0004P0W=oO09a80002000000030000M02ooml0
W4P00080000000@007Bo0?ooo`3ookl0M0000P0000001@18W;l0W4P00018W03Oool0gia800800000
00<0001809cOg`2LB0000P0000000`0007@0_ooO09a8000200000003001d_`3oool0oooo00D0oooo
00@0oooO09adW03Ooml0W4P00`0000000`1d_ol0oooo0?ooo`020?ooo`080;md0000000000000000
B02Lgol0oooo0?oog`2LB0020000000>0018W02LM0000000001d_`3oool0omnL04P000000000B9cO
0?ooo`3ogi`0B00000000018W=l20?ooo`070;md0000000000000000M02oool0oooo0;md00020000
00040000B02Lgol0oooo0=nLB080000000L0001d07Ad0000000000000000M02ooml0W4P000800000
00<004RL0=oo_`1d0000100000000`0007@0_ooo0?ooo`040?ooo`050?oo_`1d0000000000000018
W=l00`3oool01P3OW4P0000000000000000004RL09ad0080000000<0001d0;oog`2LB0000P000000
0`00B9`0gono07@00004000000030000M02oool0gia80080000000@0001d0;ooo`3oool0_g@00P00
00000`0007@0_ooo0?ooo`2D0?ooo`00d@3oool00`3ogi`0B4RL0=ooo`020?ooo`030?oOW0180000
07Bo00<0oooo01`0gia80000B02Lgol0oooo0?noM0000000M;oo0?ooo`3ogi`0B000001d_`3o_kl0
_ooO09a8M02oool0oooo0?oog`2LB00004RL0=ooo`3oool0gia80000M02oool0oooo0=nLB00007@0
_ooo0P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l00P3oool01@3OW4P0001d0;ooo`3o
ool0gia80080000000D0001d0;ooo`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3o
oml0W4P0001d_`080?ooo`040?oog`2LB000001809cOo`80oooo00<0omnL04P0001d_ol00P3oool0
9@2oM00000000018W03Oool0oooO09a8M02oool0okmd0000001d_ol0oooo0?oOW018000007Bo0?no
_`2oool0gia80000M02oool0gkno0=ooo`3OW4P0001d0;ooo`3O_kl0gooo0;md0000000007Bo0?oo
o`3ogi`0B01d0;ooo`3oool0gia80000M02oool00`3oool02`3OW4P0001d0;ooo`3oool0gia80018
W03Oool0oooo0?oog`2LB00007Bo0080oooo00<0_gA809cOo`3oool02P3oool00`3ookl0M01809cO
o`050?ooo`030=nLB0000000B9cO0080oooo00L0gia80018W03Oool0oooo0?oog`2LB00007Bo0080
oooo00<0_gA809cOo`3oool01@3oool02`3OW4P0001d0;ooo`3oool0_g@000000000M;l0oooo0?oO
W01807@0_ooo09@0oooo003A0?ooo`030?oo_`1dB9`0gooo0080oooo00@0_g@00000000004P0W=oo
0P3oool02`3OW4P0001d0;ooo`3oool0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00H0
oono07A8W03Oool0_g@00000M02oool20?ooo`070=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`02
0?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?ooo`070;md0000B9`0gooo0?ooo`3o
oml0W4P0001d_`020?ooo`030=nLB00007@0_ooo0080oooo00L0_g@00018W03Oool0oooo0?oog`2L
B00007Bo00P0oooo00<0oooO09a8B02Lgol0103oool01P2oM00004RL0=ooo`3ogi`0B000001d_`<0
oooo00T0omnL09boo`3o_g@0001809cOo`3oool0oooO09a80000M;l00P3oool00`3OW4P004RL0=oo
o`020?ooo`070=nLB000B9`0gooo0?ooo`3ogi`0B00004RLg`030?ooo`060?oOW01dW=l0oooo0=nL
B00007@0_ooo0`3oool02`3OW4P0001d0;ooo`3oool0gia80018W03Oool0oooo0?oog`2LB00007Bo
0080oooo00D0oono07@0000000000000001d_`090?ooo`030?noM000M;l0oooo00<0oooo00@0oooO
09a8000007@0_ooo0P3oool01`3OW4P004RL0=ooo`3oool0oooO09a80000M;l00P3oool01@3ookl0
M00000000000000007Bo00<0oooo00H0gia80000M02oool0omnL04P00018W=l30?ooo`030?oOW01d
W=l0oooo09<0oooo003A0?ooo`090?oo_`1dB9`0gooo0?ooo`3ogi`0B000001d_`3o_g@0B9cO0080
oooo00<0gia80000M02oool00P3oool01P3o_g@00000001d_`3ooml0W4P0001d_`<0oooo00@0omnL
07A8B0000000M;oo0`3oool01`3OW4P0001d0;ooo`3oool0gia80000M02oool00P3oool01`2oM000
04RL0=ooo`3oool0oooO09a80000M;l00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l0
0P3oool00`3OW4P0001d0;ooo`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`080?oo
o`030?oog`2LM9`0gooo00<0oooo00L0oooO09a80000B9`0gooo0?noM00004P0W=oo00H0oooo00H0
okmd00000000M;l0oooO09a80000M;l20?ooo`030=nLB000B9`0gooo0080oooo00L0gia80018W03O
ool0oooo0?noM00004P0W=oo00H0oooo00<0gia80000M02oool00`3oool02`3OW4P0001d0;ooo`3o
ool0gia80018W03Oool0oooo0?oog`2LB00007Bo0080oooo00H0gia80000M02oool0oono07@00018
W=l80?ooo`030?oo_`1d07@0_ooo00<0oooo00<0omnL04P00018W=l00`3oool01`3OW4P004RL0=oo
o`3oool0oooO09a80000M;l00P3oool01P3OW4P0001d0;ooo`3ookl0M00004RLg`80oooo00H0gia8
0000M02oool0okmd0000B02LgonI0?ooo`00d@3oool03P3ookl0M4RL0=ooo`3oool0_g@00000B02L
gol0oono07@0M02oool0oooo0=nLB00007@0_ooo103oool0103o_g@0000000000000M;l30?ooo`03
0?oo_`1d0000B9cO00@0oooo00L0gia80000M02oool0oooo0=nLB00007@0_ooo0080oooo00L0_g@0
0018W03Oool0oooo0?oog`2LB00007Bo0080oooo00L0_g@00018W03Oool0oooo0?oog`2LB00007Bo
0080oooo00<0gia80000M02oool00P3oool01`2oM00004RL0=ooo`3oool0oooO09a80000M;l03@3o
ool0203ookl0M00000000000M;l0oooo0?noM00007@0_ooo203oool0103o_g@0000000000000M;l2
0?ooo`030=nLB000B9`0gooo0080oooo00L0gia80018W03Oool0oooo0?noM00007@0_ooo00H0oooo
00<0gia80000M02oool00`3oool04`3OW4P0001d0;ooo`3oool0gia80018W03Oool0oooo0?oog`2L
B00007Bo0?ooo`3ookl0M00004RLg`3oool0oooO09a80000M;l02@3oool00`2oM000000000000003
000000030000B02Lgol0oooo0080oooo00l0gia80018W03Oool0oooo0?oog`2LB00007Bo0?ooo`3o
okl0M00004RLg`3oool0oooO09a80000M;l00P3oool01P3OW4P0001d0;ooo`3o_g@0001d0;oooiT0
oooo003A0?ooo`0F0?oo_`1dB9`0gooo0?oOW018000004RL0=ooo`3oool0gia804RLg`3oool0gia8
0000M02oool0oooo0?oo_`1d0000M;oo0?ooo`3ooml0W4P0001d_`80oooo00D0oooO09a80000B9`0
W7A809cOo`030?ooo`070=nLB00007@0_ooo0?ooo`3OW4P0001809cOo`020?ooo`070;md000007@0
_ooo0?ooo`3ooml0W4P0001d_`020?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`020?oo
o`030=nLB00004P0W=oo0080oooo00L0_g@00000M02oool0oooo0?oog`2LB00007Bo00/0oooo00X0
oono07@00000000000000018W03Oool0oooo0?oOW01804P0W=oo0P3oool02P3o_g@0001809cOo`3o
okl0M00007Boo`3oool0oooO09a80000M;l20?ooo`030=nLB000B9`0gooo0080oooo00H0gia80018
W03Oool0oooo0?oOW01800050000000504RLg`3oool0gia80000B02Lgol00`3oool04`3OW4P0001d
0;ooo`3oool0_g@00000M02oool0oooo0?oog`2LB00007Bo0?ooo`3ookl0M00004RLg`3oool0oooO
09a80018W=l02@3oool0203ogi`0B4RL0=ooo`3oool0oooO09a80000B9`0gooo0`3oool03`2oM000
001d0;ooo`3oool0oooO09a80000M;l0oooo0?oo_`1d0000B9cO0?ooo`3ooml0W4P004RLg`020?oo
o`050=nLB00007@0_ooo0?oOW01800001@0000000`18W=l0oooo0?ooo`2B0?ooo`00d@3oool01P3o
okl0M4RL0=ooo`3OW4P0000007Boo`80oooo00h0oono07@0M02oool0gia80000M02oool0oooo0?oo
_`1d0000B9cO0?ooo`3ookl0M000001d_`80oooo01P0_g@00000B02Lgol0oooO09a8M02oool0oooO
09a800000000001d0;oog`2LB00000000000B018B4P0W=oo0?oo_`1d000007Ad07@0M02oool0oono
07@00018W=l20?ooo`0E0;md0000B9`0gooo0?ooo`3ooml0W4P0001d_`3ooml0W4P00000000004P0
B4Q809cOo`3ookl0M000001dM01d07@0_ooo0?oo_`1d0000B9cO00X0oooo00D0_g@0000000000000
001809cOo`040?ooo`0>0;md0000B9`0gooo0?oog`2LB000001d0;ooo`3ookl0M00004RLg`3oool0
oono07@00000M;l20?ooo`030=nLB000B9`0gooo0080oooo00<0gia80018W03Oool00P3oool06@2o
M000B9cO0?ooo`3oool0gia80000B02Lgml0W4P00000000004P0B7BL0=noM00004P0B4P000000000
07@0_ooO09a80000000004Q804P0B02Lgol0oono07@00018W=l00P3oool0102oM00007Bo0?ooo`3o
_g@2000000030018W03Oool0oooo00H0oooo00L0oooO09a8001d_ol0oooo0?oOW0180000B9cO0080
oooo00T0oooO09a80000000004Q804P0B02Lgol0oono07@00018W=l00P3oool0102oM00007Bo0?oo
o`3o_g@2000000080018W03Oool0gia80000M02oool0oooo0;md0018W=l20?ooo`030=nLB00004P0
W=oo09@0oooo003A0?ooo`060?oo_`1dB9`0gono07@00000B9`0gooo0`3oool01@2oM000M;oo0=nL
B00007@0_ooo0080oooo00P0oono07@00000000000000000M02oool0oooO09a80080000001l007Bo
0?noM0000000001d0;ooo`3oool0_g@00000M02oool0oooo0;md000007@0_ooO09a8000000000018
09cOo`3ookl0M0000000000007@0_ooo0?oo_`1d000000000018W03Oool0omnL04P00000000007Bo
0080oooo00`0_g@00000M02ooml0W4P00000000004P0W=oo0?oo_`1d000000000000M02oool90?oo
o`040=nLB0000000000007Boo`L0oooo00D0gia8000000000000001809cOo`020?ooo`080?oo_`1d
000000000000000007@0_ooo0?ooo`2oM003000000030018W02LM00000000080000000@004RL0=oo
o`3oool0gia80P0000005@0004P0W=oo0?ooo`3oool0_g@00000M02ogi`0B0000000M02oool0_g@0
0000M02oool0oooo0;md0000B9`0gooO09a800000000001d0;ooo`020?ooo`050?oog`2LB0000000
0000001d_ol02`3oool01P3o_g@004RL0=ooo`3OW4P0001809cOo`@0oooo00L0_g@00018W03Ooml0
W4P00000000007@0_ooo0080oooo00D0oooO09a800000000000007Boo`030?ooo`030=nLB00007@0
_ooo0080oooo00D0gia8000000000000001809cOo`2E0?ooo`00d@3oool01@3ookl0M4RL0;nLB000
0000M;oo00@0oooo00D0omnL04Q8W02oW4P0001d0;ooo`110?ooo`040?oog`2LB000001d0;ooo`<0
oooo00<0oknL09cOo`3oool04@3oool00`2oM00004RL0=ooo`020?ooo`030;md0000B9`0gooo02d0
oooo00H0oono07@0B02Lgml0W4P00018W03OoolE0?ooo`030=nLB00007@0_ooo09`0oooo003A0?oo
o`050?oo_`1dB4P0B0000018W03Oool01@3oool0102oM000B4Q80000M02ooom10?ooo`030?oog`2L
B00007Bo00<0oooo00<0oooO09a8B02Lgol04P3oool00`3ookl0M4RL0=ooo`020?ooo`030?oo_`1d
B9`0gooo02h0oooo00@0gia80018B0180000B9cO5P3oool00`3OW4P0001d0;ooo`2L0?ooo`00d@3o
ool0103ogi`0B0000000001d_ol60?ooo`040?noM0000000001809cOod80oooo00P0gia80000B02L
gol0oooo0?oog`2LB000001809cOoa80oooo00<0oooO09adW03Oool00P3oool00`3ooml0W7BL0=oo
o`0^0?ooo`040?oOW0180000001809cOoaH0oooo00<0gia80000M02oool0W03oool00<l0oooo00H0
oooO09a80000000000000000M02oool60?ooo`060?oog`2LB000000000000000B9`0gooo403oool0
0`3OW4P0001809cOo`0^0?ooo`070?noM000000000000000B02Lgol0giad09cOo`0Z0?ooo`030=nL
B00004P0W=oo01`0oooo00<0_g@00018W03Oool0503oool01@3ooml0W4P00000000007@0_ooo09`0
oooo003a0?ooo`030=nLB00004P0W=oo05l0oooo00<0gia80000B02Lgol0=@3oool00`2oM000001d
0;ooo`2L0?ooo`00o`3ooooo0?ooob`0oooo003o0?ooool0oooo;03oool00?l0ooooo`3oool/0?oo
o`00o`3ooooo0?ooob`0oooo0000\
\>"],
  ImageRangeCache->{{{0, 553.063}, {341.438, 0}} -> {-1.53888, -36.1841, \
0.00747873, 1.03721}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Oblique elastic collision between unequal masses", "Subsection"],

Cell[BoxData[
    \(\(values = {m1 \[Rule] 2, m2 \[Rule] 1, m3 \[Rule] 2, qvalue \[Rule] 0, 
          p1r \[Rule] 1, theta1 \[Rule] 30\ Degree, p2r \[Rule] 3, 
          theta2 \[Rule] \(-50\)\ Degree};\)\)], "Input"],

Cell[BoxData[
    \(\(sol = \(generalSolution[p3r, p4x, 
              p4y]\)\[LeftDoubleBracket]1\[RightDoubleBracket] //. 
          values;\)\)], "Input"],

Cell[BoxData[
    \(calcCollision[sol, values, 20\ Degree]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(drawCollision @@ LabMomenta;\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.20998 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.481541 0.237368 0.635666 0.237368 [
[ 0 0 0 0 ]
[ 1 1.20998 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.20998 L
0 1.20998 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.27597 .51698 m
.48154 .63567 L
s
% Start of user PostScript
-0.8660254 -0.5 0. 0. MAarrowhead1
% End of user PostScript
.02381 1.18117 m
.48154 .63567 L
s
% Start of user PostScript
-1.9283628 2.2981333 0. 0. MAarrowhead1
% End of user PostScript
.48154 .63567 m
.97619 .8157 L
s
% Start of user PostScript
0. 0. 2.0838941 0.75847542 MAarrowhead1
% End of user PostScript
.48154 .63567 m
.65019 .02881 L
s
% Start of user PostScript
0. 0. 0.71049414 -2.5566088 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{238, 287.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005U0001[b000`400?l00000o`00003oo`3ooomV
0?ooo`00o`3ooomV0?ooo`00o`3ooomV0?ooo`00o`3ooomV0?ooo`00o`3ooomV0?ooo`00o`3ooomV
0?ooo`00o`3ooomV0?ooo`00o`3ooomV0?ooo`00o`3ooomV0?ooo`00o`3ooomV0?ooo`00i`3oool0
0`000000oooo0?ooo`1k0?ooo`00iP3oool2000007d0oooo003V0?ooo`800000O@3oool00>D0oooo
0`00001m0?ooo`00i@3oool3000007d0oooo003T0?ooo`@00000O@3oool00>@0oooo1000001m0?oo
o`00h`3oool5000007d0oooo003U0?ooo`<00000O@3oool00>@0oooo00<000000?ooo`3oool0OP3o
ool00>@0oooo00<000000?ooo`3oool0OP3oool00>@0oooo00<000000?ooo`3oool0OP3oool00>@0
oooo00<000000?ooo`3oool0OP3oool00><0oooo00<000000?ooo`3oool0O`3oool00><0oooo00<0
00000?ooo`3oool0O`3oool00><0oooo00<000000?ooo`3oool0O`3oool00><0oooo00<000000?oo
o`3oool0O`3oool00>80oooo00<000000?ooo`3oool0P03oool00>80oooo00<000000?ooo`3oool0
P03oool00>80oooo00<000000?ooo`3oool0P03oool00>40oooo00<000000?ooo`3oool0P@3oool0
0>40oooo00<000000?ooo`3oool0P@3oool00>40oooo00<000000?ooo`3oool0P@3oool00>40oooo
00<000000?ooo`3oool0P@3oool00>00oooo00<000000?ooo`3oool0PP3oool00>00oooo00<00000
0?ooo`3oool0PP3oool00>00oooo00<000000?ooo`3oool0PP3oool00=l0oooo00<000000?ooo`3o
ool0P`3oool00=l0oooo00<000000?ooo`3oool0P`3oool00=l0oooo00<000000?ooo`3oool0P`3o
ool00=l0oooo00<000000?ooo`3oool0P`3oool00=h0oooo00<000000?ooo`3oool0Q03oool00=h0
oooo00<000000?ooo`3oool0Q03oool00=h0oooo00<000000?ooo`3oool0Q03oool00=h0oooo00<0
00000?ooo`3oool0Q03oool00=d0oooo00<000000?ooo`3oool0Q@3oool00=d0oooo00<000000?oo
o`3oool0Q@3oool00=d0oooo00<000000?ooo`3oool0Q@3oool00=`0oooo00<000000?ooo`3oool0
QP3oool00=`0oooo00<000000?ooo`3oool0QP3oool00=`0oooo00<000000?ooo`3oool0QP3oool0
0=`0oooo00<000000?ooo`3oool0QP3oool00=/0oooo00<000000?ooo`3oool0Q`3oool00=/0oooo
00<000000?ooo`3oool0Q`3oool00=/0oooo00<000000?ooo`3oool0Q`3oool00=X0oooo00<00000
0?ooo`3oool0R03oool00=X0oooo00<000000?ooo`3oool0R03oool00=X0oooo00<000000?ooo`3o
ool0R03oool00=X0oooo00<000000?ooo`3oool0R03oool00=T0oooo00<000000?ooo`3oool0R@3o
ool00=T0oooo00<000000?ooo`3oool0R@3oool00=T0oooo00<000000?ooo`3oool0R@3oool00=T0
oooo00<000000?ooo`3oool0R@3oool00=P0oooo00<000000?ooo`3oool0RP3oool00=P0oooo00<0
00000?ooo`3oool0RP3oool00=P0oooo00<000000?ooo`3oool0RP3oool00=L0oooo00<000000?oo
o`3oool0R`3oool00=L0oooo00<000000?ooo`3oool0R`3oool00=L0oooo00<000000?ooo`3oool0
R`3oool00=L0oooo00<000000?ooo`3oool0R`3oool00=H0oooo00<000000?ooo`3oool0S03oool0
0=H0oooo00<000000?ooo`3oool0S03oool00=H0oooo00<000000?ooo`3oool0S03oool00=D0oooo
00<000000?ooo`3oool0S@3oool00=D0oooo00<000000?ooo`3oool0S@3oool00=D0oooo00<00000
0?ooo`3oool0S@3oool00=D0oooo00<000000?ooo`3oool0S@3oool00=@0oooo00<000000?ooo`3o
ool0SP3oool00=@0oooo00<000000?ooo`3oool0SP3oool00=@0oooo00<000000?ooo`3oool0SP3o
ool00=@0oooo00<000000?ooo`3oool0SP3oool00=<0oooo00<000000?ooo`3oool0S`3oool00=<0
oooo00<000000?ooo`3oool0S`3oool00=<0oooo00<000000?ooo`3oool0S`3oool00=80oooo00<0
00000?ooo`3oool0T03oool00=80oooo00<000000?ooo`3oool0T03oool00=80oooo00<000000?oo
o`3oool0T03oool00=80oooo00<000000?ooo`3oool0T03oool00=40oooo00<000000?ooo`3oool0
T@3oool00=40oooo00<000000?ooo`3oool0T@3oool00=40oooo00<000000?ooo`3oool0T@3oool0
0=00oooo00<000000?ooo`3oool0TP3oool00=00oooo00<000000?ooo`3oool0TP3oool00=00oooo
00<000000?ooo`3oool0TP3oool00=00oooo00<000000?ooo`3oool0TP3oool00<l0oooo00<00000
0?ooo`3oool0T`3oool00<l0oooo00<000000?ooo`3oool0T`3oool00<l0oooo00<000000?ooo`3o
ool0T`3oool00<l0oooo00<000000?ooo`3oool0T`3oool00<h0oooo00<000000?ooo`3oool0U03o
ool00<h0oooo00<000000?ooo`3oool0U03oool00<h0oooo00<000000?ooo`3oool0U03oool00<d0
oooo00<000000?ooo`3oool0U@3oool00<d0oooo00<000000?ooo`3oool0U@3oool00<d0oooo00<0
00000?ooo`3oool0U@3oool00<d0oooo00<000000?ooo`3oool0U@3oool00<`0oooo00<000000?oo
o`3oool0UP3oool00<`0oooo00<000000?ooo`3oool0UP3oool00<`0oooo00<000000?ooo`3oool0
UP3oool00</0oooo00<000000?ooo`3oool0U`3oool00</0oooo00<000000?ooo`3oool0U`3oool0
0</0oooo00<000000?ooo`3oool0U`3oool00</0oooo00<000000?ooo`3oool0U`3oool00<X0oooo
00<000000?ooo`3oool0V03oool00<X0oooo00<000000?ooo`3oool0V03oool00<X0oooo00<00000
0?ooo`3oool0V03oool00<X0oooo00<000000?ooo`3oool0V03oool00<T0oooo00<000000?ooo`3o
ool0V@3oool00<T0oooo00<000000?ooo`3oool0V@3oool00<T0oooo00<000000?ooo`3oool0V@3o
ool00<P0oooo00<000000?ooo`3oool0VP3oool00<P0oooo00<000000?ooo`3oool0VP3oool00<P0
oooo00<000000?ooo`3oool0VP3oool00<P0oooo00<000000?ooo`3oool0VP3oool00<L0oooo00<0
00000?ooo`3oool0V`3oool00<L0oooo00<000000?ooo`3oool0V`3oool00<L0oooo00<000000?oo
o`3oool0V`3oool00<H0oooo00<000000?ooo`3oool0W03oool00<H0oooo00<000000?ooo`3oool0
W03oool00<H0oooo00<000000?ooo`3oool0W03oool00<H0oooo00<000000?ooo`3oool0W03oool0
0<D0oooo00<000000?ooo`3oool0W@3oool00<D0oooo00<000000?ooo`3oool0W@3oool00<D0oooo
00<000000?ooo`3oool0W@3oool00<D0oooo00<000000?ooo`3oool0W@3oool00<@0oooo00<00000
0?ooo`3oool0WP3oool00<@0oooo00<000000?ooo`3oool0WP3oool00<@0oooo00<000000?ooo`3o
ool0WP3oool00<<0oooo00<000000?ooo`3oool0W`3oool00<<0oooo00<000000?ooo`3oool0W`3o
ool00<<0oooo00<000000?ooo`3oool0W`3oool00<<0oooo00<000000?ooo`3oool0W`3oool00<80
oooo00<000000?ooo`3oool0X03oool00<80oooo00<000000?ooo`3oool0X03oool00<80oooo00<0
00000?ooo`3oool0X03oool00<40oooo00<000000?ooo`3oool0X@3oool00<40oooo00<000000?oo
o`3oool0X@3oool00<40oooo00<000000?ooo`3oool0X@3oool00<40oooo00<000000?ooo`3oool0
X@3oool00<00oooo00<000000?ooo`3oool0XP3oool00<00oooo00<000000?ooo`3oool0XP3oool0
0<00oooo00<000000?ooo`3oool0XP3oool00<00oooo00<000000?ooo`3oool0XP3oool00;l0oooo
00<000000?ooo`3oool0X`3oool00;l0oooo00<000000?ooo`3oool0X`3oool00;l0oooo00<00000
0?ooo`3oool0X`3oool00;h0oooo00<000000?ooo`3oool0Y03oool00;h0oooo00<000000?ooo`3o
ool0Y03oool00;h0oooo00<000000?ooo`3oool0Y03oool00;h0oooo00<000000?ooo`3oool0Y03o
ool00;d0oooo00<000000?ooo`3oool0Y@3oool00;d0oooo00<000000?ooo`3oool0Y@3oool00;d0
oooo00<000000?ooo`3oool0Y@3oool00;`0oooo00<000000?ooo`3oool0YP3oool00;`0oooo00<0
00000?ooo`3oool0YP3oool00;`0oooo00<000000?ooo`3oool0YP3oool00;`0oooo00<000000?oo
o`3oool0YP3oool00;/0oooo00<000000?ooo`3oool0Y`3oool00;/0oooo00<000000?ooo`3oool0
Y`3oool00;/0oooo00<000000?ooo`3oool0Y`3oool00;/0oooo00<000000?ooo`3oool0Y`3oool0
0;X0oooo00<000000?ooo`3oool0Z03oool00;X0oooo00<000000?ooo`3oool0Z03oool00;X0oooo
00<000000?ooo`3oool0Z03oool00;T0oooo00<000000?ooo`3oool0Z@3oool00;T0oooo00<00000
0?ooo`3oool0Z@3oool00;T0oooo00<000000?ooo`3oool0Z@3oool00;T0oooo00<000000?ooo`3o
ool0Z@3oool00;P0oooo00<000000?ooo`3oool0ZP3oool00;P0oooo00<000000?ooo`3oool0ZP3o
ool00;P0oooo00<000000?ooo`3oool0ZP3oool00;L0oooo00<000000?ooo`3oool0Z`3oool00;L0
oooo00<000000?ooo`3oool0Z`3oool00;L0oooo00<000000?ooo`3oool0Z`3oool00680oooo00<0
00000?ooo`3oool0DP3oool00`000000oooo0?ooo`2[0?ooo`00H`3oool200000540oooo00<00000
0?ooo`3oool0[03oool006D0oooo0P00001?0?ooo`030000003oool0oooo0:`0oooo001W0?ooo`80
0000C@3oool00`000000oooo0?ooo`2/0?ooo`00J@3oool00`000000oooo0?ooo`1:0?ooo`030000
003oool0oooo0:`0oooo001Z0?ooo`800000B@3oool00`000000oooo0?ooo`2]0?ooo`00K03oool2
000004L0oooo00<000000?ooo`3oool0[@3oool006h0oooo0P0000150?ooo`030000003oool0oooo
0:d0oooo001`0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0[P3oool00740oooo
0P0000110?ooo`030000003oool0oooo0:h0oooo001c0?ooo`800000?`3oool00`000000oooo0?oo
o`2^0?ooo`00M@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo0:h0oooo001f0?oo
o`800000>`3oool00`000000oooo0?ooo`2_0?ooo`00N03oool2000003T0oooo00<000000?ooo`3o
ool0[`3oool007X0oooo0P00000g0?ooo`030000003oool0oooo0:l0oooo001l0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0/03oool007d0oooo0P00000c0?ooo`030000003oool0
oooo0;00oooo001o0?ooo`800000<@3oool00`000000oooo0?ooo`2`0?ooo`00P@3oool2000002l0
oooo00<000000?ooo`3oool0/03oool008<0oooo00<000000?ooo`3oool0:`3oool00`000000oooo
0?ooo`2a0?ooo`00Q03oool2000002/0oooo00<000000?ooo`3oool0/@3oool008H0oooo0P00000Y
0?ooo`030000003oool0oooo0;40oooo00280?ooo`8000009`3oool00`000000oooo0?ooo`2a0?oo
o`00RP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0;80oooo002;0?ooo`800000
8`3oool00`000000oooo0?ooo`2b0?ooo`00S@3oool200000240oooo00<000000?ooo`3oool0/P3o
ool008l0oooo0P00000N0?ooo`030000003oool0oooo0;<0oooo002A0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0/`3oool00980oooo0P00000K0?ooo`030000003oool0oooo0;<0
oooo002D0?ooo`8000006@3oool00`000000oooo0?ooo`2c0?ooo`00UP3oool2000001H0oooo00<0
00000?ooo`3oool0]03oool009P0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`2d
0?ooo`00V@3oool2000001<0oooo00<000000?ooo`3oool0]03oool009/0oooo0P00000@0?ooo`03
0000003oool0oooo0;D0oooo002M0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
]@3oool009h0oooo0P00000=0?ooo`030000003oool0oooo0;D0oooo002P0?ooo`8000002`3oool0
0`000000oooo0?ooo`2e0?ooo`00XP3oool2000000030?ooo`000000oooo00D0oooo00<000000?oo
o`3oool0]P3oool00:@0oooo0`0000050?ooo`030000003oool0oooo0;H0oooo002T0?ooo`@00000
103oool00`000000oooo0?ooo`2f0?ooo`00X`3oool700000080oooo00<000000?ooo`3oool0]P3o
ool00:D0oooo1`00002i0?ooo`00Z@3oool400000;P0oooo002Z0?ooo`030000003oool0oooo00<0
0000]@3oool00:T0oooo0P0000050?ooo`800000/`3oool00:L0oooo0`0000080?ooo`<00000/03o
ool00:H0oooo1000000;0?ooo`<00000[@3oool00:D0oooo1000000?0?ooo`<00000ZP3oool00:H0
oooo0`00000B0?ooo`800000Z03oool00:D0oooo00<000000?ooo`0000005@3oool300000:D0oooo
002T0?ooo`030000003oool0oooo01T0oooo0`00002R0?ooo`00X`3oool00`000000oooo0?ooo`0M
0?ooo`<00000W`3oool00:<0oooo00<000000?ooo`3oool0803oool2000009d0oooo002R0?ooo`03
0000003oool0oooo02<0oooo0`00002J0?ooo`00X@3oool00`000000oooo0?ooo`0W0?ooo`<00000
U`3oool00:00oooo00<000000?ooo`3oool0:`3oool3000009@0oooo002O0?ooo`030000003oool0
oooo02l0oooo0P00002B0?ooo`00WP3oool00`000000oooo0?ooo`0b0?ooo`<00000S`3oool009h0
oooo00<000000?ooo`3oool0=@3oool3000008`0oooo002M0?ooo`030000003oool0oooo03T0oooo
0`0000290?ooo`00W03oool00`000000oooo0?ooo`0m0?ooo`800000Q`3oool009/0oooo00<00000
0?ooo`3oool0@03oool3000008@0oooo002J0?ooo`030000003oool0oooo04@0oooo0`0000210?oo
o`00V@3oool00`000000oooo0?ooo`180?ooo`<00000OP3oool009T0oooo00<000000?ooo`3oool0
B`3oool2000007`0oooo002H0?ooo`030000003oool0oooo04h0oooo0`00001i0?ooo`00U`3oool0
0`000000oooo0?ooo`1B0?ooo`<00000MP3oool009H0oooo00<000000?ooo`3oool0EP3oool30000
07<0oooo002E0?ooo`030000003oool0oooo05X0oooo0P00001a0?ooo`00U03oool00`000000oooo
0?ooo`1M0?ooo`<00000KP3oool009<0oooo00<000000?ooo`3oool0H@3oool3000006/0oooo002C
0?ooo`030000003oool0oooo06@0oooo0`00001X0?ooo`00TP3oool00`000000oooo0?ooo`1X0?oo
o`800000IP3oool00940oooo00<000000?ooo`3oool0J`3oool3000006<0oooo002@0?ooo`030000
003oool0oooo06l0oooo0`00001P0?ooo`00S`3oool00`000000oooo0?ooo`1c0?ooo`<00000G@3o
ool008h0oooo00<000000?ooo`3oool0M`3oool2000005/0oooo002>0?ooo`030000003oool0oooo
07T0oooo0`00001H0?ooo`00S@3oool00`000000oooo0?ooo`1m0?ooo`<00000E@3oool008`0oooo
00<000000?ooo`3oool0P@3oool300000580oooo002;0?ooo`030000003oool0oooo08D0oooo0P00
001@0?ooo`00RP3oool00`000000oooo0?ooo`280?ooo`<00000C@3oool008T0oooo00<000000?oo
o`3oool0S03oool3000004X0oooo00290?ooo`030000003oool0oooo08l0oooo0`0000170?ooo`00
R03oool00`000000oooo0?ooo`2C0?ooo`800000A@3oool008L0oooo00<000000?ooo`3oool0UP3o
ool300000480oooo00260?ooo`030000003oool0oooo09X0oooo0`00000o0?ooo`00Q@3oool00`00
0000oooo0?ooo`2N0?ooo`<00000?03oool008@0oooo00<000000?ooo`3oool0XP3oool2000003X0
oooo00240?ooo`030000003oool0oooo0:@0oooo0`00000g0?ooo`00P`3oool00`000000oooo0?oo
o`2X0?ooo`<00000=03oool00880oooo00<000000?ooo`3oool0[03oool300000340oooo00210?oo
o`030000003oool0oooo0;00oooo0P00000_0?ooo`00P03oool00`000000oooo0?ooo`2c0?ooo`<0
0000;03oool007l0oooo00<000000?ooo`3oool0]`3oool3000002T0oooo001n0?ooo`030000003o
ool0oooo0;/0oooo0`00000V0?ooo`00OP3oool00`000000oooo0?ooo`2n0?ooo`800000903oool0
07d0oooo00<000000?ooo`3oool0`@3oool300000240oooo001l0?ooo`030000003oool0oooo0<D0
oooo0`00000N0?ooo`00N`3oool00`000000oooo0?ooo`390?ooo`<000006`3oool007X0oooo00<0
00000?ooo`3oool0c@3oool2000001T0oooo001i0?ooo`030000003oool0oooo0=00oooo0`00000F
0?ooo`00N@3oool00`000000oooo0?ooo`3C0?ooo`<000000`3oool00`000000oooo0?ooo`0=0?oo
o`00N03oool00`000000oooo0?ooo`3G0?ooo`H000003@3oool007L0oooo00<000000?ooo`3oool0
fP3oool5000000`0oooo001f0?ooo`030000003oool0oooo0=X0oooo2000000:0?ooo`00M@3oool0
0`000000oooo0?ooo`3O0?ooo`D000002@3oool007@0oooo00<000000?ooo`3oool0kP3oool007@0
oooo00<000000?ooo`3oool0kP3oool007<0oooo00<000000?ooo`3oool0k`3oool00780oooo00<0
00000?ooo`3oool0l03oool00740oooo00<000000?ooo`3oool0l@3oool00700oooo00<000000?oo
o`3oool0lP3oool006l0oooo00<000000?ooo`3oool0l`3oool006l0oooo00<000000?ooo`3oool0
l`3oool006h0oooo00<000000?ooo`3oool0m03oool006d0oooo00<000000?ooo`3oool0m@3oool0
06`0oooo00<000000?ooo`3oool0mP3oool006/0oooo00<000000?ooo`3oool0m`3oool006X0oooo
00<000000?ooo`3oool0n03oool006T0oooo00<000000?ooo`3oool0n@3oool006T0oooo00<00000
0?ooo`3oool0n@3oool006P0oooo00<000000?ooo`3oool0nP3oool006L0oooo00<000000?ooo`3o
ool0n`3oool006H0oooo00<000000?ooo`3oool0o03oool006D0oooo00<000000?ooo`3oool0o@3o
ool006@0oooo00<000000?ooo`3oool0oP3oool006@0oooo00<000000?ooo`3oool0oP3oool006<0
oooo00<000000?ooo`3oool0o`3oool00680oooo00<000000?ooo`3oool0o`3oool10?ooo`00H@3o
ool00`000000oooo0?ooo`3o0?ooo`80oooo001P0?ooo`030000003oool0oooo0?l0oooo0`3oool0
05l0oooo00<000000?ooo`3oool0o`3oool40?ooo`00G`3oool00`000000oooo0?ooo`3o0?ooo`@0
oooo001N0?ooo`030000003oool0oooo0?l0oooo1@3oool005d0oooo00<000000?ooo`3oool0o`3o
ool60?ooo`00G03oool00`000000oooo0?ooo`3o0?ooo`L0oooo001K0?ooo`030000003oool0oooo
0?l0oooo203oool005X0oooo00<000000?ooo`3oool0o`3oool90?ooo`00F@3oool00`000000oooo
0?ooo`3o0?ooo`X0oooo001I0?ooo`030000003oool0oooo0?l0oooo2P3oool005P0oooo00<00000
0?ooo`3oool0o`3oool;0?ooo`00E`3oool00`000000oooo0?ooo`3o0?ooo``0oooo001F0?ooo`03
0000003oool0oooo0?l0oooo3@3oool005D0oooo00<000000?ooo`3oool0o`3oool>0?ooo`00E03o
ool00`000000oooo0?ooo`3o0?ooo`l0oooo001D0?ooo`030000003oool0oooo0?l0oooo3`3oool0
05<0oooo00<000000?ooo`3oool0o`3oool@0?ooo`00DP3oool00`000000oooo0?ooo`3o0?oooa40
oooo001A0?ooo`030000003oool0oooo0?l0oooo4P3oool00500oooo00<000000?ooo`3oool0o`3o
oolC0?ooo`00C`3oool00`000000oooo0?ooo`3o0?oooa@0oooo001?0?ooo`030000003oool0oooo
0?l0oooo503oool004h0oooo00<000000?ooo`3oool0o`3ooolE0?ooo`00C@3oool00`000000oooo
0?ooo`3o0?oooaH0oooo001<0?ooo`030000003oool0oooo0?l0oooo5`3oool004/0oooo00<00000
0?ooo`3oool0o`3ooolH0?ooo`00BP3oool00`000000oooo0?ooo`3o0?oooaT0oooo001:0?ooo`03
0000003oool0oooo0?l0oooo6@3oool004T0oooo00<000000?ooo`3oool0o`3ooolJ0?ooo`00B03o
ool00`000000oooo0?ooo`3o0?oooa/0oooo00170?ooo`030000003oool0oooo0?l0oooo703oool0
04H0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`00A@3oool00`000000oooo0?ooo`3o0?oooah0
oooo00140?ooo`030000003oool0oooo0?l0oooo7`3oool004@0oooo00<000000?ooo`3oool0o`3o
oolO0?ooo`00@`3oool00`000000oooo0?ooo`3o0?ooob00oooo00120?ooo`030000003oool0oooo
0?l0oooo8@3oool00440oooo00<000000?ooo`3oool0o`3ooolR0?ooo`00@03oool00`000000oooo
0?ooo`3o0?ooob<0oooo000o0?ooo`030000003oool0oooo0?l0oooo903oool003l0oooo00<00000
0?ooo`3oool0o`3ooolT0?ooo`00?P3oool00`000000oooo0?ooo`3o0?ooobD0oooo000m0?ooo`03
0000003oool0oooo0?l0oooo9P3oool003`0oooo00<000000?ooo`3oool0o`3ooolW0?ooo`00>`3o
ool00`000000oooo0?ooo`3o0?ooobP0oooo000j0?ooo`030000003oool0oooo0?l0oooo:@3oool0
03X0oooo00<000000?ooo`3oool0o`3ooolY0?ooo`00>@3oool00`000000oooo0?ooo`3o0?ooobX0
oooo000h0?ooo`030000003oool0oooo0?l0oooo:`3oool003L0oooo00<000000?ooo`3oool0o`3o
ool/0?ooo`00=P3oool00`000000oooo0?ooo`3o0?ooobd0oooo000e0?ooo`030000003oool0oooo
0?l0oooo;P3oool003D0oooo00<000000?ooo`3oool0o`3oool^0?ooo`00=03oool00`000000oooo
0?ooo`3o0?ooobl0oooo000c0?ooo`030000003oool0oooo0?l0oooo<03oool00380oooo00<00000
0?ooo`3oool0o`3ooola0?ooo`00<@3oool00`000000oooo0?ooo`3o0?oooc80oooo000`0?ooo`03
0000003oool0oooo0?l0oooo<`3oool002l0oooo00<000000?ooo`3oool0o`3ooold0?ooo`00;`3o
ool00`000000oooo0?ooo`3o0?oooc@0oooo000^0?ooo`030000003oool0oooo0?l0oooo=@3oool0
02d0oooo00<000000?ooo`3oool0o`3ooolf0?ooo`00;03oool00`000000oooo0?ooo`3o0?ooocL0
oooo000[0?ooo`030000003oool0oooo0?l0oooo>03oool002X0oooo00<000000?ooo`3oool0o`3o
ooli0?ooo`00:P3oool00`000000oooo0?ooo`3o0?ooocT0oooo000Y0?ooo`030000003oool0oooo
0?l0oooo>P3oool002P0oooo00<000000?ooo`3oool0o`3ooolk0?ooo`009`3oool00`000000oooo
0?ooo`3o0?oooc`0oooo000V0?ooo`030000003oool0oooo0?l0oooo?@3oool002D0oooo00<00000
0?ooo`3oool0o`3oooln0?ooo`009@3oool00`000000oooo0?ooo`3o0?oooch0oooo000T0?ooo`03
0000003oool0oooo0?l0oooo?`3oool002<0oooo00<000000?ooo`3oool0o`3ooom00?ooo`008P3o
ool00`000000oooo0?ooo`3o0?oood40oooo000Q0?ooo`030000003oool0oooo0?l0oooo@P3oool0
0200oooo00<000000?ooo`3oool0o`3ooom30?ooo`00803oool00`000000oooo0?ooo`3o0?oood<0
oooo000O0?ooo`030000003oool0oooo0?l0ooooA03oool001h0oooo00<000000?ooo`3oool0o`3o
oom50?ooo`007@3oool00`000000oooo0?ooo`3o0?ooodH0oooo000L0?ooo`030000003oool0oooo
0?l0ooooA`3oool001/0oooo00<000000?ooo`3oool0o`3ooom80?ooo`006P3oool00`000000oooo
0?ooo`3o0?ooodT0oooo000J0?ooo`030000003oool0oooo0?l0ooooB@3oool001T0oooo00<00000
0?ooo`3oool0o`3ooom:0?ooo`00603oool00`000000oooo0?ooo`3o0?oood/0oooo000G0?ooo`03
0000003oool0oooo0?l0ooooC03oool001H0oooo00<000000?ooo`3oool0o`3ooom=0?ooo`005@3o
ool00`000000oooo0?ooo`3o0?ooodh0oooo000E0?ooo`030000003oool0oooo0?l0ooooCP3oool0
01@0oooo00<000000?ooo`3oool0o`3ooom?0?ooo`004`3oool00`000000oooo0?ooo`3o0?oooe00
oooo000B0?ooo`030000003oool0oooo0?l0ooooD@3oool00140oooo00<000000?ooo`3oool0o`3o
oomB0?ooo`00403oool00`000000oooo0?ooo`3o0?oooe<0oooo000@0?ooo`030000003oool0oooo
0?l0ooooD`3oool000l0oooo00<000000?ooo`3oool0o`3ooomD0?ooo`003P3oool00`000000oooo
0?ooo`3o0?oooeD0oooo000=0?ooo`030000003oool0oooo0?l0ooooEP3oool000`0oooo00<00000
0?ooo`3oool0o`3ooomG0?ooo`002`3oool00`000000oooo0?ooo`3o0?oooeP0oooo000;0?ooo`03
0000003oool0oooo0?l0ooooF03oool000X0oooo00<000000?ooo`3oool0o`3ooomI0?ooo`002@3o
ool00`000000oooo0?ooo`3o0?oooeX0oooo00080?ooo`030000003oool0oooo0?l0ooooF`3oool0
0?l0ooooIP3oool00?l0ooooIP3oool00?l0ooooIP3oool00?l0ooooIP3oool00?l0ooooIP3oool0
0?l0ooooIP3oool00?l0ooooIP3oool00?l0ooooIP3oool00?l0ooooIP3oool00?l0ooooIP3oool0
0001\
\>"],
  ImageRangeCache->{{{0, 356}, {430.875, 0}} -> {-2.02869, -2.67873, \
0.017751, 0.017751}}]
}, Closed]],

Cell[BoxData[
    \(\(VelocityPlot2[LabVelocities, vCM, cmVelocities];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[RGBColor[0, 1, 0]], plot1, Graphics[RGBColor[1, 0, 0]], 
        plot2, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.52246 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.493881 1.29891 0.493881 [
[ 0 0 0 0 ]
[ 1 1.52246 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.52246 L
0 1.52246 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 1 0 r
.5 Mabswid
[ ] 0 setdash
.02381 1.29891 m
.23767 1.42238 L
s
% Start of user PostScript
0. 0. 0.4330127 0.25 MAarrowhead1
% End of user PostScript
.02381 1.29891 m
.97619 .1639 L
s
% Start of user PostScript
0. 0. 1.9283628 -2.2981333 MAarrowhead1
% End of user PostScript
.02381 1.29891 m
.48384 1.00289 L
s
% Start of user PostScript
0. 0. 0.93146274 -0.59937778 MAarrowhead1
% End of user PostScript
.48384 1.00289 m
.23767 1.42238 L
s
% Start of user PostScript
0.93146274 -0.59937778 0.4330127 0.25 MAarrowhead1
% End of user PostScript
.48384 1.00289 m
.97619 .1639 L
s
% Start of user PostScript
0.93146274 -0.59937778 1.9283628 -2.2981333 MAarrowhead1
% End of user PostScript
1 0 0 r
.02381 1.29891 m
.53841 1.48621 L
s
% Start of user PostScript
0. 0. 1.041947 0.37923771 MAarrowhead1
% End of user PostScript
.02381 1.29891 m
.37471 .03625 L
s
% Start of user PostScript
0. 0. 0.71049414 -2.5566088 MAarrowhead1
% End of user PostScript
.02381 1.29891 m
.48384 1.00289 L
s
% Start of user PostScript
0. 0. 0.93146274 -0.59937778 MAarrowhead1
% End of user PostScript
.48384 1.00289 m
.53841 1.48621 L
s
% Start of user PostScript
0.93146274 -0.59937778 1.041947 0.37923771 MAarrowhead1
% End of user PostScript
.48384 1.00289 m
.37471 .03625 L
s
% Start of user PostScript
0.93146274 -0.59937778 0.71049414 -2.5566088 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{189.125, 287.875},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004K0001[b000`400?l00000o`00003oo`3ooolL
0?ooo`00o`3ooolL0?ooo`00o`3ooolL0?ooo`00o`3ooolL0?ooo`00o`3ooolL0?ooo`00o`3ooolL
0?ooo`00o`3ooolL0?ooo`00o`3ooolL0?ooo`00o`3ooolL0?ooo`00o`3ooolL0?ooo`00J@3oool0
0`3o0000oooo0?ooo`2_0?ooo`00J@3oool00`3o0000oooo0?ooo`2_0?ooo`00J03oool30?l00;00
oooo001X0?ooo`<0o`00/03oool006L0oooo1@3o002_0?ooo`00I`3oool50?l00:l0oooo001V0?oo
o`L0o`00[P3oool006H0oooo1`3o002^0?ooo`00I`3oool40?l00;00oooo001W0?ooo`040?l0003o
ool0oooo0?l00;00oooo001V0?ooo`050?l0003oool0oooo0?ooo`3o0000/03oool006H0oooo00D0
o`000?ooo`3oool0oooo0?l0002`0?ooo`00IP3oool01@3o0000oooo0?ooo`3oool0o`000;00oooo
001U0?ooo`030?l0003oool0oooo0080oooo00<0o`000?ooo`3oool0[P3oool006D0oooo00<0o`00
0?ooo`3oool00`3oool00`3o0000oooo0?ooo`2]0?ooo`00I@3oool00`3o0000oooo0?ooo`030?oo
o`030?l0003oool0oooo0:d0oooo001U0?ooo`030?l0003oool0oooo00<0oooo00<0o`000?ooo`3o
ool0[@3oool006@0oooo00<0o`000?ooo`3oool0103oool00`3o0000oooo0?ooo`2]0?ooo`00I03o
ool00`3o0000oooo0?ooo`040?ooo`030?l0003oool0oooo0:d0oooo001T0?ooo`030?l0003oool0
oooo00@0oooo00<0o`000?ooo`3oool0[@3oool006<0oooo00<0o`000?ooo`3oool01@3oool00`3o
0000oooo0?ooo`2]0?ooo`00H`3oool00`3o0000oooo0?ooo`050?ooo`030?l0003oool0oooo0:d0
oooo001S0?ooo`030?l0003oool0oooo00D0oooo00<0o`000?ooo`3oool0[@3oool006<0oooo00<0
o`000?ooo`3oool01P3oool00`3o0000oooo0?ooo`2/0?ooo`00HP3oool00`3o0000oooo0?ooo`07
0?ooo`030?l0003oool0oooo0:`0oooo001R0?ooo`030?l0003oool0oooo00L0oooo00<0o`000?oo
o`3oool0[03oool00680oooo00<0o`000?ooo`3oool01`3oool00`3o0000oooo0?ooo`2/0?ooo`00
HP3oool00`3o0000oooo0?ooo`070?ooo`030?l0003oool0oooo0:`0oooo001Q0?ooo`030?l0003o
ool0oooo00P0oooo00<0o`000?ooo`3oool0[03oool00640oooo00<0o`000?ooo`3oool0203oool0
0`3o0000oooo0?ooo`2/0?ooo`00H@3oool00`3o0000oooo0?ooo`080?ooo`030?l0003oool0oooo
0:`0oooo001P0?ooo`030?l0003oool0oooo00X0oooo00<0o`000?ooo`3oool0Z`3oool00600oooo
00<0o`000?ooo`3oool02P3oool00`3o0000oooo0?ooo`2[0?ooo`00H03oool00`3o0000oooo0?oo
o`0:0?ooo`030?l0003oool0oooo0:/0oooo001P0?ooo`030?l0003oool0oooo00X0oooo00<0o`00
0?ooo`3oool0Z`3oool005l0oooo00<0o`000?ooo`3oool02`3oool00`3o0000oooo0?ooo`2[0?oo
o`00G`3oool00`3o0000oooo0?ooo`0;0?ooo`030?l0003oool0oooo0:<0oooo00<00?l00?ooo`3o
ool01@3oool005l0oooo00<0o`000?ooo`3oool02`3oool00`3o0000oooo0?ooo`2R0?ooo`800?l0
1`3oool005h0oooo00<0o`000?ooo`3oool0303oool00`3o0000oooo0?ooo`2Q0?ooo`800?l0203o
ool005h0oooo00<0o`000?ooo`3oool0303oool00`3o0000oooo0?ooo`2O0?ooo`@00?l0203oool0
05h0oooo00<0o`000?ooo`3oool03@3oool00`3o0000oooo0?ooo`2M0?ooo`D00?l0203oool005h0
oooo00<0o`000?ooo`3oool03@3oool00`3o0000oooo0?ooo`2L0?ooo`H00?l0203oool005d0oooo
00<0o`000?ooo`3oool03P3oool00`3o0000oooo0?ooo`2M0?ooo`@00?l02@3oool005d0oooo00<0
o`000?ooo`3oool03P3oool00`3o0000oooo0?ooo`2L0?ooo`03003o003oool00?l000800?l02@3o
ool005d0oooo00<0o`000?ooo`3oool03P3oool00`3o0000oooo0?ooo`2K0?ooo`03003o003oool0
0?l00080oooo00<00?l00?ooo`3oool01`3oool005d0oooo00<0o`000?ooo`3oool03P3oool00`3o
0000oooo0?ooo`2J0?ooo`04003o003oool0oooo003o00`0oooo001L0?ooo`030?l0003oool0oooo
00l0oooo00<0o`000?ooo`3oool0VP3oool00`00o`00oooo003o000=0?ooo`00G03oool00`3o0000
oooo0?ooo`0?0?ooo`030?l0003oool0oooo09T0oooo00@00?l00?ooo`3oool00?l03@3oool005`0
oooo00<0o`000?ooo`3oool03`3oool00`3o0000oooo0?ooo`2H0?ooo`04003o003oool0oooo003o
00h0oooo001K0?ooo`030?l0003oool0oooo0140oooo00<0o`000?ooo`3oool0UP3oool01000o`00
oooo0?ooo`00o`0?0?ooo`00F`3oool00`3o0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo09D0
oooo00D00?l00?ooo`3oool0oooo003o000?0?ooo`00F`3oool00`3o0000oooo0?ooo`0A0?ooo`03
0?l0003oool0oooo09@0oooo00D00?l00?ooo`3oool0oooo003o000@0?ooo`00F`3oool00`3o0000
oooo0?ooo`0A0?ooo`030?l0003oool0oooo09@0oooo00D00?l00?ooo`3oool0oooo003o000@0?oo
o`00FP3oool00`3o0000oooo0?ooo`0B0?ooo`030?l0003oool0oooo09<0oooo00D00?l00?ooo`3o
ool0oooo003o000A0?ooo`00FP3oool00`3o0000oooo0?ooo`0B0?ooo`030?l0003oool0oooo0980
oooo00D00?l00?ooo`3oool0oooo003o000B0?ooo`00FP3oool00`3o0000oooo0?ooo`0B0?ooo`03
0?l0003oool0oooo0940oooo00<00?l00?ooo`3oool00P3oool00`00o`00oooo0?ooo`0@0?ooo`00
F@3oool00`3o0000oooo0?ooo`0C0?ooo`030?l0003oool0oooo0900oooo00<00?l00?ooo`3oool0
0P3oool00`00o`00oooo0?ooo`0A0?ooo`00F@3oool00`3o0000oooo0?ooo`0C0?ooo`030?l0003o
ool0oooo08l0oooo00<00?l00?ooo`3oool00`3oool00`00o`00oooo0?ooo`0A0?ooo`00F@3oool0
0`3o0000oooo0?ooo`0D0?ooo`030?l0003oool0oooo08h0oooo00<00?l00?ooo`3oool00P3oool0
0`00o`00oooo0?ooo`0B0?ooo`00F@3oool00`3o0000oooo0?ooo`0D0?ooo`030?l0003oool0oooo
08d0oooo00<00?l00?ooo`3oool00`3oool00`00o`00oooo0?ooo`0B0?ooo`00F03oool00`3o0000
oooo0?ooo`0E0?ooo`030?l0003oool0oooo08`0oooo00<00?l00?ooo`3oool00`3oool00`00o`00
oooo0?ooo`0C0?ooo`00F03oool00`3o0000oooo0?ooo`0E0?ooo`030?l0003oool0oooo08/0oooo
00<00?l00?ooo`3oool00`3oool00`00o`00oooo0?ooo`0D0?ooo`00F03oool00`3o0000oooo0?oo
o`0E0?ooo`030?l0003oool0oooo08X0oooo00<00?l00?ooo`3oool0103oool00`00o`00oooo0?oo
o`0D0?ooo`00F03oool00`3o0000oooo0?ooo`0E0?ooo`030?l0003oool0oooo08T0oooo00<00?l0
0?ooo`3oool0103oool00`00o`00oooo0?ooo`0E0?ooo`00E`3oool00`3o0000oooo0?ooo`0F0?oo
o`030?l0003oool0oooo08T0oooo00<00?l00?ooo`3oool0103oool00`00o`00oooo0?ooo`0E0?oo
o`00E`3oool00`3o0000oooo0?ooo`0F0?ooo`030?l0003oool0oooo08P0oooo00<00?l00?ooo`3o
ool0103oool00`00o`00oooo0?ooo`0F0?ooo`00E`3oool00`3o0000oooo0?ooo`0F0?ooo`030?l0
003oool0oooo08L0oooo00<00?l00?ooo`3oool0103oool00`00o`00oooo0?ooo`0G0?ooo`00EP3o
ool00`3o0000oooo0?ooo`0H0?ooo`030?l0003oool0oooo08D0oooo00<00?l00?ooo`3oool01@3o
ool00`00o`00oooo0?ooo`0G0?ooo`00EP3oool00`3o0000oooo0?ooo`0H0?ooo`030?l0003oool0
oooo08@0oooo00<00?l00?ooo`3oool01@3oool00`00o`00oooo0?ooo`0H0?ooo`00EP3oool00`3o
0000oooo0?ooo`0H0?ooo`030?l0003oool0oooo08<0oooo00<00?l00?ooo`3oool01P3oool00`00
o`00oooo0?ooo`0H0?ooo`00EP3oool00`3o0000oooo0?ooo`0H0?ooo`030?l0003oool0oooo08<0
oooo00<00?l00?ooo`3oool01@3oool00`00o`00oooo0?ooo`0I0?ooo`00E@3oool00`3o0000oooo
0?ooo`0I0?ooo`030?l0003oool0oooo0880oooo00<00?l00?ooo`3oool01@3oool00`00o`00oooo
0?ooo`0J0?ooo`00E@3oool00`3o0000oooo0?ooo`0I0?ooo`030?l0003oool0oooo0840oooo00<0
0?l00?ooo`3oool01P3oool00`00o`00oooo0?ooo`0J0?ooo`00E@3oool00`3o0000oooo0?ooo`0I
0?ooo`030?l0003oool0oooo0800oooo00<00?l00?ooo`3oool01P3oool00`00o`00oooo0?ooo`0K
0?ooo`00E03oool00`3o0000oooo0?ooo`0J0?ooo`030?l0003oool0oooo07l0oooo00<00?l00?oo
o`3oool01`3oool00`00o`00oooo0?ooo`0K0?ooo`00E03oool00`3o0000oooo0?ooo`0K0?ooo`03
0?l0003oool0oooo07d0oooo00<00?l00?ooo`3oool01`3oool00`00o`00oooo0?ooo`0L0?ooo`00
E03oool00`3o0000oooo0?ooo`0K0?ooo`030?l0003oool0oooo07`0oooo00<00?l00?ooo`3oool0
203oool00`00o`00oooo0?ooo`0L0?ooo`00E03oool00`3o0000oooo0?ooo`0K0?ooo`030?l0003o
ool0oooo07`0oooo00<00?l00?ooo`3oool01`3oool00`00o`00oooo0?ooo`0M0?ooo`00D`3oool0
0`3o0000oooo0?ooo`0L0?ooo`030?l0003oool0oooo07/0oooo00<00?l00?ooo`3oool01`3oool0
0`00o`00oooo0?ooo`0N0?ooo`00D`3oool00`3o0000oooo0?ooo`0L0?ooo`030?l0003oool0oooo
07X0oooo00<00?l00?ooo`3oool0203oool00`00o`00oooo0?ooo`0N0?ooo`00D`3oool00`3o0000
oooo0?ooo`0L0?ooo`030?l0003oool0oooo07T0oooo00<00?l00?ooo`3oool0203oool00`00o`00
oooo0?ooo`0O0?ooo`00D`3oool00`3o0000oooo0?ooo`0L0?ooo`030?l0003oool0oooo07P0oooo
00<00?l00?ooo`3oool02@3oool00`00o`00oooo0?ooo`0O0?ooo`00DP3oool00`3o0000oooo0?oo
o`0M0?ooo`030?l0003oool0oooo07L0oooo00<00?l00?ooo`3oool02@3oool00`00o`00oooo0?oo
o`0P0?ooo`00DP3oool00`3o0000oooo0?ooo`0M0?ooo`030?l0003oool0oooo07L0oooo00<00?l0
0?ooo`3oool0203oool00`00o`00oooo0?ooo`0Q0?ooo`00DP3oool00`3o0000oooo0?ooo`0N0?oo
o`030?l0003oool0oooo07D0oooo00<00?l00?ooo`3oool02@3oool00`00o`00oooo0?ooo`0Q0?oo
o`00D@3oool00`3o0000oooo0?ooo`0O0?ooo`030?l0003oool0oooo07@0oooo00<00?l00?ooo`3o
ool02@3oool00`00o`00oooo0?ooo`0R0?ooo`00D@3oool00`3o0000oooo0?ooo`0O0?ooo`030?l0
003oool0oooo07<0oooo00<00?l00?ooo`3oool02P3oool00`00o`00oooo0?ooo`0R0?ooo`00D@3o
ool00`3o0000oooo0?ooo`0O0?ooo`030?l0003oool0oooo0780oooo00<00?l00?ooo`3oool02P3o
ool00`00o`00oooo0?ooo`0S0?ooo`00D@3oool00`3o0000oooo0?ooo`0O0?ooo`030?l0003oool0
oooo0740oooo00<00?l00?ooo`3oool02`3oool00`00o`00oooo0?ooo`0S0?ooo`00D03oool00`3o
0000oooo0?ooo`0P0?ooo`030?l0003oool0oooo0740oooo00<00?l00?ooo`3oool02P3oool00`00
o`00oooo0?ooo`0T0?ooo`00D03oool00`3o0000oooo0?ooo`0P0?ooo`030?l0003oool0oooo0700
oooo00<00?l00?ooo`3oool02P3oool00`00o`00oooo0?ooo`0U0?ooo`00D03oool00`3o0000oooo
0?ooo`0P0?ooo`030?l0003oool0oooo06l0oooo00<00?l00?ooo`3oool02`3oool00`00o`00oooo
0?ooo`0U0?ooo`00C`3oool00`3o0000oooo0?ooo`0Q0?ooo`030?l0003oool0oooo06h0oooo00<0
0?l00?ooo`3oool02`3oool00`00o`00oooo0?ooo`0V0?ooo`00C`3oool00`3o0000oooo0?ooo`0R
0?ooo`030?l0003oool0oooo06`0oooo00<00?l00?ooo`3oool0303oool00`00o`00oooo0?ooo`0V
0?ooo`00C`3oool00`3o0000oooo0?ooo`0R0?ooo`030?l0003oool0oooo06/0oooo00<00?l00?oo
o`3oool0303oool00`00o`00oooo0?ooo`0W0?ooo`00C`3oool00`3o0000oooo0?ooo`0R0?ooo`03
0?l0003oool0oooo06/0oooo00<00?l00?ooo`3oool02`3oool00`00o`00oooo0?ooo`0X0?ooo`00
CP3oool00`3o0000oooo0?ooo`0S0?ooo`030?l0003oool0oooo06X0oooo00<00?l00?ooo`3oool0
303oool00`00o`00oooo0?ooo`0X0?ooo`00CP3oool00`3o0000oooo0?ooo`0S0?ooo`030?l0003o
ool0oooo06T0oooo00<00?l00?ooo`3oool0303oool00`00o`00oooo0?ooo`0Y0?ooo`00CP3oool0
0`3o0000oooo0?ooo`0S0?ooo`030?l0003oool0oooo06P0oooo00<00?l00?ooo`3oool03@3oool0
0`00o`00oooo0?ooo`0Y0?ooo`00CP3oool00`3o0000oooo0?ooo`0S0?ooo`030?l0003oool0oooo
06L0oooo00<00?l00?ooo`3oool03@3oool00`00o`00oooo0?ooo`0Z0?ooo`00C@3oool00`3o0000
oooo0?ooo`0T0?ooo`030?l0003oool0oooo06H0oooo00<00?l00?ooo`3oool03@3oool00`00o`00
oooo0?ooo`0[0?ooo`00C@3oool00`3o0000oooo0?ooo`0T0?ooo`030?l0003oool0oooo06H0oooo
00<00?l00?ooo`3oool03@3oool00`00o`00oooo0?ooo`0[0?ooo`00C@3oool00`3o0000oooo0?oo
o`0U0?ooo`030?l0003oool0oooo06@0oooo00<00?l00?ooo`3oool03@3oool00`00o`00oooo0?oo
o`0/0?ooo`00C03oool00`3o0000oooo0?ooo`0V0?ooo`030?l0003oool0oooo06<0oooo00<00?l0
0?ooo`3oool03P3oool00`00o`00oooo0?ooo`0/0?ooo`00C03oool00`3o0000oooo0?ooo`0V0?oo
o`030?l0003oool0oooo0680oooo00<00?l00?ooo`3oool03P3oool00`00o`00oooo0?ooo`0]0?oo
o`00C03oool00`3o0000oooo0?ooo`0V0?ooo`030?l0003oool0oooo0640oooo00<00?l00?ooo`3o
ool03`3oool00`00o`00oooo0?ooo`0]0?ooo`00C03oool00`3o0000oooo0?ooo`0V0?ooo`030?l0
003oool0oooo0600oooo00<00?l00?ooo`3oool03`3oool00`00o`00oooo0?ooo`0^0?ooo`00B`3o
ool00`3o0000oooo0?ooo`0W0?ooo`030?l0003oool0oooo0600oooo00<00?l00?ooo`3oool03P3o
ool00`00o`00oooo0?ooo`0_0?ooo`00B`3oool00`3o0000oooo0?ooo`0W0?ooo`030?l0003oool0
oooo05l0oooo00<00?l00?ooo`3oool03`3oool00`00o`00oooo0?ooo`0_0?ooo`00B`3oool00`3o
0000oooo0?ooo`0W0?ooo`030?l0003oool0oooo05h0oooo00<00?l00?ooo`3oool03`3oool00`00
o`00oooo0?ooo`0`0?ooo`00BP3oool00`3o0000oooo0?ooo`0X0?ooo`030?l0003oool0oooo05d0
oooo00<00?l00?ooo`3oool0403oool00`00o`00oooo0?ooo`0`0?ooo`00BP3oool00`3o0000oooo
0?ooo`0Y0?ooo`030?l0003oool0oooo05/0oooo00<00?l00?ooo`3oool0403oool00`00o`00oooo
0?ooo`0a0?ooo`00BP3oool00`3o0000oooo0?ooo`0Y0?ooo`030?l0003oool0oooo05X0oooo00<0
0?l00?ooo`3oool0403oool00`00o`00oooo0?ooo`0b0?ooo`00BP3oool00`3o0000oooo0?ooo`0Y
0?ooo`030?l0003oool0oooo05T0oooo00<00?l00?ooo`3oool04@3oool00`00o`00oooo0?ooo`0b
0?ooo`00B@3oool00`3o0000oooo0?ooo`0Z0?ooo`030?l0003oool0oooo05T0oooo00<00?l00?oo
o`3oool0403oool00`00o`00oooo0?ooo`0c0?ooo`00B@3oool00`3o0000oooo0?ooo`0Z0?ooo`03
0?l0003oool0oooo05P0oooo00<00?l00?ooo`3oool04@3oool00`00o`00oooo0?ooo`0c0?ooo`00
B@3oool00`3o0000oooo0?ooo`0Z0?ooo`030?l0003oool0oooo05L0oooo00<00?l00?ooo`3oool0
4@3oool00`00o`00oooo0?ooo`0d0?ooo`00B@3oool00`3o0000oooo0?ooo`0Z0?ooo`030?l0003o
ool0oooo05H0oooo00<00?l00?ooo`3oool04@3oool00`00o`00oooo0?ooo`0e0?ooo`00B03oool0
0`3o0000oooo0?ooo`0[0?ooo`030?l0003oool0oooo05D0oooo00<00?l00?ooo`3oool04P3oool0
0`00o`00oooo0?ooo`0e0?ooo`00B03oool00`3o0000oooo0?ooo`0/0?ooo`030?l0003oool0oooo
05<0oooo00<00?l00?ooo`3oool04P3oool00`00o`00oooo0?ooo`0f0?ooo`00B03oool00`3o0000
oooo0?ooo`0/0?ooo`030?l0003oool0oooo05<0oooo00<00?l00?ooo`3oool04P3oool00`00o`00
oooo0?ooo`0f0?ooo`00A`3oool00`3o0000oooo0?ooo`0]0?ooo`030?l0003oool0oooo0580oooo
00<00?l00?ooo`3oool04P3oool00`00o`00oooo0?ooo`0g0?ooo`00A`3oool00`3o0000oooo0?oo
o`0]0?ooo`030?l0003oool0oooo0540oooo00<00?l00?ooo`3oool04`3oool00`00o`00oooo0?oo
o`0g0?ooo`00A`3oool00`3o0000oooo0?ooo`0]0?ooo`030?l0003oool0oooo0500oooo00<00?l0
0?ooo`3oool04`3oool00`00o`00oooo0?ooo`0h0?ooo`00A`3oool00`3o0000oooo0?ooo`0]0?oo
o`030?l0003oool0oooo04l0oooo00<00?l00?ooo`3oool04`3oool00`00o`00oooo0?ooo`0i0?oo
o`00AP3oool00`3o0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo04h0oooo00<00?l00?ooo`3o
ool0503oool00`00o`00oooo0?ooo`0i0?ooo`00AP3oool00`3o0000oooo0?ooo`0^0?ooo`030?l0
003oool0oooo04h0oooo00<00?l00?ooo`3oool04`3oool00`00o`00oooo0?ooo`0j0?ooo`00AP3o
ool00`3o0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo04d0oooo00<00?l00?ooo`3oool0503o
ool00`00o`00oooo0?ooo`0j0?ooo`00AP3oool00`3o0000oooo0?ooo`0_0?ooo`030?l0003oool0
oooo04/0oooo00<00?l00?ooo`3oool0503oool00`00o`00oooo0?ooo`0k0?ooo`00A@3oool00`3o
0000oooo0?ooo`0`0?ooo`030?l0003oool0oooo04X0oooo00<00?l00?ooo`3oool0503oool00`00
o`00oooo0?ooo`0l0?ooo`00A@3oool00`3o0000oooo0?ooo`0`0?ooo`030?l0003oool0oooo04T0
oooo00<00?l00?ooo`3oool05@3oool00`00o`00oooo0?ooo`0l0?ooo`00A@3oool00`3o0000oooo
0?ooo`0`0?ooo`030?l0003oool0oooo04P0oooo00<00?l00?ooo`3oool05@3oool00`00o`00oooo
0?ooo`0m0?ooo`00A03oool00`3o0000oooo0?ooo`0a0?ooo`030?l0003oool0oooo04P0oooo00<0
0?l00?ooo`3oool05@3oool00`00o`00oooo0?ooo`0m0?ooo`00A03oool00`3o0000oooo0?ooo`0a
0?ooo`030?l0003oool0oooo04L0oooo00<00?l00?ooo`3oool05@3oool00`00o`00oooo0?ooo`0n
0?ooo`00A03oool00`3o0000oooo0?ooo`0a0?ooo`030?l0003oool0oooo04H0oooo00<00?l00?oo
o`3oool05P3oool00`00o`00oooo0?ooo`0n0?ooo`00A03oool00`3o0000oooo0?ooo`0a0?ooo`03
0?l0003oool0oooo04D0oooo00<00?l00?ooo`3oool05P3oool00`00o`00oooo0?ooo`0o0?ooo`00
@`3oool00`3o0000oooo0?ooo`0b0?ooo`030?l0003oool0oooo04@0oooo00<00?l00?ooo`3oool0
5P3oool00`00o`00oooo0?ooo`100?ooo`00@`3oool00`3o0000oooo0?ooo`0c0?ooo`030?l0003o
ool0oooo0480oooo00<00?l00?ooo`3oool05`3oool00`00o`00oooo0?ooo`100?ooo`00@`3oool0
0`3o0000oooo0?ooo`0c0?ooo`030?l0003oool0oooo0480oooo00<00?l00?ooo`3oool05P3oool0
0`00o`00oooo0?ooo`110?ooo`00@P3oool00`3o0000oooo0?ooo`0d0?ooo`030?l0003oool0oooo
0440oooo00<00?l00?ooo`3oool05`3oool00`00o`00oooo0?ooo`110?ooo`00@P3oool00`3o0000
oooo0?ooo`0d0?ooo`030?l0003oool0oooo0400oooo00<00?l00?ooo`3oool05`3oool00`00o`00
oooo0?ooo`120?ooo`00@P3oool00`3o0000oooo0?ooo`0d0?ooo`030?l0003oool0oooo03l0oooo
00<00?l00?ooo`3oool05`3oool00`00o`00oooo0?ooo`130?ooo`00@P3oool00`3o0000oooo0?oo
o`0d0?ooo`030?l0003oool0oooo03h0oooo00<00?l00?ooo`3oool0603oool00`00o`00oooo0?oo
o`130?ooo`00@@3oool00`3o0000oooo0?ooo`0e0?ooo`030?l0003oool0oooo03d0oooo00<00?l0
0?ooo`3oool0603oool00`00o`00oooo0?ooo`140?ooo`00@@3oool00`3o0000oooo0?ooo`0e0?oo
o`030?l0003oool0oooo03`0oooo00<00?l00?ooo`3oool06@3oool00`00o`00oooo0?ooo`140?oo
o`00@@3oool00`3o0000oooo0?ooo`0e0?ooo`030?l0003oool0oooo03`0oooo00<00?l00?ooo`3o
ool0603oool00`00o`00oooo0?ooo`150?ooo`00@@3oool00`3o0000oooo0?ooo`0f0?ooo`030?l0
003oool0oooo03X0oooo00<00?l00?ooo`3oool0603oool00`00o`00oooo0?ooo`160?ooo`00@03o
ool00`3o0000oooo0?ooo`0g0?ooo`030?l0003oool0oooo03T0oooo00<00?l00?ooo`3oool06@3o
ool00`00o`00oooo0?ooo`160?ooo`00@03oool00`3o0000oooo0?ooo`0g0?ooo`030?l0003oool0
oooo03P0oooo00<00?l00?ooo`3oool06@3oool00`00o`00oooo0?ooo`170?ooo`00@03oool00`3o
0000oooo0?ooo`0g0?ooo`030?l0003oool0oooo03L0oooo00<00?l00?ooo`3oool06P3oool00`00
o`00oooo0?ooo`170?ooo`00?`3oool00`3o0000oooo0?ooo`0h0?ooo`030?l0003oool0oooo03H0
oooo00<00?l00?ooo`3oool06P3oool00`00o`00oooo0?ooo`180?ooo`00?`3oool00`3o0000oooo
0?ooo`0h0?ooo`030?l0003oool0oooo03H0oooo00<00?l00?ooo`3oool06P3oool00`00o`00oooo
0?ooo`180?ooo`00?`3oool00`3o0000oooo0?ooo`0h0?ooo`030?l0003oool0oooo03D0oooo00<0
0?l00?ooo`3oool06P3oool00`00o`00oooo0?ooo`190?ooo`00?`3oool00`3o0000oooo0?ooo`0h
0?ooo`030?l0003oool0oooo03@0oooo00<00?l00?ooo`3oool06P3oool00`00o`00oooo0?ooo`1:
0?ooo`00?P3oool00`3o0000oooo0?ooo`0i0?ooo`030?l0003oool0oooo03<0oooo00<00?l00?oo
o`3oool06`3oool00`00o`00oooo0?ooo`1:0?ooo`00?P3oool00`3o0000oooo0?ooo`0j0?ooo`03
0?l0003oool0oooo0340oooo00<00?l00?ooo`3oool06`3oool00`00o`00oooo0?ooo`1;0?ooo`00
?P3oool00`3o0000oooo0?ooo`0j0?ooo`030?l0003oool0oooo0300oooo00<00?l00?ooo`3oool0
703oool00`00o`00oooo0?ooo`1;0?ooo`00?@3oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003o
ool0oooo0300oooo00<00?l00?ooo`3oool06`3oool00`00o`00oooo0?ooo`1<0?ooo`00?@3oool0
0`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo02l0oooo00<00?l00?ooo`3oool06`3oool0
0`00o`00oooo0?ooo`1=0?ooo`00?@3oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo
02h0oooo00<00?l00?ooo`3oool0703oool00`00o`00oooo0?ooo`1=0?ooo`00?@3oool00`3o0000
oooo0?ooo`0k0?ooo`030?l0003oool0oooo02d0oooo00<00?l00?ooo`3oool0703oool00`00o`00
oooo0?ooo`1>0?ooo`00?03oool00`3o0000oooo0?ooo`0l0?ooo`030?l0003oool0oooo02`0oooo
00<00?l00?ooo`3oool07@3oool00`00o`00oooo0?ooo`1>0?ooo`00?03oool00`3o0000oooo0?oo
o`0l0?ooo`030?l0003oool0oooo02/0oooo00<00?l00?ooo`3oool07@3oool00`00o`00oooo0?oo
o`1?0?ooo`00?03oool00`3o0000oooo0?ooo`0m0?ooo`030?l0003oool0oooo02X0oooo00<00?l0
0?ooo`3oool07@3oool00`00o`00oooo0?ooo`1?0?ooo`00?03oool00`3o0000oooo0?ooo`0m0?oo
o`030?l0003oool0oooo02T0oooo00<00?l00?ooo`3oool07@3oool00`00o`00oooo0?ooo`1@0?oo
o`00>`3oool00`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo02P0oooo00<00?l00?ooo`3o
ool07@3oool00`00o`00oooo0?ooo`1A0?ooo`00>`3oool00`3o0000oooo0?ooo`0n0?ooo`030?l0
003oool0oooo02L0oooo00<00?l00?ooo`3oool07P3oool00`00o`00oooo0?ooo`1A0?ooo`00>`3o
ool00`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo02H0oooo00<00?l00?ooo`3oool07P3o
ool00`00o`00oooo0?ooo`1B0?ooo`00>P3oool00`3o0000oooo0?ooo`0o0?ooo`030?l0003oool0
oooo02D0oooo00<00?l00?ooo`3oool07`3oool00`00o`00oooo0?ooo`1B0?ooo`00>P3oool00`3o
0000oooo0?ooo`0o0?ooo`030?l0003oool0oooo02D0oooo00<00?l00?ooo`3oool07P3oool00`00
o`00oooo0?ooo`1C0?ooo`00>P3oool00`3o0000oooo0?ooo`0o0?ooo`030?l0003oool0oooo02@0
oooo00<00?l00?ooo`3oool07P3oool00`00o`00oooo0?ooo`1D0?ooo`00>P3oool00`3o0000oooo
0?ooo`0o0?ooo`030?l0003oool0oooo02<0oooo00<00?l00?ooo`3oool07`3oool00`00o`00oooo
0?ooo`1D0?ooo`00>@3oool00`3o0000oooo0?ooo`110?ooo`030?l0003oool0oooo0240oooo00<0
0?l00?ooo`3oool07`3oool00`00o`00oooo0?ooo`1E0?ooo`00>@3oool00`3o0000oooo0?ooo`11
0?ooo`030?l0003oool0oooo0200oooo00<00?l00?ooo`3oool0803oool00`00o`00oooo0?ooo`1E
0?ooo`00>@3oool00`3o0000oooo0?ooo`110?ooo`030?l0003oool0oooo01l0oooo00<00?l00?oo
o`3oool0803oool00`00o`00oooo0?ooo`1F0?ooo`00>03oool00`3o0000oooo0?ooo`120?ooo`03
0?l0003oool0oooo01l0oooo00<00?l00?ooo`3oool07`3oool00`00o`00oooo0?ooo`1G0?ooo`00
>03oool00`3o0000oooo0?ooo`120?ooo`030?l0003oool0oooo01h0oooo00<00?l00?ooo`3oool0
803oool00`00o`00oooo0?ooo`1G0?ooo`00>03oool00`3o0000oooo0?ooo`120?ooo`030?l0003o
ool0oooo01d0oooo00<00?l00?ooo`3oool0803oool00`00o`00oooo0?ooo`1H0?ooo`00>03oool0
0`3o0000oooo0?ooo`120?ooo`030?l0003oool0oooo01`0oooo00<00?l00?ooo`3oool08@3oool0
0`00o`00oooo0?ooo`1H0?ooo`00=`3oool00`3o0000oooo0?ooo`130?ooo`030?l0003oool0oooo
01/0oooo00<00?l00?ooo`3oool08@3oool00`00o`00oooo0?ooo`1I0?ooo`00=`3oool00`3o0000
oooo0?ooo`130?ooo`030?l0003oool0oooo01X0oooo00<00?l00?ooo`3oool08P3oool00`00o`00
oooo0?ooo`1I0?ooo`00=`3oool00`3o0000oooo0?ooo`140?ooo`030?l0003oool0oooo01P0oooo
00<00?l00?ooo`3oool08P3oool00`00o`00oooo0?ooo`1J0?ooo`00=`3oool00`3o0000oooo0?oo
o`140?ooo`030?l0003oool0oooo01P0oooo00<00?l00?ooo`3oool08@3oool00`00o`00oooo0?oo
o`1K0?ooo`00=P3oool00`3o0000oooo0?ooo`150?ooo`030?l0003oool0oooo01L0oooo00<00?l0
0?ooo`3oool08P3oool00`00o`00oooo0?ooo`1K0?ooo`00=P3oool00`3o0000oooo0?ooo`150?oo
o`030?l0003oool0oooo01H0oooo00<00?l00?ooo`3oool08P3oool00`00o`00oooo0?ooo`1L0?oo
o`00=P3oool00`3o0000oooo0?ooo`150?ooo`030?l0003oool0oooo01D0oooo00<00?l00?ooo`3o
ool08`3oool00`00o`00oooo0?ooo`1L0?ooo`00=@3oool00`3o0000oooo0?ooo`160?ooo`030?l0
003oool0oooo01@0oooo00<00?l00?ooo`3oool08`3oool00`00o`00oooo0?ooo`1M0?ooo`00=@3o
ool00`3o0000oooo0?ooo`160?ooo`030?l0003oool0oooo01<0oooo00<00?l00?ooo`3oool08`3o
ool00`00o`00oooo0?ooo`1N0?ooo`00=@3oool00`3o0000oooo0?ooo`160?ooo`030?l0003oool0
oooo01<0oooo00<00?l00?ooo`3oool08`3oool00`00o`00oooo0?ooo`1N0?ooo`00=@3oool00`3o
0000oooo0?ooo`160?ooo`030?l0003oool0oooo0180oooo00<00?l00?ooo`3oool08`3oool00`00
o`00oooo0?ooo`1O0?ooo`00=03oool00`3o0000oooo0?ooo`180?ooo`030?l0003oool0oooo0100
oooo00<00?l00?ooo`3oool0903oool00`00o`00oooo0?ooo`1O0?ooo`00=03oool00`3o0000oooo
0?ooo`180?ooo`030?l0003oool0oooo00l0oooo00<00?l00?ooo`3oool0903oool00`00o`00oooo
0?ooo`1P0?ooo`00=03oool00`3o0000oooo0?ooo`180?ooo`030?l0003oool0oooo00h0oooo00<0
0?l00?ooo`3oool09@3oool00`00o`00oooo0?ooo`1P0?ooo`00<`3oool00`3o0000oooo0?ooo`19
0?ooo`030?l0003oool0oooo00d0oooo00<00?l00?ooo`3oool09@3oool00`00o`00oooo0?ooo`1Q
0?ooo`00<`3oool00`3o0000oooo0?ooo`190?ooo`030?l0003oool0oooo00d0oooo00<00?l00?oo
o`3oool0903oool00`00o`00oooo0?ooo`1R0?ooo`00<`3oool00`3o0000oooo0?ooo`190?ooo`03
0?l0003oool0oooo00`0oooo00<00?l00?ooo`3oool09@3oool00`00o`00oooo0?ooo`1R0?ooo`00
<`3oool00`3o0000oooo0?ooo`190?ooo`030?l0003oool0oooo00/0oooo00<00?l00?ooo`3oool0
9@3oool00`00o`00oooo0?ooo`1S0?ooo`00<P3oool00`3o0000oooo0?ooo`1:0?ooo`030?l0003o
ool0oooo00X0oooo00<00?l00?ooo`3oool09P3oool00`00o`00oooo0?ooo`1S0?ooo`00<P3oool0
0`3o0000oooo0?ooo`1:0?ooo`030?l0003oool0oooo00T0oooo00<00?l00?ooo`3oool09P3oool0
0`00o`00oooo0?ooo`1T0?ooo`00<P3oool00`3o0000oooo0?ooo`1;0?ooo`030?l0003oool0oooo
00L0oooo00<00?l00?ooo`3oool09P3oool00`00o`00oooo0?ooo`1U0?ooo`00<P3oool00`3o0000
oooo0?ooo`1;0?ooo`030?l0003oool0oooo00L0oooo00<00?l00?ooo`3oool09P3oool00`00o`00
oooo0?ooo`1U0?ooo`00<@3oool00`3o0000oooo0?ooo`1<0?ooo`030?l0003oool0oooo00H0oooo
00<00?l00?ooo`3oool09P3oool00`00o`00oooo0?ooo`1V0?ooo`00<@3oool00`3o0000oooo0?oo
o`1<0?ooo`030?l0003oool0oooo00D0oooo00<00?l00?ooo`3oool09`3oool00`00o`00oooo0?oo
o`1V0?ooo`00<@3oool00`3o0000oooo0?ooo`1<0?ooo`030?l0003oool0oooo00@0oooo00<00?l0
0?ooo`3oool09`3oool00`00o`00oooo0?ooo`1W0?ooo`00<03oool00`3o0000oooo0?ooo`1=0?oo
o`030?l0003oool0oooo00<0oooo00<00?l00?ooo`3oool09`3oool00`00o`00oooo0?ooo`1X0?oo
o`00<03oool00`3o0000oooo0?ooo`1=0?ooo`030?l0003oool0oooo0080oooo00<00?l00?ooo`3o
ool0:03oool00`00o`00oooo0?ooo`1X0?ooo`00<03oool00`3o0000oooo0?ooo`1=0?ooo`030?l0
003oool0oooo0080oooo00<00?l00?ooo`3oool09`3oool00`00o`00oooo0?ooo`1Y0?ooo`00<03o
ool00`3o0000oooo0?ooo`1>0?ooo`040?l0003oool0oooo003o02X0oooo00<00?l00?ooo`3oool0
J@3oool002l0oooo00<0o`000?ooo`3oool0C`3oool00`3o0000oooo003o000Z0?ooo`03003o003o
ool0oooo06X0oooo000_0?ooo`030?l0003oool0oooo04l0oooo00<0o`00003o003oool0:P3oool0
0`00o`00oooo0?ooo`1Z0?ooo`00;`3oool00`3o0000oooo0?ooo`1?0?ooo`030?l0003oool0oooo
02T0oooo00<00?l00?ooo`3oool0J`3oool002h0oooo00<0o`000?ooo`3oool0C`3oool00`00o`00
o`000?ooo`0Y0?ooo`03003o003oool0oooo06`0oooo000^0?ooo`030?l0003oool0oooo04l0oooo
00<00?l00?l0003oool0:@3oool00`00o`00oooo0?ooo`1/0?ooo`00;P3oool00`3o0000oooo0?oo
o`1>0?ooo`03003o003oool0o`0002T0oooo00<00?l00?ooo`3oool0K@3oool002h0oooo00<0o`00
0?ooo`3oool0C@3oool01000o`00oooo0?ooo`3o000Y0?ooo`03003o003oool0oooo06d0oooo000]
0?ooo`030?l0003oool0oooo04d0oooo00D00?l00?ooo`3oool0oooo0?l0000X0?ooo`03003o003o
ool0oooo06h0oooo000]0?ooo`030?l0003oool0oooo04`0oooo00<00?l00?ooo`3oool00`3oool0
0`3o0000oooo0?ooo`0T0?ooo`03003o003oool0oooo06l0oooo000]0?ooo`030?l0003oool0oooo
04/0oooo00<00?l00?ooo`3oool0103oool00`3o0000oooo0?ooo`0T0?ooo`03003o003oool0oooo
06l0oooo000]0?ooo`030?l0003oool0oooo04X0oooo00<00?l00?ooo`3oool01@3oool00`3o0000
oooo0?ooo`0S0?ooo`03003o003oool0oooo0700oooo000/0?ooo`030?l0003oool0oooo04/0oooo
00<00?l00?ooo`3oool01@3oool00`3o0000oooo0?ooo`0S0?ooo`03003o003oool0oooo0700oooo
000/0?ooo`030?l0003oool0oooo04X0oooo00<00?l00?ooo`3oool01P3oool00`3o0000oooo0?oo
o`0R0?ooo`03003o003oool0oooo0740oooo000/0?ooo`030?l0003oool0oooo04T0oooo00<00?l0
0?ooo`3oool01`3oool00`3o0000oooo0?ooo`0Q0?ooo`03003o003oool0oooo0780oooo000[0?oo
o`030?l0003oool0oooo04T0oooo00<00?l00?ooo`3oool0203oool00`3o0000oooo0?ooo`0Q0?oo
o`03003o003oool0oooo0780oooo000[0?ooo`030?l0003oool0oooo04P0oooo00<00?l00?ooo`3o
ool02@3oool00`3o0000oooo0?ooo`0P0?ooo`03003o003oool0oooo07<0oooo000[0?ooo`030?l0
003oool0oooo04L0oooo00<00?l00?ooo`3oool02P3oool00`3o0000oooo0?ooo`0P0?ooo`03003o
003oool0oooo07<0oooo000[0?ooo`030?l0003oool0oooo04L0oooo00<00?l00?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`0N0?ooo`03003o003oool0oooo07@0oooo000Z0?ooo`030?l0003oool0
oooo04L0oooo00<00?l00?ooo`3oool0303oool00`3o0000oooo0?ooo`0N0?ooo`03003o003oool0
oooo07@0oooo000Z0?ooo`030?l0003oool0oooo04H0oooo00<00?l00?ooo`3oool03@3oool00`3o
0000oooo0?ooo`0M0?ooo`03003o003oool0oooo07D0oooo000Z0?ooo`030?l0003oool0oooo04D0
oooo00<00?l00?ooo`3oool03P3oool00`3o0000oooo0?ooo`0L0?ooo`03003o003oool0oooo07H0
oooo000Y0?ooo`030?l0003oool0oooo04D0oooo00<00?l00?ooo`3oool03`3oool00`3o0000oooo
0?ooo`0L0?ooo`03003o003oool0oooo07H0oooo000Y0?ooo`030?l0003oool0oooo04@0oooo00<0
0?l00?ooo`3oool0403oool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo07L0oooo000Y
0?ooo`030?l0003oool0oooo04@0oooo00<00?l00?ooo`3oool0403oool00`3o0000oooo0?ooo`0K
0?ooo`03003o003oool0oooo07L0oooo000Y0?ooo`030?l0003oool0oooo04<0oooo00<00?l00?oo
o`3oool04@3oool00`3o0000oooo0?ooo`0J0?ooo`03003o003oool0oooo07P0oooo000X0?ooo`03
0?l0003oool0oooo04<0oooo00<00?l00?ooo`3oool04P3oool00`3o0000oooo0?ooo`0I0?ooo`03
003o003oool0oooo07T0oooo000X0?ooo`030?l0003oool0oooo0480oooo00<00?l00?ooo`3oool0
503oool00`3o0000oooo0?ooo`0H0?ooo`03003o003oool0oooo07T0oooo000X0?ooo`030?l0003o
ool0oooo0440oooo00<00?l00?ooo`3oool05@3oool00`3o0000oooo0?ooo`0G0?ooo`03003o003o
ool0oooo07X0oooo000X0?ooo`030?l0003oool0oooo0400oooo00<00?l00?ooo`3oool05P3oool0
0`3o0000oooo0?ooo`0G0?ooo`03003o003oool0oooo07X0oooo000W0?ooo`030?l0003oool0oooo
0440oooo00<00?l00?ooo`3oool05P3oool00`3o0000oooo0?ooo`0F0?ooo`03003o003oool0oooo
07/0oooo000W0?ooo`030?l0003oool0oooo0400oooo00<00?l00?ooo`3oool05`3oool00`3o0000
oooo0?ooo`0F0?ooo`03003o003oool0oooo07/0oooo000W0?ooo`030?l0003oool0oooo03l0oooo
00<00?l00?ooo`3oool0603oool00`3o0000oooo0?ooo`0E0?ooo`03003o003oool0oooo07`0oooo
000V0?ooo`030?l0003oool0oooo03l0oooo00<00?l00?ooo`3oool06@3oool00`3o0000oooo0?oo
o`0D0?ooo`03003o003oool0oooo07d0oooo000V0?ooo`030?l0003oool0oooo03h0oooo00<00?l0
0?ooo`3oool06P3oool00`3o0000oooo0?ooo`0D0?ooo`03003o003oool0oooo07d0oooo000V0?oo
o`030?l0003oool0oooo03d0oooo00<00?l00?ooo`3oool06`3oool00`3o0000oooo0?ooo`0C0?oo
o`03003o003oool0oooo07h0oooo000V0?ooo`030?l0003oool0oooo03d0oooo00<00?l00?ooo`3o
ool0703oool00`3o0000oooo0?ooo`0B0?ooo`03003o003oool0oooo07h0oooo000U0?ooo`030?l0
003oool0oooo03d0oooo00<00?l00?ooo`3oool07@3oool00`3o0000oooo0?ooo`0A0?ooo`03003o
003oool0oooo07l0oooo000U0?ooo`030?l0003oool0oooo03`0oooo00<00?l00?ooo`3oool07P3o
ool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo0800oooo000U0?ooo`030?l0003oool0
oooo03/0oooo00<00?l00?ooo`3oool07`3oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0
oooo0800oooo000U0?ooo`030?l0003oool0oooo03X0oooo00<00?l00?ooo`3oool0803oool00`3o
0000oooo0?ooo`0?0?ooo`03003o003oool0oooo0840oooo000T0?ooo`030?l0003oool0oooo03X0
oooo00<00?l00?ooo`3oool08@3oool00`3o0000oooo0?ooo`0?0?ooo`03003o003oool0oooo0840
oooo000T0?ooo`030?l0003oool0oooo03X0oooo00<00?l00?ooo`3oool08@3oool00`3o0000oooo
0?ooo`0>0?ooo`03003o003oool0oooo0880oooo000T0?ooo`030?l0003oool0oooo03T0oooo00<0
0?l00?ooo`3oool08P3oool00`3o0000oooo0?ooo`0=0?ooo`03003o003oool0oooo08<0oooo000S
0?ooo`030?l0003oool0oooo03T0oooo00<00?l00?ooo`3oool0903oool00`3o0000oooo0?ooo`0<
0?ooo`03003o003oool0oooo08<0oooo000S0?ooo`030?l0003oool0oooo03P0oooo00<00?l00?oo
o`3oool09@3oool00`3o0000oooo0?ooo`0;0?ooo`03003o003oool0oooo08@0oooo000S0?ooo`03
0?l0003oool0oooo03L0oooo00<00?l00?ooo`3oool09P3oool00`3o0000oooo0?ooo`0;0?ooo`03
003o003oool0oooo08@0oooo000S0?ooo`030?l0003oool0oooo03H0oooo00<00?l00?ooo`3oool0
9`3oool00`3o0000oooo0?ooo`0:0?ooo`03003o003oool0oooo08D0oooo000R0?ooo`030?l0003o
ool0oooo03H0oooo00<00?l00?ooo`3oool0:03oool00`3o0000oooo0?ooo`0:0?ooo`03003o003o
ool0oooo08D0oooo000R0?ooo`030?l0003oool0oooo03H0oooo00<00?l00?ooo`3oool0:03oool0
0`3o0000oooo0?ooo`090?ooo`03003o003oool0oooo08H0oooo000R0?ooo`030?l0003oool0oooo
03D0oooo00<00?l00?ooo`3oool0:@3oool00`3o0000oooo0?ooo`080?ooo`03003o003oool0oooo
08L0oooo000Q0?ooo`030?l0003oool0oooo03D0oooo00<00?l00?ooo`3oool0:P3oool00`3o0000
oooo0?ooo`080?ooo`03003o003oool0oooo08L0oooo000Q0?ooo`030?l0003oool0oooo03@0oooo
00<00?l00?ooo`3oool0:`3oool00`3o0000oooo0?ooo`070?ooo`03003o003oool0oooo08P0oooo
000Q0?ooo`030?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0;@3oool00`3o0000oooo0?oo
o`060?ooo`03003o003oool0oooo08P0oooo000Q0?ooo`030?l0003oool0oooo0380oooo00<00?l0
0?ooo`3oool0;P3oool00`3o0000oooo0?ooo`050?ooo`03003o003oool0oooo08T0oooo000P0?oo
o`030?l0003oool0oooo03<0oooo00<00?l00?ooo`3oool0;P3oool00`3o0000oooo0?ooo`040?oo
o`03003o003oool0oooo08X0oooo000P0?ooo`030?l0003oool0oooo0380oooo00<00?l00?ooo`3o
ool0;`3oool00`3o0000oooo0?ooo`040?ooo`03003o003oool0oooo08X0oooo000P0?ooo`030?l0
003oool0oooo0340oooo00<00?l00?ooo`3oool0<03oool00`3o0000oooo0?ooo`030?ooo`03003o
003oool0oooo08/0oooo000P0?ooo`030?l0003oool0oooo0300oooo00<00?l00?ooo`3oool0<@3o
ool00`3o0000oooo0?ooo`030?ooo`03003o003oool0oooo08/0oooo000O0?ooo`030?l0003oool0
oooo0300oooo00<00?l00?ooo`3oool0<P3oool00`3o0000oooo0?ooo`020?ooo`03003o003oool0
oooo08`0oooo000O0?ooo`030?l0003oool0oooo02l0oooo00<00?l00?ooo`3oool0<`3oool00`3o
0000oooo0?ooo`020?ooo`03003o003oool0oooo08`0oooo000O0?ooo`030?l0003oool0oooo02l0
oooo00<00?l00?ooo`3oool0<`3oool01@3o0000oooo0?ooo`3oool00?l008l0oooo000N0?ooo`03
0?l0003oool0oooo02l0oooo00<00?l00?ooo`3oool0=@3oool00`3o0000oooo003o002@0?ooo`00
7P3oool00`3o0000oooo0?ooo`0^0?ooo`03003o003oool0oooo03H0oooo00<0o`000?ooo`00o`00
T03oool001h0oooo00<0o`000?ooo`3oool0;@3oool00`00o`00oooo0?ooo`0g0?ooo`030?l00000
o`00oooo0900oooo000N0?ooo`030?l0003oool0oooo02`0oooo00<00?l00?ooo`3oool0>03oool0
0`3o00000?l00?ooo`2@0?ooo`007@3oool00`3o0000oooo0?ooo`0/0?ooo`03003o003oool0oooo
03P0oooo0P3o002B0?ooo`007@3oool00`3o0000oooo0?ooo`0/0?ooo`03003o003oool0oooo03H0
oooo103o002B0?ooo`007@3oool00`3o0000oooo0?ooo`0[0?ooo`03003o003oool0oooo03D0oooo
103o00000`00o`00o`000?ooo`2A0?ooo`00703oool00`3o0000oooo0?ooo`0[0?ooo`03003o003o
ool0oooo03@0oooo1@3o00000`00o`00oooo0?l0002B0?ooo`00703oool00`3o0000oooo0?ooo`0Z
0?ooo`03003o003oool0oooo03H0oooo103o00000`00o`00oooo0?l0002B0?ooo`00703oool00`3o
0000oooo0?ooo`0Y0?ooo`03003o003oool0oooo03D0oooo1@3o00001@00o`00oooo0?ooo`3oool0
o`000940oooo000L0?ooo`030?l0003oool0oooo02P0oooo00<00?l00?ooo`3oool0=03oool20?l0
00<0oooo00<0o`000?ooo`00o`000`3oool00`3o0000oooo0?ooo`2?0?ooo`006`3oool00`3o0000
oooo0?ooo`0Y0?ooo`03003o003oool0oooo03<0oooo00<0o`000?ooo`3oool0103oool00`00o`00
oooo0?ooo`020?ooo`030?l0003oool0oooo08l0oooo000K0?ooo`030?l0003oool0oooo02P0oooo
00<00?l00?ooo`3oool0<P3oool20?l000H0oooo00<00?l00?ooo`3oool00`3oool00`3o0000oooo
0?ooo`2?0?ooo`006`3oool00`3o0000oooo0?ooo`0W0?ooo`03003o003oool0oooo0380oooo00<0
o`000?ooo`3oool01P3oool00`00o`00oooo0?ooo`030?ooo`030?l0003oool0oooo08l0oooo000K
0?ooo`030?l0003oool0oooo02H0oooo00<00?l00?ooo`3oool0<@3oool20?l000P0oooo00<00?l0
0?ooo`3oool0103oool00`3o0000oooo0?ooo`2?0?ooo`006P3oool00`3o0000oooo0?ooo`0V0?oo
o`03003o003oool0oooo0340oooo00<0o`000?ooo`3oool0203oool00`00o`00oooo0?ooo`040?oo
o`030?l0003oool0oooo08l0oooo000J0?ooo`030?l0003oool0oooo02D0oooo00<00?l00?ooo`3o
ool0<03oool20?l000X0oooo00<00?l00?ooo`3oool01@3oool00`3o0000oooo0?ooo`2?0?ooo`00
6P3oool00`3o0000oooo0?ooo`0T0?ooo`03003o003oool0oooo0300oooo00<0o`000?ooo`3oool0
2@3oool00`00o`00oooo0?ooo`070?ooo`030?l0003oool0oooo08h0oooo000I0?ooo`030?l0003o
ool0oooo02D0oooo00<00?l00?ooo`3oool0;P3oool20?l000`0oooo00<00?l00?ooo`3oool01`3o
ool00`3o0000oooo0?ooo`2>0?ooo`006@3oool00`3o0000oooo0?ooo`0T0?ooo`03003o003oool0
oooo02h0oooo00<0o`000?ooo`3oool02`3oool00`00o`00oooo0?ooo`080?ooo`030?l0003oool0
oooo08h0oooo000I0?ooo`030?l0003oool0oooo02<0oooo00<00?l00?ooo`3oool0;@3oool20?l0
00h0oooo00<00?l00?ooo`3oool0203oool00`3o0000oooo0?ooo`2>0?ooo`006@3oool00`3o0000
oooo0?ooo`0R0?ooo`03003o003oool0oooo02`0oooo0P3o000?0?ooo`03003o003oool0oooo00T0
oooo00<0o`000?ooo`3oool0SP3oool001P0oooo00<0o`000?ooo`3oool08P3oool00`00o`00oooo
0?ooo`0/0?ooo`030?l0003oool0oooo00l0oooo00<00?l00?ooo`3oool02@3oool00`3o0000oooo
0?ooo`2>0?ooo`00603oool00`3o0000oooo0?ooo`0Q0?ooo`03003o003oool0oooo02/0oooo0P3o
000A0?ooo`03003o003oool0oooo00X0oooo00<0o`000?ooo`3oool0SP3oool001P0oooo00<0o`00
0?ooo`3oool08@3oool00`00o`00oooo0?ooo`0Z0?ooo`030?l0003oool0oooo0100oooo00<00?l0
0?ooo`3oool02`3oool00`3o0000oooo0?ooo`2>0?ooo`005`3oool00`3o0000oooo0?ooo`0Q0?oo
o`03003o003oool0oooo02T0oooo0P3o000C0?ooo`03003o003oool0oooo00/0oooo00<0o`000?oo
o`3oool0SP3oool001L0oooo00<0o`000?ooo`3oool0803oool00`00o`00oooo0?ooo`0Y0?ooo`03
0?l0003oool0oooo0180oooo00<00?l00?ooo`3oool03@3oool00`3o0000oooo0?ooo`2=0?ooo`00
5`3oool00`3o0000oooo0?ooo`0O0?ooo`03003o003oool0oooo02P0oooo0P3o000E0?ooo`03003o
003oool0oooo00d0oooo00<0o`000?ooo`3oool0S@3oool001L0oooo00<0o`000?ooo`3oool07P3o
ool00`00o`00oooo0?ooo`0X0?ooo`030?l0003oool0oooo01@0oooo00<00?l00?ooo`3oool03P3o
ool00`3o0000oooo0?ooo`2=0?ooo`005P3oool00`3o0000oooo0?ooo`0N0?ooo`03003o003oool0
oooo02L0oooo0P3o000G0?ooo`03003o003oool0oooo00h0oooo00<0o`000?ooo`3oool0S@3oool0
01H0oooo00<0o`000?ooo`3oool07P3oool00`00o`00oooo0?ooo`0U0?ooo`80o`00603oool00`00
o`00oooo0?ooo`0?0?ooo`030?l0003oool0oooo08d0oooo000F0?ooo`030?l0003oool0oooo01d0
oooo00<00?l00?ooo`3oool09@3oool00`3o0000oooo0?ooo`0G0?ooo`03003o003oool0oooo0100
oooo00<0o`000?ooo`3oool0S@3oool001H0oooo00<0o`000?ooo`3oool0703oool00`00o`00oooo
0?ooo`0T0?ooo`80o`006P3oool00`00o`00oooo0?ooo`0@0?ooo`030?l0003oool0oooo08d0oooo
000E0?ooo`030?l0003oool0oooo01`0oooo00<00?l00?ooo`3oool0903oool00`3o0000oooo0?oo
o`0I0?ooo`03003o003oool0oooo0140oooo00<0o`000?ooo`3oool0S@3oool001D0oooo00<0o`00
0?ooo`3oool06`3oool00`00o`00oooo0?ooo`0S0?ooo`80o`00703oool00`00o`00oooo0?ooo`0B
0?ooo`030?l0003oool0oooo08`0oooo000E0?ooo`030?l0003oool0oooo01X0oooo00<00?l00?oo
o`3oool08`3oool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo01<0oooo00<0o`000?oo
o`3oool0S03oool001@0oooo00<0o`000?ooo`3oool06`3oool00`00o`00oooo0?ooo`0Q0?ooo`80
o`007@3oool00`00o`00oooo0?ooo`0D0?ooo`030?l0003oool0oooo08`0oooo000D0?ooo`030?l0
003oool0oooo01X0oooo00<00?l00?ooo`3oool08@3oool00`3o0000oooo0?ooo`0M0?ooo`03003o
003oool0oooo01@0oooo00<0o`000?ooo`3oool0S03oool001@0oooo00<0o`000?ooo`3oool06@3o
ool00`00o`00oooo0?ooo`0P0?ooo`80o`007`3oool00`00o`00oooo0?ooo`0E0?ooo`030?l0003o
ool0oooo08`0oooo000D0?ooo`030?l0003oool0oooo01P0oooo00<00?l00?ooo`3oool0803oool0
0`3o0000oooo0?ooo`0O0?ooo`03003o003oool0oooo01D0oooo00<0o`000?ooo`3oool0S03oool0
01<0oooo00<0o`000?ooo`3oool0603oool00`00o`00oooo0?ooo`0O0?ooo`80o`008@3oool00`00
o`00oooo0?ooo`0F0?ooo`030?l0003oool0oooo08`0oooo000C0?ooo`030?l0003oool0oooo01L0
oooo00<00?l00?ooo`3oool07P3oool20?l002<0oooo00<00?l00?ooo`3oool05P3oool00`3o0000
oooo0?ooo`2<0?ooo`004`3oool00`3o0000oooo0?ooo`0G0?ooo`03003o003oool0oooo01d0oooo
00<0o`000?ooo`3oool08P3oool00`00o`00oooo0?ooo`0G0?ooo`030?l0003oool0oooo08`0oooo
000B0?ooo`030?l0003oool0oooo01L0oooo00<00?l00?ooo`3oool0703oool20?l002@0oooo00<0
0?l00?ooo`3oool06@3oool00`3o0000oooo0?ooo`2;0?ooo`004P3oool00`3o0000oooo0?ooo`0F
0?ooo`03003o003oool0oooo01`0oooo00<0o`000?ooo`3oool0903oool00`00o`00oooo0?ooo`0I
0?ooo`030?l0003oool0oooo08/0oooo000B0?ooo`030?l0003oool0oooo01D0oooo00<00?l00?oo
o`3oool06`3oool20?l002H0oooo00<00?l00?ooo`3oool06P3oool00`3o0000oooo0?ooo`2;0?oo
o`004P3oool00`3o0000oooo0?ooo`0D0?ooo`03003o003oool0oooo01/0oooo00<0o`000?ooo`3o
ool09P3oool00`00o`00oooo0?ooo`0J0?ooo`030?l0003oool0oooo08/0oooo000A0?ooo`030?l0
003oool0oooo01@0oooo00<00?l00?ooo`3oool06P3oool20?l002P0oooo00<00?l00?ooo`3oool0
6`3oool00`3o0000oooo0?ooo`2;0?ooo`004@3oool00`3o0000oooo0?ooo`0D0?ooo`03003o003o
ool0oooo01T0oooo00<0o`000?ooo`3oool09`3oool00`00o`00oooo0?ooo`0L0?ooo`030?l0003o
ool0oooo08/0oooo000A0?ooo`030?l0003oool0oooo01<0oooo00<00?l00?ooo`3oool0603oool2
0?l002X0oooo00<00?l00?ooo`3oool0703oool00`3o0000oooo0?ooo`2;0?ooo`004@3oool00`3o
0000oooo0?ooo`0B0?ooo`03003o003oool0oooo01P0oooo00<0o`000?ooo`3oool0:@3oool00`00
o`00oooo0?ooo`0M0?ooo`030?l0003oool0oooo08/0oooo000@0?ooo`030?l0003oool0oooo0180
oooo00<00?l00?ooo`3oool05`3oool20?l002`0oooo00<00?l00?ooo`3oool07@3oool00`3o0000
oooo0?ooo`2;0?ooo`00403oool00`3o0000oooo0?ooo`0A0?ooo`03003o003oool0oooo01H0oooo
0P3o000]0?ooo`03003o003oool0oooo01l0oooo00<0o`000?ooo`3oool0RP3oool00100oooo00<0
o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0F0?ooo`030?l0003oool0oooo02d0oooo00<0
0?l00?ooo`3oool07`3oool00`3o0000oooo0?ooo`2:0?ooo`003`3oool00`3o0000oooo0?ooo`0@
0?ooo`03003o003oool0oooo01D0oooo0P3o000_0?ooo`03003o003oool0oooo0200oooo00<0o`00
0?ooo`3oool0RP3oool000l0oooo00<0o`000?ooo`3oool0403oool00`00o`00oooo0?ooo`0D0?oo
o`030?l0003oool0oooo02h0oooo00<00?l00?ooo`3oool08@3oool00`3o0000oooo0?ooo`2:0?oo
o`003`3oool00`3o0000oooo0?ooo`0?0?ooo`03003o003oool0oooo01<0oooo0P3o000a0?ooo`03
003o003oool0oooo0240oooo00<0o`000?ooo`3oool0RP3oool000l0oooo00<0o`000?ooo`3oool0
3P3oool00`00o`00oooo0?ooo`0C0?ooo`030?l0003oool0oooo0300oooo00<00?l00?ooo`3oool0
8P3oool00`3o0000oooo0?ooo`2:0?ooo`003P3oool00`3o0000oooo0?ooo`0>0?ooo`03003o003o
ool0oooo0180oooo0P3o000c0?ooo`03003o003oool0oooo0280oooo00<0o`000?ooo`3oool0RP3o
ool000h0oooo00<0o`000?ooo`3oool03@3oool00`00o`00oooo0?ooo`0B0?ooo`030?l0003oool0
oooo0380oooo00<00?l00?ooo`3oool08`3oool00`3o0000oooo0?ooo`2:0?ooo`003P3oool00`3o
0000oooo0?ooo`0<0?ooo`03003o003oool0oooo0140oooo0P3o000e0?ooo`03003o003oool0oooo
02@0oooo00<0o`000?ooo`3oool0R@3oool000d0oooo00<0o`000?ooo`3oool03@3oool00`00o`00
oooo0?ooo`0@0?ooo`030?l0003oool0oooo03@0oooo00<00?l00?ooo`3oool09@3oool00`3o0000
oooo0?ooo`290?ooo`003@3oool00`3o0000oooo0?ooo`0<0?ooo`03003o003oool0oooo00l0oooo
0P3o000f0?ooo`03003o003oool0oooo02H0oooo00<0o`000?ooo`3oool0R@3oool000d0oooo00<0
o`000?ooo`3oool02`3oool00`00o`00oooo0?ooo`0>0?ooo`80o`00>03oool00`00o`00oooo0?oo
o`0V0?ooo`030?l0003oool0oooo08T0oooo000=0?ooo`030?l0003oool0oooo00X0oooo00<00?l0
0?ooo`3oool03P3oool00`3o0000oooo0?ooo`0g0?ooo`03003o003oool0oooo02L0oooo00<0o`00
0?ooo`3oool0R@3oool000`0oooo00<0o`000?ooo`3oool02P3oool00`00o`00oooo0?ooo`0=0?oo
o`80o`00>P3oool00`00o`00oooo0?ooo`0W0?ooo`030?l0003oool0oooo08T0oooo000<0?ooo`03
0?l0003oool0oooo00T0oooo00<00?l00?ooo`3oool03@3oool00`3o0000oooo0?ooo`0i0?ooo`03
003o003oool0oooo02P0oooo00<0o`000?ooo`3oool0R@3oool000`0oooo00<0o`000?ooo`3oool0
2@3oool00`00o`00oooo0?ooo`0;0?ooo`80o`00>`3oool00`00o`00oooo0?ooo`0Y0?ooo`030?l0
003oool0oooo08T0oooo000<0?ooo`030?l0003oool0oooo00P0oooo00<00?l00?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`0k0?ooo`03003o003oool0oooo02T0oooo00<0o`000?ooo`3oool0R@3o
ool000/0oooo00<0o`000?ooo`3oool0203oool00`00o`00oooo0?ooo`0:0?ooo`80o`00?@3oool0
0`00o`00oooo0?ooo`0[0?ooo`030?l0003oool0oooo08P0oooo000;0?ooo`030?l0003oool0oooo
00L0oooo00<00?l00?ooo`3oool02P3oool00`3o0000oooo0?ooo`0m0?ooo`03003o003oool0oooo
02/0oooo00<0o`000?ooo`3oool0R03oool000/0oooo00<0o`000?ooo`3oool01P3oool00`00o`00
oooo0?ooo`090?ooo`80o`00?`3oool00`00o`00oooo0?ooo`0/0?ooo`030?l0003oool0oooo08P0
oooo000:0?ooo`030?l0003oool0oooo00H0oooo00<00?l00?ooo`3oool0203oool20?l00440oooo
00<00?l00?ooo`3oool0;03oool00`3o0000oooo0?ooo`280?ooo`002P3oool00`3o0000oooo0?oo
o`060?ooo`03003o003oool0oooo00L0oooo00<0o`000?ooo`3oool0@03oool00`00o`00oooo0?oo
o`0]0?ooo`030?l0003oool0oooo08P0oooo000:0?ooo`030?l0003oool0oooo00D0oooo00<00?l0
0?ooo`3oool01P3oool20?l00480oooo00<00?l00?ooo`3oool0;P3oool00`3o0000oooo0?ooo`28
0?ooo`002P3oool00`3o0000oooo0?ooo`040?ooo`03003o003oool0oooo00H0oooo00<0o`000?oo
o`3oool0@P3oool00`00o`00oooo0?ooo`0^0?ooo`030?l0003oool0oooo08P0oooo00090?ooo`03
0?l0003oool0oooo00@0oooo00<00?l00?ooo`3oool01@3oool20?l004@0oooo00<00?l00?ooo`3o
ool0;`3oool00`3o0000oooo0?ooo`280?ooo`002@3oool00`3o0000oooo0?ooo`030?ooo`03003o
003oool0oooo00D0oooo00<0o`000?ooo`3oool0A03oool00`00o`00oooo0?ooo`0`0?ooo`030?l0
003oool0oooo08L0oooo00090?ooo`030?l0003oool0oooo0080oooo00<00?l00?ooo`3oool0103o
ool20?l004H0oooo00<00?l00?ooo`3oool0<@3oool00`3o0000oooo0?ooo`270?ooo`00203oool0
0`3o0000oooo0?ooo`030?ooo`03003o003oool0oooo00<0oooo00<0o`000?ooo`3oool0AP3oool0
0`00o`00oooo0?ooo`0a0?ooo`030?l0003oool0oooo08L0oooo00080?ooo`030?l0003oool0oooo
0080oooo00<00?l00?ooo`3oool00P3oool20?l004P0oooo00<00?l00?ooo`3oool0<P3oool00`3o
0000oooo0?ooo`270?ooo`00203oool01@3o0000oooo0?ooo`3oool00?l000@0oooo00<0o`000?oo
o`3oool0A`3oool00`00o`00oooo0?ooo`0c0?ooo`030?l0003oool0oooo08L0oooo00080?ooo`04
0?l0003oool0oooo003o00<0oooo0P3o001:0?ooo`03003o003oool0oooo03<0oooo00<0o`000?oo
o`3oool0Q`3oool000L0oooo00@0o`000?ooo`3oool00?l00P3oool20?l004/0oooo00<00?l00?oo
o`3oool0=03oool00`3o0000oooo0?ooo`270?ooo`001`3oool00`3o0000oooo003o00020?ooo`03
0?l0003oool0oooo04/0oooo00<00?l00?ooo`3oool0=03oool00`3o0000oooo0?ooo`270?ooo`00
1`3oool00`3o0000oooo003o00020?l004d0oooo00<00?l00?ooo`3oool0=@3oool00`3o0000oooo
0?ooo`270?ooo`001`3oool00`3o00000?l00?l0001>0?ooo`03003o003oool0oooo03L0oooo00<0
o`000?ooo`3oool0QP3oool000H0oooo0`3o001?0?ooo`03003o003oool0oooo03L0oooo00<0o`00
0?ooo`3oool0QP3oool000H0oooo0P3o001?0?ooo`03003o003oool0oooo03P0oooo00<0o`000?oo
o`3oool0QP3oool000L0oooo00@00?l00?l0003o0000o`00C03oool00`00o`00oooo0?ooo`0h0?oo
o`030?l0003oool0oooo08H0oooo00090?ooo`800?l00P3o00190?ooo`03003o003oool0oooo03T0
oooo00<0o`000?ooo`3oool0QP3oool000/0oooo0P00o`030?l004H0oooo00<00?l00?ooo`3oool0
>@3oool00`3o0000oooo0?ooo`260?ooo`003@3oool00`00o`00oooo0?ooo`030?l00480oooo00<0
0?l00?ooo`3oool0>P3oool00`3o0000oooo0?ooo`260?ooo`003P3oool2003o00<0oooo0`3o000n
0?ooo`03003o003oool0oooo03/0oooo00<0o`000?ooo`3oool0QP3oool00100oooo0P00o`040?oo
o`80o`00?03oool00`00o`00oooo0?ooo`0l0?ooo`030?l0003oool0oooo08D0oooo000B0?ooo`80
0?l0103oool30?l003P0oooo00<00?l00?ooo`3oool0?@3oool00`3o0000oooo0?ooo`250?ooo`00
503oool00`00o`00oooo0?ooo`040?ooo`<0o`00=@3oool00`00o`00oooo0?ooo`0m0?ooo`030?l0
003oool0oooo08D0oooo000E0?ooo`800?l01`3oool30?l00340oooo00<00?l00?ooo`3oool0?P3o
ool00`3o0000oooo0?ooo`250?ooo`005`3oool2003o00P0oooo0P3o000^0?ooo`03003o003oool0
oooo03l0oooo00<0o`000?ooo`3oool0Q@3oool001T0oooo0P00o`080?ooo`<0o`00:`3oool00`00
o`00oooo0?ooo`0o0?ooo`030?l0003oool0oooo08D0oooo000K0?ooo`03003o003oool0oooo00P0
oooo0`3o000W0?ooo`03003o003oool0oooo0400oooo00<0o`000?ooo`3oool0Q@3oool001`0oooo
0P00o`0;0?ooo`<0o`00903oool00`00o`00oooo0?ooo`100?ooo`030?l0003oool0oooo08D0oooo
000N0?ooo`800?l0303oool20?l00240oooo00<00?l00?ooo`3oool0@@3oool00`3o0000oooo0?oo
o`250?ooo`00803oool2003o00`0oooo0`3o000N0?ooo`03003o003oool0oooo0480oooo00<0o`00
0?ooo`3oool0Q03oool00280oooo00<00?l00?ooo`3oool0303oool30?l001X0oooo00<00?l00?oo
o`3oool0@`3oool00`3o0000oooo0?ooo`240?ooo`008`3oool2003o00l0oooo0`3o000F0?ooo`03
003o003oool0oooo04@0oooo00<0o`000?ooo`3oool0Q03oool002D0oooo0P00o`0@0?ooo`80o`00
503oool00`00o`00oooo0?ooo`140?ooo`030?l0003oool0oooo08@0oooo000W0?ooo`03003o003o
ool0oooo00l0oooo0`3o000@0?ooo`03003o003oool0oooo04D0oooo00<0o`000?ooo`3oool0Q03o
ool002P0oooo0P00o`0B0?ooo`<0o`003@3oool00`00o`00oooo0?ooo`150?ooo`030?l0003oool0
oooo08@0oooo000Z0?ooo`800?l04`3oool30?l000T0oooo00<00?l00?ooo`3oool0AP3oool00`3o
0000oooo0?ooo`240?ooo`00;03oool2003o01@0oooo0P3o00070?ooo`03003o003oool0oooo04H0
oooo00<0o`000?ooo`3oool0Q03oool002h0oooo00<00?l00?ooo`3oool04`3oool30?l000<0oooo
00<00?l00?ooo`3oool0A`3oool00`3o0000oooo0?ooo`240?ooo`00;`3oool2003o01H0oooo0`3o
001;0?ooo`030?l0003oool0oooo08<0oooo000a0?ooo`800?l05P3oool01000o`00o`000?l0003o
00180?ooo`030?l0003oool0oooo08<0oooo000c0?ooo`800?l04`3oool00`00o`00oooo0?ooo`02
0?ooo`80o`00AP3oool00`3o0000oooo0?ooo`230?ooo`00=@3oool00`00o`00oooo0?ooo`0@0?oo
o`03003o003oool0oooo00@0oooo0`3o00130?ooo`030?l0003oool0oooo08<0oooo000f0?ooo`80
0?l03@3oool3003o00X0oooo0`3o00100?ooo`030?l0003oool0oooo08<0oooo000h0?ooo`800?l0
2`3oool4003o00`0oooo0`3o000m0?ooo`030?l0003oool0oooo08<0oooo000j0?ooo`800?l000<0
oooo003o003oool01@3oool4003o0100oooo0`3o000j0?ooo`030?l0003oool0oooo08<0oooo000l
0?ooo`<00?l01@3oool3003o01@0oooo0P3o000h0?ooo`030?l0003oool0oooo08<0oooo000l0?oo
o`@00?l0103oool2003o01L0oooo0`3o000f0?ooo`030?l0003oool0oooo0880oooo000k0?ooo`L0
0?l00P3oool2003o01X0oooo0`3o000c0?ooo`030?l0003oool0oooo0880oooo000m0?ooo`P00?l0
7P3oool30?l00300oooo00<0o`000?ooo`3oool0PP3oool00440oooo0`00o`0R0?ooo`80o`00;P3o
ool00`3o0000oooo0?ooo`220?ooo`00J03oool30?l002/0oooo00<0o`000?ooo`3oool0PP3oool0
06/0oooo0`3o000X0?ooo`030?l0003oool0oooo0880oooo001^0?ooo`<0o`009@3oool00`3o0000
oooo0?ooo`220?ooo`00L@3oool20?l002<0oooo00<0o`000?ooo`3oool0PP3oool007<0oooo0`3o
000P0?ooo`030?l0003oool0oooo0880oooo001f0?ooo`<0o`007P3oool00`3o0000oooo0?ooo`21
0?ooo`00N@3oool30?l001/0oooo00<0o`000?ooo`3oool0P@3oool007`0oooo0P3o000I0?ooo`03
0?l0003oool0oooo0840oooo001n0?ooo`<0o`005P3oool00`3o0000oooo0?ooo`210?ooo`00P@3o
ool30?l001<0oooo0`3o00210?ooo`00Q03oool30?l000h0oooo1@3o00210?ooo`00Q`3oool20?l0
00`0oooo1@3o00210?ooo`00R@3oool30?l000X0oooo103o00210?ooo`00S03oool30?l00080oooo
00<0o`000?ooo`3oool00P3oool30?l00880oooo002?0?ooo`D0o`000`3oool20?l00880oooo002@
0?ooo`H0o`0000<0oooo0?l0003o0000PP3oool00900oooo2@3o00220?ooo`00U03oool50?l00880
oooo003o0?oooa`0oooo003o0?oooa`0oooo003o0?oooa`0oooo003o0?oooa`0oooo003o0?oooa`0
oooo003o0?oooa`0oooo003o0?oooa`0oooo003o0?oooa`0oooo003o0?oooa`0oooo003o0?oooa`0
oooo0000\
\>"],
  ImageRangeCache->{{{0, 282.688}, {430.813, 0}} -> {-0.0482191, -2.63157, \
0.010744, 0.010744}}]
}, Closed]],

Cell[BoxData[
    \(\(PlotCMmomenta[cmMomenta];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[%, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.5 0.241762 0.5 0.241762 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.5 Mabswid
[ ] 0 setdash
.5 .5 m
.25899 .91069 L
s
% Start of user PostScript
0. 0. -0.99690008 1.6987556 MAarrowhead1
% End of user PostScript
.5 .5 m
.74101 .08931 L
s
% Start of user PostScript
0. 0. 0.99690008 -1.6987556 MAarrowhead1
% End of user PostScript
.5 .5 m
.55342 .97318 L
s
% Start of user PostScript
0. 0. 0.2209686 1.957231 MAarrowhead1
% End of user PostScript
.5 .5 m
.44658 .02682 L
s
% Start of user PostScript
0. 0. -0.2209686 -1.957231 MAarrowhead1
% End of user PostScript
newpath
.5 .5 .47619 0 365.73 arc
s
newpath
.5 .5 .47619 0 365.73 arc
s
3 Mabswid
.5 .5 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00a`3ooolQ00000<P0oooo002m0?ooo`X000008@3oool:00000;h0oooo002e0?ooo`P00000
0`3oool00`000000oooo0?ooo`0_0?ooo`P00000]P3oool00;00oooo1@00000;0?ooo`800000>03o
ool500000;40oooo002[0?ooo`D00000403oool2000003d0oooo1@00002/0?ooo`00YP3oool50000
01@0oooo100000110?ooo`800000ZP3oool00:@0oooo0P00000I0?ooo`@00000@`3oool500000:D0
oooo002O0?ooo`D000006`3oool5000004L0oooo1@00002P0?ooo`00W@3oool200000200oooo1@00
001<0?ooo`<00000W@3oool009X0oooo0`00000R0?ooo`<00000D@3oool2000009/0oooo002E0?oo
o`D000009`3oool00`000000oooo0?ooo`1A0?ooo`D00000UP3oool009<0oooo0P00000/0?ooo`03
0000003oool0oooo05H0oooo0P00002D0?ooo`00T03oool3000002h0oooo00<000000?ooo`3oool0
F03oool300000940oooo002>0?ooo`800000<@3oool00`000000oooo0?ooo`1K0?ooo`800000S`3o
ool008`0oooo0P00000c0?ooo`030000003oool0oooo05d0oooo0P00002=0?ooo`00R@3oool30000
03H0oooo00<000000?ooo`3oool0GP3oool3000008X0oooo00270?ooo`800000>@3oool00`000000
oooo0?ooo`1Q0?ooo`800000R03oool008D0oooo0P00000k0?ooo`030000003oool0oooo06<0oooo
0P0000260?ooo`00PP3oool3000003d0oooo00<000000?ooo`3oool0I@3oool3000008<0oooo0020
0?ooo`800000@03oool00`000000oooo0?ooo`1X0?ooo`800000P@3oool007h0oooo0P0000120?oo
o`030000003oool0oooo06X0oooo0P00001o0?ooo`00N`3oool3000004@0oooo00<000000?ooo`3o
ool0K03oool3000007`0oooo001i0?ooo`800000A`3oool00`000000oooo0?ooo`1_0?ooo`800000
NP3oool007P0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`1a0?ooo`030000003o
ool0oooo07L0oooo001g0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0L@3oool0
0`000000oooo0?ooo`1f0?ooo`00M@3oool2000004`0oooo00<000000?ooo`3oool0LP3oool20000
07H0oooo001b0?ooo`<00000CP3oool00`000000oooo0?ooo`1d0?ooo`<00000L`3oool00700oooo
0P00001A0?ooo`030000003oool0oooo07L0oooo0P00001a0?ooo`00K`3oool00`000000oooo0?oo
o`1A0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0KP3oool006h0oooo00<00000
0?ooo`3oool0DP3oool00`000000oooo0?ooo`1h0?ooo`<00000K`3oool006`0oooo0P00001E0?oo
o`030000003oool0oooo07L0oooo0P0000020?ooo`800000K@3oool006X0oooo0P00001G0?ooo`03
0000003oool0oooo07H0oooo0`0000040?ooo`800000J`3oool006T0oooo00<000000?ooo`3oool0
E`3oool00`000000oooo0?ooo`1e0?ooo`@000001P3oool00`000000oooo0?ooo`1X0?ooo`00J03o
ool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo07<0oooo1@0000070?ooo`800000J03o
ool006D0oooo0`00001L0?ooo`030000003oool0oooo0780oooo1@00000:0?ooo`800000IP3oool0
06<0oooo0P00001O0?ooo`030000003oool0oooo07@0oooo0`00000<0?ooo`800000I03oool00680
oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0H@3oool00640oooo00<000000?ooo`3oool0H03oool00`000000oooo
0?ooo`1c0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0H03oool005l0oooo0P00
001S0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool04P3oool200000600oooo001N
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo05d0oooo001M0?ooo`030000003oool0oooo06@0oooo00<000000?oo
o`3oool0L@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo05`0oooo001K0?ooo`80
0000J03oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo01P0oooo0P00001L0?ooo`00
FP3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`1I0?ooo`00F@3oool00`000000oooo0?ooo`1Y0?ooo`030000003o
ool0oooo06h0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`1H0?ooo`00F03oool0
0`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool07@3oool0
0`000000oooo0?ooo`1G0?ooo`00E`3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo
06d0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`1F0?ooo`00E@3oool2000006h0
oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`0Q0?ooo`800000EP3oool005@0oooo
00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo02<0oooo
00<000000?ooo`3oool0D`3oool005<0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?oo
o`1[0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0DP3oool00540oooo0P00001b
0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool09P3oool200000580oooo001@0?oo
o`030000003oool0oooo07<0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0Y0?oo
o`030000003oool0oooo04l0oooo001?0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3o
ool0J03oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo04h0oooo001>0?ooo`030000
003oool0oooo07D0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo04d0oooo001=0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0I`3o
ool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo04`0oooo001<0?ooo`030000003oool0
oooo07L0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo04/0oooo001;0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0IP3oool00`00
0000oooo0?ooo`0a0?ooo`030000003oool0oooo04X0oooo001:0?ooo`030000003oool0oooo07T0
oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo04T0
oooo00190?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0I@3oool00`000000oooo
0?ooo`0d0?ooo`030000003oool0oooo04P0oooo00180?ooo`030000003oool0oooo07/0oooo00<0
00000?ooo`3oool0I03oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo04L0oooo0017
0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo04H0oooo00160?ooo`030000003oool0oooo07h0oooo00<000000?oo
o`3oool0HP3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo04D0oooo00150?ooo`03
0000003oool0oooo07l0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo04@0oooo00140?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool0
H@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo04<0oooo00130?ooo`030000003o
ool0oooo0840oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`0n0?ooo`030000003o
ool0oooo0480oooo00120?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0H03oool0
0`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0440oooo00110?ooo`030000003oool0oooo
08<0oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo
0400oooo00100?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0G`3oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo03l0oooo000o0?ooo`030000003oool0oooo08D0oooo
00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo03h0oooo
000n0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?oo
o`160?ooo`030000003oool0oooo03d0oooo000m0?ooo`030000003oool0oooo08P0oooo00<00000
0?ooo`3oool0G03oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo03`0oooo000l0?oo
o`030000003oool0oooo08T0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo03/0oooo000k0?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3o
ool0F`3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo03X0oooo000j0?ooo`030000
003oool0oooo08/0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`1;0?ooo`030000
003oool0oooo03X0oooo000j0?ooo`030000003oool0oooo08/0oooo00<000000?ooo`3oool0FP3o
ool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo03T0oooo000i0?ooo`030000003oool0
oooo08`0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0
oooo03P0oooo000h0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool0F03oool00`00
0000oooo0?ooo`1@0?ooo`030000003oool0oooo03L0oooo000g0?ooo`030000003oool0oooo08h0
oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo03L0
oooo000g0?ooo`030000003oool0oooo08l0oooo00<000000?ooo`3oool0EP3oool00`000000oooo
0?ooo`1B0?ooo`030000003oool0oooo03H0oooo000f0?ooo`030000003oool0oooo0900oooo00<0
00000?ooo`3oool0EP3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo03D0oooo000e
0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`1E
0?ooo`030000003oool0oooo03@0oooo000d0?ooo`030000003oool0oooo0980oooo00<000000?oo
o`3oool0E03oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo03<0oooo000c0?ooo`03
0000003oool0oooo09<0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1H0?ooo`03
0000003oool0oooo0380oooo000b0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0
D`3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo0380oooo000b0?ooo`030000003o
ool0oooo09@0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`1J0?ooo`030000003o
ool0oooo0340oooo000a0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0DP3oool0
0`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0300oooo000`0?ooo`030000003oool0oooo
09H0oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo
02l0oooo000_0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0D03oool00`000000
oooo0?ooo`1N0?ooo`030000003oool0oooo02l0oooo000_0?ooo`030000003oool0oooo09P0oooo
00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo02h0oooo
000^0?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?oo
o`1Q0?ooo`030000003oool0oooo02d0oooo000]0?ooo`030000003oool0oooo09X0oooo00<00000
0?ooo`3oool0CP3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo02`0oooo000/0?oo
o`030000003oool0oooo09/0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1S0?oo
o`030000003oool0oooo02`0oooo000/0?ooo`030000003oool0oooo09/0oooo00<000000?ooo`3o
ool0C@3oool00`000000oooo0?ooo`1U0?ooo`030000003oool0oooo02/0oooo000/0?ooo`030000
003oool0oooo09/0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1V0?ooo`030000
003oool0oooo02/0oooo000[0?ooo`030000003oool0oooo09`0oooo00<000000?ooo`3oool0C03o
ool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo02/0oooo000[0?ooo`030000003oool0
oooo09`0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0
oooo02X0oooo000Z0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool0BP3oool00`00
0000oooo0?ooo`1Y0?ooo`030000003oool0oooo02T0oooo000Y0?ooo`030000003oool0oooo09l0
oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo02P0
oooo000X0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0B03oool00`000000oooo
0?ooo`1/0?ooo`030000003oool0oooo02P0oooo000X0?ooo`030000003oool0oooo0:00oooo00<0
00000?ooo`3oool0B03oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo02L0oooo000W
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`1^
0?ooo`030000003oool0oooo02L0oooo000W0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0A`3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo02H0oooo000V0?ooo`03
0000003oool0oooo0:80oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`1a0?ooo`03
0000003oool0oooo02D0oooo000U0?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool0
AP3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo02@0oooo000U0?ooo`030000003o
ool0oooo0:<0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`1c0?ooo`030000003o
ool0oooo02@0oooo000T0?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool0@`3oool0
0`000000oooo0?ooo`1e0?ooo`030000003oool0oooo02<0oooo000T0?ooo`030000003oool0oooo
0:D0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`1e0?ooo`030000003oool0oooo
02<0oooo000S0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool0@P3oool00`000000
oooo0?ooo`1f0?ooo`030000003oool0oooo02<0oooo000S0?ooo`030000003oool0oooo0:H0oooo
00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo0280oooo
000R0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?oo
o`1h0?ooo`030000003oool0oooo0280oooo000R0?ooo`030000003oool0oooo0:L0oooo00<00000
0?ooo`3oool0@03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0240oooo000Q0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`1k0?oo
o`030000003oool0oooo0200oooo000P0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3o
ool0?`3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo01l0oooo000P0?ooo`030000
003oool0oooo0:X0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1m0?ooo`030000
003oool0oooo01l0oooo000O0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0?@3o
ool00`000000oooo0?ooo`1o0?ooo`030000003oool0oooo01h0oooo000O0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`200?ooo`030000003oool0
oooo01h0oooo000N0?ooo`030000003oool0oooo0:`0oooo00<000000?ooo`3oool0?03oool00`00
0000oooo0?ooo`200?ooo`030000003oool0oooo01h0oooo000N0?ooo`030000003oool0oooo0:`0
oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo01d0
oooo000M0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool0>`3oool00`000000oooo
0?ooo`220?ooo`030000003oool0oooo01d0oooo000M0?ooo`030000003oool0oooo0:d0oooo00<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo01`0oooo000M
0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`24
0?ooo`030000003oool0oooo01`0oooo000L0?ooo`030000003oool0oooo0:h0oooo00<000000?oo
o`3oool0>@3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo01/0oooo000L0?ooo`03
0000003oool0oooo0:l0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`270?ooo`03
0000003oool0oooo01/0oooo000K0?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3oool0
=`3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo01/0oooo000K0?ooo`030000003o
ool0oooo0;00oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`290?ooo`030000003o
ool0oooo01X0oooo000J0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0=P3oool0
0`000000oooo0?ooo`290?ooo`030000003oool0oooo01X0oooo000J0?ooo`030000003oool0oooo
0;40oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo
01T0oooo000J0?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0=03oool00`000000
oooo0?ooo`2<0?ooo`030000003oool0oooo01T0oooo000I0?ooo`030000003oool0oooo0;80oooo
00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo01P0oooo
000I0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?oo
o`2>0?ooo`030000003oool0oooo01P0oooo000H0?ooo`030000003oool0oooo0;<0oooo00<00000
0?ooo`3oool0<`3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo01P0oooo000H0?oo
o`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`2@0?oo
o`030000003oool0oooo01L0oooo000G0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3o
ool0<03oool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo01L0oooo000G0?ooo`030000
003oool0oooo0;D0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`2B0?ooo`030000
003oool0oooo01H0oooo000G0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo01H0oooo000F0?ooo`030000003oool0
oooo0;H0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`2D0?ooo`030000003oool0
oooo01D0oooo000F0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`2E0?ooo`030000003oool0oooo01D0oooo000E0?ooo`030000003oool0oooo0;L0
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo01D0
oooo000E0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0;@3oool00`000000oooo
0?ooo`2G0?ooo`030000003oool0oooo01@0oooo000E0?ooo`030000003oool0oooo0;L0oooo00<0
00000?ooo`3oool0;03oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo01@0oooo000D
0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`2I
0?ooo`030000003oool0oooo01<0oooo000D0?ooo`030000003oool0oooo0;T0oooo00<000000?oo
o`3oool0:P3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo01<0oooo000D0?ooo`03
0000003oool0oooo0;T0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`2J0?ooo`03
0000003oool0oooo01<0oooo000D0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0
:@3oool00`000000oooo0?ooo`2K0?ooo`030000003oool0oooo01<0oooo000C0?ooo`030000003o
ool0oooo0;X0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`2L0?ooo`030000003o
ool0oooo01<0oooo000C0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`2M0?ooo`030000003oool0oooo0180oooo000C0?ooo`030000003oool0oooo
0;X0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo
0180oooo000B0?ooo`030000003oool0oooo0;/0oooo00<000000?ooo`3oool09`3oool00`000000
oooo0?ooo`2N0?ooo`030000003oool0oooo0180oooo000B0?ooo`030000003oool0oooo0;/0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo0140oooo
000A0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0903oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0140oooo000A0?ooo`030000003oool0oooo0;d0oooo00<00000
0?ooo`3oool0903oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo0100oooo000A0?oo
o`030000003oool0oooo0;d0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`2S0?oo
o`030000003oool0oooo0100oooo000A0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3o
ool08`3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo0100oooo000A0?ooo`030000
003oool0oooo0;d0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`2T0?ooo`030000
003oool0oooo0100oooo000@0?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo0100oooo000@0?ooo`030000003oool0
oooo0;h0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`2V0?ooo`030000003oool0
oooo00l0oooo000@0?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool0803oool00`00
0000oooo0?ooo`2W0?ooo`030000003oool0oooo00l0oooo000@0?ooo`030000003oool0oooo0;h0
oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00h0
oooo000@0?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool07P3oool00`000000oooo
0?ooo`2Y0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0oooo0<00oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo00h0oooo000?
0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`2Z
0?ooo`030000003oool0oooo00h0oooo000?0?ooo`030000003oool0oooo0<00oooo00<000000?oo
o`3oool0703oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00h0oooo000>0?ooo`03
0000003oool0oooo0<40oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`2/0?ooo`03
0000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool0
6`3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003o
ool0oooo0<40oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2]0?ooo`030000003o
ool0oooo00d0oooo000>0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`2^0?ooo`030000003oool0oooo00d0oooo000>0?ooo`030000003oool0oooo
0<40oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`2^0?ooo`030000003oool0oooo
00d0oooo000=0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`2`0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0<<0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo00`0oooo
000=0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`2a0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo0<<0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`2b0?ooo`030000003oool0oooo00`0oooo000=0?oo
o`030000003oool0oooo0<<0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`2b0?oo
o`030000003oool0oooo00`0oooo000<0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000
003oool0oooo0<@0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`2e0?ooo`030000
003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0
oooo0<D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0
oooo00/0oooo000<0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`2f0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo0<D0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo00/0
oooo000<0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`2h0?ooo`030000003oool0oooo00/0oooo000;0?ooo`030000003oool0oooo0<H0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo00/0oooo000;
0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`2j
0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0<H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo00X0oooo000;0?ooo`03
0000003oool0oooo0<H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`2k0?ooo`03
0000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003o
ool0oooo0<L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`2l0?ooo`030000003o
ool0oooo00X0oooo000;0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`2m0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo
0<L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo
00X0oooo000;0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`2n0?ooo`030000003oool0oooo00X0oooo000;0?ooo`030000003oool0oooo0<L0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo00X0oooo
000;0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`2o0?ooo`030000003oool0oooo00X0oooo000:0?ooo`030000003oool0oooo0<P0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo00T0oooo000:0?oo
o`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`310?oo
o`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000
003oool0oooo0<T0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`320?ooo`030000
003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`330?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0
oooo0<T0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`340?ooo`030000003oool0
oooo00T0oooo000:0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`340?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<T0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo00T0
oooo000:0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`350?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<T0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo00T0oooo000:
0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`36
0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo0<T0oooo00D000000?oo
o`3oool0oooo000000390?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo
0<X0oooo00<000000?ooo`000000bP3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000
oooo0?ooo`3:0?ooo`030000003oool000000<X0oooo00<000000?ooo`3oool02@3oool000X0oooo
00<000000?ooo`3oool0bP3oool200000</0oooo00<000000?ooo`3oool02@3oool000X0oooo00<0
00000?ooo`3oool0b@3oool300000</0oooo00<000000?ooo`3oool02@3oool000X0oooo00<00000
0?ooo`3oool0b03oool400000</0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?oo
o`3oool0b03oool400000</0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3o
ool0b@3oool200000<`0oooo00<000000?ooo`3oool02@3oool000X0oooo00<000000?ooo`3oool0
b03oool00`000000oooo0000003<0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003o
ool0oooo0<P0oooo00<000000?ooo`000000c03oool00`000000oooo0?ooo`090?ooo`002P3oool0
0`000000oooo0?ooo`370?ooo`050000003oool0oooo0?ooo`000000b`3oool00`000000oooo0?oo
o`090?ooo`002P3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0b@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`360?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`090?oo
o`002P3oool00`000000oooo0?ooo`350?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0b@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`350?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`090?ooo`002P3o
ool00`000000oooo0?ooo`340?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0b@3o
ool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`340?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`00
0000oooo0?ooo`330?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0b@3oool00`00
0000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo
0?ooo`320?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0b03oool00`000000oooo
0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`002P3oool00`000000oooo0?ooo`31
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`09
0?ooo`002`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2n0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`00
2`3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
a`3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2m0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`002`3oool0
0`000000oooo0?ooo`2m0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0a`3oool0
0`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000
oooo0?ooo`2k0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0aP3oool00`000000
oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0aP3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`000000oooo0?oo
o`2j0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0aP3oool00`000000oooo0?oo
o`0:0?ooo`002`3oool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0aP3oool00`000000oooo0?ooo`0:0?ooo`00303oool00`000000oooo0?ooo`2h0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0;0?oo
o`00303oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool0a@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`2g0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0;0?ooo`00303o
ool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0a@3o
ool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`2f0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`2e0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0a03oool00`00
0000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo
0?ooo`2d0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0a03oool00`000000oooo
0?ooo`0;0?ooo`003@3oool00`000000oooo0?ooo`2b0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool0``3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`2b
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`0<
0?ooo`003@3oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool0``3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`2a0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`0<0?ooo`00
3@3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
``3oool00`000000oooo0?ooo`0<0?ooo`003P3oool00`000000oooo0?ooo`2^0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0=0?ooo`003P3oool0
0`000000oooo0?ooo`2^0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0`@3oool0
0`000000oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0=0?ooo`003P3oool00`000000
oooo0?ooo`2]0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0`@3oool00`000000
oooo0?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`2/0?ooo`030000003oool0oooo01`0oooo
00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0=0?ooo`003`3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0`03oool00`000000oooo0?oo
o`0>0?ooo`003`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool0`03oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo0?ooo`2Y0?oo
o`030000003oool0oooo01h0oooo00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`0>0?oo
o`003`3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3o
ool0`03oool00`000000oooo0?ooo`0>0?ooo`003`3oool00`000000oooo0?ooo`2X0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool0_`3oool00`000000oooo0?ooo`0>0?ooo`00403o
ool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0_P3o
ool00`000000oooo0?ooo`0?0?ooo`00403oool00`000000oooo0?ooo`2V0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0?0?ooo`00403oool00`00
0000oooo0?ooo`2U0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0_@3oool00`00
0000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo
0?ooo`2S0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0_@3oool00`000000oooo
0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool0_@3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`000000oooo0?ooo`2R
0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0@
0?ooo`004@3oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo02@0oooo00<000000?oo
o`3oool0_03oool00`000000oooo0?ooo`0A0?ooo`004P3oool00`000000oooo0?ooo`2P0?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`0A0?ooo`00
4P3oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
^P3oool00`000000oooo0?ooo`0B0?ooo`004`3oool00`000000oooo0?ooo`2N0?ooo`030000003o
ool0oooo02L0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`0B0?ooo`004`3oool0
0`000000oooo0?ooo`2M0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0^P3oool0
0`000000oooo0?ooo`0B0?ooo`00503oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo
02P0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`0C0?ooo`00503oool00`000000
oooo0?ooo`2K0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0^@3oool00`000000
oooo0?ooo`0C0?ooo`00503oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo02X0oooo
00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`0C0?ooo`00503oool00`000000oooo0?oo
o`2J0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?oo
o`0C0?ooo`00503oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo02/0oooo00<00000
0?ooo`3oool0^@3oool00`000000oooo0?ooo`0C0?ooo`005@3oool00`000000oooo0?ooo`2H0?oo
o`030000003oool0oooo02`0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0D0?oo
o`005@3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3o
ool0]`3oool00`000000oooo0?ooo`0D0?ooo`005@3oool00`000000oooo0?ooo`2F0?ooo`030000
003oool0oooo02h0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0E0?ooo`005P3o
ool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0]P3o
ool00`000000oooo0?ooo`0E0?ooo`005P3oool00`000000oooo0?ooo`2D0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0E0?ooo`005`3oool00`00
0000oooo0?ooo`2C0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0]@3oool00`00
0000oooo0?ooo`0F0?ooo`005`3oool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo0300
oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0F0?ooo`005`3oool00`000000oooo
0?ooo`2B0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0]03oool00`000000oooo
0?ooo`0G0?ooo`00603oool00`000000oooo0?ooo`2@0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0]03oool00`000000oooo0?ooo`0G0?ooo`00603oool00`000000oooo0?ooo`2?
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0H
0?ooo`006@3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo03<0oooo00<000000?oo
o`3oool0/P3oool00`000000oooo0?ooo`0H0?ooo`006@3oool00`000000oooo0?ooo`2=0?ooo`03
0000003oool0oooo03@0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0H0?ooo`00
6P3oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0
/@3oool00`000000oooo0?ooo`0I0?ooo`006P3oool00`000000oooo0?ooo`2;0?ooo`030000003o
ool0oooo03D0oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`0I0?ooo`006P3oool0
0`000000oooo0?ooo`2:0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0/03oool0
0`000000oooo0?ooo`0J0?ooo`006`3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo
03H0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`0J0?ooo`006`3oool00`000000
oooo0?ooo`280?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0[`3oool00`000000
oooo0?ooo`0K0?ooo`00703oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo03L0oooo
00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`0K0?ooo`00703oool00`000000oooo0?oo
o`260?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0[P3oool00`000000oooo0?oo
o`0K0?ooo`007@3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo03X0oooo00<00000
0?ooo`3oool0[@3oool00`000000oooo0?ooo`0L0?ooo`007@3oool00`000000oooo0?ooo`240?oo
o`030000003oool0oooo03X0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`0L0?oo
o`007@3oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3o
ool0[03oool00`000000oooo0?ooo`0M0?ooo`007P3oool00`000000oooo0?ooo`220?ooo`030000
003oool0oooo03/0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0M0?ooo`007P3o
ool00`000000oooo0?ooo`210?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`0N0?ooo`007`3oool00`000000oooo0?ooo`200?ooo`030000003oool0
oooo03`0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`0N0?ooo`007`3oool00`00
0000oooo0?ooo`1o0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`0N0?ooo`00803oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo03h0
oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0O0?ooo`00803oool00`000000oooo
0?ooo`1m0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo
0?ooo`0O0?ooo`008@3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0400oooo00<0
00000?ooo`3oool0Z03oool00`000000oooo0?ooo`0P0?ooo`008P3oool00`000000oooo0?ooo`1j
0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0Q
0?ooo`008P3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo0440oooo00<000000?oo
o`3oool0YP3oool00`000000oooo0?ooo`0R0?ooo`008`3oool00`000000oooo0?ooo`1g0?ooo`03
0000003oool0oooo0480oooo00<000000?ooo`3oool0YP3oool00`000000oooo0?ooo`0R0?ooo`00
8`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0
Y@3oool00`000000oooo0?ooo`0S0?ooo`00903oool00`000000oooo0?ooo`1e0?ooo`030000003o
ool0oooo04<0oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`0S0?ooo`00903oool0
0`000000oooo0?ooo`1e0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0Y@3oool0
0`000000oooo0?ooo`0S0?ooo`009@3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo
04@0oooo00<000000?ooo`3oool0Y03oool00`000000oooo0?ooo`0T0?ooo`009@3oool00`000000
oooo0?ooo`1b0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0X`3oool00`000000
oooo0?ooo`0T0?ooo`009P3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo04H0oooo
00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`0U0?ooo`009`3oool00`000000oooo0?oo
o`1_0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0V0?ooo`009`3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo04L0oooo00<00000
0?ooo`3oool0X03oool00`000000oooo0?ooo`0W0?ooo`00:03oool00`000000oooo0?ooo`1]0?oo
o`030000003oool0oooo04P0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`0W0?oo
o`00:03oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3o
ool0W`3oool00`000000oooo0?ooo`0X0?ooo`00:@3oool00`000000oooo0?ooo`1[0?ooo`030000
003oool0oooo04T0oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`0X0?ooo`00:P3o
ool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0WP3o
ool00`000000oooo0?ooo`0Y0?ooo`00:`3oool00`000000oooo0?ooo`1X0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool0W03oool00`000000oooo0?ooo`0Z0?ooo`00:`3oool00`00
0000oooo0?ooo`1W0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0V`3oool00`00
0000oooo0?ooo`0[0?ooo`00;03oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo04`0
oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`0[0?ooo`00;03oool00`000000oooo
0?ooo`1U0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0V`3oool00`000000oooo
0?ooo`0[0?ooo`00;@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo04h0oooo00<0
00000?ooo`3oool0VP3oool00`000000oooo0?ooo`0/0?ooo`00;@3oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0/
0?ooo`00;P3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0V@3oool00`000000oooo0?ooo`0]0?ooo`00;`3oool00`000000oooo0?ooo`1P0?ooo`03
0000003oool0oooo04l0oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`0^0?ooo`00
;`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0
U`3oool00`000000oooo0?ooo`0_0?ooo`00<03oool00`000000oooo0?ooo`1M0?ooo`030000003o
ool0oooo0580oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`0_0?ooo`00<@3oool0
0`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0U@3oool0
0`000000oooo0?ooo`0`0?ooo`00<P3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`0a0?ooo`00<P3oool00`000000
oooo0?ooo`1J0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0T`3oool00`000000
oooo0?ooo`0b0?ooo`00<`3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo05@0oooo
00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`0b0?ooo`00=03oool00`000000oooo0?oo
o`1G0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?oo
o`0c0?ooo`00=@3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo05D0oooo00<00000
0?ooo`3oool0T@3oool00`000000oooo0?ooo`0d0?ooo`00=P3oool00`000000oooo0?ooo`1C0?oo
o`030000003oool0oooo05H0oooo00<000000?ooo`3oool0T03oool00`000000oooo0?ooo`0e0?oo
o`00=`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo05H0oooo00<000000?ooo`3o
ool0S`3oool00`000000oooo0?ooo`0f0?ooo`00=`3oool00`000000oooo0?ooo`1A0?ooo`030000
003oool0oooo05P0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0g0?ooo`00>03o
ool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0S@3o
ool00`000000oooo0?ooo`0g0?ooo`00>@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0
oooo05T0oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`0h0?ooo`00>P3oool00`00
0000oooo0?ooo`1<0?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0R`3oool00`00
0000oooo0?ooo`0i0?ooo`00>P3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo05X0
oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`0j0?ooo`00>`3oool00`000000oooo
0?ooo`1:0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0RP3oool00`000000oooo
0?ooo`0j0?ooo`00?03oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo05/0oooo00<0
00000?ooo`3oool0R@3oool00`000000oooo0?ooo`0k0?ooo`00?@3oool00`000000oooo0?ooo`17
0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0l
0?ooo`00?P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo05`0oooo00<000000?oo
o`3oool0Q`3oool00`000000oooo0?ooo`0m0?ooo`00?`3oool00`000000oooo0?ooo`140?ooo`03
0000003oool0oooo05h0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`0n0?ooo`00
@03oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0
Q03oool00`000000oooo0?ooo`0o0?ooo`00@@3oool00`000000oooo0?ooo`110?ooo`030000003o
ool0oooo05l0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`100?ooo`00@P3oool0
0`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0PP3oool0
0`000000oooo0?ooo`110?ooo`00@`3oool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo
0600oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`120?ooo`00A03oool00`000000
oooo0?ooo`0l0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0P03oool00`000000
oooo0?ooo`130?ooo`00A@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0680oooo
00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`140?ooo`00AP3oool00`000000oooo0?oo
o`0i0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?oo
o`150?ooo`00A`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0O@3oool00`000000oooo0?ooo`160?ooo`00B03oool00`000000oooo0?ooo`0f0?oo
o`030000003oool0oooo06@0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`170?oo
o`00B@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3o
ool0NP3oool00`000000oooo0?ooo`180?ooo`00BP3oool00`000000oooo0?ooo`0b0?ooo`030000
003oool0oooo06H0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`190?ooo`00B`3o
ool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0N03o
ool00`000000oooo0?ooo`1:0?ooo`00C03oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo06L0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`1;0?ooo`00C@3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0MP3oool00`00
0000oooo0?ooo`1<0?ooo`00CP3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo06P0
oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`1=0?ooo`00C`3oool00`000000oooo
0?ooo`0[0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0M03oool00`000000oooo
0?ooo`1>0?ooo`00D03oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo06T0oooo00<0
00000?ooo`3oool0L`3oool00`000000oooo0?ooo`1?0?ooo`00D@3oool00`000000oooo0?ooo`0W
0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`1@
0?ooo`00DP3oool2000002L0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`1_0?oo
o`800000D`3oool005@0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1/0?ooo`03
0000003oool0oooo06h0oooo00<000000?ooo`3oool0D`3oool005D0oooo00<000000?ooo`3oool0
8P3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0
E03oool005H0oooo0P00000Q0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0J`3o
ool2000005L0oooo001H0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0KP3oool0
0`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo05L0oooo001I0?ooo`030000003oool0oooo
01`0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo
05P0oooo001J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0K`3oool00`000000
oooo0?ooo`1X0?ooo`030000003oool0oooo05T0oooo001K0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo05X0oooo
001L0?ooo`800000603oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo06@0oooo0P00
001M0?ooo`00GP3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0780oooo00<00000
0?ooo`3oool0H`3oool00`000000oooo0?ooo`1M0?ooo`00G`3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo0780oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`1N0?oo
o`00H03oool200000180oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`1P0?ooo`80
0000H@3oool00680oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1c0?ooo`030000
003oool0oooo05l0oooo00<000000?ooo`3oool0H@3oool006<0oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0000001f0?ooo`030000003oool0oooo05d0oooo0P00001T0?ooo`00I03oool3
000000X0oooo0`00001f0?ooo`030000003oool0oooo05/0oooo0P00001V0?ooo`00I`3oool20000
00L0oooo1P00001d0?ooo`030000003oool0oooo05T0oooo0P00001X0?ooo`00J@3oool00`000000
oooo0?ooo`040?ooo`D00000MP3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo06P0
oooo001Z0?ooo`030000003oool0oooo00<0oooo1000001g0?ooo`030000003oool0oooo05H0oooo
00<000000?ooo`3oool0J@3oool006/0oooo0P0000030?ooo`<00000N03oool00`000000oooo0?oo
o`1D0?ooo`800000K03oool006d0oooo1000001j0?ooo`030000003oool0oooo0580oooo0P00001^
0?ooo`00K`3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo0540oooo00<000000?oo
o`3oool0KP3oool00700oooo0P00001i0?ooo`030000003oool0oooo04l0oooo0P00001a0?ooo`00
LP3oool2000007L0oooo00<000000?ooo`3oool0C@3oool2000007<0oooo001d0?ooo`800000M@3o
ool00`000000oooo0?ooo`1;0?ooo`800000M@3oool007H0oooo0P00001c0?ooo`030000003oool0
oooo04T0oooo0P00001g0?ooo`00N03oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool0M`3oool007T0oooo0P00001a0?ooo`030000003oool0oooo04D0
oooo0P00001j0?ooo`00N`3oool2000006l0oooo00<000000?ooo`3oool0@`3oool2000007`0oooo
001m0?ooo`800000K@3oool00`000000oooo0?ooo`110?ooo`800000OP3oool007l0oooo0`00001Z
0?ooo`030000003oool0oooo03h0oooo0`0000200?ooo`00PP3oool2000006P0oooo00<000000?oo
o`3oool0?03oool2000008<0oooo00240?ooo`800000IP3oool00`000000oooo0?ooo`0j0?ooo`80
0000Q@3oool008H0oooo0`00001S0?ooo`030000003oool0oooo03L0oooo0`0000270?ooo`00R@3o
ool200000640oooo00<000000?ooo`3oool0=@3oool2000008X0oooo002;0?ooo`800000H03oool0
0`000000oooo0?ooo`0b0?ooo`800000S03oool008d0oooo0`00001M0?ooo`030000003oool0oooo
02l0oooo0`00002>0?ooo`00T03oool2000005/0oooo00<000000?ooo`3oool0;@3oool200000940
oooo002B0?ooo`<00000F03oool00`000000oooo0?ooo`0Z0?ooo`<00000T`3oool009D0oooo1000
001D0?ooo`030000003oool0oooo02D0oooo1@00002F0?ooo`00V@3oool300000540oooo0`00000S
0?ooo`800000V`3oool009`0oooo0P00001=0?ooo`D000008@3oool2000009d0oooo002N0?ooo`D0
0000B03oool5000001`0oooo1@00002O0?ooo`00X`3oool5000004@0oooo1000000I0?ooo`<00000
Y03oool00:P0oooo0`0000110?ooo`<000005@3oool500000:L0oooo002[0?ooo`D00000?@3oool2
00000100oooo1@00002/0?ooo`00/03oool5000003P0oooo0P00000;0?ooo`D00000/@3oool00;D0
oooo1`00000b0?ooo`030000003oool0oooo0080oooo1`00002f0?ooo`00_03oool;00000240oooo
2`00002m0?ooo`00a`3ooolQ00000<P0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-2.06817, -2.06817, 0.0143956, \
0.0143956}}]
}, Closed]],

Cell[BoxData[
    \(\(MomentumPlot2[LabMomenta];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[Graphics[RGBColor[0, 1, 0]], plot1, Graphics[RGBColor[1, 0, 0]], 
        plot2, AspectRatio \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.19058 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.251202 0.971704 0.251202 [
[ 0 0 0 0 ]
[ 1 1.19058 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 m
1 0 L
1 1.19058 L
0 1.19058 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
currentlinewidth 1 Mabswid
currentlinewidth dup scale setlinewidth
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto fill
-5. 1.25 moveto 0 0 lineto
-5. -1.25 lineto
-5. 1.25 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 1 0 r
.5 Mabswid
[ ] 0 setdash
.02381 .9717 m
.24136 1.09731 L
s
% Start of user PostScript
0. 0. 0.8660254 0.5 MAarrowhead1
% End of user PostScript
.24136 1.09731 m
.72577 .52001 L
s
% Start of user PostScript
0.8660254 0.5 2.7943882 -1.7981333 MAarrowhead1
% End of user PostScript
.02381 .9717 m
.72577 .52001 L
s
% Start of user PostScript
0. 0. 2.7943882 -1.7981333 MAarrowhead1
% End of user PostScript
.72577 .52001 m
.47534 .94674 L
s
% Start of user PostScript
2.7943882 -1.7981333 1.7974881 -0.099377776 MAarrowhead1
% End of user PostScript
.72577 .52001 m
.97619 .09328 L
s
% Start of user PostScript
2.7943882 -1.7981333 3.7912883 -3.4968889 MAarrowhead1
% End of user PostScript
1 0 0 r
.02381 .9717 m
.54729 1.16223 L
s
% Start of user PostScript
0. 0. 2.0838941 0.75847542 MAarrowhead1
% End of user PostScript
.54729 1.16223 m
.72577 .52001 L
s
% Start of user PostScript
2.0838941 0.75847542 2.7943882 -1.7981333 MAarrowhead1
% End of user PostScript
.02381 .9717 m
.72577 .52001 L
s
% Start of user PostScript
0. 0. 2.7943882 -1.7981333 MAarrowhead1
% End of user PostScript
.72577 .52001 m
.78127 1.01167 L
s
% Start of user PostScript
2.7943882 -1.7981333 3.0153568 0.15909764 MAarrowhead1
% End of user PostScript
.72577 .52001 m
.67026 .02835 L
s
% Start of user PostScript
2.7943882 -1.7981333 2.5734196 -3.7553643 MAarrowhead1
% End of user PostScript
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{241.875, 287.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005Z0001[b000`400?l00000o`00003oo`3ooom[
0?ooo`00o`3ooom[0?ooo`00o`3ooom[0?ooo`00o`3ooom[0?ooo`00o`3ooom[0?ooo`00o`3ooom[
0?ooo`00o`3ooom[0?ooo`00o`3ooom[0?ooo`00o`3ooom[0?ooo`00o`3ooom[0?ooo`00lP3oool0
0`3o0000oooo0?ooo`1e0?ooo`00lP3oool00`3o0000oooo0?ooo`1e0?ooo`00lP3oool20?l007H0
oooo003b0?ooo`80o`00MP3oool00?40oooo103o001e0?ooo`00l@3oool40?l007D0oooo003a0?oo
o`D0o`00M03oool00?40oooo1@3o001d0?ooo`00l@3oool30?l007H0oooo003c0?ooo`030?l0003o
ool0oooo07@0oooo003c0?ooo`030?l0003oool0oooo07@0oooo003c0?ooo`030?l0003oool0oooo
07@0oooo003c0?ooo`030?l0003oool0oooo07@0oooo003c0?ooo`030?l0003oool0oooo07@0oooo
003d0?ooo`030?l0003oool0oooo07<0oooo003d0?ooo`030?l0003oool0oooo07<0oooo003d0?oo
o`030?l0003oool0oooo07<0oooo003d0?ooo`030?l0003oool0oooo07<0oooo003d0?ooo`030?l0
003oool0oooo07<0oooo003d0?ooo`030?l0003oool0oooo07<0oooo003d0?ooo`030?l0003oool0
oooo07<0oooo003d0?ooo`030?l0003oool0oooo07<0oooo003d0?ooo`030?l0003oool0oooo07<0
oooo003e0?ooo`030?l0003oool0oooo06T0oooo00<00?l00?ooo`3oool01P3oool00?D0oooo00<0
o`000?ooo`3oool0J03oool2003o00P0oooo003e0?ooo`030?l0003oool0oooo06L0oooo0P00o`09
0?ooo`00m@3oool00`3o0000oooo0?ooo`1V0?ooo`<00?l02@3oool00?D0oooo00<0o`000?ooo`3o
ool0I@3oool4003o00T0oooo003e0?ooo`030?l0003oool0oooo06@0oooo1@00o`090?ooo`00m@3o
ool00`3o0000oooo0?ooo`1S0?ooo`D00?l02P3oool00?D0oooo00<0o`000?ooo`3oool0I03oool4
003o00X0oooo003e0?ooo`030?l0003oool0oooo06@0oooo00@00?l00?ooo`00o`000?l02P3oool0
0?H0oooo00<0o`000?ooo`3oool0H`3oool00`00o`00oooo0?ooo`0;0?ooo`00mP3oool00`3o0000
oooo0?ooo`1R0?ooo`03003o003oool0oooo00`0oooo003f0?ooo`030?l0003oool0oooo0680oooo
00<00?l00?ooo`3oool0303oool00?H0oooo00<0o`000?ooo`3oool0H@3oool00`00o`00oooo0?oo
o`0=0?ooo`00mP3oool00`3o0000oooo0?ooo`1P0?ooo`03003o003oool0oooo00h0oooo003f0?oo
o`030?l0003oool0oooo0600oooo00<00?l00?ooo`3oool03P3oool00?H0oooo00<0o`000?ooo`3o
ool0G`3oool00`00o`00oooo0?ooo`0?0?ooo`00mP3oool00`3o0000oooo0?ooo`1O0?ooo`03003o
003oool0oooo00l0oooo003f0?ooo`030?l0003oool0oooo05h0oooo00<00?l00?ooo`3oool0403o
ool00?L0oooo00<0o`000?ooo`3oool0G03oool00`00o`00oooo0?ooo`0A0?ooo`00m`3oool00`3o
0000oooo0?ooo`1L0?ooo`03003o003oool0oooo0140oooo003g0?ooo`030?l0003oool0oooo05/0
oooo00<00?l00?ooo`3oool04P3oool00?L0oooo00<0o`000?ooo`3oool0F`3oool00`00o`00oooo
0?ooo`0B0?ooo`00m`3oool00`3o0000oooo0?ooo`1J0?ooo`03003o003oool0oooo01<0oooo003g
0?ooo`030?l0003oool0oooo05T0oooo00<00?l00?ooo`3oool0503oool00?L0oooo00<0o`000?oo
o`3oool0F@3oool00`00o`00oooo0?ooo`0D0?ooo`00m`3oool00`3o0000oooo0?ooo`1H0?ooo`03
003o003oool0oooo01D0oooo003h0?ooo`030?l0003oool0oooo05L0oooo00<00?l00?ooo`3oool0
5@3oool00?P0oooo00<0o`000?ooo`3oool0EP3oool00`00o`00oooo0?ooo`0F0?ooo`00n03oool0
0`3o0000oooo0?ooo`1F0?ooo`03003o003oool0oooo01H0oooo003h0?ooo`030?l0003oool0oooo
05D0oooo00<00?l00?ooo`3oool05`3oool00?P0oooo00<0o`000?ooo`3oool0E03oool00`00o`00
oooo0?ooo`0H0?ooo`00n03oool00`3o0000oooo0?ooo`1D0?ooo`03003o003oool0oooo01P0oooo
003h0?ooo`030?l0003oool0oooo05<0oooo00<00?l00?ooo`3oool06@3oool00?P0oooo00<0o`00
0?ooo`3oool0D`3oool00`00o`00oooo0?ooo`0I0?ooo`00n03oool00`3o0000oooo0?ooo`1B0?oo
o`03003o003oool0oooo01X0oooo003i0?ooo`030?l0003oool0oooo0500oooo00<00?l00?ooo`3o
ool06`3oool00?T0oooo00<0o`000?ooo`3oool0D03oool00`00o`00oooo0?ooo`0K0?ooo`00n@3o
ool00`3o0000oooo0?ooo`1?0?ooo`03003o003oool0oooo01`0oooo003i0?ooo`030?l0003oool0
oooo04l0oooo00<00?l00?ooo`3oool0703oool00?T0oooo00<0o`000?ooo`3oool0CP3oool00`00
o`00oooo0?ooo`0M0?ooo`00n@3oool00`3o0000oooo0?ooo`1>0?ooo`03003o003oool0oooo01d0
oooo003i0?ooo`030?l0003oool0oooo04d0oooo00<00?l00?ooo`3oool07P3oool00?T0oooo00<0
o`000?ooo`3oool0C03oool00`00o`00oooo0?ooo`0O0?ooo`00n@3oool00`3o0000oooo0?ooo`1<
0?ooo`03003o003oool0oooo01l0oooo003j0?ooo`030?l0003oool0oooo04X0oooo00<00?l00?oo
o`3oool0803oool00?X0oooo00<0o`000?ooo`3oool0BP3oool00`00o`00oooo0?ooo`0P0?ooo`00
nP3oool00`3o0000oooo0?ooo`190?ooo`03003o003oool0oooo0240oooo003j0?ooo`030?l0003o
ool0oooo04P0oooo00<00?l00?ooo`3oool08P3oool00?X0oooo00<0o`000?ooo`3oool0B03oool0
0`00o`00oooo0?ooo`0R0?ooo`00nP3oool00`3o0000oooo0?ooo`170?ooo`03003o003oool0oooo
02<0oooo003j0?ooo`030?l0003oool0oooo04L0oooo00<00?l00?ooo`3oool08`3oool00?X0oooo
00<0o`000?ooo`3oool0AP3oool00`00o`00oooo0?ooo`0T0?ooo`00nP3oool00`3o0000oooo0?oo
o`150?ooo`03003o003oool0oooo02D0oooo003k0?ooo`030?l0003oool0oooo04@0oooo00<00?l0
0?ooo`3oool09@3oool00?/0oooo00<0o`000?ooo`3oool0@`3oool00`00o`00oooo0?ooo`0V0?oo
o`00n`3oool00`3o0000oooo0?ooo`130?ooo`03003o003oool0oooo02H0oooo003k0?ooo`030?l0
003oool0oooo0480oooo00<00?l00?ooo`3oool09`3oool00?/0oooo00<0o`000?ooo`3oool0@P3o
ool00`00o`00oooo0?ooo`0W0?ooo`00n`3oool00`3o0000oooo0?ooo`110?ooo`03003o003oool0
oooo02P0oooo003k0?ooo`030?l0003oool0oooo0400oooo00<00?l00?ooo`3oool0:@3oool00?/0
oooo00<0o`000?ooo`3oool0@03oool00`00o`00oooo0?ooo`0Y0?ooo`00n`3oool00`3o0000oooo
0?ooo`0o0?ooo`03003o003oool0oooo02X0oooo003l0?ooo`030?l0003oool0oooo03h0oooo00<0
0?l00?ooo`3oool0:P3oool00?`0oooo00<0o`000?ooo`3oool0?@3oool00`00o`00oooo0?ooo`0[
0?ooo`00o03oool00`3o0000oooo0?ooo`0l0?ooo`03003o003oool0oooo02`0oooo003l0?ooo`03
0?l0003oool0oooo03`0oooo00<00?l00?ooo`3oool0;03oool00?`0oooo00<0o`000?ooo`3oool0
>`3oool00`00o`00oooo0?ooo`0]0?ooo`00o03oool00`3o0000oooo0?ooo`0k0?ooo`03003o003o
ool0oooo02d0oooo003l0?ooo`030?l0003oool0oooo03X0oooo00<00?l00?ooo`3oool0;P3oool0
0?`0oooo00<0o`000?ooo`3oool0>@3oool00`00o`00oooo0?ooo`0_0?ooo`00o03oool00`3o0000
oooo0?ooo`0i0?ooo`03003o003oool0oooo02l0oooo003m0?ooo`030?l0003oool0oooo03L0oooo
00<00?l00?ooo`3oool0<03oool00?d0oooo00<0o`000?ooo`3oool0=`3oool00`00o`00oooo0?oo
o`0`0?ooo`00o@3oool00`3o0000oooo0?ooo`0f0?ooo`03003o003oool0oooo0340oooo003m0?oo
o`030?l0003oool0oooo03H0oooo00<00?l00?ooo`3oool0<@3oool00?d0oooo00<0o`000?ooo`3o
ool0=@3oool00`00o`00oooo0?ooo`0b0?ooo`00o@3oool00`3o0000oooo0?ooo`0d0?ooo`03003o
003oool0oooo03<0oooo003m0?ooo`030?l0003oool0oooo03@0oooo00<00?l00?ooo`3oool0<`3o
ool00?d0oooo00<0o`000?ooo`3oool0<`3oool00`00o`00oooo0?ooo`0d0?ooo`00o@3oool00`3o
0000oooo0?ooo`0c0?ooo`03003o003oool0oooo03@0oooo003n0?ooo`030?l0003oool0oooo0340
oooo00<00?l00?ooo`3oool0=@3oool00?h0oooo00<0o`000?ooo`3oool0<03oool00`00o`00oooo
0?ooo`0f0?ooo`00oP3oool00`3o0000oooo0?ooo`0`0?ooo`03003o003oool0oooo03H0oooo003n
0?ooo`030?l0003oool0oooo02l0oooo00<00?l00?ooo`3oool0=`3oool00?h0oooo00<0o`000?oo
o`3oool0;`3oool00`00o`00oooo0?ooo`0g0?ooo`00oP3oool00`3o0000oooo0?ooo`0^0?ooo`03
003o003oool0oooo03P0oooo003n0?ooo`030?l0003oool0oooo02h0oooo00<00?l00?ooo`3oool0
>03oool00?h0oooo00<0o`000?ooo`3oool0;@3oool00`00o`00oooo0?ooo`0i0?ooo`00oP3oool0
0`3o0000oooo0?ooo`0/0?ooo`03003o003oool0oooo03X0oooo003o0?ooo`030?l0003oool0oooo
02/0oooo00<00?l00?ooo`3oool0>P3oool00?l0oooo00<0o`000?ooo`3oool0:P3oool00`00o`00
oooo0?ooo`0k0?ooo`00o`3oool00`3o0000oooo0?ooo`0Z0?ooo`03003o003oool0oooo03/0oooo
003o0?ooo`030?l0003oool0oooo02T0oooo00<00?l00?ooo`3oool0?03oool00?l0oooo00<0o`00
0?ooo`3oool0:03oool00`00o`00oooo0?ooo`0m0?ooo`00o`3oool00`3o0000oooo0?ooo`0X0?oo
o`03003o003oool0oooo03d0oooo003o0?ooo`030?l0003oool0oooo02L0oooo00<00?l00?ooo`3o
ool0?P3oool00?l0oooo00<0o`000?ooo`3oool09`3oool00`00o`00oooo0?ooo`0n0?ooo`00o`3o
ool00`3o0000oooo0?ooo`0V0?ooo`03003o003oool0oooo03l0oooo003o0?ooo`40oooo00<0o`00
0?ooo`3oool0903oool00`00o`00oooo0?ooo`100?ooo`00o`3oool10?ooo`030?l0003oool0oooo
02@0oooo00<00?l00?ooo`3oool0@03oool00?l0oooo0@3oool00`3o0000oooo0?ooo`0S0?ooo`03
003o003oool0oooo0440oooo003o0?ooo`40oooo00<0o`000?ooo`3oool08`3oool00`00o`00oooo
0?ooo`110?ooo`00o`3oool10?ooo`030?l0003oool0oooo0280oooo00<00?l00?ooo`3oool0@P3o
ool00?l0oooo0@3oool00`3o0000oooo0?ooo`0R0?ooo`03003o003oool0oooo0480oooo003o0?oo
o`40oooo00<0o`000?ooo`3oool08@3oool00`00o`00oooo0?ooo`130?ooo`00o`3oool10?ooo`03
0?l0003oool0oooo0200oooo00<00?l00?ooo`3oool0A03oool00?l0oooo0@3oool00`3o0000oooo
0?ooo`0P0?ooo`03003o003oool0oooo04@0oooo003o0?ooo`80oooo00<0o`000?ooo`3oool07P3o
ool00`00o`00oooo0?ooo`150?ooo`00o`3oool20?ooo`030?l0003oool0oooo01h0oooo00<00?l0
0?ooo`3oool0A@3oool00?l0oooo0P3oool00`3o0000oooo0?ooo`0M0?ooo`03003o003oool0oooo
04H0oooo003o0?ooo`80oooo00<0o`000?ooo`3oool0703oool00`00o`00oooo0?ooo`170?ooo`00
o`3oool20?ooo`030?l0003oool0oooo01`0oooo00<00?l00?ooo`3oool0A`3oool00?l0oooo0P3o
ool00`3o0000oooo0?ooo`0K0?ooo`03003o003oool0oooo04P0oooo003o0?ooo`80oooo00<0o`00
0?ooo`3oool06`3oool00`00o`00oooo0?ooo`180?ooo`00o`3oool20?ooo`030?l0003oool0oooo
01X0oooo00<00?l00?ooo`3oool0B@3oool00?l0oooo0`3oool00`3o0000oooo0?ooo`0H0?ooo`03
003o003oool0oooo04X0oooo003o0?ooo`<0oooo00<0o`000?ooo`3oool0603oool00`00o`00oooo
0?ooo`1:0?ooo`00o`3oool30?ooo`030?l0003oool0oooo01L0oooo00<00?l00?ooo`3oool0B`3o
ool00?l0oooo0`3oool00`3o0000oooo0?ooo`0G0?ooo`03003o003oool0oooo04/0oooo003o0?oo
o`<0oooo00<0o`000?ooo`3oool05P3oool00`00o`00oooo0?ooo`1<0?ooo`00o`3oool30?ooo`03
0?l0003oool0oooo01H0oooo00<00?l00?ooo`3oool0C03oool00?l0oooo0`3oool00`3o0000oooo
0?ooo`0E0?ooo`03003o003oool0oooo04d0oooo003o0?ooo`<0oooo00<0o`000?ooo`3oool0503o
ool00`00o`00oooo0?ooo`1>0?ooo`00o`3oool30?ooo`030?l0003oool0oooo01@0oooo00<00?l0
0?ooo`3oool0CP3oool00?l0oooo103oool00`3o0000oooo0?ooo`0B0?ooo`03003o003oool0oooo
04l0oooo003o0?ooo`@0oooo00<0o`000?ooo`3oool04P3oool00`00o`00oooo0?ooo`1?0?ooo`00
o`3oool40?ooo`030?l0003oool0oooo0140oooo00<00?l00?ooo`3oool0D03oool00?l0oooo103o
ool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo0540oooo003o0?ooo`@0oooo00<0o`00
0?ooo`3oool0403oool00`00o`00oooo0?ooo`1A0?ooo`00o`3oool40?ooo`030?l0003oool0oooo
00l0oooo00<00?l00?ooo`3oool0DP3oool00?l0oooo103oool00`3o0000oooo0?ooo`0?0?ooo`03
003o003oool0oooo0580oooo003o0?ooo`@0oooo00<0o`000?ooo`3oool03P3oool00`00o`00oooo
0?ooo`1C0?ooo`00o`3oool40?ooo`030?l0003oool0oooo00h0oooo00<00?l00?ooo`3oool0D`3o
ool00?l0oooo1@3oool00`3o0000oooo0?ooo`0<0?ooo`03003o003oool0oooo05@0oooo003o0?oo
o`D0oooo00<0o`000?ooo`3oool02`3oool00`00o`00oooo0?ooo`1E0?ooo`00o`3oool50?ooo`03
0?l0003oool0oooo00/0oooo00<00?l00?ooo`3oool0E@3oool00?l0oooo1@3oool00`3o0000oooo
0?ooo`0:0?ooo`03003o003oool0oooo05H0oooo003o0?ooo`D0oooo00<0o`000?ooo`3oool02P3o
ool00`00o`00oooo0?ooo`1F0?ooo`00o`3oool50?ooo`030?l0003oool0oooo00T0oooo00<00?l0
0?ooo`3oool0E`3oool00?l0oooo1@3oool00`3o0000oooo0?ooo`080?ooo`03003o003oool0oooo
05P0oooo003o0?ooo`D0oooo00<0o`000?ooo`3oool0203oool00`00o`00oooo0?ooo`1H0?ooo`00
o`3oool50?ooo`030?l0003oool0oooo00L0oooo00<00?l00?ooo`3oool0F@3oool00?l0oooo1P3o
ool00`3o0000oooo0?ooo`060?ooo`03003o003oool0oooo05T0oooo003o0?ooo`H0oooo00<0o`00
0?ooo`3oool01@3oool00`00o`00oooo0?ooo`1J0?ooo`00o`3oool60?ooo`030?l0003oool0oooo
00@0oooo00<00?l00?ooo`3oool0F`3oool00?l0oooo1P3oool00`3o0000oooo0?ooo`040?ooo`03
003o003oool0oooo05/0oooo003o0?ooo`H0oooo00<0o`000?ooo`3oool00`3oool00`00o`00oooo
0?ooo`1L0?ooo`00o`3oool60?ooo`030?l0003oool0oooo00<0oooo00<00?l00?ooo`3oool0G03o
ool00?l0oooo1P3oool00`3o0000oooo0?ooo`020?ooo`03003o003oool0oooo05d0oooo003o0?oo
o`H0oooo00<0o`000?ooo`3oool00P3oool00`00o`00oooo0?ooo`1M0?ooo`00o`3oool60?ooo`05
0?l0003oool0oooo0?ooo`00o`00H03oool00?l0oooo1`3oool00`3o0000oooo003o001Q0?ooo`00
o`3oool70?ooo`030?l0003oool00?l00640oooo003o0?ooo`L0oooo00<0o`00003o003oool0H@3o
ool00?l0oooo1`3oool00`3o00000?l00?ooo`1Q0?ooo`00o`3oool60?ooo`80o`00H`3oool00?l0
oooo0P3oool60?l006<0oooo003o0?ooo`P0o`00H`3oool00?l0oooo203o001S0?ooo`00o`3oool8
0?l006<0oooo003n0?ooo`030?l0003oool0o`0000L0o`00HP3oool00?`0oooo0P3o00030?ooo`03
0?l00000o`00o`0000@0o`00HP3oool00?/0oooo00<0o`000?ooo`3oool00P3oool00`00o`00oooo
0?l000050?l00680oooo003i0?ooo`80o`00103oool00`00o`00oooo003o00020?ooo`@0o`00HP3o
ool00?P0oooo00<0o`000?ooo`3oool00`3oool01000o`00oooo0?ooo`00o`020?ooo`040?l0003o
ool0oooo0?l00680oooo003f0?ooo`80o`001P3oool00`00o`00oooo003o00020?ooo`050?l0003o
ool0oooo0?ooo`3o0000HP3oool00?D0oooo00<0o`000?ooo`3oool01@3oool00`00o`00oooo003o
00030?ooo`050?l0003oool0oooo0?ooo`3o0000HP3oool00?<0oooo0P3o00070?ooo`04003o003o
ool0oooo003o00<0oooo00D0o`000?ooo`3oool0oooo0?l0001R0?ooo`00l@3oool20?l000P0oooo
00@00?l00?ooo`3oool00?l00`3oool00`3o0000oooo0?ooo`020?ooo`030?l0003oool0oooo0600
oooo003`0?ooo`030?l0003oool0oooo00L0oooo00D00?l00?ooo`3oool0oooo003o00030?ooo`03
0?l0003oool0oooo00<0oooo00<0o`000?ooo`3oool0G`3oool00>h0oooo0P3o00090?ooo`05003o
003oool0oooo0?ooo`00o`00103oool00`3o0000oooo0?ooo`030?ooo`030?l0003oool0oooo05l0
oooo003]0?ooo`030?l0003oool0oooo00T0oooo00D00?l00?ooo`3oool0oooo003o00040?ooo`03
0?l0003oool0oooo00<0oooo00<0o`000?ooo`3oool0G`3oool00>/0oooo0P3o000;0?ooo`05003o
003oool0oooo0?ooo`00o`00103oool00`3o0000oooo0?ooo`040?ooo`030?l0003oool0oooo05l0
oooo003Z0?ooo`030?l0003oool0oooo00X0oooo00D00?l00?ooo`3oool0oooo003o00050?ooo`03
0?l0003oool0oooo00@0oooo00<0o`000?ooo`3oool0G`3oool00>P0oooo0P3o000<0?ooo`03003o
003oool0oooo0080oooo00<00?l00?ooo`3oool00`3oool00`3o0000oooo0?ooo`040?ooo`030?l0
003oool0oooo05l0oooo003W0?ooo`030?l0003oool0oooo00/0oooo00<00?l00?ooo`3oool00P3o
ool00`00o`00oooo0?ooo`030?ooo`030?l0003oool0oooo00D0oooo00<0o`000?ooo`3oool0G`3o
ool00>D0oooo0P3o000=0?ooo`03003o003oool0oooo00<0oooo00<00?l00?ooo`3oool00`3oool0
0`3o0000oooo0?ooo`050?ooo`030?l0003oool0oooo05l0oooo003S0?ooo`80o`003`3oool00`00
o`00oooo0?ooo`020?ooo`03003o003oool0oooo00@0oooo00<0o`000?ooo`3oool01@3oool00`3o
0000oooo0?ooo`1O0?ooo`00hP3oool00`3o0000oooo0?ooo`0>0?ooo`03003o003oool0oooo0080
oooo00<00?l00?ooo`3oool01@3oool00`3o0000oooo0?ooo`060?ooo`030?l0003oool0oooo05h0
oooo003P0?ooo`80o`00403oool00`00o`00oooo0?ooo`030?ooo`03003o003oool0oooo00@0oooo
00<0o`000?ooo`3oool01`3oool00`3o0000oooo0?ooo`1N0?ooo`00g`3oool00`3o0000oooo0?oo
o`0?0?ooo`03003o003oool0oooo00<0oooo00<00?l00?ooo`3oool01@3oool00`3o0000oooo0?oo
o`070?ooo`030?l0003oool0oooo05h0oooo003M0?ooo`80o`004@3oool00`00o`00oooo0?ooo`04
0?ooo`03003o003oool0oooo00D0oooo00<0o`000?ooo`3oool01`3oool00`3o0000oooo0?ooo`1N
0?ooo`00g03oool00`3o0000oooo0?ooo`0@0?ooo`03003o003oool0oooo00@0oooo00<00?l00?oo
o`3oool01P3oool00`3o0000oooo0?ooo`070?ooo`030?l0003oool0oooo05h0oooo003J0?ooo`80
o`004`3oool00`00o`00oooo0?ooo`030?ooo`03003o003oool0oooo00H0oooo00<0o`000?ooo`3o
ool0203oool00`3o0000oooo0?ooo`1N0?ooo`00f@3oool00`3o0000oooo0?ooo`0B0?ooo`03003o
003oool0oooo00@0oooo00<00?l00?ooo`3oool01P3oool00`3o0000oooo0?ooo`080?ooo`030?l0
003oool0oooo05h0oooo003G0?ooo`80o`00503oool00`00o`00oooo0?ooo`040?ooo`03003o003o
ool0oooo00L0oooo00<0o`000?ooo`3oool0203oool00`3o0000oooo0?ooo`1N0?ooo`00e@3oool2
0?l001D0oooo00<00?l00?ooo`3oool01@3oool00`00o`00oooo0?ooo`060?ooo`030?l0003oool0
oooo00T0oooo00<0o`000?ooo`3oool0GP3oool00=@0oooo00<0o`000?ooo`3oool0503oool00`00
o`00oooo0?ooo`050?ooo`03003o003oool0oooo00L0oooo00<0o`000?ooo`3oool02P3oool00`3o
0000oooo0?ooo`1M0?ooo`00dP3oool20?l001H0oooo00<00?l00?ooo`3oool01@3oool00`00o`00
oooo0?ooo`080?ooo`030?l0003oool0oooo00X0oooo00<0o`000?ooo`3oool0G@3oool00=40oooo
00<0o`000?ooo`3oool05P3oool00`00o`00oooo0?ooo`050?ooo`03003o003oool0oooo00P0oooo
00<0o`000?ooo`3oool02P3oool00`3o0000oooo0?ooo`1M0?ooo`00c`3oool20?l001P0oooo00<0
0?l00?ooo`3oool01@3oool00`00o`00oooo0?ooo`080?ooo`030?l0003oool0oooo00/0oooo00<0
o`000?ooo`3oool0G@3oool00<h0oooo00<0o`000?ooo`3oool05`3oool00`00o`00oooo0?ooo`06
0?ooo`03003o003oool0oooo00P0oooo00<0o`000?ooo`3oool02`3oool00`3o0000oooo0?ooo`1M
0?ooo`00c03oool20?l001T0oooo00<00?l00?ooo`3oool01P3oool00`00o`00oooo0?ooo`090?oo
o`030?l0003oool0oooo00/0oooo00<0o`000?ooo`3oool0G@3oool00</0oooo00<0o`000?ooo`3o
ool0603oool00`00o`00oooo0?ooo`070?ooo`03003o003oool0oooo00T0oooo00<0o`000?ooo`3o
ool02`3oool00`3o0000oooo0?ooo`1M0?ooo`00b@3oool20?l001X0oooo00<00?l00?ooo`3oool0
1`3oool00`00o`00oooo0?ooo`090?ooo`030?l0003oool0oooo00`0oooo00<0o`000?ooo`3oool0
G@3oool00<L0oooo0P3o000L0?ooo`03003o003oool0oooo00H0oooo00<00?l00?ooo`3oool02P3o
ool00`3o0000oooo0?ooo`0<0?ooo`030?l0003oool0oooo05d0oooo00360?ooo`030?l0003oool0
oooo01/0oooo00<00?l00?ooo`3oool01`3oool00`00o`00oooo0?ooo`0:0?ooo`030?l0003oool0
oooo00d0oooo00<0o`000?ooo`3oool0G03oool00<@0oooo0P3o000M0?ooo`03003o003oool0oooo
00L0oooo00<00?l00?ooo`3oool02P3oool00`3o0000oooo0?ooo`0>0?ooo`030?l0003oool0oooo
05`0oooo00330?ooo`030?l0003oool0oooo01`0oooo00<00?l00?ooo`3oool0203oool00`00o`00
oooo0?ooo`0:0?ooo`030?l0003oool0oooo00h0oooo00<0o`000?ooo`3oool0G03oool00<40oooo
0P3o000N0?ooo`03003o003oool0oooo00P0oooo00<00?l00?ooo`3oool02`3oool00`3o0000oooo
0?ooo`0>0?ooo`030?l0003oool0oooo05`0oooo00300?ooo`030?l0003oool0oooo01d0oooo00<0
0?l00?ooo`3oool0203oool00`00o`00oooo0?ooo`0<0?ooo`030?l0003oool0oooo00h0oooo00<0
o`000?ooo`3oool0G03oool00;h0oooo0P3o000O0?ooo`03003o003oool0oooo00T0oooo00<00?l0
0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0?0?ooo`030?l0003oool0oooo05`0oooo002l0?oo
o`80o`008@3oool00`00o`00oooo0?ooo`080?ooo`03003o003oool0oooo00`0oooo00<0o`000?oo
o`3oool03`3oool00`3o0000oooo0?ooo`1L0?ooo`00^`3oool00`3o0000oooo0?ooo`0P0?ooo`03
003o003oool0oooo00T0oooo00<00?l00?ooo`3oool0303oool00`3o0000oooo0?ooo`0?0?ooo`03
0?l0003oool0oooo05`0oooo002i0?ooo`80o`008P3oool00`00o`00oooo0?ooo`090?ooo`03003o
003oool0oooo00`0oooo00<0o`000?ooo`3oool04@3oool00`3o0000oooo0?ooo`1K0?ooo`00^03o
ool00`3o0000oooo0?ooo`0Q0?ooo`03003o003oool0oooo00T0oooo00<00?l00?ooo`3oool03@3o
ool00`3o0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo05/0oooo002f0?ooo`80o`008`3oool0
0`00o`00oooo0?ooo`0:0?ooo`03003o003oool0oooo00d0oooo00<0o`000?ooo`3oool04@3oool0
0`3o0000oooo0?ooo`1K0?ooo`00]@3oool00`3o0000oooo0?ooo`0R0?ooo`03003o003oool0oooo
00X0oooo00<00?l00?ooo`3oool03P3oool00`3o0000oooo0?ooo`0A0?ooo`030?l0003oool0oooo
05/0oooo002c0?ooo`80o`009@3oool00`00o`00oooo0?ooo`0:0?ooo`03003o003oool0oooo00d0
oooo00<0o`000?ooo`3oool04P3oool00`3o0000oooo0?ooo`1K0?ooo`00/P3oool00`3o0000oooo
0?ooo`0T0?ooo`03003o003oool0oooo00X0oooo00<00?l00?ooo`3oool03P3oool00`3o0000oooo
0?ooo`0B0?ooo`030?l0003oool0oooo05/0oooo002`0?ooo`80o`009P3oool00`00o`00oooo0?oo
o`0:0?ooo`03003o003oool0oooo00l0oooo00<0o`000?ooo`3oool04P3oool00`3o0000oooo0?oo
o`1K0?ooo`00[P3oool20?l002L0oooo00<00?l00?ooo`3oool02`3oool00`00o`00oooo0?ooo`0?
0?ooo`030?l0003oool0oooo0180oooo00<0o`000?ooo`3oool0F`3oool00:d0oooo00<0o`000?oo
o`3oool09P3oool00`00o`00oooo0?ooo`0;0?ooo`03003o003oool0oooo00l0oooo00<0o`000?oo
o`3oool04`3oool00`3o0000oooo0?ooo`1K0?ooo`00Z`3oool20?l002P0oooo00<00?l00?ooo`3o
ool0303oool00`00o`00oooo0?ooo`0?0?ooo`030?l0003oool0oooo01@0oooo00<0o`000?ooo`3o
ool0FP3oool00:X0oooo00<0o`000?ooo`3oool0:03oool00`00o`00oooo0?ooo`0;0?ooo`03003o
003oool0oooo0100oooo00<0o`000?ooo`3oool0503oool00`3o0000oooo0?ooo`1J0?ooo`00Z03o
ool20?l002X0oooo00<00?l00?ooo`3oool0303oool00`00o`00oooo0?ooo`0?0?ooo`030?l0003o
ool0oooo01D0oooo00<0o`000?ooo`3oool0FP3oool00:L0oooo00<0o`000?ooo`3oool0:@3oool0
0`00o`00oooo0?ooo`0<0?ooo`03003o003oool0oooo0100oooo00<0o`000?ooo`3oool05@3oool0
0`3o0000oooo0?ooo`1J0?ooo`00Y@3oool20?l002/0oooo00<00?l00?ooo`3oool0303oool00`00
o`00oooo0?ooo`0A0?ooo`030?l0003oool0oooo01D0oooo00<0o`000?ooo`3oool0FP3oool00:@0
oooo00<0o`000?ooo`3oool0:P3oool00`00o`00oooo0?ooo`0=0?ooo`03003o003oool0oooo0140
oooo00<0o`000?ooo`3oool05@3oool00`3o0000oooo0?ooo`1J0?ooo`00XP3oool20?l002`0oooo
00<00?l00?ooo`3oool03@3oool00`00o`00oooo0?ooo`0A0?ooo`030?l0003oool0oooo01H0oooo
00<0o`000?ooo`3oool0FP3oool00:00oooo0P3o000^0?ooo`03003o003oool0oooo00d0oooo00<0
0?l00?ooo`3oool04@3oool00`3o0000oooo0?ooo`0F0?ooo`030?l0003oool0oooo05X0oooo002O
0?ooo`030?l0003oool0oooo02d0oooo00<00?l00?ooo`3oool03@3oool00`00o`00oooo0?ooo`0B
0?ooo`030?l0003oool0oooo01H0oooo00<0o`000?ooo`3oool0FP3oool009d0oooo0P3o000_0?oo
o`03003o003oool0oooo00d0oooo00<00?l00?ooo`3oool04`3oool00`3o0000oooo0?ooo`0G0?oo
o`030?l0003oool0oooo05T0oooo002L0?ooo`030?l0003oool0oooo02h0oooo00<00?l00?ooo`3o
ool03P3oool00`00o`00oooo0?ooo`0B0?ooo`030?l0003oool0oooo01P0oooo00<0o`000?ooo`3o
ool0F@3oool009X0oooo0P3o000`0?ooo`03003o003oool0oooo00h0oooo00<00?l00?ooo`3oool0
4`3oool00`3o0000oooo0?ooo`0H0?ooo`030?l0003oool0oooo05T0oooo002I0?ooo`030?l0003o
ool0oooo02l0oooo00<00?l00?ooo`3oool03`3oool00`00o`00oooo0?ooo`0C0?ooo`030?l0003o
ool0oooo01P0oooo00<0o`000?ooo`3oool0F@3oool009L0oooo0P3o000b0?ooo`03003o003oool0
oooo00h0oooo00<00?l00?ooo`3oool04`3oool00`3o0000oooo0?ooo`0I0?ooo`030?l0003oool0
oooo05T0oooo002F0?ooo`030?l0003oool0oooo0340oooo00<00?l00?ooo`3oool03P3oool00`00
o`00oooo0?ooo`0D0?ooo`030?l0003oool0oooo01T0oooo00<0o`000?ooo`3oool0F@3oool009@0
oooo0P3o000c0?ooo`03003o003oool0oooo00l0oooo00<00?l00?ooo`3oool0503oool00`3o0000
oooo0?ooo`0I0?ooo`030?l0003oool0oooo05T0oooo002B0?ooo`80o`00=03oool00`00o`00oooo
0?ooo`0?0?ooo`03003o003oool0oooo01D0oooo00<0o`000?ooo`3oool06@3oool00`3o0000oooo
0?ooo`1I0?ooo`00T@3oool00`3o0000oooo0?ooo`0c0?ooo`03003o003oool0oooo0100oooo00<0
0?l00?ooo`3oool0503oool00`3o0000oooo0?ooo`0J0?ooo`030?l0003oool0oooo05T0oooo002?
0?ooo`80o`00=@3oool00`00o`00oooo0?ooo`0@0?ooo`03003o003oool0oooo01D0oooo00<0o`00
0?ooo`3oool06`3oool00`3o0000oooo0?ooo`1H0?ooo`00SP3oool00`3o0000oooo0?ooo`0e0?oo
o`03003o003oool0oooo00l0oooo00<00?l00?ooo`3oool05P3oool00`3o0000oooo0?ooo`0K0?oo
o`030?l0003oool0oooo05P0oooo002<0?ooo`80o`00=`3oool00`00o`00oooo0?ooo`0@0?ooo`03
003o003oool0oooo01D0oooo00<0o`000?ooo`3oool0703oool00`3o0000oooo0?ooo`1H0?ooo`00
R`3oool00`3o0000oooo0?ooo`0f0?ooo`03003o003oool0oooo0100oooo00<00?l00?ooo`3oool0
5P3oool00`3o0000oooo0?ooo`0L0?ooo`030?l0003oool0oooo05P0oooo00290?ooo`80o`00>03o
ool00`00o`00oooo0?ooo`0A0?ooo`03003o003oool0oooo01H0oooo00<0o`000?ooo`3oool0703o
ool00`3o0000oooo0?ooo`1H0?ooo`00R03oool00`3o0000oooo0?ooo`0g0?ooo`03003o003oool0
oooo0140oooo00<00?l00?ooo`3oool05`3oool00`3o0000oooo0?ooo`0L0?ooo`030?l0003oool0
oooo05P0oooo00260?ooo`80o`00>@3oool00`00o`00oooo0?ooo`0B0?ooo`03003o003oool0oooo
01H0oooo00<0o`000?ooo`3oool07@3oool00`3o0000oooo0?ooo`1H0?ooo`00Q03oool20?l003/0
oooo00<00?l00?ooo`3oool04@3oool00`00o`00oooo0?ooo`0G0?ooo`030?l0003oool0oooo01d0
oooo00<0o`000?ooo`3oool0F03oool008<0oooo00<0o`000?ooo`3oool0>P3oool00`00o`00oooo
0?ooo`0A0?ooo`03003o003oool0oooo01P0oooo00<0o`000?ooo`3oool07@3oool00`3o0000oooo
0?ooo`1H0?ooo`00P@3oool20?l003`0oooo00<00?l00?ooo`3oool04P3oool00`00o`00oooo0?oo
o`0H0?ooo`030?l0003oool0oooo01h0oooo00<0o`000?ooo`3oool0E`3oool00800oooo00<0o`00
0?ooo`3oool0>`3oool00`00o`00oooo0?ooo`0B0?ooo`03003o003oool0oooo01P0oooo00<0o`00
0?ooo`3oool07`3oool00`3o0000oooo0?ooo`1G0?ooo`00OP3oool20?l003d0oooo00<00?l00?oo
o`3oool04`3oool00`00o`00oooo0?ooo`0H0?ooo`030?l0003oool0oooo01l0oooo00<0o`000?oo
o`3oool0E`3oool007d0oooo00<0o`000?ooo`3oool0?03oool00`00o`00oooo0?ooo`0C0?ooo`03
003o003oool0oooo01T0oooo00<0o`000?ooo`3oool07`3oool00`3o0000oooo0?ooo`1G0?ooo`00
N`3oool20?l003h0oooo00<00?l00?ooo`3oool04`3oool00`00o`00oooo0?ooo`0I0?ooo`030?l0
003oool0oooo0200oooo00<0o`000?ooo`3oool0E`3oool007T0oooo0P3o00100?ooo`03003o003o
ool0oooo01<0oooo00<00?l00?ooo`3oool06@3oool00`3o0000oooo0?ooo`0P0?ooo`030?l0003o
ool0oooo05L0oooo001h0?ooo`030?l0003oool0oooo03l0oooo00<00?l00?ooo`3oool04`3oool0
0`00o`00oooo0?ooo`0J0?ooo`030?l0003oool0oooo0200oooo00<0o`000?ooo`3oool0E`3oool0
07H0oooo0P3o00110?ooo`03003o003oool0oooo01@0oooo00<00?l00?ooo`3oool06P3oool00`3o
0000oooo0?ooo`0P0?ooo`030?l0003oool0oooo05L0oooo001e0?ooo`030?l0003oool0oooo0400
oooo00<00?l00?ooo`3oool0503oool00`00o`00oooo0?ooo`0J0?ooo`030?l0003oool0oooo0240
oooo00<0o`000?ooo`3oool0E`3oool007<0oooo0P3o00120?ooo`03003o003oool0oooo01@0oooo
00<00?l00?ooo`3oool06`3oool00`3o0000oooo0?ooo`0R0?ooo`030?l0003oool0oooo05H0oooo
001b0?ooo`030?l0003oool0oooo0440oooo00<00?l00?ooo`3oool05@3oool00`00o`00oooo0?oo
o`0K0?ooo`030?l0003oool0oooo0280oooo00<0o`000?ooo`3oool0EP3oool00700oooo0P3o0014
0?ooo`03003o003oool0oooo01@0oooo00<00?l00?ooo`3oool0703oool00`3o0000oooo0?ooo`0R
0?ooo`030?l0003oool0oooo05H0oooo001_0?ooo`030?l0003oool0oooo04<0oooo00<00?l00?oo
o`3oool05@3oool00`00o`00oooo0?ooo`0K0?ooo`030?l0003oool0oooo02<0oooo00<0o`000?oo
o`3oool0EP3oool006d0oooo0P3o00150?ooo`03003o003oool0oooo01D0oooo00<00?l00?ooo`3o
ool0703oool00`3o0000oooo0?ooo`0S0?ooo`030?l0003oool0oooo05H0oooo001[0?ooo`80o`00
AP3oool00`00o`00oooo0?ooo`0E0?ooo`03003o003oool0oooo01d0oooo00<0o`000?ooo`3oool0
8`3oool00`3o0000oooo0?ooo`1F0?ooo`00JP3oool00`3o0000oooo0?ooo`150?ooo`03003o003o
ool0oooo01H0oooo00<00?l00?ooo`3oool0703oool00`3o0000oooo0?ooo`0T0?ooo`030?l0003o
ool0oooo05H0oooo001X0?ooo`80o`00A`3oool00`00o`00oooo0?ooo`0F0?ooo`03003o003oool0
oooo01d0oooo00<0o`000?ooo`3oool0903oool00`3o0000oooo0?ooo`1F0?ooo`00I`3oool00`3o
0000oooo0?ooo`170?ooo`03003o003oool0oooo01H0oooo00<00?l00?ooo`3oool07@3oool00`3o
0000oooo0?ooo`0T0?ooo`030?l0003oool0oooo05H0oooo001U0?ooo`80o`00B@3oool00`00o`00
oooo0?ooo`0F0?ooo`03003o003oool0oooo01h0oooo00<0o`000?ooo`3oool09@3oool00`3o0000
oooo0?ooo`1E0?ooo`00I03oool00`3o0000oooo0?ooo`180?ooo`03003o003oool0oooo01L0oooo
00<00?l00?ooo`3oool07@3oool00`3o0000oooo0?ooo`0V0?ooo`030?l0003oool0oooo05D0oooo
001R0?ooo`80o`00BP3oool00`00o`00oooo0?ooo`0G0?ooo`03003o003oool0oooo01h0oooo00<0
o`000?ooo`3oool09P3oool00`3o0000oooo0?ooo`1E0?ooo`00H@3oool00`3o0000oooo0?ooo`19
0?ooo`03003o003oool0oooo01L0oooo00<00?l00?ooo`3oool07`3oool00`3o0000oooo0?ooo`0V
0?ooo`030?l0003oool0oooo05D0oooo001O0?ooo`80o`00B`3oool00`00o`00oooo0?ooo`0H0?oo
o`03003o003oool0oooo01h0oooo00<0o`000?ooo`3oool09`3oool00`3o0000oooo0?ooo`1E0?oo
o`00G@3oool20?l004d0oooo00<00?l00?ooo`3oool05`3oool00`00o`00oooo0?ooo`0O0?ooo`03
0?l0003oool0oooo02L0oooo00<0o`000?ooo`3oool0E@3oool005`0oooo00<0o`000?ooo`3oool0
C03oool00`00o`00oooo0?ooo`0H0?ooo`03003o003oool0oooo01l0oooo00<0o`000?ooo`3oool0
9`3oool00`3o0000oooo0?ooo`1E0?ooo`00FP3oool20?l004h0oooo00<00?l00?ooo`3oool0603o
ool00`00o`00oooo0?ooo`0P0?ooo`030?l0003oool0oooo02L0oooo00<0o`000?ooo`3oool0E@3o
ool005T0oooo00<0o`000?ooo`3oool0C@3oool00`00o`00oooo0?ooo`0H0?ooo`03003o003oool0
oooo0200oooo00<0o`000?ooo`3oool0:03oool00`3o0000oooo0?ooo`1E0?ooo`00E`3oool20?l0
04l0oooo00<00?l00?ooo`3oool06@3oool00`00o`00oooo0?ooo`0P0?ooo`030?l0003oool0oooo
02T0oooo00<0o`000?ooo`3oool0E03oool005H0oooo00<0o`000?ooo`3oool0CP3oool00`00o`00
oooo0?ooo`0I0?ooo`03003o003oool0oooo0240oooo00<0o`000?ooo`3oool0:@3oool00`3o0000
oooo0?ooo`1D0?ooo`00E03oool20?l00540oooo00<00?l00?ooo`3oool06@3oool00`00o`00oooo
0?ooo`0Q0?ooo`030?l0003oool0oooo02T0oooo00<0o`000?ooo`3oool0E03oool005<0oooo00<0
o`000?ooo`3oool0D03oool00`00o`00oooo0?ooo`0I0?ooo`03003o003oool0oooo0240oooo00<0
o`000?ooo`3oool0:P3oool00`3o0000oooo0?ooo`1D0?ooo`00D@3oool20?l00580oooo00<00?l0
0?ooo`3oool06@3oool00`00o`00oooo0?ooo`0R0?ooo`030?l0003oool0oooo02X0oooo00<0o`00
0?ooo`3oool0E03oool004l0oooo0P3o001C0?ooo`03003o003oool0oooo01X0oooo00<00?l00?oo
o`3oool08P3oool00`3o0000oooo0?ooo`0Z0?ooo`030?l0003oool0oooo05@0oooo001>0?ooo`03
0?l0003oool0oooo0580oooo00<00?l00?ooo`3oool06P3oool00`00o`00oooo0?ooo`0R0?ooo`03
0?l0003oool0oooo02/0oooo00<0o`000?ooo`3oool0E03oool004`0oooo0P3o001D0?ooo`03003o
003oool0oooo01/0oooo00<00?l00?ooo`3oool08P3oool00`3o0000oooo0?ooo`0[0?ooo`030?l0
003oool0oooo05@0oooo001;0?ooo`030?l0003oool0oooo05@0oooo00<00?l00?ooo`3oool06P3o
ool00`00o`00oooo0?ooo`0S0?ooo`030?l0003oool0oooo02/0oooo00<0o`000?ooo`3oool0E03o
ool004T0oooo0P3o001F0?ooo`03003o003oool0oooo01X0oooo00<00?l00?ooo`3oool0903oool0
0`3o0000oooo0?ooo`0/0?ooo`030?l0003oool0oooo05<0oooo00180?ooo`030?l0003oool0oooo
05D0oooo00<00?l00?ooo`3oool06`3oool00`00o`00oooo0?ooo`0S0?ooo`030?l0003oool0oooo
02d0oooo00<0o`000?ooo`3oool0D`3oool004H0oooo0P3o001G0?ooo`03003o003oool0oooo01/0
oooo00<00?l00?ooo`3oool0903oool00`3o0000oooo0?ooo`0]0?ooo`030?l0003oool0oooo05<0
oooo00140?ooo`80o`00F03oool00`00o`00oooo0?ooo`0L0?ooo`03003o003oool0oooo02@0oooo
00<0o`000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`1C0?ooo`00@`3oool00`3o0000oooo0?oo
o`1G0?ooo`03003o003oool0oooo01`0oooo00<00?l00?ooo`3oool09@3oool00`3o0000oooo0?oo
o`0]0?ooo`030?l0003oool0oooo05<0oooo00110?ooo`80o`00F@3oool00`00o`00oooo0?ooo`0M
0?ooo`03003o003oool0oooo02@0oooo00<0o`000?ooo`3oool0;P3oool00`3o0000oooo0?ooo`1C
0?ooo`00@03oool00`3o0000oooo0?ooo`1I0?ooo`03003o003oool0oooo01`0oooo00<00?l00?oo
o`3oool09@3oool00`3o0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo05<0oooo000n0?ooo`80
o`00F`3oool00`00o`00oooo0?ooo`0L0?ooo`03003o003oool0oooo02H0oooo00<0o`000?ooo`3o
ool0;P3oool00`3o0000oooo0?ooo`1C0?ooo`00?@3oool00`3o0000oooo0?ooo`1J0?ooo`03003o
003oool0oooo01d0oooo00<00?l00?ooo`3oool09@3oool00`3o0000oooo0?ooo`0_0?ooo`030?l0
003oool0oooo05<0oooo000k0?ooo`80o`00G03oool00`00o`00oooo0?ooo`0M0?ooo`03003o003o
ool0oooo02H0oooo00<0o`000?ooo`3oool0<03oool00`3o0000oooo0?ooo`1B0?ooo`00>P3oool0
0`3o0000oooo0?ooo`1K0?ooo`03003o003oool0oooo01h0oooo00<00?l00?ooo`3oool09P3oool0
0`3o0000oooo0?ooo`0`0?ooo`030?l0003oool0oooo0580oooo000h0?ooo`80o`00G@3oool00`00
o`00oooo0?ooo`0N0?ooo`03003o003oool0oooo02L0oooo00<0o`000?ooo`3oool0<03oool00`3o
0000oooo0?ooo`1B0?ooo`00=P3oool20?l005l0oooo00<00?l00?ooo`3oool07@3oool00`00o`00
oooo0?ooo`0W0?ooo`030?l0003oool0oooo0340oooo00<0o`000?ooo`3oool0DP3oool003D0oooo
00<0o`000?ooo`3oool0GP3oool00`00o`00oooo0?ooo`0N0?ooo`03003o003oool0oooo02L0oooo
00<0o`000?ooo`3oool0<@3oool00`3o0000oooo0?ooo`1B0?ooo`00<`3oool20?l00600oooo00<0
0?l00?ooo`3oool07P3oool00`00o`00oooo0?ooo`0X0?ooo`030?l0003oool0oooo0340oooo00<0
o`000?ooo`3oool0DP3oool00380oooo00<0o`000?ooo`3oool0G`3oool00`00o`00oooo0?ooo`0O
0?ooo`03003o003oool0oooo02L0oooo00<0o`000?ooo`3oool0<P3oool00`3o0000oooo0?ooo`1B
0?ooo`00<03oool20?l00640oooo00<00?l00?ooo`3oool07`3oool00`00o`00oooo0?ooo`0X0?oo
o`030?l0003oool0oooo0380oooo00<0o`000?ooo`3oool0DP3oool002l0oooo00<0o`000?ooo`3o
ool0H03oool00`00o`00oooo0?ooo`0O0?ooo`03003o003oool0oooo02T0oooo00<0o`000?ooo`3o
ool0<`3oool00`3o0000oooo0?ooo`1A0?ooo`00;@3oool20?l006<0oooo00<00?l00?ooo`3oool0
7`3oool00`00o`00oooo0?ooo`0Y0?ooo`030?l0003oool0oooo03<0oooo00<0o`000?ooo`3oool0
D@3oool002`0oooo00<0o`000?ooo`3oool0HP3oool00`00o`00oooo0?ooo`0O0?ooo`03003o003o
ool0oooo02T0oooo00<0o`000?ooo`3oool0=03oool00`3o0000oooo0?ooo`1A0?ooo`00:P3oool2
0?l006@0oooo00<00?l00?ooo`3oool0803oool00`00o`00oooo0?ooo`0Y0?ooo`030?l0003oool0
oooo03@0oooo00<0o`000?ooo`3oool0D@3oool002P0oooo0P3o001U0?ooo`03003o003oool0oooo
0200oooo00<00?l00?ooo`3oool0:P3oool00`3o0000oooo0?ooo`0d0?ooo`030?l0003oool0oooo
0540oooo000W0?ooo`030?l0003oool0oooo06@0oooo00<00?l00?ooo`3oool0803oool00`00o`00
oooo0?ooo`0[0?ooo`030?l0003oool0oooo03@0oooo00<0o`000?ooo`3oool0D@3oool002D0oooo
0P3o001V0?ooo`03003o003oool0oooo0240oooo00<00?l00?ooo`3oool0:P3oool00`3o0000oooo
0?ooo`0e0?ooo`030?l0003oool0oooo0540oooo000T0?ooo`030?l0003oool0oooo06H0oooo00<0
0?l00?ooo`3oool0803oool00`00o`00oooo0?ooo`0[0?ooo`030?l0003oool0oooo03D0oooo00<0
o`000?ooo`3oool0D@3oool00280oooo0P3o001X0?ooo`03003o003oool0oooo0240oooo00<00?l0
0?ooo`3oool0:`3oool00`3o0000oooo0?ooo`0e0?ooo`030?l0003oool0oooo0540oooo000Q0?oo
o`030?l0003oool0oooo06L0oooo00<00?l00?ooo`3oool0803oool2003o02d0oooo00<0o`000?oo
o`3oool0=`3oool00`3o0000oooo0?ooo`1@0?ooo`007`3oool20?l006T0oooo00<00?l00?ooo`3o
ool08@3oool3003o02`0oooo00<0o`000?ooo`3oool0=`3oool00`3o0000oooo0?ooo`1@0?ooo`00
7P3oool00`3o0000oooo0?ooo`1X0?ooo`03003o003oool0oooo0240oooo1@00o`0[0?ooo`030?l0
003oool0oooo03L0oooo00<0o`000?ooo`3oool0D03oool001`0oooo0P3o001Z0?ooo`03003o003o
ool0oooo0280oooo1000o`0/0?ooo`030?l0003oool0oooo03L0oooo00<0o`000?ooo`3oool0D03o
ool001X0oooo0P3o001/0?ooo`03003o003oool0oooo0240oooo1000o`0/0?ooo`030?l0003oool0
oooo03P0oooo00<0o`000?ooo`3oool0D03oool001T0oooo00<0o`000?ooo`3oool0J`3oool00`00
o`00oooo0?ooo`0R0?ooo`<00?l0;@3oool00`3o0000oooo0?ooo`0h0?ooo`030?l0003oool0oooo
0500oooo000G0?ooo`80o`00K@3oool00`00o`00oooo0?ooo`0R0?ooo`800?l0;`3oool00`3o0000
oooo0?ooo`0h0?ooo`030?l0003oool0oooo0500oooo000F0?ooo`030?l0003oool0oooo06`0oooo
00<00?l00?ooo`3oool08`3oool00`00o`00oooo0?ooo`0^0?ooo`030?l0003oool0oooo03P0oooo
00<0o`000?ooo`3oool0D03oool001@0oooo0P3o001^0?ooo`03003o003oool0oooo05@0oooo00<0
o`000?ooo`3oool0>@3oool00`3o0000oooo0?ooo`1@0?ooo`004`3oool00`3o0000oooo0?ooo`1]
0?ooo`03003o003oool0oooo05D0oooo00<0o`000?ooo`3oool0>P3oool00`3o0000oooo0?ooo`1?
0?ooo`004@3oool20?l00700oooo00<00?l00?ooo`3oool0E@3oool00`3o0000oooo0?ooo`0j0?oo
o`030?l0003oool0oooo04l0oooo000@0?ooo`030?l0003oool0oooo06l0oooo00<00?l00?ooo`3o
ool0E@3oool00`3o0000oooo0?ooo`0k0?ooo`030?l0003oool0oooo04l0oooo000>0?ooo`80o`00
L@3oool00`00o`00oooo0?ooo`1F0?ooo`030?l0003oool0oooo03/0oooo00<0o`000?ooo`3oool0
C`3oool000`0oooo0P3o001b0?ooo`03003o003oool0oooo05L0oooo00<0o`000?ooo`3oool0>`3o
ool00`3o0000oooo0?ooo`1?0?ooo`002`3oool00`3o0000oooo0?ooo`1a0?ooo`03003o003oool0
oooo05P0oooo00<0o`000?ooo`3oool0>`3oool00`3o0000oooo0?ooo`1?0?ooo`002@3oool20?l0
07<0oooo00<00?l00?ooo`3oool0F03oool00`3o0000oooo0?ooo`0l0?ooo`030?l0003oool0oooo
04l0oooo00080?ooo`80o`00M03oool00`00o`00oooo0?ooo`1H0?ooo`030?l0003oool0oooo03`0
oooo00<0o`000?ooo`3oool0C`3oool000T0oooo00@00?l00?l0003o0000o`00L03oool00`00o`00
oooo0?ooo`1I0?ooo`030?l0003oool0oooo03`0oooo00<0o`000?ooo`3oool0C`3oool000/0oooo
0P00o`020?l006d0oooo00<00?l00?ooo`3oool0F@3oool00`3o0000oooo0?ooo`0n0?ooo`030?l0
003oool0oooo04h0oooo000=0?ooo`800?l00`3o001Y0?ooo`03003o003oool0oooo05X0oooo00<0
o`000?ooo`3oool0?P3oool00`3o0000oooo0?ooo`1>0?ooo`003`3oool00`00o`00oooo0?ooo`03
0?l006D0oooo00<00?l00?ooo`3oool0F`3oool00`3o0000oooo0?ooo`0n0?ooo`030?l0003oool0
oooo04h0oooo000@0?ooo`800?l00`3oool30?l00640oooo00<00?l00?ooo`3oool0G03oool00`3o
0000oooo0?ooo`0n0?ooo`030?l0003oool0oooo04h0oooo000B0?ooo`800?l0103oool20?l005h0
oooo00<00?l00?ooo`3oool0G03oool00`3o0000oooo0?ooo`0o0?ooo`030?l0003oool0oooo04h0
oooo000D0?ooo`03003o003oool0oooo00<0oooo0`3o001K0?ooo`03003o003oool0oooo05`0oooo
00<0o`000?ooo`3oool0?`3oool30?l004h0oooo000E0?ooo`800?l01P3oool30?l005L0oooo00<0
0?l00?ooo`3oool0G@3oool00`3o0000oooo0?ooo`0m0?ooo`D0o`00CP3oool001L0oooo0P00o`07
0?ooo`<0o`00D`3oool00`00o`00oooo0?ooo`1N0?ooo`030?l0003oool0oooo03d0oooo1@3o001>
0?ooo`006@3oool2003o00P0oooo0P3o001@0?ooo`03003o003oool0oooo05h0oooo00<0o`000?oo
o`3oool0?`3oool40?l004h0oooo000K0?ooo`03003o003oool0oooo00L0oooo0`3o001<0?ooo`03
003o003oool0oooo05l0oooo00<0o`000?ooo`3oool0?`3oool30?l004l0oooo000L0?ooo`800?l0
2P3oool30?l004P0oooo00<00?l00?ooo`3oool0H03oool00`3o0000oooo0?ooo`100?ooo`80o`00
C`3oool001h0oooo0P00o`0;0?ooo`<0o`00A@3oool00`00o`00oooo0?ooo`1O0?ooo`030?l0003o
ool0oooo0440oooo0P3o001?0?ooo`00803oool00`00o`00oooo0?ooo`0;0?ooo`80o`00@P3oool0
0`00o`00oooo0?ooo`1P0?ooo`030?l0003oool0oooo0480oooo00<0o`000?ooo`3oool0C@3oool0
0240oooo0P00o`0=0?ooo`<0o`00?P3oool00`00o`00oooo0?ooo`1Q0?ooo`030?l0003oool0oooo
0480oooo00<0o`000?ooo`3oool0C@3oool002<0oooo0P00o`0>0?ooo`<0o`00>P3oool00`00o`00
oooo0?ooo`1R0?ooo`030?l0003oool0oooo0980oooo000U0?ooo`800?l03`3oool30?l003H0oooo
00<00?l00?ooo`3oool0HP3oool00`3o0000oooo0?ooo`2C0?ooo`009`3oool00`00o`00oooo0?oo
o`0?0?ooo`80o`00<`3oool00`00o`00oooo0?ooo`1S0?ooo`030?l0003oool0oooo09<0oooo000X
0?ooo`800?l04@3oool30?l00300oooo00<00?l00?ooo`3oool0H`3oool00`3o0000oooo0?ooo`2C
0?ooo`00:P3oool2003o0180oooo0`3o000/0?ooo`03003o003oool0oooo06@0oooo00<0o`000?oo
o`3oool0T`3oool002`0oooo00<00?l00?ooo`3oool04P3oool30?l002P0oooo00<00?l00?ooo`3o
ool0I03oool00`3o0000oooo0?ooo`2D0?ooo`00;@3oool2003o01D0oooo0P3o000U0?ooo`03003o
003oool0oooo06D0oooo00<0o`000?ooo`3oool0U03oool002l0oooo0P00o`0E0?ooo`<0o`008@3o
ool00`00o`00oooo0?ooo`1V0?ooo`030?l0003oool0oooo09@0oooo000a0?ooo`800?l05P3oool3
0?l001d0oooo00<00?l00?ooo`3oool0IP3oool00`3o0000oooo0?ooo`2E0?ooo`00<`3oool00`00
o`00oooo0?ooo`0F0?ooo`<0o`006P3oool00`00o`00oooo0?ooo`1V0?ooo`030?l0003oool0oooo
09D0oooo000d0?ooo`800?l06@3oool20?l001L0oooo00<00?l00?ooo`3oool0I`3oool00`3o0000
oooo0?ooo`2E0?ooo`00=P3oool2003o01T0oooo0`3o000C0?ooo`03003o003oool0oooo06P0oooo
00<0o`000?ooo`3oool0U@3oool003P0oooo00<00?l00?ooo`3oool06@3oool30?l000l0oooo00<0
0?l00?ooo`3oool0J03oool00`3o0000oooo0?ooo`2F0?ooo`00>@3oool2003o01`0oooo0`3o000;
0?ooo`03003o003oool0oooo06T0oooo00<0o`000?ooo`3oool0UP3oool003/0oooo0P00o`0M0?oo
o`80o`00203oool00`00o`00oooo0?ooo`1Z0?ooo`030?l0003oool0oooo09H0oooo000m0?ooo`80
0?l07@3oool30?l000D0oooo00<00?l00?ooo`3oool0J@3oool00`3o0000oooo0?ooo`2G0?ooo`00
?`3oool00`00o`00oooo0?ooo`0M0?ooo`<0o`0000<0oooo003o003oool0J`3oool00`3o0000oooo
0?ooo`2G0?ooo`00@03oool2003o0200oooo0`3o001[0?ooo`030?l0003oool0oooo09L0oooo0012
0?ooo`800?l07@3oool01000o`00oooo0?ooo`3oool20?l006T0oooo00<0o`000?ooo`3oool0U`3o
ool004@0oooo00<00?l00?ooo`3oool06@3oool00`00o`00oooo0?ooo`040?ooo`<0o`00I@3oool0
0`3o0000oooo0?ooo`2H0?ooo`00A@3oool2003o01P0oooo00<00?l00?ooo`3oool0203oool30?l0
0680oooo00<0o`000?ooo`3oool0V03oool004L0oooo0P00o`0F0?ooo`03003o003oool0oooo00/0
oooo0`3o001O0?ooo`030?l0003oool0oooo09P0oooo00190?ooo`800?l04`3oool00`00o`00oooo
0?ooo`0?0?ooo`<0o`00G03oool00`3o0000oooo0?ooo`2H0?ooo`00B`3oool00`00o`00oooo0?oo
o`0?0?ooo`03003o003oool0oooo01<0oooo0P3o001I0?ooo`030?l0003oool0oooo09T0oooo001<
0?ooo`800?l00`3oool00`00o`00oooo0?ooo`080?ooo`03003o003oool0oooo01H0oooo0`3o001F
0?ooo`030?l0003oool0oooo09T0oooo001>0?ooo`D00?l0203oool00`00o`00oooo0?ooo`0J0?oo
o`<0o`00D`3oool00`3o0000oooo0?ooo`2I0?ooo`00D03oool4003o00H0oooo00<00?l00?ooo`3o
ool07P3oool30?l004l0oooo00<0o`000?ooo`3oool0VP3oool004l0oooo1P00o`050?ooo`03003o
003oool0oooo0240oooo0P3o001=0?ooo`030?l0003oool0oooo09X0oooo001?0?ooo`L00?l00`3o
ool00`00o`00oooo0?ooo`0T0?ooo`<0o`00BP3oool00`3o0000oooo0?ooo`2J0?ooo`00D@3oool6
003o00030?ooo`00o`00oooo02T0oooo0`3o00170?ooo`030?l0003oool0oooo09X0oooo001E0?oo
o`<00?l0;P3oool30?l004<0oooo00<0o`000?ooo`3oool0V`3oool008T0oooo0P3o00110?ooo`03
0?l0003oool0oooo09/0oooo002;0?ooo`<0o`00?P3oool00`3o0000oooo0?ooo`2K0?ooo`00SP3o
ool30?l003/0oooo00<0o`000?ooo`3oool0V`3oool00940oooo0`3o000g0?ooo`030?l0003oool0
oooo09`0oooo002D0?ooo`80o`00=@3oool00`3o0000oooo0?ooo`2L0?ooo`00UP3oool30?l00380
oooo00<0o`000?ooo`3oool0W03oool009T0oooo0`3o000^0?ooo`030?l0003oool0oooo09d0oooo
002L0?ooo`<0o`00:`3oool00`3o0000oooo0?ooo`2M0?ooo`00W`3oool20?l002T0oooo00<0o`00
0?ooo`3oool0W@3oool00:40oooo0`3o000V0?ooo`030?l0003oool0oooo09d0oooo002T0?ooo`<0
o`008P3oool00`3o0000oooo0?ooo`2N0?ooo`00Y`3oool30?l001l0oooo00<0o`000?ooo`3oool0
WP3oool00:X0oooo0P3o000M0?ooo`030?l0003oool0oooo09h0oooo002/0?ooo`<0o`006@3oool0
0`3o0000oooo0?ooo`2O0?ooo`00[`3oool30?l001H0oooo00<0o`000?ooo`3oool0W`3oool00;80
oooo0`3o000C0?ooo`030?l0003oool0oooo09l0oooo002e0?ooo`80o`004@3oool00`3o0000oooo
0?ooo`2O0?ooo`00]`3oool30?l000d0oooo00<0o`000?ooo`3oool0X03oool00;X0oooo0`3o0002
0?ooo`030?l0003oool0oooo00D0oooo00<0o`000?ooo`3oool0X03oool00;d0oooo1@3o00050?oo
o`030?l0003oool0oooo0:00oooo002n0?ooo`H0o`000`3oool00`3o0000oooo0?ooo`2P0?ooo`00
_P3oool90?l00:<0oooo00320?ooo`D0o`00X`3oool00?l0ooooJ`3oool00?l0ooooJ`3oool00?l0
ooooJ`3oool00?l0ooooJ`3oool00?l0ooooJ`3oool00?l0ooooJ`3oool00?l0ooooJ`3oool00?l0
ooooJ`3oool00?l0ooooJ`3oool00?l0ooooJ`3oool00001\
\>"],
  ImageRangeCache->{{{0, 361.813}, {430.875, 0}} -> {-0.0948022, -3.86884, \
0.016504, 0.016504}}]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->"EditBar",
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
PrivateNotebookOptions->{"ColorPalette"->{RGBColor, -1}},
ShowCellLabel->False,
ShowCellTags->False,
RenderingOptions->{"ObjectDithering"->True,
"RasterDithering"->False},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 83, 2, 81, "Title"],
Cell[1862, 57, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2342, 76, 31, 0, 83, "Section"],
Cell[2376, 78, 316, 5, 67, "Text"],
Cell[2695, 85, 1453, 38, 119, "Text"],
Cell[4151, 125, 429, 6, 66, "DisplayFormula"],
Cell[4583, 133, 540, 10, 93, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[5160, 148, 33, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[5218, 152, 42, 0, 68, "Subsection"],
Cell[5263, 154, 223, 4, 125, "Input"],
Cell[5489, 160, 117, 2, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[5643, 167, 33, 0, 68, "Subsection"],
Cell[5679, 169, 209, 4, 41, "Text"],
Cell[5891, 175, 164, 3, 125, "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[6104, 184, 33, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[6162, 188, 52, 0, 68, "Subsection"],
Cell[6217, 190, 221, 3, 99, "Input"],
Cell[6441, 195, 421, 7, 99, "Input"],
Cell[6865, 204, 106, 2, 73, "Input"],
Cell[6974, 208, 764, 12, 229, "Input"],
Cell[7741, 222, 771, 12, 229, "Input"],
Cell[8515, 236, 476, 8, 151, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[9028, 249, 71, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[9124, 253, 105, 2, 55, "Subsubsection"],
Cell[9232, 257, 228, 3, 97, "Input"],
Cell[9463, 262, 390, 6, 99, "Input"],
Cell[9856, 270, 243, 4, 105, "Input"],
Cell[10102, 276, 291, 5, 99, "Input"],
Cell[10396, 283, 243, 4, 126, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[10676, 292, 103, 2, 55, "Subsubsection"],
Cell[10782, 296, 160, 3, 73, "Input"],
Cell[10945, 301, 477, 9, 177, "Input"],
Cell[11425, 312, 659, 11, 203, "Input"],
Cell[12087, 325, 443, 8, 180, "Input"],
Cell[12533, 335, 649, 11, 177, "Input"],
Cell[13185, 348, 2899, 43, 609, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[16121, 396, 68, 0, 55, "Subsubsection"],
Cell[16192, 398, 899, 15, 262, "Input"],
Cell[17094, 415, 1597, 28, 359, "Input"],
Cell[18694, 445, 1622, 30, 281, "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[20365, 481, 42, 0, 68, "Subsection"],
Cell[20410, 483, 390, 6, 125, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[20837, 494, 57, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[20919, 498, 74, 0, 55, "Subsubsection"],
Cell[20996, 500, 112, 2, 73, "Input"],
Cell[21111, 504, 177, 3, 96, "Input"],

Cell[CellGroupData[{
Cell[21313, 511, 377, 6, 125, "Input"],
Cell[21693, 519, 225, 4, 76, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[21955, 528, 418, 7, 150, "Input"],
Cell[22376, 537, 283, 4, 76, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[22696, 546, 125, 2, 73, "Input"],
Cell[22824, 550, 268, 4, 76, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[23129, 559, 273, 4, 96, "Input"],
Cell[23405, 565, 324, 5, 119, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[23778, 576, 102, 2, 55, "Subsubsection"],

Cell[CellGroupData[{
Cell[23905, 582, 372, 7, 127, "Input"],
Cell[24280, 591, 182, 3, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[24499, 599, 152, 3, 73, "Input"],
Cell[24654, 604, 139, 2, 59, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[24842, 612, 85, 1, 55, "Subsubsection"],

Cell[CellGroupData[{
Cell[24952, 617, 372, 7, 155, "Input"],
Cell[25327, 626, 89, 1, 75, "Output"]
}, Closed]],
Cell[25431, 630, 611, 15, 67, "Text"],
Cell[26045, 647, 243, 4, 100, "Input"],
Cell[26291, 653, 687, 11, 120, "Text"],

Cell[CellGroupData[{
Cell[27003, 668, 202, 4, 73, "Input"],
Cell[27208, 674, 82, 1, 90, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[27327, 680, 71, 1, 73, "Input"],
Cell[27401, 683, 115, 2, 66, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[27553, 690, 117, 2, 96, "Input"],
Cell[27673, 694, 159, 2, 90, "Output"]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[27905, 704, 27, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[27957, 708, 37, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[28019, 712, 121, 2, 73, "Input"],
Cell[28143, 716, 10394, 206, 279, 1391, 90, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[38574, 927, 72, 1, 73, "Input"],
Cell[38649, 930, 27241, 424, 444, 1617, 103, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[65927, 1359, 80, 1, 73, "Input"],
Cell[66010, 1362, 10494, 217, 179, 1629, 103, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[76541, 1584, 109, 2, 73, "Input"],
Cell[76653, 1588, 11993, 239, 300, 1723, 108, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[88683, 1832, 169, 3, 73, "Input"],
Cell[88855, 1837, 398, 6, 109, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[89290, 1848, 67, 1, 73, "Input"],
Cell[89360, 1851, 8862, 196, 279, 1615, 102, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[98271, 2053, 48, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[98344, 2057, 95, 2, 55, "Subsubsection"],
Cell[98442, 2061, 221, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[98688, 2071, 85, 1, 73, "Input"],
Cell[98776, 2074, 193, 3, 81, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[99006, 2082, 156, 3, 73, "Input"],
Cell[99165, 2087, 1110, 17, 308, "Output"]
}, Closed]],
Cell[100290, 2107, 114, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[100429, 2114, 173, 3, 73, "Input"],
Cell[100605, 2119, 134, 3, 75, "Output"]
}, Closed]],
Cell[100754, 2125, 263, 7, 41, "Text"],
Cell[101020, 2134, 104, 2, 73, "Input"],

Cell[CellGroupData[{
Cell[101149, 2140, 182, 3, 96, "Input"],
Cell[101334, 2145, 388, 6, 163, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[101771, 2157, 59, 0, 55, "Subsubsection"],

Cell[CellGroupData[{
Cell[101855, 2161, 152, 2, 73, "Input"],
Cell[102010, 2165, 143, 3, 75, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[102190, 2173, 220, 3, 99, "Input"],
Cell[102413, 2178, 188, 4, 75, "Output"]
}, Closed]],
Cell[102616, 2185, 61, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[102702, 2189, 65, 1, 73, "Input"],
Cell[102770, 2192, 50, 1, 59, "Output"]
}, Closed]],
Cell[102835, 2196, 466, 14, 67, "Text"],

Cell[CellGroupData[{
Cell[103326, 2214, 77, 1, 73, "Input"],
Cell[103406, 2217, 63, 1, 59, "Output"]
}, Closed]],
Cell[103484, 2221, 629, 16, 93, "Text"],

Cell[CellGroupData[{
Cell[104138, 2241, 77, 1, 73, "Input"],
Cell[104218, 2244, 58, 1, 59, "Output"]
}, Closed]],
Cell[104291, 2248, 570, 12, 93, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[104898, 2265, 85, 1, 55, "Subsubsection"],

Cell[CellGroupData[{
Cell[105008, 2270, 149, 2, 73, "Input"],
Cell[105160, 2274, 331, 5, 138, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[105528, 2284, 155, 2, 73, "Input"],
Cell[105686, 2288, 289, 5, 103, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[106012, 2298, 78, 1, 73, "Input"],
Cell[106093, 2301, 242, 4, 103, "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[106396, 2312, 65, 0, 68, "Subsection"],
Cell[106464, 2314, 201, 3, 73, "Input"],

Cell[CellGroupData[{
Cell[106690, 2321, 186, 4, 73, "Input"],
Cell[106879, 2327, 151, 2, 61, "Output"]
}, Closed]],
Cell[107045, 2332, 71, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[107141, 2337, 43, 1, 73, "Input"],
Cell[107187, 2340, 91, 1, 82, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[107315, 2346, 147, 2, 73, "Input"],
Cell[107465, 2350, 35, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[107537, 2356, 65, 1, 73, "Input"],
Cell[107605, 2359, 9881, 219, 226, 1869, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[117523, 2583, 168, 2, 73, "Input"],
Cell[117694, 2587, 12200, 238, 199, 1638, 103, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[129931, 2830, 112, 2, 73, "Input"],
Cell[130046, 2834, 35, 1, 59, "Output"]
}, Closed]],
Cell[130096, 2838, 85, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[130206, 2843, 254, 4, 148, "Input"],
Cell[130463, 2849, 44136, 677, 444, 2527, 159, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[174614, 3529, 276, 7, 41, "Text"],

Cell[CellGroupData[{
Cell[174915, 3540, 73, 1, 73, "Input"],
Cell[174991, 3543, 90244, 1931, 279, 23893, 1107, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[265272, 5479, 72, 1, 73, "Input"],
Cell[265347, 5482, 76407, 1652, 279, 20590, 958, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[341769, 7137, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[341856, 7142, 76, 1, 73, "Input"],
Cell[341935, 7145, 35903, 542, 444, 1767, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[377887, 7693, 71, 0, 68, "Subsection"],
Cell[377961, 7695, 531, 8, 93, "Text"],

Cell[CellGroupData[{
Cell[378517, 7707, 45, 0, 55, "Subsubsection"],
Cell[378565, 7709, 206, 3, 73, "Input"],
Cell[378774, 7714, 157, 3, 73, "Input"],
Cell[378934, 7719, 71, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[379030, 7724, 65, 1, 73, "Input"],
Cell[379098, 7727, 5639, 167, 127, 1883, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[384752, 7897, 359, 9, 67, "Text"],
Cell[385114, 7908, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[385201, 7913, 76, 1, 73, "Input"],
Cell[385280, 7916, 40684, 601, 444, 1785, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[425979, 8520, 85, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[426089, 8525, 254, 4, 148, "Input"],
Cell[426346, 8531, 43645, 671, 444, 2559, 159, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[470006, 9205, 346, 6, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[470389, 9216, 44, 0, 55, "Subsubsection"],
Cell[470436, 9218, 206, 3, 73, "Input"],
Cell[470645, 9223, 157, 3, 73, "Input"],
Cell[470805, 9228, 71, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[470901, 9233, 65, 1, 73, "Input"],
Cell[470969, 9236, 7268, 187, 174, 1884, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[478252, 9426, 361, 9, 67, "Text"],
Cell[478616, 9437, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[478703, 9442, 76, 1, 73, "Input"],
Cell[478782, 9445, 46467, 672, 444, 1783, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[525264, 10120, 85, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[525374, 10125, 254, 4, 148, "Input"],
Cell[525631, 10131, 51152, 764, 444, 2560, 159, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[576798, 10898, 181, 4, 41, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[577028, 10908, 67, 0, 68, "Subsection"],
Cell[577098, 10910, 181, 4, 41, "Text"],
Cell[577282, 10916, 209, 3, 73, "Input"],

Cell[CellGroupData[{
Cell[577516, 10923, 213, 3, 73, "Input"],
Cell[577732, 10928, 528, 9, 204, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[578297, 10942, 55, 0, 55, "Subsubsection"],
Cell[578355, 10944, 319, 7, 69, "Text"],
Cell[578677, 10953, 106, 2, 73, "Input"],

Cell[CellGroupData[{
Cell[578808, 10959, 65, 1, 73, "Input"],
Cell[578876, 10962, 6916, 183, 154, 1883, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[585807, 11148, 85, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[585917, 11153, 154, 2, 73, "Input"],
Cell[586074, 11157, 25190, 436, 398, 2642, 153, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[611279, 11596, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[611366, 11601, 76, 1, 73, "Input"],
Cell[611445, 11604, 36230, 546, 444, 1791, 116, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[647712, 12155, 108, 2, 73, "Input"],
Cell[647823, 12159, 125, 2, 84, "Output"]
}, Closed]],
Cell[647963, 12164, 170, 6, 42, "Text"],
Cell[648136, 12172, 196, 3, 99, "Input"],

Cell[CellGroupData[{
Cell[648357, 12179, 57, 1, 73, "Input"],
Cell[648417, 12182, 368, 6, 60, "Message"],
Cell[648788, 12190, 53, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[648878, 12196, 119, 2, 73, "Input"],
Cell[649000, 12200, 95083, 2105, 279, 26651, 1256, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[744120, 14310, 118, 2, 73, "Input"],
Cell[744241, 14314, 90610, 2061, 279, 26156, 1261, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[834900, 16381, 55, 0, 55, "Subsubsection"],
Cell[834958, 16383, 345, 7, 68, "Text"],

Cell[CellGroupData[{
Cell[835328, 16394, 111, 2, 73, "Input"],
Cell[835442, 16398, 113768, 2475, 279, 29774, 1433, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[949247, 18878, 110, 2, 73, "Input"],
Cell[949360, 18882, 100458, 2122, 279, 24765, 1183, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[1049879, 21011, 46, 0, 68, "Subsection"],
Cell[1049928, 21013, 675, 10, 119, "Text"],

Cell[CellGroupData[{
Cell[1050628, 21027, 168, 2, 73, "Input"],
Cell[1050799, 21031, 8602, 193, 141, 1657, 103, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[1059416, 21227, 711, 19, 69, "Text"],

Cell[CellGroupData[{
Cell[1060152, 21250, 443, 8, 161, "Input"],
Cell[1060598, 21260, 125, 2, 80, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[1060760, 21267, 123, 3, 73, "Input"],
Cell[1060886, 21272, 108, 2, 83, "Output"]
}, Closed]],
Cell[1061009, 21277, 189, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[1061223, 21289, 159, 3, 74, "Input"],
Cell[1061385, 21294, 121, 2, 89, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[1061543, 21301, 115, 2, 74, "Input"],
Cell[1061661, 21305, 116, 2, 59, "Output"]
}, Closed]],
Cell[1061792, 21310, 453, 8, 67, "Text"],
Cell[1062248, 21320, 239, 5, 133, "Input"],
Cell[1062490, 21327, 487, 8, 185, "Input"],

Cell[CellGroupData[{
Cell[1063002, 21339, 198, 3, 73, "Input"],
Cell[1063203, 21344, 272974, 5000, 560, 80071, 2614, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[1336214, 26349, 502, 9, 156, "Input"],
Cell[1336719, 26360, 179105, 3687, 355, 35459, 1909, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[1515873, 30053, 70, 0, 68, "Subsection"],
Cell[1515946, 30055, 218, 3, 73, "Input"],
Cell[1516167, 30060, 157, 3, 73, "Input"],
Cell[1516327, 30065, 71, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1516423, 30070, 65, 1, 73, "Input"],
Cell[1516491, 30073, 19522, 332, 444, 1783, 108, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1536028, 30408, 85, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1536138, 30413, 154, 2, 73, "Input"],
Cell[1536295, 30417, 47112, 699, 444, 2629, 145, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1583422, 31119, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1583509, 31124, 76, 1, 73, "Input"],
Cell[1583588, 31127, 46394, 671, 444, 1814, 116, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1629997, 31801, 63, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1630085, 31806, 154, 2, 73, "Input"],
Cell[1630242, 31810, 40678, 619, 444, 2635, 145, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

