(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    602042,      17550]*)
(*NotebookOutlinePosition[    638965,      18855]*)
(*  CellTagsIndexPosition[    638921,      18851]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Driven Pendulum", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell["\<\
A driven pendulum with linear damping satisfies an equation of motion of the \
form\
\>", "Text"],

Cell[BoxData[
    FormBox[
      RowBox[{
        RowBox[{
          FractionBox[\(\[PartialD]\^2 \[Theta]\), \(\[PartialD]t\^2\),
            MultilineFunction->None], "+", 
          RowBox[{"\[Gamma]", " ", 
            FractionBox[\(\[PartialD]\[Theta]\), \(\[PartialD]t\),
              MultilineFunction->None]}], "+", 
          " ", \(\(\[Omega]\_0\^2\) Sin[\[Theta]]\)}], "  ", "\[Equal]", 
        " ", \(f\ Cos[\[Omega]\ t]\)}], TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`f\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\)]],
  " are the amplitude and frequency of the driving force and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\)]],
  " is the damping coefficient.  ",
  "The natural frequency for free small-amplitude oscillations is ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\_0 = \@\(g/\[ScriptL]\)\)]],
  ", but it is convenient to express frequency in units of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\_0\)]],
  " by setting  ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\_0 \[Rule] 1\)]],
  " and rescaling the time unit accordingly."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Off[General::spell, \ General::spell1]\)}], "Input"],

Cell["\<\
$DefaultFont={\"Times\", 12}; 
$TextStyle={FontFamily\[Rule]\"Times\", \
FontSize\[Rule]12,FontSlant\[Rule]\"Italic\"}; \
\>", "Input"],

Cell[BoxData[{
    \(\(Needs["\<Utilities`Notation`\>"];\)\), "\n", 
    \(Needs["\<Utilities`FilterOptions`\>"]\)}], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Omega]\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Theta]\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", 
      RowBox[{
        TagBox["",
          NotationBoxTag,
          Editable->True], 
        RowBox[{"Symbolize", "[", 
          TagBox[\(\[Theta]\& . \_0\),
            NotationBoxTag,
            Editable->True], "]"}]}], ";"}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Implementation", "Section"],

Cell["\<\
In this section we construct functions which define and solve the pendulum \
equation given its parameters and initial conditions.  \
\>", "Text"],

Cell[BoxData[{
    \(\(\(pendulumEq::usage\  = \ \*"\"\<pendulumEq[\[Gamma],f,\[Omega]] \
describes the motion of a driven pendulum with linear damping, \\n \!\(\
\[Theta]\^\[DoublePrime]\) + \[Gamma] \!\(\[Theta]\^\[Prime]\)+Sin[\[Theta]]\
\[Equal]f Cos[\[Omega] t], \\n where f and \[Omega] are the driving force and \
frequency and \[Gamma] is the damping coefficient.\>\"";\)\(\n\)
    \)\), "\n", 
    \(pendulumEq[\[Gamma]_, 
        f_, \[Omega]_] := \ \(\[Theta]\^\[DoublePrime]\)[
            t]\  + \ \[Gamma]\ \(\[Theta]\^\[Prime]\)[t]\  + \ 
          Sin[\[Theta][t]] \[Equal] \ f\ Cos[\[Omega]\ t]\)}], "Input"],

Cell[BoxData[{
    RowBox[{
      RowBox[{
        RowBox[{\(pendulumSol::usage\), " ", "=", " ", 
          "\"\<pendulumSol[\[Gamma],f,\[Omega],\!\(\[Theta]\_0\),\!\(
\*TagBox[\(\[Theta]\& . \_0\),
NotationBoxTag,
Editable->True]\),tmin,tmax,opts] uses NDSolve to produce a numerical \
solution to the pendulum equation given initial conditions \
\[Theta][0]\[Equal]\!\(\[Theta]\_0\) and \
\!\(\[Theta]\^\[Prime]\)[0]\[Equal]\!\(
\*TagBox[
TagBox[\(\[Theta]\& . \_0\),
NotationBoxTag,
Editable->True],
NotationBoxTag,
Editable->True]\) for {t,tmin,tmax}.  See pendulumEq for other parameter \
definitions.  The optional parameters opts are passed to NDSolve.\>\""}], 
        ";"}], "\n", 
      "\t\t"}], "\n", \(pendulumSol[\[Gamma]_, 
        f_, \[Omega]_, \(\[Theta]\_0\) : _, \(\[Theta]\& . \_0\) : _, tmin_, 
        tmax_, opts___]\  := \(NDSolve[{pendulumEq[\[Gamma], 
              f, \[Omega]], \[Theta][
                0] \[Equal] \[Theta]\_0, \(\[Theta]\^\[Prime]\)[
                0] \[Equal] \[Theta]\& . \_0}, \[Theta], {t, tmin, tmax}, 
          opts]\)\[LeftDoubleBracket]1\[RightDoubleBracket]\)}], "Input"],

Cell[BoxData[{
    RowBox[{
      RowBox[{
        RowBox[{\(pendulumPlot::usage\), " ", "=", " ", 
          "\"\<pendulumPlot[\[Gamma],f,\[Omega],\!\(\[Theta]\_0\),\!\(
\*TagBox[\(\[Theta]\& . \_0\),
NotationBoxTag,
Editable->True]\),tmin,tmax,opts] produces a phase diagram for the pendulum \
equation.  The optional parameters opts can regulate either NDSolve or \
ParametricPlot as appropriate.  See pendulumEq and pendulumSol for parameter \
definitions.\>\""}], ";"}], 
      "\n"}], "\n", \(pendulumPlot[\[Gamma]_, 
        f_, \[Omega]_, \(\[Theta]\_0\) : _, \(\[Theta]\& . \_0\) : _, tmin_, 
        tmax_, opts___]\  := \ 
      With[{optPlot\  = \ FilterOptions[ParametricPlot, opts]\ , 
          optND\  = 
            FilterOptions[NDSolve, opts]}, \[IndentingNewLine]ParametricPlot[
          Evaluate[{\[Theta][t], \(\[Theta]\^\[Prime]\)[t]} /. 
              pendulumSol[\[Gamma], 
                f, \[Omega], \[Theta]\_0, \[Theta]\& . \_0, tmin, tmax, 
                optND]], {t, tmin, tmax}, \ optPlot, 
          AspectRatio \[Rule] Automatic, Frame \[Rule] True, 
          FrameLabel \[Rule] {"\<position\>", \ "\<velocity\>"}, 
          PlotLabel \[Rule] "\<Pendulum Equation\>"]]\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples", "Section"],

Cell[CellGroupData[{

Cell["Free swing without damping", "Subsection"],

Cell["\<\
For small amplitudes we find the classic circular phase diagram \
characteristic of simple harmonic motion.  For larger amplitudes the \
nonlinearity of the restoring force distorts the trajectory and increases the \
period.\
\>", "Text"],

Cell[BoxData[
    \(\(plots\  = \ 
        Table[pendulumPlot[0, 0, 0, \[Theta]\_0, 0, 0, 15, 
            DisplayFunction \[Rule] 
              Identity], {\[Theta]\_0, \[Pi]\/8, \(7  \[Pi]\)\/8, \[Pi]\/8}];\
\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[plots, DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .71358 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.499999 0.17323 0.356792 0.17323 [
[.15354 -0.0125 -9.34375 -15.6875 ]
[.15354 -0.0125 9.34375 0 ]
[.32677 -0.0125 -9.34375 -15.6875 ]
[.32677 -0.0125 9.34375 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.67323 -0.0125 -5 -15.6875 ]
[.67323 -0.0125 5 0 ]
[.84646 -0.0125 -5 -15.6875 ]
[.84646 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.9688 -31.375 ]
[.5 -0.0125 21.9688 -15.6875 ]
[-0.0125 .01033 -18.6875 -7.84375 ]
[-0.0125 .01033 0 7.84375 ]
[-0.0125 .18356 -18.6875 -7.84375 ]
[-0.0125 .18356 0 7.84375 ]
[-0.0125 .35679 -10 -7.84375 ]
[-0.0125 .35679 0 7.84375 ]
[-0.0125 .53002 -10 -7.84375 ]
[-0.0125 .53002 0 7.84375 ]
[-0.0125 .70325 -10 -7.84375 ]
[-0.0125 .70325 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .35679 -34.375 -20.4688 ]
[-0.01875 .35679 -18.6875 20.4688 ]
[ 0 .71358 .125 0 ]
[ 1 0 .125 0 ]
[.5 .72608 -49.7813 0 ]
[.5 .72608 49.7813 15.6875 ]
[ 0 0 0 0 ]
[ 1 .71358 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15354 0 m
.15354 .00625 L
s
gsave
.15354 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.32677 0 m
.32677 .00625 L
s
gsave
.32677 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.67323 0 m
.67323 .00625 L
s
gsave
.67323 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.84646 0 m
.84646 .00625 L
s
gsave
.84646 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.18819 0 m
.18819 .00375 L
s
.22283 0 m
.22283 .00375 L
s
.25748 0 m
.25748 .00375 L
s
.29212 0 m
.29212 .00375 L
s
.36142 0 m
.36142 .00375 L
s
.39606 0 m
.39606 .00375 L
s
.43071 0 m
.43071 .00375 L
s
.46535 0 m
.46535 .00375 L
s
.53465 0 m
.53465 .00375 L
s
.56929 0 m
.56929 .00375 L
s
.60394 0 m
.60394 .00375 L
s
.63858 0 m
.63858 .00375 L
s
.70788 0 m
.70788 .00375 L
s
.74252 0 m
.74252 .00375 L
s
.77717 0 m
.77717 .00375 L
s
.81181 0 m
.81181 .00375 L
s
.11889 0 m
.11889 .00375 L
s
.08425 0 m
.08425 .00375 L
s
.0496 0 m
.0496 .00375 L
s
.01496 0 m
.01496 .00375 L
s
.88111 0 m
.88111 .00375 L
s
.91575 0 m
.91575 .00375 L
s
.9504 0 m
.9504 .00375 L
s
.98504 0 m
.98504 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.9688 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(position) show
102.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01033 m
.00625 .01033 L
s
gsave
-0.0125 .01033 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .18356 m
.00625 .18356 L
s
gsave
-0.0125 .18356 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .35679 m
.00625 .35679 L
s
gsave
-0.0125 .35679 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .53002 m
.00625 .53002 L
s
gsave
-0.0125 .53002 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .70325 m
.00625 .70325 L
s
gsave
-0.0125 .70325 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04498 m
.00375 .04498 L
s
0 .07962 m
.00375 .07962 L
s
0 .11427 m
.00375 .11427 L
s
0 .14892 m
.00375 .14892 L
s
0 .21821 m
.00375 .21821 L
s
0 .25285 m
.00375 .25285 L
s
0 .2875 m
.00375 .2875 L
s
0 .32215 m
.00375 .32215 L
s
0 .39144 m
.00375 .39144 L
s
0 .42608 m
.00375 .42608 L
s
0 .46073 m
.00375 .46073 L
s
0 .49538 m
.00375 .49538 L
s
0 .56467 m
.00375 .56467 L
s
0 .59931 m
.00375 .59931 L
s
0 .63396 m
.00375 .63396 L
s
0 .66861 m
.00375 .66861 L
s
.25 Mabswid
0 0 m
0 .71358 L
s
gsave
-0.01875 .35679 -108 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.4688 11.8438 translate
90 rotate
-81.4688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(velocity) show
99.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.15354 .70733 m
.15354 .71358 L
s
.32677 .70733 m
.32677 .71358 L
s
.5 .70733 m
.5 .71358 L
s
.67323 .70733 m
.67323 .71358 L
s
.84646 .70733 m
.84646 .71358 L
s
.125 Mabswid
.18819 .70983 m
.18819 .71358 L
s
.22283 .70983 m
.22283 .71358 L
s
.25748 .70983 m
.25748 .71358 L
s
.29212 .70983 m
.29212 .71358 L
s
.36142 .70983 m
.36142 .71358 L
s
.39606 .70983 m
.39606 .71358 L
s
.43071 .70983 m
.43071 .71358 L
s
.46535 .70983 m
.46535 .71358 L
s
.53465 .70983 m
.53465 .71358 L
s
.56929 .70983 m
.56929 .71358 L
s
.60394 .70983 m
.60394 .71358 L
s
.63858 .70983 m
.63858 .71358 L
s
.70788 .70983 m
.70788 .71358 L
s
.74252 .70983 m
.74252 .71358 L
s
.77717 .70983 m
.77717 .71358 L
s
.81181 .70983 m
.81181 .71358 L
s
.11889 .70983 m
.11889 .71358 L
s
.08425 .70983 m
.08425 .71358 L
s
.0496 .70983 m
.0496 .71358 L
s
.01496 .70983 m
.01496 .71358 L
s
.88111 .70983 m
.88111 .71358 L
s
.91575 .70983 m
.91575 .71358 L
s
.9504 .70983 m
.9504 .71358 L
s
.98504 .70983 m
.98504 .71358 L
s
.25 Mabswid
0 .71358 m
1 .71358 L
s
.99375 .01033 m
1 .01033 L
s
.99375 .18356 m
1 .18356 L
s
.99375 .35679 m
1 .35679 L
s
.99375 .53002 m
1 .53002 L
s
.99375 .70325 m
1 .70325 L
s
.125 Mabswid
.99625 .04498 m
1 .04498 L
s
.99625 .07962 m
1 .07962 L
s
.99625 .11427 m
1 .11427 L
s
.99625 .14892 m
1 .14892 L
s
.99625 .21821 m
1 .21821 L
s
.99625 .25285 m
1 .25285 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .32215 m
1 .32215 L
s
.99625 .39144 m
1 .39144 L
s
.99625 .42608 m
1 .42608 L
s
.99625 .46073 m
1 .46073 L
s
.99625 .49538 m
1 .49538 L
s
.99625 .56467 m
1 .56467 L
s
.99625 .59931 m
1 .59931 L
s
.99625 .63396 m
1 .63396 L
s
.99625 .66861 m
1 .66861 L
s
.25 Mabswid
1 0 m
1 .71358 L
s
0 .35679 m
1 .35679 L
s
.5 0 m
.5 .71358 L
s
gsave
.5 .72608 -110.781 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Pendulum) show
114.625 15.813 moveto
(Equation) show
158.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .71358 L
0 .71358 L
closepath
clip
newpath
.5 Mabswid
.56803 .35679 m
.56802 .35556 L
.56799 .35445 L
.56793 .35316 L
.56785 .35194 L
.56766 .34979 L
.56737 .34745 L
.56695 .34489 L
.5664 .34219 L
.56514 .33742 L
.56347 .3326 L
.56157 .3282 L
.5561 .31873 L
.54901 .31007 L
.54134 .30324 L
.53668 .29998 L
.53131 .29687 L
.52602 .2944 L
.52087 .2925 L
.5158 .29107 L
.51289 .29044 L
.5102 .28997 L
.50791 .28967 L
.50539 .28942 L
.50411 .28933 L
.50275 .28926 L
.50145 .28922 L
.50028 .2892 L
.49889 .28921 L
.49819 .28923 L
.49742 .28925 L
.49606 .28932 L
.49479 .2894 L
.4933 .28953 L
.49193 .28968 L
.48886 .29012 L
.48638 .29059 L
.48368 .2912 L
.47834 .29276 L
.47309 .29478 L
.4683 .29707 L
.45989 .30232 L
.4518 .30925 L
.44438 .31805 L
.43923 .32657 L
.43688 .33171 L
.43509 .33667 L
.43429 .33941 L
.43364 .34203 L
.43313 .34443 L
Mistroke
.4327 .347 L
.4325 .34847 L
.43234 .34981 L
.43221 .35116 L
.43212 .35242 L
.43207 .35312 L
.43204 .35388 L
.43201 .35466 L
.43199 .35547 L
.43197 .35619 L
.43197 .35696 L
.43198 .35769 L
.43199 .35835 L
.43203 .35958 L
.43209 .36074 L
.43218 .36201 L
.4323 .36337 L
.43263 .36614 L
.43304 .36867 L
.43404 .37324 L
.43544 .37797 L
.43736 .38301 L
.43951 .38755 L
.4456 .39721 L
.45302 .40553 L
.4612 .41219 L
.46934 .41705 L
.47421 .41928 L
.47884 .42099 L
.48363 .42237 L
.48631 .42298 L
.48887 .42346 L
.49127 .42382 L
.49254 .42397 L
.49389 .42411 L
.49502 .4242 L
.49627 .42428 L
.49742 .42433 L
.49849 .42437 L
.49977 .42438 L
.50118 .42437 L
.50188 .42436 L
.50264 .42433 L
.504 .42426 L
.50536 .42417 L
.50665 .42406 L
.50904 .42378 L
.51185 .42334 L
.51443 .42283 L
.52017 .42131 L
Mistroke
.52485 .41966 L
.52973 .41751 L
.53807 .4127 L
.54697 .40554 L
.55395 .3978 L
.55965 .38913 L
.56369 .38042 L
.56543 .37518 L
.56663 .37034 L
.56714 .36762 L
.56736 .36618 L
.56756 .36466 L
.5677 .36332 L
.56783 .36185 L
.56792 .36051 L
.56798 .35926 L
.56802 .35794 L
.56802 .35728 L
.56803 .35654 L
.56801 .35525 L
.56797 .35403 L
.56794 .35334 L
.56789 .3526 L
.5678 .3513 L
.56767 .34989 L
.56749 .34836 L
.56711 .3458 L
.56659 .34307 L
.56538 .33823 L
.56372 .33326 L
.5618 .32869 L
.55731 .32055 L
.55112 .31236 L
.54328 .30479 L
.5357 .29936 L
.52657 .29463 L
.52138 .29267 L
.51632 .2912 L
.51169 .29022 L
.50905 .28981 L
.50658 .28952 L
.50519 .2894 L
.50445 .28935 L
.50367 .2893 L
.50299 .28927 L
.50225 .28924 L
.50093 .28921 L
.49965 .2892 L
.49828 .28922 L
.49699 .28927 L
Mistroke
.49582 .28933 L
.49435 .28944 L
.493 .28956 L
.48996 .28995 L
.48729 .29041 L
.48449 .291 L
.47945 .2924 L
.47494 .29401 L
.4653 .29876 L
.45713 .30445 L
.45022 .31089 L
.44387 .31877 L
.44092 .32345 L
.43849 .32807 L
.43656 .33253 L
.43484 .33749 L
.43362 .34209 L
.43279 .3464 L
.43244 .34894 L
.4322 .35126 L
.43211 .35255 L
.43207 .35323 L
.43203 .35396 L
.43199 .35529 L
.43197 .35652 L
.43198 .35763 L
.432 .35884 L
.43205 .35998 L
.43211 .36102 L
.43219 .36218 L
.43231 .36344 L
.43261 .36597 L
.43306 .36877 L
.43358 .37129 L
.43508 .37688 L
.43691 .38193 L
.439 .38656 L
.44424 .39534 L
.45117 .4037 L
.4581 .40991 L
.46664 .41561 L
.47157 .41813 L
.47643 .42015 L
.48092 .42164 L
.48591 .4229 L
.48879 .42345 L
.49142 .42384 L
.49288 .42401 L
.49447 .42416 L
.4952 .42421 L
Mistroke
.49597 .42426 L
.49735 .42433 L
.49862 .42437 L
.49999 .42438 L
.50128 .42437 L
.50245 .42434 L
.50383 .42427 L
.5046 .42423 L
.50532 .42417 L
.50803 .42391 L
.51058 .42355 L
.51291 .42314 L
.51809 .42192 L
.52257 .42051 L
.52716 .4187 L
.5364 .41379 L
.54126 .4104 L
.5455 .4069 L
.55316 .3988 L
.55908 .39014 L
.56318 .38171 L
.56496 .37675 L
.56624 .37207 L
.56678 .36961 L
.56719 .36726 L
.56757 .36454 L
.5677 .36334 L
.56782 .36205 L
.5679 .36093 L
.56796 .3597 L
.568 .35857 L
.56802 .35752 L
.56802 .35628 L
.56801 .35514 L
.56796 .35389 L
.56789 .35258 L
.56779 .35123 L
.56768 .35001 L
.56734 .34728 L
.56693 .34476 L
.56644 .34239 L
.56506 .33715 L
.56307 .33161 L
.55866 .32273 L
.55558 .31798 L
.55233 .31378 L
.54484 .30611 L
.54016 .30236 L
.53553 .29926 L
.52596 .29438 L
Mistroke
.52063 .29242 L
.51784 .2916 L
.51478 .29084 L
.51183 .29024 L
.5091 .28982 L
.50647 .28951 L
.50514 .2894 L
.50439 .28934 L
.50368 .2893 L
.50235 .28924 L
.50114 .28921 L
.49984 .2892 L
.4991 .28921 L
.49842 .28922 L
.49729 .28926 L
.49604 .28932 L
.49474 .28941 L
.49352 .28951 L
.49095 .28981 L
.48857 .29017 L
.48322 .29132 L
.47811 .29284 L
.47281 .2949 L
.46371 .29973 L
.45616 .30524 L
.45514 .30612 L
Mfstroke
.63605 .35679 m
.63603 .35452 L
.63598 .35246 L
.63587 .35008 L
.63573 .34783 L
.63537 .34384 L
.63483 .33952 L
.63406 .33477 L
.63305 .32977 L
.63071 .32088 L
.62761 .31187 L
.62408 .30359 L
.61388 .28558 L
.60056 .26878 L
.58605 .25518 L
.57721 .24855 L
.56694 .24211 L
.55681 .23688 L
.54689 .23272 L
.53707 .22948 L
.53182 .22807 L
.5262 .22682 L
.52143 .22595 L
.51616 .2252 L
.51384 .22493 L
.51135 .22469 L
.50916 .22452 L
.50684 .22438 L
.50545 .22432 L
.50391 .22427 L
.50121 .22421 L
.49969 .22421 L
.49805 .22422 L
.4965 .22425 L
.49508 .2243 L
.49249 .22442 L
.49102 .22451 L
.48966 .22461 L
.48397 .22518 L
.47873 .22592 L
.474 .22678 L
.46352 .22931 L
.4539 .23243 L
.44401 .2365 L
.4272 .24563 L
.41079 .25783 L
.39513 .27365 L
.38313 .29023 L
.37458 .30656 L
.3709 .31594 L
Mistroke
.36926 .32098 L
.36776 .32639 L
.36655 .33163 L
.36565 .33642 L
.36493 .34124 L
.36444 .3458 L
.36426 .34805 L
.36411 .35046 L
.36401 .35273 L
.36396 .3548 L
.36395 .35723 L
.36396 .35861 L
.36398 .35988 L
.36406 .3622 L
.3642 .36467 L
.36437 .36703 L
.36461 .36955 L
.36517 .37411 L
.36607 .37952 L
.36709 .38443 L
.36924 .39254 L
.37222 .40126 L
.38013 .41822 L
.39151 .43546 L
.40416 .44966 L
.42119 .46391 L
.42958 .46943 L
.43926 .47481 L
.44816 .4789 L
.45801 .48259 L
.46357 .48429 L
.4687 .48564 L
.47397 .4868 L
.47889 .48769 L
.48344 .48834 L
.48829 .48886 L
.49081 .48906 L
.49318 .4892 L
.49578 .48931 L
.49725 .48935 L
.49859 .48937 L
.50085 .48937 L
.50295 .48934 L
.50526 .48927 L
.50773 .48915 L
.51031 .48898 L
.51275 .48876 L
.51731 .48824 L
.52223 .4875 L
.52682 .48664 L
Mistroke
.53712 .48409 L
.54694 .48084 L
.55717 .47654 L
.57463 .46677 L
.58915 .45581 L
.60344 .4416 L
.61462 .42689 L
.62433 .40945 L
.62854 .39924 L
.63156 .38977 L
.63374 .38054 L
.63457 .37585 L
.63526 .37068 L
.6355 .36841 L
.63569 .36625 L
.63586 .36376 L
.63596 .36149 L
.63603 .35936 L
.63605 .35738 L
.63604 .35521 L
.63599 .3529 L
.63589 .35043 L
.63575 .34816 L
.63555 .34568 L
.63528 .34307 L
.63471 .33868 L
.6339 .33389 L
.63138 .32319 L
.62819 .3134 L
.62437 .30421 L
.61941 .29454 L
.60835 .27794 L
.59519 .26329 L
.58121 .25142 L
.56425 .24061 L
.54706 .23278 L
.53675 .22938 L
.52712 .22701 L
.52173 .226 L
.51887 .22556 L
.51582 .22515 L
.51288 .22483 L
.51021 .2246 L
.50752 .22442 L
.50499 .2243 L
.50255 .22423 L
.49994 .22421 L
.49747 .22423 L
.49523 .22429 L
.49259 .22441 L
Mistroke
.49111 .22451 L
.48974 .22461 L
.48456 .22511 L
.47962 .22578 L
.47422 .22673 L
.46343 .22933 L
.4543 .23228 L
.44474 .23616 L
.42834 .24491 L
.41149 .25722 L
.39555 .27315 L
.3832 .29012 L
.37433 .30714 L
.37073 .31643 L
.36914 .3214 L
.36768 .3267 L
.3666 .33137 L
.36563 .33654 L
.36493 .34126 L
.36445 .34569 L
.36427 .34791 L
.36412 .35033 L
.36402 .35262 L
.36396 .35473 L
.36395 .35664 L
.36396 .35874 L
.36401 .36093 L
.3641 .36299 L
.36424 .36526 L
.36443 .36765 L
.36488 .37191 L
.36553 .37647 L
.36644 .38145 L
.36891 .39142 L
.372 .40068 L
.37969 .41743 L
.39064 .43433 L
.4054 .45086 L
.42036 .46332 L
.43879 .47457 L
.4487 .47913 L
.45973 .48314 L
.46519 .48474 L
.47043 .48605 L
.4752 .48704 L
.4804 .48792 L
.4832 .48831 L
.48577 .48861 L
.4882 .48885 L
.49086 .48906 L
Mistroke
.49325 .48921 L
.49549 .4893 L
.49785 .48936 L
.50038 .48938 L
.50314 .48934 L
.50469 .48929 L
.50613 .48924 L
.50876 .48909 L
.51155 .48887 L
.51636 .48836 L
.52165 .4876 L
.531 .48571 L
.54089 .48294 L
.55001 .47965 L
.55813 .47609 L
.57584 .46596 L
.59073 .45443 L
.605 .43978 L
.61189 .43087 L
.61753 .42227 L
.62666 .4041 L
.63026 .39414 L
.63177 .38898 L
.63314 .3834 L
.63413 .37846 L
.63485 .37393 L
.63514 .37174 L
.63541 .36936 L
.63562 .36712 L
.63578 .36501 L
.63592 .36258 L
.63597 .36129 L
.63601 .35989 L
.63604 .35851 L
.63605 .35705 L
.63605 .35581 L
.63603 .35444 L
.63596 .352 L
.6359 .35062 L
.63583 .34938 L
.63566 .34695 L
.63545 .34467 L
.63491 .34009 L
.63425 .33582 L
.63218 .32617 L
.62943 .31691 L
.62567 .30714 L
.61684 .29019 L
.60658 .27572 L
.59285 .26107 L
Mistroke
.57819 .24924 L
.56032 .23857 L
.54966 .23379 L
.53966 .23025 L
.5303 .22771 L
.52512 .2266 L
.52024 .22576 L
.51606 .22518 L
.51155 .22471 L
.50898 .22451 L
.50654 .22437 L
.50435 .22428 L
.50196 .22422 L
.4994 .22421 L
.497 .22424 L
.49437 .22433 L
.49289 .2244 L
.49153 .22448 L
.48896 .22467 L
.48615 .22493 L
.48051 .22565 L
.47559 .22647 L
.47037 .22755 L
.46111 .23002 L
Mfstroke
.70408 .35679 m
.70405 .35383 L
.70398 .35113 L
.70384 .34802 L
.70365 .34508 L
.70318 .33986 L
.70248 .3342 L
.70148 .32797 L
.70015 .32139 L
.69708 .30966 L
.693 .29766 L
.68835 .28655 L
.67483 .26196 L
.65699 .23829 L
.63735 .2184 L
.61112 .1983 L
.59681 .1897 L
.58321 .18285 L
.57039 .17745 L
.55613 .1726 L
.54813 .17038 L
.54069 .16864 L
.53306 .16716 L
.52589 .16606 L
.5197 .16532 L
.51651 .16502 L
.51304 .16475 L
.50923 .16453 L
.50563 .16439 L
.50211 .16432 L
.49886 .16431 L
.49546 .16436 L
.49226 .16446 L
.48876 .16464 L
.48499 .16489 L
.48126 .16522 L
.47733 .16565 L
.47034 .1666 L
.46267 .16795 L
.45547 .1695 L
.44219 .17311 L
.42822 .17799 L
.41354 .18437 L
.38861 .19848 L
.3677 .21403 L
.34735 .23343 L
.33113 .25325 L
.3167 .27635 L
.31021 .28977 L
.30532 .30225 L
Mistroke
.30166 .31407 L
.30003 .3206 L
.29874 .32677 L
.29776 .33255 L
.29693 .33882 L
.29664 .3416 L
.29639 .34452 L
.29622 .34703 L
.29607 .34977 L
.29601 .35145 L
.29596 .35328 L
.29592 .35661 L
.29595 .35984 L
.29603 .36283 L
.29619 .36606 L
.2963 .36789 L
.29643 .36961 L
.29717 .3768 L
.29809 .38311 L
.29915 .38888 L
.3023 .40181 L
.30611 .41354 L
.31114 .42591 L
.32265 .44763 L
.34031 .47214 L
.35983 .49263 L
.38273 .51111 L
.40621 .52551 L
.43412 .53779 L
.44884 .5424 L
.45672 .54437 L
.46538 .54615 L
.47351 .54745 L
.4771 .54791 L
.48108 .54834 L
.48457 .54866 L
.48834 .54892 L
.49233 .54912 L
.49608 .54924 L
.4995 .54927 L
.50324 .54925 L
.50714 .54914 L
.51078 .54897 L
.51418 .54875 L
.51728 .5485 L
.52421 .54775 L
.53225 .54656 L
.53962 .54517 L
.55604 .54101 L
.57017 .53622 L
Mistroke
.58469 .53005 L
.60942 .51638 L
.6335 .49854 L
.65342 .47931 L
.66899 .46017 L
.68327 .43728 L
.69328 .41518 L
.69759 .4022 L
.70055 .39032 L
.70181 .3837 L
.70236 .38021 L
.70286 .37652 L
.70323 .37328 L
.70356 .36972 L
.70379 .36648 L
.70394 .36346 L
.70404 .36028 L
.70407 .35867 L
.70408 .3569 L
.70405 .35377 L
.70397 .35083 L
.7039 .34915 L
.7038 .34738 L
.70359 .34422 L
.70328 .34082 L
.70287 .33711 L
.70198 .33088 L
.70076 .32424 L
.69786 .31235 L
.69387 .3 L
.68922 .28848 L
.67824 .26743 L
.66283 .2453 L
.64296 .22358 L
.62334 .20689 L
.59928 .19108 L
.58535 .18384 L
.57162 .17793 L
.55878 .17341 L
.54479 .16956 L
.53672 .16783 L
.52936 .16656 L
.52191 .16556 L
.51855 .1652 L
.51487 .16488 L
.51085 .16462 L
.50866 .1645 L
.50661 .16442 L
.50293 .16433 L
.499 .16431 L
Mistroke
.49679 .16434 L
.49439 .16439 L
.49212 .16447 L
.49003 .16457 L
.48607 .16481 L
.48188 .16516 L
.4743 .16603 L
.46713 .16713 L
.4538 .1699 L
.4397 .1739 L
.4247 .1794 L
.3991 .19201 L
.37721 .20648 L
.35385 .22669 L
.3351 .24794 L
.3212 .26836 L
.3147 .2802 L
.30948 .29147 L
.30485 .30363 L
.30145 .31484 L
.29862 .3274 L
.29749 .33435 L
.29708 .33749 L
.2967 .34095 L
.29643 .34403 L
.2962 .34735 L
.2961 .34918 L
.29603 .35085 L
.29594 .35415 L
.29592 .35699 L
.29595 .3601 L
.29605 .36335 L
.2962 .36637 L
.29646 .36997 L
.29678 .37336 L
.29756 .37971 L
.29869 .38653 L
.30024 .3939 L
.30357 .40604 L
.30752 .41728 L
.3195 .4423 L
.33503 .46556 L
.35572 .48874 L
.37833 .50794 L
.39144 .51693 L
.4062 .5255 L
.42105 .53264 L
.435 .5381 L
.44941 .54256 L
.45654 .54433 L
Mistroke
.46325 .54575 L
.47054 .54701 L
.47722 .54792 L
.48079 .54832 L
.48468 .54866 L
.48681 .54882 L
.48879 .54895 L
.49268 .54914 L
.49641 .54924 L
.4999 .54928 L
.50374 .54924 L
.50589 .54919 L
.50787 .54911 L
.51161 .54892 L
.5157 .54863 L
.51997 .54824 L
.52392 .54779 L
.53116 .54674 L
.53794 .54551 L
.55049 .54258 L
.56376 .53853 L
.57775 .53316 L
.60317 .52024 L
.6286 .50259 L
.65178 .48108 L
.66994 .45884 L
.67751 .44736 L
.68472 .43452 L
.69015 .42298 L
.69496 .41051 L
.69843 .39919 L
.69993 .39314 L
.7013 .38658 L
.70231 .38057 L
.70304 .37504 L
.70356 .36975 L
.70375 .36709 L
.70391 .36417 L
.70403 .36085 L
.70408 .3578 L
.70408 .3562 L
.70406 .35444 L
.70403 .3526 L
.70397 .35088 L
.70382 .34772 L
.70363 .34484 L
.70335 .34151 L
.70302 .33834 L
.70224 .33253 L
.70133 .32715 L
Mistroke
.69859 .31501 L
.69454 .30186 L
.68992 .29008 L
.67637 .26438 L
.65997 .24178 L
.64184 .22252 L
.61888 .20362 L
.59162 .18694 L
.58737 .18481 L
Mfstroke
.77211 .35679 m
.77208 .35358 L
.772 .35067 L
.77185 .3473 L
.77164 .34411 L
.77114 .33845 L
.77038 .3323 L
.76929 .32553 L
.76784 .31836 L
.76451 .30549 L
.76006 .29222 L
.75498 .27981 L
.74007 .25174 L
.72018 .22366 L
.69794 .19894 L
.66773 .17236 L
.6349 .1501 L
.6024 .13356 L
.58323 .12609 L
.56535 .12058 L
.54828 .11658 L
.53855 .11485 L
.52954 .11359 L
.52161 .11276 L
.51709 .1124 L
.51287 .11215 L
.50841 .11195 L
.50435 .11185 L
.49965 .11181 L
.4952 .11186 L
.49081 .11198 L
.48607 .1122 L
.48161 .1125 L
.47753 .11284 L
.46806 .11389 L
.45816 .11539 L
.43967 .11927 L
.41986 .12504 L
.38602 .13888 L
.35666 .15526 L
.3251 .17807 L
.29899 .20204 L
.27899 .22471 L
.2607 .2505 L
.24643 .27673 L
.24033 .29118 L
.23572 .30473 L
.23237 .3174 L
.23012 .32899 L
.22921 .33539 L
Mistroke
.22883 .33877 L
.22849 .34236 L
.22824 .34582 L
.22807 .34895 L
.22795 .3521 L
.2279 .35506 L
.2279 .35825 L
.22796 .36166 L
.22808 .36487 L
.22824 .3678 L
.22849 .37124 L
.22885 .37498 L
.22969 .38177 L
.23086 .38888 L
.23242 .39642 L
.23605 .40994 L
.24506 .43387 L
.25883 .46006 L
.27744 .48692 L
.29818 .51071 L
.32666 .53678 L
.35691 .55849 L
.38731 .57533 L
.40376 .5826 L
.42213 .58929 L
.43287 .59252 L
.44287 .59508 L
.45234 .59713 L
.46282 .59895 L
.47229 .60021 L
.47685 .60068 L
.48118 .60105 L
.4861 .60138 L
.49059 .6016 L
.49543 .60173 L
.49817 .60177 L
.5007 .60178 L
.50523 .60172 L
.51 .60157 L
.51407 .60137 L
.51852 .60108 L
.52332 .60067 L
.52844 .60012 L
.53764 .59888 L
.55555 .59545 L
.57158 .59124 L
.59009 .58501 L
.60955 .57682 L
.64284 .55864 L
.6714 .53834 L
Mistroke
.69914 .51345 L
.72358 .48561 L
.74206 .45857 L
.7495 .44517 L
.75642 .43047 L
.76148 .41747 L
.76579 .40359 L
.76756 .39651 L
.76896 .38984 L
.77007 .38334 L
.77089 .37737 L
.77147 .37169 L
.77169 .36882 L
.77188 .36568 L
.77201 .36252 L
.77209 .35914 L
.77211 .35595 L
.77207 .35304 L
.77197 .34982 L
.77182 .3468 L
.7716 .34348 L
.77129 .33991 L
.77042 .33263 L
.76937 .32597 L
.76677 .31377 L
.7631 .30094 L
.75227 .27397 L
.73562 .24477 L
.71527 .21772 L
.69307 .19417 L
.66291 .1687 L
.63146 .14811 L
.5996 .13237 L
.58274 .12592 L
.56369 .12014 L
.54419 .1158 L
.5331 .11405 L
.52782 .11339 L
.52292 .11288 L
.51793 .11246 L
.51342 .11218 L
.5084 .11195 L
.50564 .11187 L
.50305 .11183 L
.49831 .11181 L
.49397 .11188 L
.4892 .11205 L
.48417 .11232 L
.47945 .11267 L
.47433 .11315 L
Mistroke
.46507 .1143 L
.45609 .11575 L
.44786 .11738 L
.4294 .12205 L
.41029 .12844 L
.37978 .142 L
.34859 .16054 L
.31905 .18317 L
.29534 .20585 L
.27198 .23387 L
.25464 .26072 L
.24254 .2856 L
.23734 .29958 L
.23361 .31227 L
.23058 .32625 L
.22937 .33412 L
.22892 .33787 L
.22858 .34137 L
.22831 .3447 L
.22813 .34773 L
.22798 .3511 L
.22793 .35295 L
.2279 .35471 L
.2279 .35821 L
.22795 .36146 L
.22809 .36502 L
.22818 .36681 L
.22831 .3688 L
.22887 .3752 L
.22929 .37883 L
.22975 .3822 L
.2324 .39632 L
.23561 .40851 L
.24 .42153 L
.25369 .45116 L
.27053 .47772 L
.29158 .50366 L
.31453 .52644 L
.34337 .54944 L
.37765 .57047 L
.41054 .58524 L
.42986 .59166 L
.44795 .59623 L
.45618 .59785 L
.46501 .59927 L
.47011 .59995 L
.47474 .60047 L
.47946 .60091 L
.4838 .60124 L
.48794 .60148 L
Mistroke
.49231 .60166 L
.4967 .60175 L
.50152 .60177 L
.50568 .60171 L
.51015 .60157 L
.51435 .60136 L
.51818 .6011 L
.52699 .60029 L
.53666 .59903 L
.54669 .59731 L
.55603 .59534 L
.57249 .59097 L
.58984 .5851 L
.62741 .56775 L
.65889 .54782 L
.70935 .50264 L
.73166 .47464 L
.74808 .44789 L
.7548 .43418 L
.76083 .41927 L
.76535 .40517 L
.7685 .39216 L
.76956 .3865 L
.7705 .38042 L
.77094 .37693 L
.77128 .37374 L
.77155 .37072 L
.77178 .36751 L
.77195 .36418 L
.77206 .3611 L
.77211 .35778 L
.77211 .35589 L
.77209 .35412 L
.77201 .35094 L
.77189 .34801 L
.77169 .34481 L
.77143 .34142 L
.77066 .33437 L
.76968 .32779 L
Mfstroke
.84014 .35679 m
.84011 .35383 L
.84004 .35113 L
.8399 .34802 L
.83971 .34507 L
.83924 .33983 L
.83854 .33414 L
.83753 .32785 L
.83619 .32117 L
.83309 .30913 L
.82421 .28476 L
.81018 .25738 L
.79146 .2292 L
.76608 .1985 L
.70646 .14425 L
.67066 .11985 L
.65024 .10821 L
.62744 .09706 L
.60468 .0878 L
.58341 .08082 L
.57296 .07797 L
.56154 .0753 L
.55062 .07317 L
.54059 .07158 L
.52991 .07027 L
.52401 .06972 L
.51841 .06931 L
.51324 .06903 L
.50845 .06885 L
.50319 .06874 L
.4976 .06873 L
.49202 .06883 L
.48888 .06894 L
.486 .06906 L
.4807 .06937 L
.47513 .06979 L
.46334 .07106 L
.45051 .07298 L
.4382 .07536 L
.42712 .07795 L
.40267 .08521 L
.36057 .10268 L
.32227 .12426 L
.29018 .14683 L
.25702 .1751 L
.23041 .20239 L
.19232 .25321 L
.17815 .27952 L
.17203 .29395 L
.16748 .3072 L
Mistroke
.16413 .31973 L
.16279 .32612 L
.16163 .33299 L
.16109 .33693 L
.16069 .34054 L
.16038 .34393 L
.16025 .34572 L
.16013 .34763 L
.15997 .35097 L
.15989 .35407 L
.15987 .3559 L
.15987 .35763 L
.15991 .36086 L
.16002 .36378 L
.16019 .36699 L
.16044 .37035 L
.16073 .3735 L
.16142 .37913 L
.16238 .38521 L
.16542 .39911 L
.16926 .41193 L
.17976 .43746 L
.1931 .46165 L
.21312 .49042 L
.23645 .51782 L
.26348 .54444 L
.29732 .57216 L
.33171 .59517 L
.3668 .61389 L
.40646 .62964 L
.42834 .63594 L
.44072 .63876 L
.45228 .64091 L
.46443 .64267 L
.47087 .64339 L
.47773 .644 L
.48367 .6444 L
.48926 .64466 L
.49246 .64476 L
.49539 .64483 L
.49862 .64486 L
.50202 .64486 L
.50746 .64477 L
.51328 .64456 L
.51877 .64425 L
.52375 .64388 L
.53593 .64262 L
.54738 .64096 L
.56792 .63685 L
.59012 .63074 L
Mistroke
.6332 .61389 L
.67032 .59394 L
.70215 .57255 L
.76256 .51889 L
.8025 .46856 L
.81781 .44233 L
.82841 .41847 L
.83296 .40493 L
.83612 .39275 L
.83747 .38602 L
.83863 .37878 L
.83905 .37546 L
.83942 .37195 L
.83969 .36869 L
.83989 .36568 L
.84003 .3626 L
.84008 .36098 L
.84012 .35925 L
.84014 .35636 L
.84009 .35317 L
.84001 .35034 L
.83985 .34728 L
.83966 .3444 L
.83943 .34176 L
.83876 .33578 L
.83777 .32923 L
.83496 .31598 L
.83158 .30421 L
.82703 .29155 L
.81635 .26844 L
.79988 .24109 L
.77701 .21095 L
.75297 .18479 L
.72202 .15659 L
.68992 .13232 L
.65156 .10892 L
.62886 .09769 L
.60729 .08877 L
.58621 .08165 L
.57442 .07835 L
.56323 .07567 L
.55214 .07344 L
.54008 .07151 L
.53472 .07081 L
.52906 .07019 L
.52374 .0697 L
.5189 .06934 L
.51395 .06906 L
.50865 .06885 L
.50373 .06875 L
Mistroke
.49908 .06872 L
.49367 .06879 L
.48867 .06894 L
.48319 .06921 L
.47739 .06961 L
.4716 .07012 L
.46631 .07069 L
.45444 .07233 L
.4332 .07648 L
.40887 .08317 L
.38635 .09123 L
.3476 .10937 L
.30961 .13264 L
.2515 .18038 L
.2252 .20835 L
.20185 .23856 L
.18462 .26662 L
.17282 .29192 L
.16847 .30403 L
.1648 .31691 L
.16339 .32313 L
.16227 .32899 L
.16145 .33421 L
.16075 .33996 L
.16044 .3432 L
.16021 .34625 L
.16003 .34946 L
.15996 .35127 L
.15991 .35292 L
.15987 .35571 L
.15988 .35876 L
.15995 .36195 L
.16007 .36491 L
.16016 .36661 L
.16029 .36845 L
.16057 .37179 L
.16127 .37805 L
.16241 .38538 L
.16375 .39216 L
.16679 .40407 L
.17104 .41697 L
.18131 .4406 L
.19797 .46928 L
.21749 .49596 L
.26692 .54752 L
.30162 .5753 L
.33783 .59877 L
.35766 .60946 L
.38 .61976 L
.40039 .62759 L
Mistroke
.4225 .63442 L
.44291 .6392 L
.45463 .64129 L
.46564 .64281 L
.47086 .64339 L
.47581 .64385 L
.48031 .64419 L
.48509 .64448 L
.4903 .6447 L
.49511 .64482 L
.50029 .64486 L
.50601 .6448 L
.51072 .64466 L
.5151 .64447 L
.5199 .64418 L
.52501 .64378 L
.53545 .64268 L
.54489 .64136 L
.56297 .63797 L
.5798 .63379 L
.61622 .62133 L
.65346 .60364 L
Mfstroke
.90816 .35679 m
.90814 .35452 L
.90809 .35246 L
.90798 .35008 L
.90784 .34782 L
.90748 .34381 L
.90694 .33943 L
.90616 .3346 L
.90514 .32946 L
.90276 .32013 L
.8959 .30105 L
.885 .2791 L
.84995 .22886 L
.80056 .17616 L
.73419 .12127 L
.69928 .09807 L
.6586 .07571 L
.63531 .06514 L
.61348 .05674 L
.59071 .04952 L
.56932 .0442 L
.55724 .04182 L
.54389 .03971 L
.53039 .03815 L
.52299 .03753 L
.51967 .03731 L
.51607 .03711 L
.5095 .03685 L
.50328 .03672 L
.49722 .03672 L
.49162 .03681 L
.48558 .03703 L
.48214 .0372 L
.47892 .0374 L
.47244 .03789 L
.46546 .03857 L
.45348 .04008 L
.44074 .04218 L
.41806 .04717 L
.39163 .05498 L
.36827 .06367 L
.32007 .08681 L
.27853 .11239 L
.20996 .16643 L
.16178 .21507 L
.12668 .26017 L
.11245 .28372 L
.10262 .3046 L
.09893 .31469 L
.09579 .32548 L
Mistroke
.09466 .33037 L
.09367 .33553 L
.09293 .34039 L
.09243 .34477 L
.09223 .34702 L
.09206 .34939 L
.09195 .35142 L
.09188 .35364 L
.09184 .35603 L
.09184 .35737 L
.09185 .35859 L
.09191 .36096 L
.092 .36321 L
.09213 .36531 L
.09231 .36753 L
.09272 .37154 L
.09331 .37586 L
.09414 .38065 L
.09637 .39035 L
.09953 .4007 L
.10732 .4197 L
.11887 .44113 L
.13545 .46592 L
.17664 .51472 L
.23803 .57115 L
.27495 .59875 L
.3198 .62662 L
.36633 .64913 L
.38997 .65804 L
.41211 .66484 L
.43507 .6703 L
.44787 .67264 L
.46008 .67439 L
.46556 .67503 L
.4714 .6756 L
.47645 .67601 L
.48191 .67637 L
.48775 .67665 L
.49324 .67681 L
.49977 .67688 L
.50576 .67683 L
.51182 .67666 L
.51828 .67636 L
.52438 .67595 L
.52992 .67548 L
.54346 .67393 L
.55617 .67196 L
.5812 .6666 L
.60778 .6588 L
.65148 .64128 L
Mistroke
.69619 .61738 L
.77344 .56167 L
.8281 .50969 L
.86981 .45855 L
.8842 .43589 L
.89037 .42443 L
.8959 .41254 L
.90022 .4014 L
.90331 .3915 L
.90453 .38677 L
.90565 .38169 L
.90652 .3769 L
.90715 .37255 L
.90766 .36795 L
.90786 .36542 L
.908 .36303 L
.9081 .36082 L
.90815 .35878 L
.90816 .35654 L
.90813 .35416 L
.90806 .35178 L
.908 .35045 L
.90793 .34922 L
.90755 .34457 L
.90728 .3421 L
.90693 .3394 L
.90604 .33393 L
.90371 .32356 L
.90026 .31231 L
.89149 .29141 L
.87916 .26924 L
.86163 .24372 L
.81785 .19317 L
.75822 .13944 L
.72129 .11227 L
.67695 .08518 L
.63134 .06351 L
.60859 .05506 L
.58698 .04849 L
.56494 .04328 L
.55318 .04112 L
.54058 .03928 L
.53477 .03859 L
.52849 .03797 L
.52294 .03753 L
.51708 .03716 L
.51177 .03692 L
.50595 .03676 L
.50066 .03671 L
.49572 .03673 L
Mistroke
.48933 .03688 L
.4861 .03701 L
.48255 .03718 L
.47628 .03758 L
.47041 .03807 L
.45724 .03956 L
.44315 .04175 L
.41766 .04727 L
.3948 .05393 L
.35063 .07132 L
.31134 .09175 L
.23663 .14357 L
.18211 .1932 L
.13771 .2446 L
.11106 .28635 L
.10605 .29661 L
.10161 .3072 L
.09845 .31617 L
.09575 .32565 L
.09455 .33092 L
.09364 .33571 L
.09291 .34055 L
.0924 .34501 L
.09219 .34752 L
.09201 .35019 L
.09195 .35157 L
.0919 .35286 L
.09186 .35428 L
.09184 .35581 L
.09184 .35706 L
.09185 .35841 L
.09187 .35983 L
.09191 .36115 L
.09202 .36352 L
.09219 .36605 L
.09255 .37006 L
.09309 .37432 L
.09441 .382 L
.09637 .39034 L
.09917 .39964 L
.10592 .41668 L
.1143 .43324 L
Mfstroke
.97619 .35679 m
.97618 .35556 L
.97615 .35445 L
.97609 .35316 L
.97601 .35194 L
.97582 .34976 L
.97553 .34739 L
.97511 .34477 L
.97455 .34197 L
.97326 .33689 L
.96953 .3264 L
.96357 .31414 L
.94409 .28482 L
.91575 .251 L
.82625 .16445 L
.75283 .10756 L
.71266 .08176 L
.66567 .05671 L
.6399 .04545 L
.61482 .03623 L
.58976 .02879 L
.56624 .02343 L
.55454 .02136 L
.54356 .01978 L
.53205 .0185 L
.52594 .01798 L
.51938 .01754 L
.51281 .01723 L
.5091 .01711 L
.50566 .01704 L
.49935 .01699 L
.49265 .01707 L
.48667 .01725 L
.48106 .01752 L
.47492 .01791 L
.4683 .01847 L
.45556 .01989 L
.44201 .02193 L
.41785 .02688 L
.38967 .03477 L
.36473 .04362 L
.31314 .0673 L
.2685 .09337 L
.14014 .19453 L
.09708 .23715 L
.06337 .2753 L
.0419 .30486 L
.03535 .31615 L
.02996 .32762 L
.02782 .3334 L
Mistroke
.02625 .33864 L
.02512 .34355 L
.02464 .34627 L
.0243 .34871 L
.02415 .35009 L
.02402 .35154 L
.02393 .35278 L
.02386 .35413 L
.02384 .3549 L
.02382 .35559 L
.02381 .35639 L
.02381 .35715 L
.02383 .3585 L
.02385 .35927 L
.02388 .35997 L
.02396 .36123 L
.02405 .36242 L
.02433 .36512 L
.02471 .36777 L
.02524 .37063 L
.02653 .37598 L
.02811 .38103 L
.03309 .39298 L
.03991 .40548 L
.04949 .42005 L
.11298 .49279 L
.16547 .54196 L
.22908 .59315 L
.30619 .64256 L
.35029 .66406 L
.37647 .67435 L
.40203 .68256 L
.42732 .68885 L
.4418 .69162 L
.45531 .69366 L
.46669 .69496 L
.47319 .69554 L
.47931 .69597 L
.48573 .6963 L
.49163 .69649 L
.49811 .69659 L
.50493 .69656 L
.51193 .69639 L
.51829 .6961 L
.52181 .6959 L
.52563 .69563 L
.53254 .69504 L
.54585 .69351 L
.55806 .69164 L
.58581 .68581 L
Mistroke
.61458 .67743 L
.66086 .65911 L
.70722 .63502 L
.78045 .58605 L
.89493 .48474 L
.93 .4463 L
.95245 .41728 L
.96655 .39372 L
.96939 .38753 L
.97159 .38188 L
.9733 .37656 L
.97454 .37169 L
.97506 .36909 L
.97545 .36674 L
.97578 .36415 L
.97593 .36272 L
.97603 .36138 L
.97611 .36007 L
.97614 .35942 L
.97616 .35869 L
.97619 .35741 L
.97619 .35621 L
.97618 .35552 L
.97616 .3548 L
.97611 .35351 L
.97607 .35279 L
.97603 .35211 L
.9759 .35059 L
.97562 .34806 L
.97526 .34566 L
.97436 .34114 L
.97306 .33619 L
.97123 .33072 L
.96654 .31985 L
.95944 .30704 L
.93818 .27723 L
.86435 .19875 L
.807 .14837 L
.73729 .09712 L
.6615 .05477 L
.63886 .04503 L
.61403 .03597 L
.56784 .02374 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{340.875, 243.188},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40007o0001K2000`400?l00000o`00003oo`3ooooo
0?ooo`40oooo003o0?ooool0oooo0@3oool00?l0ooooo`3oool10?ooo`00o@3oool00`3ooml0W4P0
000000020000000307Boo`3oool0oooo0?X0oooo003o0?ooo`030?noM00007@0_ooo0?d0oooo003o
0?ooo`030?oo_`1d0000M;oo0?d0oooo003o0?ooo`030?oog`2LB000B9cO0?d0oooo003o0?ooo`40
oooo00<0gia80000000000000P0000000`1d_ol0oooo0?ooo`020?ooo`050?noM000000000000000
B02Lgol00P3oool00`3ookl0M000000000020000000=04RLg`3oool0oooo0?ooo`2oM00000000000
B02Lgol0oooo0?noM0000000001809cOo`020?ooo`040;md00000000001809cOo`80oooo00D0okmd
000000000000001809cOo`030?ooo`030=nLB000B9`0gooo0080oooo00<0_g@000000000M;l0a03o
ool00?l0oooo0@3oool01`3o_g@0001809cOo`3oool0oooO09a80000M;l00P3oool03`3ogi`0B018
09cOo`3oool0omnL04P00000M;l0oooo0?oog`2LB000M;oo0?ooo`3ogi`0B00007Boo`020?ooo`0G
0=nLB000B9`0gooO09adW03Oool0okmd0000B02Lgml0W7BL0=ooo`3oool0gia80018W03Ooml0W7BL
0=ooo`3ogi`0B01809cOo`3oool0omnL04P00000M;l00P3oool00`3o_g@0001d0;ooo`020?ooo`04
0=nLB000B9`0gkmd07Bool<0oooo003o0?ooo`40oooo00<0oono07@0001d_ol00P3oool01`3ooml0
W4P004RLg`3oool0gia80018W03Oool00P3oool02P3ogi`0B00004RLg`3oool0_gAd0;ooo`3oool0
oono07@0001d_ol20?ooo`030?noM00004P0W=oo0080oooo00<0oono07@0001d_ol00`3oool00`3o
_g@0001809cOo`020?ooo`030=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d
0000B9cO0080oooo00<0omnL04P0B02Lgol0a03oool00?l0oooo0@3oool00`3ooml0W4P004RLg`03
0?ooo`060?oOW01804P0W=oo0?noM00007@0_ooo0`3oool00`3OW4P0001d0;ooo`030?ooo`030=nL
B00007@0_ooo0080oooo00<0oono07@00018W=l00`3oool00`2oM00007Bo0?ooo`020?ooo`030?oo
_`1d0000B9cO0080oooo00<0okmd0000M02oool00`3oool02`3OW4P0001d0;ooo`3ooml0W4P00000
M02oool0oooo0?oog`2LB000B9cO0<@0oooo003o0?ooo`80oooo00<0gia80000M02oool00P3oool0
1P3ooml0W4P004RLg`3ookl0M00007Boo`<0oooo00<0omnL04P0B02Lgol00P3oool00`3o_g@00018
09cOo`040?ooo`030;md0000B9`0gooo0080oooo00<0gia80018W03Oool00`3oool01`2oM00004RL
0=ooo`3oool0oono07@0001d_ol00`3oool01`3ogi`0B01809cOo`3oool0gia80018B018M;l00P3o
ool00`2oM00007Bo0?ooo`330?ooo`00o`3oool20?ooo`040?noM000000004RL0=ooo`80oooo00H0
gia8001d_`3oool0gia80000M02oool20?ooo`0:0?oo_`1d0000M;oo0?ooo`3ogi`0B01809cOo`3o
ool0giad09cOo`80oooo00<0gia80000B02Lgol00P3oool00`3ogi`0B01809cOo`030?ooo`030=nL
B00004P0W=oo0080oooo00<0gia80000M02oool00P3oool0303ookl0M00007Boo`3oool0okmd0000
M02o_g@0B9cO0?ooo`3o_g@0001d0;oool<0oooo003o0?ooo`80oooo00P0oono07@0001dW9`0B01d
0;ooo`3ooml0W4P0001d_`80oooo00h0_g@00000M02oool0oooo0?noM00007@0_ooo0?ooo`3ogi`0
B01809cOo`3ooml0W4P007Boo`80oooo00<0omnL04P00018W=l00P3oool00`3ooml0W4P004RLg`03
0?ooo`030?oOW0180000B9cO00<0oooo0100_g@00000M02oool0oooo0?noM00007@0_ooo0?ooo`3o
oml0W4P007Boo`3OW4P007Bo0?oOW01804P0W=oo``3oool00?l0oooo0@3oool02@3ogi`0B0000000
0018W=l0oono07@000000000001d0;ooo`030?ooo`050?noM000000000000000001d_ol00`3oool0
1`2oM0000000000000000000B9cO0?ooo`3o_g@00P0000001000B9`0gooo0?oo_`1d000300000006
07Boo`3o_g@0000000000000B9`0gooo0`3oool03`3o_g@00000000000000000M;oo0?ooo`3ookl0
M00000000018W=l0oooo0?noM0000000001809cOo`330?ooo`00o`3ooolR0?ooo`030?noM00007@0
_ooo0=/0oooo003o0?ooob<0oooo00<0giad09cOo`3oool0fP3oool00?l0oooo7@3oool0103ookl0
M0000018W03Oool70?ooo`040?oo_`1d000004RL0=ooom@0oooo003o0?ooool0oooo0@3oool00?l0
ooooo`3oool10?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo0@3oool00?l0ooooo`3oool1
0?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo0@3oool00?l0ooooo`3oool10?ooo`00o`3o
oooo0?ooo`40oooo003o0?ooool0oooo0@3oool00?l0ooooo`3oool10?ooo`00o`3ooooo0?ooo`40
oooo002E0?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0>`3oool00`3ookl0M000
000000020000000304RLg`3oool0oooo03T0oooo00@0_g@0000000000000M;oo?P3oool00`3ookl0
M000000000020000000304RLg`3oool0oooo03d0oooo00<0gia80000000000000`0000000`18W=l0
oooo0?ooo`1E0?ooo`00UP3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo03d0oooo
00<0okmd0000001d_ol0>`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLgcl0oooo00<0okmd0000
001d_ol0@@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo05D0oooo002G0?ooo`03
0?noM000M;l0oooo0440oooo00<0oooO09a80000M;l0>P3oool00`3ooml0W4P004RLg`020?ooo`03
0?noM00004P0W=oo03h0oooo00<0oooO09a80000M;l0@P3oool00`3o_g@007Bo0?ooo`1I0?ooo`00
V03oool00`3o_g@007Bo0?ooo`110?ooo`030=nLB000B9`0gooo03T0oooo00<0oooO09a80018W=l0
0P3oool0103ooml0W4P00018W03Oooln0?ooo`030=nLB000B9`0gooo0480oooo00<0okmd001d_`3o
ool0F03oool008P0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`040?ooo`030?no
M018W=l0oooo0300oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`030?oO
W01804P0W=oo03T0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0n0?ooo`030?oO
W01804P0W=oo04<0oooo00<0okmd04RLg`3oool0E`3oool009X0oooo00<0gia804RLg`3oool0?`3o
ool00`3ookl0M00007Boo`0j0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0?P3o
ool00`3ookl0M00007Boo`140?ooo`030=nLB018W=l0oooo05H0oooo002K0?ooo`030;md001d_ol0
oooo03l0oooo00<0_g@0001d_`3oool0>@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo03l0oooo00<0_g@0001d_`3oool0A03oool00`2oM000M;oo0?ooo`1E0?ooo`00V`3oool00`3o
gi`0B01d0;ooo`0o0?ooo`030=nLB000B9`0gooo03T0oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`0n0?ooo`030=nLB000B9`0gooo04@0oooo00<0omnL04P0M02oool0E@3oool009/0
oooo00<0oooO09a8001d_ol0?`3oool00`3ogi`0B01809cOo`0i0?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0gia80018W03Oool0?P3oool00`3ogi`0B01809cOo`140?ooo`030?oog`2LB000M;oo
05D0oooo002K0?ooo`030?oo_`1d0000B9cO03l0oooo00<0oono07@0001d_ol0>P3oool00`3ogi`0
B00007Boo`020?ooo`030=nLB000M;l0oooo03h0oooo00<0oono07@0001d_ol0A03oool00`3ookl0
M00004RLg`1E0?ooo`00U`3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0@03oool00`2o
M00007Bo0?ooo`0j0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO@03oool00`2oM00007Bo0?oo
o`0o0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`1E0?ooo`00V03oool01@3ogi`0B000
000000000000M;oo03l0oooo00D0oono07@000000000001d0;ooo`0k0?ooo`040=nLB00000000000
001d_cl0oooo00D0oono07@000000000001d0;ooo`100?ooo`050?oOW018000000000000001d_ol0
EP3oool00?l0ooooo`3oool10?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo0@3oool00?l0
ooooo`3oool10?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo0@3oool00?l0ooooo`3oool1
0?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo0@3oool004D0oooo00<0gia8000000000000
0`0000000`18W=l0oooo0?ooo`3o0?oook80oooo00160?ooo`030?noM000M;l0oooo0080oooo00<0
_gAd0;ooo`3oool0203ooooo000008l000006`3oool004L0oooo00<0okmd001d_`3oool0303oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0I0?ooo`00B03oool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo
03X0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo
0480oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo
01T0oooo000h0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`3o_g@0
B9cO0?ooo`0:0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0a03oool00`000000
oooo0?ooo`0I0?ooo`00BP3oool00`3OW4P0B9cO0?ooo`090?ooo`<00000``3oool00`000000oooo
0?ooo`310?ooo`@000006`3oool004/0oooo00<0_g@007Boo`3oool0203oool00`000000oooo0?oo
o`330?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool06@3oool004/0oooo00<0omnL
04P0M02oool0203oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo0<@0oooo00<00000
0?ooo`3oool06@3oool004/0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`2d0?oo
oah00000^03oool00`000000oooo0?ooo`0I0?ooo`00B`3oool00`3ookl0M00004RLg`080?ooo`03
0000003oool0oooo0:X0oooo2P00000?0?ooo`030000003oool0oooo00`0oooo2@00002_0?ooo`03
0000003oool0oooo01T0oooo00170?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?oo
o`030000003oool0oooo0:D0oooo1@00000I0?ooo`030000003oool0oooo01D0oooo1@00002Z0?oo
o`030000003oool0oooo01T0oooo00180?ooo`050?oOW018000000000000001d_ol02@3oool00`00
0000oooo0?ooo`2Q0?ooo`@000007P3oool00`000000oooo0?ooo`0J0?ooo`D00000Y@3oool00`00
0000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`2N0?ooo`<000008P3oool00`000000
oooo0?ooo`0N0?ooo`D00000X@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo
0?ooo`2L0?ooo`8000009@3oool00`000000oooo0?ooo`0Q0?ooo`D00000WP3oool00`000000oooo
0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`2I0?ooo`<000009`3oool00`000000oooo0?oo
o`0U0?ooo`D00000VP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`2F
0?ooo`<00000903oool<000002H0oooo1000002G0?ooo`030000003oool0oooo01T0oooo001F0?oo
o`030000003oool0oooo09<0oooo0`00000J0?ooo`h000001@3oool00`000000oooo0?ooo`020?oo
o`h00000703oool3000009D0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3o
ool0S`3oool4000001T0oooo2000000?0?ooo`030000003oool0oooo00`0oooo2`00000G0?ooo`@0
0000TP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`2<0?ooo`<00000
6@3oool7000001@0oooo00<000000?ooo`3oool05@3oool6000001H0oooo0`00002@0?ooo`030000
003oool0oooo01T0oooo001F0?ooo`800000R`3oool2000001H0oooo1P00000K0?ooo`030000003o
ool0oooo01X0oooo1@00000D0?ooo`@00000R`3oool3000001/0oooo001F0?ooo`030000003oool0
oooo08P0oooo0P00000E0?ooo`D000007`3oool00`000000oooo0?ooo`0N0?ooo`D000004`3oool3
000008/0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0QP3oool20000
01@0oooo1000000S0?ooo`030000003oool0oooo0280oooo1000000B0?ooo`@00000R03oool00`00
0000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`240?ooo`8000004P3oool5000002H0
oooo00<000000?ooo`3oool0903oool500000140oooo100000260?ooo`030000003oool0oooo01T0
oooo001F0?ooo`030000003oool0oooo0840oooo0`00000B0?ooo`@00000:@3oool00`000000oooo
0?ooo`0W0?ooo`D00000403oool4000008@0oooo00<000000?ooo`3oool06@3oool005H0oooo00<0
00000?ooo`3oool0O`3oool200000180oooo1000000/0?ooo`030000003oool0oooo02X0oooo1000
000@0?ooo`@00000PP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1m
0?ooo`8000004P3oool3000002l0oooo00<000000?ooo`3oool0;@3oool300000100oooo10000020
0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo07/0oooo0P00000A0?oo
o`<00000<P3oool00`000000oooo0?ooo`0`0?ooo`800000403oool4000007h0oooo00<000000?oo
o`3oool06@3oool005H0oooo00<000000?ooo`3oool0N@3oool200000140oooo0`00000d0?ooo`03
0000003oool0oooo0380oooo0`00000?0?ooo`@00000O03oool00`000000oooo0?ooo`0I0?ooo`00
EP3oool00`000000oooo0?ooo`1g0?ooo`8000004@3oool2000002`0oooo5P00000/0?ooo`<00000
3`3oool3000007/0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0M@3o
ool200000140oooo0P00000V0?ooo`X000002@3oool00`000000oooo0?ooo`080?ooo`P000009P3o
ool3000000l0oooo0`00001i0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0
oooo07@0oooo00<000000?ooo`3oool03`3oool2000002@0oooo1`00000@0?ooo`030000003oool0
oooo00l0oooo1@00000T0?ooo`<000003`3oool3000007L0oooo00<000000?ooo`3oool06@3oool0
05H0oooo00<000000?ooo`3oool0LP3oool200000100oooo0P00000R0?ooo`H000005@3oool00`00
0000oooo0?ooo`0C0?ooo`H000008@3oool3000000l0oooo0`00001e0?ooo`030000003oool0oooo
01T0oooo001F0?ooo`800000LP3oool00`000000oooo0?ooo`0>0?ooo`8000008@3oool4000001X0
oooo00<000000?ooo`3oool0603oool5000001l0oooo0`00000>0?ooo`@00000L@3oool3000001/0
oooo001F0?ooo`030000003oool0oooo06l0oooo0P00000?0?ooo`8000007`3oool4000001h0oooo
00<000000?ooo`3oool0703oool5000001d0oooo0`00000>0?ooo`<00000LP3oool00`000000oooo
0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1]0?ooo`8000003`3oool2000001h0oooo1000
000Q0?ooo`030000003oool0oooo0200oooo1000000L0?ooo`<000003P3oool300000700oooo00<0
00000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?ooo`0=
0?ooo`8000007@3oool3000002D0oooo00<000000?ooo`3oool08`3oool4000001/0oooo0P00000?
0?ooo`800000K`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1Z0?oo
o`8000003P3oool2000001`0oooo0`00000X0?ooo`030000003oool0oooo02H0oooo0`00000K0?oo
o`8000003`3oool2000006d0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3o
ool0J03oool2000000h0oooo0P00000K0?ooo`@00000:P3oool00`000000oooo0?ooo`0Y0?ooo`<0
00006P3oool2000000l0oooo0P00001[0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo06L0oooo00<000000?ooo`3oool0303oool2000001/0oooo1000000/0?ooo`030000
003oool0oooo02/0oooo1000000I0?ooo`8000003`3oool2000006T0oooo00<000000?ooo`3oool0
6@3oool005H0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0<0?ooo`8000006P3o
ool4000002h0oooo00<000000?ooo`3oool0;@3oool4000001T0oooo0P00000>0?ooo`800000J03o
ool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1T0?ooo`8000003@3oool2
000001X0oooo0`00000a0?ooo`030000003oool0oooo02l0oooo1000000I0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool0oooo00/0oooo0`00000I0?ooo`<00000<`3oool00`00
0000oooo0?ooo`0b0?ooo`<00000603oool2000000h0oooo0P00001U0?ooo`030000003oool0oooo
01T0oooo001F0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool02`3oool2000001T0
oooo0`00000e0?ooo`030000003oool0oooo03@0oooo0`00000H0?ooo`8000003P3oool2000006<0
oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0H03oool2000000`0oooo
0`00000H0?ooo`<00000=`3oool00`000000oooo0?ooo`0f0?ooo`<000005`3oool3000000d0oooo
0P00001R0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo05l0oooo00<0
00000?ooo`3oool02P3oool4000001L0oooo0`00000i0?ooo`030000003oool0oooo03P0oooo0P00
000H0?ooo`<00000303oool200000640oooo00<000000?ooo`3oool06@3oool005H0oooo00<00000
0?ooo`3oool0GP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0000001L0oooo0`00000c
0?oooa000000=@3oool2000001L0oooo0`00000=0?ooo`800000G`3oool00`000000oooo0?ooo`0I
0?ooo`00EP3oool2000005h0oooo00<000000?ooo`3oool02P3oool3000001H0oooo0`00000^0?oo
o`P000001`3oool00`000000oooo0?ooo`050?ooo`T00000;P3oool2000001L0oooo0`00000<0?oo
o`800000G03oool3000001/0oooo001F0?ooo`030000003oool0oooo05/0oooo0P00000;0?ooo`<0
00005P3oool3000002T0oooo2000000>0?ooo`030000003oool0oooo00d0oooo1P00000[0?ooo`80
00005`3oool2000000`0oooo0P00001M0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo05X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0000000F0?ooo`800000
9`3oool6000001@0oooo00<000000?ooo`3oool04P3oool6000002L0oooo0`00000F0?ooo`800000
3@3oool2000005/0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0F@3o
ool00`000000oooo0?ooo`0:0?ooo`<000005@3oool2000002H0oooo1000000I0?ooo`030000003o
ool0oooo01H0oooo1000000W0?ooo`<000005P3oool2000000`0oooo0P00001J0?ooo`030000003o
ool0oooo01T0oooo001F0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool02P3oool2
000001D0oooo0P00000T0?ooo`@000007@3oool00`000000oooo0?ooo`0I0?ooo`@000009P3oool2
000001H0oooo0P00000<0?ooo`800000F@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`00
0000oooo0?ooo`1F0?ooo`800000303oool2000001@0oooo0`00000S0?ooo`@000007`3oool00`00
0000oooo0?ooo`0M0?ooo`<00000903oool2000001H0oooo0P00000=0?ooo`800000E`3oool00`00
0000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1E0?ooo`030000003oool0oooo00X0
oooo0P00000D0?ooo`<000008`3oool300000280oooo00<000000?ooo`3oool0803oool400000280
oooo0P00000E0?ooo`<00000303oool2000005H0oooo00<000000?ooo`3oool06@3oool005H0oooo
00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0:0?ooo`800000503oool300000280oooo
0P00000U0?ooo`030000003oool0oooo0280oooo1000000R0?ooo`030000003oool0oooo01@0oooo
0P00000<0?ooo`800000E@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?oo
o`1B0?ooo`800000303oool2000001@0oooo0P00000R0?ooo`8000009`3oool00`000000oooo0?oo
o`0U0?ooo`<000008@3oool2000001D0oooo0P00000=0?ooo`030000003oool0oooo0580oooo00<0
00000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo0180oooo0`00000P0?ooo`<00000:@3oool00`000000oooo0?ooo`0X
0?ooo`800000803oool2000001D0oooo0`00000<0?ooo`800000DP3oool00`000000oooo0?ooo`0I
0?ooo`00EP3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00X0oooo0P00000D0?oo
o`<000007`3oool3000002/0oooo00<000000?ooo`3oool0:P3oool200000200oooo00<000000?oo
o`3oool04`3oool00`000000oooo0000000<0?ooo`800000D@3oool00`000000oooo0?ooo`0I0?oo
o`00EP3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00X0oooo0P00000D0?ooo`80
00007`3oool3000002d0oooo00<000000?ooo`3oool0:`3oool3000001l0oooo0P00000D0?ooo`<0
00003@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo04d0oooo0P00000<0?ooo`030000003oool0oooo01<0oooo0P00000N0?ooo`<00000
;`3oool00`000000oooo0?ooo`0]0?ooo`<000007P3oool2000001D0oooo0P00000=0?ooo`800000
CP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1<0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0L0?ooo`<00000<@3o
ool00`000000oooo0?ooo`0_0?ooo`<000007P3oool00`000000oooo0?ooo`0C0?ooo`8000003@3o
ool2000004d0oooo00<000000?ooo`3oool06@3oool005H0oooo0P00001<0?ooo`030000003oool0
oooo00X0oooo0P00000D0?ooo`8000007P3oool2000003<0oooo00<000000?ooo`3oool0<@3oool3
000001d0oooo00<000000?ooo`3oool04`3oool2000000h0oooo00<000000?ooo`3oool0B03oool3
000001/0oooo001F0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool02P3oool20000
01@0oooo0P00000M0?ooo`800000=@3oool00`000000oooo0?ooo`0b0?ooo`@00000703oool20000
01@0oooo0`00000=0?ooo`800000BP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`180?ooo`800000303oool00`000000oooo0?ooo`0C0?ooo`800000703oool2000003L0
oooo00<000000?ooo`3oool0=03oool3000001`0oooo0P00000D0?ooo`030000003oool0000000d0
oooo0P0000190?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo04L0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01X0oooo
0P00000i0?ooo`030000003oool0oooo03H0oooo0`00000K0?ooo`800000503oool3000000d0oooo
0P0000180?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01X0oooo0P00
000j0?ooo`030000003oool0oooo03L0oooo0`00000K0?ooo`8000005@3oool2000000h0oooo0P00
00160?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo04@0oooo0P00000<
0?ooo`800000503oool2000001`0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0i
0?ooo`800000703oool00`000000oooo0?ooo`0C0?ooo`8000003P3oool2000004D0oooo00<00000
0?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0;0?oo
o`800000503oool2000001/0oooo0P00000e0?oooa<00000<P3oool3000001/0oooo00<000000?oo
o`3oool04`3oool2000000h0oooo0P0000140?ooo`030000003oool0oooo01T0oooo001F0?ooo`03
0000003oool0oooo0480oooo00<000000?ooo`3oool02`3oool2000001@0oooo0P00000K0?ooo`80
0000;P3oool9000000L0oooo00<000000?ooo`3oool01P3oool8000002h0oooo0P00000K0?ooo`03
0000003oool0oooo01<0oooo0P00000>0?ooo`800000@`3oool00`000000oooo0?ooo`0I0?ooo`00
EP3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo00/0oooo0P00000D0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool0:P3oool5000000l0oooo00<000000?ooo`3oool0
303oool6000002`0oooo0P00000J0?ooo`030000003oool0oooo01<0oooo0P00000>0?ooo`800000
@P3oool00`000000oooo0?ooo`0I0?ooo`00A@3oool00`3ookl0M000000000020000000304RLg`3o
ool0oooo00T0oooo00<000000?ooo`3oool0?`3oool2000000d0oooo00<000000?ooo`3oool04`3o
ool00`000000oooo0?ooo`0I0?ooo`800000:@3oool5000001<0oooo00<000000?ooo`3oool04@3o
ool4000002X0oooo0P00000J0?ooo`8000005@3oool00`000000oooo0?ooo`0<0?ooo`800000@@3o
ool00`000000oooo0?ooo`0I0?ooo`00A`3oool00`3o_g@0000007Boo`0<0?ooo`030000003oool0
oooo03h0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0B0?ooo`8000006`3oool2
000002P0oooo1000000F0?ooo`030000003oool0oooo01<0oooo1@00000X0?ooo`8000006P3oool2
000001D0oooo00<000000?ooo`3oool0303oool200000400oooo00<000000?ooo`3oool06@3oool0
04L0oooo00<0oooO09a80000M;l0303oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool04P3oool2000001/0oooo00<000000?ooo`3oool09P3oool30000
01T0oooo00<000000?ooo`3oool05`3oool4000002H0oooo0P00000J0?ooo`8000005@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool06@3oool004P0
oooo00<0gia80018W03Oool02`3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool04P3oool2000001X0oooo0P00000W0?ooo`800000703oool00`000000
oooo0?ooo`0J0?ooo`<000009P3oool00`000000oooo0?ooo`0H0?ooo`8000005@3oool2000000h0
oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0I0?ooo`00>03oool00`3ooml0W4P0
000000070000000304RLg`3oool0oooo00<0oooo00<0omnL04P0B02Lgol02`3oool00`000000oooo
0?ooo`0j0?ooo`8000003P3oool00`000000oooo0?ooo`0B0?ooo`8000006P3oool2000002H0oooo
0P00000N0?ooo`030000003oool0oooo01`0oooo0`00000U0?ooo`030000003oool0oooo01P0oooo
0P00000E0?ooo`8000003P3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo01T0oooo
00180?ooo`030?oo_`1d0000M;oo00/0oooo0`00000i0?ooo`030000003oool0oooo00`0oooo0P00
000D0?ooo`8000006P3oool2000002D0oooo0P00000P0?ooo`030000003oool0oooo01h0oooo0`00
000T0?ooo`030000003oool0oooo01P0oooo0P00000E0?ooo`8000003P3oool00`000000oooo0?oo
o`0g0?ooo`@000006`3oool004T0oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`0h
0?ooo`030000003oool0oooo00`0oooo0P00000D0?ooo`8000006P3oool2000002@0oooo0P00000R
0?ooo`030000003oool0oooo0200oooo0`00000S0?ooo`8000006@3oool2000001D0oooo0P00000>
0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool06@3oool004T0oooo00<0gia80018
W03Oool02P3oool00`000000oooo0?ooo`0f0?ooo`8000003P3oool2000001@0oooo0P00000J0?oo
o`8000008`3oool2000002@0oooo00<000000?ooo`3oool08P3oool2000002<0oooo0P00000J0?oo
o`030000003oool0oooo01<0oooo0P00000>0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool06@3oool004T0oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`0e0?ooo`03
0000003oool0oooo00d0oooo0P00000D0?ooo`8000006P3oool00`000000oooo0?ooo`0Q0?ooo`80
00009P3oool00`000000oooo0?ooo`0T0?ooo`8000008`3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01<0oooo0P00000>0?ooo`800000>03oool00`000000oooo0?ooo`0I0?ooo`00
B@3oool00`3ookl0M00007Boo`0:0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0
3@3oool2000001D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0P0?ooo`<00000
9`3oool00`000000oooo0?ooo`0U0?ooo`<000008P3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01<0oooo0P00000>0?ooo`800000=`3oool00`000000oooo0?ooo`0I0?ooo`00BP3o
ool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool03@3o
ool2000001D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0P0?ooo`800000:@3o
ool00`000000oooo0?ooo`0W0?ooo`<000008@3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo01<0oooo0P00000>0?ooo`800000=P3oool00`000000oooo0?ooo`0I0?ooo`00B03oool0
1@3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`0=0?ooo`8000005@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01l0oooo0P00
000[0?ooo`030000003oool0oooo02T0oooo0`00000P0?ooo`030000003oool0oooo01L0oooo0P00
000E0?ooo`8000003P3oool2000003D0oooo00<000000?ooo`3oool06@3oool005H0oooo00<00000
0?ooo`3oool0<@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool0603oool00`000000oooo0?ooo`0O0?ooo`800000;03oool00`000000oooo0?ooo`0Z
0?ooo`<00000803oool00`000000oooo0?ooo`0G0?ooo`8000005@3oool00`000000oooo0?ooo`0=
0?ooo`800000=03oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0`0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01h0oooo0`00000]0?ooo`030000003oool0oooo02`0oooo0P00000P0?oo
o`030000003oool0oooo01L0oooo0P00000D0?ooo`8000003`3oool2000003<0oooo00<000000?oo
o`3oool06@3oool005H0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0N0?ooo`80
0000;`3oool00`000000oooo0?ooo`0]0?ooo`<000007P3oool2000001T0oooo00<000000?ooo`3o
ool04`3oool2000000l0oooo0P00000b0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo02h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0C0?ooo`800000
6@3oool00`000000oooo0?ooo`0N0?ooo`800000<03oool00`000000oooo0?ooo`0^0?ooo`<00000
7P3oool2000001T0oooo00<000000?ooo`3oool04`3oool200000100oooo00<000000?ooo`3oool0
;`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0]0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool04`3oool2000001T0oooo0P00000O0?ooo`800000<@3o
ool00`000000oooo0?ooo`0`0?ooo`8000007P3oool2000001T0oooo00<000000?ooo`3oool04`3o
ool200000100oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`0a0?ooo`<000007@3oool2000001T0oooo00<000000?ooo`3oool04`3oool2
00000100oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool20000
02`0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0703oool2000003@0oooo00<000000?ooo`3oool0<P3oool30000
01d0oooo0P00000I0?ooo`030000003oool0oooo01<0oooo0P00000@0?ooo`030000003oool0oooo
02X0oooo0`00000K0?ooo`00EP3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00h0
oooo0P00000E0?ooo`8000006@3oool00`000000oooo0?ooo`0L0?ooo`800000=@3oool00`000000
oooo0?ooo`0d0?ooo`800000703oool00`000000oooo0000000H0?ooo`8000005@3oool00`000000
oooo0?ooo`0>0?ooo`800000;@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo
0?ooo`0Y0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool04`3oool2000001T0oooo
0P00000M0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0=@3oool2000001`0oooo
0P00000I0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool03P3oool2000002`0oooo
00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool05`3oool2000001d0oooo00<0
00000?ooo`3oool0;P3ooolB000002l0oooo0P00000L0?ooo`8000006@3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo00h0oooo0P00000[0?ooo`030000003oool0oooo01T0oooo001F
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01L0oooo0P00000L0?ooo`800000;03oool5000000P0oooo00<00000
0?ooo`3oool01P3oool5000002`0oooo00<000000?ooo`3oool06`3oool2000001T0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0>0?ooo`800000:P3oool00`000000oooo0?ooo`0I
0?ooo`00EP3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`3oool0503oool2000001T0oooo00<000000?ooo`3oool06P3oool2000002/0oooo1000000;0?oo
o`030000003oool0oooo00X0oooo1000000Z0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`0000006@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo0P00000Y0?oo
o`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3o
ool03`3oool2000001D0oooo0P00000I0?ooo`8000006`3oool2000002T0oooo0`00000?0?ooo`03
0000003oool0oooo00h0oooo1000000W0?ooo`030000003oool0oooo01X0oooo0P00000J0?ooo`03
0000003oool0oooo01@0oooo00<000000?ooo`3oool03P3oool2000002P0oooo00<000000?ooo`3o
ool06@3oool005H0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0?0?ooo`800000
5P3oool00`000000oooo0?ooo`0G0?ooo`8000006`3oool2000002P0oooo0P00000B0?ooo`030000
003oool0oooo0100oooo1000000V0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000
003oool0oooo01@0oooo0P00000@0?ooo`8000009`3oool00`000000oooo0?ooo`0I0?ooo`00EP3o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0G0?ooo`8000006`3oool2000002H0oooo0`00000D0?ooo`030000003o
ool0oooo01<0oooo0`00000U0?ooo`030000003oool0oooo01X0oooo0P00000I0?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool03`3oool2000002H0oooo00<000000?ooo`3oool06@3o
ool005H0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0
oooo01@0oooo0P00000I0?ooo`8000006`3oool2000002@0oooo0`00000F0?ooo`030000003oool0
oooo01D0oooo0`00000T0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`0000006@3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00l0oooo0P00000U0?ooo`030000003oool0
oooo01T0oooo001F0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0403oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo01L0oooo0P00000K0?ooo`8000008`3oool30000
01P0oooo00<000000?ooo`3oool05`3oool2000002<0oooo0P00000K0?ooo`8000006@3oool20000
01D0oooo0P00000A0?ooo`800000903oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo0100oooo0P00000F0?ooo`030000003oool0oooo01L0
oooo0P00000K0?ooo`8000008P3oool3000001X0oooo00<000000?ooo`3oool06@3oool200000280
oooo0P00000K0?ooo`8000006@3oool00`000000oooo0?ooo`0D0?ooo`8000004@3oool2000002<0
oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0803oool00`000000oooo
0?ooo`0?0?ooo`8000005P3oool2000001T0oooo00<000000?ooo`3oool06@3oool200000280oooo
0P00000L0?ooo`030000003oool0oooo01X0oooo0`00000Q0?ooo`8000006`3oool00`000000oooo
0?ooo`0G0?ooo`8000005P3oool200000140oooo0P00000R0?ooo`030000003oool0oooo01T0oooo
001F0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo01L0oooo0P00000K0?ooo`030000003oool0oooo0200oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`0L0?ooo`800000803oool2000001/0oooo0P00
000I0?ooo`8000005P3oool00`000000oooo0?ooo`0@0?ooo`8000008@3oool00`000000oooo0?oo
o`0I0?ooo`00EP3oool2000001l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo01L0oooo0P00000K0?ooo`030000003oool0oooo01l0oooo0P00000O
0?ooo`030000003oool0oooo01d0oooo0P00000P0?ooo`8000006`3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01@0oooo0P00000B0?ooo`8000007P3oool3000001/0oooo001F0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool0403oool2000001H0oooo0P00000I0?oo
o`030000003oool0oooo01T0oooo0P00000P0?ooo`800000803oool00`000000oooo0?ooo`0N0?oo
o`800000803oool2000001/0oooo00<000000?ooo`3oool05`3oool2000001H0oooo0P00000B0?oo
o`8000007`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0L0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0G0?ooo`80
00006`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3o
ool0803oool2000001l0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0M0?ooo`03
0000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`0E0?ooo`8000006@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`800000
7P3oool2000001/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0D0?ooo`800000
4P3oool2000001h0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0A0?ooo`8000005`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo0280oooo0P00000N0?ooo`8000006P3oool00`000000oooo0?ooo`0H0?ooo`8000005P3o
ool200000180oooo0P00000M0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0E0?ooo`8000006@3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01`0oooo0P00000U0?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool05@3oool200000180oooo0P00000L0?ooo`030000003oool0
oooo01T0oooo001F0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0L0?ooo`8000009P3oool00`000000oooo0?ooo`0T0?ooo`8000007@3oool20000
01X0oooo00<000000?ooo`3oool0603oool2000001H0oooo0P00000C0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0A0?ooo`8000005`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0
oooo0P00000M0?ooo`8000009`3oool00`000000oooo0?ooo`0U0?ooo`8000007@3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05@3oool2000001<0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06`3oool2000002P0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0K0?ooo`8000006P3oool00`000000oooo
0?ooo`0H0?ooo`8000005`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?oo
o`0B0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`800000703oool2000002T0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`0K0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0I0?ooo`00EP3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0180oooo0P00000H
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0K
0?ooo`800000:@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01X0oooo0P00000J
0?ooo`030000003oool0oooo01P0oooo0P00000G0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo0P00000L0?ooo`800000:P3oool00`000000oooo0?ooo`0X0?oo
o`800000703oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0B0?ooo`8000005`3oool00`000000oooo0?ooo`0I0?oo
o`00EP3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0180oooo0P00000H0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo02T0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0J0?ooo`80
00006P3oool00`000000oooo0?ooo`0H0?ooo`8000005`3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`0H0?ooo`800000703oool00`000000oooo0?ooo`0Z0?ooo`030000
003oool0oooo02X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool06@3oool005H0oooo0P00000C0?ooo`030000003o
ool0oooo01<0oooo0P00000H0?ooo`030000003oool0oooo01L0oooo0P00000J0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0Z0?ooo`8000006`3o
ool2000001X0oooo00<000000?ooo`3oool0603oool2000001L0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0A0?ooo`<000006`3oool002<0oooo00<0oono07Ad_`3oool0<03oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo0P00000L0?ooo`030000003oool0
oooo02@0oooo3`00000V0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0603oool2
000001X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool06@3oool002<0oooo00@0omnL04P0000004P0W=oo;`3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01<0oooo0P00000H0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0603oool2000001/0oooo00<000000?ooo`3oool08P3oool4
000000H0oooo00<000000?ooo`3oool0103oool3000002@0oooo0P00000K0?ooo`8000006P3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01T0oooo000U0?ooo`030;md000000000000
00<0000000<0001d0;ooo`3oool0:03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool05P3oool2000001T0oooo0P00000J0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0803oool3000000T0oooo00<000000?ooo`3oool01P3oool30000
02<0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01P0oooo0P00000H0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0I0?ooo`00:@3oool00`3ogi`0B00007Boo`0Z0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0503oool2000001P0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0H0?ooo`8000006`3oool00`000000oooo0?ooo`0O0?ooo`<000002`3oool00`000000
oooo0?ooo`080?ooo`<000008@3oool2000001/0oooo0P00000J0?ooo`030000003oool0oooo01P0
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0D0?ooo`8000004P3oool00`000000
oooo0?ooo`0I0?ooo`00:03oool00`3OW4P0001d0;ooo`0[0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0F0?ooo`8000006P3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo01T0oooo0P00000P0?ooo`8000003@3oool00`000000oooo
0?ooo`0:0?ooo`<00000803oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0I0?ooo`008`3oool01P3ogi`0B01809cO
o`3o_g@0000004RLgbd0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0D0?ooo`80
0000603oool2000001T0oooo0P00000J0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool07P3oool2000000h0oooo00<000000?ooo`3oool0303oool2000001l0oooo0P00000K0?ooo`03
0000003oool0oooo01P0oooo0P00000I0?ooo`800000603oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool06@3oool002<0oooo00@0omnL04P000000000
07Bo;`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo0P00000J0?ooo`80
00007P3oool200000100oooo00<000000?ooo`3oool03P3oool2000001h0oooo00<000000?ooo`3o
ool06@3oool2000001X0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo01@0oooo0P00000@0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01`0oooo0P00000A0?ooo`030000003oool0oooo00l0oooo0P00000N0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool05@3oool2000000l0oooo00<000000?ooo`3oool0
6@3oool005H0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0E0?ooo`800000603o
ool2000001T0oooo0P00000I0?ooo`8000006P3oool2000001d0oooo0P00000B0?ooo`030000003o
ool0oooo0100oooo0P00000M0?ooo`030000003oool0oooo01T0oooo0P00000I0?ooo`8000006@3o
ool00`000000oooo0?ooo`0G0?ooo`8000005`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo01T0oooo000V0?ooo`060=nLB000000000000000000004P0W=oo:P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool2000001T0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01/0oooo
0P00000C0?ooo`030000003oool0oooo0140oooo0P00000L0?ooo`8000006P3oool2000001X0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01D0oooo
0P00000>0?ooo`030000003oool0oooo01T0oooo000T0?ooo`090?oo_`1d0000B9cO0?oOW02L_ol0
oooo0=nLB000B9`0gooo02T0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0I0?oo
o`00903oool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d0;ooo`0Y0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo01X0oooo0P00000J0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603o
ool2000001P0oooo0P00000G0?ooo`8000003@3oool00`000000oooo0?ooo`0I0?ooo`008`3oool0
2@3ogi`0W;oo0?ooo`3ooml0W7BL0=ooo`3oool0omnL04Qd_`0Z0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`8000006@3oool2000001X0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool04`3oool2000001`0oooo00<000000?ooo`3oool0603oool2000001T0
oooo0P00000I0?ooo`8000006@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool06@3oool002<0oooo00T0oknL09cOo`3oool0_g@007Boo`3oool0gia8
0000M02oool0:P3oool2000000/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0D
0?ooo`8000006`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo0P00000:
0?ooo`<000006`3oool002<0oooo00@0oono07@000000000M;oo;`3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0I
0?ooo`8000006P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool06@3oool2000001L0oooo00<000000?oo
o`3oool05@3oool2000001X0oooo0P00000J0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0G0?ooo`800000603oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool06@3oool002L0oooo00H0okmd00000000000000000000M02ooolY0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01P0oooo0P00000H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool20000
01T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo
01T0oooo000O0?ooo`060;nLW03Oool0oooo0?ooo`3ogi`0B0000`0000001000B9`0gooo0?oo_`1d
M;lZ0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo01L0oooo0P00000I0?ooo`8000006P3oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3oool2000001X0oooo00<0
00000?ooo`3oool0603oool2000001T0oooo00<000000?ooo`3oool0603oool2000001T0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo01T0oooo000N
0?ooo`030?oog`2LB000000000<0000000<007Bo0?ooo`3oool00P3oool00`3o_g@0001d0;ooo`0Z
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H
0?ooo`8000006@3oool00`000000oooo0?ooo`0H0?ooo`8000006@3oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H
0?ooo`8000006P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool06@3oool001h0oooo00<0oono07@00018W=l0=@3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo01T0oooo00170?ooo`040;md00000000000007Boo`/0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`0I0?ooo`009P3oool01P3ogi`0B0000000000000000018
09cOoaX0oooo00H0gia8001d_`3oool0omnL04P00018W=l:0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool06@3oool002D0oooo00<0_g@00000001d_ol00P3oool00`3ooml0W7BL0=oo
o`0H0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo01T0oooo000T0?ooo`030=nLB018W=l0oooo00@0oooo00<0gibL
0;ooo`3oool05`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`0I0?ooo`008`3oool00`3ogi`0M9cO0?ooo`04
0?ooo`030?oog`2LM9`0gooo01P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`08
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool06@3oool002<0oooo00<0oknL09cO
o`3oool00`3oool00`3ooml0W4P007Boo`0J0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l0203ooooo000008l000006`3oool002<0oooo00<0omnL04Qd_`3oool00P3oool00`3OW4P0
000004RLg`0K0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0203oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo01T0oooo000T0?ooo`060?noM0000000000000000000
B9`0gooo703oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0I0?ooo`00AP3oool0103ooml0W4P00018W03Oool20?oo
o`030=nLB000B9`0gooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0I0?oo
o`00A`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0I0?ooo`009P3oool01P3ogi`0B000000000000000001809cO
oa`0oooo00H0gia80000M02oool0oooO09a80018W=l80?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool06@3oool002@0oooo00@0oooO09a80000B9`0gooo0P3oool00`3o_g@004RL0=oo
o`0L0?ooo`040=nLB00000000000001d_`T0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool2000001T0oooo
00<000000?ooo`3oool05`3oool2000001X0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool06@3oool002@0oooo00<0
_gA809cOo`3oool00`3oool00`3ooml0W4Qd0;ooo`0Y0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool05`3oool2000001T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0H0?ooo`8000006@3oool2000001X0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo01P0oooo0P00000J0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0I0?ooo`008`3oool00`3ogi`0W;oo0?ooo`04
0?ooo`030?oOW018M;l0oooo02T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01L0oooo0P00000I0?ooo`030000003oool0oooo01P0oooo0P00000J
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool06@3oool002<0oooo00<0oknL09cO
o`3oool00`3oool00`3ogi`0B4RL0=ooo`0Z0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01L0oooo0P00000J0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo0P00000I0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01T0oooo000S0?ooo`04
0?oo_`1d0000000007Boobl0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool2000001P0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool05`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01T0oooo001F0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
6@3oool2000001L0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0:0?ooo`030000
003oool0oooo01T0oooo000S0?ooo`030?oo_`2LW=l0oooo00<0oooo00@0okmd0000000007@0_ooo
:@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
603oool2000001/0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool06@3oool002<0oooo00@0oono09bLg`3ookl0M000
100000000`00B9`0gooo0?ooo`0X0?ooo`8000002`3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo01L0oooo0P00000I0?ooo`8000006@3oool2000001X0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01@0oooo0P00000K0?ooo`030000
003oool0oooo01P0oooo0P00000J0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00T0oooo0`00000K0?ooo`008`3o
ool02@3ogi`0B0000000000004P0W=oo0?ooo`3ooml0W4Qd0;ooo`0Z0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0C0?ooo`8000006`3oool20000
01X0oooo00<000000?ooo`3oool0603oool2000001T0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool06@3oool001l0
oooo00@0oooO09a8000007@0_ooo<`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
01H0oooo0P00000I0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06@3oool005H0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0
oooo0P00000I0?ooo`030000003oool0oooo01P0oooo0P00000J0?ooo`030000003oool0oooo01/0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01/0
oooo00<000000?ooo`3oool0603oool2000001X0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0I0?ooo`008`3oool01@3ogkl0_moo0?ooo`3oool0_g@000<0000000<0001d0;ooo`3o
ool0:03oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool05`3oool2000001T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01/0oooo0P00000C0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`0I0?ooo`8000006P3oool00`000000oooo0?ooo`0G0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo01T0oooo000S0?ooo`030;md0000000000000080000000D0001d0;ooo`3oool0
_ibL0=ooo`0Y0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool05@3oool2000001T0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0
oooo00<000000?ooo`3oool0703oool200000180oooo00<000000?ooo`3oool0403oool2000001d0
oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`8000006@3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0I0?ooo`008@3oool01@2oM000000000000000B9`0gooo00<0oooo00<0oooO09a8001d
_ol00P3oool00`2oM00004RL0=ooo`0U0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01`0oooo0P00000A0?ooo`03
0000003oool0oooo00l0oooo0P00000M0?ooo`8000006`3oool00`000000oooo0?ooo`0H0?ooo`80
00006@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0I0?ooo`008`3oool00`3o_i`0W=oo0?ooo`080?ooo`030?oO
W018M;l0oooo02D0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0E0?ooo`800000
603oool2000001T0oooo00<000000?ooo`3oool0603oool2000001X0oooo00<000000?ooo`3oool0
7@3oool200000100oooo00<000000?ooo`3oool03P3oool2000001h0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
6@3oool002<0oooo00<0oono07Ad_`3oool01`3oool00`3o_g@004RL0=ooo`0V0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01h0oooo0P00000>0?ooo`030000003oool0oooo00d0oooo0P00000O0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo01T0oooo000S0?ooo`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3oool09`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01D0oooo0P00000H0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01l0oooo0`00000<0?ooo`030000003oool0oooo00/0oooo0P00000P0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo01T0oooo000T0?ooo`030?oOW0180000000000D0000000<0B9cO0?ooo`3oool09`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05P3oool2
000001T0oooo00<000000?ooo`3oool0603oool2000001/0oooo00<000000?ooo`3oool0803oool2
000000/0oooo00<000000?ooo`3oool02@3oool300000240oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool06@3oool0
02T0oooo00<0okmd0000B02Lgol0:P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool05P3oool2000001X0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`0I0?ooo`8000008P3oool3000000T0oooo00<000000?ooo`3oool01`3oool30000
0280oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool06@3oool002P0oooo00<0_g@0001d_`3oool0:`3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool05P3oool2000001T0
oooo00<000000?ooo`3oool0603oool2000001/0oooo00<000000?ooo`3oool08P3oool5000000D0
oooo00<000000?ooo`3oool0103oool4000002@0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool06@3oool002<0oooo
00H0omnL04P0001d_ol0okmd0000001d_ol]0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0F0?ooo`8000006@3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo01T0oooo0P00000V0?ooo`l000009P3oool00`000000oooo0?ooo`0J0?oo
o`8000006P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01T0oooo000S0?ooo`04
0?oOW01800000000001d_bl0oooo0P00000D0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo0P00000K0?oo
o`800000;03oool00`000000oooo0?ooo`0Z0?ooo`800000703oool00`000000oooo0?ooo`0H0?oo
o`8000006@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01<0oooo00<000000?oo
o`3oool04@3oool3000001/0oooo001F0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool04`3oool00`000000oooo0?ooo`0F0?ooo`8000006@3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo01X0oooo0P00000[0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool06@3oool005H0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0603oool2000001T0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
6@3oool005H0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo01H0oooo0P00000I0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06P3o
ool2000002X0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo01<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool0
0`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06`3oool2
000002T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo0180oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06`3oool20000
02P0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo01T0oooo0P00000M0?ooo`8000009`3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0E0?ooo`8000006@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool09@3oool00`000000oooo
0?ooo`0U0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool06@3oool2000001P0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0I0?ooo`8000007P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo02@0oooo00<0
00000?ooo`3oool0703oool2000001X0oooo0P00000I0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo001F
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0E
0?ooo`8000006@3oool00`000000oooo0?ooo`0I0?ooo`8000007P3oool00`000000oooo0?ooo`0S
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0I
0?ooo`800000603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0oooo0280oooo00<000000?ooo`3oool07@3oool2000001X0oooo0P00000I0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`0E0?ooo`8000006@3oool00`000000oooo0?ooo`0I0?ooo`80
00007`3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool07@3oool2000001/0oooo0P00000H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo01d0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool06P3oool2000001l0oooo0P00000Q0?ooo`030000
003oool0oooo0200oooo00<000000?ooo`3oool07P3oool2000001X0oooo0P00000I0?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0E0?ooo`8000006@3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo01h0oooo0P00000P0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool07P3o
ool2000001/0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo0100oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool2
00000200oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0N0?ooo`8000007`3oool0
0`000000oooo0?ooo`0M0?ooo`800000803oool2000001/0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool07@3oool3
000001/0oooo001F0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool06P3oool20000
0200oooo0P00000N0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool07`3oool20000
01`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01X0oooo0P00000P0?ooo`<00000703oool00`000000
oooo0?ooo`0K0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0P0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo0240oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0P0?ooo`8000006`3oool00`000000oooo
0?ooo`0I0?ooo`8000008P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0240oooo
00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo0240oooo0P00000I0?ooo`030000003oool0oooo01L0oooo0P00
000S0?ooo`800000703oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01T0oooo001F
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0D
0?ooo`8000006@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0280oooo0P00000G
0?ooo`030000003oool0oooo01D0oooo0P00000T0?ooo`800000703oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0S
0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo02@0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool06`3oool00`000000oooo0?ooo`0S0?ooo`8000005@3oool00`000000oooo0?ooo`0C0?oo
o`8000009P3oool00`000000oooo0?ooo`0J0?ooo`8000006@3oool00`000000oooo0?ooo`0D0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0I0?oo
o`00EP3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01X0oooo0P00000V0?ooo`80
00004`3oool00`000000oooo0?ooo`0@0?ooo`@000009P3oool00`000000oooo0?ooo`0J0?ooo`80
00006@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool09@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0G0?ooo`800000
703oool00`000000oooo0?ooo`0V0?ooo`@000003`3oool00`000000oooo0?ooo`0=0?ooo`D00000
9`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0
6@3oool005H0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo01@0oooo0P00000I0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool09`3o
ool4000000d0oooo00<000000?ooo`3oool02`3oool4000002T0oooo00<000000?ooo`3oool06`3o
ool2000001P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo02L0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo02X0oooo1P0000070?ooo`030000003oool0
oooo00H0oooo1@00000/0?ooo`030000003oool0oooo01/0oooo0P00000H0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0
oooo01T0oooo001F0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0_0?oooa000000<03oool00`000000oooo0?ooo`0K0?ooo`8000006@3oool00`00
0000oooo0?ooo`0D0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0G0?ooo`8000007@3oool00`000000
oooo0?ooo`0e0?ooo`030000003oool0oooo03@0oooo0P00000M0?ooo`8000006@3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo0P00000/0?ooo`030000003oool0oooo01T0
oooo001F0?ooo`800000:`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01<0oooo
0P00000I0?ooo`8000007@3oool2000003D0oooo00<000000?ooo`3oool0<`3oool2000001d0oooo
0P00000I0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`0Z0?ooo`<000006`3oool005H0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0603oool2000001d0oooo0P00
000d0?ooo`030000003oool0oooo0380oooo0P00000N0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02d0oooo00<0
00000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0M
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0<@3oool2000001h0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0]0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0G0?oo
o`8000007`3oool200000340oooo00<000000?ooo`3oool0<03oool2000001h0oooo0P00000I0?oo
o`030000003oool0oooo01@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01L0oooo0P00000O0?ooo`80
0000<03oool00`000000oooo0?ooo`0^0?ooo`8000007`3oool2000001T0oooo00<000000?ooo`3o
ool0503oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3o
ool06@3oool005H0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool05`3oool2000001l0oooo0P00000_0?ooo`030000
003oool0oooo02d0oooo0P00000O0?ooo`8000006@3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0I0?ooo`00EP3o
ool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0G0?ooo`800000803oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo02`0oooo0P00000O0?ooo`8000006@3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo00`0oooo0P00000d0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0
oooo0340oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0C0?ooo`8000006@3oool2
00000200oooo0P00000/0?ooo`030000003oool0oooo02X0oooo0P00000P0?ooo`8000006@3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0=03oool0
0`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool04`3oool2000001T0oooo0P00000P0?ooo`800000:`3oool00`00
0000oooo0?ooo`0Y0?ooo`8000007`3oool3000001X0oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0<0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool06@3oool004D0
oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`090?ooo`030000003oool0oooo03<0
oooo0P00000?0?ooo`030000003oool0oooo01<0oooo0P00000I0?ooo`8000008@3oool2000002T0
oooo00<000000?ooo`3oool09`3oool200000200oooo00<000000?ooo`0000006P3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`0I0?ooo`00A`3oool00`3o_g@0000007Boo`0<0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`8000006@3oool200000240oooo
0`00000W0?ooo`030000003oool0oooo02H0oooo0P00000P0?ooo`<000006P3oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0=`3oool00`000000oooo
0?ooo`0I0?ooo`00A`3oool00`3ooml0W4P0001d_`0<0?ooo`030000003oool0oooo03H0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`8000006@3oool200000280oooo0`00
000U0?ooo`030000003oool0oooo02@0oooo0P00000Q0?ooo`8000006`3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?oo
o`0I0?ooo`00B03oool00`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo03L0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`8000006@3oool2000002<0oooo0`00000S
0?ooo`030000003oool0oooo0280oooo0P00000R0?ooo`8000006`3oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0I
0?ooo`00B03oool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`8000006@3oool2000002@0oooo0`00000Q0?oo
o`030000003oool0oooo0200oooo0P00000S0?ooo`8000006`3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo00/0oooo0P00000l0?ooo`030000003oool0oooo01T0oooo00180?ooo`03
0?oo_`1d0000M;oo00/0oooo0`00000i0?ooo`030000003oool0oooo00d0oooo0P00000D0?ooo`80
00006@3oool2000002D0oooo0`00000O0?ooo`030000003oool0oooo01h0oooo0P00000T0?ooo`80
00006P3oool2000001D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0i0?ooo`@0
00006`3oool004T0oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`0j0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool04P3oool2000001T0oooo0P00000V0?ooo`<00000
7@3oool00`000000oooo0?ooo`0L0?ooo`<00000903oool2000001X0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0
6@3oool004T0oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`0k0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool04P3oool2000001X0oooo00<000000?ooo`3oool09P3o
ool2000001/0oooo00<000000?ooo`3oool06P3oool3000002D0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0?P3o
ool00`000000oooo0?ooo`0I0?ooo`00B@3oool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0
oooo03`0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0B0?ooo`8000006P3oool0
0`000000oooo0?ooo`0W0?ooo`<00000603oool00`000000oooo0?ooo`0H0?ooo`<000009@3oool2
000001/0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo03l0oooo00<000000?ooo`3oool06@3oool004T0oooo00<0oono07@0001d_ol02P3oool00`00
0000oooo0?ooo`0m0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool04P3oool20000
01X0oooo0P00000Z0?ooo`<000005@3oool00`000000oooo0?ooo`0D0?ooo`@000009`3oool20000
01/0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
0400oooo00<000000?ooo`3oool06@3oool004X0oooo00<0_g@0001d_`3oool02@3oool00`000000
oooo0?ooo`0n0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool04P3oool2000001/0
oooo00<000000?ooo`3oool0:P3oool300000180oooo00<000000?ooo`3oool03`3oool6000002T0
oooo0P00000K0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool02P3oool2000004<0
oooo00<000000?ooo`3oool06@3oool004P0oooo00D0oono07@000000000001d0;ooo`090?ooo`03
0000003oool0oooo03l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0B0?ooo`80
00006`3oool2000002d0oooo1@00000=0?ooo`030000003oool0oooo00/0oooo1P00000/0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo04<0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0
@03oool2000000l0oooo00<000000?ooo`3oool04P3oool2000001/0oooo0P00000_0?ooo`T00000
1P3oool00`000000oooo0?ooo`040?ooo`P00000<03oool00`000000oooo0?ooo`0I0?ooo`800000
5P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0
6@3oool005H0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo0180oooo0P00000L0?ooo`800000=@3oool?000003D0oooo0P00000K0?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`150?ooo`030000003o
ool0oooo01T0oooo001F0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01X0oooo0P00000k0?ooo`030000003oool0
oooo03T0oooo0P00000K0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`160?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo
04@0oooo0P00000>0?ooo`030000003oool0oooo01<0oooo0P00000L0?ooo`800000>@3oool00`00
0000oooo0?ooo`0g0?ooo`<000006`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`160?ooo`030000003oool0oooo00`0oooo0P00000D0?ooo`8000007@3oool2000003L0
oooo00<000000?ooo`3oool0=P3oool2000001`0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool06@3oool005H0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0180oooo
0P00000M0?ooo`800000=P3oool00`000000oooo0?ooo`0d0?ooo`<00000703oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00T0oooo0P00001;0?ooo`030000003oool0oooo01T0oooo
001F0?ooo`030000003oool0oooo04P0oooo0P00000>0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0703oool2000003@0oooo00<000000?ooo`3oool0<`3oool3000001`0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo04/0oooo00<0
00000?ooo`3oool06@3oool005H0oooo0P00001;0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`0M0?ooo`800000<P3oool00`000000oooo0?ooo`0a
0?ooo`<00000703oool2000001H0oooo0P00000;0?ooo`030000003oool0oooo04X0oooo0`00000K
0?ooo`00EP3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool04`3oool2000001h0oooo0`00000`0?ooo`030000003oool0oooo02l0oooo0`00000M0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`1=0?oo
o`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01d0oooo0`00000^0?ooo`03
0000003oool0oooo02d0oooo0`00000N0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000
003oool0oooo04d0oooo0P00000=0?ooo`8000005@3oool00`000000oooo0?ooo`0N0?ooo`<00000
;03oool00`000000oooo0?ooo`0[0?ooo`<000007`3oool00`000000oooo0?ooo`0D0?ooo`800000
303oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003o
ool0oooo04l0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0C0?ooo`800000803o
ool3000002X0oooo00<000000?ooo`3oool0:@3oool300000200oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool06@3o
ool005H0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0
oooo01<0oooo0P00000Q0?ooo`<00000:03oool00`000000oooo0?ooo`0W0?ooo`<00000803oool2
000001D0oooo0P00000<0?ooo`800000D`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`00
0000oooo0?ooo`1A0?ooo`8000003@3oool2000001D0oooo0P00000Q0?ooo`<000009P3oool00`00
0000oooo0?ooo`0U0?ooo`<000008@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`1C0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`0P0?ooo`<00000903oool00`000000oooo0?ooo`0S0?ooo`<000008P3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0E03oool00`000000
oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool0503oool2000002<0oooo0`00000Q0?ooo`030000003oool0oooo0200oooo
0`00000S0?ooo`8000005@3oool2000000d0oooo00<000000?ooo`3oool0E@3oool00`000000oooo
0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1E0?ooo`8000003@3oool2000001D0oooo0`00
000S0?ooo`<000007`3oool00`000000oooo0?ooo`0L0?ooo`@000009@3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo00/0oooo0P00001H0?ooo`030000003oool0oooo01T0oooo001F
0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0D
0?ooo`800000903oool3000001d0oooo00<000000?ooo`3oool0603oool4000002P0oooo00<00000
0?ooo`3oool04`3oool2000000d0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0I
0?ooo`00EP3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0503oool3000002D0oooo1@00000H0?ooo`030000003oool0oooo01D0oooo1000000Y0?oo
o`8000005@3oool00`000000oooo0?ooo`0;0?ooo`800000F`3oool00`000000oooo0?ooo`0I0?oo
o`00EP3oool00`000000oooo0?ooo`1I0?ooo`8000003@3oool2000001H0oooo0`00000W0?ooo`H0
00004`3oool00`000000oooo0?ooo`0@0?ooo`D00000:`3oool2000001D0oooo0P00000=0?ooo`03
0000003oool0oooo05/0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0
F`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01@0oooo0`00000Z0?ooo`H00000
3`3oool00`000000oooo0?ooo`0;0?ooo`L00000;03oool2000001H0oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo01T0oooo001F0?ooo`800000G@3o
ool2000000d0oooo0P00000F0?ooo`<00000;03oool=000000@0oooo00<000000?ooo`3oool00`3o
ool<000002d0oooo0P00000G0?ooo`030000003oool0oooo00/0oooo0P00001M0?ooo`<000006`3o
ool005H0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`0<0?ooo`8000005P3oool2
000003L0oooo2`00000f0?ooo`<000005`3oool2000000d0oooo00<000000?ooo`3oool0G`3oool0
0`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1O0?ooo`8000003P3oool00`00
0000oooo0?ooo`0E0?ooo`800000>P3oool00`000000oooo0?ooo`0g0?ooo`<00000603oool00`00
0000oooo0?ooo`0;0?ooo`800000HP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000
oooo0?ooo`1Q0?ooo`030000003oool0oooo00`0oooo0P00000G0?ooo`800000>03oool00`000000
oooo0?ooo`0e0?ooo`<00000603oool2000000d0oooo00<000000?ooo`3oool0HP3oool00`000000
oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1R0?ooo`8000003P3oool00`000000oooo
0?ooo`0F0?ooo`800000=P3oool00`000000oooo0?ooo`0c0?ooo`<000006@3oool00`000000oooo
0?ooo`0;0?ooo`800000I@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo00`0oooo0P00000G0?ooo`@00000<`3oool00`000000oooo0?oo
o`0a0?ooo`<000006P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06D0oooo00<0
00000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0I@3oool2000000h0oooo0P00000G
0?ooo`@00000<@3oool00`000000oooo0?ooo`0_0?ooo`8000006`3oool2000000`0oooo0P00001X
0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo06L0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0F0?ooo`@00000;`3oool00`000000oooo0?ooo`0/0?oo
o`<00000703oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo06P0oooo00<000000?oo
o`3oool06@3oool005H0oooo00<000000?ooo`3oool0J03oool2000000h0oooo0P00000H0?ooo`@0
0000;@3oool00`000000oooo0?ooo`0Z0?ooo`<00000703oool2000000`0oooo0P00001[0?ooo`03
0000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0
3@3oool2000001T0oooo1000000Z0?ooo`030000003oool0oooo02L0oooo0`00000M0?ooo`800000
3@3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003o
ool0oooo06/0oooo0P00000?0?ooo`030000003oool0oooo01T0oooo0`00000X0?ooo`030000003o
ool0oooo02D0oooo0`00000M0?ooo`8000003P3oool00`000000oooo0?ooo`1/0?ooo`030000003o
ool0oooo01T0oooo001F0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool03@3oool2
000001`0oooo0`00000U0?ooo`030000003oool0oooo0280oooo0`00000N0?ooo`8000003P3oool2
000006l0oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0KP3oool20000
00l0oooo0P00000L0?ooo`@000008P3oool00`000000oooo0?ooo`0O0?ooo`@000007P3oool20000
00l0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool200000740
oooo00<000000?ooo`3oool03P3oool2000001d0oooo1@00000N0?ooo`030000003oool0oooo01`0
oooo1000000O0?ooo`8000003`3oool200000700oooo0`00000K0?ooo`00EP3oool00`000000oooo
0?ooo`1a0?ooo`800000403oool2000001h0oooo1`00000I0?ooo`030000003oool0oooo01P0oooo
1@00000P0?ooo`800000403oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo01T0oooo
001F0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool03`3oool200000240oooo1`00
000D0?ooo`030000003oool0oooo01<0oooo1`00000Q0?ooo`800000403oool2000007D0oooo00<0
00000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0M03oool200000140oooo0P00000U
0?ooo`D00000403oool00`000000oooo0?ooo`0>0?ooo`L000008`3oool300000100oooo0P00001g
0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo07H0oooo0P00000A0?oo
o`800000:03oool9000000L0oooo00<000000?ooo`3oool0103oool:000002P0oooo0P00000A0?oo
o`800000N@3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`1h0?ooo`80
00004@3oool2000002l0oooo3`00000^0?ooo`<000004@3oool2000007/0oooo00<000000?ooo`3o
ool06@3oool005H0oooo00<000000?ooo`3oool0NP3oool200000140oooo0`00000c0?ooo`030000
003oool0oooo0340oooo0P00000B0?ooo`800000O@3oool00`000000oooo0?ooo`0I0?ooo`00EP3o
ool00`000000oooo0?ooo`1l0?ooo`8000004P3oool200000340oooo00<000000?ooo`3oool0;`3o
ool200000180oooo0P00001o0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0
oooo07h0oooo0`00000A0?ooo`800000;`3oool00`000000oooo0?ooo`0/0?ooo`<000004P3oool2
00000840oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0P@3oool20000
0140oooo0P00000]0?ooo`030000003oool0oooo02X0oooo0P00000C0?ooo`800000P`3oool00`00
0000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`230?ooo`8000004@3oool3000002X0
oooo00<000000?ooo`3oool09`3oool300000180oooo0`0000250?ooo`030000003oool0oooo01T0
oooo001F0?ooo`030000003oool0oooo08D0oooo0P00000B0?ooo`<000009`3oool00`000000oooo
0?ooo`0S0?ooo`@000004`3oool2000008P0oooo00<000000?ooo`3oool06@3oool005H0oooo00<0
00000?ooo`3oool0Q`3oool300000180oooo0`00000T0?ooo`030000003oool0oooo01l0oooo1000
000E0?ooo`800000RP3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`2:
0?ooo`8000004`3oool400000200oooo00<000000?ooo`3oool06`3oool4000001L0oooo0P00002<
0?ooo`030000003oool0oooo01T0oooo001F0?ooo`800000S@3oool3000001@0oooo1000000L0?oo
o`030000003oool0oooo01H0oooo1@00000H0?ooo`<00000S03oool3000001/0oooo001F0?ooo`03
0000003oool0oooo08l0oooo0P00000F0?ooo`D000005`3oool00`000000oooo0?ooo`0A0?ooo`D0
00006`3oool200000940oooo00<000000?ooo`3oool06@3oool005H0oooo00<000000?ooo`3oool0
T@3oool3000001P0oooo2000000?0?ooo`030000003oool0oooo00`0oooo1@00000N0?ooo`800000
T`3oool00`000000oooo0?ooo`0I0?ooo`00EP3oool00`000000oooo0?ooo`2D0?ooo`8000007P3o
oolN00000240oooo0P00002E0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0
oooo09H0oooo0`00000Z0?ooo`030000003oool0oooo02X0oooo0`00002G0?ooo`030000003oool0
oooo01T0oooo001F0?ooo`030000003oool0oooo09T0oooo0P00000X0?ooo`030000003oool0oooo
02H0oooo1000002J0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo09/0
oooo1000000T0?ooo`030000003oool0oooo0280oooo1000002N0?ooo`030000003oool0oooo01T0
oooo001F0?ooo`030000003oool0oooo09l0oooo1@00000O0?ooo`030000003oool0oooo01d0oooo
1@00002R0?ooo`030000003oool0oooo01T0oooo001F0?ooo`030000003oool0oooo0:@0oooo1@00
000J0?ooo`030000003oool0oooo01L0oooo1P00002W0?ooo`030000003oool0oooo01T0oooo0015
0?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0203oool00`000000oooo0?ooo`2Y
0?ooo`H00000503oool00`000000oooo0?ooo`0B0?ooo`D00000[@3oool00`000000oooo0?ooo`0I
0?ooo`00AP3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00P0oooo00<000000?oo
o`3oool0[`3oool5000000l0oooo00<000000?ooo`3oool03@3oool500000;80oooo00<000000?oo
o`3oool06@3oool004L0oooo00<0okmd001d_`3oool0303oool00`000000oooo0?ooo`2d0?oooal0
0000]`3oool00`000000oooo0?ooo`0I0?ooo`00B03oool00`3o_g@007Bo0?ooo`0;0?ooo`030000
003oool0oooo0<<0oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`0I0?ooo`00B@3o
ool00`3o_g@0B9cO0?ooo`0:0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0a03o
ool00`000000oooo0?ooo`0I0?ooo`00BP3oool00`3OW4P0B9cO0?ooo`090?ooo`<00000``3oool0
0`000000oooo0?ooo`310?ooo`@000006`3oool004/0oooo00<0_g@007Boo`3oool0203oool00`00
0000oooo0?ooo`330?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool06@3oool004/0
oooo00<0omnL04P0M02oool0203oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0440
oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo0480
oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0I0?ooo`00B`3oool00`3ooml0W4P0
07Boo`080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo01T0oooo001;0?ooo`030?oo_`1d0000B9cO00P0ooooo`00
002?000001/0oooo00170?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`3o0?oook80oooo
00180?ooo`050?oOW018000000000000001d_ol0o`3ooonc0?ooo`00o`3ooooo0?ooo`40oooo003o
0?ooool0oooo0@3oool00?l0ooooo`3oool10?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo
0@3oool00?l0oooo<03oool01@2oM0000000000000000000B9cO0</0oooo003o0?oooc40oooo00<0
oooO09a80018W=l0c03oool00?l0oooo<P3oool00`3OW4P004RL0=ooo`3;0?ooo`00o`3ooolb0?oo
o`030?oOW01807@0_ooo0</0oooo003C0?ooo`060=nLB000000000000000000004P0W=oo1`3oool0
1@3OW4P00000000000000000M;oo0080oooo00<0gia80018W03Oool00P3oool0<02oM0000000001d
_`3oool0oono07@000000000001d0;ooo`3OW4P0000004RLg`3oool0omnL04P00000000004RL0=oo
o`2oM0000000001d_`3oool0okmd0000000007@0_ooo0?ooo`3ogi`0B00000000000B9`0gooo0;md
0000000007Bo0?ooo`3OW4P004RL0=ooo`3oool0okmd0000M02oool0oooo0?oog`2LB000000007Bo
o`D0oooo00<0okmd0000000000001`0000002@1d_ol0oooo0?oOW018000000000000M02ookl0M000
07Boo`020?ooo`0;0?oOW018000000000018W03Oool0_g@000000000M;l0oooo0?oOW01800000P00
00002@1d_ol0gia800000000B9`0gooo0?noM0000000001809cOo`020?ooo`040;md000000000018
09cOo`80oooo00D0okmd000000000000001809cOo`030?ooo`030=nLB000B9`0gooo0080oooo00<0
_g@000000000M;l0V`3oool00=@0oooo00@0oooO09a80000B9`0gooo1`3oool00`3ogi`0B00007Bo
o`020?ooo`060;md0000M;l0oooo0?noM00007@0_ooo0P3oool0:`3OW4P004RL0=noM01d_ol0gia8
0018W03Oool0oono07@0M02o_g@0001d09bLM02Lgol0omnL04P0B02Lgol0omnL04Qd_`2oM00007Bo
0?noW02Lgol0omnL04P0M02ookl0W9cO0?ooo`3ogi`0B01809cOo`3ogi`0B7Bo0;md0000M;l0oknL
09cOo`3o_g@0001809cOo`3oool0oono07@0001d_ol00P3oool0102oM00007Bo0;mdM02oool50?oo
o`040?oog`2LB00004RL0=ooo`<0oooo00/0oooO09a80000B9`0gooo0;md0000B9`0gooo0?oo_`1d
07@0M7@004RLg`020?ooo`0T0?oOW01804P0W=oo0?oOW018M;l0_g@0001d_`3o_i`0W=oo0;md0000
B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?noM00004P0W=oO09adW03Oool0oooo0=nLB000
B9`0gooO09adW03Oool0omnL04P0B02Lgol0oooo0?oOW018000007Bo0P3oool00`3o_g@0001d0;oo
o`020?ooo`040=nLB000B9`0gkmd07BooiX0oooo003E0?ooo`030?noM00004P0W=oo00L0oooo00<0
gia80018W03Oool00`3oool01@3OW9`0_ooo0?oo_`1d0000B9cO0080oooo00H0omnL04P0B02Lgol0
oooo0;md0000M;l20?ooo`0S0?oo_`1dB4P0B01809cOo`3oool0oooO09a80018W=l0oooo0?noM01d
M7@004RL0=ooo`3oool0oono07@0001d_ol0gibL0;ooo`3ooml0W4P004RLg`3oool0okmd07AdM000
B9`0gooo0?ooo`3ookl0M00004RLg`3oool0oooO09a80000M;l00P3oool00`3OW4P004RL0=ooo`07
0?ooo`030?noM00004P0W=oo00D0oooo00@0giad09cOo`2oM00007Bo0P3oool0403ookl0M0000018
W03Oool0oooo0?oog`2LB000B9cO0?ooo`3o_g@0M7Ad0018W03Oool0oooo0;md0000M;l30?ooo`07
0;md00000000M;oo0?ooo`3ookl0M00007Boo`030?ooo`030?noM00004P0W=oo0080oooo00<0gia8
0018W03Oool00P3oool01`3ogi`0B00004RLg`3oool0oono07@00018W=l00P3oool00`3ogi`0B018
09cOo`2K0?ooo`00e@3oool00`3ookl0M00004RLg`070?ooo`050?noM0000000000000000018W=l0
0`3oool0303ooml0W4P00000M02oool0oooo0?oog`2LB000B9cO0?ooo`3o_g@004RL0=ooo`80oooo
00<0okmd00000018W=l00P3oool01`2oM00004RL0=ooo`3oool0_g@00000B02Lgol00P3oool00`2o
M000B9cO0?ooo`020?ooo`0D0;md0000B9`0gooo0?ooo`2oM000001809cOo`3oool0oooO09a80000
B9`0gooo0?ooo`3OW4P0000007Boo`3oool0omnL04P0B02Lgol70?ooo`030?oo_`1d0000B9cO00L0
oooo00<0gia80018W03Oool00P3oool00`3o_g@0001d0;ooo`020?ooo`0;0;md0000B9`0gooo0?oo
o`2oM000001809cOo`3oool0okmd0000M02oool00P3oool00`3ookl0M00004RLg`020?ooo`030;md
0000M;l0oooo0080oooo00<0oono07@00018W=l00P3oool00`3o_g@0001d0;ooo`030?ooo`0;0=nL
B00007@0_ooo0?oog`2LB000001d0;ooo`3oool0oooO09a80018W=l0V`3oool00=D0oooo00<0oooO
09a80000M;l01`3oool01`3ookl0M01809cOo`3oool0gia800000018W=l00P3oool00`3OW4P004Q8
04Qd_`020?ooo`060;md0000M;l0oooo0?oo_`1d0000M;oo0P3oool00`3ooml0W4P0001d_`020?oo
o`070?noM00007@0_ooo0?ooo`3ookl0M00007Boo`020?ooo`030=nLB000B9`0gooo0080oooo00L0
okmd0000M02oool0oooo0?oo_`1d0000M;oo0080oooo00/0gia80018B01dW=l0oooo0?noM00007@0
M7Ad0;ooo`3ooml0W4P004RLg`070?ooo`030?oog`2LB00007Bo0080oooo00<0oono09bLg`3oool0
0P3oool00`3ogi`0B00007Boo`020?ooo`030?oo_`1d0000M;oo0080oooo00/0okmd0000M02oool0
oooo0?oo_`1d0000M;oo0?ooo`3ookl0M00007Boo`030?ooo`070;md0000B9`0gooo0?ooo`3OW4P0
04RL0=ooo`030?ooo`070;md0000B9`0gooo0?ooo`3ookl0M00007Boo`030?ooo`070?oOW01804P0
W=oo0?ooo`3OW4P004Q804Qd_`020?ooo`030;md0000M;l0oooo09X0oooo003F0?ooo`030=nLB000
04P0B4P000<0000001P0B9cO0?ooo`3oool0oooo0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3o
ool0okmd0000M02o_g@0B9cO0?ooo`3o_g@0001d0;ooo`3oool0okmd0000M02oool20?ooo`070=nL
B00007@0_ooo0?oOg`3Ookl0M00007Boo`020?ooo`0<0;md0000M;l0oooo0?ooo`3ogi`0B01d0;oo
o`3oool0omoO0=oo_`1d0000M;oo0P3oool0402oM00007Bo0?ooo`3oool0okmd0000M02LW7@0W=oo
0?oo_`1d0000M;no07Ad_`3oool0_g@00018W03Oool70?ooo`060=nLB00007@0_ooo0?ooo`3ogi`0
B7Bo103oool0302oM000001d0;ooo`3oool0oooO09a80018W=l0oooo0?oOg`3Ookl0M00007Boo`80
oooo00L0_g@0001d_`3oool0oooo0?noM00007@0_ooo0080oooo00L0gia80000M02oool0oooo0?oO
W01804P0W=oo00<0oooo00<0gia80000B02Lgol00P3oool00`3OW4P0001d0;ooo`020?ooo`0<0?oo
_`1d0000M;oo0?ooo`3o_g@0001d0;noM018W=l0oooo0?noM00007@0_oooVP3oool00=H0oooo00<0
okmd0000B02Lgol00P3oool0103ogi`0B0000000M02oool20?ooo`0?0=nLB00007@0_ooo0?ooo`3O
W4P004RL0=ooo`3ooml0W4P007Boo`3OW4P007Bo0?oOW01804P0W=oo0080oooo00/0gia80000M02o
ool0oooo0;md000004P0W=oo0?oOW018M;l0_g@0001d_`020?ooo`0<0?noM000B9`0gooo0?ooo`3o
okl0M00007Boo`3oool0omnL04Qd_`2oM00007Bo0P3oool0403o_g@004RL0=ooo`3oool0oono07@0
B02Lgml0W4Q809cOg`2LB000B9cO0?noM000M;l0okmd0000M02oool70?ooo`030?noM00000000000
0080000000<0001d0;ooo`3oool00P3oool0303ooml0W4P00018W03Oool0oono09a8B000B9`0gooo
0?oOW018M;l0_g@0001d_`80oooo00<0okmd0018W03Oool00P3oool02`3OW4P0001d0;ooo`3oool0
_g@00000B02Lgol0oooo0?oog`2LB000B9cO00<0oooo00<0omnL04P00018W=l00`3oool0402oM000
001d0;ooo`3oool0okmd0000M02oool0oooo0?oog`2LB000M;oo0=nLB000M;l0omnL04P0B02LgonJ
0?ooo`00eP3oool00`3ookl0M00004RLg`030?ooo`030?oOW018000007Bo00<0oooo00h0omnL04P0
00000000000007Boo`3ookl0M00000000018W=l0oooo0?noM0000000001809cOo`<0oooo00X0omnL
04P000000000B9bo09a80018W=l0oooo0?noM000000007Bo0P3oool00`3ogi`0B01809cOo`020?oo
o`070;md0018W=l0oooo0?ooo`3o_g@00000001d_`020?ooo`0@0?oOW01804P0W=oo0?oo_`1d0000
000004RLg`3ooml0W4P000000018W=l0oooo0?noM000000004RL0=ooo`L0oooo00<0oono07@0001d
_ol00P3oool00`2oM4P0W=oo0?ooo`040?ooo`0:0=nLB0000000001d0;noM00007@0_ooo0?ooo`3o
_g@00000001d_`80oooo00<0omnL04P0B02Lgol00`3oool0203ogi`0B00000000018W9`0M00004RL
g`3ookl0M0000`0000001P1d_ol0okmd00000000000004RL0=ooo`<0oooo00l0okmd000000000000
000007Boo`3oool0oono07@000000000B9cO0?ooo`3o_g@000000000B02Lgol0VP3oool00=L0oooo
00<0_g@00018W03Oool00`3oool00`2oM000001d0;ooo`0G0?ooo`030;md0000M;l0oooo00X0oooo
00<0gia80018W03Oool07P3oool01P2oM00007Bo0?ooo`3oool0omnL07BLgal0oooo00<0okmd0000
M02oool0/P3oool00=L0oooo00<0gia80000M02oool00`3oool00`2oM000001d0;ooo`0G0?ooo`03
0?noM00007@0_ooo00X0oooo00<0omnL04P0M02oool07P3oool00`3OW4P0001d0;ooo`040?ooo`03
0=nLW02oool0oooo01`0oooo00<0giad09cOo`3oool0/@3oool00=L0oooo00<0okmd0000001d_ol0
0P3oool00`3o_g@00000001d_`0H0?ooo`030?oOW01804P0W=oo00X0oooo00<0oooO09a8001d_ol0
7P3oool00`3o_g@0001809cOo`030?ooo`030?oo_`1d04P0W=oo0280oooo00@0oono07@00000B9`0
goooZ`3oool00=H0oooo00<0_g@0000000000000100000000`00M;l0oooo0?ooo`0G0?ooo`030?oo
g`2LB000B9cO00/0oooo00<0gia804RLg`3oool0703oool00`2oM0000000000000060000000307Bo
o`3oool0oooo0<l0oooo003f0?ooo`040?oo_`1d00000000001d_`T0oooo00@0oooO09a800000000
07Bon03oool00?l0ooooo`3oool10?ooo`00o`3ooooo0?ooo`40oooo003o0?ooool0oooo0@3oool0
0?l0ooooo`3oool10?ooo`00\
\>"],
  ImageRangeCache->{{{0, 510.313}, {363.75, 0}} -> {-4.13866, -2.81629, \
0.0218158, 0.0218158}}]
}, Closed]],

Cell[TextData[{
  "If we initiate the swing at the bottom with finite velocity, we find that \
when the total energy is negative the pendulum swings back and forth over a \
limited range, but for positive energy it orbits the pivot with monotonically \
increasing angle and a limited band of velocities.  The critical condition in \
which the bob just reaches the top and stops is represented by the ",
  StyleBox["separatrix",
    FontSlant->"Italic"],
  " which divides the phase space into these two distinct regions.  The \
inverted equilibrium is unstable, of course."
}], "Text"],

Cell[BoxData[
    \(\(plots\  = \ 
        Table[pendulumPlot[0, 0, 0, 0, \[Theta]\& . \_0, \(-5\), 5, 
            PlotRange \[Rule] {{\(-2\) \[Pi], 2  \[Pi]}, Automatic}, 
            DisplayFunction \[Rule] Identity], {\[Theta]\& . \_0, \(-3\), 3, 
            0.5}];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Show[plots, DisplayFunction \[Rule] $DisplayFunction];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .50134 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.0795775 0.250668 0.0795775 [
[.02254 -0.0125 -9.46875 -15.6875 ]
[.02254 -0.0125 9.46875 0 ]
[.18169 -0.0125 -9.34375 -15.6875 ]
[.18169 -0.0125 9.34375 0 ]
[.34085 -0.0125 -9.34375 -15.6875 ]
[.34085 -0.0125 9.34375 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.65915 -0.0125 -5 -15.6875 ]
[.65915 -0.0125 5 0 ]
[.81831 -0.0125 -5 -15.6875 ]
[.81831 -0.0125 5 0 ]
[.97746 -0.0125 -5.125 -15.6875 ]
[.97746 -0.0125 5.125 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.9688 -31.375 ]
[.5 -0.0125 21.9688 -15.6875 ]
[-0.0125 .01194 -18.6875 -7.84375 ]
[-0.0125 .01194 0 7.84375 ]
[-0.0125 .09151 -18.6875 -7.84375 ]
[-0.0125 .09151 0 7.84375 ]
[-0.0125 .17109 -18.6875 -7.84375 ]
[-0.0125 .17109 0 7.84375 ]
[-0.0125 .25067 -10 -7.84375 ]
[-0.0125 .25067 0 7.84375 ]
[-0.0125 .33025 -10 -7.84375 ]
[-0.0125 .33025 0 7.84375 ]
[-0.0125 .40982 -10 -7.84375 ]
[-0.0125 .40982 0 7.84375 ]
[-0.0125 .4894 -10 -7.84375 ]
[-0.0125 .4894 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .25067 -34.375 -20.4688 ]
[-0.01875 .25067 -18.6875 20.4688 ]
[ 0 .50134 .125 0 ]
[ 1 0 .125 0 ]
[.5 .51384 -49.7813 0 ]
[.5 .51384 49.7813 15.6875 ]
[ 0 0 0 0 ]
[ 1 .50134 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02254 0 m
.02254 .00625 L
s
gsave
.02254 -0.0125 -70.4688 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.18169 0 m
.18169 .00625 L
s
gsave
.18169 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.34085 0 m
.34085 .00625 L
s
gsave
.34085 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.65915 0 m
.65915 .00625 L
s
gsave
.65915 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.81831 0 m
.81831 .00625 L
s
gsave
.81831 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.97746 0 m
.97746 .00625 L
s
gsave
.97746 -0.0125 -66.125 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.06232 0 m
.06232 .00375 L
s
.10211 0 m
.10211 .00375 L
s
.1419 0 m
.1419 .00375 L
s
.22148 0 m
.22148 .00375 L
s
.26127 0 m
.26127 .00375 L
s
.30106 0 m
.30106 .00375 L
s
.38063 0 m
.38063 .00375 L
s
.42042 0 m
.42042 .00375 L
s
.46021 0 m
.46021 .00375 L
s
.53979 0 m
.53979 .00375 L
s
.57958 0 m
.57958 .00375 L
s
.61937 0 m
.61937 .00375 L
s
.69894 0 m
.69894 .00375 L
s
.73873 0 m
.73873 .00375 L
s
.77852 0 m
.77852 .00375 L
s
.8581 0 m
.8581 .00375 L
s
.89789 0 m
.89789 .00375 L
s
.93768 0 m
.93768 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.9688 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(position) show
102.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01194 m
.00625 .01194 L
s
gsave
-0.0125 .01194 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
0 .09151 m
.00625 .09151 L
s
gsave
-0.0125 .09151 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .17109 m
.00625 .17109 L
s
gsave
-0.0125 .17109 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .25067 m
.00625 .25067 L
s
gsave
-0.0125 .25067 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .33025 m
.00625 .33025 L
s
gsave
-0.0125 .33025 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .40982 m
.00625 .40982 L
s
gsave
-0.0125 .40982 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .4894 m
.00625 .4894 L
s
gsave
-0.0125 .4894 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .02785 m
.00375 .02785 L
s
0 .04377 m
.00375 .04377 L
s
0 .05968 m
.00375 .05968 L
s
0 .0756 m
.00375 .0756 L
s
0 .10743 m
.00375 .10743 L
s
0 .12334 m
.00375 .12334 L
s
0 .13926 m
.00375 .13926 L
s
0 .15518 m
.00375 .15518 L
s
0 .18701 m
.00375 .18701 L
s
0 .20292 m
.00375 .20292 L
s
0 .21884 m
.00375 .21884 L
s
0 .23475 m
.00375 .23475 L
s
0 .26658 m
.00375 .26658 L
s
0 .2825 m
.00375 .2825 L
s
0 .29841 m
.00375 .29841 L
s
0 .31433 m
.00375 .31433 L
s
0 .34616 m
.00375 .34616 L
s
0 .36208 m
.00375 .36208 L
s
0 .37799 m
.00375 .37799 L
s
0 .39391 m
.00375 .39391 L
s
0 .42574 m
.00375 .42574 L
s
0 .44165 m
.00375 .44165 L
s
0 .45757 m
.00375 .45757 L
s
0 .47349 m
.00375 .47349 L
s
.25 Mabswid
0 0 m
0 .50134 L
s
gsave
-0.01875 .25067 -108 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.4688 11.8438 translate
90 rotate
-81.4688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(velocity) show
99.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02254 .49509 m
.02254 .50134 L
s
.18169 .49509 m
.18169 .50134 L
s
.34085 .49509 m
.34085 .50134 L
s
.5 .49509 m
.5 .50134 L
s
.65915 .49509 m
.65915 .50134 L
s
.81831 .49509 m
.81831 .50134 L
s
.97746 .49509 m
.97746 .50134 L
s
.125 Mabswid
.06232 .49759 m
.06232 .50134 L
s
.10211 .49759 m
.10211 .50134 L
s
.1419 .49759 m
.1419 .50134 L
s
.22148 .49759 m
.22148 .50134 L
s
.26127 .49759 m
.26127 .50134 L
s
.30106 .49759 m
.30106 .50134 L
s
.38063 .49759 m
.38063 .50134 L
s
.42042 .49759 m
.42042 .50134 L
s
.46021 .49759 m
.46021 .50134 L
s
.53979 .49759 m
.53979 .50134 L
s
.57958 .49759 m
.57958 .50134 L
s
.61937 .49759 m
.61937 .50134 L
s
.69894 .49759 m
.69894 .50134 L
s
.73873 .49759 m
.73873 .50134 L
s
.77852 .49759 m
.77852 .50134 L
s
.8581 .49759 m
.8581 .50134 L
s
.89789 .49759 m
.89789 .50134 L
s
.93768 .49759 m
.93768 .50134 L
s
.25 Mabswid
0 .50134 m
1 .50134 L
s
.99375 .01194 m
1 .01194 L
s
.99375 .09151 m
1 .09151 L
s
.99375 .17109 m
1 .17109 L
s
.99375 .25067 m
1 .25067 L
s
.99375 .33025 m
1 .33025 L
s
.99375 .40982 m
1 .40982 L
s
.99375 .4894 m
1 .4894 L
s
.125 Mabswid
.99625 .02785 m
1 .02785 L
s
.99625 .04377 m
1 .04377 L
s
.99625 .05968 m
1 .05968 L
s
.99625 .0756 m
1 .0756 L
s
.99625 .10743 m
1 .10743 L
s
.99625 .12334 m
1 .12334 L
s
.99625 .13926 m
1 .13926 L
s
.99625 .15518 m
1 .15518 L
s
.99625 .18701 m
1 .18701 L
s
.99625 .20292 m
1 .20292 L
s
.99625 .21884 m
1 .21884 L
s
.99625 .23475 m
1 .23475 L
s
.99625 .26658 m
1 .26658 L
s
.99625 .2825 m
1 .2825 L
s
.99625 .29841 m
1 .29841 L
s
.99625 .31433 m
1 .31433 L
s
.99625 .34616 m
1 .34616 L
s
.99625 .36208 m
1 .36208 L
s
.99625 .37799 m
1 .37799 L
s
.99625 .39391 m
1 .39391 L
s
.99625 .42574 m
1 .42574 L
s
.99625 .44165 m
1 .44165 L
s
.99625 .45757 m
1 .45757 L
s
.99625 .47349 m
1 .47349 L
s
.25 Mabswid
1 0 m
1 .50134 L
s
0 .25067 m
1 .25067 L
s
.5 0 m
.5 .50134 L
s
gsave
.5 .51384 -110.781 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Pendulum) show
114.625 15.813 moveto
(Equation) show
158.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .50134 L
0 .50134 L
closepath
clip
newpath
.5 Mabswid
1 .01194 m
.99777 .01195 L
.99467 .012 L
.99293 .01204 L
.99134 .01209 L
.98841 .01222 L
.98532 .01239 L
.97878 .01288 L
.97165 .01361 L
.95854 .01548 L
.94469 .01817 L
.92023 .02452 L
.87446 .04033 L
.82875 .05734 L
.80853 .06375 L
.79775 .06662 L
.78782 .06883 L
.77873 .07045 L
.7705 .07156 L
.76582 .07203 L
.7614 .07236 L
.75899 .0725 L
.75675 .0726 L
.75429 .07268 L
.75305 .0727 L
.75168 .07272 L
.74942 .07273 L
.74731 .07271 L
.74499 .07266 L
.74369 .07262 L
.74251 .07257 L
.74013 .07245 L
.73754 .07229 L
.73283 .0719 L
.72844 .07143 L
.72378 .07082 L
.71537 .06944 L
.69627 .06508 L
.67462 .05849 L
.63039 .04217 L
.58664 .02662 L
.56393 .02019 L
.55169 .0174 L
.53851 .015 L
.53194 .01405 L
.52574 .01331 L
.52015 .01278 L
.51726 .01256 L
.51409 .01235 L
.51131 .0122 L
Mistroke
.50832 .01208 L
.50517 .01199 L
.50222 .01195 L
.49929 .01194 L
.49651 .01196 L
.49401 .01201 L
.4913 .01209 L
.48825 .01222 L
.48545 .01238 L
.47916 .01284 L
.47362 .01338 L
.46781 .01409 L
.45554 .016 L
.44226 .01871 L
.41612 .02576 L
.39262 .03362 L
.35142 .0491 L
.31002 .06332 L
.29948 .06619 L
.28849 .06869 L
.27921 .07037 L
.27391 .07114 L
.26909 .07171 L
.26426 .07216 L
.25968 .07246 L
.25717 .07258 L
.25485 .07266 L
.25352 .07269 L
.25231 .07271 L
.25091 .07273 L
.2496 .07273 L
.24738 .07271 L
.24504 .07266 L
.24304 .07259 L
.24084 .07249 L
.23848 .07236 L
.23594 .07217 L
.23136 .07176 L
.22652 .0712 L
.22136 .07046 L
.21204 .0688 L
.19059 .0635 L
.14961 .04949 L
.10621 .0332 L
.08617 .02647 L
.06364 .02012 L
.05108 .01727 L
.03929 .01512 L
.02793 .01356 L
.02155 .0129 L
Mistroke
.01575 .01245 L
.01248 .01226 L
.01069 .01217 L
.00902 .01211 L
.00607 .01201 L
.00282 .01195 L
.00101 .01194 L
Mfstroke
.00101 .01194 m
0 .01194 L
s
1 .05173 m
.99761 .05174 L
.99502 .05179 L
.99262 .05186 L
.99004 .05197 L
.98719 .05214 L
.98145 .05259 L
.97578 .05319 L
.9706 .05388 L
.95905 .05589 L
.94846 .05828 L
.93853 .06097 L
.90039 .07492 L
.86789 .0901 L
.83765 .10512 L
.80805 .11849 L
.79287 .12401 L
.7864 .12597 L
.77942 .12777 L
.77334 .12906 L
.76688 .13012 L
.76334 .13056 L
.76012 .13087 L
.75857 .131 L
.75689 .1311 L
.7553 .13118 L
.75384 .13124 L
.75236 .13128 L
.7508 .1313 L
.74947 .1313 L
.74801 .13129 L
.74644 .13125 L
.74556 .13122 L
.74476 .13119 L
.74319 .13111 L
.74173 .13102 L
.73892 .13079 L
.73587 .13047 L
.73031 .1297 L
.72422 .12857 L
.71745 .127 L
.70347 .12279 L
.67624 .11171 L
.64408 .09605 L
.61025 .07965 L
.59104 .07135 L
.56967 .06351 L
.55007 .05792 L
.53867 .05544 L
.53341 .05451 L
Mistroke
.52785 .05366 L
.5231 .05306 L
.51788 .05253 L
.51313 .05216 L
.51101 .05203 L
.5087 .05191 L
.506 .05182 L
.50345 .05175 L
.50097 .05173 L
.49868 .05173 L
.4962 .05176 L
.49347 .05183 L
.4921 .05188 L
.49063 .05195 L
.48796 .05209 L
.48267 .05248 L
.47705 .05304 L
.46698 .05444 L
.45512 .05672 L
.4445 .0593 L
.40546 .07278 L
.33746 .10521 L
.31061 .11745 L
.29631 .12287 L
.28345 .12677 L
.27706 .1283 L
.27107 .12947 L
.26778 .12999 L
.26422 .13046 L
.261 .1308 L
.25803 .13103 L
.25651 .13112 L
.25488 .1312 L
.25334 .13126 L
.25193 .13129 L
.25037 .1313 L
.24948 .1313 L
.24866 .13129 L
.24779 .13128 L
.24699 .13126 L
.24608 .13124 L
.24521 .13121 L
.24356 .13113 L
.24181 .13102 L
.23867 .13077 L
.23528 .1304 L
.23154 .12989 L
.22489 .12871 L
.21852 .12727 L
.20632 .12375 L
Mistroke
.17728 .11218 L
.11585 .08225 L
.09585 .07333 L
.07614 .06571 L
.05778 .05992 L
.04812 .05745 L
.03741 .05521 L
.03229 .05433 L
.02679 .05352 L
.02202 .05294 L
.01683 .05244 L
.01377 .0522 L
.01095 .05203 L
.00825 .0519 L
.00687 .05184 L
.00538 .0518 L
.00282 .05174 L
.00041 .05172 L
Mfstroke
.00041 .05172 m
0 .05173 L
s
.74786 .24852 m
.74678 .24745 L
.74499 .24566 L
.74241 .24308 L
.73869 .23937 L
.7327 .2334 L
.72398 .22477 L
.70992 .21101 L
.68946 .19158 L
.6607 .16598 L
.64186 .15066 L
.61737 .13287 L
.59015 .11637 L
.56232 .10356 L
.54666 .09831 L
.53929 .09634 L
.53134 .09459 L
.52688 .09378 L
.52281 .09315 L
.51867 .09261 L
.51484 .0922 L
.51035 .09185 L
.50798 .09171 L
.50678 .09166 L
.50548 .09161 L
.50331 .09155 L
.50127 .09152 L
.50003 .09151 L
.49886 .09152 L
.49665 .09155 L
.49475 .0916 L
.4927 .09168 L
.49076 .09178 L
.48899 .09189 L
.48488 .09223 L
.48057 .0927 L
.47247 .09389 L
.4637 .09564 L
.44836 .09982 L
.43453 .10479 L
.40468 .11921 L
.38048 .13432 L
.33854 .16662 L
.30997 .1921 L
.28929 .21177 L
.27606 .22473 L
.26753 .23318 L
.26148 .2392 L
.25764 .24303 L
.25495 .24572 L
Mistroke
.25326 .24741 L
.25218 .24848 L
.25214 .24852 L
Mfstroke
.3863 .30877 m
.3993 .32409 L
.41619 .33927 L
.43636 .3527 L
.44896 .35898 L
.46132 .36372 L
.46734 .36555 L
.47398 .36719 L
.48033 .36841 L
.48337 .36888 L
.48624 .36924 L
.48903 .36953 L
.49159 .36974 L
.49295 .36983 L
.49444 .3699 L
.49601 .36997 L
.49751 .37001 L
.49909 .37003 L
.50052 .37003 L
.50217 .37001 L
.50372 .36998 L
.50516 .36992 L
.50649 .36986 L
.50948 .36966 L
.51255 .36937 L
.51534 .36905 L
.52157 .36808 L
.5273 .3669 L
.53266 .36554 L
.54441 .36169 L
.55667 .35636 L
.57738 .34403 L
.5963 .32847 L
.61062 .3128 L
.62109 .29763 L
.62551 .28942 L
.62925 .28077 L
.63171 .27341 L
.63272 .26955 L
.63361 .26537 L
.634 .26309 L
.6343 .26098 L
.63453 .25901 L
.63473 .25684 L
.6348 .25591 L
.63486 .2549 L
.63491 .25395 L
.63494 .25308 L
.63496 .2521 L
.63497 .25104 L
Mistroke
.63497 .25003 L
.63496 .2491 L
.63493 .24801 L
.63491 .24739 L
.63488 .24683 L
.63475 .24469 L
.63458 .24274 L
.63434 .24069 L
.63369 .23641 L
.6327 .23171 L
.63021 .22319 L
.62714 .21542 L
.61747 .19795 L
.6053 .18221 L
.58857 .16593 L
.56946 .15207 L
.55878 .14604 L
.54634 .14039 L
.53439 .13628 L
.52858 .13474 L
.52312 .13355 L
.51712 .13253 L
.51392 .13211 L
.5105 .13176 L
.50877 .13162 L
.5069 .1315 L
.50512 .13141 L
.50349 .13135 L
.50201 .13132 L
.50039 .1313 L
.49957 .1313 L
.49869 .13131 L
.49711 .13134 L
.49559 .13138 L
.49472 .13142 L
.49391 .13146 L
.49051 .13168 L
.48758 .13195 L
.48443 .13232 L
.47729 .13347 L
.47082 .13488 L
.45908 .13837 L
.44716 .14316 L
.42607 .15493 L
.40886 .16814 L
.39287 .18431 L
.38142 .19965 L
.37276 .21565 L
.3693 .22467 L
.36808 .22867 L
Mistroke
.36701 .23297 L
.36618 .23717 L
.36562 .24101 L
.36541 .24285 L
.36524 .24486 L
.36518 .24578 L
.36512 .24676 L
.36508 .24769 L
.36506 .24852 L
.36503 .2496 L
.36503 .25014 L
.36503 .25073 L
.36503 .25178 L
.36505 .25276 L
.36507 .25335 L
.36509 .25391 L
.36514 .25497 L
.36521 .2561 L
.36531 .25733 L
.36551 .25943 L
.36579 .26165 L
.36644 .26564 L
.36732 .26971 L
.3683 .27344 L
.37117 .28186 L
.37443 .2893 L
.37862 .29714 L
.38996 .31352 L
.40353 .3283 L
.42188 .34351 L
.4412 .35529 L
.45245 .36046 L
.46326 .36434 L
.46955 .36613 L
.4755 .36751 L
.48179 .36864 L
.48514 .36911 L
.48875 .3695 L
.49194 .36976 L
.49372 .36987 L
.49533 .36994 L
.49685 .36999 L
.49848 .37002 L
.50003 .37003 L
.50144 .37003 L
.50225 .37001 L
.50311 .36999 L
.50466 .36994 L
.50561 .3699 L
.5065 .36986 L
Mistroke
.50818 .36975 L
.51157 .36947 L
.51525 .36906 L
.52153 .36809 L
.52813 .36671 L
.53973 .36337 L
.55066 .35915 L
.5619 .35365 L
.58058 .34173 L
.59904 .32578 L
.61317 .30948 L
.6137 .30877 L
Mfstroke
.41678 .25461 m
.41689 .25624 L
.41705 .25797 L
.41746 .26109 L
.41812 .26478 L
.4189 .26813 L
.42115 .27532 L
.42398 .28198 L
.42783 .28902 L
.43205 .29527 L
.43777 .30218 L
.44385 .30819 L
.4561 .3174 L
.46328 .32143 L
.47143 .325 L
.47945 .32756 L
.48411 .32865 L
.48635 .32907 L
.48848 .32941 L
.49051 .32968 L
.49271 .32991 L
.49368 .32999 L
.49471 .33007 L
.49568 .33013 L
.49656 .33017 L
.49759 .33021 L
.4987 .33024 L
.49972 .33025 L
.50069 .33024 L
.50124 .33024 L
.50182 .33022 L
.50286 .33019 L
.50344 .33017 L
.50398 .33015 L
.50522 .33007 L
.50724 .32992 L
.50913 .32972 L
.51337 .32912 L
.51748 .32831 L
.5218 .32722 L
.52947 .32465 L
.53619 .3217 L
.5497 .31345 L
.55652 .30785 L
.56222 .30219 L
.57124 .29051 L
.57531 .28341 L
.57838 .27656 L
.57964 .27307 L
.58081 .26923 L
Mistroke
.58166 .26583 L
.58237 .26217 L
.58266 .2603 L
.58288 .25854 L
.58306 .25682 L
.58318 .25524 L
.58323 .25446 L
.58327 .25373 L
.58329 .25297 L
.58332 .25214 L
.58333 .25129 L
.58333 .25052 L
.58333 .24963 L
.58331 .24879 L
.58328 .24796 L
.58324 .24707 L
.58314 .24546 L
.58296 .24355 L
.58276 .2418 L
.5822 .2382 L
.58141 .23443 L
.57961 .22819 L
.577 .22149 L
.57347 .21451 L
.56962 .20839 L
.56438 .20158 L
.55888 .19571 L
.54676 .1858 L
.53963 .18143 L
.53163 .17756 L
.52422 .17484 L
.52004 .17364 L
.51602 .17271 L
.51225 .17204 L
.51025 .17175 L
.50811 .1715 L
.50632 .17134 L
.50536 .17127 L
.50433 .17121 L
.50339 .17116 L
.50252 .17113 L
.50169 .17111 L
.50082 .1711 L
.49979 .17109 L
.49921 .17109 L
.49867 .1711 L
.49769 .17112 L
.49665 .17116 L
.49572 .17121 L
.49485 .17126 L
Mistroke
.49287 .17141 L
.49074 .17163 L
.48879 .17188 L
.48444 .17262 L
.48077 .17344 L
.47682 .17452 L
.46978 .17698 L
.46197 .18057 L
.45527 .18446 L
.44797 .18968 L
.44181 .19504 L
.4317 .20655 L
.42735 .2131 L
.42353 .22031 L
.42084 .22688 L
.41963 .23056 L
.41868 .23407 L
.41796 .23736 L
.41764 .2391 L
.41735 .24095 L
.41715 .24251 L
.41697 .24423 L
.41684 .24579 L
.41675 .24726 L
.41671 .24811 L
.41669 .24902 L
.41667 .24988 L
.41667 .25066 L
.41667 .25158 L
.41668 .2521 L
.41669 .25258 L
.41672 .25345 L
.41677 .25438 L
.41682 .25527 L
.41689 .25622 L
.41705 .25795 L
.4173 .25999 L
.41758 .26185 L
.41819 .26511 L
.41903 .26864 L
.42103 .275 L
.42407 .28216 L
.42748 .28845 L
.43223 .29551 L
.43729 .30165 L
.44873 .31226 L
.45555 .31706 L
.46325 .32142 L
.47049 .32464 L
.47459 .32612 L
Mistroke
.47854 .32732 L
.48227 .32825 L
.48637 .32907 L
.48819 .32937 L
.49013 .32963 L
.49179 .32982 L
.49362 .32999 L
.4947 .33007 L
.49571 .33013 L
.49668 .33018 L
.49771 .33021 L
.49873 .33024 L
.49968 .33025 L
.50054 .33024 L
.50147 .33023 L
.50249 .33021 L
.50307 .33019 L
.50361 .33016 L
.50479 .3301 L
.50587 .33003 L
.50781 .32986 L
.5099 .32963 L
.51362 .32908 L
.51763 .32828 L
.5248 .32632 L
.53213 .32357 L
.53987 .31978 L
.54755 .315 L
.55408 .30999 L
.56503 .29898 L
.56985 .29261 L
.57417 .28558 L
.57734 .27909 L
.58 .27197 L
.58102 .26846 L
.58181 .26513 L
.58238 .26212 L
.58285 .25885 L
.58306 .25679 L
.58315 .25575 L
.58322 .25461 L
Mfstroke
.46064 .24258 m
.46029 .24443 L
.46005 .24617 L
.45994 .24717 L
.45987 .24808 L
.45984 .24859 L
.45983 .24886 L
.45982 .24914 L
.4598 .24966 L
.45979 .25014 L
.45978 .25063 L
.45979 .25109 L
.4598 .25161 L
.45981 .25211 L
.45983 .25256 L
.45985 .25299 L
.45992 .25395 L
.46003 .25501 L
.46015 .25597 L
.46051 .25812 L
.461 .26026 L
.46155 .26222 L
.46317 .26651 L
.46502 .27014 L
.46737 .27375 L
.46982 .2768 L
.47277 .2798 L
.47884 .2844 L
.48255 .28644 L
.48459 .28736 L
.48682 .28821 L
.48879 .28885 L
.49101 .28943 L
.49305 .28985 L
.49498 .29014 L
.49599 .29025 L
.49693 .29034 L
.49747 .29038 L
.49797 .29041 L
.49825 .29042 L
.49855 .29043 L
.49908 .29045 L
.4996 .29045 L
.49989 .29046 L
.50017 .29046 L
.50046 .29045 L
.50073 .29045 L
.50103 .29044 L
.50133 .29043 L
.50184 .29041 L
Mistroke
.50232 .29039 L
.5034 .29031 L
.5044 .29021 L
.50535 .2901 L
.50711 .28982 L
.50898 .28943 L
.51099 .28891 L
.51475 .28763 L
.51871 .28581 L
.52261 .28346 L
.52597 .28091 L
.52914 .27793 L
.53182 .27483 L
.53592 .26841 L
.53754 .26481 L
.53827 .26278 L
.53885 .26084 L
.5393 .25904 L
.53963 .25737 L
.53991 .25551 L
.54002 .25452 L
.54011 .25349 L
.54015 .25299 L
.54017 .25245 L
.54019 .25193 L
.54021 .25146 L
.54021 .25104 L
.54022 .25057 L
.54021 .25008 L
.5402 .24962 L
.54018 .24909 L
.54016 .24858 L
.5401 .24763 L
.54 .24661 L
.53987 .2455 L
.53957 .24365 L
.53916 .2417 L
.53817 .23825 L
.53677 .2347 L
.53488 .23103 L
.53052 .22492 L
.52779 .22205 L
.52449 .21923 L
.52069 .21664 L
.51699 .21467 L
.5131 .21309 L
.50928 .21198 L
.50715 .21153 L
.50516 .21122 L
.50411 .21109 L
Mistroke
.50296 .21099 L
.50238 .21095 L
.50175 .21092 L
.50145 .21091 L
.50116 .2109 L
.50062 .21088 L
.5001 .21088 L
.49981 .21088 L
.49954 .21088 L
.49904 .21089 L
.49852 .21091 L
.49805 .21093 L
.49761 .21095 L
.49662 .21102 L
.49554 .21113 L
.49457 .21125 L
.49239 .21161 L
.49054 .21202 L
.48856 .21256 L
.48503 .21379 L
.48147 .21544 L
.47774 .21764 L
.4741 .22037 L
.47105 .22321 L
.46608 .22945 L
.46401 .23307 L
.46226 .23707 L
.46108 .24077 L
.46058 .24288 L
.46023 .24482 L
.46 .24664 L
.4599 .24767 L
.45987 .24816 L
.45984 .24862 L
.45982 .24906 L
.4598 .24946 L
.45979 .2499 L
.45979 .25037 L
.45979 .25083 L
.45979 .25133 L
.4598 .25179 L
.45982 .25223 L
.45984 .25275 L
.45987 .2533 L
.45995 .25429 L
.46007 .25534 L
.46023 .2565 L
.4606 .25856 L
.46105 .26046 L
.46214 .26395 L
Mistroke
.46366 .26757 L
.46567 .27123 L
.46829 .27497 L
.47089 .27796 L
.47408 .28095 L
.47724 .28336 L
.48089 .28559 L
.48499 .28752 L
.48705 .28829 L
.489 .28891 L
.49114 .28946 L
.49311 .28986 L
.49417 .29003 L
.49531 .29018 L
.49653 .29031 L
.49713 .29035 L
.49768 .29039 L
.49817 .29041 L
.49869 .29044 L
.49899 .29044 L
.49927 .29045 L
.49981 .29046 L
.50008 .29046 L
.50036 .29046 L
.50088 .29045 L
.50117 .29044 L
.50144 .29043 L
.50205 .2904 L
.50254 .29038 L
.50306 .29034 L
.50398 .29026 L
.50497 .29015 L
.50605 .28999 L
.50822 .2896 L
.51025 .28912 L
.51396 .28793 L
.51781 .28627 L
.52176 .28403 L
.5282 .27888 L
.53101 .27584 L
.53372 .27219 L
.53579 .26867 L
.5374 .26515 L
.53814 .26316 L
.53882 .26095 L
.53932 .25892 L
.53969 .257 L
.53994 .25524 L
.54005 .25425 L
.54012 .25334 L
Mistroke
.54017 .25256 L
.54019 .25213 L
.5402 .25172 L
.54021 .25128 L
.54021 .25089 L
.54021 .25043 L
.54021 .25 L
.5402 .24957 L
.54018 .2491 L
.54016 .24861 L
.54013 .24814 L
.54007 .24732 L
.53998 .24643 L
.5397 .24441 L
.53936 .24258 L
Mfstroke
.5 .25067 m
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
.5 .25067 L
Mistroke
.5 .25067 L
Mfstroke
.53936 .25875 m
.53971 .2569 L
.53995 .25517 L
.54006 .25417 L
.54013 .25325 L
.54016 .25275 L
.54017 .25248 L
.54018 .2522 L
.5402 .25168 L
.54021 .2512 L
.54022 .2507 L
.54021 .25025 L
.5402 .24973 L
.54019 .24923 L
.54017 .24877 L
.54015 .24835 L
.54008 .24739 L
.53997 .24632 L
.53985 .24536 L
.53949 .24322 L
.539 .24107 L
.53845 .23912 L
.53683 .23483 L
.53498 .2312 L
.53263 .22758 L
.53018 .22453 L
.52723 .22154 L
.52116 .21694 L
.51745 .21489 L
.51541 .21397 L
.51318 .21312 L
.51121 .21249 L
.50899 .21191 L
.50695 .21149 L
.50502 .2112 L
.50401 .21108 L
.50307 .211 L
.50253 .21096 L
.50203 .21093 L
.50175 .21092 L
.50145 .21091 L
.50092 .21089 L
.5004 .21088 L
.50011 .21088 L
.49983 .21088 L
.49954 .21088 L
.49927 .21089 L
.49897 .21089 L
.49867 .2109 L
.49816 .21092 L
Mistroke
.49768 .21095 L
.4966 .21102 L
.4956 .21112 L
.49465 .21124 L
.49289 .21152 L
.49102 .21191 L
.48901 .21242 L
.48525 .21371 L
.48129 .21553 L
.47739 .21787 L
.47403 .22043 L
.47086 .2234 L
.46818 .2265 L
.46408 .23293 L
.46246 .23653 L
.46173 .23856 L
.46115 .2405 L
.4607 .2423 L
.46037 .24397 L
.46009 .24583 L
.45998 .24681 L
.45989 .24785 L
.45985 .24835 L
.45983 .24889 L
.45981 .2494 L
.45979 .24987 L
.45979 .2503 L
.45978 .25077 L
.45979 .25126 L
.4598 .25172 L
.45982 .25225 L
.45984 .25276 L
.4599 .25371 L
.46 .25473 L
.46013 .25583 L
.46043 .25768 L
.46084 .25963 L
.46183 .26308 L
.46323 .26663 L
.46512 .27031 L
.46948 .27642 L
.47221 .27928 L
.47551 .2821 L
.47931 .28469 L
.48301 .28666 L
.4869 .28824 L
.49072 .28936 L
.49285 .28981 L
.49484 .29012 L
.49589 .29024 L
Mistroke
.49704 .29035 L
.49762 .29039 L
.49825 .29042 L
.49855 .29043 L
.49884 .29044 L
.49938 .29045 L
.4999 .29046 L
.50019 .29046 L
.50046 .29045 L
.50096 .29045 L
.50148 .29043 L
.50195 .29041 L
.50239 .29039 L
.50338 .29031 L
.50446 .29021 L
.50543 .29008 L
.50761 .28972 L
.50946 .28932 L
.51144 .28878 L
.51497 .28754 L
.51853 .2859 L
.52226 .2837 L
.5259 .28097 L
.52895 .27813 L
.53392 .27188 L
.53599 .26827 L
.53774 .26427 L
.53892 .26057 L
.53942 .25846 L
.53977 .25652 L
.54 .2547 L
.5401 .25367 L
.54013 .25318 L
.54016 .25272 L
.54018 .25228 L
.5402 .25188 L
.54021 .25144 L
.54021 .25096 L
.54021 .25051 L
.54021 .25001 L
.5402 .24954 L
.54018 .24911 L
.54016 .24859 L
.54013 .24803 L
.54005 .24705 L
.53993 .24599 L
.53977 .24484 L
.5394 .24278 L
.53895 .24088 L
.53786 .23739 L
Mistroke
.53634 .23377 L
.53433 .2301 L
.53171 .22636 L
.52911 .22338 L
.52592 .22038 L
.52276 .21798 L
.51911 .21574 L
.51501 .21381 L
.51295 .21304 L
.511 .21243 L
.50886 .21188 L
.50689 .21148 L
.50583 .21131 L
.50469 .21116 L
.50347 .21103 L
.50287 .21098 L
.50232 .21095 L
.50183 .21092 L
.50131 .2109 L
.50101 .21089 L
.50073 .21089 L
.50019 .21088 L
.49992 .21088 L
.49964 .21088 L
.49912 .21089 L
.49883 .2109 L
.49856 .21091 L
.49795 .21093 L
.49746 .21096 L
.49694 .211 L
.49602 .21108 L
.49503 .21119 L
.49395 .21134 L
.49178 .21174 L
.48975 .21222 L
.48604 .2134 L
.48219 .21507 L
.47824 .21731 L
.4718 .22245 L
.46899 .2255 L
.46628 .22914 L
.46421 .23267 L
.4626 .23618 L
.46186 .23818 L
.46118 .24039 L
.46068 .24242 L
.46031 .24434 L
.46006 .24609 L
.45995 .24709 L
.45988 .248 L
Mistroke
.45983 .24877 L
.45981 .24921 L
.4598 .24962 L
.45979 .25005 L
.45979 .25045 L
.45979 .2509 L
.45979 .25133 L
.4598 .25176 L
.45982 .25224 L
.45984 .25273 L
.45987 .25319 L
.45993 .25402 L
.46002 .25491 L
.4603 .25692 L
.46064 .25875 L
Mfstroke
.58322 .24672 m
.58311 .2451 L
.58295 .24337 L
.58254 .24025 L
.58188 .23655 L
.5811 .23321 L
.57885 .22602 L
.57602 .21936 L
.57217 .21231 L
.56795 .20606 L
.56223 .19915 L
.55615 .19315 L
.5439 .18393 L
.53672 .1799 L
.52857 .17634 L
.52055 .17377 L
.51589 .17269 L
.51365 .17227 L
.51152 .17193 L
.50949 .17166 L
.50729 .17143 L
.50632 .17134 L
.50529 .17127 L
.50432 .17121 L
.50344 .17117 L
.50241 .17113 L
.5013 .1711 L
.50028 .17109 L
.49931 .17109 L
.49876 .1711 L
.49818 .17111 L
.49714 .17114 L
.49656 .17117 L
.49602 .17119 L
.49478 .17126 L
.49276 .17142 L
.49087 .17162 L
.48663 .17222 L
.48252 .17303 L
.4782 .17412 L
.47053 .17668 L
.46381 .17964 L
.4503 .18789 L
.44348 .19349 L
.43778 .19915 L
.42876 .21082 L
.42469 .21792 L
.42162 .22478 L
.42036 .22827 L
.41919 .2321 L
Mistroke
.41834 .23551 L
.41763 .23916 L
.41734 .24103 L
.41712 .2428 L
.41694 .24452 L
.41682 .2461 L
.41677 .24688 L
.41673 .24761 L
.41671 .24837 L
.41668 .2492 L
.41667 .25004 L
.41667 .25082 L
.41667 .25171 L
.41669 .25255 L
.41672 .25338 L
.41676 .25427 L
.41686 .25588 L
.41704 .25779 L
.41724 .25954 L
.4178 .26314 L
.41859 .26691 L
.42039 .27315 L
.423 .27985 L
.42653 .28683 L
.43038 .29295 L
.43562 .29975 L
.44112 .30563 L
.45324 .31554 L
.46037 .31991 L
.46837 .32378 L
.47578 .3265 L
.47996 .32769 L
.48398 .32862 L
.48775 .3293 L
.48975 .32958 L
.49189 .32983 L
.49368 .32999 L
.49464 .33007 L
.49567 .33013 L
.49661 .33017 L
.49748 .33021 L
.49831 .33023 L
.49918 .33024 L
.50021 .33025 L
.50079 .33024 L
.50133 .33023 L
.50231 .33021 L
.50335 .33018 L
.50428 .33013 L
.50515 .33008 L
Mistroke
.50713 .32993 L
.50926 .32971 L
.51121 .32945 L
.51556 .32872 L
.51923 .3279 L
.52318 .32682 L
.53022 .32436 L
.53803 .32077 L
.54473 .31688 L
.55203 .31166 L
.55819 .3063 L
.5683 .29479 L
.57265 .28823 L
.57647 .28103 L
.57916 .27445 L
.58037 .27077 L
.58132 .26726 L
.58204 .26398 L
.58236 .26224 L
.58265 .26039 L
.58285 .25883 L
.58303 .25711 L
.58316 .25554 L
.58325 .25407 L
.58329 .25322 L
.58331 .25231 L
.58333 .25145 L
.58333 .25067 L
.58333 .24975 L
.58332 .24924 L
.58331 .24876 L
.58328 .24788 L
.58323 .24695 L
.58318 .24607 L
.58311 .24511 L
.58295 .24339 L
.5827 .24135 L
.58242 .23948 L
.58181 .23622 L
.58097 .23269 L
.57897 .22633 L
.57593 .21918 L
.57252 .21288 L
.56777 .20583 L
.56271 .19968 L
.55127 .18908 L
.54445 .18428 L
.53675 .17991 L
.52951 .1767 L
.52541 .17522 L
Mistroke
.52146 .17402 L
.51773 .17308 L
.51363 .17226 L
.51181 .17197 L
.50987 .1717 L
.50821 .17151 L
.50638 .17135 L
.5053 .17127 L
.50429 .17121 L
.50332 .17116 L
.50229 .17112 L
.50127 .1711 L
.50032 .17109 L
.49946 .17109 L
.49853 .1711 L
.49751 .17113 L
.49693 .17115 L
.49639 .17117 L
.49521 .17123 L
.49413 .17131 L
.49219 .17148 L
.4901 .17171 L
.48638 .17226 L
.48237 .17306 L
.4752 .17502 L
.46787 .17777 L
.46013 .18156 L
.45245 .18634 L
.44592 .19135 L
.43497 .20236 L
.43015 .20872 L
.42583 .21576 L
.42266 .22225 L
.42 .22937 L
.41898 .23287 L
.41819 .23621 L
.41762 .23921 L
.41715 .24249 L
.41694 .24454 L
.41685 .24558 L
.41678 .24672 L
Mfstroke
.6137 .19256 m
.6007 .17725 L
.58381 .16207 L
.56364 .14864 L
.55104 .14236 L
.53868 .13761 L
.53266 .13579 L
.52602 .13415 L
.51967 .13293 L
.51663 .13246 L
.51376 .1321 L
.51097 .13181 L
.50841 .1316 L
.50705 .13151 L
.50556 .13143 L
.50399 .13137 L
.50249 .13133 L
.50091 .13131 L
.49948 .1313 L
.49783 .13132 L
.49628 .13136 L
.49484 .13141 L
.49351 .13148 L
.49052 .13168 L
.48745 .13196 L
.48466 .13229 L
.47843 .13326 L
.4727 .13443 L
.46734 .13579 L
.45559 .13964 L
.44333 .14498 L
.42262 .1573 L
.4037 .17287 L
.38938 .18854 L
.37891 .20371 L
.37449 .21192 L
.37075 .22056 L
.36829 .22793 L
.36728 .23179 L
.36639 .23597 L
.366 .23825 L
.3657 .24036 L
.36547 .24233 L
.36527 .24449 L
.3652 .24543 L
.36514 .24644 L
.36509 .24739 L
.36506 .24825 L
.36504 .24923 L
.36503 .2503 L
Mistroke
.36503 .2513 L
.36504 .25224 L
.36507 .25333 L
.36509 .25394 L
.36512 .25451 L
.36525 .25665 L
.36542 .2586 L
.36566 .26065 L
.36631 .26493 L
.3673 .26962 L
.36979 .27814 L
.37286 .28592 L
.38253 .30339 L
.3947 .31913 L
.41143 .33541 L
.43054 .34927 L
.44122 .3553 L
.45366 .36095 L
.46561 .36505 L
.47142 .3666 L
.47688 .36779 L
.48288 .36881 L
.48608 .36922 L
.4895 .36957 L
.49123 .36971 L
.4931 .36984 L
.49488 .36992 L
.49651 .36998 L
.49799 .37002 L
.49961 .37003 L
.50043 .37003 L
.50131 .37003 L
.50289 .37 L
.50441 .36995 L
.50528 .36992 L
.50609 .36988 L
.50949 .36966 L
.51242 .36939 L
.51557 .36902 L
.52271 .36787 L
.52918 .36645 L
.54092 .36296 L
.55284 .35817 L
.57393 .3464 L
.59114 .3332 L
.60713 .31703 L
.61858 .30169 L
.62724 .28569 L
.6307 .27667 L
.63192 .27267 L
Mistroke
.63299 .26837 L
.63382 .26417 L
.63438 .26033 L
.63459 .25848 L
.63476 .25647 L
.63482 .25555 L
.63488 .25457 L
.63492 .25365 L
.63494 .25281 L
.63497 .25174 L
.63497 .2512 L
.63497 .25061 L
.63497 .24956 L
.63495 .24858 L
.63493 .24799 L
.63491 .24742 L
.63486 .24637 L
.63479 .24524 L
.63469 .244 L
.63449 .24191 L
.63421 .23969 L
.63356 .2357 L
.63268 .23163 L
.6317 .2279 L
.62883 .21947 L
.62557 .21203 L
.62138 .2042 L
.61004 .18781 L
.59647 .17304 L
.57812 .15782 L
.5588 .14605 L
.54755 .14088 L
.53674 .13699 L
.53045 .1352 L
.5245 .13382 L
.51821 .13269 L
.51486 .13223 L
.51125 .13183 L
.50806 .13157 L
.50628 .13147 L
.50467 .13139 L
.50315 .13134 L
.50152 .13131 L
.49997 .1313 L
.49856 .13131 L
.49775 .13132 L
.49689 .13134 L
.49534 .13139 L
.49439 .13143 L
.4935 .13148 L
Mistroke
.49182 .13158 L
.48843 .13186 L
.48475 .13228 L
.47847 .13325 L
.47187 .13463 L
.46027 .13796 L
.44934 .14219 L
.4381 .14768 L
.41942 .15961 L
.40096 .17556 L
.38683 .19186 L
.3863 .19256 L
Mfstroke
.25214 .25281 m
.25322 .25389 L
.25501 .25568 L
.25759 .25825 L
.26131 .26197 L
.2673 .26793 L
.27602 .27657 L
.29008 .29032 L
.31054 .30976 L
.3393 .33536 L
.35814 .35068 L
.38263 .36847 L
.40985 .38497 L
.43768 .39778 L
.45334 .40303 L
.46071 .405 L
.46866 .40675 L
.47312 .40756 L
.47719 .40819 L
.48133 .40873 L
.48516 .40913 L
.48965 .40949 L
.49202 .40962 L
.49322 .40968 L
.49452 .40973 L
.49669 .40979 L
.49873 .40982 L
.49997 .40982 L
.50114 .40982 L
.50335 .40979 L
.50525 .40974 L
.5073 .40966 L
.50924 .40956 L
.51101 .40944 L
.51512 .40911 L
.51943 .40864 L
.52753 .40745 L
.5363 .4057 L
.55164 .40152 L
.56547 .39654 L
.59532 .38212 L
.61952 .36702 L
.66146 .33471 L
.69003 .30923 L
.71071 .28957 L
.72394 .27661 L
.73247 .26816 L
.73852 .26214 L
.74236 .2583 L
.74505 .25562 L
Mistroke
.74674 .25393 L
.74782 .25285 L
.74786 .25281 L
Mfstroke
0 .44961 m
.00239 .4496 L
.00498 .44955 L
.00738 .44948 L
.00996 .44936 L
.01281 .4492 L
.01855 .44875 L
.02422 .44814 L
.0294 .44745 L
.04095 .44545 L
.05154 .44306 L
.06147 .44036 L
.09961 .42641 L
.13211 .41124 L
.16235 .39622 L
.19195 .38285 L
.20713 .37733 L
.2136 .37537 L
.22058 .37357 L
.22666 .37228 L
.23312 .37122 L
.23666 .37078 L
.23988 .37046 L
.24143 .37034 L
.24311 .37023 L
.2447 .37015 L
.24616 .3701 L
.24764 .37006 L
.2492 .37004 L
.25053 .37004 L
.25199 .37005 L
.25356 .37009 L
.25444 .37012 L
.25524 .37015 L
.25681 .37023 L
.25827 .37032 L
.26108 .37055 L
.26413 .37087 L
.26969 .37164 L
.27578 .37276 L
.28255 .37433 L
.29653 .37854 L
.32376 .38963 L
.35592 .40528 L
.38975 .42169 L
.40896 .42999 L
.43033 .43782 L
.44993 .44342 L
.46133 .44589 L
.46659 .44683 L
Mistroke
.47215 .44767 L
.4769 .44828 L
.48212 .44881 L
.48687 .44918 L
.48899 .44931 L
.4913 .44942 L
.494 .44952 L
.49655 .44958 L
.49903 .44961 L
.50132 .44961 L
.5038 .44958 L
.50653 .4495 L
.5079 .44946 L
.50937 .44939 L
.51204 .44925 L
.51733 .44886 L
.52295 .44829 L
.53302 .44689 L
.54488 .44462 L
.5555 .44204 L
.59454 .42856 L
.66254 .39613 L
.68939 .38389 L
.70369 .37847 L
.71655 .37457 L
.72294 .37303 L
.72893 .37187 L
.73222 .37135 L
.73578 .37088 L
.739 .37054 L
.74197 .3703 L
.74349 .37021 L
.74512 .37013 L
.74666 .37008 L
.74807 .37005 L
.74963 .37004 L
.75052 .37004 L
.75134 .37004 L
.75221 .37005 L
.75301 .37007 L
.75392 .3701 L
.75479 .37013 L
.75644 .37021 L
.75819 .37031 L
.76133 .37057 L
.76472 .37094 L
.76846 .37145 L
.77511 .37262 L
.78148 .37406 L
.79368 .37759 L
Mistroke
.82272 .38915 L
.88415 .41909 L
.90415 .42801 L
.92386 .43563 L
.94222 .44141 L
.95188 .44388 L
.96259 .44613 L
.96771 .44701 L
.97321 .44782 L
.97798 .4484 L
.98317 .4489 L
.98623 .44914 L
.98905 .44931 L
.99175 .44944 L
.99313 .44949 L
.99462 .44954 L
.99718 .44959 L
.99959 .44961 L
Mfstroke
.99959 .44961 m
1 .44961 L
s
0 .4894 m
.00223 .48939 L
.00533 .48934 L
.00707 .4893 L
.00866 .48924 L
.01159 .48912 L
.01468 .48895 L
.02122 .48846 L
.02835 .48773 L
.04146 .48585 L
.05531 .48317 L
.07977 .47682 L
.12554 .46101 L
.17125 .44399 L
.19147 .43758 L
.20225 .43472 L
.21218 .43251 L
.22127 .43089 L
.2295 .42978 L
.23418 .42931 L
.2386 .42897 L
.24101 .42884 L
.24325 .42874 L
.24571 .42866 L
.24695 .42864 L
.24832 .42862 L
.25058 .42861 L
.25269 .42863 L
.25501 .42868 L
.25631 .42872 L
.25749 .42877 L
.25987 .42888 L
.26246 .42904 L
.26717 .42943 L
.27156 .4299 L
.27622 .43051 L
.28463 .4319 L
.30373 .43625 L
.32538 .44285 L
.36961 .45916 L
.41336 .47472 L
.43607 .48115 L
.44831 .48394 L
.46149 .48634 L
.46806 .48728 L
.47426 .48802 L
.47985 .48855 L
.48274 .48878 L
.48591 .48899 L
.48869 .48913 L
Mistroke
.49168 .48926 L
.49483 .48934 L
.49778 .48939 L
.50071 .4894 L
.50349 .48938 L
.50599 .48933 L
.5087 .48924 L
.51175 .48911 L
.51455 .48896 L
.52084 .48849 L
.52638 .48795 L
.53219 .48725 L
.54446 .48533 L
.55774 .48262 L
.58388 .47558 L
.60738 .46772 L
.64858 .45224 L
.68998 .43801 L
.70052 .43515 L
.71151 .43265 L
.72079 .43096 L
.72609 .4302 L
.73091 .42963 L
.73574 .42918 L
.74032 .42887 L
.74283 .42875 L
.74515 .42867 L
.74648 .42864 L
.74769 .42862 L
.74909 .42861 L
.7504 .42861 L
.75262 .42863 L
.75496 .42868 L
.75696 .42875 L
.75916 .42884 L
.76152 .42898 L
.76406 .42916 L
.76864 .42958 L
.77348 .43014 L
.77864 .43087 L
.78796 .43254 L
.80941 .43784 L
.85039 .45185 L
.89379 .46813 L
.91383 .47487 L
.93636 .48122 L
.94892 .48406 L
.96071 .48621 L
.97207 .48778 L
.97845 .48843 L
Mistroke
.98425 .48888 L
.98752 .48907 L
.98931 .48916 L
.99098 .48923 L
.99393 .48932 L
.99718 .48938 L
.99899 .4894 L
Mfstroke
.99899 .4894 m
1 .4894 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{406.375, 203.688},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40009Q0001<B000`400?l00000o`00003oo`3ooooo
0?ooof<0oooo003o0?ooool0ooooH`3oool00?l0ooooo`3ooomS0?ooo`00o`3oool`0?ooo`030?oo
g`2LB00000000080000000<0M;oo0?ooo`3oool0o`3oool[0?ooo`00o`3ooolb0?ooo`030?noM000
07@0_ooo0?l0oooo;P3oool00?l0oooo<P3oool00`3ookl0M00007Boo`3o0?ooobh0oooo003o0?oo
oc80oooo00<0oooO09a80018W=l0o`3oool^0?ooo`00o`3ooolc0?ooo`030=nLB000000000000080
000000<0M;oo0?ooo`3oool00P3oool01@3o_g@000000000000004P0W=oo0080oooo00<0oono07@0
000000000P0000003@18W=l0oooo0?ooo`3oool0_g@00000000004P0W=oo0?ooo`3o_g@000000000
B02Lgol00P3oool0102oM00000000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3o
ool00`3OW4P004RL0=ooo`020?ooo`030;md0000000007Bo0?@0oooo003o0?oooc<0oooo00L0okmd
0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00l0omnL04P0B02Lgol0oooo0?oOW018000007Bo
0?ooo`3ooml0W4P007Boo`3oool0omnL04P0001d_ol00P3oool05`3OW4P004RL0=oog`2LM9`0gooo
0?noM00004P0W=oO09adW03Oool0oooo0=nLB000B9`0gooO09adW03Oool0omnL04P0B02Lgol0oooo
0?oOW018000007Bo0080oooo00<0okmd0000M02oool00P3oool0103OW4P004RL0=noM01d_ooc0?oo
o`00o`3ooolc0?ooo`030?oo_`1d0000M;oo0080oooo00L0oooO09a80018W=l0oooo0=nLB000B9`0
gooo0080oooo00X0omnL04P00018W=l0oooo0;mdM02oool0oooo0?oo_`1d0000M;oo0P3oool00`3o
_g@0001809cOo`020?ooo`030?oo_`1d0000M;oo00<0oooo00<0okmd0000B02Lgol00P3oool00`3O
W4P004RL0=ooo`020?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00004RLg`020?ooo`030?oOW018
04P0W=oo0?@0oooo003o0?oooc<0oooo00<0oooO09a80018W=l00`3oool01P3ogi`0B01809cOo`3o
_g@0001d0;ooo`<0oooo00<0gia80000M02oool00`3oool00`3OW4P0001d0;ooo`020?ooo`030?oo
_`1d0000B9cO00<0oooo00<0_g@0001d_`3oool00P3oool00`3ookl0M00004RLg`020?ooo`030?no
M00007@0_ooo00<0oooo00/0gia80000M02oool0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RL
g`3d0?ooo`00o`3ooold0?ooo`030=nLB00007@0_ooo0080oooo00H0oooO09a80018W=l0oono07@0
001d_ol30?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0000B02Lgol0103oool00`2oM00004RL
0=ooo`020?ooo`030=nLB000B9`0gooo00<0oooo00L0_g@00018W03Oool0oooo0?oo_`1d0000M;oo
00<0oooo00L0omnL04P0B02Lgol0oooo0=nLB000B4P0B7Bo0080oooo00<0_g@0001d_`3oool0l`3o
ool00?l0oooo=03oool0103o_g@000000018W03Oool20?ooo`060=nLB000M;l0oooo0=nLB00007@0
_ooo0P3oool02P3ookl0M00007Boo`3oool0omnL04P0B02Lgol0oooo0=nLM02Lgol20?ooo`030=nL
B00004P0W=oo0080oooo00<0omnL04P0B02Lgol00`3oool00`3OW4P0001809cOo`020?ooo`030=nL
B00007@0_ooo0080oooo00`0oono07@0001d_ol0oooo0?noM00007@0_kmd04RLg`3oool0okmd0000
M02ooooc0?ooo`00o`3ooold0?ooo`080?oo_`1d0000M9bL04P0M02oool0oooO09a80000M;l20?oo
o`0>0;md000007@0_ooo0?ooo`3o_g@0001d0;ooo`3oool0omnL04P0B02Lgol0oooO09a8001d_ol2
0?ooo`030?oOW0180000B9cO0080oooo00<0oooO09a80018W=l00`3oool00`3ogi`0B00004RLg`03
0?ooo`0@0;md000007@0_ooo0?ooo`3o_g@0001d0;ooo`3oool0oooO09a8001d_ol0gia8001d_`3o
gi`0B01809cOoo<0oooo003o0?oooc<0oooo00T0omnL04P000000000B9cO0?oo_`1d000000000000
M02oool00`3oool01@3o_g@00000000000000000M;oo00<0oooo00L0_g@0000000000000000004RL
g`3oool0okmd0080000000@004RL0=ooo`3ookl0M0000`0000001P1d_ol0okmd00000000000004RL
0=ooo`<0oooo00l0okmd000000000000000007Boo`3oool0oono07@000000000B9cO0?ooo`3o_g@0
00000000B02Lgol0l`3oool00?l0ooooE03oool00`3o_g@0001d0;ooo`3o0?ooo``0oooo003o0?oo
oeD0oooo00<0giad09cOo`3oool0o`3oool;0?ooo`00o`3ooom?0?ooo`040?oo_`1d000004RL0=oo
o`L0oooo00@0oono07@00000B9`0goooo`3oool50?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0
ooooH`3oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0ooooH`3oool0
0?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0ooooH`3oool00?l0ooooo`3o
oomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0ooooH`3oool00?l0ooooo`3ooomS0?ooo`00
O03oool01@3ooml0W4P000000000000007Bo04@0oooo00<0oooO09a80018W=l0@@3oool00`3OW4P0
0000000000030000000304RLg`3oool0oooo03T0oooo00@0_g@0000000000000M;oo@@3oool00`3O
W4P00000000000030000000304RLg`3oool0oooo0440oooo00<0oooO09a80018W=l0@@3oool01@3o
oml0W4P000000000000007Bo03h0oooo001l0?ooo`070=nLB000M;l0oooo0?oo_`1d000004RL0=oo
o`130?ooo`030=nLB000B9`0gooo0440oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?oo
o`0h0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO@@3oool00`3o_g@007Bo0?ooo`020?ooo`03
0;mdM02oool0oooo0480oooo00<0gia80018W03Oool0@03oool01`3OW4P007Bo0?ooo`3ookl0M000
0018W03Oool0?03oool007/0oooo00<0oooO09a8001d_ol00P3oool00`3ogi`0B00004RLg`130?oo
o`030?oOW01804P0W=oo0480oooo00<0okmd001d_`3oool0>`3oool00`3ooml0W4P004RLg`020?oo
o`030?noM00004P0W=oo0440oooo00<0okmd001d_`3oool0AP3oool00`3ogi`0B01809cOo`0o0?oo
o`030?oog`2LB000M;oo0080oooo00<0omnL04P00018W=l0?03oool007/0oooo00<0oono07@0001d
_ol00`3oool00`2oM000001d0;ooo`0n0?ooo`030?noM0000000000000@0000000<004RL0=ooo`3o
ool0@03oool00`3o_g@007Bo0?ooo`0j0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000
B9`0gooo@@3oool00`3o_g@007Bo0?ooo`110?ooo`030?noM0000000000000@0000000<004RL0=oo
o`3oool0?03oool00`3ookl0M00007Boo`030?ooo`030;md000007@0_ooo03/0oooo001^0?ooo`03
0?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0
W=oo0340oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ooml0W4Q809cOo`80oooo
00<0_g@0001d_`3oool0<P3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo
00<0okmd04RLg`3oool0>@3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo0480oooo
00<0okmd04RLg`3oool0@03oool00`3ooml0W4Q809cOo`020?ooo`030;md0000M;l0oooo03h0oooo
00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0k0?ooo`00O03oool00`2oM000001d0;oo
o`020?ooo`030=nLB00004P0W=oo03l0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo04@0
oooo00<0gia804RLg`3oool0>@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO04<0
oooo00<0gia804RLg`3oool0@03oool01`3ookl0M01d0;ooo`3oool0okmd0018W03Oool0?`3oool0
0`2oM000001d0;ooo`020?ooo`030=nLB00004P0W=oo03/0oooo001l0?ooo`080?noM0000000B9cO
0?ooo`3ooml0W4P00018W03Ooom00?ooo`060?oOW018B9`0gooo0?oOW01804P0W=ooA@3oool00`2o
M000M;oo0?ooo`0h0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0A03oool00`2o
M000M;oo0?ooo`100?ooo`060?oOW018B9`0gooo0?oOW01804P0W=oo?`3oool0203o_g@0000004RL
g`3oool0oooO09a80000B9`0gooo>`3oool007`0oooo00<0oooO09a8000000000P0000000`0007@0
_ooo0?ooo`110?ooo`050?noM000M;l0oooO09a80018W=l0A@3oool00`3ogi`0B01d0;ooo`0h0?oo
o`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool0@`3oool00`3ogi`0B01d0;ooo`110?oo
o`050?noM000M;l0oooO09a80018W=l0?`3oool00`3ooml0W4P000000002000000030000M02oool0
oooo03/0oooo001m0?ooo`030?oo_`1d000007Bo04H0oooo00@0gia804RLg`2oM00007BoA@3oool0
0`3ooml0W4P007Boo`0h0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0@`3o
ool00`3ooml0W4P007Boo`120?ooo`040=nLB018W=l0_g@0001d_d00oooo00<0oono07@00000M;l0
?`3oool007h0oooo00<0omnL04P00018W=l0AP3oool0103OW4P0M7Ad0000M02ooom40?ooo`030?oo
_`1d0000B9cO03T0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`130?ooo`030?oo
_`1d0000B9cO04<0oooo00@0gia807AdM00007@0_ooo@03oool00`3ogi`0B00004RLg`0n0?ooo`00
O`3oool0103ooml0W4P00000M02ooom50?ooo`030;md000004P0W=oo0400oooo00L0okmd0018W03O
ool0oooo0=nLB00004P0W=oo03X0oooo00H0gia80000M02oool0oooO09a80018W=m00?ooo`070?no
M000B9`0gooo0?ooo`3OW4P0001809cOo`140?ooo`030;md000004P0W=oo0440oooo00@0oooO09a8
000007@0_ooo?03oool00840oooo00@0oono07@000000000B9cO@`3oool00`3ooml0W4P004RLg`11
0?ooo`050?oOW018000000000000001d_ol0?03oool0103OW4P0000000000000M;m20?ooo`050?oO
W018000000000000001d_ol0A@3oool00`3ooml0W4P004RLg`130?ooo`040?oo_`1d0000000004RL
gcX0oooo003o0?ooool0ooooH`3oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o
0?ooool0ooooH`3oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0oooo
H`3oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0ooooH`3oool00?l0
ooooo`3ooomS0?ooo`00GP3oool00`3ookl0M000000000020000000307Boo`3oool0oooo00T0oooo
o`00002o000003@0oooo001N0?ooo`030?noM000000007Bo0080oooo00<0_gA809cOo`3oool02@3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0<P3o
ool006<0oooo00<0omnL04Q8W03Oool02@3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo04@0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`140?ooo`030000003oool0
oooo04<0oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`140?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0<P3oool006<0oooo00<0oono07@0001d_ol02@3oool00`00
0000oooo0?ooo`3L0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool0<P3oool00540
oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`050?ooo`030?oOW0180000M;oo00T0
oooo00<000000?ooo`3oool0g03oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo0380
oooo001R0?ooo`040?oOW0180000001809cOo`T0oooo300000390?oooa@00000b@3oool<000003@0
oooo001P0?ooo`050?oOW018000000000000001d_ol02P3oool00`000000oooo0?ooo`090?ooo`T0
0000^@3oool7000000X0oooo00<000000?ooo`3oool01`3oool:00000;H0oooo2@00000;0?ooo`03
0000003oool0oooo0380oooo001R0?ooo`030?oo_`1d000007Bo00X0oooo00<000000?ooo`3oool0
4P3oool400000;00oooo1@00000A0?ooo`030000003oool0oooo0140oooo1@00002^0?ooo`<00000
503oool00`000000oooo0?ooo`0b0?ooo`00I03oool00`3OW4P007Bo0?ooo`080?ooo`030000003o
ool0oooo01H0oooo0`00002X0?ooo`D000005P3oool00`000000oooo0?ooo`0F0?ooo`800000Z@3o
ool3000001L0oooo00<000000?ooo`3oool0<P3oool006@0oooo00<0omnL04P0B02Lgol0203oool0
0`000000oooo0?ooo`0I0?ooo`@00000X03oool4000001/0oooo00<000000?ooo`3oool0603oool4
00000:40oooo1000000J0?ooo`030000003oool0oooo0380oooo001Q0?ooo`060;md001d_ol0oooo
0=nLB00004P0W=oo203oool00`000000oooo0?ooo`0M0?ooo`D00000U`3oool4000001l0oooo00<0
00000?ooo`3oool0703oool5000009P0oooo1000000N0?ooo`030000003oool0oooo0380oooo001R
0?ooo`040;md00000000000004RLg`T0oooo00<000000?ooo`3oool08P3oool4000008l0oooo1000
000S0?ooo`030000003oool0oooo0240oooo1000002A0?ooo`<000008P3oool00`000000oooo0?oo
o`0b0?ooo`00K`3oool3000002H0oooo0`00002:0?ooo`8000009`3oool00`000000oooo0?ooo`0U
0?ooo`<00000R`3oool3000002<0oooo0`00000d0?ooo`00K`3oool00`000000oooo0?ooo`0Y0?oo
o`<00000Q03oool3000002T0oooo00<000000?ooo`3oool0:03oool3000008D0oooo0`00000X0?oo
o`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo02`0oooo0`00001o0?ooo`80
0000;03oool00`000000oooo0?ooo`0[0?ooo`800000P@3oool2000002/0oooo00<000000?ooo`3o
ool0<P3oool006l0oooo00<000000?ooo`3oool0;`3oool2000007X0oooo0`00000^0?ooo`030000
003oool0oooo02d0oooo0`00001k0?ooo`<00000;@3oool00`000000oooo0?ooo`0b0?ooo`00K`3o
ool00`000000oooo0?ooo`0a0?ooo`800000M@3oool300000340oooo00<000000?ooo`3oool0<03o
ool3000007D0oooo0`00000`0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0
oooo03<0oooo0`00001_0?ooo`<00000=03oool00`000000oooo0?ooo`0c0?ooo`<00000K`3oool3
000003<0oooo00<000000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0=P3oool20000
06X0oooo0`00000g0?ooo`030000003oool0oooo03H0oooo0`00001Y0?ooo`<00000=P3oool00`00
0000oooo0?ooo`0b0?ooo`00K`3oool3000003P0oooo0`00001T0?ooo`<00000>P3oool00`000000
oooo0?ooo`0i0?ooo`800000I03oool3000003L0oooo0`00000d0?ooo`00K`3oool00`000000oooo
0?ooo`0k0?ooo`800000G`3oool3000003d0oooo00<000000?ooo`3oool0>`3oool2000005l0oooo
0`00000l0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo03d0oooo0P00
001J0?ooo`<00000@03oool00`000000oooo0?ooo`0m0?ooo`<00000FP3oool2000003l0oooo00<0
00000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0?`3oool3000005@0oooo0`000013
0?ooo`030000003oool0oooo0400oooo0P00001E0?ooo`<00000@@3oool00`000000oooo0?ooo`0b
0?ooo`00K`3oool>000003L0oooo0`00001?0?ooo`800000>P3ooolI000003P0oooo0`00001?0?oo
o`<00000>@3oool<000003@0oooo001_0?ooo`030000003oool0oooo00/0oooo1P00000d0?ooo`<0
0000B@3oool3000003L0oooo1@00000<0?ooo`030000003oool0oooo00X0oooo1`00000d0?ooo`80
0000BP3oool3000003@0oooo2000000;0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000
003oool0oooo0140oooo1000000c0?ooo`<00000@`3oool3000003D0oooo1@00000A0?ooo`030000
003oool0oooo0140oooo1000000b0?ooo`800000A@3oool3000003D0oooo0P00000C0?ooo`030000
003oool0oooo0380oooo001_0?ooo`<000005@3oool400000380oooo0`00000n0?ooo`800000=03o
ool4000001H0oooo00<000000?ooo`3oool05@3oool300000340oooo0`00000o0?ooo`<00000=@3o
ool3000001<0oooo0`00000d0?ooo`00K`3oool00`000000oooo0?ooo`0I0?ooo`800000<`3oool3
000003P0oooo0`00000c0?ooo`<000006P3oool00`000000oooo0?ooo`0H0?ooo`<00000<@3oool3
000003T0oooo0`00000d0?ooo`@00000603oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`00
0000oooo0?ooo`0K0?ooo`<00000<`3oool400000340oooo0`00000c0?ooo`<000007@3oool00`00
0000oooo0?ooo`0K0?ooo`<00000<@3oool400000340oooo1000000d0?ooo`<00000703oool00`00
0000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`0N0?ooo`<00000=03oool5000002L0
oooo1@00000d0?ooo`800000803oool00`000000oooo0?ooo`0N0?ooo`<00000<P3oool5000002P0
oooo1000000e0?ooo`<000007`3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo
0?ooo`0Q0?ooo`<00000=P3oool4000001l0oooo1000000f0?ooo`<000008P3oool00`000000oooo
0?ooo`0Q0?ooo`<00000=03oool6000001h0oooo1000000g0?ooo`8000008P3oool00`000000oooo
0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`0T0?ooo`<00000=`3oool500000180oooo2000
000g0?ooo`<000009@3oool00`000000oooo0?ooo`0T0?ooo`<00000=`3oool500000180oooo1`00
000h0?ooo`<00000903oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`0W
0?ooo`<00000>@3ooolB000003`0oooo0`00000X0?ooo`030000003oool0oooo02L0oooo0P00000j
0?oooa800000?03oool3000002L0oooo00<000000?ooo`3oool0<P3oool006l0oooo0`00000Z0?oo
o`800000P`3oool2000002/0oooo00<000000?ooo`3oool0:@3oool2000008<0oooo0`00000X0?oo
o`<00000=03oool006l0oooo00<000000?ooo`3oool0;03oool2000007l0oooo0P00000]0?ooo`03
0000003oool0oooo02/0oooo0P00001o0?ooo`800000;@3oool00`000000oooo0?ooo`0b0?ooo`00
GP3oool00`3OW4P00000000000030000000304RLg`3oool0oooo00P0oooo00<000000?ooo`3oool0
;P3oool2000007/0oooo0P00000_0?ooo`030000003oool0oooo02d0oooo0`00001j0?ooo`800000
;`3oool00`000000oooo0?ooo`0b0?ooo`00G`3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02o
ool0oooo00P0oooo00<000000?ooo`3oool0<03oool3000007H0oooo0P00000a0?ooo`030000003o
ool0oooo0300oooo0P00001f0?ooo`800000<@3oool00`000000oooo0?ooo`0b0?ooo`00H03oool0
0`3o_g@007Bo0?ooo`0<0?ooo`030000003oool0oooo03<0oooo0P00001a0?ooo`<00000<`3oool0
0`000000oooo0?ooo`0b0?ooo`800000LP3oool2000003<0oooo00<000000?ooo`3oool0<P3oool0
0640oooo00<0okmd001d_`3oool02`3oool00`000000oooo0?ooo`0e0?ooo`800000K@3oool20000
03H0oooo00<000000?ooo`3oool0=03oool2000006h0oooo0P00000e0?ooo`030000003oool0oooo
0380oooo001A0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`3o_g@0
B9cO0?ooo`0:0?ooo`030000003oool0oooo03L0oooo0P00001Y0?ooo`800000>03oool00`000000
oooo0?ooo`0f0?ooo`800000JP3oool2000003L0oooo00<000000?ooo`3oool0<P3oool006<0oooo
00<0gia804RLg`3oool02@3oool4000003P0oooo0P00001U0?ooo`800000=@3oool:000003H0oooo
0P00001U0?ooo`<00000=P3oool4000003@0oooo001T0?ooo`030;md001d_ol0oooo00P0oooo00<0
00000?ooo`3oool0>`3oool200000600oooo0`00000^0?ooo`T000001@3oool00`000000oooo0?oo
o`020?ooo`P00000<03oool200000640oooo0P00000l0?ooo`030000003oool0oooo0380oooo001T
0?ooo`030?oOW01807@0_ooo00P0oooo00<000000?ooo`3oool0?@3oool2000005`0oooo0P00000^
0?ooo`<000003P3oool00`000000oooo0?ooo`0:0?ooo`H00000;03oool2000005d0oooo0P00000n
0?ooo`030000003oool0oooo0380oooo001T0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?oo
o`3oool0?`3oool2000005P0oooo0P00000/0?ooo`@000004@3oool00`000000oooo0?ooo`0@0?oo
o`<00000:`3oool2000005T0oooo0P0000100?ooo`030000003oool0oooo0380oooo001T0?ooo`03
0?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0@@3oool2000005@0oooo0P00000[0?ooo`<0
00005@3oool00`000000oooo0?ooo`0C0?ooo`800000:`3oool2000005D0oooo0P0000120?ooo`03
0000003oool0oooo0380oooo001P0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?oo
o`030000003oool0oooo04<0oooo0P00001@0?ooo`800000:P3oool3000001P0oooo00<000000?oo
o`3oool05@3oool2000002/0oooo0P00001A0?ooo`800000A03oool00`000000oooo0?ooo`0b0?oo
o`00H@3oool01@3ogi`0B000000000000000M;oo00T0oooo00<000000?ooo`3oool0A@3oool20000
04`0oooo0P00000Z0?ooo`8000006`3oool00`000000oooo0?ooo`0G0?ooo`<00000:P3oool20000
04`0oooo0`0000160?ooo`030000003oool0oooo0380oooo001_0?ooo`<00000A`3oool2000004P0
oooo0P00000Z0?ooo`8000007@3oool00`000000oooo0?ooo`0J0?ooo`800000:P3oool2000004P0
oooo0P0000170?ooo`<00000=03oool006l0oooo00<000000?ooo`3oool0B@3oool2000004@0oooo
0P00000Z0?ooo`8000007`3oool00`000000oooo0?ooo`0L0?ooo`<00000:@3oool2000004@0oooo
0P00001;0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo04/0oooo0`00
000o0?ooo`800000:P3oool200000240oooo00<000000?ooo`3oool07`3oool3000002P0oooo0`00
000o0?ooo`800000C@3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`1>
0?ooo`800000>`3oool2000002X0oooo0P00000S0?ooo`030000003oool0oooo0280oooo0P00000Y
0?ooo`800000>`3oool2000004l0oooo00<000000?ooo`3oool0<P3oool006l0oooo00<000000?oo
o`3oool0D03oool2000003L0oooo0P00000Z0?ooo`8000009@3oool00`000000oooo0?ooo`0T0?oo
o`800000:@3oool2000003L0oooo0P00001A0?ooo`030000003oool0oooo0380oooo001_0?ooo`03
0000003oool0oooo0580oooo0P00000b0?ooo`<00000:P3oool2000002L0oooo00<000000?ooo`3o
ool09P3oool2000002T0oooo0`00000a0?ooo`<00000D`3oool00`000000oooo0?ooo`0b0?ooo`00
K`3oool00`000000oooo0?ooo`1D0?ooo`800000;@3oool3000002/0oooo0P00000Y0?ooo`030000
003oool0oooo02P0oooo00<000000?ooo`3oool0:@3oool2000002d0oooo0P00001F0?ooo`030000
003oool0oooo0380oooo001_0?ooo`<00000EP3oool2000002P0oooo0`00000/0?ooo`800000:`3o
ool00`000000oooo0?ooo`0Y0?ooo`800000:`3oool3000002P0oooo0P00001F0?ooo`<00000=03o
ool006l0oooo00<000000?ooo`3oool0F03oool400000240oooo0`00000]0?ooo`800000;@3oool0
0`000000oooo0?ooo`0[0?ooo`030000003oool0oooo02/0oooo0`00000R0?ooo`<00000FP3oool0
0`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`1L0?ooo`@000006`3oool20000
02l0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0/0?ooo`800000;P3oool30000
01T0oooo1P00001M0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo0600
oooo1@00000?0?ooo`L00000;`3oool200000300oooo00<000000?ooo`3oool0;P3oool00`000000
oooo0?ooo`0^0?ooo`D000003`3oool5000006<0oooo00<000000?ooo`3oool0<P3oool006l0oooo
00<000000?ooo`3oool0I@3oool?000003@0oooo0P00000Z0?oooa000000:P3oool2000003<0oooo
3`00001X0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo0:L0oooo00<0
00000?ooo`3oool0903oool6000000P0oooo00<000000?ooo`3oool0103oool7000002H0oooo00<0
00000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2V
0?ooo`030000003oool0oooo02<0oooo1@00000;0?ooo`030000003oool0oooo00T0oooo1P00000S
0?ooo`800000Y`3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2T0?oo
o`8000008P3oool5000000l0oooo00<000000?ooo`3oool03@3oool4000002<0oooo00<000000?oo
o`3oool0Y03oool00`000000oooo0?ooo`0b0?ooo`00K`3oool300000:<0oooo00<000000?ooo`3o
ool0803oool400000180oooo00<000000?ooo`3oool0403oool300000280oooo00<000000?ooo`3o
ool0X@3oool3000003@0oooo001_0?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool0
7`3oool3000001D0oooo00<000000?ooo`3oool04P3oool300000240oooo0P00002S0?ooo`030000
003oool0oooo0380oooo001_0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool07P3o
ool2000001P0oooo00<000000?ooo`3oool0503oool300000240oooo00<000000?ooo`3oool0X03o
ool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2O0?ooo`8000007`3oool2
000001X0oooo00<000000?ooo`3oool05P3oool300000200oooo00<000000?ooo`3oool0W`3oool0
0`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo
01d0oooo0P00000L0?ooo`030000003oool0oooo01P0oooo0`00000O0?ooo`800000W`3oool00`00
0000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo01`0
oooo0`00000M0?ooo`030000003oool0oooo01X0oooo0`00000O0?ooo`030000003oool0oooo09`0
oooo00<000000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0V`3oool2000001h0oooo
0P00000O0?ooo`030000003oool0oooo01/0oooo0`00000O0?ooo`030000003oool0oooo09/0oooo
00<000000?ooo`3oool0<P3oool006l0oooo0`00002J0?ooo`030000003oool0oooo01d0oooo0P00
000P0?ooo`030000003oool0oooo01d0oooo0`00000N0?ooo`800000V@3oool3000003@0oooo001_
0?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3oool0703oool300000240oooo00<00000
0?ooo`3oool07`3oool2000001l0oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`0b
0?ooo`00GP3oool00`3ookl0M000000000020000000304RLg`3oool0oooo00T0oooo00<000000?oo
o`3oool0V03oool00`000000oooo0?ooo`0L0?ooo`8000008`3oool00`000000oooo0?ooo`0P0?oo
o`8000007`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo0380oooo001P0?ooo`03
0?noM0000000M;oo00`0oooo00<000000?ooo`3oool0UP3oool2000001h0oooo0P00000T0?ooo`03
0000003oool0oooo0240oooo0`00000N0?ooo`800000U`3oool00`000000oooo0?ooo`0b0?ooo`00
H03oool00`3ooml0W4P0001d_`0<0?ooo`030000003oool0oooo09D0oooo00<000000?ooo`3oool0
703oool3000002D0oooo00<000000?ooo`3oool08`3oool2000001l0oooo00<000000?ooo`3oool0
U03oool00`000000oooo0?ooo`0b0?ooo`00H@3oool00`3OW4P004RL0=ooo`0;0?ooo`030000003o
ool0oooo09@0oooo00<000000?ooo`3oool0703oool3000002H0oooo00<000000?ooo`3oool0903o
ool2000001l0oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`0b0?ooo`00D@3oool0
0`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0omnL04P0B02Lgol02`3oool0
0`000000oooo0?ooo`2C0?ooo`030000003oool0oooo01`0oooo0`00000W0?ooo`030000003oool0
oooo02D0oooo0P00000O0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool0<P3oool0
0640oooo00<0oono07@0001d_ol02`3oool400000940oooo00<000000?ooo`3oool0703oool20000
0280oooo3`00000R0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0SP3oool40000
03@0oooo001R0?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3oool0T03oool2000001h0
oooo0P00000N0?ooo`H000001P3oool00`000000oooo0?ooo`040?ooo`D000007P3oool2000001l0
oooo00<000000?ooo`3oool0T03oool00`000000oooo0?ooo`0b0?ooo`00HP3oool00`3OW4P004RL
0=ooo`0:0?ooo`030000003oool0oooo08l0oooo00<000000?ooo`3oool07@3oool2000001d0oooo
0`00000;0?ooo`030000003oool0oooo00P0oooo0`00000M0?ooo`8000007`3oool00`000000oooo
0?ooo`2?0?ooo`030000003oool0oooo0380oooo001R0?ooo`030?oOW01804P0W=oo00X0oooo00<0
00000?ooo`3oool0SP3oool00`000000oooo0?ooo`0M0?ooo`800000703oool3000000d0oooo00<0
00000?ooo`3oool02P3oool3000001`0oooo0P00000O0?ooo`800000S`3oool00`000000oooo0?oo
o`0b0?ooo`00HP3oool00`3ookl0M00007Boo`0:0?ooo`030000003oool0oooo08d0oooo00<00000
0?ooo`3oool07@3oool2000001/0oooo0P00000@0?ooo`030000003oool0oooo00`0oooo0`00000K
0?ooo`800000803oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo0380oooo001S0?oo
o`030;md0000M;l0oooo00T0oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`0M0?oo
o`8000006P3oool300000140oooo00<000000?ooo`3oool03P3oool3000001X0oooo0P00000P0?oo
o`030000003oool0oooo08/0oooo00<000000?ooo`3oool0<P3oool00640oooo00D0oono07@00000
0000001d0;ooo`090?ooo`030000003oool0oooo08X0oooo0P00000O0?ooo`8000006P3oool20000
01<0oooo00<000000?ooo`3oool0403oool3000001T0oooo0P00000P0?ooo`030000003oool0oooo
08X0oooo00<000000?ooo`3oool0<P3oool006l0oooo0`0000290?ooo`030000003oool0oooo01h0
oooo0P00000J0?ooo`800000503oool00`000000oooo0?ooo`0A0?ooo`<000006@3oool200000200
oooo00<000000?ooo`3oool0Q`3oool3000003@0oooo000l0?ooo`030?oo_`1dM;l0oooo0300oooo
00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0N0?ooo`8000006P3oool00`000000oooo
0?ooo`0D0?ooo`030000003oool0oooo01<0oooo0`00000H0?ooo`800000803oool00`000000oooo
0?ooo`280?ooo`030000003oool0oooo0380oooo000l0?ooo`040?oOW0180000001809cOobl0oooo
00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01L0oooo
0P00000G0?ooo`030000003oool0oooo01@0oooo0`00000G0?ooo`8000008@3oool00`000000oooo
0?ooo`270?ooo`030000003oool0oooo0380oooo000n0?ooo`030;md00000000000000<0000000<0
001d0;ooo`3oool0:03oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo01l0oooo0P00
000H0?ooo`800000603oool00`000000oooo0?ooo`0F0?ooo`8000005`3oool200000240oooo00<0
00000?ooo`3oool0QP3oool00`000000oooo0?ooo`0b0?ooo`00@P3oool00`3ogi`0B00007Boo`0Z
0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0F
0?ooo`8000006@3oool00`000000oooo0?ooo`0G0?ooo`8000005`3oool200000240oooo00<00000
0?ooo`3oool0Q@3oool00`000000oooo0?ooo`0b0?ooo`00@@3oool00`3OW4P0001d0;ooo`0[0?oo
o`030000003oool0oooo08@0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0F0?oo
o`8000006P3oool00`000000oooo0?ooo`0H0?ooo`8000005`3oool200000240oooo00<000000?oo
o`3oool0Q03oool00`000000oooo0?ooo`0b0?ooo`00?03oool01P3ogi`0B01809cOo`3o_g@00000
04RLgbd0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0P0?ooo`8000005`3oool2
000001/0oooo00<000000?ooo`3oool06@3oool2000001L0oooo00<000000?ooo`3oool0803oool2
000008@0oooo00<000000?ooo`3oool0<P3oool003`0oooo00@0omnL04P00000000007Bo;`3oool3
00000880oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0E0?ooo`800000703oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01D0oooo0P00000S0?ooo`030000003oool0
oooo07l0oooo0`00000d0?ooo`00K`3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo
0240oooo00<000000?ooo`3oool05@3oool2000001d0oooo00<000000?ooo`3oool06`3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0P03oool00`00
0000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool0503oool2000001h0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0O`3oool00`000000
oooo0?ooo`0b0?ooo`00?`3oool01P3OW4P00000000000000000001809cOobX0oooo00<000000?oo
o`3oool0O`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool06P3oool6000001`0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0S0?oo
o`030000003oool0oooo07h0oooo00<000000?ooo`3oool0<P3oool003d0oooo00T0oono07@00018
W=l0omnL09boo`3oool0gia80018W03Oool0:@3oool00`000000oooo0?ooo`1n0?ooo`030000003o
ool0oooo02<0oooo00<000000?ooo`3oool0503oool2000001P0oooo1@0000020?ooo`030000003o
ool0000000<000006P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02@0oooo00<0
00000?ooo`3oool0O@3oool00`000000oooo0?ooo`0b0?ooo`00?@3oool02@2oM7@0_ooo0?ooo`3o
_i`0W=oo0?ooo`3ookl0M01d0;ooo`0Y0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3o
ool08`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo100000050?ooo`03
0000003oool0oooo00<0oooo0`00000G0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool0903oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo0380oooo000l0?ooo`090?oO
W02L_ol0oooo0?oog`2LM9`0gooo0?ooo`3ogi`0B7Bo02X0oooo00<000000?ooo`3oool0O03oool0
0`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool05@3oool2
000000P0oooo00<000000?ooo`3oool01@3oool3000001H0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0<P3oool0
03`0oooo00T0oknL09cOo`3oool0_g@007Boo`3oool0gia80000M02oool0:P3oool3000007X0oooo
0P00000V0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool04`3oool2000000X0oooo
00<000000?ooo`3oool01`3oool2000001D0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo07T0oooo0`00000d0?ooo`00?03oool0103ookl0M0000000
001d_ol_0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool09P3oool00`000000oooo
0?ooo`0D0?ooo`800000503oool2000000/0oooo00<000000?ooo`3oool0203oool2000001D0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo07X0oooo
00<000000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?oo
o`0W0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04P3oool2000000`0oooo00<0
00000?ooo`3oool02@3oool2000001@0oooo00<000000?ooo`3oool0503oool2000002L0oooo00<0
00000?ooo`3oool0N@3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`1g
0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0503oool2000001<0oooo0P00000=
0?ooo`030000003oool0oooo00X0oooo0P00000D0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0380oooo00100?oo
o`060?noM000000000000000000007@0_ooo:@3oool00`000000oooo0?ooo`1f0?ooo`030000003o
ool0oooo02P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0B0?ooo`8000003@3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0<P3o
ool003P0oooo00H0_ibL0=ooo`3oool0oooo0?oOW0180003000000040018W03Oool0oono07Ad_bX0
oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool04@3oool2000000h0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0A0?ooo`8000005@3oool2000002T0oooo0P00001g0?ooo`030000003oool0oooo0380
oooo000g0?ooo`030?oog`2LB000000000<0000000<007Bo0?ooo`3oool00P3oool00`3o_g@0001d
0;ooo`0Z0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0:P3oool00`000000oooo
0?ooo`0C0?ooo`8000004`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02T0oooo
00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0b0?ooo`00=`3oool00`3ookl0M00004RL
g`0e0?ooo`<00000L`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02X0oooo00<0
00000?ooo`3oool0L@3oool3000003@0oooo001_0?ooo`030000003oool0oooo07<0oooo00<00000
0?ooo`3oool0:P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`8000004P3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0b
0?ooo`00?`3oool01P3ogi`0B000000000000000001809cOoa/0oooo00@0_g@0000000000000M;oo
2`3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`0@0?ooo`800000403oool00`000000oooo0?ooo`0=0?ooo`800000
4P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0
L@3oool00`000000oooo0?ooo`0b0?ooo`00?P3oool00`2oM000000007Boo`020?ooo`030?oog`2L
M9`0gooo01T0oooo00H0gia8001d_`3oool0omnL04P00018W=l:0?ooo`030000003oool0oooo0740
oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0740
oooo00<000000?ooo`3oool0<P3oool003d0oooo00<0gia804RLg`3oool0103oool00`3OW9`0_ooo
0?ooo`0G0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02@3oool00`000000oooo
0?ooo`1`0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`0@0?ooo`8000005@3oool2000002h0oooo00<000000?ooo`3oool0L03oool00`000000oooo
0?ooo`0b0?ooo`00?03oool00`3ogi`0M9cO0?ooo`040?ooo`030?oog`2LM9`0gooo01P0oooo00<0
oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0]0?ooo`800000L03oool00`000000oooo
0?ooo`0b0?ooo`00?03oool00`3o_i`0W=oo0?ooo`030?ooo`030?oog`2LB000M;oo01T0oooo00<0
oooO09a80000M;l00`3oool00`3OW4P0001809cOo`080?ooo`030000003oool0oooo06d0oooo0P00
000a0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo02h0oooo0P00001_0?ooo`030000003oool0oooo0380oooo000l
0?ooo`030?oOW018M;l0oooo0080oooo00<0gia800000018W=l06`3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00P0ooooo`00002o000003@0oooo000m0?ooo`060?noM00000000000
00000000B9`0gooo703oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00<0
00000?ooo`3oool0K@3oool200000340oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool200000180oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0^0?ooo`800000K`3oool00`000000oooo0?oo
o`0b0?ooo`00G`3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00<00000
0?ooo`3oool0K`3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0KP3oool00`000000oooo0?ooo`0b0?ooo`00G`3oool0103ooml0W4P00018W03Oool2
0?ooo`030=nLB000B9`0gooo00L0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`0]
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0C
0?ooo`800000;P3oool200000740oooo00<000000?ooo`3oool0<P3oool003l0oooo00H0omnL04P0
0000000000000000B02LgolK0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3o
ool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0;03o
ool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo0380oooo000m0?ooo`040?oog`2LB000
04RL0=ooo`80oooo00<0okmd0018W03Oool06`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0
oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool0403oool200000100oooo00<000000?ooo`3oool03@3oool200000180
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0740
oooo00<000000?ooo`3oool0<P3oool003d0oooo00<0_gA809cOo`3oool00`3oool00`3ooml0W4Qd
0;ooo`0L0?ooo`040=nLB00000000000001d_`T0oooo00<000000?ooo`3oool0L`3oool00`000000
oooo0?ooo`0Z0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo0P00000B0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo0380
oooo000l0?ooo`030?oOW02L_ol0oooo00@0oooo00<0omnL04Qd_`3oool0:@3oool3000007<0oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0740oooo
0`00000d0?ooo`00?03oool00`3o_i`0W=oo0?ooo`030?ooo`030?oOW018B9`0gooo02X0oooo00<0
00000?ooo`3oool0M03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01<0oooo0P00
000B0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`0@0?ooo`8000005P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo07@0oooo00<0
00000?ooo`3oool0<P3oool003`0oooo00@0oono07@000000000M;oo;`3oool00`000000oooo0?oo
o`1e0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?oo
o`1e0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo07H0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`0D0?ooo`8000004`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00/0oooo0P00000C0?ooo`030000003oool0oooo01@0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo0380oooo001_0?oo
o`030000003oool0oooo07L0oooo00<000000?ooo`3oool09`3oool2000001D0oooo0P00000C0?oo
o`8000003@3oool00`000000oooo0?ooo`0:0?ooo`8000004`3oool2000001H0oooo00<000000?oo
o`3oool09`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo0380oooo000l0?ooo`03
0?oo_`2LW=l0oooo00<0oooo00@0okmd0000000007@0_ooo:@3oool00`000000oooo0?ooo`1h0?oo
o`030000003oool0oooo02H0oooo0P00000F0?ooo`030000003oool0oooo0180oooo0P00000<0?oo
o`030000003oool0oooo00T0oooo0P00000D0?ooo`8000005@3oool2000002P0oooo00<000000?oo
o`3oool0N03oool00`000000oooo0?ooo`0b0?ooo`00?03oool0103ookl0W9cO0?oo_`1d00040000
00030018W03Oool0oooo02P0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo01@0oooo0P00000D0?ooo`8000002`3oool00`000000oooo0?ooo`080?oo
o`8000005@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02H0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`0b0?ooo`00?03oool02@3ogi`0B0000000000004P0W=oo
0?ooo`3ooml0W4Qd0;ooo`0Z0?ooo`<00000NP3oool00`000000oooo0?ooo`0U0?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool04`3oool2000000X0oooo00<000000?ooo`3oool01`3o
ool2000001D0oooo0P00000E0?ooo`8000009`3oool00`000000oooo0?ooo`1h0?ooo`<00000=03o
ool003P0oooo00@0oooO09a8000007@0_ooo<`3oool00`000000oooo0?ooo`1k0?ooo`030000003o
ool0oooo02D0oooo00<000000?ooo`3oool0503oool2000001H0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`060?ooo`8000005P3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo02@0oooo0P00001m0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0
oooo07`0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo01D0oooo0`0000060?ooo`030000003oool0oooo00<0oooo1000000F0?ooo`8000005@3oool2
000002D0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`0b0?ooo`00?03oool01@3o
gkl0_moo0?ooo`3oool0_g@000<0000000<0001d0;ooo`3oool0:03oool00`000000oooo0?ooo`1m
0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0503oool2000001P0oooo1@000002
0?ooo`030000003oool0000000D00000603oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`0b0?ooo`00?03oool00`2o
M000000000000002000000050000M02oool0oooo0;nLW03Oool0:@3oool00`000000oooo0?ooo`1n
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0J
0?ooo`H00000703oool2000001D0oooo0P00000T0?ooo`030000003oool0oooo07l0oooo00<00000
0?ooo`3oool0<P3oool003X0oooo00D0_g@000000000000004RL0=ooo`030?ooo`030?oog`2LB000
M;oo0080oooo00<0_g@00018W03Oool09@3oool00`000000oooo0?ooo`1o0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0
oooo01/0oooo0P00000F0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0P03oool0
0`000000oooo0?ooo`0b0?ooo`00?03oool00`3o_i`0W=oo0?ooo`080?ooo`030?oOW018M;l0oooo
02D0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo
01D0oooo0P00000M0?ooo`030000003oool0oooo01X0oooo0P00000F0?ooo`8000008`3oool00`00
0000oooo0?ooo`210?ooo`030000003oool0oooo0380oooo000l0?ooo`030?oo_`1dM;l0oooo00L0
oooo00<0okmd0018W03Oool09P3oool300000840oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0E0?ooo`800000703oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`200?ooo`<00000=03oool003`0oooo
00@0omnL04P0000007@0_ooo1@3oool00`3OW4P007Bo0?ooo`0W0?ooo`030000003oool0oooo0880
oooo0P00000R0?ooo`8000005`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool05@3oool200000280oooo00<000000?ooo`3oool0P`3oool00`000000
oooo0?ooo`0b0?ooo`00?@3oool00`3ogi`0B000000000050000000304RLg`3oool0oooo02L0oooo
00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01H0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01H0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo0380oooo
00120?ooo`030?noM00004P0W=oo02X0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?oo
o`0P0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`250?ooo`030000003oool0oooo0380oooo00110?ooo`030;md0000M;l0oooo02/0oooo00<00000
0?ooo`3oool0QP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01H0oooo0P00000H
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0O
0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0<P3oool003`0oooo00H0omnL04P0
001d_ol0okmd0000001d_ol]0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool07`3o
ool2000001P0oooo0P00000G0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0<P3o
ool003`0oooo00@0omnL04P00000000007Bo;`3oool00`000000oooo0?ooo`280?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool05`3oool2000001H0oooo00<000000?ooo`3oool0503o
ool00`000000oooo0?ooo`0G0?ooo`800000803oool00`000000oooo0?ooo`280?ooo`030000003o
ool0oooo0380oooo001_0?ooo`<00000R@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0
oooo01L0oooo0P00000E0?ooo`030000003oool0oooo0180oooo0P00000I0?ooo`800000803oool0
0`000000oooo0?ooo`270?ooo`<00000=03oool006l0oooo00<000000?ooo`3oool0RP3oool00`00
0000oooo0?ooo`0O0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0A0?ooo`8000006@3oool2000001l0oooo0P00002<0?ooo`030000003oool0oooo
0380oooo001N0?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool02@3oool00`000000
oooo0?ooo`2;0?ooo`030000003oool0oooo01h0oooo0P00000J0?ooo`8000004P3oool00`000000
oooo0?ooo`0?0?ooo`<000006P3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo08`0
oooo00<000000?ooo`3oool0<P3oool00600oooo00<0okmd0000001d_ol0303oool00`000000oooo
0?ooo`2<0?ooo`030000003oool0oooo01h0oooo0P00000K0?ooo`800000403oool00`000000oooo
0?ooo`0=0?ooo`<000006`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo08d0oooo
00<000000?ooo`3oool0<P3oool00600oooo00<0oooO09a80000M;l0303oool00`000000oooo0?oo
o`2=0?ooo`8000007`3oool2000001/0oooo0`00000>0?ooo`030000003oool0oooo00/0oooo0`00
000L0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?oo
o`0b0?ooo`00H@3oool00`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo08l0oooo00<00000
0?ooo`3oool07@3oool2000001`0oooo1000000;0?ooo`030000003oool0oooo00T0oooo0`00000M
0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0S`3oool00`000000oooo0?ooo`0b
0?ooo`00H@3oool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0oooo0900oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`0M0?ooo`D000001`3oool00`000000oooo0?ooo`050?oo
o`@000007`3oool00`000000oooo0?ooo`0L0?ooo`800000TP3oool00`000000oooo0?ooo`0b0?oo
o`00H@3oool00`3ookl0M00007Boo`0;0?ooo`@00000T03oool00`000000oooo0?ooo`0M0?ooo`03
0000003oool0oooo0200oooo4P00000O0?ooo`8000007P3oool00`000000oooo0?ooo`2?0?ooo`@0
0000=03oool00680oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`2B0?ooo`030000
003oool0oooo01d0oooo0P00000X0?ooo`030000003oool0oooo02D0oooo0P00000N0?ooo`030000
003oool0oooo09<0oooo00<000000?ooo`3oool0<P3oool00680oooo00<0gia80018W03Oool02P3o
ool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo01d0oooo0P00000W0?ooo`030000003o
ool0oooo02@0oooo0P00000N0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0<P3o
ool00680oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`2D0?ooo`030000003oool0
oooo01d0oooo0P00000V0?ooo`030000003oool0oooo0280oooo0`00000N0?ooo`030000003oool0
oooo09D0oooo00<000000?ooo`3oool0<P3oool00680oooo00<0oono07@0001d_ol02P3oool00`00
0000oooo0?ooo`2E0?ooo`8000007P3oool2000002D0oooo00<000000?ooo`3oool08@3oool30000
01h0oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`0b0?ooo`00H`3oool00`2oM000
07Bo0?ooo`090?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3oool07@3oool2000002<0
oooo00<000000?ooo`3oool0803oool2000001h0oooo0P00002I0?ooo`030000003oool0oooo0380
oooo001Q0?ooo`050?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`2H0?ooo`80
00007P3oool200000280oooo00<000000?ooo`3oool07`3oool2000001h0oooo00<000000?ooo`3o
ool0V@3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool3000009X0oooo00<000000?ooo`3oool0
7@3oool200000200oooo00<000000?ooo`3oool07@3oool3000001h0oooo00<000000?ooo`3oool0
V03oool3000003@0oooo001_0?ooo`030000003oool0oooo09/0oooo0P00000N0?ooo`8000007`3o
ool00`000000oooo0?ooo`0L0?ooo`<000007P3oool00`000000oooo0?ooo`2K0?ooo`030000003o
ool0oooo0380oooo001_0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool07@3oool2
000001d0oooo00<000000?ooo`3oool06`3oool2000001l0oooo00<000000?ooo`3oool0W03oool0
0`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0oooo
01d0oooo0P00000L0?ooo`030000003oool0oooo01T0oooo0P00000O0?ooo`800000W`3oool00`00
0000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2O0?ooo`8000007`3oool2000001X0
oooo00<000000?ooo`3oool05`3oool200000200oooo00<000000?ooo`3oool0W`3oool00`000000
oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01h0oooo
0P00000H0?ooo`030000003oool0oooo01D0oooo0P00000Q0?ooo`030000003oool0oooo0:00oooo
00<000000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0XP3oool00`000000oooo0?oo
o`0O0?ooo`<000005@3oool00`000000oooo0?ooo`0B0?ooo`<000008P3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo0:<0oooo00<00000
0?ooo`3oool0803oool3000001<0oooo00<000000?ooo`3oool03`3oool4000002<0oooo00<00000
0?ooo`3oool0XP3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool300000:@0oooo0P00000R0?oo
o`@00000403oool00`000000oooo0?ooo`0=0?ooo`@000008`3oool200000:<0oooo0`00000d0?oo
o`00K`3oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo0240oooo1P00000<0?ooo`03
0000003oool0oooo00T0oooo1000000V0?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3o
ool0<P3oool006l0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0T0?ooo`L00000
1`3oool00`000000oooo0?ooo`050?ooo`H000009`3oool00`000000oooo0?ooo`2V0?ooo`030000
003oool0oooo0380oooo001_0?ooo`030000003oool0oooo06L0oooo3@00000d0?ooo`800000:@3o
oolB000002X0oooo0P00000c0?ooo`h00000J03oool00`000000oooo0?ooo`0b0?ooo`00K`3oool0
0`000000oooo0?ooo`1Q0?ooo`H000003@3oool600000300oooo0P00000`0?ooo`030000003oool0
oooo02h0oooo0P00000_0?ooo`H000003P3oool600000680oooo00<000000?ooo`3oool0<P3oool0
06l0oooo00<000000?ooo`3oool0F`3oool6000001T0oooo0`00000_0?ooo`030000003oool0oooo
02d0oooo00<000000?ooo`3oool0;03oool2000002h0oooo0`00000J0?ooo`@00000GP3oool00`00
0000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`1I0?ooo`8000008P3oool3000002d0
oooo0P00000]0?ooo`030000003oool0oooo02X0oooo0P00000]0?ooo`<000008@3oool4000005X0
oooo00<000000?ooo`3oool0<P3oool006l0oooo0`00001G0?ooo`8000009`3oool3000002`0oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0X0?ooo`800000;03oool3000002P0oooo
0P00001F0?ooo`<00000=03oool006l0oooo00<000000?ooo`3oool0E@3oool2000002`0oooo0P00
000[0?ooo`800000:P3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo02T0oooo0`00
000]0?ooo`800000EP3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`1B
0?ooo`<00000<03oool2000002/0oooo0P00000X0?ooo`030000003oool0oooo02D0oooo0P00000Z
0?ooo`800000<P3oool2000005@0oooo00<000000?ooo`3oool0<P3oool006l0oooo00<000000?oo
o`3oool0D03oool2000003D0oooo0`00000Z0?ooo`8000009P3oool00`000000oooo0?ooo`0S0?oo
o`800000:P3oool2000003H0oooo0P00001B0?ooo`030000003oool0oooo0380oooo001_0?ooo`03
0000003oool0oooo04h0oooo0P00000j0?ooo`800000:P3oool2000002@0oooo00<000000?ooo`3o
ool08@3oool2000002X0oooo0P00000j0?ooo`800000D03oool00`000000oooo0?ooo`0b0?ooo`00
K`3oool00`000000oooo0?ooo`1<0?ooo`800000?P3oool3000002T0oooo0P00000R0?ooo`030000
003oool0oooo01l0oooo0P00000Z0?ooo`800000?P3oool3000004d0oooo00<000000?ooo`3oool0
<P3oool006l0oooo00<000000?ooo`3oool0BP3oool2000004<0oooo0P00000Y0?ooo`<000007`3o
ool00`000000oooo0?ooo`0M0?ooo`800000:P3oool2000004<0oooo0P00001;0?ooo`030000003o
ool0oooo0380oooo001_0?ooo`<00000B03oool2000004L0oooo0P00000Z0?ooo`8000007@3oool0
0`000000oooo0?ooo`0K0?ooo`800000:P3oool2000004L0oooo0P0000170?ooo`<00000=03oool0
06l0oooo00<000000?ooo`3oool0A@3oool3000004/0oooo0P00000Z0?ooo`<000006P3oool00`00
0000oooo0?ooo`0I0?ooo`800000:@3oool3000004/0oooo0P0000170?ooo`030000003oool0oooo
0380oooo001N0?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0203oool00`000000
oooo0?ooo`130?ooo`800000D03oool2000002/0oooo0P00000H0?ooo`030000003oool0oooo01H0
oooo0`00000Y0?ooo`800000D03oool2000004D0oooo00<000000?ooo`3oool0<P3oool005l0oooo
00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`080?ooo`030000003oool0oooo0440oooo
0P00001D0?ooo`800000:`3oool2000001H0oooo00<000000?ooo`3oool04`3oool3000002X0oooo
0P00001D0?ooo`800000@`3oool00`000000oooo0?ooo`0b0?ooo`00H03oool00`3o_g@007Bo0?oo
o`0<0?ooo`030000003oool0oooo03h0oooo0`00001H0?ooo`800000:`3oool3000001<0oooo00<0
00000?ooo`3oool03`3oool4000002/0oooo0P00001H0?ooo`800000@@3oool00`000000oooo0?oo
o`0b0?ooo`00H@3oool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo03`0oooo0P00001M
0?ooo`<00000:`3oool6000000d0oooo00<000000?ooo`3oool0303oool3000002d0oooo0P00001L
0?ooo`800000?`3oool00`000000oooo0?ooo`0b0?ooo`00HP3oool00`3o_g@0B9cO0?ooo`0:0?oo
o`030000003oool0oooo03X0oooo0P00001R0?ooo`800000;`3oool7000000H0oooo00<000000?oo
o`3oool00`3oool9000002h0oooo0P00001P0?ooo`800000?@3oool00`000000oooo0?ooo`0b0?oo
o`00H`3oool00`3OW4P0B9cO0?ooo`090?ooo`@00000=`3oool2000006H0oooo0P00000d0?ooo``0
0000=@3oool2000006@0oooo0P00000h0?ooo`@00000=03oool006@0oooo00<0_g@007Boo`3oool0
203oool00`000000oooo0?ooo`0f0?ooo`800000JP3oool2000003P0oooo00<000000?ooo`3oool0
=@3oool3000006P0oooo0P00000i0?ooo`030000003oool0oooo0380oooo001T0?ooo`030?oOW018
07@0_ooo00P0oooo00<000000?ooo`3oool0=03oool2000006h0oooo0P00000f0?ooo`030000003o
ool0oooo03<0oooo0P00001]0?ooo`800000=`3oool00`000000oooo0?ooo`0b0?ooo`00I03oool0
0`3ooml0W4P007Boo`080?ooo`030000003oool0oooo0380oooo0P00001b0?ooo`800000=03oool0
0`000000oooo0?ooo`0a0?ooo`800000L@3oool2000003D0oooo00<000000?ooo`3oool0<P3oool0
06@0oooo00<0oono07@00018W=l0203oool00`000000oooo0?ooo`0`0?ooo`800000MP3oool20000
0380oooo00<000000?ooo`3oool0;`3oool2000007D0oooo0`00000b0?ooo`030000003oool0oooo
0380oooo001P0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?ooo`030000003oool0
oooo02h0oooo0P00001j0?ooo`800000<03oool00`000000oooo0?ooo`0]0?ooo`800000NP3oool2
00000300oooo00<000000?ooo`3oool0<P3oool00640oooo00D0omnL04P000000000000007Boo`09
0?ooo`030000003oool0oooo02`0oooo0P00001n0?ooo`800000;P3oool00`000000oooo0?ooo`0[
0?ooo`800000OP3oool2000002h0oooo00<000000?ooo`3oool0<P3oool006l0oooo0`00000Y0?oo
o`<00000PP3oool2000002`0oooo00<000000?ooo`3oool0:@3oool200000880oooo0P00000Z0?oo
o`<00000=03oool006l0oooo00<000000?ooo`3oool09P3oool3000003`0oooo4000000k0?ooo`80
0000:P3oool00`000000oooo0?ooo`0V0?ooo`<00000?03oool>000003`0oooo0`00000Y0?ooo`03
0000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo02<0oooo0`00000i0?ooo`H00000
403oool5000003P0oooo0`00000W0?ooo`030000003oool0oooo02<0oooo0`00000h0?ooo`L00000
3P3oool8000003L0oooo0`00000V0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003o
ool0oooo0240oooo0P00000g0?ooo`D000006`3oool6000003D0oooo0`00000T0?ooo`030000003o
ool0oooo0200oooo0`00000f0?ooo`D000007@3oool5000003D0oooo0`00000S0?ooo`030000003o
ool0oooo0380oooo001_0?ooo`030000003oool0oooo01h0oooo0`00000e0?ooo`@000009P3oool6
00000380oooo1000000P0?ooo`030000003oool0oooo01d0oooo0`00000d0?ooo`D000009`3oool4
000003@0oooo0`00000P0?ooo`030000003oool0oooo0380oooo001_0?ooo`030000003oool0oooo
01/0oooo0`00000d0?ooo`@00000<03oool3000003<0oooo0`00000M0?ooo`030000003oool0oooo
01X0oooo0`00000c0?ooo`@00000<03oool4000003<0oooo0P00000N0?ooo`030000003oool0oooo
0380oooo001_0?ooo`030000003oool0oooo01P0oooo0`00000d0?ooo`<00000=`3oool400000380
oooo0P00000K0?ooo`030000003oool0oooo01L0oooo0`00000d0?ooo`800000>03oool300000380
oooo0`00000K0?ooo`030000003oool0oooo0380oooo001_0?ooo`<000005P3oool2000003@0oooo
0`00000n0?ooo`<00000<@3oool3000001P0oooo00<000000?ooo`3oool0503oool3000003@0oooo
0`00000m0?ooo`<00000<P3oool3000001H0oooo0`00000d0?ooo`00K`3oool00`000000oooo0?oo
o`0C0?ooo`<00000<`3oool3000004@0oooo0P00000b0?ooo`@00000503oool00`000000oooo0?oo
o`0?0?ooo`D00000=@3oool2000004<0oooo0`00000b0?ooo`D000004`3oool00`000000oooo0?oo
o`0b0?ooo`00K`3oool00`000000oooo0?ooo`0:0?ooo`T00000<`3oool3000004T0oooo0`00000c
0?ooo`H000003P3oool00`000000oooo0?ooo`0:0?ooo`D00000=`3oool3000004P0oooo0`00000d
0?ooo`H000003@3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool=000003T0oooo0`00001?0?oo
o`800000=`3ooolK000003T0oooo0`00001>0?ooo`<00000=`3oool>000003@0oooo001_0?ooo`03
0000003oool0oooo0400oooo0`00001D0?ooo`<00000@P3oool00`000000oooo0?ooo`110?ooo`80
0000E03oool300000440oooo00<000000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0
?@3oool3000005X0oooo0P0000100?ooo`030000003oool0oooo03h0oooo0`00001I0?ooo`800000
?`3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`0j0?ooo`<00000G`3o
ool3000003d0oooo00<000000?ooo`3oool0>`3oool3000005h0oooo0`00000l0?ooo`030000003o
ool0oooo0380oooo001_0?ooo`<00000=`3oool3000006D0oooo0P00000k0?ooo`030000003oool0
oooo03P0oooo0`00001T0?ooo`800000>03oool3000003@0oooo001_0?ooo`030000003oool0oooo
03@0oooo0`00001Z0?ooo`<00000>03oool00`000000oooo0?ooo`0d0?ooo`@00000J@3oool30000
03L0oooo00<000000?ooo`3oool0<P3oool006l0oooo00<000000?ooo`3oool0<P3oool200000700
oooo0`00000e0?ooo`030000003oool0oooo0340oooo0`00001`0?ooo`800000=@3oool00`000000
oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`0_0?ooo`<00000M@3oool2000003<0oooo
00<000000?ooo`3oool0;P3oool3000007D0oooo0`00000b0?ooo`030000003oool0oooo0380oooo
001_0?ooo`030000003oool0oooo02d0oooo0P00001j0?ooo`<00000<03oool00`000000oooo0?oo
o`0[0?ooo`<00000N`3oool200000300oooo00<000000?ooo`3oool0<P3oool006l0oooo00<00000
0?ooo`3oool0:P3oool3000007l0oooo0`00000]0?ooo`030000003oool0oooo02T0oooo0P000020
0?ooo`<00000;@3oool00`000000oooo0?ooo`0b0?ooo`00K`3oool00`000000oooo0?ooo`0X0?oo
o`800000Q@3oool2000002/0oooo00<000000?ooo`3oool09P3oool3000008D0oooo0`00000Z0?oo
o`030000003oool0oooo0380oooo001_0?ooo`<000009@3oool3000008T0oooo0`00000X0?ooo`03
0000003oool0oooo02@0oooo0P00002;0?ooo`<000009@3oool3000003@0oooo001_0?ooo`030000
003oool0oooo02<0oooo0P00002?0?ooo`@00000903oool00`000000oooo0?ooo`0Q0?ooo`<00000
T03oool4000002<0oooo00<000000?ooo`3oool0<P3oool005h0oooo00<0oono07@0000000000P00
00000`1d_ol0oooo0?ooo`090?ooo`030000003oool0oooo01l0oooo1000002E0?ooo`D000007`3o
ool00`000000oooo0?ooo`0M0?ooo`@00000U`3oool5000001h0oooo00<000000?ooo`3oool0<P3o
ool005h0oooo00<0okmd00000000M;l00P3oool00`2oM4P0W=oo0?ooo`090?ooo`030000003oool0
oooo01T0oooo1P00002N0?ooo`@000006`3oool00`000000oooo0?ooo`0I0?ooo`@00000X03oool4
000001X0oooo00<000000?ooo`3oool0<P3oool006<0oooo00<0omnL04Q8W03Oool02@3oool00`00
0000oooo0?ooo`0E0?ooo`@00000Z03oool3000001P0oooo00<000000?ooo`3oool0503oool50000
0:P0oooo0P00000H0?ooo`030000003oool0oooo0380oooo001S0?ooo`030?oo_`1d0000M;oo00T0
oooo00<000000?ooo`3oool04P3oool300000:l0oooo1000000D0?ooo`030000003oool0oooo00h0
oooo1P00002_0?ooo`@00000503oool00`000000oooo0?ooo`0b0?ooo`00H`3oool00`3ogi`0B000
07Boo`090?ooo`030000003oool0oooo00T0oooo2@00002f0?ooo`X000002P3oool00`000000oooo
0?ooo`080?ooo`H00000^@3oool9000000/0oooo00<000000?ooo`3oool0<P3oool00680oooo00@0
omnL04P0000004P0W=oo2@3oool<00000<T0oooo5@0000380?ooo``00000=03oool00600oooo00D0
omnL04P000000000000007Boo`0:0?ooo`030000003oool0oooo0=`0oooo00<000000?ooo`3oool0
f`3oool00`000000oooo0?ooo`0b0?ooo`00HP3oool00`3ookl0M000001d_`0:0?ooo`030000003o
ool0oooo0=`0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0b0?ooo`00I03oool0
0`3OW4P007Bo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0A03oool0
0`000000oooo0?ooo`140?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0@`3oool0
0`000000oooo0?ooo`140?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0b0?ooo`00I03oool00`3ogi`0B01809cOo`080?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0b0?ooo`00H@3oool01P2oM000
M;oo0?ooo`3OW4P0001809cOo`P0ooooo`00002o000003@0oooo001R0?ooo`040;md000000000000
04RLgol0ooooo03oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0oooo
H`3oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0ooooH`3oool00?l0
ooooo`3ooomS0?ooo`00o`3ooomQ0?ooo`050;md00000000000000000018W=l0o03oool00?l0oooo
HP3oool00`3ooml0W4P004RLg`3m0?ooo`00o`3ooomS0?ooo`030=nLB000B9`0gooo0?`0oooo003o
0?ooof<0oooo00<0omnL04P0M02oool0o03oool00?l0oooo1P3oool01P3OW4P00000000000000000
001809cOo`L0oooo00D0gia8000000000000000007Boo`020?ooo`030=nLB000B9`0gooo0080oooo
0300_g@000000000M;l0oooo0?oo_`1d000000000000M02oool0gia800000018W=l0oooo0?oOW018
000000000018W03Oool0_g@000000000M;l0oooo0?noM0000000001d0;ooo`3oool0omnL04P00000
000004RL0=ooo`2oM0000000001d_`3oool0gia80018W03Oool0oooo0?noM00007@0_ooo0?ooo`3o
oml0W4P00000001d_ol40?ooo`030?noM0000000000000L0000000T0M;oo0?ooo`3ogi`0B0000000
000007@0_ono07@0001d_ol00P3oool02`3ogi`0B00000000000B9`0gooo0;md0000000007Bo0?oo
o`3ogi`0B0000080000000T0M;oo0=nLB000000004RL0=ooo`3o_g@000000000B02Lgol00P3oool0
102oM00000000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3oool00`3OW4P004RL
0=ooo`020?ooo`030;md0000000007Bo0<`0oooo003o0?ooo`L0oooo00@0oooO09a80000B9`0gooo
1`3oool00`3ogi`0B00007Boo`020?ooo`060;md0000M;l0oooo0?noM00007@0_ooo0P3oool0:`3O
W4P004RL0=noM01d_ol0gia80018W03Oool0oono07@0M02o_g@0001d09bLM02Lgol0omnL04P0B02L
gol0omnL04Qd_`2oM00007Bo0?noW02Lgol0omnL04P0M02ookl0W9cO0?ooo`3ogi`0B01809cOo`3o
gi`0B7Bo0;md0000M;l0oknL09cOo`3o_g@0001809cOo`3oool0oono07@0001d_ol00P3oool0102o
M00007Bo0;mdM02oool40?ooo`040?oog`2LB00004RL0=ooo`<0oooo00/0oooO09a80000B9`0gooo
0;md0000B9`0gooo0?oo_`1d07@0M7@004RLg`020?ooo`0T0?oOW01804P0W=oo0?oOW018M;l0_g@0
001d_`3o_i`0W=oo0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?noM00004P0W=oO
09adW03Oool0oooo0=nLB000B9`0gooO09adW03Oool0omnL04P0B02Lgol0oooo0?oOW018000007Bo
0P3oool00`3o_g@0001d0;ooo`020?ooo`040=nLB000B9`0gkmd07Bool/0oooo003o0?ooo`P0oooo
00<0okmd0000B02Lgol01`3oool00`3OW4P004RL0=ooo`030?ooo`050=nLW02oool0oono07@00018
W=l00P3oool01P3ogi`0B01809cOo`3oool0_g@0001d_`80oooo02<0oono07A8B01804P0W=oo0?oo
o`3ooml0W4P004RLg`3oool0okmd07AdM000B9`0gooo0?ooo`3ookl0M00007Boo`3OW9`0_ooo0?oo
g`2LB000B9cO0?ooo`3o_g@0M7Ad0018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3ooml0W4P0001d
_`020?ooo`030=nLB000B9`0gooo00H0oooo00<0okmd0000B02Lgol01@3oool0103OW7@0W=oo0;md
0000M;l20?ooo`0@0?oo_`1d000004RL0=ooo`3oool0oooO09a80018W=l0oooo0?noM01dM7@004RL
0=ooo`3oool0_g@0001d_`<0oooo00L0_g@00000001d_ol0oooo0?oo_`1d0000M;oo00<0oooo00<0
okmd0000B02Lgol00P3oool00`3OW4P004RL0=ooo`020?ooo`070?oOW0180000B9cO0?ooo`3ookl0
M00004RLg`020?ooo`030?oOW01804P0W=oo0<`0oooo003o0?ooo`P0oooo00<0oono07@00018W=l0
1`3oool01@3o_g@00000000000000000B9cO00<0oooo00`0oooO09a8000007@0_ooo0?ooo`3ooml0
W4P004RLg`3oool0okmd0018W03Oool20?ooo`030?noM0000000B9cO0080oooo00L0_g@00018W03O
ool0oooo0;md000004P0W=oo0080oooo00<0_g@004RLg`3oool00P3oool0502oM00004RL0=ooo`3o
ool0_g@00000B02Lgol0oooo0?oog`2LB00004RL0=ooo`3oool0gia80000001d_ol0oooo0?oOW018
04P0W=oo1P3oool00`3ookl0M00004RLg`070?ooo`030=nLB000B9`0gooo0080oooo00<0okmd0000
M02oool00P3oool02`2oM00004RL0=ooo`3oool0_g@00000B02Lgol0oooo0?noM00007@0_ooo0080
oooo00<0oono07@00018W=l00P3oool00`2oM00007Bo0?ooo`020?ooo`030?oo_`1d0000B9cO0080
oooo00<0okmd0000M02oool00`3oool02`3OW4P0001d0;ooo`3ooml0W4P00000M02oool0oooo0?oo
g`2LB000B9cO0<`0oooo003o0?ooo`P0oooo00<0oooO09a80000M;l01`3oool01`3ookl0M01809cO
o`3oool0gia800000018W=l00P3oool00`3OW4P004Q804Qd_`020?ooo`060;md0000M;l0oooo0?oo
_`1d0000M;oo0P3oool00`3ooml0W4P0001d_`020?ooo`070?noM00007@0_ooo0?ooo`3ookl0M000
07Boo`020?ooo`030=nLB000B9`0gooo0080oooo00L0okmd0000M02oool0oooo0?oo_`1d0000M;oo
0080oooo00/0gia80018B01dW=l0oooo0?noM00007@0M7Ad0;ooo`3ooml0W4P004RLg`060?ooo`03
0?oog`2LB00007Bo0080oooo00<0oono09bLg`3oool00P3oool00`3ogi`0B00007Boo`020?ooo`03
0?oo_`1d0000M;oo0080oooo00/0okmd0000M02oool0oooo0?oo_`1d0000M;oo0?ooo`3ookl0M000
07Boo`030?ooo`070;md0000B9`0gooo0?ooo`3OW4P004RL0=ooo`030?ooo`070;md0000B9`0gooo
0?ooo`3ookl0M00007Boo`030?ooo`070?oOW01804P0W=oo0?ooo`3OW4P004Q804Qd_`020?ooo`03
0;md0000M;l0oooo0</0oooo003o0?ooo`T0oooo00<0gia80000B018B0000`0000006018W=l0oooo
0?ooo`3oool0okmd0018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3o_g@0001d0;noM018W=l0oooo
0?noM00007@0_ooo0?ooo`3o_g@0001d0;ooo`80oooo00L0gia80000M02oool0omoO0=oo_`1d0000
M;oo0080oooo00`0_g@0001d_`3oool0oooo0?oOW01807@0_ooo0?ooo`3ogml0gono07@0001d_ol2
0?ooo`0@0;md0000M;l0oooo0?ooo`3o_g@0001d09bLM02Lgol0oono07@0001d_kl0M7Bo0?ooo`2o
M00004RL0=ooo`H0oooo00H0gia80000M02oool0oooo0?oOW018M;l40?ooo`0<0;md000007@0_ooo
0?ooo`3ooml0W4P004RLg`3oool0omoO0=oo_`1d0000M;oo0P3oool01`2oM00007Bo0?ooo`3oool0
okmd0000M02oool00P3oool01`3OW4P0001d0;ooo`3oool0omnL04P0B02Lgol00`3oool00`3OW4P0
001809cOo`020?ooo`030=nLB00007@0_ooo0080oooo00`0oono07@0001d_ol0oooo0?noM00007@0
_kmd04RLg`3oool0okmd0000M02oooo;0?ooo`00o`3oool90?ooo`030?noM00004P0W=oo0080oooo
00@0omnL04P0000007@0_ooo0P3oool03`3OW4P0001d0;ooo`3oool0gia80018W03Oool0oooO09a8
001d_ol0gia8001d_`3ogi`0B01809cOo`020?ooo`0;0=nLB00007@0_ooo0?ooo`2oM000001809cO
o`3ogi`0B7Bo0;md0000M;l00P3oool0303o_g@004RL0=ooo`3oool0oono07@0001d_ol0oooo0?oO
W018M;l0_g@0001d_`80oooo0100okmd0018W03Oool0oooo0?oo_`1d04P0W=oO09a8B02Lgml0W4P0
04RLg`3o_g@007Bo0?noM00007@0_ooo1P3oool00`3o_g@0000000000002000000030000M02oool0
oooo0080oooo00`0oooO09a80000B9`0gooo0?oo_`2LB4P004RL0=ooo`3ogi`0B7Bo0;md0000M;l2
0?ooo`030?noM000B9`0gooo0080oooo00/0gia80000M02oool0oooo0;md000004P0W=oo0?ooo`3o
oml0W4P004RLg`030?ooo`030?oOW0180000B9cO00<0oooo0100_g@00000M02oool0oooo0?noM000
07@0_ooo0?ooo`3ooml0W4P007Boo`3OW4P007Bo0?oOW01804P0W=oob`3oool00?l0oooo2@3oool0
0`3ookl0M00004RLg`030?ooo`030?oOW018000007Bo00<0oooo00h0omnL04P000000000000007Bo
o`3ookl0M00000000018W=l0oooo0?noM0000000001809cOo`<0oooo00X0omnL04P000000000B9bo
09a80018W=l0oooo0?noM000000007Bo0P3oool00`3ogi`0B01809cOo`020?ooo`070;md0018W=l0
oooo0?ooo`3o_g@00000001d_`020?ooo`0@0?oOW01804P0W=oo0?oo_`1d0000000004RLg`3ooml0
W4P000000018W=l0oooo0?noM000000004RL0=ooo`H0oooo00<0oono07@0001d_ol00P3oool00`2o
M4P0W=oo0?ooo`040?ooo`0:0=nLB0000000001d0;noM00007@0_ooo0?ooo`3o_g@00000001d_`80
oooo00<0omnL04P0B02Lgol00`3oool0203ogi`0B00000000018W9`0M00004RLg`3ookl0M0000`00
00001P1d_ol0okmd00000000000004RL0=ooo`<0oooo00l0okmd000000000000000007Boo`3oool0
oono07@000000000B9cO0?ooo`3o_g@000000000B02Lgol0b`3oool00?l0oooo2P3oool00`2oM000
04RL0=ooo`030?ooo`030;md000007@0_ooo01L0oooo00<0_g@0001d_`3oool02P3oool00`3OW4P0
04RL0=ooo`0M0?ooo`060;md0000M;l0oooo0?ooo`3ogi`0M9cO7`3oool00`3o_g@0001d0;ooo`3S
0?ooo`00o`3oool:0?ooo`030=nLB00007@0_ooo00<0oooo00<0_g@00000M02oool05`3oool00`3o
_g@0001d0;ooo`0:0?ooo`030?oOW01807@0_ooo01d0oooo00<0gia80000M02oool0103oool00`3O
W9`0_ooo0?ooo`0L0?ooo`030=nLM02Lgol0oooo0>80oooo003o0?ooo`X0oooo00<0okmd0000001d
_ol00P3oool00`3o_g@00000001d_`0H0?ooo`030?oOW01804P0W=oo00X0oooo00<0oooO09a8001d
_ol07@3oool00`3o_g@0001809cOo`030?ooo`030?oo_`1d04P0W=oo0280oooo00@0oono07@00000
B9`0gooog03oool00?l0oooo2@3oool00`2oM00000000000000400000003001d_`3oool0oooo01L0
oooo00<0oooO09a80018W=l02`3oool00`3OW4P0B9cO0?ooo`0K0?ooo`030;md00000000000000H0
000000<0M;oo0?ooo`3oool0o`3oool10?ooo`00o`3ooolY0?ooo`040?oo_`1d00000000001d_`T0
oooo00@0oooO09a80000000007Boo`3ooolY0?ooo`00o`3ooooo0?ooof<0oooo003o0?ooool0oooo
H`3oool00?l0ooooo`3ooomS0?ooo`00o`3ooooo0?ooof<0oooo0000\
\>"],
  ImageRangeCache->{{{0, 608.563}, {304.5, 0}} -> {-9.44845, -4.6379, \
0.042416, 0.042416}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Driven pendulum with damping", "Subsection"],

Cell["\<\
For a wide range of initial conditions, the motion eventually settles upon a \
limiting orbit that depends only on the driving force.\
\>", "Text"],

Cell[BoxData[
    \(\(attractor\  = 
        pendulumPlot[0.1, 0.1, 0.5, 0, 0, 160, 200, MaxSteps \[Rule] 2000, 
          PlotPoints \[Rule] 100, PlotRange \[Rule] All, 
          PlotStyle \[Rule] RGBColor[1, 0, 0], 
          DisplayFunction \[Rule] Identity];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(plot1\  = 
        pendulumPlot[0.1, 0.1, 0.5, 0, 0, 0, 200, MaxSteps \[Rule] 2000, 
          PlotPoints \[Rule] 100, PlotRange \[Rule] All, 
          DisplayFunction \[Rule] Identity];\)\), "\n", 
    \(\(Show[{plot1, attractor}, 
        DisplayFunction \[Rule] $DisplayFunction];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .79815 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.618612 2.57555 0.425052 2.57555 [
[.1035 -0.0125 -13.8438 -15.6875 ]
[.1035 -0.0125 13.8438 0 ]
[.23228 -0.0125 -16.8438 -15.6875 ]
[.23228 -0.0125 16.8438 0 ]
[.36106 -0.0125 -13.8438 -15.6875 ]
[.36106 -0.0125 13.8438 0 ]
[.48984 -0.0125 -16.8438 -15.6875 ]
[.48984 -0.0125 16.8438 0 ]
[.61861 -0.0125 -5 -15.6875 ]
[.61861 -0.0125 5 0 ]
[.74739 -0.0125 -12.5 -15.6875 ]
[.74739 -0.0125 12.5 0 ]
[.87617 -0.0125 -9.5 -15.6875 ]
[.87617 -0.0125 9.5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.9688 -31.375 ]
[.5 -0.0125 21.9688 -15.6875 ]
[-0.0125 .03872 -33.6875 -7.84375 ]
[-0.0125 .03872 0 7.84375 ]
[-0.0125 .1675 -27.6875 -7.84375 ]
[-0.0125 .1675 0 7.84375 ]
[-0.0125 .29627 -33.6875 -7.84375 ]
[-0.0125 .29627 0 7.84375 ]
[-0.0125 .42505 -10 -7.84375 ]
[-0.0125 .42505 0 7.84375 ]
[-0.0125 .55383 -25 -7.84375 ]
[-0.0125 .55383 0 7.84375 ]
[-0.0125 .68261 -19 -7.84375 ]
[-0.0125 .68261 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .39908 -49.375 -20.4688 ]
[-0.01875 .39908 -33.6875 20.4688 ]
[ 0 .79815 .125 0 ]
[ 1 0 .125 0 ]
[.5 .81065 -49.7813 0 ]
[.5 .81065 49.7813 15.6875 ]
[ 0 0 0 0 ]
[ 1 .79815 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.1035 0 m
.1035 .00625 L
s
gsave
.1035 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.23228 0 m
.23228 .00625 L
s
gsave
.23228 -0.0125 -77.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.15) show
1.000 setlinewidth
grestore
.36106 0 m
.36106 .00625 L
s
gsave
.36106 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
.48984 0 m
.48984 .00625 L
s
gsave
.48984 -0.0125 -77.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.05) show
1.000 setlinewidth
grestore
.61861 0 m
.61861 .00625 L
s
gsave
.61861 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.74739 0 m
.74739 .00625 L
s
gsave
.74739 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.05) show
1.000 setlinewidth
grestore
.87617 0 m
.87617 .00625 L
s
gsave
.87617 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
.125 Mabswid
.12926 0 m
.12926 .00375 L
s
.15501 0 m
.15501 .00375 L
s
.18077 0 m
.18077 .00375 L
s
.20653 0 m
.20653 .00375 L
s
.25804 0 m
.25804 .00375 L
s
.28379 0 m
.28379 .00375 L
s
.30955 0 m
.30955 .00375 L
s
.3353 0 m
.3353 .00375 L
s
.38681 0 m
.38681 .00375 L
s
.41257 0 m
.41257 .00375 L
s
.43832 0 m
.43832 .00375 L
s
.46408 0 m
.46408 .00375 L
s
.51559 0 m
.51559 .00375 L
s
.54135 0 m
.54135 .00375 L
s
.5671 0 m
.5671 .00375 L
s
.59286 0 m
.59286 .00375 L
s
.64437 0 m
.64437 .00375 L
s
.67012 0 m
.67012 .00375 L
s
.69588 0 m
.69588 .00375 L
s
.72163 0 m
.72163 .00375 L
s
.77315 0 m
.77315 .00375 L
s
.7989 0 m
.7989 .00375 L
s
.82466 0 m
.82466 .00375 L
s
.85041 0 m
.85041 .00375 L
s
.90192 0 m
.90192 .00375 L
s
.92768 0 m
.92768 .00375 L
s
.95343 0 m
.95343 .00375 L
s
.97919 0 m
.97919 .00375 L
s
.07775 0 m
.07775 .00375 L
s
.05199 0 m
.05199 .00375 L
s
.02624 0 m
.02624 .00375 L
s
.00048 0 m
.00048 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.9688 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(position) show
102.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .03872 m
.00625 .03872 L
s
gsave
-0.0125 .03872 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.15) show
1.000 setlinewidth
grestore
0 .1675 m
.00625 .1675 L
s
gsave
-0.0125 .1675 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
0 .29627 m
.00625 .29627 L
s
gsave
-0.0125 .29627 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.05) show
1.000 setlinewidth
grestore
0 .42505 m
.00625 .42505 L
s
gsave
-0.0125 .42505 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .55383 m
.00625 .55383 L
s
gsave
-0.0125 .55383 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.05) show
1.000 setlinewidth
grestore
0 .68261 m
.00625 .68261 L
s
gsave
-0.0125 .68261 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .06448 m
.00375 .06448 L
s
0 .09023 m
.00375 .09023 L
s
0 .11599 m
.00375 .11599 L
s
0 .14174 m
.00375 .14174 L
s
0 .19325 m
.00375 .19325 L
s
0 .21901 m
.00375 .21901 L
s
0 .24476 m
.00375 .24476 L
s
0 .27052 m
.00375 .27052 L
s
0 .32203 m
.00375 .32203 L
s
0 .34779 m
.00375 .34779 L
s
0 .37354 m
.00375 .37354 L
s
0 .3993 m
.00375 .3993 L
s
0 .45081 m
.00375 .45081 L
s
0 .47656 m
.00375 .47656 L
s
0 .50232 m
.00375 .50232 L
s
0 .52807 m
.00375 .52807 L
s
0 .57958 m
.00375 .57958 L
s
0 .60534 m
.00375 .60534 L
s
0 .6311 m
.00375 .6311 L
s
0 .65685 m
.00375 .65685 L
s
0 .01296 m
.00375 .01296 L
s
0 .70836 m
.00375 .70836 L
s
0 .73412 m
.00375 .73412 L
s
0 .75987 m
.00375 .75987 L
s
0 .78563 m
.00375 .78563 L
s
.25 Mabswid
0 0 m
0 .79815 L
s
gsave
-0.01875 .39908 -123 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.4688 11.8438 translate
90 rotate
-81.4688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(velocity) show
99.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.1035 .7919 m
.1035 .79815 L
s
.23228 .7919 m
.23228 .79815 L
s
.36106 .7919 m
.36106 .79815 L
s
.48984 .7919 m
.48984 .79815 L
s
.61861 .7919 m
.61861 .79815 L
s
.74739 .7919 m
.74739 .79815 L
s
.87617 .7919 m
.87617 .79815 L
s
.125 Mabswid
.12926 .7944 m
.12926 .79815 L
s
.15501 .7944 m
.15501 .79815 L
s
.18077 .7944 m
.18077 .79815 L
s
.20653 .7944 m
.20653 .79815 L
s
.25804 .7944 m
.25804 .79815 L
s
.28379 .7944 m
.28379 .79815 L
s
.30955 .7944 m
.30955 .79815 L
s
.3353 .7944 m
.3353 .79815 L
s
.38681 .7944 m
.38681 .79815 L
s
.41257 .7944 m
.41257 .79815 L
s
.43832 .7944 m
.43832 .79815 L
s
.46408 .7944 m
.46408 .79815 L
s
.51559 .7944 m
.51559 .79815 L
s
.54135 .7944 m
.54135 .79815 L
s
.5671 .7944 m
.5671 .79815 L
s
.59286 .7944 m
.59286 .79815 L
s
.64437 .7944 m
.64437 .79815 L
s
.67012 .7944 m
.67012 .79815 L
s
.69588 .7944 m
.69588 .79815 L
s
.72163 .7944 m
.72163 .79815 L
s
.77315 .7944 m
.77315 .79815 L
s
.7989 .7944 m
.7989 .79815 L
s
.82466 .7944 m
.82466 .79815 L
s
.85041 .7944 m
.85041 .79815 L
s
.90192 .7944 m
.90192 .79815 L
s
.92768 .7944 m
.92768 .79815 L
s
.95343 .7944 m
.95343 .79815 L
s
.97919 .7944 m
.97919 .79815 L
s
.07775 .7944 m
.07775 .79815 L
s
.05199 .7944 m
.05199 .79815 L
s
.02624 .7944 m
.02624 .79815 L
s
.00048 .7944 m
.00048 .79815 L
s
.25 Mabswid
0 .79815 m
1 .79815 L
s
.99375 .03872 m
1 .03872 L
s
.99375 .1675 m
1 .1675 L
s
.99375 .29627 m
1 .29627 L
s
.99375 .42505 m
1 .42505 L
s
.99375 .55383 m
1 .55383 L
s
.99375 .68261 m
1 .68261 L
s
.125 Mabswid
.99625 .06448 m
1 .06448 L
s
.99625 .09023 m
1 .09023 L
s
.99625 .11599 m
1 .11599 L
s
.99625 .14174 m
1 .14174 L
s
.99625 .19325 m
1 .19325 L
s
.99625 .21901 m
1 .21901 L
s
.99625 .24476 m
1 .24476 L
s
.99625 .27052 m
1 .27052 L
s
.99625 .32203 m
1 .32203 L
s
.99625 .34779 m
1 .34779 L
s
.99625 .37354 m
1 .37354 L
s
.99625 .3993 m
1 .3993 L
s
.99625 .45081 m
1 .45081 L
s
.99625 .47656 m
1 .47656 L
s
.99625 .50232 m
1 .50232 L
s
.99625 .52807 m
1 .52807 L
s
.99625 .57958 m
1 .57958 L
s
.99625 .60534 m
1 .60534 L
s
.99625 .6311 m
1 .6311 L
s
.99625 .65685 m
1 .65685 L
s
.99625 .01296 m
1 .01296 L
s
.99625 .70836 m
1 .70836 L
s
.99625 .73412 m
1 .73412 L
s
.99625 .75987 m
1 .75987 L
s
.99625 .78563 m
1 .78563 L
s
.25 Mabswid
1 0 m
1 .79815 L
s
0 .42505 m
1 .42505 L
s
.61861 0 m
.61861 .79815 L
s
gsave
.5 .81065 -110.781 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Pendulum) show
114.625 15.813 moveto
(Equation) show
158.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.5 Mabswid
.61861 .42505 m
.61907 .44041 L
.62029 .45424 L
.62262 .46998 L
.62572 .48457 L
.62936 .49775 L
.63335 .5096 L
.64453 .53514 L
.65864 .55873 L
.67717 .58178 L
.6994 .60208 L
.72244 .61722 L
.7346 .62328 L
.74826 .6287 L
.76146 .63265 L
.77391 .63527 L
.78636 .63686 L
.79771 .63744 L
.81081 .63706 L
.82311 .6357 L
.83644 .63307 L
.85029 .62901 L
.86188 .62452 L
.87426 .61852 L
.88713 .61082 L
.9003 .60118 L
.92209 .58029 L
.93245 .56751 L
.94243 .5527 L
.95744 .52318 L
.96401 .50533 L
.96895 .48772 L
.97109 .478 L
.97294 .46755 L
.97451 .45585 L
.9755 .44493 L
.97619 .4263 L
.97604 .41568 L
.97557 .40594 L
.97372 .38674 L
.97082 .36839 L
.96602 .3468 L
.96023 .3268 L
.9423 .28172 L
.92254 .24525 L
.90007 .21219 L
.86855 .17481 L
.83545 .14304 L
.75657 .08728 L
.70759 .06281 L
Mistroke
.65824 .04426 L
.63378 .03709 L
.6069 .03065 L
.58115 .02583 L
.55725 .02252 L
.53546 .02044 L
.51147 .01918 L
.48639 .019 L
.46302 .01989 L
.44989 .02084 L
.43567 .02224 L
.40999 .02574 L
.38524 .03033 L
.36266 .0356 L
.31272 .05114 L
.26231 .07284 L
.22444 .09379 L
.18646 .11965 L
.12632 .17412 L
.09715 .20954 L
.07141 .24921 L
.05365 .28462 L
.03887 .32448 L
.03284 .34678 L
.02861 .36745 L
.02692 .37829 L
.02546 .39021 L
.02439 .40265 L
.02382 .41427 L
.02381 .4356 L
.02499 .45494 L
.02613 .46555 L
.02776 .47704 L
.03176 .4977 L
.03641 .51576 L
.04273 .53534 L
.05923 .5736 L
.07875 .60704 L
.10056 .63639 L
.13021 .66801 L
.16566 .69756 L
.19893 .71945 L
.23807 .73968 L
.28293 .75693 L
.30899 .76446 L
.33315 .76995 L
.35601 .77392 L
.369 .77567 L
.38105 .77696 L
.40328 .77856 L
Mistroke
.4269 .77915 L
.44864 .77871 L
.4723 .77716 L
.49452 .77471 L
.51497 .77159 L
.55299 .76357 L
.59235 .75212 L
.63677 .73509 L
.67622 .71594 L
.74249 .67354 L
.79516 .62746 L
.83378 .58249 L
.86498 .53281 L
.87631 .50847 L
.88149 .49496 L
.88548 .48278 L
.88864 .47122 L
.89132 .45889 L
.89232 .45308 L
.8932 .44695 L
.89389 .44064 L
.89433 .43485 L
.89462 .42453 L
.89453 .41954 L
.89424 .41424 L
.89328 .40528 L
.89189 .39745 L
.88995 .3897 L
.88768 .38292 L
.88246 .37169 L
.87903 .36623 L
.87558 .36174 L
.87121 .35711 L
.86697 .35352 L
.86248 .35051 L
.85748 .34798 L
.85272 .34625 L
.84811 .34517 L
.84271 .34458 L
.83769 .34466 L
.83481 .34496 L
.83221 .34541 L
.82923 .3461 L
.82646 .34694 L
.82131 .34899 L
.81604 .35177 L
.8072 .35825 L
.80266 .36262 L
.79872 .36712 L
.79473 .37252 L
Mistroke
.79094 .37869 L
.78818 .3841 L
.78554 .39032 L
.7835 .39624 L
.78197 .40179 L
.78059 .40844 L
.78003 .41214 L
.77962 .41567 L
.77922 .42194 L
.77917 .42548 L
.77924 .42875 L
.77969 .43506 L
.7805 .4409 L
.78153 .44601 L
.78296 .45138 L
.78488 .45702 L
.78731 .46266 L
.79259 .47183 L
.79826 .47888 L
.8017 .48222 L
.80513 .48499 L
.80846 .48721 L
.81221 .48923 L
.81571 .49067 L
.81901 .49168 L
.82296 .49244 L
.82669 .49274 L
.83078 .49258 L
.83517 .49183 L
.83946 .49049 L
.84158 .48959 L
.84386 .48843 L
.85128 .48309 L
.85505 .47926 L
.85828 .47516 L
.86146 .47016 L
.86454 .46397 L
.86683 .45802 L
.86862 .45199 L
.8701 .44515 L
.87074 .4411 L
.8712 .43728 L
.87171 .42982 L
.87178 .42255 L
.87137 .41423 L
.87036 .40499 L
.86882 .39583 L
.86684 .38695 L
.86171 .36975 L
.85316 .34864 L
Mistroke
.84112 .32578 L
.81168 .2846 L
.76891 .24223 L
.73983 .22016 L
.70985 .20142 L
.68119 .1866 L
.64891 .17299 L
.61163 .16087 L
.59452 .1565 L
.57581 .15252 L
.5581 .14951 L
.54172 .14737 L
.52596 .14588 L
.50931 .14492 L
.49094 .14458 L
.47386 .14495 L
.45518 .14611 L
.44477 .14711 L
.43535 .14823 L
.41907 .15067 L
.40202 .15392 L
.37207 .16146 L
.34092 .17196 L
.30832 .18623 L
.28182 .20075 L
.25535 .21842 L
.21352 .25521 L
.19299 .27929 L
.1747 .30652 L
.16213 .33082 L
.15164 .3585 L
.14737 .37419 L
.14445 .38869 L
.14331 .39631 L
.14237 .40467 L
.14173 .41339 L
.14145 .4215 L
.14146 .42921 L
.14172 .43648 L
.14286 .45001 L
.14387 .45754 L
.14504 .46444 L
.14849 .4799 L
.15208 .4921 L
.15674 .505 L
.16718 .52761 L
.18289 .55309 L
.20019 .57483 L
.2209 .59559 L
.2464 .61592 L
Mistroke
.27307 .63277 L
.29946 .64605 L
.32897 .65768 L
.34474 .6627 L
.3624 .66741 L
.377 .67065 L
.39284 .67351 L
.40975 .67586 L
.42565 .67743 L
.44116 .67839 L
.45807 .67882 L
.47404 .67863 L
.4888 .67797 L
.50202 .67697 L
.51653 .67544 L
.53158 .67339 L
.54559 .67104 L
.57729 .66417 L
.60956 .65491 L
.66358 .63395 L
.71532 .60663 L
.76311 .57349 L
.7988 .54187 L
.8248 .51319 L
.84235 .48932 L
.85529 .46731 L
.86002 .45744 L
.86385 .44792 L
.86629 .44042 L
.86774 .43463 L
.86832 .43162 L
.86867 .42909 L
.8688 .42785 L
.86887 .42678 L
.86892 .42569 L
.86892 .42464 L
.86889 .42381 L
.86883 .423 L
.86874 .42232 L
.86865 .42176 L
.86841 .42087 L
.86812 .42017 L
.86779 .41971 L
.8676 .41953 L
.86742 .41942 L
.86708 .41935 L
.86672 .41943 L
.86641 .41963 L
.86613 .41991 L
.86584 .42031 L
Mistroke
.86556 .42084 L
.8653 .42149 L
.86507 .42227 L
.86499 .42265 L
.86491 .42306 L
.8648 .42381 L
.86475 .42459 L
.86474 .42532 L
.86477 .42593 L
.86485 .42655 L
.86495 .42707 L
.86507 .42748 L
.86523 .42785 L
.86542 .42812 L
.86563 .42825 L
.86583 .42822 L
.86601 .42804 L
.86618 .42768 L
.86631 .42723 L
.86642 .42655 L
.86649 .42564 L
.86649 .4246 L
.86644 .42345 L
.86633 .42225 L
.86614 .42076 L
.86584 .41899 L
.86475 .41418 L
.86307 .40864 L
.85784 .39567 L
.84843 .37817 L
.83281 .35565 L
.81387 .33376 L
.788 .30939 L
.72637 .26657 L
.68395 .24551 L
.6409 .22948 L
.61569 .2223 L
.58726 .216 L
.57399 .21371 L
.55971 .21168 L
.54613 .21017 L
.53374 .20916 L
.5211 .20849 L
.50772 .20816 L
.49629 .20821 L
.48377 .2086 L
.47031 .20944 L
.45596 .21081 L
.44252 .21255 L
.43038 .21452 L
Mistroke
.40722 .21937 L
.38274 .22616 L
.34073 .24246 L
.31963 .25329 L
.29844 .26644 L
.26473 .29371 L
.24824 .31129 L
.23353 .33095 L
.22322 .34852 L
.21443 .36835 L
.21073 .37942 L
.20797 .38996 L
.20611 .39949 L
.20476 .40983 L
.20433 .41522 L
.20407 .42114 L
.20404 .42731 L
.20423 .43307 L
.2051 .44333 L
.20588 .44906 L
.20678 .45427 L
.20903 .46428 L
.2117 .47344 L
.21978 .49371 L
.22918 .51089 L
.24142 .52829 L
.25633 .54515 L
.27439 .56156 L
.31206 .587 L
.33561 .59879 L
.35893 .60815 L
.3858 .61657 L
.40123 .6204 L
.41572 .6234 L
.42831 .62557 L
.44232 .62751 L
.45653 .62901 L
.46434 .62964 L
.47168 .63011 L
.48519 .63066 L
.49758 .63083 L
.50441 .63079 L
.51187 .63064 L
.52543 .63008 L
.53862 .62919 L
.55052 .62809 L
.57691 .6247 L
.6009 .62048 L
.62298 .61563 L
.66672 .6033 L
Mistroke
.711 .58696 L
.78627 .54925 L
.83803 .51419 L
.87177 .48531 L
.88531 .47141 L
.89629 .4582 L
.90351 .44733 L
.906 .44259 L
.9078 .43833 L
.909 .43465 L
.90951 .43263 L
.90987 .43078 L
.91013 .42889 L
.9103 .42683 L
.91035 .42468 L
.91027 .42267 L
.91005 .42051 L
.90972 .41846 L
.90931 .41659 L
.90876 .41454 L
.90708 .40981 L
.90496 .40524 L
.89912 .39551 L
.8901 .38378 L
.86549 .35892 L
.81984 .3242 L
.76088 .29106 L
.72643 .27605 L
.68594 .26185 L
.665 .25585 L
.64464 .25086 L
.62603 .24701 L
.60521 .24348 L
.58164 .24049 L
.56863 .23929 L
.55643 .23846 L
.54496 .23795 L
.53449 .23771 L
.52239 .23771 L
.51095 .23797 L
.49857 .23857 L
.49234 .23899 L
.4855 .23955 L
.46224 .24223 L
.44934 .24426 L
.43563 .24687 L
.4114 .25268 L
.39066 .25901 L
.36947 .26695 L
.33423 .28419 L
Mistroke
.31482 .29646 L
.29821 .30913 L
.27204 .33529 L
.26126 .34979 L
.25196 .36573 L
.24579 .37974 L
.24302 .38781 L
.24093 .39542 L
.23998 .39965 L
.23911 .40427 L
.23849 .40835 L
.23797 .41269 L
.23742 .42091 L
.2374 .42852 L
.23759 .43271 L
.23795 .43722 L
.23904 .44536 L
.23991 .45005 L
.24102 .45497 L
.24351 .46369 L
.24957 .4791 L
.25712 .49324 L
.2758 .51829 L
.28871 .53133 L
.30208 .54265 L
.33698 .56537 L
.36896 .58042 L
.40753 .5935 L
.42801 .59866 L
.45064 .60314 L
.46179 .60491 L
.47248 .60636 L
.483 .60755 L
.49274 .60845 L
.50439 .60929 L
.51054 .60962 L
.51705 .6099 L
.52801 .6102 L
.54005 .61028 L
.55079 .61014 L
.56227 .60976 L
.57309 .60921 L
.58292 .60854 L
.60628 .60631 L
.6313 .60294 L
.67533 .59458 L
.71418 .58458 L
.75262 .57213 L
.81315 .54659 L
.86424 .51739 L
Mistroke
.88545 .50191 L
.90359 .48578 L
.91605 .47181 L
.92134 .46451 L
.926 .45678 L
.92941 .44969 L
.93176 .44329 L
.93264 .44021 L
.93342 .43687 L
.93397 .43373 L
.93433 .43083 L
.93458 .42713 L
.9346 .42362 L
.93444 .42047 L
.93407 .41703 L
.93343 .41329 L
.93249 .40931 L
.93011 .40213 L
.92679 .3948 L
.92283 .38787 L
.91324 .3748 L
.89792 .35894 L
.88015 .34428 L
.85676 .32837 L
.82713 .31172 L
.7957 .29709 L
.76389 .28473 L
.72869 .2735 L
.69395 .26464 L
.65123 .25656 L
.62797 .25342 L
.61594 .25214 L
.6029 .25102 L
.59145 .25027 L
.58094 .24977 L
.57035 .24945 L
.56026 .24933 L
.54858 .2494 L
.53774 .24968 L
.53151 .24993 L
.52587 .25022 L
.51326 .25108 L
.50266 .25203 L
.49282 .25311 L
.47071 .25625 L
.44961 .26023 L
.42761 .26549 L
.38987 .2776 L
.35836 .29143 L
.32597 .31057 L
Mistroke
.311 .32185 L
.29685 .33464 L
.28608 .34637 L
.27597 .35987 L
.26845 .37257 L
.26288 .38479 L
.26066 .39096 L
.25869 .39761 L
.25722 .40391 L
.25622 .40967 L
.25545 .41646 L
.25521 .42029 L
.25511 .42383 L
.25523 .43034 L
.2558 .43721 L
.25677 .44377 L
.25826 .45078 L
.2601 .45735 L
.26214 .46332 L
.26857 .47768 L
.27654 .49094 L
.29454 .51287 L
.31564 .53163 L
.34324 .55011 L
.37714 .56701 L
.39653 .57464 L
.41803 .58172 L
.45873 .59182 L
.48237 .59599 L
.50487 .59895 L
.51795 .60024 L
.52994 .60115 L
.5366 .60156 L
.54385 .60191 L
.55071 .60217 L
.55705 .60234 L
.5675 .60248 L
.57875 .60244 L
.58936 .60221 L
.59907 .60186 L
.60981 .6013 L
.62151 .60049 L
.64487 .59826 L
.66765 .59529 L
.69147 .59133 L
.73275 .58235 L
.77394 .57043 L
.80879 .55769 L
.84252 .54243 L
.87462 .52424 L
Mistroke
.89662 .50864 L
.91598 .4913 L
.92473 .48153 L
.93149 .47252 L
.93673 .46406 L
.94132 .45467 L
.94298 .45041 L
.94448 .4458 L
.94554 .44181 L
.94644 .43744 L
.94711 .43278 L
.94734 .43023 L
.94747 .42786 L
.94751 .42361 L
.94727 .41907 L
.94679 .41481 L
.94597 .41019 L
.94491 .40583 L
.9437 .40181 L
.94079 .39423 L
.93681 .38623 L
.9244 .36828 L
.90859 .35184 L
.88672 .33442 L
.85815 .31666 L
.7949 .28849 L
.75316 .27539 L
.73195 .27002 L
.71132 .26556 L
.69119 .26189 L
.66894 .25858 L
.64838 .2562 L
.62924 .25457 L
.61843 .25389 L
.6083 .25342 L
.59716 .25308 L
.59091 .25297 L
.58521 .25292 L
.57373 .25299 L
.56725 .25311 L
.56126 .25329 L
.55029 .25376 L
.54477 .25407 L
.53868 .25447 L
.51838 .25627 L
.49648 .25901 L
.47406 .26275 L
.4537 .26704 L
.41875 .27661 L
.38414 .28943 L
Mistroke
.34786 .30768 L
.3188 .3277 L
.29757 .34762 L
.28779 .35953 L
.28023 .37086 L
.2742 .3823 L
.26995 .39294 L
.26812 .39888 L
.26656 .4054 L
.26549 .41158 L
.26485 .41732 L
.26456 .42313 L
.2646 .4286 L
.26491 .43351 L
.26552 .43881 L
.26653 .44458 L
.26797 .45065 L
.2715 .46143 L
.27679 .4731 L
.28406 .48532 L
.30122 .50645 L
.31416 .51865 L
.32802 .5296 L
.35791 .54836 L
.38857 .56298 L
.42601 .57643 L
.44606 .58209 L
.46841 .58731 L
.48826 .59108 L
.51042 .59442 L
.5206 .59566 L
.53154 .5968 L
.54105 .59763 L
.55139 .59836 L
.55751 .59872 L
.56316 .59899 L
.57428 .59939 L
.58441 .59958 L
.59512 .59962 L
.60443 .59951 L
.6144 .59925 L
.62379 .59887 L
.63234 .59841 L
.65315 .59684 L
.67267 .59478 L
.69291 .59203 L
.7116 .58892 L
.75271 .58003 L
.79029 .56916 L
.82747 .55529 L
Mistroke
.85686 .54151 L
.8841 .52562 L
.90551 .50998 L
.92234 .49447 L
.93026 .48554 L
.93761 .47569 L
.94272 .46738 L
.94729 .45814 L
.95096 .44806 L
.95241 .44252 L
.95341 .43732 L
.95403 .43245 L
.95432 .428 L
.95435 .42286 L
.95406 .41798 L
.95351 .41342 L
.95278 .40928 L
.95166 .40452 L
.95034 .40005 L
.94734 .39212 L
.94319 .38362 L
.93024 .36466 L
.91415 .34784 L
.89367 .3314 L
.86749 .31491 L
.81028 .28897 L
.76991 .27607 L
.72866 .2663 L
.68967 .25977 L
.66704 .25709 L
.65674 .25612 L
.64583 .25528 L
.63582 .25466 L
.62483 .25415 L
.61482 .25384 L
.6055 .25368 L
.59444 .25367 L
.58441 .25381 L
.57886 .25396 L
.57284 .25417 L
.56196 .2547 L
.55253 .2553 L
.5422 .25612 L
.52152 .25828 L
.50264 .26089 L
.48528 .26385 L
.44384 .27333 L
.4085 .28455 L
.37147 .30033 L
.34075 .31784 L
Mistroke
.31559 .33692 L
.29723 .35564 L
.28897 .36652 L
.28164 .37858 L
.27628 .39021 L
.2744 .39544 L
.27272 .40106 L
.27147 .40638 L
.27059 .41125 L
.26998 .41615 L
.26965 .4208 L
.26955 .42622 L
.26976 .43122 L
.27029 .4367 L
.27122 .4425 L
.27236 .44766 L
.27391 .45318 L
.27758 .46308 L
.28335 .47469 L
.28996 .4851 L
.30699 .50528 L
.33036 .52533 L
.35695 .54253 L
.38452 .55645 L
.42069 .57057 L
.44111 .57691 L
.46356 .58277 L
.48389 .58716 L
.50678 .59116 L
.52804 .59405 L
.54818 .5961 L
.55918 .59694 L
.57092 .59763 L
.58097 .59806 L
.5921 .59835 L
.59831 .59844 L
.60402 .59846 L
.61028 .59844 L
.61687 .59834 L
.62797 .59805 L
.63425 .5978 L
.64008 .59751 L
.65138 .59682 L
.662 .59599 L
.68172 .59401 L
.70266 .59124 L
.72492 .58753 L
.76275 .57925 L
.80038 .56825 L
.83011 .5572 L
Mistroke
.85988 .54345 L
.88906 .52639 L
.91239 .50879 L
.92945 .49205 L
.94363 .47301 L
.94958 .46191 L
.95362 .45173 L
.9553 .44604 L
.95602 .44304 L
.95666 .43987 L
.95712 .43709 L
.95752 .43403 L
.95793 .42866 L
.95799 .42322 L
.95773 .41826 L
.95745 .41552 L
.95705 .41255 L
.95604 .40716 L
.95477 .40213 L
.95308 .39681 L
.94916 .38734 L
.94399 .37774 L
.9373 .3678 L
.92223 .35056 L
.90172 .33292 L
.87451 .31504 L
.84715 .30091 L
.81451 .28751 L
.78086 .27666 L
.74829 .26847 L
.70858 .2611 L
.68795 .2583 L
.67595 .25699 L
.66493 .25597 L
.65544 .25524 L
.64507 .25459 L
.63407 .25409 L
.62362 .25377 L
.61385 .25362 L
.60345 .25361 L
.59458 .25373 L
.58479 .254 L
.5734 .2545 L
.56144 .25523 L
.5402 .25708 L
.51765 .25983 L
.49345 .26374 L
.45176 .27306 L
.41548 .28428 L
.38114 .29831 L
Mistroke
.34828 .31602 L
.32207 .33481 L
.30244 .35365 L
.28878 .3716 L
.28279 .38213 L
.27838 .39215 L
.27649 .39761 L
.275 .40283 L
.27394 .40753 L
.27307 .41268 L
.27246 .41827 L
.27227 .4214 L
.2722 .42424 L
.27232 .42977 L
.27276 .43497 L
.2735 .4402 L
.27464 .44578 L
.27603 .45104 L
.27758 .4558 L
.28247 .46737 L
.28851 .47814 L
.30449 .49886 L
.32762 .51999 L
.35214 .53679 L
.38232 .55283 L
.41109 .56483 L
.44561 .57611 L
.48468 .58559 L
.50514 .58937 L
.52801 .59273 L
.54795 .59494 L
.559 .5959 L
.56948 .59664 L
.5793 .59717 L
.58996 .59759 L
.59967 .59782 L
.60863 .59791 L
.61921 .59786 L
.63057 .59762 L
.64126 .59723 L
.65101 .59672 L
.66237 .59595 L
.67469 .59489 L
.69688 .5924 L
.71963 .58903 L
.74327 .58461 L
.78392 .57459 L
.82313 .56151 L
.85571 .54737 L
.8871 .5298 L
Mistroke
.91466 .5093 L
.93363 .49012 L
.94682 .4714 L
.95237 .46038 L
.95611 .45023 L
.95764 .44457 L
.95829 .44158 L
.95886 .43841 L
.95926 .43563 L
.95959 .43257 L
.95991 .42717 L
.95987 .42172 L
.95951 .41666 L
.95876 .41111 L
.95819 .40802 L
.95756 .40522 L
.95617 .40009 L
.95431 .39456 L
.94998 .38462 L
.94458 .37501 L
.93763 .36503 L
.92208 .34772 L
.89933 .32881 L
.87377 .31251 L
.84136 .29644 L
.80215 .28163 L
.76204 .27037 L
.72253 .26236 L
.70121 .25916 L
.67724 .25643 L
.66507 .25538 L
.65186 .2545 L
.64577 .25418 L
.63934 .2539 L
.62782 .25356 L
.61682 .25341 L
.60681 .25343 L
.59527 .25364 L
.58439 .25401 L
.57434 .25452 L
.5651 .25512 L
.54437 .25694 L
.52372 .25943 L
.50139 .26292 L
.45798 .2723 L
.41901 .28414 L
.3852 .29774 L
.35798 .31171 L
.3321 .32864 L
.30899 .34868 L
Mistroke
.29846 .36051 L
.29048 .37155 L
.28437 .38211 L
.2817 .38775 L
.27928 .39374 L
.27751 .39903 L
.27595 .40485 L
.27487 .41018 L
.27416 .41515 L
.27371 .42075 L
.27361 .42358 L
.27361 .4267 L
.27389 .43221 L
.27447 .43737 L
.27494 .44032 L
.27554 .44342 L
.27689 .44894 L
.27874 .45486 L
.28122 .46129 L
.28648 .47198 L
.29357 .48319 L
.31009 .50267 L
.33235 .52186 L
.35766 .5385 L
.38363 .55202 L
.41487 .56495 L
.45126 .57657 L
.48764 .58525 L
.50934 .58923 L
.52957 .5922 L
.54089 .59356 L
.55332 .59483 L
.56427 .59574 L
.57614 .59652 L
.58267 .59686 L
.58867 .59712 L
.59442 .59732 L
.60048 .59747 L
.61102 .5976 L
.62257 .59756 L
.62905 .59745 L
.63494 .5973 L
.64173 .59706 L
.64813 .59677 L
.6595 .59611 L
.67178 .59517 L
.69377 .59292 L
.71671 .58975 L
.75657 .58213 L
.79465 .57201 L
Mistroke
.82704 .56078 L
.86095 .54571 L
.88926 .52951 L
.91153 .51324 L
.92873 .49715 L
.94353 .47861 L
.94988 .46794 L
.95441 .45812 L
.95774 .44827 L
.95911 .44264 L
.96005 .43738 L
.96068 .4318 L
.96093 .42672 L
.96094 .42391 L
.96084 .42085 L
.96038 .41529 L
.9597 .41056 L
.95864 .40536 L
.95719 .39994 L
.95551 .39488 L
.95174 .38591 L
.9466 .37636 L
.93059 .35502 L
.9204 .34472 L
.90769 .33385 L
.88343 .31705 L
.85274 .30057 L
.81722 .2859 L
.78035 .27429 L
.74191 .26529 L
.70474 .25912 L
.68343 .25661 L
.66324 .25489 L
.65158 .25417 L
.64089 .25369 L
.63498 .2535 L
.62852 .25335 L
.62242 .25326 L
.61676 .25323 L
.61083 .25325 L
.60442 .25333 L
.59838 .25345 L
.59283 .25362 L
.58231 .25406 L
.57125 .2547 L
.55938 .25559 L
.54838 .25662 L
.52368 .25961 L
.50192 .26308 L
.48226 .26694 L
Mistroke
.43979 .27786 L
.40517 .28987 L
.37195 .30485 L
.34578 .31996 L
.32458 .33542 L
.30533 .3536 L
.29706 .36349 L
.28961 .3743 L
.2836 .38534 L
.27943 .39552 L
.27786 .40047 L
.27648 .40586 L
.27549 .41096 L
.27485 .41567 L
.27442 .42118 L
.27435 .42634 L
.27462 .43199 L
.27493 .43516 L
.27532 .43805 L
.27639 .44384 L
.278 .45013 L
.2821 .46147 L
.28748 .47224 L
.29501 .48392 L
.31524 .50663 L
.33961 .52618 L
.36706 .54293 L
.39691 .55716 L
.43321 .57056 L
.45277 .57639 L
.47449 .58187 L
.49384 .58597 L
.51535 .58974 L
.53728 .59276 L
.54702 .59386 L
.55782 .59489 L
.56736 .59566 L
.57767 .59633 L
.58859 .59687 L
.59884 .59721 L
.60874 .59739 L
.61954 .59743 L
.62974 .5973 L
.6392 .59705 L
.64855 .59667 L
.657 .59621 L
.67584 .59481 L
.69584 .59272 L
.71744 .58975 L
.75965 .58165 L
Mistroke
.79654 .57175 L
.82857 .56059 L
.86286 .54523 L
.89029 .52941 L
.91689 .50932 L
.92778 .49891 L
.93667 .48883 L
.94939 .46997 L
.95434 .45971 L
.95638 .45437 L
.95823 .44851 L
.95964 .44271 L
.9606 .4373 L
.96098 .4343 L
.96123 .43157 L
.96142 .42842 L
.96148 .42545 L
.96134 .42016 L
.96092 .41528 L
.96013 .40994 L
.95894 .4043 L
.95759 .39941 L
.95611 .39496 L
.95191 .38502 L
.94697 .37593 L
.94136 .36744 L
.92564 .34903 L
.90372 .33011 L
.87538 .31169 L
.8444 .29621 L
.81342 .28405 L
.77708 .27298 L
.7345 .26351 L
.71221 .25986 L
.68727 .25675 L
.66561 .25484 L
.65338 .25407 L
.64221 .25357 L
.63178 .25327 L
.62043 .25312 L
.60971 .25315 L
.59981 .25334 L
.58894 .2537 L
.57862 .25421 L
.56939 .25481 L
.55935 .2556 L
.5371 .25791 L
.51674 .26071 L
.4808 .26741 L
.44436 .27672 L
Mistroke
.40422 .29055 L
.37031 .30606 L
.34363 .3218 L
.31923 .34047 L
.30047 .3598 L
.28743 .37876 L
.28198 .38994 L
.27832 .40028 L
.27684 .40603 L
.27621 .40908 L
.27567 .41232 L
.2753 .41515 L
.27499 .41827 L
.27474 .42379 L
.27474 .42656 L
.27485 .42959 L
.27507 .43276 L
.27539 .4357 L
.2762 .44098 L
.27744 .44659 L
.27936 .45317 L
.28163 .45934 L
.2871 .47077 L
.29372 .48142 L
.30103 .49104 L
.3217 .51189 L
.3444 .52896 L
.37312 .54567 L
.40393 .55966 L
.43485 .57073 L
.47383 .58141 L
.497 .5863 L
.51907 .59006 L
.54072 .59295 L
.56062 .59495 L
.57175 .5958 L
.57762 .59617 L
.58383 .5965 L
.59528 .59697 L
.60583 .59724 L
.61656 .59733 L
.62262 .59731 L
.62822 .59724 L
.63849 .597 L
.64936 .59657 L
.65905 .59604 L
.6681 .59542 L
.68859 .59355 L
.71034 .59085 L
.72992 .58775 L
Mistroke
.77245 .57866 L
.80486 .56928 L
.83773 .55709 L
.86871 .54243 L
.89405 .52722 L
.91802 .50868 L
.92867 .49837 L
.93838 .48712 L
.94554 .47701 L
.95193 .46569 L
.95637 .45523 L
.95806 .45011 L
.95933 .44535 L
.96038 .44038 L
.96119 .43496 L
.96161 .43017 L
.96176 .42507 L
.96164 .42053 L
.96131 .41626 L
.9607 .41159 L
.95975 .40655 L
.9585 .40153 L
.95679 .39606 L
.95227 .38511 L
.94681 .37512 L
.94065 .36597 L
.92291 .34595 L
.89805 .32565 L
.8696 .30818 L
.83948 .29379 L
.80953 .28245 L
.77391 .27192 L
.75509 .26744 L
.73429 .26328 L
.71411 .25997 L
.69549 .25751 L
.6737 .25533 L
.66164 .25444 L
.65022 .2538 L
.63959 .25338 L
.6298 .25314 L
.61903 .25304 L
.60766 .25311 L
.59618 .25339 L
.58975 .25362 L
.5839 .25389 L
.57255 .25456 L
.56186 .25537 L
.53814 .25779 L
.51279 .26138 L
Mistroke
.47162 .26958 L
.43059 .28113 L
.39832 .29312 L
.36623 .30843 L
.33909 .32513 L
.31525 .34436 L
.29905 .36189 L
.29169 .3721 L
.28595 .38194 L
.28325 .38751 L
.28075 .3936 L
.27901 .39869 L
.27745 .40432 L
.27618 .41044 L
.27567 .4138 L
.27531 .41694 L
.27497 .42257 L
.27493 .42573 L
.275 .42861 L
.27542 .4342 L
.27613 .43936 L
.27726 .44502 L
.27889 .45102 L
.283 .46205 L
.28797 .4719 L
.30217 .49207 L
.31119 .50186 L
.32246 .51226 L
.34611 .52983 L
.37126 .54448 L
.40502 .55989 L
.44532 .57375 L
.4644 .57895 L
.48582 .58391 L
.50508 .58763 L
.52627 .59098 L
.54791 .59363 L
.55747 .59456 L
.56798 .59543 L
.57733 .59606 L
.58733 .59658 L
.59793 .59698 L
.60779 .5972 L
.61394 .59726 L
.6196 .59727 L
.62581 .59722 L
.63237 .59711 L
.6434 .59679 L
.64965 .59652 L
.65547 .59622 L
Mistroke
.67617 .59475 L
.68769 .59364 L
.69813 .59246 L
.74134 .58567 L
.76486 .58059 L
.7859 .57512 L
.82234 .56327 L
.85747 .5483 L
.88616 .53249 L
.90886 .51657 L
.92944 .49777 L
.93807 .48775 L
.94586 .47678 L
.9515 .46689 L
.95558 .45776 L
.95719 .45332 L
.95868 .44849 L
.95981 .44394 L
.96066 .43967 L
.96136 .4347 L
.96181 .42922 L
.9619 .42387 L
.96181 .42087 L
.96163 .41813 L
.96109 .41325 L
.96019 .40795 L
.95904 .40295 L
.9577 .39833 L
.95333 .38703 L
.94798 .37678 L
.94066 .36573 L
.93131 .35436 L
.91198 .33604 L
.88585 .3174 L
.85856 .30233 L
.82559 .28806 L
.7914 .27659 L
.75805 .26798 L
.71859 .26054 L
.69805 .25772 L
.68609 .2564 L
.67505 .25537 L
.66562 .25464 L
.65527 .25399 L
.64483 .25351 L
.63383 .25316 L
.62408 .25301 L
.61367 .25301 L
.60384 .25314 L
.59493 .25339 L
Mistroke
.58361 .25388 L
.57168 .2546 L
.55044 .25642 L
.52789 .25912 L
.50359 .26297 L
.46174 .27207 L
.42511 .28302 L
.39071 .29649 L
.35745 .31348 L
.33062 .33146 L
.31012 .34952 L
.30137 .35919 L
.29425 .3685 L
.28872 .37713 L
.28389 .38637 L
.28178 .39124 L
.2798 .39661 L
.27827 .40157 L
.27714 .4061 L
.27611 .41148 L
.27569 .41445 L
.27539 .41726 L
.27508 .42225 L
.27508 .42768 L
.27541 .43302 L
.276 .4379 L
.27703 .44351 L
.27836 .44884 L
.28152 .45822 L
.28616 .46835 L
.29264 .47927 L
.30113 .49066 L
.32037 .51031 L
.34805 .53095 L
.37932 .54842 L
.41121 .56219 L
.44921 .57477 L
.4879 .58426 L
.5259 .59085 L
.54832 .59361 L
.56904 .59545 L
.58058 .59619 L
.58669 .5965 L
.59319 .59678 L
.59886 .59696 L
.60513 .59711 L
.61619 .59723 L
.62175 .59722 L
.62783 .59716 L
.63417 .59704 L
Mistroke
.64007 .59688 L
.65063 .59645 L
.66189 .59582 L
.67505 .59483 L
.6874 .59366 L
.71025 .59087 L
.73154 .5875 L
.75074 .58378 L
.79226 .57332 L
.8262 .56187 L
.85936 .54744 L
.88705 .53201 L
.90895 .51658 L
.93008 .49718 L
.93977 .48568 L
.94723 .47473 L
.95299 .46399 L
.957 .45413 L
.95872 .44862 L
.96017 .44263 L
.96116 .43696 L
.96173 .43174 L
.96198 .42623 L
.96189 .42102 L
.96154 .41635 L
.96087 .41127 L
.9598 .40577 L
.95828 .39995 L
.95454 .38957 L
.94997 .38012 L
.94471 .37139 L
.92931 .35208 L
.90735 .33223 L
.8828 .31544 L
.85138 .29882 L
.8139 .28372 L
.77631 .27234 L
.73993 .26416 L
.71903 .26055 L
.69592 .25742 L
.68519 .25626 L
.67376 .25522 L
.663 .25442 L
.65313 .25384 L
.64335 .25341 L
.63406 .25314 L
.62423 .25298 L
.61353 .25298 L
.60421 .25311 L
.59431 .25339 L
Mistroke
.58588 .25375 L
.57658 .25426 L
.55593 .25587 L
.54441 .25706 L
.53391 .25832 L
.49504 .26457 L
.47507 .26882 L
.45741 .27323 L
.41978 .28492 L
.38786 .29782 L
.36072 .31165 L
.33324 .32954 L
.31223 .34749 L
.29435 .36845 L
.28749 .37942 L
.28423 .38577 L
.28165 .39172 L
.27962 .39731 L
.2781 .4024 L
.27678 .40808 L
.27621 .41121 L
.27579 .41416 L
.27527 .41954 L
.27509 .42449 L
.27523 .42993 L
.27575 .43566 L
.27648 .44046 L
.27757 .44559 L
.2803 .45479 L
.28484 .46562 L
.2902 .47535 L
.30612 .49631 L
.31749 .50766 L
.32948 .51776 L
.35473 .53504 L
.38652 .55177 L
.42077 .56564 L
.45503 .57635 L
.48993 .58463 L
.52841 .59116 L
.54942 .59369 L
.5611 .59479 L
.57217 .59564 L
.58255 .59627 L
.59384 .59677 L
.60412 .59706 L
.6136 .5972 L
.62488 .59718 L
.63542 .59699 L
.64697 .5966 L
Mistroke
.65344 .5963 L
.65933 .59597 L
.67113 .59514 L
.68391 .59401 L
.70691 .59133 L
.72842 .58805 L
.7511 .58371 L
.79042 .57387 L
.82599 .56197 L
.85583 .54919 L
.88473 .5335 L
.91167 .51445 L
.92323 .5042 L
.93401 .49285 L
.94192 .48282 L
.94915 .47152 L
.95453 .46064 L
.95673 .45503 L
.95867 .44895 L
.96009 .44325 L
.96104 .43801 L
.96168 .43274 L
.96199 .42771 L
.96203 .42468 L
.96197 .42193 L
.9618 .41876 L
.96152 .41577 L
.96078 .41045 L
.96021 .40746 L
.9596 .40472 L
.95582 .39263 L
.95326 .38659 L
.94996 .38002 L
.93577 .35931 L
.92554 .34813 L
.91482 .33827 L
.8924 .32146 L
.86325 .30454 L
.83204 .29044 L
.79535 .27765 L
.75868 .26802 L
.72235 .26104 L
.70097 .258 L
.68916 .25664 L
.67808 .25556 L
.66774 .25473 L
.6583 .2541 L
.64778 .25357 L
.63651 .25318 L
.63064 .25305 L
Mistroke
.62441 .25297 L
.61324 .25297 L
.60703 .25304 L
.60123 .25317 L
.59486 .25336 L
.58803 .25364 L
.57582 .2543 L
.56428 .25514 L
.54279 .25724 L
.5208 .26014 L
.5005 .26354 L
.46386 .27156 L
.42677 .28251 L
.3901 .29683 L
.36135 .31131 L
.33494 .32831 L
.31591 .344 L
.29901 .36224 L
.29162 .37256 L
.28511 .38402 L
.28266 .38934 L
.28044 .39502 L
.27882 .39996 L
.27741 .40531 L
.27637 .41047 L
.27569 .41521 L
.27529 .41975 L
.27512 .42454 L
.27525 .4298 L
.27567 .43469 L
.27644 .44005 L
.27763 .44575 L
.27934 .45183 L
.28129 .4574 L
.28706 .46986 L
.29367 .48065 L
.3022 .49181 L
.32058 .51037 L
.34579 .52938 L
.37587 .54662 L
.40839 .56101 L
.44047 .57211 L
.47801 .58204 L
.49999 .58656 L
.52117 .5901 L
.5437 .59305 L
.55597 .59431 L
.569 .5954 L
.57945 .59608 L
.59094 .59665 L
Mistroke
.60142 .59699 L
.61119 .59716 L
.62276 .59718 L
.63328 .59703 L
.6391 .59688 L
.6454 .59666 L
.65683 .5961 L
.66749 .59541 L
.6772 .59463 L
.69887 .59236 L
.71865 .58963 L
.73712 .58649 L
.77691 .57762 L
.8174 .56516 L
.85313 .55049 L
.88653 .5324 L
.91301 .51336 L
.9324 .49472 L
.94724 .47483 L
.95282 .46451 L
.95543 .45852 L
.95744 .45302 L
.95888 .44827 L
.96014 .44309 L
.96113 .4376 L
.96173 .43239 L
.96201 .42751 L
.96201 .42233 L
.96175 .41791 L
.9612 .41303 L
.96021 .40734 L
.95875 .40137 L
.9551 .39076 L
.94969 .3795 L
.94204 .3674 L
.92391 .3465 L
.9022 .32826 L
.87504 .31083 L
.84082 .29405 L
.80451 .2805 L
.76815 .27022 L
.73342 .2629 L
.71304 .25961 L
.69364 .25711 L
.68304 .25601 L
.67292 .25511 L
.66368 .25443 L
.65381 .25384 L
.64308 .25337 L
.63311 .25309 L
Mistroke
.62699 .25298 L
.62124 .25294 L
.61049 .25298 L
.60027 .25319 L
.58935 .25357 L
.57906 .2541 L
.56973 .25471 L
.54705 .25676 L
.52589 .2594 L
.48499 .26663 L
.46281 .27183 L
.44301 .27733 L
.38074 .30119 L
.35384 .31573 L
.32828 .3334 L
.30761 .35228 L
.29838 .36306 L
.29115 .37333 L
.28561 .38305 L
.28158 .392 L
.27824 .40205 L
.27692 .40761 L
.27601 .41286 L
.27542 .41811 L
.27523 .4211 L
.27514 .42389 L
.27525 .42952 L
.27548 .43267 L
.27578 .43555 L
.27657 .44067 L
.27777 .44624 L
.27924 .45147 L
.28089 .4563 L
.28554 .46693 L
.29224 .47849 L
.31069 .50106 L
.33203 .51967 L
.35935 .5377 L
.38625 .5516 L
.41908 .56499 L
.45791 .57708 L
.48052 .58259 L
.50257 .58702 L
.52362 .59045 L
.5469 .59339 L
.55729 .59442 L
.56837 .59534 L
.57872 .59603 L
.58836 .59653 L
.59891 .59691 L
Mistroke
.61044 .59715 L
.62061 .59719 L
.63145 .59707 L
.64209 .59678 L
.65208 .59635 L
.66104 .59585 L
.67068 .59517 L
.69278 .59307 L
.71576 .59007 L
.75567 .58273 L
.79093 .57374 L
.82176 .56358 L
.85287 .55061 L
.88259 .53483 L
.90271 .52144 L
Mfstroke
1 0 0 r
.55797 .59455 m
.566 .59522 L
.57352 .59576 L
.58248 .5963 L
.59065 .59668 L
.59478 .59683 L
.59927 .59697 L
.60351 .59708 L
.60742 .59715 L
.60956 .59718 L
.61157 .5972 L
.61377 .59722 L
.61501 .59723 L
.61614 .59723 L
.61841 .59723 L
.61969 .59723 L
.62088 .59723 L
.62306 .59722 L
.62428 .59721 L
.62538 .59719 L
.62759 .59717 L
.63001 .59713 L
.63489 .59703 L
.63941 .5969 L
.64363 .59675 L
.65313 .59633 L
.66329 .59572 L
.68179 .59422 L
.69849 .59241 L
.73058 .58767 L
.76425 .58076 L
.7948 .57256 L
.82528 .56222 L
.87385 .53983 L
.89518 .5267 L
.91516 .51146 L
.93123 .49594 L
.94463 .47882 L
.94947 .47088 L
.95378 .46227 L
.95703 .45406 L
.95927 .44654 L
.96023 .44235 L
.96102 .43791 L
.96151 .4341 L
.96171 .43194 L
.96185 .42992 L
.96191 .4288 L
.96193 .42821 L
.96195 .42758 L
Mistroke
.96197 .42698 L
.96198 .42643 L
.96199 .42536 L
.96198 .42433 L
.96197 .42322 L
.96194 .42217 L
.9619 .4212 L
.96184 .42008 L
.96178 .41903 L
.96157 .41665 L
.96131 .41433 L
.96097 .41185 L
.96013 .40726 L
.95914 .40303 L
.95611 .39351 L
.95256 .3852 L
.9479 .37648 L
.93664 .36044 L
.92325 .34596 L
.88635 .31762 L
.86302 .30447 L
.83573 .29199 L
.78024 .27338 L
.74913 .26599 L
.71449 .25987 L
.6971 .25756 L
.67802 .25558 L
.6692 .25486 L
.65983 .25422 L
.65102 .25374 L
.64297 .2534 L
.63891 .25327 L
.63448 .25315 L
.63029 .25306 L
.62843 .25303 L
.62641 .25301 L
.6243 .25299 L
.62232 .25297 L
.62117 .25297 L
.61996 .25296 L
.61779 .25296 L
.61553 .25297 L
.61342 .25298 L
.6111 .253 L
.60863 .25303 L
.60448 .25311 L
.59999 .25322 L
.59184 .25349 L
.58391 .25385 L
.57662 .25426 L
Mistroke
.55985 .25552 L
.54428 .25707 L
.50947 .26195 L
.47647 .2685 L
.44761 .27596 L
.41734 .2858 L
.39123 .2963 L
.34576 .32078 L
.32548 .3356 L
.30987 .34986 L
.29663 .3653 L
.291 .37349 L
.28578 .38262 L
.28202 .39079 L
.27889 .39965 L
.27755 .40457 L
.27659 .40907 L
.27621 .41124 L
.27586 .4136 L
.27561 .41567 L
.27539 .4179 L
.27531 .41897 L
.27524 .42011 L
.27518 .42115 L
.27514 .42213 L
.27512 .42304 L
.2751 .42402 L
.27509 .42511 L
.2751 .42609 L
.27511 .42666 L
.27512 .42725 L
.27516 .42832 L
.27521 .42948 L
.27528 .43074 L
.27546 .43301 L
.27569 .43513 L
.27635 .43973 L
.27718 .44392 L
.27979 .45331 L
.28306 .46177 L
.28685 .46953 L
.29807 .48673 L
.31272 .50315 L
.32888 .51729 L
.35014 .53221 L
.37236 .54487 L
.42395 .56676 L
.45205 .57552 L
.48461 .58352 L
.51679 .58946 L
Mistroke
.53529 .59206 L
.55235 .59398 L
.56954 .59545 L
.57778 .596 L
.58553 .59642 L
.59391 .59677 L
.59837 .59692 L
.60316 .59704 L
.60556 .59709 L
.60818 .59714 L
.61065 .59717 L
.61295 .59719 L
.61522 .59721 L
.61728 .59721 L
.61843 .59721 L
.61966 .59721 L
.62191 .5972 L
.62408 .59719 L
.62606 .59717 L
.62835 .59714 L
.63051 .5971 L
.63436 .59702 L
.63855 .59691 L
.64802 .59656 L
.65676 .59612 L
.67448 .59487 L
.69315 .59303 L
.72601 .58845 L
.7559 .58268 L
.78749 .57472 L
.81488 .56604 L
.8432 .55496 L
.86735 .54339 L
.90829 .51715 L
.92587 .50162 L
.93896 .48679 L
.94948 .47094 L
.95403 .46181 L
.95733 .45329 L
.95858 .44927 L
.9597 .44494 L
.9605 .44119 L
.96118 .43709 L
.96149 .4346 L
.96172 .43229 L
.96188 .43008 L
.96194 .42884 L
.96199 .42773 L
.96201 .42662 L
.96203 .4256 L
Mistroke
.96203 .42449 L
.96202 .42393 L
.96201 .42331 L
.96198 .42227 L
.96194 .42116 L
.96189 .42011 L
.96182 .41916 L
.96166 .41716 L
.96143 .41497 L
.96079 .41053 L
.95999 .40643 L
.95906 .40259 L
.95634 .39399 L
.95243 .38482 L
.94249 .36806 L
.9284 .35105 L
.91196 .33587 L
.89401 .32253 L
.84803 .29724 L
.8189 .28545 L
.78923 .27584 L
.75794 .26785 L
.72781 .26194 L
.69614 .25742 L
.67782 .25554 L
.66053 .25424 L
.65183 .25376 L
.64724 .25355 L
.64234 .25336 L
.63807 .25322 L
.63336 .2531 L
.63129 .25306 L
.62906 .25302 L
.62697 .253 L
.62503 .25297 L
.62313 .25296 L
.62109 .25295 L
.61883 .25294 L
.61678 .25295 L
.61463 .25296 L
.61342 .25296 L
.61231 .25297 L
.61026 .253 L
.60808 .25303 L
.60426 .2531 L
.60069 .25318 L
.59255 .25345 L
.58389 .25384 L
.57446 .25439 L
.55737 .25573 L
Mistroke
.54089 .25746 L
.52576 .25942 L
.49171 .26523 L
.46205 .27202 L
.43286 .28049 L
.40332 .2912 L
.37779 .30262 L
.35253 .31652 L
.33121 .3311 L
.31454 .34528 L
.30022 .36073 L
.29359 .3696 L
.28834 .37796 L
.28434 .38561 L
.28075 .39414 L
.27803 .4028 L
.27691 .40757 L
.27611 .41206 L
.27559 .41614 L
.27538 .41846 L
.2753 .41949 L
.27524 .42061 L
.27521 .42122 L
.27518 .42178 L
.27515 .42289 L
.27513 .4239 L
.27512 .42497 L
.27513 .42611 L
.27515 .42718 L
.27517 .4278 L
.27519 .42836 L
.27524 .42962 L
.2753 .43062 L
.27538 .43167 L
.27555 .43357 L
.27608 .43788 L
.2768 .44198 L
.27762 .44569 L
.28018 .45435 L
.2833 .46225 L
.29306 .47974 L
.30505 .49508 L
.3215 .51117 L
.34178 .5267 L
.36305 .53983 L
.41333 .5629 L
.44111 .5723 L
.47321 .58093 L
.50536 .58753 L
.52225 .59026 L
Mistroke
.54072 .5927 L
.55821 .59452 L
.57437 .59577 L
.58217 .59623 L
.59071 .59664 L
.59462 .59678 L
.5988 .59692 L
.60238 .59701 L
.60632 .59709 L
.60867 .59713 L
.61087 .59716 L
.61319 .59718 L
.6145 .59719 L
.61569 .59719 L
.61781 .5972 L
.62012 .5972 L
.6223 .59719 L
.62432 .59717 L
.62663 .59715 L
.62881 .59712 L
.63376 .59702 L
.63859 .5969 L
.64378 .59672 L
.65273 .59633 L
.66212 .59578 L
.67883 .59448 L
.6952 .5928 L
.71249 .59055 L
.74297 .58538 L
.77697 .5776 L
.80737 .56861 L
.85957 .54738 L
.88303 .53456 L
.90492 .51978 L
.92251 .50491 L
.93591 .49063 L
.94732 .4747 L
.95224 .4657 L
.95601 .45705 L
.95873 .44881 L
.9599 .44417 L
.96075 .43997 L
.96134 .43602 L
.9616 .43379 L
.96179 .43174 L
.96191 .42994 L
.96196 .42892 L
.962 .42797 L
.96203 .42696 L
.96204 .42604 L
Mistroke
.96205 .42499 L
.96204 .42398 L
.96202 .42299 L
.96199 .42192 L
.96195 .42091 L
.9619 .41998 L
.96174 .41784 L
.96152 .41558 L
.96094 .41132 L
.96007 .40668 L
.95788 .39844 L
.95513 .39083 L
.95098 .38189 L
.94619 .37357 L
.93499 .35835 L
.91977 .3426 L
.90296 .32882 L
.8815 .31458 L
.85886 .30232 L
.80729 .28141 L
.77662 .27236 L
.74712 .26552 L
.71563 .25999 L
.6974 .25755 L
.68028 .25575 L
.67147 .25499 L
.66191 .25431 L
.65346 .25382 L
.64453 .25343 L
.63642 .25317 L
.63182 .25306 L
.62962 .25302 L
.62753 .25299 L
.62564 .25297 L
.62358 .25295 L
.62171 .25294 L
.61998 .25294 L
.61812 .25294 L
.61606 .25294 L
.61392 .25295 L
.61191 .25297 L
.60973 .25299 L
.6074 .25303 L
.60322 .25311 L
.59825 .25324 L
.59374 .2534 L
.58488 .25378 L
.57654 .25425 L
.56112 .2554 L
.54456 .25703 L
Mistroke
.51091 .26171 L
.49436 .26471 L
.47657 .26849 L
.44376 .27711 L
.41562 .28646 L
.36507 .30925 L
.34381 .32212 L
.32362 .33719 L
.30746 .35244 L
.29373 .36942 L
.2888 .3772 L
.28431 .38571 L
.28091 .39377 L
.27844 .40134 L
.27734 .40567 L
.27653 .40962 L
.27585 .41401 L
.27557 .41642 L
.27547 .41751 L
.27537 .41871 L
.2753 .41978 L
.27524 .42093 L
.27521 .42157 L
.27519 .42215 L
.27515 .42329 L
.27514 .42433 L
.27514 .42546 L
.27515 .42653 L
.27517 .42752 L
.27519 .42812 L
.27522 .42877 L
.27528 .42994 L
.27543 .43214 L
.27568 .43471 L
.27598 .43706 L
.27666 .44115 L
.27759 .44552 L
.27985 .45336 L
.28348 .46261 L
.28767 .47094 L
.29816 .48677 L
.31297 .50333 L
.33029 .51833 L
.35187 .53323 L
.39819 .55686 L
.42524 .56716 L
.45648 .57669 L
.48827 .58425 L
.505 .58745 L
.52319 .59039 L
Mistroke
.54077 .5927 L
.55691 .59439 L
.56476 .59506 L
.57329 .59569 L
.58077 .59615 L
.58882 .59655 L
.5933 .59673 L
.5974 .59687 L
.60156 .59698 L
.60542 .59707 L
.60755 .59711 L
.60991 .59714 L
.61227 .59716 L
.61357 .59717 L
.61479 .59718 L
.61702 .59719 L
.61907 .59719 L
.62132 .59719 L
.62245 .59718 L
.62368 .59717 L
.62617 .59715 L
.62852 .59712 L
.63291 .59704 L
.63764 .59692 L
.64277 .59675 L
.65199 .59636 L
.66048 .59588 L
.67725 .59462 L
.69283 .59306 L
.72764 .58818 L
.76064 .58162 L
.78949 .57416 L
.81977 .56431 L
.84587 .55381 L
.89134 .52933 L
.90271 .52144 L
Mfstroke
0 0 m
1 0 L
1 .79815 L
0 .79815 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{321.75, 256.813},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40007R0001PB000`400?l00000o`00003oo`3ooooS
0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00nP3oool00`3ooml0W4P0000000020000000307Bo
o`3oool0oooo0>00oooo003l0?ooo`030?noM00007@0_ooo0><0oooo003l0?ooo`030?oo_`1d0000
M;oo0><0oooo003l0?ooo`030?oog`2LB000B9cO0><0oooo003m0?ooo`030=nLB000000000000080
000000<0M;oo0?ooo`3oool00P3oool01@3o_g@000000000000004P0W=oo0080oooo00<0oono07@0
000000000P0000003@18W=l0oooo0?ooo`3oool0_g@00000000004P0W=oo0?ooo`3o_g@000000000
B02Lgol00P3oool0102oM00000000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3o
ool00`3OW4P004RL0=ooo`020?ooo`030;md0000000007Bo0:X0oooo003m0?ooo`070?noM00004P0
W=oo0?ooo`3ooml0W4P0001d_`020?ooo`0?0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`3oool0
oooO09a8001d_ol0oooo0?oOW0180000M;oo0080oooo01L0gia80018W03Ooml0W7BL0=ooo`3o_g@0
001809cOg`2LM9`0gooo0?ooo`3OW4P004RL0=oog`2LM9`0gooo0?oOW01804P0W=oo0?ooo`3ogi`0
B000001d_`020?ooo`030?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;ooZ@3oool00?d0
oooo00<0oono07@0001d_ol00P3oool01`3ooml0W4P004RLg`3oool0gia80018W03Oool00P3oool0
2P3ogi`0B00004RLg`3oool0_gAd0;ooo`3oool0oono07@0001d_ol20?ooo`030?noM00004P0W=oo
0080oooo00<0oono07@0001d_ol00`3oool00`3o_g@0001809cOo`020?ooo`030=nLB000B9`0gooo
0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol0ZP3o
ool00?d0oooo00<0oooO09a80018W=l00`3oool01P3ogi`0B01809cOo`3o_g@0001d0;ooo`<0oooo
00<0gia80000M02oool00`3oool00`3OW4P0001d0;ooo`020?ooo`030?oo_`1d0000B9cO00<0oooo
00<0_g@0001d_`3oool00P3oool00`3ookl0M00004RLg`020?ooo`030?noM00007@0_ooo00<0oooo
00/0gia80000M02oool0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`2Z0?ooo`00oP3oool0
0`3OW4P0001d0;ooo`020?ooo`060?oog`2LB000B9cO0?oo_`1d0000M;oo0`3oool00`3ogi`0B018
09cOo`020?ooo`030?noM00004P0W=oo00@0oooo00<0_g@00018W03Oool00P3oool00`3OW4P004RL
0=ooo`030?ooo`070;md0000B9`0gooo0?ooo`3ookl0M00007Boo`030?ooo`070?oOW01804P0W=oo
0?ooo`3OW4P004Q804Qd_`020?ooo`030;md0000M;l0oooo0:T0oooo003n0?ooo`040?noM0000000
04RL0=ooo`80oooo00H0gia8001d_`3oool0gia80000M02oool20?ooo`0:0?oo_`1d0000M;oo0?oo
o`3ogi`0B01809cOo`3oool0giad09cOo`80oooo00<0gia80000B02Lgol00P3oool00`3ogi`0B018
09cOo`030?ooo`030=nLB00004P0W=oo0080oooo00<0gia80000M02oool00P3oool0303ookl0M000
07Boo`3oool0okmd0000M02o_g@0B9cO0?ooo`3o_g@0001d0;ooojT0oooo003n0?ooo`080?oo_`1d
0000M9bL04P0M02oool0oooO09a80000M;l20?ooo`0>0;md000007@0_ooo0?ooo`3o_g@0001d0;oo
o`3oool0omnL04P0B02Lgol0oooO09a8001d_ol20?ooo`030?oOW0180000B9cO0080oooo00<0oooO
09a80018W=l00`3oool00`3ogi`0B00004RLg`030?ooo`0@0;md000007@0_ooo0?ooo`3o_g@0001d
0;ooo`3oool0oooO09a8001d_ol0gia8001d_`3ogi`0B01809cOojT0oooo003m0?ooo`090?oOW018
0000000004RLg`3ookl0M0000000000007@0_ooo00<0oooo00D0okmd000000000000000007Boo`03
0?ooo`070;md00000000000000000018W=l0oooo0?noM002000000040018W03Oool0oono07@000<0
000000H0M;oo0?noM000000000000018W03Oool30?ooo`0?0?noM000000000000000001d_ol0oooo
0?oo_`1d0000000004RLg`3oool0okmd0000000004P0W=oo0:T0oooo003o0?oooal0oooo00<0okmd
0000M02oool0`@3oool00?l0oooo803oool00`3OW7@0W=oo0?ooo`300?ooo`00o`3ooolJ0?ooo`04
0?oo_`1d000004RL0=ooo`L0oooo00@0oono07@00000B9`0gooo^P3oool00?l0ooooh`3oool00?l0
ooooh`3oool00?l0ooooh`3oool00?l0ooooh`3oool00?l0ooooh`3oool00?l0ooooh`3oool00?l0
ooooh`3oool00?l0ooooh`3oool00?l0ooooh`3oool00?l0ooooh`3oool00?l0ooooh`3oool00?l0
ooooh`3oool00?l0ooooh`3oool007l0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B000
07Boo`020?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0603oool0102oM0000000
0000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono07@0000000000P0000001P18W=l0
oooo0?ooo`3oool0omnL04P00080000000<004RL0=ooo`3oool06@3oool0102oM00000000000001d
_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono07@0000000000P0000000`18W=l0oooo0?oo
o`0I0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol0103oool0102oM0000000
0000001d_ol30?ooo`060?oOW0180000000000000000B9`0gooo6`3oool0102oM00000000000001d
_olQ0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol0103oool0102oM0000000
0000001d_ol30?ooo`060?oOW0180000000000000000B9`0gooo6`3oool0102oM00000000000001d
_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono07@0000000000P0000000`18W=l0oooo0?oo
o`0_0?ooo`00OP3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0
0`3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo01L0oooo00H0gia8001d_`3oool0
omnL04P00018W=l20?ooo`030?oOW0180000M;oo00@0oooo00<0okmd0000001d_ol0103oool01P3o
_g@0000007Boo`3oool0omnL04Qd_aT0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo00@0oooo00<0okmd0000001d_ol06`3oool01P3OW4P007Bo0?ooo`3ogi`0B000
04RLg`80oooo00<0omnL04P0001d_ol00`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo
00H0okmd0000001d_ol0oooo0?oOW018M;lJ0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO7`3o
ool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool01P3OW4P0
07Bo0?ooo`3ogi`0B00004RLg`80oooo00H0okmd0000001d_ol0oooo0?oOW018M;lJ0?ooo`060=nL
B000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`040?ooo`030?noM0000000M;oo
0380oooo001m0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3o_g@0
07Bo0?ooo`0J0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3ooml0
W4P0001d_`090?ooo`030?noM018W=l0oooo01H0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`080?ooo`030?oog`2LB00007Bo01X0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`060?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3o_g@0
B9cO0?ooo`0G0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol07@3oool00`3ooml0
W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`060?ooo`030?noM018W=l0oooo01L0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`080?ooo`030?oog`2LB00007Bo0380oooo001m0?ooo`030?oog`2LB000B9cO0080oooo
00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo0?ooo`0I0?ooo`030?oog`2LB000B9cO0080
oooo00@0oooO09a80000B9`0gooo203oool00`3OW4P004RL0=ooo`080?ooo`030?oo_`1d07@0_ooo
01H0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030=nLB000B9`0
gooo01T0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?ooo`030?oog`2L
B000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3ookl0M01d0;ooo`0G0?ooo`030?oo
g`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo703oool00`3ooml0W4P004RLg`020?ooo`04
0?oog`2LB00004RL0=ooo`D0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool5
0?ooo`030?oo_`1d07@0_ooo01L0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03O
ool80?ooo`030=nLB000B9`0gooo0340oooo001`0?ooo`030?oog`2LB000000000L0000000H0B9cO
0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00T0oooo00<0okmd04RLg`3oool0
2`3oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB00007Bo0`3oool00`3O
W4P0001809cOo`080?ooo`030?oOW01804P0W=oo00P0oooo00<0oooO09a8001d_ol02@3oool00`3o
oml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cO
o`080?ooo`030?oOW01804P0W=oo00`0oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?oo
o`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol01@3oool00`3ooml0W4P0001d_`030?ooo`03
0=nLB00004P0W=oo00D0oooo00<0oooO09a8001d_ol05`3oool00`3ooml0W4P0001d_`030?ooo`03
0=nLB00004P0W=oo01`0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`03
0?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01@3oool00`3ooml0W4P007Boo`0G0?ooo`03
0?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol0203oool00`3ogi`0B01809cOo`0a0?ooo`00
OP3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo00<0gia804RLg`3oool0
603oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0oono07@0001d_ol0
203oool00`3ookl0M00004RLg`0G0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
203oool00`3ookl0M00007Boo`0J0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
1P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00D0oooo00<0oono07@00018W=l0
603oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO01d0oooo00<0_g@00018W03Oool0
0P3oool00`3ogi`0B00004RLg`060?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
1@3oool00`3ookl0M00004RLg`0H0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
203oool00`3ookl0M00007Boo`0a0?ooo`00OP3oool00`3OW4P0001809cOo`020?ooo`030?oog`2L
B00007Bo00/0oooo00<0_g@007Boo`3oool05`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2L
B00007Bo00T0oooo00<0_g@0001d_`3oool01`3oool00`3o_g@0000007Boo`0G0?ooo`030=nLB000
04P0W=oo0080oooo00<0oooO09a80000M;l02@3oool00`2oM00007Bo0?ooo`0I0?ooo`030=nLB000
04P0W=oo0080oooo00<0oooO09a80000M;l01P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2L
B00007Bo00D0oooo00<0okmd0000001d_ol0603oool00`3OW4P0001809cOo`020?ooo`030?oog`2L
B00007Bo01d0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`060?ooo`030=nLB000
04P0W=oo0080oooo00<0oooO09a80000M;l01@3oool00`3o_g@0000007Boo`0H0?ooo`030=nLB000
04P0W=oo0080oooo00<0oooO09a80000M;l02@3oool00`2oM00007Bo0?ooo`0`0?ooo`00OP3oool0
0`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02oool05`3oool0
0`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00P0oooo00<0gia80018W03Oool0103oool0
1P3ogml0gooo0?noM0000000001d0;oooaL0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL
0=ooo`080?ooo`030=nLB000B9`0gooo01T0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL
0=ooo`050?ooo`030?oOW0180000B9cO00<0oooo00X0_g@00018W03Oool0oooo0?oOg`3Oool0okmd
0000000007@0_ooo603oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo01`0oooo00<0
omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`050?ooo`030?oOW0180000B9cO00<0oooo00X0
_g@00018W03Oool0oooo0?oOg`3Oool0okmd0000000007@0_ooo603oool00`3ogi`0B00004RLg`03
0?ooo`030;md0000B9`0gooo00P0oooo00<0gia80018W03Oool0<03oool007h0oooo00@0oooO09a8
0000B9`0gooo0P3oool00`3OW4P004RL0=ooo`0:0?ooo`030?oog`2LB000M;oo01L0oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`030?oOW01804P0W=oo00@0oooo00D0
oono07@000000000001809cOo`0H0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool0203oool00`3ogi`0B01809cOo`0I0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018
W03Oool01@3oool0103ooml0W4P00018W03Oool20?ooo`090=nLB000B9`0gooo0?ooo`3ookl0M000
0000000004P0W=oo01T0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`0L0?oo
o`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01@3oool0103ooml0W4P00018W03O
ool20?ooo`090=nLB000B9`0gooo0?ooo`3ookl0M0000000000004P0W=oo01T0oooo00@0oooO09a8
0000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`030?oOW01804P0W=oo0300oooo001o0?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool02P3oool00`3ookl0M00004RLg`0H0?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0203oool00`3ookl0M00007Boo`050?oo
o`030=nLW02oool0oooo01X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`080?oo
o`030?oo_`1d0000M;oo01X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`060?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool00P3oool00`3OW9`0_ooo0?ooo`0K0?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool07@3oool00`3ogi`0B00007Boo`020?oo
o`030=nLB000M;l0oooo00H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`020?oo
o`030=nLW02oool0oooo01/0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`080?oo
o`030?oo_`1d0000M;oo0300oooo00200?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol06@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`X0oooo00<0_g@0001d_`3oool0103oool00`3ogi`0M9cO0?ooo`0K0?ooo`060=nLB00007@0_ooo
0?oog`2LB000B9cO2P3oool00`2oM00007Bo0?ooo`0J0?ooo`060=nLB00007@0_ooo0?oog`2LB000
B9cO203oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00<0omnL07BLg`3oool0703oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLgal0oooo00H0gia80000M02oool0oooO09a80018W=l80?oo
o`060=nLB00007@0_ooo0?oog`2LB000B9cO0`3oool00`3ogi`0M9cO0?ooo`0L0?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO2P3oool00`2oM00007Bo0?ooo`0_0?ooo`00P@3oool0103OW4P00000
00000000M;l90?ooo`050?oOW018000000000000001d_ol06`3oool0103OW4P0000000000000M;l9
0?ooo`050?oo_`1d000000000000M02oool01@3oool01@2oM0000000000000000000B9cO01T0oooo
00@0gia800000000000007Bo2@3oool01@3ookl0M0000000000007@0_ooo01/0oooo00@0gia80000
0000000007Bo2P3oool0103OW4P0000000000000M;l50?ooo`050;md00000000000000000018W=l0
6P3oool0103OW4P0000000000000M;lQ0?ooo`040=nLB00000000000001d_`X0oooo00@0gia80000
0000000007Bo1@3oool01@2oM0000000000000000000B9cO01X0oooo00@0gia800000000000007Bo
2@3oool01@3ookl0M0000000000007@0_ooo02l0oooo003o0?ooon<0oooo003o0?ooon<0oooo003o
0?ooon<0oooo003o0?ooon<0oooo003o0?ooon<0oooo003o0?ooon<0oooo003o0?ooon<0oooo003o
0?ooon<0oooo003o0?ooon<0oooo001J0?ooool00000O`00000:0?ooo`00FP3oool2000000P0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00P0oooo
001J0?ooo`800000203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0^
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0^
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo0>T0oooo00<000000?oo
o`3oool0SP3oool00`000000oooo0?ooo`080?ooo`00FP3oool300000>T0oooo00<000000?ooo`3o
ool0S@3oool2000000X0oooo001J0?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0
SP3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`2c0?oooa800000903o
ool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0
oooo0:L0oooo3000000B0?ooo`T000006`3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0
oooo00P0oooo001J0?ooo`030000003oool0oooo09l0oooo2000000W0?ooo`T000004P3oool00`00
0000oooo0?ooo`2>0?ooo`030000003oool0oooo00P0oooo000e0?ooo`040;md00000000000007Bo
o`<0oooo00<0omnL04P0001d_ol00P3oool00`3ookl0M000000000020000000604RLg`3oool0oooo
0?ooo`3ogi`0B0000P0000000`00B9`0gooo0?ooo`090?ooo`030000003oool0oooo09P0oooo1`00
000h0?ooo`L000002`3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo00P0oooo000d
0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`040?ooo`030?no
M0000000M;oo00@0oooo00H0okmd0000001d_ol0oooo0?oOW018M;l:0?ooo`030000003oool0oooo
09<0oooo1@0000160?ooo`D000001P3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo
00P0oooo000c0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3ooml0
W4P0001d_`090?ooo`030?noM018W=l0oooo00P0oooo00<000000?ooo`3oool0S`3oool400000500
oooo1@0000000`3oool000000?ooo`2?0?ooo`030000003oool0oooo00P0oooo000c0?ooo`030?oo
g`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`3OW4P004RL0=ooo`080?ooo`03
0?oo_`1d07@0_ooo00P0oooo00<000000?ooo`3oool0R`3oool4000005T0oooo1@00002=0?ooo`03
0000003oool0oooo00P0oooo000V0?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0
oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00P0oooo00<0omnL04P0B02Lgol0203oool00`3o
oml0W4P007Boo`080?ooo`030000003oool0oooo08L0oooo1000001N0?ooo`040000003oool0oooo
0?ooo`@00000R@3oool00`000000oooo0?ooo`080?ooo`00=03oool00`2oM00004RL0=ooo`020?oo
o`030?oOW0180000B9cO00P0oooo00<0oono07@0001d_ol0203oool00`3ookl0M00004RLg`080?oo
o`@00000P`3oool300000680oooo00<000000?ooo`3oool01@3oool3000008@0oooo0`00000:0?oo
o`00=03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00T0oooo00<0_g@0001d_`3o
ool01`3oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo0800oooo1000001U0?ooo`03
0000003oool0oooo00P0oooo0`0000230?ooo`030000003oool0oooo00P0oooo000d0?ooo`030?oO
W0180000B9cO00<0oooo00<0_g@00018W03Oool0203oool00`3OW4P004RL0=ooo`040?ooo`060?oO
g`3Oool0okmd0000000007@0_ooo203oool00`000000oooo0?ooo`1m0?ooo`<00000J@3oool00`00
0000oooo0?ooo`0;0?ooo`<00000P03oool00`000000oooo0?ooo`080?ooo`00=03oool0103ooml0
W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00P0oooo00<0omnL04P0B02Lgol0103oool01@3o
okl0M0000000000004P0W=oo00T0oooo00<000000?ooo`3oool0NP3oool3000006`0oooo00<00000
0?ooo`3oool03P3oool3000007d0oooo00<000000?ooo`3oool0203oool003D0oooo00<0omnL04P0
001d_ol00P3oool00`3OW4P007Bo0?ooo`080?ooo`030?oo_`1d0000M;oo00D0oooo00<0gibL0;oo
o`3oool02P3oool00`000000oooo0?ooo`1f0?ooo`@00000K`3oool00`000000oooo0?ooo`0A0?oo
o`800000N`3oool00`000000oooo0?ooo`080?ooo`00=P3oool01P3OW4P0001d0;ooo`3ooml0W4P0
04RLg`X0oooo00<0_g@0001d_`3oool0103oool00`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0
oooo07<0oooo0`00001c0?ooo`030000003oool0oooo01<0oooo0`00001h0?ooo`030000003oool0
oooo00P0oooo000g0?ooo`040=nLB00000000000001d_`T0oooo00D0oono07@000000000001d0;oo
o`050?ooo`050;md00000000000000000018W=l01`3oool00`000000oooo0?ooo`1a0?ooo`800000
MP3oool00`000000oooo0?ooo`0F0?ooo`<00000M@3oool00`000000oooo0?ooo`080?ooo`00FP3o
ool00`000000oooo0?ooo`1_0?ooo`800000N03oool00`000000oooo0?ooo`0I0?ooo`800000L`3o
ool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`1/0?ooo`<00000NP3oool0
0`000000oooo0?ooo`0K0?ooo`<00000L03oool00`000000oooo0?ooo`080?ooo`00FP3oool00`00
0000oooo0?ooo`1Z0?ooo`800000O@3oool00`000000oooo0?ooo`0N0?ooo`<00000K@3oool00`00
0000oooo0?ooo`080?ooo`00FP3oool3000006L0oooo0`00001o0?ooo`030000003oool0oooo0240
oooo0P00001Z0?ooo`8000002P3oool005X0oooo00<000000?ooo`3oool0I@3oool200000880oooo
00<000000?ooo`3oool08`3oool2000006T0oooo00<000000?ooo`3oool0203oool005X0oooo00<0
00000?ooo`3oool0H`3oool2000008@0oooo00<000000?ooo`3oool09@3oool2000006L0oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0H@3oool2000008H0oooo00<00000
0?ooo`3oool09`3oool2000006D0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?oo
o`3oool0G`3oool2000008P0oooo00<000000?ooo`3oool0:@3oool2000006<0oooo00<000000?oo
o`3oool0203oool005X0oooo00<000000?ooo`3oool0G@3oool2000008X0oooo00<000000?ooo`3o
ool0:`3oool200000640oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0
F`3oool2000008`0oooo00<000000?ooo`3oool0;@3oool2000005l0oooo00<000000?ooo`3oool0
203oool005X0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`2<0?ooo`030000003o
ool0oooo02l0oooo0P00001M0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0
oooo05P0oooo0P00002?0?ooo`030000003oool0oooo0340oooo0P00001K0?ooo`030000003oool0
oooo00P0oooo001J0?ooo`030000003oool0oooo05H0oooo0P00002A0?ooo`030000003oool0oooo
03<0oooo0P00001I0?ooo`030000003oool0oooo00P0oooo001J0?ooo`<00000E03oool2000009<0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`1E0?ooo`8000002P3oool005X0oooo
00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo03H0oooo
00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?oo
o`1A0?ooo`800000UP3oool00`000000oooo0?ooo`0g0?ooo`800000E@3oool00`000000oooo0?oo
o`080?ooo`00FP3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo09H0oooo00<00000
0?ooo`3oool0>@3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo00P0oooo001J0?oo
o`030000003oool0oooo04h0oooo0P00002I0?ooo`030000003oool0oooo03X0oooo00<000000?oo
o`3oool0D@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`1=0?ooo`03
0000003oool0oooo09T0oooo00<000000?ooo`3oool0>`3oool200000540oooo00<000000?ooo`3o
ool0203oool005X0oooo00<000000?ooo`3oool0B`3oool2000009`0oooo00<000000?ooo`3oool0
?@3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003o
ool0oooo04X0oooo00<000000?ooo`3oool0W03oool00`000000oooo0?ooo`0n0?ooo`030000003o
ool0oooo04d0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0B03oool2
000009l0oooo00<000000?ooo`3oool0?`3oool2000004d0oooo00<000000?ooo`3oool0203oool0
05X0oooo0`0000170?ooo`030000003oool0oooo09l0oooo00<000000?ooo`3oool0@@3oool00`00
0000oooo0?ooo`190?ooo`8000002P3oool005X0oooo00<000000?ooo`3oool0A@3oool200000:80
oooo00<000000?ooo`3oool0@P3oool2000004X0oooo00<000000?ooo`3oool0203oool005X0oooo
00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo04@0oooo
00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?oo
o`130?ooo`030000003oool0oooo0:<0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?oo
o`160?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo0480oooo00<00000
0?ooo`3oool0Y03oool00`000000oooo0?ooo`160?ooo`800000AP3oool00`000000oooo0?ooo`08
0?ooo`00FP3oool00`000000oooo0?ooo`100?ooo`800000Y`3oool00`000000oooo0?ooo`180?oo
o`030000003oool0oooo04<0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3o
ool0?`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3o
ool0@P3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0n0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool0BP3oool200000480oooo00<000000?ooo`3oool0
203oool005X0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`2Y0?ooo`030000003o
ool0oooo04`0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`080?ooo`00FP3oool0
0`000000oooo0?ooo`0l0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0C@3oool0
0`000000oooo0?ooo`0n0?ooo`030000003oool0oooo00P0oooo001J0?ooo`<00000>`3oool00`00
0000oooo0?ooo`2[0?ooo`030000003oool0oooo04h0oooo0P00000m0?ooo`8000002P3oool005X0
oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1b0?oooa@000009P3oool00`000000
oooo0?ooo`1@0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0203oool005X0oooo
00<000000?ooo`3oool0>03oool2000006`0oooo2@00000D0?ooo``000006P3oool00`000000oooo
0?ooo`1A0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0203oool005X0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`1W0?ooo`D00000:@3oool7000001<0oooo00<0
00000?ooo`3oool0DP3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00P0oooo001J
0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0H@3oool7000003D0oooo1@00000>
0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`08
0?ooo`00?P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono
07@0000000000P0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool0G@3oool500000440oooo1000000:0?ooo`030000003oool0oooo05@0oooo00<00000
0?ooo`3oool0=`3oool00`000000oooo0?ooo`080?ooo`00?@3oool01P3OW4P007Bo0?ooo`3ogi`0
B00004RLg`80oooo00<0omnL04P0001d_ol0103oool00`3o_g@0000007Boo`0;0?ooo`030000003o
ool0oooo03@0oooo00<000000?ooo`3oool0FP3oool4000004X0oooo100000060?ooo`030000003o
ool0oooo05D0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`080?ooo`00?03oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0oooO09a80000M;l02`3oool0
0`000000oooo0?ooo`0c0?ooo`030000003oool0oooo05L0oooo1000001B0?ooo`D0000000<0oooo
0000003oool0E`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00P0oooo000l0?oo
o`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`3OW4P004RL0=ooo`0:
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0E@3oool3000005/0oooo0`00001H
0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0203oool002l0oooo00<0oooO09a8
000000001`0000001P18W=l0oooo0?ooo`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol0203o
ool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo0300oooo0P00001E0?ooo`<00000G`3o
ool00`000000oooo00000002000005H0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?oo
o`080?ooo`00?@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0oono
07@0001d_ol02P3oool4000002h0oooo00<000000?ooo`3oool0DP3oool300000680oooo00<00000
0?ooo`3oool00P3oool3000005@0oooo00<000000?ooo`3oool0<03oool3000000X0oooo000m0?oo
o`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02@3oool00`2oM00007Bo0?ooo`090?oo
o`030000003oool0oooo02h0oooo00<000000?ooo`3oool0D03oool3000006D0oooo00<000000?oo
o`3oool01@3oool300000580oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`080?oo
o`00?@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00P0oooo00<0gia80018W03O
ool02@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo04h0oooo0`00001X0?ooo`03
0000003oool0oooo00P0oooo0`00001@0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3o
ool0203oool003d0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`03
0?oOW01804P0W=oo00T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1=0?ooo`80
0000J`3oool00`000000oooo0?ooo`0;0?ooo`800000C`3oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo00P0oooo000n0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0
203oool00`3ookl0M00007Boo`090?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0
C03oool2000006d0oooo00<000000?ooo`3oool03@3oool2000004h0oooo00<000000?ooo`3oool0
;P3oool00`000000oooo0?ooo`080?ooo`00?`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`X0
oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo04T0
oooo0`00001_0?ooo`030000003oool0oooo00l0oooo0`00001<0?ooo`030000003oool0oooo02d0
oooo00<000000?ooo`3oool0203oool00400oooo00@0gia800000000000007Bo2@3oool01@3ookl0
M0000000000007@0_ooo00P0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`180?oo
o`800000LP3oool00`000000oooo0?ooo`0B0?ooo`800000B`3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3o
ool0A`3oool2000007@0oooo00<000000?ooo`3oool0503oool2000004T0oooo00<000000?ooo`3o
ool0;03oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0X0?ooo`030000
003oool0oooo04H0oooo0P00001f0?ooo`030000003oool0oooo01H0oooo0P0000180?ooo`030000
003oool0oooo02/0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`150?ooo`800000N03oool00`000000oooo0?ooo`0H0?ooo`800000A`3o
ool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00P0oooo001J0?ooo`<000009`3oool0
0`000000oooo0?ooo`140?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool06P3oool2
000004H0oooo00<000000?ooo`3oool0:03oool2000000X0oooo001J0?ooo`030000003oool0oooo
02H0oooo00<000000?ooo`3oool0@`3oool2000007/0oooo00<000000?ooo`3oool0703oool20000
04D0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000
oooo0?ooo`0U0?ooo`030000003oool0oooo0480oooo0P00001m0?ooo`030000003oool0oooo01h0
oooo0P0000140?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0203oool005X0oooo
00<000000?ooo`3oool0903oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo07d0oooo
00<000000?ooo`3oool0803oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0903oool00`000000oooo0?oo
o`100?ooo`800000P03oool00`000000oooo0?ooo`0Q0?ooo`800000@P3oool00`000000oooo0?oo
o`0V0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool0?`3oool200000580oooo2@00000W0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool0@03oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00P0oooo001J0?oo
o`030000003oool0oooo0280oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`120?oo
oa0000002@3oool>000001T0oooo00<000000?ooo`3oool0903oool200000440oooo00<000000?oo
o`3oool0903oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0Q0?ooo`03
0000003oool0oooo03h0oooo0P0000100?ooo`D000009`3oool6000001<0oooo00<000000?ooo`3o
ool09P3oool200000400oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`080?ooo`00
FP3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo03`0oooo0P00000m0?ooo`D00000
<P3oool9000000X0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0n0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0803o
ool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo03T0oooo100000100?ooo`@000001P3o
ool00`000000oooo0?ooo`0Y0?ooo`800000?P3oool00`000000oooo0?ooo`0R0?ooo`030000003o
ool0oooo00P0oooo001J0?ooo`<00000803oool00`000000oooo0?ooo`0k0?ooo`030000003oool0
oooo03D0oooo1@0000180?ooo`@000000P3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0
oooo03`0oooo00<000000?ooo`3oool0803oool2000000X0oooo001J0?ooo`030000003oool0oooo
01l0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0b0?ooo`@00000D@3oool30000
02h0oooo0P00000m0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo02l0
oooo1000001G0?ooo`@00000;@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool07P3oool00`000000oooo
0?ooo`0i0?ooo`800000<03oool2000005/0oooo00@000000?ooo`3oool0oooo1000000Z0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`080?ooo`00
FP3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0
;P3oool2000005d0oooo00<000000?ooo`3oool01@3oool3000002P0oooo0P00000k0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo02d0oooo0P00001O0?ooo`030000003o
ool0oooo00P0oooo0P00000X0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool07@3o
ool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0
oooo03P0oooo00<000000?ooo`3oool0:`3oool300000640oooo00<000000?ooo`3oool02P3oool3
000002H0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo00P0oooo001J0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`0Z0?ooo`800000;@3ooolW00000100oooo00<000000?ooo`3oool03@3oool30000
02@0oooo0P00000i0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02T0
oooo0P00000Y0?ooo`H000009`3oool7000000T0oooo00<000000?ooo`3oool0403oool2000002@0
oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00P0
oooo001J0?ooo`<000006P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02P0oooo
0P00000V0?ooo`D00000=03oool8000000030?ooo`000000oooo01<0oooo0`00000R0?ooo`030000
003oool0oooo03H0oooo00<000000?ooo`3oool06P3oool2000000X0oooo001J0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0=@3oool2000002T0oooo0P00000R0?ooo`H00000@@3o
ool6000001<0oooo0P00000Q0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo03D0oooo00<000000?ooo`3oool09`3oool200000200oooo1000000S0?oooad00000203oool0
0`000000oooo0?ooo`020?ooo`@000004@3oool200000200oooo00<000000?ooo`3oool0=P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0V0?ooo`8000007P3oool40000
01l0oooo2000000G0?ooo`800000103oool7000000<0o`001`3oool300000100oooo0P00000O0?oo
o`030000003oool0oooo03D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`080?oo
o`00FP3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3o
ool09@3oool2000001h0oooo0P00000M0?ooo`H00000303oool9000003<0o`000`3oool2000001h0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00P0
oooo001J0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`0T0?ooo`8000007@3oool3000001L0oooo200000090?ooo`X000003@3o000C0?ooo`030000
003oool0oooo00D0oooo3@00000=0?l001d0oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0<`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01X0oooo0`00
000C0?ooo`L000002@3oool8000000040?ooo`0000000000000000L0o`000P00000M0?ooo`030000
003oool0oooo00`0oooo4P0000080?l001L0oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool0=03oool00`000000oooo0?ooo`0R0?ooo`8000006`3oool200000180oooo1000
000:0?ooo`H000000`3oool3000000L0o`000P00000S0?ooo`030000003oool0oooo0180oooo4P00
00080?l00180oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00P0oooo001J0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0<`3oool0
0`000000oooo0?ooo`0Q0?ooo`8000006P3oool300000100oooo1000000:0?ooo`@000000`3oool3
000000L0o`0000<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0H0?oooa8000001`3o000>
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`08
0?ooo`00FP3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo03<0oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`0G0?ooo`<00000403oool3000000X0oooo100000020?oo
o`<000001P3o0002000002l0oooo00<000000?ooo`3oool0703oool5000000030?ooo`0000000000
00H0000000<0oooo0000000000000`0000050?l000/0oooo00<000000?ooo`3oool0<P3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo00P0oooo001J0?ooo`<000005@3oool00`000000
oooo0?ooo`0b0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool05P3oool200000140
oooo0P00000:0?ooo`P000001P3o0002000003<0oooo00<000000?ooo`3oool07`3oool600000003
0?ooo`000000000000L0000000<0oooo0000000000000P0000050?l000P0oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`0D0?ooo`8000002P3oool005X0oooo00<000000?ooo`3oool0
503oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0200oooo0P00000G0?ooo`800000
403oool3000000P0oooo200000060?l000030000003oool0oooo03H0oooo00<000000?ooo`3oool0
8P3oool<000000030?ooo`000000000000L000001@3o00050?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`0E0?ooo`8000003`3oool3000000P0oooo0`0000000`3oool0000000000002000000D0o`0000<0
00000?ooo`3oool0>P3oool00`000000oooo0?ooo`0U0?ooo`<0000000@0oooo0000000000000000
0P3oool4000000<0oooo1P0000050?l000<0oooo00<000000?ooo`3oool0<@3oool00`000000oooo
0?ooo`0D0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01@0oooo0P00
000>0?ooo`<00000203oool3000000030?ooo`000000000000800000103o00000`000000oooo0?oo
o`0n0?ooo`030000003oool0oooo02P0oooo0P000000103oool000000000000000040?ooo`<00000
103oool6000000@0o`0000<0oooo0000003oool0<@3oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0<@3o
ool00`000000oooo0?ooo`0N0?ooo`8000005@3oool2000000h0oooo0P0000080?ooo`<0000000<0
oooo0000000000000P0000040?l000030000003oool0oooo0440oooo00<000000?ooo`3oool0:P3o
ool2000000030?ooo`000000000000800000103oool3000000@0oooo100000000`3oool000000000
00040?l00340oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`080?ooo`00=@3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo00@0oooo00@0_g@0000000000000M;oo
0`3oool01P3ogi`0B00000000000000004RL0=ooo`X0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0a0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool04`3oool20000
00h0oooo0P0000060?ooo`@000000P3oool3000000@0o`0000<000000?ooo`3oool0A03oool00`00
0000oooo0?ooo`0/0?ooo`80000000<0oooo0000003oool00P0000050?ooo`@000000`3oool60000
00@0o`00;`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00P0oooo000d0?ooo`06
0=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`030?ooo`060=nLB000M;l0
oooo0?oOW0180000B9cO0P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_`X0oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo01d0oooo00<000000?oo
o`3oool04P3oool2000000d0oooo0`0000050?ooo`<000000`3oool3000000<0o`000P0000190?oo
o`030000003oool0oooo02h0oooo10000000103oool000000000000000060?ooo`<000000`3oool6
000000@0o`00;@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00P0oooo000c0?oo
o`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3ooml0W4P004RLg`020?oo
o`030?noM00004P0W=oo00H0oooo00<0okmd04RLg`3oool0203oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo0300oooo00<000000?ooo`3oool0703oool2000001@0oooo00<000000?oo
o`3oool02`3oool2000000H0oooo0P0000040?ooo`<000000`3o00000`000000oooo0?ooo`1:0?oo
o`030000003oool0oooo0300oooo0`0000030?ooo`<000001P3oool3000000<0oooo1P0000040?l0
02X0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`080?ooo`00<`3oool00`3ooml0
W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<0oooO09a80018W=l00P3oool0103o
oml0W4P00018W03Oool50?ooo`030?oo_`1d07@0_ooo00P0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool04P3oool2
000000`0oooo0P0000060?ooo`800000103oool2000000<0o`000P00001>0?ooo`030000003oool0
oooo0380oooo0`0000040?ooo`<000001P3oool2000000@0oooo1P0000030?l002T0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`080?ooo`009P3oool00`3ooml0W4P0000000070000
000604RLg`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2L
B00007Bo00<0oooo00<0gia80000B02Lgol01@3oool00`3ooml0W4P007Boo`080?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0L0?ooo`030000003o
ool0oooo0140oooo0P00000<0?ooo`8000001P3oool2000000<0oooo0`0000030?l000030000003o
ool0oooo04l0oooo00<000000?ooo`3oool0=03oool2000000H0oooo0`0000050?ooo`8000000`3o
ool5000000030?ooo`000000o`000080o`009`3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo00P0oooo000d0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01P3oool0
0`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00D0oooo00<0oono07@00018W=l0203oool4
000000l0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool02P3oool2000000H0oooo0P0000030?ooo`800000103o001D
0?ooo`030000003oool0oooo03H0oooo0P0000070?ooo`8000001@3oool200000080oooo00<00000
0?ooo`0000001P0000030?l002@0oooo00<000000?ooo`3oool03`3oool3000000X0oooo000d0?oo
o`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01P3oool00`3OW4P0001809cOo`020?oo
o`030?oog`2LB00007Bo00D0oooo00<0okmd0000001d_ol0203oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool2000000<0oooo0P0000040?l0
05H0oooo00<000000?ooo`3oool0>03oool2000000L0oooo0P0000050?ooo`80000000@0oooo0000
003oool0oooo0P0000000`3oool0000000000002000000<0o`008`3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo00P0oooo000d0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018
W03Oool01@3oool00`3ogi`0B00004RLg`030?ooo`0:0;md0000B9`0gooo0?ooo`3ogml0gooo0?no
M0000000001d0;ooo`P0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0^0?ooo`03
0000003oool0oooo01/0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`090?ooo`80
00001@3oool2000000<0oooo0P0000030?l000030000003oool0oooo05H0oooo00<000000?ooo`3o
ool0>@3oool2000000P0oooo0P0000050?ooo`800000103oool7000000<0o`008@3oool00`000000
oooo0?ooo`0@0?ooo`030000003oool0oooo00P0oooo000d0?ooo`040?oog`2LB00004RL0=ooo`80
oooo00<0gia80018W03Oool01@3oool0103ooml0W4P00018W03Oool20?ooo`090=nLB000B9`0gooo
0?ooo`3ookl0M0000000000004P0W=oo00T0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`0_0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool03`3oool2000000X0oooo
0P0000050?ooo`8000000`3oool2000000<0o`0000<000000?ooo`3oool0F03oool00`000000oooo
0?ooo`0k0?ooo`800000203oool2000000D0oooo0`0000040?ooo`H000000`3o000O0?ooo`030000
003oool0oooo0100oooo00<000000?ooo`3oool0203oool003D0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`060?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool00P3o
ool00`3OW9`0_ooo0?ooo`0:0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0;P3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203o
ool2000000D0oooo0P0000030?ooo`<000000P3o00000`000000oooo0?ooo`1J0?ooo`030000003o
ool0oooo03`0oooo0`0000080?ooo`800000103oool010000000oooo0000000000040?ooo`H00000
0P3o000O0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool003H0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO0`3oool0
0`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0;P3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`030?ooo`8000000`3oool300000080o`0000<000000?ooo`3oool0G03oool0
0`000000oooo0?ooo`0n0?ooo`<00000203oool2000000<0oooo00<000000?ooo`3oool00P000004
0?ooo`H000000P3o000M0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0203oool0
03L0oooo00@0gia800000000000007Bo2P3oool0103OW4P0000000000000M;l50?ooo`050;md0000
0000000000000018W=l01`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool300000080o`0000<000000?ooo`3o
ool0GP3oool00`000000oooo0?ooo`0o0?ooo`040000003oool00000000000P0oooo0P0000000`3o
ool000000?ooo`030?ooo`800000103oool5000000<0o`00703oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0;@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<000000`3o001R0?ooo`03
0000003oool0oooo0400oooo00@000000?ooo`3oool000002@3oool3000000D0oooo0P0000040?oo
o`D000000`3o000J0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`070?ooo`800000103oool200000004
0?ooo`00000000000?ooo`<0o`0000<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`110?oo
o`030000003oool0oooo008000002@3oool2000000H0oooo0P0000040?ooo`D000000P3o000I0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo00H000000?ooo`3o
ool000000?ooo`0000030?l006D0oooo00<000000?ooo`3oool0@P3oool010000000oooo0?ooo`3o
ool2000000T0oooo0P0000060?ooo`800000103oool3000000040?ooo`000000o`000?l001P0oooo
00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`080?ooo`00FP3oool3000000`0oooo00<0
00000?ooo`3oool0;P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`030?ooo`060000003oool000000000003oool0
00000P3o001W0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool00P3oool2000000L0
oooo00@000000?ooo`00000000001P3oool00`000000oooo0?ooo`030?ooo`@000000P3o000G0?oo
o`030000003oool0oooo00`0oooo0P00000:0?ooo`00FP3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo02d0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool2000000040?ooo`000000oooo0000
00<0o`00J03oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`02000000D0oooo0P0000040?ooo`D000000P3o000F0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0203oool00180oooo00<0oono07Ad_`3o
ool0A@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool00P3oool01P000000oooo0?ooo`000000oooo00000080o`00JP3oool00`000000oooo0?ooo`15
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`8000001@3oool2000000@0oooo0`0000000`3oool0o`000?l0000E0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0203oool00180oooo00@0omnL04P0000004P0W=ooA03oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00H0oooo0P0000040?ooo`060000003oool0
00000000003oool000000P3o001[0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0
0`3oool2000000D0oooo00<000000?ooo`3oool00`3oool2000000D0oooo00<000000?ooo`3oool0
0`3oool300000080o`00503oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00P0oooo
000D0?ooo`030;md00000000000000<0000000<0001d0;ooo`3oool0?@3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool01@000000oooo0000
003oool000000080o`00K@3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo
0P0000040?ooo`030000003oool0000000<0o`004`3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00P0oooo000H0?ooo`030?oOW0180000M;oo03l0oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo00D000000?ooo`000000oooo
000000020?l006h0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool01@3oool2000000D0oooo00<000000?ooo`3oool0
0P3oool010000000oooo0000000000020?l00180oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`080?ooo`005`3oool00`3OW4P0001d0;ooo`100?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`050000003oool000000?ooo`00
00000P3o001_0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0103oool200000080
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080
oooo00@000000?ooo`00000000000P3o000A0?ooo`030000003oool0oooo00/0oooo00<000000?oo
o`3oool0203oool00180oooo00H0omnL04P0B02Lgol0okmd00000018W=m20?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`050000003oool0
00000?ooo`0000000P3o001`0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool01@3o
ool010000000oooo0?ooo`0000090?ooo`800000103oool2000000<0oooo0`0000000`3oool0o`00
0?l0000@0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0203oool00180oooo00@0
omnL04P00000000007BoA03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool00`3oool010000000oooo0000003oool20?l00780oooo00<000000?ooo`3o
ool0B`3oool00`000000oooo0?ooo`020?ooo`@0000000<0oooo0000003oool02P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo100000000`3o0000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00@000000?ooo`000000
oooo0P3o001c0?ooo`030000003oool0oooo04/0oooo1@0000030?ooo`@000002`3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00D000000?ooo`00000000000?l0000?0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool005X0oooo0`00000:0?ooo`03
0000003oool0oooo02/0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0<0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000020?l0
07@0oooo00<000000?ooo`3oool0BP3oool3000000T0oooo0`00000:0?ooo`030000003oool0oooo
00<0oooo0P0000030?ooo`050000003oool000000000003o00003P3oool00`000000oooo0?ooo`09
0?ooo`8000002P3oool001D0oooo00H0gia800000000000000000000B02Lgolo0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`040000003o
ool0oooo00000080o`00M@3oool00`000000oooo0?ooo`180?ooo`8000000`3oool00`000000oooo
0?ooo`060?ooo`80000000<0oooo0000000000002@3oool00`000000oooo0?ooo`040?ooo`060000
003oool0oooo0?ooo`000000oooo0P0000000`3o0000oooo0?ooo`0;0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0203oool001<0oooo00T0oono07@00018W=l0omnL09boo`3oool0
gia80018W03Oool0?P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool00P3oool010000000oooo0?ooo`0000020?l007H0oooo00<000000?ooo`3oool0
A`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00H0oooo0`0000000`3oool00000
000000080?ooo`030000003oool0oooo00@0oooo00H000000?ooo`3oool0oooo0000003oool20000
00030?l0003oool0oooo00/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?oo
o`004`3oool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d0;ooo`0n0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`040000003o
ool0oooo000000<0o`00MP3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0103oool01P000000oooo0?ooo`3oool000000?ooo`80000000<0o`00
0?ooo`3oool02P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo000B0?oo
o`090?oOW02L_ol0oooo0?oog`2LM9`0gooo0?ooo`3ogi`0B7Bo03l0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
0P0000020?l007L0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00D0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo00@0oooo0P0000020?ooo`80000000<0oooo0?l0003o00002`3oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo000B0?ooo`090?noW02Lgol0oooo0;md001d
_ol0oooo0=nLB00007@0_ooo03l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`05
0?ooo`050000003oool0oooo0?ooo`0000000P3oool200000080o`00N03oool00`000000oooo0?oo
o`130?ooo`8000002`3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
0P0000000`3oool0o`000?l0000:0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
203oool00180oooo00@0oono07@000000000M;ooA03oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo02/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0o`000?l0
07T0oooo00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
00D0oooo00L000000?ooo`3oool0oooo0000003oool0000000T0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`02000000030?ooo`3o0000o`0000X0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo02/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo00D0oooo00P000000?ooo`3oool0oooo0000003oool000000?ooo`80o`00NP3oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`020?ooo`030000003oool0000000T0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`03000000030?l0003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo
00X000000?ooo`3oool0oooo0000003oool000000?ooo`000000o`00N`3oool00`000000oooo0?oo
o`110?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`030000003oool0000000T0oooo00<000000?ooo`3oool01@3oool01@000000oooo0000
003oool000000080o`002@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo
000F0?ooo`060?noM000000000000000000007@0_ooo?P3oool3000000P0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool01@3oool01P000000oooo0?ooo`3oool000000?ooo`80000000<0o`000000003oool0NP3oool0
0`000000oooo0?ooo`100?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`030?ooo`030000003oool0000000T0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0300000080o`00203oool00`000000oooo0?ooo`070?ooo`8000002P3oool0
00h0oooo00H0_ibL0=ooo`3oool0oooo0?oOW0180003000000040018W03Oool0oono07Ad_cl0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01H0oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`040?ooo`060000003oool0oooo0?ooo`00
0000oooo0P0000020?l007`0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool2000000X0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0200000080o`00203oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00P0oooo000=0?ooo`030?oog`2LB000000000<0000000<007Bo0?ooo`3o
ool00P3oool00`3o_g@0001d0;ooo`0o0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool01@3oool01P000000oooo0?ooo`3oool000000?ooo`80000000<0o`000?ooo`3oool0N`3oool0
0`000000oooo0?ooo`0o0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`040?ooo`8000002P3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo0?ooo`8000000P3o00080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203o
ool000d0oooo00<0oono07@00018W=l0BP3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo02X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003o0000o`0007d0oooo00<0
00000?ooo`3oool0?P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00`0000000`3o0000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0Y
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`04
0?ooo`060000003oool0oooo0?ooo`000000oooo0P0000000`3o0000oooo0?ooo`1l0?ooo`030000
003oool0oooo03h0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`040000
003oool0oooo0?ooo`8000000P3o00070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0203oool001D0oooo00H0omnL04P00000000000000000B02Lgolo0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool000000080oooo0P000000
0`3o0000oooo0?ooo`1l0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00D0oooo0P00000:0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool00`0000000`3o0000oooo0?ooo`050?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0203oool001@0oooo00<0_g@00000001d_ol00P3oool00`3o
oml0W7BL0=ooo`0n0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0:P3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool01P00
0000oooo0?ooo`3oool000000?ooo`80000000<0o`000?ooo`3oool0O@3oool00`000000oooo0?oo
o`0m0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0000000X0oooo00<000000?ooo`3oool00`3oool010000000oooo0000
003oool2000000030?l0003oool0oooo00@0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`080?ooo`004`3oool00`3OW4P0B9cO0?ooo`040?ooo`030=nLW02oool0oooo03d0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo00000080oooo
0P0000000`3o0000oooo0?ooo`1m0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<000000?ooo`000000
2P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0000000<0o`000?ooo`3oool0
103oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo000B0?ooo`030?oOW01d
W=l0oooo00@0oooo00<0oooO09adW03Oool0?P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo02X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00D0oooo00D000000?ooo`3oool000000?ooo`0200000080o`00O`3oool00`000000oooo
0?ooo`0l0?ooo`030000003oool0oooo01T0oooo00D000000?ooo`3oool0oooo000000070?ooo`04
0000003oool0oooo000000X0oooo00<000000?ooo`3oool00`3oool01@000000oooo0000003oool0
00000080o`001@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo000B0?oo
o`030?noW02Lgol0oooo00<0oooo00<0oooO09a8001d_ol0?`3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo00@0oooo00@000000?ooo`3oool000000P3oool2000000030?l0003oool0
oooo07h0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0I0?ooo`050000003oool0
oooo0?ooo`000000203oool010000000oooo0?ooo`0000090?ooo`030000003oool0oooo00<0oooo
00@000000?ooo`000000oooo0P0000000`3o0000oooo0?ooo`030?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0203oool00180oooo00<0omnL04Qd_`3oool00P3oool00`3OW4P00000
04RLg`100?ooo`<000001`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01D0oooo
00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`070000003oool0oooo0000003o
ool000000?ooo`020?l00800oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0J0?oo
o`050000003oool0oooo0?ooo`000000203oool010000000oooo0?ooo`00000:0?ooo`030000003o
ool0oooo00<0oooo00H000000?ooo`00000000000?ooo`3o00060?ooo`030000003oool0oooo00D0
oooo0P00000:0?ooo`004`3oool01P3o_g@0000000000000000004RL0=oood40oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`040?ooo`060000003oool0oooo0?ooo`000000oooo0P00
00000`3o000000000?ooo`1o0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool06`3o
ool01@000000oooo0?ooo`3oool0000000L0oooo00D000000?ooo`3oool0oooo0000000:0?ooo`03
0000003oool0oooo0080oooo00D000000?ooo`000000oooo000000020?l000D0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo02X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo00@0oooo00H000000?ooo`3oool0oooo0000003oool2000000030?l0003oool0
oooo07l0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0K0?ooo`050000003oool0
oooo0?ooo`000000203oool010000000oooo0?ooo`00000:0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`0000000`3o0000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0Z0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?oo
o`040?ooo`060000003oool0oooo0?ooo`000000oooo0P0000000`3o0000oooo0?ooo`1o0?ooo`03
0000003oool0oooo03/0oooo00<000000?ooo`3oool0703oool010000000oooo0?ooo`0000080?oo
o`050000003oool0oooo0?ooo`0000002P3oool01@000000oooo0?ooo`3oool000000080oooo00@0
00000?ooo`000000o`001@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00P0oooo
000E0?ooo`060?oOW018000000000000000004P0W=oo?`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo02X0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo00@0oooo00D000000?ooo`3oool000000?ooo`03000000030?l0003oool0oooo
07l0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0L0?ooo`040000003oool0oooo
000000P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo0080oooo00H000000?ooo`000000oooo0000003o00050?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0203oool001<0oooo00@0oooO09a80000B9`0gooo0P3oool00`3o_g@004RL
0=ooo`0`0?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`0Z0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`040?ooo`050000003oool0oooo0000003oool00P0000020?l00840oooo00<000000?oo
o`3oool0>`3oool00`000000oooo0?ooo`0L0?ooo`040000003oool0oooo000000P0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00L00000
0?ooo`000000oooo0000003o0000000000@0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`080?ooo`004`3oool00`2oM4P0W=oo0?ooo`030?ooo`030?oog`2LB7@0_ooo02l0oooo00H0
gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool0103oool01@000000oooo0?ooo`000000oooo0080000000<0o`000?ooo`3oool0P03oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`000000203oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`03000000030?l0003oool0oooo0080oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`080?ooo`004P3oool00`3ogi`0W;oo0?ooo`040?ooo`030?oOW018M;l0oooo02h0
oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`000000oooo0080000000<0o`000?oo
o`3oool0P03oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`000000203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`03000000030?l0003oool0oooo0080oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`080?ooo`004P3oool00`3o_i`0W=oo0?ooo`030?ooo`03
0?oOW018B9`0gooo02l0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`000000
oooo0080000000<0o`000?ooo`3oool0P03oool00`000000oooo0?ooo`0j0?ooo`030000003oool0
oooo01h0oooo0`0000080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0000000<0o`000?ooo`3oool00P3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00P0oooo000B0?ooo`040?oo_`1d00000000
07Booc@0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`070?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`000000oooo0080000000<0
o`000?ooo`3oool0P03oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo01h0oooo0`00
00080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00<0000000<0o`000?ooo`3oool00P3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00P0oooo001;0?ooo`030;md0000B9`0gooo0080oooo00<0omnL
04P00018W=l01`3ooomY000000030?l00000000000000?l00000100000000`3o000000000000000<
000000X0oooo001;0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01`3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool000000?ooo`02
000000030?l0003oool0oooo0800oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0N
0?ooo`<00000203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`03000000030?l0003oool0oooo0080oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`080?ooo`004P3oool00`3ookl0W9cO0?ooo`030?oo
o`040?noM0000000001d0;ooobl0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`06
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`00
0000oooo0080000000<0o`000?ooo`3oool0P03oool00`000000oooo0?ooo`0j0?ooo`030000003o
ool0oooo01l0oooo0P0000080?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0000000<0o`000?ooo`3oool00P3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00P0oooo000B0?ooo`040?oo_`2LW=l0
oono07@000@0000000<004RL0=ooo`3oool0;P3oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0Z0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`050000
003oool0oooo0000003oool00P0000000`3o0000oooo0?ooo`200?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool07`3oool2000000P0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`0000000`3o0000
oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00180oooo
00T0omnL04P000000000001809cOo`3oool0oooO09a8M02oool0<@3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0Z
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`04
0?ooo`050000003oool0oooo0000003oool00P0000000`3o0000oooo0?ooo`200?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool07`3oool2000000P0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`00
00000`3o0000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203o
ool000h0oooo00@0oooO09a8000007@0_ooo>`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0000
003oool00P0000020?l00840oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0N0?oo
o`030000003oool0000000P0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool00P0000020?l000@0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`080?ooo`00CP3oool0103OW4P0000000000000M;l80?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool02@000000oooo0?ooo`000000
oooo0000003oool000000?l000210?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0
7@3oool00`000000oooo000000080?ooo`050000003oool0oooo0?ooo`0000002P3oool00`000000
oooo0?ooo`020?ooo`060000003oool000000?ooo`000000o`001@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo00P0oooo000B0?ooo`050?oO_`2ogol0oooo0?ooo`2oM0000`000000
0`0007@0_ooo0?ooo`0m0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool0
1P000000oooo0?ooo`3oool000000?ooo`80000000<0o`000?ooo`3oool0O`3oool00`000000oooo
0?ooo`0k0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`000000203oool01@000000oooo
0?ooo`3oool0000000T0oooo00<000000?ooo`3oool00P3oool01`000000oooo0?ooo`000000oooo
0000003o00001@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00P0oooo000B0?oo
o`030;md0000000000000080000000D0001d0;ooo`3oool0_ibL0=ooo`0n0?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0103oool01P000000oooo0?ooo`3oool000000?ooo`800000
00<0o`000?ooo`3oool0O`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo01`0oooo
00@000000?ooo`3oool00000203oool01@000000oooo0?ooo`3oool0000000P0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`03000000030?l0003oool0oooo00<0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`080?ooo`00403oool01@2oM000000000000000B9`0gooo
00<0oooo00<0oooO09a8001d_ol00P3oool00`2oM00004RL0=ooo`0j0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool01@3oool01@000000oooo0?ooo`000000oooo0080000000<0o`00
0?ooo`3oool0O`3oool2000003`0oooo00<000000?ooo`3oool0703oool00`000000oooo00000009
0?ooo`050000003oool0oooo0?ooo`000000203oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0000000<0o`000?ooo`3oool00`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00P0oooo000B0?ooo`030?noW02Lgol0oooo00P0oooo00<0omnL04Qd_`3oool0>P3oool3
000000L0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`000000oooo00<0000000<0
o`000?ooo`3oool0OP3oool2000003d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0000
00080?ooo`050000003oool0oooo0?ooo`0000002@3oool00`000000oooo0?ooo`030?ooo`050000
003oool000000?ooo`0000000P3o00050?ooo`030000003oool0oooo00D0oooo0P00000:0?ooo`00
4P3oool00`3ookl0M7Bo0?ooo`070?ooo`030?noM000B9`0gooo03/0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo00000080oooo0P0000000`3o0000
oooo0?ooo`1n0?ooo`800000?@3oool00`000000oooo0?ooo`0J0?ooo`040000003oool0oooo0000
00P0oooo00@000000?ooo`3oool000002@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00<0000000<0o`000?ooo`3oool0103oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00P0oooo000B0?ooo`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3oool0?03oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00H000000?ooo`3oool0oooo0000
003oool2000000030?l000000000oooo07h0oooo0P00000m0?ooo`030000003oool0oooo01X0oooo
00@000000?ooo`3oool00000203oool010000000oooo0?ooo`0000090?ooo`030000003oool0oooo
0080oooo00D000000?ooo`3oool000000?ooo`02000000030?l0003oool0oooo00@0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`080?ooo`004`3oool00`3ogi`0B000000000050000
000304RLg`3oool0oooo03`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?oo
o`060000003oool0oooo0?ooo`000000oooo0`0000000`3o0000oooo0?ooo`1m0?ooo`800000?@3o
ool00`000000oooo0?ooo`0I0?ooo`050000003oool0oooo0?ooo`000000203oool00`000000oooo
000000090?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`0000000`3o0000oooo
0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool001P0oooo00<0
okmd0000B02Lgol0?`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo02/0oooo00<0
00000?ooo`3oool05@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00@0oooo00H0
00000?ooo`3oool0oooo0000003oool2000000030?l000000000oooo07d0oooo0P00000n0?ooo`03
0000003oool0oooo01P0oooo00@000000?ooo`3oool000002@3oool2000000X0oooo00<000000?oo
o`3oool00P3oool01@000000oooo0?ooo`000000oooo0080000000<0o`000?ooo`3oool0103oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo000G0?ooo`030;md0000M;l0oooo
0400oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`040?ooo`060000003oool0oooo
0?ooo`000000oooo0P0000020?l007h0oooo00<000000?ooo`000000?@3oool00`000000oooo0?oo
o`0G0?ooo`050000003oool0oooo0?ooo`0000002@3oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`0000000`3o0000oooo0?ooo`050?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00180oooo00H0omnL04P0001d_ol0okmd
0000001d_om20?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0:`3oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01@3oool010000000
oooo0?ooo`3oool2000000040?ooo`000000o`00000007d0oooo00<000000?ooo`000000?P3oool0
0`000000oooo0?ooo`0F0?ooo`040000003oool0oooo000000T0oooo0P0000090?ooo`030000003o
ool0oooo00<0oooo00D000000?ooo`3oool000000?ooo`02000000030?l0003oool0oooo00D0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`004P3oool0103ogi`0B0000000
0000M;m40?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0:`3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0103oool01P000000oooo
0?ooo`3oool000000?ooo`8000000P3o001m0?ooo`030000003oool0000003h0oooo00<000000?oo
o`3oool05@3oool01@000000oooo0?ooo`3oool0000000T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0000000<0o`000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`3oool2000000040?ooo`000000
o`000?l007`0oooo00<000000?ooo`000000?`3oool00`000000oooo0?ooo`0D0?ooo`040000003o
ool0oooo000000T0oooo0P0000090?ooo`030000003oool0oooo00<0oooo00D000000?ooo`3oool0
00000?ooo`02000000030?l0003oool0oooo00D0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool00P3oool4000000030?l0003oool0oooo07X0oooo00<000000?ooo`000000
@03oool00`000000oooo0?ooo`0B0?ooo`040000003oool0oooo000000T0oooo00<000000?ooo`00
0000203oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`<0000000<0o`000?oo
o`3oool01P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo001J0?ooo`<0
0000203oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`050?ooo`060000003oool0oooo0?ooo`000000oooo0P000002
0?l007/0oooo00<000000?ooo`000000@03oool00`000000oooo0?ooo`0B0?ooo`030000003oool0
000000T0oooo00<000000?ooo`000000203oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0080000000<0oooo0?l0003o0000203oool00`000000oooo0?ooo`070?ooo`8000002P3oool0
05X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo
01H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo
0?ooo`80000000@0oooo0000003o00000000NP3oool010000000oooo0?ooo`0000100?ooo`030000
003oool0oooo0100oooo00@000000?ooo`3oool000001`3oool200000080oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`030?ooo`050000003oool0oooo0000003oool00P0000000`3o
0000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`060000003oool0oooo0?oo
o`000000oooo0`0000000`3o000000000?ooo`1h0?ooo`040000003oool0oooo00000440oooo0P00
000@0?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`0000000`3o0000oooo0?oo
o`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool005X0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`060000003oool0oooo0?ooo`000000oooo
0P0000020?l007T0oooo00@000000?ooo`3oool00000@`3oool00`000000oooo0?ooo`0<0?ooo`<0
00001`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0300000080o`002@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
;03oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`020?ooo`060000003oool000000000003o00000000N03oool01000
0000oooo0?ooo`0000140?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool2
00000080oooo0P0000000`3oool0o`000?l0000:0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0[0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`060?oo
o`0;0000003oool0oooo0?ooo`000000oooo0000003oool000000?l000000000M`3oool010000000
oooo0?ooo`0000150?ooo`@000001`3oool2000000P0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00H000000?ooo`3oool0oooo0000003o
ool2000000030?l0003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`00FP3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02/0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00H00000
0?ooo`3oool0oooo0000003oool3000000030?l000000000oooo07D0oooo00D000000?ooo`3oool0
oooo000000180?ooo`P00000203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080
oooo0P0000050?ooo`030000003oool0oooo0080oooo0`0000000`3o000000000?ooo`0:0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`060?ooo`060000003oool0oooo0?ooo`000000oooo0P0000020?l0
07H0oooo00D000000?ooo`3oool0oooo0000001?0?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool01@3oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3oool00P3o
ool3000000030?l000000000oooo00/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`00FP3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02`0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool00P3oool010000000oooo0000000000020?l007D0oooo00D000000?ooo`3oool0oooo
0000001>0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01P3oool010000000oooo
0?ooo`0000060?ooo`040000003oool0oooo0?ooo`@0000000<0o`000000003oool02`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool01P3oool02@000000oooo0?ooo`3oool000000?ooo`000000oooo0000
00020?l007@0oooo00D000000?ooo`3oool0oooo0000001=0?ooo`030000003oool0oooo00@0oooo
0P0000080?ooo`040000003oool0oooo000000H0oooo00H000000?ooo`3oool0oooo0000003oool2
000000030?l000000000oooo00`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`08
0?ooo`00FP3oool3000000X0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool01P3oool02@000000oooo0?ooo`3oool0
00000?ooo`000000oooo000000020?l007<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`1:0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0203oool00`000000oooo
000000060?ooo`060000003oool0oooo0?ooo`000000oooo0P0000000`3o000000000?ooo`0=0?oo
o`030000003oool0oooo00T0oooo0P00000:0?ooo`00FP3oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo02`0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00L0oooo00T000000?ooo`3oool0oooo0000003oool000000?ooo`0000000P3o
001b0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00P0oooo0`0000060?ooo`030000003oool0oooo0080000000D0
oooo000000000000o`000000000?0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
203oool005X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo0080oooo00@000000?ooo`000000oooo0P0000000`3o0000oooo0?ooo`1_0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`0000020?ooo`04
0000003o0000o`00000000l0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`080?oo
o`00FP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool00P3oool020000000oooo000000000000oooo0000003o00000000K`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`090?ooo`8000001P3oool010000000oooo0?ooo`0000020?ooo`030000003o0000o`000140oooo
00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?oo
o`0;0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool01P000000oooo0?oo
o`000000oooo00000080o`00KP3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04D0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`<000001@3oool200000004
0?ooo`00000000000?ooo`8000000P3o000A0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0203oool005X0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo0080oooo00H000000?ooo`3oool000000?ooo`0000020?l006d0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`140?ooo`040000003oool0oooo0?ooo`8000002P3o
ool010000000oooo0?ooo`0000040?ooo`040000003oool0oooo00000080oooo0P0000020?l00180
oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0300000080o`00K03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0480oooo
00D000000?ooo`3oool0oooo0000000:0?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`04
0000003oool0oooo00000080oooo00@000000?l0003o000000004P3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool0;@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`8000000`3o001Z
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`02
000000T0oooo0P0000040?ooo`030000003oool0oooo0080000000@0oooo000000000000oooo0P00
00020?l001@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`080?ooo`00FP3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0oooo0P0000030?ooo`8000000P3oool3
00000080o`00J@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0400oooo00@00000
0?ooo`3oool000002@3oool2000000D0oooo00@000000?ooo`3oool000000P3oool010000000oooo
0?ooo`0000030?l001@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`080?ooo`00
FP3oool3000000d0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00<000000P3o00000`000000oooo0?ooo`1U0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0?P3oool010000000oooo0000000000080?ooo`8000001`3oool300000003
0?ooo`00000000000080oooo00@000000?l0003o0000o`005@3oool00`000000oooo0?ooo`0=0?oo
o`8000002P3oool005X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0^0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo0080oooo00@000000?ooo`3oool0oooo0`0000020?l006H0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`0n0?ooo`8000002@3oool00`000000oooo0?ooo`060?oo
o`80000000<0oooo0000000000000P3oool010000000o`000?l0003o000F0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`0]0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo0P0000020?ooo`@0000000<0o`000000
003oool0H`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03`0oooo0P0000080?oo
o`8000001`3oool3000000040?ooo`000000oooo0?ooo`8000000`3o000G0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0?ooo`080?ooo`800000103oool2000000050?ooo`00000000000000003oool00P3o0000
0`000000oooo0?ooo`1P0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0>P3oool2
000000P0oooo0P0000070?ooo`80000000D0oooo00000000000000000?ooo`02000000<0o`00603o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo00h0oooo0P00000:0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P000000
103oool000000?l0003o001Q0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0>03o
ool2000000L0oooo0P0000070?ooo`8000000P3oool2000000040?ooo`0000000000000000<0o`00
6@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0oooo000e0?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol0103oool0102oM00000000000001d_ol30?ooo`06
0?oOW0180000000000000000B9`0gooo2P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo02h0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool00`3oool2000000050?ooo`00000000000000003oool00`3o
001O0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0=P3oool3000000H0oooo0P00
00070?ooo`8000000`3oool5000000<0o`006P3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo00P0oooo000d0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B000
07Boo`030?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_`X0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0_0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`080?ooo`8000001@3o
ool2000000050?ooo`00000000000000003oool00`3o001M0?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0=03oool00`000000oooo000000050?ooo`8000001`3oool2000000<0oooo
1@0000030?l001`0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`080?ooo`00<`3o
ool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0oooO09a80018W=l00P3o
ool00`3o_g@0001809cOo`060?ooo`030?noM018W=l0oooo00P0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool03`3o
ool2000000X0oooo00<000000?ooo`3oool0103oool2000000050?ooo`00000000000000003oool0
0`3o001K0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0<P3oool2000000030?oo
o`000000oooo00<0oooo0P0000070?ooo`8000000`3oool5000000<0o`007@3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo00P0oooo000c0?ooo`030?oog`2LB000B9cO0080oooo00@0
oooO09a80000B9`0gooo1@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0
oooo00<0oono07@0M02oool0203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02l0
oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00P0
oooo0P0000060?ooo`030000003oool0oooo00<0000000@0oooo0?l0003o0000o`00F@3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo0300oooo00@000000?ooo`3oool000000`3oool2
000000L0oooo0P0000030?ooo`D000000`3o000O0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0203oool003<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?oo
o`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01@3oool00`3ooml0W4P007Boo`080?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo0100oooo0P00000:0?ooo`8000001@3oool200000080oooo0`0000001@3o
ool0o`000?l0003o0000000005H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0^
0?ooo`80000000<0oooo0000000000000P3oool2000000L0oooo0P0000030?ooo`@00000103o000P
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0203oool003@0oooo00<0_g@00018
W03Oool00P3oool00`3ogi`0B00004RLg`060?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l01@3oool00`3ookl0M00004RLg`080?ooo`@00000403oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo01`0oooo0P00000B0?ooo`030000003oool0oooo00T0oooo0P0000050?oo
o`8000000P3oool500000080o`0000<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo02/0oooo0P0000020?ooo`030000003oool00000008000001`3oool20000
00<0oooo100000040?l000030000003oool0oooo01l0oooo00<000000?ooo`3oool04@3oool30000
00X0oooo000d0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01P3oool00`3OW4P0
001809cOo`020?ooo`030?oog`2LB00007Bo00D0oooo00<0okmd0000001d_ol0203oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`0@0?ooo`8000002`3oool2000000D0oooo0P0000020?ooo`D000000`3o001B0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0:@3oool2000000<0oooo0`0000080?ooo`80
00000`3oool4000000@0o`0000<0oooo0000003oool0803oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo00P0oooo000d0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool0
1@3oool00`3ogi`0B00004RLg`030?ooo`0:0;md0000B9`0gooo0?ooo`3ogml0gooo0?noM0000000
001d0;ooo`P0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0`0?ooo`030000003o
ool0oooo01d0oooo00<000000?ooo`3oool04@3oool2000000/0oooo0P0000050?ooo`8000000`3o
ool4000000<0o`0000<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo02H0oooo0P0000040?ooo`8000001`3oool300000080oooo1@0000040?l000<0oooo00<0
00000?ooo`3oool07`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00P0oooo000d
0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01@3oool0103ooml0W4P00018
W03Oool20?ooo`090=nLB000B9`0gooo0?ooo`3ookl0M0000000000004P0W=oo00T0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo01d0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0:0?ooo`8000001@3oool2000000<0oooo10000004
0?l004d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0S0?ooo`800000103oool2
000000H0oooo0`0000020?ooo`D00000103o00050?ooo`030000003oool0oooo01l0oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`080?ooo`00=@3oool00`3ogi`0B00007Boo`020?oo
o`030=nLB000M;l0oooo00H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`020?oo
o`030=nLW02oool0oooo00X0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0a0?oo
o`030000003oool0oooo01h0oooo0P00000C0?ooo`800000303oool2000000D0oooo0P0000030?oo
o`@00000103o00000`000000oooo0?ooo`180?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool0803oool2000000<0oooo100000030?ooo`@000000P3oool5000000D0o`001P3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0203oool003H0
oooo00H0gia80000M02oool0oooO09a80018W=l80?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO
0`3oool00`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
<@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo0180oooo0P00000<0?ooo`<00000
103oool3000000<0oooo0P0000000`3oool000000?l000030?l000030000003oool0oooo04D0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0N0?ooo`8000000`3oool2000000040?oo
o`00000000000?ooo`<000000`3oool3000000030?ooo`000000o`0000@0o`002@3oool00`000000
oooo0?ooo`0O0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0203oool003L0oooo
00@0gia800000000000007Bo2P3oool0103OW4P0000000000000M;l50?ooo`050;md000000000000
00000018W=l01`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0380oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`0C0?ooo`8000003@3oool2000000D0oooo0`000002
0?ooo`80000000<0oooo0000000000001@3o00140?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool06`3oool2000000<0oooo0P0000020?ooo`@000000`3oool3000000030?ooo`000000
o`0000@0o`002`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0b0?oo
o`030000003oool0oooo01l0oooo0P00000E0?ooo`8000003@3oool2000000H0oooo0`0000001@3o
ool00000000000000000oooo00<000001@3o00100?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0603oool200000080oooo0`0000020?ooo`<000000P3oool6000000D0o`003P3oool0
0`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0203oool0
05X0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool0503oool3000000`0oooo0`0000060?ooo`<0000000<0oooo0000
000000001@0000050?l003d0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0E0?oo
o`80000000@0oooo00000000000000000P3oool4000000060?ooo`00000000000000003oool00000
1@3o00000`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0
603oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo03<0oooo00<000000?ooo`3oool0803oool2000001L0oooo0P00000=0?ooo`<000001P3o
ool300000080oooo1P0000060?l003T0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0C0?ooo`D0000000<0oooo0000000000000P0000000`3oool0000000000003000000H0o`0000<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool0203oool005X0oooo0`00000E0?ooo`030000003oool0oooo03@0oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0F0?ooo`<000003@3oool3000000H0oooo10000002
0?ooo`D000001`3o000e0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0403oool4
000000050?ooo`00000000000000003oool01@0000060?l000030000003oool0oooo01@0oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`0I0?ooo`8000002P3oool005X0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0603oool3000000d0oooo0`0000070?ooo`@000000P3oool5000000L0o`00<@3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00d0oooo300000060?l000030000003oool0
oooo01L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0
oooo00P0oooo001J0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0=03oool00`00
0000oooo0?ooo`0R0?ooo`8000006`3oool3000000d0oooo100000070?ooo`@000000P3oool50000
00L0o`0000<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo
00T0oooo2@0000080?l000030000003oool0oooo01X0oooo00<000000?ooo`3oool07P3oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo01L0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0S0?ooo`800000703oool3000000h0
oooo100000070?ooo`d000001`3o0002000002D0oooo00<000000?ooo`3oool05@3oool00`000000
oooo0?ooo`040?ooo`L00000203o0003000001l0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0T0?ooo`8000007@3oool3000000l0oooo
100000090?ooo``00000203o0003000001h0oooo00<000000?ooo`3oool05P3oool7000000P0o`00
0P0000040?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool07@3oool200000200oooo
00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0e0?ooo`030000003oool0oooo02D0oooo0P00000N0?ooo`<00000403oool6000000T0oooo1000
00040?ooo`@000003@3o000F0?ooo`030000003oool0oooo0180oooo100000090?l000030000003o
ool0oooo00L0oooo0P00000P0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0803o
ool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo03D0oooo00<000000?ooo`3oool09P3oool2000001l0oooo1000000B0?ooo`T00000103ooolA
00000180o`0000<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003o0000
o`000140o`0000<000000?ooo`3oool0303oool00`000000oooo0?ooo`0P0?ooo`030000003oool0
oooo01`0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool09`3oool30000
0200oooo1000000G0?ooo`L000002@3ooolE000001@0o`002P0000030?ooo`<000000`3oool00`00
0000oooo0?ooo`090?ooo`8000008P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo
0280oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0f0?ooo`800000:P3oool200000280oooo1000000J0?ooo`d00000303ooolE00000080
oooo1P0000060?ooo`<000001`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00P0
oooo001J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0>03oool00`000000oooo
0?ooo`0Y0?ooo`800000903oool4000002<0oooo4`0000080?ooo`P000002@3oool3000000/0oooo
00<000000?ooo`3oool0103oool2000002<0oooo00<000000?ooo`3oool0703oool00`000000oooo
0?ooo`0T0?ooo`030000003oool0oooo00P0oooo001J0?ooo`<000006`3oool00`000000oooo0?oo
o`0h0?ooo`030000003oool0oooo02X0oooo0P00000V0?ooo`P00000;@3oool:000000d0oooo0`00
000?0?ooo`8000000`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01/0oooo0P00
000V0?ooo`8000002P3oool005X0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0h
0?ooo`030000003oool0oooo02/0oooo0`00000[0?oooa4000003P3oool>000000T0oooo00<00000
0?ooo`3oool0203oool3000001@0oooo0`00000T0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo03T0oooo0P00000^0?ooo`<00000>@3oool>000001L0oooo00<000000?oo
o`3oool0103oool4000001H0oooo0P0000000`3oool000000000000R0?ooo`030000003oool0oooo
01/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0;P3oool4000005X0
oooo00@000000?ooo`3oool0oooo0`00000H0?ooo`8000001@3oool2000001l0oooo00<000000?oo
o`3oool06P3oool2000002/0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3o
ool07P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0340oooo1000001F0?ooo`@0
00006@3oool2000000T0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0H0?ooo`80
0000;@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0O0?ooo`030000
003oool0oooo03X0oooo0P00000e0?ooo`D00000C@3oool5000001/0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
;@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0O0?ooo`030000003o
ool0oooo03`0oooo0P00000h0?ooo`<00000@`3oool7000000@0oooo00<000000?ooo`3oool05`3o
ool2000000h0oooo0P00000I0?ooo`030000003oool0oooo01L0oooo0P00000`0?ooo`030000003o
ool0oooo00P0oooo001J0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0?@3oool0
0`000000oooo0?ooo`0h0?ooo`T00000<@3oool9000000/0oooo00<000000?ooo`3oool05@3oool2
00000180oooo0P00000F0?ooo`030000003oool0oooo01D0oooo0`00000b0?ooo`030000003oool0
oooo00P0oooo001J0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0?@3oool20000
0440oooo1@00000U0?ooo`L00000503oool00`000000oooo0?ooo`0C0?ooo`8000005P3oool20000
01<0oooo00<000000?ooo`3oool0503oool2000003D0oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0n0?ooo`800000A03ooolU000001/0
oooo00<000000?ooo`3oool04@3oool2000001X0oooo1000000>0?ooo`030000003oool0oooo0140
oooo1000000g0?ooo`030000003oool0oooo00P0oooo001J0?ooo`<000008P3oool00`000000oooo
0?ooo`100?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool03`3oool200000200oooo
1@0000080?ooo`030000003oool0oooo00d0oooo1@00000j0?ooo`8000002P3oool005X0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`100?ooo`800000P@3oool00`000000oooo0?oo
o`0=0?ooo`8000009`3oool500000080oooo00<000000?ooo`3oool02@3oool500000400oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`11
0?ooo`800000O`3oool00`000000oooo0?ooo`0;0?ooo`800000;P3oool>000004D0oooo00<00000
0?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`120?oo
o`800000O@3oool00`000000oooo0?ooo`080?ooo`<00000;`3oool200000580oooo00<000000?oo
o`3oool0203oool005X0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`130?ooo`80
0000N`3oool00`000000oooo0?ooo`060?ooo`800000<@3oool00`000000oooo0?ooo`1B0?ooo`03
0000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
A03oool3000007P0oooo00<000000?ooo`3oool00`3oool300000380oooo00<000000?ooo`3oool0
D`3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`0X0?ooo`030000003o
ool0oooo04H0oooo0P00001f0?ooo`040000003oool0oooo0?ooo`800000=03oool00`000000oooo
0?ooo`1D0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool0A`3oool2000007@0oooo1000000e0?ooo`030000003oool0oooo05D0oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`19
0?ooo`<00000L03oool2000003L0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`08
0?ooo`00FP3oool3000002X0oooo00<000000?ooo`3oool0B`3oool2000006X0oooo100000000`3o
ool000000?ooo`0e0?ooo`030000003oool0oooo05H0oooo0P00000:0?ooo`00FP3oool00`000000
oooo0?ooo`0[0?ooo`030000003oool0oooo04`0oooo0`00001S0?ooo`@000001@3oool00`000000
oooo0?ooo`0c0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0203oool005X0oooo
00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1>0?ooo`<00000G03oool4000000T0oooo
00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo00P0oooo
001J0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0D03oool4000005@0oooo1000
000=0?ooo`030000003oool0oooo0300oooo0P00001L0?ooo`030000003oool0oooo00P0oooo001J
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0D`3oool5000004/0oooo1000000A
0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`08
0?ooo`00?P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0oono
07@0000000000P0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool0E`3oool3000004@0oooo1000000E0?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0G@3oool00`000000oooo0?ooo`080?ooo`00?@3oool01P3OW4P007Bo0?ooo`3ogi`0
B00004RLg`80oooo00<0omnL04P0001d_ol0103oool00`3o_g@0000007Boo`0;0?ooo`030000003o
ool0oooo0300oooo00<000000?ooo`3oool0F@3oool4000003/0oooo1@00000I0?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`080?ooo`00?03oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0oooO09a80000M;l02`3oool0
0`000000oooo0?ooo`0a0?ooo`030000003oool0oooo05`0oooo1P00000Z0?ooo`/000007P3oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0203oool0
03`0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030=nLB000B9`0
gooo00X0oooo00<000000?ooo`3oool0<P3oool200000680oooo3000000D0?ooo`X00000:@3oool0
0`000000oooo0?ooo`0Z0?ooo`800000HP3oool00`000000oooo0?ooo`080?ooo`00?03oool00`3o
oml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<0omnL04P0B02Lgol02P3oool00`00
0000oooo0?ooo`0d0?ooo`030000003oool0oooo06/0oooo5000000c0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`080?ooo`00?@3oool00`2oM000
04RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0oono07@0001d_ol02P3oool4000003@0
oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`0W0?ooo`800000H`3oool3000000X0
oooo000m0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02@3oool00`2oM00007Bo
0?ooo`090?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0/03oool00`000000oooo
0?ooo`0U0?ooo`800000I`3oool00`000000oooo0?ooo`080?ooo`00?@3oool00`3ogi`0B00004RL
g`030?ooo`030;md0000B9`0gooo00P0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?oo
o`0g0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0903oool00`000000oooo0?oo
o`1W0?ooo`030000003oool0oooo00P0oooo000m0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0
gia80018W03Oool0203oool00`3ogi`0B01809cOo`090?ooo`030000003oool0oooo03P0oooo0P00
002_0?ooo`030000003oool0oooo0280oooo0P00001Z0?ooo`030000003oool0oooo00P0oooo000n
0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0203oool00`3ookl0M00007Boo`09
0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0203oool003l0oooo00H0gia80000
M02oool0oooO09a80018W=l:0?ooo`030;md0000M;l0oooo00P0oooo00<000000?ooo`3oool0>`3o
ool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo01l0oooo0P00001]0?ooo`030000003o
ool0oooo00P0oooo00100?ooo`040=nLB00000000000001d_`T0oooo00D0oono07@000000000001d
0;ooo`080?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0203oool005X0oooo00<0
00000?ooo`3oool0?@3oool200000:X0oooo00<000000?ooo`3oool0703oool200000700oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0?`3oool200000:P0oooo00<00000
0?ooo`3oool06`3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo00P0oooo001J0?oo
o`030000003oool0oooo0440oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`0I0?oo
o`800000L`3oool00`000000oooo0?ooo`080?ooo`00FP3oool300000480oooo0P00002U0?ooo`03
0000003oool0oooo01L0oooo0P00001d0?ooo`8000002P3oool005X0oooo00<000000?ooo`3oool0
A03oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
M@3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`150?ooo`800000XP3o
ool00`000000oooo0?ooo`0D0?ooo`800000N03oool00`000000oooo0?ooo`080?ooo`00FP3oool0
0`000000oooo0?ooo`170?ooo`800000X03oool00`000000oooo0?ooo`0B0?ooo`800000NP3oool0
0`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo
09d0oooo00<000000?ooo`3oool0403oool2000007`0oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool0BP3oool2000009d0oooo00<000000?ooo`3oool03P3oool2000007h0
oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0C03oool2000009/0oooo
00<000000?ooo`3oool0303oool200000800oooo00<000000?ooo`3oool0203oool005X0oooo00<0
00000?ooo`3oool0CP3oool2000009T0oooo00<000000?ooo`3oool02P3oool200000880oooo00<0
00000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0D03oool2000009L0oooo00<00000
0?ooo`3oool0203oool2000008@0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?oo
o`3oool0DP3oool2000009D0oooo00<000000?ooo`3oool01P3oool2000008H0oooo00<000000?oo
o`3oool0203oool005X0oooo0`00001D0?ooo`800000T`3oool00`000000oooo0?ooo`030?ooo`<0
0000Q`3oool2000000X0oooo001J0?ooo`030000003oool0oooo05H0oooo0P00002A0?ooo`040000
003oool0oooo0?ooo`800000R`3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo
0?ooo`1H0?ooo`800000S`3oool4000008d0oooo00<000000?ooo`3oool0203oool005X0oooo00<0
00000?ooo`3oool0FP3oool3000008/0oooo0P00002@0?ooo`030000003oool0oooo00P0oooo001J
0?ooo`030000003oool0oooo05d0oooo0`0000260?ooo`80000000<0oooo0000003oool0S`3oool0
0`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`1P0?ooo`800000P@3oool30000
00<0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000
oooo0?ooo`1R0?ooo`<00000O03oool2000000H0oooo00<000000?ooo`3oool0SP3oool00`000000
oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?ooo`1U0?ooo`<00000MP3oool3000000P0oooo
00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?oo
o`1X0?ooo`<00000K`3oool4000000/0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?oo
o`080?ooo`00FP3oool00`000000oooo0?ooo`1[0?ooo`<00000J03oool4000000l0oooo00<00000
0?ooo`3oool0SP3oool00`000000oooo0?ooo`080?ooo`00FP3oool3000006h0oooo1000001P0?oo
o`@000004`3oool00`000000oooo0?ooo`2=0?ooo`8000002P3oool005X0oooo00<000000?ooo`3o
ool0LP3oool4000005P0oooo1000000G0?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3o
ool0203oool005X0oooo00<000000?ooo`3oool0MP3oool400000500oooo1000000K0?ooo`030000
003oool0oooo08h0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0NP3o
ool5000004H0oooo1@00000O0?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3oool0203o
ool005X0oooo00<000000?ooo`3oool0O`3oool4000003`0oooo1P00000T0?ooo`030000003oool0
oooo08h0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0P`3oool50000
02h0oooo2@00000Z0?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3oool0203oool005X0
oooo00<000000?ooo`3oool0R03oool;000001X0oooo2@00000c0?ooo`030000003oool0oooo08h0
oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3oool0T`3ooolJ000003`0oooo
00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`080?ooo`00FP3oool00`000000oooo0?oo
o`3Y0?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3oool0203oool005X0oooo0`00003Y
0?ooo`030000003oool0oooo08d0oooo0P00000:0?ooo`00FP3oool00`000000oooo0?ooo`3Y0?oo
o`030000003oool0oooo08h0oooo00<000000?ooo`3oool0203oool005X0oooo00<000000?ooo`3o
ool0j@3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo00P0oooo001J0?ooo`030000
003oool0oooo02D0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0^0?ooo`030000
003oool0oooo02h0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0^0?ooo`030000
003oool0oooo02h0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`080?ooo`00FP3o
ool2000000P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00P0oooo001J0?ooool00000O`00000:0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00
o`3ooooS0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00
o`3oool]0?ooo`050;md00000000000000000018W=l0/@3oool00?l0oooo;P3oool00`3ooml0W4P0
04RLg`2b0?ooo`00o`3oool_0?ooo`030=nLB000B9`0gooo0;40oooo003o0?ooobl0oooo00<0omnL
04P0M02oool0/@3oool00=00oooo00H0gia800000000000000000000B02Lgol70?ooo`050=nLB000
000000000000001d_ol00P3oool00`3OW4P004RL0=ooo`020?ooo`0`0;md0000000007Bo0?ooo`3o
okl0M0000000000007@0_ooo0=nLB0000000B9cO0?ooo`3ogi`0B00000000000B9`0gooo0;md0000
000007Bo0?ooo`3o_g@000000000M02oool0oooo0?oOW018000000000018W03Oool0_g@000000000
M;l0oooo0=nLB000B9`0gooo0?ooo`3o_g@0001d0;ooo`3oool0oooO09a800000000M;oo1@3oool0
0`3o_g@00000000000070000000907Boo`3oool0omnL04P000000000001d0;oo_`1d0000M;oo0080
oooo00/0omnL04P00000000004RL0=ooo`2oM0000000001d_`3oool0omnL04P000020000000907Bo
o`3OW4P000000018W03Oool0okmd0000000004P0W=oo0080oooo00@0_g@00000000004P0W=oo0P3o
ool01@3o_g@000000000000004P0W=oo00<0oooo00<0gia80018W03Oool00P3oool00`2oM0000000
001d_`210?ooo`00d@3oool0103ooml0W4P00018W03Oool70?ooo`030?oOW0180000M;oo0080oooo
00H0_g@0001d_`3oool0okmd0000M02oool20?ooo`0[0=nLB000B9`0gkmd07Boo`3OW4P004RL0=oo
o`3ookl0M01d0;noM00007@0W9ad09cOo`3ogi`0B01809cOo`3ogi`0B7Bo0;md0000M;l0oknL09cO
o`3ogi`0B01d0;oo_`2LW=l0oooo0?oOW01804P0W=oo0?oOW018M;l0_g@0001d_`3o_i`0W=oo0?no
M00004P0W=oo0?ooo`3ookl0M00007Boo`020?ooo`040;md0000M;l0_gAd0;ooo`D0oooo00@0oooO
09a80000B9`0gooo0`3oool02`3ooml0W4P00018W03Oool0_g@00018W03Oool0oono07@0M01dM000
B9cO0080oooo02@0omnL04P0B02Lgol0omnL04Qd_`2oM00007Bo0?noW02Lgol0_g@00018W03Oool0
oooo0;md001dM7@0001d0;oOW02L_ol0okmd0000B02Lgml0W7BL0=ooo`3oool0gia80018W03Ooml0
W7BL0=ooo`3ogi`0B01809cOo`3oool0omnL04P00000M;l20?ooo`030?noM00007@0_ooo0080oooo
00@0gia80018W03O_g@0M;ooP03oool00=80oooo00<0okmd0000B02Lgol01`3oool00`3OW4P004RL
0=ooo`030?ooo`050=nLW02oool0oono07@00018W=l00P3oool01P3ogi`0B01809cOo`3oool0_g@0
001d_`80oooo02<0oono07A8B01804P0W=oo0?ooo`3ooml0W4P004RLg`3oool0okmd07AdM000B9`0
gooo0?ooo`3ookl0M00007Boo`3OW9`0_ooo0?oog`2LB000B9cO0?ooo`3o_g@0M7Ad0018W03Oool0
oooo0?oo_`1d0000B9cO0?ooo`3ooml0W4P0001d_`020?ooo`030=nLB000B9`0gooo00L0oooo00<0
okmd0000B02Lgol01@3oool0103OW7@0W=oo0;md0000M;l20?ooo`0@0?oo_`1d000004RL0=ooo`3o
ool0oooO09a80018W=l0oooo0?noM01dM7@004RL0=ooo`3oool0_g@0001d_`<0oooo00L0_g@00000
001d_ol0oooo0?oo_`1d0000M;oo00<0oooo00<0okmd0000B02Lgol00P3oool00`3OW4P004RL0=oo
o`020?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00004RLg`020?ooo`030?oOW01804P0W=oo0840
oooo003B0?ooo`030?oo_`1d0000B9cO00L0oooo00D0okmd000000000000000004RLg`030?ooo`0<
0?oog`2LB000001d0;ooo`3oool0oooO09a80018W=l0oooo0?noM000B9`0gooo0P3oool00`3o_g@0
000004RLg`020?ooo`070;md0000B9`0gooo0?ooo`2oM000001809cOo`020?ooo`030;md0018W=l0
oooo0080oooo01@0_g@00018W03Oool0oooo0;md000004P0W=oo0?ooo`3ooml0W4P00018W03Oool0
oooo0=nLB0000000M;oo0?ooo`3ogi`0B01809cOo`L0oooo00<0oono07@00018W=l01`3oool00`3O
W4P004RL0=ooo`020?ooo`030?noM00007@0_ooo0080oooo00/0_g@00018W03Oool0oooo0;md0000
04P0W=oo0?ooo`3o_g@0001d0;ooo`020?ooo`030?oo_`1d0000B9cO0080oooo00<0_g@0001d_`3o
ool00P3oool00`3ookl0M00004RLg`020?ooo`030?noM00007@0_ooo00<0oooo00/0gia80000M02o
ool0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`210?ooo`00dP3oool00`3ooml0W4P0001d
_`070?ooo`070?oo_`1d04P0W=oo0?ooo`3OW4P0000004RLg`020?ooo`030=nLB000B4P0B7Bo0080
oooo00H0_g@0001d_`3oool0oono07@0001d_ol20?ooo`030?oog`2LB00007Bo0080oooo00L0okmd
0000M02oool0oooo0?oo_`1d0000M;oo0080oooo00<0gia80018W03Oool00P3oool01`3o_g@0001d
0;ooo`3oool0oono07@0001d_ol00P3oool02`3OW4P004Q807BLg`3oool0okmd0000M01dM7@0_ooo
0?oog`2LB000B9cO00L0oooo00<0oooO09a80000M;l00P3oool00`3ookl0W9cO0?ooo`020?ooo`03
0?oOW0180000M;oo0080oooo00<0oono07@0001d_ol00P3oool02`3o_g@0001d0;ooo`3oool0oono
07@0001d_ol0oooo0?oo_`1d0000M;oo00<0oooo00L0_g@00018W03Oool0oooo0=nLB000B9`0gooo
00<0oooo00L0_g@00018W03Oool0oooo0?oo_`1d0000M;oo00<0oooo00L0omnL04P0B02Lgol0oooo
0=nLB000B4P0B7Bo0080oooo00<0_g@0001d_`3oool0P03oool00=<0oooo00<0gia80000B018B000
0`0000006018W=l0oooo0?ooo`3oool0okmd0018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3o_g@0
001d0;noM018W=l0oooo0?noM00007@0_ooo0?ooo`3o_g@0001d0;ooo`80oooo00L0gia80000M02o
ool0omoO0=oo_`1d0000M;oo0080oooo00`0_g@0001d_`3oool0oooo0?oOW01807@0_ooo0?ooo`3o
gml0gono07@0001d_ol20?ooo`0@0;md0000M;l0oooo0?ooo`3o_g@0001d09bLM02Lgol0oono07@0
001d_kl0M7Bo0?ooo`2oM00004RL0=ooo`L0oooo00H0gia80000M02oool0oooo0?oOW018M;l40?oo
o`0<0;md000007@0_ooo0?ooo`3ooml0W4P004RLg`3oool0omoO0=oo_`1d0000M;oo0P3oool01`2o
M00007Bo0?ooo`3oool0okmd0000M02oool00P3oool01`3OW4P0001d0;ooo`3oool0omnL04P0B02L
gol00`3oool00`3OW4P0001809cOo`020?ooo`030=nLB00007@0_ooo0080oooo00`0oono07@0001d
_ol0oooo0?noM00007@0_kmd04RLg`3oool0okmd0000M02ooon00?ooo`00d`3oool00`3o_g@00018
09cOo`020?ooo`040?oOW0180000001d0;ooo`80oooo00l0gia80000M02oool0oooo0=nLB000B9`0
gooo0?oog`2LB000M;oo0=nLB000M;l0omnL04P0B02Lgol00P3oool02`3OW4P0001d0;ooo`3oool0
_g@00000B02Lgol0omnL04Qd_`2oM00007Bo0080oooo00`0okmd0018W03Oool0oooo0?oo_`1d0000
M;oo0?ooo`3ogi`0B7Bo0;md0000M;l20?ooo`0@0?noM000B9`0gooo0?ooo`3ookl0M01809cOg`2L
B4P0W=oO09a80018W=l0okmd001d_`3o_g@0001d0;ooo`L0oooo00<0okmd0000000000000P000000
0`0007@0_ooo0?ooo`020?ooo`0<0?oog`2LB00004RL0=ooo`3ookl0W4Q80018W03Oool0omnL04Qd
_`2oM00007Bo0P3oool00`3o_g@004RL0=ooo`020?ooo`0;0=nLB00007@0_ooo0?ooo`2oM0000018
09cOo`3oool0oooO09a80018W=l00`3oool00`3ogi`0B00004RLg`030?ooo`0@0;md000007@0_ooo
0?ooo`3o_g@0001d0;ooo`3oool0oooO09a8001d_ol0gia8001d_`3ogi`0B01809cOoh00oooo003C
0?ooo`030?oo_`1d0000B9cO00<0oooo00<0omnL04P00000M;l00`3oool03P3ogi`0B00000000000
0000M;oo0?oo_`1d0000000004RLg`3oool0okmd0000000004P0W=oo0`3oool02P3ogi`0B0000000
0018W;l0W4P004RLg`3oool0okmd00000000M;l20?ooo`030?oOW01804P0W=oo0080oooo00L0_g@0
04RLg`3oool0oooo0?noM000000007Bo0080oooo0100omnL04P0B02Lgol0oono07@000000000B9cO
0?oog`2LB000000004RLg`3oool0okmd00000000B9`0gooo1`3oool00`3ookl0M00007Boo`020?oo
o`030;mdB02Lgol0oooo00@0oooo00X0gia80000000007@0_kmd0000M02oool0oooo0?noM0000000
07Bo0P3oool00`3ogi`0B01809cOo`030?ooo`080?oOW0180000000004RLW01d0000B9cO0?oo_`1d
00030000000607Boo`3o_g@0000000000000B9`0gooo0`3oool03`3o_g@00000000000000000M;oo
0?ooo`3ookl0M00000000018W=l0oooo0?noM0000000001809cOo`200?ooo`00e03oool00`2oM000
04RL0=ooo`030?ooo`030;md000007@0_ooo01L0oooo00<0_g@0001d_`3oool02P3oool00`3OW4P0
04RL0=ooo`0N0?ooo`060;md0000M;l0oooo0?ooo`3ogi`0M9cO7`3oool00`3o_g@0001d0;ooo`2H
0?ooo`00e03oool00`3OW4P0001d0;ooo`030?ooo`030;md000007@0_ooo01L0oooo00<0okmd0000
M02oool02P3oool00`3ogi`0B01d0;ooo`0N0?ooo`030=nLB00007@0_ooo00@0oooo00<0gibL0;oo
o`3oool0703oool00`3OW7@0W=oo0?ooo`2G0?ooo`00e03oool00`3o_g@0000007Boo`020?ooo`03
0?noM000000007Bo01P0oooo00<0omnL04P0B02Lgol02P3oool00`3ooml0W4P007Boo`0N0?ooo`03
0?noM00004P0W=oo00<0oooo00<0oono07@0B02Lgol08P3oool0103ookl0M0000018W03OoonA0?oo
o`00d`3oool00`2oM00000000000000400000003001d_`3oool0oooo01L0oooo00<0oooO09a80018
W=l02`3oool00`3OW4P0B9cO0?ooo`0L0?ooo`030;md00000000000000H0000000<0M;oo0?ooo`3o
ool0]@3oool00?<0oooo00@0oono07@00000000007Bo2@3oool0103ooml0W4P000000000M;oN0?oo
o`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00o`3ooooS0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 481.625}, {384.188, 0}} -> {-0.332767, -0.217853, \
0.00152875, 0.00152875}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(plot2\  = 
      pendulumPlot[0.1, 0.1, 0.5, 1, 1, 0, 200, MaxSteps \[Rule] 2000, 
        PlotPoints \[Rule] 100, 
        PlotRange \[Rule] {{\(-0.2\), 0.2}, {\(-0.2\), 0.2}}, 
        DisplayFunction \[Rule] Identity]; 
    Show[{plot2, attractor}, 
      DisplayFunction \[Rule] $DisplayFunction];\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 2.5 0.5 2.5 [
[.25 -0.0125 -13.8438 -15.6875 ]
[.25 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.75 -0.0125 -9.5 -15.6875 ]
[.75 -0.0125 9.5 0 ]
[1 -0.0125 -9.5 -15.6875 ]
[1 -0.0125 9.5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -21.9688 -31.375 ]
[.5 -0.0125 21.9688 -15.6875 ]
[-0.0125 .25 -27.6875 -7.84375 ]
[-0.0125 .25 0 7.84375 ]
[-0.0125 .5 -10 -7.84375 ]
[-0.0125 .5 0 7.84375 ]
[-0.0125 .75 -19 -7.84375 ]
[-0.0125 .75 0 7.84375 ]
[-0.0125 1 -19 -7.84375 ]
[-0.0125 1 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .5 -43.375 -20.4688 ]
[-0.01875 .5 -27.6875 20.4688 ]
[ 0 1 .125 0 ]
[ 1 0 .125 0 ]
[.5 1.0125 -49.7813 0 ]
[.5 1.0125 49.7813 15.6875 ]
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.125 Mabswid
.05 0 m
.05 .00375 L
s
.1 0 m
.1 .00375 L
s
.15 0 m
.15 .00375 L
s
.2 0 m
.2 .00375 L
s
.3 0 m
.3 .00375 L
s
.35 0 m
.35 .00375 L
s
.4 0 m
.4 .00375 L
s
.45 0 m
.45 .00375 L
s
.55 0 m
.55 .00375 L
s
.6 0 m
.6 .00375 L
s
.65 0 m
.65 .00375 L
s
.7 0 m
.7 .00375 L
s
.8 0 m
.8 .00375 L
s
.85 0 m
.85 .00375 L
s
.9 0 m
.9 .00375 L
s
.95 0 m
.95 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -82.9688 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(position) show
102.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .25 m
.00625 .25 L
s
gsave
-0.0125 .25 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
0 .5 m
.00625 .5 L
s
gsave
-0.0125 .5 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .75 m
.00625 .75 L
s
gsave
-0.0125 .75 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
1.000 setlinewidth
grestore
0 1 m
.00625 1 L
s
gsave
-0.0125 1 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .05 m
.00375 .05 L
s
0 .1 m
.00375 .1 L
s
0 .15 m
.00375 .15 L
s
0 .2 m
.00375 .2 L
s
0 .3 m
.00375 .3 L
s
0 .35 m
.00375 .35 L
s
0 .4 m
.00375 .4 L
s
0 .45 m
.00375 .45 L
s
0 .55 m
.00375 .55 L
s
0 .6 m
.00375 .6 L
s
0 .65 m
.00375 .65 L
s
0 .7 m
.00375 .7 L
s
0 .8 m
.00375 .8 L
s
0 .85 m
.00375 .85 L
s
0 .9 m
.00375 .9 L
s
0 .95 m
.00375 .95 L
s
.25 Mabswid
0 0 m
0 1 L
s
gsave
-0.01875 .5 -117 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.4688 11.8438 translate
90 rotate
-81.4688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(velocity) show
99.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .99375 m
0 1 L
s
.25 .99375 m
.25 1 L
s
.5 .99375 m
.5 1 L
s
.75 .99375 m
.75 1 L
s
.125 Mabswid
.05 .99625 m
.05 1 L
s
.1 .99625 m
.1 1 L
s
.15 .99625 m
.15 1 L
s
.2 .99625 m
.2 1 L
s
.3 .99625 m
.3 1 L
s
.35 .99625 m
.35 1 L
s
.4 .99625 m
.4 1 L
s
.45 .99625 m
.45 1 L
s
.55 .99625 m
.55 1 L
s
.6 .99625 m
.6 1 L
s
.65 .99625 m
.65 1 L
s
.7 .99625 m
.7 1 L
s
.8 .99625 m
.8 1 L
s
.85 .99625 m
.85 1 L
s
.9 .99625 m
.9 1 L
s
.95 .99625 m
.95 1 L
s
.25 Mabswid
0 1 m
1 1 L
s
.99375 0 m
1 0 L
s
.99375 .25 m
1 .25 L
s
.99375 .5 m
1 .5 L
s
.99375 .75 m
1 .75 L
s
.125 Mabswid
.99625 .05 m
1 .05 L
s
.99625 .1 m
1 .1 L
s
.99625 .15 m
1 .15 L
s
.99625 .2 m
1 .2 L
s
.99625 .3 m
1 .3 L
s
.99625 .35 m
1 .35 L
s
.99625 .4 m
1 .4 L
s
.99625 .45 m
1 .45 L
s
.99625 .55 m
1 .55 L
s
.99625 .6 m
1 .6 L
s
.99625 .65 m
1 .65 L
s
.99625 .7 m
1 .7 L
s
.99625 .8 m
1 .8 L
s
.99625 .85 m
1 .85 L
s
.99625 .9 m
1 .9 L
s
.99625 .95 m
1 .95 L
s
.25 Mabswid
1 0 m
1 1 L
s
0 .5 m
1 .5 L
s
.5 0 m
.5 1 L
s
gsave
.5 1.0125 -110.781 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Pendulum) show
114.625 15.813 moveto
(Equation) show
158.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
.5 Mabswid
.79469 1 m
.8562 .93199 L
.90667 .86177 L
.94525 .794 L
.97671 .72205 L
.98975 .68377 L
s
.98975 .68377 m
1 .64638 L
s
1 .36172 m
.99577 .34648 L
.98753 .32224 L
.96479 .27105 L
.93497 .22247 L
.89602 .17601 L
.84948 .13599 L
.82537 .12005 L
.80159 .10693 L
.7791 .09666 L
.75398 .08741 L
.7307 .08079 L
.70882 .07617 L
.68257 .07259 L
.65865 .07114 L
.631 .07157 L
.60508 .07401 L
.57899 .07848 L
.56432 .08189 L
.55077 .08564 L
.52159 .09571 L
.49571 .10706 L
.45091 .13269 L
.40671 .16685 L
.36448 .21051 L
.32582 .26497 L
.31088 .29199 L
.29706 .32186 L
.28694 .34833 L
.27777 .378 L
.27097 .4065 L
.26627 .43337 L
.26304 .46205 L
.26155 .49284 L
.26199 .52164 L
.26404 .54873 L
.26722 .5729 L
.27207 .59884 L
.27908 .62693 L
.28877 .65717 L
.3145 .71568 L
.34527 .76594 L
.37958 .80867 L
.41915 .84701 L
.46065 .87832 L
.50072 .90185 L
.54681 .92223 L
.57138 .93052 L
.59904 .93787 L
.62835 .94348 L
Mistroke
.65617 .9468 L
.68219 .94819 L
.7061 .94802 L
.73229 .94622 L
.75972 .94253 L
.78727 .9369 L
.8167 .92867 L
.8684 .90822 L
.91592 .88175 L
.95663 .85191 L
.9966 .81422 L
Mfstroke
.9966 .81422 m
1 .81003 L
s
1 .17003 m
.98444 .1492 L
.9319 .09414 L
.88064 .05137 L
.82002 .01094 L
s
.82002 .01094 m
.79872 0 L
s
.12092 0 m
.09422 .01399 L
.02888 .05756 L
s
.02888 .05756 m
0 .08188 L
s
0 .86873 m
.01159 .87609 L
.06308 .90031 L
.09012 .91008 L
.1205 .91889 L
.1469 .92482 L
.17557 .92955 L
.20288 .93247 L
.2279 .93382 L
.24311 .93404 L
.25928 .93378 L
.28825 .93206 L
.30426 .93041 L
.31924 .92842 L
.35307 .92233 L
.38203 .9153 L
.40813 .90751 L
.46487 .88546 L
.51112 .86181 L
.55133 .83642 L
.5951 .8026 L
.63068 .76906 L
.68985 .69478 L
.71413 .65228 L
.7316 .61209 L
.73765 .59445 L
.74301 .57571 L
.7467 .55975 L
.74973 .54267 L
.75109 .5325 L
.75202 .5232 L
.75295 .50527 L
.75279 .48954 L
.75161 .47341 L
.75051 .46476 L
.74924 .45701 L
.74545 .44028 L
.74113 .42645 L
.73563 .41285 L
.724 .39183 L
.71635 .38143 L
.70863 .37277 L
.70098 .36564 L
.69223 .35892 L
.68419 .35389 L
.67537 .34948 L
.66724 .34633 L
.65811 .34377 L
.64969 .34226 L
.64175 .34155 L
Mistroke
.63232 .34159 L
.62243 .34264 L
.61407 .34435 L
.60506 .34705 L
.59552 .35093 L
.58564 .35613 L
.56888 .36813 L
.55251 .38476 L
.53785 .4059 L
.52782 .42642 L
.52325 .43892 L
.51977 .45108 L
.51817 .45798 L
.51672 .46562 L
.51561 .47293 L
.51482 .47979 L
.51396 .49292 L
.51396 .50707 L
.51433 .51439 L
.51501 .52231 L
.51592 .5298 L
.517 .53667 L
.52034 .55253 L
.52513 .56905 L
.53513 .59399 L
.54913 .61924 L
.56526 .64123 L
.58219 .65932 L
.60452 .67782 L
.61721 .68624 L
.63126 .69408 L
.64397 .69997 L
.65838 .70538 L
.67178 .70928 L
.68453 .71204 L
.69805 .71397 L
.70565 .71462 L
.71256 .71494 L
.72602 .71481 L
.73867 .71379 L
.74583 .71282 L
.75335 .71148 L
.76668 .70828 L
.78077 .70372 L
.79581 .69736 L
.81974 .68368 L
.84322 .66491 L
.86282 .64355 L
.87816 .62123 L
.88489 .609 L
Mistroke
.89124 .59537 L
.90044 .56936 L
.90426 .55416 L
.90698 .53945 L
.90897 .52289 L
.90962 .51339 L
.90992 .50459 L
.90976 .48957 L
.90886 .47542 L
.90703 .45972 L
.90409 .4428 L
.89989 .4249 L
.89489 .40797 L
.88281 .37638 L
.86569 .34252 L
.84598 .31169 L
.81807 .2766 L
.78853 .24652 L
.71746 .1927 L
.67293 .1683 L
.62778 .14912 L
.58737 .13603 L
.56463 .1302 L
.54293 .12561 L
.52241 .12213 L
.50005 .11924 L
.47961 .11743 L
.46076 .11643 L
.43816 .1161 L
.41711 .11662 L
.39406 .11812 L
.36953 .1208 L
.35726 .12257 L
.344 .12479 L
.32013 .12966 L
.27642 .14157 L
.23455 .15689 L
.19175 .17702 L
.12178 .22203 L
.08639 .2525 L
.0538 .28736 L
.0279 .32212 L
.00847 .355 L
Mfstroke
.00847 .355 m
0 .37372 L
s
0 .61895 m
.01255 .64227 L
.03164 .66823 L
.05182 .68985 L
.07663 .71096 L
.1054 .73003 L
.13236 .74392 L
.15875 .75446 L
.18884 .76336 L
.20609 .76712 L
.22197 .76978 L
.23966 .77189 L
.24863 .77262 L
.25852 .77317 L
.27604 .77351 L
.29245 .7731 L
.30233 .77251 L
.31169 .77173 L
.32917 .76967 L
.34781 .76662 L
.36787 .76235 L
.40086 .75304 L
.43452 .74049 L
.49014 .7122 L
.54306 .67454 L
.5822 .63702 L
.61427 .59593 L
.62634 .57604 L
.63627 .55611 L
.64263 .54007 L
.64666 .52691 L
.64806 .52112 L
.6492 .51524 L
.64996 .51012 L
.6504 .50575 L
.65061 .50116 L
.65057 .49731 L
.65027 .49361 L
.64999 .49177 L
.64969 .49024 L
.64898 .48772 L
.64851 .48654 L
.64803 .48558 L
.64751 .48475 L
.64699 .4841 L
.64637 .48353 L
.64579 .48316 L
.64479 .48289 L
.64371 .48306 L
.64271 .48364 L
Mistroke
.64185 .48453 L
.6409 .48598 L
.64042 .48699 L
.64 .48805 L
.63933 .49027 L
.63879 .49301 L
.63857 .49467 L
.63842 .49628 L
.63832 .49816 L
.63829 .50024 L
.63842 .50397 L
.63861 .50621 L
.63885 .50837 L
.63951 .51258 L
.64034 .51658 L
.64308 .52603 L
.64677 .53524 L
.65116 .54387 L
.65696 .55306 L
.6646 .56281 L
.6802 .57752 L
.69003 .58427 L
.69992 .58953 L
.70567 .59196 L
.71196 .59412 L
.71802 .59573 L
.72358 .59681 L
.72703 .59729 L
.73021 .59761 L
.73372 .59781 L
.73743 .59786 L
.74366 .59756 L
.74718 .59717 L
.75046 .59665 L
.75671 .59526 L
.76253 .59345 L
.76765 .59141 L
.77308 .58874 L
.77877 .58531 L
.78456 .58107 L
.79414 .57192 L
.8016 .56215 L
.80507 .55644 L
.80834 .54999 L
.81089 .54398 L
.81328 .53705 L
.81509 .53048 L
.81646 .52403 L
.81775 .51537 L
.81841 .50718 L
Mistroke
.81858 .49804 L
.81841 .49276 L
.81809 .4878 L
.81719 .47937 L
.81565 .46988 L
.81342 .45976 L
.81075 .45013 L
.80277 .42836 L
.79279 .40786 L
.77766 .38344 L
.76092 .36164 L
.72247 .32307 L
.67172 .28654 L
.61856 .25939 L
.58606 .24703 L
.5549 .23776 L
.5267 .23137 L
.51024 .22847 L
.49512 .22632 L
.479 .22458 L
.46165 .22332 L
.44551 .22271 L
.43023 .22264 L
.4141 .2231 L
.39677 .2242 L
.3805 .22582 L
.36569 .22779 L
.3318 .23416 L
.3151 .23829 L
.29719 .24349 L
.26431 .25525 L
.23111 .27041 L
.17736 .30377 L
.15462 .32228 L
.13263 .34378 L
.11566 .36378 L
.10028 .38578 L
.08876 .40619 L
.07873 .42901 L
.07521 .43909 L
.07204 .44992 L
.06964 .46 L
.06787 .46945 L
.06634 .48106 L
.06582 .48719 L
.06548 .49378 L
.06538 .50028 L
.06551 .50707 L
.06638 .51909 L
.06789 .53007 L
Mistroke
.0699 .54012 L
.07276 .55092 L
.07656 .56219 L
.08784 .5864 L
.10174 .6075 L
.11669 .62478 L
.1353 .64159 L
.15354 .65464 L
.17534 .66699 L
.19975 .67758 L
.22381 .68533 L
.23566 .68831 L
.24708 .69069 L
.25757 .69249 L
.26885 .69403 L
.28115 .69526 L
.29262 .696 L
.29966 .69627 L
.30629 .6964 L
.3187 .69633 L
.33031 .69588 L
.34272 .69503 L
.35436 .69387 L
.36491 .69253 L
.39036 .68818 L
.41385 .68279 L
.45819 .66906 L
.50202 .6509 L
.61241 .58592 L
.6322 .57283 L
.6503 .56206 L
.65775 .55833 L
.66515 .55521 L
.67118 .55318 L
.67755 .55159 L
.68098 .55098 L
.68405 .55057 L
.68575 .55039 L
.68758 .55025 L
.6909 .5501 L
.694 .55007 L
.69728 .55014 L
.70009 .55028 L
.70319 .5505 L
.71795 .55209 L
.72539 .55289 L
.72905 .55321 L
.73238 .55344 L
.73569 .55359 L
.73881 .55365 L
Mistroke
.7421 .55362 L
.74565 .55346 L
.7474 .55332 L
.7493 .55313 L
.75312 .5526 L
.75662 .55192 L
.75983 .5511 L
.76582 .54902 L
.77103 .54648 L
.77664 .54273 L
.78146 .53832 L
.78604 .53259 L
.78805 .5293 L
.79 .52538 L
.79144 .52179 L
.79273 .51766 L
.7937 .51353 L
.79435 .50954 L
.79464 .50706 L
.79484 .50433 L
.79496 .49923 L
.79487 .49599 L
.79465 .49285 L
.79436 .48998 L
.79394 .4868 L
.79283 .48068 L
.79119 .47398 L
.78702 .46142 L
.78118 .44824 L
.77404 .4353 L
.75631 .41029 L
.73073 .38313 L
.69537 .35467 L
.65852 .33196 L
.61388 .31112 L
.57203 .29681 L
.54696 .29033 L
.52277 .28545 L
.50988 .28338 L
.49599 .28154 L
.48278 .28018 L
.4707 .27926 L
.45645 .27857 L
.4484 .27837 L
.44096 .2783 L
.4273 .27849 L
.41286 .27911 L
.3991 .28012 L
.3844 .28166 L
.37063 .28354 L
Mistroke
.35816 .28561 L
.32825 .29212 L
.30083 .30012 L
.25612 .31801 L
.21883 .33866 L
.18353 .36512 L
.15328 .39668 L
.14071 .41423 L
.13 .43325 L
.12299 .44972 L
.11989 .45912 L
.11758 .46788 L
.11653 .47275 L
.11559 .47803 L
.11488 .483 L
.11439 .48758 L
.11388 .49621 L
.11393 .50528 L
.11419 .50991 L
.11466 .51486 L
.11526 .51951 L
.11596 .52373 L
.1181 .53339 L
.12112 .54332 L
.12738 .5583 L
.13604 .57346 L
.14599 .5869 L
.15642 .59831 L
.17028 .61069 L
.18705 .62276 L
.20306 .63211 L
.22148 .64088 L
.24033 .64809 L
.2583 .65358 L
.27695 .65807 L
.29509 .66139 L
.30545 .66286 L
.31655 .66413 L
.32711 .66505 L
.3367 .66566 L
.34262 .66593 L
.34807 .66611 L
.35405 .66623 L
.36037 .66628 L
.371 .66617 L
.37702 .66601 L
.38263 .6658 L
.39369 .66521 L
.40564 .66431 L
.42715 .66207 L
Mistroke
.44776 .65921 L
.46643 .65604 L
.54778 .63672 L
.66013 .5996 L
.7096 .58129 L
.74652 .56591 L
.76156 .55822 L
.77576 .54909 L
.78142 .54455 L
.78684 .53939 L
.79128 .53422 L
.79481 .52915 L
.79837 .52241 L
.79986 .51866 L
.80113 .51453 L
.80197 .5108 L
.80253 .5072 L
.80275 .50509 L
.80289 .50313 L
.80296 .50083 L
.80295 .49861 L
.80275 .49474 L
.80227 .4905 L
.80165 .4868 L
.80071 .48261 L
.7977 .47295 L
.79365 .46349 L
.78272 .44486 L
.76554 .42358 L
.74159 .40101 L
.71382 .38042 L
.68368 .36245 L
.65151 .34693 L
.61045 .33147 L
.57005 .32023 L
.54787 .31557 L
.52666 .31205 L
.51456 .31044 L
.50298 .30917 L
.48955 .30802 L
.47722 .30728 L
.46506 .30683 L
.4521 .30667 L
.43987 .30682 L
.42881 .3072 L
.42201 .30756 L
.41578 .30797 L
.40178 .30918 L
.38953 .31058 L
.3767 .31238 L
Mistroke
.35589 .31608 L
.33412 .32104 L
.29661 .3325 L
.25571 .35003 L
.22279 .36925 L
.19345 .39217 L
.17986 .40569 L
.1688 .41886 L
.15342 .44294 L
.14745 .45601 L
.14276 .47018 L
.14102 .47762 L
.13985 .48453 L
.13937 .48849 L
.13905 .49207 L
.13883 .4962 L
.13877 .50008 L
.13884 .50385 L
.13905 .50782 L
.13979 .51489 L
.1404 .51878 L
.1411 .5224 L
.14314 .53054 L
.14536 .53734 L
.14787 .54368 L
.15363 .55526 L
.16131 .56723 L
.17133 .57959 L
.19349 .59994 L
.21859 .61652 L
.25016 .63167 L
.26948 .63875 L
.28804 .64432 L
.30564 .64866 L
.32498 .65251 L
.34671 .65579 L
.3673 .65802 L
.37852 .65889 L
.38446 .65926 L
.39077 .65959 L
.39628 .65982 L
.40237 .66001 L
.41309 .66021 L
.42351 .66023 L
.42936 .66017 L
.4347 .66007 L
.44506 .65976 L
.45488 .65932 L
.46497 .65873 L
.47583 .65795 L
Mistroke
.49535 .65614 L
.5311 .65157 L
.57129 .64455 L
.60639 .63679 L
.665 .62015 L
.71802 .59994 L
.7609 .57724 L
.77996 .5634 L
.79397 .54999 L
.80395 .53678 L
.80822 .529 L
.81126 .52158 L
.81227 .51841 L
.81317 .51495 L
.81395 .51111 L
.81447 .50745 L
.81479 .50353 L
.81488 .49934 L
.81475 .49573 L
.81439 .49174 L
.81367 .48709 L
.81274 .48276 L
.8114 .47796 L
.80963 .47283 L
.80581 .4641 L
.80039 .45438 L
.78695 .43623 L
.77035 .41924 L
.746 .39976 L
.71928 .38275 L
.68343 .36466 L
.64496 .3496 L
.6047 .33767 L
.56591 .32934 L
.54363 .32584 L
.53176 .32434 L
.51913 .32303 L
.50799 .3221 L
.50206 .3217 L
.49571 .32133 L
.48451 .32086 L
.47405 .32062 L
.46278 .32058 L
.45237 .32074 L
.44097 .32114 L
.43522 .32144 L
.42891 .32183 L
.41837 .32264 L
.40709 .32375 L
.3869 .32634 L
Mistroke
.36635 .32984 L
.34725 .3339 L
.31279 .34345 L
.27454 .35799 L
.24339 .37381 L
.21235 .39475 L
.18851 .41659 L
.17961 .42698 L
.1715 .43822 L
.16559 .44813 L
.16041 .45889 L
.15802 .46498 L
.15618 .47056 L
.15463 .47625 L
.1535 .48152 L
.15253 .48764 L
.15218 .49087 L
.15191 .49429 L
.15175 .50003 L
.15195 .50633 L
.15253 .51229 L
.1534 .5178 L
.15471 .5238 L
.15653 .53014 L
.16157 .54287 L
.16766 .55409 L
.18328 .57466 L
.20243 .5925 L
.22943 .61092 L
.25838 .62556 L
.2939 .63883 L
.31306 .64439 L
.33495 .6496 L
.35417 .65331 L
.37561 .65658 L
.39623 .65895 L
.40618 .65983 L
.41559 .66052 L
.42487 .66106 L
.43492 .6615 L
.44408 .66176 L
.45251 .6619 L
.46167 .66193 L
.47171 .66183 L
.48218 .66157 L
.49193 .66119 L
.50258 .66063 L
.5141 .65985 L
.53487 .65798 L
.55482 .65563 L
Mistroke
.57296 .653 L
.61284 .64556 L
.65268 .63558 L
.71264 .61448 L
.74156 .60059 L
.76463 .58673 L
.78338 .57257 L
.79774 .55868 L
.80941 .54351 L
.81416 .53525 L
.81811 .52625 L
.81963 .52179 L
.82096 .51684 L
.82151 .51434 L
.82199 .51163 L
.82236 .50907 L
.82261 .50673 L
.82289 .50223 L
.82287 .49731 L
.82255 .49263 L
.82199 .48829 L
.82104 .48329 L
.81984 .47858 L
.81598 .46781 L
.81073 .4573 L
.80341 .44601 L
.78684 .42684 L
.76405 .40734 L
.74045 .39156 L
.71053 .37557 L
.6751 .3607 L
.63821 .34872 L
.60274 .33999 L
.58206 .33601 L
.56228 .33292 L
.54071 .33032 L
.52914 .32925 L
.51663 .32834 L
.51035 .32798 L
.50349 .32766 L
.497 .32743 L
.49096 .32728 L
.47957 .32715 L
.47319 .32718 L
.46742 .32726 L
.45604 .32758 L
.44559 .32806 L
.43417 .32881 L
.42217 .32983 L
.40216 .33212 L
Mistroke
.38056 .33541 L
.35977 .33946 L
.33781 .34475 L
.3 .35661 L
.2664 .37075 L
.23867 .38588 L
.21285 .40407 L
.1898 .42601 L
.18051 .43761 L
.17229 .45043 L
.16627 .46278 L
.16403 .46865 L
.16227 .47423 L
.16088 .47972 L
.15974 .48569 L
.15905 .49112 L
.15869 .49613 L
.15862 .50153 L
.15892 .50746 L
.15921 .51043 L
.15964 .51362 L
.1607 .51935 L
.16201 .52465 L
.16379 .53038 L
.16799 .54066 L
.17324 .55048 L
.1791 .55936 L
.19789 .5806 L
.21961 .59825 L
.24972 .61642 L
.28237 .63108 L
.31491 .64212 L
.35326 .65171 L
.37346 .65552 L
.39559 .65882 L
.41685 .66121 L
.42633 .66203 L
.43644 .66276 L
.44579 .66329 L
.45612 .66372 L
.46553 .66398 L
.47437 .6641 L
.48476 .6641 L
.49427 .66396 L
.50469 .66365 L
.51563 .66317 L
.52456 .66264 L
.53406 .66196 L
.55103 .66041 L
.57098 .65804 L
Mistroke
.58879 .65542 L
.62644 .64818 L
.66401 .6384 L
.69387 .62846 L
.72392 .616 L
.75263 .60097 L
.77877 .5832 L
.79864 .56512 L
.81243 .54769 L
.81747 .53915 L
.82188 .52956 L
.82363 .52469 L
.82516 .5194 L
.82627 .51437 L
.82699 .50974 L
.82742 .5054 L
.82759 .50135 L
.82754 .4969 L
.82721 .49213 L
.82654 .48711 L
.82561 .48236 L
.82453 .47811 L
.82309 .47349 L
.8191 .46365 L
.81326 .4529 L
.79692 .43162 L
.77653 .41276 L
.75308 .39614 L
.72477 .38035 L
.68957 .36506 L
.65274 .3528 L
.61597 .34358 L
.59642 .33974 L
.57762 .3367 L
.56025 .33443 L
.54167 .33255 L
.53192 .33178 L
.52118 .33112 L
.51138 .33067 L
.50222 .33039 L
.49138 .33023 L
.4815 .33025 L
.47011 .33046 L
.45934 .33085 L
.44952 .33137 L
.43899 .3321 L
.42006 .33388 L
.39787 .33675 L
.37793 .3401 L
.33831 .34911 L
Mistroke
.29948 .36149 L
.26878 .37443 L
.24312 .38813 L
.21855 .40476 L
.19687 .42407 L
.18754 .43467 L
.17922 .4462 L
.17333 .45637 L
.1684 .46746 L
.16626 .47374 L
.1647 .4795 L
.16353 .48531 L
.16277 .49083 L
.1625 .49394 L
.16233 .49724 L
.16229 .50036 L
.16235 .50318 L
.16253 .5065 L
.16287 .51009 L
.16333 .51347 L
.16387 .51656 L
.16501 .5217 L
.16657 .52716 L
.17025 .5369 L
.17608 .54829 L
.18272 .55842 L
.20142 .57947 L
.22415 .59787 L
.24896 .61319 L
.28021 .6281 L
.31664 .64122 L
.35368 .65108 L
.3722 .65491 L
.39021 .65802 L
.40686 .66038 L
.42486 .66241 L
.43517 .66334 L
.44455 .66404 L
.45414 .66461 L
.46298 .66502 L
.47336 .66534 L
.47882 .66544 L
.4846 .66551 L
.49432 .6655 L
.50498 .66534 L
.51594 .66499 L
.52605 .66451 L
.53711 .66382 L
.54881 .66288 L
.56826 .66087 L
Mistroke
.5894 .658 L
.62719 .65103 L
.66089 .64257 L
.69743 .63054 L
.72788 .61765 L
.75742 .60167 L
.78363 .58305 L
.8025 .56503 L
.81555 .54779 L
.82105 .53797 L
.82343 .53273 L
.82557 .52715 L
.82711 .52223 L
.82845 .5168 L
.82935 .51183 L
.8299 .50717 L
.83019 .50256 L
.83019 .49752 L
.8299 .49276 L
.82938 .48834 L
.82862 .48396 L
.82773 .48001 L
.82502 .47117 L
.82099 .46169 L
.81537 .45156 L
.80188 .43363 L
.7829 .4153 L
.76118 .39924 L
.73319 .38303 L
.69886 .36766 L
.65865 .35408 L
.63754 .34847 L
.61736 .34396 L
.59897 .34051 L
.57826 .33735 L
.55707 .33485 L
.54535 .33378 L
.53429 .33298 L
.52423 .33241 L
.51347 .33198 L
.50329 .33173 L
.49404 .33165 L
.48315 .33172 L
.47701 .33185 L
.47136 .33201 L
.46099 .33244 L
.45007 .33308 L
.43963 .33387 L
.42849 .33491 L
.40858 .33728 L
Mistroke
.38533 .34093 L
.36463 .34502 L
.32972 .35392 L
.29435 .36594 L
.26398 .37936 L
.23888 .39337 L
.2157 .40979 L
.1952 .42893 L
.18744 .43813 L
.18044 .44805 L
.175 .45752 L
.17101 .46625 L
.16744 .4768 L
.16604 .48263 L
.1651 .48816 L
.16454 .49329 L
.1643 .49802 L
.16433 .50321 L
.16472 .50871 L
.16546 .51422 L
.16604 .5173 L
.16666 .52012 L
.16987 .5307 L
.17223 .53655 L
.17528 .54288 L
.18208 .55421 L
.19963 .57528 L
.22121 .59379 L
.24446 .60906 L
.27333 .62389 L
.30879 .63785 L
.34995 .64977 L
.38747 .65749 L
.40951 .66082 L
.43046 .66322 L
.44098 .66416 L
.45221 .66497 L
.46184 .66551 L
.47249 .66595 L
.47846 .66611 L
.48393 .66622 L
.48994 .66628 L
.49627 .66629 L
.50694 .66617 L
.51297 .66603 L
.51858 .66585 L
.52947 .66536 L
.53971 .66474 L
.55874 .66316 L
.57899 .66085 L
Mistroke
.60056 .65765 L
.63739 .65032 L
.67417 .64032 L
.70338 .63008 L
.73276 .61717 L
.76176 .6009 L
.78508 .58391 L
.80224 .56758 L
.81661 .54887 L
.8227 .53789 L
.82688 .52779 L
.82864 .52213 L
.8294 .51914 L
.83009 .51599 L
.83059 .51321 L
.83103 .51017 L
.83153 .5048 L
.83167 .49936 L
.83148 .4944 L
.83124 .49166 L
.83088 .48868 L
.82995 .48328 L
.82875 .47823 L
.82713 .4729 L
.82336 .4634 L
.81802 .45325 L
.81174 .44375 L
.79688 .42643 L
.77656 .40875 L
.74959 .39092 L
.72244 .37691 L
.69008 .36375 L
.65676 .35321 L
.62458 .34539 L
.58535 .33851 L
.56502 .33599 L
.55322 .33483 L
.54239 .33397 L
.53239 .33333 L
.52288 .33287 L
.51282 .33254 L
.50185 .33235 L
.49229 .33233 L
.48212 .33247 L
.47345 .33272 L
.4639 .33312 L
.45278 .33377 L
.44112 .33466 L
.42045 .33678 L
.39855 .33981 L
Mistroke
.37511 .34399 L
.33487 .35368 L
.30003 .3651 L
.26716 .37917 L
.2359 .3967 L
.21113 .41512 L
.19273 .43341 L
.18005 .45075 L
.17456 .46087 L
.17056 .47048 L
.16887 .47571 L
.16757 .48069 L
.16665 .48518 L
.16593 .49009 L
.16547 .49542 L
.16536 .4984 L
.16535 .5011 L
.16559 .50635 L
.16611 .5113 L
.16692 .51627 L
.16813 .52158 L
.17115 .53109 L
.17614 .54226 L
.18201 .55226 L
.19898 .57324 L
.22011 .59185 L
.2435 .60759 L
.26825 .62081 L
.29853 .63366 L
.33375 .6452 L
.37115 .6543 L
.40715 .66052 L
.42654 .66293 L
.44802 .6649 L
.45895 .66562 L
.46497 .66594 L
.47066 .66618 L
.48175 .66653 L
.49182 .66668 L
.5008 .66669 L
.51067 .66655 L
.52095 .66626 L
.53055 .66584 L
.53949 .66532 L
.54763 .66474 L
.56582 .66307 L
.58508 .66074 L
.6057 .65755 L
.64199 .65008 L
.67773 .64012 L
Mistroke
.70886 .62887 L
.73979 .61462 L
.76884 .59726 L
.79269 .57835 L
.80349 .56743 L
.81207 .55694 L
.81871 .54698 L
.82364 .53776 L
.82785 .52738 L
.82959 .52162 L
.83086 .51614 L
.83167 .51137 L
.83222 .50622 L
.83244 .50135 L
.8324 .4969 L
.83203 .49169 L
.83168 .48875 L
.83126 .48602 L
.83026 .48103 L
.82885 .47573 L
.82461 .46429 L
.81915 .45367 L
.8052 .43434 L
.78771 .41694 L
.76812 .40189 L
.74301 .38662 L
.71275 .37216 L
.67885 .35958 L
.6454 .35003 L
.62603 .34559 L
.60437 .34148 L
.58522 .33854 L
.56373 .33597 L
.55413 .33506 L
.54373 .33424 L
.5347 .33367 L
.52479 .33319 L
.5142 .33284 L
.50302 .33266 L
.49266 .33266 L
.48303 .33281 L
.47327 .33311 L
.46407 .33352 L
.45583 .334 L
.44694 .33464 L
.42711 .33652 L
.40906 .33878 L
.37144 .34534 L
.35109 .35 L
.33307 .35486 L
Mistroke
.29959 .36596 L
.27112 .378 L
.24166 .39385 L
.21448 .41312 L
.1952 .43152 L
.17984 .45224 L
.1742 .46303 L
.17173 .46899 L
.16982 .47458 L
.16838 .47978 L
.16736 .48454 L
.16655 .48983 L
.16624 .49275 L
.16605 .49551 L
.16595 .49824 L
.16594 .50124 L
.16604 .50438 L
.16625 .50732 L
.16688 .51257 L
.16737 .51551 L
.16792 .5182 L
.16937 .5239 L
.1714 .5301 L
.17689 .54245 L
.18336 .55327 L
.19048 .56287 L
.21063 .58351 L
.23419 .60124 L
.26037 .61646 L
.28952 .62981 L
.32492 .64236 L
.34391 .64778 L
.36503 .65289 L
.38382 .65669 L
.40475 .66016 L
.42592 .6629 L
.43534 .66388 L
.4458 .66481 L
.45501 .66548 L
.46499 .66606 L
.47556 .66651 L
.48551 .66678 L
.49458 .66689 L
.50452 .66687 L
.51488 .66669 L
.52452 .66637 L
.53354 .66595 L
.54173 .66546 L
.56 .66399 L
.57911 .66188 L
Mistroke
.59983 .65892 L
.64032 .65093 L
.67568 .64125 L
.70646 .63035 L
.73925 .61544 L
.76553 .60006 L
.79096 .58055 L
.80135 .57042 L
.80984 .56059 L
.81671 .55102 L
.8228 .54046 L
.82711 .53072 L
.83003 .52166 L
.83122 .51654 L
.83204 .51176 L
.83263 .50649 L
.83281 .50354 L
.83288 .50084 L
.83276 .49542 L
.83253 .49235 L
.83222 .48952 L
.83139 .48433 L
.83016 .47884 L
.82732 .46982 L
.82315 .4602 L
.81273 .44314 L
.80502 .43353 L
.79564 .42368 L
.77572 .40688 L
.7498 .39005 L
.72272 .37626 L
.68774 .36237 L
.65162 .35149 L
.61595 .34352 L
.58164 .33806 L
.56329 .33596 L
.54288 .33426 L
.53253 .33365 L
.52138 .33317 L
.51084 .3329 L
.50119 .33279 L
.49017 .33285 L
.47981 .33307 L
.47053 .33341 L
.46053 .33392 L
.44962 .33465 L
.43818 .33563 L
.41791 .33787 L
.39938 .34051 L
.38242 .34346 L
Mistroke
.34524 .35183 L
.30718 .36355 L
.27553 .37633 L
.24316 .39334 L
.21721 .41133 L
.1984 .42853 L
.18249 .44859 L
.17648 .45901 L
.17375 .46484 L
.17154 .47042 L
.16983 .47562 L
.16835 .48131 L
.16734 .48652 L
.16669 .49136 L
.16633 .49626 L
.16626 .50157 L
.16651 .50658 L
.167 .51119 L
.16766 .51535 L
.16863 .5199 L
.17135 .52905 L
.17546 .53891 L
.18118 .54933 L
.19487 .56759 L
.21574 .58744 L
.23964 .60444 L
.2705 .62124 L
.30798 .63662 L
.3437 .64762 L
.37951 .65582 L
.40079 .65953 L
.4207 .66229 L
.44291 .6646 L
.45418 .66547 L
.46019 .66586 L
.46657 .6662 L
.47774 .66666 L
.48807 .66691 L
.49919 .667 L
.50551 .66696 L
.51142 .66688 L
.52208 .66659 L
.53181 .66618 L
.543 .66552 L
.55358 .66472 L
.57289 .66281 L
.59058 .66052 L
.62955 .65357 L
.65066 .64862 L
.66984 .64329 L
Mistroke
.70363 .63173 L
.73657 .6171 L
.76749 .59905 L
.79018 .58157 L
.80725 .56415 L
.81504 .55394 L
.82116 .54407 L
.82401 .53854 L
.82667 .53249 L
.82868 .52699 L
.83023 .5218 L
.83142 .51671 L
.83235 .51122 L
.8327 .50824 L
.83296 .5051 L
.83308 .50242 L
.83311 .49949 L
.83303 .49638 L
.83286 .49354 L
.83254 .49028 L
.83212 .4872 L
.83109 .48173 L
.82957 .47583 L
.82572 .46521 L
.82016 .45413 L
.80642 .43481 L
.78764 .41612 L
.76651 .40016 L
.73777 .38335 L
.70662 .36926 L
.67498 .35807 L
.64354 .3494 L
.60712 .34189 L
.58812 .33893 L
.56675 .33633 L
.54714 .33459 L
.53594 .33387 L
.52547 .33337 L
.51436 .33303 L
.50425 .33288 L
.49864 .33286 L
.49256 .3329 L
.48153 .33311 L
.4721 .33343 L
.46175 .33395 L
.45097 .33466 L
.44093 .33549 L
.42312 .33735 L
.40423 .3399 L
.38271 .34356 L
Mistroke
.35986 .34838 L
.32033 .35928 L
.285 .37237 L
.25554 .38647 L
.22831 .40324 L
.204 .42314 L
.19475 .43279 L
.18634 .44328 L
.18023 .45255 L
.1749 .46266 L
.17272 .46777 L
.17095 .47264 L
.1695 .4774 L
.16841 .48182 L
.1676 .48601 L
.16696 .49047 L
.16657 .49496 L
.16642 .49991 L
.16658 .50525 L
.16703 .51014 L
.16786 .5155 L
.16841 .51818 L
.16911 .52112 L
.17237 .53133 L
.17654 .54073 L
.18271 .55143 L
.18965 .5611 L
.20964 .58207 L
.23195 .59928 L
.25997 .61588 L
.28853 .6291 L
.31785 .63989 L
.35251 .64982 L
.39209 .65808 L
.41281 .66127 L
.42433 .66272 L
.43526 .6639 L
.44556 .66484 L
.45674 .66566 L
.46694 .66625 L
.47634 .66664 L
.48178 .66681 L
.48761 .66694 L
.49813 .66704 L
.50456 .66702 L
.51064 .66695 L
.5161 .66683 L
.52202 .66665 L
.52778 .66643 L
.53413 .66612 L
Mistroke
.54693 .66532 L
.55873 .66435 L
.56972 .66325 L
.59108 .66055 L
.61102 .65732 L
.64722 .64961 L
.68386 .639 L
.72046 .62488 L
.74916 .61052 L
.77568 .59344 L
Mfstroke
1 0 0 r
.44113 .66452 m
.44893 .66518 L
.45623 .6657 L
.46493 .66622 L
.47286 .66659 L
.47687 .66674 L
.48122 .66688 L
.48534 .66698 L
.48914 .66705 L
.49122 .66708 L
.49316 .6671 L
.4953 .66712 L
.4965 .66712 L
.4976 .66713 L
.4998 .66713 L
.50104 .66713 L
.5022 .66713 L
.50432 .66711 L
.5055 .6671 L
.50657 .66709 L
.50871 .66707 L
.51106 .66703 L
.5158 .66693 L
.52019 .66681 L
.52429 .66667 L
.53351 .66625 L
.54337 .66567 L
.56132 .66421 L
.57754 .66245 L
.60868 .65785 L
.64137 .65114 L
.67102 .64318 L
.7006 .63314 L
.74775 .61141 L
.76845 .59866 L
.78785 .58388 L
.80344 .56881 L
.81645 .5522 L
.82115 .54448 L
.82534 .53613 L
.82849 .52816 L
.83067 .52086 L
.8316 .51679 L
.83236 .51248 L
.83284 .50879 L
.83303 .50669 L
.83317 .50472 L
.83322 .50364 L
.83325 .50306 L
.83327 .50245 L
Mistroke
.83328 .50187 L
.83329 .50134 L
.8333 .5003 L
.8333 .4993 L
.83328 .49822 L
.83326 .4972 L
.83322 .49626 L
.83316 .49517 L
.8331 .49415 L
.8329 .49184 L
.83265 .48959 L
.83231 .48719 L
.8315 .48273 L
.83054 .47862 L
.8276 .46938 L
.82415 .46131 L
.81962 .45285 L
.8087 .43728 L
.7957 .42322 L
.75988 .39572 L
.73724 .38296 L
.71075 .37084 L
.65689 .35278 L
.62669 .3456 L
.59307 .33966 L
.57618 .33742 L
.55766 .3355 L
.5491 .3348 L
.54001 .33418 L
.53146 .33371 L
.52365 .33338 L
.5197 .33325 L
.5154 .33314 L
.51134 .33305 L
.50953 .33303 L
.50757 .333 L
.50552 .33298 L
.5036 .33297 L
.50248 .33296 L
.50131 .33296 L
.4992 .33296 L
.497 .33296 L
.49496 .33298 L
.49271 .333 L
.49031 .33303 L
.48628 .3331 L
.48192 .3332 L
.47401 .33347 L
.46632 .33382 L
.45924 .33422 L
Mistroke
.44296 .33544 L
.42784 .33695 L
.39406 .34168 L
.36202 .34804 L
.33402 .35528 L
.30463 .36484 L
.27929 .37503 L
.23515 .39879 L
.21547 .41317 L
.20032 .42701 L
.18746 .442 L
.182 .44995 L
.17693 .45881 L
.17328 .46675 L
.17024 .47534 L
.16895 .48012 L
.16801 .48448 L
.16764 .48659 L
.1673 .48888 L
.16706 .49089 L
.16685 .49306 L
.16677 .4941 L
.1667 .4952 L
.16664 .49621 L
.1666 .49716 L
.16658 .49805 L
.16656 .499 L
.16656 .50005 L
.16656 .50101 L
.16657 .50156 L
.16658 .50214 L
.16662 .50318 L
.16667 .50429 L
.16674 .50552 L
.16692 .50772 L
.16713 .50978 L
.16778 .51425 L
.16858 .51832 L
.17112 .52743 L
.17429 .53564 L
.17796 .54317 L
.18886 .55987 L
.20308 .5758 L
.21877 .58953 L
.2394 .60402 L
.26097 .6163 L
.31105 .63755 L
.33832 .64606 L
.36993 .65382 L
.40117 .65958 L
Mistroke
.41912 .66211 L
.43568 .66397 L
.45237 .6654 L
.46037 .66593 L
.46789 .66634 L
.47603 .66669 L
.48035 .66683 L
.48501 .66695 L
.48733 .66699 L
.48987 .66704 L
.49227 .66707 L
.4945 .66709 L
.49671 .6671 L
.49871 .66711 L
.49982 .66711 L
.50102 .66711 L
.5032 .6671 L
.50531 .66709 L
.50723 .66707 L
.50945 .66704 L
.51155 .667 L
.51529 .66692 L
.51935 .66682 L
.52854 .66648 L
.53703 .66605 L
.55423 .66484 L
.57235 .66306 L
.60425 .65861 L
.63326 .653 L
.66392 .64528 L
.69051 .63685 L
.718 .6261 L
.74145 .61487 L
.78118 .5894 L
.79824 .57432 L
.81095 .55993 L
.82116 .54455 L
.82558 .53568 L
.82878 .52741 L
.82999 .52351 L
.83109 .51931 L
.83186 .51567 L
.83252 .51169 L
.83282 .50927 L
.83304 .50703 L
.8332 .50488 L
.83326 .50368 L
.8333 .50259 L
.83333 .50152 L
.83334 .50053 L
Mistroke
.83334 .49945 L
.83334 .49891 L
.83333 .49831 L
.8333 .4973 L
.83326 .49622 L
.8332 .4952 L
.83315 .49428 L
.83299 .49234 L
.83276 .49021 L
.83214 .48591 L
.83137 .48192 L
.83046 .4782 L
.82782 .46985 L
.82402 .46095 L
.81437 .44468 L
.8007 .42817 L
.78474 .41343 L
.76732 .40049 L
.72269 .37593 L
.69441 .3645 L
.66561 .35517 L
.63524 .34741 L
.60599 .34167 L
.57526 .33728 L
.55747 .33546 L
.54069 .3342 L
.53225 .33373 L
.52778 .33352 L
.52303 .33334 L
.51889 .33321 L
.51431 .33309 L
.51231 .33305 L
.51015 .33302 L
.50812 .33299 L
.50623 .33297 L
.50438 .33295 L
.50241 .33294 L
.50021 .33294 L
.49822 .33294 L
.49614 .33295 L
.49496 .33296 L
.49388 .33297 L
.49189 .33299 L
.48978 .33302 L
.48607 .33309 L
.4826 .33317 L
.47471 .33343 L
.46629 .33381 L
.45714 .33434 L
.44056 .33565 L
Mistroke
.42456 .33732 L
.40987 .33922 L
.37682 .34487 L
.34803 .35146 L
.3197 .35968 L
.29102 .37007 L
.26624 .38116 L
.24172 .39465 L
.22103 .4088 L
.20485 .42257 L
.19094 .43756 L
.18451 .44618 L
.17942 .45429 L
.17554 .46171 L
.17205 .47 L
.16941 .4784 L
.16832 .48303 L
.16754 .48739 L
.16704 .49135 L
.16683 .4936 L
.16676 .4946 L
.1667 .49569 L
.16667 .49628 L
.16665 .49683 L
.16661 .4979 L
.16659 .49888 L
.16658 .49992 L
.16659 .50103 L
.16661 .50207 L
.16663 .50267 L
.16665 .50321 L
.1667 .50444 L
.16676 .5054 L
.16683 .50643 L
.167 .50827 L
.16752 .51245 L
.16821 .51643 L
.16901 .52003 L
.17149 .52844 L
.17453 .5361 L
.184 .55309 L
.19563 .56798 L
.2116 .58359 L
.23128 .59866 L
.25193 .61141 L
.30074 .6338 L
.32771 .64293 L
.35886 .65131 L
.39007 .65771 L
.40646 .66036 L
Mistroke
.42439 .66273 L
.44137 .66449 L
.45705 .66571 L
.46463 .66616 L
.47292 .66655 L
.47671 .6667 L
.48076 .66682 L
.48424 .66692 L
.48806 .66699 L
.49035 .66703 L
.49249 .66706 L
.49474 .66708 L
.49601 .66709 L
.49716 .66709 L
.49922 .6671 L
.50146 .6671 L
.50358 .66709 L
.50554 .66707 L
.50778 .66705 L
.50989 .66702 L
.5147 .66693 L
.51939 .6668 L
.52443 .66663 L
.53311 .66625 L
.54223 .66572 L
.55845 .66446 L
.57434 .66282 L
.59113 .66064 L
.62071 .65562 L
.65372 .64808 L
.68322 .63935 L
.73389 .61874 L
.75666 .60629 L
.77791 .59195 L
.79499 .57752 L
.80799 .56366 L
.81906 .54819 L
.82384 .53946 L
.8275 .53106 L
.83014 .52306 L
.83128 .51856 L
.8321 .51448 L
.83268 .51064 L
.83293 .50848 L
.83311 .50649 L
.83323 .50475 L
.83328 .50375 L
.83331 .50283 L
.83334 .50185 L
.83336 .50096 L
Mistroke
.83336 .49994 L
.83336 .49896 L
.83334 .498 L
.83331 .49696 L
.83327 .49598 L
.83322 .49508 L
.83307 .493 L
.83285 .4908 L
.83229 .48667 L
.83144 .48216 L
.82932 .47417 L
.82665 .46678 L
.82262 .45811 L
.81797 .45002 L
.8071 .43525 L
.79232 .41997 L
.77601 .40659 L
.75518 .39277 L
.7332 .38086 L
.68314 .36057 L
.65337 .35179 L
.62474 .34515 L
.59417 .33978 L
.57647 .33741 L
.55986 .33566 L
.55131 .33493 L
.54203 .33427 L
.53383 .3338 L
.52515 .33341 L
.51729 .33315 L
.51282 .33305 L
.51068 .33302 L
.50866 .33299 L
.50682 .33297 L
.50482 .33295 L
.50301 .33294 L
.50133 .33293 L
.49952 .33293 L
.49753 .33294 L
.49544 .33295 L
.49349 .33296 L
.49138 .33299 L
.48912 .33302 L
.48506 .3331 L
.48023 .33323 L
.47586 .33338 L
.46725 .33375 L
.45916 .33421 L
.44419 .33532 L
.42812 .33691 L
Mistroke
.39545 .34145 L
.37939 .34436 L
.36212 .34803 L
.33028 .35639 L
.30296 .36547 L
.2539 .38759 L
.23325 .40009 L
.21366 .41472 L
.19797 .42952 L
.18465 .446 L
.17986 .45355 L
.17551 .46181 L
.1722 .46964 L
.16981 .47699 L
.16874 .48119 L
.16795 .48502 L
.16729 .48928 L
.16702 .49163 L
.16692 .49268 L
.16683 .49384 L
.16676 .49488 L
.1667 .496 L
.16667 .49662 L
.16665 .49718 L
.16662 .49829 L
.1666 .4993 L
.1666 .50039 L
.16661 .50143 L
.16663 .50239 L
.16665 .50298 L
.16668 .50361 L
.16674 .50475 L
.16688 .50688 L
.16713 .50937 L
.16742 .51166 L
.16807 .51563 L
.16899 .51987 L
.17118 .52748 L
.1747 .53646 L
.17877 .54455 L
.18894 .55991 L
.20333 .57598 L
.22013 .59054 L
.24108 .60501 L
.28604 .62794 L
.3123 .63794 L
.34262 .64719 L
.37348 .65453 L
.38972 .65764 L
.40737 .66048 L
Mistroke
.42444 .66273 L
.44011 .66437 L
.44772 .66502 L
.456 .66563 L
.46327 .66608 L
.47109 .66647 L
.47543 .66664 L
.47941 .66678 L
.48345 .66689 L
.48719 .66697 L
.48926 .66701 L
.49155 .66704 L
.49385 .66706 L
.49511 .66707 L
.49629 .66708 L
.49846 .66709 L
.50044 .66709 L
.50263 .66708 L
.50372 .66708 L
.50492 .66707 L
.50734 .66705 L
.50962 .66702 L
.51388 .66694 L
.51847 .66683 L
.52345 .66666 L
.5324 .66628 L
.54064 .66582 L
.55692 .6646 L
.57204 .66308 L
.60583 .65834 L
.63786 .65197 L
.66587 .64473 L
.69525 .63517 L
.7206 .62498 L
.76473 .60122 L
.77577 .59356 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00e@3oool00`3ooml0W4P0000000020000000307Boo`3oool0oooo0=<0
oooo003G0?ooo`030?noM00007@0_ooo0=H0oooo003G0?ooo`030?oo_`1d0000M;oo0=H0oooo003G
0?ooo`030?oog`2LB000B9cO0=H0oooo003H0?ooo`030=nLB000000000000080000000<0M;oo0?oo
o`3oool00P3oool01@3o_g@000000000000004P0W=oo0080oooo00<0oono07@0000000000P000000
3@18W=l0oooo0?ooo`3oool0_g@00000000004P0W=oo0?ooo`3o_g@000000000B02Lgol00P3oool0
102oM00000000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3oool00`3OW4P004RL
0=ooo`020?ooo`030;md0000000007Bo09d0oooo003H0?ooo`070?noM00004P0W=oo0?ooo`3ooml0
W4P0001d_`020?ooo`0?0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`3oool0oooO09a8001d_ol0
oooo0?oOW0180000M;oo0080oooo01L0gia80018W03Ooml0W7BL0=ooo`3o_g@0001809cOg`2LM9`0
gooo0?ooo`3OW4P004RL0=oog`2LM9`0gooo0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`020?oo
o`030?noM00007@0_ooo0080oooo00@0gia80018W03O_g@0M;ooW03oool00=P0oooo00<0oono07@0
001d_ol00P3oool01`3ooml0W4P004RLg`3oool0gia80018W03Oool00P3oool02P3ogi`0B00004RL
g`3oool0_gAd0;ooo`3oool0oono07@0001d_ol20?ooo`030?noM00004P0W=oo0080oooo00<0oono
07@0001d_ol00`3oool00`3o_g@0001809cOo`020?ooo`030=nLB000B9`0gooo0080oooo00L0omnL
04P00018W=l0oooo0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol0W@3oool00=P0oooo00<0
oooO09a80018W=l00`3oool01P3ogi`0B01809cOo`3o_g@0001d0;ooo`<0oooo00<0gia80000M02o
ool00`3oool00`3OW4P0001d0;ooo`020?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0001d_`3o
ool00P3oool00`3ookl0M00004RLg`020?ooo`030?noM00007@0_ooo00<0oooo00/0gia80000M02o
ool0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`2M0?ooo`00f@3oool00`3OW4P0001d0;oo
o`020?ooo`060?oog`2LB000B9cO0?oo_`1d0000M;oo0`3oool00`3ogi`0B01809cOo`020?ooo`03
0?noM00004P0W=oo00@0oooo00<0_g@00018W03Oool00P3oool00`3OW4P004RL0=ooo`030?ooo`07
0;md0000B9`0gooo0?ooo`3ookl0M00007Boo`030?ooo`070?oOW01804P0W=oo0?ooo`3OW4P004Q8
04Qd_`020?ooo`030;md0000M;l0oooo09`0oooo003I0?ooo`040?noM000000004RL0=ooo`80oooo
00H0gia8001d_`3oool0gia80000M02oool20?ooo`0:0?oo_`1d0000M;oo0?ooo`3ogi`0B01809cO
o`3oool0giad09cOo`80oooo00<0gia80000B02Lgol00P3oool00`3ogi`0B01809cOo`030?ooo`03
0=nLB00004P0W=oo0080oooo00<0gia80000M02oool00P3oool0303ookl0M00007Boo`3oool0okmd
0000M02o_g@0B9cO0?ooo`3o_g@0001d0;oooi`0oooo003I0?ooo`080?oo_`1d0000M9bL04P0M02o
ool0oooO09a80000M;l20?ooo`0>0;md000007@0_ooo0?ooo`3o_g@0001d0;ooo`3oool0omnL04P0
B02Lgol0oooO09a8001d_ol20?ooo`030?oOW0180000B9cO0080oooo00<0oooO09a80018W=l00`3o
ool00`3ogi`0B00004RLg`030?ooo`0@0;md000007@0_ooo0?ooo`3o_g@0001d0;ooo`3oool0oooO
09a8001d_ol0gia8001d_`3ogi`0B01809cOoi`0oooo003H0?ooo`090?oOW0180000000004RLg`3o
okl0M0000000000007@0_ooo00<0oooo00D0okmd000000000000000007Boo`030?ooo`070;md0000
0000000000000018W=l0oooo0?noM002000000040018W03Oool0oono07@000<0000000H0M;oo0?no
M000000000000018W03Oool30?ooo`0?0?noM000000000000000001d_ol0oooo0?oo_`1d00000000
04RLg`3oool0okmd0000000004P0W=oo09`0oooo003i0?ooo`030?noM00007@0_ooo0;@0oooo003j
0?ooo`030=nLM02Lgol0oooo0;<0oooo003d0?ooo`040?oo_`1d000004RL0=ooo`L0oooo00@0oono
07@00000B9`0gooo[@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool009/0oooo00@0_g@000000000
0000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`030?oo_`1d000000000080000000<0B9cO0?oo
o`3oool0@P3oool0102oM00000000000001d_om<0?ooo`040;md00000000000007Boo`<0oooo00<0
omnL04P0001d_ol00P3oool00`3ookl0M000000000020000000304RLg`3oool0oooo0480oooo00@0
_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB0000000000000<00000
00<0B9cO0?ooo`3oool0103oool009X0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo00@0oooo00<0okmd0000001d_ol0A03oool01P3OW4P007Bo0?ooo`3ogi`0B000
04RLgdX0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00@0oooo
00<0okmd0000001d_ol0A03oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0
001d_ol00`3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00@0oooo002I0?ooo`03
0?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3ooml0W4P0001d_`130?ooo`03
0?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0B03oool00`3ooml0W4P004RLg`020?ooo`03
0?noM00004P0W=oo00P0oooo00<0oooO09a80000M;l0@`3oool00`3ooml0W4P004RLg`020?ooo`03
0?noM00004P0W=oo00P0oooo00<0okmd001d_`3oool0203oool009T0oooo00<0oooO09a80018W=l0
0P3oool0103ooml0W4P00018W03Oool80?ooo`030=nLB000B9`0gooo0480oooo00<0oooO09a80018
W=l00P3oool0103ooml0W4P00018W03Ooom70?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a8
0000B9`0gooo203oool00`3OW4P004RL0=ooo`120?ooo`030?oog`2LB000B9cO0080oooo00@0oooO
09a80000B9`0gooo203oool00`3o_g@007Bo0?ooo`070?ooo`00S03oool00`3ooml0W4P000000007
0000000604RLg`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`080?ooo`030?oO
W01804P0W=oo0480oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`170?ooo`030?oo
g`2LB00007Bo00<0oooo00<0gia80000B02Lgol0203oool00`3ogi`0B01809cOo`120?ooo`030?oo
g`2LB00007Bo00<0oooo00<0gia80000B02Lgol02@3oool00`3o_g@0B9cO0?ooo`060?ooo`00VP3o
ool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo00<0oono07@0001d_ol0@`3o
ool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO04P0oooo00<0_g@00018W03Oool00P3o
ool00`3ogi`0B00004RLg`080?ooo`030?oo_`1d0000M;oo04<0oooo00<0_g@00018W03Oool00P3o
ool00`3ogi`0B00004RLg`0:0?ooo`030=nLB018W=l0oooo00D0oooo002J0?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02@3oool00`2oM00007Bo0?ooo`120?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l0B03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00T0oooo00<0_g@0001d_`3oool0@P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00/0oooo00<0_g@007Boo`3oool0103oool009X0oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`080?ooo`030=nLB000B9`0gooo0480oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`170?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool0203oool00`3O
W4P004RL0=ooo`120?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool02P3oool00`3o
gi`0B01d0;ooo`040?ooo`00VP3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo
00P0oooo00<0omnL04P0B02Lgol0@P3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0
gooo04L0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`030?oOW018
04P0W=oo0480oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`0:0?ooo`030?oo
g`2LB000M;oo00@0oooo002K0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0203o
ool00`3ookl0M00007Boo`130?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0B03o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00P0oooo00<0oono07@0001d_ol0@`3o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00X0oooo00<0oono07@00018W=l0103o
ool009`0oooo00H0gia80000M02oool0oooO09a80018W=l:0?ooo`030;md0000M;l0oooo04<0oooo
00H0gia80000M02oool0oooO09a80018W=m:0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO2P3o
ool00`2oM00007Bo0?ooo`130?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool01`3o_g@0
04RL0=ooo`3oool0gia80000B02Lgol0103oool009d0oooo00@0gia800000000000007Bo2@3oool0
1@3ookl0M0000000000007@0_ooo04@0oooo00@0gia800000000000007BoC03oool0103OW4P00000
00000000M;l90?ooo`050?oo_`1d000000000000M02oool0A03oool0103OW4P0000000000000M;l9
0?ooo`050?oOW018000000000000001d_ol01@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0
oooo/@3oool00?l0oooo/@3oool004L0ooooo`00001K000000l0oooo00170?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`020?ooo`8000002P3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0>0?ooo`<000003P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`00
0000oooo0?ooo`0S0?ooo`800000;P3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo
05<0oooo00<000000?ooo`3oool04@3oool200000400oooo00<000000?ooo`3oool03@3oool004L0
oooo00<000000?ooo`3oool08@3oool2000008L0oooo00<000000?ooo`3oool0J@3oool2000003h0
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool07`3oool2000008T0oooo
00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo00d0oooo
00170?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0R@3oool00`000000oooo0?oo
o`1/0?ooo`800000>`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0L
0?ooo`800000S03oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2<0?oo
o`030000003oool0oooo06l0oooo0P00000h0?ooo`030000003oool0oooo00d0oooo00170?ooo`03
0000003oool0oooo01T0oooo0P00002?0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3o
ool0=@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo08l0oooo00<000000?ooo`3oool0LP3oool2000003D0oooo00<000000?ooo`3oool0
3@3oool004L0oooo00<000000?ooo`3oool05P3oool200000980oooo00<000000?ooo`3oool0M03o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`1e0?ooo`800000<P3oool0
0`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0C0?ooo`800000U@3oool00`00
0000oooo0?ooo`1g0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool03@3oool004L0
oooo00<000000?ooo`3oool04@3oool2000009L0oooo00<000000?ooo`3oool0N03oool2000002l0
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`2G0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0>0?ooo`800000VP3oool00`000000oooo0?oo
o`1k0?ooo`800000;03oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`0Y
0?ooo`030000003oool0oooo00d0oooo00170?ooo`800000303oool2000009d0oooo00<000000?oo
o`3oool0OP3oool2000002P0oooo0P00000?0?ooo`00A`3oool00`000000oooo0?ooo`0:0?ooo`03
0000003oool0oooo09d0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0V0?ooo`03
0000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo00P0oooo0P00002P0?ooo`030000
003oool0oooo0840oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0PP3o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`230?ooo`800000903oool0
0`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`040?ooo`800000Y03oool00`00
0000oooo0?ooo`250?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool03@3oool004L0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo08H0
oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool0Q`3oool00`000000oooo
0?ooo`0O0?ooo`030000003oool0oooo00d0oooo00170?ooo`050000003oool0oooo0?ooo`000000
Z03oool00`000000oooo0?ooo`0V0?oooa/00000A`3oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo00d0oooo00170?ooo`040000003oool0000000000:T0oooo00<000000?ooo`3oool0
7@3oool9000001/0oooo1P0000120?ooo`8000007P3oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool200000:/0oooo00<000000?ooo`3oool05P3oool7000002X0oooo100000100?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0ZP3oool0
0`000000oooo0?ooo`0B0?ooo`@00000=@3oool3000003h0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo
0100oooo0P00000l0?ooo`@00000>`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
00d0oooo00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool03@3oool300000480
oooo1000000h0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03@3oool004L0oooo
00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0:0?ooo`<00000B@3oool2000003L0oooo
0P00000H0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo0:X0oooo00<0
00000?ooo`3oool01P3oool4000004h0oooo0P00000g0?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`03
0?ooo`<00000E03oool2000003H0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0=
0?ooo`00A`3oool00`000000oooo0?ooo`2Z0?ooo`040000003oool0oooo0?ooo`800000F@3oool3
000003@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool20000
0:/0oooo00@000000?ooo`0000000000GP3oool2000003<0oooo00<000000?ooo`3oool04@3oool2
000000l0oooo00170?ooo`030000003oool0oooo0:X0oooo0P00001R0?ooo`<00000<@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo0:P0
oooo0`00001V0?ooo`800000<03oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00d0
oooo00170?ooo`030000003oool0oooo0:H0oooo0P0000020?ooo`030000003oool0oooo06H0oooo
0P00000_0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool004L0oooo00<0
00000?ooo`3oool0Y03oool2000000@0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool004L0oooo00<00000
0?ooo`3oool0RP3ooolF000000<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1Y
0?ooo`800000;P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00170?oo
o`030000003oool0oooo0880oooo2000000F0?ooo`h00000JP3oool2000002d0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`1g0?ooo`/0
00007@3oool2000000T0oooo00@000000?ooo`3oool0oooo1@00001W0?ooo`030000003oool0oooo
02/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000
oooo0?ooo`1c0?ooo`@000009`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0
oooo1000001T0?ooo`800000:`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00d0
oooo00170?ooo`030000003oool0oooo06l0oooo1000000Y0?ooo`800000303oool00`000000oooo
0?ooo`0:0?ooo`D00000H@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0J`3oool4000002/0oooo0P00
000>0?ooo`030000003oool0oooo00l0oooo1P00001L0?ooo`800000:P3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo06L0oooo1000000^
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool05@3oool4000005X0oooo0P00000Y
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?oo
o`3oool0H`3oool400000340oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0I0?oo
o`@00000F03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool03@3oool004L0oooo00<000000?ooo`3oool0G`3oool4000003<0oooo0P00000B0?ooo`03
0000003oool0oooo01d0oooo0`00001F0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`1K0?ooo`@00000
=P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0200oooo0P00001E0?ooo`030000
003oool0oooo02L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool00`000000oooo0?ooo`1H0?ooo`<00000>@3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo0280oooo0`00001C0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`1F0?ooo`800000>P3oool2
000001H0oooo00<000000?ooo`3oool09@3oool300000540oooo0P00000X0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03@3oool004L0oooo0P00001D0?ooo`<00000>`3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo02P0oooo0`00001@0?ooo`030000003oool0oooo
02H0oooo00@000000?ooo`3oool0oooo0P00000?0?ooo`00A`3oool00`000000oooo0?ooo`1@0?oo
o`<00000?@3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo02/0oooo0P00001?0?oo
o`030000003oool0oooo02H0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`00A`3oool00`00
0000oooo0?ooo`1=0?ooo`<00000?P3oool2000001X0oooo00<000000?ooo`3oool0;@3oool30000
04d0oooo00<000000?ooo`3oool09@3oool01@000000oooo0?ooo`3oool0000000l0oooo00170?oo
o`030000003oool0oooo04/0oooo0P0000100?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0<03oool2000004`0oooo00<000000?ooo`3oool09@3oool010000000oooo0?ooo`00000?
0?ooo`00A`3oool00`000000oooo0?ooo`190?ooo`800000@@3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo0380oooo0`00001:0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`0000003`3oool004L0oooo00<000000?ooo`3oool0A`3oool200000440oooo0P00000N0?ooo`03
0000003oool0oooo03D0oooo0P0000190?ooo`030000003oool0oooo02D0oooo0P00000?0?ooo`00
A`3oool00`000000oooo0?ooo`150?ooo`800000@P3oool00`000000oooo0?ooo`0N0?ooo`030000
003oool0oooo03L0oooo0P0000180?ooo`8000009@3oool2000000l0oooo00170?ooo`030000003o
ool0oooo04<0oooo0P0000130?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0>@3o
ool2000004P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool0
0`000000oooo0?ooo`110?ooo`800000A03oool00`000000oooo0?ooo`0P0?ooo`030000003oool0
oooo03/0oooo0P0000170?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool03@3oool0
04L0oooo00<000000?ooo`3oool0?`3oool2000004D0oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`0m0?ooo`800000AP3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo
00d0oooo00170?ooo`030000003oool0oooo03d0oooo0P0000160?ooo`030000003oool0oooo0280
oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0?03oool00`000000oooo
0?ooo`150?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0@03oool2000004D0oooo
00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?oo
o`0j0?ooo`800000A`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0480oooo0P00
00140?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03@3oool004L0oooo00<00000
0?ooo`3oool0>@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo02D0oooo00<00000
0?ooo`3oool0A03oool200000480oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0=
0?ooo`00A`3oool00`000000oooo0?ooo`0g0?ooo`800000B@3oool00`000000oooo0?ooo`0U0?oo
o`030000003oool0oooo04H0oooo0P0000110?ooo`030000003oool0oooo01d0oooo00<000000?oo
o`3oool03@3oool004L0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`180?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0o0?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0
=03oool2000004X0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`190?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool2000003@0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0X0?ooo`030000003o
ool0oooo04X0oooo0P0000100?ooo`030000003oool0oooo01T0oooo0P00000?0?ooo`00A`3oool0
0`000000oooo0?ooo`0a0?ooo`800000B`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0
oooo04`0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo00d0oooo00170?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0BP3oool00`00
0000oooo0?ooo`0Z0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0?P3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo02h0
oooo0P00001<0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0CP3oool2000003l0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo
0?ooo`0]0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`1@0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0[0?ooo`800000C@3oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0?03oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo02X0oooo00<00000
0?ooo`3oool0C@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0580oooo0P00000m
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?oo
o`3oool0:03oool2000004l0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1D0?oo
o`030000003oool0oooo03/0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0=0?oo
o`00A`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3o
ool03`3ooolE000000/0oooo00<000000?ooo`3oool0E@3oool00`000000oooo0?ooo`0k0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0
9P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00<0oooo3000000E0?ooo``00000
F03oool2000003`0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`0000001000
000/0?ooo`H00000E@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0180oooo00<0
00000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool08`3oool2000004h0oooo1P00000a
0?ooo`030000003oool0oooo00<0oooo1P00001@0?ooo`030000003oool0oooo03T0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0oooo04T0oooo1@0000030?ooo`030000003oool0oooo0340oooo00<000000?oo
o`3oool02@3oool5000004`0oooo0P00000i0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool03@3oool002/0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?oo
o`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool0203oool00`000000oooo0?ooo`0Q0?oo
o`030000003oool0oooo04D0oooo1@0000070?ooo`030000003oool0oooo0380oooo00<000000?oo
o`3oool03P3oool4000004X0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0@0?oo
o`030000003oool0oooo00d0oooo000Z0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool0
0`3ogi`0B00007Boo`040?ooo`030?noM0000000M;oo00/0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`130?ooo`<000002`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo0180oooo0`0000180?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0=0?ooo`00:@3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo
00P0oooo00<0oooO09a80000M;l02`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
0440oooo0`00000=0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool05@3oool40000
04D0oooo0P00000g0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool002T0
oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030=nLB000B9`0gooo
00X0oooo00<000000?ooo`3oool07@3oool200000440oooo0`00000@0?ooo`030000003oool0oooo
03@0oooo00<000000?ooo`3oool06@3oool3000004@0oooo00<000000?ooo`3oool0=@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo000L0?ooo`030?oog`2LB000000000L0
000000H0B9cO0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00P0oooo00<0omnL
04P0B02Lgol02P3oool3000001`0oooo00<000000?ooo`3oool0?P3oool300000180oooo00<00000
0?ooo`3oool0=@3oool00`000000oooo0?ooo`0L0?ooo`800000@`3oool00`000000oooo0?ooo`0d
0?ooo`030000003oool0oooo00`0oooo0`00000?0?ooo`00:P3oool00`2oM00004RL0=ooo`020?oo
o`030?oOW0180000B9cO00P0oooo00<0oono07@0001d_ol02P3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo03`0oooo0`00000D0?ooo`030000003oool0oooo03H0oooo00<000000?oo
o`3oool07P3oool200000480oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00d0oooo000Z0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000
M;l02@3oool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool0>`3oool2000001L0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0P0?ooo`<0
0000@03oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool03@3oool002X0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`080?ooo`030=nL
B000B9`0gooo00T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0j0?ooo`800000
603oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02<0oooo0P00000o0?ooo`030000
003oool0oooo03<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=0?ooo`00:P3o
ool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00P0oooo00<0omnL04P0B02Lgol0
2@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo03T0oooo0P00000I0?ooo`030000
003oool0oooo03P0oooo00<000000?ooo`3oool09@3oool2000003h0oooo00<000000?ooo`3oool0
<`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00d0oooo000[0?ooo`030?oOW018
0000M;oo0080oooo00<0gia8001d_`3oool0203oool00`3ookl0M00007Boo`090?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`3oool0>03oool2000001/0oooo00<000000?ooo`3oool0>03o
ool00`000000oooo0?ooo`0W0?ooo`800000?03oool00`000000oooo0?ooo`0c0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool03@3oool002`0oooo00H0gia80000M02oool0oooO09a8
0018W=l:0?ooo`030;md0000M;l0oooo00P0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`0g0?ooo`800000703oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo02T0oooo
0P00000k0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`0=0?ooo`00;@3oool0103OW4P0000000000000M;l90?ooo`050?oo_`1d000000000000M02o
ool0203oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo03H0oooo0P00000M0?ooo`03
0000003oool0oooo03X0oooo00<000000?ooo`3oool0:`3oool2000003X0oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0M0?ooo`030000
003oool0oooo03X0oooo00<000000?ooo`3oool0;@3oool2000003T0oooo00<000000?ooo`3oool0
<P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool0=03oool2000001l0oooo00<000000?ooo`3oool0>`3o
ool00`000000oooo0?ooo`0_0?ooo`800000>03oool00`000000oooo0?ooo`0a0?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`0c0?ooo`800000803oool00`000000oooo0?ooo`0Q0?ooo`l00000303oool0
0`000000oooo0?ooo`0a0?ooo`800000=`3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0b0?ooo`8000008P3oool00`000000oooo0?ooo`0B0?ooo`l000003`3oool=0000
03D0oooo0P00000f0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0380
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0?0?ooo`@00000:P3oool600000380
oooo0P00000e0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0340oooo
0P00000S0?ooo`030000003oool0oooo00/0oooo1@00000^0?ooo`030000003oool0oooo00<0oooo
1P00000^0?ooo`800000<`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0`0?ooo`8000009@3oool00`000000oooo0?ooo`050?ooo`H00000<`3oool00`000000oooo0?oo
o`090?ooo`H00000:P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`02
0?ooo`@00000>@3oool00`000000oooo0?ooo`0?0?ooo`<00000:03oool200000380oooo00<00000
0?ooo`3oool0;`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00d0oooo00170?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0;P3oool2000002L0oooo00<000000?oo
o`0000000P00000m0?ooo`030000003oool0oooo0180oooo0`00000W0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00d0oooo00170?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0;@3oool2000002P0
oooo0`0000100?ooo`030000003oool0oooo01D0oooo0`00000U0?ooo`030000003oool0oooo0300
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00d0
oooo00170?ooo`8000003@3oool00`000000oooo0?ooo`0/0?ooo`8000009`3oool400000480oooo
00<000000?ooo`3oool0603oool3000002<0oooo0P00000a0?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool00`3oool2000000l0oooo00170?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0;03oool00`000000oooo0?ooo`0U0?ooo`8000000P3oool00`000000oooo0?oo
o`110?ooo`030000003oool0oooo01/0oooo0P00000S0?ooo`030000003oool0oooo02h0oooo00<0
00000?ooo`3oool0;P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo0017
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0S
0?ooo`<00000103oool00`000000oooo0?ooo`0R0?oooah0000000<0oooo0000003oool07P3oool3
00000240oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0Z0?ooo`8000008`3oool3000000L0oooo00<000000?ooo`3oool06@3oool90000
01h0oooo2000000L0?ooo`<000007`3oool2000002l0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0Q0?ooo`8000002@3oool00`000000
oooo0?ooo`0D0?ooo`H00000:03oool00`000000oooo0?ooo`040?ooo`H000006@3oool3000001h0
oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo01l0oooo0`00000;0?ooo`030000003oool0oooo00l0oooo
1@00000^0?ooo`030000003oool0oooo00X0oooo1@00000G0?ooo`8000007@3oool00`000000oooo
0?ooo`0]0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo02P0oooo0P00
000P0?ooo`8000003@3oool00`000000oooo0?ooo`0<0?ooo`@00000<`3oool00`000000oooo0?oo
o`0?0?ooo`@000005@3oool2000001`0oooo0P00000]0?ooo`030000003oool0oooo02d0oooo00D0
00000?ooo`3oool0oooo0000000?0?ooo`00A`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo02P0oooo00<000000?ooo`3oool07P3oool2000000l0oooo00<000000?ooo`3oool0203o
ool4000001d0oooo7P00000B0?ooo`@000004`3oool2000001`0oooo00<000000?ooo`3oool0:`3o
ool00`000000oooo0?ooo`0/0?ooo`050000003oool0oooo0?ooo`0000003`3oool004L0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01d0oooo0P00
000A0?ooo`030000003oool0oooo00D0oooo0`00000H0?ooo`T000006P3oool010000000oooo0?oo
o`3oool9000000d0oooo1000000A0?ooo`8000006`3oool00`000000oooo0?ooo`0[0?ooo`030000
003oool0oooo02/0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`00A`3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool07@3oool00`000000oooo
0?ooo`0@0?ooo`030000003oool0oooo0080oooo1000000D0?ooo`L000003`3ooolH000000T0oooo
1P00000;0?ooo`<00000403oool2000001X0oooo00<000000?ooo`3oool0:P3oool00`000000oooo
0?ooo`0/0?ooo`040000003oool0oooo000000l0oooo00170?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0L0?ooo`8000004`3oool00`000000oooo
0000000200000140oooo1`00000<0?ooo`X000001@3ooolQ000000030?ooo`000000000000@00000
203oool4000000h0oooo0P00000I0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0
:`3oool010000000oooo0?ooo`00000?0?ooo`00A`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool06`3oool2000001@0oooo0`00000?0?ooo`D00000
303ooolH000001X0o`00100000000`3oool0000000000009000000P0oooo0`00000=0?ooo`800000
603oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo02X0oooo00@000000?ooo`3oool0
00003`3oool004L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0W0?ooo`030000
003oool0oooo01T0oooo0P00000D0?ooo`<000003P3oool3000000T0oooo4P00000?0?l000`0oooo
00<000000?ooo`3oool02@3oool;0?l000/000001P3oool3000000`0oooo0P00000G0?ooo`030000
003oool0oooo02X0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0000000?0?ooo`00A`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0B0?ooo`80000000<0oooo0000003oool02`3oool3000000L0oooo1@00
00050?ooo`<0000000@0oooo00000000000000002P3o000G0?ooo`030000003oool0oooo01@0oooo
303o0003000000D0oooo0P00000<0?ooo`8000005P3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo02X0oooo00<000000?ooo`0000003`3oool004L0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo01P0oooo0P00000B0?ooo`<000000`3o
ool00`000000oooo0?ooo`070?ooo`<000001P3oool4000000H0oooo1@0000060?l0008000007`3o
ool00`000000oooo0?ooo`0L0?ooo`X0o`00300000060?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0000000l0oooo0017
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo0100oooo0P0000050?ooo`030000003oool0oooo00D0oooo0`000006
0?ooo`<000001P3oool3000000L0o`0000<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0R
0?ooo`8000001P3o00030?ooo`<000000P3oool3000000@0oooo0P00000E0?ooo`800000:@3oool0
0`000000oooo0?ooo`0Z0?ooo`8000003`3oool004L0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0U0?ooo`030000003oool0oooo01L0oooo0P00000A0?ooo`8000001`3oool00`00
0000oooo0?ooo`030?ooo`8000001P3oool3000000@0oooo0`0000060?l000030000003oool0oooo
02T0oooo00<000000?ooo`3oool0803oool3000000<0oooo0`0000050?l00080oooo0`0000020?oo
o`<000000`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool0:@3oool2000000l0oooo00170?ooo`8000000P3oool00`000000oooo0?ooo`0U0?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`3oool03`3oool2000000T0oooo00<000000?ooo`3o
ool00`0000040?ooo`@000000`3oool3000000D0o`000P00000_0?ooo`030000003oool0oooo01h0
oooo0P00000:0?ooo`<000001@3o0003000000<0oooo0`0000000`3oool000000000000E0?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool0:@3oool2000000l0oooo00170?ooo`050000
003oool0oooo0?ooo`0000009P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00h0
oooo0P00000:0?ooo`@00000103oool3000000<0oooo100000050?l000030000003oool0oooo0340
oooo00<000000?ooo`3oool0703oool200000100oooo0`0000040?l000800000103oool200000003
0?ooo`000000000001@0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0X0?ooo`80
00003`3oool004L0oooo00@000000?ooo`3oool000009`3oool00`000000oooo0?ooo`0E0?ooo`80
0000403oool00`000000oooo0?ooo`090?ooo`800000103oool3000000<0oooo0`0000050?l00003
0000003oool0oooo03D0oooo00<000000?ooo`3oool06P3oool2000001D0oooo0`0000040?l00003
0000003oool0oooo0080oooo0P0000000`3oool000000?ooo`0C0?ooo`030000003oool0oooo02D0
oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo
0000000W0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool2000000T0oooo
0`000000103oool000000?ooo`3oool200000080oooo100000050?l000030000003oool0oooo03P0
oooo00<000000?ooo`3oool0603oool2000001/0oooo0P0000040?l000070000003oool0oooo0?oo
o`000000oooo0000000D0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0000000V0?ooo`030000003oool0oooo
01@0oooo0P00000?0?ooo`8000002@3oool2000000@0oooo0`0000000`3oool00000000000030000
00@0o`00?P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01h0oooo0`0000040?l0
00030?ooo`0000000000008000004`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo
02P0oooo00<000000?ooo`3oool03@3oool004L0oooo0P00000V0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`800000103oool300000003
0?ooo`0000000000008000000`3o00000`000000oooo0?ooo`0o0?ooo`030000003oool0oooo01D0
oooo0P00000T0?ooo`<00000103o00000`000000oooo0000000C0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool2000002D0oooo
00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0=0?ooo`800000203oool2000000@0oooo
0P0000000`3oool0000000000003000000<0o`00A03oool00`000000oooo0?ooo`0D0?ooo`030000
003oool0oooo02H0oooo0`0000040?l000030?ooo`000000oooo0140oooo00<000000?ooo`3oool0
903oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003o
ool0oooo02@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00H0oooo0P0000040?ooo`80000000<0oooo0000000000000`0000030?l004H0oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0Y0?ooo`@00000103o00000`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?oo
o`0=0?ooo`00A`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`050?ooo`800000103oool2000000030?ooo`000000
000000<000000P3o00190?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0;03oool4
000000@0o`00403oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo02H0oooo00<00000
0?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo00`0oooo0P0000060?ooo`800000103oool2000000030?ooo`0000000000
00<000000P3o001;0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0;`3oool40000
00D0o`003P3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool03@3oool004L0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`80
000000<0oooo0000000000000`0000020?l000030000003oool0oooo04X0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`0a0?ooo`D00000103o000=0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`040?ooo`8000000`3oool2000000030?ooo`0000000000008000000`3o00020?ooo`030000
003oool0oooo04X0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0d0?ooo`D00000
103o000;0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0903oool00`000000oooo
0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`8000000`3oool2000000050?ooo`00
000000000?ooo`0000000`3o00040?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`0f0?ooo`H00000103o000:0?ooo`030000003oool0oooo02<0oooo
00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?oo
o`0O0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool02P3oool2000000D0oooo00<0
00000?ooo`3oool00P3oool01P000000oooo0?ooo`000000oooo000000<0o`001@3oool00`000000
oooo0?ooo`1;0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0>03oool7000000<0
o`0000<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool07`3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3oool2000000@0oooo
00<000000?ooo`0000000P0000030?l000L0oooo00<000000?ooo`3oool0B`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo03T0oooo0P0000000`3oool000000?ooo`03000000<0o`00
203oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0
3@3oool004L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0B0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`040000003o
ool00000000000<0o`002@3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool0>P3oool010000000oooo0000003oool6000000<0o`001`3oool00`000000
oooo0?ooo`0R0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool03@3oool004L0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`@000000`3o000;0?ooo`030000
003oool0oooo04/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0l0?ooo`050000
003oool000000000003oool01P0000030?l000H0oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0R0?ooo`030000003oool0oooo00d0oooo00170?ooo`8000007P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo00T0oooo0P0000050?ooo`030000003oool0oooo0080oooo
00@000000?ooo`000000oooo0P3o000<0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0m0?ooo`060000003oool0oooo0?ooo`000000oooo0P000000
103oool000000000003oool30?l000H0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?oo
o`0Q0?ooo`8000003`3oool004L0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`@000000P3o000=0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0n0?ooo`060000003oool0oooo0?ooo`000000oooo0P0000000`3oool00000
0000000200000080o`001@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo0280oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`020?ooo`80000000<0oooo0?l0003o00003`3oool00`000000oooo0?ooo`1<0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`020?ooo`050000003o
ool000000000003oool0100000020?l000D0oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo01/0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool00P3oool2000000030?ooo`3o0000o`000100oooo00<000000?ooo`3oool0C03o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool00P3o
ool010000000oooo0000000000020?ooo`@000000P3o00040?ooo`030000003oool0oooo0200oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?oo
o`0K0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo0080oooo0`0000020?l00140oooo00<000000?ooo`3oool0C03o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool00`3o
ool00`000000oooo00000002000000030?ooo`0000000000008000000P3o00030?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0`0000000`3o0000oooo0?ooo`0@
0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`13
0?ooo`030000003oool0oooo00<0oooo0`0000001@3oool000000?ooo`000000oooo008000000P3o
00030?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`0=0?ooo`00A`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`@0000000<0
o`000?ooo`3oool04@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0@`3oool00`000000oooo0?ooo`050?ooo`070000003oool0oooo0000003oool0
00000?ooo`0300000080o`0000<0oooo0000003oool08@3oool00`000000oooo0?ooo`0P0?ooo`03
0000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
4@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo00H000000?ooo`3oool0
oooo0000003oool2000000030?l0003oool0oooo0180oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool01@3oool01@00
0000oooo0?ooo`000000oooo00@0000000D0oooo0?l0003o0000oooo0000000Q0?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool03@3oool000P0oooo00<0oono07Ad_`3oool0?03oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`030?ooo`070000003oool0oooo0?ooo`000000oooo000000020?l001D0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04D0oooo
00<000000?ooo`3oool01P3oool010000000oooo0?ooo`0000020?ooo`<0000000@0oooo0?l0003o
000000008P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00d0oooo00080?ooo`04
0?oOW0180000001809cOoc/0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool00`3oool01P000000oooo0?ooo`000000
oooo00000080o`005P3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0A`3oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo00000080oooo
00@000000?ooo`00000000000P3o00000`000000oooo0?ooo`0O0?ooo`030000003oool0oooo01l0
oooo00<000000?ooo`3oool03@3oool000X0oooo00<0_g@00000000000000`0000000`0007@0_ooo
0?ooo`0d0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0403oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00<0oooo00H000000?ooo`3oool000000?ooo`0000020?l0
01H0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
04L0oooo00<000000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool000000080oooo00@00000
0?ooo`00000000000P3o000Q0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool03@3o
ool000h0oooo00<0omnL04P0001d_ol0=P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`060000003oool0
oooo0000003oool000000P3o000G0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo00L0oooo00D000000?ooo`3oool0
oooo000000020?ooo`040000003oool0000000000080o`00803oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo00d0oooo000=0?ooo`030=nLB00007@0_ooo03L0oooo00<000000?ooo`3o
ool05`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool00P3oool01@000000oooo0?ooo`000000oooo0080000000<0o`000?ooo`3oool05P3oool00`00
0000oooo0?ooo`1>0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0B@3oool00`00
0000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo00000000
00020?l001l0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0=0?ooo`00203oool0
1P3ogi`0B01809cOo`3o_g@0000004RLgcT0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool01@000000oooo
0?ooo`000000oooo008000000P3o000H0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo00P0oooo00X000000?ooo`3o
ool0oooo0000003oool000000?ooo`000000oooo0P3o000P0?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool03@3oool000P0oooo00@0omnL04P00000000007Bo>`3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01`3oool01@000000oooo
0?ooo`3oool0000000<0oooo0`0000020?l001T0oooo00<000000?ooo`3oool0CP3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`020?ooo`040000003oool000000?ooo`8000000P3o000O0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool01@000000
oooo0?ooo`000000oooo0080000000<0o`000?ooo`3oool0603oool00`000000oooo0?ooo`1>0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`090?oo
o`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0000000000020?l001h0oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01`3oool01@000000oooo0?ooo`3o
ool0000000<0oooo0`0000000`3o0000oooo0?ooo`0I0?ooo`030000003oool0oooo04h0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00P0oooo0P00
00030?ooo`050000003oool0oooo0000003oool00P0000000`3o0000oooo0?ooo`0L0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool03@3oool000/0oooo00H0gia80000000000000000
0000B02Lgolf0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`0000000`3o0000
oooo0?ooo`0I0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`1<0?ooo`030000003oool0oooo00T0oooo0P0000030?ooo`030000003oool0oooo00<0
00000P3o000M0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03@3oool000T0oooo
00T0oono07@00018W=l0omnL09boo`3oool0gia80018W03Oool0=@3oool2000001L0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`0000000`3o
ool200000080o`006`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0C03oool00`000000oooo0?ooo`090?ooo`8000000`3oool01@000000oooo0?oo
o`000000oooo008000000P3o000L0?ooo`030000003oool0oooo01d0oooo0P00000?0?ooo`002@3o
ool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d0;ooo`0e0?ooo`030000003oool0oooo
01D0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00`0000000`3o0000oooo0?ooo`0J0?ooo`030000003oool0oooo
04l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo
00T0oooo0P0000030?ooo`030000003oool0oooo00@000000P3o000L0?ooo`030000003oool0oooo
01d0oooo00<000000?ooo`3oool03@3oool000P0oooo00T0omnL09boo`3oool0oooO09adW03Oool0
oooo0?oOW018M;l0=P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool01`3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`000000
0P3o000L0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`1=0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`0000000`3oool00`000000oooo
0?ooo`0300000080o`00703oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo00d0oooo
00080?ooo`090?noW02Lgol0oooo0;md001d_ol0oooo0=nLB00007@0_ooo03H0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`03000000030?l0003oool0oooo01/0oooo00<000000?oo
o`3oool0C`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo04d0oooo00<000000?oo
o`3oool02P3oool2000000<0oooo00<000000?ooo`3oool0100000020?l001/0oooo00<000000?oo
o`3oool07@3oool00`000000oooo0?ooo`0=0?ooo`00203oool0103ookl0M0000000001d_olk0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo0080oooo00D000000?ooo`000000oooo000000020?l001d0oooo00<00000
0?ooo`3oool0C`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04h0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo000000030?ooo`050000003oool0oooo0000003oool00P00
00000`3o0000oooo0?ooo`0I0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool03@3o
ool004L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo00000002000000030?l0003oool0
oooo01`0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo04h0oooo00<000000?ooo`3oool02P3oool00`000000oooo000000040?ooo`030000003oool0
000000<000000P3o000J0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool03@3oool0
04L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0200000080o`007P3oool00`00
0000oooo0?ooo`1?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0CP3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0000000<0oooo00<000000?ooo`3oool00`0000020?l0
01/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0=0?ooo`00303oool01P3o_g@0
0000000000000000001d0;ooocD0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo00000002
000000030?l0003oool0oooo01d0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool02P3oool00`000000oooo00000004
0?ooo`040000003oool000000?ooo`80000000<0o`000?ooo`3oool06@3oool00`000000oooo0?oo
o`0L0?ooo`030000003oool0oooo00d0oooo00040?ooo`060;nLW03Oool0oooo0?ooo`3ogi`0B000
0`0000001000B9`0gooo0?oo_`1dM;lf0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0000000P3oool30000
00030?l0003oool0oooo01d0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo04l0oooo00<000000?ooo`3oool02P3oool00`000000oooo000000040?oo
o`030000003oool0000000<000000P3o000J0?ooo`030000003oool0oooo01`0oooo00<000000?oo
o`3oool03@3oool000<0oooo00<0oooO09a8000000000`0000000`00M;l0oooo0?ooo`020?ooo`03
0?noM00007@0_ooo03H0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000020?ooo`8000000P3o000N0?oo
o`030000003oool0oooo0500oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1?0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`000000103oool01P000000oooo000000000000
oooo00000080o`006P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00d0oooo0003
0?ooo`030?oo_`1d0000B9cO0440oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02
000000030?l0003oool0oooo01d0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool02`3oool00`000000oooo00000003
0?ooo`050000003oool000000000003oool00P0000000`3o0000oooo0?ooo`0H0?ooo`030000003o
ool0oooo01`0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo00000002000000030?l0003oool0oooo01d0oooo00<000000?ooo`3oool0D03oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo02T0oooo0`00000S0?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`000000103oool3000000040?ooo`00000000000?l001X0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`0=0?ooo`002`3oool01P3ogi`0B00000000000
0000001809cOobP0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`03
0000003oool000000080000000<0o`000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`05
0000003oool0oooo0?ooo`000000:`3oool010000000oooo0?ooo`00000S0?ooo`030000003oool0
oooo00/0oooo00@000000?ooo`3oool000000`3oool3000000040?ooo`00000000000?l001X0oooo
00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0=0?ooo`002P3oool00`2oM000000007Bo
o`020?ooo`030?oog`2LM9`0gooo02H0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`03
0000003oool0oooo01@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`0000000P0000000`3o0000oooo0?ooo`0M0?ooo`03
0000003oool0oooo0500oooo00D000000?ooo`3oool0oooo0000000[0?ooo`040000003oool0oooo
000002<0oooo00<000000?ooo`3oool02`3oool010000000oooo0?ooo`0000030?ooo`80000000@0
oooo00000000000000000P3o000I0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0
3@3oool000T0oooo00<0gia804RLg`3oool0103oool00`3OW9`0_ooo0?ooo`0T0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool000000080000000<0o`000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`050000003o
ool0oooo0?ooo`000000:`3oool010000000oooo0?ooo`00000S0?ooo`030000003oool0oooo00/0
oooo00@000000?ooo`3oool000000`3oool01@000000oooo000000000000oooo0080000000<0o`00
0?ooo`3oool05`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00d0oooo00080?oo
o`030?oOW01dW=l0oooo00@0oooo00<0oooO09adW03Oool09@3oool00`3ooml0W4P004RLg`020?oo
o`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000020?ooo`<0000000<0
o`000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`050000003oool0oooo0?ooo`000000
:P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0
2`3oool010000000oooo0?ooo`0000030?ooo`050000003oool000000000003oool00P0000000`3o
0000oooo0?ooo`0G0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool03@3oool000P0
oooo00<0oknL09cOo`3oool00`3oool00`3ooml0W4P007Boo`0V0?ooo`030?oog`2LB00007Bo00<0
oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool000000080oooo0`0000000`3o
0000oooo0?ooo`0M0?ooo`030000003oool0oooo0500oooo00D000000?ooo`3oool0oooo0000000Z
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0;
0?ooo`040000003oool0oooo000000<0oooo00D000000?ooo`00000000000?ooo`02000000030?l0
003oool0oooo01L0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0=0?ooo`00203o
ool00`3ogi`0B7Bo0?ooo`020?ooo`030=nLB0000000B9cO02P0oooo00<0_g@00018W03Oool00P3o
ool00`3ogi`0B00004RLg`070?ooocX0000000<0o`00000000000000h`0000000`3o000000000000
000g000000l0oooo00090?ooo`060?noM0000000000000000000B9`0gooo:@3oool00`3OW4P00018
09cOo`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0503oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000020?ooo`<0
000000<0o`000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`050000003oool0oooo0?oo
o`000000:P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0200oooo00<000000?oo
o`3oool02`3oool010000000oooo0?ooo`0000030?ooo`050000003oool000000000003oool00P00
00000`3o0000oooo0?ooo`0G0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool03@3o
ool003P0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`060?ooo`050000003oool0
oooo0?ooo`0000000P3oool3000000030?l0003oool0oooo01d0oooo00<000000?ooo`3oool0D03o
ool01@000000oooo0?ooo`3oool0000002X0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo00/0oooo00@000000?ooo`3oool000000`3oool01@000000
oooo000000000000oooo0080000000<0o`000?ooo`3oool05`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo00d0oooo000h0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018
W03Oool01P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool01P3oool01@000000oooo0?ooo`3oool000000080oooo0`0000000`3o0000oooo0?ooo`0M
0?ooo`030000003oool0oooo0500oooo00D000000?ooo`3oool0oooo0000000[0?ooo`040000003o
ool0oooo000002<0oooo00<000000?ooo`3oool02`3oool010000000oooo0?ooo`0000030?ooo`06
0000003oool000000000003oool000000P3o000I0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool03@3oool000/0oooo00H0omnL04P00000000000000000B02LgolX0?ooo`030?oOW018
0000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool000000080000000<0o`000?ooo`3oool07@3oool00`000000oooo0?ooo`1@0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0:03oool010000000oooo0?ooo`00000S0?ooo`030000
003oool0oooo00/0oooo00@000000?ooo`3oool000000`3oool2000000050?ooo`00000000000000
003o00006P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00d0oooo00090?ooo`04
0?oog`2LB00004RL0=ooo`80oooo00<0okmd0018W03Oool0:03oool01P3OW4P0001d0;ooo`3ooml0
W4P004RLg`L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool00P3oool01P000000oooo0000003oool000000?l001l0
oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo02P0
oooo00@000000?ooo`3oool000008`3oool00`000000oooo0?ooo`0;0?ooo`040000003oool0oooo
000000<0oooo0P0000001@3oool00000000000000000o`0001X0oooo00<000000?ooo`3oool0703o
ool00`000000oooo0?ooo`0=0?ooo`002@3oool00`2oM4P0W=oo0?ooo`030?ooo`030?oog`2LB7@0
_ooo02T0oooo00@0gia800000000000007Bo203oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool01P3oool01@000000oooo0?ooo`3oool000000080oooo
0P0000000`3o0000oooo0?ooo`0M0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0X0?ooo`040000003oool0oooo000002<0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo000000030?ooo`050000003oool000000000003oool00P0000000`3o
0000oooo0?ooo`0H0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool03@3oool000P0
oooo00<0omnL09boo`3oool0103oool00`3ogi`0B7Bo0?ooo`0e0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`050000003oool0oooo0?oo
o`0000000P3oool2000000030?l0003oool0oooo01d0oooo00<000000?ooo`3oool0D03oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo02P0oooo00@000000?ooo`3oool000008`3oool0
0`000000oooo0?ooo`0:0?ooo`040000003oool0oooo000000<0oooo00<000000?ooo`0000000P00
00000`3oool000000?l0000J0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool03@3o
ool000P0oooo00<0oknL09cOo`3oool00`3oool00`3ogi`0B4RL0=ooo`0f0?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`050000003oool0
oooo0?ooo`0000000P3oool200000080o`007P3oool00`000000oooo0?ooo`1@0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0:@3oool2000002@0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo000000040?ooo`040000003oool000000?ooo`8000000P3o000J0?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool03@3oool000P0oooo00@0oono07@000000000M;oo
>`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
1P3oool01@000000oooo0?ooo`3oool000000080oooo00@000000?ooo`000000o`007P3oool00`00
0000oooo0?ooo`1@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0:@3oool20000
02@0oooo00<000000?ooo`3oool02P3oool00`000000oooo000000030?ooo`050000003oool00000
0000003oool00P0000000`3o0000oooo0?ooo`0I0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020000
00030?l0003oool0oooo01`0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo02T0oooo0P00000S0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`000000103oool01@000000oooo000000000000oooo0080000000<0o`000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P0000020?l001d0oooo00<000000?ooo`3oool0D03oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0000000@0oooo00D000000?ooo`3oool000000?ooo`02
000000030?l0003oool0oooo01T0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0=
0?ooo`00203oool00`3ookl0W9cO0?ooo`030?ooo`040?noM0000000001d0;ooocD0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool00P3oool01P000000oooo0000003oool000000?l001d0oooo00<000000?ooo`3oool0
D03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02P0oooo0P00000S0?ooo`030000
003oool0oooo00P0oooo00@000000?ooo`3oool00000103oool01@000000oooo0?ooo`000000oooo
0080000000<0o`000?ooo`3oool06@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo
00d0oooo00080?ooo`040?oo_`2LW=l0oono07@000@0000000<004RL0=ooo`3oool0=03oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo008000000P3o000M0?ooo`030000003oool0oooo
04l0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool00000
0280oooo00<000000?ooo`3oool0203oool00`000000oooo000000040?ooo`030000003oool0oooo
00@000000P3o000K0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool03@3oool000P0
oooo00T0omnL04P000000000001809cOo`3oool0oooO09a8M02oool0=P3oool00`000000oooo0?oo
o`0E0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`030?ooo`060000003oool000000?ooo`000000o`007@3oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0:03oool00`000000oooo0000000R0?ooo`03
0000003oool0oooo00L0oooo00<000000?ooo`000000103oool010000000oooo0?ooo`3oool20000
00030?ooo`000000o`0001`0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`0=0?oo
o`00103oool0103ooml0W4P00000M02ooolo0?ooo`030000003oool0oooo01D0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P0000020?l001`0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo02P0oooo00@000000?ooo`3oool000008@3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool0100000020?l001`0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?oo
o`050000003oool000000?ooo`0000000P3o000K0?ooo`030000003oool0oooo04l0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool07P3oool00`000000oooo0?ooo`040?ooo`80000000<0oooo0000003oool0103oool0
0`000000oooo0?ooo`03000000030?ooo`000000o`0001`0oooo00<000000?ooo`3oool07P3oool0
0`000000oooo0?ooo`0=0?ooo`00203oool01@3ogkl0_moo0?ooo`3oool0_g@000<0000000<0001d
0;ooo`3oool0=03oool2000001L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`0000000`3o0000oooo0?ooo`0I
0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01P3oool;000000`0oooo00@00000
0?ooo`3oool0oooo100000020?ooo`030000003oool0oooo00<0oooo00P000000?ooo`3oool00000
0?ooo`000000oooo00000080o`00703oool00`000000oooo0?ooo`0M0?ooo`8000003`3oool000P0
oooo00<0_g@00000000000000P0000001@0007@0_ooo0?ooo`2oW9`0gooo03D0oooo00<000000?oo
o`3oool05P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`0200000080o`006P3oool00`000000oooo0?ooo`1?0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0080oooo0`00000;0?oooa000000103oool2000000D0oooo00@000000?oo
o`3oool0oooo1@0000000`3o0000oooo0?ooo`0K0?ooo`030000003oool0oooo01h0oooo00<00000
0?ooo`3oool03@3oool000H0oooo00D0_g@000000000000004RL0=ooo`030?ooo`030?oog`2LB000
M;oo0080oooo00<0_g@00018W03Oool0<@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`050000003oool0
00000000003oool00P3o000I0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00D0oooo1000000J0?ooo`030000003o
ool0oooo00<0oooo0P0000060?ooo`060000003oool0oooo0?ooo`000000oooo0`0000000`3o0000
oooo0?ooo`0L0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03@3oool000P0oooo
00<0oknL09cOo`3oool0203oool00`3ogi`0B7Bo0?ooo`0a0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`0000020?l001P0oooo00<000000?ooo`3oool0C`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool01@3oool2000001d0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?ooo`0:0000003oool0oooo0?ooo`00
0000oooo0000003oool000000?l001l0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0=0?ooo`00203oool00`3ookl0M7Bo0?ooo`070?ooo`030?noM000B9`0gooo0380oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`03000000030?l0003oool0oooo01H0oooo00<00000
0?ooo`3oool0C`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02@0oooo00<00000
0?ooo`3oool00`3oool200000080oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`02
000000L0oooo00D000000?ooo`3oool0oooo000000020?ooo`8000000P3o000N0?ooo`030000003o
ool0oooo01l0oooo00<000000?ooo`3oool03@3oool000P0oooo00@0omnL04P0000007@0_ooo1@3o
ool00`3OW4P007Bo0?ooo`0c0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool00000
0?ooo`02000000030?l0003oool0oooo01H0oooo00<000000?ooo`3oool0CP3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool00P3oool2000000D0oooo
00<000000?ooo`3oool06@3oool3000000P0oooo00D000000?ooo`3oool0oooo000000020?ooo`80
00000P3o000O0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool03@3oool000T0oooo
00<0omnL04P0000000001@0000000`18W=l0oooo0?ooo`0c0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0080oooo
00H000000?ooo`3oool000000?ooo`0000020?l001L0oooo00<000000?ooo`3oool0CP3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool00P0000080?oo
o`030000003oool0oooo01L0oooo0P0000080?ooo`8000000`3oool00`000000oooo0?ooo`030000
00030?l0003oool0oooo01h0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0=0?oo
o`003P3oool00`3o_g@0001809cOo`0f0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00H000000?ooo`3o
ool000000?ooo`0000020?l001H0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo0240oooo00@000000?ooo`00000000002`3oool00`000000oooo0?oo
o`0D0?ooo`<00000203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0000000<0
o`000?ooo`3oool07`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00d0oooo000=
0?ooo`030;md0000M;l0oooo03L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool01`000000oooo0?ooo`3o
ool000000?ooo`0000000P3o000E0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0Q0?ooo`8000003P3oool00`000000oooo0?ooo`0@0?ooo`<00000
0P3oool00`000000oooo0?ooo`040?ooo`800000103oool00`000000oooo00000003000000030?l0
003oool0oooo0200oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0=0?ooo`00203o
ool01P3ogi`0B00007Boo`3o_g@0000007BoocT0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool010000000
oooo0?ooo`0000020?ooo`030000003o0000oooo01@0oooo00<000000?ooo`3oool0CP3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo01l0oooo0P00000@0?ooo`030000003oool0oooo
00`0oooo0`0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool01000
0000oooo0000003oool2000000030?l0003oool0oooo0200oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`0=0?ooo`00203oool0103ogi`0B00000000000M;lk0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo0080oooo00@000000?ooo`3oool000000P3oool00`000000o`000000000C0?ooo`030000003o
ool0oooo04h0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0M0?ooo`8000004`3o
ool00`000000oooo0?ooo`090?ooo`800000203oool010000000oooo0?ooo`3oool2000000D0oooo
00@000000?ooo`000000oooo0P0000000`3o0000oooo0?ooo`0Q0?ooo`030000003oool0oooo0200
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00P3oool010000000oooo
0?ooo`0000020?ooo`030000003o0000000001<0oooo00<000000?ooo`3oool0C@3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo01X0oooo0P0000000`3oool000000?ooo`0C0?ooo`80
00001P3oool3000000T0oooo00<000000?ooo`3oool00P0000050?ooo`80000000<0oooo0000003o
ool00P0000000`3o0000oooo0?ooo`0R0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3o
ool03@3oool004L0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool2000000050?ooo`000000oooo0?ooo`00
00000P3o000B0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0I0?ooo`040000003oool0oooo000001L0oooo00<000000?ooo`3oool00`00000<0?oo
o`<000001P3oool00`000000oooo0?ooo`04000000030?l0003oool0oooo0280oooo00<000000?oo
o`3oool08@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`06
0000003oool000000000003oool000000P3o000A0?ooo`030000003oool0oooo04d0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0F0?ooo`8000000`3oool00`000000oooo0?ooo`0E
0?ooo`<000003P3oool2000000L0oooo00<000000?ooo`0000000`0000020?l002D0oooo00<00000
0?ooo`3oool08@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool02@3oool2000000@0oooo00D000000?oo
o`3oool000000?ooo`02000000030?l0003oool0oooo00h0oooo00<000000?ooo`3oool0C@3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo01@0oooo0P0000040?ooo`030000003oool0
oooo01@0oooo0P0000030?ooo`8000002P3oool3000000H0oooo0P0000000`3oool0000000000002
00000080o`009P3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00d0oooo00170?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0:0?oo
o`030000003oool0oooo0080oooo00D000000?ooo`3oool000000?ooo`0200000080o`003`3oool0
0`000000oooo0?ooo`1<0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0140oooo0`0000070?ooo`@00000103oool2
00000080oooo00<000000?ooo`3oool00P3oool2000000040?ooo`00000000000?ooo`8000000P3o
000V0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool03@3oool004L0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool00P3oool2000000030?ooo`0000000000008000000P3o000>0?ooo`030000003oool0
oooo04`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0?0?ooo`8000001P3oool0
0`000000oooo0?ooo`0?0?ooo`<000003P3oool600000080oooo00<000000?ooo`00000020000003
0?l002L0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool20000
01l0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P0000000`3oool000000?l0000=0?ooo`030000003oool0oooo
04`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0=0?ooo`8000001`3oool00`00
0000oooo0?ooo`0=0?ooo`<000004@3oool2000000@0oooo1@0000030?ooo`030000003oool00000
00<0o`00:03oool00`000000oooo0?ooo`0R0?ooo`8000003`3oool004L0oooo00<000000?ooo`3o
ool07`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool00`3oool2000000030?ooo`0000000000008000000P3o000;0?ooo`030000003oool0oooo04`0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool02`3oool300000180oooo0P0000070?ooo`800000103oool300000080
o`00:P3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000
003oool0oooo01l0oooo00<000000?ooo`3oool04`3oool2000000/0oooo0P0000050?ooo`030000
003oool0oooo00@000000P3o00090?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0
2`3oool00`000000oooo0?ooo`090?ooo`8000002@3oool00`000000oooo0?ooo`080?ooo`<00000
4P3oool3000000L0oooo0`0000030?ooo`<000000`3o000Z0?ooo`030000003oool0oooo02@0oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0803oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3oool200000080oooo0`00
00030?l000P0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00H0oooo0P00000:0?ooo`030000003oool0oooo00H0oooo0`00000C0?ooo`800000203o
ool2000000040?ooo`000000oooo0?ooo`<000000`3o000/0?ooo`030000003oool0oooo02@0oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?oo
o`0D0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`04000000<0o`001P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo0`00
000C0?ooo`<00000203oool2000000<0oooo100000030?l002d0oooo00<000000?ooo`3oool09@3o
ool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0
oooo01@0oooo0P00000;0?ooo`8000001@3oool200000080oooo100000030?l000@0oooo00<00000
0?ooo`3oool0B`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0080oooo0P00000;
0?ooo`030000003oool0oooo0080oooo0`00000C0?ooo`<00000203oool3000000@0oooo10000002
0?l000030000003oool0oooo02`0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0=
0?ooo`00A`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo01H0oooo00<000000?oo
o`3oool02P3oool2000000D0oooo0P0000020?ooo`@000000`3o00020?ooo`030000003oool0oooo
04/0oooo00<000000?ooo`3oool03@3oool01@000000oooo0?ooo`3oool0000000`0oooo00<00000
0?ooo`3oool00`00000D0?ooo`8000002@3oool2000000@0oooo1@0000020?l000030000003oool0
oooo02d0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`00
0000oooo0?ooo`0S0?ooo`030000003oool0oooo01H0oooo0P00000<0?ooo`8000001@3oool20000
0080oooo100000030?l000030000003oool0oooo04/0oooo00<000000?ooo`3oool03P3oool30000
00d0oooo0`00000D0?ooo`<00000203oool3000000@0oooo1@0000020?l000030000003oool0oooo
02l0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000
oooo0?ooo`0T0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool02`3oool2000000D0
oooo0P0000020?ooo`@000000`3o001<0?ooo`030000003oool0oooo00d0oooo0`00000;0?ooo`<0
00004`3oool4000000P0oooo0`0000040?ooo`H000000P3o00000`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool03@3oool004L0oooo0P00000V0?ooo`030000
003oool0oooo01L0oooo0P00000=0?ooo`8000001@3oool3000000030?ooo`000000000000800000
0`3o001:0?ooo`030000003oool0oooo00`0oooo00@000000?ooo`3oool00000203oool300000003
0?ooo`000000oooo0100oooo0`00000:0?ooo`8000001@3oool2000000030?ooo`000000000000<0
o`000P00000b0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool03@3oool004L0oooo
0P00000W0?ooo`030000003oool0oooo01P0oooo0P00000=0?ooo`8000001P3oool200000080oooo
0`0000030?l000030000003oool0oooo04D0oooo00<000000?ooo`3oool02P3oool2000000@0oooo
00<000000?ooo`3oool00P3oool3000000<0oooo00<000000?ooo`3oool0303oool4000000X0oooo
0`0000040?ooo`<0000000<0oooo0000000000000`3o00020?ooo`030000003oool0oooo0340oooo
00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0000
000W0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0303oool3000000D0oooo0P00
00000`3oool0000000000003000000<0o`00A@3oool00`000000oooo0?ooo`080?ooo`8000001P3o
ool00`000000oooo00000002000000H0oooo00<000000?ooo`3oool02@3oool3000000/0oooo0`00
00040?ooo`<0000000<0oooo000000000000103o00040?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0000000X
0?ooo`8000006P3oool2000000l0oooo0P0000050?ooo`<000000P3oool3000000@0o`00@P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00D0oooo0`0000080?ooo`030000003oool0
oooo00H0oooo1000000<0?ooo`8000001@3oool2000000030?ooo`000000000000D0o`001P3oool0
0`000000oooo0?ooo`0_0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool03@3oool0
04L0oooo00@000000?ooo`3oool00000:@3oool00`000000oooo0?ooo`0I0?ooo`8000003`3oool2
000000D0oooo0`0000030?ooo`<00000103o000o0?ooo`030000003oool0oooo00D0oooo0P000005
0?ooo`<000000`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo0`00000<
0?ooo`@00000103oool6000000@0o`00203oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo02/0oooo00<000000?ooo`3oool03@3oool004L0oooo00@000000?ooo`3oool00000:P3oool0
0`000000oooo0?ooo`0J0?ooo`<000003P3oool3000000<0oooo00<000000?ooo`0000000P000002
0?ooo`@000000`3o00000`000000oooo0?ooo`0j0?ooo`030000003oool0oooo00<0oooo0P000003
0?ooo`@000001P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`@000002`3o
ool4000000@0oooo1`0000040?l000/0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?oo
o`0[0?ooo`030000003oool0oooo00d0oooo00170?ooo`050000003oool0oooo0?ooo`000000:P3o
ool00`000000oooo0?ooo`0L0?ooo`8000003`3oool4000000@0oooo100000000`3oool000000000
0002000000@0o`00>@3oool00`000000oooo0?ooo`020?ooo`030000003oool0000000<000002`3o
ool010000000oooo0?ooo`3oool5000000/0oooo100000040?ooo`P00000103o000=0?ooo`030000
003oool0oooo02l0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo02P0oooo0P00000N0?ooo`800000403o
ool5000000@0oooo200000040?l000030000003oool0oooo03<0oooo00<000000?ooo`3oool01000
000>0?ooo`H000002P3oool5000000@0oooo200000050?l00100oooo00<000000?ooo`3oool0;P3o
ool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0M0?ooo`<000003P3oool0
10000000oooo0?ooo`3oool6000000<0oooo1@0000060?l000800000;`3oool4000000/0oooo1`00
00030?ooo`030000003oool0oooo00H0oooo100000030?ooo`X000001@3o000C0?ooo`030000003o
ool0oooo02h0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool2
000000@0oooo00<000000?ooo`3oool0:P3oool200000200oooo0`00000;0?ooo`030000003oool0
oooo00L0oooo1@0000040?ooo`<000001P3o0002000002L0oooo1P0000070?ooo`H000002@3oool0
0`000000oooo000000020?ooo`D0000000<0oooo000000000000100000000`3oool0000000000002
000000L0o`005P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02h0oooo0P00000?
0?ooo`00A`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo02/0oooo00<000000?oo
o`3oool0803oool5000000L0oooo00<000000?ooo`3oool02`3oool4000000<0oooo1P0000060?l0
01l0oooo100000020?ooo`80000000<0oooo0000003oool01P0000080?oooa80000000<0oooo0000
000000000P0000090?l001T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0_0?oo
o`8000003`3oool004L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0[0?ooo`80
00009@3oool6000000030?ooo`000000oooo0100oooo3@0000070?l000030000003oool0oooo0180
oooo400000070?ooo`L000001@3oool=000000X0o`007@3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo02l0oooo0P00000?0?ooo`00A`3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo02`0oooo0P00000Y0?ooo`D000006@3oool8000000L0o`004`0000020?ooo`800000
0P3oool=000000H0oooo3@00000;0?l00240oooo00<000000?ooo`3oool0;@3oool00`000000oooo
0?ooo`0`0?ooo`8000003`3oool004L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`0^0?ooo`800000:@3oool00`000000oooo0?ooo`07000001`0oooo0`00000;0?l002L00000303o
000Y0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0<@3oool2000000l0oooo0017
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0;`3oool2000002L0oooo00<00000
0?ooo`3oool01`3ooolL000000X0oooo00<000000?l0003o0000;03o00070?ooo`8000009`3oool0
0`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0340oooo00<000000?ooo`0000003`3oool0
04L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0`0?ooo`8000009P3oool00`00
0000oooo0?ooo`0d0?ooo`<000003@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo
0380oooo00<000000?ooo`0000003`3oool004L0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`0a0?ooo`800000903oool00`000000oooo0?ooo`0a0?ooo`<00000403oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool08P3oool00`000000
oooo0?ooo`0Z0?ooo`800000=@3oool00`000000oooo0000000?0?ooo`00A`3oool00`000000oooo
0?ooo`0:0?ooo`030000003oool0oooo0380oooo0P00000S0?ooo`030000003oool0oooo02d0oooo
0`00000C0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0403oool200000240oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0000000l0oooo
00170?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0<`3oool300000200oooo00<0
00000?ooo`3oool0:P3oool3000001H0oooo00<000000?ooo`3oool02P3oool2000001D0oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo03D0oooo00@0
00000?ooo`3oool000003`3oool004L0oooo00<000000?ooo`3oool0303oool00`000000oooo0?oo
o`0e0?ooo`8000007`3oool00`000000oooo0?ooo`0V0?ooo`<000006@3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`0X0?ooo`800000>03oool010000000oooo0?ooo`00000?0?ooo`00A`3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo03H0oooo0`00000L0?ooo`030000003oool0oooo02<0oooo0`00
000L0?ooo`030000003oool0oooo00L0oooo0P00000J0?ooo`8000006`3oool00`000000oooo0?oo
o`0X0?ooo`030000003oool0oooo03P0oooo00@000000?ooo`3oool000003`3oool004L0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0h0?ooo`<000006@3oool00`000000oooo0?oo
o`0O0?ooo`@000007`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01`0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0V0?ooo`800000>`3oool010000000oooo0?oo
o`00000?0?ooo`00A`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo03X0oooo1000
000F0?ooo`030000003oool0oooo01L0oooo1`00000S0?ooo`030000003oool0oooo00@0oooo0P00
000P0?ooo`8000005`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo03X0oooo00D0
00000?ooo`3oool0oooo0000000?0?ooo`00A`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo03d0oooo2000000>0?ooo`030000003oool0oooo0140oooo1P00000Z0?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool08P3oool2000001D0oooo00<000000?ooo`3oool0903o
ool2000003d0oooo00D000000?ooo`3oool0oooo0000000?0?ooo`00A`3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo04@0oooo2`0000040?ooo`030000003oool0oooo00@0oooo3000
000`0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool09@3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo02<0oooo0P00000n0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1>
0?ooo`/00000?03oool00`000000oooo0?ooo`02000002T0oooo0P00000@0?ooo`030000003oool0
oooo0280oooo0P0000100?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03@3oool0
04L0oooo0P00000D0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool0?`3oool00`00
0000oooo0000000]0?ooo`<00000303oool00`000000oooo0?ooo`0Q0?ooo`800000@P3oool01000
0000oooo0?ooo`3oool2000000l0oooo00170?ooo`030000003oool0oooo01@0oooo00<000000?oo
o`3oool0D@3oool00`000000oooo0?ooo`0o0?ooo`800000<@3oool2000000X0oooo00<000000?oo
o`3oool07P3oool3000004<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?oo
o`00A`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3o
ool0?@3oool2000003@0oooo100000050?ooo`030000003oool0oooo01d0oooo0P0000160?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo03/0oooo0P0000000`3oool00000
0?ooo`0g0?ooo`D00000703oool4000004P0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0500oooo00<0
00000?ooo`3oool0>@3oool01@000000oooo0?ooo`3oool0000003/0oooo1P00000B0?ooo`H00000
B`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003o
ool0oooo01P0oooo0P00001@0?ooo`030000003oool0oooo03L0oooo0P0000040?ooo`030000003o
ool0oooo03T0oooo00<000000?ooo`3oool00`3ooolB00000540oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0J0?ooo`800000C`3oool0
0`000000oooo0?ooo`0d0?ooo`8000001P3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0
oooo06H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`00
0000oooo0?ooo`0L0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0<P3oool20000
00P0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool07@3oool2000004d0
oooo00<000000?ooo`3oool0;`3oool2000000X0oooo00<000000?ooo`3oool0=P3oool00`000000
oooo0?ooo`1X0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03@3oool004L0oooo
00<000000?ooo`3oool07`3oool2000004/0oooo00<000000?ooo`3oool0;@3oool2000000`0oooo
00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?oo
o`190?ooo`030000003oool0oooo02X0oooo0P00000>0?ooo`030000003oool0oooo03D0oooo00<0
00000?ooo`3oool0J03oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00d0oooo0017
0?ooo`030000003oool0oooo0280oooo0P0000190?ooo`030000003oool0oooo02P0oooo0P00000@
0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00d0oooo000[0?ooo`040;md00000000000007Boo`<0oooo00<0omnL
04P0001d_ol00P3oool00`3ookl0M000000000020000000304RLg`3oool0oooo00P0oooo00<00000
0?ooo`3oool0903oool2000004P0oooo00<000000?ooo`3oool09@3oool200000180oooo00<00000
0?ooo`3oool0<`3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo00L0oooo00<00000
0?ooo`3oool03@3oool002X0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW018
0000M;oo00@0oooo00<0okmd0000001d_ol02`3oool00`000000oooo0?ooo`0V0?ooo`800000AP3o
ool00`000000oooo0?ooo`0S0?ooo`800000503oool00`000000oooo0?ooo`0b0?ooo`030000003o
ool0oooo06T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0=0?ooo`00:@3oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0oooO09a80000M;l02`3oool0
0`000000oooo0?ooo`0X0?ooo`800000A@3oool00`000000oooo0?ooo`0P0?ooo`8000005P3oool0
0`000000oooo0?ooo`0b0?ooo`030000003oool0oooo06T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`0=0?ooo`00:@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL
0=ooo`P0oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`0Z0?ooo`<00000@`3oool0
0`000000oooo0?ooo`0L0?ooo`<00000603oool00`000000oooo0?ooo`0a0?ooo`030000003oool0
oooo06T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0=0?ooo`00:@3oool00`3o
oml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<0omnL04P0B02Lgol02P3oool00`00
0000oooo0?ooo`0]0?ooo`800000@@3oool00`000000oooo0?ooo`0I0?ooo`<000006`3oool00`00
0000oooo0?ooo`0`0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`0=0?ooo`00:P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0
oooo00<0oono07@0001d_ol02P3oool3000002l0oooo0P0000100?ooo`030000003oool0oooo01D0
oooo0`00000N0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0J`3oool00`000000
oooo0?ooo`070?ooo`<000003`3oool002X0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`090?ooo`030;md0000M;l0oooo00T0oooo00<000000?ooo`3oool0<@3oool2000003h0oooo
00<000000?ooo`3oool04P3oool300000240oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`1Z0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool03@3oool002X0oooo00<0
omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`080?ooo`030=nLB000B9`0gooo00T0oooo00<0
00000?ooo`3oool0<`3oool3000003`0oooo00<000000?ooo`3oool03@3oool4000002@0oooo00<0
00000?ooo`3oool0;P3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool03@3oool002X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=oo
o`080?ooo`030?oOW01804P0W=oo00T0oooo00<000000?ooo`3oool0=P3oool3000003T0oooo00<0
00000?ooo`3oool02@3oool4000002P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`1[0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool03@3oool002/0oooo00<0omnL
04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`080?ooo`030?oo_`1d0000M;oo00T0oooo00<00000
0?ooo`3oool0>@3oool4000003H0oooo00<000000?ooo`3oool01@3oool3000002`0oooo00<00000
0?ooo`3oool0;03oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo00/0oooo00<00000
0?ooo`3oool03@3oool002`0oooo00H0gia80000M02oool0oooO09a80018W=l:0?ooo`030;md0000
M;l0oooo00P0oooo00<000000?ooo`3oool0?@3oool400000380oooo00@000000?ooo`3oool0oooo
1000000_0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0J`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00d0oooo000]0?ooo`040=nLB00000000000001d_`T0oooo
00D0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo0440oooo1P00000/0?ooo`@0
0000<`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`170?ooo`H00000
803oool600000080oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0Z0?ooo`030000
003oool0oooo06d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool00`000000oooo0?ooo`1=0?ooob0000002@3oool00`000000oooo0?ooo`0a0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00d0oooo00170?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0<03oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`3oool09P3oool200000700oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo02l0
oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0N@3oool00`000000oooo
0?ooo`0^0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0L03oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo07X0oooo00<0
00000?ooo`3oool0;@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0740oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`1k
0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1a
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?oo
o`3oool0O03oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0200oooo0P00001d0?oo
o`030000003oool0oooo0100oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3o
ool0O@3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3o
ool0L`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00d0oooo00170?ooo`800000
OP3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
M03oool00`000000oooo0?ooo`0@0?ooo`8000003`3oool004L0oooo00<000000?ooo`3oool0OP3o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0M03o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0
oooo07l0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0L0?ooo`030000003oool0
oooo07D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`00
0000oooo0?ooo`200?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool06P3oool20000
07L0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000
oooo0?ooo`210?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`1g0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03@3oool004L0oooo
00<000000?ooo`3oool0PP3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0C0?ooo`8000003`3oool004L0oooo00<0
00000?ooo`3oool0P`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0N03oool00`000000oooo0?ooo`0B0?ooo`030000003oool0000000l0oooo0017
0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0E
0?ooo`800000NP3oool00`000000oooo0?ooo`0B0?ooo`040000003oool0oooo000000l0oooo0017
0?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool04@3oool01@000000oooo0?ooo`3o
ool0000000l0oooo00170?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo
08L0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
07X0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00d0oooo00170?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`0@0?ooo`800000O@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00@0
oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool0R@3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0O03oool00`000000oooo
0?ooo`0?0?ooo`8000001`3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?oo
o`2:0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`1m0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?oo
o`0=0?ooo`00A`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0303oool2000007l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo08`0oooo00<000000?oo
o`3oool06`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo07l0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00d0oooo00170?ooo`03
0000003oool0oooo08d0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`090?ooo`80
0000P@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool03@3oool004L0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00/0oooo00<000000?ooo`3oool03@3oool004L0oooo0P00002@0?ooo`8000006@3o
ool00`000000oooo0?ooo`060?ooo`800000P`3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00/0oooo0P00000?0?ooo`00A`3oool00`000000oooo0?ooo`2A0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool0103oool2000008D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo
0980oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
08@0oooo00<000000?ooo`3oool02P3oool200000100oooo00<000000?ooo`3oool03@3oool004L0
oooo00<000000?ooo`3oool0T`3oool2000001D0oooo00@000000?ooo`3oool0oooo0P0000270?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0=0?oo
o`00A`3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo0180oooo00@000000?ooo`3o
ool00000R03oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool03@3oool004L0oooo00<000000?ooo`3oool0UP3oool200000180oooo0`0000280?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0=0?ooo`00
A`3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo00h0oooo0P00002:0?ooo`030000
003oool0oooo00L0oooo0P00000E0?ooo`030000003oool0oooo00d0oooo00170?ooo`800000VP3o
ool00`000000oooo0?ooo`0;0?ooo`80000000<0oooo0000003oool0R03oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03@3oool004L0oooo00@00000
0?ooo`0000000000V@3oool2000000T0oooo0P0000030?ooo`030000003oool0oooo08H0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00d0oooo0017
0?ooo`040000003oool0oooo0?ooo`800000V@3oool00`000000oooo0?ooo`040?ooo`8000001@3o
ool00`000000oooo0?ooo`250?ooo`030000003oool0oooo00H0oooo0P00000I0?ooo`030000003o
ool0oooo00d0oooo00170?ooo`030000003oool0oooo00<0oooo0P00002H0?ooo`8000000P3oool2
000000L0oooo00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool01@3oool20000
09P0oooo0P0000090?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo00L0
oooo0P00002D0?ooo`8000000P3oool2000000L0oooo00<000000?ooo`3oool0P`3oool00`000000
oooo0?ooo`040?ooo`8000007@3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo
0?ooo`090?ooo`<00000SP3oool3000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`230?ooo`030000003oool0oooo0080oooo0P00000O0?ooo`030000003oool0oooo00d0oooo
00170?ooo`030000003oool0oooo00`0oooo0P0000290?ooo`<000002P3oool2000000@0oooo00<0
00000?ooo`3oool0PP3oool010000000oooo0?ooo`3oool200000240oooo00<000000?ooo`3oool0
3@3oool004L0oooo00<000000?ooo`3oool03P3oool2000008D0oooo0P00000?0?ooo`8000000P3o
ool00`000000oooo0?ooo`210?ooo`050000003oool0oooo0?ooo`0000008`3oool00`000000oooo
0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0@0?ooo`800000P03oool3000001<0oooo0`00
00220?ooo`030000003oool0oooo00800000903oool00`000000oooo0?ooo`0=0?ooo`00A`3oool2
000001<0oooo0`00001j0?ooo`<00000603oool00`000000oooo0?ooo`200?ooo`<000009@3oool2
000000l0oooo00170?ooo`030000003oool0oooo01D0oooo0`00001d0?ooo`<000006`3oool30000
07l0oooo0P00000X0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo01P0
oooo0`00001^0?ooo`<000007P3oool00`000000oooo0?ooo`02000007/0oooo0P00000Z0?ooo`03
0000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo01/0oooo0`00001X0?ooo`<00000
8@3oool00`000000oooo0?ooo`020?ooo`800000M`3oool2000000030?ooo`000000oooo02T0oooo
00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3oool07P3oool400000640oooo0`00
000T0?ooo`030000003oool0oooo00@0oooo0P00001c0?ooo`8000000P3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo0280oooo1000001J
0?ooo`<000009`3oool00`000000oooo0?ooo`060?ooo`800000KP3oool3000000<0oooo00<00000
0?ooo`3oool0:P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`0V0?oo
o`@00000D`3oool3000002X0oooo00<000000?ooo`3oool0203oool2000006X0oooo0P0000060?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool03@3oool004L0oooo00<000000?ooo`3o
ool0:P3oool4000004/0oooo1000000]0?ooo`030000003oool0oooo00X0oooo0P00001U0?ooo`<0
00001`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000
003oool0oooo02h0oooo1@0000120?ooo`@00000<@3oool00`000000oooo0?ooo`0<0?ooo`<00000
G`3oool3000000T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0=0?ooo`00A`3o
ool00`000000oooo0?ooo`0c0?ooo`D00000=`3oool6000003D0oooo00<000000?ooo`3oool03`3o
ool4000005T0oooo0P00000;0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool03@3o
ool004L0oooo00<000000?ooo`3oool0>03oool7000002D0oooo2`00000k0?ooo`030000003oool0
oooo01<0oooo1000001B0?ooo`<000003@3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0
oooo00d0oooo00170?ooo`030000003oool0oooo03l0oooo2@00000E0?ooo`L00000AP3oool00`00
0000oooo0?ooo`0G0?ooo`<00000C03oool3000000l0oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`180?oooaD00000C@3oool00`000000
oooo0?ooo`0J0?ooo`@00000A03oool400000140oooo00<000000?ooo`3oool0;`3oool00`000000
oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01h0oooo
1000000l0?ooo`@00000503oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00d0oooo
00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool08P3oool500000380oooo1@00
000G0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool03@3oool004L0oooo00<00000
0?ooo`3oool0ZP3oool00`000000oooo0?ooo`0W0?ooo`L00000903oool7000001/0oooo00<00000
0?ooo`3oool0<P3oool00`000000oooo0?ooo`0=0?ooo`00A`3oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo02h0oooo9000000Q0?ooo`030000003oool0oooo03<0oooo00<000000?oo
o`3oool03@3oool004L0oooo0P00002[0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3o
ool0<`3oool2000000l0oooo00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0
L@3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003o
ool0oooo0:X0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`0f0?ooo`030000003o
ool0oooo00d0oooo00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0K`3oool0
0`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo
0:X0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo
00d0oooo00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0K@3oool00`000000
oooo0?ooo`0i0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo0:X0oooo
00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00d0oooo
00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0K03oool00`000000oooo0?oo
o`0j0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo0:X0oooo00<00000
0?ooo`3oool0J`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo00d0oooo00170?oo
o`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo00d0oooo00170?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3o
ool0J@3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo00d0oooo00170?ooo`030000
003oool0oooo0:X0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0n0?ooo`030000
003oool0oooo00d0oooo000[0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P00000000000030000000304RLg`3oool0oooo00L0oooo00<000000?ooo`3oool0
ZP3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0
3@3oool002X0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`070?ooo`030000003oool0oooo0:X0
oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo00d0
oooo000Y0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3o_g@007Bo
0?ooo`0;0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0I@3oool00`000000oooo
0?ooo`110?ooo`030000003oool0oooo00d0oooo000Y0?ooo`030?oog`2LB000B9cO0080oooo00@0
oooO09a80000B9`0gooo203oool00`3o_g@007Bo0?ooo`0:0?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`1C0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`0=0?ooo`00:@3oool00`3ooml0W4P0001d
_`030?ooo`030=nLB00004P0W=oo00T0oooo00<0okmd04RLg`3oool02@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`0=0?ooo`800000403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`00:P3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0oooo00<0gia804RLg`3oool0203ooooo000005/000003`3oool0
02X0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0;0?ooo`030;md001d_ol0oooo
0?l0ooooL@3oool002X0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`0:0?ooo`03
0?oOW01807@0_ooo0?l0ooooL@3oool002X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P0
04RL0=ooo`0:0?ooo`030?oog`2LB000M;oo0?l0ooooL@3oool002/0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO0?l0ooooL@3oool002`0oooo00H0
gia80000M02oool0oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`3o
0?ooog40oooo000]0?ooo`040=nLB00000000000001d_`T0oooo00D0omnL04P000000000000007Bo
o`3o0?ooog80oooo003o0?ooo`L0oooo00D0_g@0000000000000000004RLg`2U0?ooo`00o`3oool8
0?ooo`030?oog`2LB000B9cO0:H0oooo003o0?ooo`T0oooo00<0gia80018W03Oool0Y@3oool00?l0
oooo2@3oool00`3ogi`0B01d0;ooo`2U0?ooo`00Z`3oool01P3OW4P00000000000000000001809cO
o`L0oooo00D0gia8000000000000000007Boo`020?ooo`030=nLB000B9`0gooo0080oooo0300_g@0
00000000M;l0oooo0?oo_`1d000000000000M02oool0gia800000018W=l0oooo0?oOW01800000000
0018W03Oool0_g@000000000M;l0oooo0?noM0000000001d0;ooo`3oool0omnL04P00000000004RL
0=ooo`2oM0000000001d_`3oool0gia80018W03Oool0oooo0?noM00007@0_ooo0?ooo`3ooml0W4P0
0000001d_ol40?ooo`030?noM0000000000000L0000000T0M;oo0?ooo`3ogi`0B0000000000007@0
_ono07@0001d_ol00P3oool02`3ogi`0B00000000000B9`0gooo0;md0000000007Bo0?ooo`3ogi`0
B0000080000000T0M;oo0=nLB000000004RL0=ooo`3o_g@000000000B02Lgol00P3oool0102oM000
00000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3oool00`3OW4P004RL0=ooo`02
0?ooo`030;md0000000007Bo07D0oooo002/0?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<0omnL
04P0001d_ol00P3oool01P2oM00007Bo0?ooo`3o_g@0001d0;ooo`80oooo02/0gia80018W03O_g@0
M;oo0=nLB000B9`0gooo0?oo_`1d07@0_kmd0000M02LW7@0W=oo0?oOW01804P0W=oo0?oOW018M;l0
_g@0001d_`3o_i`0W=oo0?oOW01807@0_ono09bLg`3oool0omnL04P0B02Lgol0omnL04Qd_`2oM000
07Bo0?noW02Lgol0okmd0000B02Lgol0oooo0?oo_`1d0000M;oo0080oooo00@0_g@0001d_`2oM7@0
_ooo103oool0103ooml0W4P00018W03Oool30?ooo`0;0?oog`2LB00004RL0=ooo`2oM00004RL0=oo
o`3ookl0M01d07Ad0018W=l00P3oool0903ogi`0B01809cOo`3ogi`0B7Bo0;md0000M;l0oknL09cO
o`2oM00004RL0=ooo`3oool0_g@007AdM00007@0_mnL09boo`3o_g@0001809cOg`2LM9`0gooo0?oo
o`3OW4P004RL0=oog`2LM9`0gooo0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`80oooo00<0okmd
0000M02oool00P3oool0103OW4P004RL0=noM01d_omd0?ooo`00[@3oool00`3o_g@0001809cOo`07
0?ooo`030=nLB000B9`0gooo00<0oooo00D0gibL0;ooo`3ookl0M00004RLg`020?ooo`060?oOW018
04P0W=oo0?ooo`2oM00007Bo0P3oool08`3ookl0M4Q804P0B02Lgol0oooo0?oog`2LB000B9cO0?oo
o`3o_g@0M7Ad0018W03Oool0oooo0?oo_`1d0000M;oo0=nLW02oool0oooO09a80018W=l0oooo0?no
M01dM7@004RL0=ooo`3oool0oono07@00018W=l0oooo0?oog`2LB00007Bo0080oooo00<0gia80018
W03Oool01P3oool00`3o_g@0001809cOo`050?ooo`040=nLM02Lgol0_g@0001d_`80oooo0100oono
07@00000B9`0gooo0?ooo`3ooml0W4P004RLg`3oool0okmd07AdM000B9`0gooo0?ooo`2oM00007Bo
0`3oool01`2oM000000007Boo`3oool0oono07@0001d_ol00`3oool00`3o_g@0001809cOo`020?oo
o`030=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0?oo_`1d0000B9cO0080oooo00<0
omnL04P0B02Lgol0M@3oool00:d0oooo00<0oono07@00018W=l01`3oool01@3o_g@0000000000000
0000B9cO00<0oooo00`0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`3oool0okmd0018W03O
ool20?ooo`030?noM0000000B9cO0080oooo00L0_g@00018W03Oool0oooo0;md000004P0W=oo0080
oooo00<0_g@004RLg`3oool00P3oool0502oM00004RL0=ooo`3oool0_g@00000B02Lgol0oooo0?oo
g`2LB00004RL0=ooo`3oool0gia80000001d_ol0oooo0?oOW01804P0W=oo1P3oool00`3ookl0M000
04RLg`070?ooo`030=nLB000B9`0gooo0080oooo00<0okmd0000M02oool00P3oool02`2oM00004RL
0=ooo`3oool0_g@00000B02Lgol0oooo0?noM00007@0_ooo0080oooo00<0oono07@00018W=l00P3o
ool00`2oM00007Bo0?ooo`020?ooo`030?oo_`1d0000B9cO0080oooo00<0okmd0000M02oool00`3o
ool02`3OW4P0001d0;ooo`3ooml0W4P00000M02oool0oooo0?oog`2LB000B9cO07D0oooo002]0?oo
o`030?oog`2LB00007Bo00L0oooo00L0oono07@0B02Lgol0oooo0=nLB0000000B9cO0080oooo00<0
gia80018B018M;l00P3oool01P2oM00007Bo0?ooo`3ookl0M00007Boo`80oooo00<0oooO09a80000
M;l00P3oool01`3o_g@0001d0;ooo`3oool0oono07@0001d_ol00P3oool00`3OW4P004RL0=ooo`02
0?ooo`070?noM00007@0_ooo0?ooo`3ookl0M00007Boo`020?ooo`0;0=nLB000B4P0M9cO0?ooo`3o
_g@0001d07AdM02oool0oooO09a80018W=l01P3oool00`3ooml0W4P0001d_`020?ooo`030?oo_`2L
W=l0oooo0080oooo00<0omnL04P0001d_ol00P3oool00`3ookl0M00007Boo`020?ooo`0;0?noM000
07@0_ooo0?ooo`3ookl0M00007Boo`3oool0oono07@0001d_ol00`3oool01`2oM00004RL0=ooo`3o
ool0gia80018W03Oool00`3oool01`2oM00004RL0=ooo`3oool0oono07@0001d_ol00`3oool01`3o
gi`0B01809cOo`3oool0gia80018B018M;l00P3oool00`2oM00007Bo0?ooo`1d0?ooo`00[P3oool0
0`3OW4P0001804Q800030000000H04RLg`3oool0oooo0?ooo`3o_g@004RL0=ooo`3oool0oono07@0
0018W=l0oooo0?noM00007@0_kmd04RLg`3oool0okmd0000M02oool0oooo0?noM00007@0_ooo0P3o
ool01`3OW4P0001d0;ooo`3ogml0gono07@0001d_ol00P3oool0302oM00007Bo0?ooo`3oool0omnL
04P0M02oool0oooo0?oOg`3Ookl0M00007Boo`80oooo0100_g@0001d_`3oool0oooo0?noM00007@0
W9ad09cOo`3ookl0M00007Bo_`1dM;l0oooo0;md0000B9`0gooo1P3oool01P3OW4P0001d0;ooo`3o
ool0omnL04Qd_`@0oooo00`0_g@00000M02oool0oooo0?oog`2LB000B9cO0?ooo`3ogml0gono07@0
001d_ol20?ooo`070;md0000M;l0oooo0?ooo`3o_g@0001d0;ooo`020?ooo`070=nLB00007@0_ooo
0?ooo`3ogi`0B01809cOo`030?ooo`030=nLB00004P0W=oo0080oooo00<0gia80000M02oool00P3o
ool0303ookl0M00007Boo`3oool0okmd0000M02o_g@0B9cO0?ooo`3o_g@0001d0;ooog@0oooo002^
0?ooo`030?noM00004P0W=oo0080oooo00@0omnL04P0000007@0_ooo0P3oool03`3OW4P0001d0;oo
o`3oool0gia80018W03Oool0oooO09a8001d_ol0gia8001d_`3ogi`0B01809cOo`020?ooo`0;0=nL
B00007@0_ooo0?ooo`2oM000001809cOo`3ogi`0B7Bo0;md0000M;l00P3oool0303o_g@004RL0=oo
o`3oool0oono07@0001d_ol0oooo0?oOW018M;l0_g@0001d_`80oooo0100okmd0018W03Oool0oooo
0?oo_`1d04P0W=oO09a8B02Lgml0W4P004RLg`3o_g@007Bo0?noM00007@0_ooo1P3oool00`3o_g@0
000000000002000000030000M02oool0oooo0080oooo00`0oooO09a80000B9`0gooo0?oo_`2LB4P0
04RL0=ooo`3ogi`0B7Bo0;md0000M;l20?ooo`030?noM000B9`0gooo0080oooo00/0gia80000M02o
ool0oooo0;md000004P0W=oo0?ooo`3ooml0W4P004RLg`030?ooo`030?oOW0180000B9cO00<0oooo
0100_g@00000M02oool0oooo0?noM00007@0_ooo0?ooo`3ooml0W4P007Boo`3OW4P007Bo0?oOW018
04P0W=ooM03oool00:h0oooo00<0oono07@00018W=l00`3oool00`3ogi`0B000001d_`030?ooo`0>
0?oOW018000000000000001d_ol0oono07@000000000B9cO0?ooo`3o_g@000000000B02Lgol30?oo
o`0:0?oOW0180000000004RL_`2LB000B9cO0?ooo`3o_g@00000001d_`80oooo00<0omnL04P0B02L
gol00P3oool01`2oM000B9cO0?ooo`3oool0okmd00000000M;l00P3oool0403ogi`0B01809cOo`3o
okl0M00000000018W=l0oooO09a800000000B9cO0?ooo`3o_g@000000018W03Oool60?ooo`030?oo
_`1d0000M;oo0080oooo00<0_gA809cOo`3oool0103oool02P3OW4P000000000M02o_g@0001d0;oo
o`3oool0okmd00000000M;l20?ooo`030?oOW01804P0W=oo00<0oooo00P0omnL04P000000000B9bL
07@00018W=l0oono07@000<0000000H0M;oo0?noM000000000000018W03Oool30?ooo`0?0?noM000
000000000000001d_ol0oooo0?oo_`1d0000000004RLg`3oool0okmd0000000004P0W=oo07@0oooo
002_0?ooo`030;md0000B9`0gooo00<0oooo00<0_g@00000M02oool05`3oool00`2oM00007Bo0?oo
o`0:0?ooo`030=nLB000B9`0gooo01d0oooo00H0_g@0001d_`3oool0oooo0?oOW01dW=lO0?ooo`03
0?noM00007@0_ooo08`0oooo002_0?ooo`030=nLB00007@0_ooo00<0oooo00<0_g@00000M02oool0
5`3oool00`3o_g@0001d0;ooo`0:0?ooo`030?oOW01807@0_ooo01d0oooo00<0gia80000M02oool0
103oool00`3OW9`0_ooo0?ooo`0L0?ooo`030=nLM02Lgol0oooo08/0oooo002_0?ooo`030?noM000
0000M;oo0080oooo00<0okmd00000000M;l0603oool00`3ogi`0B01809cOo`0:0?ooo`030?oog`2L
B000M;oo01d0oooo00<0okmd0000B02Lgol00`3oool00`3ookl0M01809cOo`0R0?ooo`040?oo_`1d
000004RL0=ooohD0oooo002^0?ooo`030;md00000000000000@0000000<007Bo0?ooo`3oool05`3o
ool00`3ooml0W4P004RLg`0;0?ooo`030=nLB018W=l0oooo01/0oooo00<0_g@00000000000001P00
00000`1d_ol0oooo0?ooo`2Y0?ooo`00cP3oool0103ookl0M00000000000M;l90?ooo`040?oog`2L
B0000000001d_m40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-0.282891, -0.263335, \
0.00173807, 0.00173807}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell[CellGroupData[{

Cell["Period for free oscillation", "ExerciseTitle"],

Cell["\<\
Write a function which calculates the period for free oscillations without \
damping.  Plot the period as a function of total energy.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Parametrically-driven pendulum", "ExerciseTitle"],

Cell["\<\
Analyze the behavior of a parametrically-driven pendulum satisfying the \
equation\
\>", "ExerciseText"],

Cell[BoxData[
    FormBox[
      RowBox[{
        RowBox[{
          FractionBox[\(\[PartialD]\^2 \[Theta]\), \(\[PartialD]t\^2\),
            MultilineFunction->None], "+", 
          RowBox[{"\[Gamma]", " ", 
            FractionBox[\(\[PartialD]\[Theta]\), \(\[PartialD]t\),
              MultilineFunction->None]}], "+", 
          "  ", \(\(\(\[Omega]\_0\^2\)(\ 1\  + \ f\ Cos[\[Omega]\ t]\ )\) 
            Sin[\[Theta]]\)}], " ", "\[Equal]", " ", "0"}], 
      TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "The name describes the origin of this equation in the transformation ",
  Cell[BoxData[
      \(TraditionalForm\`\(\(\ \)\(\[Omega]\_0\^2 \[RightArrow] \ \
\(\[Omega]\_0\^2\)(1 + f\ Cos[\[Omega]\ t])\)\)\)]],
  " that could occur if the length parameter were to oscillate, for example."
}], "ExerciseText",
  CellDingbat->None]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
ScreenStyleEnvironment->"Working",
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
ShowCellLabel->False,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"]},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Notation Package Styles", "Section",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell["\<\
The cells below define certain styles needed by the Notation package. These \
styles serve to make visible otherwise invisible tagboxes.\
\>", "Text",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationTemplateStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 1, 0.850004],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationPatternWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 0.900008, 0.979995],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationMadeBoxesWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[0.900008, 0.889998, 1],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"]
  }, Closed]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 81, 2, 81, "Title"],
Cell[1860, 57, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2340, 76, 31, 0, 83, "Section"],
Cell[2374, 78, 107, 3, 41, "Text"],
Cell[2484, 83, 545, 13, 64, "DisplayFormula",
  Evaluatable->False],
Cell[3032, 98, 709, 21, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3778, 124, 33, 0, 51, "Section"],
Cell[3814, 126, 119, 2, 99, "Input"],
Cell[3936, 130, 145, 4, 95, "Input"],
Cell[4084, 136, 126, 2, 99, "Input"],
Cell[4213, 140, 538, 17, 73, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[4788, 162, 33, 0, 51, "Section"],
Cell[4824, 164, 156, 3, 41, "Text"],
Cell[4983, 169, 625, 10, 151, "Input"],
Cell[5611, 181, 1131, 25, 229, "Input"],
Cell[6745, 208, 1222, 23, 281, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[8004, 236, 27, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[8056, 240, 48, 0, 68, "Subsection"],
Cell[8107, 242, 248, 5, 67, "Text"],
Cell[8358, 249, 226, 5, 96, "Input"],

Cell[CellGroupData[{
Cell[8609, 258, 91, 1, 73, "Input"],
Cell[8703, 261, 141023, 3748, 377, 44604, 2553, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[149741, 4012, 585, 10, 93, "Text"],
Cell[150329, 4024, 285, 5, 99, "Input"],

Cell[CellGroupData[{
Cell[150639, 4033, 91, 1, 73, "Input"],
Cell[150733, 4036, 123476, 4538, 318, 61859, 3773, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[274258, 8580, 50, 0, 68, "Subsection"],
Cell[274311, 8582, 157, 3, 41, "Text"],
Cell[274471, 8587, 278, 5, 99, "Input"],

Cell[CellGroupData[{
Cell[274774, 8596, 323, 6, 125, "Input"],
Cell[275100, 8604, 171251, 4799, 398, 58989, 3409, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[446388, 13408, 334, 7, 125, "Input"],
Cell[446725, 13417, 153804, 4077, 444, 45846, 2740, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[600590, 17501, 33, 0, 54, "ExerciseMain"],

Cell[CellGroupData[{
Cell[600648, 17505, 52, 0, 58, "ExerciseTitle"],
Cell[600703, 17507, 165, 3, 41, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[600905, 17515, 55, 0, 58, "ExerciseTitle"],
Cell[600963, 17517, 114, 3, 41, "ExerciseText"],
Cell[601080, 17522, 573, 14, 64, "DisplayFormula",
  Evaluatable->False],
Cell[601656, 17538, 346, 7, 41, "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

