(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   1064191,      26543]*)
(*NotebookOutlinePosition[   1101141,      27849]*)
(*  CellTagsIndexPosition[   1101097,      27845]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["van der Pol Oscillator", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "\tAn interesting feature of many physical or biological systems is the \
appearance of limit-cycle oscillations.  The classic example is the ",
  StyleBox["van der Pol equation",
    FontSlant->"Italic"]
}], "Text"],

Cell[BoxData[
    FormBox[
      RowBox[{
        FractionBox[\(\[PartialD]\^2 x[t]\), \(\[PartialD]t\^2\),
          MultilineFunction->None], "\[Equal]", 
        RowBox[{\(-x[t]\), "+", 
          RowBox[{"\[CurlyEpsilon]", " ", \((1 - x[t]\^2)\), " ", 
            FractionBox[\(\[PartialD]x[t]\), \(\[PartialD]t\),
              MultilineFunction->None]}]}]}], 
      TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\)]],
  " is a \"damping parameter\".  When \[CurlyEpsilon]\[Rule]0, the equation \
reduces to a simple-harmonic oscillator with unit frequency.  [Note that we \
can always choose the time unit to normalize the frequency to unity.]  For ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] \[NotEqual] 0\)]],
  " there is an interesting nonlinear velocity-dependent force which has a \
damping effect for large displacements and an anti-damping effect for small \
displacements; thus, this nonconservative force supplies energy for small \
displacements or drains energy for large displacements.  The net effect is to \
drive the system toward a stable limit cycle for which the damping and \
anti-damping effects are balanced over the course of each cycle.  This \
equation was originally used by Balthasar van der Pol in the 1920s to study \
nonlinearity in electrical circuits containing vacuum tubes and iron-core \
inductors, but more recently has found application to lasers and biological \
systems as well."
}], "Text"],

Cell["\<\
\tIt is useful to re-express the second-order differential equation as a pair \
of coupled first-order equations of the form\
\>", "Text"],

Cell[BoxData[
    FormBox[
      RowBox[{"{", 
        RowBox[{
          RowBox[{
            RowBox[{
              SuperscriptBox["x", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}], 
            "\[Equal]", \(y[t]\)}], ",", 
          RowBox[{
            RowBox[{
              SuperscriptBox["y", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}], 
            "\[Equal]", \(\(-x[t]\) + \[CurlyEpsilon]\ \((1 - x[t]\^2)\)\ y[
                  t]\)}]}], "}"}], TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where a point in ",
  StyleBox["phase space",
    FontSlant->"Italic"],
  " is represented by the \"position\" vector ",
  Cell[BoxData[
      \(TraditionalForm\`r = {x, y}\)]],
  " and the corresponding phase velocity is then ",
  Cell[BoxData[
      \(TraditionalForm\`v = \(\[PartialD]\_t r = {x\^\[Prime], 
            y\^\[Prime]}\)\)]],
  ".  The equations of motion now take the form"
}], "Text"],

Cell[BoxData[{
    \(TraditionalForm\`\[PartialD]\_t x[t] \[Equal] v\_x[x, y]\), "\n", 
    \(TraditionalForm\`\[PartialD]\_t y[t] \[Equal] 
      v\_y[x, y]\)}], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where it is important to recognize that there is no explicit time \
dependence in the phase velocity.   Thus, the driving terms can be described \
as a time-independent velocity field in phase space.  Systems of this type \
are often described as ",
  StyleBox["autonomous",
    FontSlant->"Italic"],
  " because, in the absence of explicit time dependence, the phase point at \
any time uniquely determines the subsequent trajectory.  Thus, each point in \
phase space belongs to a unique trajectory and trajectories do not \
intersect."
}], "Text"],

Cell[TextData[{
  "\tThe behavior of autonomous systems can be analyzed in several steps.  \
First, the fixed points ",
  Cell[BoxData[
      \(TraditionalForm\`r\_fp\)]],
  " for the system are determined by the solutions to the system of equations \
",
  Cell[BoxData[
      \(TraditionalForm\`v \[Equal] 0\)]],
  ".  Second, near any fixed-point we can linearize the equations of motion \
by performing a series expansion of the velocity field.  The resulting \
linearized equations of motion can now be expressed in the form"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[PartialD]\_t r[t]\  \[Equal] \ 
      A\[CenterDot]\((r[t] - r\_fp)\)\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`A = \((\[PartialD]v\_i\/\[PartialD]r\_j)\)\_fp\)]],
  " is the stability matrix composed of derivatives of velocity with respect \
to position evaluated at the fixed point.  Next, we construct displacement \
eigenvectors ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\_i\)]],
  " satisfying"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(\(\ \)\(A\[CenterDot]\[Delta]\_i \[Equal] \
\(\[Lambda]\_i\) \[Delta]\_i\)\)\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\_i\)]],
  " is an eigenvalue that governs the time dependence of displacements in the \
direction of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\_i\)]],
  ".  Finally, the displacement of the phase point can be represented by the \
expansion"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`r[t] - 
        r\_fp \[Equal] \[Sum]\+i\( 
              a\_i\) \(\[ExponentialE]\^\(\(\[Lambda]\_i\) 
                t\)\) \[Delta]\_i\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where the constant coefficients ",
  Cell[BoxData[
      \(TraditionalForm\`a\_i\)]],
  " represent the initial displacement"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`r[0] - 
        r\_fp \[Equal] \[Sum]\+i\( a\_i\) \[Delta]\_i\)], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "\tThe dynamical behavior of the system near its fixed points can now be \
classified according to the properties of the eigenvalues of the stability \
matrix.  For simplicity, consider a two-dimensional system such that the \
eigenvalues are either both real or are a pair of complex conjugates.  \
Although one can classify all possibilities rather completely, for the \
present purposes it is sufficient to consider the case of complex \
eigenvalues.  If the real part of an eigenvalue is negative, the \
corresponding fixed point is considered ",
  StyleBox["stable ",
    FontSlant->"Italic"],
  "because small initial displacements are damped exponentially, whereas if \
the real part is positive the fixed point is ",
  StyleBox["unstable",
    FontSlant->"Italic"],
  " because small initial displacements grow exponentially.  However, if it \
should happen that for larger displacements there exists a closed curve in \
phase space on which the normal component of the phase velocity is everywhere \
inward, the system will eventually find itself trapped on a ",
  StyleBox["limit cycle",
    FontSlant->"Italic"],
  ".  This occurs because points inside the limit cycle are driven outward \
while points outside are driven inward, such that the trajectory relaxes \
exponentially toward the limit cycle.  This phenomenon is explored in this \
notebook.  Furthermore, this system provides an example of a ",
  StyleBox["relaxation oscillator",
    FontSlant->"Italic"],
  " for which stress accumulates slowly and then is relieved rapidly.  This \
behavior is illustrated by strobe plots in phase space."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(\(Off[General::spell, General::spell1];\)\), "\n", 
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", 
          FontSize \[Rule] 12};\)\)}], "Input"],

Cell["\<\
$DefaultFont={\"Times\", 12}; 
$TextStyle={FontFamily\[Rule]\"Times\", \
FontSize\[Rule]12,FontSlant\[Rule]\"Italic\"}; \
\>", "Input"],

Cell[BoxData[{
    \(\(Needs["\<Utilities`Notation`\>"];\)\), "\n", 
    \(\(Needs["\<Graphics`PlotField`\>"];\)\), "\n", 
    \(Needs["\<Graphics`Graphics`\>"]\)}], "Input"],

Cell[BoxData[
    \(\(Off[Symbolize::"\<boxSymbolExists\>"];\)\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(x\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(y\_0\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(t\_max\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "\n", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(x\_max\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(y\_max\),
          NotationBoxTag,
          Editable->True], "]"}]}]], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Linear stability analysis", "Section"],

Cell[CellGroupData[{

Cell["Define phase-space position and velocity", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(position = {x, y};\)\), "\n", 
    \(velocity = {y, \[CurlyEpsilon] \((1 - x\^2)\) y\  - \ x}\)}], "Input"],

Cell[BoxData[
    \({y, \(-x\) + \((1 - x\^2)\)\ y\ \[CurlyEpsilon]}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Locate fixed point", "Subsection"],

Cell["\<\
Although it is obvious by inspection that the only fixed point lies at the \
origin, perhaps it is worth using a procedure that can be generalized to \
other systems. \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(fixedPoints\  = \ 
      Solve[MapThread[Equal, {velocity, \ Table[0, {Length[velocity]}]}], 
        position]\)], "Input"],

Cell[BoxData[
    \({{y \[Rule] 0, x \[Rule] 0}}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Construct stability matrix", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(stabilityMatrix\  = \ 
      Outer[D, velocity, position] /. 
          fixedPoints\[LeftDoubleBracket]1\[RightDoubleBracket] // 
        Simplify\)], "Input"],

Cell[BoxData[
    \({{0, 1}, {\(-1\), \[CurlyEpsilon]}}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Examine eigenvalues", "Subsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(evalues\  = \ Eigenvalues[stabilityMatrix]\)], "Input"],

Cell[BoxData[
    \({1\/2\ \((\[CurlyEpsilon] - \@\(\(-4\) + \[CurlyEpsilon]\^2\))\), 
      1\/2\ \((\[CurlyEpsilon] + \@\(\(-4\) + \[CurlyEpsilon]\^2\))\)}\)], \
"Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[Re[evalues]], {\[CurlyEpsilon], \(-4\), 4}, 
        PlotStyle \[Rule] {RGBColor[1, 0, 0], 
            RGBColor[0, 0, 1]}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.119048 0.309017 0.078858 [
[.02381 .29652 -9.34375 -15.6875 ]
[.02381 .29652 9.34375 0 ]
[.2619 .29652 -9.34375 -15.6875 ]
[.2619 .29652 9.34375 0 ]
[.7381 .29652 -5 -15.6875 ]
[.7381 .29652 5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .07244 -18.6875 -7.84375 ]
[.4875 .07244 0 7.84375 ]
[.4875 .1513 -18.6875 -7.84375 ]
[.4875 .1513 0 7.84375 ]
[.4875 .23016 -18.6875 -7.84375 ]
[.4875 .23016 0 7.84375 ]
[.4875 .38787 -10 -7.84375 ]
[.4875 .38787 0 7.84375 ]
[.4875 .46673 -10 -7.84375 ]
[.4875 .46673 0 7.84375 ]
[.4875 .54559 -10 -7.84375 ]
[.4875 .54559 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.08333 .30902 m
.08333 .31277 L
s
.14286 .30902 m
.14286 .31277 L
s
.20238 .30902 m
.20238 .31277 L
s
.32143 .30902 m
.32143 .31277 L
s
.38095 .30902 m
.38095 .31277 L
s
.44048 .30902 m
.44048 .31277 L
s
.55952 .30902 m
.55952 .31277 L
s
.61905 .30902 m
.61905 .31277 L
s
.67857 .30902 m
.67857 .31277 L
s
.79762 .30902 m
.79762 .31277 L
s
.85714 .30902 m
.85714 .31277 L
s
.91667 .30902 m
.91667 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .07244 m
.50625 .07244 L
s
gsave
.4875 .07244 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.5 .1513 m
.50625 .1513 L
s
gsave
.4875 .1513 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .23016 m
.50625 .23016 L
s
gsave
.4875 .23016 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .38787 m
.50625 .38787 L
s
gsave
.4875 .38787 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .46673 m
.50625 .46673 L
s
gsave
.4875 .46673 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .54559 m
.50625 .54559 L
s
gsave
.4875 .54559 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .08821 m
.50375 .08821 L
s
.5 .10399 m
.50375 .10399 L
s
.5 .11976 m
.50375 .11976 L
s
.5 .13553 m
.50375 .13553 L
s
.5 .16707 m
.50375 .16707 L
s
.5 .18284 m
.50375 .18284 L
s
.5 .19862 m
.50375 .19862 L
s
.5 .21439 m
.50375 .21439 L
s
.5 .24593 m
.50375 .24593 L
s
.5 .2617 m
.50375 .2617 L
s
.5 .27747 m
.50375 .27747 L
s
.5 .29325 m
.50375 .29325 L
s
.5 .32479 m
.50375 .32479 L
s
.5 .34056 m
.50375 .34056 L
s
.5 .35633 m
.50375 .35633 L
s
.5 .3721 m
.50375 .3721 L
s
.5 .40365 m
.50375 .40365 L
s
.5 .41942 m
.50375 .41942 L
s
.5 .43519 m
.50375 .43519 L
s
.5 .45096 m
.50375 .45096 L
s
.5 .4825 m
.50375 .4825 L
s
.5 .49828 m
.50375 .49828 L
s
.5 .51405 m
.50375 .51405 L
s
.5 .52982 m
.50375 .52982 L
s
.5 .05667 m
.50375 .05667 L
s
.5 .0409 m
.50375 .0409 L
s
.5 .02513 m
.50375 .02513 L
s
.5 .00936 m
.50375 .00936 L
s
.5 .56136 m
.50375 .56136 L
s
.5 .57713 m
.50375 .57713 L
s
.5 .59291 m
.50375 .59291 L
s
.5 .60868 m
.50375 .60868 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.02381 .01472 m
.06244 .04251 L
.10458 .07349 L
.14415 .10357 L
.18221 .13405 L
.20342 .15222 L
.22272 .17011 L
.23233 .17986 L
.24276 .19157 L
.24754 .1976 L
.25261 .20484 L
.25694 .21233 L
.25813 .21481 L
.2594 .21787 L
.26061 .22149 L
.26171 .2269 L
.263 .23052 L
.26421 .23092 L
.26647 .23167 L
.27155 .23336 L
.2807 .23638 L
.30142 .24325 L
.33997 .25601 L
.38097 .26959 L
.42046 .28267 L
.45843 .29525 L
.49886 .30864 L
.53776 .32152 L
.57913 .33522 L
.61897 .34842 L
.6573 .36112 L
.69808 .37462 L
.71834 .38133 L
.72825 .38462 L
.7326 .38605 L
.73489 .38681 L
.73616 .38723 L
.73735 .38763 L
.7385 .3834 L
.73975 .37911 L
.74089 .37669 L
.74193 .37493 L
.7462 .36981 L
.74835 .36788 L
.75072 .36604 L
.75552 .36293 L
.76563 .35799 L
.7751 .35449 L
.79539 .34895 L
.81392 .34523 L
Mistroke
.83456 .34198 L
.85368 .33954 L
.89193 .33572 L
.93111 .33277 L
.97275 .33033 L
.97619 .33015 L
Mfstroke
0 0 1 r
.02381 .28789 m
.06244 .28568 L
.10458 .28261 L
.14415 .27875 L
.18221 .27348 L
.20178 .26972 L
.22272 .26425 L
.23293 .26063 L
.23832 .25831 L
.24402 .25537 L
.24901 .25219 L
.25165 .25016 L
.25449 .24753 L
.2571 .24449 L
.25949 .24063 L
.26064 .23787 L
.26189 .23091 L
.26308 .23055 L
.26416 .23091 L
.26666 .23173 L
.26893 .23248 L
.27401 .23417 L
.28315 .23719 L
.30387 .24406 L
.34336 .25714 L
.38133 .26971 L
.42175 .2831 L
.46066 .29599 L
.50202 .30969 L
.54187 .32288 L
.5802 .33558 L
.62098 .34909 L
.66025 .36209 L
.698 .37459 L
.71897 .38154 L
.72816 .38458 L
.73067 .38542 L
.73336 .38631 L
.73466 .38674 L
.7359 .38715 L
.737 .38751 L
.7382 .39026 L
.73937 .39647 L
.74048 .39986 L
.74176 .40298 L
.74293 .40545 L
.74736 .41315 L
.75197 .41978 L
.75705 .42624 L
.76737 .43786 L
Mistroke
.77689 .44754 L
.81513 .48177 L
.85583 .51445 L
.89501 .54424 L
.93665 .57486 L
.97619 .60332 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00
f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool3
00000=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo000>0?ooo`030?l0003oool0oooo0<P0
oooo00<000000?ooo`3oool0e03oool000l0oooo00<0o`000?ooo`3oool0a`3oool00`000000oooo
0?ooo`3D0?ooo`00403oool20?l00<L0oooo00<000000?ooo`3oool0e03oool00180oooo00<0o`00
0?ooo`3oool0a03oool00`000000oooo0?ooo`3D0?ooo`004`3oool00`3o0000oooo0?ooo`330?oo
o`<00000e03oool001@0oooo0P3o00330?ooo`030000003oool0oooo0=@0oooo000F0?ooo`030?l0
003oool0oooo0<00oooo00<000000?ooo`3oool0e03oool001L0oooo00<0o`000?ooo`3oool0_`3o
ool00`000000oooo0?ooo`3D0?ooo`00603oool20?l00;l0oooo00<000000?ooo`3oool0e03oool0
01X0oooo00<0o`000?ooo`3oool0_03oool00`000000oooo0?ooo`3D0?ooo`006`3oool00`3o0000
oooo0?ooo`2k0?ooo`030000003oool0oooo0=@0oooo000L0?ooo`80o`00^`3oool300000=@0oooo
000N0?ooo`030?l0003oool0oooo0;P0oooo00<000000?ooo`3oool0e03oool001l0oooo0P3o002h
0?ooo`030000003oool0oooo0=@0oooo000Q0?ooo`030?l0003oool0oooo0;D0oooo00<000000?oo
o`3oool0e03oool00280oooo00<0o`000?ooo`3oool0]03oool00`000000oooo0?ooo`3D0?ooo`00
8`3oool20?l00;@0oooo00<000000?ooo`3oool0e03oool002D0oooo00<0o`000?ooo`3oool0/@3o
ool00`000000oooo0?ooo`3D0?ooo`009P3oool00`3o0000oooo0?ooo`2`0?ooo`<00000e03oool0
02L0oooo0P3o002`0?ooo`030000003oool0oooo0=@0oooo000Y0?ooo`030?l0003oool0oooo09`0
oooo00<0oono07@0000000000P0000000`1d_ol0oooo0?ooo`090?ooo`030000003oool0oooo0=@0
oooo000Z0?ooo`80o`00W03oool00`3o_g@00000001d_`020?ooo`030;mdB02Lgol0oooo00T0oooo
00<000000?ooo`3oool0e03oool002`0oooo00<0o`000?ooo`3oool0WP3oool00`3ogi`0B4RL0=oo
o`090?ooo`030000003oool0oooo0=@0oooo000]0?ooo`030?l0003oool0oooo09d0oooo00<0oono
07@0001d_ol02@3oool00`000000oooo0?ooo`3D0?ooo`00;P3oool20?l008/0oooo00<0oooO09a8
000000001`0000000`18W=l0oooo0?ooo`050?ooo`030?oOW0180000M;oo00T0oooo00<000000?oo
o`3oool0e03oool00300oooo00<0o`000?ooo`3oool0V@3oool0103ogi`0B0000000B02Lgol90?oo
o`@00000d`3oool00340oooo00<0o`000?ooo`3oool0UP3oool01@3ogi`0B000000000000000M;oo
00X0oooo00<000000?ooo`3oool0e03oool00380oooo0P3o002H0?ooo`030?oo_`1d000007Bo00X0
oooo00<000000?ooo`3oool0e03oool003@0oooo00<0o`000?ooo`3oool0U`3oool00`3OW4P007Bo
0?ooo`080?ooo`030000003oool0oooo0=@0oooo000e0?ooo`030?l0003oool0oooo09H0oooo00<0
omnL04P0B02Lgol0203oool00`000000oooo0?ooo`3D0?ooo`00=P3oool20?l009<0oooo00H0_g@0
07Boo`3oool0gia80000B02Lgol80?ooo`030000003oool0oooo0=@0oooo000h0?ooo`030?l0003o
ool0oooo0940oooo00@0_g@0000000000000B9cO2@3oool300000=@0oooo000i0?ooo`030?l0003o
ool0oooo09d0oooo00<000000?ooo`3oool0e03oool003X0oooo0P3o002M0?ooo`030000003oool0
oooo0=@0oooo000l0?ooo`030?l0003oool0oooo09X0oooo00<000000?ooo`3oool0e03oool003d0
oooo00<0o`000?ooo`3oool0V@3oool00`000000oooo0?ooo`3D0?ooo`00?P3oool20?l009T0oooo
00<000000?ooo`3oool0e03oool00400oooo00<0o`000?ooo`3oool0UP3oool00`000000oooo0?oo
o`3D0?ooo`00@@3oool00`3o0000oooo0?ooo`2E0?ooo`<00000e03oool00480oooo00<0o`000?oo
o`3oool0U03oool00`000000oooo0?ooo`3D0?ooo`00@`3oool20?l009@0oooo00<000000?ooo`3o
ool0e03oool004D0oooo00<0o`000?ooo`3oool0T@3oool00`000000oooo0?ooo`3D0?ooo`00AP3o
ool00`3o0000oooo0?ooo`2@0?ooo`030000003oool0oooo0=@0oooo00170?ooo`030?l0003oool0
oooo08l0oooo00<000000?ooo`3oool0e03oool004P0oooo00<0o`000?ooo`3oool0SP3oool00`00
0000oooo0?ooo`3D0?ooo`00B@3oool20?l008h0oooo0`00003D0?ooo`00B`3oool00`3o0000oooo
0?ooo`2;0?ooo`030000003oool0oooo0=@0oooo001<0?ooo`030?l0003oool0oooo08X0oooo00<0
00000?ooo`3oool0e03oool004d0oooo00<0o`000?ooo`3oool0R@3oool00`000000oooo0?ooo`3D
0?ooo`00CP3oool20?l008T0oooo00<000000?ooo`3oool0e03oool00500oooo00<0o`000?ooo`3o
ool0QP3oool00`000000oooo0?ooo`3D0?ooo`00D@3oool00`3o0000oooo0?ooo`250?ooo`<00000
e03oool00580oooo00<0o`000?ooo`3oool0Q03oool00`000000oooo0?ooo`3D0?ooo`00D`3oool0
0`3o0000oooo0?ooo`1b0?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0203oool0
0`000000oooo0?ooo`3D0?ooo`00E03oool00`3o0000oooo0?ooo`1b0?ooo`030?noM000M;l0oooo
0080oooo00<0_gAd0;ooo`3oool0203oool00`000000oooo0?ooo`3D0?ooo`00E@3oool20?l007<0
oooo00<0okmd001d_`3oool0303oool00`000000oooo0?ooo`3D0?ooo`00E`3oool00`3o0000oooo
0?ooo`1a0?ooo`030?noM000M;l0oooo00/0oooo00<000000?ooo`3oool0e03oool005P0oooo00<0
o`000?ooo`3oool0H03oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0
okmd04RLg`3oool02P3oool00`000000oooo0?ooo`3D0?ooo`00F@3oool00`3o0000oooo0?ooo`1a
0?ooo`030=nLB018W=l0oooo00T0oooo1000003C0?ooo`00FP3oool00`3o0000oooo0?ooo`1a0?oo
o`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0e03oool005/0oooo00<0o`000?ooo`3o
ool0L03oool00`3ogi`0B01d0;ooo`080?ooo`030000003oool0oooo0=@0oooo001L0?ooo`030?l0
003oool0oooo06l0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`3D0?ooo`00G@3o
ool00`3o0000oooo0?ooo`1^0?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0e03o
ool005h0oooo0P3o001Z0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?ooo`030000
003oool0oooo0=@0oooo001P0?ooo`030?l0003oool0oooo06P0oooo00D0omnL04P0000000000000
07Boo`090?ooo`030000003oool0oooo0=@0oooo001Q0?ooo`030?l0003oool0oooo07D0oooo0`00
003D0?ooo`00HP3oool00`3o0000oooo0?ooo`1d0?ooo`030000003oool0oooo0=@0oooo001S0?oo
o`030?l0003oool0oooo07<0oooo00<000000?ooo`3oool0e03oool006@0oooo0P3o001c0?ooo`03
0000003oool0oooo0=@0oooo001V0?ooo`030?l0003oool0oooo0700oooo00<000000?ooo`3oool0
e03oool006L0oooo00<0o`000?ooo`3oool0K`3oool00`000000oooo0?ooo`3D0?ooo`00J03oool0
0`3o0000oooo0?ooo`1^0?ooo`030000003oool0oooo0=@0oooo001Y0?ooo`030?l0003oool0oooo
06d0oooo0`00003D0?ooo`00J@3oool00`3o0000oooo0?ooo`1]0?ooo`030000003oool0oooo0=@0
oooo001Z0?ooo`030?l0003oool0oooo06`0oooo00<000000?ooo`3oool0e03oool006/0oooo00<0
o`000?ooo`3oool0J`3oool00`000000oooo0?ooo`3D0?ooo`00K03oool00`3o0000oooo0?ooo`1Z
0?ooo`030000003oool0oooo0=@0oooo001/0?ooo`030?l0003oool0oooo06X0oooo00<000000?oo
o`3oool0e03oool006d0oooo00<0o`000?ooo`3oool0J@3oool300000=@0oooo001^0?ooo`030?l0
003oool0oooo06P0oooo00<000000?ooo`3oool0e03oool006h0oooo00<0o`000?ooo`3oool0J03o
ool00`000000oooo0?ooo`3D0?ooo`00K`3oool00`3o0000oooo0?ooo`1W0?ooo`030000003oool0
oooo0=@0oooo001`0?ooo`030?l0003oool0oooo06H0oooo00<000000?ooo`3oool0e03oool00700
oooo00<0o`000?ooo`3oool0IP3oool00`000000oooo0?ooo`3D0?ooo`00L@3oool00`3o0000oooo
0?ooo`1U0?ooo`030000003oool0oooo0=@0oooo001b0?ooo`030?l0003oool0oooo06@0oooo0`00
003D0?ooo`00LP3oool00`3o0000oooo0?ooo`1T0?ooo`030000003oool0oooo0=@0oooo001b0?oo
o`030?l0003oool0oooo05<0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`090?oo
o`030000003oool0oooo0=@0oooo001c0?ooo`030?l0003oool0oooo05@0oooo00<0okmd0000001d
_ol0303oool00`000000oooo0?ooo`3D0?ooo`00L`3oool00`3o0000oooo0?ooo`1D0?ooo`030?oo
g`2LB00007Bo00`0oooo00<000000?ooo`3oool0e03oool007<0oooo00<0o`000?ooo`3oool0E@3o
ool00`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo0=@0oooo001c0?ooo`030?l0003oool0
oooo04D0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW01804P0
W=oo00/0oooo00<000000?ooo`3oool0e03oool007<0oooo0P000?mF0?ooo`030?oo_`1d0000M;oo
00/0oooo1000003C0?ooo`00L`3oool01@000?l0oooo0000o`000?l0o`0005@0oooo00<0_g@0001d
_`3oool02P3oool00`000000oooo0?ooo`3D0?ooo`00L`3oool010000?l0oooo0?ooo`3oool30000
oe80oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`3D0?ooo`00L`3oool00`000?l0
oooo0?ooo`040?ooo`@0003oCP3oool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo0=@0
oooo001b0?ooo`030000o`3oool0oooo00T0oooo0`000?l00`3o0000oooo0?ooo`180?ooo`030?oo
_`1d0000M;oo00X0oooo00<000000?ooo`3oool0e03oool00740oooo00<0003o0?ooo`3oool03@3o
ool30000o`80o`00A`3oool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo0=@0oooo001a
0?ooo`030000o`3oool0oooo0100oooo0`000?l20?l00480oooo00D0oono07@000000000001d0;oo
o`090?ooo`030000003oool0oooo0=@0oooo001`0?ooo`030000o`3oool0oooo01@0oooo10000?l0
0`3o0000oooo0?ooo`1;0?ooo`<00000e03oool006l0oooo00<0003o0?ooo`3oool06@3oool30000
od/0oooo00<000000?ooo`3oool0e03oool006h0oooo00<0003o0?ooo`3oool07@3oool30000odP0
oooo00<000000?ooo`3oool0e03oool006d0oooo00<0003o0?ooo`3oool08@3oool40000od@0oooo
00<000000?ooo`3oool0e03oool006/0oooo0P000?lW0?ooo`040?l000000?l0003o0000od40oooo
00<000000?ooo`3oool0e03oool001<0oooo00<0oooO09a80018W=l0D`3oool20000o`X0oooo00<0
gia80000000000000`0000000`18W=l0oooo0?ooo`0I0?ooo`030?l000000?l0003o03l0oooo00<0
00000?ooo`3oool0GP3oool00`3OW4P00000000000030000000304RLg`3oool0oooo0600oooo00<0
oooO09a80018W=l02P3oool001@0oooo00<0gia80018W03Oool0CP3oool40000o`d0oooo00<0okmd
001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`0L0?ooo`<0003o?03oool3000005l0oooo00<0okmd
001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`1Q0?ooo`030=nLB000B9`0gooo00T0oooo000D0?oo
o`030?oOW01804P0W=oo04X0oooo10000?lB0?ooo`030?noM000M;l0oooo02<0oooo0`000?li0?oo
o`030000003oool0oooo0600oooo00<0okmd001d_`3oool0I@3oool00`3ogi`0B01809cOo`090?oo
o`00403oool00`3o_g@0000000000004000000030018W03Oool0oooo04D0oooo0P000?lG0?ooo`03
0?noM000M;l0oooo02@0oooo00<0o`000000o`000?l0=`3oool00`000000oooo0?ooo`1Q0?ooo`03
0?noM000M;l0oooo0600oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`060?ooo`00
0`3oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB4P0W=oo0P3oool00`2o
M00007Bo0?ooo`140?ooo`<0003o2@3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo
00@0oooo00<0okmd04RLg`3oool09P3oool30000oc@0oooo00<000000?ooo`3oool0HP3oool00`3o
_g@0B9cO0?ooo`1O0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0203oool00140
oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo03h0oooo1P000?lN0?ooo`030=nLB018W=l0
oooo02P0oooo0`000?la0?ooo`030000003oool0oooo06<0oooo00<0gia804RLg`3oool0G`3oool0
1`3ookl0M01d0;ooo`3oool0okmd0018W03Oool0203oool00180oooo00H0omnL04Q8W03Oool0omnL
04P0B02Lgolg0?ooo`L0003o9@3oool00`2oM000M;oo0?ooo`0Z0?ooo`@0003o;@3oool00`000000
oooo0?ooo`1T0?ooo`030;md001d_ol0oooo05l0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol8
0?ooo`004`3oool01@3o_g@007Bo0?oog`2LB000B9cO0380oooo1@000?l/0?ooo`030?oOW01807@0
_ooo02h0oooo0`000?lZ0?ooo`030000003oool0oooo06@0oooo00<0omnL04P0M02oool0H03oool0
1@3o_g@007Bo0?oog`2LB000B9cO00P0oooo000D0?ooo`040=nLB018W=l0_g@0001d_b`0oooo1P00
0?la0?ooo`030?oog`2LB000M;oo0340oooo0`000?lW0?ooo`<00000I03oool00`3ooml0W4P007Bo
o`1Q0?ooo`040=nLB018W=l0_g@0001d_`P0oooo000E0?ooo`040=nLB01dM7@0001d0;ooob00oooo
2`000?lg0?ooo`030?oo_`1d0000B9cO03@0oooo10000?lS0?ooo`030000003oool0oooo06@0oooo
00<0oono07@00018W=l0HP3oool0103OW4P0M7Ad0000M02oool70?ooo`005P3oool00`2oM0000018
09cOo`0C0?ooo`d0003o?P3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0>03oool30000
ob00oooo00<000000?ooo`3oool0H03oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0H`3o
ool00`2oM000001809cOo`070?ooo`005P3oool00`3ooml0W4P004RLg`0:0?ooo`T0003oC03oool0
1@3ogi`0B000000000000000M;oo03`0oooo0`000?lM0?ooo`030000003oool0oooo0640oooo00D0
omnL04P000000000000007Boo`1T0?ooo`030?oog`2LB000B9cO00L0oooo000G0?ooo``0003oV03o
ool0103o0000003o0000o`000?lJ0?ooo`030000003oool0oooo0=@0oooo000>0?ooo`T0003oY`3o
ool00`3o0000003o0000o`0H0?ooo`030000003oool0oooo0=@0oooo00310?ooo`<0003o5@3oool0
0`000000oooo0?ooo`3D0?ooo`00``3oool0103o0000003o0000o`000?lB0?ooo`<00000e03oool0
0<H0oooo00@0o`000000o`000?l0003o3`3oool00`000000oooo0?ooo`3D0?ooo`00b@3oool0103o
0000003o0000o`000?l<0?ooo`030000003oool0oooo0=@0oooo003=0?ooo`<0003o2@3oool00`00
0000oooo0?ooo`3D0?ooo`00d03oool30000o`H0oooo00<000000?ooo`3oool0e03oool00=<0oooo
0`000?l0103o0000oooo0?ooo`00003F0?ooo`00eP3oool30000o`030?l0003oool0oooo0=@0oooo
00030?ooomH000000`000?l00`3o000000000000003@00000040oooo000>0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`040000oa80oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`003P3oool00`000000oooo0?ooo`1R0?oo
o`030000003oool0oooo06<0oooo00<000000?ooo`3oool0103oool30000oe`0oooo00<000000?oo
o`3oool0HP3oool00`000000oooo0?ooo`090?ooo`00f@3oool00`000000oooo0?ooo`060?ooo`04
0?l000000?l0003o0000olX0oooo003I0?ooo`030000003oool0oooo00T0oooo00<0o`000000o`00
0?l00P000?o60?ooo`00f@3oool00`000000oooo0?ooo`0<0?ooo`80o`000`000?o30?ooo`00f@3o
ool3000000l0oooo0P3o00020000ol40oooo003I0?ooo`030000003oool0oooo0180oooo00@0o`00
0000o`000?l0003o_P3oool00=T0oooo00<000000?ooo`3oool05@3oool0103o0000003o0000o`00
0?nk0?ooo`00f@3oool00`000000oooo0?ooo`0H0?ooo`030?l000000?l0003o0:@0oooo2P3o000;
0?ooo`00f@3oool00`000000oooo0?ooo`0K0?ooo`<0003oS`3ooolB0?l001D0oooo003I0?ooo`03
0000003oool0oooo01h0oooo0`000?n00?ooo``0o`009`3oool00=T0oooo00<000000?ooo`3oool0
8@3oool30000o`030?l0003oool0oooo0780oooo203o000c0?ooo`00f@3oool3000002@0oooo0`00
0?l00`3o0000oooo0?ooo`1W0?ooo`P0o`00>`3oool00=T0oooo00<000000?ooo`3oool09`3oool3
0000o`030?l0003oool0oooo05/0oooo2@3o00130?ooo`00f@3oool00`000000oooo0?ooo`0Z0?oo
o`<0003o0P3o001C0?ooo`H0o`00C03oool00=T0oooo00<000000?ooo`3oool0;@3oool30000o`80
o`00C03oool40?l00580oooo003I0?ooo`030000003oool0oooo0300oooo0`000?l00`3o0000oooo
0?ooo`130?ooo`D0o`00EP3oool00=T0oooo00<000000?ooo`3oool0<`3oool30000o`030?l0003o
ool0oooo03`0oooo103o001K0?ooo`00f@3oool00`000000oooo0?ooo`0f0?ooo`<0003o00<0o`00
0?ooo`3oool0=@3oool40?l005l0oooo003I0?ooo`<00000>@3oool40000oc40oooo0`3o001S0?oo
o`00f@3oool00`000000oooo0?ooo`0l0?ooo`040?l000000?l0003o0000obd0oooo00<0o`000?oo
o`3oool0I03oool00=T0oooo00<000000?ooo`3oool0?`3oool0103o0000003o0000o`000?lX0?oo
o`80o`00I`3oool00=T0oooo00<000000?ooo`3oool0@P3oool0103o0000003o0000o`000?lT0?oo
o`030?l0003oool0oooo06L0oooo003I0?ooo`030000003oool0oooo04D0oooo00@0o`000000o`00
0?l0003o7`3oool20?l006X0oooo003I0?ooo`030000003oool0oooo04P0oooo00<0o`000000o`00
0?l0703oool00`3o0000oooo0?ooo`1Z0?ooo`00f@3oool00`000000oooo0?ooo`1;0?ooo`<0003o
603oool00`3o0000oooo0?ooo`1[0?ooo`00f@3oool3000004h0oooo0`000?lD0?ooo`030?l0003o
ool0oooo06`0oooo00380?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool02@3oool0
0`000000oooo0?ooo`1A0?ooo`<0003o403oool00`3o0000oooo0?ooo`1]0?ooo`00bP3oool00`3o
_g@0000007Boo`0<0?ooo`030000003oool0oooo05@0oooo0P000?l20?l000/0oooo00<0o`000?oo
o`3oool0KP3oool00<X0oooo00<0oooO09a80000M;l0303oool00`000000oooo0?ooo`1F0?ooo`<0
003o0P3o00080?ooo`030?l0003oool0oooo06h0oooo003;0?ooo`030=nLB000B9`0gooo00/0oooo
00<000000?ooo`3oool0F@3oool30000o`030?l0003oool0oooo00@0oooo00<0o`000?ooo`3oool0
KP3oool00</0oooo00<0omnL04P0B02Lgol02`3oool00`000000oooo0?ooo`1L0?ooo`@0003o00<0
o`000?ooo`3o0000L@3oool00</0oooo00<0oono07@0001d_ol02`3oool4000005l0oooo0`000?ma
0?ooo`00c03oool00`2oM00007Bo0?ooo`0:0?ooo`030000003oool0oooo06<0oooo00<0003o0?oo
o`3oool0KP3oool00<`0oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`1S0?ooo`03
0000o`3oool0oooo06h0oooo003<0?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool0
H`3oool00`000?l0oooo0?ooo`1^0?ooo`00c03oool00`3ookl0M00007Boo`0:0?ooo`030000003o
ool0oooo06<0oooo00<0003o0?ooo`3oool0KP3oool00<d0oooo00<0_g@0001d_`3oool02@3oool0
0`000000oooo0?ooo`1T0?ooo`030000o`3oool0oooo06d0oooo003;0?ooo`050?oo_`1d00000000
0000M02oool02@3oool00`000000oooo0?ooo`1T0?ooo`030000o`3oool0oooo06d0oooo003I0?oo
o`<00000I03oool00`000?l0oooo0?ooo`1]0?ooo`00f@3oool00`000000oooo0?ooo`1U0?ooo`03
0000o`3oool0oooo06`0oooo003I0?ooo`030000003oool0oooo06H0oooo00<0003o0?ooo`3oool0
J`3oool00=T0oooo00<000000?ooo`3oool0IP3oool00`000?l0oooo0?ooo`1[0?ooo`00f@3oool0
0`000000oooo0?ooo`1W0?ooo`030000o`3oool0oooo06X0oooo003I0?ooo`030000003oool0oooo
06P0oooo00<0003o0?ooo`3oool0J@3oool00=T0oooo00<000000?ooo`3oool0J03oool00`000?l0
oooo0?ooo`1Y0?ooo`00f@3oool3000006T0oooo00<0003o0?ooo`3oool0J03oool00=T0oooo00<0
00000?ooo`3oool0JP3oool00`000?l0oooo0?ooo`1W0?ooo`00f@3oool00`000000oooo0?ooo`1Z
0?ooo`030000o`3oool0oooo06L0oooo003I0?ooo`030000003oool0oooo06/0oooo00<0003o0?oo
o`3oool0IP3oool00=T0oooo00<000000?ooo`3oool0K03oool00`000?l0oooo0?ooo`1U0?ooo`00
f@3oool00`000000oooo0?ooo`1]0?ooo`030000o`3oool0oooo06@0oooo003I0?ooo`<00000K@3o
ool00`000?l0oooo0?ooo`1T0?ooo`00f@3oool00`000000oooo0?ooo`1^0?ooo`030000o`3oool0
oooo06<0oooo003I0?ooo`030000003oool0oooo06l0oooo00<0003o0?ooo`3oool0HP3oool00=T0
oooo00<000000?ooo`3oool0L03oool00`000?l0oooo0?ooo`1Q0?ooo`00f@3oool00`000000oooo
0?ooo`1a0?ooo`030000o`3oool0oooo0600oooo003I0?ooo`030000003oool0oooo0780oooo00<0
003o0?ooo`3oool0G`3oool00=T0oooo00<000000?ooo`3oool0L`3oool00`000?l0oooo0?ooo`1N
0?ooo`00f@3oool3000007@0oooo00<0003o0?ooo`3oool0G@3oool00=T0oooo00<000000?ooo`3o
ool0M@3oool20000oed0oooo00380?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0
203oool00`000000oooo0?ooo`1g0?ooo`030000o`3oool0oooo05X0oooo00390?ooo`030?noM000
M;l0oooo0080oooo00<0_gAd0;ooo`3oool0203oool00`000000oooo0?ooo`1h0?ooo`030000o`3o
ool0oooo05T0oooo003:0?ooo`030?noM000M;l0oooo00`0oooo00<000000?ooo`3oool0N@3oool0
0`000?l0oooo0?ooo`1H0?ooo`00b`3oool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo
07X0oooo00<0003o0?ooo`3oool0E`3oool00<`0oooo00<0okmd04RLg`3oool02P3oool00`000000
oooo0?ooo`1k0?ooo`80003oE`3oool00<d0oooo00<0gia804RLg`3oool02@3oool4000007`0oooo
00<0003o0?ooo`3oool0E03oool00<h0oooo00<0_g@007Boo`3oool0203oool00`000000oooo0?oo
o`1n0?ooo`030000o`3oool0oooo05<0oooo003>0?ooo`030?oOW01807@0_ooo00P0oooo00<00000
0?ooo`3oool0O`3oool00`000?l0oooo0?ooo`1B0?ooo`00cP3oool00`3ooml0W4P007Boo`080?oo
o`030000003oool0oooo0800oooo00<0003o0?ooo`3oool0D@3oool00<h0oooo00<0oono07@00018
W=l0203oool00`000000oooo0?ooo`210?ooo`80003oD@3oool00<X0oooo00L0okmd0018W03Oool0
oooo0=nLB00004P0W=oo00P0oooo00<000000?ooo`3oool0P`3oool00`000?l0oooo0?ooo`1>0?oo
o`00b`3oool01@3ogi`0B000000000000000M;oo00T0oooo00<000000?ooo`3oool0Q03oool00`00
0?l0oooo0?ooo`1=0?ooo`00f@3oool3000008D0oooo00<0003o0?ooo`3oool0C03oool00=T0oooo
00<000000?ooo`3oool0QP3oool20000od`0oooo003I0?ooo`030000003oool0oooo08P0oooo00<0
003o0?ooo`3oool0B@3oool00=T0oooo00<000000?ooo`3oool0R@3oool00`000?l0oooo0?ooo`18
0?ooo`00f@3oool00`000000oooo0?ooo`2:0?ooo`030000o`3oool0oooo04L0oooo003I0?ooo`03
0000003oool0oooo08/0oooo00<0003o0?ooo`3oool0AP3oool00=T0oooo0`00002<0?ooo`80003o
AP3oool00=T0oooo00<000000?ooo`3oool0SP3oool00`000?l0oooo0?ooo`130?ooo`00f@3oool0
0`000000oooo0?ooo`2?0?ooo`030000o`3oool0oooo0480oooo003I0?ooo`030000003oool0oooo
0900oooo00<0003o0?ooo`3oool0@@3oool00=T0oooo00<000000?ooo`3oool0T@3oool00`000?l0
oooo0?ooo`100?ooo`00f@3oool00`000000oooo0?ooo`2B0?ooo`80003o@03oool00=T0oooo00<0
00000?ooo`3oool0U03oool00`000?l0oooo0?ooo`0m0?ooo`00f@3oool3000009D0oooo00<0003o
0?ooo`3oool0?03oool00=T0oooo00<000000?ooo`3oool0UP3oool00`000?l0oooo0?ooo`0k0?oo
o`00f@3oool00`000000oooo0?ooo`2G0?ooo`80003o>`3oool00=T0oooo00<000000?ooo`3oool0
V@3oool00`000?l0oooo0?ooo`0h0?ooo`00f@3oool00`000000oooo0?ooo`2J0?ooo`030000o`3o
ool0oooo03L0oooo003I0?ooo`030000003oool0oooo09/0oooo0P000?lg0?ooo`00f@3oool00`00
0000oooo0?ooo`2M0?ooo`030000o`3oool0oooo03@0oooo003I0?ooo`<00000WP3oool00`000?l0
oooo0?ooo`0c0?ooo`00f@3oool00`000000oooo0?ooo`2O0?ooo`80003o<`3oool00<P0oooo00<0
oono07@0000000000P0000000`1d_ol0oooo0?ooo`090?ooo`030000003oool0oooo0:40oooo00<0
003o0?ooo`3oool0<03oool00<P0oooo00<0okmd00000000M;l00P3oool00`2oM4P0W=oo0?ooo`09
0?ooo`030000003oool0oooo0:80oooo00<0003o0?ooo`3oool0;`3oool00<d0oooo00<0omnL04Q8
W03Oool02@3oool00`000000oooo0?ooo`2S0?ooo`80003o;`3oool00<d0oooo00<0oono07@0001d
_ol02@3oool00`000000oooo0?ooo`2U0?ooo`030000o`3oool0oooo02`0oooo003=0?ooo`030?oO
W0180000M;oo00T0oooo00<000000?ooo`3oool0YP3oool00`000?l0oooo0?ooo`0[0?ooo`00c03o
ool0103ogi`0B0000000B02Lgol90?ooo`@00000YP3oool00`000?l0oooo0?ooo`0Z0?ooo`00bP3o
ool01@3ogi`0B000000000000000M;oo00X0oooo00<000000?ooo`3oool0Z03oool20000obX0oooo
003<0?ooo`030?oo_`1d000007Bo00X0oooo00<000000?ooo`3oool0ZP3oool00`000?l0oooo0?oo
o`0W0?ooo`00cP3oool00`3OW4P007Bo0?ooo`080?ooo`030000003oool0oooo0:/0oooo00<0003o
0?ooo`3oool09P3oool00<h0oooo00<0omnL04P0B02Lgol0203oool00`000000oooo0?ooo`2/0?oo
o`80003o9P3oool00</0oooo00H0_g@007Boo`3oool0gia80000B02Lgol80?ooo`030000003oool0
oooo0:h0oooo00<0003o0?ooo`3oool08`3oool00<`0oooo00@0_g@0000000000000B9cO2@3oool3
00000:l0oooo00<0003o0?ooo`3oool08P3oool00=T0oooo00<000000?ooo`3oool0/03oool20000
ob80oooo003I0?ooo`030000003oool0oooo0;80oooo00<0003o0?ooo`3oool07`3oool00=T0oooo
00<000000?ooo`3oool0/`3oool00`000?l0oooo0?ooo`0N0?ooo`00f@3oool00`000000oooo0?oo
o`2d0?ooo`80003o7P3oool00=T0oooo00<000000?ooo`3oool0]P3oool00`000?l0oooo0?ooo`0K
0?ooo`00f@3oool00`000000oooo0?ooo`2g0?ooo`030000o`3oool0oooo01X0oooo003I0?ooo`<0
0000^03oool20000oaX0oooo003I0?ooo`030000003oool0oooo0;X0oooo00<0003o0?ooo`3oool0
5`3oool00=T0oooo00<000000?ooo`3oool0^`3oool00`000?l0oooo0?ooo`0F0?ooo`00f@3oool0
0`000000oooo0?ooo`2l0?ooo`80003o5P3oool00=T0oooo00<000000?ooo`3oool0_P3oool00`00
0?l0oooo0?ooo`0C0?ooo`00f@3oool00`000000oooo0?ooo`2o0?ooo`80003o4`3oool00=T0oooo
00<000000?ooo`3oool0`@3oool00`000?l0oooo0?ooo`0@0?ooo`00f@3oool300000<80oooo0P00
0?l@0?ooo`00f@3oool00`000000oooo0?ooo`340?ooo`030000o`3oool0oooo00d0oooo003I0?oo
o`030000003oool0oooo0<D0oooo00<0003o0?ooo`3oool0303oool00=T0oooo00<000000?ooo`3o
ool0aP3oool20000o``0oooo003I0?ooo`030000003oool0oooo0<P0oooo00<0003o0?ooo`3oool0
2@3oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool0
0=T0oooo0`00003D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo
0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D
0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-4.27564, -3.9466, \
0.0294977, 0.0445312}}]
}, Closed]],

Cell[TextData[{
  "Thus, we can identify four distinct regimes.  1) For ",
  Cell[BoxData[
      \(TraditionalForm\`\(-2\) < \[CurlyEpsilon] < 0\)]],
  " the fixed-point is stable because the complex eigenvalues have negative \
real parts, such that nearby initial points will spiral inward toward the \
origin.   2) For ",
  Cell[BoxData[
      \(TraditionalForm\`0\  < \ \[CurlyEpsilon] < 2\)]],
  " the fixed-point is unstable because the complex eigenvalues have positive \
real parts, such that nearby initial points will spiral outward.  3) For ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] < \(-2\)\)]],
  " both eigenvalues are real and negative, producing exponential damping of \
initial displacements.  4) For ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] > 2\)]],
  " both eigenvalues are real and positive, producing exponential growth of \
initial displacements."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Limit-cycle oscillations", "Section"],

Cell[TextData[{
  "The most interesting behavior occurs near an unstable fixed point.  \
Although the initial motion tends to carry the system away from the fixed \
point, if restoring forces are encountered for larger displacements, the \
system can find itself trapped on a closed path through phase space known as \
a ",
  StyleBox["limit",
    FontSlant->"Italic"],
  " ",
  StyleBox["cycle",
    FontSlant->"Italic"],
  ".  This occurs because points inside the limit cycle are driven outward, \
while points outside are driven inward.  This behavior can be illustrated by \
solving the equations of motion numerically.  "
}], "Text"],

Cell[CellGroupData[{

Cell["Define display functions", "Subsection"],

Cell["\<\
Here we define several simple display functions that can be used to examine \
various aspects of limit-cycle behavior.  Although it is a relatively easy to \
generalize these functions to handle any pair of autonomous first-order \
equations, we leave that exercise to the student.\
\>", "Text"],

Cell[BoxData[{
    \(\(r[t]\  = \ position /. {x \[Rule] x[t], y \[Rule] y[t]};\)\), "\n", 
    \(\(v[t]\  = \ 
        velocity\  /. {x \[Rule] x[t], \ y \[Rule] y[t]};\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(eq1\)\(\ \)\(=\)\(\ \)\(MapThread[\ 
        Equal, {D[r[t], t], \ v[t]}\ ]\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{
            SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "\[Equal]", \(y[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["y", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "\[Equal]", \(\(-x[t]\) + \[CurlyEpsilon]\ \((1 - x[t]\^2)\)\ y[
                t]\)}]}], "}"}]], "Output"]
}, Closed]],

Cell[BoxData[
    RowBox[{\(trajectory[\[CurlyEpsilon]_, x0_, y0_, tmax_, opts___Rule]\), ":=",
       
      RowBox[{
        RowBox[{"NDSolve", "[", 
          RowBox[{
            RowBox[{"{", 
              RowBox[{
                RowBox[{
                  RowBox[{
                    SuperscriptBox["x", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}], 
                  "\[Equal]", \(y[t]\)}], ",", 
                RowBox[{
                  RowBox[{
                    SuperscriptBox["y", "\[Prime]",
                      MultilineFunction->None], "[", "t", "]"}], 
                  "\[Equal]", \(\(-x[
                        t]\) + \[CurlyEpsilon]\ \((1 - x[t]\^2)\)\ y[t]\)}], 
                ",", \(x[0] \[Equal] x0\), ",", \(y[0] \[Equal] y0\)}], "}"}],
             ",", \({x, y}\), ",", \({t, 0, tmax}\), ",", "opts", 
            ",", \(MaxSteps \[Rule] 10\^4\)}], "]"}], "\[LeftDoubleBracket]", 
        "1", "\[RightDoubleBracket]"}]}]], "Input"],

Cell[BoxData[
    \(\(velocityField[\[CurlyEpsilon]\[CurlyEpsilon]_, xmax_, ymax_, 
          opts___Rule]\  := \ 
        PlotVectorField[
          velocity /. \[CurlyEpsilon] \[Rule] \[CurlyEpsilon]\[CurlyEpsilon], \
{x, \(-xmax\), xmax}, {y, \(-ymax\), ymax}, opts];\)\)], "Input"],

Cell[BoxData[
    \(\(phaseCurve[x0_, \ y0_, \[CurlyEpsilon]_, tmax_, opts___Rule]\  := \ 
        ParametricPlot[
          Evaluate[r[t] /. trajectory[\[CurlyEpsilon], x0, y0, tmax]], {t, 0, 
            tmax}, opts];\)\)], "Input"],

Cell[BoxData[
    \(StrobePlot[x0_, \ y0_, \[CurlyEpsilon]_, tmin_, tmax_, tinc_, 
        opts___]\  := \ 
      Module[{sol}, sol = r[t] /. trajectory[\[CurlyEpsilon], x0, y0, tmax]; 
        ListPlot[Table[sol, {t, tmin, tmax, tinc}], opts];]\)], "Input"],

Cell[BoxData[
    \(\(xyCurves[x0_, \ y0_, \[CurlyEpsilon]_, tmax_]\  := 
        Plot[Evaluate[r[t] /. trajectory[\[CurlyEpsilon], x0, y0, tmax]], {t, 
            0, tmax}, 
          PlotStyle \[Rule] {RGBColor[1, 0, 0], 
              RGBColor[0, 0, 1]}];\)\)], "Input"],

Cell[BoxData[
    \(phaseAndField[x0_, \ y0_, \[CurlyEpsilon]_, tmax_, xmax_, ymax_]\  := 
      Show[{velocityField[\[CurlyEpsilon], xmax, ymax, 
            DisplayFunction \[Rule] Identity], 
          phaseCurve[x0, y0, \[CurlyEpsilon], tmax, 
            DisplayFunction \[Rule] Identity]}, 
        DisplayFunction \[Rule] $DisplayFunction]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
Examine behavior for various initial conditions and relaxation parameters\
\>", "Subsection"],

Cell[CellGroupData[{

Cell["Case 1: initial point inside limit cycle", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseAndField[1, 1, 1, 20, 3, 3];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.156265 0.5 0.156265 [
[ 0 0 0 0 ]
[ 1 1 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto fill
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto
-0.02 0.005 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.0001 w
[ ] 0 setdash
.03121 .03121 m
.02381 .09777 L
s
% Start of user PostScript
-3. -3. -3.0473278 -2.5740498 MAarrowhead1
% End of user PostScript
.03121 .09818 m
.02487 .15628 L
s
% Start of user PostScript
-3. -2.5714286 -3.0405667 -2.1995673 MAarrowhead1
% End of user PostScript
.03121 .16515 m
.02592 .2148 L
s
% Start of user PostScript
-3. -2.1428571 -3.0338056 -1.8250848 MAarrowhead1
% End of user PostScript
.03121 .23212 m
.02698 .27332 L
s
% Start of user PostScript
-3. -1.7142857 -3.0270445 -1.4506023 MAarrowhead1
% End of user PostScript
.03121 .29909 m
.02804 .33184 L
s
% Start of user PostScript
-3. -1.2857143 -3.0202833 -1.0761198 MAarrowhead1
% End of user PostScript
.03121 .36606 m
.02909 .39036 L
s
% Start of user PostScript
-3. -0.85714286 -3.0135222 -0.70163724 MAarrowhead1
% End of user PostScript
.03121 .43303 m
.03015 .44888 L
s
% Start of user PostScript
-3. -0.42857143 -3.0067611 -0.32715472 MAarrowhead1
% End of user PostScript
.03121 .5 m
.03121 .5074 L
s
% Start of user PostScript
-3. 0. -3. 0.047327797 MAarrowhead1
% End of user PostScript
.03121 .56697 m
.03226 .56591 L
s
% Start of user PostScript
-3. 0.42857143 -2.9932389 0.42181031 MAarrowhead1
% End of user PostScript
.03121 .63394 m
.03332 .62443 L
s
% Start of user PostScript
-3. 0.85714286 -2.9864778 0.79629283 MAarrowhead1
% End of user PostScript
.03121 .70091 m
.03437 .68295 L
s
% Start of user PostScript
-3. 1.2857143 -2.9797167 1.1707754 MAarrowhead1
% End of user PostScript
.03121 .76788 m
.03543 .74147 L
s
% Start of user PostScript
-3. 1.7142857 -2.9729555 1.5452579 MAarrowhead1
% End of user PostScript
.03121 .83485 m
.03649 .79999 L
s
% Start of user PostScript
-3. 2.1428571 -2.9661944 1.9197404 MAarrowhead1
% End of user PostScript
.03121 .90182 m
.03754 .85851 L
s
% Start of user PostScript
-3. 2.5714286 -2.9594333 2.2942229 MAarrowhead1
% End of user PostScript
.03121 .96879 m
.0386 .91703 L
s
% Start of user PostScript
-3. 3. -2.9526722 2.6687054 MAarrowhead1
% End of user PostScript
.09818 .03121 m
.09078 .07905 L
s
% Start of user PostScript
-2.5714286 -3. -2.6187564 -2.6938181 MAarrowhead1
% End of user PostScript
.09818 .09818 m
.09184 .14009 L
s
% Start of user PostScript
-2.5714286 -2.5714286 -2.6119953 -2.3031917 MAarrowhead1
% End of user PostScript
.09818 .16515 m
.09289 .20113 L
s
% Start of user PostScript
-2.5714286 -2.1428571 -2.6052341 -1.9125653 MAarrowhead1
% End of user PostScript
.09818 .23212 m
.09395 .26217 L
s
% Start of user PostScript
-2.5714286 -1.7142857 -2.598473 -1.5219389 MAarrowhead1
% End of user PostScript
.09818 .29909 m
.09501 .32322 L
s
% Start of user PostScript
-2.5714286 -1.2857143 -2.5917119 -1.1313125 MAarrowhead1
% End of user PostScript
.09818 .36606 m
.09606 .38426 L
s
% Start of user PostScript
-2.5714286 -0.85714286 -2.5849508 -0.74068612 MAarrowhead1
% End of user PostScript
.09818 .43303 m
.09712 .4453 L
s
% Start of user PostScript
-2.5714286 -0.42857143 -2.5781897 -0.35005972 MAarrowhead1
% End of user PostScript
.09818 .5 m
.09818 .50634 L
s
% Start of user PostScript
-2.5714286 0. -2.5714286 0.040566683 MAarrowhead1
% End of user PostScript
.09818 .56697 m
.09923 .56738 L
s
% Start of user PostScript
-2.5714286 0.42857143 -2.5646675 0.43119308 MAarrowhead1
% End of user PostScript
.09818 .63394 m
.10029 .62842 L
s
% Start of user PostScript
-2.5714286 0.85714286 -2.5579063 0.82181949 MAarrowhead1
% End of user PostScript
.09818 .70091 m
.10135 .68946 L
s
% Start of user PostScript
-2.5714286 1.2857143 -2.5511452 1.2124459 MAarrowhead1
% End of user PostScript
.09818 .76788 m
.1024 .7505 L
s
% Start of user PostScript
-2.5714286 1.7142857 -2.5443841 1.6030723 MAarrowhead1
% End of user PostScript
.09818 .83485 m
.10346 .81155 L
s
% Start of user PostScript
-2.5714286 2.1428571 -2.537623 1.9936987 MAarrowhead1
% End of user PostScript
.09818 .90182 m
.10452 .87259 L
s
% Start of user PostScript
-2.5714286 2.5714286 -2.5308619 2.3843251 MAarrowhead1
% End of user PostScript
.09818 .96879 m
.10557 .93363 L
s
% Start of user PostScript
-2.5714286 3. -2.5241008 2.7749515 MAarrowhead1
% End of user PostScript
.16515 .03121 m
.15775 .06305 L
s
% Start of user PostScript
-2.1428571 -3. -2.1901849 -2.7962007 MAarrowhead1
% End of user PostScript
.16515 .09818 m
.15881 .12623 L
s
% Start of user PostScript
-2.1428571 -2.5714286 -2.1834238 -2.3919141 MAarrowhead1
% End of user PostScript
.16515 .16515 m
.15986 .1894 L
s
% Start of user PostScript
-2.1428571 -2.1428571 -2.1766627 -1.9876275 MAarrowhead1
% End of user PostScript
.16515 .23212 m
.16092 .25258 L
s
% Start of user PostScript
-2.1428571 -1.7142857 -2.1699016 -1.5833409 MAarrowhead1
% End of user PostScript
.16515 .29909 m
.16198 .31576 L
s
% Start of user PostScript
-2.1428571 -1.2857143 -2.1631405 -1.1790543 MAarrowhead1
% End of user PostScript
.16515 .36606 m
.16303 .37893 L
s
% Start of user PostScript
-2.1428571 -0.85714286 -2.1563794 -0.77476765 MAarrowhead1
% End of user PostScript
.16515 .43303 m
.16409 .44211 L
s
% Start of user PostScript
-2.1428571 -0.42857143 -2.1496183 -0.37048104 MAarrowhead1
% End of user PostScript
.16515 .5 m
.16515 .50528 L
s
% Start of user PostScript
-2.1428571 0. -2.1428571 0.033805569 MAarrowhead1
% End of user PostScript
.16515 .56697 m
.1662 .56846 L
s
% Start of user PostScript
-2.1428571 0.42857143 -2.136096 0.43809218 MAarrowhead1
% End of user PostScript
.16515 .63394 m
.16726 .63163 L
s
% Start of user PostScript
-2.1428571 0.85714286 -2.1293349 0.84237879 MAarrowhead1
% End of user PostScript
.16515 .70091 m
.16832 .69481 L
s
% Start of user PostScript
-2.1428571 1.2857143 -2.1225738 1.2466654 MAarrowhead1
% End of user PostScript
.16515 .76788 m
.16937 .75799 L
s
% Start of user PostScript
-2.1428571 1.7142857 -2.1158127 1.650952 MAarrowhead1
% End of user PostScript
.16515 .83485 m
.17043 .82116 L
s
% Start of user PostScript
-2.1428571 2.1428571 -2.1090516 2.0552386 MAarrowhead1
% End of user PostScript
.16515 .90182 m
.17149 .88434 L
s
% Start of user PostScript
-2.1428571 2.5714286 -2.1022905 2.4595252 MAarrowhead1
% End of user PostScript
.16515 .96879 m
.17254 .94751 L
s
% Start of user PostScript
-2.1428571 3. -2.0955293 2.8638118 MAarrowhead1
% End of user PostScript
.23212 .03121 m
.22472 .04977 L
s
% Start of user PostScript
-1.7142857 -3. -1.7616135 -2.8811976 MAarrowhead1
% End of user PostScript
.23212 .09818 m
.22578 .11469 L
s
% Start of user PostScript
-1.7142857 -2.5714286 -1.7548524 -2.4657344 MAarrowhead1
% End of user PostScript
.23212 .16515 m
.22683 .17961 L
s
% Start of user PostScript
-1.7142857 -2.1428571 -1.7480913 -2.0502713 MAarrowhead1
% End of user PostScript
.23212 .23212 m
.22789 .24454 L
s
% Start of user PostScript
-1.7142857 -1.7142857 -1.7413302 -1.6348081 MAarrowhead1
% End of user PostScript
.23212 .29909 m
.22895 .30946 L
s
% Start of user PostScript
-1.7142857 -1.2857143 -1.7345691 -1.219345 MAarrowhead1
% End of user PostScript
.23212 .36606 m
.23 .37438 L
s
% Start of user PostScript
-1.7142857 -0.85714286 -1.7278079 -0.80388184 MAarrowhead1
% End of user PostScript
.23212 .43303 m
.23106 .4393 L
s
% Start of user PostScript
-1.7142857 -0.42857143 -1.7210468 -0.38841869 MAarrowhead1
% End of user PostScript
.23212 .5 m
.23212 .50423 L
s
% Start of user PostScript
-1.7142857 0. -1.7142857 0.027044455 MAarrowhead1
% End of user PostScript
.23212 .56697 m
.23317 .56915 L
s
% Start of user PostScript
-1.7142857 0.42857143 -1.7075246 0.4425076 MAarrowhead1
% End of user PostScript
.23212 .63394 m
.23423 .63407 L
s
% Start of user PostScript
-1.7142857 0.85714286 -1.7007635 0.85797075 MAarrowhead1
% End of user PostScript
.23212 .70091 m
.23529 .69899 L
s
% Start of user PostScript
-1.7142857 1.2857143 -1.6940024 1.2734339 MAarrowhead1
% End of user PostScript
.23212 .76788 m
.23634 .76392 L
s
% Start of user PostScript
-1.7142857 1.7142857 -1.6872413 1.688897 MAarrowhead1
% End of user PostScript
.23212 .83485 m
.2374 .82884 L
s
% Start of user PostScript
-1.7142857 2.1428571 -1.6804801 2.1043602 MAarrowhead1
% End of user PostScript
.23212 .90182 m
.23846 .89376 L
s
% Start of user PostScript
-1.7142857 2.5714286 -1.673719 2.5198233 MAarrowhead1
% End of user PostScript
.23212 .96879 m
.23951 .95868 L
s
% Start of user PostScript
-1.7142857 3. -1.6669579 2.9352865 MAarrowhead1
% End of user PostScript
.29909 .03121 m
.29169 .0392 L
s
% Start of user PostScript
-1.2857143 -3. -1.3330421 -2.9488087 MAarrowhead1
% End of user PostScript
.29909 .09818 m
.29275 .10549 L
s
% Start of user PostScript
-1.2857143 -2.5714286 -1.326281 -2.5246527 MAarrowhead1
% End of user PostScript
.29909 .16515 m
.29381 .17177 L
s
% Start of user PostScript
-1.2857143 -2.1428571 -1.3195199 -2.1004967 MAarrowhead1
% End of user PostScript
.29909 .23212 m
.29486 .23805 L
s
% Start of user PostScript
-1.2857143 -1.7142857 -1.3127587 -1.6763407 MAarrowhead1
% End of user PostScript
.29909 .29909 m
.29592 .30433 L
s
% Start of user PostScript
-1.2857143 -1.2857143 -1.3059976 -1.2521847 MAarrowhead1
% End of user PostScript
.29909 .36606 m
.29697 .37061 L
s
% Start of user PostScript
-1.2857143 -0.85714286 -1.2992365 -0.82802867 MAarrowhead1
% End of user PostScript
.29909 .43303 m
.29803 .43689 L
s
% Start of user PostScript
-1.2857143 -0.42857143 -1.2924754 -0.40387267 MAarrowhead1
% End of user PostScript
.29909 .5 m
.29909 .50317 L
s
% Start of user PostScript
-1.2857143 0. -1.2857143 0.020283342 MAarrowhead1
% End of user PostScript
.29909 .56697 m
.30014 .56945 L
s
% Start of user PostScript
-1.2857143 0.42857143 -1.2789532 0.44443935 MAarrowhead1
% End of user PostScript
.29909 .63394 m
.3012 .63573 L
s
% Start of user PostScript
-1.2857143 0.85714286 -1.2721921 0.86859536 MAarrowhead1
% End of user PostScript
.29909 .70091 m
.30226 .70201 L
s
% Start of user PostScript
-1.2857143 1.2857143 -1.2654309 1.2927514 MAarrowhead1
% End of user PostScript
.29909 .76788 m
.30331 .76829 L
s
% Start of user PostScript
-1.2857143 1.7142857 -1.2586698 1.7169074 MAarrowhead1
% End of user PostScript
.29909 .83485 m
.30437 .83457 L
s
% Start of user PostScript
-1.2857143 2.1428571 -1.2519087 2.1410634 MAarrowhead1
% End of user PostScript
.29909 .90182 m
.30543 .90085 L
s
% Start of user PostScript
-1.2857143 2.5714286 -1.2451476 2.5652194 MAarrowhead1
% End of user PostScript
.29909 .96879 m
.30648 .96713 L
s
% Start of user PostScript
-1.2857143 3. -1.2383865 2.9893754 MAarrowhead1
% End of user PostScript
.36606 .03121 m
.35866 .03136 L
s
% Start of user PostScript
-0.85714286 -3. -0.90447065 -2.9990341 MAarrowhead1
% End of user PostScript
.36606 .09818 m
.35972 .09861 L
s
% Start of user PostScript
-0.85714286 -2.5714286 -0.89770954 -2.5686689 MAarrowhead1
% End of user PostScript
.36606 .16515 m
.36078 .16586 L
s
% Start of user PostScript
-0.85714286 -2.1428571 -0.89094843 -2.1383037 MAarrowhead1
% End of user PostScript
.36606 .23212 m
.36183 .23311 L
s
% Start of user PostScript
-0.85714286 -1.7142857 -0.88418731 -1.7079385 MAarrowhead1
% End of user PostScript
.36606 .29909 m
.36289 .30036 L
s
% Start of user PostScript
-0.85714286 -1.2857143 -0.8774262 -1.2775734 MAarrowhead1
% End of user PostScript
.36606 .36606 m
.36395 .36761 L
s
% Start of user PostScript
-0.85714286 -0.85714286 -0.87066508 -0.84720816 MAarrowhead1
% End of user PostScript
.36606 .43303 m
.365 .43486 L
s
% Start of user PostScript
-0.85714286 -0.42857143 -0.86390397 -0.41684297 MAarrowhead1
% End of user PostScript
.36606 .5 m
.36606 .50211 L
s
% Start of user PostScript
-0.85714286 0. -0.85714286 0.013522228 MAarrowhead1
% End of user PostScript
.36606 .56697 m
.36712 .56936 L
s
% Start of user PostScript
-0.85714286 0.42857143 -0.85038174 0.44388742 MAarrowhead1
% End of user PostScript
.36606 .63394 m
.36817 .63662 L
s
% Start of user PostScript
-0.85714286 0.85714286 -0.84362063 0.87425261 MAarrowhead1
% End of user PostScript
.36606 .70091 m
.36923 .70387 L
s
% Start of user PostScript
-0.85714286 1.2857143 -0.83685952 1.3046178 MAarrowhead1
% End of user PostScript
.36606 .76788 m
.37028 .77112 L
s
% Start of user PostScript
-0.85714286 1.7142857 -0.8300984 1.734983 MAarrowhead1
% End of user PostScript
.36606 .83485 m
.37134 .83837 L
s
% Start of user PostScript
-0.85714286 2.1428571 -0.82333729 2.1653482 MAarrowhead1
% End of user PostScript
.36606 .90182 m
.3724 .90562 L
s
% Start of user PostScript
-0.85714286 2.5714286 -0.81657617 2.5957134 MAarrowhead1
% End of user PostScript
.36606 .96879 m
.37345 .97287 L
s
% Start of user PostScript
-0.85714286 3. -0.80981506 3.0260786 MAarrowhead1
% End of user PostScript
.43303 .03121 m
.42563 .02622 L
s
% Start of user PostScript
-0.42857143 -3. -0.47589923 -3.0318738 MAarrowhead1
% End of user PostScript
.43303 .09818 m
.42669 .09406 L
s
% Start of user PostScript
-0.42857143 -2.5714286 -0.46913811 -2.5977831 MAarrowhead1
% End of user PostScript
.43303 .16515 m
.42775 .16189 L
s
% Start of user PostScript
-0.42857143 -2.1428571 -0.462377 -2.1636924 MAarrowhead1
% End of user PostScript
.43303 .23212 m
.4288 .22972 L
s
% Start of user PostScript
-0.42857143 -1.7142857 -0.45561588 -1.7296017 MAarrowhead1
% End of user PostScript
.43303 .29909 m
.42986 .29756 L
s
% Start of user PostScript
-0.42857143 -1.2857143 -0.44885477 -1.295511 MAarrowhead1
% End of user PostScript
.43303 .36606 m
.43092 .36539 L
s
% Start of user PostScript
-0.42857143 -0.85714286 -0.44209366 -0.8614203 MAarrowhead1
% End of user PostScript
.43303 .43303 m
.43197 .43322 L
s
% Start of user PostScript
-0.42857143 -0.42857143 -0.43533254 -0.42732959 MAarrowhead1
% End of user PostScript
.43303 .5 m
.43303 .50106 L
s
% Start of user PostScript
-0.42857143 0. -0.42857143 0.0067611138 MAarrowhead1
% End of user PostScript
.43303 .56697 m
.43409 .56889 L
s
% Start of user PostScript
-0.42857143 0.42857143 -0.42181031 0.44085182 MAarrowhead1
% End of user PostScript
.43303 .63394 m
.43514 .63672 L
s
% Start of user PostScript
-0.42857143 0.85714286 -0.4150492 0.87494252 MAarrowhead1
% End of user PostScript
.43303 .70091 m
.4362 .70456 L
s
% Start of user PostScript
-0.42857143 1.2857143 -0.40828809 1.3090332 MAarrowhead1
% End of user PostScript
.43303 .76788 m
.43726 .77239 L
s
% Start of user PostScript
-0.42857143 1.7142857 -0.40152697 1.7431239 MAarrowhead1
% End of user PostScript
.43303 .83485 m
.43831 .84022 L
s
% Start of user PostScript
-0.42857143 2.1428571 -0.39476586 2.1772146 MAarrowhead1
% End of user PostScript
.43303 .90182 m
.43937 .90806 L
s
% Start of user PostScript
-0.42857143 2.5714286 -0.38800475 2.6113053 MAarrowhead1
% End of user PostScript
.43303 .96879 m
.44042 .97589 L
s
% Start of user PostScript
-0.42857143 3. -0.38124363 3.0453961 MAarrowhead1
% End of user PostScript
.5 .03121 m
.4926 .02381 L
s
% Start of user PostScript
0. -3. -0.047327797 -3.0473278 MAarrowhead1
% End of user PostScript
.5 .09818 m
.49366 .09184 L
s
% Start of user PostScript
0. -2.5714286 -0.040566683 -2.6119953 MAarrowhead1
% End of user PostScript
.5 .16515 m
.49472 .15986 L
s
% Start of user PostScript
0. -2.1428571 -0.033805569 -2.1766627 MAarrowhead1
% End of user PostScript
.5 .23212 m
.49577 .22789 L
s
% Start of user PostScript
0. -1.7142857 -0.027044455 -1.7413302 MAarrowhead1
% End of user PostScript
.5 .29909 m
.49683 .29592 L
s
% Start of user PostScript
0. -1.2857143 -0.020283342 -1.3059976 MAarrowhead1
% End of user PostScript
.5 .36606 m
.49789 .36395 L
s
% Start of user PostScript
0. -0.85714286 -0.013522228 -0.87066508 MAarrowhead1
% End of user PostScript
.5 .43303 m
.49894 .43197 L
s
% Start of user PostScript
0. -0.42857143 -0.0067611138 -0.43533254 MAarrowhead1
% End of user PostScript
.5 .5 m
.5 .5 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.5 .56697 m
.50106 .56803 L
s
% Start of user PostScript
0. 0.42857143 0.0067611138 0.43533254 MAarrowhead1
% End of user PostScript
.5 .63394 m
.50211 .63605 L
s
% Start of user PostScript
0. 0.85714286 0.013522228 0.87066508 MAarrowhead1
% End of user PostScript
.5 .70091 m
.50317 .70408 L
s
% Start of user PostScript
0. 1.2857143 0.020283342 1.3059976 MAarrowhead1
% End of user PostScript
.5 .76788 m
.50423 .77211 L
s
% Start of user PostScript
0. 1.7142857 0.027044455 1.7413302 MAarrowhead1
% End of user PostScript
.5 .83485 m
.50528 .84014 L
s
% Start of user PostScript
0. 2.1428571 0.033805569 2.1766627 MAarrowhead1
% End of user PostScript
.5 .90182 m
.50634 .90816 L
s
% Start of user PostScript
0. 2.5714286 0.040566683 2.6119953 MAarrowhead1
% End of user PostScript
.5 .96879 m
.5074 .97619 L
s
% Start of user PostScript
0. 3. 0.047327797 3.0473278 MAarrowhead1
% End of user PostScript
.56697 .03121 m
.55958 .02411 L
s
% Start of user PostScript
0.42857143 -3. 0.38124363 -3.0453961 MAarrowhead1
% End of user PostScript
.56697 .09818 m
.56063 .09194 L
s
% Start of user PostScript
0.42857143 -2.5714286 0.38800475 -2.6113053 MAarrowhead1
% End of user PostScript
.56697 .16515 m
.56169 .15978 L
s
% Start of user PostScript
0.42857143 -2.1428571 0.39476586 -2.1772146 MAarrowhead1
% End of user PostScript
.56697 .23212 m
.56274 .22761 L
s
% Start of user PostScript
0.42857143 -1.7142857 0.40152697 -1.7431239 MAarrowhead1
% End of user PostScript
.56697 .29909 m
.5638 .29544 L
s
% Start of user PostScript
0.42857143 -1.2857143 0.40828809 -1.3090332 MAarrowhead1
% End of user PostScript
.56697 .36606 m
.56486 .36328 L
s
% Start of user PostScript
0.42857143 -0.85714286 0.4150492 -0.87494252 MAarrowhead1
% End of user PostScript
.56697 .43303 m
.56591 .43111 L
s
% Start of user PostScript
0.42857143 -0.42857143 0.42181031 -0.44085182 MAarrowhead1
% End of user PostScript
.56697 .5 m
.56697 .49894 L
s
% Start of user PostScript
0.42857143 0. 0.42857143 -0.0067611138 MAarrowhead1
% End of user PostScript
.56697 .56697 m
.56803 .56678 L
s
% Start of user PostScript
0.42857143 0.42857143 0.43533254 0.42732959 MAarrowhead1
% End of user PostScript
.56697 .63394 m
.56908 .63461 L
s
% Start of user PostScript
0.42857143 0.85714286 0.44209366 0.8614203 MAarrowhead1
% End of user PostScript
.56697 .70091 m
.57014 .70244 L
s
% Start of user PostScript
0.42857143 1.2857143 0.44885477 1.295511 MAarrowhead1
% End of user PostScript
.56697 .76788 m
.5712 .77028 L
s
% Start of user PostScript
0.42857143 1.7142857 0.45561588 1.7296017 MAarrowhead1
% End of user PostScript
.56697 .83485 m
.57225 .83811 L
s
% Start of user PostScript
0.42857143 2.1428571 0.462377 2.1636924 MAarrowhead1
% End of user PostScript
.56697 .90182 m
.57331 .90594 L
s
% Start of user PostScript
0.42857143 2.5714286 0.46913811 2.5977831 MAarrowhead1
% End of user PostScript
.56697 .96879 m
.57437 .97378 L
s
% Start of user PostScript
0.42857143 3. 0.47589923 3.0318738 MAarrowhead1
% End of user PostScript
.63394 .03121 m
.62655 .02713 L
s
% Start of user PostScript
0.85714286 -3. 0.80981506 -3.0260786 MAarrowhead1
% End of user PostScript
.63394 .09818 m
.6276 .09438 L
s
% Start of user PostScript
0.85714286 -2.5714286 0.81657617 -2.5957134 MAarrowhead1
% End of user PostScript
.63394 .16515 m
.62866 .16163 L
s
% Start of user PostScript
0.85714286 -2.1428571 0.82333729 -2.1653482 MAarrowhead1
% End of user PostScript
.63394 .23212 m
.62972 .22888 L
s
% Start of user PostScript
0.85714286 -1.7142857 0.8300984 -1.734983 MAarrowhead1
% End of user PostScript
.63394 .29909 m
.63077 .29613 L
s
% Start of user PostScript
0.85714286 -1.2857143 0.83685952 -1.3046178 MAarrowhead1
% End of user PostScript
.63394 .36606 m
.63183 .36338 L
s
% Start of user PostScript
0.85714286 -0.85714286 0.84362063 -0.87425261 MAarrowhead1
% End of user PostScript
.63394 .43303 m
.63288 .43064 L
s
% Start of user PostScript
0.85714286 -0.42857143 0.85038174 -0.44388742 MAarrowhead1
% End of user PostScript
.63394 .5 m
.63394 .49789 L
s
% Start of user PostScript
0.85714286 0. 0.85714286 -0.013522228 MAarrowhead1
% End of user PostScript
.63394 .56697 m
.635 .56514 L
s
% Start of user PostScript
0.85714286 0.42857143 0.86390397 0.41684297 MAarrowhead1
% End of user PostScript
.63394 .63394 m
.63605 .63239 L
s
% Start of user PostScript
0.85714286 0.85714286 0.87066508 0.84720816 MAarrowhead1
% End of user PostScript
.63394 .70091 m
.63711 .69964 L
s
% Start of user PostScript
0.85714286 1.2857143 0.8774262 1.2775734 MAarrowhead1
% End of user PostScript
.63394 .76788 m
.63817 .76689 L
s
% Start of user PostScript
0.85714286 1.7142857 0.88418731 1.7079385 MAarrowhead1
% End of user PostScript
.63394 .83485 m
.63922 .83414 L
s
% Start of user PostScript
0.85714286 2.1428571 0.89094843 2.1383037 MAarrowhead1
% End of user PostScript
.63394 .90182 m
.64028 .90139 L
s
% Start of user PostScript
0.85714286 2.5714286 0.89770954 2.5686689 MAarrowhead1
% End of user PostScript
.63394 .96879 m
.64134 .96864 L
s
% Start of user PostScript
0.85714286 3. 0.90447065 2.9990341 MAarrowhead1
% End of user PostScript
.70091 .03121 m
.69352 .03287 L
s
% Start of user PostScript
1.2857143 -3. 1.2383865 -2.9893754 MAarrowhead1
% End of user PostScript
.70091 .09818 m
.69457 .09915 L
s
% Start of user PostScript
1.2857143 -2.5714286 1.2451476 -2.5652194 MAarrowhead1
% End of user PostScript
.70091 .16515 m
.69563 .16543 L
s
% Start of user PostScript
1.2857143 -2.1428571 1.2519087 -2.1410634 MAarrowhead1
% End of user PostScript
.70091 .23212 m
.69669 .23171 L
s
% Start of user PostScript
1.2857143 -1.7142857 1.2586698 -1.7169074 MAarrowhead1
% End of user PostScript
.70091 .29909 m
.69774 .29799 L
s
% Start of user PostScript
1.2857143 -1.2857143 1.2654309 -1.2927514 MAarrowhead1
% End of user PostScript
.70091 .36606 m
.6988 .36427 L
s
% Start of user PostScript
1.2857143 -0.85714286 1.2721921 -0.86859536 MAarrowhead1
% End of user PostScript
.70091 .43303 m
.69986 .43055 L
s
% Start of user PostScript
1.2857143 -0.42857143 1.2789532 -0.44443935 MAarrowhead1
% End of user PostScript
.70091 .5 m
.70091 .49683 L
s
% Start of user PostScript
1.2857143 0. 1.2857143 -0.020283342 MAarrowhead1
% End of user PostScript
.70091 .56697 m
.70197 .56311 L
s
% Start of user PostScript
1.2857143 0.42857143 1.2924754 0.40387267 MAarrowhead1
% End of user PostScript
.70091 .63394 m
.70303 .62939 L
s
% Start of user PostScript
1.2857143 0.85714286 1.2992365 0.82802867 MAarrowhead1
% End of user PostScript
.70091 .70091 m
.70408 .69567 L
s
% Start of user PostScript
1.2857143 1.2857143 1.3059976 1.2521847 MAarrowhead1
% End of user PostScript
.70091 .76788 m
.70514 .76195 L
s
% Start of user PostScript
1.2857143 1.7142857 1.3127587 1.6763407 MAarrowhead1
% End of user PostScript
.70091 .83485 m
.70619 .82823 L
s
% Start of user PostScript
1.2857143 2.1428571 1.3195199 2.1004967 MAarrowhead1
% End of user PostScript
.70091 .90182 m
.70725 .89451 L
s
% Start of user PostScript
1.2857143 2.5714286 1.326281 2.5246527 MAarrowhead1
% End of user PostScript
.70091 .96879 m
.70831 .9608 L
s
% Start of user PostScript
1.2857143 3. 1.3330421 2.9488087 MAarrowhead1
% End of user PostScript
.76788 .03121 m
.76049 .04132 L
s
% Start of user PostScript
1.7142857 -3. 1.6669579 -2.9352865 MAarrowhead1
% End of user PostScript
.76788 .09818 m
.76154 .10624 L
s
% Start of user PostScript
1.7142857 -2.5714286 1.673719 -2.5198233 MAarrowhead1
% End of user PostScript
.76788 .16515 m
.7626 .17116 L
s
% Start of user PostScript
1.7142857 -2.1428571 1.6804801 -2.1043602 MAarrowhead1
% End of user PostScript
.76788 .23212 m
.76366 .23608 L
s
% Start of user PostScript
1.7142857 -1.7142857 1.6872413 -1.688897 MAarrowhead1
% End of user PostScript
.76788 .29909 m
.76471 .30101 L
s
% Start of user PostScript
1.7142857 -1.2857143 1.6940024 -1.2734339 MAarrowhead1
% End of user PostScript
.76788 .36606 m
.76577 .36593 L
s
% Start of user PostScript
1.7142857 -0.85714286 1.7007635 -0.85797075 MAarrowhead1
% End of user PostScript
.76788 .43303 m
.76683 .43085 L
s
% Start of user PostScript
1.7142857 -0.42857143 1.7075246 -0.4425076 MAarrowhead1
% End of user PostScript
.76788 .5 m
.76788 .49577 L
s
% Start of user PostScript
1.7142857 0. 1.7142857 -0.027044455 MAarrowhead1
% End of user PostScript
.76788 .56697 m
.76894 .5607 L
s
% Start of user PostScript
1.7142857 0.42857143 1.7210468 0.38841869 MAarrowhead1
% End of user PostScript
.76788 .63394 m
.77 .62562 L
s
% Start of user PostScript
1.7142857 0.85714286 1.7278079 0.80388184 MAarrowhead1
% End of user PostScript
.76788 .70091 m
.77105 .69054 L
s
% Start of user PostScript
1.7142857 1.2857143 1.7345691 1.219345 MAarrowhead1
% End of user PostScript
.76788 .76788 m
.77211 .75546 L
s
% Start of user PostScript
1.7142857 1.7142857 1.7413302 1.6348081 MAarrowhead1
% End of user PostScript
.76788 .83485 m
.77317 .82039 L
s
% Start of user PostScript
1.7142857 2.1428571 1.7480913 2.0502713 MAarrowhead1
% End of user PostScript
.76788 .90182 m
.77422 .88531 L
s
% Start of user PostScript
1.7142857 2.5714286 1.7548524 2.4657344 MAarrowhead1
% End of user PostScript
.76788 .96879 m
.77528 .95023 L
s
% Start of user PostScript
1.7142857 3. 1.7616135 2.8811976 MAarrowhead1
% End of user PostScript
.83485 .03121 m
.82746 .05249 L
s
% Start of user PostScript
2.1428571 -3. 2.0955293 -2.8638118 MAarrowhead1
% End of user PostScript
.83485 .09818 m
.82851 .11566 L
s
% Start of user PostScript
2.1428571 -2.5714286 2.1022905 -2.4595252 MAarrowhead1
% End of user PostScript
.83485 .16515 m
.82957 .17884 L
s
% Start of user PostScript
2.1428571 -2.1428571 2.1090516 -2.0552386 MAarrowhead1
% End of user PostScript
.83485 .23212 m
.83063 .24201 L
s
% Start of user PostScript
2.1428571 -1.7142857 2.1158127 -1.650952 MAarrowhead1
% End of user PostScript
.83485 .29909 m
.83168 .30519 L
s
% Start of user PostScript
2.1428571 -1.2857143 2.1225738 -1.2466654 MAarrowhead1
% End of user PostScript
.83485 .36606 m
.83274 .36837 L
s
% Start of user PostScript
2.1428571 -0.85714286 2.1293349 -0.84237879 MAarrowhead1
% End of user PostScript
.83485 .43303 m
.8338 .43154 L
s
% Start of user PostScript
2.1428571 -0.42857143 2.136096 -0.43809218 MAarrowhead1
% End of user PostScript
.83485 .5 m
.83485 .49472 L
s
% Start of user PostScript
2.1428571 0. 2.1428571 -0.033805569 MAarrowhead1
% End of user PostScript
.83485 .56697 m
.83591 .55789 L
s
% Start of user PostScript
2.1428571 0.42857143 2.1496183 0.37048104 MAarrowhead1
% End of user PostScript
.83485 .63394 m
.83697 .62107 L
s
% Start of user PostScript
2.1428571 0.85714286 2.1563794 0.77476765 MAarrowhead1
% End of user PostScript
.83485 .70091 m
.83802 .68424 L
s
% Start of user PostScript
2.1428571 1.2857143 2.1631405 1.1790543 MAarrowhead1
% End of user PostScript
.83485 .76788 m
.83908 .74742 L
s
% Start of user PostScript
2.1428571 1.7142857 2.1699016 1.5833409 MAarrowhead1
% End of user PostScript
.83485 .83485 m
.84014 .8106 L
s
% Start of user PostScript
2.1428571 2.1428571 2.1766627 1.9876275 MAarrowhead1
% End of user PostScript
.83485 .90182 m
.84119 .87377 L
s
% Start of user PostScript
2.1428571 2.5714286 2.1834238 2.3919141 MAarrowhead1
% End of user PostScript
.83485 .96879 m
.84225 .93695 L
s
% Start of user PostScript
2.1428571 3. 2.1901849 2.7962007 MAarrowhead1
% End of user PostScript
.90182 .03121 m
.89443 .06637 L
s
% Start of user PostScript
2.5714286 -3. 2.5241008 -2.7749515 MAarrowhead1
% End of user PostScript
.90182 .09818 m
.89548 .12741 L
s
% Start of user PostScript
2.5714286 -2.5714286 2.5308619 -2.3843251 MAarrowhead1
% End of user PostScript
.90182 .16515 m
.89654 .18845 L
s
% Start of user PostScript
2.5714286 -2.1428571 2.537623 -1.9936987 MAarrowhead1
% End of user PostScript
.90182 .23212 m
.8976 .2495 L
s
% Start of user PostScript
2.5714286 -1.7142857 2.5443841 -1.6030723 MAarrowhead1
% End of user PostScript
.90182 .29909 m
.89865 .31054 L
s
% Start of user PostScript
2.5714286 -1.2857143 2.5511452 -1.2124459 MAarrowhead1
% End of user PostScript
.90182 .36606 m
.89971 .37158 L
s
% Start of user PostScript
2.5714286 -0.85714286 2.5579063 -0.82181949 MAarrowhead1
% End of user PostScript
.90182 .43303 m
.90077 .43262 L
s
% Start of user PostScript
2.5714286 -0.42857143 2.5646675 -0.43119308 MAarrowhead1
% End of user PostScript
.90182 .5 m
.90182 .49366 L
s
% Start of user PostScript
2.5714286 0. 2.5714286 -0.040566683 MAarrowhead1
% End of user PostScript
.90182 .56697 m
.90288 .5547 L
s
% Start of user PostScript
2.5714286 0.42857143 2.5781897 0.35005972 MAarrowhead1
% End of user PostScript
.90182 .63394 m
.90394 .61574 L
s
% Start of user PostScript
2.5714286 0.85714286 2.5849508 0.74068612 MAarrowhead1
% End of user PostScript
.90182 .70091 m
.90499 .67678 L
s
% Start of user PostScript
2.5714286 1.2857143 2.5917119 1.1313125 MAarrowhead1
% End of user PostScript
.90182 .76788 m
.90605 .73783 L
s
% Start of user PostScript
2.5714286 1.7142857 2.598473 1.5219389 MAarrowhead1
% End of user PostScript
.90182 .83485 m
.90711 .79887 L
s
% Start of user PostScript
2.5714286 2.1428571 2.6052341 1.9125653 MAarrowhead1
% End of user PostScript
.90182 .90182 m
.90816 .85991 L
s
% Start of user PostScript
2.5714286 2.5714286 2.6119953 2.3031917 MAarrowhead1
% End of user PostScript
.90182 .96879 m
.90922 .92095 L
s
% Start of user PostScript
2.5714286 3. 2.6187564 2.6938181 MAarrowhead1
% End of user PostScript
.96879 .03121 m
.9614 .08297 L
s
% Start of user PostScript
3. -3. 2.9526722 -2.6687054 MAarrowhead1
% End of user PostScript
.96879 .09818 m
.96246 .14149 L
s
% Start of user PostScript
3. -2.5714286 2.9594333 -2.2942229 MAarrowhead1
% End of user PostScript
.96879 .16515 m
.96351 .20001 L
s
% Start of user PostScript
3. -2.1428571 2.9661944 -1.9197404 MAarrowhead1
% End of user PostScript
.96879 .23212 m
.96457 .25853 L
s
% Start of user PostScript
3. -1.7142857 2.9729555 -1.5452579 MAarrowhead1
% End of user PostScript
.96879 .29909 m
.96563 .31705 L
s
% Start of user PostScript
3. -1.2857143 2.9797167 -1.1707754 MAarrowhead1
% End of user PostScript
.96879 .36606 m
.96668 .37557 L
s
% Start of user PostScript
3. -0.85714286 2.9864778 -0.79629283 MAarrowhead1
% End of user PostScript
.96879 .43303 m
.96774 .43409 L
s
% Start of user PostScript
3. -0.42857143 2.9932389 -0.42181031 MAarrowhead1
% End of user PostScript
.96879 .5 m
.96879 .4926 L
s
% Start of user PostScript
3. 0. 3. -0.047327797 MAarrowhead1
% End of user PostScript
.96879 .56697 m
.96985 .55112 L
s
% Start of user PostScript
3. 0.42857143 3.0067611 0.32715472 MAarrowhead1
% End of user PostScript
.96879 .63394 m
.97091 .60964 L
s
% Start of user PostScript
3. 0.85714286 3.0135222 0.70163724 MAarrowhead1
% End of user PostScript
.96879 .70091 m
.97196 .66816 L
s
% Start of user PostScript
3. 1.2857143 3.0202833 1.0761198 MAarrowhead1
% End of user PostScript
.96879 .76788 m
.97302 .72668 L
s
% Start of user PostScript
3. 1.7142857 3.0270445 1.4506023 MAarrowhead1
% End of user PostScript
.96879 .83485 m
.97408 .7852 L
s
% Start of user PostScript
3. 2.1428571 3.0338056 1.8250848 MAarrowhead1
% End of user PostScript
.96879 .90182 m
.97513 .84372 L
s
% Start of user PostScript
3. 2.5714286 3.0405667 2.1995673 MAarrowhead1
% End of user PostScript
.96879 .96879 m
.97619 .90223 L
s
% Start of user PostScript
3. 3. 3.0473278 2.5740498 MAarrowhead1
% End of user PostScript
.5 Mabswid
.65627 .65626 m
.6707 .63892 L
.68247 .61959 L
.69324 .59603 L
.70208 .56961 L
.70547 .55628 L
.70833 .54205 L
.70937 .53565 L
.71029 .52897 L
.71093 .52328 L
.71146 .51727 L
.71183 .51158 L
.71204 .50647 L
.71213 .50143 L
.71211 .49673 L
.71196 .49139 L
.71169 .48654 L
.71127 .48134 L
.71067 .47592 L
.70989 .47044 L
.7089 .4648 L
.70669 .45507 L
.70373 .44506 L
.69982 .43456 L
.69115 .41668 L
.68122 .4006 L
.62586 .33368 L
.54443 .2438 L
.48531 .18263 L
.45065 .15288 L
.42961 .1385 L
.40916 .12802 L
.39856 .12416 L
.3932 .12266 L
.3873 .12139 L
.37682 .12014 L
.36706 .12022 L
.36112 .12091 L
.35552 .12201 L
.3451 .12529 L
.33932 .12785 L
.33402 .13068 L
.32221 .13871 L
.30469 .15543 L
.28653 .17951 L
.26913 .20997 L
.25451 .24201 L
.23067 .31037 L
.2152 .37054 L
.20907 .40069 L
Mistroke
.20414 .43033 L
.20244 .4427 L
.20099 .45506 L
.19994 .46581 L
.19921 .47521 L
.19868 .48452 L
.1985 .4892 L
.19838 .49401 L
.19832 .49886 L
.19834 .50327 L
.19854 .51105 L
.19891 .51782 L
.19941 .52366 L
.20011 .52964 L
.20101 .53554 L
.20207 .54099 L
.20337 .54643 L
.20612 .55528 L
.20996 .56442 L
.21395 .57173 L
.22157 .58236 L
.23101 .59238 L
.25306 .61041 L
.38528 .71822 L
.49159 .82965 L
.52916 .86579 L
.55131 .88405 L
.57291 .89879 L
.59191 .90861 L
.60275 .91267 L
.61272 .9153 L
.62367 .91687 L
.62988 .91711 L
.63562 .91688 L
.64614 .91528 L
.65725 .91184 L
.66855 .90636 L
.67875 .89956 L
.68974 .89015 L
.70099 .87808 L
.71906 .85313 L
.73727 .82016 L
.75497 .7794 L
.77114 .73312 L
.79294 .65173 L
.80026 .61645 L
.80613 .58221 L
.80993 .55399 L
.81121 .54192 L
.81212 .5314 L
Mistroke
.81273 .52257 L
.81316 .5137 L
.81341 .50457 L
.81342 .49684 L
.81325 .4894 L
.81285 .48191 L
.81229 .4754 L
.81162 .4698 L
.81072 .46404 L
.80962 .45844 L
.80726 .4495 L
.80424 .4412 L
.80045 .43335 L
.79259 .42171 L
.78422 .41272 L
.76636 .39832 L
.74432 .3835 L
.68835 .34477 L
.60278 .26841 L
.54815 .21094 L
.5126 .17334 L
.47519 .13679 L
.44042 .10856 L
.41981 .09572 L
.41049 .09109 L
.40042 .08703 L
.38987 .0839 L
.38 .08211 L
.37394 .08159 L
.36823 .08152 L
.35754 .08257 L
.3463 .08543 L
.33603 .08973 L
.3249 .09634 L
.31335 .1055 L
.29293 .128 L
.27482 .15547 L
.2451 .21874 L
.23124 .25766 L
.22025 .29398 L
.20174 .37097 L
.19603 .40167 L
.19199 .42762 L
.18934 .44871 L
.18746 .46845 L
.18683 .47778 L
.18637 .4873 L
.18617 .49533 L
.18614 .50231 L
.18631 .51 L
Mistroke
.18662 .51647 L
.18717 .52339 L
.18785 .52942 L
.18862 .53463 L
.18959 .53983 L
.19166 .54828 L
.19459 .55698 L
.19768 .56389 L
.20123 .57021 L
.20546 .57631 L
.214 .58591 L
.23283 .6013 L
.25298 .61486 L
.38145 .71598 L
.49477 .83425 L
.53538 .87257 L
.55937 .89134 L
.57149 .89932 L
.58275 .90567 L
.59252 .91027 L
.60327 .91424 L
.61472 .91712 L
.62562 .91843 L
.63132 .91853 L
.63761 .91814 L
.64417 .91715 L
.65031 .91567 L
.6612 .91164 L
.66727 .90857 L
.6728 .90524 L
.69228 .88917 L
.7028 .87746 L
.71205 .86524 L
.7439 .80764 L
.75917 .77005 L
.77342 .72763 L
.79315 .65298 L
.80604 .58578 L
.80996 .55753 L
.81154 .54303 L
.81261 .53083 L
.81328 .52052 L
.81368 .51144 L
.81387 .50257 L
.81387 .49785 L
.81381 .49367 L
.81356 .48646 L
.81317 .4804 L
.81257 .47391 L
.81184 .46827 L
Mistroke
.811 .46314 L
.80997 .45805 L
.80778 .44979 L
.80471 .44128 L
.80153 .43455 L
.79353 .42238 L
.78484 .41297 L
.76731 .39879 L
.74537 .38406 L
.60746 .27307 L
.55562 .21891 L
.52356 .18472 L
.4857 .14652 L
.44886 .1147 L
.42668 .09958 L
.41579 .09358 L
.40538 .08886 L
.39578 .08546 L
.38531 .08289 L
.37575 .08165 L
.36693 .08152 L
.35739 .08256 L
.34694 .08517 L
.33613 .08963 L
.32615 .09545 L
.31592 .10321 L
.30664 .11193 L
.2865 .13685 L
.26998 .16408 L
.25393 .19724 L
.23037 .2603 L
.21077 .33025 L
.19694 .39631 L
.19193 .428 L
.18882 .45338 L
.18789 .46312 L
.18713 .47288 L
.18655 .48284 L
.18623 .49144 L
.18612 .50041 L
.18624 .50822 L
.18651 .51472 L
.18696 .52128 L
.18759 .52745 L
.1883 .53268 L
.1892 .53792 L
.19032 .5432 L
.19266 .55161 L
.19524 .5586 L
.19865 .56579 L
Mistroke
.20269 .57249 L
.21087 .58271 L
.22133 .59251 L
.2322 .60086 L
.25498 .61617 L
.38266 .71716 L
.49374 .8332 L
.53336 .87083 L
.55692 .88959 L
.56891 .89772 L
.58022 .90434 L
.59076 .90951 L
.60254 .91402 L
.6087 .91579 L
.6145 .91708 L
.62063 .91801 L
.62726 .9185 L
.63829 .91807 L
.64987 .9158 L
.65969 .91231 L
.6703 .90681 L
.67966 .90038 L
.68961 .8918 L
.70703 .8721 L
.72303 .84825 L
.73676 .82284 L
.76378 .75719 L
.78549 .68472 L
.79866 .62707 L
.80684 .58051 L
.80997 .55743 L
.81196 .53863 L
.81273 .52914 L
.81334 .51948 L
.8137 .51099 L
.81386 .50364 L
Mfstroke
0 0 m
1 0 L
1 1 L
0 1 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 288},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001/2000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00P@3oool0
0`000000oooo0?ooo`370?ooo`030000003oool0oooo0680oooo00200?ooo`<00000D@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool05@3oool30000
06<0oooo00200?ooo`D00000703oool3000001<0oooo0P00000L0?ooo`8000006`3oool2000001X0
oooo0P00000P0?ooo`<000005`3oool400000680oooo001R0?ooo`030000003oool0oooo01X0oooo
1`00000G0?ooo`L00000503oool3000001X0oooo1000000I0?ooo`@000006@3oool3000001`0oooo
1@00000F0?ooo`H00000H@3oool000d0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo01P0oooo0`00000J0?ooo`H000005@3oool:000001D0oooo1@00
000H0?ooo`D00000603oool5000001P0oooo1@00000G0?ooo`P000005@3oool5000001L0oooo00<0
00000?ooo`000000703oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo000=
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H
0?ooo`D000005`3oool5000001X0oooo1`00000F0?ooo`H000005P3oool6000001L0oooo1P00000H
0?ooo`H000004`3oool:000001@0oooo1000000I0?ooo`<00000703oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`D00000603oool300000200oooo0`00000G0?oo
o`@00000603oool4000001T0oooo1000000J0?ooo`@00000<P3oool3000001X0oooo1@00000J0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo1000000H0?ooo`80
0000?@3oool3000001P0oooo0`00000J0?ooo`<00000703oool300000340oooo0`00000K0?ooo`@0
00006P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`<00000
F@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
703oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01/0oooo1000000J0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool03@3oool000`0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0G0?ooo`80000000<0oooo0000003oool06@3oool300000?l0oooo0@3o
ool4000001/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0=0?ooo`00303oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01L0oooo1@00000H0?ooo`<00000o`3oool2
0?ooo`<000006P3oool3000001`0oooo00<000000?ooo`3oool03@3oool000`0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0G0?ooo`@000006@3oool200000?l0oooo0`3oool20000
01/0oooo1@00000J0?ooo`030000003oool0oooo00d0oooo000<0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool05`3oool4000001T0oooo00<000000?ooo`3oool0o`3oool20?ooo`80
00006`3oool4000001/0oooo00<000000?ooo`3oool03@3oool000`0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0G0?ooo`<00000o`3ooolO0?ooo`030000003oool0oooo01X0oooo
1000000K0?ooo`030000003oool0oooo00d0oooo000<0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool05`3oool300000?l0oooo?03oool3000001`0oooo00<000000?ooo`3oool03@3o
ool000`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`800000o`3ooolm
0?ooo`<000006`3oool00`000000oooo0?ooo`0>0?ooo`00303oool00`000000oooo0?ooo`0G0?oo
o`D00000603oool200000?l0oooo?@3oool2000001`0oooo00<000000?ooo`3oool03P3oool000`0
oooo00<000000?ooo`3oool05`3oool5000001P0oooo00<000000?ooo`3oool0o`3oooll0?ooo`80
00006P3oool300000100oooo000<0?ooo`030000003oool0oooo01L0oooo1000000I0?ooo`030000
003oool0oooo0?l0oooo?03oool00`000000oooo0?ooo`0I0?ooo`D000003P3oool000/0oooo00<0
00000?ooo`3oool0603oool400000?l0ooooF03oool00`000000oooo0?ooo`0I0?ooo`D000003P3o
ool000/0oooo00<000000?ooo`3oool0603oool400000?l0ooooM03oool4000000l0oooo000;0?oo
o`030000003oool0oooo01T0oooo0P00003o0?ooogD0oooo1000000?0?ooo`002`3oool00`000000
oooo0?ooo`0I0?ooo`800000o`3ooomf0?ooo`800000403oool000/0oooo00<000000?ooo`3oool0
6@3oool200000?l0ooooMP3oool200000100oooo00090?ooo`D000006@3oool00`000000oooo0?oo
o`3o0?ooogD0oooo0P00000@0?ooo`002@3oool5000001T0oooo00<000000?ooo`3oool0o`3ooome
0?ooo`030000003oool0oooo00l0oooo00090?ooo`@00000R`3oool@00000?H0oooo00<000000?oo
o`3oool03`3oool000T0oooo100000270?ooo`@000003P3oool500000?l0oooo1P3oool000T0oooo
1000001e0?ooo`030000003oool0oooo00d0oooo0P00000F0?ooo`<00000WP3oool00`000000oooo
0?ooo`1R0?ooo`002P3oool2000007D0oooo0`00000<0?ooo`8000006`3oool3000009/0oooo0P00
001S0?ooo`002P3oool2000007D0oooo1000000:0?ooo`030000003oool0oooo0100oooo0P00000;
0?ooo`<000008P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03h0oooo0P00000F
0?ooo`D00000H@3oool000X0oooo0P00001F0?ooo`030000003oool0oooo01/0oooo1P0000070?oo
o`8000004@3oool4000000d0oooo0`0000030?ooo`800000703oool2000001/0oooo0P00000J0?oo
o`800000803oool4000001H0oooo1P00001P0?ooo`002P3oool00`000000oooo0?ooo`1E0?ooo`<0
00006`3oool5000000L0oooo00<000000?ooo`3oool03P3oool7000000l0oooo0`0000020?ooo`@0
00006@3oool3000001X0oooo0`00000J0?ooo`<000006`3oool7000001D0oooo1P00000G0?ooo`80
0000B03oool000X0oooo00@000000?ooo`3oool00000703oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01P0oooo1@00000H0?ooo`@00000203oool00`000000oooo0?ooo`0=0?ooo`T0
00004@3oool3000000030?ooo`000000000000@000005`3oool4000001T0oooo1000000I0?ooo`D0
00005P3oool:000001@0oooo1000000I0?ooo`D000006@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0H0?ooo`@000006@3oool3000000L0oooo0P00000B0?ooo`L00000
4`3oool:000001D0oooo1P00000G0?ooo`H00000603oool6000001T0oooo1@00000D0?ooo`<00000
6P3oool4000001X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0<0?ooo`003@3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool06@3o
ool4000001P0oooo0`0000070?ooo`030000003oool0oooo01H0oooo0`00000E0?ooo`L000005`3o
ool6000001L0oooo1P00000G0?ooo`D00000<P3oool3000001/0oooo1000000I0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0H0?ooo`<000006P3oool4000001T0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`0a0?ooo`@000006@3oool4000001T0oooo1000000I0?ooo`@00000<P3oool0
0`000000oooo0?ooo`0J0?ooo`@000006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo00d0oooo000=0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0603oool50000
01P0oooo0`00000R0?ooo`030000003oool0oooo03<0oooo0`00000I0?ooo`<000006P3oool30000
01/0oooo0P00001@0?ooo`<000006@3oool5000001X0oooo00<000000?ooo`3oool03@3oool000`0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`@000006@3oool200000280
oooo00<000000?ooo`3oool0=@3oool2000001X0oooo00<000000?ooo`3oool06P3oool00`000000
oooo0?ooo`0K0?ooo`030000003oool0oooo04h0oooo0P00000J0?ooo`@000006`3oool00`000000
oooo0?ooo`0=0?ooo`00303oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo
1000000I0?ooo`8000008@3oool00`000000oooo0?ooo`0h0?ooo`800000Y03oool2000001X0oooo
1000000K0?ooo`030000003oool0oooo00d0oooo000<0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool0603oool3000001X0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`0k0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool06@3oool3000001`0oooo00<0
00000?ooo`3oool03@3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H
0?ooo`<00000>`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo0;`0oooo0`00000L
0?ooo`030000003oool0oooo00d0oooo000<0?ooo`030000003oool0oooo01L0oooo0`00000K0?oo
o`800000>`3oool2000001X0oooo1`00000O0?ooo`030000003oool0oooo0;/0oooo0P00000L0?oo
o`030000003oool0oooo00h0oooo000<0?ooo`030000003oool0oooo01L0oooo1@00000I0?ooo`80
0000>P3oool2000001D0oooo1P0000070?ooo`H000006P3oool200000;/0oooo0P00000J0?ooo`H0
00003@3oool000`0oooo00<000000?ooo`3oool05`3oool5000001T0oooo00<000000?ooo`3oool0
>03oool2000001<0oooo0`00000C0?ooo`<00000603oool200000;X0oooo00<000000?ooo`3oool0
6@3oool5000000h0oooo000<0?ooo`030000003oool0oooo01L0oooo1000000J0?ooo`030000003o
ool0oooo03P0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0G0?ooo`800000603o
ool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01T0oooo1@00000>0?ooo`002`3oool0
0`000000oooo0?ooo`0H0?ooo`@00000E03oool00`000000oooo0?ooo`0@0?ooo`800000703oool3
000001H0oooo00<000000?ooo`3oool0dP3oool4000000l0oooo000;0?ooo`030000003oool0oooo
01T0oooo0P00001D0?ooo`030000003oool0oooo00l0oooo0P00000Q0?ooo`8000005@3oool20000
0=80oooo1000000?0?ooo`002`3oool00`000000oooo0?ooo`0I0?ooo`800000D`3oool2000000l0
oooo0P00000U0?ooo`800000503oool200000=80oooo0P00000@0?ooo`002@3oool3000001/0oooo
0P00001B0?ooo`8000003P3oool2000002T0oooo0P00000C0?ooo`800000d@3oool200000100oooo
00090?ooo`D000006@3oool00`000000oooo0?ooo`1@0?ooo`8000003P3oool00`000000oooo0?oo
o`0[0?ooo`8000004P3oool200000=00oooo00<000000?ooo`3oool03`3oool000T0oooo1@00000I
0?ooo`030000003oool0oooo0500oooo0P00000=0?ooo`030000003oool0oooo02h0oooo0P00000A
0?ooo`800000c`3oool00`000000oooo0?ooo`0?0?ooo`002@3oool4000006`0oooo0P00000=0?oo
o`030000003oool0oooo0340oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`3N0?oo
o`002@3oool4000006`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0c0?ooo`80
00004@3oool00`000000oooo0?ooo`3M0?ooo`002P3oool2000006`0oooo0P00000<0?ooo`030000
003oool0oooo03H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`3L0?ooo`002P3o
ool2000006/0oooo0P00000<0?ooo`030000003oool0oooo03P0oooo0P00000@0?ooo`030000003o
ool0oooo0=/0oooo000:0?ooo`800000J`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
000003d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0
oooo0640oooo000:0?ooo`030000003oool0oooo06T0oooo0P00000:0?ooo`800000?`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo07@0oooo0`00001R0?ooo`002P3oool00`000000
oooo0?ooo`1Y0?ooo`030000003oool0oooo00P0oooo1000000o0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool0L`3oool400000640oooo001R0?ooo`030000003oool0oooo0100oooo
0P00000:0?ooo`D00000703oool200000240oooo0P00000>0?ooo`8000006`3oool00`000000oooo
0?ooo`0n0?ooo`8000005@3oool6000001H0oooo0P0000180?ooo`00HP3oool300000100oooo00<0
00000?ooo`3oool0203oool5000001/0oooo1000000D0?ooo`8000003@3oool00`000000oooo0?oo
o`0;0?ooo`<000006P3oool3000001T0oooo0P00000P0?ooo`D000005@3oool5000001L0oooo1@00
00150?ooo`00HP3oool5000000d0oooo0P00000:0?ooo`@000006@3oool7000001D0oooo0`00000<
0?ooo`030000003oool0oooo00/0oooo0`00000J0?ooo`<000006@3oool4000001X0oooo2000000D
0?ooo`@000006@3oool4000004H0oooo000=0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`@000003P3oool00`000000oooo0?ooo`080?oo
o`@00000603oool:000001D0oooo1@00000:0?ooo`030000003oool0oooo00X0oooo1000000I0?oo
o`@000006@3oool6000001D0oooo2P00000D0?ooo`<000006P3oool4000001X0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01L0oooo0P0000000`3oool000000?ooo`0I0?ooo`@000003@3oool2000000X0
oooo0P00000O0?ooo`D000005P3oool6000000T0oooo00<000000?ooo`3oool02P3oool5000001P0
oooo1@00000H0?ooo`L00000603oool5000001<0oooo0`00000K0?ooo`<000006@3oool3000001`0
oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0G0?ooo`D00000603oool4000000h0oooo00<000000?ooo`3oool0203oool2000003/0oooo
1@00000;0?ooo`030000003oool0oooo00T0oooo1P00000G0?ooo`H000005`3oool600000340oooo
00<000000?ooo`3oool06P3oool3000001X0oooo1@00000J0?ooo`030000003oool0oooo00`0oooo
000=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool05`3oool4000001T0oooo0`00
000>0?ooo`8000002@3oool2000003d0oooo1000000<0?ooo`030000003oool0oooo00T0oooo1@00
000H0?ooo`@000006@3oool4000004l0oooo0`00000J0?ooo`@000006`3oool00`000000oooo0?oo
o`0<0?ooo`003@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo1000000I
0?ooo`8000003`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03h0oooo0P00000>
0?ooo`8000002@3oool300000080oooo00<000000?ooo`3oool05@3oool3000001/0oooo0`00001?
0?ooo`8000006`3oool4000001X0oooo00<000000?ooo`3oool03@3oool000d0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`<000006P3oool2000000h0oooo0P0000090?oo
o`030000003oool0oooo03l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0oooo04h0oooo0P00000K0?ooo`<000006`3oool00`000000oooo0?ooo`0=0?oo
o`00303oool00`000000oooo0?ooo`0H0?ooo`<000006P3oool3000001X0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0580oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo01X0oooo0`00000I0?ooo`<0
00003`3oool000`0oooo00<000000?ooo`3oool0603oool5000001P0oooo0P00000Z0?ooo`800000
203oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
W`3oool2000001X0oooo1@00000=0?ooo`00303oool00`000000oooo0?ooo`0H0?ooo`D00000603o
ool2000002X0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`1E0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0WP3oool2000001X0oooo1@00000=0?ooo`00303oool0
0`000000oooo0?ooo`0H0?ooo`@000006@3oool00`000000oooo0?ooo`0X0?ooo`800000203oool0
0`000000oooo0?ooo`1G0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0W@3oool0
0`000000oooo0?ooo`0I0?ooo`@000003P3oool000`0oooo00<000000?ooo`3oool0603oool40000
01T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
05P0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo
01T0oooo1000000>0?ooo`00303oool00`000000oooo0?ooo`0I0?ooo`800000A03oool2000000P0
oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0;P0
oooo0P00000?0?ooo`002@3oool4000001/0oooo0P0000140?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0;L0oooo
0P00000?0?ooo`002@3oool6000001T0oooo0P0000130?ooo`800000203oool00`000000oooo0?oo
o`1M0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0]P3oool2000000l0oooo0009
0?ooo`H000006@3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0GP3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0;D0oooo00<00000
0?ooo`3oool03P3oool000X0oooo1000000J0?ooo`030000003oool0oooo0480oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0]03oool00`000000oooo0?ooo`0>0?ooo`002P3oool4000005h0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo00`0oooo0P0000360?ooo`00
2P3oool3000005l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1S0?ooo`030000
003oool0oooo00`0oooo0P0000350?ooo`002P3oool3000005h0oooo0P0000070?ooo`030000003o
ool0oooo06@0oooo00<000000?ooo`3oool0303oool200000<@0oooo000;0?ooo`800000GP3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0303oool2
00000<<0oooo000;0?ooo`030000003oool0oooo05`0oooo0P0000070?ooo`030000003oool0oooo
06L0oooo00<000000?ooo`3oool0303oool200000<80oooo000;0?ooo`030000003oool0oooo05`0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo00`0
oooo0P0000310?ooo`00J@3oool2000000L0oooo00<000000?ooo`3oool0JP3oool00`000000oooo
0?ooo`0<0?ooo`800000`03oool006T0oooo0P0000060?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0F`3oool00`000000oooo0?ooo`0<0?ooo`800000G03oool00`000000oooo0?oo
o`1P0?ooo`00J@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00h0oooo0`00001L
0?ooo`030000003oool0oooo00`0oooo0P00001J0?ooo`<00000603oool00`000000oooo0?ooo`16
0?ooo`00HP3oool2000000@0oooo0P0000060?ooo`030000003oool0oooo00h0oooo1@00000M0?oo
o`030000003oool0oooo03`0oooo00<000000?ooo`3oool0303oool2000005T0oooo1000000G0?oo
o`<00000AP3oool00680oooo1@0000000`3oool00000000000060?ooo`030000003oool0oooo00h0
oooo1000000L0?ooo`<00000?`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo05H0
oooo1P00000F0?ooo`D00000A03oool00680oooo100000020?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool03@3oool4000001/0oooo1@00000a0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool02`3oool2000001/0oooo00<000000?ooo`3oool07`3oool5000001D0oooo
1@00000G0?ooo`@000005`3oool2000002`0oooo00140?ooo`<000006`3oool4000000030?ooo`00
0000000000H0oooo00<000000?ooo`3oool03P3oool3000001X0oooo2000000C0?ooo`<000006`3o
ool2000000`0oooo00<000000?ooo`3oool0303oool2000001/0oooo0P00000J0?ooo`X00000503o
ool4000001T0oooo1000000G0?ooo`D00000:@3oool000d0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0G0?ooo`H00000603oool300000080oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0>0?ooo`8000006@3oool:000001@0oooo1P00000G0?ooo`@000002`3oool0
0`000000oooo0?ooo`0;0?ooo`@000006@3oool4000001T0oooo2000000E0?ooo`8000006P3oool4
000001P0oooo1000000K0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool05`3oool5000001T0oooo0`0000020?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool03P3oool2000003T0oooo2000000E0?ooo`D000002P3oool00`00
0000oooo0?ooo`0;0?ooo`D00000603oool5000001X0oooo1@00000D0?ooo`030000003oool0oooo
01X0oooo0`00000I0?ooo`@000006`3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo01L0oooo1@00000I0?ooo`8000000P3oool2000000H0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0i0?ooo`H000005`3oool6000000T0
oooo00<000000?ooo`3oool02P3oool6000001P0oooo1P00000K0?ooo`800000<@3oool2000001X0
oooo0`00000I0?ooo`@000003P3oool000d0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0H0?ooo`@000006P3oool200000080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`1=0?ooo`@00000603oool4000000/0oooo00<000000?ooo`3oool02P3oool4000001X0oooo
1000001?0?ooo`8000006P3oool3000001T0oooo1P00000<0?ooo`003@3oool00`000000oooo0?oo
o`0G0?ooo`D000006@3oool3000001X0oooo00<000000?ooo`3oool00P0000060?ooo`030000003o
ool0oooo04h0oooo0P00000I0?ooo`<000003@3oool00`000000oooo0?ooo`090?ooo`<0000000<0
oooo0000003oool0603oool300000500oooo00<000000?ooo`3oool06@3oool2000001X0oooo1@00
000=0?ooo`00303oool00`000000oooo0?ooo`0H0?ooo`D000006@3oool2000001/0oooo00<00000
0?ooo`3oool00P0000060?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00T0oooo00D000000?ooo`3oool0oooo0000
000I0?ooo`030000003oool0oooo06/0oooo0P00000J0?ooo`D000003@3oool000`0oooo00<00000
0?ooo`3oool0603oool4000001X0oooo0P00000N0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0O03oool00`000000oooo0?ooo`0<0?ooo`800000QP3oool00`000000oooo0?ooo`0I
0?ooo`@000003P3oool000`0oooo00<000000?ooo`3oool0603oool4000001X0oooo00<000000?oo
o`3oool0703oool2000000H0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`0<0?oo
o`800000Q@3oool00`000000oooo0?ooo`0J0?ooo`<000003P3oool000`0oooo00<000000?ooo`3o
ool0603oool3000001/0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo07h0oooo00<000000?ooo`3oool0303oool200000:40oooo0P00000?0?ooo`00
2P3oool5000001T0oooo0P00000j0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
P03oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo09l0oooo0P00000?0?ooo`002P3o
ool5000001T0oooo0P00000i0?ooo`8000001P3oool00`000000oooo0?ooo`210?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`0>0?ooo`002P3oool4
000001X0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo0880oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`2M0?ooo`030000003oool0
oooo00h0oooo000:0?ooo`@000006P3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo
00@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
0:d0oooo000:0?ooo`<00000E@3oool2000000H0oooo00<000000?ooo`3oool0Q03oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo0:`0oooo000;0?ooo`800000E@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`2[0?ooo`002`3oool2000005D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`270?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0ZP3oool000/0oooo00<00000
0?ooo`3oool0D`3oool2000000H0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo0:T0oooo000;0?ooo`030000003oool0oooo05<0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0Z03oool00640oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2;0?ooo`03
0000003oool0oooo00d0oooo00<000000?ooo`3oool0Y`3oool00600oooo0P0000060?ooo`030000
003oool0oooo08`0oooo00<000000?ooo`3oool03@3oool200000:L0oooo001P0?ooo`8000001@3o
ool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0Y03o
ool00600oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0X`3oool005l0oooo0P0000060?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
05X0oooo00<000000?ooo`3oool0A@3oool005l0oooo0P0000050?ooo`030000003oool0oooo01T0
oooo0`00001f0?ooo`030000003oool0oooo00h0oooo0P0000110?ooo`030000003oool0oooo01H0
oooo0`0000150?ooo`00G`3oool010000000oooo0?ooo`3oool2000000030?ooo`000000oooo01T0
oooo1@00000M0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool03`3oool2000003l0
oooo0`00000G0?ooo`D000005P3oool00`000000oooo0?ooo`0Z0?ooo`00GP3oool2000000<0oooo
1@00000I0?ooo`@00000703oool3000005P0oooo00<000000?ooo`3oool03`3oool2000003d0oooo
1000000F0?ooo`D000005`3oool3000002X0oooo001N0?ooo`8000000`3oool4000001X0oooo0`00
000L0?ooo`D00000F03oool00`000000oooo0?ooo`0?0?ooo`800000>`3oool6000001D0oooo1000
000H0?ooo`D00000:03oool004D0oooo0P00000G0?ooo`030000003oool0oooo0080oooo1000000I
0?ooo`<00000703oool600000340oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00h0oooo1000000h0?ooo`P00000503oool3000001T0oooo1000000H
0?ooo`8000003`3oool000d0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?oo
o`D00000503oool00`000000oooo0?ooo`020?ooo`<000006P3oool3000001X0oooo2@00000C0?oo
o`8000006`3oool3000001X0oooo0P00000:0?ooo`030000003oool0oooo00h0oooo1000000H0?oo
o`D000006@3oool6000001L0oooo0`00000I0?ooo`@00000603oool5000000`0oooo000=0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool0603oool4000001@0oooo0P0000030?ooo`<0
00006`3oool2000001X0oooo1@00000I0?ooo`D00000603oool3000001T0oooo0`00000:0?ooo`03
0000003oool0oooo00d0oooo1@00000H0?ooo`T000004`3oool3000001X0oooo0`00000J0?ooo`<0
00006@3oool5000000`0oooo000=0?ooo`030000003oool0oooo01L0oooo1000000J0?ooo`@00000
503oool2000000<0oooo0`00000J0?ooo`800000>P3oool7000001D0oooo1000000I0?ooo`@00000
2@3oool00`000000oooo0?ooo`0=0?ooo`D000006@3oool700000300oooo0P00000K0?ooo`<00000
6@3oool4000000d0oooo000=0?ooo`030000003oool0oooo01L0oooo1P00000H0?ooo`<000005@3o
ool00`000000oooo0?ooo`020?ooo`<000006P3oool00`000000oooo0?ooo`0j0?ooo`D000005`3o
ool5000001L0oooo1P0000080?ooo`030000003oool0oooo00`0oooo1`00000H0?ooo`D00000<@3o
ool00`000000oooo0?ooo`0J0?ooo`8000006P3oool4000000d0oooo000=0?ooo`030000003oool0
oooo01L0oooo1@00000I0?ooo`<000005@3oool00`000000oooo0?ooo`020?ooo`800000F@3oool4
000001L0oooo1P00000G0?ooo`H00000203oool00`000000oooo0?ooo`0<0?ooo`D0000000<0oooo
0000003oool0603oool3000004h0oooo0P00000K0?ooo`8000003P3oool000d0oooo00<000000?oo
o`3oool05`3oool5000001T0oooo0P00000E0?ooo`800000103oool2000005X0oooo0P00000I0?oo
o`@000006@3oool4000000X0oooo00<000000?ooo`3oool0303oool3000000040?ooo`000000oooo
000001X0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0J0?ooo`8000003P3oool0
00X0oooo1P00000G0?ooo`@000006P3oool2000001D0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`1I0?ooo`030000003oool0oooo01L0oooo0`00000J0?ooo`<00000303oool00`00
0000oooo0?ooo`0;0?ooo`<000000P3oool00`000000oooo000000250?ooo`8000003P3oool000X0
oooo1P00000H0?ooo`<000006P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0M@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`020?ooo`040000003oool000000000
08<0oooo00<000000?ooo`3oool03@3oool000X0oooo1@00000I0?ooo`8000006`3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0XP3oool00`000000
oooo0?ooo`0@0?ooo`80000000<0oooo0000003oool0P@3oool00`000000oooo0?ooo`0=0?ooo`00
2`3oool4000001T0oooo0P00000a0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
XP3oool00`000000oooo0?ooo`0B0?ooo`030000003oool000000940oooo000;0?ooo`@000006@3o
ool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0X`3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool000000900oooo000;0?ooo`<000006P3oool0
0`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0Y@3oool0
0`000000oooo0?ooo`0B0?ooo`040000003oool00000000008h0oooo000;0?ooo`<00000C@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool04P3oool0
10000000oooo0?ooo`00002=0?ooo`00303oool2000004`0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0180oooo00@000000?ooo`3oool00000S03o
ool000`0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo0:P0oooo00<000000?ooo`3oool04P3oool010000000oooo0?ooo`00002;0?ooo`00303oool0
0`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0ZP3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00800000R@3oool005X0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo0180oooo0P0000020?oo
o`030000003oool0oooo08H0oooo001I0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0[03oool00`000000oooo0?ooo`0C0?ooo`040000003oool0oooo000008L0oooo001I0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`0D0?ooo`04
0000003oool0oooo000008H0oooo001I0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0[P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00800000Q03oool005T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo01@0oooo
00D000000?ooo`3oool0oooo000000230?ooo`00F03oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo0;00oooo00<000000?ooo`3oool0503oool200000080oooo00<000000?ooo`3oool0
P03oool005P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2b0?ooo`030000003o
ool0oooo01D0oooo00@000000?ooo`3oool00000P@3oool005P0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`0P0?ooo`8000007`3oool00`000000oooo0?ooo`1_0?ooo`030000003o
ool0oooo01D0oooo00<000000?ooo`3oool00P00000j0?ooo`030000003oool0oooo01H0oooo0P00
000Z0?ooo`00F03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P00000L
0?ooo`D000006`3oool300000700oooo00<000000?ooo`3oool05P3oool200000080oooo00<00000
0?ooo`3oool0=P3oool4000001H0oooo1@00000G0?ooo`8000003P3oool005L0oooo0P0000050?oo
o`030000003oool0oooo0080oooo1@00000H0?ooo`D000006`3oool400000740oooo00<000000?oo
o`3oool05`3oool010000000oooo0?ooo`00000g0?ooo`D000005@3oool4000001P0oooo1@00000;
0?ooo`00A@3oool200000100oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?oo
o`@000006@3oool4000001/0oooo1P00001a0?ooo`030000003oool0oooo01L0oooo0P0000000`3o
ool000000?ooo`0d0?ooo`D000005P3oool4000001P0oooo1000000<0?ooo`00A@3oool5000000d0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`@000006@3oool3000001`0
oooo1`00001a0?ooo`030000003oool0oooo01P0oooo00@000000?ooo`0000000000<P3oool40000
01P0oooo0`00000I0?ooo`@00000303oool002P0oooo0P00000K0?ooo`D000003@3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo0`00000J0?ooo`<000006`3oool600000380
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0000
01P0oooo1000000D0?ooo`@00000603oool3000001X0oooo0`00000=0?ooo`003@3oool00`000000
oooo0?ooo`0H0?ooo`D00000603oool4000000d0oooo0P0000050?ooo`030000003oool0oooo00<0
oooo0`00000I0?ooo`<000006`3oool4000001P0oooo1@00000I0?ooo`8000006P3oool2000001/0
oooo0P0000070?ooo`030000003oool0oooo0180oooo0P0000040?ooo`80000000<0oooo0000003o
ool04@3oool9000001@0oooo0P00000J0?ooo`<000006P3oool3000000d0oooo000=0?ooo`030000
003oool0oooo01P0oooo1@00000H0?ooo`@000003@3oool2000000@0oooo00<000000?ooo`3oool0
103oool2000001X0oooo0P00000K0?ooo`800000703oool9000001@0oooo0`00000J0?ooo`800000
6`3oool2000000H0oooo00<000000?ooo`3oool04P3oool3000000D0oooo00<000000?ooo`000000
4`3oool7000001<0oooo00<000000?ooo`3oool06P3oool2000001/0oooo0P00000>0?ooo`002`3o
ool5000001P0oooo1000000J0?ooo`8000003P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00@0oooo0P00000J0?ooo`030000003oool0oooo03X0oooo1`00000E0?ooo`@00000603o
ool4000001T0oooo100000050?ooo`030000003oool0oooo0180oooo100000040?ooo`@000004`3o
ool500000300oooo0P00000K0?ooo`8000003P3oool000/0oooo1@00000H0?ooo`@000006P3oool2
000000h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo05L0oooo1@00000F0?ooo`H000005P3oool5000001P0oooo1@0000050?ooo`030000003oool0
oooo0140oooo1P0000040?ooo`030000003oool0000001D0oooo0P00000`0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool03@3oool000/0oooo1000000J0?ooo`8000006`3oool20000
00h0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
05T0oooo0`00000G0?ooo`H000005P3oool5000001P0oooo1@0000050?ooo`030000003oool0oooo
0140oooo1P0000040?ooo`<00000H`3oool00`000000oooo0?ooo`0=0?ooo`002`3oool4000001X0
oooo0P00000K0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`1R0?ooo`030000003oool0oooo01H0oooo1000000G0?ooo`H000005`3oool6000000D0
oooo00<000000?ooo`3oool04@3oool4000000L0oooo0P00001b0?ooo`002`3oool4000001X0oooo
0P00000K0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`1l0?ooo`<000006@3oool4000001T0oooo100000070?ooo`030000003oool0oooo0100oooo
0`0000090?ooo`<00000L03oool000`0oooo0P00000K0?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo01P0oooo0P00
000K0?ooo`8000002@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00T0oooo0`00
001_0?ooo`00303oool2000001/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0<40oooo00<000000?ooo`3oool07P3oool2000006h0oooo000<0?oo
o`800000A`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0<80oooo00<000000?oo
o`3oool07P3oool3000006`0oooo000<0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`330?ooo`030000003oool0oooo01l0oooo0P00001[0?ooo`00
303oool00`000000oooo0?ooo`150?ooo`8000001@3oool00`000000oooo0?ooo`340?ooo`030000
003oool0oooo0200oooo0P00001Y0?ooo`00E03oool2000000@0oooo00<000000?ooo`3oool0aP3o
ool00`000000oooo0?ooo`0P0?ooo`800000J03oool005@0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`360?ooo`030000003oool0oooo0280oooo0P00001V0?ooo`00E03oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool08`3oool00`00
0000oooo0?ooo`1S0?ooo`00E03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0<P0
oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1R0?ooo`00E03oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool08`3oool200000680oooo
001C0?ooo`8000001@3oool00`000000oooo0?ooo`390?ooo`030000003oool0oooo02@0oooo0P00
001Q0?ooo`00D`3oool2000000@0oooo00<000000?ooo`3oool0b`3oool00`000000oooo0?ooo`0U
0?ooo`030000003oool0oooo05h0oooo001C0?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0b`3oool00`000000oooo0?ooo`0V0?ooo`800000GP3oool005<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo02H0oooo0P00001M0?ooo`00
D`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0
QP3oool00`000000oooo0?ooo`0W0?ooo`800000G03oool005<0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`0U0?ooo`8000007@3oool2000008P0oooo00<000000?ooo`3oool09`3o
ool200000500oooo00<000000?ooo`3oool0203oool005<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`070?ooo`<000006`3oool5000001X0oooo100000260?ooo`030000003oool0
oooo02P0oooo00<000000?ooo`3oool0C@3oool3000000T0oooo00150?ooo`8000002`3oool20000
00@0oooo00<000000?ooo`3oool01`3oool6000001P0oooo1000000J0?ooo`H00000QP3oool00`00
0000oooo0?ooo`0X0?ooo`030000003oool0oooo04`0oooo100000080?ooo`00A@3oool5000000P0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`D000006@3oool4000001X0
oooo1@0000270?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0BP3oool6000000L0
oooo000X0?ooo`D00000603oool5000000P0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`080?ooo`D000006@3oool3000001/0oooo1000000O0?ooo`800000J03oool00`000000oooo
0?ooo`0X0?ooo`800000B`3oool5000000P0oooo000;0?ooo`8000006`3oool5000001P0oooo1000
00090?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool4000001X0oooo0`00
000K0?ooo`<000007P3oool4000006P0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?oo
o`180?ooo`@000002P3oool000/0oooo1@00000H0?ooo`@000006@3oool4000000T0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`<000006P3oool2000001/0oooo0`00000L
0?ooo`L00000<03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0P00000L
0?ooo`030000003oool0oooo00X0oooo0P00000>0?ooo`030000003oool0oooo03P0oooo0`00000;
0?ooo`002`3oool5000001P0oooo1000000J0?ooo`8000002P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo00T0oooo0P00000K0?ooo`8000006`3oool2000001/0oooo2@00000D0?oo
o`030000003oool0oooo01T0oooo0P00000K0?ooo`8000006P3oool3000001/0oooo0P00000<0?oo
o`030000003oool0oooo00`0oooo0P00000K0?ooo`@000006@3oool3000000`0oooo000;0?ooo`@0
00006@3oool3000001/0oooo0P00000:0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool02@3oool2000001/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0N0?ooo`D0
00005@3oool2000001T0oooo0`00000J0?ooo`<000006P3oool2000001/0oooo0`00000;0?ooo`80
00003P3oool2000001/0oooo2000000D0?ooo`030000003oool0oooo00`0oooo000;0?ooo`@00000
6P3oool2000001/0oooo0P00000:0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo05D0oooo1000000H0?ooo`<00000
6P3oool3000001T0oooo0`00000K0?ooo`<000002`3oool00`000000oooo0?ooo`0<0?ooo`@00000
6`3oool6000002<0oooo000<0?ooo`8000006`3oool2000001/0oooo00<000000?ooo`3oool0203o
ool2000000@0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1c0?ooo`D000005P3o
ool3000001X0oooo0`00000I0?ooo`@000006P3oool3000000/0oooo0P00000=0?ooo`D000006`3o
ool4000002@0oooo000<0?ooo`8000006`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1o0?ooo`H000005@3oool4
000001T0oooo1000000I0?ooo`@000006@3oool4000000/0oooo00<000000?ooo`3oool0303oool5
000001/0oooo0`00000T0?ooo`00303oool2000001/0oooo00<000000?ooo`3oool09@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0800oooo1000000F0?ooo`D00000603oool50000
01P0oooo1000000I0?ooo`D000002P3oool2000000d0oooo1P00000L0?ooo`030000003oool0oooo
0280oooo000<0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`200?ooo`<00000603oool5000001P0oooo1@00000G0?ooo`D000006@3oool5000000X0
oooo00<000000?ooo`3oool02`3oool500000480oooo000<0?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo01L0oooo
1000000I0?ooo`@00000603oool4000001X0oooo1000000;0?ooo`8000003@3oool3000004<0oooo
001A0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0W03oool2000001/0oooo0P00
000J0?ooo`<000006`3oool2000000d0oooo0P00000=0?ooo`030000003oool0oooo04<0oooo001A
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`0X
0?ooo`800000D`3oool00540oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3G0?oo
o`030000003oool0oooo02T0oooo0P00001B0?ooo`00D03oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0=L0oooo00<000000?ooo`3oool0:@3oool200000580oooo001@0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`0Y0?ooo`800000
DP3oool00500oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3H0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0D03oool00500oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`3H0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0D03oool0
0500oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`3H0?ooo`030000003oool0oooo
02T0oooo00<000000?ooo`3oool0D03oool00500oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`3I0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0D03oool00500oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3I0?ooo`030000003oool0oooo02T0oooo
0P00001A0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0=X0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`00D03oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0=X0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1?
0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0=X0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo04<0oooo1@00000H0?ooo`D00000M@3oool00`000000oooo0?ooo`0Y0?ooo`03
0000003oool0oooo04l0oooo001@0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
2@3oool6000001L0oooo1P00000G0?ooo`D00000603oool5000007D0oooo00<000000?ooo`3oool0
:@3oool00`000000oooo0?ooo`1?0?ooo`00:03oool5000001L0oooo1P0000060?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool02@3oool6000001L0oooo1P00000G0?ooo`D00000603o
ool4000007H0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`002`3oool5
000001P0oooo1@00000G0?ooo`H000001P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00X0oooo1@00000G0?ooo`D000006@3oool3000001X0oooo0`00001f0?ooo`030000003oool0
oooo02T0oooo00<000000?ooo`3oool0C`3oool000/0oooo1@00000H0?ooo`D00000603oool40000
00L0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0:0?ooo`@000006@3oool40000
01T0oooo0`00000J0?ooo`<00000MP3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo
04l0oooo000;0?ooo`@000006P3oool3000001T0oooo100000070?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool02P3oool4000001T0oooo0`00000J0?ooo`<000006P3oool3000007H0
oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool0303oool000`0oooo0`00000J0?ooo`<000006@3oool4000000L0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0;0?ooo`<000006@3oool3000001X0oooo
0`00000J0?ooo`800000M`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0303oool000`0oooo0`00000J0?ooo`<000006P3oool3000000L0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0;0?ooo`<000006@3oool3000001/0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo01X0oooo00D0
00000?ooo`3oool0oooo0000000G0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0P00000>0?ooo`00303o
ool3000001X0oooo0`00000J0?ooo`800000203oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00/0oooo0P00000K0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`050000003oool0oooo0?ooo`0000005`3oool200000180oooo
00<000000?ooo`3oool01@3oool2000001/0oooo0`00000J0?ooo`<000003@3oool000`0oooo0P00
000L0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0I0?ooo`8000000`3oool00`000000oooo0?ooo`0D0?ooo`<000004P3oool00`000000oooo0?oo
o`050?ooo`<000006P3oool3000001X0oooo0`00000=0?ooo`003@3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0S03oool2000001/0oooo0`00000J
0?ooo`<000000P3oool00`000000oooo0?ooo`0D0?ooo`<000004P3oool00`000000oooo0?ooo`05
0?ooo`<000006P3oool3000001X0oooo0`00000=0?ooo`003@3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo02<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2K0?oo
o`<000006P3oool3000001T0oooo100000020?ooo`030000003oool0oooo01@0oooo1000000A0?oo
o`030000003oool0oooo00D0oooo0`00000J0?ooo`<000006@3oool4000000d0oooo001@0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool0V`3oool3000001X0oooo0`00000I0?ooo`@0
00000P3oool00`000000oooo0?ooo`0D0?ooo`@000004@3oool00`000000oooo0?ooo`040?ooo`@0
00006@3oool5000001P0oooo1@00000<0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo09/0oooo0`00000J0?ooo`<000006@3oool400000080oooo00<000000?ooo`3oool0
4`3oool500000140oooo00<000000?ooo`3oool0103oool5000001P0oooo1@00000H0?ooo`D00000
303oool00500oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2J0?ooo`@000006@3o
ool5000001L0oooo1P0000000`3oool000000?ooo`0D0?ooo`H00000403oool00`000000oooo0?oo
o`040?ooo`D00000603oool5000002T0oooo001@0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0VP3oool5000001P0oooo1@00000G0?ooo`H0000000<0oooo0000003oool0503oool6
00000100oooo00<000000?ooo`3oool0C`3oool00500oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`2J0?ooo`D00000603oool5000001h0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`1?0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0=X0
oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`00D03oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0=X0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`1?0?ooo`00D03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0=X0oooo00<0
00000?ooo`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`00D03oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo0=T0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1?
0?ooo`00D03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0=T0oooo00<000000?oo
o`3oool0:@3oool00`000000oooo0?ooo`1?0?ooo`00D03oool2000000@0oooo00<000000?ooo`3o
ool0f03oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo04l0oooo001A0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool0f03oool00`000000oooo0?ooo`0Y0?ooo`800000
D@3oool00540oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3H0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0D03oool00540oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`3G0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0D03oool0
0540oooo0P0000040?ooo`030000003oool0oooo0=L0oooo00<000000?ooo`3oool0:@3oool00`00
0000oooo0?ooo`1@0?ooo`00D@3oool2000000@0oooo00<000000?ooo`3oool0e`3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo0500oooo001B0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool0e`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0500oooo
00140?ooo`030000003oool0oooo00/0oooo0P0000030?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool06`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01X0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo02X0oooo00<0
00000?ooo`3oool0D03oool004<0oooo0P00000=0?ooo`8000000`3oool00`000000oooo0?ooo`06
0?ooo`<000006`3oool3000001T0oooo0`00000J0?ooo`<00000603oool2000005`0oooo00<00000
0?ooo`3oool0:P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo00/0oooo00110?oo
o`D00000303oool2000000@0oooo00<000000?ooo`3oool00`3oool5000001T0oooo1@00000G0?oo
o`D00000603oool5000001L0oooo1000001K0?ooo`030000003oool0oooo02X0oooo00<000000?oo
o`3oool0@P3oool00`000000oooo0?ooo`0;0?ooo`008`3oool00`000000oooo0?ooo`0J0?ooo`H0
00003@3oool2000000<0oooo00<000000?ooo`3oool0103oool5000001T0oooo1000000H0?ooo`D0
0000603oool5000001D0oooo1@00001K0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3o
ool09@3oool00`000000oooo0?ooo`0I0?ooo`8000003@3oool00280oooo1000000K0?ooo`D00000
3@3oool2000000<0oooo00<000000?ooo`3oool01@3oool4000001T0oooo1000000I0?ooo`@00000
6@3oool4000001H0oooo1@00001J0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0P00000=0?ooo`008P3o
ool5000001/0oooo1@00000=0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3o
ool3000001X0oooo1000000I0?ooo`<000006P3oool3000001P0oooo0`00001I0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool03P3oool2000000X0oooo00<000000?ooo`3oool06@3o
ool2000001X0oooo0`00000=0?ooo`008@3oool7000001/0oooo1000000=0?ooo`8000000`3oool0
0`000000oooo0?ooo`050?ooo`<000006`3oool3000001T0oooo0`00000J0?ooo`<000006@3oool3
000005L0oooo0P00000J0?ooo`800000403oool2000000T0oooo0P00000K0?ooo`8000006P3oool4
000000`0oooo000=0?ooo`030000003oool0oooo0140oooo2@00000K0?ooo`8000003P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo0`00000K0?ooo`8000006P3oool30000
01X0oooo0`00000I0?ooo`800000503oool500000240oooo00<000000?ooo`3oool06@3oool30000
01X0oooo0P00000@0?ooo`030000003oool0oooo00P0oooo0P00000J0?ooo`<000006P3oool40000
00`0oooo000;0?ooo`<00000603oool5000001/0oooo0P00000=0?ooo`800000103oool00`000000
oooo0?ooo`050?ooo`8000006`3oool2000001/0oooo0P00000K0?ooo`8000006`3oool00`000000
oooo0?ooo`0A0?ooo`X000006`3oool2000001/0oooo0P00000J0?ooo`<00000403oool00`000000
oooo0?ooo`070?ooo`<000006P3oool4000001P0oooo1@00000<0?ooo`002P3oool3000003X0oooo
00<000000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0;@3oool8000001`0oooo0`00000J0?ooo`<000006P3oool300000100oooo
00<000000?ooo`3oool01`3oool4000001T0oooo1000000H0?ooo`D00000303oool000T0oooo0`00
001;0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0S@3oool5000001d0oooo0`00
000K0?ooo`<000006@3oool5000000l0oooo00<000000?ooo`3oool01`3oool4000001P0oooo1@00
000J0?ooo`<00000303oool000L0oooo1@00001;0?ooo`800000103oool00`000000oooo0?ooo`2<
0?ooo`8000007`3oool4000001X0oooo1000000I0?ooo`D000003P3oool2000000P0oooo1@00000H
0?ooo`D00000:@3oool000H0oooo1@00001=0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0[03oool5000001X0oooo1000000H0?ooo`H000003P3oool2000000P0oooo1@0000160?oo
o`001`3oool4000004h0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2Z0?ooo`H0
00006@3oool5000001P0oooo1P00000>0?ooo`030000003oool0oooo00T0oooo0`0000160?ooo`00
203oool2000004l0oooo0P0000040?ooo`030000003oool0oooo0:X0oooo1000000J0?ooo`D00000
6`3oool3000000h0oooo00<000000?ooo`3oool0DP3oool000T0oooo00<000000?ooo`3oool0CP3o
ool2000000@0oooo00<000000?ooo`3oool0ZP3oool3000001`0oooo0`00000/0?ooo`030000003o
ool0oooo0580oooo001K0?ooo`800000103oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1B0?ooo`00G03oool20000
00<0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0/0?ooo`800000E03oool005h0
oooo00D000000?ooo`3oool0oooo000000380?ooo`030000003oool0oooo02`0oooo0P00001D0?oo
o`00G`3oool01@000000oooo0?ooo`3oool000000<H0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`1C0?ooo`00H03oool200000080oooo00<000000?ooo`3oool0``3oool00`000000
oooo0?ooo`0]0?ooo`030000003oool0oooo05<0oooo001R0?ooo`80000000<0oooo000000000000
``3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo05<0oooo001S0?ooo`8000000P3o
ool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo02d0oooo0P00001E0?ooo`00I03oool2
00000080oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`0]0?ooo`800000E@3oool0
06D0oooo0`0000000`3oool000000?ooo`2n0?ooo`030000003oool0oooo02d0oooo00<000000?oo
o`3oool0E03oool006L0oooo0P0000000`3oool000000000002l0?ooo`030000003oool0oooo02h0
oooo00<000000?ooo`3oool0E03oool006P0oooo0P0000020?ooo`030000003oool0oooo0;T0oooo
00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo00/0oooo
001Z0?ooo`80000000<0oooo0000003oool0^03oool00`000000oooo0?ooo`0^0?ooo`800000B03o
ool00`000000oooo0?ooo`0;0?ooo`00J`3oool2000000030?ooo`000000oooo0;L0oooo00<00000
0?ooo`3oool0;P3oool2000004L0oooo0P00000=0?ooo`00K@3oool010000000oooo00000000002f
0?ooo`030000003oool0oooo02h0oooo0P00000[0?ooo`030000003oool0oooo01T0oooo0P00000=
0?ooo`00KP3oool010000000oooo0?ooo`00000:0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo04l0oooo00<00000
0?ooo`3oool0;`3oool2000002/0oooo00<000000?ooo`3oool06@3oool2000000d0oooo001_0?oo
o`80000000<0oooo0000003oool01`3oool3000001/0oooo0`00000J0?ooo`<000006P3oool20000
0540oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
01T0oooo0P00000J0?ooo`@00000303oool00740oooo00<000000?ooo`0000001P3oool4000001T0
oooo1P00000G0?ooo`H000005`3oool5000001L0oooo00<000000?ooo`3oool0=@3oool00`000000
oooo0?ooo`0_0?ooo`8000003`3oool00`000000oooo0?ooo`0I0?ooo`8000006P3oool4000000`0
oooo000>0?ooo`030000003oool0oooo0640oooo00@000000?ooo`0000000000103oool5000001T0
oooo1@00000H0?ooo`D000005P3oool6000001L0oooo0`00000e0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool02@3oool2000000h0oooo0P00000J0?ooo`<000006P3oool4000000`0
oooo000=0?ooo`800000703oool00`000000oooo0?ooo`0^0?ooo`8000005@3oool2000000040?oo
o`000000oooo0?ooo`L000006@3oool4000001T0oooo1000000G0?ooo`H000005@3oool5000003<0
oooo00<000000?ooo`3oool08`3oool2000000/0oooo0P00000>0?ooo`8000006P3oool4000001P0
oooo1@00000<0?ooo`003@3oool2000001/0oooo0P00000`0?ooo`@00000503oool2000000030?oo
o`000000oooo0080oooo1P00000I0?ooo`@000006@3oool4000001P0oooo1000000E0?ooo`L00000
<03oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01P0oooo0P00000;0?ooo`800000
3@3oool3000001X0oooo1000000H0?ooo`D00000303oool000`0oooo0`00000K0?ooo`800000703o
ool00`000000oooo0?ooo`0A0?ooo`L000004`3oool00`000000oooo000000040?ooo`@000006`3o
ool2000001/0oooo0P00000I0?ooo`@00000503oool9000001`0oooo0P00000@0?ooo`030000003o
ool0oooo00P0oooo0P00000I0?ooo`<000002`3oool2000000d0oooo1000000H0?ooo`D000006P3o
ool00`000000oooo0?ooo`0<0?ooo`00303oool3000001X0oooo0`00000J0?ooo`800000503oool9
00000180oooo00@000000?ooo`00000000001@3oool2000001/0oooo0P00000K0?ooo`8000006P3o
ool2000001T0oooo1@00000H0?ooo`@00000403oool00`000000oooo0?ooo`080?ooo`<000006@3o
ool3000000X0oooo0P00000>0?ooo`@00000603oool5000001X0oooo00<000000?ooo`3oool0303o
ool000/0oooo1000000I0?ooo`@000006@3oool3000001@0oooo1@00000G0?ooo`80000000<0oooo
0000003oool08@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool0<03oool600000140oooo00<000000?ooo`3oool01`3oool3000001T0oooo1000000:
0?ooo`8000003@3oool5000001X0oooo0`00000Y0?ooo`002`3oool4000001T0oooo0`00000H0?oo
o`@00000<P3oool2000000030?ooo`000000oooo08/0oooo1`00000A0?ooo`030000003oool0oooo
00P0oooo0`00000I0?ooo`@000002P3oool2000000d0oooo1@0000160?ooo`002P3oool5000001P0
oooo1000000F0?ooo`D00000=@3oool00`000000oooo0000002<0?ooo`H00000403oool00`000000
oooo0?ooo`080?ooo`@00000603oool5000000X0oooo0P00000?0?ooo`<00000AP3oool000X0oooo
1@00000H0?ooo`@000005@3oool6000003H0oooo00<000000?ooo`000000R`3oool500000100oooo
00<000000?ooo`3oool0203oool5000001P0oooo1@00000:0?ooo`030000003oool0oooo05L0oooo
000<0?ooo`<000005`3oool5000001H0oooo1000000h0?ooo`@00000RP3oool300000100oooo00<0
00000?ooo`3oool0203oool5000001/0oooo0`0000090?ooo`800000F@3oool002P0oooo0`00000G
0?ooo`<00000>@3oool2000000030?ooo`000000oooo08T0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0:0?ooo`@000009`3oool2000005T0oooo00130?ooo`030000003oool0oooo
03X0oooo00<000000?ooo`000000V@3oool00`000000oooo0?ooo`0=0?ooo`8000009`3oool20000
05T0oooo00210?ooo`<00000U`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo05P0
oooo00220?ooo`@00000U03oool00`000000oooo0?ooo`0g0?ooo`800000FP3oool008<0oooo0P00
00000`3oool000000?ooo`3<0?ooo`800000FP3oool008D0oooo00<000000?ooo`000000c03oool2
000005X0oooo00260?ooo`<00000b`3oool00`000000oooo0?ooo`1I0?ooo`00Q`3oool400000<P0
oooo0P00001K0?ooo`00R03oool2000000030?ooo`000000oooo0<H0oooo0P00001=0?ooo`030000
003oool0oooo00/0oooo002:0?ooo`<00000aP3oool2000004d0oooo00<000000?ooo`3oool02`3o
ool008/0oooo0`0000350?ooo`030000003oool0oooo04/0oooo0P00000=0?ooo`00S03oool40000
0<80oooo0P00001=0?ooo`<00000303oool008d0oooo0P0000000`3oool000000?ooo`300?ooo`80
0000<P3oool00`000000oooo0?ooo`0H0?ooo`<00000303oool008l0oooo0`0000300?ooo`800000
<P3oool00`000000oooo0?ooo`0G0?ooo`@00000303oool00900oooo0`00002o0?ooo`030000003o
ool0oooo0300oooo0P00000I0?ooo`@00000303oool000h0oooo00<000000?ooo`3oool0P03oool4
00000;`0oooo0P00000F0?ooo`030000003oool0oooo01T0oooo0P00000I0?ooo`D000002`3oool0
00h0oooo00<000000?ooo`3oool0P@3oool4000000@0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo07X0oooo0P00000E0?ooo`8000006P3oool30000
01P0oooo1P00000;0?ooo`003@3oool2000008D0oooo0`0000020?ooo`<000006P3oool3000001X0
oooo0P00001l0?ooo`8000005@3oool2000001X0oooo1000000G0?ooo`H000002`3oool000d0oooo
0P00000L0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0603oool7000001T0oooo
1000000I0?ooo`@00000603oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool04`3oool3000001T0oooo1@00000I0?ooo`030000003oool0oooo00`0oooo
000<0?ooo`<000006`3oool2000004h0oooo0`00000I0?ooo`L000005`3oool6000001L0oooo1@00
000H0?ooo`<00000FP3oool2000000<0oooo0P00000E0?ooo`<000006@3oool5000001T0oooo00<0
00000?ooo`3oool0303oool000`0oooo1000000J0?ooo`800000703oool00`000000oooo0?ooo`0^
0?ooo`D000006@3oool6000001P0oooo1@00000H0?ooo`D000005P3oool5000003d0oooo00<00000
0?ooo`3oool06@3oool2000000<0oooo0P00000D0?ooo`@00000603oool6000001T0oooo00<00000
0?ooo`3oool0303oool000`0oooo1000000I0?ooo`<000006`3oool200000300oooo1`00000H0?oo
o`H000006@3oool4000001T0oooo0`00000F0?ooo`H00000>`3oool2000001X0oooo0`0000030?oo
o`800000503oool4000001P0oooo1P00000I0?ooo`030000003oool0oooo00`0oooo000;0?ooo`D0
00006@3oool3000001X0oooo0`00000J0?ooo`<000004`3oool9000001P0oooo1@00000I0?ooo`@0
00006@3oool3000001L0oooo1P00000I0?ooo`D000006P3oool2000001X0oooo100000030?ooo`03
0000003oool0oooo0180oooo1@00000J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool0303oool000/0oooo1@00000H0?ooo`@000006@3oool3000001L0oooo1P00000I0?ooo`D00000
603oool4000001/0oooo0P00000J0?ooo`8000006`3oool3000001<0oooo2@00000J0?ooo`<00000
6P3oool3000000<0oooo0P00000D0?ooo`D000006P3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo00`0oooo000=0?ooo`<00000603oool4000001T0oooo0`00000D0?ooo`P00000>03o
ool4000001/0oooo0P00000K0?ooo`030000003oool0oooo02h0oooo1`00000L0?ooo`<000006@3o
ool4000000<0oooo0P00000F0?ooo`<00000AP3oool002L0oooo1@00000H0?ooo`@000005@3oool6
000003/0oooo0`00001Y0?ooo`D00000703oool3000001X0oooo100000030?ooo`030000003oool0
oooo05h0oooo000X0?ooo`@000005`3oool5000001D0oooo1@00000m0?ooo`<00000J03oool40000
01`0oooo1000000I0?ooo`D000000`3oool00`000000oooo0?ooo`1N0?ooo`00:P3oool2000001H0
oooo1@00000G0?ooo`<00000?`3oool3000006P0oooo00<000000?ooo`3oool06`3oool5000001/0
oooo0`0000020?ooo`800000H03oool004<0oooo1000000H0?ooo`030000003oool0oooo03l0oooo
100000250?ooo`<000008@3oool200000600oooo00150?ooo`800000G03oool3000008H0oooo00<0
00000?ooo`3oool07`3oool00`000000oooo0?ooo`1O0?ooo`00Y03oool300000:L0oooo00<00000
0?ooo`3oool0G`3oool00:D0oooo0`00002U0?ooo`800000H@3oool00:H0oooo0`00002T0?ooo`03
0000003oool0oooo0600oooo002W0?ooo`<00000X`3oool00`000000oooo0?ooo`1P0?ooo`00Z03o
ool300000:80oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`0:0?ooo`00Z@3oool3
00000:00oooo0P00001E0?ooo`030000003oool0oooo00X0oooo002Z0?ooo`<00000W`3oool00`00
0000oooo0?ooo`1C0?ooo`800000303oool00:/0oooo0`00002N0?ooo`030000003oool0oooo05<0
oooo0P00000<0?ooo`00[03oool2000009d0oooo0P00000i0?ooo`030000003oool0oooo01P0oooo
0`00000<0?ooo`00[@3oool2000009`0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?oo
o`0H0?ooo`@000002`3oool000l0oooo00<000000?ooo`3oool0W03oool2000009/0oooo00<00000
0?ooo`3oool0=`3oool2000001X0oooo1000000;0?ooo`003P3oool2000009l0oooo0P00002I0?oo
o`800000>@3oool2000001T0oooo1@00000;0?ooo`003P3oool200000:00oooo0P00002H0?ooo`03
0000003oool0oooo03P0oooo0P00000I0?ooo`D000002`3oool000d0oooo0`00002P0?ooo`<00000
U`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo01P0oooo1000000J0?ooo`<00000
2`3oool000d0oooo0`00000L0?ooo`030000003oool0oooo0880oooo0`00002E0?ooo`8000007@3o
ool2000001X0oooo1000000J0?ooo`030000003oool0oooo00/0oooo000<0?ooo`D000006P3oool2
000008D0oooo0`00002D0?ooo`8000007@3oool2000001X0oooo1000000J0?ooo`030000003oool0
oooo00/0oooo000<0?ooo`D000006P3oool2000006`0oooo00<000000?ooo`3oool05`3oool30000
0080oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo
05H0oooo0P0000020?ooo`030000003oool0oooo01P0oooo0`00000I0?ooo`D000006P3oool00`00
0000oooo0?ooo`0;0?ooo`002`3oool6000001T0oooo0`00000L0?ooo`030000003oool0oooo04`0
oooo0`00000I0?ooo`H000006@3oool3000001P0oooo0`00001G0?ooo`80000000<0oooo00000000
00006P3oool3000001T0oooo1@00000I0?ooo`030000003oool0oooo00`0oooo000;0?ooo`H00000
6@3oool3000001/0oooo0P00001>0?ooo`@000006@3oool5000001P0oooo1000000H0?ooo`@00000
EP3oool2000000030?ooo`000000000001T0oooo1@00000I0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0303oool000d0oooo1000000H0?ooo`@000006P3oool300000300oooo0P00
000K0?ooo`H000005`3oool7000001H0oooo1P00000F0?ooo`L00000>`3oool00`000000oooo0?oo
o`0E0?ooo`80000000@0oooo00000000000000006@3oool5000001T0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0I0?ooo`@000006@3o
ool3000001`0oooo00<000000?ooo`3oool04P3oool4000001X0oooo1@00000H0?ooo`H000005`3o
ool6000001D0oooo2000000i0?ooo`8000005`3oool6000001P0oooo1P00000I0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool0603oool5
000001T0oooo0`00000J0?ooo`800000503oool8000001T0oooo1000000I0?ooo`D00000603oool5
000001L0oooo1P00000D0?ooo`X000006@3oool2000001P0oooo00<000000?ooo`0000000`00000H
0?ooo`H000006@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo000X0?oo
o`D00000603oool4000001P0oooo1000000D0?ooo`X000006@3oool3000001X0oooo0`00000J0?oo
o`<000006`3oool3000001<0oooo2@00000I0?ooo`<000005`3oool7000001/0oooo0`0000150?oo
o`00:P3oool3000001L0oooo1@00000F0?ooo`D000005@3oool500000200oooo0P00000J0?ooo`<0
00006`3oool200000300oooo1`00000J0?ooo`<00000603oool7000006<0oooo00130?ooo`D00000
5P3oool6000005L0oooo0P00001>0?ooo`@000006`3oool4000001P0oooo1`00001S0?ooo`00A03o
ool4000001L0oooo1000001I0?ooo`800000C@3oool2000001`0oooo1@00000G0?ooo`8000000`3o
ool3000006<0oooo00160?ooo`800000603oool3000005X0oooo0P00001[0?ooo`<00000603oool0
0`000000oooo0?ooo`1X0?ooo`00H@3oool00`000000oooo0?ooo`1J0?ooo`800000K03oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo06P0oooo002o0?ooo`800000P`3oool2000006X0
oooo00300?ooo`800000PP3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo00X0oooo
00310?ooo`800000P03oool2000005h0oooo00<000000?ooo`3oool02P3oool00<80oooo0P00001o
0?ooo`030000003oool0oooo05`0oooo0P00000<0?ooo`00``3oool2000007d0oooo0P00001N0?oo
o`<000002`3oool00<@0oooo0P00001l0?ooo`030000003oool0oooo05`0oooo1000000;0?ooo`00
a@3oool2000007X0oooo0P00001N0?ooo`@000002`3oool000l0oooo00<000000?ooo`3oool0]03o
ool2000007T0oooo0P0000130?ooo`030000003oool0oooo01P0oooo1000000;0?ooo`003`3oool0
0`000000oooo0?ooo`2e0?ooo`800000N03oool00`000000oooo0?ooo`120?ooo`030000003oool0
oooo01L0oooo1P00000:0?ooo`003P3oool200000;P0oooo0P00001f0?ooo`800000@`3oool20000
01T0oooo1P00000:0?ooo`003P3oool200000;P0oooo0`00001e0?ooo`030000003oool0oooo0480
oooo0P00000K0?ooo`@000002P3oool000h0oooo0P00002i0?ooo`<00000L`3oool2000004<0oooo
0`00000K0?ooo`030000003oool0oooo00/0oooo000=0?ooo`@000006`3oool00`000000oooo0?oo
o`2K0?ooo`800000L`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01L0oooo1000
000J0?ooo`030000003oool0oooo00/0oooo000=0?ooo`@000006P3oool2000009h0oooo0P00001a
0?ooo`800000:@3oool2000001T0oooo1000000J0?ooo`030000003oool0oooo00/0oooo000=0?oo
o`@000006P3oool2000009l0oooo0P00001`0?ooo`030000003oool0oooo02P0oooo0P00000H0?oo
o`D000006P3oool00`000000oooo0?ooo`0;0?ooo`00303oool5000001T0oooo0`00002P0?ooo`80
0000KP3oool2000002T0oooo0`00000H0?ooo`D000006P3oool00`000000oooo0?ooo`0;0?ooo`00
303oool5000001T0oooo0`00000L0?ooo`030000003oool0oooo0880oooo0P00001]0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0603oool3000001X0oooo0`00000J0?ooo`030000
003oool0oooo00/0oooo000>0?ooo`030000003oool0oooo01P0oooo1000000K0?ooo`800000D03o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01@0oooo0P0000040?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0B@3oool2000000h0oooo0P00000I0?ooo`@000006P3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000>0?ooo`030000003oool0
oooo01P0oooo1000000K0?ooo`800000C`3oool3000001/0oooo0P00000G0?ooo`8000000P3oool2
000001T0oooo0`00001:0?ooo`030000003oool0oooo00d0oooo0P00000I0?ooo`@000006P3oool0
0`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo
01P0oooo1@00000J0?ooo`<00000CP3oool5000001T0oooo1000000G0?ooo`H00000603oool40000
03d0oooo00<000000?ooo`3oool0203oool2000000h0oooo0`00000H0?ooo`D000006@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003oool0oooo01P0
oooo1@00000I0?ooo`@00000703oool00`000000oooo0?ooo`0_0?ooo`H000005`3oool5000001P0
oooo1@00000G0?ooo`H00000>`3oool2000000X0oooo00<000000?ooo`3oool0303oool4000001P0
oooo1@00000I0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0303oool000d0oooo
00<000000?ooo`3oool06P3oool3000001T0oooo0`00000K0?ooo`800000503oool5000001P0oooo
1`00000H0?ooo`D00000603oool5000001H0oooo1P00000F0?ooo`D000007`3oool2000000X0oooo
0P00000>0?ooo`<000006P3oool4000001T0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo1000
000J0?ooo`<00000503oool:000001D0oooo1P00000I0?ooo`<000006@3oool4000001P0oooo1@00
000E0?ooo`X00000603oool4000000X0oooo00<000000?ooo`3oool0303oool4000001X0oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo0015
0?ooo`@00000603oool4000001D0oooo2000000J0?ooo`@000006@3oool3000001T0oooo1000000I
0?ooo`@00000503oool8000001T0oooo1000000:0?ooo`8000003P3oool4000006<0oooo00140?oo
o`D000005P3oool6000001D0oooo1@00000P0?ooo`8000006@3oool2000001X0oooo0`00000L0?oo
o`800000503oool5000001X0oooo1@0000090?ooo`8000003P3oool5000006<0oooo00160?ooo`<0
00005@3oool6000001H0oooo0P0000100?ooo`030000003oool0oooo01P0oooo0`00000a0?ooo`80
0000703oool5000000X0oooo0P00000@0?ooo`<00000H`3oool005l0oooo1@00001d0?ooo`800000
D03oool4000000T0oooo0P00001[0?ooo`030000003oool0oooo00T0oooo001P0?ooo`<00000MP3o
ool200000500oooo0P00000:0?ooo`800000J`3oool00`000000oooo0?ooo`090?ooo`00HP3oool0
0`000000oooo0?ooo`1e0?ooo`<00000C`3oool00`000000oooo0?ooo`070?ooo`800000J`3oool2
000000/0oooo003K0?ooo`<00000E`3oool2000006`0oooo0P00000;0?ooo`00g03oool3000005H0
oooo0P00001/0?ooo`8000002`3oool00=h0oooo0P00001D0?ooo`800000K03oool4000000X0oooo
003O0?ooo`800000D`3oool00`000000oooo0?ooo`1[0?ooo`@000002P3oool00100oooo00<00000
0?ooo`3oool0c@3oool200000540oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0H
0?ooo`@000002P3oool000l0oooo0P00003@0?ooo`800000C`3oool200000580oooo0P00000I0?oo
o`D000002P3oool000l0oooo0P00003A0?ooo`800000CP3oool00`000000oooo0?ooo`1A0?ooo`80
00006@3oool5000000X0oooo000>0?ooo`<00000dP3oool2000004`0oooo00<000000?ooo`3oool0
D@3oool3000001/0oooo00<000000?ooo`3oool02P3oool000h0oooo0`00003C0?ooo`800000B`3o
ool00`000000oooo0?ooo`1A0?ooo`<000006`3oool00`000000oooo0?ooo`0:0?ooo`003@3oool5
00000=<0oooo0P0000190?ooo`030000003oool0oooo0540oooo1@00000J0?ooo`030000003oool0
oooo00X0oooo000=0?ooo`D000006`3oool00`000000oooo0?ooo`2f0?ooo`800000A`3oool00`00
0000oooo0?ooo`0h0?ooo`030000003oool0oooo01L0oooo1@00000I0?ooo`030000003oool0oooo
00/0oooo000<0?ooo`H000006P3oool200000;T0oooo0`0000140?ooo`800000>@3oool2000001P0
oooo1P00000I0?ooo`030000003oool0oooo00/0oooo000<0?ooo`H000006P3oool200000;/0oooo
0P0000130?ooo`030000003oool0oooo03P0oooo0P00000H0?ooo`H000006@3oool00`000000oooo
0?ooo`0;0?ooo`003P3oool00`000000oooo0?ooo`0J0?ooo`<00000_03oool200000440oooo00<0
00000?ooo`3oool0>03oool3000001X0oooo1000000I0?ooo`030000003oool0oooo00/0oooo000>
0?ooo`030000003oool0oooo01X0oooo0`00002m0?ooo`800000?`3oool00`000000oooo0?ooo`0i
0?ooo`<000006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000>0?oo
o`030000003oool0oooo01T0oooo1000000K0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0>`3oool200000200oooo00<000000?ooo`3oool05`3oool4000001X0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`003P3oool00`000000oooo0?ooo`0I0?ooo`@0
00006P3oool200000:@0oooo0P00000j0?ooo`800000803oool2000001T0oooo1000000J0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool02`3oool000h0oooo00<000000?ooo`3oool0
603oool5000001X0oooo0P00001^0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
5`3oool3000003L0oooo0P00000Q0?ooo`800000603oool5000001X0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0;0?ooo`003P3oool00`000000oooo0?ooo`0H0?ooo`D000006@3o
ool300000500oooo00<000000?ooo`3oool06P3oool3000001X0oooo0`00000I0?ooo`<00000=@3o
ool200000240oooo0`00000H0?ooo`D000006P3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00`0oooo000=0?ooo`030000003oool0oooo01/0oooo0`00000H0?ooo`@00000703oool0
0`000000oooo0?ooo`0`0?ooo`<000006P3oool4000001T0oooo1000000H0?ooo`H00000<@3oool3
000000L0oooo00<000000?ooo`3oool05`3oool4000001X0oooo0`00000I0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool0303oool000d0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0H0?ooo`@000006`3oool300000380oooo1000000H0?ooo`H000005`3oool60000
01L0oooo1`00000E0?ooo`800000603oool3000000H0oooo0P00000I0?ooo`@000006P3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0303oool000d0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`@000006P3oool3000001@0
oooo1@00000I0?ooo`H00000603oool5000001P0oooo1@00000F0?ooo`L0000000<0oooo00000000
00004`3oool6000001<0oooo0P0000070?ooo`<000006@3oool4000001X0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool05`3oool5000001T0oooo1000000D0?ooo`X00000
5P3oool5000001T0oooo1000000I0?ooo`@000005`3oool6000000040?ooo`000000000000000140
oooo2@00000?0?ooo`8000001`3oool3000001T0oooo1@00000J0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool010000000oooo0?ooo`00000;0?ooo`00A@3oool4000001L0
oooo1@00000E0?ooo`P000006P3oool4000001T0oooo1000000I0?ooo`@000006@3oool400000080
oooo0`00000?0?ooo`L000003`3oool3000000H0oooo1@00000J0?ooo`@00000E`3oool00`000000
oooo0?ooo`090?ooo`00A`3oool2000001H0oooo1P00000F0?ooo`D000007`3oool2000001X0oooo
0P00000K0?ooo`800000703oool2000000<0oooo0`00000=0?ooo`D000003`3oool3000000L0oooo
1@00000M0?ooo`800000EP3oool2000000/0oooo001P0?ooo`D000005P3oool200000400oooo00<0
00000?ooo`3oool06P3oool00`000000oooo0?ooo`0O0?ooo`@000002`3oool200000100oooo1000
00090?ooo`@00000M@3oool2000000/0oooo001Q0?ooo`<00000VP3oool4000001P0oooo1@00000<
0?ooo`800000MP3oool2000000/0oooo001S0?ooo`030000003oool0oooo09/0oooo1@00000B0?oo
o`D000003`3oool00`000000oooo0?ooo`1c0?ooo`@000002P3oool00?l0oooo103oool7000000X0
oooo1`0000270?ooo`@000002P3oool00100oooo00<000000?ooo`3oool0l`3ooolA000008/0oooo
1000000:0?ooo`00403oool00`000000oooo0?ooo`3o0?ooogD0oooo00<000000?ooo`3oool05`3o
ool5000000X0oooo000?0?ooo`800000o`3ooomg0?ooo`030000003oool0oooo01L0oooo1@00000:
0?ooo`003`3oool200000?l0ooooMP3oool2000001/0oooo00<000000?ooo`3oool02P3oool000h0
oooo0`00003o0?ooogH0oooo0P00000K0?ooo`030000003oool0oooo00X0oooo000>0?ooo`@00000
o`3ooome0?ooo`8000006`3oool00`000000oooo0?ooo`0:0?ooo`003P3oool400000?l0ooooM03o
ool4000001X0oooo00<000000?ooo`3oool02P3oool000d0oooo1@00000K0?ooo`030000003oool0
oooo0?l0ooooEP3oool4000001X0oooo00<000000?ooo`3oool02P3oool000d0oooo1@00000J0?oo
o`800000o`3oooln0?ooo`030000003oool0oooo01L0oooo1000000I0?ooo`030000003oool0oooo
00/0oooo000?0?ooo`<000006P3oool200000?l0oooo?@3oool2000001P0oooo1@00000I0?ooo`03
0000003oool0oooo00/0oooo000?0?ooo`030000003oool0oooo01T0oooo0`00003o0?ooocd0oooo
0P00000H0?ooo`D000006@3oool00`000000oooo0?ooo`0;0?ooo`003P3oool00`000000oooo0?oo
o`0J0?ooo`<00000o`3oooll0?ooo`<000006P3oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00/0oooo000>0?ooo`030000003oool0oooo01T0oooo1000003o0?oooc/0oooo1000000I
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02`3oool000h0oooo00<000000?oo
o`3oool06@3oool4000001`0oooo00<000000?ooo`3oool0o`3ooolL0?ooo`@000006@3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000>0?ooo`030000003oool0oooo01P0
oooo1@00000K0?ooo`800000o`3oool40?ooo`030000003oool0oooo01H0oooo1@00000I0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool02`3oool000h0oooo00<000000?ooo`3oool0
603oool5000001/0oooo0P00003o0?ooo`<0oooo0P00000H0?ooo`D000006@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000>0?ooo`030000003oool0oooo01X0oooo0`00
000J0?ooo`<00000o`3oool20?ooo`<000005`3oool6000001T0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0;0?ooo`003P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01T0oooo1000003o0?ooo`40oooo0`00000J0?ooo`@000006@3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo00/0oooo000>0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0603oool4000001d0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1G0?oo
o`<000006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo00<000000?oo
o`3oool0303oool000d0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0H0?ooo`@0
00006`3oool200000380oooo0`00000L0?ooo`<000006P3oool3000001P0oooo0`00000n0?ooo`03
0000003oool0oooo01H0oooo1000000J0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0<0?ooo`003@3oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo01P0oooo1@00000J0?ooo`<00000<P3oool4000001X0oooo1000000I0?ooo`@00000
603oool4000001L0oooo0`00000Q0?ooo`800000603oool5000001X0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo000=0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool06@3oool4000001T0oooo0`00000E0?ooo`X000004`3o
ool6000001P0oooo1P00000G0?ooo`H000005P3oool6000001H0oooo1`00000K0?ooo`@000005`3o
ool5000001X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo00`0oooo000=0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool3
000001L0oooo1@00000E0?ooo`T00000503oool7000001P0oooo1@00000H0?ooo`D00000603oool5
000001D0oooo2P00000F0?ooo`D000006@3oool4000001X0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00`0oooo001P0?ooo`D000005P3oool70000
01P0oooo1P00000I0?ooo`@000006@3oool4000001T0oooo1000000D0?ooo`L000005`3oool60000
01`0oooo0P00001S0?ooo`00H@3oool4000001L0oooo1000000N0?ooo`<000006P3oool2000001/0
oooo0P00000L0?ooo`8000004`3oool3000001`0oooo100000220?ooo`00HP3oool2000001P0oooo
0P00000o0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0CP3oool300000880oooo
001S0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0P@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {431, 0}} -> {-3.19972, -3.19972, 0.0222719, \
0.0222719}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseAndField[0.1, 0, 1, 40, 2.5, 2.5];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.04745 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.186272 0.523624 0.186272 [
[ 0 0 0 0 ]
[ 1 1.04745 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto fill
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto
-0.02 0.005 lineto stroke
} ifelse
grestore
} def
% End of user PostScript
0 g
.0001 w
[ ] 0 setdash
.03432 .05794 m
.02381 .12363 L
s
% Start of user PostScript
-2.5 -2.5 -2.5564252 -2.1473426 MAarrowhead1
% End of user PostScript
.03432 .12447 m
.02531 .18228 L
s
% Start of user PostScript
-2.5 -2.1428571 -2.5483644 -1.8325187 MAarrowhead1
% End of user PostScript
.03432 .191 m
.02681 .24092 L
s
% Start of user PostScript
-2.5 -1.7857143 -2.5403037 -1.5176947 MAarrowhead1
% End of user PostScript
.03432 .25752 m
.02831 .29956 L
s
% Start of user PostScript
-2.5 -1.4285714 -2.532243 -1.2028707 MAarrowhead1
% End of user PostScript
.03432 .32405 m
.02982 .35821 L
s
% Start of user PostScript
-2.5 -1.0714286 -2.5241822 -0.88804674 MAarrowhead1
% End of user PostScript
.03432 .39057 m
.03132 .41685 L
s
% Start of user PostScript
-2.5 -0.71428571 -2.5161215 -0.57322277 MAarrowhead1
% End of user PostScript
.03432 .4571 m
.03282 .47549 L
s
% Start of user PostScript
-2.5 -0.35714286 -2.5080607 -0.25839879 MAarrowhead1
% End of user PostScript
.03432 .52362 m
.03432 .53413 L
s
% Start of user PostScript
-2.5 -3.3306691e-16 -2.5 0.056425179 MAarrowhead1
% End of user PostScript
.03432 .59015 m
.03582 .59278 L
s
% Start of user PostScript
-2.5 0.35714286 -2.4919393 0.37124915 MAarrowhead1
% End of user PostScript
.03432 .65668 m
.03732 .65142 L
s
% Start of user PostScript
-2.5 0.71428571 -2.4838785 0.68607312 MAarrowhead1
% End of user PostScript
.03432 .7232 m
.03882 .71006 L
s
% Start of user PostScript
-2.5 1.0714286 -2.4758178 1.0008971 MAarrowhead1
% End of user PostScript
.03432 .78973 m
.04033 .76871 L
s
% Start of user PostScript
-2.5 1.4285714 -2.467757 1.3157211 MAarrowhead1
% End of user PostScript
.03432 .85625 m
.04183 .82735 L
s
% Start of user PostScript
-2.5 1.7857143 -2.4596963 1.630545 MAarrowhead1
% End of user PostScript
.03432 .92278 m
.04333 .88599 L
s
% Start of user PostScript
-2.5 2.1428571 -2.4516356 1.945369 MAarrowhead1
% End of user PostScript
.03432 .9893 m
.04483 .94463 L
s
% Start of user PostScript
-2.5 2.5 -2.4435748 2.260193 MAarrowhead1
% End of user PostScript
.10085 .05794 m
.09034 .1047 L
s
% Start of user PostScript
-2.1428571 -2.5 -2.1992823 -2.2489655 MAarrowhead1
% End of user PostScript
.10085 .12447 m
.09184 .16584 L
s
% Start of user PostScript
-2.1428571 -2.1428571 -2.1912216 -1.9207755 MAarrowhead1
% End of user PostScript
.10085 .191 m
.09334 .22697 L
s
% Start of user PostScript
-2.1428571 -1.7857143 -2.1831608 -1.5925855 MAarrowhead1
% End of user PostScript
.10085 .25752 m
.09484 .2881 L
s
% Start of user PostScript
-2.1428571 -1.4285714 -2.1751001 -1.2643955 MAarrowhead1
% End of user PostScript
.10085 .32405 m
.09634 .34924 L
s
% Start of user PostScript
-2.1428571 -1.0714286 -2.1670394 -0.93620555 MAarrowhead1
% End of user PostScript
.10085 .39057 m
.09784 .41037 L
s
% Start of user PostScript
-2.1428571 -0.71428571 -2.1589786 -0.60801555 MAarrowhead1
% End of user PostScript
.10085 .4571 m
.09934 .4715 L
s
% Start of user PostScript
-2.1428571 -0.35714286 -2.1509179 -0.27982556 MAarrowhead1
% End of user PostScript
.10085 .52362 m
.10085 .53263 L
s
% Start of user PostScript
-2.1428571 -3.3306691e-16 -2.1428571 0.048364439 MAarrowhead1
% End of user PostScript
.10085 .59015 m
.10235 .59377 L
s
% Start of user PostScript
-2.1428571 0.35714286 -2.1347964 0.37655443 MAarrowhead1
% End of user PostScript
.10085 .65668 m
.10385 .6549 L
s
% Start of user PostScript
-2.1428571 0.71428571 -2.1267357 0.70474443 MAarrowhead1
% End of user PostScript
.10085 .7232 m
.10535 .71603 L
s
% Start of user PostScript
-2.1428571 1.0714286 -2.1186749 1.0329344 MAarrowhead1
% End of user PostScript
.10085 .78973 m
.10685 .77716 L
s
% Start of user PostScript
-2.1428571 1.4285714 -2.1106142 1.3611244 MAarrowhead1
% End of user PostScript
.10085 .85625 m
.10835 .8383 L
s
% Start of user PostScript
-2.1428571 1.7857143 -2.1025534 1.6893144 MAarrowhead1
% End of user PostScript
.10085 .92278 m
.10985 .89943 L
s
% Start of user PostScript
-2.1428571 2.1428571 -2.0944927 2.0175044 MAarrowhead1
% End of user PostScript
.10085 .9893 m
.11136 .96056 L
s
% Start of user PostScript
-2.1428571 2.5 -2.086432 2.3456944 MAarrowhead1
% End of user PostScript
.16737 .05794 m
.15686 .08846 L
s
% Start of user PostScript
-1.7857143 -2.5 -1.8421395 -2.3361943 MAarrowhead1
% End of user PostScript
.16737 .12447 m
.15836 .1517 L
s
% Start of user PostScript
-1.7857143 -2.1428571 -1.8340787 -1.9966945 MAarrowhead1
% End of user PostScript
.16737 .191 m
.15986 .21493 L
s
% Start of user PostScript
-1.7857143 -1.7857143 -1.826018 -1.6571948 MAarrowhead1
% End of user PostScript
.16737 .25752 m
.16137 .27817 L
s
% Start of user PostScript
-1.7857143 -1.4285714 -1.8179572 -1.3176951 MAarrowhead1
% End of user PostScript
.16737 .32405 m
.16287 .34141 L
s
% Start of user PostScript
-1.7857143 -1.0714286 -1.8098965 -0.97819542 MAarrowhead1
% End of user PostScript
.16737 .39057 m
.16437 .40465 L
s
% Start of user PostScript
-1.7857143 -0.71428571 -1.8018358 -0.63869572 MAarrowhead1
% End of user PostScript
.16737 .4571 m
.16587 .46789 L
s
% Start of user PostScript
-1.7857143 -0.35714286 -1.793775 -0.29919601 MAarrowhead1
% End of user PostScript
.16737 .52362 m
.16737 .53113 L
s
% Start of user PostScript
-1.7857143 -3.3306691e-16 -1.7857143 0.040303699 MAarrowhead1
% End of user PostScript
.16737 .59015 m
.16887 .59437 L
s
% Start of user PostScript
-1.7857143 0.35714286 -1.7776535 0.37980341 MAarrowhead1
% End of user PostScript
.16737 .65668 m
.17037 .65761 L
s
% Start of user PostScript
-1.7857143 0.71428571 -1.7695928 0.71930311 MAarrowhead1
% End of user PostScript
.16737 .7232 m
.17188 .72085 L
s
% Start of user PostScript
-1.7857143 1.0714286 -1.7615321 1.0588028 MAarrowhead1
% End of user PostScript
.16737 .78973 m
.17338 .78409 L
s
% Start of user PostScript
-1.7857143 1.4285714 -1.7534713 1.3983025 MAarrowhead1
% End of user PostScript
.16737 .85625 m
.17488 .84733 L
s
% Start of user PostScript
-1.7857143 1.7857143 -1.7454106 1.7378022 MAarrowhead1
% End of user PostScript
.16737 .92278 m
.17638 .91057 L
s
% Start of user PostScript
-1.7857143 2.1428571 -1.7373498 2.0773019 MAarrowhead1
% End of user PostScript
.16737 .9893 m
.17788 .97381 L
s
% Start of user PostScript
-1.7857143 2.5 -1.7292891 2.4168016 MAarrowhead1
% End of user PostScript
.2339 .05794 m
.22339 .07489 L
s
% Start of user PostScript
-1.4285714 -2.5 -1.4849966 -2.4090288 MAarrowhead1
% End of user PostScript
.2339 .12447 m
.22489 .13985 L
s
% Start of user PostScript
-1.4285714 -2.1428571 -1.4769359 -2.0602757 MAarrowhead1
% End of user PostScript
.2339 .191 m
.22639 .20482 L
s
% Start of user PostScript
-1.4285714 -1.7857143 -1.4688751 -1.7115226 MAarrowhead1
% End of user PostScript
.2339 .25752 m
.22789 .26978 L
s
% Start of user PostScript
-1.4285714 -1.4285714 -1.4608144 -1.3627695 MAarrowhead1
% End of user PostScript
.2339 .32405 m
.22939 .33474 L
s
% Start of user PostScript
-1.4285714 -1.0714286 -1.4527536 -1.0140164 MAarrowhead1
% End of user PostScript
.2339 .39057 m
.23089 .3997 L
s
% Start of user PostScript
-1.4285714 -0.71428571 -1.4446929 -0.66526326 MAarrowhead1
% End of user PostScript
.2339 .4571 m
.2324 .46467 L
s
% Start of user PostScript
-1.4285714 -0.35714286 -1.4366322 -0.31651015 MAarrowhead1
% End of user PostScript
.2339 .52362 m
.2339 .52963 L
s
% Start of user PostScript
-1.4285714 -3.3306691e-16 -1.4285714 0.032242959 MAarrowhead1
% End of user PostScript
.2339 .59015 m
.2354 .59459 L
s
% Start of user PostScript
-1.4285714 0.35714286 -1.4205107 0.38099607 MAarrowhead1
% End of user PostScript
.2339 .65668 m
.2369 .65956 L
s
% Start of user PostScript
-1.4285714 0.71428571 -1.4124499 0.72974917 MAarrowhead1
% End of user PostScript
.2339 .7232 m
.2384 .72452 L
s
% Start of user PostScript
-1.4285714 1.0714286 -1.4043892 1.0785023 MAarrowhead1
% End of user PostScript
.2339 .78973 m
.2399 .78948 L
s
% Start of user PostScript
-1.4285714 1.4285714 -1.3963285 1.4272554 MAarrowhead1
% End of user PostScript
.2339 .85625 m
.2414 .85444 L
s
% Start of user PostScript
-1.4285714 1.7857143 -1.3882677 1.7760085 MAarrowhead1
% End of user PostScript
.2339 .92278 m
.24291 .91941 L
s
% Start of user PostScript
-1.4285714 2.1428571 -1.380207 2.1247616 MAarrowhead1
% End of user PostScript
.2339 .9893 m
.24441 .98437 L
s
% Start of user PostScript
-1.4285714 2.5 -1.3721462 2.4735147 MAarrowhead1
% End of user PostScript
.30042 .05794 m
.28991 .064 L
s
% Start of user PostScript
-1.0714286 -2.5 -1.1278538 -2.4674692 MAarrowhead1
% End of user PostScript
.30042 .12447 m
.29141 .13031 L
s
% Start of user PostScript
-1.0714286 -2.1428571 -1.119793 -2.111519 MAarrowhead1
% End of user PostScript
.30042 .191 m
.29292 .19661 L
s
% Start of user PostScript
-1.0714286 -1.7857143 -1.1117323 -1.7555688 MAarrowhead1
% End of user PostScript
.30042 .25752 m
.29442 .26291 L
s
% Start of user PostScript
-1.0714286 -1.4285714 -1.1036715 -1.3996186 MAarrowhead1
% End of user PostScript
.30042 .32405 m
.29592 .32922 L
s
% Start of user PostScript
-1.0714286 -1.0714286 -1.0956108 -1.0436684 MAarrowhead1
% End of user PostScript
.30042 .39057 m
.29742 .39552 L
s
% Start of user PostScript
-1.0714286 -0.71428571 -1.0875501 -0.68771817 MAarrowhead1
% End of user PostScript
.30042 .4571 m
.29892 .46182 L
s
% Start of user PostScript
-1.0714286 -0.35714286 -1.0794893 -0.33176798 MAarrowhead1
% End of user PostScript
.30042 .52362 m
.30042 .52813 L
s
% Start of user PostScript
-1.0714286 -3.3306691e-16 -1.0714286 0.02418222 MAarrowhead1
% End of user PostScript
.30042 .59015 m
.30192 .59443 L
s
% Start of user PostScript
-1.0714286 0.35714286 -1.0633678 0.38013242 MAarrowhead1
% End of user PostScript
.30042 .65668 m
.30343 .66074 L
s
% Start of user PostScript
-1.0714286 0.71428571 -1.0553071 0.73608261 MAarrowhead1
% End of user PostScript
.30042 .7232 m
.30493 .72704 L
s
% Start of user PostScript
-1.0714286 1.0714286 -1.0472464 1.0920328 MAarrowhead1
% End of user PostScript
.30042 .78973 m
.30643 .79334 L
s
% Start of user PostScript
-1.0714286 1.4285714 -1.0391856 1.447983 MAarrowhead1
% End of user PostScript
.30042 .85625 m
.30793 .85965 L
s
% Start of user PostScript
-1.0714286 1.7857143 -1.0311249 1.8039332 MAarrowhead1
% End of user PostScript
.30042 .92278 m
.30943 .92595 L
s
% Start of user PostScript
-1.0714286 2.1428571 -1.0230641 2.1598834 MAarrowhead1
% End of user PostScript
.30042 .9893 m
.31093 .99225 L
s
% Start of user PostScript
-1.0714286 2.5 -1.0150034 2.5158336 MAarrowhead1
% End of user PostScript
.36695 .05794 m
.35644 .0558 L
s
% Start of user PostScript
-0.71428571 -2.5 -0.77071089 -2.5115153 MAarrowhead1
% End of user PostScript
.36695 .12447 m
.35794 .12306 L
s
% Start of user PostScript
-0.71428571 -2.1428571 -0.76265015 -2.1504244 MAarrowhead1
% End of user PostScript
.36695 .191 m
.35944 .19032 L
s
% Start of user PostScript
-0.71428571 -1.7857143 -0.75458941 -1.7893334 MAarrowhead1
% End of user PostScript
.36695 .25752 m
.36094 .25758 L
s
% Start of user PostScript
-0.71428571 -1.4285714 -0.74652867 -1.4282424 MAarrowhead1
% End of user PostScript
.36695 .32405 m
.36244 .32484 L
s
% Start of user PostScript
-0.71428571 -1.0714286 -0.73846793 -1.0671514 MAarrowhead1
% End of user PostScript
.36695 .39057 m
.36395 .3921 L
s
% Start of user PostScript
-0.71428571 -0.71428571 -0.73040719 -0.70606047 MAarrowhead1
% End of user PostScript
.36695 .4571 m
.36545 .45937 L
s
% Start of user PostScript
-0.71428571 -0.35714286 -0.72234645 -0.34496949 MAarrowhead1
% End of user PostScript
.36695 .52362 m
.36695 .52663 L
s
% Start of user PostScript
-0.71428571 -3.3306691e-16 -0.71428571 0.01612148 MAarrowhead1
% End of user PostScript
.36695 .59015 m
.36845 .59389 L
s
% Start of user PostScript
-0.71428571 0.35714286 -0.70622497 0.37721245 MAarrowhead1
% End of user PostScript
.36695 .65668 m
.36995 .66115 L
s
% Start of user PostScript
-0.71428571 0.71428571 -0.69816423 0.73830343 MAarrowhead1
% End of user PostScript
.36695 .7232 m
.37145 .72841 L
s
% Start of user PostScript
-0.71428571 1.0714286 -0.69010349 1.0993944 MAarrowhead1
% End of user PostScript
.36695 .78973 m
.37295 .79567 L
s
% Start of user PostScript
-0.71428571 1.4285714 -0.68204275 1.4604854 MAarrowhead1
% End of user PostScript
.36695 .85625 m
.37446 .86293 L
s
% Start of user PostScript
-0.71428571 1.7857143 -0.67398202 1.8215764 MAarrowhead1
% End of user PostScript
.36695 .92278 m
.37596 .93019 L
s
% Start of user PostScript
-0.71428571 2.1428571 -0.66592128 2.1826673 MAarrowhead1
% End of user PostScript
.36695 .9893 m
.37746 .99745 L
s
% Start of user PostScript
-0.71428571 2.5 -0.65786054 2.5437583 MAarrowhead1
% End of user PostScript
.43347 .05794 m
.42296 .05028 L
s
% Start of user PostScript
-0.35714286 -2.5 -0.41356804 -2.5411673 MAarrowhead1
% End of user PostScript
.43347 .12447 m
.42447 .11811 L
s
% Start of user PostScript
-0.35714286 -2.1428571 -0.4055073 -2.1769919 MAarrowhead1
% End of user PostScript
.43347 .191 m
.42597 .18595 L
s
% Start of user PostScript
-0.35714286 -1.7857143 -0.39744656 -1.8128165 MAarrowhead1
% End of user PostScript
.43347 .25752 m
.42747 .25378 L
s
% Start of user PostScript
-0.35714286 -1.4285714 -0.38938582 -1.448641 MAarrowhead1
% End of user PostScript
.43347 .32405 m
.42897 .32162 L
s
% Start of user PostScript
-0.35714286 -1.0714286 -0.38132508 -1.0844656 MAarrowhead1
% End of user PostScript
.43347 .39057 m
.43047 .38945 L
s
% Start of user PostScript
-0.35714286 -0.71428571 -0.37326434 -0.72029014 MAarrowhead1
% End of user PostScript
.43347 .4571 m
.43197 .45729 L
s
% Start of user PostScript
-0.35714286 -0.35714286 -0.3652036 -0.3561147 MAarrowhead1
% End of user PostScript
.43347 .52362 m
.43347 .52513 L
s
% Start of user PostScript
-0.35714286 -3.3306691e-16 -0.35714286 0.0080607399 MAarrowhead1
% End of user PostScript
.43347 .59015 m
.43498 .59296 L
s
% Start of user PostScript
-0.35714286 0.35714286 -0.34908212 0.37223618 MAarrowhead1
% End of user PostScript
.43347 .65668 m
.43648 .6608 L
s
% Start of user PostScript
-0.35714286 0.71428571 -0.34102138 0.73641162 MAarrowhead1
% End of user PostScript
.43347 .7232 m
.43798 .72863 L
s
% Start of user PostScript
-0.35714286 1.0714286 -0.33296064 1.1005871 MAarrowhead1
% End of user PostScript
.43347 .78973 m
.43948 .79647 L
s
% Start of user PostScript
-0.35714286 1.4285714 -0.3248999 1.4647625 MAarrowhead1
% End of user PostScript
.43347 .85625 m
.44098 .8643 L
s
% Start of user PostScript
-0.35714286 1.7857143 -0.31683916 1.8289379 MAarrowhead1
% End of user PostScript
.43347 .92278 m
.44248 .93214 L
s
% Start of user PostScript
-0.35714286 2.1428571 -0.30877842 2.1931134 MAarrowhead1
% End of user PostScript
.43347 .9893 m
.44398 .99998 L
s
% Start of user PostScript
-0.35714286 2.5 -0.30071768 2.5572888 MAarrowhead1
% End of user PostScript
.5 .05794 m
.48949 .04743 L
s
% Start of user PostScript
-3.3306691e-16 -2.5 -0.056425179 -2.5564252 MAarrowhead1
% End of user PostScript
.5 .12447 m
.49099 .11546 L
s
% Start of user PostScript
-3.3306691e-16 -2.1428571 -0.048364439 -2.1912216 MAarrowhead1
% End of user PostScript
.5 .191 m
.49249 .18349 L
s
% Start of user PostScript
-3.3306691e-16 -1.7857143 -0.040303699 -1.826018 MAarrowhead1
% End of user PostScript
.5 .25752 m
.49399 .25152 L
s
% Start of user PostScript
-3.3306691e-16 -1.4285714 -0.032242959 -1.4608144 MAarrowhead1
% End of user PostScript
.5 .32405 m
.4955 .31954 L
s
% Start of user PostScript
-3.3306691e-16 -1.0714286 -0.02418222 -1.0956108 MAarrowhead1
% End of user PostScript
.5 .39057 m
.497 .38757 L
s
% Start of user PostScript
-3.3306691e-16 -0.71428571 -0.01612148 -0.73040719 MAarrowhead1
% End of user PostScript
.5 .4571 m
.4985 .4556 L
s
% Start of user PostScript
-3.3306691e-16 -0.35714286 -0.0080607399 -0.3652036 MAarrowhead1
% End of user PostScript
.5 .52362 m
.5 .52362 L
s
% Start of user PostScript
-3.3306691e-16 -3.3306691e-16 -3.4058425e-16 -3.3306691e-16 MAarrowhead1
% End of user PostScript
.5 .59015 m
.5015 .59165 L
s
% Start of user PostScript
-3.3306691e-16 0.35714286 0.0080607399 0.3652036 MAarrowhead1
% End of user PostScript
.5 .65668 m
.503 .65968 L
s
% Start of user PostScript
-3.3306691e-16 0.71428571 0.01612148 0.73040719 MAarrowhead1
% End of user PostScript
.5 .7232 m
.5045 .72771 L
s
% Start of user PostScript
-3.3306691e-16 1.0714286 0.02418222 1.0956108 MAarrowhead1
% End of user PostScript
.5 .78973 m
.50601 .79573 L
s
% Start of user PostScript
-3.3306691e-16 1.4285714 0.032242959 1.4608144 MAarrowhead1
% End of user PostScript
.5 .85625 m
.50751 .86376 L
s
% Start of user PostScript
-3.3306691e-16 1.7857143 0.040303699 1.826018 MAarrowhead1
% End of user PostScript
.5 .92278 m
.50901 .93179 L
s
% Start of user PostScript
-3.3306691e-16 2.1428571 0.048364439 2.1912216 MAarrowhead1
% End of user PostScript
.5 .9893 m
.51051 .99981 L
s
% Start of user PostScript
-3.3306691e-16 2.5 0.056425179 2.5564252 MAarrowhead1
% End of user PostScript
.56653 .05794 m
.55602 .04727 L
s
% Start of user PostScript
0.35714286 -2.5 0.30071768 -2.5572888 MAarrowhead1
% End of user PostScript
.56653 .12447 m
.55752 .11511 L
s
% Start of user PostScript
0.35714286 -2.1428571 0.30877842 -2.1931134 MAarrowhead1
% End of user PostScript
.56653 .191 m
.55902 .18294 L
s
% Start of user PostScript
0.35714286 -1.7857143 0.31683916 -1.8289379 MAarrowhead1
% End of user PostScript
.56653 .25752 m
.56052 .25078 L
s
% Start of user PostScript
0.35714286 -1.4285714 0.3248999 -1.4647625 MAarrowhead1
% End of user PostScript
.56653 .32405 m
.56202 .31862 L
s
% Start of user PostScript
0.35714286 -1.0714286 0.33296064 -1.1005871 MAarrowhead1
% End of user PostScript
.56653 .39057 m
.56352 .38645 L
s
% Start of user PostScript
0.35714286 -0.71428571 0.34102138 -0.73641162 MAarrowhead1
% End of user PostScript
.56653 .4571 m
.56502 .45429 L
s
% Start of user PostScript
0.35714286 -0.35714286 0.34908212 -0.37223618 MAarrowhead1
% End of user PostScript
.56653 .52362 m
.56653 .52212 L
s
% Start of user PostScript
0.35714286 -3.3306691e-16 0.35714286 -0.0080607399 MAarrowhead1
% End of user PostScript
.56653 .59015 m
.56803 .58996 L
s
% Start of user PostScript
0.35714286 0.35714286 0.3652036 0.3561147 MAarrowhead1
% End of user PostScript
.56653 .65668 m
.56953 .65779 L
s
% Start of user PostScript
0.35714286 0.71428571 0.37326434 0.72029014 MAarrowhead1
% End of user PostScript
.56653 .7232 m
.57103 .72563 L
s
% Start of user PostScript
0.35714286 1.0714286 0.38132508 1.0844656 MAarrowhead1
% End of user PostScript
.56653 .78973 m
.57253 .79347 L
s
% Start of user PostScript
0.35714286 1.4285714 0.38938582 1.448641 MAarrowhead1
% End of user PostScript
.56653 .85625 m
.57403 .8613 L
s
% Start of user PostScript
0.35714286 1.7857143 0.39744656 1.8128165 MAarrowhead1
% End of user PostScript
.56653 .92278 m
.57553 .92914 L
s
% Start of user PostScript
0.35714286 2.1428571 0.4055073 2.1769919 MAarrowhead1
% End of user PostScript
.56653 .9893 m
.57704 .99697 L
s
% Start of user PostScript
0.35714286 2.5 0.41356804 2.5411673 MAarrowhead1
% End of user PostScript
.63305 .05794 m
.62254 .04979 L
s
% Start of user PostScript
0.71428571 -2.5 0.65786054 -2.5437583 MAarrowhead1
% End of user PostScript
.63305 .12447 m
.62404 .11705 L
s
% Start of user PostScript
0.71428571 -2.1428571 0.66592128 -2.1826673 MAarrowhead1
% End of user PostScript
.63305 .191 m
.62554 .18432 L
s
% Start of user PostScript
0.71428571 -1.7857143 0.67398202 -1.8215764 MAarrowhead1
% End of user PostScript
.63305 .25752 m
.62705 .25158 L
s
% Start of user PostScript
0.71428571 -1.4285714 0.68204275 -1.4604854 MAarrowhead1
% End of user PostScript
.63305 .32405 m
.62855 .31884 L
s
% Start of user PostScript
0.71428571 -1.0714286 0.69010349 -1.0993944 MAarrowhead1
% End of user PostScript
.63305 .39057 m
.63005 .3861 L
s
% Start of user PostScript
0.71428571 -0.71428571 0.69816423 -0.73830343 MAarrowhead1
% End of user PostScript
.63305 .4571 m
.63155 .45336 L
s
% Start of user PostScript
0.71428571 -0.35714286 0.70622497 -0.37721245 MAarrowhead1
% End of user PostScript
.63305 .52362 m
.63305 .52062 L
s
% Start of user PostScript
0.71428571 -3.3306691e-16 0.71428571 -0.01612148 MAarrowhead1
% End of user PostScript
.63305 .59015 m
.63455 .58788 L
s
% Start of user PostScript
0.71428571 0.35714286 0.72234645 0.34496949 MAarrowhead1
% End of user PostScript
.63305 .65668 m
.63605 .65514 L
s
% Start of user PostScript
0.71428571 0.71428571 0.73040719 0.70606047 MAarrowhead1
% End of user PostScript
.63305 .7232 m
.63756 .7224 L
s
% Start of user PostScript
0.71428571 1.0714286 0.73846793 1.0671514 MAarrowhead1
% End of user PostScript
.63305 .78973 m
.63906 .78967 L
s
% Start of user PostScript
0.71428571 1.4285714 0.74652867 1.4282424 MAarrowhead1
% End of user PostScript
.63305 .85625 m
.64056 .85693 L
s
% Start of user PostScript
0.71428571 1.7857143 0.75458941 1.7893334 MAarrowhead1
% End of user PostScript
.63305 .92278 m
.64206 .92419 L
s
% Start of user PostScript
0.71428571 2.1428571 0.76265015 2.1504244 MAarrowhead1
% End of user PostScript
.63305 .9893 m
.64356 .99145 L
s
% Start of user PostScript
0.71428571 2.5 0.77071089 2.5115153 MAarrowhead1
% End of user PostScript
.69958 .05794 m
.68907 .05499 L
s
% Start of user PostScript
1.0714286 -2.5 1.0150034 -2.5158336 MAarrowhead1
% End of user PostScript
.69958 .12447 m
.69057 .1213 L
s
% Start of user PostScript
1.0714286 -2.1428571 1.0230641 -2.1598834 MAarrowhead1
% End of user PostScript
.69958 .191 m
.69207 .1876 L
s
% Start of user PostScript
1.0714286 -1.7857143 1.0311249 -1.8039332 MAarrowhead1
% End of user PostScript
.69958 .25752 m
.69357 .25391 L
s
% Start of user PostScript
1.0714286 -1.4285714 1.0391856 -1.447983 MAarrowhead1
% End of user PostScript
.69958 .32405 m
.69507 .32021 L
s
% Start of user PostScript
1.0714286 -1.0714286 1.0472464 -1.0920328 MAarrowhead1
% End of user PostScript
.69958 .39057 m
.69657 .38651 L
s
% Start of user PostScript
1.0714286 -0.71428571 1.0553071 -0.73608261 MAarrowhead1
% End of user PostScript
.69958 .4571 m
.69808 .45282 L
s
% Start of user PostScript
1.0714286 -0.35714286 1.0633678 -0.38013242 MAarrowhead1
% End of user PostScript
.69958 .52362 m
.69958 .51912 L
s
% Start of user PostScript
1.0714286 -3.3306691e-16 1.0714286 -0.02418222 MAarrowhead1
% End of user PostScript
.69958 .59015 m
.70108 .58542 L
s
% Start of user PostScript
1.0714286 0.35714286 1.0794893 0.33176798 MAarrowhead1
% End of user PostScript
.69958 .65668 m
.70258 .65173 L
s
% Start of user PostScript
1.0714286 0.71428571 1.0875501 0.68771817 MAarrowhead1
% End of user PostScript
.69958 .7232 m
.70408 .71803 L
s
% Start of user PostScript
1.0714286 1.0714286 1.0956108 1.0436684 MAarrowhead1
% End of user PostScript
.69958 .78973 m
.70558 .78433 L
s
% Start of user PostScript
1.0714286 1.4285714 1.1036715 1.3996186 MAarrowhead1
% End of user PostScript
.69958 .85625 m
.70708 .85064 L
s
% Start of user PostScript
1.0714286 1.7857143 1.1117323 1.7555688 MAarrowhead1
% End of user PostScript
.69958 .92278 m
.70859 .91694 L
s
% Start of user PostScript
1.0714286 2.1428571 1.119793 2.111519 MAarrowhead1
% End of user PostScript
.69958 .9893 m
.71009 .98324 L
s
% Start of user PostScript
1.0714286 2.5 1.1278538 2.4674692 MAarrowhead1
% End of user PostScript
.7661 .05794 m
.75559 .06288 L
s
% Start of user PostScript
1.4285714 -2.5 1.3721462 -2.4735147 MAarrowhead1
% End of user PostScript
.7661 .12447 m
.75709 .12784 L
s
% Start of user PostScript
1.4285714 -2.1428571 1.380207 -2.1247616 MAarrowhead1
% End of user PostScript
.7661 .191 m
.7586 .1928 L
s
% Start of user PostScript
1.4285714 -1.7857143 1.3882677 -1.7760085 MAarrowhead1
% End of user PostScript
.7661 .25752 m
.7601 .25777 L
s
% Start of user PostScript
1.4285714 -1.4285714 1.3963285 -1.4272554 MAarrowhead1
% End of user PostScript
.7661 .32405 m
.7616 .32273 L
s
% Start of user PostScript
1.4285714 -1.0714286 1.4043892 -1.0785023 MAarrowhead1
% End of user PostScript
.7661 .39057 m
.7631 .38769 L
s
% Start of user PostScript
1.4285714 -0.71428571 1.4124499 -0.72974917 MAarrowhead1
% End of user PostScript
.7661 .4571 m
.7646 .45266 L
s
% Start of user PostScript
1.4285714 -0.35714286 1.4205107 -0.38099607 MAarrowhead1
% End of user PostScript
.7661 .52362 m
.7661 .51762 L
s
% Start of user PostScript
1.4285714 -3.3306691e-16 1.4285714 -0.032242959 MAarrowhead1
% End of user PostScript
.7661 .59015 m
.7676 .58258 L
s
% Start of user PostScript
1.4285714 0.35714286 1.4366322 0.31651015 MAarrowhead1
% End of user PostScript
.7661 .65668 m
.76911 .64754 L
s
% Start of user PostScript
1.4285714 0.71428571 1.4446929 0.66526326 MAarrowhead1
% End of user PostScript
.7661 .7232 m
.77061 .71251 L
s
% Start of user PostScript
1.4285714 1.0714286 1.4527536 1.0140164 MAarrowhead1
% End of user PostScript
.7661 .78973 m
.77211 .77747 L
s
% Start of user PostScript
1.4285714 1.4285714 1.4608144 1.3627695 MAarrowhead1
% End of user PostScript
.7661 .85625 m
.77361 .84243 L
s
% Start of user PostScript
1.4285714 1.7857143 1.4688751 1.7115226 MAarrowhead1
% End of user PostScript
.7661 .92278 m
.77511 .9074 L
s
% Start of user PostScript
1.4285714 2.1428571 1.4769359 2.0602757 MAarrowhead1
% End of user PostScript
.7661 .9893 m
.77661 .97236 L
s
% Start of user PostScript
1.4285714 2.5 1.4849966 2.4090288 MAarrowhead1
% End of user PostScript
.83263 .05794 m
.82212 .07344 L
s
% Start of user PostScript
1.7857143 -2.5 1.7292891 -2.4168016 MAarrowhead1
% End of user PostScript
.83263 .12447 m
.82362 .13668 L
s
% Start of user PostScript
1.7857143 -2.1428571 1.7373498 -2.0773019 MAarrowhead1
% End of user PostScript
.83263 .191 m
.82512 .19992 L
s
% Start of user PostScript
1.7857143 -1.7857143 1.7454106 -1.7378022 MAarrowhead1
% End of user PostScript
.83263 .25752 m
.82662 .26316 L
s
% Start of user PostScript
1.7857143 -1.4285714 1.7534713 -1.3983025 MAarrowhead1
% End of user PostScript
.83263 .32405 m
.82812 .3264 L
s
% Start of user PostScript
1.7857143 -1.0714286 1.7615321 -1.0588028 MAarrowhead1
% End of user PostScript
.83263 .39057 m
.82963 .38964 L
s
% Start of user PostScript
1.7857143 -0.71428571 1.7695928 -0.71930311 MAarrowhead1
% End of user PostScript
.83263 .4571 m
.83113 .45288 L
s
% Start of user PostScript
1.7857143 -0.35714286 1.7776535 -0.37980341 MAarrowhead1
% End of user PostScript
.83263 .52362 m
.83263 .51612 L
s
% Start of user PostScript
1.7857143 -3.3306691e-16 1.7857143 -0.040303699 MAarrowhead1
% End of user PostScript
.83263 .59015 m
.83413 .57936 L
s
% Start of user PostScript
1.7857143 0.35714286 1.793775 0.29919601 MAarrowhead1
% End of user PostScript
.83263 .65668 m
.83563 .6426 L
s
% Start of user PostScript
1.7857143 0.71428571 1.8018358 0.63869572 MAarrowhead1
% End of user PostScript
.83263 .7232 m
.83713 .70583 L
s
% Start of user PostScript
1.7857143 1.0714286 1.8098965 0.97819542 MAarrowhead1
% End of user PostScript
.83263 .78973 m
.83863 .76907 L
s
% Start of user PostScript
1.7857143 1.4285714 1.8179572 1.3176951 MAarrowhead1
% End of user PostScript
.83263 .85625 m
.84014 .83231 L
s
% Start of user PostScript
1.7857143 1.7857143 1.826018 1.6571948 MAarrowhead1
% End of user PostScript
.83263 .92278 m
.84164 .89555 L
s
% Start of user PostScript
1.7857143 2.1428571 1.8340787 1.9966945 MAarrowhead1
% End of user PostScript
.83263 .9893 m
.84314 .95879 L
s
% Start of user PostScript
1.7857143 2.5 1.8421395 2.3361943 MAarrowhead1
% End of user PostScript
.89915 .05794 m
.88864 .08669 L
s
% Start of user PostScript
2.1428571 -2.5 2.086432 -2.3456944 MAarrowhead1
% End of user PostScript
.89915 .12447 m
.89015 .14782 L
s
% Start of user PostScript
2.1428571 -2.1428571 2.0944927 -2.0175044 MAarrowhead1
% End of user PostScript
.89915 .191 m
.89165 .20895 L
s
% Start of user PostScript
2.1428571 -1.7857143 2.1025534 -1.6893144 MAarrowhead1
% End of user PostScript
.89915 .25752 m
.89315 .27008 L
s
% Start of user PostScript
2.1428571 -1.4285714 2.1106142 -1.3611244 MAarrowhead1
% End of user PostScript
.89915 .32405 m
.89465 .33122 L
s
% Start of user PostScript
2.1428571 -1.0714286 2.1186749 -1.0329344 MAarrowhead1
% End of user PostScript
.89915 .39057 m
.89615 .39235 L
s
% Start of user PostScript
2.1428571 -0.71428571 2.1267357 -0.70474443 MAarrowhead1
% End of user PostScript
.89915 .4571 m
.89765 .45348 L
s
% Start of user PostScript
2.1428571 -0.35714286 2.1347964 -0.37655443 MAarrowhead1
% End of user PostScript
.89915 .52362 m
.89915 .51462 L
s
% Start of user PostScript
2.1428571 -3.3306691e-16 2.1428571 -0.048364439 MAarrowhead1
% End of user PostScript
.89915 .59015 m
.90066 .57575 L
s
% Start of user PostScript
2.1428571 0.35714286 2.1509179 0.27982556 MAarrowhead1
% End of user PostScript
.89915 .65668 m
.90216 .63688 L
s
% Start of user PostScript
2.1428571 0.71428571 2.1589786 0.60801555 MAarrowhead1
% End of user PostScript
.89915 .7232 m
.90366 .69801 L
s
% Start of user PostScript
2.1428571 1.0714286 2.1670394 0.93620555 MAarrowhead1
% End of user PostScript
.89915 .78973 m
.90516 .75915 L
s
% Start of user PostScript
2.1428571 1.4285714 2.1751001 1.2643955 MAarrowhead1
% End of user PostScript
.89915 .85625 m
.90666 .82028 L
s
% Start of user PostScript
2.1428571 1.7857143 2.1831608 1.5925855 MAarrowhead1
% End of user PostScript
.89915 .92278 m
.90816 .88141 L
s
% Start of user PostScript
2.1428571 2.1428571 2.1912216 1.9207755 MAarrowhead1
% End of user PostScript
.89915 .9893 m
.90966 .94254 L
s
% Start of user PostScript
2.1428571 2.5 2.1992823 2.2489655 MAarrowhead1
% End of user PostScript
.96568 .05794 m
.95517 .10261 L
s
% Start of user PostScript
2.5 -2.5 2.4435748 -2.260193 MAarrowhead1
% End of user PostScript
.96568 .12447 m
.95667 .16126 L
s
% Start of user PostScript
2.5 -2.1428571 2.4516356 -1.945369 MAarrowhead1
% End of user PostScript
.96568 .191 m
.95817 .2199 L
s
% Start of user PostScript
2.5 -1.7857143 2.4596963 -1.630545 MAarrowhead1
% End of user PostScript
.96568 .25752 m
.95967 .27854 L
s
% Start of user PostScript
2.5 -1.4285714 2.467757 -1.3157211 MAarrowhead1
% End of user PostScript
.96568 .32405 m
.96118 .33718 L
s
% Start of user PostScript
2.5 -1.0714286 2.4758178 -1.0008971 MAarrowhead1
% End of user PostScript
.96568 .39057 m
.96268 .39583 L
s
% Start of user PostScript
2.5 -0.71428571 2.4838785 -0.68607312 MAarrowhead1
% End of user PostScript
.96568 .4571 m
.96418 .45447 L
s
% Start of user PostScript
2.5 -0.35714286 2.4919393 -0.37124915 MAarrowhead1
% End of user PostScript
.96568 .52362 m
.96568 .51311 L
s
% Start of user PostScript
2.5 -3.3306691e-16 2.5 -0.056425179 MAarrowhead1
% End of user PostScript
.96568 .59015 m
.96718 .57176 L
s
% Start of user PostScript
2.5 0.35714286 2.5080607 0.25839879 MAarrowhead1
% End of user PostScript
.96568 .65668 m
.96868 .6304 L
s
% Start of user PostScript
2.5 0.71428571 2.5161215 0.57322277 MAarrowhead1
% End of user PostScript
.96568 .7232 m
.97018 .68904 L
s
% Start of user PostScript
2.5 1.0714286 2.5241822 0.88804674 MAarrowhead1
% End of user PostScript
.96568 .78973 m
.97169 .74769 L
s
% Start of user PostScript
2.5 1.4285714 2.532243 1.2028707 MAarrowhead1
% End of user PostScript
.96568 .85625 m
.97319 .80633 L
s
% Start of user PostScript
2.5 1.7857143 2.5403037 1.5176947 MAarrowhead1
% End of user PostScript
.96568 .92278 m
.97469 .86497 L
s
% Start of user PostScript
2.5 2.1428571 2.5483644 1.8325187 MAarrowhead1
% End of user PostScript
.96568 .9893 m
.97619 .92361 L
s
% Start of user PostScript
2.5 2.5 2.5564252 2.1473426 MAarrowhead1
% End of user PostScript
.5 Mabswid
.51863 .52362 m
.5186 .52268 L
.51854 .52179 L
.51842 .52071 L
.51825 .51964 L
.51782 .51764 L
.51714 .51531 L
.51474 .50949 L
.51136 .50351 L
.50689 .49728 L
.50056 .49014 L
.49238 .48271 L
.48313 .47603 L
.47772 .47283 L
.47144 .46971 L
.46535 .46728 L
.45962 .46554 L
.45604 .46472 L
.45263 .46415 L
.45074 .46392 L
.449 .46376 L
.44708 .46365 L
.44505 .46362 L
.44342 .46365 L
.44168 .46374 L
.43855 .46408 L
.43664 .46439 L
.43485 .46476 L
.43153 .46566 L
.4287 .46667 L
.42565 .46803 L
.41974 .47162 L
.41467 .47597 L
.41035 .48096 L
.40811 .4842 L
.40596 .48792 L
.40266 .49532 L
.40117 .49986 L
.40002 .50433 L
.39905 .50951 L
.39863 .51253 L
.39835 .51535 L
.39805 .52024 L
.398 .52499 L
.39821 .53038 L
.39875 .53636 L
.39914 .53944 L
.39967 .54284 L
.4009 .54919 L
.40417 .56161 L
Mistroke
.40845 .57392 L
.41435 .58766 L
.42869 .61411 L
.45436 .6512 L
.48692 .68955 L
.52358 .72514 L
.54549 .74268 L
.55829 .75152 L
.57084 .75906 L
.58325 .76534 L
.58991 .76819 L
.59705 .77082 L
.60983 .77434 L
.62177 .77617 L
.63376 .77646 L
.64497 .7752 L
.65717 .772 L
.67007 .76635 L
.68282 .75824 L
.69617 .74671 L
.70931 .73194 L
.72065 .71601 L
.73855 .68337 L
.7538 .64529 L
.76541 .60452 L
.76992 .58241 L
.77157 .57224 L
.77284 .56297 L
.77447 .54681 L
.77499 .53859 L
.77531 .53006 L
.77537 .52143 L
.77519 .51353 L
.77424 .49947 L
.77271 .48737 L
.77042 .47528 L
.76797 .46555 L
.76475 .45536 L
.75582 .43439 L
.74535 .4162 L
.72104 .38403 L
.68704 .34682 L
.58974 .24371 L
.52913 .17805 L
.49007 .13817 L
.44927 .10165 L
.42965 .08688 L
.40817 .07342 L
.38395 .06229 L
.36118 .05645 L
Mistroke
.34066 .05563 L
.3215 .05918 L
.30076 .06829 L
.27907 .0844 L
.26659 .09703 L
.2552 .11091 L
.23501 .14147 L
.21517 .17989 L
.19621 .22568 L
.16904 .31131 L
.1519 .38444 L
.14559 .41882 L
.14069 .45174 L
.13868 .46877 L
.13729 .48315 L
.13631 .49653 L
.13572 .50826 L
.13544 .51952 L
.13549 .53052 L
.13636 .54802 L
.13723 .55661 L
.13828 .56403 L
.13966 .57152 L
.14141 .5789 L
.14462 .58927 L
.14808 .59774 L
.1527 .60662 L
.15732 .61379 L
.16788 .62655 L
.18059 .63839 L
.23464 .67742 L
.40892 .83127 L
.48556 .91266 L
.53549 .96092 L
.56196 .98267 L
.58719 .9998 L
.61102 1.01196 L
.6341 1.01927 L
.65869 1.0214 L
.6854 1.01601 L
.71221 1.00126 L
.72675 .98884 L
.73954 .97508 L
.76323 .94191 L
.78368 .90427 L
.81808 .81815 L
.84204 .73552 L
.85971 .65394 L
.86614 .61483 L
.87013 .58372 L
Mistroke
.87155 .56948 L
.87266 .55551 L
.87371 .53318 L
.87366 .51269 L
.87315 .50248 L
.8724 .49377 L
.87044 .4799 L
.86804 .46898 L
.86467 .45812 L
.86094 .44922 L
.85338 .43617 L
.84411 .42475 L
.82487 .40773 L
.77067 .37121 L
.69365 .31382 L
.57077 .19336 L
.49124 .11027 L
.46722 .08776 L
.44009 .06516 L
.4138 .04706 L
.38806 .03391 L
.36413 .0266 L
.33784 .02504 L
.31404 .0304 L
.29239 .04167 L
.26925 .06133 L
.24504 .09144 L
.22172 .13099 L
.20202 .17372 L
.17084 .26307 L
.14998 .34386 L
.13537 .42035 L
.13041 .45608 L
.12759 .48413 L
.1267 .49692 L
.12611 .50941 L
.1259 .52928 L
.12673 .54747 L
.12763 .55651 L
.12871 .56422 L
.13121 .57655 L
.13402 .58627 L
.13781 .59599 L
.14189 .60402 L
.14997 .61597 L
.1597 .62663 L
.17962 .64301 L
.23527 .68007 L
.31483 .7407 L
.44286 .86851 L
Mistroke
.52596 .95333 L
.57464 .99279 L
.58868 1.0017 L
.60399 1.00985 L
.61865 1.01593 L
.63214 1.01988 L
.65801 1.02248 L
.68188 1.01829 L
.70765 1.00558 L
.72034 .99583 L
.734 .9825 L
.75584 .95454 L
.77765 .91751 L
.81158 .83827 L
.83885 .74944 L
.85008 .70315 L
.8584 .66284 L
.86435 .62874 L
.86902 .59591 L
.87185 .56973 L
.87342 .54855 L
.87414 .52667 L
.87384 .50914 L
.87261 .4927 L
.87154 .4846 L
.87037 .47782 L
.86748 .46576 L
.86556 .45974 L
.86363 .4546 L
.85975 .4462 L
.85509 .43826 L
.84543 .42588 L
.83527 .41599 L
.80872 .39612 L
.7407 .35048 L
.65075 .27526 L
.52617 .14579 L
.48428 .10352 L
.44065 .06556 L
.41502 .04777 L
.40115 .03994 L
.38622 .03314 L
.35945 .02577 L
.33454 .02535 L
.3074 .03314 L
.29243 .04161 L
.27899 .052 L
.25393 .07914 L
.23169 .11269 L
.20832 .15901 L
Mistroke
.18927 .20661 L
.17143 .26103 L
.15619 .31732 L
.14642 .36028 L
.13819 .40319 L
.13271 .43821 L
.12925 .46633 L
.12778 .48168 L
.1268 .49508 L
.12615 .50808 L
.12586 .52099 L
.12593 .53126 L
.12631 .54126 L
.12769 .55708 L
.12966 .56956 L
.13241 .58112 L
.13554 .59055 L
.1388 .59815 L
.14759 .61285 L
.15255 .6191 L
.15835 .62531 L
.18007 .64333 L
.20699 .66149 L
.266 .70165 L
.35768 .7801 L
.42298 .84723 L
.51137 .93954 L
.55544 .97864 L
.58133 .9972 L
.60677 1.01114 L
.62139 1.01687 L
.63748 1.02098 L
.6522 1.02251 L
.66585 1.02185 L
.69259 1.01411 L
.70727 1.00584 L
.72031 .99585 L
.74443 .97022 L
.76523 .93979 L
.78621 .9002 L
.80593 .85356 L
.83303 .77083 L
.85397 .68512 L
.86127 .64711 L
.867 .61109 L
.86925 .5941 L
.87118 .57679 L
.87335 .54972 L
.87413 .52796 L
.87409 .51714 L
Mistroke
.87379 .50804 L
.87257 .49239 L
.87071 .47962 L
.86839 .46908 L
.86541 .4593 L
.86218 .45121 L
.85893 .44468 L
.84989 .43109 L
.84017 .42049 L
.81527 .4006 L
.7886 .38289 L
.71322 .32975 L
.62575 .25076 L
.56178 .18371 L
.48654 .10569 L
.43714 .06289 L
.42281 .05275 L
.40724 .04321 L
.39356 .03626 L
.3786 .03038 L
.35257 .02499 L
.32802 .02639 L
.31389 .03043 L
.29917 .03741 L
.27353 .05703 L
.25079 .0833 L
.22814 .11894 L
.19277 .19712 L
.17505 .24911 L
.16112 .29792 L
.14207 .38193 L
.13606 .41593 L
.13133 .44852 L
.12833 .47554 L
.12667 .49722 L
.12588 .51921 L
.12591 .5302 L
.12622 .53955 L
.12742 .55475 L
.12947 .56859 L
.13088 .57524 L
.13235 .58088 L
.13617 .59215 L
.14007 .60069 L
.14464 .60858 L
.15355 .62024 L
.16453 .63108 L
.27548 .70874 L
.36901 .79125 L
.5014 .92973 L
Mistroke
.54708 .97186 L
.57392 .9923 L
.60011 1.00795 L
.614 1.01419 L
.62913 1.01914 L
.64607 1.02214 L
.66163 1.02228 L
.67497 1.02021 L
.68944 1.0155 L
.70437 1.00772 L
.71799 .99782 L
.74323 .97173 L
.76552 .9393 L
.78398 .90486 L
.80207 .86347 L
.83478 .76455 L
.85524 .67895 L
.8624 .6406 L
.86742 .60808 L
.87062 .58214 L
.87286 .55726 L
.87401 .53396 L
.87415 .52246 L
.87399 .51291 L
.87298 .4964 L
.87131 .4831 L
.86916 .4722 L
.86639 .4622 L
.86366 .45467 L
.86032 .44733 L
.85177 .43353 L
.84266 .42295 L
.81829 .40274 L
.79244 .38539 L
.72014 .33516 L
.63304 .25804 L
.49568 .11461 L
.45098 .07378 L
.42459 .05394 L
.39853 .03862 L
.38356 .03212 L
.36702 .02718 L
.35187 .02494 L
.33778 .02502 L
.32324 .02747 L
.31016 .03192 L
.29524 .03978 L
.28144 .0499 L
.25788 .07414 L
.23455 .10783 L
Mistroke
.2162 .14201 L
.19798 .18366 L
.16525 .28259 L
.14456 .36928 L
.13743 .40767 L
.13195 .44382 L
.12878 .47088 L
.12669 .49688 L
.12609 .50975 L
.12586 .52059 L
.12594 .53151 L
.12635 .54199 L
.12774 .55753 L
.12888 .56528 L
.13013 .5719 L
.13307 .58332 L
.13636 .59263 L
.1397 .59998 L
.14368 .60707 L
.15379 .62051 L
.16432 .63089 L
.27596 .70911 L
.35879 .78118 L
.48949 .9177 L
.53969 .9656 L
.56647 .98702 L
.59236 1.00381 L
.61644 1.01513 L
.64257 1.02176 L
.65761 1.0225 L
.67175 1.0209 L
.69789 1.01147 L
.7224 .99402 L
.74705 .96683 L
.76678 .93718 L
.7867 .89915 L
.82363 .80238 L
.84809 .71194 L
.86146 .64604 L
.86914 .59493 L
.87164 .5721 L
.87334 .54983 L
.8741 .52956 L
.87399 .51303 L
.8733 .50008 L
.87204 .48806 L
.87059 .47896 L
.86864 .47006 L
.86617 .46151 L
.86315 .45343 L
Mistroke
.85686 .44107 L
.84986 .43105 L
.84145 .42173 L
.82432 .40714 L
.78646 .3815 L
Mfstroke
0 0 m
1 0 L
1 1.04745 L
0 1.04745 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{274.938, 287.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006L0001[b000`400?l00000o`00003oo`3ooonM
0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM
0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00Q@3oool@
00000?l0oooo203oool008<0oooo1@0000080?ooo`T00000o`3oool40?ooo`00P03oool4000001<0
oooo1P00003o0?ooo`00OP3oool3000001T0oooo1P00003l0?ooo`00O03oool3000001l0oooo1000
003j0?ooo`00N`3oool2000002<0oooo1000003h0?ooo`00N@3oool2000002L0oooo1000003f0?oo
o`00M`3oool2000002/0oooo1000003d0?ooo`00MP3oool2000002h0oooo1000003b0?ooo`00M03o
ool300000340oooo0`00000N0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0]@3o
ool007<0oooo0P0000080?ooo`030000003oool0oooo02X0oooo1@00000K0?ooo`8000006@3oool2
000001T0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`1L0?ooo`00LP3oool20000
00P0oooo0P00000^0?ooo`D000006@3oool4000001L0oooo0`00000I0?ooo`800000>03oool30000
05h0oooo001N0?ooo`030000003oool0oooo0100oooo0P0000080?ooo`@00000603oool500000180
oooo1@00000H0?ooo`D000005P3oool3000001P0oooo1000000F0?ooo`@000006`3oool5000001H0
oooo00<000000?ooo`3oool0A03oool000h0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo01L0oooo0`00000?0?ooo`800000203oool5000000X0oooo
4`00000C0?ooo`H000005P3oool6000001H0oooo1000000G0?ooo`D000005@3oool8000001D0oooo
1P00000F0?ooo`<000006@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo
000>0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0F0?ooo`@000003P3oool00`000000oooo0?ooo`070?ooo`L000000`3oool6000000/0oooo1`00
000F0?ooo`D000005P3oool4000001P0oooo1000000G0?ooo`D000005P3oool6000001<0oooo2@00
000D0?ooo`D00000603oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000>
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G
0?ooo`D00000303oool00`000000oooo0?ooo`070?ooo`H000000`3oool2000001<0oooo2000000D
0?ooo`<000005`3oool3000001T0oooo0`00000I0?ooo`<00000603oool4000001<0oooo1@00000I
0?ooo`@00000603oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00d0oooo000=0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0F0?oo
o`D00000303oool2000000P0oooo0`0000050?ooo`800000603oool2000000<0oooo0P00000B0?oo
o`@000005`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01P0oooo0`00000I0?oo
o`<00000<03oool4000001T0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0=0?oo
o`003@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool05P3oool4000000`0oooo0P00000?0?ooo`8000008@3oool200000140oooo00@000000?ooo`00
00000000CP3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo02h0oooo0`00000H0?oo
o`<000006`3oool00`000000oooo0?ooo`0=0?ooo`003@3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01H0oooo0P00000I0?ooo`<00000303oool2000000h0oooo0P00000U0?ooo`80
00004P3oool2000009/0oooo0P00000I0?ooo`D000006@3oool00`000000oooo0?ooo`0=0?ooo`00
3@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01H0oooo1@00000E0?ooo`800000
3P3oool00`000000oooo0?ooo`0;0?ooo`800000:@3oool200000140oooo0P00002I0?ooo`800000
6P3oool4000001T0oooo00<000000?ooo`3oool03P3oool000d0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`0F0?ooo`@000005P3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo00/0oooo00<000000?ooo`3oool0:`3oool200000100oooo0P00002H0?ooo`030000003o
ool0oooo01T0oooo1000000I0?ooo`030000003oool0oooo00h0oooo000=0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool05`3oool4000002@0oooo00<000000?ooo`3oool02P3oool2
00000300oooo0P00000@0?ooo`800000/@3oool4000001X0oooo00<000000?ooo`3oool03P3oool0
00`0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0F0?ooo`@00000903oool20000
00/0oooo00<000000?ooo`3oool0<P3oool2000000l0oooo0P00002`0?ooo`<000006`3oool00`00
0000oooo0?ooo`0>0?ooo`00303oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01H0
oooo0`00000T0?ooo`8000002P3oool2000003L0oooo0P00000>0?ooo`800000[`3oool2000001T0
oooo0`00000A0?ooo`00303oool00`000000oooo0?ooo`0E0?ooo`80000000<0oooo0000003oool0
5`3oool2000002@0oooo0P00000:0?ooo`030000003oool0oooo03T0oooo0P00000=0?ooo`800000
[P3oool2000001T0oooo1@00000?0?ooo`00303oool00`000000oooo0?ooo`0E0?ooo`D000005`3o
ool2000002@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0l0?ooo`030000003o
ool0oooo00/0oooo0P00002]0?ooo`030000003oool0oooo01P0oooo1000000@0?ooo`00303oool0
0`000000oooo0?ooo`0E0?ooo`D000005`3oool00`000000oooo0?ooo`0R0?ooo`8000002@3oool0
0`000000oooo0?ooo`0n0?ooo`030000003oool0oooo00/0oooo0P0000370?ooo`@00000403oool0
00/0oooo00<000000?ooo`3oool05P3oool4000003`0oooo0P0000090?ooo`030000003oool0oooo
0400oooo0P00000<0?ooo`800000aP3oool300000140oooo000;0?ooo`030000003oool0oooo01H0
oooo1000000k0?ooo`8000002P3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo00X0
oooo0P0000350?ooo`<000004@3oool000/0oooo00<000000?ooo`3oool05`3oool2000003`0oooo
0P0000090?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool02P3oool200000<@0oooo
0P00000B0?ooo`002`3oool00`000000oooo0?ooo`0G0?ooo`800000>`3oool2000000T0oooo00<0
00000?ooo`3oool0AP3oool2000000`0oooo0P0000320?ooo`8000004P3oool000/0oooo00<00000
0?ooo`3oool05`3oool2000003/0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`19
0?ooo`030000003oool0oooo00X0oooo0P0000310?ooo`030000003oool0oooo0140oooo00090?oo
o`8000006P3oool00`000000oooo0?ooo`0i0?ooo`800000203oool00`000000oooo0?ooo`1;0?oo
o`030000003oool0oooo00X0oooo0P0000300?ooo`030000003oool0oooo0140oooo00090?ooo`D0
00005`3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool0C@3oool00`000000oooo0?ooo`0:0?ooo`800000d`3oool000T0oooo1000001C0?ooo`030000
003oool0oooo00H0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0:0?ooo`800000
dP3oool000T0oooo1000001B0?ooo`8000001`3oool00`000000oooo0?ooo`1A0?ooo`8000002`3o
ool200000=40oooo00090?ooo`<00000D`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo05@0oooo00<000000?ooo`3oool02@3oool200000=00oooo00090?ooo`<00000DP3oool20000
00L0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo
00L0oooo00@000000?ooo`0000000000603oool00`000000oooo0?ooo`1F0?ooo`030000003oool0
oooo05/0oooo00090?ooo`800000D`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
01<0oooo0P00000`0?ooo`800000403oool00`000000oooo0?ooo`070?ooo`@00000603oool20000
01P0oooo0P00000j0?ooo`<000005`3oool00`000000oooo0?ooo`130?ooo`002@3oool200000580
oooo0P0000070?ooo`030000003oool0oooo0180oooo1000000`0?ooo`<000003`3oool00`000000
oooo0?ooo`060?ooo`D000005`3oool3000001P0oooo0`00000g0?ooo`@000005P3oool3000004@0
oooo00090?ooo`030000003oool0oooo0540oooo100000040?ooo`030000003oool0oooo0180oooo
1@00000D0?ooo`T00000503oool5000000d0oooo00<000000?ooo`3oool01P3oool5000001L0oooo
0`00000G0?ooo`H000005@3oool9000001D0oooo1P00000E0?ooo`@00000@`3oool000h0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01D0oooo1`00
00000`3oool000000?ooo`0C0?ooo`L00000503oool8000001D0oooo1P00000<0?ooo`030000003o
ool0oooo00H0oooo1P00000E0?ooo`D000005P3oool7000001@0oooo2000000C0?ooo`P00000503o
ool6000001L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0<0?ooo`003P3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0503oool7
00000080oooo00<000000?ooo`3oool04@3oool6000001P0oooo1P00000E0?ooo`D000003P3oool2
000000L0oooo1@00000F0?ooo`D000005P3oool5000001L0oooo1@00000C0?ooo`D00000603oool5
000001H0oooo00<000000?ooo`0000006`3oool00`000000oooo0?ooo`0<0?ooo`003P3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05@3oool60000
0080oooo00<000000?ooo`3oool04@3oool3000001h0oooo1000000F0?ooo`@00000403oool00`00
0000oooo0?ooo`040?ooo`<0000000<0oooo0000000000005@3oool4000001P0oooo0`00000J0?oo
o`<00000;`3oool4000001P0oooo0`00000J0?ooo`030000003oool0oooo00d0oooo000=0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool05P3oool3000001H0oooo0P000000103oool0
00000000000000020?ooo`030000003oool0oooo03D0oooo0P00000G0?ooo`8000004P3oool00`00
0000oooo0?ooo`040?ooo`040000003oool0oooo0?ooo`8000005@3oool00`000000oooo0?ooo`0H
0?ooo`<00000703oool00`000000oooo0?ooo`0]0?ooo`<000006@3oool5000001P0oooo00<00000
0?ooo`3oool03@3oool000d0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0F0?oo
o`D00000503oool5000000<0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00P0oooo0P00000`0?ooo`030000003oool0oooo04X0oooo0`00000J0?oo
o`@000006@3oool00`000000oooo0?ooo`0=0?ooo`003@3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01H0oooo1000000D0?ooo`80000000<0oooo0000000000000`3oool00`000000
oooo0?ooo`1U0?ooo`030000003oool0oooo00P0oooo0P00001l0?ooo`030000003oool0oooo01T0
oooo1000000J0?ooo`030000003oool0oooo00d0oooo000=0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool05P3oool4000001@0oooo0P0000000`3oool000000?ooo`030?ooo`030000
003oool0oooo06H0oooo00<000000?ooo`3oool0203oool2000009L0oooo0`00000J0?ooo`030000
003oool0oooo00h0oooo000=0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3o
ool3000001@0oooo0P0000060?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0203o
ool2000009H0oooo0`00000H0?ooo`D000003P3oool000d0oooo00<000000?ooo`3oool05@3oool3
000001T0oooo0`00000D0?ooo`8000001@3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0U03oool2000001T0oooo1000000?0?ooo`00303oool00`00
0000oooo0?ooo`0F0?ooo`D000005`3oool2000001@0oooo0P0000060?ooo`030000003oool0oooo
06/0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo
01P0oooo1000000?0?ooo`00303oool00`000000oooo0?ooo`0F0?ooo`@00000603oool2000001@0
oooo0P0000050?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`2]0?ooo`<00000403oool000`0oooo00<000000?ooo`3oool05P3oool4000001P0oooo
00<000000?ooo`3oool04P3oool2000000H0oooo00<000000?ooo`3oool0KP3oool00`000000oooo
0?ooo`070?ooo`800000[P3oool300000100oooo000<0?ooo`030000003oool0oooo01H0oooo0`00
000^0?ooo`8000001@3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo00L0oooo0P00
002]0?ooo`8000004@3oool000`0oooo00<000000?ooo`3oool05P3oool3000002h0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo00L0oooo0P00002/
0?ooo`8000004@3oool000`0oooo00<000000?ooo`3oool05P3oool2000002h0oooo0P0000050?oo
o`030000003oool0oooo0780oooo00<000000?ooo`3oool0203oool200000:/0oooo00<000000?oo
o`3oool0403oool000T0oooo0`00000I0?ooo`800000;P3oool2000000D0oooo00<000000?ooo`3o
ool0L`3oool00`000000oooo0?ooo`080?ooo`800000ZP3oool00`000000oooo0?ooo`0@0?ooo`00
2@3oool5000001L0oooo00<000000?ooo`3oool0;03oool2000000D0oooo00<000000?ooo`3oool0
M@3oool00`000000oooo0?ooo`080?ooo`800000_03oool000T0oooo1@00000G0?ooo`030000003o
ool0oooo02`0oooo0P0000050?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0203o
ool200000;/0oooo00090?ooo`@00000AP3oool2000000D0oooo00<000000?ooo`3oool0N03oool0
0`000000oooo0?ooo`080?ooo`800000^P3oool000T0oooo100000160?ooo`8000001@3oool00`00
0000oooo0?ooo`1i0?ooo`030000003oool0oooo00P0oooo0P00002i0?ooo`002P3oool2000004L0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo00P0
oooo0P00002h0?ooo`002P3oool2000004H0oooo0P0000050?ooo`030000003oool0oooo07`0oooo
00<000000?ooo`3oool0203oool200000;L0oooo000:0?ooo`800000AP3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo07h0oooo00<000000?ooo`3oool0203oool200000;H0oooo000:
0?ooo`030000003oool0oooo04@0oooo0P0000050?ooo`030000003oool0oooo0280oooo00<00000
0?ooo`3oool0FP3oool00`000000oooo0?ooo`080?ooo`800000L03oool00`000000oooo0?ooo`12
0?ooo`002P3oool00`000000oooo0?ooo`140?ooo`8000001@3oool00`000000oooo0?ooo`0Q0?oo
o`<00000B`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00L0oooo0`00000I0?oo
o`030000003oool0oooo03/0oooo00<000000?ooo`3oool0503oool3000004<0oooo001A0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`@0
0000;`3oool2000001X0oooo0P00000?0?ooo`030000003oool0oooo00L0oooo0`00000I0?ooo`80
0000>@3oool3000001H0oooo100000120?ooo`00D03oool2000000D0oooo00<000000?ooo`3oool0
1@3oool3000001T0oooo1P00000G0?ooo`D000004`3oool3000001P0oooo1000000>0?ooo`030000
003oool0oooo00H0oooo1000000H0?ooo`@000005P3oool5000001X0oooo1@00000E0?ooo`H00000
5@3oool00`000000oooo0?ooo`0Y0?ooo`003P3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool02@3oool2000000@0oooo00<000000?ooo`3oool01@3o
ool5000001L0oooo1P00000C0?ooo`X00000503oool5000001H0oooo1@00000=0?ooo`030000003o
ool0oooo00H0oooo1000000H0?ooo`D000005@3oool9000001<0oooo2000000D0?ooo`D000005P3o
ool3000001X0oooo00<000000?ooo`3oool0303oool000h0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00P0oooo0P0000050?ooo`030000003oool0
oooo00D0oooo1000000G0?ooo`D000005`3oool7000001D0oooo1`00000D0?ooo`H000003@3oool0
0`000000oooo0?ooo`060?ooo`D000005P3oool6000001H0oooo1P00000C0?ooo`T000004`3oool4
000001P0oooo1000000I0?ooo`030000003oool0oooo00`0oooo000>0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool05P3oool2000000/0oooo0P0000040?ooo`030000003oool0oooo
00H0oooo0`00000H0?ooo`<000006`3oool5000001H0oooo1@00000F0?ooo`@000003P3oool00`00
0000oooo0?ooo`070?ooo`D000005P3oool4000001P0oooo1@00000_0?ooo`<00000603oool40000
01T0oooo00<000000?ooo`3oool03@3oool000h0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0G0?ooo`D00000203oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00H0
oooo0P00000H0?ooo`800000803oool2000001L0oooo0`00000G0?ooo`<00000403oool00`000000
oooo0?ooo`060?ooo`D000005`3oool3000001T0oooo0`00000_0?ooo`<000006@3oool4000001T0
oooo00<000000?ooo`3oool03@3oool000d0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0G0?ooo`@00000203oool2000000@0oooo00<000000?ooo`3oool01P3oool3000005@0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo
00@000000?ooo`3oool000005`3oool2000001`0oooo00<000000?ooo`3oool0;@3oool00`000000
oooo0?ooo`0I0?ooo`<00000603oool3000000l0oooo000=0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool05`3oool4000000P0oooo0P0000040?ooo`030000003oool0oooo00H0oooo
0P0000230?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`1V0?ooo`8000006@3oool5000000d0oooo000=0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool05P3oool4000000P0oooo0P0000050?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0P`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo07`0oooo0P00
000J0?ooo`@000003P3oool000d0oooo00<000000?ooo`3oool05P3oool5000001H0oooo0`000009
0?ooo`800000103oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0N`3oool2000001X0oooo1000000>0?ooo`003@3oool00`000000oooo0?ooo`0F0?oo
o`@000005`3oool2000000X0oooo0P0000040?ooo`030000003oool0oooo08l0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo01T0oooo0`00000?0?oo
o`003@3oool00`000000oooo0?ooo`0F0?ooo`@000005`3oool2000000T0oooo0P0000040?ooo`03
0000003oool0oooo0940oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`2E0?ooo`<0
00003`3oool000d0oooo00<000000?ooo`3oool05P3oool3000001P0oooo00<000000?ooo`3oool0
203oool2000000@0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo09@0oooo0P00000@0?ooo`00303oool00`000000oooo0?ooo`0G0?ooo`<000008`3o
ool2000000<0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo09<0oooo0P00000@0?ooo`002P3oool3000001T0oooo0P00000S0?ooo`800000103oool0
0`000000oooo0?ooo`2E0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0TP3oool0
0`000000oooo0?ooo`0?0?ooo`002P3oool5000001L0oooo0P00000S0?ooo`8000000`3oool00`00
0000oooo0?ooo`2G0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0X`3oool000X0
oooo1@00000G0?ooo`030000003oool0oooo0280oooo0P0000030?ooo`030000003oool0oooo09P0
oooo00<000000?ooo`3oool0203oool200000:@0oooo000:0?ooo`@00000603oool00`000000oooo
0?ooo`0Q0?ooo`800000103oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo00P0oooo
0P00002S0?ooo`002P3oool4000003`0oooo0P0000030?ooo`030000003oool0oooo09X0oooo00<0
00000?ooo`3oool02@3oool200000:80oooo000;0?ooo`800000?@3oool2000000<0oooo00<00000
0?ooo`3oool0V`3oool00`000000oooo0?ooo`090?ooo`800000X@3oool000/0oooo0P00000l0?oo
o`800000103oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo00T0oooo0P00002P0?oo
o`002`3oool2000003`0oooo0P0000030?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3o
ool02@3oool2000009l0oooo000;0?ooo`030000003oool0oooo03/0oooo0P0000030?ooo`030000
003oool0oooo09l0oooo00<000000?ooo`3oool02@3oool2000009h0oooo000;0?ooo`030000003o
ool0oooo03/0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2P0?ooo`030000003o
ool0oooo00T0oooo0P00002M0?ooo`00B03oool2000000@0oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`090?ooo`800000W03oool004P0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`0]0?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool02@3oool20000
05L0oooo00<000000?ooo`3oool0@@3oool004P0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo01X0oooo0`00001e0?ooo`030000003oool0oooo00T0
oooo0P00001E0?ooo`<000005`3oool00`000000oooo0?ooo`0X0?ooo`00A`3oool2000000@0oooo
00<000000?ooo`3oool03`3oool3000001X0oooo1000000K0?ooo`800000;P3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02@3oool2000003`0oooo
0P00000F0?ooo`@000005P3oool3000002P0oooo00170?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0403oool5000001L0oooo1P00000H0?ooo`@000004P3oool2000001/0oooo0P00
000I0?ooo`8000003@3oool00`000000oooo0?ooo`090?ooo`<00000603oool2000001h0oooo1000
000E0?ooo`H000005@3oool4000002L0oooo00170?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool03`3oool5000001P0oooo1@00000F0?ooo`L000004`3oool3000001T0oooo1000000G
0?ooo`<000003@3oool00`000000oooo0?ooo`080?ooo`@00000603oool3000001T0oooo1`00000E
0?ooo`D000005@3oool4000002P0oooo000>0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool05P3oool3000000030?ooo`000000000000@0oooo00<000000?ooo`3oool03`3oool40000
01P0oooo1000000F0?ooo`T00000503oool5000001L0oooo1@00000F0?ooo`<000003@3oool00`00
0000oooo0?ooo`080?ooo`D000005`3oool5000001@0oooo2@00000D0?ooo`@000005`3oool40000
01H0oooo0P0000000`3oool000000?ooo`0=0?ooo`003P3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo01H0oooo1@0000040?ooo`030000003oool0oooo0100oooo0`00000I0?ooo`80
0000703oool5000001D0oooo1P00000E0?ooo`H000005@3oool5000000`0oooo00<000000?ooo`3o
ool0203oool6000001H0oooo1P00000F0?ooo`D00000503oool2000001T0oooo0`00000G0?ooo`D0
00003@3oool000h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0F0?ooo`D00000
103oool00`000000oooo0?ooo`0@0?ooo`<00000603oool00`000000oooo0?ooo`0g0?ooo`@00000
5`3oool4000001L0oooo1@00000<0?ooo`030000003oool0oooo00P0oooo1000000H0?ooo`@00000
;`3oool00`000000oooo0?ooo`0I0?ooo`800000603oool4000000h0oooo000=0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool05`3oool5000000@0oooo00<000000?ooo`3oool03`3o
ool3000005@0oooo0`00000G0?ooo`<00000603oool4000000h0oooo00<000000?ooo`3oool01`3o
ool5000001P0oooo0`00001:0?ooo`8000006@3oool4000000h0oooo000=0?ooo`030000003oool0
oooo01H0oooo1@00000G0?ooo`D000000`3oool00`000000oooo0?ooo`0@0?ooo`800000EP3oool0
0`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0000000800000603oool00`000000oooo0?ooo`19
0?ooo`8000006@3oool3000000l0oooo000=0?ooo`030000003oool0oooo01H0oooo1@00000G0?oo
o`@00000103oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo09`0oooo00<000000?oo
o`3oool02P3oool3000006<0oooo00<000000?ooo`3oool0603oool3000000l0oooo000=0?ooo`03
0000003oool0oooo01H0oooo1000000H0?ooo`@00000103oool00`000000oooo0?ooo`2`0?ooo`03
0000003oool0oooo00X0oooo0`00001m0?ooo`800000403oool000d0oooo00<000000?ooo`3oool0
5P3oool4000001P0oooo100000040?ooo`030000003oool0oooo0;40oooo00<000000?ooo`3oool0
2P3oool3000007`0oooo0P00000@0?ooo`002P3oool2000000030?ooo`000000oooo01L0oooo0`00
000I0?ooo`030000003oool0000000@0oooo00<000000?ooo`3oool0/`3oool00`000000oooo0?oo
o`0:0?ooo`<00000N`3oool00`000000oooo0?ooo`0?0?ooo`002P3oool5000001P0oooo0P00000J
0?ooo`800000103oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo00X0oooo0`00002<
0?ooo`002P3oool5000001P0oooo0P00000J0?ooo`800000103oool00`000000oooo0?ooo`2e0?oo
o`030000003oool0oooo00X0oooo0`00002;0?ooo`002P3oool4000001T0oooo00<000000?ooo`3o
ool06@3oool2000000<0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0;0?ooo`80
0000RP3oool000X0oooo1000000I0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo00/0oooo0P0000290?ooo`002`3o
ool2000003D0oooo0P0000040?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool02`3o
ool2000008P0oooo000;0?ooo`800000=@3oool2000000<0oooo00<000000?ooo`3oool0^`3oool0
0`000000oooo0?ooo`0;0?ooo`800000Q`3oool000/0oooo0P00000e0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`0;0?ooo`800000QP3oool000/0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0;d0
oooo00<000000?ooo`3oool0303oool2000008@0oooo000;0?ooo`030000003oool0oooo03<0oooo
0P0000040?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool0303oool2000008<0oooo
00110?ooo`8000000`3oool00`000000oooo0?ooo`300?ooo`030000003oool0oooo00`0oooo0P00
00220?ooo`00@@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0<40oooo00<00000
0?ooo`3oool0303oool200000840oooo00110?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool0`P3oool00`000000oooo0?ooo`0<0?ooo`800000P03oool00440oooo00D000000?ooo`3o
ool0oooo000000360?ooo`030000003oool0oooo00`0oooo0P00001o0?ooo`00@03oool2000000<0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo00`0
oooo0`00000k0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool09`3oool00400oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01X0oooo
1000002>0?ooo`030000003oool0oooo00`0oooo0`00000i0?ooo`8000005`3oool3000002L0oooo
00100?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool05`3oool3000001X0oooo1@00
000J0?ooo`800000LP3oool00`000000oooo0?ooo`0<0?ooo`<00000=`3oool4000001D0oooo1P00
000E0?ooo`8000003P3oool00400oooo00D000000?ooo`3oool0oooo0000000J0?ooo`D000005`3o
ool5000001T0oooo1000000^0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00P00
000e0?ooo`D000005@3oool5000001H0oooo1000000<0?ooo`00?`3oool200000080oooo0P00000J
0?ooo`@00000603oool3000001T0oooo1P00000C0?ooo`8000006P3oool2000001T0oooo0`00000I
0?ooo`8000002`3oool00`000000oooo0?ooo`0;0?ooo`80000000<0oooo0000000000005P3oool4
000001T0oooo1`00000D0?ooo`@000005`3oool3000000d0oooo000>0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool04`3oool010000000oooo0?ooo`3oool5000001H0oooo1000000H
0?ooo`<00000603oool9000001<0oooo1@00000F0?ooo`@00000603oool3000001P0oooo0`00000;
0?ooo`030000003oool0oooo00X0oooo1P00000F0?ooo`P000004`3oool6000001H0oooo0`00000H
0?ooo`<000003@3oool000h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0C0?oo
o`040000003oool0oooo0?ooo`@000005`3oool3000001T0oooo0P00000M0?ooo`D00000503oool7
000001@0oooo1@00000F0?ooo`@00000603oool3000000/0oooo00<000000?ooo`3oool02P3oool7
000001H0oooo1P00000B0?ooo`<000006@3oool3000001P0oooo0`00000>0?ooo`003P3oool00`00
0000oooo0?ooo`0F0?ooo`<000005@3oool010000000oooo0?ooo`3oool4000001L0oooo0`00000H
0?ooo`030000003oool0oooo03L0oooo1@00000F0?ooo`H000005@3oool5000001H0oooo1@00000:
0?ooo`030000003oool0oooo00X0oooo1`00000F0?ooo`@00000;`3oool2000001T0oooo0`00000>
0?ooo`003P3oool00`000000oooo0?ooo`0F0?ooo`D000004P3oool2000000<0oooo0`00000H0?oo
o`800000E03oool4000001L0oooo1000000F0?ooo`H000005P3oool5000000X0oooo00<000000?oo
o`3oool02P3oool4000000030?ooo`000000000001H0oooo0`00000_0?ooo`030000003oool0oooo
01P0oooo0P00000?0?ooo`003@3oool00`000000oooo0?ooo`0G0?ooo`@000004`3oool2000000<0
oooo0`00000H0?ooo`030000003oool0oooo05@0oooo0P00000H0?ooo`<00000603oool4000001L0
oooo1000000<0?ooo`030000003oool0oooo00T0oooo0`0000040?ooo`030000003oool0oooo01D0
oooo00<000000?ooo`3oool0B03oool2000000l0oooo000=0?ooo`030000003oool0oooo01L0oooo
1000000C0?ooo`030000003oool0oooo0080oooo0P00001a0?ooo`030000003oool0oooo01L0oooo
00<000000?ooo`3oool05`3oool2000001X0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00@0oooo0P00001P0?ooo`030000003oool0oooo00h0oooo
000;0?ooo`D000005`3oool3000001@0oooo00@000000?ooo`3oool0oooo0`00003B0?ooo`030000
003oool0oooo0100oooo0P00001`0?ooo`002`3oool5000001L0oooo0`00000C0?ooo`8000000`3o
ool200000=@0oooo00<000000?ooo`3oool0403oool2000006l0oooo000;0?ooo`@00000603oool2
000001@0oooo0P0000030?ooo`030000003oool0oooo0=@0oooo00<000000?ooo`3oool0403oool3
000006d0oooo000;0?ooo`@00000603oool2000001@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`3E0?ooo`030000003oool0oooo0140oooo0P00001/0?ooo`002`3oool3000001T0
oooo00<000000?ooo`3oool04`3oool01@000000oooo0?ooo`3oool000000=T0oooo00<000000?oo
o`3oool04@3oool3000006X0oooo000<0?ooo`800000;P3oool2000000<0oooo00<000000?ooo`3o
ool0f03oool00`000000oooo0?ooo`0A0?ooo`<00000J@3oool000`0oooo0P00000^0?ooo`800000
0`3oool00`000000oooo0?ooo`3I0?ooo`030000003oool0oooo0140oooo0`00001X0?ooo`00303o
ool00`000000oooo0?ooo`0]0?ooo`8000000`3oool00`000000oooo0?ooo`3I0?ooo`030000003o
ool0oooo01<0oooo0`00001V0?ooo`00303oool00`000000oooo0?ooo`0]0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0fP3oool00`000000oooo0?ooo`0C0?ooo`<00000I@3oool0
03`0oooo00D000000?ooo`3oool0oooo0000003N0?ooo`030000003oool0oooo01@0oooo0`00001S
0?ooo`00>`3oool2000000<0oooo00<000000?ooo`3oool0g@3oool00`000000oooo0?ooo`0D0?oo
o`<00000HP3oool003/0oooo0P0000030?ooo`030000003oool0oooo0=h0oooo00<000000?ooo`3o
ool05@3oool200000640oooo000k0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
g`3oool00`000000oooo0?ooo`0E0?ooo`<00000G`3oool003/0oooo00D000000?ooo`3oool0oooo
0000003S0?ooo`030000003oool0oooo01H0oooo0P00001N0?ooo`00>`3oool01@000000oooo0?oo
o`3oool000000>@0oooo00<000000?ooo`3oool05P3oool3000005`0oooo000k0?ooo`050000003o
ool0oooo0?ooo`000000?03oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo01L0oooo
0`00001=0?ooo`030000003oool0oooo00/0oooo000j0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0>P3oool3000001d0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?oo
o`0H0?ooo`<00000<`3oool00`000000oooo0?ooo`0E0?ooo`<000002`3oool003X0oooo00D00000
0?ooo`3oool0oooo0000000P0?ooo`8000006P3oool6000001X0oooo0P0000290?ooo`030000003o
ool0oooo01P0oooo0`00000a0?ooo`8000005P3oool5000000X0oooo000j0?ooo`050000003oool0
oooo0?ooo`000000803oool5000001L0oooo1@00000J0?ooo`<00000RP3oool00`000000oooo0?oo
o`0I0?ooo`<00000;P3oool4000001D0oooo1000000;0?ooo`00>P3oool01@000000oooo0?ooo`3o
ool0000000@0oooo1@00000G0?ooo`@00000603oool4000001T0oooo1P00001U0?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`3oool06@3oool3000002`0oooo1@00000E0?ooo`<00000303o
ool003T0oooo0P0000030?ooo`030000003oool0oooo0080oooo1000000H0?ooo`@00000603oool3
000001T0oooo1`00000^0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool06@3oool2
000001T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00X0oooo0P00000Z0?ooo`L00000503oool2000000d0oooo000>0?ooo`030000003oool0oooo
01H0oooo0`00000?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool40000
01L0oooo1000000H0?ooo`<000006@3oool6000001D0oooo1000000H0?ooo`800000603oool30000
01X0oooo0P00000H0?ooo`<000002@3oool00`000000oooo0?ooo`0=0?ooo`8000002`3oool30000
00/0oooo2@00000C0?ooo`H000005@3oool3000000d0oooo000>0?ooo`030000003oool0oooo01H0
oooo1@00000=0?ooo`050000003oool0oooo0?ooo`0000001@3oool3000001P0oooo0`00000I0?oo
o`8000006@3oool3000001X0oooo2000000C0?ooo`@000005`3oool3000001T0oooo0`00000H0?oo
o`<00000203oool00`000000oooo0?ooo`0=0?ooo`@000002`3oool3000000X0oooo2000000B0?oo
o`<000006@3oool2000000h0oooo000>0?ooo`030000003oool0oooo01H0oooo1@00000=0?ooo`05
0000003oool0oooo0?ooo`0000001@3oool3000001P0oooo0P00000J0?ooo`030000003oool0oooo
03L0oooo1P00000D0?ooo`D000005@3oool4000001P0oooo1000000G0?ooo`@00000203oool00`00
0000oooo0?ooo`0=0?ooo`D000002P3oool3000000/0oooo1@00000_0?ooo`030000003oool0oooo
00d0oooo000;0?ooo`D000005`3oool4000000d0oooo0P0000030?ooo`030000003oool0oooo00<0
oooo0P00000I0?ooo`800000E@3oool4000001D0oooo1P00000D0?ooo`H000005`3oool4000001L0
oooo1@0000070?ooo`030000003oool0oooo00d0oooo1P00000:0?ooo`<000002`3oool3000003l0
oooo000;0?ooo`D000005`3oool4000000d0oooo0P0000030?ooo`030000003oool0oooo00<0oooo
0P00000I0?ooo`030000003oool0oooo05D0oooo0`00000F0?ooo`@000005P3oool6000001H0oooo
1@00000F0?ooo`H000001`3oool00`000000oooo0?ooo`0=0?ooo`@00000303oool3000000`0oooo
00<000000?ooo`3oool0?@3oool000/0oooo1000000I0?ooo`8000003P3oool01@000000oooo0?oo
o`3oool0000000H0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0D0?ooo`<00000
5`3oool5000001P0oooo0`00000H0?ooo`@00000203oool00`000000oooo0?ooo`0=0?ooo`<00000
3`3oool2000004/0oooo000;0?ooo`@000006@3oool2000000h0oooo00D000000?ooo`3oool0oooo
000000060?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3oool05`3oool3000001T0oooo
00<000000?ooo`3oool0603oool2000000/0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`0?0?ooo`<00000B@3oool000/0oooo0`00000J0?ooo`030000003oool0oooo00d0oooo00D0
00000?ooo`3oool0oooo0000002]0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0
7`3oool3000004P0oooo000<0?ooo`8000006P3oool00`000000oooo0?ooo`0=0?ooo`050000003o
ool0oooo0?ooo`000000l`3oool00`000000oooo0?ooo`0P0?ooo`800000A`3oool000`0oooo0P00
000Y0?ooo`8000000`3oool00`000000oooo0?ooo`3a0?ooo`030000003oool0oooo0240oooo0P00
00160?ooo`00303oool00`000000oooo0?ooo`0X0?ooo`8000000P3oool00`000000oooo0?ooo`3c
0?ooo`030000003oool0oooo0240oooo0`0000140?ooo`00303oool00`000000oooo0?ooo`0X0?oo
o`050000003oool0oooo0?ooo`000000mP3oool00`000000oooo0?ooo`0R0?ooo`800000@`3oool0
03L0oooo00D000000?ooo`3oool0oooo0000003g0?ooo`030000003oool0oooo0280oooo0P000012
0?ooo`00=`3oool01@000000oooo0?ooo`3oool000000?L0oooo00<000000?ooo`3oool08`3oool2
00000440oooo000g0?ooo`050000003oool0oooo0?ooo`000000n03oool00`000000oooo0?ooo`0S
0?ooo`800000@03oool003L0oooo00D000000?ooo`3oool0oooo0000003h0?ooo`030000003oool0
oooo02@0oooo0P00000o0?ooo`00=`3oool010000000oooo0?ooo`00003j0?ooo`030000003oool0
oooo02@0oooo0P00000n0?ooo`00=P3oool200000080oooo00<000000?ooo`3oool0n03oool00`00
0000oooo0?ooo`0U0?ooo`030000003oool0oooo03`0oooo000f0?ooo`8000000P3oool00`000000
oooo0?ooo`3i0?ooo`030000003oool0oooo02@0oooo0P00000m0?ooo`00=P3oool01@000000oooo
0?ooo`3oool000000?/0oooo00<000000?ooo`3oool09@3oool2000003`0oooo000f0?ooo`050000
003oool0oooo0?ooo`000000o03oool00`000000oooo0?ooo`0U0?ooo`800000>`3oool003H0oooo
00D000000?ooo`3oool0oooo000000100?ooo`800000703oool00`000000oooo0?ooo`2K0?ooo`03
0000003oool0oooo02D0oooo0P00000k0?ooo`00=P3oool010000000oooo0?ooo`00000U0?ooo`80
00006P3oool5000001T0oooo0`00002L0?ooo`030000003oool0oooo02D0oooo0P00000j0?ooo`00
=P3oool010000000oooo0?ooo`00000U0?ooo`D000005`3oool4000001T0oooo1@00002K0?ooo`03
0000003oool0oooo02D0oooo0P00000j0?ooo`00=@3oool200000080oooo00<000000?ooo`3oool0
1`3oool5000001L0oooo1000000H0?ooo`@000006@3oool4000009d0oooo00<000000?ooo`3oool0
9@3oool2000003T0oooo000W0?ooo`800000303oool200000080oooo00<000000?ooo`3oool01`3o
ool5000001L0oooo1000000H0?ooo`<000006@3oool4000001d0oooo0P00001o0?ooo`030000003o
ool0oooo02H0oooo00<000000?ooo`3oool0=`3oool002L0oooo1@0000090?ooo`8000000P3oool0
0`000000oooo0?ooo`070?ooo`@00000603oool3000001P0oooo0`00000J0?ooo`<00000703oool4
000002h0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0:0?ooo`8000003@3oool0
0`000000oooo0?ooo`0X0?ooo`00303oool5000001H0oooo1@0000090?ooo`050000003oool0oooo
0?ooo`0000002@3oool4000001P0oooo0`00000H0?ooo`<000006P3oool2000001X0oooo1`00000C
0?ooo`030000003oool0oooo01P0oooo0P00000I0?ooo`8000006P3oool2000001T0oooo0P00000I
0?ooo`800000303oool00`000000oooo0?ooo`0;0?ooo`8000006@3oool00`000000oooo0?ooo`0=
0?ooo`00303oool5000001H0oooo1000000:0?ooo`050000003oool0oooo0?ooo`0000002@3oool3
000001T0oooo0P00000I0?ooo`8000006P3oool2000001T0oooo2P00000B0?ooo`<00000603oool3
000001P0oooo0P00000J0?ooo`8000006@3oool2000001T0oooo0P00000<0?ooo`800000303oool2
000001T0oooo0P00000>0?ooo`00303oool4000001L0oooo1000000:0?ooo`050000003oool0oooo
0?ooo`0000002P3oool2000001T0oooo0P00000I0?ooo`8000006P3oool00`000000oooo0?ooo`0M
0?ooo`D000004`3oool4000001L0oooo0`00000G0?ooo`<000006@3oool3000001P0oooo0`00000H
0?ooo`<000002`3oool2000000`0oooo0`00000H0?ooo`<000003@3oool000`0oooo1000000H0?oo
o`8000002`3oool01@000000oooo0?ooo`3oool0000000X0oooo0P00000I0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool0>03oool5000001D0oooo1@00000E0?ooo`<000005`3oool4
000001P0oooo0`00000H0?ooo`@000005`3oool4000000/0oooo00<000000?ooo`3oool02P3oool4
000001P0oooo0`00000<0?ooo`00303oool3000001T0oooo0P00000:0?ooo`8000000`3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0C`3oool40000
00D0oooo1000000A0?ooo`H00000503oool4000001L0oooo1000000G0?ooo`@00000603oool40000
01L0oooo1000000:0?ooo`030000003oool0oooo00/0oooo1000000G0?ooo`@000002`3oool000d0
oooo0P00000I0?ooo`8000002P3oool2000000<0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`1X0?ooo`8000003@3oool200000100oooo1000000E0?ooo`D000005P3oool4000001L0
oooo1000000H0?ooo`@000005`3oool4000000X0oooo0P00000<0?ooo`@000005`3oool5000000X0
oooo000=0?ooo`8000006@3oool00`000000oooo0?ooo`090?ooo`8000000`3oool00`000000oooo
0?ooo`1a0?ooo`8000004@3oool3000000d0oooo0`00000G0?ooo`D000005@3oool5000001H0oooo
1@00000G0?ooo`D000005P3oool5000000X0oooo00<000000?ooo`3oool02P3oool5000001H0oooo
1P0000090?ooo`003@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00T0oooo0P00
00030?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0503oool2000000`0oooo00<0
00000?ooo`3oool05P3oool4000001H0oooo1000000G0?ooo`<000006@3oool3000001P0oooo0`00
000<0?ooo`030000003oool0oooo00X0oooo1000000H0?ooo`@000002P3oool000d0oooo00<00000
0?ooo`3oool0903oool01@000000oooo0?ooo`3oool000000740oooo0P00000I0?ooo`8000008`3o
ool2000001P0oooo0P00000f0?ooo`030000003oool0oooo02H0oooo0P00000;0?ooo`8000006P3o
ool2000000`0oooo000d0?ooo`050000003oool0oooo0?ooo`000000L03oool00`000000oooo0?oo
o`0K0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool09P3oool2000003D0oooo000d
0?ooo`050000003oool0oooo0?ooo`000000K`3oool00`000000oooo0?ooo`0M0?ooo`800000LP3o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo03<0oooo000d0?ooo`050000003oool0
oooo0?ooo`000000KP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0700oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`0c0?ooo`00=03oool01@000000oooo0?ooo`3o
ool0000006h0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`1_0?ooo`030000003o
ool0oooo02H0oooo00<000000?ooo`3oool0<`3oool003@0oooo00D000000?ooo`3oool0oooo0000
001^0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?oo
o`0V0?ooo`030000003oool0oooo03<0oooo000d0?ooo`050000003oool0oooo0?ooo`000000K@3o
ool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool09P3o
ool00`000000oooo0?ooo`0c0?ooo`00=03oool01@000000oooo0?ooo`3oool0000006d0oooo00<0
00000?ooo`3oool09@3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo02H0oooo0P00
000d0?ooo`00=03oool01@000000oooo0?ooo`3oool0000006`0oooo00<000000?ooo`3oool09`3o
ool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo02D0oooo0P00000d0?ooo`00=03oool0
1@000000oooo0?ooo`3oool0000006`0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`1[0?ooo`030000003oool0oooo02D0oooo0P00000d0?ooo`00=03oool01@000000oooo0?ooo`3o
ool0000006`0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`1Z0?ooo`030000003o
ool0oooo02D0oooo0P00000d0?ooo`00<`3oool2000000<0oooo00<000000?ooo`3oool0J@3oool0
0`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0?ooo`0b0?ooo`00<`3oool200000080oooo00<000000?ooo`3oool0JP3oool00`00
0000oooo0?ooo`090?ooo`D000007@3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo
02H0oooo00<000000?ooo`3oool0<P3oool003<0oooo0P0000020?ooo`030000003oool0oooo05X0
oooo1@00000;0?ooo`030000003oool0oooo00T0oooo1@00000M0?ooo`030000003oool0oooo06T0
oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0b0?ooo`00<`3oool01@000000oooo
0?ooo`3oool0000002H0oooo1@00000F0?ooo`D000005P3oool5000000/0oooo00<000000?ooo`3o
ool02@3oool5000001h0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0V0?ooo`03
0000003oool0oooo0380oooo000c0?ooo`050000003oool0oooo0?ooo`0000002P3oool5000001L0
oooo1@00000F0?ooo`D000005P3oool4000000/0oooo00<000000?ooo`3oool02`3oool3000001l0
oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0380
oooo000<0?ooo`D000005P3oool5000000L0oooo00D000000?ooo`3oool0oooo0000000:0?ooo`D0
00005`3oool4000001L0oooo1000000H0?ooo`<000002`3oool00`000000oooo0?ooo`0;0?ooo`<0
0000803oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000<0?ooo`D00000
5P3oool5000000L0oooo00D000000?ooo`3oool0oooo0000000:0?ooo`@000006@3oool3000001P0
oooo0`00000H0?ooo`<000002`3oool00`000000oooo0?ooo`0;0?ooo`<00000803oool00`000000
oooo0?ooo`1W0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0303oool000`0oooo
1000000G0?ooo`@00000203oool01@000000oooo0?ooo`3oool0000000/0oooo0`00000I0?ooo`<0
0000603oool3000001P0oooo0`00000;0?ooo`030000003oool0oooo00/0oooo0`00000Q0?ooo`03
0000003oool0oooo04L0oooo00<000000?ooo`3oool06@3oool010000000oooo0?ooo`00000G0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0H0?oo
o`8000003P3oool000d0oooo0`00000H0?ooo`<00000203oool01@000000oooo0?ooo`3oool00000
00/0oooo0`00000I0?ooo`<00000603oool3000001P0oooo0P00000<0?ooo`030000003oool0oooo
00`0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`040000003oool0oooo
000001H0oooo0P00000@0?ooo`030000003oool0oooo00H0oooo0`00000I0?ooo`<000003@3oool0
00d0oooo0`00000H0?ooo`<00000203oool01@000000oooo0?ooo`3oool0000000/0oooo0`00000I
0?ooo`8000006@3oool2000001X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool2000001X0oooo0P000002
0?ooo`030000003oool0oooo01@0oooo0`00000?0?ooo`030000003oool0oooo00H0oooo0`00000I
0?ooo`<000003@3oool000d0oooo0`00000H0?ooo`<00000203oool01@000000oooo0?ooo`3oool0
000000/0oooo0P00000K0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0<`3oool2
000001T0oooo0P00000I0?ooo`<000006@3oool3000000030?ooo`000000oooo01D0oooo0`00000?
0?ooo`030000003oool0oooo00H0oooo0`00000I0?ooo`<000003@3oool000d0oooo0P00000I0?oo
o`8000002@3oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0@P3oool3
000001P0oooo0`00000H0?ooo`<000006@3oool3000000030?ooo`000000oooo01D0oooo0`00000?
0?ooo`030000003oool0oooo00H0oooo0`00000H0?ooo`@000003@3oool000h0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`070?ooo`050000003oool0oooo0?ooo`000000303oool0
0`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0480oooo0`00000H0?ooo`<00000603oool3
000001T0oooo0`0000000`3oool000000?ooo`0D0?ooo`@000003`3oool00`000000oooo0?ooo`05
0?ooo`D000005`3oool5000000`0oooo000>0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool01`3oool01@000000oooo0?ooo`3oool0000006/0oooo00<000000?ooo`3oool0@@3oool4
000001P0oooo0`00000G0?ooo`@00000603oool4000000030?ooo`000000oooo01@0oooo1@00000>
0?ooo`030000003oool0oooo00D0oooo1@00000G0?ooo`D00000303oool003<0oooo00D000000?oo
o`3oool0oooo0000001/0?ooo`030000003oool0oooo0400oooo1000000G0?ooo`@000005`3oool5
000001L0oooo1P00000E0?ooo`D000003P3oool00`000000oooo0?ooo`050?ooo`D00000:03oool0
03<0oooo00D000000?ooo`3oool0oooo0000001/0?ooo`030000003oool0oooo0400oooo1@00000F
0?ooo`D000005P3oool5000001L0oooo1P00000X0?ooo`030000003oool0oooo0380oooo000c0?oo
o`8000000`3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo03l0oooo1P00000F0?oo
o`D00000=`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0380oooo000c0?ooo`80
00000`3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo03l0oooo1P00001B0?ooo`03
0000003oool0oooo02H0oooo00<000000?ooo`3oool0<P3oool003<0oooo0P0000030?ooo`030000
003oool0oooo06T0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`0U0?ooo`800000
=03oool003<0oooo0P0000030?ooo`030000003oool0oooo06T0oooo00<000000?ooo`3oool0U`3o
ool00`000000oooo0?ooo`0U0?ooo`800000=03oool003<0oooo0P0000030?ooo`030000003oool0
oooo06T0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`0U0?ooo`800000=03oool0
03@0oooo00D000000?ooo`3oool0oooo0000001[0?ooo`030000003oool0oooo09L0oooo00<00000
0?ooo`3oool09@3oool2000003@0oooo000d0?ooo`050000003oool0oooo0?ooo`000000K03oool0
0`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02H0oooo0P00000d0?ooo`00=03oool01@00
0000oooo0?ooo`3oool0000006`0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`0V
0?ooo`800000=03oool003@0oooo00D000000?ooo`3oool0oooo0000001]0?ooo`030000003oool0
oooo09@0oooo00<000000?ooo`3oool09P3oool2000003@0oooo000d0?ooo`8000000P3oool00`00
0000oooo0?ooo`1[0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool09P3oool20000
03@0oooo000d0?ooo`8000000P3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo09@0
oooo00<000000?ooo`3oool09P3oool2000003@0oooo000d0?ooo`8000000P3oool00`000000oooo
0?ooo`1/0?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool09P3oool2000003@0oooo
000d0?ooo`8000000`3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo09<0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`0c0?ooo`00303oool00`000000oooo0?ooo`0V
0?ooo`050000003oool0oooo0?ooo`000000KP3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo08X0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0T0?ooo`030000003o
ool0oooo00`0oooo000:0?ooo`<000006P3oool2000000`0oooo00D000000?ooo`3oool0oooo0000
00090?ooo`8000006P3oool2000001T0oooo0P00000I0?ooo`8000004@3oool00`000000oooo0?oo
o`030?ooo`<000006@3oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00`0oooo0009
0?ooo`@000005`3oool5000000`0oooo0P0000030?ooo`030000003oool0oooo00<0oooo1@00000G
0?ooo`D000005P3oool5000001H0oooo1@00000A0?ooo`030000003oool0oooo0080oooo1000000G
0?ooo`@00000L03oool00`000000oooo0?ooo`0V0?ooo`8000002`3oool00`000000oooo0?ooo`0G
0?ooo`8000003P3oool000X0oooo1000000G0?ooo`@00000303oool2000000<0oooo00<000000?oo
o`3oool0103oool4000001P0oooo1000000G0?ooo`@000005`3oool400000180oooo00<000000?oo
o`3oool00P3oool4000001D0oooo1@00001`0?ooo`030000003oool0oooo01P0oooo00<000000?oo
o`3oool02`3oool2000000X0oooo0P00000I0?ooo`<000003@3oool000/0oooo0`00000H0?ooo`@0
0000303oool01@000000oooo0?ooo`3oool0000000H0oooo1000000H0?ooo`@000005`3oool40000
01L0oooo1000000B0?ooo`030000003oool0oooo0080oooo1000000F0?ooo`D00000KP3oool20000
01T0oooo0P00000=0?ooo`8000002P3oool2000001T0oooo0`00000=0?ooo`002`3oool3000001T0
oooo0`00000<0?ooo`050000003oool0oooo0?ooo`0000001`3oool3000001T0oooo0`00000H0?oo
o`<00000603oool3000001<0oooo00<000000?ooo`3oool00P3oool3000001P0oooo0`00001C0?oo
o`030000003oool0oooo01P0oooo0P00000I0?ooo`8000003@3oool2000000T0oooo0`00000I0?oo
o`<000003@3oool000`0oooo0P00000I0?ooo`<00000303oool2000000<0oooo00<000000?ooo`3o
ool01@3oool3000001P0oooo0`00000I0?ooo`<00000603oool300000180oooo00<000000?ooo`3o
ool00`3oool2000001T0oooo0`00000g0?ooo`030000003oool0oooo01L0oooo0P00000I0?ooo`<0
0000603oool3000000d0oooo0P0000090?ooo`@000005`3oool4000000d0oooo000=0?ooo`800000
6@3oool2000000d0oooo00D000000?ooo`3oool0oooo000000080?ooo`8000006@3oool2000001X0
oooo0P00000I0?ooo`8000004`3oool00`000000oooo0?ooo`030?ooo`8000006@3oool2000001<0
oooo2@00000J0?ooo`8000006@3oool2000001T0oooo0P00000I0?ooo`<000003@3oool00`000000
oooo0?ooo`080?ooo`@000005`3oool5000000`0oooo000>0?ooo`030000003oool0oooo01P0oooo
0P00000<0?ooo`8000000P3oool00`000000oooo0?ooo`060?ooo`8000006@3oool2000001X0oooo
0P00000I0?ooo`8000004`3oool00`000000oooo0?ooo`030?ooo`8000006`3oool00`000000oooo
0?ooo`0@0?ooo`P000006P3oool2000001T0oooo0`00000H0?ooo`<00000603oool5000000`0oooo
00<000000?ooo`3oool01`3oool5000001L0oooo1@00000<0?ooo`00:P3oool00`000000oooo0?oo
o`0;0?ooo`040000003oool0oooo000000T0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo02`0oooo1P00000J0?ooo`@00000603oool4000001P0oooo
1000000G0?ooo`D00000303oool00`000000oooo0?ooo`070?ooo`D00000:03oool003P0oooo0P00
00020?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0<P3oool4000001/0oooo1000
000I0?ooo`<000006@3oool4000001H0oooo1P00000<0?ooo`030000003oool0oooo00T0oooo0`00
000X0?ooo`00>@3oool010000000oooo0?ooo`00001`0?ooo`030000003oool0oooo0340oooo0P00
000L0?ooo`D00000603oool4000001P0oooo1@00000F0?ooo`H00000303oool00`000000oooo0?oo
o`0d0?ooo`00>@3oool200000080oooo00<000000?ooo`3oool0K@3oool00`000000oooo0?ooo`1>
0?ooo`H00000603oool4000001P0oooo1@00000W0?ooo`800000=P3oool003X0oooo00@000000?oo
o`3oool00000L03oool00`000000oooo0?ooo`1>0?ooo`@00000603oool5000001X0oooo0`00000W
0?ooo`800000=P3oool003X0oooo0P0000020?ooo`030000003oool0oooo06d0oooo00<000000?oo
o`3oool0D03oool2000001X0oooo0`00000J0?ooo`030000003oool0oooo02L0oooo00<000000?oo
o`3oool0=@3oool003/0oooo0P0000020?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3o
ool0R03oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo03D0oooo000k0?ooo`800000
0P3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool0
:03oool00`000000oooo0?ooo`0e0?ooo`00?03oool200000080oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`260?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0=@3o
ool003d0oooo0P0000020?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0Q@3oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo03D0oooo000n0?ooo`050000003oool0oooo
0?ooo`000000KP3oool00`000000oooo0?ooo`250?ooo`030000003oool0oooo02L0oooo0P00000g
0?ooo`00?`3oool01@000000oooo0?ooo`3oool0000006h0oooo00<000000?ooo`3oool0P`3oool0
0`000000oooo0?ooo`0X0?ooo`800000=`3oool00400oooo00D000000?ooo`3oool0oooo0000001^
0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0f
0?ooo`00@@3oool01@000000oooo0?ooo`3oool0000006d0oooo00<000000?ooo`3oool0PP3oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo03H0oooo00120?ooo`050000003oool0oooo
0?ooo`000000K@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo02P0oooo00<00000
0?ooo`3oool0=P3oool004<0oooo0P0000020?ooo`030000003oool0oooo06/0oooo00<000000?oo
o`3oool0O`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02P0oooo00<000000?oo
o`3oool02`3oool004@0oooo0P0000020?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3o
ool0O`3oool00`000000oooo0?ooo`0X0?ooo`800000:P3oool00`000000oooo0?ooo`0;0?ooo`00
A@3oool200000080oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`1n0?ooo`030000
003oool0oooo02P0oooo0P00000Y0?ooo`8000003@3oool004L0oooo0P0000000`3oool000000?oo
o`1[0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0:03oool2000000l0oooo00<0
00000?ooo`3oool05`3oool2000000d0oooo00180?ooo`80000000<0oooo000000000000JP3oool0
0`000000oooo0?ooo`1l0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool03@3oool2
000001P0oooo0`00000=0?ooo`00B@3oool200000080oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0603oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0J@3oool00`00
0000oooo0?ooo`0N0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool03@3oool20000
01P0oooo1000000<0?ooo`00BP3oool3000000030?ooo`000000oooo00/0oooo0`00000I0?ooo`<0
00006@3oool3000001P0oooo0`0000080?ooo`030000003oool0oooo00`0oooo1000000G0?ooo`03
0000003oool0oooo0500oooo00<000000?ooo`3oool07@3oool2000000X0oooo00<000000?ooo`3o
ool0303oool3000001P0oooo1000000<0?ooo`00?P3oool00`000000oooo0?ooo`0;0?ooo`800000
00<0oooo0000000000002P3oool3000001P0oooo1@00000F0?ooo`H000005P3oool5000000L0oooo
00<000000?ooo`3oool02`3oool5000001H0oooo1000001@0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool05`3oool2000000T0oooo0P00000>0?ooo`<000005`3oool5000000`0oooo
000n0?ooo`<00000303oool3000000030?ooo`000000oooo00L0oooo1@00000H0?ooo`@000005`3o
ool5000001L0oooo100000080?ooo`030000003oool0oooo00/0oooo1@00000E0?ooo`D00000CP3o
ool00`000000oooo0?ooo`030?ooo`800000603oool3000000T0oooo0P00000=0?ooo`@000005`3o
ool5000000`0oooo000?0?ooo`030000003oool0oooo02/0oooo1P00000<0?ooo`80000000<0oooo
0000003oool01`3oool5000001P0oooo1000000G0?ooo`@00000603oool4000000P0oooo00<00000
0?ooo`3oool0303oool3000001@0oooo1`00000i0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool00`3oool2000001P0oooo0`0000090?ooo`8000003@3oool4000001T0oooo00<00000
0?ooo`3oool0303oool000h0oooo0P00000I0?ooo`8000004P3oool8000000/0oooo1@0000080?oo
o`@00000603oool3000001P0oooo1000000H0?ooo`<000002@3oool00`000000oooo0?ooo`0<0?oo
o`<000004`3oool9000001X0oooo0P00000J0?ooo`8000004P3oool00`000000oooo0?ooo`030?oo
o`<000005`3oool5000000P0oooo00<000000?ooo`3oool02`3oool5000001T0oooo00<000000?oo
o`3oool0303oool000d0oooo0P00000F0?ooo`D00000603oool5000000/0oooo0P0000000`3oool0
00000?ooo`080?ooo`8000006P3oool2000001T0oooo0P00000J0?ooo`8000002@3oool00`000000
oooo0?ooo`0<0?ooo`800000603oool5000001D0oooo1@00000J0?ooo`8000004`3oool00`000000
oooo0?ooo`020?ooo`<00000603oool5000000P0oooo00<000000?ooo`3oool03@3oool3000001T0
oooo00<000000?ooo`3oool0303oool000`0oooo0`00000C0?ooo`L00000:P3oool2000000030?oo
o`000000000000T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?ooo`800000
6P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
;03oool7000001T0oooo1000000B0?ooo`030000003oool0oooo0080oooo1000000G0?ooo`H00000
1`3oool2000003X0oooo000;0?ooo`@000004`3oool6000002d0oooo0P0000000`3oool000000?oo
o`0n0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0?03oool6000001T0oooo1000
000C0?ooo`030000003oool0oooo0080oooo1000000G0?ooo`H000001`3oool2000003X0oooo000:
0?ooo`D00000503oool4000002l0oooo0P0000000`3oool000000?ooo`1T0?ooo`030000003oool0
oooo03`0oooo1000000I0?ooo`D000004P3oool00`000000oooo0?ooo`020?ooo`D00000903oool2
000003X0oooo00090?ooo`D000005@3oool300000380oooo1000001T0?ooo`030000003oool0oooo
03/0oooo0`00000I0?ooo`H000004P3oool00`000000oooo0?ooo`040?ooo`<00000903oool20000
03X0oooo000:0?ooo`@000005P3oool00`000000oooo0?ooo`0b0?ooo`80000000<0oooo0000003o
ool0H`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo01T0oooo1000000C0?ooo`03
0000003oool0oooo02/0oooo00<000000?ooo`3oool0>@3oool000`0oooo0P00001=0?ooo`<00000
I03oool00`000000oooo0?ooo`1H0?ooo`8000004P3oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo03T0oooo001L0?ooo`@00000H`3oool00`000000oooo0?ooo`1[0?ooo`030000003o
ool0oooo02/0oooo0P00000k0?ooo`00GP3oool3000006<0oooo00<000000?ooo`3oool0J@3oool0
0`000000oooo0?ooo`0/0?ooo`800000>`3oool005l0oooo0`00001R0?ooo`030000003oool0oooo
06T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0j0?ooo`00H@3oool200000680
oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo03X0
oooo001R0?ooo`800000HP3oool00`000000oooo0?ooo`1V0?ooo`030000003oool0oooo02`0oooo
0P00000l0?ooo`00I03oool200000640oooo00<000000?ooo`3oool0I03oool00`000000oooo0?oo
o`0]0?ooo`800000?03oool006D0oooo0P00001Q0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool0;@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00X0oooo001V0?oo
o`800000H@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo02h0oooo00<000000?oo
o`3oool0;@3oool2000000`0oooo001X0?ooo`800000H03oool00`000000oooo0?ooo`1P0?ooo`03
0000003oool0oooo02h0oooo00<000000?ooo`3oool0;@3oool2000000`0oooo001Y0?ooo`800000
H03oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo02h0oooo0P00000^0?ooo`<00000
303oool006X0oooo0`00001O0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0;P3o
ool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01L0oooo0`00000<0?ooo`00J`3oool3
000005l0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool05P3oool5000000/0oooo001/0?ooo`<00000G`3oool00`00
0000oooo0?ooo`1J0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool04P3oool20000
01P0oooo1@00000;0?ooo`00K@3oool4000005h0oooo00<000000?ooo`3oool0F03oool00`000000
oooo0?ooo`0[0?ooo`040000003oool0oooo0?ooo`800000503oool2000001L0oooo1P00000;0?oo
o`00K`3oool3000005h0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`0Z0?ooo`80
00000`3oool2000001<0oooo0`00000G0?ooo`H000002`3oool000l0oooo00<000000?ooo`3oool0
GP3oool3000000D0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo0480oooo00<000000?ooo`3oool0:`3oool2000000<0oooo00<000000?ooo`3oool0
4P3oool4000001T0oooo0`00000;0?ooo`003P3oool2000004X0oooo00<000000?ooo`3oool0503o
ool3000000<0oooo0`00000I0?ooo`<00000603oool4000001P0oooo0`0000050?ooo`030000003o
ool0oooo00h0oooo0`0000130?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool05P3o
ool3000000<0oooo00<000000?ooo`3oool04P3oool4000001T0oooo00<000000?ooo`3oool02`3o
ool000h0oooo0P00000K0?ooo`030000003oool0oooo02`0oooo0`00000E0?ooo`040000003oool0
00000?ooo`@000005`3oool6000001H0oooo1@00000G0?ooo`@000001P3oool00`000000oooo0?oo
o`0=0?ooo`@00000@@3oool00`000000oooo0?ooo`0A0?ooo`800000603oool300000080oooo0P00
000C0?ooo`D00000603oool00`000000oooo0?ooo`0<0?ooo`003@3oool3000001X0oooo0P00000]
0?ooo`D000005@3oool00`000000oooo00000005000001L0oooo1@00000G0?ooo`D000005@3oool6
000000H0oooo00<000000?ooo`3oool02`3oool6000003T0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0A0?ooo`800000603oool400000080oooo0P00000C0?ooo`D00000603oool0
0`000000oooo0?ooo`0<0?ooo`003@3oool3000001T0oooo0P00000J0?ooo`8000004P3oool70000
01@0oooo00<000000?ooo`0000001@00000H0?ooo`@00000603oool3000001H0oooo1@0000070?oo
o`030000003oool0oooo00X0oooo1`00000D0?ooo`D000007@3oool2000000H0oooo00<000000?oo
o`3oool0403oool3000001P0oooo100000020?ooo`800000503oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo00`0oooo000<0?ooo`@00000603oool3000001H0oooo1@00000C0?ooo`T0
00004`3oool00`000000oooo00000005000001P0oooo0`00000I0?ooo`<000005`3oool4000000L0
oooo00<000000?ooo`3oool02`3oool6000001<0oooo2P00000G0?ooo`<000001@3oool00`000000
oooo0?ooo`0@0?ooo`@000005`3oool500000080oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000<0?ooo`@000005`3oool3000001@0oooo
1`00000I0?ooo`D000004`3oool7000001T0oooo0`00000I0?ooo`8000006@3oool2000000P0oooo
00<000000?ooo`3oool03P3oool300000180oooo2000000H0?ooo`<000001@3oool00`000000oooo
0?ooo`0@0?ooo`D000006@3oool3000000030?ooo`00000000000400oooo000;0?ooo`D000005P3o
ool4000001@0oooo1P00000d0?ooo`030000003oool0000000<000006P3oool00`000000oooo0?oo
o`0I0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`0S0?ooo`D00000603oool5000000@0oooo00<000000?ooo`3oool0403oool5000001h0oooo0P00
00100?ooo`003@3oool3000001D0oooo1@00000E0?ooo`@00000=P3oool010000000oooo00000000
001L0?ooo`8000008`3oool2000001X0oooo1@0000050?ooo`030000003oool0oooo0140oooo1000
000N0?ooo`800000@03oool002H0oooo0`00000F0?ooo`<00000>03oool010000000oooo00000000
001M0?ooo`030000003oool0oooo03d0oooo100000040?ooo`030000003oool0oooo01@0oooo0P00
000N0?ooo`030000003oool0oooo03l0oooo000X0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool0>03oool00`000000oooo00000002000005`0oooo00<000000?ooo`3oool0?@3oool2
000000@0oooo00<000000?ooo`3oool0=03oool200000440oooo001l0?ooo`030000003oool00000
00800000G03oool2000003h0oooo00D000000?ooo`3oool0oooo0000000g0?ooo`800000@@3oool0
07d0oooo00<000000?ooo`0000000P00001M0?ooo`030000003oool0oooo03h0oooo00<000000?oo
o`3oool0=P3oool200000440oooo001n0?ooo`030000003oool0000000800000G@3oool00`000000
oooo0?ooo`0m0?ooo`030000003oool0oooo03H0oooo0P0000110?ooo`00O`3oool2000000030?oo
o`000000000005d0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0f0?ooo`800000
@P3oool00840oooo00@000000?ooo`0000000000G@3oool2000003X0oooo00<000000?ooo`3oool0
=`3oool2000003D0oooo00<000000?ooo`3oool02P3oool00880oooo00@000000?ooo`0000000000
GP3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo03P0oooo0P00000e0?ooo`030000
003oool0oooo00X0oooo00230?ooo`040000003oool00000000005h0oooo00<000000?ooo`3oool0
=03oool00`000000oooo0?ooo`0i0?ooo`800000=03oool2000000`0oooo00240?ooo`040000003o
ool00000000005h0oooo0P00000b0?ooo`800000>`3oool2000003D0oooo0P00000<0?ooo`00Q@3o
ool010000000oooo00000000001O0?ooo`800000;`3oool00`000000oooo0?ooo`0k0?ooo`800000
=03oool3000000`0oooo00260?ooo`040000003oool0000000000600oooo0P00000/0?ooo`030000
003oool0oooo03`0oooo0P00000K0?ooo`030000003oool0oooo01H0oooo1000000;0?ooo`00Q`3o
ool010000000oooo0?ooo`00001Q0?ooo`800000:03oool2000003h0oooo0P00000K0?ooo`800000
603oool4000000/0oooo00280?ooo`80000000<0oooo0000003oool0H@3oool2000002@0oooo0P00
00100?ooo`8000006`3oool2000001L0oooo1@00000;0?ooo`00RP3oool00`000000oooo0000001S
0?ooo`800000803oool200000480oooo0P00000J0?ooo`<000005`3oool5000000/0oooo000@0?oo
o`030000003oool0oooo07P0oooo00<000000?ooo`000000I03oool2000001`0oooo0P0000140?oo
o`030000003oool0000001T0oooo0`00000I0?ooo`<000002`3oool000l0oooo0P00001k0?ooo`04
0000003oool00000000006@0oooo1000000E0?ooo`<00000A@3oool4000001P0oooo1000000I0?oo
o`030000003oool0oooo00/0oooo000?0?ooo`800000O03oool010000000oooo00000000001W0?oo
o`D000003P3oool2000004P0oooo1000000H0?ooo`@000006@3oool00`000000oooo0?ooo`0;0?oo
o`003P3oool3000001X0oooo00<000000?ooo`3oool0H03oool010000000oooo00000000001[0?oo
o`h00000<P3oool00`000000oooo0?ooo`0E0?ooo`@000005`3oool5000001T0oooo00<000000?oo
o`3oool02`3oool000h0oooo0`00000I0?ooo`800000C03oool00`000000oooo0?ooo`0D0?ooo`04
0000003oool0000000000080oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0E03oool2000001L0oooo1000000G0?oo
o`D00000603oool00`000000oooo0?ooo`0<0?ooo`003@3oool4000001T0oooo0P00001;0?ooo`<0
00005P3oool00`000000oooo00000004000001L0oooo1000000H0?ooo`<000005`3oool3000005D0
oooo0P00000F0?ooo`D000006@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00`0
oooo000=0?ooo`@00000603oool3000001/0oooo00<000000?ooo`3oool0;03oool5000001H0oooo
1P00000F0?ooo`D000005`3oool4000001L0oooo1000000F0?ooo`8000008P3oool00`000000oooo
0?ooo`0F0?ooo`8000005`3oool5000001P0oooo00<000000?ooo`3oool06@3oool00`000000oooo
0?ooo`0<0?ooo`00303oool5000001L0oooo1000000I0?ooo`8000004`3oool4000001P0oooo1P00
000F0?ooo`H000005P3oool5000001D0oooo1P00000E0?ooo`H000005@3oool6000001`0oooo0P00
000H0?ooo`<000005`3oool5000001P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0<0?ooo`003P3oool3000001H0oooo1000000H0?ooo`@000004`3oool900000180oooo2000000F
0?ooo`D00000603oool3000001H0oooo1P00000F0?ooo`D00000503oool9000001L0oooo1000000G
0?ooo`@000005`3oool00`000000oooo00000002000001P0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`0<0?ooo`009`3oool4000001H0oooo1@00000D0?ooo`L000005`3oool60000
01L0oooo1000000H0?ooo`<000005`3oool5000001L0oooo1000000C0?ooo`L000005`3oool50000
01P0oooo1000000F0?ooo`030000003oool0oooo04H0oooo000V0?ooo`D00000503oool6000001D0
oooo1@00000M0?ooo`<000005`3oool3000001T0oooo0P00000I0?ooo`<000006P3oool200000180
oooo1@00000G0?ooo`H00000603oool4000001L0oooo00<000000?ooo`3oool0AP3oool002L0oooo
1000000E0?ooo`@000005P3oool2000003/0oooo0`00000J0?ooo`030000003oool0oooo01P0oooo
0P00000]0?ooo`8000006`3oool4000001X0oooo0`00000G0?ooo`030000003oool0oooo04H0oooo
000Y0?ooo`8000005P3oool3000005@0oooo0`0000210?ooo`<00000703oool00`000000oooo0?oo
o`0D0?ooo`800000B03oool00480oooo00<000000?ooo`3oool0E03oool300000840oooo00<00000
0?ooo`3oool0<P3oool00`000000oooo0?ooo`170?ooo`00VP3oool300000;D0oooo00<000000?oo
o`3oool0A`3oool009`0oooo0P00002c0?ooo`800000?@3oool00`000000oooo0?ooo`090?ooo`00
W@3oool200000;80oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`090?ooo`00WP3o
ool200000;40oooo00<000000?ooo`3oool0>`3oool2000000/0oooo002O0?ooo`800000[`3oool2
000003d0oooo0P00000;0?ooo`00X03oool200000:h0oooo00<000000?ooo`3oool0>`3oool30000
00/0oooo002Q0?ooo`800000[03oool2000003d0oooo1000000:0?ooo`00XP3oool200000:/0oooo
0P00000S0?ooo`030000003oool0oooo01L0oooo1000000:0?ooo`00X`3oool200000:X0oooo00<0
00000?ooo`3oool08P3oool00`000000oooo0?ooo`0F0?ooo`D000002P3oool00:@0oooo0P00002X
0?ooo`8000008`3oool2000001P0oooo1@00000:0?ooo`004@3oool00`000000oooo0?ooo`2A0?oo
o`800000Y`3oool2000002<0oooo0P00000J0?ooo`<000002P3oool00100oooo0P00002D0?ooo`80
0000YP3oool00`000000oooo0?ooo`0Q0?ooo`<000006P3oool00`000000oooo0?ooo`0:0?ooo`00
403oool2000009D0oooo0P00002T0?ooo`8000002P3oool00`000000oooo0?ooo`0F0?ooo`@00000
603oool00`000000oooo0?ooo`0;0?ooo`003`3oool3000009H0oooo0P00002S0?ooo`8000002@3o
ool2000001P0oooo1000000H0?ooo`030000003oool0oooo00/0oooo000?0?ooo`<000006P3oool0
0`000000oooo0?ooo`1j0?ooo`800000XP3oool00`000000oooo0?ooo`080?ooo`8000005`3oool5
000001P0oooo00<000000?ooo`3oool02`3oool000h0oooo1000000I0?ooo`800000O@3oool20000
0:00oooo0P0000090?ooo`<000005`3oool5000001P0oooo00<000000?ooo`3oool02`3oool000h0
oooo1000000I0?ooo`800000OP3oool200000940oooo00<000000?ooo`3oool02`3oool2000000T0
oooo0`00000H0?ooo`@00000603oool00`000000oooo0?ooo`0;0?ooo`003@3oool5000001P0oooo
0`00001o0?ooo`800000S`3oool2000000`0oooo0P0000090?ooo`@00000603oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00/0oooo000?0?ooo`<000005`3oool4000001X0oooo00<0
00000?ooo`3oool0;@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01D0oooo0P00
00020?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`1C0?ooo`8000003@3oool2000000T0oooo1000000H0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool02`3oool000l0oooo00<000000?ooo`3oool05P3oool4000001X0oooo0P00000_
0?ooo`<000006P3oool3000001L0oooo1@00000H0?ooo`<000005`3oool3000001L0oooo0P00000j
0?ooo`<000003@3oool2000000P0oooo1@00000H0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0303oool000h0oooo00<000000?ooo`3oool05`3oool4000001T0oooo0P00000_0?oo
o`D00000603oool4000001P0oooo1000000G0?ooo`@000005`3oool4000001H0oooo1000000N0?oo
o`8000005`3oool4000000`0oooo0P00000;0?ooo`<000005`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo00`0oooo000>0?ooo`030000003oool0oooo01H0oooo1@00000G0?ooo`@0
00004`3oool:00000180oooo1P00000F0?ooo`H000005P3oool6000001D0oooo1P00000E0?ooo`H0
00005@3oool7000001T0oooo0`00000G0?ooo`D00000303oool2000000X0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000>0?ooo`030000
003oool0oooo01L0oooo1000000F0?ooo`@00000503oool9000001<0oooo2000000E0?ooo`H00000
5P3oool5000001H0oooo1@00000G0?ooo`D00000503oool9000001D0oooo1000000G0?ooo`D00000
3@3oool2000000X0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo00`0oooo000Z0?ooo`8000005@3oool5000001@0oooo1`00000I0?ooo`D000005P3o
ool5000001L0oooo1000000G0?ooo`@00000603oool4000001<0oooo1`00000E0?ooo`D000006@3o
ool4000000`0oooo0P00001A0?ooo`00@P3oool4000001D0oooo1000000h0?ooo`<000006@3oool2
000001T0oooo0P00000K0?ooo`8000004P3oool3000001X0oooo1000000K0?ooo`800000303oool2
00000540oooo00130?ooo`8000005P3oool2000003`0oooo0P00000I0?ooo`8000006P3oool00`00
0000oooo0?ooo`190?ooo`<00000:@3oool2000004H0oooo00<000000?ooo`3oool02@3oool004@0
oooo00<000000?ooo`3oool0KP3oool2000006H0oooo00<000000?ooo`3oool0:03oool2000004H0
oooo00<000000?ooo`3oool02@3oool00;H0oooo0P00002?0?ooo`800000AP3oool2000000/0oooo
002g0?ooo`030000003oool0oooo08d0oooo0P0000160?ooo`8000002`3oool00;P0oooo00<00000
0?ooo`3oool0S03oool2000004D0oooo0`00000;0?ooo`00^@3oool00`000000oooo0?ooo`2:0?oo
o`800000AP3oool4000000X0oooo002j0?ooo`030000003oool0oooo08T0oooo0P0000160?ooo`@0
00002P3oool00;/0oooo00<000000?ooo`3oool0Q`3oool2000002h0oooo00<000000?ooo`3oool0
5@3oool5000000X0oooo002k0?ooo`800000R03oool2000002d0oooo0P00000G0?ooo`D000002P3o
ool00140oooo00<000000?ooo`3oool0Z03oool2000008L0oooo0P00000]0?ooo`8000006@3oool3
000000X0oooo000@0?ooo`800000Z`3oool2000008D0oooo0P00000]0?ooo`<000006@3oool00`00
0000oooo0?ooo`0:0?ooo`00403oool200000:`0oooo0P0000240?ooo`800000;@3oool3000001T0
oooo00<000000?ooo`3oool02P3oool000l0oooo0`00002]0?ooo`800000PP3oool2000002d0oooo
1000000H0?ooo`030000003oool0oooo00/0oooo000?0?ooo`<00000[P3oool200000840oooo0P00
000E0?ooo`030000003oool0oooo01D0oooo1000000H0?ooo`030000003oool0oooo00/0oooo000>
0?ooo`@00000[`3oool2000007l0oooo0P00000E0?ooo`8000005P3oool5000001P0oooo00<00000
0?ooo`3oool02`3oool000h0oooo1000000K0?ooo`030000003oool0oooo0980oooo0P00001n0?oo
o`8000005@3oool2000001H0oooo1@00000H0?ooo`030000003oool0oooo00/0oooo000=0?ooo`D0
00006P3oool2000009D0oooo0P00001m0?ooo`030000003oool0oooo01<0oooo0`00000H0?ooo`<0
0000603oool00`000000oooo0?ooo`0;0?ooo`003@3oool5000001X0oooo0P00002F0?ooo`800000
N`3oool2000001@0oooo1000000H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
2`3oool000l0oooo0`00000I0?ooo`<00000U`3oool2000007L0oooo00@000000?ooo`3oool00000
503oool4000001P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0;0?ooo`003`3o
ool00`000000oooo0?ooo`0H0?ooo`<000006`3oool00`000000oooo0?ooo`1k0?ooo`800000M@3o
ool2000000030?ooo`000000000001@0oooo1000000H0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool02`3oool000l0oooo00<000000?ooo`3oool05`3oool4000001X0oooo0P00001<
0?ooo`030000003oool0oooo02l0oooo0P00000N0?ooo`030000003oool0oooo0580oooo0P000002
0?ooo`030000003oool0oooo0180oooo1@00000H0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0303oool000l0oooo00<000000?ooo`3oool05`3oool4000001T0oooo0P00000`0?oo
o`030000003oool0oooo01T0oooo0`00000J0?ooo`030000003oool0oooo01@0oooo0P0000030?oo
o`030000003oool0oooo01H0oooo0`00001B0?ooo`<0000000<0oooo0000000000005@3oool40000
01P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0<0?ooo`003P3oool00`000000
oooo0?ooo`0G0?ooo`D000005`3oool4000002l0oooo1000000H0?ooo`@000006@3oool3000001H0
oooo00<000000?ooo`3oool00`00000F0?ooo`@000005`3oool200000200oooo0P00000F0?ooo`@0
000000<0oooo0000003oool05P3oool3000001L0oooo00<000000?ooo`3oool06@3oool00`000000
oooo0?ooo`0<0?ooo`003P3oool00`000000oooo0?ooo`0H0?ooo`@000005P3oool4000001T0oooo
1000000C0?ooo`D000005`3oool5000001L0oooo1000000G0?ooo`D000005P3oool5000001H0oooo
1000000L0?ooo`<000005P3oool7000001H0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo00`0oooo000>0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool05P3oool5000001@0oooo2000000C0?ooo`L000005@3oool7000001D0oooo1P00
000F0?ooo`H00000503oool7000001@0oooo2000000F0?ooo`D000005P3oool5000000030?ooo`00
0000oooo01H0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00`0oooo00110?ooo`H00000503oool6000001D0oooo2@00000E0?ooo`H000005@3oool6
000001H0oooo1@00000F0?ooo`H000004`3oool:00000180oooo1P00000H0?ooo`D00000D`3oool0
0`000000oooo0?ooo`080?ooo`00@P3oool4000001D0oooo1@00000K0?ooo`D000005P3oool40000
01L0oooo1@00000G0?ooo`@000005`3oool4000001<0oooo1@00000H0?ooo`@000006`3oool30000
05<0oooo00<000000?ooo`3oool0203oool004@0oooo0P00000F0?ooo`<00000>P3oool2000001P0
oooo0`00000H0?ooo`<000006P3oool2000002l0oooo0`00000M0?ooo`030000003oool0oooo0540
oooo0P00000:0?ooo`00G03oool00`000000oooo0?ooo`1F0?ooo`800000603oool00`000000oooo
0000001;0?ooo`030000003oool0oooo01/0oooo0P00001C0?ooo`8000002P3oool00=00oooo00<0
00000?ooo`3oool0J03oool00`000000oooo0?ooo`1A0?ooo`<000002P3oool00=40oooo0P00001W
0?ooo`800000D`3oool4000000T0oooo003B0?ooo`800000IP3oool00`000000oooo0?ooo`1A0?oo
o`D000002@3oool00=<0oooo0P00001T0?ooo`800000>`3oool00`000000oooo0?ooo`0E0?ooo`D0
00002@3oool00180oooo00<000000?ooo`3oool0_`3oool200000680oooo0P00000k0?ooo`800000
6@3oool3000000T0oooo000A0?ooo`800000`P3oool200000640oooo00<000000?ooo`3oool0>P3o
ool2000001T0oooo00<000000?ooo`3oool02@3oool00140oooo0P0000330?ooo`800000G`3oool2
000003/0oooo0`00000I0?ooo`030000003oool0oooo00T0oooo000@0?ooo`<00000a03oool20000
05d0oooo0P00000l0?ooo`<00000603oool00`000000oooo0?ooo`0:0?ooo`00403oool300000<D0
oooo0P00001L0?ooo`030000003oool0oooo03X0oooo1000000H0?ooo`030000003oool0oooo00X0
oooo000?0?ooo`@00000aP3oool2000005X0oooo0P00000l0?ooo`@00000603oool00`000000oooo
0?ooo`0:0?ooo`003`3oool400000<L0oooo0P00001H0?ooo`800000903oool00`000000oooo0?oo
o`0E0?ooo`D00000603oool00`000000oooo0?ooo`0:0?ooo`003P3oool5000001X0oooo00<00000
0?ooo`3oool0[03oool2000005H0oooo00<000000?ooo`3oool08P3oool2000001L0oooo1@00000H
0?ooo`030000003oool0oooo00X0oooo000>0?ooo`D000006@3oool200000:l0oooo0P00001D0?oo
o`800000903oool2000001T0oooo0`00000H0?ooo`030000003oool0oooo00X0oooo000@0?ooo`<0
00006@3oool200000;00oooo0P00001B0?ooo`800000903oool3000001P0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`0;0?ooo`00403oool00`000000oooo0?ooo`0H0?ooo`<00000
/@3oool200000500oooo0`00000T0?ooo`<00000603oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo00/0oooo000@0?ooo`030000003oool0oooo01P0oooo0P00002c0?ooo`800000C`3o
ool2000002@0oooo1000000H0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool02`3o
ool000l0oooo00<000000?ooo`3oool0603oool3000001`0oooo00<000000?ooo`3oool0UP3oool0
0`000000oooo0?ooo`1;0?ooo`8000003@3oool00`000000oooo0?ooo`0E0?ooo`@00000603oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00/0oooo000?0?ooo`030000003oool0oooo
01P0oooo0`00000K0?ooo`800000V@3oool2000004X0oooo0P00000=0?ooo`8000005P3oool50000
01L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0;0?ooo`003`3oool00`000000
oooo0?ooo`0G0?ooo`@000006P3oool2000009/0oooo0P0000180?ooo`8000003@3oool2000001P0
oooo1000000G0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02`3oool000l0oooo
00<000000?ooo`3oool0603oool3000001P0oooo1000001<0?ooo`030000003oool0oooo04d0oooo
0`0000150?ooo`8000003@3oool3000001P0oooo00@000000?ooo`00000000005`3oool00`000000
oooo0?ooo`0I0?ooo`030000003oool0oooo00/0oooo000?0?ooo`030000003oool0oooo01P0oooo
00<000000?ooo`3oool05`3oool400000300oooo00<000000?ooo`3oool06@3oool3000001/0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0F0?ooo`<00000603oool2000001l0oooo
0`0000070?ooo`8000003@3oool3000001T0oooo00<000000?ooo`3oool0603oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo00`0oooo000>0?ooo`030000003oool0oooo01T0oooo00<0
00000?ooo`3oool05P3oool5000001T0oooo1000000C0?ooo`<000006@3oool4000001T0oooo0`00
000H0?ooo`<000005`3oool4000001L0oooo1000000I0?ooo`H000001`3oool2000000d0oooo1000
000I0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`0<0?ooo`003P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo01H0oooo1P00000D
0?ooo`P000004`3oool6000001H0oooo1P00000G0?ooo`@000005`3oool4000001H0oooo1P00000E
0?ooo`L00000503oool8000000L0oooo0P00000=0?ooo`D00000603oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0303oool000h0oooo00<000000?oo
o`3oool0603oool00`000000oooo0?ooo`0G0?ooo`@000005@3oool6000001D0oooo2000000D0?oo
o`H000005P3oool6000001D0oooo1P00000E0?ooo`L00000503oool900000180oooo1`0000070?oo
o`<000003P3oool3000001T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo00`0oooo00150?ooo`8000005@3oool5000001H0oooo2@00000E0?ooo`D0
00005P3oool5000001H0oooo1P00000F0?ooo`L000004P3oool5000001L0oooo100000080?ooo`<0
00004@3oool00`000000oooo0?ooo`1M0?ooo`00G@3oool3000003L0oooo1000000G0?ooo`@00000
5`3oool5000001L0oooo1P00000]0?ooo`<000001`3oool400000780oooo001M0?ooo`030000003o
ool0oooo03X0oooo0P00000H0?ooo`8000006@3oool3000001X0oooo1@00000/0?ooo`030000003o
ool0oooo00D0oooo1000001c0?ooo`00]P3oool00`000000oooo0?ooo`0H0?ooo`8000007@3oool2
00000380oooo0`00001e0?ooo`00l@3oool3000002h0oooo1000001f0?ooo`00l`3oool3000002/0
oooo1000001g0?ooo`00m@3oool3000002L0oooo1000001i0?ooo`00m`3oool3000002<0oooo1@00
001j0?ooo`00n@3oool4000001l0oooo1000001l0?ooo`00n`3oool5000001X0oooo1000001n0?oo
o`00o@3oool6000001@0oooo100000210?ooo`00o`3oool20?ooo`P00000203oool8000008<0oooo
003o0?ooo`H0oooo4@0000260?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00
o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00o`3ooonM0?ooo`00
o`3ooonM0?ooo`00o`3ooonM0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 411.375}, {430.875, 0}} -> {-2.6844, -2.8111, \
0.0195763, 0.0195763}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(xyCurves[0.1, 0, 1, 40];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.0238095 0.30891 0.10992 [
[.2619 .29641 -8 -15.6875 ]
[.2619 .29641 8 0 ]
[.5 .29641 -8 -15.6875 ]
[.5 .29641 8 0 ]
[.7381 .29641 -8 -15.6875 ]
[.7381 .29641 8 0 ]
[.97619 .29641 -8 -15.6875 ]
[.97619 .29641 8 0 ]
[.01131 .08907 -18.6875 -7.84375 ]
[.01131 .08907 0 7.84375 ]
[.01131 .19899 -18.6875 -7.84375 ]
[.01131 .19899 0 7.84375 ]
[.01131 .41883 -10 -7.84375 ]
[.01131 .41883 0 7.84375 ]
[.01131 .52875 -10 -7.84375 ]
[.01131 .52875 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .30891 m
.2619 .31516 L
s
gsave
.2619 .29641 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.5 .30891 m
.5 .31516 L
s
gsave
.5 .29641 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.7381 .30891 m
.7381 .31516 L
s
gsave
.7381 .29641 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
.97619 .30891 m
.97619 .31516 L
s
gsave
.97619 .29641 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(40) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30891 m
.07143 .31266 L
s
.11905 .30891 m
.11905 .31266 L
s
.16667 .30891 m
.16667 .31266 L
s
.21429 .30891 m
.21429 .31266 L
s
.30952 .30891 m
.30952 .31266 L
s
.35714 .30891 m
.35714 .31266 L
s
.40476 .30891 m
.40476 .31266 L
s
.45238 .30891 m
.45238 .31266 L
s
.54762 .30891 m
.54762 .31266 L
s
.59524 .30891 m
.59524 .31266 L
s
.64286 .30891 m
.64286 .31266 L
s
.69048 .30891 m
.69048 .31266 L
s
.78571 .30891 m
.78571 .31266 L
s
.83333 .30891 m
.83333 .31266 L
s
.88095 .30891 m
.88095 .31266 L
s
.92857 .30891 m
.92857 .31266 L
s
.25 Mabswid
0 .30891 m
1 .30891 L
s
.02381 .08907 m
.03006 .08907 L
s
gsave
.01131 .08907 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .19899 m
.03006 .19899 L
s
gsave
.01131 .19899 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .41883 m
.03006 .41883 L
s
gsave
.01131 .41883 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .52875 m
.03006 .52875 L
s
gsave
.01131 .52875 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .11105 m
.02756 .11105 L
s
.02381 .13304 m
.02756 .13304 L
s
.02381 .15502 m
.02756 .15502 L
s
.02381 .17701 m
.02756 .17701 L
s
.02381 .22097 m
.02756 .22097 L
s
.02381 .24296 m
.02756 .24296 L
s
.02381 .26494 m
.02756 .26494 L
s
.02381 .28693 m
.02756 .28693 L
s
.02381 .33089 m
.02756 .33089 L
s
.02381 .35288 m
.02756 .35288 L
s
.02381 .37486 m
.02756 .37486 L
s
.02381 .39685 m
.02756 .39685 L
s
.02381 .44081 m
.02756 .44081 L
s
.02381 .4628 m
.02756 .4628 L
s
.02381 .48478 m
.02756 .48478 L
s
.02381 .50677 m
.02756 .50677 L
s
.02381 .06709 m
.02756 .06709 L
s
.02381 .0451 m
.02756 .0451 L
s
.02381 .02312 m
.02756 .02312 L
s
.02381 .00113 m
.02756 .00113 L
s
.02381 .55073 m
.02756 .55073 L
s
.02381 .57272 m
.02756 .57272 L
s
.02381 .5947 m
.02756 .5947 L
s
.02381 .61669 m
.02756 .61669 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.02381 .3199 m
.02499 .31989 L
.02605 .31985 L
.02729 .31978 L
.02846 .31968 L
.02954 .31956 L
.03053 .31942 L
.03279 .31902 L
.03506 .31849 L
.03756 .31773 L
.04262 .31561 L
.04755 .31275 L
.05285 .30872 L
.06244 .29896 L
.07277 .2853 L
.08405 .26872 L
.0892 .26164 L
.09469 .25525 L
.09736 .25278 L
.09988 .25093 L
.10214 .24971 L
.10341 .24923 L
.10458 .24892 L
.10583 .24875 L
.10697 .24872 L
.10828 .24887 L
.10952 .24919 L
.11067 .24964 L
.11173 .2502 L
.11414 .25197 L
.11655 .25449 L
.11922 .25819 L
.12204 .26319 L
.12463 .26882 L
.12988 .28356 L
.13555 .30484 L
.1458 .35806 L
.15495 .4144 L
.15998 .44117 L
.16245 .45158 L
.16473 .45924 L
.16601 .46272 L
.16742 .46583 L
.16884 .46824 L
.16962 .46926 L
.17035 .47003 L
.17163 .47096 L
.17281 .47137 L
.17346 .47141 L
.17416 .47133 L
Mistroke
.17545 .47082 L
.17677 .46986 L
.17796 .46863 L
.18063 .46473 L
.1831 .45983 L
.18543 .45419 L
.1901 .44005 L
.19522 .42022 L
.20021 .39602 L
.20555 .36344 L
.21511 .28149 L
.22403 .18118 L
.2266 .15453 L
.22933 .13103 L
.23166 .11574 L
.23299 .10905 L
.23423 .10404 L
.23559 .09986 L
.23633 .09809 L
.23702 .09676 L
.23824 .09504 L
.23888 .09445 L
.23957 .09402 L
.24031 .0938 L
.24112 .0938 L
.24188 .09402 L
.24258 .09439 L
.24321 .09485 L
.24391 .09549 L
.24531 .09715 L
.24794 .10138 L
.25033 .10625 L
.25572 .11989 L
.26544 .15204 L
.27534 .19608 L
.28451 .25357 L
.28928 .29382 L
.29456 .34897 L
.29977 .41191 L
.30261 .44582 L
.30526 .47382 L
.30778 .49542 L
.31007 .51008 L
.31138 .51636 L
.31261 .52099 L
.31392 .52465 L
.31465 .52621 L
.31533 .52735 L
.31644 .52865 L
.31747 .52931 L
Mistroke
.3186 .5295 L
.31982 .52916 L
.3211 .52829 L
.32231 .52704 L
.32458 .52378 L
.32708 .5191 L
.32975 .51314 L
.33455 .50044 L
.34543 .46369 L
.35572 .41624 L
.36152 .38081 L
.36686 .33961 L
.37729 .22759 L
.38239 .16631 L
.3848 .14133 L
.38703 .12241 L
.38927 .10797 L
.39045 .10216 L
.3917 .09726 L
.39279 .09399 L
.39381 .09166 L
.39492 .08985 L
.39613 .08866 L
.39731 .08816 L
.3986 .08827 L
.39926 .08856 L
.39995 .08901 L
.40121 .09018 L
.40247 .09174 L
.40365 .09353 L
.40587 .09756 L
.41579 .12302 L
.42654 .1613 L
.43601 .20692 L
.4414 .24081 L
.44643 .28039 L
.45723 .3978 L
.46255 .46064 L
.46514 .48562 L
.46748 .50323 L
.46864 .51013 L
.4699 .51623 L
.47099 .52045 L
.47218 .52402 L
.47345 .52677 L
.47415 .52785 L
.4748 .52861 L
.47595 .52945 L
.47719 .52969 L
.4779 .52956 L
Mistroke
.47856 .52928 L
.47928 .52881 L
.48004 .52815 L
.48132 .5267 L
.48274 .52468 L
.48778 .51467 L
.49735 .48774 L
.50632 .45462 L
.51128 .43221 L
.51669 .40303 L
.52147 .37167 L
.52656 .33024 L
.53616 .22486 L
.5405 .17238 L
.54287 .14695 L
.54508 .12705 L
.54763 .10955 L
.54898 .10266 L
.55043 .09698 L
.55107 .095 L
.55176 .09318 L
.55301 .09066 L
.55423 .0891 L
.55536 .08833 L
.55668 .08814 L
.55792 .08856 L
.55863 .08903 L
.55928 .08959 L
.56075 .09125 L
.56354 .09568 L
.56648 .10173 L
.57611 .12826 L
.58665 .16746 L
.59193 .19224 L
.59692 .22024 L
.60625 .29096 L
.61642 .40456 L
.61911 .43713 L
.62199 .46885 L
.62466 .4928 L
.62598 .50227 L
.6272 .50962 L
.62845 .51578 L
.62958 .52023 L
.63084 .52403 L
.63153 .52566 L
.63217 .52689 L
.6333 .52846 L
.63452 .52941 L
.63559 .5297 L
Mistroke
.63672 .5295 L
.63797 .52878 L
.63867 .52817 L
.63931 .5275 L
.64172 .52413 L
.64454 .51885 L
.6472 .51283 L
.65604 .4876 L
.66565 .45174 L
.67507 .40465 L
.68043 .36939 L
.68546 .32796 L
.6957 .21387 L
.6984 .18104 L
.70128 .14928 L
.70396 .12526 L
.70527 .11579 L
.70647 .10847 L
.70763 .10266 L
.70891 .09758 L
.71018 .09373 L
.71088 .0921 L
.71154 .09086 L
.7127 .08928 L
.71375 .08845 L
.71498 .08812 L
.71613 .08835 L
.7174 .08913 L
.71805 .0897 L
.71876 .09046 L
.72119 .09391 L
.72402 .0993 L
.72669 .10538 L
.73559 .13095 L
.7452 .16704 L
.75475 .21528 L
.76017 .25148 L
.76521 .29383 L
.77605 .41658 L
.77864 .44732 L
.78138 .47608 L
.78372 .49591 L
.78507 .50506 L
.7863 .51199 L
.78746 .51733 L
.78873 .5219 L
.78943 .52389 L
.79008 .52544 L
.79135 .52772 L
.79256 .52903 L
Mistroke
.79365 .5296 L
.79491 .52964 L
.79609 .52914 L
.79738 .52811 L
.79804 .52742 L
.79875 .52654 L
.8012 .52272 L
.80672 .51069 L
.81633 .48215 L
.8253 .44743 L
.83031 .42352 L
.83575 .39219 L
.84566 .3132 L
.85523 .20304 L
.85992 .14946 L
.86197 .13042 L
.86421 .11386 L
.86642 .10206 L
.86762 .09735 L
.86874 .09397 L
.86998 .09124 L
.8711 .08956 L
.87239 .08846 L
.87362 .08812 L
.87488 .08839 L
.87551 .08873 L
.87621 .08925 L
.87744 .0905 L
.8786 .092 L
.8812 .09635 L
.88399 .10221 L
.89497 .13337 L
.905 .17208 L
.90979 .19511 L
.91488 .22437 L
.92403 .29547 L
.934 .408 L
.93924 .46839 L
.94188 .49218 L
.94339 .50303 L
.94477 .51109 L
.94688 .52024 L
.94804 .52379 L
.94914 .52628 L
.95015 .52792 L
.95109 .52893 L
.95212 .52954 L
.95321 .52969 L
.95413 .52946 L
.95511 .5289 L
Mistroke
.95605 .5281 L
.9569 .52717 L
.95898 .5242 L
.96096 .52063 L
.96824 .50306 L
.97619 .47795 L
Mfstroke
0 0 1 r
.02381 .30891 m
.04262 .29704 L
.06244 .28082 L
.06747 .27736 L
.07024 .27581 L
.0716 .27517 L
.07286 .27466 L
.07401 .27426 L
.07529 .27391 L
.07657 .27367 L
.07728 .27358 L
.07793 .27352 L
.0791 .27351 L
.08016 .27358 L
.08139 .27378 L
.08255 .27408 L
.08327 .27432 L
.08394 .27459 L
.0852 .27521 L
.08801 .27712 L
.09045 .2794 L
.09308 .28253 L
.09907 .29221 L
.10458 .30402 L
.1234 .36099 L
.13421 .40477 L
.13933 .4274 L
.14407 .4462 L
.14639 .45309 L
.14757 .4556 L
.14887 .45745 L
.15016 .45815 L
.15089 .458 L
.15156 .4575 L
.15279 .45559 L
.15409 .45216 L
.15522 .44796 L
.15643 .44225 L
.15862 .42903 L
.16358 .38921 L
.18449 .25021 L
.20584 .14326 L
.21089 .10051 L
.21563 .05763 L
.21685 .04822 L
.21801 .04082 L
.21864 .0376 L
.21934 .03491 L
.22056 .03265 L
.22128 .03298 L
Mistroke
.22193 .03444 L
.22267 .03758 L
.22338 .04201 L
.22463 .05348 L
.22599 .07074 L
.23106 .164 L
.23384 .21709 L
.23642 .25905 L
.23881 .28979 L
.24099 .31147 L
.24341 .32982 L
.246 .34444 L
.24837 .35461 L
.25055 .36217 L
.25548 .37562 L
.26446 .39691 L
.2691 .40995 L
.27411 .42789 L
.27883 .45061 L
.28316 .47845 L
.28837 .52276 L
.29321 .57033 L
.29447 .58155 L
.29565 .59066 L
.29631 .59484 L
.29702 .5985 L
.29826 .60227 L
.29892 .60269 L
.29965 .60169 L
.30042 .59889 L
.30113 .59459 L
.30241 .58272 L
.30312 .57377 L
.30377 .56427 L
.30854 .46961 L
.31301 .37982 L
.31526 .34519 L
.31768 .31641 L
.32037 .29315 L
.32284 .27778 L
.32534 .26623 L
.32808 .25665 L
.33361 .24247 L
.34359 .22035 L
.34886 .20575 L
.35385 .18758 L
.35832 .16569 L
.36317 .13351 L
.36846 .0868 L
.37136 .05796 L
Mistroke
.37277 .04437 L
.37407 .03312 L
.37517 .02501 L
.37637 .01841 L
.37764 .0148 L
.37885 .01533 L
.38017 .02102 L
.38091 .02669 L
.38158 .03339 L
.38289 .05052 L
.38412 .07103 L
.38917 .17741 L
.39389 .26352 L
.39626 .29408 L
.39881 .31834 L
.40163 .33746 L
.40421 .34996 L
.40657 .35868 L
.40876 .36523 L
.4137 .37712 L
.42269 .39695 L
.42733 .40944 L
.43232 .42677 L
.43704 .44881 L
.44137 .4759 L
.44653 .51894 L
.44937 .54676 L
.45204 .57299 L
.45326 .58364 L
.45454 .59315 L
.45521 .5971 L
.45595 .60045 L
.45725 .60328 L
.45842 .60185 L
.45969 .5955 L
.46088 .58469 L
.46197 .57081 L
.47104 .39002 L
.47352 .34987 L
.47616 .31757 L
.47865 .29542 L
.48091 .28062 L
.48354 .26782 L
.48639 .25756 L
.49154 .24402 L
.50162 .22182 L
.50623 .20964 L
.51116 .19298 L
.51657 .16791 L
.52161 .13513 L
Mistroke
.52728 .08509 L
.53007 .05729 L
.5313 .0455 L
.53265 .03364 L
.53377 .02528 L
.53498 .01855 L
.53626 .01482 L
.53746 .01524 L
.53813 .01734 L
.53886 .02128 L
.53955 .02662 L
.54019 .03298 L
.5414 .04854 L
.54269 .06972 L
.54769 .17464 L
.55042 .22888 L
.55297 .26975 L
.55573 .30276 L
.5572 .31618 L
.55876 .32803 L
.56155 .34409 L
.56409 .35479 L
.56905 .36952 L
.57445 .38163 L
.58426 .40452 L
.5888 .41863 L
.59358 .43815 L
.59883 .4677 L
.60364 .50488 L
.60891 .55568 L
.61169 .58187 L
.6125 .58828 L
.61325 .59355 L
.61394 .59752 L
.61467 .60072 L
.61531 .60252 L
.61601 .60332 L
.61725 .60129 L
.61794 .59806 L
.61869 .59283 L
.61939 .58619 L
.62004 .57872 L
.6226 .53725 L
.62503 .48642 L
.62975 .38887 L
.63239 .34682 L
.63477 .31796 L
.63701 .29767 L
.63943 .28138 L
.64171 .26988 L
Mistroke
.64381 .26163 L
.64856 .24785 L
.65377 .23619 L
.66429 .21134 L
.6695 .19416 L
.67225 .18273 L
.67517 .16819 L
.68052 .13316 L
.68546 .0897 L
.68988 .04613 L
.69109 .03536 L
.69236 .02569 L
.69345 .0194 L
.69464 .01529 L
.69594 .01493 L
.69668 .01692 L
.69734 .0202 L
.69863 .03076 L
.69985 .04569 L
.70221 .08605 L
.70475 .14003 L
.70721 .19282 L
.70984 .24235 L
.71233 .2793 L
.7146 .30495 L
.71715 .32621 L
.71957 .34091 L
.72218 .35266 L
.72502 .36229 L
.73465 .38494 L
.7437 .40673 L
.74821 .42135 L
.75308 .44224 L
.75848 .47459 L
.76344 .5153 L
.76873 .56722 L
.77006 .57944 L
.77131 .58952 L
.77244 .59672 L
.77366 .60181 L
.77495 .60325 L
.77615 .6003 L
.77744 .59199 L
.77809 .58571 L
.7788 .57722 L
.78171 .52814 L
.78444 .46986 L
.7894 .37159 L
.79221 .33158 L
.79482 .30462 L
Mistroke
.79733 .28593 L
.7996 .27339 L
.80222 .26249 L
.80469 .25453 L
.82322 .21054 L
.82777 .19559 L
.83209 .17713 L
.83706 .14837 L
.8416 .11306 L
.8466 .06497 L
.84923 .0397 L
.85068 .02801 L
.85139 .02329 L
.85203 .01971 L
.85319 .01549 L
.85444 .01472 L
.85573 .01875 L
.85638 .02281 L
.85709 .02886 L
.85827 .04241 L
.85952 .06145 L
.86177 .10483 L
.86715 .21836 L
.86979 .26284 L
.87222 .29422 L
.87444 .3157 L
.87678 .33271 L
.87913 .34543 L
.88172 .35595 L
.88697 .37101 L
.8926 .38345 L
.89789 .39516 L
.90278 .40799 L
.90769 .4245 L
.91292 .44843 L
.91591 .46624 L
.91872 .48641 L
.92397 .5333 L
.92867 .57908 L
.92981 .5884 L
.93104 .59643 L
.93174 .59982 L
.93238 .60199 L
.93363 .60323 L
.93433 .60199 L
.93497 .5995 L
.93568 .59523 L
.93643 .58889 L
.93768 .57415 L
.93907 .55246 L
Mistroke
.94407 .44844 L
.94605 .40766 L
.94821 .36902 L
.95011 .34104 L
.95215 .31721 L
.95415 .2989 L
.95598 .28591 L
.95954 .26763 L
.96168 .25967 L
.96366 .25363 L
.96811 .2426 L
.97619 .22504 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`008@3oool00`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`800000
o`3ooon>0?ooo`008@3oool00`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000003oool0
oooo0?l0ooooS@3oool00240oooo00<000000?ooo`3oool0o`3ooon=0?ooo`008@3oool00`000000
oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000003oool0oooo08`0oooo00<0003o0?ooo`3oool0
?@3oool00`000?l0oooo0?ooo`0n0?ooo`030000o`3oool0oooo03d0oooo00<0003o0?ooo`3oool0
?03oool00240oooo00<000000?ooo`3oool0R`3oool20000ocl0oooo0P000?ln0?ooo`80003o?`3o
ool00`000?l0oooo0?ooo`0l0?ooo`008@3oool00`000000oooo0?ooo`2;0?ooo`80003o?`3oool2
0000och0oooo0P000?ln0?ooo`030000o`3oool0003o03d0oooo000Q0?ooo`030000003oool0oooo
08/0oooo00<0003o0?ooo`000?l0?P3oool20000och0oooo00<0003o0?ooo`000?l0?@3oool00`00
0?l0oooo0000o`0m0?ooo`008@3oool2000008`0oooo00<0003o0?ooo`000?l0?@3oool00`000?l0
oooo0000o`0m0?ooo`040000o`3oool0oooo0000ocd0oooo00<0003o0?ooo`000?l0?@3oool00240
oooo00<000000?ooo`3oool0R`3oool00`000?l0oooo0000o`0m0?ooo`040000o`3oool0oooo0000
oc`0oooo00@0003o0?ooo`3oool0003o?@3oool00`000?l0oooo0000o`0m0?ooo`008@3oool00`00
0000oooo0?ooo`2;0?ooo`030000o`3oool0003o03d0oooo00@0003o0?ooo`3oool0003o?03oool0
10000?l0oooo0?ooo`000?lm0?ooo`030000o`3oool0003o03d0oooo000Q0?ooo`030000003oool0
oooo08/0oooo00@0003o0?ooo`3oool0003o?03oool010000?l0oooo0?ooo`000?ll0?ooo`040000
o`3oool0oooo0000ocd0oooo00<0003o0?ooo`000?l0?@3oool00240oooo00<000000?ooo`3oool0
C03oool20000ocd0oooo00@0003o0?ooo`3oool0003o?03oool010000?l0oooo0?ooo`000?ll0?oo
o`040000o`3oool0oooo0000oc`0oooo00@0003o0?ooo`3oool0003o?@3oool00240oooo00<00000
0?ooo`3oool0C03oool20000ocd0oooo00@0003o0?ooo`3oool0003o?03oool010000?l0oooo0?oo
o`000?ll0?ooo`040000o`3oool0oooo0000oc`0oooo00@0003o0?ooo`3oool0003o?@3oool00240
oooo00<000000?ooo`3oool0C03oool20000oc`0oooo00D0003o0?ooo`3oool0oooo0000o`0l0?oo
o`040000o`3oool0oooo0000oc`0oooo00@0003o0?ooo`3oool0003o?03oool01@000?l0oooo0?oo
o`3oool0003o03`0oooo000Q0?ooo`030000003oool0oooo04/0oooo00<0003o0?ooo`000?l0?03o
ool01@000?l0oooo0?ooo`3oool0003o03/0oooo00D0003o0?ooo`3oool0oooo0000o`0k0?ooo`03
0000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>@3oool01@000?l0oooo0?ooo`3oool0003o
03`0oooo000Q0?ooo`030000003oool0oooo04/0oooo00<0003o0?ooo`000?l0?03oool01@000?l0
oooo0?ooo`3oool0003o03/0oooo00D0003o0?ooo`3oool0oooo0000o`0k0?ooo`030000o`3oool0
oooo0080oooo00<0003o0?ooo`3oool0>@3oool01@000?l0oooo0?ooo`3oool0003o03`0oooo000Q
0?ooo`800000C03oool00`000?l0oooo0000o`0l0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3o
ool01@000?l0oooo0?ooo`3oool0003o03/0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0i0?ooo`050000o`3oool0oooo0?ooo`000?l0?03oool00240oooo00<000000?ooo`3oool0
B`3oool010000?l0oooo0?ooo`000?lk0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3oool01@00
0?l0oooo0?ooo`3oool0003o03/0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0i
0?ooo`050000o`3oool0oooo0?ooo`000?l0?03oool00240oooo00<000000?ooo`3oool0B`3oool0
10000?l0oooo0?ooo`000?lj0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>@3o
ool01@000?l0oooo0?ooo`3oool0003o03/0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0i0?ooo`050000o`3oool0oooo0?ooo`000?l0?03oool00240oooo00<000000?ooo`3oool0
BP3oool01@000?l0oooo0?ooo`3oool0003o03X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0
oooo0?ooo`0i0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3oool00`000?l0oooo0?ooo`020?oo
o`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0j0?oo
o`008@3oool00`000000oooo0?ooo`1:0?ooo`050000o`3oool0oooo0?ooo`000?l0>P3oool00`00
0?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03T0oooo00D0003o0?ooo`3oool0oooo0000o`0k
0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>03oool00`000?l0oooo0?ooo`02
0?ooo`030000o`3oool0oooo03X0oooo000Q0?ooo`030000003oool0oooo04X0oooo00D0003o0?oo
o`3oool0oooo0000o`0j0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>@3oool0
1@000?l0oooo0?ooo`3oool0003o03/0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?oo
o`0h0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>P3oool00240oooo00<00000
0?ooo`3oool0BP3oool01@000?l0oooo0?ooo`3oool0003o03X0oooo00<0003o0?ooo`3oool00P3o
ool00`000?l0oooo0?ooo`0i0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>03o
ool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3o
ool00`000?l0oooo0?ooo`0j0?ooo`008@3oool00`000000oooo0?ooo`1:0?ooo`030000o`3oool0
oooo0080oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0
oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0h0?ooo`030000o`3oool0
oooo0080oooo00<0003o0?ooo`3oool0>03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0
oooo03X0oooo000Q0?ooo`030000003oool0oooo04X0oooo00<0003o0?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0>03oool00`00
0?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`0h0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>P3oool00240
oooo0P00001;0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0=`3oool00`000?l0
oooo0?ooo`030?ooo`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0
oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0>03oool00`000?l0
oooo0?ooo`020?ooo`030000o`3oool0oooo03X0oooo000Q0?ooo`030000003oool0oooo04T0oooo
00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo
00<0003o0?ooo`3oool0>03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03L0oooo
00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0h0?ooo`030000o`3oool0oooo0080oooo
00<0003o0?ooo`3oool0>P3oool00240oooo00<000000?ooo`3oool0B@3oool00`000?l0oooo0?oo
o`030?ooo`030000o`3oool0oooo03H0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?oo
o`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?oo
o`030?ooo`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?oo
o`0j0?ooo`00403oool00`3OW4P00000000000030000000304RLg`3oool0oooo00P0oooo00<00000
0?ooo`3oool0B@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03H0oooo00<0003o
0?ooo`3oool0103oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o
0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03P0oooo00<0003o
0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0j0?ooo`004@3oool00`3o_g@007Bo0?ooo`020?oo
o`030;mdM02oool0oooo00P0oooo00<000000?ooo`3oool0B@3oool00`000?l0oooo0?ooo`030?oo
o`030000o`3oool0oooo03H0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`0g0?oo
o`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`030?oo
o`030000o`3oool0oooo03L0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0j0?oo
o`004P3oool00`3o_g@007Bo0?ooo`0<0?ooo`030000003oool0oooo04T0oooo00<0003o0?ooo`3o
ool00`3oool00`000?l0oooo0?ooo`0f0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3o
ool0=`3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03L0oooo00<0003o0?ooo`3o
ool00`3oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3o
ool0>P3oool001<0oooo00<0okmd001d_`3oool02`3oool00`000000oooo0?ooo`190?ooo`030000
o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0=P3oool00`000?l0oooo0?ooo`040?ooo`030000
o`3oool0oooo03L0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0g0?ooo`030000
o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`030?ooo`030000
o`3oool0oooo03X0oooo00030?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103o
ool00`3o_g@0B9cO0?ooo`0:0?ooo`030000003oool0oooo04T0oooo00<0003o0?ooo`3oool00`3o
ool00`000?l0oooo0?ooo`0f0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0=`3o
ool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03L0oooo00<0003o0?ooo`3oool00`3o
ool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0>@3o
ool001D0oooo00<0gia804RLg`3oool02@3oool3000004T0oooo00<0003o0?ooo`3oool00`3oool0
0`000?l0oooo0?ooo`0f0?ooo`030000o`3oool0oooo00@0oooo00@0003o0?ooo`3oool0oooo0`3o
000c0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool00P3oool20?l003<0oooo00<0
003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`020?ooo`<0o`00<P3oool00`000?l0oooo0?oo
o`040?ooo`040000o`3oool0oooo0?ooo`<0o`00=@3oool001H0oooo00<0_g@007Boo`3oool0203o
ool00`000000oooo0?ooo`180?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0=P3o
ool00`000?l0oooo0?ooo`040?ooo`050000o`3oool0oooo0?ooo`3o00000P3oool00`3o0000oooo
0?ooo`0`0?ooo`030000o`3oool0oooo00<0oooo00D0003o0?ooo`3oool0oooo0?l000020?ooo`03
0?l0003oool0oooo0300oooo00<0003o0?ooo`3oool00`3oool010000?l0oooo0?ooo`3oool20?l0
00030?ooo`3o0000oooo0340oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`020?l0
00030?ooo`3o0000oooo03@0oooo000F0?ooo`030?oOW01807@0_ooo00P0oooo00<000000?ooo`3o
ool0B03oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo0080oooo0P3o000b0?ooo`03
0000o`3oool0oooo00@0oooo00@0003o0?ooo`3oool0o`000`3oool20?l00340oooo00<0003o0?oo
o`3oool00`3oool01@000?l0oooo0?ooo`3oool0o`000080oooo00<0o`000?ooo`3oool0<03oool0
0`000?l0oooo0?ooo`030?ooo`050000o`3oool0oooo0?ooo`3o00000`3oool00`3o0000oooo0?oo
o`0_0?ooo`030000o`3oool0oooo00@0oooo00@0003o0?ooo`3oool0o`000`3oool00`3o0000oooo
0?ooo`0b0?ooo`005P3oool00`3ooml0W4P007Boo`080?ooo`030000003oool0oooo04P0oooo00<0
003o0?ooo`3oool01@3oool00`000?l0oooo0?ooo`040?l00340oooo00<0003o0?ooo`3oool0103o
ool010000?l0oooo0?ooo`3o00040?ooo`030?l0003oool0oooo02h0oooo00<0003o0?ooo`3oool0
103oool010000?l0oooo0?ooo`3o00040?ooo`030?l0003oool0oooo02l0oooo00<0003o0?ooo`3o
ool00`3oool010000?l0oooo0?ooo`3o00040?ooo`030?l0003oool0oooo02l0oooo00<0003o0?oo
o`3oool0103oool010000?l0oooo0?ooo`3o00040?ooo`030?l0003oool0oooo0340oooo000F0?oo
o`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0B03oool00`000?l0oooo0?ooo`050?oo
o`040000o`3oool0oooo0?l00080oooo00<0o`000?ooo`3oool0;`3oool00`000?l0oooo0?ooo`05
0?ooo`030000o`3oool0o`0000@0oooo00<0o`000?ooo`3oool0;P3oool00`000?l0oooo0?ooo`05
0?ooo`030000o`3oool0o`0000@0oooo00<0o`000?ooo`3oool0;`3oool00`000?l0oooo0?ooo`03
0?ooo`040000o`3oool0oooo0?l000D0oooo00<0o`000?ooo`3oool0;P3oool00`000?l0oooo0?oo
o`040?ooo`040000o`3oool0oooo0?l000@0oooo00<0o`000?ooo`3oool0<@3oool00180oooo00L0
okmd0018W03Oool0oooo0=nLB00004P0W=oo00P0oooo00<000000?ooo`3oool0B03oool00`000?l0
oooo0?ooo`050?ooo`030000o`3oool0o`0000@0oooo00<0o`000?ooo`3oool0;@3oool00`000?l0
oooo0?ooo`060?ooo`030000o`3o0000oooo00@0oooo00<0o`000?ooo`3oool0;P3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3o0000oooo00D0oooo00<0o`000?ooo`3oool0;P3oool00`000?l0
oooo0?ooo`040?ooo`030000o`3oool0o`0000D0oooo00<0o`000?ooo`3oool0;P3oool00`000?l0
oooo0?ooo`040?ooo`030000o`3oool0o`0000H0oooo00<0o`000?ooo`3oool0<03oool001<0oooo
00D0omnL04P000000000000007Boo`090?ooo`030000003oool0oooo04P0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0oooo0?l000040?ooo`030?l0003oool0oooo02d0oooo00<0003o0?ooo`3o
ool01P3oool00`000?l0o`000?ooo`050?ooo`030?l0003oool0oooo02d0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0o`000?ooo`050?ooo`030?l0003oool0oooo02d0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0oooo0?l000050?ooo`030?l0003oool0oooo02d0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0oooo0?l000060?ooo`030?l0003oool0oooo0300oooo000Q0?ooo`030000
003oool0oooo04P0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?l000050?ooo`030?l0
003oool0oooo02`0oooo00<0003o0?ooo`3oool01P3oool00`000?l0o`000?ooo`050?ooo`030?l0
003oool0oooo02d0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`050?ooo`030?l0
003oool0oooo02d0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`060?ooo`030?l0
003oool0oooo02`0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?l000060?ooo`030?l0
003oool0oooo0300oooo000Q0?ooo`030000003oool0oooo04P0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`050?ooo`030?l0003oool0oooo02`0oooo00<0003o0?ooo`3oool01P3o
ool00`000?l0o`000?ooo`060?ooo`030?l0003oool0oooo02`0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`060?ooo`030?l0003oool0oooo02`0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`060?ooo`030?l0003oool0oooo02`0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`070?ooo`030?l0003oool0oooo02l0oooo000Q0?ooo`800000B@3oool0
0`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00H0oooo00<0o`000?ooo`3oool0:`3oool0
0`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo00H0oooo00<0o`000?ooo`3oool0:`3oool0
0`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo00H0oooo00<0o`000?ooo`3oool0;03oool0
0`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00L0oooo00<0o`000?ooo`3oool0:`3oool0
0`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00L0oooo00<0o`000?ooo`3oool0;`3oool0
0240oooo00<000000?ooo`3oool0A`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo
00H0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo
00H0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo
00L0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo
00L0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo
00P0oooo00<0o`000?ooo`3oool0;P3oool00240oooo00<000000?ooo`3oool0A`3oool00`000?l0
oooo0?ooo`060?ooo`030000o`3o0000oooo00L0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0
oooo0?ooo`060?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0
oooo0?ooo`060?ooo`030000o`3o0000oooo00L0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3o0000oooo00P0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3o0000oooo00P0oooo00<0o`000?ooo`3oool0;P3oool00240oooo
00<000000?ooo`3oool0A`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo00L0oooo
00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00L0oooo
00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?ooo`060?ooo`030000o`3o0000oooo00L0oooo
00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00P0oooo
00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00P0oooo
00<0o`000?ooo`3oool0;P3oool00240oooo00<000000?ooo`3oool0A`3oool00`000?l0oooo0?oo
o`060?ooo`030000o`3o0000oooo00P0oooo00<0o`000?ooo`3oool0:@3oool00`000?l0oooo0?oo
o`060?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool0:@3oool00`000?l0oooo0?oo
o`060?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?oo
o`050?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?oo
o`050?ooo`030000o`3o0000oooo00T0oooo00<0o`000?ooo`3oool0;@3oool00240oooo00<00000
0?ooo`3oool0A`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00P0oooo00<0o`00
0?ooo`3oool0:03oool00`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo00P0oooo00<0o`00
0?ooo`3oool0:@3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00P0oooo00<0o`00
0?ooo`3oool0:@3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00T0oooo00<0o`00
0?ooo`3oool0:03oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00T0oooo00<0o`00
0?ooo`3oool0;@3oool00240oooo00<000000?ooo`3oool0A`3oool00`000?l0oooo0?ooo`060?oo
o`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool09`3oool00`000?l0oooo0?ooo`070?oo
o`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool0:@3oool00`000?l0oooo0?ooo`060?oo
o`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool0:03oool00`000?l0oooo0?ooo`060?oo
o`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool0:03oool00`000?l0oooo0?ooo`060?oo
o`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool0;03oool00240oooo00<000000?ooo`3o
ool0A`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3o
ool09`3oool00`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3o
ool09`3oool00`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3o
ool0:03oool00`000?l0oooo0?ooo`060?ooo`030?l000000?l0oooo00X0oooo00<0o`000?ooo`3o
ool09`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3o
ool0;03oool00240oooo00<000000?ooo`3oool0A`3oool00`000?l0oooo0?ooo`060?ooo`030000
o`3oool0oooo00T0oooo00<0o`000?ooo`3oool09`3oool00`000?l0oooo0?ooo`070?ooo`030000
o`3oool0oooo00T0oooo00<0o`000?ooo`3oool09`3oool00`000?l0oooo0?ooo`070?ooo`030000
o`3oool0oooo00T0oooo00<0o`000?ooo`3oool0:03oool00`000?l0oooo0?ooo`060?ooo`030?l0
00000?l0oooo00X0oooo00<0o`000?ooo`3oool09`3oool00`000?l0oooo0?ooo`060?ooo`030000
o`3oool0oooo00/0oooo00<0o`000?ooo`3oool0:`3oool00240oooo0P0000170?ooo`030000o`3o
ool0oooo00L0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3o
ool0oooo00H0oooo00<0o`000000o`3oool02P3oool00`3o0000oooo0?ooo`0W0?ooo`030000o`3o
ool0oooo00L0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`0W0?ooo`030000o`3o
ool0oooo00H0oooo00<0o`000000o`3oool02P3oool00`3o0000oooo0?ooo`0W0?ooo`030000o`3o
ool0oooo00H0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0[0?ooo`008@3oool0
0`000000oooo0?ooo`160?ooo`030000o`3oool0oooo00L0oooo00<0003o0?ooo`3oool02P3oool0
0`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3oool0
0`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00L0oooo00<0003o0?ooo`3oool02P3oool0
0`3o0000oooo0?ooo`0W0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3oool0
0`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3oool0
0`3o0000oooo0?ooo`0[0?ooo`008@3oool00`000000oooo0?ooo`160?ooo`030000o`3oool0oooo
00L0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo
00H0oooo00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo
00H0oooo00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0W0?ooo`030000o`3oool0oooo
00D0oooo00<0o`000?ooo`000?l0303oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo
00H0oooo00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0[0?ooo`008@3oool00`000000
oooo0?ooo`160?ooo`030000o`3oool0oooo00L0oooo00<0003o0?ooo`3oool02P3oool00`3o0000
oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3oool00`3o0000
oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool0303oool00`3o0000
oooo0?ooo`0V0?ooo`030000o`3oool0oooo00D0oooo00<0o`000?ooo`000?l0303oool00`3o0000
oooo0?ooo`0U0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l03@3oool00`3o0000
oooo0?ooo`0Z0?ooo`008@3oool00`000000oooo0?ooo`160?ooo`030000o`3oool0oooo00H0oooo
00<0o`000?ooo`000?l0303oool00`3o0000oooo0?ooo`0T0?ooo`030000o`3oool0oooo00L0oooo
00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo
00<0o`000000o`3oool0303oool00`3o0000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00H0oooo
00<0o`000?ooo`000?l0303oool00`3o0000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00H0oooo
00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?ooo`0Z0?ooo`008@3oool00`000000oooo0?oo
o`160?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l0303oool00`3o0000oooo0?oo
o`0T0?ooo`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool0303oool00`3o0000oooo0?oo
o`0T0?ooo`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool0303oool00`3o0000oooo0?oo
o`0U0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?oo
o`0T0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?oo
o`0Z0?ooo`008@3oool00`000000oooo0?ooo`160?ooo`030000o`3oool0oooo00H0oooo00<0o`00
0?ooo`000?l0303oool00`3o0000oooo0?ooo`0T0?ooo`030000o`3oool0oooo00L0oooo00<0o`00
0000o`3oool0303oool00`3o0000oooo0?ooo`0T0?ooo`030000o`3oool0oooo00L0oooo00<0o`00
0000o`3oool0303oool00`3o0000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00H0oooo00<0o`00
0?ooo`000?l03@3oool00`3o0000oooo0?ooo`0T0?ooo`030000o`3oool0oooo00H0oooo00<0o`00
0?ooo`000?l03@3oool00`3o0000oooo0?ooo`0Z0?ooo`008@3oool00`000000oooo0?ooo`150?oo
o`030000o`3oool0oooo00L0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?ooo`0S0?oo
o`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool0303oool00`3o0000oooo0?ooo`0T0?oo
o`030000o`3oool0oooo00L0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?ooo`0T0?oo
o`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?ooo`0T0?oo
o`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l03P3oool00`3o0000oooo0?ooo`0Y0?oo
o`008@3oool00`000000oooo0?ooo`150?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`00
0?l03@3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`00
0?l03@3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`00
0?l03@3oool00`3o0000oooo0?ooo`0T0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`00
0?l03@3oool00`3o0000oooo0?ooo`0T0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`00
0?l03P3oool00`3o0000oooo0?ooo`0Y0?ooo`008@3oool2000004H0oooo00<0003o0?ooo`3oool0
1`3oool00`3o0000oooo0000o`0=0?ooo`030?l0003oool0oooo02<0oooo00<0003o0?ooo`3oool0
1P3oool0103o0000oooo0?ooo`000?l=0?ooo`030?l0003oool0oooo02<0oooo00<0003o0?ooo`3o
ool01`3oool00`3o0000oooo0000o`0=0?ooo`030?l0003oool0oooo02@0oooo00<0003o0?ooo`3o
ool01P3oool00`3o0000oooo0000o`0>0?ooo`030?l0003oool0oooo0280oooo00<0003o0?ooo`3o
ool01`3oool00`3o0000oooo0000o`0>0?ooo`030?l0003oool0oooo02T0oooo000Q0?ooo`030000
003oool0oooo04D0oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0=0?ooo`030?l0
003oool0oooo0280oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?ooo`000?l=0?ooo`03
0?l0003oool0oooo02<0oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0>0?ooo`03
0?l0003oool0oooo02<0oooo00<0003o0?ooo`3oool01P3oool00`3o0000oooo0000o`0>0?ooo`03
0?l0003oool0oooo0280oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0>0?ooo`03
0?l0003oool0oooo02T0oooo000Q0?ooo`030000003oool0oooo04D0oooo00<0003o0?ooo`3oool0
1`3oool00`3o0000oooo0000o`0>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool0
1`3oool0103o0000oooo0?ooo`000?l>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3o
ool01`3oool0103o0000oooo0?ooo`000?l>0?ooo`030?l0003oool0oooo0280oooo00<0003o0?oo
o`3oool01`3oool00`3o0000oooo0000o`0>0?ooo`030?l0003oool0oooo0280oooo00<0003o0?oo
o`3oool01`3oool00`3o0000oooo0000o`0?0?ooo`030?l0003oool0oooo02P0oooo000Q0?ooo`03
0000003oool0oooo04D0oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0>0?ooo`03
0?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?ooo`000?l>0?oo
o`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?ooo`000?l>
0?ooo`030?l0003oool0oooo0280oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0>
0?ooo`030?l0003oool0oooo0280oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0?
0?ooo`030?l0003oool0oooo02P0oooo000Q0?ooo`030000003oool0oooo04@0oooo00<0003o0?oo
o`3oool0203oool00`3o0000oooo0000o`0>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?oo
o`3oool01`3oool0103o0000oooo0?ooo`000?l>0?ooo`030?l0003oool0oooo0240oooo00<0003o
0?ooo`3oool01`3oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003oool0oooo0240oooo00<0
003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0?0?ooo`030?l0003oool0oooo0240oooo00<0
003o0?ooo`3oool01P3oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003oool0oooo02P0oooo
000Q0?ooo`030000003oool0oooo04@0oooo00<0003o0?ooo`3oool0203oool00`3o0000oooo0000
o`0>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?oo
o`000?l>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo
0?ooo`000?l?0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01P3oool01@3o0000
oooo0?ooo`3oool0003o00h0oooo00<0o`000?ooo`3oool08@3oool00`000?l0oooo0?ooo`060?oo
o`040?l0003oool0oooo0000o`l0oooo00<0o`000?ooo`3oool0:03oool00240oooo00<000000?oo
o`3oool0A03oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000o`l0oooo00<0o`00
0?ooo`3oool0803oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000o`l0oooo00<0
o`000?ooo`3oool0803oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000o`l0oooo
00<0o`000?ooo`3oool08@3oool00`000?l0oooo0?ooo`060?ooo`050?l0003oool0oooo0?ooo`00
0?l03P3oool00`3o0000oooo0?ooo`0P0?ooo`030000o`3oool0oooo00L0oooo00@0o`000?ooo`3o
ool0003o403oool00`3o0000oooo0?ooo`0W0?ooo`008@3oool00`000000oooo0?ooo`140?ooo`03
0000o`3oool0oooo00L0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?ooo`0O0?oo
o`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?ooo`0P
0?ooo`030000o`3oool0oooo00L0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?oo
o`0P0?ooo`030000o`3oool0oooo00L0oooo00D0o`000?ooo`3oool0oooo0000o`0>0?ooo`030?l0
003oool0oooo0200oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?ooo`000?l@0?ooo`03
0?l0003oool0oooo02L0oooo000Q0?ooo`030000003oool0oooo04@0oooo00<0003o0?ooo`3oool0
1`3oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003oool0oooo01l0oooo00<0003o0?ooo`3o
ool0203oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003oool0oooo01l0oooo00<0003o0?oo
o`3oool0203oool0103o0000oooo0?ooo`000?l@0?ooo`030?l0003oool0oooo01l0oooo00<0003o
0?ooo`3oool01`3oool01@3o0000oooo0?ooo`3oool0003o00l0oooo00<0o`000?ooo`3oool07`3o
ool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool0
9`3oool00240oooo0P0000140?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0003o
3`3oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0
003o3`3oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3o
ool0003o403oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00L0oooo00D0o`000?oo
o`3oool0oooo0000o`0?0?ooo`030?l0003oool0oooo01l0oooo00<0003o0?ooo`3oool01`3oool0
103o0000oooo0?ooo`000?l@0?ooo`030?l0003oool0oooo02L0oooo000Q0?ooo`030000003oool0
oooo04<0oooo00<0003o0?ooo`3oool0203oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003o
ool0oooo01l0oooo00<0003o0?ooo`3oool0203oool0103o0000oooo0?ooo`000?l@0?ooo`030?l0
003oool0oooo01h0oooo00<0003o0?ooo`3oool0203oool0103o0000oooo0?ooo`000?l@0?ooo`03
0?l0003oool0oooo01h0oooo00<0003o0?ooo`3oool0203oool01@3o0000oooo0?ooo`3oool0003o
00l0oooo00<0o`000?ooo`3oool07`3oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo
0000oa40oooo00<0o`000?ooo`3oool09P3oool00240oooo00<000000?ooo`3oool0@`3oool00`00
0?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool07@3oool0
0`000?l0oooo0?ooo`090?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool07P3o
ool00`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool0
7P3oool00`000?l0oooo0?ooo`080?ooo`050?l0003oool0oooo0?ooo`000?l03`3oool00`3o0000
oooo0?ooo`0N0?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0003o4@3oool00`3o
0000oooo0?ooo`0V0?ooo`008@3oool00`000000oooo0?ooo`130?ooo`030000o`3oool0oooo00P0
oooo00@0o`000?ooo`3oool0003o403oool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo
00P0oooo00D0o`000?ooo`3oool0oooo0000o`0@0?ooo`030?l0003oool0oooo01h0oooo00<0003o
0?ooo`3oool0203oool0103o0000oooo0?ooo`000?l@0?ooo`030?l0003oool0oooo01h0oooo00<0
003o0?ooo`3oool0203oool01@3o0000oooo0?ooo`3oool0003o00l0oooo00<0o`000?ooo`3oool0
7P3oool00`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa40oooo00<0o`000?ooo`3o
ool09P3oool00100oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`090?ooo`030000
003oool0oooo04<0oooo00<0003o0?ooo`3oool0203oool0103o0000oooo0?ooo`000?l@0?ooo`03
0?l0003oool0oooo01d0oooo00<0003o0?ooo`3oool0203oool01@3o0000oooo0?ooo`3oool0003o
0100oooo00<0o`000?ooo`3oool07P3oool00`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo
0?ooo`000?l04@3oool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo00P0oooo00D0o`00
0?ooo`3oool0oooo0000o`0@0?ooo`030?l0003oool0oooo01d0oooo00<0003o0?ooo`3oool0203o
ool0103o0000oooo0?ooo`000?lA0?ooo`030?l0003oool0oooo02H0oooo000B0?ooo`030?noM000
0000M;oo00`0oooo00<000000?ooo`3oool0@`3oool00`000?l0oooo0?ooo`080?ooo`040?l0003o
ool0oooo0000oa00oooo00<0o`000?ooo`3oool07@3oool00`000?l0oooo0?ooo`080?ooo`050?l0
003oool0oooo0?ooo`000?l0403oool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo00P0
oooo00D0o`000?ooo`3oool0oooo0000o`0A0?ooo`030?l0003oool0oooo01`0oooo00<0003o0?oo
o`3oool02@3oool01@3o0000oooo0?ooo`3oool0003o0100oooo00<0o`000?ooo`3oool07@3oool0
0`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa80oooo00<0o`000?ooo`3oool09@3o
ool00180oooo00<0oooO09a80000M;l0303oool00`000000oooo0?ooo`120?ooo`030000o`3oool0
oooo00T0oooo00D0o`000?ooo`3oool0oooo0000o`0@0?ooo`030?l0003oool0oooo01/0oooo00<0
003o0?ooo`3oool02@3oool01@3o0000oooo0?ooo`3oool0003o0140oooo00<0o`000?ooo`3oool0
703oool00`000?l0oooo0?ooo`080?ooo`050?l0003oool0oooo0?ooo`000?l04@3oool00`3o0000
oooo0?ooo`0L0?ooo`030000o`3oool0oooo00T0oooo00D0o`000?ooo`3oool0oooo0000o`0@0?oo
o`030?l0003oool0oooo01`0oooo00<0003o0?ooo`3oool0203oool00`3o0000oooo0?ooo`020?oo
o`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool09@3oool001<0oooo00<0gia80018W03O
ool02`3oool00`000000oooo0?ooo`120?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3o
ool00P3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo01/0oooo00<0003o0?ooo`3o
ool02@3oool01@3o0000oooo0?ooo`3oool0003o0140oooo00<0o`000?ooo`3oool0703oool00`00
0?l0oooo0?ooo`080?ooo`050?l0003oool0oooo0?ooo`000?l04@3oool00`3o0000oooo0?ooo`0L
0?ooo`030000o`3oool0oooo00T0oooo00D0o`000?ooo`3oool0oooo0000o`0@0?ooo`030?l0003o
ool0oooo01`0oooo00<0003o0?ooo`3oool0203oool00`3o0000oooo0?ooo`020?ooo`030000o`3o
ool0oooo00l0oooo00<0o`000?ooo`3oool09@3oool000<0oooo00<0oooO09a8000000001`000000
0`18W=l0oooo0?ooo`030?ooo`030?oOW01804P0W=oo00/0oooo00<000000?ooo`3oool0@P3oool0
0`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03P3oool0
0`3o0000oooo0?ooo`0K0?ooo`030000o`3oool0oooo00T0oooo00D0o`000?ooo`3oool0oooo0000
o`0A0?ooo`030?l0003oool0oooo01`0oooo00<0003o0?ooo`3oool0203oool01@3o0000oooo0?oo
o`3oool0003o0140oooo00<0o`000?ooo`3oool0703oool00`000?l0oooo0?ooo`090?ooo`050?l0
003oool0oooo0?ooo`000?l0403oool00`3o0000oooo0?ooo`0L0?ooo`030000o`3oool0oooo00P0
oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo02D0
oooo000C0?ooo`030?oo_`1d0000M;oo00/0oooo0`0000120?ooo`030000o`3oool0oooo00P0oooo
00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo01/0oooo
00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00h0oooo
00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?ooo`090?ooo`050?l0003oool0oooo0?ooo`00
0?l04@3oool00`3o0000oooo0?ooo`0K0?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3o
ool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01/0oooo00<0003o0?ooo`3o
ool0203oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3o
ool09@3oool001@0oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`120?ooo`030000
o`3oool0oooo00P0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0
003oool0oooo01X0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`020?ooo`030000
o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06P3oool00`000?l0oooo0?ooo`090?ooo`050?l0
003oool0oooo0?ooo`000?l04P3oool00`3o0000oooo0?ooo`0J0?ooo`030000o`3oool0oooo00T0
oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01/0
oooo00<0003o0?ooo`3oool0203oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0
oooo00<0o`000?ooo`3oool09@3oool001@0oooo00<0gia80018W03Oool02P3oool00`000000oooo
0?ooo`110?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0>0?ooo`030?l0003oool0oooo01X0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo
0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06P3oool00`000?l0oooo
0?ooo`090?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo
0?ooo`0J0?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0?0?ooo`030?l0003oool0oooo01X0oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo
0?ooo`020?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0903oool001@0oooo00<0
omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`110?ooo`030000o`3oool0oooo00T0oooo00<0
o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01T0oooo00<0
003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0
o`000?ooo`3oool06@3oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo0080oooo00<0
003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0J0?ooo`030000o`3oool0oooo00T0oooo00<0
o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01X0oooo00<0
003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo0100oooo00<0
o`000?ooo`3oool0903oool001@0oooo00<0oono07@0001d_ol02P3oool00`000000oooo0?ooo`11
0?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?
0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`02
0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06@3oool00`000?l0oooo0?ooo`0:
0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0I
0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?
0?ooo`030?l0003oool0oooo01X0oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`02
0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0903oool001D0oooo00<0_g@0001d
_`3oool02@3oool00`000000oooo0?ooo`110?ooo`030000o`3oool0oooo00T0oooo00<0o`000?oo
o`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01P0oooo00<0003o0?oo
o`3oool02`3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?oo
o`3oool06@3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo00<0oooo00<0003o0?oo
o`3oool03`3oool00`3o0000oooo0?ooo`0I0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?oo
o`3oool00P3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?oo
o`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo0100oooo00<0o`000?oo
o`3oool0903oool001<0oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo
0440oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo
00l0oooo00<0o`000?ooo`3oool0603oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo
00<0oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo
00X0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo
01T0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo
0100oooo00<0o`000?ooo`3oool06@3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo
0080oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0T0?ooo`008@3oool00`000000
oooo0?ooo`110?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool00P3oool00`000?l0
oooo0?ooo`0?0?ooo`030?l0003oool0oooo01P0oooo00<0003o0?ooo`3oool02P3oool00`3o0000
oooo0?ooo`030?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool05`3oool00`000?l0
oooo0?ooo`0:0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0403oool00`3o0000
oooo0?ooo`0G0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool00P3oool00`000?l0
oooo0?ooo`0@0?ooo`030?l0003oool0oooo01P0oooo00<0003o0?ooo`3oool02P3oool00`3o0000
oooo0?ooo`020?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0903oool00240oooo
00<000000?ooo`3oool0@03oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo00<0oooo
00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0G0?ooo`030000o`3oool0oooo00/0oooo
00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01H0oooo
00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0100oooo
00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo00<0oooo
00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo00X0oooo
00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo02<0oooo
000Q0?ooo`800000@@3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo00<0oooo00<0
003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0G0?ooo`030000o`3oool0oooo00/0oooo00<0
o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01H0oooo00<0
003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0100oooo00<0
o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00<0oooo00<0
003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0G0?ooo`030000o`3oool0oooo00/0oooo00<0
o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo02<0oooo000Q
0?ooo`030000003oool0oooo0400oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`03
0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0;
0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0F
0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@
0?ooo`030?l0003oool0oooo01H0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`03
0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0:
0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0S
0?ooo`008@3oool00`000000oooo0?ooo`100?ooo`030000o`3oool0oooo00T0oooo00<0o`000?oo
o`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01D0oooo00<0003o0?oo
o`3oool0303oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0100oooo00<0o`000?oo
o`3oool05@3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?oo
o`3oool0403oool00`3o0000oooo0?ooo`0F0?ooo`030000o`3oool0oooo00`0oooo00<0o`000?oo
o`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01D0oooo00<0003o0?oo
o`3oool02`3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0140oooo00<0o`000?oo
o`3oool05@3oool00`000?l0oooo0?ooo`0;0?ooo`008@3oool00`000000oooo0?ooo`100?ooo`03
0000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`03
0?l0003oool0oooo01D0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`030?ooo`03
0000o`3oool0oooo0100oooo00<0o`000?ooo`3oool05@3oool00`000?l0oooo0?ooo`0<0?ooo`03
0?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0E0?ooo`03
0000o`3oool0oooo00`0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0@0?ooo`03
0?l0003oool0oooo01D0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`030?ooo`03
0000o`3oool0oooo0140oooo00<0o`000?ooo`3oool05@3oool00`000?l0oooo0?ooo`0;0?ooo`00
8@3oool00`000000oooo0?ooo`0o0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool0
0`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01@0oooo00<0003o0?ooo`3oool0
3@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0
503oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0
4@3oool00`3o0000oooo0?ooo`0D0?ooo`030000o`3oool0oooo00`0oooo00<0o`000?ooo`3oool0
103oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01@0oooo00<0003o0?ooo`3oool0
303oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0
503oool00`000?l0oooo0?ooo`0<0?ooo`008@3oool00`000000oooo0?ooo`0o0?ooo`030000o`3o
ool0oooo00X0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003o
ool0oooo01@0oooo00<0003o0?ooo`3oool03@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3o
ool0oooo0140oooo00<0o`000?ooo`3oool04`3oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003o
ool0oooo00<0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0C0?ooo`030000o`3o
ool0oooo00d0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003o
ool0oooo01@0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`030?ooo`030000o`3o
ool0oooo0140oooo00<0o`000?ooo`3oool0503oool00`000?l0oooo0?ooo`0<0?ooo`008@3oool0
0`000000oooo0?ooo`0o0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00`3oool0
0`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01<0oooo00<0003o0?ooo`3oool03P3oool0
0`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool04P3oool0
0`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool04@3oool0
0`3o0000oooo0?ooo`0C0?ooo`030000o`3oool0oooo00d0oooo00<0o`000?ooo`3oool0103oool0
0`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01<0oooo00<0003o0?ooo`3oool03@3oool0
0`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool04P3oool0
0`000?l0oooo0?ooo`0=0?ooo`008@3oool00`000000oooo0?ooo`0o0?ooo`030000o`3oool0oooo
00X0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo
01<0oooo00<0003o0?ooo`3oool03P3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo
0140oooo00<0o`000?ooo`3oool04P3oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003oool0oooo
00@0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0C0?ooo`030000o`3oool0oooo
00d0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo
01<0oooo00<0003o0?ooo`3oool03@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo
0180oooo00<0o`000?ooo`3oool04P3oool00`000?l0oooo0?ooo`0=0?ooo`008@3oool00`000000
oooo0?ooo`0o0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool0103oool00`000?l0
oooo0?ooo`0?0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool03`3oool00`3o0000
oooo0?ooo`030?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool04@3oool00`000?l0
oooo0?ooo`0>0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool04@3oool00`3o0000
oooo0?ooo`0B0?ooo`030000o`3oool0oooo00h0oooo00<0o`000?ooo`3oool0103oool00`000?l0
oooo0?ooo`0@0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool03P3oool00`3o0000
oooo0?ooo`030?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool04@3oool00`000?l0
oooo0?ooo`0>0?ooo`008@3oool200000400oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo
0?ooo`050?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool04P3oool00`000?l0oooo
0?ooo`0>0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo
0?ooo`0A0?ooo`030000o`3oool0oooo00h0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo
0?ooo`0A0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool03P3oool00`3o0000oooo
0?ooo`040?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo
0?ooo`0>0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`0A0?ooo`030000o`3oool0oooo00h0oooo000Q0?ooo`030000003oool0oooo03h0oooo00<0
003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0100oooo00<0
o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo00D0oooo00<0
003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0A0?ooo`030000o`3oool0oooo00h0oooo00<0
o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo0140oooo00<0
003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140oooo00<0
o`000?ooo`3oool0403oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00@0oooo00<0
003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo00l0oooo000Q
0?ooo`030000003oool0oooo01d0oooo0`3o000N0?ooo`030000o`3oool0oooo00X0oooo00<0o`00
0?ooo`3oool01@3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo0100oooo00<0003o
0?ooo`3oool03`3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0100oooo00<0o`00
0?ooo`3oool0403oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00@0oooo00<0003o
0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo00l0oooo00<0o`00
0?ooo`3oool0103oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo0100oooo00<0003o
0?ooo`3oool03`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140oooo00<0o`00
0?ooo`3oool0403oool00`000?l0oooo0?ooo`0?0?ooo`008@3oool00`000000oooo0?ooo`0K0?oo
o`80o`000`3oool20?l001`0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`050?oo
o`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0403oool00`000?l0oooo0?ooo`0?0?oo
o`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0@0?oo
o`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0A0?oo
o`030?l0003oool0oooo0100oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`040?oo
o`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool03`3oool00`000?l0oooo0?ooo`0@0?oo
o`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0?0?oo
o`030000o`3oool0oooo0100oooo000Q0?ooo`030000003oool0oooo01X0oooo00<0o`000?ooo`3o
ool01@3oool00`3o0000oooo0?ooo`0I0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3o
ool01@3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00l0oooo00<0003o0?ooo`3o
ool0403oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3o
ool03`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3o
ool04@3oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3o
ool0103oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00l0oooo00<0003o0?ooo`3o
ool0403oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3o
ool03`3oool00`000?l0oooo0?ooo`0@0?ooo`008@3oool00`000000oooo0?ooo`0I0?ooo`030?l0
003oool0oooo00H0oooo00<0o`000?ooo`3oool06@3oool00`000?l0oooo0?ooo`0:0?ooo`030?l0
003oool0oooo00D0oooo00<0003o0?oo_`1d00000`0000000`18W=l0oooo0?ooo`030?ooo`040;md
00000000000007Boo`<0oooo00<0o`000?ooo`3oool03`3oool00`000?l0oooo0?ooo`0@0?ooo`03
0?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0;0?ooo`04
0=nLB000000000000000o`80000000<0B9cO0?ooo`3oool00P3oool0102oM00000000000001d_ol7
0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0?
0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0103oool010000?l0oooo0?oo_`1d
00030000000307Boo`3oool0oooo00<0oooo00@0_g@0000000000000M;oo0`3oool00`3o0000oooo
0?ooo`0?0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo
0?ooo`0A0?ooo`030?l0003oool0oooo00h0oooo00<0oooO0000o`18W=l01@3oool0102oM0000000
0000001d_ol80?ooo`008@3oool00`000000oooo0?ooo`0H0?ooo`030?l0003oool0oooo00P0oooo
00<0o`000?ooo`3oool0603oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo00D0oooo
00H0003o0?ooo`3oool0okmd0000001d_ol50?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3o
ool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool01@3o
ool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00`0oooo00<0okmd001d_`000?l00P3o
ool02P2oM7@0_ooo0?ooo`3oool0gia8001d_`3oool0omnL04P00018W=l60?ooo`030?l0003oool0
oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0>0?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool01@3oool01@000?l0oooo0?noM000000007Bo0080oooo00<0
_gA809cOo`3oool00P3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0o`000?ooo`3o
ool03P3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3o
ool04@3oool00`3o0000oooo0?ooo`0?0?ooo`030000o`00B9`0gooo00<0oooo00H0gia8001d_`3o
ool0omnL04P00018W=l70?ooo`008@3oool00`000000oooo0?ooo`0G0?ooo`030?l0003oool0oooo
00T0oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo
00D0oooo00H0003o0?ooo`3oool0oooO09a80000M;l40?ooo`030?oog`2LB000B9cO0080oooo00D0
okmd0000B02Lgol0oooo0?l0000@0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0
1@3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00d0oooo00<0okmd0000o`3oool0
103oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00D0oooo00<0o`000?ooo`3oool0
1@3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3oool0
403oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo00@0oooo00L0omnL04Q8W03Oool0
oooo0?oog`2LB000B9cO0080oooo00D0okmd0000B02Lgol0oooo0?l0000@0?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0
oooo00h0oooo00<0003o04P0B02Lgol00P3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0
W=oo00H0oooo000Q0?ooo`030000003oool0oooo01H0oooo00<0o`000?ooo`3oool02`3oool00`3o
0000oooo0?ooo`0F0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool01@3oool01`00
0?l0oooo0?ooo`3oool0gia80018W03Oool00`3oool00`3ooml0W4P004RLg`020?ooo`050?oog`2L
B00004RL0=ooo`3o0000403oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00D0oooo
00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0=0?ooo`030000o`3o_g@007Bo00@0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool40?ooo`030?l0003oool0oooo00D0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0>0?ooo`030000o`3oool0oooo0100
oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`030?ooo`070?oo_`1d0000M;oo0?oo
o`3ooml0W4P004RLg`020?ooo`060?oog`2LB00004RL0=ooo`3oool0o`003`3oool00`000?l0oooo
0?ooo`0@0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo
0?ooo`0:0?ooo`050?noM00000000000000000000?l00P0000001@00B9`0gooo0?oog`2LB000B9cO
0080oooo00@0oooO09a80000B9`0gooo1@3oool00240oooo0P00000G0?ooo`030?l0003oool0oooo
00/0oooo00<0o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo
00D0oooo00L0003o0?ooo`3oool0oooo0?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3o
ool01@3OW4P0001809cOo`3oool0o`0000l0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`050?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool03@3oool010000?l0oooo
0?noM018W=l30?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol00`3oool00`3o0000
oooo0?ooo`060?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0303oool00`000?l0
oooo0?ooo`0A0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool00`3oool01`3ogi`0
B00007Boo`3oool0oooO09a80000M;l00`3oool01@3OW4P0001809cOo`3oool0o`0000l0oooo00<0
003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0180oooo00<0
o`000?ooo`3oool02P3oool01`3ooml0W4Q809cOo`000?l0oooo0;md0000M;l00P3oool00`3ooml0
W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo000Q0?ooo`030000003oool0oooo01D0oooo
00<0o`000?ooo`3oool03@3oool00`3o0000oooo0?ooo`0E0?ooo`030000o`3oool0oooo00X0oooo
00<0o`000?ooo`3oool01P3oool01`000?l0oooo0?ooo`3oool0oono07@0001d_ol0103oool00`2o
M00004RL0=ooo`020?ooo`050?oOW0180000B9cO0?ooo`3o00003`3oool00`000?l0oooo0?ooo`0A
0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0=
0?ooo`050000o`3oool0oooo0=nLB018W=l00`3oool00`2oM00004RL0=ooo`020?ooo`030?oOW018
0000B9cO00<0oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003o
ool0oooo00`0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`060?ooo`030000o`3o
ool0oooo0080oooo00@0omnL04P0000004P0W=oo0P3oool00`2oM00004RL0=ooo`020?ooo`050?oO
W0180000B9cO0?ooo`3o00003P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00D0
oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0;0?ooo`070?oo_`1d07@0003o0?oo
o`3o_g@004RL0=ooo`020?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01@3oool0
0240oooo00<000000?ooo`3oool0503oool00`3o0000oooo0?ooo`0>0?ooo`030?l0003oool0oooo
01D0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo
0080oooo00<0_g@0001d_`3oool00`3oool00`3OW4P0001809cOo`020?ooo`050?oog`2LB00007Bo
0?ooo`3o00003`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00H0oooo00<0003o
0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0=0?ooo`060000o`3oool0oooo0?ooo`2oM000M;oo
0P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00<0oooo00<0o`000?ooo`3oool0
1P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00`0oooo00<0003o0?ooo`3oool0
4@3oool00`3o0000oooo0?ooo`060?ooo`050000o`3oool0oooo0?oOW01800000P0000001`1d_ol0
oooo0?ooo`3oool0gia80000B02Lgol00P3oool01@3ooml0W4P0001d_`3oool0o`0000h0oooo00<0
003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0180oooo00<0
o`000?ooo`3oool0303oool01P3ogi`0003o0=ooo`3ogi`0B01809cOo`80oooo00<0gia80000B02L
gol00P3oool00`3ooml0W4P0001d_`050?ooo`008@3oool00`000000oooo0?ooo`0A0?ooo`@0003o
4@3oool00`3o0000oooo0?ooo`0D0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool0
1P3oool00`000?l0oooo0?ooo`020?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018W=l0
0`3oool0102oM00004RL0=ooo`3o000>0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3o
ool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00`0oooo00/0003o0?ooo`3o
ool0oooo0?oOW01807@0_ooo0?ooo`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo0080oooo
00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00`0oooo
00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0080oooo
00<0oono07@00000M;l00`3oool00`3ogi`0B00004RLg`030?ooo`040;md0000B9`0gooo0?l000h0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0180
oooo00<0o`000?ooo`3oool0303oool01P000?l0okmd001d_`3ooml0W4P004RLg`80oooo00<0omnL
04P00018W=l00`3oool00`2oM00004RL0=ooo`040?ooo`008@3oool00`000000oooo0?ooo`0?0?oo
o`80003o0P3oool00`3o0000oooo0000o`0@0?ooo`030?l0003oool0oooo01<0oooo00<0003o0?oo
o`3oool02`3oool00`3o0000oooo0?ooo`070?ooo`070000o`3oool0oooo0?ooo`3ogi`0B01809cO
o`030?ooo`040?oog`2LB00004RL0=ooo`80oooo00@0gia80018W03Oool0o`003P3oool00`000?l0
oooo0?ooo`0A0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool04P3oool00`3o0000
oooo0?ooo`0;0?ooo`030000o`3oool0oooo0080oooo00P0oooO09a8001d_ol0oooo0?oog`2LB000
04RL0=ooo`80oooo00<0gia80018W03Oool00P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3o
ool0oooo0180oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003o
ool0oooo00H0oooo00<0003o0?ooo`3oool0103oool0203OW4P007Bo0?ooo`3oool0oooO09a80000
B9`0gooo0P3oool0103OW4P004RL0=ooo`3o000>0?ooo`030000o`3oool0oooo0140oooo00<0o`00
0?ooo`3oool01@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00`0oooo00H0003o
0?ooo`3OW4P0B9cO0;md0000M;l20?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool0103oool00240oooo00<000000?ooo`3oool03P3oool01@000?l0oooo0?ooo`3oool0o`0000<0
oooo0P000?l?0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool02`3oool00`3o0000
oooo0?ooo`070?ooo`070000o`3oool0oooo0?ooo`3ookl0M00007Boo`040?ooo`030?oOW0180000
M;oo0080oooo00@0gia8001d_`3oool0o`003P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003o
ool0oooo00H0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0;0?ooo`030000o`3o
ool0oooo0080oooo00<0oono07@00018W=l00P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000
M;l0oooo0080oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003o
ool0oooo00/0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3o
ool0oooo00@0oooo00<0omnL04P0B02Lgol00P3oool00`3ogi`0B00007Boo`020?ooo`040=nLB000
M;l0oooo0?l000h0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`050?ooo`030000
o`3oool0oooo0180oooo00<0o`000?ooo`3oool0303oool01`000?l0oooo0?ooo`3OW4P0M7Ad0000
M02oool00P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00@0oooo000Q0?ooo`03
0000003oool0oooo00`0oooo0P000?l30?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3o
ool0303oool00`3o0000oooo0?ooo`0B0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3o
ool01`3oool00`000?l0oooo0?ooo`020?ooo`030;md0000M;l0oooo00@0oooo00P0gia80000M02o
ool0oooO09a80018W=l0oooo0?l000h0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?oo
o`070?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool02`3oool020000?l0okmd0018
W03Oool0oooo0=nLB00004P0W=oo0`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00<0
o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00/0oooo00<0
003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`060?ooo`0:0000o`3oool0oooo0?ooo`2oM000
M;oo0?ooo`3OW4P0001809cOo`<0oooo00P0gia80000M02oool0oooO09a80018W=l0oooo0?l000d0
oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0140
oooo00<0o`000?ooo`3oool0303oool01`000?l0oooo0?ooo`3oool0_g@00000B02Lgol00`3oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`D0oooo000Q0?ooo`030000003oool0oooo00/0oooo00<0
003o0?ooo`3oool00`3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo00`0oooo00<0
o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo00L0oooo00P0
003o0?ooo`3oool0oono07@000000000001d0;ooo`D0oooo00@0gia800000000000007Bo0P3oool0
0`3o0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool01`3oool0
0`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00/0oooo00@0003o0?ooo`3ogi`0B0000P00
00000`1d_ol0oooo0?ooo`030?ooo`040=nLB00000000000001d_`@0oooo00<0o`000?ooo`3oool0
1P3oool00`000?l0oooo0?ooo`0C0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool0
4P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0080oooo00@0_g@0000000000000
B9cO1@3oool0103OW4P0000000000000M;l30?ooo`030?l0003oool0oooo00X0oooo00<0003o0?oo
o`3oool04P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0140oooo00<0o`000?oo
o`3oool02`3oool00`000?l0oooo0?ooo`020?ooo`030?oog`2LB000B9cO00@0oooo00@0gia80000
0000000007Bo1P3oool00240oooo00<000000?ooo`3oool02P3oool00`000?l0oooo0?ooo`030?oo
o`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`0A0?oo
o`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0A0?oo
o`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?oo
o`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0A0?oo
o`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool04`3oool00`3o0000oooo0?ooo`0:0?oo
o`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0B0?oo
o`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`060?oo
o`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0C0?oo
o`008@3oool2000000X0oooo00<0003o0?ooo`3oool00`3oool00`3o0000oooo0?ooo`080?ooo`03
0000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0;0?ooo`03
0?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0:0?ooo`03
0000o`3oool0oooo0180oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0A0?ooo`03
0?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?ooo`03
0000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0A0?ooo`03
0?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0:0?ooo`03
0000o`3oool0oooo0180oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`03
0?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04`3oool00240oooo00<000000?ooo`3oool0
1`3oool20000o`D0oooo00<0o`000?ooo`3oool02@3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0
003oool0oooo00l0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`070?ooo`030000
o`3oool0oooo0140oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0
003oool0oooo00L0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0:0?ooo`030000
o`3oool0oooo0140oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0B0?ooo`030?l0
003oool0oooo00T0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?ooo`030000
o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0
003oool0oooo00H0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0:0?ooo`030000
o`3oool0oooo01<0oooo000Q0?ooo`030000003oool0oooo00H0oooo00<0003o0?ooo`3oool01@3o
ool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool03`3o
ool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04@3o
ool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool01`3o
ool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04@3o
ool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02@3o
ool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3o
ool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool01P3o
ool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04`3o
ool00240oooo00<000000?ooo`3oool01@3oool00`000?l0oooo0?ooo`050?ooo`030?l0003oool0
oooo00/0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0
oooo00/0oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo00X0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0
oooo0140oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo00P0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0
oooo0180oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo00T0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0
oooo0180oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0C0?ooo`008@3oool00`00
0000oooo0?ooo`040?ooo`030000o`3oool0oooo00D0oooo00<0o`000?ooo`3oool03@3oool00`00
0?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo00l0oooo00<0003o0?ooo`3oool02`3oool00`3o
0000oooo0?ooo`080?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool02P3oool00`00
0?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`090?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0203oool00`00
0?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`080?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool02@3oool00`00
0?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo01<0oooo000Q0?ooo`030000003oool0oooo0080
oooo0P000?l70?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3oool02`3oool00`3o0000
oooo0?ooo`0>0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool0203oool00`000?l0
oooo0?ooo`0A0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04@3oool00`3o0000
oooo0?ooo`080?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02@3oool00`000?l0
oooo0?ooo`0A0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool00`3o0000
oooo0?ooo`090?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0203oool00`000?l0
oooo0?ooo`0A0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3oool04P3oool00`3o0000
oooo0?ooo`070?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool00`000?l0
oooo0?ooo`0C0?ooo`008@3oool01@000000oooo0?ooo`3oool0003o00P0oooo00<0o`000?ooo`3o
ool0403oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3o
ool02`3oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3o
ool02P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3o
ool04P3oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3o
ool0203oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3o
ool04P3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3o
ool02@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3o
ool04P3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo01<0oooo000Q0?ooo`040000
003oool0oooo0000o`P0oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0:0?ooo`03
0?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`090?ooo`03
0000o`3oool0oooo0100oooo00<0o`000?ooo`3oool02@3oool00`000?l0oooo0?ooo`0B0?ooo`03
0?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`090?ooo`03
0000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0C0?ooo`03
0?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`080?ooo`03
0000o`3oool0oooo0140oooo00<0o`000?ooo`3oool02@3oool00`000?l0oooo0?ooo`0B0?ooo`03
0?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0:0?ooo`03
0000o`3oool0oooo01<0oooo000Q0?ooo`03000000000?l0003o00P0oooo00<0o`000?ooo`3oool0
4P3oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0
303oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0
203oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3oool0
4@3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0
203oool00`000?l0oooo0?ooo`0C0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool0
4`3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0
2@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool0
4@3oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo01@0oooo000G0?ooo`X0000000<0
003o0000000000001`0000000`3o000000000000000D000000030000o`000000000000T0000000<0
o`000000000000003@0000000`000?l000000000000<000000030?l000000000000000T0000000<0
003o0000000000004@0000000`3o0000000000000008000000030000o`00000000000180000000<0
o`000000000000002@0000000`000?l000000000000A000000030?l000000000000000P0000000<0
003o0000000000004P0000000`3o0000000000000009000000030000o`00000000000180000000<0
o`000000000000001`0000000`000?l000000000000C000000030?l000000000000000P0000000<0
003o0000000000004P0000000`3o0000000000000008000000030000o`00000000000180000000<0
o`00000000000000200000000`000?l000000000000B000000030?l000000000000000P0000000<0
003o0000000000004@0000030?ooo`008@3oool00`000000oooo0?ooo`060?ooo`030?l0003oool0
oooo00L0oooo00<000000?ooo`3oool02`3oool00`000?l0oooo0?ooo`020?ooo`030000003oool0
oooo00D0oooo00<0o`000?ooo`3oool0203oool01@000000oooo0?ooo`3oool0003o00d0oooo00<0
o`000?ooo`0000002P3oool00`000?l0oooo0?ooo`050?ooo`030000003oool0oooo00T0oooo00<0
o`000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0003o00l0oooo00<000000?ooo`3oool0
0P3oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool000000140oooo00<0o`000000003oool0
203oool00`000?l0oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0o`000?ooo`3oool0
1@3oool01@000000oooo0?ooo`3oool0003o00h0oooo00<000000?ooo`3oool00`3oool00`3o0000
oooo0?ooo`070?ooo`040000o`3oool0oooo00000180oooo00<0o`000000003oool0203oool00`00
0?l0oooo0?ooo`060?ooo`030000003oool0oooo00T0oooo00<0o`000?ooo`3oool0103oool01@00
0000oooo0?ooo`3oool0003o00l0oooo00<000000?ooo`3oool00P3oool00`3o0000oooo0?ooo`08
0?ooo`040000o`3oool0oooo00000140oooo00<0o`000000003oool0203oool00`000?l0oooo0?oo
o`060?ooo`030000003oool0oooo00/0oooo000Q0?ooo`030000003oool0oooo00@0oooo0P3o000H
0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool0303oool00`000?l0oooo0?ooo`0;
0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool01@3oool00`000000oooo0?ooo`09
0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`09
0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`07
0?ooo`030000003oool0oooo00P0oooo00<0o`000?ooo`3oool02@3oool00`000?l0oooo0?ooo`0B
0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`09
0?ooo`030000o`3oool0oooo00H0oooo00<000000?ooo`3oool02@3oool00`3o0000oooo0?ooo`08
0?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0B
0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool01P3oool00`000000oooo0?ooo`0;
0?ooo`008@3oool010000000oooo0?ooo`3oool30?l001/0oooo00<0003o0?ooo`3oool02@3oool0
0`3o0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool02P3oool0
0`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool0
0`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0203oool0
0`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04P3oool0
0`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool0
0`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool0
0`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0203oool0
0`000?l0oooo0?ooo`0D0?ooo`008@3oool40?l001h0oooo00<0003o0?ooo`3oool02@3oool00`3o
0000oooo0?ooo`0;0?ooo`030000o`3oool0oooo00`0oooo00<0o`000?ooo`3oool02P3oool00`00
0?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`090?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0203oool00`00
0?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`070?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool00`00
0?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04`3oool00`3o
0000oooo0?ooo`080?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0203oool00`00
0?l0oooo0?ooo`0D0?ooo`008@3oool00`000000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00T0
oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00/0
oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo0180
oooo00<0o`000?ooo`3oool02@3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00P0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo0180
oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00X0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo0180
oooo00<0o`000?ooo`3oool02@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00P0
oooo00<0003o0?ooo`3oool0503oool00240oooo00<000000?ooo`3oool0803oool00`000?l0oooo
0?ooo`090?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo
0?ooo`0;0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo
0?ooo`0B0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`080?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo
0?ooo`0B0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo
0?ooo`0:0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo
0?ooo`0B0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`080?ooo`030000o`3oool0oooo01@0oooo000Q0?ooo`030000003oool0oooo0200oooo00<0
003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo00`0oooo00<0
o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00P0oooo00<0
003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo0180oooo00<0
o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00X0oooo00<0
003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo0180oooo00<0
o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00L0oooo00<0
003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo0140oooo00<0
o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0D0?ooo`008@3oool200000240oooo00<0003o
0?ooo`3oool02@3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo00`0oooo00<0o`00
0?ooo`3oool02`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00P0oooo00<0003o
0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0;0?ooo`030000o`3oool0oooo0140oooo00<0o`00
0?ooo`3oool01`3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00X0oooo00<0003o
0?ooo`3oool04P3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo0180oooo00<0o`00
0?ooo`3oool02P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00L0oooo00<0003o
0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo0140oooo00<0o`00
0?ooo`3oool01`3oool00`000?l0oooo0?ooo`0E0?ooo`008@3oool00`000000oooo0?ooo`0Q0?oo
o`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0<0?oo
o`030?l0003oool0oooo00/0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`070?oo
o`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0A0?oo
o`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0:0?oo
o`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0C0?oo
o`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`070?oo
o`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0A0?oo
o`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool05@3oool00240oooo00<000000?ooo`3o
ool08@3oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3o
ool02`3oool00`3o0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3o
ool01`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00/0oooo00<0003o0?ooo`3o
ool04@3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3o
ool02P3oool00`000?l0oooo0?ooo`0C0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3o
ool04`3oool00`3o0000oooo0?ooo`0;0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3o
ool01`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3o
ool04@3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo01D0oooo000Q0?ooo`030000
003oool0oooo0280oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0?ooo`090?ooo`030000
o`3oool0oooo00`0oooo00<0o`000?ooo`3oool0303oool00`000?l0oooo0?ooo`0A0?ooo`030?l0
003oool0oooo00H0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0;0?ooo`030000
o`3oool0oooo0140oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0
003oool0oooo00/0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`050?ooo`030000
o`3oool0oooo01<0oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0
003oool0oooo00L0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0:0?ooo`030000
o`3oool0oooo0140oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0E0?ooo`008@3o
ool00`000000oooo0?ooo`0R0?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool0203o
ool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0403o
ool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02`3o
ool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool04P3o
ool00`3o0000oooo0?ooo`0;0?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool01@3o
ool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00`0oooo00<0003o0?ooo`3oool04@3o
ool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool02`3o
ool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool05@3o
ool00240oooo00<000000?ooo`3oool08P3oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0
oooo00P0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`0=0?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0
oooo00/0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0
oooo0140oooo00<0o`000?ooo`3oool0303oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0
oooo00D0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0<0?ooo`030000o`3oool0
oooo0140oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo00`0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0
oooo01D0oooo000Q0?ooo`030000003oool0oooo02<0oooo00<0003o0?ooo`3oool01`3oool00`3o
0000oooo0?ooo`080?ooo`030000o`3oool0oooo00`0oooo00<0o`000?ooo`3oool03@3oool00`00
0?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool01P3oool00`00
0?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00`0oooo00<0003o0?ooo`3oool04P3oool00`3o
0000oooo0?ooo`050?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0303oool00`00
0?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool04@3oool00`3o
0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool01P3oool00`00
0?l0oooo0?ooo`0E0?ooo`008@3oool00`000000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00L0
oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00d0
oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0140
oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00H0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo0180
oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00d0
oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo0140
oooo00<0o`000?ooo`3oool0303oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00H0
oooo00<0003o0?ooo`3oool05@3oool00240oooo00<000000?ooo`3oool08`3oool00`000?l0oooo
0?ooo`070?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo
0?ooo`0=0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo
0?ooo`0A0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo
0?ooo`060?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo
0?ooo`0A0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo
0?ooo`0=0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo
0?ooo`0A0?ooo`030?l0003oool0oooo00`0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`060?ooo`030000o`3oool0oooo01D0oooo000Q0?ooo`8000009@3oool00`000?l0oooo0?oo
o`070?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool03@3oool00`3o0000oooo0?oo
o`0>0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?oo
o`0A0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?oo
o`060?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo0?oo
o`0A0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?oo
o`0=0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?oo
o`0B0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?oo
o`060?ooo`030000o`3oool0oooo01D0oooo000Q0?ooo`030000003oool0oooo02@0oooo00<0003o
0?ooo`3oool01`3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo00`0oooo00<0o`00
0?ooo`3oool03`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00H0oooo00<0003o
0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0=0?ooo`030000o`3oool0oooo0100oooo00<0o`00
0?ooo`3oool01P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00h0oooo00<0003o
0?ooo`3oool0403oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0180oooo00<0o`00
0?ooo`3oool03@3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo00D0oooo00<0003o
0?ooo`3oool04P3oool00`3o0000oooo0?ooo`0=0?ooo`030000o`3oool0oooo0100oooo00<0o`00
0?ooo`3oool01P3oool00`000?l0oooo0?ooo`0E0?ooo`008@3oool00`000000oooo0?ooo`0T0?oo
o`030000o`3oool0oooo00L0oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0<0?oo
o`030?l0003oool0oooo00l0oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`060?oo
o`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool03P3oool00`000?l0oooo0?ooo`0@0?oo
o`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0>0?oo
o`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0B0?oo
o`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`050?oo
o`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo0?ooo`0@0?oo
o`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool05P3oool00240oooo00<000000?ooo`3o
ool09@3oool00`000?l0oooo0?ooo`060?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3o
ool0303oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3o
ool01P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3o
ool0403oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3o
ool03`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3o
ool04@3oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3o
ool01@3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3o
ool0403oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo01H0oooo000Q0?ooo`030000
003oool0oooo02D0oooo00<0003o0?ooo`3oool01P3oool00`3o0000oooo0?ooo`060?ooo`030000
o`3oool0oooo00`0oooo00<0o`000?ooo`3oool03`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0
003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0?0?ooo`030000
o`3oool0oooo0100oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0
003oool0oooo0100oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`040?ooo`030000
o`3oool0oooo0140oooo00<0o`000?ooo`3oool03`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0
003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0?0?ooo`030000
o`3oool0oooo00l0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0F0?ooo`008@3o
ool00`000000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`3oool01@3o
ool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00l0oooo00<0003o0?ooo`3oool03`3o
ool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool0403o
ool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool0403o
ool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0103o
ool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00l0oooo00<0003o0?ooo`3oool03`3o
ool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool03`3o
ool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool05P3o
ool00240oooo00<000000?ooo`3oool09@3oool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0
oooo00@0oooo00<0003o0?ooo`3oool03@3oool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0
oooo00l0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo0100oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0
oooo00@0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0
oooo00h0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo00l0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0
oooo01H0oooo000Q0?ooo`030000003oool0oooo02H0oooo00<0003o0?ooo`3oool01P3oool00`3o
0000oooo0?ooo`040?ooo`030000o`3oool0oooo00d0oooo00<0o`000?ooo`3oool0403oool00`00
0?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool04@3oool00`3o
0000oooo0?ooo`0A0?ooo`030000o`3oool0oooo00h0oooo00<0o`000?ooo`3oool0103oool00`00
0?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo0140oooo00<0003o0?ooo`3oool03`3oool00`3o
0000oooo0?ooo`030?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool0403oool00`00
0?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool04@3oool00`3o
0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool0103oool00`00
0?l0oooo0?ooo`0F0?ooo`008@3oool00`000000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0
oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003oool0oooo0100
oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140
oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo00@0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0A0?ooo`030000o`3oool0oooo00l0
oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo0140
oooo00<0003o0?ooo`3oool03P3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140
oooo00<0o`000?ooo`3oool0403oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo00@0
oooo00<0003o0?ooo`3oool05P3oool00240oooo0P00000W0?ooo`030000o`3oool0oooo00H0oooo
00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo0180oooo
00<0003o0?ooo`3oool03P3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140oooo
00<0o`000?ooo`3oool04P3oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003oool0oooo00@0oooo
00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0B0?ooo`030000o`3oool0oooo00h0oooo
00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo0140oooo
00<0003o0?ooo`3oool03P3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0140oooo
00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo00@0oooo
00<0003o0?ooo`3oool05P3oool00240oooo00<000000?ooo`3oool09P3oool00`000?l0oooo0?oo
o`060?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?oo
o`0B0?ooo`030000o`3oool0oooo00h0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?oo
o`0@0?ooo`030?l0003oool0oooo01<0oooo00<0003o0?ooo`3oool03@3oool00`3o0000oooo0?oo
o`040?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool04P3oool00`000?l0oooo0?oo
o`0>0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?oo
o`0C0?ooo`030000o`3oool0oooo00d0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?oo
o`0A0?ooo`030?l0003oool0oooo0140oooo00<0003o0?ooo`3oool03P3oool00`3o0000oooo0?oo
o`040?ooo`030000o`3oool0oooo01H0oooo000Q0?ooo`030000003oool0oooo02L0oooo00<0003o
0?ooo`3oool01P3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo00`0oooo00<0o`00
0?ooo`3oool04P3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo00@0oooo00<0003o
0?ooo`3oool0403oool00`3o0000oooo0?ooo`0C0?ooo`030000o`3oool0oooo00d0oooo00<0o`00
0?ooo`3oool0103oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo01<0oooo00<0003o
0?ooo`3oool03@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo0140oooo00<0o`00
0?ooo`3oool04`3oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003oool0oooo00@0oooo00<0003o
0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0B0?ooo`030000o`3oool0oooo00d0oooo00<0o`00
0?ooo`3oool0103oool00`000?l0oooo0?ooo`0F0?ooo`008@3oool00`000000oooo0?ooo`0W0?oo
o`030000o`3oool0oooo00H0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0=0?oo
o`030?l0003oool0oooo01<0oooo00<0003o0?ooo`3oool03@3oool00`3o0000oooo0?ooo`040?oo
o`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0503oool00`000?l0oooo0?ooo`0<0?oo
o`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0D0?oo
o`030000o`3oool0oooo00d0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0A0?oo
o`030?l0003oool0oooo01@0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`040?oo
o`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool04`3oool00`000?l0oooo0?ooo`0=0?oo
o`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool05P3oool00240oooo00<000000?ooo`3o
ool09`3oool00`000?l0oooo0?ooo`060?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3o
ool03@3oool00`3o0000oooo0?ooo`0C0?ooo`030000o`3oool0oooo00d0oooo00<0o`000?ooo`3o
ool0103oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01@0oooo00<0003o0?ooo`3o
ool0303oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3o
ool0503oool00`000?l0oooo0?ooo`0=0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3o
ool04@3oool00`3o0000oooo0?ooo`0D0?ooo`030000o`3oool0oooo00`0oooo00<0o`000?ooo`3o
ool0103oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01@0oooo00<0003o0?ooo`3o
ool0303oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo01H0oooo000Q0?ooo`030000
003oool0oooo02L0oooo00<0003o0?ooo`3oool01P3oool00`3o0000oooo0?ooo`020?ooo`030000
o`3oool0oooo00d0oooo00<0o`000?ooo`3oool0503oool00`000?l0oooo0?ooo`0<0?ooo`030?l0
003oool0oooo00@0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0D0?ooo`030000
o`3oool0oooo00d0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0
003oool0oooo01D0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`030?ooo`030000
o`3oool0oooo0100oooo00<0o`000?ooo`3oool05@3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0
003oool0oooo00@0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0D0?ooo`030000
o`3oool0oooo00`0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0F0?ooo`008@3o
ool00`000000oooo0?ooo`0X0?ooo`030000o`3oool0oooo00D0oooo00<0o`000?ooo`3oool00P3o
ool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo01D0oooo00<0003o0?ooo`3oool0303o
ool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool05@3o
ool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0403o
ool00`3o0000oooo0?ooo`0E0?ooo`030000o`3oool0oooo00`0oooo00<0o`000?ooo`3oool00`3o
ool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01H0oooo00<0003o0?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool05@3o
ool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool05`3o
ool00240oooo00<000000?ooo`3oool0:03oool00`000?l0oooo0?ooo`050?ooo`030?l0003oool0
oooo0080oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`0F0?ooo`030000o`3oool0
oooo00/0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0
oooo01H0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0003oool0
oooo0080oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0F0?ooo`030000o`3oool0
oooo00/0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0
oooo01D0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0
oooo01L0oooo000Q0?ooo`030000003oool0oooo02P0oooo00<0003o0?ooo`3oool01P3oool01@3o
0000oooo0?ooo`3oool0003o00h0oooo00<0o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0;
0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0G
0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0@
0?ooo`030?l0003oool0oooo01H0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`02
0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0;
0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0F
0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0G
0?ooo`008@3oool2000002T0oooo00<0003o0?ooo`3oool01P3oool01@3o0000oooo0?ooo`3oool0
003o00h0oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0
oooo00<0oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0G0?ooo`030000o`3oool0
oooo00/0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0
oooo01P0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0
oooo00<0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`0F0?ooo`030000o`3oool0
oooo00/0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0G0?ooo`008@3oool00`00
0000oooo0?ooo`0Y0?ooo`030000o`3oool0oooo00D0oooo00@0o`000?ooo`3oool0003o3`3oool0
0`3o0000oooo0?ooo`0G0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool00`3oool0
0`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01L0oooo00<0003o0?ooo`3oool02`3oool0
0`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool0603oool0
0`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03`3oool0
0`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00`3oool0
0`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo01P0oooo00<0003o0?ooo`3oool02P3oool0
0`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo01L0oooo000Q0?ooo`030000003oool0oooo
02T0oooo00<0003o0?ooo`3oool01@3oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003oool0
oooo01P0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0
oooo00l0oooo00<0o`000?ooo`3oool0603oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0
oooo00<0oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0I0?ooo`030000o`3oool0
oooo00X0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0
oooo01P0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool0603oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0
oooo0080oooo00<0003o0?ooo`3oool05`3oool00240oooo00<000000?ooo`3oool0:@3oool00`00
0?l0oooo0?ooo`050?ooo`040?l0003oool0oooo0000o`h0oooo00<0o`000?ooo`3oool06@3oool0
0`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool0403oool0
0`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00`3oool0
0`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?ooo`3oool02P3oool0
0`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06@3oool0
0`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool0403oool0
0`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00P3oool0
0`000?l0oooo0?ooo`0G0?ooo`00403oool00`3ookl0M000000000020000000304RLg`3oool0oooo
00T0oooo00<000000?ooo`3oool0:@3oool00`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo
0000o`h0oooo00<0o`000?ooo`3oool06P3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0
oooo0080oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0I0?ooo`030000o`3oool0
oooo00T0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0
oooo01T0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0
oooo00l0oooo00<0o`000?ooo`3oool06@3oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0
oooo0080oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`0I0?ooo`030000o`3oool0
oooo00T0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0G0?ooo`004P3oool00`3o
_g@0000007Boo`0<0?ooo`030000003oool0oooo02T0oooo00<0003o0?ooo`3oool01P3oool00`3o
0000oooo0000o`0>0?ooo`030?l0003oool0oooo01X0oooo00<0003o0?ooo`3oool02@3oool00`3o
0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06P3oool00`00
0?l0oooo0?ooo`090?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool03`3oool00`3o
0000oooo0?ooo`0J0?ooo`030000o`3oool0oooo00T0oooo00<0o`000?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?ooo`3oool02P3oool00`3o
0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06P3oool00`00
0?l0oooo0?ooo`090?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool05`3oool00180
oooo00<0oooO09a80000M;l0303oool00`000000oooo0?ooo`0Z0?ooo`030000o`3oool0oooo00D0
oooo00<0o`000?ooo`000?l03P3oool00`3o0000oooo0?ooo`0J0?ooo`030000o`3oool0oooo00T0
oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01X0
oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo00h0
oooo00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo0080
oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0J0?ooo`030000o`3oool0oooo00T0
oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0?0?ooo`030?l0003oool0oooo01X0
oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo01L0
oooo000C0?ooo`030=nLB000B9`0gooo00/0oooo00<000000?ooo`3oool0:P3oool00`000?l0oooo
0?ooo`050?ooo`030?l0003oool0003o00h0oooo00<0o`000?ooo`3oool06P3oool00`000?l0oooo
0?ooo`090?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo
0?ooo`0K0?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0?0?ooo`030?l0003oool0oooo01/0oooo00<0003o0?ooo`3oool02@3oool01@3o0000oooo
0?ooo`3oool0003o0140oooo00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?ooo`090?ooo`03
0?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03`3oool00`3o0000oooo0?ooo`0K0?ooo`03
0000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0G0?ooo`00
4`3oool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0oooo02X0oooo00<0003o0?ooo`3oool0
1@3oool00`3o0000oooo0000o`0=0?ooo`030?l0003oool0oooo01`0oooo00<0003o0?ooo`3oool0
203oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool0
6`3oool00`000?l0oooo0?ooo`090?ooo`050?l0003oool0oooo0?ooo`000?l04@3oool00`3o0000
oooo0?ooo`0L0?ooo`030000o`3oool0oooo00P0oooo00D0o`000?ooo`3oool0oooo0000o`0A0?oo
o`030?l0003oool0oooo01/0oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`020?oo
o`030000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?ooo`080?oo
o`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool05`3oool001<0oooo00<0oono07@0001d
_ol02`3oool3000002X0oooo00<0003o0?ooo`3oool01@3oool00`3o0000003o0?ooo`0=0?ooo`03
0?l0003oool0oooo01`0oooo00<0003o0?ooo`3oool0203oool00`3o0000oooo0?ooo`020?ooo`03
0000o`3oool0oooo00l0oooo00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?ooo`090?ooo`05
0?l0003oool0oooo0?ooo`000?l04@3oool00`3o0000oooo0?ooo`0L0?ooo`030000o`3oool0oooo
00P0oooo00D0o`000?ooo`3oool0oooo0000o`0A0?ooo`030?l0003oool0oooo01`0oooo00<0003o
0?ooo`3oool0203oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00h0oooo00<0o`00
0?ooo`3oool0703oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo0080oooo00<0003o
0?ooo`3oool05`3oool001@0oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`0Z0?oo
o`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool0303oool00`3o0000oooo0?ooo`0L0?oo
o`030000o`3oool0oooo00T0oooo00D0o`000?ooo`3oool0oooo0000o`0A0?ooo`030?l0003oool0
oooo01/0oooo00<0003o0?ooo`3oool02@3oool01@3o0000oooo0?ooo`3oool0003o0140oooo00<0
o`000?ooo`3oool0703oool00`000?l0oooo0?ooo`090?ooo`040?l0003oool0oooo0000oa40oooo
00<0o`000?ooo`3oool0703oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo0080oooo
00<0003o0?ooo`3oool03P3oool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo00P0oooo
00D0o`000?ooo`3oool0oooo0000o`0I0?ooo`00503oool00`3OW4P004RL0=ooo`0:0?ooo`030000
003oool0oooo02/0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?ooo`0<0?ooo`030?l0
003oool0oooo01`0oooo00<0003o0?ooo`3oool02@3oool01@3o0000oooo0?ooo`3oool0003o0100
oooo00<0o`000?ooo`3oool07@3oool00`000?l0oooo0?ooo`080?ooo`050?l0003oool0oooo0?oo
o`000?l04@3oool00`3o0000oooo0?ooo`0L0?ooo`030000o`3oool0oooo00T0oooo00@0o`000?oo
o`3oool0003o4@3oool00`3o0000oooo0?ooo`0L0?ooo`030000o`3oool0oooo00P0oooo00<0o`00
0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0>0?ooo`030?l0003oool0oooo01d0oooo00<0003o
0?ooo`3oool0203oool01@3o0000oooo0?ooo`3oool0003o01T0oooo000D0?ooo`030?oOW01804P0
W=oo00X0oooo00<000000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0
oooo00/0oooo00<0o`000?ooo`3oool07P3oool00`000?l0oooo0?ooo`080?ooo`050?l0003oool0
oooo0?ooo`000?l0403oool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo00P0oooo00D0
o`000?ooo`3oool0oooo0000o`0@0?ooo`030?l0003oool0oooo01d0oooo00<0003o0?ooo`3oool0
2@3oool0103o0000oooo0?ooo`000?lA0?ooo`030?l0003oool0oooo01`0oooo00<0003o0?ooo`3o
ool0203oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo00h0oooo00<0o`000?ooo`3o
ool07@3oool00`000?l0oooo0?ooo`080?ooo`050?l0003oool0oooo0?ooo`000?l06@3oool001@0
oooo00<0oono07@0001d_ol02P3oool00`000000oooo0?ooo`0[0?ooo`030000o`3oool0oooo00D0
oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0N0?ooo`030000o`3oool0oooo00P0
oooo00D0o`000?ooo`3oool0oooo0000o`0@0?ooo`030?l0003oool0oooo01d0oooo00<0003o0?oo
o`3oool0203oool01@3o0000oooo0?ooo`3oool0003o0100oooo00<0o`000?ooo`3oool07@3oool0
0`000?l0oooo0?ooo`090?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool07P3o
ool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool03P3o
ool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0003o
6P3oool001D0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`0[0?ooo`030000o`3o
ool0oooo00D0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0N0?ooo`030000o`3o
ool0oooo00P0oooo00D0o`000?ooo`3oool0oooo0000o`0@0?ooo`030?l0003oool0oooo01d0oooo
00<0003o0?ooo`3oool0203oool01@3o0000oooo0?ooo`3oool0003o0100oooo00<0o`000?ooo`3o
ool07P3oool00`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?oo
o`3oool07P3oool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0oooo0080oooo00<0003o0?oo
o`3oool03@3oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00L0oooo00@0o`000?oo
o`3oool0003o6P3oool001<0oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0
oooo02`0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`0;0?ooo`030?l0003oool0
oooo01h0oooo00<0003o0?ooo`3oool0203oool01@3o0000oooo0?ooo`3oool0003o0100oooo00<0
o`000?ooo`3oool07P3oool00`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo0?ooo`000?l0
403oool00`3o0000oooo0?ooo`0N0?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0
003o403oool00`3o0000oooo0?ooo`0N0?ooo`030000o`3oool0oooo00P0oooo00D0o`000?ooo`3o
ool0oooo0000o`0?0?ooo`030?l0003oool0oooo01l0oooo00<0003o0?ooo`3oool01`3oool0103o
0000oooo0?ooo`000?lJ0?ooo`008@3oool00`000000oooo0?ooo`0/0?ooo`030000o`3oool0oooo
00<0oooo00<0003o0?ooo`3o00002`3oool00`3o0000oooo0?ooo`0P0?ooo`030000o`3oool0oooo
00L0oooo00D0o`000?ooo`3oool0oooo0000o`0?0?ooo`030?l0003oool0oooo01l0oooo00<0003o
0?ooo`3oool01`3oool01@3o0000oooo0?ooo`3oool0003o0100oooo00<0o`000?ooo`3oool07P3o
ool00`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool0
7P3oool00`000?l0oooo0?ooo`080?ooo`050?l0003oool0oooo0?ooo`000?l03`3oool00`3o0000
oooo0?ooo`0O0?ooo`030000o`3oool0oooo00L0oooo00@0o`000?ooo`3oool0003o6P3oool00240
oooo00<000000?ooo`3oool0;03oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0o`0000/0
oooo00<0o`000?ooo`3oool0803oool00`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000
o`l0oooo00<0o`000?ooo`3oool07`3oool00`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo
0?ooo`000?l03`3oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00P0oooo00@0o`00
0?ooo`3oool0003o403oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00L0oooo00D0
o`000?ooo`3oool0oooo0000o`0?0?ooo`030?l0003oool0oooo01l0oooo00<0003o0?ooo`3oool0
1`3oool0103o0000oooo0?ooo`000?lJ0?ooo`008@3oool2000002d0oooo00<0003o0?ooo`3oool0
0`3oool00`000?l0oooo0?l0000;0?ooo`030?l0003oool0oooo0200oooo00<0003o0?ooo`3oool0
203oool0103o0000oooo0?ooo`000?l?0?ooo`030?l0003oool0oooo01l0oooo00<0003o0?ooo`3o
ool01`3oool01@3o0000oooo0?ooo`3oool0003o00l0oooo00<0o`000?ooo`3oool07`3oool00`00
0?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000oa00oooo00<0o`000?ooo`3oool07`3oool0
0`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo0?ooo`000?l03`3oool00`3o0000oooo0?oo
o`0O0?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`000?l06P3oool00240oooo00<00000
0?ooo`3oool0;@3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0o`0000/0oooo00<0o`00
0?ooo`3oool0803oool00`000?l0oooo0?ooo`080?ooo`040?l0003oool0oooo0000o`l0oooo00<0
o`000?ooo`3oool0803oool00`000?l0oooo0?ooo`060?ooo`050?l0003oool0oooo0?ooo`000?l0
3`3oool00`3o0000oooo0?ooo`0O0?ooo`030000o`3oool0oooo00P0oooo00@0o`000?ooo`3oool0
003o3`3oool00`3o0000oooo0?ooo`0P0?ooo`030000o`3oool0oooo00L0oooo00D0o`000?ooo`3o
ool0oooo0000o`0>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool00`3o
0000oooo0000o`0J0?ooo`008@3oool00`000000oooo0?ooo`0]0?ooo`030000o`3oool0oooo0080
oooo00<0003o0?ooo`3o00002P3oool00`3o0000oooo0?ooo`0R0?ooo`030000o`3oool0oooo00L0
oooo00@0o`000?ooo`3oool0003o3P3oool00`3o0000oooo0?ooo`0Q0?ooo`030000o`3oool0oooo
00L0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?ooo`0P0?ooo`030000o`3oool0
oooo00L0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?ooo`0P0?ooo`030000o`3o
ool0oooo00L0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?ooo`0Q0?ooo`030000
o`3oool0oooo00L0oooo00<0o`000?ooo`000?l06P3oool00240oooo00<000000?ooo`3oool0;@3o
ool00`000?l0oooo0?ooo`020?ooo`040000o`3oool0oooo0?l000T0oooo00<0o`000?ooo`3oool0
8P3oool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0003o00l0oooo00<0o`000?ooo`3oool0
8@3oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000o`l0oooo00<0o`000?ooo`3o
ool0803oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000o`l0oooo00<0o`000?oo
o`3oool0803oool00`000?l0oooo0?ooo`070?ooo`040?l0003oool0oooo0000o`l0oooo00<0o`00
0?ooo`3oool08@3oool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0003o01X0oooo000Q0?oo
o`030000003oool0oooo02h0oooo00@0003o0?ooo`3oool0003o0`3oool00`3o0000oooo0?ooo`06
0?ooo`030?l0003oool0oooo02<0oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0000o`0?
0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?ooo`00
0?l>0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo0?oo
o`000?l?0?ooo`030?l0003oool0oooo0200oooo00<0003o0?ooo`3oool01`3oool0103o0000oooo
0?ooo`000?l?0?ooo`030?l0003oool0oooo0240oooo00<0003o0?ooo`3oool01`3oool00`3o0000
oooo0000o`0J0?ooo`008@3oool00`000000oooo0?ooo`0^0?ooo`040000o`3oool0oooo0000o`<0
oooo00<0o`000?ooo`3oool01P3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00L0
oooo00<0o`000?ooo`000?l03`3oool00`3o0000oooo0?ooo`0Q0?ooo`030000o`3oool0oooo00L0
oooo00@0o`000?ooo`3oool0003o3P3oool00`3o0000oooo0?ooo`0Q0?ooo`030000o`3oool0oooo
00P0oooo00<0o`000?ooo`000?l03`3oool00`3o0000oooo0?ooo`0Q0?ooo`030000o`3oool0oooo
00H0oooo00@0o`000?ooo`3oool0003o3`3oool00`3o0000oooo0?ooo`0Q0?ooo`030000o`3oool0
oooo00L0oooo00<0o`000?ooo`000?l06P3oool00240oooo00<000000?ooo`3oool0;`3oool00`00
0?l0oooo0000o`040?ooo`030?l0003oool0oooo00D0oooo00<0o`000?ooo`3oool08`3oool00`00
0?l0oooo0?ooo`070?ooo`030?l0003oool0003o00l0oooo00<0o`000?ooo`3oool08P3oool00`00
0?l0oooo0?ooo`060?ooo`040?l0003oool0oooo0000o`h0oooo00<0o`000?ooo`3oool08P3oool0
0`000?l0oooo0?ooo`070?ooo`030?l0003oool0003o00h0oooo00<0o`000?ooo`3oool08P3oool0
0`000?l0oooo0?ooo`070?ooo`030?l0003oool0003o00h0oooo00<0o`000?ooo`3oool08P3oool0
0`000?l0oooo0?ooo`070?ooo`030?l0003oool0003o01X0oooo000Q0?ooo`030000003oool0oooo
02l0oooo0P000?l50?ooo`030?l0003oool0oooo00@0oooo00<0o`000?ooo`3oool09@3oool00`00
0?l0oooo0?ooo`060?ooo`030?l0003oool0003o00h0oooo00<0o`000?ooo`3oool08`3oool00`00
0?l0oooo0?ooo`060?ooo`030?l0003oool0003o00l0oooo00<0o`000?ooo`3oool08P3oool00`00
0?l0oooo0?ooo`070?ooo`030?l0003oool0003o00h0oooo00<0o`000?ooo`3oool08P3oool00`00
0?l0oooo0?ooo`070?ooo`030?l0003oool0003o00h0oooo00<0o`000?ooo`3oool08`3oool00`00
0?l0oooo0?ooo`060?ooo`030?l0003oool0003o01X0oooo000Q0?ooo`030000003oool0oooo03H0
oooo00<0o`000?ooo`3oool0103oool00`3o0000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00L0
oooo00<0o`000000o`3oool03@3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00H0
oooo00<0o`000?ooo`000?l03P3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00L0
oooo00<0o`000?ooo`000?l03P3oool00`3o0000oooo0?ooo`0R0?ooo`030000o`3oool0oooo00L0
oooo00<0o`000?ooo`000?l03P3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo00L0
oooo00<0o`000000o`3oool06@3oool00240oooo0P00000g0?ooo`030?l0003oool0oooo00<0oooo
00<0o`000?ooo`3oool09P3oool00`000?l0oooo0?ooo`070?ooo`030?l000000?l0oooo00d0oooo
00<0o`000?ooo`3oool08`3oool00`000?l0oooo0?ooo`060?ooo`030?l0003oool0003o00h0oooo
00<0o`000?ooo`3oool08`3oool00`000?l0oooo0?ooo`070?ooo`030?l000000?l0oooo00h0oooo
00<0o`000?ooo`3oool08P3oool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0003o00h0oooo
00<0o`000?ooo`3oool08`3oool00`000?l0oooo0?ooo`070?ooo`030?l000000?l0oooo01T0oooo
000Q0?ooo`030000003oool0oooo03L0oooo00<0o`000?ooo`3oool00P3oool00`3o0000oooo0?oo
o`0V0?ooo`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool03@3oool00`3o0000oooo0?oo
o`0S0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`000?l03P3oool00`3o0000oooo0?oo
o`0T0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool03@3oool00`3o0000oooo0?oo
o`0S0?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?oo
o`0T0?ooo`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool06@3oool00240oooo00<00000
0?ooo`3oool0=`3oool20?l00080oooo00<0o`000?ooo`3oool09`3oool00`000?l0oooo0?ooo`07
0?ooo`030?l000000?l0oooo00`0oooo00<0o`000?ooo`3oool0903oool00`000?l0oooo0?ooo`06
0?ooo`030?l0003oool0003o00d0oooo00<0o`000?ooo`3oool09@3oool00`000?l0oooo0?ooo`06
0?ooo`030?l000000?l0oooo00d0oooo00<0o`000?ooo`3oool0903oool00`000?l0oooo0?ooo`06
0?ooo`030?l0003oool0003o00d0oooo00<0o`000?ooo`3oool0903oool00`000?l0oooo0?ooo`07
0?ooo`030?l000000?l0oooo01T0oooo000Q0?ooo`030000003oool0oooo03P0oooo103o000Y0?oo
o`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool0303oool00`3o0000oooo0?ooo`0U0?oo
o`030000o`3oool0oooo00D0oooo00<0o`000?ooo`000?l03@3oool00`3o0000oooo0?ooo`0U0?oo
o`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool03@3oool00`3o0000oooo0?ooo`0T0?oo
o`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool0303oool00`3o0000oooo0?ooo`0T0?oo
o`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool06@3oool00240oooo00<000000?ooo`3o
ool0IP3oool00`000?l0oooo0?ooo`060?ooo`030?l000000?l0oooo00`0oooo00<0o`000?ooo`3o
ool09@3oool00`000?l0oooo0?ooo`050?ooo`030?l0003oool0003o00d0oooo00<0o`000?ooo`3o
ool09@3oool00`000?l0oooo0?ooo`060?ooo`030?l000000?l0oooo00d0oooo00<0o`000?ooo`3o
ool0903oool00`000?l0oooo0?ooo`070?ooo`030?l000000?l0oooo00`0oooo00<0o`000?ooo`3o
ool0903oool00`000?l0oooo0?ooo`070?ooo`030?l000000?l0oooo01T0oooo000Q0?ooo`030000
003oool0oooo06H0oooo00<0003o0?ooo`3oool01P3oool00`3o0000003o0?ooo`0<0?ooo`030?l0
003oool0oooo02D0oooo00<0003o0?ooo`3oool01P3oool00`3o0000003o0?ooo`0<0?ooo`030?l0
003oool0oooo02D0oooo00<0003o0?ooo`3oool01P3oool00`3o0000003o0?ooo`0<0?ooo`030?l0
003oool0oooo02D0oooo00<0003o0?ooo`3oool01`3oool00`3o0000003o0?ooo`0<0?ooo`030?l0
003oool0oooo02@0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`0I0?ooo`008@3o
ool00`000000oooo0?ooo`1V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3o
ool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3o
ool00`3o0000oooo0?ooo`0W0?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool0303o
ool00`3o0000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool02`3o
ool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`0;0?ooo`008@3oool00`000000oooo0?ooo`1V0?ooo`030000o`3oool0
oooo00H0oooo00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0
oooo00H0oooo00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0W0?ooo`030000o`3oool0
oooo00D0oooo00<0o`000000o`3oool0303oool00`3o0000oooo0?ooo`0U0?ooo`030000o`3oool0
oooo00L0oooo00<0o`000000o`3oool02`3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0
oooo00H0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?ooo`0;0?ooo`008@3oool00`00
0000oooo0?ooo`1V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3oool00`3o
0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0o`000000o`3oool02`3oool00`3o
0000oooo0?ooo`0W0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool02`3oool00`3o
0000oooo0?ooo`0U0?ooo`030000o`3oool0oooo00L0oooo00<0o`000000o`3oool02`3oool00`3o
0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?l0003oool02P3oool00`3o
0000oooo0?ooo`0<0?ooo`008@3oool2000006L0oooo00<0003o0?ooo`3oool01`3oool00`000?l0
oooo0?ooo`0:0?ooo`030?l0003oool0oooo02H0oooo00<0003o0?ooo`3oool01P3oool00`3o0000
003o0?ooo`0:0?ooo`030?l0003oool0oooo02P0oooo00<0003o0?ooo`3oool01P3oool00`000?l0
oooo0?ooo`0:0?ooo`030?l0003oool0oooo02H0oooo00<0003o0?ooo`3oool01`3oool00`3o0000
003o0?ooo`0;0?ooo`030?l0003oool0oooo02H0oooo00<0003o0?ooo`3oool01P3oool00`000?l0
o`000?ooo`0:0?ooo`030?l0003oool0oooo00`0oooo000Q0?ooo`030000003oool0oooo06H0oooo
00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo02P0oooo
00<0003o0?ooo`3oool01@3oool00`3o0000003o0?ooo`0:0?ooo`030?l0003oool0oooo02P0oooo
00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo02L0oooo
00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo02L0oooo
00<0003o0?ooo`3oool01P3oool00`000?l0o`000?ooo`0:0?ooo`030?l0003oool0oooo00`0oooo
000Q0?ooo`030000003oool0oooo06L0oooo00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?oo
o`090?ooo`030?l0003oool0oooo02P0oooo00<0003o0?ooo`3oool01@3oool00`3o0000003o0?oo
o`0:0?ooo`030?l0003oool0oooo02P0oooo00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?oo
o`0:0?ooo`030?l0003oool0oooo02L0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?oo
o`090?ooo`030?l0003oool0oooo02L0oooo00<0003o0?ooo`3oool01P3oool00`000?l0o`000?oo
o`090?ooo`030?l0003oool0oooo00d0oooo000Q0?ooo`030000003oool0oooo06L0oooo00<0003o
0?ooo`3oool01P3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo02P0oooo00<0003o
0?ooo`3oool01@3oool00`3o0000003o0?ooo`090?ooo`030?l0003oool0oooo02T0oooo00<0003o
0?ooo`3oool01P3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo02P0oooo00<0003o
0?ooo`3oool01`3oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo02L0oooo00<0003o
0?ooo`3oool01P3oool00`000?l0o`000?ooo`090?ooo`030?l0003oool0oooo00d0oooo000Q0?oo
o`030000003oool0oooo06L0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`0:0?oo
o`030?l0003oool0oooo02P0oooo00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?ooo`080?oo
o`030?l0003oool0oooo02T0oooo00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?ooo`090?oo
o`030?l0003oool0oooo02P0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`080?oo
o`030?l0003oool0oooo02P0oooo00<0003o0?ooo`3oool01P3oool00`000?l0o`000?ooo`090?oo
o`030?l0003oool0oooo00d0oooo000Q0?ooo`030000003oool0oooo06L0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0o`000?ooo`090?ooo`030?l0003oool0oooo02T0oooo00<0003o0?ooo`3o
ool01P3oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo02X0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo02T0oooo00<0003o0?ooo`3o
ool01`3oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo02T0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0o`000?ooo`090?ooo`030?l0003oool0oooo00d0oooo000Q0?ooo`030000
003oool0oooo06L0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`090?ooo`030?l0
003oool0oooo02T0oooo00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?ooo`080?ooo`030?l0
003oool0oooo02X0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`080?ooo`030?l0
003oool0oooo02T0oooo00<0003o0?ooo`3oool01P3oool00`000?l0o`000?ooo`080?ooo`030?l0
003oool0oooo02X0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`080?ooo`030?l0
003oool0oooo00h0oooo000Q0?ooo`030000003oool0oooo06L0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`090?ooo`030?l0003oool0oooo02T0oooo00<0003o0?ooo`3oool01P3o
ool00`000?l0oooo0?ooo`070?ooo`030?l0003oool0oooo02/0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`080?ooo`030?l0003oool0oooo02T0oooo00<0003o0?ooo`3oool01P3o
ool00`000?l0o`000?ooo`080?ooo`030?l0003oool0oooo02X0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0o`000?ooo`080?ooo`030?l0003oool0oooo00h0oooo000Q0?ooo`030000003oool0
oooo06L0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?l000080?ooo`030?l0003oool0
oooo02X0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`080?ooo`030?l0003oool0
oooo02/0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`070?ooo`030?l0003oool0
oooo02X0oooo00<0003o0?ooo`3oool01P3oool00`000?l0o`000?ooo`080?ooo`030?l0003oool0
oooo02X0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`080?ooo`030?l0003oool0
oooo00h0oooo000Q0?ooo`800000J03oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0o`00
00P0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0o`00
00P0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo
00L0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo
00L0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0o`00
00L0oooo00<0o`000?ooo`3oool03`3oool00240oooo00<000000?ooo`3oool0I`3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3oool0o`0000L0oooo00<0o`000?ooo`3oool0;03oool00`000?l0
oooo0?ooo`040?ooo`030000o`3oool0o`0000P0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3o0000oooo00H0oooo00<0o`000?ooo`3oool0;03oool00`000?l0
oooo0?ooo`050?ooo`030000o`3oool0o`0000L0oooo00<0o`000?ooo`3oool0:`3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3oool0o`0000L0oooo00<0o`000?ooo`3oool03`3oool00240oooo
00<000000?ooo`3oool0J03oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0o`0000L0oooo
00<0o`000?ooo`3oool0;03oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0o`0000L0oooo
00<0o`000?ooo`3oool0;03oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0o`0000L0oooo
00<0o`000?ooo`3oool0;03oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0o`0000L0oooo
00<0o`000?ooo`3oool0:`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0o`0000H0oooo
00<0o`000?ooo`3oool0403oool00240oooo00<000000?ooo`3oool0J03oool00`000?l0oooo0?oo
o`040?ooo`030000o`3oool0o`0000H0oooo00<0o`000?ooo`3oool0;@3oool00`000?l0oooo0?oo
o`040?ooo`030000o`3oool0o`0000L0oooo00<0o`000?ooo`3oool0;03oool00`000?l0oooo0?oo
o`040?ooo`040000o`3oool0oooo0?l000H0oooo00<0o`000?ooo`3oool0;03oool00`000?l0oooo
0?ooo`050?ooo`030000o`3oool0o`0000H0oooo00<0o`000?ooo`3oool0;03oool00`000?l0oooo
0?ooo`050?ooo`030000o`3oool0o`0000H0oooo00<0o`000?ooo`3oool0403oool00100oooo00<0
gia80000000000000`0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo06P0oooo00<0
003o0?ooo`3oool0103oool00`000?l0oooo0?l000060?ooo`030?l0003oool0oooo02d0oooo00<0
003o0?ooo`3oool0103oool00`000?l0oooo0?l000060?ooo`030?l0003oool0oooo02d0oooo00<0
003o0?ooo`3oool0103oool010000?l0oooo0?ooo`3o00050?ooo`030?l0003oool0oooo02d0oooo
00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?l000060?ooo`030?l0003oool0oooo02`0oooo
00<0003o0?ooo`3oool01@3oool010000?l0oooo0?ooo`3o00040?ooo`030?l0003oool0oooo0140
oooo000A0?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool0203oool00`000000oooo
0?ooo`1X0?ooo`030000o`3oool0oooo00@0oooo00@0003o0?ooo`3oool0o`00103oool00`3o0000
oooo0?ooo`0^0?ooo`030000o`3oool0oooo00@0oooo00@0003o0?ooo`3oool0o`001@3oool00`3o
0000oooo0?ooo`0^0?ooo`030000o`3oool0oooo00<0oooo00@0003o0?ooo`3oool0o`001@3oool0
0`3o0000oooo0?ooo`0]0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3o00001@3oool0
0`3o0000oooo0?ooo`0]0?ooo`030000o`3oool0oooo00D0oooo00@0003o0?ooo`3oool0o`00103o
ool00`3o0000oooo0?ooo`0A0?ooo`004P3oool00`3o_g@007Bo0?ooo`0<0?ooo`030000003oool0
oooo06P0oooo00<0003o0?ooo`3oool0103oool010000?l0oooo0?ooo`3o00030?ooo`030?l0003o
ool0oooo02l0oooo00<0003o0?ooo`3oool0103oool010000?l0oooo0?ooo`3o00040?ooo`030?l0
003oool0oooo02l0oooo00<0003o0?ooo`3oool00`3oool010000?l0oooo0?ooo`3o00040?ooo`03
0?l0003oool0oooo02h0oooo00<0003o0?ooo`3oool01@3oool010000?l0oooo0?ooo`3o00040?oo
o`030?l0003oool0oooo02h0oooo00<0003o0?ooo`3oool0103oool010000?l0oooo0?ooo`3o0004
0?ooo`030?l0003oool0oooo0140oooo000C0?ooo`030?noM000M;l0oooo00/0oooo00<000000?oo
o`3oool0J03oool00`000?l0oooo0?ooo`040?ooo`040000o`3oool0oooo0?l000<0oooo00<0o`00
0?ooo`3oool0;`3oool00`000?l0oooo0?ooo`040?ooo`040000o`3oool0oooo0?l000@0oooo00<0
o`000?ooo`3oool0;`3oool00`000?l0oooo0?ooo`030?ooo`050000o`3oool0oooo0?ooo`3o0000
0`3oool00`3o0000oooo0?ooo`0^0?ooo`030000o`3oool0oooo00D0oooo00@0003o0?ooo`3oool0
o`000`3oool00`3o0000oooo0?ooo`0_0?ooo`030000o`3oool0oooo00@0oooo00@0003o0?ooo`3o
ool0o`00103oool00`3o0000oooo0?ooo`0A0?ooo`00503oool00`3o_g@0B9cO0?ooo`0:0?ooo`03
0000003oool0oooo06P0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`020?l00003
0?ooo`3o0000o`000340oooo00<0003o0?ooo`3oool0103oool01P000?l0oooo0?ooo`3oool0o`00
0?ooo`80o`00<P3oool00`000?l0oooo0?ooo`030?ooo`040000o`3oool0oooo0?ooo`80o`0000<0
oooo0?l0003oool0<03oool00`000?l0oooo0?ooo`050?ooo`060000o`3oool0oooo0?ooo`3o0000
oooo0P3o000a0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool00P3oool0103o0000
oooo0?l0003o000D0?ooo`005@3oool00`3OW4P0B9cO0?ooo`090?ooo`<00000J03oool00`000?l0
oooo0?ooo`040?ooo`040000o`3oool0oooo0?ooo`<0o`00<P3oool00`000?l0oooo0?ooo`040?oo
o`040000o`3oool0oooo0?ooo`<0o`00<`3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0
oooo0080oooo0P3o000c0?ooo`030000o`3oool0oooo00@0oooo00@0003o0?ooo`3oool0oooo0`3o
000b0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool00P3oool30?l001D0oooo000F
0?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0J03oool00`000?l0oooo0?ooo`04
0?ooo`030000o`3oool0oooo03H0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`0g
0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`04
0?ooo`030000o`3oool0oooo03H0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0J
0?ooo`005P3oool00`3ogi`0B01d0;ooo`080?ooo`030000003oool0oooo06T0oooo00<0003o0?oo
o`3oool00`3oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?oo
o`3oool0=`3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03L0oooo00<0003o0?oo
o`3oool0103oool00`000?l0oooo0?ooo`0f0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?oo
o`3oool06P3oool001H0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`1Y0?ooo`03
0000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`030?ooo`03
0000o`3oool0oooo03L0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0g0?ooo`03
0000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0=P3oool00`000?l0oooo0?ooo`030?ooo`03
0000o`3oool0oooo01X0oooo000F0?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0
J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03L0oooo00<0003o0?ooo`3oool0
0`3oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0
=`3oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo03H0oooo00<0003o0?ooo`3oool0
0`3oool00`000?l0oooo0?ooo`0J0?ooo`004P3oool01`3o_g@004RL0=ooo`3oool0gia80000B02L
gol0203oool00`000000oooo0?ooo`1Y0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3o
ool0>03oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03L0oooo00<0003o0?ooo`3o
ool00`3oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3o
ool0=P3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo01X0oooo000C0?ooo`050?oO
W018000000000000001d_ol02@3oool00`000000oooo0?ooo`1Y0?ooo`030000o`3oool0oooo0080
oooo00<0003o0?ooo`3oool0>03oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo03P0
oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0g0?ooo`030000o`3oool0oooo00@0
oooo00<0003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo01X0
oooo000Q0?ooo`030000003oool0oooo06T0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0h0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0>03oool00`000?l0oooo
0?ooo`020?ooo`030000o`3oool0oooo03L0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo
0?ooo`0g0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool06P3oool00240oooo00<0
00000?ooo`3oool0J@3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03P0oooo00<0
003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0i0?ooo`030000o`3oool0oooo0080oooo00<0
003o0?ooo`3oool0=`3oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo03L0oooo00<0
003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0J0?ooo`008@3oool2000006X0oooo00<0003o
0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0h0?ooo`030000o`3oool0oooo0080oooo00<0003o
0?ooo`3oool0>@3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03L0oooo00<0003o
0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0h0?ooo`030000o`3oool0oooo0080oooo00<0003o
0?ooo`3oool06P3oool00240oooo00<000000?ooo`3oool0J@3oool00`000?l0oooo0?ooo`020?oo
o`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0i0?oo
o`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>03oool00`000?l0oooo0?ooo`020?oo
o`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0J0?oo
o`008@3oool00`000000oooo0?ooo`1Y0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3o
ool0>03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03T0oooo00<0003o0?ooo`3o
ool00P3oool00`000?l0oooo0?ooo`0h0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3o
ool0>03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo01X0oooo000Q0?ooo`030000
003oool0oooo06X0oooo00D0003o0?ooo`3oool0oooo0000o`0j0?ooo`030000o`3oool0oooo0080
oooo00<0003o0?ooo`3oool0>@3oool01@000?l0oooo0?ooo`3oool0003o03/0oooo00<0003o0?oo
o`3oool00P3oool00`000?l0oooo0?ooo`0h0?ooo`030000o`3oool0oooo0080oooo00<0003o0?oo
o`3oool06P3oool00240oooo00<000000?ooo`3oool0JP3oool01@000?l0oooo0?ooo`3oool0003o
03/0oooo00D0003o0?ooo`3oool0oooo0000o`0k0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3o
ool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo03P0oooo00<0003o0?ooo`3oool00P3o
ool00`000?l0oooo0?ooo`0J0?ooo`008@3oool00`000000oooo0?ooo`1Z0?ooo`050000o`3oool0
oooo0?ooo`000?l0>`3oool01@000?l0oooo0?ooo`3oool0003o03/0oooo00D0003o0?ooo`3oool0
oooo0000o`0k0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>03oool00`000?l0
oooo0?ooo`020?ooo`030000o`3oool0oooo01X0oooo000Q0?ooo`030000003oool0oooo06X0oooo
00D0003o0?ooo`3oool0oooo0000o`0k0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3oool01@00
0?l0oooo0?ooo`3oool0003o03/0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0h
0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool06P3oool00240oooo00<000000?oo
o`3oool0JP3oool01@000?l0oooo0?ooo`3oool0003o03/0oooo00D0003o0?ooo`3oool0oooo0000
o`0k0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3oool00`000?l0oooo0?ooo`020?ooo`030000
o`3oool0oooo03T0oooo00D0003o0?ooo`3oool0oooo0000o`0L0?ooo`008@3oool2000006/0oooo
00D0003o0?ooo`3oool0oooo0000o`0k0?ooo`050000o`3oool0oooo0?ooo`000?l0?03oool01000
0?l0oooo0?ooo`000?lk0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0>@3oool0
1@000?l0oooo0?ooo`3oool0003o01`0oooo000Q0?ooo`030000003oool0oooo06X0oooo00D0003o
0?ooo`3oool0oooo0000o`0k0?ooo`050000o`3oool0oooo0?ooo`000?l0?03oool010000?l0oooo
0?ooo`000?ll0?ooo`050000o`3oool0oooo0?ooo`000?l0>`3oool01@000?l0oooo0?ooo`3oool0
003o01`0oooo000Q0?ooo`030000003oool0oooo06X0oooo00D0003o0?ooo`3oool0oooo0000o`0k
0?ooo`050000o`3oool0oooo0?ooo`000?l0?03oool010000?l0oooo0?ooo`000?ll0?ooo`050000
o`3oool0oooo0?ooo`000?l0>`3oool01@000?l0oooo0?ooo`3oool0003o01`0oooo000Q0?ooo`03
0000003oool0oooo06X0oooo00@0003o0?ooo`3oool0003o?03oool01@000?l0oooo0?ooo`3oool0
003o03`0oooo00@0003o0?ooo`3oool0003o?03oool01@000?l0oooo0?ooo`3oool0003o03/0oooo
00D0003o0?ooo`3oool0oooo0000o`0L0?ooo`008@3oool00`000000oooo0?ooo`1Z0?ooo`040000
o`3oool0oooo0000ocd0oooo00@0003o0?ooo`3oool0003o?03oool010000?l0oooo0?ooo`000?ll
0?ooo`040000o`3oool0oooo0000oc`0oooo00@0003o0?ooo`3oool0003o7@3oool00240oooo00<0
00000?ooo`3oool0JP3oool010000?l0oooo0?ooo`000?lm0?ooo`040000o`3oool0oooo0000oc`0
oooo00@0003o0?ooo`3oool0003o?03oool010000?l0oooo0?ooo`000?ll0?ooo`040000o`3oool0
oooo0000oad0oooo000Q0?ooo`030000003oool0oooo06X0oooo00@0003o0?ooo`3oool0003o?@3o
ool010000?l0oooo0?ooo`000?ll0?ooo`040000o`3oool0oooo0000oc`0oooo00@0003o0?ooo`3o
ool0003o?@3oool00`000?l0oooo0000o`0M0?ooo`008@3oool00`000000oooo0?ooo`1[0?ooo`03
0000o`3oool0003o03d0oooo00@0003o0?ooo`3oool0003o?03oool010000?l0oooo0?ooo`000?ll
0?ooo`040000o`3oool0oooo0000ocd0oooo00<0003o0?ooo`000?l07@3oool00240oooo00<00000
0?ooo`3oool0J`3oool00`000?l0oooo0000o`0m0?ooo`030000o`3oool0003o03d0oooo00@0003o
0?ooo`3oool0003o?03oool010000?l0oooo0?ooo`000?lm0?ooo`030000o`3oool0003o01d0oooo
000Q0?ooo`800000K03oool00`000?l0oooo0000o`0m0?ooo`030000o`3oool0003o03d0oooo00@0
003o0?ooo`3oool0003o?@3oool00`000?l0oooo0000o`0m0?ooo`030000o`3oool0003o01d0oooo
000Q0?ooo`030000003oool0oooo06/0oooo00<0003o0?ooo`000?l0?@3oool00`000?l0oooo0000
o`0n0?ooo`80003o?P3oool20000och0oooo00<0003o0?ooo`000?l07@3oool00240oooo00<00000
0?ooo`3oool0J`3oool00`000?l0oooo0000o`0m0?ooo`030000o`3oool0003o03h0oooo0P000?ln
0?ooo`80003o?P3oool00`000?l0oooo0000o`0M0?ooo`008@3oool00`000000oooo0?ooo`1/0?oo
o`030000o`3oool0oooo03d0oooo0P000?ln0?ooo`80003o?P3oool20000ocl0oooo0P000?lM0?oo
o`008@3oool00`000000oooo0?ooo`1/0?ooo`030000o`3oool0oooo03d0oooo00<0003o0?ooo`3o
ool0?P3oool00`000?l0oooo0?ooo`0m0?ooo`030000o`3oool0oooo03d0oooo00<0003o0?ooo`3o
ool0703oool00240oooo00<000000?ooo`3oool0o`3ooon=0?ooo`008@3oool00`000000oooo0?oo
o`3o0?ooohd0oooo000Q0?ooo`030000003oool0oooo0?l0ooooS@3oool00240oooo00<000000?oo
o`3oool0o`3ooon=0?ooo`008@3oool200000?l0ooooSP3oool00240oooo00<000000?ooo`3oool0
o`3ooon=0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.43073, -2.98378, \
0.155481, 0.0336783}}]
}, Closed]],

Cell[TextData[
"Even systems which start very near the fixed point evolve toward the limit \
cycle quite quickly.  Notice that when the relaxation parameter \
\[CurlyEpsilon] is relatively small, the oscillation period is near the \
2\[Pi] period of the driving term."], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Case 2: initial point outside limit cycle", "Subsubsection"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseAndField[3, 1, 1, 20, 3.5, 3];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .85853 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.13473 0.429267 0.13473 [
[ 0 0 0 0 ]
[ 1 .85853 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto fill
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto
-0.02 0.005 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.0001 w
[ ] 0 setdash
.02844 .02508 m
.02381 .08263 L
s
% Start of user PostScript
-3.5 -3. -3.5344044 -2.5728117 MAarrowhead1
% End of user PostScript
.02844 .08282 m
.02447 .13292 L
s
% Start of user PostScript
-3.5 -2.5714286 -3.5294895 -2.1995331 MAarrowhead1
% End of user PostScript
.02844 .14056 m
.02513 .18322 L
s
% Start of user PostScript
-3.5 -2.1428571 -3.5245746 -1.8262545 MAarrowhead1
% End of user PostScript
.02844 .1983 m
.0258 .23351 L
s
% Start of user PostScript
-3.5 -1.7142857 -3.5196597 -1.4529759 MAarrowhead1
% End of user PostScript
.02844 .25604 m
.02646 .2838 L
s
% Start of user PostScript
-3.5 -1.2857143 -3.5147448 -1.0796973 MAarrowhead1
% End of user PostScript
.02844 .31378 m
.02712 .33409 L
s
% Start of user PostScript
-3.5 -0.85714286 -3.5098298 -0.70641872 MAarrowhead1
% End of user PostScript
.02844 .37153 m
.02778 .38438 L
s
% Start of user PostScript
-3.5 -0.42857143 -3.5049149 -0.33314011 MAarrowhead1
% End of user PostScript
.02844 .42927 m
.02844 .43467 L
s
% Start of user PostScript
-3.5 0. -3.5 0.040138494 MAarrowhead1
% End of user PostScript
.02844 .48701 m
.02911 .48497 L
s
% Start of user PostScript
-3.5 0.42857143 -3.4950851 0.4134171 MAarrowhead1
% End of user PostScript
.02844 .54475 m
.02977 .53526 L
s
% Start of user PostScript
-3.5 0.85714286 -3.4901702 0.7866957 MAarrowhead1
% End of user PostScript
.02844 .60249 m
.03043 .58555 L
s
% Start of user PostScript
-3.5 1.2857143 -3.4852552 1.1599743 MAarrowhead1
% End of user PostScript
.02844 .66023 m
.03109 .63584 L
s
% Start of user PostScript
-3.5 1.7142857 -3.4803403 1.5332529 MAarrowhead1
% End of user PostScript
.02844 .71797 m
.03176 .68613 L
s
% Start of user PostScript
-3.5 2.1428571 -3.4754254 1.9065315 MAarrowhead1
% End of user PostScript
.02844 .77572 m
.03242 .73643 L
s
% Start of user PostScript
-3.5 2.5714286 -3.4705105 2.2798101 MAarrowhead1
% End of user PostScript
.02844 .83346 m
.03308 .78672 L
s
% Start of user PostScript
-3.5 3. -3.4655956 2.6530887 MAarrowhead1
% End of user PostScript
.09581 .02508 m
.09117 .06679 L
s
% Start of user PostScript
-3. -3. -3.0344044 -2.6903602 MAarrowhead1
% End of user PostScript
.09581 .08282 m
.09184 .11924 L
s
% Start of user PostScript
-3. -2.5714286 -3.0294895 -2.3011081 MAarrowhead1
% End of user PostScript
.09581 .14056 m
.0925 .17168 L
s
% Start of user PostScript
-3. -2.1428571 -3.0245746 -1.911856 MAarrowhead1
% End of user PostScript
.09581 .1983 m
.09316 .22413 L
s
% Start of user PostScript
-3. -1.7142857 -3.0196597 -1.5226039 MAarrowhead1
% End of user PostScript
.09581 .25604 m
.09382 .27657 L
s
% Start of user PostScript
-3. -1.2857143 -3.0147448 -1.1333518 MAarrowhead1
% End of user PostScript
.09581 .31378 m
.09449 .32901 L
s
% Start of user PostScript
-3. -0.85714286 -3.0098298 -0.74409975 MAarrowhead1
% End of user PostScript
.09581 .37153 m
.09515 .38146 L
s
% Start of user PostScript
-3. -0.42857143 -3.0049149 -0.35484766 MAarrowhead1
% End of user PostScript
.09581 .42927 m
.09581 .4339 L
s
% Start of user PostScript
-3. 0. -3. 0.034404423 MAarrowhead1
% End of user PostScript
.09581 .48701 m
.09647 .48635 L
s
% Start of user PostScript
-3. 0.42857143 -2.9950851 0.42365651 MAarrowhead1
% End of user PostScript
.09581 .54475 m
.09713 .53879 L
s
% Start of user PostScript
-3. 0.85714286 -2.9901702 0.8129086 MAarrowhead1
% End of user PostScript
.09581 .60249 m
.0978 .59123 L
s
% Start of user PostScript
-3. 1.2857143 -2.9852552 1.2021607 MAarrowhead1
% End of user PostScript
.09581 .66023 m
.09846 .64368 L
s
% Start of user PostScript
-3. 1.7142857 -2.9803403 1.5914128 MAarrowhead1
% End of user PostScript
.09581 .71797 m
.09912 .69612 L
s
% Start of user PostScript
-3. 2.1428571 -2.9754254 1.9806649 MAarrowhead1
% End of user PostScript
.09581 .77572 m
.09978 .74857 L
s
% Start of user PostScript
-3. 2.5714286 -2.9705105 2.3699169 MAarrowhead1
% End of user PostScript
.09581 .83346 m
.10045 .80101 L
s
% Start of user PostScript
-3. 3. -2.9655956 2.759169 MAarrowhead1
% End of user PostScript
.16317 .02508 m
.15854 .05327 L
s
% Start of user PostScript
-2.5 -3. -2.5344044 -2.7907064 MAarrowhead1
% End of user PostScript
.16317 .08282 m
.1592 .10754 L
s
% Start of user PostScript
-2.5 -2.5714286 -2.5294895 -2.3879383 MAarrowhead1
% End of user PostScript
.16317 .14056 m
.15986 .1618 L
s
% Start of user PostScript
-2.5 -2.1428571 -2.5245746 -1.9851702 MAarrowhead1
% End of user PostScript
.16317 .1983 m
.16053 .21607 L
s
% Start of user PostScript
-2.5 -1.7142857 -2.5196597 -1.5824021 MAarrowhead1
% End of user PostScript
.16317 .25604 m
.16119 .27033 L
s
% Start of user PostScript
-2.5 -1.2857143 -2.5147448 -1.179634 MAarrowhead1
% End of user PostScript
.16317 .31378 m
.16185 .3246 L
s
% Start of user PostScript
-2.5 -0.85714286 -2.5098298 -0.77686587 MAarrowhead1
% End of user PostScript
.16317 .37153 m
.16251 .37886 L
s
% Start of user PostScript
-2.5 -0.42857143 -2.5049149 -0.37409776 MAarrowhead1
% End of user PostScript
.16317 .42927 m
.16317 .43313 L
s
% Start of user PostScript
-2.5 0. -2.5 0.028670353 MAarrowhead1
% End of user PostScript
.16317 .48701 m
.16384 .48739 L
s
% Start of user PostScript
-2.5 0.42857143 -2.4950851 0.43143846 MAarrowhead1
% End of user PostScript
.16317 .54475 m
.1645 .54166 L
s
% Start of user PostScript
-2.5 0.85714286 -2.4901702 0.83420658 MAarrowhead1
% End of user PostScript
.16317 .60249 m
.16516 .59592 L
s
% Start of user PostScript
-2.5 1.2857143 -2.4852552 1.2369747 MAarrowhead1
% End of user PostScript
.16317 .66023 m
.16582 .65019 L
s
% Start of user PostScript
-2.5 1.7142857 -2.4803403 1.6397428 MAarrowhead1
% End of user PostScript
.16317 .71797 m
.16649 .70445 L
s
% Start of user PostScript
-2.5 2.1428571 -2.4754254 2.0425109 MAarrowhead1
% End of user PostScript
.16317 .77572 m
.16715 .75872 L
s
% Start of user PostScript
-2.5 2.5714286 -2.4705105 2.445279 MAarrowhead1
% End of user PostScript
.16317 .83346 m
.16781 .81298 L
s
% Start of user PostScript
-2.5 3. -2.4655956 2.8480471 MAarrowhead1
% End of user PostScript
.23054 .02508 m
.2259 .04207 L
s
% Start of user PostScript
-2. -3. -2.0344044 -2.8738504 MAarrowhead1
% End of user PostScript
.23054 .08282 m
.22657 .09783 L
s
% Start of user PostScript
-2. -2.5714286 -2.0294895 -2.4600238 MAarrowhead1
% End of user PostScript
.23054 .14056 m
.22723 .15358 L
s
% Start of user PostScript
-2. -2.1428571 -2.0245746 -2.0461971 MAarrowhead1
% End of user PostScript
.23054 .1983 m
.22789 .20934 L
s
% Start of user PostScript
-2. -1.7142857 -2.0196597 -1.6323704 MAarrowhead1
% End of user PostScript
.23054 .25604 m
.22855 .26509 L
s
% Start of user PostScript
-2. -1.2857143 -2.0147448 -1.2185437 MAarrowhead1
% End of user PostScript
.23054 .31378 m
.22922 .32085 L
s
% Start of user PostScript
-2. -0.85714286 -2.0098298 -0.80471707 MAarrowhead1
% End of user PostScript
.23054 .37153 m
.22988 .3766 L
s
% Start of user PostScript
-2. -0.42857143 -2.0049149 -0.39089039 MAarrowhead1
% End of user PostScript
.23054 .42927 m
.23054 .43236 L
s
% Start of user PostScript
-2. 0. -2. 0.022936282 MAarrowhead1
% End of user PostScript
.23054 .48701 m
.2312 .48811 L
s
% Start of user PostScript
-2. 0.42857143 -1.9950851 0.43676296 MAarrowhead1
% End of user PostScript
.23054 .54475 m
.23186 .54387 L
s
% Start of user PostScript
-2. 0.85714286 -1.9901702 0.85058963 MAarrowhead1
% End of user PostScript
.23054 .60249 m
.23253 .59962 L
s
% Start of user PostScript
-2. 1.2857143 -1.9852552 1.2644163 MAarrowhead1
% End of user PostScript
.23054 .66023 m
.23319 .65538 L
s
% Start of user PostScript
-2. 1.7142857 -1.9803403 1.678243 MAarrowhead1
% End of user PostScript
.23054 .71797 m
.23385 .71113 L
s
% Start of user PostScript
-2. 2.1428571 -1.9754254 2.0920697 MAarrowhead1
% End of user PostScript
.23054 .77572 m
.23451 .76689 L
s
% Start of user PostScript
-2. 2.5714286 -1.9705105 2.5058963 MAarrowhead1
% End of user PostScript
.23054 .83346 m
.23518 .82264 L
s
% Start of user PostScript
-2. 3. -1.9655956 2.919723 MAarrowhead1
% End of user PostScript
.2979 .02508 m
.29327 .03319 L
s
% Start of user PostScript
-1.5 -3. -1.5344044 -2.9397923 MAarrowhead1
% End of user PostScript
.2979 .08282 m
.29393 .0901 L
s
% Start of user PostScript
-1.5 -2.5714286 -1.5294895 -2.5173645 MAarrowhead1
% End of user PostScript
.2979 .14056 m
.29459 .14702 L
s
% Start of user PostScript
-1.5 -2.1428571 -1.5245746 -2.0949367 MAarrowhead1
% End of user PostScript
.2979 .1983 m
.29526 .20393 L
s
% Start of user PostScript
-1.5 -1.7142857 -1.5196597 -1.6725089 MAarrowhead1
% End of user PostScript
.2979 .25604 m
.29592 .26084 L
s
% Start of user PostScript
-1.5 -1.2857143 -1.5147448 -1.2500811 MAarrowhead1
% End of user PostScript
.2979 .31378 m
.29658 .31776 L
s
% Start of user PostScript
-1.5 -0.85714286 -1.5098298 -0.82765335 MAarrowhead1
% End of user PostScript
.2979 .37153 m
.29724 .37467 L
s
% Start of user PostScript
-1.5 -0.42857143 -1.5049149 -0.40522557 MAarrowhead1
% End of user PostScript
.2979 .42927 m
.2979 .43158 L
s
% Start of user PostScript
-1.5 0. -1.5 0.017202212 MAarrowhead1
% End of user PostScript
.2979 .48701 m
.29857 .4885 L
s
% Start of user PostScript
-1.5 0.42857143 -1.4950851 0.43962999 MAarrowhead1
% End of user PostScript
.2979 .54475 m
.29923 .54541 L
s
% Start of user PostScript
-1.5 0.85714286 -1.4901702 0.86205777 MAarrowhead1
% End of user PostScript
.2979 .60249 m
.29989 .60233 L
s
% Start of user PostScript
-1.5 1.2857143 -1.4852552 1.2844856 MAarrowhead1
% End of user PostScript
.2979 .66023 m
.30055 .65924 L
s
% Start of user PostScript
-1.5 1.7142857 -1.4803403 1.7069133 MAarrowhead1
% End of user PostScript
.2979 .71797 m
.30122 .71615 L
s
% Start of user PostScript
-1.5 2.1428571 -1.4754254 2.1293411 MAarrowhead1
% End of user PostScript
.2979 .77572 m
.30188 .77307 L
s
% Start of user PostScript
-1.5 2.5714286 -1.4705105 2.5517689 MAarrowhead1
% End of user PostScript
.2979 .83346 m
.30254 .82998 L
s
% Start of user PostScript
-1.5 3. -1.4655956 2.9741967 MAarrowhead1
% End of user PostScript
.36527 .02508 m
.36063 .02662 L
s
% Start of user PostScript
-1. -3. -1.0344044 -2.9885319 MAarrowhead1
% End of user PostScript
.36527 .08282 m
.3613 .08436 L
s
% Start of user PostScript
-1. -2.5714286 -1.0294895 -2.5599604 MAarrowhead1
% End of user PostScript
.36527 .14056 m
.36196 .1421 L
s
% Start of user PostScript
-1. -2.1428571 -1.0245746 -2.131389 MAarrowhead1
% End of user PostScript
.36527 .1983 m
.36262 .19985 L
s
% Start of user PostScript
-1. -1.7142857 -1.0196597 -1.7028176 MAarrowhead1
% End of user PostScript
.36527 .25604 m
.36328 .25759 L
s
% Start of user PostScript
-1. -1.2857143 -1.0147448 -1.2742461 MAarrowhead1
% End of user PostScript
.36527 .31378 m
.36395 .31533 L
s
% Start of user PostScript
-1. -0.85714286 -1.0098298 -0.84567472 MAarrowhead1
% End of user PostScript
.36527 .37153 m
.36461 .37307 L
s
% Start of user PostScript
-1. -0.42857143 -1.0049149 -0.41710329 MAarrowhead1
% End of user PostScript
.36527 .42927 m
.36527 .43081 L
s
% Start of user PostScript
-1. 0. -1. 0.011468141 MAarrowhead1
% End of user PostScript
.36527 .48701 m
.36593 .48855 L
s
% Start of user PostScript
-1. 0.42857143 -0.99508508 0.44003957 MAarrowhead1
% End of user PostScript
.36527 .54475 m
.36659 .54629 L
s
% Start of user PostScript
-1. 0.85714286 -0.99017016 0.868611 MAarrowhead1
% End of user PostScript
.36527 .60249 m
.36726 .60404 L
s
% Start of user PostScript
-1. 1.2857143 -0.98525525 1.2971824 MAarrowhead1
% End of user PostScript
.36527 .66023 m
.36792 .66178 L
s
% Start of user PostScript
-1. 1.7142857 -0.98034033 1.7257539 MAarrowhead1
% End of user PostScript
.36527 .71797 m
.36858 .71952 L
s
% Start of user PostScript
-1. 2.1428571 -0.97542541 2.1543253 MAarrowhead1
% End of user PostScript
.36527 .77572 m
.36924 .77726 L
s
% Start of user PostScript
-1. 2.5714286 -0.97051049 2.5828967 MAarrowhead1
% End of user PostScript
.36527 .83346 m
.36991 .835 L
s
% Start of user PostScript
-1. 3. -0.96559558 3.0114681 MAarrowhead1
% End of user PostScript
.43263 .02508 m
.428 .02237 L
s
% Start of user PostScript
-0.5 -3. -0.53440442 -3.0200692 MAarrowhead1
% End of user PostScript
.43263 .08282 m
.42866 .08061 L
s
% Start of user PostScript
-0.5 -2.5714286 -0.52948951 -2.5878116 MAarrowhead1
% End of user PostScript
.43263 .14056 m
.42932 .13885 L
s
% Start of user PostScript
-0.5 -2.1428571 -0.52457459 -2.155554 MAarrowhead1
% End of user PostScript
.43263 .1983 m
.42999 .19709 L
s
% Start of user PostScript
-0.5 -1.7142857 -0.51965967 -1.7232964 MAarrowhead1
% End of user PostScript
.43263 .25604 m
.43065 .25533 L
s
% Start of user PostScript
-0.5 -1.2857143 -0.51474475 -1.2910388 MAarrowhead1
% End of user PostScript
.43263 .31378 m
.43131 .31356 L
s
% Start of user PostScript
-0.5 -0.85714286 -0.50982984 -0.85878116 MAarrowhead1
% End of user PostScript
.43263 .37153 m
.43197 .3718 L
s
% Start of user PostScript
-0.5 -0.42857143 -0.50491492 -0.42652355 MAarrowhead1
% End of user PostScript
.43263 .42927 m
.43263 .43004 L
s
% Start of user PostScript
-0.5 0. -0.5 0.0057340705 MAarrowhead1
% End of user PostScript
.43263 .48701 m
.4333 .48828 L
s
% Start of user PostScript
-0.5 0.42857143 -0.49508508 0.43799169 MAarrowhead1
% End of user PostScript
.43263 .54475 m
.43396 .54652 L
s
% Start of user PostScript
-0.5 0.85714286 -0.49017016 0.8702493 MAarrowhead1
% End of user PostScript
.43263 .60249 m
.43462 .60475 L
s
% Start of user PostScript
-0.5 1.2857143 -0.48525525 1.3025069 MAarrowhead1
% End of user PostScript
.43263 .66023 m
.43528 .66299 L
s
% Start of user PostScript
-0.5 1.7142857 -0.48034033 1.7347645 MAarrowhead1
% End of user PostScript
.43263 .71797 m
.43595 .72123 L
s
% Start of user PostScript
-0.5 2.1428571 -0.47542541 2.1670222 MAarrowhead1
% End of user PostScript
.43263 .77572 m
.43661 .77947 L
s
% Start of user PostScript
-0.5 2.5714286 -0.47051049 2.5992798 MAarrowhead1
% End of user PostScript
.43263 .83346 m
.43727 .83771 L
s
% Start of user PostScript
-0.5 3. -0.46559558 3.0315374 MAarrowhead1
% End of user PostScript
.5 .02508 m
.49536 .02044 L
s
% Start of user PostScript
0. -3. -0.034404423 -3.0344044 MAarrowhead1
% End of user PostScript
.5 .08282 m
.49603 .07884 L
s
% Start of user PostScript
0. -2.5714286 -0.029489506 -2.6009181 MAarrowhead1
% End of user PostScript
.5 .14056 m
.49669 .13725 L
s
% Start of user PostScript
0. -2.1428571 -0.024574588 -2.1674317 MAarrowhead1
% End of user PostScript
.5 .1983 m
.49735 .19565 L
s
% Start of user PostScript
0. -1.7142857 -0.01965967 -1.7339454 MAarrowhead1
% End of user PostScript
.5 .25604 m
.49801 .25406 L
s
% Start of user PostScript
0. -1.2857143 -0.014744753 -1.300459 MAarrowhead1
% End of user PostScript
.5 .31378 m
.49868 .31246 L
s
% Start of user PostScript
0. -0.85714286 -0.0098298352 -0.86697269 MAarrowhead1
% End of user PostScript
.5 .37153 m
.49934 .37086 L
s
% Start of user PostScript
0. -0.42857143 -0.0049149176 -0.43348635 MAarrowhead1
% End of user PostScript
.5 .42927 m
.5 .42927 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.5 .48701 m
.50066 .48767 L
s
% Start of user PostScript
0. 0.42857143 0.0049149176 0.43348635 MAarrowhead1
% End of user PostScript
.5 .54475 m
.50132 .54607 L
s
% Start of user PostScript
0. 0.85714286 0.0098298352 0.86697269 MAarrowhead1
% End of user PostScript
.5 .60249 m
.50199 .60448 L
s
% Start of user PostScript
0. 1.2857143 0.014744753 1.300459 MAarrowhead1
% End of user PostScript
.5 .66023 m
.50265 .66288 L
s
% Start of user PostScript
0. 1.7142857 0.01965967 1.7339454 MAarrowhead1
% End of user PostScript
.5 .71797 m
.50331 .72128 L
s
% Start of user PostScript
0. 2.1428571 0.024574588 2.1674317 MAarrowhead1
% End of user PostScript
.5 .77572 m
.50397 .77969 L
s
% Start of user PostScript
0. 2.5714286 0.029489506 2.6009181 MAarrowhead1
% End of user PostScript
.5 .83346 m
.50464 .83809 L
s
% Start of user PostScript
0. 3. 0.034404423 3.0344044 MAarrowhead1
% End of user PostScript
.56737 .02508 m
.56273 .02083 L
s
% Start of user PostScript
0.5 -3. 0.46559558 -3.0315374 MAarrowhead1
% End of user PostScript
.56737 .08282 m
.56339 .07907 L
s
% Start of user PostScript
0.5 -2.5714286 0.47051049 -2.5992798 MAarrowhead1
% End of user PostScript
.56737 .14056 m
.56405 .1373 L
s
% Start of user PostScript
0.5 -2.1428571 0.47542541 -2.1670222 MAarrowhead1
% End of user PostScript
.56737 .1983 m
.56472 .19554 L
s
% Start of user PostScript
0.5 -1.7142857 0.48034033 -1.7347645 MAarrowhead1
% End of user PostScript
.56737 .25604 m
.56538 .25378 L
s
% Start of user PostScript
0.5 -1.2857143 0.48525525 -1.3025069 MAarrowhead1
% End of user PostScript
.56737 .31378 m
.56604 .31202 L
s
% Start of user PostScript
0.5 -0.85714286 0.49017016 -0.8702493 MAarrowhead1
% End of user PostScript
.56737 .37153 m
.5667 .37026 L
s
% Start of user PostScript
0.5 -0.42857143 0.49508508 -0.43799169 MAarrowhead1
% End of user PostScript
.56737 .42927 m
.56737 .42849 L
s
% Start of user PostScript
0.5 0. 0.5 -0.0057340705 MAarrowhead1
% End of user PostScript
.56737 .48701 m
.56803 .48673 L
s
% Start of user PostScript
0.5 0.42857143 0.50491492 0.42652355 MAarrowhead1
% End of user PostScript
.56737 .54475 m
.56869 .54497 L
s
% Start of user PostScript
0.5 0.85714286 0.50982984 0.85878116 MAarrowhead1
% End of user PostScript
.56737 .60249 m
.56935 .60321 L
s
% Start of user PostScript
0.5 1.2857143 0.51474475 1.2910388 MAarrowhead1
% End of user PostScript
.56737 .66023 m
.57001 .66145 L
s
% Start of user PostScript
0.5 1.7142857 0.51965967 1.7232964 MAarrowhead1
% End of user PostScript
.56737 .71797 m
.57068 .71968 L
s
% Start of user PostScript
0.5 2.1428571 0.52457459 2.155554 MAarrowhead1
% End of user PostScript
.56737 .77572 m
.57134 .77792 L
s
% Start of user PostScript
0.5 2.5714286 0.52948951 2.5878116 MAarrowhead1
% End of user PostScript
.56737 .83346 m
.572 .83616 L
s
% Start of user PostScript
0.5 3. 0.53440442 3.0200692 MAarrowhead1
% End of user PostScript
.63473 .02508 m
.63009 .02353 L
s
% Start of user PostScript
1. -3. 0.96559558 -3.0114681 MAarrowhead1
% End of user PostScript
.63473 .08282 m
.63076 .08127 L
s
% Start of user PostScript
1. -2.5714286 0.97051049 -2.5828967 MAarrowhead1
% End of user PostScript
.63473 .14056 m
.63142 .13901 L
s
% Start of user PostScript
1. -2.1428571 0.97542541 -2.1543253 MAarrowhead1
% End of user PostScript
.63473 .1983 m
.63208 .19676 L
s
% Start of user PostScript
1. -1.7142857 0.98034033 -1.7257539 MAarrowhead1
% End of user PostScript
.63473 .25604 m
.63274 .2545 L
s
% Start of user PostScript
1. -1.2857143 0.98525525 -1.2971824 MAarrowhead1
% End of user PostScript
.63473 .31378 m
.63341 .31224 L
s
% Start of user PostScript
1. -0.85714286 0.99017016 -0.868611 MAarrowhead1
% End of user PostScript
.63473 .37153 m
.63407 .36998 L
s
% Start of user PostScript
1. -0.42857143 0.99508508 -0.44003957 MAarrowhead1
% End of user PostScript
.63473 .42927 m
.63473 .42772 L
s
% Start of user PostScript
1. 0. 1. -0.011468141 MAarrowhead1
% End of user PostScript
.63473 .48701 m
.63539 .48546 L
s
% Start of user PostScript
1. 0.42857143 1.0049149 0.41710329 MAarrowhead1
% End of user PostScript
.63473 .54475 m
.63605 .5432 L
s
% Start of user PostScript
1. 0.85714286 1.0098298 0.84567472 MAarrowhead1
% End of user PostScript
.63473 .60249 m
.63672 .60095 L
s
% Start of user PostScript
1. 1.2857143 1.0147448 1.2742461 MAarrowhead1
% End of user PostScript
.63473 .66023 m
.63738 .65869 L
s
% Start of user PostScript
1. 1.7142857 1.0196597 1.7028176 MAarrowhead1
% End of user PostScript
.63473 .71797 m
.63804 .71643 L
s
% Start of user PostScript
1. 2.1428571 1.0245746 2.131389 MAarrowhead1
% End of user PostScript
.63473 .77572 m
.6387 .77417 L
s
% Start of user PostScript
1. 2.5714286 1.0294895 2.5599604 MAarrowhead1
% End of user PostScript
.63473 .83346 m
.63937 .83191 L
s
% Start of user PostScript
1. 3. 1.0344044 2.9885319 MAarrowhead1
% End of user PostScript
.7021 .02508 m
.69746 .02855 L
s
% Start of user PostScript
1.5 -3. 1.4655956 -2.9741967 MAarrowhead1
% End of user PostScript
.7021 .08282 m
.69812 .08547 L
s
% Start of user PostScript
1.5 -2.5714286 1.4705105 -2.5517689 MAarrowhead1
% End of user PostScript
.7021 .14056 m
.69878 .14238 L
s
% Start of user PostScript
1.5 -2.1428571 1.4754254 -2.1293411 MAarrowhead1
% End of user PostScript
.7021 .1983 m
.69945 .19929 L
s
% Start of user PostScript
1.5 -1.7142857 1.4803403 -1.7069133 MAarrowhead1
% End of user PostScript
.7021 .25604 m
.70011 .25621 L
s
% Start of user PostScript
1.5 -1.2857143 1.4852552 -1.2844856 MAarrowhead1
% End of user PostScript
.7021 .31378 m
.70077 .31312 L
s
% Start of user PostScript
1.5 -0.85714286 1.4901702 -0.86205777 MAarrowhead1
% End of user PostScript
.7021 .37153 m
.70143 .37004 L
s
% Start of user PostScript
1.5 -0.42857143 1.4950851 -0.43962999 MAarrowhead1
% End of user PostScript
.7021 .42927 m
.7021 .42695 L
s
% Start of user PostScript
1.5 0. 1.5 -0.017202212 MAarrowhead1
% End of user PostScript
.7021 .48701 m
.70276 .48386 L
s
% Start of user PostScript
1.5 0.42857143 1.5049149 0.40522557 MAarrowhead1
% End of user PostScript
.7021 .54475 m
.70342 .54078 L
s
% Start of user PostScript
1.5 0.85714286 1.5098298 0.82765335 MAarrowhead1
% End of user PostScript
.7021 .60249 m
.70408 .59769 L
s
% Start of user PostScript
1.5 1.2857143 1.5147448 1.2500811 MAarrowhead1
% End of user PostScript
.7021 .66023 m
.70474 .6546 L
s
% Start of user PostScript
1.5 1.7142857 1.5196597 1.6725089 MAarrowhead1
% End of user PostScript
.7021 .71797 m
.70541 .71152 L
s
% Start of user PostScript
1.5 2.1428571 1.5245746 2.0949367 MAarrowhead1
% End of user PostScript
.7021 .77572 m
.70607 .76843 L
s
% Start of user PostScript
1.5 2.5714286 1.5294895 2.5173645 MAarrowhead1
% End of user PostScript
.7021 .83346 m
.70673 .82535 L
s
% Start of user PostScript
1.5 3. 1.5344044 2.9397923 MAarrowhead1
% End of user PostScript
.76946 .02508 m
.76482 .03589 L
s
% Start of user PostScript
2. -3. 1.9655956 -2.919723 MAarrowhead1
% End of user PostScript
.76946 .08282 m
.76549 .09165 L
s
% Start of user PostScript
2. -2.5714286 1.9705105 -2.5058963 MAarrowhead1
% End of user PostScript
.76946 .14056 m
.76615 .1474 L
s
% Start of user PostScript
2. -2.1428571 1.9754254 -2.0920697 MAarrowhead1
% End of user PostScript
.76946 .1983 m
.76681 .20316 L
s
% Start of user PostScript
2. -1.7142857 1.9803403 -1.678243 MAarrowhead1
% End of user PostScript
.76946 .25604 m
.76747 .25891 L
s
% Start of user PostScript
2. -1.2857143 1.9852552 -1.2644163 MAarrowhead1
% End of user PostScript
.76946 .31378 m
.76814 .31467 L
s
% Start of user PostScript
2. -0.85714286 1.9901702 -0.85058963 MAarrowhead1
% End of user PostScript
.76946 .37153 m
.7688 .37042 L
s
% Start of user PostScript
2. -0.42857143 1.9950851 -0.43676296 MAarrowhead1
% End of user PostScript
.76946 .42927 m
.76946 .42618 L
s
% Start of user PostScript
2. 0. 2. -0.022936282 MAarrowhead1
% End of user PostScript
.76946 .48701 m
.77012 .48193 L
s
% Start of user PostScript
2. 0.42857143 2.0049149 0.39089039 MAarrowhead1
% End of user PostScript
.76946 .54475 m
.77078 .53769 L
s
% Start of user PostScript
2. 0.85714286 2.0098298 0.80471707 MAarrowhead1
% End of user PostScript
.76946 .60249 m
.77145 .59344 L
s
% Start of user PostScript
2. 1.2857143 2.0147448 1.2185437 MAarrowhead1
% End of user PostScript
.76946 .66023 m
.77211 .6492 L
s
% Start of user PostScript
2. 1.7142857 2.0196597 1.6323704 MAarrowhead1
% End of user PostScript
.76946 .71797 m
.77277 .70495 L
s
% Start of user PostScript
2. 2.1428571 2.0245746 2.0461971 MAarrowhead1
% End of user PostScript
.76946 .77572 m
.77343 .76071 L
s
% Start of user PostScript
2. 2.5714286 2.0294895 2.4600238 MAarrowhead1
% End of user PostScript
.76946 .83346 m
.7741 .81646 L
s
% Start of user PostScript
2. 3. 2.0344044 2.8738504 MAarrowhead1
% End of user PostScript
.83683 .02508 m
.83219 .04555 L
s
% Start of user PostScript
2.5 -3. 2.4655956 -2.8480471 MAarrowhead1
% End of user PostScript
.83683 .08282 m
.83285 .09981 L
s
% Start of user PostScript
2.5 -2.5714286 2.4705105 -2.445279 MAarrowhead1
% End of user PostScript
.83683 .14056 m
.83351 .15408 L
s
% Start of user PostScript
2.5 -2.1428571 2.4754254 -2.0425109 MAarrowhead1
% End of user PostScript
.83683 .1983 m
.83418 .20834 L
s
% Start of user PostScript
2.5 -1.7142857 2.4803403 -1.6397428 MAarrowhead1
% End of user PostScript
.83683 .25604 m
.83484 .26261 L
s
% Start of user PostScript
2.5 -1.2857143 2.4852552 -1.2369747 MAarrowhead1
% End of user PostScript
.83683 .31378 m
.8355 .31687 L
s
% Start of user PostScript
2.5 -0.85714286 2.4901702 -0.83420658 MAarrowhead1
% End of user PostScript
.83683 .37153 m
.83616 .37114 L
s
% Start of user PostScript
2.5 -0.42857143 2.4950851 -0.43143846 MAarrowhead1
% End of user PostScript
.83683 .42927 m
.83683 .4254 L
s
% Start of user PostScript
2.5 0. 2.5 -0.028670353 MAarrowhead1
% End of user PostScript
.83683 .48701 m
.83749 .47967 L
s
% Start of user PostScript
2.5 0.42857143 2.5049149 0.37409776 MAarrowhead1
% End of user PostScript
.83683 .54475 m
.83815 .53393 L
s
% Start of user PostScript
2.5 0.85714286 2.5098298 0.77686587 MAarrowhead1
% End of user PostScript
.83683 .60249 m
.83881 .5882 L
s
% Start of user PostScript
2.5 1.2857143 2.5147448 1.179634 MAarrowhead1
% End of user PostScript
.83683 .66023 m
.83947 .64246 L
s
% Start of user PostScript
2.5 1.7142857 2.5196597 1.5824021 MAarrowhead1
% End of user PostScript
.83683 .71797 m
.84014 .69673 L
s
% Start of user PostScript
2.5 2.1428571 2.5245746 1.9851702 MAarrowhead1
% End of user PostScript
.83683 .77572 m
.8408 .75099 L
s
% Start of user PostScript
2.5 2.5714286 2.5294895 2.3879383 MAarrowhead1
% End of user PostScript
.83683 .83346 m
.84146 .80526 L
s
% Start of user PostScript
2.5 3. 2.5344044 2.7907064 MAarrowhead1
% End of user PostScript
.90419 .02508 m
.89955 .05752 L
s
% Start of user PostScript
3. -3. 2.9655956 -2.759169 MAarrowhead1
% End of user PostScript
.90419 .08282 m
.90022 .10997 L
s
% Start of user PostScript
3. -2.5714286 2.9705105 -2.3699169 MAarrowhead1
% End of user PostScript
.90419 .14056 m
.90088 .16241 L
s
% Start of user PostScript
3. -2.1428571 2.9754254 -1.9806649 MAarrowhead1
% End of user PostScript
.90419 .1983 m
.90154 .21486 L
s
% Start of user PostScript
3. -1.7142857 2.9803403 -1.5914128 MAarrowhead1
% End of user PostScript
.90419 .25604 m
.9022 .2673 L
s
% Start of user PostScript
3. -1.2857143 2.9852552 -1.2021607 MAarrowhead1
% End of user PostScript
.90419 .31378 m
.90287 .31974 L
s
% Start of user PostScript
3. -0.85714286 2.9901702 -0.8129086 MAarrowhead1
% End of user PostScript
.90419 .37153 m
.90353 .37219 L
s
% Start of user PostScript
3. -0.42857143 2.9950851 -0.42365651 MAarrowhead1
% End of user PostScript
.90419 .42927 m
.90419 .42463 L
s
% Start of user PostScript
3. 0. 3. -0.034404423 MAarrowhead1
% End of user PostScript
.90419 .48701 m
.90485 .47708 L
s
% Start of user PostScript
3. 0.42857143 3.0049149 0.35484766 MAarrowhead1
% End of user PostScript
.90419 .54475 m
.90551 .52952 L
s
% Start of user PostScript
3. 0.85714286 3.0098298 0.74409975 MAarrowhead1
% End of user PostScript
.90419 .60249 m
.90618 .58196 L
s
% Start of user PostScript
3. 1.2857143 3.0147448 1.1333518 MAarrowhead1
% End of user PostScript
.90419 .66023 m
.90684 .63441 L
s
% Start of user PostScript
3. 1.7142857 3.0196597 1.5226039 MAarrowhead1
% End of user PostScript
.90419 .71797 m
.9075 .68685 L
s
% Start of user PostScript
3. 2.1428571 3.0245746 1.911856 MAarrowhead1
% End of user PostScript
.90419 .77572 m
.90816 .7393 L
s
% Start of user PostScript
3. 2.5714286 3.0294895 2.3011081 MAarrowhead1
% End of user PostScript
.90419 .83346 m
.90883 .79174 L
s
% Start of user PostScript
3. 3. 3.0344044 2.6903602 MAarrowhead1
% End of user PostScript
.97156 .02508 m
.96692 .07182 L
s
% Start of user PostScript
3.5 -3. 3.4655956 -2.6530887 MAarrowhead1
% End of user PostScript
.97156 .08282 m
.96758 .12211 L
s
% Start of user PostScript
3.5 -2.5714286 3.4705105 -2.2798101 MAarrowhead1
% End of user PostScript
.97156 .14056 m
.96824 .1724 L
s
% Start of user PostScript
3.5 -2.1428571 3.4754254 -1.9065315 MAarrowhead1
% End of user PostScript
.97156 .1983 m
.96891 .22269 L
s
% Start of user PostScript
3.5 -1.7142857 3.4803403 -1.5332529 MAarrowhead1
% End of user PostScript
.97156 .25604 m
.96957 .27298 L
s
% Start of user PostScript
3.5 -1.2857143 3.4852552 -1.1599743 MAarrowhead1
% End of user PostScript
.97156 .31378 m
.97023 .32328 L
s
% Start of user PostScript
3.5 -0.85714286 3.4901702 -0.7866957 MAarrowhead1
% End of user PostScript
.97156 .37153 m
.97089 .37357 L
s
% Start of user PostScript
3.5 -0.42857143 3.4950851 -0.4134171 MAarrowhead1
% End of user PostScript
.97156 .42927 m
.97156 .42386 L
s
% Start of user PostScript
3.5 0. 3.5 -0.040138494 MAarrowhead1
% End of user PostScript
.97156 .48701 m
.97222 .47415 L
s
% Start of user PostScript
3.5 0.42857143 3.5049149 0.33314011 MAarrowhead1
% End of user PostScript
.97156 .54475 m
.97288 .52444 L
s
% Start of user PostScript
3.5 0.85714286 3.5098298 0.70641872 MAarrowhead1
% End of user PostScript
.97156 .60249 m
.97354 .57473 L
s
% Start of user PostScript
3.5 1.2857143 3.5147448 1.0796973 MAarrowhead1
% End of user PostScript
.97156 .66023 m
.9742 .62503 L
s
% Start of user PostScript
3.5 1.7142857 3.5196597 1.4529759 MAarrowhead1
% End of user PostScript
.97156 .71797 m
.97487 .67532 L
s
% Start of user PostScript
3.5 2.1428571 3.5245746 1.8262545 MAarrowhead1
% End of user PostScript
.97156 .77572 m
.97553 .72561 L
s
% Start of user PostScript
3.5 2.5714286 3.5294895 2.1995331 MAarrowhead1
% End of user PostScript
.97156 .83346 m
.97619 .7759 L
s
% Start of user PostScript
3.5 3. 3.5344044 2.5728117 MAarrowhead1
% End of user PostScript
.5 Mabswid
.90419 .564 m
.90907 .50521 L
.91072 .48101 L
.91175 .46235 L
.91234 .44804 L
.91264 .43529 L
.91266 .42579 L
.91251 .41847 L
.91216 .4113 L
.91168 .40581 L
.91102 .40088 L
.91018 .3966 L
.9092 .39304 L
.90819 .39034 L
.90722 .38835 L
.90611 .38657 L
.90488 .38502 L
.90369 .38382 L
.90138 .38208 L
.89882 .38071 L
.89597 .37961 L
.89071 .37815 L
.88575 .3771 L
.83556 .3674 L
.80764 .36061 L
.77851 .35196 L
.74663 .34013 L
.70615 .32058 L
.65022 .28318 L
.57552 .21411 L
.52687 .16246 L
.49545 .13012 L
.46268 .10046 L
.44827 .08955 L
.43262 .07968 L
.42426 .07538 L
.41501 .07152 L
.40652 .06886 L
.39858 .06722 L
.38899 .06641 L
.37993 .0669 L
.37182 .06846 L
.36309 .0714 L
.35367 .07614 L
.34385 .08294 L
.32684 .09965 L
.30992 .12314 L
.29515 .15008 L
.27144 .20833 L
.25227 .2734 L
Mistroke
.24421 .30815 L
.23859 .33689 L
.23484 .3597 L
.23191 .38163 L
.23075 .39252 L
.22993 .40181 L
.22937 .41009 L
.22898 .4182 L
.22881 .4247 L
.22878 .4312 L
.22889 .43691 L
.22909 .44179 L
.22946 .44714 L
.22999 .45252 L
.23062 .45723 L
.23129 .46128 L
.23314 .46949 L
.23517 .47601 L
.23773 .48228 L
.24074 .48805 L
.24647 .49633 L
.25351 .50394 L
.27194 .51849 L
.2919 .53164 L
.40707 .62471 L
.45428 .67407 L
.48381 .70548 L
.51905 .74038 L
.55324 .76815 L
.56309 .77449 L
.57384 .78034 L
.58414 .7848 L
.59364 .78779 L
.60295 .7896 L
.6116 .79019 L
.6211 .78954 L
.62588 .78866 L
.63115 .78725 L
.63988 .78382 L
.64919 .77859 L
.65783 .77218 L
.6656 .76505 L
.68256 .74461 L
.69695 .72147 L
.71112 .69273 L
.72474 .65864 L
.74422 .59596 L
.75711 .54089 L
.76258 .51128 L
.76626 .48673 L
Mistroke
.76881 .46442 L
.76969 .45389 L
.77005 .4484 L
.7703 .44353 L
.77057 .43548 L
.77063 .43148 L
.77063 .4274 L
.77058 .42369 L
.77048 .42033 L
.77018 .41442 L
.76974 .40901 L
.76922 .40439 L
.76854 .39967 L
.76769 .39505 L
.76563 .38678 L
.7633 .38002 L
.76093 .37467 L
.7581 .36951 L
.75091 .35974 L
.74339 .35219 L
.70894 .32759 L
.59387 .23482 L
.55029 .18941 L
.49321 .12983 L
.46587 .10499 L
.44995 .09267 L
.43443 .08264 L
.42542 .07788 L
.41552 .07367 L
.40645 .07082 L
.39803 .06912 L
.39006 .06843 L
.3813 .06875 L
.37217 .0704 L
.36369 .0732 L
.35478 .07756 L
.34523 .08397 L
.32837 .10001 L
.31286 .12073 L
.29948 .14378 L
.27341 .20512 L
.26168 .24157 L
.25284 .27395 L
.23944 .33577 L
.23465 .36531 L
.23297 .3779 L
.23173 .3888 L
.23081 .39847 L
.23019 .40668 L
.22971 .41517 L
Mistroke
.22954 .41957 L
.22944 .42356 L
.22938 .43106 L
.22951 .43747 L
.22983 .44395 L
.23005 .44702 L
.23035 .45026 L
.23099 .45562 L
.23171 .46022 L
.23271 .46526 L
.23374 .46947 L
.23594 .47646 L
.23864 .48295 L
.24455 .49304 L
.25095 .50083 L
.26757 .51511 L
.28874 .52943 L
.40864 .62622 L
.45406 .67379 L
.48368 .70528 L
.51516 .73671 L
.54739 .76394 L
.56672 .77654 L
.57618 .7814 L
.58517 .78512 L
.59273 .7875 L
.60093 .78925 L
.61007 .79011 L
.61845 .7898 L
.62669 .78842 L
.63564 .7856 L
.6449 .78115 L
.6534 .7756 L
.67029 .76004 L
.67943 .74886 L
.68756 .73715 L
.70138 .71308 L
.71521 .68315 L
.73534 .62684 L
.75118 .56804 L
.76244 .51198 L
.76642 .4854 L
.7678 .47408 L
.76881 .46429 L
.76958 .45506 L
.77009 .44727 L
.77045 .43921 L
.77061 .43129 L
.77058 .42488 L
.77039 .41846 L
Mistroke
.77005 .41285 L
.76963 .40805 L
.76899 .40279 L
.76827 .39819 L
.76662 .39043 L
.76446 .38318 L
.76172 .37636 L
.75594 .3662 L
.74979 .3585 L
.74233 .35127 L
.73332 .34408 L
.71271 .33006 L
.59725 .23818 L
.50576 .14235 L
.47356 .11157 L
.45452 .09601 L
.43578 .08343 L
.42707 .07869 L
.41773 .07451 L
.41282 .07271 L
.40738 .07107 L
.39778 .06909 L
.38903 .06841 L
.37964 .06895 L
.3708 .07077 L
.36278 .07358 L
.35357 .07827 L
.34394 .08497 L
.32718 .10139 L
.31022 .12489 L
.29311 .15662 L
.26768 .22211 L
.25023 .28456 L
.24326 .31594 L
.23766 .34592 L
.23413 .36898 L
.23154 .39068 L
.23056 .40162 L
.22993 .4108 L
.22955 .41933 L
.22939 .42677 L
.22938 .43073 L
.22944 .43474 L
.22968 .44149 L
.2299 .44499 L
.23015 .44813 L
.23088 .45478 L
.23161 .45964 L
.23252 .46441 L
.23448 .47207 L
Mistroke
.23728 .4799 L
.24019 .48601 L
.2468 .49604 L
.25391 .50383 L
.26903 .51617 L
.28787 .52886 L
.29777 .5353 L
Mfstroke
0 0 m
1 0 L
1 .85853 L
0 .85853 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 247.25},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`0001LR000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00P03oool0
0`000000oooo0?ooo`2_0?ooo`030000003oool0oooo07/0oooo00200?ooo`<000007`3oool00`00
0000oooo0?ooo`2<0?ooo`<000005P3oool00`000000oooo0?ooo`1S0?ooo`00P03oool5000001/0
oooo0`00002=0?ooo`@000005P3oool3000006<0oooo001o0?ooo`D000006`3oool500000340oooo
00<000000?ooo`3oool0E`3oool6000001D0oooo1@00001Q0?ooo`00H@3oool2000001`0oooo1000
000K0?ooo`H00000503oool2000001`0oooo0P00000J0?ooo`030000003oool0oooo03X0oooo1@00
000F0?ooo`@00000HP3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01P0oooo1@00000I0?ooo`<000006P3oool9000001@0oooo1000000I
0?ooo`<000006P3oool2000001X0oooo2P00000E0?ooo`@00000603oool4000001P0oooo0`00000L
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo1000000J0?oo
o`<000006@3oool5000001X0oooo1P00000G0?ooo`@00000603oool4000001T0oooo2@00000D0?oo
o`<000006@3oool4000001T0oooo1@00000J0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool02`3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01P0oooo1000000I0?ooo`<00000>P3oool7000001D0oooo1P00000G0?ooo`D0
00006@3oool6000001@0oooo0P00000K0?ooo`<000006P3oool4000001/0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0J0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool0603oool3000001X0oooo0P00000l0?ooo`D00000
5`3oool6000001L0oooo1P00000I0?ooo`@00000<@3oool2000001/0oooo1000000K0?ooo`030000
003oool0oooo01X0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0G0?ooo`D000006@3oool3000001X0oooo00<000000?ooo`3oool0?03o
ool3000001T0oooo1000000I0?ooo`@00000703oool200000340oooo0P00000K0?ooo`<000006`3o
ool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool0603oool5000001T0oooo0P00001K0?ooo`030000003oool0
oooo01P0oooo0`00000J0?ooo`<00000D03oool00`000000oooo0?ooo`0J0?ooo`<000006`3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool0603oool4000001X0oooo0P00001g0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0K03oool2000001X0oooo1@00000J0?ooo`030000003oool0oooo
00`0oooo000;0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0603oool4000001X0
oooo00<000000?ooo`3oool0o`3oool30?ooo`8000006P3oool5000001X0oooo00<000000?ooo`3o
ool0303oool000/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`@00000
6P3oool00`000000oooo0?ooo`3o0?ooo`<0oooo00<000000?ooo`3oool06@3oool4000001/0oooo
00<000000?ooo`3oool0303oool000/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`0I0?ooo`800000o`3ooolQ0?ooo`030000003oool0oooo01T0oooo1000000K0?ooo`030000003o
ool0oooo00`0oooo000;0?ooo`030000003oool0oooo01P0oooo1@00000I0?ooo`800000o`3ooolm
0?ooo`<00000703oool00`000000oooo0?ooo`0<0?ooo`002`3oool00`000000oooo0?ooo`0H0?oo
o`D000006@3oool200000?l0oooo?P3oool2000001`0oooo00<000000?ooo`3oool0303oool000/0
oooo00<000000?ooo`3oool0603oool4000001X0oooo00<000000?ooo`3oool0o`3ooolm0?ooo`80
00006P3oool5000000`0oooo000;0?ooo`030000003oool0oooo01P0oooo1000000J0?ooo`030000
003oool0oooo0?l0oooo?@3oool00`000000oooo0?ooo`0I0?ooo`D00000303oool000/0oooo00<0
00000?ooo`3oool0603oool300000?l0ooooF`3oool00`000000oooo0?ooo`0I0?ooo`@000003@3o
ool000/0oooo00<000000?ooo`3oool06@3oool200000?l0ooooM`3oool4000000d0oooo00080?oo
o`@000006`3oool200000?l0ooooM`3oool4000000d0oooo00080?ooo`H000006@3oool00`000000
oooo0?ooo`3o0?ooogL0oooo0P00000>0?ooo`00203oool6000001T0oooo00<000000?ooo`3oool0
N03oool700000?L0oooo0P00000>0?ooo`002@3oool400000980oooo4P00003_0?ooo`8000003P3o
ool000T0oooo1000001c0?ooo`030000003oool0oooo01T0oooo1`00000:0?ooo`H00000O`3oool0
0`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0580oooo00<000000?ooo`3oool03@3oool0
00T0oooo0`00001d0?ooo`<000005`3oool5000000<0oooo00<000000?ooo`3oool02`3oool50000
07`0oooo0`00000G0?ooo`<00000DP3oool00`000000oooo0?ooo`0=0?ooo`002@3oool3000007<0
oooo1P00000C0?ooo`@00000103oool300000100oooo1000001i0?ooo`D000005P3oool500000600
oooo000:0?ooo`800000E@3oool2000001`0oooo1@00000C0?ooo`<000001@3oool500000140oooo
1000001f0?ooo`L00000503oool500000640oooo000:0?ooo`030000003oool0oooo05@0oooo1@00
000I0?ooo`@000004P3oool3000000H0oooo1P00000C0?ooo`@000006P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo03T0oooo1P00000F0?ooo`@00000603oool2000004P0oooo000:
0?ooo`030000003oool0000001`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H
0?ooo`@000006P3oool300000180oooo0`0000050?ooo`T00000503oool4000001T0oooo0P00000K
0?ooo`8000006P3oool5000001T0oooo1000000I0?ooo`<000006@3oool5000001X0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool0603oool4000001T0oooo0`00000B0?oo
o`8000001P3oool5000001X0oooo1P00000F0?ooo`@000006@3oool4000001T0oooo2@00000C0?oo
o`800000703oool3000001T0oooo1000000K0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool02`3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0G0?ooo`@0
00006P3oool3000001X0oooo0P00000B0?ooo`8000009`3oool7000001D0oooo1@00000H0?ooo`D0
00006@3oool700000300oooo0`00000J0?ooo`@000006`3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
5`3oool6000001P0oooo0`00000J0?ooo`030000003oool0oooo0100oooo0P00000Z0?ooo`D00000
5P3oool6000001L0oooo1P00000H0?ooo`D00000<@3oool2000001/0oooo0`00000J0?ooo`800000
7@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo01P0oooo1@00000I0?ooo`800000;@3oool2000002`0oooo0`0000000`3oool000000?oo
o`0E0?ooo`@000006@3oool4000001/0oooo0`00000a0?ooo`030000003oool0oooo01X0oooo0`00
000J0?ooo`D000006P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0I
0?ooo`030000003oool0oooo01P0oooo1@00000I0?ooo`800000;03oool2000002h0oooo00<00000
0?ooo`0000000P00000E0?ooo`<000006P3oool3000001h0oooo00<000000?ooo`3oool0C03oool2
000001/0oooo1000000J0?ooo`030000003oool0oooo00`0oooo000;0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0603oool4000001X0oooo00<000000?ooo`3oool0:P3oool20000
0380oooo1000000D0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0K03oool20000
01/0oooo1000000J0?ooo`030000003oool0oooo00`0oooo000;0?ooo`030000003oool0oooo01P0
oooo0`00000K0?ooo`<000006P3oool00`000000oooo0?ooo`0Y0?ooo`800000=@3oool00`000000
oooo0000002O0?ooo`030000003oool0oooo01X0oooo0`00000K0?ooo`030000003oool0oooo00`0
oooo000;0?ooo`030000003oool0oooo01P0oooo1@00000I0?ooo`800000AP3oool2000003L0oooo
0`00002N0?ooo`030000003oool0oooo01X0oooo0`00000I0?ooo`<000003P3oool000/0oooo00<0
00000?ooo`3oool0603oool5000001T0oooo0P0000150?ooo`800000>@3oool300000;X0oooo0P00
000J0?ooo`D00000303oool000/0oooo00<000000?ooo`3oool0603oool4000001X0oooo00<00000
0?ooo`3oool0A03oool00`000000oooo0?ooo`0j0?ooo`800000^@3oool2000001X0oooo1@00000<
0?ooo`002`3oool00`000000oooo0?ooo`0H0?ooo`@000006P3oool00`000000oooo0?ooo`130?oo
o`800000?@3oool200000;P0oooo00<000000?ooo`3oool06@3oool4000000d0oooo000;0?ooo`03
0000003oool0oooo01T0oooo0P00001P0?ooo`800000?`3oool200000;L0oooo00<000000?ooo`3o
ool06@3oool4000000d0oooo00090?ooo`<000006`3oool2000005l0oooo0P0000110?ooo`800000
d`3oool2000000h0oooo00090?ooo`D000006@3oool2000005l0oooo00<000000?ooo`3oool0@@3o
ool200000=80oooo0P00000>0?ooo`002@3oool5000001T0oooo00<000000?ooo`3oool0G@3oool2
000004@0oooo0`00003@0?ooo`8000003P3oool000T0oooo1000000J0?ooo`030000003oool0oooo
05`0oooo0P0000160?ooo`<00000c`3oool00`000000oooo0?ooo`0=0?ooo`002@3oool4000007T0
oooo00<000000?ooo`3oool0A`3oool200000<h0oooo00<000000?ooo`3oool03@3oool000X0oooo
0P00001i0?ooo`030000003oool0oooo04T0oooo0P00003M0?ooo`002P3oool2000007@0oooo00@0
00000?ooo`3oool0oooo0P00001<0?ooo`800000M`3oool00`000000oooo0?ooo`1R0?ooo`002P3o
ool2000007@0oooo0`0000000`3oool000000?ooo`0L0?ooo`030000003oool0oooo02h0oooo0P00
001N0?ooo`030000003oool0oooo01D0oooo0`00001R0?ooo`002P3oool00`000000oooo0?ooo`1c
0?ooo`D00000703oool200000340oooo0P00001L0?ooo`8000005`3oool500000600oooo000:0?oo
o`030000003oool0oooo05@0oooo0`00000L0?ooo`@00000703oool400000340oooo0P00001J0?oo
o`@000005P3oool4000001P0oooo0P0000170?ooo`00H@3oool6000001T0oooo1000000K0?ooo`D0
0000<P3oool2000005P0oooo1@00000F0?ooo`@00000603oool5000004@0oooo001Q0?ooo`D00000
6@3oool4000001/0oooo1`00000D0?ooo`8000006`3oool3000001X0oooo00<000000?ooo`3oool0
>@3oool7000001@0oooo1000000I0?ooo`@00000A@3oool000`0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo1@00000I0?ooo`@000006P3o
ool6000001L0oooo1@00000H0?ooo`<000006P3oool2000001X0oooo1000000J0?ooo`H000005P3o
ool3000001X0oooo1000000J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02`3o
ool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`D00000603oool4
000001X0oooo0`00000J0?ooo`<00000703oool7000001D0oooo1000000I0?ooo`@000006@3oool8
000001@0oooo0`00000J0?ooo`8000006`3oool3000001T0oooo1@00000J0?ooo`030000003oool0
oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0603oool40000
01T0oooo0`00000K0?ooo`<00000>P3oool5000001L0oooo1@00000H0?ooo`D000006@3oool60000
0340oooo0P00000K0?ooo`<000006@3oool5000001X0oooo00<000000?ooo`3oool02`3oool000`0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`@000006@3oool3000001/0
oooo0`00000k0?ooo`@000005`3oool6000001L0oooo1P00000I0?ooo`@00000<P3oool00`000000
oooo0?ooo`0J0?ooo`8000006P3oool4000001/0oooo00<000000?ooo`3oool02`3oool000`0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0I0?ooo`<000006P3oool2000001`0oooo
0P00000m0?ooo`8000006@3oool5000001P0oooo1000000K0?ooo`<00000C`3oool2000001X0oooo
1000000K0?ooo`030000003oool0oooo00/0oooo000;0?ooo`030000003oool0oooo01P0oooo1@00
000I0?ooo`<000006P3oool2000001`0oooo0P00000n0?ooo`030000003oool0oooo01L0oooo0`00
00000`3oool000000000000G0?ooo`<000007P3oool00`000000oooo0?ooo`1=0?ooo`030000003o
ool0oooo01T0oooo1000000H0?ooo`@000003@3oool000/0oooo00<000000?ooo`3oool0603oool5
000001T0oooo0P00000K0?ooo`030000003oool0oooo01X0oooo0P00001J0?ooo`040000003oool0
oooo0?ooo`8000005`3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo01X0oooo0P00
000I0?ooo`H000002`3oool000/0oooo00<000000?ooo`3oool0603oool4000001X0oooo0P00000h
0?ooo`800000G`3oool200000:<0oooo0P00000I0?ooo`H000002`3oool000/0oooo00<000000?oo
o`3oool0603oool4000001X0oooo00<000000?ooo`3oool0=P3oool200000640oooo0P00002R0?oo
o`8000006P3oool4000000`0oooo000;0?ooo`030000003oool0oooo01P0oooo1000000J0?ooo`03
0000003oool0oooo03H0oooo0P00001R0?ooo`800000X@3oool00`000000oooo0?ooo`0I0?ooo`@0
0000303oool000T0oooo1@00000I0?ooo`800000E03oool00`000000oooo0?ooo`1R0?ooo`800000
X03oool00`000000oooo0?ooo`0I0?ooo`<000003@3oool000T0oooo1@00000I0?ooo`800000D`3o
ool2000006D0oooo0P00002k0?ooo`<000003@3oool000T0oooo1000000J0?ooo`800000D`3oool0
0`000000oooo0?ooo`1U0?ooo`800000^`3oool2000000d0oooo00090?ooo`@000006P3oool00`00
0000oooo0?ooo`1A0?ooo`800000J03oool200000;X0oooo00<000000?ooo`3oool0303oool000T0
oooo1000000J0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0J03oool200000;T0
oooo00<000000?ooo`3oool0303oool000X0oooo0P00001_0?ooo`030000003oool0oooo06T0oooo
0P0000370?ooo`002P3oool2000006h0oooo0P00001/0?ooo`800000aP3oool000X0oooo0P00001^
0?ooo`030000003oool0oooo06`0oooo0P0000350?ooo`002P3oool00`000000oooo0?ooo`1/0?oo
o`800000K`3oool200000<@0oooo000:0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04H0oooo0P00001O0?ooo`03
0000003oool0oooo0640oooo001i0?ooo`030000003oool0oooo00@0oooo0`00000N0?ooo`800000
B@3oool2000004H0oooo00<000000?ooo`3oool05@3oool300000640oooo001R0?ooo`800000503o
ool00`000000oooo0?ooo`050?ooo`D000006`3oool4000004T0oooo00<000000?ooo`3oool0@P3o
ool3000001H0oooo1P00000F0?ooo`800000A`3oool00680oooo1@00000A0?ooo`030000003oool0
oooo00D0oooo1000000K0?ooo`D00000B@3oool200000480oooo1@00000E0?ooo`D000005`3oool5
000004@0oooo001R0?ooo`@000004@3oool2000000H0oooo1000000L0?ooo`H00000B@3oool20000
0400oooo1P00000E0?ooo`@00000603oool4000001P0oooo0P00000[0?ooo`00A03oool2000001`0
oooo1000000A0?ooo`030000003oool0oooo00D0oooo0`00000L0?ooo`D00000<`3oool00`000000
oooo0?ooo`0F0?ooo`8000000P3oool00`000000oooo0?ooo`0J0?ooo`800000703oool9000001@0
oooo0`00000I0?ooo`@00000603oool5000002P0oooo000<0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool0603oool5000001T0oooo0`00000A0?ooo`8000001`3oool3000001/0oooo
1000000H0?ooo`D000006@3oool2000001L0oooo0P000000103oool0000000000000000K0?ooo`<0
00006@3oool:000001<0oooo0`00000J0?ooo`<000006@3oool5000001X0oooo00<000000?ooo`3o
ool02`3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0H0?ooo`D00000
6@3oool300000140oooo0P0000070?ooo`8000006`3oool2000001`0oooo2@00000D0?ooo`@00000
5P3oool2000000040?ooo`0000000000000001/0oooo1@00000c0?ooo`8000006`3oool3000001T0
oooo1000000K0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01L0oooo
1000000J0?ooo`@000006P3oool200000140oooo0P0000080?ooo`030000003oool0oooo03X0oooo
1P00000F0?ooo`D000005@3oool6000001/0oooo1P00000a0?ooo`030000003oool0oooo01X0oooo
0P00000J0?ooo`@000006@3oool5000000/0oooo000<0?ooo`030000003oool0oooo01L0oooo1P00
000H0?ooo`@000006P3oool200000140oooo0P0000160?ooo`D000005`3oool6000001@0oooo1`00
000I0?ooo`D00000C`3oool2000001/0oooo0P00000J0?ooo`D000002`3oool000`0oooo00<00000
0?ooo`3oool05`3oool6000001T0oooo0P00000K0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0AP3oool3000001T0oooo1000000F0?ooo`L000006@3oool4000004l0oooo00<00000
0?ooo`3oool06P3oool2000001X0oooo1000000<0?ooo`00303oool00`000000oooo0?ooo`0H0?oo
o`@000006P3oool2000001/0oooo00<000000?ooo`3oool03`3oool2000004X0oooo00<000000?oo
o`3oool05`3oool3000001P0oooo1@00000K0?ooo`800000K@3oool2000001X0oooo1000000<0?oo
o`002@3oool2000000030?ooo`000000oooo01T0oooo1000000J0?ooo`800000;@3oool2000006D0
oooo00<000000?ooo`3oool0603oool3000001d0oooo00<000000?ooo`3oool0J`3oool00`000000
oooo0?ooo`0I0?ooo`@00000303oool000T0oooo1@00000I0?ooo`<000006`3oool00`000000oooo
0?ooo`0/0?ooo`030000003oool0oooo0800oooo0P00002;0?ooo`030000003oool0oooo01X0oooo
0P00000=0?ooo`002@3oool5000001T0oooo0`00000K0?ooo`030000003oool0oooo02/0oooo0P00
00230?ooo`800000Y`3oool2000000d0oooo000:0?ooo`<000006`3oool2000004T0oooo00<00000
0?ooo`3oool0P`3oool200000:H0oooo0P00000=0?ooo`002P3oool3000001/0oooo00<000000?oo
o`3oool0B03oool00`000000oooo0?ooo`240?ooo`800000Y@3oool00`000000oooo0?ooo`0<0?oo
o`002P3oool3000001/0oooo00<000000?ooo`3oool0A`3oool2000008H0oooo0`00002T0?ooo`03
0000003oool0oooo00`0oooo000:0?ooo`<00000I@3oool00`000000oooo0?ooo`260?ooo`<00000
/P3oool000/0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`270?ooo`<00000/@3o
ool000/0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`280?ooo`<00000/03oool0
00/0oooo00<000000?ooo`3oool0H`3oool2000008/0oooo1000002^0?ooo`00L@3oool00`000000
oooo0?ooo`2;0?ooo`80000000<0oooo0000003oool0[03oool00740oooo00<000000?ooo`3oool0
S03oool2000000030?ooo`000000oooo0:/0oooo001`0?ooo`8000003P3oool00`000000oooo0?oo
o`1n0?ooo`80000000<0oooo0000003oool0A`3oool00`000000oooo0?ooo`1P0?ooo`00L03oool0
0`000000oooo0?ooo`0=0?ooo`<000007`3oool00`000000oooo0?ooo`1N0?ooo`<00000A`3oool3
000001P0oooo0P0000160?ooo`00HP3oool2000000`0oooo00<000000?ooo`3oool03@3oool50000
01`0oooo0`00001P0?ooo`<00000A@3oool6000001H0oooo1@0000130?ooo`00HP3oool5000000T0
oooo00<000000?ooo`3oool03@3oool4000001`0oooo1000001Q0?ooo`<00000A03oool5000001L0
oooo1000000H0?ooo`800000:P3oool004@0oooo0P00000L0?ooo`@000002@3oool00`000000oooo
0?ooo`0>0?ooo`@000006`3oool600000640oooo0`00000[0?ooo`8000005@3oool5000001P0oooo
1000000H0?ooo`D000009`3oool004@0oooo1@00000I0?ooo`@000002@3oool00`000000oooo0?oo
o`0=0?ooo`@00000703oool5000006@0oooo0P00000X0?ooo`@000005@3oool4000001T0oooo0`00
000I0?ooo`@00000603oool2000000h0oooo00140?ooo`D000006@3oool3000000X0oooo00<00000
0?ooo`3oool03@3oool3000001`0oooo1000000d0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool04`3oool2000000D0oooo00<000000?ooo`3oool0703oool7000001D0oooo0P00000J
0?ooo`<000006P3oool4000001P0oooo1@00000;0?ooo`00303oool00`000000oooo0?ooo`0H0?oo
o`D00000603oool4000001X0oooo0`0000090?ooo`8000003`3oool2000001`0oooo0`00000I0?oo
o`X00000503oool2000001X0oooo0`00000D0?ooo`8000001@3oool2000001X0oooo2P00000C0?oo
o`8000006`3oool3000001X0oooo0`00000I0?ooo`D000002`3oool000`0oooo00<000000?ooo`3o
ool0603oool5000001P0oooo1000000J0?ooo`8000002P3oool00`000000oooo0?ooo`0>0?ooo`80
00006`3oool2000001`0oooo2@00000D0?ooo`@000006@3oool3000001@0oooo0P0000050?ooo`<0
00007@3oool5000001<0oooo00<000000?ooo`3oool06P3oool2000001/0oooo0`00000I0?ooo`@0
0000303oool000`0oooo00<000000?ooo`3oool0603oool4000001X0oooo0P00000K0?ooo`800000
2P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo03X0oooo1P00000F0?ooo`D00000
5`3oool5000001@0oooo0P0000040?ooo`@00000D03oool2000001/0oooo0P00000J0?ooo`@00000
303oool000T0oooo1P00000H0?ooo`@000006P3oool2000001/0oooo00<000000?ooo`3oool0203o
ool2000004h0oooo1@00000F0?ooo`L000005P3oool6000001<0oooo0P0000030?ooo`H00000CP3o
ool00`000000oooo0?ooo`0J0?ooo`8000006`3oool2000000d0oooo00090?ooo`H00000603oool4
000001X0oooo0P00000K0?ooo`030000003oool0oooo00P0oooo0P00001?0?ooo`<00000603oool5
000001L0oooo1`00000C0?ooo`8000000`3oool4000006`0oooo00<000000?ooo`3oool06P3oool2
000000d0oooo00090?ooo`D000006P3oool2000001/0oooo00<000000?ooo`3oool09@3oool20000
0540oooo00<000000?ooo`3oool05`3oool3000001T0oooo1@00000E0?ooo`8000000`3oool30000
06`0oooo00<000000?ooo`3oool06P3oool2000000d0oooo000:0?ooo`@000006P3oool2000001/0
oooo00<000000?ooo`3oool0903oool2000006`0oooo0`00000I0?ooo`@00000603oool200000080
oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0<0?ooo`002P3oool4000001X0oooo
0P0000120?ooo`800000K@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01T0oooo
0P00002<0?ooo`030000003oool0oooo00`0oooo000:0?ooo`<000006`3oool00`000000oooo0?oo
o`110?ooo`800000YP3oool2000009X0oooo000:0?ooo`<000006`3oool00`000000oooo0?ooo`11
0?ooo`800000Y`3oool2000009T0oooo000;0?ooo`800000GP3oool300000:P0oooo0`00002G0?oo
o`002`3oool00`000000oooo0?ooo`1M0?ooo`800000ZP3oool3000009H0oooo000;0?ooo`030000
003oool0oooo05d0oooo0P00002[0?ooo`<00000U@3oool006/0oooo0P00002/0?ooo`<00000U03o
ool006/0oooo0P00002^0?ooo`<00000TP3oool006X0oooo0P00002`0?ooo`<00000T@3oool006X0
oooo0P00002b0?ooo`800000T03oool006X0oooo0P00000d0?ooo`030000003oool0oooo07`0oooo
0P0000180?ooo`030000003oool0oooo04@0oooo001Z0?ooo`800000503oool2000001d0oooo0`00
001o0?ooo`800000;@3oool00`000000oooo0?ooo`0F0?ooo`<00000A03oool00680oooo0P000005
0?ooo`8000005@3oool5000001X0oooo1@00001n0?ooo`800000:`3oool3000001L0oooo1@00000F
0?ooo`D000009`3oool00680oooo1@0000020?ooo`8000005@3oool4000001X0oooo1`00001n0?oo
o`800000:P3oool3000001L0oooo1000000G0?ooo`D00000603oool2000000d0oooo00140?ooo`D0
00006@3oool4000000<0oooo0P00000E0?ooo`@000006P3oool600000800oooo0`00000W0?ooo`D0
00005P3oool4000001L0oooo1000000I0?ooo`D000002P3oool004@0oooo1@00000I0?ooo`@00000
0`3oool2000001D0oooo0`00000J0?ooo`D00000Q03oool2000002D0oooo1`00000D0?ooo`@00000
603oool4000001T0oooo1@00000:0?ooo`009`3oool5000001P0oooo1000000J0?ooo`<00000103o
ool2000001@0oooo0`00000K0?ooo`@00000QP3oool300000280oooo1P00000F0?ooo`<000006@3o
ool3000001X0oooo1000000;0?ooo`009`3oool5000001P0oooo1000000J0?ooo`<000000`3oool2
000001D0oooo0`00000J0?ooo`<000007`3oool5000001<0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01D0oooo0`00000Q0?ooo`@00000603oool2
000001/0oooo0P00000J0?ooo`@000002`3oool000X0oooo1@00000H0?ooo`@000006@3oool30000
01/0oooo0P0000040?ooo`8000005@3oool2000001/0oooo0P00000K0?ooo`X00000503oool20000
01X0oooo0P00000L0?ooo`8000005P3oool9000001T0oooo0P00000K0?ooo`8000006`3oool20000
01/0oooo0P00000<0?ooo`002P3oool5000001P0oooo1000000J0?ooo`8000006`3oool2000000@0
oooo0P00000E0?ooo`800000>P3oool8000001@0oooo1000000I0?ooo`8000006`3oool3000001L0
oooo3000000a0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool2000000`0
oooo000:0?ooo`@000006@3oool3000001/0oooo0P00000K0?ooo`030000003oool0oooo00<0oooo
0P00000E0?ooo`030000003oool0oooo03`0oooo1@00000E0?ooo`D000005`3oool4000001X0oooo
1000000F0?ooo`/00000CP3oool00`000000oooo0?ooo`0J0?ooo`800000303oool000X0oooo1000
000J0?ooo`8000006`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0EP3oool2000001D0oooo1`00000F0?ooo`@000006@3oool6000001D0oooo0P00
00000`3oool0000000000004000006`0oooo00<000000?ooo`3oool02`3oool000X0oooo1000000J
0?ooo`8000006`3oool00`000000oooo0?ooo`0O0?ooo`800000L03oool5000001L0oooo1@00000I
0?ooo`H000005@3oool200000080oooo1000001/0?ooo`030000003oool0oooo00/0oooo000;0?oo
o`8000006`3oool00`000000oooo0?ooo`0l0?ooo`800000L@3oool3000001T0oooo1@00000I0?oo
o`@00000603oool200000080oooo0P00001j0?ooo`002`3oool2000001/0oooo00<000000?ooo`3o
ool0?03oool200000740oooo0`00000I0?ooo`@000006P3oool3000001X0oooo0P0000030?ooo`80
0000N03oool000/0oooo0P00001J0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01X0oooo0P0000040?ooo`800000
MP3oool000/0oooo00<000000?ooo`3oool0F03oool200000<`0oooo0P0000040?ooo`800000M03o
ool000/0oooo00<000000?ooo`3oool0F03oool200000<d0oooo0P0000050?ooo`800000LP3oool0
06H0oooo0P00003?0?ooo`8000001@3oool200000700oooo001V0?ooo`800000d03oool2000000H0
oooo0`00001]0?ooo`00IP3oool200000=80oooo0P0000070?ooo`<00000JP3oool006H0oooo00<0
00000?ooo`3oool0dP3oool2000000T0oooo0P00001X0?ooo`00IP3oool00`000000oooo0?ooo`3D
0?ooo`030000003oool0oooo00P0oooo0`00001U0?ooo`00I@3oool200000=L0oooo0P00000;0?oo
o`<00000HP3oool006D0oooo0P00003H0?ooo`8000003@3oool3000005l0oooo001U0?ooo`800000
603oool2000001d0oooo0P00002P0?ooo`8000003`3oool3000003D0oooo00<000000?ooo`3oool0
5P3oool3000000/0oooo001Q0?ooo`<0000000<0oooo000000000000603oool5000001X0oooo1@00
002N0?ooo`8000004@3oool400000300oooo0`00000G0?ooo`H00000203oool004@0oooo0`00000J
0?ooo`H00000603oool5000001X0oooo1000000N0?ooo`030000003oool0oooo07l0oooo0P00000D
0?ooo`<00000;@3oool4000001H0oooo1@0000090?ooo`009`3oool2000001/0oooo1P00000G0?oo
o`H00000603oool4000001/0oooo1000000M0?ooo`800000PP3oool2000001H0oooo0`00000Y0?oo
o`H000005@3oool5000000T0oooo000W0?ooo`D00000603oool6000001P0oooo1000000I0?ooo`@0
00006P3oool4000001d0oooo100000220?ooo`030000003oool0oooo01L0oooo1000000U0?ooo`D0
00005P3oool4000000X0oooo000:0?ooo`D00000603oool5000001T0oooo1000000I0?ooo`@00000
6P3oool2000001/0oooo0`00000M0?ooo`D00000PP3oool2000001`0oooo1000000P0?ooo`@00000
603oool3000000/0oooo000:0?ooo`D000006@3oool3000001X0oooo1000000I0?ooo`@000006P3o
ool2000001/0oooo0P00000M0?ooo`L00000PP3oool2000001l0oooo1@00000K0?ooo`<000006@3o
ool3000000/0oooo000:0?ooo`@000006P3oool3000001X0oooo0`00000J0?ooo`@000006P3oool2
000001/0oooo0P00000L0?ooo`H000005P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool07@3oool5000001D0oooo0`00000J0?ooo`800000303oool0
00/0oooo0`00000J0?ooo`<000006P3oool3000001/0oooo0P00000K0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool06P3oool3000001X0oooo0`00000J0?ooo`8000006`3oool20000
01/0oooo0P00000F0?ooo`8000000`3oool2000001/0oooo100000040?ooo`H000003`3oool00`00
0000oooo0?ooo`0J0?ooo`800000303oool000/0oooo0`00000J0?ooo`<000006`3oool2000001/0
oooo0P00000K0?ooo`030000003oool0oooo05D0oooo1000000H0?ooo`<000006P3oool2000001/0
oooo0`00000F0?ooo`030000003oool0oooo0080oooo0P00000K0?ooo`P000001@3oool5000002L0
oooo00<000000?ooo`3oool02`3oool000/0oooo0`00000K0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool06P3oool2000007@0oooo1@00000G0?ooo`<000006@3oool3000001/0oooo
0`00000E0?ooo`8000000`3oool4000001/0oooo1P00000:0?ooo`<00000<P3oool000/0oooo0P00
000L0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`1b0?ooo`L000005@3oool4000001P0oooo1000000J0?ooo`<000005P3oool010000000oooo0?oo
o`3oool5000001/0oooo1000000>0?ooo`D00000;@3oool000`0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0g0?ooo`800000M@3oool5000001H0oooo1@00000H0?ooo`@000006@3o
ool4000001D0oooo0P0000030?ooo`D000006`3oool3000001<0oooo0P00000[0?ooo`00303oool0
0`000000oooo0?ooo`1D0?ooo`800000MP3oool3000001L0oooo1P00000G0?ooo`@000006@3oool5
000001D0oooo00@000000?ooo`3oool0oooo1P00000L0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0:03oool006<0oooo0P00001f0?ooo`<00000603oool4000001P0oooo1@00000I
0?ooo`D00000503oool010000000oooo0?ooo`3oool5000003@0oooo00<000000?ooo`3oool09`3o
ool006<0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0G0?ooo`<000006@3oool4
000001X0oooo1000000E0?ooo`8000000`3oool3000003D0oooo00<000000?ooo`3oool09`3oool0
06<0oooo00<000000?ooo`3oool0T03oool00`000000oooo0?ooo`0I0?ooo`800000703oool20000
01L0oooo0P0000030?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool09P3oool006<0
oooo00<000000?ooo`3oool0i03oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo02H0
oooo001S0?ooo`030000003oool0oooo0>@0oooo0P00000l0?ooo`030000003oool0oooo02D0oooo
001S0?ooo`030000003oool0oooo0>@0oooo0P00000l0?ooo`030000003oool0oooo02D0oooo001S
0?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0U
0?ooo`00H`3oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo03X0oooo00<000000?oo
o`3oool09@3oool006<0oooo00<000000?ooo`3oool0i@3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo02D0oooo001R0?ooo`800000i`3oool00`000000oooo0?ooo`0j0?ooo`030000
003oool0oooo02D0oooo001R0?ooo`800000E03oool5000008h0oooo00<000000?ooo`3oool0>`3o
ool00`000000oooo0?ooo`0T0?ooo`00H@3oool5000001P0oooo1@00000H0?ooo`D00000603oool5
000008h0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0T0?ooo`002P3oool50000
01P0oooo1@00000H0?ooo`D00000603oool5000001P0oooo1@00000H0?ooo`D00000603oool50000
08h0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0T0?ooo`002P3oool5000001P0
oooo1@00000H0?ooo`D00000603oool5000001P0oooo1@00000H0?ooo`@000006P3oool3000008l0
oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0T0?ooo`002P3oool5000001P0oooo
1@00000H0?ooo`@000006P3oool3000001X0oooo0`00000J0?ooo`<000006P3oool3000008l0oooo
0P00000l0?ooo`030000003oool0oooo02@0oooo000;0?ooo`<000006P3oool3000001X0oooo0`00
000J0?ooo`<000006P3oool3000001X0oooo0`00000J0?ooo`<00000S`3oool2000003`0oooo00<0
00000?ooo`3oool0903oool000/0oooo0`00000J0?ooo`<000006P3oool3000001X0oooo0`00000J
0?ooo`<000006P3oool3000001X0oooo0`00002@0?ooo`030000003oool0oooo03X0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`0;0?ooo`002`3oool3000001X0oooo0`00000J0?oo
o`<000006P3oool3000001X0oooo0`00000J0?ooo`800000703oool00`000000oooo0?ooo`2>0?oo
o`8000006`3oool00`000000oooo0?ooo`0J0?ooo`050000003oool0oooo0?ooo`000000603oool0
0`000000oooo0?ooo`0;0?ooo`002`3oool3000001X0oooo0`00000J0?ooo`8000006`3oool20000
01`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
03L0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
01X0oooo0P00000K0?ooo`030000003oool0oooo01X0oooo00D000000?ooo`3oool0oooo0000000H
0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`800000703oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?oo
o`8000006`3oool00`000000oooo0?ooo`0I0?ooo`8000000`3oool00`000000oooo0?ooo`0E0?oo
o`<00000303oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo01T0oooo0P00000L0?ooo`030000003oool0oooo0740oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`0I0?ooo`8000006`3oool3000001X0oooo0`00000J0?ooo`<0
00000P3oool00`000000oooo0?ooo`0E0?ooo`<00000303oool000`0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0oooo08h0oooo0`00000J0?ooo`<00000
6P3oool3000001X0oooo0`00000J0?ooo`<000006P3oool300000080oooo00<000000?ooo`3oool0
5@3oool3000000`0oooo001R0?ooo`030000003oool0oooo08h0oooo0`00000J0?ooo`<000006P3o
ool3000001X0oooo0`00000J0?ooo`<000006P3oool300000080oooo00<000000?ooo`3oool05@3o
ool3000000`0oooo001R0?ooo`800000S`3oool3000001X0oooo0`00000J0?ooo`<000006P3oool3
000001X0oooo0`00000I0?ooo`@000000P3oool00`000000oooo0?ooo`0D0?ooo`D000002`3oool0
0680oooo0P00002?0?ooo`<000006P3oool3000001T0oooo1000000I0?ooo`D00000603oool50000
01P0oooo1@0000000`3oool000000?ooo`0E0?ooo`D000002`3oool00680oooo0P00002>0?ooo`D0
0000603oool5000001P0oooo1@00000H0?ooo`D00000603oool5000001P0oooo1@0000000`3oool0
00000?ooo`0E0?ooo`D000002`3oool00680oooo0P00002>0?ooo`D00000603oool5000001P0oooo
1@00000H0?ooo`D00000603oool5000001h0oooo00<000000?ooo`3oool0903oool00680oooo0P00
002>0?ooo`D00000603oool5000003L0oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?oo
o`0T0?ooo`00H`3oool00`000000oooo0?ooo`3U0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool0903oool006<0oooo00<000000?ooo`3oool0i@3oool00`000000oooo0?ooo`0k0?oo
o`030000003oool0oooo02@0oooo001S0?ooo`030000003oool0oooo0>D0oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`0U0?ooo`00H`3oool00`000000oooo0?ooo`3U0?ooo`030000
003oool0oooo03X0oooo00<000000?ooo`3oool09@3oool006<0oooo00<000000?ooo`3oool0i@3o
ool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo02D0oooo001S0?ooo`030000003oool0
oooo0>D0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0U0?ooo`00H`3oool20000
0>H0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0U0?ooo`00I03oool00`000000
oooo0?ooo`0G0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0[03oool00`000000
oooo0?ooo`0j0?ooo`030000003oool0oooo02D0oooo001P0?ooo`040000003oool0oooo0?ooo`80
00005P3oool3000001/0oooo0`00000K0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0M03oool2000003`0oooo00<000000?ooo`3oool09@3oool005h0oooo0`0000030?ooo`800000
503oool5000001T0oooo1@00000I0?ooo`<000006@3oool2000007H0oooo0P00000l0?ooo`030000
003oool0oooo02D0oooo000o0?ooo`030000003oool0oooo01X0oooo1P0000030?ooo`030000003o
ool0oooo01<0oooo1@00000I0?ooo`@000005`3oool5000001P0oooo1000001e0?ooo`800000?03o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00/0oooo000n0?ooo`<00000703oool5
000000<0oooo0P00000E0?ooo`@000006@3oool4000001P0oooo1@00000F0?ooo`D00000M@3oool0
0`000000oooo0?ooo`0h0?ooo`040000003oool0oooo000001T0oooo00<000000?ooo`3oool02`3o
ool003h0oooo1000000L0?ooo`@000000`3oool2000001H0oooo0`00000J0?ooo`@00000603oool4
000001L0oooo1@00001d0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06P3oool0
10000000oooo0?ooo`00000H0?ooo`8000003@3oool003d0oooo1`00000K0?ooo`@000000`3oool0
0`000000oooo0?ooo`0D0?ooo`<000006`3oool3000001T0oooo0`00000I0?ooo`<00000M03oool2
000001`0oooo00<000000?ooo`3oool06P3oool010000000oooo0?ooo`00000H0?ooo`<00000303o
ool003d0oooo2000000L0?ooo`8000000`3oool2000001H0oooo0`00000K0?ooo`8000006P3oool2
000001X0oooo0`00001G0?ooo`030000003oool0oooo01T0oooo0P00000K0?ooo`8000006`3oool3
000000030?ooo`000000oooo01L0oooo0`00000<0?ooo`00303oool00`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo01@0oooo1`00000K0?ooo`8000000`3oool00`000000oooo0?ooo`0E0?oo
o`8000006`3oool2000001/0oooo0P00000J0?ooo`<000006`3oool2000001`0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`8000006`3oool3000001X0oooo0`0000000`3o
ool000000?ooo`0G0?ooo`<00000303oool000/0oooo0P00000J0?ooo`8000006`3oool3000001`0
oooo00@000000?ooo`3oool0oooo0P00000G0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01D0oooo1@00000L0?oo
o`8000006`3oool2000001/0oooo0P00000K0?ooo`<000006P3oool3000000030?ooo`000000oooo
01H0oooo1000000<0?ooo`002`3oool2000001T0oooo0`00000o0?ooo`030000003oool0oooo0840
oooo1`00000M0?ooo`8000006`3oool2000001/0oooo0P00000J0?ooo`@000006P3oool300000003
0?ooo`000000oooo01H0oooo1@00000;0?ooo`002P3oool3000001P0oooo0`0000110?ooo`030000
003oool0oooo0800oooo1P00000M0?ooo`<000006P3oool3000001X0oooo1000000I0?ooo`@00000
6@3oool5000001P0oooo1@00000;0?ooo`002@3oool3000001L0oooo1@0000120?ooo`030000003o
ool0oooo0800oooo1000000M0?ooo`@000006P3oool4000001T0oooo1000000I0?ooo`D00000603o
ool5000001X0oooo0`00000;0?ooo`00203oool4000001H0oooo1@0000140?ooo`030000003oool0
oooo07l0oooo0`00000M0?ooo`@000006`3oool4000001T0oooo1000000H0?ooo`H00000603oool5
000002P0oooo00080?ooo`@000005`3oool4000004@0oooo0P0000210?ooo`030000003oool0oooo
01`0oooo1000000J0?ooo`D00000603oool5000001P0oooo1P00000L0?ooo`030000003oool0oooo
02H0oooo00070?ooo`D00000603oool2000004H0oooo0P00002N0?ooo`D000006P3oool5000001P0
oooo1@00000j0?ooo`030000003oool0oooo02H0oooo00090?ooo`<000006@3oool00`000000oooo
0?ooo`150?ooo`800000WP3oool4000001`0oooo0`00000I0?ooo`030000003oool0oooo03/0oooo
00<000000?ooo`3oool09P3oool006h0oooo0P00002O0?ooo`800000=`3oool00`000000oooo0?oo
o`0l0?ooo`030000003oool0oooo02H0oooo001_0?ooo`800000e`3oool00`000000oooo0?ooo`0l
0?ooo`030000003oool0oooo02H0oooo001`0?ooo`800000eP3oool00`000000oooo0?ooo`0l0?oo
o`030000003oool0oooo02H0oooo001b0?ooo`030000003oool0oooo0=<0oooo00<000000?ooo`3o
ool0?03oool00`000000oooo0?ooo`0V0?ooo`00L`3oool200000=<0oooo00<000000?ooo`3oool0
?03oool00`000000oooo0?ooo`0V0?ooo`00M03oool200000=40oooo0P00000n0?ooo`030000003o
ool0oooo02H0oooo001f0?ooo`800000c`3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0
oooo02L0oooo001g0?ooo`<00000c@3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo
02L0oooo001i0?ooo`<00000b`3oool00`000000oooo0?ooo`0l0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool02P3oool007X0oooo0`00000L0?ooo`030000003oool0oooo01/0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`1a0?ooo`030000003oool0oooo03`0oooo
00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0:0?ooo`00N`3oool3000001X0oooo0`00
000K0?ooo`8000006P3oool300000740oooo0P00000n0?ooo`030000003oool0oooo01T0oooo0P00
000<0?ooo`00N@3oool010000000oooo0?ooo`3oool3000001L0oooo1000000J0?ooo`@00000603o
ool400000740oooo0P00000n0?ooo`030000003oool0oooo01T0oooo0P00000<0?ooo`00303oool0
0`000000oooo0?ooo`1Z0?ooo`8000000`3oool3000001D0oooo1P00000H0?ooo`D000005`3oool6
00000700oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0I0?ooo`8000006P3oool3
000000`0oooo000<0?ooo`030000003oool0oooo06T0oooo1@0000020?ooo`8000005P3oool50000
01T0oooo1@00000G0?ooo`D000005P3oool2000005P0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`0I0?ooo`8000006`3oool2000001X0oooo1000000;0?ooo`002`3oool2000001`0
oooo00<000000?ooo`3oool0C03oool6000000<0oooo00<000000?ooo`3oool05@3oool4000001X0
oooo0`00000I0?ooo`@000005@3oool5000005@0oooo0P0000040?ooo`8000006`3oool2000001X0
oooo0`00000J0?ooo`@000002`3oool000/0oooo0P00000K0?ooo`800000703oool00`000000oooo
0?ooo`0_0?ooo`L000000`3oool00`000000oooo0?ooo`0E0?ooo`@000006P3oool3000001T0oooo
1000000D0?ooo`P00000>`3oool00`000000oooo0?ooo`0C0?ooo`800000103oool2000001X0oooo
0`00000J0?ooo`@00000603oool5000000/0oooo000:0?ooo`<000006`3oool2000001/0oooo0P00
000L0?ooo`030000003oool0oooo01D0oooo1P0000020?ooo`8000005`3oool2000001/0oooo0P00
000K0?ooo`800000503oool:000001/0oooo0P00000K0?ooo`8000005@3oool2000000<0oooo0`00
000J0?ooo`<000006P3oool4000001P0oooo1@00000;0?ooo`002P3oool4000001T0oooo0`00000K
0?ooo`8000006P3oool2000001`0oooo0`0000020?ooo`800000603oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool04@3oool7000001d0oooo0P00000L
0?ooo`8000005@3oool00`000000oooo0?ooo`020?ooo`<000006@3oool5000001P0oooo1@00000X
0?ooo`002P3oool4000001T0oooo0`00000J0?ooo`<00000603oool4000002<0oooo00<000000?oo
o`3oool0H`3oool300000200oooo0`00000K0?ooo`<000005@3oool010000000oooo0?ooo`3oool4
000001T0oooo1@00000H0?ooo`D00000:03oool000T0oooo1@00000H0?ooo`D00000603oool40000
01H0oooo1@00000U0?ooo`800000Q03oool4000001/0oooo0`00000E0?ooo`8000000`3oool40000
01P0oooo1P00000J0?ooo`<00000:03oool000T0oooo1@00000H0?ooo`D000005`3oool4000001D0
oooo1P00000W0?ooo`800000P@3oool6000001X0oooo1000000E0?ooo`8000000P3oool5000001P0
oooo1P00000J0?ooo`030000003oool0oooo02P0oooo000;0?ooo`<000005`3oool6000001L0oooo
1000000F0?ooo`@00000:@3oool2000007l0oooo1P00000K0?ooo`@000005@3oool010000000oooo
0?ooo`3oool5000003P0oooo00<000000?ooo`3oool0:03oool002D0oooo1P00000F0?ooo`D00000
5`3oool3000002/0oooo00<000000?ooo`3oool0O@3oool5000001X0oooo1@00000E0?ooo`030000
003oool0oooo00<0oooo0`00000h0?ooo`030000003oool0oooo02P0oooo000X0?ooo`<000005`3o
ool4000001P0oooo00<000000?ooo`3oool0:`3oool2000007h0oooo0`00000M0?ooo`<000005@3o
ool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo02P0oooo00140?ooo`800000A`3oool2
000007l0oooo00<000000?ooo`3oool0<P3oool200000400oooo00<000000?ooo`3oool0:03oool0
08h0oooo0P00002c0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0:03oool00900
oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`1Z0?ooo`00T@3oool200000;00oooo
00<000000?ooo`3oool0JP3oool00980oooo0P00002^0?ooo`800000K03oool009<0oooo0P00002]
0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool02P3oool009D0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo00X0oooo002F0?ooo`80
0000ZP3oool00`000000oooo0?ooo`1M0?ooo`800000303oool009L0oooo0P00002X0?ooo`800000
@`3oool00`000000oooo0?ooo`0I0?ooo`<000002`3oool009P0oooo0P00002W0?ooo`030000003o
ool0oooo0480oooo00<000000?ooo`3oool06@3oool3000000/0oooo000=0?ooo`030000003oool0
oooo08X0oooo0P00000K0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0J`3oool0
0`000000oooo0?ooo`110?ooo`8000006P3oool4000000/0oooo000=0?ooo`030000003oool0oooo
08P0oooo00<000000?ooo`3oool00P00000I0?ooo`8000006P3oool3000006`0oooo00<000000?oo
o`3oool09@3oool00`000000oooo0?ooo`0I0?ooo`8000006P3oool4000000/0oooo000<0?ooo`80
0000703oool00`000000oooo0?ooo`1Z0?ooo`<000000P3oool2000001H0oooo1@00000H0?ooo`@0
0000J`3oool2000002L0oooo00<000000?ooo`3oool06@3oool2000001X0oooo1@00000:0?ooo`00
303oool2000001/0oooo0P00001/0?ooo`@000000P3oool2000001@0oooo1P00000G0?ooo`H00000
5`3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
6@3oool2000001X0oooo1000000H0?ooo`H000002P3oool000/0oooo0`00000K0?ooo`800000703o
ool00`000000oooo0?ooo`1<0?ooo`H000000`3oool2000001<0oooo1P00000G0?ooo`D000005`3o
ool300000500oooo00<000000?ooo`3oool0203oool2000001/0oooo0P00000J0?ooo`@00000603o
ool6000000X0oooo000;0?ooo`@000006@3oool3000001/0oooo0P00001>0?ooo`H00000103oool2
000001@0oooo1000000I0?ooo`@000005@3oool6000003d0oooo00<000000?ooo`3oool03@3oool2
000000X0oooo0P00000J0?ooo`<000006P3oool4000001X0oooo00<000000?ooo`3oool02`3oool0
00/0oooo1000000I0?ooo`<000006`3oool2000001`0oooo00<000000?ooo`3oool04@3oool50000
01/0oooo1@0000040?ooo`800000503oool4000001T0oooo1000000D0?ooo`P00000>P3oool20000
00l0oooo00<000000?ooo`3oool0203oool3000001X0oooo1000000H0?ooo`D000006P3oool00`00
0000oooo0?ooo`0;0?ooo`002P3oool5000001P0oooo1000000J0?ooo`<000006`3oool2000001<0
oooo2P00000H0?ooo`@00000103oool2000001D0oooo0P00000K0?ooo`800000503oool:000001/0
oooo0P00000K0?ooo`8000003`3oool00`000000oooo0?ooo`080?ooo`<000006P3oool4000001P0
oooo1@00000J0?ooo`030000003oool0oooo00/0oooo000:0?ooo`D00000603oool4000001T0oooo
1000000J0?ooo`800000503oool8000001d0oooo0P0000050?ooo`8000005@3oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool05`3oool3000001/0oooo
0`00000?0?ooo`030000003oool0oooo00L0oooo1000000I0?ooo`D00000A@3oool000`0oooo0`00
000G0?ooo`D000006@3oool3000001T0oooo1000000E0?ooo`D000009`3oool2000006D0oooo1@00
000K0?ooo`@000003P3oool2000000T0oooo1000000I0?ooo`D00000A@3oool002H0oooo1@00000H
0?ooo`@00000603oool4000001H0oooo0P00000[0?ooo`800000HP3oool6000001/0oooo1000000?
0?ooo`030000003oool0oooo00L0oooo1@00000K0?ooo`<00000A@3oool002P0oooo0`00000H0?oo
o`@000005`3oool5000004@0oooo0`00001P0?ooo`H000006`3oool4000000l0oooo00<000000?oo
o`3oool01`3oool5000006<0oooo00120?ooo`D000005P3oool6000004H0oooo0P00001P0?ooo`@0
00006`3oool5000000l0oooo00<000000?ooo`3oool02@3oool3000006<0oooo00140?ooo`<00000
5`3oool4000004P0oooo0P00001O0?ooo`<000007@3oool4000000h0oooo00<000000?ooo`3oool0
L03oool00600oooo0P0000190?ooo`800000G`3oool00`000000oooo0?ooo`0N0?ooo`8000003P3o
ool00`000000oooo0?ooo`1`0?ooo`00[03oool3000008h0oooo00<000000?ooo`3oool0L03oool0
0:h0oooo0P00002<0?ooo`800000LP3oool00:l0oooo0P00002;0?ooo`030000003oool0oooo06@0
oooo00<000000?ooo`3oool02P3oool00;00oooo0P00002:0?ooo`030000003oool0oooo06@0oooo
00<000000?ooo`3oool02P3oool00;40oooo0P0000290?ooo`030000003oool0oooo06<0oooo0P00
000<0?ooo`00/P3oool2000008L0oooo00<000000?ooo`3oool0I03oool3000000/0oooo002c0?oo
o`800000QP3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo01P0oooo1000000;0?oo
o`003@3oool00`000000oooo0?ooo`2T0?ooo`800000Q@3oool00`000000oooo0?ooo`180?ooo`03
0000003oool0oooo01P0oooo1000000;0?ooo`003@3oool00`000000oooo0?ooo`2U0?ooo`030000
003oool0oooo0880oooo00<000000?ooo`3oool0B03oool2000001X0oooo1000000;0?ooo`00303o
ool200000:P0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo01T0oooo0`00000H0?ooo`H000002P3oool000`0oooo0P00000L0?ooo`030000003oool0
oooo08X0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0/0?ooo`030000003oool0
oooo01P0oooo1000000H0?ooo`H000002P3oool000/0oooo0`00000K0?ooo`800000S03oool20000
01/0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`0/0?ooo`8000006P3oool40000
01X0oooo1000000:0?ooo`002`3oool4000001X0oooo0P00001]0?ooo`030000003oool0oooo01/0
oooo1000000I0?ooo`<00000603oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo0100
oooo00<000000?ooo`3oool06@3oool2000001X0oooo1000000J0?ooo`030000003oool0oooo00/0
oooo000;0?ooo`@000006@3oool3000001`0oooo00<000000?ooo`3oool0C@3oool3000001/0oooo
1P00000G0?ooo`@000005`3oool3000004T0oooo00<000000?ooo`3oool03`3oool2000001/0oooo
0P00000I0?ooo`H000006@3oool00`000000oooo0?ooo`0;0?ooo`002P3oool5000001T0oooo0`00
000K0?ooo`800000CP3oool5000001T0oooo2000000E0?ooo`H000005P3oool4000004L0oooo0P00
000A0?ooo`8000006P3oool4000001P0oooo1P00000I0?ooo`030000003oool0oooo00/0oooo000:
0?ooo`D00000603oool5000001X0oooo0P00000L0?ooo`030000003oool0oooo02h0oooo1`00000I
0?ooo`D0000000<0oooo0000000000005@3oool5000001D0oooo1`00000l0?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool03`3oool3000001X0oooo1000000J0?ooo`@000006@3oool0
0`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0H0?ooo`D000006@3oool30000
01/0oooo0P00000b0?ooo`H000006P3oool4000000030?ooo`000000000001H0oooo1000000D0?oo
o`P000007@3oool2000001/0oooo0P0000080?ooo`030000003oool0oooo00l0oooo0`00000J0?oo
o`@000006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000<0?ooo`03
0000003oool0oooo01L0oooo1P00000I0?ooo`<000006P3oool3000001P0oooo1@00000H0?ooo`@0
00006`3oool200000080oooo0P00000G0?ooo`8000005`3oool6000001X0oooo0`00000L0?ooo`80
00001`3oool200000100oooo1000000I0?ooo`D000006P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo00/0oooo000V0?ooo`H00000603oool4000001T0oooo0`00000D0?ooo`T00000
703oool2000001`0oooo00@000000?ooo`3oool000006@3oool00`000000oooo0?ooo`0H0?ooo`<0
00005P3oool5000001`0oooo0`0000070?ooo`030000003oool0oooo00l0oooo1000000I0?ooo`D0
0000A@3oool002T0oooo0`00000H0?ooo`@000006@3oool3000001@0oooo1`0000100?ooo`030000
003oool0oooo04P0oooo1P00000L0?ooo`<00000203oool00`000000oooo0?ooo`0>0?ooo`D00000
H`3oool004<0oooo1@00000H0?ooo`@000005@3oool500000480oooo00<000000?ooo`3oool0A`3o
ool6000001/0oooo100000070?ooo`030000003oool0oooo00l0oooo1@00001S0?ooo`00@`3oool5
000001L0oooo1@00000E0?ooo`@00000A03oool00`000000oooo0?ooo`170?ooo`@00000703oool4
000000L0oooo00<000000?ooo`3oool04@3oool3000006<0oooo00150?ooo`<000005P3oool50000
01L0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`<00000703oool50000
00L0oooo00<000000?ooo`3oool0M`3oool005l0oooo1000001P0?ooo`030000003oool0oooo04H0
oooo00<000000?ooo`3oool07@3oool3000000H0oooo00<000000?ooo`3oool0N03oool00640oooo
0P00001Q0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0K03oool00`000000oooo
0?ooo`090?ooo`00a@3oool00`000000oooo0?ooo`1/0?ooo`800000KP3oool00`000000oooo0?oo
o`090?ooo`00aP3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo06`0oooo0P00000;
0?ooo`00a`3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo06`0oooo0P00000;0?oo
o`003@3oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo06P0oooo0P00001]0?ooo`<0
00002`3oool000d0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`1W0?ooo`030000
003oool0oooo0540oooo00<000000?ooo`3oool0603oool4000000X0oooo000<0?ooo`800000^`3o
ool2000006L0oooo0P00001C0?ooo`030000003oool0oooo01P0oooo1000000:0?ooo`00303oool2
00000;`0oooo0P00001V0?ooo`030000003oool0oooo0540oooo0P00000I0?ooo`D000002P3oool0
00`0oooo0P00002m0?ooo`800000I03oool2000005<0oooo0P00000I0?ooo`D000002P3oool000/0
oooo1000000K0?ooo`030000003oool0oooo09l0oooo0P00001S0?ooo`030000003oool0oooo03H0
oooo00<000000?ooo`3oool0603oool3000001/0oooo00<000000?ooo`3oool02P3oool000/0oooo
1000000K0?ooo`030000003oool0oooo0:00oooo0P00001Q0?ooo`800000=`3oool2000001X0oooo
1000000J0?ooo`030000003oool0oooo00X0oooo000;0?ooo`@000006P3oool200000:<0oooo0P00
001P0?ooo`030000003oool0oooo03H0oooo0P00000J0?ooo`@000006P3oool00`000000oooo0?oo
o`0:0?ooo`002P3oool5000001X0oooo0P00000L0?ooo`030000003oool0oooo06d0oooo00<00000
0?ooo`3oool05@3oool2000000<0oooo00<000000?ooo`3oool0F03oool2000001`0oooo00<00000
0?ooo`3oool0603oool3000001T0oooo1@00000J0?ooo`030000003oool0oooo00X0oooo000:0?oo
o`D000006@3oool3000001/0oooo0P00001?0?ooo`030000003oool0oooo01`0oooo0`00000G0?oo
o`80000000@0oooo00000000000000006@3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo01X0oooo0P00000J0?ooo`<000006@3oool5000001X0oooo00<000000?ooo`3oool02P3oool0
00`0oooo00<000000?ooo`3oool06@3oool4000001X0oooo0P00001?0?ooo`800000703oool40000
01P0oooo1@00000H0?ooo`<00000?@3oool2000001`0oooo0P00000I0?ooo`@000006`3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01T0
oooo1000000I0?ooo`<00000703oool00`000000oooo0?ooo`0_0?ooo`@000006P3oool6000001L0
oooo1P00000G0?ooo`@00000?03oool2000001/0oooo0`00000I0?ooo`@000006P3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01P0oooo
1@00000I0?ooo`<000006`3oool200000340oooo1P00000I0?ooo`D00000603oool5000001H0oooo
1P00000j0?ooo`<000006P3oool4000001P0oooo1@00000J0?ooo`030000003oool0oooo01X0oooo
00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0603oool5000001P0oooo1000
000K0?ooo`8000006`3oool2000001@0oooo1`00000J0?ooo`@000006@3oool4000001D0oooo1`00
000M0?ooo`8000006P3oool3000001X0oooo0`00000I0?ooo`D000006P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool0603oool4000001X0oooo0`00000I0?ooo`<000005`3oool6000001/0oooo0P00000J
0?ooo`<000005`3oool6000001P0oooo1@00000J0?ooo`@000006@3oool4000001/0oooo0`00000J
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02`3oool004<0oooo1@00000I0?oo
o`@000005P3oool5000001`0oooo0`00000L0?ooo`030000003oool0oooo01P0oooo0`00000J0?oo
o`<00000503oool7000001/0oooo1000000I0?ooo`@00000H`3oool004<0oooo1@00000H0?ooo`@0
00005@3oool6000005P0oooo0P00000a0?ooo`H000006`3oool4000001T0oooo1@00001S0?ooo`00
A@3oool3000001P0oooo1000000F0?ooo`@00000FP3oool200000340oooo1000000L0?ooo`@00000
6`3oool3000005L0oooo00<000000?ooo`3oool02@3oool005l0oooo1@00000F0?ooo`<00000G03o
ool200000300oooo0`00000L0?ooo`D00000M@3oool00`000000oooo0?ooo`090?ooo`00H03oool4
000001L0oooo00<000000?ooo`3oool0G03oool200000300oooo00<000000?ooo`3oool06`3oool5
000007@0oooo0P00000;0?ooo`00HP3oool2000007L0oooo0P00001<0?ooo`040000003oool0oooo
0?ooo`800000M03oool2000000/0oooo003L0?ooo`800000BP3oool2000007T0oooo0P00000;0?oo
o`003P3oool00`000000oooo0?ooo`3<0?ooo`800000B@3oool00`000000oooo0?ooo`1g0?ooo`@0
00002P3oool000h0oooo00<000000?ooo`3oool0c@3oool2000004L0oooo00<000000?ooo`3oool0
G@3oool00`000000oooo0?ooo`0H0?ooo`@000002P3oool000d0oooo0P00003@0?ooo`800000A@3o
ool2000005l0oooo00<000000?ooo`3oool0603oool4000000X0oooo000=0?ooo`800000d@3oool2
000004@0oooo00<000000?ooo`3oool0G@3oool2000001T0oooo1@00000:0?ooo`00303oool30000
0=80oooo0P0000120?ooo`030000003oool0oooo05h0oooo0P00000I0?ooo`D000002P3oool000`0
oooo1000000K0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0?P3oool00`000000
oooo0?ooo`1O0?ooo`8000006`3oool00`000000oooo0?ooo`0:0?ooo`00303oool4000001X0oooo
0P00002h0?ooo`800000?@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo01P0oooo
1000000J0?ooo`030000003oool0oooo00X0oooo000;0?ooo`D000006P3oool200000;T0oooo0P00
000k0?ooo`030000003oool0oooo04@0oooo0P00000J0?ooo`@000006P3oool00`000000oooo0?oo
o`0:0?ooo`002`3oool5000001T0oooo0`00002j0?ooo`800000>P3oool00`000000oooo0?ooo`14
0?ooo`8000006P3oool4000001X0oooo00<000000?ooo`3oool02P3oool000d0oooo0`00000I0?oo
o`<00000703oool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo03H0oooo00<000000?oo
o`3oool0A03oool3000001T0oooo1@00000J0?ooo`030000003oool0oooo00X0oooo000=0?ooo`03
0000003oool0oooo01P0oooo1000000K0?ooo`800000X03oool00`000000oooo0?ooo`0d0?ooo`03
0000003oool0oooo02X0oooo00<000000?ooo`3oool0603oool3000001T0oooo1@00000J0?ooo`03
0000003oool0oooo00X0oooo000=0?ooo`030000003oool0oooo01P0oooo1000000K0?ooo`800000
KP3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
<P3oool00`000000oooo0?ooo`0Z0?ooo`8000006@3oool5000001X0oooo00<000000?ooo`3oool0
6P3oool00`000000oooo0?ooo`0:0?ooo`003@3oool00`000000oooo0?ooo`0G0?ooo`D000006P3o
ool3000004h0oooo00<000000?ooo`3oool0703oool3000001X0oooo0`00000E0?ooo`8000000P3o
ool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo02/0oooo0P00000I0?ooo`D000006P3o
ool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0
oooo01P0oooo1@00000J0?ooo`<00000703oool00`000000oooo0?ooo`0_0?ooo`<000006`3oool4
000001T0oooo1000000F0?ooo`D00000;03oool00`000000oooo0?ooo`0[0?ooo`<00000603oool6
000001T0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`0J0?ooo`<000006@3oool4000001/0oooo0P00000a0?ooo`@000006@3oool60000
01L0oooo1P00000G0?ooo`@00000:P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
01P0oooo0`00000H0?ooo`H000006@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
00/0oooo000<0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06@3oool4000001X0
oooo0`00000`0?ooo`L00000603oool6000001L0oooo1@00000F0?ooo`L000009`3oool00`000000
oooo0?ooo`0A0?ooo`8000006@3oool4000001X0oooo1000000I0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0H0?ooo`D000006@3oool3000001`0oooo0P00000C0?ooo`T00000603oool5000001P0oooo
1000000E0?ooo`P000006P3oool5000000H0oooo00<000000?ooo`3oool04@3oool2000001X0oooo
1000000J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo1@00
000I0?ooo`<000006@3oool4000001@0oooo2P00000I0?ooo`<000006@3oool4000001L0oooo1P00
000D0?ooo`T000001@3oool2000001<0oooo0`00000I0?ooo`D000006P3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`0000002`3oool004H0oooo0`00000H
0?ooo`@000005P3oool6000003X0oooo0P00000J0?ooo`8000006`3oool3000001<0oooo1`000006
0?ooo`8000004`3oool4000001T0oooo1@00001F0?ooo`030000003oool0oooo00T0oooo001P0?oo
o`D00000503oool7000005T0oooo00<000000?ooo`3oool06`3oool200000180oooo1@0000050?oo
o`800000503oool5000001/0oooo0`00001E0?ooo`8000002`3oool005l0oooo1@00000F0?ooo`D0
0000N@3oool300000100oooo100000040?ooo`8000005@3oool5000007@0oooo0`00000:0?ooo`00
H03oool4000001L0oooo0`00001l0?ooo`@000003P3oool010000000oooo0?ooo`3oool3000001P0
oooo1000001c0?ooo`@000002P3oool000h0oooo00<000000?ooo`3oool0D@3oool2000001P0oooo
00<000000?ooo`3oool0OP3oool4000000`0oooo1000000L0?ooo`800000L`3oool4000000X0oooo
000>0?ooo`030000003oool0oooo0?00oooo3000002E0?ooo`@000002P3oool000d0oooo0P00003o
0?ooogT0oooo00<000000?ooo`3oool05`3oool6000000T0oooo000=0?ooo`800000o`3ooomi0?oo
o`030000003oool0oooo01L0oooo1P0000090?ooo`00303oool300000?l0ooooN03oool2000001/0
oooo100000090?ooo`00303oool400000?l0ooooM`3oool2000001/0oooo00<000000?ooo`3oool0
2P3oool000`0oooo1000000K0?ooo`030000003oool0oooo0?l0ooooF03oool3000001/0oooo00<0
00000?ooo`3oool02P3oool000/0oooo1@00000J0?ooo`800000o`3ooomJ0?ooo`@000006P3oool0
0`000000oooo0?ooo`0:0?ooo`002`3oool5000001X0oooo0P00003o0?ooocl0oooo00<000000?oo
o`3oool0603oool4000001X0oooo00<000000?ooo`3oool02P3oool000d0oooo00<000000?ooo`3o
ool06@3oool300000?l0oooo?`3oool00`000000oooo0?ooo`0G0?ooo`D000006P3oool00`000000
oooo0?ooo`0:0?ooo`003@3oool00`000000oooo0?ooo`0I0?ooo`<00000o`3oooln0?ooo`800000
6@3oool5000001X0oooo00<000000?ooo`3oool02P3oool000d0oooo00<000000?ooo`3oool0603o
ool5000001/0oooo00<000000?ooo`3oool0o`3ooolO0?ooo`8000006`3oool3000001X0oooo00<0
00000?ooo`3oool02P3oool000d0oooo00<000000?ooo`3oool0603oool5000001X0oooo0P00003o
0?ooo`D0oooo00<000000?ooo`3oool0603oool3000001/0oooo00<000000?ooo`3oool06P3oool0
0`000000oooo0?ooo`0:0?ooo`003@3oool00`000000oooo0?ooo`0G0?ooo`H000006P3oool20000
0?l0oooo103oool2000001X0oooo1000000J0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool02P3oool000d0oooo00<000000?ooo`3oool05`3oool6000001T0oooo0`00001^0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool0M@3oool2000001X0oooo1000000J0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool02`3oool000`0oooo00<000000?ooo`3oool0
6`3oool3000001T0oooo0`00000L0?ooo`030000003oool0oooo04h0oooo0`00000J0?ooo`<00000
6P3oool00`000000oooo0?ooo`1H0?ooo`<000006@3oool5000001X0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0K0?ooo`030000003o
ool0oooo01P0oooo1000000K0?ooo`800000<@3oool00`000000oooo0?ooo`0K0?ooo`@000006@3o
ool4000001T0oooo0`00000n0?ooo`030000003oool0oooo01P0oooo0`00000I0?ooo`D000006@3o
ool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00/0oooo000<0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06@3oool4000001/0oooo0P00000a0?ooo`<000006P3oool6
000001L0oooo1P00000G0?ooo`D00000?03oool2000001T0oooo1000000J0?ooo`030000003oool0
oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0oooo1@00000J0?ooo`8000007@3oool00`00
0000oooo0?ooo`0A0?ooo`H000006@3oool5000001P0oooo1@00000G0?ooo`H00000>P3oool30000
01T0oooo1000000J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0;0?ooo`00303oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01P0
oooo1@00000I0?ooo`<000006`3oool2000001@0oooo2000000I0?ooo`@000006@3oool4000001D0
oooo2000000J0?ooo`D000006@3oool3000001T0oooo1@00000J0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0;0?ooo`00303oool00`000000oooo
0?ooo`0J0?ooo`030000003oool0oooo01X0oooo0`00000H0?ooo`@000006@3oool3000001D0oooo
2P00000H0?ooo`@000006@3oool4000001L0oooo1P00000D0?ooo`T000006P3oool3000001T0oooo
1@00000J0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`0;0?ooo`00H@3oool4000001L0oooo1000000j0?ooo`8000006`3oool2000001/0oooo0`00
000C0?ooo`L000006`3oool4000001/0oooo0`00001R0?ooo`00H03oool4000001H0oooo1P00000l
0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0;`3oool5000001/0oooo1@000020
0?ooo`00H@3oool3000001L0oooo1000002<0?ooo`@000006`3oool500000840oooo001S0?ooo`03
0000003oool0oooo01D0oooo0`00002>0?ooo`030000003oool0oooo01`0oooo100000210?ooo`00
O03oool00`000000oooo0?ooo`2^0?ooo`800000P@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {369.875, 0}} -> {-3.71265, -3.18613, \
0.0258421, 0.0258421}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(xyCurves[3, 1, 1, 20];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.047619 0.290082 0.102266 [
[.2619 .27758 -5 -15.6875 ]
[.2619 .27758 5 0 ]
[.5 .27758 -8 -15.6875 ]
[.5 .27758 8 0 ]
[.7381 .27758 -8 -15.6875 ]
[.7381 .27758 8 0 ]
[.97619 .27758 -8 -15.6875 ]
[.97619 .27758 8 0 ]
[.01131 .08555 -18.6875 -7.84375 ]
[.01131 .08555 0 7.84375 ]
[.01131 .18782 -18.6875 -7.84375 ]
[.01131 .18782 0 7.84375 ]
[.01131 .39235 -10 -7.84375 ]
[.01131 .39235 0 7.84375 ]
[.01131 .49461 -10 -7.84375 ]
[.01131 .49461 0 7.84375 ]
[.01131 .59688 -10 -7.84375 ]
[.01131 .59688 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .29008 m
.2619 .29633 L
s
gsave
.2619 .27758 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.5 .29008 m
.5 .29633 L
s
gsave
.5 .27758 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.7381 .29008 m
.7381 .29633 L
s
gsave
.7381 .27758 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.97619 .29008 m
.97619 .29633 L
s
gsave
.97619 .27758 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .29008 m
.07143 .29383 L
s
.11905 .29008 m
.11905 .29383 L
s
.16667 .29008 m
.16667 .29383 L
s
.21429 .29008 m
.21429 .29383 L
s
.30952 .29008 m
.30952 .29383 L
s
.35714 .29008 m
.35714 .29383 L
s
.40476 .29008 m
.40476 .29383 L
s
.45238 .29008 m
.45238 .29383 L
s
.54762 .29008 m
.54762 .29383 L
s
.59524 .29008 m
.59524 .29383 L
s
.64286 .29008 m
.64286 .29383 L
s
.69048 .29008 m
.69048 .29383 L
s
.78571 .29008 m
.78571 .29383 L
s
.83333 .29008 m
.83333 .29383 L
s
.88095 .29008 m
.88095 .29383 L
s
.92857 .29008 m
.92857 .29383 L
s
.25 Mabswid
0 .29008 m
1 .29008 L
s
.02381 .08555 m
.03006 .08555 L
s
gsave
.01131 .08555 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .18782 m
.03006 .18782 L
s
gsave
.01131 .18782 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .39235 m
.03006 .39235 L
s
gsave
.01131 .39235 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.02381 .49461 m
.03006 .49461 L
s
gsave
.01131 .49461 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .59688 m
.03006 .59688 L
s
gsave
.01131 .59688 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .106 m
.02756 .106 L
s
.02381 .12646 m
.02756 .12646 L
s
.02381 .14691 m
.02756 .14691 L
s
.02381 .16736 m
.02756 .16736 L
s
.02381 .20827 m
.02756 .20827 L
s
.02381 .22872 m
.02756 .22872 L
s
.02381 .24918 m
.02756 .24918 L
s
.02381 .26963 m
.02756 .26963 L
s
.02381 .31054 m
.02756 .31054 L
s
.02381 .33099 m
.02756 .33099 L
s
.02381 .35144 m
.02756 .35144 L
s
.02381 .37189 m
.02756 .37189 L
s
.02381 .4128 m
.02756 .4128 L
s
.02381 .43325 m
.02756 .43325 L
s
.02381 .45371 m
.02756 .45371 L
s
.02381 .47416 m
.02756 .47416 L
s
.02381 .51507 m
.02756 .51507 L
s
.02381 .53552 m
.02756 .53552 L
s
.02381 .55597 m
.02756 .55597 L
s
.02381 .57643 m
.02756 .57643 L
s
.02381 .0651 m
.02756 .0651 L
s
.02381 .04464 m
.02756 .04464 L
s
.02381 .02419 m
.02756 .02419 L
s
.02381 .00374 m
.02756 .00374 L
s
.02381 .61733 m
.02756 .61733 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.02381 .59688 m
.02499 .59908 L
.02605 .60059 L
.02729 .60183 L
.02846 .60262 L
.02954 .60306 L
.03053 .60327 L
.03163 .60332 L
.03279 .6032 L
.03395 .60293 L
.0352 .60251 L
.03746 .60148 L
.04262 .59834 L
.06244 .58288 L
.10458 .54478 L
.1458 .49933 L
.1664 .47149 L
.18551 .44041 L
.204 .40242 L
.22371 .34635 L
.23399 .30645 L
.24512 .25214 L
.25473 .19768 L
.25983 .16886 L
.26529 .14088 L
.27068 .11832 L
.2756 .1031 L
.27797 .09758 L
.28054 .09283 L
.28298 .08944 L
.28523 .08714 L
.28643 .08621 L
.28773 .08543 L
.28896 .08488 L
.29008 .08452 L
.29133 .08429 L
.29203 .08423 L
.29269 .08421 L
.29339 .08423 L
.29402 .08429 L
.29475 .0844 L
.29544 .08455 L
.29676 .08492 L
.29815 .08544 L
.30065 .08668 L
.30629 .09066 L
.31573 .09999 L
.32587 .11261 L
.34418 .141 L
.35431 .16011 L
Mistroke
.36535 .18465 L
.3759 .21305 L
.38546 .24462 L
.40517 .33454 L
.4161 .39659 L
.4209 .42271 L
.42615 .44783 L
.43101 .46638 L
.43554 .47917 L
.43804 .48443 L
.44032 .4882 L
.44158 .4899 L
.44295 .49145 L
.44415 .49258 L
.44544 .49356 L
.44673 .49433 L
.44794 .49486 L
.44864 .49509 L
.44937 .49528 L
.45069 .49547 L
.45136 .49551 L
.45207 .4955 L
.45336 .49537 L
.45414 .49523 L
.45488 .49505 L
.45627 .4946 L
.45877 .49349 L
.46107 .49214 L
.46627 .48817 L
.47599 .47821 L
.48507 .46672 L
.50573 .43395 L
.51616 .41337 L
.52605 .39042 L
.5446 .33372 L
.55424 .29392 L
.56449 .24198 L
.57409 .18738 L
.5829 .14092 L
.58737 .12197 L
.59225 .10596 L
.5948 .09963 L
.59717 .09489 L
.59967 .09101 L
.60106 .08931 L
.60235 .08801 L
.60347 .08706 L
.60471 .08623 L
.60595 .08558 L
.60727 .08509 L
Mistroke
.60842 .08482 L
.60967 .08468 L
.61084 .08469 L
.61192 .08479 L
.61302 .085 L
.61405 .08528 L
.61518 .08567 L
.61638 .08618 L
.62062 .08863 L
.62598 .09294 L
.63091 .09781 L
.64205 .11119 L
.66222 .14234 L
.68106 .18083 L
.69175 .2088 L
.70173 .2409 L
.7221 .33249 L
.7311 .38341 L
.74066 .435 L
.74537 .45572 L
.74788 .46491 L
.75058 .4733 L
.75305 .47963 L
.75531 .48436 L
.75746 .48795 L
.75976 .4909 L
.76107 .49222 L
.76226 .49321 L
.76347 .49401 L
.76459 .49458 L
.76589 .49506 L
.76658 .49524 L
.76731 .49538 L
.76853 .49549 L
.76923 .49549 L
.76987 .49545 L
.77053 .49538 L
.77125 .49525 L
.77256 .49493 L
.77383 .49449 L
.77501 .494 L
.77768 .49256 L
.78052 .49063 L
.79084 .48097 L
.80199 .46724 L
.82218 .43549 L
.84217 .39338 L
.85223 .36598 L
.86325 .32865 L
.88281 .23629 L
Mistroke
.8937 .17442 L
.90375 .12554 L
.9087 .10838 L
.9112 .1017 L
.91396 .09582 L
.91671 .09136 L
.91824 .08945 L
.91968 .08798 L
.92102 .08688 L
.92248 .08596 L
.92312 .08564 L
.92381 .08535 L
.92504 .08496 L
.9257 .08483 L
.9264 .08473 L
.92766 .08467 L
.92835 .0847 L
.929 .08476 L
.92972 .08488 L
.93048 .08504 L
.93186 .08544 L
.93316 .08595 L
.9356 .08717 L
.94019 .09031 L
.94518 .09472 L
.96021 .11236 L
.97619 .13658 L
Mfstroke
0 0 1 r
.02381 .39235 m
.02499 .36699 L
.02605 .34773 L
.02721 .33047 L
.02846 .31519 L
.02961 .3037 L
.03068 .29496 L
.03279 .28189 L
.03396 .27645 L
.03506 .27228 L
.03636 .26825 L
.03756 .26529 L
.04 .26083 L
.04134 .25905 L
.04262 .25767 L
.045 .25574 L
.04633 .25492 L
.04753 .25429 L
.04991 .25331 L
.0521 .25261 L
.05466 .25195 L
.05743 .25137 L
.06244 .25048 L
.08253 .24731 L
.10458 .24312 L
.125 .23797 L
.14407 .2314 L
.15452 .2267 L
.16442 .2212 L
.17324 .21513 L
.18298 .20667 L
.19263 .19569 L
.20289 .17991 L
.21239 .15975 L
.22131 .13398 L
.23978 .0563 L
.24484 .03423 L
.24768 .02426 L
.24902 .02054 L
.25025 .01782 L
.25142 .01597 L
.25266 .01486 L
.2539 .01472 L
.2546 .0151 L
.25526 .01577 L
.25645 .0178 L
.25754 .0206 L
.2588 .02497 L
.25999 .03026 L
.2624 .04421 L
Mistroke
.26497 .06344 L
.26958 .10628 L
.28003 .20921 L
.28574 .25283 L
.29111 .28312 L
.29593 .30289 L
.30113 .31838 L
.30606 .32914 L
.31066 .33686 L
.32111 .34982 L
.34219 .37131 L
.35267 .38441 L
.36258 .40061 L
.37148 .42005 L
.38115 .44863 L
.39057 .48564 L
.40096 .53315 L
.40351 .54382 L
.40619 .55339 L
.40759 .55744 L
.40833 .55925 L
.40912 .56089 L
.40985 .56212 L
.41052 .563 L
.41181 .56395 L
.41312 .56383 L
.41433 .56268 L
.41496 .56166 L
.41565 .5602 L
.41705 .55615 L
.41828 .55133 L
.4196 .54487 L
.42198 .52999 L
.43157 .44174 L
.44177 .34778 L
.44708 .31245 L
.44991 .29766 L
.45296 .28438 L
.45829 .26652 L
.46331 .25438 L
.46814 .24556 L
.47342 .23797 L
.48407 .22608 L
.50313 .20632 L
.51338 .1928 L
.52271 .17673 L
.53348 .15125 L
.54365 .11756 L
.55336 .07585 L
.55867 .05127 L
Mistroke
.56127 .04001 L
.56364 .03084 L
.56567 .02435 L
.56673 .02155 L
.56787 .01914 L
.56908 .01731 L
.57021 .01637 L
.57137 .01622 L
.57242 .01686 L
.57367 .01865 L
.57481 .02131 L
.57607 .02538 L
.5774 .031 L
.57964 .04346 L
.58209 .06109 L
.59296 .16599 L
.5983 .21537 L
.60325 .25242 L
.60773 .27839 L
.61249 .29914 L
.61726 .31451 L
.62253 .32694 L
.62758 .33584 L
.6322 .34235 L
.64262 .35397 L
.66146 .37348 L
.67127 .38624 L
.6804 .40156 L
.69041 .42422 L
.7011 .45808 L
.71136 .5013 L
.71675 .52633 L
.71971 .53928 L
.72247 .54993 L
.72488 .55734 L
.72623 .56046 L
.72746 .56251 L
.72816 .5633 L
.72881 .56377 L
.72952 .56399 L
.73028 .56386 L
.731 .56337 L
.73168 .56256 L
.73296 .56016 L
.73413 .55684 L
.73538 .55217 L
.7376 .54089 L
.74 .52479 L
.74262 .50314 L
.75352 .39559 L
Mistroke
.75888 .34932 L
.76382 .31574 L
.76831 .29258 L
.77307 .27422 L
.77785 .26062 L
.78312 .24953 L
.78818 .24142 L
.79281 .23538 L
.80324 .22421 L
.8221 .20433 L
.83194 .19089 L
.84108 .17459 L
.8511 .15034 L
.8618 .11428 L
.87207 .06926 L
.87748 .04459 L
.88044 .03267 L
.88188 .02765 L
.8832 .0237 L
.88435 .02081 L
.88561 .01839 L
.8863 .01741 L
.88694 .01675 L
.8882 .01617 L
.88947 .01664 L
.89019 .0174 L
.89087 .01846 L
.89158 .01994 L
.89223 .02164 L
.8937 .02664 L
.89595 .03751 L
.89841 .05359 L
.90337 .09655 L
.91345 .19647 L
.9184 .23743 L
.92297 .26729 L
.92823 .29304 L
.93394 .31276 L
.93878 .32476 L
.94391 .33432 L
.95195 .34536 L
.96059 .35465 L
.97619 .37057 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`008@3oool00`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`800000o`3ooon>0?ooo`00
8@3oool00`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000003oool0oooo0?l0ooooS@3o
ool00240oooo00<000000?ooo`3oool0o`3ooon=0?ooo`008@3oool00`000000oooo0?ooo`3o0?oo
ohd0oooo000Q0?ooo`030000003oool0oooo05T0oooo0P000?mo0?ooo`030000o`3oool0oooo07h0
oooo00<0003o0?ooo`3oool0;P3oool00240oooo00<000000?ooo`3oool0F03oool010000?l0oooo
0?ooo`000?mn0?ooo`80003oOP3oool00`000?l0oooo0000o`0_0?ooo`008@3oool00`000000oooo
0?ooo`1H0?ooo`040000o`3oool0oooo0000ogd0oooo00@0003o0?ooo`3oool0003oO03oool01000
0?l0oooo0?ooo`000?l_0?ooo`008@3oool00`000000oooo0?ooo`1G0?ooo`050000o`3oool0oooo
0?ooo`000?l0O03oool01@000?l0oooo0?ooo`3oool0003o07`0oooo00@0003o0?ooo`3oool0003o
;`3oool00240oooo0P00001H0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0N@3o
ool01@000?l0oooo0?ooo`3oool0003o07/0oooo00D0003o0?ooo`3oool0oooo0000o`0_0?ooo`00
8@3oool00`000000oooo0?ooo`1G0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0
N03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo07T0oooo00<0003o0?ooo`3oool0
0P3oool00`000?l0oooo0?ooo`0/0?ooo`008@3oool00`000000oooo0?ooo`1F0?ooo`030000o`3o
ool0oooo00<0oooo00<0003o0?ooo`3oool0N03oool00`000?l0oooo0?ooo`030?ooo`030000o`3o
ool0oooo07P0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0/0?ooo`008@3oool0
0`000000oooo0?ooo`1F0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0N03oool0
0`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo07P0oooo00<0003o0?ooo`3oool00P3oool0
0`000?l0oooo0?ooo`0/0?ooo`008@3oool00`000000oooo0?ooo`1F0?ooo`030000o`3oool0oooo
00<0oooo00<0003o0?ooo`3oool0M`3oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo
07P0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0[0?ooo`008@3oool00`000000
oooo0?ooo`1F0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0M`3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3oool0oooo07L0oooo00<0003o0?ooo`3oool00`3oool00`000?l0
oooo0?ooo`0[0?ooo`008@3oool00`000000oooo0?ooo`1F0?ooo`030000o`3oool0oooo00@0oooo
00<0003o0?ooo`3oool0MP3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0oooo07H0oooo
00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`0[0?ooo`008@3oool00`000000oooo0?oo
o`1E0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0MP3oool00`000?l0oooo0?oo
o`050?ooo`030000o`3oool0oooo07H0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?oo
o`0[0?ooo`008@3oool2000005H0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?ooo`1f
0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0MP3oool00`000?l0oooo0?ooo`04
0?ooo`030000o`3oool0oooo02/0oooo000Q0?ooo`030000003oool0oooo05D0oooo00<0003o0?oo
o`3oool01@3oool00`000?l0oooo0?ooo`1e0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?oo
o`3oool0MP3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0oooo02X0oooo000Q0?ooo`03
0000003oool0oooo05D0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?ooo`1e0?ooo`03
0000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool0M@3oool00`000?l0oooo0?ooo`060?ooo`03
0000o`3oool0oooo02X0oooo000Q0?ooo`030000003oool0oooo05@0oooo00<0003o0?ooo`3oool0
1P3oool00`000?l0oooo0?ooo`1e0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool0
M@3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo02X0oooo000Q0?ooo`030000003o
ool0oooo05@0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`1d0?ooo`030000o`3o
ool0oooo00L0oooo00<0003o0?ooo`3oool0M03oool00`000?l0oooo0?ooo`060?ooo`030000o`3o
ool0oooo02X0oooo000Q0?ooo`030000003oool0oooo05@0oooo00<0003o0?ooo`3oool01`3oool0
0`000?l0oooo0?ooo`1d0?ooo`030000o`3oool0oooo00L0oooo00<0003o0?ooo`3oool0L`3oool0
0`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo02X0oooo000Q0?ooo`030000003oool0oooo
05@0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`1c0?ooo`030000o`3oool0oooo
00P0oooo00<0003o0?ooo`3oool0L`3oool00`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo
02X0oooo000Q0?ooo`030000003oool0oooo05@0oooo00<0003o0?ooo`3oool01`3oool00`000?l0
oooo0?ooo`1c0?ooo`030000o`3oool0oooo00P0oooo00<0003o0?ooo`3oool0L`3oool00`000?l0
oooo0?ooo`070?ooo`030000o`3oool0oooo02X0oooo000Q0?ooo`800000E03oool00`000?l0oooo
0?ooo`080?ooo`030000o`3oool0oooo07<0oooo00<0003o0?ooo`3oool0203oool00`000?l0oooo
0?ooo`1c0?ooo`030000o`3oool0oooo00P0oooo00<0003o0?ooo`3oool0:@3oool00240oooo00<0
00000?ooo`3oool0D`3oool00`000?l0oooo0?ooo`080?ooo`030000o`3oool0oooo07<0oooo00<0
003o0?ooo`3oool0203oool00`000?l0oooo0?ooo`1b0?ooo`030000o`3oool0oooo00T0oooo00<0
003o0?ooo`3oool0:@3oool00240oooo00<000000?ooo`3oool0D`3oool00`000?l0oooo0?ooo`08
0?ooo`030000o`3oool0oooo0780oooo00<0003o0?ooo`3oool02@3oool00`000?l0oooo0?ooo`1b
0?ooo`030000o`3oool0oooo00T0oooo00<0003o0?ooo`3oool0:@3oool00100oooo00<0gia80000
000000000`0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo05<0oooo00<0003o0?oo
o`3oool0203oool00`000?l0oooo0?ooo`1b0?ooo`030000o`3oool0oooo00T0oooo00<0003o0?oo
o`3oool0LP3oool00`000?l0oooo0?ooo`090?ooo`030000o`3oool0oooo02T0oooo000A0?ooo`03
0?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool0203oool00`000000oooo0?ooo`1B0?ooo`03
0000o`3oool0oooo00X0oooo00<0003o0?ooo`3oool0L@3oool00`000?l0oooo0?ooo`090?ooo`03
0000o`3oool0oooo0780oooo00<0003o0?ooo`3oool02@3oool00`000?l0oooo0?ooo`0Y0?ooo`00
4P3oool00`3o_g@007Bo0?ooo`0<0?ooo`030000003oool0oooo0580oooo00<0003o0?ooo`3oool0
2P3oool00`000?l0oooo0?ooo`1a0?ooo`030000o`3oool0oooo00X0oooo00<0003o0?ooo`3oool0
L03oool00`000?l0oooo0?ooo`0:0?ooo`030000o`3oool0oooo02T0oooo000C0?ooo`030?noM000
M;l0oooo00/0oooo00<000000?ooo`3oool0DP3oool00`000?l0oooo0?ooo`0:0?ooo`030000o`3o
ool0oooo0740oooo00<0003o0?ooo`3oool02P3oool00`000?l0oooo0?ooo`1`0?ooo`030000o`3o
ool0oooo00X0oooo00<0003o0?ooo`3oool0:@3oool000<0oooo00<0oooO09a8000000001`000000
0`18W=l0oooo0?ooo`040?ooo`030?noM018W=l0oooo00X0oooo00<000000?ooo`3oool0DP3oool0
0`000?l0oooo0?ooo`0:0?ooo`030000o`3oool0oooo0700oooo00<0003o0?ooo`3oool02`3oool0
0`000?l0oooo0?ooo`1`0?ooo`030000o`3oool0oooo00X0oooo00<0003o0?ooo`3oool0:@3oool0
01D0oooo00<0gia804RLg`3oool02@3oool300000580oooo00<0003o0?ooo`3oool02P3oool00`00
0?l0oooo0?ooo`060?ooo`D0o`00I@3oool00`000?l0oooo0?ooo`0;0?ooo`030000o`3oool0oooo
00D0oooo1@3o001V0?ooo`030000o`3oool0oooo00X0oooo00<0003o0?ooo`3oool01P3oool50?l0
01h0oooo000F0?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0D@3oool00`000?l0
oooo0?ooo`0;0?ooo`030000o`3oool0oooo00D0oooo00<0o`000?ooo`3oool00`3oool00`3o0000
oooo0?ooo`1R0?ooo`030000o`3oool0oooo00/0oooo00<0003o0?ooo`3oool0103oool20?l000@0
oooo00<0o`000?ooo`3oool0HP3oool00`000?l0oooo0?ooo`0<0?ooo`030000o`3oool0oooo00@0
oooo0P3o00030?ooo`80o`007@3oool001H0oooo00<0omnL04P0M02oool0203oool00`000000oooo
0?ooo`1A0?ooo`030000o`3oool0oooo00/0oooo00<0003o0?ooo`3oool0103oool00`3o0000oooo
0?ooo`050?ooo`80o`00HP3oool00`000?l0oooo0?ooo`0;0?ooo`030000o`3oool0oooo00<0oooo
00<0o`000?ooo`3oool01@3oool20?l00680oooo00<0003o0?ooo`3oool0303oool00`000?l0oooo
0?ooo`030?ooo`030?l0003oool0oooo00D0oooo0P3o000K0?ooo`005P3oool00`3ooml0W4P007Bo
o`080?ooo`030000003oool0oooo0540oooo00<0003o0?ooo`3oool02`3oool00`000?l0oooo0?oo
o`030?ooo`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool0GP3oool00`000?l0oooo0?oo
o`0<0?ooo`030000o`3oool0oooo00<0oooo00<0o`000?ooo`3oool01`3oool20?l00600oooo00<0
003o0?ooo`3oool0303oool00`000?l0oooo0?ooo`020?ooo`030?l0003oool0oooo00P0oooo00<0
o`000?ooo`3oool0603oool001H0oooo00<0oono07@00018W=l0203oool00`000000oooo0?ooo`1A
0?ooo`030000o`3oool0oooo00/0oooo00<0003o0?ooo`3oool00P3oool00`3o0000oooo0?ooo`0:
0?ooo`030?l0003oool0oooo05d0oooo00<0003o0?ooo`3oool0303oool00`000?l0oooo0?ooo`02
0?ooo`030?l0003oool0oooo00X0oooo00<0o`000?ooo`3oool0G@3oool00`000?l0oooo0?ooo`0<
0?ooo`050000o`3oool0oooo0?ooo`3o0000303oool00`3o0000oooo0?ooo`0G0?ooo`004P3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0203oool00`000000oooo0?ooo`1@0?ooo`030000
o`3oool0oooo00d0oooo00D0003o0?ooo`3oool0oooo0?l0000=0?ooo`030?l0003oool0oooo05`0
oooo00<0003o0?ooo`3oool0303oool01@000?l0oooo0?ooo`3oool0o`0000h0oooo00<0o`000?oo
o`3oool0G03oool00`000?l0oooo0?ooo`0<0?ooo`050000o`3oool0oooo0?ooo`3o00003@3oool0
0`3o0000oooo0?ooo`0F0?ooo`004`3oool01@3ogi`0B000000000000000M;oo00T0oooo00<00000
0?ooo`3oool0D03oool00`000?l0oooo0?ooo`0=0?ooo`050000o`3oool0oooo0?ooo`3o00003P3o
ool00`3o0000oooo0?ooo`1K0?ooo`030000o`3oool0oooo00`0oooo00D0003o0?ooo`3oool0oooo
0?l0000?0?ooo`030?l0003oool0oooo05X0oooo00<0003o0?ooo`3oool03@3oool010000?l0oooo
0?ooo`3o000?0?ooo`030?l0003oool0oooo01D0oooo000Q0?ooo`030000003oool0oooo0500oooo
00<0003o0?ooo`3oool03@3oool010000?l0oooo0?ooo`3o000@0?ooo`030?l0003oool0oooo05T0
oooo00<0003o0?ooo`3oool03P3oool010000?l0oooo0?ooo`3o000?0?ooo`030?l0003oool0oooo
05X0oooo00<0003o0?ooo`3oool03@3oool010000?l0oooo0?ooo`3o000@0?ooo`030?l0003oool0
oooo01@0oooo000Q0?ooo`030000003oool0oooo0500oooo00<0003o0?ooo`3oool03@3oool01000
0?l0oooo0?ooo`3o000A0?ooo`030?l0003oool0oooo05P0oooo00<0003o0?ooo`3oool03P3oool0
0`000?l0oooo0?l0000A0?ooo`030?l0003oool0oooo05T0oooo00<0003o0?ooo`3oool03@3oool0
10000?l0oooo0?ooo`3o000@0?ooo`030?l0003oool0oooo01@0oooo000Q0?ooo`800000D03oool0
0`000?l0oooo0?ooo`0>0?ooo`040000o`3oool0oooo0?l00140oooo00<0o`000?ooo`3oool0F03o
ool00`000?l0oooo0?ooo`0>0?ooo`030000o`3oool0o`000180oooo00<0o`000?ooo`3oool0F03o
ool00`000?l0oooo0?ooo`0=0?ooo`030000o`3oool0o`000180oooo00<0o`000?ooo`3oool04`3o
ool00240oooo00<000000?ooo`3oool0C`3oool00`000?l0oooo0?ooo`0>0?ooo`030000o`3oool0
o`0001<0oooo00<0o`000?ooo`3oool0E`3oool00`000?l0oooo0?ooo`0>0?ooo`030000o`3oool0
o`000180oooo00<0o`000?ooo`3oool0E`3oool00`000?l0oooo0?ooo`0>0?ooo`030000o`3oool0
o`0001<0oooo00<0o`000?ooo`3oool04P3oool00240oooo00<000000?ooo`3oool0C`3oool00`00
0?l0oooo0?ooo`0>0?ooo`030000o`3oool0o`0001@0oooo00<0o`000?ooo`3oool0E@3oool00`00
0?l0oooo0?ooo`0?0?ooo`030000o`3o0000oooo01<0oooo00<0o`000?ooo`3oool0EP3oool00`00
0?l0oooo0?ooo`0?0?ooo`030000o`3o0000oooo01<0oooo00<0o`000?ooo`3oool04@3oool00240
oooo00<000000?ooo`3oool0C`3oool00`000?l0oooo0?ooo`0>0?ooo`030000o`3oool0o`0001D0
oooo00<0o`000?ooo`3oool0E03oool00`000?l0oooo0?ooo`0?0?ooo`030000o`3o0000oooo01@0
oooo00<0o`000?ooo`3oool0E@3oool00`000?l0oooo0?ooo`0?0?ooo`030000o`3oool0oooo01@0
oooo00<0o`000?ooo`3oool0403oool00240oooo00<000000?ooo`3oool0C`3oool00`000?l0oooo
0?ooo`0>0?ooo`030000o`3o0000oooo01D0oooo00<0o`000?ooo`3oool0E03oool00`000?l0oooo
0?ooo`0?0?ooo`030000o`3o0000oooo01@0oooo00<0o`000?ooo`3oool0E@3oool00`000?l0oooo
0?ooo`0?0?ooo`030000o`3oool0oooo01@0oooo00<0o`000?ooo`3oool0403oool00240oooo00<0
00000?ooo`3oool0CP3oool00`000?l0oooo0?ooo`0?0?ooo`030000o`3o0000oooo01H0oooo00<0
o`000?ooo`3oool0D`3oool00`000?l0oooo0?ooo`0?0?ooo`030000o`3oool0oooo01D0oooo00<0
o`000?ooo`3oool0D`3oool00`000?l0oooo0?ooo`0@0?ooo`030000o`3oool0oooo01D0oooo00<0
o`000?ooo`3oool03`3oool00240oooo00<000000?ooo`3oool0CP3oool00`000?l0oooo0?ooo`0@
0?ooo`030000o`3oool0oooo01D0oooo00<0o`000?ooo`3oool0DP3oool00`000?l0oooo0?ooo`0@
0?ooo`030000o`3oool0oooo01H0oooo00<0o`000?ooo`3oool0DP3oool00`000?l0oooo0?ooo`0@
0?ooo`030000o`3oool0oooo01D0oooo00<0o`000?ooo`3oool03`3oool00240oooo00<000000?oo
o`3oool0CP3oool00`000?l0oooo0?ooo`0?0?ooo`030?l000000?l0oooo01L0oooo00<0o`000?oo
o`3oool0D@3oool00`000?l0oooo0?ooo`0@0?ooo`030000o`3oool0oooo01H0oooo00<0o`000?oo
o`3oool0DP3oool00`000?l0oooo0?ooo`0?0?ooo`030?l000000?l0oooo01L0oooo00<0o`000?oo
o`3oool03P3oool00240oooo0P00001?0?ooo`030000o`3oool0oooo00l0oooo00<0o`000000o`3o
ool05`3oool00`3o0000oooo0?ooo`1A0?ooo`030000o`3oool0oooo00l0oooo00<0o`000000o`3o
ool0603oool00`3o0000oooo0?ooo`1A0?ooo`030000o`3oool0oooo00l0oooo00<0o`000000o`3o
ool0603oool00`3o0000oooo0?ooo`0=0?ooo`008@3oool00`000000oooo0?ooo`1=0?ooo`030000
o`3oool0oooo0100oooo00<0o`000000o`3oool0603oool00`3o0000oooo0?ooo`1@0?ooo`030000
o`3oool0oooo00l0oooo00<0o`000?ooo`000?l06@3oool00`3o0000oooo0?ooo`1?0?ooo`030000
o`3oool0oooo0100oooo00<0o`000000o`3oool0603oool00`3o0000oooo0?ooo`0=0?ooo`008@3o
ool00`000000oooo0?ooo`1=0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`000?l06P3o
ool00`3o0000oooo0?ooo`1>0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`000?l06@3o
ool00`3o0000oooo0?ooo`1?0?ooo`030000o`3oool0oooo0100oooo00<0o`000000o`3oool06@3o
ool00`3o0000oooo0?ooo`0<0?ooo`008@3oool00`000000oooo0?ooo`1=0?ooo`030000o`3oool0
oooo00l0oooo00<0o`000?ooo`000?l06P3oool00`3o0000oooo0?ooo`1>0?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`000?l06P3oool00`3o0000oooo0?ooo`1>0?ooo`030000o`3oool0
oooo00l0oooo00<0o`000?ooo`000?l06P3oool00`3o0000oooo0?ooo`0<0?ooo`008@3oool00`00
0000oooo0?ooo`1=0?ooo`030000o`3oool0oooo00l0oooo00<0o`000?ooo`000?l06`3oool00`3o
0000oooo0?ooo`1=0?ooo`030000o`3oool0oooo00l0oooo00@0o`000?ooo`3oool0003o6`3oool0
0`3o0000oooo0?ooo`1=0?ooo`030000o`3oool0oooo00l0oooo00@0o`000?ooo`3oool0003o6P3o
ool00`3o0000oooo0?ooo`0;0?ooo`008@3oool00`000000oooo0?ooo`1<0?ooo`030000o`3oool0
oooo00l0oooo00@0o`000?ooo`3oool0003o6`3oool00`3o0000oooo0?ooo`1<0?ooo`030000o`3o
ool0oooo0100oooo00@0o`000?ooo`3oool0003o6`3oool00`3o0000oooo0?ooo`1<0?ooo`030000
o`3oool0oooo0100oooo00@0o`000?ooo`3oool0003o:03oool00240oooo00<000000?ooo`3oool0
C03oool00`000?l0oooo0?ooo`0?0?ooo`040?l0003oool0oooo0000oa`0oooo00<0o`000?ooo`3o
ool0B`3oool00`000?l0oooo0?ooo`0@0?ooo`040?l0003oool0oooo0000oa`0oooo00<0o`000?oo
o`3oool0B`3oool00`000?l0oooo0?ooo`0@0?ooo`040?l0003oool0oooo0000obP0oooo000Q0?oo
o`030000003oool0oooo04/0oooo00<0003o0?ooo`3oool0403oool0103o0000oooo0?ooo`000?lM
0?ooo`030?l0003oool0oooo04X0oooo00<0003o0?ooo`3oool0403oool0103o0000oooo0?ooo`00
0?lL0?ooo`030?l0003oool0oooo04/0oooo00<0003o0?ooo`3oool0403oool0103o0000oooo0?oo
o`000?lX0?ooo`008@3oool2000004`0oooo00<0003o0?ooo`3oool0403oool01@3o0000oooo0?oo
o`3oool0003o01`0oooo00<0o`000?ooo`3oool0BP3oool00`000?l0oooo0?ooo`0@0?ooo`040?l0
003oool0oooo0000oad0oooo00<0o`000?ooo`3oool0BP3oool00`000?l0oooo0?ooo`0?0?ooo`05
0?l0003oool0oooo0?ooo`000?l0:03oool00240oooo00<000000?ooo`3oool0B`3oool00`000?l0
oooo0?ooo`0?0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool06`3oool00`3o0000
oooo0?ooo`180?ooo`030000o`3oool0oooo0140oooo00@0o`000?ooo`3oool0003o7@3oool00`3o
0000oooo0?ooo`190?ooo`030000o`3oool0oooo0100oooo00D0o`000?ooo`3oool0oooo0000o`0X
0?ooo`008@3oool00`000000oooo0?ooo`1:0?ooo`030000o`3oool0oooo0100oooo00<0o`000?oo
o`3oool00P3oool00`000?l0oooo0?ooo`0K0?ooo`030?l0003oool0oooo04P0oooo00<0003o0?oo
o`3oool0403oool01@3o0000oooo0?ooo`3oool0003o01h0oooo00<0o`000?ooo`3oool0B03oool0
0`000?l0oooo0?ooo`0@0?ooo`050?l0003oool0oooo0?ooo`000?l0:03oool00240oooo00<00000
0?ooo`3oool0BP3oool00`000?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo0080oooo00<0003o
0?ooo`3oool0703oool00`3o0000oooo0?ooo`170?ooo`030000o`3oool0oooo0100oooo00<0o`00
0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0K0?ooo`030?l0003oool0oooo04P0oooo00<0003o
0?ooo`3oool0403oool01@3o0000oooo0?ooo`3oool0003o02P0oooo000Q0?ooo`030000003oool0
oooo04T0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0
oooo01d0oooo00<0o`000?ooo`3oool0A@3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0
oooo0080oooo00<0003o0?ooo`3oool0703oool00`3o0000oooo0?ooo`160?ooo`030000o`3oool0
oooo0100oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0V0?ooo`008@3oool00`00
0000oooo0?ooo`190?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`0M0?ooo`030?l0003oool0oooo04D0oooo00<0003o0?ooo`3oool04@3oool00`3o
0000oooo0?ooo`020?ooo`030000o`3oool0oooo01`0oooo00<0o`000?ooo`3oool0AP3oool00`00
0?l0oooo0?ooo`0@0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool09P3oool00240
oooo00<000000?ooo`3oool0B@3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo0080
oooo00<0003o0?ooo`3oool07P3oool00`3o0000oooo0?ooo`130?ooo`030000o`3oool0oooo0180
oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0M0?ooo`030?l0003oool0oooo04@0
oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo02D0
oooo000Q0?ooo`030000003oool0oooo04P0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`030?ooo`030000o`3oool0oooo01h0oooo00<0o`000?ooo`3oool0@`3oool00`000?l0oooo
0?ooo`0B0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool07@3oool00`3o0000oooo
0?ooo`140?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo
0?ooo`0U0?ooo`008@3oool00`000000oooo0?ooo`180?ooo`030000o`3oool0oooo0140oooo00<0
o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0O0?ooo`030?l0003oool0oooo0480oooo00<0
003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo01h0oooo00<0
o`000?ooo`3oool0@P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00<0oooo00<0
003o0?ooo`3oool09@3oool00240oooo0P0000180?ooo`030000o`3oool0oooo0180oooo00<0o`00
0?ooo`3oool00`3oool00`000?l0oooo0?ooo`0O0?ooo`030?l0003oool0oooo0440oooo00<0003o
0?ooo`3oool04P3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo01h0oooo00<0o`00
0?ooo`3oool0@P3oool00`000?l0oooo0?ooo`0A0?ooo`030?l0003oool0oooo00@0oooo00<0003o
0?ooo`3oool09@3oool00240oooo00<000000?ooo`3oool0A`3oool00`000?l0oooo0?ooo`0B0?oo
o`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0803oool00`3o0000oooo0?ooo`100?oo
o`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0O0?oo
o`030?l0003oool0oooo0400oooo00<0003o0?ooo`3oool04P3oool00`3o0000oooo0?ooo`040?oo
o`030000o`3oool0oooo02D0oooo000Q0?ooo`030000003oool0oooo04H0oooo00<0003o0?ooo`3o
ool04`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo01l0oooo00<0o`000?ooo`3o
ool0?`3oool00`000?l0oooo0?ooo`0C0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3o
ool07`3oool00`3o0000oooo0?ooo`100?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3o
ool0103oool00`000?l0oooo0?ooo`0U0?ooo`00403oool00`3ookl0M000000000020000000304RL
g`3oool0oooo00T0oooo00<000000?ooo`3oool0AP3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0
003oool0oooo00D0oooo00<0003o0?ooo`3oool07`3oool00`3o0000oooo0?ooo`0o0?ooo`030000
o`3oool0oooo01<0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0P0?ooo`030?l0
003oool0oooo03h0oooo00<0003o0?ooo`3oool04`3oool00`3o0000oooo0?ooo`040?ooo`030000
o`3oool0oooo02D0oooo000B0?ooo`030?noM0000000M;oo00`0oooo00<000000?ooo`3oool0A@3o
ool00`000?l0oooo0?ooo`0C0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool0803o
ool00`3o0000oooo0?ooo`0m0?ooo`030000o`3oool0oooo01@0oooo00<0o`000?ooo`3oool00`3o
ool00`000?l0oooo0?ooo`0P0?ooo`030?l0003oool0oooo03h0oooo00<0003o0?ooo`3oool04`3o
ool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo02D0oooo000B0?ooo`030?oog`2LB000
07Bo00`0oooo00<000000?ooo`3oool0A@3oool00`000?l0oooo0?ooo`0C0?ooo`030?l0003oool0
oooo00D0oooo00<0003o0?ooo`3oool0803oool00`3o0000oooo0?ooo`0m0?ooo`030000o`3oool0
oooo01<0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0P0?ooo`030?l0003oool0
oooo03`0oooo00<0003o0?ooo`3oool0503oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0
oooo02D0oooo000C0?ooo`030=nLB000B9`0gooo00/0oooo00<000000?ooo`3oool0A03oool00`00
0?l0oooo0?ooo`0D0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool08@3oool00`3o
0000oooo0?ooo`0k0?ooo`030000o`3oool0oooo01@0oooo00<0o`000?ooo`3oool01@3oool00`00
0?l0oooo0?ooo`0P0?ooo`030?l0003oool0oooo03`0oooo00<0003o0?ooo`3oool04`3oool00`3o
0000oooo0?ooo`050?ooo`030000o`3oool0oooo02D0oooo00030?ooo`030?oog`2LB000000000L0
000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0oooo04@0
oooo00<0003o0?ooo`3oool0503oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo0240
oooo00<0o`000?ooo`3oool0>`3oool00`000?l0oooo0?ooo`0D0?ooo`030?l0003oool0oooo00D0
oooo00<0003o0?ooo`3oool08@3oool00`3o0000oooo0?ooo`0j0?ooo`030000o`3oool0oooo01@0
oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0T0?ooo`004`3oool00`3ookl0M000
07Boo`0;0?ooo`<00000@`3oool00`000?l0oooo0?ooo`0E0?ooo`030?l0003oool0oooo00D0oooo
00<0003o0?ooo`3oool08@3oool00`3o0000oooo0?ooo`0j0?ooo`030000o`3oool0oooo01D0oooo
00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0Q0?ooo`030?l0003oool0oooo03X0oooo
00<0003o0?ooo`3oool0503oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo02@0oooo
000D0?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3oool0@P3oool00`000?l0oooo0?oo
o`0E0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool08P3oool00`3o0000oooo0?oo
o`0i0?ooo`030000o`3oool0oooo01D0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?oo
o`0Q0?ooo`030?l0003oool0oooo03T0oooo00<0003o0?ooo`3oool05@3oool00`3o0000oooo0?oo
o`060?ooo`030000o`3oool0oooo02@0oooo000D0?ooo`030=nLB000B9`0gooo00X0oooo00<00000
0?ooo`3oool0@P3oool00`000?l0oooo0?ooo`0E0?ooo`030?l0003oool0oooo00H0oooo00<0003o
0?ooo`3oool08P3oool00`3o0000oooo0?ooo`0h0?ooo`030000o`3oool0oooo01H0oooo00<0o`00
0?ooo`3oool01@3oool00`000?l0oooo0?ooo`0R0?ooo`030?l0003oool0oooo03L0oooo00<0003o
0?ooo`3oool05P3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo02@0oooo000D0?oo
o`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool0@@3oool00`000?l0oooo0?ooo`0F0?oo
o`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool08P3oool00`3o0000oooo0?ooo`0g0?oo
o`030000o`3oool0oooo01H0oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0R0?oo
o`030?l0003oool0oooo03L0oooo00<0003o0?ooo`3oool05P3oool00`3o0000oooo0?ooo`060?oo
o`030000o`3oool0oooo02@0oooo000D0?ooo`030?oo_`1d0000M;oo00X0oooo00<000000?ooo`3o
ool0@03oool00`000?l0oooo0?ooo`0G0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3o
ool08P3oool00`3o0000oooo0?ooo`0e0?ooo`030000o`3oool0oooo01L0oooo00<0o`000?ooo`3o
ool01P3oool00`000?l0oooo0?ooo`0S0?ooo`030?l0003oool0oooo03D0oooo00<0003o0?ooo`3o
ool05P3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo02@0oooo000E0?ooo`030;md
0000M;l0oooo00T0oooo00<000000?ooo`3oool0?`3oool00`000?l0oooo0?ooo`0H0?ooo`030?l0
003oool0oooo00L0oooo00<0003o0?ooo`3oool08P3oool00`3o0000oooo0?ooo`0e0?ooo`030000
o`3oool0oooo01L0oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo0?ooo`0S0?ooo`030?l0
003oool0oooo03@0oooo00<0003o0?ooo`3oool05`3oool00`3o0000oooo0?ooo`070?ooo`030000
o`3oool0oooo02@0oooo000C0?ooo`050?oo_`1d000000000000M02oool02@3oool00`000000oooo
0?ooo`0n0?ooo`030000o`3oool0oooo01T0oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo
0?ooo`0S0?ooo`030?l0003oool0oooo03<0oooo00<0003o0?ooo`3oool0603oool00`3o0000oooo
0?ooo`060?ooo`030000o`3oool0oooo02<0oooo00<0o`000?ooo`3oool0=03oool00`000?l0oooo
0?ooo`0G0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool0903oool00240oooo00<0
00000?ooo`3oool0?@3oool00`000?l0oooo0?ooo`0I0?ooo`030?l0003oool0oooo00P0oooo00<0
003o0?ooo`3oool08`3oool00`3o0000oooo0?ooo`0b0?ooo`030000o`3oool0oooo01T0oooo00<0
o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0S0?ooo`030?l0003oool0oooo0380oooo00<0
003o0?ooo`3oool0603oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo02@0oooo000Q
0?ooo`800000?@3oool00`000?l0oooo0?ooo`0J0?ooo`030?l0003oool0oooo00P0oooo00<0003o
0?ooo`3oool08`3oool00`3o0000oooo0?ooo`0a0?ooo`030000o`3oool0oooo01X0oooo00<0o`00
0?ooo`3oool01`3oool00`000?l0oooo0?ooo`0S0?ooo`030?l0003oool0oooo0340oooo00<0003o
0?ooo`3oool06@3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo02<0oooo000Q0?oo
o`030000003oool0oooo03/0oooo00<0003o0?ooo`3oool06`3oool00`3o0000oooo0?ooo`080?oo
o`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool0;`3oool00`000?l0oooo0?ooo`0J0?oo
o`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool08`3oool00`3o0000oooo0?ooo`0`0?oo
o`030000o`3oool0oooo01X0oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0S0?oo
o`008@3oool00`000000oooo0?ooo`0j0?ooo`030000o`3oool0oooo01`0oooo00<0o`000?ooo`3o
ool0203oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003oool0oooo02h0oooo00<0003o0?ooo`3o
ool06`3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3o
ool0;P3oool00`000?l0oooo0?ooo`0J0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3o
ool08`3oool00240oooo00<000000?ooo`3oool0>@3oool00`000?l0oooo0?ooo`0M0?ooo`030?l0
003oool0oooo00P0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo0?ooo`0]0?ooo`030000
o`3oool0oooo01`0oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0T0?ooo`030?l0
003oool0oooo02d0oooo00<0003o0?ooo`3oool06`3oool00`3o0000oooo0?ooo`090?ooo`030000
o`3oool0oooo02<0oooo000Q0?ooo`030000003oool0oooo03L0oooo0P000?lP0?ooo`030?l0003o
ool0oooo00T0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo0?ooo`0[0?ooo`030000o`3o
ool0oooo01d0oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003o
ool0oooo02`0oooo00<0003o0?ooo`3oool0703oool00`3o0000oooo0?ooo`090?ooo`030000o`3o
ool0oooo02<0oooo000Q0?ooo`030000003oool0oooo03H0oooo00<0003o0?ooo`3oool07`3oool0
0`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool0:P3oool0
0`000?l0oooo0?ooo`0N0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool09@3oool0
0`3o0000oooo0?ooo`0Z0?ooo`030000o`3oool0oooo01d0oooo00<0o`000?ooo`3oool02@3oool0
0`000?l0oooo0?ooo`0S0?ooo`008@3oool00`000000oooo0?ooo`0e0?ooo`030000o`3oool0oooo
0200oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003oool0oooo
02T0oooo00<0003o0?ooo`3oool07P3oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo
02D0oooo00<0o`000?ooo`3oool0:@3oool00`000?l0oooo0?ooo`0N0?ooo`030?l0003oool0oooo
00T0oooo00<0003o0?ooo`3oool08`3oool00240oooo00<000000?ooo`3oool0<`3oool20000ob<0
oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0U0?ooo`030?l0003oool0oooo02L0
oooo00<0003o0?ooo`3oool07`3oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo02@0
oooo00<0o`000?ooo`3oool0:03oool00`000?l0oooo0?ooo`0O0?ooo`030?l0003oool0oooo00T0
oooo00<0003o0?ooo`3oool08`3oool00240oooo0P00000b0?ooo`80003o9@3oool00`3o0000oooo
0?ooo`0:0?ooo`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3oool09P3oool00`000?l0oooo
0?ooo`0P0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo
0?ooo`0W0?ooo`030000o`3oool0oooo01l0oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo
0?ooo`0R0?ooo`008@3oool00`000000oooo0?ooo`0_0?ooo`80003o9`3oool00`3o0000oooo0?oo
o`0:0?ooo`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3oool09@3oool00`000?l0oooo0?oo
o`0Q0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?oo
o`0U0?ooo`030000o`3oool0oooo0200oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?oo
o`0R0?ooo`008@3oool00`000000oooo0?ooo`0/0?ooo`<0003o:03oool00`3o0000oooo0?ooo`0<
0?ooo`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool09@3oool00`000?l0oooo0?ooo`0Q
0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?ooo`0T
0?ooo`030000o`3oool0oooo0240oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0R
0?ooo`008@3oool00`000000oooo0?ooo`0X0?ooo`@0003o:`3oool00`3o0000oooo0?ooo`0<0?oo
o`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3oool08`3oool00`000?l0oooo0?ooo`0R0?oo
o`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?ooo`0S0?oo
o`030000o`3oool0oooo0280oooo00<0o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0R0?oo
o`008@3oool00`000000oooo0?ooo`0T0?ooo`@0003o;`3oool0103o0000oooo0?oOW01800020000
00030018W03Oool0oooo00H0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?ooo`0R0?oo
o`040000o`3oool0oono07@000<0000000<0B9cO0?ooo`3oool00`3oool0102oM00000000000001d
_olD0?ooo`030?l0003oool0oooo00/0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?oo
o`030?ooo`030?oo_`1d000000000080000000H0B9cO0?ooo`3oool0oooo0?oOW018000200000003
0018W03Oool0oooo00h0oooo00<0003o0?ooo`3oool08`3oool00`3o0000oooo0?ooo`0;0?ooo`03
0000o`3oool0oooo00/0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`020?ooo`04
0;md00000000000007Boo`P0oooo000Q0?ooo`030000003oool0oooo0200oooo10000?lc0?ooo`08
0?l0003oool0okmd0000001d_ol0oooo0?oOW018M;l70?ooo`030000o`3oool0oooo02D0oooo00<0
o`000?ooo`3oool08@3oool00`000?l0oooo0?ooo`020?ooo`030?noM0000000M;oo00D0oooo00H0
gia8001d_`3oool0omnL04P00018W=lC0?ooo`030?l0003oool0oooo00/0oooo00<0003o0?ooo`3o
ool09P3oool00`3o0000oooo0?ooo`050?ooo`030?noM0000000M;oo00@0oooo00H0okmd0000001d
_ol0oooo0?oOW018M;l?0?ooo`030000o`3oool0oooo02<0oooo00<0o`000?ooo`3oool02`3oool0
0`000?l0oooo0?ooo`0<0?ooo`030?noM000M;l0oooo0080oooo00X0_gAd0;ooo`3oool0oooo0=nL
B000M;l0oooo0?oOW0180000B9cO1`3oool00240oooo00<000000?ooo`3oool06`3oool50000ocL0
oooo00<0o`000?ooo`3oool0103oool00`3o_g@0B9cO0?ooo`050?ooo`030000o`3oool0oooo02H0
oooo00<0o`000?ooo`3oool0803oool00`000?l0oooo0?ooo`020?ooo`030?oog`2LB00007Bo00@0
oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0B0?ooo`030?l0003oool0oooo00/0
oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`050?ooo`030?oog`2LB00007Bo00T0
oooo00<0okmd04RLg`3oool0303oool00`000?l0oooo0?ooo`0S0?ooo`030?l0003oool0oooo00`0
oooo00<0003o0?ooo`3oool03@3oool00`3o_g@007Bo0?ooo`040?ooo`030?oog`2LB000B9cO0080
oooo00<0okmd0000B02Lgol01P3oool00240oooo00<000000?ooo`3oool05`3oool40000oc`0oooo
00<0o`000?ooo`3oool0103oool00`3ookl0M01d0;ooo`050?ooo`030000o`3oool0oooo02H0oooo
00<0o`000?ooo`3oool07`3oool00`000?l0oooo0?ooo`040?ooo`030=nLB000B9`0gooo00<0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03OoolA0?ooo`030?l0003oool0oooo00`0
oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?ooo`060?ooo`030=nLB000B9`0gooo00P0
oooo00<0oono07@0M02oool02`3oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003oool0oooo00`0
oooo00<0003o0?ooo`3oool03P3oool00`3o_g@007Bo0?ooo`030?ooo`030?oog`2LB000B9cO0080
oooo00@0oooO09a80000B9`0gooo1@3oool00240oooo00<000000?ooo`3oool04@3oool60000ocl0
oooo00<0o`000?ooo`3oool01@3oool00`3ooml0W4P007Boo`060?ooo`030000o`3oool0oooo02D0
oooo00<0o`000?ooo`3oool07P3oool00`000?l0oooo0?ooo`050?ooo`030?oOW01804P0W=oo00<0
oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0A0?ooo`030?l0003oool0oooo00`0
oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`050?ooo`030?oOW01804P0W=oo00P0
oooo00<0oooO09a8001d_ol02`3oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003oool0oooo00d0
oooo00<0003o0?ooo`3oool03P3oool00`3o_g@0B9cO0?ooo`020?ooo`030?oog`2LB00007Bo00<0
oooo00<0gia80000B02Lgol01@3oool00240oooo0P00000=0?ooo`D0003oA@3oool00`3o0000oooo
0?ooo`050?ooo`030?oo_`1d0000B9cO00H0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo
0?ooo`0N0?ooo`030000o`3oool0oooo00D0oooo00<0oono07@0001d_ol0103oool00`2oM00004RL
0=ooo`020?ooo`030?oOW0180000B9cO0100oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo
0?ooo`0V0?ooo`030?l0003oool0oooo00D0oooo00<0oono07@0001d_ol0203oool00`3ookl0M000
04RLg`0:0?ooo`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo
0?ooo`0?0?ooo`030=nLB018W=l0oooo0080oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B000
04RLg`050?ooo`008@3oool00`000000oooo0?ooo`070?ooo`D0003oBP3oool00`3o0000oooo0?oo
o`050?ooo`030?noM0000000M;oo00H0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?oo
o`0L0?ooo`030000o`3oool0oooo00L0oooo00<0_g@0001d_`3oool00`3oool00`3OW4P0001809cO
o`020?ooo`030?oog`2LB00007Bo0100oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo0?oo
o`0V0?ooo`030?l0003oool0oooo00H0oooo00<0_g@0001d_`3oool01`3oool00`3o_g@0000007Bo
o`090?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo0?oo
o`0@0?ooo`070;md001d_ol0oooo0?ooo`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00D0
oooo000Q0?ooo`030000003oool0oooo00D0oooo0P000?m?0?ooo`030?l0003oool0oooo0080oooo
00H0omoO0=ooo`3o_g@000000000M02oool60?ooo`030000o`3oool0oooo02H0oooo00<0o`000?oo
o`3oool0703oool00`000?l0oooo0?ooo`070?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P0
0018W=l00`3oool00`2oM00004RL0=ooo`0?0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?oo
o`3oool09P3oool00`3o0000oooo0?ooo`060?ooo`030=nLB000B9`0gooo00@0oooo00H0omoO0=oo
o`3o_g@000000000M02oool90?ooo`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3oool03P3o
ool00`000?l0oooo0?ooo`0@0?ooo`070?oOW01807@0_ooo0?ooo`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo00@0oooo000Q0?ooo`030000003oool0oooo00@0oooo00<0003o0?ooo`3oool0
C`3oool00`3o0000oooo0?ooo`020?ooo`050?oo_`1d000000000000B02Lgol01`3oool00`000?l0
oooo0?ooo`0V0?ooo`030?l0003oool0oooo01`0oooo00<0003o0?ooo`3oool01`3oool00`3ogi`0
B01809cOo`030?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool03`3oool00`3o
0000oooo0?ooo`0=0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool01@3oool00`3o
gi`0B01809cOo`040?ooo`050?oo_`1d000000000000B02Lgol02@3oool00`000?l0oooo0?ooo`0V
0?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3oool0403oool0203ooml0W4P007Boo`3o
ool0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`040?ooo`008@3oool00`000000oooo
0?ooo`030?ooo`80003oD03oool00`3o0000oooo0?ooo`040?ooo`030=nLW02oool0oooo00P0oooo
00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0K0?ooo`030000o`3oool0oooo00P0oooo
00<0oono07@0001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00l0oooo
00<0o`000?ooo`3oool03P3oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo00D0oooo
00<0oono07@0001d_ol01@3oool00`3OW9`0_ooo0?ooo`0:0?ooo`030000o`3oool0oooo02H0oooo
00<0o`000?ooo`3oool03P3oool00`000?l0oooo0?ooo`0@0?ooo`030?oo_`1d0000B9cO0080oooo
00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`040?ooo`008@3oool00`000000oooo0?oo
o`020?ooo`030000o`3oool0oooo0500oooo00<0o`000?ooo`3oool0103oool00`3ogi`0M9cO0?oo
o`090?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool06P3oool00`000?l0oooo0?oo
o`090?ooo`030;md0000M;l0oooo00@0oooo00H0gia80000M02oool0oooO09a80018W=l?0?ooo`03
0?l0003oool0oooo00l0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`060?ooo`03
0;md0000M;l0oooo00@0oooo00<0omnL07BLg`3oool02P3oool00`000?l0oooo0?ooo`0V0?ooo`03
0?l0003oool0oooo00l0oooo00<0003o0?ooo`3oool02`3oool01`3o_g@004RL0=ooo`3oool0gia8
0000B02Lgol00`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`D0oooo000Q0?ooo`030000003o
ool0oooo0080oooo00<0003o0?ooo`3oool0D03oool00`3o0000oooo0?ooo`050?ooo`050;md0000
0000000000000018W=l01P3oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo01T0oooo
00<0003o0?ooo`3oool0203oool01@3ookl0M0000000000007@0_ooo00D0oooo00@0gia800000000
000007Bo403oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?oo
o`3oool0103oool01@3ookl0M0000000000007@0_ooo00D0oooo00D0_g@0000000000000000004RL
g`060?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool0403oool00`000?l0oooo0?oo
o`0<0?ooo`050?oOW018000000000000001d_ol01@3oool0103OW4P0000000000000M;l60?ooo`00
8@3oool00`000000oooo0?ooo`020?ooo`030000o`3oool0oooo0500oooo00<0o`000?ooo`3oool0
403oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?ooo`3oool0
9P3oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool0
603oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo0100oooo00<0003o0?ooo`3oool0
803oool00240oooo0P0000020?ooo`030000o`3oool0oooo0540oooo00<0o`000?ooo`3oool0403o
ool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?ooo`3oool09P3o
ool00`3o0000oooo0?ooo`0?0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool0603o
ool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo0100oooo00<0003o0?ooo`3oool0803o
ool00240oooo00D000000?ooo`3oool0oooo0000o`1C0?ooo`030?l0003oool0oooo0100oooo00<0
003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo02L0oooo00<0
o`000?ooo`3oool03`3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo01L0oooo00<0
003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo0200oooo000Q
0?ooo`050000003oool0oooo0?ooo`000?l0DP3oool00`3o0000oooo0?ooo`0A0?ooo`030000o`3o
ool0oooo02L0oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003o
ool0oooo0100oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0G0?ooo`030000o`3o
ool0oooo02L0oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0O0?ooo`008@3oool0
1@000000oooo0?ooo`3oool0003o0580oooo00<0o`000?ooo`3oool04P3oool00`000?l0oooo0?oo
o`0V0?ooo`030?l0003oool0oooo01H0oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?oo
o`0A0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool05P3oool00`000?l0oooo0?oo
o`0V0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool07`3oool00240oooo00@00000
0?ooo`3oool0003oD`3oool00`3o0000oooo0?ooo`0B0?ooo`030000o`3oool0oooo02H0oooo00<0
o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo0140oooo00<0
003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0E0?ooo`030000o`3oool0oooo02L0oooo00<0
o`000?ooo`3oool04P3oool00`000?l0oooo0?ooo`0O0?ooo`008@3oool010000000oooo0?ooo`00
0?mC0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?oo
o`0E0?ooo`030000o`3oool0oooo02P0oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?oo
o`0W0?ooo`030?l0003oool0oooo01D0oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?oo
o`0B0?ooo`030000o`3oool0oooo01l0oooo000Q0?ooo`040000003oool0oooo0000oe<0oooo00<0
o`000?ooo`3oool04P3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo01@0oooo00<0
003o0?ooo`3oool0:03oool00`3o0000oooo0?ooo`0A0?ooo`030000o`3oool0oooo02L0oooo00<0
o`000?ooo`3oool05@3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo0180oooo00<0
003o0?ooo`3oool07`3oool00240oooo00@000000?ooo`3oool0003oDP3oool00`3o0000oooo0?oo
o`0C0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool0503oool00`000?l0oooo0?oo
o`0X0?ooo`030?l0003oool0oooo0180oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?oo
o`0D0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool04`3oool00`000?l0oooo0?oo
o`0N0?ooo`005`3oool<000000030000o`00000000000540000000<0o`0000000000000050000000
0`000?l000000000000V000000030?l000000000000001@0000000<0003o0000000000009`000000
0`3o000000000000000C000000030000o`000000000002L0000000<0o`000000000000004`000000
0`000?l000000000000W000000030?l000000000000001@0000000<0003o0000000000006`000003
0?ooo`008@3oool00`000000oooo0000o`0@0?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<0o`000?oo
o`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000o`3oool0oooo00<0oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0<0?ooo`050?l0003oool0oooo0?ooo`0000004@3oool0
0`000?l000000?ooo`0B0?ooo`030000003oool0oooo0100oooo00@000000?ooo`3oool0o`003`3o
ool00`000000oooo0?ooo`030?ooo`030000o`3oool0oooo00X0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`060?ooo`030?l0003oool0oooo00L0oooo00<000000?ooo`3oool02@3o
ool00`000?l0oooo0?ooo`040?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0303o
ool01@3o0000oooo0?ooo`3oool000000180oooo00<0003o0?ooo`3oool0403oool00`000000oooo
0?ooo`0;0?ooo`008@3oool00`000000oooo0000o`0@0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<0
o`000?ooo`3oool01`3oool00`000000oooo0?ooo`0:0?ooo`030000o`3oool0oooo00<0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0<0?ooo`050?l0003oool0oooo0?ooo`000000
4@3oool00`000?l000000?ooo`0B0?ooo`030000003oool0oooo0100oooo00@000000?ooo`3oool0
o`003`3oool00`000000oooo0?ooo`030?ooo`030000o`3oool0oooo00X0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`060?ooo`030?l0003oool0oooo00L0oooo00<000000?ooo`3o
ool02@3oool00`000?l0oooo0?ooo`040?ooo`030000003oool0oooo0140oooo00<000000?ooo`3o
ool0303oool01@3o0000oooo0?ooo`3oool000000180oooo00<0003o0?ooo`3oool0403oool00`00
0000oooo0?ooo`0;0?ooo`008@3oool00`000000oooo0000o`1C0?ooo`030?l0003oool0oooo00L0
oooo00<000000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo0180
oooo00<0003o0?ooo`3oool04P3oool00`000000oooo0?ooo`0C0?ooo`030?l0003oool0oooo01@0
oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`060?ooo`030000003oool0oooo00T0
oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0D0?ooo`030000o`3oool0oooo0100
oooo00<000000?ooo`3oool02`3oool00240oooo00<000000?ooo`000?l0D`3oool00`3o0000oooo
0?ooo`0E0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool04P3oool00`000?l0oooo
0?ooo`0W0?ooo`030?l0003oool0oooo01D0oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo
0?ooo`0A0?ooo`030000o`3oool0oooo02P0oooo00<0o`000?ooo`3oool05@3oool00`000?l0oooo
0?ooo`0M0?ooo`008@3oool00`000000oooo0000o`1B0?ooo`030?l0003oool0oooo01H0oooo00<0
003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0B0?ooo`030000o`3oool0oooo02L0oooo00<0
o`000?ooo`3oool05@3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo0140oooo00<0
003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0F0?ooo`030000o`3oool0oooo01d0oooo000Q
0?ooo`030000003oool0003o0580oooo00<0o`000?ooo`3oool05P3oool00`000?l0oooo0?ooo`0V
0?ooo`030?l0003oool0oooo0140oooo00<0003o0?ooo`3oool0:03oool00`3o0000oooo0?ooo`0F
0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool04@3oool00`000?l0oooo0?ooo`0W
0?ooo`030?l0003oool0oooo01H0oooo00<0003o0?ooo`3oool07@3oool00240oooo00<000000?oo
o`000?l0DP3oool00`3o0000oooo0?ooo`0G0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?oo
o`3oool0403oool00`000?l0oooo0?ooo`0X0?ooo`030?l0003oool0oooo01H0oooo00<0003o0?oo
o`3oool09`3oool00`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?oo
o`3oool05P3oool00`000?l0oooo0?ooo`0M0?ooo`008@3oool00`000000003o0?ooo`1B0?ooo`03
0?l0003oool0oooo01L0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0@0?ooo`03
0000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool05`3oool00`000?l0oooo0?ooo`0W0?ooo`03
0?l0003oool0oooo0100oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0G0?ooo`03
0000o`3oool0oooo01`0oooo000Q0?ooo`03000000000?l0oooo0540oooo00<0o`000?ooo`3oool0
603oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo0100oooo00<0003o0?ooo`3oool0
9`3oool00`3o0000oooo0?ooo`0H0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool0
3`3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo01P0oooo00<0003o0?ooo`3oool0
703oool00240oooo00<000000000o`3oool0D@3oool00`3o0000oooo0?ooo`0H0?ooo`030000o`3o
ool0oooo02H0oooo00<0o`000?ooo`3oool0403oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003o
ool0oooo01P0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0?0?ooo`030000o`3o
ool0oooo02L0oooo00<0o`000?ooo`3oool0603oool00`000?l0oooo0?ooo`0L0?ooo`008@3oool0
0`000000003o0?ooo`1A0?ooo`030?l0003oool0oooo01T0oooo00<0003o0?ooo`3oool09@3oool0
0`3o0000oooo0?ooo`0@0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool0603oool0
0`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3oool09`3oool0
0`3o0000oooo0?ooo`0I0?ooo`030000o`3oool0oooo01/0oooo000Q0?ooo`03000000000?l0oooo
0540oooo00<0o`000?ooo`3oool06@3oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo
00l0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0J0?ooo`030000o`3oool0oooo
02H0oooo00<0o`000?ooo`3oool03P3oool00`000?l0oooo0?ooo`0W0?ooo`030?l0003oool0oooo
01T0oooo00<0003o0?ooo`3oool06`3oool00240oooo00<000000000o`3oool0D03oool00`3o0000
oooo0?ooo`0J0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool03P3oool00`000?l0
oooo0?ooo`0W0?ooo`030?l0003oool0oooo01X0oooo00<0003o0?ooo`3oool09P3oool00`3o0000
oooo0?ooo`0>0?ooo`030000o`3oool0oooo02L0oooo00<0o`000?ooo`3oool06P3oool00`000?l0
oooo0?ooo`0J0?ooo`008@3oool00`000000003o0?ooo`1@0?ooo`030?l0003oool0oooo01/0oooo
00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?ooo`0>0?ooo`030000o`3oool0oooo02L0oooo
00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?ooo`0U0?ooo`030?l0003oool0oooo00h0oooo
00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0K0?ooo`030000o`3oool0oooo01X0oooo
000Q0?ooo`03000000000?l0oooo0500oooo00<0o`000?ooo`3oool06`3oool00`000?l0oooo0?oo
o`0U0?ooo`030?l0003oool0oooo00h0oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?oo
o`0K0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool0303oool00`000?l0oooo0?oo
o`0W0?ooo`030?l0003oool0oooo01/0oooo00<0003o0?ooo`3oool06P3oool00240oooo00<00000
0000o`3oool0D03oool00`3o0000oooo0?ooo`0L0?ooo`030000o`3oool0oooo02D0oooo00<0o`00
0?ooo`3oool03@3oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo01`0oooo00<0003o
0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0<0?ooo`030000o`3oool0oooo02L0oooo00<0o`00
0?ooo`3oool0703oool00`000?l0oooo0?ooo`0I0?ooo`008@3oool00`000000003o0?ooo`1?0?oo
o`030?l0003oool0oooo01h0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo0?ooo`0=0?oo
o`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool07@3oool00`000?l0oooo0?ooo`0U0?oo
o`030?l0003oool0oooo00`0oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0L0?oo
o`030000o`3oool0oooo01T0oooo000Q0?ooo`03000000000?l0oooo04l0oooo00<0o`000?ooo`3o
ool07P3oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003oool0oooo00d0oooo00<0003o0?ooo`3o
ool09P3oool00`3o0000oooo0?ooo`0M0?ooo`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3o
ool0303oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo01h0oooo00<0003o0?ooo`3o
ool0603oool00240oooo00<000000000o`3oool0C`3oool00`3o0000oooo0?ooo`0O0?ooo`030000
o`3oool0oooo02<0oooo00<0o`000?ooo`3oool03@3oool00`000?l0oooo0?ooo`0V0?ooo`030?l0
003oool0oooo01h0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo0?ooo`0<0?ooo`030000
o`3oool0oooo02H0oooo00<0o`000?ooo`3oool07`3oool00`000?l0oooo0?ooo`0G0?ooo`008@3o
ool00`000000003o0?ooo`1?0?ooo`030?l0003oool0oooo0200oooo00<0003o0?ooo`3oool08P3o
ool00`3o0000oooo0?ooo`0=0?ooo`030000o`3oool0oooo02D0oooo00<0o`000?ooo`3oool0803o
ool00`000?l0oooo0?ooo`0S0?ooo`030?l0003oool0oooo00`0oooo00<0003o0?ooo`3oool09P3o
ool00`3o0000oooo0?ooo`0P0?ooo`030000o`3oool0oooo01H0oooo000Q0?ooo`03000000000?l0
oooo04h0oooo00<0o`000?ooo`3oool08P3oool00`000?l0oooo0?ooo`0R0?ooo`030?l0003oool0
oooo00/0oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0Q0?ooo`030000o`3oool0
oooo02<0oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0V0?ooo`030?l0003oool0
oooo0240oooo00<0003o0?ooo`3oool05P3oool00240oooo00<000000000o`3oool0CP3oool00`3o
0000oooo0?ooo`0R0?ooo`030000o`3oool0oooo0280oooo00<0o`000?ooo`3oool02`3oool00`00
0?l0oooo0?ooo`0V0?ooo`030?l0003oool0oooo0280oooo00<0003o0?ooo`3oool08P3oool00`3o
0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo02H0oooo00<0o`000?ooo`3oool08P3oool00`00
0?l0oooo0?ooo`0E0?ooo`008@3oool00`000000003o0?ooo`1>0?ooo`030?l0003oool0oooo02<0
oooo00<0003o0?ooo`3oool08@3oool00`3o0000oooo0?ooo`0;0?ooo`030000o`3oool0oooo02D0
oooo00<0o`000?ooo`3oool0903oool00`000?l0oooo0?ooo`0Q0?ooo`030?l0003oool0oooo00X0
oooo00<0003o0?ooo`3oool09P3oool00`3o0000oooo0?ooo`0S0?ooo`030000o`3oool0oooo01@0
oooo000Q0?ooo`03000000000?l0oooo04h0oooo00<0o`000?ooo`3oool0903oool00`000?l0oooo
0?ooo`0P0?ooo`030?l0003oool0oooo00/0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo
0?ooo`0T0?ooo`030000o`3oool0oooo0240oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo
0?ooo`0U0?ooo`030?l0003oool0oooo02D0oooo00<0003o0?ooo`3oool04`3oool00240oooo00<0
00000000o`3oool0C@3oool00`3o0000oooo0?ooo`0V0?ooo`030000o`3oool0oooo01l0oooo00<0
o`000?ooo`3oool02`3oool00`000?l0oooo0?ooo`0U0?ooo`030?l0003oool0oooo02D0oooo00<0
003o0?ooo`3oool0803oool00`3o0000oooo0?ooo`0:0?ooo`030000o`3oool0oooo02D0oooo00<0
o`000?ooo`3oool09P3oool00`000?l0oooo0?ooo`0B0?ooo`008@3oool00`000000003o0?ooo`1=
0?ooo`030?l0003oool0oooo02L0oooo00<0003o0?ooo`3oool07`3oool00`3o0000oooo0?ooo`0:
0?ooo`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool09`3oool00`000?l0oooo0?ooo`0O
0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?ooo`3oool09@3oool00`3o0000oooo0?ooo`0W
0?ooo`030000o`3oool0oooo0140oooo000Q0?ooo`03000000000?l0oooo04d0oooo00<0o`000?oo
o`3oool0:03oool00`000?l0oooo0?ooo`0N0?ooo`030?l0003oool0oooo00X0oooo00<0003o0?oo
o`3oool0903oool00`3o0000oooo0?ooo`0X0?ooo`030000o`3oool0oooo01h0oooo00<0o`000?oo
o`3oool02P3oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003oool0oooo02T0oooo00<0003o0?oo
o`3oool0403oool00240oooo00<0003o0?ooo`3oool0C03oool00`3o0000oooo0?ooo`0Z0?ooo`03
0000o`3oool0oooo01d0oooo00<0o`000?ooo`3oool02P3oool00`000?l0oooo0?ooo`0T0?ooo`03
0?l0003oool0oooo02T0oooo00<0003o0?ooo`3oool07P3oool00`3o0000oooo0?ooo`090?ooo`03
0000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool0:P3oool00`000?l0oooo0?ooo`0?0?ooo`00
8@3oool00`000?l0oooo0?ooo`1<0?ooo`030?l0003oool0oooo02/0oooo00<0003o0?ooo`3oool0
703oool00`3o0000oooo0?ooo`090?ooo`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool0
:`3oool00`000?l0oooo0?ooo`0M0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool0
9@3oool00`3o0000oooo0?ooo`0[0?ooo`030000o`3oool0oooo00h0oooo000Q0?ooo`030000o`3o
ool0oooo04`0oooo00<0o`000?ooo`3oool0;03oool00`000?l0oooo0?ooo`0K0?ooo`030?l0003o
ool0oooo00T0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo0?ooo`0/0?ooo`030000o`3o
ool0oooo01`0oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0T0?ooo`030?l0003o
ool0oooo02`0oooo00<0003o0?ooo`3oool03P3oool00240oooo00<0003o0?ooo`3oool0B`3oool0
0`3o0000oooo0?ooo`0^0?ooo`030000o`3oool0oooo01/0oooo00<0o`000?ooo`3oool0203oool0
0`000?l0oooo0?ooo`0S0?ooo`030?l0003oool0oooo02h0oooo00<0003o0?ooo`3oool06`3oool0
0`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo02@0oooo00<0o`000?ooo`3oool0;@3oool0
0`000?l0oooo0?ooo`0=0?ooo`008@3oool00`000?l0oooo0?ooo`1;0?ooo`030?l0003oool0oooo
02l0oooo00<0003o0?ooo`3oool06P3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo
02<0oooo00<0o`000?ooo`3oool0;`3oool00`000?l0oooo0?ooo`0J0?ooo`030?l0003oool0oooo
00P0oooo00<0003o0?ooo`3oool0903oool00`3o0000oooo0?ooo`0^0?ooo`030000o`3oool0oooo
00`0oooo000Q0?ooo`030000o`000000oooo04/0oooo00<0o`000?ooo`3oool0<03oool00`000?l0
oooo0?ooo`0I0?ooo`030?l0003oool0oooo00P0oooo00<0003o0?ooo`3oool08`3oool00`3o0000
oooo0?ooo`0`0?ooo`030000o`3oool0oooo01T0oooo00<0o`000?ooo`3oool0203oool00`000?l0
oooo0?ooo`0S0?ooo`030?l0003oool0oooo0300oooo00<0003o0?ooo`3oool02`3oool00240oooo
00<0003o0?ooo`3oool0BP3oool00`3o0000oooo0?ooo`0b0?ooo`030000o`3oool0oooo01P0oooo
00<0o`000?ooo`3oool0203oool00`000?l0oooo0?ooo`0R0?ooo`030?l0003oool0oooo0380oooo
00<0003o0?ooo`3oool0603oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo02<0oooo
00<0o`000?ooo`3oool0?P3oool00240oooo00<0003o0?ooo`3oool0BP3oool00`3o0000oooo0?oo
o`0b0?ooo`030000o`3oool0oooo01P0oooo00<0o`000?ooo`3oool0203oool00`000?l0oooo0?oo
o`0R0?ooo`030?l0003oool0oooo03<0oooo00<0003o0?ooo`3oool0603oool00`3o0000oooo0?oo
o`070?ooo`030000o`3oool0oooo02<0oooo00<0o`000?ooo`3oool0?P3oool00240oooo00<0003o
0?ooo`3oool0B@3oool00`3o0000oooo0?ooo`0d0?ooo`030000o`3oool0oooo01P0oooo00<0o`00
0?ooo`3oool01`3oool00`000?l0oooo0?ooo`0R0?ooo`030?l0003oool0oooo03@0oooo00<0003o
0?ooo`3oool05`3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo0280oooo00<0o`00
0?ooo`3oool0?`3oool00100oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`090?oo
o`030000o`3oool0oooo04T0oooo00<0o`000?ooo`3oool0=@3oool00`000?l0oooo0?ooo`0G0?oo
o`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool08@3oool00`3o0000oooo0?ooo`0e0?oo
o`030000o`3oool0oooo01L0oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0R0?oo
o`030?l0003oool0oooo03l0oooo000B0?ooo`030?noM0000000M;oo00`0oooo00<0003o0?ooo`3o
ool0B@3oool00`3o0000oooo0?ooo`0e0?ooo`030000o`3oool0oooo01L0oooo00<0o`000?ooo`3o
ool01`3oool00`000?l0oooo0?ooo`0Q0?ooo`030?l0003oool0oooo03H0oooo00<0003o0?ooo`3o
ool05P3oool00`3o0000oooo0?ooo`070?ooo`030000o`3oool0oooo0240oooo00<0o`000?ooo`3o
ool0@03oool00180oooo00<0oooO09a80000M;l0303oool00`000?l0oooo0?ooo`180?ooo`030?l0
003oool0oooo03L0oooo00<0003o0?ooo`3oool05P3oool00`3o0000oooo0?ooo`060?ooo`030000
o`3oool0oooo0280oooo00<0o`000?ooo`3oool0=`3oool00`000?l0oooo0?ooo`0E0?ooo`030?l0
003oool0oooo00H0oooo00<0003o0?ooo`3oool08P3oool00`3o0000oooo0?ooo`100?ooo`004`3o
ool00`3OW4P004RL0=ooo`0;0?ooo`030000o`3oool0oooo04P0oooo00<0o`000?ooo`3oool0>03o
ool00`000?l0oooo0?ooo`0E0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool08@3o
ool00`3o0000oooo0?ooo`0i0?ooo`030000o`3oool0oooo01@0oooo00<0o`000?ooo`3oool01P3o
ool00`000?l0oooo0?ooo`0R0?ooo`030?l0003oool0oooo0400oooo000C0?ooo`030?oOW01804P0
W=oo00/0oooo00<0003o0?ooo`3oool0B03oool00`3o0000oooo0?ooo`0h0?ooo`030000o`3oool0
oooo01H0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0Q0?ooo`030?l0003oool0
oooo03T0oooo00<0003o0?ooo`3oool05@3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0
oooo0240oooo00<0o`000?ooo`3oool0@@3oool001<0oooo00<0oono07@0001d_ol02`3oool00`00
0?l00000000000170?ooo`030?l0003oool0oooo03X0oooo00<0003o0?ooo`3oool05@3oool00`3o
0000oooo0?ooo`050?ooo`030000o`3oool0oooo0240oooo00<0o`000?ooo`3oool0>P3oool00`00
0?l0oooo0?ooo`0D0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool08@3oool00`3o
0000oooo0?ooo`110?ooo`00503oool00`2oM00007Bo0?ooo`0:0?ooo`030000003oool0oooo04L0
oooo00<0o`000?ooo`3oool0>`3oool00`000?l0oooo0?ooo`0D0?ooo`030?l0003oool0oooo00D0
oooo00<0003o0?ooo`3oool0803oool00`3o0000oooo0?ooo`0l0?ooo`030000o`3oool0oooo01<0
oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0Q0?ooo`030?l0003oool0oooo0440
oooo000D0?ooo`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool0A`3oool00`3o0000oooo
0?ooo`0k0?ooo`030000o`3oool0oooo01@0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo
0?ooo`0P0?ooo`030?l0003oool0oooo03`0oooo00<0003o0?ooo`3oool04`3oool00`3o0000oooo
0?ooo`050?ooo`030000o`3oool0oooo0200oooo00<0o`000?ooo`3oool0@P3oool001@0oooo00<0
omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`160?ooo`030?l0003oool0oooo03d0oooo00<0
003o0?ooo`3oool04`3oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo01l0oooo00<0
o`000?ooo`3oool0?P3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo00D0oooo00<0
003o0?ooo`3oool0803oool00`3o0000oooo0?ooo`120?ooo`00503oool00`3ookl0M00007Boo`0:
0?ooo`030000003oool0oooo04H0oooo00<0o`000?ooo`3oool0?@3oool00`000?l0oooo0?ooo`0C
0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool07`3oool00`3o0000oooo0?ooo`0n
0?ooo`030000o`3oool0oooo01<0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0O
0?ooo`030?l0003oool0oooo04<0oooo000E0?ooo`030;md0000M;l0oooo00T0oooo00<000000?oo
o`3oool0A@3oool00`3o0000oooo0?ooo`0o0?ooo`030000o`3oool0oooo0180oooo00<0o`000?oo
o`3oool01@3oool00`000?l0oooo0?ooo`0N0?ooo`030?l0003oool0oooo0400oooo00<0003o0?oo
o`3oool04P3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo01l0oooo00<0o`000?oo
o`3oool0@`3oool001<0oooo00D0oono07@000000000001d0;ooo`090?ooo`030000003oool0oooo
04D0oooo00<0o`000?ooo`3oool0?`3oool00`000?l0oooo0?ooo`0B0?ooo`030?l0003oool0oooo
00@0oooo00<0003o0?ooo`3oool07`3oool00`3o0000oooo0?ooo`100?ooo`030000o`3oool0oooo
0180oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0N0?ooo`030?l0003oool0oooo
04@0oooo000Q0?ooo`030000003oool0oooo04@0oooo00<0o`000?ooo`3oool0@03oool00`000?l0
oooo0?ooo`0C0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool07P3oool00`3o0000
oooo0?ooo`120?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool00`3oool00`000?l0
oooo0?ooo`0O0?ooo`030?l0003oool0oooo04@0oooo000Q0?ooo`800000A@3oool00`3o0000oooo
0?ooo`110?ooo`030000o`3oool0oooo0180oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo
0?ooo`0N0?ooo`030?l0003oool0oooo0480oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo
0?ooo`030?ooo`030000o`3oool0oooo01h0oooo00<0o`000?ooo`3oool0A@3oool00240oooo00<0
00000?ooo`3oool0@`3oool00`3o0000oooo0?ooo`120?ooo`030000o`3oool0oooo0180oooo00<0
o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0N0?ooo`030?l0003oool0oooo0480oooo00<0
003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo01h0oooo00<0
o`000?ooo`3oool0A@3oool00240oooo00<000000?ooo`3oool0@`3oool00`3o0000oooo0?ooo`13
0?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0M
0?ooo`030?l0003oool0oooo04@0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0?ooo`02
0?ooo`030000o`3oool0oooo01d0oooo00<0o`000?ooo`3oool0AP3oool00240oooo00<000000?oo
o`3oool0@P3oool00`3o0000oooo0?ooo`140?ooo`030000o`3oool0oooo0140oooo00<0o`000?oo
o`3oool00`3oool00`000?l0oooo0?ooo`0M0?ooo`030?l0003oool0oooo04@0oooo00<0003o0?oo
o`3oool04@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo01d0oooo00<0o`000?oo
o`3oool0AP3oool00240oooo00<000000?ooo`3oool0@P3oool00`3o0000oooo0?ooo`140?ooo`03
0000o`3oool0oooo0140oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0L0?ooo`03
0?l0003oool0oooo04H0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`020?ooo`03
0000o`3oool0oooo01`0oooo00<0o`000?ooo`3oool0A`3oool00240oooo00<000000?ooo`3oool0
@@3oool00`3o0000oooo0?ooo`160?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`3oool0
0`3oool00`000?l0oooo0?ooo`0L0?ooo`030?l0003oool0oooo04H0oooo00<0003o0?ooo`3oool0
403oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo01`0oooo00<0o`000?ooo`3oool0
A`3oool00240oooo00<000000?ooo`3oool0@@3oool00`3o0000oooo0?ooo`160?ooo`030000o`3o
ool0oooo0140oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0K0?ooo`030?l0003o
ool0oooo04L0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0?ooo`020?ooo`030000o`3o
ool0oooo01/0oooo00<0o`000?ooo`3oool0B03oool00240oooo00<000000?ooo`3oool0@03oool0
0`3o0000oooo0?ooo`170?ooo`030000o`3oool0oooo0140oooo00<0o`000?ooo`3oool00P3oool0
0`000?l0oooo0?ooo`0K0?ooo`030?l0003oool0oooo04P0oooo00<0003o0?ooo`3oool0403oool0
1@3o0000oooo0?ooo`3oool0003o01d0oooo00<0o`000?ooo`3oool0B03oool00240oooo0P000011
0?ooo`030?l0003oool0oooo04P0oooo00<0003o0?ooo`3oool0403oool01@3o0000oooo0?ooo`3o
ool0003o01d0oooo00<0o`000?ooo`3oool0B@3oool00`000?l0oooo0?ooo`0@0?ooo`050?l0003o
ool0oooo0?ooo`000?l0703oool00`3o0000oooo0?ooo`190?ooo`008@3oool00`000000oooo0?oo
o`0o0?ooo`030?l0003oool0oooo04T0oooo00<0003o0?ooo`3oool0403oool01@3o0000oooo0?oo
o`3oool0003o01d0oooo00<0o`000?ooo`3oool0B@3oool00`000?l0oooo0?ooo`0@0?ooo`050?l0
003oool0oooo0?ooo`000?l0703oool00`3o0000oooo0?ooo`190?ooo`008@3oool00`000000oooo
0?ooo`0o0?ooo`030?l0003oool0oooo04T0oooo00<0003o0?ooo`3oool04@3oool0103o0000oooo
0?ooo`000?lL0?ooo`030?l0003oool0oooo04/0oooo00<0003o0?ooo`3oool03`3oool0103o0000
oooo0?ooo`000?lL0?ooo`030?l0003oool0oooo04X0oooo000Q0?ooo`030000003oool0oooo03h0
oooo00<0o`000?ooo`3oool0B`3oool00`000?l0oooo0?ooo`0@0?ooo`040?l0003oool0oooo0000
oa`0oooo00<0o`000?ooo`3oool0B`3oool00`000?l0oooo0?ooo`0?0?ooo`040?l0003oool0oooo
0000oa`0oooo00<0o`000?ooo`3oool0BP3oool00240oooo00<000000?ooo`3oool0?@3oool00`3o
0000oooo0?ooo`1<0?ooo`030000o`3oool0oooo0100oooo00@0o`000?ooo`3oool0003o6`3oool0
0`3o0000oooo0?ooo`1<0?ooo`030000o`3oool0oooo0100oooo00<0o`000?ooo`000?l06`3oool0
0`3o0000oooo0?ooo`1;0?ooo`008@3oool00`000000oooo0?ooo`0m0?ooo`030?l0003oool0oooo
04`0oooo00<0003o0?ooo`3oool04@3oool00`3o0000oooo0000o`0K0?ooo`030?l0003oool0oooo
04`0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0000o`0K0?ooo`030?l0003oool0oooo
04/0oooo000Q0?ooo`030000003oool0oooo03`0oooo00<0o`000?ooo`3oool0CP3oool00`000?l0
oooo0?ooo`0@0?ooo`030?l0003oool0003o01X0oooo00<0o`000?ooo`3oool0CP3oool00`000?l0
oooo0?ooo`0?0?ooo`030?l0003oool0003o01X0oooo00<0o`000?ooo`3oool0C03oool00240oooo
00<000000?ooo`3oool0?03oool00`3o0000oooo0?ooo`1>0?ooo`030000o`3oool0oooo0100oooo
00<0o`000?ooo`000?l06@3oool00`3o0000oooo0?ooo`1?0?ooo`030000o`3oool0oooo00l0oooo
00<0o`000?ooo`000?l06@3oool00`3o0000oooo0?ooo`1=0?ooo`008@3oool00`000000oooo0?oo
o`0k0?ooo`030?l0003oool0oooo04l0oooo00<0003o0?ooo`3oool0403oool00`3o0000oooo0000
o`0I0?ooo`030?l0003oool0oooo04l0oooo00<0003o0?ooo`3oool0403oool00`3o0000003o0?oo
o`0H0?ooo`030?l0003oool0oooo04d0oooo000Q0?ooo`800000>`3oool00`3o0000oooo0?ooo`1A
0?ooo`030000o`3oool0oooo0100oooo00<0003o0?ooo`3oool05`3oool00`3o0000oooo0?ooo`1A
0?ooo`030000o`3oool0oooo00l0oooo00<0o`000000o`3oool05`3oool00`3o0000oooo0?ooo`1>
0?ooo`008@3oool00`000000oooo0?ooo`0j0?ooo`030?l0003oool0oooo0540oooo00<0003o0?oo
o`3oool0403oool00`000?l0oooo0?ooo`0F0?ooo`030?l0003oool0oooo0580oooo00<0003o0?oo
o`3oool03`3oool00`3o0000003o0?ooo`0F0?ooo`030?l0003oool0oooo04l0oooo000Q0?ooo`03
0000003oool0oooo03T0oooo00<0o`000?ooo`3oool0DP3oool00`000?l0oooo0?ooo`0@0?ooo`03
0000o`3oool0oooo01H0oooo00<0o`000?ooo`3oool0DP3oool00`000?l0oooo0?ooo`0?0?ooo`03
0?l000000?l0oooo01H0oooo00<0o`000?ooo`3oool0C`3oool00240oooo00<000000?ooo`3oool0
>03oool00`3o0000oooo0?ooo`1C0?ooo`030000o`3oool0oooo0100oooo00<0003o0?ooo`3oool0
5@3oool00`3o0000oooo0?ooo`1C0?ooo`030000o`3oool0oooo0100oooo00<0003o0?ooo`3oool0
503oool00`3o0000oooo0?ooo`1@0?ooo`008@3oool00`000000oooo0?ooo`0h0?ooo`030?l0003o
ool0oooo05@0oooo00<0003o0?ooo`3oool03`3oool00`000?l0o`000?ooo`0E0?ooo`030?l0003o
ool0oooo05@0oooo00<0003o0?ooo`3oool03P3oool00`000?l0o`000?ooo`0E0?ooo`030?l0003o
ool0oooo0500oooo000Q0?ooo`030000003oool0oooo03L0oooo00<0o`000?ooo`3oool0E@3oool0
0`000?l0oooo0?ooo`0?0?ooo`030000o`3o0000oooo01@0oooo00<0o`000?ooo`3oool0E@3oool0
0`000?l0oooo0?ooo`0>0?ooo`030000o`3o0000oooo01@0oooo00<0o`000?ooo`3oool0D@3oool0
0240oooo00<000000?ooo`3oool0=`3oool00`3o0000oooo0?ooo`1E0?ooo`030000o`3oool0oooo
00l0oooo00<0003o0?l0003oool04`3oool00`3o0000oooo0?ooo`1F0?ooo`030000o`3oool0oooo
00h0oooo00<0003o0?l0003oool04`3oool00`3o0000oooo0?ooo`1B0?ooo`008@3oool00`000000
oooo0?ooo`0f0?ooo`030?l0003oool0oooo05H0oooo00<0003o0?ooo`3oool03`3oool00`000?l0
o`000?ooo`0B0?ooo`030?l0003oool0oooo05L0oooo00<0003o0?ooo`3oool03P3oool00`000?l0
oooo0?l0000B0?ooo`030?l0003oool0oooo05<0oooo000Q0?ooo`800000=P3oool00`3o0000oooo
0?ooo`1H0?ooo`030000o`3oool0oooo00h0oooo00<0003o0?ooo`3o00004P3oool00`3o0000oooo
0?ooo`1H0?ooo`030000o`3oool0oooo00d0oooo00<0003o0?ooo`3o00004@3oool00`3o0000oooo
0?ooo`1D0?ooo`008@3oool00`000000oooo0?ooo`0e0?ooo`030?l0003oool0oooo05P0oooo00<0
003o0?ooo`3oool03@3oool010000?l0oooo0?ooo`3o000A0?ooo`030?l0003oool0oooo05T0oooo
00<0003o0?ooo`3oool03@3oool00`000?l0oooo0?l0000A0?ooo`030?l0003oool0oooo05@0oooo
000Q0?ooo`030000003oool0oooo03@0oooo00<0o`000?ooo`3oool0F@3oool00`000?l0oooo0?oo
o`0=0?ooo`040000o`3oool0oooo0?l00100oooo00<0o`000?ooo`3oool0FP3oool00`000?l0oooo
0?ooo`0=0?ooo`040000o`3oool0oooo0?l000l0oooo00<0o`000?ooo`3oool0E@3oool00100oooo
00<0gia80000000000000`0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo03<0oooo
00<0o`000?ooo`3oool0FP3oool00`000?l0oooo0?ooo`0=0?ooo`040000o`3oool0oooo0?l000l0
oooo00<0o`000?ooo`3oool0F`3oool00`000?l0oooo0?ooo`0=0?ooo`040000o`3oool0oooo0?l0
00h0oooo00<0o`000?ooo`3oool0EP3oool00140oooo00<0okmd001d_`3oool00P3oool00`2oM7@0
_ooo0?ooo`080?ooo`030000003oool0oooo0380oooo00<0o`000?ooo`3oool0G03oool00`000?l0
oooo0?ooo`0<0?ooo`050000o`3oool0oooo0?ooo`3o00003@3oool00`3o0000oooo0?ooo`1L0?oo
o`030000o`3oool0oooo00d0oooo00D0003o0?ooo`3oool0oooo0?l0000<0?ooo`030?l0003oool0
oooo05L0oooo000B0?ooo`030?noM000M;l0oooo00`0oooo00<000000?ooo`3oool0<P3oool00`3o
0000oooo0?ooo`1L0?ooo`030000o`3oool0oooo00`0oooo00D0003o0?ooo`3oool0oooo0?l0000<
0?ooo`030?l0003oool0oooo05h0oooo00<0003o0?ooo`3oool0303oool00`000?l0oooo0?ooo`02
0?ooo`030?l0003oool0oooo00P0oooo00<0o`000?ooo`3oool0F03oool001<0oooo00<0okmd001d
_`3oool02`3oool00`000000oooo0?ooo`0a0?ooo`030?l0003oool0oooo05d0oooo00<0003o0?oo
o`3oool0303oool00`000?l0oooo0?ooo`020?ooo`80o`002@3oool00`3o0000oooo0?ooo`1O0?oo
o`030000o`3oool0oooo00`0oooo00<0003o0?ooo`3oool00P3oool00`3o0000oooo0?ooo`070?oo
o`030?l0003oool0oooo05T0oooo000D0?ooo`030?noM018W=l0oooo00X0oooo00<000000?ooo`3o
ool0<03oool00`3o0000oooo0?ooo`1N0?ooo`030000o`3oool0oooo00`0oooo00<0003o0?ooo`3o
ool00`3oool20?l000H0oooo0P3o001R0?ooo`030000o`3oool0oooo00/0oooo00<0003o0?ooo`3o
ool0103oool00`3o0000oooo0?ooo`040?ooo`80o`00G03oool001D0oooo00<0gia804RLg`3oool0
2@3oool3000002l0oooo00<0o`000?ooo`3oool0H03oool00`000?l0oooo0?ooo`0;0?ooo`030000
o`3oool0oooo00@0oooo0`3o0000103oool0o`000?l0003o001T0?ooo`030000o`3oool0oooo00/0
oooo00<0003o0?ooo`3oool01@3oool60?l005h0oooo000F0?ooo`030;md001d_ol0oooo00P0oooo
00<000000?ooo`3oool0;`3oool00`3o0000oooo0?ooo`1P0?ooo`030000o`3oool0oooo00/0oooo
00<0003o0?ooo`3oool01`3oool00`3o0000oooo0?ooo`1V0?ooo`030000o`3oool0oooo00X0oooo
00<0003o0?ooo`3oool01`3oool20?l00600oooo000F0?ooo`030?oOW01807@0_ooo00P0oooo00<0
00000?ooo`3oool0;P3oool00`3o0000oooo0?ooo`1Q0?ooo`030000o`3oool0oooo00X0oooo00<0
003o0?ooo`3oool0L@3oool00`000?l0oooo0?ooo`0:0?ooo`030000o`3oool0oooo06T0oooo000F
0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?ooo`3oool0;@3oool00`3o0000oooo0?ooo`1R
0?ooo`030000o`3oool0oooo00X0oooo00<0003o0?ooo`3oool0L@3oool00`000?l0oooo0?ooo`0:
0?ooo`030000o`3oool0oooo06T0oooo000F0?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?oo
o`3oool0;03oool00`3o0000oooo0?ooo`1S0?ooo`030000o`3oool0oooo00X0oooo00<0003o0?oo
o`3oool0L@3oool00`000?l0oooo0?ooo`0:0?ooo`030000o`3oool0oooo06T0oooo000B0?ooo`07
0?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?ooo`030000003oool0oooo02/0oooo00<0o`00
0?ooo`3oool0I@3oool00`000?l0oooo0?ooo`090?ooo`030000o`3oool0oooo0740oooo00<0003o
0?ooo`3oool02P3oool00`000?l0oooo0?ooo`1Y0?ooo`004`3oool01@3ogi`0B000000000000000
M;oo00T0oooo00<000000?ooo`3oool0:P3oool00`3o0000oooo0?ooo`1V0?ooo`030000o`3oool0
oooo00T0oooo00<0003o0?ooo`3oool0LP3oool00`000?l0oooo0?ooo`090?ooo`030000o`3oool0
oooo06T0oooo000Q0?ooo`030000003oool0oooo02T0oooo00<0o`000?ooo`3oool0I`3oool00`00
0?l0oooo0?ooo`090?ooo`030000o`3oool0oooo0780oooo00<0003o0?ooo`3oool02@3oool00`00
0?l0oooo0?ooo`1Y0?ooo`008@3oool2000002T0oooo00<0o`000?ooo`3oool0J03oool00`000?l0
oooo0?ooo`090?ooo`030000o`3oool0oooo0780oooo00<0003o0?ooo`3oool0203oool00`000?l0
oooo0?ooo`1Z0?ooo`008@3oool00`000000oooo0?ooo`0W0?ooo`030?l0003oool0oooo06X0oooo
00<0003o0?ooo`3oool0203oool00`000?l0oooo0?ooo`1b0?ooo`030000o`3oool0oooo00P0oooo
00<0003o0?ooo`3oool0JP3oool00240oooo00<000000?ooo`3oool09P3oool00`3o0000oooo0?oo
o`1[0?ooo`030000o`3oool0oooo00P0oooo00<0003o0?ooo`3oool0LP3oool00`000?l0oooo0?oo
o`080?ooo`030000o`3oool0oooo06X0oooo000Q0?ooo`030000003oool0oooo02D0oooo00<0o`00
0?ooo`3oool0K03oool00`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo07@0oooo00<0003o
0?ooo`3oool01`3oool00`000?l0oooo0?ooo`1Z0?ooo`008@3oool00`000000oooo0?ooo`0U0?oo
o`030?l0003oool0oooo06`0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`1d0?oo
o`030000o`3oool0oooo00L0oooo00<0003o0?ooo`3oool0JP3oool00240oooo00<000000?ooo`3o
ool0903oool00`3o0000oooo0?ooo`1]0?ooo`030000o`3oool0oooo00L0oooo00<0003o0?ooo`3o
ool0M03oool00`000?l0oooo0?ooo`070?ooo`030000o`3oool0oooo06X0oooo000Q0?ooo`030000
003oool0oooo02<0oooo00<0o`000?ooo`3oool0K`3oool00`000?l0oooo0?ooo`060?ooo`030000
o`3oool0oooo07@0oooo00<0003o0?ooo`3oool01`3oool00`000?l0oooo0?ooo`1Z0?ooo`008@3o
ool00`000000oooo0?ooo`0R0?ooo`030?l0003oool0oooo0700oooo00<0003o0?ooo`3oool01P3o
ool00`000?l0oooo0?ooo`1e0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool0JP3o
ool00240oooo00<000000?ooo`3oool08@3oool00`3o0000oooo0?ooo`1a0?ooo`030000o`3oool0
oooo00H0oooo00<0003o0?ooo`3oool0M@3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0
oooo06/0oooo000Q0?ooo`8000008@3oool00`3o0000oooo0?ooo`1b0?ooo`030000o`3oool0oooo
00H0oooo00<0003o0?ooo`3oool0MP3oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo
06/0oooo000Q0?ooo`030000003oool0oooo01l0oooo00<0o`000?ooo`3oool0L`3oool00`000?l0
oooo0?ooo`050?ooo`030000o`3oool0oooo07L0oooo00<0003o0?ooo`3oool0103oool00`000?l0
oooo0?ooo`1[0?ooo`008@3oool00`000000oooo0?ooo`0N0?ooo`030?l0003oool0oooo07D0oooo
00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`1g0?ooo`030000o`3oool0oooo00@0oooo
00<0003o0?ooo`3oool0J`3oool00240oooo00<000000?ooo`3oool07@3oool00`3o0000oooo0?oo
o`1f0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0M`3oool00`000?l0oooo0?oo
o`040?ooo`030000o`3oool0oooo06/0oooo000Q0?ooo`030000003oool0oooo01`0oooo00<0o`00
0?ooo`3oool0M`3oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo07P0oooo00<0003o
0?ooo`3oool00P3oool00`000?l0oooo0?ooo`1/0?ooo`008@3oool00`000000oooo0?ooo`0K0?oo
o`030?l0003oool0oooo07T0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`1i0?oo
o`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0K03oool00240oooo00<000000?ooo`3o
ool06P3oool00`3o0000oooo0?ooo`1j0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3o
ool0N@3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06`0oooo000Q0?ooo`030000
003oool0oooo01T0oooo00<0o`000?ooo`3oool0O03oool01@000?l0oooo0?ooo`3oool0003o07/0
oooo00D0003o0?ooo`3oool0oooo0000o`1_0?ooo`008@3oool2000001T0oooo00<0o`000?ooo`3o
ool0O@3oool01@000?l0oooo0?ooo`3oool0003o07`0oooo00@0003o0?ooo`3oool0003oK`3oool0
0240oooo00<000000?ooo`3oool05`3oool00`3o0000oooo0?ooo`1n0?ooo`040000o`3oool0oooo
0000ogd0oooo00@0003o0?ooo`3oool0003oK`3oool00240oooo00<000000?ooo`3oool05P3oool0
0`3o0000oooo0?ooo`200?ooo`030000o`3oool0003o07d0oooo00@0003o0?ooo`3oool0003oK`3o
ool00240oooo00<000000?ooo`3oool0503oool20?l008<0oooo0`000?mn0?ooo`80003oL03oool0
0240oooo00<000000?ooo`3oool04`3oool00`3o0000oooo0?ooo`3o0?ooogL0oooo000Q0?ooo`03
0000003oool0oooo0180oooo00<0o`000?ooo`3oool0o`3ooomh0?ooo`008@3oool00`000000oooo
0?ooo`0A0?ooo`030?l0003oool0oooo0?l0ooooN@3oool00240oooo00<000000?ooo`3oool0403o
ool00`3o0000oooo0?ooo`3o0?ooogX0oooo000Q0?ooo`800000403oool00`3o0000oooo0?ooo`3o
0?ooog/0oooo000Q0?ooo`030000003oool0oooo00h0oooo00<0o`000?ooo`3oool0o`3oooml0?oo
o`008@3oool00`000000oooo0?ooo`0=0?ooo`030?l0003oool0oooo0?l0ooooO@3oool00240oooo
00<000000?ooo`3oool0303oool00`3o0000oooo0?ooo`3o0?ooogh0oooo000@0?ooo`030?oo_`1d
000000000080000000<0M;oo0?ooo`3oool02@3oool00`000000oooo0?ooo`0;0?ooo`030?l0003o
ool0oooo0?l0ooooO`3oool00100oooo00<0okmd00000000M;l00P3oool00`2oM4P0W=oo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo0P3o003o0?oooh80oooo000E0?ooo`030?oOW018B9`0gooo
00T0oooo00<000000?ooo`3oool0203oool00`3o0000oooo0?ooo`3o0?oooh80oooo000E0?ooo`03
0?oo_`1d0000M;oo00T0oooo00<000000?ooo`3oool01`3oool00`3o0000oooo0?ooo`3o0?oooh<0
oooo000E0?ooo`030?oOW0180000M;oo00T0oooo00<000000?ooo`3oool01@3oool20?l00?l0oooo
QP3oool001@0oooo00@0omnL04P0000004P0W=oo2@3oool00`3o00000000000000040?ooo`030?l0
003oool0oooo0?l0ooooQP3oool00180oooo00D0omnL04P000000000000007Boo`0:0?ooo`030000
003o0000oooo0080oooo0P3o003o0?ooohT0oooo000D0?ooo`030?oo_`1d000007Bo00X0oooo00<0
00000?l0003o00000P3o003o0?oooh/0oooo000F0?ooo`030=nLB000M;l0oooo00P0oooo00<00000
0?ooo`3oool0o`3ooon=0?ooo`005P3oool00`3ogi`0B01809cOo`080?ooo`030000003oool0oooo
0?l0ooooS@3oool001<0oooo00H0_g@007Boo`3oool0gia80000B02Lgol80?ooo`030000003oool0
oooo0?l0ooooS@3oool001@0oooo00@0_g@0000000000000B9cO2@3oool00`000000oooo0?ooo`3o
0?ooohd0oooo000Q0?ooo`030000003oool0oooo0?l0ooooS@3oool00240oooo0P00003o0?ooohh0
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.71536, -2.98443, \
0.0777404, 0.0361989}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData["Case 3: large \[CurlyEpsilon]"], "Subsubsection"],

Cell[TextData[
"The shape of the limit cycle depends strongly upon \[CurlyEpsilon]. "], 
  "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseCurve[1, 1, 3, 20];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.500002 0.235355 0.308954 0.0581196 [
[.02929 .29645 -9.34375 -15.6875 ]
[.02929 .29645 9.34375 0 ]
[.26465 .29645 -9.34375 -15.6875 ]
[.26465 .29645 9.34375 0 ]
[.73536 .29645 -5 -15.6875 ]
[.73536 .29645 5 0 ]
[.97071 .29645 -5 -15.6875 ]
[.97071 .29645 5 0 ]
[.4875 .07648 -18.6875 -7.84375 ]
[.4875 .07648 0 7.84375 ]
[.4875 .19271 -18.6875 -7.84375 ]
[.4875 .19271 0 7.84375 ]
[.4875 .42519 -10 -7.84375 ]
[.4875 .42519 0 7.84375 ]
[.4875 .54143 -10 -7.84375 ]
[.4875 .54143 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02929 .30895 m
.02929 .3152 L
s
gsave
.02929 .29645 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.26465 .30895 m
.26465 .3152 L
s
gsave
.26465 .29645 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.73536 .30895 m
.73536 .3152 L
s
gsave
.73536 .29645 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.97071 .30895 m
.97071 .3152 L
s
gsave
.97071 .29645 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.07636 .30895 m
.07636 .3127 L
s
.12343 .30895 m
.12343 .3127 L
s
.17051 .30895 m
.17051 .3127 L
s
.21758 .30895 m
.21758 .3127 L
s
.31172 .30895 m
.31172 .3127 L
s
.35879 .30895 m
.35879 .3127 L
s
.40586 .30895 m
.40586 .3127 L
s
.45293 .30895 m
.45293 .3127 L
s
.54707 .30895 m
.54707 .3127 L
s
.59414 .30895 m
.59414 .3127 L
s
.64121 .30895 m
.64121 .3127 L
s
.68829 .30895 m
.68829 .3127 L
s
.78243 .30895 m
.78243 .3127 L
s
.8295 .30895 m
.8295 .3127 L
s
.87657 .30895 m
.87657 .3127 L
s
.92364 .30895 m
.92364 .3127 L
s
.25 Mabswid
0 .30895 m
1 .30895 L
s
.5 .07648 m
.50625 .07648 L
s
gsave
.4875 .07648 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.5 .19271 m
.50625 .19271 L
s
gsave
.4875 .19271 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .42519 m
.50625 .42519 L
s
gsave
.4875 .42519 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .54143 m
.50625 .54143 L
s
gsave
.4875 .54143 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .10554 m
.50375 .10554 L
s
.5 .1346 m
.50375 .1346 L
s
.5 .16366 m
.50375 .16366 L
s
.5 .22177 m
.50375 .22177 L
s
.5 .25083 m
.50375 .25083 L
s
.5 .27989 m
.50375 .27989 L
s
.5 .33801 m
.50375 .33801 L
s
.5 .36707 m
.50375 .36707 L
s
.5 .39613 m
.50375 .39613 L
s
.5 .45425 m
.50375 .45425 L
s
.5 .48331 m
.50375 .48331 L
s
.5 .51237 m
.50375 .51237 L
s
.5 .04742 m
.50375 .04742 L
s
.5 .01836 m
.50375 .01836 L
s
.5 .57049 m
.50375 .57049 L
s
.5 .59955 m
.50375 .59955 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.73536 .36707 m
.74616 .3639 L
.75696 .35962 L
.77394 .35023 L
.78252 .34388 L
.7893 .33778 L
.79856 .32679 L
.80167 .32162 L
.80293 .31895 L
.80385 .31656 L
.80444 .31462 L
.80489 .31262 L
.80516 .3105 L
.80522 .30866 L
.80508 .30666 L
.80475 .3049 L
.80448 .30393 L
.80418 .30307 L
.80336 .30123 L
.80246 .29972 L
.80131 .29816 L
.79879 .29553 L
.7955 .29292 L
.79191 .29066 L
.78208 .2859 L
.75542 .27616 L
.72417 .26501 L
.67664 .24443 L
.60285 .20174 L
.55236 .16659 L
.4796 .11289 L
.44026 .08511 L
.39195 .05502 L
.36165 .03963 L
.33244 .02814 L
.30354 .02069 L
.27255 .01777 L
.2462 .02 L
.21852 .02762 L
.19331 .03976 L
.17153 .05463 L
.14768 .07597 L
.12436 .10238 L
.08951 .15325 L
.04492 .24307 L
.03464 .27018 L
.02907 .28779 L
.02721 .29502 L
.02652 .29821 L
.02596 .30119 L
Mistroke
.02538 .30557 L
.02522 .30879 L
.02526 .31042 L
.0254 .31196 L
.02562 .31321 L
.02589 .31422 L
.02654 .31582 L
.02738 .31715 L
.02836 .31818 L
.02933 .31892 L
.03054 .3196 L
.03176 .32011 L
.03403 .32079 L
.03537 .32107 L
.03663 .32129 L
.03933 .32166 L
.04239 .32196 L
.04877 .32241 L
.07409 .32385 L
.09975 .32551 L
.12808 .32783 L
.14413 .32943 L
.16289 .33162 L
.18413 .33463 L
.20613 .33846 L
.23198 .34411 L
.26154 .35236 L
.30328 .36782 L
.36744 .40073 L
.46712 .46872 L
.53275 .51706 L
.57486 .54593 L
.62557 .57532 L
.65337 .58779 L
.66862 .59326 L
.68537 .59802 L
.71526 .60284 L
.74355 .60252 L
.7725 .59659 L
.80203 .58401 L
.82597 .56845 L
.85038 .54717 L
.89592 .49106 L
.93091 .43111 L
.96426 .35499 L
.97127 .33394 L
.97347 .32593 L
.97499 .31924 L
.97577 .31463 L
.97615 .31061 L
Mistroke
.97617 .30772 L
.97595 .30554 L
.97553 .30372 L
.97499 .30232 L
.9744 .30128 L
.97368 .30035 L
.97281 .29953 L
.9718 .29883 L
.97079 .29829 L
.96983 .29789 L
.96757 .2972 L
.96623 .29691 L
.96496 .29669 L
.96232 .29633 L
.95943 .29604 L
.95404 .29564 L
.949 .29534 L
.92368 .29393 L
.89567 .29206 L
.86623 .28954 L
.83275 .28572 L
.8131 .28284 L
.78823 .27832 L
.75923 .27156 L
.72616 .26147 L
.67738 .24131 L
.59784 .19526 L
.47287 .10489 L
.42099 .06932 L
.39247 .05218 L
.36451 .0378 L
.33786 .02684 L
.30865 .01853 L
.27871 .01472 L
.24568 .0169 L
.21558 .02556 L
.18816 .03967 L
.16494 .05672 L
.14155 .07907 L
.09781 .13643 L
.06842 .18811 L
.03795 .25696 L
.03041 .27852 L
.02595 .29436 L
.02523 .29761 L
.02464 .30066 L
.02401 .30523 L
.02384 .30744 L
.02381 .30926 L
.02401 .31205 L
Mistroke
.02444 .3141 L
.02501 .31559 L
.02579 .3169 L
.02623 .31745 L
.02675 .31799 L
.02792 .3189 L
.02908 .31956 L
.0305 .32014 L
.03191 .32057 L
.03325 .32089 L
.03452 .32114 L
.03749 .32156 L
.04035 .32185 L
.0431 .32207 L
.04835 .32241 L
.06057 .32309 L
.11181 .32643 L
.14375 .32939 L
.1623 .33155 L
.18462 .33471 L
.20886 .339 L
.23495 .34484 L
.26708 .35415 L
.30654 .36923 L
.4526 .45805 L
.52689 .51284 L
.57607 .54671 L
.60609 .56491 L
.63593 .58031 L
.66357 .59157 L
.69417 .59995 L
.71195 .60256 L
.72884 .60332 L
.76034 .59981 L
.78869 .59056 L
.81379 .577 L
.83967 .55721 L
.86468 .53196 L
.90387 .47884 L
.93428 .42443 L
.96421 .35512 L
.97206 .3312 L
.97416 .32312 L
.97536 .31727 L
.97595 .31308 L
.97619 .30952 L
.97617 .30784 L
.97607 .30645 L
.97569 .30429 L
.97512 .3026 L
Mistroke
.97472 .3018 L
.97431 .30114 L
.97338 .30004 L
.97281 .29953 L
.97226 .29912 L
.97124 .29851 L
.97006 .29797 L
.96889 .29756 L
.96778 .29725 L
.96522 .29673 L
.96372 .2965 L
.96228 .29632 L
.95611 .29578 L
.95077 .29545 L
.94479 .29511 L
.93229 .29443 L
.90832 .29296 L
.8799 .29079 L
.8446 .28721 L
.82453 .28458 L
.7997 .28054 L
.77195 .27475 L
.74061 .26622 L
.69979 .25138 L
.64634 .22514 L
.54227 .1561 L
.46487 .09915 L
.421 .06932 L
.36671 .03883 L
.34925 .03116 L
.33018 .02426 L
.31357 .01963 L
.29495 .01615 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003of03oool0
0`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000
oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?oo
o`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00LP3oool;000005/0oooo00<000000?oo
o`3oool0e@3oool006T0oooo4@0000030?ooo`L00000E03oool00`000000oooo0?ooo`3E0?ooo`00
IP3oool3000000030?ooo`000000000000D000002@3oool6000000<0oooo1`00001>0?ooo`800000
eP3oool006<0oooo0`000000103oool0000000000000000F0?ooo`H000000P3oool6000004X0oooo
00<000000?ooo`3oool0e@3oool00600oooo0`000000103oool0000000000000000O0?ooo`@00000
0P3oool5000004L0oooo00<000000?ooo`3oool0e@3oool005h0oooo0P000000103oool000000000
0000000V0?ooo`@000000P3oool4000004@0oooo00<000000?ooo`3oool0e@3oool005`0oooo1@00
000]0?ooo`@0000000@0oooo0000000000000000@P3oool00`000000oooo0?ooo`3E0?ooo`00FP3o
ool4000003@0oooo0P0000020?ooo`<00000?`3oool00`000000oooo0?ooo`3E0?ooo`00F03oool4
000003P0oooo0P0000020?ooo`<00000?@3oool00`000000oooo0?ooo`3E0?ooo`00EP3oool40000
03`0oooo0`0000020?ooo`800000>`3oool00`000000oooo0?ooo`3E0?ooo`00E03oool400000440
oooo0P0000020?ooo`800000>@3oool00`000000oooo0?ooo`3E0?ooo`00DP3oool4000004D0oooo
0`0000000`3oool000000000000g0?ooo`030000003oool0oooo0=D0oooo001@0?ooo`@00000BP3o
ool2000000030?ooo`000000000003D0oooo00<000000?ooo`3oool0e@3oool004l0oooo00<00000
0?ooo`000000CP3oool2000000030?ooo`000000000003<0oooo00<000000?ooo`3oool0e@3oool0
04h0oooo0`00001A0?ooo`D00000<@3oool200000=H0oooo001<0?ooo`<00000E@3oool5000002l0
oooo00<000000?ooo`3oool0e@3oool004/0oooo0`00001H0?ooo`D00000;@3oool00`000000oooo
0?ooo`3E0?ooo`00B@3oool3000005`0oooo1@00000[0?ooo`030000003oool0oooo0=D0oooo0018
0?ooo`030000003oool0000005l0oooo1000000Z0?ooo`030000003oool0oooo0=D0oooo00170?oo
o`<00000HP3oool4000002P0oooo00<000000?ooo`3oool0e@3oool004H0oooo0P00001V0?ooo`<0
00009`3oool00`000000oooo0?ooo`3E0?ooo`00A@3oool2000006P0oooo1000000U0?ooo`030000
003oool0oooo0=D0oooo00140?ooo`800000J`3oool3000001D0oooo00<0oooO09a80018W=l0303o
ool00`000000oooo0?ooo`3E0?ooo`00@`3oool2000006d0oooo1000000D0?ooo`030=nLB000B9`0
gooo00/0oooo00<000000?ooo`3oool0e@3oool00480oooo0P00001`0?ooo`<000004`3oool00`3o
gi`0B01809cOo`0;0?ooo`030000003oool0oooo0=D0oooo00110?ooo`030000003oool0oooo0780
oooo0`00000=0?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0203oool00`000000
oooo0?ooo`3E0?ooo`00@03oool00`000000oooo0?ooo`1d0?ooo`<0000000<0W4P0000000000000
1P0000001P18W=l0oooo0?ooo`3ooml0W4Q809cOo`80oooo00<0_g@0001d_`3oool02P3oool00`00
0000oooo0?ooo`3E0?ooo`00?`3oool00`000000oooo0?ooo`1g0?ooo`<000002`3oool01`3ookl0
M01d0;ooo`3oool0okmd0018W03Oool02P3oool300000=D0oooo000n0?ooo`030000003oool0oooo
07T0oooo1000000:0?ooo`060?oOW018B9`0gooo0?oOW01804P0W=oo2P3oool00`000000oooo0?oo
o`3E0?ooo`00?@3oool00`000000oooo0?ooo`1l0?ooo`<000002P3oool01@3o_g@007Bo0?oog`2L
B000B9cO00X0oooo00<000000?ooo`3oool0e@3oool003`0oooo0P0000200?ooo`<000002@3oool0
103OW4P0B9cO0;md0000M;l:0?ooo`030000003oool0oooo0=D0oooo000l0?ooo`030000003oool0
oooo0800oooo0`0000090?ooo`040=nLB01dM7@0001d0;ooo`T0oooo00<000000?ooo`3oool0e@3o
ool003/0oooo00<000000?ooo`3oool0P`3oool3000000P0oooo00<0_g@00000B02Lgol02@3oool0
0`000000oooo0?ooo`3E0?ooo`00>P3oool00`000000oooo0?ooo`250?ooo`@000001P3oool00`3o
oml0W4P004RLg`090?ooo`030000003oool0oooo0=D0oooo000i0?ooo`030000003oool0oooo08L0
oooo1000000A0?ooo`030000003oool0oooo0=D0oooo000h0?ooo`030000003oool0oooo08X0oooo
1000000?0?ooo`030000003oool0oooo0=D0oooo000h0?ooo`030000003oool0oooo08/0oooo1000
000>0?ooo`030000003oool0oooo0=D0oooo000g0?ooo`030000003oool0oooo08h0oooo0`00000=
0?ooo`030000003oool0oooo0=D0oooo000f0?ooo`030000003oool0oooo0900oooo1000000;0?oo
o`030000003oool0oooo0=D0oooo000e0?ooo`030000003oool0oooo09<0oooo0`00000:0?ooo`80
0000eP3oool003@0oooo0P00002F0?ooo`<000002@3oool00`000000oooo0?ooo`3E0?ooo`00=03o
ool00`000000oooo0?ooo`2F0?ooo`@000001`3oool00`000000oooo0?ooo`3E0?ooo`00<`3oool0
0`000000oooo0?ooo`2I0?ooo`<000001P3oool00`000000oooo0?ooo`3E0?ooo`00<P3oool20000
09`0oooo0`0000050?ooo`030000003oool0oooo0=D0oooo000a0?ooo`800000W`3oool3000000<0
oooo00<000000?ooo`3oool0e@3oool00340oooo00<000000?ooo`3oool0W`3oool300000080oooo
00<000000?ooo`3oool0e@3oool00300oooo0P00002R0?ooo`<0000000<0oooo0000003oool0eP3o
ool002l0oooo0P00002U0?ooo`<00000e`3oool002h0oooo0P00002W0?ooo`<00000eP3oool002d0
oooo0P00002Y0?ooo`<00000e@3oool002d0oooo0P00002Y0?ooo`030000003oool0000000800000
d`3oool002`0oooo0P00002Z0?ooo`030000003oool0oooo00<00000dP3oool002/0oooo0P00002[
0?ooo`8000000P3oool300000=40oooo000Z0?ooo`030000003oool000000:/0oooo00<000000?oo
o`3oool00`3oool300000<l0oooo000Z0?ooo`800000[03oool00`000000oooo0?ooo`040?ooo`03
0000003oool000000<h0oooo000Y0?ooo`800000[@3oool00`000000oooo0?ooo`050?ooo`<00000
c@3oool002T0oooo0P00002]0?ooo`030000003oool0oooo00L0oooo0`00003;0?ooo`00:03oool2
00000:h0oooo00<000000?ooo`3oool0203oool00`000000oooo0000003:0?ooo`00:03oool00`00
0000oooo0?ooo`2]0?ooo`030000003oool0oooo00T0oooo0`0000390?ooo`009`3oool200000:l0
oooo00<000000?ooo`3oool02`3oool300000<L0oooo000V0?ooo`800000/03oool00`000000oooo
0?ooo`0<0?ooo`030000003oool000000<H0oooo000V0?ooo`030000003oool0oooo0:l0oooo00<0
00000?ooo`3oool03@3oool400000<@0oooo000U0?ooo`800000/@3oool00`000000oooo0?ooo`0?
0?ooo`<00000``3oool002D0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`0@0?oo
o`030000003oool000000<80oooo000T0?ooo`800000/P3oool200000180oooo100000300?ooo`00
903oool00`000000oooo0?ooo`2a0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`000000
_`3oool002<0oooo0P00002c0?ooo`030000003oool0oooo01@0oooo1000002m0?ooo`008`3oool0
0`000000oooo0?ooo`2b0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`000000_03oool0
0280oooo0P00002d0?ooo`030000003oool0oooo01L0oooo1000002j0?ooo`008@3oool200000;D0
oooo00<000000?ooo`3oool06@3oool300000;T0oooo000Q0?ooo`800000]@3oool00`000000oooo
0?ooo`0J0?ooo`@00000]`3oool00200oooo0P00002f0?ooo`030000003oool0oooo01`0oooo0`00
002f0?ooo`00803oool200000:@0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`09
0?ooo`030000003oool0oooo01d0oooo1000002d0?ooo`007`3oool200000:H0oooo00<0okmd001d
_`3oool00P3oool00`2oM7@0_ooo0?ooo`090?ooo`030000003oool0oooo01h0oooo1000002c0?oo
o`007`3oool200000:L0oooo00<0okmd001d_`3oool03@3oool00`000000oooo0?ooo`0P0?ooo`@0
0000/@3oool001h0oooo0P00002Y0?ooo`030?noM000M;l0oooo00`0oooo00<000000?ooo`3oool0
8@3oool400000;00oooo000N0?ooo`800000V@3oool00`3ooml0W4P0000000070000000304RLg`3o
ool0oooo00@0oooo00<0okmd04RLg`3oool02`3oool00`000000oooo0?ooo`0S0?ooo`@00000[P3o
ool001d0oooo0P00002/0?ooo`030=nLB018W=l0oooo00X0oooo0`00000T0?ooo`@00000[@3oool0
01d0oooo0P00002]0?ooo`030;md001d_ol0oooo00T0oooo00<000000?ooo`3oool09P3oool40000
0:/0oooo000L0?ooo`800000[P3oool00`3ogi`0B01d0;ooo`090?ooo`030000003oool0oooo02L0
oooo1000002Z0?ooo`00703oool200000:h0oooo00<0oooO09a8001d_ol02@3oool00`000000oooo
0?ooo`0Y0?ooo`@00000Z03oool001/0oooo0P00002_0?ooo`030?oo_`1d0000B9cO00T0oooo00<0
00000?ooo`3oool0:P3oool400000:L0oooo000K0?ooo`800000Z`3oool01`3o_g@004RL0=ooo`3o
ool0gia80000B02Lgol02@3oool00`000000oooo0?ooo`0/0?ooo`@00000Y@3oool001X0oooo0P00
002]0?ooo`050?oOW018000000000000001d_ol02P3oool00`000000oooo0?ooo`0^0?ooo`030000
003oool000000:@0oooo000J0?ooo`800000_03oool00`000000oooo0?ooo`0_0?ooo`@00000XP3o
ool001X0oooo00<000000?ooo`3oool0^`3oool00`000000oooo0?ooo`0a0?ooo`@00000X03oool0
01T0oooo0P00002m0?ooo`030000003oool0oooo03<0oooo1000002N0?ooo`006@3oool00`000000
oooo0?ooo`2l0?ooo`030000003oool0oooo03@0oooo1000002M0?ooo`00603oool200000;h0oooo
00<000000?ooo`3oool0=P3oool4000009/0oooo000H0?ooo`030000003oool0oooo0;d0oooo0P00
000i0?ooo`@00000V@3oool001L0oooo0P00002o0?ooo`030000003oool0oooo03T0oooo1@00002G
0?ooo`005`3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo03/0oooo1@00002E0?oo
o`005P3oool200000<00oooo00<000000?ooo`3oool0?@3oool4000009@0oooo000F0?ooo`030000
003oool0oooo0;l0oooo00<000000?ooo`3oool0?`3oool400000980oooo000E0?ooo`800000`@3o
ool00`000000oooo0?ooo`100?ooo`D00000T03oool001D0oooo00<000000?ooo`3oool0`03oool0
0`000000oooo0?ooo`120?ooo`D00000SP3oool001D0oooo00<000000?ooo`3oool0`03oool00`00
0000oooo0?ooo`140?ooo`030000003oool0000000800000S03oool001@0oooo00<000000?ooo`3o
ool0`@3oool00`000000oooo0?ooo`150?ooo`80000000@0oooo0000000000000000RP3oool001@0
oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`170?ooo`80000000@0oooo00000000
00000000R03oool001<0oooo0P0000330?ooo`030000003oool0oooo04T0oooo0P000000103oool0
00000000000000260?ooo`004`3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo04/0
oooo0P0000000`3oool0000000000002000008<0oooo000B0?ooo`800000a03oool00`000000oooo
0?ooo`1=0?ooo`80000000<0oooo0000000000000P0000210?ooo`004P3oool200000<@0oooo0P00
001@0?ooo`80000000<0oooo0000000000000P00001o0?ooo`004@3oool200000<D0oooo00<00000
0?ooo`3oool0D@3oool200000080oooo1000001l0?ooo`00403oool00`3OW4P00000000000030000
000304RLg`3oool0oooo05`0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`1K0?oo
o`030000003oool0oooo05<0oooo0`0000020?ooo`<000000P3oool00`3ookl0M000000000020000
000304RLg`3oool0oooo05d0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`0:0?oo
o`004@3oool2000000<0oooo00<0_gAd0;ooo`3oool0GP3oool00`3o_g@0000007Boo`1N0?ooo`03
0000003oool0oooo05H0oooo0P0000030?ooo`8000000P3oool00`3o_g@0000007Boo`1Q0?ooo`03
0?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool02P3oool00100oooo0P0000000`3o_g@007Bo
0?ooo`1R0?ooo`030?oog`2LB00007Bo05h0oooo00<000000?ooo`3oool0F03oool2000000<0oooo
0`0000000`2LB00007Bo0?ooo`1Q0?ooo`030?noM000M;l0oooo00h0oooo000@0?ooo`80000000<0
oooo0?noM000M;l0H`3oool00`3OW4P004RL0=ooo`1M0?ooo`030000003oool0oooo05X0oooo0P00
00040?ooo`@00000H@3oool00`3o_g@007Bo0?ooo`0=0?ooo`000`3oool00`3ooml0W4P000000007
0000000304RLg`3oool0oooo008000000P3oool00`3o_g@0B9cO0?ooo`1A0?ooo`030?oog`2LB000
000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`1M0?ooo`030000003oool0
oooo05`0oooo0`000000103oool0omnL04P0B02Lgol4000005l0oooo00<0okmd04RLg`3oool0303o
ool000l0oooo0P0000040?ooo`030=nLB018W=l0oooo0600oooo00<0oono07@0001d_ol0G@3oool0
0`000000oooo0?ooo`1O0?ooo`<0000000@0M;oo0?ooo`3oool0oooo1@00001L0?ooo`030=nLB018
W=l0oooo00/0oooo000?0?ooo`8000001@3oool00`2oM000M;oo0?ooo`1P0?ooo`030;md0000M;l0
oooo05`0oooo00<000000?ooo`3oool0H@3oool00`2oM000000000000002000000@0oooo1@00001I
0?ooo`030;md001d_ol0oooo00X0oooo000?0?ooo`8000001@3oool00`3ogi`0B01d0;ooo`1P0?oo
o`030=nLB000B9`0gooo05`0oooo00<000000?ooo`3oool0H@3oool00`3OW4P004RL0=ooo`020?oo
o`<00000103oool6000005D0oooo00<0omnL04P0M02oool02P3oool000h0oooo0P0000060?ooo`03
0?oog`2LB000M;oo0600oooo00<0omnL04P0B02Lgol0G03oool00`000000oooo0?ooo`1Q0?ooo`03
0?oOW01804P0W=oo00D0oooo0`0000050?ooo`H00000D@3oool00`3ooml0W4P007Boo`0:0?ooo`00
3P3oool2000000H0oooo00<0oono07@00018W=l0H03oool00`3ookl0M00007Boo`1L0?ooo`030000
003oool0oooo0640oooo00<0oono07@0001d_ol0203oool2000000L0oooo1`00001<0?ooo`030?oo
_`1d0000B9cO00X0oooo000>0?ooo`0;0000003oool0oooo0?ooo`3o_g@004RL0=ooo`3oool0gia8
0000B02Lgol0H@3oool00`2oM00007Bo0?ooo`1K0?ooo`800000H`3oool00`2oM00007Bo0?ooo`09
0?ooo`8000002`3oool6000004<0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00X0oooo
000=0?ooo`800000103oool01@3ogi`0B000000000000000M;oo0600oooo00D0oono07@000000000
001d0;ooo`1K0?ooo`030000003oool0oooo0600oooo00D0oono07@000000000001d0;ooo`0;0?oo
o`<000003@3oool6000003l0oooo00D0omnL04P000000000000007Boo`0;0?ooo`003@3oool00`00
0000oooo0?ooo`380?ooo`030000003oool0oooo07<0oooo0P00000A0?ooo`L00000B03oool000d0
oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`1e0?ooo`8000004`3oool;00000400
oooo000=0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3oool0M`3oool2000001X0oooo
3P00000d0?ooo`00303oool200000<X0oooo00<000000?ooo`3oool0N@3oool200000280oooo4000
000X0?ooo`00303oool200000<X0oooo00<000000?ooo`3oool0N`3oool2000002T0oooo4`00000L
0?ooo`00303oool200000<X0oooo00<000000?ooo`3oool0O@3oool00`000000oooo0?ooo`0i0?oo
o`h000003P3oool000`0oooo00<000000?ooo`3oool0b@3oool00`000000oooo0?ooo`1n0?ooo`03
0000003oool0oooo04D0oooo0`00000<0?ooo`00303oool00`000000oooo0?ooo`390?ooo`030000
003oool0oooo07l0oooo00<000000?ooo`3oool0AP3oool2000000/0oooo000;0?ooo`800000b`3o
ool00`000000oooo0?ooo`1o0?ooo`800000B03oool00`000000oooo0?ooo`090?ooo`002`3oool2
00000</0oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`160?ooo`030000003oool0
oooo00T0oooo000;0?ooo`800000b`3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo
04H0oooo00<000000?ooo`3oool02@3oool000030?ooo`00000000000?l00000[@0000010?ooo`00
2`3oool2000000030?ooo`000000oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0000000/0oooo000;0?ooo`80000000<0oooo0000003oool0H`3o
ool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`150?ooo`030000003oool0000000/0oooo000<0?ooo`800000bP3oool0
0`000000oooo0?ooo`1o0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool02@3oool0
00d0oooo0`0000380?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0A`3oool00`00
0000oooo0?ooo`090?ooo`003`3oool=00000;`0oooo00<000000?ooo`3oool0OP3oool00`000000
oooo0?ooo`170?ooo`8000002`3oool001T0oooo3P00002a0?ooo`030000003oool0oooo07d0oooo
00<000000?ooo`3oool0B03oool00`000000oooo0?ooo`0:0?ooo`009P3ooolB00000:00oooo00<0
00000?ooo`3oool0O@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo00X0oooo000b
0?oooa400000U@3oool00`000000oooo0?ooo`1l0?ooo`030000003oool0oooo04P0oooo0P00000<
0?ooo`00@`3oool7000008h0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?ooo`190?oo
o`030000003oool0oooo00/0oooo001:0?ooo`@00000RP3oool00`000000oooo0?ooo`1k0?ooo`03
0000003oool0oooo04T0oooo00<000000?ooo`3oool02`3oool004h0oooo200000220?ooo`030000
003oool0oooo07X0oooo00<000000?ooo`3oool0B@3oool2000000d0oooo001E0?ooo`P00000N`3o
ool2000007X0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0<0?ooo`00F`3oool6
000007L0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0
oooo00`0oooo001O0?ooo`H00000L`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo
04/0oooo0P00000>0?ooo`00H`3oool500000700oooo00<000000?ooo`3oool0MP3oool00`000000
oooo0?ooo`1<0?ooo`030000003oool0oooo00d0oooo001W0?ooo`D00000K03oool00`000000oooo
0?ooo`1d0?ooo`800000C`3oool00`000000oooo0?ooo`0=0?ooo`00J`3oool4000006T0oooo00<0
00000?ooo`3oool0LP3oool200000540oooo00<000000?ooo`3oool03@3oool006l0oooo1000001U
0?ooo`030000003oool0oooo0700oooo0P00001B0?ooo`030000003oool0oooo00h0oooo001c0?oo
o`<00000HP3oool00`000000oooo0?ooo`1^0?ooo`800000E03oool00`000000oooo0?ooo`0>0?oo
o`00M@3oool4000005l0oooo00<000000?ooo`3oool0K03oool2000005H0oooo00<000000?ooo`3o
ool03P3oool007P0oooo1000001L0?ooo`030000003oool0oooo06X0oooo0P00001G0?ooo`030000
003oool0oooo00l0oooo001j0?ooo`@00000FP3oool00`000000oooo0?ooo`1X0?ooo`800000F@3o
ool00`000000oooo0?ooo`0?0?ooo`00O@3oool4000005L0oooo00<000000?ooo`3oool0I@3oool3
000005X0oooo00<000000?ooo`3oool0403oool00800oooo1000001D0?ooo`800000H`3oool30000
05d0oooo00<000000?ooo`3oool0403oool00880oooo1000001B0?ooo`030000003oool0oooo0<40
oooo00<000000?ooo`3oool04@3oool008@0oooo1000001@0?ooo`030000003oool0oooo0<40oooo
00<000000?ooo`3oool04@3oool008H0oooo0`00001?0?ooo`030000003oool0oooo0<40oooo00<0
00000?ooo`3oool04@3oool008P0oooo0`00001=0?ooo`030000003oool0oooo0<00oooo00<00000
0?ooo`3oool04P3oool008X0oooo0`00001;0?ooo`030000003oool0oooo0<00oooo00<000000?oo
o`3oool04P3oool008`0oooo0P00001:0?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3o
ool04`3oool008h0oooo0P0000180?ooo`030000003oool0oooo0;l0oooo00<000000?ooo`3oool0
4`3oool00900oooo0P0000160?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool0503o
ool00940oooo0`0000140?ooo`030000003oool0oooo0;h0oooo00<000000?ooo`3oool0503oool0
09<0oooo0`0000120?ooo`030000003oool0oooo0;d0oooo0P00000F0?ooo`00U@3oool300000400
oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0E0?ooo`00UP3oool4000003h0oooo
00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0E0?ooo`00V03oool010000000oooo0000
0000000l0?ooo`800000_@3oool00`000000oooo0?ooo`0F0?ooo`00V@3oool2000000030?ooo`00
0000000003X0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`0F0?ooo`00V`3oool2
000000030?ooo`000000000003P0oooo00<000000?ooo`3oool0^`3oool2000001P0oooo002M0?oo
o`030000003oool0oooo00800000=P3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo
01L0oooo002N0?ooo`8000000P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0;X0
oooo0P00000I0?ooo`00X03oool00`000000oooo0?ooo`02000003<0oooo00<000000?ooo`3oool0
^P3oool00`000000oooo0?ooo`0H0?ooo`00X@3oool200000080oooo00<000000?ooo`3oool0<03o
ool00`000000oooo0?ooo`2j0?ooo`030000003oool0oooo01P0oooo002S0?ooo`030000003oool0
oooo00800000<03oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo01T0oooo002T0?oo
o`8000000P3oool00`000000oooo0?ooo`0K0?ooo`030=nLB0000000000000<0000000<0B9cO0?oo
o`3oool02@3oool00`000000oooo0?ooo`2i0?ooo`030000003oool0oooo01T0oooo002V0?ooo`80
000000<0oooo000000000000703oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00T0
oooo00<000000?ooo`3oool0^03oool2000001/0oooo002X0?ooo`040000003oool0oooo000001`0
oooo00<0okmd001d_`3oool03@3oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01X0
oooo002Y0?ooo`80000000<0oooo0000000000006`3oool00`3o_g@007Bo0?ooo`0<0?ooo`030000
003oool0oooo0;L0oooo0P00000L0?ooo`00Z`3oool010000000oooo0?ooo`00000K0?ooo`030?no
M018W=l0oooo00/0oooo0`00002g0?ooo`030000003oool0oooo01/0oooo002/0?ooo`80000000<0
oooo0000000000006P3oool00`3OW4P0B9cO0?ooo`0:0?ooo`030000003oool0oooo0;H0oooo0P00
000M0?ooo`00[P3oool2000000030?ooo`000000oooo01T0oooo00<0_g@007Boo`3oool02@3oool0
0`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01`0oooo002`0?ooo`040000003oool00000
000001P0oooo00<0omnL04P0M02oool02@3oool00`000000oooo0?ooo`2e0?ooo`8000007P3oool0
0;40oooo0P0000000`3oool000000?ooo`0F0?ooo`030?oog`2LB000M;oo00T0oooo00<000000?oo
o`3oool0]@3oool00`000000oooo0?ooo`0M0?ooo`00/`3oool00`000000oooo0000000F0?ooo`03
0?oo_`1d0000B9cO00T0oooo00<000000?ooo`3oool0]03oool2000001l0oooo002d0?ooo`@00000
403oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol02@3oool00`000000oooo0?ooo`2d0?oo
o`030000003oool0oooo01h0oooo002f0?ooo`<00000403oool01@3ogi`0B000000000000000M;oo
00X0oooo00<000000?ooo`3oool0/`3oool200000200oooo002h0?ooo`<000007@3oool00`000000
oooo0?ooo`2c0?ooo`030000003oool0oooo01l0oooo002i0?ooo`<00000703oool00`000000oooo
0?ooo`2b0?ooo`8000008@3oool00;/0oooo0`00000J0?ooo`030000003oool0oooo0;80oooo00<0
00000?ooo`3oool0803oool00;`0oooo0`00000I0?ooo`030000003oool0oooo0;40oooo00<00000
0?ooo`3oool08@3oool00;h0oooo0`00000G0?ooo`030000003oool0oooo0;00oooo0P00000S0?oo
o`00`03oool2000001H0oooo0P00002a0?ooo`030000003oool0oooo0280oooo00310?ooo`<00000
503oool00`000000oooo0?ooo`2_0?ooo`800000903oool00<<0oooo0P00000C0?ooo`030000003o
ool0oooo0:l0oooo00<000000?ooo`3oool08`3oool00<@0oooo0`00000A0?ooo`030000003oool0
oooo0:h0oooo00<000000?ooo`3oool0903oool00<H0oooo0P00000@0?ooo`030000003oool0oooo
0:h0oooo00<000000?ooo`3oool0903oool00<L0oooo0`00000>0?ooo`030000003oool0oooo0:d0
oooo00<000000?ooo`3oool09@3oool00<P0oooo0`00000=0?ooo`030000003oool0oooo0:d0oooo
00<000000?ooo`3oool09@3oool00<X0oooo0P00000<0?ooo`030000003oool0oooo0:`0oooo00<0
00000?ooo`3oool09P3oool00</0oooo0`00000:0?ooo`030000003oool0oooo0:`0oooo00<00000
0?ooo`3oool09P3oool00<d0oooo0P0000090?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool09`3oool00<h0oooo0P0000080?ooo`030000003oool0oooo0:X0oooo0P00000Y0?ooo`00
c`3oool3000000H0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0X0?ooo`00d@3o
ool2000000D0oooo0P00002Z0?ooo`030000003oool0oooo02T0oooo003B0?ooo`800000103oool0
0`000000oooo0?ooo`2X0?ooo`800000:`3oool00=<0oooo0`0000020?ooo`030000003oool0oooo
0:P0oooo00<000000?ooo`3oool0:P3oool00=D0oooo0P0000000`3oool000000?ooo`2X0?ooo`80
0000;03oool00=H0oooo0`00002X0?ooo`800000;@3oool00=P0oooo0P00002V0?ooo`800000;P3o
ool00=P0oooo0`00002U0?ooo`030000003oool0oooo02d0oooo003H0?ooo`040000003oool00000
00000:<0oooo00<000000?ooo`3oool0;P3oool00=P0oooo00@000000?ooo`3oool0oooo0P00002P
0?ooo`800000<03oool00=P0oooo00<000000?ooo`3oool00P3oool2000009h0oooo0P00000a0?oo
o`00f03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo09X0oooo0P00000b0?ooo`00
f03oool00`000000oooo0?ooo`050?ooo`800000VP3oool00`000000oooo0?ooo`0a0?ooo`00f03o
ool00`000000oooo0?ooo`060?ooo`800000V03oool00`000000oooo0?ooo`0b0?ooo`00f03oool2
000000T0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`0c0?ooo`00f03oool00`00
0000oooo0?ooo`090?ooo`800000T`3oool2000003D0oooo003H0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0T03oool00`000000oooo0?ooo`0d0?ooo`00f03oool00`000000oooo
0?ooo`0<0?ooo`800000S`3oool00`000000oooo0?ooo`0e0?ooo`00f03oool00`000000oooo0?oo
o`0=0?ooo`800000S@3oool00`000000oooo0?ooo`0f0?ooo`00f03oool00`000000oooo0?ooo`0?
0?ooo`800000RP3oool00`000000oooo0?ooo`0g0?ooo`00f03oool00`000000oooo0?ooo`0@0?oo
o`800000R03oool2000003T0oooo00390?ooo`030?oog`2LB000B9cO00`0oooo00<000000?ooo`3o
ool04P3oool2000008H0oooo00<000000?ooo`3oool0>03oool00<X0oooo00<0gia80018W03Oool0
2`3oool00`000000oooo0?ooo`0C0?ooo`800000Q03oool00`000000oooo0?ooo`0i0?ooo`00bP3o
ool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0oooo01@0oooo0`0000210?ooo`030000003o
ool0oooo03X0oooo00360?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0203oool0
0`000000oooo0?ooo`0F0?ooo`800000O`3oool00`000000oooo0?ooo`0k0?ooo`00aP3oool00`3o
oml0W4Q809cOo`020?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3oool05`3oool30000
07`0oooo00<000000?ooo`3oool0?03oool00<L0oooo00L0oono07@0M02oool0oooo0?noM000B9`0
gooo00X0oooo0`00000I0?ooo`800000NP3oool00`000000oooo0?ooo`0m0?ooo`00b03oool01P3o
gi`0B4RL0=ooo`3ogi`0B01809cOo`X0oooo00<000000?ooo`3oool06P3oool3000007L0oooo0P00
000o0?ooo`00b@3oool01@3o_g@007Bo0?oog`2LB000B9cO00X0oooo00<000000?ooo`3oool06`3o
ool3000007H0oooo00<000000?ooo`3oool0?P3oool00<X0oooo00@0gia804RLg`2oM00007Bo2P3o
ool00`000000oooo0?ooo`0M0?ooo`<00000L`3oool00`000000oooo0?ooo`0o0?ooo`00b`3oool0
103OW4P0M7Ad0000M02oool90?ooo`030000003oool0oooo01h0oooo1000001`0?ooo`030000003o
ool0oooo0400oooo003<0?ooo`030;md000004P0W=oo00T0oooo00<000000?ooo`3oool0803oool3
000006h0oooo00<000000?ooo`3oool0@@3oool00<`0oooo00<0oooO09a80018W=l02@3oool00`00
0000oooo0?ooo`0R0?ooo`<00000J`3oool00`000000oooo0?ooo`120?ooo`00f03oool00`000000
oooo0?ooo`0T0?ooo`<00000J03oool00`000000oooo0?ooo`130?ooo`00f03oool00`000000oooo
0?ooo`0V0?ooo`800000I@3oool2000004H0oooo003H0?ooo`030000003oool0oooo02P0oooo0P00
001R0?ooo`800000A`3oool00=P0oooo00<000000?ooo`3oool0:P3oool2000005l0oooo00<00000
0?ooo`3oool0A`3oool00=P0oooo00<000000?ooo`3oool0:`3oool3000005`0oooo00<000000?oo
o`3oool0B03oool00=P0oooo00<000000?ooo`3oool0;@3oool2000005X0oooo00<000000?ooo`3o
ool0B@3oool00=P0oooo0P00000`0?ooo`800000EP3oool2000004`0oooo003H0?ooo`030000003o
ool0oooo0340oooo0P00001C0?ooo`800000C@3oool00=P0oooo00<000000?ooo`3oool0<`3oool2
000004l0oooo0P00001?0?ooo`00f03oool00`000000oooo0?ooo`0e0?ooo`800000B`3oool30000
0500oooo003H0?ooo`030000003oool0oooo03L0oooo0P0000180?ooo`<00000D@3oool00=P0oooo
00<000000?ooo`3oool0>@3oool3000004<0oooo0`00001C0?ooo`00f03oool00`000000oooo0?oo
o`0k0?ooo`<00000?`3oool3000005D0oooo003H0?ooo`030000003oool0oooo03d0oooo0`00000k
0?ooo`<00000E`3oool00=P0oooo00<000000?ooo`3oool0@03oool2000003L0oooo0`00001I0?oo
o`00f03oool00`000000oooo0?ooo`120?ooo`<00000<@3oool4000005/0oooo003H0?ooo`030000
003oool0oooo04@0oooo1000000/0?ooo`@00000G@3oool00=P0oooo00<000000?ooo`3oool0A`3o
ool5000002@0oooo1000001Q0?ooo`00f03oool2000004`0oooo1@00000J0?ooo`H00000I@3oool0
0=P0oooo00<000000?ooo`3oool0C`3oool7000000d0oooo2@00001Y0?ooo`00f03oool00`000000
oooo0?ooo`1D0?oooa@00000K@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<0
00000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?oo
o`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0
e@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.13866, -5.3237, \
0.0148369, 0.0600818}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseCurve[0.1, 0, 3, 20];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.500011 0.235359 0.309307 0.0581752 [
[.02929 .29681 -9.34375 -15.6875 ]
[.02929 .29681 9.34375 0 ]
[.26465 .29681 -9.34375 -15.6875 ]
[.26465 .29681 9.34375 0 ]
[.73537 .29681 -5 -15.6875 ]
[.73537 .29681 5 0 ]
[.97073 .29681 -5 -15.6875 ]
[.97073 .29681 5 0 ]
[.48751 .07661 -18.6875 -7.84375 ]
[.48751 .07661 0 7.84375 ]
[.48751 .19296 -18.6875 -7.84375 ]
[.48751 .19296 0 7.84375 ]
[.48751 .42566 -10 -7.84375 ]
[.48751 .42566 0 7.84375 ]
[.48751 .54201 -10 -7.84375 ]
[.48751 .54201 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02929 .30931 m
.02929 .31556 L
s
gsave
.02929 .29681 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.26465 .30931 m
.26465 .31556 L
s
gsave
.26465 .29681 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.73537 .30931 m
.73537 .31556 L
s
gsave
.73537 .29681 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.97073 .30931 m
.97073 .31556 L
s
gsave
.97073 .29681 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.07636 .30931 m
.07636 .31306 L
s
.12344 .30931 m
.12344 .31306 L
s
.17051 .30931 m
.17051 .31306 L
s
.21758 .30931 m
.21758 .31306 L
s
.31172 .30931 m
.31172 .31306 L
s
.35879 .30931 m
.35879 .31306 L
s
.40587 .30931 m
.40587 .31306 L
s
.45294 .30931 m
.45294 .31306 L
s
.54708 .30931 m
.54708 .31306 L
s
.59415 .30931 m
.59415 .31306 L
s
.64123 .30931 m
.64123 .31306 L
s
.6883 .30931 m
.6883 .31306 L
s
.78244 .30931 m
.78244 .31306 L
s
.82951 .30931 m
.82951 .31306 L
s
.87659 .30931 m
.87659 .31306 L
s
.92366 .30931 m
.92366 .31306 L
s
.25 Mabswid
0 .30931 m
1 .30931 L
s
.50001 .07661 m
.50626 .07661 L
s
gsave
.48751 .07661 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.50001 .19296 m
.50626 .19296 L
s
gsave
.48751 .19296 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.50001 .42566 m
.50626 .42566 L
s
gsave
.48751 .42566 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.50001 .54201 m
.50626 .54201 L
s
gsave
.48751 .54201 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.50001 .10569 m
.50376 .10569 L
s
.50001 .13478 m
.50376 .13478 L
s
.50001 .16387 m
.50376 .16387 L
s
.50001 .22204 m
.50376 .22204 L
s
.50001 .25113 m
.50376 .25113 L
s
.50001 .28022 m
.50376 .28022 L
s
.50001 .3384 m
.50376 .3384 L
s
.50001 .36748 m
.50376 .36748 L
s
.50001 .39657 m
.50376 .39657 L
s
.50001 .45475 m
.50376 .45475 L
s
.50001 .48383 m
.50376 .48383 L
s
.50001 .51292 m
.50376 .51292 L
s
.50001 .04752 m
.50376 .04752 L
s
.50001 .01843 m
.50376 .01843 L
s
.50001 .5711 m
.50376 .5711 L
s
.50001 .60018 m
.50376 .60018 L
s
.25 Mabswid
.50001 0 m
.50001 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.52355 .30931 m
.52354 .30916 L
.52352 .30901 L
.52348 .30883 L
.52342 .30865 L
.52328 .30829 L
.52304 .30785 L
.52215 .30659 L
.52076 .30504 L
.50411 .29126 L
.44175 .24812 L
.41223 .23024 L
.37158 .20951 L
.35788 .20381 L
.34418 .19888 L
.33027 .19473 L
.32333 .19301 L
.31695 .19164 L
.30411 .18954 L
.29698 .18878 L
.29023 .18833 L
.28285 .18815 L
.27605 .18829 L
.26862 .1888 L
.2608 .18973 L
.24853 .19207 L
.24171 .19386 L
.23551 .1958 L
.22465 .19994 L
.21318 .20542 L
.19486 .2167 L
.17786 .23028 L
.16257 .24553 L
.14967 .26124 L
.14132 .27342 L
.13477 .28488 L
.13093 .29327 L
.12945 .29725 L
.12843 .30063 L
.128 .30234 L
.12764 .30407 L
.12725 .30685 L
.12713 .30937 L
.12724 .31144 L
.1275 .31317 L
.12789 .31464 L
.12844 .3161 L
.12918 .31751 L
.13009 .31884 L
.13108 .31997 L
Mistroke
.1332 .32181 L
.13617 .32363 L
.1393 .32506 L
.14584 .32719 L
.15372 .32905 L
.16303 .33082 L
.20167 .33764 L
.22897 .34355 L
.26026 .3522 L
.30295 .36797 L
.37246 .40407 L
.4877 .4844 L
.52797 .51406 L
.57796 .54839 L
.62672 .57638 L
.6567 .58958 L
.68596 .59868 L
.71302 .60318 L
.74169 .60323 L
.76599 .59898 L
.79192 .58963 L
.81877 .57418 L
.84545 .55242 L
.88694 .5044 L
.92358 .44547 L
.94688 .39807 L
.96915 .34107 L
.97183 .33229 L
.97384 .32471 L
.97499 .31944 L
.97576 .3148 L
.97602 .31245 L
.97614 .31054 L
.9761 .30743 L
.9758 .30521 L
.97526 .30335 L
.97457 .30193 L
.97367 .30073 L
.97272 .29983 L
.97178 .29917 L
.97066 .29859 L
.96938 .29807 L
.96798 .29765 L
.96664 .29734 L
.96414 .29691 L
.96172 .2966 L
.95706 .29619 L
.95189 .29585 L
.94614 .29552 L
.92228 .29418 L
Mistroke
.89305 .29219 L
.86123 .28937 L
.82583 .28509 L
.80294 .28144 L
.77758 .27636 L
.75035 .26943 L
.71296 .25695 L
.6527 .22892 L
.55924 .16865 L
.49687 .12264 L
.40912 .06203 L
.38304 .04714 L
.3535 .033 L
.33589 .02623 L
.3195 .02121 L
.28678 .01531 L
.25716 .01539 L
.22532 .02209 L
.19626 .03493 L
.17083 .05203 L
.14633 .07419 L
.12595 .09723 L
.08809 .15244 L
.04641 .23597 L
.0343 .26725 L
.02802 .28678 L
.026 .29446 L
.02482 .3 L
.02408 .30488 L
.02388 .30709 L
.02381 .30891 L
.02391 .31157 L
.02428 .31387 L
.02456 .31483 L
.02491 .31573 L
.02567 .31709 L
.02655 .31816 L
.02748 .31896 L
.02838 .31955 L
.02941 .32007 L
.03071 .32058 L
.03193 .32094 L
.03338 .32128 L
.03478 .32154 L
.03734 .32191 L
.04017 .3222 L
.04567 .32261 L
.05094 .32293 L
.06333 .3236 L
.09032 .32523 L
Mistroke
.11974 .32746 L
.15061 .33052 L
.18896 .33578 L
.21537 .34071 L
.24476 .34781 L
.28064 .35926 L
.32571 .37851 L
.39992 .42161 L
.46242 .46575 L
.54319 .52503 L
.59281 .55774 L
.61967 .5729 L
.64573 .58528 L
.67717 .59649 L
.69414 .60057 L
.71229 .60322 L
.74184 .60332 L
.77395 .59677 L
.79179 .5898 L
.80822 .58109 L
.8374 .5598 L
.8658 .53126 L
.89255 .49657 L
.92918 .43499 L
.9512 .38833 L
.96317 .35843 L
.97076 .33608 L
.97333 .32695 L
.97485 .32039 L
.97572 .31543 L
.97615 .31131 L
.97619 .30816 L
.97591 .30554 L
.9754 .30362 L
.97479 .30223 L
.97435 .30152 L
.97391 .30094 L
.97281 .29986 L
.97174 .29913 L
.97054 .29851 L
.96941 .29807 L
.96812 .29768 L
.96696 .2974 L
.9657 .29715 L
.96324 .29678 L
.96091 .29652 L
.95502 .29605 L
.94953 .29571 L
.92745 .29449 L
.89962 .29268 L
Mistroke
.87067 .2903 L
.8377 .28669 L
.81852 .28401 L
.79455 .2799 L
.76652 .27375 L
.73513 .26479 L
.68688 .24603 L
.61812 .20862 L
.50719 .1303 L
.46512 .09947 L
.41199 .06379 L
.38477 .04806 L
.35441 .03339 L
.3273 .02344 L
.29677 .01648 L
.26799 .01472 L
.24192 .01769 L
.21243 .02695 L
.18578 .04129 L
.16251 .05891 L
.14261 .07807 L
.10399 .12716 L
.07419 .17724 L
.05166 .22376 L
.03708 .25958 L
.02942 .28203 L
.02709 .29015 L
.02537 .29728 L
.02434 .30287 L
.02403 .30538 L
.02386 .30745 L
.02384 .31057 L
.02407 .31285 L
.02456 .31483 L
.02527 .31646 L
.026 .31754 L
.02688 .31847 L
.02779 .31918 L
.02867 .31971 L
.02975 .32022 L
.03098 .32067 L
.03332 .32127 L
.03466 .32152 L
.03593 .32172 L
.03889 .32208 L
.04184 .32234 L
.04508 .32257 L
.05668 .32324 L
.11097 .32673 L
.11164 .32679 L
Mistroke
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003of03oool0
0`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000
oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?oo
o`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00K@3ooolB000005T0oooo00<000000?oo
o`3oool0e@3oool006P0oooo1`00000;0?ooo`T00000E@3oool00`000000oooo0?ooo`3E0?ooo`00
I@3oool4000001X0oooo1@00001@0?ooo`800000eP3oool00640oooo1000000R0?ooo`D00000C03o
ool00`000000oooo0?ooo`3E0?ooo`00GP3oool3000002/0oooo100000180?ooo`030000003oool0
oooo0=D0oooo001L0?ooo`<00000<@3oool3000004D0oooo00<000000?ooo`3oool0e@3oool005X0
oooo0`00000f0?ooo`800000@`3oool00`000000oooo0?ooo`3E0?ooo`00F@3oool2000003X0oooo
0`0000100?ooo`030000003oool0oooo0=D0oooo001G0?ooo`800000?`3oool3000003d0oooo00<0
00000?ooo`3oool0e@3oool005D0oooo0P0000130?ooo`<00000>`3oool00`000000oooo0?ooo`3E
0?ooo`00D`3oool2000004L0oooo0`00000i0?ooo`030000003oool0oooo0=D0oooo001B0?ooo`80
0000BP3oool3000003L0oooo00<000000?ooo`3oool0e@3oool00500oooo0P00001>0?ooo`<00000
=@3oool00`000000oooo0?ooo`3E0?ooo`00CP3oool300000540oooo0`00000c0?ooo`030000003o
ool0oooo0=D0oooo001=0?ooo`800000E@3oool300000340oooo0P00003F0?ooo`00B`3oool30000
05P0oooo0`00000_0?ooo`030000003oool0oooo0=D0oooo001:0?ooo`<00000F`3oool3000002d0
oooo00<000000?ooo`3oool0e@3oool004T0oooo0P00001O0?ooo`800000;03oool00`000000oooo
0?ooo`3E0?ooo`00B03oool200000680oooo0P00000Z0?ooo`030000003oool0oooo0=D0oooo0017
0?ooo`030000003oool0oooo06@0oooo0P00000X0?ooo`030000003oool0oooo0=D0oooo00160?oo
o`030000003oool0oooo06L0oooo0P00000V0?ooo`030000003oool0oooo0=D0oooo00150?ooo`03
0000003oool0oooo06X0oooo0P00000T0?ooo`030000003oool0oooo0=D0oooo00140?ooo`030000
003oool0oooo06`0oooo0P00000D0?ooo`030?oog`2LB000B9cO00`0oooo00<000000?ooo`3oool0
e@3oool00480oooo0P00001a0?ooo`030000003oool0oooo0180oooo00<0gia80018W03Oool02`3o
ool00`000000oooo0?ooo`3E0?ooo`00@@3oool2000007<0oooo0P00000B0?ooo`030?oOW01804P0
W=oo00/0oooo00<000000?ooo`3oool0e@3oool00400oooo0P00001f0?ooo`030000003oool0oooo
00/0oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`080?ooo`030000003oool0oooo
0=D0oooo000o0?ooo`800000N03oool00`000000W4P0000000070000000604RLg`3oool0oooo0?oo
g`2LB4P0W=oo0P3oool00`2oM00007Bo0?ooo`0:0?ooo`030000003oool0oooo0=D0oooo000n0?oo
o`800000NP3oool2000000/0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo00X0oooo0`00
003E0?ooo`00?@3oool2000007d0oooo00<000000?ooo`3oool02@3oool01P3ogi`0B4RL0=ooo`3o
gi`0B01809cOo`X0oooo00<000000?ooo`3oool0e@3oool003`0oooo0P00001o0?ooo`8000002P3o
ool01@3o_g@007Bo0?oog`2LB000B9cO00X0oooo00<000000?ooo`3oool0e@3oool003`0oooo00<0
00000?ooo`3oool0P03oool00`000000oooo0?ooo`080?ooo`040=nLB018W=l0_g@0001d_`X0oooo
00<000000?ooo`3oool0e@3oool003/0oooo00<000000?ooo`3oool0PP3oool2000000T0oooo00@0
gia807AdM00007@0_ooo2@3oool00`000000oooo0?ooo`3E0?ooo`00>P3oool2000008D0oooo0P00
00090?ooo`030;md000004P0W=oo00T0oooo00<000000?ooo`3oool0e@3oool003T0oooo0P000028
0?ooo`030000003oool0oooo00H0oooo00<0oooO09a80018W=l02@3oool00`000000oooo0?ooo`3E
0?ooo`00>@3oool00`000000oooo0?ooo`280?ooo`8000004P3oool00`000000oooo0?ooo`3E0?oo
o`00>03oool00`000000oooo0?ooo`2:0?ooo`8000004@3oool00`000000oooo0?ooo`3E0?ooo`00
=`3oool00`000000oooo0?ooo`2=0?ooo`8000003`3oool00`000000oooo0?ooo`3E0?ooo`00=P3o
ool00`000000oooo0?ooo`2?0?ooo`8000003P3oool00`000000oooo0?ooo`3E0?ooo`00=P3oool0
0`000000oooo0?ooo`2@0?ooo`<00000303oool00`000000oooo0?ooo`3E0?ooo`00=@3oool00`00
0000oooo0?ooo`2C0?ooo`8000002`3oool200000=H0oooo000d0?ooo`030000003oool0oooo09D0
oooo0P00000:0?ooo`030000003oool0oooo0=D0oooo000c0?ooo`800000V@3oool2000000P0oooo
00<000000?ooo`3oool0e@3oool003<0oooo00<000000?ooo`3oool0V@3oool2000000L0oooo00<0
00000?ooo`3oool0e@3oool00380oooo00<000000?ooo`3oool0V`3oool2000000H0oooo00<00000
0?ooo`3oool0e@3oool00340oooo00<000000?ooo`3oool0WP3oool2000000@0oooo00<000000?oo
o`3oool0e@3oool00300oooo0P00002Q0?ooo`8000000`3oool00`000000oooo0?ooo`3E0?ooo`00
<03oool00`000000oooo0?ooo`2Q0?ooo`<0000000<0oooo0000003oool0eP3oool002l0oooo00<0
00000?ooo`3oool0Y03oool300000=L0oooo000^0?ooo`800000Y`3oool200000=L0oooo000]0?oo
o`800000Z@3oool300000=D0oooo000]0?ooo`030000003oool0oooo0:P0oooo00@000000?ooo`00
00000000e03oool002`0oooo0P00002Z0?ooo`030000003oool0oooo00<00000dP3oool002`0oooo
00<000000?ooo`3oool0Z@3oool2000000<0oooo0P00003A0?ooo`00:`3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo00<0oooo0P00003@0?ooo`00:P3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00D0oooo0P00003>0?ooo`00:P3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00H0oooo0P00003=0?ooo`00:@3oool00`000000oooo0?ooo`2/0?ooo`03
0000003oool0oooo00P0oooo0P00003;0?ooo`00:03oool200000:h0oooo00<000000?ooo`3oool0
2@3oool200000<X0oooo000X0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3oool02P3o
ool300000<P0oooo000W0?ooo`030000003oool0oooo0:h0oooo00<000000?ooo`3oool0303oool2
00000<L0oooo000V0?ooo`800000/03oool00`000000oooo0?ooo`0=0?ooo`800000aP3oool002H0
oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`0?0?ooo`800000a03oool002D0oooo
0P00002a0?ooo`030000003oool0oooo0100oooo0P0000330?ooo`009@3oool00`000000oooo0?oo
o`2`0?ooo`030000003oool0oooo0180oooo0P0000310?ooo`00903oool200000;80oooo0P00000D
0?ooo`800000`03oool002@0oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`0D0?oo
o`<00000_P3oool002<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0F0?ooo`80
0000_@3oool002<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0G0?ooo`800000
_03oool00280oooo00<000000?ooo`3oool0/`3oool00`000000oooo0?ooo`0I0?ooo`800000^P3o
ool00280oooo00<000000?ooo`3oool0/`3oool00`000000oooo0?ooo`0J0?ooo`800000^@3oool0
0240oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0L0?ooo`800000]`3oool00200
oooo0P00002f0?ooo`030000003oool0oooo01d0oooo0P00002f0?ooo`00803oool00`000000oooo
0?ooo`2S0?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool02@3oool00`000000oooo
0?ooo`0O0?ooo`800000]03oool001l0oooo0P00002V0?ooo`030?noM000M;l0oooo0080oooo00<0
_gAd0;ooo`3oool02@3oool00`000000oooo0?ooo`0P0?ooo`800000/`3oool001l0oooo00<00000
0?ooo`3oool0YP3oool00`3o_g@007Bo0?ooo`0=0?ooo`030000003oool0oooo0280oooo0P00002a
0?ooo`007P3oool2000004l0oooo600000120?ooo`030?noM000M;l0oooo00`0oooo00<000000?oo
o`3oool08`3oool200000;00oooo000N0?ooo`030000003oool0oooo04X0oooo1000000H0?ooo`@0
0000;P3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0okmd04RLg`3o
ool02`3oool00`000000oooo0?ooo`0U0?ooo`800000[P3oool001d0oooo0P0000190?ooo`<00000
803oool6000003X0oooo00<0gia804RLg`3oool02P3oool3000002H0oooo0P00002]0?ooo`007@3o
ool00`000000oooo0?ooo`150?ooo`<00000:@3oool3000003P0oooo00<0_g@007Boo`3oool02@3o
ool00`000000oooo0?ooo`0X0?ooo`800000Z`3oool001`0oooo0P0000150?ooo`800000;`3oool3
000003D0oooo00<0omnL04P0M02oool02@3oool00`000000oooo0?ooo`0Y0?ooo`800000ZP3oool0
01`0oooo00<000000?ooo`3oool0@P3oool2000003@0oooo0`00000b0?ooo`030?oog`2LB000M;oo
00T0oooo00<000000?ooo`3oool0:`3oool200000:P0oooo000K0?ooo`800000@P3oool2000003T0
oooo0P00000`0?ooo`030?oo_`1d0000B9cO00T0oooo00<000000?ooo`3oool0;03oool200000:L0
oooo000K0?ooo`030000003oool0oooo03l0oooo0P00000m0?ooo`800000:P3oool01`3o_g@004RL
0=ooo`3oool0gia80000B02Lgol02@3oool00`000000oooo0?ooo`0]0?ooo`<00000Y@3oool001X0
oooo0P00000o0?ooo`800000@@3oool2000002T0oooo00D0omnL04P000000000000007Boo`0:0?oo
o`030000003oool0oooo02l0oooo0P00002T0?ooo`006P3oool00`000000oooo0?ooo`0m0?ooo`03
0000003oool0oooo04<0oooo0P00000f0?ooo`030000003oool0oooo0300oooo0`00002R0?ooo`00
6@3oool2000003d0oooo0P0000180?ooo`800000=03oool00`000000oooo0?ooo`0b0?ooo`800000
X@3oool001T0oooo00<000000?ooo`3oool0>P3oool2000004`0oooo0P00000b0?ooo`030000003o
ool0oooo03@0oooo0P00002O0?ooo`00603oool2000003/0oooo00<000000?ooo`3oool0CP3oool2
00000300oooo00<000000?ooo`3oool0=P3oool2000009d0oooo000H0?ooo`030000003oool0oooo
03T0oooo00<000000?ooo`3oool0D@3oool2000002h0oooo00<000000?ooo`3oool0=`3oool20000
09`0oooo000G0?ooo`800000>@3oool2000005H0oooo0P00000/0?ooo`800000>P3oool2000009X0
oooo000G0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0F03oool2000002X0oooo
00<000000?ooo`3oool0>`3oool2000009P0oooo000G0?ooo`030000003oool0oooo03H0oooo00<0
00000?ooo`3oool0F`3oool2000002P0oooo00<000000?ooo`3oool0?03oool2000009L0oooo000F
0?ooo`030000003oool0oooo03D0oooo0P00001P0?ooo`8000009P3oool00`000000oooo0?ooo`0n
0?ooo`800000U@3oool001H0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`1R0?oo
o`030000003oool0oooo02<0oooo00<000000?ooo`3oool0@03oool2000009<0oooo000E0?ooo`03
0000003oool0oooo03@0oooo00<000000?ooo`3oool0I03oool2000002<0oooo00<000000?ooo`3o
ool0@P3oool200000940oooo000E0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0
I`3oool200000240oooo00<000000?ooo`3oool0A03oool2000008l0oooo000E0?ooo`030000003o
ool0oooo0380oooo00<000000?ooo`3oool0JP3oool2000001l0oooo00<000000?ooo`3oool0AP3o
ool2000008d0oooo000D0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0K@3oool2
000001d0oooo00<000000?ooo`3oool0A`3oool3000008/0oooo000D0?ooo`030000003oool0oooo
0340oooo00<000000?ooo`3oool0L03oool2000001/0oooo00<000000?ooo`3oool0B@3oool40000
08P0oooo000C0?ooo`800000<P3oool00`000000oooo0?ooo`1c0?ooo`8000006@3oool00`000000
oooo0?ooo`1;0?ooo`@00000QP3oool001<0oooo00<000000?ooo`3oool0<03oool00`000000oooo
0?ooo`1f0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0C@3oool4000008@0oooo
000C0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0N03oool2000001H0oooo00<0
00000?ooo`3oool0C`3oool400000880oooo000B0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool0N`3oool00`000000oooo0?ooo`0C0?ooo`800000DP3oool400000800oooo000B0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo05@0oooo0`00001n0?ooo`004@3oool200000300oooo00<000000?ooo`3o
ool0OP3oool200000180oooo00<000000?ooo`3oool0E`3oool3000007/0oooo000@0?ooo`030=nL
B0000000000000<0000000<0B9cO0?ooo`3oool0:@3oool00`000000oooo0?ooo`0`0?ooo`030?oo
_`1d000000000080000000<0B9cO0?ooo`3oool0B@3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo05T0oooo0`0000000`3oool0oono07@000030000000304RLg`3oool0oooo05d0oooo
00<0gia80000000000000`0000000`18W=l0oooo0?ooo`0:0?ooo`004@3oool00`00000007Bo0?oo
o`020?ooo`030;mdM02oool0oooo02P0oooo00<000000?ooo`3oool0<`3oool00`3o_g@0000007Bo
o`1=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0G03oool3000000030?noM000
0000M;oo0640oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`0:0?ooo`00403oool2
000000030?noM000M;l0oooo02/0oooo00<000000?ooo`3oool0=03oool00`3ooml0W4P0001d_`1>
0?ooo`8000003P3oool00`000000oooo0?ooo`1O0?ooo`<00000HP3oool00`3o_g@007Bo0?ooo`0>
0?ooo`00403oool01@000000oooo0?ooo`3o_g@007Bo02/0oooo00<000000?ooo`3oool0=@3oool0
0`3OW4P004RL0=ooo`1?0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0H03oool0
0`3OW4P000000000000200000600oooo00<0okmd001d_`3oool03@3oool000<0oooo00<0oooO09a8
000000001`0000001018W=l0oooo0?ooo`0000030?ooo`030?noM018W=l0oooo02P0oooo00<00000
0?ooo`3oool09P3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0omnL
04P0B02Lgol0D03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0600oooo00@0omnL
04P0B02Lgol0oooo1@00001M0?ooo`030?noM018W=l0oooo00`0oooo000?0?ooo`800000103oool0
0`3OW4P0B9cO0?ooo`0W0?ooo`030000003oool0oooo03H0oooo00<0oono07@0001d_ol0D@3oool2
000000X0oooo00<000000?ooo`3oool0H03oool00`3ookl0M00007Boo`040?ooo`H00000FP3oool0
0`3OW4P0B9cO0?ooo`0;0?ooo`003`3oool00`000000oooo0?ooo`040?ooo`030;md001d_ol0oooo
02D0oooo00<000000?ooo`3oool0>03oool00`2oM00007Bo0?ooo`1B0?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0H@3oool00`2oM00007Bo0?ooo`060?ooo`L00000E`3oool00`2o
M000M;oo0?ooo`0:0?ooo`003`3oool00`000000oooo0?ooo`040?ooo`030?oOW01807@0_ooo02@0
oooo00<000000?ooo`3oool0>@3oool00`3OW4P004RL0=ooo`1C0?ooo`8000001`3oool00`000000
oooo0?ooo`1Q0?ooo`030=nLB000B9`0gooo00X0oooo1`00001C0?ooo`030?oOW01807@0_ooo00X0
oooo000>0?ooo`8000001P3oool00`3ooml0W4P007Boo`0T0?ooo`030000003oool0oooo03T0oooo
00<0omnL04P0B02Lgol0E@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0640oooo
00<0omnL04P0B02Lgol03P3oool7000004l0oooo00<0oooO09a8001d_ol02P3oool000h0oooo00<0
00000?ooo`3oool01@3oool00`3ookl0M00004RLg`0S0?ooo`030000003oool0oooo03X0oooo00<0
oono07@0001d_ol0EP3oool00`000000oooo0?ooo`030?ooo`800000HP3oool00`3ookl0M00007Bo
o`0C0?ooo`L00000BP3oool00`3ookl0M00004RLg`0:0?ooo`003P3oool02`000000oooo0?ooo`3o
ool0okmd0018W03Oool0oooo0=nLB00004P0W=oo02<0oooo00<000000?ooo`3oool0>`3oool00`2o
M00007Bo0?ooo`1F0?ooo`8000000`3oool00`000000oooo0?ooo`1R0?ooo`030;md0000M;l0oooo
01P0oooo1`0000100?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`0:0?ooo`003@3oool2
000000@0oooo00D0omnL04P000000000000007Boo`0S0?ooo`030000003oool0oooo03X0oooo00D0
oono07@000000000001d0;ooo`1H0?ooo`040000003oool0oooo00000680oooo00D0oono07@00000
0000001d0;ooo`0O0?ooo`T00000>03oool01@3ogi`0B000000000000000M;oo00/0oooo000=0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0000002<0?oo
o`/00000@03oool000d0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`2J0?ooo`80
0000U03oool=000003H0oooo000<0?ooo`800000:`3oool00`000000oooo0?ooo`2L0?ooo`800000
VP3ooolC000002T0oooo000<0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0W03o
ool00`000000oooo0000002R0?oooaL00000703oool000`0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`2L0?ooo`030000003oool0oooo00800000]@3oool?000000l0oooo000<0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool0W03oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0<00oooo0`00000=0?ooo`002`3oool2000002/0oooo00<000000?ooo`3o
ool0W@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3o
ool02P3oool000/0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`2M0?ooo`030000
003oool0oooo00@0oooo0P0000330?ooo`030000003oool0oooo00T0oooo000;0?ooo`030000003o
ool0oooo02X0oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo0<00oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool0:P3oool0
0`000000oooo0?ooo`2M0?ooo`030000003oool0oooo00H0oooo0P0000310?ooo`030000003oool0
oooo00T0oooo00000`3oool000000000003o00000:d000000@3oool000/0oooo00@000000?ooo`3o
ool000004`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool000000140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool04P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0000000/0oooo000;0?ooo`80
000000<0oooo0000003oool0:03oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo0680
oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`1R0?ooo`030000003oool0000000/0
oooo000<0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0W@3oool00`000000oooo
0?ooo`390?ooo`030000003oool0oooo00T0oooo000<0?ooo`<00000:P3oool00`000000oooo0?oo
o`2L0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool02@3oool000h0oooo2@00000S
0?ooo`030000003oool0oooo09/0oooo00<000000?ooo`3oool0b03oool2000000/0oooo000F0?oo
o`l000005P3oool4000009T0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`0:0?oo
o`008P3ooolF000000L0oooo0`00002F0?ooo`030000003oool0oooo0<P0oooo00<000000?ooo`3o
ool02P3oool003P0oooo1`0000030?ooo`<00000T`3oool00`000000oooo0?ooo`370?ooo`800000
303oool003l0oooo2`00002>0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02`3o
ool004H0oooo2P0000280?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02`3oool0
04h0oooo200000220?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool02`3oool005D0
oooo1P00001m0?ooo`800000a`3oool2000000d0oooo001K0?ooo`H00000M`3oool00`000000oooo
0?ooo`360?ooo`030000003oool0oooo00`0oooo001P0?ooo`D00000L`3oool00`000000oooo0?oo
o`360?ooo`030000003oool0oooo00`0oooo001T0?ooo`D00000K`3oool00`000000oooo0?ooo`35
0?ooo`8000003P3oool006P0oooo1000001/0?ooo`030000003oool0oooo0<D0oooo00<000000?oo
o`3oool03@3oool006`0oooo1000001X0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3o
ool03@3oool006l0oooo1000001U0?ooo`030000003oool0oooo0<@0oooo0P00000?0?ooo`00LP3o
ool3000006<0oooo00<000000?ooo`3oool0a03oool2000000l0oooo001e0?ooo`<00000H03oool0
0`000000oooo0?ooo`330?ooo`800000403oool007P0oooo0`00001M0?ooo`030000003oool0oooo
0<<0oooo0P00000@0?ooo`00NP3oool3000005/0oooo00<000000?ooo`3oool0``3oool200000100
oooo001m0?ooo`<00000F03oool00`000000oooo0?ooo`320?ooo`8000004@3oool00800oooo0P00
001F0?ooo`030000003oool0oooo0<80oooo0P00000A0?ooo`00PP3oool3000005<0oooo0P000033
0?ooo`030000003oool0oooo0100oooo00240?ooo`<00000D@3oool00`000000oooo0?ooo`310?oo
o`8000004P3oool008H0oooo1000001>0?ooo`030000003oool0oooo0<40oooo0P00000B0?ooo`00
R03oool4000004`0oooo00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0A0?ooo`00RP3o
ool4000004X0oooo00<000000?ooo`3oool0`03oool2000001<0oooo002<0?ooo`@00000B03oool0
0`000000oooo0?ooo`300?ooo`8000004`3oool008h0oooo100000160?ooo`030000003oool0oooo
0<00oooo00<000000?ooo`3oool04P3oool00900oooo0`0000150?ooo`030000003oool0oooo0;l0
oooo0P00000D0?ooo`00TP3oool3000004<0oooo00<000000?ooo`3oool0_`3oool00`000000oooo
0?ooo`0C0?ooo`00U03oool300000440oooo00<000000?ooo`3oool0_P3oool2000001D0oooo002F
0?ooo`800000@03oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo01@0oooo002H0?oo
o`800000?P3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo01@0oooo002J0?ooo`80
0000?03oool200000;h0oooo00<000000?ooo`3oool05@3oool009`0oooo0P00000j0?ooo`030000
003oool0oooo0;d0oooo00<000000?ooo`3oool05@3oool009d0oooo0`00000h0?ooo`030000003o
ool0oooo0;`0oooo00<000000?ooo`3oool05P3oool009l0oooo0`00000f0?ooo`030000003oool0
oooo0;`0oooo00<000000?ooo`3oool05P3oool00:40oooo0`00000d0?ooo`030000003oool0oooo
0;/0oooo00<000000?ooo`3oool05`3oool00:<0oooo0P00000c0?ooo`030000003oool0oooo0;/0
oooo00<000000?ooo`3oool05`3oool00:@0oooo0P00000b0?ooo`030000003oool0oooo0;X0oooo
00<000000?ooo`3oool0603oool00:H0oooo0P00000`0?ooo`030000003oool0oooo0;X0oooo00<0
00000?ooo`3oool0603oool00:P0oooo00<000000?ooo`3oool06`3oool00`3OW4P0000000000003
0000000304RLg`3oool0oooo00T0oooo00<000000?ooo`3oool0^@3oool00`000000oooo0?ooo`0I
0?ooo`00Z@3oool2000001`0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`090?oo
o`030000003oool0oooo0;T0oooo00<000000?ooo`3oool06@3oool00:/0oooo0P00000K0?ooo`03
0?noM000M;l0oooo00d0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`0J0?ooo`00
[03oool2000001/0oooo00<0okmd001d_`3oool0303oool00`000000oooo0?ooo`2h0?ooo`030000
003oool0oooo01X0oooo002]0?ooo`<000006P3oool00`3o_g@0B9cO0?ooo`0;0?ooo`030000003o
ool0oooo0;L0oooo00<000000?ooo`3oool06`3oool00:l0oooo0P00000J0?ooo`030=nLB018W=l0
oooo00X0oooo0`00002g0?ooo`030000003oool0oooo01/0oooo002`0?ooo`8000006P3oool00`2o
M000M;oo0?ooo`090?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0703oool00;80
oooo0P00000H0?ooo`030?oOW01807@0_ooo00T0oooo00<000000?ooo`3oool0]P3oool00`000000
oooo0?ooo`0L0?ooo`00/`3oool2000001L0oooo00<0oooO09a8001d_ol02@3oool00`000000oooo
0?ooo`2e0?ooo`030000003oool0oooo01d0oooo002d0?ooo`<000005@3oool00`3ookl0M00004RL
g`090?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool07@3oool00;H0oooo0P00000@
0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`090?ooo`030000003oool0oooo0;@0oooo
00<000000?ooo`3oool07P3oool00;L0oooo0`00000?0?ooo`050?oOW018000000000000001d_ol0
2P3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo01h0oooo002i0?ooo`8000007@3o
ool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01l0oooo002j0?ooo`800000703oool0
0`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01l0oooo002k0?ooo`<000006P3oool00`00
0000oooo0?ooo`2b0?ooo`030000003oool0oooo0200oooo002m0?ooo`8000006@3oool00`000000
oooo0?ooo`2a0?ooo`8000008P3oool00;h0oooo0`00000G0?ooo`030000003oool0oooo0;40oooo
00<000000?ooo`3oool08@3oool00<00oooo0P00000F0?ooo`800000/@3oool00`000000oooo0?oo
o`0R0?ooo`00`@3oool3000001@0oooo00<000000?ooo`3oool0[`3oool2000002@0oooo00320?oo
o`<000004`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo02<0oooo00340?ooo`80
00004P3oool00`000000oooo0?ooo`2^0?ooo`8000009@3oool00<D0oooo0`00000@0?ooo`030000
003oool0oooo0:h0oooo00<000000?ooo`3oool0903oool00<L0oooo0P00000?0?ooo`030000003o
ool0oooo0:d0oooo00<000000?ooo`3oool09@3oool00<P0oooo0`00000=0?ooo`030000003oool0
oooo0:`0oooo0P00000W0?ooo`00b@3oool3000000`0oooo00<000000?ooo`3oool0[03oool00`00
0000oooo0?ooo`0V0?ooo`00b`3oool2000000/0oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`0W0?ooo`00c03oool3000000T0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo
0?ooo`0W0?ooo`00cP3oool2000000P0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`0X0?ooo`00c`3oool2000000L0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`0Y
0?ooo`00d03oool3000000D0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`0Y0?oo
o`00dP3oool2000000@0oooo0P00002Y0?ooo`030000003oool0oooo02X0oooo003C0?ooo`800000
0`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo02/0oooo003E0?ooo`80000000<0
oooo0000003oool0Z03oool00`000000oooo0?ooo`0[0?ooo`00eP3oool300000:P0oooo00<00000
0?ooo`3oool0;03oool00=L0oooo0P00002X0?ooo`030000003oool0oooo02`0oooo003H0?ooo`<0
0000Y@3oool00`000000oooo0?ooo`0]0?ooo`00f03oool010000000oooo00000000002S0?ooo`03
0000003oool0oooo02h0oooo003H0?ooo`030000003oool0oooo00<00000X@3oool00`000000oooo
0?ooo`0^0?ooo`00f03oool00`000000oooo0?ooo`020?ooo`800000W`3oool00`000000oooo0?oo
o`0_0?ooo`00f03oool00`000000oooo0?ooo`030?ooo`800000W@3oool00`000000oooo0?ooo`0`
0?ooo`00f03oool00`000000oooo0?ooo`050?ooo`800000VP3oool00`000000oooo0?ooo`0a0?oo
o`00f03oool00`000000oooo0?ooo`060?ooo`800000V03oool2000003<0oooo003H0?ooo`800000
203oool2000009L0oooo00<000000?ooo`3oool0<P3oool00=P0oooo00<000000?ooo`3oool02@3o
ool2000009@0oooo00<000000?ooo`3oool0<`3oool00=P0oooo00<000000?ooo`3oool02P3oool2
00000980oooo00<000000?ooo`3oool0=03oool00=P0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`2>0?ooo`800000=P3oool00=P0oooo00<000000?ooo`3oool03@3oool2000008d0
oooo0P00000g0?ooo`00f03oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo08T0oooo
0P00000h0?ooo`00f03oool00`000000oooo0?ooo`0@0?ooo`800000R03oool2000003T0oooo003H
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0Q@3oool2000003T0oooo00390?oo
o`030?oog`2LB000B9cO00`0oooo00<000000?ooo`3oool04`3oool2000008@0oooo0P00000j0?oo
o`00bP3oool00`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo01D0oooo0P0000210?ooo`80
0000>`3oool00<X0oooo00<0omnL04P0B02Lgol02`3oool00`000000oooo0?ooo`0F0?ooo`800000
O`3oool2000003`0oooo00360?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0203o
ool00`000000oooo0?ooo`0H0?ooo`800000O03oool2000003d0oooo00360?ooo`030?oog`2LB4P0
W=oo0080oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`0I0?ooo`800000NP3oool2
000003h0oooo00370?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`0:0?ooo`<000006`3o
ool2000007L0oooo0P00000o0?ooo`00b03oool01P3ogi`0B4RL0=ooo`3ogi`0B01809cOo`X0oooo
00<000000?ooo`3oool0703oool2000007H0oooo00<000000?ooo`3oool0?P3oool00<T0oooo00D0
okmd001d_`3ooml0W4P004RLg`0:0?ooo`030000003oool0oooo01h0oooo0P00001c0?ooo`030000
003oool0oooo03l0oooo003:0?ooo`040=nLB018W=l0_g@0001d_`X0oooo00<000000?ooo`3oool0
7`3oool300000700oooo00<000000?ooo`3oool0@03oool00</0oooo00@0gia807AdM00007@0_ooo
2@3oool00`000000oooo0?ooo`0Q0?ooo`800000KP3oool00`000000oooo0?ooo`110?ooo`00c03o
ool00`2oM000001809cOo`090?ooo`030000003oool0oooo02<0oooo0P00001[0?ooo`030000003o
ool0oooo0480oooo003<0?ooo`030?oog`2LB000B9cO00T0oooo00<000000?ooo`3oool09@3oool0
0`000000oooo0?ooo`1V0?ooo`800000A@3oool00=P0oooo00<000000?ooo`3oool09P3oool20000
06D0oooo0P0000160?ooo`00f03oool00`000000oooo0?ooo`0X0?ooo`800000HP3oool2000004L0
oooo003H0?ooo`030000003oool0oooo02T0oooo0`00001O0?ooo`030000003oool0oooo04L0oooo
003H0?ooo`030000003oool0oooo02/0oooo0P00001L0?ooo`800000BP3oool00=P0oooo00<00000
0?ooo`3oool0;@3oool2000005T0oooo0P00001;0?ooo`00f03oool2000002l0oooo0`00001F0?oo
o`030000003oool0oooo04/0oooo003H0?ooo`030000003oool0oooo0300oooo0`00001B0?ooo`80
0000CP3oool00=P0oooo00<000000?ooo`3oool0<P3oool3000004l0oooo0P00001?0?ooo`00f03o
ool00`000000oooo0?ooo`0d0?ooo`<00000B`3oool200000540oooo003H0?ooo`030000003oool0
oooo03H0oooo0`0000180?ooo`030000003oool0oooo0540oooo003H0?ooo`030000003oool0oooo
03P0oooo0`0000140?ooo`800000E03oool00=P0oooo00<000000?ooo`3oool0>P3oool300000400
oooo0P00001F0?ooo`00f03oool00`000000oooo0?ooo`0m0?ooo`<00000>`3oool2000005P0oooo
003H0?ooo`030000003oool0oooo03l0oooo0`00000g0?ooo`800000FP3oool00=P0oooo00<00000
0?ooo`3oool0@P3oool300000340oooo0`00001L0?ooo`00f03oool00`000000oooo0?ooo`150?oo
o`@00000:`3oool3000005h0oooo003H0?ooo`030000003oool0oooo04P0oooo1000000T0?ooo`@0
0000H@3oool00=P0oooo00<000000?ooo`3oool0B`3oool5000001`0oooo1@00001T0?ooo`00f03o
ool200000500oooo1`00000B0?ooo`D00000J03oool00=P0oooo00<000000?ooo`3oool0E03ooolD
000006d0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0
oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H
0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.13866, -5.32468, \
0.0148366, 0.0600243}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseAndField[0.1, 0, 3, 20, 3, 5];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: 1.67657 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.157814 0.839071 0.157814 [
[ 0 0 0 0 ]
[ 1 1.67657 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto fill
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto
-0.02 0.005 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
0 0 Mdot
grestore
} def
% End of user PostScript
0 g
.0001 w
[ ] 0 setdash
.02656 .05 m
.02381 .11758 L
s
% Start of user PostScript
-3. -5. -3.0174072 -4.5717822 MAarrowhead1
% End of user PostScript
.02656 .16272 m
.0242 .22088 L
s
% Start of user PostScript
-3. -4.2857143 -3.0149205 -3.9171784 MAarrowhead1
% End of user PostScript
.02656 .27545 m
.02459 .32419 L
s
% Start of user PostScript
-3. -3.5714286 -3.0124337 -3.2625746 MAarrowhead1
% End of user PostScript
.02656 .38817 m
.02499 .4275 L
s
% Start of user PostScript
-3. -2.8571429 -3.009947 -2.6079708 MAarrowhead1
% End of user PostScript
.02656 .5009 m
.02538 .5308 L
s
% Start of user PostScript
-3. -2.1428571 -3.0074602 -1.953367 MAarrowhead1
% End of user PostScript
.02656 .61362 m
.02577 .63411 L
s
% Start of user PostScript
-3. -1.4285714 -3.0049735 -1.2987633 MAarrowhead1
% End of user PostScript
.02656 .72635 m
.02616 .73741 L
s
% Start of user PostScript
-3. -0.71428571 -3.0024867 -0.64415946 MAarrowhead1
% End of user PostScript
.02656 .83907 m
.02656 .84072 L
s
% Start of user PostScript
-3. 0. -3. 0.010444336 MAarrowhead1
% End of user PostScript
.02656 .9518 m
.02695 .94403 L
s
% Start of user PostScript
-3. 0.71428571 -2.9975133 0.66504813 MAarrowhead1
% End of user PostScript
.02656 1.06452 m
.02734 1.04733 L
s
% Start of user PostScript
-3. 1.4285714 -2.9950265 1.3196519 MAarrowhead1
% End of user PostScript
.02656 1.17725 m
.02773 1.15064 L
s
% Start of user PostScript
-3. 2.1428571 -2.9925398 1.9742557 MAarrowhead1
% End of user PostScript
.02656 1.28997 m
.02813 1.25394 L
s
% Start of user PostScript
-3. 2.8571429 -2.990053 2.6288595 MAarrowhead1
% End of user PostScript
.02656 1.40269 m
.02852 1.35725 L
s
% Start of user PostScript
-3. 3.5714286 -2.9875663 3.2834633 MAarrowhead1
% End of user PostScript
.02656 1.51542 m
.02891 1.46056 L
s
% Start of user PostScript
-3. 4.2857143 -2.9850795 3.9380671 MAarrowhead1
% End of user PostScript
.02656 1.62814 m
.0293 1.56386 L
s
% Start of user PostScript
-3. 5. -2.9825928 4.5926709 MAarrowhead1
% End of user PostScript
.09419 .05 m
.09144 .09766 L
s
% Start of user PostScript
-2.5714286 -5. -2.5888358 -4.6979669 MAarrowhead1
% End of user PostScript
.09419 .16272 m
.09184 .20378 L
s
% Start of user PostScript
-2.5714286 -4.2857143 -2.5863491 -4.0255498 MAarrowhead1
% End of user PostScript
.09419 .27545 m
.09223 .3099 L
s
% Start of user PostScript
-2.5714286 -3.5714286 -2.5838623 -3.3531328 MAarrowhead1
% End of user PostScript
.09419 .38817 m
.09262 .41602 L
s
% Start of user PostScript
-2.5714286 -2.8571429 -2.5813756 -2.6807158 MAarrowhead1
% End of user PostScript
.09419 .5009 m
.09301 .52213 L
s
% Start of user PostScript
-2.5714286 -2.1428571 -2.5788888 -2.0082988 MAarrowhead1
% End of user PostScript
.09419 .61362 m
.09341 .62825 L
s
% Start of user PostScript
-2.5714286 -1.4285714 -2.5764021 -1.3358818 MAarrowhead1
% End of user PostScript
.09419 .72635 m
.0938 .73437 L
s
% Start of user PostScript
-2.5714286 -0.71428571 -2.5739153 -0.66346473 MAarrowhead1
% End of user PostScript
.09419 .83907 m
.09419 .84048 L
s
% Start of user PostScript
-2.5714286 0. -2.5714286 0.0089522879 MAarrowhead1
% End of user PostScript
.09419 .9518 m
.09458 .9466 L
s
% Start of user PostScript
-2.5714286 0.71428571 -2.5689418 0.68136931 MAarrowhead1
% End of user PostScript
.09419 1.06452 m
.09498 1.05272 L
s
% Start of user PostScript
-2.5714286 1.4285714 -2.5664551 1.3537863 MAarrowhead1
% End of user PostScript
.09419 1.17725 m
.09537 1.15884 L
s
% Start of user PostScript
-2.5714286 2.1428571 -2.5639683 2.0262034 MAarrowhead1
% End of user PostScript
.09419 1.28997 m
.09576 1.26495 L
s
% Start of user PostScript
-2.5714286 2.8571429 -2.5614816 2.6986204 MAarrowhead1
% End of user PostScript
.09419 1.40269 m
.09615 1.37107 L
s
% Start of user PostScript
-2.5714286 3.5714286 -2.5589948 3.3710374 MAarrowhead1
% End of user PostScript
.09419 1.51542 m
.09655 1.47719 L
s
% Start of user PostScript
-2.5714286 4.2857143 -2.5565081 4.0434544 MAarrowhead1
% End of user PostScript
.09419 1.62814 m
.09694 1.5833 L
s
% Start of user PostScript
-2.5714286 5. -2.5540213 4.7158714 MAarrowhead1
% End of user PostScript
.16183 .05 m
.15908 .08078 L
s
% Start of user PostScript
-2.1428571 -5. -2.1602644 -4.804968 MAarrowhead1
% End of user PostScript
.16183 .16272 m
.15947 .18927 L
s
% Start of user PostScript
-2.1428571 -4.2857143 -2.1577776 -4.1174783 MAarrowhead1
% End of user PostScript
.16183 .27545 m
.15986 .29777 L
s
% Start of user PostScript
-2.1428571 -3.5714286 -2.1552909 -3.4299885 MAarrowhead1
% End of user PostScript
.16183 .38817 m
.16026 .40627 L
s
% Start of user PostScript
-2.1428571 -2.8571429 -2.1528041 -2.7424988 MAarrowhead1
% End of user PostScript
.16183 .5009 m
.16065 .51476 L
s
% Start of user PostScript
-2.1428571 -2.1428571 -2.1503174 -2.055009 MAarrowhead1
% End of user PostScript
.16183 .61362 m
.16104 .62326 L
s
% Start of user PostScript
-2.1428571 -1.4285714 -2.1478306 -1.3675193 MAarrowhead1
% End of user PostScript
.16183 .72635 m
.16143 .73175 L
s
% Start of user PostScript
-2.1428571 -0.71428571 -2.1453439 -0.68002951 MAarrowhead1
% End of user PostScript
.16183 .83907 m
.16183 .84025 L
s
% Start of user PostScript
-2.1428571 0. -2.1428571 0.0074602399 MAarrowhead1
% End of user PostScript
.16183 .9518 m
.16222 .94874 L
s
% Start of user PostScript
-2.1428571 0.71428571 -2.1403704 0.69494999 MAarrowhead1
% End of user PostScript
.16183 1.06452 m
.16261 1.05724 L
s
% Start of user PostScript
-2.1428571 1.4285714 -2.1378836 1.3824397 MAarrowhead1
% End of user PostScript
.16183 1.17725 m
.163 1.16574 L
s
% Start of user PostScript
-2.1428571 2.1428571 -2.1353969 2.0699295 MAarrowhead1
% End of user PostScript
.16183 1.28997 m
.1634 1.27423 L
s
% Start of user PostScript
-2.1428571 2.8571429 -2.1329102 2.7574192 MAarrowhead1
% End of user PostScript
.16183 1.40269 m
.16379 1.38273 L
s
% Start of user PostScript
-2.1428571 3.5714286 -2.1304234 3.444909 MAarrowhead1
% End of user PostScript
.16183 1.51542 m
.16418 1.49122 L
s
% Start of user PostScript
-2.1428571 4.2857143 -2.1279367 4.1323987 MAarrowhead1
% End of user PostScript
.16183 1.62814 m
.16457 1.59972 L
s
% Start of user PostScript
-2.1428571 5. -2.1254499 4.8198885 MAarrowhead1
% End of user PostScript
.22946 .05 m
.22671 .06692 L
s
% Start of user PostScript
-1.7142857 -5. -1.7316929 -4.8927857 MAarrowhead1
% End of user PostScript
.22946 .16272 m
.22711 .17736 L
s
% Start of user PostScript
-1.7142857 -4.2857143 -1.7292062 -4.1929637 MAarrowhead1
% End of user PostScript
.22946 .27545 m
.2275 .2878 L
s
% Start of user PostScript
-1.7142857 -3.5714286 -1.7267194 -3.4931417 MAarrowhead1
% End of user PostScript
.22946 .38817 m
.22789 .39825 L
s
% Start of user PostScript
-1.7142857 -2.8571429 -1.7242327 -2.7933197 MAarrowhead1
% End of user PostScript
.22946 .5009 m
.22828 .50869 L
s
% Start of user PostScript
-1.7142857 -2.1428571 -1.721746 -2.0934978 MAarrowhead1
% End of user PostScript
.22946 .61362 m
.22868 .61913 L
s
% Start of user PostScript
-1.7142857 -1.4285714 -1.7192592 -1.3936758 MAarrowhead1
% End of user PostScript
.22946 .72635 m
.22907 .72957 L
s
% Start of user PostScript
-1.7142857 -0.71428571 -1.7167725 -0.69385379 MAarrowhead1
% End of user PostScript
.22946 .83907 m
.22946 .84001 L
s
% Start of user PostScript
-1.7142857 0. -1.7142857 0.0059681919 MAarrowhead1
% End of user PostScript
.22946 .9518 m
.22985 .95046 L
s
% Start of user PostScript
-1.7142857 0.71428571 -1.711799 0.70579018 MAarrowhead1
% End of user PostScript
.22946 1.06452 m
.23025 1.0609 L
s
% Start of user PostScript
-1.7142857 1.4285714 -1.7093122 1.4056122 MAarrowhead1
% End of user PostScript
.22946 1.17725 m
.23064 1.17134 L
s
% Start of user PostScript
-1.7142857 2.1428571 -1.7068255 2.1054341 MAarrowhead1
% End of user PostScript
.22946 1.28997 m
.23103 1.28178 L
s
% Start of user PostScript
-1.7142857 2.8571429 -1.7043387 2.8052561 MAarrowhead1
% End of user PostScript
.22946 1.40269 m
.23142 1.39222 L
s
% Start of user PostScript
-1.7142857 3.5714286 -1.701852 3.5050781 MAarrowhead1
% End of user PostScript
.22946 1.51542 m
.23182 1.50267 L
s
% Start of user PostScript
-1.7142857 4.2857143 -1.6993652 4.2049001 MAarrowhead1
% End of user PostScript
.22946 1.62814 m
.23221 1.61311 L
s
% Start of user PostScript
-1.7142857 5. -1.6968785 4.9047221 MAarrowhead1
% End of user PostScript
.2971 .05 m
.29435 .05609 L
s
% Start of user PostScript
-1.2857143 -5. -1.3031215 -4.9614199 MAarrowhead1
% End of user PostScript
.2971 .16272 m
.29474 .16804 L
s
% Start of user PostScript
-1.2857143 -4.2857143 -1.3006348 -4.2520062 MAarrowhead1
% End of user PostScript
.2971 .27545 m
.29513 .28 L
s
% Start of user PostScript
-1.2857143 -3.5714286 -1.298148 -3.5425925 MAarrowhead1
% End of user PostScript
.2971 .38817 m
.29553 .39195 L
s
% Start of user PostScript
-1.2857143 -2.8571429 -1.2956613 -2.8331787 MAarrowhead1
% End of user PostScript
.2971 .5009 m
.29592 .50391 L
s
% Start of user PostScript
-1.2857143 -2.1428571 -1.2931745 -2.123765 MAarrowhead1
% End of user PostScript
.2971 .61362 m
.29631 .61587 L
s
% Start of user PostScript
-1.2857143 -1.4285714 -1.2906878 -1.4143513 MAarrowhead1
% End of user PostScript
.2971 .72635 m
.2967 .72782 L
s
% Start of user PostScript
-1.2857143 -0.71428571 -1.288201 -0.70493758 MAarrowhead1
% End of user PostScript
.2971 .83907 m
.2971 .83978 L
s
% Start of user PostScript
-1.2857143 0. -1.2857143 0.0044761439 MAarrowhead1
% End of user PostScript
.2971 .9518 m
.29749 .95173 L
s
% Start of user PostScript
-1.2857143 0.71428571 -1.2832275 0.71388986 MAarrowhead1
% End of user PostScript
.2971 1.06452 m
.29788 1.06369 L
s
% Start of user PostScript
-1.2857143 1.4285714 -1.2807408 1.4233036 MAarrowhead1
% End of user PostScript
.2971 1.17725 m
.29827 1.17564 L
s
% Start of user PostScript
-1.2857143 2.1428571 -1.278254 2.1327173 MAarrowhead1
% End of user PostScript
.2971 1.28997 m
.29867 1.2876 L
s
% Start of user PostScript
-1.2857143 2.8571429 -1.2757673 2.842131 MAarrowhead1
% End of user PostScript
.2971 1.40269 m
.29906 1.39956 L
s
% Start of user PostScript
-1.2857143 3.5714286 -1.2732806 3.5515447 MAarrowhead1
% End of user PostScript
.2971 1.51542 m
.29945 1.51151 L
s
% Start of user PostScript
-1.2857143 4.2857143 -1.2707938 4.2609585 MAarrowhead1
% End of user PostScript
.2971 1.62814 m
.29984 1.62347 L
s
% Start of user PostScript
-1.2857143 5. -1.2683071 4.9703722 MAarrowhead1
% End of user PostScript
.36473 .05 m
.36198 .04828 L
s
% Start of user PostScript
-0.85714286 -5. -0.87455008 -5.0108706 MAarrowhead1
% End of user PostScript
.36473 .16272 m
.36238 .16132 L
s
% Start of user PostScript
-0.85714286 -4.2857143 -0.87206334 -4.2946057 MAarrowhead1
% End of user PostScript
.36473 .27545 m
.36277 .27436 L
s
% Start of user PostScript
-0.85714286 -3.5714286 -0.86957659 -3.5783407 MAarrowhead1
% End of user PostScript
.36473 .38817 m
.36316 .38739 L
s
% Start of user PostScript
-0.85714286 -2.8571429 -0.86708984 -2.8620758 MAarrowhead1
% End of user PostScript
.36473 .5009 m
.36355 .50043 L
s
% Start of user PostScript
-0.85714286 -2.1428571 -0.8646031 -2.1458108 MAarrowhead1
% End of user PostScript
.36473 .61362 m
.36395 .61347 L
s
% Start of user PostScript
-0.85714286 -1.4285714 -0.86211635 -1.4295458 MAarrowhead1
% End of user PostScript
.36473 .72635 m
.36434 .72651 L
s
% Start of user PostScript
-0.85714286 -0.71428571 -0.8596296 -0.71328087 MAarrowhead1
% End of user PostScript
.36473 .83907 m
.36473 .83954 L
s
% Start of user PostScript
-0.85714286 0. -0.85714286 0.002984096 MAarrowhead1
% End of user PostScript
.36473 .9518 m
.36512 .95258 L
s
% Start of user PostScript
-0.85714286 0.71428571 -0.85465611 0.71924906 MAarrowhead1
% End of user PostScript
.36473 1.06452 m
.36552 1.06562 L
s
% Start of user PostScript
-0.85714286 1.4285714 -0.85216936 1.435514 MAarrowhead1
% End of user PostScript
.36473 1.17725 m
.36591 1.17865 L
s
% Start of user PostScript
-0.85714286 2.1428571 -0.84968262 2.151779 MAarrowhead1
% End of user PostScript
.36473 1.28997 m
.3663 1.29169 L
s
% Start of user PostScript
-0.85714286 2.8571429 -0.84719587 2.8680439 MAarrowhead1
% End of user PostScript
.36473 1.40269 m
.36669 1.40473 L
s
% Start of user PostScript
-0.85714286 3.5714286 -0.84470912 3.5843089 MAarrowhead1
% End of user PostScript
.36473 1.51542 m
.36709 1.51776 L
s
% Start of user PostScript
-0.85714286 4.2857143 -0.84222238 4.3005739 MAarrowhead1
% End of user PostScript
.36473 1.62814 m
.36748 1.6308 L
s
% Start of user PostScript
-0.85714286 5. -0.83973563 5.0168388 MAarrowhead1
% End of user PostScript
.43237 .05 m
.42962 .04351 L
s
% Start of user PostScript
-0.42857143 -5. -0.44597865 -5.0411379 MAarrowhead1
% End of user PostScript
.43237 .16272 m
.43001 .15719 L
s
% Start of user PostScript
-0.42857143 -4.2857143 -0.44349191 -4.3207622 MAarrowhead1
% End of user PostScript
.43237 .27545 m
.4304 .27088 L
s
% Start of user PostScript
-0.42857143 -3.5714286 -0.44100516 -3.6003865 MAarrowhead1
% End of user PostScript
.43237 .38817 m
.4308 .38456 L
s
% Start of user PostScript
-0.42857143 -2.8571429 -0.43851842 -2.8800108 MAarrowhead1
% End of user PostScript
.43237 .5009 m
.43119 .49825 L
s
% Start of user PostScript
-0.42857143 -2.1428571 -0.43603167 -2.1596351 MAarrowhead1
% End of user PostScript
.43237 .61362 m
.43158 .61194 L
s
% Start of user PostScript
-0.42857143 -1.4285714 -0.43354492 -1.4392594 MAarrowhead1
% End of user PostScript
.43237 .72635 m
.43197 .72562 L
s
% Start of user PostScript
-0.42857143 -0.71428571 -0.43105818 -0.71888366 MAarrowhead1
% End of user PostScript
.43237 .83907 m
.43237 .83931 L
s
% Start of user PostScript
-0.42857143 0. -0.42857143 0.001492048 MAarrowhead1
% End of user PostScript
.43237 .9518 m
.43276 .95299 L
s
% Start of user PostScript
-0.42857143 0.71428571 -0.42608468 0.72186775 MAarrowhead1
% End of user PostScript
.43237 1.06452 m
.43315 1.06668 L
s
% Start of user PostScript
-0.42857143 1.4285714 -0.42359794 1.4422435 MAarrowhead1
% End of user PostScript
.43237 1.17725 m
.43354 1.18036 L
s
% Start of user PostScript
-0.42857143 2.1428571 -0.42111119 2.1626192 MAarrowhead1
% End of user PostScript
.43237 1.28997 m
.43394 1.29405 L
s
% Start of user PostScript
-0.42857143 2.8571429 -0.41862444 2.8829949 MAarrowhead1
% End of user PostScript
.43237 1.40269 m
.43433 1.40774 L
s
% Start of user PostScript
-0.42857143 3.5714286 -0.4161377 3.6033706 MAarrowhead1
% End of user PostScript
.43237 1.51542 m
.43472 1.52142 L
s
% Start of user PostScript
-0.42857143 4.2857143 -0.41365095 4.3237463 MAarrowhead1
% End of user PostScript
.43237 1.62814 m
.43511 1.63511 L
s
% Start of user PostScript
-0.42857143 5. -0.4111642 5.044122 MAarrowhead1
% End of user PostScript
.5 .05 m
.49725 .04176 L
s
% Start of user PostScript
0. -5. -0.017407226 -5.0522217 MAarrowhead1
% End of user PostScript
.5 .16272 m
.49765 .15566 L
s
% Start of user PostScript
0. -4.2857143 -0.01492048 -4.3304757 MAarrowhead1
% End of user PostScript
.5 .27545 m
.49804 .26956 L
s
% Start of user PostScript
0. -3.5714286 -0.012433733 -3.6087298 MAarrowhead1
% End of user PostScript
.5 .38817 m
.49843 .38346 L
s
% Start of user PostScript
0. -2.8571429 -0.0099469865 -2.8869838 MAarrowhead1
% End of user PostScript
.5 .5009 m
.49882 .49737 L
s
% Start of user PostScript
0. -2.1428571 -0.0074602399 -2.1652379 MAarrowhead1
% End of user PostScript
.5 .61362 m
.49922 .61127 L
s
% Start of user PostScript
0. -1.4285714 -0.0049734933 -1.4434919 MAarrowhead1
% End of user PostScript
.5 .72635 m
.49961 .72517 L
s
% Start of user PostScript
0. -0.71428571 -0.0024867466 -0.72174595 MAarrowhead1
% End of user PostScript
.5 .83907 m
.5 .83907 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.5 .9518 m
.50039 .95297 L
s
% Start of user PostScript
0. 0.71428571 0.0024867466 0.72174595 MAarrowhead1
% End of user PostScript
.5 1.06452 m
.50078 1.06688 L
s
% Start of user PostScript
0. 1.4285714 0.0049734933 1.4434919 MAarrowhead1
% End of user PostScript
.5 1.17725 m
.50118 1.18078 L
s
% Start of user PostScript
0. 2.1428571 0.0074602399 2.1652379 MAarrowhead1
% End of user PostScript
.5 1.28997 m
.50157 1.29468 L
s
% Start of user PostScript
0. 2.8571429 0.0099469865 2.8869838 MAarrowhead1
% End of user PostScript
.5 1.40269 m
.50196 1.40858 L
s
% Start of user PostScript
0. 3.5714286 0.012433733 3.6087298 MAarrowhead1
% End of user PostScript
.5 1.51542 m
.50235 1.52248 L
s
% Start of user PostScript
0. 4.2857143 0.01492048 4.3304757 MAarrowhead1
% End of user PostScript
.5 1.62814 m
.50275 1.63638 L
s
% Start of user PostScript
0. 5. 0.017407226 5.0522217 MAarrowhead1
% End of user PostScript
.56763 .05 m
.56489 .04304 L
s
% Start of user PostScript
0.42857143 -5. 0.4111642 -5.044122 MAarrowhead1
% End of user PostScript
.56763 .16272 m
.56528 .15672 L
s
% Start of user PostScript
0.42857143 -4.2857143 0.41365095 -4.3237463 MAarrowhead1
% End of user PostScript
.56763 .27545 m
.56567 .27041 L
s
% Start of user PostScript
0.42857143 -3.5714286 0.4161377 -3.6033706 MAarrowhead1
% End of user PostScript
.56763 .38817 m
.56606 .38409 L
s
% Start of user PostScript
0.42857143 -2.8571429 0.41862444 -2.8829949 MAarrowhead1
% End of user PostScript
.56763 .5009 m
.56646 .49778 L
s
% Start of user PostScript
0.42857143 -2.1428571 0.42111119 -2.1626192 MAarrowhead1
% End of user PostScript
.56763 .61362 m
.56685 .61146 L
s
% Start of user PostScript
0.42857143 -1.4285714 0.42359794 -1.4422435 MAarrowhead1
% End of user PostScript
.56763 .72635 m
.56724 .72515 L
s
% Start of user PostScript
0.42857143 -0.71428571 0.42608468 -0.72186775 MAarrowhead1
% End of user PostScript
.56763 .83907 m
.56763 .83884 L
s
% Start of user PostScript
0.42857143 0. 0.42857143 -0.001492048 MAarrowhead1
% End of user PostScript
.56763 .9518 m
.56803 .95252 L
s
% Start of user PostScript
0.42857143 0.71428571 0.43105818 0.71888366 MAarrowhead1
% End of user PostScript
.56763 1.06452 m
.56842 1.06621 L
s
% Start of user PostScript
0.42857143 1.4285714 0.43354492 1.4392594 MAarrowhead1
% End of user PostScript
.56763 1.17725 m
.56881 1.17989 L
s
% Start of user PostScript
0.42857143 2.1428571 0.43603167 2.1596351 MAarrowhead1
% End of user PostScript
.56763 1.28997 m
.5692 1.29358 L
s
% Start of user PostScript
0.42857143 2.8571429 0.43851842 2.8800108 MAarrowhead1
% End of user PostScript
.56763 1.40269 m
.5696 1.40726 L
s
% Start of user PostScript
0.42857143 3.5714286 0.44100516 3.6003865 MAarrowhead1
% End of user PostScript
.56763 1.51542 m
.56999 1.52095 L
s
% Start of user PostScript
0.42857143 4.2857143 0.44349191 4.3207622 MAarrowhead1
% End of user PostScript
.56763 1.62814 m
.57038 1.63464 L
s
% Start of user PostScript
0.42857143 5. 0.44597865 5.0411379 MAarrowhead1
% End of user PostScript
.63527 .05 m
.63252 .04734 L
s
% Start of user PostScript
0.85714286 -5. 0.83973563 -5.0168388 MAarrowhead1
% End of user PostScript
.63527 .16272 m
.63291 .16038 L
s
% Start of user PostScript
0.85714286 -4.2857143 0.84222238 -4.3005739 MAarrowhead1
% End of user PostScript
.63527 .27545 m
.63331 .27342 L
s
% Start of user PostScript
0.85714286 -3.5714286 0.84470912 -3.5843089 MAarrowhead1
% End of user PostScript
.63527 .38817 m
.6337 .38645 L
s
% Start of user PostScript
0.85714286 -2.8571429 0.84719587 -2.8680439 MAarrowhead1
% End of user PostScript
.63527 .5009 m
.63409 .49949 L
s
% Start of user PostScript
0.85714286 -2.1428571 0.84968262 -2.151779 MAarrowhead1
% End of user PostScript
.63527 .61362 m
.63448 .61253 L
s
% Start of user PostScript
0.85714286 -1.4285714 0.85216936 -1.435514 MAarrowhead1
% End of user PostScript
.63527 .72635 m
.63488 .72556 L
s
% Start of user PostScript
0.85714286 -0.71428571 0.85465611 -0.71924906 MAarrowhead1
% End of user PostScript
.63527 .83907 m
.63527 .8386 L
s
% Start of user PostScript
0.85714286 0. 0.85714286 -0.002984096 MAarrowhead1
% End of user PostScript
.63527 .9518 m
.63566 .95164 L
s
% Start of user PostScript
0.85714286 0.71428571 0.8596296 0.71328087 MAarrowhead1
% End of user PostScript
.63527 1.06452 m
.63605 1.06467 L
s
% Start of user PostScript
0.85714286 1.4285714 0.86211635 1.4295458 MAarrowhead1
% End of user PostScript
.63527 1.17725 m
.63645 1.17771 L
s
% Start of user PostScript
0.85714286 2.1428571 0.8646031 2.1458108 MAarrowhead1
% End of user PostScript
.63527 1.28997 m
.63684 1.29075 L
s
% Start of user PostScript
0.85714286 2.8571429 0.86708984 2.8620758 MAarrowhead1
% End of user PostScript
.63527 1.40269 m
.63723 1.40379 L
s
% Start of user PostScript
0.85714286 3.5714286 0.86957659 3.5783407 MAarrowhead1
% End of user PostScript
.63527 1.51542 m
.63762 1.51682 L
s
% Start of user PostScript
0.85714286 4.2857143 0.87206334 4.2946057 MAarrowhead1
% End of user PostScript
.63527 1.62814 m
.63802 1.62986 L
s
% Start of user PostScript
0.85714286 5. 0.87455008 5.0108706 MAarrowhead1
% End of user PostScript
.7029 .05 m
.70016 .05467 L
s
% Start of user PostScript
1.2857143 -5. 1.2683071 -4.9703722 MAarrowhead1
% End of user PostScript
.7029 .16272 m
.70055 .16663 L
s
% Start of user PostScript
1.2857143 -4.2857143 1.2707938 -4.2609585 MAarrowhead1
% End of user PostScript
.7029 .27545 m
.70094 .27859 L
s
% Start of user PostScript
1.2857143 -3.5714286 1.2732806 -3.5515447 MAarrowhead1
% End of user PostScript
.7029 .38817 m
.70133 .39054 L
s
% Start of user PostScript
1.2857143 -2.8571429 1.2757673 -2.842131 MAarrowhead1
% End of user PostScript
.7029 .5009 m
.70173 .5025 L
s
% Start of user PostScript
1.2857143 -2.1428571 1.278254 -2.1327173 MAarrowhead1
% End of user PostScript
.7029 .61362 m
.70212 .61445 L
s
% Start of user PostScript
1.2857143 -1.4285714 1.2807408 -1.4233036 MAarrowhead1
% End of user PostScript
.7029 .72635 m
.70251 .72641 L
s
% Start of user PostScript
1.2857143 -0.71428571 1.2832275 -0.71388986 MAarrowhead1
% End of user PostScript
.7029 .83907 m
.7029 .83836 L
s
% Start of user PostScript
1.2857143 0. 1.2857143 -0.0044761439 MAarrowhead1
% End of user PostScript
.7029 .9518 m
.7033 .95032 L
s
% Start of user PostScript
1.2857143 0.71428571 1.288201 0.70493758 MAarrowhead1
% End of user PostScript
.7029 1.06452 m
.70369 1.06228 L
s
% Start of user PostScript
1.2857143 1.4285714 1.2906878 1.4143513 MAarrowhead1
% End of user PostScript
.7029 1.17725 m
.70408 1.17423 L
s
% Start of user PostScript
1.2857143 2.1428571 1.2931745 2.123765 MAarrowhead1
% End of user PostScript
.7029 1.28997 m
.70447 1.28619 L
s
% Start of user PostScript
1.2857143 2.8571429 1.2956613 2.8331787 MAarrowhead1
% End of user PostScript
.7029 1.40269 m
.70487 1.39814 L
s
% Start of user PostScript
1.2857143 3.5714286 1.298148 3.5425925 MAarrowhead1
% End of user PostScript
.7029 1.51542 m
.70526 1.5101 L
s
% Start of user PostScript
1.2857143 4.2857143 1.3006348 4.2520062 MAarrowhead1
% End of user PostScript
.7029 1.62814 m
.70565 1.62206 L
s
% Start of user PostScript
1.2857143 5. 1.3031215 4.9614199 MAarrowhead1
% End of user PostScript
.77054 .05 m
.76779 .06504 L
s
% Start of user PostScript
1.7142857 -5. 1.6968785 -4.9047221 MAarrowhead1
% End of user PostScript
.77054 .16272 m
.76818 .17548 L
s
% Start of user PostScript
1.7142857 -4.2857143 1.6993652 -4.2049001 MAarrowhead1
% End of user PostScript
.77054 .27545 m
.76858 .28592 L
s
% Start of user PostScript
1.7142857 -3.5714286 1.701852 -3.5050781 MAarrowhead1
% End of user PostScript
.77054 .38817 m
.76897 .39636 L
s
% Start of user PostScript
1.7142857 -2.8571429 1.7043387 -2.8052561 MAarrowhead1
% End of user PostScript
.77054 .5009 m
.76936 .5068 L
s
% Start of user PostScript
1.7142857 -2.1428571 1.7068255 -2.1054341 MAarrowhead1
% End of user PostScript
.77054 .61362 m
.76975 .61725 L
s
% Start of user PostScript
1.7142857 -1.4285714 1.7093122 -1.4056122 MAarrowhead1
% End of user PostScript
.77054 .72635 m
.77015 .72769 L
s
% Start of user PostScript
1.7142857 -0.71428571 1.711799 -0.70579018 MAarrowhead1
% End of user PostScript
.77054 .83907 m
.77054 .83813 L
s
% Start of user PostScript
1.7142857 0. 1.7142857 -0.0059681919 MAarrowhead1
% End of user PostScript
.77054 .9518 m
.77093 .94857 L
s
% Start of user PostScript
1.7142857 0.71428571 1.7167725 0.69385379 MAarrowhead1
% End of user PostScript
.77054 1.06452 m
.77132 1.05901 L
s
% Start of user PostScript
1.7142857 1.4285714 1.7192592 1.3936758 MAarrowhead1
% End of user PostScript
.77054 1.17725 m
.77172 1.16946 L
s
% Start of user PostScript
1.7142857 2.1428571 1.721746 2.0934978 MAarrowhead1
% End of user PostScript
.77054 1.28997 m
.77211 1.2799 L
s
% Start of user PostScript
1.7142857 2.8571429 1.7242327 2.7933197 MAarrowhead1
% End of user PostScript
.77054 1.40269 m
.7725 1.39034 L
s
% Start of user PostScript
1.7142857 3.5714286 1.7267194 3.4931417 MAarrowhead1
% End of user PostScript
.77054 1.51542 m
.77289 1.50078 L
s
% Start of user PostScript
1.7142857 4.2857143 1.7292062 4.1929637 MAarrowhead1
% End of user PostScript
.77054 1.62814 m
.77329 1.61122 L
s
% Start of user PostScript
1.7142857 5. 1.7316929 4.8927857 MAarrowhead1
% End of user PostScript
.83817 .05 m
.83543 .07842 L
s
% Start of user PostScript
2.1428571 -5. 2.1254499 -4.8198885 MAarrowhead1
% End of user PostScript
.83817 .16272 m
.83582 .18692 L
s
% Start of user PostScript
2.1428571 -4.2857143 2.1279367 -4.1323987 MAarrowhead1
% End of user PostScript
.83817 .27545 m
.83621 .29541 L
s
% Start of user PostScript
2.1428571 -3.5714286 2.1304234 -3.444909 MAarrowhead1
% End of user PostScript
.83817 .38817 m
.8366 .40391 L
s
% Start of user PostScript
2.1428571 -2.8571429 2.1329102 -2.7574192 MAarrowhead1
% End of user PostScript
.83817 .5009 m
.837 .51241 L
s
% Start of user PostScript
2.1428571 -2.1428571 2.1353969 -2.0699295 MAarrowhead1
% End of user PostScript
.83817 .61362 m
.83739 .6209 L
s
% Start of user PostScript
2.1428571 -1.4285714 2.1378836 -1.3824397 MAarrowhead1
% End of user PostScript
.83817 .72635 m
.83778 .7294 L
s
% Start of user PostScript
2.1428571 -0.71428571 2.1403704 -0.69494999 MAarrowhead1
% End of user PostScript
.83817 .83907 m
.83817 .83789 L
s
% Start of user PostScript
2.1428571 0. 2.1428571 -0.0074602399 MAarrowhead1
% End of user PostScript
.83817 .9518 m
.83857 .94639 L
s
% Start of user PostScript
2.1428571 0.71428571 2.1453439 0.68002951 MAarrowhead1
% End of user PostScript
.83817 1.06452 m
.83896 1.05489 L
s
% Start of user PostScript
2.1428571 1.4285714 2.1478306 1.3675193 MAarrowhead1
% End of user PostScript
.83817 1.17725 m
.83935 1.16338 L
s
% Start of user PostScript
2.1428571 2.1428571 2.1503174 2.055009 MAarrowhead1
% End of user PostScript
.83817 1.28997 m
.83974 1.27188 L
s
% Start of user PostScript
2.1428571 2.8571429 2.1528041 2.7424988 MAarrowhead1
% End of user PostScript
.83817 1.40269 m
.84014 1.38037 L
s
% Start of user PostScript
2.1428571 3.5714286 2.1552909 3.4299885 MAarrowhead1
% End of user PostScript
.83817 1.51542 m
.84053 1.48887 L
s
% Start of user PostScript
2.1428571 4.2857143 2.1577776 4.1174783 MAarrowhead1
% End of user PostScript
.83817 1.62814 m
.84092 1.59736 L
s
% Start of user PostScript
2.1428571 5. 2.1602644 4.804968 MAarrowhead1
% End of user PostScript
.90581 .05 m
.90306 .09484 L
s
% Start of user PostScript
2.5714286 -5. 2.5540213 -4.7158714 MAarrowhead1
% End of user PostScript
.90581 .16272 m
.90345 .20096 L
s
% Start of user PostScript
2.5714286 -4.2857143 2.5565081 -4.0434544 MAarrowhead1
% End of user PostScript
.90581 .27545 m
.90385 .30707 L
s
% Start of user PostScript
2.5714286 -3.5714286 2.5589948 -3.3710374 MAarrowhead1
% End of user PostScript
.90581 .38817 m
.90424 .41319 L
s
% Start of user PostScript
2.5714286 -2.8571429 2.5614816 -2.6986204 MAarrowhead1
% End of user PostScript
.90581 .5009 m
.90463 .51931 L
s
% Start of user PostScript
2.5714286 -2.1428571 2.5639683 -2.0262034 MAarrowhead1
% End of user PostScript
.90581 .61362 m
.90502 .62542 L
s
% Start of user PostScript
2.5714286 -1.4285714 2.5664551 -1.3537863 MAarrowhead1
% End of user PostScript
.90581 .72635 m
.90542 .73154 L
s
% Start of user PostScript
2.5714286 -0.71428571 2.5689418 -0.68136931 MAarrowhead1
% End of user PostScript
.90581 .83907 m
.90581 .83766 L
s
% Start of user PostScript
2.5714286 0. 2.5714286 -0.0089522879 MAarrowhead1
% End of user PostScript
.90581 .9518 m
.9062 .94378 L
s
% Start of user PostScript
2.5714286 0.71428571 2.5739153 0.66346473 MAarrowhead1
% End of user PostScript
.90581 1.06452 m
.90659 1.04989 L
s
% Start of user PostScript
2.5714286 1.4285714 2.5764021 1.3358818 MAarrowhead1
% End of user PostScript
.90581 1.17725 m
.90699 1.15601 L
s
% Start of user PostScript
2.5714286 2.1428571 2.5788888 2.0082988 MAarrowhead1
% End of user PostScript
.90581 1.28997 m
.90738 1.26213 L
s
% Start of user PostScript
2.5714286 2.8571429 2.5813756 2.6807158 MAarrowhead1
% End of user PostScript
.90581 1.40269 m
.90777 1.36824 L
s
% Start of user PostScript
2.5714286 3.5714286 2.5838623 3.3531328 MAarrowhead1
% End of user PostScript
.90581 1.51542 m
.90816 1.47436 L
s
% Start of user PostScript
2.5714286 4.2857143 2.5863491 4.0255498 MAarrowhead1
% End of user PostScript
.90581 1.62814 m
.90856 1.58048 L
s
% Start of user PostScript
2.5714286 5. 2.5888358 4.6979669 MAarrowhead1
% End of user PostScript
.97344 .05 m
.9707 .11428 L
s
% Start of user PostScript
3. -5. 2.9825928 -4.5926709 MAarrowhead1
% End of user PostScript
.97344 .16272 m
.97109 .21759 L
s
% Start of user PostScript
3. -4.2857143 2.9850795 -3.9380671 MAarrowhead1
% End of user PostScript
.97344 .27545 m
.97148 .32089 L
s
% Start of user PostScript
3. -3.5714286 2.9875663 -3.2834633 MAarrowhead1
% End of user PostScript
.97344 .38817 m
.97187 .4242 L
s
% Start of user PostScript
3. -2.8571429 2.990053 -2.6288595 MAarrowhead1
% End of user PostScript
.97344 .5009 m
.97227 .52751 L
s
% Start of user PostScript
3. -2.1428571 2.9925398 -1.9742557 MAarrowhead1
% End of user PostScript
.97344 .61362 m
.97266 .63081 L
s
% Start of user PostScript
3. -1.4285714 2.9950265 -1.3196519 MAarrowhead1
% End of user PostScript
.97344 .72635 m
.97305 .73412 L
s
% Start of user PostScript
3. -0.71428571 2.9975133 -0.66504813 MAarrowhead1
% End of user PostScript
.97344 .83907 m
.97344 .83742 L
s
% Start of user PostScript
3. 0. 3. -0.010444336 MAarrowhead1
% End of user PostScript
.97344 .9518 m
.97384 .94073 L
s
% Start of user PostScript
3. 0.71428571 3.0024867 0.64415946 MAarrowhead1
% End of user PostScript
.97344 1.06452 m
.97423 1.04403 L
s
% Start of user PostScript
3. 1.4285714 3.0049735 1.2987633 MAarrowhead1
% End of user PostScript
.97344 1.17725 m
.97462 1.14734 L
s
% Start of user PostScript
3. 2.1428571 3.0074602 1.953367 MAarrowhead1
% End of user PostScript
.97344 1.28997 m
.97501 1.25065 L
s
% Start of user PostScript
3. 2.8571429 3.009947 2.6079708 MAarrowhead1
% End of user PostScript
.97344 1.40269 m
.97541 1.35395 L
s
% Start of user PostScript
3. 3.5714286 3.0124337 3.2625746 MAarrowhead1
% End of user PostScript
.97344 1.51542 m
.9758 1.45726 L
s
% Start of user PostScript
3. 4.2857143 3.0149205 3.9171784 MAarrowhead1
% End of user PostScript
.97344 1.62814 m
.97619 1.56056 L
s
% Start of user PostScript
3. 5. 3.0174072 4.5717822 MAarrowhead1
% End of user PostScript
.5 Mabswid
.51578 .83907 m
.51578 .83867 L
.51576 .83827 L
.51574 .83778 L
.5157 .83729 L
.5156 .83631 L
.51544 .83511 L
.51484 .8317 L
.51391 .8275 L
.50275 .79011 L
.46093 .6731 L
.44114 .62457 L
.41389 .56835 L
.4047 .55289 L
.39551 .53951 L
.38619 .52826 L
.38153 .52358 L
.37725 .51987 L
.36864 .51418 L
.36386 .5121 L
.35934 .51088 L
.35439 .51041 L
.34983 .51079 L
.34484 .51216 L
.3396 .51469 L
.33137 .52105 L
.32681 .52589 L
.32265 .53115 L
.31537 .5424 L
.30767 .55725 L
.29539 .58785 L
.28399 .62468 L
.27373 .66606 L
.26509 .70867 L
.25949 .74173 L
.2551 .77282 L
.25252 .79556 L
.25153 .80637 L
.25084 .81554 L
.25056 .82016 L
.25032 .82487 L
.25006 .83241 L
.24997 .83924 L
.25005 .84486 L
.25023 .84955 L
.25048 .85353 L
.25085 .85749 L
.25135 .86133 L
.25196 .86492 L
.25262 .86799 L
Mistroke
.25405 .87298 L
.25604 .87794 L
.25813 .88179 L
.26252 .88759 L
.2678 .89262 L
.27405 .89744 L
.29995 .91594 L
.31826 .93195 L
.33924 .95543 L
.36786 .9982 L
.41447 1.09613 L
.49175 1.31405 L
.51875 1.39451 L
.55227 1.48765 L
.58496 1.56358 L
.60506 1.59938 L
.62469 1.62407 L
.64282 1.63627 L
.66205 1.6364 L
.67835 1.62488 L
.69574 1.59951 L
.71373 1.5576 L
.73163 1.49857 L
.75945 1.3683 L
.78401 1.20844 L
.79964 1.07986 L
.81457 .92523 L
.81636 .90141 L
.81771 .88086 L
.81849 .86655 L
.819 .85397 L
.81918 .84759 L
.81926 .84243 L
.81923 .83398 L
.81903 .82796 L
.81866 .82291 L
.8182 .81906 L
.8176 .81579 L
.81697 .81337 L
.81633 .81158 L
.81558 .80999 L
.81472 .8086 L
.81378 .80745 L
.81288 .8066 L
.81121 .80543 L
.80959 .80461 L
.80646 .80349 L
.803 .80257 L
.79914 .80168 L
.78314 .79804 L
Mistroke
.76354 .79265 L
.74221 .785 L
.71847 .77338 L
.70312 .76349 L
.68612 .7497 L
.66786 .73089 L
.64279 .69704 L
.60238 .62101 L
.53971 .45751 L
.4979 .33269 L
.43905 .16827 L
.42157 .12787 L
.40176 .08953 L
.38995 .07117 L
.37897 .05752 L
.35702 .04152 L
.33716 .04176 L
.31581 .05992 L
.29633 .09476 L
.27928 .14115 L
.26285 .20125 L
.24918 .26377 L
.22379 .41353 L
.19585 .64012 L
.18773 .72499 L
.18352 .77796 L
.18217 .79879 L
.18137 .81383 L
.18088 .82707 L
.18074 .83305 L
.1807 .83799 L
.18076 .8452 L
.18101 .85145 L
.1812 .85405 L
.18143 .85651 L
.18194 .86019 L
.18253 .86308 L
.18316 .86526 L
.18376 .86685 L
.18445 .86828 L
.18532 .86964 L
.18614 .87064 L
.18711 .87156 L
.18805 .87227 L
.18977 .87325 L
.19167 .87404 L
.19536 .87516 L
.19889 .87602 L
.2072 .87786 L
.22529 .88227 L
Mistroke
.24502 .88831 L
.26572 .8966 L
.29143 .9109 L
.30914 .92426 L
.32885 .94352 L
.3529 .97457 L
.38312 1.02679 L
.43288 1.14372 L
.4748 1.26347 L
.52895 1.42427 L
.56222 1.51299 L
.58024 1.55412 L
.59771 1.58771 L
.61879 1.61812 L
.63017 1.6292 L
.64234 1.63638 L
.66216 1.63665 L
.68368 1.61888 L
.69565 1.59998 L
.70667 1.57635 L
.72623 1.5186 L
.74527 1.44117 L
.76321 1.34706 L
.78777 1.18002 L
.80253 1.05344 L
.81056 .97233 L
.81565 .91171 L
.81737 .88694 L
.81839 .86914 L
.81898 .85568 L
.81926 .84451 L
.81929 .83597 L
.8191 .82886 L
.81876 .82364 L
.81835 .81987 L
.81805 .81794 L
.81776 .81637 L
.81702 .81345 L
.81631 .81145 L
.8155 .80978 L
.81474 .80858 L
.81388 .80753 L
.8131 .80677 L
.81225 .80609 L
.81061 .80508 L
.80905 .80437 L
.8051 .8031 L
.80141 .80219 L
.78661 .79887 L
.76795 .79398 L
Mistroke
.74854 .7875 L
.72643 .77772 L
.71357 .77045 L
.69749 .75928 L
.6787 .74261 L
.65765 .71829 L
.6253 .66742 L
.5792 .56593 L
.50481 .35346 L
.47661 .26983 L
.44098 .17305 L
.42273 .13037 L
.40237 .09058 L
.3842 .06358 L
.36372 .04471 L
.34442 .03992 L
.32694 .04799 L
.30717 .0731 L
.2893 .112 L
.2737 .15979 L
.26036 .21178 L
.23446 .34496 L
.21448 .4808 L
.19937 .607 L
.18959 .70417 L
.18446 .76508 L
.18289 .78711 L
.18174 .80646 L
.18105 .82161 L
.18084 .82843 L
.18073 .83402 L
.18071 .84251 L
.18087 .84868 L
.1812 .85405 L
.18168 .85846 L
.18217 .8614 L
.18275 .86393 L
.18336 .86585 L
.18395 .86729 L
.18468 .86867 L
.1855 .86989 L
.18708 .87153 L
.18797 .87221 L
.18882 .87275 L
.19081 .87371 L
.19279 .87443 L
.19496 .87506 L
.20274 .87687 L
.23914 .88634 L
.23959 .88649 L
Mistroke
Mfstroke
0 0 m
1 0 L
1 1.67657 L
0 1.67657 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{171.75, 287.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000410001[b000`400?l00000o`00003oo`3oool2
0?ooo`00o`3oool20?ooo`00o`3oool20?ooo`00o`3oool20?ooo`00o`3oool20?ooo`00o`3oool2
0?ooo`00o`3oool20?ooo`00o`3oool20?ooo`00o`3oool20?ooo`00C03oool00`000000oooo0?oo
o`1V0?ooo`030000003oool0oooo04T0oooo001<0?ooo`<00000203oool4000005X0oooo0`000019
0?ooo`00C03oool2000000L0oooo2@00000@0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0Q0?ooo`<000003@3oool2000003/0oooo000i0?ooo`@0
00003P3oool2000000H0oooo0`0000060?ooo`<000003`3oool2000000l0oooo0P00000?0?ooo`80
0000403oool00`000000oooo0?ooo`0?0?ooo`8000003P3oool4000003T0oooo00060?ooo`030000
003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`@00000
3P3oool2000000D0oooo0`0000080?ooo`<000003P3oool2000000l0oooo0P00000?0?ooo`800000
4@3oool00`000000oooo0?ooo`0=0?ooo`8000003`3oool4000000l0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`<000003`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo1000000=0?ooo`8000003`3o
ool2000000l0oooo0P00000@0?ooo`<000003@3oool00`000000oooo0?ooo`0>0?ooo`<00000403o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01@3o
ool000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0<0?ooo`<00000403oool2
000001D0oooo00<000000?ooo`3oool02`3oool3000000d0oooo0`00000>0?ooo`<000003P3oool3
00000100oooo0`00000N0?ooo`8000003P3oool4000000l0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo
00`0oooo1000000?0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03@3oool00`00
0000oooo0000000<0?ooo`8000003`3oool2000000l0oooo0P00000A0?ooo`8000007`3oool00`00
0000oooo0?ooo`0=0?ooo`@000003`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0303oool4000000l0
oooo00<000000?ooo`3oool04P3oool200000180oooo00<000000?ooo`3oool0?`3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo00d0oooo0`00000@0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`0<0?ooo`<000009@3oool00`000000oooo0?ooo`0A0?ooo`800000L@3oool200000100oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00h0oooo0P00000T0?ooo`800000503oool200000700oooo00<0
00000?ooo`3oool03@3oool4000000l0oooo00<000000?ooo`3oool01@3oool000H0oooo00<00000
0?ooo`3oool03@3oool4000000h0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool03@3oool4000000l0oooo00<00000
0?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03@3oool4000000h0oooo00<000000?oo
o`3oool08P3oool2000001L0oooo00<000000?ooo`3oool0O@3oool300000100oooo00<000000?oo
o`3oool01@3oool000H0oooo00<000000?ooo`3oool03@3oool3000003@0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`1n0?ooo`800000403oool00`000000oooo0?ooo`050?ooo`00
1P3oool00`000000oooo0?ooo`0>0?ooo`800000<`3oool2000001T0oooo00<000000?ooo`3oool0
O@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo07d0oooo00<000000?ooo`3oool03@3oool4000000H0oooo00050?ooo`<000003`3oool0
0`000000oooo0?ooo`0b0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0S03oool4
000000H0oooo00050?ooo`<00000@`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
08d0oooo0`0000060?ooo`001@3oool3000004<0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`2<0?ooo`8000001`3oool000H0oooo00<000000?ooo`3oool0@@3oool00`000000oooo
0?ooo`0L0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool01@3oool000H0oooo00<0
00000?ooo`3oool0@@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo08`0oooo00<0
00000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool0@03oool2000001l0oooo00<00000
0?ooo`3oool0U03oool004T0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`2C0?oo
o`00B@3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo09<0oooo00180?ooo`800000
8P3oool00`000000oooo0?ooo`2B0?ooo`00B03oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo0980oooo00170?ooo`800000903oool00`000000oooo0?ooo`2A0?ooo`00A`3oool20000
02@0oooo00<000000?ooo`3oool0T@3oool004L0oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`2@0?ooo`00AP3oool2000000@0oooo00<000000?ooo`3oool07`3oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo04T0oooo00160?ooo`800000103oool3000001l0oooo0P00
00150?ooo`<00000B@3oool003T0oooo0P00000;0?ooo`030000003oool0oooo00<0oooo0P00000P
0?ooo`8000003`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0200oooo0`00000=
0?ooo`@00000>@3oool003T0oooo100000090?ooo`030000003oool0oooo0080oooo0P00000?0?oo
o`800000403oool3000000h0oooo0P00000@0?ooo`8000003`3oool00`000000oooo0?ooo`0?0?oo
o`8000003P3oool4000003T0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0=0?ooo`<000002@3oool2000000@0oooo0P00000@0?ooo`@0
00003@3oool3000000h0oooo0P00000@0?ooo`800000403oool00`000000oooo0?ooo`0=0?ooo`80
00003`3oool300000100oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
3@3oool200000100oooo0P0000090?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
403oool5000000/0oooo1000000>0?ooo`8000003`3oool3000000l0oooo0`00000=0?ooo`030000
003oool0oooo00l0oooo0P00000>0?ooo`@000003`3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3o
ool3000000l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0G0?ooo`<000003@3o
ool3000000h0oooo0`00000>0?ooo`<00000403oool3000001h0oooo00<000000?ooo`3oool03@3o
ool4000000l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo0`00000?0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0<0?ooo`<000003P3oool2
000000l0oooo1000000?0?ooo`8000007`3oool00`000000oooo0?ooo`0=0?ooo`<00000403oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03@3oool2000001X0oooo00<000000?ooo`3oool0:03oool40000
0380oooo00<000000?ooo`3oool0;P3oool2000000h0oooo0`00000@0?ooo`030000003oool0oooo
00D0oooo00060?ooo`030000003oool0oooo00d0oooo1000000>0?ooo`8000006P3oool00`000000
oooo0?ooo`0Z0?ooo`800000H`3oool00`000000oooo0?ooo`0=0?ooo`@000003`3oool00`000000
oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0=0?ooo`@000003P3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0H@3oool00`000000oooo
0?ooo`0=0?ooo`@000003`3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?oo
o`0=0?ooo`<000003`3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo02/0oooo0P00
001b0?ooo`<00000403oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0>
0?ooo`800000:P3oool00`000000oooo0?ooo`0/0?ooo`800000L`3oool200000100oooo00<00000
0?ooo`3oool01@3oool000D0oooo0P00000@0?ooo`030000003oool0oooo02T0oooo00<000000?oo
o`3oool0;@3oool200000780oooo00<000000?ooo`3oool03P3oool3000000H0oooo00050?ooo`<0
00003`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02d0oooo0P00001b0?ooo`03
0000003oool0oooo00h0oooo0`0000060?ooo`001@3oool3000003X0oooo0P00000`0?ooo`030000
003oool0oooo0840oooo0`0000060?ooo`001@3oool3000003X0oooo00<000000?ooo`3oool0;`3o
ool2000008<0oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool0>@3oool0
0`000000oooo0?ooo`0_0?ooo`800000P`3oool00`000000oooo0?ooo`050?ooo`001P3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0P@3oool00`00
0000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0h0?ooo`800000<P3oool2000008X0
oooo00110?ooo`030000003oool0oooo0340oooo0P00002:0?ooo`00@@3oool00`000000oooo0?oo
o`0b0?ooo`030000003oool0oooo08P0oooo00110?ooo`030000003oool0oooo0380oooo0P000029
0?ooo`00@@3oool00`000000oooo0?ooo`0b0?ooo`800000R@3oool00400oooo0P00000e0?ooo`03
0000003oool0oooo08L0oooo00100?ooo`030000003oool0oooo03@0oooo0P0000280?ooo`00@03o
ool00`000000oooo0?ooo`0d0?ooo`800000R03oool00400oooo00<000000?ooo`3oool0=@3oool2
000008L0oooo000o0?ooo`800000=`3oool2000003/0oooo00<000000?ooo`3oool0B@3oool003l0
oooo0P00000;0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0>@3oool2000004X0
oooo000o0?ooo`8000002`3oool3000002X0oooo0P00000i0?ooo`@00000303oool4000003T0oooo
000i0?ooo`@000000P3oool00`000000oooo0?ooo`0:0?ooo`800000803oool00`000000oooo0?oo
o`080?ooo`800000103oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0200oooo0`00
000=0?ooo`@00000>@3oool003T0oooo100000020?ooo`030000003oool0oooo00T0oooo0P00000@
0?ooo`8000003`3oool2000000X0oooo00<000000?ooo`3oool00P3oool200000100oooo0P00000?
0?ooo`030000003oool0oooo00h0oooo0`00000>0?ooo`<00000>P3oool000H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0=0?ooo`<000003P3oool3000000<0oooo00<000000?oo
o`3oool02@3oool200000100oooo1000000=0?ooo`8000002P3oool2000000<0oooo0P00000@0?oo
o`800000403oool2000000h0oooo0P00000@0?ooo`8000003P3oool4000000l0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo00d0oooo0`00000?0?ooo`8000000P3oool2000000/0oooo00<000000?ooo`3o
ool04@3oool4000000/0oooo0`0000090?ooo`800000103oool2000000l0oooo0`00000?0?ooo`@0
00007P3oool00`000000oooo0?ooo`0=0?ooo`@000003`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
3@3oool3000000l0oooo00@000000?ooo`3oool0oooo0P00000P0?ooo`8000003@3oool2000000X0
oooo00<000000?ooo`3oool00P3oool3000000h0oooo0`00000@0?ooo`@000007@3oool00`000000
oooo0?ooo`0=0?ooo`<00000403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0
oooo00060?ooo`030000003oool0oooo00d0oooo1000000?0?ooo`030000003oool0oooo00h0oooo
00D000000?ooo`3oool0oooo0000000R0?ooo`030000003oool0oooo00/0oooo0`0000090?ooo`80
00000`3oool2000000l0oooo1000000?0?ooo`<00000;`3oool2000000h0oooo1000000?0?ooo`03
0000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00d0oooo1000000?0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool0;P3oool2000000X0oooo0P00000D0?ooo`800000
4P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00d0oooo1000000?0?ooo`030000
003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00d0oooo0`00000@0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1G0?ooo`030000003o
ool0oooo00d0oooo0`00000@0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0
oooo00h0oooo0P00000U0?ooo`030000003oool0oooo03/0oooo0P00001Y0?ooo`8000003`3oool3
000000H0oooo00050?ooo`<000003`3oool00`000000oooo0?ooo`0S0?ooo`800000?@3oool20000
06T0oooo00<000000?ooo`3oool03P3oool3000000H0oooo00050?ooo`<000003`3oool00`000000
oooo0?ooo`0S0?ooo`800000?P3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo00h0
oooo0`0000060?ooo`001@3oool3000003D0oooo00<000000?ooo`3oool0?@3oool00`000000oooo
0?ooo`1i0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo03@0oooo00<0
00000?ooo`3oool0?P3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo00D0oooo0006
0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`1h
0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo03@0oooo00<000000?oo
o`3oool0?P3oool00`000000oooo0?ooo`200?ooo`00?03oool200000440oooo00<000000?ooo`3o
ool0O`3oool003`0oooo0P0000110?ooo`030000003oool0oooo07l0oooo000l0?ooo`030000003o
ool0oooo0400oooo00<000000?ooo`3oool0O`3oool003`0oooo00<000000?ooo`3oool0@@3oool0
0`000000oooo0?ooo`1n0?ooo`00?03oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo
07h0oooo000l0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0OP3oool003`0oooo
00<000000?ooo`3oool0@P3oool00`000000oooo0?ooo`1m0?ooo`00>`3oool2000004@0oooo00<0
00000?ooo`3oool0O@3oool003/0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`1m
0?ooo`00>`3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo07`0oooo000k0?ooo`03
0000003oool0oooo04@0oooo00<000000?ooo`3oool0O03oool003/0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0
B@3oool003/0oooo00<000000?ooo`3oool03P3oool3000003@0oooo00<000000?ooo`3oool0;`3o
ool2000000d0oooo0P00000k0?ooo`00>P3oool3000000l0oooo0P00000e0?ooo`030000003oool0
oooo02h0oooo1000000<0?ooo`@00000>@3oool003X0oooo0`00000>0?ooo`<00000<@3oool01@00
0000oooo0?ooo`3oool000000300oooo0`00000=0?ooo`@000003@3oool2000002X0oooo000X0?oo
o`<000003`3oool2000000l0oooo0P00000@0?ooo`8000003`3oool00`000000oooo0?ooo`0>0?oo
o`8000000`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03P3oool3000000h0oooo0`00000>0?ooo`@00000:03oool000H0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`0=0?ooo`<000003`3oool2000000l0oooo0P00000@0?ooo`@0
00003@3oool2000000l0oooo0P0000030?ooo`030000003oool0oooo00X0oooo0P00000@0?ooo`80
00003P3oool200000100oooo0P00000>0?ooo`@000003`3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
3@3oool3000000l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0A0?ooo`@00000
2`3oool2000000l0oooo0`0000020?ooo`8000002`3oool200000100oooo1000000N0?ooo`030000
003oool0oooo00d0oooo0`00000?0?ooo`<000003`3oool00`000000oooo0?ooo`050?ooo`001P3o
ool00`000000oooo0?ooo`0=0?ooo`@000003`3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo02<0oooo0P00000=0?ooo`8000003`3oool2000000<0oooo00<000000?ooo`3oool02@3o
ool300000100oooo1000000M0?ooo`030000003oool0oooo00h0oooo0P00000?0?ooo`<000003`3o
ool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0=0?ooo`@000003`3oool0
0`000000oooo0?ooo`0=0?ooo`8000009P3oool00`000000oooo0?ooo`0;0?ooo`<000003P3oool3
00000080oooo00<000000?ooo`3oool02@3oool4000000l0oooo0`00000_0?ooo`030000003oool0
oooo00h0oooo0`00000?0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo
00d0oooo0`00000@0?ooo`030000003oool0oooo00d0oooo0P00000d0?ooo`8000003`3oool20000
00@0oooo00<000000?ooo`3oool0203oool200000180oooo00<000000?ooo`3oool0;P3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo0`0000060?ooo`001@3oool3000000l0
oooo0P00000P0?ooo`800000B`3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo00d0
oooo0`0000060?ooo`001@3oool3000000l0oooo00<000000?ooo`3oool07`3oool00`000000oooo
0?ooo`1:0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool03@3oool3000000H0oooo
00050?ooo`<000003`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04/0oooo00<0
00000?ooo`3oool0K`3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0`
0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`05
0?ooo`001P3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo04/0oooo00<000000?oo
o`3oool0K`3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo04`0oooo00<000000?ooo`3oool0MP3oool003P0oooo0P00001>0?ooo`030000
003oool0oooo07H0oooo000h0?ooo`800000CP3oool00`000000oooo0?ooo`1f0?ooo`00>03oool0
0`000000oooo0?ooo`1>0?ooo`030000003oool0oooo07D0oooo000h0?ooo`030000003oool0oooo
04h0oooo00<000000?ooo`3oool0M@3oool003P0oooo00<000000?ooo`3oool0CP3oool00`000000
oooo0?ooo`1e0?ooo`00>03oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo07@0oooo
000h0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0M03oool003P0oooo00<00000
0?ooo`3oool0C`3oool00`000000oooo0?ooo`1d0?ooo`00=`3oool200000580oooo00<000000?oo
o`3oool0L`3oool003L0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`1c0?ooo`00
=`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo0780oooo000g0?ooo`030000003o
ool0oooo0580oooo00<000000?ooo`3oool0LP3oool003L0oooo00<000000?ooo`3oool0DP3oool0
0`000000oooo0?ooo`1b0?ooo`00=`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
03h0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`180?ooo`00=`3oool010000000
oooo0?ooo`00000A0?ooo`<00000?P3oool00`000000oooo0?ooo`0V0?ooo`8000003@3oool30000
03T0oooo000g0?ooo`030000003oool0oooo00<000003`3oool2000003l0oooo00<000000?ooo`3o
ool09@3oool4000000`0oooo0`00000>0?ooo`<00000:03oool002P0oooo0`00000;0?ooo`800000
0P3oool2000000l0oooo0P0000110?ooo`030000003oool0oooo02@0oooo0`00000=0?ooo`800000
3`3oool3000002P0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3o
ool3000000/0oooo00@000000?ooo`3oool0oooo0P00000?0?ooo`800000403oool3000000h0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0;0?ooo`040000003oool0oooo00000100
oooo0P00000@0?ooo`<000003P3oool2000000l0oooo0P00000?0?ooo`<000003`3oool00`000000
oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0=0?ooo`@000003P3oool3000000/0oooo
00D000000?ooo`3oool0oooo0000000@0?ooo`030000003oool0oooo0100oooo1@00000;0?ooo`80
0000403oool00`000000oooo0?ooo`0;0?ooo`80000000<0oooo0000000000003`3oool200000100
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`8000003`3oool3000000l0
oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03@3oool4000000l0oooo
00<000000?ooo`3oool02P3oool01@000000oooo0?ooo`3oool0000002@0oooo0`00000<0?ooo`80
0000403oool2000000d0oooo00@000000?ooo`0000000000403oool3000001l0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`<000003P3oool3000000H0oooo00050?ooo`<0
00003P3oool300000100oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0O0?ooo`H0
00000`3oool00`000000oooo0?ooo`0;0?ooo`8000003`3oool2000000d0oooo00<000000?ooo`00
00000P00000?0?ooo`@00000;`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0
oooo0`0000060?ooo`001@3oool3000000l0oooo0P00000@0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo
0`00000>0?ooo`<00000303oool6000000l0oooo0P0000120?ooo`030000003oool0oooo00d0oooo
0`0000060?ooo`001@3oool3000000l0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`0M0?ooo`030000003oool0oooo00H0oooo0P00000>0?ooo`8000003`3oool2000000h0oooo0`00
000A0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`050?ooo`001P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0703oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02`0oooo00<00000
0?ooo`3oool0IP3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0/0?oo
o`800000703oool2000000d0oooo00<000000?ooo`3oool0:`3oool2000006L0oooo00<000000?oo
o`3oool01@3oool000H0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1]0?ooo`00
=@3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
:P3oool2000006h0oooo000e0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo06`0oooo000e0?ooo`030000003oool0
oooo01P0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0
oooo06`0oooo000e0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`0Y0?ooo`800000K@3oool003D0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0J`3oool003D0oooo
00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo02T0oooo
0P00001/0?ooo`00=@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0:03oool2000006`0oooo000d0?ooo`030000003oool0oooo01L0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo06X0oooo000d0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0X0?oo
o`800000J`3oool003@0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0J0?ooo`03
0000003oool0oooo02L0oooo0P00001[0?ooo`00=03oool00`000000oooo0?ooo`0E0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1Y0?ooo`00=03o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool09`3o
ool2000006X0oooo000d0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0703oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo06P0oooo000d0?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0W0?ooo`800000J@3oool003@0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool0I`3oool003<0oooo00<000000?ooo`3oool05@3oool00`000000oooo
0?ooo`0O0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0I`3oool003<0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0000000P0?ooo`030000003oool0oooo02P0oooo00<0
00000?ooo`3oool0:P3oool00`000000oooo0?ooo`0i0?ooo`00<`3oool00`000000oooo0?ooo`04
0?ooo`<000003@3oool00`000000oooo00000002000001l0oooo00<000000?ooo`3oool09`3oool0
0`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00/0oooo0`00000>0?ooo`<00000:03oool0
02P0oooo0`0000080?ooo`030000003oool0oooo00@0oooo0`00000=0?ooo`040000003oool00000
00000200oooo00<000000?ooo`3oool09`3oool2000001`0oooo0`00000<0?ooo`8000003`3oool3
000000h0oooo0`00000G0?ooo`005P3oool4000000h0oooo0`0000080?ooo`030000003oool0oooo
00@0oooo0P00000=0?ooo`030000003oool0oooo008000008@3oool00`000000oooo0?ooo`0W0?oo
o`030000003oool0oooo01T0oooo1@00000;0?ooo`8000003`3oool2000000l0oooo0`00000G0?oo
o`001@3oool2000000l0oooo1000000>0?ooo`<00000203oool00`000000oooo0?ooo`040?ooo`80
00003@3oool010000000oooo0?ooo`00000@0?ooo`<000003P3oool200000100oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`050?ooo`8000001`3oool200000100oooo1000000=0?oo
o`030000003oool0oooo00h0oooo0P00000?0?ooo`<000003P3oool4000000D0oooo00050?ooo`@0
00003@3oool300000100oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00`0oooo00@000000?ooo`3oool000004@3oool5000000/0oooo0`00000?0?oo
o`8000003`3oool2000000L0oooo00<000000?ooo`3oool01@3oool200000100oooo0P00000?0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0=0?oo
o`@000001@3oool000D0oooo1000000>0?ooo`800000403oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0D0?ooo`<0
0000303oool3000000l0oooo0P00000?0?ooo`8000001`3oool00`000000oooo0?ooo`060?ooo`80
0000<@3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo0`0000060?ooo`00
1@3oool3000000l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`060?ooo`800000
503oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00/0oooo0`00000>0?ooo`<00000
3P3oool3000000H0oooo0P0000070?ooo`<00000@P3oool00`000000oooo0?ooo`0>0?ooo`800000
1P3oool000H0oooo0P00000?0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool04`3o
ool00`000000oooo0?ooo`0T0?ooo`<000003P3oool4000000d0oooo100000060?ooo`030000003o
ool0oooo00H0oooo0`00001B0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0
oooo02T0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0T0?ooo`80000000<0oooo
0000003oool0303oool2000000l0oooo0P0000080?ooo`8000001`3oool2000005<0oooo00<00000
0?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0B0?oo
o`030000003oool0oooo02P0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`1R0?oo
o`00<P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3o
ool09P3oool2000006<0oooo000b0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
:@3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0640oooo000b0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0W0?ooo`800000HP3o
ool00380oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0
oooo02L0oooo00<000000?ooo`3oool0H03oool00380oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0Z0?ooo`030000003oool0oooo02L0oooo0P00001Q0?ooo`00<@3oool2000001<0
oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo05l0
oooo000a0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`0W0?ooo`800000H03oool00340oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0/0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0GP3oool00340oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo02L0oooo0P00001O
0?ooo`00<@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo02d0oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`1M0?ooo`00<@3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo02h0oooo00<000000?ooo`3oool0:03oool2000005h0oooo000a0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0X0?ooo`030000
003oool0oooo05`0oooo000a0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`0X0?ooo`800000G@3oool00300oooo0P00000C0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool0:@3oool2000005`0oooo000`0?ooo`8000004`3oool00`00
0000oooo0?ooo`0`0?ooo`030000003oool0oooo02P0oooo0P00001L0?ooo`00<03oool2000001<0
oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0Y0?ooo`800000F`3oool00300oooo
0P00000B0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0:@3oool2000005X0oooo
000`0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool2000002`0oooo00<0
00000?ooo`3oool0:P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo03T0oooo000`
0?ooo`030000003oool0oooo00L0oooo0`0000070?ooo`030000003oool0oooo00@0oooo1000000A
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0:P3oool2000001d0oooo0`00000>
0?ooo`@000003@3oool2000001P0oooo000F0?ooo`@000003P3oool4000000@0oooo00<000000?oo
o`3oool01`3oool3000000L0oooo00<000000?ooo`3oool0103oool400000100oooo0`00000H0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool06`3oool3000000h0oooo0`00000>0?oo
o`@000003@3oool4000000D0oooo00050?ooo`@000003@3oool4000000h0oooo100000040?ooo`03
0000003oool0oooo00L0oooo0P0000070?ooo`030000003oool0oooo00D0oooo0`00000@0?ooo`D0
00005`3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01X0oooo0P00000?0?ooo`<0
00003P3oool4000000d0oooo100000050?ooo`001@3oool4000000h0oooo0`00000>0?ooo`<00000
1@3oool00`000000oooo0?ooo`070?ooo`8000001`3oool00`000000oooo0?ooo`060?ooo`800000
403oool3000001T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0;0?ooo`<00000
2`3oool2000000l0oooo0P00000?0?ooo`<000003P3oool3000000H0oooo00050?ooo`<000003`3o
ool200000100oooo0P0000050?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00h0oooo0P00000?0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00h0oooo0P0000050?ooo`8000002@3oool6000000/0oooo00<000000?ooo`3oool03P3o
ool200000100oooo0P00000?0?ooo`8000001P3oool000H0oooo0P00000@0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0O0?ooo`8000002`3oool0
0`000000oooo0?ooo`020?ooo`8000003`3oool00`000000oooo0?ooo`0>0?ooo`8000001P3oool2
000000X0oooo1000000;0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00l0oooo00<000000?ooo`3oool0:P3oool2000000X0oooo00<000000?ooo`3oool00P3oool20000
00l0oooo0P00000@0?ooo`8000001P3oool00`000000oooo0?ooo`0:0?ooo`800000;P3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo02H0
oooo0P00000A0?ooo`030000003oool0oooo02X0oooo0`00000:0?ooo`040000003oool0oooo0?oo
o`<000003P3oool200000100oooo0`0000060?ooo`030000003oool0oooo0580oooo000_0?ooo`80
00004@3oool00`000000oooo0?ooo`0[0?ooo`<000002@3oool010000000oooo0?ooo`3oool40000
00d0oooo0`00000@0?ooo`<000001P3oool200000580oooo000_0?ooo`8000004@3oool00`000000
oooo0?ooo`0[0?ooo`8000002`3oool00`000000oooo0?ooo`02000000l0oooo0`00000@0?ooo`80
0000203oool200000540oooo000_0?ooo`8000004@3oool00`000000oooo0?ooo`0h0?ooo`030000
003oool0oooo02l0oooo0P00001@0?ooo`00;`3oool00`000000oooo0?ooo`0@0?ooo`030000003o
ool0oooo03P0oooo00<000000?ooo`3oool0<03oool2000004l0oooo000_0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`0`0?ooo`800000CP3oool0
02l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo
0340oooo0P00001=0?ooo`00;`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool0<P3oool2000004`0oooo000_0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`0c0?ooo`800000BP3oool002l0oooo00<0
00000?ooo`3oool03`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo03D0oooo0P00
00180?ooo`00;P3oool200000140oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0e
0?ooo`<00000AP3oool002h0oooo0P00000A0?ooo`030000003oool0oooo03`0oooo00<000000?oo
o`3oool0=`3oool3000004@0oooo000^0?ooo`800000403oool00`000000oooo0?ooo`0m0?ooo`03
0000003oool0oooo03T0oooo100000110?ooo`00;P3oool200000100oooo00<000000?ooo`3oool0
?P3oool00`000000oooo0?ooo`0k0?ooo`<00000?`3oool002h0oooo00<000000?ooo`3oool03`3o
ool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo03h0oooo0`00000l0?ooo`00;P3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0@03oool3
000003X0oooo000^0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0?P3oool00`00
0000oooo0?ooo`130?ooo`@00000=P3oool002h0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0o0?ooo`030000003oool0oooo04H0oooo1@00000a0?ooo`00;P3oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0B`3oool2000002l0oooo
000^0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?oo
o`1<0?ooo`030000003oool0oooo02d0oooo000^0?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool0?`3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo02d0oooo000F0?oo
o`@000003P3oool3000000<0oooo00<000000?ooo`3oool0203oool4000000<0oooo00<000000?oo
o`3oool01`3oool4000000h0oooo0`00000>0?ooo`@000004P3oool00`000000oooo0?ooo`1;0?oo
o`800000;P3oool000D0oooo1000000=0?ooo`@000003P3oool3000000<0oooo00<000000?ooo`3o
ool0203oool4000000<0oooo00<000000?ooo`3oool01`3oool4000000h0oooo0`00000>0?ooo`@0
00004P3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo02`0oooo00050?ooo`@00000
3P3oool3000000h0oooo0`0000030?ooo`030000003oool0oooo00P0oooo0`0000040?ooo`030000
003oool0oooo00P0oooo0`00000>0?ooo`<000003P3oool3000001<0oooo00<000000?ooo`3oool0
C03oool00`000000oooo0?ooo`0/0?ooo`001@3oool3000000l0oooo0P00000?0?ooo`800000103o
ool00`000000oooo0?ooo`080?ooo`<00000103oool00`000000oooo0?ooo`080?ooo`800000403o
ool00`000000oooo0?ooo`0>0?ooo`800000503oool00`000000oooo0?ooo`1;0?ooo`030000003o
ool0oooo02`0oooo00060?ooo`8000003`3oool200000100oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`090?ooo`800000103oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo04/0oooo00<000000?ooo`3oool0;03oool000H0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000000<
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo04h0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`050?ooo`00;P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo04h0oooo00<000000?ooo`3oool03P3oool200000100oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0:0?ooo`040000003oool0oooo0?ooo`800000403oool00`00
0000oooo0?ooo`0>0?ooo`8000001P3oool002h0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`1=0?ooo`<000003`3oool3000000h0oooo0`00000>0?ooo`<000002`3oool010000000
oooo0?ooo`3oool3000000h0oooo0`00000>0?ooo`<000001P3oool002h0oooo00<000000?ooo`3o
ool03`3oool00`000000oooo0?ooo`1=0?ooo`<000003P3oool4000000h0oooo0`00000>0?ooo`<0
00002`3oool00`000000oooo0?ooo`04000000h0oooo0`00000>0?ooo`@000001@3oool002h0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1=0?ooo`<000003P3oool4000000h0oooo
0`00000>0?ooo`<000002`3oool00`000000oooo0?ooo`04000000h0oooo0`00000>0?ooo`@00000
1@3oool002h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`2?0?ooo`030000003o
ool0oooo02`0oooo000^0?ooo`800000403oool00`000000oooo0?ooo`2?0?ooo`030000003oool0
oooo02`0oooo000_0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0S`3oool00`00
0000oooo0?ooo`0/0?ooo`00<03oool3000000h0oooo00<000000?ooo`3oool0S@3oool2000002h0
oooo000b0?ooo`D000002P3oool00`000000oooo0?ooo`2=0?ooo`800000;P3oool003L0oooo1000
00070?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0;@3oool003/0oooo0`000004
0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0;@3oool003d0oooo0`0000030?oo
o`030000003oool0oooo08/0oooo00<000000?ooo`3oool0;@3oool00400oooo0`0000000`3oool0
00000?ooo`2;0?ooo`030000003oool0oooo02d0oooo00130?ooo`@00000RP3oool00`000000oooo
0?ooo`0]0?ooo`00A@3oool3000008T0oooo00<000000?ooo`3oool0;@3oool004L0oooo0P000028
0?ooo`030000003oool0oooo02d0oooo00180?ooo`<00000QP3oool00`000000oooo0?ooo`0]0?oo
o`00BP3oool2000008D0oooo00<000000?ooo`3oool0;@3oool004/0oooo0P0000240?ooo`030000
003oool0oooo02d0oooo001=0?ooo`800000PP3oool00`000000oooo0?ooo`0]0?ooo`00CP3oool2
00000840oooo00<000000?ooo`3oool0;@3oool00500oooo00<000000?ooo`3oool0OP3oool00`00
0000oooo0?ooo`0]0?ooo`00D@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0100
oooo0P00000?0?ooo`8000003P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo02d0
oooo001B0?ooo`030000003oool0oooo00D0oooo0P00000@0?ooo`@000003@3oool4000000d0oooo
0P0000110?ooo`030000003oool0oooo02d0oooo001C0?ooo`030000003oool0oooo00<0oooo1000
000@0?ooo`<000003P3oool3000000`0oooo100000100?ooo`030000003oool0oooo02D0oooo00<0
00000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`03
0?ooo`<00000403oool3000000h0oooo0`00000=0?ooo`<00000?`3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02T0oooo00<000000?oo
o`3oool00`3oool300000100oooo0P00000?0?ooo`8000003P3oool3000003<0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`<0
0000203oool2000000H0oooo00<000000?ooo`3oool03P3oool200000100oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0=0?ooo`<00000403oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00d0oooo0P0000070?ooo`001@3oool3000000l0oooo0P00000?0?ooo`800000
403oool00`000000oooo0?ooo`0:0?ooo`D000002P3oool2000004D0oooo1P00000A0?ooo`030000
003oool0oooo00d0oooo0P00000:0?ooo`030000003oool0oooo00<0oooo0P00000?0?ooo`<00000
3P3oool3000000H0oooo00050?ooo`<000003`3oool3000000h0oooo0P00000?0?ooo`800000303o
ool4000000`0oooo0P0000150?ooo`@000004@3oool2000000l0oooo0P00000:0?ooo`030000003o
ool0oooo00<0oooo0`00000>0?ooo`<000003P3oool3000000H0oooo00050?ooo`<000003P3oool4
000000d0oooo1000000>0?ooo`<00000303oool2000000h0oooo00<000000?ooo`3oool0@`3oool3
00000180oooo0P00000>0?ooo`@000002@3oool00`000000oooo0?ooo`020?ooo`@000003P3oool3
000000l0oooo0P0000060?ooo`005P3oool4000000d0oooo1000000=0?ooo`@00000303oool00`00
0000oooo0?ooo`0=0?ooo`800000A@3oool00`000000oooo0?ooo`0@0?ooo`<000003P3oool40000
00T0oooo00<000000?ooo`3oool00P3oool4000002P0oooo000Y0?ooo`8000003@3oool4000001d0
oooo0P00001G0?ooo`<00000403oool2000000T0oooo00<000000?ooo`3oool0;P3oool005X0oooo
00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0^0?ooo`00F`3oool00`000000oooo0?oo
o`1b0?ooo`030000003oool0oooo02h0oooo001K0?ooo`030000003oool0oooo0740oooo0P00000`
0?ooo`00G03oool00`000000oooo0?ooo`1`0?ooo`030000003oool0oooo02l0oooo001L0?ooo`80
0000L@3oool00`000000oooo0?ooo`0_0?ooo`00G@3oool00`000000oooo0?ooo`1_0?ooo`030000
003oool0oooo02l0oooo001N0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3oool0;`3o
ool005h0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0_0?ooo`00G`3oool00`00
0000oooo0?ooo`1]0?ooo`030000003oool0oooo02l0oooo001O0?ooo`800000KP3oool00`000000
oooo0?ooo`0_0?ooo`00H03oool00`000000oooo0?ooo`1[0?ooo`800000<@3oool00600oooo0P00
001/0?ooo`800000<@3oool00640oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`0`
0?ooo`00H@3oool2000006/0oooo00<000000?ooo`3oool0<03oool00680oooo00<000000?ooo`3o
ool0J@3oool00`000000oooo0?ooo`0`0?ooo`00HP3oool2000006X0oooo00<000000?ooo`3oool0
<03oool006<0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0`0?ooo`00H`3oool2
000006P0oooo0P00000b0?ooo`00G03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool03P3oool2000000l0oooo00<000000?ooo`3oool0>P3oool20000
02/0oooo00<000000?ooo`3oool0103oool000L0oooo00<000000?ooo`3oool0D@3oool2000000L0
oooo0P0000060?ooo`<000003P3oool4000000d0oooo0`00000l0?ooo`8000006@3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo00@0oooo00070?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0?`3oool3000000P0oooo00<000000?ooo`3oool01@3oool2000000l0oooo
0`00000>0?ooo`8000003@3oool00`000000oooo0?ooo`0/0?ooo`8000006@3oool00`000000oooo
0?ooo`0>0?ooo`8000001P3oool000H0oooo0P00000@0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool0;`3oool3000000L0oooo00<000000?ooo`3oool01@3oool2000000l0oooo0`00
000>0?ooo`8000003@3oool3000002`0oooo0P0000080?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool03P3oool3000000D0oooo00060?ooo`<000003P3oool200000100oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`0N0?ooo`800000203oool00`000000oooo0?ooo`05
0?ooo`8000003`3oool2000000l0oooo0P00000;0?ooo`H000008P3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool03@3oool2000000l0oooo10000005
0?ooo`001@3oool4000000h0oooo0`00000>0?ooo`800000403oool2000000l0oooo0P00000@0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool2000000l0oooo0P00000?0?oo
o`8000002`3oool700000100oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00H0oooo0P00000?0?ooo`<000003P3oool4000000D0oooo00050?ooo`@0
00003@3oool4000000h0oooo0`00000?0?ooo`8000003@3oool3000001/0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool07`3oool2000000l0oooo0P0000070?ooo`030000003oool0oooo00H0oooo0`00000>0?ooo`<0
00005`3oool001H0oooo1000000=0?ooo`@000003P3oool3000000/0oooo1000000L0?ooo`030000
003oool0oooo04P0oooo0`00000?0?ooo`<000001P3oool00`000000oooo0?ooo`050?ooo`@00000
3`3oool2000001L0oooo000W0?ooo`@000003P3oool3000000`0oooo0`00000M0?ooo`030000003o
ool0oooo04L0oooo0`00000>0?ooo`@000001@3oool2000000L0oooo1000000X0?ooo`00>03oool4
000000d0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`160?ooo`@000003P3oool4
000000D0oooo0P00000c0?ooo`00>P3oool2000002`0oooo0P0000190?ooo`8000005`3oool20000
03<0oooo001Y0?ooo`030000003oool0oooo0600oooo0P00000c0?ooo`00J@3oool200000640oooo
0P00000c0?ooo`00JP3oool00`000000oooo0?ooo`1O0?ooo`800000<`3oool006X0oooo00<00000
0?ooo`3oool0G`3oool00`000000oooo0?ooo`0b0?ooo`00J`3oool00`000000oooo0?ooo`1N0?oo
o`030000003oool0oooo0380oooo001[0?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3o
ool0<P3oool006/0oooo0P00001O0?ooo`030000003oool0oooo0380oooo001/0?ooo`030000003o
ool0oooo05d0oooo00<000000?ooo`3oool0<P3oool006`0oooo00<000000?ooo`3oool0G03oool2
000003@0oooo001/0?ooo`800000G@3oool2000003@0oooo001]0?ooo`030000003oool0oooo05/0
oooo0P00000d0?ooo`00K@3oool2000005`0oooo00<000000?ooo`3oool0<`3oool006h0oooo00<0
00000?ooo`3oool0FP3oool00`000000oooo0?ooo`0c0?ooo`00KP3oool2000005/0oooo00<00000
0?ooo`3oool0<`3oool006h0oooo0P00001K0?ooo`030000003oool0oooo03<0oooo001_0?ooo`03
0000003oool0oooo05T0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`040?ooo`00
1`3oool00`000000oooo0?ooo`1U0?ooo`800000FP3oool00`000000oooo0?ooo`0/0?ooo`030000
003oool0oooo00@0oooo00070?ooo`030000003oool0oooo06D0oooo0P00001J0?ooo`030000003o
ool0oooo02/0oooo0P0000060?ooo`001P3oool200000100oooo00<000000?ooo`3oool0E@3oool0
0`000000oooo0?ooo`1H0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool03P3oool2
000000H0oooo00060?ooo`<000003`3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo
0100oooo00@000000?ooo`0000000000303oool3000000l0oooo00<000000?ooo`3oool0=`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool4
000000D0oooo00050?ooo`@000003P3oool200000100oooo00<000000?ooo`3oool0;@3oool30000
0100oooo0`0000000`3oool000000000000<0?ooo`<000003@3oool3000000h0oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03@3oool4000000D0oooo00050?ooo`@000003P3oool200000100oooo00<000000?ooo`3o
ool0;P3oool300000100oooo0P0000020?ooo`030000003oool0oooo00/0oooo0P00000>0?ooo`80
00003@3oool3000002D0oooo00D000000?ooo`3oool0oooo0000000;0?ooo`8000003`3oool30000
00l0oooo0`0000050?ooo`001P3oool00`000000oooo0?ooo`0=0?ooo`@000003P3oool200000140
oooo00<000000?ooo`3oool07@3oool200000100oooo0`0000000`3oool000000000000<0?ooo`80
00003P3oool3000000`0oooo1@00000A0?ooo`030000003oool0oooo00h0oooo0P0000030?ooo`03
0000003oool0oooo00T0oooo0P00000?0?ooo`<000003`3oool00`000000oooo0?ooo`050?ooo`00
1P3oool00`000000oooo0?ooo`0=0?ooo`@000003P3oool3000000l0oooo0P00000@0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool2000000030?ooo`000000000000d0oooo
00<000000?ooo`3oool03@3oool2000000l0oooo0`00000?0?ooo`800000403oool2000000<0oooo
00<000000?ooo`3oool0203oool4000000h0oooo0`00000?0?ooo`030000003oool0oooo00D0oooo
000H0?ooo`8000003@3oool4000000h0oooo0`00000?0?ooo`030000003oool0oooo0240oooo00@0
00000?ooo`3oool000003@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01l0oooo
0P00000?0?ooo`<000000`3oool00`000000oooo0?ooo`080?ooo`@00000:03oool002L0oooo1000
000>0?ooo`<000003@3oool3000002H0oooo0P00000n0?ooo`<000003`3oool3000000<0oooo00<0
00000?ooo`3oool02P3oool2000002P0oooo000h0?ooo`@00000303oool3000002P0oooo00<00000
0?ooo`3oool0?03oool3000000h0oooo100000030?ooo`030000003oool0oooo03@0oooo000j0?oo
o`8000003@3oool2000002P0oooo00<000000?ooo`3oool0>`3oool4000001@0oooo0P00000f0?oo
o`00L`3oool2000003h0oooo0P00000D0?ooo`030000003oool0oooo03D0oooo001d0?ooo`030000
003oool0oooo0580oooo00<000000?ooo`3oool0=@3oool007@0oooo00<000000?ooo`3oool0DP3o
ool00`000000oooo0?ooo`0e0?ooo`00M03oool00`000000oooo0?ooo`1B0?ooo`030000003oool0
oooo03D0oooo001e0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0=@3oool007D0
oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`0e0?ooo`00M@3oool00`000000oooo
0?ooo`1@0?ooo`030000003oool0oooo03H0oooo001f0?ooo`030000003oool0oooo04l0oooo00<0
00000?ooo`3oool0=P3oool007H0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0f
0?ooo`00MP3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo03H0oooo001g0?ooo`03
0000003oool0oooo04h0oooo00<000000?ooo`3oool0=P3oool007L0oooo00<000000?ooo`3oool0
CP3oool00`000000oooo0?ooo`0f0?ooo`00M`3oool2000004h0oooo00<000000?ooo`3oool0=`3o
ool007P0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0g0?ooo`00N03oool00`00
0000oooo0?ooo`1<0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0103oool000L0
oooo00<000000?ooo`3oool0KP3oool2000004d0oooo00<000000?ooo`3oool0<03oool00`000000
oooo0?ooo`040?ooo`001`3oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool0;`3oool2000000H0oooo00060?ooo`800000L@3oool00`000000oooo0?oo
o`1;0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03P3oool3000000D0oooo0006
0?ooo`<000003`3oool00`000000oooo0?ooo`1N0?ooo`800000B`3oool00`000000oooo0?ooo`0O
0?ooo`030000003oool0oooo00d0oooo100000050?ooo`001@3oool4000000l0oooo00<000000?oo
o`3oool0G`3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`@000001@3oool000D0oooo1000000>0?ooo`80
0000403oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`020?ooo`8000003`3oool2000003@0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0=0?ooo`<000003`3oool00`000000oooo0?ooo`050?ooo`00
1P3oool00`000000oooo0?ooo`0>0?ooo`<000003`3oool00`000000oooo0?ooo`0^0?ooo`<00000
3`3oool3000000/0oooo0P0000000`3oool0000000000002000000d0oooo1000000=0?ooo`030000
003oool0oooo02@0oooo00<000000?ooo`3oool03@3oool2000000l0oooo0`00000?0?ooo`030000
003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00d0oooo1000000?0?ooo`030000003o
ool0oooo00l0oooo00<000000?ooo`3oool06`3oool400000100oooo0P00000<0?ooo`030000003o
ool0oooo00<000003P3oool3000000`0oooo0`00000T0?ooo`8000003`3oool2000000l0oooo0`00
000?0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00d0oooo1000000>
0?ooo`<000003`3oool200000100oooo00<000000?ooo`3oool0303oool3000000l0oooo0`00000;
0?ooo`030000003oool0oooo00<000003`3oool3000000/0oooo1@00000A0?ooo`030000003oool0
oooo00h0oooo0P00000>0?ooo`@000003`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool30000
00l0oooo0P00000?0?ooo`800000403oool2000000l0oooo0P00000;0?ooo`8000000P3oool20000
0100oooo0P00000>0?ooo`<000003`3oool2000000l0oooo0`00000>0?ooo`@000003`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo000X0?ooo`<000003`3oool2000000h0
oooo0P00000S0?ooo`030000003oool0oooo00X0oooo00D000000?ooo`3oool0oooo0000000A0?oo
o`030000003oool0oooo01d0oooo0`00000?0?ooo`<00000403oool2000002P0oooo000i0?ooo`<0
00003@3oool300000300oooo0P00000d0?ooo`<000003P3oool4000003X0oooo000i0?ooo`<00000
303oool300000340oooo0P00000c0?ooo`@00000403oool2000003X0oooo00190?ooo`800000<P3o
ool00`000000oooo0?ooo`0c0?ooo`800000403oool00`000000oooo0?ooo`0i0?ooo`00O@3oool2
000004H0oooo00<000000?ooo`3oool0>@3oool007d0oooo0P0000150?ooo`030000003oool0oooo
03X0oooo001n0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0>P3oool007h0oooo
0P0000140?ooo`030000003oool0oooo03X0oooo001n0?ooo`800000A03oool00`000000oooo0?oo
o`0j0?ooo`00O`3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo03X0oooo001o0?oo
o`800000@`3oool00`000000oooo0?ooo`0j0?ooo`00O`3oool200000480oooo00<000000?ooo`3o
ool0>`3oool00800oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0k0?ooo`00P03o
ool200000440oooo00<000000?ooo`3oool0>`3oool00840oooo00<000000?ooo`3oool0?`3oool0
0`000000oooo0?ooo`0k0?ooo`00P@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo
03@0oooo00<000000?ooo`3oool0103oool000L0oooo00<000000?ooo`3oool0M`3oool200000400
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`040?ooo`001`3oool00`000000oooo
0?ooo`1h0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0=03oool2000000H0oooo
00060?ooo`800000NP3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo03@0oooo0P00
00060?ooo`001P3oool3000007T0oooo0P00000n0?ooo`030000003oool0oooo02<0oooo00<00000
0?ooo`3oool03P3oool3000000D0oooo00050?ooo`@000003`3oool00`000000oooo0?ooo`1X0?oo
o`030000003oool0oooo03`0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0=0?oo
o`@000001@3oool000D0oooo1000000?0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3o
ool0?03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00d0oooo100000050?ooo`00
1P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo06P0oooo0P00000l0?ooo`800000
503oool00`000000oooo0?ooo`0=0?ooo`<000003`3oool00`000000oooo0?ooo`050?ooo`001P3o
ool00`000000oooo0?ooo`0>0?ooo`<00000403oool00`000000oooo0?ooo`1G0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0=0?ooo`<000003`3o
ool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0>0?ooo`<00000403oool0
0`000000oooo0?ooo`0^0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03P3oool2
000000<0oooo00<000000?ooo`3oool02@3oool2000002l0oooo00<000000?ooo`3oool04P3oool2
000000l0oooo0`00000?0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo
00h0oooo0`00000?0?ooo`800000403oool00`000000oooo0?ooo`0L0?ooo`<000003`3oool30000
00h0oooo100000030?ooo`800000203oool4000000d0oooo00<000000?ooo`3oool07`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00d0oooo0`00000?0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0>0?ooo`800000403oool00`000000oooo0?ooo`0K0?ooo`@00000403oool2000000l0
oooo0`0000040?ooo`030000003oool0oooo00L0oooo0`00000<0?ooo`<00000803oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00`0oooo1000000?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo
0?ooo`0=0?ooo`<000003`3oool200000100oooo00<000000?ooo`3oool0303oool3000000l0oooo
0`00000>0?ooo`<00000103oool00`000000oooo0?ooo`080?ooo`<000002`3oool500000140oooo
00<000000?ooo`3oool02@3oool2000000<0oooo0P00000>0?ooo`@000003`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool03@3oool3000000l0oooo0P00000?0?ooo`800000403oool2000000l0oooo0P00
000?0?ooo`800000103oool2000000X0oooo0P00000>0?ooo`<000003`3oool2000000/0oooo0P00
00030?ooo`800000403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool01@3oool003T0oooo0`00000>0?ooo`<000008P3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0M
0?ooo`<000002`3oool010000000oooo0?ooo`3oool3000003X0oooo000i0?ooo`<000003@3oool3
000003X0oooo00<000000?ooo`3oool0:@3oool3000000/0oooo00@000000?ooo`3oool0oooo0`00
000j0?ooo`00>P3oool2000000`0oooo1000000j0?ooo`800000:@3oool4000000/0oooo00<00000
0?ooo`3oool0?P3oool004X0oooo0P00000k0?ooo`030000003oool0oooo02T0oooo0P00000:0?oo
o`800000@03oool008L0oooo00<000000?ooo`3oool0=@3oool200000400oooo00270?ooo`800000
=P3oool200000400oooo00280?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0?`3o
ool008P0oooo00<000000?ooo`3oool0<`3oool200000440oooo00290?ooo`030000003oool0oooo
0380oooo0P0000110?ooo`00R@3oool00`000000oooo0?ooo`0b0?ooo`800000@@3oool008T0oooo
00<000000?ooo`3oool0<P3oool200000440oooo002:0?ooo`030000003oool0oooo0340oooo00<0
00000?ooo`3oool0@03oool008X0oooo00<000000?ooo`3oool0<03oool200000480oooo002:0?oo
o`800000<@3oool2000003/0oooo00<000000?ooo`3oool0103oool000L0oooo00<000000?ooo`3o
ool0P@3oool00`000000oooo0?ooo`0_0?ooo`800000>`3oool00`000000oooo0?ooo`040?ooo`00
1`3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0
>@3oool2000000H0oooo00060?ooo`800000P`3oool200000300oooo00<000000?ooo`3oool0>@3o
ool3000000D0oooo00060?ooo`<00000P`3oool00`000000oooo0?ooo`0]0?ooo`800000:`3oool0
0`000000oooo0?ooo`0<0?ooo`@000001@3oool000D0oooo1000000?0?ooo`030000003oool0oooo
0740oooo00<000000?ooo`3oool0;@3oool2000002/0oooo00<000000?ooo`3oool0303oool40000
00D0oooo00050?ooo`@000003`3oool00`000000oooo0?ooo`1a0?ooo`800000;P3oool00`000000
oooo0?ooo`0Y0?ooo`800000403oool00`000000oooo0?ooo`040?ooo`001`3oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0;03oool00`000000oooo
0?ooo`0Y0?ooo`800000403oool00`000000oooo0?ooo`040?ooo`001P3oool00`000000oooo0?oo
o`0>0?ooo`<00000L`3oool2000002d0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`0=0?ooo`@000003P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0>
0?ooo`<00000403oool00`000000oooo0?ooo`1P0?ooo`800000;03oool2000001X0oooo00<00000
0?ooo`3oool03@3oool4000000h0oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?oo
o`3oool03P3oool300000100oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo01P0oooo0P00000@0?ooo`<000003P3oool00`000000oooo0?ooo`050?oo
o`001P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo0P00000`0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0;0?ooo`@0
0000:@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00d0oooo0`00000?0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool0
3`3oool00`000000oooo0?ooo`0>0?ooo`800000403oool00`000000oooo0?ooo`0L0?ooo`<00000
3`3oool3000000l0oooo0`00000=0?ooo`@000003@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0303oool4000000l0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo00d0oooo0`00000?0?ooo`8000007@3oool400000100oooo0P00000?0?ooo`<000003P3o
ool3000000`0oooo0`00000C0?ooo`030000003oool0oooo00<0oooo0P0000090?ooo`8000003P3o
ool4000000l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo0`00000?0?ooo`800000403oool0
0`000000oooo0?ooo`0<0?ooo`<000003`3oool3000000l0oooo0P00000?0?ooo`<000002`3oool5
00000100oooo0P0000050?ooo`030000003oool0oooo00P0oooo0P00000@0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool30000
00l0oooo0P00000@0?ooo`8000003`3oool2000000l0oooo0P00000?0?ooo`<000003P3oool30000
00l0oooo0P0000050?ooo`030000003oool0oooo00L0oooo0`00000@0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`00>@3oool3000000h0
oooo0`00000R0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool2000001l0
oooo0`0000050?ooo`030000003oool0oooo00L0oooo0`00000j0?ooo`00>P3oool2000000h0oooo
0P00000d0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool07P3oool3000000@0oooo
0P0000160?ooo`00B@3oool3000004D0oooo0P00000N0?ooo`@00000103oool00`000000oooo0?oo
o`150?ooo`00BP3oool2000004H0oooo00<000000?ooo`3oool07P3oool2000000@0oooo00<00000
0?ooo`3oool0A@3oool00980oooo0P00000U0?ooo`030000003oool0oooo04D0oooo002C0?ooo`03
0000003oool0oooo0280oooo00<000000?ooo`3oool0AP3oool009<0oooo0P00000S0?ooo`030000
003oool0oooo04H0oooo002C0?ooo`8000008P3oool2000004P0oooo002D0?ooo`8000008@3oool2
000004P0oooo002D0?ooo`8000008@3oool00`000000oooo0?ooo`170?ooo`00U@3oool2000001l0
oooo0P0000120?ooo`030000003oool0oooo00@0oooo00070?ooo`030000003oool0oooo08/0oooo
0P00000O0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0103oool000L0oooo00<0
00000?ooo`3oool0S03oool2000001h0oooo00<000000?ooo`3oool0@03oool2000000H0oooo0007
0?ooo`030000003oool0oooo08`0oooo0P00000M0?ooo`030000003oool0oooo0440oooo0P000006
0?ooo`001P3oool3000008h0oooo0P00000L0?ooo`030000003oool0oooo0440oooo0`0000050?oo
o`001P3oool3000008l0oooo00<000000?ooo`3oool06@3oool200000480oooo100000050?ooo`00
1P3oool3000000l0oooo00<000000?ooo`3oool0O@3oool2000001X0oooo00<000000?ooo`3oool0
<P3oool00`000000oooo0?ooo`0<0?ooo`@000001@3oool000L0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`1n0?ooo`030000003oool0oooo01L0oooo0P00000d0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0103oool000L0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`1o0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0<P3oool2
00000100oooo00<000000?ooo`3oool0103oool000L0oooo00<000000?ooo`3oool03@3oool30000
0800oooo00<000000?ooo`3oool05@3oool2000003@0oooo0`00000>0?ooo`030000003oool0oooo
00D0oooo00060?ooo`030000003oool0oooo00h0oooo0`00000@0?ooo`030000003oool0oooo06h0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo00d0
oooo1000000>0?ooo`030000003oool0oooo00D0oooo00060?ooo`030000003oool0oooo00h0oooo
0`00000@0?ooo`030000003oool0oooo06h0oooo0P00000D0?ooo`800000903oool00`000000oooo
0?ooo`0=0?ooo`@000003P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo00h0oooo0P00001a0?ooo`030000003oool0oooo0180oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>
0?ooo`800000LP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`3oool03@3oool3000000l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`05
0?ooo`001P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0oooo0`00000@0?oo
o`030000003oool0oooo01d0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0P0?oo
o`030000003oool0oooo00T0oooo0P0000000`3oool000000?ooo`0>0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool03@3oool3000000l0oooo00<000000?ooo`3oool03P3oool00`00
0000oooo0?ooo`050?ooo`001P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00d0
oooo0`00000@0?ooo`030000003oool0oooo01`0oooo0`00000?0?ooo`<000003`3oool3000000h0
oooo0`00000<0?ooo`@000003@3oool00`000000oooo0?ooo`0C0?ooo`8000003`3oool3000000l0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`050?ooo`001P3oool00`000000oooo
0?ooo`0?0?ooo`030000003oool0oooo00h0oooo0P00000?0?ooo`8000007@3oool400000100oooo
0P00000@0?ooo`8000003`3oool2000000d0oooo0`00000<0?ooo`8000001@3oool00`000000oooo
0?ooo`0=0?ooo`800000403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo00h0oooo0P00000@0?ooo`030000003oool0oooo00`0oooo0`00
000?0?ooo`800000403oool2000000l0oooo0`00000;0?ooo`D000002P3oool2000000D0oooo0P00
000>0?ooo`<00000403oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool01@3oool000H0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo00d0oooo0`00000?0?ooo`800000403oool2000000l0oooo0P00000@
0?ooo`8000003`3oool2000000h0oooo0`0000070?ooo`<000001@3oool3000000h0oooo0`00000@
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`05
0?ooo`00>@3oool3000000h0oooo0`00000Q0?ooo`800000403oool200000100oooo00<000000?oo
o`3oool03@3oool2000000H0oooo0P0000060?ooo`<00000403oool2000003X0oooo00190?ooo`<0
00008`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0200oooo1P0000070?ooo`@0
0000C03oool004X0oooo0P00001W0?ooo`800000C03oool004/0oooo00<000000?ooo`3oool0/`3o
ool00?l0oooo0P3oool00?l0oooo0P3oool00?l0oooo0P3oool00?l0oooo0P3oool00?l0oooo0P3o
ool00?l0oooo0P3oool00?l0oooo0P3oool00?l0oooo0P3oool00001\
\>"],
  ImageRangeCache->{{{0, 256.625}, {430.875, 0}} -> {-3.16832, -5.32461, \
0.0370383, 0.0370383}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(phaseCurve[\(-3\), \(-3\), 3, 20];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.600406 0.185738 0.30897 0.0581398 [
[.04319 .29647 -9.34375 -15.6875 ]
[.04319 .29647 9.34375 0 ]
[.22893 .29647 -9.34375 -15.6875 ]
[.22893 .29647 9.34375 0 ]
[.41467 .29647 -9.34375 -15.6875 ]
[.41467 .29647 9.34375 0 ]
[.78614 .29647 -5 -15.6875 ]
[.78614 .29647 5 0 ]
[.97188 .29647 -5 -15.6875 ]
[.97188 .29647 5 0 ]
[.58791 .07641 -18.6875 -7.84375 ]
[.58791 .07641 0 7.84375 ]
[.58791 .19269 -18.6875 -7.84375 ]
[.58791 .19269 0 7.84375 ]
[.58791 .42525 -10 -7.84375 ]
[.58791 .42525 0 7.84375 ]
[.58791 .54153 -10 -7.84375 ]
[.58791 .54153 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.04319 .30897 m
.04319 .31522 L
s
gsave
.04319 .29647 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.22893 .30897 m
.22893 .31522 L
s
gsave
.22893 .29647 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.41467 .30897 m
.41467 .31522 L
s
gsave
.41467 .29647 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.78614 .30897 m
.78614 .31522 L
s
gsave
.78614 .29647 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.97188 .30897 m
.97188 .31522 L
s
gsave
.97188 .29647 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.08034 .30897 m
.08034 .31272 L
s
.11749 .30897 m
.11749 .31272 L
s
.15463 .30897 m
.15463 .31272 L
s
.19178 .30897 m
.19178 .31272 L
s
.26608 .30897 m
.26608 .31272 L
s
.30323 .30897 m
.30323 .31272 L
s
.34037 .30897 m
.34037 .31272 L
s
.37752 .30897 m
.37752 .31272 L
s
.45182 .30897 m
.45182 .31272 L
s
.48896 .30897 m
.48896 .31272 L
s
.52611 .30897 m
.52611 .31272 L
s
.56326 .30897 m
.56326 .31272 L
s
.63755 .30897 m
.63755 .31272 L
s
.6747 .30897 m
.6747 .31272 L
s
.71185 .30897 m
.71185 .31272 L
s
.749 .30897 m
.749 .31272 L
s
.82329 .30897 m
.82329 .31272 L
s
.86044 .30897 m
.86044 .31272 L
s
.89759 .30897 m
.89759 .31272 L
s
.93473 .30897 m
.93473 .31272 L
s
.00604 .30897 m
.00604 .31272 L
s
.25 Mabswid
0 .30897 m
1 .30897 L
s
.60041 .07641 m
.60666 .07641 L
s
gsave
.58791 .07641 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.60041 .19269 m
.60666 .19269 L
s
gsave
.58791 .19269 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.60041 .42525 m
.60666 .42525 L
s
gsave
.58791 .42525 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.60041 .54153 m
.60666 .54153 L
s
gsave
.58791 .54153 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.60041 .10548 m
.60416 .10548 L
s
.60041 .13455 m
.60416 .13455 L
s
.60041 .16362 m
.60416 .16362 L
s
.60041 .22176 m
.60416 .22176 L
s
.60041 .25083 m
.60416 .25083 L
s
.60041 .2799 m
.60416 .2799 L
s
.60041 .33804 m
.60416 .33804 L
s
.60041 .36711 m
.60416 .36711 L
s
.60041 .39618 m
.60416 .39618 L
s
.60041 .45432 m
.60416 .45432 L
s
.60041 .48339 m
.60416 .48339 L
s
.60041 .51246 m
.60416 .51246 L
s
.60041 .04734 m
.60416 .04734 L
s
.60041 .01827 m
.60416 .01827 L
s
.60041 .5706 m
.60416 .5706 L
s
.60041 .59967 m
.60416 .59967 L
s
.25 Mabswid
.60041 0 m
.60041 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.04319 .13455 m
.02808 .25994 L
.02527 .28715 L
.02414 .30066 L
.02381 .30748 L
.02387 .31148 L
.02413 .31345 L
.02446 .31449 L
.02492 .31518 L
.0254 .31553 L
.02594 .31574 L
.02653 .31585 L
.02683 .31589 L
.02716 .31592 L
.02776 .31596 L
.02835 .31598 L
.0289 .31599 L
.03012 .31602 L
.03127 .31603 L
.03344 .31607 L
.03834 .31614 L
.0591 .31649 L
.079 .31685 L
.1013 .31731 L
.12407 .31784 L
.14743 .31846 L
.17428 .31931 L
.20261 .3204 L
.23647 .32206 L
.25543 .32324 L
.27465 .32468 L
.29536 .32659 L
.32053 .32958 L
.33409 .3316 L
.35036 .33451 L
.38647 .34339 L
.4111 .35195 L
.44649 .36859 L
.50186 .40541 L
.59103 .48438 L
.66078 .54708 L
.6837 .56471 L
.70968 .58158 L
.73306 .5932 L
.74673 .59815 L
.7597 .60141 L
.78502 .60332 L
.80895 .59901 L
.82993 .58972 L
.8518 .57393 L
Mistroke
.87383 .55106 L
.89488 .52204 L
.92558 .46554 L
.96143 .37281 L
.96954 .34508 L
.97364 .32802 L
.97479 .3221 L
.97551 .31757 L
.97597 .31367 L
.97619 .31021 L
.97617 .3074 L
.97599 .30533 L
.97561 .30342 L
.97512 .30198 L
.97454 .30084 L
.97382 .29986 L
.97308 .29912 L
.97236 .29856 L
.97151 .29806 L
.97055 .29762 L
.9695 .29724 L
.96851 .29696 L
.96637 .29651 L
.96437 .29622 L
.96201 .29595 L
.95974 .29574 L
.94261 .29451 L
.92397 .29308 L
.90109 .29089 L
.8749 .2876 L
.85837 .28492 L
.84084 .28143 L
.81984 .27616 L
.79566 .26827 L
.76821 .25653 L
.72745 .23307 L
.66668 .18585 L
.61383 .13738 L
.54693 .07656 L
.50463 .04466 L
.48141 .03107 L
.46786 .02478 L
.45482 .02002 L
.44044 .0164 L
.42726 .01472 L
.41301 .0148 L
.39763 .0173 L
.38433 .02164 L
.37052 .02843 L
.34676 .04602 L
Mistroke
.32635 .06764 L
.30654 .09496 L
.27711 .14842 L
.25415 .20263 L
.23884 .24685 L
.23032 .27653 L
.22643 .29364 L
.22543 .29947 L
.22483 .30443 L
.22461 .3081 L
.22465 .31078 L
.22476 .31205 L
.22493 .31325 L
.22536 .31505 L
.22569 .31595 L
.22604 .31669 L
.22679 .31784 L
.22721 .31832 L
.22767 .31878 L
.22856 .31945 L
.22958 .32002 L
.23075 .32051 L
.23175 .32083 L
.23282 .3211 L
.2348 .32148 L
.23698 .32179 L
.23942 .32205 L
.24819 .32273 L
.26885 .32421 L
.28971 .32602 L
.31442 .32877 L
.34205 .33295 L
.35872 .33624 L
.37918 .34128 L
.40003 .34782 L
.42867 .35954 L
.54183 .43887 L
.58966 .48309 L
.62125 .51258 L
.66059 .54693 L
.68236 .56374 L
.70415 .57831 L
.72467 .58947 L
.74765 .59843 L
.77008 .60294 L
.79391 .60246 L
.81405 .59725 L
.8355 .58631 L
.84578 .57893 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3oool3
0?ooo`030000003oool0oooo0:/0oooo003o0?ooo`<0oooo00<000000?ooo`3oool0Z`3oool00?l0
oooo0`3oool00`000000oooo0?ooo`2[0?ooo`00o`3oool30?ooo`030000003oool0oooo0:/0oooo
003o0?ooo`<0oooo00<000000?ooo`3oool0Z`3oool00?l0oooo0`3oool00`000000oooo0?ooo`2[
0?ooo`00[`3oool<000004L0oooo00<000000?ooo`3oool0Z`3oool00:X0oooo1@00000<0?ooo`H0
0000@@3oool300000:/0oooo002W0?ooo`<000005`3oool5000003`0oooo00<000000?ooo`3oool0
Z`3oool00:D0oooo0P00000O0?ooo`800000>P3oool00`000000oooo0?ooo`2[0?ooo`00X`3oool2
000002<0oooo0P00000h0?ooo`030000003oool0oooo0:/0oooo002Q0?ooo`8000009`3oool20000
03H0oooo00<000000?ooo`3oool0Z`3oool009l0oooo0P00000[0?ooo`800000=03oool00`000000
oooo0?ooo`2[0?ooo`00W@3oool2000002l0oooo0P00000b0?ooo`030000003oool0oooo0:/0oooo
002L0?ooo`030000003oool0oooo0340oooo0P00000`0?ooo`030000003oool0oooo0:/0oooo002K
0?ooo`030000003oool0oooo03@0oooo0P00000^0?ooo`030000003oool0oooo0:/0oooo002I0?oo
o`800000>@3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0:/0oooo002H0?ooo`03
0000003oool0oooo03X0oooo0P00000[0?ooo`030000003oool0oooo0:/0oooo002F0?ooo`800000
?`3oool2000002T0oooo00<000000?ooo`3oool0Z`3oool009D0oooo00<000000?ooo`3oool0@@3o
ool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0:/0oooo002D0?ooo`030000003oool0
oooo04<0oooo00<000000?ooo`3oool09@3oool300000:/0oooo002C0?ooo`030000003oool0oooo
04D0oooo0P00000U0?ooo`030000003oool0oooo0:/0oooo002B0?ooo`030000003oool0oooo04P0
oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`2[0?ooo`00T@3oool00`000000oooo
0?ooo`1:0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0Z`3oool00900oooo00<0
00000?ooo`3oool0C03oool200000240oooo00<000000?ooo`3oool0Z`3oool00900oooo00<00000
0?ooo`3oool0CP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo0:/0oooo002?0?oo
o`030000003oool0oooo0500oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`2[0?oo
o`00SP3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0Z`3oool008d0oooo00<000000?ooo`3oool0E03oool2000000h0oooo00<0oooO09a80018W=l0
2`3oool00`000000oooo0?ooo`2[0?ooo`00S03oool00`000000oooo0?ooo`1G0?ooo`030000003o
ool0oooo00`0oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`2[0?ooo`00R`3oool0
0`000000oooo0?ooo`1I0?ooo`030000003oool0oooo00/0oooo00<0omnL04P0B02Lgol02P3oool0
0`000000oooo0?ooo`2[0?ooo`00R`3oool00`000000oooo0?ooo`1J0?ooo`8000001`3oool00`3o
_g@0000000000004000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0Z`3oool008X0
oooo00<000000?ooo`3oool0E`3oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo0?oo
g`2LB4P0W=oo0P3oool00`2oM00007Bo0?ooo`090?ooo`@00000ZP3oool008T0oooo00<000000?oo
o`3oool0G`3oool00`000000oooo0?ooo`040?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=oo
o`090?ooo`030000003oool0oooo0:/0oooo00280?ooo`030000003oool0oooo0640oooo00<00000
0?ooo`3oool0103oool01P3ogi`0B4RL0=ooo`3ogi`0B01809cOo`T0oooo00<000000?ooo`3oool0
Z`3oool008P0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`040?ooo`050?noM000
M;l0oooO09a80018W=l02@3oool00`000000oooo0?ooo`2[0?ooo`00Q`3oool00`000000oooo0?oo
o`1T0?ooo`030000003oool0oooo00@0oooo00@0gia804RLg`2oM00007Bo2@3oool00`000000oooo
0?ooo`2[0?ooo`00QP3oool00`000000oooo0?ooo`1V0?ooo`8000001@3oool0103OW4P0M7Ad0000
M02oool80?ooo`030000003oool0oooo0:/0oooo00250?ooo`030000003oool0oooo06T0oooo00<0
00000?ooo`3oool00`3oool00`2oM000001809cOo`080?ooo`030000003oool0oooo0:/0oooo0025
0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool00P3oool00`3ooml0W4P004RLg`08
0?ooo`030000003oool0oooo0:/0oooo00240?ooo`030000003oool0oooo06`0oooo00<000000?oo
o`3oool0303oool00`000000oooo0?ooo`2[0?ooo`00P`3oool00`000000oooo0?ooo`1^0?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0Z`3oool008<0oooo00<000000?ooo`3oool0
K`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0:/0oooo00220?ooo`030000003o
ool0oooo0740oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`2[0?ooo`00PP3oool0
0`000000oooo0?ooo`1b0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0Z`3oool0
0840oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`070?ooo`<00000Z`3oool00840
oooo00<000000?ooo`3oool0M@3oool2000000L0oooo00<000000?ooo`3oool0Z`3oool00800oooo
00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0:/0oooo
001o0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`2[0?ooo`00O`3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0Z`3oool007h0oooo00<000000?ooo`3oool0O@3oool01@000000oooo0?ooo`3oool0
00000:d0oooo001n0?ooo`030000003oool0oooo07h0oooo00@000000?ooo`3oool00000[@3oool0
07d0oooo00<000000?ooo`3oool0P03oool00`000000oooo0000002]0?ooo`00O@3oool00`000000
oooo0?ooo`210?ooo`800000[@3oool007`0oooo00<000000?ooo`3oool0P`3oool00`000000oooo
0?ooo`2[0?ooo`00O03oool00`000000oooo0?ooo`230?ooo`<00000Z`3oool007/0oooo00<00000
0?ooo`3oool0Q03oool010000000oooo0?ooo`00002Z0?ooo`00NP3oool00`000000oooo0?ooo`25
0?ooo`050000003oool0oooo0?ooo`000000Z@3oool00180oooo00<000000?ooo`3oool0I@3oool0
0`000000oooo0?ooo`250?ooo`<000000P3oool00`000000oooo0?ooo`2V0?ooo`004P3oool00`00
0000oooo0?ooo`1T0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`2U0?ooo`004P3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo08H0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2T0?ooo`004P3oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`2S0?ooo`004P3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo08L0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`2R0?ooo`004@3oool00`000000oooo0?ooo`1S
0?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`2Q
0?ooo`004@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo08P0oooo00<000000?oo
o`3oool0203oool200000:40oooo000A0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3o
ool0R@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo09h0oooo000A0?ooo`030000
003oool0oooo0680oooo00<000000?ooo`3oool0R@3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo09d0oooo000A0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool0RP3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo09`0oooo000A0?ooo`030000003oool0
oooo0640oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo09/0oooo000A0?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0R`3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo09X0oooo000A0?ooo`030000003oool0oooo0600
oooo00<000000?ooo`3oool0R`3oool3000000l0oooo00<000000?ooo`3oool0V@3oool00140oooo
00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0V03oool00100oooo00<000000?ooo`3oool0H03oool00`000000oooo0?oo
o`2<0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0U`3oool00100oooo00<00000
0?ooo`3oool0H03oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0UP3oool00100oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`2=0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool0U@3oool00100oooo00<000000?ooo`3o
ool0G`3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo01@0oooo0P00002E0?ooo`00
403oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`2B0?ooo`00403oool00`000000oooo0?ooo`1N0?ooo`030000003o
ool0oooo08h0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`2A0?ooo`00403oool0
0`000000oooo0?ooo`1N0?ooo`030000003oool0oooo07d0oooo00<0gia80000000000000`000000
0`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0T03oool0
0100oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`1o0?ooo`030?noM000M;l0oooo
0080oooo00<0_gAd0;ooo`3oool0203oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
08l0oooo000@0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0P03oool00`3o_g@0
07Bo0?ooo`0<0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0SP3oool000l0oooo
00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`220?ooo`030?noM000M;l0oooo00/0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`2=0?ooo`003`3oool00`000000oooo0?oo
o`1M0?ooo`030000003oool0oooo0780oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?oo
o`040?ooo`030?noM018W=l0oooo00X0oooo00<000000?ooo`3oool0703oool2000008d0oooo000?
0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0Q03oool00`3OW4P0B9cO0?ooo`09
0?ooo`@000007@3oool00`000000oooo0?ooo`2:0?ooo`003`3oool00`000000oooo0?ooo`1L0?oo
o`030000003oool0oooo08H0oooo00<0_g@007Boo`3oool0203oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo08T0oooo000?0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3o
ool0QP3oool00`3ogi`0B01d0;ooo`080?ooo`030000003oool0oooo0200oooo0P0000290?ooo`00
3`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo08L0oooo00<0oooO09a8001d_ol0
203oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo08H0oooo000?0?ooo`030000003o
ool0oooo05/0oooo00<000000?ooo`3oool0Q`3oool00`3ookl0M00004RLg`080?ooo`030000003o
ool0oooo02<0oooo00<000000?ooo`3oool0Q@3oool000l0oooo00<000000?ooo`3oool0F`3oool0
0`000000oooo0?ooo`230?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?ooo`030000
003oool0oooo02@0oooo0P0000250?ooo`003P3oool00`000000oooo0?ooo`1K0?ooo`030000003o
ool0oooo08D0oooo00D0omnL04P000000000000007Boo`090?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0PP3oool000h0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?oo
o`2C0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0P@3oool000h0oooo00<00000
0?ooo`3oool0F`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo02P0oooo00<00000
0?ooo`3oool0P03oool000h0oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`2D0?oo
o`030000003oool0oooo02T0oooo0P0000200?ooo`003P3oool00`000000oooo0?ooo`1J0?ooo`03
0000003oool0oooo09@0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1m0?ooo`00
3P3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo09@0oooo00<000000?ooo`3oool0
;03oool00`000000oooo0?ooo`1l0?ooo`003P3oool00`000000oooo0?ooo`1I0?ooo`030000003o
ool0oooo09D0oooo0`00000]0?ooo`800000O03oool000h0oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`2E0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0N@3oool0
00h0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo
0300oooo00<000000?ooo`3oool0N03oool000d0oooo00<000000?ooo`3oool0FP3oool00`000000
oooo0?ooo`2E0?ooo`030000003oool0oooo0340oooo0P00001h0?ooo`003@3oool00`000000oooo
0?ooo`1I0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0<`3oool00`000000oooo
0?ooo`1e0?ooo`003@3oool00`000000oooo0?ooo`1I0?ooo`030000003oool0oooo09H0oooo00<0
00000?ooo`3oool0=03oool00`000000oooo0?ooo`1d0?ooo`003@3oool00`000000oooo0?ooo`1I
0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0=@3oool2000007@0oooo000=0?oo
o`030000003oool0oooo05P0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`0g0?oo
o`800000LP3oool000d0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`2G0?ooo`03
0000003oool0oooo03T0oooo00<000000?ooo`3oool0K`3oool000d0oooo00<000000?ooo`3oool0
F03oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo03X0oooo0P00001_0?ooo`003@3o
ool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool0?03o
ool2000006d0oooo000=0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0V03oool0
0`000000oooo0?ooo`0n0?ooo`800000J`3oool000`0oooo00<000000?ooo`3oool0F03oool00`00
0000oooo0?ooo`2H0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0J03oool000`0
oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`2I0?ooo`<00000@@3oool2000006P0
oooo000<0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0V@3oool00`000000oooo
0?ooo`130?ooo`800000IP3oool000`0oooo00<000000?ooo`3oool01P3oool00`3ookl0M0000000
00020000000307Boo`3oool0oooo04L0oooo00<0gia80000000000000`0000000`18W=l0oooo0?oo
o`170?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool0AP3oool00`000000oooo0?oo
o`150?ooo`8000000P3oool00`3ookl0M000000000020000000304RLg`3oool0oooo04P0oooo00<0
gia80000000000000`0000000`18W=l0oooo0?ooo`090?ooo`00303oool00`000000oooo0?ooo`06
0?ooo`030?noM000000007Bo0080oooo00<0_gA809cOo`3oool0B03oool00`00000007Bo0?ooo`02
0?ooo`030;mdM02oool0oooo04T0oooo00<0okmd0000001d_ol0B@3oool00`000000oooo0?ooo`17
0?ooo`8000000P3oool00`3o_g@0000007Boo`1<0?ooo`030?noM000M;l0oooo0080oooo00<0_gAd
0;ooo`3oool02@3oool000`0oooo00<000000?ooo`3oool02`3oool00`3ogi`0B4RL0=ooo`180?oo
o`030000003o_g@007Bo04h0oooo00<0oooO09a80000M;l0B@3oool00`000000oooo0?ooo`190?oo
o`80000000<0oooO09a80000M;l0C@3oool00`3o_g@007Bo0?ooo`0=0?ooo`00303oool00`000000
oooo0?ooo`0;0?ooo`030?oo_`1d0000M;oo04P0oooo00@000000?ooo`3o_g@007BoCP3oool00`3O
W4P004RL0=ooo`180?ooo`030000003oool0oooo04/0oooo0`0000000`3Oool0oooo0?ooo`1;0?oo
o`030?noM000M;l0oooo00`0oooo00080?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3o
ool01@3oool00`3ogi`0B00007Boo`0j0?ooo`030?oog`2LB000000000L0000000D0B9cO0?ooo`3o
ool0oooo000000020?ooo`030?noM018W=l0oooo03`0oooo00<0oooO09a8000000001`0000000`18
W=l0oooo0?ooo`030?ooo`030?oOW01804P0W=oo04P0oooo00<000000?ooo`3oool0C03oool0103o
gi`0B01800000000001=0?ooo`030?noM018W=l0oooo00/0oooo000<0?ooo`030000003oool0oooo
00X0oooo00@0omnL04P0000004P0W=ooA`3oool00`000000oooo0?ooo`020?ooo`030=nLB018W=l0
oooo04/0oooo00<0oono07@0001d_ol0B03oool00`000000oooo0?ooo`1<0?ooo`040?oo_`1d0000
M;oo0?ooo`<00000B`3oool00`3OW4P0B9cO0?ooo`0:0?ooo`00303oool00`000000oooo0?ooo`08
0?ooo`050?oOW018000000000000001d_ol0B03oool00`000000oooo0?ooo`030?ooo`030;md001d
_ol0oooo04/0oooo00<0_g@0001d_`3oool0A`3oool00`000000oooo0?ooo`1=0?ooo`030;md0000
M;l0oooo00<0oooo100000180?ooo`030;md001d_ol0oooo00T0oooo000;0?ooo`030000003oool0
oooo00/0oooo00<0oono07@00000M;l0B03oool00`000000oooo0?ooo`030?ooo`030?oOW01807@0
_ooo04/0oooo00<0gia80018W03Oool0A`3oool00`000000oooo0?ooo`1=0?ooo`030=nLB000B9`0
gooo00L0oooo100000140?ooo`030?oOW01807@0_ooo00T0oooo000;0?ooo`030000003oool0oooo
00d0oooo00<0gia8001d_`3oool0A@3oool00`000000oooo0?ooo`040?ooo`030?oog`2LB000M;oo
04/0oooo00<0omnL04P0B02Lgol0A`3oool00`000000oooo0?ooo`1=0?ooo`030?oOW01804P0W=oo
00/0oooo0`0000110?ooo`030?oog`2LB000M;oo00T0oooo000;0?ooo`030000003oool0oooo00d0
oooo00<0omnL04P0B02Lgol0A@3oool00`000000oooo0?ooo`040?ooo`030?oo_`1d0000B9cO04/0
oooo00<0oono07@0001d_ol0A`3oool00`000000oooo0?ooo`1=0?ooo`030?oo_`1d0000M;oo00h0
oooo0`00000n0?ooo`030?oo_`1d0000B9cO00T0oooo000;0?ooo`030000003oool0oooo00X0oooo
00H0_g@007Boo`3oool0gia80000B02Lgom50?ooo`0:0000003oool0oooo0?noM000B9`0gooo0?oo
o`3OW4P0001809cOod`0oooo00<0_g@0001d_`3oool0AP3oool3000004h0oooo00<0_g@0001d_`3o
ool0403oool3000003L0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00T0oooo000;0?oo
o`030000003oool0oooo00/0oooo00@0_g@0000000000000B9cOA@3oool00`000000oooo0?ooo`02
0?ooo`050?oOW018000000000000001d_ol0B`3oool01@3ookl0M0000000000007@0_ooo04H0oooo
00<000000?ooo`3oool0C03oool01@3ookl0M0000000000007@0_ooo01<0oooo1@00000c0?ooo`05
0?oOW018000000000000001d_ol02P3oool000/0oooo00<000000?ooo`3oool0E03oool00`000000
oooo0?ooo`2M0?ooo`030000003oool0oooo06T0oooo1`00000k0?ooo`002`3oool00`000000oooo
0?ooo`1D0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0L03oool6000003D0oooo
000;0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?oo
o`1f0?ooo`H00000;`3oool000/0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`2N
0?ooo`030000003oool0oooo07`0oooo2000000W0?ooo`002P3oool00`000000oooo0?ooo`1D0?oo
o`030000003oool0oooo09h0oooo00<000000?ooo`3oool0Q03oool@000001L0oooo000:0?ooo`03
0000003oool0oooo05@0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`2D0?ooo`T0
00003P3oool000X0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`2N0?ooo`030000
003oool0oooo09`0oooo0`00000<0?ooo`002P3oool00`000000oooo0?ooo`1D0?ooo`030000003o
ool0oooo09h0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`0:0?ooo`002P3oool0
0`000000oooo0?ooo`1D0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool0WP3oool2
000000/0oooo000:0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0WP3oool00`00
0000oooo0?ooo`2O0?ooo`030000003oool0oooo00T0oooo000:0?ooo`030000003oool0oooo05@0
oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo00T0
oooo003o00000;0000000@3oool000<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool2000000l0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0000000;0?ooo`002P3oool2000000H0oooo00<00000
0?ooo`3oool0C03oool2000004l0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`1=
0?ooo`030000003oool0oooo04d0oooo00<000000?ooo`0000002`3oool000/0oooo;`00000W0?oo
o`030000003oool0oooo09h0oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`090?oo
o`00>P3ooolU000000<0oooo0P00002N0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3o
ool02P3oool005l0oooo4@00002B0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool0
2P3oool006X0oooo4@0000270?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool02P3o
ool007T0oooo3000001m0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool02P3oool0
0880oooo2`00001e0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool02P3oool008X0
oooo2000001`0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool02P3oool00900oooo
1`00001[0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool02`3oool009H0oooo1`00
001U0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool02`3oool009X0oooo1`00001Q
0?ooo`<00000W@3oool00`000000oooo0?ooo`0;0?ooo`00W`3oool6000005d0oooo00<000000?oo
o`3oool0W@3oool00`000000oooo0?ooo`0;0?ooo`00Y03oool4000005X0oooo00<000000?ooo`3o
ool0W03oool00`000000oooo0?ooo`0<0?ooo`00Z03oool3000005L0oooo00<000000?ooo`3oool0
W03oool00`000000oooo0?ooo`0<0?ooo`00Z`3oool3000005@0oooo00<000000?ooo`3oool0W03o
ool00`000000oooo0?ooo`0<0?ooo`00[@3oool300000580oooo00<000000?ooo`3oool0V`3oool0
0`000000oooo0?ooo`0=0?ooo`00/03oool3000004l0oooo00<000000?ooo`3oool0V`3oool00`00
0000oooo0?ooo`0=0?ooo`00/P3oool3000004d0oooo00<000000?ooo`3oool0V`3oool00`000000
oooo0?ooo`0=0?ooo`00]@3oool2000004/0oooo00<000000?ooo`3oool0V`3oool00`000000oooo
0?ooo`0=0?ooo`00]`3oool2000004T0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?oo
o`0>0?ooo`00^@3oool2000004L0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0>
0?ooo`00^`3oool2000004D0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0>0?oo
o`00_03oool3000004<0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0>0?ooo`00
_P3oool300000440oooo0`00002I0?ooo`030000003oool0oooo00l0oooo002o0?ooo`040000003o
ool00000000003l0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`0?0?ooo`00`03o
ool2000000030?ooo`000000oooo03d0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?oo
o`0?0?ooo`00`P3oool010000000oooo00000000000l0?ooo`030000003oool0oooo09T0oooo00<0
00000?ooo`3oool03`3oool00<<0oooo0P0000000`3oool000000?ooo`0j0?ooo`030000003oool0
oooo09P0oooo00<000000?ooo`3oool0403oool00<D0oooo00@000000?ooo`0000000000>@3oool0
0`000000oooo0?ooo`2H0?ooo`030000003oool0oooo0100oooo00360?ooo`80000000<0oooo0000
003oool0=`3oool00`000000oooo0?ooo`2G0?ooo`030000003oool0oooo0140oooo00380?ooo`04
0000003oool00000000003H0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`0A0?oo
o`00b@3oool2000000030?ooo`000000oooo03@0oooo00<000000?ooo`3oool0U`3oool00`000000
oooo0?ooo`0A0?ooo`00b`3oool010000000oooo00000000000c0?ooo`030000003oool0oooo09H0
oooo00<000000?ooo`3oool04P3oool00<`0oooo00@000000?ooo`3oool00000<P3oool00`000000
oooo0?ooo`2F0?ooo`030000003oool0oooo0180oooo003=0?ooo`80000000<0oooo000000000000
<03oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo01<0oooo003?0?ooo`040000003o
ool0oooo000002l0oooo0`00002E0?ooo`030000003oool0oooo01<0oooo003@0?ooo`80000000<0
oooo000000000000;@3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo01<0oooo003B
0?ooo`040000003oool0oooo000002`0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?oo
o`0D0?ooo`00d`3oool2000000030?ooo`000000000002X0oooo00<000000?ooo`3oool0U03oool0
0`000000oooo0?ooo`0D0?ooo`00e@3oool010000000oooo0?ooo`00000Y0?ooo`030000003oool0
oooo09<0oooo00<000000?ooo`3oool05@3oool00=H0oooo00@000000?ooo`3oool00000:03oool0
0`000000oooo0?ooo`2C0?ooo`030000003oool0oooo01D0oooo003G0?ooo`80000000<0oooo0000
003oool09P3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo01D0oooo003I0?ooo`03
0000003oool0000002H0oooo00<000000?ooo`3oool0TP3oool00`000000oooo0?ooo`0F0?ooo`00
fP3oool4000001<0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`080?ooo`030000
003oool0oooo0980oooo00<000000?ooo`3oool05P3oool00=`0oooo00<000000?ooo`0000004`3o
ool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00P0oooo00<000000?ooo`3oool0T@3o
ool00`000000oooo0?ooo`0G0?ooo`00g@3oool3000001<0oooo00<0okmd001d_`3oool0303oool0
0`000000oooo0?ooo`2A0?ooo`030000003oool0oooo01L0oooo003O0?ooo`8000004`3oool00`3o
_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool05`3oool00>00
oooo0P00000C0?ooo`030?noM018W=l0oooo00X0oooo00<000000?ooo`3oool0T03oool00`000000
oooo0?ooo`0H0?ooo`00hP3oool00`000000oooo0?ooo`0A0?ooo`030=nLB018W=l0oooo00T0oooo
1000002?0?ooo`030000003oool0oooo01P0oooo003S0?ooo`030000003oool0oooo0140oooo00<0
_g@007Boo`3oool0203oool00`000000oooo0?ooo`2?0?ooo`030000003oool0oooo01T0oooo003T
0?ooo`8000004@3oool00`3ogi`0B01d0;ooo`080?ooo`030000003oool0oooo08l0oooo00<00000
0?ooo`3oool06@3oool00>D0oooo0P00000@0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?oo
o`3oool0S`3oool00`000000oooo0?ooo`0I0?ooo`00iP3oool3000000h0oooo00<0oono07@00018
W=l0203oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo01X0oooo003X0?ooo`800000
2@3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol0203oool00`000000oooo0?ooo`2>0?oo
o`030000003oool0oooo01X0oooo003Y0?ooo`8000002@3oool01@3ogi`0B000000000000000M;oo
00T0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0K0?ooo`00jP3oool2000001H0
oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0K0?ooo`00j`3oool2000001D0oooo
00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`0K0?ooo`00k03oool2000001@0oooo00<0
00000?ooo`3oool0S03oool00`000000oooo0?ooo`0L0?ooo`00k@3oool300000180oooo00<00000
0?ooo`3oool0S03oool00`000000oooo0?ooo`0L0?ooo`00kP3oool00`000000oooo0000000A0?oo
o`030000003oool0oooo08/0oooo00<000000?ooo`3oool07@3oool00>l0oooo0`00000@0?ooo`<0
0000R`3oool00`000000oooo0?ooo`0M0?ooo`00l@3oool2000000l0oooo00<000000?ooo`3oool0
R`3oool00`000000oooo0?ooo`0M0?ooo`00lP3oool2000000h0oooo00<000000?ooo`3oool0RP3o
ool00`000000oooo0?ooo`0N0?ooo`00l`3oool2000000d0oooo00<000000?ooo`3oool0RP3oool0
0`000000oooo0?ooo`0N0?ooo`00m03oool2000000`0oooo00<000000?ooo`3oool0R@3oool00`00
0000oooo0?ooo`0O0?ooo`00m@3oool2000000/0oooo00<000000?ooo`3oool0R@3oool00`000000
oooo0?ooo`0O0?ooo`00mP3oool2000000X0oooo00<000000?ooo`3oool0R03oool00`000000oooo
0?ooo`0P0?ooo`00m`3oool3000000P0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?oo
o`0P0?ooo`00n03oool00`000000oooo000000070?ooo`030000003oool0oooo08L0oooo00<00000
0?ooo`3oool08@3oool00?T0oooo0`0000060?ooo`030000003oool0oooo08L0oooo00<000000?oo
o`3oool08@3oool00?/0oooo0P0000050?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3o
ool08P3oool00?`0oooo0P0000040?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0
8P3oool00?d0oooo0P0000030?ooo`030000003oool0oooo08D0oooo00<000000?ooo`3oool08`3o
ool00?h0oooo0P0000020?ooo`<00000Q@3oool00`000000oooo0?ooo`0S0?ooo`00o`3oool20000
00030?ooo`000000oooo08D0oooo00<000000?ooo`3oool0903oool00?l0oooo0@3oool3000008H0
oooo00<000000?ooo`3oool0903oool00?l0oooo0P3oool2000008D0oooo00<000000?ooo`3oool0
9@3oool00?l0oooo0`3oool2000008<0oooo00<000000?ooo`3oool09P3oool00?l0oooo0`3oool3
00000880oooo00<000000?ooo`3oool09P3oool00?l0oooo0`3oool010000000oooo0?ooo`000020
0?ooo`030000003oool0oooo02L0oooo003o0?ooo`<0oooo00D000000?ooo`3oool0oooo0000001o
0?ooo`030000003oool0oooo02L0oooo003o0?ooo`<0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`1k0?ooo`030000003oool0oooo02P0oooo003o0?ooo`<0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo02P0oooo003o0?ooo`<0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo02T0oooo
003o0?ooo`<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1g0?ooo`030000003o
ool0oooo02T0oooo003o0?ooo`<0oooo0`0000060?ooo`030000003oool0oooo07D0oooo00<00000
0?ooo`3oool0:P3oool00?l0oooo0`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
07@0oooo00<000000?ooo`3oool0:P3oool00?l0oooo0`3oool00`000000oooo0?ooo`080?ooo`80
0000L`3oool00`000000oooo0?ooo`0[0?ooo`00o`3oool30?ooo`030000003oool0oooo00T0oooo
0P00001b0?ooo`030000003oool0oooo02/0oooo003o0?ooo`<0oooo00<000000?ooo`3oool02P3o
ool200000700oooo00<000000?ooo`3oool0;03oool00?l0oooo0`3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo06`0oooo00<000000?ooo`3oool0;@3oool00?l0oooo0`3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo06/0oooo00<000000?ooo`3oool0;@3oool00?l0
oooo0`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo06T0oooo00<000000?ooo`3o
ool0;P3oool00?@0oooo00<0oooO09a80018W=l02`3oool00`000000oooo0?ooo`0?0?ooo`030000
003oool0oooo06L0oooo00<000000?ooo`3oool0;`3oool00?D0oooo00<0gia80018W03Oool02P3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo06H0oooo00<000000?ooo`3oool0;`3o
ool00?D0oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo06@0oooo00<000000?ooo`3oool0<03oool00?40oooo00<0okmd000000000000100000000`00
B9`0gooo0?ooo`070?ooo`030000003oool0oooo0180oooo0P00001S0?ooo`030000003oool0oooo
0340oooo003a0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool02@3oool00`000000
oooo0?ooo`0D0?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool0<P3oool00?80oooo
00L0oono07@0M02oool0oooo0?noM000B9`0gooo00T0oooo1000000D0?ooo`030000003oool0oooo
05h0oooo00<000000?ooo`3oool0<P3oool00?<0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol9
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`0c
0?ooo`00m03oool01@3o_g@007Bo0?oog`2LB000B9cO00T0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo03@0oooo003e0?ooo`040=nLB018W=l0_g@0
001d_`T0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1I0?ooo`030000003oool0
oooo03@0oooo003f0?ooo`040=nLB01dM7@0001d0;ooo`P0oooo00<000000?ooo`3oool06@3oool2
000005P0oooo00<000000?ooo`3oool0=@3oool00?L0oooo00<0_g@00000B02Lgol0203oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0=P3oool00?L0
oooo00<0oooO09a80018W=l0203oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0580
oooo00<000000?ooo`3oool0=`3oool00?l0oooo0`3oool00`000000oooo0?ooo`0M0?ooo`800000
D@3oool00`000000oooo0?ooo`0h0?ooo`00o`3oool30?ooo`030000003oool0oooo01h0oooo0P00
001?0?ooo`030000003oool0oooo03T0oooo003o0?ooo`<0oooo00<000000?ooo`3oool0803oool0
0`000000oooo0?ooo`1;0?ooo`030000003oool0oooo03X0oooo003o0?ooo`<0oooo00<000000?oo
o`3oool08@3oool2000004X0oooo00<000000?ooo`3oool0>`3oool00?l0oooo0`3oool00`000000
oooo0?ooo`0R0?ooo`<00000A`3oool00`000000oooo0?ooo`0l0?ooo`00o`3oool30?ooo`<00000
903oool2000004D0oooo00<000000?ooo`3oool0?@3oool00?l0oooo0`3oool00`000000oooo0?oo
o`0U0?ooo`<00000@P3oool00`000000oooo0?ooo`0n0?ooo`00o`3oool30?ooo`030000003oool0
oooo02L0oooo0`00000o0?ooo`030000003oool0oooo03l0oooo003o0?ooo`<0oooo00<000000?oo
o`3oool0:@3oool2000003`0oooo0P0000120?ooo`00o`3oool30?ooo`030000003oool0oooo02X0
oooo0`00000i0?ooo`800000@`3oool00?l0oooo0`3oool00`000000oooo0?ooo`0/0?ooo`<00000
=@3oool3000004@0oooo003o0?ooo`<0oooo00<000000?ooo`3oool0;@3oool400000380oooo0P00
00160?ooo`00o`3oool30?ooo`030000003oool0oooo02l0oooo1000000^0?ooo`800000B03oool0
0?l0oooo0`3oool00`000000oooo0?ooo`0a0?ooo`@00000:P3oool3000004T0oooo003o0?ooo`<0
oooo00<000000?ooo`3oool0<`3oool4000002H0oooo0`00001;0?ooo`00o`3oool30?ooo`030000
003oool0oooo03D0oooo1@00000Q0?ooo`<00000C@3oool00?l0oooo0`3oool00`000000oooo0?oo
o`0h0?ooo`@000007@3oool3000004l0oooo003o0?ooo`<0oooo00<000000?ooo`3oool0>`3oool5
000001D0oooo1@00001A0?ooo`00o`3oool30?ooo`<00000?`3oool9000000P0oooo1`00001D0?oo
o`00o`3oool30?ooo`030000003oool0oooo04@0oooo3`00001H0?ooo`00o`3oool30?ooo`030000
003oool0oooo0:/0oooo003o0?ooo`<0oooo00<000000?ooo`3oool0Z`3oool00?l0oooo0`3oool0
0`000000oooo0?ooo`2[0?ooo`00o`3oool30?ooo`030000003oool0oooo0:/0oooo003o0?ooo`<0
oooo00<000000?ooo`3oool0Z`3oool00?l0oooo0`3oool00`000000oooo0?ooo`2[0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.23759, -5.31431, \
0.0187727, 0.0599728}}]
}, Closed]],

Cell[TextData[{
  "It is also important to recognize that limit cycles for large \
\[CurlyEpsilon] are traversed at rather uneven rates and that their periods \
are much longer than the underlying 2\[Pi] period for free oscillations (",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] = 0\)]],
  ")."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(xyCurves[\(-3\), \(-3\), 3, 30];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.031746 0.309021 0.0581083 [
[.18254 .29652 -5 -15.6875 ]
[.18254 .29652 5 0 ]
[.34127 .29652 -8 -15.6875 ]
[.34127 .29652 8 0 ]
[.5 .29652 -8 -15.6875 ]
[.5 .29652 8 0 ]
[.65873 .29652 -8 -15.6875 ]
[.65873 .29652 8 0 ]
[.81746 .29652 -8 -15.6875 ]
[.81746 .29652 8 0 ]
[.97619 .29652 -8 -15.6875 ]
[.97619 .29652 8 0 ]
[.01131 .07659 -18.6875 -7.84375 ]
[.01131 .07659 0 7.84375 ]
[.01131 .1928 -18.6875 -7.84375 ]
[.01131 .1928 0 7.84375 ]
[.01131 .42524 -10 -7.84375 ]
[.01131 .42524 0 7.84375 ]
[.01131 .54145 -10 -7.84375 ]
[.01131 .54145 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .30902 m
.18254 .31527 L
s
gsave
.18254 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.34127 .30902 m
.34127 .31527 L
s
gsave
.34127 .29652 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.5 .30902 m
.5 .31527 L
s
gsave
.5 .29652 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(15) show
1.000 setlinewidth
grestore
.65873 .30902 m
.65873 .31527 L
s
gsave
.65873 .29652 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(20) show
1.000 setlinewidth
grestore
.81746 .30902 m
.81746 .31527 L
s
gsave
.81746 .29652 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(25) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(30) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 .30902 m
.05556 .31277 L
s
.0873 .30902 m
.0873 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.15079 .30902 m
.15079 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.24603 .30902 m
.24603 .31277 L
s
.27778 .30902 m
.27778 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.37302 .30902 m
.37302 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.43651 .30902 m
.43651 .31277 L
s
.46825 .30902 m
.46825 .31277 L
s
.53175 .30902 m
.53175 .31277 L
s
.56349 .30902 m
.56349 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.62698 .30902 m
.62698 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.72222 .30902 m
.72222 .31277 L
s
.75397 .30902 m
.75397 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.84921 .30902 m
.84921 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.9127 .30902 m
.9127 .31277 L
s
.94444 .30902 m
.94444 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .07659 m
.03006 .07659 L
s
gsave
.01131 .07659 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.02381 .1928 m
.03006 .1928 L
s
gsave
.01131 .1928 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .42524 m
.03006 .42524 L
s
gsave
.01131 .42524 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.02381 .54145 m
.03006 .54145 L
s
gsave
.01131 .54145 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .10564 m
.02756 .10564 L
s
.02381 .1347 m
.02756 .1347 L
s
.02381 .16375 m
.02756 .16375 L
s
.02381 .22186 m
.02756 .22186 L
s
.02381 .25091 m
.02756 .25091 L
s
.02381 .27997 m
.02756 .27997 L
s
.02381 .33807 m
.02756 .33807 L
s
.02381 .36713 m
.02756 .36713 L
s
.02381 .39618 m
.02756 .39618 L
s
.02381 .45429 m
.02756 .45429 L
s
.02381 .48335 m
.02756 .48335 L
s
.02381 .5124 m
.02756 .5124 L
s
.02381 .04753 m
.02756 .04753 L
s
.02381 .01848 m
.02756 .01848 L
s
.02381 .57051 m
.02756 .57051 L
s
.02381 .59956 m
.02756 .59956 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.02381 .1347 m
.02499 .13054 L
.02605 .12914 L
.02729 .12866 L
.02846 .12865 L
.02954 .12879 L
.03053 .12897 L
.03279 .12944 L
.04262 .13162 L
.06244 .13612 L
.10458 .1462 L
.14393 .1564 L
.18177 .16714 L
.22205 .17996 L
.26083 .19436 L
.28054 .20289 L
.30205 .21375 L
.31202 .21962 L
.32264 .22677 L
.3326 .23477 L
.34176 .24396 L
.34666 .25007 L
.35204 .25835 L
.35745 .26937 L
.36041 .27724 L
.36317 .2864 L
.36578 .29732 L
.36826 .31042 L
.3729 .34437 L
.37787 .38893 L
.37922 .39931 L
.38069 .4084 L
.38207 .41488 L
.38334 .41919 L
.3845 .42199 L
.38559 .42381 L
.38677 .42515 L
.38804 .426 L
.38868 .42627 L
.38938 .42646 L
.39064 .42659 L
.39186 .42653 L
.39299 .42636 L
.39422 .42609 L
.39537 .42577 L
.39798 .42492 L
.40325 .42291 L
.42167 .41491 L
.44025 .40562 L
.46049 .39324 L
Mistroke
.47148 .38479 L
.48167 .37488 L
.4914 .36182 L
.49684 .3515 L
.50177 .33854 L
.50431 .32966 L
.50704 .31767 L
.5096 .30318 L
.51194 .28668 L
.51707 .24145 L
.51848 .22915 L
.51998 .21778 L
.52137 .20929 L
.52268 .20321 L
.52376 .19944 L
.52493 .19646 L
.52622 .19421 L
.52741 .19288 L
.52861 .19207 L
.52992 .1916 L
.53058 .19149 L
.53128 .19145 L
.53255 .19152 L
.53376 .19171 L
.53507 .19201 L
.53745 .19273 L
.54189 .19435 L
.58049 .21222 L
.60041 .22431 L
.61056 .23196 L
.62146 .24225 L
.62673 .2485 L
.63172 .25572 L
.6362 .26385 L
.63852 .26895 L
.64108 .27558 L
.64569 .29141 L
.64822 .30324 L
.65059 .31722 L
.65483 .3504 L
.6594 .39134 L
.66179 .40766 L
.66313 .41415 L
.66435 .41851 L
.66549 .42145 L
.66672 .42366 L
.66788 .42503 L
.66896 .42583 L
.67021 .42636 L
.67138 .42656 L
Mistroke
.67266 .42657 L
.67338 .4265 L
.67404 .4264 L
.67536 .42611 L
.6768 .42572 L
.67942 .42485 L
.70051 .41601 L
.72184 .40543 L
.74118 .39363 L
.75086 .38634 L
.76134 .3766 L
.76615 .37106 L
.77123 .36406 L
.77557 .35667 L
.78033 .34616 L
.78284 .33908 L
.78557 .32959 L
.78798 .31912 L
.79054 .3051 L
.79538 .2683 L
.79985 .2281 L
.80232 .21092 L
.80371 .20408 L
.80498 .1995 L
.8061 .19661 L
.80733 .1944 L
.80861 .19292 L
.8098 .19209 L
.81048 .19179 L
.81123 .19159 L
.81193 .19148 L
.81258 .19145 L
.8138 .19152 L
.8151 .19173 L
.81578 .19188 L
.8165 .19206 L
.8178 .19244 L
.82077 .19346 L
.84053 .20176 L
.86207 .2124 L
.88271 .22505 L
.89191 .23205 L
.90185 .24133 L
.90738 .24774 L
.91245 .25491 L
.91731 .26357 L
.91978 .269 L
.92242 .27588 L
.92483 .28346 L
.92702 .29181 L
Mistroke
.93131 .31388 L
.93371 .33063 L
.93632 .3526 L
.9386 .37358 L
.94105 .3945 L
.94217 .40243 L
.94324 .40875 L
.94427 .41373 L
.94524 .41738 L
.94615 .42006 L
.94711 .42224 L
.94819 .42397 L
.94916 .42506 L
.95031 .42589 L
.95137 .42633 L
.95254 .42656 L
.95377 .42658 L
.95484 .42647 L
.95583 .42629 L
.95693 .42603 L
.9581 .4257 L
.97619 .4185 L
Mfstroke
0 0 1 r
.02381 .1347 m
.02499 .24359 L
.02605 .28541 L
.02729 .3048 L
.02846 .31167 L
.02954 .3142 L
.03053 .3152 L
.03163 .31567 L
.03279 .31588 L
.03395 .31597 L
.0352 .31602 L
.03638 .31604 L
.03746 .31605 L
.04262 .31611 L
.06244 .31634 L
.10458 .31692 L
.14429 .31762 L
.18248 .31851 L
.20215 .31911 L
.22313 .31989 L
.24454 .32093 L
.25435 .32152 L
.26471 .32225 L
.27428 .32307 L
.28479 .32416 L
.29434 .32541 L
.30326 .32691 L
.30827 .32794 L
.31371 .32929 L
.31851 .33072 L
.32361 .33259 L
.32819 .3347 L
.33325 .33768 L
.33785 .34126 L
.34217 .34578 L
.34686 .35268 L
.34952 .35798 L
.352 .36422 L
.35462 .37274 L
.357 .38286 L
.35975 .39853 L
.36235 .41893 L
.36459 .44266 L
.36697 .47568 L
.37123 .55397 L
.37237 .57494 L
.3736 .59304 L
.37476 .60244 L
.37581 .60167 L
.37705 .58737 L
Mistroke
.37775 .57293 L
.3784 .55603 L
.38086 .47527 L
.38333 .39945 L
.38458 .37107 L
.38595 .34775 L
.38667 .33834 L
.38743 .33007 L
.38814 .32381 L
.38882 .31894 L
.38948 .31498 L
.39019 .31146 L
.39145 .30682 L
.39265 .30373 L
.39396 .30137 L
.39463 .30048 L
.39534 .29969 L
.39663 .2986 L
.39785 .29787 L
.39913 .29731 L
.40022 .29695 L
.40142 .29664 L
.40387 .29618 L
.40609 .29588 L
.41107 .29536 L
.42004 .29448 L
.42938 .2934 L
.43926 .29196 L
.44442 .29103 L
.4501 .28982 L
.46007 .287 L
.4653 .285 L
.47026 .28257 L
.47471 .27978 L
.47956 .27576 L
.48482 .26959 L
.48748 .26537 L
.49041 .25946 L
.49291 .253 L
.49559 .24397 L
.49813 .23265 L
.50046 .21892 L
.5096 .10766 L
.51223 .05688 L
.51352 .03454 L
.51471 .01987 L
.5158 .01472 L
.51694 .02046 L
.5181 .03891 L
.51936 .07209 L
Mistroke
.52203 .16147 L
.5235 .2065 L
.5249 .24071 L
.52619 .26461 L
.52688 .27453 L
.5276 .28334 L
.52889 .2952 L
.53008 .30296 L
.53137 .30886 L
.5321 .31133 L
.53277 .31317 L
.53401 .31575 L
.53531 .31764 L
.53648 .31884 L
.53757 .31966 L
.53877 .32032 L
.54006 .32084 L
.54128 .32121 L
.54262 .32153 L
.54535 .32198 L
.54787 .32229 L
.55024 .32254 L
.56131 .32363 L
.57093 .32476 L
.58144 .32635 L
.59128 .32835 L
.59664 .32976 L
.60163 .33137 L
.60624 .33318 L
.61059 .33531 L
.61559 .33845 L
.62014 .34225 L
.62478 .34754 L
.62741 .35149 L
.62988 .35606 L
.63242 .36199 L
.63484 .36914 L
.63763 .37997 L
.64017 .3933 L
.64239 .40871 L
.64476 .43044 L
.64699 .45757 L
.64901 .48871 L
.65139 .53288 L
.65264 .55729 L
.65397 .58078 L
.65513 .59621 L
.6564 .60332 L
.65755 .5977 L
.65863 .58094 L
Mistroke
.66366 .42481 L
.66504 .38811 L
.6665 .3586 L
.66774 .34031 L
.66845 .33231 L
.66912 .326 L
.67041 .31675 L
.67159 .31081 L
.67285 .30633 L
.67357 .30444 L
.67424 .30301 L
.67546 .30101 L
.67656 .29971 L
.67784 .29863 L
.67904 .2979 L
.68011 .29741 L
.6813 .297 L
.68253 .29667 L
.68368 .29643 L
.68575 .29609 L
.68796 .29581 L
.69761 .29485 L
.70778 .29375 L
.71884 .29223 L
.72929 .29028 L
.73394 .28918 L
.73896 .28777 L
.74398 .28604 L
.74858 .28407 L
.75298 .28172 L
.75762 .27854 L
.76008 .27644 L
.76271 .27379 L
.76745 .2675 L
.77035 .26228 L
.77309 .25593 L
.77555 .24861 L
.77821 .23823 L
.78057 .22609 L
.7831 .20865 L
.78549 .18642 L
.78767 .15958 L
.79029 .11777 L
.79307 .06444 L
.79443 .03986 L
.7957 .02231 L
.79684 .01494 L
.79808 .01942 L
.79939 .04003 L
.80005 .05611 L
Mistroke
.80077 .07717 L
.80324 .16051 L
.80465 .20391 L
.80616 .24106 L
.80758 .26684 L
.80887 .28363 L
.81018 .29561 L
.8114 .30337 L
.81203 .30646 L
.81273 .30927 L
.81346 .31168 L
.81414 .31347 L
.81538 .31597 L
.81603 .31696 L
.81671 .31783 L
.81786 .31896 L
.81912 .31985 L
.82044 .32051 L
.82166 .32096 L
.823 .32134 L
.82442 .32164 L
.8268 .32201 L
.82942 .32232 L
.83876 .32325 L
.84972 .32445 L
.85972 .32586 L
.86972 .32772 L
.8792 .33015 L
.88439 .33192 L
.8892 .33397 L
.89448 .33686 L
.9001 .34106 L
.90497 .34618 L
.90945 .35286 L
.91179 .3575 L
.91436 .36388 L
.91704 .37251 L
.91954 .38311 L
.92191 .39642 L
.92416 .41301 L
.92618 .43236 L
.92837 .45948 L
.93324 .5448 L
.93446 .56806 L
.93563 .58709 L
.93667 .59897 L
.93781 .60327 L
.93888 .59678 L
.94002 .57773 L
.9411 .54968 L
Mistroke
.94208 .51848 L
.94447 .43784 L
.94667 .37915 L
.94785 .35674 L
.94909 .33903 L
.95034 .32621 L
.95104 .32067 L
.9517 .31642 L
.9529 .3105 L
.95399 .30664 L
.95525 .30347 L
.95644 .30137 L
.95757 .29995 L
.95882 .29881 L
.96013 .29798 L
.96136 .29741 L
.9625 .29701 L
.96354 .29672 L
.9659 .29625 L
.9683 .29592 L
.97091 .29562 L
.97619 .29511 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`008@3oool00`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000
003oool0oooo0?l0ooooS@3oool00240oooo00<000000?ooo`3oool0o`3ooon=0?ooo`008@3oool0
0`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000003oool0oooo0?l0ooooS@3oool00240
oooo00<000000?ooo`3oool0o`3ooon=0?ooo`008@3oool00`000000oooo0?ooo`340?ooo`030000
o`3oool0oooo06l0oooo00<0003o0?ooo`3oool0D`3oool00240oooo00<000000?ooo`3oool0a03o
ool00`000?l0oooo0?ooo`1_0?ooo`030000o`3oool0oooo05<0oooo000Q0?ooo`800000a@3oool0
0`000?l0oooo0?ooo`1^0?ooo`80003oE@3oool00240oooo00<000000?ooo`3oool0a03oool00`00
0?l0oooo0?ooo`1^0?ooo`80003oE@3oool00240oooo00<000000?ooo`3oool0a03oool00`000?l0
oooo0?ooo`1^0?ooo`80003oE@3oool00240oooo00<000000?ooo`3oool0``3oool20000og00oooo
0P000?mE0?ooo`008@3oool00`000000oooo0?ooo`330?ooo`80003oL03oool20000oeD0oooo000Q
0?ooo`030000003oool0oooo0<<0oooo00<0003o0?ooo`000?l0K`3oool00`000?l0oooo0000o`1D
0?ooo`008@3oool00`000000oooo0?ooo`330?ooo`030000o`3oool0003o06l0oooo00<0003o0?oo
o`000?l0E03oool00240oooo00<000000?ooo`3oool0``3oool00`000?l0oooo0000o`1_0?ooo`03
0000o`3oool0003o05@0oooo000Q0?ooo`030000003oool0oooo0<<0oooo00<0003o0?ooo`000?l0
K`3oool00`000?l0oooo0000o`1D0?ooo`008@3oool00`000000oooo0?ooo`330?ooo`030000o`3o
ool0003o06l0oooo00<0003o0?ooo`000?l0E03oool00240oooo00<000000?ooo`3oool0``3oool0
0`000?l0oooo0000o`1_0?ooo`030000o`3oool0003o05@0oooo000Q0?ooo`030000003oool0oooo
0<<0oooo00<0003o0?ooo`000?l0K`3oool00`000?l0oooo0000o`1D0?ooo`008@3oool200000<@0
oooo00<0003o0?ooo`000?l0K`3oool00`000?l0oooo0000o`1D0?ooo`008@3oool00`000000oooo
0?ooo`330?ooo`030000o`3oool0003o06l0oooo00<0003o0?ooo`000?l0E03oool00240oooo00<0
00000?ooo`3oool0``3oool00`000?l0oooo0000o`1^0?ooo`040000o`3oool0oooo0000oe@0oooo
000Q0?ooo`030000003oool0oooo0<<0oooo00<0003o0?ooo`000?l0KP3oool010000?l0oooo0?oo
o`000?mD0?ooo`008@3oool00`000000oooo0?ooo`330?ooo`030000o`3oool0003o06h0oooo00@0
003o0?ooo`3oool0003oE03oool00240oooo00<000000?ooo`3oool0``3oool00`000?l0oooo0000
o`1^0?ooo`040000o`3oool0oooo0000oe@0oooo000C0?ooo`030?oog`2LB000B9cO00/0oooo00<0
00000?ooo`3oool0``3oool00`000?l0oooo0000o`1^0?ooo`040000o`3oool0oooo0000oe@0oooo
000D0?ooo`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool0``3oool00`000?l0oooo0000
o`1^0?ooo`040000o`3oool0oooo0000oe@0oooo000D0?ooo`030?oOW01804P0W=oo00X0oooo00<0
00000?ooo`3oool0``3oool00`000?l0oooo0000o`1^0?ooo`040000o`3oool0oooo0000oe@0oooo
000@0?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool01`3oool00`000000oooo0?oo
o`320?ooo`040000o`3oool0oooo0000ofh0oooo00@0003o0?ooo`3oool0003oE03oool000<0oooo
00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ooml0W4Q809cOo`80oooo00<0_g@0001d
_`3oool02@3oool00`000000oooo0?ooo`320?ooo`040000o`3oool0oooo0000ofh0oooo00@0003o
0?ooo`3oool0003oE03oool00140oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo00T0oooo
0`0000320?ooo`040000o`3oool0oooo0000ofh0oooo00@0003o0?ooo`3oool0003oE03oool00180
oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol90?ooo`030000003oool0oooo0<80oooo00@0003o
0?ooo`3oool0003oKP3oool010000?l0oooo0?ooo`000?mD0?ooo`004`3oool01@3o_g@007Bo0?oo
g`2LB000B9cO00T0oooo00<000000?ooo`3oool0`P3oool010000?l0oooo0?ooo`000?m^0?ooo`04
0000o`3oool0oooo0000oe@0oooo000D0?ooo`040=nLB018W=l0_g@0001d_`T0oooo00<000000?oo
o`3oool0`P3oool010000?l0oooo0?ooo`000?m^0?ooo`040000o`3oool0oooo0000oe@0oooo000E
0?ooo`040=nLB01dM7@0001d0;ooo`P0oooo00<000000?ooo`3oool0`P3oool010000?l0oooo0?oo
o`000?m^0?ooo`040000o`3oool0oooo0000oe@0oooo000F0?ooo`030;md000004P0W=oo00P0oooo
00<000000?ooo`3oool0`P3oool010000?l0oooo0?ooo`000?m^0?ooo`040000o`3oool0oooo0000
oe@0oooo000F0?ooo`030?oog`2LB000B9cO00P0oooo00<000000?ooo`3oool0`P3oool010000?l0
oooo0?ooo`000?m]0?ooo`050000o`3oool0oooo0?ooo`000?l0E03oool00240oooo00<000000?oo
o`3oool0`P3oool010000?l0oooo0?ooo`000?m]0?ooo`050000o`3oool0oooo0?ooo`000?l0E03o
ool00240oooo00<000000?ooo`3oool0`@3oool01@000?l0oooo0?ooo`3oool0003o06d0oooo00D0
003o0?ooo`3oool0oooo0000o`1D0?ooo`008@3oool00`000000oooo0?ooo`310?ooo`050000o`3o
ool0oooo0?ooo`000?l0K@3oool01@000?l0oooo0?ooo`3oool0003o05@0oooo000Q0?ooo`030000
003oool0oooo0<40oooo00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`050000o`3oool0oooo0?oo
o`000?l0E03oool00240oooo00<000000?ooo`3oool0`@3oool01@000?l0oooo0?ooo`3oool0003o
06d0oooo00D0003o0?ooo`3oool0oooo0000o`1D0?ooo`008@3oool200000<80oooo00D0003o0?oo
o`3oool0oooo0000o`1]0?ooo`050000o`3oool0oooo0?ooo`000?l0E03oool00240oooo00<00000
0?ooo`3oool0`@3oool01@000?l0oooo0?ooo`3oool0003o06d0oooo00D0003o0?ooo`3oool0oooo
0000o`1D0?ooo`008@3oool00`000000oooo0?ooo`310?ooo`050000o`3oool0oooo0?ooo`000?l0
K@3oool01@000?l0oooo0?ooo`3oool0003o05@0oooo000Q0?ooo`030000003oool0oooo0<40oooo
00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`050000o`3oool0oooo0?ooo`000?l0E03oool00240
oooo00<000000?ooo`3oool0`@3oool01@000?l0oooo0?ooo`3oool0003o06d0oooo00D0003o0?oo
o`3oool0oooo0000o`1D0?ooo`008@3oool00`000000oooo0?ooo`310?ooo`050000o`3oool0oooo
0?ooo`000?l0K@3oool01@000?l0oooo0?ooo`3oool0003o05@0oooo000Q0?ooo`030000003oool0
oooo0<40oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`1Z0?ooo`030000o`3oool0
oooo0080oooo00<0003o0?ooo`3oool0D@3oool00240oooo00<000000?ooo`3oool0`@3oool00`00
0?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`1A0?ooo`008@3oool00`000000oooo0?ooo`310?ooo`030000o`3oool0oooo0080
oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo0540
oooo000Q0?ooo`030000003o0000oooo0<00oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo
0?ooo`1Z0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0D@3oool00240oooo1@3o
002n0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?oo
o`020?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030?l0003oool0oooo0080oooo103o002j
0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`02
0?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`000000oooo00H0oooo103o002f0?oo
o`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?oo
o`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo00X0oooo1@3o002a0?ooo`03
0000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`03
0000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo00l0oooo103o002]0?ooo`030000
o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000
o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo01<0oooo103o002Y0?ooo`030000o`3o
ool0oooo00<0oooo00<0003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3o
ool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo01L0oooo103o002U0?ooo`030000o`3oool0
oooo00<0oooo00<0003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0
oooo0540oooo000Q0?ooo`030000o`3oool0oooo01/0oooo103o002Q0?ooo`030000o`3oool0oooo
00<0oooo00<0003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo
0540oooo000Q0?ooo`030000o`3oool0oooo01l0oooo103o002M0?ooo`030000o`3oool0oooo00<0
oooo00<0003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540
oooo000Q0?ooo`030000o`3oool0oooo02<0oooo103o002I0?ooo`030000o`3oool0oooo00<0oooo
00<0003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo
000Q0?ooo`030000o`3oool0oooo02L0oooo103o002E0?ooo`030000o`3oool0oooo00<0oooo00<0
003o0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q
0?ooo`030000o`3oool0oooo02/0oooo103o002A0?ooo`030000o`3oool0oooo00<0oooo00<0003o
0?ooo`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?oo
o`030000o`3oool0oooo02l0oooo103o002=0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?oo
o`3oool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`03
0000o`3oool0oooo03<0oooo103o00290?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3o
ool0J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000
o`000000oooo03L0oooo103o00240?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0
J@3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3o
ool0oooo03/0oooo103o00200?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J@3o
ool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0
oooo03l0oooo0`3o001m0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J@3oool0
0`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo
0480oooo0`3o001j0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J@3oool00`00
0?l0oooo0?ooo`030?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo04D0
oooo103o001f0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J03oool00`000?l0
oooo0?ooo`040?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo04T0oooo
0`3o001c0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J03oool00`000?l0oooo
0?ooo`040?ooo`030000o`3oool0oooo0540oooo000Q0?ooo`030000o`3oool0oooo04`0oooo0`3o
001`0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J03oool00`000?l0oooo0?oo
o`040?ooo`030000o`3oool0oooo0540oooo000@0?ooo`030=nLB0000000000000<0000000<0B9cO
0?ooo`3oool0203oool00`000?l0oooo0?ooo`1?0?ooo`80o`00KP3oool00`000?l0oooo0?ooo`04
0?ooo`030000o`3oool0oooo06P0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`1A
0?ooo`004@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00P0oooo00<0003o0?oo
o`3oool0D@3oool30?l006/0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`1X0?oo
o`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0D03oool00180oooo00<0okmd001d_`3o
ool0303oool00`000?l0oooo0?ooo`1D0?ooo`<0o`00J03oool00`000?l0oooo0?ooo`050?ooo`03
0000o`3oool0oooo06L0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?ooo`1@0?ooo`00
4`3oool00`3o_g@007Bo0?ooo`0;0?ooo`030000o`3oool0oooo05L0oooo0P3o001V0?ooo`030000
o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0I`3oool00`000?l0oooo0?ooo`050?ooo`030000
o`3oool0oooo0500oooo00030?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool0103o
ool00`3o_g@0B9cO0?ooo`0:0?ooo`030000o`000000000005T0oooo0`3o001S0?ooo`030000o`3o
ool0oooo00D0oooo00<0003o0?ooo`3o0000103o001S0?ooo`030000o`3oool0oooo00D0oooo00<0
003o0?ooo`3o00000`3o001=0?ooo`005@3oool00`3OW4P0B9cO0?ooo`090?ooo`030000o`3oool0
oooo05`0oooo0P3o001Q0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?l0003o0000103oool3
0?l00600oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?l000030?ooo`<0o`00BP3oool0
01H0oooo00<0_g@007Boo`3oool0203oool00`000?l0oooo0?ooo`1N0?ooo`80o`00G`3oool00`00
0?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00L0oooo0P3o001N0?ooo`030000o`3oool0oooo
00D0oooo00<0003o0?l0003oool01P3oool20?l004P0oooo000F0?ooo`030?oOW01807@0_ooo00P0
oooo00<0003o0?ooo`3oool0H03oool20?l005d0oooo00<0003o0?ooo`3oool01@3oool00`000?l0
oooo0?ooo`090?ooo`80o`00F`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00P0
oooo0`3o00150?ooo`005P3oool00`3ooml0W4P007Boo`080?ooo`030000o`3oool0oooo0680oooo
0P3o001J0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool02`3oool20?l005T0oooo
00<0003o0?ooo`3oool01P3oool00`000?l0oooo0?ooo`0;0?ooo`80o`00@`3oool001H0oooo00<0
oono07@00018W=l0203oool00`000?l0oooo0?ooo`1T0?ooo`80o`00F03oool00`000?l0oooo0?oo
o`060?ooo`030000o`3oool0oooo00d0oooo0P3o001G0?ooo`030000o`3oool0oooo00H0oooo00<0
003o0?ooo`3oool03@3oool20?l00440oooo000B0?ooo`070?noM000B9`0gooo0?ooo`3OW4P00018
09cOo`080?ooo`030000o`3oool0oooo06H0oooo0P3o001F0?ooo`030000o`3oool0oooo00D0oooo
00<0o`000000o`3oool0403oool20?l005D0oooo00<0003o0?ooo`3oool01P3oool00`000?l0oooo
0?ooo`0?0?ooo`80o`00?`3oool001<0oooo00D0omnL04P000000000000007Boo`090?ooo`030000
o`3oool0oooo06P0oooo0P3o001D0?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool0
4P3oool20?l005<0oooo00<0003o0?ooo`3oool01@3oool00`3o0000003o0?ooo`0B0?ooo`80o`00
?@3oool00240oooo00<0003o0?ooo`3oool0JP3oool20?l00580oooo00<0003o0?ooo`3oool01@3o
ool00`3o0000003o0?ooo`0D0?ooo`80o`00D@3oool00`000?l0oooo0?ooo`050?ooo`030?l00000
0?l0oooo01@0oooo0P3o000k0?ooo`008@3oool00`000?l0oooo0?ooo`1/0?ooo`80o`00D03oool0
0`000?l0oooo0?ooo`050?ooo`030?l000000?l0oooo01H0oooo0P3o001?0?ooo`030000o`3oool0
oooo00D0oooo00<0o`000000o`3oool05P3oool20?l003T0oooo000Q0?ooo`030000o`3oool0oooo
06h0oooo0P3o001>0?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool0603oool00`3o
0000oooo0?ooo`1<0?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool0603oool00`3o
0000oooo0?ooo`0f0?ooo`008@3oool00`000?l0oooo0?ooo`1`0?ooo`80o`00C03oool00`000?l0
oooo0?ooo`050?ooo`030?l000000?l0oooo01T0oooo0P3o001;0?ooo`030000o`3oool0oooo00H0
oooo00<0o`000000o`3oool06@3oool20?l003H0oooo000Q0?ooo`030000o`000000oooo0780oooo
0P3o001:0?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool06`3oool20?l004T0oooo
00<0003o0?ooo`3oool01@3oool00`3o0000oooo0000o`0L0?ooo`030?l0003oool0oooo03<0oooo
000Q0?ooo`030000o`3oool0oooo07@0oooo0P3o00180?ooo`030000o`3oool0oooo00@0oooo00<0
o`000?ooo`000?l07P3oool00`3o0000oooo0?ooo`160?ooo`030000o`3oool0oooo00D0oooo00@0
o`000?ooo`3oool0003o703oool20?l003<0oooo000Q0?ooo`030000o`3oool0oooo07H0oooo00<0
o`000?ooo`3oool0A03oool00`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo0000oah0oooo
00<0o`000?ooo`3oool0A@3oool00`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo0000oah0
oooo00<0o`000?ooo`3oool0<03oool00240oooo00<0003o0?ooo`3oool0M`3oool00`3o0000oooo
0?ooo`130?ooo`030000o`3oool0oooo00D0oooo00@0o`000?ooo`3oool0003o7`3oool20?l004D0
oooo00<0003o0?ooo`3oool01@3oool0103o0000oooo0?ooo`000?lO0?ooo`80o`00<03oool00240
oooo00<0003o0?ooo`3oool0N03oool20?l004<0oooo00<0003o0?ooo`3oool01@3oool0103o0000
oooo0?ooo`000?lQ0?ooo`030?l0003oool0oooo0480oooo00<0003o0?ooo`3oool01@3oool0103o
0000oooo0?ooo`000?lQ0?ooo`030?l0003oool0oooo02d0oooo000Q0?ooo`030000o`3oool0oooo
07X0oooo00<0o`000?ooo`3oool0@03oool00`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo
0000ob80oooo00<0o`000?ooo`3oool0@03oool00`000?l0oooo0?ooo`060?ooo`040?l0003oool0
oooo0000ob80oooo00<0o`000?ooo`3oool0;03oool00240oooo00<0003o0?ooo`3oool0N`3oool0
0`3o0000oooo0?ooo`0o0?ooo`030000o`3oool0oooo00@0oooo00D0o`000?ooo`3oool0oooo0000
o`0S0?ooo`030?l0003oool0oooo03l0oooo00<0003o0?ooo`3oool01P3oool0103o0000oooo0?oo
o`000?lS0?ooo`030?l0003oool0oooo02/0oooo000Q0?ooo`030000o`3oool0oooo07/0oooo00<0
o`000?ooo`3oool0?P3oool00`000?l0oooo0?ooo`050?ooo`050?l0003oool0oooo0?ooo`000?l0
903oool00`3o0000oooo0?ooo`0n0?ooo`030000o`3oool0oooo00D0oooo00D0o`000?ooo`3oool0
oooo0000o`0T0?ooo`030?l0003oool0oooo02X0oooo000Q0?ooo`030000o`3oool0oooo07`0oooo
00<0o`000?ooo`3oool0?@3oool00`000?l0oooo0?ooo`050?ooo`050?l0003oool0oooo0?ooo`00
0?l09@3oool00`3o0000oooo0?ooo`0m0?ooo`030000o`3oool0oooo00D0oooo00D0o`000?ooo`3o
ool0oooo0000o`0U0?ooo`030?l0003oool0oooo02T0oooo000Q0?ooo`030000o`3oool0oooo07d0
oooo00<0o`000?ooo`3oool0?03oool00`000?l0oooo0?ooo`050?ooo`050?l0003oool0oooo0?oo
o`000?l09P3oool00`3o0000oooo0?ooo`0k0?ooo`030000o`3oool0oooo00H0oooo00D0o`000?oo
o`3oool0oooo0000o`0V0?ooo`030?l0003oool0oooo02P0oooo000Q0?ooo`030000o`3oool0oooo
07h0oooo00<0o`000?ooo`3oool0>`3oool00`000?l0oooo0?ooo`050?ooo`050?l0003oool0oooo
0?ooo`000?l09`3oool00`3o0000oooo0?ooo`0j0?ooo`030000o`3oool0oooo00H0oooo00D0o`00
0?ooo`3oool0oooo0000o`0W0?ooo`030?l0003oool0oooo02L0oooo000Q0?ooo`030000o`3oool0
oooo07h0oooo00<0o`000?ooo`3oool0>P3oool00`000?l0oooo0?ooo`060?ooo`050?l0003oool0
oooo0?ooo`000?l0:03oool00`3o0000oooo0?ooo`0i0?ooo`030000o`3oool0oooo00H0oooo00D0
o`000?ooo`3oool0oooo0000o`0X0?ooo`030?l0003oool0oooo02H0oooo000Q0?ooo`030000o`00
0000oooo07l0oooo00<0o`000?ooo`3oool0>@3oool00`000?l0oooo0?ooo`060?ooo`050?l0003o
ool0oooo0?ooo`000?l0:@3oool00`3o0000oooo0?ooo`0h0?ooo`030000o`3oool0oooo00H0oooo
00D0o`000?ooo`3oool0oooo0000o`0Y0?ooo`030?l0003oool0oooo02D0oooo000Q0?ooo`030000
o`3oool0oooo0800oooo00<0o`000?ooo`3oool0>03oool00`000?l0oooo0?ooo`050?ooo`030?l0
003oool0oooo0080oooo00<0003o0?ooo`3oool09`3oool00`3o0000oooo0?ooo`0g0?ooo`030000
o`3oool0oooo00L0oooo00D0o`000?ooo`3oool0oooo0000o`0Y0?ooo`030?l0003oool0oooo02D0
oooo000Q0?ooo`030000o`3oool0oooo03P0oooo00H0omnL04P00000000000000018W03Ooolf0?oo
o`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool00`3oool0102oM00000000?l0001d_olb
0?ooo`030?oo_`1d000000000080000000H0003o0?ooo`3oool0oooo0?oOW0180002000000030018
W03o0000oooo00<0oooo00<0003o0?ooo`3oool0:03oool00`3o0000oooo0=nLB0050000000304RL
g`3oool0oooo0080oooo00@0_g@0000000000000M;oo:03oool00`000?l0oooo0?ooo`060?ooo`03
0=nLB03o000000000080000000H0003o04RLg`3oool0oooo0?oOW0180002000000030018W03Oool0
oooo0200oooo00<0o`000?ooo`3oool03@3oool00`3ookl0M000000000020000000307Boo`3oool0
oooo00<0oooo00@0_g@0000000000000M;oo203oool00240oooo00<0003o0?ooo`3oool0>03oool0
1P3o_g@0000007Boo`3oool0omnL04Qd_cP0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo
0?ooo`3ogi`0o`0004RLgc<0oooo00@0okmd0000001d_ol0003o0`3oool01P3o_g@0000007Boo`3o
ool0omnL0?l000@0oooo00<0003o0?ooo`3oool0:@3oool0103o0000oooo0?noM000M;l30?ooo`0:
0;mdM02oool0oooo0?ooo`3OW4P007Bo0?ooo`3ogi`0B00004RLgbL0oooo00<0003o0?ooo`3oool0
1P3oool00`3o0000okmd001d_`020?ooo`0:0000o`2oM7@0_ooo0?ooo`3o_g@0000007Boo`3oool0
omnL04Qd_b80oooo00<0o`000?ooo`3oool0303oool00`3o_g@00000001d_`020?ooo`030;mdB02L
gol0oooo0080oooo00H0gia8001d_`3oool0omnL04P00018W=l70?ooo`008@3oool00`000?l0oooo
0?ooo`0m0?ooo`030?noM018W=l0oooo03H0oooo00<0oooO09a80000M;l0103oool00`3ooml0W4P0
04RLg`020?ooo`030?l0000004P0W=oo0380oooo00<0oooO09a800000?l02@3oool00`3o0000B9cO
0?ooo`020?ooo`030000o`3oool0oooo02T0oooo00D0o`000?ooo`3oool0okmd001d_`050?ooo`03
0?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol09@3oool00`000?l0oooo0?ooo`070?ooo`06
0?l0003oool0okmd001d_`3oool0003o203oool00`3o_g@0B9cO0?ooo`0P0?ooo`030?l0003oool0
oooo0140oooo00L0omnL04Q8W03Oool0oooo0?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0
1P3oool00240oooo00<0003o0?ooo`3oool0?@3oool00`3ookl0M01d0;ooo`0g0?ooo`030=nLB000
B9`0gooo00<0oooo00<0oooO09a80018W=l00P3oool0103ooml0o`000018W03Ooolb0?ooo`030=nL
B0000?l0gooo00P0oooo00<0o`0007@0M02oool00P3oool00`000?l0oooo0?ooo`0Z0?ooo`050?l0
003oool0oooo0?noM000M;l0103oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=oo
ob@0oooo00<0003o0?ooo`3oool01`3oool01P3o0000oooo0?ooo`3o_g@007Bo0000o`P0oooo00<0
oono07@0M02oool08@3oool00`3o0000oooo0?ooo`0@0?ooo`070?oo_`1d0000M;oo0?ooo`3ooml0
W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo000Q0?ooo`03000000000?l0oooo03d0
oooo00<0oooO09a8001d_ol0=`3oool00`3ogi`0B01809cOo`030?ooo`030?oog`2LB00007Bo00<0
oooo00<0o`000000B02Lgol0<P3oool00`000?l0B01809cOo`080?ooo`030?l0002LB000M;oo0080
oooo00<0003o0?ooo`3oool0:P3oool01P3o0000oooo0?ooo`3oool0okmd04RLg`<0oooo00<0oooO
09a80000M;l00`3oool00`3OW4P0001809cOo`0S0?ooo`030000o`3oool0oooo00P0oooo00H0o`00
0?ooo`3oool0oooo0?noM0000?l80?ooo`030?oog`2LB000M;oo0240oooo00<0o`000?ooo`3oool0
403oool01`3ogi`0B00007Boo`3oool0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?oo
o`008@3oool00`000000003o0?ooo`0m0?ooo`030?oo_`1d0000B9cO03L0oooo00<0oono07@0001d
_ol0103oool00`2oM00004RL0=ooo`020?ooo`030?oOW03o0000B9cO0340oooo00@0003o0?oo_`1d
0000M;oo203oool00`3o0000M00004RLg`020?ooo`030000o`3oool0oooo02/0oooo00H0o`000?oo
o`3oool0oooo0=nLB018W=l30?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l08`3o
ool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo0080oooo00<0003o04RLg`3oool01P3o
ool00`3ookl0M00004RLg`0R0?ooo`030?l0003oool0oooo00h0oooo00@0omnL04P0000004P0W=oo
0P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00D0oooo000Q0?ooo`0300000000
0?l0oooo03d0oooo00<0okmd0000001d_ol0>03oool00`2oM00007Bo0?ooo`030?ooo`030=nLB000
04P0W=oo0080oooo00<0oooO0?l00000M;l0<@3oool010000?l0oooo0;md0000M;l80?ooo`030?l0
00000000M;oo0080oooo00<0003o0?ooo`3oool0:`3oool00`3o0000oooo0?ooo`020?ooo`070;md
001d_ol0oooo0?ooo`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo0280oooo00<0003o0?oo
o`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`2oM000M;oo00H0oooo00<0okmd0000
001d_ol08P3oool00`3o0000oooo0?ooo`0<0?ooo`050?oOW018000000000000001d_ol00`3oool0
0`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00D0oooo000Q0?ooo`03000000000?l0oooo
03X0oooo00H0omoO0=ooo`3o_g@000000000M02ooolh0?ooo`030=nLB000B9`0gooo00<0oooo00<0
omnL04P00018W=l00`3oool00`3o000004RL0=ooo`0_0?ooo`060000o`3oool0oooo0=nLB000B9`0
gooo103oool01P3ogml0gooo0?noM03o0000001d0;ooo`80oooo00<0003o0?ooo`3oool0;03oool0
2`3o0000oooo0?ooo`3oool0omnL04P0M02oool0oooo0?oOW0180000B9cO00<0oooo00<0_g@00018
W03Oool0803oool00`000?l0oooo0?ooo`0:0?ooo`030?l0003oool0oooo0080oooo00@0003o0?oO
W01807@0_ooo0P3oool01P3ogml0gooo0?noM0000000001d0;ooob<0oooo00<0o`000?ooo`3oool0
3@3oool00`3ookl0M000001d_`030?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool0
103oool00240oooo00<000000000o`3oool0>P3oool01@3ookl0M0000000000004P0W=oo03T0oooo
00<0omnL04P0B02Lgol00`3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB03o0000gooo02h0
oooo00L0003o0?ooo`3oool0oooo0?oOW01804P0W=oo00@0oooo00D0oono07@000000000o`0009cO
o`030?ooo`030000o`3oool0oooo02`0oooo00`0o`000?ooo`3oool0oooo0?oog`2LB000M;oo0?oo
o`3ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo01l0oooo00<0003o0?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`030?ooo`030000o`2LB000M;oo0080oooo00D0oono07@0000000000018
09cOo`0T0?ooo`030?l0003oool0oooo00l0oooo00P0gia8001d_`3oool0oooo0?oog`2LB00004RL
0=ooo`80oooo00<0gia80018W03Oool0103oool00240oooo00<000000000o`3oool0>`3oool00`3O
W9`0_ooo0?ooo`0j0?ooo`030?oo_`1d0000M;oo00@0oooo00<0omnL04P0001d_ol00P3oool00`3O
W4P0o`000?ooo`0]0?ooo`030000o`3oool0oooo0080oooo00<0oono07@0001d_ol01@3oool00`3O
W9`0o`000?ooo`040?ooo`030000o`3oool0oooo02`0oooo00L0o`000?ooo`3oool0oooo0?oo_`1d
0000B9cO0080oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`0N0?ooo`030000o`3o
ool0oooo00/0oooo00<0o`000?ooo`3oool0103oool00`000?l0M00004RLg`030?ooo`030=nLW02o
ool0oooo02D0oooo00<0o`000?ooo`3oool03`3oool00`3ogi`0B01809cOo`020?ooo`030?oOW018
0000M;oo0080oooo00<0gia8001d_`3oool0103oool00240oooo00<000000000o`3oool0>`3oool0
0`3ogi`0M9cO0?ooo`0k0?ooo`030;md0000M;l0oooo00@0oooo00H0gia80000M02oool0oooO09a8
003o000]0?ooo`030000o`3oool0oooo00@0oooo00<0_g@0001d_`3oool0103oool00`3ogi`0o`00
0?ooo`050?ooo`030000o`3oool0oooo02/0oooo00L0okmd0?l0003Oool0oooo0=nLB00004P0W=oo
00<0oooo00H0gia80000M02oool0oooO09a80018W=lN0?ooo`030000o`3oool0oooo00`0oooo00X0
o`000?ooo`3oool0okmd0018W03Oool0oooo0000o`0004P0W=oo0`3oool00`3ogi`0M9cO0?ooo`0V
0?ooo`030?l0003oool0oooo00/0oooo00H0_g@007Boo`3oool0gia80000B02Lgol30?ooo`060=nL
B00007@0_ooo0?oog`2LB000B9cO1@3oool00240oooo00<000000000o`3oool0?03oool01@2oM000
0000000000000000B9cO03H0oooo00D0oono07@000000000001d0;ooo`050?ooo`060=nLB0000000
0000001d_`3oool0o`00:P3oool20000o`D0oooo00D0oono07@000000000001d0;ooo`050?ooo`05
0?l000000000000000000018W=l00P3oool00`000?l0oooo0?ooo`0/0?ooo`050?l0001800000000
0000001d_ol01@3oool0103OW4P0000000000000M;lL0?ooo`<0003o3`3oool02@3o0000oooo0?oo
o`3oool0omnL04P000000000003o07Boo`050?ooo`050;md00000000000000000018W=l08`3oool0
0`3o0000oooo0?ooo`0<0?ooo`040;md00000000000004RLg`D0oooo00@0gia800000000000007Bo
1P3oool00240oooo00<000000000o`3oool0QP3oool00`3o0000oooo0?ooo`0V0?ooo`80003o4@3o
ool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo02`0oooo00<0o`000?ooo`3oool09@3o
ool20000oa80oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0/0?ooo`030?l0003o
ool0oooo01l0oooo000Q0?ooo`03000000000?l0oooo08H0oooo00<0o`000?ooo`3oool08`3oool3
0000oa<0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0]0?ooo`030?l0003oool0
oooo0280oooo0P000?lD0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool0;03oool0
0`3o0000oooo0?ooo`0O0?ooo`008@3oool00`000000003o0?ooo`260?ooo`030?l0003oool0oooo
01h0oooo1@000?lF0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool0;@3oool00`3o
0000oooo0?ooo`0M0?ooo`D0003o5P3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo
02d0oooo00<0o`000?ooo`3oool07P3oool00240oooo00<000000000o`3oool0Q`3oool00`3o0000
oooo0?ooo`0F0?ooo`L0003o6`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo02d0
oooo00<0o`000?ooo`3oool05@3oool80000oa/0oooo00<0o`000?ooo`3oool0103oool00`000?l0
oooo0?ooo`0]0?ooo`030?l0003oool0oooo01h0oooo000Q0?ooo`03000000000?l0oooo08L0oooo
00<0o`000?ooo`3oool03@3oool90000ob40oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo
0?ooo`0]0?ooo`030?l0003oool0oooo00`0oooo2@000?lS0?ooo`030?l0003oool0oooo00@0oooo
00<0003o0?ooo`3oool0;@3oool00`3o0000oooo0?ooo`0<0?ooo`D0003o3@3oool00240oooo00<0
00000000o`3oool0Q`3oool00`3o0000oooo0?ooo`0:0?ooo`<0003o:P3oool00`3o0000oooo0?oo
o`050?ooo`030000o`3oool0oooo02h0oooo00<0o`000?ooo`3oool02@3oool30000obX0oooo00<0
o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0]0?ooo`030?l0003oool0oooo00T0oooo0`00
0?lB0?ooo`008@3oool00`000000003o0?ooo`270?ooo`030?l0003oool0oooo00T0oooo00<0003o
0?ooo`3oool0:`3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo02d0oooo00<0o`00
0?ooo`3oool0203oool00`000?l0oooo0?ooo`0[0?ooo`030?l0003oool0oooo00H0oooo00<0003o
0?ooo`3oool0;03oool00`3o0000oooo0?ooo`080?ooo`80003o503oool00240oooo00<000000000
o`3oool0Q`3oool00`3o0000oooo0?ooo`080?ooo`030000o`3oool0oooo02`0oooo00<0o`000?oo
o`3oool01P3oool00`000?l0oooo0?ooo`0]0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?oo
o`3oool0;03oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo02`0oooo00<0o`000?oo
o`3oool0203oool00`000?l0oooo0?ooo`0C0?ooo`008@3oool00`000000003o0?ooo`280?ooo`03
0?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool0;03oool00`3o0000oooo0?ooo`060?ooo`03
0000o`3oool0oooo02d0oooo00<0o`000?ooo`3oool01`3oool00`000?l0oooo0?ooo`0/0?ooo`03
0?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool0;@3oool00`3o0000oooo0?ooo`070?ooo`03
0000o`3oool0oooo01<0oooo000Q0?ooo`03000000000?l0oooo08P0oooo00<0o`000?ooo`3oool0
1`3oool00`000?l0oooo0?ooo`0/0?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool0
;@3oool00`3o0000oooo0?ooo`060?ooo`030000o`3oool0oooo02d0oooo00<0o`000?ooo`3oool0
1P3oool00`000?l0oooo0?ooo`0]0?ooo`030?l0003oool0oooo00L0oooo00<0003o0?ooo`3oool0
4`3oool001L0oooo2`0000000`000?l0000000000027000000030?l000000000000000H0000000<0
003o000000000000;00000000`3o0000000000000008000000030000o`000000000002`0000000<0
o`000000000000001P0000000`000?l000000000000]000000030?l000000000000000H0000000<0
003o000000000000;@0000000`3o0000000000000006000000030000o`0000000000014000000`3o
ool00240oooo00<000000000o`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`080?ooo`030?l0003oool0
000000H0oooo00<0003o0?ooo`3oool00`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?l000090?ooo`03000000000?l0
oooo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00L0oooo00@0o`000?ooo`3oool00000103oool00`000?l0oooo0?ooo`050?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0:0?ooo`030000003o
0000oooo00T0oooo00<0003o0000003oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0103oool00`3o0000oooo0?ooo`030?ooo`040000003o
ool0oooo0000o`P0oooo00<000000?ooo`3oool02`3oool00240oooo00<000000?ooo`000?l02000
0?le0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0203oool00`3o0000oooo0?oo
o`060?ooo`030000o`3oool0oooo02X0oooo00<000000?ooo`3o00002P3oool00`000?l0oooo0?oo
o`0]0?ooo`040?l0003oool0oooo000000@0oooo00<0003o0?ooo`3oool0;03oool00`3o0000oooo
0?ooo`080?ooo`030000o`000000oooo02`0oooo00<0o`000?ooo`3oool01P3oool00`000?l0oooo
0?ooo`060?ooo`030000003oool0oooo00/0oooo000Q0?ooo`030000003oool0oooo00P0oooo>@00
0?m70?ooo`030?l0003oool0oooo00H0oooo00<0003o0?ooo`3oool0;03oool00`3o0000oooo0?oo
o`090?ooo`030000o`3oool0oooo02`0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?oo
o`0/0?ooo`030?l0003oool0oooo00T0oooo00<0003o0?ooo`3oool0:`3oool00`3o0000oooo0?oo
o`060?ooo`030000o`3oool0oooo01@0oooo000Q0?ooo`030000003oool0oooo0440oooo5`000?la
0?ooo`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool0;03oool00`3o0000oooo0?ooo`09
0?ooo`<0003o;03oool00`3o0000oooo0?ooo`050?ooo`030000o`3oool0oooo02`0oooo00<0o`00
0?ooo`3oool02@3oool30000ob/0oooo00<0o`000?ooo`3oool01@3oool00`000?l0oooo0?ooo`0E
0?ooo`008@3oool00`000000oooo0?ooo`1H0?ooo``0003o9@3oool00`3o0000oooo0?ooo`040?oo
o`030000o`3oool0oooo02d0oooo00<0o`000?ooo`3oool02`3oool:0000ob<0oooo00<0o`000?oo
o`3oool0103oool00`000?l0oooo0?ooo`0]0?ooo`030?l0003oool0oooo00/0oooo2@000?lS0?oo
o`030?l0003oool0oooo00D0oooo00<0003o0?ooo`3oool05@3oool00240oooo00<000000?ooo`3o
ool0I03oool80000oad0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0/0?ooo`03
0?l0003oool0oooo01H0oooo1@000?lN0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3o
ool0;03oool00`3o0000oooo0?ooo`0E0?ooo`P0003o703oool00`3o0000oooo0?ooo`040?ooo`03
0000o`3oool0oooo01D0oooo000Q0?ooo`030000003oool0oooo06`0oooo1@000?lH0?ooo`030?l0
003oool0oooo00@0oooo00<0003o0?ooo`3oool0;03oool00`3o0000oooo0?ooo`0K0?ooo`L0003o
5`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0oooo02`0oooo00<0o`000?ooo`3oool0
7@3oool40000oaP0oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0E0?ooo`008@3o
ool00`000000oooo0?ooo`1a0?ooo`@0003o503oool00`3o0000oooo0?ooo`040?ooo`030000o`3o
ool0oooo02`0oooo00<0o`000?ooo`3oool08P3oool20000oaD0oooo00<0o`000?ooo`3oool0103o
ool00`000?l0oooo0?ooo`0/0?ooo`030?l0003oool0oooo0240oooo0`000?lE0?ooo`030?l0003o
ool0oooo00@0oooo00<0003o0?ooo`3oool05@3oool00240oooo00<000000?ooo`3oool0M@3oool2
0000oa80oooo00<0o`000?ooo`3oool0103oool00`000?l0oooo0?ooo`0/0?ooo`030?l0003oool0
oooo02@0oooo10000?lA0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool0;03oool0
0`3o0000oooo0?ooo`0T0?ooo`80003o4`3oool00`3o0000oooo0?ooo`040?ooo`030000o`3oool0
oooo01D0oooo000Q0?ooo`030000003oool0oooo07L0oooo0P000?l@0?ooo`030?l0003oool0oooo
00@0oooo00<0003o0?ooo`3oool0:`3oool00`3o0000oooo0?ooo`0Y0?ooo`030000o`3oool0oooo
00l0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0[0?ooo`030?l0003oool0oooo
02L0oooo0P000?lA0?ooo`030?l0003oool0oooo00@0oooo00<0003o0?ooo`3oool05@3oool00240
oooo0P00001j0?ooo`80003o3`3oool00`3o0000oooo0?ooo`030?ooo`030000o`3oool0oooo02/0
oooo00<0o`000?ooo`3oool0:P3oool20000o`l0oooo00<0o`000?ooo`3oool00`3oool00`000?l0
oooo0?ooo`0[0?ooo`030?l0003oool0oooo02T0oooo0P000?l?0?ooo`030?l0003oool0oooo00@0
oooo00<0003o0?ooo`3oool05@3oool00240oooo00<000000?ooo`3oool0N`3oool00`000?l0oooo
0?ooo`0<0?ooo`030?l0003oool0oooo00<0oooo00<0003o0?ooo`3oool0:`3oool00`3o0000oooo
0?ooo`0/0?ooo`030000o`3oool0oooo00`0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo
0?ooo`0[0?ooo`030?l0003oool0oooo02/0oooo0P000?l=0?ooo`030?l0003oool0oooo00@0oooo
00<0003o0?ooo`3oool05@3oool00240oooo00<000000?ooo`3oool0O03oool00`000?l0oooo0?oo
o`0;0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool0;03oool00`3o0000oooo0?oo
o`0]0?ooo`030000o`3oool0oooo00/0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?oo
o`0[0?ooo`030?l0003oool0oooo02d0oooo00<0003o0?ooo`3oool02`3oool00`3o0000oooo0?oo
o`030?ooo`030000o`3oool0oooo01D0oooo000Q0?ooo`030000003oool0oooo07d0oooo0P000?l;
0?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool0:`3oool00`3o0000oooo0?ooo`0_
0?ooo`030000o`3oool0oooo00X0oooo00<0o`000?ooo`3oool00`3oool00`000?l0oooo0?ooo`0Z
0?ooo`030?l0003oool0oooo02l0oooo00<0003o0?ooo`3oool02P3oool00`3o0000oooo0?ooo`03
0?ooo`030000o`3oool0oooo01D0oooo000Q0?ooo`030000003oool0oooo07h0oooo00<0003o0?oo
o`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo02/0oooo00<0o`000?oo
o`3oool0<03oool00`000?l0oooo0?ooo`090?ooo`030?l0003oool0oooo00<0oooo00<0003o0?oo
o`3oool0:P3oool00`3o0000oooo0?ooo`0`0?ooo`030000o`3oool0oooo00T0oooo00<0o`000?oo
o`3oool00`3oool00`000?l0oooo0?ooo`0E0?ooo`008@3oool00`000000oooo0?ooo`1o0?ooo`03
0000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0Z0?ooo`03
0?l0003oool0oooo0340oooo00<0003o0?ooo`3oool02@3oool00`3o0000oooo0?ooo`020?ooo`03
0000o`3oool0oooo02X0oooo00<0o`000?ooo`3oool0<@3oool00`000?l0oooo0?ooo`090?ooo`03
0?l0003oool0oooo0080oooo00<0003o0?ooo`3oool05P3oool00240oooo00<000000?ooo`3oool0
O`3oool00`000?l0oooo0?ooo`080?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool0
:P3oool00`3o0000oooo0?ooo`0b0?ooo`030000o`3oool0oooo00P0oooo00<0o`000?ooo`3oool0
0P3oool00`000?l0oooo0?ooo`0Z0?ooo`030?l0003oool0oooo0380oooo00<0003o0?ooo`3oool0
203oool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo01H0oooo000Q0?ooo`030000003o
ool0oooo0800oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0?ooo`020?ooo`030000o`3o
ool0oooo02T0oooo00<0o`000?ooo`3oool0<`3oool00`000?l0oooo0?ooo`080?ooo`030?l0003o
ool0oooo0080oooo00<0003o0?ooo`3oool0:@3oool00`3o0000oooo0?ooo`0c0?ooo`030000o`3o
ool0oooo00P0oooo00<0o`000?ooo`3oool00P3oool00`000?l0oooo0?ooo`0F0?ooo`008@3oool0
0`000000oooo0?ooo`210?ooo`030000o`3oool0oooo00L0oooo00D0o`000?ooo`3oool0oooo0000
o`0[0?ooo`030?l0003oool0oooo03@0oooo00<0003o0?ooo`3oool01`3oool00`3o0000oooo0?oo
o`020?ooo`030000o`3oool0oooo02T0oooo00<0o`000?ooo`3oool0=03oool00`000?l0oooo0?oo
o`070?ooo`030?l0003oool0oooo0080oooo00<0003o0?ooo`3oool05P3oool00240oooo00<00000
0?ooo`3oool0P@3oool00`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo0?ooo`000?l0:P3o
ool00`3o0000oooo0?ooo`0e0?ooo`030000o`3oool0oooo00L0oooo00<0o`000?ooo`3oool00P3o
ool00`000?l0oooo0?ooo`0X0?ooo`030?l0003oool0oooo03D0oooo00<0003o0?ooo`3oool01`3o
ool00`3o0000oooo0?ooo`020?ooo`030000o`3oool0oooo01H0oooo000Q0?ooo`030000003oool0
oooo0880oooo00<0003o0?ooo`3oool01P3oool01@3o0000oooo0?ooo`3oool0003o02X0oooo00<0
o`000?ooo`3oool0=P3oool00`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo0?ooo`000?l0
:P3oool00`3o0000oooo0?ooo`0f0?ooo`030000o`3oool0oooo00H0oooo00<0o`000?ooo`3oool0
0P3oool00`000?l0oooo0?ooo`0F0?ooo`008@3oool00`000000oooo0?ooo`220?ooo`030000o`3o
ool0oooo00H0oooo00D0o`000?ooo`3oool0oooo0000o`0Y0?ooo`030?l0003oool0oooo03L0oooo
00<0003o0?ooo`3oool01`3oool01@3o0000oooo0?ooo`3oool0003o02T0oooo00<0o`000?ooo`3o
ool0=`3oool00`000?l0oooo0?ooo`070?ooo`050?l0003oool0oooo0?ooo`000?l0603oool00240
oooo0P0000230?ooo`030000o`3oool0oooo00H0oooo00D0o`000?ooo`3oool0oooo0000o`0X0?oo
o`030?l0003oool0oooo03T0oooo00<0003o0?ooo`3oool01P3oool01@3o0000oooo0?ooo`3oool0
003o02P0oooo00<0o`000?ooo`3oool0>@3oool00`000?l0oooo0?ooo`060?ooo`050?l0003oool0
oooo0?ooo`000?l0603oool00240oooo00<000000?ooo`3oool0P`3oool00`000?l0oooo0?ooo`05
0?ooo`050?l0003oool0oooo0?ooo`000?l0:03oool00`3o0000oooo0?ooo`0i0?ooo`030000o`3o
ool0oooo00H0oooo00D0o`000?ooo`3oool0oooo0000o`0X0?ooo`030?l0003oool0oooo03T0oooo
00<0003o0?ooo`3oool01P3oool01@3o0000oooo0?ooo`3oool0003o01P0oooo000Q0?ooo`030000
003oool0oooo08<0oooo00<0003o0?ooo`3oool01@3oool01@3o0000oooo0?ooo`3oool0003o02L0
oooo00<0o`000?ooo`3oool0>P3oool00`000?l0oooo0?ooo`060?ooo`050?l0003oool0oooo0?oo
o`000?l09`3oool00`3o0000oooo0?ooo`0k0?ooo`030000o`3oool0oooo00D0oooo00D0o`000?oo
o`3oool0oooo0000o`0H0?ooo`008@3oool00`000000oooo0?ooo`230?ooo`030000o`3oool0oooo
00D0oooo00D0o`000?ooo`3oool0oooo0000o`0V0?ooo`030?l0003oool0oooo03/0oooo00<0003o
0?ooo`3oool01P3oool0103o0000oooo0?ooo`000?lW0?ooo`030?l0003oool0oooo03`0oooo00<0
003o0?ooo`3oool01@3oool01@3o0000oooo0?ooo`3oool0003o01P0oooo000Q0?ooo`030000003o
ool0oooo08<0oooo00<0003o0?ooo`3oool01@3oool01@3o0000oooo0?ooo`3oool0003o02D0oooo
00<0o`000?ooo`3oool0?@3oool00`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo0000obH0
oooo00<0o`000?ooo`3oool0?@3oool00`000?l0oooo0?ooo`050?ooo`050?l0003oool0oooo0?oo
o`000?l0603oool00240oooo00<000000?ooo`3oool0Q03oool00`000?l0oooo0?ooo`050?ooo`04
0?l0003oool0oooo0000ob@0oooo00<0o`000?ooo`3oool0?P3oool00`000?l0oooo0?ooo`050?oo
o`040?l0003oool0oooo0000obD0oooo00<0o`000?ooo`3oool0?`3oool00`000?l0oooo0?ooo`04
0?ooo`050?l0003oool0oooo0?ooo`000?l0603oool00240oooo00<000000?ooo`3oool0Q03oool0
0`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo0000ob<0oooo00<0o`000?ooo`3oool0?`3o
ool00`000?l0oooo0?ooo`060?ooo`030?l0003oool0003o02@0oooo00<0o`000?ooo`3oool0@03o
ool00`000?l0oooo0?ooo`040?ooo`050?l0003oool0oooo0?ooo`000?l0603oool00240oooo00<0
00000?ooo`3oool0Q03oool00`000?l0oooo0?ooo`050?ooo`040?l0003oool0oooo0000ob80oooo
00<0o`000?ooo`3oool0@03oool00`000?l0oooo0?ooo`060?ooo`030?l0003oool0003o02<0oooo
00<0o`000?ooo`3oool0@@3oool00`000?l0oooo0?ooo`040?ooo`050?l0003oool0oooo0?ooo`00
0?l0603oool00240oooo00<000000?ooo`3oool0Q03oool00`000?l0oooo0?ooo`050?ooo`030?l0
003oool0003o0280oooo00<0o`000?ooo`3oool0@@3oool00`000?l0oooo0?ooo`060?ooo`030?l0
003oool0003o0280oooo00<0o`000?ooo`3oool0@P3oool00`000?l0oooo0?ooo`050?ooo`040?l0
003oool0oooo0000oaP0oooo000Q0?ooo`030000003oool0oooo08@0oooo00<0003o0?ooo`3oool0
1@3oool00`3o0000oooo0000o`0Q0?ooo`030?l0003oool0oooo04<0oooo00<0003o0?ooo`3oool0
1@3oool00`3o0000oooo0000o`0Q0?ooo`030?l0003oool0oooo04<0oooo00<0003o0?ooo`3oool0
1@3oool0103o0000oooo0?ooo`000?lH0?ooo`008@3oool00`000000oooo0?ooo`250?ooo`030000
o`3oool0oooo00@0oooo00<0o`000?ooo`000?l0803oool00`3o0000oooo0?ooo`140?ooo`030000
o`3oool0oooo00D0oooo00<0o`000?ooo`000?l07`3oool20?l004L0oooo00<0003o0?ooo`3oool0
103oool0103o0000oooo0?ooo`000?lH0?ooo`008@3oool00`000000oooo0?ooo`250?ooo`030000
o`3oool0oooo00@0oooo00<0o`000?ooo`000?l07`3oool00`3o0000oooo0?ooo`150?ooo`030000
o`3oool0oooo00D0oooo00<0o`000?ooo`000?l07P3oool00`3o0000oooo0?ooo`170?ooo`030000
o`3oool0oooo00@0oooo00@0o`000?ooo`3oool0003o603oool00240oooo0P0000260?ooo`030000
o`3oool0oooo00D0oooo00<0o`000000o`3oool0703oool20?l004P0oooo00<0003o0?ooo`3oool0
1@3oool00`3o0000oooo0000o`0L0?ooo`80o`00BP3oool00`000?l0oooo0?ooo`040?ooo`040?l0
003oool0oooo0000oaP0oooo000Q0?ooo`030000003oool0oooo08D0oooo00<0003o0?ooo`3oool0
1@3oool00`3o0000003o0?ooo`0J0?ooo`80o`00BP3oool00`000?l0oooo0?ooo`050?ooo`030?l0
003oool0003o01/0oooo00<0o`000?ooo`3oool0BP3oool00`000?l0oooo0?ooo`050?ooo`030?l0
003oool0003o01P0oooo000Q0?ooo`030000003oool0oooo08D0oooo00<0003o0?ooo`3oool01@3o
ool00`3o0000003o0?ooo`0H0?ooo`80o`00C03oool00`000?l0oooo0?ooo`060?ooo`030?l00000
0?l0oooo01P0oooo0P3o001=0?ooo`030000o`3oool0oooo00D0oooo00<0o`000?ooo`000?l0603o
ool00240oooo00<000000?ooo`3oool0Q@3oool00`000?l0oooo0?ooo`050?ooo`030?l000000?l0
oooo01H0oooo0P3o001?0?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool05P3oool2
0?l004l0oooo00<0003o0?ooo`3oool01@3oool00`3o0000oooo0000o`0H0?ooo`008@3oool00`00
0000oooo0?ooo`250?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool0503oool20?l0
0540oooo00<0003o0?ooo`3oool01@3oool00`3o0000003o0?ooo`0E0?ooo`030?l0003oool0oooo
04l0oooo00<0003o0?ooo`3oool01@3oool00`3o0000oooo0000o`0H0?ooo`008@3oool00`000000
oooo0?ooo`250?ooo`030000o`3oool0oooo00D0oooo00<0o`000000o`3oool04P3oool20?l005<0
oooo00<0003o0?ooo`3oool01@3oool00`3o0000003o0?ooo`0C0?ooo`80o`00DP3oool00`000?l0
oooo0?ooo`050?ooo`030?l000000?l0oooo01P0oooo000Q0?ooo`030000003oool0oooo08H0oooo
00<0003o0?ooo`3oool0103oool00`3o0000003o0?ooo`0@0?ooo`80o`00E@3oool00`000?l0oooo
0?ooo`050?ooo`030?l000000?l0oooo0140oooo0P3o001D0?ooo`030000o`3oool0oooo00D0oooo
00<0o`000000o`3oool0603oool00100oooo00<0gia80000000000000`0000000`18W=l0oooo0?oo
o`080?ooo`030000003oool0oooo08H0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?oo
o`0=0?ooo`80o`00E`3oool00`000?l0oooo0?ooo`060?ooo`030000o`3oool0oooo00h0oooo0P3o
001F0?ooo`030000o`3oool0oooo00H0oooo00<0003o0?ooo`3oool05`3oool00140oooo00<0okmd
001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`080?ooo`030000003oool0oooo08H0oooo00<0003o
0?ooo`3oool01@3oool00`000?l0oooo0?ooo`0;0?ooo`80o`00F@3oool00`000?l0oooo0?ooo`06
0?ooo`030000o`3oool0oooo00/0oooo0`3o001H0?ooo`030000o`3oool0oooo00H0oooo00<0003o
0?ooo`3oool05`3oool00180oooo00<0okmd001d_`3oool0303oool00`000000oooo0?ooo`260?oo
o`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool02@3oool20?l005/0oooo00<0003o0?oo
o`3oool01P3oool00`000?l0oooo0?ooo`080?ooo`<0o`00F`3oool00`000?l0oooo0?ooo`060?oo
o`030000o`3oool0oooo00P0oooo0P3o000=0?ooo`004`3oool00`3o_g@007Bo0?ooo`0;0?ooo`03
0000003oool0oooo08H0oooo00<0003o0?ooo`3oool01@3oool00`000?l0o`000?ooo`070?ooo`80
o`00GP3oool00`000?l0oooo0?ooo`050?ooo`030000o`3o0000oooo00D0oooo0`3o001N0?ooo`03
0000o`3oool0oooo00H0oooo00<0003o0?l0003oool01P3oool20?l000l0oooo000D0?ooo`030?no
M018W=l0oooo00X0oooo00<000000?ooo`3oool0QP3oool00`000?l0oooo0?ooo`050?ooo`030000
o`3o0000oooo00D0oooo0P3o001P0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?l0003oool0
0`3oool20?l00640oooo00<0003o0?ooo`3oool01P3oool00`000?l0o`000?ooo`040?ooo`80o`00
4@3oool001D0oooo00<0gia804RLg`3oool02@3oool3000008H0oooo00<0003o0?ooo`3oool01@3o
ool00`000?l0oooo0?l000050?l00680oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?l0
00030?l006@0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?l000040?l001<0oooo000F
0?ooo`030;md001d_ol0oooo00P0oooo00<000000?ooo`3oool0QP3oool00`000?l0oooo0?ooo`05
0?ooo`030000o`3oool0oooo06L0oooo00<0003o0?ooo`3oool01@3oool00`000?l0oooo0?ooo`1W
0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool05`3oool001H0oooo00<0omnL04P0
M02oool0203oool00`000000oooo0?ooo`260?ooo`030000o`3oool0oooo00D0oooo00<0003o0?oo
o`3oool0I`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0oooo06L0oooo00<0003o0?oo
o`3oool01@3oool00`000?l0oooo0?ooo`0G0?ooo`005P3oool00`3ooml0W4P007Boo`080?ooo`03
0000003oool0oooo08L0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`1W0?ooo`03
0000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0I`3oool00`000?l0oooo0?ooo`050?ooo`03
0000o`3oool0oooo01L0oooo000F0?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0
Q`3oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo06L0oooo00<0003o0?ooo`3oool0
1@3oool00`000?l0oooo0?ooo`1W0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0
5`3oool00180oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00P0oooo00<000000?ooo`3o
ool0Q`3oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo06P0oooo00<0003o0?ooo`3o
ool01@3oool00`000?l0oooo0?ooo`1W0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3o
ool05`3oool001<0oooo00D0omnL04P000000000000007Boo`090?ooo`030000003oool0oooo08L0
oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`1X0?ooo`030000o`3oool0oooo00D0
oooo00<0003o0?ooo`3oool0I`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0oooo01L0
oooo000Q0?ooo`030000003oool0oooo08L0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo
0?ooo`1X0?ooo`030000o`3oool0oooo00D0oooo00<0003o0?ooo`3oool0I`3oool00`000?l0oooo
0?ooo`050?ooo`030000o`3oool0oooo01L0oooo000Q0?ooo`030000003oool0oooo08L0oooo00<0
003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00@0oooo00<0
003o0?ooo`3oool0I`3oool00`000?l0oooo0?ooo`050?ooo`030000o`3oool0oooo01L0oooo000Q
0?ooo`030000003oool0oooo08L0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`1Y
0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool0J03oool00`000?l0oooo0?ooo`04
0?ooo`030000o`3oool0oooo01L0oooo000Q0?ooo`030000003oool0oooo08L0oooo00<0003o0?oo
o`3oool00`3oool00`000?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?oo
o`3oool0J03oool00`000?l0oooo0?ooo`040?ooo`030000o`3oool0oooo01L0oooo000Q0?ooo`80
0000R03oool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo06T0oooo00<0003o0?ooo`3o
ool0103oool00`000?l0oooo0?ooo`1X0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3o
ool05`3oool00240oooo00<000000?ooo`3oool0Q`3oool00`000?l0oooo0?ooo`030?ooo`030000
o`3oool0oooo06T0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`1X0?ooo`030000
o`3oool0oooo00@0oooo00<0003o0?ooo`3oool05`3oool00240oooo00<000000?ooo`3oool0Q`3o
ool00`000?l0oooo0?ooo`030?ooo`030000o`3oool0oooo06T0oooo00<0003o0?ooo`3oool0103o
ool00`000?l0oooo0?ooo`1X0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool05`3o
ool00240oooo00<000000?ooo`3oool0R03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0
oooo06T0oooo00<0003o0?ooo`3oool0103oool00`000?l0oooo0?ooo`1X0?ooo`030000o`3oool0
oooo00@0oooo00<0003o0?ooo`3oool05`3oool00240oooo00<000000?ooo`3oool0R03oool00`00
0?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06T0oooo00<0003o0?ooo`3oool00`3oool00`00
0?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool05`3oool00240
oooo00<000000?ooo`3oool0R03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06T0
oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00@0
oooo00<0003o0?ooo`3oool05`3oool00240oooo00<000000?ooo`3oool0R03oool00`000?l0oooo
0?ooo`020?ooo`030000o`3oool0oooo06T0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo
0?ooo`1Y0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool05`3oool00240oooo00<0
00000?ooo`3oool0R03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06T0oooo00<0
003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00@0oooo00<0
003o0?ooo`3oool05`3oool00240oooo00<000000?ooo`3oool0R03oool00`000?l0oooo0?ooo`02
0?ooo`030000o`3oool0oooo06T0oooo00<0003o0?ooo`3oool00`3oool00`000?l0oooo0?ooo`1Y
0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?ooo`3oool05`3oool00240oooo00<000000?oo
o`3oool0R03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?oo
o`3oool00P3oool00`000?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00@0oooo00<0003o0?oo
o`3oool05`3oool00240oooo00<000000?ooo`3oool0R03oool00`000?l0oooo0?ooo`020?ooo`03
0000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`1Y0?ooo`03
0000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0603oool00240oooo00<000000?ooo`3oool0
R03oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool0
0P3oool00`000?l0oooo0?ooo`1Y0?ooo`030000o`3oool0oooo00<0oooo00<0003o0?ooo`3oool0
603oool00240oooo0P0000290?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0JP3o
ool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3o
ool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`280?ooo`030000o`3oool0
oooo0080oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0
oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`00
0000oooo0?ooo`280?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0JP3oool00`00
0?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`280?ooo`030000o`3oool0oooo0080
oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0
oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo
0?ooo`280?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo
0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo
0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`280?ooo`030000o`3oool0oooo0080oooo00<0
003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0
003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`28
0?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`02
0?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H
0?ooo`008@3oool00`000000oooo0?ooo`280?ooo`030000o`3oool0oooo0080oooo00<0003o0?oo
o`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?oo
o`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`280?ooo`03
0000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`03
0000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`00
8@3oool00`000000oooo0?ooo`280?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0
JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool0
0P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`280?ooo`030000o`3o
ool0oooo0080oooo00<0003o0?ooo`3oool0JP3oool00`000?l0oooo0?ooo`020?ooo`030000o`3o
ool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool0
0`000000oooo0?ooo`280?ooo`030000o`3oool0oooo0080oooo00<0003o0?ooo`3oool0JP3oool0
0`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo06X0oooo00<0003o0?ooo`3oool00P3oool0
0`000?l0oooo0?ooo`0H0?ooo`008@3oool2000008T0oooo00<0003o0?ooo`3oool00P3oool00`00
0?l0oooo0?ooo`1[0?ooo`050000o`3oool0oooo0?ooo`000?l0K03oool00`000?l0oooo0?ooo`02
0?ooo`030000o`3oool0oooo01P0oooo000Q0?ooo`030000003oool0oooo08P0oooo00<0003o0?oo
o`3oool00P3oool00`000?l0oooo0?ooo`1[0?ooo`050000o`3oool0oooo0?ooo`000?l0K03oool0
0`000?l0oooo0?ooo`020?ooo`030000o`3oool0oooo01P0oooo000Q0?ooo`030000003oool0oooo
08T0oooo00@0003o0?ooo`3oool0003oKP3oool01@000?l0oooo0?ooo`3oool0003o06`0oooo00<0
003o0?ooo`3oool00P3oool00`000?l0oooo0?ooo`0H0?ooo`008@3oool00`000000oooo0?ooo`29
0?ooo`040000o`3oool0oooo0000ofh0oooo00D0003o0?ooo`3oool0oooo0000o`1/0?ooo`030000
o`3oool0oooo0080oooo00<0003o0?ooo`3oool0603oool00240oooo00<000000?ooo`3oool0R@3o
ool010000?l0oooo0?ooo`000?m^0?ooo`050000o`3oool0oooo0?ooo`000?l0K03oool00`000?l0
oooo0?ooo`020?ooo`030000o`3oool0oooo01P0oooo000Q0?ooo`030000003oool0oooo08T0oooo
00@0003o0?ooo`3oool0003oKP3oool01@000?l0oooo0?ooo`3oool0003o06d0oooo00D0003o0?oo
o`3oool0oooo0000o`0J0?ooo`008@3oool00`000000oooo0?ooo`290?ooo`040000o`3oool0oooo
0000ofh0oooo00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`050000o`3oool0oooo0?ooo`000?l0
6P3oool001<0oooo00<0oooO09a80018W=l02`3oool00`000000oooo0?ooo`290?ooo`040000o`3o
ool0oooo0000ofh0oooo00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`050000o`3oool0oooo0?oo
o`000?l06P3oool001@0oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`290?ooo`04
0000o`3oool0oooo0000ofh0oooo00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`050000o`3oool0
oooo0?ooo`000?l06P3oool001@0oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`29
0?ooo`040000o`3oool0oooo0000ofh0oooo00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`040000
o`3oool0oooo0000oa/0oooo000@0?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0
1`3oool00`000000oooo0?ooo`290?ooo`040000o`3oool0oooo0000ofh0oooo00D0003o0?ooo`3o
ool0oooo0000o`1]0?ooo`040000o`3oool0oooo0000oa/0oooo000@0?ooo`030?oog`2LB4P0W=oo
0080oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`290?ooo`040000o`3oool0oooo
0000ofh0oooo00D0003o0?ooo`3oool0oooo0000o`1]0?ooo`040000o`3oool0oooo0000oa/0oooo
000A0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`090?ooo`<00000R@3oool010000?l0
oooo0?ooo`000?m^0?ooo`040000o`3oool0oooo0000ofh0oooo00@0003o0?ooo`3oool0003o6`3o
ool00180oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol90?ooo`030000003oool0oooo08T0oooo
00@0003o0?ooo`3oool0003oKP3oool010000?l0oooo0?ooo`000?m^0?ooo`040000o`3oool0oooo
0000oa/0oooo000C0?ooo`050?noM000M;l0oooO09a80018W=l02@3oool00`000000oooo0?ooo`29
0?ooo`040000o`3oool0oooo0000ofh0oooo00@0003o0?ooo`3oool0003oKP3oool010000?l0oooo
0?ooo`000?lK0?ooo`00503oool0103OW4P0B9cO0;md0000M;l90?ooo`030000003oool0oooo08T0
oooo00@0003o0?ooo`3oool0003oKP3oool010000?l0oooo0?ooo`000?m^0?ooo`040000o`3oool0
oooo0000oa/0oooo000E0?ooo`040=nLB01dM7@0001d0;ooo`P0oooo00<000000?ooo`3oool0R@3o
ool010000?l0oooo0?ooo`000?m^0?ooo`040000o`3oool0oooo0000ofh0oooo00@0003o0?ooo`3o
ool0003o6`3oool001H0oooo00<0_g@00000B02Lgol0203oool00`000000oooo0?ooo`290?ooo`04
0000o`3oool0oooo0000ofh0oooo00@0003o0?ooo`3oool0003oKP3oool010000?l0oooo0?ooo`00
0?lK0?ooo`005P3oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo08T0oooo00@0003o
0?ooo`3oool0003oKP3oool010000?l0oooo0?ooo`000?m^0?ooo`040000o`3oool0oooo0000oa/0
oooo000Q0?ooo`030000003oool0oooo08T0oooo00@0003o0?ooo`3oool0003oKP3oool010000?l0
oooo0?ooo`000?m_0?ooo`030000o`3oool0003o01/0oooo000Q0?ooo`030000003oool0oooo08T0
oooo00@0003o0?ooo`3oool0003oKP3oool010000?l0oooo0?ooo`000?m_0?ooo`030000o`3oool0
003o01/0oooo000Q0?ooo`030000003oool0oooo08T0oooo00@0003o0?ooo`3oool0003oKP3oool0
10000?l0oooo0?ooo`000?m_0?ooo`030000o`3oool0003o01/0oooo000Q0?ooo`030000003oool0
oooo08X0oooo00<0003o0?ooo`000?l0KP3oool010000?l0oooo0?ooo`000?m_0?ooo`030000o`3o
ool0003o01/0oooo000Q0?ooo`030000003oool0oooo08X0oooo00<0003o0?ooo`000?l0KP3oool0
10000?l0oooo0?ooo`000?m_0?ooo`030000o`3oool0003o01/0oooo000Q0?ooo`800000R`3oool0
0`000?l0oooo0000o`1_0?ooo`030000o`3oool0003o06l0oooo00<0003o0?ooo`000?l06`3oool0
0240oooo00<000000?ooo`3oool0RP3oool00`000?l0oooo0000o`1_0?ooo`030000o`3oool0003o
06l0oooo00<0003o0?ooo`000?l06`3oool00240oooo00<000000?ooo`3oool0RP3oool00`000?l0
oooo0000o`1_0?ooo`030000o`3oool0003o06l0oooo00<0003o0?ooo`000?l06`3oool00240oooo
00<000000?ooo`3oool0RP3oool00`000?l0oooo0000o`1_0?ooo`030000o`3oool0003o06l0oooo
00<0003o0?ooo`000?l06`3oool00240oooo00<000000?ooo`3oool0RP3oool00`000?l0oooo0000
o`1_0?ooo`030000o`3oool0003o06l0oooo00<0003o0?ooo`000?l06`3oool00240oooo00<00000
0?ooo`3oool0RP3oool00`000?l0oooo0000o`1_0?ooo`030000o`3oool0003o06l0oooo00<0003o
0?ooo`000?l06`3oool00240oooo00<000000?ooo`3oool0RP3oool00`000?l0oooo0000o`1_0?oo
o`030000o`3oool0003o06l0oooo0P000?lL0?ooo`008@3oool00`000000oooo0?ooo`2:0?ooo`03
0000o`3oool0003o06l0oooo00<0003o0?ooo`000?l0K`3oool20000oa`0oooo000Q0?ooo`030000
003oool0oooo08X0oooo00<0003o0?ooo`000?l0K`3oool20000og00oooo0P000?lL0?ooo`008@3o
ool00`000000oooo0?ooo`2:0?ooo`80003oL03oool20000og00oooo0P000?lL0?ooo`008@3oool0
0`000000oooo0?ooo`2:0?ooo`80003oL03oool20000og00oooo0P000?lL0?ooo`008@3oool20000
08`0oooo00<0003o0?ooo`3oool0KP3oool20000og00oooo0P000?lL0?ooo`008@3oool00`000000
oooo0?ooo`2;0?ooo`030000o`3oool0oooo06l0oooo00<0003o0?ooo`3oool0KP3oool20000oa`0
oooo000Q0?ooo`030000003oool0oooo0?d0oooo00<0003o0?ooo`3oool0K`3oool00`000?l0oooo
0?ooo`0J0?ooo`008@3oool00`000000oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000003oool0
oooo0?l0ooooS@3oool00240oooo00<000000?ooo`3oool0o`3ooon=0?ooo`008@3oool00`000000
oooo0?ooo`3o0?ooohd0oooo000Q0?ooo`030000003oool0oooo0?l0ooooS@3oool00240oooo00<0
00000?ooo`3oool0o`3ooon=0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00
o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.57305, -5.64616, \
0.116611, 0.0637072}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(StrobePlot[\(-3\), \(-3\), 3, 15, 24, 0.1, 
        PlotRange \[Rule] All];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.500034 0.235382 0.310706 0.0584415 [
[.02927 .29821 -9.34375 -15.6875 ]
[.02927 .29821 9.34375 0 ]
[.26465 .29821 -9.34375 -15.6875 ]
[.26465 .29821 9.34375 0 ]
[.73542 .29821 -5 -15.6875 ]
[.73542 .29821 5 0 ]
[.9708 .29821 -5 -15.6875 ]
[.9708 .29821 5 0 ]
[.48753 .07694 -18.6875 -7.84375 ]
[.48753 .07694 0 7.84375 ]
[.48753 .19382 -18.6875 -7.84375 ]
[.48753 .19382 0 7.84375 ]
[.48753 .42759 -10 -7.84375 ]
[.48753 .42759 0 7.84375 ]
[.48753 .54447 -10 -7.84375 ]
[.48753 .54447 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02927 .31071 m
.02927 .31696 L
s
gsave
.02927 .29821 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.26465 .31071 m
.26465 .31696 L
s
gsave
.26465 .29821 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.73542 .31071 m
.73542 .31696 L
s
gsave
.73542 .29821 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.9708 .31071 m
.9708 .31696 L
s
gsave
.9708 .29821 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.07635 .31071 m
.07635 .31446 L
s
.12342 .31071 m
.12342 .31446 L
s
.1705 .31071 m
.1705 .31446 L
s
.21758 .31071 m
.21758 .31446 L
s
.31173 .31071 m
.31173 .31446 L
s
.3588 .31071 m
.3588 .31446 L
s
.40588 .31071 m
.40588 .31446 L
s
.45296 .31071 m
.45296 .31446 L
s
.54711 .31071 m
.54711 .31446 L
s
.59419 .31071 m
.59419 .31446 L
s
.64126 .31071 m
.64126 .31446 L
s
.68834 .31071 m
.68834 .31446 L
s
.78249 .31071 m
.78249 .31446 L
s
.82957 .31071 m
.82957 .31446 L
s
.87665 .31071 m
.87665 .31446 L
s
.92372 .31071 m
.92372 .31446 L
s
.25 Mabswid
0 .31071 m
1 .31071 L
s
.50003 .07694 m
.50628 .07694 L
s
gsave
.48753 .07694 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.50003 .19382 m
.50628 .19382 L
s
gsave
.48753 .19382 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.50003 .42759 m
.50628 .42759 L
s
gsave
.48753 .42759 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.50003 .54447 m
.50628 .54447 L
s
gsave
.48753 .54447 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.125 Mabswid
.50003 .10616 m
.50378 .10616 L
s
.50003 .13538 m
.50378 .13538 L
s
.50003 .1646 m
.50378 .1646 L
s
.50003 .22304 m
.50378 .22304 L
s
.50003 .25226 m
.50378 .25226 L
s
.50003 .28149 m
.50378 .28149 L
s
.50003 .33993 m
.50378 .33993 L
s
.50003 .36915 m
.50378 .36915 L
s
.50003 .39837 m
.50378 .39837 L
s
.50003 .45681 m
.50378 .45681 L
s
.50003 .48603 m
.50378 .48603 L
s
.50003 .51525 m
.50378 .51525 L
s
.50003 .04772 m
.50378 .04772 L
s
.50003 .0185 m
.50378 .0185 L
s
.50003 .57369 m
.50378 .57369 L
s
.50003 .60291 m
.50378 .60291 L
s
.25 Mabswid
.50003 0 m
.50003 .61803 L
s
.008 w
.64065 .22315 Mdot
.6007 .19826 Mdot
.54849 .16155 Mdot
.47847 .10958 Mdot
.3851 .04843 Mdot
.27071 .01472 Mdot
.15813 .06305 Mdot
.07924 .16864 Mdot
.04032 .25213 Mdot
.02635 .29425 Mdot
.02381 .31188 Mdot
.0259 .31888 Mdot
.02985 .32172 Mdot
.03457 .32297 Mdot
.03965 .32361 Mdot
.04493 .32402 Mdot
.05037 .32436 Mdot
.05593 .32467 Mdot
.06161 .32498 Mdot
.06742 .3253 Mdot
.07337 .32564 Mdot
.07945 .32601 Mdot
.08569 .3264 Mdot
.0921 .32682 Mdot
.09868 .32728 Mdot
.10546 .32778 Mdot
.11244 .32833 Mdot
.11966 .32893 Mdot
.12714 .3296 Mdot
.13489 .33034 Mdot
.14296 .33117 Mdot
.15139 .3321 Mdot
.16022 .33316 Mdot
.1695 .33438 Mdot
.17931 .33579 Mdot
.18973 .33743 Mdot
.20088 .33939 Mdot
.21289 .34174 Mdot
.22594 .34461 Mdot
.24029 .34818 Mdot
.25626 .35272 Mdot
.27431 .35861 Mdot
.29511 .36648 Mdot
.31963 .37729 Mdot
.34934 .39257 Mdot
.3865 .41476 Mdot
.43458 .44752 Mdot
.49869 .49489 Mdot
.5848 .55531 Mdot
.69426 .60332 Mdot
.81099 .58207 Mdot
.90205 .48454 Mdot
.95163 .38924 Mdot
.97124 .33621 Mdot
.97619 .31319 Mdot
.97505 .30398 Mdot
.97149 .3003 Mdot
.96694 .29872 Mdot
.96195 .29796 Mdot
.95671 .2975 Mdot
.95132 .29715 Mdot
.9458 .29684 Mdot
.94015 .29653 Mdot
.93438 .29621 Mdot
.92847 .29588 Mdot
.92243 .29552 Mdot
.91623 .29513 Mdot
.90988 .29472 Mdot
.90335 .29427 Mdot
.89663 .29378 Mdot
.88971 .29325 Mdot
.88256 .29266 Mdot
.87517 .29202 Mdot
.8675 .2913 Mdot
.85952 .2905 Mdot
.85121 .2896 Mdot
.84251 .28858 Mdot
.83336 .28741 Mdot
.82372 .28607 Mdot
.81348 .28449 Mdot
.80256 .28264 Mdot
.79083 .28042 Mdot
.7781 .27772 Mdot
.76417 .27438 Mdot
.74872 .27016 Mdot
.73134 .26471 Mdot
.71144 .25751 Mdot
.68814 .2477 Mdot
.66016 .23395 Mdot
.6255 .21412 Mdot
.5811 .18493 Mdot
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003of03oool0
0`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000
oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?oo
o`3E0?ooo`00M@3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0=D0oooo001d0?oo
o`<00000H@3oool00`000000oooo0?ooo`3E0?ooo`00M@3oool00`000000oooo0?ooo`1P0?ooo`03
0000003oool0oooo0=D0oooo003H0?ooo`800000eP3oool00=P0oooo00<000000?ooo`3oool0e@3o
ool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0
oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<0
00000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?oo
o`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0
e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00:H0oooo00<000000?ooo`3oool0;`3oool2
00000=H0oooo002U0?ooo`<00000<03oool00`000000oooo0?ooo`3E0?ooo`00YP3oool00`000000
oooo0?ooo`0_0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo
003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo00150?oo
o`030000003oool0oooo0900oooo00<000000?ooo`3oool0e@3oool004@0oooo0`00002A0?ooo`03
0000003oool0oooo0=D0oooo00150?ooo`030000003oool0oooo0840oooo00<0oooO09a80018W=l0
303oool00`000000oooo0?ooo`3E0?ooo`00bP3oool00`3OW4P004RL0=ooo`0;0?ooo`030000003o
ool0oooo0=D0oooo003:0?ooo`030?oOW01804P0W=oo00/0oooo00<000000?ooo`3oool0e@3oool0
0<H0oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`080?ooo`030000003oool0oooo
0=D0oooo002i0?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0oooO09a8B02Lgol2
0?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3oool0e@3oool00<L0oooo00L0oono07@0
M02oool0oooo0?noM000B9`0gooo00X0oooo0`00003E0?ooo`00b03oool01P3ogi`0B4RL0=ooo`3o
gi`0B01809cOo`X0oooo00<000000?ooo`3oool0e@3oool00<T0oooo00D0okmd001d_`3ooml0W4P0
04RLg`0:0?ooo`030000003oool0oooo0=D0oooo003:0?ooo`040=nLB018W=l0_g@0001d_`X0oooo
00<000000?ooo`3oool0e@3oool00</0oooo00@0gia807AdM00007@0_ooo2@3oool00`000000oooo
0?ooo`3E0?ooo`00c03oool00`2oM000001809cOo`090?ooo`030000003oool0oooo0=D0oooo003<
0?ooo`030?oog`2LB000B9cO00T0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?oo
o`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0
e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool0
0=P0oooo0P00003F0?ooo`00cP3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0=D0
oooo003=0?ooo`<00000203oool00`000000oooo0?ooo`3E0?ooo`00cP3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?oo
o`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000
003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0
oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0
oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`800000eP3oool00=P0oooo00<00000
0?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3o
ool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3o
ool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0
oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`310?ooo`00f03oool00`000000oooo0?ooo`0@
0?ooo`<00000`P3oool00=P0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`310?oo
o`008`3oool00`000000oooo0?ooo`2b0?ooo`800000eP3oool00280oooo0`00002c0?ooo`030000
003oool0oooo0=D0oooo000S0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0e@3o
ool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0
oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00<H0oooo00<0
gia80000000000000`0000000`18W=l0oooo0?ooo`090?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool0/`3oool00<L0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`09
0?ooo`030000003oool0oooo01h0oooo0`00002d0?ooo`00b03oool00`3o_g@007Bo0?ooo`0=0?oo
o`030000003oool0oooo01l0oooo00<000000?ooo`3oool0/`3oool00<T0oooo00<0okmd001d_`3o
ool0303oool00`000000oooo0?ooo`3E0?ooo`00^@3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00@0oooo00<0okmd04RLg`3oool02`3oool00`000000oooo0?ooo`3E0?ooo`00b`3o
ool00`3OW4P0B9cO0?ooo`0:0?ooo`<00000e@3oool00<`0oooo00<0_g@007Boo`3oool02@3oool0
0`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0:X0oooo003<0?ooo`030?oOW01807@0_ooo
00T0oooo00<000000?ooo`3oool09`3oool300000:/0oooo003<0?ooo`030?oog`2LB000M;oo00T0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`2Z0?ooo`00c03oool00`3ookl0M000
04RLg`090?ooo`030000003oool0oooo0=D0oooo00380?ooo`070?noM000B9`0gooo0?ooo`3OW4P0
001809cOo`090?ooo`030000003oool0oooo0=D0oooo00390?ooo`050?oOW018000000000000001d
_ol02P3oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03o
ool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0:00oooo003H0?ooo`030000003oool0
oooo0340oooo0`00002Q0?ooo`00f03oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo
0:00oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool0V@3oool00=P0oooo0P00000i0?ooo`<00000VP3oool00=P0oooo00<00000
0?ooo`3oool0>@3oool00`000000oooo0?ooo`2I0?ooo`00f03oool00`000000oooo0?ooo`3E0?oo
o`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`110?ooo`030000
003oool0oooo0940oooo003H0?ooo`030000003oool0oooo0400oooo0`00002B0?ooo`00f03oool0
0`000000oooo0?ooo`110?ooo`030000003oool0oooo0940oooo003H0?ooo`030000003oool0oooo
0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0Q@3oool00=P0oooo00<000000?ooo`3oool0C03oool3000008H0oooo000B
0?ooo`030000003oool0oooo0<<0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`25
0?ooo`004@3oool300000<@0oooo0P00003F0?ooo`004P3oool00`000000oooo0?ooo`330?ooo`03
0000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3oool0
N`3oool00100oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`1L0?ooo`030?oo_`1d
000000000080000000<0B9cO0?ooo`3oool0F`3oool00`000000oooo0?ooo`1F0?ooo`<00000103o
ool00`3ookl0M000000000020000000304RLg`3oool0oooo05d0oooo00<0gia80000000000000`00
00000`18W=l0oooo0?ooo`0:0?ooo`004@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0
oooo05h0oooo00<0okmd0000001d_ol0GP3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0
oooo00D0oooo00<0okmd0000001d_ol0H@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0
oooo00X0oooo000B0?ooo`030?noM000M;l0oooo0680oooo00<0oooO09a80000M;l0GP3oool00`00
0000oooo0?ooo`1O0?ooo`030?oog`00000007Bo0680oooo00<0okmd001d_`3oool03P3oool001<0
oooo00<0okmd001d_`3oool0HP3oool00`3OW4P004RL0=ooo`1M0?ooo`030000003oool0oooo05l0
oooo0`0000000`3Oool0oooo0?ooo`1P0?ooo`030?noM000M;l0oooo00d0oooo00030?ooo`030?oo
g`2LB000000000L0000000<0B9cO0?ooo`3oool0103oool00`3o_g@0B9cO0?ooo`1A0?ooo`030?oo
g`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`1M0?ooo`030000
003oool0oooo0600oooo00<0000004P0B02Lgol0103oool00`000000oooo0?ooo`1L0?ooo`030?no
M018W=l0oooo00`0oooo000E0?ooo`030=nLB018W=l0oooo0600oooo00<0oono07@0001d_ol0G@3o
ool00`000000oooo0?ooo`1P0?ooo`030?oo_`1d0000M;oo00<0oooo0`00001N0?ooo`030=nLB018
W=l0oooo00/0oooo000F0?ooo`030;md001d_ol0oooo0600oooo00<0_g@0001d_`3oool0G03oool0
0`000000oooo0?ooo`1Q0?ooo`030;md0000M;l0oooo00<0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`1G0?ooo`030;md001d_ol0oooo00X0oooo000F0?ooo`030?oOW01807@0_ooo
0600oooo00<0gia80018W03Oool0G03oool00`000000oooo0?ooo`1Q0?ooo`030=nLB000B9`0gooo
00T0oooo0`0000040?ooo`030000003oool0oooo0540oooo00<0omnL04P0M02oool02P3oool001H0
oooo00<0oooO09a8001d_ol0H03oool00`3ogi`0B01809cOo`1L0?ooo`030000003oool0oooo0640
oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`020?ooo`<000000`3oool00`000000
oooo0?ooo`1<0?ooo`030?oog`2LB000M;oo00X0oooo000F0?ooo`030?oo_`1d0000B9cO0600oooo
00<0oono07@0001d_ol0G03oool00`000000oooo0?ooo`1Q0?ooo`030?oo_`1d0000M;oo0100oooo
00@000000?ooo`3oool0oooo0`0000040?ooo`030000003oool0oooo04H0oooo00<0oono07@00018
W=l02P3oool00180oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo0640oooo00<0_g@0001d
_`3oool0F`3oool2000006<0oooo00<0_g@0001d_`3oool0503oool00`000000oooo0?ooo`020?oo
o`<000000P3oool00`000000oooo0?ooo`0n0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cO
o`0:0?ooo`004`3oool01@3ogi`0B000000000000000M;oo0600oooo00D0oono07@000000000001d
0;ooo`1K0?ooo`030000003oool0oooo0600oooo00D0oono07@000000000001d0;ooo`0J0?ooo`03
0000003oool0oooo00<000000`3oool01@000000oooo0?ooo`3oool0000003P0oooo00D0omnL04P0
00000000000007Boo`0;0?ooo`00f03oool00`000000oooo0?ooo`230?ooo`040000003oool0oooo
0?ooo`<0000000@0oooo00000000000000000P3oool010000000oooo0?ooo`0000110?ooo`00f03o
ool00`000000oooo0?ooo`280?ooo`050000003oool0oooo0?ooo`0000000P3oool600000080oooo
00@000000?ooo`3oool000000`3oool00`000000oooo0?ooo`0d0?ooo`00303oool00`000000oooo
0?ooo`390?ooo`030000003oool0oooo0900oooo00@000000?ooo`3oool000000P3oool600000006
0?ooo`00000000000000003oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02P0oooo000;0?ooo`<00000bP3oool00`000000oooo0?ooo`2G0?ooo`040000003oool0
oooo000000<0oooo00<000000?ooo`0000002`000000103oool000000?ooo`0000020?ooo`060000
003oool0oooo0000003oool000000P3oool00`000000oooo0?ooo`0H0?ooo`00303oool00`000000
oooo0?ooo`390?ooo`030000003oool0oooo0:40oooo00@000000?ooo`3oool000000P3oool01@00
0000oooo0?ooo`000000oooo0140000000T0oooo0000003oool0oooo0000003oool000000?ooo`00
00003`3oool00=P0oooo00<000000?ooo`3oool0[@3oool00`000000oooo000000020?ooo`060000
003oool0oooo0000003oool000000P3oool>000000d0oooo003H0?ooo`030000003oool0oooo0;`0
oooo00<000000?ooo`0000000P3oool010000000oooo0000003oool4000000`0oooo003H0?ooo`03
0000003oool0oooo0<L0oooo00<000000?ooo`0000002`3oool00=P0oooo00<000000?ooo`3oool0
b03oool3000000X0oooo003H0?ooo`030000003oool0oooo0<T0oooo00<000000?ooo`3oool02@3o
ool000030?ooo`00000000000?l00000[@0000010?ooo`002P3oool3000000030?ooo`000000oooo
0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0000000;0?ooo`002`3oool010000000
oooo0?ooo`00000C0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo1000000:0?ooo`00303oool00`000000
oooo0000001T0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0HP3oool00`000000
oooo0?ooo`1R0?ooo`030000003oool0000000/0oooo000;0?ooo`@00000b@3oool00`000000oooo
0?ooo`3E0?ooo`00303oool5000000060?ooo`000000oooo0000003oool000000P3oool00`000000
oooo0?ooo`2l0?ooo`030000003oool0oooo0=D0oooo000>0?ooo`h000000P3oool00`000000oooo
000000020?ooo`060000003oool0oooo0000003oool00000[`3oool00`000000oooo0?ooo`3E0?oo
o`00403oool01`000000oooo0000003oool000000?ooo`0000000P3ooolA000000050?ooo`000000
oooo0?ooo`0000000P3oool010000000oooo0?ooo`00002S0?ooo`030000003oool0oooo0=D0oooo
000K0?ooo`060000003oool0oooo0000003oool000000P3oool01`000000oooo0?ooo`000000oooo
0000003oool0300000000`3oool000000?ooo`020?ooo`040000003oool0oooo000009T0oooo00<0
00000?ooo`3oool0e@3oool002/0oooo00@000000?ooo`3oool000000P3oool01@000000oooo0?oo
o`000000oooo00<0000000<0oooo000000000000100000020?ooo`050000003oool0oooo0?ooo`00
0000T@3oool00`000000oooo0?ooo`3E0?ooo`00=`3oool01@000000oooo0?ooo`3oool000000080
oooo00<000000?ooo`3oool00`000000103oool000000000000000020?ooo`050000003oool0oooo
0?ooo`000000R@3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00/0oooo00120?oo
o`050000003oool0oooo0?ooo`0000000P3oool3000000040?ooo`000000000000000080oooo00<0
00000?ooo`3oool0P`3oool00`000000oooo0?ooo`360?ooo`<00000303oool004X0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`<000000`3oool00`000000oooo0?ooo`1n0?ooo`030000003o
ool0oooo0<L0oooo00<000000?ooo`3oool02`3oool00580oooo00@000000?ooo`3oool0oooo0`00
00030?ooo`030000003oool0oooo07T0oooo0P00003F0?ooo`00E`3oool010000000oooo0?ooo`3o
ool3000000@0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`3E0?ooo`00G03oool0
0`000000oooo0?ooo`020?ooo`<00000M03oool00`000000oooo0?ooo`3E0?ooo`00HP3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0e@3oool006L0
oooo0`00001^0?ooo`030000003oool0oooo0=D0oooo001X0?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`3E0?ooo`00KP3oool3000006L0oooo00<0
00000?ooo`3oool0e@3oool006l0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1N
0?ooo`030000003oool0oooo0=D0oooo001f0?ooo`<00000G`3oool00`000000oooo0?ooo`3E0?oo
o`00M`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo0=D0oooo00200?ooo`030000
003oool0oooo05D0oooo00<000000?ooo`3oool0e@3oool007l0oooo0`00001F0?ooo`030000003o
ool0oooo0=D0oooo00200?ooo`030000003oool0oooo05D0oooo0P00003F0?ooo`00f03oool00`00
0000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00RP3oool00`000000oooo
0?ooo`1;0?ooo`030000003oool0oooo0=D0oooo00290?ooo`<00000C03oool00`000000oooo0?oo
o`3E0?ooo`00RP3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0=D0oooo003H0?oo
o`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000
003oool0oooo0;l0oooo00<000000?ooo`3oool04`3oool00=P0oooo00<000000?ooo`3oool0_P3o
ool3000001@0oooo002G0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0_`3oool0
0`000000oooo0?ooo`0C0?ooo`00UP3oool3000003l0oooo00<000000?ooo`3oool0e@3oool009L0
oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`3E0?ooo`00f03oool200000=H0oooo
003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?oo
o`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000
003oool0oooo0=D0oooo002W0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0e@3o
ool00:H0oooo0`00000_0?ooo`030000003oool0oooo0=D0oooo002W0?ooo`030000003oool0oooo
01`0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`090?ooo`030000003oool0oooo
0=D0oooo00370?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool02@3oool00`000000
oooo0?ooo`3E0?ooo`00b03oool00`3o_g@007Bo0?ooo`0=0?ooo`030000003oool0oooo0=D0oooo
00390?ooo`030?noM000M;l0oooo00`0oooo00<000000?ooo`3oool0e@3oool00<X0oooo00<0okmd
04RLg`3oool02`3oool300000=D0oooo003;0?ooo`030=nLB018W=l0oooo00X0oooo00<000000?oo
o`3oool0e@3oool00<`0oooo00<0_g@007Boo`3oool02@3oool00`000000oooo0?ooo`3E0?ooo`00
c03oool00`3ogi`0B01d0;ooo`090?ooo`030000003oool0oooo0=D0oooo003<0?ooo`030?oog`2L
B000M;oo00T0oooo00<000000?ooo`3oool0e@3oool00<`0oooo00<0oono07@00018W=l02@3oool0
0`000000oooo0?ooo`3E0?ooo`00b03oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol02@3o
ool00`000000oooo0?ooo`3E0?ooo`00b@3oool01@3ogi`0B000000000000000M;oo00X0oooo00<0
00000?ooo`3oool0e@3oool00;/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`3E
0?ooo`00^P3oool3000001/0oooo00<000000?ooo`3oool0e@3oool00;/0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03o
ool00`000000oooo0?ooo`3E0?ooo`00f03oool200000=H0oooo003H0?ooo`030000003oool0oooo
0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo
003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?oo
o`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000
003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0
oooo0=D0oooo003H0?ooo`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0:@3oool00=P0
oooo00<000000?ooo`3oool0Z03oool3000002X0oooo003H0?ooo`800000ZP3oool00`000000oooo
0?ooo`0Y0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E
0?ooo`00e`3oool200000=L0oooo003F0?ooo`<00000e`3oool00=L0oooo0P00003G0?ooo`00f03o
ool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`00
0000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo
0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool200000=H0oooo003H0?oo
o`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000
003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0
oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0
oooo00390?ooo`030?oog`2LB000B9cO00`0oooo00<000000?ooo`3oool0e@3oool00<X0oooo00<0
gia80018W03Oool02`3oool00`000000oooo0?ooo`3E0?ooo`00bP3oool00`3ogi`0B01809cOo`0;
0?ooo`030000003oool0oooo0=D0oooo00360?ooo`030?noM0000000000000@0000000<004RL0=oo
o`3oool0203oool00`000000oooo0?ooo`3E0?ooo`00aP3oool00`3ooml0W4Q809cOo`020?ooo`03
0;md0000M;l0oooo00X0oooo00<000000?ooo`3oool0e@3oool00<L0oooo00L0oono07@0M02oool0
oooo0?noM000B9`0gooo00X0oooo0`00003E0?ooo`00b03oool01P3ogi`0B4RL0=ooo`3ogi`0B018
09cOo`X0oooo00<000000?ooo`3oool0e@3oool00<T0oooo00D0okmd001d_`3ooml0W4P004RLg`0:
0?ooo`030000003oool0oooo0=D0oooo003:0?ooo`040=nLB018W=l0_g@0001d_`X0oooo00<00000
0?ooo`3oool0e@3oool00</0oooo00@0gia807AdM00007@0_ooo2@3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0;40oooo003<0?ooo`030;md000004P0W=oo00T0oooo00<000000?oo
o`3oool0803oool300000;80oooo003<0?ooo`030?oog`2LB000B9cO00T0oooo00<000000?ooo`3o
ool08@3oool00`000000oooo0?ooo`2a0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03o
ool00`000000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool00`00
0000oooo0?ooo`3E0?ooo`00f03oool00`000000oooo0?ooo`3E0?ooo`00f03oool200000=H0oooo
003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?oo
o`030000003oool0oooo0880oooo00<000000?ooo`3oool0D03oool00=P0oooo00<000000?ooo`3o
ool0P@3oool300000540oooo003H0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0
D03oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool0
0=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo
00<000000?ooo`3oool0e@3oool00=P0oooo00<000000?ooo`3oool0e@3oool00=P0oooo00<00000
0?ooo`3oool0D03oool00`000000oooo0?ooo`220?ooo`00f03oool200000500oooo0`0000230?oo
o`00f03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0880oooo003H0?ooo`030000
003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0
oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0oooo003H0?ooo`030000003oool0oooo0=D0
oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.13864, -5.32446, \
0.0148355, 0.0597521}}]
}, Closed]],

Cell[TextData[{
  "These figures illustrate the typical behavior of a ",
  StyleBox["relaxation oscillator",
    FontSlant->"Italic"],
  " for which stress accumulates slowly and is then released rapidly.  Thus, \
intervals of relatively slow variation alternate with brief intervals of \
rather rapid variation, with the ratio between these intervals being governed \
by a relaxation parameter.  The characteristics of relaxation oscillators are \
investigated in the next section. "
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Phase shift", "ExerciseTitle"],

Cell[TextData[{
  "a) For a given value of ",
  StyleBox["\[CurlyEpsilon]", "InlineInput",
    FontWeight->"Bold"],
  " in the van der Pol equation, compare two ",
  StyleBox["x[t]", "InlineInput",
    FontWeight->"Bold"],
  " curves for different initial conditions and show that for large times \
they differ only by a time shift.  Interpret this result.\nb) Suppose that a \
van der Pol oscillator has already settled upon its limit cycle, but a clumsy \
passerby bumps into the table and kicks the system off cycle.  Describe the \
subsequent motion qualitatively.  How does the perturbation alter the long \
term behavior of the system?"
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Convergence upon the limit cycle", "ExerciseTitle"],

Cell["\<\
Produce a figure in which phase-space trajectories for a two-dimensional \
array of starting conditions are plotted together.  The starting conditions \
should cover a wide range but need not be closely spaced.  Interpret your \
result.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["van der Pol equation without driving term", "ExerciseTitle"],

Cell["Study the behavior of the modified van der Pol equation", "ExerciseText"],

Cell[BoxData[
    FormBox[
      RowBox[{
        FractionBox[\(\[PartialD]\^2 x[t]\), \(\[PartialD]t\^2\),
          MultilineFunction->None], "\[Equal]", 
        RowBox[{"\[CurlyEpsilon]", " ", \((1 - x[t]\^2)\), " ", 
          FractionBox[\(\[PartialD]x[t]\), \(\[PartialD]t\),
            MultilineFunction->None]}]}], TraditionalForm]], "DisplayFormula",\

  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "in which the underlying (unit frequency) oscillation has been removed.  \
Does this system still display limit-cycle behavior?  Why isn't the behavior \
the same as that of the original equation in the limit of large ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\)]],
  " where one might suppose that the term proportional to ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\)]],
  " would dominate?"
}], "ExerciseText",
  CellDingbat->None]
}, Closed]],

Cell[CellGroupData[{

Cell["Lienard equation", "ExerciseTitle"],

Cell["Study the behavior of the Lienard equation", "ExerciseText"],

Cell[BoxData[
    FormBox[
      RowBox[{
        RowBox[{
          FractionBox[\(\[PartialD]\^2 x[t]\), \(\[PartialD]t\^2\),
            MultilineFunction->None], "+", " ", 
          RowBox[{\((x[t]\^2 - \[Beta])\), " ", 
            FractionBox[\(\[PartialD]x[t]\), \(\[PartialD]t\),
              MultilineFunction->None]}], "+", \(x[t]\)}], " ", "\[Equal]", 
        " ", "0"}], TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "for ",
  Cell[BoxData[
      \(TraditionalForm\`\[Beta] > 0\)]],
  "."
}], "ExerciseText",
  CellDingbat->None]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Relaxation oscillators", "Section"],

Cell[CellGroupData[{

Cell["Introduction", "Subsection"],

Cell[TextData[{
  "There are an infinite variety of transformations which represent a \
second-order differential equation as a pair of first-order differential \
equations.  Further insight into the behavior of the van der Pol equation can \
be obtained by representing it in the form of a ",
  StyleBox["relaxation oscillator",
    FontSlant->"Italic"],
  " described by a pair of first-order equations of the form"
}], "Text"],

Cell[BoxData[
    FormBox[
      RowBox[{"{", 
        RowBox[{
          RowBox[{
            RowBox[{"\[Alpha]", " ", 
              FractionBox[\(\[PartialD]x[t]\), \(\[PartialD]t\),
                MultilineFunction->None]}], "\[Equal]", \(y[t] - f[x]\)}], 
          ",", "   ", 
          RowBox[{
            FractionBox[\(\[PartialD]y[t]\), \(\[PartialD]t\),
              MultilineFunction->None], "\[Equal]", \(-x[t]\)}]}], "}"}], 
      TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where the relaxation parameter ",
  Cell[BoxData[
      \(TraditionalForm\`0 < \[Alpha] \[LessLess] 1\)]],
  " is small and where ",
  Cell[BoxData[
      \(TraditionalForm\`f[x]\)]],
  " is a continuous function.  An important feature of this representation is \
that the null clines are readily apparent.  A ",
  StyleBox["null cline",
    FontSlant->"Italic"],
  " is a curve in phase space upon which one of the phase variables is \
constant.  For this system it is clear that"
}], "Text"],

Cell[BoxData[{
    \(TraditionalForm\`x \[Equal] 
      0\ \ \[AlignmentMarker]\[AlignmentMarker]\[DoubleLongRightArrow]\ \ \(\
\[PartialD]y\/\[PartialD]t\) \[Equal] 0\), "\[IndentingNewLine]", 
    \(TraditionalForm\`y \[Equal] 
      f[x]\[AlignmentMarker]\ \[AlignmentMarker]\[DoubleLongRightArrow]\ \ \(\
\[PartialD]x\/\[PartialD]t\) \[Equal] 0\)}], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->AlignmentMarker,
  TextJustification->0],

Cell[TextData[{
  "represent the null clines.  Fixed points occur where null clines \
intersect.  For this type of system, one of the null clines is simply the ",
  StyleBox["y",
    FontSlant->"Italic"],
  "-axis, and is trivial, while the nontrivial function ",
  Cell[BoxData[
      \(TraditionalForm\`f[x]\)]],
  " specifies the dynamics.  "
}], "Text"],

Cell[TextData[{
  "Suppose that ",
  StyleBox["y",
    FontSlant->"Italic"],
  " is near the null cline ",
  Cell[BoxData[
      \(TraditionalForm\`f[x]\)]],
  " , where nearness is gauged in terms of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha]\^\(-1\)\)]],
  "; in other words, we consider the phase point to be near the null cline \
when ",
  Cell[BoxData[
      \(TraditionalForm\`\[LeftBracketingBar]y - 
            f[x]\[RightBracketingBar] \[LessTilde] \[Alpha]\^\(-1\)\)]],
  ".  Under those circumstances both components of the phase velocity will be \
of order unity.  Conversely, when ",
  Cell[BoxData[
      \(TraditionalForm\`\[LeftBracketingBar]y - 
            f[x]\[RightBracketingBar] \[GreaterGreater] \[Alpha]\^\(-1\)\)]],
  " the velocity scaled by \[Alpha] becomes very large while the other \
velocity remains of order unity.  Thus, the relaxation parameter \[Alpha] \
regulates the rate at which various portions of the limit cycle are \
traversed."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Relaxation representation of van der Pol equation", "Subsection"],

Cell["\<\
First, we show that the van der Pol equation can be represented as a \
relaxation oscillator.  It is useful to define a function which establishes \
the equations for any choice of null cline.\
\>", "Text"],

Cell[BoxData[
    \(RelaxationEquations[x_, y_, f_, 
        t_, \[Alpha]_] := \ {\[Alpha]\ D[x[t], t] \[Equal] y[t] - f[x[t]], 
        D[y[t], t] \[Equal] \(-x[t]\)}\)], "Input"],

Cell[TextData[{
  "Next, we show that a judicious choice of ",
  Cell[BoxData[
      \(TraditionalForm\`f[x]\)]],
  " gives an equation"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(testEq = 
      RelaxationEquations[x, y, \((\(1\/3\) #\^3 - #)\) &, 
        t, \[Alpha]]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{
        RowBox[{
          RowBox[{"\[Alpha]", " ", 
            RowBox[{
              SuperscriptBox["x", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}]}], 
          "\[Equal]", \(x[t] - x[t]\^3\/3 + y[t]\)}], ",", 
        RowBox[{
          RowBox[{
            SuperscriptBox["y", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}], 
          "\[Equal]", \(-x[t]\)}]}], "}"}]], "Output"]
}, Closed]],

Cell[TextData[{
  "which can be transformed into the van der Pol equation by means of the \
variable changes ",
  Cell[BoxData[
      \(TraditionalForm\`{t \[Rule] \[Alpha]\ t, \[Alpha] \[Rule] \
\[CurlyEpsilon]\^\(-2\)}\)]],
  ": "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(D[testEq\[LeftDoubleBracket]1\[RightDoubleBracket], 
            t] /. \((testEq\[LeftDoubleBracket]2\[RightDoubleBracket] /. 
              Equal \[Rule] Rule)\)\) /. {\(x\^\[Prime]\)[
              t] \[Rule] \[CurlyEpsilon]\ \(x\^\[Prime]\)[
                t], \(x\^\[DoublePrime]\)[
              t] \[Rule] \(\[CurlyEpsilon]\^2\) \(x\^\[DoublePrime]\)[
                t], \[Alpha] \[Rule] \[CurlyEpsilon]\^\(-2\)} // 
      Simplify\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{\(x[t]\), "+", 
        RowBox[{"\[CurlyEpsilon]", " ", \(x[t]\^2\), " ", 
          RowBox[{
            SuperscriptBox["x", "\[Prime]",
              MultilineFunction->None], "[", "t", "]"}]}], "+", 
        RowBox[{
          SuperscriptBox["x", "\[Prime]\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}], "\[Equal]", 
      RowBox[{"\[CurlyEpsilon]", " ", 
        RowBox[{
          SuperscriptBox["x", "\[Prime]",
            MultilineFunction->None], "[", "t", "]"}]}]}]], "Output"]
}, Closed]],

Cell[TextData[
"Thus, we have identified the characteristic null cline and have related the \
relaxation parameter to the damping parameter found in the original form van \
der Pol equation.  It is important to recognize that the time scale must be \
scaled by \[Alpha] to represent the van der Pol equation as a relaxation \
oscillator. "], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Characteristic behavior of relaxation oscillators", "Subsection"],

Cell["\<\
The following functions are designed to illustrate the characteristic \
behavior of relaxation oscillators by plotting the velocity field, the null \
clines, and phase trajectories together.  The first shows the phase \
trajectory as a continuous curve, while the second shows the phase velocities \
stroboscopically using points at equal time intervals.\
\>", "Text"],

Cell[BoxData[{
    \(\(\(Clear[RelaxationPlot];\)\(\n\)
    \)\), "\n", 
    \(RelaxationPlot[f_, \[Alpha]_, xmax_, ymax_, x0_, y0_, tmax_, 
        opts___]\  := 
      Show[{ParametricPlot[
            Evaluate[{x[t], y[t]} /. 
                NDSolve[
                  Join[RelaxationEquations[x, y, f, 
                      t, \[Alpha]], {x[0] \[Equal] x0, y[0] \[Equal] y0}], {x[
                      t], y[t]}, {t, 0, tmax}, MaxSteps \[Rule] 10\^4]], {t, 
              0, tmax}, DisplayFunction \[Rule] Identity], 
          PlotVectorField[{\(y - f[x]\)\/\[Alpha], \(-x\)}, {x, \(-xmax\), 
              xmax}, {y, \(-ymax\), ymax}, DisplayFunction \[Rule] Identity], 
          Plot[f[x], {x, \(-xmax\), xmax}, 
            PlotStyle \[Rule] RGBColor[1, 0, 0], 
            DisplayFunction \[Rule] Identity]}, opts, 
        DisplayFunction \[Rule] $DisplayFunction]\)}], "Input"],

Cell[BoxData[{
    \(\(\(Clear[RelaxationStrobePlot];\)\(\n\)
    \)\), "\n", 
    \(RelaxationStrobePlot[f_, \[Alpha]_, xmax_, ymax_, x0_, y0_, tmin_, 
        tmax_, tinc_, opts___]\  := 
      Module[{sol}, 
        sol = \(NDSolve[
              Join[RelaxationEquations[x, y, f, 
                  t, \[Alpha]], {x[0] \[Equal] x0, y[0] \[Equal] y0}], {x[t], 
                y[t]}, {t, 0, tmax}, 
              MaxSteps \[Rule] 
                10\^4]\)\[LeftDoubleBracket]1\[RightDoubleBracket]; \n\t
        Show[{Plot[f[x], {x, \(-xmax\), xmax}, 
              PlotStyle \[Rule] RGBColor[1, 0, 0], 
              DisplayFunction \[Rule] Identity], \ 
            ListPlot[Table[{x[t], y[t]} /. sol, {t, tmin, tmax, tinc}], 
              DisplayFunction \[Rule] Identity]}, opts, 
          DisplayFunction \[Rule] $DisplayFunction];]\)}], "Input"],

Cell[TextData[
"We illustrate this function using the van der Pol oscillator for a fairly \
small relaxation parameter.  The null cline is a cubic equation with two \
turning points \[LongDash] the presence of these turning points is \
responsible for the appearance of limit cycle oscillations."], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(RelaxationPlot[\((\(1\/3\) #\^3 - #)\) &, 0.1, 4, 2, 0.1, 0. , 20, 
        PlotRange \[Rule] {{\(-3\), 3}, {\(-1.5\), 1.5}}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.166667 0.309017 0.206011 [
[0 .29652 -9.34375 -15.6875 ]
[0 .29652 9.34375 0 ]
[.16667 .29652 -9.34375 -15.6875 ]
[.16667 .29652 9.34375 0 ]
[.33333 .29652 -9.34375 -15.6875 ]
[.33333 .29652 9.34375 0 ]
[.66667 .29652 -5 -15.6875 ]
[.66667 .29652 5 0 ]
[.83333 .29652 -5 -15.6875 ]
[.83333 .29652 5 0 ]
[1 .29652 -5 -15.6875 ]
[1 .29652 5 0 ]
[.4875 0 -27.6875 -7.84375 ]
[.4875 0 0 7.84375 ]
[.4875 .10301 -18.6875 -7.84375 ]
[.4875 .10301 0 7.84375 ]
[.4875 .20601 -27.6875 -7.84375 ]
[.4875 .20601 0 7.84375 ]
[.4875 .41202 -19 -7.84375 ]
[.4875 .41202 0 7.84375 ]
[.4875 .51503 -10 -7.84375 ]
[.4875 .51503 0 7.84375 ]
[.4875 .61803 -19 -7.84375 ]
[.4875 .61803 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
0 .30902 m
0 .31527 L
s
gsave
0 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.16667 .30902 m
.16667 .31527 L
s
gsave
.16667 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.33333 .30902 m
.33333 .31527 L
s
gsave
.33333 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.66667 .30902 m
.66667 .31527 L
s
gsave
.66667 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.83333 .30902 m
.83333 .31527 L
s
gsave
.83333 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
1 .30902 m
1 .31527 L
s
gsave
1 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.03333 .30902 m
.03333 .31277 L
s
.06667 .30902 m
.06667 .31277 L
s
.1 .30902 m
.1 .31277 L
s
.13333 .30902 m
.13333 .31277 L
s
.2 .30902 m
.2 .31277 L
s
.23333 .30902 m
.23333 .31277 L
s
.26667 .30902 m
.26667 .31277 L
s
.3 .30902 m
.3 .31277 L
s
.36667 .30902 m
.36667 .31277 L
s
.4 .30902 m
.4 .31277 L
s
.43333 .30902 m
.43333 .31277 L
s
.46667 .30902 m
.46667 .31277 L
s
.53333 .30902 m
.53333 .31277 L
s
.56667 .30902 m
.56667 .31277 L
s
.6 .30902 m
.6 .31277 L
s
.63333 .30902 m
.63333 .31277 L
s
.7 .30902 m
.7 .31277 L
s
.73333 .30902 m
.73333 .31277 L
s
.76667 .30902 m
.76667 .31277 L
s
.8 .30902 m
.8 .31277 L
s
.86667 .30902 m
.86667 .31277 L
s
.9 .30902 m
.9 .31277 L
s
.93333 .30902 m
.93333 .31277 L
s
.96667 .30902 m
.96667 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 0 m
.50625 0 L
s
gsave
.4875 0 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.5 .10301 m
.50625 .10301 L
s
gsave
.4875 .10301 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .20601 m
.50625 .20601 L
s
gsave
.4875 .20601 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .41202 m
.50625 .41202 L
s
gsave
.4875 .41202 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .51503 m
.50625 .51503 L
s
gsave
.4875 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .61803 m
.50625 .61803 L
s
gsave
.4875 .61803 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .0206 m
.50375 .0206 L
s
.5 .0412 m
.50375 .0412 L
s
.5 .0618 m
.50375 .0618 L
s
.5 .0824 m
.50375 .0824 L
s
.5 .12361 m
.50375 .12361 L
s
.5 .14421 m
.50375 .14421 L
s
.5 .16481 m
.50375 .16481 L
s
.5 .18541 m
.50375 .18541 L
s
.5 .22661 m
.50375 .22661 L
s
.5 .24721 m
.50375 .24721 L
s
.5 .26781 m
.50375 .26781 L
s
.5 .28842 m
.50375 .28842 L
s
.5 .32962 m
.50375 .32962 L
s
.5 .35022 m
.50375 .35022 L
s
.5 .37082 m
.50375 .37082 L
s
.5 .39142 m
.50375 .39142 L
s
.5 .43262 m
.50375 .43262 L
s
.5 .45322 m
.50375 .45322 L
s
.5 .47383 m
.50375 .47383 L
s
.5 .49443 m
.50375 .49443 L
s
.5 .53563 m
.50375 .53563 L
s
.5 .55623 m
.50375 .55623 L
s
.5 .57683 m
.50375 .57683 L
s
.5 .59743 m
.50375 .59743 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% Start of user PostScript
/mathtops {
gsave
MBeginOrig
moveto
MEndOrig
currentpoint
grestore
} bind def
/MAtocoords {
mathtops 4 2 roll mathtops
4 copy pop pop
3 -1 roll sub
/arry exch def
exch sub
/arrx exch def
arrx dup mul
arry dup mul
add sqrt
/arrl exch def
translate
} bind def
/MAarrowhead1 {
gsave
MAtocoords
arrl 0. eq
{ 0 0 Mdot }
{
[ arrx arrl div
arry arrl div
-1 arry mul arrl div
arrx arrl div
0 0 ]
concat
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto fill
-0.02 0.005 moveto 0 0 lineto
-0.02 -0.005 lineto
-0.02 0.005 lineto stroke
} ifelse
grestore
} def
/MAarrowzero1 {
gsave
mathtops translate
0 0 Mdot
grestore
} def
% End of user PostScript
.5 Mabswid
.51667 .30902 m
.54226 .30568 L
.593 .29839 L
.63617 .29117 L
.68061 .28167 L
.70262 .27532 L
.7127 .27169 L
.722 .26767 L
.73483 .26019 L
.74279 .25298 L
.74778 .24517 L
.74993 .23792 L
.75027 .22951 L
.74904 .22154 L
.74694 .21425 L
.74437 .20755 L
.7369 .19267 L
.71588 .16304 L
.70119 .1478 L
.68551 .13499 L
.66589 .12293 L
.65416 .1174 L
.63928 .11183 L
.62437 .10756 L
.61505 .10543 L
.60531 .10357 L
.58328 .10047 L
.55846 .09834 L
.54189 .0975 L
.52089 .09691 L
.49534 .09675 L
.46745 .09708 L
.40745 .09907 L
.36858 .10111 L
.33124 .10361 L
.29083 .10709 L
.25654 .11097 L
.22513 .11584 L
.20008 .12161 L
.18255 .12813 L
.17249 .13456 L
.16723 .14046 L
.16422 .14694 L
.16285 .15805 L
.16378 .16844 L
.16589 .17979 L
.16873 .19189 L
.18152 .23803 L
.20741 .31587 L
.23917 .38878 L
Mistroke
.26085 .42601 L
.2727 .44259 L
.28508 .45738 L
.29838 .47067 L
.31502 .48399 L
.33311 .49502 L
.3535 .50406 L
.36729 .50857 L
.38225 .51235 L
.40179 .51594 L
.41462 .51764 L
.42777 .51897 L
.45974 .52091 L
.4819 .52147 L
.50566 .52159 L
.55898 .52057 L
.59093 .51934 L
.62949 .51735 L
.70011 .51212 L
.75808 .50533 L
.80173 .49622 L
.81669 .49062 L
.82565 .48534 L
.83478 .47433 L
.83667 .46803 L
.83722 .46221 L
.83653 .4515 L
.83455 .44024 L
.82985 .42055 L
.82495 .40242 L
.81306 .36225 L
.78361 .27917 L
.76692 .24147 L
.74954 .20871 L
.73022 .17931 L
.7054 .15089 L
.68981 .13757 L
.67249 .12608 L
.64891 .11489 L
.63245 .10938 L
.61432 .10495 L
.59741 .10198 L
.57568 .09938 L
.55052 .09758 L
.5224 .09663 L
.48098 .09655 L
.42361 .09808 L
.35821 .10144 L
.29471 .1064 L
.23788 .11333 L
Mistroke
.20065 .12112 L
.17747 .13057 L
.16645 .14128 L
.16296 .15271 L
.16313 .16362 L
.16458 .17338 L
.16681 .18397 L
.19149 .2701 L
.21765 .34196 L
.23421 .37891 L
.25458 .41619 L
.27735 .44843 L
.29016 .46275 L
.30337 .47502 L
.31925 .48687 L
.33718 .49709 L
.34789 .50188 L
.3608 .50659 L
.37735 .51123 L
.39568 .51496 L
.41438 .51761 L
.43888 .51983 L
.4676 .52117 L
.50025 .5216 L
.54537 .52097 L
.60647 .51861 L
.67922 .51392 L
.74313 .50743 L
.79409 .49834 L
.81217 .49259 L
.82319 .48704 L
.8337 .47648 L
.83706 .46508 L
.83691 .45484 L
.83553 .44528 L
.83337 .43485 L
.83067 .42373 L
.80445 .33565 L
.77515 .25925 L
.75576 .2197 L
.73492 .18582 L
.7231 .17015 L
.71072 .15619 L
.69806 .14423 L
.68272 .13248 L
.67363 .12674 L
.66254 .12081 L
.65104 .11572 L
.63917 .11144 L
.62286 .10685 L
Mistroke
.61327 .10474 L
.6017 .10264 L
.58969 .10092 L
.57726 .09953 L
.54803 .09746 L
.51093 .09648 L
.46549 .09681 L
.41475 .09844 L
.35006 .10197 L
.27853 .10804 L
.24387 .11241 L
.21825 .11686 L
.18524 .12653 L
.17018 .13635 L
.16438 .14587 L
.16278 .15604 L
.16325 .16473 L
.16474 .17425 L
.16692 .18444 L
.16958 .19528 L
.17477 .21462 L
.2021 .30133 L
.23197 .37425 L
.24937 .40746 L
.27082 .44012 L
.29729 .46966 L
.31605 .48471 L
.3258 .49096 L
.33777 .49737 L
.34872 .50221 L
.36206 .50699 L
.37854 .51151 L
.3969 .51516 L
.41924 .51815 L
.44544 .52024 L
.47473 .52135 L
.51469 .52153 L
.57052 .52017 L
.60804 .51853 L
.64597 .51633 L
.68686 .51329 L
.72261 .50987 L
.75882 .50522 L
.78617 .50023 L
.81706 .49044 L
.8307 .48072 L
.83631 .46975 L
.83717 .46357 L
.83715 .45796 L
.83591 .44742 L
Mistroke
.83399 .43764 L
.82857 .41566 L
.81622 .37249 L
.79162 .29959 L
.75989 .22745 L
.73843 .19095 L
.72785 .17617 L
.71494 .16068 L
.70243 .14811 L
.68698 .13547 L
.67729 .12895 L
.66745 .1233 L
.65626 .11791 L
.6445 .11325 L
.6319 .10923 L
.6188 .10591 L
.60186 .10267 L
.58005 .09981 L
.55406 .09777 L
.51829 .09656 L
.47471 .09664 L
.42431 .09805 L
.35864 .10141 L
.28486 .10738 L
.2273 .11512 L
.19304 .1235 L
.17241 .13424 L
.1669 .14054 L
.16419 .14645 L
.16277 .15735 L
.16306 .16286 L
.16382 .16894 L
.16485 .1748 L
.16613 .18095 L
.16872 .19188 L
.17522 .21621 L
.18793 .25898 L
.21401 .33297 L
.23031 .37076 L
.24757 .40431 L
.26595 .43344 L
.27688 .44786 L
.28977 .46235 L
.30196 .47382 L
.31736 .48561 L
.33388 .49543 L
.35256 .50371 L
.36621 .50826 L
.38332 .51259 L
.40275 .51608 L
Mistroke
.42416 .51864 L
.43966 .51988 L
.45589 .52075 L
.47632 .52138 L
.50143 .5216 L
.55344 .52074 L
.58885 .51944 L
.62528 .51759 L
.6992 .5122 L
.76073 .50493 L
.79914 .49698 L
.82265 .48739 L
.83375 .4764 L
.83605 .47075 L
.83711 .46449 L
.83676 .45341 L
.83512 .44305 L
.8331 .43366 L
.8306 .42347 L
.82516 .40316 L
.81476 .36774 L
.78886 .29237 L
.77271 .25381 L
.75568 .21956 L
.73811 .19047 L
.71623 .16211 L
.70291 .14855 L
.68592 .1347 L
.66737 .12326 L
.6459 .11376 L
.63003 .1087 L
.62112 .10644 L
.61205 .10449 L
.59258 .1013 L
.56633 .09858 L
.5484 .09748 L
.52893 .09677 L
.48408 .09652 L
.45311 .09711 L
.42066 .0982 L
.38104 .10009 L
.33632 .10293 L
.26782 .10925 L
.23542 .11372 L
.21157 .11832 L
.18295 .12759 L
.16864 .1381 L
.16382 .14775 L
.16277 .15687 L
.16352 .16688 L
Mistroke
.16541 .1776 L
.17108 .20106 L
.17713 .22297 L
.1892 .26298 L
.21577 .33736 L
.24439 .39861 L
.26362 .43011 L
.28507 .45736 L
.29795 .47027 L
.31361 .48299 L
.32292 .48921 L
.33431 .49565 L
.34595 .50107 L
.35852 .50583 L
.37189 .50985 L
.38895 .51374 L
.40748 .51675 L
.42809 .519 L
.45531 .52073 L
.49358 .52159 L
.54575 .52096 L
.60768 .51854 L
.66988 .51465 L
.73336 .50864 L
.77678 .50216 L
.80845 .49401 L
.82678 .48447 L
.83498 .47386 L
.83716 .4637 L
.83687 .45444 L
.83531 .44406 L
.83326 .43439 L
.82775 .41262 L
.8154 .3698 L
.79078 .29738 L
.759 .22574 L
.73741 .18944 L
.72556 .17322 L
.71364 .15927 L
.70112 .14692 L
.68557 .13446 L
.67583 .12806 L
.66589 .12249 L
.65457 .11718 L
.64259 .11258 L
.62985 .10865 L
.61653 .10542 L
.59927 .10226 L
.57693 .0995 L
.55034 .09757 L
Mistroke
.51355 .0965 L
.46873 .09675 L
.4169 .09835 L
.35091 .10191 L
.2781 .10809 L
.22276 .11596 L
.19043 .12443 L
.17197 .13463 L
.16686 .14061 L
.16426 .14622 L
.16277 .15766 L
.16318 .16406 L
.16397 .16986 L
.16502 .17567 L
.16632 .18179 L
.16893 .1927 L
.17544 .21701 L
.1882 .25982 L
.2141 .33321 L
.23031 .37077 L
.24754 .40426 L
.2658 .43323 L
.27667 .4476 L
.28953 .4621 L
.30301 .47471 L
.31699 .48535 L
.33345 .4952 L
.35211 .50353 L
.3648 .50784 L
.38057 .51198 L
.39824 .51538 L
.41747 .51796 L
.44532 .52023 L
.46433 .52107 L
.48454 .52151 L
.52975 .52131 L
.59088 .51935 L
.65982 .51538 L
.73339 .50864 L
.78627 .5002 L
.81524 .49129 L
.82536 .48555 L
.83122 .4801 L
.83491 .47403 L
.83675 .46751 L
.83723 .46127 L
.83696 .45539 L
.83538 .44443 L
.83343 .43513 L
.83098 .42498 L
Mistroke
.82559 .40472 L
.81525 .36931 L
.78942 .29382 L
.77349 .25553 L
.75666 .22136 L
.73933 .1923 L
.71775 .16382 L
.70473 .15026 L
.68815 .13632 L
.67014 .12476 L
.64937 .11507 L
.63503 .11014 L
.61712 .10554 L
.59873 .10218 L
.57402 .09922 L
.55721 .09796 L
.53968 .0971 L
.51619 .09653 L
.4902 .09647 L
.43317 .09773 L
.39475 .09937 L
.35627 .10156 L
.28452 .10741 L
.22807 .11498 L
.19447 .12302 L
.1745 .13258 L
.16535 .14341 L
.16342 .14956 L
.1628 .15514 L
.16341 .16603 L
.16513 .17623 L
.16973 .19585 L
.17449 .21361 L
.18598 .25271 L
.21516 .33583 L
.2315 .37328 L
.24816 .40536 L
.26551 .43283 L
.28777 .46026 L
.30226 .47408 L
.31781 .48591 L
.33548 .49624 L
.35879 .50592 L
.37376 .51034 L
.38988 .51392 L
.41119 .51723 L
.42375 .5186 L
.43928 .51985 L
.47097 .52126 L
.51461 .52153 L
Mistroke
.56752 .52028 L
.62552 .51758 L
.66459 .51504 L
.70559 .51161 L
.76836 .50368 L
.79355 .49848 L
.81049 .49325 L
.82287 .48725 L
.8308 .48059 L
.83458 .47477 L
.8366 .46838 L
.83722 .46233 L
.83707 .45671 L
.83569 .44613 L
.83339 .43493 L
Mfstroke
.0001 w
% Start of user PostScript
-4. -2. -3.7734475 -1.9940899 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. -1.7142857 -3.769226 -1.7083756 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. -1.4285714 -3.7650045 -1.4226614 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. -1.1428571 -3.7607831 -1.1369471 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. -0.85714286 -3.7565616 -0.85123279 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. -0.57142857 -3.7523401 -0.56551851 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. -0.28571429 -3.7481186 -0.27980422 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 0. -3.7438972 0.0059100652 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 0.28571429 -3.7396757 0.29162435 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 0.57142857 -3.7354542 0.57733864 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 0.85714286 -3.7312327 0.86305292 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 1.1428571 -3.7270113 1.1487672 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 1.4285714 -3.7227898 1.4344815 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 1.7142857 -3.7185683 1.7201958 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-4. 2. -3.7143468 2.0059101 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -2. -3.310284 -1.9949342 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -1.7142857 -3.3060625 -1.7092199 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -1.4285714 -3.301841 -1.4235057 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -1.1428571 -3.2976195 -1.1377914 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -0.85714286 -3.2933981 -0.85207709 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -0.57142857 -3.2891766 -0.5663628 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 -0.28571429 -3.2849551 -0.28064852 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 0. -3.2807336 0.0050657702 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 0.28571429 -3.2765122 0.29078006 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 0.57142857 -3.2722907 0.57649434 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 0.85714286 -3.2680692 0.86220863 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 1.1428571 -3.2638477 1.1479229 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 1.4285714 -3.2596263 1.4336372 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 1.7142857 -3.2554048 1.7193515 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-3.4285714 2. -3.2511833 2.0050658 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.8571429 -2. -2.8140379 -1.9957785 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.8571429 -1.7142857 -2.8098164 -1.7100642 MAarrowhead1
% End of user PostScript
.02381 .01472 m
.0324 .01558 L
s
% Start of user PostScript
-2.8571429 -1.4285714 -2.8055949 -1.42435 MAarrowhead1
% End of user PostScript
.02381 .07358 m
.0331 .07445 L
s
% Start of user PostScript
-2.8571429 -1.1428571 -2.8013734 -1.1386357 MAarrowhead1
% End of user PostScript
.02381 .13244 m
.03381 .13331 L
s
% Start of user PostScript
-2.8571429 -0.85714286 -2.797152 -0.85292138 MAarrowhead1
% End of user PostScript
.02381 .1913 m
.03451 .19217 L
s
% Start of user PostScript
-2.8571429 -0.57142857 -2.7929305 -0.5672071 MAarrowhead1
% End of user PostScript
.02381 .25016 m
.03522 .25103 L
s
% Start of user PostScript
-2.8571429 -0.28571429 -2.788709 -0.28149281 MAarrowhead1
% End of user PostScript
.02381 .30902 m
.03592 .30989 L
s
% Start of user PostScript
-2.8571429 0. -2.7844875 0.0042214752 MAarrowhead1
% End of user PostScript
.02381 .36788 m
.03662 .36875 L
s
% Start of user PostScript
-2.8571429 0.28571429 -2.7802661 0.28993576 MAarrowhead1
% End of user PostScript
.02381 .42674 m
.03733 .42761 L
s
% Start of user PostScript
-2.8571429 0.57142857 -2.7760446 0.57565005 MAarrowhead1
% End of user PostScript
.02381 .4856 m
.03803 .48647 L
s
% Start of user PostScript
-2.8571429 0.85714286 -2.7718231 0.86136433 MAarrowhead1
% End of user PostScript
.02381 .54446 m
.03873 .54533 L
s
% Start of user PostScript
-2.8571429 1.1428571 -2.7676016 1.1470786 MAarrowhead1
% End of user PostScript
.02381 .60332 m
.03944 .60419 L
s
% Start of user PostScript
-2.8571429 1.4285714 -2.7633802 1.4327929 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.8571429 1.7142857 -2.7591587 1.7185072 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.8571429 2. -2.7549372 2.0042215 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.2857143 -2. -2.2902229 -1.9966228 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.2857143 -1.7142857 -2.2860015 -1.7109085 MAarrowhead1
% End of user PostScript
.11905 .01472 m
.1197 .01541 L
s
% Start of user PostScript
-2.2857143 -1.4285714 -2.28178 -1.4251942 MAarrowhead1
% End of user PostScript
.11905 .07358 m
.12041 .07427 L
s
% Start of user PostScript
-2.2857143 -1.1428571 -2.2775585 -1.13948 MAarrowhead1
% End of user PostScript
.11905 .13244 m
.12111 .13313 L
s
% Start of user PostScript
-2.2857143 -0.85714286 -2.273337 -0.85376568 MAarrowhead1
% End of user PostScript
.11905 .1913 m
.12181 .19199 L
s
% Start of user PostScript
-2.2857143 -0.57142857 -2.2691156 -0.56805139 MAarrowhead1
% End of user PostScript
.11905 .25016 m
.12252 .25085 L
s
% Start of user PostScript
-2.2857143 -0.28571429 -2.2648941 -0.28233711 MAarrowhead1
% End of user PostScript
.11905 .30902 m
.12322 .30971 L
s
% Start of user PostScript
-2.2857143 0. -2.2606726 0.0033771801 MAarrowhead1
% End of user PostScript
.11905 .36788 m
.12392 .36857 L
s
% Start of user PostScript
-2.2857143 0.28571429 -2.2564511 0.28909147 MAarrowhead1
% End of user PostScript
.11905 .42674 m
.12463 .42743 L
s
% Start of user PostScript
-2.2857143 0.57142857 -2.2522297 0.57480575 MAarrowhead1
% End of user PostScript
.11905 .4856 m
.12533 .48629 L
s
% Start of user PostScript
-2.2857143 0.85714286 -2.2480082 0.86052004 MAarrowhead1
% End of user PostScript
.11905 .54446 m
.12604 .54515 L
s
% Start of user PostScript
-2.2857143 1.1428571 -2.2437867 1.1462343 MAarrowhead1
% End of user PostScript
.11905 .60332 m
.12674 .60401 L
s
% Start of user PostScript
-2.2857143 1.4285714 -2.2395652 1.4319486 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.2857143 1.7142857 -2.2353438 1.7176629 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-2.2857143 2. -2.2311223 2.0033772 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.7142857 -2. -1.744353 -1.9974671 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.7142857 -1.7142857 -1.7401315 -1.7117528 MAarrowhead1
% End of user PostScript
.21429 .01472 m
.21068 .01524 L
s
% Start of user PostScript
-1.7142857 -1.4285714 -1.73591 -1.4260385 MAarrowhead1
% End of user PostScript
.21429 .07358 m
.21139 .0741 L
s
% Start of user PostScript
-1.7142857 -1.1428571 -1.7316885 -1.1403243 MAarrowhead1
% End of user PostScript
.21429 .13244 m
.21209 .13296 L
s
% Start of user PostScript
-1.7142857 -0.85714286 -1.7274671 -0.85460997 MAarrowhead1
% End of user PostScript
.21429 .1913 m
.21279 .19182 L
s
% Start of user PostScript
-1.7142857 -0.57142857 -1.7232456 -0.56889569 MAarrowhead1
% End of user PostScript
.21429 .25016 m
.2135 .25068 L
s
% Start of user PostScript
-1.7142857 -0.28571429 -1.7190241 -0.2831814 MAarrowhead1
% End of user PostScript
.21429 .30902 m
.2142 .30954 L
s
% Start of user PostScript
-1.7142857 0. -1.7148026 0.0025328851 MAarrowhead1
% End of user PostScript
.21429 .36788 m
.2149 .3684 L
s
% Start of user PostScript
-1.7142857 0.28571429 -1.7105812 0.28824717 MAarrowhead1
% End of user PostScript
.21429 .42674 m
.21561 .42726 L
s
% Start of user PostScript
-1.7142857 0.57142857 -1.7063597 0.57396146 MAarrowhead1
% End of user PostScript
.21429 .4856 m
.21631 .48612 L
s
% Start of user PostScript
-1.7142857 0.85714286 -1.7021382 0.85967574 MAarrowhead1
% End of user PostScript
.21429 .54446 m
.21701 .54498 L
s
% Start of user PostScript
-1.7142857 1.1428571 -1.6979167 1.14539 MAarrowhead1
% End of user PostScript
.21429 .60332 m
.21772 .60384 L
s
% Start of user PostScript
-1.7142857 1.4285714 -1.6936953 1.4311043 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.7142857 1.7142857 -1.6894738 1.7168186 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.7142857 2. -1.6852523 2.0025329 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.1428571 -2. -1.1819417 -1.9983114 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.1428571 -1.7142857 -1.1777202 -1.7125971 MAarrowhead1
% End of user PostScript
.30952 .01472 m
.30442 .01506 L
s
% Start of user PostScript
-1.1428571 -1.4285714 -1.1734987 -1.4268828 MAarrowhead1
% End of user PostScript
.30952 .07358 m
.30512 .07392 L
s
% Start of user PostScript
-1.1428571 -1.1428571 -1.1692773 -1.1411686 MAarrowhead1
% End of user PostScript
.30952 .13244 m
.30582 .13278 L
s
% Start of user PostScript
-1.1428571 -0.85714286 -1.1650558 -0.85545427 MAarrowhead1
% End of user PostScript
.30952 .1913 m
.30653 .19164 L
s
% Start of user PostScript
-1.1428571 -0.57142857 -1.1608343 -0.56973998 MAarrowhead1
% End of user PostScript
.30952 .25016 m
.30723 .2505 L
s
% Start of user PostScript
-1.1428571 -0.28571429 -1.1566128 -0.2840257 MAarrowhead1
% End of user PostScript
.30952 .30902 m
.30793 .30936 L
s
% Start of user PostScript
-1.1428571 0. -1.1523914 0.0016885901 MAarrowhead1
% End of user PostScript
.30952 .36788 m
.30864 .36823 L
s
% Start of user PostScript
-1.1428571 0.28571429 -1.1481699 0.28740288 MAarrowhead1
% End of user PostScript
.30952 .42674 m
.30934 .42709 L
s
% Start of user PostScript
-1.1428571 0.57142857 -1.1439484 0.57311716 MAarrowhead1
% End of user PostScript
.30952 .4856 m
.31005 .48595 L
s
% Start of user PostScript
-1.1428571 0.85714286 -1.1397269 0.85883145 MAarrowhead1
% End of user PostScript
.30952 .54446 m
.31075 .54481 L
s
% Start of user PostScript
-1.1428571 1.1428571 -1.1355055 1.1445457 MAarrowhead1
% End of user PostScript
.30952 .60332 m
.31145 .60367 L
s
% Start of user PostScript
-1.1428571 1.4285714 -1.131284 1.43026 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.1428571 1.7142857 -1.1270625 1.7159743 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-1.1428571 2. -1.122841 2.0016886 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-0.57142857 -2. -0.60850289 -1.9991557 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-0.57142857 -1.7142857 -0.60428141 -1.7134414 MAarrowhead1
% End of user PostScript
.40476 .01472 m
.39999 .01489 L
s
% Start of user PostScript
-0.57142857 -1.4285714 -0.60005994 -1.4277271 MAarrowhead1
% End of user PostScript
.40476 .07358 m
.40069 .07375 L
s
% Start of user PostScript
-0.57142857 -1.1428571 -0.59583846 -1.1420128 MAarrowhead1
% End of user PostScript
.40476 .13244 m
.4014 .13261 L
s
% Start of user PostScript
-0.57142857 -0.85714286 -0.59161699 -0.85629856 MAarrowhead1
% End of user PostScript
.40476 .1913 m
.4021 .19147 L
s
% Start of user PostScript
-0.57142857 -0.57142857 -0.58739551 -0.57058428 MAarrowhead1
% End of user PostScript
.40476 .25016 m
.4028 .25033 L
s
% Start of user PostScript
-0.57142857 -0.28571429 -0.58317404 -0.28486999 MAarrowhead1
% End of user PostScript
.40476 .30902 m
.40351 .30919 L
s
% Start of user PostScript
-0.57142857 0. -0.57895256 0.00084429504 MAarrowhead1
% End of user PostScript
.40476 .36788 m
.40421 .36805 L
s
% Start of user PostScript
-0.57142857 0.28571429 -0.57473109 0.28655858 MAarrowhead1
% End of user PostScript
.40476 .42674 m
.40492 .42691 L
s
% Start of user PostScript
-0.57142857 0.57142857 -0.57050961 0.57227287 MAarrowhead1
% End of user PostScript
.40476 .4856 m
.40562 .48577 L
s
% Start of user PostScript
-0.57142857 0.85714286 -0.56628814 0.85798715 MAarrowhead1
% End of user PostScript
.40476 .54446 m
.40632 .54463 L
s
% Start of user PostScript
-0.57142857 1.1428571 -0.56206666 1.1437014 MAarrowhead1
% End of user PostScript
.40476 .60332 m
.40703 .60349 L
s
% Start of user PostScript
-0.57142857 1.4285714 -0.55784519 1.4294157 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-0.57142857 1.7142857 -0.55362371 1.71513 MAarrowhead1
% End of user PostScript
% Start of user PostScript
-0.57142857 2. -0.54940223 2.0008443 MAarrowhead1
% End of user PostScript
% Start of user PostScript
0. -2. -0.029550326 -2. MAarrowhead1
% End of user PostScript
% Start of user PostScript
0. -1.7142857 -0.025328851 -1.7142857 MAarrowhead1
% End of user PostScript
.5 .01472 m
.49648 .01472 L
s
% Start of user PostScript
0. -1.4285714 -0.021107376 -1.4285714 MAarrowhead1
% End of user PostScript
.5 .07358 m
.49719 .07358 L
s
% Start of user PostScript
0. -1.1428571 -0.016885901 -1.1428571 MAarrowhead1
% End of user PostScript
.5 .13244 m
.49789 .13244 L
s
% Start of user PostScript
0. -0.85714286 -0.012664426 -0.85714286 MAarrowhead1
% End of user PostScript
.5 .1913 m
.49859 .1913 L
s
% Start of user PostScript
0. -0.57142857 -0.0084429504 -0.57142857 MAarrowhead1
% End of user PostScript
.5 .25016 m
.4993 .25016 L
s
% Start of user PostScript
0. -0.28571429 -0.0042214752 -0.28571429 MAarrowhead1
% End of user PostScript
.5 .30902 m
.5 .30902 L
s
% Start of user PostScript
0. 0. MAarrowzero1
% End of user PostScript
.5 .36788 m
.5007 .36788 L
s
% Start of user PostScript
0. 0.28571429 0.0042214752 0.28571429 MAarrowhead1
% End of user PostScript
.5 .42674 m
.50141 .42674 L
s
% Start of user PostScript
0. 0.57142857 0.0084429504 0.57142857 MAarrowhead1
% End of user PostScript
.5 .4856 m
.50211 .4856 L
s
% Start of user PostScript
0. 0.85714286 0.012664426 0.85714286 MAarrowhead1
% End of user PostScript
.5 .54446 m
.50281 .54446 L
s
% Start of user PostScript
0. 1.1428571 0.016885901 1.1428571 MAarrowhead1
% End of user PostScript
.5 .60332 m
.50352 .60332 L
s
% Start of user PostScript
0. 1.4285714 0.021107376 1.4285714 MAarrowhead1
% End of user PostScript
% Start of user PostScript
0. 1.7142857 0.025328851 1.7142857 MAarrowhead1
% End of user PostScript
% Start of user PostScript
0. 2. 0.029550326 2. MAarrowhead1
% End of user PostScript
% Start of user PostScript
0.57142857 -2. 0.54940223 -2.0008443 MAarrowhead1
% End of user PostScript
% Start of user PostScript
0.57142857 -1.7142857 0.55362371 -1.71513 MAarrowhead1
% End of user PostScript
.59524 .01472 m
.59297 .01454 L
s
% Start of user PostScript
0.57142857 -1.4285714 0.55784519 -1.4294157 MAarrowhead1
% End of user PostScript
.59524 .07358 m
.59368 .0734 L
s
% Start of user PostScript
0.57142857 -1.1428571 0.56206666 -1.1437014 MAarrowhead1
% End of user PostScript
.59524 .13244 m
.59438 .13226 L
s
% Start of user PostScript
0.57142857 -0.85714286 0.56628814 -0.85798715 MAarrowhead1
% End of user PostScript
.59524 .1913 m
.59508 .19112 L
s
% Start of user PostScript
0.57142857 -0.57142857 0.57050961 -0.57227287 MAarrowhead1
% End of user PostScript
.59524 .25016 m
.59579 .24998 L
s
% Start of user PostScript
0.57142857 -0.28571429 0.57473109 -0.28655858 MAarrowhead1
% End of user PostScript
.59524 .30902 m
.59649 .30884 L
s
% Start of user PostScript
0.57142857 0. 0.57895256 -0.00084429504 MAarrowhead1
% End of user PostScript
.59524 .36788 m
.5972 .3677 L
s
% Start of user PostScript
0.57142857 0.28571429 0.58317404 0.28486999 MAarrowhead1
% End of user PostScript
.59524 .42674 m
.5979 .42656 L
s
% Start of user PostScript
0.57142857 0.57142857 0.58739551 0.57058428 MAarrowhead1
% End of user PostScript
.59524 .4856 m
.5986 .48542 L
s
% Start of user PostScript
0.57142857 0.85714286 0.59161699 0.85629856 MAarrowhead1
% End of user PostScript
.59524 .54446 m
.59931 .54428 L
s
% Start of user PostScript
0.57142857 1.1428571 0.59583846 1.1420128 MAarrowhead1
% End of user PostScript
.59524 .60332 m
.60001 .60314 L
s
% Start of user PostScript
0.57142857 1.4285714 0.60005994 1.4277271 MAarrowhead1
% End of user PostScript
% Start of user PostScript
0.57142857 1.7142857 0.60428141 1.7134414 MAarrowhead1
% End of user PostScript
% Start of user PostScript
0.57142857 2. 0.60850289 1.9991557 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.1428571 -2. 1.122841 -2.0016886 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.1428571 -1.7142857 1.1270625 -1.7159743 MAarrowhead1
% End of user PostScript
.69048 .01472 m
.68855 .01437 L
s
% Start of user PostScript
1.1428571 -1.4285714 1.131284 -1.43026 MAarrowhead1
% End of user PostScript
.69048 .07358 m
.68925 .07323 L
s
% Start of user PostScript
1.1428571 -1.1428571 1.1355055 -1.1445457 MAarrowhead1
% End of user PostScript
.69048 .13244 m
.68995 .13209 L
s
% Start of user PostScript
1.1428571 -0.85714286 1.1397269 -0.85883145 MAarrowhead1
% End of user PostScript
.69048 .1913 m
.69066 .19095 L
s
% Start of user PostScript
1.1428571 -0.57142857 1.1439484 -0.57311716 MAarrowhead1
% End of user PostScript
.69048 .25016 m
.69136 .24981 L
s
% Start of user PostScript
1.1428571 -0.28571429 1.1481699 -0.28740288 MAarrowhead1
% End of user PostScript
.69048 .30902 m
.69207 .30867 L
s
% Start of user PostScript
1.1428571 0. 1.1523914 -0.0016885901 MAarrowhead1
% End of user PostScript
.69048 .36788 m
.69277 .36753 L
s
% Start of user PostScript
1.1428571 0.28571429 1.1566128 0.2840257 MAarrowhead1
% End of user PostScript
.69048 .42674 m
.69347 .42639 L
s
% Start of user PostScript
1.1428571 0.57142857 1.1608343 0.56973998 MAarrowhead1
% End of user PostScript
.69048 .4856 m
.69418 .48525 L
s
% Start of user PostScript
1.1428571 0.85714286 1.1650558 0.85545427 MAarrowhead1
% End of user PostScript
.69048 .54446 m
.69488 .54411 L
s
% Start of user PostScript
1.1428571 1.1428571 1.1692773 1.1411686 MAarrowhead1
% End of user PostScript
.69048 .60332 m
.69558 .60297 L
s
% Start of user PostScript
1.1428571 1.4285714 1.1734987 1.4268828 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.1428571 1.7142857 1.1777202 1.7125971 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.1428571 2. 1.1819417 1.9983114 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.7142857 -2. 1.6852523 -2.0025329 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.7142857 -1.7142857 1.6894738 -1.7168186 MAarrowhead1
% End of user PostScript
.78571 .01472 m
.78228 .01419 L
s
% Start of user PostScript
1.7142857 -1.4285714 1.6936953 -1.4311043 MAarrowhead1
% End of user PostScript
.78571 .07358 m
.78299 .07305 L
s
% Start of user PostScript
1.7142857 -1.1428571 1.6979167 -1.14539 MAarrowhead1
% End of user PostScript
.78571 .13244 m
.78369 .13191 L
s
% Start of user PostScript
1.7142857 -0.85714286 1.7021382 -0.85967574 MAarrowhead1
% End of user PostScript
.78571 .1913 m
.78439 .19077 L
s
% Start of user PostScript
1.7142857 -0.57142857 1.7063597 -0.57396146 MAarrowhead1
% End of user PostScript
.78571 .25016 m
.7851 .24963 L
s
% Start of user PostScript
1.7142857 -0.28571429 1.7105812 -0.28824717 MAarrowhead1
% End of user PostScript
.78571 .30902 m
.7858 .3085 L
s
% Start of user PostScript
1.7142857 0. 1.7148026 -0.0025328851 MAarrowhead1
% End of user PostScript
.78571 .36788 m
.7865 .36736 L
s
% Start of user PostScript
1.7142857 0.28571429 1.7190241 0.2831814 MAarrowhead1
% End of user PostScript
.78571 .42674 m
.78721 .42622 L
s
% Start of user PostScript
1.7142857 0.57142857 1.7232456 0.56889569 MAarrowhead1
% End of user PostScript
.78571 .4856 m
.78791 .48508 L
s
% Start of user PostScript
1.7142857 0.85714286 1.7274671 0.85460997 MAarrowhead1
% End of user PostScript
.78571 .54446 m
.78861 .54394 L
s
% Start of user PostScript
1.7142857 1.1428571 1.7316885 1.1403243 MAarrowhead1
% End of user PostScript
.78571 .60332 m
.78932 .6028 L
s
% Start of user PostScript
1.7142857 1.4285714 1.73591 1.4260385 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.7142857 1.7142857 1.7401315 1.7117528 MAarrowhead1
% End of user PostScript
% Start of user PostScript
1.7142857 2. 1.744353 1.9974671 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.2857143 -2. 2.2311223 -2.0033772 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.2857143 -1.7142857 2.2353438 -1.7176629 MAarrowhead1
% End of user PostScript
.88095 .01472 m
.87326 .01402 L
s
% Start of user PostScript
2.2857143 -1.4285714 2.2395652 -1.4319486 MAarrowhead1
% End of user PostScript
.88095 .07358 m
.87396 .07288 L
s
% Start of user PostScript
2.2857143 -1.1428571 2.2437867 -1.1462343 MAarrowhead1
% End of user PostScript
.88095 .13244 m
.87467 .13174 L
s
% Start of user PostScript
2.2857143 -0.85714286 2.2480082 -0.86052004 MAarrowhead1
% End of user PostScript
.88095 .1913 m
.87537 .1906 L
s
% Start of user PostScript
2.2857143 -0.57142857 2.2522297 -0.57480575 MAarrowhead1
% End of user PostScript
.88095 .25016 m
.87608 .24946 L
s
% Start of user PostScript
2.2857143 -0.28571429 2.2564511 -0.28909147 MAarrowhead1
% End of user PostScript
.88095 .30902 m
.87678 .30832 L
s
% Start of user PostScript
2.2857143 0. 2.2606726 -0.0033771801 MAarrowhead1
% End of user PostScript
.88095 .36788 m
.87748 .36718 L
s
% Start of user PostScript
2.2857143 0.28571429 2.2648941 0.28233711 MAarrowhead1
% End of user PostScript
.88095 .42674 m
.87819 .42604 L
s
% Start of user PostScript
2.2857143 0.57142857 2.2691156 0.56805139 MAarrowhead1
% End of user PostScript
.88095 .4856 m
.87889 .4849 L
s
% Start of user PostScript
2.2857143 0.85714286 2.273337 0.85376568 MAarrowhead1
% End of user PostScript
.88095 .54446 m
.87959 .54376 L
s
% Start of user PostScript
2.2857143 1.1428571 2.2775585 1.13948 MAarrowhead1
% End of user PostScript
.88095 .60332 m
.8803 .60262 L
s
% Start of user PostScript
2.2857143 1.4285714 2.28178 1.4251942 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.2857143 1.7142857 2.2860015 1.7109085 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.2857143 2. 2.2902229 1.9966228 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.8571429 -2. 2.7549372 -2.0042215 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.8571429 -1.7142857 2.7591587 -1.7185072 MAarrowhead1
% End of user PostScript
.97619 .01472 m
.96056 .01385 L
s
% Start of user PostScript
2.8571429 -1.4285714 2.7633802 -1.4327929 MAarrowhead1
% End of user PostScript
.97619 .07358 m
.96127 .07271 L
s
% Start of user PostScript
2.8571429 -1.1428571 2.7676016 -1.1470786 MAarrowhead1
% End of user PostScript
.97619 .13244 m
.96197 .13157 L
s
% Start of user PostScript
2.8571429 -0.85714286 2.7718231 -0.86136433 MAarrowhead1
% End of user PostScript
.97619 .1913 m
.96267 .19043 L
s
% Start of user PostScript
2.8571429 -0.57142857 2.7760446 -0.57565005 MAarrowhead1
% End of user PostScript
.97619 .25016 m
.96338 .24929 L
s
% Start of user PostScript
2.8571429 -0.28571429 2.7802661 -0.28993576 MAarrowhead1
% End of user PostScript
.97619 .30902 m
.96408 .30815 L
s
% Start of user PostScript
2.8571429 0. 2.7844875 -0.0042214752 MAarrowhead1
% End of user PostScript
.97619 .36788 m
.96478 .36701 L
s
% Start of user PostScript
2.8571429 0.28571429 2.788709 0.28149281 MAarrowhead1
% End of user PostScript
.97619 .42674 m
.96549 .42587 L
s
% Start of user PostScript
2.8571429 0.57142857 2.7929305 0.5672071 MAarrowhead1
% End of user PostScript
.97619 .4856 m
.96619 .48473 L
s
% Start of user PostScript
2.8571429 0.85714286 2.797152 0.85292138 MAarrowhead1
% End of user PostScript
.97619 .54446 m
.9669 .54359 L
s
% Start of user PostScript
2.8571429 1.1428571 2.8013734 1.1386357 MAarrowhead1
% End of user PostScript
.97619 .60332 m
.9676 .60245 L
s
% Start of user PostScript
2.8571429 1.4285714 2.8055949 1.42435 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.8571429 1.7142857 2.8098164 1.7100642 MAarrowhead1
% End of user PostScript
% Start of user PostScript
2.8571429 2. 2.8140379 1.9957785 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -2. 3.2511833 -2.0050658 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -1.7142857 3.2554048 -1.7193515 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -1.4285714 3.2596263 -1.4336372 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -1.1428571 3.2638477 -1.1479229 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -0.85714286 3.2680692 -0.86220863 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -0.57142857 3.2722907 -0.57649434 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 -0.28571429 3.2765122 -0.29078006 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 0. 3.2807336 -0.0050657702 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 0.28571429 3.2849551 0.28064852 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 0.57142857 3.2891766 0.5663628 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 0.85714286 3.2933981 0.85207709 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 1.1428571 3.2976195 1.1377914 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 1.4285714 3.301841 1.4235057 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 1.7142857 3.3060625 1.7092199 MAarrowhead1
% End of user PostScript
% Start of user PostScript
3.4285714 2. 3.310284 1.9949342 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -2. 3.7143468 -2.0059101 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -1.7142857 3.7185683 -1.7201958 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -1.4285714 3.7227898 -1.4344815 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -1.1428571 3.7270113 -1.1487672 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -0.85714286 3.7312327 -0.86305292 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -0.57142857 3.7354542 -0.57733864 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. -0.28571429 3.7396757 -0.29162435 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 0. 3.7438972 -0.0059100652 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 0.28571429 3.7481186 0.27980422 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 0.57142857 3.7523401 0.56551851 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 0.85714286 3.7565616 0.85123279 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 1.1428571 3.7607831 1.1369471 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 1.4285714 3.7650045 1.4226614 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 1.7142857 3.769226 1.7083756 MAarrowhead1
% End of user PostScript
% Start of user PostScript
4. 2. 3.7734475 1.9940899 MAarrowhead1
% End of user PostScript
1 0 0 r
.5 Mabswid
.12906 0 m
.16639 .17066 L
.1932 .2599 L
.2223 .33461 L
.25007 .38638 L
.26408 .40586 L
.27951 .42256 L
.28678 .42879 L
.2937 .43378 L
.29998 .43756 L
.30668 .44081 L
.31048 .44231 L
.31393 .44346 L
.31743 .44442 L
.32066 .44514 L
.32443 .44576 L
.32641 .446 L
.32743 .4461 L
.32851 .44618 L
.33034 .44629 L
.33204 .44635 L
.33308 .44636 L
.33406 .44635 L
.33591 .44631 L
.33751 .44623 L
.33922 .4461 L
.34085 .44595 L
.34234 .44577 L
.34579 .44524 L
.34943 .4445 L
.35628 .44263 L
.36378 .4399 L
.37714 .43337 L
.38957 .42554 L
.41821 .402 L
.44424 .37537 L
.50021 .30876 L
.55406 .24453 L
.58134 .21646 L
.59581 .20364 L
.61135 .19186 L
.62624 .18282 L
.63985 .17672 L
.64639 .1746 L
.64985 .1737 L
.65353 .17292 L
.65706 .17235 L
.66027 .17198 L
.66181 .17185 L
.66349 .17175 L
Mistroke
.66506 .1717 L
.66651 .17168 L
.66829 .1717 L
.66997 .17176 L
.67148 .17185 L
.67311 .17199 L
.6749 .17219 L
.67679 .17245 L
.68018 .17307 L
.68624 .17463 L
.69292 .17706 L
.70012 .18053 L
.70786 .1853 L
.72187 .19677 L
.7353 .21141 L
.74998 .23173 L
.77954 .28749 L
.80705 .35886 L
.8364 .45786 L
.86433 .57598 L
Mfstroke
.86433 .57598 m
.87245 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00_03oool00`3ookl0M000000000020000000704RLg`3oool0oooo0?ooo`3ogi`0B00007Bo
o`020?ooo`060?oOW0180000000000000000B9`0goooh03oool00;h0oooo00<0okmd0000001d_ol0
103oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Boh03oool00;h0
oooo00<0oooO09a80000M;l03P3oool00`3o_g@0B9cO0?ooo`3N0?ooo`00_`3oool00`3OW4P004RL
0=ooo`0=0?ooo`030?oo_`1d07@0_ooo0=h0oooo002_0?ooo`030?oog`2LB000000000L0000000<0
B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`0=0?ooo`030?oog`2LB000M;oo0=h0oooo0011
0?ooo`030000003oool0oooo00D0oooo00<0o`000?ooo`3oool0L`3oool00`3ookl0M00007Boo`0=
0?ooo`030?oo_`1d0000B9cO00P0oooo1000003B0?ooo`00?`3oool4000000H0oooo00<0o`000?oo
o`3oool0M03oool00`2oM00007Bo0?ooo`0<0?ooo`030?noM0000000M;oo00P0oooo00<000000?oo
o`3oool0d`3oool003h0oooo1@0000060?ooo`030?l0003oool0oooo07@0oooo00<0gia80018W03O
ool02@3oool01P3ogml0gooo0?noM0000000001d0;ooo`P0oooo00<000000?ooo`3oool0d`3oool0
03l0oooo1@0000060?ooo`030?l0003oool0oooo0280oooo0P00001?0?ooo`030?oOW01804P0W=oo
00T0oooo00D0oono07@000000000001809cOo`090?ooo`030000003oool0oooo0=<0oooo000K0?oo
o`800000903oool4000000D0oooo00<0o`000?ooo`3oool0803oool4000002<0oooo0P00000T0?oo
o`800000103oool00`3ookl0M00007Boo`0:0?ooo`030=nLW02oool0oooo00X0oooo00<000000?oo
o`3oool00`3oool200000;@0oooo0P00000H0?ooo`006`3oool4000002@0oooo0P0000050?ooo`03
0?l0003oool0oooo01h0oooo1P00000O0?ooo`H00000803oool6000000D0oooo00<0_g@0001d_`3o
ool02@3oool00`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0000000D000008P3oool400000240
oooo1000000Q0?ooo`@00000803oool4000001`0oooo1P00000H0?ooo`006`3oool900000240oooo
00<000000?ooo`3oool00P3oool00`3o0000oooo0?ooo`0L0?ooo`T00000703oool8000001d0oooo
2@0000030?ooo`050?oo_`1d000000000000M02oool02P3oool01@2oM0000000000000000000B9cO
00H0oooo2@00000M0?ooo`T00000703oool9000001`0oooo2@00000K0?ooo`T000006@3oool90000
01P0oooo000K0?ooo`T000009P3oool00`3o0000oooo0?ooo`0P0?ooo`D000007P3oool7000001h0
oooo1`00000N0?ooo`P000007`3oool6000001l0oooo1`00000N0?ooo`H000007P3oool6000001`0
oooo1`00000H0?ooo`006`3oool5000002/0oooo00<0o`000?ooo`3oool0AP3oool300000280oooo
0`00000N0?ooo`<000000P3oool300000240oooo1000000Q0?ooo`D00000803oool400000200oooo
1000000P0?ooo`<00000603oool004/0oooo00<0o`000?ooo`3oool0S03oool00`000000oooo0?oo
o`0X0?ooo`800000903oool200000280oooo0P00000R0?ooo`800000>`3oool004/0oooo00<0o`00
0?ooo`3oool0S03oool00`000000oooo0?ooo`3C0?ooo`00B`3oool00`3o0000oooo0?ooo`2<0?oo
o`030000003oool0oooo0=<0oooo001<0?ooo`030?l0003oool0oooo08/0oooo00<000000?ooo`3o
ool0d`3oool004`0oooo00<0o`000?ooo`3oool0R`3oool00`000000oooo0?ooo`3C0?ooo`00C03o
ool00`3o0000oooo0?ooo`2;0?ooo`030000003oool0oooo0=<0oooo001<0?ooo`030?l0003oool0
oooo08/0oooo00<000000?ooo`3oool0d`3oool004`0oooo00<0o`000?ooo`3oool0R`3oool00`00
0000oooo0?ooo`3C0?ooo`00C@3oool00`3o0000oooo0?ooo`2:0?ooo`<00000d`3oool004d0oooo
00<0o`000?ooo`3oool0RP3oool00`000000oooo0?ooo`3C0?ooo`00C@3oool00`3o0000oooo0?oo
o`2:0?ooo`030000003oool0oooo0=<0oooo001=0?ooo`030?l0003oool0oooo08X0oooo00<00000
0?ooo`3oool0d`3oool004d0oooo00<0o`000?ooo`3oool0RP3oool00`000000oooo0?ooo`3C0?oo
o`00CP3oool00`3o0000oooo0?ooo`290?ooo`030000003oool0oooo0=<0oooo001>0?ooo`030?l0
003oool0oooo08T0oooo00<000000?ooo`3oool0d`3oool00400oooo00<000000?ooo`3oool02`3o
ool00`3o0000oooo0?ooo`290?ooo`030000003oool0oooo0=<0oooo000o0?ooo`<00000303oool0
0`3o0000oooo0?ooo`290?ooo`<00000d`3oool003l0oooo1000000;0?ooo`030?l0003oool0oooo
01h0oooo0P00001Y0?ooo`030000003oool0oooo0=<0oooo000L0?ooo`800000803oool5000000`0
oooo00<0o`000?ooo`3oool06`3oool4000002<0oooo0P00000T0?ooo`8000007P3oool00`000000
oooo0?ooo`030?ooo`800000]03oool2000001P0oooo000L0?ooo`@00000803oool4000000/0oooo
00<0o`000?ooo`3oool06@3oool6000001l0oooo1P00000P0?ooo`H000007P3oool00`000000oooo
0000000500000280oooo100000160?ooo`@00000803oool4000001`0oooo1P00000H0?ooo`006`3o
ool700000200oooo100000090?ooo`030?l0003oool0oooo01L0oooo2@00000L0?ooo`P000007@3o
ool9000001d0oooo2@00000M0?ooo`T00000703oool:000001/0oooo2@00000K0?ooo`T000006@3o
ool9000001P0oooo000K0?ooo`T00000803oool2000000T0oooo00<0o`000?ooo`3oool06`3oool5
000001h0oooo1`00000N0?ooo`L000007P3oool8000001l0oooo1`00000M0?ooo`T000007@3oool7
000001d0oooo1`00000K0?ooo`L00000603oool001/0oooo1@00000_0?ooo`030?l0003oool0oooo
0480oooo0`00000R0?ooo`<000007P3oool00`000000oooo0?ooo`020?ooo`<000008@3oool50000
01l0oooo1P00000P0?ooo`D000007`3oool5000001l0oooo0`00000H0?ooo`00D03oool00`3o0000
oooo0?ooo`270?ooo`030000003oool0oooo02T0oooo0P00000Q0?ooo`@000008`3oool200000280
oooo0P00000j0?ooo`00D03oool00`3o0000oooo0?ooo`270?ooo`<00000CP3oool2000008<0oooo
001@0?ooo`030?l0003oool0oooo08L0oooo00<000000?ooo`3oool0d`3oool00500oooo00<0o`00
0?ooo`3oool0Q`3oool00`000000oooo0?ooo`3C0?ooo`00D@3oool00`3o0000oooo0?ooo`1e0?oo
o`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool02@3oool00`000000oooo0?ooo`3C0?oo
o`00D@3oool00`3o0000oooo0?ooo`1g0?ooo`030?noM0000000M;oo00`0oooo00<000000?ooo`3o
ool0d`3oool00540oooo00<0o`000?ooo`3oool0KP3oool/00000<80oooo001A0?ooo`030?l0003o
ool0oooo05H0oooo8P0000000`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo00T0oooo5`00
002c0?ooo`00D@3oool00`3o0000oooo0?ooo`170?oooaT00000203oool00`3ooml0W4P000000007
0000000304RLg`3oool0oooo00<0oooo00<0omnL04P0B02Lgol02`3oool00`000000oooo0?ooo`0H
0?ooo`h00000[@3oool00580oooo00<0o`000?ooo`3oool0>@3ooolF000002P0oooo00<0oono07@0
001d_ol02`3oool400000200oooo2`00002W0?ooo`00DP3oool00`3o0000oooo0?ooo`0_0?oooa40
0000>03oool00`2oM00007Bo0?ooo`0:0?ooo`030000003oool0oooo02T0oooo1P00002T0?ooo`00
DP3oool00`3o0000oooo0?ooo`0X0?ooo`h00000@P3oool00`3OW4P004RL0=ooo`0:0?ooo`030000
003oool0oooo02`0oooo1`00002P0?ooo`00DP3oool00`3o0000oooo0?ooo`0O0?ooo`d00000C03o
ool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo0340oooo1P00002L0?ooo`00DP3oool0
0`3o0000oooo0?ooo`0J0?ooo`X00000E03oool00`3ookl0M00007Boo`0:0?ooo`030000003oool0
oooo03@0oooo1@00002J0?ooo`00D`3oool00`3o0000oooo0?ooo`0E0?ooo`L00000G03oool00`2o
M00007Bo0?ooo`090?ooo`030000003oool0oooo03L0oooo1000002H0?ooo`00D`3oool00`3o0000
oooo0?ooo`0B0?ooo`D00000G`3oool01@3ookl0M0000000000007@0_ooo00T0oooo00<000000?oo
o`3oool0>P3oool3000009H0oooo000o0?ooo`030000003oool0oooo0140oooo00<0o`000?ooo`3o
ool03P3oool500000740oooo00<000000?ooo`3oool0?03oool3000009@0oooo000o0?ooo`<00000
4@3oool00`3o0000oooo0?ooo`0:0?ooo`D00000303oool00`000000oooo0?ooo`1V0?ooo`<00000
?P3oool300000980oooo000L0?ooo`800000803oool500000100oooo00<0o`000?ooo`3oool0203o
ool4000000d0oooo0`00000S0?ooo`8000008`3oool2000001h0oooo00<000000?ooo`3oool00`3o
ool2000003X0oooo1000002@0?ooo`00703oool4000001h0oooo1P00000@0?ooo`030?l0003oool0
oooo00H0oooo0`00000=0?ooo`D000007`3oool6000001l0oooo1P00000N0?ooo`030000003oool0
000000D00000?03oool300000540oooo1000000N0?ooo`@00000603oool001/0oooo1`00000L0?oo
o`P000003P3oool00`3o0000oooo0?ooo`040?ooo`<00000303oool8000001`0oooo2@00000M0?oo
o`P000007@3oool9000001d0oooo2@00000H0?ooo`<000000P3oool2000002<0oooo1000000O0?oo
o`T000006@3oool9000001P0oooo000K0?ooo`T000007P3oool5000000d0oooo00<0o`000?ooo`3o
ool00`3oool2000000d0oooo2@00000N0?ooo`L000007`3oool6000001h0oooo2000000N0?ooo`P0
00006P3oool3000000040?ooo`000000000000000280oooo2000000L0?ooo`L000006`3oool70000
01P0oooo000K0?ooo`D00000=03oool00`3o0000oooo0?ooo`020?ooo`800000>@3oool3000002<0
oooo0P00000N0?ooo`030000003oool0oooo0080oooo0`00000P0?ooo`H00000703oool200000003
0?ooo`0000000000008000008P3oool5000001l0oooo1@00000O0?ooo`<00000603oool005@0oooo
00@0o`000?ooo`3oool0oooo0P0000200?ooo`030000003oool0oooo02L0oooo1000000M0?ooo`P0
00008@3oool400000280oooo0P00000j0?ooo`00E@3oool0103o0000oooo0000000000210?ooo`03
0000003oool0oooo02T0oooo0P00000N0?ooo`H00000903oool00`000000oooo0?ooo`1M0?ooo`00
E@3oool00`3o0000oooo000000220?ooo`<00000B`3oool4000008@0oooo001E0?ooo`030?l00000
000000000880oooo00<000000?ooo`3oool0C03oool2000008D0oooo001E0?ooo`030?l000000000
00000880oooo00<000000?ooo`3oool0C@3oool2000008@0oooo001F0?ooo`030?l0003oool0oooo
0840oooo00<000000?ooo`3oool0CP3oool2000008<0oooo001F0?ooo`030?l0003oool0oooo0840
oooo00<000000?ooo`3oool0C`3oool200000880oooo001F0?ooo`030?l0003oool0oooo0840oooo
00<000000?ooo`3oool0D03oool200000840oooo001F0?ooo`030?l0003oool0oooo0840oooo00<0
00000?ooo`3oool0D@3oool200000800oooo001F0?ooo`030?l0003oool0oooo0840oooo00<00000
0?ooo`3oool0DP3oool2000007l0oooo001F0?ooo`030000003o0000oooo0840oooo0`00001C0?oo
o`800000OP3oool005H0oooo00<000000?l0003oool0P@3oool00`000000oooo0?ooo`1C0?ooo`80
0000OP3oool005H0oooo00<000000?l0003oool0P@3oool00`000000oooo0?ooo`1D0?ooo`800000
O@3oool005H0oooo00<000000?l0003oool0P@3oool00`000000oooo0?ooo`0i0?ooo``0o`00403o
ool2000007`0oooo001G0?ooo`030000003o0000oooo0800oooo00<000000?ooo`3oool0=P3oool3
0?l000`0oooo0`3o000=0?ooo`<00000N`3oool005L0oooo00<000000?l0003oool0P03oool00`00
0000oooo0?ooo`0c0?ooo`<0o`004P3oool30?l000/0oooo0`00001j0?ooo`00E`3oool00`000000
o`000?ooo`0F0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0<@3oool20?l001P0
oooo0P3o000:0?ooo`800000NP3oool003l0oooo00<000000?ooo`3oool05@3oool2000000030?l0
003oool0oooo01<0oooo0P00001Y0?ooo`030000003oool0oooo02l0oooo0P3o000L0?ooo`80o`00
203oool3000007T0oooo000L0?ooo`8000008@3oool3000001D0oooo0P0000000`3o0000oooo0?oo
o`0B0?ooo`@000008`3oool2000002@0oooo0P00000M0?ooo`<00000103oool2000002L0oooo0P3o
000P0?ooo`030?l0003oool0oooo00H0oooo0`00001h0?ooo`00703oool4000001h0oooo1P00000C
0?ooo`80000000<0o`000?ooo`3oool0403oool600000240oooo1000000P0?ooo`H000007@3oool0
0`000000oooo0?ooo`06000002H0oooo00<0o`000?ooo`3oool08@3oool00`3o0000oooo0?ooo`06
0?ooo`800000>P3oool4000001l0oooo1000000G0?ooo`00703oool7000001/0oooo2000000B0?oo
o`80000000<0o`000?ooo`3oool03@3oool9000001d0oooo1`00000M0?ooo`T000007@3oool90000
01d0oooo00<000000?ooo`3oool0103oool20?l001`0oooo00<000000?ooo`3oool01P3oool00`3o
0000oooo0?ooo`050?ooo`<000004@3oool4000001l0oooo2@00000J0?ooo`T000005`3oool001`0
oooo2@00000I0?ooo`T000004@3oool2000000030?l0003oool0oooo00d0oooo1@00000O0?ooo`T0
00007`3oool7000000050?ooo`2oM00000000000001d_ol00`3oool00`3ogi`0B00007Boo`020?oo
o`060?oOW0180000000000000000B9`0gooo2P3oool00`000000oooo00000006000001h0oooo00<0
00000?ooo`3oool00P3oool00`3o0000oooo0?ooo`0K0?ooo`8000002@3oool00`3o0000oooo0?oo
o`050?ooo`<000004@3oool8000001`0oooo1`00000L0?ooo`H00000603oool001`0oooo1@00000g
0?ooo`80000000<0o`000?ooo`3oool0=@3oool5000002<0oooo0`0000001P3OW4P007Bo0?ooo`3o
gi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0omnL04Qd
_`X0oooo00<000000?ooo`3oool00`3oool3000001h0oooo0P0000000`3oool0o`000?l0000M0?oo
o`8000002`3oool00`3o0000oooo0?ooo`040?ooo`<000004P3oool6000001l0oooo1@00000N0?oo
o`@00000603oool005T0oooo00<000000?l0003oool0H03oool00`3ooml0W4P004RLg`020?ooo`03
0?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool0203oool00`000000oooo0?ooo`0U0?ooo`03
0000003o0000000001d0oooo0`00000<0?ooo`030?l0003oool0oooo00@0oooo0`00000B0?ooo`D0
00008P3oool200000200oooo0P00000H0?ooo`00F@3oool2000000030?l0003oool0oooo05h0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d07@0_ooo00P0
oooo00<000000?ooo`3oool0903oool20?l000<00000703oool3000000d0oooo00<0o`000?ooo`3o
ool00`3oool010000000oooo00000000000C0?ooo`800000G`3oool005T0oooo0P0000000`3o0000
oooo0?ooo`1A0?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0oooO09a80000M;l3
0?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`0S
0?ooo`030?l0003oool0oooo00@000006P3oool4000000h0oooo00<0o`000?ooo`3oool00`3oool3
000001@0oooo00<000000?ooo`3oool0G@3oool005T0oooo0P0000000`3o0000oooo0?ooo`1O0?oo
o`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0M00004RLg`080?oo
o`@000008@3oool0103o0000oooo0?ooo`3oool3000001X0oooo1000000@0?ooo`80o`000`3oool0
10000000oooo00000000001c0?ooo`00FP3oool00`000000oooo0?l0001P0?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`080?ooo`030000003oool0
oooo0240oooo00<0o`000?ooo`3oool00`3oool00`000000oooo0?ooo`0J0?ooo`<000004P3oool0
103o0000oooo0?ooo`3oool3000007<0oooo001J0?ooo`80000000<0o`000?ooo`3oool0GP3oool0
0`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@000000000
M02oool80?ooo`030000003oool0oooo01l0oooo0P3o000U0?ooo`030000003oool0oooo0140oooo
00D0o`000?ooo`3oool000000?ooo`0200000780oooo001J0?ooo`80000000<0o`000?ooo`3oool0
GP3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@00000
0000001809cOo`090?ooo`030000003oool0oooo01h0oooo00<0o`000?ooo`3oool0>P3oool01@3o
0000oooo0?ooo`000000oooo00800000L@3oool005X0oooo0P0000000`3oool0o`000?ooo`1O0?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?ooo`0:0?oo
o`030000003oool0oooo01d0oooo00<0o`000?ooo`3oool0>`3oool01@3o0000oooo0?ooo`000000
oooo00800000L@3oool005/0oooo0P0000000`3o0000oooo0?ooo`1O0?ooo`060=nLB00007@0_ooo
0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0oooo01`0oooo00<0
o`000?ooo`3oool0?@3oool01@3o0000oooo0?ooo`000000oooo00800000L03oool005/0oooo0P00
00000`3o0000oooo0?ooo`1P0?ooo`040=nLB00000000000001d_`X0oooo00D0_g@0000000000000
000004RLg`070?ooo`030000003oool0oooo01/0oooo00<0o`000?ooo`3oool0?P3oool01@3o0000
oooo0?ooo`000000oooo00800000L03oool005/0oooo0P0000000`3oool0o`000?ooo`1j0?ooo`03
0000003oool0oooo01X0oooo00<0o`000?ooo`3oool0@03oool00`3o0000oooo000000020?ooo`80
0000K`3oool005`0oooo00<000000?ooo`3o0000N`3oool3000001T0oooo00<0o`000?ooo`3oool0
@P3oool0103o000000000?ooo`3oool2000006l0oooo001L0?ooo`80000000<0o`000?ooo`3oool0
N@3oool00`000000oooo0?ooo`0H0?ooo`030?l0003oool0oooo04<0oooo00<0o`000000003oool0
0P3oool2000006h0oooo001L0?ooo`80000000<0oooo0?l0003oool0N@3oool00`000000oooo0?oo
o`0G0?ooo`030?l0003oool0oooo04D0oooo00@0o`000?ooo`3oool0oooo0P00001^0?ooo`00G03o
ool2000000030?ooo`3o0000oooo07T0oooo00<000000?ooo`3oool05P3oool00`3o0000oooo0?oo
o`160?ooo`030000003o0000oooo0080oooo0P00001]0?ooo`00G@3oool00`000000oooo0?l0000@
0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool05@3oool00`3o0000oooo0?ooo`17
0?ooo`030000003o0000oooo0080oooo0P00001]0?ooo`00G@3oool2000000030?l0003oool0oooo
00d0oooo0`00001X0?ooo`030000003oool0oooo01@0oooo00<0o`000?ooo`3oool0B03oool00`00
0000oooo0?l000030?ooo`030000003oool0oooo06/0oooo000o0?ooo`030000003oool0oooo01/0
oooo0P0000000`3oool0o`000?ooo`0<0?ooo`@00000J03oool00`000000oooo0?ooo`0C0?ooo`03
0?l0003oool0oooo04T0oooo00<000000?ooo`3o00000`3oool2000006`0oooo000L0?ooo`800000
8@3oool3000001`0oooo00<000000?ooo`3o0000303oool600000280oooo0P00000T0?ooo`800000
7@3oool00`000000oooo0?ooo`0B0?ooo`030?l0003oool0oooo04T0oooo00D000000?ooo`3oool0
oooo0?l000030?ooo`030000003oool0oooo06X0oooo000L0?ooo`@000007P3oool6000001X0oooo
0P0000000`3o0000oooo0?ooo`090?ooo`D000008P3oool400000280oooo1000000M0?ooo`<00000
0P3oool4000000/0oooo00<0o`000?ooo`3oool0BP3oool01@000000oooo0?ooo`3oool0o`0000<0
oooo0P00000^0?ooo`@000007P3oool4000001L0oooo000L0?ooo`L000006`3oool8000001P0oooo
0P0000000`3oool0o`000?ooo`070?ooo`D000008@3oool7000001h0oooo2000000M0?ooo`T00000
2P3oool00`3o0000oooo0?ooo`080?ooo`T000008@3oool4000001@0oooo00<000000?ooo`3oool0
0`3oool01@3o0000oooo0?ooo`3oool0000000D0oooo0P00000R0?ooo`T000006@3oool9000001L0
oooo000L0?ooo`T000006@3oool9000001P0oooo00<000000?ooo`3o0000203oool2000002<0oooo
2@00000L0?ooo`T000007@3oool00`000000oooo00000006000000T0oooo00<0o`000?ooo`3oool0
2@3oool8000001d0oooo2000000E0?ooo`030000003oool0oooo00<0oooo00@0o`000?ooo`3oool0
oooo0P0000050?ooo`<000008P3oool6000001`0oooo1`00000G0?ooo`00603oool00`3ookl0M000
00000006000003T0oooo00<0gia80000000000000`0000000`18W=l0o`000?ooo`0`0?ooo`D00000
0`3oool00`3ookl0M000000000020000000304RLg`3oool0oooo01D0oooo1@00000M0?ooo`030000
003oool0oooo00<0oooo0`0000090?ooo`030?l0003oool0oooo00T0oooo1P00000I0?ooo`030?oo
_`1d000000000080000000<0B9cO0000000000001@00000E0?ooo`030000003oool0oooo00D0oooo
00D0o`000?ooo`3oool0oooo000000050?ooo`D00000203oool00`3OW4P000000000000300000003
04RLg`3oool0oooo0140oooo1000000N0?ooo`D0000000<0oooo0?oo_`1d00000`0000000`1d_ol0
oooo0?ooo`0>0?ooo`00603oool00`3o_g@00000001d_`020?ooo`030;mdB02Lgol0oooo03/0oooo
00@0okmd001d_`3oool0oooo0P0000000`2oool0o`000?ooo`0i0?ooo`030?noM0000000M;oo03X0
oooo00<000000?ooo`3oool03P3oool00`3o0000oooo0?ooo`0;0?ooo`<000007@3oool00`3o_g@0
000007Boo`020?ooo`D00000503oool2000000P0oooo00@0o`000?ooo`3oool0oooo0P0000050?oo
o`H000001`3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo01<0oooo0P00000Q0?oo
o`80000000@0oooo0?noM000000007Bo0P3oool00`2oM4P0W=oo0?ooo`0>0?ooo`007@3oool00`3o
gi`0B4RL0=ooo`0l0?ooo`030?noM000M;l0oooo0080000000<0oooo0?l0003oool0>@3oool00`3o
oml0W4P0001d_`0j0?ooo`030000003oool0oooo00d0oooo00<0o`000?ooo`3oool0303oool00`00
0000oooo0?ooo`0M0?ooo`030?oog`2LB00007Bo0080oooo0`00000E0?ooo`030000003oool0oooo
00T0oooo00D0o`000?ooo`3oool0oooo000000060?ooo`@000002@3oool00`3o_g@007Bo0?ooo`12
0?ooo`030?oOW018B9`0gooo00h0oooo000M0?ooo`030?oo_`1d0000M;oo03d0oooo00H0okmd001d
_`3oool000000?ooo`3o000k0?ooo`030=nLB000B9`0gooo03T0oooo00<000000?ooo`3oool0303o
ool00`3o0000oooo0?ooo`0^0?ooo`030=nLB000B9`0gooo0080oooo00<000000?ooo`3oool04P3o
ool2000000d0oooo00<0o`000?ooo`3oool00P0000060?ooo`<000002P3oool00`3o_g@007Bo0?oo
o`110?ooo`030?oo_`1d0000M;oo00h0oooo000;0?ooo`030?oog`2LB000000000L0000000<0B9cO
0?ooo`3oool01@3oool00`3ogi`0B00007Boo`0]0?ooo`030?oog`2LB000000000L0000000<0B9cO
0?ooo`3oool0103oool01P3o_g@0B9cO000000000000oooo0?l002X0oooo00<0oooO09a800000000
1`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW01804P0W=oo03T0oooo00<000000?ooo`3oool0
2`3oool00`3o0000oooo0?ooo`0_0?ooo`030?oOW01804P0W=oo01D0oooo0P00000?0?ooo`030?l0
003oool0oooo008000001P3oool2000000`0oooo00<0okmd04RLg`3oool0@03oool00`3ogi`0B000
07Boo`0>0?ooo`00703oool0103ogi`0B0000000B02Lgolo0?ooo`050=nLB000000000000?ooo`3o
0000>P3oool00`3ookl0M00007Boo`0i0?ooo`<000002P3oool00`3o0000oooo0?ooo`0`0?ooo`03
0?oo_`1d0000M;oo01@0oooo00<000000?ooo`3oool0403oool00`3o0000oooo0?ooo`02000000H0
oooo00<000000?ooo`3oool02`3oool00`3OW4P0B9cO0?ooo`0n0?ooo`040?oOW0180000001809cO
o`h0oooo000J0?ooo`050?oOW018000000000000001d_ol0@@3oool0102oM00000000000003o000k
0?ooo`030;md0000M;l0oooo03P0oooo00<000000?ooo`3oool02P3oool00`3o0000oooo0?ooo`0a
0?ooo`030;md0000M;l0oooo0140oooo0P00000C0?ooo`030?l0003oool0oooo008000005@3oool0
0`2oM000M;oo0?ooo`0k0?ooo`050?oOW018000000000000001d_ol03`3oool001`0oooo00<0oono
07@00000M;l0@@3oool01@3ogi`000000000003oool0o`0003X0oooo00<0gia80018W03Oool0>03o
ool00`000000oooo0?ooo`090?ooo`030?l0003oool0oooo0380oooo00<0gia80018W03Oool03`3o
ool2000001H0oooo00@0o`000?ooo`3oool000005@3oool00`3ogi`0B01d0;ooo`0m0?ooo`030?oo
_`1d000007Bo00l0oooo000N0?ooo`030=nLB000M;l0oooo03l0oooo00D0oooO09a800000000oooo
0?l0000j0?ooo`030?oOW01804P0W=oo03P0oooo00<000000?ooo`3oool0203oool00`3o0000oooo
0?ooo`0c0?ooo`030?oOW01804P0W=oo00/0oooo1000000H0?ooo`030?l0003oool0oooo00800000
503oool00`3ooml0W4P007Boo`0o0?ooo`030=nLB000M;l0oooo00d0oooo000N0?ooo`030?oOW018
04P0W=oo03l0oooo00H0oono07@00000000000000?ooo`3o000i0?ooo`030?oo_`1d0000M;oo03P0
oooo00<000000?ooo`3oool01`3oool00`3o0000oooo0?ooo`0d0?ooo`030?oo_`1d0000M;oo00P0
oooo0`00000M0?ooo`040?l0003oool0oooo000001@0oooo00<0oono07@00018W=l0?`3oool00`3o
gi`0B01809cOo`0=0?ooo`006`3oool01P2oM000M;oo0?ooo`3OW4P0001809cOoc/0oooo00H0okmd
0018W03Oool0oooo0=nLB00004P2000000030?ooo`3o0000oooo03T0oooo00<0_g@0001d_`3oool0
=`3oool00`000000oooo0?ooo`060?ooo`030?l0003oool0oooo03H0oooo00<0_g@0001d_`3oool0
103oool300000200oooo00<0o`000?ooo`3oool00P00000?0?ooo`070?noM000B9`0gooo0?ooo`3O
W4P0001809cOo`0l0?ooo`060;md001d_ol0oooo0=nLB00004P0W=oo3@3oool001`0oooo00@0_g@0
000000000000B9cO?@3oool02@3ogi`0B000000000000000M;oo0?ooo`000000oooo0?l0000h0?oo
o`050?oo_`1d000000000000M02oool0=`3oool00`000000oooo0?ooo`050?ooo`030?l0003oool0
oooo03D0oooo00D0oono07@000000000001d0;ooo`04000002@0oooo00@0o`000?ooo`0000000000
403oool01@3ogi`0B000000000000000M;oo03h0oooo00@0_g@0000000000000B9cO3P3oool006<0
oooo0P0000000`3oool0o`000?ooo`1b0?ooo`030000003oool0oooo00D0oooo00<0o`000?ooo`3o
ool0=P3oool4000002P0oooo00<0o`000?ooo`3oool00P00001T0?ooo`00H`3oool2000000030?oo
o`3o0000oooo0780oooo0`0000040?ooo`030?l0003oool0oooo03<0oooo1000000]0?ooo`040?l0
003oool00000000006@0oooo001T0?ooo`040000003oool0o`000?ooo`D00000K@3oool00`000000
oooo0?ooo`030?ooo`030?l0003oool0oooo0300oooo1000000a0?ooo`030?l0003oool000000080
0000H`3oool006@0oooo0P0000000`3oool0o`0000000004000006d0oooo00<000000?ooo`3oool0
0P3oool00`3o0000oooo0?ooo`0Z0?ooo`L00000=P3oool0103o0000oooo00000000001S0?ooo`00
I03oool2000000030?ooo`3o0000000000<00000KP3oool01@000000oooo0?ooo`3oool0o`0002D0
oooo2000000m0?ooo`040?l0003oool00000000006<0oooo001T0?ooo`<0000000<0o`000?ooo`00
00000P00001^0?ooo`040000003oool0oooo0?l001h0oooo200000160?ooo`040?l0003oool00000
00000680oooo000o0?ooo`030000003oool0oooo02<0oooo0P0000000`3oool0o`00000000020000
02X0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0K0?ooo`040000003oool0oooo
0?l001P0oooo1P00001>0?ooo`040?l0003oool0000000000680oooo000M0?ooo`800000803oool3
000002<0oooo0P0000000`3oool0o`0000000002000002P0oooo0`00000S0?ooo`<000007@3oool0
0`000000oooo0?l0000B0?ooo`L00000E03oool0103o0000oooo00000000001R0?ooo`007@3oool4
000001d0oooo1P00000Q0?ooo`<0000000<0oooo0?l0000000009`3oool600000200oooo1@00000M
0?ooo`030000003o0000oooo00`0oooo1P00001L0?ooo`040?l0003oool00000000004H0oooo1000
000G0?ooo`00703oool7000001/0oooo2000000P0?ooo`80000000<0oooo0?l0000000009@3oool8
000001h0oooo1`00000M0?ooo`030?l0003oool0oooo00L0oooo1@00000?0?ooo`T00000703oool:
000002<0oooo00<000000?l0003oool00P00000O0?ooo`T000006@3oool9000001L0oooo000F0?oo
oe<0000000<0o`00000000000000K@0000000`3o000000000000001`000000030?l0000000000000
050000004@3oool001H0oooo00<000000?ooo`3oool00`3oool500000080oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0:0?ooo`060000003oool000000000003oool0o`001`3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo00<0o`000?ooo`000000303oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo100000040?ooo`030000003oool0
oooo00/0oooo00<000000?ooo`3oool02P3oool8000000D0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00<0oooo0P0000000`3oool0o`000?ooo`03
000000`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`050?ooo`@0000000<0oooo
0000003oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00D000001@3oool0
0`000000oooo0?ooo`0?0?ooo`005P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo
00`0oooo0P0000000`3oool0o`000?ooo`0]0?ooo`030000003oool0oooo03`0oooo00<0o`000?oo
o`0000007P3oool200000240oooo00<000000?ooo`0000000`00000V0?ooo`<0000000<0oooo0?l0
003oool00P00000=0?ooo`030000003oool0oooo01@0oooo0P00000Q0?ooo`8000001@3oool00`00
0000oooo0?ooo`0?0?ooo`00J03oool00`000000oooo0?l0001/0?ooo`040?l0003oool0oooo0000
04<0oooo0P00000X0?ooo`<000000P3oool00`3o0000oooo0000001P0?ooo`00J03oool200000003
0?ooo`3o0000oooo06T0oooo00D0o`000?ooo`3oool0oooo0000001/0?ooo`@000000P3oool0103o
0000oooo00000000001O0?ooo`00J03oool2000000030?ooo`3o0000oooo06P0oooo00<0o`000?oo
o`3oool00P3oool00`000000oooo0?ooo`1Z0?ooo`@000000P3oool0103o0000oooo00000000001O
0?ooo`00J@3oool2000000030?ooo`3o0000oooo06H0oooo00<0o`000?ooo`3oool00`3oool00`00
0000oooo0?ooo`1Y0?ooo`D000000`3oool00`3o0000oooo0000001O0?ooo`00J@3oool200000003
0?ooo`3o0000oooo06D0oooo00<0o`000?ooo`3oool0103oool00`000000oooo0?ooo`1[0?ooo`<0
00000`3oool0103o0000oooo00000000001N0?ooo`00JP3oool00`000000oooo0?l0001V0?ooo`03
0?l0003oool0oooo00@0oooo0`00001b0?ooo`030?l000000000000005h0oooo001Z0?ooo`800000
00<0oooo0?l0003oool0H`3oool00`3o0000oooo0?ooo`050?ooo`030000003oool0oooo0780oooo
00@0o`000?ooo`0000000000G@3oool006X0oooo0P0000000`3oool0o`000?ooo`1R0?ooo`030?l0
003oool0oooo00H0oooo00<000000?ooo`3oool0LP3oool0103o0000oooo00000000001M0?ooo`00
J`3oool2000000030?ooo`3o0000oooo0600oooo00<0o`000?ooo`3oool01`3oool00`000000oooo
0?ooo`1c0?ooo`030?l0003oool0000005d0oooo001[0?ooo`80000000<0oooo0?l0003oool0G`3o
ool00`3o0000oooo0?ooo`080?ooo`030000003oool0oooo07<0oooo00@0o`000?ooo`0000000000
G03oool006`0oooo0P0000000`3oool0o`000?ooo`1N0?ooo`030?l0003oool0oooo00P0oooo00<0
00000?ooo`3oool0M03oool00`3o000000000000001L0?ooo`00K03oool2000000030?ooo`3o0000
oooo05d0oooo00<0o`000?ooo`3oool02@3oool00`000000oooo0?ooo`1d0?ooo`030?l0003oool0
000005`0oooo001]0?ooo`040000003oool0oooo0?l005`0oooo00<0o`000?ooo`3oool02P3oool0
0`000000oooo0?ooo`1d0?ooo`030?l0003oool0000005`0oooo001]0?ooo`80000000<0oooo0?l0
003oool0FP3oool00`3o0000oooo0?ooo`0;0?ooo`<00000M@3oool00`3o000000000000001K0?oo
o`00I@3oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo0?l005T0oooo00<0o`000?oo
o`3oool0303oool00`000000oooo0?ooo`1e0?ooo`030?l0003oool0000005/0oooo001T0?ooo`<0
00001`3oool2000000030?ooo`3o0000oooo02<0oooo00<000000?ooo`3oool08`3oool00`000000
oooo0?ooo`0<0?ooo`030?l0003oool0oooo00`0oooo00<000000?ooo`3oool0MP3oool00`3o0000
00000?ooo`1J0?ooo`00I03oool4000000H0oooo0P0000020?ooo`030?l0003oool0oooo01l0oooo
0`00000S0?ooo`<000003@3oool00`3o0000oooo0?ooo`0=0?ooo`030000003oool0oooo07H0oooo
00<0o`00000000000000FP3oool003l0oooo0P00000R0?ooo`D000001`3oool010000000oooo0?oo
o`3o000P0?ooo`@000008P3oool4000000`0oooo00<0o`000?ooo`3oool03P3oool00`000000oooo
0?ooo`1f0?ooo`030?l0003oool0000005X0oooo000M0?ooo`800000803oool400000280oooo1000
00060?ooo`8000000P3oool00`3o0000oooo0?ooo`0L0?ooo`H00000803oool6000000X0oooo00<0
o`000?ooo`3oool0203oool2000000D0oooo00<000000?ooo`3oool06`3oool2000005X0oooo00<0
o`000000003oool0?P3oool4000001L0oooo000M0?ooo`H00000703oool700000240oooo0`000006
0?ooo`050000003oool0oooo0?ooo`3o00006`3oool8000001h0oooo200000090?ooo`030?l0003o
ool0oooo00T0oooo1P0000000`3oool000000?ooo`0L0?ooo`H00000803oool4000002T0oooo00<0
00000?ooo`3oool01P3oool00`3o000000000000000G0?ooo`T000006P3oool8000001L0oooo000M
0?ooo`T000006@3oool900000240oooo0P0000050?ooo`8000000P3oool00`3o0000oooo0?ooo`0H
0?ooo`D000008@3oool5000000`0oooo00<0o`000?ooo`3oool02P3oool9000001`0oooo2@00000M
0?ooo`T000008P3oool2000000T0oooo00<0o`000?ooo`000000603oool8000001`0oooo1P00000G
0?ooo`007@3oool5000001d0oooo1@00000]0?ooo`050000003oool0oooo0?ooo`3o0000D03oool0
0`3o0000oooo0?ooo`0:0?ooo`X00000703oool6000001l0oooo1`00000R0?ooo`@000002@3oool0
0`3o0000oooo0000000J0?ooo`D000007`3oool4000001L0oooo001a0?ooo`8000000P3oool00`3o
0000oooo0?ooo`1=0?ooo`030?l0003oool0oooo00/0oooo0`0000040?ooo`030000003oool0oooo
01`0oooo0P00000S0?ooo`D000008P3oool5000000/0oooo00<0o`000000000000006`3oool30000
0240oooo0P00000G0?ooo`00LP3oool01@000000oooo0?ooo`3oool0o`0004d0oooo00<0o`000?oo
o`3oool04`3oool00`000000oooo0?ooo`110?ooo`8000008`3oool6000000`0oooo00<0o`000?oo
o`0000007@3oool00`000000oooo0?ooo`0h0?ooo`00LP3oool01@000000oooo0?ooo`3oool0o`00
04`0oooo00<0o`000?ooo`3oool0503oool00`000000oooo0?ooo`1W0?ooo`@000003@3oool00`3o
0000oooo0000001H0?ooo`00L`3oool01@000000oooo0?ooo`3oool0o`0004X0oooo00<0o`000?oo
o`3oool05@3oool00`000000oooo0?ooo`1X0?ooo`<000003P3oool00`3o000000000000001G0?oo
o`00L`3oool01@000000oooo0?ooo`3oool0o`0004T0oooo00<0o`000?ooo`3oool05P3oool00`00
0000oooo0?ooo`1Y0?ooo`030000003oool0oooo00d0oooo00<0o`00000000000000E`3oool007@0
oooo00D000000?ooo`3oool0oooo0?l000170?ooo`030?l0003oool0oooo01L0oooo00<000000?oo
o`3oool0N@3oool00`3o0000oooo0000001G0?ooo`00M03oool00`000000oooo0?ooo`020?ooo`03
0?l0003oool0oooo04<0oooo00<0o`000?ooo`3oool0603oool00`000000oooo0?ooo`1i0?ooo`03
0?l0003oool0000005L0oooo001e0?ooo`050000003oool0oooo0?ooo`3o0000A03oool00`3o0000
oooo0?ooo`0I0?ooo`<00000NP3oool00`3o000000000000001F0?ooo`00M@3oool2000000<0oooo
00<0o`000?ooo`3oool0@03oool00`3o0000oooo0?ooo`0J0?ooo`030000003oool0oooo07X0oooo
00<0o`000?ooo`000000EP3oool007H0oooo00<000000?ooo`3oool00P3oool00`3o0000oooo0?oo
o`0n0?ooo`030?l0003oool0oooo01/0oooo00<000000?ooo`3oool0NP3oool00`3o0000oooo0000
001F0?ooo`00MP3oool2000000@0oooo00<0o`000?ooo`3oool0?03oool0103o0000oooo0?ooo`2o
M0020000000707Boo`3oool0oooo0?ooo`3ogi`0B00007Boo`020?ooo`060?oOW018000000000000
0000B9`0gooo2P3oool00`000000oooo0?ooo`1k0?ooo`030?l000000000000005D0oooo001g0?oo
o`030000003oool0oooo0080oooo00<0o`000?ooo`3oool0>`3oool02@3o0000oooo0?ooo`3OW4P0
07Bo0?ooo`3ogi`0B00004RLg`020?ooo`030?oOW0180000M;oo0080oooo00H0okmd0000001d_ol0
oooo0?oOW018M;l:0?ooo`030000003oool0oooo07/0oooo00<0o`00000000000000E@3oool007L0
oooo0P0000040?ooo`030?l0003oool0oooo03T0oooo00H0o`000?ooo`3oool0oooO09a80018W=l2
0?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool0203oool00`000000oooo0?ooo`1k
0?ooo`030?l0003oool0000005D0oooo001h0?ooo`800000103oool00`3o0000oooo0?ooo`0f0?oo
o`80o`000`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono
07@0M02oool0203oool00`000000oooo0?ooo`1k0?ooo`040?l0003oool00000000005@0oooo001h
0?ooo`8000001@3oool00`3o0000oooo0?ooo`0B0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool00P3oool00`3o0000oooo0?ooo`030?ooo`030?oog`2LB00007Bo00<0oooo00<0gia8
0000B02Lgol02P3oool00`3ooml0W4P007Boo`080?ooo`030000003oool0oooo07`0oooo00<0o`00
000000000000E03oool007T0oooo0P0000050?ooo`030?l0003oool0oooo0100oooo0`00000J0?oo
o`<000000P3oool00`3o0000oooo0?ooo`050?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l02P3oool00`3ookl0M00004RLg`080?ooo`@00000N`3oool00`3o0000oooo0000001D0?oo
o`00I03oool00`000000oooo0?ooo`0B0?ooo`8000001P3oool00`3o0000oooo0?ooo`0?0?ooo`<0
00006@3oool4000000030?ooo`3o0000oooo00L0oooo00<0gia80000B02Lgol00P3oool00`3ooml0
W4P0001d_`0:0?ooo`030?noM0000000M;oo00P0oooo00<000000?ooo`3oool0O03oool00`3o0000
oooo0000001D0?ooo`00I03oool3000001<0oooo0P0000060?ooo`030?l0003oool0oooo00d0oooo
1@00000G0?ooo`D0000000<0o`000?ooo`3oool01`3oool00`3ogi`0B00004RLg`030?ooo`030;md
0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@000000000M02oool80?ooo`030000003oool0oooo
07`0oooo00@0o`000?ooo`0000000000D`3oool003l0oooo0P00000R0?ooo`D000004`3oool00`00
0000oooo0?ooo`050?ooo`030?l0003oool0oooo00/0oooo1@00000I0?ooo`8000000P3o00000`00
0000oooo0?ooo`070?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo00T0oooo00<000000?ooo`3oool0O@3oool00`3o000000000000
001C0?ooo`007@3oool200000200oooo1000000P0?ooo`H000004P3oool2000000L0oooo00<0o`00
0?ooo`3oool02@3oool4000001`0oooo00<0o`000000000000000P0000090?ooo`030?oOW0180000
M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?ooo`030?ooo`8000001@3oool0
0`000000oooo0?ooo`0L0?ooo`800000G`3oool00`3o0000oooo0000000i0?ooo`@000005P3oool0
01d0oooo1P00000L0?ooo`L000007@3oool800000140oooo0P0000070?ooo`030?l0003oool0oooo
00P0oooo0P00000M0?ooo`040?l0003oool0oooo0?ooo`8000002P3oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`P0oooo00<0omnL07BLg`3oool00`3oool6000000030?ooo`000000oooo01d0oooo
1P00000O0?ooo`@00000>03oool00`3o0000oooo0000000B0?ooo`T000006@3oool9000001H0oooo
000M0?ooo`T000006@3oool9000001l0oooo1@00000A0?ooo`030000003oool0oooo00H0oooo00<0
o`000?ooo`3oool01P3oool2000001`0oooo0P3o00060?ooo`030000003oool0oooo00P0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO00T000007@3oool9000001`0
oooo2@00000L0?ooo`T000003`3oool00`3o000000000?ooo`0B0?ooo`P000006`3oool6000001L0
oooo000M0?ooo`L000006`3oool5000003T0oooo0P0000080?ooo`030?l0003oool0oooo0240oooo
0P3o000V0?ooo`P000007P3oool7000001h0oooo1`00000N0?ooo`P00000403oool00`3o00000000
0000000D0?ooo`D000007P3oool4000001L0oooo000M0?ooo`<00000GP3oool2000000P0oooo0P3o
000P0?ooo`030?l0003oool0oooo02H0oooo0`0000040?ooo`<00000703oool300000280oooo0`00
000R0?ooo`H000004P3oool00`3o0000oooo0000000F0?ooo`<00000803oool2000001L0oooo001o
0?ooo`030000003oool0oooo00P0oooo0P3o000L0?ooo`80o`00<03oool00`000000oooo0?ooo`1W
0?ooo`<00000503oool00`3o0000oooo0000000H0?ooo`030000003oool0oooo03L0oooo001o0?oo
o`8000002`3oool00`3o0000oooo0?ooo`0G0?ooo`80o`00<P3oool00`000000oooo0?ooo`1W0?oo
o`030000003oool0oooo01D0oooo00<0o`000000003oool0D@3oool00800oooo0P00000;0?ooo`80
o`005@3oool20?l003@0oooo00<000000?ooo`3oool0O`3oool00`3o000000000000001A0?ooo`00
P@3oool2000000`0oooo0`3o000?0?ooo`<0o`00=P3oool00`000000oooo0?ooo`1o0?ooo`030?l0
0000000000000540oooo00220?ooo`030000003oool0oooo00d0oooo3`3o000i0?ooo`030000003o
ool0oooo0800oooo00<0o`000000003oool0D03oool008<0oooo00<000000?ooo`3oool0E03oool0
0`000000oooo0?ooo`200?ooo`030?l000000000oooo0500oooo00230?ooo`800000E@3oool00`00
0000oooo0?ooo`200?ooo`030?l000000000oooo0500oooo00240?ooo`800000E03oool300000800
oooo00<0o`000000003oool0D03oool008D0oooo00<000000?ooo`3oool0DP3oool00`000000oooo
0?ooo`210?ooo`030?l0003oool0oooo04l0oooo00260?ooo`030000003oool0oooo0540oooo00<0
00000?ooo`3oool0P@3oool00`3o0000oooo0?ooo`1?0?ooo`00Q`3oool00`000000oooo0?ooo`1@
0?ooo`030000003oool0oooo0840oooo00<0o`000?ooo`3oool0C`3oool008P0oooo00<000000?oo
o`3oool0C`3oool00`000000oooo0?ooo`210?ooo`030?l0003oool0oooo04l0oooo00290?ooo`03
0000003oool0oooo04h0oooo00<000000?ooo`3oool0P@3oool00`000000o`000?ooo`1?0?ooo`00
RP3oool3000004d0oooo00<000000?ooo`3oool0P@3oool00`000000o`000?ooo`1?0?ooo`00R`3o
ool3000004`0oooo00<000000?ooo`3oool0P03oool2000000030?l0003oool0oooo04h0oooo002:
0?ooo`@00000C03oool300000800oooo0P0000000`3o0000oooo0?ooo`1>0?ooo`00I03oool00`00
0000oooo0?ooo`0R0?ooo`H00000B`3oool00`000000oooo0?ooo`200?ooo`040000003oool0oooo
0?l004l0oooo00100?ooo`8000008P3oool3000002<0oooo1P00000N0?ooo`800000:P3oool00`00
0000oooo0?ooo`1o0?ooo`8000000P3oool00`3o0000oooo0?ooo`1=0?ooo`007@3oool200000240
oooo1000000O0?ooo`H000008`3oool6000001`0oooo1000000Q0?ooo`8000001@3oool00`000000
oooo0?ooo`0L0?ooo`8000008`3oool2000003/0oooo0P0000030?ooo`030?l0003oool0oooo03<0
oooo1000000F0?ooo`007@3oool6000001`0oooo1`00000M0?ooo`P000008`3oool2000000030?oo
o`000000000001X0oooo1`00000O0?ooo`H0000000<0oooo0000003oool07@3oool6000001l0oooo
1P00000O0?ooo`@000004`3oool2000000@0oooo00<0o`000?ooo`3oool0;P3oool9000001H0oooo
000M0?ooo`T000006@3oool9000001/0oooo2@00000T0?ooo`030000003oool0000000800000603o
ool9000001d0oooo2@00000M0?ooo`T00000703oool9000001`0oooo2@00000<0?ooo`<000001P3o
ool00`3o0000oooo0?ooo`0;0?ooo`T000006`3oool7000001H0oooo000M0?ooo`L000006`3oool5
000004l0oooo1000000F0?ooo`D000008@3oool8000001h0oooo1`00000O0?ooo`H000007P3oool7
000000`0oooo100000070?ooo`030?l0003oool0oooo00`0oooo2000000M0?ooo`D000005P3oool0
01d0oooo0`00001e0?ooo`@00000>P3oool3000000@0oooo00<000000?ooo`3oool0703oool30000
02<0oooo0P00000S0?ooo`@00000303oool4000000T0oooo00<0o`000?ooo`3oool03P3oool50000
0240oooo0P00000F0?ooo`00U`3oool300000400oooo0`00001W0?ooo`800000303oool4000000/0
oooo00<0o`000?ooo`3oool0403oool3000003T0oooo002I0?ooo`<00000?P3oool00`000000oooo
0?ooo`1a0?ooo`H000003P3oool00`3o0000oooo0?ooo`0A0?ooo`030000003oool0oooo03L0oooo
002K0?ooo`<00000?03oool00`000000oooo0?ooo`1/0?ooo`L000004P3oool00`3o0000oooo0?oo
o`1;0?ooo`00W@3oool4000002P0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`09
0?ooo`030000003oool0oooo06H0oooo2P00000E0?ooo`030?l0003oool0oooo04/0oooo002O0?oo
o`@00000:03oool00`3o_g@0000007Boo`0<0?ooo`030000003oool0oooo0600oooo2`00000J0?oo
o`030?l0003oool0oooo04/0oooo002R0?ooo`@000009@3oool00`3ooml0W4P0001d_`0<0?ooo`03
0000003oool0oooo05P0oooo3@00000Q0?ooo`030?l0003oool0oooo04X0oooo002T0?ooo`H00000
8P3oool00`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo0500oooo3@00000Y0?ooo`030?l0
003oool0oooo04X0oooo002X0?ooo`L000007@3oool00`3ogi`0B01809cOo`0;0?ooo`030000003o
ool0oooo04@0oooo3`00000c0?ooo`030?l0003oool0oooo04X0oooo002/0?ooo`X000005P3oool0
0`3ookl0M00007Boo`0;0?ooo`@00000=@3ooolE000003/0oooo00<0o`000?ooo`3oool0BP3oool0
0;80oooo3P00000=0?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3oool0903ooolK0000
04L0oooo00<0o`000?ooo`3oool0BP3oool00;T0oooo4P0000020?ooo`030=nLB000B9`0gooo00X0
oooo00<000000?ooo`3oool03P3ooolR000005L0oooo00<0o`000?ooo`3oool0B@3oool00<@0oooo
=@00001[0?ooo`030?l0003oool0oooo04T0oooo003=0?ooo`030?oo_`1d0000M;oo00X0oooo00<0
00000?ooo`3oool0Q`3oool00`3o0000oooo0?ooo`190?ooo`00cP3oool00`2oM00007Bo0?ooo`09
0?ooo`030000003oool0oooo08L0oooo00<0o`000?ooo`3oool0B@3oool00<`0oooo00D0oono07@0
00000000001d0;ooo`090?ooo`030000003oool0oooo08P0oooo00<0o`000?ooo`3oool0B03oool0
0=X0oooo00<000000?ooo`3oool0R03oool00`3o0000oooo0?ooo`180?ooo`00R@3oool00`000000
oooo0?ooo`1>0?ooo`<00000R03oool00`3o0000oooo0?ooo`180?ooo`00I03oool2000002<0oooo
0`00000R0?ooo`800000:P3oool00`000000oooo0?ooo`280?ooo`030?l0003oool0oooo04P0oooo
000N0?ooo`800000803oool200000280oooo1000000P0?ooo`H00000803oool400000240oooo0P00
00050?ooo`030000003oool0oooo01`0oooo0P00001[0?ooo`030?l0003oool0oooo02l0oooo0P00
000F0?ooo`007P3oool6000001`0oooo1P00000M0?ooo`L000007P3oool8000001h0oooo1`00000N
0?ooo`H0000000<0oooo0000003oool07@3oool600000200oooo1000000P0?ooo`@000007`3oool0
0`3o0000oooo0?ooo`0[0?ooo`H000005P3oool001d0oooo2@00000J0?ooo`T000006P3oool90000
01`0oooo2@00000M0?ooo`T00000703oool9000001d0oooo2@00000M0?ooo`T000006`3oool90000
01X0oooo00<0o`000?ooo`3oool01P3oool200000200oooo2@00000F0?ooo`007@3oool7000001`0
oooo1@00000N0?ooo`D00000AP3oool500000200oooo2000000N0?ooo`L000007`3oool7000001d0
oooo1`00000L0?ooo`030?l0003oool0oooo00L0oooo1@00000N0?ooo`L000005P3oool001d0oooo
0`00002c0?ooo`<00000103oool00`000000oooo0?ooo`0L0?ooo`<000008`3oool500000200oooo
1000000O0?ooo`030?l0003oool0oooo00L0oooo1`00000O0?ooo`<000005P3oool00=X0oooo00<0
00000?ooo`3oool0@P3oool2000002<0oooo0P00000Q0?ooo`030?l0003oool0oooo00P0oooo1@00
000i0?ooo`00fP3oool3000008X0oooo00<0o`000?ooo`3oool02P3oool3000003T0oooo003J0?oo
o`030000003oool0oooo08X0oooo00<0o`000?ooo`3oool02`3oool00`000000oooo0?ooo`0h0?oo
o`00fP3oool00`000000oooo0?ooo`2;0?ooo`030?l0003oool0oooo04D0oooo003J0?ooo`030000
003oool0oooo08/0oooo00<0o`000?ooo`3oool0A@3oool00=X0oooo00<000000?ooo`3oool0R`3o
ool00`3o0000oooo0?ooo`150?ooo`00fP3oool00`000000oooo0?ooo`2;0?ooo`030?l0003oool0
oooo04D0oooo003J0?ooo`030000003oool0oooo08`0oooo00<0o`000?ooo`3oool0A03oool00=X0
oooo00<000000?ooo`3oool0S03oool00`3o0000oooo0?ooo`140?ooo`00fP3oool00`000000oooo
0?ooo`2<0?ooo`030?l0003oool0oooo04@0oooo003J0?ooo`<00000S03oool00`3o0000oooo0?oo
o`140?ooo`00fP3oool00`000000oooo0?ooo`2<0?ooo`030?l0003oool0oooo04@0oooo003J0?oo
o`030000003oool0oooo08d0oooo00<0o`000?ooo`3oool0@`3oool00=X0oooo00<000000?ooo`3o
ool0S@3oool00`3o0000oooo0?ooo`130?ooo`00fP3oool00`000000oooo0?ooo`2=0?ooo`030?l0
003oool0oooo04<0oooo003J0?ooo`030000003oool0oooo08d0oooo00<0o`000?ooo`3oool0@`3o
ool00=X0oooo00<000000?ooo`3oool0S@3oool00`3o0000oooo0?ooo`130?ooo`00I03oool20000
02<0oooo0P00001?0?ooo`030000003oool0oooo08d0oooo00<0o`000?ooo`3oool0@`3oool001h0
oooo0P00000P0?ooo`8000008P3oool400000240oooo1000000Q0?ooo`800000903oool2000000@0
oooo0`00000M0?ooo`8000008`3oool2000004X0oooo00<0o`000?ooo`3oool0@P3oool001h0oooo
1P00000L0?ooo`H000007P3oool7000001d0oooo1`00000O0?ooo`H00000803oool7000001l0oooo
1P00000O0?ooo`H000007`3oool4000002<0oooo00<0o`000?ooo`3oool0:@3oool4000001D0oooo
000N0?ooo`T000006@3oool9000001/0oooo2@00000K0?ooo`T000007@3oool9000001d0oooo2@00
000M0?ooo`P000007@3oool9000001`0oooo2@00000N0?ooo`040?l0003oool0oooo0?ooo`800000
8@3oool9000001D0oooo000N0?ooo`L000006`3oool7000001d0oooo1@00000O0?ooo`D000008@3o
ool5000000T0oooo00<0oono07@0000000000P0000001`18W=l0oooo0?ooo`3oool0omnL04P0001d
_ol00P3oool01P3ogi`0B00000000000000004RL0=ooo`@0oooo1`00000O0?ooo`H000007`3oool7
000001h0oooo1`00000P0?ooo`030?l0003oool0oooo0080oooo0P00000R0?ooo`H000005P3oool0
01h0oooo0`00000O0?ooo`<00000N`3oool00`3o_g@0000007Boo`040?ooo`030?oOW0180000M;oo
0080oooo00H0okmd0000001d_ol0oooo0?oOW018M;l40?ooo`<000000`3oool00`000000oooo0?oo
o`0M0?ooo`8000008`3oool300000280oooo1@00000R0?ooo`030?l0003oool0oooo0080oooo1000
000T0?ooo`8000005P3oool00;h0oooo00<0oooO09a80000M;l03P3oool00`3o_g@0B9cO0?ooo`08
0?ooo`030000003oool0oooo06L0oooo0P00000U0?ooo`030?l0003oool0oooo00<0oooo1@00000j
0?ooo`00_`3oool00`3OW4P004RL0=ooo`0=0?ooo`030?oo_`1d07@0_ooo00P0oooo00<000000?oo
o`3oool0S`3oool00`3o0000oooo0?ooo`030?ooo`D00000>@3oool00;l0oooo00<0omnL04P0B02L
gol03@3oool00`3ooml0W4P007Boo`080?ooo`030000003oool0oooo08l0oooo00<0o`000?ooo`3o
ool0103oool3000003X0oooo002o0?ooo`030?oo_`1d0000M;oo00d0oooo00<0oono07@00018W=l0
203oool4000008h0oooo00<0o`000?ooo`3oool0103oool3000003X0oooo00300?ooo`030;md0000
M;l0oooo00`0oooo00<0okmd0000001d_ol0gP3oool00<00oooo00<0gia80018W03Oool02@3oool0
1P3ogml0gooo0?noM0000000001d0;ooomh0oooo00300?ooo`030?oOW01804P0W=oo00T0oooo00D0
oono07@000000000001809cOo`3O0?ooo`00`03oool00`3ookl0M00007Boo`0:0?ooo`030=nLW02o
ool0oooo0>00oooo00310?ooo`030;md0000M;l0oooo00T0oooo00<0omnL07BLg`3oool0h03oool0
0;l0oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l0g@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.34731, -1.64567, \
0.0229522, 0.0185688}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(RelaxationStrobePlot[\((\(1\/3\) #\^3 - #)\) &, 0.01, 4, 2, 0.1, 0. , 
        18, 20, 0.01, 
        PlotRange \[Rule] {{\(-3\), 3}, {\(-1.5\), 1.5}}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.166667 0.309017 0.206011 [
[0 .29652 -9.34375 -15.6875 ]
[0 .29652 9.34375 0 ]
[.16667 .29652 -9.34375 -15.6875 ]
[.16667 .29652 9.34375 0 ]
[.33333 .29652 -9.34375 -15.6875 ]
[.33333 .29652 9.34375 0 ]
[.66667 .29652 -5 -15.6875 ]
[.66667 .29652 5 0 ]
[.83333 .29652 -5 -15.6875 ]
[.83333 .29652 5 0 ]
[1 .29652 -5 -15.6875 ]
[1 .29652 5 0 ]
[.4875 0 -27.6875 -7.84375 ]
[.4875 0 0 7.84375 ]
[.4875 .10301 -18.6875 -7.84375 ]
[.4875 .10301 0 7.84375 ]
[.4875 .20601 -27.6875 -7.84375 ]
[.4875 .20601 0 7.84375 ]
[.4875 .41202 -19 -7.84375 ]
[.4875 .41202 0 7.84375 ]
[.4875 .51503 -10 -7.84375 ]
[.4875 .51503 0 7.84375 ]
[.4875 .61803 -19 -7.84375 ]
[.4875 .61803 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
0 .30902 m
0 .31527 L
s
gsave
0 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.16667 .30902 m
.16667 .31527 L
s
gsave
.16667 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.33333 .30902 m
.33333 .31527 L
s
gsave
.33333 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.66667 .30902 m
.66667 .31527 L
s
gsave
.66667 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.83333 .30902 m
.83333 .31527 L
s
gsave
.83333 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
1 .30902 m
1 .31527 L
s
gsave
1 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
.03333 .30902 m
.03333 .31277 L
s
.06667 .30902 m
.06667 .31277 L
s
.1 .30902 m
.1 .31277 L
s
.13333 .30902 m
.13333 .31277 L
s
.2 .30902 m
.2 .31277 L
s
.23333 .30902 m
.23333 .31277 L
s
.26667 .30902 m
.26667 .31277 L
s
.3 .30902 m
.3 .31277 L
s
.36667 .30902 m
.36667 .31277 L
s
.4 .30902 m
.4 .31277 L
s
.43333 .30902 m
.43333 .31277 L
s
.46667 .30902 m
.46667 .31277 L
s
.53333 .30902 m
.53333 .31277 L
s
.56667 .30902 m
.56667 .31277 L
s
.6 .30902 m
.6 .31277 L
s
.63333 .30902 m
.63333 .31277 L
s
.7 .30902 m
.7 .31277 L
s
.73333 .30902 m
.73333 .31277 L
s
.76667 .30902 m
.76667 .31277 L
s
.8 .30902 m
.8 .31277 L
s
.86667 .30902 m
.86667 .31277 L
s
.9 .30902 m
.9 .31277 L
s
.93333 .30902 m
.93333 .31277 L
s
.96667 .30902 m
.96667 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 0 m
.50625 0 L
s
gsave
.4875 0 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.5 .10301 m
.50625 .10301 L
s
gsave
.4875 .10301 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .20601 m
.50625 .20601 L
s
gsave
.4875 .20601 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .41202 m
.50625 .41202 L
s
gsave
.4875 .41202 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .51503 m
.50625 .51503 L
s
gsave
.4875 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .61803 m
.50625 .61803 L
s
gsave
.4875 .61803 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .0206 m
.50375 .0206 L
s
.5 .0412 m
.50375 .0412 L
s
.5 .0618 m
.50375 .0618 L
s
.5 .0824 m
.50375 .0824 L
s
.5 .12361 m
.50375 .12361 L
s
.5 .14421 m
.50375 .14421 L
s
.5 .16481 m
.50375 .16481 L
s
.5 .18541 m
.50375 .18541 L
s
.5 .22661 m
.50375 .22661 L
s
.5 .24721 m
.50375 .24721 L
s
.5 .26781 m
.50375 .26781 L
s
.5 .28842 m
.50375 .28842 L
s
.5 .32962 m
.50375 .32962 L
s
.5 .35022 m
.50375 .35022 L
s
.5 .37082 m
.50375 .37082 L
s
.5 .39142 m
.50375 .39142 L
s
.5 .43262 m
.50375 .43262 L
s
.5 .45322 m
.50375 .45322 L
s
.5 .47383 m
.50375 .47383 L
s
.5 .49443 m
.50375 .49443 L
s
.5 .53563 m
.50375 .53563 L
s
.5 .55623 m
.50375 .55623 L
s
.5 .57683 m
.50375 .57683 L
s
.5 .59743 m
.50375 .59743 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
1 0 0 r
.5 Mabswid
.12906 0 m
.16639 .17066 L
.1932 .2599 L
.2223 .33461 L
.25007 .38638 L
.26408 .40586 L
.27951 .42256 L
.28678 .42879 L
.2937 .43378 L
.29998 .43756 L
.30668 .44081 L
.31048 .44231 L
.31393 .44346 L
.31743 .44442 L
.32066 .44514 L
.32443 .44576 L
.32641 .446 L
.32743 .4461 L
.32851 .44618 L
.33034 .44629 L
.33204 .44635 L
.33308 .44636 L
.33406 .44635 L
.33591 .44631 L
.33751 .44623 L
.33922 .4461 L
.34085 .44595 L
.34234 .44577 L
.34579 .44524 L
.34943 .4445 L
.35628 .44263 L
.36378 .4399 L
.37714 .43337 L
.38957 .42554 L
.41821 .402 L
.44424 .37537 L
.50021 .30876 L
.55406 .24453 L
.58134 .21646 L
.59581 .20364 L
.61135 .19186 L
.62624 .18282 L
.63985 .17672 L
.64639 .1746 L
.64985 .1737 L
.65353 .17292 L
.65706 .17235 L
.66027 .17198 L
.66181 .17185 L
.66349 .17175 L
Mistroke
.66506 .1717 L
.66651 .17168 L
.66829 .1717 L
.66997 .17176 L
.67148 .17185 L
.67311 .17199 L
.6749 .17219 L
.67679 .17245 L
.68018 .17307 L
.68624 .17463 L
.69292 .17706 L
.70012 .18053 L
.70786 .1853 L
.72187 .19677 L
.7353 .21141 L
.74998 .23173 L
.77954 .28749 L
.80705 .35886 L
.8364 .45786 L
.86433 .57598 L
Mfstroke
.86433 .57598 m
.87245 .61803 L
s
0 g
.008 w
.19093 .25469 Mdot
.1922 .2585 Mdot
.19348 .2623 Mdot
.19476 .26608 Mdot
.19606 .26984 Mdot
.19737 .27359 Mdot
.19869 .27733 Mdot
.20001 .28104 Mdot
.20135 .28474 Mdot
.20271 .28842 Mdot
.20407 .29209 Mdot
.20544 .29574 Mdot
.20683 .29937 Mdot
.20823 .30299 Mdot
.20965 .30659 Mdot
.21107 .31017 Mdot
.21251 .31373 Mdot
.21397 .31727 Mdot
.21544 .3208 Mdot
.21693 .32431 Mdot
.21843 .3278 Mdot
.21994 .33127 Mdot
.22148 .33472 Mdot
.22303 .33815 Mdot
.2246 .34157 Mdot
.22619 .34496 Mdot
.2278 .34834 Mdot
.22943 .35169 Mdot
.23109 .35502 Mdot
.23276 .35834 Mdot
.23446 .36163 Mdot
.23618 .3649 Mdot
.23793 .36815 Mdot
.23971 .37138 Mdot
.24151 .37459 Mdot
.24334 .37777 Mdot
.24521 .38093 Mdot
.2471 .38407 Mdot
.24904 .38718 Mdot
.25101 .39027 Mdot
.25302 .39334 Mdot
.25507 .39638 Mdot
.25716 .39939 Mdot
.25931 .40238 Mdot
.2615 .40534 Mdot
.26376 .40828 Mdot
.26607 .41118 Mdot
.26845 .41406 Mdot
.2709 .41691 Mdot
.27342 .41972 Mdot
.27604 .42251 Mdot
.27875 .42526 Mdot
.28156 .42798 Mdot
.2845 .43066 Mdot
.28758 .43331 Mdot
.29081 .43591 Mdot
.29423 .43848 Mdot
.29786 .441 Mdot
.30174 .44347 Mdot
.30594 .4459 Mdot
.31053 .44827 Mdot
.3156 .45058 Mdot
.32133 .45282 Mdot
.32793 .45499 Mdot
.33581 .45707 Mdot
.34564 .45904 Mdot
.3587 .46088 Mdot
.37782 .46251 Mdot
.41023 .46384 Mdot
.47893 .46459 Mdot
.65074 .46393 Mdot
.81594 .46084 Mdot
.83539 .45676 Mdot
.83532 .45261 Mdot
.83427 .44848 Mdot
.83316 .44435 Mdot
.83205 .44024 Mdot
.83093 .43614 Mdot
.8298 .43206 Mdot
.82867 .42799 Mdot
.82753 .42393 Mdot
.82639 .41989 Mdot
.82524 .41586 Mdot
.82408 .41185 Mdot
.82291 .40785 Mdot
.82174 .40387 Mdot
.82055 .3999 Mdot
.81937 .39594 Mdot
.81817 .392 Mdot
.81697 .38808 Mdot
.81575 .38417 Mdot
.81453 .38027 Mdot
.8133 .37639 Mdot
.81207 .37253 Mdot
.81082 .36868 Mdot
.80957 .36484 Mdot
.8083 .36102 Mdot
.80703 .35722 Mdot
.80574 .35343 Mdot
.80445 .34966 Mdot
.80315 .34591 Mdot
.80183 .34217 Mdot
.80051 .33845 Mdot
.79917 .33474 Mdot
.79782 .33105 Mdot
.79647 .32738 Mdot
.7951 .32372 Mdot
.79371 .32008 Mdot
.79232 .31646 Mdot
.79091 .31286 Mdot
.78949 .30927 Mdot
.78805 .3057 Mdot
.7866 .30215 Mdot
.78514 .29861 Mdot
.78366 .2951 Mdot
.78216 .2916 Mdot
.78065 .28812 Mdot
.77912 .28466 Mdot
.77758 .28122 Mdot
.77601 .2778 Mdot
.77443 .2744 Mdot
.77283 .27102 Mdot
.77121 .26766 Mdot
.76956 .26431 Mdot
.7679 .26099 Mdot
.76621 .25769 Mdot
.7645 .25441 Mdot
.76276 .25115 Mdot
.76099 .24792 Mdot
.7592 .2447 Mdot
.75738 .24151 Mdot
.75553 .23834 Mdot
.75364 .23519 Mdot
.75172 .23207 Mdot
.74977 .22897 Mdot
.74778 .22589 Mdot
.74574 .22284 Mdot
.74366 .21982 Mdot
.74154 .21682 Mdot
.73936 .21385 Mdot
.73713 .2109 Mdot
.73485 .20798 Mdot
.73249 .2051 Mdot
.73007 .20224 Mdot
.72758 .19941 Mdot
.725 .19661 Mdot
.72233 .19385 Mdot
.71955 .19112 Mdot
.71666 .18842 Mdot
.71364 .18576 Mdot
.71048 .18314 Mdot
.70713 .18056 Mdot
.70359 .17802 Mdot
.69981 .17553 Mdot
.69574 .17308 Mdot
.69132 .17069 Mdot
.68646 .16835 Mdot
.68101 .16608 Mdot
.67478 .16388 Mdot
.66746 .16176 Mdot
.65851 .15975 Mdot
.64693 .15786 Mdot
.63069 .15613 Mdot
.60491 .15466 Mdot
.55532 .15364 Mdot
.43371 .15358 Mdot
.22318 .15574 Mdot
.16666 .15965 Mdot
.16437 .1638 Mdot
.16531 .16794 Mdot
.1664 .17207 Mdot
.16751 .17619 Mdot
.16863 .18029 Mdot
.16975 .18438 Mdot
.17088 .18845 Mdot
.17202 .19251 Mdot
.17316 .19656 Mdot
.17431 .20059 Mdot
.17547 .20461 Mdot
.17663 .20862 Mdot
.1778 .21261 Mdot
.17898 .21658 Mdot
.18017 .22054 Mdot
.18136 .22449 Mdot
.18256 .22842 Mdot
.18377 .23234 Mdot
.18499 .23624 Mdot
.18621 .24012 Mdot
.18745 .24399 Mdot
.18869 .24785 Mdot
.18994 .25169 Mdot
.1912 .25551 Mdot
.19247 .25932 Mdot
.19375 .26312 Mdot
.19504 .26689 Mdot
.19634 .27066 Mdot
.19765 .2744 Mdot
.19897 .27813 Mdot
.2003 .28184 Mdot
.20165 .28554 Mdot
.203 .28922 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00_03oool00`3ookl0M000000000020000000704RLg`3oool0oooo0?ooo`3ogi`0B00007Bo
o`020?ooo`060?oOW0180000000000000000B9`0goooh03oool00;h0oooo00<0okmd0000001d_ol0
103oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Boh03oool00;h0
oooo00<0oooO09a80000M;l03P3oool00`3o_g@0B9cO0?ooo`3N0?ooo`00_`3oool00`3OW4P004RL
0=ooo`0=0?ooo`030?oo_`1d07@0_ooo0=h0oooo002_0?ooo`030?oog`2LB000000000L0000000<0
B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`0=0?ooo`030?oog`2LB000M;oo0=h0oooo0019
0?ooo`030?l0003oool0oooo07<0oooo00<0oono07@0001d_ol03@3oool00`3ookl0M00004RLg`08
0?ooo`@00000dP3oool004T0oooo00<0o`000?ooo`3oool0M03oool00`2oM00007Bo0?ooo`0<0?oo
o`030?noM0000000M;oo00P0oooo00<000000?ooo`3oool0d`3oool004T0oooo00<0o`000?ooo`3o
ool0M03oool00`3OW4P004RL0=ooo`090?ooo`060?oOg`3Oool0okmd0000000007@0_ooo203oool0
0`000000oooo0?ooo`3C0?ooo`00BP3oool00`3o0000oooo0?ooo`1c0?ooo`030?oOW01804P0W=oo
00T0oooo00D0oono07@000000000001809cOo`090?ooo`030000003oool0oooo0=<0oooo001:0?oo
o`030?l0003oool0oooo07<0oooo00<0oono07@0001d_ol02P3oool00`3OW9`0_ooo0?ooo`0:0?oo
o`030000003oool0oooo0=<0oooo001:0?ooo`030?l0003oool0oooo07@0oooo00<0_g@0001d_`3o
ool02@3oool00`3ogi`0M9cO0?ooo`0:0?ooo`030000003oool0oooo0=<0oooo001:0?ooo`030?l0
003oool0oooo0780oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md0000000000000000
0018W=l01`3oool00`000000oooo0?ooo`3C0?ooo`00BP3oool00`3o0000oooo0?ooo`2=0?ooo`03
0000003oool0oooo0=<0oooo001;0?ooo`030?l0003oool0oooo08`0oooo0`00003C0?ooo`00B`3o
ool00`3o0000oooo0?ooo`2<0?ooo`030000003oool0oooo0=<0oooo001;0?ooo`030?l0003oool0
oooo08`0oooo00<000000?ooo`3oool0d`3oool004/0oooo00<0o`000?ooo`3oool0S03oool00`00
0000oooo0?ooo`3C0?ooo`00C03oool00`3o0000oooo0?ooo`2;0?ooo`030000003oool0oooo0=<0
oooo001<0?ooo`030?l0003oool0oooo08/0oooo00<000000?ooo`3oool0d`3oool004`0oooo00<0
o`000?ooo`3oool0R`3oool00`000000oooo0?ooo`3C0?ooo`00C03oool00`3o0000oooo0?ooo`2;
0?ooo`030000003oool0oooo0=<0oooo001<0?ooo`030?l0003oool0oooo08/0oooo00<000000?oo
o`3oool0d`3oool004d0oooo00<0o`000?ooo`3oool0RP3oool300000=<0oooo001=0?ooo`030?l0
003oool0oooo08X0oooo00<000000?ooo`3oool0d`3oool004d0oooo00<0o`000?ooo`3oool0RP3o
ool00`000000oooo0?ooo`3C0?ooo`00C@3oool00`3o0000oooo0?ooo`2:0?ooo`030000003oool0
oooo0=<0oooo001=0?ooo`030?l0003oool0oooo08X0oooo00<000000?ooo`3oool0d`3oool004h0
oooo00<0o`000?ooo`3oool0R@3oool00`000000oooo0?ooo`3C0?ooo`00CP3oool00`3o0000oooo
0?ooo`290?ooo`030000003oool0oooo0=<0oooo001>0?ooo`030?l0003oool0oooo08T0oooo00<0
00000?ooo`3oool0d`3oool004h0oooo00<0o`000?ooo`3oool0R@3oool300000=<0oooo001>0?oo
o`030?l0003oool0oooo08T0oooo00<000000?ooo`3oool0d`3oool004l0oooo00<0o`000?ooo`3o
ool0R03oool00`000000oooo0?ooo`3C0?ooo`00C`3oool00`3o0000oooo0?ooo`280?ooo`030000
003oool0oooo0=<0oooo001?0?ooo`030?l0003oool0oooo08P0oooo00<000000?ooo`3oool0d`3o
ool004l0oooo00<0o`000?ooo`3oool0R03oool00`000000oooo0?ooo`3C0?ooo`00C`3oool00`3o
0000oooo0?ooo`280?ooo`030000003oool0oooo0=<0oooo001@0?ooo`030?l0003oool0oooo08L0
oooo00<000000?ooo`3oool0d`3oool00500oooo00<0o`000?ooo`3oool0Q`3oool300000=<0oooo
001@0?ooo`030?l0003oool0oooo08L0oooo00<000000?ooo`3oool0d`3oool00500oooo00<0o`00
0?ooo`3oool0Q`3oool00`000000oooo0?ooo`3C0?ooo`00D@3oool00`3o0000oooo0?ooo`1e0?oo
o`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool02@3oool00`000000oooo0?ooo`3C0?oo
o`00D@3oool00`3o0000oooo0?ooo`1g0?ooo`030?noM0000000M;oo00`0oooo00<000000?ooo`3o
ool0d`3oool00540oooo00<0o`000?ooo`3oool0M`3oool00`3ooml0W4P0001d_`0<0?ooo`030000
003oool0oooo0=<0oooo001A0?ooo`030?l0003oool0oooo07P0oooo00<0gia80018W03Oool02`3o
ool00`000000oooo0?ooo`3C0?ooo`00D@3oool00`3o0000oooo0?ooo`1X0?ooo`030?oog`2LB000
000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0
oooo0=<0oooo001B0?ooo`030?l0003oool0oooo07L0oooo00<0oono07@0001d_ol02`3oool40000
0=80oooo001B0?ooo`030?l0003oool0oooo07P0oooo00<0_g@0001d_`3oool02P3oool00`000000
oooo0?ooo`3C0?ooo`00DP3oool00`3o0000oooo0?ooo`1h0?ooo`030=nLB000B9`0gooo00X0oooo
00<000000?ooo`3oool0d`3oool00580oooo00<0o`000?ooo`3oool0N03oool00`3ogi`0B01809cO
o`0:0?ooo`030000003oool0oooo0=<0oooo001B0?ooo`030?l0003oool0oooo07P0oooo00<0oono
07@0001d_ol02P3oool00`000000oooo0?ooo`3C0?ooo`00D`3oool00`3o0000oooo0?ooo`1h0?oo
o`030;md0000M;l0oooo00T0oooo00<000000?ooo`3oool0d`3oool005<0oooo00<0o`000?ooo`3o
ool0MP3oool01@3ookl0M0000000000007@0_ooo00T0oooo00<000000?ooo`3oool0d`3oool005<0
oooo00<0o`000?ooo`3oool0Q03oool00`000000oooo0?ooo`3C0?ooo`00D`3oool00`3o0000oooo
0?ooo`240?ooo`<00000d`3oool005<0oooo00<0o`000?ooo`3oool0Q03oool00`000000oooo0?oo
o`3C0?ooo`00E03oool00`3o0000oooo0?ooo`230?ooo`030000003oool0oooo0=<0oooo001D0?oo
o`030?l0003oool0oooo08<0oooo00<000000?ooo`3oool0d`3oool005@0oooo00<0o`000?ooo`3o
ool0P`3oool00`000000oooo0?ooo`3C0?ooo`00E03oool00`3o0000oooo0?ooo`230?ooo`030000
003oool0oooo0=<0oooo001D0?ooo`030?l0003oool0oooo08<0oooo00<000000?ooo`3oool0d`3o
ool005D0oooo00<0o`000?ooo`3oool0PP3oool00`000000oooo0?ooo`3C0?ooo`00E@3oool00`3o
0000oooo0?ooo`220?ooo`<00000d`3oool005D0oooo00<0o`000?ooo`3oool0PP3oool00`000000
oooo0?ooo`3C0?ooo`00E@3oool00`3o0000oooo0?ooo`220?ooo`030000003oool0oooo0=<0oooo
001F0?ooo`030?l0003oool0oooo06L0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0ZP3oool005H0oooo00<0o`00
0?ooo`3oool05@3oool00`000000oooo0?ooo`1>0?ooo`<00000603oool00`000000oooo0?ooo`0B
0?ooo`<00000403oool3000000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2I
0?ooo`00EP3oool00`3o000000000?ooo`0D0?ooo`<00000D03oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`060?oo
o`<00000103oool300000080oooo00<000000?ooo`3oool0U@3oool005H0oooo0`00000E0?ooo`03
0000003oool0oooo06T0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00<000000P3oool00`000000oooo0?ooo`2A0?ooo`00EP3oool00`3o00000000
0?ooo`210?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool00`0000000`3oool00000
0?ooo`2?0?ooo`00EP3oool300000840oooo0`00000o0?ooo`030000003oool0000000<0000000<0
oooo0000003oool0R`3oool005H0oooo0`0000210?ooo`030000003oool0oooo0480oooo1`00002:
0?ooo`00E`3oool00`000000oooo0?ooo`200?ooo`030000003oool0oooo04@0oooo00<000000?oo
o`000000100000280?ooo`00EP3oool300000840oooo00<000000?ooo`3oool0>@3oool<0?l000<0
oooo1@0000260?ooo`00E`3oool200000840oooo00<000000?ooo`3oool0=P3oool30?l000`0oooo
0`3o00020?ooo`@00000Q@3oool005L0oooo0`0000200?ooo`030000003oool0oooo03<0oooo0`3o
000B0?ooo`<0o`0000@0oooo0000000000000000Q03oool005L0oooo0`0000200?ooo`030000003o
ool0oooo0340oooo0P3o000H0?ooo`80o`00100000220?ooo`00F03oool200000800oooo00<00000
0?ooo`3oool0;`3oool20?l001`0oooo00<0o`000000000000000P0000210?ooo`00F03oool30000
07l0oooo0`00000]0?ooo`80o`00803oool0103o000000000000000000200?ooo`00F03oool30000
07l0oooo00<000000?ooo`3oool0;03oool00`3o0000oooo0?ooo`0Q0?ooo`040?l0000000000000
000007l0oooo001I0?ooo`800000O`3oool00`000000oooo0?ooo`0Z0?ooo`80o`009@3oool0103o
0000000000000000001n0?ooo`00F@3oool300000680oooo00@0_g@0000000000000M;oo0`3oool0
0`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000B9`0gooo2P3oool00`000000oooo
0?ooo`0Y0?ooo`030?l0003oool0oooo02H0oooo00@0o`000000000000000000O@3oool005X0oooo
00<000000?ooo`3oool0H03oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0
001d_ol00P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_`X0oooo00<000000?ooo`3oool09`3o
ool20?l002X0oooo00<0o`00000000000000O@3oool005T0oooo0`00001P0?ooo`030?oog`2LB000
B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`080?ooo`030000003oool0
oooo02H0oooo00<0o`000?ooo`3oool0:`3oool3000007`0oooo001I0?ooo`<00000H03oool00`3o
oml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02oool0203oool0
0`000000oooo0?ooo`0T0?ooo`80o`00;`3oool3000007/0oooo001J0?ooo`800000D`3oool00`3o
oml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cO
o`0:0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?ooo`3oool08`3oool00`3o0000oooo0?oo
o`0`0?ooo`<00000NP3oool005X0oooo0`00001P0?ooo`030;md0000B9`0gooo0080oooo00<0omnL
04P00018W=l02P3oool00`3ookl0M00004RLg`080?ooo`@000008@3oool00`3o0000oooo0?ooo`0b
0?ooo`<00000N@3oool005X0oooo0`00001P0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l02P3oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo0240oooo00<0o`000?oo
o`3oool0=03oool3000007P0oooo001K0?ooo`800000H03oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@000000000M02oool80?ooo`030000003oool0
oooo01l0oooo0P3o000h0?ooo`<00000M`3oool005/0oooo0`00001O0?ooo`040?oog`2LB00004RL
0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0M0000000000004P0W=oo00T0oooo00<0
00000?ooo`3oool07P3oool00`3o0000oooo0?ooo`0i0?ooo`<00000MP3oool005/0oooo0`00001P
0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?ooo`0:
0?ooo`030000003oool0oooo01d0oooo00<0o`000?ooo`3oool0>`3oool3000007D0oooo001L0?oo
o`800000H@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool02P3o
ool00`000000oooo0?ooo`0L0?ooo`030?l0003oool0oooo03d0oooo0P00001e0?ooo`00G03oool3
00000640oooo00@0gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO00L0oooo
00<000000?ooo`3oool06`3oool00`3o0000oooo0?ooo`0n0?ooo`<00000M03oool005`0oooo0`00
001k0?ooo`030000003oool0oooo01X0oooo00<0o`000?ooo`3oool0@03oool3000007<0oooo001M
0?ooo`800000N`3oool3000001T0oooo00<0o`000?ooo`3oool0@@3oool3000007<0oooo001M0?oo
o`<00000NP3oool00`000000oooo0?ooo`0H0?ooo`030?l0003oool0oooo04<0oooo0`00001b0?oo
o`00GP3oool00`000000o`000?ooo`1i0?ooo`030000003oool0oooo01L0oooo00<0o`000?ooo`3o
ool0A@3oool300000740oooo001M0?ooo`<00000NP3oool00`000000oooo0?ooo`0F0?ooo`030?l0
003oool0oooo04L0oooo0P00001a0?ooo`00GP3oool3000007T0oooo00<000000?ooo`3oool05@3o
ool00`3o0000oooo0?ooo`180?ooo`<00000L03oool005l0oooo00<000000?ooo`3oool0N03oool0
0`000000oooo0?ooo`0D0?ooo`030?l0003oool0oooo04T0oooo0`00001`0?ooo`00GP3oool30000
07T0oooo00<000000?ooo`3oool04`3oool00`3o0000oooo0?ooo`1;0?ooo`<00000K`3oool005l0
oooo0`00001h0?ooo`030000003oool0oooo0180oooo00<0o`000?ooo`3oool0C@3oool3000006h0
oooo001P0?ooo`030000003oool0oooo07L0oooo0`00000A0?ooo`030?l0003oool0oooo04h0oooo
00<0o`00000000000000KP3oool005l0oooo0`00001h0?ooo`030000003oool0oooo0100oooo00<0
o`000?ooo`3oool0D03oool3000006d0oooo001P0?ooo`<00000M`3oool00`000000oooo0?ooo`0?
0?ooo`030?l0003oool0oooo0540oooo0`00001]0?ooo`00603oool00`3ookl0M000000000020000
000307Boo`3oool0oooo03X0oooo00<0gia80000000000001P00000h0?ooo`030?oo_`1d00000000
0080000000<0B9cO0?ooo`3oool0=`3oool00`000000oooo0?ooo`0?0?ooo`030?l0003oool0oooo
02P0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`0R0?ooo`<00000503oool00`3O
W4P00000000000030000000304RLg`3oool0oooo03T0oooo00<0oono07@0000000000P0000000`1d
_ol0oooo0?ooo`0>0?ooo`00603oool00`3o_g@00000001d_`020?ooo`030;mdB02Lgol0oooo03/0
oooo00<0okmd001d_`3oool00P3oool3000003X0oooo00<0okmd0000001d_ol0>P3oool00`000000
oooo0?ooo`0>0?ooo`030?l0003oool0oooo02/0oooo00<0okmd0000001d_ol09@3oool0103o0000
000000000000000D0?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool0>@3oool00`3o
_g@00000001d_`020?ooo`030;mdB02Lgol0oooo00h0oooo000M0?ooo`030?oOW018B9`0gooo03`0
oooo00<0okmd001d_`3oool00P3oool3000003T0oooo00<0oooO09a80000M;l0>P3oool00`000000
oooo0?ooo`0=0?ooo`030?l0003oool0oooo02`0oooo00<0oooO09a80000M;l09P3oool00`3o0000
00000?ooo`0E0?ooo`030?noM000M;l0oooo0480oooo00<0omnL04Q8W03Oool03P3oool001d0oooo
00<0oono07@0001d_ol0?@3oool0103o_g@007Bo0?ooo`3oool3000003X0oooo00<0gia80018W03O
ool0>@3oool00`000000oooo0?ooo`0<0?ooo`030?l0003oool0oooo02h0oooo00<0gia80018W03O
ool09@3oool3000001H0oooo00<0okmd001d_`3oool0@@3oool00`3ookl0M00007Boo`0>0?ooo`00
2`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00D0oooo00<0omnL04P0001d_ol0
;@3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0okmd04RLg`3oool0
0`00000Z0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B018
09cOo`0i0?ooo`030000003oool0oooo00/0oooo00<0o`000?ooo`3oool0;`3oool00`3ogi`0B018
09cOo`0V0?ooo`<000005P3oool00`3o_g@0B9cO0?ooo`100?ooo`030?oOW0180000M;oo00h0oooo
000L0?ooo`040?oOW0180000001809cOocl0oooo00<0gia804RLg`3oool00`00000i0?ooo`030?oo
_`1d0000M;oo03T0oooo0`00000:0?ooo`030?l0003oool0oooo0300oooo00<0oono07@0001d_ol0
9`3oool3000001H0oooo00<0gia804RLg`3oool0?P3oool0103ogi`0B0000000B02Lgol>0?ooo`00
6P3oool01@3ogi`0B000000000000000M;oo0440oooo00<0_g@007Boo`0000000P00000j0?ooo`03
0;md0000M;l0oooo03P0oooo00<000000?ooo`3oool02P3oool00`3o0000oooo0?ooo`0a0?ooo`03
0;md0000M;l0oooo02H0oooo00<0o`000000003oool05`3oool00`2oM000M;oo0?ooo`0k0?ooo`05
0?oOW018000000000000001d_ol03`3oool001`0oooo00<0oono07@00000M;l0@@3oool00`3ogi`0
B01d00000002000003X0oooo00<0gia80018W03Oool0>03oool00`000000oooo0?ooo`090?ooo`03
0?l0003oool0oooo0380oooo00<0gia80018W03Oool09P3oool3000001L0oooo00<0omnL04P0M02o
ool0?@3oool00`3ookl0M000001d_`0?0?ooo`007P3oool00`3OW4P007Bo0?ooo`0o0?ooo`030?oo
g`2LB000M;oo00<00000>@3oool00`3ogi`0B01809cOo`0h0?ooo`030000003oool0oooo00P0oooo
00<0o`000?ooo`3oool0<`3oool00`3ogi`0B01809cOo`0W0?ooo`<000005P3oool00`3ooml0W4P0
07Boo`0o0?ooo`030=nLB000M;l0oooo00d0oooo000N0?ooo`030?oOW01804P0W=oo03l0oooo00<0
oono07@00018W=l00`00000i0?ooo`030?oo_`1d0000M;oo03P0oooo00<000000?ooo`3oool01`3o
ool00`3o0000oooo0?ooo`0d0?ooo`030?oo_`1d0000M;oo02L0oooo0`00000F0?ooo`030?oo_`1d
0000B9cO03l0oooo00<0omnL04P0B02Lgol03@3oool001/0oooo00H0_g@007Boo`3oool0gia80000
B02Lgolk0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`03000003X0oooo00<0_g@0001d
_`3oool0=`3oool00`000000oooo0?ooo`060?ooo`030?l0003oool0oooo03H0oooo00<0_g@0001d
_`3oool09`3oool200000180oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo03`0oooo00H0
_g@007Boo`3oool0gia80000B02Lgol=0?ooo`00703oool0102oM000000000000018W=lm0?ooo`05
0?oOW018000000000000001d_ol00P3oool3000003L0oooo00D0oono07@000000000001d0;ooo`0g
0?ooo`030000003oool0oooo00D0oooo00<0o`000?ooo`3oool0=@3oool01@3ookl0M00000000000
07@0_ooo02L0oooo0`00000B0?ooo`050?oOW018000000000000001d_ol0?P3oool0102oM0000000
00000018W=l>0?ooo`00I03oool3000007<0oooo00<000000?ooo`3oool01@3oool00`3o0000oooo
0?ooo`1R0?ooo`<00000IP3oool006D0oooo0P00001c0?ooo`<00000103oool00`3o0000oooo0?oo
o`1S0?ooo`<00000IP3oool006D0oooo0`00001b0?ooo`030000003oool0oooo00<0oooo00<0o`00
0?ooo`3oool0I@3oool2000006H0oooo001U0?ooo`<00000LP3oool00`000000oooo0?ooo`020?oo
o`030?l0003oool0oooo06H0oooo0`00001U0?ooo`00IP3oool300000740oooo00D000000?ooo`3o
ool0oooo0?l0001Y0?ooo`<00000I@3oool006L0oooo00<000000?ooo`3oool0L03oool010000000
oooo0?ooo`3o001[0?ooo`800000I@3oool006H0oooo0`00001a0?ooo`040000003oool0oooo0?l0
06/0oooo0`00001T0?ooo`00I`3oool300000700oooo00<000000?ooo`3o0000K@3oool3000006<0
oooo001X0?ooo`030000003o0000oooo06l0oooo00<000000?l0003oool0K@3oool3000006<0oooo
001W0?ooo`<00000L03oool00`3o0000oooo0?ooo`1^0?ooo`800000H`3oool001H0oooo``000000
0`3o000000000000003300000140oooo000F0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00<000001`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00P0oooo00<0o`000?ooo`000000303oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo00D0oooo0`0000020?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00l0oooo000F0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool03P3oool2000002h0
oooo00<000000?ooo`3oool0?03oool00`3o0000oooo000000110?ooo`030000003oool0oooo02d0
oooo0P00000?0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool03`3oool006T0oooo
0`00001[0?ooo`040?l0003oool0oooo00000740oooo0`00001Q0?ooo`00JP3oool3000006T0oooo
00D0o`000?ooo`3oool0oooo0000001a0?ooo`<00000H@3oool006/0oooo00<000000?ooo`3oool0
I`3oool00`3o0000oooo0?ooo`020?ooo`030000003oool0oooo0700oooo0`00001P0?ooo`00JP3o
ool3000006L0oooo00<0o`000?ooo`3oool00`3oool00`000000oooo0?ooo`1a0?ooo`030000003o
ool0oooo05l0oooo001[0?ooo`<00000I@3oool00`3o0000oooo0?ooo`040?ooo`030000003oool0
oooo0700oooo0`00001P0?ooo`00J`3oool3000006D0oooo00<0o`000?ooo`3oool0103oool30000
0740oooo0`00001O0?ooo`00K03oool2000006@0oooo00<0o`000?ooo`3oool01@3oool00`000000
oooo0?ooo`1b0?ooo`800000G`3oool006`0oooo0`00001R0?ooo`030?l0003oool0oooo00H0oooo
00<000000?ooo`3oool0LP3oool3000005h0oooo001]0?ooo`<00000H03oool00`3o0000oooo0?oo
o`070?ooo`030000003oool0oooo0780oooo0`00001N0?ooo`00K@3oool3000005l0oooo00<0o`00
0?ooo`3oool0203oool00`000000oooo0?ooo`1c0?ooo`800000GP3oool006h0oooo0P00001O0?oo
o`030?l0003oool0oooo00P0oooo00<000000?ooo`3oool0L`3oool3000005d0oooo001^0?ooo`<0
0000G@3oool00`3o0000oooo0?ooo`090?ooo`030000003oool0oooo07<0oooo0`00001M0?ooo`00
KP3oool3000005`0oooo00<0o`000?ooo`3oool02P3oool00`000000oooo0?ooo`1d0?ooo`800000
G@3oool006l0oooo0`00001J0?ooo`030?l0003oool0oooo00/0oooo0`00001d0?ooo`<00000G03o
ool00700oooo0P00001I0?ooo`030?l0003oool0oooo00`0oooo00<000000?ooo`3oool0M03oool3
000005`0oooo001`0?ooo`<00000F03oool00`3o0000oooo0?ooo`0<0?ooo`030000003oool0oooo
07D0oooo0P00001L0?ooo`00L03oool3000005L0oooo00<0o`000?ooo`3oool03@3oool00`000000
oooo0?ooo`1e0?ooo`<00000F`3oool00740oooo0`00001E0?ooo`030?l0003oool0oooo00h0oooo
00<000000?ooo`3oool0M@3oool3000005/0oooo001b0?ooo`800000E03oool00`3o0000oooo0?oo
o`0?0?ooo`030000003oool0oooo07H0oooo0P00001K0?ooo`00LP3oool300000580oooo00<0o`00
0?ooo`3oool0403oool00`000000oooo0?ooo`1f0?ooo`<00000FP3oool00780oooo0`00001A0?oo
o`030?l0003oool0oooo0140oooo00<000000?ooo`3oool0MP3oool3000005X0oooo001c0?ooo`<0
0000D03oool00`3o0000oooo0?ooo`0A0?ooo`<00000M`3oool2000005X0oooo001d0?ooo`<00000
CP3oool00`3o0000oooo0?ooo`0B0?ooo`030000003oool0oooo07L0oooo0`00001I0?ooo`00M@3o
ool2000004d0oooo00<0o`000?ooo`3oool04`3oool00`000000oooo0?ooo`1g0?ooo`<00000F@3o
ool007D0oooo0`00001;0?ooo`030?l0003oool0oooo01@0oooo00<000000?ooo`3oool0N03oool2
000005T0oooo001e0?ooo`<00000BP3oool00`3o0000oooo0?ooo`0E0?ooo`030000003oool0oooo
07P0oooo0`00001H0?ooo`00MP3oool3000004P0oooo00<0o`000?ooo`3oool05P3oool00`000000
oooo0?ooo`1h0?ooo`<00000F03oool007L0oooo0`0000160?ooo`030?l0003oool0oooo01L0oooo
00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1G0?ooo`00N03oool3000004@0oooo00<0
o`000?ooo`3oool0603oool00`000000oooo0?ooo`1h0?ooo`<00000F03oool007T0oooo00<00000
0?ooo`3oool0@P3oool00`3o0000oooo0?ooo`0I0?ooo`<00000N@3oool3000005L0oooo001h0?oo
o`<00000@P3oool00`3o0000oooo0?ooo`0J0?ooo`030000003oool0oooo07X0oooo00<000000?oo
o`3oool0EP3oool007T0oooo0`0000100?ooo`030?l0003oool0oooo01/0oooo00<000000?ooo`3o
ool0N@3oool3000005L0oooo001j0?ooo`<00000?P3oool0103o0000oooo0?ooo`2oM00200000007
07Boo`3oool0oooo0?ooo`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000B9`0gooo
2P3oool00`000000oooo0?ooo`1j0?ooo`<00000EP3oool007/0oooo0`00000l0?ooo`090?l0003o
ool0oooo0=nLB000M;l0oooo0?oOW0180000B9cO0080oooo00<0omnL04P0001d_ol00P3oool01P3o
_g@0000007Boo`3oool0omnL04Qd_`X0oooo00<000000?ooo`3oool0N`3oool00`000000oooo0?oo
o`1E0?ooo`00O03oool3000003X0oooo00H0o`000?ooo`3oool0oooO09a80018W=l20?ooo`030?no
M00004P0W=oo00/0oooo00<0okmd04RLg`3oool0203oool00`000000oooo0?ooo`1j0?ooo`<00000
EP3oool007d0oooo0P00000h0?ooo`80o`000`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2L
B00004RL0=ooo`X0oooo00<0oono07@0M02oool0203oool00`000000oooo0?ooo`1k0?ooo`<00000
E@3oool007d0oooo0`00000f0?ooo`030?l0003oool0oooo00<0oooo00<0oooO09a80000M;l00`3o
ool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?ooo`3oool0O03o
ool00`000000oooo0?ooo`1D0?ooo`00O@3oool3000000030?l0003oool0oooo0380oooo00<0o`00
0?ooo`3oool01@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo00<0oono
07@00018W=l0203oool4000007X0oooo0`00001E0?ooo`00OP3oool3000000030?l0003oool0oooo
0300oooo00<0o`000?ooo`3oool01P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo
00X0oooo00<0okmd0000001d_ol0203oool00`000000oooo0?ooo`1l0?ooo`800000E@3oool007l0
oooo0`0000000`3o0000oooo0?ooo`0^0?ooo`030?l0003oool0oooo00L0oooo00<0omnL04P00018
W=l00`3oool00`2oM00004RL0=ooo`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo203oool0
0`000000oooo0?ooo`1l0?ooo`<00000E03oool00800oooo0`0000000`3o0000oooo0?ooo`0[0?oo
o`80o`002P3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono
07@000000000001809cOo`090?ooo`030000003oool0oooo07`0oooo0`00001D0?ooo`00P@3oool3
000000030?l0003oool0oooo02T0oooo00<0o`000?ooo`3oool02`3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3oool02P3oool00`000000oooo0?ooo`1m
0?ooo`800000E03oool00880oooo1000000Y0?ooo`030?l0003oool0oooo00d0oooo00H0gia80000
M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo00X0oooo00<000000?ooo`3oool0O@3o
ool3000005<0oooo00230?ooo`@000009P3oool20?l00140oooo00@0gia800000000000007Bo2P3o
ool01@2oM0000000000000000000B9cO00L0oooo00<000000?ooo`3oool0O@3oool3000005<0oooo
00250?ooo`<000008`3oool20?l002d0oooo00<000000?ooo`3oool0OP3oool2000005<0oooo0026
0?ooo`<0000000<0o`000?ooo`3oool07P3oool00`3o0000oooo0?ooo`0]0?ooo`<00000OP3oool3
00000580oooo00270?ooo`@0000000<0o`000?ooo`3oool06P3oool20?l00300oooo00<000000?oo
o`3oool0OP3oool300000580oooo00280?ooo`@0000000<0o`000?ooo`3oool05`3oool20?l00380
oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`1A0?ooo`00RP3oool300000080o`00
5@3oool20?l003@0oooo00<000000?ooo`3oool0OP3oool300000580oooo002;0?ooo`@000000`3o
000?0?ooo`<0o`00=P3oool00`000000oooo0?ooo`1o0?ooo`800000DP3oool008`0oooo1@000000
0`3oool0o`000?l0000=0?l003T0oooo00<000000?ooo`3oool0O`3oool300000540oooo002>0?oo
o`D00000A`3oool00`000000oooo0?ooo`1o0?ooo`<00000D@3oool00900oooo1@0000150?ooo`03
0000003oool0oooo0800oooo0P00001A0?ooo`00TP3oool4000000030?ooo`000000oooo0440oooo
0`0000200?ooo`<00000D03oool009@0oooo00<000000?ooo`0000000P0000000`3oool000000?oo
o`020?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?oo
o`1?0?ooo`00U`3oool00`000000oooo00000002000000040?ooo`0000000000000000<0oooo00<0
00000?ooo`3oool0=03oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo00@0oooo0`00
001@0?ooo`00VP3oool01@000000oooo0?ooo`3oool0000000<0oooo0`0000050?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`0h0?ooo`030000003o
ool0oooo03d0oooo0`0000060?ooo`030000003oool0oooo04l0oooo002S0?ooo`030000003oool0
oooo00<0oooo0`00000:0?ooo`<000006@3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo03L0oooo0`00000o0?ooo`030000003oool0oooo00D0oooo00<0o`000?ooo`3oool0C`3oool0
0:X0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0G0?ooo`<000001P3oool00`00
0000oooo0?ooo`0h0?ooo`030000003oool0oooo04L0oooo00<0o`000?ooo`3oool0CP3oool00=80
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`220?ooo`030?l0003oool0oooo04h0
oooo003J0?ooo`030000003oool0oooo0880oooo00<0o`000?ooo`3oool0CP3oool00=X0oooo0`00
00220?ooo`030?l0003oool0oooo04h0oooo003J0?ooo`030000003oool0oooo08<0oooo00<0o`00
0?ooo`3oool0C@3oool00=X0oooo00<000000?ooo`3oool0P`3oool00`3o0000oooo0?ooo`1=0?oo
o`00fP3oool00`000000oooo0?ooo`230?ooo`030?l0003oool0oooo04d0oooo003J0?ooo`030000
003oool0oooo08<0oooo00<0o`000?ooo`3oool0C@3oool00=X0oooo00<000000?ooo`3oool0Q03o
ool00`3o0000oooo0?ooo`1<0?ooo`00fP3oool00`000000oooo0?ooo`240?ooo`030?l0003oool0
oooo04`0oooo003J0?ooo`030000003oool0oooo08@0oooo00<0o`000?ooo`3oool0C03oool00=X0
oooo0`0000240?ooo`030?l0003oool0oooo04`0oooo003J0?ooo`030000003oool0oooo08D0oooo
00<0o`000?ooo`3oool0B`3oool00=X0oooo00<000000?ooo`3oool0Q@3oool00`3o0000oooo0?oo
o`1;0?ooo`00b@3oool00`3ookl0M000000000020000000304RLg`3oool0oooo00T0oooo00<00000
0?ooo`3oool0Q@3oool00`3o0000oooo0?ooo`1;0?ooo`00b`3oool00`3o_g@0000007Boo`0<0?oo
o`030000003oool0oooo08D0oooo00<0o`000?ooo`3oool0B`3oool00</0oooo00<0oooO09a80000
M;l0303oool00`000000oooo0?ooo`260?ooo`030?l0003oool0oooo04X0oooo003<0?ooo`030=nL
B000B9`0gooo00/0oooo00<000000?ooo`3oool0QP3oool00`3o0000oooo0?ooo`1:0?ooo`00c03o
ool00`3ogi`0B01809cOo`0;0?ooo`030000003oool0oooo08H0oooo00<0o`000?ooo`3oool0BP3o
ool00<`0oooo00<0oono07@0001d_ol02`3oool4000008D0oooo00<0o`000?ooo`3oool0BP3oool0
0<d0oooo00<0_g@0001d_`3oool02P3oool00`000000oooo0?ooo`260?ooo`030?l0003oool0oooo
04X0oooo003=0?ooo`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool0Q`3oool00`3o0000
oooo0?ooo`190?ooo`00c@3oool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo08L0oooo
00<0o`000?ooo`3oool0B@3oool00<d0oooo00<0oono07@0001d_ol02P3oool00`000000oooo0?oo
o`270?ooo`030?l0003oool0oooo04T0oooo003>0?ooo`030;md0000M;l0oooo00T0oooo00<00000
0?ooo`3oool0Q`3oool00`3o0000oooo0?ooo`190?ooo`00c03oool01@3ookl0M0000000000007@0
_ooo00T0oooo00<000000?ooo`3oool0R03oool00`3o0000oooo0?ooo`180?ooo`00fP3oool00`00
0000oooo0?ooo`280?ooo`030?l0003oool0oooo04P0oooo003J0?ooo`<00000R03oool00`3o0000
oooo0?ooo`180?ooo`00fP3oool00`000000oooo0?ooo`280?ooo`030?l0003oool0oooo04P0oooo
003J0?ooo`030000003oool0oooo08T0oooo00<0o`000?ooo`3oool0A`3oool00=X0oooo00<00000
0?ooo`3oool0R@3oool00`3o0000oooo0?ooo`170?ooo`00fP3oool00`000000oooo0?ooo`290?oo
o`030?l0003oool0oooo04L0oooo003J0?ooo`030000003oool0oooo08T0oooo00<0o`000?ooo`3o
ool0A`3oool00=X0oooo00<000000?ooo`3oool0RP3oool00`3o0000oooo0?ooo`160?ooo`00fP3o
ool00`000000oooo0?ooo`2:0?ooo`030?l0003oool0oooo04H0oooo003J0?ooo`<00000RP3oool0
0`3o0000oooo0?ooo`160?ooo`00fP3oool00`000000oooo0?ooo`2:0?ooo`030?l0003oool0oooo
04H0oooo003J0?ooo`030000003oool0oooo08/0oooo00<0o`000?ooo`3oool0A@3oool00=X0oooo
00<000000?ooo`3oool0R`3oool00`3o0000oooo0?ooo`150?ooo`00fP3oool00`000000oooo0?oo
o`2;0?ooo`030?l0003oool0oooo04D0oooo003J0?ooo`030000003oool0oooo08/0oooo00<0o`00
0?ooo`3oool0A@3oool00=X0oooo00<000000?ooo`3oool0S03oool00`3o0000oooo0?ooo`140?oo
o`00fP3oool00`000000oooo0?ooo`2<0?ooo`030?l0003oool0oooo04@0oooo003J0?ooo`030000
003oool0oooo08`0oooo00<0o`000?ooo`3oool0A03oool00=X0oooo0`00002<0?ooo`030?l0003o
ool0oooo04@0oooo003J0?ooo`030000003oool0oooo08`0oooo00<0o`000?ooo`3oool0A03oool0
0=X0oooo00<000000?ooo`3oool0S@3oool00`3o0000oooo0?ooo`130?ooo`00fP3oool00`000000
oooo0?ooo`2=0?ooo`030?l0003oool0oooo04<0oooo003J0?ooo`030000003oool0oooo08d0oooo
00<0o`000?ooo`3oool0@`3oool00=X0oooo00<000000?ooo`3oool0S@3oool00`3o0000oooo0?oo
o`130?ooo`00fP3oool00`000000oooo0?ooo`2=0?ooo`030?l0003oool0oooo04<0oooo003J0?oo
o`030000003oool0oooo08d0oooo00<0o`000?ooo`3oool0@`3oool00=X0oooo0`00002>0?ooo`03
0?l0003oool0oooo0480oooo003J0?ooo`030000003oool0oooo08h0oooo00<0o`000?ooo`3oool0
@P3oool00=X0oooo00<000000?ooo`3oool0SP3oool00`3o0000oooo0?ooo`120?ooo`00_03oool0
0`3ookl0M000000000020000000704RLg`3oool0oooo0?ooo`3ogi`0B00007Boo`020?ooo`060?oO
W0180000000000000000B9`0gooo2P3oool00`000000oooo0?ooo`2>0?ooo`030?l0003oool0oooo
0480oooo002n0?ooo`030?noM0000000M;oo00@0oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0
000007Boo`3oool0omnL04Qd_`X0oooo00<000000?ooo`3oool0SP3oool00`3o0000oooo0?ooo`12
0?ooo`00_P3oool00`3ooml0W4P0001d_`0>0?ooo`030?noM018W=l0oooo00P0oooo00<000000?oo
o`3oool0SP3oool00`3o0000oooo0?ooo`120?ooo`00_`3oool00`3OW4P004RL0=ooo`0=0?ooo`03
0?oo_`1d07@0_ooo00P0oooo00<000000?ooo`3oool0S`3oool00`3o0000oooo0?ooo`110?ooo`00
_`3oool00`3ogi`0B01809cOo`0=0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?ooo`3oool0
S`3oool00`3o0000oooo0?ooo`110?ooo`00_`3oool00`3ookl0M00007Boo`0=0?ooo`030?oo_`1d
0000B9cO00P0oooo1000002>0?ooo`030?l0003oool0oooo0440oooo00300?ooo`030;md0000M;l0
oooo00`0oooo00<0okmd0000001d_ol0gP3oool00<00oooo00<0gia80018W03Oool02@3oool01P3o
gml0gooo0?noM0000000001d0;ooomh0oooo00300?ooo`030?oOW01804P0W=oo00T0oooo00D0oono
07@000000000001809cOo`3O0?ooo`00`03oool00`3ookl0M00007Boo`0:0?ooo`030=nLW02oool0
oooo0>00oooo00310?ooo`030;md0000M;l0oooo00T0oooo00<0omnL07BLg`3oool0h03oool00;l0
oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l0g@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-3.34731, -1.64567, \
0.0229522, 0.0185688}}]
}, Closed]],

Cell[TextData[{
  "The limit cycle now resembles a slightly curved parallelogram.  On the \
left side, ",
  StyleBox["x",
    FontSlant->"Italic"],
  " being negative requires the ",
  StyleBox["y",
    FontSlant->"Italic"],
  " velocity to be positive so that the phase point moves slowly upward and \
remains near the null cline.  However, once the phase point passes the peak \
of the null cline the positive velocity brings it into a region where the \
small relaxation parameter strongly amplifies the horizontal velocity.  The \
system traverses the top segment very rapidly until it passes through the \
right side of the null cline where the large positive horizontal velocity \
abruptly changes to a small negative value.  Now that the phase point is \
below the null cline, the vertical velocity is negative and it remains small \
as long as the system can remain near the null cline.  However, once the \
phase point overshoots the bottom of the null cline, it once again enters a \
regime with large horizontal velocity.  The system races to the left until \
crossing the null cline again where it slowly retraces the left segment of \
the limit cycle.  The smaller the value of \[Alpha] the more tightly the \
trajectory adheres to the null cline in its vertical segments and the flatter \
are its horizontal segments."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData["Period for \[Alpha]\[Rule]0"], "ExerciseTitle"],

Cell[TextData[{
  "Deduce the period and amplitude of limit-cycle van der Pol oscillations in \
the limit \[Alpha]\[Rule]0.  What is the time scale?  [Hint: most of the time \
is spent where ",
  Cell[BoxData[
      \(TraditionalForm\`y[t] \[TildeTilde] f[x]\)]],
  ".  In this region we can write ",
  Cell[BoxData[
      \(TraditionalForm
      \`\[DifferentialD]t \[TildeTilde] 
        \(-\(\[PartialD]f\/\[PartialD]x\)\) \[DifferentialD]x\/x\)]],
  " and integrate over the relevant range of ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  ".]"
}], "ExerciseText"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell[CellGroupData[{

Cell["Period and amplitude for limit-cycle oscillations", "ExerciseTitle"],

Cell["\<\
a) Write a function which finds approximate values for the amplitude and \
period of a limit cycle.  Make sure that several oscillations are present and \
use late oscillations to be sure that transients have decayed.\
\>", "ExerciseText"],

Cell[TextData[{
  "b) Produce a plot which shows the dependence of the period for van der Pol \
limit-cycle oscillations upon the damping parameter ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\)]],
  "."
}], "ExerciseText"],

Cell[TextData[{
  "c) Produce a plot which shows the dependence of the amplitude for van der \
Pol limit-cycle oscillations upon the damping parameter ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\)]],
  "."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["BZ reaction", "ExerciseTitle"],

Cell["\<\
The Belousov-Zhabotinskii reaction is a chemical reaction that exhibits \
stable oscillations and wave phenomena.  A simplified model, proposed by \
Field and Noyes, represents this reaction in terms of a system of 3 \
differential equations\
\>", "ExerciseText"],

Cell[BoxData[
    FormBox[
      RowBox[{"\t\t", 
        RowBox[{
          RowBox[{
            RowBox[{"\[CurlyEpsilon]", " ", 
              RowBox[{
                SuperscriptBox["x", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], 
            "\[AlignmentMarker]", 
            "\[Equal]", \(\((1 - x[t])\)\ x[t] + q\ y[t] - x[t]\ y[t]\)}], 
          "\[IndentingNewLine]", "\t", 
          RowBox[{
            RowBox[{"\[Delta]", " ", 
              RowBox[{
                SuperscriptBox["y", "\[Prime]",
                  MultilineFunction->None], "[", "t", "]"}]}], 
            "\[AlignmentMarker]", 
            "\[Equal]", \(\(-q\)\ y[t] - x[t]\ y[t] + 2\ f\ z[t]\)}], "\n", 
          "\t", 
          RowBox[{
            RowBox[{
              SuperscriptBox["z", "\[Prime]",
                MultilineFunction->None], "[", "t", "]"}], 
            "\[AlignmentMarker]", "\[Equal]", \(x[t] - z[t]\)}]}]}], 
      TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->AlignmentMarker,
  TextJustification->0],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " represent the concentrations of malonic acid ",
  Cell[BoxData[
      StyleBox[\(HBrO\_2\),
        "InlineInput"]]],
  " and bromide ions ",
  Cell[BoxData[
      StyleBox[\(Br\^-\),
        "InlineInput"]]],
  ", while ",
  Cell[BoxData[
      \(TraditionalForm\`z\)]],
  " represents the ratio between the concentrations of the ",
  Cell[BoxData[
      StyleBox[\(Ce\^\(\(4\)\(+\)\)\),
        "InlineInput"]]],
  " and ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(Ce\^\(\(3\)\(+\)\)\),
          "InlineInput"], InputForm]]],
  " states of the cerium catalyst.  The parameters \[CurlyEpsilon], \[Delta], \
and ",
  Cell[BoxData[
      \(TraditionalForm\`q\)]],
  " are related to reaction rates while the parameter ",
  Cell[BoxData[
      \(TraditionalForm\`f\)]],
  " is stoichiometric.  Typical values of these parameters are ",
  Cell[BoxData[
      \(TraditionalForm\`\(\(\ \)\({f \[Rule] 0.5, \[Delta] \[Rule] 1\/3, 
        q \[Rule] 5\[Times]10\^\(-3\), \ \[CurlyEpsilon] \[Rule] 
          0.01}\)\)\)]],
  ".\na) Show that for the typical reaction parameters there is only one \
nontrivial fixed point that has positive values of the concentration \
variables.  Evaluate and characterize the eigenvalues of the stability matrix \
at this fixed point.\nb) Demonstrate that this system exhibits limit-cycle \
oscillations."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["General display functions", "ExerciseTitle"],

Cell["\<\
Any pair of autonomous first-order differential equations can be represented \
in the form\
\>", "ExerciseText"],

Cell[BoxData[{
    \(TraditionalForm\`\[PartialD]\_t x[t]\[AlignmentMarker] \[Equal] 
      v\_x[x, y, \[CurlyEpsilon]]\), "\n", 
    \(TraditionalForm\`\[PartialD]\_t y[t]\[AlignmentMarker] \[Equal] 
      v\_y[x, y, \[CurlyEpsilon]]\)}], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where \[CurlyEpsilon] is a parameter.  Develop ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions which produce the types of figures shown in this notebook given \
appropriate definitions for the velocity field.  Your functions should accept \
any definitions for ",
  Cell[BoxData[
      \(TraditionalForm\`{v\_x[x, y, \[CurlyEpsilon]], \ 
        v\_y[x, y, \[CurlyEpsilon]]}\)]],
  ", not just the van der Pol equation, but should be tested by reproducing \
some of the previous results for the van der Pol equation.  Then use your \
functions to study the behavior of the following equation."
}], "ExerciseText"],

Cell[BoxData[
    FormBox[
      RowBox[{
        FractionBox[\(\[PartialD]\^2 x[t]\), \(\[PartialD]t\^2\),
          MultilineFunction->None], "\[Equal]", 
        RowBox[{\(-x[t]\), "+", 
          RowBox[{"\[CurlyEpsilon]", " ", 
            RowBox[{"(", 
              RowBox[{"1", "-", 
                SuperscriptBox[
                  RowBox[{"(", 
                    FractionBox[\(\[PartialD]x[t]\), \(\[PartialD]t\),
                      MultilineFunction->None], ")"}], "2"]}], ")"}], " ", 
            FractionBox[\(\[PartialD]x[t]\), \(\[PartialD]t\),
              MultilineFunction->None]}]}]}], 
      TraditionalForm]], "DisplayFormula",
  Evaluatable->False,
  TextAlignment->Left,
  TextJustification->0]
}, Closed]],

Cell[CellGroupData[{

Cell["Effect of a pulse", "ExerciseTitle"],

Cell[TextData[{
  "Suppose that a van der Pol system experiences a force pulse of the form ",
  Cell[BoxData[
      \(TraditionalForm\`f[
          t] = \[Beta]\ Exp[\(-\((t - t\_0)\)\^2\)/\[Sigma]\^2]\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Beta]\)]],
  " is a small amplitude, \[Sigma] measures the duration, and ",
  Cell[BoxData[
      \(TraditionalForm\`t\_0\)]],
  " is the time of the peak force.  Study the effect of such a perturbation \
upon the motion.  How is the motion for ",
  Cell[BoxData[
      \(TraditionalForm\`\((t - t\_0)\)/\[Sigma] \[GreaterGreater] 1\)]],
  " affected?"
}], "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->{"RulerBar", "EditBar"},
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
PrintingCopies->1,
PrintingPageRange->{Automatic, Automatic},
ShowCellLabel->False,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"]},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Notation Package Styles", "Section",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell["\<\
The cells below define certain styles needed by the Notation package. These \
styles serve to make visible otherwise invisible tagboxes.\
\>", "Text",
    GeneratedCell->True,
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationTemplateStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 1, 0.850004],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationPatternWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[1, 0.900008, 0.979995],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationMadeBoxesWrapperStyle"],
    GeneratedCell->True,
    StyleMenuListing->None,
    Background->RGBColor[0.900008, 0.889998, 1],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"]
  }, Closed]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 88, 2, 81, "Title"],
Cell[1867, 57, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2347, 76, 31, 0, 83, "Section"],
Cell[2381, 78, 234, 5, 41, "Text"],
Cell[2618, 85, 478, 12, 64, "DisplayFormula",
  Evaluatable->False],
Cell[3099, 99, 1118, 19, 145, "Text"],
Cell[4220, 120, 148, 3, 41, "Text"],
Cell[4371, 125, 615, 17, 43, "DisplayFormula",
  Evaluatable->False],
Cell[4989, 144, 422, 12, 67, "Text"],
Cell[5414, 158, 249, 6, 69, "DisplayFormula",
  Evaluatable->False],
Cell[5666, 166, 569, 11, 93, "Text"],
Cell[6238, 179, 541, 12, 93, "Text"],
Cell[6782, 193, 199, 5, 44, "DisplayFormula",
  Evaluatable->False],
Cell[6984, 200, 378, 10, 88, "Text"],
Cell[7365, 212, 206, 5, 41, "DisplayFormula",
  Evaluatable->False],
Cell[7574, 219, 342, 10, 41, "Text"],
Cell[7919, 231, 265, 7, 65, "DisplayFormula",
  Evaluatable->False],
Cell[8187, 240, 156, 5, 41, "Text"],
Cell[8346, 247, 189, 5, 65, "DisplayFormula",
  Evaluatable->False],
Cell[8538, 254, 1643, 29, 223, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[10218, 288, 33, 0, 51, "Section"],
Cell[10254, 290, 221, 4, 125, "Input"],
Cell[10478, 296, 145, 4, 95, "Input"],
Cell[10626, 302, 174, 3, 125, "Input"],
Cell[10803, 307, 75, 1, 73, "Input"],
Cell[10881, 310, 679, 21, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[11597, 336, 44, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[11666, 340, 62, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[11753, 344, 130, 2, 99, "Input"],
Cell[11886, 348, 82, 1, 61, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[12017, 355, 40, 0, 68, "Subsection"],
Cell[12060, 357, 191, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[12276, 365, 144, 3, 73, "Input"],
Cell[12423, 370, 62, 1, 59, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[12534, 377, 48, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[12607, 381, 179, 4, 73, "Input"],
Cell[12789, 387, 69, 1, 59, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[12907, 394, 41, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[12973, 398, 75, 1, 73, "Input"],
Cell[13051, 401, 173, 3, 75, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[13261, 409, 175, 3, 73, "Input"],
Cell[13439, 414, 36229, 991, 279, 15083, 725, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[49683, 1408, 917, 19, 119, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[50649, 1433, 43, 0, 51, "Section"],
Cell[50695, 1435, 639, 14, 93, "Text"],

Cell[CellGroupData[{
Cell[51359, 1453, 46, 0, 68, "Subsection"],
Cell[51408, 1455, 305, 5, 67, "Text"],
Cell[51716, 1462, 185, 3, 99, "Input"],

Cell[CellGroupData[{
Cell[51926, 1469, 122, 2, 73, "Input"],
Cell[52051, 1473, 481, 13, 61, "Output"]
}, Closed]],
Cell[52547, 1489, 1000, 22, 99, "Input"],
Cell[53550, 1513, 285, 5, 99, "Input"],
Cell[53838, 1520, 234, 4, 99, "Input"],
Cell[54075, 1526, 258, 4, 99, "Input"],
Cell[54336, 1532, 274, 5, 99, "Input"],
Cell[54613, 1539, 359, 6, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[55009, 1550, 103, 2, 68, "Subsection"],

Cell[CellGroupData[{
Cell[55137, 1556, 65, 0, 55, "Subsubsection"],

Cell[CellGroupData[{
Cell[55227, 1560, 70, 1, 73, "Input"],
Cell[55300, 1563, 105292, 2579, 444, 36967, 1731, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[160629, 4147, 76, 1, 73, "Input"],
Cell[160708, 4150, 110021, 2766, 444, 40082, 1898, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[270766, 6921, 61, 1, 73, "Input"],
Cell[270830, 6924, 140127, 2611, 279, 20526, 1130, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[410972, 9538, 278, 4, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[411287, 9547, 66, 0, 55, "Subsubsection"],

Cell[CellGroupData[{
Cell[411378, 9551, 72, 1, 73, "Input"],
Cell[411453, 9554, 89104, 2371, 383, 35271, 1702, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[500594, 11930, 59, 1, 73, "Input"],
Cell[500656, 11933, 84481, 1748, 279, 17586, 918, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[585186, 13687, 64, 0, 55, "Subsubsection"],
Cell[585253, 13689, 98, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[585376, 13695, 61, 1, 73, "Input"],
Cell[585440, 13698, 38238, 1023, 279, 14415, 724, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[623715, 14726, 63, 1, 73, "Input"],
Cell[623781, 14729, 39340, 1049, 279, 14808, 742, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[663158, 15783, 72, 1, 73, "Input"],
Cell[663233, 15786, 92900, 2395, 444, 36544, 1695, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[756170, 18186, 71, 1, 73, "Input"],
Cell[756244, 18189, 42440, 1047, 279, 14583, 699, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[798699, 19239, 321, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[799045, 19250, 69, 1, 73, "Input"],
Cell[799117, 19253, 71660, 1675, 279, 19759, 1030, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[870814, 20933, 112, 2, 73, "Input"],
Cell[870929, 20937, 31135, 812, 279, 12499, 577, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[902079, 21752, 496, 9, 93, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[902624, 21767, 36, 0, 58, "ExerciseTitle"],
Cell[902663, 21769, 662, 13, 128, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[903362, 21787, 57, 0, 58, "ExerciseTitle"],
Cell[903422, 21789, 268, 5, 67, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[903727, 21799, 66, 0, 58, "ExerciseTitle"],
Cell[903796, 21801, 79, 0, 41, "ExerciseText"],
Cell[903878, 21803, 432, 11, 64, "DisplayFormula",
  Evaluatable->False],
Cell[904313, 21816, 485, 11, 67, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[904835, 21832, 41, 0, 58, "ExerciseTitle"],
Cell[904879, 21834, 66, 0, 41, "ExerciseText"],
Cell[904948, 21836, 490, 12, 64, "DisplayFormula",
  Evaluatable->False],
Cell[905441, 21850, 130, 6, 41, "ExerciseText"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[905620, 21862, 41, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[905686, 21866, 34, 0, 68, "Subsection"],
Cell[905723, 21868, 429, 8, 67, "Text"],
Cell[906155, 21878, 553, 15, 61, "DisplayFormula",
  Evaluatable->False],
Cell[906711, 21895, 512, 13, 67, "Text"],
Cell[907226, 21910, 451, 9, 108, "DisplayFormula",
  Evaluatable->False],
Cell[907680, 21921, 357, 9, 67, "Text"],
Cell[908040, 21932, 994, 24, 93, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[909071, 21961, 71, 0, 68, "Subsection"],
Cell[909145, 21963, 216, 4, 41, "Text"],
Cell[909364, 21969, 180, 3, 73, "Input"],
Cell[909547, 21974, 148, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[909720, 21983, 123, 3, 96, "Input"],
Cell[909846, 21988, 481, 13, 80, "Output"]
}, Closed]],
Cell[910342, 22004, 244, 7, 41, "Text"],

Cell[CellGroupData[{
Cell[910611, 22015, 474, 8, 74, "Input"],
Cell[911088, 22025, 548, 13, 61, "Output"]
}, Closed]],
Cell[911651, 22041, 349, 5, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[912037, 22051, 71, 0, 68, "Subsection"],
Cell[912111, 22053, 378, 6, 67, "Text"],
Cell[912492, 22061, 892, 17, 295, "Input"],
Cell[913387, 22080, 856, 17, 255, "Input"],
Cell[914246, 22099, 306, 4, 67, "Text"],

Cell[CellGroupData[{
Cell[914577, 22107, 163, 2, 96, "Input"],
Cell[914743, 22111, 92454, 2842, 279, 52515, 2344, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[1007234, 24958, 189, 3, 96, "Input"],
Cell[1007426, 24963, 48198, 1330, 279, 20494, 983, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[1055639, 26296, 1344, 23, 197, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[1057020, 26324, 62, 0, 58, "ExerciseTitle"],
Cell[1057085, 26326, 574, 15, 72, "ExerciseText"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[1057708, 26347, 33, 0, 54, "ExerciseMain"],

Cell[CellGroupData[{
Cell[1057766, 26351, 74, 0, 58, "ExerciseTitle"],
Cell[1057843, 26353, 249, 4, 67, "ExerciseText"],
Cell[1058095, 26359, 238, 6, 41, "ExerciseText"],
Cell[1058336, 26367, 241, 6, 41, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[1058614, 26378, 36, 0, 58, "ExerciseTitle"],
Cell[1058653, 26380, 273, 5, 67, "ExerciseText"],
Cell[1058929, 26387, 1076, 28, 91, "DisplayFormula",
  Evaluatable->False],
Cell[1060008, 26417, 1502, 44, 192, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[1061547, 26466, 50, 0, 58, "ExerciseTitle"],
Cell[1061600, 26468, 122, 3, 41, "ExerciseText"],
Cell[1061725, 26473, 326, 7, 69, "DisplayFormula",
  Evaluatable->False],
Cell[1062054, 26482, 653, 13, 93, "ExerciseText"],
Cell[1062710, 26497, 724, 18, 66, "DisplayFormula",
  Evaluatable->False]
}, Closed]],

Cell[CellGroupData[{
Cell[1063471, 26520, 42, 0, 58, "ExerciseTitle"],
Cell[1063516, 26522, 635, 16, 67, "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

