(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   1116359,      29916]*)
(*NotebookOutlinePosition[   1154319,      31242]*)
(*  CellTagsIndexPosition[   1154181,      31235]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Introduction to Data Analysis", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Off[General::spell, General::spell1]\)}], "Input"],

Cell[BoxData[{
    \(\($DefaultFont = {"\<Times\>", 12};\)\), "\n", 
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", FontSize \[Rule] 12, 
          FontSlant \[Rule] "\<Italic\>"};\)\)}], "Input"],

Cell[BoxData[{
    \(\(Needs["\<Graphics`\>"];\)\), "\n", 
    \(\(Needs["\<Utilities`Notation`\>"];\)\), "\[IndentingNewLine]", 
    \(\(Needs["\<Statistics`\>"];\)\)}], "Input"],

Cell[BoxData[{
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Sigma]\_x\),
          NotationBoxTag,
          TagStyle->"NotationTemplateStyle"], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Sigma]\_y\),
          NotationBoxTag,
          TagStyle->"NotationTemplateStyle"], "]"}], 
      ";"}], "\[IndentingNewLine]", 
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(S\_x\),
          NotationBoxTag,
          TagStyle->"NotationTemplateStyle"], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(S\_y\),
          NotationBoxTag,
          TagStyle->"NotationTemplateStyle"], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(S\_\(x, x\)\),
          NotationBoxTag,
          TagStyle->"NotationTemplateStyle"], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(S\_\(x, y\)\),
          NotationBoxTag,
          TagStyle->"NotationTemplateStyle"], "]"}], ";"}]}], "Input"],

Cell[BoxData[
    RowBox[{"Symbolize", "[", 
      TagBox[\(y\&_\),
        NotationBoxTag,
        TagStyle->"NotationTemplateStyle"], "]"}]], "Input"],

Cell[TextData[{
  "It is also useful to append the directory containing this notebook and \
related files to ",
  StyleBox["$Path", "InlineInput"],
  "."
}], "Text"],

Cell[BoxData[
    \(\(AppendTo[$Path, "\<C:\Documents and Settings\Jim\My Documents\My \
eBooks\EssentialMathematica\Data analysis\>"];\)\)], "Input"],

Cell["\<\
You will need to adjust this directory name for your computer.\
\>", "Text",
  FontSize->14,
  FontColor->RGBColor[1, 0, 0]]
}, Closed]],

Cell[CellGroupData[{

Cell["Principle of maximum likelihood", "Section"],

Cell[TextData[{
  "\tLet ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " represent an independent variable, such as an applied voltage that we can \
vary freely and measure accurately, and let ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " represent the dependent variable, such as the current in a circuit.  We \
often think of the dependent variable as the response of a system to a \
variable that is controlled externally.  We usually have less control over \
the response and measure the dependent variable less accurately than the \
independent variable; therefore, we assume that the random errors, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\_x\)]],
  ", are negligible for the independent variable and consider only random \
errors, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\_y\)]],
  ", in the dependent variable.  Our data set then consists of the set of ",
  Cell[BoxData[
      \(TraditionalForm\`N\)]],
  " measurements ",
  Cell[BoxData[
      \(TraditionalForm\`{{x\_1, y\_1, \[Sigma]\_1}, {x\_2, 
          y\_2, \[Sigma]\_2}, \[CenterEllipsis], {x\_N, 
          y\_N, \[Sigma]\_N}}\)]],
  "."
}], "Text"],

Cell["\<\
\tWe imagine that our data set is merely one outcome out of an infinite \
number of possible outcomes and that there exists a parent probability \
function which determines the likelihood of our particular data set.  \
Furthermore, we assume that each measurement is independent such that the \
joint probability for the entire data set\
\>", "Text"],

Cell[BoxData[
    \(TraditionalForm\`P[{y\_i}]\  = \ \[Product]\+\(i = 1\)\%N P\_i[
          y\_i]\)], "DisplayFormula"],

Cell[TextData[{
  "factors into a product of probabilities for each individual measurement.  \
",
  "For counting experiments we can be pretty sure that ",
  Cell[BoxData[
      \(TraditionalForm\`P\_i\)]],
  " should be a Poisson distribution, but usually we do not know the proper \
",
  Cell[BoxData[
      \(TraditionalForm\`P\_i\)]],
  " functions very accurately.  In the absence of additional information, we \
usually assume that ",
  Cell[BoxData[
      \(TraditionalForm\`P\_i\)]],
  " is a normal or Gaussian distribution of the form"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`P\_i[y\_i]\  \[Proportional] \ \(\[Sigma]\_i\^\(-1\)\) 
          Exp[\(-\((\(y\_i - y[x\_i; a]\)\/\[Sigma]\_i)\)\^2\)]\ \ \ \
\[DoubleLongRightArrow]\ \ \ Log[
              P]\  = \ \(-\(\[Sum]\+\(i = 1\)\%N\((\(y\_i - y[x\_i; a]\)\/\
\[Sigma]\_i)\)\^2\)\)\  + \ constant\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`y[x; a]\)]],
  " is a ",
  StyleBox["fitting function ",
    FontSlant->"Italic"],
  "that depends upon the independent variable ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and a set of ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " parameters ",
  Cell[BoxData[
      \(TraditionalForm\`a = {a\_\[Mu], \[Mu] = 1, m}\)]],
  " that represent the properties of the system that we seek to determine \
from our data set.  Such a fitting function is based upon a physical model of \
the system.  Unless one is paranoid, it is reasonable to assume that our data \
set is close to the most likely result for an experiment of this type.  \
Therefore, the ",
  StyleBox["principle of maximum likelihood",
    FontSlant->"Italic"],
  " claims that the best estimates of the model parameters are those that \
maximize the joint probability for our data set.  Because Log is a monotonic \
function, maximizing ",
  Cell[BoxData[
      \(TraditionalForm\`Log[P]\)]],
  " is equivalent to maximizing ",
  Cell[BoxData[
      \(TraditionalForm\`P\)]],
  ", and is much simpler.  Thus, one must solve the system of equations"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[PartialD]Log[P]\/\[PartialD]a\_\[Mu] \[Equal] 
      0\ \ \[DoubleLongRightArrow]\ \ \ \(\[Sum]\+\(i = 1\)\%N\(\[PartialD]y[
                    x\_i; a]\/\[PartialD]a\_\[Mu]\) \(y\_i - y[x\_i; a]\)\/\
\[Sigma]\_i\^2\)\  \[Equal] \ 0\)], "DisplayFormula"],

Cell[TextData[{
  "for the parameters that maximize the likelihood.  These solutions are \
denoted ",
  Cell[BoxData[
      \(TraditionalForm\`a\&^ = {a\&^\_\[Mu]}\)]],
  " and are described as ",
  StyleBox["maximum-likelihood estimators",
    FontSlant->"Italic"],
  " of the model parameters.  To minimize confusion, it is useful to employ \
latin indices (",
  Cell[BoxData[
      \(TraditionalForm\`i, j\)]],
  ") to enumerate data points and greek indices (",
  Cell[BoxData[
      \(TraditionalForm\`\[Mu], \[Nu]\)]],
  ") to enumerate model parameters."
}], "Text"],

Cell[TextData[{
  "\tThe method of maximum likelihood is quite general and can be used for a \
wide variety of sample probability distributions ",
  Cell[BoxData[
      \(TraditionalForm\`P\_i\)]],
  ".  When these are assumed to be Gaussian, the most common situation, \
maximizing ",
  Cell[BoxData[
      \(TraditionalForm\`Log[P]\)]],
  " becomes equivalent to minimizing ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Chi]\^2\)[a]\)]],
  " where"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(\[Chi]\^2\)[
        a]\  = \ \[Sum]\+\(i = 1\)\%N\((\(y\_i - y[x\_i; a]\)\/\[Sigma]\_i)\)\
\^2\)], "DisplayFormula"],

Cell[TextData[{
  "is a weighted sum of the squares of the difference between observed and \
fitted quantities.  Thus, for normal distributions we use the ",
  StyleBox["method of least-squares",
    FontSlant->"Italic"],
  ".  Although not strictly correct, one often uses least squares to fit data \
based upon the Poisson distribution also."
}], "Text"],

Cell[TextData[{
  "\tIt is important to recognize that the values we obtain for ",
  Cell[BoxData[
      \(TraditionalForm\`{a\&^\_\[Mu]}\)]],
  " are merely estimates of the physical parameters based upon a particular \
finite set of measurements.  We expect that these estimates will become \
increasingly accurate as the number and/or precision of the measurements \
increases, but we can never be absolutely certain that measured values are \
exactly equal to true values.  Therefore, we describe the fitted parameters \
as maximum-likelihood estimators and often adorn the symbols ",
  Cell[BoxData[
      \(TraditionalForm\`{a\&^\_\[Mu]}\)]],
  " with carets to distinguish estimators from true parameters of the \
physical system.  The estimators are functions of measured quantities that \
are subject to random errors and, consequently, are themselves subject to \
random errors.  If we repeat the experiment many times, with each repetition \
containing the same sequence of measurements, we will obtain a distribution \
of results for each of the estimators.  Nevertheless, estimates of the \
accuracy of the parameters can be obtained from the curvature of ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Chi]\^2\)[a]\)]],
  " near its minimum.  If the minimum is deep and the curvature with respect \
to ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Mu]\)]],
  " is large, then parameter ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Mu]\)]],
  " is determined well by the data but if the curvature is small the \
uncertainty in that parameter will be large. "
}], "Text"],

Cell[TextData[{
  "\tThe quality of the fit is gauged by the ",
  StyleBox["reduced chi-square",
    FontSlant->"Italic"]
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Chi]\_red\%2\  = \ \(\(1\/\(N - m\)\) \(\[Chi]\^2\)[
            a\&^]\  = \ \(1\/\(N - 
                m\)\) \(\[Sum]\+\(i = 1\)\%N\((\(y\_i - y[x\_i; a\&^]\)\/\
\[Sigma]\_i)\)\^2\)\)\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " is the number of parameters and ",
  Cell[BoxData[
      \(TraditionalForm\`N - m\)]],
  " is the number of degrees of freedom.  If the model is reliable and our \
estimates of the measurement uncertainties are accurate, we expect to find ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\_red\%2 \[TildeTilde] 1\)]],
  " for a good fit.  However, this statistic is a random variable with its \
own distribution \[LongDash] the larger the data set the narrower this \
distribution becomes."
}], "Text"],

Cell["\<\
\tMaximum likelihood methods are quite versatile and can be applied to \
problems in which each measurement involves several dependent variables and \
produces several independent variables.  The probability distributions for \
each measurement can also be much more sophisticated.  It may be useful to \
apply constraints upon the fitted parameters.  Naturally, the most efficient \
methods for solving these maximum-likelihood problems depend upon the nature \
of the data set and the fitting function.  In the next few sections we \
consider increasingly sophisticated methods, starting with the simplest case \
of a linear function.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Linear least squares", "Section"],

Cell[CellGroupData[{

Cell[TextData[{
  "Two-parameter model, linear wrt ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]]
}], "Subsection"],

Cell[TextData[{
  "\tSuppose that our physical model of the system suggests that ",
  Cell[BoxData[
      \(TraditionalForm\`y[x] = a\_0 + a\_1\ x\)]],
  " is a linear function of two parameters ",
  Cell[BoxData[
      \(TraditionalForm\`{a\_0, a\_1}\)]],
  ".  Note that in interests of streamlining the notation, it is customary to \
leave the dependence of the fitting function upon the model parameters \
implicit.  The most important feature of this model is that the fitting \
function is linear with respect to the properties of the system that we seek \
to measure, namely the parameters ",
  Cell[BoxData[
      \(TraditionalForm\`{a\_0, a\_1}\)]],
  ".  The fact that it is also linear with respect to ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " simplifies some of the algebra in this section, but is not an essential \
feature of the method of linear least squares."
}], "Text"],

Cell[TextData[{
  "\tMaximizing the probability provides the system of ",
  StyleBox["normal equations",
    FontSlant->"Italic"]
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[PartialD]\[Chi]\^2\/\[PartialD]a\_\[Mu] \[Equal] 
      0\ \ \[DoubleLongRightArrow]\ \ \ \(\[Sum]\+\(i = 1\)\%N\(\[PartialD]y[
                    x\_i]\/\[PartialD]a\_\[Mu]\) \(y\_i - y[x\_i]\)\/\[Sigma]\
\_i\^2\)\  \[Equal] \ 0\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Mu] \[Element] {0, 1}\)]],
  " for our two-parameter model.  Thus,"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(\(\ \ \)\(\[Sum]\+\(i = 1\)\%N\( y\_i - a\_0 - \
\(a\_1\) x\_i\)\/\[Sigma]\_i\^2\  \[Equal] \ 
      0\ \ \[DoubleLongRightArrow]\ \ \ S\_y - \(a\_0\) W - \(a\_1\) 
          S\_x\  \[Equal] \ 0\[IndentingNewLine]
    \ \ \[Sum]\+\(i = 1\)\%N\( 
              x\_i\) \(y\_i - a\_0 - \(a\_1\) x\_i\)\/\[Sigma]\_i\^2\  \
\[Equal] \ 
      0\ \ \[DoubleLongRightArrow]\ \ \ S\_\(x, y\) - \(a\_0\) 
          S\_x - \(a\_1\) S\_\(x, x\)\  \[Equal] \ 0\)\)\)], "DisplayFormula"],

Cell["where", "Text"],

Cell[BoxData[{
    \(TraditionalForm\`S\_x\  = \ \[Sum]\+\(i = 1\)\%N\( w\_i\) 
          x\_i\), "\[IndentingNewLine]", 
    \(TraditionalForm\`S\_y\  = \ \[Sum]\+\(i = 1\)\%N\( w\_i\) 
          y\_i\), "\[IndentingNewLine]", 
    \(TraditionalForm\`S\_\(x, x\)\  = \ \[Sum]\+\(i = 1\)\%N\( w\_i\) 
          x\_i\^2\), "\[IndentingNewLine]", 
    \(TraditionalForm\`S\_\(x, y\)\  = \ \[Sum]\+\(i = 1\)\%N\( 
              w\_i\) \(x\_i\) y\_i\), "\[IndentingNewLine]", 
    \(TraditionalForm\`W = \[Sum]\+\(i = 1\)\%N w\_i\)}], "DisplayFormula"],

Cell[TextData[{
  "are expressed in terms of the weights ",
  Cell[BoxData[
      \(TraditionalForm\`w\_i = 1/\[Sigma]\_i\^2\)]],
  ".  Therefore, we obtain the ",
  StyleBox["maximum-likelihood estimators",
    FontSlant->"Italic"],
  " "
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`a\&^\_0\  = \ \(\(S\_y\) S\_\(x, x\) - \(S\_x\) \
S\_\(x, y\)\)\/\[CapitalDelta]\ , \ \ \ \ a\&^\_1\  = \ \(W\ S\_\(x, y\) - \
\(S\_x\) S\_y\)\/\[CapitalDelta]\)], "DisplayFormula"],

Cell["where", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[CapitalDelta]\  = \ 
      W\ S\_\(x, x\)\  - \ S\_x\%2\)], "DisplayFormula"],

Cell[TextData[{
  "is the discriminant for this pair of equations.  These are simple formulas \
that can easily be evaluated using a spreadsheet, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", or other software.  They can even be evaluated by a hand calculator or \
(gasp!) using pencil and paper if ",
  Cell[BoxData[
      \(TraditionalForm\`N\)]],
  " is not large. "
}], "Text"],

Cell[TextData[{
  "\tThe uncertainty in a derived quantity, such as ",
  Cell[BoxData[
      \(TraditionalForm\`a\&^\_0\)]],
  ", can be estimated using our familiar techniques for error propagation.  \
We assume that the measurements are independent and write"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][a\&^\_0]\^2\  = \ \[Sum]\+\(i = 1\)\%N\((\(\
\[PartialD]a\&^\_0\/\[PartialD]y\_i\) \[Sigma]\_i)\)\^2\ , \ \ \ \ \[Sigma][a\
\&^\_1]\^2\  = \ \[Sum]\+\(i = 1\)\%N\((\(\[PartialD]a\&^\_1\/\[PartialD]y\_i\
\) \[Sigma]\_i)\)\^2\)], "DisplayFormula"],

Cell[TextData[{
  "where the contribution of the uncertainty in measurement ",
  Cell[BoxData[
      \(TraditionalForm\`y\_i\)]],
  " is determined by the variation of ",
  Cell[BoxData[
      \(TraditionalForm\`a\&^\_\[Mu]\)]],
  " with respect to ",
  Cell[BoxData[
      \(TraditionalForm\`y\_i\)]],
  ".  Thus, substituting"
}], "Text"],

Cell[BoxData[{
    \(TraditionalForm\`\[PartialD]a\&^\_0\/\[PartialD]y\_i\  = \ \(\(\
\[CapitalDelta]\^\(-1\)\)(\(S\_\(x, 
                  x\)\) \[PartialD]S\_y\/\[PartialD]y\_i - \(S\_x\) \
\[PartialD]S\_\(x, y\)\/\[PartialD]y\_i)\ \  = \ \(w\_i\/\[CapitalDelta]\) \
\((S\_\(x, x\) - \(S\_x\) x\_i)\)\)\), "\[IndentingNewLine]", 
    \(TraditionalForm\`\[PartialD]a\&^\_1\/\[PartialD]y\_i\  = \ \(\(\
\[CapitalDelta]\^\(-1\)\)(
          W\ \[PartialD]S\_\(x, y\)\/\[PartialD]y\_i - \(S\_x\) \
\[PartialD]S\_y\/\[PartialD]y\_i)\ \  = \ \(w\_i\/\[CapitalDelta]\) \((W\ \
x\_i - S\_x)\)\)\)}], "DisplayFormula"],

Cell["into", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][a\&^\_0]\^2\  = \ \(\[CapitalDelta]\^\(-2\)\
\ \ \(\[Sum]\+\(i = 1\)\%N\(\( w\_i\)(S\_\(x, x\) - \(S\_x\) x\_i)\)\^2\)\  = \
\ \(\[CapitalDelta]\^\(-2\)\ \ \(\[Sum]\+\(i = 1\)\%N\( w\_i\)(
                S\_\(x, x\)\^2 - 
                  2 \( S\_x\) \(S\_\(x, x\)\) x\_i + \(S\_x\^2\) 
                    x\_i\^2)\)\  = \ \(\(S\_\(x, x\)\)(W\ S\_\(x, x\) - \
S\_x\^2)\)\/\[CapitalDelta]\^2\)\)\)], "DisplayFormula"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][a\&^\_1]\^2\  = \ \(\[CapitalDelta]\^\(-2\)\
\ \ \(\[Sum]\+\(i = 1\)\%N\(\( w\_i\)(W\ x\_i - S\_x)\)\^2\)\  = \ \(\
\[CapitalDelta]\^\(-2\)\ \ \(\[Sum]\+\(i = 1\)\%N\( w\_i\)(\(W\^2\) x\_i\^2 - 
                  2  W\ \(S\_x\) x\_i + 
                  S\_x\^2)\)\  = \ \(W(W\ S\_\(x, x\) - S\_x\^2)\)\/\
\[CapitalDelta]\^2\)\)\)], "DisplayFormula"],

Cell["we obtain", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][a\&^\_0]\^2\  = \ 
      S\_\(x, x\)\/\[CapitalDelta]\ , \ \ \ \ \ \ \[Sigma][a\&^\_1]\^2\  = \ 
      W\/\[CapitalDelta]\)], "DisplayFormula"],

Cell["relatively easily.  ", "Text"],

Cell[TextData[{
  "\tOften we do not know ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\_i\)]],
  " but believe that the uncertainties are pretty uniform.  Because the \
weights appear linearly in both numerator and denominator, only relative \
weights affect the fitted values for the parameter; hence, we can set all ",
  Cell[BoxData[
      \(TraditionalForm\`w\_i \[Rule] 1\)]],
  " if the weights are uniform.  The measurement uncertainties can then be \
estimated using the standard deviation of the data from the model, according \
to "
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma]\_y\^2 = \ \(1\/\(N - 
              2\)\) \(\[Sum]\+\(i = 1\)\%N\((y\_i - y\&_\_i)\)\^2\)\)], \
"DisplayFormula"],

Cell[TextData[{
  "where the expected or ",
  StyleBox["fit values",
    FontSlant->"Italic"]
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`y\&_\_i\  = \ y[x\_i; a\&^]\)], "DisplayFormula"],

Cell[TextData[{
  "are based upon our best estimates of the model parameters and the number \
of ",
  StyleBox["degrees of freedom",
    FontSlant->"Italic"],
  " is reduced by the number of parameters fit to the data, here  two.  Under \
these circumstances the uncertainties in fitted parameters become"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`w\_i \[Rule] 
        1\ \ \ \ \[DoubleLongRightArrow]\ \ \ \[Sigma][
            a\&^\_0]\  = \ \(\[Sigma]\_y\) \@\(S\_\(x, x\)\/\[CapitalDelta]\)\
\ , \ \ \ \ \ \ \[Sigma][
        a\&^\_1]\  = \ \(\[Sigma]\_y\) \@\(N\/\[CapitalDelta]\)\)], \
"DisplayFormula"],

Cell[TextData[{
  "The quality of the fit should be judged graphically.  If the data scatter \
fairly symmetrically about the fitted line, the model is probably reliable \
and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\_y\)]],
  " provides an estimate of the quality of the data.  However, if the scatter \
is not uniform the assumption of a linear model becomes dubious."
}], "Text"],

Cell[CellGroupData[{

Cell["example: unweighted linear fit", "Subsubsection"],

Cell[TextData[{
  "\tTo illustrate this procedure, we construct a data set with uniformly \
spaced values of the independent variable and values of the dependent \
variable based upon a linear function with random fluctuations drawn from a \
normal distribution with prescribed ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  ".  Pretending that we do not know ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  ", the weights are initialized to unity."
}], "Text"],

Cell[BoxData[{
    \(\(x = Range[\(-0.9\), 0.9, 0.1];\)\), "\[IndentingNewLine]", 
    \(\(y = 
        117\  - \ 23  x + 
          Table[Random[
              NormalDistribution[0, 2. ]], {Length[
                x]}];\)\), "\[IndentingNewLine]", 
    \(\(w = Table[1, {Length[x]}];\)\)}], "Input"],

Cell[TextData[{
  "Next we evaluate the sums defined above using the listability of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions"
}], "Text"],

Cell[BoxData[{
    \(\(S\_x = w . x;\)\), "\[IndentingNewLine]", 
    \(\(S\_y = w . y;\)\), "\[IndentingNewLine]", 
    \(\(S\_\(x, x\) = w . x\^2;\)\), "\[IndentingNewLine]", 
    \(\(S\_\(x, y\) = w . \((x*y)\);\)\), "\[IndentingNewLine]", 
    \(\(W = Plus @@ w;\)\), "\[IndentingNewLine]", 
    \(\(\[CapitalDelta] = W\ S\_\(x, x\) - S\_x\^2;\)\)}], "Input"],

Cell["and evaluate the estimators and their uncertainties.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({a\_0, 
        a\_1} = {\(\(S\_y\) S\_\(x, x\) - \(S\_x\) S\_\(x, y\)\)\/\
\[CapitalDelta], \ \(W\ S\_\(x, y\) - \(S\_x\) S\_y\)\/\[CapitalDelta]} // 
        N\)], "Input"],

Cell[BoxData[
    \({118.03652247715455`, \(-23.431675964936083`\)}\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(\(y\&_ = a\_0 + \(a\_1\) x;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Sigma]\_y\  = \ 
      Sqrt[Plus @@ \(\((y - y\&_)\)\^2\)/\((Length[y] - 2)\)]\)], "Input"],

Cell[BoxData[
    \(1.9097890074360775`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \({\[Sigma]\_0, \[Sigma]\_1} = \(\[Sigma]\_y\) 
          Sqrt /@ {S\_\(x, x\)\/\[CapitalDelta], \ \(\(W\)\(\ \)\)\/\
\[CapitalDelta]} // N\)], "Input"],

Cell[BoxData[
    \({0.4381356466788795`, 0.799922589777119`}\)], "Output"]
}, Closed]],

Cell["Finally, we compare the fit with the data.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[\[IndentingNewLine]ListPlot[
          Evaluate[Transpose[{x, y}]], 
          PlotStyle \[Rule] PointSize[0.02]], \[IndentingNewLine]Plot[
          a\_0 + \(a\_1\) x, {x, \(-1\), 1}], Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 -1.17352 0.01256 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .08248 -22 -7.84375 ]
[-0.0125 .08248 0 7.84375 ]
[-0.0125 .20808 -22 -7.84375 ]
[-0.0125 .20808 0 7.84375 ]
[-0.0125 .33368 -22 -7.84375 ]
[-0.0125 .33368 0 7.84375 ]
[-0.0125 .45928 -22 -7.84375 ]
[-0.0125 .45928 0 7.84375 ]
[-0.0125 .58488 -22 -7.84375 ]
[-0.0125 .58488 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -37.6875 -5.09375 ]
[-0.01875 .30902 -22 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .08248 m
.00625 .08248 L
s
gsave
-0.0125 .08248 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
0 .20808 m
.00625 .20808 L
s
gsave
-0.0125 .20808 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(110) show
1.000 setlinewidth
grestore
0 .33368 m
.00625 .33368 L
s
gsave
-0.0125 .33368 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(120) show
1.000 setlinewidth
grestore
0 .45928 m
.00625 .45928 L
s
gsave
-0.0125 .45928 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(130) show
1.000 setlinewidth
grestore
0 .58488 m
.00625 .58488 L
s
gsave
-0.0125 .58488 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(140) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .1076 m
.00375 .1076 L
s
0 .13272 m
.00375 .13272 L
s
0 .15784 m
.00375 .15784 L
s
0 .18296 m
.00375 .18296 L
s
0 .2332 m
.00375 .2332 L
s
0 .25832 m
.00375 .25832 L
s
0 .28344 m
.00375 .28344 L
s
0 .30856 m
.00375 .30856 L
s
0 .3588 m
.00375 .3588 L
s
0 .38392 m
.00375 .38392 L
s
0 .40904 m
.00375 .40904 L
s
0 .43416 m
.00375 .43416 L
s
0 .4844 m
.00375 .4844 L
s
0 .50952 m
.00375 .50952 L
s
0 .53464 m
.00375 .53464 L
s
0 .55976 m
.00375 .55976 L
s
0 .05736 m
.00375 .05736 L
s
0 .03224 m
.00375 .03224 L
s
0 .00712 m
.00375 .00712 L
s
0 .61 m
.00375 .61 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95.9375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .08248 m
1 .08248 L
s
.99375 .20808 m
1 .20808 L
s
.99375 .33368 m
1 .33368 L
s
.99375 .45928 m
1 .45928 L
s
.99375 .58488 m
1 .58488 L
s
.125 Mabswid
.99625 .1076 m
1 .1076 L
s
.99625 .13272 m
1 .13272 L
s
.99625 .15784 m
1 .15784 L
s
.99625 .18296 m
1 .18296 L
s
.99625 .2332 m
1 .2332 L
s
.99625 .25832 m
1 .25832 L
s
.99625 .28344 m
1 .28344 L
s
.99625 .30856 m
1 .30856 L
s
.99625 .3588 m
1 .3588 L
s
.99625 .38392 m
1 .38392 L
s
.99625 .40904 m
1 .40904 L
s
.99625 .43416 m
1 .43416 L
s
.99625 .4844 m
1 .4844 L
s
.99625 .50952 m
1 .50952 L
s
.99625 .53464 m
1 .53464 L
s
.99625 .55976 m
1 .55976 L
s
.99625 .05736 m
1 .05736 L
s
.99625 .03224 m
1 .03224 L
s
.99625 .00712 m
1 .00712 L
s
.99625 .61 m
1 .61 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .08248 m
1 .08248 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.07143 .57757 Mdot
.11905 .57533 Mdot
.16667 .52409 Mdot
.21429 .46977 Mdot
.2619 .43596 Mdot
.30952 .43096 Mdot
.35714 .4235 Mdot
.40476 .36737 Mdot
.45238 .29049 Mdot
.5 .32399 Mdot
.54762 .24627 Mdot
.59524 .28482 Mdot
.64286 .20394 Mdot
.69048 .18126 Mdot
.7381 .12659 Mdot
.78571 .15054 Mdot
.83333 .11264 Mdot
.88095 .08151 Mdot
.92857 .06474 Mdot
.5 Mabswid
.02381 .60332 m
.06244 .57944 L
.10458 .5534 L
.14415 .52894 L
.18221 .50542 L
.22272 .48039 L
.26171 .45629 L
.30316 .43067 L
.34309 .406 L
.3815 .38225 L
.42237 .357 L
.46172 .33268 L
.49955 .30929 L
.53984 .2844 L
.57861 .26043 L
.61984 .23495 L
.65954 .21041 L
.69774 .18681 L
.73838 .16169 L
.77751 .13751 L
.81909 .11181 L
.85916 .08704 L
.89771 .06322 L
.93871 .03788 L
.97619 .01472 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00lP3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo]P3oool00?@0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;D0oooo003e0?ooo`040;mdM02LW4P004RL0=oookL0
oooo003e0?ooo`030?oo_`1d000007Bo0;P0oooo003e0?ooo`030?oo_`1d0000M;oo0;P0oooo003e
0?ooo`040?noM000000004RL0=oookL0oooo003e0?ooo`050;md0000B9`0gmnL04Q8W03Oool0]P3o
ool00?<0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]03oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005<0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`160?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
od40oooo00@0_g@0000000000000M;ooB03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom50?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool04@3oool005D0oooo00<0okmd0000001d_ol0B03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_d00oooo00H0gia8001d_`3oool0omnL04P00018W=m60?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
A`3oool00`3o_g@0000007Boo`0D0?ooo`00E@3oool00`3ooml0W4P0001d_`170?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0m0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0A03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0A@3oool00`3ooml0W4P0001d_`0D0?ooo`00EP3oool0
0`3OW4P004RL0=ooo`160?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0m0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
@`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0AP3oool00`3OW4P004RL0=ooo`0C0?ooo`00AP3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0>@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03d0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`130?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`160?ooo`030?oOW01804P0W=oo
01<0oooo001F0?ooo`030?oo_`1d0000M;oo04L0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03h0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`140?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`160?ooo`030?oo_`1d0000M;oo01<0oooo001G0?ooo`030;md0000M;l0oooo04H0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03h0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`140?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`170?ooo`030;md0000M;l0oooo0180oooo
001G0?ooo`030=nLB000B9`0gooo04H0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo?P3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo04<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_oooA`3oool00`3OW4P004RL0=ooo`0B0?ooo`00E`3oool00`3o
gi`0B01809cOo`160?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`130?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04P0oooo00<0omnL04P0B02Lgol04P3oool005L0oooo00<0oono07@0001d
_ol0A`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0@@3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo04@0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04T0oooo00<0oono07@0001d
_ol04P3oool005P0oooo00<0_g@0001d_`3oool0A`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0@P3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgdH0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04X0oooo00<0_g@0001d_`3o
ool04@3oool005H0oooo00D0oono07@000000000001d0;ooo`180?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`100?ooo`040=nLB00000000000001d_dP0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO04D0oooo00D0oono07@00000
0000001d0;ooo`0A0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00B@3ooooo000005d000002`3oool004T0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`090?ooo`00B@3oool2000000H0oooo00<000000?ooo`3oool0D03oool00`000000oooo
0?ooo`1?0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool0C`3oool00`000000oooo
0?ooo`040?ooo`<000002`3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<00000
0?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo0:80oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`090?oo
o`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0:00oooo0P0000090?ooo`03
0000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0
WP3oool2000000/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02@3o
ool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`2K0?ooo`8000003P3oool0
0`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo
09X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`090?ooo`00B@3oool200000:/0
oooo00<000000?ooo`3oool0V03oool2000000l0oooo0`00000;0?ooo`00B@3oool00`000000oooo
0?ooo`2Z0?ooo`030000003oool0oooo09H0oooo0P00000C0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<00000
0?ooo`3oool0T`3oool2000001H0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo01H0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`2@0?ooo`80
00006@3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo08l0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo08d0oooo0P00000L0?ooo`030000003o
ool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0R`3oool2
000001h0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002[0?ooo`030000003oool0oooo
08X0oooo00<000000?ooo`3oool0103oool2000001H0oooo0`00000;0?ooo`00B@3oool00`000000
oooo0?ooo`2Z0?ooo`030000003oool0oooo08P0oooo0P0000060?ooo`@000005`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo08L0oooo
00<000000?ooo`3oool01@3oool6000001H0oooo00<000000?ooo`3oool02@3oool004T0oooo00<0
00000?ooo`3oool0ZP3oool00`000000oooo0?ooo`250?ooo`800000203oool6000001H0oooo00<0
00000?ooo`3oool02@3oool002L0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`03
0?ooo`040;md00000000000007Boo`D0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo
0?ooo`2Z0?ooo`030000003oool0oooo08<0oooo0P00000;0?ooo`@000005`3oool00`000000oooo
0?ooo`090?ooo`00:@3oool00`3o_g@0000007Boo`050?ooo`060=nLB000M;l0oooo0?oOW0180000
B9cO0`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`T0oooo00<000000?ooo`3oool0ZP3oool0
0`000000oooo0?ooo`220?ooo`030000003oool0oooo00`0oooo0P00000H0?ooo`030000003oool0
oooo00T0oooo000Y0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool01`3o
_g@0001809cOo`3oool0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000003o
ool0oooo0:X0oooo00<000000?ooo`3oool0P03oool3000002P0oooo00<000000?ooo`3oool02@3o
ool002X0oooo00<0gia80018W03Oool00`3oool00`3ooml0W4P004RLg`020?ooo`070?oog`2LB000
04RL0=ooo`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3o
ool0ZP3oool00`000000oooo0?ooo`1o0?ooo`D000009`3oool00`000000oooo0?ooo`090?ooo`00
:P3oool00`3ogi`0B01809cOo`030?ooo`030?oog`2LB00007Bo00<0oooo00H0gia80000B02Lgol0
oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool0ZP3oool00`00
0000oooo0?ooo`1m0?ooo`P000009P3oool00`000000oooo0?ooo`090?ooo`00:P3oool00`3ookl0
M00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0;md0000B9`0
gooo0080oooo00<0omnL04P00018W=l01`3ooooo000005d000002`3oool002/0oooo00<0_g@0001d
_`3oool00`3oool00`3OW4P0001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P0001809cO
o`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`1j0?ooo`030000003oool0oooo00<0oooo1000000W0?ooo`030000003oool0oooo00T0oooo000[
0?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018W=l00`3oool01P2oM00004RL0=ooo`3o
gi`0B00004RLg`<0oooo00<0_g@00018W03Oool01P3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo07P0oooo0P0000070?ooo`800000:03oool00`000000oooo0?ooo`090?ooo`00:`3o
ool00`3ogi`0B01809cOo`030?ooo`040?oog`2LB00004RL0=ooo`80oooo00L0gia80018W03Oool0
oooO09a80000B9`0gooo0080oooo00<0gia80018W03Oool01P3oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo07L0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`090?oo
o`00:`3oool00`3ookl0M00007Boo`040?ooo`030?oOW0180000M;oo0080oooo00L0gia8001d_`3o
ool0oooo0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`2Z
0?ooo`030000003oool0oooo07D0oooo0P00000d0?ooo`030000003oool0oooo00T0oooo000/0?oo
o`030;md0000M;l0oooo00@0oooo00H0gia80000M02oool0oooO09a80018W=l30?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo07@0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`090?ooo`00:P3oool01@3ookl0M000
0000000007@0_ooo00D0oooo00@0gia800000000000007Bo1@3oool0103OW4P0000000000000M;l8
0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0LP3oool2000003L0oooo00<00000
0?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`1`0?oo
o`800000>@3oool00`000000oooo0?ooo`090?ooo`00B@3oool200000:/0oooo00<000000?ooo`3o
ool0K`3oool3000003L0oooo0`00000;0?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo06d0oooo1P00000h0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003o
ool0oooo0:X0oooo00<000000?ooo`3oool0K03oool00`000000oooo00000005000003L0oooo00<0
00000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`1Z
0?ooo`8000000P3oool6000003L0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo00<0oooo1000000h0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3o
ool0C`3oool2000001H0oooo0P0000070?ooo`800000>@3oool00`000000oooo0?ooo`090?ooo`00
B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo04h0oooo1000000D0?ooo`030000
003oool0oooo0480oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`1=0?ooo`H000004@3oool2000004D0oooo00<000000?ooo`3oool02@3o
ool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`1=0?ooo`H000003`3oool2
000004L0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002[0?ooo`030000003oool0oooo
04h0oooo1000000?0?ooo`030000003oool0oooo04D0oooo0`00000;0?ooo`00B@3oool00`000000
oooo0?ooo`2Z0?ooo`030000003oool0oooo04l0oooo0P00000>0?ooo`800000BP3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo05h0oooo
00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo05`0oooo0P00001=0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0FP3oool2000000@0oooo0P000019
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?oo
o`3oool0F03oool2000000D0oooo100000180?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo0:X0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`040?ooo`H0
0000A`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo05D0oooo0P0000070?ooo`H00000A`3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo05<0oooo0P00000:0?ooo`@00000B03o
ool00`000000oooo0?ooo`090?ooo`00B@3oool200000:/0oooo00<000000?ooo`3oool0D@3oool2
000000d0oooo0P0000170?ooo`<000002`3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`00
0000oooo0?ooo`1@0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`1>0?ooo`800000F`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo04/0oooo0P00001N0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`1N
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?oo
o`3oool0?`3oool2000000L0oooo0P00001Q0?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo0:X0oooo00<000000?ooo`3oool0?P3oool4000000@0oooo0P00001S0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`800000Z`3oool00`000000oooo0?ooo`0m0?ooo`H00000
0P3oool00`000000oooo0?ooo`1Q0?ooo`<000002`3oool004T0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`0m0?ooo`P00000IP3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo03h0oooo1@00001X0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0?`3oool30000
06T0oooo00<000000?ooo`3oool02@3oool002L0oooo00<0oono07@0000000000P0000001P18W=l0
oooo0?ooo`3oool0oono07@000<0000000<0B9cO0?ooo`3oool00`3oool0102oM00000000000001d
_ol:0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0?P3oool2000006/0oooo00<0
00000?ooo`3oool02@3oool002T0oooo00<0okmd0000001d_ol01P3oool00`3o_g@0000007Boo`05
0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo03d0oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`090?ooo`00:@3o
ool00`3ooml0W4P0001d_`060?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3o
ool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0;P3o
ool2000000/0oooo0P00001^0?ooo`030000003oool0oooo00T0oooo000Z0?ooo`030=nLB000B9`0
gooo00H0oooo00<0gia80018W03Oool00`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB000
04RL0=ooo`L0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0]0?ooo`@000002@3o
ool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo00T0oooo000Z0?ooo`030?oOW01804P0
W=oo00H0oooo00<0omnL04P0B02Lgol00`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0
W=oo00L0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0/0?ooo`H000001P3oool2
00000740oooo00<000000?ooo`3oool02@3oool002X0oooo00<0oono07@0001d_ol01P3oool00`3o
okl0M00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool30000
0:X0oooo00<000000?ooo`3oool0;03oool6000000@0oooo0P00001a0?ooo`<000002`3oool002/0
oooo00<0_g@0001d_`3oool01P3oool00`2oM00007Bo0?ooo`030?ooo`030=nLB00004P0W=oo0080
oooo00<0oooO09a80000M;l01`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02d0
oooo100000040?ooo`030000003oool0oooo07<0oooo00<000000?ooo`3oool02@3oool002/0oooo
00<0gia80018W03Oool01P3oool00`3OW4P004RL0=ooo`030?ooo`030?oOW0180000B9cO00<0oooo
00<0_g@00018W03Oool01P3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02h0oooo
0P0000030?ooo`800000MP3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ogi`0B01809cO
o`060?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`060?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0<@3oool2000007P0oooo
00<000000?ooo`3oool02@3oool002/0oooo00<0oono07@0001d_ol01P3oool00`3ookl0M00007Bo
o`040?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo02l0oooo0P00001j0?ooo`030000003oool0oooo00T0oooo000/
0?ooo`030;md0000M;l0oooo00H0oooo00<0_g@0001d_`3oool0103oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0^0?ooo`030000
003oool0oooo07X0oooo00<000000?ooo`3oool02@3oool002X0oooo00D0oono07@000000000001d
0;ooo`040?ooo`050?oo_`1d000000000000M02oool01@3oool0103OW4P0000000000000M;l80?oo
o`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0;03oool2000007d0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0Z0?ooo`80
0000O`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo02P0oooo0P0000210?ooo`030000003oool0oooo00T0oooo00190?ooo`800000Z`3o
ool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo07l0oooo0`00000;0?ooo`00B@3oool0
0`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02D0oooo0P0000240?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool03@3oool20000
01D0oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`2Z0?ooo`030000003oool0oooo00`0oooo1000000B0?ooo`800000Q`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo00/0oooo
1P00000@0?ooo`030000003oool0oooo08L0oooo00<000000?ooo`3oool02@3oool004T0oooo00<0
00000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0;0?ooo`H000003P3oool2000008X0oooo00<0
00000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0<
0?ooo`@000003@3oool2000008`0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`0=0?ooo`8000003@3oool00`000000oooo0?ooo`2<0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3o
ool06P3oool2000008l0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002[0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0S@3oool3000000/0oooo00190?ooo`030000003o
ool0oooo0:X0oooo00<000000?ooo`3oool05`3oool200000980oooo00<000000?ooo`3oool02@3o
ool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo0980oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`00
0000oooo0?ooo`0D0?ooo`800000U@3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`2Z0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0U@3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0140oooo
0P00002H0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<0
00000?ooo`3oool03`3oool2000009X0oooo00<000000?ooo`3oool02@3oool004T0oooo00<00000
0?ooo`3oool0ZP3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00`0oooo0P00002<
0?ooo`030000003oool0oooo00T0oooo00190?ooo`800000Z`3oool00`000000oooo0?ooo`0<0?oo
o`8000003P3oool4000008T0oooo0`00000;0?ooo`00B@3oool00`000000oooo0?ooo`2I0?ooo`80
00003`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00d0oooo1P00002:0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo09P0oooo1000000>0?ooo`030000
003oool0oooo00T0oooo0P00000@0?ooo`H00000RP3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2G0?ooo`H000003@3oool00`000000oooo0?ooo`070?ooo`8000004`3o
ool4000008/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0U`3oool6
000000d0oooo00<000000?ooo`3oool01@3oool2000001H0oooo0P00002<0?ooo`030000003oool0
oooo00T0oooo000N0?ooo`030;md0000B9`0gooo02P0oooo00<000000?ooo`3oool0V03oool40000
00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo
00T0oooo000N0?ooo`030?oOW018M;l0oooo02P0oooo00<000000?ooo`3oool0V@3oool2000000l0
oooo00<000000?ooo`3oool00P3oool200000:L0oooo00<000000?ooo`3oool02@3oool001<0oooo
00<0oono07Ad_`3oool01`3oool00`3o_g@004RL0=ooo`0Y0?ooo`030000003oool0oooo0:X0oooo
00<000000?ooo`3oool00P00002Y0?ooo`030000003oool0oooo00T0oooo000C0?ooo`040?oOW018
0000001d0;ooo`D0oooo00<0gia8001d_`3oool0:P3oool00`000000oooo0?ooo`2Z0?ooo`<00000
Z`3oool00`000000oooo0?ooo`090?ooo`00503oool00`3ogi`0B000000000050000000304RLg`3o
ool0oooo02X0oooo0P00002[0?ooo`030000003oool0oooo0:T0oooo0`00000;0?ooo`006@3oool0
0`3o_g@0001809cOo`0]0?ooo`030000003oool0oooo0:P0oooo0`00002]0?ooo`030000003oool0
oooo00T0oooo000H0?ooo`030;md0000M;l0oooo02h0oooo00<000000?ooo`3oool0YP3oool20000
0080oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`004`3oool01P3ogi`0
B00007Boo`3o_g@0000007Booc00oooo00<000000?ooo`3oool0Y@3oool010000000oooo0?ooo`3o
ool200000:d0oooo00<000000?ooo`3oool02@3oool001<0oooo00@0omnL04P00000000007Bo403o
ool00`3ookl0M000000000020000000504RLg`3oool0oooo0?ooo`3OW4P01@0000000`18W=l0oooo
0?ooo`020?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool0X`3oool2000000<0
oooo1000002/0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`030?noM0000000M;oo00D0oooo
00<0okmd001d_`3oool00P3oool02P2oM7@0_ooo0?ooo`3oool0gia8001d_`3oool0omnL04P00018
W=l90?ooo`030000003oool0oooo0:80oooo00<000000?ooo`3oool00P3oool600000:/0oooo00<0
00000?ooo`3oool02@3oool002T0oooo00<0oooO09a80000M;l01P3oool00`3o_g@007Bo0?ooo`04
0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`2P
0?ooo`8000001@3oool600000:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00<0gia80018
W03Oool01P3oool00`3o_g@007Bo0?ooo`030?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a8
0000B9`0gooo1`3oool00`000000oooo0?ooo`2N0?ooo`800000203oool400000:`0oooo00<00000
0?ooo`3oool02@3oool002X0oooo00<0omnL04P0B02Lgol01`3oool00`3o_g@0B9cO0?ooo`020?oo
o`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2M0?oo
o`030000003oool0oooo00T0oooo0P00002]0?ooo`030000003oool0oooo00T0oooo000Z0?ooo`03
0?oo_`1d0000M;oo00P0oooo00<0gia804RLg`3oool00P3oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO00L0oooo0`00002K0?ooo`8000003@3oool00`000000oooo0?ooo`2Y0?ooo`<0
00002`3oool002/0oooo00<0_g@0001d_`3oool0203oool01`2oM000M;oo0?ooo`3oool0gia80000
B02Lgol00P3oool00`3ooml0W4P0001d_`070?ooo`030000003oool0oooo09X0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000[0?ooo`03
0=nLB000B9`0gooo00P0oooo00L0omnL04P0M02oool0oooo0?oOW0180000B9cO00<0oooo00<0_g@0
0018W03Oool01P3oool00`000000oooo0?ooo`2H0?ooo`800000403oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00T0oooo000[0?ooo`030?oOW01804P0W=oo00P0oooo00P0oooO09a8
001d_ol0oooo0?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`000000oooo
0?ooo`2G0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0Z`3oool00`000000oooo
0?ooo`090?ooo`00:`3oool00`3ookl0M00007Boo`080?ooo`030?oo_`1d0000B9cO0080oooo00<0
omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0oooo09H0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000/
0?ooo`030;md0000M;l0oooo00<0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00<0oooo
00H0gia80000M02oool0oooO09a80018W=l70?ooo`030000003oool0oooo09@0oooo0P00000D0?oo
o`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00D0oono07@00000
0000001d0;ooo`040?ooo`050?oOW018000000000000001d_ol01@3oool0103OW4P0000000000000
M;l80?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool0503oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0940oooo0P00000G
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002A0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool0Z@3oool3000000/0oooo00190?ooo`03
0000003oool0oooo08P0oooo0P0000040?ooo`8000006P3oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo08L0oooo100000000`3oool00000
0000000L0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<0
00000?ooo`3oool0QP3oool6000001h0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`260?ooo`H000007P3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo08L0oooo1000000O0?oo
o`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3o
ool0QP3oool010000000oooo00000000000P0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0Q03oool2000002@0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`220?ooo`800000
9P3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`800000PP3o
ool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0:T0oooo0`00000;0?ooo`00B@3oool0
0`000000oooo0?ooo`1o0?ooo`800000:@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo07d0oooo0P00000[0?ooo`030000003oool0oooo
0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0O03oool00`000000
oooo0?ooo`0[0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo
00<000000?ooo`3oool0NP3oool2000002h0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo
0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo02h0oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1g
0?ooo`800000<@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`030000003oool0oooo07D0oooo0P00000c0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`0c0?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00001c0?ooo`800000
=P3oool00`000000oooo0?ooo`2Y0?ooo`<000002`3oool004T0oooo00<000000?ooo`3oool0L03o
ool2000003P0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`1^0?ooo`800000>P3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0203oool20000
0300oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`1[0?ooo`8000002P3oool4000002l0oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1W0?ooo`@000002`3oool6000002h0oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`1V0?ooo`@00000303oool6000002h0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`1G0?ooo`800000303oool6000000`0oooo1000000_
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00001G0?oo
o`@000002P3oool7000000d0oooo0P00000`0?ooo`030000003oool0oooo0:T0oooo0`00000;0?oo
o`00B@3oool00`000000oooo0?ooo`1E0?ooo`H00000203oool00`000000oooo0?ooo`0400000400
oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo
0?ooo`1E0?ooo`H000001P3oool2000000@0oooo0P0000110?ooo`030000003oool0oooo0:/0oooo
00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0EP3oool4000000H0oooo00<0
00000?ooo`3oool0A`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000W
0?ooo`030?oo_`1d000000000080000000H0B9cO0?ooo`3oool0oooo0?oo_`1d00030000000307Bo
o`3oool0oooo00<0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo0?ooo`1G0?ooo`80
00001@3oool2000004X0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00
:@3oool00`3o_g@0000007Boo`040?ooo`030?noM000000007Bo0080oooo00<0_gA809cOo`3oool0
0P3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`T0oooo00<000000?ooo`3oool0G03oool20000
04`0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:@3oool00`3ooml0
W4P0001d_`090?ooo`070?oOW018B9`0gooo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0
W=oo00P0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00<0gia80018W03Oool0203oool01`3o
okl0M00007Boo`3oool0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`030000
003oool0oooo05T0oooo0P00001?0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool002X0oooo00<0omnL04P0B02Lgol0203oool01`3ogi`0B00007Boo`3oool0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`070?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3o
ool0C`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000Z0?ooo`030?oo
_`1d0000M;oo00L0oooo00@0omnL04P0000004P0W=oo0P3oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO00L0oooo0`00001F0?ooo`800000DP3oool00`000000oooo0?ooo`2Y0?ooo`<0
00002`3oool002/0oooo00<0_g@0001d_`3oool0103oool01@3ogi`0B000000000000000M;oo00<0
oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`070?ooo`030000003oool0oooo05D0
oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0
oooo000[0?ooo`030=nLB000B9`0gooo00H0oooo00<0oono07@00000M;l00`3oool00`3ogi`0B000
04RLg`030?ooo`030;md0000B9`0gooo00H0oooo00<000000?ooo`3oool0AP3oool2000000/0oooo
0P00001E0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002/0oooo00<0
omnL04P0B02Lgol0203oool0203OW4P007Bo0?ooo`3oool0oooO09a80000B9`0gooo0P3oool00`3O
W4P004RL0=ooo`060?ooo`030000003oool0oooo04D0oooo100000090?ooo`030000003oool0oooo
05D0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ookl0
M00007Boo`080?ooo`030?oOW01804P0W=oo0080oooo00<0omnL04P0001d_ol00P3oool00`3OW4P0
07Bo0?ooo`060?ooo`030000003oool0oooo04@0oooo1P0000060?ooo`800000F03oool00`000000
oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000/0?ooo`030;md0000M;l0oooo00@0oooo
00H0_g@007Boo`3oool0gia80000B02Lgol30?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3o
ool00`000000oooo0?ooo`140?ooo`H00000103oool2000005X0oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00:P3oool01@3ookl0M0000000000007@0_ooo00D0oooo00@0
_g@0000000000000B9cO1@3oool0103OW4P0000000000000M;l80?ooo`030000003oool0oooo04D0
oooo100000040?ooo`030000003oool0oooo05X0oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`160?ooo`8000000`3oool2000005d0oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool2000004T0oooo0P00001P0?ooo`030000003oool0oooo0:T0oooo0`00000;
0?ooo`00B@3oool00`000000oooo0?ooo`160?ooo`800000HP3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo04@0oooo0P00001T0?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0
@`3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool004T0oooo00<000000?ooo`3oool0@@3oool2000006L0oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0o0?ooo`800000J@3oool0
0`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo
03d0oooo0P00001[0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo0:/0
oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0>P3oool2000006h0oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2000003T0oooo0P00
001`0?ooo`030000003oool0oooo0:T0oooo0`00000;0?ooo`00B@3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`09
0?ooo`00B@3oool00`000000oooo0?ooo`0e0?ooo`800000L`3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo03@0oooo00@000000?ooo`00
00000000LP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo0380oooo0P0000000`3oool000000000000200000740oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0`0?ooo`800000
0P3oool600000700oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo0080oooo1P00001`0?ooo`030000003o
ool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0;@3oool2
000000H0oooo1000001a0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool0
04T0oooo0P00000]0?ooo`030000003oool0oooo00L0oooo0P00001b0?ooo`030000003oool0oooo
0:T0oooo0`00000;0?ooo`00B@3oool00`000000oooo0?ooo`0Z0?ooo`800000OP3oool00`000000
oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo02T0oooo
00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo02L0oooo0P0000210?ooo`030000003oool0oooo0:/0oooo00<0
00000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool09@3oool2000008<0oooo00<00000
0?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0T0?oo
o`030000003oool0oooo08<0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?oo
o`00B@3oool00`000000oooo0?ooo`0R0?ooo`800000QP3oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0
QP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003o
ool0oooo01l0oooo0P0000290?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3o
ool004T0oooo0P00000O0?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0Z@3oool3
000000/0oooo00190?ooo`030000003oool0oooo01`0oooo0P00002<0?ooo`030000003oool0oooo
0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`2<0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo
00<000000?ooo`3oool06@3oool2000000X0oooo0P0000230?ooo`030000003oool0oooo0:/0oooo
00<000000?ooo`3oool02@3oool002L0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?oo
o`040?ooo`030?oog`2LB000B9cO00D0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo
0?ooo`0E0?ooo`@000002`3oool400000880oooo00<000000?ooo`3oool0Z`3oool00`000000oooo
0?ooo`090?ooo`00:@3oool00`3o_g@0000007Boo`080?ooo`030=nLB000B9`0gooo00<0oooo00H0
gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo01@0oooo1000000;0?ooo`H0
0000P@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`030?oo
g`2LB00007Bo00P0oooo00<0omnL04P0B02Lgol00P3oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00P0oooo00<000000?ooo`3oool04`3oool6000000X0oooo1P0000210?ooo`030000
003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00<0gia80018W03Oool00`3o
ool00`3o_g@0000000000004000000050018W03Oool0oooO09a80018W=l00P3oool0103ooml0W4P0
0018W03Oool70?ooo`030000003oool0oooo01<0oooo1P00000;0?ooo`@00000PP3oool00`000000
oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000Z0?ooo`030?oOW01804P0W=oo00<0oooo
00<0oooO09a8B02Lgol00P3oool01`2oM00007Bo0?ooo`3oool0oooO09a80000M;l00`3oool00`3O
W4P0001809cOo`070?ooo`030000003oool0oooo0140oooo0P0000000`3oool00000000000020000
00d0oooo0P0000230?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0
oooo00<0oono07@0001d_ol0103oool01`3ookl0M01d0;ooo`3oool0okmd0018W03Oool00P3oool0
0`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo0`00000?0?ooo`800000103oool2
000009<0oooo00<000000?ooo`3oool0Z@3oool3000000/0oooo000[0?ooo`030;md0000M;l0oooo
00@0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol20?ooo`030=nLB00004P0W=oo0080oooo00<0
oooO09a80000M;l01`3oool00`000000oooo0?ooo`0=0?ooo`800000V`3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo000[0?ooo`030=nLB000B9`0gooo00D0oooo00D0okmd
001d_`3ooml0W4P004RLg`020?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01P3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo09/0oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ogi`0B01809cOo`060?ooo`040=nLB018W=l0
_g@0001d_`80oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030000
003oool0oooo00X0oooo0P00002N0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool002/0oooo00<0oono07@0001d_ol01`3oool0103OW4P0M7Ad0000M02oool20?ooo`030?oO
W0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`080?ooo`800000
X03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000/0?ooo`030;md0000
M;l0oooo00L0oooo00<0_g@00000B02Lgol00`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0
oooo00<000000?ooo`3oool01P3oool200000:80oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`090?ooo`00:P3oool01@3ookl0M0000000000007@0_ooo00L0oooo00<0oooO09a80018
W=l0103oool0103OW4P0000000000000M;l80?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool0XP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo0:X0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00
B@3oool200000:/0oooo00<000000?ooo`3oool0Z@3oool3000000/0oooo00190?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo0500oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00T0oooo00190?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3oool0
04T0ooooo`00001M000000/0oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.49611, 81.6265, \
0.00908859, 0.344578}}]
}, Closed]],

Cell["\<\
The data do appear to scatter pretty symmetrically about the fit, but a \
clearer appraisal is provided by plotting the residuals\
\>", "Text"],

Cell[BoxData[
    \(TraditionalForm\`r\_i = \(y\_i - y\&_\_i\)\/\[Sigma]\_i\)], \
"DisplayFormula"],

Cell[TextData[{
  "If the model is reliable, the residuals should be scattered uniformly and \
if the data are normally distributed approximately two-thirds should be \
contained within ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[PlusMinus]1\)\)]],
  ".  The present data conform with these expectations by construction."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        ListPlot[Evaluate[Transpose[{x, \(y - y\&_\)\/\[Sigma]\_y}]], 
          PlotRange \[Rule] {Automatic, {\(-3\), 3}}, 
          PlotStyle \[Rule] PointSize[0.02], Frame \[Rule] True, 
          FrameLabel \[Rule] {"\<x\>", "\<residuals\>"}], 
        Plot[{\(-1\), 1}, {x, \(-1\), 1}, 
          PlotStyle \[Rule] Dashing[{0.02, 0.02}]]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.103006 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .10301 -18.6875 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -18.6875 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -10 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[-0.0125 .61803 -10 -7.84375 ]
[-0.0125 .61803 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.375 -23.9688 ]
[-0.01875 .30902 -18.6875 23.9688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .61803 m
.00625 .61803 L
s
gsave
-0.0125 .61803 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .0206 m
.00375 .0206 L
s
0 .0412 m
.00375 .0412 L
s
0 .0618 m
.00375 .0618 L
s
0 .0824 m
.00375 .0824 L
s
0 .12361 m
.00375 .12361 L
s
0 .14421 m
.00375 .14421 L
s
0 .16481 m
.00375 .16481 L
s
0 .18541 m
.00375 .18541 L
s
0 .22661 m
.00375 .22661 L
s
0 .24721 m
.00375 .24721 L
s
0 .26781 m
.00375 .26781 L
s
0 .28842 m
.00375 .28842 L
s
0 .32962 m
.00375 .32962 L
s
0 .35022 m
.00375 .35022 L
s
0 .37082 m
.00375 .37082 L
s
0 .39142 m
.00375 .39142 L
s
0 .43262 m
.00375 .43262 L
s
0 .45322 m
.00375 .45322 L
s
0 .47383 m
.00375 .47383 L
s
0 .49443 m
.00375 .49443 L
s
0 .53563 m
.00375 .53563 L
s
0 .55623 m
.00375 .55623 L
s
0 .57683 m
.00375 .57683 L
s
0 .59743 m
.00375 .59743 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -111.5 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
84.9688 11.8438 translate
90 rotate
-84.9688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(residuals) show
106.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .0206 m
1 .0206 L
s
.99625 .0412 m
1 .0412 L
s
.99625 .0618 m
1 .0618 L
s
.99625 .0824 m
1 .0824 L
s
.99625 .12361 m
1 .12361 L
s
.99625 .14421 m
1 .14421 L
s
.99625 .16481 m
1 .16481 L
s
.99625 .18541 m
1 .18541 L
s
.99625 .22661 m
1 .22661 L
s
.99625 .24721 m
1 .24721 L
s
.99625 .26781 m
1 .26781 L
s
.99625 .28842 m
1 .28842 L
s
.99625 .32962 m
1 .32962 L
s
.99625 .35022 m
1 .35022 L
s
.99625 .37082 m
1 .37082 L
s
.99625 .39142 m
1 .39142 L
s
.99625 .43262 m
1 .43262 L
s
.99625 .45322 m
1 .45322 L
s
.99625 .47383 m
1 .47383 L
s
.99625 .49443 m
1 .49443 L
s
.99625 .53563 m
1 .53563 L
s
.99625 .55623 m
1 .55623 L
s
.99625 .57683 m
1 .57683 L
s
.99625 .59743 m
1 .59743 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .30902 m
1 .30902 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.07143 .32483 Mdot
.11905 .44157 Mdot
.16667 .34793 Mdot
.21429 .24104 Mdot
.2619 .22223 Mdot
.30952 .32714 Mdot
.35714 .42151 Mdot
.40476 .30684 Mdot
.45238 .10307 Mdot
.5 .3733 Mdot
.54762 .16593 Mdot
.59524 .45786 Mdot
.64286 .23695 Mdot
.69048 .26591 Mdot
.7381 .15755 Mdot
.78571 .38675 Mdot
.83333 .35039 Mdot
.88095 .34309 Mdot
.92857 .39744 Mdot
.5 Mabswid
[ .02 .02 ] 0 setdash
.02381 .20601 m
.06244 .20601 L
.10458 .20601 L
.14415 .20601 L
.18221 .20601 L
.22272 .20601 L
.26171 .20601 L
.30316 .20601 L
.34309 .20601 L
.3815 .20601 L
.42237 .20601 L
.46172 .20601 L
.49955 .20601 L
.53984 .20601 L
.57861 .20601 L
.61984 .20601 L
.65954 .20601 L
.69774 .20601 L
.73838 .20601 L
.77751 .20601 L
.81909 .20601 L
.85916 .20601 L
.89771 .20601 L
.93871 .20601 L
.97619 .20601 L
s
.02381 .41202 m
.06244 .41202 L
.10458 .41202 L
.14415 .41202 L
.18221 .41202 L
.22272 .41202 L
.26171 .41202 L
.30316 .41202 L
.34309 .41202 L
.3815 .41202 L
.42237 .41202 L
.46172 .41202 L
.49955 .41202 L
.53984 .41202 L
.57861 .41202 L
.61984 .41202 L
.65954 .41202 L
.69774 .41202 L
.73838 .41202 L
.77751 .41202 L
.81909 .41202 L
.85916 .41202 L
.89771 .41202 L
.93871 .41202 L
.97619 .41202 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00k`3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo^@3oool00?40oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;P0oooo003b0?ooo`040;mdM02LW4P004RL0=oookX0
oooo003b0?ooo`030?oo_`1d000007Bo0;/0oooo003b0?ooo`030?oo_`1d0000M;oo0;/0oooo003b
0?ooo`040?noM000000004RL0=oookX0oooo003b0?ooo`050;md0000B9`0gmnL04Q8W03Oool0^@3o
ool00?00oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]`3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005T0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`120?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
oc`0oooo00@0_g@0000000000000M;ooA03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom10?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool0703oool005/0oooo00<0okmd0000001d_ol0A03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_c/0oooo00H0gia8001d_`3oool0omnL04P00018W=m20?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
@`3oool00`3o_g@0000007Boo`0O0?ooo`00F`3oool00`3ooml0W4P0001d_`130?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0h0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0@@3oool00`3ooml0W4P0001d_`0O0?ooo`00G03oool0
0`3OW4P004RL0=ooo`120?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0h0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
?`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0@P3oool00`3OW4P004RL0=ooo`0N0?ooo`00C03oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0o0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`120?ooo`030?oOW01804P0W=oo
01h0oooo001L0?ooo`030?oo_`1d0000M;oo04<0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03T0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`100?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`120?ooo`030?oo_`1d0000M;oo01h0oooo001M0?ooo`030;md0000M;l0oooo0480oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03T0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`100?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`130?ooo`030;md0000M;l0oooo01d0oooo
001M0?ooo`030=nLB000B9`0gooo0480oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo>@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo03l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo@`3oool00`3OW4P004RL0=ooo`0M0?ooo`00G@3oool00`3o
gi`0B01809cOo`120?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`0o0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04@0oooo00<0omnL04P0B02Lgol07@3oool005d0oooo00<0oono07@0001d
_ol0@`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0?03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0400oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04D0oooo00<0oono07@0001d
_ol07@3oool005h0oooo00<0_g@0001d_`3oool0@`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0?@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd80oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04H0oooo00<0_g@0001d_`3o
ool0703oool005`0oooo00D0oono07@000000000001d0;ooo`140?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0k0?ooo`040=nLB00000000000001d_d@0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO0440oooo00D0oono07@00000
0000001d0;ooo`0L0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00D03ooooo000004X000005`3oool00500oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04`0oooo
00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool0
0500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0
X03oool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0E0?ooo`00@03oool00`3OW4P00000000000030000000304RLg`3oool0
oooo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo00110?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool01`3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00480
oooo00<0okmd001d_`3oool02`3oool00`000000oooo0?ooo`2@0?ooo`8000003`3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00130?ooo`030?noM000M;l0oooo00X0oooo
00<000000?ooo`3oool0S`3oool4000000h0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`00<`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00@0oooo00<0
okmd04RLg`3oool02@3oool00`000000oooo0?ooo`2>0?ooo`H000003@3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo01D0oooo00150?ooo`030=nLB018W=l0oooo00P0oooo0`00002>
0?ooo`H000003@3oool00`000000oooo0?ooo`2O0?ooo`<000005`3oool004H0oooo00<0_g@007Bo
o`3oool01`3oool00`000000oooo0?ooo`2?0?ooo`@000003P3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo00160?ooo`030?oOW01807@0_ooo00L0oooo00<000000?ooo`3o
ool0T03oool2000000l0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
AP3oool00`3ooml0W4P007Boo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RLg`070?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0omnL04P0000000000000
07Boo`080?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0
oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`1:0?ooo`800000E@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo04T0oooo1000001D0?ooo`030000003oool0oooo01D0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0B03oool6000005<0oooo00<0
00000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0<
0?ooo`800000>P3oool6000005<0oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0;0?ooo`@00000>P3oool4000005@0oooo00<000000?oo
o`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo00X0oooo1P00000j0?ooo`80
0000E03oool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
2P3oool600000940oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0;0?ooo`@00000TP3oool00`000000oooo0?ooo`0E0?ooo`00D03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00`0oooo0P00002C0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool00400oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`08
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00@P3oool00`3o_g@0000007Boo`0;0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool00`3ooml0W4P0001d_`0;0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
@`3oool00`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00<`3oool00`3ooml0W4P0000000070000000304RLg`3o
ool0oooo00<0oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00<0oono07@0001d_ol02P3oool3
000000@0oooo200000060?ooo`L000001P3oool7000000H0oooo1`0000060?ooo`L000001P3oool8
000000D0oooo200000050?ooo`P000001@3oool8000000H0oooo1`0000060?ooo`L000001P3oool7
000000D0oooo2@0000050?ooo`P000001@3oool8000000D0oooo200000060?ooo`L000001P3oool7
000000H0oooo1`0000060?ooo`P000001@3oool8000000D0oooo200000050?ooo`P000001@3oool8
000000T0oooo0`00000G0?ooo`00A03oool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00A03oool00`3OW4P0
04RL0=ooo`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0E0?ooo`009@3oool0103ookl0M0000000M02ooolK0?ooo`030?oOW01804P0W=oo00T0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0
oooo000O0?ooo`040?oo_`2LW=l0oono07@000@0000000<0M;oo0?ooo`3oool06P3oool00`3ookl0
M00007Boo`090?ooo`030000003oool0oooo0580oooo0P00001=0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool001l0oooo00H0omnL04P00000000000000000B02LgolP0?oo
o`030;md0000M;l0oooo00P0oooo00<000000?ooo`3oool0D@3oool4000004`0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3ookl0M00007Boo`0P0?ooo`05
0?oo_`1d000000000000M02oool0203oool00`000000oooo0?ooo`1@0?ooo`H00000B`3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000P0?ooo`030;mdM02oool0oooo02d0
oooo0P00001A0?ooo`H00000B`3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool001l0oooo
00@0okmd00000000B9`0gooo;@3oool00`000000oooo0?ooo`1A0?ooo`@00000C03oool00`000000
oooo0?ooo`0[0?ooo`800000M03oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo
0?ooo`1B0?ooo`800000C@3oool00`000000oooo0?ooo`0Z0?ooo`@00000L`3oool00`000000oooo
0?ooo`0E0?ooo`008P3oool01P3OW4P00000000000000000001809cOobP0oooo00<000000?ooo`3o
ool0@P3oool2000005d0oooo00<000000?ooo`3oool0:@3oool600000780oooo00<000000?ooo`3o
ool05@3oool00200oooo00T0oono07@00018W=l0omnL09boo`3oool0gia80018W03Oool09`3oool0
0`000000oooo0?ooo`110?ooo`@00000G03oool00`000000oooo0?ooo`0Y0?ooo`H00000LP3oool0
0`000000oooo0?ooo`0E0?ooo`00803oool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d
0;ooo`0W0?ooo`030000003oool0oooo0400oooo1P00001K0?ooo`030000003oool0oooo02X0oooo
1000001c0?ooo`030000003oool0oooo01D0oooo000O0?ooo`090?oOW02L_ol0oooo0?oog`2LM9`0
gooo0?ooo`3ogi`0B7Bo02P0oooo00<000000?ooo`3oool0@03oool6000005/0oooo00<000000?oo
o`3oool0:`3oool2000007@0oooo00<000000?ooo`3oool05@3oool001l0oooo00T0oknL09cOo`3o
ool0_g@007Boo`3oool0gia80000M02oool0:03oool200000480oooo1000001L0?ooo`030000003o
ool0oooo0:00oooo0P00000G0?ooo`007`3oool0103ookl0M0000000001d_ol]0?ooo`030000003o
ool0oooo0480oooo0P00001M0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo000T0?ooo`050?oog`2LB00000000018W03Oool09`3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00200oooo00@0_g@00000
000007@0_ooo0P3oool00`3ookl0M7Bo0?ooo`0W0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool0>`3oool2000006@0oooo00<000000?ooo`3oool05@3oool001l0oooo00H0omnL09bo
o`3ooml0W4P00000M02oool20?ooo`030=nLW02oool0oooo02H0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0j0?ooo`@00000H`3oool00`000000oooo0?ooo`0E0?ooo`007`3oool0
2P3o_i`0W=oo0?ooo`3ooml0W4P00000B02Lgol0oono07A8W03OoolW0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0>@3oool600000680oooo00<000000?ooo`3oool05@3oool001l0
oooo00<0oono07Ad_`3oool00P3oool0102oM00000000000M02ooolX0?ooo`800000XP3oool00`00
0000oooo0?ooo`0i0?ooo`H00000H@3oool2000001L0oooo000O0?ooo`040?oOW0180000000007Bo
obd0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0j0?ooo`@00000H`3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo03/0
oooo0P00001T0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ookl0W9cO0?oo
o`030?ooo`040?noM0000000001d0;ooobL0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000O0?ooo`040?oo_`2LW=l0oono07@000@00000
00<004RL0=ooo`3oool09P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool05@3oool001l0oooo00T0omnL04P000000000001809cOo`3oool0oooO09a8
M02oool0:03oool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo000K0?ooo`04
0?oog`2LB000001d0;oooc40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo000S0?ooo`050=nLB000000000000000001d_ol06P3oool0102o
M00000000000001d_ol:0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`008@3oool00`3ogi`0B00004RLg`020?ooo`030?oOW01807@0_ooo
01P0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3ogi`0B4RL0=ooo`03
0?ooo`030?oo_`1dB9`0gooo01L0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`08
0?ooo`030000003oool0oooo0840oooo0P00000N0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool001l0oooo00<0oooO09adW03Oool0103oool00`2oM000M;oo0?ooo`0G0?oo
o`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`20
0?ooo`@000007@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000O0?oo
o`030?oO_`2ogol0oooo0080oooo00<0oooO09a8001d_ol06@3oool00`3ooml0W4P0001d_`030?oo
o`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool0O`3oool6000001`0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool02P3o_g@0M;oo0?ooo`3oool0omnL
04P0B02Lgkl0M0000018W03OoolH0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
1`3ooooo000004X000005`3oool001l0oooo00<0oooO09a8000000001@0000000`00B9`0gooo0?oo
o`0G0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01`3oool00`000000oooo0?oo
o`200?ooo`@000007@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000K
0?ooo`040;nLW03Oool0oono07@000@0000000D0M;oo0?ooo`3oool0_gAd0;ooo`0I0?ooo`030?oO
W0180000B9cO00<0oooo00<0_g@00018W03Oool01P3oool00`000000oooo0?ooo`210?ooo`800000
7P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000J0?ooo`030?oog`2L
B00000000080000000<0M;oo0?ooo`3oool07`3oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo00H0oooo00<000000?ooo`3oool04`3oool2000008`0oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`006P3oool00`3ookl0W9cO0?ooo`0U0?ooo`030?oOW018
0000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`0B0?ooo`@00000B`3o
ool2000003h0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool0
1@2oM00004RL0=ooo`3ookl0M0000080000000<0001d0;ooo`3oool06@3oool01P3OW4P0001d0;oo
o`3ooml0W4P004RLg`L0oooo00<000000?ooo`3oool04@3oool6000004T0oooo1000000m0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool001l0oooo00<0omnL04P000000000
0`0000001018W=l0oooo0;nLW03OoolK0?ooo`040=nLB00000000000001d_`P0oooo00<000000?oo
o`3oool04@3oool6000004P0oooo1P00000l0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool001l0oooo00@0oooO09a8000007@0_ooo0`3oool00`3o_g@0B9cO0?ooo`0W0?oo
o`8000004`3oool4000004T0oooo1P00000l0?ooo`030000003oool0oooo0:00oooo0P00000G0?oo
o`009@3oool00`3o_g@007Bo0?ooo`0X0?ooo`030000003oool0oooo01<0oooo0P00001;0?ooo`@0
0000?@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000S0?ooo`030?oo
g`2o_kl0W9bo02X0oooo00<000000?ooo`3oool0H@3oool2000003h0oooo00<000000?ooo`3oool0
N@3oool2000002H0oooo00<000000?ooo`3oool05@3oool002<0oooo00H0omno09adB00000000000
0000M02ooolW0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0N03oool4000002D0
oooo00<000000?ooo`3oool05@3oool001l0oooo00<0omnL04P0000000000P0000001@0004P0W=oo
0?ooo`2oW9`0gooo02L0oooo00<000000?ooo`3oool0<`3oool2000006`0oooo00<000000?ooo`3o
ool0M`3oool6000002@0oooo00<000000?ooo`3oool05@3oool001l0oooo00<0okmd0000M02oool0
103oool00`2oM000M;oo0?ooo`0W0?ooo`030000003oool0oooo0380oooo1000001[0?ooo`030000
003oool0oooo06T0oooo0P00000<0?ooo`H00000903oool00`000000oooo0?ooo`0E0?ooo`009P3o
ool00`3Ogml0oooo0?ooo`0W0?ooo`030000003oool0oooo0340oooo1P00001Z0?ooo`030000003o
ool0oooo06P0oooo1000000<0?ooo`@000009@3oool00`000000oooo0?ooo`0E0?ooo`008`3oool0
1@2oM0000000000000000000B9cO02P0oooo0P00000b0?ooo`H00000JP3oool00`000000oooo0?oo
o`1W0?ooo`H00000303oool2000002D0oooo0P00000G0?ooo`008@3oool00`3OW4P0000007Boo`02
0?ooo`030?oOW01807@0_ooo02L0oooo00<000000?ooo`3oool0<P3oool4000006/0oooo00<00000
0?ooo`3oool0I`3oool6000003@0oooo00<000000?ooo`3oool05@3oool00200oooo00<0gia804RL
g`3oool00`3oool00`3ogi`0B7Bo0?ooo`0W0?ooo`030000003oool0oooo03<0oooo0P00001/0?oo
o`030000003oool0oooo06P0oooo1000000e0?ooo`030000003oool0oooo01D0oooo000O0?ooo`03
0?oOW01dW=l0oooo00<0oooo00<0oknL04QdW03Oool0:03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo06T0oooo0P00000f0?ooo`030000003oool0oooo01D0oooo000O0?ooo`030?no
M000M;l0oooo0080oooo00D0omnL04QdM01d000004RL0=ooo`0W0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3o_g@00000
00000004000000030018W03Oool0oooo02H0oooo00<000000?ooo`3oool0X03oool200000:<0oooo
00<000000?ooo`3oool05@3oool001l0oooo00D0omnL04P000000000000004RLg`020?ooo`030;md
B02Lgol0oooo02L0oooo00<000000?ooo`3oool0W`3oool400000:80oooo00<000000?ooo`3oool0
5@3oool00500oooo0P00002O0?ooo`H00000X03oool2000001L0oooo001@0?ooo`030000003oool0
oooo09h0oooo1P00002Q0?ooo`030000003oool0oooo01D0oooo000S0?ooo`060?noM00000000000
0000000007@0_ooo9`3oool00`000000oooo0?ooo`2O0?ooo`@00000XP3oool00`000000oooo0?oo
o`0E0?ooo`006`3oool01P2oW9`0gooo0?ooo`3oool0omnL04P000<0000000@004RL0=ooo`3ookl0
M7Bo:03oool00`000000oooo0?ooo`2P0?ooo`800000G03oool2000004D0oooo00<000000?ooo`3o
ool05@3oool001X0oooo00<0oooO09a8000000000`0000000`00M;l0oooo0?ooo`020?ooo`030?no
M00007@0_ooo02P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1I0?ooo`@00000
A03oool00`000000oooo0?ooo`0E0?ooo`006P3oool00`3ookl0M00004RLg`0c0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0F03oool6000004<0oooo00<000000?ooo`3oool05@3o
ool002@0oooo00D0oooO09a80000000004RL0=ooo`0W0?ooo`800000XP3oool00`000000oooo0?oo
o`1H0?ooo`H00000:`3oool2000001D0oooo0P00000G0?ooo`00803oool0102oM00000000000M02o
ool20?ooo`030?oo_`1dM;l0oooo02L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`1I0?ooo`@00000:`3oool4000001D0oooo00<000000?ooo`3oool05@3oool001l0oooo00H0omnL
09boo`3ooml0W4P00000M02oool20?ooo`030=nLW02oool0oooo01H0oooo00<0oono07@000000000
0P0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
FP3oool2000002/0oooo1P00000D0?ooo`030000003oool0oooo01D0oooo000O0?ooo`0:0?noW02L
gol0oooo0?oog`2LB000001809cOo`3ookl0M4RL0=oooaT0oooo00<0okmd0000001d_ol02`3oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo08L0oooo1P00000D0?ooo`030000003oool0
oooo01D0oooo000O0?ooo`030?oo_`1dM;l0oooo0080oooo00@0_g@00000000007@0_ooo6P3oool0
0`3ooml0W4P0001d_`0;0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0R03oool4
000001D0oooo00<000000?ooo`3oool05@3oool001l0oooo00@0omnL04P000000000M;oo803oool0
0`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0R@3oool2
000001H0oooo00<000000?ooo`3oool05@3oool004<0oooo00<0omnL04P0B02Lgol02P3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0
oooo00<0oono07@0001d_ol02P3oool3000000@0oooo200000060?ooo`L000001P3oool7000000H0
oooo1`0000060?ooo`L000001P3oool8000000D0oooo200000050?ooo`P000001@3oool8000000H0
oooo1`0000060?ooo`L000001P3oool7000000D0oooo2@0000050?ooo`P000001@3oool8000000D0
oooo200000060?ooo`L000001P3oool7000000H0oooo1`0000060?ooo`P000001@3oool8000000D0
oooo200000050?ooo`P000001@3oool8000000T0oooo0`00000G0?ooo`00A03oool00`2oM00007Bo
0?ooo`090?ooo`030000003oool0oooo0740oooo0P00000^0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool05@3oool004@0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?oo
o`1`0?ooo`@00000;@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo0014
0?ooo`030?oOW01804P0W=oo00T0oooo00<000000?ooo`3oool0K`3oool6000002`0oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00A03oool00`3ookl0M00007Boo`090?oo
o`030000003oool0oooo06l0oooo1P00000/0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool004D0oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`1`0?ooo`@0
0000;@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00130?ooo`050?oo
_`1d000000000000M02oool0203oool00`000000oooo0?ooo`1a0?ooo`800000;P3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo
0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool08`3oool2000007`0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`0R
0?ooo`@00000N`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0240oooo1P00001j0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool08@3oool6000007X0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`0R0?ooo`@00000
N`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003o
ool0oooo02<0oooo0P00001l0?ooo`030000003oool0oooo01/0oooo0P0000240?ooo`030000003o
ool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?ooo`0J0?ooo`@00000PP3oool2
000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool06@3oool60000
0880oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0I0?ooo`H00000PP3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01X0oooo100000230?ooo`030000003oool0oooo01D0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool06`3oool2000008@0oooo00<0
00000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?ooo`2P0?oo
o`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo
0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00400oooo00<0gia8
0000000000000`0000000`18W=l0oooo0?ooo`070?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@@3oool00`3o_g@007Bo0?ooo`020?oo
o`030;mdM02oool0oooo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo00120?ooo`030?noM000M;l0oooo00/0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00130?ooo`030?no
M000M;l0oooo00X0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo01D0oooo00140?ooo`030?noM018W=l0oooo00T0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00150?ooo`030=nLB018W=l0
oooo00P0oooo0`00002Q0?ooo`030000003oool0oooo09l0oooo0`00000G0?ooo`00AP3oool00`2o
M000M;oo0?ooo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00AP3oool00`3ogi`0B01d0;ooo`070?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ooml0W4P0
07Boo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RLg`070?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool01`3o_g@004RL0=ooo`3o
ool0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool05@3oool004<0oooo00D0omnL04P000000000000007Boo`080?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03o
ool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo001@0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0
oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80
oooo00<000000?ooo`3oool0X03oool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo00100?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0203oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00400
oooo00<0okmd00000000M;l00P3oool00`2oM4P0W=oo0?ooo`080?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00A@3oool00`3ogi`0B4RL
0=ooo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`00A@3oool00`3ookl0M00007Boo`080?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0C03oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo00<0
00000?ooo`3oool0B`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01D0oooo0015
0?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E
0?ooo`00A03oool0103ogi`0B0000000B02Lgol80?ooool00000BP00000G0?ooo`00@P3oool01@3o
gi`0B000000000000000M;oo0?l0ooooJP3oool004@0oooo00<0oono07@00000M;l0o`3ooomZ0?oo
o`00AP3oool00`3OW4P007Bo0?ooo`3o0?ooofP0oooo00160?ooo`030?oOW01804P0W=oo0?l0oooo
J03oool004<0oooo00H0_g@007Boo`3oool0gia80000B02Lgooo0?ooofP0oooo00140?ooo`040;md
00000000000004RLgol0ooooJ@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.56215, -4.51278, \
0.00959293, 0.0443476}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["example: linear fit to nonlinear data", "Subsubsection"],

Cell["\tNext suppose we construct a data set for a nonlinear model", "Text"],

Cell[BoxData[{
    \(\(x = Range[\(-0.9\), 0.9, 0.1];\)\), "\[IndentingNewLine]", 
    \(\(y = 
        117\  - \ 23  x + 15  x\^2 + 
          Table[Random[
              NormalDistribution[0, 2. ]], {Length[
                x]}];\)\), "\[IndentingNewLine]", 
    \(\(w = Table[1, {Length[x]}];\)\)}], "Input"],

Cell["but attempt to fit those data using linear least squares", "Text"],

Cell[BoxData[{
    \(\(S\_x = w . x;\)\), "\[IndentingNewLine]", 
    \(\(S\_y = w . y;\)\), "\[IndentingNewLine]", 
    \(\(S\_\(x, x\) = w . x\^2;\)\), "\[IndentingNewLine]", 
    \(\(S\_\(x, y\) = w . \((x*y)\);\)\), "\[IndentingNewLine]", 
    \(\(W = Plus @@ w;\)\), "\[IndentingNewLine]", 
    \(\(\[CapitalDelta] = W\ S\_\(x, x\) - S\_x\^2;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \({a\_0, 
        a\_1} = {\(\(S\_y\) S\_\(x, x\) - \(S\_x\) S\_\(x, y\)\)\/\
\[CapitalDelta], \ \(W\ S\_\(x, y\) - \(S\_x\) S\_y\)\/\[CapitalDelta]} // 
        N\)], "Input"],

Cell[BoxData[
    \({120.97816520980534`, \(-21.789026309579313`\)}\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(\(y\&_ = a\_0 + \(a\_1\) x;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Sigma]\_y\  = \ 
      Sqrt[Plus @@ \(\((y - y\&_)\)\^2\)/\((Length[y] - 2)\)]\)], "Input"],

Cell[BoxData[
    \(5.549751889606385`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \({\[Sigma]\_0, \[Sigma]\_1} = \(\[Sigma]\_y\) 
          Sqrt /@ {S\_\(x, x\)\/\[CapitalDelta], \ \(\(W\)\(\ \)\)\/\
\[CapitalDelta]} // N\)], "Input"],

Cell[BoxData[
    \({1.2732004025535857`, 2.3245352690108567`}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The standard deviation is now substantially larger and comparisons between \
the fit and the data show pronounced systematic deviations, not just random \
fluctuations.  Although we cannot reject the linear model with absolute \
certainty without reliable estimates for the measurement uncertainties, these \
figures cast considerable doubt upon the hypothesis that ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " is a linear function of ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and suggest that the model should be expanded to include an ",
  Cell[BoxData[
      \(TraditionalForm\`x\^2\)]],
  " contribution (note that parabolic nature of the residuals)."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[\[IndentingNewLine]ListPlot[
          Evaluate[Transpose[{x, y}]], 
          PlotStyle \[Rule] PointSize[0.02]], \[IndentingNewLine]Plot[
          a\_0 + \(a\_1\) x, {x, \(-1\), 1}], Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 -1.11558 0.0113953 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .02396 -22 -7.84375 ]
[-0.0125 .02396 0 7.84375 ]
[-0.0125 .13791 -22 -7.84375 ]
[-0.0125 .13791 0 7.84375 ]
[-0.0125 .25186 -22 -7.84375 ]
[-0.0125 .25186 0 7.84375 ]
[-0.0125 .36581 -22 -7.84375 ]
[-0.0125 .36581 0 7.84375 ]
[-0.0125 .47977 -22 -7.84375 ]
[-0.0125 .47977 0 7.84375 ]
[-0.0125 .59372 -22 -7.84375 ]
[-0.0125 .59372 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -37.6875 -5.09375 ]
[-0.01875 .30902 -22 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .02396 m
.00625 .02396 L
s
gsave
-0.0125 .02396 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
1.000 setlinewidth
grestore
0 .13791 m
.00625 .13791 L
s
gsave
-0.0125 .13791 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(110) show
1.000 setlinewidth
grestore
0 .25186 m
.00625 .25186 L
s
gsave
-0.0125 .25186 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(120) show
1.000 setlinewidth
grestore
0 .36581 m
.00625 .36581 L
s
gsave
-0.0125 .36581 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(130) show
1.000 setlinewidth
grestore
0 .47977 m
.00625 .47977 L
s
gsave
-0.0125 .47977 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(140) show
1.000 setlinewidth
grestore
0 .59372 m
.00625 .59372 L
s
gsave
-0.0125 .59372 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04675 m
.00375 .04675 L
s
0 .06954 m
.00375 .06954 L
s
0 .09233 m
.00375 .09233 L
s
0 .11512 m
.00375 .11512 L
s
0 .1607 m
.00375 .1607 L
s
0 .18349 m
.00375 .18349 L
s
0 .20628 m
.00375 .20628 L
s
0 .22907 m
.00375 .22907 L
s
0 .27465 m
.00375 .27465 L
s
0 .29744 m
.00375 .29744 L
s
0 .32023 m
.00375 .32023 L
s
0 .34302 m
.00375 .34302 L
s
0 .38861 m
.00375 .38861 L
s
0 .4114 m
.00375 .4114 L
s
0 .43419 m
.00375 .43419 L
s
0 .45698 m
.00375 .45698 L
s
0 .50256 m
.00375 .50256 L
s
0 .52535 m
.00375 .52535 L
s
0 .54814 m
.00375 .54814 L
s
0 .57093 m
.00375 .57093 L
s
0 .00116 m
.00375 .00116 L
s
0 .61651 m
.00375 .61651 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95.9375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .02396 m
1 .02396 L
s
.99375 .13791 m
1 .13791 L
s
.99375 .25186 m
1 .25186 L
s
.99375 .36581 m
1 .36581 L
s
.99375 .47977 m
1 .47977 L
s
.99375 .59372 m
1 .59372 L
s
.125 Mabswid
.99625 .04675 m
1 .04675 L
s
.99625 .06954 m
1 .06954 L
s
.99625 .09233 m
1 .09233 L
s
.99625 .11512 m
1 .11512 L
s
.99625 .1607 m
1 .1607 L
s
.99625 .18349 m
1 .18349 L
s
.99625 .20628 m
1 .20628 L
s
.99625 .22907 m
1 .22907 L
s
.99625 .27465 m
1 .27465 L
s
.99625 .29744 m
1 .29744 L
s
.99625 .32023 m
1 .32023 L
s
.99625 .34302 m
1 .34302 L
s
.99625 .38861 m
1 .38861 L
s
.99625 .4114 m
1 .4114 L
s
.99625 .43419 m
1 .43419 L
s
.99625 .45698 m
1 .45698 L
s
.99625 .50256 m
1 .50256 L
s
.99625 .52535 m
1 .52535 L
s
.99625 .54814 m
1 .54814 L
s
.99625 .57093 m
1 .57093 L
s
.99625 .00116 m
1 .00116 L
s
.99625 .61651 m
1 .61651 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.07143 .60332 Mdot
.11905 .49459 Mdot
.16667 .48777 Mdot
.21429 .45131 Mdot
.2619 .38923 Mdot
.30952 .3559 Mdot
.35714 .24869 Mdot
.40476 .25147 Mdot
.45238 .21833 Mdot
.5 .19341 Mdot
.54762 .17367 Mdot
.59524 .13562 Mdot
.64286 .12718 Mdot
.69048 .1969 Mdot
.7381 .1426 Mdot
.78571 .1498 Mdot
.83333 .14799 Mdot
.88095 .13289 Mdot
.92857 .09649 Mdot
.5 Mabswid
.02381 .5113 m
.06244 .49116 L
.10458 .46919 L
.14415 .44855 L
.18221 .42871 L
.22272 .40759 L
.26171 .38726 L
.30316 .36565 L
.34309 .34483 L
.3815 .3248 L
.42237 .30349 L
.46172 .28297 L
.49955 .26324 L
.53984 .24224 L
.57861 .22202 L
.61984 .20052 L
.65954 .17982 L
.69774 .1599 L
.73838 .13871 L
.77751 .11831 L
.81909 .09663 L
.85916 .07574 L
.89771 .05564 L
.93871 .03426 L
.97619 .01472 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00lP3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo]P3oool00?@0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;D0oooo003e0?ooo`040;mdM02LW4P004RL0=oookL0
oooo003e0?ooo`030?oo_`1d000007Bo0;P0oooo003e0?ooo`030?oo_`1d0000M;oo0;P0oooo003e
0?ooo`040?noM000000004RL0=oookL0oooo003e0?ooo`050;md0000B9`0gmnL04Q8W03Oool0]P3o
ool00?<0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]03oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005H0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`140?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
od00oooo00@0_g@0000000000000M;ooAP3oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom50?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool04`3oool005P0oooo00<0okmd0000001d_ol0AP3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_cl0oooo00H0gia8001d_`3oool0omnL04P00018W=m40?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
A`3oool00`3o_g@0000007Boo`0F0?ooo`00F03oool00`3ooml0W4P0001d_`150?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0l0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@P3oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0A@3oool00`3ooml0W4P0001d_`0F0?ooo`00F@3oool0
0`3OW4P004RL0=ooo`140?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0l0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
@@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0AP3oool00`3OW4P004RL0=ooo`0E0?ooo`00B@3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=`3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03`0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`110?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`160?ooo`030?oOW01804P0W=oo
01D0oooo001I0?ooo`030?oo_`1d0000M;oo04D0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03d0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`120?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`160?ooo`030?oo_`1d0000M;oo01D0oooo001J0?ooo`030;md0000M;l0oooo04@0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03d0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`120?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`170?ooo`030;md0000M;l0oooo01@0oooo
001J0?ooo`030=nLB000B9`0gooo04@0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo?@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo0440oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_oooA`3oool00`3OW4P004RL0=ooo`0D0?ooo`00FP3oool00`3o
gi`0B01809cOo`140?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03h0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`110?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04P0oooo00<0omnL04P0B02Lgol0503oool005X0oooo00<0oono07@0001d
_ol0A@3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0@03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0480oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04T0oooo00<0oono07@0001d
_ol0503oool005/0oooo00<0_g@0001d_`3oool0A@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0@@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd@0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04X0oooo00<0_g@0001d_`3o
ool04`3oool005T0oooo00D0oono07@000000000001d0;ooo`160?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0o0?ooo`040=nLB00000000000001d_dH0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO04D0oooo00D0oono07@00000
0000001d0;ooo`0C0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00C03ooooo000005L000003P3oool004`0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo04h0oooo
00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo04h0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`00:@3oool00`3ookl0M0000000
00020000000304RLg`3oool0oooo00<0oooo00@0_g@0000000000000M;oo1@3oool0102oM0000000
0000001d_ol;0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0Z03oool00`000000
oooo0?ooo`0<0?ooo`00:`3oool00`3o_g@0000007Boo`050?ooo`060=nLB000M;l0oooo0?oOW018
0000B9cO0`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`X0oooo00<000000?ooo`3oool0Y`3o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000[0?ooo`030?oog`2LB000
07Bo00@0oooo00<0oooO09a80018W=l00P3oool01`3o_g@0001809cOo`3oool0oooO09a80018W=l0
0P3oool00`3o_g@0001809cOo`090?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0
X03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00`0oooo000/0?ooo`030=nLB000
B9`0gooo00<0oooo00<0oooO09a80018W=l00P3oool01`3ooml0W4P00018W03Oool0oooO09a80018
W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo0:L0oooo00<000000?oo
o`3oool0WP3oool2000000P0oooo00<000000?ooo`3oool0303oool002`0oooo00<0omnL04P0B02L
gol00`3oool00`3ooml0W4P0001d_`030?ooo`060=nLB00004P0W=oo0?oog`2LB00007Bo0`3oool0
0`3OW4P0001809cOo`080?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0W03oool2
000000X0oooo00<000000?ooo`3oool0303oool002`0oooo00<0oono07@0001d_ol0103oool00`2o
M00004RL0=ooo`020?ooo`070?oOW0180000B9cO0?ooo`2oM00004RL0=ooo`020?ooo`030?oOW018
0000B9cO00P0oooo0`00002W0?ooo`030000003oool0oooo09X0oooo0P00000:0?ooo`<000003P3o
ool002d0oooo00<0_g@0001d_`3oool00`3oool00`3OW4P0001809cOo`020?ooo`070?oog`2LB000
07Bo0?ooo`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00<000000?ooo`3oool0
Y`3oool00`000000oooo0?ooo`2H0?ooo`8000003P3oool00`000000oooo0?ooo`0<0?ooo`00;@3o
ool00`3OW4P004RL0=ooo`030?ooo`030?oOW0180000B9cO00<0oooo00H0_g@00018W03Oool0omnL
04P00018W=l30?ooo`030;md0000B9`0gooo00L0oooo00<000000?ooo`3oool0Y`3oool00`000000
oooo0?ooo`2F0?ooo`800000403oool00`000000oooo0?ooo`0<0?ooo`00;@3oool00`3ogi`0B018
09cOo`030?ooo`040?oog`2LB00004RL0=ooo`80oooo00L0gia80018W03Oool0oooO09a80000B9`0
gooo0080oooo00<0gia80018W03Oool01`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0
oooo09@0oooo0P00000B0?ooo`030000003oool0oooo00`0oooo000]0?ooo`030?oo_`1d0000M;oo
00@0oooo00<0omnL04P0001d_ol00P3oool01`3OW4P007Bo0?ooo`3oool0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`070?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0T`3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00`0oooo000^0?ooo`030;md0000M;l0
oooo00@0oooo00H0gia80000M02oool0oooO09a80018W=l30?ooo`060=nLB00007@0_ooo0?oog`2L
B000B9cO203oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0940oooo0P00000E0?oo
o`030000003oool0oooo00`0oooo000/0?ooo`050?oo_`1d000000000000M02oool01@3oool0103O
W4P0000000000000M;l50?ooo`040=nLB00000000000001d_`T0oooo00<000000?ooo`3oool0Y`3o
ool00`000000oooo0?ooo`2?0?ooo`8000005`3oool00`000000oooo0?ooo`0<0?ooo`00C03oool0
0`000000oooo0?ooo`2W0?ooo`030000003oool0oooo08d0oooo0P00000I0?ooo`030000003oool0
oooo00`0oooo001<0?ooo`800000Z03oool00`000000oooo0?ooo`2;0?ooo`8000006P3oool20000
00h0oooo001<0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0R@3oool2000001d0
oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo
0?ooo`270?ooo`8000007`3oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?oo
o`2W0?ooo`030000003oool0oooo08D0oooo0P00000Q0?ooo`030000003oool0oooo00`0oooo001<
0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0P`3oool2000002<0oooo00<00000
0?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`210?oo
o`8000009@3oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`03
0000003oool0oooo07l0oooo0P00000W0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000
003oool0oooo0:L0oooo00<000000?ooo`3oool0O@3oool2000002T0oooo00<000000?ooo`3oool0
303oool004`0oooo0P00002X0?ooo`030000003oool0oooo07/0oooo0P00000Z0?ooo`8000003P3o
ool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`1i0?ooo`800000;@3oool0
0`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo
07P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000
oooo0?ooo`2W0?ooo`030000003oool0oooo07H0oooo0P00000`0?ooo`030000003oool0oooo00`0
oooo001<0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0M03oool200000380oooo
00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?oo
o`1b0?ooo`800000=03oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W
0?ooo`030000003oool0oooo0700oooo0P00000f0?ooo`030000003oool0oooo00`0oooo001<0?oo
o`800000Z03oool00`000000oooo0?ooo`1^0?ooo`8000007`3oool2000001H0oooo0P00000>0?oo
o`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo06`0oooo0P00000P0?ooo`@0
00005P3oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000
003oool0oooo06X0oooo0P00000Q0?ooo`H000005@3oool00`000000oooo0?ooo`0<0?ooo`00C03o
ool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo06P0oooo0P00000S0?ooo`H000005@3o
ool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0
oooo06H0oooo0P00000V0?ooo`@000005P3oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`00
0000oooo0?ooo`2W0?ooo`030000003oool0oooo06@0oooo0P00000Y0?ooo`8000005`3oool00`00
0000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0680
oooo0P0000140?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo0:L0oooo
00<000000?ooo`3oool0H03oool2000004H0oooo00<000000?ooo`3oool0303oool004`0oooo0P00
002X0?ooo`030000003oool0oooo05h0oooo0P0000170?ooo`8000003P3oool004`0oooo00<00000
0?ooo`3oool0Y`3oool00`000000oooo0?ooo`1L0?ooo`800000BP3oool00`000000oooo0?ooo`0<
0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo02d0oooo0P00000[0?oo
o`800000C03oool00`000000oooo0?ooo`0<0?ooo`00:@3oool00`3ookl0M0000000000200000006
04RLg`3oool0oooo0?ooo`3ookl0M0000`0000000`18W=l0oooo0?ooo`030?ooo`040;md00000000
000007Boo`/0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0/0?ooo`@00000:03o
ool2000004h0oooo00<000000?ooo`3oool0303oool002/0oooo00<0okmd0000001d_ol01P3oool0
0`3o_g@0000007Boo`050?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2P3oool00`000000oooo
0?ooo`2W0?ooo`030000003oool0oooo02/0oooo1P00000U0?ooo`8000009`3oool2000002L0oooo
00<000000?ooo`3oool0303oool002/0oooo00<0oooO09a80000M;l01P3oool00`3ooml0W4P0001d
_`040?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?oo
o`2W0?ooo`030000003oool0oooo01d0oooo0P00000<0?ooo`H000008`3oool2000002P0oooo1000
000V0?ooo`030000003oool0oooo00`0oooo000/0?ooo`030=nLB000B9`0gooo00H0oooo00<0gia8
0018W03Oool00`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0
00000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0L0?ooo`@00000303oool400000280oooo0P00
000Y0?ooo`H000009@3oool00`000000oooo0?ooo`0<0?ooo`00;03oool00`3ogi`0B01809cOo`06
0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`08
0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool06`3oool6000000`0oooo0P00000Q
0?ooo`800000:`3oool6000002D0oooo00<000000?ooo`3oool0303oool002`0oooo00<0oono07@0
001d_ol01P3oool00`3ookl0M00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l0203oool300000:L0oooo00<000000?ooo`3oool06`3oool6000002d0oooo0P00000^0?oo
o`@00000903oool3000000h0oooo000]0?ooo`030;md0000M;l0oooo00H0oooo00<0_g@0001d_`3o
ool00`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00<000000?ooo`3o
ool0Y`3oool00`000000oooo0?ooo`0L0?ooo`@00000;@3oool4000001d0oooo0P00000?0?ooo`80
00009`3oool00`000000oooo0?ooo`0<0?ooo`00;@3oool00`3OW4P004RL0=ooo`060?ooo`030=nL
B000B9`0gooo00<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?ooo`030000
003oool0oooo0:L0oooo00<000000?ooo`3oool07@3oool2000002`0oooo1`00000<0?ooo`800000
3@3oool4000003L0oooo00<000000?ooo`3oool0303oool002d0oooo00<0omnL04P0B02Lgol01P3o
ool00`3ogi`0B01809cOo`030?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0
1`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo04T0oooo0P0000000`3oool00000
00000004000000/0oooo1000000;0?ooo`H00000=P3oool00`000000oooo0?ooo`0<0?ooo`00;@3o
ool00`3ookl0M00007Boo`060?ooo`030?oo_`1d0000M;oo00@0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`070?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0B03o
ool00`000000oooo0?ooo`020?ooo`@000002`3oool6000000X0oooo1P00000f0?ooo`030000003o
ool0oooo00`0oooo000^0?ooo`030;md0000M;l0oooo00H0oooo00<0_g@0001d_`3oool0103oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo
0?ooo`160?ooo`8000001P3oool2000000`0oooo1P00000;0?ooo`@00000=`3oool00`000000oooo
0?ooo`0<0?ooo`00;03oool01@3ookl0M0000000000007@0_ooo00@0oooo00D0oono07@000000000
001d0;ooo`050?ooo`040=nLB00000000000001d_`T0oooo00<000000?ooo`3oool0Y`3oool00`00
0000oooo0?ooo`140?ooo`8000005`3oool4000000d0oooo0P00000h0?ooo`030000003oool0oooo
00`0oooo001<0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0@P3oool2000001X0
oooo0P0000180?ooo`030000003oool0oooo00`0oooo001<0?ooo`800000Z03oool00`000000oooo
0?ooo`100?ooo`800000I@3oool2000000h0oooo001<0?ooo`030000003oool0oooo0:L0oooo00<0
00000?ooo`3oool0?P3oool2000006P0oooo00<000000?ooo`3oool0303oool004`0oooo00<00000
0?ooo`3oool0Y`3oool00`000000oooo0?ooo`0=0?ooo`800000;@3oool2000006X0oooo00<00000
0?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0<0?oo
o`@00000:P3oool2000006`0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3o
ool0Y`3oool00`000000oooo0?ooo`0;0?ooo`H000009`3oool2000006h0oooo00<000000?ooo`3o
ool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0;0?ooo`H00000
9@3oool200000700oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3o
ool00`000000oooo0?ooo`0<0?ooo`@00000903oool200000780oooo00<000000?ooo`3oool0303o
ool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0=0?ooo`8000008`3oool2
000007@0oooo00<000000?ooo`3oool0303oool004`0oooo0P00002X0?ooo`030000003oool0oooo
0300oooo0P00001e0?ooo`8000003P3oool004`0oooo00<000000?ooo`3oool0Y`3oool2000002l0
oooo0P00001h0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo0:H0oooo
1000000/0?ooo`800000403oool2000006P0oooo00<000000?ooo`3oool0303oool004`0oooo00<0
00000?ooo`3oool0Y@3oool6000002T0oooo0P00000A0?ooo`@00000I`3oool00`000000oooo0?oo
o`0<0?ooo`00C03oool00`000000oooo0?ooo`2U0?ooo`H000009`3oool200000180oooo1P00001V
0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo0:H0oooo1000000V0?oo
o`800000503oool6000006H0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3o
ool0Y`3oool2000002D0oooo0P00000G0?ooo`@00000I`3oool00`000000oooo0?ooo`0<0?ooo`00
C03oool200000:P0oooo00<000000?ooo`3oool08P3oool2000001X0oooo0P00001W0?ooo`800000
3P3oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`0P0?ooo`800000QP3o
ool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2F0?ooo`8000003`3oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0303oool0
04`0oooo00<000000?ooo`3oool0U@3oool4000000h0oooo00<000000?ooo`3oool07@3oool20000
08T0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0U03oool6000000d0
oooo00<000000?ooo`3oool06`3oool2000008/0oooo00<000000?ooo`3oool0303oool004`0oooo
00<000000?ooo`3oool0U03oool6000000d0oooo00<000000?ooo`3oool06@3oool2000008d0oooo
00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0U@3oool4000000h0oooo00<0
00000?ooo`3oool0603oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo00`0oooo001<
0?ooo`030000003oool0oooo09H0oooo0P00000?0?ooo`030000003oool0oooo01H0oooo0P00002@
0?ooo`030000003oool0oooo00`0oooo001<0?ooo`800000Z03oool00`000000oooo0?ooo`0D0?oo
o`800000T@3oool2000000h0oooo001<0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3o
ool04P3oool2000009@0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0
Y`3oool00`000000oooo0?ooo`0@0?ooo`800000UP3oool00`000000oooo0?ooo`0<0?ooo`00:@3o
ool00`3ookl0M000000000020000000504RLg`3oool0oooo0?ooo`3OW4P01@0000000`18W=l0oooo
0?ooo`020?ooo`040;md00000000000007Boo`/0oooo00<000000?ooo`3oool0Y`3oool00`000000
oooo0?ooo`0>0?ooo`800000V03oool00`000000oooo0?ooo`0<0?ooo`00:`3oool00`3o_g@00000
07Boo`050?ooo`030?noM000M;l0oooo0080oooo00X0_gAd0;ooo`3oool0oooo0=nLB000M;l0oooo
0?oOW0180000B9cO2P3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo00`0oooo0P00
002J0?ooo`030000003oool0oooo00`0oooo000[0?ooo`030?oog`2LB00007Bo00H0oooo00<0okmd
001d_`3oool0103oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00T0oooo00<00000
0?ooo`3oool0MP3oool2000002l0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`2J
0?ooo`030000003oool0oooo00`0oooo000/0?ooo`030=nLB000B9`0gooo00H0oooo00<0okmd001d
_`3oool00`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<00000
0?ooo`3oool0M@3oool4000000d0oooo0P00000O0?ooo`030000003oool0oooo00T0oooo0P00002M
0?ooo`030000003oool0oooo00`0oooo000/0?ooo`030?oOW01804P0W=oo00L0oooo00<0okmd04RL
g`3oool00P3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<000000?oo
o`3oool0M03oool6000000/0oooo1000000N0?ooo`030000003oool0oooo00L0oooo0P00002O0?oo
o`030000003oool0oooo00`0oooo000/0?ooo`030?oo_`1d0000M;oo00P0oooo00<0gia804RLg`3o
ool00P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo0`00001d0?ooo`H0
00002P3oool6000001d0oooo00<000000?ooo`3oool01@3oool2000009l0oooo0`00000>0?ooo`00
;@3oool00`2oM00007Bo0?ooo`080?ooo`070;md001d_ol0oooo0?ooo`3OW4P0001809cOo`020?oo
o`030?oog`2LB00007Bo00P0oooo00<000000?ooo`3oool0M@3oool4000000/0oooo1P00000M0?oo
o`030000003oool0oooo00<0oooo0P00002S0?ooo`030000003oool0oooo00`0oooo000]0?ooo`03
0=nLB000B9`0gooo00P0oooo00L0omnL04P0M02oool0oooo0?oOW0180000B9cO00<0oooo00<0_g@0
0018W03Oool01`3oool00`000000oooo0?ooo`1f0?ooo`8000003@3oool4000001h0oooo00@00000
0?ooo`3oool0oooo0P00002U0?ooo`030000003oool0oooo00`0oooo000]0?ooo`030?oOW01804P0
W=oo00P0oooo00P0oooO09a8001d_ol0oooo0?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool01`3oool00`000000oooo0?ooo`260?ooo`8000007`3oool010000000oooo00000000002W0?oo
o`030000003oool0oooo00`0oooo000]0?ooo`030?oo_`1d0000M;oo00P0oooo00<0oono07@00018
W=l00P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<000000?ooo`3o
ool0Y`3oool200000:T0oooo00<000000?ooo`3oool0303oool002h0oooo00<0_g@0001d_`3oool0
0`3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol00`3oool01P3OW4P0001d0;ooo`3ooml0
W4P004RLg`P0oooo00<000000?ooo`3oool0Y@3oool300000:X0oooo00<000000?ooo`3oool0303o
ool002`0oooo00D0oono07@000000000001d0;ooo`040?ooo`050?oOW018000000000000001d_ol0
1@3oool0103OW4P0000000000000M;l90?ooo`030000003oool0oooo0:<0oooo0P0000020?ooo`03
0000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0
X@3oool2000000@0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03o
ool200000:00oooo0P0000060?ooo`030000003oool0oooo0:L0oooo0P00000>0?ooo`00C03oool0
0`000000oooo0?ooo`2M0?ooo`800000203oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo00`0oooo001<0?ooo`030000003oool0oooo09/0oooo0P00000:0?ooo`030000003oool0oooo
0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0VP3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo
00<000000?ooo`3oool0V03oool2000000d0oooo00<000000?ooo`3oool0Z03oool00`000000oooo
0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2F0?ooo`8000003`3oool00`000000oooo0?oo
o`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo09@0oooo0P00000A
0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool00200oooo00<0_g@00018
W03Oool0:@3oool00`000000oooo0?ooo`2B0?ooo`8000004`3oool00`000000oooo0?ooo`2X0?oo
o`030000003oool0oooo00`0oooo000P0?ooo`030?oOW018M;l0oooo02T0oooo0P00002A0?ooo`80
00005@3oool00`000000oooo0?ooo`2W0?ooo`8000003P3oool001D0oooo00<0oono07Ad_`3oool0
1`3oool00`3o_g@004RL0=ooo`0Z0?ooo`030000003oool0oooo08h0oooo0P00000G0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool001D0oooo00@0omnL04P0000007@0_ooo
1@3oool00`3OW4P007Bo0?ooo`0[0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000F0?ooo`030?oOW018
0000000000D0000000<0B9cO0?ooo`3oool0:`3oool00`000000oooo0?ooo`2;0?ooo`8000006P3o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000K0?ooo`030?noM00004P0
W=oo02h0oooo00<000000?ooo`3oool0R@3oool2000001`0oooo00<000000?ooo`3oool0Z03oool0
0`000000oooo0?ooo`0<0?ooo`006P3oool00`2oM00007Bo0?ooo`0_0?ooo`030000003oool0oooo
08L0oooo0P00000N0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool001D0
oooo00H0omnL04P0001d_ol0okmd0000001d_ola0?ooo`030000003oool0oooo08D0oooo0P00000P
0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool001D0oooo00@0omnL04P0
0000000007Bo<`3oool2000008@0oooo0P00000R0?ooo`030000003oool0oooo0:L0oooo0P00000>
0?ooo`00C03oool00`000000oooo0?ooo`210?ooo`800000903oool00`000000oooo0?ooo`2X0?oo
o`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo07l0oooo0P00000V0?ooo`03
0000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0
O@3oool2000002P0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03o
ool00`000000oooo0?ooo`1k0?ooo`800000:P3oool00`000000oooo0?ooo`2X0?ooo`030000003o
ool0oooo00`0oooo001<0?ooo`030000003oool0oooo07T0oooo0P00000/0?ooo`030000003oool0
oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0M`3oool20000
02h0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000
oooo0?ooo`1e0?ooo`800000<03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0
oooo001<0?ooo`800000M03oool200000380oooo00<000000?ooo`3oool0Y`3oool2000000h0oooo
001<0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?oo
o`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo06D0oooo0P000009
0?ooo`800000=@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000Y0?oo
o`030?oo_`1d000000000080000000H0B9cO0?ooo`3oool0oooo0?oo_`1d00030000000307Boo`3o
ool0oooo00<0oooo00@0_g@0000000000000M;oo2`3oool00`000000oooo0?ooo`1T0?ooo`@00000
1P3oool2000003L0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00:`3o
ool00`3o_g@0000007Boo`040?ooo`030?noM000000007Bo0080oooo00<0_gA809cOo`3oool00P3o
ool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`X0oooo00<000000?ooo`3oool0H`3oool6000000<0
oooo0P00000i0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool002/0oooo
00<0oooO09a80000M;l02@3oool01`3ogi`0B4RL0=ooo`3oool0oooO09a80018W=l00P3oool00`3o
_g@0001809cOo`090?ooo`030000003oool0oooo06<0oooo1P0000000`3oool000000000000k0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool002`0oooo00<0gia80018W03O
ool0203oool01`3ookl0M00007Boo`3oool0oooO09a80018W=l00P3oool0103ooml0W4P00018W03O
ool80?ooo`030000003oool0oooo06@0oooo1P00000m0?ooo`030000003oool0oooo0:P0oooo00<0
00000?ooo`3oool0303oool002`0oooo00<0omnL04P0B02Lgol0203oool01`3ogi`0B00007Boo`3o
ool0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`080?ooo`030000003oool0oooo06D0oooo
0`00000o0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool002`0oooo00<0
oono07@0001d_ol01`3oool0103ogi`0B0000000B02Lgol20?ooo`030;md0000B9`0gooo0080oooo
00<0omnL04P00018W=l0203oool3000006@0oooo0P0000110?ooo`030000003oool0oooo0:H0oooo
0`00000>0?ooo`00;@3oool00`2oM00007Bo0?ooo`040?ooo`050?oOW018000000000000001d_ol0
0`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00<000000?ooo`3oool0
HP3oool2000004<0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00;@3o
ool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d000007Bo00<0oooo00<0omnL04P00018W=l00`3o
ool00`2oM00004RL0=ooo`070?ooo`030000003oool0oooo0600oooo0P0000150?ooo`030000003o
ool0oooo0:P0oooo00<000000?ooo`3oool0303oool002d0oooo00<0omnL04P0B02Lgol0203oool0
203OW4P007Bo0?ooo`3oool0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`070?ooo`03
0000003oool0oooo05h0oooo0P0000170?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0303oool002d0oooo00<0oono07@0001d_ol0203oool00`3ogi`0B01809cOo`020?ooo`030?oO
W0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`1L0?ooo`800000
B@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000^0?ooo`030;md0000
M;l0oooo00@0oooo00H0_g@007Boo`3oool0gia80000B02Lgol30?ooo`060=nLB00007@0_ooo0?oo
g`2LB000B9cO203oool00`000000oooo0?ooo`1J0?ooo`800000B`3oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00`0oooo000/0?ooo`050?oo_`1d000000000000M02oool01@3oool0
102oM000000000000018W=l50?ooo`040=nLB00000000000001d_`T0oooo00<000000?ooo`3oool0
E@3oool2000000030?ooo`000000000004d0oooo00<000000?ooo`3oool0Z03oool00`000000oooo
0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`1D0?ooo`@00000C`3oool00`000000oooo0?oo
o`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`800000E03oool6000004h0oooo00<00000
0?ooo`3oool0Y`3oool2000000h0oooo001<0?ooo`030000003oool0oooo0580oooo1`00001>0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3o
ool0D03oool200000080oooo1000001?0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0303oool004`0oooo00<000000?ooo`3oool0CP3oool2000000D0oooo0P00001@0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0C03o
ool2000005T0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool0
0`000000oooo0?ooo`1:0?ooo`800000F`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo00`0oooo001<0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0F`3oool00`00
0000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`800000B03oool2000005h0
oooo00<000000?ooo`3oool0Y`3oool2000000h0oooo001<0?ooo`030000003oool0oooo04D0oooo
0P00001P0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<0
00000?ooo`3oool0@`3oool200000680oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?oo
o`0<0?ooo`00C03oool00`000000oooo0?ooo`110?ooo`800000I03oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo03l0oooo0P00001V0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3o
ool0?@3oool2000006P0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00
C03oool00`000000oooo0?ooo`0k0?ooo`800000JP3oool00`000000oooo0?ooo`2X0?ooo`030000
003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo03T0oooo0P00001/0?ooo`030000003o
ool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo0P00000h0?ooo`800000KP3oool0
0`000000oooo0?ooo`2W0?ooo`8000003P3oool004`0oooo00<000000?ooo`3oool0=@3oool20000
0700oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000
oooo0?ooo`0c0?ooo`800000LP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0
oooo001<0?ooo`030000003oool0oooo0340oooo0P00001d0?ooo`030000003oool0oooo0:P0oooo
00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0;`3oool2000001@0oooo0P00
001P0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`0C0?ooo`@00000G`3oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo02`0oooo0P00000E0?oo
o`H00000GP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`03
0000003oool0oooo02X0oooo0P00000G0?ooo`H00000GP3oool00`000000oooo0?ooo`2X0?ooo`03
0000003oool0oooo00`0oooo001<0?ooo`800000:@3oool2000001X0oooo1000001O0?ooo`030000
003oool0oooo0:L0oooo0P00000>0?ooo`00C03oool00`000000oooo0?ooo`0V0?ooo`8000007@3o
ool200000600oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool0
0`000000oooo0?ooo`0T0?ooo`800000P@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo00`0oooo000Y0?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool0103oool00`3o
oml0W4P004RLg`050?ooo`040;md00000000000007Boo`/0oooo00<000000?ooo`3oool08P3oool2
000008<0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00:`3oool00`3o
_g@0000007Boo`080?ooo`030=nLB000B9`0gooo00<0oooo00H0gia8001d_`3oool0omnL04P00018
W=l:0?ooo`030000003oool0oooo0200oooo0P0000250?ooo`030000003oool0oooo0:P0oooo00<0
00000?ooo`3oool0303oool002/0oooo00<0oooO09a80000M;l0203oool00`3ogi`0B01809cOo`02
0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02@3oool00`000000oooo0?ooo`0N
0?ooo`800000Q`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000/0?oo
o`030=nLB000B9`0gooo00<0oooo00<0okmd000000000000100000001@00B9`0gooo0?oog`2LB000
B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?ooo`0L0?ooo`800000R@3o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000/0?ooo`030?oOW01804P0
W=oo00<0oooo00<0oooO09a8B02Lgol00P3oool01`2oM00007Bo0?ooo`3oool0oooO09a80000M;l0
0`3oool00`3OW4P0001809cOo`080?ooo`030000003oool0oooo01X0oooo0P00002;0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool002`0oooo00<0oono07@0001d_ol0103o
ool01`3ookl0M01d0;ooo`3oool0okmd0018W03Oool00P3oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO00P0oooo0`00000I0?ooo`030000003oool0oooo01T0oooo0P00001`0?ooo`03
0000003oool0oooo0:H0oooo0`00000>0?ooo`00;@3oool00`2oM00007Bo0?ooo`040?ooo`060?oO
W018B9`0gooo0?oOW01804P0W=oo0P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo
00P0oooo00<000000?ooo`3oool05`3oool2000001/0oooo1000001_0?ooo`030000003oool0oooo
0:P0oooo00<000000?ooo`3oool0303oool002d0oooo00<0gia80018W03Oool01@3oool01@3o_g@0
07Bo0?oog`2LB000B9cO0080oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?oo
o`030000003oool0oooo01D0oooo0P00000L0?ooo`H00000KP3oool00`000000oooo0?ooo`2X0?oo
o`030000003oool0oooo00`0oooo000]0?ooo`030?oOW01804P0W=oo00H0oooo00@0gia804RLg`2o
M00007Bo0P3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00L0oooo00<00000
0?ooo`3oool04`3oool2000000l0oooo0P00000=0?ooo`H00000KP3oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00`0oooo000]0?ooo`030?oo_`1d0000M;oo00L0oooo00@0gia807Ad
M00007@0_ooo0P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0>0?ooo`@000003@3oool4000006l0oooo00<00000
0?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00;P3oool00`2oM00007Bo0?ooo`070?oo
o`030;md000004P0W=oo00<0oooo00H0gia80000M02oool0oooO09a80018W=l80?ooo`030000003o
ool0oooo0100oooo0P00000@0?ooo`H000003@3oool200000700oooo00<000000?ooo`3oool0Z03o
ool00`000000oooo0?ooo`0<0?ooo`00;03oool01@3ookl0M0000000000007@0_ooo00L0oooo00<0
oooO09a80018W=l0103oool0103OW4P0000000000000M;l90?ooo`030000003oool0oooo00h0oooo
0P00000B0?ooo`H00000O`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo
001<0?ooo`030000003oool0oooo00`0oooo0P00000E0?ooo`@00000P03oool00`000000oooo0?oo
o`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`8000002`3oool2000001P0oooo0P000021
0?ooo`030000003oool0oooo0:L0oooo0P00000>0?ooo`00C03oool00`000000oooo0?ooo`080?oo
o`800000W@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`03
0000003oool0oooo00H0oooo0P00002O0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0303oool004`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`2O0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0
oooo0:L0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`00
0000oooo0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0
oooo0P00002X0?ooo`030000003oool0oooo0:L0oooo0P00000>0?ooo`00C03oool00`000000oooo
0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<0
00000?ooo`3oool0Y`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<
0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<
0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?oo
o`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`2X0?ooo`03
0000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0
Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003o
ool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo0P00002X0?ooo`030000003oool0
oooo0:L0oooo0P00000>0?ooo`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo
0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000
oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<0?ooo`030000003oool0oooo0:L0oooo
00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00C03oool00`000000oooo0?oo
o`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool004`0oooo00<00000
0?ooo`3oool0Y`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo001<0?oo
o`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`0<0?oo
o`00C03oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0303oool004`0oooo0P00002X0?ooo`030000003oool0oooo0:L0oooo0P00000>0?ooo`00C03o
ool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303o
ool004`0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo00`0oooo000Y0?ooo`030?oo_`1d000000000080000000H0B9cO0?ooo`3oool0oooo0?oOW018
0002000000030018W03Oool0oooo00@0oooo00@0_g@0000000000000M;oo2`3oool00`000000oooo
0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool002/0oooo00<0
okmd0000001d_ol0103oool01P3o_g@0000007Boo`3oool0omnL04Qd_`@0oooo00H0gia8001d_`3o
ool0omnL04P00018W=l:0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0Z03oool0
0`000000oooo0?ooo`0<0?ooo`00:`3oool00`3ooml0W4P0001d_`090?ooo`070?noM018W=l0oooo
0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00T0oooo00<000000?ooo`3oool0Y`3o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000/0?ooo`030=nLB000B9`0
gooo00P0oooo00L0oono07@0M02oool0oooo0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0
gooo203oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0303oool002`0oooo00<0omnL04P0B02Lgol0203oool01`3ooml0W4P007Boo`3oool0oooO09a8
0000M;l00`3oool00`3OW4P0001809cOo`080?ooo`030000003oool0oooo0:L0oooo00<000000?oo
o`3oool0Z03oool00`000000oooo0?ooo`0<0?ooo`00;03oool00`3ookl0M00007Boo`080?ooo`03
0?oo_`1d0000B9cO0080oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`080?ooo`<0
0000Y`3oool00`000000oooo0?ooo`2V0?ooo`<000003P3oool002d0oooo00<0_g@0001d_`3oool0
1`3oool00`3o_g@0000007Boo`020?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0
203oool00`000000oooo0?ooo`0D0?ooo`800000T@3oool00`000000oooo0?ooo`2X0?ooo`030000
003oool0oooo00`0oooo000]0?ooo`030=nLB000B9`0gooo00@0oooo00H0omoO0=ooo`3o_g@00000
0000M02oool20?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01`3oool00`000000
oooo0?ooo`0C0?ooo`@00000T03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0
oooo000]0?ooo`030?oOW01804P0W=oo00@0oooo00D0oono07@000000000001809cOo`030?ooo`04
0?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01`3oool00`000000oooo0?ooo`0B0?oo
o`H00000S`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00`0oooo000]0?ooo`03
0?oo_`1d0000M;oo00D0oooo00<0gibL0;ooo`3oool01@3oool00`3ogi`0B00007Boo`020?ooo`03
0=nLB000M;l0oooo00L0oooo00<000000?ooo`3oool04P3oool6000008l0oooo00<000000?ooo`3o
ool0Z03oool00`000000oooo0?ooo`0<0?ooo`00;P3oool00`2oM00007Bo0?ooo`040?ooo`030?oO
W01dW=l0oooo00H0oooo00H0gia80000M02oool0oooO09a80018W=l80?ooo`030000003oool0oooo
01<0oooo1000002@0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0303oool002`0
oooo00D0oono07@000000000001d0;ooo`050?ooo`050;md00000000000000000018W=l0103oool0
103OW4P0000000000000M;l90?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0303o
ool200000400oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`1?0?ooo`030000003o
ool0oooo04h0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`00C03oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00`0oooo001<0?ooool00000E`00000>0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.51796, 84.7086, \
0.00922308, 0.385417}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        ListPlot[Evaluate[Transpose[{x, \(y - y\&_\)\/\[Sigma]\_y}]], 
          PlotRange \[Rule] {Automatic, {\(-3\), 3}}, 
          PlotStyle \[Rule] PointSize[0.02], Frame \[Rule] True, 
          FrameLabel \[Rule] {"\<x\>", "\<residuals\>"}], 
        Plot[{\(-1\), 1}, {x, \(-1\), 1}, 
          PlotStyle \[Rule] Dashing[{0.02, 0.02}]]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.103006 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .10301 -18.6875 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -18.6875 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -10 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[-0.0125 .61803 -10 -7.84375 ]
[-0.0125 .61803 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.375 -23.9688 ]
[-0.01875 .30902 -18.6875 23.9688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .61803 m
.00625 .61803 L
s
gsave
-0.0125 .61803 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .0206 m
.00375 .0206 L
s
0 .0412 m
.00375 .0412 L
s
0 .0618 m
.00375 .0618 L
s
0 .0824 m
.00375 .0824 L
s
0 .12361 m
.00375 .12361 L
s
0 .14421 m
.00375 .14421 L
s
0 .16481 m
.00375 .16481 L
s
0 .18541 m
.00375 .18541 L
s
0 .22661 m
.00375 .22661 L
s
0 .24721 m
.00375 .24721 L
s
0 .26781 m
.00375 .26781 L
s
0 .28842 m
.00375 .28842 L
s
0 .32962 m
.00375 .32962 L
s
0 .35022 m
.00375 .35022 L
s
0 .37082 m
.00375 .37082 L
s
0 .39142 m
.00375 .39142 L
s
0 .43262 m
.00375 .43262 L
s
0 .45322 m
.00375 .45322 L
s
0 .47383 m
.00375 .47383 L
s
0 .49443 m
.00375 .49443 L
s
0 .53563 m
.00375 .53563 L
s
0 .55623 m
.00375 .55623 L
s
0 .57683 m
.00375 .57683 L
s
0 .59743 m
.00375 .59743 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -111.5 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
84.9688 11.8438 translate
90 rotate
-84.9688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(residuals) show
106.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .0206 m
1 .0206 L
s
.99625 .0412 m
1 .0412 L
s
.99625 .0618 m
1 .0618 L
s
.99625 .0824 m
1 .0824 L
s
.99625 .12361 m
1 .12361 L
s
.99625 .14421 m
1 .14421 L
s
.99625 .16481 m
1 .16481 L
s
.99625 .18541 m
1 .18541 L
s
.99625 .22661 m
1 .22661 L
s
.99625 .24721 m
1 .24721 L
s
.99625 .26781 m
1 .26781 L
s
.99625 .28842 m
1 .28842 L
s
.99625 .32962 m
1 .32962 L
s
.99625 .35022 m
1 .35022 L
s
.99625 .37082 m
1 .37082 L
s
.99625 .39142 m
1 .39142 L
s
.99625 .43262 m
1 .43262 L
s
.99625 .45322 m
1 .45322 L
s
.99625 .47383 m
1 .47383 L
s
.99625 .49443 m
1 .49443 L
s
.99625 .53563 m
1 .53563 L
s
.99625 .55623 m
1 .55623 L
s
.99625 .57683 m
1 .57683 L
s
.99625 .59743 m
1 .59743 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .30902 m
1 .30902 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.07143 .49933 Mdot
.11905 .36269 Mdot
.16667 .39202 Mdot
.21429 .37307 Mdot
.2619 .31239 Mdot
.30952 .29856 Mdot
.35714 .16437 Mdot
.40476 .20934 Mdot
.45238 .19581 Mdot
.5 .19565 Mdot
.54762 .20394 Mdot
.59524 .18241 Mdot
.64286 .20911 Mdot
.69048 .36311 Mdot
.7381 .3151 Mdot
.78571 .36727 Mdot
.83333 .40477 Mdot
.88095 .42061 Mdot
.92857 .40177 Mdot
.5 Mabswid
[ .02 .02 ] 0 setdash
.02381 .20601 m
.06244 .20601 L
.10458 .20601 L
.14415 .20601 L
.18221 .20601 L
.22272 .20601 L
.26171 .20601 L
.30316 .20601 L
.34309 .20601 L
.3815 .20601 L
.42237 .20601 L
.46172 .20601 L
.49955 .20601 L
.53984 .20601 L
.57861 .20601 L
.61984 .20601 L
.65954 .20601 L
.69774 .20601 L
.73838 .20601 L
.77751 .20601 L
.81909 .20601 L
.85916 .20601 L
.89771 .20601 L
.93871 .20601 L
.97619 .20601 L
s
.02381 .41202 m
.06244 .41202 L
.10458 .41202 L
.14415 .41202 L
.18221 .41202 L
.22272 .41202 L
.26171 .41202 L
.30316 .41202 L
.34309 .41202 L
.3815 .41202 L
.42237 .41202 L
.46172 .41202 L
.49955 .41202 L
.53984 .41202 L
.57861 .41202 L
.61984 .41202 L
.65954 .41202 L
.69774 .41202 L
.73838 .41202 L
.77751 .41202 L
.81909 .41202 L
.85916 .41202 L
.89771 .41202 L
.93871 .41202 L
.97619 .41202 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00k`3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo^@3oool00?40oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;P0oooo003b0?ooo`040;mdM02LW4P004RL0=oookX0
oooo003b0?ooo`030?oo_`1d000007Bo0;/0oooo003b0?ooo`030?oo_`1d0000M;oo0;/0oooo003b
0?ooo`040?noM000000004RL0=oookX0oooo003b0?ooo`050;md0000B9`0gmnL04Q8W03Oool0^@3o
ool00?00oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]`3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005T0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`120?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
oc`0oooo00@0_g@0000000000000M;ooA03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom10?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool0703oool005/0oooo00<0okmd0000001d_ol0A03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_c/0oooo00H0gia8001d_`3oool0omnL04P00018W=m20?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
@`3oool00`3o_g@0000007Boo`0O0?ooo`00F`3oool00`3ooml0W4P0001d_`130?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0h0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0@@3oool00`3ooml0W4P0001d_`0O0?ooo`00G03oool0
0`3OW4P004RL0=ooo`120?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0h0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
?`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0@P3oool00`3OW4P004RL0=ooo`0N0?ooo`00C03oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0o0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`120?ooo`030?oOW01804P0W=oo
01h0oooo001L0?ooo`030?oo_`1d0000M;oo04<0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03T0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`100?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`120?ooo`030?oo_`1d0000M;oo01h0oooo001M0?ooo`030;md0000M;l0oooo0480oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03T0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`100?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`130?ooo`030;md0000M;l0oooo01d0oooo
001M0?ooo`030=nLB000B9`0gooo0480oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo>@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo03l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo@`3oool00`3OW4P004RL0=ooo`0M0?ooo`00G@3oool00`3o
gi`0B01809cOo`120?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`0o0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04@0oooo00<0omnL04P0B02Lgol07@3oool005d0oooo00<0oono07@0001d
_ol0@`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0?03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0400oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04D0oooo00<0oono07@0001d
_ol07@3oool005h0oooo00<0_g@0001d_`3oool0@`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0?@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd80oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04H0oooo00<0_g@0001d_`3o
ool0703oool005`0oooo00D0oono07@000000000001d0;ooo`140?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0k0?ooo`040=nLB00000000000001d_d@0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO0440oooo00D0oono07@00000
0000001d0;ooo`0L0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00D03ooooo000004X000005`3oool00500oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04`0oooo
00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool0
0500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0
X03oool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0E0?ooo`00@03oool00`3OW4P00000000000030000000304RLg`3oool0
oooo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo00110?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool01`3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00480
oooo00<0okmd001d_`3oool02`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool004<0oooo00<0okmd001d_`3oool02P3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool003<0oooo00<0
oooO09a8000000001`0000000`18W=l0oooo0?ooo`040?ooo`030?noM018W=l0oooo00T0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo0015
0?ooo`030=nLB018W=l0oooo00P0oooo0`00002Q0?ooo`030000003oool0oooo09l0oooo0`00000G
0?ooo`00AP3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ogi`0B01d0;ooo`070?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
AP3oool00`3ooml0W4P007Boo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RLg`070?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0omnL04P0000000000000
07Boo`080?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0
oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0L@3oool2000002h0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00D03oool00`000000oooo0?ooo`1`0?ooo`@00000;@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo001@0?ooo`800000L03oool6000002`0oooo00<000000?ooo`3o
ool0X03oool2000001L0oooo001@0?ooo`030000003oool0oooo06l0oooo1P00000/0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0L03o
ool4000002d0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool0
0`000000oooo0?ooo`1a0?ooo`800000;P3oool00`000000oooo0?ooo`0K0?ooo`800000Q03oool0
0`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
01X0oooo100000230?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool06@3oool600000880oooo00<000000?ooo`3oool05@3oool00500oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0I0?ooo`H00000PP3oool00`000000oooo
0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool06P3oool400000880oooo0P00
000G0?ooo`00@03oool00`3ookl0M000000000020000000304RLg`3oool0oooo00P0oooo00<00000
0?ooo`3oool0T03oool2000000h0oooo0P00000M0?ooo`800000Q03oool00`000000oooo0?ooo`0E
0?ooo`00@P3oool00`3o_g@0000007Boo`0;0?ooo`030000003oool0oooo08l0oooo1000000<0?oo
o`@00000XP3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool00`3ooml0W4P0001d_`0;0?ooo`03
0000003oool0oooo08h0oooo1P00000:0?ooo`H00000X@3oool00`000000oooo0?ooo`0E0?ooo`00
@`3oool00`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo08h0oooo1P00000:0?ooo`H00000
303oool2000009<0oooo00<000000?ooo`3oool05@3oool003<0oooo00<0oooO09a8000000001`00
00000`18W=l0oooo0?ooo`030?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool0P@3o
ool2000000`0oooo1000000<0?ooo`@00000303oool4000001`0oooo0P00001d0?ooo`030000003o
ool0oooo01D0oooo00130?ooo`030?oo_`1d0000M;oo00X0oooo0`0000040?ooo`P000001P3oool7
000000H0oooo1`0000060?ooo`L000001P3oool7000000H0oooo200000050?ooo`P000001@3oool8
000000D0oooo200000060?ooo`T00000103oool7000000030?ooo`000000000000<0oooo1`000004
0?ooo`X00000103oool9000000D0oooo200000050?ooo`T000001@3oool7000000H0oooo1`000006
0?ooo`L000001P3oool8000000D0oooo200000050?ooo`P000001@3oool8000000D0oooo20000009
0?ooo`<000005`3oool004@0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`1o0?oo
o`H00000703oool00`000000oooo0?ooo`0:0?ooo`H000006@3oool600000780oooo00<000000?oo
o`3oool05@3oool004@0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?ooo`1o0?ooo`H0
0000703oool00`000000oooo0?ooo`0;0?ooo`@000006P3oool600000780oooo00<000000?ooo`3o
ool05@3oool002D0oooo00@0oono07@0000007@0_ooo6`3oool00`3ogi`0B01809cOo`090?ooo`03
0000003oool0oooo0800oooo1000000M0?ooo`030000003oool0oooo00`0oooo0P00000L0?ooo`@0
0000L`3oool00`000000oooo0?ooo`0E0?ooo`007`3oool0103ookl0W9cO0?oo_`1d000400000003
07Boo`3oool0oooo01X0oooo00<0oono07@0001d_ol02@3oool00`000000oooo0?ooo`210?ooo`80
00007P3oool00`000000oooo0?ooo`0[0?ooo`800000M03oool00`000000oooo0?ooo`0E0?ooo`00
7`3oool01P3ogi`0B000000000000000001809cOob00oooo00<0_g@0001d_`3oool0203oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00200
oooo00<0oono07@0001d_ol0803oool01@3ookl0M0000000000007@0_ooo00P0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000P0?ooo`03
0;mdM02oool0oooo02d0oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00
7`3oool0103o_g@000000018W03Oool]0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00280oooo00H0gia80000000000000000
0000B02LgolX0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0E0?ooo`00803oool02@3ookl0M00004RLg`3ogi`0W;oo0?ooo`3OW4P004RL0=ooo`0W
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00803oool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d0;ooo`0W0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3o
ool02@3ogi`0W;oo0?ooo`3ooml0W7BL0=ooo`3oool0omnL04Qd_`0X0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool02@3o_i`0
W=oo0?ooo`2oM000M;oo0?ooo`3OW4P0001d0;ooo`0X0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool001l0oooo00@0oono07@000000000M;oo;@3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000T0?oo
o`050?oog`2LB00000000018W03Oool09`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool05@3oool00200oooo00@0_g@00000000007@0_ooo0P3oool0
0`3ookl0M7Bo0?ooo`0W0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`007`3oool01P3ogi`0W;oo0?oog`2LB000001d0;ooo`80oooo00<0
gibL0;ooo`3oool09P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool001l0oooo00X0oknL09cOo`3oool0oooO09a8000004P0W=oo0?oo_`1d
B9`0gooo9`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool001l0oooo00<0oono07Ad_`3oool00P3oool0102oM00000000000M02ooolX0?oo
o`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool001l0oooo00@0omnL04P00000
0000M;oo;@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ookl0W9cO0?ooo`030?ooo`040?noM000
0000001d0;ooobL0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo01D0oooo000O0?ooo`040?oo_`2LW=l0oono07@000@0000000<004RL0=ooo`3oool0
9P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
5@3oool001l0oooo00T0omnL04P000000000001809cOo`3oool0oooO09a8M02oool0:03oool20000
0:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo000K0?ooo`040?oog`2LB000001d0;oo
oc40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
01D0oooo000S0?ooo`050=nLB000000000000000001d_ol06P3oool0102oM00000000000001d_ol:
0?ooo`030000003oool0oooo0640oooo0P00000n0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00240oooo00<0omnL04P00018W=l00P3oool00`3ogi`0B01d0;ooo`0H0?oo
o`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo0?ooo`1P0?ooo`@00000?@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000P0?ooo`030?oOW018B9`0
gooo00<0oooo00<0oono07A8W03Oool05`3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0
W=oo00P0oooo00<000000?ooo`3oool0G`3oool6000003`0oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`007`3oool00`3ooml0W7BL0=ooo`040?ooo`030;md001d_ol0oooo
01L0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`030000003oool0
oooo05l0oooo1P00000l0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool0
01l0oooo00<0omno0;oOo`3oool00P3oool00`3ooml0W4P007Boo`0I0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`1B0?ooo`800000303oool40000
03d0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool02P3o_g@0
M;oo0?ooo`3oool0omnL04P0B02Lgkl0M0000018W03OoolH0?ooo`030;md0000B9`0gooo0080oooo
00<0omnL04P00018W=l01`3ooooo000004X000005`3oool001l0oooo00<0oooO09a8000000001@00
00000`00B9`0gooo0?ooo`0G0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01`3o
ool00`000000oooo0?ooo`1@0?ooo`H00000B`3oool00`000000oooo0?ooo`190?ooo`@00000E03o
ool00`000000oooo0?ooo`0E0?ooo`006`3oool0102oW9`0gooo0?oo_`1d00040000000507Boo`3o
ool0oooo0;mdM02oool06@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo
00<000000?ooo`3oool0D03oool6000004/0oooo00<000000?ooo`3oool0B03oool6000005<0oooo
00<000000?ooo`3oool05@3oool001X0oooo00<0oooO09a8000000000P0000000`1d_ol0oooo0?oo
o`0O0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`000000oooo
0?ooo`1A0?ooo`@00000C03oool00`000000oooo0?ooo`180?ooo`H00000D`3oool00`000000oooo
0?ooo`0E0?ooo`006P3oool00`3ookl0W9cO0?ooo`0U0?ooo`030?oOW0180000M;oo0080oooo00<0
gia8001d_`3oool01P3oool00`000000oooo0?ooo`1B0?ooo`800000C@3oool00`000000oooo0?oo
o`190?ooo`@00000E03oool00`000000oooo0?ooo`0E0?ooo`00803oool01@2oM00004RL0=ooo`3o
okl0M0000080000000<0001d0;ooo`3oool06@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1:0?ooo`800000E@3oool00`000000
oooo0?ooo`0E0?ooo`007`3oool00`3ogi`0B000000000030000000404RLg`3oool0_ibL0=oooa/0
oooo00@0gia800000000000007Bo203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool001l0oooo00@0oooO09a8000007@0_ooo0`3oool00`3o
_g@0B9cO0?ooo`0W0?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool002D0
oooo00<0okmd001d_`3oool0:03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool002<0oooo00<0oooO0;no_`2LW;l0:P3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool002<0oooo00H0
omno09adB000000000000000M02ooolW0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ogi`0B00000000002000000050000
B02Lgol0oooo0;nLW03Oool09`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool001l0oooo00<0okmd0000M02oool0103oool00`2oM000M;oo
0?ooo`0W0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`009P3oool00`3Ogml0oooo0?ooo`0W0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`008`3oool01@2oM000000000000000
0000B9cO02P0oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`008@3oool0
0`3OW4P0000007Boo`020?ooo`030?oOW01807@0_ooo02L0oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000P0?ooo`030=nLB018W=l0oooo
00<0oooo00<0omnL04Qd_`3oool09`3oool00`000000oooo0?ooo`0S0?ooo`800000O03oool00`00
0000oooo0?ooo`0k0?ooo`800000I03oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ogi`0
M9cO0?ooo`030?ooo`030?noW018M9`0gooo02P0oooo00<000000?ooo`3oool08P3oool4000007/0
oooo00<000000?ooo`3oool0>P3oool4000001`0oooo0P0000150?ooo`030000003oool0oooo01D0
oooo000O0?ooo`030?noM000M;l0oooo0080oooo00D0omnL04QdM01d000004RL0=ooo`0W0?ooo`03
0000003oool0oooo0240oooo1P00001j0?ooo`030000003oool0oooo03T0oooo1P00000J0?ooo`@0
0000A03oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3o_g@0000000000004000000030018
W03Oool0oooo02H0oooo00<000000?ooo`3oool08@3oool6000001/0oooo0P00001M0?ooo`030000
003oool0oooo03T0oooo1P00000I0?ooo`H00000@`3oool00`000000oooo0?ooo`0E0?ooo`007`3o
ool01@3ogi`0B000000000000000B9cO0080oooo00<0_gA809cOo`3oool09`3oool00`000000oooo
0?ooo`0R0?ooo`@000006`3oool4000005`0oooo00<000000?ooo`3oool0>P3oool4000001X0oooo
1P0000130?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000903oool2000001/0oooo1P00
001K0?ooo`030000003oool0oooo03/0oooo0P00000L0?ooo`@00000@`3oool2000001L0oooo001@
0?ooo`030000003oool0oooo0400oooo1P00001K0?ooo`030000003oool0oooo05X0oooo0P000015
0?ooo`030000003oool0oooo01D0oooo000S0?ooo`060?noM000000000000000000007@0_ooo9`3o
ool00`000000oooo0?ooo`110?ooo`@00000G03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo000K0?ooo`060;nLW03Oool0oooo0?ooo`3ogi`0B0000`0000001000B9`0gooo
0?oo_`1dM;lX0?ooo`030000003oool0oooo0480oooo0P00001M0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool001X0oooo00<0oooO09a8000000000`0000000`00M;l0oooo
0?ooo`020?ooo`030?noM00007@0_ooo02P0oooo00<000000?ooo`3oool0<`3oool2000006`0oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`006P3oool00`3ookl0M00004RL
g`0c0?ooo`030000003oool0oooo0380oooo1000001[0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool002@0oooo00D0oooO09a80000000004RL0=ooo`0W0?ooo`800000<P3o
ool6000006X0oooo00<000000?ooo`3oool0X03oool2000001L0oooo000P0?ooo`040;md00000000
001d0;ooo`80oooo00<0oono07Ad_`3oool09`3oool00`000000oooo0?ooo`0a0?ooo`H00000JP3o
ool00`000000oooo0?ooo`290?ooo`8000005P3oool00`000000oooo0?ooo`0E0?ooo`007`3oool0
1P3ogi`0W;oo0?oog`2LB000001d0;ooo`80oooo00<0gibL0;ooo`3oool05P3oool00`3ookl0M000
000000020000000304RLg`3oool0oooo00P0oooo00<000000?ooo`3oool0<P3oool4000006/0oooo
00<000000?ooo`3oool0J@3oool2000001d0oooo1000000E0?ooo`030000003oool0oooo01D0oooo
000O0?ooo`0:0?noW02Lgol0oooo0?oog`2LB000001809cOo`3ookl0M4RL0=oooaT0oooo00<0okmd
0000001d_ol02`3oool00`000000oooo0?ooo`0c0?ooo`800000K03oool00`000000oooo0?ooo`1X
0?ooo`@000006`3oool6000001@0oooo00<000000?ooo`3oool05@3oool001l0oooo00<0oono07Ad
_`3oool00P3oool0102oM00000000000M02ooolJ0?ooo`030?oog`2LB00007Bo00/0oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`1W0?ooo`H000006P3oool6000001@0oooo00<00000
0?ooo`3oool05@3oool001l0oooo00@0omnL04P000000000M;oo803oool00`3OW4P004RL0=ooo`0:
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0I`3oool6000001/0oooo1000000E
0?ooo`030000003oool0oooo01D0oooo00130?ooo`030?oOW01804P0W=oo00X0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`1X0?ooo`@000007@3oool2000001H0oooo00<000000?oo
o`3oool05@3oool004<0oooo00<0oono07@0001d_ol02P3oool3000000@0oooo200000060?ooo`L0
00001P3oool7000000H0oooo1`0000060?ooo`L000001P3oool8000000D0oooo200000050?ooo`P0
00001@3oool8000000H0oooo1`0000060?ooo`L000001P3oool7000000D0oooo2@0000050?ooo`P0
00001@3oool8000000D0oooo200000060?ooo`L000001P3oool7000000H0oooo1`0000060?ooo`P0
00001@3oool8000000D0oooo200000050?ooo`P000001@3oool8000000T0oooo0`00000G0?ooo`00
A03oool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
N@3oool2000002H0oooo00<000000?ooo`3oool05@3oool004@0oooo00<0gia80018W03Oool02@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo07P0oooo1000000U0?ooo`030000003o
ool0oooo01D0oooo00140?ooo`030?oOW01804P0W=oo00T0oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`1g0?ooo`H00000903oool00`000000oooo0?ooo`0E0?ooo`00A03oool00`3o
okl0M00007Boo`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0M`3oool60000
02@0oooo00<000000?ooo`3oool05@3oool004D0oooo00<0_g@0001d_`3oool0203oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo07P0oooo1000000U0?ooo`030000003oool0oooo01D0
oooo00130?ooo`050?oo_`1d000000000000M02oool0203oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo07T0oooo0P00000V0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000
XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0
oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`00
0000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool05@3oool00500oooo0P00000D0?ooo`800000S03oool00`000000oooo0?ooo`2P0?ooo`800000
5`3oool00500oooo00<000000?ooo`3oool04P3oool4000008/0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0E0?ooo`00@03oool00`3OW4P00000000000030000000304RLg`3oool0
oooo00L0oooo00<000000?ooo`3oool04@3oool6000008X0oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`00@@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo
00L0oooo00<000000?ooo`3oool04@3oool6000008X0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00@P3oool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo0180
oooo1000002;0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo
00<0okmd001d_`3oool02P3oool00`000000oooo0?ooo`0C0?ooo`800000S03oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00140?ooo`030?noM018W=l0oooo00T0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo0015
0?ooo`030=nLB018W=l0oooo00P0oooo0`00002Q0?ooo`030000003oool0oooo09l0oooo0`00000G
0?ooo`00AP3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ogi`0B01d0;ooo`070?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
AP3oool00`3ooml0W4P007Boo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RLg`070?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0omnL04P0000000000000
07Boo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo001@
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0
oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo001@0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo00100?ooo`030?oo_`1d000000000080000000<0M;oo0?oo
o`3oool0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00400oooo00<0okmd00000000M;l00P3oool00`2oM4P0W=oo0?ooo`080?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
A@3oool00`3ogi`0B4RL0=ooo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00A@3oool00`3ookl0M00007Boo`080?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`1;0?ooo`030000003o
ool0oooo04/0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo01D0oooo00150?ooo`030?oOW0180000M;oo00P0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0E0?ooo`00A03oool0103ogi`0B0000000B02Lgol80?ooool00000BP00000G
0?ooo`00@P3oool01@3ogi`0B000000000000000M;oo0?l0ooooJP3oool004@0oooo00<0oono07@0
0000M;l0o`3ooomZ0?ooo`00AP3oool00`3OW4P007Bo0?ooo`3o0?ooofP0oooo00160?ooo`030?oO
W01804P0W=oo0?l0ooooJ03oool004<0oooo00H0_g@007Boo`3oool0gia80000B02Lgooo0?ooofP0
oooo00140?ooo`040;md00000000000004RLgol0ooooJ@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.56215, -4.51278, \
0.00959293, 0.0443476}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Basis functions", "Subsection"],

Cell["\<\
\tNext we consider a more general fitting function of the form\
\>", "Text"],

Cell[BoxData[
    \(TraditionalForm\`y[
        x]\  = \ \[Sum]\+\(\[Mu] = 1\)\%m\( a\_\[Mu]\) 
          f\_\[Mu][x]\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`a = {a\_\[Mu]}\)]],
  " is the set of parameters we seek and ",
  Cell[BoxData[
      \(TraditionalForm\`f = {f\_\[Mu]}\)]],
  " is a set of ",
  StyleBox["basis functions",
    FontSlant->"Italic"],
  " that depend upon the independent variable ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " but not upon the parameters ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Mu]\)]],
  ".  Therefore, ",
  Cell[BoxData[
      \(TraditionalForm\`y[x]\)]],
  " is a linear function of ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " parameters.  The normal equations"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[PartialD]\[Chi]\^2\/\[PartialD]a\_\[Mu] \[Equal] 
      0\ \ \[DoubleLongRightArrow]\ \ \ \(\[Sum]\+\(i = 1\)\%N\(\[PartialD]y[
                    x\_i]\/\[PartialD]a\_\[Mu]\) \(y\_i - y[x\_i]\)\/\[Sigma]\
\_i\^2\)\  \[Equal] \ 
      0\ \ \[DoubleLongRightArrow]\ \ \(\[Sum]\+\(\[Nu] = 1\)\%m\( 
                a\_\[Nu]\) \(\[Sum]\+\(i = 1\)\%N\( w\_i\) f\_\[Mu][x\_i] 
                f\_\[Nu][x\_i]\)\)\ \  \[Equal] \ \[Sum]\+\(i = 1\)\%N\( 
              w\_i\) f\_\[Mu][x\_i] y\_i\)], "DisplayFormula"],

Cell["can be expressed in the matrix form", "Text"],

Cell[BoxData[
    \(TraditionalForm\`M\ a\  \[Equal] \ b\)], "DisplayFormula"],

Cell["where", "Text"],

Cell[BoxData[
    \(TraditionalForm\`M\_\(\[Mu], \[Nu]\)\  = \ \[Sum]\+\(i = 1\)\%N\( 
              w\_i\) f\_\[Mu][x\_i] f\_\[Nu][x\_i]\)], "DisplayFormula"],

Cell[TextData[{
  "is the ",
  StyleBox["curvature matrix",
    FontSlant->"Italic"],
  " for the experiment, ",
  Cell[BoxData[
      \(TraditionalForm\`a = {a\_\[Mu], \[Mu] = 1, m}\)]],
  " is the parameter vector, and"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`b\_\[Mu]\  = \ \[Sum]\+\(i = 1\)\%N\( w\_i\) \(y\_i\) 
          f\_\[Mu][x\_i]\)], "DisplayFormula"],

Cell[TextData[{
  "is the ",
  StyleBox["gradient vector",
    FontSlant->"Italic"],
  ".  The terminology for ",
  Cell[BoxData[
      \(TraditionalForm\`\[Beta]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`M\)]],
  " arises from their origin in first and second derivatives of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  ".  Notice that the curvature matrix depends upon the values of the \
independent variable and the precision of the measurements, but not upon the \
actual results of those measurements.  It represents the sensitivity of the \
proposed measurements to the parameters of the system.  The gradient vector \
projects the actual measurements onto the basis functions.  Thus, the \
maximum-likelihood estimates of the model parameters are given by"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(a\&^\) = \(M\^\(-1\)\) b\)], "DisplayFormula"],

Cell["\<\
and the problem is reduced to inversion of a symmetric matrix.  One can \
easily show that the previous results are recovered for a two-parameter \
linear model.  (Try it!)\
\>", "Text"],

Cell[TextData[{
  "\tThe error analysis applies the same technique used for the \
two-dimensional problem.  It is convenient to identify the ",
  StyleBox["error matrix",
    FontSlant->"Italic"],
  " ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] = M\^\(-1\)\)]],
  " with the inverse of the curvature matrix; the reason for this terminology \
will become apparent momentarily.  Starting with"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][a\&^\_\[Mu]]\^2\  = \ \[Sum]\+\(i = 1\)\%N\((\
\(\[PartialD]a\&^\_\[Mu]\/\[PartialD]y\_i\) \[Sigma]\_i)\)\^2\)], \
"DisplayFormula"],

Cell["and using", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[PartialD]a\&^\_\[Mu]\/\[PartialD]y\_i\  = \ \(\[Sum]\
\+\(\[Nu] = 1\)\%m\( \[CurlyEpsilon]\_\(\[Mu], \[Nu]\)\) \
\[PartialD]b\_\[Nu]\/\[PartialD]y\_i\  = \ \ \[Sum]\+\(\[Nu] = 1\)\%m\( \
\[CurlyEpsilon]\_\(\[Mu], \[Nu]\)\) \(w\_i\) 
            f\_\[Nu][x\_i]\)\)], "DisplayFormula"],

Cell["we soon obtain", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][a\&^\_\[Mu]]\^2\  = \ \(\[Sum]\+i\( 
                w\_i\) \(\[Sum]\+\(\[Nu], \[Gamma]\)\((\(\[CurlyEpsilon]\_\(\
\[Mu], \[Nu]\)\) f\_\[Nu][x\_i] \(\[CurlyEpsilon]\_\(\[Mu], \[Gamma]\)\) 
                  f\_\[Gamma][
                    x\_i])\)\)\  = \ \(\[Sum]\+\(\[Nu], \[Gamma]\)\(\
\[CurlyEpsilon]\_\(\[Mu], \[Nu]\)\) \(\[CurlyEpsilon]\_\(\[Mu], \[Gamma]\)\) \
\(\[Sum]\+i\( w\_i\) f\_\[Nu][x\_i] 
                  f\_\[Gamma][
                    x\_i]\)\  = \(\[Sum]\+\(\[Nu], \
\[Gamma]\)\(\[CurlyEpsilon]\_\(\[Mu], \[Nu]\)\) \(\[CurlyEpsilon]\_\(\[Mu], \
\[Gamma]\)\) 
                M\_\(\[Gamma], \[Nu]\)\ \  = \ \[CurlyEpsilon]\_\(\[Mu], \
\[Mu]\)\)\)\)\)], "DisplayFormula"],

Cell[TextData[{
  "where in the last step contraction of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Gamma]\)]],
  " index yields the unit matrix ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\_\(\[Mu], \[Nu]\)\)]],
  " and contraction of the ",
  Cell[BoxData[
      \(TraditionalForm\`\[Nu]\)]],
  " index isolates the appropriate diagonal element.  Therefore, diagonal \
elements of the error matrix give the variances in the corresponding \
parameters. "
}], "Text"],

Cell[BoxData[{
    RowBox[{
      RowBox[{\(linfit::usage\), "=", 
        "\"\<linfit[data,f,var] performs linear least-squares fitting.  The \
\!\(\*
StyleBox[\"data\",\nFontSlant->\"Italic\"]\) array should be a list of either \
\!\(\*
StyleBox[\"{\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"x\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"y\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"dy\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"}\",\nFontSlant->\"Italic\"]\) or \!\(\*
StyleBox[\"{\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"x\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"y\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"}\",\nFontSlant->\"Italic\"]\) for each measurement, where the fit \
is unweighted if uncertainties in the dependent variable are omitted.  The \
list of basis functions is \!\(\*
StyleBox[\"f\",\nFontSlant->\"Italic\"]\) and \!\(\*
StyleBox[\"var\",\nFontSlant->\"Italic\"]\) is the symbol used in \!\(\*
StyleBox[\"f\",\nFontSlant->\"Italic\"]\) for the independent \
variable.\>\""}], ";"}], "\[IndentingNewLine]", 
    RowBox[{\(linfit::BadDim = "\<Bad dimensions in data array: `1`\>";\), 
      "\[IndentingNewLine]"}], "\[IndentingNewLine]", \(linfit[data_, 
        f_?VectorQ, var_, opts___Rule] := 
      Module[{x, y, dy, w, M, b, \[CurlyEpsilon], basis, ndata, nparam, ndim, 
          i, j, params, yfit, redchi, errors, 
          stdev}, \[IndentingNewLine]\[IndentingNewLine] (*\ Extract\ x, y, 
          dy\ vectors\ from\ data\ \(\(array\)\(.\)\)\ *) \[IndentingNewLine]\
\[IndentingNewLine]{ndata, ndim} = Dimensions[data]; \[IndentingNewLine]If[
          ndata \[LessEqual] Length[f] || ndim < 2 || 
            ndim > 3, \((Message[linfit::BadDim, {ndata, ndim}]; 
            Return[])\)]; \[IndentingNewLine]If[
          ndim \[Equal] 2, \[IndentingNewLine]\(({x, y} = Transpose[data]; 
            w = Table[1, {ndata}])\), \[IndentingNewLine]\(({x, y, dy} = 
              Transpose[data]; 
            w = 1/dy\^2)\)]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Accumulate\ the\ gradient\ vector\ and\ the\ curvature\ \
\(\(matrix\)\(.\)\)\ *) \[IndentingNewLine]\[IndentingNewLine]nparam = 
          Length[f]; \[IndentingNewLine]M = 
          Table[0, {i, nparam}, {j, nparam}]; \[IndentingNewLine]b = 
          Table[0, {i, 
              nparam}]; \[IndentingNewLine]Do[\[IndentingNewLine]basis = 
            f /. var \[Rule] 
                x\[LeftDoubleBracket]i\[RightDoubleBracket]; \
\[IndentingNewLine]M = 
            M + w\[LeftDoubleBracket]i\[RightDoubleBracket] 
                Outer[Times, basis, basis]; \[IndentingNewLine]b = 
            b + w\[LeftDoubleBracket]i\[RightDoubleBracket] 
                y\[LeftDoubleBracket]i\[RightDoubleBracket] 
                basis, \[IndentingNewLine]{i, 1, 
            ndata}]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Evaluate\ fit\ parameters, \ error\ matrix, \ 
          and\ reduced\ \(\(\[Chi]\^2\)\(.\)\)\ *) \[IndentingNewLine]\
\[IndentingNewLine]\[CurlyEpsilon] = 
          PseudoInverse[M]; \[IndentingNewLine]params = \[CurlyEpsilon] . 
            b; \[IndentingNewLine]errors = 
          Table[\@\[CurlyEpsilon]\[LeftDoubleBracket]i, \
i\[RightDoubleBracket], {i, 
              nparam}]; \[IndentingNewLine]yfit = \((params . f)\) /. 
            var \[Rule] x; \[IndentingNewLine]redchi = 
          w . \((y - yfit)\)\^2/\((ndata - 
                nparam)\); \[IndentingNewLine]stdev = \@redchi; \
\[IndentingNewLine]\[IndentingNewLine] (*\ 
          Return\ results\ as\ replacement\ \(\(rules\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]If[
          ndim \[Equal] 3, {Parameters \[Rule] params, 
            Uncertainties \[Rule] errors, 
            ReducedChiSquare \[Rule] redchi}, {Parameters \[Rule] params, 
            Uncertainties \[Rule] errors*stdev, 
            StandardDeviation \[Rule] stdev}]]\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?linfit\)\)], "Input"],

Cell[BoxData["\<\"linfit[data,f,var] performs linear least-squares fitting.  \
The \!\(\* StyleBox[\\\"data\\\",\\nFontSlant->\\\"Italic\\\"]\) array should \
be a list of either \!\(\* StyleBox[\\\"{\\\",\\nFontSlant->\\\"Italic\\\"]\)\
\!\(\* StyleBox[\\\"x\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"y\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"dy\\\",\
\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Italic\\\"]\) or \!\(\* StyleBox[\\\"{\\\
\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"x\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"y\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Italic\\\"]\) for each measurement, \
where the fit is unweighted if uncertainties in the dependent variable are \
omitted.  The list of basis functions is \!\(\* \
StyleBox[\\\"f\\\",\\nFontSlant->\\\"Italic\\\"]\) and \!\(\* \
StyleBox[\\\"var\\\",\\nFontSlant->\\\"Italic\\\"]\) is the symbol used in \!\
\(\* StyleBox[\\\"f\\\",\\nFontSlant->\\\"Italic\\\"]\) for the independent \
variable.\"\>"], "Print",
  CellTags->"Info3351438211-6959969"]
}, Closed]],

Cell[CellGroupData[{

Cell["example", "Subsubsection"],

Cell[TextData[{
  "\tWe return to the same data set used in the preceding example based upon \
a parabolic model.  If we fit these data using a quadratic, we obtain reduced \
chi-square near unity and recover the model parameters with deviations \
similar to their estimated uncertainties.  The points scatter normally around \
the fit with residuals that show no systematic dependence upon ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  ".  Life is good!"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(f = {1, z, z\^2};\)\), "\[IndentingNewLine]", 
    \(\(dy = Table[2. , {Length[y]}];\)\), "\[IndentingNewLine]", 
    \(\(data = Transpose[{x, y, dy}];\)\), "\[IndentingNewLine]", 
    \(results = linfit[data, f, z]\)}], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {115.70263221247653`, \(-21.78902630957929`\), 
          17.585109991096164`}, 
      Uncertainties \[Rule] {0.6898518836724719`, 0.8377078165833909`, 
          1.7171336116297728`}, 
      ReducedChiSquare \[Rule] 1.6263437235968718`}\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(\(y\&_ = \(Parameters . f /. results\) /. z \[Rule] x;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[\[IndentingNewLine]ErrorListPlot[
          data], \[IndentingNewLine]Plot[
          Evaluate[Parameters . f /. results], {z, \(-1\), 1}], 
        Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 -1.19665 0.0116069 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .08012 -22 -7.84375 ]
[-0.0125 .08012 0 7.84375 ]
[-0.0125 .19619 -22 -7.84375 ]
[-0.0125 .19619 0 7.84375 ]
[-0.0125 .31225 -22 -7.84375 ]
[-0.0125 .31225 0 7.84375 ]
[-0.0125 .42832 -22 -7.84375 ]
[-0.0125 .42832 0 7.84375 ]
[-0.0125 .54439 -22 -7.84375 ]
[-0.0125 .54439 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -37.6875 -5.09375 ]
[-0.01875 .30902 -22 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .08012 m
.00625 .08012 L
s
gsave
-0.0125 .08012 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(110) show
1.000 setlinewidth
grestore
0 .19619 m
.00625 .19619 L
s
gsave
-0.0125 .19619 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(120) show
1.000 setlinewidth
grestore
0 .31225 m
.00625 .31225 L
s
gsave
-0.0125 .31225 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(130) show
1.000 setlinewidth
grestore
0 .42832 m
.00625 .42832 L
s
gsave
-0.0125 .42832 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(140) show
1.000 setlinewidth
grestore
0 .54439 m
.00625 .54439 L
s
gsave
-0.0125 .54439 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .10333 m
.00375 .10333 L
s
0 .12654 m
.00375 .12654 L
s
0 .14976 m
.00375 .14976 L
s
0 .17297 m
.00375 .17297 L
s
0 .2194 m
.00375 .2194 L
s
0 .24261 m
.00375 .24261 L
s
0 .26583 m
.00375 .26583 L
s
0 .28904 m
.00375 .28904 L
s
0 .33547 m
.00375 .33547 L
s
0 .35868 m
.00375 .35868 L
s
0 .3819 m
.00375 .3819 L
s
0 .40511 m
.00375 .40511 L
s
0 .45154 m
.00375 .45154 L
s
0 .47475 m
.00375 .47475 L
s
0 .49797 m
.00375 .49797 L
s
0 .52118 m
.00375 .52118 L
s
0 .0569 m
.00375 .0569 L
s
0 .03369 m
.00375 .03369 L
s
0 .01047 m
.00375 .01047 L
s
0 .56761 m
.00375 .56761 L
s
0 .59082 m
.00375 .59082 L
s
0 .61403 m
.00375 .61403 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95.9375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .08012 m
1 .08012 L
s
.99375 .19619 m
1 .19619 L
s
.99375 .31225 m
1 .31225 L
s
.99375 .42832 m
1 .42832 L
s
.99375 .54439 m
1 .54439 L
s
.125 Mabswid
.99625 .10333 m
1 .10333 L
s
.99625 .12654 m
1 .12654 L
s
.99625 .14976 m
1 .14976 L
s
.99625 .17297 m
1 .17297 L
s
.99625 .2194 m
1 .2194 L
s
.99625 .24261 m
1 .24261 L
s
.99625 .26583 m
1 .26583 L
s
.99625 .28904 m
1 .28904 L
s
.99625 .33547 m
1 .33547 L
s
.99625 .35868 m
1 .35868 L
s
.99625 .3819 m
1 .3819 L
s
.99625 .40511 m
1 .40511 L
s
.99625 .45154 m
1 .45154 L
s
.99625 .47475 m
1 .47475 L
s
.99625 .49797 m
1 .49797 L
s
.99625 .52118 m
1 .52118 L
s
.99625 .0569 m
1 .0569 L
s
.99625 .03369 m
1 .03369 L
s
.99625 .01047 m
1 .01047 L
s
.99625 .56761 m
1 .56761 L
s
.99625 .59082 m
1 .59082 L
s
.99625 .61403 m
1 .61403 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .08012 m
1 .08012 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.002 w
.07143 .53096 m
.07143 .57738 L
s
.015 w
.07143 .55417 Mdot
.002 w
.11905 .42021 m
.11905 .46664 L
s
.015 w
.11905 .44342 Mdot
.002 w
.16667 .41326 m
.16667 .45969 L
s
.015 w
.16667 .43648 Mdot
.002 w
.21429 .37612 m
.21429 .42255 L
s
.015 w
.21429 .39934 Mdot
.002 w
.2619 .31289 m
.2619 .35932 L
s
.015 w
.2619 .3361 Mdot
.002 w
.30952 .27895 m
.30952 .32537 L
s
.015 w
.30952 .30216 Mdot
.002 w
.35714 .16974 m
.35714 .21617 L
s
.015 w
.35714 .19295 Mdot
.002 w
.40476 .17257 m
.40476 .219 L
s
.015 w
.40476 .19579 Mdot
.002 w
.45238 .13882 m
.45238 .18525 L
s
.015 w
.45238 .16203 Mdot
.002 w
.5 .11343 m
.5 .15986 L
s
.015 w
.5 .13664 Mdot
.002 w
.54762 .09332 m
.54762 .13975 L
s
.015 w
.54762 .11654 Mdot
.002 w
.59524 .05457 m
.59524 .101 L
s
.015 w
.59524 .07779 Mdot
.002 w
.64286 .04598 m
.64286 .0924 L
s
.015 w
.64286 .06919 Mdot
.002 w
.69048 .11699 m
.69048 .16342 L
s
.015 w
.69048 .14021 Mdot
.002 w
.7381 .06168 m
.7381 .10811 L
s
.015 w
.7381 .08489 Mdot
.002 w
.78571 .06901 m
.78571 .11544 L
s
.015 w
.78571 .09222 Mdot
.002 w
.83333 .06717 m
.83333 .1136 L
s
.015 w
.83333 .09038 Mdot
.002 w
.88095 .05179 m
.88095 .09822 L
s
.015 w
.88095 .075 Mdot
.002 w
.92857 .01472 m
.92857 .06114 L
s
.015 w
.92857 .03793 Mdot
.5 Mabswid
.02381 .60332 m
.06244 .55102 L
.10458 .49705 L
.14415 .44928 L
.18221 .40599 L
.22272 .36278 L
.26171 .32397 L
.30316 .28573 L
.34309 .25181 L
.3815 .22188 L
.42237 .19296 L
.46172 .16796 L
.49955 .14654 L
.53984 .12658 L
.57861 .11012 L
.61984 .09559 L
.65954 .08448 L
.69774 .07648 L
.71741 .07339 L
.73838 .07085 L
.74867 .0699 L
.75979 .06908 L
.7649 .06878 L
.7703 .06851 L
.77535 .06831 L
.77996 .06817 L
.7823 .06811 L
.78483 .06806 L
.78723 .06802 L
.78944 .06799 L
.79065 .06798 L
.79178 .06797 L
.79303 .06797 L
.79373 .06797 L
.79437 .06797 L
.79559 .06797 L
.79628 .06797 L
.79692 .06797 L
.79822 .06797 L
.79894 .06798 L
.7996 .06798 L
.80196 .06801 L
.80413 .06804 L
.80651 .06808 L
.80903 .06814 L
.81327 .06827 L
.81794 .06844 L
.8281 .06895 L
.83738 .06958 L
.85828 .07157 L
Mistroke
.8773 .07406 L
.89804 .07752 L
.93873 .08656 L
.97619 .09751 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00lP3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo]P3oool00?@0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;D0oooo003e0?ooo`040;mdM02LW4P004RL0=oookL0
oooo003e0?ooo`030?oo_`1d000007Bo0;P0oooo003e0?ooo`030?oo_`1d0000M;oo0;P0oooo003e
0?ooo`040?noM000000004RL0=oookL0oooo003e0?ooo`050;md0000B9`0gmnL04Q8W03Oool0]P3o
ool00?<0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]03oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005<0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`160?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
od40oooo00@0_g@0000000000000M;ooB03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom50?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool04@3oool005D0oooo00<0okmd0000001d_ol0B03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_d00oooo00H0gia8001d_`3oool0omnL04P00018W=m60?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
A`3oool00`3o_g@0000007Boo`0D0?ooo`00E@3oool00`3ooml0W4P0001d_`170?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0m0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0A03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0A@3oool00`3ooml0W4P0001d_`0D0?ooo`00EP3oool0
0`3OW4P004RL0=ooo`160?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0m0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
@`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0AP3oool00`3OW4P004RL0=ooo`0C0?ooo`00AP3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0>@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03d0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`130?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`160?ooo`030?oOW01804P0W=oo
01<0oooo001F0?ooo`030?oo_`1d0000M;oo04L0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03h0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`140?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`160?ooo`030?oo_`1d0000M;oo01<0oooo001G0?ooo`030;md0000M;l0oooo04H0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03h0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`140?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`170?ooo`030;md0000M;l0oooo0180oooo
001G0?ooo`030=nLB000B9`0gooo04H0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo?P3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo04<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_oooA`3oool00`3OW4P004RL0=ooo`0B0?ooo`00E`3oool00`3o
gi`0B01809cOo`160?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`130?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04P0oooo00<0omnL04P0B02Lgol04P3oool005L0oooo00<0oono07@0001d
_ol0A`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0@@3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo04@0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04T0oooo00<0oono07@0001d
_ol04P3oool005P0oooo00<0_g@0001d_`3oool0A`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0@P3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgdH0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04X0oooo00<0_g@0001d_`3o
ool04@3oool005H0oooo00D0oono07@000000000001d0;ooo`180?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`100?ooo`040=nLB00000000000001d_dP0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO04D0oooo00D0oono07@00000
0000001d0;ooo`0A0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00B@3ooooo000005d000002`3oool004T0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0500oooo
00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo0500oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002[
0?ooo`030000003oool0oooo0:X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo0980oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?oo
o`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo0980oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo0980oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`00
0000oooo0?ooo`2Z0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0940
oooo0`00000G0?ooo`030000003oool0oooo00T0oooo00190?ooo`800000Z`3oool00`000000oooo
0?ooo`2@0?ooo`D000005@3oool2000000/0oooo00190?ooo`030000003oool0oooo0:X0oooo00<0
00000?ooo`3oool0T03oool5000001H0oooo00<000000?ooo`3oool02@3oool004T0oooo00<00000
0?ooo`3oool0ZP3oool00`000000oooo0?ooo`2@0?ooo`D000005P3oool00`000000oooo0?ooo`09
0?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0940oooo0`00000G0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3o
ool0;`3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0_0?ooo`030000
003oool0oooo0600oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0D03o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool02@3o
ool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002[0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool05@3oool2000000/0oooo00190?ooo`030000003oool0oooo0:X0
oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo01h0
oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01H0
oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo00d0oooo0`00000O0?ooo`030000003oool0oooo02l0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`090?ooo`009`3oool00`3ookl0M0000000
00020000000604RLg`3oool0oooo0?ooo`3ookl0M0000`0000000`18W=l0oooo0?ooo`030?ooo`04
0;md00000000000007Boo`X0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0N0?oo
o`030000003oool0oooo00`0oooo1@00000N0?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`03
0?noM0000000M;oo00H0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B000
04RLg`T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0
oooo00`0oooo1@00000N0?ooo`030000003oool0oooo02@000002P3oool3000002L0oooo00<00000
0?ooo`3oool02@3oool002T0oooo00<0oooO09a80000M;l01P3oool00`3ooml0W4P0001d_`040?oo
o`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo01d0oooo0`00000=0?ooo`D000005P3oool;000000h0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`030?ooo`H000000`3oool5000002H0oooo00<000000?oo
o`3oool02@3oool002X0oooo00<0gia80018W03Oool01P3oool00`3OW4P004RL0=ooo`030?ooo`03
0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo01`0oooo1@00000=0?ooo`<000004`3oool4000000P0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00T0oooo2000000V0?oo
o`030000003oool0oooo00T0oooo000Z0?ooo`030?oOW01804P0W=oo00H0oooo00<0omnL04P0B02L
gol00`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3o
ool0ZP3oool00`000000oooo0?ooo`0L0?ooo`D000003P3oool00`000000oooo0?ooo`0=0?ooo`D0
0000303oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool0303oool;00000200oooo00<000000?ooo`3oool02@3oool002X0oooo00<0oono07@0001d_ol0
1P3oool00`3ookl0M00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
1`3ooooo000005d000002`3oool002/0oooo00<0_g@0001d_`3oool01P3oool00`2oM00007Bo0?oo
o`030?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01`3oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo01d0oooo0`00000?0?ooo`030000003oool0oooo0080oooo1@00
000E0?ooo`D000003P3oool00`000000oooo0?ooo`0<0?ooo`<000003`3oool00`000000oooo0?oo
o`0:0?ooo`D000005`3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3OW4P004RL0=ooo`06
0?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`06
0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0>
0?ooo`D000006P3oool5000000d0oooo0`00000<0?ooo`D000003P3oool00`000000oooo0?ooo`0?
0?ooo`@000004`3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ogi`0B01809cOo`060?oo
o`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`06
0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0<
0?ooo`<000007P3oool5000000`0oooo1@00000;0?ooo`D000003P3oool00`000000oooo0?ooo`0C
0?ooo`<00000403oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ookl0M00007Boo`060?oo
o`030?oo_`1d0000M;oo00@0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`060?oo
o`030000003oool0oooo0:X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00P0oooo100000020?ooo`030000003oool0oooo01d0oooo0`00000=0?oo
o`D000002`3oool5000000h0oooo00<000000?ooo`3oool05P3oool4000000`0oooo00<000000?oo
o`3oool02@3oool002`0oooo00<0_g@0001d_`3oool01P3oool00`2oM00007Bo0?ooo`040?ooo`06
0=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`050?ooo`<000009`3oool0
0`000000oooo0?ooo`0<0?ooo`D00000303oool3000000l0oooo00<000000?ooo`3oool06P3oool3
000000T0oooo00<000000?ooo`3oool02@3oool002X0oooo00D0oono07@000000000001d0;ooo`04
0?ooo`050?oo_`1d000000000000M02oool01@3oool0103OW4P0000000000000M;l80?ooo`030000
003oool0oooo0:X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo0080oooo0`00000Z0?ooo`030000003oool0oooo00d0oooo0`00000>0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool07@3oool2000000L0oooo00<000000?ooo`3oool0
2@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool00P00000]0?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`800000Z`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo0`00000_
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0f
0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0>0?oo
o`030000003oool0oooo00X0oooo0`00000b0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo0:X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`070?ooo`<0
0000=@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool0=`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo00d0oooo0`0000060?ooo`800000B@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo03L0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`0<0?ooo`D000000`3oool2000004/0oooo00<000000?ooo`3oool0A`3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo00`0oooo2000000/0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool02@3oool0
04T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0<0?ooo`D00000;`3oool00`00
0000oooo0?ooo`1X0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0
oooo00<000000?ooo`3oool03@3oool300000300oooo00<000000?ooo`3oool0J03oool00`000000
oooo0?ooo`090?ooo`00B@3oool200000:/0oooo00<000000?ooo`3oool02`3oool2000000030?oo
o`000000oooo0300oooo00<000000?ooo`3oool0I`3oool2000000/0oooo00190?ooo`030000003o
ool0oooo0:X0oooo00<000000?ooo`3oool02@3oool2000000<0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0
oooo0:T0oooo0`0000080?ooo`8000001@3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0
oooo06P0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0Z03oool50000
00D0oooo0P0000070?ooo`030000003oool0oooo02d0oooo0`00001Z0?ooo`030000003oool0oooo
00T0oooo00190?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool02`3oool5000000<0
oooo0P0000090?ooo`030000003oool0oooo02`0oooo1@00001Y0?ooo`030000003oool0oooo00T0
oooo00190?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool02`3oool5000000030?oo
o`000000000000/0oooo00<000000?ooo`3oool0;03oool5000006T0oooo00<000000?ooo`3oool0
2@3oool004T0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`0<0?ooo`D00000?03o
ool5000006T0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0VP3oool0
0`000000oooo0?ooo`0=0?ooo`800000?`3oool3000006X0oooo00<000000?ooo`3oool02@3oool0
04T0oooo0P00002K0?ooo`030000003oool0oooo00/0oooo0`0000120?ooo`030000003oool0oooo
06L0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00T0
oooo0P0000020?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0J03oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2I0?ooo`<00000203oool2000000@0oooo
00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`1X0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo09P0oooo1@0000050?ooo`8000001P3oool00`000000oooo0?oo
o`100?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool02@3oool004T0oooo00<00000
0?ooo`3oool0V03oool5000000<0oooo0P0000080?ooo`030000003oool0oooo0400oooo00<00000
0?ooo`3oool0J03oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2H0?oo
o`D0000000<0oooo0000000000002P3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo
06P0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0V@3oool300000003
0?ooo`000000oooo00/0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00
B@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo01h0oooo0`00000=0?ooo`030000
003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00001j0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo0:X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool0303oool3000000l0oooo00<000000?ooo`3oool0Z`3oool0
0`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool02`3oool010000000oooo0?ooo`00000?0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool02@3oool002L0oooo00<0oono07@0000000000P0000001@18
W=l0oooo0?ooo`3oool0gia800D0000000<0B9cO0?ooo`3oool00P3oool0102oM00000000000001d
_ol:0?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`090?ooo`8000000`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0:/0oooo00<0
00000?ooo`3oool02@3oool002T0oooo00<0okmd0000001d_ol01@3oool00`3o_g@007Bo0?ooo`02
0?ooo`0:0;mdM02oool0oooo0?ooo`3OW4P007Bo0?ooo`3ogi`0B00004RLg`T0oooo00<000000?oo
o`3oool0N@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00L0oooo0P0000050?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?oo
o`00:@3oool00`3ooml0W4P0001d_`060?ooo`030?noM000M;l0oooo00@0oooo00<0oooO09a80018
W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo07P0oooo0`00000>0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo000Z0?ooo`030=nLB000B9`0gooo00H0oooo00<0okmd001d_`3oool0
0`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3o
ool0M`3oool5000000`0oooo0`0000050?ooo`800000603oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo000Z0?ooo`030?oOW01804P0W=oo00L0oooo00<0okmd04RLg`3oool0
0P3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool0
M`3oool5000000/0oooo1@0000030?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
Z`3oool00`000000oooo0?ooo`090?ooo`00:P3oool00`3ookl0M00007Boo`080?ooo`030=nLB018
W=l0oooo0080oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`070?ooo`<00000M`3o
ool5000000/0oooo1@0000000`3oool000000000000K0?ooo`030000003oool0oooo0:T0oooo0`00
000;0?ooo`00:`3oool00`2oM00007Bo0?ooo`080?ooo`070;md001d_ol0oooo0?ooo`3OW4P00018
09cOo`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0N03oool3000000`0oooo
1P00000M0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002/0oooo00<0
gia80018W03Oool0203oool01`3ogi`0B01d0;ooo`3oool0omnL04P00018W=l00`3oool00`2oM000
04RL0=ooo`060?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0303oool4000001h0
oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ogi`0B018
09cOo`080?ooo`080?oog`2LB000M;oo0?ooo`3ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0
gooo00H0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0=0?ooo`8000007`3oool0
0`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000[0?ooo`030?oo_`1d0000M;oo
00P0oooo00<0oono07@00018W=l00P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo
00H0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`0<0?ooo`800000803oool00`00
0000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000/0?ooo`030;md0000M;l0oooo00<0
oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00<0oooo00H0gia80000M02oool0oooO09a8
0018W=l70?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool02P3oool2000000030?oo
o`000000oooo01l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:P3o
ool01@3ookl0M0000000000007@0_ooo00@0oooo00D0omnL04P000000000000007Boo`050?ooo`04
0=nLB00000000000001d_`P0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`090?oo
o`050000003oool0oooo0?ooo`000000803oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`090?ooo`00B@3oool2000008<0oooo0P0000050?ooo`030000003oool0oooo01h0
oooo00<000000?ooo`3oool0ZP3oool2000000/0oooo00190?ooo`030000003oool0oooo0840oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo0800oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo07h0oooo0P00000Z
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0O@3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`0[0?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0
NP3oool2000002h0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`1i0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool2000007T0oooo00<000000?ooo`3oool0;`3oool0
0`000000oooo0?ooo`2Z0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0MP3oool20000
0380oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`1e0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo03<0oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`1c0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1`0?oo
o`800000>03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo06l0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`800000K`3oool00`000000oooo0?ooo`0i0?ooo`030000
003oool0oooo0:X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`1]0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`1/0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0Z`3oool0
0`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo
03`0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`1Y0?ooo`800000?`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0
oooo00190?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo
0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo06L0oooo0P00
00110?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<00000
0?ooo`3oool0IP3oool00`000000oooo000000110?ooo`030000003oool0oooo0:/0oooo00<00000
0?ooo`3oool02@3oool004T0oooo0P00001V0?ooo`040000003oool0oooo00000440oooo00<00000
0?ooo`3oool0ZP3oool2000000/0oooo00190?ooo`030000003oool0oooo06@0oooo00D000000?oo
o`3oool0oooo000000110?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool0
04T0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
03l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`007P3oool00`2oM000
04RL0=ooo`060?ooo`030?oo_`1d000000000080000000H0B9cO0?ooo`3oool0oooo0?oo_`1d0003
0000000307Boo`3oool0oooo00<0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo0?oo
o`1R0?ooo`030000003oool0oooo0080oooo0`0000100?ooo`030000003oool0oooo0:/0oooo00<0
00000?ooo`3oool02@3oool001h0oooo00<0omnL04Qd_`3oool0203oool00`3o_g@0000007Boo`04
0?ooo`030?noM000000007Bo0080oooo00<0_gA809cOo`3oool00P3oool01P3OW4P007Bo0?ooo`3o
gi`0B00004RLg`T0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`020?ooo`D00000
?`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000C0?ooo`030?oo_`1d
M;l0oooo00L0oooo00<0okmd0018W03Oool02@3oool00`3ooml0W4P0001d_`090?ooo`070?oOW018
B9`0gooo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3o
ool0H03oool00`000000oooo0?ooo`030?ooo`D00000?`3oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo000C0?ooo`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3o
ool02`3oool00`3OW4P004RL0=ooo`080?ooo`070?oo_`1d0000M;oo0?ooo`3ooml0W4P004RLg`02
0?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool0GP3oool2000000H0oooo1@00
000o0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool001@0oooo00<0omnL
04P0000000001@0000000`18W=l0oooo0?ooo`0;0?ooo`030?oOW01804P0W=oo00P0oooo00L0omnL
04P0001d_ol0oooo0?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo
0?ooo`1M0?ooo`030000003oool0oooo00L0oooo0`0000100?ooo`030000003oool0oooo0:/0oooo
00<000000?ooo`3oool02@3oool001T0oooo00<0okmd0000B02Lgol03P3oool00`3ookl0M00007Bo
o`070?ooo`040?oOW0180000001809cOo`80oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B000
04RLg`070?ooo`<00000G03oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo03l0oooo
00<000000?ooo`3oool0Z@3oool3000000/0oooo000H0?ooo`030;md0000M;l0oooo0100oooo00<0
_g@0001d_`3oool0103oool01@3ogi`0B000000000000000M;oo00<0oooo00<0gia80000B02Lgol0
0P3oool00`3ooml0W4P0001d_`070?ooo`030000003oool0oooo05P0oooo00@000000?ooo`3oool0
0000303oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool001<0oooo00H0omnL04P0001d_ol0okmd0000001d_olB0?ooo`030=nLB000B9`0gooo
00H0oooo00<0oono07@00000M;l00`3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo
00H0oooo00<000000?ooo`3oool0F03oool00`000000oooo0000000=0?ooo`030000003oool0oooo
03l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`004`3oool0103ogi`0
B00000000000M;lD0?ooo`030?oOW01804P0W=oo00P0oooo00P0gia8001d_`3oool0oooo0?oog`2L
B00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`000000oooo0?ooo`1H0?ooo`800000
3P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool002/0oooo00<0oono07@0001d_ol0203oool00`3ogi`0B01809cOo`020?ooo`030?oOW018
0000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`1H0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`2[0?ooo`030000003o
ool0oooo00T0oooo000/0?ooo`030;md0000M;l0oooo00@0oooo00H0_g@007Boo`3oool0gia80000
B02Lgol30?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`1G0?oo
o`8000003`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool002X0oooo00D0oono07@000000000001d0;ooo`050?ooo`040;md000000000000
04RLg`D0oooo00@0gia800000000000007Bo203oool00`000000oooo0?ooo`1F0?ooo`030000003o
ool000000540oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`1E0?ooo`030000003oool0000000800000D03oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00T0oooo00190?ooo`800000E@3oool00`000000oooo000000040000
04l0oooo00<000000?ooo`3oool0ZP3oool2000000/0oooo00190?ooo`030000003oool0oooo05<0
oooo00<000000?ooo`3oool01@00001?0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool004T0oooo00<000000?ooo`3oool0DP3oool010000000oooo0?ooo`3oool5000004l0
oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo
0?ooo`1A0?ooo`030000003oool0oooo00<0oooo0`00001@0?ooo`030000003oool0oooo0:/0oooo
00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0C`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`030000003oool0oooo04h0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`1?0?oo
o`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3o
ool0C@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2000004d0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo0:X0oooo0P00000;0?ooo`00B@3o
ool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0C`3o
ool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0
oooo04X0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0GP3oool00`00
0000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo04P0
oooo00<000000?ooo`3oool0G`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0
oooo00190?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0H03oool00`000000oooo
0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo04H0oooo00<0
00000?ooo`3oool0H@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`000000HP3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo04@0oooo00@000000?oo
o`3oool00000HP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`800000A03oool01@000000oooo0?ooo`3oool000000680oooo00<000000?ooo`3oool0ZP3oool2
000000/0oooo00190?ooo`030000003oool0oooo04<0oooo00D000000?ooo`3oool0oooo0000001R
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0@P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0600oooo00<000000?oo
o`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`110?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0H03oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0
0`3oool300000640oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00<0oooo1@00001P0?ooo`030000003o
ool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0?P3oool0
0`000000oooo0?ooo`040?ooo`D00000H03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool01@3oool50000
0600oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2000003d0
oooo00<000000?ooo`3oool01`3oool300000640oooo00<000000?ooo`3oool0ZP3oool2000000/0
oooo00190?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`1P0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<0
00000?ooo`3oool0=`3oool010000000oooo0?ooo`00000<0?ooo`030000003oool0oooo0600oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`009`3oool00`3ookl0M0000000
00020000000304RLg`3oool0oooo00@0oooo00<0oooO09a80018W=l01@3oool0102oM00000000000
001d_ol:0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`0000003@3oool00`000000oooo
0?ooo`1P0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002T0oooo00<0
okmd0000001d_ol0203oool00`3OW4P004RL0=ooo`030?ooo`060=nLB000M;l0oooo0?oOW0180000
B9cO2@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0000000d0oooo00<000000?ooo`3o
ool0H03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`030?oo
g`2LB00007Bo00P0oooo00<0omnL04P0B02Lgol00P3oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00P0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0>0?ooo`800000
3P3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool002X0oooo00<0gia80018W03Oool00`3oool00`3o_g@0000000000004000000050018W03O
ool0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo02H0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0600
oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:P3oool00`3ogi`0B018
09cOo`030?ooo`030?oog`2LB4P0W=oo0080oooo00L0_g@0001d_`3oool0oooo0?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo
00d0oooo0P00001b0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0
oooo00<0oono07@0001d_ol0103oool01`3ookl0M01d0;ooo`3oool0okmd0018W03Oool00P3oool0
0`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo0`00000V0?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`000000LP3oool00`000000oooo0?ooo`2Y0?ooo`<000002`3oool0
02/0oooo00<0_g@0001d_`3oool0103oool01P3ogi`0B4RL0=ooo`3ogi`0B01809cOo`80oooo00<0
gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`070?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0000000200000740oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00:`3oool00`3OW4P004RL0=ooo`050?ooo`050?noM000M;l0
oooO09a80018W=l00P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0000000@00000L03o
ool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000[0?ooo`030?oOW01804P0
W=oo00H0oooo00@0gia804RLg`2oM00007Bo0P3oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo00H0oooo00<000000?ooo`3oool09@3oool3000000X0oooo00<000000?ooo`3oool0
1@00001`0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002/0oooo00<0
oono07@0001d_ol01`3oool0103OW4P0M7Ad0000M02oool20?ooo`030?oOW0180000M;oo0080oooo
00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`0T0?ooo`D000002@3oool00`000000oooo
0?ooo`0500000700oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00;03o
ool00`2oM00007Bo0?ooo`070?ooo`030;md000004P0W=oo00<0oooo00H0gia80000M02oool0oooO
09a80018W=l70?ooo`030000003oool0oooo02@0oooo1@0000080?ooo`030000003oool0oooo0080
oooo0`00001a0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo
00D0oono07@000000000001d0;ooo`070?ooo`030?oog`2LB000B9cO00@0oooo00@0gia800000000
000007Bo203oool00`000000oooo0?ooo`0T0?ooo`D000001`3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0700oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?oo
o`00B@3oool00`000000oooo0?ooo`0U0?ooo`<000001`3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0700oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00
B@3oool2000002L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool0ZP3oool2000000/0oooo00190?ooo`030000003o
ool0oooo02H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo0700oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`1`0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool0
04T0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0L03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo
00T0oooo00190?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`1l0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo
00<000000?ooo`3oool09P3oool01@000000oooo0?ooo`3oool0000007l0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0Y0?ooo`030000
003oool0oooo07h0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo07l0oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool2000002T0oooo00<000000?ooo`3oool0O`3oool0
0`000000oooo0?ooo`2Z0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool09`3oool00`00
0000oooo0?ooo`200?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0:/0
oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool09@3oool00`000000oooo
0?ooo`220?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<0
00000?ooo`3oool0903oool00`000000oooo0?ooo`230?ooo`030000003oool0oooo0:/0oooo00<0
00000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`24
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool08`3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`250?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00000R0?ooo`030000
003oool0oooo08H0oooo00<000000?ooo`3oool0ZP3oool2000000/0oooo00190?ooo`030000003o
ool0oooo0200oooo00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`2[0?ooo`030000003o
ool0oooo00T0oooo00190?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0R03oool0
0`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo
01l0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo
00T0oooo00190?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0R@3oool00`000000
oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo01d0oooo
00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0R`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`800000703oool00`000000oooo0?ooo`2<0?ooo`030000003oool0oooo0:X0oooo0P00000;0?oo
o`00B@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo08h0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`009`3o
ool00`3ookl0M000000000020000000604RLg`3oool0oooo0?ooo`3ogi`0B0000P0000000`00B9`0
gooo0?ooo`040?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool05P3oool00`00
0000oooo0000002A0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002T0
oooo00<0okmd0000001d_ol0103oool01P3o_g@0000007Boo`3oool0omnL04Qd_`@0oooo00H0gia8
001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo01H0oooo00<000000?ooo`000000
T@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`030?oog`2L
B00007Bo00T0oooo00L0okmd04RLg`3oool0oooo0?oog`2LB000B9cO0080oooo00<0okmd0000B02L
gol0203oool00`000000oooo0?ooo`0F0?ooo`800000TP3oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo000Z0?ooo`030=nLB000B9`0gooo00P0oooo00L0oono07@0M02oool0
oooo0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`09
0?ooo`00:P3oool00`3ogi`0B01809cOo`080?ooo`070?oog`2LB000M;oo0?ooo`3ooml0W4P0001d
_`030?ooo`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool05@3oool2000009<0oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:P3oool00`3ookl0M00007Boo`08
0?ooo`030?oo_`1d0000B9cO0080oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`07
0?ooo`<000005@3oool2000009<0oooo00<000000?ooo`3oool0Z@3oool3000000/0oooo000[0?oo
o`030;md0000M;l0oooo00L0oooo00<0okmd0000001d_ol00P3oool00`3OW4P0001809cOo`020?oo
o`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0503oool400000980oooo00<000000?oo
o`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3OW4P004RL0=ooo`040?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo0P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0
gooo00H0oooo00<000000?ooo`3oool04`3oool600000940oooo00<000000?ooo`3oool0Z`3oool0
0`000000oooo0?ooo`090?ooo`00:`3oool00`3ogi`0B01809cOo`040?ooo`050?oo_`1d00000000
0000B02Lgol00`3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0000000400000940oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ookl0M00007Boo`050?ooo`030=nLW02oool0
oooo00D0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool01@00002A0?ooo`030000003oool0oooo0:/0oooo00<00000
0?ooo`3oool02@3oool002`0oooo00<0_g@0001d_`3oool0103oool00`3ogi`0M9cO0?ooo`060?oo
o`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`0A0?ooo`040000003o
ool0oooo0?ooo`<00000TP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo
000Z0?ooo`050?oo_`1d000000000000M02oool01@3oool01@2oM0000000000000000000B9cO00@0
oooo00@0gia800000000000007Bo203oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo
00T0oooo00190?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`2A0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo
0P00000?0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0T@3oool00`000000oooo
0?ooo`2Z0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo0940oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool0T@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`2A0?oo
o`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool004T0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`2M0?ooo`030000
003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`2M0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3o
ool004T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`2N0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo0P0000090?ooo`030000003oool0oooo
09l0oooo00<000000?ooo`3oool0ZP3oool2000000/0oooo00190?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0
oooo00190?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0X03oool00`000000oooo
0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?oo
o`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0
D03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
2@3oool004T0oooo0P0000060?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`040?ooo`8000002`3oool004T0ooooo`00001M000000/0oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.49425, 90.3368, \
0.00907712, 0.372403}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        ListPlot[Evaluate[Transpose[{x, \(y - y\&_\)\/dy}]], 
          PlotRange \[Rule] {Automatic, {\(-3\), 3}}, 
          PlotStyle \[Rule] PointSize[0.02], Frame \[Rule] True, 
          FrameLabel \[Rule] {"\<x\>", "\<residuals\>"}], 
        Plot[{\(-1\), 1}, {x, \(-1\), 1}, 
          PlotStyle \[Rule] Dashing[{0.02, 0.02}]]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.103006 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .10301 -18.6875 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -18.6875 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -10 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[-0.0125 .61803 -10 -7.84375 ]
[-0.0125 .61803 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.375 -23.9688 ]
[-0.01875 .30902 -18.6875 23.9688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .61803 m
.00625 .61803 L
s
gsave
-0.0125 .61803 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .0206 m
.00375 .0206 L
s
0 .0412 m
.00375 .0412 L
s
0 .0618 m
.00375 .0618 L
s
0 .0824 m
.00375 .0824 L
s
0 .12361 m
.00375 .12361 L
s
0 .14421 m
.00375 .14421 L
s
0 .16481 m
.00375 .16481 L
s
0 .18541 m
.00375 .18541 L
s
0 .22661 m
.00375 .22661 L
s
0 .24721 m
.00375 .24721 L
s
0 .26781 m
.00375 .26781 L
s
0 .28842 m
.00375 .28842 L
s
0 .32962 m
.00375 .32962 L
s
0 .35022 m
.00375 .35022 L
s
0 .37082 m
.00375 .37082 L
s
0 .39142 m
.00375 .39142 L
s
0 .43262 m
.00375 .43262 L
s
0 .45322 m
.00375 .45322 L
s
0 .47383 m
.00375 .47383 L
s
0 .49443 m
.00375 .49443 L
s
0 .53563 m
.00375 .53563 L
s
0 .55623 m
.00375 .55623 L
s
0 .57683 m
.00375 .57683 L
s
0 .59743 m
.00375 .59743 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -111.5 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
84.9688 11.8438 translate
90 rotate
-84.9688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(residuals) show
106.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .0206 m
1 .0206 L
s
.99625 .0412 m
1 .0412 L
s
.99625 .0618 m
1 .0618 L
s
.99625 .0824 m
1 .0824 L
s
.99625 .12361 m
1 .12361 L
s
.99625 .14421 m
1 .14421 L
s
.99625 .16481 m
1 .16481 L
s
.99625 .18541 m
1 .18541 L
s
.99625 .22661 m
1 .22661 L
s
.99625 .24721 m
1 .24721 L
s
.99625 .26781 m
1 .26781 L
s
.99625 .28842 m
1 .28842 L
s
.99625 .32962 m
1 .32962 L
s
.99625 .35022 m
1 .35022 L
s
.99625 .37082 m
1 .37082 L
s
.99625 .39142 m
1 .39142 L
s
.99625 .43262 m
1 .43262 L
s
.99625 .45322 m
1 .45322 L
s
.99625 .47383 m
1 .47383 L
s
.99625 .49443 m
1 .49443 L
s
.99625 .53563 m
1 .53563 L
s
.99625 .55623 m
1 .55623 L
s
.99625 .57683 m
1 .57683 L
s
.99625 .59743 m
1 .59743 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .30902 m
1 .30902 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.07143 .37523 Mdot
.11905 .15001 Mdot
.16667 .36725 Mdot
.21429 .43242 Mdot
.2619 .36367 Mdot
.30952 .40679 Mdot
.35714 .09783 Mdot
.40476 .26791 Mdot
.45238 .25752 Mdot
.5 .26615 Mdot
.54762 .28009 Mdot
.59524 .19319 Mdot
.64286 .22198 Mdot
.69048 .58592 Mdot
.7381 .37119 Mdot
.78571 .41632 Mdot
.83333 .40263 Mdot
.88095 .31075 Mdot
.92857 .1045 Mdot
.5 Mabswid
[ .02 .02 ] 0 setdash
.02381 .20601 m
.06244 .20601 L
.10458 .20601 L
.14415 .20601 L
.18221 .20601 L
.22272 .20601 L
.26171 .20601 L
.30316 .20601 L
.34309 .20601 L
.3815 .20601 L
.42237 .20601 L
.46172 .20601 L
.49955 .20601 L
.53984 .20601 L
.57861 .20601 L
.61984 .20601 L
.65954 .20601 L
.69774 .20601 L
.73838 .20601 L
.77751 .20601 L
.81909 .20601 L
.85916 .20601 L
.89771 .20601 L
.93871 .20601 L
.97619 .20601 L
s
.02381 .41202 m
.06244 .41202 L
.10458 .41202 L
.14415 .41202 L
.18221 .41202 L
.22272 .41202 L
.26171 .41202 L
.30316 .41202 L
.34309 .41202 L
.3815 .41202 L
.42237 .41202 L
.46172 .41202 L
.49955 .41202 L
.53984 .41202 L
.57861 .41202 L
.61984 .41202 L
.65954 .41202 L
.69774 .41202 L
.73838 .41202 L
.77751 .41202 L
.81909 .41202 L
.85916 .41202 L
.89771 .41202 L
.93871 .41202 L
.97619 .41202 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00k`3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo^@3oool00?40oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;P0oooo003b0?ooo`040;mdM02LW4P004RL0=oookX0
oooo003b0?ooo`030?oo_`1d000007Bo0;/0oooo003b0?ooo`030?oo_`1d0000M;oo0;/0oooo003b
0?ooo`040?noM000000004RL0=oookX0oooo003b0?ooo`050;md0000B9`0gmnL04Q8W03Oool0^@3o
ool00?00oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]`3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005T0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`120?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
oc`0oooo00@0_g@0000000000000M;ooA03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom10?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool0703oool005/0oooo00<0okmd0000001d_ol0A03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_c/0oooo00H0gia8001d_`3oool0omnL04P00018W=m20?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
@`3oool00`3o_g@0000007Boo`0O0?ooo`00F`3oool00`3ooml0W4P0001d_`130?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0h0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0@@3oool00`3ooml0W4P0001d_`0O0?ooo`00G03oool0
0`3OW4P004RL0=ooo`120?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0h0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
?`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0@P3oool00`3OW4P004RL0=ooo`0N0?ooo`00C03oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0o0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`120?ooo`030?oOW01804P0W=oo
01h0oooo001L0?ooo`030?oo_`1d0000M;oo04<0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03T0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`100?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`120?ooo`030?oo_`1d0000M;oo01h0oooo001M0?ooo`030;md0000M;l0oooo0480oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03T0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`100?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`130?ooo`030;md0000M;l0oooo01d0oooo
001M0?ooo`030=nLB000B9`0gooo0480oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo>@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo03l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo@`3oool00`3OW4P004RL0=ooo`0M0?ooo`00G@3oool00`3o
gi`0B01809cOo`120?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`0o0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04@0oooo00<0omnL04P0B02Lgol07@3oool005d0oooo00<0oono07@0001d
_ol0@`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0?03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0400oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04D0oooo00<0oono07@0001d
_ol07@3oool005h0oooo00<0_g@0001d_`3oool0@`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0?@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd80oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04H0oooo00<0_g@0001d_`3o
ool0703oool005`0oooo00D0oono07@000000000001d0;ooo`140?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0k0?ooo`040=nLB00000000000001d_d@0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO0440oooo00D0oono07@00000
0000001d0;ooo`0L0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00D03ooooo000004X000005`3oool00500oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04`0oooo
00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool0
0500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0
X03oool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0E0?ooo`00@03oool00`3OW4P00000000000030000000304RLg`3oool0
oooo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo00110?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool01`3oool00`00
0000oooo0?ooo`1a0?ooo`800000;P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
01D0oooo00120?ooo`030?noM000M;l0oooo00/0oooo00<000000?ooo`3oool0L03oool4000002d0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@`3oool00`3o_g@007Bo
0?ooo`0:0?ooo`030000003oool0oooo06l0oooo1P00000/0?ooo`030000003oool0oooo08T0oooo
0P00000F0?ooo`030000003oool0oooo01D0oooo000c0?ooo`030?oog`2LB000000000L0000000<0
B9cO0?ooo`3oool0103oool00`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo06l0oooo1P00
000/0?ooo`030000003oool0oooo08P0oooo1000000E0?ooo`030000003oool0oooo01D0oooo0015
0?ooo`030=nLB018W=l0oooo00P0oooo0`00001`0?ooo`@00000;@3oool00`000000oooo0?ooo`27
0?ooo`H000004P3oool3000001L0oooo00160?ooo`030;md001d_ol0oooo00L0oooo00<000000?oo
o`3oool0L@3oool2000002h0oooo00<000000?ooo`3oool0Q`3oool6000001@0oooo00<000000?oo
o`3oool05@3oool004H0oooo00<0omnL04P0M02oool01`3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo08P0oooo1000000E0?ooo`030000003oool0oooo01D0oooo00160?ooo`030?oo
g`2LB000M;oo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`290?ooo`800000
5P3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RLg`070?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0omnL04P0000000000000
07Boo`080?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00000T0?ooo`800000O03oool0
0`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool08P3oool40000
07/0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000
oooo0?ooo`0Q0?ooo`H00000NP3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0
oooo001@0?ooo`030000003oool0oooo0240oooo1P00001j0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool08P3oool4000007/0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`0S
0?ooo`800000O03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo001@0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`0K0?ooo`800000P`3oool2000001L0oooo00100?ooo`030?oo_`1d000000000080000000<0B9cO
0?ooo`3oool0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01X0oooo10000023
0?ooo`030000003oool0oooo01D0oooo00120?ooo`030?noM0000000M;oo00/0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0I0?ooo`H00000PP3oool00`000000oooo0?ooo`0E0?oo
o`00@P3oool00`3ooml0W4P0001d_`0;0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool06@3oool600000880oooo00<000000?ooo`3oool05@3oool004<0oooo00<0gia80018W03Oool0
2P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01X0oooo100000230?ooo`030000
003oool0oooo01D0oooo000c0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3o
ool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool06`3o
ool2000008@0oooo00<000000?ooo`3oool05@3oool004<0oooo00<0oono07@0001d_ol02P3oool3
000000@0oooo200000060?ooo`L000001P3oool7000000H0oooo1`0000060?ooo`L000001P3oool8
000000D0oooo200000050?ooo`P000001@3oool8000000H0oooo1`0000060?ooo`L000001P3oool7
000000D0oooo2@0000050?ooo`P000001@3oool8000000D0oooo200000060?ooo`L000001P3oool7
000000H0oooo1`0000060?ooo`P000001@3oool8000000D0oooo200000050?ooo`P000001@3oool8
000000T0oooo0`00000G0?ooo`00A03oool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00A03oool00`3OW4P0
04RL0=ooo`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0E0?ooo`009@3oool0103ookl0M0000000M02ooolK0?ooo`030?oOW01804P0W=oo00T0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0[0?ooo`800000M03oool00`000000
oooo0?ooo`0E0?ooo`007`3oool0103ookl0W9cO0?oo_`1d00040000000307Boo`3oool0oooo01X0
oooo00<0oono07@0001d_ol02@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo02X0
oooo1000001c0?ooo`030000003oool0oooo01D0oooo000O0?ooo`060?oOW0180000000000000000
04P0W=oo803oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0:@3oool600000780oooo00<000000?ooo`3oool05@3oool00200oooo00<0oono07@0001d
_ol0803oool01@3ookl0M0000000000007@0_ooo00P0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0Y0?ooo`H00000LP3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`2oM7@0
_ooo0?ooo`0]0?ooo`800000XP3oool00`000000oooo0?ooo`0Z0?ooo`@00000LP3oool2000001L0
oooo000O0?ooo`040?noM000000004RL0=ooobd0oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0[0?ooo`800000M03oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00280oooo00H0
gia800000000000000000000B02LgolX0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool02@3ookl0M00004RLg`3ogi`0W;oo0?oo
o`3OW4P004RL0=ooo`0W0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`00803oool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d
0;ooo`0W0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0E0?ooo`007`3oool02@3ogi`0W;oo0?ooo`3ooml0W7BL0=ooo`3oool0omnL04Qd_`0X0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`007`3oool02@3o_i`0W=oo0?ooo`2oM000M;oo0?ooo`3OW4P0001d0;ooo`0X0?ooo`800000XP3o
ool00`000000oooo0?ooo`2P0?ooo`8000005`3oool001l0oooo00@0oono07@000000000M;oo;@3o
ool00`000000oooo0?ooo`2@0?ooo`8000003`3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo001@0?ooo`030000003oool0oooo08l0oooo1000000>0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool05@3oool002@0oooo00D0oooO09a80000000004RL0=ooo`0W
0?ooo`030000003oool0oooo08h0oooo1P00000=0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00200oooo00@0_g@00000000007@0_ooo0P3oool00`3ookl0M7Bo0?ooo`0W
0?ooo`030000003oool0oooo08h0oooo1P00000<0?ooo`800000X`3oool00`000000oooo0?ooo`0E
0?ooo`007`3oool01P3ogi`0W;oo0?oog`2LB000001d0;ooo`80oooo00<0gibL0;ooo`3oool09P3o
ool00`000000oooo0?ooo`210?ooo`800000303oool4000000`0oooo1000002R0?ooo`030000003o
ool0oooo01D0oooo000O0?ooo`0:0?noW02Lgol0oooo0?oog`2LB000001809cOo`3ookl0M4RL0=oo
obL0oooo00<000000?ooo`3oool0P03oool4000000`0oooo0P00000<0?ooo`H00000X@3oool00`00
0000oooo0?ooo`0E0?ooo`007`3oool00`3ookl0M7Bo0?ooo`020?ooo`040;md00000000001d0;oo
obP0oooo0P0000200?ooo`H000006@3oool600000:00oooo0P00000G0?ooo`007`3oool0103ogi`0
B0000000001d_ol]0?ooo`030000003oool0oooo07l0oooo1P00000J0?ooo`@00000XP3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`200?ooo`@00000703oool2000000h0
oooo0P00002C0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0840oooo
0P00000N0?ooo`030000003oool0oooo00/0oooo1000002B0?ooo`030000003oool0oooo01D0oooo
000O0?ooo`030?oo_`2LW=l0oooo00<0oooo00@0okmd0000000007@0_ooo9`3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo00X0oooo1P00002A0?ooo`030000003oool0oooo01D0oooo
000O0?ooo`040?oo_`2LW=l0oono07@000@0000000<004RL0=ooo`3oool09P3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo00X0oooo1P00002A0?ooo`030000003oool0oooo01D0oooo
000O0?ooo`090?oOW018000000000000B02Lgol0oooo0?oog`2LB7@0_ooo02P0oooo0P00002R0?oo
o`030000003oool0oooo00/0oooo1000002A0?ooo`8000005`3oool001/0oooo00@0oooO09a80000
07@0_ooo<@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00`0oooo0P00002C0?oo
o`030000003oool0oooo01D0oooo000S0?ooo`050=nLB000000000000000001d_ol06P3oool0102o
M00000000000001d_ol:0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`008@3oool00`3ogi`0B00004RLg`020?ooo`030?oOW01807@0_ooo
01P0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3ogi`0B4RL0=ooo`03
0?ooo`030?oo_`1dB9`0gooo01L0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`08
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`007`3oool00`3ooml0W7BL0=ooo`040?ooo`030;md001d_ol0oooo01L0oooo00<0oooO09a8
0018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ogkl0_moo0?ooo`020?oo
o`030?oog`2LB000M;oo01T0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`070?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0N@3oool2000002H0oooo00<000000?oo
o`3oool05@3oool001l0oooo00X0okmd07Boo`3oool0oooo0?oOW01804P0W=no07@00000B9`0gooo
603oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0ooooo`00001:000001L0oooo
000O0?ooo`030?oog`2LB000000000D0000000<004RL0=ooo`3oool05`3oool00`3OW4P0001809cO
o`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`1g0?ooo`H00000903oool00`000000oooo0?ooo`0E0?ooo`006`3oool0102oW9`0gooo0?oo_`1d
00040000000507Boo`3oool0oooo0;mdM02oool06@3oool00`3ogi`0B00004RLg`030?ooo`030;md
0000B9`0gooo00H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1g0?ooo`H00000
903oool00`000000oooo0?ooo`0E0?ooo`006P3oool00`3ooml0W4P0000000020000000307Boo`3o
ool0oooo01l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0N03oool4000002D0oooo00<000000?ooo`3oool0
5@3oool001X0oooo00<0oono09bLg`3oool09@3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000
M;l0oooo00H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1i0?ooo`8000009P3o
ool00`000000oooo0?ooo`0E0?ooo`00803oool01@2oM00004RL0=ooo`3ookl0M0000080000000<0
001d0;ooo`3oool06@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000O0?ooo`030?oO
W0180000000000<0000000@0B9cO0?ooo`2oW9`0gooo6`3oool0103OW4P0000000000000M;l80?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`007`3oool0103ooml0W4P00000M02oool30?ooo`030?noM018W=l0oooo02L0oooo0P00002R0?oo
o`030000003oool0oooo0:00oooo0P00000G0?ooo`009@3oool00`3o_g@007Bo0?ooo`0X0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
8`3oool00`3ooml0_kno09bL_`0Z0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`008`3oool01P3ogkl0W7A8000000000000001d0;ooobL0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0
oooo000O0?ooo`030?oOW018000000000080000000D0001809cOo`3oool0_ibL0=ooo`0W0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
7`3oool00`3o_g@0001d0;ooo`040?ooo`030;md001d_ol0oooo02L0oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000V0?ooo`030=oOg`3o
ool0oooo02L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo000S0?ooo`050;md00000000000000000018W=l0:03oool200000:80oooo00<0
00000?ooo`3oool0X03oool2000001L0oooo000Q0?ooo`030=nLB0000000M;oo0080oooo00<0omnL
04P0M02oool09`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00200oooo00<0gia804RLg`3oool00`3oool00`3ogi`0B7Bo0?ooo`0W0?oo
o`030000003oool0oooo0580oooo0P00001=0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool001l0oooo00<0omnL07BLg`3oool00`3oool00`3o_i`0B7BL0=ooo`0X0?ooo`03
0000003oool0oooo03<0oooo0P00000L0?ooo`@00000C03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo000O0?ooo`030?noM000M;l0oooo0080oooo00D0omnL04QdM01d0000
04RL0=ooo`0W0?ooo`030000003oool0oooo0380oooo1000000J0?ooo`H00000B`3oool00`000000
oooo0?ooo`1:0?ooo`800000E@3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3o_g@00000
00000004000000030018W03Oool0oooo02H0oooo00<000000?ooo`3oool0<@3oool6000001T0oooo
1P00001;0?ooo`030000003oool0oooo04T0oooo1000001D0?ooo`030000003oool0oooo01D0oooo
000O0?ooo`050?oOW0180000000000000018W=l00P3oool00`2oM4P0W=oo0?ooo`0W0?ooo`030000
003oool0oooo01<0oooo0P00000L0?ooo`H000006P3oool4000004`0oooo00<000000?ooo`3oool0
B03oool6000005<0oooo00<000000?ooo`3oool05@3oool00500oooo0P00000C0?ooo`@00000703o
ool4000001`0oooo0P00001=0?ooo`030000003oool0oooo04P0oooo1P00001B0?ooo`8000005`3o
ool00500oooo00<000000?ooo`3oool04@3oool6000001`0oooo0P00001/0?ooo`030000003oool0
oooo04T0oooo1000001D0?ooo`030000003oool0oooo01D0oooo000S0?ooo`060?noM00000000000
0000000007@0_ooo9`3oool00`000000oooo0?ooo`0A0?ooo`H00000RP3oool00`000000oooo0?oo
o`1:0?ooo`800000E@3oool00`000000oooo0?ooo`0E0?ooo`006`3oool01P2oW9`0gooo0?ooo`3o
ool0omnL04P000<0000000@004RL0=ooo`3ookl0M7Bo:03oool00`000000oooo0?ooo`0B0?ooo`@0
0000R`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000J0?ooo`030?oo
g`2LB000000000<0000000<007Bo0?ooo`3oool00P3oool00`3o_g@0001d0;ooo`0X0?ooo`030000
003oool0oooo01<0oooo0P00002<0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
5@3oool001X0oooo00<0oono07@00018W=l0<`3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool002@0oooo00D0oooO09a80000000004RL0=oo
o`0W0?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00200oooo00@0_g@0
0000000007@0_ooo0P3oool00`3ookl0M7Bo0?ooo`0W0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool01P3ogi`0W;oo0?oog`2L
B000001d0;ooo`80oooo00<0gibL0;ooo`3oool05P3oool00`3ookl0M000000000020000000304RL
g`3oool0oooo00P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1Y0?ooo`800000
=P3oool00`000000oooo0?ooo`0E0?ooo`007`3oool02P3o_i`0W=oo0?ooo`3ooml0W4P00000B02L
gol0oono07A8W03OoolI0?ooo`030?noM0000000M;oo00/0oooo00<000000?ooo`3oool0H@3oool2
000003h0oooo00<000000?ooo`3oool0J03oool4000003D0oooo00<000000?ooo`3oool05@3oool0
01l0oooo00<0oono07Ad_`3oool00P3oool0102oM00000000000M02ooolJ0?ooo`030?oog`2LB000
07Bo00/0oooo00<000000?ooo`3oool0H03oool4000003d0oooo00<000000?ooo`3oool0I`3oool6
000003@0oooo00<000000?ooo`3oool05@3oool001l0oooo00@0omnL04P000000000M;oo803oool0
0`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo05l0oooo1P00000l0?ooo`030000003oool0
oooo06L0oooo1P00000d0?ooo`030000003oool0oooo01D0oooo00130?ooo`030?oOW01804P0W=oo
00X0oooo00<000000?ooo`3oool0G`3oool6000003`0oooo00<000000?ooo`3oool0FP3oool20000
00`0oooo1000000e0?ooo`030000003oool0oooo01D0oooo00130?ooo`030?oo_`1d0000M;oo00X0
oooo0`0000040?ooo`P000001P3oool7000000H0oooo1`0000060?ooo`L000001P3oool7000000H0
oooo200000050?ooo`P000001@3oool8000000D0oooo200000060?ooo`L000001P3oool7000000H0
oooo1`0000050?ooo`T000001@3oool8000000D0oooo200000050?ooo`P000001P3oool7000000H0
oooo1`0000060?ooo`L000001@3oool9000000D0oooo200000050?ooo`P000001@3oool8000000D0
oooo200000090?ooo`<000005`3oool004@0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo
0?ooo`1Q0?ooo`800000?P3oool00`000000oooo0?ooo`1H0?ooo`H00000@`3oool00`000000oooo
0?ooo`0E0?ooo`00A03oool00`3OW4P004RL0=ooo`090?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0F03oool6000004<0oooo00<000000?ooo`3oool05@3oool004@0oooo00<0omnL
04P0B02Lgol02@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo05T0oooo10000014
0?ooo`030000003oool0oooo01D0oooo00140?ooo`030?oo_`1d0000M;oo00T0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`1J0?ooo`800000A@3oool00`000000oooo0?ooo`0E0?oo
o`00A@3oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo0480oooo0P00001M0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0oono07@000000000
001d0;ooo`080?ooo`030000003oool0oooo0440oooo1000001L0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo0P0000110?ooo`H00000F`3oool00`000000oooo
0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0@03oool6000005/0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`11
0?ooo`@00000G03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0480oooo0P00001M0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0
oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo
0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool00400oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`07
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00@@3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00L0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00120?ooo`03
0?noM000M;l0oooo00/0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo00130?ooo`030?noM000M;l0oooo00X0oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00140?ooo`030?noM018
W=l0oooo00T0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo00150?ooo`030=nLB018W=l0oooo00P0oooo0`00002Q0?ooo`030000003oool0
oooo09l0oooo0`00000G0?ooo`00AP3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ogi`0
B01d0;ooo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0E0?ooo`00AP3oool00`3ooml0W4P007Boo`070?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RL
g`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0E0?ooo`00@P3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol01`3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0
omnL04P000000000000007Boo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0X03o
ool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R
0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0
oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0>`3oool2000006@0oooo
00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0j0?ooo`@00000H`3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo03T0oooo1P00001R0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0>@3oool600000680oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0j0?ooo`@0
0000H`3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo03/0oooo0P00001T0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3o
ool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00100?ooo`030?oo_`1d00000000
0080000000<0M;oo0?ooo`3oool0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00400oooo00<0okmd00000000M;l00P3oool00`2oM4P0
W=oo0?ooo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0E0?ooo`00A@3oool00`3ogi`0B4RL0=ooo`080?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00A@3oool00`3ookl0M00007Bo
o`080?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?oo
o`1;0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo01D0oooo00150?ooo`030?oOW0180000M;oo00P0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool0303oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`00A03oool0103ogi`0B0000000B02Lgol8
0?ooool00000BP00000G0?ooo`00@P3oool01@3ogi`0B000000000000000M;oo0?l0ooooJP3oool0
04@0oooo00<0oono07@00000M;l0o`3ooomZ0?ooo`00AP3oool00`3OW4P007Bo0?ooo`3o0?ooofP0
oooo00160?ooo`030?oOW01804P0W=oo0?l0ooooJ03oool004<0oooo00H0_g@007Boo`3oool0gia8
0000B02Lgooo0?ooofP0oooo00140?ooo`040;md00000000000004RLgol0ooooJ@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.56215, -4.51278, \
0.00959293, 0.0443476}}]
}, Closed]],

Cell["\<\
\tWhat happens if we try to fit too many parameters?  The uncertainties in \
the parameters begin to grow and some of the parameters become unstable \
without improving the fit.  Clearly, it does not pay to include too many \
parameters.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(linfit[data, {1, z, z\^2, z\^3}, z]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {115.70263221247653`, \(-18.68463040203636`\), 
          17.58510999109616`, \(-5.77025261625082`\)}, 
      Uncertainties \[Rule] {0.6898518836724719`, 2.1150099867766783`, 
          1.7171336116297728`, 3.6097366647323104`}, 
      ReducedChiSquare \[Rule] 1.5644145929960696`}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(linfit[data, {1, z, z\^2, z\^3, z\^4}, z]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {114.2404895427755`, \(-18.684630402036216`\), 
          34.04652262345922`, \(-5.770252616251042`\), \(-21.538296902155313`\
\)}, Uncertainties \[Rule] {0.8660448325027396`, 2.11500998677668`, 
          6.139713497432504`, 3.609736664732313`, 7.712696682897895`}, 
      ReducedChiSquare \[Rule] 1.119123884087042`}\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Linearizing a model", "Subsection"],

Cell["\<\
\tOften a model that does not appear to be a linear function of its \
parameters can be transformed to linear form, but one must be careful to \
transform the uncertainties in a consistent manner.  For example, consider a \
radioactive decay model\
\>", "Text"],

Cell[BoxData[
    \(TraditionalForm\`y[t]\  = \ A\ Exp[\(-t\)/\[Tau]]\)], "DisplayFormula"],

Cell[TextData[{
  "with parameters ",
  Cell[BoxData[
      \(TraditionalForm\`{A, \[Tau]}\)]],
  ", where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Tau]\)]],
  " is the lifetime we seek to measure and ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " is the number of counts in an interval around time ",
  Cell[BoxData[
      \(TraditionalForm\`t\)]],
  ".  To apply the method of linear least squares, we define"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`z\  = \ \(Log[y]\  = \ 
        A\  - \ t/\[Tau]\)\)], "DisplayFormula"],

Cell["with uncertainties", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma]\_z\  = \ \(\(\[PartialD]z\/\[PartialD]y\) \
\[Sigma]\_y\  = \ \[Sigma]\_y\/y\)\)], "DisplayFormula"],

Cell["\<\
The uncertainties for a counting experiment are generally governed by the \
Poisson distribution, such that\
\>", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma]\_y = \(\(\@y\)\ \ \ \[DoubleLongRightArrow]\ \
\ \[Sigma]\_z\  = \ \(\@y\/y = 1\/\[Sigma]\_y\)\)\)], "DisplayFormula"],

Cell[TextData[{
  "\tTo illustrate this technique, we generate a set of data using random \
number based upon the Poisson distribution.  Although ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " is based upon the Gaussian distribution and is not strictly correct for \
data drawn from the Poisson distribution, it is common to ignore this \
subtlety and to fit counting data by minimizing ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  ".  Perhaps we will return to that issue later.  Also note that we apply a \
slight offset ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " in order to avoid numerical errors due to ",
  Cell[BoxData[
      \(TraditionalForm\`Log[0]\)]],
  " or ",
  Cell[BoxData[
      \(TraditionalForm\`1/0\)]],
  ", but that the results are insensitive to the value of the offset if it is \
small enough."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(t = Range[0, 1, 0.02];\)\), "\[IndentingNewLine]", 
    \(\(p = {10\^3, 0.1};\)\), "\[IndentingNewLine]", 
    \(\(yfit[t_] := 
        p\[LeftDoubleBracket]1\[RightDoubleBracket] 
          Exp[\(-t\)/
              p\[LeftDoubleBracket]2\[RightDoubleBracket]];\)\), "\
\[IndentingNewLine]", 
    \(\(y = \(10\^\(-8\) + Random[PoissonDistribution[yfit[#]]] &\) /@ 
          t;\)\), "\[IndentingNewLine]", 
    \(dy = \@y; \ z = Log[y]; dz = Abs[dy/y];\), "\[IndentingNewLine]", 
    \(zresults = linfit[Transpose[{t, z, dz}], {1, q}, q]\)}], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {6.917840992048016`, \(-9.859793642537403`\)}, 
      Uncertainties \[Rule] {0.01805443588404917`, 0.13395180282448335`}, 
      ReducedChiSquare \[Rule] 0.39495057122217564`}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The parameters returned by ",
  StyleBox["linfit", "InlineInput",
    FontWeight->"Bold"],
  " represent ",
  Cell[BoxData[
      \(TraditionalForm\`Log[A]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\(-1\)/\[Tau]\)]],
  " because those are the coefficients of the basis functions.  The desired \
parameters and their uncertainties are obtained according to"
}], "Text"],

Cell[BoxData[{
    \(TraditionalForm\`A = 
      Exp[a\_0]\ , \ \[Sigma]\_A\^2\  = \ \
\(\((\(\[PartialD]A\/\[PartialD]a\_0\) \[Sigma]\_0)\)\^2\ \ \
\[DoubleLongRightArrow]\ \ \ \[Sigma]\_A\  = \ 
        A\ \[Sigma]\_\(\_0\)\)\), "\[IndentingNewLine]", 
    \(TraditionalForm\`\[Tau] = \(-\(1\/a\_1\)\), \ \ \[Sigma]\_\[Tau]\^2\  = \
\ \(\((\(\[PartialD]\[Tau]\/\[PartialD]a\_1\) \[Sigma]\_1)\)\^2\ \ \
\[DoubleLongRightArrow]\ \ \[Sigma]\_\[Tau]\  = \ \[Sigma]\_1\/a\_1\^2\)\)}], \
"DisplayFormula"],

Cell[TextData[{
  "where the parameters are labeled ",
  Cell[BoxData[
      \(TraditionalForm\`{a\_0, a\_1}\)]],
  " and their uncertainties ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Sigma]\_0, \[Sigma]\_1}\)]],
  ".  Thus, for this example we obtain"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\({pz, dpz} = {Parameters, Uncertainties} /. 
          zresults;\)\), "\[IndentingNewLine]", 
    \({{Exp[pz\[LeftDoubleBracket]1\[RightDoubleBracket]], \(-1\)/
          pz\[LeftDoubleBracket]2\[RightDoubleBracket]}, {Exp[
            pz\[LeftDoubleBracket]1\[RightDoubleBracket]] 
          dpz\[LeftDoubleBracket]1\[RightDoubleBracket], 
        dpz\[LeftDoubleBracket]2\[RightDoubleBracket]/
          pz\[LeftDoubleBracket]2\[RightDoubleBracket]\^2}}\)}], "Input"],

Cell[BoxData[
    \({{1010.1367452910694`, 0.10142200093172046`}, {18.23744910197972`, 
        0.0013778848080814538`}}\)], "Output"]
}, Closed]],

Cell["\<\
and recover the model parameters within their estimated uncertainties.  Note, \
however, that an unweighted fit\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(linfit[Transpose[{t, z}], {1, q}, q]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {11.211020051437226`, \(-25.592221142133436`\)}, 
      Uncertainties \[Rule] {1.4149935460868128`, 2.438677657914202`}, 
      StandardDeviation \[Rule] 5.127026169126479`}\)], "Output"]
}, Closed]],

Cell["would fail badly.", "Text"],

Cell["\tBelow we compare the fit to data either in linear", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[\[IndentingNewLine]ErrorListPlot[
          Transpose[{t, y, dy}]], \[IndentingNewLine]Plot[
          Evaluate[Exp[Parameters . {1, tt}] /. zresults], {tt, 0, 1}], 
        PlotRange \[Rule] All, Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<t\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147152 0.000571123 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.21429 -0.0125 -9.5 -15.6875 ]
[.21429 -0.0125 9.5 0 ]
[.40476 -0.0125 -9.5 -15.6875 ]
[.40476 -0.0125 9.5 0 ]
[.59524 -0.0125 -9.625 -15.6875 ]
[.59524 -0.0125 9.625 0 ]
[.78571 -0.0125 -9.5 -15.6875 ]
[.78571 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -3.78125 -31.375 ]
[.5 -0.0125 3.78125 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .12894 -22 -7.84375 ]
[-0.0125 .12894 0 7.84375 ]
[-0.0125 .24316 -22 -7.84375 ]
[-0.0125 .24316 0 7.84375 ]
[-0.0125 .35739 -22 -7.84375 ]
[-0.0125 .35739 0 7.84375 ]
[-0.0125 .47161 -22 -7.84375 ]
[-0.0125 .47161 0 7.84375 ]
[-0.0125 .58584 -28 -7.84375 ]
[-0.0125 .58584 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -43.6875 -5.09375 ]
[-0.01875 .30902 -28 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.21429 0 m
.21429 .00625 L
s
gsave
.21429 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.40476 0 m
.40476 .00625 L
s
gsave
.40476 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.59524 0 m
.59524 .00625 L
s
gsave
.59524 -0.0125 -70.625 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.78571 0 m
.78571 .00625 L
s
gsave
.78571 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.5 0 m
.5 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -64.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(t) show
66.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .12894 m
.00625 .12894 L
s
gsave
-0.0125 .12894 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(200) show
1.000 setlinewidth
grestore
0 .24316 m
.00625 .24316 L
s
gsave
-0.0125 .24316 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(400) show
1.000 setlinewidth
grestore
0 .35739 m
.00625 .35739 L
s
gsave
-0.0125 .35739 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(600) show
1.000 setlinewidth
grestore
0 .47161 m
.00625 .47161 L
s
gsave
-0.0125 .47161 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(800) show
1.000 setlinewidth
grestore
0 .58584 m
.00625 .58584 L
s
gsave
-0.0125 .58584 -89 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1000) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04327 m
.00375 .04327 L
s
0 .07183 m
.00375 .07183 L
s
0 .10038 m
.00375 .10038 L
s
0 .1575 m
.00375 .1575 L
s
0 .18605 m
.00375 .18605 L
s
0 .21461 m
.00375 .21461 L
s
0 .27172 m
.00375 .27172 L
s
0 .30028 m
.00375 .30028 L
s
0 .32883 m
.00375 .32883 L
s
0 .38595 m
.00375 .38595 L
s
0 .4145 m
.00375 .4145 L
s
0 .44306 m
.00375 .44306 L
s
0 .50017 m
.00375 .50017 L
s
0 .52873 m
.00375 .52873 L
s
0 .55728 m
.00375 .55728 L
s
0 .61439 m
.00375 .61439 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -101.938 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.21429 .61178 m
.21429 .61803 L
s
.40476 .61178 m
.40476 .61803 L
s
.59524 .61178 m
.59524 .61803 L
s
.78571 .61178 m
.78571 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .12894 m
1 .12894 L
s
.99375 .24316 m
1 .24316 L
s
.99375 .35739 m
1 .35739 L
s
.99375 .47161 m
1 .47161 L
s
.99375 .58584 m
1 .58584 L
s
.125 Mabswid
.99625 .04327 m
1 .04327 L
s
.99625 .07183 m
1 .07183 L
s
.99625 .10038 m
1 .10038 L
s
.99625 .1575 m
1 .1575 L
s
.99625 .18605 m
1 .18605 L
s
.99625 .21461 m
1 .21461 L
s
.99625 .27172 m
1 .27172 L
s
.99625 .30028 m
1 .30028 L
s
.99625 .32883 m
1 .32883 L
s
.99625 .38595 m
1 .38595 L
s
.99625 .4145 m
1 .4145 L
s
.99625 .44306 m
1 .44306 L
s
.99625 .50017 m
1 .50017 L
s
.99625 .52873 m
1 .52873 L
s
.99625 .55728 m
1 .55728 L
s
.99625 .61439 m
1 .61439 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
.002 w
.02381 .56722 m
.02381 .60332 L
s
.015 w
.02381 .58527 Mdot
.002 w
.04286 .47791 m
.04286 .51101 L
s
.015 w
.04286 .49446 Mdot
.002 w
.0619 .39211 m
.0619 .42205 L
s
.015 w
.0619 .40708 Mdot
.002 w
.08095 .32718 m
.08095 .35447 L
s
.015 w
.08095 .34083 Mdot
.002 w
.1 .27802 m
.1 .30312 L
s
.015 w
.1 .29057 Mdot
.002 w
.11905 .20837 m
.11905 .22998 L
s
.015 w
.11905 .21918 Mdot
.002 w
.1381 .16286 m
.1381 .18183 L
s
.015 w
.1381 .17235 Mdot
.002 w
.15714 .14183 m
.15714 .15945 L
s
.015 w
.15714 .15064 Mdot
.002 w
.17619 .12693 m
.17619 .14352 L
s
.015 w
.17619 .13522 Mdot
.002 w
.19524 .10601 m
.19524 .12103 L
s
.015 w
.19524 .11352 Mdot
.002 w
.21429 .09339 m
.21429 .10738 L
s
.015 w
.21429 .10038 Mdot
.002 w
.23333 .08027 m
.23333 .09309 L
s
.015 w
.23333 .08668 Mdot
.002 w
.25238 .06395 m
.25238 .07514 L
s
.015 w
.25238 .06954 Mdot
.002 w
.27143 .05422 m
.27143 .06431 L
s
.015 w
.27143 .05926 Mdot
.002 w
.29048 .04723 m
.29048 .05644 L
s
.015 w
.29048 .05184 Mdot
.002 w
.30952 .04136 m
.30952 .04975 L
s
.015 w
.30952 .04556 Mdot
.002 w
.32857 .03237 m
.32857 .03932 L
s
.015 w
.32857 .03585 Mdot
.002 w
.34762 .03237 m
.34762 .03932 L
s
.015 w
.34762 .03585 Mdot
.002 w
.36667 .02717 m
.36667 .0331 L
s
.015 w
.36667 .03014 Mdot
.002 w
.38571 .02717 m
.38571 .0331 L
s
.015 w
.38571 .03014 Mdot
.002 w
.40476 .02358 m
.40476 .02869 L
s
.015 w
.40476 .02614 Mdot
.002 w
.42381 .02157 m
.42381 .02614 L
s
.015 w
.42381 .02385 Mdot
.002 w
.44286 .01862 m
.44286 .02223 L
s
.015 w
.44286 .02043 Mdot
.002 w
.4619 .01862 m
.4619 .02223 L
s
.015 w
.4619 .02043 Mdot
.002 w
.48095 .01674 m
.48095 .01954 L
s
.015 w
.48095 .01814 Mdot
.002 w
.5 .01862 m
.5 .02223 L
s
.015 w
.5 .02043 Mdot
.002 w
.51905 .01629 m
.51905 .01885 L
s
.015 w
.51905 .01757 Mdot
.002 w
.5381 .01629 m
.5381 .01885 L
s
.015 w
.5381 .01757 Mdot
.002 w
.55714 .01505 m
.55714 .01667 L
s
.015 w
.55714 .01586 Mdot
.002 w
.57619 .01544 m
.57619 .01742 L
s
.015 w
.57619 .01643 Mdot
.002 w
.59524 .01505 m
.59524 .01667 L
s
.015 w
.59524 .01586 Mdot
.002 w
.61429 .01472 m
.61429 .01586 L
s
.015 w
.61429 .01529 Mdot
.002 w
.63333 .01472 m
.63333 .01586 L
s
.015 w
.63333 .01529 Mdot
.002 w
.65238 .01505 m
.65238 .01667 L
s
.015 w
.65238 .01586 Mdot
.002 w
.67143 .01472 m
.67143 .01586 L
s
.015 w
.67143 .01529 Mdot
.002 w
.69048 .01472 m
.69048 .01586 L
s
.015 w
.69048 .01529 Mdot
.002 w
.70952 .01472 m
.70952 .01586 L
s
.015 w
.70952 .01529 Mdot
.002 w
.72857 .01472 m
.72857 .01472 L
s
.015 w
.72857 .01472 Mdot
.002 w
.74762 .01472 m
.74762 .01586 L
s
.015 w
.74762 .01529 Mdot
.002 w
.76667 .01472 m
.76667 .01586 L
s
.015 w
.76667 .01529 Mdot
.002 w
.78571 .01472 m
.78571 .01472 L
s
.015 w
.78571 .01472 Mdot
.002 w
.80476 .01472 m
.80476 .01472 L
s
.015 w
.80476 .01472 Mdot
.002 w
.82381 .01472 m
.82381 .01472 L
s
.015 w
.82381 .01472 Mdot
.002 w
.84286 .01472 m
.84286 .01586 L
s
.015 w
.84286 .01529 Mdot
.002 w
.8619 .01472 m
.8619 .01472 L
s
.015 w
.8619 .01472 Mdot
.002 w
.88095 .01472 m
.88095 .01472 L
s
.015 w
.88095 .01472 Mdot
.002 w
.9 .01472 m
.9 .01472 L
s
.015 w
.9 .01472 Mdot
.002 w
.91905 .01472 m
.91905 .01472 L
s
.015 w
.91905 .01472 Mdot
.002 w
.9381 .01472 m
.9381 .01472 L
s
.015 w
.9381 .01472 Mdot
.002 w
.95714 .01472 m
.95714 .01472 L
s
.015 w
.95714 .01472 Mdot
.002 w
.97619 .01472 m
.97619 .01472 L
s
.015 w
.97619 .01472 Mdot
.5 Mabswid
.02381 .59163 m
.04262 .48956 L
.06244 .40144 L
.08426 .32326 L
.10458 .26472 L
.14264 .18331 L
.16305 .15119 L
.18163 .12731 L
.20239 .10553 L
.22156 .08919 L
.25997 .06475 L
.2787 .05593 L
.29932 .04801 L
.32036 .04149 L
.33961 .03665 L
.37838 .0294 L
.39737 .02678 L
.41809 .02445 L
.43941 .02252 L
.45874 .02111 L
.49787 .01898 L
.51711 .01821 L
.53793 .01753 L
.55748 .01701 L
.57894 .01656 L
.61842 .01594 L
.63793 .01571 L
.65885 .01552 L
.67866 .01537 L
.70021 .01524 L
.74005 .01506 L
.75983 .015 L
.78084 .01494 L
.81859 .01487 L
.83793 .01484 L
.85879 .01482 L
.87844 .0148 L
.89993 .01478 L
.91909 .01477 L
.93955 .01476 L
.97619 .01475 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00n03oool0103o_g@000000000B02Lgond0?ooo`00n03oool01@3o_g@0001809cOg`2LM9`0
gooo0;<0oooo003h0?ooo`030?oo_`1d0000M;oo0;D0oooo003i0?ooo`030;md0000M;l0oooo0;@0
oooo003i0?ooo`030=nLB000B9`0gooo0;@0oooo003i0?ooo`030?oOW01804P0W=oo0;@0oooo003i
0?ooo`030?oog`2LB000B9cO0;@0oooo003h0?ooo`030?oo_`1d000000000080000000<0M;oo0?oo
o`3oool0/03oool00?X0oooo00<0okmd0000M02oool0/`3oool00?/0oooo00<0giad09cOo`3oool0
/P3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool005@0oooo
00@0_g@0000000000000M;oo=`3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo
0080oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`0]0?ooo`040;md000000000000
07Boo`<0oooo00<0omnL04P0001d_ol01@3oool00`3ooml0W4P004RLg`0_0?ooo`040;md00000000
000007Boo`<0oooo00<0omnL04P0001d_ol00`3oool01@3ooml0W4P000000000000007Bo0340oooo
00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`030?ooo`060?oo_`1d000000000000
0000B9`0gooo<`3oool00`3ookl0M000000000020000000304RLg`3oool0oooo00d0oooo001C0?oo
o`060=nLB000M;l0oooo0?oOW0180000B9cO=@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo02`0
oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00H0oooo00<0gia8
0018W03Oool0;@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0
0`3oool01`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool0;P3oool01P3OW4P007Bo0?ooo`3ogi`0
B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool00`3ookl0M00004RLg`020?ooo`030=nLB000
07@0_ooo03@0oooo00<0okmd0000001d_ol0403oool00580oooo00<0oooO09a80018W=l00P3oool0
0`3o_g@0001809cOo`0c0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool0
0`3o_g@007Bo0?ooo`0_0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02P3oool0
0`3ogi`0B01809cOo`0/0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool0
0`3ooml0W4P007Boo`020?ooo`030?oOW0180000B9cO02d0oooo00<0oooO09a80018W=l00P3oool0
0`3o_g@0001809cOo`060?ooo`030?noM000B9`0gooo00<0oooo00<0_g@004RLg`3oool0<`3oool0
0`3ooml0W4P0001d_`0@0?ooo`00DP3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL
0=oooc80oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?noM000
M;l0oooo02h0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?ooo`030?no
M0000000000000@0000000<004RL0=ooo`3oool0:@3oool00`3ooml0W4P004RLg`020?ooo`040?oo
g`2LB00004RL0=ooo`D0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`0/0?ooo`03
0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3o_g@004RL0=ooo`030?oo
o`030;md0000M;l0oooo03@0oooo00<0gia80018W03Oool03`3oool00580oooo00<0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`0b0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol02@3oool00`3o_g@0B9cO0?ooo`0]0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol01@3oool00`3ooml0W4Q809cOo`020?ooo`030;md0000M;l0oooo02/0oooo00<0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol0;03oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a8001d
_ol00P3oool00`3ookl0M00004RLg`0e0?ooo`030?oOW01804P0W=oo00l0oooo001C0?ooo`030;md
0000B9`0gooo0080oooo00<0omnL04P00018W=l0<`3oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00X0oooo00<0gia804RLg`3oool0;@3oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00H0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo02`0oooo00<0_g@00018
W03Oool00P3oool00`3ogi`0B00004RLg`060?ooo`030;md000007@0_ooo0080oooo00<0gia80000
B02Lgol0;@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo00H0_g@00000
B02Lgml0W4P00000M02ooole0?ooo`030?oo_`1d0000M;oo00l0oooo001C0?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l0<`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00/0oooo00<0_g@007Boo`3oool0;03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00L0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol/0?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l01P3oool0203o_g@0000004RLg`3oool0oooO09a80000B9`0gooo;@3oool0
0`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00D0oooO09a80000000004RL0=oo
o`0f0?ooo`030;md0000M;l0oooo00h0oooo001C0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@0
0018W03Oool0<P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL
04P0M02oool0;03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00D0okmd
001d_`3ooml0W4P004RLg`0/0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3o
ool00`3ooml0W4P000000002000000030000M02oool0oooo02d0oooo00<0omnL04P00018W=l00`3o
ool00`2oM00004RL0=ooo`070?ooo`060;md000004P0W=oo0?oOW018B9`0gooo=@3oool00`3OW4P0
04RL0=ooo`0>0?ooo`00D`3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo0380
oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`0:0?ooo`030?oog`2LB000M;oo
02`0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`040=nLB018W=l0
_g@0001d_b`0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oo
_`1d000007Bo0340oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`03
0?oOW0180000M;oo0080oooo00<0okmd04RLg`3oool0=03oool00`3ogi`0B01809cOo`0>0?ooo`00
E03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo03<0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO02d0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`090?ooo`040=nLB01dM7@0001d0;ooob`0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030?oOW0180000B9cO0340oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03O
ool0=03oool00`3ookl0M00007Boo`0>0?ooo`00E@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
gcD0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0
001809cOo`0^0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO2`3oool00`2oM000001809cOo`0]
0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO2@3oool0103ooml0W4P00000M02oool`0?ooo`06
0=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`3ooml0W4P004RLg`020?ooo`030;md0000M;l0
oooo03D0oooo00<0_g@0001d_`3oool03@3oool005H0oooo00@0gia800000000000007Bo=`3oool0
103OW4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol0<03oool0103OW4P00000
00000000M;l<0?ooo`030?oog`2LB000B9cO02h0oooo00@0gia800000000000007Bo303oool0103o
okl0M00000000018W=l_0?ooo`040=nLB00000000000001d_`T0oooo00D0oooO09a8000000000000
001d_`0e0?ooo`050?oo_`1d000000000000M02oool03@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool00400oooo00@0_g@0000000000000M;oo2P3ooooo0000
05`000001`3oool003l0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool01@3oool003h0oooo00<0oooO09a80018W=l00P3oool00`3o
_g@0001809cOo`080?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0?`3oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0?`3oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool003h0
oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo
0=00oooo0`0000040?ooo`<000002P3oool3000000@0oooo0`0000030?ooo`<00000103oool30000
00@0oooo0`0000030?ooo`<00000103oool3000000<0oooo0`0000040?ooo`<000000`3oool30000
00@0oooo0`0000040?ooo`<000000`3oool3000000@0oooo0`0000030?ooo`<00000103oool30000
00@0oooo0`0000060?ooo`030000003oool0oooo00D0oooo000n0?ooo`030?oog`2LB00007Bo00<0
oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2R0?ooo`<000002P3oool3000000@0
oooo0`0000040?ooo`<000000`3oool3000000@0oooo0`0000020?ooo`D000000P3oool5000000<0
oooo0`0000020?ooo`D000000P3oool5000000030?ooo`000000000000<000000P3oool500000080
oooo1@0000000`3oool000000000000300000080oooo1@0000000`3oool000000000000300000080
oooo1@0000000`3oool000000000000300000080oooo1@0000020?ooo`D0000000<0oooo00000000
00000`0000020?ooo`D0000000<0oooo0000000000000`0000020?ooo`D000000P3oool5000000D0
oooo00<000000?ooo`3oool01@3oool003l0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B000
04RLg`070?ooo`<00000U@3oool3000000@0oooo0`0000020?ooo`D000000`3oool300000080oooo
1@0000020?ooo`D000000P3oool5000000030?ooo`000000000000<000000P3oool5000000030?oo
o`000000000000<000000P3oooml000000<0oooo0`0000070?ooo`00?`3oool00`3OW4P0001809cO
o`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0SP3oool3000000<0oooo1@00
00020?ooo`D0000000<0oooo0000000000000`0000020?oooc<000000P3oool5000000030?ooo`00
0000000000<000000P3oool5000000030?ooo`000000000000<000000P3oool500000080oooo1@00
00000`3oool000000000000300000080oooo1@0000000`3oool000000000000300000080oooo1@00
00000`3oool000000000000300000080oooo1@0000020?ooo`D0000000<0oooo0000000000000`00
00020?ooo`D0000000<0oooo0000000000000`0000020?ooo`D000000P3oool5000000D0oooo00<0
00000?ooo`3oool01@3oool003l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`06
0?ooo`030000003oool0oooo08P0oooo0`0000020?ooo`D000000P3ooolI000000030?ooo`000000
000000<000000P3oool500000080oooo1@0000000`3oool000000000000300000080oooo1@000002
0?ooo`<00000103oool3000000<0oooo1@0000020?ooo`<00000103oool3000000<0oooo0`000004
0?ooo`<00000103oool3000000<0oooo0`0000040?ooo`<000000`3oool3000000@0oooo0`000003
0?ooo`<00000103oool3000000@0oooo0`0000030?ooo`<00000103oool3000000<0oooo0`000004
0?ooo`<00000103oool3000000H0oooo00<000000?ooo`3oool01@3oool003l0oooo00@0oooO09a8
0000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030000003oool0oooo08L0oooo4P000002
0?ooo`D000000P3oool3000000<0oooo1@0000020?ooo`<00000103oool3000000@0oooo0`000003
0?ooo`<00000103oool300000140oooo0`00001f0?ooo`030000003oool0oooo00D0oooo00100?oo
o`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`000000oooo0?ooo`1k0?oo
o`<000000`3oool;000000030?ooo`000000000000<000000`3oool3000000@0oooo0`00000:0?oo
o`<00000Z`3oool00`000000oooo0?ooo`050?ooo`00@@3oool01P3OW4P0001d0;ooo`3ooml0W4P0
04RLg`L0oooo00<000000?ooo`3oool0KP3oool3000000<0oooo0`0000030?ooo`/000000P3oool5
00000080oooo0`0000360?ooo`030000003oool0oooo00D0oooo00120?ooo`040=nLB00000000000
001d_`P0oooo00<000000?ooo`3oool0K@3oool5000000030?ooo`000000000000<000000P3oool5
000000030?ooo`000000000000<000000`3oool300000<`0oooo00<000000?ooo`3oool01@3oool0
04h0oooo00<000000?ooo`3oool0K@3oool5000000030?ooo`000000000000X0000000<0oooo0000
000000000`00003B0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo06d0
oooo2`0000030?ooo`<000000`3oool300000=<0oooo00<000000?ooo`3oool01@3oool004h0oooo
00<000000?ooo`3oool0I`3oool3000000030?ooo`000000000000@000000`3oool300000>00oooo
00<000000?ooo`3oool01@3oool004h0oooo0P00001W0?ooo`D00000j`3oool2000000L0oooo001>
0?ooo`030000003oool0oooo0600oooo0`0000030?ooo`D00000k03oool00`000000oooo0?ooo`05
0?ooo`00CP3oool00`000000oooo0?ooo`1O0?ooo``00000k03oool00`000000oooo0?ooo`050?oo
o`00CP3oool00`000000oooo0?ooo`1O0?ooo`D000000`3oool300000>d0oooo00<000000?ooo`3o
ool01@3oool004h0oooo00<000000?ooo`3oool0FP3oool:00000?<0oooo00<000000?ooo`3oool0
1@3oool004h0oooo00<000000?ooo`3oool0F@3oool500000080oooo0`00003d0?ooo`030000003o
ool0oooo00D0oooo001>0?ooo`030000003oool0oooo05T0oooo1@00003i0?ooo`030000003oool0
oooo00D0oooo001>0?ooo`030000003oool0oooo05<0oooo0`0000000`3oool00000000000050000
0?T0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0DP3oool5000000<0
oooo0`00003j0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0580oooo
1@00003o0?ooo`40oooo00<000000?ooo`3oool01@3oool004h0oooo0P00001C0?ooo`D00000o`3o
ool2000000L0oooo001>0?ooo`030000003oool0oooo0540oooo00<000000?ooo`0000000P00003o
0?ooo`80oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0D03oool00`00
0000oooo0?ooo`3o0?ooo`D0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3o
ool0C@3oool300000?l0oooo203oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo
0?ooo`1<0?ooo`D00000o`3oool70?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003o
ool0oooo04/0oooo1P00003o0?ooo`L0oooo00<000000?ooo`3oool01@3oool004h0oooo00<00000
0?ooo`3oool0BP3oool00`000000oooo0000000400000?l0oooo1`3oool00`000000oooo0?ooo`05
0?ooo`00CP3oool00`000000oooo0?ooo`170?ooo`030000003oool0000000<0oooo0`00003o0?oo
o`P0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0AP3oool300000?l0
oooo3`3oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`150?ooo`D00000
o`3oool>0?ooo`030000003oool0oooo00D0oooo001>0?ooo`800000AP3oool500000?l0oooo3@3o
ool2000000L0oooo001>0?ooo`030000003oool0oooo04@0oooo1P00003o0?ooo`h0oooo00<00000
0?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0?`3oool3000000040?ooo`000000oooo
0?ooo`<00000o`3oool?0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo
03h0oooo1@00003o0?oooaD0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3o
ool0?P3oool500000?l0oooo5@3oool00`000000oooo0?ooo`050?ooo`00;03oool00`3OW4P00000
000000030000000304RLg`3oool0oooo0080oooo00@0_g@0000000000000M;oo1@3oool0102oM000
00000000001d_ol:0?ooo`030000003oool0oooo03h0oooo1@00003o0?oooaD0oooo00<000000?oo
o`3oool01@3oool002d0oooo00<0okmd001d_`3oool00P3oool02P2oM7@0_ooo0?ooo`3oool0gia8
001d_`3oool0omnL04P00018W=l30?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`00
0000oooo0?ooo`0o0?ooo`<00000o`3ooolF0?ooo`030000003oool0oooo00D0oooo000^0?ooo`03
0?noM000M;l0oooo00@0oooo00<0oooO09a80018W=l00P3oool01`3o_g@0001809cOo`3oool0oooO
09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo03h0oooo00<00000
0?ooo`000000o`3ooolG0?ooo`030000003oool0oooo00D0oooo000_0?ooo`030?noM000M;l0oooo
00<0oooo00<0oooO09a80018W=l00P3oool01`3ooml0W4P00018W03Oool0oooO09a80018W=l00P3o
ool0103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0
o`3ooolH0?ooo`030000003oool0oooo00D0oooo000`0?ooo`030?noM018W=l0oooo0080oooo00<0
oooO09a80000M;l00`3oool01P3OW4P0001809cOo`3ooml0W4P0001d_`<0oooo00<0gia80000B02L
gol01`3oool00`000000oooo0?ooo`0j0?ooo`030000003oool000000?l0oooo6`3oool00`000000
oooo0?ooo`050?ooo`00<@3oool00`3OW4P0B9cO0?ooo`020?ooo`030;md0000B9`0gooo0080oooo
00L0omnL04P00018W=l0oooo0;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool30000
03T0oooo0`00003o0?oooaX0oooo0`0000070?ooo`00<P3oool01`2oM000M;oo0?ooo`3oool0gia8
0000B02Lgol00P3oool01`3ooml0W4P0001d_`3oool0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`070?ooo`030000003oool0oooo03P0oooo1@00003o0?oooa/0oooo00<000000?ooo`3oool0
1@3oool00380oooo00L0omnL04P0M02oool0oooo0?oOW0180000B9cO00<0oooo00H0_g@00018W03O
ool0omnL04P00018W=l30?ooo`030;md0000B9`0gooo00H0oooo00<000000?ooo`3oool0>03oool5
00000?l0oooo6`3oool00`000000oooo0?ooo`050?ooo`00<P3oool0203ooml0W4P007Boo`3oool0
oooO09a80000B9`0gooo0P3oool01`3OW4P004RL0=ooo`3ooml0W4P00018W03Oool00P3oool00`3O
W4P004RL0=ooo`060?ooo`030000003oool0oooo03P0oooo1@00003o0?oooa/0oooo00<000000?oo
o`3oool01@3oool00380oooo00<0oono07@00018W=l00P3oool00`3ogi`0B00007Boo`020?ooo`07
0=nLB000M;l0oooo0?ooo`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00H0oooo00<00000
0?ooo`3oool0<`3oool00`000000oooo0?ooo`020?ooo`@00000o`3ooolL0?ooo`030000003oool0
oooo00D0oooo000^0?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`030?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO0`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<00000
0?ooo`3oool0<P3oool300000080oooo00@000000?ooo`3oool00000o`3ooolM0?ooo`030000003o
ool0oooo00D0oooo000_0?ooo`050?oOW018000000000000001d_ol01@3oool0103OW4P000000000
0000M;l50?ooo`040=nLB00000000000001d_`P0oooo00<000000?ooo`3oool0<@3oool600000?l0
oooo8@3oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0a0?ooo`D00000
o`3ooolR0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0340oooo1@00
003o0?ooob80oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0<P3oool3
00000?l0oooo8`3oool00`000000oooo0?ooo`050?ooo`00CP3oool2000003@0oooo00<000000?oo
o`3oool0o`3ooolQ0?ooo`8000001`3oool004h0oooo00<000000?ooo`3oool0;@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0?l0oooo8`3oool00`000000oooo0?ooo`050?ooo`00
CP3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
o`3ooolS0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo02`0oooo0`00
00020?ooo`030000003oool0oooo0?l0oooo903oool00`000000oooo0?ooo`050?ooo`00CP3oool0
0`000000oooo0?ooo`0[0?ooo`D0000000<0oooo0000003oool0o`3ooolU0?ooo`030000003oool0
oooo00D0oooo001>0?ooo`030000003oool0oooo02/0oooo1P00003o0?ooobL0oooo00<000000?oo
o`3oool01@3oool004h0oooo00<000000?ooo`3oool0:`3oool500000?l0oooo:03oool00`000000
oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0/0?ooo`@00000o`3ooolX0?ooo`030000
003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo02d0oooo0P00003o0?ooobT0oooo00<0
00000?ooo`3oool01@3oool004h0oooo0P00000_0?ooo`030000003oool0oooo0?l0oooo9P3oool2
000000L0oooo001>0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0o`3ooolX0?oo
o`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3o
ool0o`3ooolX0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool0o`3ooolY0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool0o`3ooolY0?ooo`030000003oool0oooo00D0oooo001>
0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0o`3ooolZ0?ooo`030000003oool0
oooo00D0oooo001>0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0o`3ooolZ0?oo
o`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3o
ool0o`3oool[0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo02H0oooo
00D000000?ooo`3oool0oooo0000003o0?ooobd0oooo00<000000?ooo`3oool01@3oool004h0oooo
00<000000?ooo`3oool09P3oool01@000000oooo0?ooo`3oool000000?l0oooo;@3oool00`000000
oooo0?ooo`050?ooo`00CP3oool2000002H0oooo0`0000000`3oool000000?ooo`3o0?ooob`0oooo
0P0000070?ooo`00CP3oool00`000000oooo0?ooo`0T0?ooo`H00000o`3oool^0?ooo`030000003o
ool0oooo00D0oooo001>0?ooo`030000003oool0oooo02@0oooo1@00003o0?ooobl0oooo00<00000
0?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0903oool500000?l0oooo;`3oool00`00
0000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0U0?ooo`<00000o`3oool`0?ooo`03
0000003oool0oooo00D0oooo000_0?ooo`030?oog`2LB000B9cO00D0oooo00@0_g@0000000000000
M;oo1@3oool0102oM00000000000001d_ol:0?ooo`030000003oool0oooo02H0oooo0P00003o0?oo
oc00oooo00<000000?ooo`3oool01@3oool00300oooo00<0gia80018W03Oool00`3oool01P3OW4P0
07Bo0?ooo`3ogi`0B00004RLg`<0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool0o`3oool_0?ooo`030000003oool0oooo00D0oooo
000`0?ooo`030?oOW01804P0W=oo0080oooo00<0oooO09a80018W=l00P3oool01`3o_g@0001809cO
o`3oool0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo02H0
oooo00<000000?ooo`3oool0o`3oool_0?ooo`030000003oool0oooo00D0oooo000/0?ooo`030?no
M0000000000000@0000000D004RL0=ooo`3ooml0W4P004RLg`020?ooo`070?oog`2LB00004RL0=oo
o`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool09@3o
ool00`000000oooo0?ooo`3o0?oooc00oooo00<000000?ooo`3oool01@3oool002`0oooo00<0oooO
09a8B02Lgol00P3oool01`2oM00007Bo0?ooo`3oool0oooO09a80000M;l00`3oool01P3OW4P00018
09cOo`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`0U0?oo
o`030000003oool0oooo0?l0oooo<03oool00`000000oooo0?ooo`050?ooo`00;@3oool01`3ookl0
M01d0;ooo`3oool0okmd0018W03Oool00P3oool00`2oM00004RL0=ooo`020?ooo`070?oOW0180000
B9cO0?ooo`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo0`00000U0?ooo`030000
003oool0oooo0?l0oooo;P3oool3000000L0oooo000^0?ooo`060?oOW018B9`0gooo0?oOW01804P0
W=oo0P3oool00`3OW4P0001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P0001809cOo`02
0?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`3o
0?oooc40oooo00<000000?ooo`3oool01@3oool002l0oooo00D0okmd001d_`3ooml0W4P004RLg`02
0?ooo`030?oOW0180000B9cO00<0oooo00H0_g@00018W03Oool0omnL04P00018W=l30?ooo`030;md
0000B9`0gooo00H0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`3o0?oooc40oooo
00<000000?ooo`3oool01@3oool00300oooo00@0gia804RLg`2oM00007Bo0P3oool0103ooml0W4P0
0018W03Oool20?ooo`070=nLB000B9`0gooo0?oog`2LB00004RL0=ooo`020?ooo`030=nLB000B9`0
gooo00H0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?oooc80oooo00<00000
0?ooo`3oool01@3oool00340oooo00@0gia807AdM00007@0_ooo0P3oool00`3ogi`0B00007Boo`02
0?ooo`070=nLB000M;l0oooo0?ooo`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00H0oooo
00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?oooc80oooo00<000000?ooo`3oool0
1@3oool00380oooo00<0_g@00000B02Lgol00`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`<0
oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`030000003oool0oooo0280oooo00<00000
0?ooo`3oool0o`3ooolc0?ooo`030000003oool0oooo00D0oooo000b0?ooo`030?oog`2LB000B9cO
00@0oooo00@0gia800000000000007Bo1@3oool0103OW4P0000000000000M;l80?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`3oool0o`3ooolc0?ooo`030000003oool0oooo00D0oooo001>
0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0o`3ooold0?ooo`030000003oool0
oooo00D0oooo001>0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0o`3ooold0?oo
o`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool0o`3ooold0?ooo`030000003oool0oooo00D0oooo001>0?ooo`8000008@3oool00`000000oooo
0?ooo`3o0?oooc@0oooo0P0000070?ooo`00CP3oool00`000000oooo0?ooo`0P0?ooo`030000003o
ool0oooo0?l0oooo=@3oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0P
0?ooo`030000003oool0oooo0?l0oooo=@3oool00`000000oooo0?ooo`050?ooo`00CP3oool00`00
0000oooo0?ooo`0O0?ooo`800000o`3ooolg0?ooo`030000003oool0oooo00D0oooo001>0?ooo`03
0000003oool0oooo01l0oooo0P00003o0?ooocL0oooo00<000000?ooo`3oool01@3oool004h0oooo
00<000000?ooo`3oool07P3oool300000?l0oooo=`3oool00`000000oooo0?ooo`050?ooo`00CP3o
ool00`000000oooo0?ooo`0M0?ooo`D00000o`3ooolf0?ooo`030000003oool0oooo00D0oooo001>
0?ooo`030000003oool0oooo01d0oooo1@00003o0?ooocH0oooo00<000000?ooo`3oool01@3oool0
04h0oooo00<000000?ooo`3oool07@3oool500000?l0oooo=P3oool00`000000oooo0?ooo`050?oo
o`006P3oool00`2oM00004RL0=ooo`0a0?ooo`030000003oool0oooo01d0oooo1000003o0?ooocL0
oooo00<000000?ooo`3oool01@3oool001X0oooo00<0omnL04Qd_`3oool0<@3oool2000001h0oooo
00@000000?ooo`3oool00000o`3ooolf0?ooo`8000001`3oool000l0oooo00<0oono07Ad_`3oool0
1`3oool00`3o_g@004RL0=ooo`0b0?ooo`030000003oool0oooo01`0oooo00D000000?ooo`3oool0
oooo0000003o0?ooocL0oooo00<000000?ooo`3oool01@3oool000l0oooo00@0omnL04P0000007@0
_ooo1@3oool00`3OW4P007Bo0?ooo`0c0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0o`3oooli0?ooo`030000003oool0oooo00D0oooo000@0?ooo`030?oOW0180000000000D00000
00<0B9cO0?ooo`3oool0<`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0?l0oooo
>@3oool00`000000oooo0?ooo`050?ooo`005@3oool00`3o_g@0001809cOo`0f0?ooo`030000003o
ool0oooo01/0oooo00<000000?ooo`3oool0o`3ooolj0?ooo`030000003oool0oooo00D0oooo000D
0?ooo`030;md0000M;l0oooo03L0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`3o
0?ooocX0oooo00<000000?ooo`3oool01@3oool000l0oooo00H0omnL04P0001d_ol0okmd0000001d
_oli0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0o`3ooolj0?ooo`030000003o
ool0oooo00D0oooo000?0?ooo`040?oOW01800000000001d_c/0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`3o0?oooc/0oooo00<000000?ooo`3oool01@3oool004h0oooo00<00000
0?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?oooc/0oooo00<000000?ooo`3oool01@3oool0
04h0oooo00<000000?ooo`3oool06@3oool200000?l0oooo?@3oool00`000000oooo0?ooo`050?oo
o`00CP3oool2000001X0oooo00<000000?ooo`3oool0o`3ooolk0?ooo`8000001`3oool004h0oooo
00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`3o0?oooc`0oooo00<000000?ooo`3oool0
1@3oool004h0oooo00<000000?ooo`3oool0603oool300000?l0oooo?@3oool00`000000oooo0?oo
o`050?ooo`00CP3oool00`000000oooo0?ooo`0G0?ooo`D00000o`3oooll0?ooo`030000003oool0
oooo00D0oooo001>0?ooo`030000003oool0oooo01L0oooo1@00003o0?oooc`0oooo00<000000?oo
o`3oool01@3oool002d0oooo00D0oooO09a8000000000000001d_`050?ooo`040;md000000000000
07Boo`D0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo0?ooo`0G0?ooo`D00000o`3o
ooll0?ooo`030000003oool0oooo00D0oooo000]0?ooo`070=nLB000M;l0oooo0?oo_`1d000004RL
0=ooo`020?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0`3oool01P3OW4P007Bo0?ooo`3ogi`0
B00004RLg`T0oooo00<000000?ooo`3oool0603oool300000?l0oooo?@3oool00`000000oooo0?oo
o`050?ooo`00;03oool00`3ooml0W4P007Boo`020?ooo`070?oOW0180000B9cO0?ooo`3ooml0W4P0
04RLg`020?ooo`070?noM00004P0W=oo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo
00P0oooo00<000000?ooo`3oool0603oool200000?l0oooo?P3oool00`000000oooo0?ooo`050?oo
o`00;03oool00`3ookl0M00007Boo`030?ooo`060;md000007@0_ooo0?oog`2LB000B9cO0P3oool0
1`3ooml0W4P00018W03Oool0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`03
0000003oool0oooo01L0oooo00<000000?ooo`000000o`3oooln0?ooo`030000003oool0oooo00D0
oooo000/0?ooo`030?oog`2LB00007Bo00<0oooo00H0gia80000B02Lgol0oooO09a80000M;l30?oo
o`060=nLB00004P0W=oo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`070?ooo`030000003o
ool0oooo01L0oooo00<000000?ooo`000000o`3oooln0?ooo`030000003oool0oooo00D0oooo000]
0?ooo`030;md000007@0_ooo0080oooo00L0gia80000B02Lgol0oooo0;md0000B9`0gooo0080oooo
00L0omnL04P00018W=l0oooo0;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool30000
01L0oooo00<000000?ooo`3oool0o`3oooll0?ooo`<000001`3oool002d0oooo00`0okmd00000018
W=l0oooo0?oog`2LB00004RL0=ooo`3oool0gia80000B02Lgol20?ooo`070?oog`2LB00007Bo0?oo
o`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`3o0?oooch0oooo00<000000?ooo`3oool01@3oool002d0oooo00<0oooO09a8
000000000P0000001`0007@0_ooo0?ooo`3oool0omnL04P00018W=l00`3oool01P2oM00004RL0=oo
o`3ogi`0B00004RLg`<0oooo00<0_g@00018W03Oool01P3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo0?l0oooo?`3oool00`000000oooo0?ooo`050?ooo`00;P3oool00`3ookl0M000
001d_`050?ooo`040?oog`2LB00004RL0=ooo`80oooo00L0gia80018W03Oool0oooO09a80000B9`0
gooo0080oooo00<0gia80018W03Oool01P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo0?l0oooo?`3oool00`000000oooo0?ooo`050?ooo`00;`3oool00`3ogi`0B00004RLg`050?oo
o`030?oOW0180000M;oo0080oooo00L0gia8001d_`3oool0oooo0?oOW0180000M;oo0080oooo00<0
gia8001d_`3oool01P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0?l0oooo?`3o
ool00`000000oooo0?ooo`050?ooo`00<03oool0103ooml0W4P00000M02oool40?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO0`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?ooocl0oooo00<000000?ooo`3oool01@3oool0
0380oooo00@0oono07@000000000B9cO0`3oool0103OW4P0000000000000M;l50?ooo`040=nLB000
00000000001d_`P0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`3o0?oood00oooo
00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?oo
o`3o0?oood00oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`3o0?oood00oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`3o0?oood00oooo00<000000?ooo`3oool01@3oool004h0
oooo0P00000E0?ooo`030000003oool0oooo0?l0oooo@03oool2000000L0oooo001>0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool0o`3ooom10?ooo`030000003oool0oooo00D0oooo
001>0?ooo`030000003oool0oooo0180oooo00<000000?ooo`000000o`3ooom30?ooo`030000003o
ool0oooo00D0oooo001>0?ooo`030000003oool0oooo0180oooo00<000000?ooo`000000o`3ooom3
0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0180oooo0P00003o0?oo
od@0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool04@3oool300000?l0
ooooA03oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0@0?ooo`D00000
o`3ooom30?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0100oooo1@00
003o0?oood<0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0403oool5
00000?l0oooo@`3oool00`000000oooo0?ooo`050?ooo`00CP3oool200000180oooo0`00003o0?oo
od<0oooo0P0000070?ooo`00CP3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0?l0
oooo@`3oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo0?l0oooo@`3oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?oo
o`0A0?ooo`800000o`3ooom50?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool0o`3ooom40?ooo`030000003oool0oooo00D0oooo001>0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool0o`3ooom40?ooo`030000003oool0oooo
00D0oooo001>0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0o`3ooom40?ooo`03
0000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
o`3ooom50?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool0o`3ooom50?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0
oooo0100oooo00<000000?ooo`3oool0o`3ooom50?ooo`030000003oool0oooo00D0oooo001>0?oo
o`8000004@3oool00`000000oooo0?ooo`3o0?oood@0oooo0P0000070?ooo`00CP3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooAP3oool00`000000oooo0?ooo`050?ooo`00
CP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooAP3oool00`000000oooo
0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0ooooAP3o
ool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo0?l0ooooAP3oool00`000000oooo0?ooo`050?ooo`00;@3oool01P3ookl0M000000000000000
04RL0=ooo`@0oooo00@0_g@0000000000000M;oo1@3oool0102oM00000000000001d_ol:0?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0o`3ooom60?ooo`030000003oool0oooo00D0
oooo000/0?ooo`030?oo_`1d0000B9cO0080oooo00<0gia80000M02oool00P3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`<0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0o`3ooom70?ooo`030000003oool0oooo00D0oooo000/
0?ooo`030?noM000B9`0gooo00<0oooo00H0_g@004RLg`3oool0oooO09a80018W=l20?ooo`070?no
M00004P0W=oo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`3o0?ooodL0oooo00<000000?ooo`3oool01@3oool002`0
oooo00<0okmd0018W03Oool00`3oool01P2oM00007Bo0?ooo`3ooml0W4P004RLg`80oooo00L0oooO
09a80000B9`0gooo0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo0?l0ooooA`3oool00`000000oooo0?ooo`050?ooo`00
;03oool00`3ooml0W4P007Boo`020?ooo`070?oo_`1d0000B9cO0?ooo`3ooml0W4P0001d_`030?oo
o`060=nLB00004P0W=oo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`070?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0o`3ooom70?ooo`030000003oool0oooo00D0oooo000^
0?ooo`060;md000004P0W=oO09a8000007@0_ooo0P3oool00`2oM00004RL0=ooo`020?ooo`070?oO
W0180000B9cO0?ooo`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo0`00000=0?oo
o`030000003oool0oooo0?l0ooooAP3oool3000000L0oooo000_0?ooo`050?oog`2LB00000000018
W03Oool00P3oool00`3OW4P0001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P0001809cO
o`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`3o0?ooodP0oooo00<000000?ooo`3oool01@3oool002l0oooo00X0_g@00000B02Lgol0omnL04Q8
W03Oool0oooo0?oOW0180000B9cO0`3oool01P2oM00004RL0=ooo`3ogi`0B00004RLg`<0oooo00<0
_g@00018W03Oool01P3oool00`000000oooo0?ooo`0<0?ooo`800000o`3ooom:0?ooo`030000003o
ool0oooo00D0oooo000^0?ooo`030?oOW0180000M;oo0080oooo00L0okmd04RLg`3oool0oooO09a8
0000B9`0gooo0080oooo00L0gia80018W03Oool0oooO09a80000B9`0gooo0080oooo00<0gia80018
W03Oool01P3oool00`000000oooo0?ooo`0<0?ooo`800000o`3ooom:0?ooo`030000003oool0oooo
00D0oooo000^0?ooo`030?oOW01804P0W=oo0080oooo00L0okmd0018W03Oool0oooo0?oOW0180000
M;oo0080oooo00L0gia8001d_`3oool0oooo0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0
1P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0ooooB@3oool00`000000oooo
0?ooo`050?ooo`00;P3oool00`3ooml0W4P004RLg`020?ooo`030;md0000M;l0oooo0080oooo00H0
gia80000M02oool0oooO09a80018W=l30?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0ooooB@3oool00`000000oooo0?ooo`05
0?ooo`00;`3oool01@3ooml0W4P000000000000007Bo00D0oooo00@0gia800000000000007Bo1@3o
ool0103OW4P0000000000000M;l80?ooo`030000003oool0oooo00/0oooo0`00003o0?ooodX0oooo
00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool02P3oool500000?l0ooooB@3o
ool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0:0?ooo`D00000o`3ooom9
0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo00X0oooo1@00003o0?oo
odT0oooo00<000000?ooo`3oool01@3oool004h0oooo0P00000<0?ooo`<00000o`3ooom90?ooo`80
00001`3oool004h0oooo00<000000?ooo`3oool02`3oool200000?l0ooooB`3oool00`000000oooo
0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`0;0?ooo`800000o`3ooom;0?ooo`030000003o
ool0oooo00D0oooo001>0?ooo`030000003oool0oooo00/0oooo0P00003o0?oood/0oooo00<00000
0?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0000003o0?oo
od/0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`3o0?oood/0oooo00<000000?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`3o0?oood/0oooo00<000000?ooo`3oool01@3oool004h0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`3o0?oood/0oooo00<000000?ooo`3oool01@3o
ool004h0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`3o0?oood/0oooo00<00000
0?ooo`3oool01@3oool004h0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?oo
od`0oooo00<000000?ooo`3oool01@3oool004h0oooo0P00000:0?ooo`030000003oool0oooo0?l0
ooooB`3oool2000000L0oooo001>0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
o`3ooom<0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0o`3ooom<0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0o`3ooom<0?ooo`030000003oool0oooo00D0oooo001>0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooom=0?ooo`030000003oool0oooo
00D0oooo001>0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooom=0?ooo`03
0000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
o`3ooom=0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0o`3ooom=0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0o`3ooom=0?ooo`030000003oool0oooo00D0oooo001>0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooom=0?ooo`030000003oool0oooo
00D0oooo001>0?ooo`800000203oool00`000000oooo0?ooo`3o0?ooodd0oooo0P0000070?ooo`00
CP3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0?l0ooooCP3oool00`000000oooo
0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0?l0ooooCP3o
ool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`050?ooo`030000003oool0
00000?l0ooooD03oool00`000000oooo0?ooo`050?ooo`00CP3oool00`000000oooo0?ooo`050?oo
o`030000003oool000000?l0ooooD03oool00`000000oooo0?ooo`050?ooo`008`3oool00`3ookl0
M000000000020000000304RLg`3oool0oooo00<0oooo00@0_g@0000000000000M;oo1@3oool0102o
M00000000000001d_ol50?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool01@3o
ool200000?l0ooooD@3oool00`000000oooo0?ooo`050?ooo`009@3oool00`3o_g@0000007Boo`05
0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RL
g`<0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo00D0oooo0P00
003o0?oooe40oooo00<000000?ooo`3oool01@3oool002D0oooo00<0oooO09a80000M;l0103oool0
0`3ooml0W4P004RLg`020?ooo`070?noM00004P0W=oo0?ooo`3ooml0W4P004RLg`020?ooo`070?no
M00004P0W=oo0?ooo`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<000000?oo
o`3oool01@3oool200000?l0ooooD@3oool00`000000oooo0?ooo`050?ooo`009P3oool00`3OW4P0
04RL0=ooo`030?ooo`030?oog`2LB000B9cO0080oooo00L0oooO09a80000B9`0gooo0?oog`2LB000
B9cO0080oooo00L0oooO09a80000B9`0gooo0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0
gooo1`3oool00`000000oooo0?ooo`040?ooo`<00000o`3ooomA0?ooo`030000003oool0oooo00D0
oooo000V0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool01P3OW4P00018
09cOo`3ooml0W4P0001d_`<0oooo00H0gia80000B02Lgol0oooO09a80000M;l30?ooo`030=nLB000
04P0W=oo00L0oooo00<000000?ooo`3oool00`3oool500000?l0ooooD03oool00`000000oooo0?oo
o`050?ooo`009P3oool00`3ookl0M00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00L0omnL
04P00018W=l0oooo0;md0000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0;md0000B9`0gooo
0080oooo00<0omnL04P00018W=l01`3oool3000000<0oooo1@00003o0?ooodh0oooo0`0000070?oo
o`009`3oool00`2oM00007Bo0?ooo`030?ooo`030=nLB00004P0W=oo0080oooo00L0oooO09a80000
M;l0oooo0=nLB00004P0W=oo0080oooo00L0oooO09a80000M;l0oooo0=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l01`3oool00`000000oooo0?ooo`030?ooo`D00000o`3ooom@0?ooo`030000
003oool0oooo00D0oooo000W0?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018W=l00`3o
ool01P2oM00004RL0=ooo`3ogi`0B00004RLg`<0oooo00H0_g@00018W03Oool0omnL04P00018W=l3
0?ooo`030;md0000B9`0gooo00H0oooo00<000000?ooo`3oool0103oool300000?l0ooooD@3oool0
0`000000oooo0?ooo`050?ooo`009`3oool00`3ogi`0B01809cOo`030?ooo`040?oog`2LB00004RL
0=ooo`80oooo00L0gia80018W03Oool0oooO09a80000B9`0gooo0080oooo00L0gia80018W03Oool0
oooO09a80000B9`0gooo0080oooo00<0gia80018W03Oool01P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo0?l0ooooD03oool00`000000oooo0?ooo`050?ooo`009`3oool00`3ookl0
M00007Boo`040?ooo`030?oOW0180000M;oo0080oooo00L0gia8001d_`3oool0oooo0?oOW0180000
M;oo0080oooo00L0gia8001d_`3oool0oooo0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0
1P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0?l0ooooD03oool00`000000oooo
0?ooo`050?ooo`00:03oool00`2oM00007Bo0?ooo`040?ooo`060=nLB00007@0_ooo0?oog`2LB000
B9cO0`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00H0gia80000M02oool0oooO09a8
0018W=l70?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3ooom@0?ooo`030000
003oool0oooo00D0oooo000V0?ooo`050?oo_`1d000000000000M02oool01@3oool0103OW4P00000
00000000M;l50?ooo`040=nLB00000000000001d_`D0oooo00@0gia800000000000007Bo203oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0?l0ooooD03oool00`000000oooo0?ooo`05
0?ooo`00CP3oool00`000000oooo0?ooo`3o0?oooeP0oooo00<000000?ooo`3oool01@3oool004h0
oooo00<000000?ooo`3oool0o`3ooomH0?ooo`030000003oool0oooo00D0oooo001>0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0o0?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0o0?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool004h0oooo0P0000060?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`8000001`3o
ool004h0ooooo`00001L000000L0oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.262897, -285.796, \
0.0045516, 7.59006}}]
}, Closed]],

Cell[TextData[{
  "or semilog formats.  Unfortunately, although the semilog format is \
preferable for this example, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " still lacks a method for plotting error bars on semilog plots."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[\[IndentingNewLine]LogListPlot[
          Evaluate[Transpose[{t, y}]], 
          PlotStyle \[Rule] PointSize[0.02]], \[IndentingNewLine]LogPlot[
          Evaluate[Exp[Parameters . {1, tt} /. zresults]], {tt, 0, 1}], 
        PlotRange \[Rule] {Automatic, {\(-1\), 3.2}}, Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<t\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.147151 0.147151 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.21429 -0.0125 -9.5 -15.6875 ]
[.21429 -0.0125 9.5 0 ]
[.40476 -0.0125 -9.5 -15.6875 ]
[.40476 -0.0125 9.5 0 ]
[.59524 -0.0125 -9.625 -15.6875 ]
[.59524 -0.0125 9.625 0 ]
[.78571 -0.0125 -9.5 -15.6875 ]
[.78571 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -3.78125 -31.375 ]
[.5 -0.0125 3.78125 -15.6875 ]
[-0.0125 0 -19 -7.84375 ]
[-0.0125 0 0 7.84375 ]
[-0.0125 .14715 -10 -7.84375 ]
[-0.0125 .14715 0 7.84375 ]
[-0.0125 .2943 -16 -7.84375 ]
[-0.0125 .2943 0 7.84375 ]
[-0.0125 .44145 -22 -7.84375 ]
[-0.0125 .44145 0 7.84375 ]
[-0.0125 .5886 -28 -7.84375 ]
[-0.0125 .5886 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -43.6875 -5.09375 ]
[-0.01875 .30902 -28 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ -0.0005 0 0 0 ]
[ 1.0005 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.21429 0 m
.21429 .00625 L
s
gsave
.21429 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.40476 0 m
.40476 .00625 L
s
gsave
.40476 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.59524 0 m
.59524 .00625 L
s
gsave
.59524 -0.0125 -70.625 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.78571 0 m
.78571 .00625 L
s
gsave
.78571 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.2619 0 m
.2619 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.5 0 m
.5 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.7381 0 m
.7381 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -64.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(t) show
66.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
.00625 0 L
s
gsave
-0.0125 0 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.1) show
78.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .14715 m
.00625 .14715 L
s
gsave
-0.0125 .14715 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
69.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .2943 m
.00625 .2943 L
s
gsave
-0.0125 .2943 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
75.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .44145 m
.00625 .44145 L
s
gsave
-0.0125 .44145 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(100) show
81.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .5886 m
.00625 .5886 L
s
gsave
-0.0125 .5886 -89 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1000) show
87.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.001 w
0 .0443 m
.00375 .0443 L
s
0 .07021 m
.00375 .07021 L
s
0 .08859 m
.00375 .08859 L
s
0 .10285 m
.00375 .10285 L
s
0 .11451 m
.00375 .11451 L
s
0 .12436 m
.00375 .12436 L
s
0 .13289 m
.00375 .13289 L
s
0 .14042 m
.00375 .14042 L
s
0 .19145 m
.00375 .19145 L
s
0 .21736 m
.00375 .21736 L
s
0 .23574 m
.00375 .23574 L
s
0 .25001 m
.00375 .25001 L
s
0 .26166 m
.00375 .26166 L
s
0 .27151 m
.00375 .27151 L
s
0 .28004 m
.00375 .28004 L
s
0 .28757 m
.00375 .28757 L
s
0 .3386 m
.00375 .3386 L
s
0 .36451 m
.00375 .36451 L
s
0 .3829 m
.00375 .3829 L
s
0 .39716 m
.00375 .39716 L
s
0 .40881 m
.00375 .40881 L
s
0 .41866 m
.00375 .41866 L
s
0 .42719 m
.00375 .42719 L
s
0 .43472 m
.00375 .43472 L
s
0 .48575 m
.00375 .48575 L
s
0 .51166 m
.00375 .51166 L
s
0 .53005 m
.00375 .53005 L
s
0 .54431 m
.00375 .54431 L
s
0 .55596 m
.00375 .55596 L
s
0 .56581 m
.00375 .56581 L
s
0 .57434 m
.00375 .57434 L
s
0 .58187 m
.00375 .58187 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -101.938 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.21429 .61178 m
.21429 .61803 L
s
.40476 .61178 m
.40476 .61803 L
s
.59524 .61178 m
.59524 .61803 L
s
.78571 .61178 m
.78571 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.2619 .61428 m
.2619 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.7381 .61428 m
.7381 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .14715 m
1 .14715 L
s
.99375 .2943 m
1 .2943 L
s
.99375 .44145 m
1 .44145 L
s
.99375 .5886 m
1 .5886 L
s
.001 w
.99625 .0443 m
1 .0443 L
s
.99625 .07021 m
1 .07021 L
s
.99625 .08859 m
1 .08859 L
s
.99625 .10285 m
1 .10285 L
s
.99625 .11451 m
1 .11451 L
s
.99625 .12436 m
1 .12436 L
s
.99625 .13289 m
1 .13289 L
s
.99625 .14042 m
1 .14042 L
s
.99625 .19145 m
1 .19145 L
s
.99625 .21736 m
1 .21736 L
s
.99625 .23574 m
1 .23574 L
s
.99625 .25001 m
1 .25001 L
s
.99625 .26166 m
1 .26166 L
s
.99625 .27151 m
1 .27151 L
s
.99625 .28004 m
1 .28004 L
s
.99625 .28757 m
1 .28757 L
s
.99625 .3386 m
1 .3386 L
s
.99625 .36451 m
1 .36451 L
s
.99625 .3829 m
1 .3829 L
s
.99625 .39716 m
1 .39716 L
s
.99625 .40881 m
1 .40881 L
s
.99625 .41866 m
1 .41866 L
s
.99625 .42719 m
1 .42719 L
s
.99625 .43472 m
1 .43472 L
s
.99625 .48575 m
1 .48575 L
s
.99625 .51166 m
1 .51166 L
s
.99625 .53005 m
1 .53005 L
s
.99625 .54431 m
1 .54431 L
s
.99625 .55596 m
1 .55596 L
s
.99625 .56581 m
1 .56581 L
s
.99625 .57434 m
1 .57434 L
s
.99625 .58187 m
1 .58187 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.02381 .58854 Mdot
.04286 .57746 Mdot
.0619 .56461 Mdot
.08095 .55279 Mdot
.1 .5421 Mdot
.11905 .52296 Mdot
.1381 .50633 Mdot
.15714 .49687 Mdot
.17619 .48917 Mdot
.19524 .47648 Mdot
.21429 .46736 Mdot
.23333 .45622 Mdot
.25238 .43884 Mdot
.27143 .42557 Mdot
.29048 .41392 Mdot
.30952 .40207 Mdot
.32857 .37791 Mdot
.34762 .37791 Mdot
.36667 .35778 Mdot
.38571 .35778 Mdot
.40476 .3386 Mdot
.42381 .32434 Mdot
.44286 .2943 Mdot
.4619 .2943 Mdot
.48095 .26166 Mdot
.5 .2943 Mdot
.51905 .25001 Mdot
.5381 .25001 Mdot
.55714 .19145 Mdot
.57619 .21736 Mdot
.59524 .19145 Mdot
.61429 .14715 Mdot
.63333 .14715 Mdot
.65238 .19145 Mdot
.67143 .14715 Mdot
.69048 .14715 Mdot
.70952 .14715 Mdot
.74762 .14715 Mdot
.76667 .14715 Mdot
.84286 .14715 Mdot
.5 Mabswid
.02381 .58925 m
.06244 .56369 L
.10458 .53581 L
.14415 .50963 L
.18221 .48445 L
.22272 .45765 L
.26171 .43185 L
.30316 .40443 L
.34309 .37801 L
.3815 .3526 L
.42237 .32556 L
.46172 .29952 L
.49955 .27449 L
.53984 .24784 L
.57861 .22218 L
.61984 .19491 L
.65954 .16864 L
.69774 .14337 L
.73838 .11648 L
.77751 .09059 L
.81909 .06308 L
.85916 .03657 L
.89771 .01106 L
s
.89771 .01106 m
.91443 0 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00n03oool0103o_g@000000000B02Lgond0?ooo`00n03oool01@3o_g@0001809cOg`2LM9`0
gooo0;<0oooo003h0?ooo`030?oo_`1d0000M;oo0;D0oooo003i0?ooo`030;md0000M;l0oooo0;@0
oooo003i0?ooo`030=nLB000B9`0gooo0;@0oooo003i0?ooo`030?oOW01804P0W=oo0;@0oooo003i
0?ooo`030?oog`2LB000B9cO0;@0oooo003h0?ooo`030?oo_`1d000000000080000000<0M;oo0?oo
o`3oool0/03oool00?X0oooo00<0okmd0000M02oool0/`3oool00?/0oooo00<0giad09cOo`3oool0
/P3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool005@0oooo
00@0_g@0000000000000M;oo=`3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo
0080oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`0]0?ooo`040;md000000000000
07Boo`<0oooo00<0omnL04P0001d_ol01@3oool00`3ooml0W4P004RLg`0_0?ooo`040;md00000000
000007Boo`<0oooo00<0omnL04P0001d_ol00`3oool01@3ooml0W4P000000000000007Bo0300oooo
00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`030?ooo`060?oo_`1d000000000000
0000B9`0gooo<`3oool00`3ookl0M000000000020000000304RLg`3oool0oooo00h0oooo001C0?oo
o`060=nLB000M;l0oooo0?oOW0180000B9cO=@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80
oooo00<0omnL04P0001d_ol00`3oool00`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo02`0
oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00H0oooo00<0gia8
0018W03Oool0;@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0
0`3oool01`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool0;@3oool01P3OW4P007Bo0?ooo`3ogi`0
B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool00`3ookl0M00004RLg`020?ooo`030=nLB000
07@0_ooo03@0oooo00<0okmd0000001d_ol04@3oool00580oooo00<0oooO09a80018W=l00P3oool0
0`3o_g@0001809cOo`0c0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool0
0`3o_g@007Bo0?ooo`0_0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02P3oool0
0`3ogi`0B01809cOo`0/0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool0
0`3ooml0W4P007Boo`020?ooo`030?oOW0180000B9cO02`0oooo00<0oooO09a80018W=l00P3oool0
0`3o_g@0001809cOo`060?ooo`030?noM000B9`0gooo00<0oooo00<0_g@004RLg`3oool0<`3oool0
0`3ooml0W4P0001d_`0A0?ooo`00DP3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL
0=oooc80oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?noM000
M;l0oooo02h0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?ooo`030?no
M0000000000000@0000000<004RL0=ooo`3oool0:@3oool00`3ooml0W4P004RLg`020?ooo`040?oo
g`2LB00004RL0=ooo`D0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`0[0?ooo`03
0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3o_g@004RL0=ooo`030?oo
o`030;md0000M;l0oooo03@0oooo00<0gia80018W03Oool0403oool00580oooo00<0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`0b0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol02@3oool00`3o_g@0B9cO0?ooo`0]0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol01@3oool00`3ooml0W4Q809cOo`020?ooo`030;md0000M;l0oooo02/0oooo00<0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol0:`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a8001d
_ol00P3oool00`3ookl0M00004RLg`0e0?ooo`030?oOW01804P0W=oo0100oooo001C0?ooo`030;md
0000B9`0gooo0080oooo00<0omnL04P00018W=l0<`3oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00X0oooo00<0gia804RLg`3oool0;@3oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00H0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo02`0oooo00<0_g@00018
W03Oool00P3oool00`3ogi`0B00004RLg`060?ooo`030;md000007@0_ooo0080oooo00<0gia80000
B02Lgol0;03oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo00H0_g@00000
B02Lgml0W4P00000M02ooole0?ooo`030?oo_`1d0000M;oo0100oooo001C0?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l0<`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00/0oooo00<0_g@007Boo`3oool0;03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00L0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol/0?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l01P3oool0203o_g@0000004RLg`3oool0oooO09a80000B9`0gooo;03oool0
0`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00D0oooO09a80000000004RL0=oo
o`0f0?ooo`030;md0000M;l0oooo00l0oooo001C0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@0
0018W03Oool0<P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL
04P0M02oool0;03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00D0okmd
001d_`3ooml0W4P004RLg`0/0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3o
ool00`3ooml0W4P000000002000000030000M02oool0oooo02`0oooo00<0omnL04P00018W=l00`3o
ool00`2oM00004RL0=ooo`070?ooo`060;md000004P0W=oo0?oOW018B9`0gooo=@3oool00`3OW4P0
04RL0=ooo`0?0?ooo`00D`3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo0380
oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`0:0?ooo`030?oog`2LB000M;oo
02`0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`040=nLB018W=l0
_g@0001d_b`0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oo
_`1d000007Bo0300oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`03
0?oOW0180000M;oo0080oooo00<0okmd04RLg`3oool0=03oool00`3ogi`0B01809cOo`0?0?ooo`00
E03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo03<0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO02d0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`090?ooo`040=nLB01dM7@0001d0;ooob`0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030?oOW0180000B9cO0300oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03O
ool0=03oool00`3ookl0M00007Boo`0?0?ooo`00E@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
gcD0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0
001809cOo`0^0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO2`3oool00`2oM000001809cOo`0]
0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO2@3oool0103ooml0W4P00000M02oool_0?ooo`06
0=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`3ooml0W4P004RLg`020?ooo`030;md0000M;l0
oooo03D0oooo00<0_g@0001d_`3oool03P3oool005H0oooo00@0gia800000000000007Bo=`3oool0
103OW4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol0<03oool0103OW4P00000
00000000M;l<0?ooo`030?oog`2LB000B9cO02h0oooo00@0gia800000000000007Bo303oool0103o
okl0M00000000018W=l^0?ooo`040=nLB00000000000001d_`T0oooo00D0oooO09a8000000000000
001d_`0e0?ooo`050?oo_`1d000000000000M02oool03P3oool00?l0oooo/@3oool00?l0oooo/@3o
ool00?l0oooo/@3oool00?l0oooo/@3oool003<0oooo00@0_g@0000000000000M;oo0`3oool00`3o
gi`0B00007Boo`020?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool0o`3ooomZ0?oo
o`00<P3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0103oool0
0`3o_g@0000007Boo`3o0?ooofd0oooo000a0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000
B02Lgol0203oool00`3ooml0W4P0001d_`3o0?ooofd0oooo000a0?ooo`030?oog`2LB000B9cO0080
oooo00@0oooO09a80000B9`0gooo203oool00`3OW4P004RL0=ooo`3o0?ooof`0oooo000a0?ooo`03
0?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol0203oool00`3ogi`0B01809cOo`3o0?ooof`0
oooo000b0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0203oool00`3ookl0M000
07Boo`0:0?ooool00000FP0000080?ooo`00<P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2L
B00007Bo00T0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00L0oooo0P0000050?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`060?ooo`00<P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0
gooo00P0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo03h0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0
oooo03h0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool01P3oool00380oooo00@0oooO09a80000B9`0gooo0P3oool0
0`3OW4P004RL0=ooo`080?ooo`030?oOW01804P0W=oo00T0oooo00<000000?ooo`3oool0o`3ooold
0?ooo`800000803oool00`000000oooo0?ooo`060?ooo`00<`3oool00`3ogi`0B00007Boo`020?oo
o`030=nLB000M;l0oooo00P0oooo00<0oono07@0001d_ol02@3oool00`000000oooo0?ooo`3o0?oo
oc<0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`060?ooo`00=03oool01P3OW4P0
001d0;ooo`3ooml0W4P004RLg`X0oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`3o
0?oooc40oooo0P00000S0?ooo`030000003oool0oooo00H0oooo000e0?ooo`040=nLB00000000000
001d_`T0oooo00D0oono07@000000000001d0;ooo`080?ooo`030000003oool0oooo0?l0oooo<03o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0
oooo0?l0oooo;P3oool2000002H0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?oo
o`3oool0o`3oool]0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool01P3oool004l0
oooo00<000000?ooo`3oool0o`3oool/0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool01P3oool004l0oooo00<000000?ooo`3oool0o`3ooolZ0?ooo`800000:P3oool00`000000oooo
0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`3o0?ooobT0oooo00<000000?ooo`3oool0:P3o
ool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`3o0?ooobL0oooo0P00000]
0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?l0oooo9P3oool00`00
0000oooo0?ooo`0]0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?l0
oooo903oool200000300oooo00<000000?ooo`3oool01P3oool004l0oooo0P00003o0?ooob@0oooo
00<000000?ooo`3oool0;`3oool2000000P0oooo001?0?ooo`030000003oool0oooo0?l0oooo8@3o
ool2000003<0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0o`3ooolP
0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?oo
o`3oool0o`3ooolN0?ooo`800000=P3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000
oooo0?ooo`3o0?oooa`0oooo0P00000h0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000
003oool0oooo0?l0oooo6`3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo00H0oooo
001?0?ooo`030000003oool0oooo0?l0oooo6@3oool2000003/0oooo00<000000?ooo`3oool01P3o
ool004l0oooo00<000000?ooo`3oool0o`3ooolH0?ooo`030000003oool0oooo03/0oooo00<00000
0?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0o`3ooolF0?ooo`800000?P3oool00`00
0000oooo0?ooo`060?ooo`00C`3oool200000?l0oooo5@3oool2000003l0oooo0P0000080?ooo`00
C`3oool00`000000oooo0?ooo`3o0?oooa<0oooo00<000000?ooo`3oool0@03oool00`000000oooo
0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`3o0?oooa40oooo0P0000130?ooo`030000003o
ool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?l0oooo3`3oool2000004D0oooo00<00000
0?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0o`3oool>0?ooo`030000003oool0oooo
04D0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0o`3oool<0?ooo`80
0000B03oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`3o0?ooo`X0oooo
0P00001:0?ooo`030000003oool0oooo00H0oooo001?0?ooo`800000o`3oool:0?ooo`030000003o
ool0oooo04T0oooo0P0000080?ooo`00C`3oool00`000000oooo0?ooo`3o0?ooo`L0oooo0P00001=
0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?l0oooo1P3oool00`00
0000oooo0?ooo`1=0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?l0
oooo103oool200000500oooo00<000000?ooo`3oool01P3oool004l0oooo0P00003o0?ooo`@0oooo
00<000000?ooo`3oool0C`3oool2000000P0oooo001?0?ooo`030000003oool0oooo0?l0oooo0P3o
ool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0
oooo0?l0oooo0P00001D0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo
0?h0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`060?ooo`00C`3oool200000?d0
oooo0P00001F0?ooo`800000203oool004l0oooo00<000000?ooo`3oool0n`3oool00`000000oooo
0?ooo`1G0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?T0oooo0P00
001J0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?P0oooo00<00000
0?ooo`3oool0FP3oool00`000000oooo0?ooo`060?ooo`00C`3oool200000?L0oooo0P00001L0?oo
o`800000203oool004l0oooo00<000000?ooo`3oool0m@3oool00`000000oooo0?ooo`1M0?ooo`03
0000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0?<0oooo0P00001P0?ooo`030000
003oool0oooo00H0oooo000n0?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool02@3o
ool200000?80oooo0P00001Q0?ooo`800000203oool00400oooo00<0okmd0000001d_ol0303oool0
0`000000oooo0?ooo`3`0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool01P3oool0
0400oooo00<0oooO09a80000M;l0303oool200000>l0oooo0P00001T0?ooo`800000203oool00440
oooo00<0gia80018W03Oool02`3oool00`000000oooo0?ooo`3?0?ooo`8000001@3oool2000000/0
oooo0P0000050?ooo`80000000@0oooo0000003oool0oooo0P00000;0?ooo`8000001@3oool20000
01P0oooo0P00000e0?ooo`030000003oool0oooo00H0oooo00110?ooo`030?oOW01804P0W=oo00/0
oooo00<000000?ooo`3oool0cP3oool4000000<0oooo100000090?ooo`@000000`3oool400000080
oooo100000090?ooo`@000000`3oool4000001H0oooo1000000d0?ooo`030000003oool0oooo00H0
oooo00110?ooo`030?oo_`1d0000M;oo00/0oooo0`00003=0?ooo`H0000000<0oooo000000000000
100000070?ooo`H0000000<0oooo0000000000002P0000070?ooo`H0000000<0oooo000000000000
1000000D0?ooo`H00000<@3oool3000000P0oooo00120?ooo`030;md0000M;l0oooo00X0oooo00<0
00000?ooo`3oool0c@3oool6000000030?ooo`000000000000@000001`3oool6000000030?ooo`00
0000000000X000001`3oool6000000030?ooo`000000000000@00000503oool6000003<0oooo00<0
00000?ooo`3oool01P3oool00480oooo00<0gia80018W03Oool02P3oool00`000000oooo0?ooo`3>
0?ooo`@000000`3oool4000000T0oooo100000000`3oool000000?ooo`0400000080oooo10000009
0?ooo`@000000`3oool4000001H0oooo1000000d0?ooo`030000003oool0oooo00H0oooo00120?oo
o`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool0c`3oool2000000D0oooo0P00000;0?oo
o`80000000<0oooo0000003oool00P3oool2000000@0oooo0P00000;0?ooo`8000001@3oool20000
01P0oooo0P00000e0?ooo`030000003oool0oooo00H0oooo00120?ooo`030?oo_`1d0000M;oo00X0
oooo00<000000?ooo`3oool0i03oool2000006l0oooo00<000000?ooo`3oool01P3oool004<0oooo
00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`3S0?ooo`030000003oool0oooo06l0oooo
00<000000?ooo`3oool01P3oool00440oooo00D0oono07@000000000001d0;ooo`090?ooo`030000
003oool0oooo0>40oooo0P00001b0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003o
ool0oooo0>00oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`060?ooo`00C`3oool0
0`000000oooo0?ooo`3N0?ooo`800000M@3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`00
0000oooo0?ooo`3M0?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool01P3oool004l0
oooo00<000000?ooo`3oool0f`3oool2000007P0oooo00<000000?ooo`3oool01P3oool004l0oooo
00<000000?ooo`3oool0fP3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo00H0oooo
001?0?ooo`030000003oool0oooo0=P0oooo0P00001k0?ooo`030000003oool0oooo00H0oooo001?
0?ooo`030000003oool0oooo0;`0oooo0P00000;0?ooo`8000002`3oool2000000@0oooo0P00001g
0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0;/0oooo100000090?oo
o`@000002@3oool00`000000oooo0?ooo`030?ooo`@00000MP3oool00`000000oooo0?ooo`060?oo
o`00C`3oool200000;/0oooo1P0000070?ooo`H000001P3oool2000000D0oooo1P00001d0?ooo`80
0000203oool004l0oooo00<000000?ooo`3oool0^P3oool6000000L0oooo1P0000050?ooo`030000
003oool0oooo00D0oooo1P00001e0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003o
ool0oooo0;/0oooo100000090?ooo`@00000103oool2000000T0oooo1000001f0?ooo`030000003o
ool0oooo00H0oooo001?0?ooo`030000003oool0oooo0;`0oooo0P00000;0?ooo`800000103oool0
0`000000oooo0?ooo`0:0?ooo`800000M`3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`00
0000oooo0?ooo`3>0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool01P3oool004l0
oooo00<000000?ooo`3oool0c03oool2000008L0oooo00<000000?ooo`3oool01P3oool004l0oooo
00<000000?ooo`3oool0b`3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo00H0oooo
001?0?ooo`030000003oool0oooo0<80oooo0P0000050?ooo`800000RP3oool00`000000oooo0?oo
o`060?ooo`00C`3oool00`000000oooo0?ooo`310?ooo`@000000`3oool00`000000oooo0?ooo`2:
0?ooo`030000003oool0oooo00H0oooo001?0?ooo`800000`@3oool6000000030?ooo`000000oooo
08/0oooo0P0000080?ooo`00C`3oool00`000000oooo0?ooo`300?ooo`L00000SP3oool00`000000
oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`310?ooo`@00000T03oool00`000000oooo
0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`320?ooo`800000T@3oool00`000000oooo0?oo
o`060?ooo`00C`3oool00`000000oooo0?ooo`300?ooo`800000T`3oool00`000000oooo0?ooo`06
0?ooo`00C`3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo09<0oooo00<000000?oo
o`3oool01P3oool004l0oooo0P00002n0?ooo`800000U@3oool2000000P0oooo001?0?ooo`030000
003oool0oooo0;`0oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`060?ooo`00C`3o
ool00`000000oooo0?ooo`2j0?ooo`800000V@3oool00`000000oooo0?ooo`060?ooo`00C`3oool0
0`000000oooo0?ooo`2_0?ooo`800000103oool2000000030?ooo`000000000009/0oooo00<00000
0?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0[P3oool400000080oooo1000002M0?oo
o`030000003oool0oooo00H0oooo001?0?ooo`800000[P3oool<000009/0oooo0P0000080?ooo`00
C`3oool00`000000oooo0?ooo`2]0?ooo``00000W03oool00`000000oooo0?ooo`060?ooo`00C`3o
ool00`000000oooo0?ooo`2Q0?ooo`8000002`3oool4000000030?ooo`000000000000<00000W@3o
ool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`2P0?ooo`@000002`3oool4
00000080oooo0P00002N0?ooo`030000003oool0oooo00H0oooo001?0?ooo`800000X03oool60000
00/0oooo00<000000?ooo`3oool0X@3oool2000000P0oooo001?0?ooo`030000003oool0oooo09l0
oooo1P0000090?ooo`800000Y@3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo
0?ooo`2P0?ooo`@000002@3oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo00H0oooo
001?0?ooo`030000003oool0oooo0:40oooo0P0000090?ooo`030000003oool0oooo0:H0oooo00<0
00000?ooo`3oool01P3oool004l0oooo0P00002[0?ooo`800000Z03oool2000000P0oooo001?0?oo
o`030000003oool0oooo0:T0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`060?oo
o`00=@3oool00`3ookl0M000000000020000000304RLg`3oool0oooo00<0oooo00@0_g@000000000
0000M;oo2`3oool200000:P0oooo0P00002[0?ooo`800000203oool003L0oooo00<0okmd0000001d
_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`X0oooo00<000000?ooo`3oool0Y@3oool2
00000:h0oooo00<000000?ooo`3oool01P3oool003L0oooo00<0oooO09a80000M;l0103oool00`3o
oml0W4P004RLg`020?ooo`030?noM00004P0W=oo00T0oooo00<000000?ooo`3oool0Y03oool00`00
0000oooo0?ooo`2^0?ooo`030000003oool0oooo00H0oooo000h0?ooo`030=nLB000B9`0gooo00<0
oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`800000U@3oool20000
00D0oooo0P0000050?ooo`800000103oool200000:X0oooo0P0000080?ooo`00>03oool00`3ogi`0
B01809cOo`030?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol0203oool00`000000
oooo0?ooo`2C0?ooo`@000000`3oool4000000<0oooo00<000000?ooo`3oool00`3oool400000:X0
oooo00<000000?ooo`3oool01P3oool003P0oooo00<0oono07@0001d_ol0103oool00`2oM00004RL
0=ooo`020?ooo`030?oOW0180000B9cO00P0oooo0`00002B0?ooo`H0000000<0oooo000000000000
1P0000050?ooo`H00000Y`3oool3000000P0oooo000J0?ooo`030;md0000B9`0gooo01`0oooo00<0
_g@0001d_`3oool00`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00<0
00000?ooo`3oool0TP3oool6000000030?ooo`000000000000@000001`3oool600000:T0oooo00<0
00000?ooo`3oool01P3oool001X0oooo00<0omnL04Qd_`3oool0703oool00`3OW4P004RL0=ooo`03
0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01`3oool00`000000oooo0?ooo`2C
0?ooo`@000000`3oool4000000T0oooo1000002Z0?ooo`030000003oool0oooo00H0oooo000?0?oo
o`030?oo_`1dM;l0oooo00L0oooo00<0okmd0018W03Oool07@3oool00`3ogi`0B01809cOo`030?oo
o`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01`3oool00`000000oooo0?ooo`2D
0?ooo`800000103oool3000000/0oooo0P00002[0?ooo`030000003oool0oooo00H0oooo000?0?oo
o`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3oool07P3oool00`3ookl0M00007Boo`04
0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`2I
0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool01P3oool00100oooo00<0omnL04P0
000000001@0000000`18W=l0oooo0?ooo`0O0?ooo`030;md0000M;l0oooo00@0oooo00H0gia80000
M02oool0oooO09a80018W=l80?ooo`030000003oool0oooo09L0oooo0P00002l0?ooo`030000003o
ool0oooo00H0oooo000E0?ooo`030?noM00004P0W=oo0200oooo00D0oono07@000000000001d0;oo
o`050?ooo`040=nLB00000000000001d_`T0oooo00<000000?ooo`3oool0UP3oool00`000000oooo
0?ooo`2l0?ooo`030000003oool0oooo00H0oooo000D0?ooo`030;md0000M;l0oooo03P0oooo00<0
00000?ooo`3oool0U03oool200000;l0oooo00<000000?ooo`3oool01P3oool000l0oooo00H0omnL
04P0001d_ol0okmd0000001d_olj0?ooo`030000003oool0oooo0980oooo0P0000310?ooo`030000
003oool0oooo00H0oooo000?0?ooo`040?oOW01800000000001d_c`0oooo00<000000?ooo`3oool0
SP3oool2000000030?ooo`000000oooo0<80oooo00<000000?ooo`3oool01P3oool004l0oooo00<0
00000?ooo`3oool0S@3oool400000<@0oooo00<000000?ooo`3oool01P3oool004l0oooo00<00000
0?ooo`3oool0S03oool600000<<0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?oo
o`3oool0S03oool600000<<0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?ooo`3o
ool0R`3oool00`000000oooo0000000300000<@0oooo00<000000?ooo`3oool01P3oool004l0oooo
00<000000?ooo`3oool0Q`3oool4000000<0oooo0P0000350?ooo`030000003oool0oooo00H0oooo
001?0?ooo`030000003oool0oooo08H0oooo1000003;0?ooo`030000003oool0oooo00H0oooo001?
0?ooo`800000QP3oool600000<T0oooo0P0000080?ooo`00C`3oool00`000000oooo0?ooo`240?oo
o`L00000bP3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`230?ooo`03
0000003oool0oooo00@00000b`3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo
0?ooo`210?ooo`800000103oool200000<`0oooo00<000000?ooo`3oool01P3oool004l0oooo00<0
00000?ooo`3oool0NP3oool2000000@0oooo0`00003B0?ooo`030000003oool0oooo00H0oooo001?
0?ooo`030000003oool0oooo07T0oooo100000000`3oool000000000000400000=40oooo00<00000
0?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0N03oool6000000030?ooo`0000000000
00@00000d03oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`1h0?ooo`H0
000000<0oooo0000000000001000003@0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000
003oool0oooo07T0oooo100000030?ooo`@00000d@3oool00`000000oooo0?ooo`060?ooo`00C`3o
ool2000007X0oooo0`0000050?ooo`800000d@3oool2000000P0oooo001?0?ooo`030000003oool0
oooo07L0oooo0P00003L0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo
06d0oooo0P0000040?ooo`80000000<0oooo0000003oool0g@3oool00`000000oooo0?ooo`060?oo
o`00C`3oool00`000000oooo0?ooo`1/0?ooo`@000000P3oool400000=l0oooo00<000000?ooo`3o
ool01P3oool004l0oooo00<000000?ooo`3oool0J`3oool<00000=h0oooo00<000000?ooo`3oool0
1P3oool004l0oooo00<000000?ooo`3oool0J`3oool<00000=h0oooo00<000000?ooo`3oool01P3o
ool004l0oooo0P00001]0?ooo`D0000000<0oooo0000000000000P00003N0?ooo`800000203oool0
04l0oooo00<000000?ooo`3oool0K@3oool3000000<0oooo0P00003P0?ooo`030000003oool0oooo
00H0oooo001?0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0i@3oool00`000000
oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`1[0?ooo`800000j03oool00`000000oooo
0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`1V0?ooo`80000000<0oooo000000000000jP3o
ool00`000000oooo0?ooo`060?ooo`00C`3oool2000006H0oooo1000003[0?ooo`800000203oool0
04l0oooo00<000000?ooo`3oool0I03oool600000>/0oooo00<000000?ooo`3oool01P3oool004l0
oooo00<000000?ooo`3oool0I03oool600000>/0oooo00<000000?ooo`3oool01P3oool004l0oooo
00<000000?ooo`3oool0H`3oool600000>`0oooo00<000000?ooo`3oool01P3oool004l0oooo0P00
001Q0?ooo`<000000`3oool200000>`0oooo0P0000080?ooo`00C`3oool00`000000oooo0?ooo`1O
0?ooo`@00000lP3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`1N0?oo
o`H00000l@3oool00`000000oooo0?ooo`060?ooo`00C`3oool2000005l0oooo1P00003`0?ooo`80
0000203oool004l0oooo00<000000?ooo`3oool0F@3oool2000000040?ooo`00000000000?ooo`@0
0000lP3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`1H0?ooo`@00000
103oool200000?<0oooo00<000000?ooo`3oool01P3oool002`0oooo00<0oono07@0000000000P00
00000`18W=l0oooo0?ooo`030?ooo`040;md00000000000007Boo`D0oooo00@0_g@0000000000000
M;oo2`3oool2000005P0oooo1P00003g0?ooo`800000203oool002h0oooo00<0okmd0000001d_ol0
1@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`<0oooo00H0gia8001d_`3oool0omnL04P00018
W=l:0?ooo`030000003oool0oooo05L0oooo1P00003h0?ooo`030000003oool0oooo00H0oooo000^
0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool01`3o_g@0001809cOo`3o
ool0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`090?ooo`030000003oool0oooo05<0oooo
0P0000020?ooo`D00000n@3oool00`000000oooo0?ooo`060?ooo`00;`3oool00`3OW4P004RL0=oo
o`030?ooo`030?oog`2LB000B9cO0080oooo00L0oooO09a80000B9`0gooo0?oog`2LB000B9cO0080
oooo00@0oooO09a80000B9`0gooo203oool2000005<0oooo1@0000020?ooo`800000n@3oool20000
00P0oooo000_0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool01P3OW4P0
001809cOo`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol0203oool00`000000oooo0?ooo`1A
0?ooo`H00000oP3oool00`000000oooo0?ooo`060?ooo`00;`3oool00`3ookl0M00007Boo`040?oo
o`030;md0000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0;md0000B9`0gooo0080oooo00<0
omnL04P00018W=l0203oool300000540oooo1P00003l0?ooo`<00000203oool00300oooo00<0_g@0
001d_`3oool00`3oool00`3OW4P0001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P00018
09cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00<000000?ooo`3oool0D@3oool500000?l0oooo
00<000000?ooo`3oool01P3oool00300oooo00<0gia80018W03Oool00`3oool00`3ogi`0B00004RL
g`030?ooo`060;md0000B9`0gooo0?oOW0180000B9cO0`3oool00`2oM00004RL0=ooo`070?ooo`03
0000003oool0oooo04l0oooo0P0000020?ooo`800000o`3oool10?ooo`030000003oool0oooo00H0
oooo000`0?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0gooo0P3oool01`3OW4P0
04RL0=ooo`3ooml0W4P00018W03Oool00P3oool00`3OW4P004RL0=ooo`070?ooo`030000003oool0
oooo04`0oooo0`00003o0?ooo`L0oooo00<000000?ooo`3oool01P3oool00300oooo00<0oono07@0
001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`070=nLB000M;l0oooo0?ooo`3ogi`0B00007Bo
o`020?ooo`030=nLB000M;l0oooo00L0oooo00<000000?ooo`3oool0B`3oool400000?l0oooo1`3o
ool00`000000oooo0?ooo`060?ooo`00<@3oool00`2oM00007Bo0?ooo`040?ooo`060=nLB00007@0
_ooo0?oog`2LB000B9cO0`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<000000?oo
o`3oool0BP3oool600000?l0oooo1P3oool00`000000oooo0?ooo`060?ooo`00;`3oool01@3ookl0
M0000000000007@0_ooo00D0oooo00@0gia800000000000007Bo1@3oool0103OW4P0000000000000
M;l90?ooo`030000003oool0oooo04T0oooo1`00003o0?ooo`H0oooo00<000000?ooo`3oool01P3o
ool004l0oooo00<000000?ooo`3oool0AP3oool300000080oooo1000003o0?ooo`L0oooo00<00000
0?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0A@3oool4000000<0oooo0P00003o0?oo
o`P0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool0A03oool600000?l0
oooo303oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`0o0?ooo`800000
0P3oool700000?l0oooo303oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?oo
o`0n0?ooo`D000000P3oool400000?l0oooo3@3oool00`000000oooo0?ooo`060?ooo`00C`3oool0
0`000000oooo0?ooo`0m0?ooo`H000000`3oool200000?l0oooo3P3oool00`000000oooo0?ooo`06
0?ooo`00C`3oool00`000000oooo0?ooo`0m0?ooo`H00000o`3ooolC0?ooo`030000003oool0oooo
00H0oooo001?0?ooo`030000003oool0oooo03P0oooo0P0000020?ooo`H00000o`3ooolD0?ooo`03
0000003oool0oooo00H0oooo001?0?ooo`800000>03oool5000000<0oooo0P00003o0?oooa@0oooo
0P0000080?ooo`00C`3oool00`000000oooo0?ooo`0f0?ooo`H00000o`3ooolJ0?ooo`030000003o
ool0oooo00H0oooo001?0?ooo`030000003oool0oooo0380oooo0P0000020?ooo`H00000o`3ooolJ
0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0340oooo100000000`3o
ool000000000000300000?l0oooo6`3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000
oooo0?ooo`0`0?ooo`H000000P3oool200000?l0oooo703oool00`000000oooo0?ooo`060?ooo`00
C`3oool00`000000oooo0?ooo`0[0?ooo`8000000`3oool600000?l0oooo803oool00`000000oooo
0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`0Z0?ooo`@000000`3oool400000?l0oooo8@3o
ool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`0Y0?ooo`H000000P3oool3
00000?l0oooo8P3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`0Y0?oo
o`P00000o`3ooolU0?ooo`030000003oool0oooo00H0oooo001?0?ooo`800000:`3oool500000?l0
oooo9P3oool2000000P0oooo001?0?ooo`030000003oool0oooo02/0oooo0`00003o0?ooobP0oooo
00<000000?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool09@3oool2000000@0oooo00<0
00000?ooo`3oool0o`3ooolX0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0
oooo02@0oooo100000000`3oool000000000003o0?ooob/0oooo00<000000?ooo`3oool01P3oool0
04l0oooo00<000000?ooo`3oool08`3oool600000?l0oooo;@3oool00`000000oooo0?ooo`060?oo
o`00C`3oool00`000000oooo0?ooo`0S0?ooo`H00000o`3oool]0?ooo`030000003oool0oooo00H0
oooo001?0?ooo`030000003oool0oooo02@0oooo1000003o0?ooobh0oooo00<000000?ooo`3oool0
1P3oool004l0oooo0P00000U0?ooo`<00000o`3oool^0?ooo`800000203oool004l0oooo00<00000
0?ooo`3oool08P3oool200000?l0oooo<P3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`00
0000oooo0?ooo`0N0?ooo`80000000<0oooo0000003oool0o`3ooolc0?ooo`030000003oool0oooo
00H0oooo001?0?ooo`030000003oool0oooo01d0oooo1000003o0?ooocD0oooo00<000000?ooo`3o
ool01P3oool004l0oooo0P00000M0?ooo`H00000o`3ooolc0?ooo`800000203oool004l0oooo00<0
00000?ooo`3oool0603oool200000080oooo1P00003o0?oooc@0oooo00<000000?ooo`3oool01P3o
ool004l0oooo00<000000?ooo`3oool05`3oool5000000030?ooo`000000000000800000o`3ooole
0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo01H0oooo1P0000020?oo
o`800000o`3ooolf0?ooo`030000003oool0oooo00H0oooo001?0?ooo`8000005`3oool600000?l0
oooo>@3oool2000000P0oooo001?0?ooo`030000003oool0oooo0140oooo0P0000020?ooo`H00000
o`3ooolk0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0oooo0100oooo1@00
00030?ooo`800000o`3oooll0?ooo`030000003oool0oooo00H0oooo001?0?ooo`030000003oool0
oooo00l0oooo1P00003o0?oood40oooo00<000000?ooo`3oool01P3oool004l0oooo0P00000@0?oo
o`H00000o`3ooom00?ooo`800000203oool004l0oooo00<000000?ooo`3oool03`3oool500000?l0
oooo@P3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`0;0?ooo`@00000
0P3oool200000?l0oooo@`3oool00`000000oooo0?ooo`060?ooo`008`3oool00`3ookl0M0000000
00020000000304RLg`3oool0oooo00<0oooo00@0_g@0000000000000M;oo1@3oool0102oM0000000
0000001d_ol50?ooo`040;md00000000000007Boo`/0oooo0P00000;0?ooo`@00000o`3ooom70?oo
o`800000203oool002D0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B000
04RLg`<0oooo00H0gia8001d_`3oool0omnL04P00018W=l30?ooo`060=nLB000M;l0oooo0?oOW018
0000B9cO2P3oool00`000000oooo0?ooo`090?ooo`H00000o`3ooom70?ooo`030000003oool0oooo
00H0oooo000U0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool01`3o_g@0
001809cOo`3oool0oooO09a80018W=l00P3oool01`3o_g@0001809cOo`3oool0oooO09a80018W=l0
0P3oool00`3o_g@0001809cOo`090?ooo`8000002P3oool600000?l0ooooAP3oool2000000P0oooo
000V0?ooo`030=nLB000B9`0gooo00<0oooo00<0oooO09a80018W=l00P3oool01`3ooml0W4P00018
W03Oool0oooO09a80018W=l00P3oool01`3ooml0W4P00018W03Oool0oooO09a80018W=l00P3oool0
103ooml0W4P00018W03Oool80?ooo`030000003oool0oooo00@0oooo0P0000020?ooo`030000003o
ool0000000<00000o`3ooom80?ooo`030000003oool0oooo00H0oooo000V0?ooo`030?oOW01804P0
W=oo00<0oooo00<0oooO09a80000M;l00`3oool01P3OW4P0001809cOo`3ooml0W4P0001d_`<0oooo
00H0gia80000B02Lgol0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00P0oooo00<000000?oo
o`3oool00`3oool5000000<0oooo0P00003o0?ooodT0oooo00<000000?ooo`3oool01P3oool002H0
oooo00<0oono07@0001d_ol0103oool00`2oM00004RL0=ooo`020?ooo`070?oOW0180000B9cO0?oo
o`2oM00004RL0=ooo`020?ooo`070?oOW0180000B9cO0?ooo`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00P0oooo0`0000020?ooo`H00000o`3ooom<0?ooo`<00000203oool002L0oooo00<0
_g@0001d_`3oool00`3oool00`3OW4P0001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P0
001809cOo`020?ooo`070?oog`2LB00007Bo0?ooo`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00P0oooo00<000000?ooo`3oool00P3oool600000?l0ooooCP3oool00`000000oooo0?ooo`06
0?ooo`009`3oool00`3OW4P004RL0=ooo`030?ooo`030?oOW0180000B9cO00<0oooo00H0_g@00018
W03Oool0omnL04P00018W=l30?ooo`060;md0000B9`0gooo0?oOW0180000B9cO0`3oool00`2oM000
04RL0=ooo`070?ooo`030000003oool0oooo00<0oooo1000003o0?ooodl0oooo00<000000?ooo`3o
ool01P3oool002L0oooo00<0omnL04P0B02Lgol00`3oool0103ooml0W4P00018W03Oool20?ooo`07
0=nLB000B9`0gooo0?oog`2LB00004RL0=ooo`020?ooo`070=nLB000B9`0gooo0?oog`2LB00004RL
0=ooo`020?ooo`030=nLB000B9`0gooo00L0oooo00<000000?ooo`3oool0103oool200000?l0oooo
D03oool00`000000oooo0?ooo`060?ooo`009`3oool00`3ookl0M00007Boo`040?ooo`030?oOW018
0000M;oo0080oooo00L0gia8001d_`3oool0oooo0?oOW0180000M;oo0080oooo00L0gia8001d_`3o
ool0oooo0?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`000000oooo0?ooo`3o
0?oooeH0oooo00<000000?ooo`3oool01P3oool002P0oooo00<0_g@0001d_`3oool0103oool01P3O
W4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00H0gia80000M02oool0oooO09a80018W=l30?ooo`06
0=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`000000oooo0?ooo`3o0?oooeH0oooo00<00000
0?ooo`3oool01P3oool002H0oooo00D0oono07@000000000001d0;ooo`050?ooo`040=nLB0000000
0000001d_`D0oooo00@0gia800000000000007Bo1@3oool0103OW4P0000000000000M;l90?ooo`03
0000003oool0oooo0?l0ooooEP3oool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo
0?ooo`3o0?oooeH0oooo00<000000?ooo`3oool01P3oool004l0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0?`3o
ool00`000000oooo0?ooo`0n0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`060?ooo`00C`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00H0oooo001?0?ooool00000FP0000080?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.26637, -2.01349, \
0.0045726, 0.0295945}}]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Including uncertainties in the independent variable", "Subsection"],

Cell[TextData[{
  "\tOften measurement of the independent variable is also subject to \
nonnegligible random errors, but most standard software for linear \
least-squares does not accommodate uncertainties in both variables \
simultaneously.  For example, the ",
  StyleBox["LinearRegression", "InlineInput",
    FontWeight->"Bold"],
  " package in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides many sophisticated options, but the use of uncertainties in more \
than one variable is not among them.  Therefore, we need to develop our own \
fitting procedure.  "
}], "Text"],

Cell[TextData[{
  "\tSuppose that our data set takes the form ",
  Cell[BoxData[
      \(TraditionalForm\`{{\(x\_\(\(1\)\(,\)\)\) \
\[Delta]\[VeryThinSpace]x\_1, 
          y\_1, \[Delta]\[VeryThinSpace]y\_1}, \[CenterEllipsis], {x\_N, \
\[Delta]\[VeryThinSpace]x\_N, y\_N, \[Delta]\[VeryThinSpace]y\_N}}\)]],
  " while the model is given by ",
  Cell[BoxData[
      \(TraditionalForm\`y[
          x]\  = \ \[Sum]\+\(\[Mu] = 1\)\%m\( a\_\[Mu]\) f\_\[Mu][x]\)]],
  ".  The measurements for both variables "
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`x\_i = 
      x\& . \_i + \[Delta]\[ThinSpace]x\_i\ , \ \ \ y\_i = 
      y\& . \_i + \[Delta]\[ThinSpace]y\_i\)], "DisplayFormula"],

Cell[TextData[{
  "differ from the true values, indicated by dots, by random fluctuations \
based upon normal distributions with variances ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Delta]\[ThinSpace]x\_i, \[Delta]\[ThinSpace]y\_i}\
\)]],
  ".  If the model parameters were known, the expected ",
  Cell[BoxData[
      \(TraditionalForm\`y\_i\)]],
  " could be computed from the expected ",
  Cell[BoxData[
      \(TraditionalForm\`x\_i\)]],
  " according to"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`y\& . \_i\  = \ 
      y[x\& . \_i]\  \[TildeTilde] \ 
        y[x\_i]\  - \ \(\((\[PartialD]y\/\[PartialD]x)\)\_\(x = 
                  x\_i\)\) \[Delta]\[ThinSpace]x\_i\)], "DisplayFormula"],

Cell[TextData[{
  "where the final term is the first-order change due to the fluctuation in \
",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  ".  The deviation in ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " then has two contributions"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`y\_i - 
        y\_i[x\_i]\  \[TildeTilde] \ \[Delta]\[ThinSpace]y\_i\  - \ \(\((\
\[PartialD]y\/\[PartialD]x)\)\_\(x = 
                x\_i\)\) \[Delta]\[ThinSpace]x\_i\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[ThinSpace]x\_i\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[ThinSpace]y\_i\)]],
  " are both random variables.  Therefore, the variance for a measurement is \
given by "
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma]\_i\^2 \[TildeTilde] \ \
\((\[Delta]\[ThinSpace]y\_i\ )\)\^2 + \ \((\(\((\[PartialD]y\/\[PartialD]x)\)\
\_\(x = x\_i\)\) \[Delta]\[ThinSpace]x\_i)\)\^2\)], "DisplayFormula"],

Cell[TextData[{
  "Application of this result requires iteration.  First we fit without using \
the information about ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[VeryThinSpace]x\_i\)]],
  " and obtain a set of parameters ",
  Cell[BoxData[
      \(TraditionalForm\`{a\_\[Mu]\%\((1)\)}\)]],
  ".  These parameters are used to compute ",
  Cell[BoxData[
      \(TraditionalForm\`\[PartialD]y/\[PartialD]x\)]],
  " and the contributions to ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\_i\)]],
  " from ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[ThinSpace]x\_i\)]],
  ".  Then we refit and obtain new parameters ",
  Cell[BoxData[
      \(TraditionalForm\`{a\_\[Mu]\%\((2)\)}\)]],
  ".  This set is then used to re-evaluate ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Sigma]\_i}\)]],
  " and the procedure is repeated until the changes in all of the parameters \
become acceptably small or until our patience is exhausted and a maximum \
number of iterations is reached."
}], "Text"],

Cell[TextData[{
  "\tThis iterative procedure is implemented in the ",
  StyleBox["linfit2", "InlineInput",
    FontWeight->"Bold"],
  " program listed below.  The fits are made with ",
  StyleBox["linfit", "InlineInput",
    FontWeight->"Bold"],
  " and the rest of the code adds the uncertainties due to the two variables \
and tests for convergence."
}], "Text"],

Cell[BoxData[{
    RowBox[{
      RowBox[{\(linfit2::usage\), "=", 
        "\"\<linfit2[data,f,var] performs linear least-squares fitting to \
data with one independent and one dependent variable with uncertainties in \
both.  The \!\(\*
StyleBox[\"data\",\nFontSlant->\"Italic\"]\) array should be a list of \!\(\*
StyleBox[\"{\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"x\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"dx\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"y\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"dy\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"}\",\nFontSlant->\"Italic\"]\) for each measurement.  The list of \
basis functions is \!\(\*
StyleBox[\"f\",\nFontSlant->\"Italic\"]\) and \!\(\*
StyleBox[\"var\",\nFontSlant->\"Italic\"]\) is the symbol used in \!\(\*
StyleBox[\"f\",\nFontSlant->\"Italic\"]\) for the independent \
variable.\>\""}], 
      ";"}], "\[IndentingNewLine]", \(linfit2::BadDim = "\<Bad dimensions in \
data array: `1`\>";\), "\[IndentingNewLine]", 
    RowBox[{\(linfit2::conv = "\<Warning: failed to converge.  Maximum \
relative change = `1`.\>";\), "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
  
    RowBox[{\(Options[linfit2] = {MaxNITS \[Rule] 10, Conv \[Rule] 0.001};\), 
      "\[IndentingNewLine]"}], "\[IndentingNewLine]", \(linfit2[data_, 
        f_?VectorQ, var_, opts___Rule] := 
      Module[{x, dx, y, dy, \[Sigma], ndata, nparam, ndim, i, j, params, 
          slopes, fp, errors, yfit, redchi, oldparams, nits, mnits, change, 
          conv}, \[IndentingNewLine]\[IndentingNewLine] (*\ Extract\ x, dx, 
          y, dy\ vectors\ from\ data\ \(\(array\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]{ndata, ndim} = 
          Dimensions[data]; \[IndentingNewLine]If[
          ndata \[LessEqual] Length[f] || 
            ndim \[NotEqual] 4, \((Message[linfit2::BadDim, {ndata, ndim}]; 
            Return[])\)]; \[IndentingNewLine]{x, dx, y, dy} = 
          Transpose[data]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Derivative\ of\ basis\ \(\(functions\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]nparam = 
          Length[f]; \[IndentingNewLine]slopes = 
          Table[0, {ndata}]; \[IndentingNewLine]fp = \[PartialD]\_var 
            f; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Loop\ using\ weights\ that\ include\ slope\ of\ fit\ \(\(function\)\
\(.\)\)\ *) \[IndentingNewLine]\[IndentingNewLine]{mnits, 
            conv} = \({MaxNITS, Conv} /. {opts}\) /. 
            Options[linfit2]; \[IndentingNewLine]oldparams = 
          Table[0, {i, 
              nparam}]; \[IndentingNewLine]\[IndentingNewLine]For[\((nits = 
              0; change = 1)\), \((nits < mnits && 
              change > 
                conv)\), \(nits++\), \[IndentingNewLine]\[Sigma] = \@\(dy\^2 \
+ \((slopes*dx)\)\^2\); \[IndentingNewLine]{params, 
              errors} = {Parameters, Uncertainties} /. 
              linfit[Transpose[{x, y, \[Sigma]}], f, 
                var]; \[IndentingNewLine]yfit = 
            params . f /. var \[Rule] x; \[IndentingNewLine]slopes = 
            params . fp /. var \[Rule] x; \[IndentingNewLine]change = 
            Max[Abs[\((params - oldparams)\)/
                  params]]; \[IndentingNewLine]oldparams = 
            params]; \[IndentingNewLine]If\ [mnits > 1 && change > conv, 
          Message[linfit2::conv, 
            NumberForm[change, 
              2]]]; \[IndentingNewLine]redchi = \[Sigma]\^\(-2\) . \((y - \
yfit)\)\^2/\((ndata - nparam)\); \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Return\ results\ as\ replacement\ \(\(rules\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]{Parameters \[Rule] params, 
          Uncertainties \[Rule] errors, ReducedChiSquare \[Rule] redchi, 
          NITS \[Rule] nits}]\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?linfit2\)\)], "Input"],

Cell[BoxData["\<\"linfit2[data,f,var] performs linear least-squares fitting \
to data with one independent and one dependent variable with uncertainties in \
both.  The \!\(\* StyleBox[\\\"data\\\",\\nFontSlant->\\\"Italic\\\"]\) array \
should be a list of \!\(\* StyleBox[\\\"{\\\",\\nFontSlant->\\\"Italic\\\"]\)\
\!\(\* StyleBox[\\\"x\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"dx\\\",\
\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"y\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"dy\\\",\
\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Italic\\\"]\) for each measurement.  The \
list of basis functions is \!\(\* StyleBox[\\\"f\\\",\\nFontSlant->\\\"Italic\
\\\"]\) and \!\(\* StyleBox[\\\"var\\\",\\nFontSlant->\\\"Italic\\\"]\) is \
the symbol used in \!\(\* StyleBox[\\\"f\\\",\\nFontSlant->\\\"Italic\\\"]\) \
for the independent variable.\"\>"], "Print",
  CellTags->"Info3351438212-4114851"]
}, Closed]],

Cell[CellGroupData[{

Cell["\<\
example: fit Legendre coefficients to data with error bars in both \
variables\
\>", "Subsubsection"],

Cell["\tConsider a model of the form", "Text"],

Cell[BoxData[
    \(TraditionalForm\`y[
        x]\  = \ \[Sum]\+\[Mu]\( a\_\[Mu]\) P\_\[Mu][x]\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`P\_\[Mu]\)]],
  " is the Legendre polynomial with degree ",
  Cell[BoxData[
      \(TraditionalForm\`\[Mu]\)]],
  ".  The model is linear with respect to the coefficients ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Mu]\)]],
  ", so data can be fit using the method of linear least squares.  Below we \
construct and fit a sample data set with Gaussian fluctuations in both ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  ".  Note that we assumed that the relative uncertainty in ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " is uniform."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[z];\)\), "\[IndentingNewLine]", 
    \(\(p = {3.02, 7.3, \(-9.2\)};\)\), "\[IndentingNewLine]", 
    \(\(f = LegendreP[{0, 1, 2}, z];\)\), "\[IndentingNewLine]", 
    \(\(x = Range[\(-0.95\), 0.95, 0.05];\)\), "\[IndentingNewLine]", 
    \(\(dx = Table[0.05, {Length[x]}];\)\), "\[IndentingNewLine]", 
    \(\(y = p . f /. z \[Rule] x;\)\), "\[IndentingNewLine]", 
    \(\(dy = 0.1\ Abs[y];\)\), "\[IndentingNewLine]", 
    \(\(x = 
        x + \((\(Random[NormalDistribution[0, #]] &\) /@ 
              dx)\);\)\), "\[IndentingNewLine]", 
    \(\(y = 
        y + \((\(Random[NormalDistribution[0, #]] &\) /@ 
              dy)\);\)\), "\[IndentingNewLine]", 
    \(\(data = Transpose[{x, dx, y, dy}];\)\), "\[IndentingNewLine]", 
    \(results = 
      linfit2[data, f, z, MaxNITS \[Rule] 10, 
        Conv \[Rule] 10\^\(-4\)]\)}], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {2.9434870458796087`, 
          6.941009559385762`, \(-9.196613575583598`\)}, 
      Uncertainties \[Rule] {0.17150096797935352`, 0.3312836379302557`, 
          0.43538944157814613`}, 
      ReducedChiSquare \[Rule] 0.8841263774406405`, 
      NITS \[Rule] 4}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        MultipleListPlot[
          Evaluate[
            Transpose[{x, dx, y, dy}] /. {xx_, dxx_, yy_, 
                  dyy_} \[Rule] {{xx, yy}, ErrorBar[dxx, dyy]}], 
          PlotRange \[Rule] All, 
          PlotStyle \[Rule] PointSize[0.02]], \[IndentingNewLine]Plot[
          Evaluate[Parameters . f /. results], {z, \(-1\), 1}], 
        Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.507963 0.468228 0.352085 0.0255697 [
[.03974 -0.0125 -9.34375 -15.6875 ]
[.03974 -0.0125 9.34375 0 ]
[.27385 -0.0125 -13.8438 -15.6875 ]
[.27385 -0.0125 13.8438 0 ]
[.50796 -0.0125 -5 -15.6875 ]
[.50796 -0.0125 5 0 ]
[.74208 -0.0125 -9.5 -15.6875 ]
[.74208 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .09639 -24.6875 -7.84375 ]
[-0.0125 .09639 0 7.84375 ]
[-0.0125 .22424 -18.6875 -7.84375 ]
[-0.0125 .22424 0 7.84375 ]
[-0.0125 .35208 -10 -7.84375 ]
[-0.0125 .35208 0 7.84375 ]
[-0.0125 .47993 -10 -7.84375 ]
[-0.0125 .47993 0 7.84375 ]
[-0.0125 .60778 -16 -7.84375 ]
[-0.0125 .60778 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -40.375 -5.09375 ]
[-0.01875 .30902 -24.6875 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03974 0 m
.03974 .00625 L
s
gsave
.03974 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.27385 0 m
.27385 .00625 L
s
gsave
.27385 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.50796 0 m
.50796 .00625 L
s
gsave
.50796 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.74208 0 m
.74208 .00625 L
s
gsave
.74208 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.08656 0 m
.08656 .00375 L
s
.13338 0 m
.13338 .00375 L
s
.1802 0 m
.1802 .00375 L
s
.22703 0 m
.22703 .00375 L
s
.32067 0 m
.32067 .00375 L
s
.36749 0 m
.36749 .00375 L
s
.41432 0 m
.41432 .00375 L
s
.46114 0 m
.46114 .00375 L
s
.55479 0 m
.55479 .00375 L
s
.60161 0 m
.60161 .00375 L
s
.64843 0 m
.64843 .00375 L
s
.69525 0 m
.69525 .00375 L
s
.7889 0 m
.7889 .00375 L
s
.83572 0 m
.83572 .00375 L
s
.88255 0 m
.88255 .00375 L
s
.92937 0 m
.92937 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .09639 m
.00625 .09639 L
s
gsave
-0.0125 .09639 -85.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .22424 m
.00625 .22424 L
s
gsave
-0.0125 .22424 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .35208 m
.00625 .35208 L
s
gsave
-0.0125 .35208 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .47993 m
.00625 .47993 L
s
gsave
-0.0125 .47993 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .60778 m
.00625 .60778 L
s
gsave
-0.0125 .60778 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .12196 m
.00375 .12196 L
s
0 .14753 m
.00375 .14753 L
s
0 .1731 m
.00375 .1731 L
s
0 .19867 m
.00375 .19867 L
s
0 .24981 m
.00375 .24981 L
s
0 .27538 m
.00375 .27538 L
s
0 .30095 m
.00375 .30095 L
s
0 .32651 m
.00375 .32651 L
s
0 .37765 m
.00375 .37765 L
s
0 .40322 m
.00375 .40322 L
s
0 .42879 m
.00375 .42879 L
s
0 .45436 m
.00375 .45436 L
s
0 .5055 m
.00375 .5055 L
s
0 .53107 m
.00375 .53107 L
s
0 .55664 m
.00375 .55664 L
s
0 .58221 m
.00375 .58221 L
s
0 .07082 m
.00375 .07082 L
s
0 .04525 m
.00375 .04525 L
s
0 .01968 m
.00375 .01968 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -98.625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.03974 .61178 m
.03974 .61803 L
s
.27385 .61178 m
.27385 .61803 L
s
.50796 .61178 m
.50796 .61803 L
s
.74208 .61178 m
.74208 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.08656 .61428 m
.08656 .61803 L
s
.13338 .61428 m
.13338 .61803 L
s
.1802 .61428 m
.1802 .61803 L
s
.22703 .61428 m
.22703 .61803 L
s
.32067 .61428 m
.32067 .61803 L
s
.36749 .61428 m
.36749 .61803 L
s
.41432 .61428 m
.41432 .61803 L
s
.46114 .61428 m
.46114 .61803 L
s
.55479 .61428 m
.55479 .61803 L
s
.60161 .61428 m
.60161 .61803 L
s
.64843 .61428 m
.64843 .61803 L
s
.69525 .61428 m
.69525 .61803 L
s
.7889 .61428 m
.7889 .61803 L
s
.83572 .61428 m
.83572 .61803 L
s
.88255 .61428 m
.88255 .61803 L
s
.92937 .61428 m
.92937 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .09639 m
1 .09639 L
s
.99375 .22424 m
1 .22424 L
s
.99375 .35208 m
1 .35208 L
s
.99375 .47993 m
1 .47993 L
s
.99375 .60778 m
1 .60778 L
s
.125 Mabswid
.99625 .12196 m
1 .12196 L
s
.99625 .14753 m
1 .14753 L
s
.99625 .1731 m
1 .1731 L
s
.99625 .19867 m
1 .19867 L
s
.99625 .24981 m
1 .24981 L
s
.99625 .27538 m
1 .27538 L
s
.99625 .30095 m
1 .30095 L
s
.99625 .32651 m
1 .32651 L
s
.99625 .37765 m
1 .37765 L
s
.99625 .40322 m
1 .40322 L
s
.99625 .42879 m
1 .42879 L
s
.99625 .45436 m
1 .45436 L
s
.99625 .5055 m
1 .5055 L
s
.99625 .53107 m
1 .53107 L
s
.99625 .55664 m
1 .55664 L
s
.99625 .58221 m
1 .58221 L
s
.99625 .07082 m
1 .07082 L
s
.99625 .04525 m
1 .04525 L
s
.99625 .01968 m
1 .01968 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .35208 m
1 .35208 L
s
.50796 0 m
.50796 .61803 L
s
.001 w
.07063 .06467 m
.02381 .06467 L
s
.07063 .06467 0 1.5 Mabsadd m
.07063 .06467 0 -1.5 Mabsadd L s
.02381 .06467 0 1.5 Mabsadd m
.02381 .06467 0 -1.5 Mabsadd L s
.04722 .09476 m
.04722 .03457 L
s
.04722 .09476 1.5 0 Mabsadd m
.04722 .09476 -1.5 0 Mabsadd L s
.04722 .03457 1.5 0 Mabsadd m
.04722 .03457 -1.5 0 Mabsadd L s
.11329 .09901 m
.06647 .09901 L
s
.11329 .09901 0 1.5 Mabsadd m
.11329 .09901 0 -1.5 Mabsadd L s
.06647 .09901 0 1.5 Mabsadd m
.06647 .09901 0 -1.5 Mabsadd L s
.08988 .12491 m
.08988 .07311 L
s
.08988 .12491 1.5 0 Mabsadd m
.08988 .12491 -1.5 0 Mabsadd L s
.08988 .07311 1.5 0 Mabsadd m
.08988 .07311 -1.5 0 Mabsadd L s
.16258 .13879 m
.11575 .13879 L
s
.16258 .13879 0 1.5 Mabsadd m
.16258 .13879 0 -1.5 Mabsadd L s
.11575 .13879 0 1.5 Mabsadd m
.11575 .13879 0 -1.5 Mabsadd L s
.13917 .16067 m
.13917 .11692 L
s
.13917 .16067 1.5 0 Mabsadd m
.13917 .16067 -1.5 0 Mabsadd L s
.13917 .11692 1.5 0 Mabsadd m
.13917 .11692 -1.5 0 Mabsadd L s
.15122 .20929 m
.1044 .20929 L
s
.15122 .20929 0 1.5 Mabsadd m
.15122 .20929 0 -1.5 Mabsadd L s
.1044 .20929 0 1.5 Mabsadd m
.1044 .20929 0 -1.5 Mabsadd L s
.12781 .22733 m
.12781 .19126 L
s
.12781 .22733 1.5 0 Mabsadd m
.12781 .22733 -1.5 0 Mabsadd L s
.12781 .19126 1.5 0 Mabsadd m
.12781 .19126 -1.5 0 Mabsadd L s
.18656 .22432 m
.13973 .22432 L
s
.18656 .22432 0 1.5 Mabsadd m
.18656 .22432 0 -1.5 Mabsadd L s
.13973 .22432 0 1.5 Mabsadd m
.13973 .22432 0 -1.5 Mabsadd L s
.16314 .23868 m
.16314 .20995 L
s
.16314 .23868 1.5 0 Mabsadd m
.16314 .23868 -1.5 0 Mabsadd L s
.16314 .20995 1.5 0 Mabsadd m
.16314 .20995 -1.5 0 Mabsadd L s
.2299 .23278 m
.18307 .23278 L
s
.2299 .23278 0 1.5 Mabsadd m
.2299 .23278 0 -1.5 Mabsadd L s
.18307 .23278 0 1.5 Mabsadd m
.18307 .23278 0 -1.5 Mabsadd L s
.20648 .24365 m
.20648 .22191 L
s
.20648 .24365 1.5 0 Mabsadd m
.20648 .24365 -1.5 0 Mabsadd L s
.20648 .22191 1.5 0 Mabsadd m
.20648 .22191 -1.5 0 Mabsadd L s
.20119 .26587 m
.15436 .26587 L
s
.20119 .26587 0 1.5 Mabsadd m
.20119 .26587 0 -1.5 Mabsadd L s
.15436 .26587 0 1.5 Mabsadd m
.15436 .26587 0 -1.5 Mabsadd L s
.17778 .27343 m
.17778 .25832 L
s
.17778 .27343 1.5 0 Mabsadd m
.17778 .27343 -1.5 0 Mabsadd L s
.17778 .25832 1.5 0 Mabsadd m
.17778 .25832 -1.5 0 Mabsadd L s
.24696 .31182 m
.20014 .31182 L
s
.24696 .31182 0 1.5 Mabsadd m
.24696 .31182 0 -1.5 Mabsadd L s
.20014 .31182 0 1.5 Mabsadd m
.20014 .31182 0 -1.5 Mabsadd L s
.22355 .31624 m
.22355 .3074 L
s
.22355 .31624 1.5 0 Mabsadd m
.22355 .31624 -1.5 0 Mabsadd L s
.22355 .3074 1.5 0 Mabsadd m
.22355 .3074 -1.5 0 Mabsadd L s
.28852 .34026 m
.2417 .34026 L
s
.28852 .34026 0 1.5 Mabsadd m
.28852 .34026 0 -1.5 Mabsadd L s
.2417 .34026 0 1.5 Mabsadd m
.2417 .34026 0 -1.5 Mabsadd L s
.26511 .34172 m
.26511 .33881 L
s
.26511 .34172 1.5 0 Mabsadd m
.26511 .34172 -1.5 0 Mabsadd L s
.26511 .33881 1.5 0 Mabsadd m
.26511 .33881 -1.5 0 Mabsadd L s
.26585 .36655 m
.21903 .36655 L
s
.26585 .36655 0 1.5 Mabsadd m
.26585 .36655 0 -1.5 Mabsadd L s
.21903 .36655 0 1.5 Mabsadd m
.21903 .36655 0 -1.5 Mabsadd L s
.24244 .36788 m
.24244 .36522 L
s
.24244 .36788 1.5 0 Mabsadd m
.24244 .36788 -1.5 0 Mabsadd L s
.24244 .36522 1.5 0 Mabsadd m
.24244 .36522 -1.5 0 Mabsadd L s
.32507 .39098 m
.27825 .39098 L
s
.32507 .39098 0 1.5 Mabsadd m
.32507 .39098 0 -1.5 Mabsadd L s
.27825 .39098 0 1.5 Mabsadd m
.27825 .39098 0 -1.5 Mabsadd L s
.30166 .39492 m
.30166 .38704 L
s
.30166 .39492 1.5 0 Mabsadd m
.30166 .39492 -1.5 0 Mabsadd L s
.30166 .38704 1.5 0 Mabsadd m
.30166 .38704 -1.5 0 Mabsadd L s
.34336 .4098 m
.29653 .4098 L
s
.34336 .4098 0 1.5 Mabsadd m
.34336 .4098 0 -1.5 Mabsadd L s
.29653 .4098 0 1.5 Mabsadd m
.29653 .4098 0 -1.5 Mabsadd L s
.31995 .41617 m
.31995 .40342 L
s
.31995 .41617 1.5 0 Mabsadd m
.31995 .41617 -1.5 0 Mabsadd L s
.31995 .40342 1.5 0 Mabsadd m
.31995 .40342 -1.5 0 Mabsadd L s
.36639 .43378 m
.31957 .43378 L
s
.36639 .43378 0 1.5 Mabsadd m
.36639 .43378 0 -1.5 Mabsadd L s
.31957 .43378 0 1.5 Mabsadd m
.31957 .43378 0 -1.5 Mabsadd L s
.34298 .44241 m
.34298 .42515 L
s
.34298 .44241 1.5 0 Mabsadd m
.34298 .44241 -1.5 0 Mabsadd L s
.34298 .42515 1.5 0 Mabsadd m
.34298 .42515 -1.5 0 Mabsadd L s
.33 .44391 m
.28318 .44391 L
s
.33 .44391 0 1.5 Mabsadd m
.33 .44391 0 -1.5 Mabsadd L s
.28318 .44391 0 1.5 Mabsadd m
.28318 .44391 0 -1.5 Mabsadd L s
.30659 .45462 m
.30659 .4332 L
s
.30659 .45462 1.5 0 Mabsadd m
.30659 .45462 -1.5 0 Mabsadd L s
.30659 .4332 1.5 0 Mabsadd m
.30659 .4332 -1.5 0 Mabsadd L s
.42086 .45518 m
.37404 .45518 L
s
.42086 .45518 0 1.5 Mabsadd m
.42086 .45518 0 -1.5 Mabsadd L s
.37404 .45518 0 1.5 Mabsadd m
.37404 .45518 0 -1.5 Mabsadd L s
.39745 .46779 m
.39745 .44256 L
s
.39745 .46779 1.5 0 Mabsadd m
.39745 .46779 -1.5 0 Mabsadd L s
.39745 .44256 1.5 0 Mabsadd m
.39745 .44256 -1.5 0 Mabsadd L s
.44504 .4751 m
.39822 .4751 L
s
.44504 .4751 0 1.5 Mabsadd m
.44504 .4751 0 -1.5 Mabsadd L s
.39822 .4751 0 1.5 Mabsadd m
.39822 .4751 0 -1.5 Mabsadd L s
.42163 .48944 m
.42163 .46076 L
s
.42163 .48944 1.5 0 Mabsadd m
.42163 .48944 -1.5 0 Mabsadd L s
.42163 .46076 1.5 0 Mabsadd m
.42163 .46076 -1.5 0 Mabsadd L s
.45734 .50451 m
.41051 .50451 L
s
.45734 .50451 0 1.5 Mabsadd m
.45734 .50451 0 -1.5 Mabsadd L s
.41051 .50451 0 1.5 Mabsadd m
.41051 .50451 0 -1.5 Mabsadd L s
.43392 .5204 m
.43392 .48862 L
s
.43392 .5204 1.5 0 Mabsadd m
.43392 .5204 -1.5 0 Mabsadd L s
.43392 .48862 1.5 0 Mabsadd m
.43392 .48862 -1.5 0 Mabsadd L s
.52074 .53419 m
.47391 .53419 L
s
.52074 .53419 0 1.5 Mabsadd m
.52074 .53419 0 -1.5 Mabsadd L s
.47391 .53419 0 1.5 Mabsadd m
.47391 .53419 0 -1.5 Mabsadd L s
.49732 .55146 m
.49732 .51693 L
s
.49732 .55146 1.5 0 Mabsadd m
.49732 .55146 -1.5 0 Mabsadd L s
.49732 .51693 1.5 0 Mabsadd m
.49732 .51693 -1.5 0 Mabsadd L s
.52735 .55673 m
.48053 .55673 L
s
.52735 .55673 0 1.5 Mabsadd m
.52735 .55673 0 -1.5 Mabsadd L s
.48053 .55673 0 1.5 Mabsadd m
.48053 .55673 0 -1.5 Mabsadd L s
.50394 .5752 m
.50394 .53827 L
s
.50394 .5752 1.5 0 Mabsadd m
.50394 .5752 -1.5 0 Mabsadd L s
.50394 .53827 1.5 0 Mabsadd m
.50394 .53827 -1.5 0 Mabsadd L s
.50178 .54875 m
.45496 .54875 L
s
.50178 .54875 0 1.5 Mabsadd m
.50178 .54875 0 -1.5 Mabsadd L s
.45496 .54875 0 1.5 Mabsadd m
.45496 .54875 0 -1.5 Mabsadd L s
.47837 .56823 m
.47837 .52927 L
s
.47837 .56823 1.5 0 Mabsadd m
.47837 .56823 -1.5 0 Mabsadd L s
.47837 .52927 1.5 0 Mabsadd m
.47837 .52927 -1.5 0 Mabsadd L s
.55838 .58299 m
.51155 .58299 L
s
.55838 .58299 0 1.5 Mabsadd m
.55838 .58299 0 -1.5 Mabsadd L s
.51155 .58299 0 1.5 Mabsadd m
.51155 .58299 0 -1.5 Mabsadd L s
.53497 .60332 m
.53497 .56266 L
s
.53497 .60332 1.5 0 Mabsadd m
.53497 .60332 -1.5 0 Mabsadd L s
.53497 .56266 1.5 0 Mabsadd m
.53497 .56266 -1.5 0 Mabsadd L s
.58731 .56607 m
.54049 .56607 L
s
.58731 .56607 0 1.5 Mabsadd m
.58731 .56607 0 -1.5 Mabsadd L s
.54049 .56607 0 1.5 Mabsadd m
.54049 .56607 0 -1.5 Mabsadd L s
.5639 .58707 m
.5639 .54507 L
s
.5639 .58707 1.5 0 Mabsadd m
.5639 .58707 -1.5 0 Mabsadd L s
.5639 .54507 1.5 0 Mabsadd m
.5639 .54507 -1.5 0 Mabsadd L s
.62977 .56397 m
.58295 .56397 L
s
.62977 .56397 0 1.5 Mabsadd m
.62977 .56397 0 -1.5 Mabsadd L s
.58295 .56397 0 1.5 Mabsadd m
.58295 .56397 0 -1.5 Mabsadd L s
.60636 .58546 m
.60636 .54248 L
s
.60636 .58546 1.5 0 Mabsadd m
.60636 .58546 -1.5 0 Mabsadd L s
.60636 .54248 1.5 0 Mabsadd m
.60636 .54248 -1.5 0 Mabsadd L s
.62923 .53897 m
.5824 .53897 L
s
.62923 .53897 0 1.5 Mabsadd m
.62923 .53897 0 -1.5 Mabsadd L s
.5824 .53897 0 1.5 Mabsadd m
.5824 .53897 0 -1.5 Mabsadd L s
.60581 .56077 m
.60581 .51716 L
s
.60581 .56077 1.5 0 Mabsadd m
.60581 .56077 -1.5 0 Mabsadd L s
.60581 .51716 1.5 0 Mabsadd m
.60581 .51716 -1.5 0 Mabsadd L s
.63138 .56463 m
.58455 .56463 L
s
.63138 .56463 0 1.5 Mabsadd m
.63138 .56463 0 -1.5 Mabsadd L s
.58455 .56463 0 1.5 Mabsadd m
.58455 .56463 0 -1.5 Mabsadd L s
.60796 .58658 m
.60796 .54268 L
s
.60796 .58658 1.5 0 Mabsadd m
.60796 .58658 -1.5 0 Mabsadd L s
.60796 .54268 1.5 0 Mabsadd m
.60796 .54268 -1.5 0 Mabsadd L s
.66553 .57836 m
.61871 .57836 L
s
.66553 .57836 0 1.5 Mabsadd m
.66553 .57836 0 -1.5 Mabsadd L s
.61871 .57836 0 1.5 Mabsadd m
.61871 .57836 0 -1.5 Mabsadd L s
.64212 .60027 m
.64212 .55645 L
s
.64212 .60027 1.5 0 Mabsadd m
.64212 .60027 -1.5 0 Mabsadd L s
.64212 .55645 1.5 0 Mabsadd m
.64212 .55645 -1.5 0 Mabsadd L s
.66415 .56817 m
.61733 .56817 L
s
.66415 .56817 0 1.5 Mabsadd m
.66415 .56817 0 -1.5 Mabsadd L s
.61733 .56817 0 1.5 Mabsadd m
.61733 .56817 0 -1.5 Mabsadd L s
.64074 .58987 m
.64074 .54648 L
s
.64074 .58987 1.5 0 Mabsadd m
.64074 .58987 -1.5 0 Mabsadd L s
.64074 .54648 1.5 0 Mabsadd m
.64074 .54648 -1.5 0 Mabsadd L s
.78372 .54478 m
.7369 .54478 L
s
.78372 .54478 0 1.5 Mabsadd m
.78372 .54478 0 -1.5 Mabsadd L s
.7369 .54478 0 1.5 Mabsadd m
.7369 .54478 0 -1.5 Mabsadd L s
.76031 .56609 m
.76031 .52348 L
s
.76031 .56609 1.5 0 Mabsadd m
.76031 .56609 -1.5 0 Mabsadd L s
.76031 .52348 1.5 0 Mabsadd m
.76031 .52348 -1.5 0 Mabsadd L s
.75074 .57845 m
.70392 .57845 L
s
.75074 .57845 0 1.5 Mabsadd m
.75074 .57845 0 -1.5 Mabsadd L s
.70392 .57845 0 1.5 Mabsadd m
.70392 .57845 0 -1.5 Mabsadd L s
.72733 .59919 m
.72733 .55771 L
s
.72733 .59919 1.5 0 Mabsadd m
.72733 .59919 -1.5 0 Mabsadd L s
.72733 .55771 1.5 0 Mabsadd m
.72733 .55771 -1.5 0 Mabsadd L s
.7787 .54504 m
.73187 .54504 L
s
.7787 .54504 0 1.5 Mabsadd m
.7787 .54504 0 -1.5 Mabsadd L s
.73187 .54504 0 1.5 Mabsadd m
.73187 .54504 0 -1.5 Mabsadd L s
.75529 .56503 m
.75529 .52504 L
s
.75529 .56503 1.5 0 Mabsadd m
.75529 .56503 -1.5 0 Mabsadd L s
.75529 .52504 1.5 0 Mabsadd m
.75529 .52504 -1.5 0 Mabsadd L s
.78894 .55011 m
.74212 .55011 L
s
.78894 .55011 0 1.5 Mabsadd m
.78894 .55011 0 -1.5 Mabsadd L s
.74212 .55011 0 1.5 Mabsadd m
.74212 .55011 0 -1.5 Mabsadd L s
.76553 .56919 m
.76553 .53103 L
s
.76553 .56919 1.5 0 Mabsadd m
.76553 .56919 -1.5 0 Mabsadd L s
.76553 .53103 1.5 0 Mabsadd m
.76553 .53103 -1.5 0 Mabsadd L s
.80064 .5413 m
.75382 .5413 L
s
.80064 .5413 0 1.5 Mabsadd m
.80064 .5413 0 -1.5 Mabsadd L s
.75382 .5413 0 1.5 Mabsadd m
.75382 .5413 0 -1.5 Mabsadd L s
.77723 .55928 m
.77723 .52332 L
s
.77723 .55928 1.5 0 Mabsadd m
.77723 .55928 -1.5 0 Mabsadd L s
.77723 .52332 1.5 0 Mabsadd m
.77723 .52332 -1.5 0 Mabsadd L s
.80778 .52466 m
.76095 .52466 L
s
.80778 .52466 0 1.5 Mabsadd m
.80778 .52466 0 -1.5 Mabsadd L s
.76095 .52466 0 1.5 Mabsadd m
.76095 .52466 0 -1.5 Mabsadd L s
.78436 .54136 m
.78436 .50795 L
s
.78436 .54136 1.5 0 Mabsadd m
.78436 .54136 -1.5 0 Mabsadd L s
.78436 .50795 1.5 0 Mabsadd m
.78436 .50795 -1.5 0 Mabsadd L s
.85208 .53194 m
.80526 .53194 L
s
.85208 .53194 0 1.5 Mabsadd m
.85208 .53194 0 -1.5 Mabsadd L s
.80526 .53194 0 1.5 Mabsadd m
.80526 .53194 0 -1.5 Mabsadd L s
.82867 .5472 m
.82867 .51668 L
s
.82867 .5472 1.5 0 Mabsadd m
.82867 .5472 -1.5 0 Mabsadd L s
.82867 .51668 1.5 0 Mabsadd m
.82867 .51668 -1.5 0 Mabsadd L s
.83027 .46621 m
.78345 .46621 L
s
.83027 .46621 0 1.5 Mabsadd m
.83027 .46621 0 -1.5 Mabsadd L s
.78345 .46621 0 1.5 Mabsadd m
.78345 .46621 0 -1.5 Mabsadd L s
.80686 .47984 m
.80686 .45257 L
s
.80686 .47984 1.5 0 Mabsadd m
.80686 .47984 -1.5 0 Mabsadd L s
.80686 .45257 1.5 0 Mabsadd m
.80686 .45257 -1.5 0 Mabsadd L s
.86496 .46618 m
.81813 .46618 L
s
.86496 .46618 0 1.5 Mabsadd m
.86496 .46618 0 -1.5 Mabsadd L s
.81813 .46618 0 1.5 Mabsadd m
.81813 .46618 0 -1.5 Mabsadd L s
.84154 .47801 m
.84154 .45434 L
s
.84154 .47801 1.5 0 Mabsadd m
.84154 .47801 -1.5 0 Mabsadd L s
.84154 .45434 1.5 0 Mabsadd m
.84154 .45434 -1.5 0 Mabsadd L s
.93899 .45075 m
.89216 .45075 L
s
.93899 .45075 0 1.5 Mabsadd m
.93899 .45075 0 -1.5 Mabsadd L s
.89216 .45075 0 1.5 Mabsadd m
.89216 .45075 0 -1.5 Mabsadd L s
.91558 .4606 m
.91558 .44089 L
s
.91558 .4606 1.5 0 Mabsadd m
.91558 .4606 -1.5 0 Mabsadd L s
.91558 .44089 1.5 0 Mabsadd m
.91558 .44089 -1.5 0 Mabsadd L s
.95261 .42865 m
.90579 .42865 L
s
.95261 .42865 0 1.5 Mabsadd m
.95261 .42865 0 -1.5 Mabsadd L s
.90579 .42865 0 1.5 Mabsadd m
.90579 .42865 0 -1.5 Mabsadd L s
.9292 .43635 m
.9292 .42095 L
s
.9292 .43635 1.5 0 Mabsadd m
.9292 .43635 -1.5 0 Mabsadd L s
.9292 .42095 1.5 0 Mabsadd m
.9292 .42095 -1.5 0 Mabsadd L s
.95562 .39838 m
.9088 .39838 L
s
.95562 .39838 0 1.5 Mabsadd m
.95562 .39838 0 -1.5 Mabsadd L s
.9088 .39838 0 1.5 Mabsadd m
.9088 .39838 0 -1.5 Mabsadd L s
.93221 .40375 m
.93221 .39301 L
s
.93221 .40375 1.5 0 Mabsadd m
.93221 .40375 -1.5 0 Mabsadd L s
.93221 .39301 1.5 0 Mabsadd m
.93221 .39301 -1.5 0 Mabsadd L s
.0001 w
.04722 .06467 -1.875 0 Mabsadd m
.04722 .06467 0 2.5 Mabsadd L
.04722 .06467 1.875 0 Mabsadd L
.04722 .06467 0 -2.5 Mabsadd L
.04722 .06467 -1.875 0 Mabsadd L closepath F
.08988 .09901 -1.875 0 Mabsadd m
.08988 .09901 0 2.5 Mabsadd L
.08988 .09901 1.875 0 Mabsadd L
.08988 .09901 0 -2.5 Mabsadd L
.08988 .09901 -1.875 0 Mabsadd L closepath F
.13917 .13879 -1.875 0 Mabsadd m
.13917 .13879 0 2.5 Mabsadd L
.13917 .13879 1.875 0 Mabsadd L
.13917 .13879 0 -2.5 Mabsadd L
.13917 .13879 -1.875 0 Mabsadd L closepath F
.12781 .20929 -1.875 0 Mabsadd m
.12781 .20929 0 2.5 Mabsadd L
.12781 .20929 1.875 0 Mabsadd L
.12781 .20929 0 -2.5 Mabsadd L
.12781 .20929 -1.875 0 Mabsadd L closepath F
.16314 .22432 -1.875 0 Mabsadd m
.16314 .22432 0 2.5 Mabsadd L
.16314 .22432 1.875 0 Mabsadd L
.16314 .22432 0 -2.5 Mabsadd L
.16314 .22432 -1.875 0 Mabsadd L closepath F
.20648 .23278 -1.875 0 Mabsadd m
.20648 .23278 0 2.5 Mabsadd L
.20648 .23278 1.875 0 Mabsadd L
.20648 .23278 0 -2.5 Mabsadd L
.20648 .23278 -1.875 0 Mabsadd L closepath F
.17778 .26587 -1.875 0 Mabsadd m
.17778 .26587 0 2.5 Mabsadd L
.17778 .26587 1.875 0 Mabsadd L
.17778 .26587 0 -2.5 Mabsadd L
.17778 .26587 -1.875 0 Mabsadd L closepath F
.22355 .31182 -1.875 0 Mabsadd m
.22355 .31182 0 2.5 Mabsadd L
.22355 .31182 1.875 0 Mabsadd L
.22355 .31182 0 -2.5 Mabsadd L
.22355 .31182 -1.875 0 Mabsadd L closepath F
.26511 .34026 -1.875 0 Mabsadd m
.26511 .34026 0 2.5 Mabsadd L
.26511 .34026 1.875 0 Mabsadd L
.26511 .34026 0 -2.5 Mabsadd L
.26511 .34026 -1.875 0 Mabsadd L closepath F
.24244 .36655 -1.875 0 Mabsadd m
.24244 .36655 0 2.5 Mabsadd L
.24244 .36655 1.875 0 Mabsadd L
.24244 .36655 0 -2.5 Mabsadd L
.24244 .36655 -1.875 0 Mabsadd L closepath F
.30166 .39098 -1.875 0 Mabsadd m
.30166 .39098 0 2.5 Mabsadd L
.30166 .39098 1.875 0 Mabsadd L
.30166 .39098 0 -2.5 Mabsadd L
.30166 .39098 -1.875 0 Mabsadd L closepath F
.31995 .4098 -1.875 0 Mabsadd m
.31995 .4098 0 2.5 Mabsadd L
.31995 .4098 1.875 0 Mabsadd L
.31995 .4098 0 -2.5 Mabsadd L
.31995 .4098 -1.875 0 Mabsadd L closepath F
.34298 .43378 -1.875 0 Mabsadd m
.34298 .43378 0 2.5 Mabsadd L
.34298 .43378 1.875 0 Mabsadd L
.34298 .43378 0 -2.5 Mabsadd L
.34298 .43378 -1.875 0 Mabsadd L closepath F
.30659 .44391 -1.875 0 Mabsadd m
.30659 .44391 0 2.5 Mabsadd L
.30659 .44391 1.875 0 Mabsadd L
.30659 .44391 0 -2.5 Mabsadd L
.30659 .44391 -1.875 0 Mabsadd L closepath F
.39745 .45518 -1.875 0 Mabsadd m
.39745 .45518 0 2.5 Mabsadd L
.39745 .45518 1.875 0 Mabsadd L
.39745 .45518 0 -2.5 Mabsadd L
.39745 .45518 -1.875 0 Mabsadd L closepath F
.42163 .4751 -1.875 0 Mabsadd m
.42163 .4751 0 2.5 Mabsadd L
.42163 .4751 1.875 0 Mabsadd L
.42163 .4751 0 -2.5 Mabsadd L
.42163 .4751 -1.875 0 Mabsadd L closepath F
.43392 .50451 -1.875 0 Mabsadd m
.43392 .50451 0 2.5 Mabsadd L
.43392 .50451 1.875 0 Mabsadd L
.43392 .50451 0 -2.5 Mabsadd L
.43392 .50451 -1.875 0 Mabsadd L closepath F
.49732 .53419 -1.875 0 Mabsadd m
.49732 .53419 0 2.5 Mabsadd L
.49732 .53419 1.875 0 Mabsadd L
.49732 .53419 0 -2.5 Mabsadd L
.49732 .53419 -1.875 0 Mabsadd L closepath F
.50394 .55673 -1.875 0 Mabsadd m
.50394 .55673 0 2.5 Mabsadd L
.50394 .55673 1.875 0 Mabsadd L
.50394 .55673 0 -2.5 Mabsadd L
.50394 .55673 -1.875 0 Mabsadd L closepath F
.47837 .54875 -1.875 0 Mabsadd m
.47837 .54875 0 2.5 Mabsadd L
.47837 .54875 1.875 0 Mabsadd L
.47837 .54875 0 -2.5 Mabsadd L
.47837 .54875 -1.875 0 Mabsadd L closepath F
.53497 .58299 -1.875 0 Mabsadd m
.53497 .58299 0 2.5 Mabsadd L
.53497 .58299 1.875 0 Mabsadd L
.53497 .58299 0 -2.5 Mabsadd L
.53497 .58299 -1.875 0 Mabsadd L closepath F
.5639 .56607 -1.875 0 Mabsadd m
.5639 .56607 0 2.5 Mabsadd L
.5639 .56607 1.875 0 Mabsadd L
.5639 .56607 0 -2.5 Mabsadd L
.5639 .56607 -1.875 0 Mabsadd L closepath F
.60636 .56397 -1.875 0 Mabsadd m
.60636 .56397 0 2.5 Mabsadd L
.60636 .56397 1.875 0 Mabsadd L
.60636 .56397 0 -2.5 Mabsadd L
.60636 .56397 -1.875 0 Mabsadd L closepath F
.60581 .53897 -1.875 0 Mabsadd m
.60581 .53897 0 2.5 Mabsadd L
.60581 .53897 1.875 0 Mabsadd L
.60581 .53897 0 -2.5 Mabsadd L
.60581 .53897 -1.875 0 Mabsadd L closepath F
.60796 .56463 -1.875 0 Mabsadd m
.60796 .56463 0 2.5 Mabsadd L
.60796 .56463 1.875 0 Mabsadd L
.60796 .56463 0 -2.5 Mabsadd L
.60796 .56463 -1.875 0 Mabsadd L closepath F
.64212 .57836 -1.875 0 Mabsadd m
.64212 .57836 0 2.5 Mabsadd L
.64212 .57836 1.875 0 Mabsadd L
.64212 .57836 0 -2.5 Mabsadd L
.64212 .57836 -1.875 0 Mabsadd L closepath F
.64074 .56817 -1.875 0 Mabsadd m
.64074 .56817 0 2.5 Mabsadd L
.64074 .56817 1.875 0 Mabsadd L
.64074 .56817 0 -2.5 Mabsadd L
.64074 .56817 -1.875 0 Mabsadd L closepath F
.76031 .54478 -1.875 0 Mabsadd m
.76031 .54478 0 2.5 Mabsadd L
.76031 .54478 1.875 0 Mabsadd L
.76031 .54478 0 -2.5 Mabsadd L
.76031 .54478 -1.875 0 Mabsadd L closepath F
.72733 .57845 -1.875 0 Mabsadd m
.72733 .57845 0 2.5 Mabsadd L
.72733 .57845 1.875 0 Mabsadd L
.72733 .57845 0 -2.5 Mabsadd L
.72733 .57845 -1.875 0 Mabsadd L closepath F
.75529 .54504 -1.875 0 Mabsadd m
.75529 .54504 0 2.5 Mabsadd L
.75529 .54504 1.875 0 Mabsadd L
.75529 .54504 0 -2.5 Mabsadd L
.75529 .54504 -1.875 0 Mabsadd L closepath F
.76553 .55011 -1.875 0 Mabsadd m
.76553 .55011 0 2.5 Mabsadd L
.76553 .55011 1.875 0 Mabsadd L
.76553 .55011 0 -2.5 Mabsadd L
.76553 .55011 -1.875 0 Mabsadd L closepath F
.77723 .5413 -1.875 0 Mabsadd m
.77723 .5413 0 2.5 Mabsadd L
.77723 .5413 1.875 0 Mabsadd L
.77723 .5413 0 -2.5 Mabsadd L
.77723 .5413 -1.875 0 Mabsadd L closepath F
.78436 .52466 -1.875 0 Mabsadd m
.78436 .52466 0 2.5 Mabsadd L
.78436 .52466 1.875 0 Mabsadd L
.78436 .52466 0 -2.5 Mabsadd L
.78436 .52466 -1.875 0 Mabsadd L closepath F
.82867 .53194 -1.875 0 Mabsadd m
.82867 .53194 0 2.5 Mabsadd L
.82867 .53194 1.875 0 Mabsadd L
.82867 .53194 0 -2.5 Mabsadd L
.82867 .53194 -1.875 0 Mabsadd L closepath F
.80686 .46621 -1.875 0 Mabsadd m
.80686 .46621 0 2.5 Mabsadd L
.80686 .46621 1.875 0 Mabsadd L
.80686 .46621 0 -2.5 Mabsadd L
.80686 .46621 -1.875 0 Mabsadd L closepath F
.84154 .46618 -1.875 0 Mabsadd m
.84154 .46618 0 2.5 Mabsadd L
.84154 .46618 1.875 0 Mabsadd L
.84154 .46618 0 -2.5 Mabsadd L
.84154 .46618 -1.875 0 Mabsadd L closepath F
.91558 .45075 -1.875 0 Mabsadd m
.91558 .45075 0 2.5 Mabsadd L
.91558 .45075 1.875 0 Mabsadd L
.91558 .45075 0 -2.5 Mabsadd L
.91558 .45075 -1.875 0 Mabsadd L closepath F
.9292 .42865 -1.875 0 Mabsadd m
.9292 .42865 0 2.5 Mabsadd L
.9292 .42865 1.875 0 Mabsadd L
.9292 .42865 0 -2.5 Mabsadd L
.9292 .42865 -1.875 0 Mabsadd L closepath F
.93221 .39838 -1.875 0 Mabsadd m
.93221 .39838 0 2.5 Mabsadd L
.93221 .39838 1.875 0 Mabsadd L
.93221 .39838 0 -2.5 Mabsadd L
.93221 .39838 -1.875 0 Mabsadd L closepath F
.5 Mabswid
.03974 .01472 m
.07772 .08403 L
.11916 .15433 L
.15807 .21532 L
.19549 .26939 L
.23532 .32198 L
.27366 .36779 L
.31441 .41129 L
.35367 .44814 L
.39144 .47892 L
.43163 .50662 L
.47032 .52838 L
.50752 .54476 L
.52821 .55194 L
.54714 .55731 L
.56668 .56164 L
.58526 .56461 L
.59469 .5657 L
.59986 .56617 L
.60473 .56654 L
.60929 .56681 L
.61168 .56693 L
.61422 .56704 L
.61635 .56711 L
.61869 .56717 L
.61972 .56719 L
.62082 .56721 L
.62185 .56723 L
.62281 .56724 L
.62401 .56725 L
.62465 .56725 L
.62533 .56725 L
.62601 .56725 L
.62673 .56725 L
.62801 .56724 L
.6292 .56723 L
.63031 .56722 L
.63152 .5672 L
.63282 .56717 L
.63575 .56709 L
.63852 .56699 L
.6437 .56673 L
.64864 .56641 L
.65319 .56604 L
.66348 .56496 L
.67301 .56366 L
.68193 .56217 L
.70223 .55784 L
.72218 .55229 L
.7404 .5461 L
Mistroke
.7795 .52922 L
.8171 .50834 L
.85712 .48113 L
.89565 .45006 L
.93659 .41181 L
.97603 .36985 L
.97619 .36967 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0.625, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00m03oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo]03oool00?H0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;<0oooo003g0?ooo`040;mdM02LW4P004RL0=oookD0
oooo003g0?ooo`030?oo_`1d000007Bo0;H0oooo003g0?ooo`030?oo_`1d0000M;oo0;H0oooo003g
0?ooo`040?noM000000004RL0=oookD0oooo003g0?ooo`050;md0000B9`0gmnL04Q8W03Oool0]03o
ool00?D0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0/P3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00640oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`110?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
oc`0oooo00@0_g@0000000000000M;ooA03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom10?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool05@3oool006<0oooo00<0okmd0000001d_ol0@`3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_c/0oooo00H0gia8001d_`3oool0omnL04P00018W=m20?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
@`3oool00`3o_g@0000007Boo`0H0?ooo`00H`3oool00`3ooml0W4P0001d_`120?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0h0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0@@3oool00`3ooml0W4P0001d_`0H0?ooo`00I03oool0
0`3OW4P004RL0=ooo`110?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0h0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
?`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0@P3oool00`3OW4P004RL0=ooo`0G0?ooo`00E03oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=03oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0o0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`120?ooo`030?oOW01804P0W=oo
01L0oooo001T0?ooo`030?oo_`1d0000M;oo0480oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03T0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`100?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`120?ooo`030?oo_`1d0000M;oo01L0oooo001U0?ooo`030;md0000M;l0oooo0440oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03T0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`100?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`130?ooo`030;md0000M;l0oooo01H0oooo
001U0?ooo`030=nLB000B9`0gooo0440oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo>@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo03l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo@`3oool00`3OW4P004RL0=ooo`0F0?ooo`00I@3oool00`3o
gi`0B01809cOo`110?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`0o0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04@0oooo00<0omnL04P0B02Lgol05P3oool006D0oooo00<0oono07@0001d
_ol0@P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0?03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0400oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04D0oooo00<0oono07@0001d
_ol05P3oool006H0oooo00<0_g@0001d_`3oool0@P3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0?@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd80oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04H0oooo00<0_g@0001d_`3o
ool05@3oool006@0oooo00D0oono07@000000000001d0;ooo`130?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0k0?ooo`040=nLB00000000000001d_d@0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO0440oooo00D0oono07@00000
0000001d0;ooo`0E0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00D@3ooooo000005000000403oool00540oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?oo
o`2W0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool00540oooo00<00000
0?ooo`3oool0Y`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo00h0oooo001A0?ooo`<00000303oool00`000000oooo0?ooo`2H0?ooo`03
0000003oool0oooo0:00oooo0P00000@0?ooo`00D@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo09P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3o
ool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo09L0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo09L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`0;0?ooo`D00000U`3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`000000U`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo
001A0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`000000U`3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo00d0oooo00@00000
0?ooo`3oool00000UP3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A
0?ooo`<000003@3oool010000000oooo0?ooo`00002F0?ooo`030000003oool0oooo0:00oooo0P00
000@0?ooo`00D@3oool00`000000oooo0?ooo`0=0?ooo`050000003oool0oooo0?ooo`000000U@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0
oooo00d0oooo00D000000?ooo`3oool0oooo0000002E0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`02
0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>
0?ooo`00D@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo0P0000040?oo
o`040000003oool0oooo00000940oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>
0?ooo`00D@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo100000040?oo
o`030000003oool000000940oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?oo
o`00D@3oool00`000000oooo0?ooo`050?oooa400000T@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
0`3oool5000000@0oooo0P00002A0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
3P3oool00540oooo0`0000050?ooo`030000003oool0oooo00@0oooo0`0000050?ooo`800000T@3o
ool00`000000oooo0?ooo`2P0?ooo`800000403oool00540oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3oool0oooo1P00
00280?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool00540oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0R@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo08T0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?oo
o`00>03oool00`3ookl0M000000000020000000304RLg`3oool0oooo00<0oooo00@0_g@000000000
0000M;oo2P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool03P3oool003X0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B000
04RLg`T0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`070?ooo`050000003oool0
oooo0?ooo`000000R`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo000j
0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`08
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`00
002;0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool003/0oooo00<0gia8
0018W03Oool00`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo000008/0oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00:`3oool00`3ooml0W4P00000
00070000000304RLg`3oool0oooo00<0oooo00<0omnL04P0B02Lgol00`3oool00`3ooml0W4P0001d
_`030?ooo`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool02`3oool5000000@0oooo00<0
00000?ooo`3oool00P3oool4000000H0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo00h0oooo000k0?ooo`030?oo_`1d0000M;oo00@0oooo00<0_g@0
0018W03Oool00P3oool00`3ogi`0B00004RLg`070?ooo`@000004`3oool00`000000oooo0?ooo`02
0?ooo`D000001@3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo09l0oooo0`00000@
0?ooo`00?03oool00`2oM00007Bo0?ooo`030?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l01`3oool00`000000oooo0?ooo`0D0?oooa000000P`3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo00h0oooo000l0?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018
W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool00`3oool4000000D0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo00h0oooo000l0?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0
gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool00`3oool3000000H0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo00h0oooo000l0?ooo`030?oo_`1d0000M;oo00@0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0oooo01/0oooo0P00002:0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool003d0oooo00<0_g@0001d_`3oool0103o
ool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<000000?ooo`3oool06`3oool2000008X0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00>`3oool01@3ookl0M000
0000000007@0_ooo00D0oooo00@0gia800000000000007Bo203oool00`000000oooo0?ooo`0K0?oo
o`030000003oool0000008T0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?oo
o`00D@3oool00`000000oooo0?ooo`0K0?ooo`040000003oool0oooo000000/0oooo1@00001h0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3o
ool06`3oool010000000oooo0?ooo`00000=0?ooo`030000003oool0oooo07P0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool3000001/0oooo00D000000?ooo`3o
ool0oooo0000000<0?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3oool0X03oool20000
0100oooo001A0?ooo`030000003oool0oooo01T0oooo1P0000000`3oool000000?ooo`0:0?ooo`03
0000003oool0oooo07P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00
D@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
N03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003o
ool0oooo0240oooo00@000000?ooo`3oool000001P3oool3000000H0oooo00<000000?ooo`3oool0
L03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003o
ool0oooo0240oooo00@000000?ooo`3oool000001@3oool5000000D0oooo00<000000?ooo`3oool0
L03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`0000004000001b0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool08`3oool2000000D0oooo1@000005
0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>
0?ooo`00D@3oool00`000000oooo0?ooo`0S0?ooo`8000001P3oool3000000H0oooo00<000000?oo
o`3oool0L03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`<0
0000903oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo07P0oooo00<000000?ooo`3o
ool0X03oool200000100oooo001A0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
3P3oool00540oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo07P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0N03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo
02H0oooo00@000000?ooo`3oool0oooo1@00001h0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`1m0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3o
ool09`3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool03P3oool00540oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1l0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0:@3o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3o
ool00540oooo0`00000Y0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0X03oool2
00000100oooo001A0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0NP3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo02/0
oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0
oooo001A0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0N@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0N03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A
0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0N03oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo02H0oooo1@0000020?oo
o`030000003oool0oooo07L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?oo
o`00D@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0MP3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`<00000
:03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0
X03oool200000100oooo001A0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`040?ooo`800000MP3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo00h0oooo001A0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0103oool3
000000H0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo00h0oooo001A0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool00`3oool50000
00D0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
00h0oooo00110?ooo`060?oOW0180000000000000000B9`0gooo2P3oool00`000000oooo0?ooo`0P
0?oooaL00000L03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo00110?oo
o`060?noM0000000M;oo0?ooo`3ogi`0B7Bo2P3oool00`000000oooo0?ooo`0P0?ooo`030000003o
ool0oooo00<0oooo100000060?ooo`8000000P3oool00`000000oooo0?ooo`1`0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool03P3oool004H0oooo00<0okmd04RLg`3oool0203oool0
0`000000oooo0?ooo`0P0?ooo`030000003oool0oooo00@0oooo0P0000030?ooo`090000003oool0
oooo0?ooo`000000oooo0000003oool0000000L0oooo00<000000?ooo`3oool0J03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo00160?ooo`030?oo_`1d07@0_ooo00P0oooo
00<000000?ooo`3oool0:03oool01@000000oooo0?ooo`3oool0000000H0oooo0`0000060?ooo`03
0000003oool0oooo06P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00
=03oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00D0oooo00<0oooO09a8001d_ol0
203oool00`000000oooo0?ooo`0X0?ooo`050000003oool0oooo0?ooo`0000001@3oool5000000D0
oooo00@000000?ooo`3oool0oooo1P00001Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool03P3oool004H0oooo00<0oono07@00018W=l0203oool4000002L0oooo00@000000?ooo`3o
ool0oooo4@0000050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0FP3oool00`00
0000oooo0?ooo`2O0?ooo`<00000403oool004H0oooo00<0okmd0000001d_ol0203oool00`000000
oooo0?ooo`0V0?ooo`D0000000<0oooo0000003oool0103oool4000000D0oooo0P0000040?ooo`<0
00001P3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool03P3oool004<0oooo00H0omoO0=ooo`3o_g@000000000M02oool80?ooo`030000003oool0oooo
02`0oooo00<000000?ooo`3oool0103oool2000000030?ooo`000000oooo00<0oooo0P0000040?oo
o`@000001@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool03P3oool004<0oooo00D0oono07@000000000001809cOo`090?ooo`030000003oool0oooo
03@0oooo00<000000?ooo`000000103oool@000005`0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0>0?ooo`00A03oool00`3OW9`0_ooo0?ooo`0:0?ooo`030000003oool0oooo03@0
oooo00@000000?ooo`3oool000000`3oool00`000000oooo0?ooo`030?ooo`<000001P3oool00`00
0000oooo0?ooo`1J0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool004@0
oooo00<0omnL07BLg`3oool02P3oool00`000000oooo0?ooo`0b0?ooo`D0000000D0oooo0000003o
ool0oooo000000060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0FP3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo00150?ooo`050;md0000000000000000
0018W=l01`3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo00D0oooo1P00001Q0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool03P3oool00540oooo0`00000j0?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool0
X03oool200000100oooo001A0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0J@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0
oooo03/0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo00h0oooo001A0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool00`3oool60000
00@0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
00h0oooo001A0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0103oool300000080
oooo00D000000?ooo`3oool0oooo0000001U0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool03P3oool00540oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`030?ooo`D0
00000P3oool010000000oooo0?ooo`00001U0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool03P3oool00540oooo00<000000?ooo`3oool0<@3ooolA000006D0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo00<0oooo100000040?ooo`030000003oool0000006D0oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`0a0?ooo`030000003o
ool0oooo00<0oooo1@0000040?ooo`800000I@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo00h0oooo001A0?ooo`<00000@@3oool00`000000oooo0?ooo`1S0?ooo`030000003oool0
oooo0:00oooo0P00000@0?ooo`00D@3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0640oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?oo
o`140?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0>0?ooo`00D@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo0600oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`150?oo
o`030000003oool0oooo05l0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?oo
o`008P3oool00`2oM00004RL0=ooo`0/0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3o
ool0GP3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo000R0?ooo`030?oO
W018M;l0oooo02`0oooo0`0000170?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool0
X03oool200000100oooo000G0?ooo`030?oo_`1dM;l0oooo00L0oooo00<0okmd0018W03Oool0;@3o
ool00`000000oooo0?ooo`170?ooo`800000GP3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo00h0oooo000G0?ooo`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3oool0;P3o
ool00`000000oooo0?ooo`100?ooo`030000003oool0oooo00<0oooo1@0000050?ooo`030000003o
ool0oooo05@0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00603oool0
0`3ogi`0B000000000050000000304RLg`3oool0oooo02h0oooo00<000000?ooo`3oool0@03oool0
0`000000oooo0?ooo`030?ooo`D000001@3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool03P3oool001d0oooo00<0okmd0000B02Lgol0<@3oool00`00
0000oooo0?ooo`100?oooa400000EP3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
00h0oooo000L0?ooo`030;md0000M;l0oooo0380oooo00<000000?ooo`3oool0@03oool00`000000
oooo0?ooo`030?ooo`D000001@3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool03P3oool001L0oooo00H0omnL04P0001d_ol0okmd0000001d_old0?oo
o`030000003oool0oooo0400oooo00<000000?ooo`3oool0103oool3000000030?ooo`000000oooo
00<0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
00h0oooo000G0?ooo`040?oOW01800000000001d_cH0oooo00<000000?ooo`3oool0B03oool01@00
0000oooo0?ooo`3oool0000005X0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>
0?ooo`00D@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo05L0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool3000004h0oooo00<000000?ooo`3o
ool0EP3oool00`000000oooo0?ooo`2P0?ooo`800000403oool00540oooo00<000000?ooo`3oool0
CP3oool00`000000oooo0?ooo`1F0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
3P3oool00540oooo00<000000?ooo`3oool0CP3oool2000000D0oooo0P0000070?ooo`030000003o
ool0oooo04H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00@`3oool0
102oM00000000000001d_ol:0?ooo`030000003oool0oooo04h0oooo00<000000?ooo`0000000`3o
ool4000000H0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo00h0oooo00120?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo
0?ooo`1>0?oooa400000B03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo
00110?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?oo
o`1>0?ooo`060000003oool0oooo0?ooo`000000oooo1@0000050?ooo`030000003oool0oooo04H0
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00@@3oool00`3ooml0W4P0
04RLg`020?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool0CP3oool01@000000
oooo0?ooo`3oool000000080oooo0`0000060?ooo`030000003oool0oooo04H0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0>0?ooo`00@@3oool00`3ooml0W4P0001d_`030?ooo`03
0=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`020?ooo`04
0000003oool0oooo000000L0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo00h0oooo00120?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018
W=l01`3ooooo000005000000403oool00480oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`070?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0C`3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo00120?ooo`030?oOW0180000B9cO00<0oooo00<0
_g@00018W03Oool01P3oool00`000000oooo0?ooo`1>0?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0CP3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo0012
0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`000000oooo0?oo
o`170?ooo`030000003oool0oooo00<0oooo0`0000060?ooo`800000C`3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo00h0oooo00130?ooo`030?oOW0180000M;oo0080oooo00<0gia8
001d_`3oool01P3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo00<0oooo10000005
0?ooo`800000C`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo00140?oo
o`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`170?oooa00000000<0
oooo0000003oool0C@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00h0oooo0015
0?ooo`040=nLB00000000000001d_`P0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?oo
o`020?ooo`H00000103oool010000000oooo0?ooo`00001=0?ooo`030000003oool0oooo09T0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?oo
o`170?ooo`030000003oool0oooo00<0oooo0`0000060?ooo`050000003oool0oooo0?ooo`000000
C03oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
3P3oool00540oooo00<000000?ooo`3oool0CP3oool00`000000oooo0?ooo`0:0?ooo`030000003o
ool0oooo04T0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00h0oooo001A0?ooo`<00000G03oool00`000000oooo0?ooo`180?ooo`030000003oool0
oooo09H0oooo00<000000?ooo`3oool01`3oool200000100oooo001A0?ooo`030000003oool0oooo
05d0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0FP3oool01@000000
oooo0?ooo`3oool000000080oooo0P0000070?ooo`030000003oool0oooo03X0oooo00<000000?oo
o`3oool0U03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00h0oooo001A0?ooo`03
0000003oool0oooo05X0oooo00<000000?ooo`3oool00P3oool6000000D0oooo00<000000?ooo`3o
ool0>P3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool03P3oool00540oooo00<000000?ooo`3oool0FP3ooolA000003`0oooo00<000000?ooo`3oool0
R`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
3P3oool00540oooo00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`030?ooo`D000001@3o
ool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool00P3o
ool6000000@0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0>0?ooo`00D@3oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo00<0oooo1@0000050?ooo`030000003oool0
oooo03X0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`030?ooo`@00000103oool2
000000h0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0FP3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0>P3oool00`00
0000oooo0?ooo`230?oooa0000003P3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool300000600
oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
03@0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`030?ooo`@000000P3oool01000
0000oooo0?ooo`00000=0?ooo`800000403oool00540oooo00<000000?ooo`3oool0H03oool01P00
0000oooo0?ooo`3oool000000?ooo`D000001@3oool00`000000oooo0?ooo`0d0?ooo`030000003o
ool0oooo08<0oooo00<000000?ooo`3oool00P3oool7000000<0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0
oooo0080oooo1P0000050?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0R`3oool0
0`000000oooo0000000C0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo
0600oooo4@00000f0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool04P3oool00`00
0000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo00<0
oooo100000060?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0R`3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo0600oooo
00<000000?ooo`3oool00`3oool5000000D0oooo00<000000?ooo`3oool0=03oool00`000000oooo
0?ooo`2:0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool03P3oool00540oooo00<0
00000?ooo`3oool0J@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo08L0oooo1P00
000D0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo06X0oooo00<00000
0?ooo`3oool0>P3oool00`000000oooo0?ooo`220?ooo`030000003oool0oooo00<0oooo0`000006
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?oo
o`3oool0J03oool010000000oooo0?ooo`0000020?ooo`D000001@3oool00`000000oooo0?ooo`0/
0?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool00P3oool5000000D0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool3000006P0oooo00@000000?oo
o`3oool0oooo0P000000103oool000000000000000060?ooo`030000003oool0oooo02`0oooo00<0
00000?ooo`3oool0PP3oool@000000h0oooo0P00000@0?ooo`00D@3oool00`000000oooo0?ooo`1R
0?ooo`D0000000<0oooo0000003oool0103oool5000000D0oooo00<000000?ooo`3oool0;03oool0
0`000000oooo0?ooo`220?ooo`040000003oool0000000000080oooo100000050?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0I03oool0
10000000oooo0?ooo`3ooolA000002h0oooo00<000000?ooo`3oool0PP3oool2000000D0oooo0P00
00060?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03P3oool00540oooo00<00000
0?ooo`3oool0G03oool00`000000oooo0?ooo`040?ooo`<000000P3oool01P000000oooo0?ooo`3o
ool000000?ooo`@000001P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0880oooo
00<000000?ooo`3oool00P3oool6000001@0oooo00<000000?ooo`3oool03P3oool00540oooo00<0
00000?ooo`3oool0G03oool00`000000oooo0?ooo`030?ooo`D0000000<0oooo0000003oool00P3o
ool00`000000oooo0?ooo`03000000H0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?oo
o`210?ooo`030000003oool0000000@000006@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool0
0`000000oooo0?ooo`1L0?oooa40000000<0oooo0000000000000`00000=0?ooo`D000008P3oool0
0`000000oooo0?ooo`1m0?ooo`040000003oool0oooo000000<0oooo0P0000070?ooo`030000003o
ool0oooo0140oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0G03oool0
0`000000oooo0?ooo`030?ooo`D000001@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0
000000<0oooo100000060?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool03P3oool0
0540oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`040?ooo`<000001P3oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool30000
00H0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`1m0?oooa4000004`3oool00`00
0000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`1T0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`D000001@3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo05h0oooo1P00000I0?ooo`030000003oool0oooo00<0
oooo1@0000050?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool03P3oool00540oooo
0`00001R0?ooo`D000003`3oool200000080oooo4@00000L0?ooo`030000003oool0oooo0640oooo
00<000000?ooo`3oool01P3oool6000000`0oooo0P0000060?ooo`<000001P3oool00`000000oooo
0?ooo`0@0?ooo`800000403oool00540oooo00<000000?ooo`3oool0N03oool00`000000oooo0000
00050?ooo`D000001@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo05T0oooo00<0
00000?ooo`3oool0103oool2000000<0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool2000000030?ooo`000000oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00h0oooo
001A0?ooo`030000003oool0oooo07T0oooo0P0000060?ooo`<00000103oool5000001X0oooo00<0
00000?ooo`3oool0F@3oool00`000000oooo0?ooo`030?ooo`@000000P3oool01P000000oooo0?oo
o`3oool000000?ooo`<000001P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00L0
oooo1@00000I0?ooo`030000003oool0oooo00h0oooo00110?ooo`060?oOW0180000000000000000
B9`0gooo2P3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo05T0oooo700000030?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`3oool03P3oool00440oooo00H0okmd0000001d
_ol0oooo0?oOW018M;l:0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool00P3oool5
000000@0oooo0P0000070?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`030?ooo`D0000000<0oooo0000003oool00P3oool00`000000oooo00000003
000000D0oooo00@000000?ooo`3oool00000:@3oool00`000000oooo0?ooo`0>0?ooo`00AP3oool0
0`3o_g@0B9cO0?ooo`080?ooo`030000003oool0oooo07`0oooo0P0000040?ooo`030000003oool0
oooo00<0oooo100000060?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0F@3oool0
0`000000oooo0?ooo`040?ooo`<000000P3oool01P000000oooo0?ooo`3oool000000?ooo`<00000
1P3oool00`000000oooo0000000Z0?ooo`030000003oool0oooo00h0oooo00160?ooo`030?oo_`1d
07@0_ooo00P0oooo00<000000?ooo`3oool0OP3oool010000000oooo0?ooo`3ooolA000001@0oooo
00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`00
00000`3oool010000000oooo0?ooo`0000070?ooo`800000:`3oool00`000000oooo0?ooo`0>0?oo
o`00AP3oool00`3ooml0W4P007Boo`080?ooo`030000003oool0oooo07l0oooo00@000000?ooo`3o
ool000001@3oool5000000D0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1Q0?oo
o`030000003oool0oooo00H0oooo1P0000030?ooo`030000003oool0oooo02/0oooo00<000000?oo
o`3oool03P3oool004H0oooo00<0oono07@00018W=l0203oool4000007l0oooo00<000000?ooo`00
00001P3oool3000000H0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1N0?ooo`H0
00003P3oool00`000000oooo0?ooo`0Z0?ooo`<00000403oool004H0oooo00<0okmd0000001d_ol0
203oool00`000000oooo0?ooo`210?ooo`8000001`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool0L03oool2000002l0oooo00<000000?ooo`3oool0
3P3oool004<0oooo00H0omoO0=ooo`3o_g@000000000M02oool80?ooo`030000003oool0oooo08<0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo06l0
oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0>0?ooo`00@`3oool01@3ookl0M000
0000000004P0W=oo00T0oooo00<000000?ooo`3oool0Q03oool200000080oooo2P00000E0?ooo`03
0000003oool0oooo06h0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0>0?ooo`00
A03oool00`3OW9`0_ooo0?ooo`0:0?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo06`0oooo0P00000c0?ooo`030000
003oool0oooo00h0oooo00140?ooo`030?oOW01dW=l0oooo00X0oooo00<000000?ooo`3oool0Q`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0J`3o
ool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00h0oooo00150?ooo`050;md00000000
000000000018W=l01`3oool00`000000oooo0?ooo`260?ooo`040000003oool00000000000<0oooo
0P0000070?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0J@3oool2000003H0oooo
00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0QP3oool01P000000oooo0?oo
o`3oool000000?ooo`@000001P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo06P0
oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo
0?ooo`260?oooa400000403oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo03L0oooo
00<000000?ooo`3oool03P3oool00540oooo0`0000260?ooo`030000003oool0oooo00<0oooo1@00
00050?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0E`3oool5000000T0oooo0P00
000i0?ooo`800000403oool00540oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`04
0?ooo`<000001P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo05T0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo00h0oooo001A0?oo
o`030000003oool0oooo08H0oooo00<000000?ooo`3oool01@3oool010000000oooo000000000004
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`06
0?ooo`800000?@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`2>0?oo
o`040000003oool0oooo0?ooo`8000003@3oool7000001d0oooo1P00000h0?ooo`030000003oool0
oooo00@0oooo0P0000040?ooo`D00000=P3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`00
0000oooo0?ooo`2>0?ooo`030000003oool0oooo00<0oooo0P00000=0?ooo`050000003oool0oooo
0?ooo`000000803oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo00@0oooo0`000004
0?ooo`030000003oool0000000H0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0>
0?ooo`00D@3oool00`000000oooo0?ooo`2<0?ooo`H00000103oool2000000/0oooo00D000000?oo
o`3oool0oooo0000000P0?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool00`3oool5
000000030?ooo`00000000000080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo09P0oooo0P0000090?oo
o`050000003oool0oooo0?ooo`000000803oool00`000000oooo0?ooo`0/0?oooaD000001@3oool2
000000L0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`00
0000oooo0?ooo`2J0?ooo`<000001@3oool300000080oooo00D000000?ooo`3oool0oooo0000000L
0?ooo`030000003oool0oooo02l0oooo0P0000050?ooo`D00000103oool2000000@0oooo10000006
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool03P3oool00540oooo0`00002K0?oo
o`D000000P3oool4000000030?ooo`000000oooo0080oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0_0?ooo`H0000000<0oooo0000000000000P0000050?oooa400000;`3oool20000
0100oooo001A0?ooo`030000003oool0oooo09`0oooo4000000D0?ooo`030000003oool0oooo00@0
oooo0P0000060?ooo`030000003oool0oooo02L0oooo0`0000000`3oool000000?ooo`0300000003
0?ooo`000000oooo00@0oooo0P0000050?ooo`D000001@3oool00`000000oooo0?ooo`0^0?ooo`03
0000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo09`0oooo0P0000020?ooo`H00000
00<0oooo0000003oool00P3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00<0oooo
100000050?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool00`3oool6000000030?oo
o`0000000000008000001@3oool2000000H0oooo0`0000060?ooo`030000003oool0oooo02h0oooo
00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0W03oool2000000@0oooo1`00
00020?ooo`030000003oool0oooo0180oooo4000000R0?ooo`8000001@3oool5000000030?ooo`00
0000000000<000001@3oool2000000L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`0^0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo09D0oooo00<00000
0?ooo`3oool0103oool3000000@0oooo1@00000H0?ooo`030000003oool0oooo00<0oooo1P000003
0?ooo`030000003oool0oooo0200oooo0P0000000`3oool000000?ooo`0C000000P0oooo00<00000
0?ooo`3oool0=P3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo0?ooo`2E0?oo
o`030000003oool0oooo00<0oooo1@0000030?ooo`030000003oool0000000<000003P3oool60000
00<0oooo00<000000?ooo`3oool0103oool2000000H0oooo00<000000?ooo`0000001@00000K0?oo
oa80000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
03H0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0U@3ooolC00000003
0?ooo`000000000000<00000303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool06P3oool4000000030?ooo`000000000000l000000`3oool00`000000
oooo0?ooo`040?ooo`D00000=P3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo
0?ooo`2E0?ooo`030000003oool0oooo00<0oooo1@0000000`3oool0000000000005000000D0oooo
0`00000:0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0G0?ooo`<000000P3oool2000000030?ooo`000000oooo0080oooo1`0000001P3oool00000
0?ooo`000000oooo000004D0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3o
ool0U@3oool00`000000oooo0?ooo`040?ooo`<000001@3oool4000000D0oooo00<000000?ooo`3o
ool00`0000070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`0D0?ooo`<00000203oool00`000000oooo0?ooo`020?ooo`D000000P3oool01@000000
oooo0?ooo`3oool0000004D0oooo00<000000?ooo`3oool03P3oool00540oooo0`00002M0?oooa40
00001@3oool500000080oooo00<000000?ooo`3oool00`3oool2000000D0oooo0`0000060?ooo`03
0000003oool0000000@00000403oool3000000@0oooo1@0000020?ooo`030000003oool0oooo0080
oooo0P0000000`3oool000000?ooo`020?ooo`050000003oool0oooo0?ooo`000000A03oool20000
0100oooo001A0?ooo`030000003oool0oooo09d0oooo0P0000050?ooo`@000001@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo1P0000020?ooo`8000000`3oool7000000@0
oooo00<000000?ooo`3oool00P0000060?ooo`030000003oool0oooo00D0oooo100000090?ooo`03
0000003oool0oooo00L0oooo0P0000000`3oool000000?ooo`05000004L0oooo00<000000?ooo`3o
ool03P3oool00540oooo00<000000?ooo`3oool0W@3oool2000000H0oooo0`0000050?ooo`H00000
1@3oool400000080oooo4`000000103oool000000000000000050?ooo`T000003@3oool00`000000
oooo0?ooo`040?ooo`L00000C@3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool00`000000oooo
0?ooo`2M0?ooo`030000003oool0oooo00H0oooo0P0000080?ooo`030000003oool0000000l00000
00<0oooo0000000000005@0000000`3oool000000?ooo`0C0?ooo`030000003oool0oooo00H0oooo
1@00001=0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo09/0oooo1@00
00060?ooo`800000203oool00`000000oooo000000050?ooo`@00000103oool2000000D0oooo0`00
00000`3oool000000000000?000001@0oooo00<000000?ooo`3oool0203oool5000004/0oooo00<0
00000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0YP3oool2000000P0oooo00<00000
0?ooo`0000001P3oool2000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
000000<0oooo1P0000050?ooo`8000002`3oool00`000000oooo0?ooo`040?ooo`<000001P3oool0
0`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00h0oooo001A0?ooo`030000003oool0oooo
0:<0oooo1P0000060?ooo`8000001`3oool00`000000oooo0000000=0?ooo`030000003oool0oooo
008000001@3oool5000000@0oooo0P00000;0?ooo`030000003oool0oooo00<0oooo1@0000050?oo
o`030000003oool0oooo0500oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3o
ool0Y`3oool2000000D0oooo100000060?ooo`030000003oool0000000d0oooo00@000000?ooo`3o
ool0oooo4@00000;0?oooa400000DP3oool00`000000oooo0?ooo`0>0?ooo`00D@3oool300000:L0
oooo4P0000000`3oool000000?ooo`0<0?ooo`050000003oool0oooo0?ooo`0000001@3oool50000
00D0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`D000001@3oool00`00
0000oooo0?ooo`1?0?ooo`800000403oool00540oooo00<000000?ooo`3oool0Y`3oool2000000D0
oooo1@0000050?ooo`030000003oool0000000/0oooo1@0000000`3oool000000?ooo`050?ooo`<0
00001P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0oooo0`0000060?ooo`03
0000003oool0oooo0500oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0
Y`3oool2000000H0oooo0`0000050?ooo`H000002@3oool6000000L0oooo0P00000K0?ooo`030000
003oool0oooo05P0oooo00<000000?ooo`3oool03P3oool00540oooo00<000000?ooo`3oool0Y`3o
ool2000000L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0F0?ooo`H000006@3o
ool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo00h0oooo000h0?ooo`030?oo_`1d0000
00000080000000<0B9cO0?ooo`3oool00`3oool0102oM00000000000001d_ol:0?ooo`030000003o
ool0oooo0:L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo01T0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0>0?ooo`00>P3oool0
0`3o_g@0000007Boo`050?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo
0?ooo`2W0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool08@3oool00`000000oooo
0?ooo`0I0?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool03P3oool003X0oooo00<0
oooO09a80000M;l0103oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0
00000?ooo`3oool0Y`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01l0oooo1@00
000G0?ooo`D00000F03oool00`000000oooo0?ooo`0>0?ooo`00>`3oool00`3OW4P004RL0=ooo`03
0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?oo
o`2W0?ooo`030000003oool0oooo00@0oooo1@00002H0?ooo`030000003oool0oooo00h0oooo000k
0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`07
0?ooo`030000003oool0oooo0:L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0>
0?ooo`00>`3oool00`3ookl0M00007Boo`040?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P0
0018W=l01`3oool400000:H0oooo00<000000?ooo`3oool0W`3oool300000100oooo000l0?ooo`03
0;md0000M;l0oooo00<0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`070?ooo`03
0000003oool0oooo00/0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`03
0000003oool0oooo04/0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00h0oooo000l0?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018W=l0
0`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
303oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
3P3oool003`0oooo00<0omnL04P0B02Lgol00`3oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo00H0ooooo`00001@00000100oooo000l0?ooo`030?oo_`1d0000M;oo00@0oooo00<0
omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`3o0?ooofH0oooo000m0?ooo`030;md0000M;l0
oooo00@0oooo00H0gia80000M02oool0oooO09a80018W=oo0?ooofL0oooo000k0?ooo`050?oo_`1d
000000000000M02oool01@3oool0103OW4P0000000000000M;oo0?ooofP0oooo003o0?oook40oooo
003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.60902, -19.7727, \
0.00959748, 0.175747}}]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Nonlinear least squares", "Section"],

Cell[TextData[{
  "\tOften the model cannot be transformed into a linear function of its \
parameters and we must employ more sophisticated algorithms to search for the \
minimum in ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Chi]\^2\)[a]\)]],
  ".  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides a very powerful function for this purpose, ",
  StyleBox["NonlinearRegress", "InlineInput",
    FontWeight->"Bold"],
  ", but a serious experimentalist must always be prepared to develop a \
custom search procedure whenever the problem at hand does not abide by the \
limitations of this function or other commercial software.  For example, \
uncertainties in the independent variable are not considered by ",
  StyleBox["NonlinearRegress", "InlineInput",
    FontWeight->"Bold"],
  ".  Furthermore, it is more instructive to write our routine than to use a \
black box, even if that black box provides more sophisticated output.  \
Therefore, we develop and implement a nonlinear least squares algorithm \
before discussing the built-in version."
}], "Text"],

Cell[CellGroupData[{

Cell["Levenberg-Marquardt algorithm", "Subsection"],

Cell[TextData[{
  "\tThe basic problem is to locate the minimum in a function of several \
variables, ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Chi]\^2\)[a]\)]],
  ", which can be described as a surface in a multidimensional space.  Many \
strategies and programs can be found in the literature, but we will focus \
upon the Levenberg-Marquardt algorithm \[LongDash] it is relatively simple \
but I have found that it provides a very effective method for a wide variety \
of physics problems.  This algorithm uses a single parameter, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  ", to smoothly combine two distinct strategies.  Imagine that you are a \
hiker in a mountain range seeking the comfort of a lodge locating in the \
deepest valley.  If we start our search at location on the ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " surface with a steep slope, the most effective strategy would be to head \
directly downhill.  On the other hand, if our current location is already \
near a minimum the slope is small and the gradient search becomes \
inefficient.  On the other hand, the shape of the surface near a minimum can \
be well approximated by parabolic function of all variables and we can use \
that information to locate the minimum accurately.  Iteration is still needed \
because we can only approximate the derivatives at the minimum by those at \
our current location, but the number of iterations required to obtain the \
desired precision is much smaller than needed for the gradient method.  The \
combination of these methods, gradient where the slope is large and parabolic \
where it is small, will almost always find a minimum from almost any starting \
point.  However, we cannot guarantee that this minimum is actually the global \
minimum that we seek \[LongDash] the ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " surface for a multidimensional search can be quite complicated and many \
contain many local minima.  If the reduced chi-square for our minimum is \
substantially greater than unity, we should probably vary the starting \
conditions and repeat the fit several times to determine whether there is a \
deeper minimum.  Although we will not implement this extension of the method, \
the most robust methods often perform fits on a grid of initial conditions in \
order to maximize the chances of finding the true global minimum.  If we \
understand the physics of our problem relatively well already and can obtain \
reasonably good starting conditions by combining insight with visual \
inspection of the data, the Levenberg-Marquardt algorithm will usually \
provide the correct minimum without more laborious testing of local minima.  \
I have found that it is usually fairly tolerant of poor starting conditions \
\[LongDash] they should be reasonable but need not be especially precise."
}], "Text"],

Cell[TextData[{
  "\tLet ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\_0\%2 = \(\[Chi]\^2\)[a\^\((0)\)]\)]],
  " represent the value of chi-square at the starting point ",
  Cell[BoxData[
      \(TraditionalForm\`a\^\((0)\) = {a\_\[Mu]\%\((0)\), \[Mu] = 1, m}\)]],
  ".  The shape of the surface can be approximated by the Taylor expansion "
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Chi]\^2\  \[TildeTilde] \ \[Chi]\_0\%2\  + \ \
\[Sum]\+\[Nu]\[PartialD]\[Chi]\_0\%2\/\[PartialD]a\_\[Nu]\ \[CapitalDelta]\
\[VeryThinSpace]a\_\[Nu]\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\_\[Nu] = 
        a\_\[Nu] - a\_\[Nu]\%\((0)\)\)]],
  " is the displacement in the parameter with index ",
  Cell[BoxData[
      \(TraditionalForm\`\[Nu]\)]],
  " relative to its initial value.  If we assume that this approximation is \
valid for a sufficiently large region to include a minimum, we could locate \
that minimum by solving the ",
  StyleBox["normal equations",
    FontSlant->"Italic"]
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[PartialD]\[Chi]\^2\/\[PartialD]a\_\[Mu] \[Equal] 
      0\ \ \[DoubleLongRightArrow]\ \ \ \ \(\[PartialD]\[Chi]\_0\%2\/\
\[PartialD]a\_\[Mu]\)\  + \ \[Sum]\+\[Nu]\[PartialD]\^2 \[Chi]\_0\%2\/\(\
\[PartialD]a\_\[Mu] \[PartialD]a\_\[Nu]\)\ \[CapitalDelta]\[VeryThinSpace]a\_\
\[Nu]\  \[Equal] \ 0\)], "DisplayFormula"],

Cell["It is convenient to define", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Beta]\_\[Mu]\  = \ \(\(-\(1\/2\)\) \[PartialD]\[Chi]\
\_0\%2\/\[PartialD]a\_\[Mu]\  = \ \[Sum]\+\(i = 1\)\%N\(\( w\_i\)(
              y - y\&_\_i)\) \[PartialD]y\&_\_i\/\[PartialD]a\_\[Mu]\)\)], \
"DisplayFormula"],

Cell["proportional to the gradient vector and", "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Alpha]\_\(\[Mu], \[Nu]\)\  = \ \(\(1\/2\) \
\[PartialD]\^2 \[Chi]\_0\%2\/\(\[PartialD]a\_\[Mu] \[PartialD]a\_\[Nu]\)\  = \
\ \[Sum]\+\(i = 1\)\%N\( 
                w\_i\)(\(\[PartialD]y\&_\_i\/\[PartialD]a\_\[Mu]\) \
\[PartialD]y\&_\_i\/\[PartialD]a\_\[Nu] - \((y\_i - 
                    y\&_\_i)\) \[PartialD]\^2 y\&_\_i\/\(\[PartialD]a\_\[Mu] \
\[PartialD]a\_\[Nu]\))\)\)], "DisplayFormula"],

Cell[TextData[{
  "as the ",
  StyleBox["curvature matrix,",
    FontSlant->"Italic"],
  " such that the normal equations take the form"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Beta]\  \[Equal] \ \[Alpha]\ \[CapitalDelta]\
\[VeryThinSpace]a\ \ \ \[DoubleLongRightArrow]\ \ \[CapitalDelta]\
\[VeryThinSpace]a\  = \ \[CurlyEpsilon]\ \[Beta]\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] = \[Alpha]\^\(-1\)\)]],
  " is the ",
  StyleBox["error matrix",
    FontSlant->"Italic"],
  ".  Ideally, the new parameters"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`a\  = \ 
      a\^\((0)\) + \[CapitalDelta]\[VeryThinSpace]a\)], "DisplayFormula"],

Cell[TextData[{
  "should be closer to the minimum than the initial parameters, but are not \
likely to be exact because our representation of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " was only approximate.  Even if the direction of ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\)]],
  " is good, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " might actually increase if the step is too large.  Therefore, we should \
compare ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Chi]\^2\)[a]\)]],
  " with ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Chi]\^2\)[a\^\((0)\)]\)]],
  " and reduce the step size if ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " increases.  If ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " decreases, we replace ",
  Cell[BoxData[
      \(TraditionalForm\`a\^\((0)\)\)]],
  " with ",
  Cell[BoxData[
      \(TraditionalForm\`a\)]],
  " and repeat the procedure, treating the new location as the starting point \
for the next iteration."
}], "Text"],

Cell[TextData[{
  "\tThe Levenberg-Marquardt algorithm improves upon this procedure by \
modifying the curvature matrix.  If ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha]\)]],
  " happened to be a diagonal matrix, then each ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\_\[Mu]\)]],
  " would only depend upon the corresponding derivative ",
  Cell[BoxData[
      \(TraditionalForm\`\[Beta]\_\[Mu]\)]],
  ", with a scale factor determined by ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon]\_\(\[Mu], \[Mu]\) \[Rule] 
        1/\[Alpha]\_\(\[Mu], \[Mu]\)\)]],
  ".  The search path then becomes similar to the gradient search.  This \
modification is accomplished by the simple expedient of multiplying the \
diagonal elements of the curvature matrix by a scale factor"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Alpha]\&~\_\(\[Mu], \[Nu]\) = \ \((1 + \[Lambda]\ \
\[Delta]\_\(\[Mu], \[Nu]\))\) \[Alpha]\_\(\[Mu], \[Nu]\)\ , \ \ \ \
\[CurlyEpsilon]\&~ = \ \[Alpha]\&~\^\(-1\), \ \ \ \ \[CapitalDelta]\
\[VeryThinSpace]a\  = \ \[CurlyEpsilon]\&~\ \[Beta]\)], "DisplayFormula"],

Cell[TextData[{
  "If ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " is large the diagonal dominates and the search is of gradient type with \
relatively small step, while for small ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " the search is essentially parabolic.  Assuming that the user will supply \
reasonably good initial conditions, we start with small ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  ", say ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda] = 0.001\)]],
  ".  Each time we evaluate ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\)]],
  " we adjust ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  ": if ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " decreases we take the step ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\)]],
  " and reduce ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " by a factor of 10, but if ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " increases we skip that step, increase ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " by a factor of 10, and try again.  Increasing ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " reduces the step size and tends to align its direction with the \
gradient."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Covariance and correlation matrices", "Subsection"],

Cell[TextData[{
  "\tRigorous analysis of the uncertainties in fitted parameters becomes \
rather difficult for nonlinear least squares because we do not have an \
analytical relationship between the fitted parameters and the data with their \
uncertainties.  Nor can one be certain that the fitted parameters are \
distributed in a Gaussian fashion.  The distribution of fitted parameters can \
be ",
  StyleBox["simulated",
    FontSlant->"Italic"],
  " by pretending that the results obtained by analyzing an experimental data \
set are correct and using those parameters, together with a random number \
generator, to produce many sets of pseudodata containing the same set of \
measurements as the experimental data but with random fluctuations in the \
results.  The distribution of fluctuations in the parameters fitted to \
pseudodata can then by used to analyze the uncertainties in and the \
correlations between fitted parameters.  However, this type of analysis is \
too laborious and too sophisticated for the present purposes.  Simpler \
estimates can be obtained by assuming that the fitting function can be \
linearized within a small region of the parameter space that is centered upon \
",
  Cell[BoxData[
      \(TraditionalForm\`a\&^\)]],
  "."
}], "Text"],

Cell["\<\
\tBeing optimistic, we assume that the uncertainties in the fitted parameters \
are small and that their values are close to the true values.  We can then \
linearize the fitting function according to\
\>", "Text"],

Cell[BoxData[
    \(TraditionalForm\`y[x; a]\  \[TildeTilde] \ 
      y[x; a\&^]\  + \ \(\(\(\ \)\(\[Sum]\)\)\+\[Mu]\) \(\((\[PartialD]y\/\
\[PartialD]a\_\[Mu])\)\_\(a = 
                a\&^\)\) \[CapitalDelta]\[VeryThinSpace]a\_\[Mu]\)], \
"DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a = a - a\&^\)]],
  ".  If we were to fit other experimental data sets from the same parent \
distribution using this linearized version of the model, we would expect the \
deviations ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[VeryThinSpace]a\_\[Mu]\)]],
  " to be normally distributed about 0 and could interpret the variance in \
the distribution of deviations as the standard error in ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Mu]\)]],
  ".  However, we already know that the variance in the parameters for linear \
least-squares is given by"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\[Sigma][\(a\&^\)\_\[Mu]]\^2\  = \ \[CurlyEpsilon]\_\(\
\[Mu], \[Mu]\)\)], "DisplayFormula"],

Cell[TextData[{
  "and expect that this relationship remains valid whenever ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\)]],
  " is small enough to employ the linear approximation to the fitting \
function."
}], "Text"],

Cell[TextData[{
  "\tSimilarly, near the minimal ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " we expand"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`\(\[Chi]\^2\)[
          a]\  \[TildeTilde] \ \(\[Chi]\^2\)[
            a\&^]\  + \ \(1\/2\) \(\[Sum]\+\(\[Mu], \
\[Nu]\)\(\((\[PartialD]\^2 \[Chi]\^2\/\(\[PartialD]a\_\[Mu] \[PartialD]a\_\
\[Nu]\))\)\_\(a = 
                      a\&^\)\) \[CapitalDelta]\[VeryThinSpace]\(a\_\[Mu]\) \
\[CapitalDelta]\[VeryThinSpace]a\_\[Nu]\)\  = \ \(\[Chi]\^2\)[
          a\&^]\  + \[Sum]\+\(\[Mu], \[Nu]\)\ \[CapitalDelta]\[VeryThinSpace]\
\(a\_\[Mu]\) \(\[Alpha]\_\(\[Mu], \[Nu]\)\) \
\[CapitalDelta]\[VeryThinSpace]a\_\[Nu]\)], "DisplayFormula"],

Cell[TextData[{
  "where the first-order contribution vanishes.  Remembering that ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " is closely related to the joint probability distribution for the \
experiment, we write"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`P\  \[Proportional] \ 
      Exp[\(-\(\[Chi]\^2\)[a]\)]\  \[Proportional] \ 
      Exp[\(-\(\[Sum]\+\(\[Mu], \[Nu]\)\ \[CapitalDelta]\[VeryThinSpace]\(a\_\
\[Mu]\) \(\[Alpha]\_\(\[Mu], \[Nu]\)\) \
\[CapitalDelta]\[VeryThinSpace]a\_\[Nu]\)\)]\)], "DisplayFormula"],

Cell[TextData[{
  "in the form of a multivariate Gaussian distribution and identify ",
  Cell[BoxData[
      \(TraditionalForm\`\[CurlyEpsilon] = \[Alpha]\^\(-1\)\)]],
  " as the covariance matrix.  Thus, the statistical correlation between \
fitted parameters is given by"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`C\_\(\[Mu], \[Nu]\)\  = \ \(\[LeftAngleBracket]\
\[CapitalDelta]\[VeryThinSpace]\(a\_\[Mu]\) \
\[CapitalDelta]\[VeryThinSpace]a\_\[Nu]\[RightAngleBracket]\/\@\(\
\[LeftAngleBracket]\((\[CapitalDelta]\[VeryThinSpace]a\_\[Mu])\)\^2\
\[RightAngleBracket] \[LeftAngleBracket]\((\[CapitalDelta]\[VeryThinSpace]a\_\
\[Nu])\)\^2\[RightAngleBracket]\)\ \  = \ \[CurlyEpsilon]\_\(\[Mu], \[Nu]\)\/\
\@\(\(\[CurlyEpsilon]\_\(\[Mu], \[Mu]\)\) \[CurlyEpsilon]\_\(\[Nu], \[Nu]\)\)\
\)\)], "DisplayFormula"],

Cell[TextData[{
  "where we leave the evaluation of the integrals as an exercise for the \
reader.   The diagonal elements have unit values while nondiagonal elements \
are limited to the range ",
  Cell[BoxData[
      \(TraditionalForm\`\((\(-1\), 1)\)\)]],
  ".  If two parameters are uncorrelated, the product ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]\(a\_\[Mu]\) \
\[CapitalDelta]\[VeryThinSpace]a\_\[Nu]\)]],
  " exhibits fluctuations with either positive or negative signs with equal \
likelihood such that ",
  Cell[BoxData[
      \(TraditionalForm\`C\_\(\[Mu], \[Nu]\)\)]],
  " averages to zero.  If a fluctuation in ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Mu]\)]],
  " is more likely to accompanied by a fluctuation in ",
  Cell[BoxData[
      \(TraditionalForm\`a\_\[Nu]\)]],
  " that has the same sign, then ",
  Cell[BoxData[
      \(TraditionalForm\`C\_\(\[Mu], \[Nu]\) > 0\)]],
  " represents a positive correlation between those parameters whereas a \
negative correlation with ",
  Cell[BoxData[
      \(TraditionalForm\`C\_\(\[Mu], \[Nu]\) < 0\)]],
  " indicates that a fluctuation in ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[ThinSpace]a\_\[Mu]\)]],
  " is more likely to be compensated by a fluctuation in ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalDelta]\[VeryThinSpace]a\_\[Nu]\)]],
  " with opposite sign. "
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Implementation", "Subsection"],

Cell[TextData[{
  "\tThe program below implements the Levenberg-Marquardt algorithm for \
problems in which there is a single independent variable with negligible \
uncertainty and a single dependent variable.  We also assume that derivatives \
of the fitting function with respect to each parameter can be obtained \
symbolically and we neglect the contributions of second derivatives to the \
curvature matrix.  This approximation is reasonable when the fit is good, \
such that the coefficient of the second derivative is small, fluctuates with \
random sign and tends to average to zero.  Furthermore, experience has shown \
that neglecting the second derivatives often improves the performance of the \
algorithm when the initial conditions are inaccurate or the data contain some \
points with greater than normal deviations (",
  StyleBox["outliers",
    FontSlant->"Italic"],
  ").  Finally, evaluation of second derivatives can be computationally \
expensive.  Therefore, it is customary to omit the second derivatives, but we \
provide an option for their inclusion.  In the interests of simplicity, we \
use a simple ",
  StyleBox["For", "InlineInput",
    FontWeight->"Bold"],
  " loop and have not attempted to minimize the number of evaluations of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Beta]\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha]\)]],
  " (re-evaluation is not needed when ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " is increased).  It should not be difficult to modify this procedure to \
accommodate multiple independent or dependent variables or to embed this \
procedure in a loop that accounts for uncertainties in the independent \
variable(s), but we leave those extensions as an exercise to the reader that \
needs them."
}], "Text"],

Cell[BoxData[{
    RowBox[{
      RowBox[{\(curfit::usage\), "=", 
        "\"\<curfit[data,f,var,par,start] performs nonlinear least-squares \
fitting using the Levenberg-Marquardt algorithm. The \!\(\*
StyleBox[\"data\",\nFontSlant->\"Italic\"]\) array should be a list of \!\(\*
StyleBox[\"{\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"x\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"y\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\",\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"dy\",\nFontSlant->\"Italic\"]\)\!\(\*
StyleBox[\"}\",\nFontSlant->\"Italic\"]\) for each measurement.  The fitting \
function \!\(\*
StyleBox[\"f\",\nFontSlant->\"Italic\"]\) is expressed in terms of the \
independent variable \!\(\*
StyleBox[\"var\",\nFontSlant->\"Italic\"]\) and a list of parameters \!\(\*
StyleBox[\"par\",\nFontSlant->\"Italic\"]\). The initial conditions are given \
as a list \!\(\*
StyleBox[\"start\",\nFontSlant->\"Italic\"]\).\>\""}], 
      ";", \(curfit::BadDim = "\<Bad dimensions in data array: `1`\>"\), 
      ";"}], "\[IndentingNewLine]", \(curfit::conv = "\<Warning: failed to \
converge.  Relative change = `1`; \[Lambda] = `2`\>";\), \
"\[IndentingNewLine]", 
    RowBox[{\(curfit::start = "\<Incorrect number of initial values.\>";\), 
      "\[IndentingNewLine]"}], "\[IndentingNewLine]", 
    RowBox[{\(Options[curfit] = {MaxNITS \[Rule] 10, Conv \[Rule] 0.001, 
            PrintIterations \[Rule] False, 
            IncludeSecondDerivatives \[Rule] False};\), 
      "\[IndentingNewLine]"}], "\[IndentingNewLine]", \(curfit[data_, f_, 
        var_, par_, start_, opts___Rule] := 
      Module[{npar, ndata, ndim, nits, a, da, \[Lambda], x, y, dy, w, yfit, 
          derivs, dmat, deriv2, dmat2, \[Beta], \[Alpha], \[CurlyEpsilon], 
          errors, corr, redchi, change, conv, mnits, i, j, oldchi, newchi, 
          printIterations, 
          includeSecondDerivatives}, \[IndentingNewLine]npar\  = \ 
          Length[par]; \[IndentingNewLine]\[IndentingNewLine] (*\ Extract\ x, 
          y, dy\ vectors\ from\ data\ \(\(array\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]{ndata, ndim} = 
          Dimensions[data]; \[IndentingNewLine]If[
          ndata \[LessEqual] npar || ndim < 2 || 
            ndim > 3, \((Message[curfit::BadDim, {ndata, ndim}]; 
            Return[])\)]; \[IndentingNewLine]If[
          Length[start] \[NotEqual] npar, \((Message[curfit::start]; 
            Return[])\)]; \[IndentingNewLine]If[
          ndim \[Equal] 2, \[IndentingNewLine]\(({x, y} = Transpose[data]; 
            w = Table[1, {ndata}])\), \[IndentingNewLine]\(({x, y, dy} = 
              Transpose[data]; 
            w = 1/dy\^2)\)]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Initialize\ derivatives\ and\ search\ \(\(parameters\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]{mnits, conv, printIterations, 
            includeSecondDerivatives} = \({MaxNITS, Conv, PrintIterations, 
                IncludeSecondDerivatives} /. {opts}\) /. 
            Options[curfit]; \[IndentingNewLine]\[IndentingNewLine]derivs = 
          Table[Simplify[
              D[f, par\[LeftDoubleBracket]i\[RightDoubleBracket]]], {i, 
              npar}]; \[IndentingNewLine]If[
          includeSecondDerivatives, \[IndentingNewLine]deriv2 = 
            Table[Simplify[
                D[f, par\[LeftDoubleBracket]i\[RightDoubleBracket], 
                  par\[LeftDoubleBracket]j\[RightDoubleBracket]]], {i, 
                npar}, {j, npar}]]; \[IndentingNewLine]change = 
          1. ; \[IndentingNewLine]\[Lambda] = 
          0.001; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Search\ for\ \(\(minimum\)\(.\)\)\ *) \[IndentingNewLine]\
\[IndentingNewLine]For[\((nits = 0; 
            a = start)\), \((nits \[LessEqual] mnits && 
              change > conv)\), \(nits++\), \[IndentingNewLine]yfit = \(f /. 
                Thread[par \[Rule] a]\) /. 
              var \[Rule] x; \[IndentingNewLine]oldchi = 
            w . \((y - yfit)\)\^2; \[IndentingNewLine]dmat = 
            Table[\(derivs /. Thread[par \[Rule] a]\) /. 
                var \[Rule] x\[LeftDoubleBracket]i\[RightDoubleBracket], {i, 
                ndata}]; \[IndentingNewLine]\[Beta] = \((w \((y - 
                      yfit)\))\) . dmat; \[IndentingNewLine]\[Alpha] = 
            Transpose[w\ dmat] . dmat; \[IndentingNewLine]If[
            includeSecondDerivatives, \[IndentingNewLine]dmat2 = 
              Table[\(deriv2 /. Thread[par \[Rule] a]\) /. 
                  var \[Rule] 
                    x\[LeftDoubleBracket]i\[RightDoubleBracket], {i, 
                  ndata}]; \[IndentingNewLine]\[Alpha] = \[Alpha] - \((w \((y \
- yfit)\))\) . dmat2]; \[IndentingNewLine]Do[\[Alpha]\[LeftDoubleBracket]i, 
                i\[RightDoubleBracket] = \((1 + \[Lambda])\) \[Alpha]\
\[LeftDoubleBracket]i, i\[RightDoubleBracket], {i, 
              npar}]; \[IndentingNewLine]\[CurlyEpsilon] = 
            PseudoInverse[\[Alpha]]; \[IndentingNewLine]da = \[CurlyEpsilon] \
. \[Beta]; \[IndentingNewLine]yfit = \(f /. 
                Thread[par \[Rule] \((a + da)\)]\) /. 
              var \[Rule] x; \[IndentingNewLine]newchi = 
            w . \((y - yfit)\)\^2; \[IndentingNewLine]If[
            newchi < oldchi, \((\[Lambda] = \[Lambda]/10; a = a + da; 
              change = \((oldchi - newchi)\)/oldchi)\), \[Lambda] = 
              10  \[Lambda]]; \[IndentingNewLine]If[printIterations, 
            Print[{oldchi, newchi, \[Lambda], change, 
                a + da}]];\[IndentingNewLine]]; \[IndentingNewLine]If\ [
          mnits > 1 && change > conv, 
          Message[curfit::conv, NumberForm[change, 2], 
            NumberForm[\[Lambda], 
              2]]]; \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Evaluate\ proper\ error\ \(\(matrix\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]\[Alpha] = 
          Transpose[w\ dmat] . dmat; \[IndentingNewLine]If[
          includeSecondDerivatives, \[IndentingNewLine]dmat2 = 
            Table[\(deriv2 /. Thread[par \[Rule] a]\) /. 
                var \[Rule] x\[LeftDoubleBracket]i\[RightDoubleBracket], {i, 
                ndata}]; \[IndentingNewLine]\[Alpha] = \[Alpha] \((w \((y - 
                        yfit)\))\) . 
                dmat2]; \[IndentingNewLine]\[CurlyEpsilon] = 
          PseudoInverse[\[Alpha]]; \[IndentingNewLine]errors = 
          Table[\@\[CurlyEpsilon]\[LeftDoubleBracket]i, \
i\[RightDoubleBracket], {i, npar}]; \[IndentingNewLine]corr = 
          Table[\[CurlyEpsilon]\[LeftDoubleBracket]i, \
j\[RightDoubleBracket]\/\(errors\[LeftDoubleBracket]i\[RightDoubleBracket] \
errors\[LeftDoubleBracket]j\[RightDoubleBracket]\), {i, npar}, {j, 
              npar}]; \[IndentingNewLine]yfit = \(f /. 
              Thread[par \[Rule] a]\) /. 
            var \[Rule] x; \[IndentingNewLine]redchi = 
          w . \((y - yfit)\)\^2/\((ndata - 
                npar)\); \[IndentingNewLine]\[IndentingNewLine] (*\ 
          Return\ results\ as\ replacement\ \(\(rules\)\(.\)\)\ \
*) \[IndentingNewLine]\[IndentingNewLine]{Parameters \[Rule] a, 
          Uncertainties \[Rule] errors, CorrelationMatrix \[Rule] corr, 
          ReducedChiSquare \[Rule] redchi, 
          NITS \[Rule] nits}\[IndentingNewLine]]\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Performance tests", "Subsection"],

Cell["\<\
Below we construct pseudodata for a radioactive decay experiment\
\>", "Text"],

Cell[BoxData[{
    \(\(Clear[t, a, \[Tau]];\)\), "\[IndentingNewLine]", 
    \(\(tdata = Range[0, 1, 0.02];\)\), "\[IndentingNewLine]", 
    \(\(p = {10\^3, 0.1};\)\), "\[IndentingNewLine]", 
    \(\(yfit[t_] := 
        p\[LeftDoubleBracket]1\[RightDoubleBracket] 
          Exp[\(-t\)/
              p\[LeftDoubleBracket]2\[RightDoubleBracket]];\)\), "\
\[IndentingNewLine]", 
    \(\(ydata = \(Random[PoissonDistribution[yfit[#]]] &\) /@ tdata // 
          N;\)\), "\[IndentingNewLine]", 
    \(\(dy = \@\(1.  + ydata\);\)\)}], "Input"],

Cell["\<\
and attempt to fit, with reasonable but inaccurate initial conditions, \
including second derivatives.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(results = 
      curfit[Transpose[{tdata, ydata, dy}], a\ Exp[\(-t\)/\[Tau]], 
        t, {a, \[Tau]}, {1500, 0.7}, MaxNITS \[Rule] 20, 
        PrintIterations \[Rule] True, 
        IncludeSecondDerivatives \[Rule] True]\)], "Input"],

Cell[BoxData[
    \({4.48087150539474`*^6, 610836.2978098851`, 0.0001`, 
      0.8636791309292245`, {73.34603062907604`, 
        0.34546746132221706`}}\)], "Print"],

Cell[BoxData[
    \({610836.2978098851`, 100055.04533857777`, 0.00001`, 
      0.8361999021713038`, {148.29530527005852`, 
        0.2921521923567454`}}\)], "Print"],

Cell[BoxData[
    \({100055.04533857777`, 19755.172154046384`, 1.0000000000000002`*^-6, 
      0.8025569616484949`, {138.0885503604192`, 
        0.2543158164162186`}}\)], "Print"],

Cell[BoxData[
    \({19755.172154046384`, 6071.485756100608`, 1.0000000000000002`*^-7, 
      0.6926634853517586`, {104.2591695131502`, 
        0.23676467448320937`}}\)], "Print"],

Cell[BoxData[
    \({6071.485756100608`, 4116.502386302978`, 1.0000000000000004`*^-8, 
      0.32199422815630746`, {47.481643737968994`, 
        0.2551864163889735`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4605.183697714892`, 1.0000000000000004`*^-7, 
      0.32199422815630746`, {40.74424955459294`, 
        0.3133455012655728`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4605.183884101856`, 1.0000000000000004`*^-6, 
      0.32199422815630746`, {40.744234680448145`, 
        0.31334551644213104`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4605.185747984247`, 0.000010000000000000004`, 
      0.32199422815630746`, {40.744085938140714`, 
        0.3133456682084759`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4605.204387525586`, 0.00010000000000000005`, 
      0.32199422815630746`, {40.74259847443547`, 
        0.31334718590926386`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4605.390854957113`, 0.0010000000000000005`, 
      0.32199422815630746`, {40.72771975302072`, 
        0.3133623666652128`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4607.262763936974`, 0.010000000000000005`, 
      0.32199422815630746`, {40.578523011614436`, 
        0.31351455004901113`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4626.73938535102`, 0.10000000000000006`, 
      0.32199422815630746`, {39.04450833321579`, 
        0.3150749878150635`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 4957.316135837111`, 1.0000000000000007`, 
      0.32199422815630746`, {18.01455731131506`, 
        0.3359248204563594`}}\)], "Print"],

Cell[BoxData[
    \({4116.502386302978`, 3375.5904713700525`, 0.10000000000000007`, 
      0.17998578535948262`, {179.99646251163955`, 
        0.16935134268871133`}}\)], "Print"],

Cell[BoxData[
    \({3375.5904713700525`, 1.849041862467656`*^8, 1.0000000000000007`, 
      0.17998578535948262`, {\(-4301.299589691637`\), 
        2.7614647912679304`}}\)], "Print"],

Cell[BoxData[
    \({3375.5904713700525`, 2757.330500068138`, 0.10000000000000007`, 
      0.18315609566553284`, {262.89768278434815`, 
        0.15385162749654493`}}\)], "Print"],

Cell[BoxData[
    \({2757.330500068138`, 3060.896292364004`, 1.0000000000000007`, 
      0.18315609566553284`, {662.8983527756957`, 
        0.035605571839804784`}}\)], "Print"],

Cell[BoxData[
    \({2757.330500068138`, 2091.3943643587863`, 0.10000000000000007`, 
      0.24151480415310947`, {374.72463556512963`, 
        0.14352613393917196`}}\)], "Print"],

Cell[BoxData[
    \({2091.3943643587863`, 896.5570451893893`, 0.010000000000000009`, 
      0.571311341147144`, {1058.1448432417894`, 
        0.007942812832394897`}}\)], "Print"],

Cell[BoxData[
    \({896.5570451893893`, 1368.1060462330502`, 0.10000000000000009`, 
      0.571311341147144`, {1054.6795597739667`, 
        0.12700409535780266`}}\)], "Print"],

Cell[BoxData[
    \({896.5570451893893`, 961.2554104161821`, 1.0000000000000009`, 
      0.571311341147144`, {1026.0116151537018`, 
        0.12399460969500183`}}\)], "Print"],

Cell[BoxData[
    \(curfit::"conv" \(\(:\)\(\ \)\) 
      "Warning: failed to converge.  Relative change = \!\(\"0.57\"\); \
\[Lambda] = \!\(\"1.\"\)"\)], "Message"],

Cell[BoxData[
    \({Parameters \[Rule] {674.709457459627`, 0.08651286625959878`}, 
      Uncertainties \[Rule] {\(\(0.`\)\(\[InvisibleSpace]\)\) + 
            2.921332835695056`*^-14\ \[ImaginaryI], \(\(0.`\)\(\
\[InvisibleSpace]\)\) + 3.5223025809255395`*^-6\ \[ImaginaryI]}, 
      CorrelationMatrix \[Rule] \
{{\(\(0.9999999999999998`\)\(\[InvisibleSpace]\)\) + 
              0.`\ \[ImaginaryI], \
\(\(0.9999999951178615`\)\(\[InvisibleSpace]\)\) + 
              0.`\ \[ImaginaryI]}, {\(\(1.0000000048821385`\)\(\
\[InvisibleSpace]\)\) + 
              0.`\ \[ImaginaryI], \
\(\(1.0000000000000002`\)\(\[InvisibleSpace]\)\) + 0.`\ \[ImaginaryI]}}, 
      ReducedChiSquare \[Rule] 18.297082554885495`, 
      NITS \[Rule] 21}\)], "Output"]
}, Closed]],

Cell["\<\
Although this search fails to converge, simply removing the contribution to \
the curvature matrix due to second-derivatives of the fitting function\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(results = 
      curfit[Transpose[{tdata, ydata, dy}], a\ Exp[\(-t\)/\[Tau]], 
        t, {a, \[Tau]}, {1500, 0.7}, MaxNITS \[Rule] 20, 
        PrintIterations \[Rule] True]\)], "Input"],

Cell[BoxData[
    \({4.48087150539474`*^6, 27262.01340901739`, 0.0001`, 
      0.9939159127022064`, {\(-1248.62130457911`\), 
        0.5961692319242708`}}\)], "Print"],

Cell[BoxData[
    \({27262.01340901739`, 3726.1806246232572`, 0.00001`, 
      0.8633196833733946`, {105.80059244974345`, \
\(-0.331152774046227`\)}}\)], "Print"],

Cell[BoxData[
    \({3726.1806246232572`, 5.645468746022697`*^12, 0.0001`, 
      0.8633196833733946`, {821.025775016333`, \(-0.1370707595296422`\)}}\)], \
"Print"],

Cell[BoxData[
    \({3726.1806246232572`, 5.730206991694716`*^12, 0.001`, 
      0.8633196833733946`, {820.7866258967915`, \
\(-0.13691721185604516`\)}}\)], "Print"],

Cell[BoxData[
    \({3726.1806246232572`, 6.65926006835331`*^12, 0.01`, 
      0.8633196833733946`, {818.4055509585006`, \
\(-0.1353887782118104`\)}}\)], "Print"],

Cell[BoxData[
    \({3726.1806246232572`, 3.434299664110102`*^13, 0.1`, 
      0.8633196833733946`, {795.5914425611533`, \
\(-0.12077813834841239`\)}}\)], "Print"],

Cell[BoxData[
    \({3726.1806246232572`, 1.1907003532331667`*^47, 1.`, 
      0.8633196833733946`, {636.1554550140316`, \
\(-0.020981013465594278`\)}}\)], "Print"],

Cell[BoxData[
    \({3726.1806246232572`, 1962.5405290752024`, 0.1`, 
      0.4733104143941952`, {495.38124939776014`, 
        0.13658640420080104`}}\)], "Print"],

Cell[BoxData[
    \({1962.5405290752024`, 918.5665113601084`, 0.010000000000000002`, 
      0.5319502972033093`, {1139.1426289061471`, 
        0.013718161004579954`}}\)], "Print"],

Cell[BoxData[
    \({918.5665113601084`, 116.2737751702434`, 0.0010000000000000002`, 
      0.8734182296738877`, {1195.9611257140589`, 
        0.1369018367179321`}}\)], "Print"],

Cell[BoxData[
    \({116.2737751702434`, 40.290812575840775`, 0.00010000000000000003`, 
      0.6534832337141494`, {1009.0551198510462`, 
        0.08926384931678064`}}\)], "Print"],

Cell[BoxData[
    \({40.290812575840775`, 40.20849379619024`, 0.000010000000000000004`, 
      0.0020431153999583324`, {994.5242511735486`, 
        0.09837695245438006`}}\)], "Print"],

Cell[BoxData[
    \({40.20849379619024`, 40.20844235106101`, 1.0000000000000004`*^-6, 
      1.2794592478682177`*^-6, {989.1689793726828`, 
        0.09871494488673607`}}\)], "Print"],

Cell[BoxData[
    \({Parameters \[Rule] {989.2420671508966`, 0.09870494162843006`}, 
      Uncertainties \[Rule] {18.289563357554783`, 0.0013920694606020228`}, 
      CorrelationMatrix \[Rule] {{1.0000000000000002`, \(-0.6736802054543967`\
\)}, {\(-0.673680205454559`\), 1.`}}, 
      ReducedChiSquare \[Rule] 0.8205804561441021`, 
      NITS \[Rule] 13}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "produces a successful fit even though it was necessary to increase ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\)]],
  " several times to compensate for the poor initial conditions.  Starting \
with the gradient search gets the fit going in the right direction, and then \
it settles into the minimum quickly.  Experience shows that this algorithm is \
quite robust!  With relatively simple modifications, one can include \
constraints upon the parameters that prevent the fit from attempting to move \
in unphysical directions.  For this problem, ",
  Cell[BoxData[
      \(TraditionalForm\`\[Tau] < 0\)]],
  " represents exponential growth and the initial difficulties in the search \
path arise when poor initial conditions result in parameter increments that \
venture into unphysical territory.  Adding ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\^2\)]],
  " a penalty for unphysical parameters could mitigate this problem.  We do \
not investigate those techniques here but simply re-iterate that the use of \
one's own code instead of generic commercial software provides that \
flexibility."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["example", "Subsection"],

Cell[TextData[{
  "\tThe scattering of  high-energy electrons by protons or neutrons is \
governed by electric and magnetic ",
  StyleBox["form factors",
    FontSlant->"Italic"],
  " which nonrelativistically are given by Fourier transforms of charge and \
magnetization densities.  Until recently it was believed that the proton \
electric form factor, ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep\)]],
  ", was approximately equal to the so-called dipole form factor, ",
  Cell[BoxData[
      \(TraditionalForm\`G\_D\)]],
  ", such that"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`G\_Ep[Q\^2]\  \[TildeTilde] \ G\_D[Q\^2]\  = \ 
      1\/\((1 + Q\^2/\[CapitalLambda]\^2)\)\^2\)], "DisplayFormula"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`Q\^2\)]],
  " is the square of the momentum transferred in the electron-proton \
collision and ",
  Cell[BoxData[
      \(TraditionalForm\`\[CapitalLambda]\^2 = 0.71\)]],
  " ",
  Cell[BoxData[
      \(TraditionalForm\`\((GeV/c)\)\^2\)]],
  " is the characteristic scale related to the size of the proton.  However, \
more recent data using a new technique shows that the ratio ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep/G\_D\)]],
  " decreases fairly rapidly for ",
  Cell[BoxData[
      \(TraditionalForm\`Q\^2 > 1\)]],
  ".  A simple parametrization is the rational function of ",
  Cell[BoxData[
      \(TraditionalForm\`Q\^2\)]]
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`G\_Ep[
        Q\^2]\  = \ \(\[Sum]\+\(\[Mu] = 0\)\%m\( a\_\[Mu]\) Q\^\(2  \[Mu]\)\)\
\/\(\[Sum]\+\(\[Nu] = 0\)\%\(m + 2\)\(b\_\[Nu]\) Q\^\(2  \[Nu]\)\)\)], \
"DisplayFormula"],

Cell[TextData[{
  "where the symmetry of the problem limits the expansions to even powers.  \
Recognizing that ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep[0] = 1\)]],
  " because the charge must be one elementary unit, we can constrain ",
  Cell[BoxData[
      \(TraditionalForm\`a\_0 = \(b\_0 = 1\)\)]],
  ", but if may be useful to vary ",
  Cell[BoxData[
      \(TraditionalForm\`a\_0\)]],
  " if the systematic normalization uncertainties are nonnegligible.  \
Finally, there are theoretical reasons based upon the quark model to require \
"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`Q\^2 \[Rule] \[Infinity]\ \ \[DoubleLongRightArrow]\ \ \
G\_Ep \[Proportional] Q\^\(-4\)\)], "DisplayFormula"],

Cell["\<\
This is the reason that we include two more terms in the denominator than in \
the numerator.\
\>", "Text"],

Cell[TextData[{
  "\tWe are now ready to analyze some data.  The following cell reads a data \
file assuming that it is located in a directory contain in ",
  StyleBox["$Path", "InlineInput"],
  "; if that is not true, you will need to include the proper directory in \
the file name."
}], "Text"],

Cell[BoxData[
    \(\(gepdata = Import["\<gep.dat\>"];\)\)], "Input"],

Cell["Next we try to fit the simplest version of this model", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(curfit[gepdata, 1\/\(1 + b1\ qq\  + \ b2\ qq\^2\), 
      qq, {b1, b2}, {1.4, 0}, MaxNITS \[Rule] 20]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {2.9513166307195906`, 2.5502433025309394`}, 
      Uncertainties \[Rule] {0.024372317971975185`, 0.04299062735234152`}, 
      CorrelationMatrix \[Rule] {{1.`, \(-0.5871348315997659`\)}, \
{\(-0.587134831599766`\), 1.`}}, ReducedChiSquare \[Rule] 7.376189744199092`, 
      NITS \[Rule] 6}\)], "Output"]
}, Closed]],

Cell["\<\
but find the results quite unsatisfactory.  Therefore, we add another term to \
both numerator and denominator and obtain a satisfactory fit.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(gepresults = 
      curfit[gepdata, \(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + b3\ \
qq\^3\), qq, {a1, b1, b2, b3}, {0, 1.4, 0, 0}, MaxNITS \[Rule] 20]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {\(-0.06741957500379343`\), 3.1171437164045246`, 
          1.0340714229621064`, 0.5029938181512407`}, 
      Uncertainties \[Rule] {0.035017685288137916`, 0.053168339929658165`, 
          0.09031904317044523`, 0.15572094294077274`}, 
      CorrelationMatrix \[Rule] {{1.`, 0.851549757443654`, 
            0.20352612016955865`, 0.9543774694839172`}, {0.8515497574436542`, 
            1.`, \(-0.18219252454055435`\), 
            0.8880368351329994`}, {0.20352612016955932`, \
\(-0.1821925245405533`\), 
            1.`, \(-0.06266072955377126`\)}, {0.9543774694839169`, 
            0.8880368351329988`, \(-0.06266072955377187`\), 
            1.0000000000000002`}}, 
      ReducedChiSquare \[Rule] 0.8454721516463208`, 
      NITS \[Rule] 9}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "It is useful to display the fit in the form of ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep/G\_D\)]],
  ".  Clearly, this parametrization describes the data very well.  The fact \
that ",
  Cell[BoxData[
      \(TraditionalForm\`\[Chi]\_red\%2\)]],
  " is significantly below 1 for a relatively large data set combining \
results from several different experiments suggests that some of the \
experimentalists overestimated the random uncertainties in their data or \
included systematic uncertainties in the published error bars.  "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\({x, y, dy} = Transpose[gepdata];\)\), "\[IndentingNewLine]", 
    \(\(y = y\ \((1 + x\/0.71)\)\^2;\)\), "\[IndentingNewLine]", 
    \(\(dy = dy\ \((1 + x\/0.71)\)\^2;\)\), "\[IndentingNewLine]", 
    \(\(ratiodata = Transpose[{x, y, dy}];\)\), "\[IndentingNewLine]", 
    \(\(DisplayTogether[
        Plot[Evaluate[\(\(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + 
                      b3\ qq\^3\)\) \((1 + qq\/0.71)\)\^2 /. 
              Thread[{a1, b1, b2, 
                    b3} \[Rule] \((Parameters /. gepresults)\)]], {qq, 0, 
            10}, PlotRange \[Rule] {{0, 10}, {0, 1.2}}, Frame \[Rule] True, 
          FrameLabel \[Rule] {\*"\"\<\!\(Q\^2\)\>\"", \
\*"\"\<\!\(G\_Ep\)/\!\(G\_D\)\>\""}], \[IndentingNewLine]ErrorListPlot[
          ratiodata]];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
-2.77556e-017 0.1 -2.18738e-017 0.515028 [
[.2 -0.0125 -5 -15.6875 ]
[.2 -0.0125 5 0 ]
[.4 -0.0125 -5 -15.6875 ]
[.4 -0.0125 5 0 ]
[.6 -0.0125 -5.125 -15.6875 ]
[.6 -0.0125 5.125 0 ]
[.8 -0.0125 -5 -15.6875 ]
[.8 -0.0125 5 0 ]
[1 -0.0125 -8 -15.6875 ]
[1 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -8.71875 -31.375 ]
[.5 -0.0125 8.71875 -15.6875 ]
[-0.0125 .10301 -19 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -19 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -19.25 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -19 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -35.6875 -20.9375 ]
[-0.01875 .30902 -19.25 20.9375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 0 m
.2 .00625 L
s
gsave
.2 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.4 0 m
.4 .00625 L
s
gsave
.4 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.6 0 m
.6 .00625 L
s
gsave
.6 -0.0125 -66.125 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.8 0 m
.8 .00625 L
s
gsave
.8 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.05 0 m
.05 .00375 L
s
.1 0 m
.1 .00375 L
s
.15 0 m
.15 .00375 L
s
.25 0 m
.25 .00375 L
s
.3 0 m
.3 .00375 L
s
.35 0 m
.35 .00375 L
s
.45 0 m
.45 .00375 L
s
.5 0 m
.5 .00375 L
s
.55 0 m
.55 .00375 L
s
.65 0 m
.65 .00375 L
s
.7 0 m
.7 .00375 L
s
.75 0 m
.75 .00375 L
s
.85 0 m
.85 .00375 L
s
.9 0 m
.9 .00375 L
s
.95 0 m
.95 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -69.7188 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Q) show
71.625 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
76.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .02575 m
.00375 .02575 L
s
0 .0515 m
.00375 .0515 L
s
0 .07725 m
.00375 .07725 L
s
0 .12876 m
.00375 .12876 L
s
0 .15451 m
.00375 .15451 L
s
0 .18026 m
.00375 .18026 L
s
0 .23176 m
.00375 .23176 L
s
0 .25751 m
.00375 .25751 L
s
0 .28327 m
.00375 .28327 L
s
0 .33477 m
.00375 .33477 L
s
0 .36052 m
.00375 .36052 L
s
0 .38627 m
.00375 .38627 L
s
0 .43777 m
.00375 .43777 L
s
0 .46353 m
.00375 .46353 L
s
0 .48928 m
.00375 .48928 L
s
0 .54078 m
.00375 .54078 L
s
0 .56653 m
.00375 .56653 L
s
0 .59228 m
.00375 .59228 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -109.406 -12.2188 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.9375 12.2188 translate
90 rotate
-81.9375 -12.2188 translate
0 24.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
71.688 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Ep) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
81.625 15.813 moveto
(\\220) show
85.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
94.250 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(D) show
100.875 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .61178 m
0 .61803 L
s
.2 .61178 m
.2 .61803 L
s
.4 .61178 m
.4 .61803 L
s
.6 .61178 m
.6 .61803 L
s
.8 .61178 m
.8 .61803 L
s
.125 Mabswid
.05 .61428 m
.05 .61803 L
s
.1 .61428 m
.1 .61803 L
s
.15 .61428 m
.15 .61803 L
s
.25 .61428 m
.25 .61803 L
s
.3 .61428 m
.3 .61803 L
s
.35 .61428 m
.35 .61803 L
s
.45 .61428 m
.45 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.55 .61428 m
.55 .61803 L
s
.65 .61428 m
.65 .61803 L
s
.7 .61428 m
.7 .61803 L
s
.75 .61428 m
.75 .61803 L
s
.85 .61428 m
.85 .61803 L
s
.9 .61428 m
.9 .61803 L
s
.95 .61428 m
.95 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .02575 m
1 .02575 L
s
.99625 .0515 m
1 .0515 L
s
.99625 .07725 m
1 .07725 L
s
.99625 .12876 m
1 .12876 L
s
.99625 .15451 m
1 .15451 L
s
.99625 .18026 m
1 .18026 L
s
.99625 .23176 m
1 .23176 L
s
.99625 .25751 m
1 .25751 L
s
.99625 .28327 m
1 .28327 L
s
.99625 .33477 m
1 .33477 L
s
.99625 .36052 m
1 .36052 L
s
.99625 .38627 m
1 .38627 L
s
.99625 .43777 m
1 .43777 L
s
.99625 .46353 m
1 .46353 L
s
.99625 .48928 m
1 .48928 L
s
.99625 .54078 m
1 .54078 L
s
.99625 .56653 m
1 .56653 L
s
.99625 .59228 m
1 .59228 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
0 .51503 m
.00236 .51107 L
.00488 .5078 L
.00726 .50544 L
.00942 .50379 L
.01197 .50234 L
.01341 .50171 L
.01473 .50124 L
.01717 .50059 L
.0184 .50036 L
.01975 .50017 L
.02113 .50003 L
.02245 .49995 L
.02364 .49991 L
.02493 .4999 L
.02566 .49991 L
.02632 .49993 L
.02709 .49995 L
.02782 .49998 L
.0291 .50006 L
.0305 .50015 L
.04057 .50113 L
.04336 .50139 L
.04636 .50165 L
.04914 .50184 L
.05051 .50191 L
.05177 .50197 L
.05319 .50202 L
.05393 .50204 L
.05472 .50206 L
.056 .50207 L
.05673 .50207 L
.05741 .50207 L
.05817 .50206 L
.05897 .50205 L
.05965 .50203 L
.0604 .50201 L
.06206 .50194 L
.06361 .50185 L
.06637 .50163 L
.06898 .50135 L
.07386 .50064 L
.07911 .4996 L
.08481 .49816 L
.09475 .49486 L
.10568 .49018 L
.12777 .47783 L
.17013 .4467 L
.2109 .41231 L
.25424 .37508 L
Mistroke
.29598 .34069 L
.33614 .30988 L
.37887 .27987 L
.42002 .25368 L
.46373 .22856 L
.50586 .20679 L
.54639 .18784 L
.5895 .16961 L
.63101 .1537 L
.67094 .13976 L
.71344 .12621 L
.75435 .11428 L
.79783 .10267 L
.83972 .09239 L
.88003 .08327 L
.9229 .07429 L
.96419 .06628 L
1 .05979 L
Mfstroke
.002 w
.00054 .5144 m
.00054 .51667 L
s
.015 w
.00054 .51554 Mdot
.002 w
.00078 .51284 m
.00078 .51541 L
s
.015 w
.00078 .51413 Mdot
.002 w
.00097 .5134 m
.00097 .51566 L
s
.015 w
.00097 .51453 Mdot
.002 w
.00117 .50966 m
.00117 .51407 L
s
.015 w
.00117 .51187 Mdot
.002 w
.00136 .51306 m
.00136 .51532 L
s
.015 w
.00136 .51419 Mdot
.002 w
.00156 .50904 m
.00156 .51417 L
s
.015 w
.00156 .51161 Mdot
.002 w
.00175 .51088 m
.00175 .51251 L
s
.015 w
.00175 .51169 Mdot
.002 w
.00195 .51102 m
.00195 .51419 L
s
.015 w
.00195 .51261 Mdot
.002 w
.00214 .51036 m
.00214 .512 L
s
.015 w
.00214 .51118 Mdot
.002 w
.00234 .51063 m
.00234 .51371 L
s
.015 w
.00234 .51217 Mdot
.002 w
.00253 .50982 m
.00253 .51165 L
s
.015 w
.00253 .51074 Mdot
.002 w
.00273 .50724 m
.00273 .5108 L
s
.015 w
.00273 .50902 Mdot
.002 w
.00312 .50738 m
.00312 .51013 L
s
.015 w
.00312 .50875 Mdot
.002 w
.00331 .50855 m
.00331 .51181 L
s
.015 w
.00331 .51018 Mdot
.002 w
.0035 .50976 m
.0035 .51355 L
s
.015 w
.0035 .51166 Mdot
.002 w
.00389 .50243 m
.00389 .51053 L
s
.015 w
.00389 .50648 Mdot
.002 w
.00467 .50224 m
.00467 .5083 L
s
.015 w
.00467 .50527 Mdot
.002 w
.00545 .50351 m
.00545 .50756 L
s
.015 w
.00545 .50553 Mdot
.002 w
.013 .49459 m
.013 .50602 L
s
.015 w
.013 .5003 Mdot
.002 w
.019 .4974 m
.019 .5056 L
s
.015 w
.019 .5015 Mdot
.002 w
.027 .48672 m
.027 .50038 L
s
.015 w
.027 .49355 Mdot
.002 w
.033 .48851 m
.033 .4994 L
s
.015 w
.033 .49396 Mdot
.002 w
.039 .50568 m
.039 .51792 L
s
.015 w
.039 .5118 Mdot
.002 w
.045 .50252 m
.045 .52856 L
s
.015 w
.045 .51554 Mdot
.002 w
.053 .49594 m
.053 .53145 L
s
.015 w
.053 .5137 Mdot
.002 w
.058 .49263 m
.058 .50708 L
s
.015 w
.058 .49985 Mdot
.002 w
.065 .47904 m
.065 .5225 L
s
.015 w
.065 .50077 Mdot
.002 w
.072 .52816 m
.072 .60003 L
s
.015 w
.072 .56409 Mdot
.002 w
.078 .47758 m
.078 .50684 L
s
.015 w
.078 .49221 Mdot
.002 w
.094 .52387 m
.094 .56648 L
s
.015 w
.094 .54518 Mdot
.002 w
.11 .45531 m
.11 .48858 L
s
.015 w
.11 .47194 Mdot
.002 w
.135 .47145 m
.135 .51706 L
s
.015 w
.135 .49426 Mdot
.002 w
.175 .40139 m
.175 .48028 L
s
.015 w
.175 .44083 Mdot
.002 w
.0373 .49251 m
.0373 .52127 L
s
.015 w
.0373 .50689 Mdot
.002 w
.0401 .49939 m
.0401 .52965 L
s
.015 w
.0401 .51452 Mdot
.002 w
.0441 .46155 m
.0441 .49404 L
s
.015 w
.0441 .47779 Mdot
.002 w
.038 .45641 m
.038 .50739 L
s
.015 w
.038 .4819 Mdot
.002 w
.05 .49811 m
.05 .54897 L
s
.015 w
.05 .52354 Mdot
.002 w
.049 .49418 m
.049 .51066 L
s
.015 w
.049 .50242 Mdot
.002 w
.079 .49585 m
.079 .50872 L
s
.015 w
.079 .50228 Mdot
.002 w
.118 .47152 m
.118 .48539 L
s
.015 w
.118 .47845 Mdot
.002 w
.148 .42209 m
.148 .45345 L
s
.015 w
.148 .43777 Mdot
.002 w
.177 .42258 m
.177 .44897 L
s
.015 w
.177 .43578 Mdot
.002 w
.188 .41683 m
.188 .44287 L
s
.015 w
.188 .42985 Mdot
.002 w
.213 .39604 m
.213 .43229 L
s
.015 w
.213 .41417 Mdot
.002 w
.247 .37741 m
.247 .40221 L
s
.015 w
.247 .38981 Mdot
.002 w
.297 .32293 m
.297 .35614 L
s
.015 w
.297 .33954 Mdot
.002 w
.347 .309 m
.347 .35542 L
s
.015 w
.347 .33221 Mdot
.002 w
.35 .27179 m
.35 .354 L
s
.015 w
.35 .31289 Mdot
.002 w
.397 .23357 m
.397 .29113 L
s
.015 w
.397 .26235 Mdot
.002 w
.475 .17678 m
.475 .23446 L
s
.015 w
.475 .20562 Mdot
.002 w
.554 .0975 m
.554 .194 L
s
.015 w
.554 .14575 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00i@3oool01@3ogkl0_moo0?ooo`3oool0gia800@0000000<0001d0;oo
o`3oool0_`3oool00>D0oooo00<0oooO09a8000000000P0000001`0007@0_ooo0?ooo`3oool0oono
07@0M02oool0_`3oool00>L0oooo00<0_g@004RLg`3oool01@3oool00`3ookl0M7Bo0?ooo`2n0?oo
o`00j03oool00`3OW4P007Bo0?ooo`350?ooo`00i`3oool00`3ogi`0B00000000002000000030018
W03Oool0oooo0<40oooo003V0?ooo`040?oOW018000004RL0=ooo`80oooo00<0_g@000000018W=l0
`@3oool00>H0oooo00<0_g@00000B02Lgol0103oool00`3o_g@00000001d_`300?ooo`00i@3oool0
103ooml0W4P00018W03Oool50?ooo`030=nLB0000000M;oo0;l0oooo003U0?ooo`040?oog`2LB000
001d0;ooo`D0oooo00@0oooO09a8000007@0_ooo_P3oool00>H0oooo00<0_g@00000B02Lgol01P3o
ool00`3OW4P0000004RLg`2n0?ooo`00iP3oool00`3OW4P0000004RLg`060?ooo`040?oOW0180000
04RL0=oookd0oooo003V0?ooo`040?oo_`1d000004RL0=ooo`D0oooo00@0oooO09a80000B9`0gooo
_@3oool00>L0oooo00<0okmd00000018W=l01@3oool0103ooml0W4P00018W02WM003000000030000
M02oool0oooo0;L0oooo003X0?ooo`030=nLB0000000M;oo00@0oooo00T0oono07@00000M;l0oooo
0?noM000B9`0gooo0?oOW02L_ol0^03oool00>T0oooo00<0okmd00000000M;l00P3oool0103ooml0
W4P00000B02Lgol20?ooo`030?oOW018B9`0gooo0;T0oooo003[0?ooo`060;md0000000000000000
000007@0_ooo103oool00`3ogi`0B4RL0=ooo`2h0?ooo`00mP3oool00`3ogi`0M9cO0?ooo`2g0?oo
o`00m`3oool00`2oM4P0W=oo0?ooo`2f0?ooo`00l`3oool01P3ogml0gooo0?ooo`3oool0gia804RL
gkL0oooo003c0?ooo`060?oo_`1dB9`0gooo0?oog`2LB000B9cO]`3oool00?@0oooo00D0omnL04P0
0000000004RL0=ooo`2g0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00PP3oool00`3OW4P00000
000000030000000304RLg`3oool0oooo0400oooo00<0oooO09a80018W=l0@03oool01@3ooml0W4P0
00000000000007Bo0440oooo00H0oono07@00000000000000018W03Ooolj0?ooo`030?oo_`1d0000
00000080000000<0B9cO0?ooo`3oool00`3oool0102oM00000000000001d_ol=0?ooo`00P`3oool0
0`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo0440oooo00<0gia80018W03Oool0?`3oool0
1`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool0?P3oool00`3ookl0M00004RLg`020?ooo`030=nL
B00007@0_ooo03/0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RL
g``0oooo00240?ooo`030?noM000M;l0oooo04D0oooo00<0omnL04P0B02Lgol0?P3oool00`3ooml0
W4P007Boo`020?ooo`030?oOW0180000B9cO03h0oooo00<0okmd0018W03Oool00`3oool00`2oM000
B9cO0?ooo`0j0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`0;0?ooo`00Q@3oool00`3o_g@007Bo0?ooo`100?ooo`030?noM0000000000000@00000
00<004RL0=ooo`3oool0>`3oool00`3ookl0M00007Boo`030?ooo`030;md000007@0_ooo03d0oooo
00<0okmd0018W03Oool00`3oool00`2oM00007Bo0?ooo`0k0?ooo`030=nLB000B9`0gooo00<0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`00QP3oool00`3o_g@0B9cO
0?ooo`0o0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0?@3oool00`3ooml0W4P0
001d_`030?ooo`030=nLB00004P0W=oo03d0oooo00<0oooO09a8001d_ol00P3oool00`3ookl0M000
04RLg`0l0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P00018
09cOo`0:0?ooo`00Q`3oool00`3OW4P0B9cO0?ooo`0o0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@0
04RL0=ooo`0n0?ooo`030;md000007@0_ooo0080oooo00<0gia80000B02Lgol0?`3oool01P2oM000
001809cOg`2LB000001d0;oooc`0oooo00<0oono07@0001d_ol0103oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0oooo00280?ooo`030;md001d_ol0oooo03l0oooo00H0omnL04Q8
W03Oool0omnL04P0B02Lgoln0?ooo`080?noM0000000B9cO0?ooo`3ooml0W4P00018W03Ooom00?oo
o`050?oog`2LB00000000018W03Oool0?@3oool00`2oM00007Bo0?ooo`030?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02P3oool008P0oooo00<0omnL04P0M02oool0@03oool01@3o
_g@007Bo0?oog`2LB000B9cO03h0oooo00<0oooO09a8000000000P0000000`0007@0_ooo0?ooo`10
0?ooo`060;md000004P0W=oo0?oOW018B9`0gooo?03oool00`3OW4P004RL0=ooo`030?ooo`030?oO
W0180000B9cO00<0oooo00<0_g@00018W03Oool02@3oool008P0oooo00<0oooO09a8001d_ol0@@3o
ool0103OW4P0B9cO0;md0000M;lo0?ooo`030?oo_`1d000007Bo04<0oooo00<0omnL04P0001d_ol0
0P3oool00`3o_g@0B9cO0?ooo`0k0?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0
gooo0P3oool00`3OW4P004RL0=ooo`090?ooo`00R03oool00`3ookl0M00004RLg`120?ooo`040=nL
B01dM7@0001d0;ooocl0oooo00<0omnL04P00018W=l0@P3oool00`3ogi`0B01809cOo`020?ooo`03
0?noM000B9`0gooo03/0oooo00<0oono07@0001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`03
0=nLB000M;l0oooo00T0oooo00240?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`130?oo
o`030;md000004P0W=oo0400oooo00@0oooO09a8000007@0_ooo@03oool00`3ooml0W4P004RLg`02
0?ooo`030;md0000M;l0oooo03`0oooo00<0_g@0001d_`3oool0103oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`X0oooo00250?ooo`050?oOW018000000000000001d_ol0A03oool00`3ooml0W4P0
04RLg`120?ooo`040?oo_`1d0000000004RLgcl0oooo00D0oooO09a8000000000000001d_`0l0?oo
o`050?oo_`1d000000000000M02oool01@3oool0103OW4P0000000000000M;l;0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00@P3ooooo000005d000004P3o
ool00480oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`130?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0@`3oool00`00
0000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3o
ool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?oo
o`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0
ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool200000?l0ooooF03oool300000180oooo
00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?oo
o`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0
ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003o
ool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@
0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480
oooo0P00003o0?oooeP0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003o
ool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0ooooEP3oool400000180oooo00120?oo
o`030000003oool0oooo0?l0ooooD03oool6000000<0oooo00<000000?ooo`3oool0403oool00480
oooo00<000000?ooo`3oool0o`3ooom:0?ooo`H000002@3oool00`000000oooo0?ooo`0@0?ooo`00
@P3oool00`000000oooo0?ooo`3o0?ooodH0oooo1000000?0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0?l0oooo@@3oool5000001<0oooo00<000000?ooo`3oool0403o
ool00480oooo00<000000?ooo`3oool0o`3oooll0?ooo`D00000603oool00`000000oooo0?ooo`0@
0?ooo`00@P3oool200000?l0oooo>03oool5000001/0oooo0`00000B0?ooo`00@P3oool00`000000
oooo0?ooo`3o0?oooc80oooo1@00000R0?ooo`030000003oool0oooo0100oooo00120?ooo`030000
003oool0oooo0?l0oooo;@3oool5000002L0oooo00<000000?ooo`3oool0403oool00480oooo00<0
00000?ooo`3oool0o`3ooolY0?ooo`@00000;03oool00`000000oooo0?ooo`0@0?ooo`009P3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0gia80000000000000`00
00000`18W=l0oooo0?ooo`070?ooo`030000003oool0oooo0?l0oooo903oool500000300oooo00<0
00000?ooo`3oool0403oool002D0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`070?ooo`030000
003oool0oooo0?l0oooo803oool4000003D0oooo00<000000?ooo`3oool0403oool002@0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?noM000M;l0oooo00/0oooo00<0
00000?ooo`3oool0o`3ooolL0?ooo`@00000>@3oool00`000000oooo0?ooo`0@0?ooo`00903oool0
0`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool02P3o
ool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo05L0oooo1000000m0?ooo`030000003o
ool0oooo0100oooo000T0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02@3oool0
0`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0D`3oool4
00000440oooo00<000000?ooo`3oool0403oool002D0oooo00<0_g@00018W03Oool00P3oool00`3o
gi`0B00004RLg`0:0?ooo`030=nLB018W=l0oooo00P0oooo0`00002m0?ooo`030000003oool0oooo
0500oooo0`0000120?ooo`@000004P3oool002D0oooo00<0gia80000B02Lgol00P3oool00`3ooml0
W4P0001d_`0;0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0_@3oool00`000000
oooo0?ooo`1<0?ooo`@00000B03oool00`000000oooo0?ooo`0@0?ooo`009@3oool00`3ogi`0B000
04RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02oool01`3oool00`000000oooo
0?ooo`2m0?ooo`030000003oool0oooo04P0oooo1000001<0?ooo`030000003oool0oooo0100oooo
000U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3ooml0W4P0
07Boo`070?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0A03oool400000500oooo
00<000000?ooo`3oool0403oool002H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?oo
o`0:0?ooo`030?oo_`1d0000B9cO00L0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`110?ooo`<00000E03oool00`000000oooo0?ooo`0@0?ooo`009`3oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00L0oooo00<000000?oo
o`3oool0_@3oool00`000000oooo0?ooo`0m0?ooo`@00000E`3oool00`000000oooo0?ooo`0@0?oo
o`00:03oool0103OW4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol0203oool0
0`000000oooo0?ooo`2m0?ooo`030000003oool0oooo03X0oooo0`00001K0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0=P3oool40000
05h0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000
oooo0?ooo`0c0?ooo`<00000HP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool200000;h0oooo
00<000000?ooo`3oool0<03oool3000006<0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?oo
o`2m0?ooo`030000003oool0oooo02d0oooo0`00001X0?ooo`030000003oool0oooo0100oooo0012
0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0:P3oool3000006/0oooo00<00000
0?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0W0?oo
o`<00000KP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2l0?ooo`<0
00009@3oool300000740oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0
^`3oool500000200oooo1000001d0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0;/0oooo1@00000M0?ooo`<00000N03oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`2k0?ooo`D000006@3oool4000007/0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool0_03oool3000001L0oooo0`00001o0?ooo`030000003oool0oooo
0100oooo00120?ooo`800000_P3oool00`000000oooo0?ooo`0D0?ooo`800000P03oool300000180
oooo00120?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool04P3oool2000008@0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`0?0?ooo`<00000QP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2m
0?ooo`030000003oool0oooo00d0oooo0P0000290?ooo`030000003oool0oooo0100oooo00120?oo
o`030000003oool0oooo0;d0oooo00<000000?ooo`3oool02`3oool2000008/0oooo00<000000?oo
o`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`080?ooo`<0
0000S@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2m0?ooo`030000
003oool0oooo00H0oooo0P00002@0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`<00000TP3o
ool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01T0oooo00@000000?ooo`3oool0oooo0P00002E0?ooo`030000003oool0oooo0100oooo0012
0?ooo`800000XP3oool00`000000oooo0?ooo`0I0?ooo`@00000U@3oool300000180oooo00120?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0603oool2000009X0oooo00<000000?oo
o`3oool0403oool00480oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0F0?ooo`80
000000<0oooo0000003oool0V@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01@0oooo0P0000030?ooo`030000003oool0oooo09P0oooo
00<000000?ooo`3oool0403oool002H0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B000
07Boo`050?ooo`030?oog`2LB000B9cO00X0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0B0?ooo`8000001@3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo0100oooo
000U0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`060?ooo`03
0=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0@0?ooo`80
0000XP3oool00`000000oooo0?ooo`0@0?ooo`00903oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00X0oooo00<0omnL04P0B02Lgol02@3oool00`000000oooo0?ooo`2P0?ooo`<00000
3`3oool200000:@0oooo00<000000?ooo`3oool0403oool002@0oooo00<0oooO09a80018W=l00P3o
ool0103ooml0W4P00018W03Oool50?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0
1P3oool00`000000oooo0?ooo`2O0?ooo`D00000303oool200000:H0oooo00<000000?ooo`3oool0
403oool002@0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2L
B4P0W=oo0080oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`2O0?ooo`D000002P3o
ool200000:P0oooo00<000000?ooo`3oool0403oool002D0oooo00<0_g@00018W03Oool00P3oool0
0`3ogi`0B00004RLg`060?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`080?ooo`<00000
W`3oool5000000L0oooo0`00002W0?ooo`@000004P3oool002D0oooo00<0gia80000B02Lgol00P3o
ool00`3ooml0W4P0001d_`070?ooo`060?oOW018B9`0gooo0?oOW01804P0W=oo203oool00`000000
oooo0?ooo`2P0?ooo`<000001P3oool200000:d0oooo00<000000?ooo`3oool0403oool002D0oooo
00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?ooo`050?noM000M;l0oooO09a80018
W=l0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00<0oooo0P00002_0?ooo`03
0000003oool0oooo0100oooo000U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool0203oool0103OW4P0B9cO0;md0000M;l80?ooo`030000003oool0oooo0:40oooo00@000000?oo
o`3oool0oooo0P00002a0?ooo`030000003oool0oooo0100oooo000V0?ooo`030?oOW0180000M;oo
0080oooo00<0gia8001d_`3oool02@3oool0103OW4P0M7Ad0000M02oool70?ooo`030000003oool0
oooo0:40oooo00@000000?ooo`0000000000/`3oool00`000000oooo0?ooo`0@0?ooo`009`3oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`/0oooo00<0_g@00000B02Lgol01`3oool00`000000oooo
0?ooo`2Q0?ooo`800000]@3oool00`000000oooo0?ooo`0@0?ooo`00:03oool0103OW4P000000000
0000M;l<0?ooo`030?oog`2LB000B9cO00L0oooo00<000000?ooo`3oool0W`3oool300000;H0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0W@3oool200000080oooo00<0
00000?ooo`3oool0]03oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000009`0oooo0P000004
0?ooo`030000003oool0oooo0;80oooo0`00000B0?ooo`003`3oool00`3OW4P00000000000020000
000304RLg`3oool0oooo02/0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0@0?oo
o`8000001P3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0100oooo000=0?ooo`03
0?noM0000000B9cO00@0oooo00<0gia8001d_`3oool0:`3oool00`000000oooo0?ooo`260?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0@0?ooo`00
303oool00`3OW4P007Bo0?ooo`060?ooo`030?noM01d_ol0oooo02X0oooo00<000000?ooo`3oool0
QP3oool00`000000oooo0?ooo`0=0?ooo`800000`03oool00`000000oooo0?ooo`0@0?ooo`002`3o
ool00`3o_g@0B9cO0?ooo`080?ooo`030;nLW03Oool0oooo02T0oooo00<000000?ooo`3oool0QP3o
ool00`000000oooo0?ooo`0;0?ooo`800000`P3oool00`000000oooo0?ooo`0@0?ooo`002P3oool0
0`3ooml0W9bo0?ooo`090?ooo`030=nLW02oool0oooo02T0oooo00<000000?ooo`3oool0QP3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool0403oool0
00X0oooo00<0omnL09boo`3oool02@3oool00`3OW9`0_ooo0?ooo`0Y0?ooo`030000003oool0oooo
08H0oooo00<000000?ooo`3oool0203oool200000<D0oooo00<000000?ooo`3oool0403oool000X0
oooo00<0oknL09cOo`3oool02@3oool00`2oM7@0_ooo0?ooo`0Y0?ooo`030000003oool0oooo08H0
oooo00<000000?ooo`3oool01P3oool200000<L0oooo00<000000?ooo`3oool0403oool000X0oooo
00<0gibL0;ooo`3oool00`3oool00`3o_g@0M;oo0?ooo`020?ooo`030?oog`2LM9`0gooo02X0oooo
00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0<L0oooo
00<000000?ooo`3oool0403oool000X0oooo00<0okmd07Boo`3oool00`3oool0103o_g@0B9cO0?oo
g`2LB0020000000304RLg`3oool0oooo02T0oooo0P0000260?ooo`<00000103oool200000<P0oooo
0`00000B0?ooo`002P3oool00`3ookl0M01d0;ooo`030?ooo`060?noM000000000000000000007@0
_ooo;03oool00`000000oooo0?ooo`240?ooo`D000000P3oool00`000000oooo0?ooo`3:0?ooo`03
0000003oool0oooo0100oooo000:0?ooo`090?oo_`1d000000000000001d_ol0oooo0?noM000B9`0
gooo02l0oooo00<000000?ooo`3oool0Q03oool700000<d0oooo00<000000?ooo`3oool0403oool0
00X0oooo00<0gia80000001d_ol00`3oool00`3o_g@0M;oo0?ooo`0_0?ooo`030000003oool0oooo
08@0oooo1@00003?0?ooo`030000003oool0oooo0100oooo000I0?ooo`030;mdM02oool0oooo02H0
oooo00<000000?ooo`3oool0Q@3oool300000=00oooo00<000000?ooo`3oool0403oool001H0oooo
00D0okmd000000000000001d0;ooo`0W0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3o
ool02`3oool300000=40oooo00<000000?ooo`3oool0403oool00140oooo00<0gia8000000000000
0P0000001@00M;l0oooo0?ooo`3OW9`0_ooo02L0oooo00<000000?ooo`3oool0MP3oool00`000000
oooo0?ooo`0:0?ooo`040000003oool0oooo00000=40oooo00<000000?ooo`3oool0403oool00140
oooo00H0gia807Boo`3oool0oooO09adW03Oool20?ooo`030=nLW02oool0oooo02H0oooo00<00000
0?ooo`3oool0MP3oool00`000000oooo0?ooo`080?ooo`8000000`3oool00`000000oooo0?ooo`3?
0?ooo`030000003oool0oooo0100oooo000A0?ooo`0:0=nLW02oool0oooo0?oog`2LB000000007Bo
o`3ooml0W4Qd0;ooobL0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`0@0?ooo`00
4@3oool01@3OW9`0_ooo0?oo_`1d000007Bo0080oooo00<0_g@00000M02oool09`3oool2000007L0
oooo00<000000?ooo`3oool01@3oool2000000H0oooo00<000000?ooo`3oool0c@3oool300000180
oooo000A0?ooo`030=nLW02oool0oooo00<0oooo00<0gia80018W03Oool00P3oool00`2oW9`0gooo
0?ooo`0S0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo0<l0oooo00<000000?ooo`3oool0403oool00140oooo00<0
gia800000018W=l01`3oool00`3ogi`0B4RL0=ooo`0T0?ooo`030000003oool0oooo07H0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0<l0oooo00<0
00000?ooo`3oool0403oool001@0oooo00@0_ibL0=ooo`3oool0okmd100000000`00B9`0gooo0?oo
o`0S0?ooo`030000003oool0oooo07H0oooo00@000000?ooo`3oool0oooo0P00000:0?ooo`030000
003oool0oooo0<l0oooo00<000000?ooo`3oool0403oool001@0oooo00L0_g@000000000000004RL
0=ooo`2oM7@0_ooo00X0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`030?oo
o`050?oog`2LB000000000000000M;l02`3oool00`000000oooo0?ooo`1f0?ooo`040000003oool0
oooo00000=h0oooo00<000000?ooo`3oool0403oool001@0oooo00<0oono07Ad_`3oool00P3oool0
0`3OW9`0_ooo0?ooo`080?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B000
07Boo`030?ooo`070=nLB000M;l0oooo0?oo_`1d000004RL0=ooo`090?ooo`030000003oool0oooo
07H0oooo00<000000?ooo`000000g`3oool00`000000oooo0?ooo`0@0?ooo`00503oool01P3OW9`0
_ooo0?ooo`3oool0oono07Ad_`T0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`06
0?ooo`030?oog`2LB000M;oo0080oooo00<0omnL04P00018W=l02@3oool00`000000oooo0?ooo`1f
0?ooo`800000h03oool00`000000oooo0?ooo`0@0?ooo`004`3oool01`3ooml0W7BL0=ooo`3oool0
okmd0000M02oool02@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo
00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`080?ooo`030000003oool0oooo07D0oooo
0P00003Q0?ooo`030000003oool0oooo0100oooo000D0?ooo`040=nLB00000000000001d_`/0oooo
00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo
00<0gia80000B02Lgol0203oool00`000000oooo0?ooo`1d0?ooo`030000003oool000000>40oooo
00<000000?ooo`3oool0403oool002@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RL
g`060?ooo`030;md000007@0_ooo0080oooo00<0gia80000B02Lgol0203oool300000780oooo0P00
0000103oool0000000000000003M0?ooo`@000004P3oool001P0oooo00@0oooO09a8000007@0_ooo
203oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00H0oooo00P0okmd00000018W=l0
oooo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`05
00000=l0oooo00<000000?ooo`3oool0403oool001D0oooo00D0okmd000000000000001809cOo`0:
0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3oool00`3ooml0W4P000000002
000000030000M02oool0oooo00P0oooo00<000000?ooo`3oool0K`3oool2000000<0oooo1@00003O
0?ooo`030000003oool0oooo0100oooo000B0?ooo`040;md000000000000001d_`h0oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d000007Bo00`0oooo00<0
00000?ooo`3oool0KP3oool00`000000oooo0?ooo`030?ooo`D00000g`3oool00`000000oooo0?oo
o`0@0?ooo`003P3oool01@3ogi`0B000000000000000M;oo0180oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`070?ooo`030?oOW0180000B9cO00/0oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`050?ooo`<00000h03oool00`000000oooo0?ooo`0@0?ooo`002`3oool0
1@3OW4P0000000000000B9`0gooo01H0oooo00H0gia80000M02oool0oooO09a80018W=l90?ooo`04
0?oog`2LB000001d0;ooo`T0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`040?oo
o`800000203oool200000>40oooo00<000000?ooo`3oool0403oool000T0oooo00<0omnL04P00018
W=l06`3oool0103OW4P0000000000000M;l<0?ooo`040?oo_`1d0000000004RLg`L0oooo00<00000
0?ooo`3oool0I03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00L0oooo0`00003Q
0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo06@0oooo00@000000?oo
o`3oool0oooo0P0000090?ooo`D00000h03oool00`000000oooo0?ooo`0@0?ooo`003`3oool00`3O
W4P00000000000020000000304RLg`3oool0oooo02/0oooo00<000000?ooo`3oool0I03oool01000
0000oooo0?ooo`00000;0?ooo`D00000h03oool00`000000oooo0?ooo`0@0?ooo`003@3oool00`3o
_g@0000004RLg`040?ooo`030=nLB000M;l0oooo02/0oooo0P00001T0?ooo`@00000303oool50000
0=h0oooo0`00000B0?ooo`00303oool00`3OW4P007Bo0?ooo`060?ooo`030?noM01d_ol0oooo02X0
oooo00<000000?ooo`3oool0HP3oool5000000d0oooo0`00003Q0?ooo`030000003oool0oooo0100
oooo000;0?ooo`030?noM018W=l0oooo00P0oooo00<0_ibL0=ooo`3oool0:@3oool00`000000oooo
0?ooo`1R0?ooo`D000003P3oool200000>40oooo00<000000?ooo`3oool0403oool000X0oooo00<0
oooO09bL_`3oool02@3oool00`3OW9`0_ooo0?ooo`0Y0?ooo`030000003oool0oooo0680oooo1@00
000>0?ooo`800000h@3oool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3ogi`0W;oo0?ooo`09
0?ooo`030=nLW02oool0oooo02T0oooo00<000000?ooo`3oool0H@3oool00`000000oooo00000002
000000l0oooo0P00003Q0?ooo`030000003oool0oooo0100oooo000:0?ooo`030?noW02Lgol0oooo
00T0oooo00<0_gAd0;ooo`3oool0:@3oool00`000000oooo0?ooo`1O0?ooo`8000000`3oool00`00
0000oooo0?ooo`0>0?ooo`800000h@3oool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3OW9`0
_ooo0?ooo`030?ooo`030?noM01d_ol0oooo0080oooo00<0oooO09adW03Oool0:P3oool00`000000
oooo0?ooo`1N0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool200000>40
oooo00<000000?ooo`3oool0403oool000X0oooo00<0okmd07Boo`3oool00`3oool0103o_g@0B9cO
0?oog`2LB0020000000304RLg`3oool0oooo02T0oooo00<000000?ooo`3oool0G@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo0P00003Q0?ooo`030000003oool0oooo0100
oooo000:0?ooo`030?oo_`1d07@0_ooo00<0oooo00H0okmd00000000000000000000M02oool/0?oo
o`030000003oool0oooo05`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3a0?oo
o`030000003oool0oooo0100oooo000:0?ooo`090?oo_`1d000000000000001d_ol0oooo0?noM000
B9`0gooo02l0oooo0P00001L0?ooo`030000003oool0oooo0?P0oooo0`00000B0?ooo`002P3oool0
0`3OW4P0000007Boo`030?ooo`030?noM01d_ol0oooo00H0oooo00<0omoO0=ooo`3oool09P3oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0?/0oooo00<000000?ooo`3oool0403oool0
01T0oooo00<0_gAd0;ooo`3oool09P3oool00`000000oooo0?ooo`1H0?ooo`800000oP3oool00`00
0000oooo0?ooo`0@0?ooo`004@3oool01P3OW9`0_ooo0?ooo`3oool0oooO09a80080000000<0001d
0;ooo`3oool09P3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo0?h0oooo00<00000
0?ooo`3oool0403oool00140oooo00H0gia800000000000000000000B02Lgol20?ooo`030=nLW02o
ool0oooo02H0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`3o0?ooo`030000003o
ool0oooo0100oooo000A0?ooo`030=nLW02oool0oooo00D0oooo00<0gibL0;ooo`3oool09P3oool0
0`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0?l0oooo0@3oool00`000000oooo0?ooo`0@
0?ooo`004@3oool00`3OW9`0_ooo0?ooo`050?ooo`030;nLW03Oool0oooo02H0oooo00<000000?oo
o`3oool0DP3oool00`000000oooo0000003o0?ooo`@0oooo00<000000?ooo`3oool0403oool00140
oooo00<0gibL0;ooo`3oool0103oool00`3ookl0M7Bo0?ooo`0W0?ooo`030000003oool0oooo0580
oooo0P00003o0?ooo`D0oooo00<000000?ooo`3oool0403oool00140oooo00<0okmd07Boo`3oool0
103oool00`2oM4P0W=oo0?ooo`0W0?ooo`030000003oool0oooo0540oooo0`00003o0?ooo`D0oooo
00<000000?ooo`3oool0403oool00140oooo00<0oono07@0B02Lgol00P3oool00`3ookl0M00007Bo
o`0Y0?ooo`800000D@3oool500000?l0oooo0P3oool300000180oooo000B0?ooo`060?oog`2LB000
00000000000007@0_ooo:P3oool00`000000oooo0?ooo`1?0?ooo`H00000o`3oool40?ooo`030000
003oool0oooo0100oooo00120?ooo`030000003oool0oooo04h0oooo00<000000?ooo`0000001000
003o0?ooo`@0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`040?ooo`040000003oool0oooo0?ooo`<00000o`3oool50?ooo`030000003o
ool0oooo0100oooo000V0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00`3o
ool01P3ookl0M00000000000000004RL0=ooo`T0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3oool40?ooo`03
0000003oool0oooo0100oooo000U0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3o
gi`0B00007Boo`020?ooo`030?oo_`1d0000B9cO0080oooo00<0gia80000M02oool0203oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0?l0oooo103oool00`000000oooo0?ooo`0@0?oo
o`00903oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0okmd0018W03O
ool00`3oool00`2oM000B9cO0?ooo`070?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`02000000L0oooo00<000000?ooo`3oool0o`3oool40?ooo`03
0000003oool0oooo0100oooo000T0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0
gooo1@3oool00`3o_g@004RL0=ooo`030?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool000000?l0oooo403oool00`000000
oooo0?ooo`0@0?ooo`00903oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo
00<0oooO09a8001d_ol00P3oool00`3ookl0M00004RLg`080?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool02@3oool300000?l0oooo4@3oool00`000000oooo0?ooo`0@0?ooo`009@3o
ool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo00H0_g@00000B02Lgml0W4P0
0000M02oool80?ooo`<00000>@3oool00`000000oooo0?ooo`080?ooo`D00000o`3oool=0?ooo`@0
00004P3oool002D0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`080?ooo`050?oo
g`2LB00000000018W03Oool0203oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00P0
oooo1@00003o0?oooa00oooo00<000000?ooo`3oool0403oool002D0oooo00<0omnL04P00018W=l0
0`3oool00`2oM00004RL0=ooo`070?ooo`060;md000004P0W=oo0?oOW018B9`0gooo1`3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool50000
0?l0oooo403oool00`000000oooo0?ooo`0@0?ooo`009@3oool0103ooml0W4P00018W03Oool20?oo
o`030=nLB000B9`0gooo00H0oooo00<0omnL04P0001d_ol00P3oool00`3o_g@0B9cO0?ooo`060?oo
o`030000003oool0oooo03T0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool2
000000040?ooo`000000000000000?l0oooo4@3oool00`000000oooo0?ooo`0@0?ooo`009P3oool0
0`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00H0oooo00<0omnL04P0B02Lgol00P3oool0
0`3o_g@004RL0=ooo`060?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool01P3oool2
00000080oooo0`0000000`3oool000000?ooo`030?ooo`030000003oool0oooo0?l0oooo403oool0
0`000000oooo0?ooo`0@0?ooo`009`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<0
oooO09a80018W=l00P3oool00`2oM00007Bo0?ooo`060?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool01P3oool2000000030?ooo`000000000000<000001@3oool00`000000oooo0?oo
o`3o0?oooa00oooo00<000000?ooo`3oool0403oool002P0oooo00@0gia800000000000007Bo2@3o
ool01@3ooml0W4P000000000000007Bo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo
0?ooo`060?ooo`P000001@3oool00`000000oooo0?ooo`3o0?oooa00oooo00<000000?ooo`3oool0
403oool00480oooo00<000000?ooo`3oool0;`3oool3000000H0oooo2@0000050?ooo`030000003o
ool0oooo0?l0oooo403oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0^
0?ooo`D000001@3oool800000?l0oooo6@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool20000
02l0oooo1@0000040?ooo`H0000000<0oooo0000003oool0o`3ooolG0?ooo`<000004P3oool00480
oooo00<000000?ooo`3oool0;P3oool5000000@0oooo1@0000020?ooo`030000003oool0oooo0?l0
oooo603oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0_0?ooo`<00000
1@3oool500000080oooo00<000000?ooo`3oool0o`3ooolH0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool01@3oool300000?l0oooo7P3o
ool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0
oooo00<0oooo0P0000000`3oool000000000003o0?oooah0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo
0?ooo`000000o`3ooolO0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0<0?ooo`80
00001@3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0403oool00480oooo
00<000000?ooo`3oool02P3oool00`000000oooo0000000F0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0
403oool00480oooo0P00000;0?ooo`030000003oool0000001H0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0?l0oooo6`3oool300000180oooo0012
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0000005@3oool3000000T0oooo0P000009
0?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0000001@0oooo1@0000070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`03
0000003oool0oooo00X0oooo1000000C0?ooo`H00000103oool2000000`0oooo00<000000?ooo`3o
ool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo00T0oooo
1P00000B0?ooo`L000000P3oool2000000d0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000
003oool0oooo0100oooo00120?ooo`030000003oool0oooo00P0oooo1`0000040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool01`3oool6000000040?ooo`000000oooo000000d0oooo
00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo00P0oooo1`0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203o
ool600000080oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3o0?oooad0oooo00<0
00000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0203oool6000000D0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`@00000103oool00`000000oooo0?ooo`3o
0?ooob/0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool01@3oool30000
00030?ooo`0000000000008000001P3oool010000000oooo0?ooo`3oool3000000T0oooo0P000006
0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool20000
00D0oooo1`0000000`3oool000000?ooo`050?ooo`030000003oool0oooo00D000001P3oool30000
00H0oooo0`00003o0?ooobX0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?ooo`040?ooo`L0
000000<0oooo0000003oool00`3oool00`000000oooo000000020?ooo`D00000103oool2000000P0
oooo1@00003o0?ooob/0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0
2`0000000`3oool0000000000007000000030?ooo`000000000000<0000000@0oooo000000000000
00002P3oool500000?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3ooolO000000d0oooo
1@00003o0?ooob/0oooo00<000000?ooo`3oool0403oool00340oooo00<0oono07@0000000000P00
00000`18W=l0oooo0?ooo`090?oooah000003`3oool300000?l0oooo;03oool00`000000oooo0?oo
o`0@0?ooo`00<`3oool00`3o_g@0000007Boo`0<0?ooo`T0000000<0oooo0000000000004P00000@
0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00<`3oool00`3o
oml0W4P0001d_`0<0?ooo`P000000P3oool>00000080oooo0`00000A0?ooo`030000003oool0oooo
0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00=03oool00`3OW4P004RL0=ooo`0;0?ooo`@0
00001P3oool;000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`3o0?ooob/0oooo00<000000?ooo`3oool0403oool003@0oooo00<0omnL04P0B02Lgol0
2`3oool4000000L0oooo2P0000000`3oool000000?ooo`0F0?ooo`030000003oool0oooo0?l0oooo
:`3oool00`000000oooo0?ooo`0@0?ooo`00=03oool00`3ookl0M00007Boo`0;0?ooo`@000001`3o
ool:000000030?ooo`000000oooo01H0oooo00<000000?ooo`3oool0o`3ooolX0?ooo`@000004P3o
ool003D0oooo00<0_g@0001d_`3oool02P3oool3000000P0oooo2@0000020?ooo`030000003oool0
oooo0?l0oooo@`3oool00`000000oooo0?ooo`0@0?ooo`00=@3oool00`3OW4P004RL0=ooo`0:0?oo
o`8000002P3oool800000080oooo00<000000?ooo`3oool0o`3ooom30?ooo`030000003oool0oooo
0100oooo000e0?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool02P3oool00`000000
oooo00000004000000`0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100
oooo000e0?ooo`030?oo_`1d0000M;oo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo
00000004000000@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o0?ooocT0oooo
00<000000?ooo`3oool0403oool003H0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00<000001@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00=03oool01@3ookl0M00000000000
07@0_ooo00T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100oooo0012
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool30000
0?l0oooo>P3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000000l0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`030?ooo`D00000o`3ooolg0?ooo`<000004P3oool00480oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo
1@00003o0?ooocT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo1@00003o0?ooocT0oooo00<0
00000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo0`00003o0?ooocX0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100
oooo00120?ooo`030000003oool0oooo01@0oooo0`0000060?ooo`030000003oool0oooo0?l0oooo
>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0C0?ooo`D000001@3o
ool00`000000oooo0?ooo`3o0?ooocT0oooo00<000000?ooo`3oool0403oool00480oooo00<00000
0?ooo`3oool04`3oool5000000D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0
oooo0100oooo00120?ooo`800000503oool5000000D0oooo00<000000?ooo`3oool0o`3ooolg0?oo
o`<000004P3oool00480oooo00<000000?ooo`3oool0503oool300000?l0oooo@P3oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo
@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?oo
o`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo
@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000001H0oooo00<00000
0?ooo`3oool0o`3ooolo0?ooo`<000004P3oool00480oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3oool0403oool00480oooo
00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o
0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`120?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`130?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`00@P3ooooo000005d00000
4P3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.90205, -0.287596, \
0.0432244, 0.00839264}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["NonlinearRegress", "Section"],

Cell[TextData[{
  "\t",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides several functions for fitting data, but here we discuss only ",
  ButtonBox["NonlinearRegress",
    ButtonStyle->"RefGuideLink"],
  " because it is the most general and subsumes the other functions as \
special cases.  The basic syntax and list of options are given in the \
reference guide, obtained by following the highlighted link.  We illustrate \
some of the options by returning to an earlier example.  For use in this \
function, we provide data in the form ",
  Cell[BoxData[
      \(TraditionalForm\`{{x\_1, y\_1}, \[CenterEllipsis], {x\_N, y\_N}}\)]],
  " and a separate list of weights ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Sigma]\_1\^\(-2\), \[CenterEllipsis], \[Sigma]\_N\
\^\(-2\)}\)]],
  ".  The fitting function is given explicitly, followed by a list of \
independent variables (here only 1) and a list of parameters.  The output is \
returned as a list of replacement rules."
}], "Text"],

Cell[BoxData[{
    \(\(x = Range[\(-0.9\), 0.9, 0.1];\)\), "\n", 
    \(\(y = 
        117\  - \ 23  x + 15  x\^2 + 
          Table[Random[
              NormalDistribution[0, 2. ]], {Length[
                x]}];\)\), "\[IndentingNewLine]", 
    \(\(dy = Table[2. , {Length[y]}];\)\), "\[IndentingNewLine]", 
    \(\(data = Transpose[{x, y}];\)\), "\[IndentingNewLine]", 
    \(\(w = 1/dy\^2;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(Clear[a, b, c, z];\)\), "\n", 
    \(NonlinearRegress[data, a + b\ z + c\ z\^2, z, {a, b, c}, 
      Weights \[Rule] w]\)}], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\(BestFitParameters \[Rule] {a \[Rule] 117.64946583990914`, 
            b \[Rule] \(-23.1180758142575`\), 
            c \[Rule] 14.644683708627799`}\), ",", 
        RowBox[{"ParameterCITable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Estimate\"\>", "\<\"Asymptotic SE\"\>", "\<\
\"CI\"\>"},
                {"a", "117.64946583990914`", 
                  "0.7926122273910067`", \({115.96920297883538`, 
                    119.3297287009829`}\)},
                {"b", \(-23.1180758142575`\), 
                  "0.9624927815957984`", \({\(-25.15846936242464`\), \
\(-21.077682266090363`\)}\)},
                {"c", "14.644683708627799`", 
                  "1.9729178521574118`", \({10.462284698911088`, 
                    18.82708271834451`}\)}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ 
              BoxForm`e$, TableDepth -> 2, TableHeadings -> {{a, b, c}, {
                "Estimate", "Asymptotic SE", "CI"}}]]]}], 
        ",", \(EstimatedVariance \[Rule] 1.3201091053392249`\), ",", 
        RowBox[{"ANOVATable", "\[Rule]", 
          InterpretationBox[GridBox[{
                {"\<\"\"\>", "\<\"DF\"\>", "\<\"SumOfSq\"\>", \
"\<\"MeanSq\"\>"},
                {"\<\"Model\"\>", "3", "71583.01730285482`", 
                  "23861.00576761827`"},
                {"\<\"Error\"\>", "16", "21.121745685427598`", 
                  "1.3201091053392249`"},
                {"\<\"Uncorrected Total\"\>", "19", 
                  "71604.13904854024`", "\<\"\"\>"},
                {"\<\"Corrected Total\"\>", "18", 
                  "855.442884482496`", "\<\"\"\>"}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            
            TableForm[ {{3, 71583.017302854816, 23861.00576761827}, {16, 
              21.121745685427598, 1.3201091053392249}, {19, 
              71604.139048540237}, {18, 855.44288448249597}}, 
              TableHeadings -> {{"Model", "Error", "Uncorrected Total", 
              "Corrected Total"}, {"DF", "SumOfSq", "MeanSq"}}]]}], ",", 
        RowBox[{"AsymptoticCorrelationMatrix", "\[Rule]", 
          TagBox[
            RowBox[{"(", "\[NoBreak]", GridBox[{
                  {"1.`", 
                    "6.060561463845374`*^-16", \(-0.7467401273829244`\)},
                  {"6.060561463845376`*^-16", 
                    "0.9999999999999999`", \(-7.40325885658818`*^-16\)},
                  {\(-0.7467401273829246`\), \(-7.403258856588182`*^-16\), 
                    "1.`"}
                  },
                RowSpacings->1,
                ColumnSpacings->1,
                ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
            Function[ BoxForm`e$, 
              MatrixForm[ BoxForm`e$]]]}], ",", 
        RowBox[{"FitCurvatureTable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Curvature\"\>"},
                {"\<\"Max Intrinsic\"\>", "0"},
                {"\<\"Max Parameter-Effects\"\>", "0"},
                {"\<\"95. % Confidence Region\"\>", "0.5556523300254959`"}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ 
              BoxForm`e$, TableDepth -> 2, TableHeadings -> {{"Max Intrinsic",
                 "Max Parameter-Effects", "95. % Confidence Region"}, {
                "Curvature"}}]]]}]}], "}"}]], "Output"]
}, Closed]],

Cell[TextData[{
  "Obviously, ",
  StyleBox["BestFitParameters", "InlineInput",
    FontWeight->"Bold"],
  " returns the fitted parameters.  The ",
  StyleBox["ParameterCITable", "InlineInput",
    FontWeight->"Bold"],
  " returns the standard errors (",
  Cell[BoxData[
      \(TraditionalForm\`\[Sigma]\)]],
  " for each parameter) and a ",
  StyleBox["confidence interval",
    FontSlant->"Italic"],
  " that should include the true model parameters for 95% of data samples \
from the parent distribution.  The desired confidence level is an adjustable \
option.  The ",
  StyleBox["ANOVATable", "InlineInput",
    FontWeight->"Bold"],
  " is a more advanced feature, standing for ",
  StyleBox["analysis of variance",
    FontSlant->"Italic"],
  ", that we do not discuss; nor do we discuss the curvature table.  The \
correlation matrix assesses correlations between fitted parameters with \
nondiagonal elements in the range ",
  Cell[BoxData[
      \(TraditionalForm\`\((\(-1\), 1)\)\)]],
  "."
}], "Text"],

Cell[TextData[{
  "\tThe contents of the ",
  StyleBox["regression report",
    FontSlant->"Italic"],
  " can be expanded or compressed; the default report above is labelled a \
summary report.  The quantities that can be included in the regression report \
are listed below; hopefully, most of the names should be relatively clear."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(RegressionReportValues[NonlinearRegress]\)], "Input"],

Cell[BoxData[
    \({ANOVATable, AsymptoticCorrelationMatrix, AsymptoticCovarianceMatrix, 
      BestFit, BestFitParameters, EstimatedVariance, FitCurvatureTable, 
      FitResiduals, HatDiagonal, MeanPredictionCITable, ParameterBias, 
      ParameterConfidenceRegion, ParameterCITable, ParameterTable, 
      PredictedResponse, SinglePredictionCITable, StartingParameters, 
      StandardizedResiduals, SummaryReport}\)], "Output"]
}, Closed]],

Cell["\<\
Below we use the regression report to compare the fit to the data and to \
display the residuals.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(results = 
      NonlinearRegress[data, a + b\ z + c\ z\^2, z, {a, b, c}, 
        Weights \[Rule] w, 
        RegressionReport \[Rule] {BestFitParameters, ParameterCITable, 
            EstimatedVariance, BestFit, FitResiduals, 
            StandardizedResiduals}]\)], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\(BestFitParameters \[Rule] {a \[Rule] 117.64946583990914`, 
            b \[Rule] \(-23.1180758142575`\), 
            c \[Rule] 14.644683708627799`}\), ",", 
        RowBox[{"ParameterCITable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Estimate\"\>", "\<\"Asymptotic SE\"\>", "\<\
\"CI\"\>"},
                {"a", "117.64946583990914`", 
                  "0.7926122273910067`", \({115.96920297883538`, 
                    119.3297287009829`}\)},
                {"b", \(-23.1180758142575`\), 
                  "0.9624927815957984`", \({\(-25.15846936242464`\), \
\(-21.077682266090363`\)}\)},
                {"c", "14.644683708627799`", 
                  "1.9729178521574118`", \({10.462284698911088`, 
                    18.82708271834451`}\)}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ 
              BoxForm`e$, TableDepth -> 2, TableHeadings -> {{a, b, c}, {
                "Estimate", "Asymptotic SE", "CI"}}]]]}], 
        ",", \(EstimatedVariance \[Rule] 1.3201091053392249`\), 
        ",", \(BestFit \[Rule] \(\(117.64946583990914`\)\(\[InvisibleSpace]\)\
\) - 23.1180758142575`\ z + 14.644683708627799`\ z\^2\), 
        ",", \(FitResiduals \[Rule] {\(-1.559681713870873`\), 
            5.424593019663604`, \(-0.6075350173215668`\), \
\(-3.08553987041725`\), \(-2.590645455306202`\), 
            1.7325228958756327`, \(-0.6790904465835155`\), \
\(-0.48790758799263756`\), 0.9019645888758276`, \(-0.37718015642644787`\), 
            3.234583563746682`, \(-0.6173905794929055`\), \
\(-2.370356629202405`\), \(-1.468273136272515`\), 1.1681859901592873`, 
            2.9045945429154045`, \(-0.058499863543787`\), 
            0.23617406345357495`, \(-1.7005182082603056`\)}\), 
        ",", \(StandardizedResiduals \[Rule] {\(-0.8665264276961709`\), 
            2.7260756595029165`, \(-0.2893656941797566`\), \
\(-1.4301154403994136`\), \(-1.187272045985008`\), 
            
            0.7926029632597946`, \(-0.3116704780130289`\), \
\(-0.224999069725845`\), 0.41756261548558016`, \(-0.17487170380523887`\), 
            1.4974436796550052`, \(-0.2847102801063636`\), \
\(-1.0878818681688087`\), \(-0.6717127037424657`\), 0.5353702753059693`, 
            1.3462491746578775`, \(-0.02786317354742548`\), 
            0.11868694360164439`, \(-0.9447722282894516`\)}\)}], 
      "}"}]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        ErrorListPlot[Transpose[{x, y, dy}]], \[IndentingNewLine]Plot[
          Evaluate[BestFit /. results], {z, \(-1\), 1}], Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 -1.21142 0.0116769 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .07304 -22 -7.84375 ]
[-0.0125 .07304 0 7.84375 ]
[-0.0125 .18981 -22 -7.84375 ]
[-0.0125 .18981 0 7.84375 ]
[-0.0125 .30658 -22 -7.84375 ]
[-0.0125 .30658 0 7.84375 ]
[-0.0125 .42335 -22 -7.84375 ]
[-0.0125 .42335 0 7.84375 ]
[-0.0125 .54012 -22 -7.84375 ]
[-0.0125 .54012 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -37.6875 -5.09375 ]
[-0.01875 .30902 -22 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .07304 m
.00625 .07304 L
s
gsave
-0.0125 .07304 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(110) show
1.000 setlinewidth
grestore
0 .18981 m
.00625 .18981 L
s
gsave
-0.0125 .18981 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(120) show
1.000 setlinewidth
grestore
0 .30658 m
.00625 .30658 L
s
gsave
-0.0125 .30658 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(130) show
1.000 setlinewidth
grestore
0 .42335 m
.00625 .42335 L
s
gsave
-0.0125 .42335 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(140) show
1.000 setlinewidth
grestore
0 .54012 m
.00625 .54012 L
s
gsave
-0.0125 .54012 -83 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(150) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .0964 m
.00375 .0964 L
s
0 .11975 m
.00375 .11975 L
s
0 .1431 m
.00375 .1431 L
s
0 .16646 m
.00375 .16646 L
s
0 .21317 m
.00375 .21317 L
s
0 .23652 m
.00375 .23652 L
s
0 .25987 m
.00375 .25987 L
s
0 .28323 m
.00375 .28323 L
s
0 .32994 m
.00375 .32994 L
s
0 .35329 m
.00375 .35329 L
s
0 .37664 m
.00375 .37664 L
s
0 .4 m
.00375 .4 L
s
0 .44671 m
.00375 .44671 L
s
0 .47006 m
.00375 .47006 L
s
0 .49341 m
.00375 .49341 L
s
0 .51677 m
.00375 .51677 L
s
0 .04969 m
.00375 .04969 L
s
0 .02634 m
.00375 .02634 L
s
0 .00298 m
.00375 .00298 L
s
0 .56347 m
.00375 .56347 L
s
0 .58683 m
.00375 .58683 L
s
0 .61018 m
.00375 .61018 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -95.9375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .07304 m
1 .07304 L
s
.99375 .18981 m
1 .18981 L
s
.99375 .30658 m
1 .30658 L
s
.99375 .42335 m
1 .42335 L
s
.99375 .54012 m
1 .54012 L
s
.125 Mabswid
.99625 .0964 m
1 .0964 L
s
.99625 .11975 m
1 .11975 L
s
.99625 .1431 m
1 .1431 L
s
.99625 .16646 m
1 .16646 L
s
.99625 .21317 m
1 .21317 L
s
.99625 .23652 m
1 .23652 L
s
.99625 .25987 m
1 .25987 L
s
.99625 .28323 m
1 .28323 L
s
.99625 .32994 m
1 .32994 L
s
.99625 .35329 m
1 .35329 L
s
.99625 .37664 m
1 .37664 L
s
.99625 .4 m
1 .4 L
s
.99625 .44671 m
1 .44671 L
s
.99625 .47006 m
1 .47006 L
s
.99625 .49341 m
1 .49341 L
s
.99625 .51677 m
1 .51677 L
s
.99625 .04969 m
1 .04969 L
s
.99625 .02634 m
1 .02634 L
s
.99625 .00298 m
1 .00298 L
s
.99625 .56347 m
1 .56347 L
s
.99625 .58683 m
1 .58683 L
s
.99625 .61018 m
1 .61018 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .07304 m
1 .07304 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.002 w
.07143 .50227 m
.07143 .54897 L
s
.015 w
.07143 .52562 Mdot
.002 w
.11905 .52776 m
.11905 .57446 L
s
.015 w
.11905 .55111 Mdot
.002 w
.16667 .40467 m
.16667 .45138 L
s
.015 w
.16667 .42803 Mdot
.002 w
.21429 .32651 m
.21429 .37322 L
s
.015 w
.21429 .34987 Mdot
.002 w
.2619 .28649 m
.2619 .33319 L
s
.015 w
.2619 .30984 Mdot
.002 w
.30952 .29458 m
.30952 .34129 L
s
.015 w
.30952 .31794 Mdot
.002 w
.35714 .22746 m
.35714 .27416 L
s
.015 w
.35714 .25081 Mdot
.002 w
.40476 .19414 m
.40476 .24085 L
s
.015 w
.40476 .2175 Mdot
.002 w
.45238 .17825 m
.45238 .22496 L
s
.015 w
.45238 .2016 Mdot
.002 w
.5 .13461 m
.5 .18131 L
s
.015 w
.5 .15796 Mdot
.002 w
.54762 .1515 m
.54762 .1982 L
s
.015 w
.54762 .17485 Mdot
.002 w
.59524 .08465 m
.59524 .13136 L
s
.015 w
.59524 .10801 Mdot
.002 w
.64286 .04574 m
.64286 .09245 L
s
.015 w
.64286 .06909 Mdot
.002 w
.69048 .04125 m
.69048 .08796 L
s
.015 w
.69048 .0646 Mdot
.002 w
.7381 .06043 m
.7381 .10714 L
s
.015 w
.7381 .08378 Mdot
.002 w
.78571 .07252 m
.78571 .11923 L
s
.015 w
.78571 .09587 Mdot
.002 w
.83333 .03316 m
.83333 .07986 L
s
.015 w
.83333 .05651 Mdot
.002 w
.88095 .03525 m
.88095 .08196 L
s
.015 w
.88095 .05861 Mdot
.002 w
.92857 .01472 m
.92857 .06142 L
s
.015 w
.92857 .03807 Mdot
.5 Mabswid
.02381 .60332 m
.06244 .55479 L
.10458 .50444 L
.14415 .45959 L
.18221 .41868 L
.22272 .37754 L
.26171 .34027 L
.30316 .30318 L
.34309 .26989 L
.3815 .24013 L
.42237 .21092 L
.46172 .18517 L
.49955 .16262 L
.53984 .14098 L
.57861 .12246 L
.61984 .10526 L
.65954 .09112 L
.69774 .07976 L
.73838 .07008 L
.77751 .06312 L
.79892 .06029 L
.80873 .05923 L
.81909 .05826 L
.82867 .05751 L
.83917 .05685 L
.84873 .05639 L
.85336 .05621 L
.85765 .05608 L
.86265 .05596 L
.86529 .05591 L
.8681 .05588 L
.86926 .05586 L
.87052 .05585 L
.87171 .05584 L
.8728 .05584 L
.87401 .05583 L
.87533 .05583 L
.876 .05583 L
.87671 .05583 L
.878 .05583 L
.87925 .05584 L
.8804 .05585 L
.88166 .05586 L
.883 .05587 L
.88525 .0559 L
.88772 .05594 L
.89228 .05603 L
.89655 .05615 L
.90696 .05656 L
.91684 .0571 L
Mistroke
.93538 .0585 L
.95522 .06058 L
.97362 .06304 L
.97619 .06342 L
Mfstroke
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00lP3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo]P3oool00?@0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;D0oooo003e0?ooo`040;mdM02LW4P004RL0=oookL0
oooo003e0?ooo`030?oo_`1d000007Bo0;P0oooo003e0?ooo`030?oo_`1d0000M;oo0;P0oooo003e
0?ooo`040?noM000000004RL0=oookL0oooo003e0?ooo`050;md0000B9`0gmnL04Q8W03Oool0]P3o
ool00?<0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]03oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005<0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`160?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
od40oooo00@0_g@0000000000000M;ooB03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom50?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool04@3oool005D0oooo00<0okmd0000001d_ol0B03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_d00oooo00H0gia8001d_`3oool0omnL04P00018W=m60?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
A`3oool00`3o_g@0000007Boo`0D0?ooo`00E@3oool00`3ooml0W4P0001d_`170?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0m0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0A03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0A@3oool00`3ooml0W4P0001d_`0D0?ooo`00EP3oool0
0`3OW4P004RL0=ooo`160?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0m0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
@`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0AP3oool00`3OW4P004RL0=ooo`0C0?ooo`00AP3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0>@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03d0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`130?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`160?ooo`030?oOW01804P0W=oo
01<0oooo001F0?ooo`030?oo_`1d0000M;oo04L0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03h0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`140?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`160?ooo`030?oo_`1d0000M;oo01<0oooo001G0?ooo`030;md0000M;l0oooo04H0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03h0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`140?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`170?ooo`030;md0000M;l0oooo0180oooo
001G0?ooo`030=nLB000B9`0gooo04H0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo?P3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo04<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_oooA`3oool00`3OW4P004RL0=ooo`0B0?ooo`00E`3oool00`3o
gi`0B01809cOo`160?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`130?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04P0oooo00<0omnL04P0B02Lgol04P3oool005L0oooo00<0oono07@0001d
_ol0A`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0@@3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo04@0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04T0oooo00<0oono07@0001d
_ol04P3oool005P0oooo00<0_g@0001d_`3oool0A`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0@P3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgdH0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04X0oooo00<0_g@0001d_`3o
ool04@3oool005H0oooo00D0oono07@000000000001d0;ooo`180?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`100?ooo`040=nLB00000000000001d_dP0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO04D0oooo00D0oono07@00000
0000001d0;ooo`0A0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00B@3ooooo000005d000002`3oool004T0oooo0P0000060?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`8000002`3oool004T0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo04l0oooo
00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<00000
0?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?oo
o`030000003oool0oooo0980oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?oo
o`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo0980oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0980oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool200000:/0oooo00<000000?ooo`3oool0TP3oool0
0`000000oooo0?ooo`0E0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`00
0000oooo0?ooo`2B0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`2A0?ooo`<000005`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0740oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0;0?ooo`D000005P3oool00`000000oooo
0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0740oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`0;0?ooo`D000005P3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0400oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00/0oooo1@00000F
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?oo
o`3oool0@03oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00h0oooo00<000000?oo
o`3oool0303oool3000001L0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3o
ool0ZP3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0;P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`090?ooo`00B@3oool200000:/0oooo00<000000?ooo`3oool0
;`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01D0oooo0P00000;0?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`0]0?ooo`<000003P3oool3000000h0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0/0?ooo`D000001P3oool=
000000/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`090?ooo`009`3oool00`3o
okl0M000000000020000000604RLg`3oool0oooo0?ooo`3ookl0M0000`0000000`18W=l0oooo0?oo
o`030?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo
0?ooo`0_0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0:03oool?000000H0oooo
1@0000020?oooa4000004`3oool00`000000oooo0?ooo`090?ooo`00:@3oool00`3o_g@0000007Bo
o`060?ooo`030?noM0000000M;oo00D0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`03
0000003oool0oooo0:X0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0<0?ooo`<0
00003`3oool00`000000oooo0?ooo`0?0?ooo`T00000103oool5000000`0oooo1@00000=0?ooo`03
0000003oool0oooo00<0oooo1`00000<0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`030?oo
g`2LB00007Bo00H0oooo00<0oooO09a80000M;l0103oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00P0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0^0?ooo`<00000
303oool5000000h0oooo00<000000?ooo`3oool0203oool7000000h0oooo0`00000>0?ooo`<00000
6`3oool5000000L0oooo00<000000?ooo`3oool02@3oool002X0oooo00<0gia80018W03Oool01P3o
ool00`3OW4P004RL0=ooo`030?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
1`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02d0oooo1@00000;0?ooo`D00000
3P3oool010000000oooo0?ooo`3oool7000001H0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo00T0oooo000Z0?ooo`030?oOW01804P0W=oo00H0oooo
00<0omnL04P0B02Lgol00`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00L0oooo
00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0]0?ooo`D000002`3oool5000000d0oooo
1@00000M0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool09P3oool00`000000oooo
0?ooo`090?ooo`00:P3oool00`3ookl0M00007Boo`060?ooo`030?oo_`1d0000M;oo00@0oooo00<0
_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`070?ooool00000G@00000;0?ooo`00:`3oool0
0`2oM00007Bo0?ooo`060?ooo`030;md0000M;l0oooo00<0oooo00<0gia80000B02Lgol00P3oool0
0`3ooml0W4P0001d_`070?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0;P3oool3
000000l0oooo00<000000?ooo`3oool01@3oool3000000D0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0?ooo`090?ooo`00:`3oool00`3OW4P004RL0=ooo`060?ooo`030=nLB000B9`0gooo
00<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo
0:X0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0>0?ooo`040000003oool0oooo
0?ooo`@000001`3oool3000000l0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool02@3oool002/0oooo00<0omnL04P0
B02Lgol01P3oool00`3ogi`0B01809cOo`030?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia8
0018W03Oool01P3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02l0oooo00<00000
0?ooo`3oool03P3oool4000000X0oooo1@00000>0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00T0oooo000[0?oo
o`030?oo_`1d0000M;oo00H0oooo00<0oono07@0001d_ol0103oool00`3ogi`0B00007Boo`020?oo
o`030=nLB000M;l0oooo00H0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0N0?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0303oool3000000d0oooo1@00000>0?oo
o`030000003oool0oooo04L0oooo00<000000?ooo`3oool02@3oool002`0oooo00<0_g@0001d_`3o
ool01P3oool00`2oM00007Bo0?ooo`040?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool0
0`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03P3oool0
0`000000oooo0?ooo`080?ooo`@000000P3oool00`000000oooo0?ooo`0;0?ooo`D000003P3oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo00T0oooo000Z0?ooo`050?oo_`1d00000000
0000M02oool0103oool01@3ookl0M0000000000007@0_ooo00D0oooo00@0gia800000000000007Bo
203oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`050?ooo`<000001P3oool00`000000oooo0?ooo`0<0?ooo`<00000
3P3oool3000004P0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00P3o
ool3000001T0oooo00<000000?ooo`3oool0303oool5000004L0oooo00<000000?ooo`3oool02@3o
ool004T0oooo0P00002[0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0403oool3
000001`0oooo00<000000?ooo`3oool0303oool5000004H0oooo0P00000;0?ooo`00B@3oool00`00
0000oooo0?ooo`2Z0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool03P3oool20000
01l0oooo00<000000?ooo`3oool0303oool5000004L0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo00/0
oooo0`00000Q0?ooo`030000003oool0oooo00d0oooo0`0000180?ooo`030000003oool0oooo00T0
oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool07@3oool3000000T0oooo
0`00000T0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0A`3oool00`000000oooo
0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01`0oooo1@00
00050?ooo`<000009`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo04L0oooo00<0
00000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0L
0?ooo`D000000`3oool2000003/0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`09
0?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01`0oooo1@0000000`3o
ool000000000000m0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0M0?ooo`D00000?`3oool00`000000
oooo0?ooo`170?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo
00<000000?ooo`3oool07@3oool200000480oooo00<000000?ooo`3oool0A`3oool00`000000oooo
0?ooo`090?ooo`00B@3oool200000:/0oooo00<000000?ooo`3oool06`3oool2000000030?ooo`00
0000oooo08X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo01P0oooo0`0000030?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3oool02@3oool0
04T0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0F0?ooo`8000001P3oool00`00
0000oooo0?ooo`2:0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0
oooo00<000000?ooo`3oool0503oool2000000P0oooo00<000000?ooo`3oool0RP3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo0140oooo
0`00000:0?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3oool02@3oool004T0oooo00<0
00000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0?0?ooo`800000VP3oool00`000000oooo0?oo
o`090?ooo`00B@3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo00d0oooo0P00002L
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo0:X0oooo00<000000?oo
o`3oool02`3oool2000009h0oooo00<000000?ooo`3oool02@3oool004T0oooo0P00002[0?ooo`03
0000003oool0oooo00T0oooo0P00002O0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0
ZP3oool00`000000oooo0?ooo`070?ooo`800000XP3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo00D0oooo0P00002T0?ooo`030000003o
ool0oooo00T0oooo00190?ooo`030000003oool0oooo0:T0oooo0`0000040?ooo`8000002@3oool0
0`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo
0:P0oooo1@0000000`3oool000000000000;0?ooo`030000003oool0oooo09X0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0Z03oool6000000d0oooo00<000000?ooo`3o
ool0VP3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2X0?ooo`D00000
3P3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003o
ool0oooo0:P0oooo1000000?0?ooo`030000003oool0oooo09X0oooo00<000000?ooo`3oool02@3o
ool004T0oooo0P00002W0?ooo`8000000P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo09T0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool03@3oool3000009/0oooo00<000000?ooo`3oool02@3oool004T0
oooo00<000000?ooo`3oool0X`3oool2000000D0oooo00<000000?ooo`3oool0303oool5000009X0
oooo00<000000?ooo`3oool02@3oool002L0oooo00<0oono07@0000000000P0000001@18W=l0oooo
0?ooo`3oool0gia800D0000000<0B9cO0?ooo`3oool00P3oool0102oM00000000000001d_ol:0?oo
o`030000003oool0oooo0:80oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?oo
o`D00000VP3oool00`000000oooo0?ooo`090?ooo`00:@3oool00`3o_g@0000007Boo`050?ooo`03
0?noM000M;l0oooo0080oooo00X0_gAd0;ooo`3oool0oooo0=nLB000M;l0oooo0?oOW0180000B9cO
2@3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00<0oooo0P0000080?ooo`030000
003oool0oooo00`0oooo1@00002J0?ooo`030000003oool0oooo00T0oooo000Y0?ooo`030?oog`2L
B00007Bo00H0oooo00<0okmd001d_`3oool0103oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00P0oooo00<000000?ooo`3oool0VP3oool010000000oooo0?ooo`3oool2000000X0oooo
00<000000?ooo`3oool03@3oool3000009/0oooo00<000000?ooo`3oool02@3oool002X0oooo00<0
gia80018W03Oool01P3oool00`3o_g@007Bo0?ooo`030?ooo`030?oog`2LB000B9cO0080oooo00@0
oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`2J0?ooo`040000003oool0oooo000000`0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00T0
oooo000Z0?ooo`030?oOW01804P0W=oo00L0oooo00<0okmd04RLg`3oool00P3oool00`3ooml0W4P0
001d_`030?ooo`030=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool0VP3oool3000000d0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00T0oooo
000Z0?ooo`030?oo_`1d0000M;oo00P0oooo00<0gia804RLg`3oool00P3oool00`2oM00004RL0=oo
o`020?ooo`030?oOW0180000B9cO00L0oooo0`00002J0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`2H0?ooo`<000002`3oool002/0oooo00<0_g@0
001d_`3oool0203oool01`2oM000M;oo0?ooo`3oool0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`070?ooo`030000003oool0oooo08T0oooo00<000000?ooo`3oool0303oool3000000l0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo00T0oooo
000[0?ooo`030=nLB000B9`0gooo00P0oooo00L0omnL04P0M02oool0oooo0?oOW0180000B9cO00<0
oooo00<0_g@00018W03Oool01P3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`0000000P00000>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3o
ool0VP3oool00`000000oooo0?ooo`090?ooo`00:`3oool00`3ogi`0B01809cOo`080?ooo`080?oo
g`2LB000M;oo0?ooo`3ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00<00000
0?ooo`3oool0R@3oool00`000000oooo0?ooo`090?ooo`80000000<0oooo0000000000000`00000=
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0VP3oool00`000000oooo0?ooo`09
0?ooo`00:`3oool00`3ookl0M00007Boo`080?ooo`030?oo_`1d0000B9cO0080oooo00<0omnL04P0
001d_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0oooo08T0oooo00<000000?oo
o`3oool01`3oool2000000<0oooo1@00000=0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool002`0oooo00<0_g@0001d_`3oool00`3oool01`3o_g@004RL0=ooo`3oool0gia8
0000B02Lgol00`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<000000?ooo`3oool0
R@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00<0oooo1@00000=0?ooo`030000
003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00D0oono07@000000000001d
0;ooo`040?ooo`050?oOW018000000000000001d_ol01@3oool0103OW4P0000000000000M;l80?oo
o`030000003oool0oooo08T0oooo00<000000?ooo`3oool0103oool2000000L0oooo0`00000>0?oo
o`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3o
ool0R@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`800000
R@3oool300000080oooo0P00000;0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
ZP3oool2000000/0oooo00190?ooo`030000003oool0oooo08L0oooo1P00000=0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`270?ooo`D000003P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0Q`3oool50000
00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo
00T0oooo00190?ooo`030000003oool0oooo08P0oooo0`00000?0?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo
0?ooo`1i0?ooo`030000003oool0oooo00X0oooo0P0000000`3oool000000?ooo`0O0?ooo`030000
003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0N@3o
ool00`000000oooo0?ooo`090?ooo`050000003oool0oooo0?ooo`000000803oool00`000000oooo
0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo07T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo01h0oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2000007X0oooo00<00000
0?ooo`3oool01P3oool2000000D0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`2Z
0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo07T0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1h0?ooo`<00000
103oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool004T0oooo00<000000?ooo`3oool0M`3oool5000000030?ooo`000000000002/0oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1g
0?ooo`H00000;@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`030000003oool0oooo07L0oooo1@00000^0?ooo`030000003oool0oooo0:/0oooo00<000000?oo
o`3oool02@3oool004T0oooo00<000000?ooo`3oool0N03oool3000002l0oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2000007X0oooo00<000000?ooo`3oool0
;P3oool00`000000oooo0?ooo`2Z0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0M`3o
ool300000300oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`1f0?ooo`040000003oool0oooo00000300oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1e0?ooo`050000003oool0
oooo0?ooo`000000<03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0^
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0L`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02h0oooo00<000000?oo
o`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`1a0?ooo`80
0000=`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000
003oool0oooo0700oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`2[0?ooo`030000
003oool0oooo00T0oooo00190?ooo`800000L03oool00`000000oooo0?ooo`0h0?ooo`030000003o
ool0oooo0:X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0
oooo00T0oooo00190?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool04@3oool20000
03`0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`009`3oool00`3ookl0
M000000000020000000604RLg`3oool0oooo0?ooo`3ookl0M0000`0000000`1d_ol0oooo0?ooo`03
0?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00:@3oool00`3o_g@0000007Boo`040?ooo`030?noM000000007Bo0080oooo00<0_gA8
09cOo`3oool00P3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`T0oooo00<000000?ooo`3oool0
F03oool00`000000oooo0?ooo`0=0?ooo`030000003oool0000003l0oooo00<000000?ooo`3oool0
Z`3oool00`000000oooo0?ooo`090?ooo`007P3oool00`2oM00004RL0=ooo`080?ooo`030?oog`2L
B00007Bo00T0oooo00L0omnL04Q8W03Oool0oooo0?oog`2LB000B9cO0080oooo00<0okmd0000B02L
gol0203oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo00d0oooo0P0000100?ooo`03
0000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool001h0oooo00<0omnL04Qd_`3oool0
2@3oool00`3OW4P004RL0=ooo`080?ooo`070?oo_`1d0000M;oo0?ooo`3ooml0W4P004RLg`020?oo
o`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0<
0?ooo`800000@@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000C0?oo
o`030?oo_`1dM;l0oooo00L0oooo00<0okmd0018W03Oool02P3oool00`3ogi`0B01809cOo`080?oo
o`070?oOW0180000M;oo0?ooo`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00L0oooo00<0
00000?ooo`3oool0F03oool00`000000oooo0?ooo`0;0?ooo`030000003oool000000440oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`004`3oool0103ogi`0B0000000M02o
ool50?ooo`030=nLB000M;l0oooo00/0oooo00<0oono07@0001d_ol01`3oool0103ogi`0B0000000
B02Lgol20?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool3000005L0oooo
0`00000;0?ooo`040000003oool0oooo00000440oooo00<000000?ooo`3oool0Z@3oool3000000/0
oooo000D0?ooo`030?oOW0180000000000D0000000<0B9cO0?ooo`3oool0303oool00`2oM00007Bo
0?ooo`040?ooo`050?oOW018000000000000001d_ol00`3oool00`3OW4P0001809cOo`020?ooo`03
0?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0EP3oool5000000T0oooo00D000000?ooo`3o
ool0oooo000000110?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool001T0
oooo00<0okmd0000B02Lgol03`3oool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d000007Bo00<0
oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo05H0
oooo1@0000080?ooo`040000003oool0oooo0?ooo`<00000@03oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo000H0?ooo`030;md0000M;l0oooo0100oooo00<0omnL04P0B02L
gol0203oool0203OW4P007Bo0?ooo`3oool0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=oo
o`060?ooo`030000003oool0oooo05H0oooo1@0000070?ooo`040000003oool0oooo0?ooo`D00000
?`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000C0?ooo`060?oOW018
0000M;oo0?noM0000000M;oo4P3oool00`3ookl0M00007Boo`080?ooo`030?oOW01804P0W=oo0080
oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0oooo05L0
oooo0`0000070?ooo`030000003oool0oooo0080oooo1@00000o0?ooo`030000003oool0oooo0:/0
oooo00<000000?ooo`3oool02@3oool001<0oooo00@0omnL04P00000000007Bo5@3oool00`2oM000
07Bo0?ooo`040?ooo`060;md001d_ol0oooo0=nLB00004P0W=oo0`3oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00<0oooo1@00000o0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool002X0oooo00D0oono07@000000000001d0;ooo`050?ooo`040;md00000000000004RLg`D0
oooo00@0gia800000000000007Bo203oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo
00<0oooo0P0000070?ooo`<00000@03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo
00T0oooo00190?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0?`3oool00`000000
oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo04L0oooo
00<000000?ooo`3oool03P3oool01@000000oooo0?ooo`3oool0000000/0oooo00<000000?ooo`3o
ool0?`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`800000
B03oool00`000000oooo0?ooo`0>0?ooo`040000003oool0oooo000000`0oooo00<000000?ooo`3o
ool0?`3oool00`000000oooo0?ooo`2Z0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0
A`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0000000d0oooo00<000000?ooo`3oool0
?`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003o
ool0oooo04L0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo03l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0o0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool0
04T0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
0500oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`160?ooo`<000003@3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo0:/0
oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0A@3oool5000000/0oooo
00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo04D0oooo1@00000:0?ooo`030000003oool0oooo05<0oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2000004H0oooo1@000009
0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0ZP3oool2000000/0oooo00190?oo
o`030000003oool0oooo04H0oooo0`0000090?ooo`030000003oool0oooo05D0oooo00<000000?oo
o`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`170?ooo`03
0000003oool0oooo00H0oooo0P00001H0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool004T0oooo00<000000?ooo`3oool0A`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo05P0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3o
ool00`000000oooo0?ooo`170?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0F@3o
ool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0
oooo04L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0A`3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`170?ooo`050000003oool0oooo0?oo
o`000000GP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`80
0000B`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo0:X0oooo0P00000;0?ooo`00
B@3oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo05h0oooo00<000000?ooo`3oool0
Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`180?ooo`030000003o
ool0oooo05l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`170?ooo`030000003oool0oooo0600oooo00<000000?ooo`3oool0Z`3oool0
0`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo
0640oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`150?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo06<0oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`130?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?oo
o`090?ooo`00B@3oool2000004<0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`2Z
0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo06H0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?oo
o`00B@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool0I`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000W0?ooo`030?oo
_`1d000000000080000000<0B9cO0?ooo`3oool0103oool00`3ooml0W4P004RLg`050?ooo`040;md
00000000000007Boo`X0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo06P0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00
:@3oool00`3o_g@0000007Boo`080?ooo`030=nLB000B9`0gooo00<0oooo00H0gia8001d_`3oool0
omnL04P00018W=l90?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`1Y0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002T0
oooo00<0oooO09a80000M;l0203oool00`3ogi`0B01809cOo`020?ooo`030?oog`2LB000B9cO0080
oooo00<0okmd0000B02Lgol0203oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0
oooo000Z0?ooo`030=nLB000B9`0gooo00<0oooo00<0okmd000000000000100000001@00B9`0gooo
0?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`0g0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo000Z0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a8B02L
gol00P3oool01`2oM00007Bo0?ooo`3oool0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`07
0?ooo`030000003oool0oooo03L0oooo00D000000?ooo`3oool0oooo0000001^0?ooo`030000003o
ool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00<0oono07@0001d_ol0103oool0
1`3ookl0M01d0;ooo`3oool0okmd0018W03Oool00P3oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00L0oooo0`00000f0?ooo`<0000000<0oooo0000003oool0KP3oool00`000000oooo
0?ooo`2Y0?ooo`<000002`3oool002/0oooo00<0_g@0001d_`3oool0103oool01P3ogi`0B4RL0=oo
o`3ogi`0B01809cOo`80oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`070?ooo`03
0000003oool0oooo03D0oooo1@00001`0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool002/0oooo00<0gia80018W03Oool01@3oool01@3o_g@007Bo0?oog`2LB000B9cO0080
oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo03D0
oooo1@00001`0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002/0oooo
00<0omnL04P0B02Lgol01P3oool0103OW4P0B9cO0;md0000M;l20?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0gia80018W03Oool01P3oool00`000000oooo0?ooo`0e0?ooo`D00000L03oool00`00
0000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000[0?ooo`030?oo_`1d0000M;oo00L0
oooo00@0gia807AdM00007@0_ooo0P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo
00H0oooo00<000000?ooo`3oool0=P3oool300000740oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`090?ooo`00;03oool00`2oM00007Bo0?ooo`070?ooo`030;md000004P0W=oo00<0
oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`030000003oool0oooo03H0oooo0P00001b
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool002X0oooo00D0oono07@0
00000000001d0;ooo`070?ooo`030?oog`2LB000B9cO00@0oooo00@0gia800000000000007Bo203o
ool00`000000oooo0?ooo`0e0?ooo`030000003oool000000780oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0d0?ooo`040000003oool0
oooo00000780oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool2
000003@0oooo00D000000?ooo`3oool0oooo0000001b0?ooo`030000003oool0oooo0:X0oooo0P00
000;0?ooo`00B@3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0L03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`030000003oool0oooo0380oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1`0?oo
o`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3o
ool0<@3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool004T0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`1g0?ooo`030000
003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0;`3o
ool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3o
ool004T0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1i0?ooo`030000003oool0
oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0
oooo0P00000]0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool0ZP3oool2000000/0
oooo00190?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0O03oool00`000000oooo
0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo02X0oooo00<0
00000?ooo`3oool0O@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo0019
0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0OP3oool00`000000oooo0?ooo`2[
0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo02P0oooo00<000000?oo
o`3oool0O`3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`2[0?ooo`03
0000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0
P03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003o
ool0oooo02H0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2[0?ooo`030000003o
ool0oooo00T0oooo00190?ooo`8000009P3oool00`000000oooo0?ooo`220?ooo`030000003oool0
oooo0:X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo
08<0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000
oooo0?ooo`0S0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0Z`3oool00`000000
oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0Q@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo
00190?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`260?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<00000
0?ooo`3oool05P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo08L0oooo00<00000
0?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`280?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool004T0oooo0P00000G0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
R@3oool00`000000oooo0?ooo`2Z0?ooo`8000002`3oool004T0oooo00<000000?ooo`3oool05@3o
ool3000000D0oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`2[0?ooo`030000003o
ool0oooo00T0oooo00190?ooo`030000003oool0oooo01@0oooo1@0000030?ooo`030000003oool0
oooo08/0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`00
0000oooo0?ooo`0D0?ooo`D000000`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo
0:/0oooo00<000000?ooo`3oool02@3oool002L0oooo00<0oono07@0000000000P0000001P18W=l0
oooo0?ooo`3oool0omnL04P00080000000<004RL0=ooo`3oool0103oool0102oM00000000000001d
_ol:0?ooo`030000003oool0oooo01@0oooo1@0000020?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0P@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo000Y
0?ooo`030?noM0000000M;oo00@0oooo00H0okmd0000001d_ol0oooo0?oOW018M;l40?ooo`060=nL
B000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo0?ooo`0E0?ooo`<000000P3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`090?ooo`00:@3oool00`3ooml0W4P0001d_`090?ooo`070?noM018W=l0oooo0?oo
o`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool05P3oool0
10000000oooo0?ooo`00000<0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0Z`3o
ool00`000000oooo0?ooo`090?ooo`00:P3oool00`3OW4P004RL0=ooo`080?ooo`070?oo_`1d07@0
_ooo0?ooo`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0000000=0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3o
ool0Z`3oool00`000000oooo0?ooo`090?ooo`00:P3oool00`3ogi`0B01809cOo`080?ooo`070?oo
g`2LB000M;oo0?ooo`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00L0oooo0`00000F0?oo
o`030000003oool0000000d0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2Y0?oo
o`<000002`3oool002X0oooo00<0oono07@0001d_ol0203oool00`3ookl0M00004RLg`020?ooo`03
0;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool00`000000oooo0?ooo`0F0?ooo`80
00003P3oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3o
ool02@3oool002/0oooo00<0_g@0001d_`3oool01`3oool00`3o_g@0000007Boo`020?ooo`030=nL
B00004P0W=oo0080oooo00<0oooO09a80000M;l01`3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo00`0oooo0`0000220?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool0
2@3oool002/0oooo00<0gia80018W03Oool0103oool01P3ogml0gooo0?noM0000000001d0;ooo`80
oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo01D0
oooo0P00000=0?ooo`D00000P@3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0
oooo000[0?ooo`030?oOW01804P0W=oo00@0oooo00D0oono07@000000000001809cOo`030?ooo`04
0?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`000000oooo0?ooo`0E0?oo
o`8000003@3oool500000840oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?oo
o`00:`3oool00`3ookl0M00007Boo`050?ooo`030=nLW02oool0oooo00D0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3o
ool03@3oool500000840oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00
;03oool00`2oM00007Bo0?ooo`040?ooo`030?oOW01dW=l0oooo00H0oooo00H0gia80000M02oool0
oooO09a80018W=l70?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03`3oool30000
0880oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00:P3oool01@3ookl0
M0000000000007@0_ooo00D0oooo00D0_g@0000000000000000004RLg`040?ooo`040=nLB0000000
0000001d_`P0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo0840oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0P@3oool0
0`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`8000004@3oool00`00
0000oooo0?ooo`0C0?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0ZP3oool20000
00/0oooo00190?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`210?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0840oooo
00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?oo
o`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool0VP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`2[0?oo
o`030000003oool0oooo00T0oooo00190?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0W03oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`2[0?ooo`030000
003oool0oooo00T0oooo00190?ooo`8000002P3oool00`000000oooo0?ooo`2N0?ooo`030000003o
ool0oooo0:X0oooo0P00000;0?ooo`00B@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo09l0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo09l0oooo00<000000?ooo`3oool0Z`3oool00`00
0000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0:00
oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0Z`3oool00`000000oooo
0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0:80oooo00<0
00000?ooo`3oool0Z`3oool00`000000oooo0?ooo`090?ooo`00B@3oool00`000000oooo0?ooo`2Z
0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool02@3oool004T0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`2[0?ooo`030000003oool0oooo00T0oooo00190?ooo`80
00001P3oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3o
ool0D03oool00`000000oooo0?ooo`1@0?ooo`030000003oool0oooo00@0oooo0P00000;0?ooo`00
B@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0
3P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00T0oooo00190?ooool00000G@00
000;0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.49425, 91.0601, \
0.00907712, 0.37017}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        ListPlot[Evaluate[Transpose[{x, StandardizedResiduals /. results}]], 
          PlotRange \[Rule] {Automatic, {\(-3\), 3}}, 
          PlotStyle \[Rule] PointSize[0.02], Frame \[Rule] True, 
          FrameLabel \[Rule] {"\<x\>", "\<residuals\>"}], 
        Plot[{\(-1\), 1}, {x, \(-1\), 1}, 
          PlotStyle \[Rule] Dashing[{0.02, 0.02}]]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.103006 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .10301 -18.6875 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -18.6875 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -10 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[-0.0125 .61803 -10 -7.84375 ]
[-0.0125 .61803 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -34.375 -23.9688 ]
[-0.01875 .30902 -18.6875 23.9688 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
0 .61803 m
.00625 .61803 L
s
gsave
-0.0125 .61803 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .0206 m
.00375 .0206 L
s
0 .0412 m
.00375 .0412 L
s
0 .0618 m
.00375 .0618 L
s
0 .0824 m
.00375 .0824 L
s
0 .12361 m
.00375 .12361 L
s
0 .14421 m
.00375 .14421 L
s
0 .16481 m
.00375 .16481 L
s
0 .18541 m
.00375 .18541 L
s
0 .22661 m
.00375 .22661 L
s
0 .24721 m
.00375 .24721 L
s
0 .26781 m
.00375 .26781 L
s
0 .28842 m
.00375 .28842 L
s
0 .32962 m
.00375 .32962 L
s
0 .35022 m
.00375 .35022 L
s
0 .37082 m
.00375 .37082 L
s
0 .39142 m
.00375 .39142 L
s
0 .43262 m
.00375 .43262 L
s
0 .45322 m
.00375 .45322 L
s
0 .47383 m
.00375 .47383 L
s
0 .49443 m
.00375 .49443 L
s
0 .53563 m
.00375 .53563 L
s
0 .55623 m
.00375 .55623 L
s
0 .57683 m
.00375 .57683 L
s
0 .59743 m
.00375 .59743 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -111.5 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
84.9688 11.8438 translate
90 rotate
-84.9688 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(residuals) show
106.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .0206 m
1 .0206 L
s
.99625 .0412 m
1 .0412 L
s
.99625 .0618 m
1 .0618 L
s
.99625 .0824 m
1 .0824 L
s
.99625 .12361 m
1 .12361 L
s
.99625 .14421 m
1 .14421 L
s
.99625 .16481 m
1 .16481 L
s
.99625 .18541 m
1 .18541 L
s
.99625 .22661 m
1 .22661 L
s
.99625 .24721 m
1 .24721 L
s
.99625 .26781 m
1 .26781 L
s
.99625 .28842 m
1 .28842 L
s
.99625 .32962 m
1 .32962 L
s
.99625 .35022 m
1 .35022 L
s
.99625 .37082 m
1 .37082 L
s
.99625 .39142 m
1 .39142 L
s
.99625 .43262 m
1 .43262 L
s
.99625 .45322 m
1 .45322 L
s
.99625 .47383 m
1 .47383 L
s
.99625 .49443 m
1 .49443 L
s
.99625 .53563 m
1 .53563 L
s
.99625 .55623 m
1 .55623 L
s
.99625 .57683 m
1 .57683 L
s
.99625 .59743 m
1 .59743 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .30902 m
1 .30902 L
s
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.07143 .21976 Mdot
.11905 .58982 Mdot
.16667 .27921 Mdot
.21429 .16171 Mdot
.2619 .18672 Mdot
.30952 .39066 Mdot
.35714 .27691 Mdot
.40476 .28584 Mdot
.45238 .35203 Mdot
.5 .291 Mdot
.54762 .46326 Mdot
.59524 .27969 Mdot
.64286 .19696 Mdot
.69048 .23983 Mdot
.7381 .36416 Mdot
.78571 .44769 Mdot
.83333 .30615 Mdot
.88095 .32124 Mdot
.92857 .2117 Mdot
.5 Mabswid
[ .02 .02 ] 0 setdash
.02381 .20601 m
.06244 .20601 L
.10458 .20601 L
.14415 .20601 L
.18221 .20601 L
.22272 .20601 L
.26171 .20601 L
.30316 .20601 L
.34309 .20601 L
.3815 .20601 L
.42237 .20601 L
.46172 .20601 L
.49955 .20601 L
.53984 .20601 L
.57861 .20601 L
.61984 .20601 L
.65954 .20601 L
.69774 .20601 L
.73838 .20601 L
.77751 .20601 L
.81909 .20601 L
.85916 .20601 L
.89771 .20601 L
.93871 .20601 L
.97619 .20601 L
s
.02381 .41202 m
.06244 .41202 L
.10458 .41202 L
.14415 .41202 L
.18221 .41202 L
.22272 .41202 L
.26171 .41202 L
.30316 .41202 L
.34309 .41202 L
.3815 .41202 L
.42237 .41202 L
.46172 .41202 L
.49955 .41202 L
.53984 .41202 L
.57861 .41202 L
.61984 .41202 L
.65954 .41202 L
.69774 .41202 L
.73838 .41202 L
.77751 .41202 L
.81909 .41202 L
.85916 .41202 L
.89771 .41202 L
.93871 .41202 L
.97619 .41202 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00k`3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo^@3oool00?40oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;P0oooo003b0?ooo`040;mdM02LW4P004RL0=oookX0
oooo003b0?ooo`030?oo_`1d000007Bo0;/0oooo003b0?ooo`030?oo_`1d0000M;oo0;/0oooo003b
0?ooo`040?noM000000004RL0=oookX0oooo003b0?ooo`050;md0000B9`0gmnL04Q8W03Oool0^@3o
ool00?00oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0]`3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005T0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`120?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
oc`0oooo00@0_g@0000000000000M;ooA03oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom10?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool0703oool005/0oooo00<0okmd0000001d_ol0A03oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_c/0oooo00H0gia8001d_`3oool0omnL04P00018W=m20?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
@`3oool00`3o_g@0000007Boo`0O0?ooo`00F`3oool00`3ooml0W4P0001d_`130?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0h0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@03oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0@@3oool00`3ooml0W4P0001d_`0O0?ooo`00G03oool0
0`3OW4P004RL0=ooo`120?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0h0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
?`3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0@P3oool00`3OW4P004RL0=ooo`0N0?ooo`00C03oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=@3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0o0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`120?ooo`030?oOW01804P0W=oo
01h0oooo001L0?ooo`030?oo_`1d0000M;oo04<0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03T0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`100?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`120?ooo`030?oo_`1d0000M;oo01h0oooo001M0?ooo`030;md0000M;l0oooo0480oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03T0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`100?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`130?ooo`030;md0000M;l0oooo01d0oooo
001M0?ooo`030=nLB000B9`0gooo0480oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo>@3oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo03l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo@`3oool00`3OW4P004RL0=ooo`0M0?ooo`00G@3oool00`3o
gi`0B01809cOo`120?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03X0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`0o0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04@0oooo00<0omnL04P0B02Lgol07@3oool005d0oooo00<0oono07@0001d
_ol0@`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0?03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0400oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04D0oooo00<0oono07@0001d
_ol07@3oool005h0oooo00<0_g@0001d_`3oool0@`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0?@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd80oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04H0oooo00<0_g@0001d_`3o
ool0703oool005`0oooo00D0oono07@000000000001d0;ooo`140?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0k0?ooo`040=nLB00000000000001d_d@0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO0440oooo00D0oono07@00000
0000001d0;ooo`0L0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00D03ooooo000004X000005`3oool00500oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0<0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04`0oooo
00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo0?oo
o`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool0
0500oooo0P00002R0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0
X03oool2000001L0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0E0?ooo`00@03oool00`3OW4P00000000000030000000304RLg`3oool0
oooo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo00110?ooo`030?noM000M;l0oooo0080oooo00<0_gAd0;ooo`3oool01`3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00480
oooo00<0okmd001d_`3oool02`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool004<0oooo00<0okmd001d_`3oool02P3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool003<0oooo00<0
oooO09a8000000001`0000000`18W=l0oooo0?ooo`040?ooo`030?noM018W=l0oooo00T0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo0015
0?ooo`030=nLB018W=l0oooo00P0oooo0`00002Q0?ooo`030000003oool0oooo09l0oooo0`00000G
0?ooo`00AP3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ogi`0B01d0;ooo`070?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
AP3oool00`3ooml0W4P007Boo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00AP3oool00`3ookl0M00004RLg`070?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@P3oool0
1`3o_g@004RL0=ooo`3oool0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0omnL04P0000000000000
07Boo`080?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E
0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0
oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?oo
o`120?ooo`800000G@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@
0?ooo`030000003oool0oooo0440oooo1000001L0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0@03oool6000005/0oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool200000440oooo1P00001K0?ooo`03
0000003oool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`110?ooo`@00000
G03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003o
ool0oooo0480oooo0P00001M0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3o
ool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`1B0?ooo`800000C@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0540oooo
1000001<0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00
001A0?ooo`H00000B`3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00400oooo00<0oono
07@0000000000P0000000`18W=l0oooo0?ooo`080?ooo`030000003oool0oooo0500oooo1P00001;
0?ooo`030000003oool0oooo02/0oooo0P00001d0?ooo`030000003oool0oooo01D0oooo00120?oo
o`030?noM0000000M;oo00/0oooo00<000000?ooo`3oool0D@3oool4000004`0oooo00<000000?oo
o`3oool0:P3oool4000007<0oooo00<000000?ooo`3oool05@3oool00480oooo00<0oooO09a80000
M;l02`3oool00`000000oooo0?ooo`1B0?ooo`800000C@3oool00`000000oooo0?ooo`0Y0?ooo`H0
0000LP3oool00`000000oooo0?ooo`0E0?ooo`00@`3oool00`3OW4P004RL0=ooo`0:0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0:@3oool600000780oooo00<000000?ooo`3oool0
5@3oool003<0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW018
04P0W=oo00X0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0Z0?ooo`@00000L`3o
ool00`000000oooo0?ooo`0E0?ooo`00@`3oool00`3ookl0M00007Boo`0:0?ooo`<00000103oool8
000000H0oooo1`0000060?ooo`L000001P3oool7000000H0oooo1`0000060?ooo`P000001@3oool8
000000D0oooo200000050?ooo`P000001P3oool7000000H0oooo1`0000060?ooo`L000001@3oool9
000000D0oooo200000050?ooo`P000001@3oool8000000H0oooo1`0000060?ooo`L000001P3oool7
000000H0oooo200000050?ooo`P000001@3oool8000000D0oooo2P0000030?ooo`P000002@3oool3
000001L0oooo00140?ooo`030;md0000M;l0oooo00T0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`280?ooo`@000005@3oool00`000000oooo0?ooo`0E0?ooo`00A03oool00`3OW4P0
04RL0=ooo`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0Q`3oool6000001@0
oooo00<000000?ooo`3oool05@3oool002D0oooo00@0oono07@0000007@0_ooo6`3oool00`3ogi`0
B01809cOo`090?ooo`030000003oool0oooo01<0oooo0P00002<0?ooo`030000003oool0oooo08L0
oooo1P00000D0?ooo`030000003oool0oooo01D0oooo000O0?ooo`040?oo_`2LW=l0oono07@000@0
000000<0M;oo0?ooo`3oool06P3oool00`3ookl0M00007Boo`090?ooo`030000003oool0oooo0180
oooo1000002;0?ooo`030000003oool0oooo08P0oooo1000000E0?ooo`030000003oool0oooo01D0
oooo000O0?ooo`060?oOW018000000000000000004P0W=oo803oool00`2oM00007Bo0?ooo`080?oo
o`030000003oool0oooo0140oooo1P00002:0?ooo`030000003oool0oooo08T0oooo0P00000F0?oo
o`030000003oool0oooo01D0oooo000P0?ooo`030?oo_`1d0000M;oo0200oooo00D0oono07@00000
0000001d0;ooo`080?ooo`030000003oool0oooo0140oooo1P00002:0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00200oooo00<0_gAd0;ooo`3oool0;@3oool2000001<0
oooo1000002;0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`007`3oool0103o_g@00000
0018W03Oool]0?ooo`030000003oool0oooo01<0oooo0P00002<0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0k0?ooo`800000I03oool00`000000oooo0?ooo`0E0?ooo`008P3oool01P3OW4P000000000
00000000001809cOobP0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0j0?ooo`@0
0000H`3oool00`000000oooo0?ooo`0E0?ooo`00803oool02@3ookl0M00004RLg`3ogi`0W;oo0?oo
o`3OW4P004RL0=ooo`0W0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0>@3oool6
00000680oooo00<000000?ooo`3oool05@3oool00200oooo00T0_gAd0;ooo`3oool0oknL09cOo`3o
ool0oono07@0M02oool09`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo03T0oooo
1P00001R0?ooo`030000003oool0oooo01D0oooo000O0?ooo`090?oOW02L_ol0oooo0?oog`2LM9`0
gooo0?ooo`3ogi`0B7Bo02P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0j0?oo
o`@00000H`3oool00`000000oooo0?ooo`0E0?ooo`007`3oool02@3o_i`0W=oo0?ooo`2oM000M;oo
0?ooo`3OW4P0001d0;ooo`0X0?ooo`800000XP3oool00`000000oooo0?ooo`0k0?ooo`800000H`3o
ool2000001L0oooo000O0?ooo`040?oo_`1d0000000007Boobd0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00903oool01@3o
oml0W4P000000000B9`0gooo02L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo01D0oooo000P0?ooo`040;md00000000001d0;ooo`80oooo00<0oono
07Ad_`3oool09`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool001l0oooo00H0omnL09boo`3ooml0W4P00000M02oool20?ooo`030=nLW02o
ool0oooo02H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo000O0?ooo`0:0?noW02Lgol0oooo0?oog`2LB000001809cOo`3ookl0M4RL0=oo
obL0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
01D0oooo000O0?ooo`030?oo_`1dM;l0oooo0080oooo00@0_g@00000000007@0_ooo:03oool20000
0780oooo0P00000^0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`007`3oool0103ogi`0
B0000000001d_ol]0?ooo`030000003oool0oooo03<0oooo0P00000k0?ooo`@00000;@3oool00`00
0000oooo0?ooo`0K0?ooo`800000Q03oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000
oooo0?ooo`0b0?ooo`@00000>@3oool6000002`0oooo00<000000?ooo`3oool06P3oool4000008<0
oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0<@3oool6000003P0oooo
1P00000<0?ooo`8000007P3oool00`000000oooo0?ooo`0I0?ooo`H00000PP3oool00`000000oooo
0?ooo`0E0?ooo`007`3oool00`3ookl0W9cO0?ooo`030?ooo`040?noM0000000001d0;ooobL0oooo
00<000000?ooo`3oool0<@3oool6000003T0oooo1000000<0?ooo`@000007@3oool00`000000oooo
0?ooo`0I0?ooo`H00000PP3oool00`000000oooo0?ooo`0E0?ooo`007`3oool0103ookl0W9cO0?oo
_`1d0004000000030018W03Oool0oooo02H0oooo00<000000?ooo`3oool0<P3oool4000003/0oooo
0P00000<0?ooo`H000006`3oool2000001`0oooo100000230?ooo`030000003oool0oooo01D0oooo
000O0?ooo`090?oOW018000000000000B02Lgol0oooo0?oog`2LB7@0_ooo02P0oooo0P00000d0?oo
o`800000BP3oool6000001X0oooo1000000L0?ooo`800000P`3oool2000001L0oooo000K0?ooo`04
0?oog`2LB000001d0;oooc40oooo00<000000?ooo`3oool0P03oool4000001X0oooo1P00002Q0?oo
o`030000003oool0oooo01D0oooo000S0?ooo`050=nLB000000000000000001d_ol06P3oool0102o
M00000000000001d_ol:0?ooo`030000003oool0oooo0840oooo0P00000K0?ooo`H00000X@3oool0
0`000000oooo0?ooo`0E0?ooo`008@3oool00`3ogi`0B00004RLg`020?ooo`030?oOW01807@0_ooo
01P0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003oool0oooo09l0oooo1000
002R0?ooo`030000003oool0oooo01D0oooo000P0?ooo`030?oOW018B9`0gooo00<0oooo00<0oono
07A8W03Oool05`3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<00000
0?ooo`3oool0X03oool2000006/0oooo0P00000f0?ooo`030000003oool0oooo01D0oooo000O0?oo
o`030?oog`2LM9`0gooo00@0oooo00<0_g@007Boo`3oool05`3oool00`3ooml0W4P004RLg`020?oo
o`040?oog`2LB00004RL0=ooo`L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1X
0?ooo`@00000=@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ogkl0_moo0?ooo`020?oo
o`030?oog`2LB000M;oo01T0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`070?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0I`3oool6000003@0oooo00<000000?oo
o`3oool05@3oool001l0oooo00X0okmd07Boo`3oool0oooo0?oOW01804P0W=no07@00000B9`0gooo
603oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0ooooo`00001:000001L0oooo
000O0?ooo`030?oog`2LB000000000D0000000<004RL0=ooo`3oool05`3oool00`3OW4P0001809cO
o`020?ooo`030?oog`2LB00007Bo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`1X0?ooo`@00000=@3oool00`000000oooo0?ooo`0E0?ooo`006`3oool0102oW9`0gooo0?oo_`1d
00040000000507Boo`3oool0oooo0;mdM02oool06@3oool00`3ogi`0B00004RLg`030?ooo`030;md
0000B9`0gooo00H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1Y0?ooo`800000
3P3oool2000002H0oooo00<000000?ooo`3oool05@3oool001X0oooo00<0oooO09a8000000000P00
00000`1d_ol0oooo0?ooo`0O0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0
1P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo07P0oooo1000000U0?ooo`030000
003oool0oooo01D0oooo000J0?ooo`030?oo_`2LW=l0oooo02D0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`060?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0M`3o
ool6000002@0oooo00<000000?ooo`3oool05@3oool00200oooo00D0_g@00018W03Oool0oono07@0
0002000000030000M02oool0oooo01T0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0M`3oool6000002@0oooo00<000000?ooo`3o
ool05@3oool001l0oooo00<0omnL04P0000000000`0000001018W=l0oooo0;nLW03OoolK0?ooo`04
0=nLB00000000000001d_`P0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1h0?oo
o`@000009@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool0103ooml0W4P00000M02oool30?oo
o`030?noM018W=l0oooo02L0oooo0P00002R0?ooo`030000003oool0oooo07T0oooo0P00000U0?oo
o`8000005`3oool002D0oooo00<0okmd001d_`3oool0:03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool002<0oooo00<0oooO0;no_`2LW;l0
:P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
5@3oool002<0oooo00H0omno09adB000000000000000M02ooolW0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`007`3oool00`3ogi`0B000
00000002000000050000B02Lgol0oooo0;nLW03Oool09`3oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool001l0oooo00<0okmd0000M02oool0
103oool00`2oM000M;oo0?ooo`0W0?ooo`030000003oool0oooo0900oooo0P00000?0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool002H0oooo00<0gmoO0?ooo`3oool09`3o
ool00`000000oooo0?ooo`2?0?ooo`@000003P3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo000S0?ooo`050;md00000000000000000018W=l0:03oool2000008l0oooo1P00
000=0?ooo`030000003oool0oooo0:00oooo0P00000G0?ooo`008@3oool00`3OW4P0000007Boo`02
0?ooo`030?oOW01807@0_ooo02L0oooo00<000000?ooo`3oool0SP3oool6000000d0oooo00<00000
0?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00803oool00`3OW4P0B9cO0?ooo`030?oo
o`030?oOW018M;l0oooo02L0oooo00<000000?ooo`3oool0S`3oool4000000h0oooo00<000000?oo
o`3oool0BP3oool2000005D0oooo00<000000?ooo`3oool05@3oool001l0oooo00<0omnL07BLg`3o
ool00`3oool00`3o_i`0B7BL0=ooo`0X0?ooo`030000003oool0oooo0900oooo0P00000?0?ooo`03
0000003oool0oooo04T0oooo1000001D0?ooo`030000003oool0oooo01D0oooo000O0?ooo`030?no
M000M;l0oooo0080oooo00D0omnL04QdM01d000004RL0=ooo`0W0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool0B03oool6000005<0oooo00<000000?ooo`3oool05@3oool00200oooo
00<0okmd000000000000100000000`00B9`0gooo0?ooo`0V0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool0B03oool6000005<0oooo00<000000?ooo`3oool05@3oool001l0oooo00D0
omnL04P000000000000004RLg`020?ooo`030;mdB02Lgol0oooo02L0oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`190?ooo`@00000E03oool00`000000oooo0?ooo`0E0?ooo`00D03o
ool200000:80oooo00<000000?ooo`3oool0BP3oool2000005@0oooo0P00000G0?ooo`00D03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool0
02<0oooo00H0okmd00000000000000000000M02ooolW0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`006`3oool01P2oW9`0gooo0?ooo`3o
ool0omnL04P000<0000000@004RL0=ooo`3ookl0M7Bo:03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool001X0oooo00<0oooO09a800000000
0`0000000`00M;l0oooo0?ooo`020?ooo`030?noM00007@0_ooo02P0oooo00<000000?ooo`3oool0
H@3oool2000003h0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`006P3o
ool00`3ookl0M00004RLg`0c0?ooo`030000003oool0oooo0600oooo1000000m0?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool05@3oool002@0oooo00D0oooO09a80000000004RL0=oo
o`0W0?ooo`800000H03oool6000003`0oooo00<000000?ooo`3oool0X03oool2000001L0oooo000P
0?ooo`040;md00000000001d0;ooo`80oooo00<0oono07Ad_`3oool09`3oool00`000000oooo0?oo
o`1O0?ooo`H00000?03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000O
0?ooo`060?oOW02L_ol0oooO09a8000007@0_ooo0P3oool00`3OW9`0_ooo0?ooo`0F0?ooo`030?oo
_`1d000000000080000000<0B9cO0?ooo`3oool0203oool00`000000oooo0?ooo`1P0?ooo`@00000
?@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo000O0?ooo`0:0?noW02L
gol0oooo0?oog`2LB000001809cOo`3ookl0M4RL0=oooaT0oooo00<0okmd0000001d_ol02`3oool0
0`000000oooo0?ooo`1Q0?ooo`800000?P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo000O0?ooo`030?oo_`1dM;l0oooo0080oooo00@0_g@00000000007@0_ooo6P3oool0
0`3ooml0W4P0001d_`0;0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`007`3oool0103ogi`0B0000000001d_olP0?ooo`030=nLB000B9`0
gooo00X0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01D0oooo00130?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00130?ooo`030?oo_`1d0000M;oo00X0
oooo0`0000040?ooo`P000001P3oool7000000H0oooo1`0000060?ooo`L000001P3oool7000000H0
oooo200000050?ooo`P000001@3oool8000000D0oooo200000060?ooo`L000001P3oool7000000H0
oooo1`0000050?ooo`T000001@3oool8000000D0oooo200000050?ooo`P000001P3oool7000000H0
oooo1`0000060?ooo`L000001P3oool8000000D0oooo200000050?ooo`P000001@3oool8000000D0
oooo200000090?ooo`<000005`3oool004@0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004@0oooo00<0
gia80018W03Oool02@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool004@0oooo00<0omnL04P0B02Lgol02@3oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004@0oooo00<0oono07@0
001d_ol02@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool004D0oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004<0oooo00D0oono07@000000000
001d0;ooo`080?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0X03oool2000001L0oooo
001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1J0?oo
o`800000A@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`03
0000003oool0oooo05T0oooo100000140?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0F03oool6000004<0oooo00<000000?ooo`3oool0
5@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`1H0?ooo`H00000@`3o
ool00`000000oooo0?ooo`0E0?ooo`00D03oool200000:80oooo00<000000?ooo`3oool0F@3oool4
000004<0oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
00`0oooo0P00001<0?ooo`800000A@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo00/0oooo1000002B0?ooo`030000003oool0oooo01D0
oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool02P3oool600000940oooo
00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`0:0?ooo`H00000T@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo00/0oooo1000002B0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`800000XP3oool00`000000oooo0?ooo`0<0?ooo`800000TP3oool2000001L0oooo001@0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
5@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000XP3oool00`000000oooo
0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo01D0oooo00100?ooo`030=nLB0000000000000<0000000<0B9cO
0?ooo`3oool01`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool05@3oool00440oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`070?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?oo
o`00@P3oool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@`3oool00`3o_g@007Bo0?ooo`0:0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00A03o
ool00`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0E0?ooo`00A@3oool00`3OW4P0B9cO0?ooo`080?ooo`<00000X@3oool0
0`000000oooo0?ooo`2O0?ooo`<000005`3oool004H0oooo00<0_g@007Boo`3oool01`3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004H0
oooo00<0omnL04P0M02oool01`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool004H0oooo00<0oooO09a8001d_ol01`3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004H0oooo00<0
oono07@00018W=l01`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool05@3oool00480oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00L0oooo
00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo
00130?ooo`050?oOW018000000000000001d_ol0203oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003o
ool0oooo0:00oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool00500oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@
0?ooo`800000XP3oool00`000000oooo0?ooo`2P0?ooo`8000005`3oool00500oooo00<000000?oo
o`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00
D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
5@3oool00500oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2Q0?ooo`030000003o
ool0oooo01D0oooo001@0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
0:40oooo00<000000?ooo`3oool05@3oool00500oooo0P00002R0?ooo`030000003oool0oooo0:00
oooo0P00000G0?ooo`00D03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo
00<000000?ooo`3oool05@3oool00500oooo00<000000?ooo`3oool08`3oool2000007`0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`0R
0?ooo`@00000N`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?oo
o`030000003oool0oooo0240oooo1P00001j0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool05@3oool00500oooo00<000000?ooo`3oool08@3oool6000007X0oooo00<000000?ooo`3o
ool0X@3oool00`000000oooo0?ooo`0E0?ooo`00D03oool00`000000oooo0?ooo`0R0?ooo`@00000
N`3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo001@0?ooo`800000903o
ool2000007`0oooo00<000000?ooo`3oool0X03oool2000001L0oooo001@0?ooo`030000003oool0
oooo0:40oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0E0?ooo`00@03oool00`3o
okl0M000000000020000000307Boo`3oool0oooo00P0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`2Q0?ooo`030000003oool0oooo01D0oooo00100?ooo`030?noM000000007Bo0080
oooo00<0_gA809cOo`3oool0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool05@3oool004D0oooo00<0omnL04Q8W03Oool0203oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool05@3oool004D0oooo00<0
oono07@0001d_ol0203oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo04`0oooo00<0
00000?ooo`3oool0B`3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo04/0oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0E0?ooo`00A@3oool00`3ogi`0B00007Boo`08
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool004@0oooo00@0omnL04P0
000004P0W=oo203ooooo000004X000005`3oool00480oooo00D0omnL04P000000000000007Boo`3o
0?ooofX0oooo00140?ooo`030?oo_`1d000007Bo0?l0ooooJP3oool004H0oooo00<0gia8001d_`3o
ool0o`3ooomX0?ooo`00AP3oool00`3ogi`0B01809cOo`3o0?ooofP0oooo00130?ooo`060;md001d
_ol0oooo0=nLB00004P0W=ooo`3ooomX0?ooo`00A03oool0102oM000000000000018W=oo0?ooofT0
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.56215, -4.51278, \
0.00959293, 0.0443476}}]
}, Closed]],

Cell["\<\
Notice that the estimated variance is equivalent to reduced chisquare.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(w . \((FitResiduals /. results)\)\^2\/\(Length[y] - 3\)\)], "Input"],

Cell[BoxData[
    \(1.3201091053392249`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "\tNext we consider the fit to ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep\)]],
  ".  In order to provide initial values for the fitting parameters, we use a \
poorly documented feature of the syntax.  Because the fitting code employs ",
  StyleBox["FindMinimum", "InlineInput",
    FontWeight->"Bold"],
  ", the fitting parameters can be listed in the form ",
  Cell[BoxData[
      \(TraditionalForm\`{{a, a\_0}, {b, b\_0}, \[CenterEllipsis]}\)]],
  ".  To use ",
  StyleBox["NonlinearRegress", "InlineInput",
    FontWeight->"Bold"],
  " we also have to separate the uncertainties and compute the weights."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\({x, y, dy} = Transpose[gepdata];\)\), "\[IndentingNewLine]", 
    \(gepresults2 = 
      NonlinearRegress[
        Transpose[{x, 
            y}], \(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + b3\ qq\^3\), 
        qq, {{a1, 0}, {b1, 1.4}, {b2, 0}, {b3, 0}}, 
        Weights \[Rule] dy\^\(-2\)]\)}], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\(BestFitParameters \[Rule] {a1 \[Rule] \(-0.06703155898295088`\
\), b1 \[Rule] 3.117673243156375`, b2 \[Rule] 1.03407507158637`, 
            b3 \[Rule] 0.5047708094973022`}\), ",", 
        RowBox[{"ParameterCITable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Estimate\"\>", "\<\"Asymptotic SE\"\>", "\<\
\"CI\"\>"},
                {"a1", \(-0.06703155898295088`\), 
                  "0.03268019008877416`", \({\(-0.13273948506120703`\), \
\(-0.001323632904694727`\)}\)},
                {"b1", "3.117673243156375`", 
                  "0.04934499111809321`", \({3.018458488899678`, 
                    3.216887997413072`}\)},
                {"b2", "1.03407507158637`", 
                  "0.08342444840992602`", \({0.8663389759777435`, 
                    1.2018111671949967`}\)},
                {"b3", "0.5047708094973022`", 
                  "0.1449316197525924`", \({0.21336625734396175`, 
                    0.7961753616506427`}\)}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ 
              BoxForm`e$, TableDepth -> 2, TableHeadings -> {{a1, b1, b2, 
                b3}, {"Estimate", "Asymptotic SE", "CI"}}]]]}], 
        ",", \(EstimatedVariance \[Rule] 0.8454699647048753`\), ",", 
        RowBox[{"ANOVATable", "\[Rule]", 
          InterpretationBox[GridBox[{
                {"\<\"\"\>", "\<\"DF\"\>", "\<\"SumOfSq\"\>", \
"\<\"MeanSq\"\>"},
                {"\<\"Model\"\>", "4", "2.7544198845825577`*^6", 
                  "688604.9711456394`"},
                {"\<\"Error\"\>", "48", "40.58255830583402`", 
                  "0.8454699647048753`"},
                {"\<\"Uncorrected Total\"\>", "52", 
                  "2.7544604671408636`*^6", "\<\"\"\>"},
                {"\<\"Corrected Total\"\>", "51", 
                  "1.6582070697390493`*^6", "\<\"\"\>"}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            
            TableForm[ {{4, 2754419.8845825577, 688604.97114563943}, {48, 
              40.58255830583402, .84546996470487534}, {52, 
              2754460.4671408636}, {51, 1658207.0697390493}}, 
              TableHeadings -> {{"Model", "Error", "Uncorrected Total", 
              "Corrected Total"}, {"DF", "SumOfSq", "MeanSq"}}]]}], ",", 
        RowBox[{"AsymptoticCorrelationMatrix", "\[Rule]", 
          TagBox[
            RowBox[{"(", "\[NoBreak]", GridBox[{
                  {"0.9999999999999998`", "0.8543384231837428`", 
                    "0.2119984961424832`", "0.9550277107710882`"},
                  {"0.854338423183743`", 
                    "1.0000000000000002`", \(-0.17058544740320455`\), 
                    "0.890023560823461`"},
                  {"0.2119984961424832`", \(-0.17058544740320458`\), 
                    "1.`", \(-0.05211400788279216`\)},
                  {"0.9550277107710884`", 
                    "0.890023560823461`", \(-0.05211400788279218`\), "1.`"}
                  },
                RowSpacings->1,
                ColumnSpacings->1,
                ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
            Function[ BoxForm`e$, 
              MatrixForm[ BoxForm`e$]]]}], ",", 
        RowBox[{"FitCurvatureTable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Curvature\"\>"},
                {"\<\"Max Intrinsic\"\>", "91.71315937289158`"},
                {"\<\"Max Parameter-Effects\"\>", "240.27747890331784`"},
                {"\<\"95. % Confidence Region\"\>", "0.6243612699707356`"}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ 
              BoxForm`e$, TableDepth -> 2, TableHeadings -> {{"Max Intrinsic",
                 "Max Parameter-Effects", "95. % Confidence Region"}, {
                "Curvature"}}]]]}]}], "}"}]], "Output"]
}, Closed]],

Cell[BoxData[{
    \(\({x, y, dy} = Transpose[gepdata];\)\), "\[IndentingNewLine]", 
    \(\(y = y\ \((1 + x\/0.71)\)\^2;\)\), "\[IndentingNewLine]", 
    \(\(dy = dy\ \((1 + x\/0.71)\)\^2;\)\), "\[IndentingNewLine]", 
    \(\(ratiodata = Transpose[{x, y, dy}];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        Plot[Evaluate[\(\(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + 
                      b3\ qq\^3\)\) \((1 + qq\/0.71)\)\^2 /. 
              Thread[{a1, b1, b2, 
                    b3} \[Rule] \((Parameters /. gepresults)\)]], {qq, 0, 
            10}, PlotRange \[Rule] {{0, 10}, {0, 1.2}}, Frame \[Rule] True, 
          FrameLabel \[Rule] {\*"\"\<\!\(Q\^2\)\>\"", \
\*"\"\<\!\(G\_Ep\)/\!\(G\_D\)\>\""}], \[IndentingNewLine]Plot[
          Evaluate[\(\(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + 
                      b3\ qq\^3\)\) \((1 + qq\/0.71)\)\^2 /. \
\((BestFitParameters /. gepresults2)\)], {qq, 0, 10}, 
          PlotStyle \[Rule] Hue[0.3]], \[IndentingNewLine]ErrorListPlot[
          ratiodata]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
-2.77556e-017 0.1 -2.18738e-017 0.515028 [
[.2 -0.0125 -5 -15.6875 ]
[.2 -0.0125 5 0 ]
[.4 -0.0125 -5 -15.6875 ]
[.4 -0.0125 5 0 ]
[.6 -0.0125 -5.125 -15.6875 ]
[.6 -0.0125 5.125 0 ]
[.8 -0.0125 -5 -15.6875 ]
[.8 -0.0125 5 0 ]
[1 -0.0125 -8 -15.6875 ]
[1 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -8.71875 -31.375 ]
[.5 -0.0125 8.71875 -15.6875 ]
[-0.0125 .10301 -19 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -19 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -19.25 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -19 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -35.6875 -20.9375 ]
[-0.01875 .30902 -19.25 20.9375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 0 m
.2 .00625 L
s
gsave
.2 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.4 0 m
.4 .00625 L
s
gsave
.4 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.6 0 m
.6 .00625 L
s
gsave
.6 -0.0125 -66.125 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.8 0 m
.8 .00625 L
s
gsave
.8 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.05 0 m
.05 .00375 L
s
.1 0 m
.1 .00375 L
s
.15 0 m
.15 .00375 L
s
.25 0 m
.25 .00375 L
s
.3 0 m
.3 .00375 L
s
.35 0 m
.35 .00375 L
s
.45 0 m
.45 .00375 L
s
.5 0 m
.5 .00375 L
s
.55 0 m
.55 .00375 L
s
.65 0 m
.65 .00375 L
s
.7 0 m
.7 .00375 L
s
.75 0 m
.75 .00375 L
s
.85 0 m
.85 .00375 L
s
.9 0 m
.9 .00375 L
s
.95 0 m
.95 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -69.7188 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Q) show
71.625 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
76.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .02575 m
.00375 .02575 L
s
0 .0515 m
.00375 .0515 L
s
0 .07725 m
.00375 .07725 L
s
0 .12876 m
.00375 .12876 L
s
0 .15451 m
.00375 .15451 L
s
0 .18026 m
.00375 .18026 L
s
0 .23176 m
.00375 .23176 L
s
0 .25751 m
.00375 .25751 L
s
0 .28327 m
.00375 .28327 L
s
0 .33477 m
.00375 .33477 L
s
0 .36052 m
.00375 .36052 L
s
0 .38627 m
.00375 .38627 L
s
0 .43777 m
.00375 .43777 L
s
0 .46353 m
.00375 .46353 L
s
0 .48928 m
.00375 .48928 L
s
0 .54078 m
.00375 .54078 L
s
0 .56653 m
.00375 .56653 L
s
0 .59228 m
.00375 .59228 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -109.406 -12.2188 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.9375 12.2188 translate
90 rotate
-81.9375 -12.2188 translate
0 24.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
71.688 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Ep) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
81.625 15.813 moveto
(\\220) show
85.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
94.250 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(D) show
100.875 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .61178 m
0 .61803 L
s
.2 .61178 m
.2 .61803 L
s
.4 .61178 m
.4 .61803 L
s
.6 .61178 m
.6 .61803 L
s
.8 .61178 m
.8 .61803 L
s
.125 Mabswid
.05 .61428 m
.05 .61803 L
s
.1 .61428 m
.1 .61803 L
s
.15 .61428 m
.15 .61803 L
s
.25 .61428 m
.25 .61803 L
s
.3 .61428 m
.3 .61803 L
s
.35 .61428 m
.35 .61803 L
s
.45 .61428 m
.45 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.55 .61428 m
.55 .61803 L
s
.65 .61428 m
.65 .61803 L
s
.7 .61428 m
.7 .61803 L
s
.75 .61428 m
.75 .61803 L
s
.85 .61428 m
.85 .61803 L
s
.9 .61428 m
.9 .61803 L
s
.95 .61428 m
.95 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .02575 m
1 .02575 L
s
.99625 .0515 m
1 .0515 L
s
.99625 .07725 m
1 .07725 L
s
.99625 .12876 m
1 .12876 L
s
.99625 .15451 m
1 .15451 L
s
.99625 .18026 m
1 .18026 L
s
.99625 .23176 m
1 .23176 L
s
.99625 .25751 m
1 .25751 L
s
.99625 .28327 m
1 .28327 L
s
.99625 .33477 m
1 .33477 L
s
.99625 .36052 m
1 .36052 L
s
.99625 .38627 m
1 .38627 L
s
.99625 .43777 m
1 .43777 L
s
.99625 .46353 m
1 .46353 L
s
.99625 .48928 m
1 .48928 L
s
.99625 .54078 m
1 .54078 L
s
.99625 .56653 m
1 .56653 L
s
.99625 .59228 m
1 .59228 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
0 .51503 m
.00236 .51107 L
.00488 .5078 L
.00726 .50544 L
.00942 .50379 L
.01197 .50234 L
.01341 .50171 L
.01473 .50124 L
.01717 .50059 L
.0184 .50036 L
.01975 .50017 L
.02113 .50003 L
.02245 .49995 L
.02364 .49991 L
.02493 .4999 L
.02566 .49991 L
.02632 .49993 L
.02709 .49995 L
.02782 .49998 L
.0291 .50006 L
.0305 .50015 L
.04057 .50113 L
.04336 .50139 L
.04636 .50165 L
.04914 .50184 L
.05051 .50191 L
.05177 .50197 L
.05319 .50202 L
.05393 .50204 L
.05472 .50206 L
.056 .50207 L
.05673 .50207 L
.05741 .50207 L
.05817 .50206 L
.05897 .50205 L
.05965 .50203 L
.0604 .50201 L
.06206 .50194 L
.06361 .50185 L
.06637 .50163 L
.06898 .50135 L
.07386 .50064 L
.07911 .4996 L
.08481 .49816 L
.09475 .49486 L
.10568 .49018 L
.12777 .47783 L
.17013 .4467 L
.2109 .41231 L
.25424 .37508 L
Mistroke
.29598 .34069 L
.33614 .30988 L
.37887 .27987 L
.42002 .25368 L
.46373 .22856 L
.50586 .20679 L
.54639 .18784 L
.5895 .16961 L
.63101 .1537 L
.67094 .13976 L
.71344 .12621 L
.75435 .11428 L
.79783 .10267 L
.83972 .09239 L
.88003 .08327 L
.9229 .07429 L
.96419 .06628 L
1 .05979 L
Mfstroke
.2 1 0 r
0 .51503 m
.00236 .51107 L
.00488 .5078 L
.00726 .50544 L
.00942 .50379 L
.01197 .50234 L
.01341 .50171 L
.01473 .50124 L
.01717 .5006 L
.0184 .50037 L
.01975 .50017 L
.02113 .50004 L
.02245 .49996 L
.02364 .49992 L
.02493 .49991 L
.02566 .49992 L
.02632 .49993 L
.02709 .49996 L
.02782 .49999 L
.0291 .50007 L
.0305 .50016 L
.04057 .50114 L
.04336 .50141 L
.04636 .50166 L
.04914 .50185 L
.05051 .50193 L
.05177 .50199 L
.05319 .50204 L
.05393 .50206 L
.05472 .50207 L
.056 .50209 L
.05673 .50209 L
.05741 .50209 L
.05817 .50208 L
.05897 .50207 L
.05965 .50205 L
.0604 .50203 L
.06206 .50196 L
.06361 .50187 L
.06637 .50165 L
.06898 .50137 L
.07386 .50066 L
.07911 .49962 L
.08481 .49817 L
.09475 .49487 L
.10568 .49018 L
.12777 .47781 L
.17013 .44667 L
.2109 .41227 L
.25424 .37505 L
Mistroke
.29598 .34068 L
.33614 .3099 L
.37887 .27992 L
.42002 .25376 L
.46373 .2287 L
.50586 .20696 L
.54639 .18805 L
.5895 .16985 L
.63101 .15399 L
.67094 .14008 L
.71344 .12657 L
.75435 .11467 L
.79783 .10308 L
.83972 .09284 L
.88003 .08374 L
.9229 .07479 L
.96419 .06681 L
1 .06034 L
Mfstroke
0 g
.002 w
.00054 .5144 m
.00054 .51667 L
s
.015 w
.00054 .51554 Mdot
.002 w
.00078 .51284 m
.00078 .51541 L
s
.015 w
.00078 .51413 Mdot
.002 w
.00097 .5134 m
.00097 .51566 L
s
.015 w
.00097 .51453 Mdot
.002 w
.00117 .50966 m
.00117 .51407 L
s
.015 w
.00117 .51187 Mdot
.002 w
.00136 .51306 m
.00136 .51532 L
s
.015 w
.00136 .51419 Mdot
.002 w
.00156 .50904 m
.00156 .51417 L
s
.015 w
.00156 .51161 Mdot
.002 w
.00175 .51088 m
.00175 .51251 L
s
.015 w
.00175 .51169 Mdot
.002 w
.00195 .51102 m
.00195 .51419 L
s
.015 w
.00195 .51261 Mdot
.002 w
.00214 .51036 m
.00214 .512 L
s
.015 w
.00214 .51118 Mdot
.002 w
.00234 .51063 m
.00234 .51371 L
s
.015 w
.00234 .51217 Mdot
.002 w
.00253 .50982 m
.00253 .51165 L
s
.015 w
.00253 .51074 Mdot
.002 w
.00273 .50724 m
.00273 .5108 L
s
.015 w
.00273 .50902 Mdot
.002 w
.00312 .50738 m
.00312 .51013 L
s
.015 w
.00312 .50875 Mdot
.002 w
.00331 .50855 m
.00331 .51181 L
s
.015 w
.00331 .51018 Mdot
.002 w
.0035 .50976 m
.0035 .51355 L
s
.015 w
.0035 .51166 Mdot
.002 w
.00389 .50243 m
.00389 .51053 L
s
.015 w
.00389 .50648 Mdot
.002 w
.00467 .50224 m
.00467 .5083 L
s
.015 w
.00467 .50527 Mdot
.002 w
.00545 .50351 m
.00545 .50756 L
s
.015 w
.00545 .50553 Mdot
.002 w
.013 .49459 m
.013 .50602 L
s
.015 w
.013 .5003 Mdot
.002 w
.019 .4974 m
.019 .5056 L
s
.015 w
.019 .5015 Mdot
.002 w
.027 .48672 m
.027 .50038 L
s
.015 w
.027 .49355 Mdot
.002 w
.033 .48851 m
.033 .4994 L
s
.015 w
.033 .49396 Mdot
.002 w
.039 .50568 m
.039 .51792 L
s
.015 w
.039 .5118 Mdot
.002 w
.045 .50252 m
.045 .52856 L
s
.015 w
.045 .51554 Mdot
.002 w
.053 .49594 m
.053 .53145 L
s
.015 w
.053 .5137 Mdot
.002 w
.058 .49263 m
.058 .50708 L
s
.015 w
.058 .49985 Mdot
.002 w
.065 .47904 m
.065 .5225 L
s
.015 w
.065 .50077 Mdot
.002 w
.072 .52816 m
.072 .60003 L
s
.015 w
.072 .56409 Mdot
.002 w
.078 .47758 m
.078 .50684 L
s
.015 w
.078 .49221 Mdot
.002 w
.094 .52387 m
.094 .56648 L
s
.015 w
.094 .54518 Mdot
.002 w
.11 .45531 m
.11 .48858 L
s
.015 w
.11 .47194 Mdot
.002 w
.135 .47145 m
.135 .51706 L
s
.015 w
.135 .49426 Mdot
.002 w
.175 .40139 m
.175 .48028 L
s
.015 w
.175 .44083 Mdot
.002 w
.0373 .49251 m
.0373 .52127 L
s
.015 w
.0373 .50689 Mdot
.002 w
.0401 .49939 m
.0401 .52965 L
s
.015 w
.0401 .51452 Mdot
.002 w
.0441 .46155 m
.0441 .49404 L
s
.015 w
.0441 .47779 Mdot
.002 w
.038 .45641 m
.038 .50739 L
s
.015 w
.038 .4819 Mdot
.002 w
.05 .49811 m
.05 .54897 L
s
.015 w
.05 .52354 Mdot
.002 w
.049 .49418 m
.049 .51066 L
s
.015 w
.049 .50242 Mdot
.002 w
.079 .49585 m
.079 .50872 L
s
.015 w
.079 .50228 Mdot
.002 w
.118 .47152 m
.118 .48539 L
s
.015 w
.118 .47845 Mdot
.002 w
.148 .42209 m
.148 .45345 L
s
.015 w
.148 .43777 Mdot
.002 w
.177 .42258 m
.177 .44897 L
s
.015 w
.177 .43578 Mdot
.002 w
.188 .41683 m
.188 .44287 L
s
.015 w
.188 .42985 Mdot
.002 w
.213 .39604 m
.213 .43229 L
s
.015 w
.213 .41417 Mdot
.002 w
.247 .37741 m
.247 .40221 L
s
.015 w
.247 .38981 Mdot
.002 w
.297 .32293 m
.297 .35614 L
s
.015 w
.297 .33954 Mdot
.002 w
.347 .309 m
.347 .35542 L
s
.015 w
.347 .33221 Mdot
.002 w
.35 .27179 m
.35 .354 L
s
.015 w
.35 .31289 Mdot
.002 w
.397 .23357 m
.397 .29113 L
s
.015 w
.397 .26235 Mdot
.002 w
.475 .17678 m
.475 .23446 L
s
.015 w
.475 .20562 Mdot
.002 w
.554 .0975 m
.554 .194 L
s
.015 w
.554 .14575 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00i@3oool01@3ogkl0_moo0?ooo`3oool0gia800@0000000<0001d0;oo
o`3oool0_`3oool00>D0oooo00<0oooO09a8000000000P0000001`0007@0_ooo0?ooo`3oool0oono
07@0M02oool0_`3oool00>L0oooo00<0_g@004RLg`3oool01@3oool00`3ookl0M7Bo0?ooo`2n0?oo
o`00j03oool00`3OW4P007Bo0?ooo`350?ooo`00i`3oool00`3ogi`0B00000000002000000030018
W03Oool0oooo0<40oooo003V0?ooo`040?oOW018000004RL0=ooo`80oooo00<0_g@000000018W=l0
`@3oool00>H0oooo00<0_g@00000B02Lgol0103oool00`3o_g@00000001d_`300?ooo`00i@3oool0
103ooml0W4P00018W03Oool50?ooo`030=nLB0000000M;oo0;l0oooo003U0?ooo`040?oog`2LB000
001d0;ooo`D0oooo00@0oooO09a8000007@0_ooo_P3oool00>H0oooo00<0_g@00000B02Lgol01P3o
ool00`3OW4P0000004RLg`2n0?ooo`00iP3oool00`3OW4P0000004RLg`060?ooo`040?oOW0180000
04RL0=oookd0oooo003V0?ooo`040?oo_`1d000004RL0=ooo`D0oooo00@0oooO09a80000B9`0gooo
_@3oool00>L0oooo00<0okmd00000018W=l01@3oool0103ooml0W4P00018W02WM003000000030000
M02oool0oooo0;L0oooo003X0?ooo`030=nLB0000000M;oo00@0oooo00T0oono07@00000M;l0oooo
0?noM000B9`0gooo0?oOW02L_ol0^03oool00>T0oooo00<0okmd00000000M;l00P3oool0103ooml0
W4P00000B02Lgol20?ooo`030?oOW018B9`0gooo0;T0oooo003[0?ooo`060;md0000000000000000
000007@0_ooo103oool00`3ogi`0B4RL0=ooo`2h0?ooo`00mP3oool00`3ogi`0M9cO0?ooo`2g0?oo
o`00m`3oool00`2oM4P0W=oo0?ooo`2f0?ooo`00l`3oool01P3ogml0gooo0?ooo`3oool0gia804RL
gkL0oooo003c0?ooo`060?oo_`1dB9`0gooo0?oog`2LB000B9cO]`3oool00?@0oooo00D0omnL04P0
0000000004RL0=ooo`2g0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00PP3oool00`3OW4P00000
000000030000000304RLg`3oool0oooo0400oooo00<0oooO09a80018W=l0@03oool01@3ooml0W4P0
00000000000007Bo0440oooo00H0oono07@00000000000000018W03Ooolj0?ooo`030?oo_`1d0000
00000080000000<0B9cO0?ooo`3oool00`3oool0102oM00000000000001d_ol=0?ooo`00P`3oool0
0`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo0440oooo00<0gia80018W03Oool0?`3oool0
1`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool0?P3oool00`3ookl0M00004RLg`020?ooo`030=nL
B00007@0_ooo03/0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RL
g``0oooo00240?ooo`030?noM000M;l0oooo04D0oooo00<0omnL04P0B02Lgol0?P3oool00`3ooml0
W4P007Boo`020?ooo`030?oOW0180000B9cO03h0oooo00<0okmd0018W03Oool00`3oool00`2oM000
B9cO0?ooo`0j0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`0;0?ooo`00Q@3oool00`3o_g@007Bo0?ooo`100?ooo`030?noM0000000000000@00000
00<004RL0=ooo`3oool0>`3oool00`3ookl0M00007Boo`030?ooo`030;md000007@0_ooo03d0oooo
00<0okmd0018W03Oool00`3oool00`2oM00007Bo0?ooo`0k0?ooo`030=nLB000B9`0gooo00<0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`00QP3oool00`3o_g@0B9cO
0?ooo`0o0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0?@3oool00`3ooml0W4P0
001d_`030?ooo`030=nLB00004P0W=oo03d0oooo00<0oooO09a8001d_ol00P3oool00`3ookl0M000
04RLg`0l0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P00018
09cOo`0:0?ooo`00Q`3oool00`3OW4P0B9cO0?ooo`0o0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@0
04RL0=ooo`0n0?ooo`030;md000007@0_ooo0080oooo00<0gia80000B02Lgol0?`3oool01P2oM000
001809cOg`2LB000001d0;oooc`0oooo00<0oono07@0001d_ol0103oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0oooo00280?ooo`030;md001d_ol0oooo03l0oooo00H0omnL04Q8
W03Oool0omnL04P0B02Lgoln0?ooo`080?noM0000000B9cO0?ooo`3ooml0W4P00018W03Ooom00?oo
o`050?oog`2LB00000000018W03Oool0?@3oool00`2oM00007Bo0?ooo`030?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02P3oool008P0oooo00<0omnL04P0M02oool0@03oool01@3o
_g@007Bo0?oog`2LB000B9cO03h0oooo00<0oooO09a8000000000P0000000`0007@0_ooo0?ooo`10
0?ooo`060;md000004P0W=oo0?oOW018B9`0gooo?03oool00`3OW4P004RL0=ooo`030?ooo`030?oO
W0180000B9cO00<0oooo00<0_g@00018W03Oool02@3oool008P0oooo00<0oooO09a8001d_ol0@@3o
ool0103OW4P0B9cO0;md0000M;lo0?ooo`030?oo_`1d000007Bo04<0oooo00<0omnL04P0001d_ol0
0P3oool00`3o_g@0B9cO0?ooo`0k0?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0
gooo0P3oool00`3OW4P004RL0=ooo`090?ooo`00R03oool00`3ookl0M00004RLg`120?ooo`040=nL
B01dM7@0001d0;ooocl0oooo00<0omnL04P00018W=l0@P3oool00`3ogi`0B01809cOo`020?ooo`03
0?noM000B9`0gooo03/0oooo00<0oono07@0001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`03
0=nLB000M;l0oooo00T0oooo00240?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`130?oo
o`030;md000004P0W=oo0400oooo00@0oooO09a8000007@0_ooo@03oool00`3ooml0W4P004RLg`02
0?ooo`030;md0000M;l0oooo03`0oooo00<0_g@0001d_`3oool0103oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`X0oooo00250?ooo`050?oOW018000000000000001d_ol0A03oool00`3ooml0W4P0
04RLg`120?ooo`040?oo_`1d0000000004RLgcl0oooo00D0oooO09a8000000000000001d_`0l0?oo
o`050?oo_`1d000000000000M02oool01@3oool0103OW4P0000000000000M;l;0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00@P3ooooo000005d000004P3o
ool00480oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`130?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0@`3oool00`00
0000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3o
ool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?oo
o`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0
ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool200000?l0ooooF03oool300000180oooo
00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?oo
o`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0
ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003o
ool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@
0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480
oooo0P00003o0?oooeP0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003o
ool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0ooooEP3oool403?o0180oooo00120?oo
o`030000003oool0oooo0?l0ooooD03oool603?o00<0oooo00<000000?ooo`3oool0403oool00480
oooo00<000000?ooo`3oool0o`3ooom:0?ooo`H0<ol02@3oool00`000000oooo0?ooo`0@0?ooo`00
@P3oool00`000000oooo0?ooo`3o0?ooodH0oooo100co`0?0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0?l0oooo@@3oool503?o01<0oooo00<000000?ooo`3oool0403o
ool00480oooo00<000000?ooo`3oool0o`3oooll0?ooo`D0<ol0603oool00`000000oooo0?ooo`0@
0?ooo`00@P3oool200000?l0oooo>03oool503?o01/0oooo0`00000B0?ooo`00@P3oool00`000000
oooo0?ooo`3o0?oooc80oooo1@0co`0R0?ooo`030000003oool0oooo0100oooo00120?ooo`030000
003oool0oooo0?l0oooo;@3oool503?o02L0oooo00<000000?ooo`3oool0403oool00480oooo00<0
00000?ooo`3oool0o`3ooolY0?ooo`@0<ol0;03oool00`000000oooo0?ooo`0@0?ooo`009P3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0gia80000000000000`00
00000`18W=l0oooo0?ooo`070?ooo`030000003oool0oooo0?l0oooo903oool503?o0300oooo00<0
00000?ooo`3oool0403oool002D0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`070?ooo`030000
003oool0oooo0?l0oooo803oool403?o03D0oooo00<000000?ooo`3oool0403oool002@0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?noM000M;l0oooo00/0oooo00<0
00000?ooo`3oool0o`3ooolL0?ooo`@0<ol0>@3oool00`000000oooo0?ooo`0@0?ooo`00903oool0
0`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool02P3o
ool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo05L0oooo100co`0m0?ooo`030000003o
ool0oooo0100oooo000T0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02@3oool0
0`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0D`3oool4
03?o0440oooo00<000000?ooo`3oool0403oool002D0oooo00<0_g@00018W03Oool00P3oool00`3o
gi`0B00004RLg`0:0?ooo`030=nLB018W=l0oooo00P0oooo0`00002m0?ooo`030000003oool0oooo
0500oooo0`0co`120?ooo`@000004P3oool002D0oooo00<0gia80000B02Lgol00P3oool00`3ooml0
W4P0001d_`0;0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0_@3oool00`000000
oooo0?ooo`1<0?ooo`@0<ol0B03oool00`000000oooo0?ooo`0@0?ooo`009@3oool00`3ogi`0B000
04RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02oool01`3oool00`000000oooo
0?ooo`2m0?ooo`030000003oool0oooo04P0oooo100co`1<0?ooo`030000003oool0oooo0100oooo
000U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3ooml0W4P0
07Boo`070?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0A03oool403?o0500oooo
00<000000?ooo`3oool0403oool002H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?oo
o`0:0?ooo`030?oo_`1d0000B9cO00L0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`110?ooo`<0<ol0E03oool00`000000oooo0?ooo`0@0?ooo`009`3oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00L0oooo00<000000?oo
o`3oool0_@3oool00`000000oooo0?ooo`0m0?ooo`@0<ol0E`3oool00`000000oooo0?ooo`0@0?oo
o`00:03oool0103OW4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol0203oool0
0`000000oooo0?ooo`2m0?ooo`030000003oool0oooo03X0oooo0`0co`1K0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0=P3oool403?o
05h0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000
oooo0?ooo`0c0?ooo`<0<ol0HP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool200000;h0oooo
00<000000?ooo`3oool0<03oool303?o06<0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?oo
o`2m0?ooo`030000003oool0oooo02d0oooo0`0co`1X0?ooo`030000003oool0oooo0100oooo0012
0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0:P3oool303?o06/0oooo00<00000
0?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0W0?oo
o`<0<ol0KP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2l0?ooo`<0
00009@3oool303?o0740oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0
^`3oool500000200oooo100co`1d0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0;/0oooo1@00000M0?ooo`<0<ol0N03oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`2k0?ooo`D000006@3oool403?o07/0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool0_03oool3000001L0oooo0`0co`1o0?ooo`030000003oool0oooo
0100oooo00120?ooo`800000_P3oool00`000000oooo0?ooo`0D0?ooo`80<ol0P03oool300000180
oooo00120?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool04P3oool203?o08@0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`0?0?ooo`<0<ol0QP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2m
0?ooo`030000003oool0oooo00d0oooo0P0co`290?ooo`030000003oool0oooo0100oooo00120?oo
o`030000003oool0oooo0;d0oooo00<000000?ooo`3oool02`3oool203?o08/0oooo00<000000?oo
o`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`080?ooo`<0
<ol0S@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2m0?ooo`030000
003oool0oooo00H0oooo0P0co`2@0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`<0<ol0TP3o
ool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01T0oooo00@000000?ooo`3oool0oooo0P0co`2E0?ooo`030000003oool0oooo0100oooo0012
0?ooo`800000XP3oool00`000000oooo0?ooo`0I0?ooo`040000000co`00<ol003?o09D0oooo0`00
000B0?ooo`00@P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo01P0oooo00<0<ol0
0000003oool0V@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo01H0oooo0P0co`000`3oool000000?ooo`2I0?ooo`030000003oool0oooo
0100oooo00120?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0503oool203?o00<0
oooo00<000000?ooo`3oool0V03oool00`000000oooo0?ooo`0@0?ooo`009P3oool0102oM0000000
0000001d_ol30?ooo`030?oOW0180000M;oo00D0oooo00<0oooO09a80018W=l02P3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0180oooo0P0co`050?ooo`030000003oool0oooo09P0
oooo00<000000?ooo`3oool0403oool002D0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?oo
o`030?oOW0180000M;oo00H0oooo00<0gia80018W03Oool02@3oool00`000000oooo0?ooo`2Q0?oo
o`030000003oool0oooo0100oooo0P0co`2R0?ooo`030000003oool0oooo0100oooo000T0?ooo`03
0?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02P3oool00`3ogi`0B01809cOo`090?ooo`03
0000003oool0oooo0:00oooo0`00000?0?ooo`80<ol0Y03oool00`000000oooo0?ooo`0@0?ooo`00
903oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<0okmd00000000
0000100000000`00B9`0gooo0?ooo`060?ooo`030000003oool0oooo09l0oooo1@00000<0?ooo`80
<ol0YP3oool00`000000oooo0?ooo`0@0?ooo`00903oool00`3ooml0W4P0001d_`030?ooo`030=nL
B00004P0W=oo00D0oooo00<0oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`080?ooo`030000
003oool0oooo09l0oooo1@00000:0?ooo`80<ol0Z03oool00`000000oooo0?ooo`0@0?ooo`009@3o
ool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00H0oooo00L0oono07@0M02oool0oooo
0?noM000B9`0gooo00P0oooo0`00002O0?ooo`D000001`3oool303?o0:L0oooo1000000B0?ooo`00
9@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00L0oooo00H0omnL04Q8W03Oool0
omnL04P0B02Lgol80?ooo`030000003oool0oooo0:00oooo0`0000060?ooo`80<ol0[@3oool00`00
0000oooo0?ooo`0@0?ooo`009@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0
oooo00D0okmd001d_`3ooml0W4P004RLg`080?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool00`3oool203?o0:l0oooo00<000000?ooo`3oool0403oool002D0oooo00@0oooO09a80000
B9`0gooo0P3oool00`3OW4P004RL0=ooo`080?ooo`040=nLB018W=l0_g@0001d_`P0oooo00<00000
0?ooo`3oool0X@3oool010000000oooo0?ooo`3oool203?o0;40oooo00<000000?ooo`3oool0403o
ool002H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`090?ooo`040=nLB01dM7@0
001d0;ooo`L0oooo00<000000?ooo`3oool0X@3oool010000000oooo03?o000co`2c0?ooo`030000
003oool0oooo0100oooo000W0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO2`3oool00`2oM000
001809cOo`070?ooo`030000003oool0oooo0:40oooo00<0000003?o003oool0]03oool00`000000
oooo0?ooo`0@0?ooo`00:03oool0103OW4P0000000000000M;l<0?ooo`030?oog`2LB000B9cO00L0
oooo00<000000?ooo`3oool0W`3oool203?o00030000003oool0oooo0;@0oooo00<000000?ooo`3o
ool0403oool00480oooo00<000000?ooo`3oool0W@3oool203?o0080oooo00<000000?ooo`3oool0
]03oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000009`0oooo0P0co`040?ooo`030000003o
ool0oooo0;80oooo0`00000B0?ooo`003`3oool00`3OW4P00000000000020000000304RLg`3oool0
oooo02/0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0@0?ooo`80<ol01P3oool0
0`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0100oooo000=0?ooo`030?noM0000000B9cO
00@0oooo00<0gia8001d_`3oool0:`3oool00`000000oooo0?ooo`260?ooo`030000003oool0oooo
00l0oooo00<0<ol00?ooo`3oool0_@3oool00`000000oooo0?ooo`0@0?ooo`00303oool00`3OW4P0
07Bo0?ooo`060?ooo`030?noM01d_ol0oooo02X0oooo00<000000?ooo`3oool0QP3oool00`000000
oooo0?ooo`0=0?ooo`80<ol0`03oool00`000000oooo0?ooo`0@0?ooo`002`3oool00`3o_g@0B9cO
0?ooo`080?ooo`030;nLW03Oool0oooo02T0oooo00<000000?ooo`3oool0QP3oool00`000000oooo
0?ooo`0;0?ooo`80<ol0`P3oool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3ooml0W9bo0?oo
o`090?ooo`030=nLW02oool0oooo02T0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?oo
o`0:0?ooo`0303?o003oool0oooo0<80oooo00<000000?ooo`3oool0403oool000X0oooo00<0omnL
09boo`3oool02@3oool00`3OW9`0_ooo0?ooo`0Y0?ooo`030000003oool0oooo08H0oooo00<00000
0?ooo`3oool0203oool203?o0<D0oooo00<000000?ooo`3oool0403oool000X0oooo00<0oknL09cO
o`3oool02@3oool00`2oM7@0_ooo0?ooo`0Y0?ooo`030000003oool0oooo08H0oooo00<000000?oo
o`3oool01P3oool203?o0<L0oooo00<000000?ooo`3oool0403oool000X0oooo00<0gibL0;ooo`3o
ool00`3oool00`3o_g@0M;oo0?ooo`020?ooo`030?oog`2LM9`0gooo02X0oooo00<000000?ooo`3o
ool0QP3oool00`000000oooo0?ooo`050?ooo`0303?o003oool0oooo0<L0oooo00<000000?ooo`3o
ool0403oool000X0oooo00<0okmd07Boo`3oool00`3oool0103o_g@0B9cO0?oog`2LB00200000003
04RLg`3oool0oooo02T0oooo0P0000260?ooo`<00000103oool203?o0<P0oooo0`00000B0?ooo`00
2P3oool00`3ookl0M01d0;ooo`030?ooo`060?noM000000000000000000007@0_ooo;03oool00`00
0000oooo0?ooo`240?ooo`D000000P3oool00`0co`00oooo0?ooo`3:0?ooo`030000003oool0oooo
0100oooo000:0?ooo`090?oo_`1d000000000000001d_ol0oooo0?noM000B9`0gooo02l0oooo00<0
00000?ooo`3oool0Q03oool500000080<ol0c@3oool00`000000oooo0?ooo`0@0?ooo`002P3oool0
0`3OW4P0000007Boo`030?ooo`030?noM01d_ol0oooo02l0oooo00<000000?ooo`3oool0Q03oool5
00000<l0oooo00<000000?ooo`3oool0403oool001T0oooo00<0_gAd0;ooo`3oool09P3oool00`00
0000oooo0?ooo`250?ooo`<00000d03oool00`000000oooo0?ooo`0@0?ooo`005P3oool01@3o_g@0
00000000000007@0_ooo02L0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`0;0?oo
o`80<ol000<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`3OW4P0
00000000000200000005001d_`3oool0oooo0=nLW02oool09`3oool00`000000oooo0?ooo`1f0?oo
o`030000003oool0oooo00X0oooo00@0<ol00?ooo`3oool00000d@3oool00`000000oooo0?ooo`0@
0?ooo`004@3oool01P3OW4P0M;oo0?ooo`3ooml0W7BL0=ooo`80oooo00<0gibL0;ooo`3oool09P3o
ool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo00P0oooo0P0co`030?ooo`030000003o
ool0oooo0<l0oooo00<000000?ooo`3oool0403oool00140oooo00X0gibL0;ooo`3oool0oooO09a8
00000000M;oo0?oog`2LB7@0_ooo9`3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo
00L0oooo00<0<ol00?ooo`3oool00`3oool00`000000oooo0?ooo`3?0?ooo`030000003oool0oooo
0100oooo000A0?ooo`050=nLW02oool0oono07@00000M;l00P3oool00`2oM000001d0;ooo`0W0?oo
o`800000M`3oool00`000000oooo0?ooo`050?ooo`80<ol01P3oool00`000000oooo0?ooo`3=0?oo
o`<000004P3oool00140oooo00<0gibL0;ooo`3oool00`3oool00`3OW4P004RL0=ooo`020?ooo`03
0;nLW03Oool0oooo02<0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`040?ooo`03
03?o003oool0oooo00H0oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`0@0?ooo`00
4@3oool00`3OW4P0000004RLg`070?ooo`030?oOW018B9`0gooo02@0oooo00<000000?ooo`3oool0
MP3oool00`000000oooo0?ooo`030?ooo`0303?o003oool0oooo00L0oooo00<000000?ooo`3oool0
c`3oool00`000000oooo0?ooo`0@0?ooo`00503oool0102oW9`0gooo0?ooo`3o_g@4000000030018
W03Oool0oooo02<0oooo00<000000?ooo`3oool0MP3oool010000000oooo0?ooo`3oool203?o00X0
oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`0@0?ooo`00503oool01`2oM0000000
00000000B9`0gooo0;mdM02oool02P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000
M;oo00<0oooo00D0oooO09a8000000000000001d_`0;0?ooo`030000003oool0oooo07H0oooo00@0
00000?ooo`3oool0<ol0gP3oool00`000000oooo0?ooo`0@0?ooo`00503oool00`3ookl0M7Bo0?oo
o`020?ooo`030=nLW02oool0oooo00P0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo00<0oooo00L0gia8001d_`3oool0oono07@00000B9`0gooo00T0oooo00<00000
0?ooo`3oool0MP3oool00`000000oooo03?o003O0?ooo`030000003oool0oooo0100oooo000D0?oo
o`060=nLW02oool0oooo0?ooo`3ookl0M7Bo2@3oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00H0oooo00<0oooO09a8001d_ol00P3oool00`3ogi`0B00004RLg`090?ooo`030000003o
ool0oooo07H0oooo00<0000003?o003oool0g`3oool00`000000oooo0?ooo`0@0?ooo`004`3oool0
1`3ooml0W7BL0=ooo`3oool0okmd0000M02oool02@3oool00`3ooml0W4P004RLg`020?ooo`040?oo
g`2LB00004RL0=ooo`D0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`080?ooo`03
0000003oool0oooo07D0oooo00<0<ol00000003oool0h03oool00`000000oooo0?ooo`0@0?ooo`00
503oool0103OW4P0000000000000M;l;0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol01@3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<000000?ooo`3o
ool0M03oool00`0co`00oooo0000003Q0?ooo`030000003oool0oooo0100oooo000T0?ooo`030;md
0000B9`0gooo0080oooo00<0omnL04P00018W=l01P3oool00`2oM000001d0;ooo`020?ooo`030=nL
B00004P0W=oo00P0oooo0`00001b0?ooo`80<ol000@0oooo0000000000000000g@3oool400000180
oooo000H0?ooo`040?oog`2LB000001d0;ooo`P0oooo00<0gia80000B02Lgol00P3oool00`3ooml0
W4P0001d_`060?ooo`080?noM0000000B9cO0?ooo`3ooml0W4P00018W03Oool80?ooo`030000003o
ool0oooo0740oooo00<0<ol00?ooo`3oool01@00003O0?ooo`030000003oool0oooo0100oooo000E
0?ooo`050?noM000000000000000B02Lgol02P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000
B9`0gooo00D0oooo00<0oooO09a8000000000P0000000`0007@0_ooo0?ooo`080?ooo`030000003o
ool0oooo06l0oooo0P0co`030?ooo`D00000g`3oool00`000000oooo0?ooo`0@0?ooo`004P3oool0
102oM000000000000000M;l>0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0
1P3oool00`3ookl0M000001d_`0<0?ooo`030000003oool0oooo06h0oooo00<0<ol00?ooo`3oool0
0`3oool500000=l0oooo00<000000?ooo`3oool0403oool000h0oooo00D0omnL04P0000000000000
07Boo`0B0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3ogi`0B000
04RLg`0;0?ooo`030000003oool0oooo06d0oooo00<0<ol00?ooo`3oool01@3oool300000>00oooo
00<000000?ooo`3oool0403oool000/0oooo00D0gia800000000000004RL0=ooo`0F0?ooo`060=nL
B00007@0_ooo0?oog`2LB000B9cO2@3oool0103ooml0W4P00000M02oool90?ooo`030000003oool0
oooo06@0oooo00<000000?ooo`3oool0103oool203?o00P0oooo0P00003Q0?ooo`030000003oool0
oooo0100oooo00090?ooo`030?oOW0180000B9cO01/0oooo00@0gia800000000000007Bo303oool0
103ookl0M00000000018W=l70?ooo`030000003oool0oooo06@0oooo00<000000?ooo`3oool00`3o
ool00`0co`00oooo0?ooo`070?ooo`<00000h@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`1T0?ooo`040000003oool0oooo0?ooo`80<ol02@3oool500000>00oooo00<0
00000?ooo`3oool0403oool000l0oooo00<0gia80000000000000P0000000`18W=l0oooo0?ooo`0[
0?ooo`030000003oool0oooo06@0oooo00@000000?ooo`3oool0<ol02`3oool500000>00oooo00<0
00000?ooo`3oool0403oool000d0oooo00<0okmd00000018W=l0103oool00`3OW4P007Bo0?ooo`0[
0?ooo`800000I03oool30000000303?o003oool0oooo00X0oooo1@00003N0?ooo`<000004P3oool0
00`0oooo00<0gia8001d_`3oool01P3oool00`3o_g@0M;oo0?ooo`0Z0?ooo`030000003oool0oooo
0680oooo1@00000=0?ooo`<00000h@3oool00`000000oooo0?ooo`0@0?ooo`002`3oool00`3o_g@0
B9cO0?ooo`080?ooo`030;nLW03Oool0oooo02T0oooo00<000000?ooo`3oool0HP3oool5000000h0
oooo0P00003Q0?ooo`030000003oool0oooo0100oooo000:0?ooo`030?oog`2LW;l0oooo00T0oooo
00<0gibL0;ooo`3oool0:@3oool00`000000oooo0?ooo`1R0?ooo`D000003P3oool200000>40oooo
00<000000?ooo`3oool0403oool000X0oooo00<0omnL09boo`3oool02@3oool00`3OW9`0_ooo0?oo
o`0Y0?ooo`030000003oool0oooo0640oooo00<0<ol00?ooo`0000000P00000?0?ooo`800000h@3o
ool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3o_i`0W=oo0?ooo`090?ooo`030;mdM02oool0
oooo02T0oooo00<000000?ooo`3oool0G`3oool203?o00<0oooo00<000000?ooo`3oool03P3oool2
00000>40oooo00<000000?ooo`3oool0403oool000X0oooo00<0gibL0;ooo`3oool00`3oool00`3o
_g@0M;oo0?ooo`020?ooo`030?oog`2LM9`0gooo02X0oooo00<000000?ooo`3oool0GP3oool00`0c
o`00oooo0?ooo`030?ooo`030000003oool0oooo00h0oooo0P00003Q0?ooo`030000003oool0oooo
0100oooo000:0?ooo`030?noM01d_ol0oooo00<0oooo00@0okmd04RLg`3ooml0W4P00P0000000`18
W=l0oooo0?ooo`0Y0?ooo`030000003oool0oooo05d0oooo00<0<ol00?ooo`3oool0103oool00`00
0000oooo0?ooo`0>0?ooo`800000h@3oool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3ookl0
M01d0;ooo`030?ooo`060?noM000000000000000000007@0_ooo;03oool00`000000oooo0?ooo`1L
0?ooo`0303?o003oool0oooo00D0oooo00<000000?ooo`3oool0l@3oool00`000000oooo0?ooo`0@
0?ooo`002P3oool02@3ookl0M000000000000000M;oo0?ooo`3o_g@004RL0=ooo`0_0?ooo`800000
G03oool00`0co`00oooo0?ooo`3h0?ooo`<000004P3oool000X0oooo00<0gia80000001d_ol00`3o
ool00`3o_g@0M;oo0?ooo`060?ooo`030?oOg`3Oool0oooo02H0oooo00<000000?ooo`3oool0FP3o
ool00`0co`00oooo0?ooo`3k0?ooo`030000003oool0oooo0100oooo000I0?ooo`030;mdM02oool0
oooo02H0oooo00<000000?ooo`3oool0F03oool203?o0?h0oooo00<000000?ooo`3oool0403oool0
0140oooo00H0gibL0;ooo`3oool0oooo0?oog`2LB002000000030000M02oool0oooo02H0oooo00<0
00000?ooo`3oool0E`3oool00`0co`00oooo0?ooo`3n0?ooo`030000003oool0oooo0100oooo000A
0?ooo`060=nLB000000000000000000004P0W=oo0P3oool00`3OW9`0_ooo0?ooo`0V0?ooo`030000
003oool0oooo05H0oooo00<0<ol00?ooo`3oool0o`3oool00`000000oooo0?ooo`0@0?ooo`004@3o
ool00`3OW9`0_ooo0?ooo`050?ooo`030=nLW02oool0oooo02H0oooo00<000000?ooo`3oool0E@3o
ool00`0co`00oooo0?ooo`3o0?ooo`40oooo00<000000?ooo`3oool0403oool00140oooo00<0gibL
0;ooo`3oool01@3oool00`2oW9`0gooo0?ooo`0V0?ooo`030000003oool0oooo0580oooo00<00000
0?ooo`0co`00o`3oool40?ooo`030000003oool0oooo0100oooo000A0?ooo`030=nLW02oool0oooo
00@0oooo00<0oono07Ad_`3oool09`3oool00`000000oooo0?ooo`1B0?ooo`030000000co`00oooo
0?l0oooo103oool00`000000oooo0?ooo`0@0?ooo`004@3oool00`3o_g@0M;oo0?ooo`040?ooo`03
0;mdB02Lgol0oooo02L0oooo00<000000?ooo`3oool0D@3oool300000?l0oooo1@3oool00`000000
oooo0?ooo`0@0?ooo`004@3oool00`3ookl0M01809cOo`020?ooo`030?oo_`1d0000M;oo02T0oooo
0P00001A0?ooo`D00000o`3oool20?ooo`<000004P3oool00180oooo00H0oooO09a8000000000000
0000M02ooolZ0?ooo`030000003oool0oooo04l0oooo00<0<ol00000000000000`00003o0?ooo`@0
oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0CP3oool00`0co`00oooo
0000000400000?l0oooo103oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?oo
o`160?ooo`030000003oool0oooo00@0oooo00@0<ol00?ooo`3oool0oooo0`00003o0?ooo`D0oooo
00<000000?ooo`3oool0403oool002H0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B000
07Boo`030?ooo`060?oo_`1d0000000000000000B9`0gooo2@3oool00`000000oooo0?ooo`160?oo
o`030000003oool0oooo00<0oooo00<0<ol00?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?oo
o`@0oooo00<000000?ooo`3oool0403oool002D0oooo00H0gia8001d_`3oool0omnL04P00018W=l2
0?ooo`030?oOW0180000M;oo0080oooo00<0oono07@00018W=l00P3oool00`3OW4P0001d0;ooo`08
0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`02
0?ooo`0303?o003oool0oooo00@0oooo00<000000?ooo`3oool0o`3oool40?ooo`030000003oool0
oooo0100oooo000T0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3o
_g@004RL0=ooo`030?ooo`030;md0018W=l0oooo00L0oooo00<000000?ooo`3oool0>@3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0oooo0080<ol01`3oool00`000000oooo0?ooo`3o0?oo
o`@0oooo00<000000?ooo`3oool0403oool002@0oooo00<0oooO09a80018W=l00P3oool0103ooml0
W4P00018W03Oool50?ooo`030?noM000B9`0gooo00<0oooo00<0_g@0001d_`3oool01`3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0co`00o`3oool@0?oo
o`030000003oool0oooo0100oooo000T0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol01@3oool00`3ooml0W4P007Boo`020?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`090?ooo`<00000o`3ooolA0?ooo`030000003oool0oooo0100
oooo000U0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool01P2oM0000018
09cOg`2LB000001d0;ooo`P0oooo0`00000i0?ooo`030000003oool0oooo00P0oooo1@00003o0?oo
o`d0oooo1000000B0?ooo`009@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00P0
oooo00D0oooO09a80000000004RL0=ooo`080?ooo`030000003oool0oooo03T0oooo00<000000?oo
o`3oool0203oool500000?l0oooo403oool00`000000oooo0?ooo`0@0?ooo`009@3oool00`3ogi`0
B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00H0_g@00000B02Lgol0omnL04Q8W03Oool7
0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`03
0?ooo`D00000o`3oool@0?ooo`030000003oool0oooo0100oooo000U0?ooo`040?oog`2LB00004RL
0=ooo`80oooo00<0gia80018W03Oool01P3oool00`3ogi`0B00007Boo`020?ooo`030?noM018W=l0
oooo00H0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo0?ooo`80<ol000@0oooo0000000000000000o`3ooolA0?ooo`030000003oool0oooo0100oooo
000V0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`3ogi`0B01809cO
o`020?ooo`030?noM000B9`0gooo00H0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?oo
o`060?ooo`8000000P3oool3000000030?ooo`0co`00oooo00<0oooo00<000000?ooo`3oool0o`3o
ool@0?ooo`030000003oool0oooo0100oooo000W0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO
1`3oool00`3ooml0W4P004RLg`020?ooo`030;md0000M;l0oooo00H0oooo00<000000?ooo`3oool0
<03oool00`000000oooo0?ooo`060?ooo`80000000<0oooo0000000000000`0000050?ooo`030000
003oool0oooo0?l0oooo403oool00`000000oooo0?ooo`0@0?ooo`00:03oool0103OW4P000000000
0000M;l90?ooo`050?oog`2LB000000000000000M;l0203oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0oooo00H0oooo200000050?ooo`030000003oool0oooo0?l0oooo403oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0_0?ooo`<000001P3oool9000000D0oooo
00<000000?ooo`3oool0o`3oool@0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo02h0oooo1@0000050?ooo`P00000o`3ooolI0?ooo`030000003oool0oooo0100oooo0012
0?ooo`800000;`3oool5000000@0oooo1P0000000`3oool000000?ooo`3o0?oooaL0oooo0`00000B
0?ooo`00@P3oool00`000000oooo0?ooo`0^0?ooo`D00000103oool500000080oooo00<000000?oo
o`3oool0o`3ooolH0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo02l0
oooo0`0000050?ooo`D000000P3oool00`000000oooo0?ooo`3o0?oooaP0oooo00<000000?ooo`3o
ool0403oool00480oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`050?ooo`<00000
o`3ooolN0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool00`3oool203?o00030?ooo`00000000000?l0oooo7P3oool00`000000oooo0?oo
o`0@0?ooo`00@P3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0080oooo00D0<ol0
0?ooo`3oool0oooo0000003o0?oooal0oooo00<000000?ooo`3oool0403oool00480oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo00h0oooo00<0<ol0
0?ooo`3oool00P3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo
00`0oooo0P0co`050?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000oooo0?ooo`0@0?oo
o`00@P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0000001H0oooo00<000000?ooo`3o
ool02`3oool00`0co`00oooo0?ooo`050?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool2000000/0oooo00<000000?ooo`0000005P3oool00`000000oooo
0?ooo`0:0?ooo`0303?o003oool0oooo00H0oooo00<000000?ooo`3oool0o`3ooolK0?ooo`<00000
4P3oool00480oooo00<000000?ooo`3oool02P3oool00`000000oooo0000000E0?ooo`<000002@3o
ool203?o00T0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo0012
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`000000503oool5000000L0oooo00<0<ol0
0?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool02P3oool4000001<0oooo1P0000040?ooo`030000000co`00oooo
00/0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`03
0000003oool0oooo00T0oooo1P00000B0?ooo`L000000P3oool00`0co`0000000?ooo`0<0?ooo`03
0000003oool0oooo0?l0oooo7@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo
0?ooo`080?ooo`L00000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00L0oooo
1P000000103oool0<ol00?ooo`00000=0?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`080?ooo`L00000103oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo00P0oooo1@000000100co`00oooo0?ooo`00000=0?ooo`03
0000003oool0oooo0?l0oooo7@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo
0?ooo`080?ooo`H000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo
100000040?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3o
ool00`000000oooo0?ooo`050?ooo`<0000000<0oooo0000000000000P0000060?ooo`040000003o
ool0oooo0?ooo`<000002@3oool00`000000<ol00?ooo`050?ooo`030000003oool0oooo0?l0oooo
:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000000D0oooo1`0000000`3oool000000?oo
o`050?ooo`030000003oool0oooo00D000001P3oool203?o00030000003oool0oooo00@0oooo0`00
003o0?ooobX0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?ooo`040?ooo`L0000000<0oooo
0000003oool00`3oool00`000000oooo000000020?ooo`D00000103oool203?o00P0oooo1@00003o
0?ooob/0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool02`0000000`3o
ool0000000000007000000030?ooo`000000000000<0000000@0oooo03?o000co`00<ol02P3oool5
00000?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3ooolN0000000303?o003oool0oooo
00/0oooo1@00003o0?ooob/0oooo00<000000?ooo`3oool0403oool00340oooo00<0oono07@00000
00000P0000000`18W=l0oooo0?ooo`090?oooah000003`3oool300000?l0oooo;03oool00`000000
oooo0?ooo`0@0?ooo`00<`3oool00`3o_g@0000007Boo`0<0?ooo`T0000000<0oooo000000000000
4P00000@0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00<`3o
ool00`3ooml0W4P0001d_`0<0?ooo`P000000P3oool>00000080oooo0`00000A0?ooo`030000003o
ool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00=03oool00`3OW4P004RL0=ooo`0;
0?ooo`@000001P3oool;000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`3o0?ooob/0oooo00<000000?ooo`3oool0403oool003@0oooo00<0omnL04P0
B02Lgol02`3oool4000000L0oooo2P0000000`3oool000000?ooo`0F0?ooo`030000003oool0oooo
0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00=03oool00`3ookl0M00007Boo`0;0?ooo`@0
00001`3oool:000000030?ooo`000000oooo01H0oooo00<000000?ooo`3oool0o`3ooolX0?ooo`@0
00004P3oool003D0oooo00<0_g@0001d_`3oool02P3oool3000000P0oooo2@0000020?ooo`030000
003oool0oooo0?l0oooo@`3oool00`000000oooo0?ooo`0@0?ooo`00=@3oool00`3OW4P004RL0=oo
o`0:0?ooo`8000002P3oool800000080oooo00<000000?ooo`3oool0o`3ooom30?ooo`030000003o
ool0oooo0100oooo000e0?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool02P3oool0
0`000000oooo00000004000000`0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0
oooo0100oooo000e0?ooo`030?oo_`1d0000M;oo00X0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo00000004000000@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o0?oo
ocT0oooo00<000000?ooo`3oool0403oool003H0oooo00<0_g@0001d_`3oool02@3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00<000001@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00=03oool01@3ookl0M000
0000000007@0_ooo00T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100
oooo00120?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103o
ool300000?l0oooo>P3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000000l0oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`D00000o`3ooolg0?ooo`<000004P3oool0
0480oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
00<0oooo1@00003o0?ooocT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo1@00003o0?ooocT0
oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool03P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00@0oooo0`00003o0?ooocX0oooo00<000000?ooo`3oool0
403oool00480oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo01@0oooo0`0000060?ooo`030000003oool0oooo
0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0C0?ooo`D0
00001@3oool00`000000oooo0?ooo`3o0?ooocT0oooo00<000000?ooo`3oool0403oool00480oooo
00<000000?ooo`3oool04`3oool5000000D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000
003oool0oooo0100oooo00120?ooo`800000503oool5000000D0oooo00<000000?ooo`3oool0o`3o
oolg0?ooo`<000004P3oool00480oooo00<000000?ooo`3oool0503oool300000?l0oooo@P3oool0
0`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo
0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000oooo0?oo
o`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool0
0`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo
0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000001H0oooo
00<000000?ooo`3oool0o`3ooolo0?ooo`<000004P3oool00480oooo00<000000?ooo`3oool05@3o
ool00`000000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3oool0403oool00480oooo00<00000
0?ooo`3oool05@3oool00`000000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`03
0000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo
0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0@P3o
ool00`000000oooo0?ooo`120?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@P3o
ool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`00@P3ooooo0000
05d000004P3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.90205, -0.287596, \
0.0432244, 0.00839264}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        Plot[Evaluate[\(\(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + 
                      b3\ qq\^3\)\) \((1 + qq\/0.71)\)\^2 /. 
              Thread[{a1, b1, b2, 
                    b3} \[Rule] \((Parameters /. gepresults)\)]], {qq, 0, 
            10}, PlotRange \[Rule] {{0, 2}, {0.8, 1.2}}, Frame \[Rule] True, 
          FrameLabel \[Rule] {\*"\"\<\!\(Q\^2\)\>\"", \
\*"\"\<\!\(G\_Ep\)/\!\(G\_D\)\>\""}], \[IndentingNewLine]Plot[
          Evaluate[\(\(1 + a1\ qq\)\/\(1 + b1\ qq\  + \ b2\ qq\^2 + 
                      b3\ qq\^3\)\) \((1 + qq\/0.71)\)\^2 /. \
\((BestFitParameters /. gepresults2)\)], {qq, 0, 10}, 
          PlotStyle \[Rule] Hue[0.3]], \[IndentingNewLine]ErrorListPlot[
          ratiodata]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0 0.5 -1.23607 1.54508 [
[.125 -0.0125 -12.5 -15.6875 ]
[.125 -0.0125 12.5 0 ]
[.25 -0.0125 -9.5 -15.6875 ]
[.25 -0.0125 9.5 0 ]
[.375 -0.0125 -12.5 -15.6875 ]
[.375 -0.0125 12.5 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.625 -0.0125 -12.5 -15.6875 ]
[.625 -0.0125 12.5 0 ]
[.75 -0.0125 -9.5 -15.6875 ]
[.75 -0.0125 9.5 0 ]
[.875 -0.0125 -12.5 -15.6875 ]
[.875 -0.0125 12.5 0 ]
[1 -0.0125 -5 -15.6875 ]
[1 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -8.71875 -31.375 ]
[.5 -0.0125 8.71875 -15.6875 ]
[-0.0125 .07725 -25 -7.84375 ]
[-0.0125 .07725 0 7.84375 ]
[-0.0125 .15451 -19 -7.84375 ]
[-0.0125 .15451 0 7.84375 ]
[-0.0125 .23176 -25 -7.84375 ]
[-0.0125 .23176 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .38627 -25 -7.84375 ]
[-0.0125 .38627 0 7.84375 ]
[-0.0125 .46353 -19 -7.84375 ]
[-0.0125 .46353 0 7.84375 ]
[-0.0125 .54078 -25 -7.84375 ]
[-0.0125 .54078 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -41.4375 -20.9375 ]
[-0.01875 .30902 -25 20.9375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.125 0 m
.125 .00625 L
s
gsave
.125 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.25 0 m
.25 .00625 L
s
gsave
.25 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.375 0 m
.375 .00625 L
s
gsave
.375 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.625 0 m
.625 .00625 L
s
gsave
.625 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.25) show
1.000 setlinewidth
grestore
.75 0 m
.75 .00625 L
s
gsave
.75 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.875 0 m
.875 .00625 L
s
gsave
.875 -0.0125 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.75) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.125 Mabswid
.025 0 m
.025 .00375 L
s
.05 0 m
.05 .00375 L
s
.075 0 m
.075 .00375 L
s
.1 0 m
.1 .00375 L
s
.15 0 m
.15 .00375 L
s
.175 0 m
.175 .00375 L
s
.2 0 m
.2 .00375 L
s
.225 0 m
.225 .00375 L
s
.275 0 m
.275 .00375 L
s
.3 0 m
.3 .00375 L
s
.325 0 m
.325 .00375 L
s
.35 0 m
.35 .00375 L
s
.4 0 m
.4 .00375 L
s
.425 0 m
.425 .00375 L
s
.45 0 m
.45 .00375 L
s
.475 0 m
.475 .00375 L
s
.525 0 m
.525 .00375 L
s
.55 0 m
.55 .00375 L
s
.575 0 m
.575 .00375 L
s
.6 0 m
.6 .00375 L
s
.65 0 m
.65 .00375 L
s
.675 0 m
.675 .00375 L
s
.7 0 m
.7 .00375 L
s
.725 0 m
.725 .00375 L
s
.775 0 m
.775 .00375 L
s
.8 0 m
.8 .00375 L
s
.825 0 m
.825 .00375 L
s
.85 0 m
.85 .00375 L
s
.9 0 m
.9 .00375 L
s
.925 0 m
.925 .00375 L
s
.95 0 m
.95 .00375 L
s
.975 0 m
.975 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -69.7188 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Q) show
71.625 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
76.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .07725 m
.00625 .07725 L
s
gsave
-0.0125 .07725 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.85) show
1.000 setlinewidth
grestore
0 .15451 m
.00625 .15451 L
s
gsave
-0.0125 .15451 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.9) show
1.000 setlinewidth
grestore
0 .23176 m
.00625 .23176 L
s
gsave
-0.0125 .23176 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.95) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .38627 m
.00625 .38627 L
s
gsave
-0.0125 .38627 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.05) show
1.000 setlinewidth
grestore
0 .46353 m
.00625 .46353 L
s
gsave
-0.0125 .46353 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.1) show
1.000 setlinewidth
grestore
0 .54078 m
.00625 .54078 L
s
gsave
-0.0125 .54078 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.15) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .01545 m
.00375 .01545 L
s
0 .0309 m
.00375 .0309 L
s
0 .04635 m
.00375 .04635 L
s
0 .0618 m
.00375 .0618 L
s
0 .09271 m
.00375 .09271 L
s
0 .10816 m
.00375 .10816 L
s
0 .12361 m
.00375 .12361 L
s
0 .13906 m
.00375 .13906 L
s
0 .16996 m
.00375 .16996 L
s
0 .18541 m
.00375 .18541 L
s
0 .20086 m
.00375 .20086 L
s
0 .21631 m
.00375 .21631 L
s
0 .24721 m
.00375 .24721 L
s
0 .26266 m
.00375 .26266 L
s
0 .27812 m
.00375 .27812 L
s
0 .29357 m
.00375 .29357 L
s
0 .32447 m
.00375 .32447 L
s
0 .33992 m
.00375 .33992 L
s
0 .35537 m
.00375 .35537 L
s
0 .37082 m
.00375 .37082 L
s
0 .40172 m
.00375 .40172 L
s
0 .41717 m
.00375 .41717 L
s
0 .43262 m
.00375 .43262 L
s
0 .44807 m
.00375 .44807 L
s
0 .47898 m
.00375 .47898 L
s
0 .49443 m
.00375 .49443 L
s
0 .50988 m
.00375 .50988 L
s
0 .52533 m
.00375 .52533 L
s
0 .55623 m
.00375 .55623 L
s
0 .57168 m
.00375 .57168 L
s
0 .58713 m
.00375 .58713 L
s
0 .60258 m
.00375 .60258 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -115.156 -12.2188 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.9375 12.2188 translate
90 rotate
-81.9375 -12.2188 translate
0 24.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
71.688 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Ep) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
81.625 15.813 moveto
(\\220) show
85.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
94.250 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(D) show
100.875 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .61178 m
0 .61803 L
s
.125 .61178 m
.125 .61803 L
s
.25 .61178 m
.25 .61803 L
s
.375 .61178 m
.375 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.625 .61178 m
.625 .61803 L
s
.75 .61178 m
.75 .61803 L
s
.875 .61178 m
.875 .61803 L
s
.125 Mabswid
.025 .61428 m
.025 .61803 L
s
.05 .61428 m
.05 .61803 L
s
.075 .61428 m
.075 .61803 L
s
.1 .61428 m
.1 .61803 L
s
.15 .61428 m
.15 .61803 L
s
.175 .61428 m
.175 .61803 L
s
.2 .61428 m
.2 .61803 L
s
.225 .61428 m
.225 .61803 L
s
.275 .61428 m
.275 .61803 L
s
.3 .61428 m
.3 .61803 L
s
.325 .61428 m
.325 .61803 L
s
.35 .61428 m
.35 .61803 L
s
.4 .61428 m
.4 .61803 L
s
.425 .61428 m
.425 .61803 L
s
.45 .61428 m
.45 .61803 L
s
.475 .61428 m
.475 .61803 L
s
.525 .61428 m
.525 .61803 L
s
.55 .61428 m
.55 .61803 L
s
.575 .61428 m
.575 .61803 L
s
.6 .61428 m
.6 .61803 L
s
.65 .61428 m
.65 .61803 L
s
.675 .61428 m
.675 .61803 L
s
.7 .61428 m
.7 .61803 L
s
.725 .61428 m
.725 .61803 L
s
.775 .61428 m
.775 .61803 L
s
.8 .61428 m
.8 .61803 L
s
.825 .61428 m
.825 .61803 L
s
.85 .61428 m
.85 .61803 L
s
.9 .61428 m
.9 .61803 L
s
.925 .61428 m
.925 .61803 L
s
.95 .61428 m
.95 .61803 L
s
.975 .61428 m
.975 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .07725 m
1 .07725 L
s
.99375 .15451 m
1 .15451 L
s
.99375 .23176 m
1 .23176 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .38627 m
1 .38627 L
s
.99375 .46353 m
1 .46353 L
s
.99375 .54078 m
1 .54078 L
s
.125 Mabswid
.99625 .01545 m
1 .01545 L
s
.99625 .0309 m
1 .0309 L
s
.99625 .04635 m
1 .04635 L
s
.99625 .0618 m
1 .0618 L
s
.99625 .09271 m
1 .09271 L
s
.99625 .10816 m
1 .10816 L
s
.99625 .12361 m
1 .12361 L
s
.99625 .13906 m
1 .13906 L
s
.99625 .16996 m
1 .16996 L
s
.99625 .18541 m
1 .18541 L
s
.99625 .20086 m
1 .20086 L
s
.99625 .21631 m
1 .21631 L
s
.99625 .24721 m
1 .24721 L
s
.99625 .26266 m
1 .26266 L
s
.99625 .27812 m
1 .27812 L
s
.99625 .29357 m
1 .29357 L
s
.99625 .32447 m
1 .32447 L
s
.99625 .33992 m
1 .33992 L
s
.99625 .35537 m
1 .35537 L
s
.99625 .37082 m
1 .37082 L
s
.99625 .40172 m
1 .40172 L
s
.99625 .41717 m
1 .41717 L
s
.99625 .43262 m
1 .43262 L
s
.99625 .44807 m
1 .44807 L
s
.99625 .47898 m
1 .47898 L
s
.99625 .49443 m
1 .49443 L
s
.99625 .50988 m
1 .50988 L
s
.99625 .52533 m
1 .52533 L
s
.99625 .55623 m
1 .55623 L
s
.99625 .57168 m
1 .57168 L
s
.99625 .58713 m
1 .58713 L
s
.99625 .60258 m
1 .60258 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
0 .30902 m
.01179 .29714 L
.0244 .28733 L
.0363 .28026 L
.04712 .27531 L
.05985 .27095 L
.06703 .26907 L
.07366 .26765 L
.08583 .26572 L
.09198 .26502 L
.09874 .26444 L
.10567 .26403 L
.11226 .26378 L
.11818 .26367 L
.12463 .26364 L
.1283 .26367 L
.13162 .26371 L
.13546 .26379 L
.13908 .26389 L
.14552 .2641 L
.15249 .26439 L
.20283 .26731 L
.21682 .26811 L
.23181 .26887 L
.2457 .26945 L
.25253 .26967 L
.25886 .26984 L
.26593 .26999 L
.26966 .27005 L
.27362 .2701 L
.28001 .27015 L
.28366 .27015 L
.28707 .27014 L
.29086 .27012 L
.29485 .27008 L
.29825 .27003 L
.30198 .26996 L
.3103 .26975 L
.31805 .26948 L
.33186 .26882 L
.3449 .26799 L
.36929 .26586 L
.39557 .26275 L
.42404 .25841 L
.47373 .24852 L
.52839 .23447 L
.63887 .19741 L
.85065 .10403 L
s
.85065 .10403 m
1 .02844 L
s
.2 1 0 r
0 .30902 m
.01179 .29714 L
.0244 .28733 L
.0363 .28025 L
.04712 .2753 L
.05985 .27095 L
.06703 .26907 L
.07366 .26765 L
.08583 .26572 L
.09198 .26503 L
.09874 .26445 L
.10567 .26404 L
.11226 .2638 L
.11818 .26368 L
.12463 .26366 L
.1283 .26369 L
.13162 .26373 L
.13546 .26381 L
.13908 .26391 L
.14552 .26413 L
.15249 .26442 L
.20283 .26735 L
.21682 .26816 L
.23181 .26892 L
.2457 .26949 L
.25253 .26972 L
.25886 .26989 L
.26593 .27004 L
.26966 .2701 L
.27362 .27015 L
.28001 .27019 L
.28366 .2702 L
.28707 .27019 L
.29086 .27017 L
.29485 .27013 L
.29825 .27008 L
.30198 .27001 L
.3103 .2698 L
.31805 .26953 L
.33186 .26887 L
.3449 .26804 L
.36929 .26591 L
.39557 .26279 L
.42404 .25844 L
.47373 .24854 L
.52839 .23447 L
.63887 .19737 L
.85065 .10393 L
s
.85065 .10393 m
1 .02832 L
s
0 g
.002 w
.00273 .30715 m
.00273 .31395 L
s
.015 w
.00273 .31055 Mdot
.002 w
.0039 .30246 m
.0039 .31017 L
s
.015 w
.0039 .30631 Mdot
.002 w
.00487 .30412 m
.00487 .31091 L
s
.015 w
.00487 .30752 Mdot
.002 w
.00585 .29293 m
.00585 .30614 L
s
.015 w
.00585 .29954 Mdot
.002 w
.0068 .30312 m
.0068 .30989 L
s
.015 w
.0068 .30651 Mdot
.002 w
.0078 .29107 m
.0078 .30643 L
s
.015 w
.0078 .29875 Mdot
.002 w
.00875 .29656 m
.00875 .30146 L
s
.015 w
.00875 .29901 Mdot
.002 w
.00975 .29699 m
.00975 .30652 L
s
.015 w
.00975 .30175 Mdot
.002 w
.0107 .29502 m
.0107 .29994 L
s
.015 w
.0107 .29748 Mdot
.002 w
.0117 .29582 m
.0117 .30505 L
s
.015 w
.0117 .30043 Mdot
.002 w
.01265 .29339 m
.01265 .29889 L
s
.015 w
.01265 .29614 Mdot
.002 w
.01365 .28565 m
.01365 .29634 L
s
.015 w
.01365 .29099 Mdot
.002 w
.0156 .28606 m
.0156 .29431 L
s
.015 w
.0156 .29018 Mdot
.002 w
.01655 .28957 m
.01655 .29936 L
s
.015 w
.01655 .29447 Mdot
.002 w
.0175 .29322 m
.0175 .30458 L
s
.015 w
.0175 .2989 Mdot
.002 w
.01945 .27123 m
.01945 .29554 L
s
.015 w
.01945 .28338 Mdot
.002 w
.02335 .27066 m
.02335 .28884 L
s
.015 w
.02335 .27975 Mdot
.002 w
.02725 .27445 m
.02725 .2866 L
s
.015 w
.02725 .28052 Mdot
.002 w
.065 .24769 m
.065 .28199 L
s
.015 w
.065 .26484 Mdot
.002 w
.095 .25615 m
.095 .28073 L
s
.015 w
.095 .26844 Mdot
.002 w
.135 .22411 m
.135 .26508 L
s
.015 w
.135 .24459 Mdot
.002 w
.165 .22946 m
.165 .26214 L
s
.015 w
.165 .2458 Mdot
.002 w
.195 .28097 m
.195 .31769 L
s
.015 w
.195 .29933 Mdot
.002 w
.225 .27149 m
.225 .34961 L
s
.015 w
.225 .31055 Mdot
.002 w
.265 .25176 m
.265 .35827 L
s
.015 w
.265 .30502 Mdot
.002 w
.29 .24181 m
.29 .28516 L
s
.015 w
.29 .26349 Mdot
.002 w
.325 .20105 m
.325 .33144 L
s
.015 w
.325 .26624 Mdot
.002 w
.36 .3484 m
.36 .56401 L
s
.015 w
.36 .45621 Mdot
.002 w
.39 .19666 m
.39 .28444 L
s
.015 w
.39 .24055 Mdot
.002 w
.47 .33555 m
.47 .46339 L
s
.015 w
.47 .39947 Mdot
.002 w
.55 .12986 m
.55 .22967 L
s
.015 w
.55 .17976 Mdot
.002 w
.675 .17829 m
.675 .31512 L
s
.015 w
.675 .24671 Mdot
.002 w
.875 0 m
.875 .20477 L
s
.015 w
.875 .08643 Mdot
.002 w
.1865 .24146 m
.1865 .32773 L
s
.015 w
.1865 .28459 Mdot
.002 w
.2005 .26209 m
.2005 .35289 L
s
.015 w
.2005 .30749 Mdot
.002 w
.2205 .14858 m
.2205 .24604 L
s
.015 w
.2205 .19731 Mdot
.002 w
.19 .13316 m
.19 .28611 L
s
.015 w
.19 .20964 Mdot
.002 w
.25 .25828 m
.25 .41085 L
s
.015 w
.25 .33456 Mdot
.002 w
.245 .24648 m
.245 .29591 L
s
.015 w
.245 .27119 Mdot
.002 w
.395 .25147 m
.395 .29009 L
s
.015 w
.395 .27078 Mdot
.002 w
.59 .17849 m
.59 .2201 L
s
.015 w
.59 .1993 Mdot
.002 w
.74 .03021 m
.74 .12429 L
s
.015 w
.74 .07725 Mdot
.002 w
.885 .03168 m
.885 .11085 L
s
.015 w
.885 .07126 Mdot
.002 w
.94 .01443 m
.94 .09256 L
s
.015 w
.94 .05349 Mdot
.002 w
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00k03oool01@3ogkl0_moo0?ooo`3oool0gia800@0000000<0001d0;oo
o`3oool0^03oool00>`0oooo00<0oooO09a8000000000P0000001`0007@0_ooo0?ooo`3oool0oono
07@0M02oool0^03oool00>h0oooo00<0_g@004RLg`3oool01@3oool00`3ookl0M7Bo0?ooo`2g0?oo
o`00k`3oool00`3OW4P007Bo0?ooo`2n0?ooo`00kP3oool00`3ogi`0B00000000002000000030018
W03Oool0oooo0;X0oooo003]0?ooo`040?oOW018000004RL0=ooo`80oooo00<0_g@000000018W=l0
^P3oool00>d0oooo00<0_g@00000B02Lgol0103oool00`3o_g@00000001d_`2i0?ooo`00k03oool0
103ooml0W4P00018W03Oool50?ooo`030=nLB0000000M;oo0;P0oooo003/0?ooo`040?oog`2LB000
001d0;ooo`D0oooo00@0oooO09a8000007@0_ooo]`3oool00>d0oooo00<0_g@00000B02Lgol01P3o
ool00`3OW4P0000004RLg`2g0?ooo`00k@3oool00`3OW4P0000004RLg`060?ooo`040?oOW0180000
04RL0=oookH0oooo003]0?ooo`040?oo_`1d000004RL0=ooo`D0oooo00@0oooO09a80000B9`0gooo
]P3oool00>h0oooo00<0okmd00000018W=l01@3oool00`3ooml0W4P0000P0003000000030000M02o
ool0oooo0;40oooo003_0?ooo`030=nLB0000000M;oo00@0oooo00P0oono07@00000M;l0okmd0018
W03Oool0omnL09book80oooo003`0?ooo`030?noM000000007Bo0080oooo00P0oooO09a8000004P0
W=oo0?ooo`3ogi`0B4RL0=oook<0oooo003b0?ooo`060;md0000000000000000000007@0_ooo0`3o
ool00`3ogi`0B4RL0=ooo`2b0?ooo`00o03oool00`3ogi`0M9cO0?ooo`2a0?ooo`00o@3oool00`2o
M4P0W=oo0?ooo`2`0?ooo`00n@3oool01P3ogml0gooo0?ooo`3oool0gia804RLgk40oooo003i0?oo
o`060?oo_`1dB9`0gooo0?oog`2LB000B9cO/@3oool00?X0oooo00D0omnL04P00000000004RL0=oo
o`2a0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00IP3oool0102oM00000000000001d_ol30?oo
o`030?oOW0180000M;oo0080oooo00<0gia80000000000000`0000001@18W=l0oooo0?ooo`3ogi`0
B0000080000000<004RL0=ooo`3oool0503oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03OoolD0?ooo`040;md00000000000007Bo
o`<0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0M4RL0=ooo`050?ooo`060?oOW01800000000
00000000B9`0gooo6P3oool00`3ookl0M000000000020000000304RLg`3oool0oooo01P0oooo00<0
oono07@0000000000P0000001`18W=l0oooo0?ooo`3oool0omnL04P0001d_ol00P3oool00`3OW4P0
0000000000030000000504RLg`3oool0oooo0?oOW01800000P0000000`00B9`0gooo0?ooo`0B0?oo
o`030?oo_`1d000000000080000000L0B9cO0?ooo`3oool0oooo0?oOW0180000M;oo0080oooo00H0
omnL04P00000000000000018W03OoolC0?ooo`030?oo_`1d000000000080000000L0B9cO0?ooo`3o
ool0oooo0?oOW0180000M;oo00<0oooo00<0oono07A8W03Oool01@3oool01P3ogi`0B00000000000
000004RL0=oooaX0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`080?ooo`00I@3o
ool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool00`3o_g@0
07Bo0?ooo`020?ooo`090;mdM02oool0oooo0?noM0000000M;oo0?ooo`3ogi`0B7Bo01@0oooo00H0
gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080oooo00H0okmd0000001d
_ol0oooo0?oOW018M;lC0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B000
07Boo`040?ooo`030=nLB018W=l0oooo00@0oooo00H0okmd0000001d_ol0oooo0?oOW018M;lL0?oo
o`030?noM0000000M;oo01d0oooo00<0okmd0000001d_ol0103oool00`3ogi`0B00007Boo`030?oo
o`030?noM000M;l0oooo0080oooo00T0_gAd0;ooo`3oool0okmd0000001d_ol0oooo0?oOW018M;l0
5@3oool00`3o_g@0000007Boo`040?ooo`030?oOW0180000M;oo0080oooo00H0okmd0000001d_ol0
oooo0?oOW018M;lE0?ooo`030?noM0000000M;oo00@0oooo00<0omnL04P0001d_ol0103oool00`3O
W4P0B9cO0?ooo`040?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo6`3oool00`3o_g@007Bo0?oo
o`020?ooo`030;mdM02oool0oooo00P0oooo001T0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd
0000B02Lgol0203oool00`3o_g@007Bo0?ooo`090?ooo`030?noM018W=l0oooo0140oooo00<0oooO
09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?ooo`030?noM018W=l0oooo0100oooo00<0oooO
09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<0okmd
04RLg`3oool06P3oool00`3ooml0W4P0001d_`0M0?ooo`030?oog`2LB00007Bo00/0oooo00<0okmd
001d_`3oool02@3oool00`3o_g@0B9cO0?ooo`0C0?ooo`030?oog`2LB00007Bo00h0oooo00<0okmd
04RLg`3oool04`3oool00`3ooml0W4P0001d_`0;0?ooo`030?oo_`1d07@0_ooo00T0oooo00<0okmd
04RLg`3oool06P3oool00`3o_g@007Bo0?ooo`0<0?ooo`00I03oool00`3ooml0W4P004RLg`020?oo
o`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool0203oool00`3ookl0M01d0;ooo`0A
0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3oool00`3ookl0M01d0;oo
o`0@0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo
0?ooo`080?ooo`030?oo_`1d07@0_ooo01/0oooo00<0gia80018W03Oool07@3oool00`3OW4P004RL
0=ooo`0;0?ooo`030?noM000M;l0oooo00P0oooo00<0oono07@0M02oool0503oool00`3OW4P004RL
0=ooo`0=0?ooo`030?oo_`1d07@0_ooo01@0oooo00<0gia80018W03Oool02`3oool00`3o_g@007Bo
0?ooo`080?ooo`030?oo_`1d07@0_ooo01/0oooo00<0okmd001d_`3oool02`3oool006@0oooo00<0
oooO09a80000M;l00`3oool00`3OW4P0001809cOo`090?ooo`030?noM018W=l0oooo00L0oooo00<0
oooO09a8001d_ol04@3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00X0oooo00<0
oooO09a8001d_ol0403oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<0
oooO09a8B02Lgol0203oool00`3ooml0W4P007Boo`0K0?ooo`030?oOW01804P0W=oo01d0oooo00<0
omnL04P0B02Lgol0303oool00`3o_g@0B9cO0?ooo`070?ooo`030?oog`2LB000M;oo01@0oooo00<0
omnL04P0B02Lgol03@3oool00`3ooml0W4P007Boo`0D0?ooo`030?oOW01804P0W=oo00/0oooo00<0
oooO09a8B02Lgol0203oool00`3ooml0W4P007Boo`0L0?ooo`030?noM018W=l0oooo00X0oooo001U
0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3OW4P0B9cO0?ooo`06
0?ooo`030?oo_`1d0000B9cO0180oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:
0?ooo`030?oo_`1d0000B9cO0140oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`09
0?ooo`030?oOW018M;l0oooo00L0oooo00<0oono07@00018W=l06`3oool00`3ookl0M00007Boo`0M
0?ooo`030?oo_`1d0000M;oo00d0oooo00<0gia804RLg`3oool01P3oool00`3ookl0M00004RLg`0D
0?ooo`030?oo_`1d0000M;oo00d0oooo00<0oono07@00018W=l0503oool00`3ookl0M00007Boo`0<
0?ooo`030?oOW018M;l0oooo00L0oooo00<0oono07@00018W=l07@3oool00`3OW4P0B9cO0?ooo`09
0?ooo`00I@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00/0oooo00<0_g@007Bo
o`3oool01@3oool00`3o_g@0000007Boo`0B0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l02P3oool00`3o_g@0000007Boo`0A0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l02P3oool00`2oM000M;oo0?ooo`060?ooo`030?noM0000000M;oo01`0oooo00<0_g@0001d
_`3oool07@3oool00`2oM00007Bo0?ooo`0=0?ooo`030;md001d_ol0oooo00D0oooo00<0okmd0000
001d_ol05@3oool00`2oM00007Bo0?ooo`0<0?ooo`030?noM0000000M;oo01D0oooo00<0_g@0001d
_`3oool0303oool00`2oM000M;oo0?ooo`060?ooo`030?noM0000000M;oo01h0oooo00<0_g@007Bo
o`3oool0203oool006D0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`0:0?ooo`03
0?oOW01807@0_ooo0080oooo00H0omoO0=ooo`3o_g@000000000M02ooolB0?ooo`030?oOW0180000
B9cO00<0oooo00<0_g@00018W03Oool01P3oool01P3ogml0gooo0?noM0000000001d0;oooa40oooo
00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`090?ooo`030?oOW018B9`0gooo00<0oooo
00H0omoO0=ooo`3o_g@000000000M02ooolL0?ooo`030=nLB000B9`0gooo01d0oooo00<0gia80018
W03Oool03@3oool00`3ogi`0B01d0;ooo`020?ooo`060?oOg`3Oool0okmd0000000007@0_ooo5@3o
ool00`3OW4P004RL0=ooo`090?ooo`060?oOg`3Oool0okmd0000000007@0_ooo5@3oool00`3OW4P0
04RL0=ooo`0<0?ooo`030?oOW018B9`0gooo00<0oooo00H0omoO0=ooo`3o_g@000000000M02ooolN
0?ooo`030?oOW01807@0_ooo00P0oooo001U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia8
0018W03Oool02P3oool00`3ooml0W4P007Boo`020?ooo`050?oo_`1d000000000000B02Lgol04`3o
ool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000
001809cOo`0B0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3O
W4P0B9cO0?ooo`020?ooo`050?oo_`1d000000000000B02Lgol07@3oool00`3ogi`0B01809cOo`0M
0?ooo`030?oOW01804P0W=oo00d0oooo00<0oooO09a8001d_ol00P3oool01@3ookl0M00000000000
04P0W=oo01H0oooo00<0omnL04P0B02Lgol02@3oool01@3ookl0M0000000000004P0W=oo01H0oooo
00<0omnL04P0B02Lgol03@3oool00`3OW4P0B9cO0?ooo`020?ooo`050?oo_`1d000000000000B02L
gol07`3oool00`3ooml0W4P007Boo`080?ooo`00IP3oool00`3ogi`0B00007Boo`020?ooo`030=nL
B000M;l0oooo00X0oooo00<0oono07@00018W=l00`3oool00`3OW9`0_ooo0?ooo`0E0?ooo`030?oO
W0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?ooo`0D0?ooo`030?oO
W0180000M;oo0080oooo00<0gia8001d_`3oool01@3oool00`3ooml0gmoo0?ooo`020?ooo`030?oo
_`1d07@0_ooo00<0oooo00<0gibL0;ooo`3oool07P3oool00`3ookl0M00007Boo`0M0?ooo`030?oo
_`1d0000M;oo00d0oooo00<0oono07@00018W=l00`3oool00`3OW9`0_ooo0?ooo`0G0?ooo`030?oo
_`1d0000M;oo00X0oooo00<0gibL0;ooo`3oool05`3oool00`3ookl0M00007Boo`080?ooo`030?oo
g`3Ogol0oooo0080oooo00<0oono07@0M02oool00`3oool00`3OW9`0_ooo0?ooo`0P0?ooo`030?oo
_`1d0000B9cO00P0oooo001W0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool01`3o_g@0
04RL0=ooo`3oool0gia80000B02Lgol00`3oool00`3ogi`0M9cO0?ooo`0F0?ooo`060=nLB00007@0
_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0E0?ooo`060=nLB00007@0_ooo0?oog`2L
B000B9cO1`3oool00`2oM7@0_ooo0?ooo`020?ooo`030?noM000M;l0oooo0080oooo00<0omnL07BL
g`3oool07`3oool00`2oM00007Bo0?ooo`0M0?ooo`030;md0000M;l0oooo00P0oooo00L0okmd0018
W03Oool0oooo0=nLB00004P0W=oo00<0oooo00<0omnL07BLg`3oool0603oool00`2oM00007Bo0?oo
o`090?ooo`030?oOW01dW=l0oooo01P0oooo00<0_g@0001d_`3oool0203oool00`2oM7@0_ooo0?oo
o`020?ooo`030?noM000M;l0oooo0080oooo00<0omnL07BLg`3oool0703oool01`3o_g@004RL0=oo
o`3oool0gia80000B02Lgol0203oool006P0oooo00@0gia800000000000007Bo2@3oool01@3ogi`0
B000000000000000M;oo00D0oooo00D0_g@0000000000000000004RLg`0D0?ooo`040=nLB0000000
0000001d_`X0oooo00D0_g@0000000000000000004RLg`0C0?ooo`040=nLB00000000000001d_`P0
oooo00<0oono07@0000000000`0000000`0004P0W=oo0?ooo`020?ooo`050;md0000000000000000
0018W=l06P3oool01@3ookl0M0000000000007@0_ooo01/0oooo00D0oono07@000000000001d0;oo
o`090?ooo`050?oOW018000000000000001d_ol01@3oool01@2oM0000000000000000000B9cO01<0
oooo00D0oono07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l04`3oool0
1@3ookl0M0000000000007@0_ooo00P0oooo00<0oono07@0000000000`0000000`0004P0W=oo0?oo
o`020?ooo`050;md00000000000000000018W=l06P3oool01@3ogi`0B000000000000000M;oo00T0
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40
oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo00180?ooool0
0000G@00000<0?ooo`00B03oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00X0oooo
00180?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?oo
o`0X0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`0X0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?oo
o`0:0?ooo`00B03oool00`000000oooo0?ooo`3o0?ooobh0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`3o0?ooobh0oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00B03oool200000?l0oooo;`3oool00`000000
oooo0?ooo`0C0?ooo`030000003oool0oooo0140oooo0P00000<0?ooo`00B03oool00`000000oooo
0?ooo`3o0?ooobh0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0B0?ooo`030000
003oool0oooo00X0oooo00180?ooo`030000003oool0oooo0?l0oooo;P3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool02P3oool004P0oooo00<00000
0?ooo`3oool0o`3oool^0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`3o0?ooobh0oooo00<000000?oo
o`3oool04`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo00180?ooo`03
0000003oool0oooo0?l0oooo;P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0180
oooo00<0<ol00?ooo`3oool02P3oool004P0oooo0P00003o0?ooo`030000003oool0oooo02`0oooo
00@000000?ooo`3oool000004P3oool00`000000oooo0?ooo`0@0?ooo`80<ol000<000000?ooo`3o
ool02P3oool004P0oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0/0?ooo`040000
003oool0oooo00000180oooo00<000000?ooo`3oool03P3oool203?o0080oooo00<000000?ooo`3o
ool02P3oool004P0oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0/0?ooo`040000
003oool0oooo00000180oooo00<000000?ooo`3oool0303oool203?o00@0oooo00<000000?ooo`3o
ool02P3oool004P0oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0/0?ooo`040000
003oool0oooo00000180oooo00<000000?ooo`3oool02P3oool203?o00H0oooo00<000000?ooo`3o
ool02P3oool004P0oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0/0?ooo`040000
003oool0oooo00000180oooo00<000000?ooo`3oool0203oool203?o00P0oooo00<000000?ooo`3o
ool02P3oool004P0oooo0P00003o0?ooo`030000003oool0oooo02`0oooo00@000000?ooo`3oool0
00004P3oool00`000000oooo0?ooo`060?ooo`80<ol02@3oool2000000`0oooo00180?ooo`030000
003oool0oooo0?h0oooo00<000000?ooo`3oool0;03oool010000000oooo0?ooo`00000A0?ooo`<0
00001@3oool203?o00`0oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0
oP3oool00`000000oooo0?ooo`0/0?ooo`040000003oool0oooo00000100oooo1@0000020?ooo`80
<ol03P3oool00`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`3n0?ooo`030000
003oool0oooo02`0oooo00@000000?ooo`3oool00000403oool500000080<ol0403oool00`000000
oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`3n0?ooo`030000003oool0oooo02`0oooo
00@000000?ooo`3oool00000403oool500000180oooo00<000000?ooo`3oool02P3oool004P0oooo
0P00003o0?ooo`030000003oool0oooo02`0oooo00@000000?ooo`3oool000004@3oool300000180
oooo0P00000<0?ooo`008`3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo00<0
oooo00H0oono07@00000000000000018W03Oool20?ooo`060?oOW0180000000000000000B9`0gooo
2P3oool00`000000oooo0?ooo`3n0?ooo`030000003oool0oooo02`0oooo00@000000?ooo`3oool0
00003`3oool203?o00030?ooo`000000oooo01<0oooo00<000000?ooo`3oool02P3oool00280oooo
00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080oooo00<0oono07@0
0018W=l00P3oool02P3OW4P0001d0;ooo`3oool0okmd0000001d_ol0oooo0?oOW018M;l:0?ooo`03
0000003oool0oooo0?h0oooo00<000000?ooo`3oool0;03oool00`000000oooo00000002000000`0
oooo0P0co`030?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool02P3oool00240oooo
00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`060?ooo`030?noM000B9`0gooo00<0oooo
00<0_g@004RLg`3oool01@3oool00`3o_g@0B9cO0?ooo`080?ooo`030000003oool0oooo0?h0oooo
00<000000?ooo`3oool0;03oool6000000T0oooo0P0co`050?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool02P3oool00240oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018
W03Oool50?ooo`030?noM000B9`0gooo00<0oooo00<0_g@0001d_`3oool01@3oool00`3ookl0M01d
0;ooo`080?ooo`030000003oool0oooo0?d0oooo0`00000]0?ooo`H000001`3oool203?o00L0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0:0?ooo`008@3oool00`3ooml0W4P0001d
_`030?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a8001d_ol00P3oool00`3ookl0M00004RL
g`060?ooo`030?oog`2LB000M;oo00P0oooo00<000000?ooo`3oool0o03oool5000002`0oooo1P00
00050?ooo`80<ol02@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00X0oooo000R
0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01`3oool01P2oM000001809cOg`2L
B000001d0;ooo`H0oooo00<0oono07@00018W=l0203oool300000?`0oooo1@00000/0?ooo`030000
003oool0000000800000103oool203?o00/0oooo00<000000?ooo`3oool0403oool3000000`0oooo
000R0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0203oool01@3ooml0W4P00000
0000B9`0gooo00H0oooo00<0okmd0000001d_ol0203oool00`000000oooo0?ooo`3l0?ooo`D00000
:`3oool3000000030?ooo`000000oooo0080oooo0P0co`0=0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool02P3oool00280oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`070?ooo`060;md000004P0W=oo0?oOW018B9`0gooo0P3oool01P3ogml0gooo0?noM0000000001d
0;ooo`P0oooo00<000000?ooo`3oool0o@3oool3000002/0oooo1P0000000`3oool0<ol003?o000?
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool02P3oool00280oooo00@0oooO09a8
0000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oOW0180000M;oo0080oooo00T0okmd
04RLg`3oool0oooo0?oo_`1d000000000000B02Lgol02@3oool00`000000oooo0?ooo`3n0?ooo`03
0000003oool0oooo02X0oooo1P0000000`0co`00oooo0?ooo`0?0?ooo`030000003oool0oooo0180
oooo00<000000?ooo`3oool02P3oool002<0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo
0?ooo`060?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03Oool00P3oool00`3OW9`0_ooo
0?ooo`0:0?ooo`030000003oool0oooo0?h0oooo00<000000?ooo`3oool0:P3oool600000180oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0:0?ooo`00903oool01P3OW4P0001d0;oo
o`3ooml0W4P004RLg`L0oooo00<0oooO09a80018W=l00P3oool00`2oM00007Bo0?ooo`020?ooo`03
0?oOW01dW=l0oooo00X0oooo0P00003o0?ooo`030000003oool0oooo02/0oooo0`0000000`3oool0
00000?ooo`0A0?ooo`030000003oool0oooo0140oooo0P00000<0?ooo`009@3oool0103OW4P00000
00000000M;l90?ooo`050?oog`2LB000000000000000M;l01@3oool01@2oM0000000000000000000
B9cO00L0oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0Y0?ooo`80<ol000D0oooo
0000003oool0oooo0000000W0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003oool0
oooo0?h0oooo00<000000?ooo`3oool09`3oool203?o00<0oooo00@000000?ooo`3oool000009`3o
ool00`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`3n0?ooo`030000003oool0
oooo02D0oooo0P0co`050?ooo`040000003oool0oooo000002L0oooo00<000000?ooo`3oool02P3o
ool004P0oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0R0?ooo`<0<ol01`3oool0
10000000oooo0?ooo`00000W0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003oool0
oooo0?h0oooo00<000000?ooo`3oool0803oool203?o00X0oooo00@000000?ooo`3oool000009`3o
ool00`000000oooo0?ooo`0:0?ooo`00B03oool200000?l0oooo00<000000?ooo`3oool07P3oool2
03?o00`0oooo00@000000?ooo`3oool000009P3oool2000000`0oooo00180?ooo`030000003oool0
oooo0?h0oooo00<000000?ooo`3oool0703oool203?o00h0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`3n0?ooo`030000003oool0oooo
01X0oooo0P0co`0@0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool02P3oool004P0
oooo00<000000?ooo`3oool0oP3oool00`000000oooo0?ooo`0G0?ooo`<0<ol04P3oool00`000000
oooo0?ooo`0X0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003oool0oooo0?h0oooo
00<000000?ooo`3oool05@3oool203?o01D0oooo00<000000?ooo`3oool0:03oool00`000000oooo
0?ooo`0:0?ooo`00B03oool200000?l0oooo00<000000?ooo`3oool04`3oool203?o01L0oooo00<0
00000?ooo`3oool09`3oool2000000`0oooo00180?ooo`030000003oool0oooo0?l0oooo4`3oool2
03?o01T0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00B03oool00`00
0000oooo0?ooo`2l0?ooo`030000003oool0oooo0500oooo0`0co`0K0?ooo`030000003oool0oooo
02P0oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0?`3oool00`000000
oooo0?ooo`1j0?ooo`030000003oool0oooo04h0oooo0P0co`0N0?ooo`030000003oool0oooo02P0
oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo
0?ooo`1j0?ooo`030000003oool0oooo04`0oooo0P0co`0P0?ooo`030000003oool0oooo02P0oooo
00<000000?ooo`3oool02P3oool004P0oooo0P0000100?ooo`030000003oool0oooo07X0oooo00<0
00000?ooo`3oool0BP3oool203?o0280oooo00<000000?ooo`3oool09`3oool2000000`0oooo000/
0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool0103OW4P000000000
M02oool<0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0NP3oool00`000000oooo
0?ooo`180?ooo`80<ol0903oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00X0oooo
000[0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`040?ooo`03
0?oo_`1d000007Bo00/0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`1j0?ooo`03
0000003oool0oooo04D0oooo0`0co`0V0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3o
ool02P3oool002X0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0:0?ooo`030;md
000007@0_ooo00T0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`1j0?ooo`030000
003oool0oooo04<0oooo0P0co`0Y0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
2P3oool002X0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool90?ooo`040?oo
g`2LB000001d0;ooo`P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo06l0oooo00<000000?ooo`3oool0@@3oool203?o02/0oooo00<000000?ooo`3o
ool0:03oool00`000000oooo0?ooo`0:0?ooo`00:P3oool00`3ooml0W4P0001d_`030?ooo`030=nL
B00004P0W=oo00L0oooo00<0omnL04P0000000000P0000000`00M;l0oooo0?ooo`060?ooo`030000
003oool0oooo03l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`1_0?ooo`030000
003oool0oooo03l0oooo0P0co`0]0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
2P3oool002/0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`060?ooo`080?oo_`1d
000004RL0=ooo`3oool0_g@00000B02Lgol70?ooo`<00000?`3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo06l0oooo00<000000?ooo`3oool0?@3oool203?o02l0oooo00<000000?oo
o`3oool09P3oool3000000`0oooo000[0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000
M;l01P3oool00`3ogi`0B00007Boo`020?ooo`030?oOW0180000B9cO00L0oooo00<000000?ooo`3o
ool0?`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3o
ool0>P3oool303?o0340oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00
:`3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00D0oooo00<0omnL04P0001d_ol0
0P3oool00`3ooml0W4P0001d_`070?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo03P0oooo0P0co`0d0?ooo`030000
003oool0oooo02P0oooo00<000000?ooo`3oool02P3oool002/0oooo00@0oooO09a80000B9`0gooo
0P3oool00`3OW4P004RL0=ooo`050?ooo`030?oo_`1d000007Bo00<0oooo00<0gia80018W03Oool0
1P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
K`3oool00`000000oooo0?ooo`0f0?ooo`80<ol0=P3oool00`000000oooo0?ooo`0X0?ooo`030000
003oool0oooo00X0oooo000/0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3o
ool00`2oM000001d0;ooo`020?ooo`030=nLB000M;l0oooo00H0oooo00<000000?ooo`3oool0?`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool0=03o
ool203?o03P0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00;@3oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00L0oono07@00000B9`0gooo0?oog`2LB000M;oo
00L0oooo0P0000100?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0K`3oool00`00
0000oooo0?ooo`0b0?ooo`80<ol0>P3oool00`000000oooo0?ooo`0W0?ooo`800000303oool002h0
oooo00@0gia800000000000007Bo2P3oool0102oM000000000000018W=l80?ooo`030000003oool0
oooo03l0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`1^0?ooo`<00000<03oool3
03?o03`0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00B03oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0K@3oool50000
02d0oooo0P0co`0o0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool02P3oool004P0
oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06d0
oooo1@00000;0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool00P0co`110?ooo`03
0000003oool0oooo02P0oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0
?`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06d0oooo1@00000;0?ooo`030000
003oool0oooo01X0oooo00<0000003?o000co`00@`3oool00`000000oooo0?ooo`0X0?ooo`030000
003oool0oooo00X0oooo00180?ooo`800000@03oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo06h0oooo0`00000<0?ooo`030000003oool0oooo01P0oooo0P0co`000`000000oooo0?oo
o`130?ooo`030000003oool0oooo02L0oooo0P00000<0?ooo`00B03oool00`000000oooo0?ooo`0o
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo01H0oooo0P0co`020?ooo`030000003oool0oooo04<0oooo00<00000
0?ooo`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`0o0?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo01@0oooo0P0co`040?ooo`030000003oool0oooo04<0oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`0o0?ooo`03
0000003oool0oooo00L0oooo0`00001`0?ooo`030000003oool0oooo00X0oooo0`00000C0?ooo`80
<ol01P3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3o
ool02P3oool004P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`060?ooo`D00000
=`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00T0oooo1@00000@0?ooo`80<ol0
203oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
2P3oool004P0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`060?ooo`D00000=`3o
ool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00T0oooo1@00000=0?ooo`<0<ol02P3o
ool00`000000oooo0?ooo`130?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool02P3o
ool004P0oooo0P0000100?ooo`030000003oool0oooo00H0oooo1@00000Q0?ooo`030000003oool0
oooo01<0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`090?ooo`D000002P3oool3
03?o00d0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0W0?ooo`800000303oool0
04P0oooo00<000000?ooo`3oool0?P3oool3000000P0oooo0`00000R0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0:0?ooo`<000001`3oool403?o
0100oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo
00X0oooo00180?ooo`030000003oool0oooo03d0oooo1@0000080?ooo`030000003oool0oooo0240
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool00`3oool303?o01@0oooo00<000000?ooo`3oool0KP3oool00`000000
oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`0m0?ooo`D00000203oool00`000000oooo
0?ooo`0Q0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0=@3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00<0<ol05`3oool00`000000oooo0?ooo`1^0?ooo`030000
003oool0oooo00X0oooo00180?ooo`030000003oool0oooo03d0oooo1@0000080?ooo`030000003o
ool0oooo0240oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0e0?ooo`030000003o
ool0oooo00/0oooo00<0000003?o000co`006P3oool00`000000oooo0?ooo`1^0?ooo`030000003o
ool0oooo00X0oooo000S0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3o
ool0103OW4P000000000M02oool50?ooo`060?oOW0180000000000000000B9`0gooo2P3oool20000
03l0oooo0`0000090?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool04`3oool00`00
0000oooo0?ooo`0e0?ooo`030000003oool0oooo00L0oooo100co`000`000000oooo0?ooo`0J0?oo
o`030000003oool0oooo06d0oooo0P00000<0?ooo`008P3oool01P3OW4P007Bo0?ooo`3ogi`0B000
04RLg`80oooo00<0omnL04P0001d_ol0103oool00`3ookl0M000001d_`040?ooo`060?noM0000000
M;oo0?ooo`3ogi`0B7Bo2P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo03D0oooo
00<000000?ooo`3oool0103oool303?o00@0oooo00<000000?ooo`3oool06P3oool00`000000oooo
0?ooo`1^0?ooo`030000003oool0oooo00X0oooo000Q0?ooo`030?oog`2LB000B9cO0080oooo00<0
okmd0000B02Lgol02P3oool00`2oM000001d0;ooo`070?ooo`030?noM018W=l0oooo00P0oooo00<0
00000?ooo`3oool0?`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool04`3oool00`000000oooo0?ooo`0e0?ooo`040000003oool0oooo0?ooo`<0<ol0
903oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo00X0oooo000Q0?ooo`030?oog`2L
B000B9cO0080oooo00@0oooO09a80000B9`0gooo2@3oool0103ooml0W4P00000M02oool60?ooo`03
0?oo_`1d07@0_ooo00P0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo03D0oooo00@0000003?o000co`00<ol09`3oool00`000000oooo0?ooo`1^0?oo
o`030000003oool0oooo00X0oooo000Q0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol01`3oool00`3ogi`0B0000000000200000003001d_`3oool0oooo00@0oooo00<0oooO09a8001d
_ol0203oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3o
ool0<P3oool303?o00030000003oool0oooo02P0oooo00<000000?ooo`3oool0KP3oool00`000000
oooo0?ooo`0:0?ooo`008P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00H0oooo
00P0oono07@00000B9`0gooo0?ooo`2oM000001809cOo`D0oooo00<0oono07@00018W=l0203oool3
000002`0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo02l0oooo0`0co`030?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0K03oool3
000000`0oooo000=0?ooo`030=nLB000000000000080000000<0B9cO0?ooo`3oool03@3oool00`3O
W4P0001809cOo`020?ooo`030?oog`2LB00007Bo00H0oooo00<0omnL04P0001d_ol00P3oool00`3o
gi`0B00004RLg`050?ooo`030?noM0000000M;oo00P0oooo00<000000?ooo`3oool0;03oool00`00
0000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0Q0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0;03oool303?o
0340oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0:0?ooo`002`3oool00`3o_g@0
000004RLg`040?ooo`030=nLB000M;l0oooo00d0oooo00<0omnL04P00018W=l00`3oool00`2oM000
04RL0=ooo`050?ooo`030?oOW0180000M;oo0080oooo00<0oooO09a80000M;l00P3oool01P3ogml0
gooo0?noM0000000001d0;ooo`P0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo0180oooo0`00000Y0?ooo`@0<ol0=03oool00`000000oooo0?ooo`1^
0?ooo`030000003oool0oooo00X0oooo000:0?ooo`030=nLB000M;l0oooo00H0oooo00<0okmd07Bo
o`3oool0303oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00D0oooo00<0oono
07@00000M;l00`3oool02@3OW4P004RL0=ooo`3oool0oono07@000000000001809cOo`090?ooo`03
0000003oool0oooo02/0oooo0`0000070?ooo`<000001`3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo0240oooo00<000000?ooo`3oool04@3oool5000002@0oooo100co`0h0?ooo`03
0000003oool0oooo06h0oooo00<000000?ooo`3oool02P3oool000T0oooo00<0okmd04RLg`3oool0
203oool00`2oW9`0gooo0?ooo`0<0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0
1P3oool00`2oM000001d0;ooo`020?ooo`030=nLB000M;l0oooo0080oooo00<0gibL0;ooo`3oool0
2P3oool00`000000oooo0?ooo`0Z0?ooo`D000001@3oool5000000D0oooo0P00000:0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0A0?ooo`D00000
803oool403?o03/0oooo0`00001_0?ooo`030000003oool0oooo00X0oooo00080?ooo`030?oog`2L
W;l0oooo00T0oooo00<0gibL0;ooo`3oool03@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0
oooo00L0oono07@00000B9`0gooo0?oog`2LB000M;oo00<0oooo00<0omnL07BLg`3oool02P3oool0
0`000000oooo0?ooo`0Z0?ooo`D000001@3oool5000000D0oooo0P00000:0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0A0?ooo`D00000703oool4
03?o03h0oooo1@00001^0?ooo`030000003oool0oooo00X0oooo00080?ooo`030?oOW02L_ol0oooo
00T0oooo00<0gibL0;ooo`3oool03P3oool0103OW4P0000000000000M;l:0?ooo`040;md00000000
000004RLg`D0oooo00D0_g@0000000000000000004RLg`070?ooo`8000005@3oool00`000000oooo
0?ooo`0C0?ooo`D000001@3oool5000000D0oooo0P00000B0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0B0?ooo`<000006@3oool403?o0480oooo
1@00001]0?ooo`800000303oool000P0oooo00<0oknL09cOo`3oool02@3oool00`2oM7@0_ooo0?oo
o`0a0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0503oool3000000L0oooo0`00
00060?ooo`8000004P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0000001@0oooo100c
o`160?ooo`D00000KP3oool00`000000oooo0?ooo`0:0?ooo`00203oool00`3OW9`0_ooo0?ooo`03
0?ooo`030?noM01d_ol0oooo0080oooo00<0oooO09adW03Oool0<P3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`05
0?ooo`8000004P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool000000100oooo100co`1;
0?ooo`<00000K`3oool00`000000oooo0?ooo`0:0?ooo`00203oool00`3o_g@0M;oo0?ooo`030?oo
o`040?noM018W=l0oooO09a80080000000<0B9cO0?ooo`3oool0<@3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo00D0oooo0P00000B0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01<0oooo00<00000
0?ooo`000000303oool403?o0500oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0:
0?ooo`00203oool00`3ookl0M01d0;ooo`030?ooo`060?noM000000000000000000007@0_ooo=03o
ool00`000000oooo0?ooo`0C0?ooo`<00000203oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01@3oool200000180oooo00<000000?ooo`000000103o
ool00`000000oooo0?ooo`050?ooo`<000002@3oool3000001@0oooo00<000000?ooo`0000001@3o
ool703?o05@0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0:0?ooo`00203oool0
2@3ookl0M000000000000000M;oo0?ooo`3o_g@004RL0=ooo`0g0?ooo`8000004`3oool5000000H0
oooo0`00000<0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01@3oool2000000<0
oooo00<000000?ooo`3oool0303oool00`000000oooo000000040?ooo`030000003oool0oooo00@0
oooo1@0000070?ooo`D000004@3oool203?o00030000000co`00000000D0<ol0F`3oool00`000000
oooo0?ooo`1]0?ooo`800000303oool000P0oooo00<0gia80000001d_ol00`3oool00`3o_g@0M;oo
0?ooo`0g0?ooo`030000003oool0oooo0180oooo1@0000030?ooo`80<ol01@00000;03?o00030000
000co`00<ol000`0<ol00`3oool2000000<0oooo00<000000?ooo`3oool02`3oool4000000@0oooo
00<000000?ooo`3oool0103oool5000000L0oooo1@0000090?ooo`P0<ol00P3oool4000005l0oooo
00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0:0?ooo`005`3oool00`2oM7@0_ooo0?oo
o`0^0?ooo`030000003oool0oooo0180oooo1@00000303?o0080oooo1@00000J0?ooo`<0<ol00P00
000303?o00030000000co`00<ol000X0<ol01@0000040?ooo`030000003oool0oooo00@0oooo1@00
00070?ooo`D000002@0co`0:0?ooo`D00000GP3oool00`000000oooo0?ooo`1^0?ooo`030000003o
ool0oooo00X0oooo000D0?ooo`050?noM000000000000000M02oool0;`3oool00`000000oooo0?oo
o`040?ooo`8000002`3oool203?o00<000001P3oool5000001d0oooo0P0000030?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool00P3oool5000000@0<ol000<0000003?o000co`001@0c
o`03000000T0<ol00`00000D0?ooo`D00000GP3oool00`000000oooo0?ooo`1^0?ooo`030000003o
ool0oooo00X0oooo000?0?ooo`030=nLB000000000000080000000D007Bo0?ooo`3oool0gibL0;oo
o`0_0?ooo`030000003oool0oooo00@0oooo100000070?ooo`80<ol00`3oool00`000000oooo0?oo
o`060?ooo`<000007P3oool2000000<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?oo
o`020?ooo`D00000103oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool04`3oool5000005h0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?oo
o`0:0?ooo`003`3oool01P3OW4P0M;oo0?ooo`3ooml0W7BL0=ooo`80oooo00<0gibL0;ooo`3oool0
;P3oool00`000000oooo0?ooo`030?ooo`H000000`3oool303?o00D0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0M0?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00<0oooo100000040?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`0C0?ooo`@00000G`3oool00`000000oooo0?ooo`1^0?ooo`030000
003oool0oooo00X0oooo000?0?ooo`0:0=nLW02oool0oooo0?oog`2LB000000007Boo`3ooml0W4Qd
0;ooobl0oooo0P0000030?ooo`L0000000<0oooo03?o000co`00203oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo01`0oooo0`0000030?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0103oool00`000000oooo000000040?ooo`030000003oool0oooo00H0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0C0?ooo`030000003oool000000600oooo00<00000
0?ooo`3oool0K@3oool2000000`0oooo000?0?ooo`050=nLW02oool0oono07@00000M;l00P3oool0
0`2oM000001d0;ooo`0_0?ooo`030000003oool0oooo0080oooo1`0000000`0co`00oooo0?ooo`08
0?ooo`030000003oool0oooo02D0oooo1P0000000`3oool000000?ooo`060?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`000000103oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool04`3oool00`000000oooo0000001P0?ooo`030000003oool0
oooo06h0oooo00<000000?ooo`3oool02P3oool000l0oooo00<0gibL0;ooo`3oool00`3oool00`3O
W4P004RL0=ooo`020?ooo`030;nLW03Oool0oooo02/0oooo00@000000?ooo`3oool0oooo1`000000
0`0co`00oooo0?ooo`0a0?ooo`H0000000<0oooo0000003oool01P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0000000@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`000000H03oool00`000000oooo0?ooo`1^
0?ooo`030000003oool0oooo00X0oooo000?0?ooo`030=nLB0000000B9cO00L0oooo00<0omnL04Q8
W03Oool0;03oool00`000000oooo0?ooo`060000000303?o003oool0oooo03<0oooo1P0000000`3o
ool000000?ooo`060?ooo`030000003oool0oooo00@0oooo00<000000?ooo`000000103oool00`00
0000oooo0?ooo`0B0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0GP3oool00`00
0000oooo0?ooo`1^0?ooo`030000003oool0oooo00X0oooo000B0?ooo`040;nLW03Oool0oooo0?no
M0@0000000<004RL0=ooo`3oool0:`3oool00`000000oooo0?ooo`06000003L0oooo0`000000103o
ool000000?ooo`0000070?ooo`030000003oool0oooo00@0oooo00<000000?ooo`000000103oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0GP3oool0
0`000000oooo0?ooo`1^0?ooo`030000003oool0oooo00X0oooo000B0?ooo`070;md000000000000
0018W03Oool0_gAd0;ooo`0O0?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool0203o
ool9000003P0oooo00<000000?ooo`0000000`0000070?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`000000103oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo07H0oooo00<0
00000?ooo`3oool0K@3oool2000000`0oooo000B0?ooo`030?oo_`1dM;l0oooo0080oooo00<0gibL
0;ooo`3oool0803oool00`3o_g@0000007Boo`0;0?ooo`T00000>03oool6000000L0oooo00<00000
0?ooo`3oool0103oool00`000000oooo000000040?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0MP3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo00X0oooo000B0?oo
o`060=nLW02oool0oooo0?ooo`3ookl0M7Bo8P3oool00`3ooml0W4P0001d_`0;0?ooo`T00000>03o
ool6000000L0oooo00<000000?ooo`3oool01P3oool010000000oooo0?ooo`3oool3000001<0oooo
00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo00X0oooo
000A0?ooo`070?oog`2LM9`0gooo0?ooo`3o_g@0001d0;ooo`0S0?ooo`030=nLB000B9`0gooo00X0
oooo2@00000h0?ooo`L000001P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00D0
00004P3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3o
ool02P3oool00180oooo00@0gia800000000000007Bo9@3oool00`3ogi`0B01809cOo`0:0?ooo`P0
0000>@3oool00`000000oooo00000005000000@0oooo0`0000070?ooo`030000003oool0oooo00D0
00004P3oool00`000000oooo0?ooo`1f0?ooo`030000003oool0oooo06h0oooo00<000000?ooo`3o
ool02P3oool003/0oooo00<0oono07@0001d_ol02P3oool5000003`0oooo00<000000?ooo`3oool0
1@0000030?ooo`D000001P3oool00`000000oooo0?ooo`0500000180oooo00<000000?ooo`3oool0
MP3oool00`000000oooo0?ooo`1/0?ooo`<00000303oool001H0oooo00@0oooO09a8000007@0_ooo
8P3oool00`2oM00007Bo0?ooo`090?ooo`D00000?03oool00`000000oooo0?ooo`05000000<0oooo
1@0000060?ooo`040000003oool0oooo0?ooo`<000004`3oool00`000000oooo0?ooo`1f0?ooo`03
0000003oool0oooo06h0oooo00<000000?ooo`3oool02P3oool001<0oooo00D0okmd000000000000
001809cOo`0T0?ooo`030=nLB000B9`0gooo00T0oooo1000000m0?ooo`030000003oool0oooo00@0
0000103oool5000000H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo07H0oooo00<000000?ooo`3oool0KP3oool00`000000oooo0?ooo`0:0?ooo`00
403oool0102oM000000000000000M;lX0?ooo`030?oOW01804P0W=oo00T0oooo0`00000n0?ooo`06
0000003oool0oooo0000003oool000001P3oool3000000L0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool02P3oool0
00`0oooo00D0omnL04P000000000000007Boo`0[0?ooo`030?oo_`1d0000M;oo00T0oooo00<00000
0?ooo`3oool0?P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0i`3oool00`000000oooo0?ooo`0:0?ooo`002@3oool01@3OW4P0000000000000B9`0
gooo02l0oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`0n0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`060?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`3W0?ooo`030000003oool0
oooo00X0oooo00070?ooo`030?oOW0180000B9cO0340oooo00D0oono07@000000000001d0;ooo`08
0?ooo`800000?`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool0iP3oool2000000`0oooo00180?ooo`030000003oool0oooo03h0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo0`0000030?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`0:0?oo
o`003@3oool00`3OW4P00000000000020000000304RLg`3oool0oooo03<0oooo00<000000?ooo`3o
ool0@`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo1@0000020?ooo`03
0000003oool0oooo0180oooo00<000000?ooo`3oool0i`3oool00`000000oooo0?ooo`0:0?ooo`00
2`3oool00`3o_g@0000004RLg`040?ooo`030=nLB000M;l0oooo03<0oooo00<000000?ooo`3oool0
@`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo1@0000020?ooo`030000
003oool0oooo04@0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0:0?ooo`002P3o
ool00`3OW4P007Bo0?ooo`060?ooo`030?noM01d_ol0oooo0380oooo00<000000?ooo`3oool0@`3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo1@0000020?ooo`030000003o
ool0oooo04@0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0:0?ooo`002@3oool0
0`3o_g@0B9cO0?ooo`080?ooo`030;nLW03Oool0oooo0340oooo0P0000140?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool01@3oool3000000<0oooo00<000000?ooo`3oool0A03oool0
0`000000oooo0?ooo`2d0?ooo`800000303oool000P0oooo00<0oooO09bL_`3oool02@3oool00`3O
W9`0_ooo0?ooo`0a0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0A03oool00`00
0000oooo0?ooo`2e0?ooo`030000003oool0oooo00X0oooo00080?ooo`030?oOW02L_ol0oooo00T0
oooo00<0gibL0;ooo`3oool0<@3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo04@0
oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0:0?ooo`00203oool00`3o_i`0W=oo
0?ooo`090?ooo`030;mdM02oool0oooo0340oooo00<000000?ooo`3oool0@`3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0]@3oool00`000000oooo
0?ooo`0:0?ooo`00203oool00`3OW9`0_ooo0?ooo`030?ooo`030?noM01d_ol0oooo0080oooo00<0
oooO09adW03Oool0<P3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo00h0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool0]@3oool00`000000oooo0?ooo`0:0?ooo`00203oool00`3o_g@0M;oo0?ooo`03
0?ooo`040?noM018W=l0oooO09a80080000000<0B9cO0?ooo`3oool0<@3oool2000005D0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool0]03oool2000000`0oooo00080?ooo`030?oo_`1d07@0_ooo00<0oooo00H0okmd
00000000000000000000M02ooold0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0
]@3oool00`000000oooo0?ooo`0:0?ooo`00203oool02@3ookl0M000000000000000M;oo0?ooo`3o
_g@004RL0=ooo`0g0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool08`3oool00`00
0000oooo0?ooo`0S0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool02P3oool000P0
oooo00<0gia80000001d_ol00`3oool00`3o_g@0M;oo0?ooo`060?ooo`030?oOg`3Oool0oooo02h0
oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0:0?ooo`005`3oool00`2oM7@0_ooo
0?ooo`0^0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`0S0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool02P3oool000l0oooo00H0
gibL0;ooo`3oool0oooo0?oog`2LB002000000030000M02oool0oooo02h0oooo00<000000?ooo`3o
ool0E03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3o
ool0]@3oool00`000000oooo0?ooo`0:0?ooo`003`3oool01P3OW4P00000000000000000001809cO
o`80oooo00<0gibL0;ooo`3oool01`3oool00`3ookl0M000000000020000000704RLg`3oool0oooo
0?ooo`3ogi`0B00007Boo`040?ooo`040;md00000000000007Boo`<0oooo00H0omnL04P000000000
00000018W03Oool:0?ooo`800000E@3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
02<0oooo00<000000?ooo`3oool0]03oool2000000`0oooo000?0?ooo`030=nLW02oool0oooo00D0
oooo00<0gibL0;ooo`3oool02@3oool00`3o_g@0000007Boo`040?ooo`030?oOW0180000M;oo00<0
oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
2P3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0
8`3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo00X0oooo000?0?ooo`030=nLW02o
ool0oooo00D0oooo00<0_ibL0=ooo`3oool02@3oool00`3ooml0W4P0001d_`090?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3o_g@0B9cO0?ooo`080?ooo`030000003o
ool0oooo05@0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo0;D0oooo00<000000?ooo`3oool02P3oool000l0oooo00<0gibL0;ooo`3oool0103oool0
0`3ookl0M7Bo0?ooo`0;0?ooo`030=nLB000B9`0gooo00P0oooo00<0oooO09a80018W=l00P3oool0
103ooml0W4P00018W03Oool50?ooo`030?oo_`1d07@0_ooo00P0oooo00<000000?ooo`3oool0E03o
ool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0]@3o
ool00`000000oooo0?ooo`0:0?ooo`003`3oool00`3o_g@0M;oo0?ooo`040?ooo`030;mdB02Lgol0
oooo00/0oooo00<0omnL04P0B02Lgol0203oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0
W=oo00D0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`1D0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0
oooo00X0oooo000?0?ooo`030?oo_`1d04P0W=oo0080oooo00<0oono07@0001d_ol03@3oool00`3o
okl0M00007Boo`090?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01@3oool00`3o
okl0M00004RLg`080?ooo`<00000E03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
02<0oooo00<000000?ooo`3oool0/`3oool3000000`0oooo000@0?ooo`060?oog`2LB00000000000
000007@0_ooo3`3oool00`2oM00007Bo0?ooo`080?ooo`030=nLB00004P0W=oo0080oooo00<0oooO
09a80000M;l01@3oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo05@0oooo00<00000
0?ooo`3oool08`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0;D0oooo00<00000
0?ooo`3oool02P3oool002D0oooo00<0gia80018W03Oool0203oool00`3ogi`0B00004RLg`030?oo
o`0:0;md0000B9`0gooo0?ooo`3ogml0gooo0?noM0000000001d0;ooo`P0oooo00<000000?ooo`3o
ool0E03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3o
ool0]@3oool00`000000oooo0?ooo`0:0?ooo`009@3oool00`3ogi`0B01809cOo`080?ooo`040?oo
g`2LB00004RL0=ooo`80oooo00T0gia80018W03Oool0oooo0?oo_`1d000000000000B02Lgol02@3o
ool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool08P3o
ool300000;H0oooo00<000000?ooo`3oool02P3oool002D0oooo00<0oono07@0001d_ol02@3oool0
0`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0080oooo00<0gibL0;ooo`3oool02P3oool0
0`000000oooo0?ooo`1D0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool08@3oool5
00000;D0oooo00<000000?ooo`3oool02P3oool002H0oooo00<0_g@0001d_`3oool02@3oool01P3O
W4P0001d0;ooo`3ooml0W4P004RLg`<0oooo00<0omnL07BLg`3oool02P3oool2000005D0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0Q0?ooo`D00000]03oool2000000`0oooo000T
0?ooo`050?oo_`1d000000000000M02oool02P3oool0103OW4P0000000000000M;l50?ooo`050;md
00000000000000000018W=l01`3oool00`000000oooo0?ooo`1D0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool08@3oool500000;D0oooo00<000000?ooo`3oool02P3oool004P0oooo
00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0280oooo
0`00002f0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003oool0oooo05@0oooo00<0
00000?ooo`3oool08`3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0;D0oooo00<0
00000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`0S
0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0:
0?ooo`00B03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo02<0oooo00<000000?oo
o`3oool0]@3oool00`000000oooo0?ooo`0:0?ooo`00B03oool2000007/0oooo00<000000?ooo`3o
ool08`3oool00`000000oooo0?ooo`2d0?ooo`800000303oool004P0oooo00<000000?ooo`3oool0
NP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0
2P3oool004P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo0;D0oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0NP3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool02P3oool0
04P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
0;D0oooo00<000000?ooo`3oool02P3oool004P0oooo0P00001k0?ooo`030000003oool0oooo02<0
oooo00<000000?ooo`3oool0]03oool2000000`0oooo00180?ooo`030000003oool0oooo07X0oooo
00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo00X0oooo
00180?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`2e0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003oool0oooo07X0oooo00<00000
0?ooo`3oool08`3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo00X0oooo00180?oo
o`030000003oool0oooo07X0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`2e0?oo
o`030000003oool0oooo00X0oooo00180?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3o
ool08`3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo00X0oooo000Z0?ooo`030?oo
_`1d000000000080000000L0B9cO0?ooo`3oool0oooo0?oOW0180000M;oo0080oooo00<0oono07@0
000000000P0000000`18W=l0oooo0?ooo`080?ooo`800000NP3oool3000002@0oooo00<000000?oo
o`3oool0]03oool2000000`0oooo000/0?ooo`030?noM0000000M;oo00@0oooo00<0omnL04P0001d
_ol0103oool00`3o_g@0000007Boo`0;0?ooo`030000003oool0oooo07P0oooo1@00000S0?ooo`03
0000003oool0oooo0;D0oooo00<000000?ooo`3oool02P3oool002`0oooo00<0oooO09a80000M;l0
2`3oool00`3ooml0W4P0001d_`0;0?ooo`030000003oool0oooo07P0oooo1@00000S0?ooo`030000
003oool0oooo0;D0oooo00<000000?ooo`3oool02P3oool002d0oooo00<0gia80018W03Oool02`3o
ool00`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo07P0oooo1@00000S0?ooo`030000003o
ool0oooo0;D0oooo00<000000?ooo`3oool02P3oool002d0oooo00<0omnL04P0B02Lgol02`3oool0
0`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo07T0oooo0`00000T0?ooo`030000003oool0
oooo0;D0oooo00<000000?ooo`3oool02P3oool002d0oooo00<0oono07@0001d_ol02`3oool00`3o
okl0M00007Boo`0:0?ooo`<00000NP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
0;<0oooo0`00000<0?ooo`00;P3oool00`2oM00007Bo0?ooo`0;0?ooo`030;md0000M;l0oooo00T0
oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo00X0
oooo000^0?ooo`030=nLB000B9`0gooo00/0oooo00<0gia80018W03Oool02@3oool00`000000oooo
0?ooo`1j0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool02P3oool002h0oooo00<0
omnL04P0B02Lgol02`3oool00`3ogi`0B01809cOo`090?ooo`030000003oool0oooo07X0oooo00<0
00000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`00;P3oool00`3ookl0M00007Boo`0;
0?ooo`030?oo_`1d0000M;oo00T0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3K
0?ooo`030000003oool0oooo00X0oooo000_0?ooo`030;md0000M;l0oooo00/0oooo00<0_g@0001d
_`3oool0203oool2000007/0oooo00<000000?ooo`3oool0fP3oool2000000`0oooo000]0?ooo`05
0?oo_`1d000000000000M02oool02@3oool01@3ookl0M0000000000007@0_ooo00P0oooo00<00000
0?ooo`3oool0NP3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo00X0oooo00180?oo
o`030000003oool0oooo07X0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?oo
o`00B03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3o
ool02P3oool004P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`3K0?ooo`030000
003oool0oooo00X0oooo00180?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0f`3o
ool00`000000oooo0?ooo`0:0?ooo`00B03oool2000007/0oooo00<000000?ooo`3oool0fP3oool2
000000`0oooo00180?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0f`3oool00`00
0000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo0=/0
oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0NP3oool00`000000oooo
0?ooo`3K0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003oool0oooo07X0oooo00<0
00000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`00B03oool2000007/0oooo00<00000
0?ooo`3oool0fP3oool2000000`0oooo00180?ooo`030000003oool0oooo07X0oooo00<000000?oo
o`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`00B03oool00`000000oooo0?ooo`1j0?ooo`03
0000003oool0oooo0=/0oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0
NP3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo00X0oooo00180?ooo`030000003o
ool0oooo07X0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`00B03oool2
000007/0oooo00<000000?ooo`3oool0fP3oool2000000`0oooo000Q0?ooo`030?oo_`1d00000000
0080000000L0B9cO0?ooo`3oool0oooo0?oOW0180000M;oo0080oooo00<0oono07@0000000000P00
00001P18W=l0oooo0?ooo`3oool0omnL04P00080000000<004RL0=ooo`3oool02@3oool00`000000
oooo0?ooo`1j0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool02P3oool002<0oooo
00<0okmd0000001d_ol0103oool00`3ogi`0B00007Boo`040?ooo`030?noM0000000M;oo00@0oooo
00H0okmd0000001d_ol0oooo0?oOW018M;l:0?ooo`030000003oool0oooo07X0oooo00<000000?oo
o`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`008`3oool00`3ooml0W4P0001d_`0;0?ooo`03
0?oog`2LB00007Bo00T0oooo00<0okmd04RLg`3oool0203oool00`000000oooo0?ooo`1j0?ooo`03
0000003oool0oooo0=/0oooo00<000000?ooo`3oool02P3oool002@0oooo00<0gia80018W03Oool0
2`3oool00`3OW4P004RL0=ooo`080?ooo`030?oo_`1d07@0_ooo00P0oooo00<000000?ooo`3oool0
NP3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo00X0oooo000T0?ooo`030?oOW018
04P0W=oo00/0oooo00<0omnL04P0B02Lgol0203oool00`3ooml0W4P007Boo`080?ooo`030000003o
ool0oooo07X0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`00903oool0
0`3ookl0M00007Boo`0;0?ooo`030?oo_`1d0000M;oo00P0oooo00<0oono07@00018W=l0203oool3
000007X0oooo00<000000?ooo`3oool0f@3oool3000000`0oooo000U0?ooo`030;md0000M;l0oooo
00/0oooo00<0_g@0001d_`3oool01`3oool00`3o_g@0000007Boo`080?ooo`030000003oool0oooo
07X0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`009@3oool00`3OW4P0
04RL0=ooo`0;0?ooo`030=nLB000B9`0gooo00@0oooo00H0omoO0=ooo`3o_g@000000000M02oool8
0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:
0?ooo`009@3oool00`3ogi`0B01809cOo`0;0?ooo`030?oOW01804P0W=oo00@0oooo00D0oono07@0
00000000001809cOo`090?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0f`3oool0
0`000000oooo0?ooo`0:0?ooo`009@3oool00`3ookl0M00007Boo`0;0?ooo`030?oo_`1d0000M;oo
00D0oooo00<0gibL0;ooo`3oool02P3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo
0=/0oooo00<000000?ooo`3oool02P3oool002H0oooo00<0_g@0001d_`3oool02`3oool00`2oM000
07Bo0?ooo`040?ooo`030?oOW01dW=l0oooo00X0oooo0P00001k0?ooo`030000003oool0oooo0=X0
oooo0P00000<0?ooo`00903oool01@3ookl0M0000000000007@0_ooo00T0oooo00D0oono07@00000
0000001d0;ooo`050?ooo`050;md00000000000000000018W=l01`3oool00`000000oooo0?ooo`1j
0?ooo`030000003oool0oooo0=/0oooo00<000000?ooo`3oool02P3oool004P0oooo00<000000?oo
o`3oool0NP3oool00`000000oooo0?ooo`3K0?ooo`030000003oool0oooo00X0oooo00180?ooo`03
0000003oool0oooo07X0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`0:0?ooo`00
B03oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool02P3oool004P0oooo0P00
003o0?oooeT0oooo0P00000<0?ooo`00B03oool00`000000oooo0?ooo`3o0?oooeT0oooo00<00000
0?ooo`3oool02P3oool004P0oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo
00X0oooo00180?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0:0?ooo`00
B03oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool02P3oool004P0oooo00<0
00000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo00X0oooo00180?ooo`800000o`3ooomI
0?ooo`800000303oool004P0oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo
00X0oooo00180?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0:0?ooo`00
B03oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool02P3oool004P0oooo00<0
00000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo00X0oooo00180?ooo`800000o`3ooomI
0?ooo`800000303oool004P0oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo
00X0oooo00180?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0:0?ooo`00
B03oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0
:03oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
:03oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
2P3oool004P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0:0?ooo`00B03ooooo
000005d00000303oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.418231, 0.704138, \
0.00864489, 0.00279755}}]
}, Closed]],

Cell[TextData[{
  "These results agree very well with those obtained using our custom \
program; you cannot distinguish the two curves.  For historical reasons, we \
note that the agreement was not as good with ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " versions prior to 5.2 because there was a bug in ",
  StyleBox["NonlinearRegress", "InlineInput",
    FontWeight->"Bold"],
  " \[LongDash] while searching for the best fit, earlier versions minimized \
",
  Cell[BoxData[
      \(TraditionalForm\`\[Sum]\+i\(\( w\_i\^2\)(y\_i - y\&_\_i)\)\^2\)]],
  " instead of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Sum]\+i w\_i\ \((y\_i - y\&_\_i)\)\^2\)]],
  " yet computed the estimated variance based upon the latter instead of the \
former.  Fortunately, this bug has now been corrected, but there are still \
advantages to writing programs whose inner workings you understand well and \
can explore or extend, which is not so easy with commercial software.  For \
example, none of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  "'s fitting functions handle uncertainties in the independent variable, \
which is sometimes important.  ",
  StyleBox["No software is perfect \[LongDash] you must be able to develop \
your own programs whenever faced with either limitations or reliability \
issues in commercial software! ",
    FontSlant->"Italic"]
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["The Curfit package", "Section"],

Cell[TextData[{
  "Now that we have functions that perform both linear and nonlinear \
least-squares fits, it is appropriate to package them for use in other \
notebooks.  Improved versions of both ",
  StyleBox["Linfit", "InlineInput"],
  " and ",
  StyleBox["Curfit", "InlineInput"],
  " are available in a package provided for this course.  Both functions \
accommodate data sets with ",
  Cell[BoxData[
      \(TraditionalForm\`2 - 4\)]],
  " columns in the forms ",
  Cell[BoxData[
      \(TraditionalForm\`{x, y}\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`{x, y, \[Sigma]\_y}\)]],
  ", or ",
  Cell[BoxData[
      \(TraditionalForm\`{x, \[Sigma]\_x, y, \[Sigma]\_y}\)]],
  ".  However, because we used many of the same symbols from the prototype \
versions presented in preceding sections, it is useful to start with a fresh \
slate."
}], "Text"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\[IndentingNewLine]", 
    \(Remove["\<Global`*\>"]\)}], "Input"],

Cell[BoxData[
    \(Needs["\<Curfit`\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Curfit`*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["Curfit`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ChiSquare",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`ChiSquare"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["Correlations",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351438214-2222095", 
                    "Curfit`Correlations"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["IncludeSecondDerivatives",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351438214-2222095", 
                    "Curfit`IncludeSecondDerivatives"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["NITS",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`NITS"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["PrintIterations",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351438214-2222095", 
                    "Curfit`PrintIterations"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"]},
              {
                ButtonBox[
                  StyleBox["Conv",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`Conv"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["Curfit",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`Curfit"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["Linfit",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`Linfit"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["NitsDx",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`NitsDx"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["ReducedChiSquare",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351438214-2222095", 
                    "Curfit`ReducedChiSquare"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"]},
              {
                ButtonBox[
                  StyleBox["ConvDx",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`ConvDx"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["DxNits",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`DxNits"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["MaxNITS",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`MaxNITS"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["Parameters",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438214-2222095", "Curfit`Parameters"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["Uncertainties",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351438214-2222095", 
                    "Curfit`Uncertainties"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"]}
              },
            RowMinHeight->{1, 1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell["\<\
The following subsections illustrate this package using some of the preceding \
examples.  In all cases we find good agreement with previous results, but the \
package provides extended features and simpler access to the relevant \
functions.  Feel free to use or modify this package for your future data \
analysis needs.\
\>", "Text"],

Cell[CellGroupData[{

Cell["\<\
Legendre coefficients \[LongDash] compare Linfit and Curfit for a linear \
model\
\>", "Subsection"],

Cell[TextData[{
  "First we verify that the same results are obtained using either ",
  StyleBox["Linfit", "InlineInput"],
  " or ",
  StyleBox["Curfit", "InlineInput"],
  " when the fit function is linear in its parameters.  It is convenient to \
use the same Legendre expansion as above.  The subsections below demonstrate \
that this test succeeds whether or not uncertainties are included for the \
independent variable.  Note that version of ",
  StyleBox["Curfit", "InlineInput"],
  " in the package has been generalized to provide that feature."
}], "Text"],

Cell[CellGroupData[{

Cell["error bars only for dependent variable", "Subsubsection"],

Cell[BoxData[{
    \(\(Clear[z];\)\), "\[IndentingNewLine]", 
    \(\(p = {3.02, 7.3, \(-9.2\)};\)\), "\[IndentingNewLine]", 
    \(\(f = LegendreP[{0, 1, 2}, z];\)\), "\[IndentingNewLine]", 
    \(\(x = Range[\(-0.95\), 0.95, 0.05];\)\), "\[IndentingNewLine]", 
    \(\(y = p . f /. z \[Rule] x;\)\), "\[IndentingNewLine]", 
    \(\(dy = 0.1\ Abs[y];\)\), "\[IndentingNewLine]", 
    \(\(y = 
        y + \((\(Random[NormalDistribution[0, #]] &\) /@ 
              dy)\);\)\), "\[IndentingNewLine]", 
    \(\(data = Transpose[{x, y, dy}];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(results1 = Linfit[data, f, z]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {2.974378756774328`, 
          7.218238791190833`, \(-9.078011013091778`\)}, 
      Uncertainties \[Rule] {0.05568564252442358`, 0.12170190009219685`, 
          0.199171405246533`}, 
      ReducedChiSquare \[Rule] 1.3821780046632806`}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(start = 
        Table[{a[i], Random[]}, {i, 0, 2}];\)\), "\[IndentingNewLine]", 
    \(results2 = 
      Curfit[data, Sum[a[i] LegendreP[i, z], {i, 0, 2}], z, 
        start]\)}], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {a[0] \[Rule] 2.9743787559734574`, 
          a[1] \[Rule] 7.218238789263152`, 
          a[2] \[Rule] \(-9.078011010850481`\)}, 
      Uncertainties \[Rule] {0.05568564252442358`, 0.12170190009219685`, 
          0.199171405246533`}, ReducedChiSquare \[Rule] 1.382178004663281`, 
      ChiSquare \[Rule] 49.758408167878116`, NITS \[Rule] 3, 
      Correlations \[Rule] {{1.`, 
            0.7877248568896436`, \(-0.3943532225833449`\)}, \
{0.7877248568896433`, 
            0.9999999999999999`, \(-0.6391893302144795`\)}, \
{\(-0.39435322258334465`\), \(-0.6391893302144793`\), 1.`}}}\)], "Output"]
}, Closed]],

Cell["\<\
Note that the initial guesses are irrelevant when the model is linear in its \
parameters; therefore, we just used random numbers.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["error bars for both independent and dependent variables", \
"Subsubsection"],

Cell[BoxData[{
    \(\(Clear[z];\)\), "\[IndentingNewLine]", 
    \(\(p = {3.02, 7.3, \(-9.2\)};\)\), "\[IndentingNewLine]", 
    \(\(f = LegendreP[{0, 1, 2}, z];\)\), "\[IndentingNewLine]", 
    \(\(x = Range[\(-0.95\), 0.95, 0.05];\)\), "\[IndentingNewLine]", 
    \(\(dx = Table[0.05, {Length[x]}];\)\), "\[IndentingNewLine]", 
    \(\(y = p . f /. z \[Rule] x;\)\), "\[IndentingNewLine]", 
    \(\(dy = 0.1\ Abs[y];\)\), "\[IndentingNewLine]", 
    \(\(x = 
        x + \((\(Random[NormalDistribution[0, #]] &\) /@ 
              dx)\);\)\), "\[IndentingNewLine]", 
    \(\(y = 
        y + \((\(Random[NormalDistribution[0, #]] &\) /@ 
              dy)\);\)\), "\[IndentingNewLine]", 
    \(\(data = Transpose[{x, dx, y, dy}];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(results1 = Linfit[data, f, z]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {2.8238576237067385`, 
          7.2047905207147265`, \(-8.957516830273583`\)}, 
      Uncertainties \[Rule] {0.169921259979984`, 0.32913701324586536`, 
          0.4200358636463869`}, ReducedChiSquare \[Rule] 0.8746516386163016`, 
      NITS \[Rule] 4}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(start = 
        Table[{a[i], Random[]}, {i, 0, 2}];\)\), "\[IndentingNewLine]", 
    \(results2 = 
      Curfit[data, Sum[a[i] LegendreP[i, z], {i, 0, 2}], z, 
        start]\)}], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {a[0] \[Rule] 2.823857712813643`, 
          a[1] \[Rule] 7.2047899952514864`, 
          a[2] \[Rule] \(-8.957516071052812`\)}, 
      Uncertainties \[Rule] {0.16992125980037567`, 0.32913701290641206`, 
          0.4200358632302085`}, ReducedChiSquare \[Rule] 0.8746516400783058`, 
      ChiSquare \[Rule] 31.487459042819008`, NITS \[Rule] 1, 
      Correlations \[Rule] {{1.`, \(-0.40629450856421834`\), 
            0.4184327529029214`}, {\(-0.40629450856421856`\), 
            1.`, \(-0.33431152990494317`\)}, {0.4184327529029214`, \
\(-0.33431152990494306`\), 1.`}}, DxNits \[Rule] {3, 2, 2, 1}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        MultipleListPlot[
          Evaluate[
            Transpose[{x, dx, y, dy}] /. {xx_, dxx_, yy_, 
                  dyy_} \[Rule] {{xx, yy}, ErrorBar[dxx, dyy]}], 
          PlotRange \[Rule] All, 
          PlotStyle \[Rule] PointSize[0.02]], \[IndentingNewLine]Plot[
          Evaluate[Parameters . f /. results1], {z, \(-1\), 1}], 
        Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", "\<y\>"}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.501883 0.46435 0.348991 0.0240553 [
[.03753 -0.0125 -9.34375 -15.6875 ]
[.03753 -0.0125 9.34375 0 ]
[.26971 -0.0125 -13.8438 -15.6875 ]
[.26971 -0.0125 13.8438 0 ]
[.50188 -0.0125 -5 -15.6875 ]
[.50188 -0.0125 5 0 ]
[.73406 -0.0125 -9.5 -15.6875 ]
[.73406 -0.0125 9.5 0 ]
[.96623 -0.0125 -5 -15.6875 ]
[.96623 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .10844 -24.6875 -7.84375 ]
[-0.0125 .10844 0 7.84375 ]
[-0.0125 .22871 -18.6875 -7.84375 ]
[-0.0125 .22871 0 7.84375 ]
[-0.0125 .34899 -10 -7.84375 ]
[-0.0125 .34899 0 7.84375 ]
[-0.0125 .46927 -10 -7.84375 ]
[-0.0125 .46927 0 7.84375 ]
[-0.0125 .58954 -16 -7.84375 ]
[-0.0125 .58954 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -40.375 -5.09375 ]
[-0.01875 .30902 -24.6875 5.09375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.03753 0 m
.03753 .00625 L
s
gsave
.03753 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.26971 0 m
.26971 .00625 L
s
gsave
.26971 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.50188 0 m
.50188 .00625 L
s
gsave
.50188 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.73406 0 m
.73406 .00625 L
s
gsave
.73406 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.96623 0 m
.96623 .00625 L
s
gsave
.96623 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.08397 0 m
.08397 .00375 L
s
.1304 0 m
.1304 .00375 L
s
.17684 0 m
.17684 .00375 L
s
.22327 0 m
.22327 .00375 L
s
.31614 0 m
.31614 .00375 L
s
.36258 0 m
.36258 .00375 L
s
.40901 0 m
.40901 .00375 L
s
.45545 0 m
.45545 .00375 L
s
.54832 0 m
.54832 .00375 L
s
.59475 0 m
.59475 .00375 L
s
.64119 0 m
.64119 .00375 L
s
.68762 0 m
.68762 .00375 L
s
.78049 0 m
.78049 .00375 L
s
.82693 0 m
.82693 .00375 L
s
.87336 0 m
.87336 .00375 L
s
.9198 0 m
.9198 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10844 m
.00625 .10844 L
s
gsave
-0.0125 .10844 -85.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
0 .22871 m
.00625 .22871 L
s
gsave
-0.0125 .22871 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .34899 m
.00625 .34899 L
s
gsave
-0.0125 .34899 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .46927 m
.00625 .46927 L
s
gsave
-0.0125 .46927 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
0 .58954 m
.00625 .58954 L
s
gsave
-0.0125 .58954 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .13249 m
.00375 .13249 L
s
0 .15655 m
.00375 .15655 L
s
0 .1806 m
.00375 .1806 L
s
0 .20466 m
.00375 .20466 L
s
0 .25277 m
.00375 .25277 L
s
0 .27682 m
.00375 .27682 L
s
0 .30088 m
.00375 .30088 L
s
0 .32494 m
.00375 .32494 L
s
0 .37305 m
.00375 .37305 L
s
0 .3971 m
.00375 .3971 L
s
0 .42116 m
.00375 .42116 L
s
0 .44521 m
.00375 .44521 L
s
0 .49332 m
.00375 .49332 L
s
0 .51738 m
.00375 .51738 L
s
0 .54143 m
.00375 .54143 L
s
0 .56549 m
.00375 .56549 L
s
0 .08438 m
.00375 .08438 L
s
0 .06033 m
.00375 .06033 L
s
0 .03627 m
.00375 .03627 L
s
0 .01222 m
.00375 .01222 L
s
0 .6136 m
.00375 .6136 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -98.625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
66.0938 11.8438 translate
90 rotate
-66.0938 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.938 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(y) show
69.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.03753 .61178 m
.03753 .61803 L
s
.26971 .61178 m
.26971 .61803 L
s
.50188 .61178 m
.50188 .61803 L
s
.73406 .61178 m
.73406 .61803 L
s
.96623 .61178 m
.96623 .61803 L
s
.125 Mabswid
.08397 .61428 m
.08397 .61803 L
s
.1304 .61428 m
.1304 .61803 L
s
.17684 .61428 m
.17684 .61803 L
s
.22327 .61428 m
.22327 .61803 L
s
.31614 .61428 m
.31614 .61803 L
s
.36258 .61428 m
.36258 .61803 L
s
.40901 .61428 m
.40901 .61803 L
s
.45545 .61428 m
.45545 .61803 L
s
.54832 .61428 m
.54832 .61803 L
s
.59475 .61428 m
.59475 .61803 L
s
.64119 .61428 m
.64119 .61803 L
s
.68762 .61428 m
.68762 .61803 L
s
.78049 .61428 m
.78049 .61803 L
s
.82693 .61428 m
.82693 .61803 L
s
.87336 .61428 m
.87336 .61803 L
s
.9198 .61428 m
.9198 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .10844 m
1 .10844 L
s
.99375 .22871 m
1 .22871 L
s
.99375 .34899 m
1 .34899 L
s
.99375 .46927 m
1 .46927 L
s
.99375 .58954 m
1 .58954 L
s
.125 Mabswid
.99625 .13249 m
1 .13249 L
s
.99625 .15655 m
1 .15655 L
s
.99625 .1806 m
1 .1806 L
s
.99625 .20466 m
1 .20466 L
s
.99625 .25277 m
1 .25277 L
s
.99625 .27682 m
1 .27682 L
s
.99625 .30088 m
1 .30088 L
s
.99625 .32494 m
1 .32494 L
s
.99625 .37305 m
1 .37305 L
s
.99625 .3971 m
1 .3971 L
s
.99625 .42116 m
1 .42116 L
s
.99625 .44521 m
1 .44521 L
s
.99625 .49332 m
1 .49332 L
s
.99625 .51738 m
1 .51738 L
s
.99625 .54143 m
1 .54143 L
s
.99625 .56549 m
1 .56549 L
s
.99625 .08438 m
1 .08438 L
s
.99625 .06033 m
1 .06033 L
s
.99625 .03627 m
1 .03627 L
s
.99625 .01222 m
1 .01222 L
s
.99625 .6136 m
1 .6136 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .34899 m
1 .34899 L
s
.50188 0 m
.50188 .61803 L
s
.001 w
.07024 .04303 m
.02381 .04303 L
s
.07024 .04303 0 1.5 Mabsadd m
.07024 .04303 0 -1.5 Mabsadd L s
.02381 .04303 0 1.5 Mabsadd m
.02381 .04303 0 -1.5 Mabsadd L s
.04703 .07134 m
.04703 .01472 L
s
.04703 .07134 1.5 0 Mabsadd m
.04703 .07134 -1.5 0 Mabsadd L s
.04703 .01472 1.5 0 Mabsadd m
.04703 .01472 -1.5 0 Mabsadd L s
.10361 .09502 m
.05718 .09502 L
s
.10361 .09502 0 1.5 Mabsadd m
.10361 .09502 0 -1.5 Mabsadd L s
.05718 .09502 0 1.5 Mabsadd m
.05718 .09502 0 -1.5 Mabsadd L s
.0804 .11938 m
.0804 .07066 L
s
.0804 .11938 1.5 0 Mabsadd m
.0804 .11938 -1.5 0 Mabsadd L s
.0804 .07066 1.5 0 Mabsadd m
.0804 .07066 -1.5 0 Mabsadd L s
.10349 .13802 m
.05706 .13802 L
s
.10349 .13802 0 1.5 Mabsadd m
.10349 .13802 0 -1.5 Mabsadd L s
.05706 .13802 0 1.5 Mabsadd m
.05706 .13802 0 -1.5 Mabsadd L s
.08028 .15861 m
.08028 .11744 L
s
.08028 .15861 1.5 0 Mabsadd m
.08028 .15861 -1.5 0 Mabsadd L s
.08028 .11744 1.5 0 Mabsadd m
.08028 .11744 -1.5 0 Mabsadd L s
.15631 .18774 m
.10987 .18774 L
s
.15631 .18774 0 1.5 Mabsadd m
.15631 .18774 0 -1.5 Mabsadd L s
.10987 .18774 0 1.5 Mabsadd m
.10987 .18774 0 -1.5 Mabsadd L s
.13309 .20471 m
.13309 .17078 L
s
.13309 .20471 1.5 0 Mabsadd m
.13309 .20471 -1.5 0 Mabsadd L s
.13309 .17078 1.5 0 Mabsadd m
.13309 .17078 -1.5 0 Mabsadd L s
.21756 .21219 m
.17112 .21219 L
s
.21756 .21219 0 1.5 Mabsadd m
.21756 .21219 0 -1.5 Mabsadd L s
.17112 .21219 0 1.5 Mabsadd m
.17112 .21219 0 -1.5 Mabsadd L s
.19434 .2257 m
.19434 .19868 L
s
.19434 .2257 1.5 0 Mabsadd m
.19434 .2257 -1.5 0 Mabsadd L s
.19434 .19868 1.5 0 Mabsadd m
.19434 .19868 -1.5 0 Mabsadd L s
.16956 .24132 m
.12312 .24132 L
s
.16956 .24132 0 1.5 Mabsadd m
.16956 .24132 0 -1.5 Mabsadd L s
.12312 .24132 0 1.5 Mabsadd m
.12312 .24132 0 -1.5 Mabsadd L s
.14634 .25155 m
.14634 .2311 L
s
.14634 .25155 1.5 0 Mabsadd m
.14634 .25155 -1.5 0 Mabsadd L s
.14634 .2311 1.5 0 Mabsadd m
.14634 .2311 -1.5 0 Mabsadd L s
.23723 .27927 m
.19079 .27927 L
s
.23723 .27927 0 1.5 Mabsadd m
.23723 .27927 0 -1.5 Mabsadd L s
.19079 .27927 0 1.5 Mabsadd m
.19079 .27927 0 -1.5 Mabsadd L s
.21401 .28638 m
.21401 .27216 L
s
.21401 .28638 1.5 0 Mabsadd m
.21401 .28638 -1.5 0 Mabsadd L s
.21401 .27216 1.5 0 Mabsadd m
.21401 .27216 -1.5 0 Mabsadd L s
.26442 .30203 m
.21798 .30203 L
s
.26442 .30203 0 1.5 Mabsadd m
.26442 .30203 0 -1.5 Mabsadd L s
.21798 .30203 0 1.5 Mabsadd m
.21798 .30203 0 -1.5 Mabsadd L s
.2412 .30619 m
.2412 .29788 L
s
.2412 .30619 1.5 0 Mabsadd m
.2412 .30619 -1.5 0 Mabsadd L s
.2412 .29788 1.5 0 Mabsadd m
.2412 .29788 -1.5 0 Mabsadd L s
.26977 .33443 m
.22333 .33443 L
s
.26977 .33443 0 1.5 Mabsadd m
.26977 .33443 0 -1.5 Mabsadd L s
.22333 .33443 0 1.5 Mabsadd m
.22333 .33443 0 -1.5 Mabsadd L s
.24655 .3358 m
.24655 .33306 L
s
.24655 .3358 1.5 0 Mabsadd m
.24655 .3358 -1.5 0 Mabsadd L s
.24655 .33306 1.5 0 Mabsadd m
.24655 .33306 -1.5 0 Mabsadd L s
.26995 .36203 m
.22352 .36203 L
s
.26995 .36203 0 1.5 Mabsadd m
.26995 .36203 0 -1.5 Mabsadd L s
.22352 .36203 0 1.5 Mabsadd m
.22352 .36203 0 -1.5 Mabsadd L s
.24673 .36328 m
.24673 .36078 L
s
.24673 .36328 1.5 0 Mabsadd m
.24673 .36328 -1.5 0 Mabsadd L s
.24673 .36078 1.5 0 Mabsadd m
.24673 .36078 -1.5 0 Mabsadd L s
.33809 .39132 m
.29166 .39132 L
s
.33809 .39132 0 1.5 Mabsadd m
.33809 .39132 0 -1.5 Mabsadd L s
.29166 .39132 0 1.5 Mabsadd m
.29166 .39132 0 -1.5 Mabsadd L s
.31487 .39502 m
.31487 .38761 L
s
.31487 .39502 1.5 0 Mabsadd m
.31487 .39502 -1.5 0 Mabsadd L s
.31487 .38761 1.5 0 Mabsadd m
.31487 .38761 -1.5 0 Mabsadd L s
.36198 .4049 m
.31554 .4049 L
s
.36198 .4049 0 1.5 Mabsadd m
.36198 .4049 0 -1.5 Mabsadd L s
.31554 .4049 0 1.5 Mabsadd m
.31554 .4049 0 -1.5 Mabsadd L s
.33876 .41089 m
.33876 .3989 L
s
.33876 .41089 1.5 0 Mabsadd m
.33876 .41089 -1.5 0 Mabsadd L s
.33876 .3989 1.5 0 Mabsadd m
.33876 .3989 -1.5 0 Mabsadd L s
.36023 .42435 m
.3138 .42435 L
s
.36023 .42435 0 1.5 Mabsadd m
.36023 .42435 0 -1.5 Mabsadd L s
.3138 .42435 0 1.5 Mabsadd m
.3138 .42435 0 -1.5 Mabsadd L s
.33702 .43247 m
.33702 .41624 L
s
.33702 .43247 1.5 0 Mabsadd m
.33702 .43247 -1.5 0 Mabsadd L s
.33702 .41624 1.5 0 Mabsadd m
.33702 .41624 -1.5 0 Mabsadd L s
.41041 .45566 m
.36397 .45566 L
s
.41041 .45566 0 1.5 Mabsadd m
.41041 .45566 0 -1.5 Mabsadd L s
.36397 .45566 0 1.5 Mabsadd m
.36397 .45566 0 -1.5 Mabsadd L s
.38719 .46574 m
.38719 .44559 L
s
.38719 .46574 1.5 0 Mabsadd m
.38719 .46574 -1.5 0 Mabsadd L s
.38719 .44559 1.5 0 Mabsadd m
.38719 .44559 -1.5 0 Mabsadd L s
.3866 .47346 m
.34016 .47346 L
s
.3866 .47346 0 1.5 Mabsadd m
.3866 .47346 0 -1.5 Mabsadd L s
.34016 .47346 0 1.5 Mabsadd m
.34016 .47346 0 -1.5 Mabsadd L s
.36338 .48532 m
.36338 .46159 L
s
.36338 .48532 1.5 0 Mabsadd m
.36338 .48532 -1.5 0 Mabsadd L s
.36338 .46159 1.5 0 Mabsadd m
.36338 .46159 -1.5 0 Mabsadd L s
.44245 .48602 m
.39602 .48602 L
s
.44245 .48602 0 1.5 Mabsadd m
.44245 .48602 0 -1.5 Mabsadd L s
.39602 .48602 0 1.5 Mabsadd m
.39602 .48602 0 -1.5 Mabsadd L s
.41924 .49951 m
.41924 .47253 L
s
.41924 .49951 1.5 0 Mabsadd m
.41924 .49951 -1.5 0 Mabsadd L s
.41924 .47253 1.5 0 Mabsadd m
.41924 .47253 -1.5 0 Mabsadd L s
.47716 .48477 m
.43073 .48477 L
s
.47716 .48477 0 1.5 Mabsadd m
.47716 .48477 0 -1.5 Mabsadd L s
.43073 .48477 0 1.5 Mabsadd m
.43073 .48477 0 -1.5 Mabsadd L s
.45395 .49972 m
.45395 .46982 L
s
.45395 .49972 1.5 0 Mabsadd m
.45395 .49972 -1.5 0 Mabsadd L s
.45395 .46982 1.5 0 Mabsadd m
.45395 .46982 -1.5 0 Mabsadd L s
.45421 .51322 m
.40777 .51322 L
s
.45421 .51322 0 1.5 Mabsadd m
.45421 .51322 0 -1.5 Mabsadd L s
.40777 .51322 0 1.5 Mabsadd m
.40777 .51322 0 -1.5 Mabsadd L s
.43099 .52946 m
.43099 .49697 L
s
.43099 .52946 1.5 0 Mabsadd m
.43099 .52946 -1.5 0 Mabsadd L s
.43099 .49697 1.5 0 Mabsadd m
.43099 .49697 -1.5 0 Mabsadd L s
.52942 .52081 m
.48299 .52081 L
s
.52942 .52081 0 1.5 Mabsadd m
.52942 .52081 0 -1.5 Mabsadd L s
.48299 .52081 0 1.5 Mabsadd m
.48299 .52081 0 -1.5 Mabsadd L s
.50621 .53818 m
.50621 .50344 L
s
.50621 .53818 1.5 0 Mabsadd m
.50621 .53818 -1.5 0 Mabsadd L s
.50621 .50344 1.5 0 Mabsadd m
.50621 .50344 -1.5 0 Mabsadd L s
.52997 .51852 m
.48353 .51852 L
s
.52997 .51852 0 1.5 Mabsadd m
.52997 .51852 0 -1.5 Mabsadd L s
.48353 .51852 0 1.5 Mabsadd m
.48353 .51852 0 -1.5 Mabsadd L s
.50675 .53685 m
.50675 .50019 L
s
.50675 .53685 1.5 0 Mabsadd m
.50675 .53685 -1.5 0 Mabsadd L s
.50675 .50019 1.5 0 Mabsadd m
.50675 .50019 -1.5 0 Mabsadd L s
.5612 .52 m
.51477 .52 L
s
.5612 .52 0 1.5 Mabsadd m
.5612 .52 0 -1.5 Mabsadd L s
.51477 .52 0 1.5 Mabsadd m
.51477 .52 0 -1.5 Mabsadd L s
.53799 .53913 m
.53799 .50088 L
s
.53799 .53913 1.5 0 Mabsadd m
.53799 .53913 -1.5 0 Mabsadd L s
.53799 .50088 1.5 0 Mabsadd m
.53799 .50088 -1.5 0 Mabsadd L s
.6123 .55006 m
.56587 .55006 L
s
.6123 .55006 0 1.5 Mabsadd m
.6123 .55006 0 -1.5 Mabsadd L s
.56587 .55006 0 1.5 Mabsadd m
.56587 .55006 0 -1.5 Mabsadd L s
.58908 .56981 m
.58908 .5303 L
s
.58908 .56981 1.5 0 Mabsadd m
.58908 .56981 -1.5 0 Mabsadd L s
.58908 .5303 1.5 0 Mabsadd m
.58908 .5303 -1.5 0 Mabsadd L s
.59684 .54015 m
.55041 .54015 L
s
.59684 .54015 0 1.5 Mabsadd m
.59684 .54015 0 -1.5 Mabsadd L s
.55041 .54015 0 1.5 Mabsadd m
.55041 .54015 0 -1.5 Mabsadd L s
.57362 .56036 m
.57362 .51993 L
s
.57362 .56036 1.5 0 Mabsadd m
.57362 .56036 -1.5 0 Mabsadd L s
.57362 .51993 1.5 0 Mabsadd m
.57362 .51993 -1.5 0 Mabsadd L s
.62129 .56305 m
.57485 .56305 L
s
.62129 .56305 0 1.5 Mabsadd m
.62129 .56305 0 -1.5 Mabsadd L s
.57485 .56305 0 1.5 Mabsadd m
.57485 .56305 0 -1.5 Mabsadd L s
.59807 .58356 m
.59807 .54253 L
s
.59807 .58356 1.5 0 Mabsadd m
.59807 .58356 -1.5 0 Mabsadd L s
.59807 .54253 1.5 0 Mabsadd m
.59807 .54253 -1.5 0 Mabsadd L s
.69417 .58267 m
.64774 .58267 L
s
.69417 .58267 0 1.5 Mabsadd m
.69417 .58267 0 -1.5 Mabsadd L s
.64774 .58267 0 1.5 Mabsadd m
.64774 .58267 0 -1.5 Mabsadd L s
.67096 .60332 m
.67096 .56203 L
s
.67096 .60332 1.5 0 Mabsadd m
.67096 .60332 -1.5 0 Mabsadd L s
.67096 .56203 1.5 0 Mabsadd m
.67096 .56203 -1.5 0 Mabsadd L s
.68344 .55794 m
.63701 .55794 L
s
.68344 .55794 0 1.5 Mabsadd m
.68344 .55794 0 -1.5 Mabsadd L s
.63701 .55794 0 1.5 Mabsadd m
.63701 .55794 0 -1.5 Mabsadd L s
.66022 .57855 m
.66022 .53733 L
s
.66022 .57855 1.5 0 Mabsadd m
.66022 .57855 -1.5 0 Mabsadd L s
.66022 .53733 1.5 0 Mabsadd m
.66022 .53733 -1.5 0 Mabsadd L s
.68709 .57915 m
.64065 .57915 L
s
.68709 .57915 0 1.5 Mabsadd m
.68709 .57915 0 -1.5 Mabsadd L s
.64065 .57915 0 1.5 Mabsadd m
.64065 .57915 0 -1.5 Mabsadd L s
.66387 .59956 m
.66387 .55874 L
s
.66387 .59956 1.5 0 Mabsadd m
.66387 .59956 -1.5 0 Mabsadd L s
.66387 .55874 1.5 0 Mabsadd m
.66387 .55874 -1.5 0 Mabsadd L s
.72357 .52858 m
.67713 .52858 L
s
.72357 .52858 0 1.5 Mabsadd m
.72357 .52858 0 -1.5 Mabsadd L s
.67713 .52858 0 1.5 Mabsadd m
.67713 .52858 0 -1.5 Mabsadd L s
.70035 .54863 m
.70035 .50854 L
s
.70035 .54863 1.5 0 Mabsadd m
.70035 .54863 -1.5 0 Mabsadd L s
.70035 .50854 1.5 0 Mabsadd m
.70035 .50854 -1.5 0 Mabsadd L s
.73151 .53096 m
.68508 .53096 L
s
.73151 .53096 0 1.5 Mabsadd m
.73151 .53096 0 -1.5 Mabsadd L s
.68508 .53096 0 1.5 Mabsadd m
.68508 .53096 0 -1.5 Mabsadd L s
.70829 .55047 m
.70829 .51145 L
s
.70829 .55047 1.5 0 Mabsadd m
.70829 .55047 -1.5 0 Mabsadd L s
.70829 .51145 1.5 0 Mabsadd m
.70829 .51145 -1.5 0 Mabsadd L s
.74033 .54864 m
.69389 .54864 L
s
.74033 .54864 0 1.5 Mabsadd m
.74033 .54864 0 -1.5 Mabsadd L s
.69389 .54864 0 1.5 Mabsadd m
.69389 .54864 0 -1.5 Mabsadd L s
.71711 .56745 m
.71711 .52983 L
s
.71711 .56745 1.5 0 Mabsadd m
.71711 .56745 -1.5 0 Mabsadd L s
.71711 .52983 1.5 0 Mabsadd m
.71711 .52983 -1.5 0 Mabsadd L s
.75556 .52851 m
.70912 .52851 L
s
.75556 .52851 0 1.5 Mabsadd m
.75556 .52851 0 -1.5 Mabsadd L s
.70912 .52851 0 1.5 Mabsadd m
.70912 .52851 0 -1.5 Mabsadd L s
.73234 .54646 m
.73234 .51057 L
s
.73234 .54646 1.5 0 Mabsadd m
.73234 .54646 -1.5 0 Mabsadd L s
.73234 .51057 1.5 0 Mabsadd m
.73234 .51057 -1.5 0 Mabsadd L s
.76857 .5102 m
.72214 .5102 L
s
.76857 .5102 0 1.5 Mabsadd m
.76857 .5102 0 -1.5 Mabsadd L s
.72214 .5102 0 1.5 Mabsadd m
.72214 .5102 0 -1.5 Mabsadd L s
.74536 .52712 m
.74536 .49329 L
s
.74536 .52712 1.5 0 Mabsadd m
.74536 .52712 -1.5 0 Mabsadd L s
.74536 .49329 1.5 0 Mabsadd m
.74536 .49329 -1.5 0 Mabsadd L s
.80798 .49956 m
.76155 .49956 L
s
.80798 .49956 0 1.5 Mabsadd m
.80798 .49956 0 -1.5 Mabsadd L s
.76155 .49956 0 1.5 Mabsadd m
.76155 .49956 0 -1.5 Mabsadd L s
.78476 .51528 m
.78476 .48384 L
s
.78476 .51528 1.5 0 Mabsadd m
.78476 .51528 -1.5 0 Mabsadd L s
.78476 .48384 1.5 0 Mabsadd m
.78476 .48384 -1.5 0 Mabsadd L s
.79959 .49947 m
.75315 .49947 L
s
.79959 .49947 0 1.5 Mabsadd m
.79959 .49947 0 -1.5 Mabsadd L s
.75315 .49947 0 1.5 Mabsadd m
.75315 .49947 0 -1.5 Mabsadd L s
.77637 .51382 m
.77637 .48511 L
s
.77637 .51382 1.5 0 Mabsadd m
.77637 .51382 -1.5 0 Mabsadd L s
.77637 .48511 1.5 0 Mabsadd m
.77637 .48511 -1.5 0 Mabsadd L s
.88528 .50269 m
.83884 .50269 L
s
.88528 .50269 0 1.5 Mabsadd m
.88528 .50269 0 -1.5 Mabsadd L s
.83884 .50269 0 1.5 Mabsadd m
.83884 .50269 0 -1.5 Mabsadd L s
.86206 .51551 m
.86206 .48986 L
s
.86206 .51551 1.5 0 Mabsadd m
.86206 .51551 -1.5 0 Mabsadd L s
.86206 .48986 1.5 0 Mabsadd m
.86206 .48986 -1.5 0 Mabsadd L s
.85547 .45603 m
.80904 .45603 L
s
.85547 .45603 0 1.5 Mabsadd m
.85547 .45603 0 -1.5 Mabsadd L s
.80904 .45603 0 1.5 Mabsadd m
.80904 .45603 0 -1.5 Mabsadd L s
.83225 .46717 m
.83225 .4449 L
s
.83225 .46717 1.5 0 Mabsadd m
.83225 .46717 -1.5 0 Mabsadd L s
.83225 .4449 1.5 0 Mabsadd m
.83225 .4449 -1.5 0 Mabsadd L s
.93183 .44669 m
.88539 .44669 L
s
.93183 .44669 0 1.5 Mabsadd m
.93183 .44669 0 -1.5 Mabsadd L s
.88539 .44669 0 1.5 Mabsadd m
.88539 .44669 0 -1.5 Mabsadd L s
.90861 .45596 m
.90861 .43741 L
s
.90861 .45596 1.5 0 Mabsadd m
.90861 .45596 -1.5 0 Mabsadd L s
.90861 .43741 1.5 0 Mabsadd m
.90861 .43741 -1.5 0 Mabsadd L s
.91355 .41526 m
.86711 .41526 L
s
.91355 .41526 0 1.5 Mabsadd m
.91355 .41526 0 -1.5 Mabsadd L s
.86711 .41526 0 1.5 Mabsadd m
.86711 .41526 0 -1.5 Mabsadd L s
.89033 .4225 m
.89033 .40801 L
s
.89033 .4225 1.5 0 Mabsadd m
.89033 .4225 -1.5 0 Mabsadd L s
.89033 .40801 1.5 0 Mabsadd m
.89033 .40801 -1.5 0 Mabsadd L s
.97619 .40071 m
.92976 .40071 L
s
.97619 .40071 0 1.5 Mabsadd m
.97619 .40071 0 -1.5 Mabsadd L s
.92976 .40071 0 1.5 Mabsadd m
.92976 .40071 0 -1.5 Mabsadd L s
.95297 .40576 m
.95297 .39565 L
s
.95297 .40576 1.5 0 Mabsadd m
.95297 .40576 -1.5 0 Mabsadd L s
.95297 .39565 1.5 0 Mabsadd m
.95297 .39565 -1.5 0 Mabsadd L s
.0001 w
.04703 .04303 -1.875 0 Mabsadd m
.04703 .04303 0 2.5 Mabsadd L
.04703 .04303 1.875 0 Mabsadd L
.04703 .04303 0 -2.5 Mabsadd L
.04703 .04303 -1.875 0 Mabsadd L closepath F
.0804 .09502 -1.875 0 Mabsadd m
.0804 .09502 0 2.5 Mabsadd L
.0804 .09502 1.875 0 Mabsadd L
.0804 .09502 0 -2.5 Mabsadd L
.0804 .09502 -1.875 0 Mabsadd L closepath F
.08028 .13802 -1.875 0 Mabsadd m
.08028 .13802 0 2.5 Mabsadd L
.08028 .13802 1.875 0 Mabsadd L
.08028 .13802 0 -2.5 Mabsadd L
.08028 .13802 -1.875 0 Mabsadd L closepath F
.13309 .18774 -1.875 0 Mabsadd m
.13309 .18774 0 2.5 Mabsadd L
.13309 .18774 1.875 0 Mabsadd L
.13309 .18774 0 -2.5 Mabsadd L
.13309 .18774 -1.875 0 Mabsadd L closepath F
.19434 .21219 -1.875 0 Mabsadd m
.19434 .21219 0 2.5 Mabsadd L
.19434 .21219 1.875 0 Mabsadd L
.19434 .21219 0 -2.5 Mabsadd L
.19434 .21219 -1.875 0 Mabsadd L closepath F
.14634 .24132 -1.875 0 Mabsadd m
.14634 .24132 0 2.5 Mabsadd L
.14634 .24132 1.875 0 Mabsadd L
.14634 .24132 0 -2.5 Mabsadd L
.14634 .24132 -1.875 0 Mabsadd L closepath F
.21401 .27927 -1.875 0 Mabsadd m
.21401 .27927 0 2.5 Mabsadd L
.21401 .27927 1.875 0 Mabsadd L
.21401 .27927 0 -2.5 Mabsadd L
.21401 .27927 -1.875 0 Mabsadd L closepath F
.2412 .30203 -1.875 0 Mabsadd m
.2412 .30203 0 2.5 Mabsadd L
.2412 .30203 1.875 0 Mabsadd L
.2412 .30203 0 -2.5 Mabsadd L
.2412 .30203 -1.875 0 Mabsadd L closepath F
.24655 .33443 -1.875 0 Mabsadd m
.24655 .33443 0 2.5 Mabsadd L
.24655 .33443 1.875 0 Mabsadd L
.24655 .33443 0 -2.5 Mabsadd L
.24655 .33443 -1.875 0 Mabsadd L closepath F
.24673 .36203 -1.875 0 Mabsadd m
.24673 .36203 0 2.5 Mabsadd L
.24673 .36203 1.875 0 Mabsadd L
.24673 .36203 0 -2.5 Mabsadd L
.24673 .36203 -1.875 0 Mabsadd L closepath F
.31487 .39132 -1.875 0 Mabsadd m
.31487 .39132 0 2.5 Mabsadd L
.31487 .39132 1.875 0 Mabsadd L
.31487 .39132 0 -2.5 Mabsadd L
.31487 .39132 -1.875 0 Mabsadd L closepath F
.33876 .4049 -1.875 0 Mabsadd m
.33876 .4049 0 2.5 Mabsadd L
.33876 .4049 1.875 0 Mabsadd L
.33876 .4049 0 -2.5 Mabsadd L
.33876 .4049 -1.875 0 Mabsadd L closepath F
.33702 .42435 -1.875 0 Mabsadd m
.33702 .42435 0 2.5 Mabsadd L
.33702 .42435 1.875 0 Mabsadd L
.33702 .42435 0 -2.5 Mabsadd L
.33702 .42435 -1.875 0 Mabsadd L closepath F
.38719 .45566 -1.875 0 Mabsadd m
.38719 .45566 0 2.5 Mabsadd L
.38719 .45566 1.875 0 Mabsadd L
.38719 .45566 0 -2.5 Mabsadd L
.38719 .45566 -1.875 0 Mabsadd L closepath F
.36338 .47346 -1.875 0 Mabsadd m
.36338 .47346 0 2.5 Mabsadd L
.36338 .47346 1.875 0 Mabsadd L
.36338 .47346 0 -2.5 Mabsadd L
.36338 .47346 -1.875 0 Mabsadd L closepath F
.41924 .48602 -1.875 0 Mabsadd m
.41924 .48602 0 2.5 Mabsadd L
.41924 .48602 1.875 0 Mabsadd L
.41924 .48602 0 -2.5 Mabsadd L
.41924 .48602 -1.875 0 Mabsadd L closepath F
.45395 .48477 -1.875 0 Mabsadd m
.45395 .48477 0 2.5 Mabsadd L
.45395 .48477 1.875 0 Mabsadd L
.45395 .48477 0 -2.5 Mabsadd L
.45395 .48477 -1.875 0 Mabsadd L closepath F
.43099 .51322 -1.875 0 Mabsadd m
.43099 .51322 0 2.5 Mabsadd L
.43099 .51322 1.875 0 Mabsadd L
.43099 .51322 0 -2.5 Mabsadd L
.43099 .51322 -1.875 0 Mabsadd L closepath F
.50621 .52081 -1.875 0 Mabsadd m
.50621 .52081 0 2.5 Mabsadd L
.50621 .52081 1.875 0 Mabsadd L
.50621 .52081 0 -2.5 Mabsadd L
.50621 .52081 -1.875 0 Mabsadd L closepath F
.50675 .51852 -1.875 0 Mabsadd m
.50675 .51852 0 2.5 Mabsadd L
.50675 .51852 1.875 0 Mabsadd L
.50675 .51852 0 -2.5 Mabsadd L
.50675 .51852 -1.875 0 Mabsadd L closepath F
.53799 .52 -1.875 0 Mabsadd m
.53799 .52 0 2.5 Mabsadd L
.53799 .52 1.875 0 Mabsadd L
.53799 .52 0 -2.5 Mabsadd L
.53799 .52 -1.875 0 Mabsadd L closepath F
.58908 .55006 -1.875 0 Mabsadd m
.58908 .55006 0 2.5 Mabsadd L
.58908 .55006 1.875 0 Mabsadd L
.58908 .55006 0 -2.5 Mabsadd L
.58908 .55006 -1.875 0 Mabsadd L closepath F
.57362 .54015 -1.875 0 Mabsadd m
.57362 .54015 0 2.5 Mabsadd L
.57362 .54015 1.875 0 Mabsadd L
.57362 .54015 0 -2.5 Mabsadd L
.57362 .54015 -1.875 0 Mabsadd L closepath F
.59807 .56305 -1.875 0 Mabsadd m
.59807 .56305 0 2.5 Mabsadd L
.59807 .56305 1.875 0 Mabsadd L
.59807 .56305 0 -2.5 Mabsadd L
.59807 .56305 -1.875 0 Mabsadd L closepath F
.67096 .58267 -1.875 0 Mabsadd m
.67096 .58267 0 2.5 Mabsadd L
.67096 .58267 1.875 0 Mabsadd L
.67096 .58267 0 -2.5 Mabsadd L
.67096 .58267 -1.875 0 Mabsadd L closepath F
.66022 .55794 -1.875 0 Mabsadd m
.66022 .55794 0 2.5 Mabsadd L
.66022 .55794 1.875 0 Mabsadd L
.66022 .55794 0 -2.5 Mabsadd L
.66022 .55794 -1.875 0 Mabsadd L closepath F
.66387 .57915 -1.875 0 Mabsadd m
.66387 .57915 0 2.5 Mabsadd L
.66387 .57915 1.875 0 Mabsadd L
.66387 .57915 0 -2.5 Mabsadd L
.66387 .57915 -1.875 0 Mabsadd L closepath F
.70035 .52858 -1.875 0 Mabsadd m
.70035 .52858 0 2.5 Mabsadd L
.70035 .52858 1.875 0 Mabsadd L
.70035 .52858 0 -2.5 Mabsadd L
.70035 .52858 -1.875 0 Mabsadd L closepath F
.70829 .53096 -1.875 0 Mabsadd m
.70829 .53096 0 2.5 Mabsadd L
.70829 .53096 1.875 0 Mabsadd L
.70829 .53096 0 -2.5 Mabsadd L
.70829 .53096 -1.875 0 Mabsadd L closepath F
.71711 .54864 -1.875 0 Mabsadd m
.71711 .54864 0 2.5 Mabsadd L
.71711 .54864 1.875 0 Mabsadd L
.71711 .54864 0 -2.5 Mabsadd L
.71711 .54864 -1.875 0 Mabsadd L closepath F
.73234 .52851 -1.875 0 Mabsadd m
.73234 .52851 0 2.5 Mabsadd L
.73234 .52851 1.875 0 Mabsadd L
.73234 .52851 0 -2.5 Mabsadd L
.73234 .52851 -1.875 0 Mabsadd L closepath F
.74536 .5102 -1.875 0 Mabsadd m
.74536 .5102 0 2.5 Mabsadd L
.74536 .5102 1.875 0 Mabsadd L
.74536 .5102 0 -2.5 Mabsadd L
.74536 .5102 -1.875 0 Mabsadd L closepath F
.78476 .49956 -1.875 0 Mabsadd m
.78476 .49956 0 2.5 Mabsadd L
.78476 .49956 1.875 0 Mabsadd L
.78476 .49956 0 -2.5 Mabsadd L
.78476 .49956 -1.875 0 Mabsadd L closepath F
.77637 .49947 -1.875 0 Mabsadd m
.77637 .49947 0 2.5 Mabsadd L
.77637 .49947 1.875 0 Mabsadd L
.77637 .49947 0 -2.5 Mabsadd L
.77637 .49947 -1.875 0 Mabsadd L closepath F
.86206 .50269 -1.875 0 Mabsadd m
.86206 .50269 0 2.5 Mabsadd L
.86206 .50269 1.875 0 Mabsadd L
.86206 .50269 0 -2.5 Mabsadd L
.86206 .50269 -1.875 0 Mabsadd L closepath F
.83225 .45603 -1.875 0 Mabsadd m
.83225 .45603 0 2.5 Mabsadd L
.83225 .45603 1.875 0 Mabsadd L
.83225 .45603 0 -2.5 Mabsadd L
.83225 .45603 -1.875 0 Mabsadd L closepath F
.90861 .44669 -1.875 0 Mabsadd m
.90861 .44669 0 2.5 Mabsadd L
.90861 .44669 1.875 0 Mabsadd L
.90861 .44669 0 -2.5 Mabsadd L
.90861 .44669 -1.875 0 Mabsadd L closepath F
.89033 .41526 -1.875 0 Mabsadd m
.89033 .41526 0 2.5 Mabsadd L
.89033 .41526 1.875 0 Mabsadd L
.89033 .41526 0 -2.5 Mabsadd L
.89033 .41526 -1.875 0 Mabsadd L closepath F
.95297 .40071 -1.875 0 Mabsadd m
.95297 .40071 0 2.5 Mabsadd L
.95297 .40071 1.875 0 Mabsadd L
.95297 .40071 0 -2.5 Mabsadd L
.95297 .40071 -1.875 0 Mabsadd L closepath F
.5 Mabswid
.03753 .02813 m
.07521 .09251 L
.1163 .15787 L
.15488 .21465 L
.19199 .26504 L
.23149 .31415 L
.26952 .35699 L
.30993 .39778 L
.34887 .43245 L
.38633 .46151 L
.42618 .48781 L
.46455 .50863 L
.50145 .52449 L
.52196 .53155 L
.54073 .53689 L
.56011 .54131 L
.57854 .54446 L
.58789 .54567 L
.59785 .54667 L
.6027 .54705 L
.60726 .54734 L
.61135 .54755 L
.61578 .54772 L
.6181 .54779 L
.61937 .54782 L
.62058 .54784 L
.6217 .54786 L
.62292 .54787 L
.62403 .54788 L
.62505 .54789 L
.62626 .54789 L
.6274 .54789 L
.62864 .54788 L
.62934 .54788 L
.62999 .54787 L
.6312 .54786 L
.63254 .54783 L
.63523 .54777 L
.63792 .54769 L
.64037 .5476 L
.64587 .54732 L
.65096 .54699 L
.65577 .5466 L
.66525 .54563 L
.67528 .54431 L
.69328 .54118 L
.71236 .53681 L
.73022 .53173 L
.76808 .51779 L
.80833 .49827 L
Mistroke
.8471 .47486 L
.8844 .4481 L
.92408 .41504 L
.96229 .37875 L
.96623 .37476 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00m03oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo]03oool00?H0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;<0oooo003g0?ooo`040;mdM02LW4P004RL0=oookD0
oooo003g0?ooo`030?oo_`1d000007Bo0;H0oooo003g0?ooo`030?oo_`1d0000M;oo0;H0oooo003g
0?ooo`040?noM000000004RL0=oookD0oooo003g0?ooo`050;md0000B9`0gmnL04Q8W03Oool0]03o
ool00?D0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0/P3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool005/0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`130?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
ocl0oooo00@0_g@0000000000000M;ooA@3oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooom20?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool0503oool005d0oooo00<0okmd0000001d_ol0A@3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_ch0oooo00H0gia8001d_`3oool0omnL04P00018W=m30?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
A03oool00`3o_g@0000007Boo`0G0?ooo`00G@3oool00`3ooml0W4P0001d_`140?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0k0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0@@3oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0@P3oool00`3ooml0W4P0001d_`0G0?ooo`00GP3oool0
0`3OW4P004RL0=ooo`130?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0k0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
@03oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0@`3oool00`3OW4P004RL0=ooo`0F0?ooo`00CP3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0=P3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03/0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`100?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`130?ooo`030?oOW01804P0W=oo
01H0oooo001N0?ooo`030?oo_`1d0000M;oo04@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03`0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`110?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`130?ooo`030?oo_`1d0000M;oo01H0oooo001O0?ooo`030;md0000M;l0oooo04<0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03`0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`110?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`140?ooo`030;md0000M;l0oooo01D0oooo
001O0?ooo`030=nLB000B9`0gooo04<0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo?03oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo0400oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_oooA03oool00`3OW4P004RL0=ooo`0E0?ooo`00G`3oool00`3o
gi`0B01809cOo`130?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03d0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`100?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo04D0oooo00<0omnL04P0B02Lgol05@3oool005l0oooo00<0oono07@0001d
_ol0A03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0?`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo0440oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo04H0oooo00<0oono07@0001d
_ol05@3oool00600oooo00<0_g@0001d_`3oool0A03oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0@03oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgd<0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo04L0oooo00<0_g@0001d_`3o
ool0503oool005h0oooo00D0oono07@000000000001d0;ooo`150?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0n0?ooo`040=nLB00000000000001d_dD0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO0480oooo00D0oono07@00000
0000001d0;ooo`0D0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00C03ooooo000005X000002`3oool004`0oooo00<000000?ooo`3oool02P3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo0`00002Z
0?ooo`030000003oool0oooo0:L0oooo0P00000;0?ooo`00C03oool00`000000oooo0?ooo`0;0?oo
o`H00000V@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`03
0000003oool0oooo00h0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`2X0?ooo`03
0000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
V@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`2X0?ooo`030000003o
ool0oooo00T0oooo001<0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0V@3oool0
0`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo
00X0oooo00D000000?ooo`3oool0oooo0000002K0?ooo`030000003oool0oooo0:P0oooo00<00000
0?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool02`3oool010000000oooo0?ooo`00002K
0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo0`00000;0?oo
o`040000003oool0oooo000009/0oooo00<000000?ooo`3oool0Y`3oool2000000/0oooo001<0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool3000000L0oooo00<000000?oo
o`3oool0T@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool4000000H0oooo00<000000?ooo`3o
ool0T@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000
003oool0oooo00H0oooo4@00002C0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0
2@3oool004`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`030?ooo`@000001P3o
ool00`000000oooo0?ooo`2A0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3o
ool004`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`8000001`3oool0
0`000000oooo0?ooo`2A0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool0
04`0oooo00<000000?ooo`3oool03P3oool2000009X0oooo00<000000?ooo`3oool0Z03oool00`00
0000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0>0?ooo`800000VP3oool00`000000
oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`000000V@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo
001<0?ooo`<000003P3oool00`000000oooo0000002I0?ooo`030000003oool0oooo0:L0oooo0P00
000;0?ooo`00C03oool00`000000oooo0?ooo`0>0?ooo`040000003oool0oooo000009P0oooo00<0
00000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0>
0?ooo`050000003oool0oooo0?ooo`000000U`3oool00`000000oooo0?ooo`2X0?ooo`030000003o
ool0oooo00T0oooo001<0?ooo`030000003oool0oooo00h0oooo00D000000?ooo`3oool0oooo0000
00040?ooo`D00000SP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<
0?ooo`030000003oool0oooo00/0oooo1P0000020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0SP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2>0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3o
ool0503oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo08h0oooo00<000000?ooo`3o
ool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0E0?ooo`050000
003oool0oooo0?ooo`000000T03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0
oooo001<0?ooo`<000005@3oool01@000000oooo0?ooo`3oool000000900oooo00<000000?ooo`3o
ool0Y`3oool2000000/0oooo001<0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
0P3oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo08H0oooo00<000000?ooo`3o
ool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`0000000P0000060?ooo`030000003oool0oooo08H0oooo
00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00<`3oool00`3ookl0M0000000
00020000000304RLg`3oool0oooo00<0oooo00@0_g@0000000000000M;oo2P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo00<0oooo1@0000050?ooo`030000003oool0oooo08H0oooo
00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00=@3oool00`3o_g@0000007Bo
o`050?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo0?ooo`0A0?oooa40
0000R03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000e0?ooo`030?oo
g`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool00`3oool4000000H0oooo00<000000?ooo`3oool0
QP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000f0?ooo`030=nLB000
B9`0gooo00<0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool70?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool0103oool2000000L0oooo00<000000?ooo`3oool0
QP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000V0?ooo`030?oog`2L
B000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`030?ooo`030?oog`2L
B00007Bo00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`0I0?ooo`800000S`3o
ool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000f0?ooo`030?oo_`1d0000
M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`070?ooo`@00000603oool2
000008l0oooo00<000000?ooo`3oool0YP3oool3000000/0oooo000g0?ooo`030;md0000M;l0oooo
00<0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`070?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`000000SP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo
00T0oooo000g0?ooo`030=nLB000B9`0gooo00<0oooo00<0omnL04P00018W=l00`3oool00`2oM000
04RL0=ooo`060?ooo`030000003oool0oooo01T0oooo00<000000?ooo`000000SP3oool00`000000
oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000g0?ooo`030?oOW01804P0W=oo00<0oooo
00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030000003oool0oooo01L0
oooo1P00002=0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003L0oooo
00<0oono07@0001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00H0oooo
00<000000?ooo`3oool05`3oool5000000030?ooo`000000oooo08/0oooo00<000000?ooo`3oool0
Z03oool00`000000oooo0?ooo`090?ooo`00>03oool00`2oM00007Bo0?ooo`040?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`0I0?ooo`050000003oool0oooo0?oo
o`000000S03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000f0?ooo`05
0?oo_`1d000000000000M02oool01@3oool0103OW4P0000000000000M;l80?ooo`030000003oool0
oooo01T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`290?ooo`030000003oool0
oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool0Z03oool00`00
0000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00@0
oooo0P0000050?ooo`030000003oool0000008P0oooo00<000000?ooo`3oool0Z03oool00`000000
oooo0?ooo`090?ooo`00C03oool300000140oooo00<000000?ooo`3oool00`3oool4000000D0oooo
0P0000280?ooo`030000003oool0oooo0:L0oooo0P00000;0?ooo`00C03oool00`000000oooo0?oo
o`0A0?oooa400000R03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool00`3oool5000000D0oooo00<00000
0?ooo`3oool0QP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool3000000H0oooo0P0000270?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo0P0000270?ooo`03
0000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0
6@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3oool0
Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0I0?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0P`3oool00`000000oooo0?ooo`2X0?ooo`030000003o
ool0oooo00T0oooo001<0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`230?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool0
04`0oooo0`00000I0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0PP3oool00`00
0000oooo0?ooo`2W0?ooo`8000002`3oool004`0oooo00<000000?ooo`3oool05`3oool5000000X0
oooo00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0
oooo001<0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0P@3oool00`000000oooo
0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo02L0oooo00<0
00000?ooo`3oool0P03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<
0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0O`3oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo02P0oooo1P00001l0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3o
ool0:@3oool00`000000oooo0000001n0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool02@3oool004`0oooo00<000000?ooo`3oool0:P3oool2000007h0oooo00<000000?ooo`3oool0
Z03oool00`000000oooo0?ooo`090?ooo`00C03oool3000002<0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`050?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0Y`3o
ool2000000/0oooo001<0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0103oool3
000000H0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo00T0oooo001<0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool00`3oool50000
00D0oooo00<000000?ooo`3oool0M03oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo
00T0oooo001<0?ooo`030000003oool0oooo02<0oooo4@00001f0?ooo`030000003oool0oooo0:P0
oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool08`3oool00`000000oooo
0?ooo`030?ooo`D000001@3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo0:P0oooo
00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`040?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo07@0oooo00<0
00000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0[
0?ooo`050000003oool0oooo0?ooo`0000003P3oool6000006H0oooo00<000000?ooo`3oool0Z03o
ool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0[0?ooo`050000003oool0
oooo0?ooo`000000403oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo0:P0oooo00<0
00000?ooo`3oool02@3oool004`0oooo0`00000Y0?ooo`D000000P3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool0Y`3oool2000000/0oooo001<0?oo
o`030000003oool0oooo0340oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`<000001P3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0oooo0:P0oooo00<000000?oo
o`3oool02@3oool004`0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00<0oooo1@0000050?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3o
ool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`0b0?ooo`030000
003oool0oooo00<0oooo4@00001Q0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool0
2@3oool003`0oooo00H0omnL04P00000000000000018W03Oool:0?ooo`030000003oool0oooo03<0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`D000001@3oool00`000000
oooo0?ooo`1O0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003`0oooo
00H0okmd0000001d_ol0oooo0?oOW018M;l:0?ooo`030000003oool0oooo03@0oooo00D000000?oo
o`3oool0oooo000000060?ooo`<000001P3oool00`000000oooo0?ooo`1O0?ooo`030000003oool0
oooo0:P0oooo00<000000?ooo`3oool02@3oool00440oooo00<0okmd04RLg`3oool0203oool00`00
0000oooo0?ooo`0d0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0I`3oool00`00
0000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo00110?ooo`030?oo_`1d07@0_ooo00P0
oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo06L0
oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00;`3oool00`3ooml0W4P0
000000070000000304RLg`3oool0oooo00D0oooo00<0oooO09a8001d_ol0203oool00`000000oooo
0?ooo`0f0?ooo`030000003oool0oooo00D0oooo1P00001V0?ooo`030000003oool0oooo0:P0oooo
00<000000?ooo`3oool02@3oool00440oooo00<0oono07@00018W=l0203oool4000002d0oooo1@00
00040?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0YP3oool3000000/0oooo0011
0?ooo`030?noM0000000M;oo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`09
0?ooo`00?P3oool01P3ogml0gooo0?noM0000000001d0;ooo`P0oooo00<000000?ooo`3oool0:03o
ool00`000000oooo0?ooo`040?ooo`<000001P3oool00`000000oooo0?ooo`1_0?ooo`030000003o
ool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003h0oooo00D0oono07@000000000001809cO
o`090?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool00`3oool5000000D0oooo0P00
001`0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003l0oooo00<0gibL
0;ooo`3oool02P3oool00`000000oooo0?ooo`0X0?oooa800000L03oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00T0oooo000o0?ooo`030?oOW01dW=l0oooo00X0oooo00<000000?oo
o`3oool0:03oool00`000000oooo0?ooo`030?ooo`D000001@3oool00`000000oooo0000001_0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool00400oooo00D0_g@000000000
0000000004RLg`070?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0103oool30000
00H0oooo00@000000?ooo`3oool00000KP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0
oooo00T0oooo001<0?ooo`030000003oool0oooo02h0oooo1@0000080?ooo`030000003oool0oooo
06`0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool3000003`0
oooo00<000000?ooo`3oool0J`3oool00`000000oooo0?ooo`2W0?ooo`8000002`3oool004`0oooo
00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo0:P0oooo
00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?oo
o`1Y0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<00000
0?ooo`3oool0?P3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo0:P0oooo00<00000
0?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`1X0?oo
o`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool02@3oool004`0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0600oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0000000<0oooo1@0000050?ooo`030000003o
ool0oooo05P0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool3
000003l0oooo00@000000?ooo`3oool000000P3oool5000000D0oooo00<000000?ooo`3oool0F03o
ool00`000000oooo0?ooo`2W0?ooo`8000002`3oool004`0oooo00<000000?ooo`3oool0?`3ooolA
000005X0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`00
0000oooo0?ooo`0o0?ooo`030000003oool0oooo0080oooo1P0000050?ooo`030000003oool0oooo
05P0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000
oooo0?ooo`0o0?ooo`030000003oool0oooo00<0oooo1@0000050?ooo`030000003oool0oooo05P0
oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo
0?ooo`160?ooo`800000HP3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo
001<0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?oo
o`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo04L0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo0:P0oooo00<00000
0?ooo`3oool02@3oool001d0oooo00<0_g@00018W03Oool0;03oool00`000000oooo0?ooo`180?oo
o`030000003oool0oooo00<0oooo1P0000040?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0Z03oool00`000000oooo0?ooo`090?ooo`007@3oool00`3ogi`0B7Bo0?ooo`0/0?ooo`<0
0000B03oool2000000D0oooo100000050?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3o
ool0Y`3oool2000000/0oooo000B0?ooo`030?oo_`1dM;l0oooo00L0oooo00<0okmd0018W03Oool0
;@3oool00`000000oooo0?ooo`180?oooa400000D@3oool00`000000oooo0?ooo`2X0?ooo`030000
003oool0oooo00T0oooo000B0?ooo`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3oool0
;P3oool00`000000oooo0?ooo`180?ooo`040000003oool0oooo00000080oooo1P0000040?ooo`03
0000003oool0oooo04l0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00
4`3oool00`3ogi`0B000000000050000000304RLg`3oool0oooo02h0oooo00<000000?ooo`3oool0
B03oool010000000oooo0?ooo`0000030?ooo`<000001P3oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool001P0oooo00<0okmd0000B02Lgol0
<@3oool00`000000oooo0?ooo`1<0?ooo`050000003oool0oooo0?ooo`000000F@3oool00`000000
oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000G0?ooo`030;md0000M;l0oooo0380oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0:P0oooo
00<000000?ooo`3oool02@3oool00180oooo00H0omnL04P0001d_ol0okmd0000001d_old0?ooo`03
0000003oool0oooo04h0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`2X0?ooo`03
0000003oool0oooo00T0oooo000B0?ooo`040?oOW01800000000001d_cH0oooo00<000000?ooo`3o
ool0C`3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool02@3oool004`0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`1G0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool004`0oooo0`00001@0?ooo`030000003o
ool0000005L0oooo00<000000?ooo`3oool0Y`3oool2000000/0oooo001<0?ooo`030000003oool0
oooo04X0oooo00<000000?ooo`3oool0103oool3000000H0oooo00<000000?ooo`3oool0C@3oool0
0`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo
04X0oooo00<000000?ooo`3oool00`3oool5000000D0oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000n0?ooo`040;md00000000000007Bo
o`X0oooo00<000000?ooo`3oool0BP3ooolA000004l0oooo00<000000?ooo`3oool0Z03oool00`00
0000oooo0?ooo`090?ooo`00?@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`T0oooo00<00000
0?ooo`3oool0BP3oool00`000000oooo0?ooo`030?ooo`D000001@3oool00`000000oooo0?ooo`1=
0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003`0oooo00<0oooO09a8
0018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030000003oool0oooo04X0oooo00<000000?oo
o`3oool0103oool3000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool0C@3oool00`00
0000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000l0?ooo`030?oog`2LB000B9cO0080
oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`1B0?ooo`050000003oool0oooo
0?ooo`000000D`3oool00`000000oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo000l0?oo
o`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`1G0?oo
o`030000003oool0oooo0500oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?oo
o`00?@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0ooooo`00001J000000/0
oooo000m0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01`3oool00`000000oooo
0?ooo`1H0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0Z03oool00`000000oooo
0?ooo`090?ooo`00?@3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00<0
00000?ooo`3oool0BP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo0P00
001?0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003d0oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030000003oool0oooo04X0oooo00<0
00000?ooo`3oool0103oool3000000H0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?oo
o`2X0?ooo`030000003oool0oooo00T0oooo000n0?ooo`030?oOW0180000M;oo0080oooo00<0gia8
001d_`3oool01P3oool00`000000oooo0?ooo`1:0?oooa800000CP3oool00`000000oooo0?ooo`2X
0?ooo`030000003oool0oooo00T0oooo000o0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3o
ool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo00<0oooo1P0000040?ooo`030000003o
ool0000004d0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`090?ooo`00@03oool0
103OW4P0000000000000M;l80?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool0103o
ool3000000H0oooo00@000000?ooo`3oool00000C03oool00`000000oooo0?ooo`2X0?ooo`030000
003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool01@3o
ool00`000000oooo0?ooo`050?ooo`050000003oool0oooo0?ooo`000000B`3oool00`000000oooo
0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`<00000G`3oool00`000000oooo0?oo
o`180?ooo`030000003oool0oooo0:L0oooo0P00000;0?ooo`00C03oool00`000000oooo0?ooo`1P
0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo0640oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`2K0?ooo`8000002`3oool00`000000oooo0?ooo`090?oo
o`00C03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3o
ool0VP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000
003oool0oooo0680oooo0P0000060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
=@3oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
2@3oool004`0oooo00<000000?ooo`3oool0HP3oool00`000000oooo000000020?ooo`H000001@3o
ool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`1R0?ooo`040000003oool0
oooo00000080oooo1@0000050?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0U`3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0
oooo0680oooo4@00000g0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`0000003`3oool0
0`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`1R0?ooo`030000003oool0oooo
0080oooo1P0000050?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool0T03oool00`00
0000oooo0?ooo`020?ooo`H000001@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00T0oooo001<0?ooo`<00000HP3oool00`000000oooo0?ooo`030?ooo`<000001@3oool5000000D0
oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`2@0?ooo`060000003oool0oooo0?oo
o`000000oooo1@0000050?ooo`030000003oool0oooo00@0oooo0P00000;0?ooo`00C03oool00`00
0000oooo0?ooo`1Y0?ooo`8000001@3oool5000000D0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`2@0?oooa4000001`3oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000
oooo0?ooo`1Z0?oooa400000;`3oool00`000000oooo0?ooo`2@0?ooo`030000003oool0000000<0
oooo1@0000050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02@3oool004`0oooo
00<000000?ooo`3oool0JP3oool2000000@0oooo1@0000050?ooo`030000003oool0oooo02d0oooo
00<000000?ooo`3oool0P03oool6000000X0oooo0P0000040?ooo`D000001@3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo06X0oooo00<0
00000?ooo`0000000`3oool5000000D0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`1k0?ooo`030000003oool0oooo00@0oooo0P0000070?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00T0oooo001<
0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0e
0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool00`3oool4000000H0oooo00D00000
0?ooo`3oool0oooo0000000H0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0
oooo06h0oooo1`00000e0?ooo`030000003oool0oooo07/0oooo4@0000020?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0J@3oool00`00
0000oooo0?ooo`040?ooo`<000001P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo
07/0oooo00<000000?ooo`3oool00`3oool4000000H0oooo00<000000?ooo`0000006P3oool00`00
0000oooo0?ooo`090?ooo`00C03oool3000006T0oooo00<000000?ooo`3oool0103oool4000000D0
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0080
oooo1P0000050?ooo`8000006P3oool2000000/0oooo001<0?ooo`030000003oool0oooo06T0oooo
4@00000`0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0103oool2000001`0oooo
00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?oo
o`040?ooo`@000001@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo08T0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`1Y
0?ooo`030000003oool0oooo00@0oooo0`0000000`3oool000000?ooo`030?ooo`030000003oool0
oooo02h0oooo00<000000?ooo`3oool0R03oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo00T0oooo001<0?ooo`030000003oool0oooo06l0oooo1`00000d0?ooo`030000003oool0oooo
08L0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000
oooo0?ooo`1f0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0QP3oool6000001`0
oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0M`3oool200000340oooo
00<000000?ooo`3oool0P@3oool00`000000oooo0?ooo`02000000<0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003o
ool0oooo07T0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`210?ooo`030000003o
ool0000000@0oooo0`0000060?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool02@3o
ool004`0oooo0`00001j0?ooo`030000003oool0oooo00<0oooo1P00000T0?ooo`030000003oool0
oooo06`0oooo1P00000?0?ooo`800000103oool5000000D0oooo00<000000?ooo`3oool04`3oool2
000000/0oooo001<0?ooo`030000003oool0oooo07/0oooo0P0000060?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`1_0?ooo`030000003oool0oooo
00l0oooo4@00000F0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo07/0
oooo00<000000?ooo`000000103oool2000000L0oooo00<000000?ooo`3oool0703oool00`000000
oooo0?ooo`1W0?ooo`030000003oool0oooo00@0oooo0P0000070?ooo`030000003oool0oooo00D0
oooo0`0000050?ooo`@000001P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo00T0
oooo000l0?ooo`060?oOW0180000000000000000B9`0gooo2P3oool00`000000oooo0?ooo`1k0?oo
o`040000003oool0oooo00000080oooo100000060?ooo`030000003oool0oooo01`0oooo00<00000
0?ooo`3oool0I`3oool00`000000oooo0?ooo`030?ooo`@000001P3oool00`000000oooo0?ooo`04
0?ooo`040000003oool0oooo000000H0oooo0P0000070?ooo`030000003oool0oooo01@0oooo00<0
00000?ooo`3oool02@3oool003`0oooo00H0okmd0000001d_ol0oooo0?oOW018M;l:0?ooo`030000
003oool0oooo07/0oooo4@00000N0?ooo`030000003oool0oooo06L0oooo4@0000040?ooo`800000
2@3oool5000001`0oooo00<000000?ooo`3oool02@3oool00440oooo00<0okmd04RLg`3oool0203o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo00<0oooo100000060?ooo`030000003o
ool0oooo01`0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`030?ooo`@000001P3o
ool01@000000oooo0?ooo`3oool0000002`0oooo00<000000?ooo`3oool02@3oool00440oooo00<0
oono07@0M02oool0203oool00`000000oooo0?ooo`1h0?ooo`H00000103oool2000000L0oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo00@0oooo0P00
00070?ooo`040000003oool0oooo000002d0oooo00<000000?ooo`3oool02@3oool00440oooo00<0
oooO09a8001d_ol0203oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo00<0oooo1P00
000T0?ooo`030000003oool0oooo06l0oooo00<000000?ooo`3oool01P3oool2000002h0oooo00<0
00000?ooo`3oool02@3oool00440oooo00<0oono07@00018W=l0203oool400000740oooo00<00000
0?ooo`3oool0103oool3000000H0oooo00@000000?ooo`3oool00000903oool00`000000oooo0?oo
o`1/0?ooo`H000001@3oool00`000000oooo0?ooo`0/0?ooo`<000002`3oool00440oooo00<0okmd
0000001d_ol0203oool00`000000oooo0?ooo`1b0?ooo`030000003oool0oooo00@0oooo10000005
0?ooo`040000003oool0oooo0?ooo`8000000`3oool6000000H0oooo1P00000=0?ooo`030000003o
ool0oooo07D0oooo0P00000a0?ooo`030000003oool0oooo00T0oooo000n0?ooo`060?oOg`3Oool0
okmd0000000007@0_ooo203oool00`000000oooo0?ooo`1b0?oooa4000001@3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool03@3oool00`000000oooo
0?ooo`1d0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool02@3oool003h0oooo00D0
oono07@000000000001809cOo`090?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0
103oool4000000D0oooo00<000000?ooo`3oool0103oool2000000<0oooo00D000000?ooo`3oool0
oooo000000070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`1b0?ooo`800000=03oool00`000000oooo0?ooo`090?ooo`00?`3oool00`3OW9`0_ooo
0?ooo`0:0?ooo`030000003oool0oooo0780oooo00<000000?ooo`3oool0103oool3000000H0oooo
00D000000?ooo`3oool0oooo000000040?ooo`@000000`3oool01@000000oooo0?ooo`3oool00000
0080oooo0P0000070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0L03oool20000
03H0oooo00<000000?ooo`3oool02@3oool003l0oooo00<0omnL07BLg`3oool02P3oool00`000000
oooo0?ooo`1j0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`3oool400000080
oooo00H000000?ooo`3oool0oooo0000003oool4000000H0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`1L0?ooo`H00000303oool2000003P0oooo00<000000?ooo`3oool02@3oool0
0400oooo00D0_g@0000000000000000004RLg`070?ooo`030000003oool0oooo07P0oooo1P000008
0?oooad000001`3oool00`000000oooo0?ooo`1I0?ooo`H000003P3oool00`000000oooo0?ooo`0h
0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo08H0oooo00<000000?oo
o`3oool00`3oool5000000030?ooo`000000oooo0080oooo00<000000?ooo`0000000`0000060?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0F`3oool010000000oooo0?ooo`00000<
0?ooo`8000002P3oool5000002`0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?oo
o`3oool0QP3oool00`000000oooo0?ooo`040?ooo`8000000P3oool2000000<0oooo00<000000?oo
o`3oool00P0000070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0F`3oool01000
0000oooo0?ooo`00000:0?ooo`8000003P3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0
oooo00T0oooo001<0?ooo`<00000SP3oool00`000000oooo0?ooo`020?ooo`8000001@3oool00`00
0000oooo0?ooo`0=0?ooo`030000003oool0oooo04l0oooo1@0000020?ooo`040000003oool0oooo
0?ooo`H000000`3oool00`000000oooo0?ooo`0300000100oooo00<000000?ooo`3oool0:`3oool2
000000/0oooo001<0?ooo`030000003oool0oooo08h0oooo2@0000030?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool0D@3oool00`000000oooo000000020?ooo`040000003oool0oooo
0?ooo`L000000P3oool010000000oooo0?ooo`00000:0?ooo`030000003oool0oooo00@0oooo0`00
00060?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool02@3oool004`0oooo00<00000
0?ooo`3oool0R`3oool6000000030?ooo`000000oooo00<0oooo1P00000<0?ooo`H000001@3oool6
000004@0oooo00<000000?ooo`0000004`00000:0?ooo`030000003oool0oooo00<0oooo1@000005
0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?oo
o`3oool0TP3oool00`000000oooo0?ooo`040?ooo`8000003P3oool6000000L0oooo00<000000?oo
o`3oool0?@3oool00`000000oooo0?ooo`040?ooo`80000000D0oooo0000003oool0oooo00000002
0?ooo`P0000000<0oooo0000000000000P3oool00`000000oooo0?ooo`080?oooa4000009P3oool0
0`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`2:0?ooo`030000003oool0oooo
00@0oooo0P0000070?ooo`<000003@3oool00`000000oooo000000090?ooo`030000003oool0oooo
03d0oooo00<000000?ooo`3oool00`3oool500000080oooo00<000000?ooo`3oool02@0000001@3o
ool000000?ooo`3oool0000000X0oooo00<000000?ooo`3oool00`3oool5000000D0oooo00<00000
0?ooo`3oool0903oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`2:0?oo
o`030000003oool0oooo00<0oooo100000060?ooo`030000003oool0oooo008000002`3oool00`00
0000oooo000000090?ooo`030000003oool0oooo03<0oooo1P0000040?oooa40000000@0oooo0000
003oool0oooo0P00000A0?ooo`030000003oool0oooo00@0oooo0`0000060?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0RP3ooolA0000
00@0oooo0P0000030?ooo`030000003oool0oooo00<0oooo0`0000070?ooo`030000003oool00000
03L0oooo1P0000020?ooo`/000001@3oool00`000000oooo00000003000001X0oooo00<000000?oo
o`3oool0;03oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?ooo`2:0?ooo`03
0000003oool0oooo00<0oooo100000060?ooo`030000003oool0oooo00@0oooo100000050?ooo`D0
00001@3oool4000000H0oooo00<000000?ooo`3oool0;@3oool010000000oooo0?ooo`0000040?oo
o`040000003oool0oooo000000<0oooo0`0000060?ooo`T000005@3oool5000002`0oooo00<00000
0?ooo`3oool02@3oool004`0oooo0`00002:0?ooo`030000003oool0oooo00@0oooo0P0000070?oo
o`030000003oool0oooo00L0oooo5@0000050?ooo`030000003oool0oooo02d0oooo00@000000?oo
o`3oool00000103oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00@0oooo0P00
001=0?ooo`8000002`3oool004`0oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01`3ooolK000000030?ooo`000000
000000@00000:03oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo0080oooo00D0
00000?ooo`3oool000000?ooo`0200000500oooo00<000000?ooo`3oool02@3oool004`0oooo00<0
00000?ooo`3oool0TP3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo1P00
00050?ooo`D000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0200oooo00@0
00000?ooo`3oool000000`3oool3000000030?ooo`000000oooo00<0oooo00<000000?ooo`000000
0P0000030?ooo`030000003oool0000000800000DP3oool00`000000oooo0?ooo`090?ooo`00C03o
ool00`000000oooo0?ooo`2B0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00`3o
ool5000000D0oooo100000060?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0803o
ool010000000oooo0?ooo`0000030?ooo`H000000`3oool00`000000oooo00000009000005<0oooo
00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0S`3oool6000001D0oooo00@0
00000?ooo`000000oooo0P0000060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
1@00000K0?oooa`00000D`3oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo0?oo
o`2Z0?ooo`030000003oool0000000<0oooo0`0000030?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo000001X0oooo
00<000000?ooo`3oool04P0000060?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0
2@3oool004`0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo000000060?ooo`@00000103o
ool00`000000oooo0?ooo`040?ooo`<000000`3oool010000000oooo0?ooo`00000J0?ooo`040000
003oool0oooo000000<0oooo1P0000000`3oool0000000000002000000030?ooo`000000oooo00H0
oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo
0?ooo`2Y0?ooo`H000001@3oool8000000D0oooo1@0000020?ooo`040000003oool0oooo00000180
oooo1P0000050?ooo`030000003oool0oooo0080oooo00<000000?ooo`000000100000040?ooo`03
0000003oool0oooo05T0oooo00<000000?ooo`3oool02@3oool004`0oooo0`00002Z0?ooo`030000
003oool0oooo00h0oooo4@00000E0?ooo`030000003oool0oooo00H0oooo1P0000001@3oool00000
0?ooo`3oool0000000@0oooo00@000000?ooo`3oool00000E`3oool2000000/0oooo001<0?ooo`03
0000003oool0oooo0:X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`H0
000000<0oooo0000000000001@0000020?ooo`030000003oool0oooo00h0oooo00@000000?ooo`3o
ool0oooo1@0000020?ooo`030000003oool000000080oooo00<000000?ooo`0000000P0000030?oo
o`040000003oool0oooo000005P0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?oo
o`3oool0ZP3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00@000000?oo
o`00000000000P3ooolO000000L0oooo00<000000?ooo`0000000P3oool<000005P0oooo00<00000
0?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0C0?oo
oa400000403oool00`000000oooo0?ooo`080?oooa400000F03oool00`000000oooo0?ooo`090?oo
o`00C03oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo01<0oooo00@000000?ooo`3o
ool000000P3oool6000000@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?oo
o`8000001P3oool01@000000oooo0?ooo`3oool0000000D0oooo100000060?ooo`030000003oool0
oooo05H0oooo00<000000?ooo`3oool02@3oool004`0oooo00<000000?ooo`3oool0ZP3oool00`00
0000oooo0?ooo`0C0?ooo`040000003oool0oooo000000<0oooo0`0000000`3oool000000?ooo`03
0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool00`3oool4000000D0oooo00D0
00000?ooo`3oool0oooo000000060?ooo`8000001`3oool00`000000oooo0?ooo`1F0?ooo`030000
003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool00000008000001P3oool00`000000oooo0?oo
o`030?oooa0000002`3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo00T0oooo001<
0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool04`3oool600000080oooo1P000005
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool8000000030?ooo`000000
oooo00X0oooo00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`090?ooo`00C03oool30000
0:X0oooo00<000000?ooo`3oool05P3ooolA000000D0oooo00<000000?ooo`3oool0103oool30000
00030?ooo`000000oooo0080oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1M0?oo
o`8000002`3oool004`0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo0080oooo1@0000060?ooo`030000003oool0oooo00/0oooo0P0000000`3oool0
00000?ooo`0<0?ooo`D00000GP3oool00`000000oooo0?ooo`090?ooo`00C03oool00`000000oooo
0?ooo`2Z0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool01P0000070?ooo`030000
003oool0oooo00/0oooo0P0000000`3oool000000?ooo`1_0?ooo`030000003oool0oooo00T0oooo
001<0?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00@0oooo100000050?ooo`030000003oool0oooo06P0oooo00<0
00000?ooo`3oool02@3oool003<0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`03
0?ooo`040;md00000000000007Boo`X0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`0000000`3oool5000000D0oooo00<0
00000?ooo`000000J03oool00`000000oooo0?ooo`090?ooo`00=@3oool00`3o_g@0000007Boo`05
0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO2@3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool0303ooolA000000030?ooo`000000oooo06L0oooo
00<000000?ooo`3oool02@3oool003D0oooo00<0oooO09a80000M;l0103oool00`3ooml0W4P004RL
g`020?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`0N0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`0000004000001X0?ooo`030000003o
ool0oooo00T0oooo000f0?ooo`030=nLB000B9`0gooo00<0oooo00<0oooO09a80018W=l00P3oool0
103ooml0W4P00018W03Oool70?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0703o
ool5000000`0oooo00<000000?ooo`000000103oool6000000<0oooo00<000000?ooo`000000J03o
ool00`000000oooo0?ooo`090?ooo`00=P3oool00`3ogi`0B01809cOo`030?ooo`030?oog`2LB000
07Bo00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`2Z0?ooo`030000003oool0
oooo02l0oooo00<000000?ooo`3oool00`3oool4000000H0oooo00<000000?ooo`3oool0IP3oool0
0`000000oooo0?ooo`090?ooo`00=P3oool00`3ookl0M00007Boo`040?ooo`030;md0000B9`0gooo
0080oooo00<0omnL04P00018W=l01`3oool400000:T0oooo00<000000?ooo`3oool0;`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool0I03oool30000
00/0oooo000g0?ooo`030;md0000M;l0oooo00<0oooo00<0gia80000B02Lgol00P3oool00`3ooml0
W4P0001d_`070?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0=@3oool00`000000
oooo0000001`0?ooo`030000003oool0oooo00T0oooo000g0?ooo`030=nLB000B9`0gooo00<0oooo
00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`030000003oool0oooo0:X0oooo
00<000000?ooo`3oool0=@3oool00`000000oooo0000001`0?ooo`030000003oool0oooo00T0oooo
000g0?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`060?ooo`030000003oool0oooo0:X0oooo00<000000?ooo`3oool0<`3oool500000700oooo
00<000000?ooo`3oool02@3oool003L0oooo00<0oono07@0001d_ol0103oool00`3ogi`0B00007Bo
o`020?ooo`030=nLB000M;l0oooo00H0oooo00<000000?ooo`3oool0ZP3oool00`000000oooo0?oo
o`0e0?ooo`D00000KP3oool00`000000oooo0?ooo`090?ooo`00>03oool00`2oM00007Bo0?ooo`04
0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`000000oooo0?ooo`2Z0?ooo`030000
003oool0oooo0:P0oooo00<000000?ooo`3oool02@3oool003H0oooo00D0oono07@000000000001d
0;ooo`050?ooo`040=nLB00000000000001d_`P0oooo00<000000?ooo`3oool0ZP3oool00`000000
oooo0?ooo`2X0?ooo`030000003oool0oooo00T0oooo001<0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo04d0oooo
00<000000?ooo`3oool0C@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
001<0?ooo`<000002P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool01`3oool2000000/0oooo001<0?ooo`030000003oool0oooo00X0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool02@3oool004`0ooooo`00001J000000/0oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.56275, -20.7165, \
0.00939305, 0.181318}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["GEp test", "Subsection"],

Cell[TextData[{
  "Finally, we reproduce the previous fit to ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep\)]],
  " to illustrate the use of ",
  StyleBox["Curfit", "InlineInput"],
  " for a nonlinear model."
}], "Text"],

Cell[BoxData[
    \(\(gepdata = Import["\<gep.dat\>"];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(pade[
          Q2_] = \(1 + a[1]\ Q2\)\/\(1 + b[1]\ Q2\  + \ b[2]\ Q2\^2 + b[3]\ \
Q2\^3\);\)\), "\[IndentingNewLine]", 
    \(gepresults = 
      Curfit[gepdata, pade[Q2], 
        Q2, {{a[1], 0}, {b[1], 1.4}, {b[2], 0}, {b[3], 0}}]\)}], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {a[1] \[Rule] \(-0.06741957500379343`\), 
          b[1] \[Rule] 3.1171437164045246`, b[2] \[Rule] 1.0340714229621064`, 
          b[3] \[Rule] 0.5029938181512407`}, 
      Uncertainties \[Rule] {0.035017685288137916`, 0.053168339929658165`, 
          0.09031904317044523`, 0.15572094294077274`}, 
      ReducedChiSquare \[Rule] 0.8454721516463208`, 
      ChiSquare \[Rule] 40.5826632790234`, NITS \[Rule] 9, 
      Correlations \[Rule] {{1.`, 0.851549757443654`, 0.20352612016955865`, 
            0.9543774694839172`}, {0.8515497574436542`, 
            1.`, \(-0.18219252454055435`\), 
            0.8880368351329994`}, {0.20352612016955932`, \
\(-0.1821925245405533`\), 
            1.`, \(-0.06266072955377126`\)}, {0.9543774694839169`, 
            0.8880368351329988`, \(-0.06266072955377187`\), 
            1.0000000000000002`}}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\({x, y, dy} = Transpose[gepdata];\)\), "\[IndentingNewLine]", 
    \(\(y = y\ \((1 + x\/0.71)\)\^2;\)\), "\[IndentingNewLine]", 
    \(\(dy = dy\ \((1 + x\/0.71)\)\^2;\)\), "\[IndentingNewLine]", 
    \(\(ratiodata = Transpose[{x, y, dy}];\)\), "\[IndentingNewLine]", 
    \(\(fit[Q2_] = 
        pade[Q2] \((1 + Q2\/0.71)\)\^2 /. \((Parameters /. 
              gepresults)\);\)\), "\[IndentingNewLine]", 
    \(\(DisplayTogether[
        Plot[fit[Q2], {Q2, 0, 10}, PlotRange \[Rule] {{0, 10}, {0, 1.2}}, 
          Frame \[Rule] True, 
          FrameLabel \[Rule] {\*"\"\<\!\(Q\^2\)\>\"", \
\*"\"\<\!\(G\_Ep\)/\!\(G\_D\)\>\""}], \[IndentingNewLine]ErrorListPlot[
          ratiodata]];\)\)}], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
-2.77556e-017 0.1 -2.18738e-017 0.515028 [
[.2 -0.0125 -5 -15.6875 ]
[.2 -0.0125 5 0 ]
[.4 -0.0125 -5 -15.6875 ]
[.4 -0.0125 5 0 ]
[.6 -0.0125 -5.125 -15.6875 ]
[.6 -0.0125 5.125 0 ]
[.8 -0.0125 -5 -15.6875 ]
[.8 -0.0125 5 0 ]
[1 -0.0125 -8 -15.6875 ]
[1 -0.0125 8 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -8.71875 -31.375 ]
[.5 -0.0125 8.71875 -15.6875 ]
[-0.0125 .10301 -19 -7.84375 ]
[-0.0125 .10301 0 7.84375 ]
[-0.0125 .20601 -19 -7.84375 ]
[-0.0125 .20601 0 7.84375 ]
[-0.0125 .30902 -19.25 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .41202 -19 -7.84375 ]
[-0.0125 .41202 0 7.84375 ]
[-0.0125 .51503 -10 -7.84375 ]
[-0.0125 .51503 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -35.6875 -20.9375 ]
[-0.01875 .30902 -19.25 20.9375 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2 0 m
.2 .00625 L
s
gsave
.2 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.4 0 m
.4 .00625 L
s
gsave
.4 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.6 0 m
.6 .00625 L
s
gsave
.6 -0.0125 -66.125 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.8 0 m
.8 .00625 L
s
gsave
.8 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
1 0 m
1 .00625 L
s
gsave
1 -0.0125 -69 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.05 0 m
.05 .00375 L
s
.1 0 m
.1 .00375 L
s
.15 0 m
.15 .00375 L
s
.25 0 m
.25 .00375 L
s
.3 0 m
.3 .00375 L
s
.35 0 m
.35 .00375 L
s
.45 0 m
.45 .00375 L
s
.5 0 m
.5 .00375 L
s
.55 0 m
.55 .00375 L
s
.65 0 m
.65 .00375 L
s
.7 0 m
.7 .00375 L
s
.75 0 m
.75 .00375 L
s
.85 0 m
.85 .00375 L
s
.9 0 m
.9 .00375 L
s
.95 0 m
.95 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -69.7188 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Q) show
71.625 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
76.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .10301 m
.00625 .10301 L
s
gsave
-0.0125 .10301 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .20601 m
.00625 .20601 L
s
gsave
-0.0125 .20601 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .41202 m
.00625 .41202 L
s
gsave
-0.0125 .41202 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .51503 m
.00625 .51503 L
s
gsave
-0.0125 .51503 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .02575 m
.00375 .02575 L
s
0 .0515 m
.00375 .0515 L
s
0 .07725 m
.00375 .07725 L
s
0 .12876 m
.00375 .12876 L
s
0 .15451 m
.00375 .15451 L
s
0 .18026 m
.00375 .18026 L
s
0 .23176 m
.00375 .23176 L
s
0 .25751 m
.00375 .25751 L
s
0 .28327 m
.00375 .28327 L
s
0 .33477 m
.00375 .33477 L
s
0 .36052 m
.00375 .36052 L
s
0 .38627 m
.00375 .38627 L
s
0 .43777 m
.00375 .43777 L
s
0 .46353 m
.00375 .46353 L
s
0 .48928 m
.00375 .48928 L
s
0 .54078 m
.00375 .54078 L
s
0 .56653 m
.00375 .56653 L
s
0 .59228 m
.00375 .59228 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -109.406 -12.2188 Mabsadd m
1 1 Mabs scale
currentpoint translate
81.9375 12.2188 translate
90 rotate
-81.9375 -12.2188 translate
0 24.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
71.688 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Ep) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
81.625 15.813 moveto
(\\220) show
85.438 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
94.250 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(D) show
100.875 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .61178 m
0 .61803 L
s
.2 .61178 m
.2 .61803 L
s
.4 .61178 m
.4 .61803 L
s
.6 .61178 m
.6 .61803 L
s
.8 .61178 m
.8 .61803 L
s
.125 Mabswid
.05 .61428 m
.05 .61803 L
s
.1 .61428 m
.1 .61803 L
s
.15 .61428 m
.15 .61803 L
s
.25 .61428 m
.25 .61803 L
s
.3 .61428 m
.3 .61803 L
s
.35 .61428 m
.35 .61803 L
s
.45 .61428 m
.45 .61803 L
s
.5 .61428 m
.5 .61803 L
s
.55 .61428 m
.55 .61803 L
s
.65 .61428 m
.65 .61803 L
s
.7 .61428 m
.7 .61803 L
s
.75 .61428 m
.75 .61803 L
s
.85 .61428 m
.85 .61803 L
s
.9 .61428 m
.9 .61803 L
s
.95 .61428 m
.95 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 0 m
1 0 L
s
.99375 .10301 m
1 .10301 L
s
.99375 .20601 m
1 .20601 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .41202 m
1 .41202 L
s
.99375 .51503 m
1 .51503 L
s
.125 Mabswid
.99625 .02575 m
1 .02575 L
s
.99625 .0515 m
1 .0515 L
s
.99625 .07725 m
1 .07725 L
s
.99625 .12876 m
1 .12876 L
s
.99625 .15451 m
1 .15451 L
s
.99625 .18026 m
1 .18026 L
s
.99625 .23176 m
1 .23176 L
s
.99625 .25751 m
1 .25751 L
s
.99625 .28327 m
1 .28327 L
s
.99625 .33477 m
1 .33477 L
s
.99625 .36052 m
1 .36052 L
s
.99625 .38627 m
1 .38627 L
s
.99625 .43777 m
1 .43777 L
s
.99625 .46353 m
1 .46353 L
s
.99625 .48928 m
1 .48928 L
s
.99625 .54078 m
1 .54078 L
s
.99625 .56653 m
1 .56653 L
s
.99625 .59228 m
1 .59228 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
0 .51503 m
.00236 .51107 L
.00488 .5078 L
.00726 .50544 L
.00942 .50379 L
.01197 .50234 L
.01341 .50171 L
.01473 .50124 L
.01717 .50059 L
.0184 .50036 L
.01975 .50017 L
.02113 .50003 L
.02245 .49995 L
.02364 .49991 L
.02493 .4999 L
.02566 .49991 L
.02632 .49993 L
.02709 .49995 L
.02782 .49998 L
.0291 .50006 L
.0305 .50015 L
.04057 .50113 L
.04336 .50139 L
.04636 .50165 L
.04914 .50184 L
.05051 .50191 L
.05177 .50197 L
.05319 .50202 L
.05393 .50204 L
.05472 .50206 L
.056 .50207 L
.05673 .50207 L
.05741 .50207 L
.05817 .50206 L
.05897 .50205 L
.05965 .50203 L
.0604 .50201 L
.06206 .50194 L
.06361 .50185 L
.06637 .50163 L
.06898 .50135 L
.07386 .50064 L
.07911 .4996 L
.08481 .49816 L
.09475 .49486 L
.10568 .49018 L
.12777 .47783 L
.17013 .4467 L
.2109 .41231 L
.25424 .37508 L
Mistroke
.29598 .34069 L
.33614 .30988 L
.37887 .27987 L
.42002 .25368 L
.46373 .22856 L
.50586 .20679 L
.54639 .18784 L
.5895 .16961 L
.63101 .1537 L
.67094 .13976 L
.71344 .12621 L
.75435 .11428 L
.79783 .10267 L
.83972 .09239 L
.88003 .08327 L
.9229 .07429 L
.96419 .06628 L
1 .05979 L
Mfstroke
.002 w
.00054 .5144 m
.00054 .51667 L
s
.015 w
.00054 .51554 Mdot
.002 w
.00078 .51284 m
.00078 .51541 L
s
.015 w
.00078 .51413 Mdot
.002 w
.00097 .5134 m
.00097 .51566 L
s
.015 w
.00097 .51453 Mdot
.002 w
.00117 .50966 m
.00117 .51407 L
s
.015 w
.00117 .51187 Mdot
.002 w
.00136 .51306 m
.00136 .51532 L
s
.015 w
.00136 .51419 Mdot
.002 w
.00156 .50904 m
.00156 .51417 L
s
.015 w
.00156 .51161 Mdot
.002 w
.00175 .51088 m
.00175 .51251 L
s
.015 w
.00175 .51169 Mdot
.002 w
.00195 .51102 m
.00195 .51419 L
s
.015 w
.00195 .51261 Mdot
.002 w
.00214 .51036 m
.00214 .512 L
s
.015 w
.00214 .51118 Mdot
.002 w
.00234 .51063 m
.00234 .51371 L
s
.015 w
.00234 .51217 Mdot
.002 w
.00253 .50982 m
.00253 .51165 L
s
.015 w
.00253 .51074 Mdot
.002 w
.00273 .50724 m
.00273 .5108 L
s
.015 w
.00273 .50902 Mdot
.002 w
.00312 .50738 m
.00312 .51013 L
s
.015 w
.00312 .50875 Mdot
.002 w
.00331 .50855 m
.00331 .51181 L
s
.015 w
.00331 .51018 Mdot
.002 w
.0035 .50976 m
.0035 .51355 L
s
.015 w
.0035 .51166 Mdot
.002 w
.00389 .50243 m
.00389 .51053 L
s
.015 w
.00389 .50648 Mdot
.002 w
.00467 .50224 m
.00467 .5083 L
s
.015 w
.00467 .50527 Mdot
.002 w
.00545 .50351 m
.00545 .50756 L
s
.015 w
.00545 .50553 Mdot
.002 w
.013 .49459 m
.013 .50602 L
s
.015 w
.013 .5003 Mdot
.002 w
.019 .4974 m
.019 .5056 L
s
.015 w
.019 .5015 Mdot
.002 w
.027 .48672 m
.027 .50038 L
s
.015 w
.027 .49355 Mdot
.002 w
.033 .48851 m
.033 .4994 L
s
.015 w
.033 .49396 Mdot
.002 w
.039 .50568 m
.039 .51792 L
s
.015 w
.039 .5118 Mdot
.002 w
.045 .50252 m
.045 .52856 L
s
.015 w
.045 .51554 Mdot
.002 w
.053 .49594 m
.053 .53145 L
s
.015 w
.053 .5137 Mdot
.002 w
.058 .49263 m
.058 .50708 L
s
.015 w
.058 .49985 Mdot
.002 w
.065 .47904 m
.065 .5225 L
s
.015 w
.065 .50077 Mdot
.002 w
.072 .52816 m
.072 .60003 L
s
.015 w
.072 .56409 Mdot
.002 w
.078 .47758 m
.078 .50684 L
s
.015 w
.078 .49221 Mdot
.002 w
.094 .52387 m
.094 .56648 L
s
.015 w
.094 .54518 Mdot
.002 w
.11 .45531 m
.11 .48858 L
s
.015 w
.11 .47194 Mdot
.002 w
.135 .47145 m
.135 .51706 L
s
.015 w
.135 .49426 Mdot
.002 w
.175 .40139 m
.175 .48028 L
s
.015 w
.175 .44083 Mdot
.002 w
.0373 .49251 m
.0373 .52127 L
s
.015 w
.0373 .50689 Mdot
.002 w
.0401 .49939 m
.0401 .52965 L
s
.015 w
.0401 .51452 Mdot
.002 w
.0441 .46155 m
.0441 .49404 L
s
.015 w
.0441 .47779 Mdot
.002 w
.038 .45641 m
.038 .50739 L
s
.015 w
.038 .4819 Mdot
.002 w
.05 .49811 m
.05 .54897 L
s
.015 w
.05 .52354 Mdot
.002 w
.049 .49418 m
.049 .51066 L
s
.015 w
.049 .50242 Mdot
.002 w
.079 .49585 m
.079 .50872 L
s
.015 w
.079 .50228 Mdot
.002 w
.118 .47152 m
.118 .48539 L
s
.015 w
.118 .47845 Mdot
.002 w
.148 .42209 m
.148 .45345 L
s
.015 w
.148 .43777 Mdot
.002 w
.177 .42258 m
.177 .44897 L
s
.015 w
.177 .43578 Mdot
.002 w
.188 .41683 m
.188 .44287 L
s
.015 w
.188 .42985 Mdot
.002 w
.213 .39604 m
.213 .43229 L
s
.015 w
.213 .41417 Mdot
.002 w
.247 .37741 m
.247 .40221 L
s
.015 w
.247 .38981 Mdot
.002 w
.297 .32293 m
.297 .35614 L
s
.015 w
.297 .33954 Mdot
.002 w
.347 .309 m
.347 .35542 L
s
.015 w
.347 .33221 Mdot
.002 w
.35 .27179 m
.35 .354 L
s
.015 w
.35 .31289 Mdot
.002 w
.397 .23357 m
.397 .29113 L
s
.015 w
.397 .26235 Mdot
.002 w
.475 .17678 m
.475 .23446 L
s
.015 w
.475 .20562 Mdot
.002 w
.554 .0975 m
.554 .194 L
s
.015 w
.554 .14575 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00i@3oool01@3ogkl0_moo0?ooo`3oool0gia800@0000000<0001d0;oo
o`3oool0_`3oool00>D0oooo00<0oooO09a8000000000P0000001`0007@0_ooo0?ooo`3oool0oono
07@0M02oool0_`3oool00>L0oooo00<0_g@004RLg`3oool01@3oool00`3ookl0M7Bo0?ooo`2n0?oo
o`00j03oool00`3OW4P007Bo0?ooo`350?ooo`00i`3oool00`3ogi`0B00000000002000000030018
W03Oool0oooo0<40oooo003V0?ooo`040?oOW018000004RL0=ooo`80oooo00<0_g@000000018W=l0
`@3oool00>H0oooo00<0_g@00000B02Lgol0103oool00`3o_g@00000001d_`300?ooo`00i@3oool0
103ooml0W4P00018W03Oool50?ooo`030=nLB0000000M;oo0;l0oooo003U0?ooo`040?oog`2LB000
001d0;ooo`D0oooo00@0oooO09a8000007@0_ooo_P3oool00>H0oooo00<0_g@00000B02Lgol01P3o
ool00`3OW4P0000004RLg`2n0?ooo`00iP3oool00`3OW4P0000004RLg`060?ooo`040?oOW0180000
04RL0=oookd0oooo003V0?ooo`040?oo_`1d000004RL0=ooo`D0oooo00@0oooO09a80000B9`0gooo
_@3oool00>L0oooo00<0okmd00000018W=l01@3oool0103ooml0W4P00018W02WM003000000030000
M02oool0oooo0;L0oooo003X0?ooo`030=nLB0000000M;oo00@0oooo00T0oono07@00000M;l0oooo
0?noM000B9`0gooo0?oOW02L_ol0^03oool00>T0oooo00<0okmd00000000M;l00P3oool0103ooml0
W4P00000B02Lgol20?ooo`030?oOW018B9`0gooo0;T0oooo003[0?ooo`060;md0000000000000000
000007@0_ooo103oool00`3ogi`0B4RL0=ooo`2h0?ooo`00mP3oool00`3ogi`0M9cO0?ooo`2g0?oo
o`00m`3oool00`2oM4P0W=oo0?ooo`2f0?ooo`00l`3oool01P3ogml0gooo0?ooo`3oool0gia804RL
gkL0oooo003c0?ooo`060?oo_`1dB9`0gooo0?oog`2LB000B9cO]`3oool00?@0oooo00D0omnL04P0
0000000004RL0=ooo`2g0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00PP3oool00`3OW4P00000
000000030000000304RLg`3oool0oooo0400oooo00<0oooO09a80018W=l0@03oool01@3ooml0W4P0
00000000000007Bo0440oooo00H0oono07@00000000000000018W03Ooolj0?ooo`030?oo_`1d0000
00000080000000<0B9cO0?ooo`3oool00`3oool0102oM00000000000001d_ol=0?ooo`00P`3oool0
0`3o_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo0440oooo00<0gia80018W03Oool0?`3oool0
1`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool0?P3oool00`3ookl0M00004RLg`020?ooo`030=nL
B00007@0_ooo03/0oooo00<0okmd0000001d_ol01@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RL
g``0oooo00240?ooo`030?noM000M;l0oooo04D0oooo00<0omnL04P0B02Lgol0?P3oool00`3ooml0
W4P007Boo`020?ooo`030?oOW0180000B9cO03h0oooo00<0okmd0018W03Oool00`3oool00`2oM000
B9cO0?ooo`0j0?ooo`030?oog`2LB00007Bo00@0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`0;0?ooo`00Q@3oool00`3o_g@007Bo0?ooo`100?ooo`030?noM0000000000000@00000
00<004RL0=ooo`3oool0>`3oool00`3ookl0M00007Boo`030?ooo`030;md000007@0_ooo03d0oooo
00<0okmd0018W03Oool00`3oool00`2oM00007Bo0?ooo`0k0?ooo`030=nLB000B9`0gooo00<0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`00QP3oool00`3o_g@0B9cO
0?ooo`0o0?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0?@3oool00`3ooml0W4P0
001d_`030?ooo`030=nLB00004P0W=oo03d0oooo00<0oooO09a8001d_ol00P3oool00`3ookl0M000
04RLg`0l0?ooo`030?oOW01804P0W=oo00<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P00018
09cOo`0:0?ooo`00Q`3oool00`3OW4P0B9cO0?ooo`0o0?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@0
04RL0=ooo`0n0?ooo`030;md000007@0_ooo0080oooo00<0gia80000B02Lgol0?`3oool01P2oM000
001809cOg`2LB000001d0;oooc`0oooo00<0oono07@0001d_ol0103oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0oooo00280?ooo`030;md001d_ol0oooo03l0oooo00H0omnL04Q8
W03Oool0omnL04P0B02Lgoln0?ooo`080?noM0000000B9cO0?ooo`3ooml0W4P00018W03Ooom00?oo
o`050?oog`2LB00000000018W03Oool0?@3oool00`2oM00007Bo0?ooo`030?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02P3oool008P0oooo00<0omnL04P0M02oool0@03oool01@3o
_g@007Bo0?oog`2LB000B9cO03h0oooo00<0oooO09a8000000000P0000000`0007@0_ooo0?ooo`10
0?ooo`060;md000004P0W=oo0?oOW018B9`0gooo?03oool00`3OW4P004RL0=ooo`030?ooo`030?oO
W0180000B9cO00<0oooo00<0_g@00018W03Oool02@3oool008P0oooo00<0oooO09a8001d_ol0@@3o
ool0103OW4P0B9cO0;md0000M;lo0?ooo`030?oo_`1d000007Bo04<0oooo00<0omnL04P0001d_ol0
0P3oool00`3o_g@0B9cO0?ooo`0k0?ooo`030?oOW01804P0W=oo00<0oooo00@0oooO09a80000B9`0
gooo0P3oool00`3OW4P004RL0=ooo`090?ooo`00R03oool00`3ookl0M00004RLg`120?ooo`040=nL
B01dM7@0001d0;ooocl0oooo00<0omnL04P00018W=l0@P3oool00`3ogi`0B01809cOo`020?ooo`03
0?noM000B9`0gooo03/0oooo00<0oono07@0001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`03
0=nLB000M;l0oooo00T0oooo00240?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`130?oo
o`030;md000004P0W=oo0400oooo00@0oooO09a8000007@0_ooo@03oool00`3ooml0W4P004RLg`02
0?ooo`030;md0000M;l0oooo03`0oooo00<0_g@0001d_`3oool0103oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`X0oooo00250?ooo`050?oOW018000000000000001d_ol0A03oool00`3ooml0W4P0
04RLg`120?ooo`040?oo_`1d0000000004RLgcl0oooo00D0oooO09a8000000000000001d_`0l0?oo
o`050?oo_`1d000000000000M02oool01@3oool0103OW4P0000000000000M;l;0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00@P3ooooo000005d000004P3o
ool00480oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`130?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0@`3oool00`00
0000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3o
ool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?oo
o`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0
ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool200000?l0ooooF03oool300000180oooo
00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?oo
o`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0
ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003o
ool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@
0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool0403oool00480
oooo0P00003o0?oooeP0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?ooo`3o0?oooeT0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0o`3ooomI0?ooo`030000003o
ool0oooo0100oooo00120?ooo`030000003oool0oooo0?l0ooooEP3oool400000180oooo00120?oo
o`030000003oool0oooo0?l0ooooD03oool6000000<0oooo00<000000?ooo`3oool0403oool00480
oooo00<000000?ooo`3oool0o`3ooom:0?ooo`H000002@3oool00`000000oooo0?ooo`0@0?ooo`00
@P3oool00`000000oooo0?ooo`3o0?ooodH0oooo1000000?0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0?l0oooo@@3oool5000001<0oooo00<000000?ooo`3oool0403o
ool00480oooo00<000000?ooo`3oool0o`3oooll0?ooo`D00000603oool00`000000oooo0?ooo`0@
0?ooo`00@P3oool200000?l0oooo>03oool5000001/0oooo0`00000B0?ooo`00@P3oool00`000000
oooo0?ooo`3o0?oooc80oooo1@00000R0?ooo`030000003oool0oooo0100oooo00120?ooo`030000
003oool0oooo0?l0oooo;@3oool5000002L0oooo00<000000?ooo`3oool0403oool00480oooo00<0
00000?ooo`3oool0o`3ooolY0?ooo`@00000;03oool00`000000oooo0?ooo`0@0?ooo`009P3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0gia80000000000000`00
00000`18W=l0oooo0?ooo`070?ooo`030000003oool0oooo0?l0oooo903oool500000300oooo00<0
00000?ooo`3oool0403oool002D0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oO
W0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`070?ooo`030000
003oool0oooo0?l0oooo803oool4000003D0oooo00<000000?ooo`3oool0403oool002@0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?noM000M;l0oooo00/0oooo00<0
00000?ooo`3oool0o`3ooolL0?ooo`@00000>@3oool00`000000oooo0?ooo`0@0?ooo`00903oool0
0`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool02P3o
ool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo05L0oooo1000000m0?ooo`030000003o
ool0oooo0100oooo000T0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02@3oool0
0`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0D`3oool4
00000440oooo00<000000?ooo`3oool0403oool002D0oooo00<0_g@00018W03Oool00P3oool00`3o
gi`0B00004RLg`0:0?ooo`030=nLB018W=l0oooo00P0oooo0`00002m0?ooo`030000003oool0oooo
0500oooo0`0000120?ooo`@000004P3oool002D0oooo00<0gia80000B02Lgol00P3oool00`3ooml0
W4P0001d_`0;0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0_@3oool00`000000
oooo0?ooo`1<0?ooo`@00000B03oool00`000000oooo0?ooo`0@0?ooo`009@3oool00`3ogi`0B000
04RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02oool01`3oool00`000000oooo
0?ooo`2m0?ooo`030000003oool0oooo04P0oooo1000001<0?ooo`030000003oool0oooo0100oooo
000U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3ooml0W4P0
07Boo`070?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0A03oool400000500oooo
00<000000?ooo`3oool0403oool002H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?oo
o`0:0?ooo`030?oo_`1d0000B9cO00L0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`110?ooo`<00000E03oool00`000000oooo0?ooo`0@0?ooo`009`3oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00L0oooo00<000000?oo
o`3oool0_@3oool00`000000oooo0?ooo`0m0?ooo`@00000E`3oool00`000000oooo0?ooo`0@0?oo
o`00:03oool0103OW4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol0203oool0
0`000000oooo0?ooo`2m0?ooo`030000003oool0oooo03X0oooo0`00001K0?ooo`030000003oool0
oooo0100oooo00120?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0=P3oool40000
05h0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000
oooo0?ooo`0c0?ooo`<00000HP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool200000;h0oooo
00<000000?ooo`3oool0<03oool3000006<0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?oo
o`2m0?ooo`030000003oool0oooo02d0oooo0`00001X0?ooo`030000003oool0oooo0100oooo0012
0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0:P3oool3000006/0oooo00<00000
0?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0W0?oo
o`<00000KP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2l0?ooo`<0
00009@3oool300000740oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0
^`3oool500000200oooo1000001d0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0;/0oooo1@00000M0?ooo`<00000N03oool00`000000oooo0?ooo`0@0?ooo`00@P3oool0
0`000000oooo0?ooo`2k0?ooo`D000006@3oool4000007/0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool0_03oool3000001L0oooo0`00001o0?ooo`030000003oool0oooo
0100oooo00120?ooo`800000_P3oool00`000000oooo0?ooo`0D0?ooo`800000P03oool300000180
oooo00120?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool04P3oool2000008@0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?oo
o`0?0?ooo`<00000QP3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2m
0?ooo`030000003oool0oooo00d0oooo0P0000290?ooo`030000003oool0oooo0100oooo00120?oo
o`030000003oool0oooo0;d0oooo00<000000?ooo`3oool02`3oool2000008/0oooo00<000000?oo
o`3oool0403oool00480oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`080?ooo`<0
0000S@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2m0?ooo`030000
003oool0oooo00H0oooo0P00002@0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0:40oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`030?ooo`<00000TP3o
ool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0
oooo01T0oooo00@000000?ooo`3oool0oooo0P00002E0?ooo`030000003oool0oooo0100oooo0012
0?ooo`800000XP3oool00`000000oooo0?ooo`0I0?ooo`@00000U@3oool300000180oooo00120?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0603oool2000009X0oooo00<000000?oo
o`3oool0403oool00480oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0F0?ooo`80
000000<0oooo0000003oool0V@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01@0oooo0P0000030?ooo`030000003oool0oooo09P0oooo
00<000000?ooo`3oool0403oool002H0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B000
07Boo`050?ooo`030?oog`2LB000B9cO00X0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`0B0?ooo`8000001@3oool00`000000oooo0?ooo`2H0?ooo`030000003oool0oooo0100oooo
000U0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`060?ooo`03
0=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0@0?ooo`80
0000XP3oool00`000000oooo0?ooo`0@0?ooo`00903oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00X0oooo00<0omnL04P0B02Lgol02@3oool00`000000oooo0?ooo`2P0?ooo`<00000
3`3oool200000:@0oooo00<000000?ooo`3oool0403oool002@0oooo00<0oooO09a80018W=l00P3o
ool0103ooml0W4P00018W03Oool50?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0
1P3oool00`000000oooo0?ooo`2O0?ooo`D00000303oool200000:H0oooo00<000000?ooo`3oool0
403oool002@0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2L
B4P0W=oo0080oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`2O0?ooo`D000002P3o
ool200000:P0oooo00<000000?ooo`3oool0403oool002D0oooo00<0_g@00018W03Oool00P3oool0
0`3ogi`0B00004RLg`060?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`080?ooo`<00000
W`3oool5000000L0oooo0`00002W0?ooo`@000004P3oool002D0oooo00<0gia80000B02Lgol00P3o
ool00`3ooml0W4P0001d_`070?ooo`060?oOW018B9`0gooo0?oOW01804P0W=oo203oool00`000000
oooo0?ooo`2P0?ooo`<000001P3oool200000:d0oooo00<000000?ooo`3oool0403oool002D0oooo
00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?ooo`050?noM000M;l0oooO09a80018
W=l0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00<0oooo0P00002_0?ooo`03
0000003oool0oooo0100oooo000U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03O
ool0203oool0103OW4P0B9cO0;md0000M;l80?ooo`030000003oool0oooo0:40oooo00@000000?oo
o`3oool0oooo0P00002a0?ooo`030000003oool0oooo0100oooo000V0?ooo`030?oOW0180000M;oo
0080oooo00<0gia8001d_`3oool02@3oool0103OW4P0M7Ad0000M02oool70?ooo`030000003oool0
oooo0:40oooo00@000000?ooo`0000000000/`3oool00`000000oooo0?ooo`0@0?ooo`009`3oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`/0oooo00<0_g@00000B02Lgol01`3oool00`000000oooo
0?ooo`2Q0?ooo`800000]@3oool00`000000oooo0?ooo`0@0?ooo`00:03oool0103OW4P000000000
0000M;l<0?ooo`030?oog`2LB000B9cO00L0oooo00<000000?ooo`3oool0W`3oool300000;H0oooo
00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0W@3oool200000080oooo00<0
00000?ooo`3oool0]03oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000009`0oooo0P000004
0?ooo`030000003oool0oooo0;80oooo0`00000B0?ooo`003`3oool00`3OW4P00000000000020000
000304RLg`3oool0oooo02/0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0@0?oo
o`8000001P3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0100oooo000=0?ooo`03
0?noM0000000B9cO00@0oooo00<0gia8001d_`3oool0:`3oool00`000000oooo0?ooo`260?ooo`03
0000003oool0oooo00l0oooo00<000000?ooo`3oool0_@3oool00`000000oooo0?ooo`0@0?ooo`00
303oool00`3OW4P007Bo0?ooo`060?ooo`030?noM01d_ol0oooo02X0oooo00<000000?ooo`3oool0
QP3oool00`000000oooo0?ooo`0=0?ooo`800000`03oool00`000000oooo0?ooo`0@0?ooo`002`3o
ool00`3o_g@0B9cO0?ooo`080?ooo`030;nLW03Oool0oooo02T0oooo00<000000?ooo`3oool0QP3o
ool00`000000oooo0?ooo`0;0?ooo`800000`P3oool00`000000oooo0?ooo`0@0?ooo`002P3oool0
0`3ooml0W9bo0?ooo`090?ooo`030=nLW02oool0oooo02T0oooo00<000000?ooo`3oool0QP3oool0
0`000000oooo0?ooo`0:0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool0403oool0
00X0oooo00<0omnL09boo`3oool02@3oool00`3OW9`0_ooo0?ooo`0Y0?ooo`030000003oool0oooo
08H0oooo00<000000?ooo`3oool0203oool200000<D0oooo00<000000?ooo`3oool0403oool000X0
oooo00<0oknL09cOo`3oool02@3oool00`2oM7@0_ooo0?ooo`0Y0?ooo`030000003oool0oooo08H0
oooo00<000000?ooo`3oool01P3oool200000<L0oooo00<000000?ooo`3oool0403oool000X0oooo
00<0gibL0;ooo`3oool00`3oool00`3o_g@0M;oo0?ooo`020?ooo`030?oog`2LM9`0gooo02X0oooo
00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0<L0oooo
00<000000?ooo`3oool0403oool000X0oooo00<0okmd07Boo`3oool00`3oool0103o_g@0B9cO0?oo
g`2LB0020000000304RLg`3oool0oooo02T0oooo0P0000260?ooo`<00000103oool200000<P0oooo
0`00000B0?ooo`002P3oool00`3ookl0M01d0;ooo`030?ooo`060?noM000000000000000000007@0
_ooo;03oool00`000000oooo0?ooo`240?ooo`D000000P3oool00`000000oooo0?ooo`3:0?ooo`03
0000003oool0oooo0100oooo000:0?ooo`090?oo_`1d000000000000001d_ol0oooo0?noM000B9`0
gooo02l0oooo00<000000?ooo`3oool0Q03oool700000<d0oooo00<000000?ooo`3oool0403oool0
00X0oooo00<0gia80000001d_ol00`3oool00`3o_g@0M;oo0?ooo`0_0?ooo`030000003oool0oooo
08@0oooo1@00003?0?ooo`030000003oool0oooo0100oooo000I0?ooo`030;mdM02oool0oooo02H0
oooo00<000000?ooo`3oool0Q@3oool300000=00oooo00<000000?ooo`3oool0403oool001H0oooo
00D0okmd000000000000001d0;ooo`0W0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3o
ool02`3oool300000=40oooo00<000000?ooo`3oool0403oool00140oooo00<0gia8000000000000
0P0000001@00M;l0oooo0?ooo`3OW9`0_ooo02L0oooo00<000000?ooo`3oool0MP3oool00`000000
oooo0?ooo`0:0?ooo`040000003oool0oooo00000=40oooo00<000000?ooo`3oool0403oool00140
oooo00H0gia807Boo`3oool0oooO09adW03Oool20?ooo`030=nLW02oool0oooo02H0oooo00<00000
0?ooo`3oool0MP3oool00`000000oooo0?ooo`080?ooo`8000000`3oool00`000000oooo0?ooo`3?
0?ooo`030000003oool0oooo0100oooo000A0?ooo`0:0=nLW02oool0oooo0?oog`2LB000000007Bo
o`3ooml0W4Qd0;ooobL0oooo00<000000?ooo`3oool0MP3oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0c`3oool00`000000oooo0?ooo`0@0?ooo`00
4@3oool01@3OW9`0_ooo0?oo_`1d000007Bo0080oooo00<0_g@00000M02oool09`3oool2000007L0
oooo00<000000?ooo`3oool01@3oool2000000H0oooo00<000000?ooo`3oool0c@3oool300000180
oooo000A0?ooo`030=nLW02oool0oooo00<0oooo00<0gia80018W03Oool00P3oool00`2oW9`0gooo
0?ooo`0S0?ooo`030000003oool0oooo07H0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo0<l0oooo00<000000?ooo`3oool0403oool00140oooo00<0
gia800000018W=l01`3oool00`3ogi`0B4RL0=ooo`0T0?ooo`030000003oool0oooo07H0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0<l0oooo00<0
00000?ooo`3oool0403oool001@0oooo00@0_ibL0=ooo`3oool0okmd100000000`00B9`0gooo0?oo
o`0S0?ooo`030000003oool0oooo07H0oooo00@000000?ooo`3oool0oooo0P00000:0?ooo`030000
003oool0oooo0<l0oooo00<000000?ooo`3oool0403oool001@0oooo00L0_g@000000000000004RL
0=ooo`2oM7@0_ooo00X0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`030?oo
o`050?oog`2LB000000000000000M;l02`3oool00`000000oooo0?ooo`1f0?ooo`040000003oool0
oooo00000=h0oooo00<000000?ooo`3oool0403oool001@0oooo00<0oono07Ad_`3oool00P3oool0
0`3OW9`0_ooo0?ooo`080?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B000
07Boo`030?ooo`070=nLB000M;l0oooo0?oo_`1d000004RL0=ooo`090?ooo`030000003oool0oooo
07H0oooo00<000000?ooo`000000g`3oool00`000000oooo0?ooo`0@0?ooo`00503oool01P3OW9`0
_ooo0?ooo`3oool0oono07Ad_`T0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`06
0?ooo`030?oog`2LB000M;oo0080oooo00<0omnL04P00018W=l02@3oool00`000000oooo0?ooo`1f
0?ooo`800000h03oool00`000000oooo0?ooo`0@0?ooo`004`3oool01`3ooml0W7BL0=ooo`3oool0
okmd0000M02oool02@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo
00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`080?ooo`030000003oool0oooo07D0oooo
0P00003Q0?ooo`030000003oool0oooo0100oooo000D0?ooo`040=nLB00000000000001d_`/0oooo
00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo
00<0gia80000B02Lgol0203oool00`000000oooo0?ooo`1d0?ooo`030000003oool000000>40oooo
00<000000?ooo`3oool0403oool002@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RL
g`060?ooo`030;md000007@0_ooo0080oooo00<0gia80000B02Lgol0203oool300000780oooo0P00
0000103oool0000000000000003M0?ooo`@000004P3oool001P0oooo00@0oooO09a8000007@0_ooo
203oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00H0oooo00P0okmd00000018W=l0
oooo0?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`05
00000=l0oooo00<000000?ooo`3oool0403oool001D0oooo00D0okmd000000000000001809cOo`0:
0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3oool00`3ooml0W4P000000002
000000030000M02oool0oooo00P0oooo00<000000?ooo`3oool0K`3oool2000000<0oooo1@00003O
0?ooo`030000003oool0oooo0100oooo000B0?ooo`040;md000000000000001d_`h0oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d000007Bo00`0oooo00<0
00000?ooo`3oool0KP3oool00`000000oooo0?ooo`030?ooo`D00000g`3oool00`000000oooo0?oo
o`0@0?ooo`003P3oool01@3ogi`0B000000000000000M;oo0180oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`070?ooo`030?oOW0180000B9cO00/0oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`050?ooo`<00000h03oool00`000000oooo0?ooo`0@0?ooo`002`3oool0
1@3OW4P0000000000000B9`0gooo01H0oooo00H0gia80000M02oool0oooO09a80018W=l90?ooo`04
0?oog`2LB000001d0;ooo`T0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?ooo`040?oo
o`800000203oool200000>40oooo00<000000?ooo`3oool0403oool000T0oooo00<0omnL04P00018
W=l06`3oool0103OW4P0000000000000M;l<0?ooo`040?oo_`1d0000000004RLg`L0oooo00<00000
0?ooo`3oool0I03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00L0oooo0`00003Q
0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo06@0oooo00@000000?oo
o`3oool0oooo0P0000090?ooo`D00000h03oool00`000000oooo0?ooo`0@0?ooo`003`3oool00`3O
W4P00000000000020000000304RLg`3oool0oooo02/0oooo00<000000?ooo`3oool0I03oool01000
0000oooo0?ooo`00000;0?ooo`D00000h03oool00`000000oooo0?ooo`0@0?ooo`003@3oool00`3o
_g@0000004RLg`040?ooo`030=nLB000M;l0oooo02/0oooo0P00001T0?ooo`@00000303oool50000
0=h0oooo0`00000B0?ooo`00303oool00`3OW4P007Bo0?ooo`060?ooo`030?noM01d_ol0oooo02X0
oooo00<000000?ooo`3oool0HP3oool5000000d0oooo0`00003Q0?ooo`030000003oool0oooo0100
oooo000;0?ooo`030?noM018W=l0oooo00P0oooo00<0_ibL0=ooo`3oool0:@3oool00`000000oooo
0?ooo`1R0?ooo`D000003P3oool200000>40oooo00<000000?ooo`3oool0403oool000X0oooo00<0
oooO09bL_`3oool02@3oool00`3OW9`0_ooo0?ooo`0Y0?ooo`030000003oool0oooo0680oooo1@00
000>0?ooo`800000h@3oool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3ogi`0W;oo0?ooo`09
0?ooo`030=nLW02oool0oooo02T0oooo00<000000?ooo`3oool0H@3oool00`000000oooo00000002
000000l0oooo0P00003Q0?ooo`030000003oool0oooo0100oooo000:0?ooo`030?noW02Lgol0oooo
00T0oooo00<0_gAd0;ooo`3oool0:@3oool00`000000oooo0?ooo`1O0?ooo`8000000`3oool00`00
0000oooo0?ooo`0>0?ooo`800000h@3oool00`000000oooo0?ooo`0@0?ooo`002P3oool00`3OW9`0
_ooo0?ooo`030?ooo`030?noM01d_ol0oooo0080oooo00<0oooO09adW03Oool0:P3oool00`000000
oooo0?ooo`1N0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03P3oool200000>40
oooo00<000000?ooo`3oool0403oool000X0oooo00<0okmd07Boo`3oool00`3oool0103o_g@0B9cO
0?oog`2LB0020000000304RLg`3oool0oooo02T0oooo00<000000?ooo`3oool0G@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo0P00003Q0?ooo`030000003oool0oooo0100
oooo000:0?ooo`030?oo_`1d07@0_ooo00<0oooo00H0okmd00000000000000000000M02oool/0?oo
o`030000003oool0oooo05`0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3a0?oo
o`030000003oool0oooo0100oooo000:0?ooo`090?oo_`1d000000000000001d_ol0oooo0?noM000
B9`0gooo02l0oooo0P00001L0?ooo`030000003oool0oooo0?P0oooo0`00000B0?ooo`002P3oool0
0`3OW4P0000007Boo`030?ooo`030?noM01d_ol0oooo00H0oooo00<0omoO0=ooo`3oool09P3oool0
0`000000oooo0?ooo`1J0?ooo`030000003oool0oooo0?/0oooo00<000000?ooo`3oool0403oool0
01T0oooo00<0_gAd0;ooo`3oool09P3oool00`000000oooo0?ooo`1H0?ooo`800000oP3oool00`00
0000oooo0?ooo`0@0?ooo`004@3oool01P3OW9`0_ooo0?ooo`3oool0oooO09a80080000000<0001d
0;ooo`3oool09P3oool00`000000oooo0?ooo`1G0?ooo`030000003oool0oooo0?h0oooo00<00000
0?ooo`3oool0403oool00140oooo00H0gia800000000000000000000B02Lgol20?ooo`030=nLW02o
ool0oooo02H0oooo00<000000?ooo`3oool0EP3oool00`000000oooo0?ooo`3o0?ooo`030000003o
ool0oooo0100oooo000A0?ooo`030=nLW02oool0oooo00D0oooo00<0gibL0;ooo`3oool09P3oool0
0`000000oooo0?ooo`1E0?ooo`030000003oool0oooo0?l0oooo0@3oool00`000000oooo0?ooo`0@
0?ooo`004@3oool00`3OW9`0_ooo0?ooo`050?ooo`030;nLW03Oool0oooo02H0oooo00<000000?oo
o`3oool0DP3oool00`000000oooo0000003o0?ooo`@0oooo00<000000?ooo`3oool0403oool00140
oooo00<0gibL0;ooo`3oool0103oool00`3ookl0M7Bo0?ooo`0W0?ooo`030000003oool0oooo0580
oooo0P00003o0?ooo`D0oooo00<000000?ooo`3oool0403oool00140oooo00<0okmd07Boo`3oool0
103oool00`2oM4P0W=oo0?ooo`0W0?ooo`030000003oool0oooo0540oooo0`00003o0?ooo`D0oooo
00<000000?ooo`3oool0403oool00140oooo00<0oono07@0B02Lgol00P3oool00`3ookl0M00007Bo
o`0Y0?ooo`800000D@3oool500000?l0oooo0P3oool300000180oooo000B0?ooo`060?oog`2LB000
00000000000007@0_ooo:P3oool00`000000oooo0?ooo`1?0?ooo`H00000o`3oool40?ooo`030000
003oool0oooo0100oooo00120?ooo`030000003oool0oooo04h0oooo00<000000?ooo`0000001000
003o0?ooo`@0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0AP3oool0
0`000000oooo0?ooo`040?ooo`040000003oool0oooo0?ooo`<00000o`3oool50?ooo`030000003o
ool0oooo0100oooo000V0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00`3o
ool01P3ookl0M00000000000000004RL0=ooo`T0oooo00<000000?ooo`3oool0AP3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3oool40?ooo`03
0000003oool0oooo0100oooo000U0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3o
gi`0B00007Boo`020?ooo`030?oo_`1d0000B9cO0080oooo00<0gia80000M02oool0203oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0?l0oooo103oool00`000000oooo0?ooo`0@0?oo
o`00903oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0okmd0018W03O
ool00`3oool00`2oM000B9cO0?ooo`070?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`02000000L0oooo00<000000?ooo`3oool0o`3oool40?ooo`03
0000003oool0oooo0100oooo000T0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0
gooo1@3oool00`3o_g@004RL0=ooo`030?ooo`030;md0000M;l0oooo00L0oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool000000?l0oooo403oool00`000000
oooo0?ooo`0@0?ooo`00903oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo
00<0oooO09a8001d_ol00P3oool00`3ookl0M00004RLg`080?ooo`030000003oool0oooo03T0oooo
00<000000?ooo`3oool02@3oool300000?l0oooo4@3oool00`000000oooo0?ooo`0@0?ooo`009@3o
ool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo00H0_g@00000B02Lgml0W4P0
0000M02oool80?ooo`<00000>@3oool00`000000oooo0?ooo`080?ooo`D00000o`3oool=0?ooo`@0
00004P3oool002D0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`080?ooo`050?oo
g`2LB00000000018W03Oool0203oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo00P0
oooo1@00003o0?oooa00oooo00<000000?ooo`3oool0403oool002D0oooo00<0omnL04P00018W=l0
0`3oool00`2oM00004RL0=ooo`070?ooo`060;md000004P0W=oo0?oOW018B9`0gooo1`3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool50000
0?l0oooo403oool00`000000oooo0?ooo`0@0?ooo`009@3oool0103ooml0W4P00018W03Oool20?oo
o`030=nLB000B9`0gooo00H0oooo00<0omnL04P0001d_ol00P3oool00`3o_g@0B9cO0?ooo`060?oo
o`030000003oool0oooo03T0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool2
000000040?ooo`000000000000000?l0oooo4@3oool00`000000oooo0?ooo`0@0?ooo`009P3oool0
0`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00H0oooo00<0omnL04P0B02Lgol00P3oool0
0`3o_g@004RL0=ooo`060?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool01P3oool2
00000080oooo0`0000000`3oool000000?ooo`030?ooo`030000003oool0oooo0?l0oooo403oool0
0`000000oooo0?ooo`0@0?ooo`009`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00<0
oooO09a80018W=l00P3oool00`2oM00007Bo0?ooo`060?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool01P3oool2000000030?ooo`000000000000<000001@3oool00`000000oooo0?oo
o`3o0?oooa00oooo00<000000?ooo`3oool0403oool002P0oooo00@0gia800000000000007Bo2@3o
ool01@3ooml0W4P000000000000007Bo00P0oooo00<000000?ooo`3oool0<03oool00`000000oooo
0?ooo`060?ooo`P000001@3oool00`000000oooo0?ooo`3o0?oooa00oooo00<000000?ooo`3oool0
403oool00480oooo00<000000?ooo`3oool0;`3oool3000000H0oooo2@0000050?ooo`030000003o
ool0oooo0?l0oooo403oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0^
0?ooo`D000001@3oool800000?l0oooo6@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool20000
02l0oooo1@0000040?ooo`H0000000<0oooo0000003oool0o`3ooolG0?ooo`<000004P3oool00480
oooo00<000000?ooo`3oool0;P3oool5000000@0oooo1@0000020?ooo`030000003oool0oooo0?l0
oooo603oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0_0?ooo`<00000
1@3oool500000080oooo00<000000?ooo`3oool0o`3ooolH0?ooo`030000003oool0oooo0100oooo
00120?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool01@3oool300000?l0oooo7P3o
ool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0
oooo00<0oooo0P0000000`3oool000000000003o0?oooah0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo
0?ooo`000000o`3ooolO0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`03
0000003oool0oooo00X0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0<0?ooo`80
00001@3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0403oool00480oooo
00<000000?ooo`3oool02P3oool00`000000oooo0000000F0?ooo`030000003oool0oooo00/0oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o0?oooad0oooo00<000000?ooo`3oool0
403oool00480oooo0P00000;0?ooo`030000003oool0000001H0oooo00<000000?ooo`3oool02P3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0?l0oooo6`3oool300000180oooo0012
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`0000005@3oool3000000T0oooo0P000009
0?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`00
0000oooo0?ooo`0:0?ooo`030000003oool0000001@0oooo1@0000070?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`03
0000003oool0oooo00X0oooo1000000C0?ooo`H00000103oool2000000`0oooo00<000000?ooo`3o
ool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo00T0oooo
1P00000B0?ooo`L000000P3oool2000000d0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`030000
003oool0oooo0100oooo00120?ooo`030000003oool0oooo00P0oooo1`0000040?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool01`3oool6000000040?ooo`000000oooo000000d0oooo
00<000000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo00P0oooo1`0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203o
ool600000080oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3o0?oooad0oooo00<0
00000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0203oool6000000D0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`080?ooo`@00000103oool00`000000oooo0?ooo`3o
0?ooob/0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool01@3oool30000
00030?ooo`0000000000008000001P3oool010000000oooo0?ooo`3oool3000000T0oooo0P000006
0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool20000
00D0oooo1`0000000`3oool000000?ooo`050?ooo`030000003oool0oooo00D000001P3oool30000
00H0oooo0`00003o0?ooobX0oooo0`00000B0?ooo`00@P3oool00`000000oooo0?ooo`040?ooo`L0
000000<0oooo0000003oool00`3oool00`000000oooo000000020?ooo`D00000103oool2000000P0
oooo1@00003o0?ooob/0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0
2`0000000`3oool0000000000007000000030?ooo`000000000000<0000000@0oooo000000000000
00002P3oool500000?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00@P3ooolO000000d0oooo
1@00003o0?ooob/0oooo00<000000?ooo`3oool0403oool00340oooo00<0oono07@0000000000P00
00000`18W=l0oooo0?ooo`090?oooah000003`3oool300000?l0oooo;03oool00`000000oooo0?oo
o`0@0?ooo`00<`3oool00`3o_g@0000007Boo`0<0?ooo`T0000000<0oooo0000000000004P00000@
0?ooo`030000003oool0oooo0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00<`3oool00`3o
oml0W4P0001d_`0<0?ooo`P000000P3oool>00000080oooo0`00000A0?ooo`030000003oool0oooo
0?l0oooo:`3oool00`000000oooo0?ooo`0@0?ooo`00=03oool00`3OW4P004RL0=ooo`0;0?ooo`@0
00001P3oool;000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`3o0?ooob/0oooo00<000000?ooo`3oool0403oool003@0oooo00<0omnL04P0B02Lgol0
2`3oool4000000L0oooo2P0000000`3oool000000?ooo`0F0?ooo`030000003oool0oooo0?l0oooo
:`3oool00`000000oooo0?ooo`0@0?ooo`00=03oool00`3ookl0M00007Boo`0;0?ooo`@000001`3o
ool:000000030?ooo`000000oooo01H0oooo00<000000?ooo`3oool0o`3ooolX0?ooo`@000004P3o
ool003D0oooo00<0_g@0001d_`3oool02P3oool3000000P0oooo2@0000020?ooo`030000003oool0
oooo0?l0oooo@`3oool00`000000oooo0?ooo`0@0?ooo`00=@3oool00`3OW4P004RL0=ooo`0:0?oo
o`8000002P3oool800000080oooo00<000000?ooo`3oool0o`3ooom30?ooo`030000003oool0oooo
0100oooo000e0?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool02P3oool00`000000
oooo00000004000000`0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100
oooo000e0?ooo`030?oo_`1d0000M;oo00X0oooo00<000000?ooo`3oool02P3oool00`000000oooo
00000004000000@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o0?ooocT0oooo
00<000000?ooo`3oool0403oool003H0oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?oo
o`0:0?ooo`030000003oool0oooo00<000001@3oool00`000000oooo0?ooo`050?ooo`030000003o
ool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00=03oool01@3ookl0M00000000000
07@0_ooo00T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100oooo0012
0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`00
0000oooo0?ooo`0>0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool30000
0?l0oooo>P3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000000l0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`030?ooo`D00000o`3ooolg0?ooo`<000004P3oool00480oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo
1@00003o0?ooocT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool03P3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo1@00003o0?ooocT0oooo00<0
00000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo0`00003o0?ooocX0oooo00<000000?ooo`3oool0403oool0
0480oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
0?l0oooo>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0oooo0100
oooo00120?ooo`030000003oool0oooo01@0oooo0`0000060?ooo`030000003oool0oooo0?l0oooo
>@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0C0?ooo`D000001@3o
ool00`000000oooo0?ooo`3o0?ooocT0oooo00<000000?ooo`3oool0403oool00480oooo00<00000
0?ooo`3oool04`3oool5000000D0oooo00<000000?ooo`3oool0o`3oooli0?ooo`030000003oool0
oooo0100oooo00120?ooo`800000503oool5000000D0oooo00<000000?ooo`3oool0o`3ooolg0?oo
o`<000004P3oool00480oooo00<000000?ooo`3oool0503oool300000?l0oooo@P3oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo
@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?oo
o`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo@@3oool00`000000
oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0oooo
@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0?l0oooo@@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool2000001H0oooo00<00000
0?ooo`3oool0o`3ooolo0?ooo`<000004P3oool00480oooo00<000000?ooo`3oool05@3oool00`00
0000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`3o0?oood40oooo00<000000?ooo`3oool0403oool00480oooo
00<000000?ooo`3oool0o`3ooomI0?ooo`030000003oool0oooo0100oooo00120?ooo`030000003o
ool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0@0?ooo`00@P3oool00`000000oooo0?ooo`3o
0?oooeT0oooo00<000000?ooo`3oool0403oool00480oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`120?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@P3oool00`00
0000oooo0?ooo`130?ooo`030000003oool0oooo0100oooo00120?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`00@P3ooooo000005d00000
4P3oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.90205, -0.287596, \
0.0432244, 0.00839264}}]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
WindowToolbars->"EditBar",
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
ShowSelection->True,
ShowCellLabel->False,
InputAliases->{"intt"->RowBox[ {"\[Integral]", 
        RowBox[ {"\[SelectionPlaceholder]", 
          RowBox[ {"\[DifferentialD]", "\[Placeholder]"}]}]}], 
    "dintt"->RowBox[ {
        SubsuperscriptBox[ 
        "\[Integral]", "\[SelectionPlaceholder]", "\[Placeholder]"], 
        RowBox[ {"\[Placeholder]", 
          RowBox[ {"\[DifferentialD]", "\[Placeholder]"}]}]}], 
    "sumt"->RowBox[ {
        UnderoverscriptBox[ "\[Sum]", 
          RowBox[ {"\[SelectionPlaceholder]", "=", "\[Placeholder]"}], 
          "\[Placeholder]"], "\[Placeholder]"}], "prodt"->RowBox[ {
        UnderoverscriptBox[ "\[Product]", 
          RowBox[ {"\[SelectionPlaceholder]", "=", "\[Placeholder]"}], 
          "\[Placeholder]"], "\[Placeholder]"}], "dt"->RowBox[ {
        SubscriptBox[ "\[PartialD]", "\[Placeholder]"], " ", 
        "\[SelectionPlaceholder]"}], "notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {"\"\[Placeholder]\"", "\[Rule]", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"]},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Notation Package Styles", "Section",
    CellTags->"NotationPackage"],
  
  Cell["\<\
The cells below define certain styles needed by the Notation package. These \
styles serve to make visible otherwise invisible tagboxes.\
\>", "Text",
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationTemplateStyle"],
    StyleMenuListing->None,
    Background->RGBColor[1, 1, 0.850004],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationPatternWrapperStyle"],
    StyleMenuListing->None,
    Background->RGBColor[1, 0.900008, 0.979995],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"],
  
  Cell[StyleData["NotationMadeBoxesWrapperStyle"],
    StyleMenuListing->None,
    Background->RGBColor[0.900008, 0.889998, 1],
    TagBoxOptions->{SyntaxForm->"symbol"},
    CellTags->"NotationPackage"]
  }, Closed]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Info3351438211-6959969"->{
    Cell[250141, 6514, 1326, 20, 85, "Print",
      CellTags->"Info3351438211-6959969"]},
  "Info3351438212-4114851"->{
    Cell[496343, 13071, 1137, 16, 85, "Print",
      CellTags->"Info3351438212-4114851"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Info3351438211-6959969", 1153945, 31225},
  {"Info3351438212-4114851", 1154065, 31228}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 95, 2, 81, "Title"],
Cell[1874, 57, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2354, 76, 33, 0, 83, "Section"],
Cell[2390, 78, 117, 2, 99, "Input"],
Cell[2510, 82, 203, 3, 99, "Input"],
Cell[2716, 87, 179, 3, 125, "Input"],
Cell[2898, 92, 970, 27, 102, "Input"],
Cell[3871, 121, 152, 4, 73, "Input"],
Cell[4026, 127, 165, 5, 42, "Text"],
Cell[4194, 134, 150, 2, 73, "Input"],
Cell[4347, 138, 134, 4, 44, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[4518, 147, 50, 0, 51, "Section"],
Cell[4571, 149, 1158, 28, 119, "Text"],
Cell[5732, 179, 360, 6, 67, "Text"],
Cell[6095, 187, 121, 2, 80, "DisplayFormula"],
Cell[6219, 191, 557, 15, 67, "Text"],
Cell[6779, 208, 332, 5, 80, "DisplayFormula"],
Cell[7114, 215, 1209, 32, 145, "Text"],
Cell[8326, 249, 294, 4, 94, "DisplayFormula"],
Cell[8623, 255, 573, 16, 67, "Text"],
Cell[9199, 273, 464, 13, 67, "Text"],
Cell[9666, 288, 156, 3, 80, "DisplayFormula"],
Cell[9825, 293, 356, 7, 67, "Text"],
Cell[10184, 302, 1591, 30, 197, "Text"],
Cell[11778, 334, 133, 4, 41, "Text"],
Cell[11914, 340, 246, 4, 80, "DisplayFormula"],
Cell[12163, 346, 583, 14, 67, "Text"],
Cell[12749, 362, 660, 10, 119, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[13446, 377, 39, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[13510, 381, 118, 4, 68, "Subsection"],
Cell[13631, 387, 901, 19, 119, "Text"],
Cell[14535, 408, 141, 4, 41, "Text"],
Cell[14679, 414, 291, 4, 94, "DisplayFormula"],
Cell[14973, 420, 150, 5, 41, "Text"],
Cell[15126, 427, 512, 9, 153, "DisplayFormula"],
Cell[15641, 438, 21, 0, 41, "Text"],
Cell[15665, 440, 548, 9, 372, "DisplayFormula"],
Cell[16216, 451, 251, 8, 41, "Text"],
Cell[16470, 461, 218, 3, 62, "DisplayFormula"],
Cell[16691, 466, 21, 0, 41, "Text"],
Cell[16715, 468, 116, 2, 45, "DisplayFormula"],
Cell[16834, 472, 395, 10, 67, "Text"],
Cell[17232, 484, 273, 6, 67, "Text"],
Cell[17508, 492, 291, 4, 80, "DisplayFormula"],
Cell[17802, 498, 340, 11, 41, "Text"],
Cell[18145, 511, 612, 10, 121, "DisplayFormula"],
Cell[18760, 523, 20, 0, 41, "Text"],
Cell[18783, 525, 465, 7, 80, "DisplayFormula"],
Cell[19251, 534, 396, 6, 80, "DisplayFormula"],
Cell[19650, 542, 25, 0, 41, "Text"],
Cell[19678, 544, 189, 3, 61, "DisplayFormula"],
Cell[19870, 549, 36, 0, 41, "Text"],
Cell[19909, 551, 562, 12, 93, "Text"],
Cell[20474, 565, 159, 3, 80, "DisplayFormula"],
Cell[20636, 570, 105, 4, 41, "Text"],
Cell[20744, 576, 86, 1, 42, "DisplayFormula"],
Cell[20833, 579, 317, 7, 67, "Text"],
Cell[21153, 588, 299, 6, 75, "DisplayFormula"],
Cell[21455, 596, 394, 8, 67, "Text"],

Cell[CellGroupData[{
Cell[21874, 608, 55, 0, 55, "Subsubsection"],
Cell[21932, 610, 486, 11, 67, "Text"],
Cell[22421, 623, 300, 7, 125, "Input"],
Cell[22724, 632, 165, 5, 41, "Text"],
Cell[22892, 639, 363, 6, 206, "Input"],
Cell[23258, 647, 68, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[23351, 651, 194, 4, 97, "Input"],
Cell[23548, 657, 81, 1, 59, "Output"]
}, Closed]],
Cell[23644, 661, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[23731, 666, 113, 2, 75, "Input"],
Cell[23847, 670, 53, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[23937, 676, 170, 3, 96, "Input"],
Cell[24110, 681, 75, 1, 59, "Output"]
}, Closed]],
Cell[24200, 685, 58, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[24283, 689, 307, 5, 125, "Input"],
Cell[24593, 696, 54146, 1325, 279, 17586, 869, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[78754, 2024, 153, 3, 41, "Text"],
Cell[78910, 2029, 99, 2, 64, "DisplayFormula"],
Cell[79012, 2033, 337, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[79374, 2044, 400, 7, 165, "Input"],
Cell[79777, 2053, 51016, 1367, 279, 19715, 976, "GraphicsData", "PostScript",
   "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[130842, 3426, 62, 0, 55, "Subsubsection"],
Cell[130907, 3428, 76, 0, 41, "Text"],
Cell[130986, 3430, 311, 7, 125, "Input"],
Cell[131300, 3439, 72, 0, 41, "Text"],
Cell[131375, 3441, 363, 6, 206, "Input"],

Cell[CellGroupData[{
Cell[131763, 3451, 194, 4, 97, "Input"],
Cell[131960, 3457, 81, 1, 59, "Output"]
}, Closed]],
Cell[132056, 3461, 62, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[132143, 3466, 113, 2, 75, "Input"],
Cell[132259, 3470, 52, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[132348, 3476, 170, 3, 96, "Input"],
Cell[132521, 3481, 76, 1, 59, "Output"]
}, Closed]],
Cell[132612, 3485, 709, 15, 93, "Text"],

Cell[CellGroupData[{
Cell[133346, 3504, 307, 5, 125, "Input"],
Cell[133656, 3511, 55411, 1377, 279, 18750, 920, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[189104, 4893, 400, 7, 165, "Input"],
Cell[189507, 4902, 50955, 1366, 279, 19715, 976, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[240523, 6275, 37, 0, 68, "Subsection"],
Cell[240563, 6277, 86, 2, 41, "Text"],
Cell[240652, 6281, 139, 3, 80, "DisplayFormula"],
Cell[240794, 6286, 656, 23, 67, "Text"],
Cell[241453, 6311, 548, 8, 94, "DisplayFormula"],
Cell[242004, 6321, 51, 0, 41, "Text"],
Cell[242058, 6323, 78, 1, 41, "DisplayFormula"],
Cell[242139, 6326, 21, 0, 41, "Text"],
Cell[242163, 6328, 159, 2, 80, "DisplayFormula"],
Cell[242325, 6332, 233, 8, 41, "Text"],
Cell[242561, 6342, 138, 2, 80, "DisplayFormula"],
Cell[242702, 6346, 807, 19, 119, "Text"],
Cell[243512, 6367, 85, 1, 43, "DisplayFormula"],
Cell[243600, 6370, 196, 4, 41, "Text"],
Cell[243799, 6376, 419, 10, 67, "Text"],
Cell[244221, 6388, 178, 3, 94, "DisplayFormula"],
Cell[244402, 6393, 25, 0, 41, "Text"],
Cell[244430, 6395, 322, 5, 77, "DisplayFormula"],
Cell[244755, 6402, 30, 0, 41, "Text"],
Cell[244788, 6404, 738, 13, 82, "DisplayFormula"],
Cell[245529, 6419, 478, 13, 67, "Text"],
Cell[246010, 6434, 4059, 73, 1136, "Input"],

Cell[CellGroupData[{
Cell[250094, 6511, 44, 1, 73, "Input"],
Cell[250141, 6514, 1326, 20, 85, "Print",
  CellTags->"Info3351438211-6959969"]
}, Closed]],

Cell[CellGroupData[{
Cell[251504, 6539, 32, 0, 55, "Subsubsection"],
Cell[251539, 6541, 474, 9, 93, "Text"],

Cell[CellGroupData[{
Cell[252038, 6554, 252, 4, 152, "Input"],
Cell[252293, 6560, 291, 5, 59, "Output"]
}, Closed]],
Cell[252599, 6568, 89, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[252713, 6573, 270, 5, 125, "Input"],
Cell[252986, 6580, 60630, 1522, 279, 19085, 1005, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[313653, 8107, 391, 7, 162, "Input"],
Cell[314047, 8116, 50975, 1366, 279, 19715, 976, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[365037, 9485, 261, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[365323, 9494, 68, 1, 74, "Input"],
Cell[365394, 9497, 335, 5, 84, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[365766, 9507, 74, 1, 74, "Input"],
Cell[365843, 9510, 372, 5, 84, "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[366276, 9522, 41, 0, 68, "Subsection"],
Cell[366320, 9524, 271, 5, 67, "Text"],
Cell[366594, 9531, 91, 1, 41, "DisplayFormula"],
Cell[366688, 9534, 436, 14, 67, "Text"],
Cell[367127, 9550, 109, 2, 41, "DisplayFormula"],
Cell[367239, 9554, 34, 0, 41, "Text"],
Cell[367276, 9556, 146, 2, 64, "DisplayFormula"],
Cell[367425, 9560, 131, 3, 41, "Text"],
Cell[367559, 9565, 164, 2, 72, "DisplayFormula"],
Cell[367726, 9569, 874, 22, 119, "Text"],

Cell[CellGroupData[{
Cell[368625, 9595, 575, 11, 210, "Input"],
Cell[369203, 9608, 229, 3, 59, "Output"]
}, Closed]],
Cell[369447, 9614, 402, 12, 68, "Text"],
Cell[369852, 9628, 501, 9, 119, "DisplayFormula"],
Cell[370356, 9639, 269, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[370650, 9651, 491, 8, 99, "Input"],
Cell[371144, 9661, 134, 2, 59, "Output"]
}, Closed]],
Cell[371293, 9666, 135, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[371453, 9673, 69, 1, 73, "Input"],
Cell[371525, 9676, 227, 3, 59, "Output"]
}, Closed]],
Cell[371767, 9682, 33, 0, 41, "Text"],
Cell[371803, 9684, 67, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[371895, 9688, 318, 5, 151, "Input"],
Cell[372216, 9695, 60593, 1683, 279, 22247, 1205, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[432824, 11381, 251, 6, 41, "Text"],

Cell[CellGroupData[{
Cell[433100, 11391, 394, 6, 151, "Input"],
Cell[433497, 11399, 54313, 1461, 279, 21288, 1049, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[487859, 12866, 73, 0, 68, "Subsection"],
Cell[487935, 12868, 595, 13, 94, "Text"],
Cell[488533, 12883, 518, 12, 69, "Text"],
Cell[489054, 12897, 169, 3, 42, "DisplayFormula"],
Cell[489226, 12902, 475, 13, 67, "Text"],
Cell[489704, 12917, 230, 4, 64, "DisplayFormula"],
Cell[489937, 12923, 262, 9, 41, "Text"],
Cell[490202, 12934, 229, 4, 64, "DisplayFormula"],
Cell[490434, 12940, 282, 9, 41, "Text"],
Cell[490719, 12951, 218, 3, 66, "DisplayFormula"],
Cell[490940, 12956, 1011, 26, 95, "Text"],
Cell[491954, 12984, 365, 9, 68, "Text"],
Cell[492322, 12995, 3948, 69, 1168, "Input"],

Cell[CellGroupData[{
Cell[496295, 13068, 45, 1, 73, "Input"],
Cell[496343, 13071, 1137, 16, 85, "Print",
  CellTags->"Info3351438212-4114851"]
}, Closed]],

Cell[CellGroupData[{
Cell[497517, 13092, 111, 3, 55, "Subsubsection"],
Cell[497631, 13097, 46, 0, 41, "Text"],
Cell[497680, 13099, 117, 2, 68, "DisplayFormula"],
Cell[497800, 13103, 696, 21, 67, "Text"],

Cell[CellGroupData[{
Cell[498521, 13128, 872, 17, 333, "Input"],
Cell[499396, 13147, 315, 6, 84, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[499748, 13158, 471, 10, 151, "Input"],
Cell[500222, 13170, 80884, 2176, 279, 37236, 1633, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[581179, 15354, 42, 0, 51, "Section"],
Cell[581224, 15356, 1081, 22, 147, "Text"],

Cell[CellGroupData[{
Cell[582330, 15382, 51, 0, 68, "Subsection"],
Cell[582384, 15384, 2899, 45, 405, "Text"],
Cell[585286, 15431, 359, 8, 42, "Text"],
Cell[585648, 15441, 208, 3, 83, "DisplayFormula"],
Cell[585859, 15446, 515, 13, 68, "Text"],
Cell[586377, 15461, 355, 5, 83, "DisplayFormula"],
Cell[586735, 15468, 42, 0, 41, "Text"],
Cell[586780, 15470, 254, 4, 94, "DisplayFormula"],
Cell[587037, 15476, 55, 0, 41, "Text"],
Cell[587095, 15478, 435, 7, 94, "DisplayFormula"],
Cell[587533, 15487, 148, 5, 41, "Text"],
Cell[587684, 15494, 222, 3, 41, "DisplayFormula"],
Cell[587909, 15499, 222, 8, 41, "Text"],
Cell[588134, 15509, 119, 2, 43, "DisplayFormula"],
Cell[588256, 15513, 1075, 32, 93, "Text"],
Cell[589334, 15547, 829, 18, 93, "Text"],
Cell[590166, 15567, 301, 4, 46, "DisplayFormula"],
Cell[590470, 15573, 1355, 41, 93, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[591862, 15619, 57, 0, 68, "Subsection"],
Cell[591922, 15621, 1276, 23, 197, "Text"],
Cell[593201, 15646, 224, 4, 67, "Text"],
Cell[593428, 15652, 259, 5, 77, "DisplayFormula"],
Cell[593690, 15659, 675, 15, 93, "Text"],
Cell[594368, 15676, 129, 2, 46, "DisplayFormula"],
Cell[594500, 15680, 252, 6, 41, "Text"],
Cell[594755, 15688, 134, 5, 41, "Text"],
Cell[594892, 15695, 573, 10, 85, "DisplayFormula"],
Cell[595468, 15707, 245, 6, 41, "Text"],
Cell[595716, 15715, 300, 5, 69, "DisplayFormula"],
Cell[596019, 15722, 285, 6, 41, "Text"],
Cell[596307, 15730, 529, 8, 79, "DisplayFormula"],
Cell[596839, 15740, 1413, 34, 119, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[598289, 15779, 36, 0, 68, "Subsection"],
Cell[598328, 15781, 1810, 34, 250, "Text"],
Cell[600141, 15817, 7326, 123, 1992, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[607504, 15945, 39, 0, 68, "Subsection"],
Cell[607546, 15947, 88, 2, 41, "Text"],
Cell[607637, 15951, 540, 11, 209, "Input"],
Cell[608180, 15964, 126, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[608331, 15971, 255, 5, 99, "Input"],
Cell[608589, 15978, 165, 3, 37, "Print"],
Cell[608757, 15983, 165, 3, 35, "Print"],
Cell[608925, 15988, 180, 3, 37, "Print"],
Cell[609108, 15993, 180, 3, 37, "Print"],
Cell[609291, 15998, 180, 3, 37, "Print"],
Cell[609474, 16003, 179, 3, 37, "Print"],
Cell[609656, 16008, 181, 3, 37, "Print"],
Cell[609840, 16013, 181, 3, 35, "Print"],
Cell[610024, 16018, 180, 3, 35, "Print"],
Cell[610207, 16023, 178, 3, 35, "Print"],
Cell[610388, 16028, 179, 3, 35, "Print"],
Cell[610570, 16033, 175, 3, 35, "Print"],
Cell[610748, 16038, 175, 3, 35, "Print"],
Cell[610926, 16043, 179, 3, 35, "Print"],
Cell[611108, 16048, 184, 3, 37, "Print"],
Cell[611295, 16053, 179, 3, 35, "Print"],
Cell[611477, 16058, 177, 3, 35, "Print"],
Cell[611657, 16063, 179, 3, 35, "Print"],
Cell[611839, 16068, 179, 3, 35, "Print"],
Cell[612021, 16073, 177, 3, 35, "Print"],
Cell[612201, 16078, 175, 3, 35, "Print"],
Cell[612379, 16083, 165, 3, 35, "Message"],
Cell[612547, 16088, 745, 14, 86, "Output"]
}, Closed]],
Cell[613307, 16105, 172, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[613504, 16112, 207, 4, 99, "Input"],
Cell[613714, 16118, 168, 3, 37, "Print"],
Cell[613885, 16123, 162, 3, 35, "Print"],
Cell[614050, 16128, 164, 3, 37, "Print"],
Cell[614217, 16133, 165, 3, 37, "Print"],
Cell[614385, 16138, 162, 3, 37, "Print"],
Cell[614550, 16143, 163, 3, 37, "Print"],
Cell[614716, 16148, 164, 3, 37, "Print"],
Cell[614883, 16153, 163, 3, 35, "Print"],
Cell[615049, 16158, 180, 3, 35, "Print"],
Cell[615232, 16163, 178, 3, 35, "Print"],
Cell[615413, 16168, 181, 3, 35, "Print"],
Cell[615597, 16173, 184, 3, 35, "Print"],
Cell[615784, 16178, 183, 3, 37, "Print"],
Cell[615970, 16183, 368, 6, 84, "Output"]
}, Closed]],
Cell[616353, 16192, 1146, 21, 171, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[617536, 16218, 29, 0, 68, "Subsection"],
Cell[617568, 16220, 557, 14, 93, "Text"],
Cell[618128, 16236, 153, 2, 66, "DisplayFormula"],
Cell[618284, 16240, 721, 21, 67, "Text"],
Cell[619008, 16263, 213, 4, 75, "DisplayFormula"],
Cell[619224, 16269, 563, 14, 93, "Text"],
Cell[619790, 16285, 147, 2, 46, "DisplayFormula"],
Cell[619940, 16289, 117, 3, 41, "Text"],
Cell[620060, 16294, 297, 6, 68, "Text"],
Cell[620360, 16302, 69, 1, 73, "Input"],
Cell[620432, 16305, 69, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[620526, 16309, 134, 2, 99, "Input"],
Cell[620663, 16313, 343, 5, 84, "Output"]
}, Closed]],
Cell[621021, 16321, 165, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[621211, 16328, 188, 3, 99, "Input"],
Cell[621402, 16333, 799, 14, 109, "Output"]
}, Closed]],
Cell[622216, 16350, 569, 12, 93, "Text"],

Cell[CellGroupData[{
Cell[622810, 16366, 803, 13, 347, "Input"],
Cell[623616, 16381, 58765, 1703, 279, 22384, 1249, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[682442, 18091, 35, 0, 51, "Section"],
Cell[682480, 18093, 1004, 21, 120, "Text"],
Cell[683487, 18116, 410, 9, 177, "Input"],

Cell[CellGroupData[{
Cell[683922, 18129, 156, 3, 99, "Input"],
Cell[684081, 18134, 3778, 81, 477, "Output"]
}, Closed]],
Cell[687874, 18218, 1013, 27, 121, "Text"],
Cell[688890, 18247, 345, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[689260, 18258, 73, 1, 73, "Input"],
Cell[689336, 18261, 432, 6, 134, "Output"]
}, Closed]],
Cell[689783, 18270, 121, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[689929, 18277, 298, 6, 100, "Input"],
Cell[690230, 18285, 2574, 49, 293, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[692841, 18339, 250, 4, 99, "Input"],
Cell[693094, 18345, 60795, 1525, 279, 19072, 1005, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[753926, 19875, 407, 7, 125, "Input"],
Cell[754336, 19884, 50986, 1367, 279, 19713, 976, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[805337, 21254, 94, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[805456, 21260, 88, 1, 103, "Input"],
Cell[805547, 21263, 53, 1, 59, "Output"]
}, Closed]],
Cell[805615, 21267, 644, 15, 94, "Text"],

Cell[CellGroupData[{
Cell[806284, 21286, 333, 7, 157, "Input"],
Cell[806620, 21295, 4234, 87, 533, "Output"]
}, Closed]],
Cell[810869, 21385, 277, 4, 202, "Input"],

Cell[CellGroupData[{
Cell[811171, 21393, 776, 13, 277, "Input"],
Cell[811950, 21408, 60112, 1778, 279, 23480, 1321, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[872099, 23191, 777, 13, 277, "Input"],
Cell[872879, 23206, 84941, 2325, 279, 29350, 1634, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[957835, 25534, 1384, 29, 175, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[959256, 25568, 37, 0, 51, "Section"],
Cell[959296, 25570, 863, 23, 94, "Text"],
Cell[960162, 25595, 120, 2, 99, "Input"],
Cell[960285, 25599, 53, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[960363, 25604, 46, 1, 73, "Input"],
Cell[960412, 25607, 7349, 181, 150, "Print"]
}, Closed]],
Cell[967776, 25791, 346, 6, 67, "Text"],

Cell[CellGroupData[{
Cell[968147, 25801, 110, 3, 68, "Subsection"],
Cell[968260, 25806, 564, 11, 96, "Text"],

Cell[CellGroupData[{
Cell[968849, 25821, 63, 0, 55, "Subsubsection"],
Cell[968915, 25823, 555, 10, 255, "Input"],

Cell[CellGroupData[{
Cell[969495, 25837, 62, 1, 73, "Input"],
Cell[969560, 25840, 290, 5, 84, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[969887, 25850, 212, 5, 99, "Input"],
Cell[970102, 25857, 639, 11, 109, "Output"]
}, Closed]],
Cell[970756, 25871, 154, 3, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[970947, 25879, 82, 1, 55, "Subsubsection"],
Cell[971032, 25882, 748, 14, 307, "Input"],

Cell[CellGroupData[{
Cell[971805, 25900, 62, 1, 73, "Input"],
Cell[971870, 25903, 307, 5, 84, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[972214, 25913, 212, 5, 99, "Input"],
Cell[972429, 25920, 655, 10, 109, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[973121, 25935, 472, 10, 151, "Input"],
Cell[973596, 25947, 81597, 2194, 279, 37305, 1643, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[1055254, 28148, 30, 0, 68, "Subsection"],
Cell[1055287, 28150, 224, 7, 42, "Text"],
Cell[1055514, 28159, 69, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[1055608, 28164, 271, 6, 128, "Input"],
Cell[1055882, 28172, 894, 15, 134, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[1056813, 28192, 726, 13, 339, "Input"],
Cell[1057542, 28207, 58765, 1703, 279, 22384, 1249, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

