(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    796671,      20007]*)
(*NotebookOutlinePosition[    831640,      21258]*)
(*  CellTagsIndexPosition[    831369,      21248]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell[TextData[{
  "MathLite: enough ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " for an introductory physics laboratory"
}], "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell[CellGroupData[{

Cell["Introduction", "Section"],

Cell[TextData[{
  "In this notebook we introduce enough of the features of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " for a student in an introductory physics laboratory to analyze data and \
prepare lab reports.  It is impossible in a single session to present or to \
absorb a mathematical software system that is so vast.  A full semester \
course on ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is available at ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  " in tutorial format.  Further information and many practical examples can \
be found there or in the online documentation available through the ",
  StyleBox["Help Browser",
    FontSlant->"Italic"],
  ".  Hopefully, by the end of this session you will be sufficiently familiar \
with the basic structure of ",
  StyleBox["Mathematica ",
    FontSlant->"Italic"],
  "to make efficient use of those resources. "
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["The notebook interface", "Section"],

Cell[TextData[{
  "I assume that you are reading this document on a monitor coupled to a \
computer capable of running a notebook interface.  If so, you will see that \
the present document is divided into a number of sections with clearly \
identifiable section headings; some sections are also divided into \
subsections.  The details of the appearance of these headings will depend \
upon the particular computer system that you are using and perhaps upon some \
of the options that were selected when ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " was installed.  Nevertheless, the logical structure should be \
unmistakable."
}], "Text"],

Cell[TextData[{
  "The descriptive sections of this notebook assume that you are using a ",
  StyleBox["Microsoft Windows",
    FontSlant->"Italic"],
  " system.  Notebook interfaces for other systems are generally quite \
similar, but some of the shortcut keys and details of the appearance may vary \
between interfaces.  We hope that it will not take long to become accustomed \
to these differences."
}], "Text"],

Cell[TextData[{
  "Each section consists of a collection of ",
  StyleBox["cells",
    FontWeight->"Bold"],
  ", of which there are several types.  The present paragraph sits in a ",
  StyleBox["text cell. ",
    FontSlant->"Italic"],
  " Input intended for transmission to the kernel should occupy an ",
  StyleBox["input cell ",
    FontSlant->"Italic"],
  "and the results returned by the kernel will be placed in an ",
  StyleBox["output cell",
    FontSlant->"Italic"],
  ".  Depending upon the style options selected for a particular document, \
there will usually be several other cell types available, such as title, \
section, subsection, etc.  The vertical extent of a cell is generally \
delineated by a bracket that appears at the right side of the display.  The \
innermost bracket identifies an individual cell, while outer brackets \
indicate increasingly inclusive logical groupings of cells.  A cell or group \
of cells can be selected by using the mouse to click on the appropriate \
bracket, which should then change to a bolder appearance.  "
}], "Text"],

Cell[TextData[{
  "Notebooks are often saved with the cells ",
  StyleBox["closed",
    FontSlant->"Italic"],
  " \[LongDash] the logical structure of the notebook is then revealed by the \
organization of its section headings while the details are hidden.  Closed \
cells are identified by dark hooks at the base of the cell bracket.  To open \
a cell or group of cells, double click of the appropriate cell bracket.  \
Repeating that operation then closes the cell.  Alternatively, you can select \
the cell and use the keystroke sequence \[ControlKey] { to open or \
\[ControlKey] } to close all subcells within your selection.  Notice that \
many shortcuts are designed for easy memorization; in this case, an \
open-bracket opens a set of cells while a closed-bracket closes the cell."
}], "Text"],

Cell[TextData[{
  "At the top of the screen you should see a familiar menu bar with the \
headings File, Edit, Format, etc.  Under Edit you see the usual copy, cut, \
paste, and related actions together with suggested keystroke alternatives.  \
The most common editing functions have the same keystroke shortcuts found in \
most other ",
  StyleBox["Windows",
    FontSlant->"Italic"],
  " applications, such as \[ControlKey]C for copy, \[ControlKey]X for cut, or \
\[ControlKey]V for paste.  The copy and cut functions act either on text \
highlighted by repeated mouse clicks or dragging the cursor or on cells \
selected by clicking on the appropriate cell bracket.  The paste function \
inserts text into an existing cell at the current ",
  StyleBox["insertion point",
    FontSlant->"Italic"],
  ", indicated by the cursor in the form of a long vertical bar.  Note that \
the insertion point may not be visible on the screen if you have used page \
buttons or scroll bars, but you can move it to the desired location by \
clicking within the desired cell.  Similarly,  by moving the cursor between \
cells and clicking when you see a horizontal bar, you create an insertion \
point for a new cell and can paste a selection into it.  By default, new \
cells are created with the input type, but you can select a different type \
either with the menu activated by the right mouse button.  "
}], "Text"],

Cell["\<\
Take a few minutes to practice these operations and familiarize yourself with \
the appearance of the cursor at various positions in the notebook. \
\>", "Exercise"],

Cell[TextData[{
  "Create a new notebook using the File menu.  Activate the toolbar for the \
new notebook: under the Format menu click ShowToolbar to toggle its status \
(checked means on).    The box on the left can be used to choose the cell \
type for a new cell or to change the cell type for a selected cell.  Write a \
title and subtitle for your new notebook into cells with those types.  Type \
some text into a cell type.   Then save the notebook using the File menu (use \
Save As to name it) .  Close the notebook using either the File menu or \
clicking the ",
  Cell[BoxData[
      FormBox[
        FrameBox["x",
          BoxMargins->{{0.2, 0.2}, {0.4, 0.4}}], TraditionalForm]]],
  " in the top right corner.  Re-open your notebook."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Help!", "Section"],

Cell[CellGroupData[{

Cell["Help Browser", "Subsection"],

Cell[TextData[{
  "Probably the most important source of help is the ",
  StyleBox["Help Browser",
    FontWeight->"Bold"],
  " provided with ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " itself.  The help browser can be accessed from the menu bar and its use \
should be self-explanatory.  Simply type the name of the function or keyword \
in the dialog box and search either the appropriate resource index, usually \
",
  StyleBox["Built-in Functions",
    FontWeight->"Bold"],
  " or ",
  StyleBox["The Mathematica Book",
    FontSlant->"Italic"],
  " itself, whose entire text of more than 1000 pages is available online.  \
Many parts of the Help Browser include statements which you can modify or \
execute without actually changing anything in your system.  Don't worry \
\[LongDash] you won't break it!  You can also copy cells from the browser \
into your own notebook."
}], "Text"],

Cell["\<\
Find the examples of Integrate in the reference guide.  Execute some of the \
input statements by clicking within the cell and pressing Shift Enter \
simultaneously.  The Enter key from the numeric keypad should also work.\
\>", "Exercise"],

Cell["\<\
Use the Help Browser to see the syntax for Plot and play with a few of the \
examples provided.\
\>", "Exercise"],

Cell[TextData[{
  "We also provide hyperlinks to either the ",
  StyleBox["Reference Guide",
    FontSlant->"Italic"],
  " or to an appropriate section of the ",
  StyleBox["Mathematica Book",
    FontSlant->"Italic"],
  " are provided whenever a new function is introduced or used in a new \
manner.  Simply clicking on the highlighted button takes you to the \
appropriate documentation for that feature.  The Help Browser will then \
provide the definition of the function, its options, and details of its \
syntax and properties.  In many cases there will also be examples provided \
and links to related information throughout the ",
  StyleBox["Mathematica Book",
    FontSlant->"Italic"],
  ".  Using this method it will not be necessary to provide such details \
within these notebooks themselves.  For example, by clicking on this \
hyperlink, ",
  ButtonBox["Range",
    ButtonStyle->"RefGuideLink"],
  ", one finds that ",
  StyleBox["Range[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["nmin,nmax", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " produces a list of integers between ",
  StyleBox["nmin", "InlineInput"],
  " and ",
  StyleBox["nmax", "InlineInput"],
  " (Try it!).   "
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Information escape", "Subsection"],

Cell[TextData[{
  "Quick help on a built-in function or on a symbol that you have created can \
be obtained by typing ",
  StyleBox["?symbol", "InlineInput",
    FontWeight->"Bold"],
  " or, for more information, ",
  StyleBox["??symbol", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Fibonacci\)\)], "Input"],

Cell[BoxData[
    RowBox[{"\<\"Fibonacci[n] gives the nth Fibonacci number. Fibonacci[n, x] \
gives the nth Fibonacci polynomial, using x as the variable.\"\>", " ", 
      ButtonBox[
        StyleBox["More\[Ellipsis]",
          "SR"],
        ButtonData:>"Fibonacci",
        Active->True,
        ButtonStyle->"RefGuideLink"]}]], "Print",
  CellTags->"Info3351235691-3603827"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(?? Fibonacci\)], "Input"],

Cell[BoxData[
    RowBox[{"\<\"Fibonacci[n] gives the nth Fibonacci number. Fibonacci[n, x] \
gives the nth Fibonacci polynomial, using x as the variable.\"\>", " ", 
      ButtonBox[
        StyleBox["More\[Ellipsis]",
          "SR"],
        ButtonData:>"Fibonacci",
        Active->True,
        ButtonStyle->"RefGuideLink"]}]], "Print",
  CellTags->"Info3351235691-3662103"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {\(Attributes[Fibonacci] = {Listable, NumericFunction, Protected, 
                ReadProtected}\)}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "Fibonacci"],
      Editable->False]], "Print",
  CellTags->"Info3351235691-3662103"]
}, Closed]],

Cell["\<\
Assuming that you already know the definition for Fibonacci numbers or \
polynomials, this method provides the information needed to use the built-in \
function.  However, if you do not know that definition, more complete \
information can be obtained using the Help Browser.  A suitable hyperlink is \
often included with the output.\
\>", "Text"],

Cell[TextData[{
  "Information about the available functions can be obtained by using \
wildcards.  Thus, we find the following symbols with names beginning with ",
  StyleBox["Plot", "InlineInput"],
  ".  Notice that clicking on any of these links provides a brief explanation \
of the function and a link to more detailed information."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?Plot*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["Plot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`Plot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3Matrix",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235709-6637170", 
                    "System`Plot3Matrix"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotJoined",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`PlotJoined"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`PlotPoints"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRegion",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`PlotRegion"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], "", "", "", ""},
              {
                ButtonBox[
                  StyleBox["Plot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`Plot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotDivision",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235709-6637170", 
                    "System`PlotDivision"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotLabel",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`PlotLabel"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRange",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`PlotRange"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotStyle",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235709-6637170", "System`PlotStyle"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], "", "", "", ""}
              },
            RowMinHeight->{1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Plot*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ArrayPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`ArrayPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`ListContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`ListPlot3D"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`ParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3Matrix",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`Plot3Matrix"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotLabel",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`PlotLabel"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRegion",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`PlotRegion"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["ContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`ContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`ListDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["MaxPlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`MaxPlotPoints"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`Plot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotDivision",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`PlotDivision"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`PlotPoints"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotStyle",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`PlotStyle"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["DensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`DensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`ListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351235712-2028794", 
                    "System`ParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`Plot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotJoined",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`PlotJoined"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRange",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351235712-2028794", "System`PlotRange"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Input conventions", "Section"],

Cell[CellGroupData[{

Cell["Symbols", "Subsection"],

Cell[TextData[{
  "First note that ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is case-sensitive, meaning that a symbol named ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " is distinct from a symbol named ",
  Cell[BoxData[
      \(TraditionalForm\`Y\)]],
  ".  All built-in symbols are assigned names that begin with capital \
letters.  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " generally names its symbols quite descriptively with names closely \
related to the function of each symbol.  For example, the function designed \
to simplify algebraic expressions is called ",
  ButtonBox["Simplify",
    ButtonStyle->"RefGuideLink"],
  ".  Sometimes the names become very long, but most users find it easier to \
type a verbose name than to remember abbreviations which can be chosen almost \
arbitrarily.  To avoid confusing your variables with names reserved by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " it is good practice to choose names that begin with lowercase letters, \
but if you also choose long and descriptive variable names that is usually no \
longer necessary."
}], "Text"],

Cell[TextData[{
  "Symbol names must begin with a letter and cannot include punctuation marks \
(_,@,!, etc.), operators (+,-,/, etc.), or embedded spaces.  Numerals may be \
used anywhere except the first character.  Names cannot duplicate any of \
those reserved by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ", which include"
}], "Text"],

Cell[TextData[{
  "\t",
  StyleBox["N", "InlineInput",
    FontWeight->"Bold"],
  "\tnumerical evaluation, confused with number\n\t",
  StyleBox["I", "InlineInput",
    FontWeight->"Bold"],
  "\timaginary unit ",
  Cell[BoxData[
      \(TraditionalForm\`\@\(-1\)\)], "InlineFormula"],
  "\n\t",
  StyleBox["E", "InlineInput",
    FontWeight->"Bold"],
  "\texponential constant ",
  Cell[BoxData[
      \(TraditionalForm\`e\)], "InlineFormula"],
  " (base of natural logarithms), confused with energy\n\t",
  StyleBox["C", "InlineInput",
    FontWeight->"Bold"],
  "\tconstant of integration \n\t",
  StyleBox["D", "InlineInput",
    FontWeight->"Bold"],
  "\tpartial derivative\n\t",
  StyleBox["Pi", "InlineInput",
    FontWeight->"Bold"],
  "\t\[Pi]"
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Assignments", "Subsection"],

Cell[TextData[{
  "A value is assigned to a symbol by ",
  StyleBox["symbol=value", "InlineInput"],
  ".  Several symbols can be assigned the same value in a single line: ",
  StyleBox["symbol1=symbol2=symbol3=value", "InlineInput"],
  ".  To evaluate a ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " statement, click in the cell or select its cell bracket and press \
\[ShiftKey]\[EnterKey]."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(a = \(b = \(c = Pi\)\)\)], "Input"],

Cell[BoxData[
    \(\[Pi]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Notice that ",
  StyleBox["Mathematica ",
    FontSlant->"Italic"],
  "performs some automatic typesetting, replacing ",
  StyleBox["Pi", "InlineInput",
    FontWeight->"Bold"],
  " by \[Pi] but does not display a numerical value because it recognizes \
that a decimal representation would only be an approximation because \[Pi] is \
an exact number, even if it is irrational, and can in principle be evaluated \
with arbitrary precision.  Nevertheless, if you can a decimal representation, \
the numerical evaluation function ",
  StyleBox["N", "InlineInput",
    FontWeight->"Bold"],
  " will provide it"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(N[a]\)], "Input"],

Cell[BoxData[
    \(3.141592653589793`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(N[a, 100]\)], "Input"],

Cell[BoxData[
    \(3.1415926535897932384626433832795028841971693993751058209749445923078164\
062862089986280348253421170679821480865124059`100. \)], "Output"]
}, Closed]],

Cell["to any desired precision.", "Text"],

Cell[TextData[{
  "Be aware that assignments remain in force for an entire ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " session, until reset or cleared, and are not confined to the notebook in \
which they are made.  One of  the most common errors is to reuse a symbol \
whose previous assignment has been forgotten.  For example, if we try to \
solve the quadratic equation symbolically"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(Solve[a*x\^2 + b*x + c \[Equal] 0, x]\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(-\((\(-1\))\)\^\(1/3\)\)}, {x \[Rule] \((\(-1\))\)\^\(2/3\
\)}}\)], "Output"]
}, Closed]],

Cell["\<\
we obtain rather specific, and clearly unwanted results, because assignments \
have already been made for the coefficients.  Therefore, we must clear those \
assignments first! \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(\(Clear[a, b, c];\)\[IndentingNewLine]
    Solve[a*x\^2 + b*x + c \[Equal] 0, x]\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(\(-b\) - \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)}, {x \[Rule] \(\
\(-b\) + \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)}}\)], "Output"]
}, Closed]],

Cell["It is good practice to clear symbols before using them!", "Advice"],

Cell[TextData[{
  "Notice that ",
  StyleBox["==", "InlineInput"],
  " is used to indicate an ",
  StyleBox["equation",
    FontSlant->"Italic"],
  " instead of an assignment.  The typeset form is obtained using \
\[EscapeKey]==\[EscapeKey].  Similarly, the typeset form of the exponent ",
  StyleBox["x^n", "InlineInput"],
  " is obtained using \[ControlKey]^.",
  Cell[BoxData[
      FormBox[Cell[""], TraditionalForm]]],
  "  Finally, a space between two symbols indicates multiplication."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(a = 2.34;\)\), "\[IndentingNewLine]", 
    \(\(b = 5.67;\)\), "\[IndentingNewLine]", 
    \(a\ b\)}], "Input"],

Cell[BoxData[
    \(13.2678`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Several statements, separated by semicolons, can be included in a single \
input cell.  Here only the final result is printed because the semicolon \
suppresses output.  Note that ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " aficionados rarely type ",
  StyleBox["*", "InlineInput"],
  " for multiplication, although it is allowed and it sometimes improves \
readability."
}], "Text"],

Cell[BoxData[
    \(Clear[a, b]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Replacements", "Subsection"],

Cell[TextData[{
  "One of the most powerful features of a symbolic manipulator is its ability \
to replace one expression with another using replacement rules defined by the \
user.  A replacement rule is given in the form ",
  StyleBox["OldExpression -> NewExpression", "InlineInput"],
  "; a prettier version the arrow, ",
  Cell[BoxData[
      \(TraditionalForm\` \[Rule] \)]],
  ", is entered as \[EscapeKey]->\[EscapeKey].  The usual syntax for applying \
a replacement rule to an expression is to use the replacement operator ",
  StyleBox["/.", "InlineInput"],
  " after an expression.  Many ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions return their results in the form of replacement rules.  For \
example, the roots of the quadratic equation"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(sol\)\(=\)\(Solve[a*x\^2 + b*x + c \[Equal] 0, 
        x]\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(\(-b\) - \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)}, {x \[Rule] \(\
\(-b\) + \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "are returned as a list of replacement rules for the independent variable \
",
  StyleBox["x", "InlineInput",
    FontWeight->"Bold"],
  ".  Thus,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[a\ x\^2 + b\ x\  + c\  /. sol]\)], "Input"],

Cell[BoxData[
    \({0, 0}\)], "Output"]
}, Closed]],

Cell["\<\
substitutes those solutions into the original expression and the simplify \
function confirms that both solutions are indeed roots.  One should use \
replacement rules to avoid making unnecessary assignments.  For example, if \
we define the expression\
\>", "Text"],

Cell[BoxData[
    \(\(quad = a\ x\^2 + \ b\ x\  + \ c;\)\)], "Input"],

Cell["numerical values can be substituted using", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(quad /. {a \[Rule] 2, b \[Rule] \(-2\), c \[Rule] 4}\) /. 
      x \[Rule] 0.5\)], "Input"],

Cell[BoxData[
    \(3.5`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "without making explicit assignments to the parameters ",
  StyleBox["a,b,c", "InlineInput"],
  " or the variable ",
  StyleBox["x", "InlineInput"],
  " that would alter the meaning of ",
  StyleBox["quad", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[TextData[{
  "Many ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions accept options in the form of replacement rules.  For example,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[x\^2 + 1]\)], "Input"],

Cell[BoxData[
    \(1 + x\^2\)], "Output"]
}, Closed]],

Cell["\<\
is returned unevaluated because this expression cannot be factored using just \
integers.  There is an option, though, that does enable factorization\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[x\^2 + 1, GaussianIntegers \[Rule] True]\)], "Input"],

Cell[BoxData[
    \(\((\(-\[ImaginaryI]\) + x)\)\ \((\[ImaginaryI] + x)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "by including integer multiples of ",
  Cell[BoxData[
      \(TraditionalForm\`\[ImaginaryI] = \@\(-1\)\)]],
  ".  The function ",
  StyleBox["Options", "InlineInput",
    FontWeight->"Bold"],
  " provides a list of available options.  Some functions have many more \
options and you would  need to consult the Help Browser for information about \
those whose names are not self-explanatory."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[Factor]\)], "Input"],

Cell[BoxData[
    \({Extension \[Rule] None, GaussianIntegers \[Rule] False, 
      Modulus \[Rule] 0, Trig \[Rule] False}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Delimiters", "Subsection"],

Cell[TextData[{
  "Traditional mathematical notation can be ambiguous.  For example, should \
",
  Cell[BoxData[
      \(TraditionalForm\`f(a + b)\)]],
  " be interpreted as ",
  Cell[BoxData[
      \(TraditionalForm\`a*f + b*f\)]],
  " or as a function ",
  Cell[BoxData[
      \(TraditionalForm\`f(x)\)]],
  " evaluated for the argument ",
  Cell[BoxData[
      \(TraditionalForm\`x \[Rule] a + b\)]],
  "?  Humans that understand the context usually have little difficulty \
parsing such a statement, but machines are much less tolerant of ambiguity.  \
Therefore, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " assigns rigid meanings to delimiters.  ",
  StyleBox["Parentheses are used only for grouping expressions while square \
brackets are used only for packaging the arguments of a function.",
    FontWeight->"Bold"],
  "  Thus,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f \((a + b)\)\)], "Input"],

Cell[BoxData[
    \(\((a + b)\)\ f\)], "Output"]
}, Closed]],

Cell[TextData[{
  "satisfies the commutative law of multiplication.  Using the built-in \
function ",
  ButtonBox["Expand",
    ButtonStyle->"RefGuideLink"],
  " with an argument enclosed by square brackets"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Expand[f \((a + b)\)]\)], "Input"],

Cell[BoxData[
    \(a\ f + b\ f\)], "Output"]
}, Closed]],

Cell["\<\
then produces a fully expanded result.  The table below provides a more \
complete list of the available delimiters.\
\>", "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {
            StyleBox["Delimiters",
              FontFamily->"Times New Roman",
              FontWeight->"Bold"]},
          {
            StyleBox[GridBox[{
                  {
                    StyleBox["delimiter",
                      "Text"], " ", 
                    StyleBox["purpose",
                      "Text"], 
                    StyleBox["example",
                      "Text"]},
                  {
                    StyleBox["parentheses",
                      "Text"], \((\ )\), 
                    StyleBox["grouping",
                      "Text"], \(\((a + b)\)*c\)},
                  {
                    StyleBox["brackets",
                      "Text"], \([\ ]\), 
                    StyleBox["functions",
                      "Text"], \(Sqrt[a]\)},
                  {
                    StyleBox["braces",
                      "Text"], \({\ }\), 
                    StyleBox["lists",
                      "Text"], \({a, b, c}\)},
                  {
                    StyleBox[\(double\ brackets\),
                      "Text"], \(\([\([\ ]\)]\)\(\ \)\), 
                    StyleBox[\(part\ of\ expression\),
                      "Text"], 
                    RowBox[{
                      StyleBox["list",
                        FontSlant->"Italic"], "[", \([2]\), "]"}]},
                  {
                    StyleBox["comma",
                      "Text"], ",", 
                    StyleBox[\(separate\ elements\),
                      "Text"], \({a, b, c}\)},
                  {
                    StyleBox["semicolon",
                      "Text"], ";", 
                    StyleBox[\(separate\ expressions\),
                      "Text"], \(a*b; \ b - c\)},
                  {
                    StyleBox["quotes",
                      "Text"], "\"\< \>\"", 
                    StyleBox[\(text\ strings\),
                      "Text"], 
                    StyleBox["\"\<Have a nice day!\>\"",
                      "Text"]},
                  {
                    StyleBox["comment",
                      "Text"], \( (*\ *) \), 
                    StyleBox["documentation",
                      "Text"], 
                    StyleBox[\( (*\ this\ does\  ... \ *) \),
                      "Text"]}
                  },
                ColumnAlignments->{Left, Center, Left},
                GridFrame->True,
                RowLines->{True, False}],
              "Input",
              Background->None]}
          }],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Text",
  TextAlignment->Center,
  TextJustification->0]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Functions", "Section"],

Cell[CellGroupData[{

Cell["Standard mathematical functions", "Subsection"],

Cell[TextData[{
  "Practically all standard mathematical functions (arithmetic, \
trigonometric, exponential, logarithmic, hyperbolic, etc.) are provided.  The \
names are capitalized and are either standard abbreviations, such as ",
  StyleBox["Sin", "InlineInput",
    FontWeight->"Bold"],
  ",",
  " or are spelled completely, such as ",
  StyleBox["Factorial", "InlineInput",
    FontWeight->"Bold"],
  ".  A brief summary of these functions is given in the accompanying table.  \
Representative argument sequences have been included for some functions, but \
we have not attempted to be exhaustive at this stage.  A vast array of \
special functions is also available \[LongDash] consult the Help Browser for \
any function that you might need and it will probably be available!"
}], "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {
            StyleBox[\(Standard\ Mathematical\ Functions\),
              FontFamily->"Times New Roman",
              FontWeight->"Bold"]},
          {
            StyleBox[GridBox[{
                  {
                    StyleBox["exponential",
                      "Text"], \(Exp[x]\)},
                  {
                    StyleBox[\(natural\ logarithm\),
                      "Text"], \(Log[x]\)},
                  {
                    StyleBox[\(logarithm\ to\ base\ b\),
                      "Text"], \(Log[b, x]\)},
                  {
                    StyleBox[\(\(trigonometric\)\(\ \)\),
                      "Text"], \(Sin, Cos, Tan, Sec, Csc, Cot\)},
                  {
                    StyleBox[\(\(inverse\)\(\ \)\(trig\)\(\ \)\),
                      "Text"], \(ArcSin, ArcCos, ArcTan, \)},
                  {" ", \(ArcSec, ArcCsc, ArcCot\)},
                  {
                    StyleBox[\(hyperbolic\ trig\),
                      "Text"], \(Sinh, Cosh, Tanh, Sech, Csch, Coth\)},
                  {
                    StyleBox[\(inverse\ hyperbolic\ trig\),
                      "Text"], \(ArcSinh, ArcCosh, ArcTanh, \)},
                  {" ", \(ArcSech, ArcCsch, ArcCoth\)}
                  },
                ColumnAlignments->{Left},
                GridFrame->True],
              "Input",
              Background->None]}
          }],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Text",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "Compute ",
  Cell[BoxData[
      \(TraditionalForm\`\(log\_10\) e\)]],
  " and check the result."
}], "Exercise"],

Cell[TextData[{
  "Evaluate ",
  Cell[BoxData[
      \(TraditionalForm\`sin(\ \(tan\^\(-1\)\)(x)\ )\)]],
  ".  [Remember to use ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " input conventions.]"
}], "Exercise"],

Cell[TextData[{
  "By default, trigonometric functions assume that the argument is expressed \
in radians.  For those of us who think in degrees (who doesn't?), the \
built-in symbol ",
  StyleBox["Degree", "InlineInput",
    FontWeight->"Bold"],
  " converts from degrees to radians and can be included as a factor in the \
argument."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Sin[30\ Degree]\)], "Input"],

Cell[BoxData[
    \(1\/2\)], "Output"]
}, Closed]],

Cell["Use Degree to obtain the number of radians in 32.5\[Degree].", \
"Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Creating functions", "Subsection"],

Cell[TextData[{
  "You will need to create your functions, but there are a few subtleties \
that must be understood.  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " evaluates expressions by using pattern recognition to compare an \
expression with tables of definitions and rules.  The obvious definition  "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f[x] = \(x\^2\) Exp[\(-x\^2\)/2]\)], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(-\(x\^2\/2\)\)\ x\^2\)], "Output"]
}, Closed]],

Cell["\<\
does not work quite as its author probably intends.  For example,\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f[2]\)], "Input"],

Cell[BoxData[
    \(f[2]\)], "Output"]
}, Closed]],

Cell["does not return the same result as", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f[x] /. x \[Rule] 2. \)], "Input"],

Cell[BoxData[
    \(0.5413411329464508`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The problem is that we provided a definition for the argument ",
  StyleBox["x", "InlineInput",
    FontWeight->"Bold"],
  " but not for any other argument, such as"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f[y]\)], "Input"],

Cell[BoxData[
    \(f[y]\)], "Output"]
}, Closed]],

Cell["\<\
For a system based upon pattern recognition, there is not preventing us from \
supplying a completely independent definition \
\>", "Text"],

Cell[BoxData[
    \(\(f[y]\  = \ Sin[y\^3];\)\)], "Input"],

Cell[TextData[{
  "for an independent argument.  The definitions associated with the symbol \
",
  StyleBox["f", "InlineInput",
    FontWeight->"Bold"],
  " can be displayed using"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?f\)\)], "Input"],

Cell["Global`f", "Print",
  CellTags->"Info3303546715-7418521"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {GridBox[{
                {\(f[x] = \[ExponentialE]\^\(-\(x\^2\/2\)\)\ x\^2\)},
                {" "},
                {\(f[y] = Sin[y\^3]\)}
                },
              GridBaseline->{Baseline, {1, 1}},
              ColumnWidths->0.999,
              ColumnAlignments->{Left}]}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      Definition[ "f"],
      Editable->False]], "Print",
  CellTags->"Info3303546715-7418521"]
}, Closed]],

Cell[TextData[{
  "If we want our function to substitute any argument into a specific place \
in its definition, we need to formulate the definition using a ",
  StyleBox["blank pattern",
    FontSlant->"Italic"],
  " which will match any argument presented.  Thus,"
}], "Text"],

Cell[BoxData[
    \(\(g[x_] = \(x\^2\) Exp[\(-x\^2\)/2];\)\)], "Input"],

Cell["will produce the desired behavior.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(g[2. ]\)], "Input"],

Cell[BoxData[
    \(0.5413411329464508`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(g[y]\)], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(-\(y\^2\/2\)\)\ y\^2\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(g[a\ x\  + \ b\ y]\)], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(\(-\(1\/2\)\)\ \((a\ x + b\ y)\)\^2\)\ \((a\ x + b\ \
y)\)\^2\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Appending an underscore to the argument ",
  StyleBox["x_", "InlineInput",
    FontWeight->"Bold"],
  " on the left-hand side identifies the argument as a ",
  StyleBox["blank pattern",
    FontSlant->"Italic"],
  " that will be named ",
  StyleBox["x", "InlineInput",
    FontWeight->"Bold"],
  " on the right-hand side.  (Do not use the underscore on both sides.)"
}], "Text"],

Cell["\<\
Sometimes it is necessary to postpone evaluation of the right-hand side until \
an argument is presented.  For example, the function\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(h[x_, y_] = D[x, y]/x\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Closed]],

Cell["attempts to define the logarithmic derivative", "Text"],

Cell[BoxData[
    \(h[x, y] = \ \(1\/x\) \[PartialD]x\/\[PartialD]y\)], "DisplayFormula",
  FontFamily->"Times New Roman"],

Cell[TextData[{
  "but does not work correctly because the partial derivative of ",
  StyleBox["x", "InlineInput",
    FontWeight->"Bold"],
  " with respect to ",
  StyleBox["y", "InlineInput",
    FontWeight->"Bold"],
  " appears to vanish because we do not know yet how ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " will be related to ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  ".  An alternative version"
}], "Text"],

Cell[BoxData[
    \(h[x_, y_] := D[x, y]/x\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(h[Sin[x], x]\)], "Input"],

Cell[BoxData[
    \(Cot[x]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "does work correctly because the ",
  StyleBox[":=", "InlineInput",
    FontWeight->"Bold"],
  " operator, standing for ",
  StyleBox["SetDelayed", "InlineInput",
    FontWeight->"Bold"],
  ", delays evaluation of the right-hand side until the arguments are \
presented."
}], "Text"],

Cell["\<\
Most functions should be defined using blank patterns and SetDelayed, :=.\
\>", "Advice"]
}, Closed]],

Cell[CellGroupData[{

Cell["Postfix, Prefix, Infix", "Subsection"],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides a vast library of functions and you will soon create many of \
your own.  There are several methods for supplying input to functions.  The \
most obvious method, which was used above, is to use the form ",
  StyleBox["function[args,opts]", "InlineInput"],
  " where ",
  StyleBox["args", "InlineInput"],
  " represents the list of positional arguments and ",
  StyleBox["opts", "InlineInput"],
  " represents a subset of the available options.  Positional arguments are \
interpreted according to the order they appear in the function definition and \
must agree in type.  Options may follow the positional arguments, but are \
optional (in other words, they may be omitted).  Omitted options will be \
assigned default values.  "
}], "Text"],

Cell[TextData[{
  "For example, ",
  StyleBox["Expand", "InlineInput",
    FontWeight->"Bold"],
  " expands an algebraic expression. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Expand[\ \((a + b\ x)\) \((c\ x\  - \ d\ x\^2)\)\ \((e\  + \ f\  + \ 
            g)\)\ ]\)}], "Input"],

Cell[BoxData[
    \(a\ c\ e\ x + a\ c\ f\ x + a\ c\ g\ x + b\ c\ e\ x\^2 - a\ d\ e\ x\^2 + 
      b\ c\ f\ x\^2 - a\ d\ f\ x\^2 + b\ c\ g\ x\^2 - a\ d\ g\ x\^2 - 
      b\ d\ e\ x\^3 - b\ d\ f\ x\^3 - b\ d\ g\ x\^3\)], "Output"]
}, Closed]],

Cell["\<\
However, the expression we wish to expand often arises as the result of an \
earlier evaluation.  If we then want to use this method to apply another \
operation to that result, we would need to wrap a new function around the \
previous result.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[\ 
      Expand[\ \((a + b\ x)\) \((c\ x\  - \ d\ x\^2)\)\ \((e\  + \ f\  + \ 
              g)\)\ ]\ ]\)], "Input"],

Cell[BoxData[
    \(\((e + f + g)\)\ x\ \((a + b\ x)\)\ \((c - d\ x)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Although this obviously works, complicated algebraic manipulations may \
require many steps.  It then becomes difficult to read nested functions from \
the inside out.  I prefer to use the ",
  StyleBox["Postfix", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "form in which ",
  StyleBox["f[x]", "InlineInput"],
  " is represented by ",
  StyleBox["x//f", "InlineInput"],
  " where ",
  StyleBox["//", "InlineInput"],
  " is referred to as the postfix operator. ",
  StyleBox[" ",
    FontColor->RGBColor[1, 0, 1]],
  "For the example, above we pass an expression first into ",
  StyleBox["Expand", "InlineInput",
    FontWeight->"Bold"],
  " and then into ",
  StyleBox["Factor", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((a + b\ x)\) \((c\ x\  - \ d\ x\^2)\)\ \((e\  + \ f\  + \ g)\)\  // 
        Expand\) // Factor\)], "Input"],

Cell[BoxData[
    \(\((e + f + g)\)\ x\ \((a + b\ x)\)\ \((c - d\ x)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "As we develop our calculation, we may attach other functions or \
replacement rules at the end until we achieve the desired final form.  ",
  StyleBox["Postfix", "InlineInput"],
  " form is especially convenient for tacking numerical evaluation, \
simplification, or formatting functions onto the end of an expression.  I \
find this style easier to read and to understand than the nested \
representation, but of course that is merely a matter of taste."
}], "Text"],

Cell[TextData[{
  "Alternatively, it is sometimes useful to employ the ",
  StyleBox["Prefix", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "form based upon the prefix operator ",
  StyleBox["@", "InlineInput"],
  ", such that ",
  StyleBox["f[x]", "InlineInput"],
  " is represented by ",
  StyleBox["f@x", "InlineInput"],
  ".  The example above would then be expressed as"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(Factor\ @\ \(Expand@\((a + b\ x)\)\)\) \((c\ x\  - \ 
          d\ x\^2)\)\(\ \)\((e\  + \ f\  + \ g)\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \(\((e + f + g)\)\ \((a + b\ x)\)\ \((c\ x - d\ x\^2)\)\)], "Output"]
}, Closed]],

Cell["\<\
An important consideration in the use of postfix versus prefix notation is \
the precedence of operators.  The postfix operator is assigned a very low \
precedence so that the entire preceding expression is treated as single \
entity that is to be passed into the function.\
\>", "Text"],

Cell[BoxData[
    \(Clear[f]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(x + y // f\)], "Input"],

Cell[BoxData[
    \(f[x + y]\)], "Output"]
}, Closed]],

Cell["\<\
Because arithmetic and logical operators take precedence over postfix, the \
sum becomes the argument rather than the last term.  This behavior is to be \
contrasted with the prefix operator, which is assigned a higher precedence \
than arithmetic or logical operations.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f\ @\ x + y\)], "Input"],

Cell[BoxData[
    \(y + f[x]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Thus, with prefix notation the first term becomes the argument.  Note that \
",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " has its own rules for ordering the terms upon output.  ",
  StyleBox["Therefore, one must be very careful to consider precedence when \
using prefix notation.",
    FontColor->RGBColor[1, 0, 1]],
  "  I find the prefix form to be more prone to error than the postfix form.  \
However, even with postfix operations it is often necessary to use \
parentheses to ensure that the evaluation proceeds in the proper sequence."
}], "Text"],

Cell["Use explicit grouping (parentheses) to avoid mistakes.", "Advice"],

Cell[CellGroupData[{

Cell[BoxData[
    \(f@\((x + y)\)\)], "Input"],

Cell[BoxData[
    \(f[x + y]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Finally, ",
  StyleBox["Infix", "InlineInput",
    FontWeight->"Bold"],
  " notation can be used for functions which take two arguments.  In this \
case, ",
  StyleBox["f[x,y]", "InlineInput"],
  " is represented as ",
  StyleBox["x~f~y", "InlineInput"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(x~f~y\)], "Input"],

Cell[BoxData[
    \(f[x, y]\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Infix", "InlineInput"],
  " notation also takes precedence over arithmetic or logical operators."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(a + b~f~c + d\)], "Input"],

Cell[BoxData[
    \(a + d + f[b, c]\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Lists", "Section"],

Cell[CellGroupData[{

Cell["Creating and manipulating lists", "Subsection"],

Cell[TextData[{
  "Lists play a central role in data analysis.  Suppose that we create a \
simple list using ",
  ButtonBox["Range",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["min,max,increment", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ", such as"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(alist = Range[0, 1, 0.1]\)], "Input"],

Cell[BoxData[
    \({0, 0.1`, 0.2`, 0.30000000000000004`, 0.4`, 0.5`, 0.6000000000000001`, 
      0.7000000000000001`, 0.8`, 0.9`, 1.`}\)], "Output"]
}, Closed]],

Cell["The following operations should have obvious results.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(First[alist]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Last[alist]\)], "Input"],

Cell[BoxData[
    \(1.`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Length[alist]\)], "Input"],

Cell[BoxData[
    \(11\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Drop[alist, 2]\)], "Input"],

Cell[BoxData[
    \({0.2`, 0.30000000000000004`, 0.4`, 0.5`, 0.6000000000000001`, 
      0.7000000000000001`, 0.8`, 0.9`, 1.`}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Drop[alist, \(-2\)]\)], "Input"],

Cell[BoxData[
    \({0, 0.1`, 0.2`, 0.30000000000000004`, 0.4`, 0.5`, 0.6000000000000001`, 
      0.7000000000000001`, 0.8`}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Reverse[alist]\)], "Input"],

Cell[BoxData[
    \({1.`, 0.9`, 0.8`, 0.7000000000000001`, 0.6000000000000001`, 0.5`, 0.4`, 
      0.30000000000000004`, 0.2`, 0.1`, 0}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Take[alist, {3, 6}]\)], "Input"],

Cell[BoxData[
    \({0.2`, 0.30000000000000004`, 0.4`, 0.5`}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Drop[alist, {3, 6}]\)], "Input"],

Cell[BoxData[
    \({0, 0.1`, 0.6000000000000001`, 0.7000000000000001`, 0.8`, 0.9`, 
      1.`}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "A specific element of a list is obtained by double square brackets.  The \
typeset forms ",
  StyleBox["\[LeftDoubleBracket]\[RightDoubleBracket]", "InlineInput"],
  " are written using \[EscapeKey][[\[EscapeKey] and \[EscapeKey]]]\
\[EscapeKey]"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(alist[\([3]\)]\)], "Input"],

Cell[BoxData[
    \(0.2`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(alist\[LeftDoubleBracket]3\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \(0.2`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "A multidimensional list can be created using ",
  StyleBox["Table", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(blist = Table[i\ j\^2, {i, 1, 10}, {j, 1, 5}]\)], "Input"],

Cell[BoxData[
    \({{1, 4, 9, 16, 25}, {2, 8, 18, 32, 50}, {3, 12, 27, 48, 75}, {4, 16, 
        36, 64, 100}, {5, 20, 45, 80, 125}, {6, 24, 54, 96, 150}, {7, 28, 63, 
        112, 175}, {8, 32, 72, 128, 200}, {9, 36, 81, 144, 225}, {10, 40, 90, 
        160, 250}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dimensions[blist]\)], "Input"],

Cell[BoxData[
    \({10, 5}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(blist\[LeftDoubleBracket]2, 3\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \(18\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["MatrixForm", "InlineInput",
    FontWeight->"Bold"],
  " displays a two-dimensional list in the form of  matrix."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(MatrixForm[blist]\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"1", "4", "9", "16", "25"},
            {"2", "8", "18", "32", "50"},
            {"3", "12", "27", "48", "75"},
            {"4", "16", "36", "64", "100"},
            {"5", "20", "45", "80", "125"},
            {"6", "24", "54", "96", "150"},
            {"7", "28", "63", "112", "175"},
            {"8", "32", "72", "128", "200"},
            {"9", "36", "81", "144", "225"},
            {"10", "40", "90", "160", "250"}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Transpose[blist] // MatrixForm\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"1", "2", "3", "4", "5", "6", "7", "8", "9", "10"},
            {"4", "8", "12", "16", "20", "24", "28", "32", "36", "40"},
            {"9", "18", "27", "36", "45", "54", "63", "72", "81", "90"},
            {"16", "32", "48", "64", "80", "96", "112", "128", "144", "160"},
            {"25", "50", "75", "100", "125", "150", "175", "200", "225", 
              "250"}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Listable functions", "Subsection"],

Cell[TextData[{
  "Many ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions are ",
  StyleBox["listable",
    FontSlant->"Italic"],
  ", meaning that they take arguments in the form of lists.  Listable \
functions with two arguments, such as ",
  StyleBox["Plus", "InlineInput",
    FontWeight->"Bold"],
  " or ",
  StyleBox["Times", "InlineInput",
    FontWeight->"Bold"],
  ", can receive either a scalar and a list or two lists of the same \
structure.  Given a scalar and a list, a listable function of two arguments \
is evaluated for each element of the list paired with the scalar.  Given two \
lists of the same dimensions, a listable function of two arguments is \
evaluated using corresponding members of the two lists.  In both cases the \
output is a list of results."
}], "Text"],

Cell["\<\
Predict the output of each of the following operations before evaluating \
them.  Be sure to understand each result before moving on to the next!\
\>", "Exercise"],

Cell[BoxData[{
    \(\(list1 = {1, 2, 3, 4, 5};\)\), "\[IndentingNewLine]", 
    \(\(list2 = {a, b, c, d, e};\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(list1*x\)], "Input"],

Cell[BoxData[
    \({x, 2\ x, 3\ x, 4\ x, 5\ x}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(list1*list2\)], "Input"],

Cell[BoxData[
    \({a, 2\ b, 3\ c, 4\ d, 5\ e}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(list2/list1\)], "Input"],

Cell[BoxData[
    \({a, b\/2, c\/3, d\/4, e\/5}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(list1\^2\)], "Input"],

Cell[BoxData[
    \({1, 4, 9, 16, 25}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\@list2\)], "Input"],

Cell[BoxData[
    \({\@a, \@b, \@c, \@d, \@e}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(list2\^list1\)], "Input"],

Cell[BoxData[
    \({a, b\^2, c\^3, d\^4, e\^5}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((list1 + list2)\)\^2\)], "Input"],

Cell[BoxData[
    \({\((1 + a)\)\^2, \((2 + b)\)\^2, \((3 + c)\)\^2, \((4 + d)\)\^2, \((5 + \
e)\)\^2}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Application of a function to a list", "Subsection"],

Cell[TextData[{
  "A function can be applied to a list as a whole using ",
  StyleBox["Apply[f,list]", "InlineInput",
    FontWeight->"Bold"],
  " or ",
  StyleBox["f@@list", "InlineInput",
    FontWeight->"Bold"],
  ".  For example, the sum of squares for the first five counting numbers is"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Plus @@ \(Range[5]\^2\)\)], "Input"],

Cell[BoxData[
    \(55\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Alternatively, a function can be mapped onto the individual elements of a \
list using ",
  StyleBox["Map[f,list]", "InlineInput",
    FontWeight->"Bold"],
  " or ",
  StyleBox["f/@list", "InlineInput",
    FontWeight->"Bold"],
  ".  For example, the first five square roots are"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Sqrt /@ Range[5]\)], "Input"],

Cell[BoxData[
    \({1, \@2, \@3, 2, \@5}\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Contraction", "Subsection"],

Cell[TextData[{
  "The ",
  StyleBox["Dot", "InlineInput",
    FontWeight->"Bold"],
  " product between two vectors of  the same length is given by"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dot[list1, list2]\)], "Input"],

Cell[BoxData[
    \(a + 2\ b + 3\ c + 4\ d + 5\ e\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(list1 . list2\)], "Input"],

Cell[BoxData[
    \(a + 2\ b + 3\ c + 4\ d + 5\ e\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Similarly, multiplication between a matrix and a vector with compatible \
dimensions is also performed using ",
  StyleBox["Dot", "InlineInput",
    FontWeight->"Bold"],
  ".  The dot products can be described as ",
  StyleBox["contractions",
    FontSlant->"Italic"],
  " because the number of indices needed to enumerate the result is one less \
than the number of indices for the input. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({a, b, c, d} . {e, f, g, h}\)], "Input"],

Cell[BoxData[
    \(a\ e + b\ f + c\ g + d\ h\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \({{a, b, c}, {e, f, g}} . {x, y, z}\)], "Input"],

Cell[BoxData[
    \({a\ x + b\ y + c\ z, e\ x + f\ y + g\ z}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \({x, y} . {{a, b, c}, {e, f, g}}\)], "Input"],

Cell[BoxData[
    \({a\ x + e\ y, b\ x + f\ y, c\ x + g\ y}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell["example: variance", "Subsubsection"],

Cell[TextData[{
  "Suppose that we create a set of pseudodata (fake data) based upon a known \
function but with random fluctuations applied.  The following cell creates \
the data set, using ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " comment delimiters to explain each step."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\( (*\ 
      Create\ the\ parent\ \(\(function\)\(.\)\)\ \ \
*) \)\(\[IndentingNewLine]\)\(\(f[x_] = \(x\^2\) 
          Sin[x];\)\[IndentingNewLine]\[IndentingNewLine] (*\ 
      Create\ a\ set\ of\ equally - 
        spaced\ values\ for\ the\ independent\ \(\(variable\)\(.\)\)\ \
*) \[IndentingNewLine]
    \(xdata = 
        Range[\(-0.9\), 0.9, 
          0.1];\)\[IndentingNewLine]\[IndentingNewLine] (*\ 
      For\ each\ value\ of\ the\ independent\ variable, \ 
      compute\ the\ expected\ value\ of\ the\ measurement\ and\ add\ a\ \
random\ real\ number\ in\ the\ range\ \((\(-1\), \(+1\))\) . \ \ Note\ that\ \
listability\ is\ used\ both\ to\ evaluate\ f\ and\ to\ add\ a\ list\ of\ \
\(\(fluctuations\)\(.\)\)\ *) \[IndentingNewLine]\[IndentingNewLine]
    ydata = 
      f[xdata] + 
        Table[Random[Real, {\(-0.1\), 0.1}], {Length[xdata]}]\)\)\)], "Input"],

Cell[BoxData[
    \({\(-0.6396153771873664`\), \(-0.5093703002622796`\), \
\(-0.28574803833504625`\), \(-0.19928057406649513`\), \(-0.06220550225086635`\
\), \(-0.006254527916428693`\), \(-0.027397354039524625`\), 
      0.08981053051779507`, 0.06126445888347217`, 0.017291146460391454`, 
      0.016349263730213375`, \(-0.05664070987732542`\), 
      0.026772273453490137`, 0.07978634968050469`, 0.21313914476702586`, 
      0.24139129807287726`, 0.3530324754758935`, 0.5096014479457212`, 
      0.6483942976968665`}\)], "Output"]
}, Closed]],

Cell["The variance is defined by", "Text"],

Cell[BoxData[
    \(\[Sigma]\^2\  = \ 
      1\/n\ \(\[Sum]\+i\((y\_i - f[x\_i])\)\^2\)\)], "DisplayFormula",
  FontFamily->"Times New Roman"],

Cell[TextData[{
  "and can be evaluated in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " style as follows."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(residuals = ydata - f[xdata];\)\), "\[IndentingNewLine]", 
    \(variance = residuals . residuals/Length[residuals]\)}], "Input"],

Cell[BoxData[
    \(0.0022457056821015907`\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\[Sigma] = \@variance\)], "Input"],

Cell[BoxData[
    \(0.047388877198152636`\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Here ",
  StyleBox["residuals", "InlineInput",
    FontWeight->"Bold"],
  " is a vector containing the difference between observed and predicted \
values; it is contracted with itself (sum of squares) and divided by the \
number of elements to obtain the mean-square deviation between measurements \
and predictions.  The same result can be obtained in other ways, such as \
using ",
  StyleBox["Sum", "InlineInput",
    FontWeight->"Bold"],
  ", but this style is more appealing to the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " purist.  Also, note that you might want to save typing by abbreviating \
the name for residuals to ",
  StyleBox["r", "InlineInput",
    FontWeight->"Bold"],
  ", but the longer name will be more readable if you revisit a notebook \
after some time."
}], "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Creating ntuples", "Subsection"],

Cell[TextData[{
  "Suppose that we want to combine a list of ",
  Cell[BoxData[
      \(TraditionalForm\`{x\_i}\)]],
  " and a list of ",
  Cell[BoxData[
      \(TraditionalForm\`{y\_i}\)]],
  ", each containing ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " elements into a list of pairs ",
  Cell[BoxData[
      \(TraditionalForm\`{{x\_i, y\_i}, \[CenterEllipsis]}\)]],
  ".  The solution is to use ",
  ButtonBox["Transpose",
    ButtonStyle->"RefGuideLink"],
  " to transform the ",
  Cell[BoxData[
      \(TraditionalForm\`2\[Times]n\)]],
  " list ",
  Cell[BoxData[
      \(TraditionalForm\`{x, y}\)]],
  " into an ",
  Cell[BoxData[
      \(TraditionalForm\`n\[Times]2\)]],
  " list."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(data = Transpose[{xdata, ydata}]\)], "Input"],

Cell[BoxData[
    \({{\(-0.9`\), \(-0.6396153771873664`\)}, {\(-0.8`\), \
\(-0.5093703002622796`\)}, {\(-0.7`\), \(-0.28574803833504625`\)}, \
{\(-0.6`\), \(-0.19928057406649513`\)}, {\(-0.5`\), \
\(-0.06220550225086635`\)}, {\(-0.4`\), \(-0.006254527916428693`\)}, {\(-0.3`\
\), \(-0.027397354039524625`\)}, {\(-0.19999999999999998`\), 
        0.08981053051779507`}, {\(-0.09999999999999998`\), 
        0.06126445888347217`}, {2.7755575615628914`*^-17, 
        0.017291146460391454`}, {0.10000000000000003`, 
        0.016349263730213375`}, {0.20000000000000004`, \
\(-0.05664070987732542`\)}, {0.30000000000000004`, 
        0.026772273453490137`}, {0.4`, 0.07978634968050469`}, {0.5`, 
        0.21313914476702586`}, {0.6000000000000001`, 
        0.24139129807287726`}, {0.7000000000000001`, 
        0.3530324754758935`}, {0.8`, 0.5096014479457212`}, {0.9`, 
        0.6483942976968665`}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Dimensions[data]\)], "Input"],

Cell[BoxData[
    \({19, 2}\)], "Output"]
}, Closed]],

Cell["\<\
Similarly, if we already have a list of pairs, vectors can be extracted using\
\
\>", "Text"],

Cell[BoxData[
    \(\({a, b} = Transpose[data];\)\)], "Input"],

Cell["and below we verify that the original data are recovered. ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(a \[Equal] xdata\  && \ b \[Equal] ydata\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Packages", "Section"],

Cell[TextData[{
  "One of the most important features of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " is that it is an extensible system.  In addition to the built-in \
functions available to any ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " session, one can add new features by loading a package that contains the \
necessary definitions.  A ",
  StyleBox["package",
    FontSlant->"Italic"],
  " is a special kind of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " file that contains the definitions and specifies the rules for \
manipulating symbols and functions relevant to a particular purpose.  \
However, it is important to load the package before you attempt to use the \
functions it contains \[LongDash] failure to load the package first creates \
global symbols that override the name of the function that you want.  If you \
inadvertently invoke a function before its package is loaded, use ",
  StyleBox["Remove", "InlineInput",
    FontWeight->"Bold"],
  " to get rid of the nonfunctioning symbol and then load the package; the \
desired functions should then work properly. "
}], "Text"],

Cell[TextData[{
  "The best way to load a package is to use the ",
  ButtonBox["Needs",
    ButtonStyle->"RefGuideLink"],
  " function, which reads the package if it is not yet on the context path.  \
The argument is a string of the form ",
  StyleBox["\"Category`Package`\"", "InlineInput",
    FontWeight->"Bold"],
  ", where the first part identifies the folder and the second the file.  \
Note that the back-quotes are essential.  Many categories provide a master \
package that provides access to all of the functions in all of its related \
packages without loading each package individual.  Loading the master package \
does not burden the system memory unnecessarily because the associated code \
is not loaded until it is needed.   "
}], "Text"],

Cell["\<\
Before we load any packages, let's ask what types of plots and probability \
distributions are available as standard built-in functions.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Plot*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ArrayPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`ArrayPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`ListContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`ListPlot3D"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`ParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3Matrix",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`Plot3Matrix"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotLabel",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`PlotLabel"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRegion",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`PlotRegion"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["ContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`ContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`ListDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["MaxPlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`MaxPlotPoints"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`Plot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotDivision",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`PlotDivision"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`PlotPoints"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotStyle",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`PlotStyle"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["DensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`DensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`ListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237299-4558179", 
                    "System`ParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`Plot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotJoined",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`PlotJoined"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRange",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237299-4558179", "System`PlotRange"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1.05}]}
        },
      RowSpacings->{2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Distr*\)\)], "Input"],

Cell[BoxData[
    RowBox[{"\<\"Distribute[f[x1, x2, ... ]] distributes f over Plus \
appearing in any of the xi. Distribute[expr, g] distributes over g. \
Distribute[expr, g, f] performs the distribution only if the head of expr is \
f.\"\>", " ", 
      ButtonBox[
        StyleBox["More\[Ellipsis]",
          "SR"],
        ButtonData:>"Distribute",
        Active->True,
        ButtonStyle->"RefGuideLink"]}]], "Print",
  CellTags->"Info3351237306-3667888"]
}, Closed]],

Cell["\<\
We have a relatively small number of plotting functions (some of the symbols \
are options, not functions) and no probability distributions available at \
this time.  Prior to Version 5.1, the following cell would load packages\
\>", "Text"],

Cell[BoxData[{
    \(\(Needs["\<Graphics`Master`\>"];\)\), "\[IndentingNewLine]", 
    \(\(Needs["\<Statistics`Master`\>"];\)\)}], "Input",
  Evaluatable->False],

Cell["while with later versions it is sufficient to enter", "Text"],

Cell[BoxData[{
    \(\(Needs["\<Graphics`\>"];\)\), "\[IndentingNewLine]", 
    \(\(Needs["\<Statistics`\>"];\)\)}], "Input"],

Cell["\<\
to obtain a expanded arsenal of graphical and statistical functions.  (Note \
that we disabled the evaluation of the early version.)  The lists of \
available plots\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Plot*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ArrayPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`ArrayPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`ListContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`ListPlot3D"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`ParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3Matrix",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`Plot3Matrix"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotLabel",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`PlotLabel"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRegion",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`PlotRegion"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["ContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`ContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`ListDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["MaxPlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`MaxPlotPoints"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`Plot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotDivision",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`PlotDivision"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotPoints",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`PlotPoints"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotStyle",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`PlotStyle"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"]},
              {
                ButtonBox[
                  StyleBox["DensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`DensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`ListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["ParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "System`ParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["Plot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`Plot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotJoined",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`PlotJoined"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], 
                ButtonBox[
                  StyleBox["PlotRange",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", "System`PlotRange"},
                  ButtonFrame->"None",
                  ButtonNote->"System`"], ""}
              },
            RowMinHeight->{1, 1, 1.05}]},
        {
          StyleBox["Graphics`Animation`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["MovieContourPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Animation`MovieContourPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MovieDensityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Animation`MovieDensityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MovieParametricPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Animation`MovieParametricPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MoviePlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Animation`MoviePlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], 
                ButtonBox[
                  StyleBox["MoviePlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Animation`MoviePlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Animation`"], ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`ContourPlot3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ContourPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`ContourPlot3D`ContourPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ContourPlot3D`"], 
                ButtonBox[
                  StyleBox["ListContourPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`ContourPlot3D`ListContourPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ContourPlot3D`"], "", "", "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`FilledPlot`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["FilledListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`FilledPlot`FilledListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`FilledPlot`"], 
                ButtonBox[
                  StyleBox["FilledPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`FilledPlot`FilledPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`FilledPlot`"], 
                ButtonBox[
                  StyleBox["ListFilledPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`FilledPlot`ListFilledPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`FilledPlot`"], "", "", "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`Graphics`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ErrorListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`ErrorListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LinearLogPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LinearLogPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLinearPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LogLinearPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLogPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LogLogPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["PolarPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`PolarPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["TextListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`TextListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], ""},
              {
                ButtonBox[
                  StyleBox["LabeledListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LabeledListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["ListAndCurvePlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`ListAndCurvePlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LogListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LogPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["ScaledListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`ScaledListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], "", ""},
              {
                ButtonBox[
                  StyleBox["LinearLogListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LinearLogListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLinearListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LogLinearListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["LogLogListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`LogLogListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["PolarListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`PolarListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], 
                ButtonBox[
                  StyleBox["ScaledPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics`ScaledPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics`"], "", ""}
              },
            RowMinHeight->{1, 1, 1.05}]},
        {
          StyleBox["Graphics`Graphics3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ListShadowPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics3D`ListShadowPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], 
                ButtonBox[
                  StyleBox["ListSurfacePlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics3D`ListSurfacePlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], 
                ButtonBox[
                  StyleBox["ScatterPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics3D`ScatterPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], 
                ButtonBox[
                  StyleBox["ShadowPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Graphics3D`ShadowPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Graphics3D`"], "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`ImplicitPlot`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ImplicitPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`ImplicitPlot`ImplicitPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ImplicitPlot`"], "", "", "", "", "", 
                "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`InequalityGraphics`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ComplexInequalityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`InequalityGraphics`ComplexInequalityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`InequalityGraphics`"], 
                ButtonBox[
                  StyleBox["InequalityPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`InequalityGraphics`InequalityPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`InequalityGraphics`"], 
                ButtonBox[
                  StyleBox["InequalityPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`InequalityGraphics`InequalityPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`InequalityGraphics`"], "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`Legend`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["PlotLegend",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`Legend`PlotLegend"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`Legend`"], "", "", "", "", "", "", "",
                 "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`MultipleListPlot`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["MultipleListPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`MultipleListPlot`MultipleListPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`MultipleListPlot`"], 
                ButtonBox[
                  StyleBox["PlotSymbol",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`MultipleListPlot`PlotSymbol"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`MultipleListPlot`"], "", "", "", "", 
                ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`ParametricPlot3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["CylindricalPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`ParametricPlot3D`CylindricalPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ParametricPlot3D`"], 
                ButtonBox[
                  StyleBox["PointParametricPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`ParametricPlot3D`PointParametricPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ParametricPlot3D`"], 
                ButtonBox[
                  StyleBox["SphericalPlot3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`ParametricPlot3D`SphericalPlot3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`ParametricPlot3D`"], "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`PlotField`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ListPlotVectorField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField`ListPlotVectorField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotGradientField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField`PlotGradientField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotHamiltonianField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField`PlotHamiltonianField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotPolyaField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField`PlotPolyaField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], 
                ButtonBox[
                  StyleBox["PlotVectorField",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField`PlotVectorField"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField`"], ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Graphics`PlotField3D`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ListPlotVectorField3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField3D`ListPlotVectorField3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField3D`"], 
                ButtonBox[
                  StyleBox["PlotGradientField3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField3D`PlotGradientField3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField3D`"], 
                ButtonBox[
                  StyleBox["PlotVectorField3D",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Graphics`PlotField3D`PlotVectorField3D"},
                  ButtonFrame->"None",
                  ButtonNote->"Graphics`PlotField3D`"], "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Statistics`ClusterAnalysis`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["DendrogramPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Statistics`ClusterAnalysis`DendrogramPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ClusterAnalysis`"], "", "", "", "", 
                "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Statistics`StatisticsPlots`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["BoxWhiskerPlot",
                    "Hyperlink"],
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Statistics`StatisticsPlots`BoxWhiskerPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`StatisticsPlots`"], 
                ButtonBox[
                  StyleBox["PairwiseScatterPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Statistics`StatisticsPlots`PairwiseScatterPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`StatisticsPlots`"], 
                ButtonBox[
                  StyleBox["ParetoPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Statistics`StatisticsPlots`ParetoPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`StatisticsPlots`"], 
                ButtonBox[
                  StyleBox["QuantilePlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Statistics`StatisticsPlots`QuantilePlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`StatisticsPlots`"], 
                ButtonBox[
                  StyleBox["StemLeafPlot",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237524-6943720", 
                    "Statistics`StatisticsPlots`StemLeafPlot"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`StatisticsPlots`"], ""}
              },
            RowMinHeight->1.05]}
        },
      RowSpacings->{
        2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 
          3, 2, 3, 2, 3, 2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell["and probability distributions", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*Distr*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["System`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["Distribute",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351237528-9912737", "System`Distribute"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"System`"], "", "", "", "", "", "", "", "", "", 
                ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Statistics`ContinuousDistributions`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["BetaDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`BetaDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["GammaDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`GammaDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["NoncentralChiSquareDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
NoncentralChiSquareDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["UniformDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`UniformDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"]},
              {
                ButtonBox[
                  StyleBox["CauchyDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`CauchyDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["HalfNormalDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
HalfNormalDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["NoncentralFRatioDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
NoncentralFRatioDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["WeibullDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`WeibullDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"]},
              {
                ButtonBox[
                  StyleBox["ChiDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`ChiDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["LaplaceDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`LaplaceDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["NoncentralStudentTDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
NoncentralStudentTDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], ""},
              {
                ButtonBox[
                  StyleBox["ExponentialDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
ExponentialDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["LogisticDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`LogisticDistribution"}\
,
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["ParetoDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`ParetoDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], ""},
              {
                ButtonBox[
                  StyleBox["ExtremeValueDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
ExtremeValueDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["LogNormalDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`\
LogNormalDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], 
                ButtonBox[
                  StyleBox["RayleighDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`ContinuousDistributions`RayleighDistribution"}\
,
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`ContinuousDistributions`"], ""}
              },
            RowMinHeight->{1, 1, 1, 1, 1.05}]},
        {
          StyleBox["Statistics`DiscreteDistributions`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["BernoulliDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`BernoulliDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["DiscreteUniformDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`\
DiscreteUniformDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["HypergeometricDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`\
HypergeometricDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["NegativeBinomialDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`\
NegativeBinomialDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"]},
              {
                ButtonBox[
                  StyleBox["BinomialDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`BinomialDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["GeometricDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`GeometricDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["LogSeriesDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`LogSeriesDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["PoissonDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`DiscreteDistributions`PoissonDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`DiscreteDistributions`"]}
              },
            RowMinHeight->{1, 1.05}]},
        {
          StyleBox["Statistics`MultiDiscreteDistributions`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["MultinomialDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultiDiscreteDistributions`\
MultinomialDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultiDiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["MultiPoissonDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultiDiscreteDistributions`\
MultiPoissonDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultiDiscreteDistributions`"], 
                ButtonBox[
                  StyleBox["NegativeMultinomialDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultiDiscreteDistributions`\
NegativeMultinomialDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultiDiscreteDistributions`"], ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Statistics`MultinormalDistribution`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["HotellingTSquareDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultinormalDistribution`\
HotellingTSquareDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultinormalDistribution`"], 
                ButtonBox[
                  StyleBox["MultivariateTDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultinormalDistribution`\
MultivariateTDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultinormalDistribution`"], 
                ButtonBox[
                  StyleBox["WishartDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultinormalDistribution`WishartDistribution"},
                  
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultinormalDistribution`"], ""},
              {
                ButtonBox[
                  StyleBox["MultinormalDistribution",
                    "Hyperlink"],
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultinormalDistribution`\
MultinormalDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultinormalDistribution`"], 
                ButtonBox[
                  StyleBox["QuadraticFormDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`MultinormalDistribution`\
QuadraticFormDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`MultinormalDistribution`"], "", ""}
              },
            RowMinHeight->{1, 1.05}]},
        {
          StyleBox["Statistics`NormalDistribution`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["ChiSquareDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`NormalDistribution`ChiSquareDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`NormalDistribution`"], 
                ButtonBox[
                  StyleBox["FRatioDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`NormalDistribution`FRatioDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`NormalDistribution`"], 
                ButtonBox[
                  StyleBox["NormalDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`NormalDistribution`NormalDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`NormalDistribution`"], 
                ButtonBox[
                  StyleBox["StudentTDistribution",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  
                  ButtonData:>{"Info3351237528-9912737", 
                    "Statistics`NormalDistribution`StudentTDistribution"},
                  ButtonFrame->"None",
                  ButtonNote->"Statistics`NormalDistribution`"], ""}
              },
            RowMinHeight->1.05]}
        },
      RowSpacings->{2, 3, 2, 3, 2, 3, 2, 3, 2, 3, 2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}]
}, Closed]],

Cell["\<\
are now much larger.  We will not need most of  these, but some will be quite \
useful for our work.  Clicking any of the links will provide more information \
on a specific symbol and a link to online documentation for further \
information.   (Try it!)\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Initialization section", "Section"],

Cell[TextData[{
  "Most of my notebooks begin with a section named Initialization that \
contains cells which select some useful options and load the packages needed \
by that notebook.  The cells below provide a somewhat delayed example \
\[LongDash] normally they would be at the top, but would have been mysterious \
to new users of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  ".  The first cell clears all global variables, giving a clean start, and \
turns off spell-checking warnings that I find annoying; you might want to \
omit that if your typing is error prone.  The next cell appends the directory \
containing this notebook, a sample data file ",
  StyleBox["gep.dat",
    FontSlant->"Italic"],
  ", and a new package ",
  StyleBox["Curfit.m",
    FontSlant->"Italic"],
  " to ",
  ButtonBox["$Path",
    ButtonStyle->"RefGuideLink"],
  ", which is a default list of directories used to find external files.  \
This will save typing very long directory names, but you will need to modify \
this directory name for your own system.  ",
  "The next cell affects the fonts used in graphics output.  Finally, we load \
the packages needed by this notebook.  Sometimes this section might include \
other packages or definitions for some of the functions and rules that will \
be used extensively later on.  The contents of this section will depend upon \
your preferences and could be more extensive, but will usually begin with \
something similar to this example."
}], "Text"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Off[General::spell, General::spell1]\)}], "Input"],

Cell[BoxData[
    \(\(AppendTo[$Path, "\<C:\Documents and Settings\Jim\My Documents\My \
eBooks\EssentialMathematica\Data analysis\>"];\)\)], "Input"],

Cell[BoxData[{
    \(\($DefaultFont = {"\<Times\>", 12};\)\), "\n", 
    \(\($TextStyle = {FontFamily \[Rule] "\<Times\>", FontSize \[Rule] 12, 
          FontSlant \[Rule] "\<Italic\>"};\)\)}], "Input"],

Cell[BoxData[{
    \(\(Needs["\<Graphics`\>"];\)\), "\[IndentingNewLine]", 
    \(\(Needs["\<Statistics`\>"];\)\)}], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Plotting", "Section"],

Cell[CellGroupData[{

Cell["Curves", "Subsection"],

Cell[TextData[{
  "Obviously, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides a large assortment of plotting functions.  We only have time to \
introduce a few of the most useful and refer you to the Help Browser or to ",
  ButtonBox["plotting.nb",
    ButtonData:>{
      "www.physics.umd.edu/courses/CourseWare/EssentialMathematica/Primary \
topics/plotting.nb", None},
    ButtonStyle->"Hyperlink"],
  " for further information.  The basic syntax for plotting a function of a \
single variable as a curve is"
}], "Text"],

Cell[TextData[{
  ButtonBox["Plot",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[f[x],{x,xmin,xmax,opts]", "InlineInput"]
}], "Text",
  TextAlignment->Left,
  TextJustification->0],

Cell[TextData[{
  "where ",
  StyleBox["f[x]", "InlineInput"],
  " is any built-in or user-defined function which evaluates to a real number \
given a numerical argument ",
  StyleBox["x", "InlineInput"],
  " between ",
  StyleBox["xmin", "InlineInput"],
  " and ",
  StyleBox["xmax", "InlineInput"],
  ", and where ",
  StyleBox["opts", "InlineInput"],
  " is a sequence of options given in the form of replacement rules.  Thus,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[ArcSin[x], {x, \(-1\), 1}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.187407 [
[.02381 .29652 -9.34375 -15.6875 ]
[.02381 .29652 9.34375 0 ]
[.2619 .29652 -13.8438 -15.6875 ]
[.2619 .29652 13.8438 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .02791 -27.6875 -7.84375 ]
[.4875 .02791 0 7.84375 ]
[.4875 .12161 -18.6875 -7.84375 ]
[.4875 .12161 0 7.84375 ]
[.4875 .21531 -27.6875 -7.84375 ]
[.4875 .21531 0 7.84375 ]
[.4875 .40272 -19 -7.84375 ]
[.4875 .40272 0 7.84375 ]
[.4875 .49642 -10 -7.84375 ]
[.4875 .49642 0 7.84375 ]
[.4875 .59013 -19 -7.84375 ]
[.4875 .59013 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .02791 m
.50625 .02791 L
s
gsave
.4875 .02791 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.5 .12161 m
.50625 .12161 L
s
gsave
.4875 .12161 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .21531 m
.50625 .21531 L
s
gsave
.4875 .21531 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .40272 m
.50625 .40272 L
s
gsave
.4875 .40272 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .49642 m
.50625 .49642 L
s
gsave
.4875 .49642 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.5 .59013 m
.50625 .59013 L
s
gsave
.4875 .59013 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .04665 m
.50375 .04665 L
s
.5 .06539 m
.50375 .06539 L
s
.5 .08413 m
.50375 .08413 L
s
.5 .10287 m
.50375 .10287 L
s
.5 .14035 m
.50375 .14035 L
s
.5 .15909 m
.50375 .15909 L
s
.5 .17783 m
.50375 .17783 L
s
.5 .19657 m
.50375 .19657 L
s
.5 .23405 m
.50375 .23405 L
s
.5 .25279 m
.50375 .25279 L
s
.5 .27154 m
.50375 .27154 L
s
.5 .29028 m
.50375 .29028 L
s
.5 .32776 m
.50375 .32776 L
s
.5 .3465 m
.50375 .3465 L
s
.5 .36524 m
.50375 .36524 L
s
.5 .38398 m
.50375 .38398 L
s
.5 .42146 m
.50375 .42146 L
s
.5 .4402 m
.50375 .4402 L
s
.5 .45894 m
.50375 .45894 L
s
.5 .47768 m
.50375 .47768 L
s
.5 .51516 m
.50375 .51516 L
s
.5 .53391 m
.50375 .53391 L
s
.5 .55265 m
.50375 .55265 L
s
.5 .57139 m
.50375 .57139 L
s
.5 .00917 m
.50375 .00917 L
s
.5 .60887 m
.50375 .60887 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .02781 L
.02605 .03284 L
.02846 .04084 L
.03279 .05108 L
.04262 .06749 L
.05281 .08038 L
.06244 .09065 L
.10458 .1254 L
.14299 .15017 L
.18386 .17296 L
.22321 .19277 L
.26105 .2105 L
.30133 .22836 L
.34011 .24484 L
.38133 .26182 L
.42104 .2778 L
.45923 .29295 L
.49988 .30897 L
.53901 .32439 L
.58059 .34089 L
.62066 .35702 L
.65921 .3729 L
.70021 .39034 L
.7397 .40787 L
.77767 .42569 L
.81809 .44611 L
.857 .46787 L
.89836 .49473 L
.91906 .51065 L
.93821 .52804 L
.95662 .54949 L
.96592 .56441 L
.96847 .56961 L
.97122 .57629 L
.97248 .57999 L
.97381 .58466 L
.97495 .58986 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00
f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool300000=@0oooo003I0?ooo`030000003oool0
oooo0=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo000>0?ooo`030000003oool0oooo0<P0
oooo00<000000?ooo`3oool0e03oool000h0oooo00<000000?ooo`3oool0Z@3oool00`3ookl0M000
000000020000000704RLg`3oool0oooo0?ooo`3ogi`0B00007Boo`020?ooo`060?oOW01800000000
00000000B9`0gooo2`3oool00`000000oooo0?ooo`3D0?ooo`003P3oool00`000000oooo0?ooo`2[
0?ooo`030?noM0000000M;oo00@0oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3o
ool0omnL04Qd_`/0oooo00<000000?ooo`3oool0e03oool000h0oooo00<000000?ooo`3oool0Z`3o
ool00`3ooml0W4P0001d_`0>0?ooo`030?noM018W=l0oooo00T0oooo00<000000?ooo`3oool0e03o
ool000h0oooo00<000000?ooo`3oool0[03oool00`3OW4P004RL0=ooo`0=0?ooo`030?oo_`1d07@0
_ooo00T0oooo00<000000?ooo`3oool0e03oool000h0oooo00<000000?ooo`3oool0W03oool00`3o
oml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0omnL04P0B02Lgol03@3oool00`3o
oml0W4P007Boo`090?ooo`@00000d`3oool000h0oooo00<000000?ooo`3oool0[03oool00`3ookl0
M00007Boo`0=0?ooo`030?oo_`1d0000B9cO00T0oooo00<000000?ooo`3oool0e03oool000h0oooo
00<000000?ooo`3oool0[@3oool00`2oM00007Bo0?ooo`0<0?ooo`030?noM0000000M;oo00T0oooo
00<000000?ooo`3oool0e03oool000h0oooo00<000000?ooo`3oool0[@3oool00`3OW4P004RL0=oo
o`090?ooo`060?oOg`3Oool0okmd0000000007@0_ooo2@3oool00`000000oooo0?ooo`3D0?ooo`00
3`3oool00`000000oooo0?ooo`2/0?ooo`030?oOW01804P0W=oo00T0oooo00D0oono07@000000000
001809cOo`0:0?ooo`030000003oool0oooo0=@0oooo000?0?ooo`030000003oool0oooo0:`0oooo
00<0oono07@0001d_ol02P3oool00`3OW9`0_ooo0?ooo`0;0?ooo`030000003oool0oooo0=@0oooo
000@0?ooo`030000003oool0oooo0:`0oooo00<0_g@0001d_`3oool02@3oool00`3ogi`0M9cO0?oo
o`0;0?ooo`030000003oool0oooo0=@0oooo000@0?ooo`030000003oool0oooo0:X0oooo00D0oono
07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l0203oool00`000000oooo
0?ooo`3D0?ooo`00403oool00`000000oooo0?ooo`360?ooo`<00000e03oool00140oooo00<00000
0?ooo`3oool0a@3oool00`000000oooo0?ooo`3D0?ooo`004@3oool00`000000oooo0?ooo`350?oo
o`030000003oool0oooo0=@0oooo000B0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3o
ool0e03oool00180oooo00<000000?ooo`3oool0a03oool00`000000oooo0?ooo`3D0?ooo`004`3o
ool00`000000oooo0?ooo`330?ooo`030000003oool0oooo0=@0oooo000D0?ooo`030000003oool0
oooo0<80oooo00<000000?ooo`3oool0e03oool001D0oooo00<000000?ooo`3oool0`@3oool00`00
0000oooo0?ooo`3D0?ooo`005@3oool00`000000oooo0?ooo`310?ooo`<00000e03oool001H0oooo
00<000000?ooo`3oool0`03oool00`000000oooo0?ooo`3D0?ooo`005`3oool00`000000oooo0?oo
o`2o0?ooo`030000003oool0oooo0=@0oooo000G0?ooo`030000003oool0oooo0;l0oooo00<00000
0?ooo`3oool0e03oool001P0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?ooo`3D0?oo
o`006@3oool00`000000oooo0?ooo`2m0?ooo`030000003oool0oooo0=@0oooo000I0?ooo`030000
003oool0oooo0;d0oooo00<000000?ooo`3oool0e03oool001X0oooo00<000000?ooo`3oool0_03o
ool00`000000oooo0?ooo`3D0?ooo`006`3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0
oooo0=@0oooo000L0?ooo`030000003oool0oooo0;X0oooo0`00003D0?ooo`007@3oool00`000000
oooo0?ooo`2i0?ooo`030000003oool0oooo0=@0oooo000N0?ooo`030000003oool0oooo0;P0oooo
00<000000?ooo`3oool0e03oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?oo
o`3D0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0=@0oooo000Q0?oo
o`800000]P3oool00`000000oooo0?ooo`3D0?ooo`008`3oool00`000000oooo0?ooo`2c0?ooo`03
0000003oool0oooo0=@0oooo000T0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0
e03oool002D0oooo00<000000?ooo`3oool0/@3oool300000=@0oooo000V0?ooo`030000003oool0
oooo0;00oooo00<000000?ooo`3oool0e03oool002L0oooo0P00002`0?ooo`030000003oool0oooo
0=@0oooo000Y0?ooo`030000003oool0oooo09`0oooo00<0oono07@0000000000P0000000`18W=l0
oooo0?ooo`090?ooo`030000003oool0oooo0=@0oooo000Z0?ooo`030000003oool0oooo09d0oooo
00<0okmd0000001d_ol0303oool00`000000oooo0?ooo`3D0?ooo`00:`3oool00`000000oooo0?oo
o`2L0?ooo`030?oog`2LB00007Bo00`0oooo00<000000?ooo`3oool0e03oool002`0oooo00<00000
0?ooo`3oool0W03oool00`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo0=@0oooo000]0?oo
o`800000S03oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0omnL04P0
B02Lgol02`3oool00`000000oooo0?ooo`3D0?ooo`00;`3oool00`000000oooo0?ooo`2I0?ooo`03
0?oo_`1d0000M;oo00/0oooo1000003C0?ooo`00<03oool00`000000oooo0?ooo`2I0?ooo`030;md
0000M;l0oooo00X0oooo00<000000?ooo`3oool0e03oool00340oooo0P00002I0?ooo`030=nLB000
B9`0gooo00X0oooo00<000000?ooo`3oool0e03oool003<0oooo00<000000?ooo`3oool0UP3oool0
0`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo0=@0oooo000d0?ooo`800000UP3oool00`3o
okl0M00007Boo`0:0?ooo`030000003oool0oooo0=@0oooo000f0?ooo`030000003oool0oooo09@0
oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`3D0?ooo`00=`3oool00`000000oooo
0?ooo`2A0?ooo`050?oo_`1d000000000000M02oool02@3oool00`000000oooo0?ooo`3D0?ooo`00
>03oool2000009l0oooo00<000000?ooo`3oool0e03oool003X0oooo00<000000?ooo`3oool0W03o
ool300000=@0oooo000k0?ooo`800000W03oool00`000000oooo0?ooo`3D0?ooo`00?@3oool00`00
0000oooo0?ooo`2I0?ooo`030000003oool0oooo0=@0oooo000n0?ooo`800000V@3oool00`000000
oooo0?ooo`3D0?ooo`00@03oool00`000000oooo0?ooo`2F0?ooo`030000003oool0oooo0=@0oooo
00110?ooo`800000UP3oool00`000000oooo0?ooo`3D0?ooo`00@`3oool2000009@0oooo00<00000
0?ooo`3oool0e03oool004D0oooo0P00002B0?ooo`030000003oool0oooo0=@0oooo00170?ooo`80
0000T03oool300000=@0oooo00190?ooo`800000SP3oool00`000000oooo0?ooo`3D0?ooo`00B`3o
ool2000008`0oooo00<000000?ooo`3oool0e03oool004d0oooo0P00002:0?ooo`030000003oool0
oooo0=@0oooo001?0?ooo`800000R03oool00`000000oooo0?ooo`3D0?ooo`00D@3oool2000008H0
oooo00<000000?ooo`3oool0e03oool005<0oooo0P0000240?ooo`030000003oool0oooo0=@0oooo
001E0?ooo`800000PP3oool00`000000oooo0?ooo`3D0?ooo`00E`3oool200000800oooo0`00003D
0?ooo`00F@3oool2000007h0oooo00<000000?ooo`3oool0e03oool005/0oooo0P00001l0?ooo`03
0000003oool0oooo0=@0oooo001M0?ooo`800000NP3oool00`000000oooo0?ooo`3D0?ooo`00G`3o
ool2000007P0oooo00<000000?ooo`3oool0e03oool00640oooo0P00001f0?ooo`030000003oool0
oooo0=@0oooo001S0?ooo`800000M03oool00`000000oooo0?ooo`3D0?ooo`00I@3oool200000780
oooo00<000000?ooo`3oool0e03oool006L0oooo0P00001`0?ooo`<00000e03oool006T0oooo0P00
001^0?ooo`030000003oool0oooo0=@0oooo001[0?ooo`<00000J`3oool00`000000oooo0?ooo`3D
0?ooo`00KP3oool2000004`0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`02
0?ooo`060?oOW0180000000000000000B9`0gooo2`3oool00`000000oooo0?ooo`3D0?ooo`00L03o
ool2000004T0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080
oooo00H0okmd0000001d_ol0oooo0?oOW018M;l;0?ooo`030000003oool0oooo0=@0oooo001b0?oo
o`<00000A@3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RL
g`3oool02@3oool00`000000oooo0?ooo`3D0?ooo`00M@3oool2000004<0oooo00<0oooO09a80018
W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d07@0_ooo00T0oooo00<000000?oo
o`3oool0e03oool007L0oooo0P00000d0?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3o
ool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8001d_ol02@3oool0
0`000000oooo0?ooo`3D0?ooo`00N@3oool200000400oooo00<0_g@00018W03Oool00P3oool00`3o
gi`0B00004RLg`0:0?ooo`030?oo_`1d0000B9cO00T0oooo1000003C0?ooo`00N`3oool2000003h0
oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo00T0
oooo00<000000?ooo`3oool0e03oool007d0oooo0P00000l0?ooo`030?oOW0180000B9cO00<0oooo
00<0_g@00018W03Oool01P3oool01P3ogml0gooo0?noM0000000001d0;ooo`T0oooo00<000000?oo
o`3oool0e03oool007l0oooo0P00000j0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018
W03Oool01P3oool01@3ookl0M0000000000004P0W=oo00X0oooo00<000000?ooo`3oool0e03oool0
0840oooo0P00000i0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3O
W9`0_ooo0?ooo`0;0?ooo`030000003oool0oooo0=@0oooo00230?ooo`<00000=`3oool01P3OW4P0
001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool02`3oool00`000000oooo0?ooo`3D
0?ooo`00QP3oool2000003H0oooo00@0gia800000000000007Bo2P3oool01@2oM000000000000000
0000B9cO00P0oooo00<000000?ooo`3oool0e03oool008P0oooo0`00001>0?ooo`030000003oool0
oooo0=@0oooo002;0?ooo`800000C03oool300000=@0oooo002=0?ooo`800000BP3oool00`000000
oooo0?ooo`3D0?ooo`00S`3oool3000004L0oooo00<000000?ooo`3oool0e03oool00980oooo0P00
00150?ooo`030000003oool0oooo0=@0oooo002D0?ooo`<00000@P3oool00`000000oooo0?ooo`3D
0?ooo`00U`3oool200000400oooo00<000000?ooo`3oool0e03oool009T0oooo0`00000m0?ooo`03
0000003oool0oooo0=@0oooo002L0?ooo`800000>`3oool00`000000oooo0?ooo`3D0?ooo`00WP3o
ool2000003T0oooo0`00003D0?ooo`00X03oool3000003H0oooo00<000000?ooo`3oool0e03oool0
0100oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`1H0?ooo`040;md000000000000
07Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=ooob40oooo
0P00000d0?ooo`030000003oool0oooo05T0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0
B00007Boo`020?ooo`060?oOW0180000000000000000B9`0goooF@3oool00`3ookl0M00000000002
0000000304RLg`3oool0oooo00P0oooo000B0?ooo`030?noM0000000M;oo05X0oooo00H0gia8001d
_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080oooo00H0okmd0000001d_ol0oooo
0?oOW018M;lS0?ooo`<00000<@3oool00`000000oooo0?ooo`1H0?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
F`3oool00`3o_g@0000007Boo`0;0?ooo`004P3oool00`3ooml0W4P0001d_`1I0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0T0?ooo`800000;`3o
ool00`000000oooo0?ooo`1G0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02`3o
ool00`3o_g@0B9cO0?ooo`1I0?ooo`030?oog`2LB00007Bo00/0oooo000C0?ooo`030=nLB000B9`0
gooo05P0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d
07@0_ooo02H0oooo0`00000/0?ooo`030000003oool0oooo05L0oooo00<0oooO09a80018W=l00P3o
ool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d07@0_ooo05X0oooo00<0gia80018W03Oool0
2P3oool000<0oooo00<0oooO09a8000000001`0000000`18W=l0oooo0?ooo`030?ooo`030?oOW018
04P0W=oo04/0oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ooml0W4P0001d_`<0
oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`0Y0?ooo`800000:P3oool00`000000
oooo0?ooo`1G0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0
W4P007Boo`1J0?ooo`030?oOW01804P0W=oo00X0oooo000C0?ooo`030?oo_`1d0000M;oo05T0oooo
00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030?oo_`1d0000B9cO02/0oooo
0P00000X0?ooo`030000003oool0oooo05P0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B000
04RLg`0:0?ooo`030?oo_`1d0000B9cO05X0oooo00<0oono07@0001d_ol02P3oool001@0oooo00<0
_g@0001d_`3oool0F03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00X0oooo00<0
okmd0000001d_ol0;@3oool3000002D0oooo0`00001H0?ooo`030=nLB00004P0W=oo0080oooo00<0
oooO09a80000M;l02P3oool00`3o_g@0000007Boo`1K0?ooo`030;md0000M;l0oooo00T0oooo000D
0?ooo`030=nLB000B9`0gooo05P0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`06
0?ooo`060?oOg`3Oool0okmd0000000007@0_ooo<03oool2000002<0oooo00<000000?ooo`3oool0
F03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@0
00000000M02ooomK0?ooo`030=nLB000B9`0gooo00T0oooo000D0?ooo`030?oOW01804P0W=oo05P0
oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`050?oo_`1d00000000
0000B02Lgol0<`3oool300000200oooo00<000000?ooo`3oool0F03oool0103ooml0W4P00018W03O
ool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000001809cOo`1L0?ooo`030?oO
W01804P0W=oo00T0oooo000D0?ooo`030?oo_`1d0000M;oo05T0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo03L0oooo0P00000N0?ooo`030000003o
ool0oooo05T0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02o
ool0oooo05d0oooo00<0oono07@0001d_ol02@3oool001D0oooo00<0_g@0001d_`3oool0F@3oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool0>@3oool3000001/0oooo
00<000000?ooo`3oool0FP3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BL
g`3oool0GP3oool00`2oM00007Bo0?ooo`080?ooo`004`3oool01@3ookl0M0000000000007@0_ooo
05X0oooo00@0gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO03T0oooo0P00
000I0?ooo`030000003oool0oooo05/0oooo00@0gia800000000000007Bo2P3oool01@2oM0000000
000000000000B9cO05T0oooo00D0oono07@000000000001d0;ooo`080?ooo`00`03oool2000001L0
oooo00<000000?ooo`3oool0e03oool00<80oooo0`00000D0?ooo`030000003oool0oooo0=@0oooo
00350?ooo`8000004P3oool300000=@0oooo00370?ooo`<000003`3oool00`000000oooo0?ooo`3D
0?ooo`00bP3oool2000000d0oooo00<000000?ooo`3oool0e03oool00<`0oooo0`00000:0?ooo`03
0000003oool0oooo0=@0oooo003?0?ooo`800000203oool00`000000oooo0?ooo`3D0?ooo`00d@3o
ool2000000H0oooo00<000000?ooo`3oool0e03oool00=<0oooo0`0000030?ooo`030000003oool0
oooo0=@0oooo003F0?ooo`80000000<0oooo0000003oool0e@3oool000<0ooooo`00002]00000040
oooo000>0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`0000000P00000?0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo00T0oooo000>0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0H`3o
ool00`000000oooo0?ooo`020?ooo`<00000GP3oool00`000000oooo0?ooo`1R0?ooo`030000003o
ool0oooo00T0oooo003I0?ooo`030000003oool0oooo00D0oooo0P00003=0?ooo`00f@3oool00`00
0000oooo0?ooo`070?ooo`<00000bP3oool00=T0oooo00<000000?ooo`3oool02P3oool300000<L0
oooo003I0?ooo`030000003oool0oooo00d0oooo0`0000340?ooo`00f@3oool00`000000oooo0?oo
o`0@0?ooo`800000`P3oool00=T0oooo0`00000B0?ooo`<00000_`3oool00=T0oooo00<000000?oo
o`3oool05@3oool200000;d0oooo003I0?ooo`030000003oool0oooo01L0oooo0P00002k0?ooo`00
f@3oool00`000000oooo0?ooo`0I0?ooo`<00000^03oool00=T0oooo00<000000?ooo`3oool0703o
ool200000;H0oooo003I0?ooo`030000003oool0oooo01h0oooo0`00002c0?ooo`00f@3oool00`00
0000oooo0?ooo`0Q0?ooo`800000/@3oool00=T0oooo00<000000?ooo`3oool08`3oool300000:h0
oooo003I0?ooo`<000009P3oool200000:`0oooo003I0?ooo`030000003oool0oooo02P0oooo0P00
002Z0?ooo`00f@3oool00`000000oooo0?ooo`0Z0?ooo`<00000Y`3oool00=T0oooo00<000000?oo
o`3oool0;@3oool200000:D0oooo003I0?ooo`030000003oool0oooo02l0oooo0`00002R0?ooo`00
f@3oool00`000000oooo0?ooo`0b0?ooo`800000X03oool00=T0oooo00<000000?ooo`3oool0=03o
ool3000009d0oooo003I0?ooo`030000003oool0oooo03L0oooo0P00002K0?ooo`00f@3oool30000
03T0oooo0P00002I0?ooo`00f@3oool00`000000oooo0?ooo`0k0?ooo`<00000UP3oool00=T0oooo
00<000000?ooo`3oool0?P3oool2000009@0oooo003I0?ooo`030000003oool0oooo0400oooo0`00
002A0?ooo`00f@3oool00`000000oooo0?ooo`130?ooo`800000S`3oool00=T0oooo00<000000?oo
o`3oool0A@3oool3000008`0oooo003I0?ooo`030000003oool0oooo04P0oooo0P00002:0?ooo`00
f@3oool00`000000oooo0?ooo`1:0?ooo`800000R03oool00=T0oooo0`00001<0?ooo`<00000Q@3o
ool00=T0oooo00<000000?ooo`3oool0C`3oool2000008<0oooo003I0?ooo`030000003oool0oooo
0540oooo0`0000200?ooo`00_03oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo
0080oooo00H0omnL04P00000000000000018W03Oool;0?ooo`030000003oool0oooo05@0oooo0P00
001n0?ooo`00^`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0
0P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_`/0oooo00<000000?ooo`3oool0EP3oool20000
07`0oooo002j0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0
B9cO0?ooo`090?ooo`030000003oool0oooo05P0oooo0P00001j0?ooo`00^P3oool00`3ooml0W4P0
04RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02oool02@3oool00`000000
oooo0?ooo`1J0?ooo`800000N03oool00;X0oooo00<0oooO09a80000M;l00`3oool00`3OW4P00018
09cOo`0:0?ooo`030?oog`2LB000M;oo00T0oooo00<000000?ooo`3oool0G03oool2000007H0oooo
002k0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0M00004RL
g`090?ooo`@00000G@3oool2000007@0oooo002k0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO
09a80000M;l02P3oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo0600oooo0P00001b
0?ooo`00^`3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00H0omoO0=oo
o`3o_g@000000000M02oool90?ooo`030000003oool0oooo0680oooo0`00001_0?ooo`00^`3oool0
103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@0000000000018
09cOo`0:0?ooo`030000003oool0oooo06D0oooo0P00001]0?ooo`00_03oool00`3ogi`0B00007Bo
o`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3oool02`3oool00`000000oooo0?oo
o`1W0?ooo`<00000JP3oool00;d0oooo00H0gia80000M02oool0oooO09a80018W=l80?ooo`030?oO
W01dW=l0oooo00/0oooo00<000000?ooo`3oool0JP3oool2000006P0oooo002n0?ooo`040=nLB000
00000000001d_`X0oooo00D0_g@0000000000000000004RLg`080?ooo`030000003oool0oooo06`0
oooo0P00001V0?ooo`00f@3oool00`000000oooo0?ooo`1^0?ooo`<00000H`3oool00=T0oooo0`00
001a0?ooo`800000H@3oool00=T0oooo00<000000?ooo`3oool0L`3oool2000005l0oooo003I0?oo
o`030000003oool0oooo07D0oooo0P00001M0?ooo`00f@3oool00`000000oooo0?ooo`1g0?ooo`80
0000F`3oool00=T0oooo00<000000?ooo`3oool0N@3oool2000005T0oooo003I0?ooo`030000003o
ool0oooo07/0oooo0P00001G0?ooo`00f@3oool00`000000oooo0?ooo`1m0?ooo`800000E@3oool0
0=T0oooo00<000000?ooo`3oool0O`3oool2000005<0oooo003I0?ooo`<00000P@3oool200000540
oooo003I0?ooo`030000003oool0oooo08<0oooo0P00001?0?ooo`00f@3oool00`000000oooo0?oo
o`250?ooo`030000003oool0oooo04`0oooo003I0?ooo`030000003oool0oooo08H0oooo0P00001<
0?ooo`00f@3oool00`000000oooo0?ooo`280?ooo`800000BP3oool00=T0oooo00<000000?ooo`3o
ool0RP3oool2000004P0oooo003I0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool0
A@3oool00=T0oooo00<000000?ooo`3oool0S@3oool2000004D0oooo003I0?ooo`<00000S`3oool2
000004<0oooo003I0?ooo`030000003oool0oooo0940oooo0P0000110?ooo`00f@3oool00`000000
oooo0?ooo`2C0?ooo`800000?`3oool00=T0oooo00<000000?ooo`3oool0U@3oool00`000000oooo
0?ooo`0l0?ooo`00f@3oool00`000000oooo0?ooo`2F0?ooo`800000?03oool00=T0oooo00<00000
0?ooo`3oool0V03oool00`000000oooo0?ooo`0i0?ooo`00f@3oool00`000000oooo0?ooo`2I0?oo
o`800000>@3oool00=T0oooo00<000000?ooo`3oool0V`3oool00`000000oooo0?ooo`0f0?ooo`00
f@3oool3000009`0oooo0P00000f0?ooo`00f@3oool00`000000oooo0?ooo`2N0?ooo`030000003o
ool0oooo03<0oooo003I0?ooo`030000003oool0oooo09l0oooo0P00000c0?ooo`00b03oool00`3o
okl0M000000000020000000304RLg`3oool0oooo00T0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0`0?ooo`00bP3oool00`3o_g@0000007Boo`0<0?ooo`030000003oool0oooo0:80
oooo0P00000`0?ooo`00bP3oool00`3ooml0W4P0001d_`0<0?ooo`030000003oool0oooo0:@0oooo
00<000000?ooo`3oool0;@3oool00</0oooo00<0gia80018W03Oool02`3oool00`000000oooo0?oo
o`2U0?ooo`800000;@3oool00</0oooo00<0omnL04P0B02Lgol02`3oool00`000000oooo0?ooo`2W
0?ooo`030000003oool0oooo02X0oooo003;0?ooo`030?oo_`1d0000M;oo00/0oooo1000002W0?oo
o`030000003oool0oooo02T0oooo003<0?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3o
ool0Z@3oool2000002T0oooo003<0?ooo`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool0
Z`3oool00`000000oooo0?ooo`0V0?ooo`00c03oool00`3ogi`0B01809cOo`0:0?ooo`030000003o
ool0oooo0:`0oooo00<000000?ooo`3oool09@3oool00<`0oooo00<0oono07@0001d_ol02P3oool0
0`000000oooo0?ooo`2]0?ooo`8000009@3oool00<d0oooo00<0_g@0001d_`3oool02@3oool00`00
0000oooo0?ooo`2_0?ooo`030000003oool0oooo0280oooo003;0?ooo`050?oo_`1d000000000000
M02oool02@3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo0240oooo003I0?ooo`03
0000003oool0oooo0;40oooo00<000000?ooo`3oool0803oool00=T0oooo0`00002b0?ooo`030000
003oool0oooo01l0oooo003I0?ooo`030000003oool0oooo0;<0oooo00<000000?ooo`3oool07P3o
ool00=T0oooo00<000000?ooo`3oool0]03oool2000001h0oooo003I0?ooo`030000003oool0oooo
0;H0oooo00<000000?ooo`3oool06`3oool00=T0oooo00<000000?ooo`3oool0]`3oool00`000000
oooo0?ooo`0J0?ooo`00f@3oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01T0oooo
003I0?ooo`030000003oool0oooo0;T0oooo00<000000?ooo`3oool0603oool00=T0oooo00<00000
0?ooo`3oool0^P3oool00`000000oooo0?ooo`0G0?ooo`00f@3oool300000;/0oooo00<000000?oo
o`3oool05P3oool00=T0oooo00<000000?ooo`3oool0_03oool00`000000oooo0?ooo`0E0?ooo`00
f@3oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo01D0oooo003I0?ooo`030000003o
ool0oooo0;d0oooo00<000000?ooo`3oool0503oool00=T0oooo00<000000?ooo`3oool0_P3oool0
0`000000oooo0?ooo`0C0?ooo`00f@3oool00`000000oooo0?ooo`2o0?ooo`030000003oool0oooo
0180oooo003I0?ooo`030000003oool0oooo0<00oooo00<000000?ooo`3oool04@3oool00=T0oooo
00<000000?ooo`3oool0`@3oool00`000000oooo0?ooo`0@0?ooo`00f@3oool300000<40oooo00<0
00000?ooo`3oool0403oool00=T0oooo00<000000?ooo`3oool0`P3oool00`000000oooo0?ooo`0?
0?ooo`00f@3oool00`000000oooo0?ooo`320?ooo`030000003oool0oooo00l0oooo003I0?ooo`03
0000003oool0oooo0<<0oooo00<000000?ooo`3oool03P3oool00=T0oooo00<000000?ooo`3oool0
``3oool00`000000oooo0?ooo`0>0?ooo`00f@3oool00`000000oooo0?ooo`340?ooo`030000003o
ool0oooo00d0oooo003I0?ooo`030000003oool0oooo0<@0oooo00<000000?ooo`3oool03@3oool0
0=T0oooo00<000000?ooo`3oool0a@3oool00`000000oooo0?ooo`0<0?ooo`00f@3oool300000<D0
oooo00<000000?ooo`3oool0303oool00=T0oooo00<000000?ooo`3oool0aP3oool00`000000oooo
0?ooo`0;0?ooo`00f@3oool00`000000oooo0?ooo`360?ooo`030000003oool0oooo00/0oooo002j
0?ooo`030?oo_`1d000000000080000000L0B9cO0?ooo`3oool0oooo0?oOW0180000M;oo0080oooo
00H0omnL04P00000000000000018W03Oool;0?ooo`030000003oool0oooo0<L0oooo00<000000?oo
o`3oool02P3oool00;`0oooo00<0okmd0000001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`06
0?noM0000000M;oo0?ooo`3ogi`0B7Bo2`3oool00`000000oooo0?ooo`370?ooo`030000003oool0
oooo00X0oooo002l0?ooo`030?oog`2LB00007Bo00h0oooo00<0okmd04RLg`3oool02@3oool00`00
0000oooo0?ooo`370?ooo`030000003oool0oooo00X0oooo002m0?ooo`030=nLB000B9`0gooo00d0
oooo00<0oono07@0M02oool02@3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo00X0
oooo002m0?ooo`030?oOW01804P0W=oo00d0oooo00<0oooO09a8001d_ol02@3oool00`000000oooo
0?ooo`380?ooo`030000003oool0oooo00T0oooo002m0?ooo`030?oo_`1d0000M;oo00d0oooo00<0
oono07@00018W=l02@3oool400000<L0oooo00<000000?ooo`3oool02@3oool00;h0oooo00<0_g@0
001d_`3oool0303oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo0<P0oooo00<00000
0?ooo`3oool02@3oool00;h0oooo00<0gia80018W03Oool02@3oool01P3ogml0gooo0?noM0000000
001d0;ooo`T0oooo00<000000?ooo`3oool0b03oool00`000000oooo0?ooo`090?ooo`00_P3oool0
0`3ogi`0B01809cOo`090?ooo`050?oo_`1d000000000000B02Lgol02P3oool00`000000oooo0?oo
o`380?ooo`030000003oool0oooo00T0oooo002n0?ooo`030?oo_`1d0000M;oo00X0oooo00<0gibL
0;ooo`3oool02`3oool00`000000oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo002o0?oo
o`030;md0000M;l0oooo00T0oooo00<0omnL07BLg`3oool02`3oool00`000000oooo0?ooo`380?oo
o`030000003oool0oooo00T0oooo002m0?ooo`050?oo_`1d000000000000M02oool02P3oool01@2o
M0000000000000000000B9cO00P0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?oo
o`3oool0e03oool00=T0oooo0`00003D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3o
ool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`00
0000oooo0?ooo`3D0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.06891, -1.66067, \
0.00737444, 0.018738}}]
}, Closed]],

Cell["produces a basic plot while", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[ArcSin[x], {x, \(-1\), 1}, Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<x\>", \*"\"\<\!\(sin\^\(-1\)\)x\>\""}, 
        GridLines \[Rule] Automatic, 
        PlotLabel \[Rule] "\<My First Plot\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.187407 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -4.78125 -31.375 ]
[.5 -0.0125 4.78125 -15.6875 ]
[-0.0125 .02791 -27.6875 -7.84375 ]
[-0.0125 .02791 0 7.84375 ]
[-0.0125 .12161 -18.6875 -7.84375 ]
[-0.0125 .12161 0 7.84375 ]
[-0.0125 .21531 -27.6875 -7.84375 ]
[-0.0125 .21531 0 7.84375 ]
[-0.0125 .30902 -10 -7.84375 ]
[-0.0125 .30902 0 7.84375 ]
[-0.0125 .40272 -19 -7.84375 ]
[-0.0125 .40272 0 7.84375 ]
[-0.0125 .49642 -10 -7.84375 ]
[-0.0125 .49642 0 7.84375 ]
[-0.0125 .59013 -19 -7.84375 ]
[-0.0125 .59013 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -43.375 -17.0625 ]
[-0.01875 .30902 -27.6875 17.0625 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -34.5625 0 ]
[.5 .63053 34.5625 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.2619 0 m
.2619 .61803 L
s
.5 0 m
.5 .61803 L
s
.7381 0 m
.7381 .61803 L
s
0 .02791 m
1 .02791 L
s
0 .12161 m
1 .12161 L
s
0 .21531 m
1 .21531 L
s
0 .30902 m
1 .30902 L
s
0 .40272 m
1 .40272 L
s
0 .49642 m
1 .49642 L
s
0 .59013 m
1 .59013 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -65.7813 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
68.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .02791 m
.00625 .02791 L
s
gsave
-0.0125 .02791 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
0 .12161 m
.00625 .12161 L
s
gsave
-0.0125 .12161 -79.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .21531 m
.00625 .21531 L
s
gsave
-0.0125 .21531 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .30902 m
.00625 .30902 L
s
gsave
-0.0125 .30902 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .40272 m
.00625 .40272 L
s
gsave
-0.0125 .40272 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .49642 m
.00625 .49642 L
s
gsave
-0.0125 .49642 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
0 .59013 m
.00625 .59013 L
s
gsave
-0.0125 .59013 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1.5) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04665 m
.00375 .04665 L
s
0 .06539 m
.00375 .06539 L
s
0 .08413 m
.00375 .08413 L
s
0 .10287 m
.00375 .10287 L
s
0 .14035 m
.00375 .14035 L
s
0 .15909 m
.00375 .15909 L
s
0 .17783 m
.00375 .17783 L
s
0 .19657 m
.00375 .19657 L
s
0 .23405 m
.00375 .23405 L
s
0 .25279 m
.00375 .25279 L
s
0 .27154 m
.00375 .27154 L
s
0 .29028 m
.00375 .29028 L
s
0 .32776 m
.00375 .32776 L
s
0 .3465 m
.00375 .3465 L
s
0 .36524 m
.00375 .36524 L
s
0 .38398 m
.00375 .38398 L
s
0 .42146 m
.00375 .42146 L
s
0 .4402 m
.00375 .4402 L
s
0 .45894 m
.00375 .45894 L
s
0 .47768 m
.00375 .47768 L
s
0 .51516 m
.00375 .51516 L
s
0 .53391 m
.00375 .53391 L
s
0 .55265 m
.00375 .55265 L
s
0 .57139 m
.00375 .57139 L
s
0 .00917 m
.00375 .00917 L
s
0 .60887 m
.00375 .60887 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -113.594 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
78.0625 11.8438 translate
90 rotate
-78.0625 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(sin) show
77.000 11.000 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
82.750 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
87.875 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(x) show
93.125 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .02791 m
1 .02791 L
s
.99375 .12161 m
1 .12161 L
s
.99375 .21531 m
1 .21531 L
s
.99375 .30902 m
1 .30902 L
s
.99375 .40272 m
1 .40272 L
s
.99375 .49642 m
1 .49642 L
s
.99375 .59013 m
1 .59013 L
s
.125 Mabswid
.99625 .04665 m
1 .04665 L
s
.99625 .06539 m
1 .06539 L
s
.99625 .08413 m
1 .08413 L
s
.99625 .10287 m
1 .10287 L
s
.99625 .14035 m
1 .14035 L
s
.99625 .15909 m
1 .15909 L
s
.99625 .17783 m
1 .17783 L
s
.99625 .19657 m
1 .19657 L
s
.99625 .23405 m
1 .23405 L
s
.99625 .25279 m
1 .25279 L
s
.99625 .27154 m
1 .27154 L
s
.99625 .29028 m
1 .29028 L
s
.99625 .32776 m
1 .32776 L
s
.99625 .3465 m
1 .3465 L
s
.99625 .36524 m
1 .36524 L
s
.99625 .38398 m
1 .38398 L
s
.99625 .42146 m
1 .42146 L
s
.99625 .4402 m
1 .4402 L
s
.99625 .45894 m
1 .45894 L
s
.99625 .47768 m
1 .47768 L
s
.99625 .51516 m
1 .51516 L
s
.99625 .53391 m
1 .53391 L
s
.99625 .55265 m
1 .55265 L
s
.99625 .57139 m
1 .57139 L
s
.99625 .00917 m
1 .00917 L
s
.99625 .60887 m
1 .60887 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .30902 m
1 .30902 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -95.5625 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(My) show
81.313 15.813 moveto
(First) show
107.938 15.813 moveto
(Plot) show
128.125 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .02781 L
.02605 .03284 L
.02846 .04084 L
.03279 .05108 L
.04262 .06749 L
.05281 .08038 L
.06244 .09065 L
.10458 .1254 L
.14299 .15017 L
.18386 .17296 L
.22321 .19277 L
.26105 .2105 L
.30133 .22836 L
.34011 .24484 L
.38133 .26182 L
.42104 .2778 L
.45923 .29295 L
.49988 .30897 L
.53901 .32439 L
.58059 .34089 L
.62066 .35702 L
.65921 .3729 L
.70021 .39034 L
.7397 .40787 L
.77767 .42569 L
.81809 .44611 L
.857 .46787 L
.89836 .49473 L
.91906 .51065 L
.93821 .52804 L
.95662 .54949 L
.96592 .56441 L
.96847 .56961 L
.97122 .57629 L
.97248 .57999 L
.97381 .58466 L
.97495 .58986 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00mP3oool0203OW4P00000001d_`3oool0oooO09a800000000M;oo/P3oool00?P0oooo00L0
okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0;40oooo003i0?ooo`040;mdM02LW4P004RL0=oook<0
oooo003i0?ooo`030?oo_`1d000007Bo0;@0oooo003i0?ooo`030?oo_`1d0000M;oo0;@0oooo003i
0?ooo`040?noM000000004RL0=oook<0oooo003i0?ooo`050;md0000B9`0gmnL04Q8W03Oool0/P3o
ool00?L0oooo00T0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0/03oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool006`0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`0k0?ooo`040;md0000
0000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=oo
ocL0oooo00@0_g@0000000000000M;oo?@3oool0102oM00000000000001d_ol30?ooo`030?oOW018
0000M;oo0080oooo00H0omnL04P00000000000000018W03Ooolk0?ooo`030?oo_`1d000000000080
000000<0B9cO0?ooo`3oool08P3oool006h0oooo00<0okmd0000001d_ol0?@3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_cH0oooo00H0gia8001d_`3oool0omnL04P00018W=lk0?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
?@3oool00`3o_g@0000007Boo`0U0?ooo`00KP3oool00`3ooml0W4P0001d_`0l0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0c0?ooo`030?oog`2L
B000B9cO0080oooo00<0okmd0000B02Lgol0>@3oool00`3ooml0W4P004RLg`020?ooo`030?noM000
04P0W=oo00/0oooo00<0okmd04RLg`3oool0>`3oool00`3ooml0W4P0001d_`0U0?ooo`00K`3oool0
0`3OW4P004RL0=ooo`0k0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3o
ool00`3ookl0M01d0;ooo`0c0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
>03oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02o
ool0?03oool00`3OW4P004RL0=ooo`0T0?ooo`00G`3oool00`3ooml0W4P0000000070000000304RL
g`3oool0oooo00<0oooo00<0omnL04P0B02Lgol0;P3oool00`3ooml0W4P0000000070000000604RL
g`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo
03<0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0h0?ooo`030?oog`2LB00007Bo
00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`0l0?ooo`030?oOW01804P0W=oo
02@0oooo001_0?ooo`030?oo_`1d0000M;oo03`0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0:0?ooo`030?oo_`1d0000B9cO03@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0
B00004RLg`0i0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0
M00004RLg`0l0?ooo`030?oo_`1d0000M;oo02@0oooo001`0?ooo`030;md0000M;l0oooo03/0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo03@0oooo
00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0i0?ooo`030=nLB00004P0W=oo0080oooo
00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`0m0?ooo`030;md0000M;l0oooo02<0oooo
001`0?ooo`030=nLB000B9`0gooo03/0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo=03oool00`3ogi`0B00004RLg`030?ooo`03
0;md0000B9`0gooo03P0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo?@3oool00`3OW4P004RL0=ooo`0S0?ooo`00L03oool00`3o
gi`0B01809cOo`0k0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo03D0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL
0=ooo`0h0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0
M0000000000004P0W=oo03h0oooo00<0omnL04P0B02Lgol08`3oool00700oooo00<0oono07@0001d
_ol0?03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3o
ool0=`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo03T0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo03l0oooo00<0oono07@0001d
_ol08`3oool00740oooo00<0_g@0001d_`3oool0?03oool01P3OW4P0001d0;ooo`3ooml0W4P004RL
g`P0oooo00<0omnL07BLg`3oool0>03oool01P3OW4P0001d0;ooo`3ooml0W4P004RLgc/0oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030?oOW01dW=l0oooo0400oooo00<0_g@0001d_`3o
ool08P3oool006l0oooo00D0oono07@000000000001d0;ooo`0m0?ooo`040=nLB00000000000001d
_`X0oooo00D0_g@0000000000000000004RLg`0f0?ooo`040=nLB00000000000001d_cd0oooo00@0
gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO03/0oooo00D0oono07@00000
0000001d0;ooo`0R0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00HP3ooooo
0000038000007@3oool00680oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0K0?oo
o`00HP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3o
ool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0oooo001R0?ooo`<00000
C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0
B`3oool2000001d0oooo00150?ooo`030?oo_`1d000000000080000000L0B9cO0?ooo`3oool0oooo
0?oOW0180000M;oo0080oooo00H0omnL04P00000000000000018W03Oool90?ooo`030000003oool0
oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0
oooo04`0oooo00<000000?ooo`3oool06`3oool004L0oooo00<0okmd0000001d_ol0103oool00`3o
gi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo2@3oool00`000000oooo0?oo
o`050?ooo`030000003oool0oooo04D0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?oo
o`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool004L0oooo00<0oooO
09a80000M;l03P3oool00`3o_g@0B9cO0?ooo`070?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool0A@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o
0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00B03oool00`3OW4P004RL0=ooo`0=0?oo
o`030?oo_`1d07@0_ooo00L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`150?oo
o`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?oo
o`030000003oool0oooo01/0oooo000h0?ooo`030?oog`2LB000000000L0000000<0B9cO0?ooo`3o
ool00`3oool00`3ogi`0B01809cOo`0=0?ooo`030?oog`2LB000M;oo00L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`150?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3o
ool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo00180?ooo`030?oo
_`1d0000M;oo00d0oooo00<0oono07@00018W=l01`3oool3000000D0001o00<000000000O`0007l0
S@0007l00`000000001o0000O`2C0000O`<000007@3oool004T0oooo00<0_g@0001d_`3oool0303o
ool00`3o_g@0000007Boo`070?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0A@3o
ool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03o
ool00`000000oooo0?ooo`0K0?ooo`00B@3oool00`3OW4P004RL0=ooo`090?ooo`060?oOg`3Oool0
okmd0000000007@0_ooo1`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo04D0oooo
00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo
00<000000?ooo`3oool06`3oool004T0oooo00<0omnL04P0B02Lgol02@3oool01@3ookl0M0000000
000004P0W=oo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`140?ooo`030000
O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000
003oool0oooo01/0oooo00190?ooo`030?oo_`1d0000M;oo00X0oooo00<0gibL0;ooo`3oool02@3o
ool00`000000oooo0?ooo`060?ooo`030000003oool0oooo04@0oooo00<0001o0?ooo`3oool0A@3o
ool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3o
ool004X0oooo00<0_g@0001d_`3oool02@3oool00`3ogi`0M9cO0?ooo`090?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool0A03oool00`0007l0oooo0?ooo`150?ooo`030000003oool0
oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00B03oool01@3o
okl0M0000000000007@0_ooo00X0oooo00D0_g@0000000000000000004RLg`060?ooo`<000001`3o
ool00`000000oooo0?ooo`130?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3o
ool00`0007l0oooo0?ooo`1;0?ooo`8000007@3oool00680oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`130?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool0
0`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0@P3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo
04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo0480oooo00<0001o0?ooo`3oool0A@3oool00`000000
oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool00680oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`110?ooo`030000O`3oool0oooo04D0oooo
00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo
001R0?ooo`<000002@3oool00`000000oooo0?ooo`110?ooo`030000O`3oool0oooo04D0oooo00<0
00000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1;0?ooo`8000007@3oool00680oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`100?ooo`030000O`3oool0oooo04D0oooo00<00000
0?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0?`3oool00`0007l0oooo0?ooo`150?oo
o`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?oo
o`00HP3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo03l0oooo00<0001o0?ooo`3o
ool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3o
ool06`3oool00680oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0n0?ooo`030000
O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000
003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0?@3o
ool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03o
ool00`000000oooo0?ooo`0K0?ooo`00HP3oool3000000h0oooo00<000000?ooo`3oool0?03oool0
0`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0B`3oool2
000001d0oooo001R0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0>`3oool00`00
07l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`00
0000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo03X0
oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0
oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0i0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo
0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo0180oooo00<0
00000?ooo`3oool0>03oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0
001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0C
0?ooo`800000>03oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o
0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00D`3oool00`3ookl0M000000000020000
000304RLg`3oool0oooo00L0oooo0`00000E0?ooo`030000003oool0oooo03D0oooo00<0001o0?oo
o`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04/0oooo0P00000M0?oo
o`00E@3oool00`3o_g@0000007Boo`0:0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3o
ool0=03oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3o
ool0C03oool00`000000oooo0?ooo`0K0?ooo`00E@3oool00`3ooml0W4P0001d_`0:0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool0<`3oool00`0007l0oooo0?ooo`150?ooo`030000
003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00EP3o
ool00`3OW4P004RL0=ooo`090?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<P3o
ool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03o
ool00`000000oooo0?ooo`0K0?ooo`00AP3oool00`3ooml0W4P0000000070000000304RLg`3oool0
oooo00<0oooo00<0omnL04P0B02Lgol02@3oool00`000000oooo0?ooo`0I0?ooo`800000<P3oool0
0`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool0
0`000000oooo0?ooo`0K0?ooo`00EP3oool00`3ookl0M00007Boo`090?ooo`<000006`0007l00`00
0000001o0000O`1g0000O`030000000007l0001o09<0001o0`00000M0?ooo`00E`3oool00`2oM000
07Bo0?ooo`080?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0;P3oool00`0007l0
oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000
oooo0?ooo`0K0?ooo`00E`3oool00`3OW4P004RL0=ooo`080?ooo`030000003oool0oooo01d0oooo
00<000000?ooo`3oool0;@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo
00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00E`3oool00`3ogi`0B01809cO
o`080?ooo`030000003oool0oooo01h0oooo0P00000]0?ooo`030000O`3oool0oooo04D0oooo00<0
00000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001G
0?ooo`030?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool0803oool2000002/0oooo00<0001o
0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<00000
0?ooo`3oool06`3oool005P0oooo00<0_g@0001d_`3oool01`3oool00`000000oooo0?ooo`0R0?oo
o`030000003oool0oooo02P0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?oo
o`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool005H0oooo00D0oono07@00000
0000001d0;ooo`070?ooo`<000008`3oool2000002P0oooo00<0001o0?ooo`3oool0A@3oool00`00
0000oooo0?ooo`160?ooo`030000O`3oool0oooo04/0oooo0P00000M0?ooo`00HP3oool00`000000
oooo0?ooo`0U0?ooo`8000009P3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0
oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo
0?ooo`0W0?ooo`800000903oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo
00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?oo
o`0Y0?ooo`030000003oool0oooo0240oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?oo
o`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool00680oooo00<00000
0?ooo`3oool0:P3oool200000240oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`16
0?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?oo
o`3oool0;03oool2000001l0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?oo
o`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool00680oooo0`00000^0?ooo`03
0000003oool0oooo01`0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`03
0000O`3oool0oooo04/0oooo0P00000M0?ooo`00HP3oool00`000000oooo0?ooo`0_0?ooo`800000
703oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0
C03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0a0?ooo`8000006P3o
ool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03o
ool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0c0?ooo`800000603oool0
0`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool0
0`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo
01D0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo
04`0oooo00<000000?ooo`3oool06`3oool00680oooo0`00000f0?ooo`8000005@3oool00`0007l0
oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0B`3oool2000001d0
oooo001R0?ooo`030000003oool0oooo03P0oooo0P00000C0?ooo`030000O`3oool0oooo04D0oooo
00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo
001R0?ooo`030000003oool0oooo03X0oooo0P00000A0?ooo`030000O`3oool0oooo04D0oooo00<0
00000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R
0?ooo`030000003oool0oooo03`0oooo0P00000?0?ooo`030000O`3oool0oooo04D0oooo00<00000
0?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?oo
o`030000003oool0oooo03h0oooo0P00000=0?ooo`030000O`3oool0oooo04D0oooo00<000000?oo
o`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?ooo`03
0000003oool0oooo0400oooo0`00000:0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3o
ool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?ooo`<00000
@`3oool2000000P0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000
O`3oool0oooo04/0oooo0P00000M0?ooo`00A`3oool0102oM00000000000001d_ol30?ooo`030?oO
W0180000M;oo0080oooo00H0omnL04P00000000000000018W03Oool90?ooo`030000003oool0oooo
04D0oooo0P0000060?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`00
07l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo00160?ooo`060=nLB000M;l0oooo0?oO
W0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
2@3oool00`000000oooo0?ooo`170?ooo`<000000`3oool00`0007l0oooo0?ooo`150?ooo`030000
003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00A@3o
ool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool01`3o
ool00`000000oooo0?ooo`1:0?ooo`80000000<0oooo0000O`3oool0AP3oool00`000000oooo0?oo
o`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool004D0oooo00<0oooO
09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d07@0_ooo00L0oooo00<0
00000?ooo`3oool0C03oool2000004L0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?oo
o`1<0?ooo`030000003oool0oooo01/0oooo000h0?ooo`030?oog`2LB000000000L0000000H0B9cO
0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8001d_ol0
1`3oool00`000000oooo0?ooo`1=0?ooo`030000O`000000000004D0oooo00<000000?ooo`3oool0
AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo00160?ooo`030;md0000
B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0M00004RLg`070?ooo`<00000D000
07l2000004<0001o00<000000000O`0007l0T`0007l3000001d0oooo00160?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`070?ooo`030000003oool0
oooo04d0oooo00<0001o0?ooo`3oool00P3oool200000440oooo00<000000?ooo`3oool0AP3oool0
0`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo00160?ooo`030?oOW0180000B9cO
00<0oooo00<0_g@00018W03Oool01P3oool01P3ogml0gooo0?noM0000000001d0;ooo`L0oooo00<0
00000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`040?ooo`800000?`3oool00`000000oooo0?oo
o`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool004H0oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`050?oo_`1d000000000000B02Lgol0
203oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo00H0oooo0P00000m0?ooo`030000
003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00A`3o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3oool02@3o
ool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo00P0oooo0`00000j0?ooo`030000003o
ool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00B03oool0
1P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool02@3oool00`000000oooo
0?ooo`1=0?ooo`030000O`3oool0oooo00/0oooo0P00000h0?ooo`030000003oool0oooo04H0oooo
00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00B@3oool0103OW4P000000000
0000M;l:0?ooo`050;md00000000000000000018W=l01P3oool3000004d0oooo00<0001o0?ooo`3o
ool03@3oool2000003H0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1;0?ooo`80
00007@3oool002T0oooo00D0oooO09a80000000004RL0=ooo`0d0?ooo`030000003oool0oooo04d0
oooo00<0001o0?ooo`3oool03`3oool3000003<0oooo00<000000?ooo`3oool0AP3oool00`0007l0
oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000U0?ooo`040;md00000000001d0;ooo`80
oooo00<0oono07Ad_`3oool0=03oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo0180
oooo0P00000a0?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000
oooo0?ooo`0K0?ooo`00903oool01P3ogi`0W;oo0?oog`2LB000001d0;ooo`80oooo00<0gibL0;oo
o`3oool0<`3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo01@0oooo0`00000^0?oo
o`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?oo
o`00903oool02P3o_i`0W=oo0?ooo`3ooml0W4P00000B02Lgol0oono07A8W03Ooold0?ooo`030000
003oool0oooo04d0oooo00<0001o0?ooo`3oool05`3oool2000002`0oooo00<000000?ooo`3oool0
AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000T0?ooo`030?oo_`1d
M;l0oooo0080oooo00@0_g@00000000007@0_ooo=@3oool3000004d0oooo00<0001o0?ooo`3oool0
6@3oool2000002X0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1;0?ooo`800000
7@3oool002@0oooo00@0omnL04P000000000M;oo>P3oool00`000000oooo0?ooo`1=0?ooo`030000
O`3oool0oooo01/0oooo0`00000W0?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0
C03oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`1=0?ooo`030000O`3o
ool0oooo01h0oooo0P00000U0?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03o
ool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0
oooo0200oooo0`00000R0?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0C03oool0
0`000000oooo0?ooo`0K0?ooo`00903oool00`3ookl0W9cO0?ooo`030?ooo`040?noM0000000001d
0;oooc@0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`0S0?ooo`800000803oool0
0`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool0
02@0oooo00@0oono09bLg`3ookl0M000100000000`00B9`0gooo0?ooo`0c0?ooo`030000003oool0
oooo04d0oooo00<0001o0?ooo`3oool09@3oool2000001h0oooo00<000000?ooo`3oool0AP3oool0
0`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000T0?ooo`090?oOW01800000000
0000B02Lgol0oooo0?oog`2LB7@0_ooo03D0oooo0`00001=0?ooo`030000O`3oool0oooo02L0oooo
0`00000K0?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0B`3oool2000001d0oooo
000P0?ooo`040?oog`2LB000001d0;oooch0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo
0?ooo`0Z0?ooo`8000006@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo
00<000000?ooo`3oool06`3oool002X0oooo00@0oono07@0000007@0_ooo=03oool00`000000oooo
0?ooo`1=0?ooo`030000O`3oool0oooo02`0oooo0`00000F0?ooo`030000003oool0oooo04H0oooo
00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00903oool0103ookl0W9cO0?oo
_`1d0003000000030000B02Lgol0oooo03@0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo
0?ooo`0_0?ooo`<000004`3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo
00<000000?ooo`3oool06`3oool002@0oooo00<0omnL04P0000000000P0000000`00M;l0oooo0?oo
o`0f0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0<P3oool300000100oooo00<0
00000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000V
0?ooo`030?oOW01807@0_ooo03T0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`0e
0?ooo`<000003@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<00000
0?ooo`3oool06`3oool002D0oooo00<0oono07A8W03Oool0;@3oool0102oM00000000000001d_ol9
0?ooo`<00000C@3oool00`0007l0oooo0?ooo`0h0?ooo`<000002P3oool00`000000oooo0?ooo`16
0?ooo`030000O`3oool0oooo04/0oooo0P00000M0?ooo`00903oool00`3ooml0W4Qd0;ooo`030?oo
o`040?noM0000000001d0;ooobH0oooo00H0gia8001d_`3oool0omnL04P00018W=l80?ooo`030000
003oool0oooo04d0oooo00<0001o0?ooo`3oool0>`3oool2000000P0oooo00<000000?ooo`3oool0
AP3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000T0?ooo`040?noM000
B9`0gono07@000@0000000<004RL0=ooo`3oool0903oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00L0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`0m0?ooo`800000
1P3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool0
6`3oool002@0oooo00T0omnL04P000000000000004RLg`3oool0oooO09a8M02oool09P3oool00`3o
oml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`H0oooo00<000000?ooo`3oool0C@3oool0
0`0007l0oooo0?ooo`0o0?ooo`<000000`3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0
oooo04`0oooo00<000000?ooo`3oool06`3oool005<0oooo00<0oooO09a80000M;l00`3oool00`3O
W4P0001809cOo`060?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0@P3oool20000
00030?ooo`000000oooo04L0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?oo
o`00E03oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00H0ooooo`00000b000001d0
oooo000R0?ooo`030;mdM02oool0oooo02l0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`060?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool010000000oooo
0000000000150?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool00280oooo
00<0_gAd0;ooo`3oool0;`3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00D0oooo
00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`040000003oool0oooo0?ooo`80
0000@`3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000R0?ooo`030;md
M02oool0oooo02l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`050?ooo`03
0000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`030?ooo`<0
0000@03oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000R0?ooo`030;md
M02oool0oooo0300oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`050?ooo`030000
003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`060?ooo`800000
?P3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000R0?ooo`030;mdM02o
ool0oooo0340oooo00H0gia80000M02oool0oooO09a80018W=l60?ooo`030000003oool0oooo04d0
oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`080?ooo`<00000>`3oool00`0007l0
oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000R0?ooo`030;mdM02oool0oooo0380oooo
00@0gia800000000000007Bo1`3oool3000004d0oooo00<0001o0?ooo`3oool0A@3oool00`000000
oooo0?ooo`0;0?ooo`800000>@3oool00`0007l0oooo0?ooo`1;0?ooo`8000007@3oool00280oooo
00<0_gAd0;ooo`3oool0?@3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo
00<000000?ooo`3oool03@3oool3000003H0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo
0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<0
00000?ooo`3oool0403oool3000003<0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?oo
o`0K0?ooo`009@3oool00`2oM7@0_ooo0?ooo`0j0?ooo`030000003oool0oooo04d0oooo00<0001o
0?ooo`3oool0A@3oool00`000000oooo0?ooo`0C0?ooo`800000<@3oool00`0007l0oooo0?ooo`1<
0?ooo`030000003oool0oooo01/0oooo000M0?ooo`030?oo_`2LW=l0oooo0080oooo00D0oono07@0
00000000001d0;ooo`0k0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool0
0`000000oooo0?ooo`0E0?ooo`<00000;P3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0
oooo01/0oooo000M0?ooo`030?noM000000000000080000000D0001d0;ooo`3oool0gibL0;ooo`0k
0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`0H
0?ooo`800000;03oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000M0?oo
o`030=nLB000M;l0oooo0480oooo0`00001=0?ooo`030000O`3oool0oooo04D0oooo00<000000?oo
o`3oool06P3oool2000002X0oooo00<0001o0?ooo`3oool0B`3oool2000001d0oooo001R0?ooo`03
0000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`0L0?ooo`<0
00009`3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000
003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`0O0?ooo`800000
9@3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000[0?ooo`030=nLB000
07@0_ooo03@0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003o
ool0oooo0240oooo0`00000R0?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3o
ool002@0oooo00<0oono07Ad_`3oool0103oool00`3ookl0M4RL0=ooo`0d0?ooo`030000003oool0
oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`0T0?ooo`800000803oool0
0`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo01/0oooo000T0?ooo`030?oOW01807@0_ooo
00<0oooo00<0okmd0000M02oool0=@3oool3000004d0oooo00<0001o0?ooo`3oool0A@3oool00`00
0000oooo0?ooo`0V0?ooo`8000007P3oool00`0007l0oooo0?ooo`1;0?ooo`8000007@3oool002@0
oooo00L0omnL04P000000000001809cOo`3ogi`0B7Bo03L0oooo00<000000?ooo`3oool0C@3oool0
0`0007l0oooo0?ooo`150?ooo`030000003oool0oooo02P0oooo0`00000K0?ooo`030000O`3oool0
oooo04`0oooo00<000000?ooo`3oool06`3oool002H0oooo00<0oono07@0000000000P0000000`1d
_ol0oooo0?ooo`0d0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`00
0000oooo0?ooo`0[0?ooo`8000006@3oool00`0007l0oooo0?ooo`1<0?ooo`030000003oool0oooo
01/0oooo000U0?ooo`090?oo_`1d07@0_ooo0?ooo`3ooml0W4P00000000007@0_ooo03@0oooo00<0
00000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo02d0oooo0`00
000F0?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3oool002@0oooo00<0oono
07@0M02oool0103oool00`3ogi`0B7Bo0?ooo`0d0?ooo`030000003oool0oooo04d0oooo00<0001o
0?ooo`3oool0A@3oool00`000000oooo0?ooo`0`0?ooo`800000503oool00`0007l0oooo0?ooo`1<
0?ooo`030000003oool0oooo01/0oooo000T0?ooo`030?noM0000000M;oo00<0oooo00<0oono07A8
W03Oool0=@3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?oo
o`3oool0<P3oool200000180oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?oo
o`00HP3oool3000004d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`0d0?ooo`<0
00003`3oool00`0007l0oooo0?ooo`1;0?ooo`8000007@3oool004L0oooo00@0_g@0000000000000
M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000B9`0gooo2@3oool0
0`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0=`3oool2
000000d0oooo00<0001o0?ooo`3oool0C03oool00`000000oooo0?ooo`0K0?ooo`00AP3oool01P3O
W4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Bo
o`3oool0omnL04Qd_`T0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`03
0000003oool0oooo03T0oooo0`00000:0?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3o
ool06`3oool004D0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?ooo`030?no
M018W=l0oooo00L0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000
003oool0oooo03`0oooo0P0000080?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool0
6`3oool004D0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo
_`1d07@0_ooo00L0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000
003oool0oooo03h0oooo0P0000060?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool0
6`3oool004D0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2L
B000M;oo00L0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003o
ool0oooo0400oooo0`0000030?ooo`030000O`3oool0oooo04`0oooo00<000000?ooo`3oool06`3o
ool004H0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030?oo_`1d0000
B9cO00L0oooo0`00002E0000O`030000000007l0001o04<0001o0P00001>0000O`<000007@3oool0
04H0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo
00L0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo
04D0oooo0P00001>0?ooo`030000003oool0oooo01/0oooo00160?ooo`030?oOW0180000B9cO00<0
oooo00<0_g@00018W03Oool01P3oool01P3ogml0gooo0?noM0000000001d0;ooo`L0oooo00<00000
0?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o
000000000000C03oool00`000000oooo0?ooo`0K0?ooo`00AP3oool0103ooml0W4P00018W03Oool2
0?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000001809cOo`080?ooo`030000003o
ool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3o
ool0oooo00800000BP3oool00`000000oooo0?ooo`0K0?ooo`00A`3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3oool02@3oool00`000000oooo0?ooo`1=
0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`02
0?ooo`800000B03oool00`000000oooo0?ooo`0K0?ooo`00B03oool01P3OW4P0001d0;ooo`3ooml0
W4P004RLg`P0oooo00<0omnL07BLg`3oool02@3oool3000004d0oooo00<0001o0?ooo`3oool0A@3o
ool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo00@0oooo0P0000150?ooo`8000007@3o
ool004T0oooo00@0gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO00H0oooo
00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo
00<0001o0?ooo`3oool01P3oool2000004@0oooo00<000000?ooo`3oool06`3oool00680oooo00<0
00000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0
001o0?ooo`3oool0203oool200000480oooo00<000000?ooo`3oool06`3oool00680oooo00<00000
0?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o
0?ooo`3oool02P3oool200000400oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?oo
o`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?oo
o`3oool0303oool2000003h0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3o
ool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3o
ool03P3oool2000003`0oooo00<000000?ooo`3oool06`3oool00680oooo0`00001=0?ooo`030000
O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`0@0?ooo`800000
>@3oool2000001d0oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3o
ool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo0180oooo0P00000h0?ooo`030000003o
ool0oooo01/0oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool0
0`000000oooo0?ooo`160?ooo`030000O`3oool0oooo01@0oooo0P00000f0?ooo`030000003oool0
oooo01/0oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`00
0000oooo0?ooo`160?ooo`030000O`3oool0oooo01H0oooo0P00000d0?ooo`030000003oool0oooo
01/0oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000
oooo0?ooo`160?ooo`030000O`3oool0oooo01P0oooo0P00000b0?ooo`030000003oool0oooo01/0
oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo
0?ooo`160?ooo`030000O`3oool0oooo01X0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0K0?ooo`00HP3oool3000004d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?oo
o`160?ooo`030000O`3oool0oooo01/0oooo0P00000^0?ooo`8000007@3oool00680oooo00<00000
0?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o
0?ooo`3oool07@3oool2000002d0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?oo
o`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?oo
o`3oool07`3oool2000002/0oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3o
ool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3o
ool08@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000
003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000
O`3oool0oooo0280oooo0P00000X0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003o
ool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3o
ool0oooo02@0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0K0?ooo`00D`3oool0
0`3ookl0M000000000020000000304RLg`3oool0oooo00L0oooo0`00001=0?ooo`030000O`3oool0
oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`0U0?ooo`800000903oool2
000001d0oooo001E0?ooo`030?noM0000000M;oo00X0oooo00<000000?ooo`3oool0C@3oool00`00
07l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool09`3oool00`00
0000oooo0?ooo`0R0?ooo`030000003oool0oooo01/0oooo001E0?ooo`030?oog`2LB00007Bo00X0
oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0
oooo00<0001o0?ooo`3oool0:03oool200000280oooo00<000000?ooo`3oool06`3oool005H0oooo
00<0gia80018W03Oool02@3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo
00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`0Z0?ooo`030000003oool0oooo01l0oooo
00<000000?ooo`3oool06`3oool005H0oooo00<0omnL04P0B02Lgol02@3oool00`000000oooo0?oo
o`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?oo
o`0[0?ooo`8000007`3oool00`000000oooo0?ooo`0K0?ooo`00EP3oool00`3ookl0M00007Boo`09
0?ooo`<00000U@0007l00`000000001o0000O`1f0000O`030000000007l0001o01X0001o0`00000M
0?ooo`00E`3oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo04d0oooo00<0001o0?oo
o`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo02h0oooo0P00000L0?oo
o`030000003oool0oooo01/0oooo001G0?ooo`030=nLB000B9`0gooo00P0oooo00<000000?ooo`3o
ool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3o
ool0<03oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01/0oooo001G0?ooo`030?oO
W01804P0W=oo00P0oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000
003oool0oooo04H0oooo00<0001o0?ooo`3oool0<@3oool00`000000oooo0?ooo`0H0?ooo`030000
003oool0oooo01/0oooo001G0?ooo`030?oo_`1d0000M;oo00P0oooo00<000000?ooo`3oool0C@3o
ool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0<P3o
ool2000001P0oooo00<000000?ooo`3oool06`3oool005P0oooo00<0_g@0001d_`3oool01`3oool0
0`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool0
0`0007l0oooo0?ooo`0d0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool06`3oool0
05H0oooo00D0oono07@000000000001d0;ooo`070?ooo`<00000C@3oool00`0007l0oooo0?ooo`15
0?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0=@3oool00`000000oooo0?ooo`0C
0?ooo`8000007@3oool00680oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?oo
o`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0=P3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3o
ool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo03L0oooo00<000000?ooo`3o
ool04P3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`1=0?ooo`030000
O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`0h0?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool06`3oool00680oooo00<000000?ooo`3oool0C@3o
ool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0>@3o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01/0oooo001R0?ooo`030000003oool0
oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0
oooo03X0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool30000
04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo
03/0oooo00<000000?ooo`3oool03@3oool2000001d0oooo001R0?ooo`030000003oool0oooo04d0
oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo03`0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo
0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo
0?ooo`0m0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06`3oool00680oooo00<0
00000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0
001o0?ooo`3oool0?P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo01/0oooo001R
0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`16
0?ooo`030000O`3oool0oooo03l0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0K
0?ooo`00HP3oool3000004d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?oo
o`030000O`3oool0oooo0400oooo00<000000?ooo`3oool0203oool2000001d0oooo001R0?ooo`03
0000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`03
0000O`3oool0oooo0400oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0K0?ooo`00
HP3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0
AP3oool00`0007l0oooo0?ooo`110?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
6`3oool00680oooo00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003o
ool0oooo04H0oooo00<0001o0?ooo`3oool0@@3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo01/0oooo001R0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool0
0`000000oooo0?ooo`160?ooo`030000O`3oool0oooo0480oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo
04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`130?ooo`030000003oool0oooo
00H0oooo00<000000?ooo`3oool06`3oool00680oooo0`00001=0?ooo`030000O`3oool0oooo04D0
oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`130?ooo`030000003oool0oooo00D0
oooo0P00000M0?ooo`00A@3oool00`3ookl0M000000000020000000704RLg`3oool0oooo0?ooo`3o
gi`0B00007Boo`020?ooo`060?oOW0180000000000000000B9`0gooo2@3oool00`000000oooo0?oo
o`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?oo
o`130?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool06`3oool004L0oooo00<0okmd
0000001d_ol0103oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo
2@3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0
AP3oool00`0007l0oooo0?ooo`130?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0
6`3oool004L0oooo00<0oooO09a80000M;l03P3oool00`3o_g@0B9cO0?ooo`070?ooo`030000003o
ool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3o
ool0oooo04@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0K0?ooo`00B03oool0
0`3OW4P004RL0=ooo`0=0?ooo`030?oo_`1d07@0_ooo00L0oooo00<000000?ooo`3oool0C@3oool0
0`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo00<0001o0?ooo`3oool0A03oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo01/0oooo00180?ooo`030?oOW01804P0W=oo
00d0oooo00<0oooO09a8001d_ol01`3oool00`000000oooo0?ooo`1=0?ooo`030000O`3oool0oooo
04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`140?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool06`3oool004P0oooo00<0oono07@0001d_ol03@3oool00`3ookl0
M00004RLg`070?ooo`<00000U@0007l00`000000001o0000O`2=0000O`030000000007l0001o00<0
001o0`00000M0?ooo`00B@3oool00`2oM00007Bo0?ooo`0<0?ooo`030?noM0000000M;oo00L0oooo
00<000000?ooo`3oool0C@3oool00`0007l0oooo0?ooo`150?ooo`030000003oool0oooo04H0oooo
00<0001o0?ooo`3oool0A03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01/0oooo
00190?ooo`030=nLB000B9`0gooo00T0oooo00H0omoO0=ooo`3o_g@000000000M02oool70?ooo`03
0000003oool0oooo04d0oooo00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`03
0000O`3oool0oooo04@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0K0?ooo`00
B@3oool00`3ogi`0B01809cOo`090?ooo`050?oo_`1d000000000000B02Lgol0203oool00`000000
oooo0?ooo`1=0?ooo`030000O`3oool0oooo04D0oooo00<000000?ooo`3oool0AP3oool00`0007l0
oooo0?ooo`150?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool06`3oool004T0oooo
00<0oono07@0001d_ol02P3oool00`3OW9`0_ooo0?ooo`090?ooo`030000003oool0oooo04d0oooo
00<0001o0?ooo`3oool0A@3oool00`000000oooo0?ooo`160?ooo`030000O`3oool0oooo04D0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0K0?ooo`00BP3oool00`2oM00007Bo0?oo
o`090?ooo`030?oOW01dW=l0oooo00T0oooo0`00001=0?ooo`030000O`3oool0oooo04D0oooo00<0
00000?ooo`3oool0AP3oool00`0007l0oooo0?ooo`1;0?ooo`8000007@3oool004P0oooo00D0oono
07@000000000001d0;ooo`0:0?ooo`050;md00000000000000000018W=l01P3oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0A@3oool00`000000oooo
0?ooo`160?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`0K0?ooo`00HP3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo01/0oooo001R
0?ooool00000<P00000M0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00eP3oool0103OW4P000000018W03OoooF0?oo
o`00eP3oool01@3o_g@004RL0=oOW018B9`0gooo0=D0oooo003I0?ooo`030?oOW018M;l0oooo0=@0
oooo003J0?ooo`030?noM01d_ol0oooo0=<0oooo00380?ooo`0:0=nLB000000000000000001d_ol0
oooo0=nLB01d_ol0oooo0=nLB0@0000000L0B9cO0?ooo`3oool0oooo0?noM0000000M;oo00L0oooo
00H0gia800000000000000000000B02Lgol70?ooo`070;md00000000001809cOo`3oool0_g@0001d
_`040?ooo`030?oo_`1d000000000080000000L0B9cO0?ooo`3oool0okmd0000000004P0W=oo00@0
oooo00H0gia800000000000000000000B02Lgol60?ooo`040?noM0000000001d0;ooo`80oooo00D0
okmd000000000000001809cOo`030?ooo`040?noM0000000001809cOohH0oooo00390?ooo`030?oo
g`2LB7@0_ooo0080oooo00P0_g@00000M02oool0oooo0?oog`2LB000001809cOo`@0oooo00@0gia8
0000M01dM4P0W=oo1`3oool0103ooml0W4P00018W03Oool80?ooo`080=nLB000B9`0gooO09adW03O
ool0okmd0000M02oool30?ooo`0=0?oog`2LB000M;oo0?ooo`3ogi`0B00007Boo`3oool0okmd0000
B02Lgml0W7BL0=ooo`040?ooo`040?oog`2LB00004RL0=ooo`L0oooo00`0omnL04P0M02ookl0W9cO
0?ooo`3ogi`0B01809cOo`3oool0omnL04P00000M;l20?ooo`050?noM00004P0W=oO09adW03Oool0
Q@3oool00<X0oooo00L0giad09cOo`3oool0oooo0;md0000B4P0B7Bo0080oooo00<0okmd0000B02L
gol0103oool01@3OW4P004RL0=oog`2LM9`0gooo00L0oooo00<0okmd0000B02Lgol0203oool00`3o
_g@0001809cOo`020?ooo`030?oOW0180000M;oo00@0oooo00X0_gAd0;ooo`3oool0oono07@0001d
_ol0oooo0?oo_`1d0000M;oo1`3oool00`3o_g@0001809cOo`070?ooo`080?oo_`1d0000M;oo0=nL
W02oool0gia80018W03Oool20?ooo`070?oOW0180000B9cO0?ooo`3ookl0M00007Boo`270?ooo`00
bP3oool0303ogi`0M9cO0?ooo`3ooml0W4P0001d_`3o_g@0B9cO0?ooo`3ookl0M00004RLg`@0oooo
00D0_g@00018W03Oool0omnL07BLg`070?ooo`030?oOW0180000M;oo00P0oooo00<0oono07@00018
W=l00P3oool0103ooml0W4P00018W03Oool60?ooo`030=nLB00007@0_ooo0080oooo00<0_g@0001d
_`3oool01P3oool00`3ookl0M00004RLg`080?ooo`070;md0018W=l0oooo0?ooo`3o_g@0001d0;oo
o`030?ooo`060=nLB00007@0_ooo0?ooo`2oM00007BoQ`3oool00<X0oooo00d0oooO09adW03Oool0
oono07@00018W=l0oooo0=nLM02Lgol0oooo0;md0000B9`0gooo0080oooo00<0oooO09a80000M;l0
0P3oool00`3OW9`0_ooo0?ooo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0_ibL0=ooo`3oool0
0`3oool00`2oM00004RL0=ooo`020?ooo`030;md0018M7@0M9cO00D0oooo00<0okmd0000B02Lgol0
0`3oool00`3OW4P004RL0=ooo`060?ooo`030?oog`2LB00007Bo00P0oooo00L0gia80018W03Oool0
oooo0?oo_`1d0000M;oo00<0oooo00L0omnL04P0B02Lgol0oooo0=nLB000B9`0gooo08H0oooo003;
0?ooo`0<0=nLM02Lgol0omnL04P00018W=l0oooo0?oog`2LM9`0gooo0=nLB00004P0W=oo0P3oool0
0`3ooml0W4P0001d_`020?ooo`030?oo_`1dM;l0oooo00H0oooo00L0_g@00018W03Oool0oooo0?oo
g`2LB4P0W=oo00@0oooo00<0gia80000B02Lgol00P3oool0103o_g@004RL0;nLB01d_ol30?ooo`0:
0?oOW01804P0W=oo0?ooo`3OW7@0W=oo0?ooo`3ogi`0B01809cOo`L0oooo00<0gia80000B018B000
0`0000001`18W=l0oooo0?ooo`3oool0omnL04P0M02oool00P3oool00`3OW4P0001d0;ooo`020?oo
o`070?oo_`1d0000M;oo0?ooo`3ogi`0B01809cOo`260?ooo`00b`3oool01@3ogi`0M9cO0?oOW018
0000M;oo0080oooo00D0omnL07BLg`3ogi`0B00004RLg`020?ooo`030?oo_`1d0000B9cO0080oooo
00<0omnL04P0M02oool01P3oool00`3o_g@00000000000030000000307Boo`3oool0oooo0080oooo
00<0omnL04P00018W=l00P3oool04@3ookl0M01809cOo`2oM000M7Ad0018W03Oool0omnL04P0B02L
gol0oooO09a8001d_ol0oooo0?oog`2LB000B9cO00L0oooo00<0okmd0000B02Lgol00P3oool0203o
gi`0B0000000M02oool0oooo0?oo_`1d0000M;oo0`3oool02`2oM000001d0;ooo`3oool0okmd0000
M02oool0oooo0?oog`2LB000B9cO08H0oooo003;0?ooo`050?oog`2LM9`0gkmd0000B02Lgol00`3o
ool02@3OW4P0M;oO09a80000B9`0gooo0;md00000000001809cOo`020?ooo`030=nLB000B9`0gooo
00H0oooo00<0oono07@0001d_ol00P3oool00`3OW4P0B9cO0?ooo`020?ooo`0?0?noM00000000000
0018W03Oool0omnL04P000000000M;oo0?ooo`2oM000001d0;ooo`3oool0_g@000<0000000<0B9cO
0?oo_`1d00000`0000000`1d_ol0oooo0?ooo`030?ooo`030?oo_`1d0000B9cO00<0oooo00<0omnL
04P00000M;l00P3oool00`2oM000B9cO0?ooo`030?ooo`080?noM000000000000000001d_ol0oooo
0?oo_`1d00030000000307Boo`3oool0oooo0880oooo003<0?ooo`040=nLM01dM4P0001809cOo`<0
oooo00D0oooO09a8M02LW4P0001809cOo`0?0?ooo`030?oog`2LB00007Bo0080oooo00<0oooO09ad
W03Oool05`3oool00`3o_g@0001d0;ooo`070?ooo`030;md0000B9`0gooo00<0oooo00L0_g@00000
M02oool0oooo0=nLB000B9`0gooo00/0oooo00<0okmd0000M02oool0Q@3oool00<`0oooo00@0omnL
04P0000007@0_ooo103oool0103ookl0M4Q804P00018W=l@0?ooo`030=nLB000B9`0gooo00@0oooo
00<0omnL09boo`3oool05@3oool00`3OW7@0W=oo0?ooo`060?ooo`030=nLB00007@0_ooo00<0oooo
00L0_g@00000M02oool0oooo0?oOW01807@0_ooo00`0oooo00<0giad09cOo`3oool0Q03oool00<`0
oooo00@0oooO09a80000B9`0gooo1@3oool0103o_g@000000000M02oool?0?ooo`030?noM00004P0
W=oo00@0oooo00P0gia8001d_`3oool0oooo0?oo_`1d000004RL0=oooaT0oooo00<0okmd0000001d
_ol00P3oool00`3o_g@00000001d_`020?ooo`030?oog`2LB000M;oo09<0oooo003;0?ooo`050?no
M000000000000018W03Oool01P3oool0102oM000000000000000M;l=0?ooo`030;md000000000000
00H0000000<0001d0;ooo`3oool0703oool00`2oM00000000000000400000003001d_`3oool0oooo
0080oooo00<0gia804RLg`3oool0TP3oool00?l0oooo6P3oool0103ooml0W4P000000000M;nC0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.7339, -2.54255, \
0.0103725, 0.0263561}}]
}, Closed]],

Cell["\<\
adds frame, with a list of labels given as strings, adds gridlines, and \
prints a title.  Many more options are available. \
\>", "Text"],

Cell[TextData[{
  "Often it is necessary to use ",
  ButtonBox["Evaluate",
    ButtonStyle->"RefGuideLink"],
  " to prepare a function for plotting.  For example,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[D[ArcSin[x], x], {x, \(-1\), 1}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(General::"ivar"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(-0.9999999166666667`\\) is not a valid variable. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::ivar\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"ivar"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(-0.9999999166666667`\\) is not a valid variable. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::ivar\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"ivar"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(-0.9999999166666667`\\) is not a valid variable. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::ivar\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(General :: \\\"ivar\\\"\\) will be suppressed during this calculation. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(\\(\[PartialD]\\_\\(x\\)\\) \\(\\(ArcSin[x]\\)\\)\\) is not a \
machine-size real number at \\!\\(x\\) = \\!\\(-0.9999999166666667`\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(\\(\[PartialD]\\_\\(x\\)\\) \\(\\(ArcSin[x]\\)\\)\\) is not a \
machine-size real number at \\!\\(x\\) = \\!\\(-0.9188660168541684`\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), \
"\<\"\\!\\(\\(\[PartialD]\\_\\(x\\)\\) \\(\\(ArcSin[x]\\)\\)\\) is not a \
machine-size real number at \\!\\(x\\) = \\!\\(-0.8303824002812525`\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(Plot :: \\\"plnr\\\"\\) will be suppressed during this calculation. \\!\
\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.952381 0.0147151 0.588604 [
[.21429 .00222 -9.5 -15.6875 ]
[.21429 .00222 9.5 0 ]
[.40476 .00222 -9.5 -15.6875 ]
[.40476 .00222 9.5 0 ]
[.59524 .00222 -9.625 -15.6875 ]
[.59524 .00222 9.625 0 ]
[.78571 .00222 -9.5 -15.6875 ]
[.78571 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.01131 .13244 -19 -7.84375 ]
[.01131 .13244 0 7.84375 ]
[.01131 .25016 -19 -7.84375 ]
[.01131 .25016 0 7.84375 ]
[.01131 .36788 -19.25 -7.84375 ]
[.01131 .36788 0 7.84375 ]
[.01131 .4856 -19 -7.84375 ]
[.01131 .4856 0 7.84375 ]
[.01131 .60332 -10 -7.84375 ]
[.01131 .60332 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.21429 .01472 m
.21429 .02097 L
s
gsave
.21429 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.40476 .01472 m
.40476 .02097 L
s
gsave
.40476 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.59524 .01472 m
.59524 .02097 L
s
gsave
.59524 .00222 -70.625 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.78571 .01472 m
.78571 .02097 L
s
gsave
.78571 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.2619 .01472 m
.2619 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.5 .01472 m
.5 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.7381 .01472 m
.7381 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .13244 m
.03006 .13244 L
s
gsave
.01131 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.02381 .25016 m
.03006 .25016 L
s
gsave
.01131 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.02381 .36788 m
.03006 .36788 L
s
gsave
.01131 .36788 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.02381 .4856 m
.03006 .4856 L
s
gsave
.01131 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.02381 .60332 m
.03006 .60332 L
s
gsave
.01131 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10301 m
.02756 .10301 L
s
.02381 .16187 m
.02756 .16187 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .22073 m
.02756 .22073 L
s
.02381 .27959 m
.02756 .27959 L
s
.02381 .30902 m
.02756 .30902 L
s
.02381 .33845 m
.02756 .33845 L
s
.02381 .39731 m
.02756 .39731 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .45617 m
.02756 .45617 L
s
.02381 .51503 m
.02756 .51503 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .57389 m
.02756 .57389 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00K03oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00<0gia8
0000000000000`0000000`18W=l0oooo0?ooo`0d0?ooo`040;md00000000000007Boo`<0oooo00<0
omnL04P0001d_ol01@3oool00`3ooml0W4P004RLg`0g0?ooo`040;md00000000000007Boo`<0oooo
00<0omnL04P0001d_ol00`3oool01@3ooml0W4P000000000000007Bo03P0oooo00@0_g@000000000
0000M;oo0`3oool00`3ogi`0B00007Boo`030?ooo`060?oo_`1d0000000000000000B9`0gooo>P3o
ool00`3ookl0M000000000020000000304RLg`3oool0oooo01<0oooo001[0?ooo`060=nLB000M;l0
oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`030?ooo`030?noM000M;l0oooo0080oooo
00<0_gAd0;ooo`3oool0<`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0
001d_ol01P3oool00`3OW4P004RL0=ooo`0e0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3o
ool00`3ogi`0B00007Boo`030?ooo`070=nLB000M;l0oooo0?oo_`1d000004RL0=ooo`0e0?ooo`06
0=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`030?oo_`1d0000
B9cO0080oooo00<0gia80000M02oool0>`3oool00`3o_g@0000007Boo`0F0?ooo`00JP3oool00`3o
oml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0okmd001d_`3oool0=P3oool00`3o
oml0W4P004RLg`020?ooo`030?noM00004P0W=oo00X0oooo00<0omnL04P0B02Lgol0=03oool00`3o
oml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0oooO09a8001d_ol00P3oool00`3o
gi`0B00004RLg`0d0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3o
_g@004RL0=ooo`030?ooo`030;md0018W=l0oooo03X0oooo00<0oooO09a80000M;l05P3oool006X0
oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?noM000M;l0oooo
03D0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?ooo`030?noM0000000
000000@0000000<004RL0=ooo`3oool0<@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB000
04RL0=ooo`D0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`0c0?ooo`030?oog`2L
B000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3o_g@004RL0=ooo`030?ooo`030;md
0000M;l0oooo03/0oooo00<0gia80018W03Oool05@3oool006X0oooo00<0oooO09a80000M;l00`3o
ool00`3OW4P0001809cOo`090?ooo`030?noM018W=l0oooo03@0oooo00<0oooO09a80000M;l00`3o
ool00`3OW4P0001809cOo`050?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0<`3o
ool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a80000M;l00`3o
ool00`3OW4P0001809cOo`0c0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01@3o
ool00`3ooml0W4P007Boo`020?ooo`030?oo_`1d0000B9cO03`0oooo00<0omnL04P0B02Lgol05@3o
ool006/0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030=nLB018W=l0
oooo03@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`060?ooo`070?oo_`1d07@0
_ooo0?ooo`3o_g@004RL0=ooo`0d0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0
1P3oool00`2oM000001d0;ooo`020?ooo`030=nLB00004P0W=oo03@0oooo00<0_g@00018W03Oool0
0P3oool00`3ogi`0B00004RLg`070?ooo`060;md000004P0W=oO09a8000007@0_ooo?03oool00`3o
okl0M00007Boo`0E0?ooo`00J`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00/0
oooo00<0_g@007Boo`3oool0<`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00L0
oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgold0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO
09a80000M;l01P3oool0203o_g@0000004RLg`3oool0oooO09a80000B9`0gooo=03oool00`3OW4P0
001809cOo`020?ooo`030?oog`2LB00007Bo00P0oooo00D0oooO09a80000000004RL0=ooo`0m0?oo
o`030;md0000M;l0oooo01@0oooo001[0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03O
ool02P3oool00`3ogi`0B01d0;ooo`0c0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03O
ool01`3oool01@3o_g@007Bo0?oog`2LB000B9cO03@0oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`050?ooo`030?oog`2LB00000000080000000<0001d0;ooo`3oool0=03oool00`3o
gi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00H0_g@00000B02Lgol0omnL04Q8W03O
ooll0?ooo`030=nLB000B9`0gooo01@0oooo001[0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0
gia80018W03Oool02P3oool00`3ooml0W4P007Boo`0c0?ooo`040?oog`2LB00004RL0=ooo`80oooo
00<0gia80018W03Oool0203oool0103OW4P0B9cO0;md0000M;ld0?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0gia80018W03Oool01P3oool00`3ookl0M000001d_`0h0?ooo`040?oog`2LB00004RL
0=ooo`80oooo00<0gia80018W03Oool01P3oool00`3ogi`0B00007Boo`020?ooo`030?noM018W=l0
oooo03/0oooo00<0omnL04P0B02Lgol0503oool006`0oooo00<0omnL04P0001d_ol00P3oool00`3O
W4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO03@0oooo00<0omnL04P0001d_ol00P3oool00`3O
W4P007Bo0?ooo`090?ooo`040=nLB01dM7@0001d0;oooc@0oooo00<0omnL04P0001d_ol00P3oool0
0`3OW4P007Bo0?ooo`070?ooo`030?oOW0180000B9cO03P0oooo00<0omnL04P0001d_ol00P3oool0
0`3OW4P007Bo0?ooo`060?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03Oool0>`3oool0
0`3ookl0M00007Boo`0D0?ooo`00K@3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00L0
okmd0018W03Oool0oooo0=nLB00004P0W=oo03D0oooo00H0gia80000M02oool0oooO09a80018W=l;
0?ooo`030;md000004P0W=oo03D0oooo00H0gia80000M02oool0oooO09a80018W=l90?ooo`040?oo
g`2LB000001d0;ooocL0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`030?oog`2LB000
B9cO0080oooo00<0_g@0001d_`3oool0?03oool00`2oM00007Bo0?ooo`0C0?ooo`00KP3oool0103O
W4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol0=`3oool0103OW4P000000000
0000M;l<0?ooo`030?oog`2LB000B9cO03H0oooo00@0gia800000000000007Bo303oool0103ookl0
M00000000018W=lf0?ooo`040=nLB00000000000001d_`T0oooo00D0oooO09a8000000000000001d
_`0l0?ooo`050?oo_`1d000000000000M02oool04`3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3o
ool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool0
02d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40
oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002<0ooooo`000022000000`0oooo000]
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0?
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0C
0?ooo`00;@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04H0oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo04H0oooo00<000000?oo
o`3oool04`3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo
0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<00000
0?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`03
0000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3o
ool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`800000o`3ooon20?ooo`00;@3oool00`00
0000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo
00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]
0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0oooo
P@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o
0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo0P00003o0?oooh80
oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3o
oon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo
0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo
0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<00000
0?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`03
0000003oool0oooo0?l0ooooP@3oool002d0oooo0P00003o0?oooh80oooo000]0?ooo`030000003o
ool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`00
0000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo
00<000000?ooo`3oool0o`3ooon10?ooo`00403oool0102oM00000000000001d_ol30?ooo`030?oO
W0180000M;oo0080oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`080?ooo`030000
003oool0oooo0?l0ooooP@3oool000l0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`080?ooo`03
0000003oool0oooo0?l0ooooP@3oool000h0oooo00<0oooO09a80018W=l00P3oool00`3o_g@00018
09cOo`080?ooo`030?noM000M;l0oooo00`0oooo00<000000?ooo`3oool0o`3ooon10?ooo`003P3o
ool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool0
2`3oool00`000000oooo0?ooo`3o0?oooh40oooo000>0?ooo`030?oog`2LB00007Bo00<0oooo00<0
gia80000B02Lgol02@3oool00`3o_g@0B9cO0?ooo`0:0?ooo`030000003oool0oooo0?l0ooooP@3o
ool000l0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030=nLB018W=l0
oooo00T0oooo0`00003o0?oooh40oooo000?0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l02`3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo0?l0ooooP@3oool000l0
oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`0:0?ooo`030?oOW01807@0_ooo00P0
oooo00<000000?ooo`3oool0o`3ooon10?ooo`003`3oool0103ooml0W4P00018W03Oool20?ooo`03
0=nLB000B9`0gooo00X0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`3o0?oooh40
oooo000@0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool02P3oool00`3ookl0M000
04RLg`080?ooo`030000003oool0oooo0?l0ooooP@3oool00140oooo00H0gia80000M02oool0oooO
09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?ooo`030000003oool0
oooo0?l0ooooP@3oool00180oooo00@0gia800000000000007Bo2@3oool01@3ogi`0B00000000000
0000M;oo00T0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o
0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3o
ool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`800000o`3o
oon20?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo
0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo
0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<00000
0?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`03
0000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3o
ool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool0
02d0oooo0P00003o0?oooh80oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo
00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]
0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0oooo
P@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o
0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo0P00003o0?oooh80
oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3o
oon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo
0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00403oool0102oM0000000
0000001d_ol30?ooo`030?oOW0180000M;oo00D0oooo00<0oooO09a80018W=l02`3oool00`000000
oooo0?ooo`3o0?oooh40oooo000?0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3o
gi`0B00007Boo`060?ooo`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`003P3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00X0oooo00<0omnL04P0B02L
gol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000>0?ooo`030?oog`2LB000B9cO0080oooo
00@0oooO09a80000B9`0gooo1@3oool00`3o_g@0000000000004000000030018W03Oool0oooo00L0
oooo00<000000?ooo`3oool0o`3ooon10?ooo`003P3oool00`3ooml0W4P0001d_`030?ooo`030=nL
B00004P0W=oo00D0oooo00<0oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`090?ooo`030000
003oool0oooo0?l0ooooP@3oool000l0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RL
g`060?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`090?ooo`<00000o`3ooon10?ooo`00
3`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00L0oooo00H0omnL04Q8W03Oool0
omnL04P0B02Lgol90?ooo`030000003oool0oooo0?l0ooooP@3oool000l0oooo00<0omnL04P00018
W=l00`3oool00`2oM00004RL0=ooo`070?ooo`050?noM000M;l0oooO09a80018W=l02@3oool00`00
0000oooo0?ooo`3o0?oooh40oooo000?0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018
W03Oool0203oool0103OW4P0B9cO0;md0000M;l90?ooo`030000003oool0oooo0?l0ooooP@3oool0
0100oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`090?ooo`040=nLB01dM7@0001d
0;ooo`P0oooo00<000000?ooo`3oool0o`3ooon10?ooo`004@3oool01P3OW4P0001d0;ooo`3ooml0
W4P004RLg`/0oooo00<0_g@00000B02Lgol0203oool00`000000oooo0?ooo`3o0?oooh40oooo000B
0?ooo`040=nLB00000000000001d_``0oooo00<0oooO09a80018W=l0203oool00`000000oooo0?oo
o`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?oo
o`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000
003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool2
00000?l0ooooPP3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000
oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<0
00000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?oo
o`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00
;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3o
ool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool200000?l0ooooPP3oool002d0
oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo
000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon1
0?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0
ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?oo
o`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?oo
o`3oool0o`3ooon10?ooo`00;@3oool200000?l0ooooPP3oool002d0oooo00<000000?ooo`3oool0
o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0
oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000
oooo0?ooo`3o0?oooh40oooo000@0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d
_ol00`3oool01@3ooml0W4P000000000000007Bo00/0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`003`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool0
1`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool02@3oool00`000000oooo0?ooo`3o0?oooh40oooo
000>0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3ooml0W4P007Bo
o`020?ooo`030?oOW0180000B9cO00T0oooo00<000000?ooo`3oool0o`3ooon10?ooo`003P3oool0
0`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<0oono07@0001d_ol00`3o
ool00`2oM000001d0;ooo`080?ooo`030000003oool0oooo0?l0ooooP@3oool000h0oooo00<0oooO
09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0gia8
0000B02Lgol0203oool00`000000oooo0?ooo`3o0?oooh40oooo000?0?ooo`030;md0000B9`0gooo
0080oooo00<0omnL04P00018W=l01P3oool00`2oM000001d0;ooo`020?ooo`030=nLB00004P0W=oo
00P0oooo0`00003o0?oooh40oooo000?0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000
M;l01P3oool0203o_g@0000004RLg`3oool0oooO09a80000B9`0gooo203oool00`000000oooo0?oo
o`3o0?oooh40oooo000?0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3oool0
0`3ooml0W4P000000002000000030000M02oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooon1
0?ooo`003`3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00<0oono
07@00000M;l0303oool00`000000oooo0?ooo`3o0?oooh40oooo000@0?ooo`030?oOW0180000M;oo
0080oooo00<0gia8001d_`3oool01`3oool00`3ogi`0B00004RLg`0;0?ooo`030000003oool0oooo
0?l0ooooP@3oool00140oooo00H0gia80000M02oool0oooO09a80018W=l90?ooo`040?oog`2LB000
001d0;ooo`T0oooo00<000000?ooo`3oool0o`3ooon10?ooo`004P3oool0103OW4P0000000000000
M;l<0?ooo`040?oo_`1d0000000004RLg`L0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3o
ool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool0
02d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40
oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo0P00003o0?oooh80oooo000]
0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0oooo
P@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o
0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3o
ool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003o
ool0oooo0?l0ooooP@3oool002d0oooo0P00003o0?oooh80oooo000]0?ooo`030000003oool0oooo
0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo
0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<00000
0?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`03
0000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3o
ool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool0
02d0oooo0P00003o0?oooh80oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo
00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]
0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000@0?ooo`040;md00000000000007Boo`<0
oooo00<0omnL04P0001d_ol00`3oool01P3ookl0M00000000000000004RL0=ooo`X0oooo00<00000
0?ooo`3oool0o`3ooon10?ooo`003`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0
omnL04P0001d_ol00P3oool00`3ookl0M00004RLg`020?ooo`030=nLB00007@0_ooo00T0oooo00<0
00000?ooo`3oool0o`3ooon10?ooo`003P3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0
W=oo00H0oooo00<0okmd0018W03Oool00`3oool00`2oM000B9cO0?ooo`080?ooo`030000003oool0
oooo0?l0ooooP@3oool000h0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool5
0?ooo`030?noM000B9`0gooo00<0oooo00<0_g@0001d_`3oool0203oool00`000000oooo0?ooo`3o
0?oooh40oooo000>0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01@3oool00`3o
oml0W4P007Boo`020?ooo`030?oo_`1d0000B9cO00T0oooo00<000000?ooo`3oool0o`3ooon10?oo
o`003`3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00L0oooo00H0_g@00000B02L
gml0W4P00000M02oool90?ooo`<00000o`3ooon10?ooo`003`3oool00`3OW4P0001809cOo`020?oo
o`030?oog`2LB00007Bo00P0oooo00D0oooO09a80000000004RL0=ooo`090?ooo`030000003oool0
oooo0?l0ooooP@3oool000l0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?oo
o`060;md000004P0W=oo0?oOW018B9`0gooo203oool00`000000oooo0?ooo`3o0?oooh40oooo000?
0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`3ogi`0B00007Bo
o`020?ooo`030?noM018W=l0oooo00L0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00403oool0
0`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00H0oooo00<0omnL04P0B02Lgol00P3oool0
0`3o_g@004RL0=ooo`070?ooo`030000003oool0oooo0?l0ooooP@3oool00140oooo00H0gia80000
M02oool0oooO09a80018W=l70?ooo`030?oog`2LB000B9cO0080oooo00<0_g@0001d_`3oool01`3o
ool00`000000oooo0?ooo`3o0?oooh40oooo000B0?ooo`040=nLB00000000000001d_`T0oooo00D0
oooO09a8000000000000001d_`090?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<0
00000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?oo
o`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00
;@3oool200000?l0ooooPP3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool0
0`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0
oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo
000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon1
0?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0
ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool200000?l0ooooPP3o
ool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oo
oh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0
o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0
oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool00`000000
oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool002d0oooo00<0
00000?ooo`3oool0o`3ooon10?ooo`00;@3oool200000?l0ooooPP3oool002d0oooo00<000000?oo
o`3oool0o`3ooon10?ooo`00;@3oool00`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000
003oool0oooo0?l0ooooP@3oool002d0oooo00<000000?ooo`3oool0o`3ooon10?ooo`00;@3oool0
0`000000oooo0?ooo`3o0?oooh40oooo000]0?ooo`030000003oool0oooo0?l0ooooP@3oool001`0
oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`090?ooo`030000003oool0oooo0?l0
ooooP@3oool001h0oooo00<0okmd0000001d_ol0303oool00`000000oooo0?ooo`3o0?oooh40oooo
000N0?ooo`030?oog`2LB00007Bo00`0oooo00<000000?ooo`3oool0o`3ooon10?ooo`007`3oool0
0`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo0?l0ooooP@3oool001l0oooo00<0omnL04P0
B02Lgol02`3oool00`000000oooo0?ooo`3o0?oooh40oooo000O0?ooo`030?oo_`1d0000M;oo00/0
oooo0`00003o0?oooh40oooo000P0?ooo`030;md0000M;l0oooo00X0oooo00<000000?ooo`3oool0
o`3ooon10?ooo`00803oool00`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo0?l0ooooP@3o
ool00200oooo00<0omnL04P0B02Lgol02P3oool00`000000oooo0?ooo`3o0?oooh40oooo000P0?oo
o`030?oo_`1d0000M;oo00X0oooo00<000000?ooo`3oool0o`3ooon10?ooo`008@3oool00`2oM000
07Bo0?ooo`090?ooo`030000003oool0oooo0?l0ooooP@3oool001l0oooo00D0oono07@000000000
001d0;ooo`3o0?ooohd0oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.12347, -0.125454, \
0.00410595, 0.00664356}}]
}, Closed]],

Cell[TextData[{
  "fails because ",
  StyleBox["x", "InlineInput"],
  " is substituted before the derivative is evaluated, while"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[D[ArcSin[x], x]], {x, \(-1\), 1}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.0147151 0.0543735 [
[.02381 .00222 -9.34375 -15.6875 ]
[.02381 .00222 9.34375 0 ]
[.2619 .00222 -13.8438 -15.6875 ]
[.2619 .00222 13.8438 0 ]
[.7381 .00222 -9.5 -15.6875 ]
[.7381 .00222 9.5 0 ]
[.97619 .00222 -5 -15.6875 ]
[.97619 .00222 5 0 ]
[.4875 .12346 -10 -7.84375 ]
[.4875 .12346 0 7.84375 ]
[.4875 .23221 -10 -7.84375 ]
[.4875 .23221 0 7.84375 ]
[.4875 .34096 -10.25 -7.84375 ]
[.4875 .34096 0 7.84375 ]
[.4875 .4497 -10 -7.84375 ]
[.4875 .4497 0 7.84375 ]
[.4875 .55845 -16 -7.84375 ]
[.4875 .55845 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .01472 m
.02381 .02097 L
s
gsave
.02381 .00222 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .01472 m
.2619 .02097 L
s
gsave
.2619 .00222 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .01472 m
.7381 .02097 L
s
gsave
.7381 .00222 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .01472 m
.97619 .02097 L
s
gsave
.97619 .00222 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .01472 m
.07143 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.16667 .01472 m
.16667 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.35714 .01472 m
.35714 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.45238 .01472 m
.45238 .01847 L
s
.54762 .01472 m
.54762 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.64286 .01472 m
.64286 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.83333 .01472 m
.83333 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.92857 .01472 m
.92857 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.5 .12346 m
.50625 .12346 L
s
gsave
.4875 .12346 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 .23221 m
.50625 .23221 L
s
gsave
.4875 .23221 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.5 .34096 m
.50625 .34096 L
s
gsave
.4875 .34096 -71.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.5 .4497 m
.50625 .4497 L
s
gsave
.4875 .4497 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(8) show
1.000 setlinewidth
grestore
.5 .55845 m
.50625 .55845 L
s
gsave
.4875 .55845 -77 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(10) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .0419 m
.50375 .0419 L
s
.5 .06909 m
.50375 .06909 L
s
.5 .09628 m
.50375 .09628 L
s
.5 .15065 m
.50375 .15065 L
s
.5 .17784 m
.50375 .17784 L
s
.5 .20502 m
.50375 .20502 L
s
.5 .2594 m
.50375 .2594 L
s
.5 .28658 m
.50375 .28658 L
s
.5 .31377 m
.50375 .31377 L
s
.5 .36814 m
.50375 .36814 L
s
.5 .39533 m
.50375 .39533 L
s
.5 .42252 m
.50375 .42252 L
s
.5 .47689 m
.50375 .47689 L
s
.5 .50408 m
.50375 .50408 L
s
.5 .53126 m
.50375 .53126 L
s
.5 .58564 m
.50375 .58564 L
s
.5 .61282 m
.50375 .61282 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02584 .61803 m
.02605 .5753 L
.02729 .4653 L
.02846 .40487 L
.02954 .36638 L
.03053 .33945 L
.03279 .29608 L
.03396 .27949 L
.03506 .26638 L
.03756 .24264 L
.04 .22505 L
.04262 .21012 L
.04793 .18777 L
.05297 .17252 L
.0575 .16189 L
.06244 .15252 L
.07302 .13753 L
.08429 .12619 L
.09493 .11814 L
.10458 .1123 L
.12383 .10339 L
.14465 .0964 L
.16458 .09132 L
.18565 .0871 L
.22362 .08149 L
.24313 .07929 L
.26405 .07731 L
.30541 .07429 L
.34525 .07221 L
.38358 .07079 L
.40336 .07024 L
.42436 .06979 L
.43373 .06962 L
.44383 .06947 L
.4533 .06935 L
.46211 .06926 L
.47169 .06918 L
.47675 .06915 L
.48215 .06913 L
.48488 .06912 L
.48784 .06911 L
.49063 .0691 L
.49196 .0691 L
.49321 .06909 L
.49442 .06909 L
.49553 .06909 L
.49676 .06909 L
.49744 .06909 L
.49809 .06909 L
.49937 .06909 L
Mistroke
.50056 .06909 L
.50186 .06909 L
.50259 .06909 L
.50325 .06909 L
.50446 .06909 L
.50559 .06909 L
.50814 .0691 L
.51064 .0691 L
.5133 .06911 L
.51814 .06913 L
.52254 .06915 L
.53324 .06922 L
.54337 .06932 L
.56332 .06958 L
.58175 .06991 L
.60046 .07034 L
.62107 .07094 L
.66133 .07251 L
.70007 .07463 L
.74126 .07778 L
.76183 .07981 L
.78094 .08206 L
.8191 .08797 L
.83959 .09228 L
.85819 .09723 L
.87742 .10389 L
.88733 .10819 L
.89823 .11388 L
.90784 .12005 L
.91814 .12835 L
.92903 .14003 L
.93431 .14731 L
.9392 .15542 L
.94352 .16408 L
.94829 .17594 L
.95309 .19143 L
.95575 .2023 L
.95819 .21435 L
.96032 .22709 L
.96265 .24438 L
.965 .26699 L
.96629 .28279 L
.96747 .30013 L
.96852 .31881 L
.96965 .34396 L
.97085 .37886 L
.97199 .42477 L
.97305 .48874 L
.97417 .60595 L
Mfstroke
.97417 .60595 m
.97422 .61803 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona
0?ooo`00803oool00`3ookl0M000000000020000000304RLg`3oool0oooo0540oooo00@0_g@00000
00000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000B9`0gooo
X`3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00H0omnL04P00000
000000000018W03Ooom@0?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3oool0603oool0
0280oooo00<0okmd0000001d_ol0D`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0
omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_j80oooo00H0gia8001d_`3o
ool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080oooo00H0okmd0000001d_ol0oooo0?oO
W018M;mB0?ooo`030?noM0000000M;oo01/0oooo000R0?ooo`030?oog`2LB00007Bo0580oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?ooo`030?noM018W=l0oooo09l0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?ooo`030?noM018W=l0oooo0500oooo00<0
oooO09a80000M;l06`3oool002<0oooo00<0gia80018W03Oool0D@3oool00`3ooml0W4P004RLg`02
0?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02oool0W`3oool00`3ooml0W4P004RL
g`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02oool0D@3oool00`3OW4P004RL
0=ooo`0J0?ooo`004`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0oooo00<0
omnL04P0B02Lgol0A03oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB000
07Bo0`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo09l0oooo00<0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo0540oooo00<0omnL04P0B02L
gol06P3oool002<0oooo00<0oono07@0001d_ol0DP3oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00X0oooo00<0oono07@00018W=l0X03oool00`2oM00004RL0=ooo`020?ooo`030?oO
W0180000B9cO00X0oooo00<0oono07@00018W=l0D@3oool00`3ookl0M00007Boo`0J0?ooo`00903o
ool00`2oM00007Bo0?ooo`1A0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02P3o
ool00`3o_g@0000007Boo`2P0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02P3o
ool00`3o_g@0000007Boo`1B0?ooo`030;md0000M;l0oooo01T0oooo000T0?ooo`030=nLB000B9`0
gooo0540oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`060?oOg`3Oool0
okmd0000000007@0_oooX03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo
00H0omoO0=ooo`3o_g@000000000M02ooomB0?ooo`030=nLB000B9`0gooo01T0oooo000T0?ooo`03
0?oOW01804P0W=oo0540oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?oo
o`050?oo_`1d000000000000B02Lgol0X@3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000
B9`0gooo00H0oooo00D0oono07@000000000001809cOo`1C0?ooo`030?oOW01804P0W=oo01T0oooo
000T0?ooo`030?oo_`1d0000M;oo0580oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?oo
o`070?ooo`030=nLW02oool0oooo0:<0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?oo
o`070?ooo`030=nLW02oool0oooo05@0oooo00<0oono07@0001d_ol06@3oool002D0oooo00<0_g@0
001d_`3oool0DP3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool0
Y03oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool0E@3oool00`2o
M00007Bo0?ooo`0H0?ooo`008`3oool01@3ookl0M0000000000007@0_ooo05<0oooo00@0gia80000
0000000007Bo2P3oool01@2oM0000000000000000000B9cO0:80oooo00@0gia800000000000007Bo
2P3oool01@2oM0000000000000000000B9cO0500oooo00D0oono07@000000000001d0;ooo`0H0?oo
o`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00f@3oool00`00
0000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo
0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D
0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`005@3ooooo000008/000004@3oool001h0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0100oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01P0oooo
000N0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0403oool00`000000oooo0?oo
o`0H0?ooo`007P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo05X0oooo00<00000
0?ooo`3oool0F`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo01P0oooo003I0?oo
o`030000003oool0oooo0=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo003I0?ooo`030000
003oool0oooo0=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo003I0?ooo`030000003oool0
oooo0=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo003I0?ooo`030000003oool0oooo0=@0
oooo003I0?ooo`<00000e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<00000
0?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3o
ool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03o
ool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0
oooo00<000000?ooo`3oool0e03oool00=00oooo5000003<0?ooo`00Y@3oool[000000T0oooo00<0
00000?ooo`3oool0203oool]000009l0oooo002E0?oooa000000=03oool00`000000oooo0?ooo`0e
0?oooa000000S`3oool008D0oooo400000140?ooo`030000003oool0oooo04D0oooo3`0000200?oo
o`00L@3ooolD000005@0oooo00<000000?ooo`3oool0E03oool<000007@0oooo001[0?ooo`H00000
J03oool00`000000oooo0?ooo`1P0?ooo`P00000K03oool006H0oooo1@00001^0?ooo`030000003o
ool0oooo06P0oooo2000001T0?ooo`00H@3oool5000007<0oooo00<000000?ooo`3oool0L03oool8
000005`0oooo001J0?ooo`L00000N03oool00`000000oooo0?ooo`1h0?ooo`D00000E`3oool005@0
oooo1P00001o0?ooo`030000003oool0oooo07d0oooo1000001C0?ooo`00D03oool4000008D0oooo
00<000000?ooo`3oool0P@3oool4000004l0oooo001<0?ooo`@00000R@3oool3000008D0oooo1000
001;0?ooo`00BP3oool2000008d0oooo00<000000?ooo`3oool0R@3oool3000004P0oooo00170?oo
o`<00000S`3oool00`000000oooo0?ooo`2<0?ooo`@00000A03oool004@0oooo0`00002B0?ooo`03
0000003oool0oooo0900oooo0`0000110?ooo`00@P3oool2000009D0oooo00<000000?ooo`3oool0
T`3oool2000003l0oooo00100?ooo`800000U`3oool00`000000oooo0?ooo`2E0?ooo`800000?@3o
ool003h0oooo0P0000280?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0203oool0
0`000000oooo0?ooo`2G0?ooo`800000>`3oool003d0oooo00<000000?ooo`3oool0R@3oool00`3o
_g@007Bo0?ooo`020?ooo`030;mdM02oool0oooo00P0oooo00<000000?ooo`3oool0V@3oool20000
03T0oooo000k0?ooo`800000S@3oool00`3o_g@007Bo0?ooo`0<0?ooo`030000003oool0oooo09/0
oooo00<000000?ooo`3oool0=P3oool003X0oooo00<000000?ooo`3oool0SP3oool00`3o_g@007Bo
0?ooo`0;0?ooo`030000003oool0oooo09`0oooo0P00000f0?ooo`00>@3oool00`000000oooo0?oo
o`2@0?ooo`030?noM018W=l0oooo00X0oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?oo
o`0c0?ooo`00=`3oool2000009@0oooo00<0gia804RLg`3oool02@3oool4000009h0oooo00<00000
0?ooo`3oool0<P3oool003H0oooo00<000000?ooo`3oool0U@3oool00`2oM000M;oo0?ooo`080?oo
o`030000003oool0oooo0:00oooo0P00000b0?ooo`00=@3oool00`000000oooo0?ooo`2F0?ooo`03
0?oOW01807@0_ooo00P0oooo00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`0_0?ooo`00
<`3oool2000009T0oooo00<0oooO09a8001d_ol0203oool00`000000oooo0?ooo`2S0?ooo`030000
003oool0oooo02h0oooo000b0?ooo`030000003oool0oooo09T0oooo00<0oono07@00018W=l0203o
ool00`000000oooo0?ooo`2T0?ooo`030000003oool0oooo02d0oooo000a0?ooo`030000003oool0
oooo09H0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00P0oooo00<000000?ooo`3oool0
Y@3oool00`000000oooo0?ooo`0/0?ooo`00<03oool00`000000oooo0?ooo`2H0?ooo`050?oOW018
000000000000001d_ol02@3oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo02/0oooo
000`0?ooo`030000003oool0oooo0:H0oooo00<000000?ooo`3oool0Y`3oool00`000000oooo0?oo
o`0Z0?ooo`00;`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo0:L0oooo00<00000
0?ooo`3oool0:P3oool002h0oooo00<000000?ooo`3oool0Z03oool00`000000oooo0?ooo`2X0?oo
o`030000003oool0oooo02T0oooo000^0?ooo`030000003oool0oooo0:P0oooo00<000000?ooo`3o
ool0Z@3oool00`000000oooo0?ooo`0X0?ooo`00;@3oool00`000000oooo0?ooo`2Y0?ooo`<00000
Z@3oool00`000000oooo0?ooo`0X0?ooo`00;03oool00`000000oooo0?ooo`2Z0?ooo`030000003o
ool0oooo0:X0oooo00<000000?ooo`3oool09`3oool002`0oooo00<000000?ooo`3oool0ZP3oool0
0`000000oooo0?ooo`2Z0?ooo`030000003oool0oooo02L0oooo000[0?ooo`030000003oool0oooo
0:/0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`0V0?ooo`00:`3oool00`000000
oooo0?ooo`2[0?ooo`030000003oool0oooo0:/0oooo00<000000?ooo`3oool09P3oool002X0oooo
00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`2/0?ooo`030000003oool0oooo02D0oooo
000Z0?ooo`030000003oool0oooo0:`0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?oo
o`0U0?ooo`00:@3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo0:d0oooo00<00000
0?ooo`3oool0903oool002T0oooo00<000000?ooo`3oool0[@3oool00`000000oooo0?ooo`2]0?oo
o`030000003oool0oooo02@0oooo000Y0?ooo`030000003oool0oooo0:d0oooo00<000000?ooo`3o
ool0[P3oool00`000000oooo0?ooo`0S0?ooo`00:03oool00`000000oooo0?ooo`2^0?ooo`<00000
[P3oool00`000000oooo0?ooo`0S0?ooo`00:03oool00`000000oooo0?ooo`2^0?ooo`030000003o
ool0oooo0:h0oooo00<000000?ooo`3oool08`3oool002P0oooo00<000000?ooo`3oool0[P3oool0
0`000000oooo0?ooo`2_0?ooo`030000003oool0oooo0280oooo000W0?ooo`030000003oool0oooo
0:l0oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`0R0?ooo`009`3oool00`000000
oooo0?ooo`2_0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool08P3oool002L0oooo
00<000000?ooo`3oool0[`3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo0240oooo
000W0?ooo`030000003oool0oooo0:l0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?oo
o`0Q0?ooo`009P3oool00`000000oooo0?ooo`2`0?ooo`030000003oool0oooo0;00oooo00<00000
0?ooo`3oool08@3oool002H0oooo00<000000?ooo`3oool0/03oool00`000000oooo0?ooo`2`0?oo
o`030000003oool0oooo0240oooo000V0?ooo`030000003oool0oooo0;00oooo00<000000?ooo`3o
ool0/@3oool00`000000oooo0?ooo`0P0?ooo`009P3oool00`000000oooo0?ooo`2`0?ooo`030000
003oool0oooo0;40oooo00<000000?ooo`3oool0803oool002D0oooo00<000000?ooo`3oool0/@3o
ool300000;40oooo00<000000?ooo`3oool0803oool002D0oooo00<000000?ooo`3oool0/@3oool0
0`000000oooo0?ooo`2a0?ooo`030000003oool0oooo0200oooo000U0?ooo`030000003oool0oooo
0;40oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0O0?ooo`009@3oool00`000000
oooo0?ooo`2a0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool07`3oool002D0oooo
00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`2b0?ooo`030000003oool0oooo01l0oooo
000T0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?oo
o`0O0?ooo`00903oool00`000000oooo0?ooo`2T0?ooo`030?oog`2LB000B9cO00/0oooo00<00000
0?ooo`3oool0/P3oool00`000000oooo0?ooo`0O0?ooo`00903oool00`000000oooo0?ooo`2U0?oo
o`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`0O0?oo
o`00903oool00`000000oooo0?ooo`2U0?ooo`030?oOW01804P0W=oo00X0oooo00<000000?ooo`3o
ool0/P3oool00`000000oooo0?ooo`0O0?ooo`00903oool00`000000oooo0?ooo`2Q0?ooo`030?no
M0000000000000@0000000<004RL0=ooo`3oool01`3oool00`000000oooo0?ooo`2b0?ooo`030000
003oool0oooo01l0oooo000T0?ooo`030000003oool0oooo0:40oooo00<0oooO09a8B02Lgol00P3o
ool00`2oM00007Bo0?ooo`090?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool07`3o
ool002@0oooo00<000000?ooo`3oool0XP3oool01`3ookl0M01d0;ooo`3oool0okmd0018W03Oool0
2@3oool400000;40oooo00<000000?ooo`3oool07`3oool002<0oooo00<000000?ooo`3oool0Y03o
ool01P3ogi`0B4RL0=ooo`3ogi`0B01809cOo`T0oooo00<000000?ooo`3oool0/P3oool00`000000
oooo0?ooo`0O0?ooo`008`3oool00`000000oooo0?ooo`2U0?ooo`050?noM000M;l0oooO09a80018
W=l02@3oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01h0oooo000S0?ooo`030000
003oool0oooo0:H0oooo00@0gia804RLg`2oM00007Bo2@3oool00`000000oooo0?ooo`2c0?ooo`03
0000003oool0oooo01h0oooo000S0?ooo`030000003oool0oooo0:L0oooo00@0gia807AdM00007@0
_ooo203oool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01h0oooo000S0?ooo`030000
003oool0oooo0:P0oooo00<0_g@00000B02Lgol0203oool00`000000oooo0?ooo`2c0?ooo`030000
003oool0oooo01h0oooo000S0?ooo`030000003oool0oooo0:P0oooo00<0oooO09a80018W=l0203o
ool00`000000oooo0?ooo`2c0?ooo`030000003oool0oooo01h0oooo000S0?ooo`030000003oool0
oooo0;<0oooo00<000000?ooo`3oool0/`3oool00`000000oooo0?ooo`0N0?ooo`008`3oool00`00
0000oooo0?ooo`2c0?ooo`030000003oool0oooo0;<0oooo00<000000?ooo`3oool07P3oool00280
oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo01d0
oooo000R0?ooo`030000003oool0oooo0;@0oooo0`00002d0?ooo`030000003oool0oooo01d0oooo
000R0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0]03oool00`000000oooo0?oo
o`0M0?ooo`008P3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0;@0oooo00<00000
0?ooo`3oool07@3oool00280oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2d0?oo
o`030000003oool0oooo01d0oooo000R0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3o
ool0]03oool00`000000oooo0?ooo`0M0?ooo`008P3oool00`000000oooo0?ooo`2d0?ooo`030000
003oool0oooo0;@0oooo00<000000?ooo`3oool07@3oool00280oooo00<000000?ooo`3oool0]03o
ool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo01d0oooo000R0?ooo`030000003oool0
oooo0;@0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0L0?ooo`008P3oool00`00
0000oooo0?ooo`2d0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0703oool00280
oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo01`0
oooo000R0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0]@3oool00`000000oooo
0?ooo`0L0?ooo`008P3oool00`000000oooo0?ooo`2d0?ooo`<00000]@3oool00`000000oooo0?oo
o`0L0?ooo`008@3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo0;D0oooo00<00000
0?ooo`3oool0703oool00240oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`2e0?oo
o`030000003oool0oooo01`0oooo000Q0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3o
ool0]@3oool00`000000oooo0?ooo`0L0?ooo`008@3oool00`000000oooo0?ooo`2e0?ooo`030000
003oool0oooo0;D0oooo00<000000?ooo`3oool0703oool00240oooo00<000000?ooo`3oool0]@3o
ool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo01`0oooo000Q0?ooo`030000003oool0
oooo0;D0oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`0L0?ooo`008@3oool00`00
0000oooo0?ooo`2e0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0703oool00240
oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo01`0
oooo000Q0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool0]@3oool00`000000oooo
0?ooo`0L0?ooo`008@3oool00`000000oooo0?ooo`2e0?ooo`030000003oool0oooo0;H0oooo00<0
00000?ooo`3oool06`3oool00240oooo00<000000?ooo`3oool0]@3oool300000;H0oooo00<00000
0?ooo`3oool06`3oool00240oooo00<000000?ooo`3oool0]@3oool00`000000oooo0?ooo`2f0?oo
o`030000003oool0oooo01/0oooo000Q0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3o
ool0]P3oool00`000000oooo0?ooo`0K0?ooo`008@3oool00`000000oooo0?ooo`2e0?ooo`030000
003oool0oooo0;H0oooo00<000000?ooo`3oool06`3oool00240oooo00<000000?ooo`3oool0]@3o
ool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01/0oooo000Q0?ooo`030000003oool0
oooo0;D0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`0K0?ooo`008@3oool00`00
0000oooo0?ooo`2U0?ooo`050?oog`2LB000000000000000M;l02`3oool00`000000oooo0?ooo`2f
0?ooo`030000003oool0oooo01/0oooo000Q0?ooo`030000003oool0oooo0:D0oooo00L0gia8001d
_`3oool0oono07@00000B9`0gooo00T0oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?oo
o`0K0?ooo`008@3oool00`000000oooo0?ooo`2T0?ooo`030?oog`2LB000M;oo0080oooo00<0omnL
04P00018W=l02@3oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01/0oooo000Q0?oo
o`030000003oool0oooo0:@0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`080?oo
o`030000003oool0oooo0;H0oooo00<000000?ooo`3oool06`3oool00240oooo00<000000?ooo`3o
ool0Y03oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo1000002e0?ooo`03
0000003oool0oooo01/0oooo000Q0?ooo`030000003oool0oooo0:D0oooo00<0_g@00000M02oool0
0P3oool00`3OW4P0001809cOo`080?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0
6`3oool00240oooo00<000000?ooo`3oool0Y@3oool0203o_g@0000004RLg`3oool0oooO09a80000
B9`0gooo203oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01/0oooo000Q0?ooo`03
0000003oool0oooo0:D0oooo00<0oooO09a8000000000P0000000`0007@0_ooo0?ooo`080?ooo`03
0000003oool0oooo0;H0oooo00<000000?ooo`3oool06`3oool00240oooo00<000000?ooo`3oool0
YP3oool00`3ookl0M000001d_`0<0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0
6`3oool00200oooo00<000000?ooo`3oool0Z03oool00`3ogi`0B00004RLg`0;0?ooo`030000003o
ool0oooo0;H0oooo00<000000?ooo`3oool06`3oool00200oooo00<000000?ooo`3oool0Z@3oool0
103ooml0W4P00000M02oool90?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool06`3o
ool00200oooo00<000000?ooo`3oool0Z`3oool0103ookl0M00000000018W=l70?ooo`030000003o
ool0oooo0;H0oooo00<000000?ooo`3oool06`3oool00200oooo00<000000?ooo`3oool0]P3oool0
0`000000oooo0?ooo`2f0?ooo`030000003oool0oooo01/0oooo000P0?ooo`030000003oool0oooo
0;H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000
oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00200oooo
00<000000?ooo`3oool0]P3oool300000;L0oooo00<000000?ooo`3oool06P3oool00200oooo00<0
00000?ooo`3oool0]P3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000P
0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J
0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0oooo00<000000?oo
o`3oool06P3oool00200oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`2g0?ooo`03
0000003oool0oooo01X0oooo000P0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0
]`3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003o
ool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00200oooo00<000000?ooo`3oool0]P3oool0
0`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000P0?ooo`030000003oool0oooo
0;H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000
oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00200oooo
00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo
000P0?ooo`030000003oool0oooo0;H0oooo0`00002g0?ooo`030000003oool0oooo01X0oooo000P
0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J
0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0oooo00<000000?oo
o`3oool06P3oool00200oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`2g0?ooo`03
0000003oool0oooo01X0oooo000P0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0
]`3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003o
ool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00200oooo00<000000?ooo`3oool0]P3oool0
0`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000P0?ooo`030000003oool0oooo
0;H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000
oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00200oooo
00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo
000P0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?oo
o`0J0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`<00000]`3oool00`000000oooo0?ooo`0J
0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003oool0oooo0;L0oooo00<000000?oo
o`3oool06P3oool00200oooo00<000000?ooo`3oool0]P3oool00`000000oooo0?ooo`2g0?ooo`03
0000003oool0oooo01X0oooo000P0?ooo`030000003oool0oooo0;H0oooo00<000000?ooo`3oool0
]`3oool00`000000oooo0?ooo`0J0?ooo`00803oool00`000000oooo0?ooo`2f0?ooo`030000003o
ool0oooo0;L0oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0Y`3oool0
1P3ookl0M00000000000000004RL0=ooo`X0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2V0?ooo`030?oo_`1d0000B9cO0080oooo00<0
gia80000M02oool02@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O
0?ooo`030000003oool0oooo0:H0oooo00<0okmd0018W03Oool00`3oool00`2oM000B9cO0?ooo`08
0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?oo
o`3oool0YP3oool00`3o_g@004RL0=ooo`030?ooo`030;md0000M;l0oooo00P0oooo00<000000?oo
o`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2V0?ooo`03
0?oog`2LB000M;oo0080oooo00<0oono07@00018W=l02@3oool00`000000oooo0?ooo`2g0?ooo`03
0000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0:P0oooo00H0_g@00000B02Lgml0
W4P00000M02oool90?ooo`@00000]P3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000
oooo0?ooo`2Y0?ooo`050?oog`2LB00000000018W03Oool02@3oool00`000000oooo0?ooo`2g0?oo
o`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0:T0oooo00H0_g@00000B02L
gol0omnL04Q8W03Oool80?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool0
01l0oooo00<000000?ooo`3oool0Z03oool00`3ogi`0B00007Boo`020?ooo`030?noM018W=l0oooo
00L0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000
oooo0?ooo`2X0?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03Oool01`3oool00`000000
oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0:P0oooo
00<0oooO09a80018W=l00P3oool00`2oM00007Bo0?ooo`070?ooo`030000003oool0oooo0;L0oooo
00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0Z@3oool01@3ooml0W4P00000
0000000007Bo00T0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3o
ool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3o
ool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0
oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`00
0000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0
oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0]`3oool300000;L0oooo
00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?oo
o`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<00000
0?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?oo
o`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3o
ool0]`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000
003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3o
ool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3o
ool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0
oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`00
0000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0
oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo0`00
002g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<00000
0?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?oo
o`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3o
ool0]`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000
003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3o
ool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3o
ool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0
oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`00
0000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0
oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo0`00
002g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<00000
0?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2g0?oo
o`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool001l0oooo00<000000?ooo`3o
ool0]`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo000O0?ooo`030000
003oool0oooo0;L0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`0J0?ooo`007`3o
ool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3o
ool001l0oooo00<000000?ooo`3oool0W@3oool00`3ookl0M000000000020000000304RLg`3oool0
oooo00<0oooo00@0_g@0000000000000M;oo2`3oool00`000000oooo0?ooo`2g0?ooo`030000003o
ool0oooo01X0oooo000O0?ooo`030000003oool0oooo09l0oooo00<0okmd0000001d_ol01@3oool0
1P3OW4P007Bo0?ooo`3ogi`0B00004RLg`X0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`0J0?ooo`007`3oool00`000000oooo0?ooo`2O0?ooo`030?oog`2LB00007Bo00@0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`090?ooo`030000003oool0oooo0;P0oooo00<0
00000?ooo`3oool06@3oool001l0oooo00<000000?ooo`3oool0X03oool00`3OW4P004RL0=ooo`03
0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`000000oooo0?oo
o`2h0?ooo`030000003oool0oooo01T0oooo000O0?ooo`030000003oool0oooo0:00oooo00<0omnL
04P0B02Lgol00`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<00000
0?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2P0?oo
o`030?oo_`1d0000M;oo00@0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`080?oo
o`@00000]`3oool00`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2Q0?ooo`03
0;md0000M;l0oooo00<0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`080?ooo`03
0000003oool0oooo0;P0oooo00<000000?ooo`3oool06@3oool001l0oooo00<000000?ooo`3oool0
X@3oool00`3OW4P004RL0=ooo`030?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool0
1`3oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01T0oooo000O0?ooo`030000003o
ool0oooo0:40oooo00<0omnL04P0B02Lgol00`3oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo00L0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`007`3o
ool00`000000oooo0?ooo`2Q0?ooo`030?oo_`1d0000M;oo00@0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`070?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool06@3o
ool001l0oooo00<000000?ooo`3oool0XP3oool00`2oM00007Bo0?ooo`040?ooo`060=nLB00007@0
_ooo0?oog`2LB000B9cO203oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01T0oooo
000O0?ooo`030000003oool0oooo0:00oooo00D0oono07@000000000001d0;ooo`050?ooo`040=nL
B00000000000001d_`T0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`00
7`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0
6@3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`2h0?ooo`030000003o
ool0oooo01T0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0^03oool0
0`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo
0;P0oooo00<000000?ooo`3oool06@3oool001l0oooo00<000000?ooo`3oool0]`3oool300000;P0
oooo00<000000?ooo`3oool06@3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`2h0?ooo`030000003oool0oooo01T0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<0
00000?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2g
0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool06@3oool001l0oooo00<000000?oo
o`3oool0]`3oool00`000000oooo0?ooo`2h0?ooo`030000003oool0oooo01T0oooo000O0?ooo`03
0000003oool0oooo0;L0oooo00<000000?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`00
7`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo0;P0oooo00<000000?ooo`3oool0
6@3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo0?ooo`2h0?ooo`030000003o
ool0oooo01T0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0^03oool0
0`000000oooo0?ooo`0I0?ooo`007`3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo
0;P0oooo00<000000?ooo`3oool06@3oool001l0oooo00<000000?ooo`3oool0]`3oool300000;P0
oooo00<000000?ooo`3oool06@3oool001l0oooo00<000000?ooo`3oool0]`3oool00`000000oooo
0?ooo`2h0?ooo`030000003oool0oooo01T0oooo000O0?ooo`030000003oool0oooo0;L0oooo00<0
00000?ooo`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.16236, -1.32956, \
0.00800441, 0.0701008}}]
}, Closed]],

Cell[TextData[{
  "succeeds because the function is evaluated first and is ready to receive \
numerical values and produce numerical results for plotting.  Using ",
  StyleBox["Evaluate", "InlineInput",
    FontWeight->"Bold"],
  " will correct most errors that say \"not a machine-size real number\".  If \
not, make sure that substitution of a number into your function does produce \
a valid numerical result \[LongDash] often one forgets to substitute \
numerical values for parameters in a function and symbols appear in the \
output."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Data", "Subsection"],

Cell[TextData[{
  "Next, suppose that we wish to plot a set of data ",
  Cell[BoxData[
      \(TraditionalForm\`{{x\_i, y\_i}, \[CenterEllipsis]}\)]],
  ".  First we create such a set by modifying a previous cell.  The comments \
were removed and ",
  StyleBox["Transpose", "InlineInput",
    FontWeight->"Bold"],
  " was used to package the data as a list of pairs.  Also, now we use the \
Gaussian probability distribution that came with the statistics package to \
add a more normal set of fluctuations.  The parameters are the mean, here \
zero, and the standard deviation, \[Sigma], for the fluctuations.  The plot \
is made with ",
  ButtonBox["ListPlot",
    ButtonStyle->"RefGuideLink"],
  " and an option is used to make the points large enough for aging eyes."
}], "Text"],

Cell[BoxData[{
    \(\(f[x_] = \(x\^2\) Sin[x];\)\), "\[IndentingNewLine]", 
    \(\(xdata = Range[\(-0.9\), 0.9, 0.1];\)\), "\[IndentingNewLine]", 
    \(\(ydata = 
        f[xdata] + 
          Table[Random[
              NormalDistribution[0. , 0.1]], {Length[
                xdata]}];\)\), "\[IndentingNewLine]", 
    \(\(data = Transpose[{xdata, ydata}];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ListPlot[data, PlotStyle \[Rule] PointSize[0.02]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.529101 0.28942 0.405596 [
[.10317 .27692 -16.8438 -15.6875 ]
[.10317 .27692 16.8438 0 ]
[.23545 .27692 -13.8438 -15.6875 ]
[.23545 .27692 13.8438 0 ]
[.36772 .27692 -16.8438 -15.6875 ]
[.36772 .27692 16.8438 0 ]
[.63228 .27692 -12.5 -15.6875 ]
[.63228 .27692 12.5 0 ]
[.76455 .27692 -9.5 -15.6875 ]
[.76455 .27692 9.5 0 ]
[.89683 .27692 -12.5 -15.6875 ]
[.89683 .27692 12.5 0 ]
[.4875 .04606 -27.9375 -7.84375 ]
[.4875 .04606 0 7.84375 ]
[.4875 .12718 -27.6875 -7.84375 ]
[.4875 .12718 0 7.84375 ]
[.4875 .2083 -27.6875 -7.84375 ]
[.4875 .2083 0 7.84375 ]
[.4875 .37054 -19 -7.84375 ]
[.4875 .37054 0 7.84375 ]
[.4875 .45166 -19 -7.84375 ]
[.4875 .45166 0 7.84375 ]
[.4875 .53278 -19.25 -7.84375 ]
[.4875 .53278 0 7.84375 ]
[.4875 .6139 -19 -7.84375 ]
[.4875 .6139 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.10317 .28942 m
.10317 .29567 L
s
gsave
.10317 .27692 -77.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.23545 .28942 m
.23545 .29567 L
s
gsave
.23545 .27692 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.36772 .28942 m
.36772 .29567 L
s
gsave
.36772 .27692 -77.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.63228 .28942 m
.63228 .29567 L
s
gsave
.63228 .27692 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.76455 .28942 m
.76455 .29567 L
s
gsave
.76455 .27692 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.89683 .28942 m
.89683 .29567 L
s
gsave
.89683 .27692 -73.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.125 Mabswid
.12963 .28942 m
.12963 .29317 L
s
.15608 .28942 m
.15608 .29317 L
s
.18254 .28942 m
.18254 .29317 L
s
.20899 .28942 m
.20899 .29317 L
s
.2619 .28942 m
.2619 .29317 L
s
.28836 .28942 m
.28836 .29317 L
s
.31481 .28942 m
.31481 .29317 L
s
.34127 .28942 m
.34127 .29317 L
s
.39418 .28942 m
.39418 .29317 L
s
.42063 .28942 m
.42063 .29317 L
s
.44709 .28942 m
.44709 .29317 L
s
.47354 .28942 m
.47354 .29317 L
s
.52646 .28942 m
.52646 .29317 L
s
.55291 .28942 m
.55291 .29317 L
s
.57937 .28942 m
.57937 .29317 L
s
.60582 .28942 m
.60582 .29317 L
s
.65873 .28942 m
.65873 .29317 L
s
.68519 .28942 m
.68519 .29317 L
s
.71164 .28942 m
.71164 .29317 L
s
.7381 .28942 m
.7381 .29317 L
s
.79101 .28942 m
.79101 .29317 L
s
.81746 .28942 m
.81746 .29317 L
s
.84392 .28942 m
.84392 .29317 L
s
.87037 .28942 m
.87037 .29317 L
s
.07672 .28942 m
.07672 .29317 L
s
.05026 .28942 m
.05026 .29317 L
s
.02381 .28942 m
.02381 .29317 L
s
.92328 .28942 m
.92328 .29317 L
s
.94974 .28942 m
.94974 .29317 L
s
.97619 .28942 m
.97619 .29317 L
s
.25 Mabswid
0 .28942 m
1 .28942 L
s
.5 .04606 m
.50625 .04606 L
s
gsave
.4875 .04606 -88.9375 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.5 .12718 m
.50625 .12718 L
s
gsave
.4875 .12718 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.5 .2083 m
.50625 .2083 L
s
gsave
.4875 .2083 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.5 .37054 m
.50625 .37054 L
s
gsave
.4875 .37054 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
.5 .45166 m
.50625 .45166 L
s
gsave
.4875 .45166 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
.5 .53278 m
.50625 .53278 L
s
gsave
.4875 .53278 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
.5 .6139 m
.50625 .6139 L
s
gsave
.4875 .6139 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .06634 m
.50375 .06634 L
s
.5 .08662 m
.50375 .08662 L
s
.5 .1069 m
.50375 .1069 L
s
.5 .14746 m
.50375 .14746 L
s
.5 .16774 m
.50375 .16774 L
s
.5 .18802 m
.50375 .18802 L
s
.5 .22858 m
.50375 .22858 L
s
.5 .24886 m
.50375 .24886 L
s
.5 .26914 m
.50375 .26914 L
s
.5 .3097 m
.50375 .3097 L
s
.5 .32998 m
.50375 .32998 L
s
.5 .35026 m
.50375 .35026 L
s
.5 .39082 m
.50375 .39082 L
s
.5 .4111 m
.50375 .4111 L
s
.5 .43138 m
.50375 .43138 L
s
.5 .47194 m
.50375 .47194 L
s
.5 .49222 m
.50375 .49222 L
s
.5 .5125 m
.50375 .5125 L
s
.5 .55306 m
.50375 .55306 L
s
.5 .57334 m
.50375 .57334 L
s
.5 .59362 m
.50375 .59362 L
s
.5 .02578 m
.50375 .02578 L
s
.5 .0055 m
.50375 .0055 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.02 w
.02381 .01472 Mdot
.07672 .12773 Mdot
.12963 .17531 Mdot
.18254 .19696 Mdot
.23545 .24709 Mdot
.28836 .29619 Mdot
.34127 .24671 Mdot
.39418 .25728 Mdot
.44709 .29764 Mdot
.5 .2911 Mdot
.55291 .31105 Mdot
.60582 .22102 Mdot
.65873 .22652 Mdot
.71164 .2813 Mdot
.76455 .33629 Mdot
.81746 .40404 Mdot
.87037 .40353 Mdot
.92328 .54102 Mdot
.97619 .60332 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oe`3oool0
0`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`004@3oool200000<@0
oooo0`00003F0?ooo`003`3oool600000<80oooo00<000000?ooo`3oool0eP3oool000l0oooo1P00
00320?ooo`030000003oool0oooo0=H0oooo000>0?ooo`P00000`@3oool00`000000oooo0?ooo`3F
0?ooo`003P3oool800000<40oooo00<000000?ooo`3oool0eP3oool000l0oooo1P0000320?ooo`03
0000003oool0oooo0=H0oooo000?0?ooo`H00000`P3oool00`000000oooo0?ooo`3F0?ooo`004@3o
ool200000<@0oooo00<000000?ooo`3oool0eP3oool00=L0oooo0`00003F0?ooo`00e`3oool00`00
0000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo
0?ooo`3F0?ooo`00^P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo00<0oooo
00D0oooO09a8000000000000001d_`0;0?ooo`030000003oool0oooo0=H0oooo002i0?ooo`060=nL
B000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`030?ooo`070=nLB000M;l0oooo
0?oo_`1d000004RL0=ooo`090?ooo`030000003oool0oooo0=H0oooo002h0?ooo`030?oog`2LB000
B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3ooml0W4P007Boo`020?ooo`030?oOW0180000
B9cO00T0oooo00<000000?ooo`3oool0eP3oool00;P0oooo00<0oooO09a80018W=l00P3oool0103o
oml0W4P00018W03Oool50?ooo`030?oo_`1d0000M;oo00<0oooo00<0_g@00000M02oool0203oool0
0`000000oooo0?ooo`3F0?ooo`00Z`3oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo
0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0
gia80000B02Lgol0203oool00`000000oooo0?ooo`3F0?ooo`00^@3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00H0oooo00<0_g@00000M02oool00P3oool00`3OW4P0001809cOo`08
0?ooo`@00000e@3oool00;T0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`060?oo
o`080?noM0000000B9cO0?ooo`3ooml0W4P00018W03Oool80?ooo`030000003oool0oooo0=H0oooo
002i0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3oool00`3ooml0W4P00000
0002000000030000M02oool0oooo00P0oooo00<000000?ooo`3oool0eP3oool00;T0oooo00@0oooO
09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oo_`1d000007Bo00`0oooo00<0
00000?ooo`3oool0eP3oool00;X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`07
0?ooo`030?oOW0180000B9cO00/0oooo00<000000?ooo`3oool0eP3oool00;/0oooo00H0gia80000
M02oool0oooO09a80018W=l90?ooo`040?oog`2LB000001d0;ooo`T0oooo00<000000?ooo`3oool0
eP3oool00;`0oooo00@0gia800000000000007Bo303oool0103ookl0M00000000018W=l70?ooo`03
0000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`<00000eP3o
ool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0
oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<0
00000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?oo
o`3oool0eP3oool00=L0oooo0`00003F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3o
ool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`00
0000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo
0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F
0?ooo`00e`3oool300000=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000
003oool0oooo0=H0oooo002j0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol0
1@3oool00`3ooml0W4P004RLg`0;0?ooo`030000003oool0oooo0=H0oooo002i0?ooo`060=nLB000
M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`060?ooo`030=nLB000B9`0gooo00X0
oooo00<000000?ooo`3oool0eP3oool002L0oooo0P00002?0?ooo`030?oog`2LB000B9cO0080oooo
00<0okmd0000B02Lgol02P3oool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0oooo0=H0oooo
000U0?ooo`H00000S@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo
00<0okmd000000000000100000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo0=H0oooo
000U0?ooo`H00000P03oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB000
07Bo0`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3o
ool02@3oool00`000000oooo0?ooo`3F0?ooo`00903oool8000008d0oooo00<0_g@00018W03Oool0
0P3oool00`3ogi`0B00004RLg`060?ooo`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`090?oo
o`@00000e@3oool002@0oooo2000002=0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000
M;l01`3oool01P3ogi`0B4RL0=ooo`3ogi`0B01809cOo`T0oooo00<000000?ooo`3oool0eP3oool0
02D0oooo1P00002>0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01`3oool01@3o
_g@007Bo0?oog`2LB000B9cO00T0oooo00<000000?ooo`3oool0eP3oool002D0oooo1P00002>0?oo
o`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0203oool0103OW4P0B9cO0;md0000
M;l90?ooo`030000003oool0oooo0=H0oooo000W0?ooo`800000T@3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo00T0oooo00@0gia807AdM00007@0_ooo203oool00`000000oooo0?oo
o`3F0?ooo`00^`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`/0oooo00<0_g@00000B02Lgol0
203oool00`000000oooo0?ooo`3F0?ooo`00_03oool0103OW4P0000000000000M;l<0?ooo`030?oo
g`2LB000B9cO00P0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3o
ool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo0`00003F0?ooo`00e`3oool00`000000
oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?oo
o`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?oo
o`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00?@3o
ool2000009P0oooo0`00003F0?ooo`00>`3oool6000009H0oooo00<000000?ooo`3oool0eP3oool0
03/0oooo1P00002F0?ooo`030000003oool0oooo0=H0oooo000j0?ooo`P00000U@3oool00`000000
oooo0?ooo`3F0?ooo`00>P3oool8000009D0oooo00<000000?ooo`3oool0eP3oool003/0oooo1P00
002F0?ooo`030000003oool0oooo0=H0oooo000k0?ooo`H00000UP3oool00`000000oooo0?ooo`3F
0?ooo`00?@3oool2000009P0oooo00<000000?ooo`3oool0eP3oool00=L0oooo0`00003F0?ooo`00
D`3oool200000880oooo00<000000?ooo`3oool0eP3oool00540oooo1P0000200?ooo`030000003o
ool0oooo0=H0oooo001A0?ooo`H00000P03oool00`000000oooo0?ooo`3F0?ooo`00D03oool80000
0680oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB0000000
000000<0000000<0B9cO0?ooo`3oool0203oool00`000000oooo0?ooo`3F0?ooo`00D03oool80000
0640oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0oooo00<0
okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`080?ooo`030000003oool0oooo0=H0oooo001A
0?ooo`H00000H@3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0okmd
001d_`3oool0303oool00`000000oooo0?ooo`3F0?ooo`00D@3oool600000640oooo00<0oooO09a8
0018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?noM000M;l0oooo00/0oooo00<00000
0?ooo`3oool0eP3oool005<0oooo0P00001F0?ooo`030?oog`2LB000000000L0000000H0B9cO0?oo
o`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00T0oooo00<0okmd04RLg`3oool02P3o
ool00`000000oooo0?ooo`3F0?ooo`00^@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000
B9cO00X0oooo00<0gia804RLg`3oool02@3oool400000=D0oooo002i0?ooo`030=nLB00004P0W=oo
0080oooo00<0oooO09a80000M;l02`3oool00`2oM000M;oo0?ooo`080?ooo`030000003oool0oooo
0=H0oooo002i0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool02P3oool00`3ogi`0
B01d0;ooo`080?ooo`030000003oool0oooo02P0oooo0P00002/0?ooo`00^@3oool0103ooml0W4P0
0018W03Oool20?ooo`030=nLB000B9`0gooo00X0oooo00<0oooO09a8001d_ol0203oool00`000000
oooo0?ooo`0V0?ooo`H00000ZP3oool00;X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo
0?ooo`0:0?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool09P3oool600000180oooo
0P00002F0?ooo`00^`3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00L0okmd0018W03O
ool0oooo0=nLB00004P0W=oo00P0oooo00<000000?ooo`3oool09@3oool8000000l0oooo1P00002D
0?ooo`00_03oool0103OW4P0000000000000M;l90?ooo`050?oOW018000000000000001d_ol02@3o
ool00`000000oooo0?ooo`0U0?ooo`P000003`3oool6000009@0oooo003G0?ooo`030000003oool0
oooo02H0oooo1P00000?0?ooo`P00000T`3oool00=L0oooo0`00000V0?ooo`H000003`3oool80000
09<0oooo003G0?ooo`030000003oool0oooo02P0oooo0P00000B0?ooo`H00000U03oool00=L0oooo
00<000000?ooo`3oool0?03oool6000009@0oooo003G0?ooo`030000003oool0oooo03h0oooo0P00
002F0?ooo`00;03oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo00<0oooo00<0
oono07A8W03Oool01@3oool01P3ogi`0B00000000000000004RL0=ooob00oooo00@0_g@000000000
0000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000B9`0gooo703o
ool200000080oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`030=nL
B0000000000000<0000000D0B9cO0?ooo`3oool0omnL04P00002000000030018W03Oool0oooo0280
oooo00<000000?ooo`3oool09P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo
0080oooo00<0gia80000000000000`0000001@18W=l0oooo0?ooo`3ogi`0B0000080000000<004RL
0=ooo`3oool07`3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00H0
omnL04P00000000000000018W03OoolQ0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0
001d_ol00`3oool00`3ookl0M4RL0=ooo`050?ooo`060?oOW0180000000000000000B9`0gooo9`3o
ool002/0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00@0oooo
00<0gia804RLg`3oool0103oool01P3o_g@0000007Boo`3oool0omnL04Qd_al0oooo00<0gia8001d
_`3oool00P0000001P18W=l0oooo0?ooo`3ogi`0B00007Boo`80oooo00H0okmd0000001d_ol0oooo
0?oOW018M;lJ0?ooo`H0000000D007Bo0?ooo`3ogi`0B00004RLg`020?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd001d_`3oool00P3oool02@2oM7@0_ooo0?ooo`3o_g@0000007Boo`3oool0omnL
04Qd_`0S0?ooo`030000003oool0oooo02D0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?oo
o`030?oOW0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool02@2oM7@0_ooo0?ooo`3o_g@0
000007Boo`3oool0omnL04Qd_`0O0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3o
gi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo803oool01P3OW4P007Bo0?oo
o`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0103oool00`3OW4P0B9cO0?ooo`040?ooo`06
0?noM0000000M;oo0?ooo`3ogi`0B7Bo9`3oool002X0oooo00<0oooO09a80018W=l00P3oool00`3o
_g@0001809cOo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<0okmd04RLg`3oool0703oool00`3o
oml0W4P000000005000000/0oooo00<0okmd04RLg`3oool0603oool60000000604RLg`3oool0oooo
0?noM00004P0W=oo203oool00`3o_g@007Bo0?ooo`090?ooo`030?noM018W=l0oooo0240oooo00<0
00000?ooo`3oool0903oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0
okmd001d_`3oool02@3oool00`3o_g@0B9cO0?ooo`0L0?ooo`030?oog`2LB000B9cO0080oooo00<0
okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`0M0?ooo`030?oog`2LB000B9cO0080oooo00<0
okmd0000B02Lgol0203oool00`3ookl0M01d0;ooo`090?ooo`030?noM018W=l0oooo02D0oooo000Z
0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo0?oo
o`080?ooo`030?oo_`1d07@0_ooo01`0oooo00<0oooO09a8000000001@0000000`3Oool0oooo0?oo
o`080?ooo`030?oo_`1d07@0_ooo01L0oooo200000020?ooo`040?oog`2LB00004RL0=ooo`P0oooo
00<0okmd001d_`3oool0203oool00`3ookl0M01d0;ooo`0Q0?ooo`030000003oool0oooo02@0oooo
00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?noM000M;l0oooo00P0
oooo00<0oono07@0M02oool0703oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=oo
o`X0oooo00<0oono07@0M02oool07@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL
0=ooo`P0oooo00<0okmd001d_`3oool0203oool00`3ookl0M01d0;ooo`0U0?ooo`007@3oool00`3o
oml0W4P0000000070000000604RLg`3oool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cO
o`080?ooo`030?oog`2LB4P0W=oo00P0oooo00<0oooO09a8001d_ol03`3oool00`3ooml0W4P00000
00070000000404RLg`3oool0oooo0?oog`P000002P3oool00`3ooml0W4P007Boo`0?0?ooo`030?oo
g`2LB000000000d000000`3oool00`3OW4P0001809cOo`090?ooo`030?noM018W=l0oooo00L0oooo
00<0oooO09a8001d_ol08@3oool00`000000oooo0?ooo`0T0?ooo`030?oog`2LB00007Bo00<0oooo
00<0gia80000B02Lgol02@3oool00`3o_g@0B9cO0?ooo`070?ooo`030?oog`2LB000M;oo01`0oooo
00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo01d0oooo
00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`080?ooo`030?oog`2LB4P0W=oo00P0oooo
00<0oooO09a8001d_ol09@3oool002/0oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RL
g`090?ooo`030?oOW018M;l0oooo00L0oooo00<0oono07@00018W=l07@3oool8000000X0oooo00<0
oono07@00018W=l0603oool6000000070018W03Oool0oooo0?ooo`3ogi`0B00004RLg`0:0?ooo`03
0=nLB0000000000000H0oooo00<0oono07@00018W=l08@3oool3000002D0oooo00<0_g@00018W03O
ool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030=nLB018W=l0oooo00H0oooo00<0oono07@00018
W=l07@3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo00<0oono07@00018
W=l07P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00T0oooo00<0omnL04Qd_`3o
ool01`3oool00`3ookl0M00004RLg`0U0?ooo`00:`3oool00`3OW4P0001809cOo`020?ooo`030?oo
g`2LB00007Bo00X0oooo00<0_g@007Boo`3oool01P3oool00`3o_g@0000007Boo`0M0?ooo`030=nL
B0000000000000@0000000<007Bo0?ooo`3oool0203oool00`3o_g@0000007Boo`0H0?ooo`H00000
00L0001809cOo`3oool0oooo0?oog`2LB00007Bo00T0oooo1P0000040?ooo`030?noM0000000M;oo
0240oooo00<000000?ooo`3oool09@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo
00/0oooo00<0_g@007Boo`3oool01@3oool00`3o_g@0000007Boo`0M0?ooo`030=nLB00004P0W=oo
0080oooo00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`0N0?ooo`030=nLB00004P0W=oo
0080oooo00<0oooO09a80000M;l02P3oool00`2oM000M;oo0?ooo`060?ooo`030?noM0000000M;oo
02D0oooo000[0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool02@3oool00`3ogi`0
B4RL0=ooo`030?ooo`060?oOg`3Oool0okmd0000000007@0_ooo7@3oool00`3ogi`0000000000004
000000030018W03Oool0oooo00D0oooo00H0omoO0=ooo`3o_g@000000000M02ooolJ0?ooo`800000
00@0oooo0?oOW0180000B9cO0`3oool00`2oM00004RL0=ooo`080?ooo`H0000000L0oooo0?oOg`3O
ool0okmd0000000007@0_ooo0240oooo00<000000?ooo`3oool09@3oool00`3ogi`0B00004RLg`03
0?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02oool00P3oool01P3ogml0gooo0?noM000
0000001d0;oooad0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`060?oO
g`3Oool0okmd0000000007@0_ooo7P3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo
00T0oooo00<0omnL04Q8W03Oool00`3oool01P3ogml0gooo0?noM0000000001d0;ooobD0oooo000[
0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3OW4P0B9cO0?oo
o`020?ooo`050?oo_`1d000000000000B02Lgol07P3oool00`3ooml0W4P00018W002000000040?oo
o`3OW4P004RL0=ooo`H0oooo00D0oono07@000000000001809cOo`0N0?ooo`040?oog`2LB00004RL
0=ooo`80oooo00<0gia80018W03Oool01`3oool8000000050?oo_`1d000000000000B02Lgol08P3o
ool00`000000oooo0?ooo`0U0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0
2P3oool00`3ooml0W4P007Boo`020?ooo`050?oo_`1d000000000000B02Lgol07P3oool0103ooml0
W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000001809cOo`0O
0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3OW4P0B9cO0?oo
o`020?ooo`050?oo_`1d000000000000B02Lgol09P3oool002`0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`050?ooo`030?oog`3Ogol0oooo0080oooo00<0oono07@0M02oool00`3o
ool00`3OW9`0_ooo0?ooo`0P0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3o
ool00`3OW9`0_ooo0?ooo`0P0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3o
ool8000000030?ooo`3OW9`0_ooo02@0oooo00<000000?ooo`3oool09P3oool00`3ogi`0B00007Bo
o`020?ooo`030=nLB000M;l0oooo00X0oooo00<0oono07@00018W=l00`3oool00`3OW9`0_ooo0?oo
o`0P0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0_ooo0?oo
o`0Q0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01@3oool00`3ooml0gmoo0?oo
o`020?ooo`030?oo_`1d07@0_ooo00<0oooo00<0gibL0;ooo`3oool09`3oool002d0oooo00H0gia8
0000M02oool0oooO09a80018W=l70?ooo`030;mdM02oool0oooo0080oooo00<0okmd001d_`3oool0
0P3oool00`3ogi`0M9cO0?ooo`0Q0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`3o
gi`0M9cO0?ooo`0Q0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`3o_g@004RL0000
000500000080oooo00<0omnL07BLg`3oool08`3oool00`000000oooo0?ooo`0W0?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO1`3oool01`3o_g@004RL0=ooo`3oool0gia80000B02Lgol00`3oool0
0`3ogi`0M9cO0?ooo`0Q0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO
0?ooo`0R0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool00`2oM7@0_ooo0?ooo`020?oo
o`030?noM000M;l0oooo0080oooo00<0omnL07BLg`3oool09`3oool002h0oooo00@0gia800000000
000007Bo203oool00`3ookl0M00000000003000000030000B02Lgol0oooo0080oooo00D0_g@00000
00000000000004RLg`0O0?ooo`040=nLB00000000000001d_`X0oooo00D0_g@00000000000000000
04RLg`0O0?ooo`040=nLB00000000000001d_`T0oooo00<0omnL000000000000100000030?ooo`05
0;md00000000000000000018W=l0803oool00`000000oooo0?ooo`0X0?ooo`040=nLB00000000000
001d_`T0oooo00D0omnL04P000000000000007Boo`050?ooo`050;md00000000000000000018W=l0
7`3oool0103OW4P0000000000000M;l:0?ooo`050;md00000000000000000018W=l0803oool0103O
W4P0000000000000M;l80?ooo`030?oo_`1d0000000000<0000000<0001809cOo`3oool00P3oool0
1@2oM0000000000000000000B9cO02@0oooo002[0?ooo`800000:P3oool00`000000oooo0?ooo`3F
0?ooo`00e`3oool300000=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000
003oool0oooo05@0oooo0P0000200?ooo`00e`3oool00`000000oooo0?ooo`1B0?ooo`H00000OP3o
ool00=L0oooo00<000000?ooo`3oool0DP3oool6000007h0oooo003G0?ooo`030000003oool0oooo
0540oooo2000001m0?ooo`00eP3oool2000005<0oooo2000001m0?ooo`00e03oool600000580oooo
1P00001n0?ooo`00203ooooo00000:0000002@3oool00180oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool6
000000L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`H00000203oool0
0`000000oooo0?ooo`040?ooo`P000001`3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`8000002P3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool0
0`000000oooo0?ooo`0A0?ooo`004P3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo
00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`H000001`3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00D0oooo1P0000080?ooo`030000003oool0oooo
00@0oooo200000070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0203oool00`00
0000oooo0?ooo`0A0?ooo`00<`3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo0100
oooo2000000L0?ooo`030000003oool0oooo01X0oooo2000000?0?ooo`H00000=03oool00`000000
oooo0?ooo`0d0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0<P3oool007`0oooo
2000000i0?ooo`P000003`3oool600000=H0oooo001m0?ooo`H00000>`3oool600000180oooo0P00
003H0?ooo`00O@3oool6000003/0oooo1P00000C0?ooo`030000003oool0oooo0180oooo0P000032
0?ooo`00O`3oool2000003l0oooo0P00000E0?ooo`030000003oool0oooo0100oooo1P0000300?oo
o`00e`3oool00`000000oooo0?ooo`0@0?ooo`H00000`03oool00=L0oooo0`00000?0?ooo`P00000
_`3oool00=L0oooo00<000000?ooo`3oool03`3oool800000;l0oooo003G0?ooo`030000003oool0
oooo0100oooo1P0000300?ooo`00e`3oool00`000000oooo0?ooo`0@0?ooo`H00000`03oool00=L0
oooo00<000000?ooo`3oool04P3oool200000<80oooo003G0?ooo`030000003oool0oooo0=H0oooo
003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo06X0oooo0P00001Z
0?ooo`00e`3oool3000006P0oooo1P00001X0?ooo`00e`3oool00`000000oooo0?ooo`1X0?ooo`H0
0000J03oool00=L0oooo00<000000?ooo`3oool0I`3oool8000006L0oooo003G0?ooo`030000003o
ool0oooo06L0oooo2000001W0?ooo`00e`3oool00`000000oooo0?ooo`1X0?ooo`H00000J03oool0
0=L0oooo00<000000?ooo`3oool0J03oool6000006P0oooo003G0?ooo`030000003oool0oooo06X0
oooo0P00001Z0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?oo
o`3F0?ooo`00e`3oool300000=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`03
0000003oool0oooo0=H0oooo002j0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P00000000000030000000304RLg`3oool0oooo00P0oooo00<000000?ooo`3o
ool0eP3oool00;T0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`080?ooo`030000003oool0oooo
0=H0oooo002h0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3o_g@0
07Bo0?ooo`0<0?ooo`030000003oool0oooo0=H0oooo002h0?ooo`030?oog`2LB000B9cO0080oooo
00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo0=H0
oooo002h0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02@3oool00`3o_g@0B9cO
0?ooo`0:0?ooo`030000003oool0oooo0=H0oooo002i0?ooo`030;md0000B9`0gooo0080oooo00<0
omnL04P00018W=l02P3oool00`3OW4P0B9cO0?ooo`090?ooo`@00000e@3oool00;T0oooo00<0gia8
0000B02Lgol00P3oool00`3ooml0W4P0001d_`0;0?ooo`030;md001d_ol0oooo00P0oooo00<00000
0?ooo`3oool0eP3oool00;T0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`0:0?oo
o`030?oOW01807@0_ooo00P0oooo00<000000?ooo`3oool0eP3oool00;T0oooo00@0oooO09a80000
B9`0gooo0P3oool00`3OW4P004RL0=ooo`0:0?ooo`030?oog`2LB000M;oo00P0oooo00<000000?oo
o`3oool0eP3oool00;X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`0:0?ooo`03
0?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0eP3oool00;/0oooo00H0gia80000M02oool0
oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`080?ooo`030000003o
ool0oooo0=H0oooo002l0?ooo`040=nLB00000000000001d_`T0oooo00D0omnL04P0000000000000
07Boo`090?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G
0?ooo`<00000eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3o
ool0P03oool2000001@0oooo0P00000n0?ooo`00e`3oool00`000000oooo0?ooo`1n0?ooo`H00000
403oool6000003`0oooo003G0?ooo`030000003oool0oooo07h0oooo1P00000@0?ooo`H00000?03o
ool00=L0oooo00<000000?ooo`3oool0O@3oool8000000h0oooo2000000k0?ooo`00e`3oool00`00
0000oooo0?ooo`1m0?ooo`P000003P3oool8000003/0oooo003G0?ooo`030000003oool0oooo07h0
oooo1P00000@0?ooo`H00000?03oool00=L0oooo00<000000?ooo`3oool0OP3oool600000100oooo
1P00000l0?ooo`00e`3oool300000800oooo0P00000D0?ooo`800000?P3oool00=L0oooo00<00000
0?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3o
ool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3o
ool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0
oooo0`00003F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?oo
o`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00^P3oool0102oM00000000000001d_ol3
0?ooo`030?oOW0180000M;oo00D0oooo00<0oooO09a80018W=l02`3oool00`000000oooo0?ooo`3F
0?ooo`00^@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol01P3o
ool00`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo0=H0oooo002h0?ooo`030?oog`2LB000
B9cO0080oooo00<0okmd0000B02Lgol02P3oool00`3ogi`0B01809cOo`0:0?ooo`030000003oool0
oooo0=H0oooo002h0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool0
0`3o_g@0000000000004000000030018W03Oool0oooo00L0oooo00<000000?ooo`3oool0eP3oool0
0;P0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB4P0W=oo
0080oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`3F0?ooo`00^@3oool00`2oM000
04RL0=ooo`020?ooo`030?oOW0180000B9cO00H0oooo00L0oono07@0M02oool0oooo0?noM000B9`0
gooo00T0oooo1000003E0?ooo`00^@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo
00L0oooo00H0omnL04Q8W03Oool0omnL04P0B02Lgol90?ooo`030000003oool0oooo0=H0oooo002i
0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01`3oool01@3o_g@007Bo0?oog`2L
B000B9cO00T0oooo00<000000?ooo`3oool0eP3oool00;T0oooo00@0oooO09a80000B9`0gooo0P3o
ool00`3OW4P004RL0=ooo`080?ooo`040=nLB018W=l0_g@0001d_`T0oooo00<000000?ooo`3oool0
eP3oool00;X0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`090?ooo`040=nLB01d
M7@0001d0;ooo`P0oooo00<000000?ooo`3oool0eP3oool00;/0oooo00H0gia80000M02oool0oooO
09a80018W=l;0?ooo`030;md000004P0W=oo00P0oooo00<000000?ooo`3oool0eP3oool00;`0oooo
00@0gia800000000000007Bo303oool00`3ooml0W4P004RLg`080?ooo`030000003oool0oooo0=H0
oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`<00000eP3oool00=L0oooo00<00000
0?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3o
ool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3o
ool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0
oooo0`00003F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?oo
o`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?oo
o`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3o
ool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool30000
0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`030000003oool0oooo0=H0oooo
002j0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00`3oool01@3ooml0W4P0
00000000000007Bo00/0oooo00<000000?ooo`3oool0eP3oool00;T0oooo00H0gia8001d_`3oool0
omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0oooo00L0gia8001d_`3oool0oono07@00000
B9`0gooo00T0oooo00<000000?ooo`3oool0eP3oool00;P0oooo00<0oooO09a80018W=l00P3oool0
0`3o_g@0001809cOo`060?ooo`030?oog`2LB000M;oo0080oooo00<0omnL04P00018W=l02@3oool0
0`000000oooo0?ooo`3F0?ooo`00^03oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL
0=ooo`D0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d0;ooo`080?ooo`030000003oool0
oooo0=H0oooo002h0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol01@3oool00`3o
oml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<000000?ooo`3oool0eP3oool00;T0
oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`060?ooo`030;md000007@0_ooo0080
oooo00<0gia80000B02Lgol0203oool400000:/0oooo0P00000X0?ooo`00^@3oool00`3OW4P00018
09cOo`020?ooo`030?oog`2LB00007Bo00H0oooo00P0okmd00000018W=l0oooo0?oog`2LB00004RL
0=ooo`P0oooo00<000000?ooo`3oool0ZP3oool6000002H0oooo002i0?ooo`030?oOW0180000B9cO
00<0oooo00<0_g@00018W03Oool01@3oool00`3ooml0W4P000000002000000030000M02oool0oooo
00P0oooo00<000000?ooo`3oool0ZP3oool6000002H0oooo002i0?ooo`040?oog`2LB00004RL0=oo
o`80oooo00<0gia80018W03Oool01P3oool00`3ookl0M000001d_`0<0?ooo`030000003oool0oooo
0:T0oooo2000000U0?ooo`00^P3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0
oooo00<0omnL04P00018W=l02`3oool00`000000oooo0?ooo`2Y0?ooo`P000009@3oool00;/0oooo
00H0gia80000M02oool0oooO09a80018W=l90?ooo`040?oog`2LB000001d0;ooo`T0oooo00<00000
0?ooo`3oool0ZP3oool6000002H0oooo002l0?ooo`040=nLB00000000000001d_``0oooo00@0oono
07@000000000B9cO1`3oool00`000000oooo0?ooo`2Z0?ooo`H000009P3oool00=L0oooo00<00000
0?ooo`3oool0[03oool2000002P0oooo003G0?ooo`030000003oool0oooo0=H0oooo003G0?ooo`<0
0000eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3o
ool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0
oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<000000?ooo`3oool0eP3oool00=L0oooo00<0
00000?ooo`3oool0eP3oool00=L0oooo0`00003F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?oo
o`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3o
ool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool00`00
0000oooo0?ooo`3F0?ooo`00e`3oool00`000000oooo0?ooo`3F0?ooo`00e`3oool300000=H0oooo
003G0?ooo`030000003oool0oooo0<40oooo0P00000C0?ooo`00e`3oool00`000000oooo0?ooo`2o
0?ooo`H000004@3oool00=L0oooo00<000000?ooo`3oool0_`3oool600000140oooo002j0?ooo`04
0;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00`3oool01P3ookl0M000000000000000
04RL0=ooo`X0oooo00<000000?ooo`3oool0_P3oool800000100oooo002i0?ooo`060=nLB000M;l0
oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`030?oo_`1d0000B9cO0080oooo
00<0gia80000M02oool02@3oool00`000000oooo0?ooo`2n0?ooo`P00000403oool00;P0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`060?ooo`030?noM000B9`0gooo00<0oooo00<0
_g@004RLg`3oool0203oool00`000000oooo0?ooo`2o0?ooo`H000004@3oool00;P0oooo00<0oooO
09a80018W=l00P3oool0103ooml0W4P00018W03Oool50?ooo`030?noM000B9`0gooo00<0oooo00<0
_g@0001d_`3oool0203oool00`000000oooo0?ooo`2o0?ooo`H000004@3oool00;P0oooo00<0oooO
09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB000M;oo0080oooo00<0oono
07@00018W=l02@3oool00`000000oooo0?ooo`310?ooo`8000004`3oool00;T0oooo00<0_g@00018
W03Oool00P3oool00`3ogi`0B00004RLg`070?ooo`060;md000004P0W=oO09a8000007@0_ooo2@3o
ool400000=D0oooo002i0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0203oool0
1@3ooml0W4P000000000B9`0gooo00T0oooo00<000000?ooo`3oool0eP3oool00;T0oooo00<0omnL
04P00018W=l00`3oool00`2oM00004RL0=ooo`070?ooo`060;md000004P0W=oo0?oOW018B9`0gooo
h@3oool00;T0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`030?oO
W0180000M;oo0080oooo00<0okmd04RLg`3oool0h03oool00;X0oooo00<0omnL04P0001d_ol00P3o
ool00`3OW4P007Bo0?ooo`060?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03Oool0h03o
ool00;/0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`030?oog`2LB000B9cO0080oooo
00<0_g@0001d_`3oool0h03oool00;`0oooo00@0gia800000000000007Bo2@3oool01@3ooml0W4P0
00000000000007Bo0>80oooo003o0?oook40oooo003o0?oook40oooo003o0?oook40oooo003o0?oo
ok40oooo003o0?oook40oooo0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-0.983988, -0.713575, \
0.00684911, 0.00893468}}]
}, Closed]],

Cell["\<\
Next, we want to plot the theoretical curve together with the data.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[ListPlot[data, PlotStyle \[Rule] PointSize[0.02]], 
        Plot[f[x], {x, \(-1\), 1}], PlotRange \[Rule] All];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.309017 0.349747 [
[.02381 .29652 -9.34375 -15.6875 ]
[.02381 .29652 9.34375 0 ]
[.2619 .29652 -13.8438 -15.6875 ]
[.2619 .29652 13.8438 0 ]
[.7381 .29652 -9.5 -15.6875 ]
[.7381 .29652 9.5 0 ]
[.97619 .29652 -5 -15.6875 ]
[.97619 .29652 5 0 ]
[.4875 .04671 -33.6875 -7.84375 ]
[.4875 .04671 0 7.84375 ]
[.4875 .13414 -27.6875 -7.84375 ]
[.4875 .13414 0 7.84375 ]
[.4875 .22158 -33.6875 -7.84375 ]
[.4875 .22158 0 7.84375 ]
[.4875 .39645 -25 -7.84375 ]
[.4875 .39645 0 7.84375 ]
[.4875 .48389 -19 -7.84375 ]
[.4875 .48389 0 7.84375 ]
[.4875 .57133 -25 -7.84375 ]
[.4875 .57133 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30902 m
.02381 .31527 L
s
gsave
.02381 .29652 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30902 m
.2619 .31527 L
s
gsave
.2619 .29652 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30902 m
.7381 .31527 L
s
gsave
.7381 .29652 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30902 m
.97619 .31527 L
s
gsave
.97619 .29652 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.5 .04671 m
.50625 .04671 L
s
gsave
.4875 .04671 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.5 .13414 m
.50625 .13414 L
s
gsave
.4875 .13414 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .22158 m
.50625 .22158 L
s
gsave
.4875 .22158 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.5 .39645 m
.50625 .39645 L
s
gsave
.4875 .39645 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.5 .48389 m
.50625 .48389 L
s
gsave
.4875 .48389 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .57133 m
.50625 .57133 L
s
gsave
.4875 .57133 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .06419 m
.50375 .06419 L
s
.5 .08168 m
.50375 .08168 L
s
.5 .09917 m
.50375 .09917 L
s
.5 .11666 m
.50375 .11666 L
s
.5 .15163 m
.50375 .15163 L
s
.5 .16912 m
.50375 .16912 L
s
.5 .18661 m
.50375 .18661 L
s
.5 .20409 m
.50375 .20409 L
s
.5 .23907 m
.50375 .23907 L
s
.5 .25655 m
.50375 .25655 L
s
.5 .27404 m
.50375 .27404 L
s
.5 .29153 m
.50375 .29153 L
s
.5 .3265 m
.50375 .3265 L
s
.5 .34399 m
.50375 .34399 L
s
.5 .36148 m
.50375 .36148 L
s
.5 .37897 m
.50375 .37897 L
s
.5 .41394 m
.50375 .41394 L
s
.5 .43143 m
.50375 .43143 L
s
.5 .44892 m
.50375 .44892 L
s
.5 .4664 m
.50375 .4664 L
s
.5 .50138 m
.50375 .50138 L
s
.5 .51887 m
.50375 .51887 L
s
.5 .53635 m
.50375 .53635 L
s
.5 .55384 m
.50375 .55384 L
s
.5 .02922 m
.50375 .02922 L
s
.5 .01173 m
.50375 .01173 L
s
.5 .58881 m
.50375 .58881 L
s
.5 .6063 m
.50375 .6063 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
.02 w
.07143 .07214 Mdot
.11905 .16959 Mdot
.16667 .21062 Mdot
.21429 .22929 Mdot
.2619 .27251 Mdot
.30952 .31486 Mdot
.35714 .27219 Mdot
.40476 .2813 Mdot
.45238 .31611 Mdot
.5 .31047 Mdot
.54762 .32767 Mdot
.59524 .25004 Mdot
.64286 .25478 Mdot
.69048 .30201 Mdot
.7381 .34944 Mdot
.78571 .40785 Mdot
.83333 .40741 Mdot
.88095 .52598 Mdot
.92857 .57969 Mdot
.5 Mabswid
.02381 .01472 m
.06244 .07428 L
.10458 .13099 L
.14415 .17628 L
.18221 .21261 L
.22272 .2438 L
.26171 .267 L
.28158 .27644 L
.30316 .28501 L
.32216 .29122 L
.34309 .29673 L
.36292 .30079 L
.38395 .304 L
.40336 .30611 L
.41426 .30699 L
.42424 .30761 L
.43408 .30809 L
.44322 .30843 L
.45325 .30869 L
.45886 .30879 L
.46395 .30887 L
.46882 .30892 L
.47149 .30894 L
.47397 .30896 L
.47837 .30898 L
.4807 .30899 L
.48319 .309 L
.48587 .30901 L
.48739 .30901 L
.48879 .30901 L
.49001 .30901 L
.49136 .30901 L
.49203 .30902 L
.49276 .30902 L
.49408 .30902 L
.49529 .30902 L
.49661 .30902 L
.49727 .30902 L
.49798 .30902 L
.49926 .30902 L
.50054 .30902 L
.50175 .30902 L
.50284 .30902 L
.50401 .30902 L
.50525 .30902 L
.50637 .30902 L
.50767 .30902 L
.5089 .30902 L
.51004 .30902 L
.51109 .30902 L
Mistroke
.51348 .30902 L
.51484 .30903 L
.51613 .30903 L
.51852 .30904 L
.5211 .30905 L
.52389 .30906 L
.52923 .3091 L
.53191 .30912 L
.53486 .30915 L
.54007 .30923 L
.54493 .30931 L
.54996 .30942 L
.55541 .30957 L
.56533 .30992 L
.57498 .31038 L
.58389 .31092 L
.6026 .31249 L
.61324 .31368 L
.62321 .31501 L
.64343 .31843 L
.66195 .32251 L
.68252 .32823 L
.70163 .33478 L
.73979 .35181 L
.77889 .37533 L
.82044 .40772 L
.86047 .44665 L
.89899 .49151 L
.93996 .54725 L
.97619 .60332 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00
f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00f@3oool300000=@0oooo000>0?ooo`030000003oool0oooo0<P0
oooo00<000000?ooo`3oool0e03oool000l0oooo00<000000?ooo`3oool0a`3oool00`000000oooo
0?ooo`3D0?ooo`003`3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo0=@0oooo000@
0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0e03oool00140oooo00<000000?oo
o`3oool0a@3oool00`000000oooo0?ooo`3D0?ooo`004@3oool00`000000oooo0?ooo`350?ooo`03
0000003oool0oooo0=@0oooo000B0?ooo`030000003oool0oooo0<@0oooo0`00003D0?ooo`004P3o
ool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0=@0oooo000C0?ooo`030000003oool0
oooo0<<0oooo00<000000?ooo`3oool0e03oool001@0oooo00<000000?ooo`3oool0W03oool0102o
M00000000000001d_ol30?ooo`030?oOW0180000M;oo00<0oooo00<0oono07A8W03Oool01@3oool0
1P3ogi`0B00000000000000004RL0=ooo`/0oooo00<000000?ooo`3oool0e03oool001@0oooo00<0
00000?ooo`3oool0V`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d
_ol0103oool00`3OW4P0B9cO0?ooo`040?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7Bo2`3oool0
0`000000oooo0?ooo`3D0?ooo`005@3oool00`000000oooo0?ooo`2I0?ooo`030?oog`2LB000B9cO
0080oooo00<0okmd0000B02Lgol0203oool00`3ookl0M01d0;ooo`090?ooo`030?noM018W=l0oooo
00T0oooo00<000000?ooo`3oool0e03oool001H0oooo00<000000?ooo`3oool0V03oool00`3ooml0
W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool0203oool00`3o
okl0M01d0;ooo`090?ooo`030000003oool0oooo0=@0oooo000F0?ooo`030000003oool0oooo08/0
oooo00<0oooO09a8000000001`0000001P18W=l0oooo0?ooo`3ooml0W4P0001d_`<0oooo00<0gia8
0000B02Lgol0203oool00`3ooml0W4Q809cOo`080?ooo`030?oog`2LB000M;oo00T0oooo00<00000
0?ooo`3oool0e03oool001L0oooo00<000000?ooo`3oool0V03oool00`2oM00004RL0=ooo`020?oo
o`030?oOW0180000B9cO00T0oooo00<0omnL04Qd_`3oool01`3oool00`3ookl0M00004RLg`090?oo
o`@00000d`3oool001P0oooo00<000000?ooo`3oool0U`3oool00`3OW4P0001809cOo`020?ooo`03
0?oog`2LB00007Bo00X0oooo00<0_g@007Boo`3oool01P3oool00`3o_g@0000007Boo`090?ooo`03
0000003oool0oooo0=@0oooo000H0?ooo`030000003oool0oooo09L0oooo00<0omnL04P00018W=l0
0`3oool00`2oM00004RL0=ooo`090?ooo`030?oOW018B9`0gooo00<0oooo00H0omoO0=ooo`3o_g@0
00000000M02oool90?ooo`030000003oool0oooo0=@0oooo000I0?ooo`030000003oool0oooo09H0
oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`0:0?ooo`030=nLB018W=l0oooo
0080oooo00D0oono07@000000000001809cOo`0:0?ooo`030000003oool0oooo0=@0oooo000J0?oo
o`030000003oool0oooo09H0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`050?oo
o`030?oog`3Ogol0oooo0080oooo00<0oono07@0M02oool00`3oool00`3OW9`0_ooo0?ooo`0;0?oo
o`030000003oool0oooo0=@0oooo000J0?ooo`030000003oool0oooo09L0oooo00H0gia80000M02o
ool0oooO09a80018W=l70?ooo`030;mdM02oool0oooo0080oooo00<0okmd001d_`3oool00P3oool0
0`3ogi`0M9cO0?ooo`0;0?ooo`030000003oool0oooo0=@0oooo000K0?ooo`030000003oool0oooo
09L0oooo00@0gia800000000000007Bo203oool00`3ookl0M00000000003000000030000B02Lgol0
oooo0080oooo00D0_g@0000000000000000004RLg`080?ooo`030000003oool0oooo0=@0oooo000K
0?ooo`030000003oool0oooo00<0oooo0P00002f0?ooo`<00000e03oool001`0oooo00<000000?oo
o`3oool01P00002d0?ooo`030000003oool0oooo0=@0oooo000M0?ooo`030000003oool0000000D0
0000]03oool00`000000oooo0?ooo`3D0?ooo`007@3oool900000;<0oooo00<000000?ooo`3oool0
e03oool001h0oooo2000002c0?ooo`030000003oool0oooo0=@0oooo000O0?ooo`H00000]03oool0
0`000000oooo0?ooo`3D0?ooo`007`3oool600000;@0oooo00<000000?ooo`3oool0e03oool00200
oooo0`00002f0?ooo`<00000e03oool00240oooo00<000000?ooo`3oool0]@3oool00`000000oooo
0?ooo`3D0?ooo`008P3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0=@0oooo000R
0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3oool0e03oool002<0oooo00<000000?oo
o`3oool0/`3oool00`000000oooo0?ooo`3D0?ooo`00903oool00`000000oooo0?ooo`2b0?ooo`03
0000003oool0oooo0=@0oooo000T0?ooo`030000003oool0oooo0;80oooo00<000000?ooo`3oool0
e03oool002D0oooo00<000000?ooo`3oool0/@3oool00`000000oooo0?ooo`3D0?ooo`009P3oool0
0`000000oooo0?ooo`2`0?ooo`<00000e03oool002L0oooo00<000000?ooo`3oool0[`3oool00`00
0000oooo0?ooo`3D0?ooo`009`3oool00`000000oooo0?ooo`2_0?ooo`030000003oool0oooo0=@0
oooo000X0?ooo`030000003oool0oooo0:h0oooo00<000000?ooo`3oool0e03oool002T0oooo00<0
00000?ooo`3oool0[@3oool00`000000oooo0?ooo`3D0?ooo`00:P3oool00`000000oooo0?ooo`2/
0?ooo`030000003oool0oooo0=@0oooo000Z0?ooo`030000003oool0oooo0:`0oooo00<000000?oo
o`3oool0e03oool002/0oooo00<000000?ooo`3oool0Z`3oool300000=@0oooo000/0?ooo`030000
003oool0oooo0:X0oooo00<000000?ooo`3oool0e03oool002`0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`3D0?ooo`00;@3oool00`000000oooo0?ooo`2<0?ooo`040;md00000000
000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=ooo`/0
oooo00<000000?ooo`3oool0e03oool002h0oooo00<000000?ooo`3oool0RP3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_`/0oooo00<000000?ooo`3oool0e03oool002l0oooo00<000000?ooo`3oool0R03oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool02@3oool0
0`000000oooo0?ooo`3D0?ooo`00;`3oool00`000000oooo0?ooo`280?ooo`030?oog`2LB000B9cO
0080oooo00@0oooO09a80000B9`0gooo2P3oool00`3ookl0M01d0;ooo`090?ooo`030000003oool0
oooo0=@0oooo000`0?ooo`030000003oool0oooo07X0oooo00<0oooO09a8000000001`0000001P18
W=l0oooo0?ooo`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Bo
o`090?ooo`030000003oool0oooo0=@0oooo000a0?ooo`030000003oool0oooo08L0oooo00<0_g@0
0018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030?oo_`1d0000B9cO00T0oooo1000003C
0?ooo`00<P3oool00`000000oooo0?ooo`260?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l02P3oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo0=@0oooo000c0?ooo`03
0000003oool0oooo08D0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo2@3oool00`000000oooo0?ooo`3D0?ooo`00=03oool00`00
0000oooo0?ooo`240?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool0
1@3ookl0M0000000000004P0W=oo00X0oooo00<000000?ooo`3oool0e03oool003@0oooo00<00000
0?ooo`3oool0Q@3oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL
0;ooo`3oool02`3oool00`000000oooo0?ooo`3D0?ooo`00=@3oool00`000000oooo0?ooo`250?oo
o`060=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0;0?ooo`030000003o
ool0oooo0=@0oooo000f0?ooo`030000003oool0oooo08D0oooo00@0gia800000000000007Bo2P3o
ool01@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3oool0e03oool003L0oooo00<0
00000?ooo`3oool0W`3oool300000=@0oooo000h0?ooo`030000003oool0oooo09h0oooo00<00000
0?ooo`3oool0e03oool003T0oooo00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`3D0?oo
o`00>P3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo0=@0oooo000k0?ooo`030000
003oool0oooo09/0oooo00<000000?ooo`3oool0e03oool003D0oooo0P0000050?ooo`030000003o
ool0oooo09X0oooo00<000000?ooo`3oool0e03oool003<0oooo1P0000040?ooo`030000003oool0
oooo09T0oooo00<000000?ooo`3oool0e03oool003<0oooo1P0000040?ooo`030000003oool0oooo
09T0oooo00<000000?ooo`3oool0e03oool00380oooo200000040?ooo`030000003oool0oooo09P0
oooo0`00003D0?ooo`00<P3oool8000000D0oooo00<000000?ooo`3oool0U`3oool00`000000oooo
0?ooo`3D0?ooo`00<`3oool6000000L0oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?oo
o`3D0?ooo`00<`3oool6000000P0oooo00<000000?ooo`3oool0U@3oool00`000000oooo0?ooo`3D
0?ooo`00=@3oool2000000/0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`3D0?oo
o`00@`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo0=@0oooo00140?ooo`030000
003oool0oooo0980oooo00<000000?ooo`3oool0e03oool004D0oooo00<000000?ooo`3oool0T@3o
ool300000=@0oooo00160?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0e03oool0
04L0oooo00<000000?ooo`3oool0S`3oool00`000000oooo0?ooo`3D0?ooo`00B03oool00`000000
oooo0?ooo`2>0?ooo`030000003oool0oooo0=@0oooo00190?ooo`800000SP3oool00`000000oooo
0?ooo`3D0?ooo`00B`3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo0=@0oooo001<
0?ooo`030000003oool0oooo08X0oooo00<000000?ooo`3oool0e03oool004X0oooo0P0000000`3o
ool000000?ooo`2:0?ooo`030000003oool0oooo0=@0oooo00180?ooo`L00000RP3oool300000=@0
oooo00180?ooo`H0000000<0oooo0000003oool0R03oool00`000000oooo0?ooo`3D0?ooo`00A`3o
ool8000000030?ooo`000000oooo0640oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B000
07Boo`020?ooo`030=nLB0000000000000<0000000D0B9cO0?ooo`3oool0omnL04P0000200000003
0018W03Oool0oooo00X0oooo00<000000?ooo`3oool0e03oool004L0oooo200000020?ooo`030000
003oool0oooo05h0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo
00<0oooo00<0okmd001d_`3oool00P3oool02@2oM7@0_ooo0?ooo`3o_g@0000007Boo`3oool0omnL
04Qd_`0;0?ooo`030000003oool0oooo0=@0oooo00180?ooo`H00000103oool00`000000oooo0?oo
o`1L0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3o_g@007Bo0?oo
o`090?ooo`030?noM018W=l0oooo00T0oooo00<000000?ooo`3oool0e03oool004P0oooo1P000005
0?ooo`800000G03oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0
okmd001d_`3oool0203oool00`3ookl0M01d0;ooo`090?ooo`030000003oool0oooo0=@0oooo001:
0?ooo`8000002@3oool00`000000oooo0?ooo`1<0?ooo`030?oog`2LB000000000L0000000H0B9cO
0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00T0oooo00<0okmd04RLg`3oool0
1`3oool00`3ooml0W4P007Boo`090?ooo`030000003oool0oooo0=@0oooo001F0?ooo`030000003o
ool0oooo00D0oooo0P00001B0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3o
ool00`3OW4P0B9cO0?ooo`060?ooo`030?oo_`1d0000B9cO00T0oooo1000003C0?ooo`00E`3oool0
0`000000oooo0?ooo`020?ooo`H00000D03oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB000
07Bo00/0oooo00<0_g@007Boo`3oool01@3oool00`3o_g@0000007Boo`090?ooo`030000003oool0
oooo0=@0oooo001H0?ooo`040000003oool0oooo0?ooo`H00000D03oool00`3ogi`0B00004RLg`03
0?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02oool00P3oool01P3ogml0gooo0?noM000
0000001d0;ooo`T0oooo00<000000?ooo`3oool0e03oool005T0oooo2P00001?0?ooo`040?oog`2L
B00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3ooml0W4P007Boo`020?ooo`050?oo
_`1d000000000000B02Lgol02P3oool00`000000oooo0?ooo`3D0?ooo`00F`3oool800000500oooo
00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gibL0;ooo`3oool02`3oool00`000000oooo0?ooo`3D0?ooo`00G03oool600000580oooo00H0
gia80000M02oool0oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`03
0?ooo`030?oOW01dW=l0oooo00/0oooo00<000000?ooo`3oool0e03oool005`0oooo1P00001C0?oo
o`040=nLB00000000000001d_`T0oooo00D0omnL04P000000000000007Boo`050?ooo`050;md0000
0000000000000018W=l0203oool00`000000oooo0?ooo`3D0?ooo`00GP3oool2000007T0oooo00<0
00000?ooo`3oool0e03oool005l0oooo0P00001h0?ooo`<00000e03oool00640oooo00<000000?oo
o`3oool0M@3oool00`000000oooo0?ooo`0U0?ooo`800000[@3oool00680oooo00<000000?ooo`3o
ool0M03oool00`000000oooo0?ooo`0S0?ooo`H00000Z`3oool006<0oooo0P00001d0?ooo`030000
003oool0oooo02<0oooo1P00000@0?ooo`800000V@3oool006D0oooo0P00001b0?ooo`030000003o
ool0oooo0280oooo2000000=0?ooo`H00000U`3oool006L0oooo00<000000?ooo`3oool0K`3oool0
0`000000oooo0?ooo`0R0?ooo`P000003@3oool6000009L0oooo001X0?ooo`800000K`3oool00`00
0000oooo0?ooo`0S0?ooo`H000003@3oool8000009H0oooo001Z0?ooo`800000K@3oool3000002<0
oooo1P00000=0?ooo`P00000UP3oool00100oooo00<0oono07@0000000000P0000000`18W=l0oooo
0?ooo`1D0?ooo`8000000P3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080
oooo00H0omnL04P00000000000000018W03OoomG0?ooo`030000003oool0oooo02D0oooo0P00000@
0?ooo`H00000703oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00H0
omnL04P00000000000000018W03OoomI0?ooo`030?oo_`1d000000000080000000<0B9cO0?ooo`3o
ool0203oool00180oooo00<0okmd0000001d_ol0F@3oool01`000000gia8001d_`3oool0omnL04P0
0018W=l00P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7BoE`3o
ool00`000000oooo0?ooo`0g0?ooo`H000006`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80
oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_e/0oooo00<0okmd
0000001d_ol02`3oool00180oooo00<0oooO09a80000M;l0F@3oool00`3ooml00000000000020?oo
o`030?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool0E@3oool00`000000oooo0?ooo`0i0?oo
o`800000703oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RL
g`3oool0F@3oool00`3ooml0W4P0001d_`0;0?ooo`004`3oool00`3OW4P004RL0=ooo`1H0?ooo`03
0?oog`2LB000B9cO00<0000000<0W4P00018W03Oool02P3oool00`3ookl0M01d0;ooo`0G0?ooo`80
0000?03oool00`000000oooo0?ooo`1G0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000
B9`0gooo2P3oool00`3ookl0M01d0;ooo`1J0?ooo`030=nLB000B9`0gooo00X0oooo00030?ooo`03
0?oog`2LB000000000L0000000<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`1;0?ooo`03
0?oog`2LB000000000L0000000D0B9cO0?ooo`3oool0oooO09a800060000000309cOo`3oool0oooo
00P0oooo00<0oooO09a8001d_ol05@3oool6000003X0oooo00<000000?ooo`3oool0E`3oool00`3o
oml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8001d_ol0FP3oool00`3o
gi`0B01809cOo`0:0?ooo`004`3oool00`3ookl0M00007Boo`1I0?ooo`030;md00000000000000D0
00002P3oool00`3ookl0M00004RLg`0E0?ooo`H00000>P3oool00`000000oooo0?ooo`1H0?ooo`03
0;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0M00004RLg`1J0?ooo`03
0?oo_`1d0000M;oo00X0oooo000D0?ooo`030;md0000M;l0oooo05P0oooo2P0000080?ooo`030?no
M0000000M;oo01@0oooo2000000?0?ooo`800000:03oool00`000000oooo0?ooo`1H0?ooo`030=nL
B00004P0W=oo0080oooo00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`1K0?ooo`030;md
0000M;l0oooo00T0oooo000D0?ooo`030=nLB000B9`0gooo05P0oooo20000000103Oool0oooo0000
000000030?ooo`060?oOg`3Oool0okmd0000000007@0_ooo503oool8000000d0oooo1P00000V0?oo
o`<00000F03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00H0omoO0=oo
o`3o_g@000000000M02ooomK0?ooo`030=nLB000B9`0gooo00T0oooo000D0?ooo`030?oOW01804P0
W=oo05P0oooo00<0oooO000000000000100000000`00B9`0gooo0?ooo`020?ooo`<0000000D0oono
07@000000000001809cOo`0F0?ooo`H000003P3oool6000002H0oooo00<000000?ooo`3oool0F03o
ool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000
001809cOo`1L0?ooo`030?oOW01804P0W=oo00T0oooo000D0?ooo`030?oo_`1d0000M;oo05T0oooo
1P0000000`00M;l0oooo0?ooo`050?ooo`<00000603oool6000000d0oooo2000000U0?ooo`030000
003oool0oooo05T0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nL
W02oool0oooo05d0oooo00<0oono07@0001d_ol02@3oool001D0oooo00<0_g@0001d_`3oool0F@3o
ool01P3OW4P000000000003ooml0W4P004RLg`P0oooo00<0omnL07BLg`0000000P00000G0?ooo`80
00003`3oool8000002D0oooo00<000000?ooo`3oool0FP3oool01P3OW4P0001d0;ooo`3ooml0W4P0
04RLg`P0oooo00<0omnL07BLg`3oool0GP3oool00`2oM00007Bo0?ooo`080?ooo`004`3oool01@3o
okl0M0000000000007@0_ooo05X0oooo00@0gia800000000000007Bo2P3oool00`2oM00000000000
0004000002H0oooo1P00000V0?ooo`030000003oool0oooo05/0oooo00@0gia800000000000007Bo
2P3oool01@2oM0000000000000000000B9cO05T0oooo00D0oono07@000000000001d0;ooo`080?oo
o`00Q`3oool2000002@0oooo1P00000V0?ooo`030000003oool0oooo0=@0oooo00290?ooo`<00000
8`3oool2000002P0oooo00<000000?ooo`3oool0e03oool008`0oooo100000190?ooo`<00000e03o
ool00900oooo100000150?ooo`030000003oool0oooo04h0oooo0P0000240?ooo`00U03oool40000
0440oooo00<000000?ooo`3oool0C03oool600000880oooo002H0?ooo`@00000?@3oool00`000000
oooo0?ooo`1<0?ooo`H00000PP3oool009`0oooo1`00000f0?ooo`030000003oool0oooo04/0oooo
200000210?ooo`00X`3oool9000002`0oooo0P00001=0?ooo`P00000P@3oool00:`0oooo1P00000T
0?ooo`H00000C03oool600000880oooo00270?ooo`800000:@3ooolI000000/0oooo1P00001<0?oo
o`H00000PP3oool000<0ooooo`00002]00000040oooo000>0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0?0?ooo`H000004@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo00l0oooo1P00000=0?ooo`P00000403oool00`000000oooo
0?ooo`0A0?ooo`H000003`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00T0oooo
000>0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool03P3oool8000003D0oooo2000
000=0?ooo`H00000:`3oool9000002l0oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?oo
o`090?ooo`00Q03oool8000003D0oooo2000000=0?ooo`H00000=03oool4000009`0oooo00250?oo
o`H00000=`3oool600000100oooo0P00000C0?ooo`8000009@3oool6000009H0oooo00250?ooo`H0
0000=`3oool600000140oooo00<000000?ooo`3oool03`3oool6000002T0oooo1P00002@0?ooo`00
Q`3oool2000003/0oooo0P00000C0?ooo`030000003oool0oooo00l0oooo1P00000_0?ooo`<00000
S@3oool00=T0oooo0`00000>0?ooo`P00000<@3oool3000008X0oooo003I0?ooo`030000003oool0
oooo00h0oooo2000000d0?ooo`<00000Q`3oool00=T0oooo00<000000?ooo`3oool03`3oool60000
03P0oooo0P0000250?ooo`00f@3oool00`000000oooo0?ooo`0?0?ooo`H00000>P3oool300000880
oooo003I0?ooo`030000003oool0oooo0140oooo0P00000o0?ooo`<00000O`3oool00=T0oooo00<0
00000?ooo`3oool0E@3oool2000007d0oooo003I0?ooo`030000003oool0oooo05L0oooo0P00001k
0?ooo`00f@3oool00`000000oooo0?ooo`1I0?ooo`8000001`3oool200000700oooo003I0?ooo`<0
0000F`3oool300000080oooo1P00001^0?ooo`00f@3oool00`000000oooo0?ooo`1N0?ooo`P00000
KP3oool00=T0oooo00<000000?ooo`3oool0G`3oool8000006d0oooo003I0?ooo`030000003oool0
oooo05l0oooo2000001]0?ooo`00f@3oool00`000000oooo0?ooo`1P0?ooo`H00000KP3oool00=T0
oooo00<000000?ooo`3oool0H03oool8000006`0oooo003I0?ooo`030000003oool0oooo0680oooo
0P0000040?ooo`030000003oool0oooo06T0oooo003I0?ooo`<00000J@3oool2000006T0oooo003I
0?ooo`030000003oool0oooo06/0oooo0P00001W0?ooo`00f@3oool00`000000oooo0?ooo`1]0?oo
o`800000I@3oool00=T0oooo00<000000?ooo`3oool0K`3oool00`000000oooo0?ooo`1R0?ooo`00
f@3oool00`000000oooo0?ooo`1`0?ooo`800000HP3oool00=T0oooo00<000000?ooo`3oool0LP3o
ool200000600oooo003I0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0G@3oool0
0=T0oooo0`00001e0?ooo`030000003oool0oooo05`0oooo003I0?ooo`030000003oool0oooo07H0
oooo0P00001L0?ooo`00f@3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo05T0oooo
002c0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P00000
000000030000000504RLg`3oool0oooo0?oOW01800000P0000000`00B9`0gooo0?ooo`0:0?ooo`03
0000003oool0oooo07T0oooo00<000000?ooo`3oool0F03oool00;80oooo00H0gia8001d_`3oool0
omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool02@2o
M7@0_ooo0?ooo`3o_g@0000007Boo`3oool0omnL04Qd_`0;0?ooo`030000003oool0oooo07X0oooo
0P00001H0?ooo`00/@3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0
okmd001d_`3oool02@3oool00`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo07`0oooo00<0
00000?ooo`3oool0E@3oool00;40oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03O
ool80?ooo`030?noM000M;l0oooo00P0oooo00<0oono07@0M02oool02@3oool00`000000oooo0?oo
o`1m0?ooo`030000003oool0oooo05@0oooo002a0?ooo`030?oog`2LB00007Bo00<0oooo00<0gia8
0000B02Lgol02@3oool00`3o_g@0B9cO0?ooo`070?ooo`030?oog`2LB000M;oo00T0oooo00<00000
0?ooo`3oool0OP3oool00`000000oooo0?ooo`1C0?ooo`00/P3oool00`2oM00004RL0=ooo`020?oo
o`030?oOW0180000B9cO00X0oooo00<0gia804RLg`3oool01P3oool00`3ookl0M00004RLg`090?oo
o`@00000OP3oool2000005<0oooo002b0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000
M;l02`3oool00`2oM000M;oo0?ooo`050?ooo`030?noM0000000M;oo00T0oooo00<000000?ooo`3o
ool0P@3oool00`000000oooo0?ooo`070?ooo`800000A`3oool00;80oooo00<0omnL04P00018W=l0
0`3oool00`2oM00004RL0=ooo`0:0?ooo`030?oOW01807@0_ooo0080oooo00H0omoO0=ooo`3o_g@0
00000000M02oool90?ooo`030000003oool0oooo07H0oooo0P00000:0?ooo`030000003oool0oooo
00@0oooo1P0000150?ooo`00/P3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo
00X0oooo00<0oooO09a8001d_ol00P3oool01@3ookl0M0000000000004P0W=oo00X0oooo00<00000
0?ooo`3oool0M03oool6000000T0oooo0P0000040?ooo`H00000A@3oool00;<0oooo00<0omnL04P0
001d_ol00P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO00<0oooo00<0gibL0;oo
o`3oool02`3oool00`000000oooo0?ooo`1d0?ooo`H000002`3oool00`000000oooo0?ooo`080000
04@0oooo002d0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool01`3o_g@004RL0=ooo`3o
ool0gia80000B02Lgol00`3oool00`3ogi`0M9cO0?ooo`0;0?ooo`030000003oool0oooo07<0oooo
2000000;0?ooo`030000003oool0000000L00000A03oool00;D0oooo00@0gia800000000000007Bo
2@3oool01@3ogi`0B000000000000000M;oo00D0oooo00D0_g@0000000000000000004RLg`080?oo
o`030000003oool0oooo07<0oooo2000000<0?ooo`030000003oool0000000D00000A@3oool00=T0
oooo0`00001d0?ooo`H000003P3oool7000004D0oooo003I0?ooo`030000003oool0oooo07@0oooo
1P00000?0?ooo`040000003oool00000000004L0oooo003I0?ooo`030000003oool0oooo07H0oooo
0P00000B0?ooo`030000003oool0oooo04L0oooo003I0?ooo`030000003oool0oooo08/0oooo00<0
00000?ooo`3oool0AP3oool00=T0oooo00<000000?ooo`3oool0S03oool00`000000oooo0?ooo`15
0?ooo`00f@3oool00`000000oooo0?ooo`2=0?ooo`030000003oool0oooo04@0oooo003I0?ooo`03
0000003oool0oooo08h0oooo0P0000140?ooo`00f@3oool00`000000oooo0?ooo`2@0?ooo`030000
003oool0oooo0440oooo003I0?ooo`<00000T@3oool00`000000oooo0?ooo`100?ooo`00f@3oool0
0`000000oooo0?ooo`2B0?ooo`030000003oool0oooo03l0oooo003I0?ooo`030000003oool0oooo
09<0oooo00<000000?ooo`3oool0?P3oool00=T0oooo00<000000?ooo`3oool0U03oool00`000000
oooo0?ooo`0m0?ooo`00f@3oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo03`0oooo
003I0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0>`3oool00=T0oooo00<00000
0?ooo`3oool0U`3oool00`000000oooo0?ooo`0j0?ooo`00f@3oool3000009P0oooo00<000000?oo
o`3oool0>@3oool00=T0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`0h0?ooo`00
f@3oool00`000000oooo0?ooo`2I0?ooo`030000003oool0oooo03P0oooo003I0?ooo`030000003o
ool0oooo09X0oooo00<000000?ooo`3oool0=`3oool00=T0oooo00<000000?ooo`3oool0V`3oool0
0`000000oooo0?ooo`0f0?ooo`00f@3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0oooo
03D0oooo003I0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0=03oool00=T0oooo
00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`0d0?ooo`00f@3oool3000009h0oooo00<0
00000?ooo`3oool0<`3oool00=T0oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?ooo`0b
0?ooo`00_03oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00H0omnL
04P00000000000000018W03Oool;0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0
<@3oool00;/0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080
oooo00H0okmd0000001d_ol0oooo0?oOW018M;l;0?ooo`030000003oool0oooo0:40oooo00<00000
0?ooo`3oool0<03oool00;X0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?oo
o`030?noM018W=l0oooo00T0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0`0?oo
o`00^P3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0
M02oool02@3oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo02l0oooo002j0?ooo`03
0?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`090?ooo`03
0000003oool0oooo0:<0oooo00<000000?ooo`3oool0;P3oool00;/0oooo00<0_g@00018W03Oool0
0P3oool00`3ogi`0B00004RLg`0:0?ooo`030?oo_`1d0000B9cO00T0oooo1000002S0?ooo`030000
003oool0oooo02d0oooo002k0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02P3o
ool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool0;03o
ool00;/0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`060?oOg`3Oool0
okmd0000000007@0_ooo2@3oool00`000000oooo0?ooo`2U0?ooo`030000003oool0oooo02`0oooo
002k0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool01@3ookl0M000
0000000004P0W=oo00X0oooo00<000000?ooo`3oool0YP3oool00`000000oooo0?ooo`0[0?ooo`00
_03oool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3oool0
2`3oool00`000000oooo0?ooo`2W0?ooo`030000003oool0oooo02X0oooo002m0?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0;0?ooo`030000003oool0oooo0:P0
oooo00<000000?ooo`3oool0:@3oool00;h0oooo00@0gia800000000000007Bo2P3oool01@2oM000
0000000000000000B9cO00P0oooo00<000000?ooo`3oool0Z@3oool00`000000oooo0?ooo`0X0?oo
o`00f@3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo02P0oooo003I0?ooo`<00000
ZP3oool00`000000oooo0?ooo`0W0?ooo`00f@3oool00`000000oooo0?ooo`2[0?ooo`030000003o
ool0oooo02H0oooo003I0?ooo`030000003oool0oooo0:`0oooo00<000000?ooo`3oool09@3oool0
0=T0oooo00<000000?ooo`3oool0[03oool00`000000oooo0?ooo`0U0?ooo`00f@3oool00`000000
oooo0?ooo`2]0?ooo`030000003oool0oooo02@0oooo003I0?ooo`030000003oool0oooo0:h0oooo
00<000000?ooo`3oool08`3oool00=T0oooo00<000000?ooo`3oool0[`3oool00`000000oooo0?oo
o`0R0?ooo`00f@3oool3000009l0oooo0P00000?0?ooo`030000003oool0oooo0240oooo003I0?oo
o`030000003oool0oooo09d0oooo1P00000=0?ooo`030000003oool0oooo0240oooo003I0?ooo`03
0000003oool0oooo09d0oooo1P00000>0?ooo`030000003oool0oooo0200oooo003I0?ooo`030000
003oool0oooo09`0oooo2000000>0?ooo`030000003oool0oooo01l0oooo003I0?ooo`030000003o
ool0oooo09`0oooo2000000?0?ooo`030000003oool0oooo01h0oooo003I0?ooo`030000003oool0
oooo09d0oooo1P00000A0?ooo`030000003oool0oooo01d0oooo003I0?ooo`030000003oool0oooo
09d0oooo1P00000A0?ooo`030000003oool0oooo01d0oooo003I0?ooo`030000003oool0oooo09l0
oooo0P00000D0?ooo`030000003oool0oooo01`0oooo003I0?ooo`<00000]P3oool00`000000oooo
0?ooo`0K0?ooo`00f@3oool00`000000oooo0?ooo`2g0?ooo`030000003oool0oooo01X0oooo003I
0?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool06P3oool00=T0oooo00<000000?oo
o`3oool0^03oool00`000000oooo0?ooo`0I0?ooo`00f@3oool00`000000oooo0?ooo`2i0?ooo`03
0000003oool0oooo01P0oooo003I0?ooo`030000003oool0oooo0;X0oooo00<000000?ooo`3oool0
5`3oool00=T0oooo00<000000?ooo`3oool0^P3oool00`000000oooo0?ooo`0G0?ooo`00f@3oool3
00000;/0oooo00<000000?ooo`3oool05P3oool00=T0oooo00<000000?ooo`3oool0^`3oool00`00
0000oooo0?ooo`0F0?ooo`00f@3oool00`000000oooo0?ooo`2l0?ooo`030000003oool0oooo01D0
oooo002c0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00`3oool00`3ookl0
M4RL0=ooo`050?ooo`060?oOW0180000000000000000B9`0gooo2`3oool00`000000oooo0?ooo`2m
0?ooo`030000003oool0oooo01@0oooo002b0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3o
ool00`3ogi`0B00007Boo`040?ooo`030=nLB018W=l0oooo00@0oooo00H0okmd0000001d_ol0oooo
0?oOW018M;l;0?ooo`030000003oool0oooo0;d0oooo00<000000?ooo`3oool0503oool00;40oooo
00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?oo_`1d07@0_ooo00T0oooo
00<0okmd04RLg`3oool02@3oool00`000000oooo0?ooo`2n0?ooo`030000003oool0oooo01<0oooo
002a0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo
0?ooo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<000000?ooo`3oool0_`3oool00`000000oooo
0?ooo`0B0?ooo`00/@3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00P0oooo00<0
oooO09a8B02Lgol0203oool00`3ooml0W4P007Boo`090?ooo`030000003oool0oooo0;l0oooo00<0
00000?ooo`3oool04P3oool00;80oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`09
0?ooo`030?oOW018M;l0oooo00L0oooo00<0oono07@00018W=l02@3oool400000;80oooo0P00000;
0?ooo`030000003oool0oooo0140oooo002b0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a8
0000M;l02P3oool00`2oM000M;oo0?ooo`060?ooo`030?noM0000000M;oo00T0oooo00<000000?oo
o`3oool0/@3oool6000000T0oooo00<000000?ooo`3oool04@3oool00;80oooo00<0omnL04P00018
W=l00`3oool00`2oM00004RL0=ooo`090?ooo`030?oOW018B9`0gooo00<0oooo00H0omoO0=ooo`3o
_g@000000000M02oool90?ooo`030000003oool0oooo0;40oooo1P00000:0?ooo`030000003oool0
oooo0100oooo002b0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool0
0`3OW4P0B9cO0?ooo`020?ooo`050?oo_`1d000000000000B02Lgol02P3oool00`000000oooo0?oo
o`2`0?ooo`P000002P3oool00`000000oooo0?ooo`0?0?ooo`00/`3oool00`3ogi`0B00007Boo`02
0?ooo`030=nLB000M;l0oooo00D0oooo00<0oooO0=oOo`3oool00P3oool00`3ookl0M01d0;ooo`03
0?ooo`030=nLW02oool0oooo00/0oooo00<000000?ooo`3oool0/03oool8000000X0oooo00<00000
0?ooo`3oool03`3oool00;@0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`030;mdM02o
ool0oooo0080oooo00<0okmd001d_`3oool00P3oool00`3ogi`0M9cO0?ooo`0;0?ooo`030000003o
ool0oooo0;40oooo1P00000<0?ooo`030000003oool0oooo00h0oooo002e0?ooo`040=nLB0000000
0000001d_`P0oooo00<0oono07@0000000000`0000000`0004P0W=oo0?ooo`020?ooo`050;md0000
0000000000000018W=l0203oool00`000000oooo0?ooo`2a0?ooo`H000003@3oool00`000000oooo
0?ooo`0=0?ooo`00f@3oool300000;<0oooo0P00000?0?ooo`030000003oool0oooo00d0oooo003I
0?ooo`030000003oool0oooo0<D0oooo00<000000?ooo`3oool0303oool00=T0oooo00<000000?oo
o`3oool0a@3oool00`000000oooo0?ooo`0<0?ooo`00f@3oool00`000000oooo0?ooo`360?ooo`03
0000003oool0oooo00/0oooo003I0?ooo`030000003oool0oooo0<L0oooo00<000000?ooo`3oool0
2P3oool00=T0oooo00<000000?ooo`3oool0a`3oool00`000000oooo0?ooo`0:0?ooo`00f@3oool0
0`000000oooo0?ooo`380?ooo`030000003oool0oooo00T0oooo003I0?ooo`030000003oool0oooo
0=@0oooo003I0?ooo`<00000e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<0
00000?ooo`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool00=T0oooo00<000000?oo
o`3oool0e03oool00=T0oooo00<000000?ooo`3oool0e03oool00001\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.06891, -0.889847, \
0.00737444, 0.0100405}}]
}, Closed]],

Cell[TextData[{
  "The simplest way to add error bars to the data points is to use ",
  ButtonBox["ErrorListPlot",
    ButtonStyle->"RefGuideLink"],
  " with data of the form ",
  Cell[BoxData[
      \(TraditionalForm\`{{x\_i, 
          y\_i, \[Delta]\[VeryThinSpace]y\_i}, \[CenterEllipsis]}\)]],
  ".  Here we assume that ",
  Cell[BoxData[
      \(TraditionalForm\`\[Delta]\[VeryThinSpace]y\_i = \(\[Sigma] = 
          0.1\)\)]],
  " for all points."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(data3 = 
      Transpose[{xdata, ydata, Table[0.1, {Length[ydata]}]}]\)], "Input"],

Cell[BoxData[
    \({{\(-0.9`\), \(-0.6772868307328634`\), 
        0.1`}, {\(-0.8`\), \(-0.3986379615930573`\), 
        0.1`}, {\(-0.7`\), \(-0.2813480696442982`\), 
        0.1`}, {\(-0.6`\), \(-0.22795405728072682`\), 
        0.1`}, {\(-0.5`\), \(-0.10437189722648448`\), 0.1`}, {\(-0.4`\), 
        0.016693822902229774`, 0.1`}, {\(-0.3`\), \(-0.10530884161805215`\), 
        0.1`}, {\(-0.19999999999999998`\), \(-0.07925205289484344`\), 
        0.1`}, {\(-0.09999999999999998`\), 0.020268861116419686`, 
        0.1`}, {2.7755575615628914`*^-17, 0.00414321051343284`, 
        0.1`}, {0.10000000000000003`, 0.05332116872243335`, 
        0.1`}, {0.20000000000000004`, \(-0.1686324585313897`\), 
        0.1`}, {0.30000000000000004`, \(-0.15506826773175045`\), 
        0.1`}, {0.4`, \(-0.020024002012701173`\), 0.1`}, {0.5`, 
        0.11556364877402414`, 0.1`}, {0.6000000000000001`, 
        0.28259794010736183`, 0.1`}, {0.7000000000000001`, 
        0.28132821748077974`, 0.1`}, {0.8`, 0.6203349609727172`, 
        0.1`}, {0.9`, 0.7739211493057384`, 0.1`}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[ErrorListPlot[data3], Plot[f[x], {x, \(-1\), 1}], 
        PlotRange \[Rule] All];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.30345 0.343131 [
[.02381 .29095 -9.34375 -15.6875 ]
[.02381 .29095 9.34375 0 ]
[.2619 .29095 -13.8438 -15.6875 ]
[.2619 .29095 13.8438 0 ]
[.7381 .29095 -9.5 -15.6875 ]
[.7381 .29095 9.5 0 ]
[.97619 .29095 -5 -15.6875 ]
[.97619 .29095 5 0 ]
[.4875 .0461 -33.6875 -7.84375 ]
[.4875 .0461 0 7.84375 ]
[.4875 .13188 -27.6875 -7.84375 ]
[.4875 .13188 0 7.84375 ]
[.4875 .21767 -33.6875 -7.84375 ]
[.4875 .21767 0 7.84375 ]
[.4875 .38923 -25 -7.84375 ]
[.4875 .38923 0 7.84375 ]
[.4875 .47502 -19 -7.84375 ]
[.4875 .47502 0 7.84375 ]
[.4875 .5608 -25 -7.84375 ]
[.4875 .5608 0 7.84375 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .30345 m
.02381 .3097 L
s
gsave
.02381 .29095 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 .30345 m
.2619 .3097 L
s
gsave
.2619 .29095 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.7381 .30345 m
.7381 .3097 L
s
gsave
.7381 .29095 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 .30345 m
.97619 .3097 L
s
gsave
.97619 .29095 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 .30345 m
.07143 .3072 L
s
.11905 .30345 m
.11905 .3072 L
s
.16667 .30345 m
.16667 .3072 L
s
.21429 .30345 m
.21429 .3072 L
s
.30952 .30345 m
.30952 .3072 L
s
.35714 .30345 m
.35714 .3072 L
s
.40476 .30345 m
.40476 .3072 L
s
.45238 .30345 m
.45238 .3072 L
s
.54762 .30345 m
.54762 .3072 L
s
.59524 .30345 m
.59524 .3072 L
s
.64286 .30345 m
.64286 .3072 L
s
.69048 .30345 m
.69048 .3072 L
s
.78571 .30345 m
.78571 .3072 L
s
.83333 .30345 m
.83333 .3072 L
s
.88095 .30345 m
.88095 .3072 L
s
.92857 .30345 m
.92857 .3072 L
s
.25 Mabswid
0 .30345 m
1 .30345 L
s
.5 .0461 m
.50625 .0461 L
s
gsave
.4875 .0461 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.5 .13188 m
.50625 .13188 L
s
gsave
.4875 .13188 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .21767 m
.50625 .21767 L
s
gsave
.4875 .21767 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.5 .38923 m
.50625 .38923 L
s
gsave
.4875 .38923 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
.5 .47502 m
.50625 .47502 L
s
gsave
.4875 .47502 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 .5608 m
.50625 .5608 L
s
gsave
.4875 .5608 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.125 Mabswid
.5 .06326 m
.50375 .06326 L
s
.5 .08041 m
.50375 .08041 L
s
.5 .09757 m
.50375 .09757 L
s
.5 .11473 m
.50375 .11473 L
s
.5 .14904 m
.50375 .14904 L
s
.5 .1662 m
.50375 .1662 L
s
.5 .18335 m
.50375 .18335 L
s
.5 .20051 m
.50375 .20051 L
s
.5 .23482 m
.50375 .23482 L
s
.5 .25198 m
.50375 .25198 L
s
.5 .26914 m
.50375 .26914 L
s
.5 .28629 m
.50375 .28629 L
s
.5 .32061 m
.50375 .32061 L
s
.5 .33776 m
.50375 .33776 L
s
.5 .35492 m
.50375 .35492 L
s
.5 .37208 m
.50375 .37208 L
s
.5 .40639 m
.50375 .40639 L
s
.5 .42355 m
.50375 .42355 L
s
.5 .4407 m
.50375 .4407 L
s
.5 .45786 m
.50375 .45786 L
s
.5 .49217 m
.50375 .49217 L
s
.5 .50933 m
.50375 .50933 L
s
.5 .52648 m
.50375 .52648 L
s
.5 .54364 m
.50375 .54364 L
s
.5 .02895 m
.50375 .02895 L
s
.5 .01179 m
.50375 .01179 L
s
.5 .57795 m
.50375 .57795 L
s
.5 .59511 m
.50375 .59511 L
s
.5 .61227 m
.50375 .61227 L
s
.25 Mabswid
.5 0 m
.5 .61803 L
s
.002 w
.07143 .03674 m
.07143 .10536 L
s
.015 w
.07143 .07105 Mdot
.002 w
.11905 .13235 m
.11905 .20098 L
s
.015 w
.11905 .16666 Mdot
.002 w
.16667 .1726 m
.16667 .24122 L
s
.015 w
.16667 .20691 Mdot
.002 w
.21429 .19092 m
.21429 .25954 L
s
.015 w
.21429 .22523 Mdot
.002 w
.2619 .23332 m
.2619 .30195 L
s
.015 w
.2619 .26764 Mdot
.002 w
.30952 .27486 m
.30952 .34349 L
s
.015 w
.30952 .30918 Mdot
.002 w
.35714 .233 m
.35714 .30163 L
s
.015 w
.35714 .26731 Mdot
.002 w
.40476 .24194 m
.40476 .31057 L
s
.015 w
.40476 .27626 Mdot
.002 w
.45238 .27609 m
.45238 .34472 L
s
.015 w
.45238 .3104 Mdot
.002 w
.5 .27056 m
.5 .33918 L
s
.015 w
.5 .30487 Mdot
.002 w
.54762 .28743 m
.54762 .35606 L
s
.015 w
.54762 .32175 Mdot
.002 w
.59524 .21127 m
.59524 .2799 L
s
.015 w
.59524 .24559 Mdot
.002 w
.64286 .21593 m
.64286 .28455 L
s
.015 w
.64286 .25024 Mdot
.002 w
.69048 .26227 m
.69048 .33089 L
s
.015 w
.69048 .29658 Mdot
.002 w
.7381 .30879 m
.7381 .37742 L
s
.015 w
.7381 .3431 Mdot
.002 w
.78571 .3661 m
.78571 .43473 L
s
.015 w
.78571 .40042 Mdot
.002 w
.83333 .36567 m
.83333 .4343 L
s
.015 w
.83333 .39998 Mdot
.002 w
.88095 .48199 m
.88095 .55062 L
s
.015 w
.88095 .51631 Mdot
.002 w
.92857 .53469 m
.92857 .60332 L
s
.015 w
.92857 .56901 Mdot
.5 Mabswid
.02381 .01472 m
.06244 .07315 L
.10458 .12879 L
.14415 .17322 L
.18221 .20887 L
.22272 .23947 L
.26171 .26222 L
.28158 .27149 L
.30316 .2799 L
.32216 .28599 L
.34309 .29139 L
.36292 .29538 L
.38395 .29853 L
.40336 .3006 L
.41426 .30146 L
.42424 .30207 L
.43408 .30254 L
.44322 .30287 L
.45325 .30313 L
.45886 .30323 L
.46395 .3033 L
.46882 .30335 L
.47149 .30338 L
.47397 .30339 L
.47837 .30342 L
.4807 .30343 L
.48319 .30343 L
.48587 .30344 L
.48739 .30344 L
.48879 .30345 L
.49001 .30345 L
.49136 .30345 L
.49203 .30345 L
.49276 .30345 L
.49408 .30345 L
.49529 .30345 L
.49661 .30345 L
.49727 .30345 L
.49798 .30345 L
.49926 .30345 L
.50054 .30345 L
.50175 .30345 L
.50284 .30345 L
.50401 .30345 L
.50525 .30345 L
.50637 .30345 L
.50767 .30345 L
.5089 .30345 L
.51004 .30345 L
.51109 .30345 L
Mistroke
.51348 .30346 L
.51484 .30346 L
.51613 .30346 L
.51852 .30347 L
.5211 .30348 L
.52389 .30349 L
.52923 .30353 L
.53191 .30355 L
.53486 .30358 L
.54007 .30365 L
.54493 .30374 L
.54996 .30385 L
.55541 .30399 L
.56533 .30433 L
.57498 .30478 L
.58389 .30532 L
.6026 .30685 L
.61324 .30802 L
.62321 .30933 L
.64343 .31269 L
.66195 .31669 L
.68252 .3223 L
.70163 .32873 L
.73979 .34543 L
.77889 .3685 L
.82044 .40029 L
.86047 .43848 L
.89899 .48249 L
.93996 .53718 L
.97619 .59218 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00
f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool0
0`000000oooo0?ooo`3D0?ooo`00f@3oool300000=@0oooo000>0?ooo`030000003oool0oooo0<P0
oooo00<000000?ooo`3oool0e03oool000l0oooo00<000000?ooo`3oool0a`3oool00`000000oooo
0?ooo`3D0?ooo`003`3oool00`000000oooo0?ooo`370?ooo`030000003oool0oooo0=@0oooo000@
0?ooo`030000003oool0oooo0<H0oooo00<000000?ooo`3oool0e03oool00140oooo00<000000?oo
o`3oool0a@3oool00`000000oooo0?ooo`3D0?ooo`004@3oool00`000000oooo0?ooo`350?ooo`03
0000003oool0oooo0=@0oooo000B0?ooo`030000003oool0oooo0<@0oooo0`00003D0?ooo`004P3o
ool00`000000oooo0?ooo`340?ooo`030000003oool0oooo0=@0oooo000C0?ooo`030000003oool0
oooo09d0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`030?ooo`030?oo_`1d
B9`0gooo00D0oooo00H0omnL04P00000000000000018W03Oool;0?ooo`030000003oool0oooo0=@0
oooo000D0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0S@3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol0103oool00`3OW4P0B9cO0?ooo`040?oo
o`060?noM0000000M;oo0?ooo`3ogi`0B7Bo2`3oool00`000000oooo0?ooo`3D0?ooo`00503oool0
0`000000oooo0?ooo`0;0?ooo`030000003oool0oooo08`0oooo00<0oooO09a80018W=l00P3oool0
0`3o_g@0001809cOo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<0okmd04RLg`3oool02@3oool0
0`000000oooo0?ooo`3D0?ooo`005@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
08`0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?noM000M;l0
oooo00P0oooo00<0oono07@0M02oool02@3oool00`000000oooo0?ooo`3D0?ooo`005P3oool00`00
0000oooo0?ooo`090?ooo`030000003oool0oooo07l0oooo00<0oooO09a8000000001`0000001P18
W=l0oooo0?ooo`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol0203oool00`3ooml0W4Q809cO
o`080?ooo`030?oog`2LB000M;oo00T0oooo00<000000?ooo`3oool0e03oool001H0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`2=0?ooo`030;md0000B9`0gooo0080oooo00<0omnL
04P00018W=l02@3oool00`3ogi`0B7Bo0?ooo`070?ooo`030?oo_`1d0000B9cO00T0oooo1000003C
0?ooo`005`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo08d0oooo00<0gia80000
B02Lgol00P3oool00`3ooml0W4P0001d_`0:0?ooo`030;md001d_ol0oooo00H0oooo00<0okmd0000
001d_ol02@3oool00`000000oooo0?ooo`3D0?ooo`00603oool00`000000oooo0?ooo`070?ooo`03
0000003oool0oooo08d0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`090?ooo`03
0?oOW018B9`0gooo00<0oooo00H0omoO0=ooo`3o_g@000000000M02oool90?ooo`030000003oool0
oooo0=@0oooo000H0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0S@3oool0103o
oml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00X0oooo00<0gia804RLg`3oool00P3oool0
1@3ookl0M0000000000004P0W=oo00X0oooo00<000000?ooo`3oool0e03oool001T0oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`2>0?ooo`030?oOW0180000M;oo0080oooo00<0gia8
001d_`3oool01@3oool00`3ooml0gmoo0?ooo`020?ooo`030?oo_`1d07@0_ooo00<0oooo00<0gibL
0;ooo`3oool02`3oool00`000000oooo0?ooo`3D0?ooo`006P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo08l0oooo00H0gia80000M02oool0oooO09a80018W=l70?ooo`030;mdM02o
ool0oooo0080oooo00<0okmd001d_`3oool00P3oool00`3ogi`0M9cO0?ooo`0;0?ooo`030000003o
ool0oooo0=@0oooo000J0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0T03oool0
103OW4P0000000000000M;l80?ooo`030?oo_`1d0000000000<0000000<0001809cOo`3oool00P3o
ool01@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3oool0e03oool001/0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo0=@0oooo000K
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0]03oool300000=@0oooo000L0?oo
o`030000003oool0oooo0080oooo0P00002f0?ooo`030000003oool0oooo0=@0oooo000M0?ooo`03
0000003oool0oooo00@00000]@3oool00`000000oooo0?ooo`3D0?ooo`007@3oool00`000000oooo
0000000500000;@0oooo00<000000?ooo`3oool0e03oool001h0oooo1`00002d0?ooo`030000003o
ool0oooo0=@0oooo000O0?ooo`D00000]@3oool00`000000oooo0?ooo`3D0?ooo`007`3oool01000
0000oooo00000000002f0?ooo`030000003oool0oooo0=@0oooo000P0?ooo`030000003oool00000
0;H0oooo0`00003D0?ooo`008@3oool200000;H0oooo00<000000?ooo`3oool0e03oool00280oooo
00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`3D0?ooo`008P3oool00`000000oooo0?oo
o`2d0?ooo`030000003oool0oooo0=@0oooo000R0?ooo`800000]@3oool00`000000oooo0?ooo`3D
0?ooo`008P3oool00`000000oooo0000002d0?ooo`030000003oool0oooo0=@0oooo000R0?ooo`04
0000003oool0oooo00000;<0oooo00<000000?ooo`3oool0e03oool00280oooo00@000000?ooo`3o
ool00000/`3oool300000=@0oooo000R0?ooo`050000003oool0oooo0?ooo`000000/P3oool00`00
0000oooo0?ooo`3D0?ooo`008P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0:l0
oooo00<000000?ooo`3oool0e03oool00280oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`2^0?ooo`030000003oool0oooo0=@0oooo000R0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0[P3oool00`000000oooo0?ooo`3D0?ooo`00:@3oool00`000000oooo0?ooo`2]
0?ooo`030000003oool0oooo0=@0oooo000Z0?ooo`030000003oool0oooo0:`0oooo00<000000?oo
o`3oool0e03oool002/0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`3D0?ooo`00
:`3oool00`000000oooo0?ooo`2[0?ooo`<00000e03oool002`0oooo00<000000?ooo`3oool0ZP3o
ool00`000000oooo0?ooo`3D0?ooo`00;@3oool00`000000oooo0?ooo`2<0?ooo`040;md00000000
000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=ooo`/0
oooo00<000000?ooo`3oool0e03oool002h0oooo00<000000?ooo`3oool0RP3oool01P3OW4P007Bo
0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0
omnL04Qd_`/0oooo00<000000?ooo`3oool0e03oool002h0oooo00<000000?ooo`3oool0R@3oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool02@3oool0
0`000000oooo0?ooo`3D0?ooo`00;`3oool00`000000oooo0?ooo`280?ooo`030?oog`2LB000B9cO
0080oooo00@0oooO09a80000B9`0gooo2P3oool00`3ookl0M01d0;ooo`090?ooo`030000003oool0
oooo0=@0oooo000`0?ooo`030000003oool0oooo07X0oooo00<0oooO09a8000000001`0000001P18
W=l0oooo0?ooo`3ooml0W4P0001d_`<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Bo
o`090?ooo`030000003oool0oooo0=@0oooo000a0?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0PP3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo00<0oono
07@00018W=l02@3oool400000=<0oooo000b0?ooo`050000003oool0oooo0?ooo`000000Q03oool0
0`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00X0oooo00<0okmd0000001d_ol02@3oool0
0`000000oooo0?ooo`3D0?ooo`00<`3oool010000000oooo0?ooo`0000240?ooo`030?oOW0180000
B9cO00<0oooo00<0_g@00018W03Oool01P3oool01P3ogml0gooo0?noM0000000001d0;ooo`T0oooo
00<000000?ooo`3oool0e03oool003@0oooo00<000000?ooo`000000Q03oool0103ooml0W4P00018
W03Oool20?ooo`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000001809cOo`0:0?ooo`03
0000003oool0oooo0=@0oooo000d0?ooo`030000003oool0000008D0oooo00<0omnL04P0001d_ol0
0P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo00/0oooo00<000000?ooo`3oool0
e03oool003D0oooo0P0000260?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0
M9cO0?ooo`0;0?ooo`030000003oool0oooo0=@0oooo000f0?ooo`030000003oool0oooo08D0oooo
00@0gia800000000000007Bo2P3oool01@2oM0000000000000000000B9cO00P0oooo00<000000?oo
o`3oool0e03oool003H0oooo0P00002Q0?ooo`<00000e03oool003H0oooo00<000000?ooo`000000
X03oool00`000000oooo0?ooo`3D0?ooo`00=P3oool010000000oooo0?ooo`00002O0?ooo`030000
003oool0oooo0=@0oooo000f0?ooo`050000003oool0oooo0?ooo`000000WP3oool00`000000oooo
0?ooo`3D0?ooo`00=P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo09/0oooo00<0
00000?ooo`3oool0e03oool003H0oooo0P0000040?ooo`030000003oool0oooo09X0oooo00<00000
0?ooo`3oool0e03oool003D0oooo100000040?ooo`030000003oool0oooo09T0oooo00<000000?oo
o`3oool0e03oool003@0oooo1P0000030?ooo`030000003oool0oooo09T0oooo00<000000?ooo`3o
ool0e03oool003@0oooo1P0000040?ooo`030000003oool0oooo09P0oooo0`00003D0?ooo`00=@3o
ool4000000H0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`3D0?ooo`00=P3oool2
000000P0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`2;0?ooo`030000003oool0
oooo0=@0oooo000f0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`00
0000oooo0?ooo`2;0?ooo`030000003oool0oooo0=@0oooo000f0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo0=@0
oooo000f0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01@3oool00`000000oooo
0?ooo`2;0?ooo`030000003oool0oooo0=@0oooo000f0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`2;0?ooo`030000003oool0oooo0=@0oooo000f
0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2;
0?ooo`<00000e03oool003H0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo08/0oooo00<000000?ooo`3oool0e03oool003H0oooo00<000000?ooo`3o
ool03P3oool01@000000oooo0?ooo`3oool0000008d0oooo00<000000?ooo`3oool0e03oool003H0
oooo00<000000?ooo`3oool03`3oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo
07L0oooo00<000000?ooo`3oool0e03oool003H0oooo00<000000?ooo`3oool0403oool3000001<0
oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`3D0?ooo`00=P3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0M`3oool00`000000oooo
0?ooo`3D0?ooo`00=P3oool00`000000oooo0?ooo`0B0?ooo`8000004P3oool00`000000oooo0?oo
o`1g0?ooo`030000003oool0oooo0=@0oooo000f0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`0000004@3oool00`000000oooo0?ooo`1g0?ooo`<00000e03oool004X0oooo0P0000020?oo
o`030000003oool0oooo00h0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`3D0?oo
o`00B@3oool400000080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`1g0?ooo`03
0000003oool0oooo0=@0oooo00180?ooo`H000000P3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo0540oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`03
0=nLB0000000000000<0000000D0B9cO0?ooo`3oool0omnL04P00002000000030018W03Oool0oooo
00X0oooo00<000000?ooo`3oool0e03oool004P0oooo1P0000030?ooo`030000003oool0oooo00/0
oooo00<000000?ooo`3oool0D03oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL
04P0001d_ol00`3oool00`3o_g@007Bo0?ooo`020?ooo`090;mdM02oool0oooo0?noM0000000M;oo
0?ooo`3ogi`0B7Bo00/0oooo00<000000?ooo`3oool0e03oool004T0oooo100000050?ooo`030000
003oool0oooo00X0oooo00<000000?ooo`3oool0C`3oool00`3ooml0W4P004RLg`020?ooo`030?no
M00004P0W=oo00P0oooo00<0okmd001d_`3oool02@3oool00`3o_g@0B9cO0?ooo`090?ooo`030000
003oool0oooo02H0oooo00<000000?ooo`3oool0Z`3oool004X0oooo0P0000070?ooo`8000002P3o
ool00`000000oooo0?ooo`1?0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo
203oool00`3o_g@007Bo0?ooo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<000000?ooo`3oool0
9P3oool00`000000oooo0?ooo`2[0?ooo`00B`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0@P3oool00`3ooml0W4P0000000070000000604RLg`3o
ool0oooo0?oog`2LB00007Bo0`3oool00`3OW4P0001809cOo`090?ooo`030?noM018W=l0oooo00L0
oooo00<0oooO09a8001d_ol02@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool0U`3oool004/0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`050?ooo`800000DP3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo
00<0gia804RLg`3oool01P3oool00`3ookl0M00004RLg`090?ooo`@000009@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo09L0oooo001;0?ooo`030000003oool0oooo00T0oooo0P00
00040?ooo`@00000D@3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00/0oooo00<0
_g@007Boo`3oool01@3oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`2G0?ooo`00B`3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00H00000D03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0
gooo00X0oooo00<0omnL04P0M02oool00P3oool01P3ogml0gooo0?noM0000000001d0;ooo`T0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo09L0oooo
001;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`0000001@00001@0?ooo`040?oog`2L
B00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3ooml0W4P007Boo`020?ooo`050?oo
_`1d000000000000B02Lgol02P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool0U`3oool004/0oooo00<000000?ooo`3oool03@3oool600000580oooo
00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo_`1d0000B9cO00<0oooo
00<0gibL0;ooo`3oool02`3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0U`3oool004/0oooo00<000000?ooo`3oool03`3oool3000005@0oooo00H0
gia80000M02oool0oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cOo`03
0?ooo`030?oOW01dW=l0oooo00/0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo09L0oooo001;0?ooo`030000003oool0oooo0100oooo0P00000C0?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`3oool05P3oool0103OW4P0000000000000M;l9
0?ooo`050?oOW018000000000000001d_ol01@3oool01@2oM0000000000000000000B9cO00P0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo09L0oooo
001;0?ooo`030000003oool0oooo0140oooo0P00000B0?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool0>P3oool3000002H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`2G0?ooo`00B`3oool00`000000oooo0?ooo`0A0?ooo`030000003oool000000140oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo02H0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`2G0?ooo`00B`3oool00`000000oooo0?ooo`0A0?oo
o`040000003oool0oooo00000100oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0j
0?ooo`030000003oool0oooo02D0oooo0P00000C0?ooo`030000003oool0oooo09L0oooo001;0?oo
o`030000003oool0oooo0140oooo00@000000?ooo`3oool0oooo0P00000>0?ooo`030000003oool0
oooo02H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0
oooo02@0oooo1000000B0?ooo`030000003oool0oooo09L0oooo001O0?ooo`030000003oool0oooo
00<0oooo0P00000<0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`0V0?ooo`030000003oool0oooo02<0oooo1P00000@0?ooo`800000V@3oool005l0
oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02H0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02<0
oooo1P00000?0?ooo`@00000V03oool005l0oooo00<000000?ooo`3oool01P3oool2000000T0oooo
00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo
00<000000?ooo`3oool0903oool4000000l0oooo1P00002G0?ooo`00403oool00`3ookl0M0000000
00020000000304RLg`3oool0oooo04L0oooo00<000000?ooo`3oool0203oool2000000@0oooo00@0
_g@000000000000000000`3oool00`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000
B9`0gooo6P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo02H0oooo0`00000U0?oo
o`800000403oool6000001`0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`02
0?ooo`060?oOW0180000000000000000B9`0goooF@3oool00`3ookl0M000000000020000000304RL
g`3oool0oooo00P0oooo000B0?ooo`030?noM0000000M;oo04X0oooo00<000000?ooo`3oool02P3o
ool2000000070?ooo`3OW4P007Bo0?ooo`3ogi`0000004RLg`020?ooo`030?oOW0180000M;oo0080
oooo00H0okmd0000001d_ol0oooo0?oOW018M;lJ0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo00l0oooo1000000L
0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?no
M0000000M;oo0?ooo`3ogi`0B7BoF`3oool00`3o_g@0000007Boo`0;0?ooo`004P3oool00`3ooml0
W4P0001d_`1:0?ooo`030000003oool0oooo00`0oooo00<0000009a80018W=l00P3oool00`000000
001809cOo`0;0?ooo`030?noM018W=l0oooo01P0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0403oool2000001`0
oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?ooo`030?noM018W=l0oooo05T0
oooo00<0oooO09a80000M;l02`3oool001<0oooo00<0gia80018W03Oool0B@3oool00`000000oooo
0?ooo`0<0?ooo`030?oog`00000000000080oooo00@0000009a80000B9`0gooo2P3oool00`3ookl0
M01d0;ooo`0H0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool09P3oool00`000000
oooo0?ooo`0V0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool06P3oool00`3ooml0
W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono07@0M02oool0FP3oool00`3O
W4P004RL0=ooo`0:0?ooo`000`3oool00`3ooml0W4P0000000070000000304RLg`3oool0oooo00<0
oooo00<0omnL04P0B02Lgol0B`3oool00`3ooml0W4P0000000070000000604RLg`3oool0oooo0?oo
g`2LB00007Bo0`0000000`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo01L0oooo0P00000C
0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo01X0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0:
0?ooo`030?oog`2LB000M;oo05X0oooo00<0omnL04P0B02Lgol02P3oool001<0oooo00<0oono07@0
001d_ol0F@3oool0102oM00004RL0=ooo`3oool20000000304P00018W=l0oooo00T0oooo00<0oono
07@00018W=l05P3oool400000180oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0V
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`07
0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02P3oool00`3ookl0M00004RLg`1J
0?ooo`030?oo_`1d0000M;oo00X0oooo000D0?ooo`030;md0000M;l0oooo05P0oooo00<0gia80000
B02Lgol01@00000:0?ooo`030?noM0000000M;oo01D0oooo1P00000A0?ooo`030000003oool0oooo
02H0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool01`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo
00X0oooo00<0okmd0000001d_ol0F`3oool00`2oM00007Bo0?ooo`090?ooo`00503oool00`3OW4P0
04RL0=ooo`1H0?ooo`030?oOW0180000000000L000001@3oool01P3ogml0gooo0?noM0000000001d
0;oooaD0oooo1P00000A0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool09P3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool0
0`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@000000000
M02ooomK0?ooo`030=nLB000B9`0gooo00T0oooo000D0?ooo`030?oOW01804P0W=oo05P0oooo00<0
oooO09a8000000001@000000103Oool0oooo0000000000030?ooo`050?oo_`1d000000000000B02L
gol05`3oool400000180oooo0P00000W0?ooo`<000009P3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool01`3oool0103ooml0W4P00018W03Oool20?oo
o`030=nLB000B9`0gooo00H0oooo00D0oono07@000000000001809cOo`1L0?ooo`030?oOW01804P0
W=oo00T0oooo000D0?ooo`030?oo_`1d0000M;oo05T0oooo00<0omnL04P0000000000`0000000`00
M;l0oooo0?ooo`020?ooo`<0000000<0oooo0=nLW02oool06P3oool200000180oooo1000000V0?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00P0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`070?oo
o`030=nLW02oool0oooo05d0oooo00<0oono07@0001d_ol02@3oool001D0oooo00<0_g@0001d_`3o
ool0F@3oool01P3OW4P0001d000000000000W4P004RLg`L0oooo0`0000070?ooo`030000003oool0
oooo0140oooo00<000000?ooo`3oool03`3oool600000140oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`090?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO
0?ooo`1N0?ooo`030;md0000M;l0oooo00P0oooo000C0?ooo`050?oo_`1d000000000000M02oool0
FP3oool0103OW4P0000000000000M;l:0?ooo`050;md00000000000000000018W=l00`3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo00l0oooo1P00000A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool02P3oool0103OW4P0000000000000M;l:0?ooo`050;md00000000
000000000018W=l0F@3oool01@3ookl0M0000000000007@0_ooo00P0oooo001c0?ooo`030000003o
ool0oooo00h0oooo0`0000000`3oool000000?ooo`0B0?ooo`030000003oool0oooo0100oooo1000
000B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool09P3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0P`3oool007<0oooo00<00000
0?ooo`3oool04@3oool2000001<0oooo00<000000?ooo`3oool04@3oool2000001<0oooo00<00000
0?ooo`3oool04@3oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool0P`3oool007<0oooo00<000000?ooo`3oool04P3oool400000100oooo00<000000?oo
o`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool0>P3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo08<0oooo001c0?ooo`03
0000003oool0oooo0180oooo00@000000?ooo`3oool0oooo1000000<0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0A0?ooo`<000004@3oool00`00
0000oooo0?ooo`0j0?ooo`030000003oool0oooo08<0oooo001c0?ooo`030000003oool0oooo0180
oooo00<000000?ooo`3oool01@3oool4000000P0oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0j0?ooo`030000003oool0oooo08<0oooo001c0?ooo`030000003oool0oooo0180oooo
00<000000?ooo`3oool02@3oool4000000@0oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0j0?ooo`800000Q03oool007<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0=0?ooo`D000004`3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0i0?ooo`@00000P`3oool007<0oooo00<00000
0?ooo`3oool04P3oool00`000000oooo0?ooo`0A0?ooo`L000003@3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0h
0?ooo`H00000PP3oool007<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00@0oooo2@0000040?ooo`030000003oool0oooo0180oooo00<000000?oo
o`3oool0403oool2000001<0oooo00<000000?ooo`3oool0>03oool600000880oooo001c0?ooo`03
0000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0=0?ooo`/0
00003P3oool00`000000oooo0?ooo`0?0?ooo`@000004P3oool00`000000oooo0?ooo`0i0?ooo`@0
0000P`3oool000<0ooooo`00002]00000040oooo000>0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool04@3oool00`000000oooo0?ooo`0@0?ooo`@000004P3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo0140oooo0P00000@0?ooo`H000004@3oool00`000000oooo0?oo
o`030?ooo`h0000000<0oooo0000003oool04P3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo0180oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`030000003o
ool0oooo00T0oooo000>0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0?0?ooo`H000004@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo0100oooo1000000@0?ooo`@000004P3oool00`000000oooo0?ooo`0A0?ooo`/000002P3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00T0oooo000>0?ooo`030000003oool0oooo
0680oooo00<000000?ooo`3oool03`3oool6000002D0oooo00<000000?ooo`3oool03`3oool60000
0100oooo0P00000C0?ooo`030000003oool0oooo01`0oooo1P00000H0?ooo`030000003oool0oooo
0180oooo00<000000?ooo`3oool0HP3oool00`000000oooo0?ooo`090?ooo`00QP3oool4000003P0
oooo1P00000A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool08P3oool600000180
oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1^0?ooo`00Q`3oool2000003X0oooo
1000000B0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0:03oool4000000h0oooo
00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`1^0?ooo`00R03oool00`000000oooo0?oo
o`0i0?ooo`8000004`3oool00`000000oooo0?ooo`0A0?ooo`800000;@3oool5000000T0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`1^0?ooo`00R03oool00`000000oooo0?ooo`0j
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0403oool400000340oooo10000005
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0KP3oool008P0oooo00<000000?oo
o`3oool0>P3oool00`000000oooo0?ooo`0A0?ooo`<000003`3oool6000003@0oooo100000000`3o
ool000000?ooo`0C0?ooo`030000003oool0oooo06h0oooo00280?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0?0?ooo`H00000>03oool2000001@0
oooo00<000000?ooo`3oool0KP3oool008P0oooo00<000000?ooo`3oool0>P3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo0100oooo1000000j0?ooo`@000004@3oool00`000000oooo
0?ooo`1^0?ooo`00R03oool00`000000oooo0?ooo`0j0?ooo`030000003oool0oooo0140oooo00<0
00000?ooo`3oool04@3oool2000003/0oooo00@000000?ooo`3oool0oooo0`00000>0?ooo`030000
003oool0oooo06h0oooo00280?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool0103o
ool2000000`0oooo00<000000?ooo`3oool0KP3oool008P0oooo00<000000?ooo`3oool0>P3oool0
0`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0@`3oool2
000000X0oooo00<000000?ooo`3oool0KP3oool008P0oooo00<000000?ooo`3oool0>P3oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0A@3oool20000
00P0oooo00<000000?ooo`3oool0KP3oool008P0oooo00<000000?ooo`3oool0>P3oool00`000000
oooo0?ooo`0A0?ooo`<000004@3oool00`000000oooo0?ooo`170?ooo`<00000103oool200000700
oooo00280?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool04@3oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo04X0oooo0P0000000`3oool0000000000002000006l0oooo
00280?ooo`030000003oool0oooo03X0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo04`0oooo1P00001^0?ooo`00a@3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0C03oool6000006h0oooo003I0?oo
o`030000003oool0oooo0140oooo00<000000?ooo`3oool0C@3oool5000006h0oooo003I0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool0CP3oool200000080oooo0P00001/0?ooo`00
f@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`1Y0?ooo`00f@3oool300000140oooo00<000000?ooo`3oool0C`3o
ool00`000000oooo0?ooo`030?ooo`800000J@3oool00=T0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00D0oooo0P00001W0?ooo`00f@3oool00`00
0000oooo0?ooo`1S0?ooo`030000003oool0oooo00L0oooo0P00001U0?ooo`00f@3oool00`000000
oooo0?ooo`1S0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0HP3oool00=T0oooo
00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0:0?ooo`800000HP3oool00=T0oooo00<0
00000?ooo`3oool0H`3oool00`000000oooo0?ooo`0<0?ooo`8000000`3oool00`000000oooo0?oo
o`0A0?ooo`030000003oool0oooo04H0oooo003I0?ooo`030000003oool0oooo06<0oooo00<00000
0?ooo`3oool03P3oool010000000oooo0?ooo`00000C0?ooo`030000003oool0oooo04H0oooo003I
0?ooo`<00000H`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0000001<0oooo00<00000
0?ooo`3oool0AP3oool00=T0oooo00<000000?ooo`3oool0H`3oool00`000000oooo0?ooo`0@0?oo
o`8000004`3oool00`000000oooo0?ooo`160?ooo`00f@3oool00`000000oooo0?ooo`1S0?ooo`03
0000003oool0oooo0140oooo0P00000B0?ooo`030000003oool0oooo04H0oooo002c0?ooo`040;md
00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P000000000000300000005
04RLg`3oool0oooo0?oOW01800000P0000000`00B9`0gooo0?ooo`0:0?ooo`030000003oool0oooo
06<0oooo00<000000?ooo`3oool04@3oool00`000000oooo0000000A0?ooo`030000003oool0oooo
04H0oooo002b0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`03
0?ooo`030?noM000M;l0oooo0080oooo00T0_gAd0;ooo`3oool0okmd0000001d_ol0oooo0?oOW018
M;l02`3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo008000003`3oool00`000000
oooo0?ooo`160?ooo`00/@3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo
00<0okmd001d_`3oool02@3oool00`3o_g@0B9cO0?ooo`090?ooo`030000003oool0oooo07L0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo04H0oooo
002a0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo
0?ooo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<000000?ooo`3oool0M`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0AP3oool00;40oooo00<0
oooO09a80000M;l00`3oool00`3OW4P0001809cOo`090?ooo`030?noM018W=l0oooo00L0oooo00<0
oooO09a8001d_ol02@3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool02P3oool00`000000oooo0?ooo`160?ooo`00/P3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0oooo00<0gia804RLg`3oool01P3oool00`3ookl0M00004RLg`09
0?ooo`@00000MP3oool00`000000oooo0?ooo`050?ooo`8000002P3oool00`000000oooo0?ooo`16
0?ooo`00/P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00/0oooo00<0_g@007Bo
o`3oool01@3oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo07L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo04H0oooo002b0?ooo`03
0?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool02P3oool00`3ogi`0B01d0;ooo`020?ooo`06
0?oOg`3Oool0okmd0000000007@0_ooo2@3oool00`000000oooo0?ooo`1f0?ooo`8000002P3oool0
0`000000oooo0?ooo`060?ooo`800000A`3oool00;80oooo00@0oooO09a80000B9`0gooo0P3oool0
0`3OW4P004RL0=ooo`0:0?ooo`030?oog`2LB000M;oo0080oooo00D0oono07@000000000001809cO
o`0:0?ooo`030000003oool0oooo07D0oooo1000000:0?ooo`8000001@3oool4000004H0oooo002c
0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool02P3oool00`3ookl0M00004RLg`03
0?ooo`030=nLW02oool0oooo00/0oooo00<000000?ooo`3oool0M03oool6000000/0oooo00@00000
0?ooo`3oool0oooo1P0000150?ooo`00]03oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo
00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00<0oooo00<0omnL07BLg`3oool02`3oool00`00
0000oooo0?ooo`1d0?ooo`H00000303oool00`000000oooo0?ooo`06000004D0oooo002e0?ooo`04
0=nLB00000000000001d_`T0oooo00D0omnL04P000000000000007Boo`050?ooo`050;md00000000
000000000018W=l0203oool00`000000oooo0?ooo`1e0?ooo`@000003P3oool00`000000oooo0?oo
o`04000004H0oooo003I0?ooo`<00000MP3oool200000100oooo00<000000?ooo`3oool00P000017
0?ooo`00f@3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo00l0oooo00<000000?oo
o`000000B03oool00=T0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0@0?ooo`80
0000B03oool00=T0oooo00<000000?ooo`3oool0M`3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo04H0oooo003I0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool04@3o
ool2000004L0oooo003I0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool04@3oool0
0`000000oooo000000160?ooo`00f@3oool00`000000oooo0?ooo`1g0?ooo`030000003oool0oooo
0140oooo00<000000?ooo`3oool00P0000140?ooo`00f@3oool3000007L0oooo00<000000?ooo`3o
ool04@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0440oooo003I0?ooo`030000
003oool0oooo07L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0400oooo003I0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo03l0oooo003I0?ooo`030000003oool0
oooo07L0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo03h0oooo003I0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool04@3oool00`00
0000oooo0?ooo`060?ooo`030000003oool0oooo03d0oooo003I0?ooo`030000003oool0oooo07L0
oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo03`0
oooo003I0?ooo`030000003oool0oooo09H0oooo00<000000?ooo`3oool0>`3oool00=T0oooo00<0
00000?ooo`3oool0U`3oool00`000000oooo0?ooo`0j0?ooo`00f@3oool3000009P0oooo00<00000
0?ooo`3oool0>@3oool00=T0oooo00<000000?ooo`3oool0V@3oool00`000000oooo0?ooo`0h0?oo
o`00f@3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo03L0oooo003I0?ooo`030000
003oool0oooo09X0oooo00<000000?ooo`3oool0=`3oool00=T0oooo00<000000?ooo`3oool0V`3o
ool00`000000oooo0?ooo`0f0?ooo`00f@3oool00`000000oooo0?ooo`2L0?ooo`030000003oool0
oooo03D0oooo003I0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0=03oool00=T0
oooo0`00002N0?ooo`030000003oool0oooo03<0oooo003I0?ooo`030000003oool0oooo09l0oooo
00<000000?ooo`3oool0<P3oool00=T0oooo00<000000?ooo`3oool0W`3oool00`000000oooo0?oo
o`0b0?ooo`00_03oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo0080oooo00H0
omnL04P00000000000000018W03Oool;0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3o
ool0<@3oool00;/0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo
0080oooo00H0okmd0000001d_ol0oooo0?oOW018M;l;0?ooo`030000003oool0oooo0:40oooo00<0
00000?ooo`3oool0<03oool00;X0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;
0?ooo`030?noM018W=l0oooo00T0oooo00<000000?ooo`3oool0XP3oool00`000000oooo0?ooo`0_
0?ooo`00^P3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`X0oooo00<0oono
07@0M02oool02@3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo02h0oooo002j0?oo
o`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol02P3oool00`3ooml0W4P007Boo`090?oo
o`@00000X`3oool00`000000oooo0?ooo`0]0?ooo`00^`3oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO00X0oooo00<0oono07@00018W=l02@3oool00`000000oooo0?ooo`2T0?ooo`03
0000003oool0oooo02d0oooo002k0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0
2P3oool00`3o_g@0000007Boo`090?ooo`030000003oool0oooo0:D0oooo00<000000?ooo`3oool0
;03oool00;/0oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`060?oOg`3O
ool0okmd0000000007@0_ooo2@3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0:`3oool00;/0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P0
04RL0=ooo`060?ooo`050?oo_`1d000000000000B02Lgol02P3oool00`000000oooo0?ooo`2P0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0:P3oool00;`0oooo00<0omnL04P0001d
_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo00/0oooo00<000000?ooo`3o
ool0X03oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02T0oooo002m0?ooo`060=nL
B00007@0_ooo0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0;0?ooo`030000003oool0oooo
0:00oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0X0?ooo`00_P3oool0103OW4P0
000000000000M;l:0?ooo`050;md00000000000000000018W=l0203oool00`000000oooo0?ooo`2P
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0:03oool00=T0oooo0`00002P0?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool09`3oool00=T0oooo00<000000?ooo`3o
ool0X03oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo02H0oooo003I0?ooo`030000
003oool0oooo0:00oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0U0?ooo`00f@3o
ool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool09@3o
ool00=T0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo02@0oooo003I0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool02`3oool00`00
0000oooo0?ooo`0S0?ooo`00f@3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool08P3oool00=T0oooo0`00002P0?ooo`030000003oool0oooo00d0oooo
00<000000?ooo`3oool08@3oool00=T0oooo00<000000?ooo`3oool0W`3oool2000000l0oooo00<0
00000?ooo`3oool08@3oool00=T0oooo00<000000?ooo`3oool0WP3oool4000000l0oooo00<00000
0?ooo`3oool0803oool00=T0oooo00<000000?ooo`3oool0W@3oool6000000l0oooo00<000000?oo
o`3oool07`3oool00=T0oooo00<000000?ooo`3oool0W@3oool600000100oooo00<000000?ooo`3o
ool07P3oool00=T0oooo00<000000?ooo`3oool0WP3oool400000180oooo00<000000?ooo`3oool0
7@3oool00=T0oooo00<000000?ooo`3oool0W`3oool2000001<0oooo00<000000?ooo`3oool07@3o
ool00=T0oooo0`00002P0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0703oool0
0=T0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo
01/0oooo003I0?ooo`030000003oool0oooo0:00oooo00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`0J0?ooo`00f@3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool06P3oool00=T0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?oo
o`0A0?ooo`050000003oool0oooo0?ooo`0000006`3oool00=T0oooo00<000000?ooo`3oool0X03o
ool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0603o
ool00=T0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool05`3oool00=T0oooo00<000000?ooo`3oool0X03oool00`00
0000oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool05`3oool00=T0
oooo0`00002P0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`0F0?ooo`00f@3oool00`000000oooo0?ooo`2P0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0F0?ooo`00/`3oool0102oM00000000000
001d_ol30?ooo`030?oOW0180000M;oo00<0oooo00<0oono07A8W03Oool01@3oool01P3ogi`0B000
00000000000004RL0=ooo`/0oooo00<000000?ooo`3oool0X03oool00`000000oooo0?ooo`0A0?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool00;80oooo00H0gia8001d_`3o
ool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00@0oooo00<0gia804RLg`3oool0103oool0
1P3o_g@0000007Boo`3oool0omnL04Qd_`/0oooo00<000000?ooo`3oool0X03oool00`000000oooo
0?ooo`0A0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0503oool00;40oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<0
okmd04RLg`3oool02@3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool0503oool00;40oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03O
ool80?ooo`030?noM000M;l0oooo00P0oooo00<0oono07@0M02oool02@3oool00`000000oooo0?oo
o`2d0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool04`3oool00;40oooo00<0oooO
09a80000M;l00`3oool00`3OW4P0001809cOo`080?ooo`030?oog`2LB4P0W=oo00P0oooo00<0oooO
09a8001d_ol02@3oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo00P0oooo00<00000
0?ooo`3oool04P3oool00;80oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`090?oo
o`030?oOW018M;l0oooo00L0oooo00<0oono07@00018W=l02@3oool400000;<0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`0B0?ooo`00/P3oool00`3OW4P0001809cOo`020?ooo`03
0?oog`2LB00007Bo00X0oooo00<0_g@007Boo`3oool01P3oool00`3o_g@0000007Boo`090?ooo`03
0000003oool0oooo0;<0oooo0P00000;0?ooo`030000003oool0oooo0140oooo002b0?ooo`030?oO
W0180000B9cO00<0oooo00<0_g@00018W03Oool02@3oool00`3ogi`0B4RL0=ooo`030?ooo`060?oO
g`3Oool0okmd0000000007@0_ooo2@3oool00`000000oooo0?ooo`2b0?ooo`@000002P3oool00`00
0000oooo0?ooo`0A0?ooo`00/P3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo
00X0oooo00<0gia804RLg`3oool00P3oool01@3ookl0M0000000000004P0W=oo00X0oooo00<00000
0?ooo`3oool0/@3oool6000000X0oooo00<000000?ooo`3oool0403oool00;<0oooo00<0omnL04P0
001d_ol00P3oool00`3OW4P007Bo0?ooo`050?ooo`030?oog`3Ogol0oooo0080oooo00<0oono07@0
M02oool00`3oool00`3OW9`0_ooo0?ooo`0;0?ooo`030000003oool0oooo0;40oooo1P00000;0?oo
o`030000003oool0oooo00l0oooo002d0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3oool0
0`2oM7@0_ooo0?ooo`020?ooo`030?noM000M;l0oooo0080oooo00<0omnL07BLg`3oool02`3oool0
0`000000oooo0?ooo`2b0?ooo`@00000303oool00`000000oooo0?ooo`0?0?ooo`00]@3oool0103O
W4P0000000000000M;l80?ooo`030?oo_`1d0000000000<0000000<0001809cOo`3oool00P3oool0
1@2oM0000000000000000000B9cO00P0oooo00<000000?ooo`3oool0/`3oool2000000h0oooo00<0
00000?ooo`3oool03P3oool00=T0oooo0`00002d0?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool03@3oool00=T0oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo00d0oooo003I0?ooo`030000003oool0oooo0;@0oooo00<000000?ooo`3o
ool03P3oool00`000000oooo0?ooo`0<0?ooo`00f@3oool00`000000oooo0?ooo`2d0?ooo`030000
003oool0oooo00h0oooo00<000000?ooo`3oool0303oool00=T0oooo00<000000?ooo`3oool0]03o
ool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo00/0oooo003I0?ooo`030000003oool0
oooo0;@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0:0?ooo`00f@3oool00`00
0000oooo0?ooo`2d0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02P3oool00=T0
oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo00T0
oooo003I0?ooo`<00000]03oool00`000000oooo0?ooo`0M0?ooo`00f@3oool00`000000oooo0?oo
o`2d0?ooo`030000003oool0oooo01d0oooo003I0?ooo`030000003oool0oooo0;@0oooo00<00000
0?ooo`3oool07@3oool00=T0oooo00<000000?ooo`3oool0]03oool00`000000oooo0?ooo`0M0?oo
o`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3oool00`000000oooo0?ooo`3D0?ooo`00f@3o
ool00`000000oooo0?ooo`3D0?ooo`00f@3oool300000=@0oooo003I0?ooo`030000003oool0oooo
0=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo003I0?ooo`030000003oool0oooo0=@0oooo
0000\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-1.06891, -0.89078, \
0.00737444, 0.0102341}}]
}, Closed]],

Cell[TextData[{
  "Sometime we need to plot several data sets with different symbols or with \
error bars in both variables, but the options offered by ",
  StyleBox["ErrorListPlot", "InlineInput",
    FontWeight->"Bold"],
  " are quite limited.  A more versatile function for plotting data is ",
  ButtonBox["MultipleListPlot",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput"],
  StyleBox["dataset1,dataset2,\[CenterEllipsis], options", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" where each data set should have the form ",
    FontWeight->"Plain"],
  StyleBox["{{point,error},\[CenterEllipsis]}", "InlineInput"],
  StyleBox[" where ",
    FontWeight->"Plain"],
  StyleBox["point", "InlineInput"],
  StyleBox[" is an {",
    FontWeight->"Plain"],
  StyleBox["x,y",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  StyleBox["} pair and ",
    FontWeight->"Plain"],
  StyleBox["error", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" is an ",
    FontWeight->"Plain"],
  ButtonBox["ErrorBar",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" object of the form ",
    FontWeight->"Plain"],
  StyleBox["ErrorBar[", "InlineInput"],
  StyleBox["yerr", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" or ",
    FontWeight->"Plain"],
  StyleBox["ErrorBar[", "InlineInput"],
  StyleBox["xerr,yerr", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[".  Each error is either a single value representing a symmetric \
uncertainty or an interval about zero of the form ",
    FontWeight->"Plain"],
  StyleBox["{negerr,poserr}", "InlineInput"],
  StyleBox[" which can be used for asymmetric error bars.  ",
    FontWeight->"Plain"]
}], "Text"],

Cell[TextData[{
  StyleBox["To illustrate this function we create two data sets based upon \
the same function.  First we construct grids for the independent variable ",
    FontWeight->"Plain"],
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and compute the predicted values for ",
  Cell[BoxData[
      \(TraditionalForm\`y\_i = f[x\_i]\)]],
  ".  We then add random measurement errors based upon the normal \
distribution to both ",
  Cell[BoxData[
      \(TraditionalForm\`x\_i\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`y\_i\)]],
  ".  Different standard deviations are used for each data set.  Next, we \
package the data points and their two-dimensional error bars in the form \
required for ",
  StyleBox["MultipleListPlot", "InlineInput",
    FontWeight->"Bold"],
  ".  Finally, we plot both data sets with different symbols and add the \
theoretical curve."
}], "Text"],

Cell[BoxData[{
    \(\(xdata1 = Range[\(-0.9\), 0.9, 0.2];\)\), "\[IndentingNewLine]", 
    \(\(xdata2 = Range[\(-0.8\), 0.8, 0.2];\)\)}], "Input"],

Cell[BoxData[{
    \(\(ydata1 = 
        f[xdata1] + 
          Table[Random[
              NormalDistribution[0, 0.1]], {Length[
                xdata1]}];\)\), "\[IndentingNewLine]", 
    \(\(ydata2 = 
        f[xdata2] + 
          Table[Random[
              NormalDistribution[0, 0.05]], {Length[xdata2]}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(xdata1 = 
        xdata1\  + \ 
          Table[Random[
              NormalDistribution[0, 0.03]], {Length[
                xdata1]}];\)\), "\[IndentingNewLine]", 
    \(\(xdata2 = 
        xdata2\  + 
          Table[Random[
              NormalDistribution[0, 0.01]], {Length[xdata2]}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(data1 = Transpose[{xdata1, ydata1}];\)\), "\[IndentingNewLine]", 
    \(\(data2 = Transpose[{xdata2, ydata2}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(edata1 = 
        Table[ErrorBar[0.03, 
            0.1], {Length[xdata1]}];\)\), "\[IndentingNewLine]", 
    \(\(edata2 = Table[ErrorBar[0.01, 0.05], {Length[xdata2]}];\)\)}], "Input"],

Cell[BoxData[{
    \(\(set1 = Transpose[{data1, edata1}];\)\), "\[IndentingNewLine]", 
    \(\(set2 = Transpose[{data2, edata2}];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[
        MultipleListPlot[set1, set2, 
          SymbolShape \[Rule] {PlotSymbol[Triangle, 4], 
              PlotSymbol[Diamond, 4, Filled \[Rule] False]}, 
          SymbolStyle \[Rule] {RGBColor[0, 1, 0], RGBColor[1, 0, 0]}], 
        Plot[f[x], {x, \(-1\), 1}], PlotRange \[Rule] All, 
        Frame \[Rule] True, 
        FrameLabel \[Rule] {"\<put xlabel here with units\>", "\<put ylabel \
here with units\>"}, PlotLabel \[Rule] "\<put plot label here\>", 
        GridLines \[Rule] Automatic];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.5 0.476191 0.31113 0.347236 [
[.02381 -0.0125 -9.34375 -15.6875 ]
[.02381 -0.0125 9.34375 0 ]
[.2619 -0.0125 -13.8438 -15.6875 ]
[.2619 -0.0125 13.8438 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.7381 -0.0125 -9.5 -15.6875 ]
[.7381 -0.0125 9.5 0 ]
[.97619 -0.0125 -5 -15.6875 ]
[.97619 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -63.7813 -31.375 ]
[.5 -0.0125 63.7813 -15.6875 ]
[-0.0125 .0507 -33.6875 -7.84375 ]
[-0.0125 .0507 0 7.84375 ]
[-0.0125 .13751 -27.6875 -7.84375 ]
[-0.0125 .13751 0 7.84375 ]
[-0.0125 .22432 -33.6875 -7.84375 ]
[-0.0125 .22432 0 7.84375 ]
[-0.0125 .31113 -10 -7.84375 ]
[-0.0125 .31113 0 7.84375 ]
[-0.0125 .39794 -25 -7.84375 ]
[-0.0125 .39794 0 7.84375 ]
[-0.0125 .48475 -19 -7.84375 ]
[-0.0125 .48475 0 7.84375 ]
[-0.0125 .57156 -25 -7.84375 ]
[-0.0125 .57156 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -49.375 -64.0938 ]
[-0.01875 .30902 -33.6875 64.0938 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -47.625 0 ]
[.5 .63053 47.625 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 0 .5 r
.25 Mabswid
[ ] 0 setdash
.2619 0 m
.2619 .61803 L
s
.5 0 m
.5 .61803 L
s
.7381 0 m
.7381 .61803 L
s
0 .0507 m
1 .0507 L
s
0 .13751 m
1 .13751 L
s
0 .22432 m
1 .22432 L
s
0 .31113 m
1 .31113 L
s
0 .39794 m
1 .39794 L
s
0 .48475 m
1 .48475 L
s
0 .57156 m
1 .57156 L
s
0 g
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -70.3438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.2619 0 m
.2619 .00625 L
s
gsave
.2619 -0.0125 -74.8438 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.7381 0 m
.7381 .00625 L
s
gsave
.7381 -0.0125 -70.5 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
.07143 0 m
.07143 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.16667 0 m
.16667 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.35714 0 m
.35714 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.45238 0 m
.45238 .00375 L
s
.54762 0 m
.54762 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.64286 0 m
.64286 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.83333 0 m
.83333 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.92857 0 m
.92857 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -124.781 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(put) show
83.188 15.813 moveto
(xlabel) show
115.438 15.813 moveto
(here) show
139.625 15.813 moveto
(with) show
163.250 15.813 moveto
(units) show
186.563 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .0507 m
.00625 .0507 L
s
gsave
-0.0125 .0507 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
0 .13751 m
.00625 .13751 L
s
gsave
-0.0125 .13751 -88.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .22432 m
.00625 .22432 L
s
gsave
-0.0125 .22432 -94.6875 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Mathematica1
%%IncludeFont: Mathematica1
/Mathematica1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(-) show
71.688 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
0 .31113 m
.00625 .31113 L
s
gsave
-0.0125 .31113 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .39794 m
.00625 .39794 L
s
gsave
-0.0125 .39794 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.25) show
1.000 setlinewidth
grestore
0 .48475 m
.00625 .48475 L
s
gsave
-0.0125 .48475 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.5) show
1.000 setlinewidth
grestore
0 .57156 m
.00625 .57156 L
s
gsave
-0.0125 .57156 -86 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.75) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .06806 m
.00375 .06806 L
s
0 .08543 m
.00375 .08543 L
s
0 .10279 m
.00375 .10279 L
s
0 .12015 m
.00375 .12015 L
s
0 .15487 m
.00375 .15487 L
s
0 .17224 m
.00375 .17224 L
s
0 .1896 m
.00375 .1896 L
s
0 .20696 m
.00375 .20696 L
s
0 .24168 m
.00375 .24168 L
s
0 .25904 m
.00375 .25904 L
s
0 .27641 m
.00375 .27641 L
s
0 .29377 m
.00375 .29377 L
s
0 .32849 m
.00375 .32849 L
s
0 .34585 m
.00375 .34585 L
s
0 .36322 m
.00375 .36322 L
s
0 .38058 m
.00375 .38058 L
s
0 .4153 m
.00375 .4153 L
s
0 .43266 m
.00375 .43266 L
s
0 .45002 m
.00375 .45002 L
s
0 .46739 m
.00375 .46739 L
s
0 .50211 m
.00375 .50211 L
s
0 .51947 m
.00375 .51947 L
s
0 .53683 m
.00375 .53683 L
s
0 .5542 m
.00375 .5542 L
s
0 .03334 m
.00375 .03334 L
s
0 .01598 m
.00375 .01598 L
s
0 .58892 m
.00375 .58892 L
s
0 .60628 m
.00375 .60628 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -166.625 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
125.094 11.8438 translate
90 rotate
-125.094 -11.8438 translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(put) show
83.813 15.813 moveto
(ylabel) show
116.063 15.813 moveto
(here) show
140.250 15.813 moveto
(with) show
163.875 15.813 moveto
(units) show
187.188 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.2619 .61178 m
.2619 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.7381 .61178 m
.7381 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.07143 .61428 m
.07143 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.16667 .61428 m
.16667 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.35714 .61428 m
.35714 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.45238 .61428 m
.45238 .61803 L
s
.54762 .61428 m
.54762 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.64286 .61428 m
.64286 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.83333 .61428 m
.83333 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.92857 .61428 m
.92857 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .0507 m
1 .0507 L
s
.99375 .13751 m
1 .13751 L
s
.99375 .22432 m
1 .22432 L
s
.99375 .31113 m
1 .31113 L
s
.99375 .39794 m
1 .39794 L
s
.99375 .48475 m
1 .48475 L
s
.99375 .57156 m
1 .57156 L
s
.125 Mabswid
.99625 .06806 m
1 .06806 L
s
.99625 .08543 m
1 .08543 L
s
.99625 .10279 m
1 .10279 L
s
.99625 .12015 m
1 .12015 L
s
.99625 .15487 m
1 .15487 L
s
.99625 .17224 m
1 .17224 L
s
.99625 .1896 m
1 .1896 L
s
.99625 .20696 m
1 .20696 L
s
.99625 .24168 m
1 .24168 L
s
.99625 .25904 m
1 .25904 L
s
.99625 .27641 m
1 .27641 L
s
.99625 .29377 m
1 .29377 L
s
.99625 .32849 m
1 .32849 L
s
.99625 .34585 m
1 .34585 L
s
.99625 .36322 m
1 .36322 L
s
.99625 .38058 m
1 .38058 L
s
.99625 .4153 m
1 .4153 L
s
.99625 .43266 m
1 .43266 L
s
.99625 .45002 m
1 .45002 L
s
.99625 .46739 m
1 .46739 L
s
.99625 .50211 m
1 .50211 L
s
.99625 .51947 m
1 .51947 L
s
.99625 .53683 m
1 .53683 L
s
.99625 .5542 m
1 .5542 L
s
.99625 .03334 m
1 .03334 L
s
.99625 .01598 m
1 .01598 L
s
.99625 .58892 m
1 .58892 L
s
.99625 .60628 m
1 .60628 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
0 .31113 m
1 .31113 L
s
.5 0 m
.5 .61803 L
s
gsave
.5 .63053 -108.625 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
64.313 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(put) show
84.188 15.813 moveto
(plot) show
106.063 15.813 moveto
(label) show
133.063 15.813 moveto
(here) show
154.250 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.001 w
.0836 .04944 m
.05503 .04944 L
s
.0836 .04944 0 1.5 Mabsadd m
.0836 .04944 0 -1.5 Mabsadd L s
.05503 .04944 0 1.5 Mabsadd m
.05503 .04944 0 -1.5 Mabsadd L s
.06932 .08416 m
.06932 .01472 L
s
.06932 .08416 1.5 0 Mabsadd m
.06932 .08416 -1.5 0 Mabsadd L s
.06932 .01472 1.5 0 Mabsadd m
.06932 .01472 -1.5 0 Mabsadd L s
.1686 .20364 m
.14002 .20364 L
s
.1686 .20364 0 1.5 Mabsadd m
.1686 .20364 0 -1.5 Mabsadd L s
.14002 .20364 0 1.5 Mabsadd m
.14002 .20364 0 -1.5 Mabsadd L s
.15431 .23836 m
.15431 .16892 L
s
.15431 .23836 1.5 0 Mabsadd m
.15431 .23836 -1.5 0 Mabsadd L s
.15431 .16892 1.5 0 Mabsadd m
.15431 .16892 -1.5 0 Mabsadd L s
.24895 .27248 m
.22038 .27248 L
s
.24895 .27248 0 1.5 Mabsadd m
.24895 .27248 0 -1.5 Mabsadd L s
.22038 .27248 0 1.5 Mabsadd m
.22038 .27248 0 -1.5 Mabsadd L s
.23467 .30721 m
.23467 .23776 L
s
.23467 .30721 1.5 0 Mabsadd m
.23467 .30721 -1.5 0 Mabsadd L s
.23467 .23776 1.5 0 Mabsadd m
.23467 .23776 -1.5 0 Mabsadd L s
.36203 .3236 m
.33346 .3236 L
s
.36203 .3236 0 1.5 Mabsadd m
.36203 .3236 0 -1.5 Mabsadd L s
.33346 .3236 0 1.5 Mabsadd m
.33346 .3236 0 -1.5 Mabsadd L s
.34775 .35833 m
.34775 .28888 L
s
.34775 .35833 1.5 0 Mabsadd m
.34775 .35833 -1.5 0 Mabsadd L s
.34775 .28888 1.5 0 Mabsadd m
.34775 .28888 -1.5 0 Mabsadd L s
.47185 .29361 m
.44328 .29361 L
s
.47185 .29361 0 1.5 Mabsadd m
.47185 .29361 0 -1.5 Mabsadd L s
.44328 .29361 0 1.5 Mabsadd m
.44328 .29361 0 -1.5 Mabsadd L s
.45757 .32834 m
.45757 .25889 L
s
.45757 .32834 1.5 0 Mabsadd m
.45757 .32834 -1.5 0 Mabsadd L s
.45757 .25889 1.5 0 Mabsadd m
.45757 .25889 -1.5 0 Mabsadd L s
.55151 .28601 m
.52293 .28601 L
s
.55151 .28601 0 1.5 Mabsadd m
.55151 .28601 0 -1.5 Mabsadd L s
.52293 .28601 0 1.5 Mabsadd m
.52293 .28601 0 -1.5 Mabsadd L s
.53722 .32073 m
.53722 .25128 L
s
.53722 .32073 1.5 0 Mabsadd m
.53722 .32073 -1.5 0 Mabsadd L s
.53722 .25128 1.5 0 Mabsadd m
.53722 .25128 -1.5 0 Mabsadd L s
.66305 .32671 m
.63448 .32671 L
s
.66305 .32671 0 1.5 Mabsadd m
.66305 .32671 0 -1.5 Mabsadd L s
.63448 .32671 0 1.5 Mabsadd m
.63448 .32671 0 -1.5 Mabsadd L s
.64876 .36144 m
.64876 .29199 L
s
.64876 .36144 1.5 0 Mabsadd m
.64876 .36144 -1.5 0 Mabsadd L s
.64876 .29199 1.5 0 Mabsadd m
.64876 .29199 -1.5 0 Mabsadd L s
.75568 .34163 m
.72711 .34163 L
s
.75568 .34163 0 1.5 Mabsadd m
.75568 .34163 0 -1.5 Mabsadd L s
.72711 .34163 0 1.5 Mabsadd m
.72711 .34163 0 -1.5 Mabsadd L s
.74139 .37635 m
.74139 .30691 L
s
.74139 .37635 1.5 0 Mabsadd m
.74139 .37635 -1.5 0 Mabsadd L s
.74139 .30691 1.5 0 Mabsadd m
.74139 .30691 -1.5 0 Mabsadd L s
.84557 .44879 m
.817 .44879 L
s
.84557 .44879 0 1.5 Mabsadd m
.84557 .44879 0 -1.5 Mabsadd L s
.817 .44879 0 1.5 Mabsadd m
.817 .44879 0 -1.5 Mabsadd L s
.83128 .48352 m
.83128 .41407 L
s
.83128 .48352 1.5 0 Mabsadd m
.83128 .48352 -1.5 0 Mabsadd L s
.83128 .41407 1.5 0 Mabsadd m
.83128 .41407 -1.5 0 Mabsadd L s
.90252 .54022 m
.87395 .54022 L
s
.90252 .54022 0 1.5 Mabsadd m
.90252 .54022 0 -1.5 Mabsadd L s
.87395 .54022 0 1.5 Mabsadd m
.87395 .54022 0 -1.5 Mabsadd L s
.88824 .57494 m
.88824 .50549 L
s
.88824 .57494 1.5 0 Mabsadd m
.88824 .57494 -1.5 0 Mabsadd L s
.88824 .50549 1.5 0 Mabsadd m
.88824 .50549 -1.5 0 Mabsadd L s
0 1 0 r
.0001 w
.06932 .04944 0 4 Mabsadd m
.06932 .04944 -3.464 -2 Mabsadd L
.06932 .04944 3.464 -2 Mabsadd L
.06932 .04944 0 4 Mabsadd L closepath F
.15431 .20364 0 4 Mabsadd m
.15431 .20364 -3.464 -2 Mabsadd L
.15431 .20364 3.464 -2 Mabsadd L
.15431 .20364 0 4 Mabsadd L closepath F
.23467 .27248 0 4 Mabsadd m
.23467 .27248 -3.464 -2 Mabsadd L
.23467 .27248 3.464 -2 Mabsadd L
.23467 .27248 0 4 Mabsadd L closepath F
.34775 .3236 0 4 Mabsadd m
.34775 .3236 -3.464 -2 Mabsadd L
.34775 .3236 3.464 -2 Mabsadd L
.34775 .3236 0 4 Mabsadd L closepath F
.45757 .29361 0 4 Mabsadd m
.45757 .29361 -3.464 -2 Mabsadd L
.45757 .29361 3.464 -2 Mabsadd L
.45757 .29361 0 4 Mabsadd L closepath F
.53722 .28601 0 4 Mabsadd m
.53722 .28601 -3.464 -2 Mabsadd L
.53722 .28601 3.464 -2 Mabsadd L
.53722 .28601 0 4 Mabsadd L closepath F
.64876 .32671 0 4 Mabsadd m
.64876 .32671 -3.464 -2 Mabsadd L
.64876 .32671 3.464 -2 Mabsadd L
.64876 .32671 0 4 Mabsadd L closepath F
.74139 .34163 0 4 Mabsadd m
.74139 .34163 -3.464 -2 Mabsadd L
.74139 .34163 3.464 -2 Mabsadd L
.74139 .34163 0 4 Mabsadd L closepath F
.83128 .44879 0 4 Mabsadd m
.83128 .44879 -3.464 -2 Mabsadd L
.83128 .44879 3.464 -2 Mabsadd L
.83128 .44879 0 4 Mabsadd L closepath F
.88824 .54022 0 4 Mabsadd m
.88824 .54022 -3.464 -2 Mabsadd L
.88824 .54022 3.464 -2 Mabsadd L
.88824 .54022 0 4 Mabsadd L closepath F
0 g
.001 w
.12567 .16844 m
.11615 .16844 L
s
.12567 .16844 0 1.5 Mabsadd m
.12567 .16844 0 -1.5 Mabsadd L s
.11615 .16844 0 1.5 Mabsadd m
.11615 .16844 0 -1.5 Mabsadd L s
.12091 .1858 m
.12091 .15108 L
s
.12091 .1858 1.5 0 Mabsadd m
.12091 .1858 -1.5 0 Mabsadd L s
.12091 .15108 1.5 0 Mabsadd m
.12091 .15108 -1.5 0 Mabsadd L s
.22204 .24757 m
.21252 .24757 L
s
.22204 .24757 0 1.5 Mabsadd m
.22204 .24757 0 -1.5 Mabsadd L s
.21252 .24757 0 1.5 Mabsadd m
.21252 .24757 0 -1.5 Mabsadd L s
.21728 .26493 m
.21728 .23021 L
s
.21728 .26493 1.5 0 Mabsadd m
.21728 .26493 -1.5 0 Mabsadd L s
.21728 .23021 1.5 0 Mabsadd m
.21728 .23021 -1.5 0 Mabsadd L s
.31255 .28497 m
.30302 .28497 L
s
.31255 .28497 0 1.5 Mabsadd m
.31255 .28497 0 -1.5 Mabsadd L s
.30302 .28497 0 1.5 Mabsadd m
.30302 .28497 0 -1.5 Mabsadd L s
.30779 .30233 m
.30779 .26761 L
s
.30779 .30233 1.5 0 Mabsadd m
.30779 .30233 -1.5 0 Mabsadd L s
.30779 .26761 1.5 0 Mabsadd m
.30779 .26761 -1.5 0 Mabsadd L s
.39911 .28236 m
.38958 .28236 L
s
.39911 .28236 0 1.5 Mabsadd m
.39911 .28236 0 -1.5 Mabsadd L s
.38958 .28236 0 1.5 Mabsadd m
.38958 .28236 0 -1.5 Mabsadd L s
.39434 .29972 m
.39434 .265 L
s
.39434 .29972 1.5 0 Mabsadd m
.39434 .29972 -1.5 0 Mabsadd L s
.39434 .265 1.5 0 Mabsadd m
.39434 .265 -1.5 0 Mabsadd L s
.50447 .27217 m
.49495 .27217 L
s
.50447 .27217 0 1.5 Mabsadd m
.50447 .27217 0 -1.5 Mabsadd L s
.49495 .27217 0 1.5 Mabsadd m
.49495 .27217 0 -1.5 Mabsadd L s
.49971 .28953 m
.49971 .2548 L
s
.49971 .28953 1.5 0 Mabsadd m
.49971 .28953 -1.5 0 Mabsadd L s
.49971 .2548 1.5 0 Mabsadd m
.49971 .2548 -1.5 0 Mabsadd L s
.60243 .29258 m
.59291 .29258 L
s
.60243 .29258 0 1.5 Mabsadd m
.60243 .29258 0 -1.5 Mabsadd L s
.59291 .29258 0 1.5 Mabsadd m
.59291 .29258 0 -1.5 Mabsadd L s
.59767 .30995 m
.59767 .27522 L
s
.59767 .30995 1.5 0 Mabsadd m
.59767 .30995 -1.5 0 Mabsadd L s
.59767 .27522 1.5 0 Mabsadd m
.59767 .27522 -1.5 0 Mabsadd L s
.70141 .33798 m
.69188 .33798 L
s
.70141 .33798 0 1.5 Mabsadd m
.70141 .33798 0 -1.5 Mabsadd L s
.69188 .33798 0 1.5 Mabsadd m
.69188 .33798 0 -1.5 Mabsadd L s
.69664 .35534 m
.69664 .32061 L
s
.69664 .35534 1.5 0 Mabsadd m
.69664 .35534 -1.5 0 Mabsadd L s
.69664 .32061 1.5 0 Mabsadd m
.69664 .32061 -1.5 0 Mabsadd L s
.78381 .38773 m
.77429 .38773 L
s
.78381 .38773 0 1.5 Mabsadd m
.78381 .38773 0 -1.5 Mabsadd L s
.77429 .38773 0 1.5 Mabsadd m
.77429 .38773 0 -1.5 Mabsadd L s
.77905 .40509 m
.77905 .37037 L
s
.77905 .40509 1.5 0 Mabsadd m
.77905 .40509 -1.5 0 Mabsadd L s
.77905 .37037 1.5 0 Mabsadd m
.77905 .37037 -1.5 0 Mabsadd L s
.88457 .44405 m
.87505 .44405 L
s
.88457 .44405 0 1.5 Mabsadd m
.88457 .44405 0 -1.5 Mabsadd L s
.87505 .44405 0 1.5 Mabsadd m
.87505 .44405 0 -1.5 Mabsadd L s
.87981 .46141 m
.87981 .42669 L
s
.87981 .46141 1.5 0 Mabsadd m
.87981 .46141 -1.5 0 Mabsadd L s
.87981 .42669 1.5 0 Mabsadd m
.87981 .42669 -1.5 0 Mabsadd L s
1 0 0 r
.0001 w
.12091 .16844 -3 0 Mabsadd m
.12091 .16844 0 4 Mabsadd L
.12091 .16844 3 0 Mabsadd L
.12091 .16844 0 -4 Mabsadd L
.12091 .16844 -3 0 Mabsadd L s
.21728 .24757 -3 0 Mabsadd m
.21728 .24757 0 4 Mabsadd L
.21728 .24757 3 0 Mabsadd L
.21728 .24757 0 -4 Mabsadd L
.21728 .24757 -3 0 Mabsadd L s
.30779 .28497 -3 0 Mabsadd m
.30779 .28497 0 4 Mabsadd L
.30779 .28497 3 0 Mabsadd L
.30779 .28497 0 -4 Mabsadd L
.30779 .28497 -3 0 Mabsadd L s
.39434 .28236 -3 0 Mabsadd m
.39434 .28236 0 4 Mabsadd L
.39434 .28236 3 0 Mabsadd L
.39434 .28236 0 -4 Mabsadd L
.39434 .28236 -3 0 Mabsadd L s
.49971 .27217 -3 0 Mabsadd m
.49971 .27217 0 4 Mabsadd L
.49971 .27217 3 0 Mabsadd L
.49971 .27217 0 -4 Mabsadd L
.49971 .27217 -3 0 Mabsadd L s
.59767 .29258 -3 0 Mabsadd m
.59767 .29258 0 4 Mabsadd L
.59767 .29258 3 0 Mabsadd L
.59767 .29258 0 -4 Mabsadd L
.59767 .29258 -3 0 Mabsadd L s
.69664 .33798 -3 0 Mabsadd m
.69664 .33798 0 4 Mabsadd L
.69664 .33798 3 0 Mabsadd L
.69664 .33798 0 -4 Mabsadd L
.69664 .33798 -3 0 Mabsadd L s
.77905 .38773 -3 0 Mabsadd m
.77905 .38773 0 4 Mabsadd L
.77905 .38773 3 0 Mabsadd L
.77905 .38773 0 -4 Mabsadd L
.77905 .38773 -3 0 Mabsadd L s
.87981 .44405 -3 0 Mabsadd m
.87981 .44405 0 4 Mabsadd L
.87981 .44405 3 0 Mabsadd L
.87981 .44405 0 -4 Mabsadd L
.87981 .44405 -3 0 Mabsadd L s
0 g
.5 Mabswid
.02381 .01894 m
.06244 .07808 L
.10458 .13438 L
.14415 .17934 L
.18221 .21541 L
.22272 .24638 L
.26171 .26941 L
.28158 .27879 L
.30316 .2873 L
.32216 .29346 L
.34309 .29893 L
.36292 .30296 L
.38395 .30615 L
.40336 .30825 L
.41426 .30911 L
.42424 .30974 L
.43408 .31021 L
.44322 .31054 L
.45325 .3108 L
.45886 .31091 L
.46395 .31098 L
.46882 .31103 L
.47149 .31106 L
.47397 .31107 L
.47837 .3111 L
.4807 .31111 L
.48319 .31111 L
.48587 .31112 L
.48739 .31112 L
.48879 .31113 L
.49001 .31113 L
.49136 .31113 L
.49203 .31113 L
.49276 .31113 L
.49408 .31113 L
.49529 .31113 L
.49661 .31113 L
.49727 .31113 L
.49798 .31113 L
.49926 .31113 L
.50054 .31113 L
.50175 .31113 L
.50284 .31113 L
.50401 .31113 L
.50525 .31113 L
.50637 .31113 L
.50767 .31113 L
.5089 .31113 L
.51004 .31113 L
.51109 .31113 L
Mistroke
.51348 .31114 L
.51484 .31114 L
.51613 .31114 L
.51852 .31115 L
.5211 .31116 L
.52389 .31117 L
.52923 .31121 L
.53191 .31123 L
.53486 .31127 L
.54007 .31134 L
.54493 .31142 L
.54996 .31153 L
.55541 .31168 L
.56533 .31202 L
.57498 .31248 L
.58389 .31302 L
.6026 .31458 L
.61324 .31576 L
.62321 .31708 L
.64343 .32048 L
.66195 .32453 L
.68252 .33021 L
.70163 .33671 L
.73979 .35362 L
.77889 .37696 L
.82044 .40913 L
.86047 .44777 L
.89899 .49231 L
.93996 .54765 L
.97619 .60332 L
Mfstroke
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{448.75, 277.375},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000:Q0001X2000`400?l00000o`00003oo`3ooooo
0?oooj<0oooo003o0?ooool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00o`3ooolM0?ooo`030?oo
g`2LB00000000080000000<0M;oo0?ooo`3oool0o`3ooomn0?ooo`00o`3ooolO0?ooo`030?noM000
07@0_ooo0?l0ooooP@3oool00?l0oooo7`3oool00`3ookl0M00007Boo`3o0?oooh40oooo003o0?oo
oal0oooo00<0oooO09a80018W=l0o`3ooon10?ooo`00o`3ooolP0?ooo`030=nLB000000000000080
00000140M;oo0?ooo`3oool0oooo0?oOW018000000000018W03Oool0_g@000000000M;l0oooo0?no
M0000000001809cOo`050?ooo`0@0=nLB000000007Bo0?ooo`3ooml0W4P00000001d_ol0oooo0?no
M0000000001d0;ooo`3oool0omnL04P00080000000L0M;oo0=nLB000000004RL0=ooo`3oool0gia8
0080000000<0001809cOo`3oool00`3oool01@3OW4P00000000000000000M;oo0080oooo00@0okmd
0000000007@0_ooo1@3oool00`3OW4P004RL0=ooo`020?ooo`030;md0000000007Bo0080oooo00D0
gia8000000000000000007Boo`020?ooo`030;md0000M;l0oooo00D0oooo00D0gia8000000000000
000007Boo`090?ooo`070;mdM02oool0oooo0?ooo`3ogi`0B4RL0=ooo`040?ooo`0=0;md00000000
001809cOo`3oool0okmd0000000004P0W=oo0?ooo`3OW4P004RL0=ooo`020?ooo`030;md00000000
07Bo00@0oooo00`0omnL04P00000000004RL0=ooo`2oM0000000001d_`3oool0gia80018W03Oool2
0?ooo`030;md0000000007Bo0080oooo00/0_g@00000000004P0W=oo0?ooo`3o_g@000000000B02L
gol0oono07@000030000000304RLg`3oool0oooo0<X0oooo003o0?ooob00oooo00L0okmd0000B02L
gol0oooo0?oog`2LB00007Bo0080oooo00h0omnL04P0B02Lgol0omnL04Qd_`2oM00007Bo0?noW02L
gol0okmd0000B02Lgml0W7BL0=ooo`H0oooo01`0okmd07Boo`3ogi`0B01809cOo`2oM7@0_ooo0?oO
W01807@0_ono09bLg`3oool0_g@00018W03Oool0oooo0;md001dM7@0001d0;oOW02L_ol0omnL04P0
M02oool0oooo0?oo_`1d0000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`070;md0000M;l0oooo
0?oOW01807@0_ono09bLg`050?ooo`030?noM00007@0_ooo0080oooo00L0gia80018W03O_g@0M;oo
0?oOW0180000M;oo0080oooo00H0_g@0001d_`3oool0okmd0000M02oool40?ooo`030?oOW0180000
M;oo0080oooo00<0_g@0001d_`3oool01`3oool0202oM000001d0;ooo`3oool0okmd00000000B9`0
gooo0`3oool03@3OW4P004RL0=oog`2LM9`0gooo0?noM00004P0W=oO09adW03Oool0okmd0000M02o
ool00P3oool0103OW4P004RL0=noM01d_ol30?ooo`0<0?oOW01804P0W=oo0?oOW018M;l0_g@0001d
_`3o_i`0W=oo0?noM00007@0_ooo0P3oool05@3OW4P004RL0=noM01d_ol0oooo0=nLB000B9`0gooO
09adW03Oool0okmd0000B02Lgml0W7BL0=oog`2LB000M;oo0?ooo`3ogi`0B00007Boo`3;0?ooo`00
o`3ooolP0?ooo`030?oo_`1d0000M;oo0080oooo0100oooO09a80018W=l0oooo0?oog`2LB000B9cO
0?ooo`3o_g@0M7Ad0018W03Oool0oooo0?oo_`1d0000M;oo2@3oool0102oM7@0W9a80018W03Oool2
0?ooo`070?oo_`1d0000M;oo0=nLW02oool0_g@0001d_`030?ooo`070;md00000000M;oo0?ooo`3o
okl0M00007Boo`020?ooo`070?oo_`1d0000M;oo0?ooo`3OW4P004RL0=ooo`030?ooo`070=nLW02o
ool0oono07@0001d_ol0gibL0;ooo`040?ooo`030?oo_`1d0000M;oo0080oooo00L0omnL04P0B02L
gol0oooo0=nLB000B9`0gooo00<0oooo00D0gibL0;ooo`3ogi`0B00007Boo`040?ooo`030=nLB000
B9`0gooo00<0oooo00<0gibL0;ooo`3oool01@3oool02@3ooml0W4P0001dM02LW=l0oooo0?noM000
07@0_mnL07BLg`030?ooo`030?noM00004P0W=oo0080oooo00<0oono07@0001d_ol00P3oool00`3o
okl0M00007Boo`020?ooo`030?oOW01804P0W=oo00@0oooo00`0oooO09a80018W=l0oooo0?noM01d
M7@004RL0=ooo`3oool0oono07@00018W=l20?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0000
B02Lgol00P3oool00`3ookl0M00007Boo`020?ooo`060;mdM02oool0oooo0?oo_`1d0000M;oob`3o
ool00?l0oooo803oool00`3ooml0W4P004RLg`030?ooo`0;0?oOW01804P0W=oo0?ooo`2oM00004RL
0=ooo`3oool0_g@00000B02Lgol00P3oool00`2oM00007Bo0?ooo`080?ooo`030?oo_`1d000007Bo
00@0oooo00L0_g@004RLg`3oool0oooo0?noM00007@0_ooo0080oooo00<0oono07@00018W=l00P3o
ool00`2oM000B9cO0?ooo`020?ooo`040=nLB000B9`0gooo0?noM0<0000000<0B9cO0?ooo`3oool0
0P3oool00`2oM000B9cO0?ooo`050?ooo`0:0?oog`2LB000001d0;ooo`3oool0oooO09a80018W=l0
oooo0?noM0<0000000P0B9cO0?ooo`3oool0oooo0?oog`2LB00004RL0=ooo`<0oooo00D0okmd0000
00000000000004RLg`090?ooo`0:0?oog`2LB000B9cO0=nLM02Lgol0gia80000M02oool0giad09cO
o`80oooo00<0oono07@00018W=l00`3oool0302oM00007Bo0?ooo`3oool0oooO09a8000007@0_ooo
0?ooo`3ooml0W4P004RLg`D0oooo0100_g@00018W03Oool0oooo0;md000004P0W=oo0?ooo`3ooml0
W4P00000M02oool0oooo0?oog`2LB000B9cO0P3oool00`3ookl0M00004RLg`030?ooo`030;md0000
M;l0oooo00@0oooo00<0gia80000M02oool0b`3oool00?l0oooo8@3oool00`3OW4P0001d0;ooo`02
0?ooo`0;0?oog`2LB000B9cO0?ooo`3o_g@0001d0;ooo`3oool0oono07@0001d_ol00P3oool00`3O
W4P004RL0=ooo`080?ooo`030?oo_`1d0000M;oo00@0oooo00L0gia80018W03Oool0oooo0?oo_`1d
0000M;oo00<0oooo00L0_g@00018W03Oool0oooo0=nLB000B9`0gooo0080oooo00X0omnL04P0B02L
gol0oono07@0B02Lgol0oooo0=nLB0000000B9cO0P3oool00`3OW4P004RL0=ooo`060?ooo`030=nL
B000B4P0B7Bo0080oooo00X0_g@0001d_`3oool0oono07@0B02Lgol0oooo0=nLB0000000B9cO0P3o
ool00`2oM000B7Ad07BLg`030?ooo`070?oo_`1d04P0W=oo0?ooo`3OW4P0000004RLg`070?ooo`0:
0?oo_`1d0000B9cO0?oog`2LW;l0gia80000M02oool0oooO09bL_`<0oooo00<0_g@00018W03Oool0
0P3oool00`3OW4P004RL0=ooo`020?ooo`030=nLB000B4P0B7Bo0080oooo00<0_g@0001d_`3oool0
103oool01`3o_g@0001d0;ooo`3oool0oono07@0001d_ol00P3oool00`3OW4P004Q804Qd_`020?oo
o`030;md0000M;l0oooo0080oooo00<0_g@00018W03Oool00P3oool00`3OW4P004RL0=ooo`030?oo
o`030?noM00004P0W=oo0<`0oooo003o0?ooob40oooo00@0okmd00000000B9`0gooo0P3oool01P3O
W4P007Bo0?oOg`3Ookl0M00007Boo`80oooo00L0_g@0001d_`3oool0oooo0?oOW01804P0W=oo00P0
oooo00@0okmd00000000B9`0gooo0`3oool00`3ogi`0B01d0;ooo`020?ooo`030?noM00007@0_ooo
0080oooo00L0gia80000M02oool0oooo0?oOW01804P0M;oO0080oooo00l0oono07@0001d_ol0oooo
0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0omnL04P0M02oool01P3oool05@3o_g@0001d
0;noM018W=l0oooo0?noM00007@0_ooo0?ooo`3o_g@004RL0=ooo`3oool0oono07@00018W=l0oooo
0?noM000B9`0_ia807Boo`030?ooo`070?noM000B9`0gooo0?ooo`3ookl0M00004RLg`060?ooo`0;
0?oo_`1d0000M;oo0?ooo`3o_g@000000018W03Oool0oooo0?noM01d_ol00P3oool00`3OW4P00018
09cOo`020?ooo`030?oOW01804P0W=oo0080oooo00P0okmd0000M02o_g@0B9cO0?ooo`3o_g@0001d
0;ooo`<0oooo00@0omoO0=oo_`1d0000M;oo0P3oool0302oM00007Bo0?ooo`3oool0okmd0000M02o
_g@0B9cO0?ooo`3o_g@0001d0;ooo`80oooo00<0gia80000B02Lgol00P3oool00`3ogi`0B01809cO
o`020?ooo`060?oOW01804P0W=oo0?ooo`3OW7@0W=oobP3oool00?l0oooo8@3oool0303ookl0M000
07BLW01807@0_ooo0?oog`2LB00007Bo0?oOW018M;l0_g@0001d_`80oooo00L0okmd0018W03Oool0
oooo0?oog`2LB000B9cO00P0oooo00D0_g@00018W03Ogi`0B4RL0=ooo`020?ooo`030?oo_`1d0000
M;oo00<0oooo0100gia80000M02oool0oooo0;md000004P0W=oo0?ooo`3ookl0M00004Q8B018W=l0
oooo0?noM00004P0W=oo0P3oool02P3OW4P0001d0;ooo`3oool0gia80018W03Oool0oono07@0001d
_ol60?ooo`080?oo_`1d04P0W=oo0=nLB000M;l0omnL04P0B02Lgol20?ooo`0F0=nLB00007@0_ooo
0?ooo`3OW4P004RL0=ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooo0=nLB00007@0_ooo
0?ooo`3OW4P004RL0=ooo`D0oooo00<0okmd0000B02Lgol00P3oool01P2oM00004RL0=ooo`3oool0
omnL04Qd_`80oooo00<0omnL04P00018W=l00P3oool00`3ooml0W4P004RLg`020?ooo`080?oo_`1d
04P0W=oo0=nLB000M;l0omnL04P0B02Lgol30?ooo`040?oOW018M;l0_g@0001d_`80oooo00`0okmd
0018W03Oool0oooo0?oog`2LB000M;oo0=nLB000M;l0omnL04P0B02Lgol20?ooo`030?oOW0180000
B9cO0080oooo00<0oooO09a80018W=l00P3oool01P3ogi`0B01809cOo`3ooml0W4P007BoolX0oooo
003o0?ooob00oooo00d0omnL04P000000000B9cO0?oo_`1d000000000000M02oool0oooo0?noM000
000007Bo0080oooo00D0omnL04P0B02Lgol0oono07@000030000000307Boo`3oool0oooo0080oooo
00`0okmd000000000000M;oo0?ooo`3ogi`0B0000018W03Oool0oooo0;md0018W=l40?ooo`060?oO
W0180000000004RLW01d0000B9cO0P3oool00`2oM00007Bo0=nLB0020000000304RLg`3oool0oooo
0080oooo00D0omnL04P000000000000007Boo`020?ooo`030;md0018W=l0oooo00D0oooo00P0oooO
09a80018W=l0oooo0?noM0000000001809cOo`<0oooo00d0omnL04P000000000000007Boo`3ogi`0
B0000000001d_ol0oooo0;md000007@0_ooo0080oooo00D0omnL04P000000000000007Boo`040?oo
o`050?oog`2LB00000000000M02oool00P3oool0203ogi`0B7Bo0?ooo`3ookl0M000001d_`3oool0
okmd0P0000001000B9`0gooo0?oo_`1d00030000000907Boo`3ooml0W4P004RLg`3oool0okmd0000
000004P0W=oo00@0oooo00<0okmd00000000M;l00P3oool03P3ogi`0B01809cOo`3ookl0M0000000
0018W=l0oooo0?noM0000000001809cOo`3oool0okmd0P0000001000B9`0gooo0?oo_`1d00030000
000307Boo`3oool0_g@000<0000000<0B9cO0?ooo`3oool0b03oool00?l0oooo=03oool00`3o_g@0
001d0;ooo`0?0?ooo`030=nLB000B9`0gooo00/0oooo00<0gia80018W03Oool03P3oool00`3OW4P0
04RL0=ooo`060?ooo`030=nLB000M;l0oooo03@0oooo00<0okmd0000M02oool00P3oool00`3OW4P0
07Bo0?ooo`0Q0?ooo`030?noM00007@0_ooo0=40oooo003o0?ooocD0oooo00<0giad09cOo`3oool0
3P3oool00`3ogi`0B01d0;ooo`0;0?ooo`030?oOW01804P0W=oo00h0oooo00<0omnL04P0M02oool0
1P3oool00`3o_g@0001d0;ooo`0e0?ooo`070=nLM02Lgol0oooo0?ooo`3o_g@0001d0;ooo`0R0?oo
o`030=nLM02Lgol0oooo0=00oooo003o0?ooodH0oooo00<0oooO09a8001d_ol02`3oool00`3ookl0
M00004RLg`0>0?ooo`030?oog`2LB000M;oo00H0oooo00<0oono07@0B02Lgol0;`3oool0103ookl0
M0000018W03Oool60?ooo`030?oo_`1d04P0W=oo01`0oooo00@0oono07@00000B9`0goooe@3oool0
0?l0ooooA`3oool00`3OW4P0B9cO0?ooo`0;0?ooo`030=nLB000M;l0oooo00h0oooo00<0gia804RL
g`3oool01P3oool00`2oM000M;oo0?ooo`0i0?ooo`030;md001d_ol0oooo0?@0oooo003o0?ooodD0
oooo00@0oooO09a80000000007Bo2P3oool01@3ooml0W4P00000000007@0_ooo00`0oooo00@0oooO
09a80000000007Bo1@3oool0103ookl0M00000000000M;lh0?ooo`040?oo_`1d00000000001d_oD0
oooo003o0?ooool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?oo
ool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?ooool0ooooX`3o
ool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?ooool0ooooX`3oool00?l0oooo
o`3ooonS0?ooo`00O03oool00`3ookl0M000000000020000000304RLg`3oool0oooo07<0oooo00@0
_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`060?oOW0180000000000000000
B9`0goooKP3oool0102oM00000000000001d_ome0?ooo`040;md00000000000007Boo`<0oooo00<0
omnL04P0001d_ol00P3oool01P3ogi`0B00000000000000004RL0=ooog80oooo00<0oono07@00000
00000P0000000`18W=l0oooo0?ooo`0U0?ooo`00OP3oool00`3o_g@0000007Boo`1e0?ooo`060=nL
B000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B00007Boo`020?ooo`060?noM0000000M;oo
0?ooo`3ogi`0B7BoK@3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLgg<0oooo00H0gia8001d_`3o
ool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo0080oooo00H0okmd0000001d_ol0oooo0?oO
W018M;md0?ooo`030?noM0000000M;oo02P0oooo001n0?ooo`030?oog`2LB00007Bo07@0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0;0?ooo`030?noM018W=l0oooo06X0oooo00<0
oooO09a80018W=l00P3oool00`3o_g@0001809cOo`1a0?ooo`030?oog`2LB000B9cO0080oooo00<0
okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?ooo`1b0?ooo`030?oog`2LB00007Bo02P0oooo001o
0?ooo`030=nLB000B9`0gooo07<0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03O
ool:0?ooo`030?oo_`1d07@0_ooo06X0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018
W03Ooom`0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo2P3oool00`3ookl0
M01d0;ooo`1c0?ooo`030=nLB000B9`0gooo02L0oooo001_0?ooo`030?oog`2LB000000000L00000
00<0B9cO0?ooo`3oool00`3oool00`3ogi`0B01809cOo`1V0?ooo`030?oog`2LB000000000L00000
00H0B9cO0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8
001d_ol0JP3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo0700oooo00<0oooO09a8
0000M;l00`3oool00`3OW4P0001809cOo`0:0?ooo`030?oog`2LB000M;oo07<0oooo00<0omnL04P0
B02Lgol09`3oool007l0oooo00<0oono07@0001d_ol0M03oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO00X0oooo00<0oono07@00018W=l0J`3oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO0740oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`03
0?oo_`1d0000B9cO07<0oooo00<0oono07@0001d_ol09`3oool00800oooo00<0_g@0001d_`3oool0
L`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00X0oooo00<0okmd0000001d_ol0
J`3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo0740oooo00<0gia80000B02Lgol0
0P3oool00`3ooml0W4P0001d_`0:0?ooo`030?noM0000000M;oo07@0oooo00<0_g@0001d_`3oool0
9P3oool00800oooo00<0gia80018W03Oool0L`3oool00`3ogi`0B00004RLg`030?ooo`030;md0000
B9`0gooo00H0oooo00H0omoO0=ooo`3o_g@000000000M02ooom[0?ooo`030?oOW0180000B9cO00<0
oooo00<0_g@00018W03Oool0L03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0
oooo00H0omoO0=ooo`3o_g@000000000M02ooomd0?ooo`030=nLB000B9`0gooo02H0oooo00200?oo
o`030?oOW01804P0W=oo07<0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`06
0?ooo`050?oo_`1d000000000000B02Lgol0K03oool0103ooml0W4P00018W03Oool20?ooo`030=nL
B000B9`0gooo0700oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`05
0?oo_`1d000000000000B02Lgol0M@3oool00`3ogi`0B01809cOo`0V0?ooo`00P03oool00`3ookl0
M00007Boo`1d0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01`3oool00`3OW9`0
_ooo0?ooo`1^0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool0L@3oool00`3ogi`0
B00007Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<0gibL0;ooo`3oool0MP3oool00`3ookl0
M00007Boo`0V0?ooo`00P@3oool00`2oM00007Bo0?ooo`1d0?ooo`060=nLB00007@0_ooo0?oog`2L
B000B9cO203oool00`3ogi`0M9cO0?ooo`1_0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cOL`3o
ool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`P0oooo00<0omnL07BLg`3oool0M`3oool00`2oM000
07Bo0?ooo`0U0?ooo`00O`3oool01@3ookl0M0000000000007@0_ooo07D0oooo00@0gia800000000
000007Bo2P3oool01@2oM0000000000000000000B9cO06d0oooo00@0gia800000000000007BoM@3o
ool0103OW4P0000000000000M;l:0?ooo`050;md00000000000000000018W=l0LP3oool01@3ookl0
M0000000000007@0_ooo02D0oooo003o0?ooool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00o`3o
oooo0?oooj<0oooo003o0?ooool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0
oooo003o0?ooool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?oo
ool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00KP3ooooo00000?l000006P00000K0?ooo`00KP3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool02@3oool00`00
0000oooo0?ooo`1m0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`00
0000oooo0?ooo`1m0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool06@3oool006h0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo07d0
oooo00<000000?ooo`3oool0O03oool00`000000oooo0?ooo`1m0?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo
0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo
0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<0
00000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^
0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l
0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?oo
o`3oool0803oool5000006@0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?oo
o`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo0`00000R0?ooo`03
0000003oool0oooo06@0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`03
0000O`3oool0oooo08L0oooo0`00000K0?ooo`00KP3oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo01H0oooo00<000000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000
003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3o
ool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0I03o
ool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3o
ool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo01D0oooo00<000000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`00
0000oooo0?ooo`0;0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0I03oool00`00
07l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`00
0000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0
oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0I03oool00`0007l0oooo
0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo
0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0180oooo00<0
00000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0>
0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m
0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I
0?ooo`00KP3oool3000000h0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`1T0?oo
o`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`270?oo
o`<000006`3oool006h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo06@0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`03
0000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0
403oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo06@0oooo00<0001o0?ooo`3oool0
O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool0
6@3oool006h0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo06@0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3o
ool0oooo08T0oooo00<000000?ooo`3oool06@3oool004L0oooo00@0_g@0000000000000M;oo0`3o
ool00`3ogi`0B00007Boo`030?ooo`030?oo_`1dB9`0gooo00D0oooo00H0omnL04P0000000000000
0018W03Oool<0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`1T0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0
oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo00160?ooo`060=nLB000M;l0oooo0?oOW018
0000B9cO0P3oool00`3ogi`0B00007Boo`040?ooo`030=nLB018W=l0oooo00@0oooo00H0okmd0000
001d_ol0oooo0?oOW018M;l<0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`1T0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo00150?ooo`030?oog`2LB000
B9cO0080oooo00<0okmd0000B02Lgol0203oool00`3ookl0M01d0;ooo`090?ooo`030?noM018W=l0
oooo00X0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`1M0?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0
oooo01T0oooo00150?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203oool0
0`3o_g@007Bo0?ooo`080?ooo`030?oo_`1d07@0_ooo00X0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`040?ooo`040000003oool0oooo0?ooo`T00?l00P3oool00`000000oooo0?oo
o`1M0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?oo
o`290?ooo`030000003oool0oooo01T0oooo000h0?ooo`030?oog`2LB000000000L0000000H0B9cO
0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0W=oo00P0oooo00<0oooO09a8B02Lgol0
203oool00`3ooml0W4P007Boo`0:0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`020?ooo`L00?l00`3oool00`000000oooo0?ooo`1M0?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000
003oool0oooo01T0oooo00160?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l02@3o
ool00`3ogi`0B7Bo0?ooo`070?ooo`030?oo_`1d0000B9cO00X0oooo1000000C0000O`0300000000
07l0001o00<0001o1@000007003o00@00000g`0007l00`000000001o0000O`3o0000O`H0001o1000
000K0?ooo`00AP3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00X0oooo00<0_g@0
07Boo`3oool01P3oool00`3o_g@0000007Boo`0:0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`D00?l0103oool00`000000oooo0?ooo`1M
0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`29
0?ooo`030000003oool0oooo01T0oooo00160?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018
W03Oool02@3oool00`3ogi`0B4RL0=ooo`030?ooo`060?oOg`3Oool0okmd0000000007@0_ooo2P3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3o
ool5003o00@0oooo00<000000?ooo`3oool0G@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003o
ool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00AP3oool0
103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00X0oooo00<0gia804RLg`3oool00P3o
ool01@3ookl0M0000000000004P0W=oo00/0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`080?ooo`<00?l0I@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo
00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00A`3oool00`3ogi`0B00007Bo
o`020?ooo`030=nLB000M;l0oooo00D0oooo00<0oooO0=oOo`3oool00P3oool00`3ookl0M01d0;oo
o`030?ooo`030=nLW02oool0oooo00`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?oo
o`070?ooo`<00?l0I@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00B03oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00<0_gAd0;ooo`3oool00P3oool00`3o_g@007Bo0?ooo`020?ooo`030?oO
W01dW=l0oooo00`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`080?ooo`03003o
003oool0oooo06@0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool004T0oooo00@0gia800000000000007Bo
203oool00`3ookl0M00000000003000000030000B02Lgol0oooo0080oooo00D0_g@0000000000000
000004RLg`090?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`1T0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0
oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`1T0?ooo`030000O`3oool0oooo07d0oooo
00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo
001^0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`1T0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?oo
o`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`1T0?ooo`030000O`3oool0oooo07d0oooo00<00000
0?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?oo
o`<000006`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo06@0oooo00<0001o0?oo
o`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08L0oooo0`00000K0?oo
o`00KP3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3o
ool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000
003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3o
ool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0I03o
ool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3o
ool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo0080oooo00<000000?ooo`3oool0I03oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`00
0000oooo0?ooo`0N0?ooo`050000003oool0oooo0?ooo`000000IP3oool00`0007l0oooo0?ooo`1m
0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I
0?ooo`00KP3oool00`000000oooo0?ooo`0O0?ooo`040000003oool0oooo000006H0oooo00<0001o
0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<00000
0?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0803oool00`000000oooo0000001V0?oo
o`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?oo
o`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo0200oooo1@00001T0?ooo`03
0000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`03
0000003oool0oooo01T0oooo001^0?ooo`<000008@3oool00`000000oooo0?ooo`1U0?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`270?ooo`<00000
6`3oool006h0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`1T0?ooo`030000O`3o
ool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003o
ool0oooo01T0oooo001^0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0H`3oool0
0`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool0
0`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo
06<0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo
08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0903oool00`000000
oooo0?ooo`1R0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0
oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo02D0oooo
00<000000?ooo`3oool0H@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo
00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?oo
o`0V0?ooo`030000003oool0oooo0600oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<00000
0?ooo`3oool09P3oool00`000000oooo0?ooo`1P0?ooo`030000O`3oool0oooo07d0oooo00<00000
0?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?oo
o`030000003oool0oooo02L0oooo00<000000?ooo`3oool0G`3oool00`0007l0oooo0?ooo`1m0?oo
o`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?oo
o`00KP3oool3000002P0oooo00<000000?ooo`3oool0GP3oool00`0007l0oooo0?ooo`1m0?ooo`03
0000003oool0oooo07`0oooo00<0001o0?ooo`3oool0Q`3oool3000001/0oooo001^0?ooo`030000
003oool0oooo02T0oooo00<000000?ooo`3oool0G@3oool00`0007l0oooo0?ooo`1m0?ooo`030000
003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo05d0oooo00<0001o0?ooo`3oool0O@3o
ool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3o
ool006h0oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`1L0?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0
oooo01T0oooo001^0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0F`3oool00`00
07l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`00
0000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo05X0
oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0
oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0;@3oool00`000000oooo
0?ooo`1I0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo02d0oooo00<0
00000?ooo`3oool0F@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`0^
0?ooo`030000003oool0oooo05P0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l
0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`1G0?ooo`030000O`3oool0oooo07d0oooo00<000000?oo
o`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`<0
0000<03oool00`000000oooo0?ooo`1F0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3o
ool0O03oool00`0007l0oooo0?ooo`270?ooo`<000006`3oool006h0oooo00<000000?ooo`3oool0
<03oool00`000000oooo0?ooo`1F0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0
O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003o
ool0oooo0340oooo00<000000?ooo`3oool0E@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003o
ool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool0
0`000000oooo0?ooo`0b0?ooo`030000003oool0oooo05@0oooo00<0001o0?ooo`3oool0O@3oool0
0`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool0
0500oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`060?oOW0180000
000000000000B9`0gooo303oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo05<0oooo
00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo
00<000000?ooo`3oool06@3oool004l0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo0080oooo00H0okmd0000001d_ol0oooo0?oOW018M;l<0?ooo`030000003oool0
oooo03<0oooo00<000000?ooo`3oool0D`3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00CP3oool00`3o
oml0W4P004RLg`020?ooo`030?noM00004P0W=oo00/0oooo00<0okmd04RLg`3oool02P3oool00`00
0000oooo0?ooo`0d0?ooo`030000003oool0oooo0580oooo00<0001o0?ooo`3oool0O@3oool00`00
0000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool004h0
oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d07@0_ooo
00X0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`1A0?ooo`030000O`3oool0oooo
07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo
01T0oooo00110?ooo`030?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0oooO09a80000M;l3
0?ooo`030=nLB00004P0W=oo00X0oooo00<0oooO09a8001d_ol02P3oool00`000000oooo0?ooo`0f
0?ooo`030000003oool0oooo0500oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l
0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool002h0oooo00<0_ibL0=oo
o`3oool07P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo00<0oono07@0
0018W=l02P3oool4000003H0001o00<000000000O`0007l0c`0007l00`000000001o0000O`3o0000
O`H0001o1000000K0?ooo`00;03oool0103ookl0M0000018W03OoolO0?ooo`030=nLB00004P0W=oo
0080oooo00<0oooO09a80000M;l02P3oool00`3o_g@0000007Boo`0:0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool0CP3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo
07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00:03oool00`3ookl0
M00000000003000000030018W03Oool0oooo01h0oooo00<0omnL04P00018W=l00`3oool00`2oM000
04RL0=ooo`060?ooo`060?oOg`3Oool0okmd0000000007@0_ooo2P3oool00`000000oooo0?ooo`0h
0?ooo`030000003oool0oooo04h0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l
0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool00280oooo00@0oono09bL
g`3ookl0M000100000001P00M;l0oooo0?ooo`3oool0_ibL0=oooal0oooo00@0oooO09a80000B9`0
gooo0P3oool00`3OW4P004RL0=ooo`060?ooo`050?oo_`1d000000000000B02Lgol02`3oool00`00
0000oooo0?ooo`0i0?ooo`030000003oool0oooo04d0oooo00<0001o0?ooo`3oool0O@3oool00`00
0000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool00280
oooo00D0omnL04P0000000000000001d_`030?ooo`030;mdM02oool0oooo02<0oooo00<0omnL04P0
001d_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo00`0oooo00<000000?oo
o`3oool0>P3oool00`000000oooo0?ooo`1<0?ooo`030000O`3oool0oooo07d0oooo00<000000?oo
o`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000S0?ooo`03
0?oo_`1dB9`0gooo00@0oooo00<0gibL0;ooo`3oool0903oool01P3OW4P0001d0;ooo`3ooml0W4P0
04RLg`P0oooo00<0omnL07BLg`3oool0303oool00`000000oooo0?ooo`0k0?ooo`030000003oool0
oooo04/0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo08T0oooo00<000000?ooo`3oool06@3oool00280oooo00<0oooO09a8M02oool0103oool00`3o
okl0M7Bo0?ooo`0V0?ooo`040=nLB00000000000001d_`X0oooo00D0_g@0000000000000000004RL
g`090?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0BP3oool00`0007l0oooo0?oo
o`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?oo
o`0I0?ooo`008P3oool00`3ogi`0M9cO0?ooo`030?ooo`030?oo_`1d07@0_ooo04<0oooo00<00000
0?ooo`3oool0>`3oool6000004P0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l
0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool00280oooo00P0okmd001d
_`3oool0oooo0?oog`2LB000001d0;oood@0oooo00<000000?ooo`3oool0?@3oool2000004X0oooo
00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo
00<000000?ooo`3oool06@3oool002<0oooo00D0_g@0000000000000001809cOo`160?ooo`<00000
?@3oool00`000000oooo000000190?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0
O03oool00`0007l0oooo0?ooo`270?ooo`<000006`3oool006h0oooo00<000000?ooo`3oool0?@3o
ool00`3o0000oooo000000190?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0
oooo03`0oooo00D0o`000000003o0000oooo000000180?ooo`030000O`3oool0oooo07d0oooo00<0
00000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000S
0?ooo`050;md0000B9`0gooo0?oo_`1d00000P0000000`0007@0_ooo0?ooo`110?ooo`030000003o
ool0oooo03`0oooo00H0o`000000003oool0o`000?ooo`0000170?ooo`030000O`3oool0oooo07d0
oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0
oooo000R0?ooo`030?oOW0180000000000<0000000@0B9cO0?ooo`2oW9`0gooo@P3oool00`000000
oooo0?ooo`0k0?ooo`050?l0003oool000000?ooo`3o00000P3oool00`000000oooo0?ooo`140?oo
o`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?oo
o`030000003oool0oooo01T0oooo000R0?ooo`040?oog`2LB000001d0;ooo`<0oooo00<0okmd04RL
g`3oool0@P3oool00`000000oooo0?ooo`0j0?ooo`040?l000000000oooo00000080oooo00@0o`00
0?ooo`3oool00000A@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00:03oool00`3o_g@007Bo0?ooo`13
0?ooo`030000003oool0oooo03X0oooo00@0o`000000003oool000000P3oool01@000000o`000?oo
o`3oool0000004@0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool002H0oooo00<0oooO0;no_`2LW;l0A@3o
ool00`000000oooo0?ooo`0i0?ooo`030?l0003oool0000000D0000000D0oooo0?l0003oool0oooo
000000130?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`290?ooo`030000003oool0oooo01T0oooo000V0?ooo`060?oO_`2LM4P000000000000007@0
_ooo@P3oool00`000000oooo0?ooo`0j0?ooo`040?l000000000oooo00000080oooo00<000000?l0
003oool00`3oool00`000000oooo0?ooo`040?ooo`H00000=P3oool00`0007l0oooo0?ooo`1m0?oo
o`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?oo
o`008P3oool00`3ogi`0B00000000002000000050000B02Lgol0oooo0;nLW03Oool0@P3oool30000
03X0oooo00@0o`000000003oool000000P3oool00`3o0000oooo0?ooo`030?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0Q`3oool3000001/0oooo000R0?ooo`030?noM00007@0_ooo
00@0oooo00<0_g@007Boo`3oool0@P3oool00`000000oooo0?ooo`0k0?ooo`060?l0003oool00000
0?ooo`3o000000001P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo03L0oooo00<0
001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<0
00000?ooo`3oool06@3oool002T0oooo00<0gmoO0?ooo`3oool0@P3oool00`000000oooo0?ooo`0l
0?ooo`040?l000000000oooo0?l000P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0g0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?oo
o`290?ooo`030000003oool0oooo01T0oooo000R0?ooo`050?oO_`2ogol0oooo0?ooo`2oM0000`00
00000`0007@0_ooo0?ooo`110?ooo`030000003oool0oooo03`0oooo00<0o`000000003o00002P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03L0oooo00<0001o0?ooo`3oool0O@3o
ool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3o
ool00280oooo00<0_g@00000000000000P0000001@0007@0_ooo0?ooo`2oW9`0gooo0480oooo00<0
00000?ooo`3oool0?@3oool00`3o0000oooo0?ooo`0:0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0=`3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00803oool01@2oM000000000000000
B9`0gooo00<0oooo00<0oooO09a8001d_ol0@`3oool00`000000oooo0?ooo`0m0?ooo`030000003o
ool0oooo00/0oooo00D000000?ooo`3oool0oooo0000000i0?ooo`030000O`3oool0oooo07d0oooo
00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo
000R0?ooo`030?noW02Lgol0oooo04T0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?oo
o`0<0?ooo`040000003oool0oooo000003T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo
0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<0
00000?ooo`3oool0?@3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0000003T0oooo00<0
001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<0
00000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0>`3oool6000000d0oooo0P00000i
0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`29
0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo04l0oooo00<000000?oo
o`3oool0=`3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?oo
o`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool3000004l0oooo0P00000h0?ooo`03
0000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`270?ooo`<0
00006`3oool002d0oooo00<0_g@00018W03Oool0?P3oool00`000000oooo0?ooo`1?0?ooo`030000
003oool0000003L0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool002d0oooo00<0omnL04Qd_`3oool0?P3o
ool00`000000oooo0?ooo`1?0?ooo`040000003oool0oooo000003H0oooo00<0001o0?ooo`3oool0
O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool0
6@3oool00280oooo00<0oono07Ad_`3oool01`3oool00`3o_g@004RL0=ooo`0o0?ooo`030000003o
ool0oooo04l0oooo00D000000?ooo`3oool0oooo0000000e0?ooo`030000O`3oool0oooo07d0oooo
00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo
000R0?ooo`040?oOW0180000001d0;ooo`D0oooo00<0gia8001d_`3oool0@03oool00`000000oooo
0?ooo`1?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0<P3oool00`0007l0oooo
0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo
0?ooo`0I0?ooo`008`3oool00`3ogi`0B000000000050000000304RLg`3oool0oooo0400oooo00<0
00000?ooo`3oool0B03oool00`000000oooo0?ooo`0:003o00030000003oool000000340oooo00<0
001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<0
00000?ooo`3oool06@3oool002P0oooo00<0okmd0000B02Lgol0@`3oool00`000000oooo0?ooo`18
0?ooo`040000003oool0oooo0?ooo`P00?l00P3oool200000340oooo00<0001o0?ooo`3oool0O@3o
ool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3o
ool002L0oooo00<0_g@0001d_`3oool0A03oool00`000000oooo0?ooo`180?ooo`@000001`00o`05
00000340oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo08T0oooo00<000000?ooo`3oool06@3oool00280oooo00H0omnL04P0001d_ol0okmd0000001d
_om60?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3oool00P3oool6003o00@0oooo0P00
000`0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?oo
o`290?ooo`030000003oool0oooo01T0oooo000R0?ooo`040?oOW01800000000001d_dP0oooo0`00
00180?ooo`030000003oool0oooo0080oooo1@00o`050?ooo`030000003oool0000002l0oooo00<0
001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08L0oooo0`00
000K0?ooo`00KP3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo00<0oooo0`00o`06
0?ooo`040000003oool0oooo000002h0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0oooo00<00000
0?ooo`3oool0CP3oool3003o00X0oooo00<000000?ooo`3oool0:`3oool00`0007l0oooo0?ooo`1m
0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I
0?ooo`009P3oool01P3o_g@00000000000000000001d0;oood80oooo00<000000?ooo`3oool0C`3o
ool00`00o`00oooo0?ooo`0:0?ooo`030000003oool0oooo02X0oooo00<0001o0?ooo`3oool0O@3o
ool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3o
ool001h0oooo00H0_ibL0=ooo`3oool0oooo0?oOW0180003000000040018W03Oool0oono07Ad_a`0
oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB00000000000
00<0000000D0B9cO0?ooo`3oool0omnL04P00002000000030018W03Oool0oooo00/0oooo00<00000
0?ooo`3oool0C`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo02T0oooo00<0001o
0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<00000
0?ooo`3oool06@3oool001d0oooo00<0oooO09a8000000000`0000000`00M;l0oooo0?ooo`020?oo
o`030?noM00007@0_ooo01/0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW018
0000M;oo00<0oooo00<0okmd001d_`3oool00P3oool02@2oM7@0_ooo0?ooo`3o_g@0000007Boo`3o
ool0omnL04Qd_`0<0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0303oool00`00
0000oooo0?ooo`0X0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`00
07l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000M0?ooo`030?oo_`1d0000B9cO02D0
oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`080?ooo`030?noM000M;l0oooo00T0
oooo00<0okmd04RLg`3oool02P3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo00d0
oooo0P00000X0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0
oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo00150?ooo`030?oog`2LB000B9cO0080oooo
00@0oooO09a80000B9`0gooo203oool00`3o_g@007Bo0?ooo`080?ooo`030?oo_`1d07@0_ooo00X0
oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo02D0
oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0
oooo00<000000?ooo`3oool06@3oool002H0oooo00D0_g@0000000000000000004RLg`0=0?ooo`03
0?oog`2LB000000000L0000000H0B9cO0?ooo`3oool0oooO09a80000M;l30?ooo`030=nLB00004P0
W=oo00T0oooo00<0okmd04RLg`3oool01`3oool00`3ooml0W4P007Boo`0:0?ooo`030000003oool0
oooo04l0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0T0?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0
oooo01T0oooo000T0?ooo`030=nLB0000000M;oo0080oooo00<0omnL04P0M02oool06P3oool00`2o
M00004RL0=ooo`020?ooo`030?oOW0180000B9cO00X0oooo00<0gia804RLg`3oool01P3oool00`3o
okl0M00004RLg`0:0?ooo`@00000CP0007l00`000000001o0000O`0A0000O`800000Y00007l00`00
0000001o0000O`3o0000O`H0001o1000000K0?ooo`008`3oool00`3OW4P0B9cO0?ooo`030?ooo`03
0?oOW018M;l0oooo01X0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`0;0?ooo`03
0;md001d_ol0oooo00D0oooo00<0okmd0000001d_ol02P3oool00`000000oooo0?ooo`1?0?ooo`03
0000003oool0oooo01<0oooo00<000000?ooo`3oool08@3oool00`0007l0oooo0?ooo`1m0?ooo`03
0000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00
8P3oool00`3ogi`0M9cO0?ooo`030?ooo`030?noW018M9`0gooo01/0oooo00<0omnL04P00018W=l0
0`3oool00`2oM00004RL0=ooo`0:0?ooo`030?oOW01807@0_ooo0080oooo00H0omoO0=ooo`3o_g@0
00000000M02oool:0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`0P0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`00
07l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000R0?ooo`030?noM000M;l0oooo0080
oooo00D0omnL04QdM01d000004RL0=ooo`0J0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia8
0018W03Oool02P3oool00`3ooml0W4P007Boo`020?ooo`050?oo_`1d000000000000B02Lgol02`3o
ool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo01D0oooo0P00000P0?ooo`030000O`3o
ool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003o
ool0oooo01T0oooo000S0?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool06P3oool0
0`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00X0oooo00<0oono07@00018W=l00`3oool0
0`3OW9`0_ooo0?ooo`0<0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool05`3oool0
0`000000oooo0?ooo`030?ooo`D000005@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool01@3o
gi`0B000000000000000B9cO0080oooo00<0_gA809cOo`3oool0703oool01P3OW4P0001d0;ooo`3o
oml0W4P004RLg`L0oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00<0oooo00<0omnL07BL
g`3oool0303oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo01P0oooo0P0000050?oo
o`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?oo
o`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool004T0oooo00@0gia800000000
000007Bo2@3oool01@3ogi`0B000000000000000M;oo00D0oooo00D0_g@0000000000000000004RL
g`090?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?oo
o`020?ooo`030000003oool0oooo01D0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool002P0oooo00<0omnL
04P0B02Lgol0@`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo01/0oooo00D00000
0?ooo`3oool0oooo0?l0000G0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000T0?ooo`030?oOW0180000
00000080000000<004Qd07BL_`3oool0@P3oool00`000000oooo0?ooo`1=0?ooo`H000006`3oool2
000000030?l000000000o`0000D0oooo1P00000;0?ooo`030000O`3oool0oooo07d0oooo00<00000
0?ooo`3oool0O03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000N0?oo
o`040;nLW03Oool0oono07@000<0000000<0001d0;ooo`3oool00P3oool00`3OW9`0_ooo0?ooo`11
0?ooo`030000003oool0oooo0700oooo0P0000000`3oool0o`000?ooo`050?ooo`030000003oool0
oooo00`0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo08T0oooo00<000000?ooo`3oool06@3oool001d0oooo00<0oooO09a8000000000P0000000`1d
_ol0okmd04RLg`050?ooo`030=nLW02oool0oooo0440oooo0`00001_0?ooo`050?l0003oool00000
0?ooo`3o00001P3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0oooo07d0oooo00<00000
0?ooo`3oool0O03oool00`0007l0oooo0?ooo`270?ooo`<000006`3oool001d0oooo00<0oono09bL
g`3oool00P3oool00`3ookl0M7Bo0?ooo`040?ooo`030?oo_`1dM;l0oooo0480oooo00<000000?oo
o`3oool0KP3oool00`3o000000000?ooo`03000000030?l0003oool0oooo00<0oooo00<000000?oo
o`3oool0303oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?oo
o`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool00`3o_g@0B9cO0?ooo`030?ooo`03
0?oo_`1d07@0_ooo04<0oooo00<000000?ooo`3oool0KP3oool0103o000000000?ooo`0000020?oo
o`030000003o0000oooo00<0oooo00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`1m0?oo
o`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?oo
o`008P3oool00`3ogi`0B4RL0=ooo`020?ooo`030;md000007@0_ooo04@0oooo00<000000?ooo`3o
ool0K@3oool00`3o0000oooo00000007000000<0oooo00<000000?ooo`3oool0303oool00`0007l0
oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0R@3oool00`000000
oooo0?ooo`0I0?ooo`008`3oool01@3OW4P00000000000000000M;oo04H0oooo00<000000?ooo`3o
ool0KP3oool0103o000000000?ooo`0000020?ooo`030000003o0000oooo0080000000<0oooo0000
003oool03@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?oo
o`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`1^0?ooo`04
0?l000000000oooo00000080oooo00<0o`000?ooo`3oool00P3oool2000000h0oooo00<0001o0?oo
o`3oool0O@3oool00`000000oooo0?ooo`0>0?ooo`H00000J03oool00`0007l0oooo0?ooo`290?oo
o`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo06l0oooo00D0o`000?ooo`00
0000oooo0?l000060?ooo`8000003@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo
0100oooo00<000000?ooo`3oool0J@3oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo
01T0oooo000U0?ooo`060=nLB000000000000000000004P0W=oo@`3oool00`000000oooo0?ooo`1`
0?ooo`040?l000000000oooo0?l000H0oooo00<000000?ooo`000000303oool00`0007l0oooo0?oo
o`1j0?ooo`H00000403oool00`000000oooo0?ooo`1Y0?ooo`030000O`3oool0oooo08T0oooo00<0
00000?ooo`3oool06@3oool002<0oooo00T0oono07@00018W=l0omnL09boo`3oool0gia80018W03O
ool0@P3oool00`000000oooo0?ooo`1`0?ooo`030?l000000000o`0000L0oooo00<000000?ooo`3o
ool00P00000:0?ooo`030000O`3oool0oooo07`0oooo0P00000B0?ooo`030000003oool0oooo06T0
oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`008`3oool02@2oM7@0_ooo
0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d0;ooo`120?ooo`<00000L@3oool00`3o0000oooo0?ooo`06
0?ooo`030000003oool0oooo0080oooo0P0000080?ooo`030000O`3oool0oooo06@0oooo1@00000C
0?ooo`8000004P3oool00`000000oooo0?ooo`1Y0?ooo`030000O`3oool0oooo08L0oooo0`00000K
0?ooo`008P3oool02@3ogi`0W;oo0?ooo`3ooml0W7BL0=ooo`3oool0omnL04Qd_`130?ooo`030000
003oool0oooo0740oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00D0oooo00<0001o0?ooo`3oool0IP3oool00`000000oooo0?ooo`0C0?ooo`030?l0
00000000oooo0140oooo00<000000?ooo`3oool0J@3oool00`0007l0oooo0?ooo`290?ooo`030000
003oool0oooo01T0oooo000R0?ooo`090?noW02Lgol0oooo0;md001d_ol0oooo0=nLB00007@0_ooo
04<0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo
00D0oooo0P0000050?ooo`030000O`3oool0oooo06H0oooo00<000000?ooo`3oool04P3oool00`3o
000000000?l0000B0?ooo`030000003oool0oooo06T0oooo00<0001o0?ooo`3oool0R@3oool00`00
0000oooo0?ooo`0I0?ooo`008P3oool0103ookl0M0000000001d_om80?ooo`030000003oool0oooo
06l0oooo1@0000060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3oool00`00
07l0oooo0?ooo`120?ooo`D000007`3oool00`000000oooo0?ooo`0B0?ooo`040?l0000000000000
0?l00140oooo00<000000?ooo`3oool0J@3oool00`0007l0oooo0?ooo`290?ooo`030000003oool0
oooo01T0oooo001^0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0203oool20000
0080oooo00<0001o0?ooo`3oool0A03oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo
0140oooo00D0o`000?ooo`00000000000?l0000A0?ooo`030000003oool0oooo06T0oooo00<0001o
0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`1c0?oo
o`030000003oool0oooo00X00?l00P3oool00`000000oooo0?ooo`020?ooo`80000000<0001o0?oo
o`3oool04`3oool6000002/0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0@0?oo
o`030?l000000000oooo0080000000<0oooo0?l0003oool03`3oool00`000000oooo0?ooo`1Y0?oo
o`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool002H0oooo00H0okmd00000000
000000000000M02ooom20?ooo`030000003oool0oooo07<0oooo00@000000?ooo`3oool0oooo2000
o`030?ooo`030000003oool0oooo00@0oooo0P00000G0?ooo`030000003oool0oooo02/0oooo00<0
o`000?ooo`3oool07`3oool00`000000oooo0?ooo`0@0?ooo`030?l000000000oooo0080000000<0
oooo0000003o00003`3oool00`000000oooo0?ooo`1Y0?ooo`030000O`3oool0oooo08T0oooo00<0
00000?ooo`3oool06@3oool001h0oooo00H0_ibL0=ooo`3oool0oooo0?oOW0180003000000040018
W03Oool0oono07Ad_d<0oooo00<000000?ooo`3oool0L`3oool4000000L00?l01@0000060?ooo`04
0000O`3oool00000000001D0oooo00<000000?ooo`3oool0:P3oool00`3o000000000?l0000P0?oo
o`030000003oool0oooo00l0oooo00<0o`000?ooo`0000001@0000000`3oool0o`000?ooo`0=0?oo
o`030000003oool0oooo06T0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?oo
o`007@3oool00`3ooml0W4P00000000300000003001d_`3oool0oooo0080oooo00<0okmd0000M02o
ool0@`3oool00`000000oooo0?ooo`1c0?ooo`030000003oool0oooo0080oooo1P00o`040?ooo`03
0000003oool0oooo00@0oooo00@0001o0?ooo`3oool0oooo0P00000C0?ooo`030?l0003oool0oooo
02T0oooo00@0o`000?ooo`000000o`00803oool00`000000oooo0?ooo`0@0?ooo`030?l000000000
oooo0080000000<0oooo0000003o00003`3oool00`000000oooo0?ooo`1Y0?ooo`030000O`3oool0
oooo08T0oooo00<000000?ooo`3oool06@3oool001d0oooo00<0oono07@00018W=l0CP3oool30000
07<0oooo00<000000?ooo`3oool00P3oool5003o00D0oooo00<000000?ooo`3oool0103oool00`00
07l0oooo0?ooo`030?ooo`800000403oool00`3o000000000?l0000Y0?ooo`060?l0003oool0oooo
0000003oool0o`007`3oool00`000000oooo0?ooo`0@0?ooo`030?l000000000oooo0080000000<0
oooo0?l0003oool03`3oool00`000000oooo0?ooo`0L0?ooo`D00000B03oool00`0007l0oooo0?oo
o`270?ooo`<000006`3oool006h0oooo00<000000?ooo`3oool0N@3oool3003o00d0oooo00<0001o
0?ooo`3oool01@3oool2000000d0oooo00@0o`000?ooo`000000o`00:@3oool01`3o0000oooo0?oo
o`000000oooo0000003o00007P3oool00`000000oooo0?ooo`0A0?ooo`050?l0003oool000000000
003o00004@3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo04P0oooo00<0001o0?oo
o`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`1i0?ooo`<0
0?l03@3oool00`0007l0oooo0?ooo`070?ooo`<000002@3oool01P3o0000oooo0?ooo`000000oooo
0?l002L0oooo00P0o`000000003oool0oooo0000003oool000000?l001h0oooo00<000000?ooo`3o
ool04P3oool0103o000000000000003o000A0?ooo`030000003oool0oooo01h0oooo00<000000?oo
o`3oool0B03oool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`03
0000003oool0oooo07X0oooo00<00?l00?ooo`3oool0303oool00`0007l0oooo0?ooo`0:0?ooo`<0
00001P3oool01`3o0000oooo0?ooo`000000oooo0000003o00009@3oool00`3o0000oooo00000005
000000030?ooo`3o0000oooo01`0oooo00<000000?ooo`3oool04P3oool00`3o000000000?l0000;
0?ooo`030000003oool0oooo00X00?l00P3oool00`000000oooo0?ooo`0F0?ooo`030?l0003oool0
oooo04P0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`00
0000oooo0?ooo`1j0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool03@3oool20000
00<0oooo00P0o`000000003oool0oooo0000003oool000000?l002H0oooo00P0o`000000003oool0
oooo0000003oool000000?l001h0oooo00<000000?ooo`3oool04`3oool00`3o000000000?ooo`0:
0?ooo`040000003oool0oooo0?ooo`P00?l00`3oool00`000000oooo0?ooo`0E0?ooo`030?l00000
0000o`0004T0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00:03oool0
103ookl0M0000000M02ooom20?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0303o
ool00`0007l0oooo0?ooo`0?0?ooo`<0000000<0oooo000000000000100000000`3oool0o`000?oo
o`0U0?ooo`070?l0003oool0oooo0000003oool000000?l0000N0?ooo`030000003oool0oooo01<0
oooo0P00000;0?ooo`@000001`00o`05000001H0oooo00@0o`000?ooo`000000o`00B@3oool00`00
07l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo000T0?ooo`030?oo_`1d000000000080
000000<0001809cOo`3oool0@P3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo00`0
oooo00<0001o0?ooo`3oool04P3oool3000000050?ooo`000000oooo0000003o00009`3oool01P3o
0000oooo0?ooo`000000oooo0?l001l0oooo00<000000?ooo`3oool04`3oool2000000/0oooo00<0
00000?ooo`3oool00P3oool6003o00@0oooo00<000000?ooo`3oool04`3oool01P3o0000oooo0?oo
o`000000oooo0?l004P0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00
7P3oool0102oW9`0gooo0?oo_`1d0003000000030000B018B00007Bo04H0oooo00<000000?ooo`3o
ool0NP3oool00`000000oooo0?ooo`0<0?ooo`030000O`3oool0oooo01<0oooo00<0o`000?ooo`00
00000`0000000`3o0000oooo0?ooo`0=0?ooo`D00000503oool0103o0000oooo0000003o000H0?oo
o`040000003oool0oooo0?ooo`T00?l00P3oool00`000000oooo0?ooo`0:0?ooo`H000002@3oool0
0`000000oooo0?ooo`020?ooo`D00?l01@3oool00`000000oooo0?ooo`0C0?ooo`070?l0003oool0
oooo0000003oool000000?l000170?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool0
6@3oool001d0oooo00X0oooO09a80000000000000018W03Oool0oooo0?oOW01804P0W=ooA`3oool0
0`000000oooo0?ooo`1j0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool04`3oool0
1@3o0000oooo0?ooo`000000oooo00@000003`3oool00`000000oooo0?ooo`0E0?ooo`030?l00000
0000o`0001P0oooo00<000000?ooo`3oool00P3oool7003o00<0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00<0oooo0`00o`060?ooo`030000003o
ool0oooo0180oooo00P0o`000000003oool0oooo0000003oool000000?l004L0oooo00<0001o0?oo
o`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`008`3oool00`3ookl0M4RL0=ooo`180?ooo`03
0000003oool0oooo07X0oooo00<000000?ooo`3oool0303oool00`0007l0oooo0?ooo`0D0?ooo`04
0?l0003oool000000?l000@0oooo0`00000<0?ooo`030000003oool0oooo01H0oooo00<0o`000?oo
o`3oool05`3oool5000000L00?l01000000?0?ooo`030000003oool0oooo00l0oooo0`00o`0J0?oo
o`030?l0003oool0000000D0000000<0oooo0?l0003oool04`3oool5000002d0oooo00<0001o0?oo
o`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool00`3ooml0W4Qd0;ooo`030?ooo`04
0?noM0000000001d0;oood80oooo0`00001j0?ooo`030000003oool0oooo00`0oooo00<0001o0?oo
o`3oool05@3oool00`3o000000000?l000070?ooo`@00000203oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo01L0oooo00<000000?ooo`3oool00`3oool5003o00@0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0@0?ooo`03003o003oool0oooo01X0oooo00P0o`000000
003oool0oooo0000003oool000000?l001L0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo
0?ooo`270?ooo`<000006`3oool00280oooo00@0okmd0018W03Ookl0M000100000000`00B9`0gooo
0?ooo`110?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool0303oool00`0007l0oooo
0?ooo`0F0?ooo`030?l0003oool0oooo00X0oooo100000040?ooo`030000003oool0oooo01H0oooo
00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`030?ooo`D00?l0103oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool06`3oool01`3o0000oooo
0?ooo`000000oooo0000003o00005`3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo
08T0oooo00<000000?ooo`3oool06@3oool00280oooo00T0omnL04P000000000000004RLg`3oool0
oooO09a8M02oool0@`3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo00`0oooo00<0
001o0?ooo`3oool05P3oool00`000000oooo0?ooo`0>0?ooo`D00000603oool00`000000oooo0?oo
o`0G0?ooo`030000003oool0oooo00@0oooo0`00o`050?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0403oool00`000000oooo0?ooo`0K0?ooo`060?l0003oool0oooo0000003oool0
o`00603oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3o
ool06@3oool006h0oooo00<000000?ooo`3oool0NP3oool00`000000oooo0?ooo`0<0?ooo`030000
O`3oool0oooo01H0oooo00<000000?ooo`3oool04P3oool600000140oooo1@00000N0?ooo`<00?l0
5@3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01`0oooo00@0o`000?ooo`000000
o`006@3oool00`000000oooo0?ooo`0]0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3o
ool06@3oool005d0oooo00@0_g@0000000000000M;oo3@3oool00`000000oooo0?ooo`1j0?ooo`03
0000003oool0oooo00`0oooo00<0001o0?ooo`3oool04`3oool600000180oooo00<000000?ooo`3o
ool00`3oool500000300oooo00<00?l00?ooo`3oool0503oool00`000000oooo0?ooo`0@0?ooo`03
0000003oool0oooo01d0oooo00<0o`000000003o00006@3oool00`000000oooo0?ooo`0]0?ooo`03
0000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool002D0oooo00H0gia8000000000000
00000000B02Lgola0?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO303oool00`000000oooo0?oo
o`1j0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?oo
o`080?ooo`H00000:P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo0100oooo00<0
00000?ooo`3oool07P3oool00`3o0000oooo0?ooo`0H0?ooo`030000003oool0oooo02d0oooo00<0
001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`008`3oool02@3ookl0M00004RLg`3o
gi`0W;oo0?ooo`3OW4P004RL0=ooo`0_0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02L
gol02`3oool00`000000oooo0?ooo`1j0?ooo`030000003oool0oooo00`0oooo00<0001o0?ooo`3o
ool0:`3oool00`000000oooo0?ooo`0>0?ooo`P000008P3oool00`000000oooo0?ooo`0D0?ooo`03
0000003oool0oooo0100oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0H0?ooo`03
0000003oool0oooo02d0oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00
8`3oool02@2oM7@0_ooo0?ooo`3o_i`0W=oo0?ooo`3ookl0M01d0;ooo`0_0?ooo`030?oog`2LB000
B9cO0080oooo00@0oooO09a80000B9`0gooo2P3oool00`000000oooo0?ooo`1h0?ooo`H000002`3o
ool00`0007l0oooo0?ooo`0[0?ooo`030000003oool0oooo01H0oooo2000000J0?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0;@3oool5000008L0oooo00<000000?ooo`3oool06@3o
ool00280oooo00T0omnL09boo`3oool0oooO09adW03Oool0oooo0?oOW018M;l0<03oool00`3ooml0
W4P0001d_`030?ooo`030=nLB00004P0W=oo00X0oooo00<000000?ooo`3oool0R@3oool00`0007l0
oooo0?ooo`0[0?ooo`030000003oool0oooo01h0oooo4000000:0?ooo`030000003oool0oooo01@0
oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`0L0?ooo`D00000603oool00`000000
oooo0?ooo`0]0?ooo`030000O`3oool0000008T0oooo00<000000?ooo`3oool06@3oool00280oooo
00T0oknL09cOo`3oool0_g@007Boo`3oool0gia80000M02oool0<@3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0ooooo`00003o000001X000006`3oool00280oooo00@0oono07@0
00000000M;oo=P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00X0oooo00<00000
0?ooo`3oool0R@3oool00`0007l0oooo0?ooo`0[0?ooo`030000003oool0oooo03P0oooo00<00000
0?ooo`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0100oooo2`00000N
0?ooo`030000003oool0oooo02d0oooo00<0001o0?ooo`000000R@3oool00`000000oooo0?ooo`0I
0?ooo`00G03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00T0oooo00<000000?oo
o`3oool0R@3oool00`0007l0oooo0?ooo`0[0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01/0oooo3P00000@0?oo
o`030000003oool0oooo02d0oooo00<0001o0?ooo`000000R@3oool00`000000oooo0?ooo`0I0?oo
o`00G03oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00T0oooo00<000000?oo
o`3oool0R@3oool00`0007l0oooo0?ooo`0[0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02T0oooo1P00000:0?oo
o`030000003oool0oooo02d0oooo00<0001o0?ooo`000000R@3oool00`000000oooo0?ooo`0I0?oo
o`00:03oool0103ookl0M0000000M02ooola0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d
_`3oool02@3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo02<0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo0300oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo02l0oooo1@0000050?oo
o`030000003oool0oooo02d0oooo00<0001o0?ooo`000000R@3oool00`000000oooo0?ooo`0I0?oo
o`008P3oool0103ookl0W9cO0?oo_`1d00040000000307Boo`3oool0oooo0340oooo00H0gia80000
M02oool0oooO09a80018W=l:0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool08`3o
ool010000000oooo0?ooo`3oool9003o00<0oooo00<000000?ooo`3oool0<03oool00`000000oooo
0?ooo`0D0?ooo`030000003oool0oooo00h0oooo1P00000c0?ooo`H000005`3oool5000001<0oooo
00<0001o0?ooo`000000R@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool01P3ogi`0B0000000
00000000001809cOocL0oooo00@0gia800000000000007Bo2`3oool00`000000oooo0?ooo`290?oo
o`030000O`3oool0oooo02<0oooo00<000000?ooo`3oool00P3oool7003o00@0oooo00<000000?oo
o`3oool0<03oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo04D0oooo00<000000?oo
o`3oool01000o`05000000<0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0C0?oo
o`030000O`3oool0000008T0oooo00<000000?ooo`3oool06@3oool002<0oooo00<0oono07@0001d
_ol0B03oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo02<0oooo1@000007003o00D0
0000<P3oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo04D0oooo00@000000?ooo`3o
ool0oooo1`00o`000`3oool000000000000200000140oooo00<000000?ooo`3oool04`3oool00`00
07l0oooo000000290?ooo`030000003oool0oooo01T0oooo000S0?ooo`030;mdM02oool0oooo04P0
oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`0S0?ooo`030000003oool0oooo00<0
oooo1@00o`050?ooo`030000003oool0oooo0300oooo00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`150?ooo`@000001`00o`09000000d0oooo00<0o`000?ooo`3oool04`3oool00`0007l0
oooo000000290?ooo`030000003oool0oooo01T0oooo000R0?ooo`040?noM000000004RL0=ooodP0
oooo0`0000290?ooo`030000O`3oool0oooo02<0oooo00<000000?ooo`3oool00`3oool5003o00D0
oooo00<000000?ooo`3oool0;P3oool5000001@0oooo00<000000?ooo`3oool0A@3oool00`000000
oooo0?ooo`020?ooo`D00?l01@3oool00`000000oooo0?ooo`020?ooo`@00000203oool00`3o0000
00000?l0000D0?ooo`030000O`3oool0000008L0oooo0`00000K0?ooo`00KP3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo02X0oooo0`00o`1@0?ooo`030000003oool0oooo04D0oooo
00<000000?ooo`3oool00P3oool5003o00D0oooo00<000000?ooo`3oool01P3oool3000000@0oooo
00@0o`000?ooo`000000o`00503oool00`0007l0oooo000000290?ooo`030000003oool0oooo01T0
oooo000U0?ooo`060=nLB000000000000000000004P0W=oo@`3oool00`000000oooo0?ooo`290?oo
o`030000O`3oool0oooo02X0oooo0`00o`1@0?ooo`030000003oool0oooo04/0oooo0`00o`0B0?oo
o`<0000000H0o`000?ooo`3oool000000?ooo`3o000C0?ooo`030000O`3oool0000008T0oooo00<0
00000?ooo`3oool06@3oool002<0oooo00T0oono07@00018W=l0omnL09boo`3oool0gia80018W03O
ool0@P3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo02/0oooo00<00?l00?ooo`3o
ool0C`3oool00`000000oooo0?ooo`1;0?ooo`<00?l05@3oool2000000050?ooo`000000oooo0000
003o00004P3oool00`0007l0oooo000000290?ooo`030000003oool0oooo01T0oooo000S0?ooo`09
0;mdM02oool0oooo0?noW02Lgol0oooo0?oo_`1d07@0_ooo0480oooo00<000000?ooo`3oool0R@3o
ool00`0007l0oooo0?ooo`0[0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0C03o
ool00`00o`00oooo0?ooo`0C0?ooo`030?l000000000oooo00@0000000<0o`000?ooo`3oool0403o
ool00`0007l0oooo000000290?ooo`030000003oool0oooo01T0oooo000R0?ooo`090?oOW02L_ol0
oooo0?oog`2LM9`0gooo0?ooo`3ogi`0B7Bo04<0oooo00<000000?ooo`3oool0R@3oool00`0007l0
oooo0?ooo`0[0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0C03oool00`000000
oooo0?ooo`0B0?ooo`030?l0003oool0000000L000002`3oool010000000oooo0?ooo`3oool9003o
0080oooo00<000000?ooo`3oool0P03oool00`000000oooo0?ooo`0I0?ooo`008P3oool02@3o_i`0
W=oo0?ooo`2oM000M;oo0?ooo`3OW4P0001d0;ooo`130?ooo`030000003oool0oooo08T0oooo00<0
001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo04`0oooo00<0
00000?ooo`3oool04`3oool02@3o000000000?ooo`3oool000000?ooo`000000o`000?ooo`020000
00T0oooo00<000000?ooo`3oool00P3oool7003o00<0oooo00<000000?ooo`3oool0P03oool00`00
0000oooo0?ooo`0I0?ooo`008P3oool0103ookl0M0000000001d_om80?ooo`030000003oool0oooo
08T0oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo
04`0oooo00<000000?ooo`3oool0503oool01`3o0000oooo0?ooo`000000oooo0000003o00000`3o
ool2000000L0oooo1@000007003o00@00000PP3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool0
0`000000oooo0?ooo`290?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool0C`3oool0
0`000000oooo0?ooo`1<0?ooo`030000003oool0oooo01@0oooo00H0o`000?ooo`3oool000000?oo
o`3o00060?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`D00?l0103oool00`000000oooo
0?ooo`200?ooo`030000003oool0oooo01T0oooo001^0?ooo`<00000R@3oool00`0007l0oooo0?oo
o`0[0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?oo
o`0E0?ooo`040?l0003oool000000?l000T0oooo0P0000030?ooo`030000003oool0oooo00<0oooo
1@00o`040?ooo`030000003oool0oooo07h0oooo0`00000K0?ooo`00KP3oool00`000000oooo0?oo
o`290?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?oo
o`1<0?ooo`030000003oool0oooo01H0oooo00<0o`000000003o00002`3oool3000000H0oooo00@0
001o003o0000o`000?l0R03oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?oo
o`290?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?oo
o`1<0?ooo`030000003oool0oooo01L0oooo00<0o`000?ooo`3oool03@3oool2000000@0oooo00@0
001o003o0000o`000?l0R03oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?oo
o`290?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?oo
o`1<0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool03`3oool200000080oooo00<0
001o0?ooo`00o`00R@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool00`3ookl0M7Bo0?ooo`19
0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0:`3oool00`000000oooo0?ooo`1?
0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0A
0?ooo`80000000<0001o0?ooo`000000R@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool00`3o
gi`0B00007Boo`190?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0:`3oool00`00
0000oooo0?ooo`1?0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool05`3oool00`00
0000oooo0?ooo`0C0?ooo`<00000R@3oool00`000000oooo0?ooo`0I0?ooo`008`3oool00`3OW4P0
000000000004000000030018W03Oool0oooo0440oooo00<000000?ooo`3oool0R@3oool00`0007l0
oooo0?ooo`0[0?ooo`030000003oool0oooo04l0oooo00<000000?ooo`3oool0C03oool00`000000
oooo0?ooo`0E0?ooo`D000004`3oool0100007l0oooo0000000000280?ooo`030000003oool0oooo
01T0oooo000W0?ooo`030=nLB00004P0W=oo04@0oooo00<000000?ooo`3oool0R@3oool00`0007l0
oooo0?ooo`0Y0?ooo`D00000C`3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo02d0
oooo00@0001o0?ooo`000000oooo0P0000260?ooo`030000003oool0oooo01T0oooo000U0?ooo`03
0?oo_`1d0000B9cO04H0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`03
0000003oool0oooo04`0oooo00<000000?ooo`3oool0;@3oool00`0007l0oooo000000030?ooo`80
0000Q03oool00`000000oooo0?ooo`0I0?ooo`00903oool00`3o_g@0001d0;ooo`170?ooo`030000
003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1:0?ooo`D00000
;@3oool00`0007l0oooo000000050?ooo`030000003oool0oooo0840oooo00<000000?ooo`3oool0
6@3oool00280oooo00<0omnL04P0000000001@0000000`0007@0_ooo0?ooo`110?ooo`<00000R@3o
ool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`0000001P3o
ool2000007l0oooo0`00000K0?ooo`00:03oool00`3ogi`0B00007Boo`130?ooo`030000003oool0
oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
000000P0oooo0P00001o0?ooo`030000003oool0oooo01T0oooo000W0?ooo`030;md000007@0_ooo
04@0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo
07`0oooo00<0001o0?ooo`0000002P3oool2000007d0oooo00<000000?ooo`3oool06@3oool00280
oooo00H0omnL04P0M02oool0gia80000001d_om60?ooo`030000003oool0oooo08T0oooo00<0001o
0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0000000`0oooo00<00000
0?ooo`3oool0NP3oool00`000000oooo0?ooo`0I0?ooo`008P3oool0103ogi`0B00000000018W=m8
0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l
0?ooo`030000O`3oool0000000d0oooo0P0000020?ooo`D00000L`3oool00`000000oooo0?ooo`0I
0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?oo
o`3oool0O03oool00`0007l0oooo0000000?0?ooo`8000000P3oool00`000000oooo0?ooo`1c0?oo
o`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3o
ool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool000000140oooo0`00001e0?ooo`03
0000003oool0oooo01T0oooo000R0?ooo`030?oo_`2LW=l0oooo00<0oooo00@0okmd0000000007@0
_ooo@P3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3o
ool0O03oool500000140oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0I0?ooo`00
8P3oool0103ookl0W9cO0?oo_`1d0004000000030018W03Oool0oooo0440oooo00<000000?ooo`3o
ool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3o
ool04P3oool00`3o000000000000001d0?ooo`030000003oool0oooo01T0oooo000R0?ooo`090?oO
W018000000000000B02Lgol0oooo0?oog`2LB7@0_ooo04<0oooo0`0000290?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0B0?ooo`030?l000000000
oooo00800000L03oool3000001/0oooo000N0?ooo`040?oog`2LB000001d0;oood`0oooo00<00000
0?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o
0?ooo`3oool04@3oool01`3o0000oooo0000003oool0o`000?ooo`000000L@3oool00`000000oooo
0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<0
00000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0@0?ooo`040?l0003oool0oooo00000080oooo
00<0o`000?ooo`000000L03oool00`000000oooo0?ooo`0I0?ooo`008P3oool01@3ogkl0_moo0?oo
o`3oool0_g@000<0000000<0001d0;ooo`3oool0@@3oool00`000000oooo0?ooo`290?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0@0?ooo`040?l0
003oool0oooo00000080oooo00<000000?l0003oool00P00001^0?ooo`030000003oool0oooo01T0
oooo000R0?ooo`030;md0000000000000080000000D0001d0;ooo`3oool0_ibL0=ooo`0K0?ooo`04
0;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P00000000000030000
000504RLg`3oool0oooo0?oOW01800000P0000000`00B9`0gooo0?ooo`0;0?ooo`030000003oool0
oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo00l0oooo00<0o`000000000000001@0000001@3oool0o`000?ooo`3oool0000006d0oooo00<0
00000?ooo`3oool06@3oool00200oooo00D0_g@000000000000004RL0=ooo`030?ooo`030?oog`2L
B000M;oo01/0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0
oooo00<0okmd001d_`3oool00P3oool02@2oM7@0_ooo0?ooo`3o_g@0000007Boo`3oool0omnL04Qd
_`0<0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo0100oooo00@0o`000?ooo`3oool000000P3oool00`000000o`00
0?ooo`030?ooo`030000003oool0oooo06X0oooo00<000000?ooo`3oool06@3oool00280oooo00<0
oknL09cOo`3oool0803oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0
okmd001d_`3oool02@3oool00`3o_g@0B9cO0?ooo`0:0?ooo`030000003oool0oooo08T0oooo00<0
001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo0100oooo00@0
o`000?ooo`3oool000000P3oool00`3o0000oooo0?ooo`040?ooo`030000003oool0oooo06T0oooo
00<000000?ooo`3oool06@3oool002P0oooo00@0oono07@0000007@0_ooo6@3oool00`3ooml0W4P0
04RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<0okmd001d_`3oool0203oool00`3ookl0
M01d0;ooo`0:0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000
oooo0?ooo`1l0?ooo`030000O`3oool0oooo0140oooo00D0o`000?ooo`000000oooo0?l000080?oo
o`800000J@3oool00`000000oooo0?ooo`0I0?ooo`00903oool00`3ookl0M0000000000200000003
0000B02Lgol0oooo01T0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`090?ooo`03
0?noM018W=l0oooo00L0oooo00<0oooO09a8001d_ol02P3oool00`000000oooo0?ooo`290?ooo`03
0000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0B0?ooo`04
0?l000000000oooo0?l000X0oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0I0?oo
o`007P3oool0102oW9`0gooo0?oo_`1d0003000000030000B018B00007Bo01h0oooo00<0_g@00018
W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030=nLB018W=l0oooo00H0oooo00<0oono07@0
0018W=l02P3oool400000?l0001o2@0007l00`000000001o0000O`2A0000O`030?l000000000o`00
00`0001o00<000000000O`0007l0HP0007l4000001/0oooo000M0?ooo`0:0?oog`2LB00000000000
0000B9`0gooo0?ooo`3ogi`0B01809cOoal0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`0;0?ooo`030;md001d_ol0oooo00D0oooo00<0okmd0000001d_ol02P3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`0C0?ooo`030?l0003oool0oooo00`0oooo0P00001U0?ooo`030000003oool0oooo01T0oooo
000S0?ooo`030?oo_`1dB9`0gooo0200oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=oo
o`0:0?ooo`030?oOW01807@0_ooo0080oooo00H0omoO0=ooo`3o_g@000000000M02oool:0?ooo`03
0000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`03
0000O`3oool0oooo01<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`1R0?ooo`03
0000003oool0oooo01T0oooo000R0?ooo`030?oog`2LB7@0_ooo00<0oooo00@0okmd0000000007@0
_ooo6P3oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0gooo00X0oooo00<0oooO09a8
001d_ol00P3oool01@3ookl0M0000000000004P0W=oo00/0oooo00<000000?ooo`3oool0R@3oool0
0`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool04`3oool0
0`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool06@3oool0
0280oooo00@0okmd0018W03Ookl0M000100000000`00B9`0gooo0?ooo`0J0?ooo`030?oOW0180000
M;oo0080oooo00<0gia8001d_`3oool02P3oool00`3ookl0M00004RLg`030?ooo`030=nLW02oool0
oooo00`0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool04@3oool500000100oooo0P00001Q0?ooo`030000003oool0
oooo01T0oooo000R0?ooo`090?oOW0180000000000000018W=l0oooo0?oog`2LB7@0_ooo01d0oooo
00H0gia80000M02oool0oooO09a80018W=l70?ooo`070?noM000B9`0gooo0?ooo`3OW4P0001809cO
o`030?ooo`030?oOW01dW=l0oooo00`0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?oo
o`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0:03oool00`000000oooo0?oo
o`1N0?ooo`030000003oool0oooo01T0oooo00190?ooo`040=nLB00000000000001d_`T0oooo00D0
omnL04P000000000000007Boo`050?ooo`050;md00000000000000000018W=l02@3oool00`000000
oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0
oooo0?ooo`0Y0?ooo`030000003oool0oooo05d0oooo00<000000?ooo`3oool06@3oool006h0oooo
00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo
00<0001o0?ooo`3oool0:P3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo01T0oooo
001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo02/0oooo00<000000?ooo`3oool0F`3oool00`000000oooo0?oo
o`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<00000
0?ooo`3oool0O03oool00`0007l0oooo0?ooo`0/0?ooo`H00000E`3oool00`000000oooo0?ooo`0I
0?ooo`00KP3oool3000008T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?oo
o`030000O`3oool0oooo02d0oooo00<000000?ooo`000000E`3oool3000001/0oooo001^0?ooo`03
0000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`03
0000O`3oool0oooo02h0oooo0P00001I0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000
003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo02l0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`0I0?ooo`008`3o
ool01@2oM00004RL0=ooo`3ookl0M0000080000000<0001d0;ooo`3oool0@@3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`0_0?ooo`800000F03oool00`000000oooo0?ooo`0I0?ooo`008P3oool00`3ogi`0B0000000
00030000000404RLg`3oool0_ibL0=oood80oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo
0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo
0000001G0?ooo`030000003oool0oooo01T0oooo000R0?ooo`040?oog`2LB000001d0;ooo`<0oooo
00<0okmd04RLg`3oool0@P3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo
00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0_0?ooo`040000003oool0oooo000005H0
oooo00<000000?ooo`3oool06@3oool002P0oooo00<0okmd001d_`3oool0@`3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`0_0?ooo`050000003oool0oooo0?ooo`0000004`3oool5000003d0oooo00<000000?ooo`3o
ool06@3oool002H0oooo00<0oooO0;no_`2LW;l0A@3oool00`000000oooo0?ooo`290?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo0080oooo0P00000C0?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool0
6@3oool002H0oooo00H0omno09adB000000000000000M02ooom20?ooo`030000003oool0oooo08T0
oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo02l0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo03d0
oooo00<000000?ooo`3oool06@3oool00280oooo00<0omnL04P0000000000P0000001@0004P0W=oo
0?ooo`2oW9`0gooo0480oooo0`0000290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3o
ool0O03oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool03`3oool00`3o0000oooo0?ooo`0k0?ooo`<000006`3oool00280oooo00<0okmd0000M02oool0
103oool00`2oM000M;oo0?ooo`120?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0
O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool0
1P3oool00`000000oooo0?ooo`0=0?ooo`030?l000000000o`0003h0oooo00<000000?ooo`3oool0
6@3oool002P0oooo00@0oono06H0000007@0_ooo@P3oool00`000000oooo0?ooo`290?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0_0?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool02`3oool0103o0000oooo0000003o000n0?ooo`03
0000003oool0oooo01T0oooo000R0?ooo`040?oo_`2LW=l0oono07@000<0000000<0001809cOo`3o
ool0@P3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3o
ool0O03oool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool02@3oool01P3o0000oooo0?ooo`000000oooo0?l003d0oooo00<000000?ooo`3oool06@3oool0
0280oooo00<0omnL04P0000000000P0000000`00M;l0oooo0?ooo`140?ooo`030000003oool0oooo
08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo
02l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`040?l0003oool0oooo
00000080oooo00<0o`000?ooo`3oool0>P3oool00`000000oooo0?ooo`0I0?ooo`00903oool00`3o
gi`0B01d0;ooo`170?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`00
0000oooo0?ooo`1l0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`060?ooo`050?l000000000oooo0?ooo`0000000P3oool00`3o0000oooo0?ooo`0j
0?ooo`030000003oool0oooo01T0oooo000S0?ooo`030?oo_`1dB9`0gooo04P0oooo00<000000?oo
o`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?oo
o`3oool09`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`040?ooo`030?l0003oool0000000H0000000<0o`000?oo
o`3oool0>@3oool00`000000oooo0?ooo`0I0?ooo`008P3oool00`3ooml0W4Qd0;ooo`030?ooo`04
0?noM0000000001d0;oood80oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?oo
o`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool09`3oool010000000oooo0?ooo`3oool9
003o00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`050?l000000000
oooo0?ooo`0000000P3oool00`3o0000oooo0?ooo`0j0?ooo`030000003oool0oooo01T0oooo000R
0?ooo`040?noM000B9`0gono07@000@0000000<004RL0=ooo`3oool0@@3oool00`000000oooo0?oo
o`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?oo
o`0W0?ooo`030000003oool0oooo0080oooo1`00o`040?ooo`030000003oool0oooo00D0oooo00<0
00000?ooo`3oool0103oool0103o0000oooo0?ooo`0000020?ooo`030?l0003oool0oooo03X0oooo
00<000000?ooo`3oool06@3oool00280oooo00T0omnL04P000000000000004RLg`3oool0oooO09a8
M02oool0@`3oool3000008T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?oo
o`030000O`3oool0oooo02L0oooo1@000007003o00D00000203oool00`000000oooo0?ooo`030?oo
o`060?l0003oool0oooo0000003oool0o`00>`3oool3000001/0oooo001^0?ooo`030000003oool0
oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo02L0oooo00<000000?ooo`3oool00`3oool5003o00D0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`030?ooo`040?l0003oool000000?l003h0oooo00<000000?ooo`3oool06@3o
ool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool09`3oool00`000000oooo0?ooo`030?ooo`D00?l01@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<0o`000000003o0000?P3oool0
0`000000oooo0?ooo`0I0?ooo`008P3oool00`3ookl0W9cO0?ooo`030?ooo`040?noM0000000001d
0;oood80oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0;P3oool3003o0140oooo00<000000?ooo`3oool0103oool0
0`3o0000oooo0?ooo`0m0?ooo`030000003oool0oooo01T0oooo000R0?ooo`040?oo_`2LW=l0oono
07@000@0000000<004RL0=ooo`3oool0@@3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0^0?ooo`<00?l04P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo03d0oooo00<000000?ooo`3oool06@3oool0
0280oooo00T0omnL04P000000000001809cOo`3oool0oooO09a8M02oool0@`3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`0_0?ooo`03003o003oool0oooo0180oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0m0?ooo`030000003oool0oooo01T0oooo000N0?ooo`040?oog`2LB000001d0;oood`0oooo
00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo
00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0000000@00000
?@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3o
ool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0_0?ooo`030000003o
ool0oooo01@0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?ooo`0I0?ooo`008P3oool0
1@3ogkl0_moo0?ooo`3oool0_g@000<0000000<0001d0;ooo`3oool0@@3oool00`000000oooo0?oo
o`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?oo
o`0_0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0@03oool00`000000oooo0?oo
o`0I0?ooo`008P3oool00`2oM000000000000002000000050000M02oool0oooo0;nLW03Oool0@P3o
ool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`0007l0oooo0?ooo`0_0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0?`3o
ool00`000000oooo0?ooo`0I0?ooo`00803oool01@2oM000000000000000B9`0gooo00<0oooo00<0
oooO09a8001d_ol0@`3oool3000008T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?oo
o`0l0?ooo`<000006`3oool00280oooo00<0oknL09cOo`3oool0B@3oool00`000000oooo0?ooo`29
0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`0_
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?ooo`0I
0?ooo`009`3oool01@3ooml0W4P000000000B9`0gooo0480oooo00<000000?ooo`3oool0R@3oool0
0`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0;`3oool0
0`000000oooo0?ooo`0H0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool06@3oool0
02<0oooo00@0_g@00000000007@0_ooo0P3oool00`3ookl0M7Bo0?ooo`120?ooo`030000003oool0
oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo02l0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0
oooo01T0oooo000R0?ooo`060?oOW02L_ol0oooO09a8000007@0_ooo0P3oool00`3OW9`0_ooo0?oo
o`0S0?ooo`040;md00000000000007Boo`<0oooo00<0omnL04P0001d_ol00P3oool01P3ogi`0B000
00000000000004RL0=ooo``0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?oo
o`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0I0?oo
o`030000003oool0oooo03/0oooo00<000000?ooo`3oool06@3oool00280oooo00X0oknL09cOo`3o
ool0oooO09a8000004P0W=oo0?oo_`1dB9`0gooo8`3oool01P3OW4P007Bo0?ooo`3ogi`0B00004RL
g`80oooo00<0omnL04P0001d_ol00P3oool01P3o_g@0000007Boo`3oool0omnL04Qd_``0oooo00<0
00000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03X0oooo00<0
00000?ooo`3oool06@3oool00280oooo00<0oono07Ad_`3oool00P3oool0102oM00000000000M02o
oolS0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol02`3oool00`3o_g@0B9cO0?oo
o`0:0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo02l0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?oo
o`0i0?ooo`030000003oool0oooo01T0oooo000R0?ooo`040?oOW0180000000007BoobP0oooo00<0
oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool:0?ooo`030?oo_`1d07@0_ooo00X0oooo
00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo
00<0001o0?ooo`3oool0;`3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo03P0oooo
00<000000?ooo`3oool06@3oool004h0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cO
o`0:0?ooo`030?oog`2LB000M;oo00X0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?oo
o`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0;@3oool5000001d0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0I0?ooo`00C`3oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO00X0oooo00<0oono07@00018W=l02P3oool400000?l0001o2@0007l0
0`000000001o0000O`3?0000O`030000000007l0001o03<0001o1000000K0?ooo`00C`3oool00`3O
W4P0001809cOo`020?ooo`030?oog`2LB00007Bo00X0oooo00<0okmd0000001d_ol02P3oool00`00
0000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`00
07l0oooo0?ooo`1@0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool06@3oool004l0
oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`060?ooo`060?oOg`3Oool0okmd0000
000007@0_ooo2P3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<00000
0?ooo`3oool0O03oool00`0007l0oooo0?ooo`1A0?ooo`030000003oool0oooo03D0oooo00<00000
0?ooo`3oool06@3oool004l0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=ooo`06
0?ooo`050?oo_`1d000000000000B02Lgol02`3oool00`000000oooo0?ooo`290?ooo`030000O`3o
ool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1B0?ooo`030000003o
ool0oooo03@0oooo00<000000?ooo`3oool06@3oool00500oooo00<0omnL04P0001d_ol00P3oool0
0`3OW4P007Bo0?ooo`070?ooo`030=nLW02oool0oooo00`0oooo00<000000?ooo`3oool0R@3oool0
0`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0D`3oool0
0`000000oooo0?ooo`0c0?ooo`030000003oool0oooo01T0oooo001A0?ooo`060=nLB00007@0_ooo
0?oog`2LB000B9cO203oool00`3ogi`0M9cO0?ooo`0<0?ooo`030000003oool0oooo08T0oooo00<0
001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo05@0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`0I0?ooo`00DP3oool0103OW4P0000000000000
M;l:0?ooo`050;md00000000000000000018W=l02@3oool00`000000oooo0?ooo`290?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1D0?ooo`030000
003oool0oooo0380oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0R@3o
ool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0E@3o
ool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0
oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0
oooo05H0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0I0?ooo`00KP3oool30000
08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo
05L0oooo00<000000?ooo`3oool0;@3oool3000001/0oooo001^0?ooo`030000003oool0oooo08T0
oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo05L0
oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`1;0?ooo`H000001`3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo01T0oooo
001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo04h0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o
0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo04h0oooo00<00000
0?ooo`3oool02@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo01T0oooo001^0?oo
o`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?oo
o`030000O`3oool0oooo04h0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3o
ool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo04h0oooo00<000000?ooo`3o
ool02P3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000
003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo04h0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0Z0?ooo`030000
003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3o
ool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo04h0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo01T0oooo001^0?ooo`<00000R@3oool0
0`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0CP3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo02L0oooo0`00000K0?ooo`00KP3oool00`00
0000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`00
07l0oooo0?ooo`1>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0:03oool00`00
0000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0
oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1>0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo
0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo
0?ooo`1>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<0
00000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1>0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool09P3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`29
0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1>
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0I
0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?oo
o`3oool0O03oool00`0007l0oooo0?ooo`1>0?ooo`030000003oool0oooo0140oooo00<000000?oo
o`3oool0903oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`03
0000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1>0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0I0?ooo`00
KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0
O03oool00`0007l0oooo0?ooo`1>0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0
8`3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3o
ool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`160?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0<0?ooo`030000003o
ool0oooo0280oooo00<000000?ooo`3oool06@3oool006h0oooo0`0000290?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`160?ooo`030000003oool0
oooo00X00?l00P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo01l0oooo0`00000K
0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?oo
o`3oool0O03oool00`0007l0oooo0?ooo`160?ooo`040000003oool0oooo0?ooo`P00?l00`3oool0
0`000000oooo0?ooo`0=0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool06@3oool0
06h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo
07`0oooo00<0001o0?ooo`3oool0AP3oool4000000P00?l01000000@0?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0
oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0AP3oool00`000000
oooo0?ooo`020?ooo`H00?l0103oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo01l0
oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo
0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0AP3oool00`000000oooo
0?ooo`030?ooo`D00?l0103oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?oo
o`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0C03oool4003o01P0oooo00<0
00000?ooo`3oool07P3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`29
0?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1=
0?ooo`<00?l06@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01T0oooo001^0?oo
o`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?oo
o`030000O`3oool0oooo04h0oooo00<00?l00?ooo`3oool06@3oool00`000000oooo0?ooo`0L0?oo
o`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3o
ool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo04h0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo01T0oooo001^0?ooo`<00000
R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0
CP3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo01T0oooo0`00000K0?ooo`00KP3o
ool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`0007l0oooo0?ooo`1>0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06P3o
ool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0
oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1>0?ooo`030000003oool0
oooo01/0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`00
0000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`00
07l0oooo0?ooo`1>0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool06@3oool00`00
0000oooo0?ooo`0I0?ooo`00A`3oool0102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo
00<0oooo00<0oono07A8W03Oool01@3oool01P3ogi`0B00000000000000004RL0=ooo``0oooo00<0
00000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0
001o0?ooo`3oool0CP3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool06@3oool004H0oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oO
W0180000M;oo00@0oooo00<0gia804RLg`3oool0103oool01P3o_g@0000007Boo`3oool0omnL04Qd
_``0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo
07`0oooo00<0001o0?ooo`3oool0CP3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool06@3oool004D0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0
001809cOo`080?ooo`030?oo_`1d07@0_ooo00T0oooo00<0okmd04RLg`3oool02P3oool00`000000
oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0
oooo0?ooo`1>0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`0I0?ooo`00A@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`P0
oooo00<0okmd001d_`3oool0203oool00`3ookl0M01d0;ooo`0:0?ooo`030000003oool0oooo08T0
oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo04h0
oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01T0
oooo00150?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02Lgol0203oool00`3ooml0W4Q8
09cOo`080?ooo`030?oog`2LB000M;oo00X0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo
0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0CP3oool00`000000oooo
0?ooo`0P0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool06@3oool004H0oooo00<0
_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`090?ooo`030?oOW018M;l0oooo00L0oooo00<0
oono07@00018W=l02P3oool400000?l0001o2@0007l00`000000001o0000O`3=0000O`0300000000
07l0001o0200001o00<000000000O`0007l04P0007l4000001/0oooo00160?ooo`030=nLB00004P0
W=oo0080oooo00<0oooO09a80000M;l02P3oool00`2oM000M;oo0?ooo`060?ooo`030?noM0000000
M;oo00X0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0
oooo07`0oooo00<0001o0?ooo`3oool0CP3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo01@0oooo00<000000?ooo`3oool06@3oool004H0oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`090?ooo`030?oOW018B9`0gooo00<0oooo00H0omoO0=ooo`3o_g@000000000M02o
ool:0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?oo
o`1l0?ooo`030000O`3oool0oooo04/0oooo1P00000R0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool06@3oool004H0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=oo
o`0:0?ooo`030=nLB018W=l0oooo0080oooo00D0oono07@000000000001809cOo`0;0?ooo`030000
003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo07<0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0I0?ooo`00A`3o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo00D0oooo00<0oooO0=oOo`3oool00P3o
ool00`3ookl0M01d0;ooo`030?ooo`030=nLW02oool0oooo00`0oooo00<000000?ooo`3oool0R@3o
ool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0M03o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo01T0oooo00180?ooo`060=nLB00007@0
_ooo0?oog`2LB000B9cO1`3oool00`2oM7@0_ooo0?ooo`020?ooo`030?noM000M;l0oooo0080oooo
00<0omnL07BLg`3oool0303oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo
00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1e0?ooo`030000003oool0oooo0140oooo
00<000000?ooo`3oool06@3oool004T0oooo00@0gia800000000000007Bo203oool00`3ookl0M000
00000003000000030000B02Lgol0oooo0080oooo00D0_g@0000000000000000004RLg`090?ooo`03
0000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`03
0000O`3oool0oooo07D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0I0?ooo`00
KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0
O03oool00`0007l0oooo0?ooo`1f0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool0
6@3oool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003o
ool0oooo07`0oooo00<0001o0?ooo`3oool0M`3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool0
0`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo07L0oooo00<000000?ooo`3oool03`3oool0
0`000000oooo0?ooo`0I0?ooo`00KP3oool3000008T0oooo00<0001o0?ooo`3oool0O@3oool00`00
0000oooo0?ooo`1l0?ooo`030000O`3oool0oooo07P0oooo00<000000?ooo`3oool0303oool30000
01/0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000
oooo0?ooo`1l0?ooo`030000O`3oool0oooo07T0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo
00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1j0?ooo`030000003oool0oooo00`0oooo
00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?oo
o`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0NP3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o
0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000O`3oool0oooo07/0oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`290?oo
o`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03oool00`0007l0oooo0?ooo`1l0?oo
o`030000003oool0oooo00X0oooo00<000000?ooo`3oool06@3oool006h0oooo00<000000?ooo`3o
ool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3o
ool0O03oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo01T0oooo001^0?ooo`030000
003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`000000oooo0?ooo`1l0?ooo`030000
O`3oool0oooo07d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0I0?ooo`00KP3o
ool00`000000oooo0?ooo`290?ooo`030000O`3oool0oooo07d0oooo00<000000?ooo`3oool0O03o
ool00`0007l0oooo0?ooo`290?ooo`030000003oool0oooo01T0oooo001^0?ooo`<00000R@3oool0
0`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0oooo00<0001o0?ooo`3oool0Q`3oool3
000001/0oooo001^0?ooo`030000003oool0oooo08T0oooo00<0001o0?ooo`3oool0O@3oool00`00
0000oooo0?ooo`1l0?ooo`030000O`3oool0oooo08T0oooo00<000000?ooo`3oool06@3oool006h0
oooo00<000000?ooo`3oool0R@3oool00`0007l0oooo0?ooo`1m0?ooo`030000003oool0oooo07`0
oooo00<0001o0?ooo`3oool0R@3oool00`000000oooo0?ooo`0I0?ooo`00KP3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0O@3oool00`000000oooo
0?ooo`1l0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`0I0?ooo`00KP3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01T0oooo001^
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool06@3oool006h0ooooo`00003o0000
01X000006`3oool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?ooool0ooooX`3o
ool00?l0ooooo`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?ooool0ooooX`3oool00?l0oooo
o`3ooonS0?ooo`00o`3ooooo0?oooj<0oooo003o0?ooool0ooooX`3oool00?l0oooo=@3oool00`3o
oml0W4P0000000020000000307Boo`3oool0oooo01H0oooo00<0oooO09a8000000000P0000000`1d
_ol0oooo0?ooo`3o0?ooodP0oooo003o0?ooocL0oooo00<0okmd0000M02oool06`3oool00`3o_g@0
001d0;ooo`3o0?oood/0oooo003o0?ooocL0oooo00<0oono07@0001d_ol06`3oool00`3ookl0M000
07Boo`3o0?oood/0oooo003o0?ooocL0oooo00<0oooO09a80018W=l06`3oool00`3ooml0W4P004RL
g`3o0?oood/0oooo003o0?ooocP0oooo00<0gia80000000000000P0000004@1d_ol0oooo0?ooo`3o
ool0omnL04P00000000004RL0=ooo`2oM0000000001d_`3oool0okmd0000000004P0W=oo00P0oooo
00<0gia80000000000000P000000201d_ol0oooo0?ooo`3oool0okmd0000000007@0_ooo0P3oool0
1@3o_g@000000000000004P0W=oo00<0oooo00@0okmd0000000004P0W=oo1P3oool01`3o_g@00000
0000M02oool0oooo0?oOW01800000P0000001`1d_ol0gia800000000B9`0gooo0?ooo`3OW4P00P00
00000`0004P0W=oo0?ooo`030?ooo`050=nLB000000000000000001d_ol00P3oool0103o_g@00000
0000M02oool50?ooo`030=nLB000B9`0gooo0080oooo00<0_g@000000000M;l00P3oool01@3OW4P0
0000000000000000M;oo0080oooo00<0_g@0001d_`3oool01@3oool01@3OW4P00000000000000000
M;oo0>D0oooo003o0?ooocP0oooo00L0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00h0
omnL04P0B02Lgol0omnL04Qd_`2oM00007Bo0?noW02Lgol0okmd0000B02Lgml0W7BL0=ooo`L0oooo
00L0okmd0000B02Lgol0oooo0?oog`2LB00007Bo0080oooo00`0omnL04P0M02ookl0W9cO0?ooo`3o
gi`0B01809cOo`3oool0omnL04P00000M;l20?ooo`050?noM00004P0W=oO09adW03Oool01@3oool0
5@3ogi`0B01d0;oo_`2LW=l0oooo0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?oO
W01807@0_ooo0?ooo`3ookl0M00004RLg`020?ooo`030?oOW0180000M;oo0080oooo00L0_g@0001d
_`3oool0omnL04P0M02ookl0W9cO00D0oooo00<0okmd0000M02oool00P3oool01`3OW4P004RL0=no
M01d_ol0omnL04P0001d_ol00P3oool01P2oM00007Bo0?ooo`3o_g@0001d0;ooo`@0oooo00<0omnL
04P0001d_ol00P3oool00`2oM00007Bo0?ooo`3S0?ooo`00o`3ooolh0?ooo`030?oo_`1d0000M;oo
0080oooo0100oooO09a80018W=l0oooo0?oog`2LB000B9cO0?ooo`3o_g@0M7Ad0018W03Oool0oooo
0?oo_`1d0000M;oo2@3oool00`3ookl0M00007Boo`020?ooo`0<0?oog`2LB000B9cO0?ooo`3ookl0
M00007Boo`3OW9`0_ooo0=nLB000B9`0gooo0P3oool01`3ogi`0B00004RLg`3oool0oono07@0001d
_ol01`3oool01`3ookl0M00007Boo`3OW9`0_ooo0;md0000M;l00`3oool01`2oM000000007Boo`3o
ool0oono07@0001d_ol00P3oool01`3ookl0M00007Boo`3oool0gia80018W03Oool00`3oool01`3O
W9`0_ooo0?oo_`1d0000M;oo0=nLW02oool0103oool00`3ookl0M00007Boo`020?ooo`070?oOW018
04P0W=oo0?ooo`3OW4P004RL0=ooo`030?ooo`050=nLW02oool0omnL04P0001d_ol0103oool00`3O
W4P004RL0=ooo`030?ooo`030=nLW02oool0oooo0>80oooo003o0?ooocP0oooo00<0oooO09a80018
W=l00`3oool02`3ogi`0B01809cOo`3oool0_g@00018W03Oool0oooo0;md000004P0W=oo0080oooo
00<0_g@0001d_`3oool0203oool00`3ooml0W4P004RLg`030?ooo`060?oOW01804P0W=oo0?ooo`2o
M000B9cO0P3oool00`3o_g@0001d0;ooo`030?ooo`060=nLB00007@0_ooo0?ooo`2oM00007Bo203o
ool01`2oM000B9cO0?ooo`3oool0okmd0000M02oool00P3oool00`3ookl0M00004RLg`020?ooo`03
0;md0018W=l0oooo0080oooo00@0gia80018W03Oool0okmd0`0000000`18W=l0oooo0?ooo`020?oo
o`030;md0018W=l0oooo00D0oooo00X0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`3oool0
okmd0`0000002018W=l0oooo0?ooo`3oool0oooO09a80000B9`0gooo0`3oool01@3o_g@000000000
00000000B9cO0>H0oooo003o0?ooocT0oooo00<0gia80000M02oool00P3oool02`3ooml0W4P004RL
g`3oool0okmd0000M02oool0oooo0?oo_`1d0000M;oo0080oooo00<0gia80018W03Oool02@3oool0
0`3OW4P0001d0;ooo`020?ooo`0;0?oog`2LB000B9cO0?ooo`3OW4P004RL0=ooo`3oool0oono07@0
001d_ol00`3oool01`3ogi`0B01809cOo`3oool0gia80018W03Oool01`3oool01`3OW4P004RL0=oo
o`3oool0oono07@0001d_ol00`3oool01`2oM00004RL0=ooo`3oool0gia80018W03Oool00P3oool0
2P3ogi`0B01809cOo`3ookl0M01809cOo`3oool0gia800000018W=l20?ooo`030=nLB000B9`0gooo
00H0oooo00<0gia80018B018M;l00P3oool02P2oM00007Bo0?ooo`3ookl0M01809cOo`3oool0gia8
00000018W=l20?ooo`030;md0018M7@0M9cO00<0oooo00L0oono07@0B02Lgol0oooo0=nLB0000000
B9cO0>@0oooo003o0?ooocT0oooo00@0okmd00000000B9`0gooo0P3oool01P3OW4P007Bo0?oOg`3O
okl0M00007Boo`80oooo00L0_g@0001d_`3oool0oooo0?oOW01804P0W=oo00T0oooo00@0okmd0000
0000B9`0gooo0P3oool01P3OW4P007Bo0?ooo`3ogi`0B01d0;ooo`80oooo00<0gia80000M02oool0
0P3oool01`3ookl0M00007Boo`3oool0omnL04P0B02Lgol01`3oool00`3ogi`0B01d0;ooo`020?oo
o`030?noM00007@0_ooo0080oooo00L0gia80000M02oool0oooo0?oOW01804P0M;oO0080oooo00l0
oono07@0001d_ol0oooo0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0omnL04P0M02oool0
1P3oool05@3o_g@0001d0;noM018W=l0oooo0?noM00007@0_ooo0?ooo`3o_g@004RL0=ooo`3oool0
oono07@00018W=l0oooo0?noM000B9`0_ia807Boo`030?ooo`070?noM000B9`0gooo0?ooo`3ookl0
M00004RLg`3S0?ooo`00o`3oooli0?ooo`0<0?oo_`1d0000M9bL04P0M02oool0oooO09a80000M;l0
omnL04Qd_`2oM00007Bo0P3oool01`3o_g@004RL0=ooo`3oool0oooO09a80018W=l02@3oool0303o
okl0M00007BLW01807@0_ooo0?oog`2LB00007Bo0?ooo`3ookl0M00007Boo`<0oooo00/0_g@00000
M02oool0oooo0?noM00007@0_ooo0?ooo`3ooml0W4P004RLg`070?ooo`030?oo_`1d0000M;oo00<0
oooo0100gia80000M02oool0oooo0;md000004P0W=oo0?ooo`3ookl0M00004Q8B018W=l0oooo0?no
M00004P0W=oo0P3oool02P3OW4P0001d0;ooo`3oool0gia80018W03Oool0oono07@0001d_ol60?oo
o`080?oo_`1d04P0W=oo0=nLB000M;l0omnL04P0B02Lgol20?ooo`0F0=nLB00007@0_ooo0?ooo`3O
W4P004RL0=ooo`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooo0=nLB00007@0_ooo0?ooo`3O
W4P004RL0=ooon80oooo003o0?ooocP0oooo00d0omnL04P000000000B9cO0?oo_`1d000000000000
M02oool0oooo0?noM000000007Bo0080oooo00D0omnL04P0B02Lgol0oono07@000030000000307Bo
o`3oool0oooo00@0oooo00T0omnL04P000000000B9cO0?oo_`1d000000000000M02oool00P3oool0
0`2oM000B9cO0?ooo`030?ooo`080?noM000000000000000001d_ol0oooo0?oo_`1d000300000003
07Boo`3oool0oooo00@0oooo00<0_g@004RLg`3oool00`3oool01P3ogi`0B00000000018W9`0M000
04RLg`80oooo00<0_g@0001d_`3OW4P00P0000000`18W=l0oooo0?ooo`020?ooo`050?oOW0180000
00000000001d_ol00P3oool00`2oM000B9cO0?ooo`050?ooo`080?oog`2LB000B9cO0?ooo`3o_g@0
00000000B02Lgol30?ooo`0=0?oOW018000000000000001d_ol0omnL04P000000000M;oo0?ooo`2o
M000001d0;ooo`020?ooo`050?oOW018000000000000001d_ol0h`3oool00?l0ooooC03oool00`3o
_g@0001d0;ooo`0B0?ooo`030=nLB000B9`0gooo00/0oooo00<0okmd0000M02oool01`3oool00`3O
W4P004RL0=ooo`0;0?ooo`030=nLB000B9`0gooo00h0oooo00<0gia80018W03Oool01P3oool00`3O
W4P007Bo0?ooo`3n0?ooo`00o`3ooom=0?ooo`030=nLM02Lgol0oooo0140oooo00<0omnL04P0M02o
ool0303oool00`3OW7@0W=oo0?ooo`060?ooo`030?oOW01807@0_ooo00/0oooo00<0omnL04P0B02L
gol03P3oool00`3ogi`0B01d0;ooo`060?ooo`030?noM00007@0_ooo0?h0oooo003o0?ooof40oooo
00<0oooO09a8001d_ol05@3oool00`3ooml0W4P007Boo`0;0?ooo`030?oo_`1d0000B9cO00h0oooo
00<0oooO09a8001d_ol01P3oool00`3ookl0M01809cOo`3n0?ooo`00o`3ooomR0?ooo`030=nLB018
W=l0oooo01D0oooo00<0gia804RLg`3oool02`3oool00`3OW4P007Bo0?ooo`0>0?ooo`030=nLB018
W=l0oooo00H0oooo00<0_g@007Boo`3oool0o@3oool00?l0ooooH03oool0103ooml0W4P000000000
M;lD0?ooo`040?oog`2LB0000000001d_`X0oooo00D0oooO09a800000000001d0;ooo`0<0?ooo`04
0?oog`2LB0000000001d_`D0oooo00@0oono07@00000000007BooP3oool00?l0ooooo`3ooonS0?oo
o`00o`3ooooo0?oooj<0oooo003o0?ooool0ooooX`3oool00?l0ooooo`3ooonS0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 672.125}, {415.063, 0}} -> {-1.48191, -1.18504, \
0.00588025, 0.00806403}}]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Reading a data file", "Section"],

Cell[TextData[{
  "For the present purpose we assume that data are listed in a file that \
contains simply two or three columns of numbers.  The spacing between those \
columns does not matter and need not be uniform, but we assume that every \
line contains the same number of numerical entries.  Although there are \
methods in ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " for handling headers, comments, variations in file structure, and other \
variations, it is simpler for now to edit the data file so that it has the \
simplest possible structure.  The command ",
  ButtonBox["Import",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[file]", "InlineInput"],
  " then reads the contents of ",
  StyleBox["file", "InlineInput"],
  " into a list where each member of the list is a list of the numbers on one \
line of ",
  StyleBox["file", "InlineInput"],
  ".  The file name is a string.  If you have not properly adjusted ",
  StyleBox["$Path", "InlineInput"],
  ", you will need to edit the following example to specify the correct file \
location for your system.  Once successful, the statement"
}], "Text"],

Cell[BoxData[
    \(\(gepdata = Import["\<gep.dat\>"];\)\)], "Input"],

Cell[TextData[{
  "reads a data for the electric form factor of the proton, ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep\)]],
  ", into the symbol ",
  StyleBox["gepdata", "InlineInput"],
  ".  Below we plot the data as a function of ",
  Cell[BoxData[
      \(TraditionalForm\`Q\^2\)]],
  ", the square of the momentum transfer for scattering a high-energy \
electron from the internal charge distribution of the proton.  The charge \
density is related to the Fourier transform of the form factor."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(ErrorListPlot[gepdata, Frame \[Rule] True, 
        FrameLabel \[Rule] {\*"\"\<\!\(Q\^2\) [(GeV/c\!\(\()\^2\)\)]\>\"", \ \
\*"\"\<\!\(G\_Ep\)\>\""}, 
        PlotLabel \[Rule] "\<Proton Electric Form Factor\>"];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.17191 0.0147151 0.595771 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.19572 -0.0125 -5 -15.6875 ]
[.19572 -0.0125 5 0 ]
[.36763 -0.0125 -5 -15.6875 ]
[.36763 -0.0125 5 0 ]
[.53954 -0.0125 -5 -15.6875 ]
[.53954 -0.0125 5 0 ]
[.71145 -0.0125 -5 -15.6875 ]
[.71145 -0.0125 5 0 ]
[.88336 -0.0125 -5 -15.6875 ]
[.88336 -0.0125 5 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -36.1563 -31.375 ]
[.5 -0.0125 36.1563 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .13387 -19 -7.84375 ]
[-0.0125 .13387 0 7.84375 ]
[-0.0125 .25302 -19 -7.84375 ]
[-0.0125 .25302 0 7.84375 ]
[-0.0125 .37218 -19.25 -7.84375 ]
[-0.0125 .37218 0 7.84375 ]
[-0.0125 .49133 -19 -7.84375 ]
[-0.0125 .49133 0 7.84375 ]
[-0.0125 .61049 -10 -7.84375 ]
[-0.0125 .61049 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -35.6875 -11.3125 ]
[-0.01875 .30902 -19.25 11.3125 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -71.7188 0 ]
[.5 .63053 71.7188 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.19572 0 m
.19572 .00625 L
s
gsave
.19572 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.36763 0 m
.36763 .00625 L
s
gsave
.36763 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.53954 0 m
.53954 .00625 L
s
gsave
.53954 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.71145 0 m
.71145 .00625 L
s
gsave
.71145 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.88336 0 m
.88336 .00625 L
s
gsave
.88336 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.125 Mabswid
.05819 0 m
.05819 .00375 L
s
.09257 0 m
.09257 .00375 L
s
.12696 0 m
.12696 .00375 L
s
.16134 0 m
.16134 .00375 L
s
.2301 0 m
.2301 .00375 L
s
.26448 0 m
.26448 .00375 L
s
.29887 0 m
.29887 .00375 L
s
.33325 0 m
.33325 .00375 L
s
.40201 0 m
.40201 .00375 L
s
.43639 0 m
.43639 .00375 L
s
.47078 0 m
.47078 .00375 L
s
.50516 0 m
.50516 .00375 L
s
.57392 0 m
.57392 .00375 L
s
.6083 0 m
.6083 .00375 L
s
.64269 0 m
.64269 .00375 L
s
.67707 0 m
.67707 .00375 L
s
.74583 0 m
.74583 .00375 L
s
.78021 0 m
.78021 .00375 L
s
.8146 0 m
.8146 .00375 L
s
.84898 0 m
.84898 .00375 L
s
.91774 0 m
.91774 .00375 L
s
.95212 0 m
.95212 .00375 L
s
.98651 0 m
.98651 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -97.1563 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Q) show
71.625 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
79.438 15.813 moveto
(@) show
83.375 15.813 moveto
(H) show
87.188 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(GeV) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
109.688 15.813 moveto
(\\220) show
113.500 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(c) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
118.750 15.813 moveto
(L) show
122.563 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
127.375 15.813 moveto
(D) show
131.313 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .13387 m
.00625 .13387 L
s
gsave
-0.0125 .13387 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .25302 m
.00625 .25302 L
s
gsave
-0.0125 .25302 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .37218 m
.00625 .37218 L
s
gsave
-0.0125 .37218 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .49133 m
.00625 .49133 L
s
gsave
-0.0125 .49133 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .61049 m
.00625 .61049 L
s
gsave
-0.0125 .61049 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .0445 m
.00375 .0445 L
s
0 .07429 m
.00375 .07429 L
s
0 .10408 m
.00375 .10408 L
s
0 .16366 m
.00375 .16366 L
s
0 .19345 m
.00375 .19345 L
s
0 .22323 m
.00375 .22323 L
s
0 .28281 m
.00375 .28281 L
s
0 .3126 m
.00375 .3126 L
s
0 .34239 m
.00375 .34239 L
s
0 .40197 m
.00375 .40197 L
s
0 .43175 m
.00375 .43175 L
s
0 .46154 m
.00375 .46154 L
s
0 .52112 m
.00375 .52112 L
s
0 .55091 m
.00375 .55091 L
s
0 .5807 m
.00375 .5807 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -99.7813 -12.2188 Mabsadd m
1 1 Mabs scale
currentpoint translate
72.3125 12.2188 translate
90 rotate
-72.3125 -12.2188 translate
0 24.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
71.688 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Ep) show
81.625 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.19572 .61178 m
.19572 .61803 L
s
.36763 .61178 m
.36763 .61803 L
s
.53954 .61178 m
.53954 .61803 L
s
.71145 .61178 m
.71145 .61803 L
s
.88336 .61178 m
.88336 .61803 L
s
.125 Mabswid
.05819 .61428 m
.05819 .61803 L
s
.09257 .61428 m
.09257 .61803 L
s
.12696 .61428 m
.12696 .61803 L
s
.16134 .61428 m
.16134 .61803 L
s
.2301 .61428 m
.2301 .61803 L
s
.26448 .61428 m
.26448 .61803 L
s
.29887 .61428 m
.29887 .61803 L
s
.33325 .61428 m
.33325 .61803 L
s
.40201 .61428 m
.40201 .61803 L
s
.43639 .61428 m
.43639 .61803 L
s
.47078 .61428 m
.47078 .61803 L
s
.50516 .61428 m
.50516 .61803 L
s
.57392 .61428 m
.57392 .61803 L
s
.6083 .61428 m
.6083 .61803 L
s
.64269 .61428 m
.64269 .61803 L
s
.67707 .61428 m
.67707 .61803 L
s
.74583 .61428 m
.74583 .61803 L
s
.78021 .61428 m
.78021 .61803 L
s
.8146 .61428 m
.8146 .61803 L
s
.84898 .61428 m
.84898 .61803 L
s
.91774 .61428 m
.91774 .61803 L
s
.95212 .61428 m
.95212 .61803 L
s
.98651 .61428 m
.98651 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .13387 m
1 .13387 L
s
.99375 .25302 m
1 .25302 L
s
.99375 .37218 m
1 .37218 L
s
.99375 .49133 m
1 .49133 L
s
.99375 .61049 m
1 .61049 L
s
.125 Mabswid
.99625 .0445 m
1 .0445 L
s
.99625 .07429 m
1 .07429 L
s
.99625 .10408 m
1 .10408 L
s
.99625 .16366 m
1 .16366 L
s
.99625 .19345 m
1 .19345 L
s
.99625 .22323 m
1 .22323 L
s
.99625 .28281 m
1 .28281 L
s
.99625 .3126 m
1 .3126 L
s
.99625 .34239 m
1 .34239 L
s
.99625 .40197 m
1 .40197 L
s
.99625 .43175 m
1 .43175 L
s
.99625 .46154 m
1 .46154 L
s
.99625 .52112 m
1 .52112 L
s
.99625 .55091 m
1 .55091 L
s
.99625 .5807 m
1 .5807 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -132.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Proton) show
99.438 15.813 moveto
(Electric) show
140.188 15.813 moveto
(Form) show
169.875 15.813 moveto
(Factor) show
202.438 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.002 w
.02475 .60073 m
.02475 .60332 L
s
.015 w
.02475 .60203 Mdot
.002 w
.02515 .59515 m
.02515 .59806 L
s
.015 w
.02515 .5966 Mdot
.002 w
.02548 .59265 m
.02548 .5952 L
s
.015 w
.02548 .59392 Mdot
.002 w
.02582 .58532 m
.02582 .59025 L
s
.015 w
.02582 .58779 Mdot
.002 w
.02615 .58611 m
.02615 .58863 L
s
.015 w
.02615 .58737 Mdot
.002 w
.02649 .57852 m
.02649 .58419 L
s
.015 w
.02649 .58135 Mdot
.002 w
.02682 .57759 m
.02682 .57939 L
s
.015 w
.02682 .57849 Mdot
.002 w
.02716 .57467 m
.02716 .57815 L
s
.015 w
.02716 .57641 Mdot
.002 w
.02749 .57105 m
.02749 .57283 L
s
.015 w
.02749 .57194 Mdot
.002 w
.02783 .56831 m
.02783 .57164 L
s
.015 w
.02783 .56997 Mdot
.002 w
.02816 .56458 m
.02816 .56655 L
s
.015 w
.02816 .56556 Mdot
.002 w
.0285 .55883 m
.0285 .56265 L
s
.015 w
.0285 .56074 Mdot
.002 w
.02917 .55326 m
.02917 .55618 L
s
.015 w
.02917 .55472 Mdot
.002 w
.0295 .55175 m
.0295 .55519 L
s
.015 w
.0295 .55347 Mdot
.002 w
.02983 .55029 m
.02983 .55427 L
s
.015 w
.02983 .55228 Mdot
.002 w
.0305 .5371 m
.0305 .54553 L
s
.015 w
.0305 .54132 Mdot
.002 w
.03184 .5262 m
.03184 .53237 L
s
.015 w
.03184 .52928 Mdot
.002 w
.03318 .51708 m
.03318 .52111 L
s
.015 w
.03318 .51909 Mdot
.002 w
.04616 .42346 m
.04616 .4329 L
s
.015 w
.04616 .42818 Mdot
.002 w
.05647 .3728 m
.05647 .3787 L
s
.015 w
.05647 .37575 Mdot
.002 w
.07023 .31024 m
.07023 .31853 L
s
.015 w
.07023 .31439 Mdot
.002 w
.08054 .27809 m
.08054 .28396 L
s
.015 w
.08054 .28102 Mdot
.002 w
.09085 .25842 m
.09085 .26431 L
s
.015 w
.09085 .26136 Mdot
.002 w
.10117 .23249 m
.10117 .24377 L
s
.015 w
.10117 .23813 Mdot
.002 w
.11492 .2028 m
.11492 .21626 L
s
.015 w
.11492 .20953 Mdot
.002 w
.12352 .18734 m
.12352 .1924 L
s
.015 w
.12352 .18987 Mdot
.002 w
.13555 .16574 m
.13555 .17945 L
s
.015 w
.13555 .17259 Mdot
.002 w
.14758 .16533 m
.14758 .18582 L
s
.015 w
.14758 .17557 Mdot
.002 w
.1579 .14015 m
.1579 .14784 L
s
.015 w
.1579 .144 Mdot
.002 w
.1854 .12692 m
.1854 .13605 L
s
.015 w
.1854 .13149 Mdot
.002 w
.21291 .09576 m
.21291 .10168 L
s
.015 w
.21291 .09872 Mdot
.002 w
.25589 .0795 m
.25589 .08577 L
s
.015 w
.25589 .08263 Mdot
.002 w
.32465 .05339 m
.32465 .06099 L
s
.015 w
.32465 .05719 Mdot
.002 w
.08793 .25958 m
.08793 .27388 L
s
.015 w
.08793 .26673 Mdot
.002 w
.09275 .25064 m
.09275 .26494 L
s
.015 w
.09275 .25779 Mdot
.002 w
.09962 .21787 m
.09962 .23217 L
s
.015 w
.09962 .22502 Mdot
.002 w
.08914 .23872 m
.08914 .26375 L
s
.015 w
.08914 .25124 Mdot
.002 w
.10976 .21311 m
.10976 .23336 L
s
.015 w
.10976 .22323 Mdot
.002 w
.10805 .21483 m
.10805 .22151 L
s
.015 w
.10805 .21817 Mdot
.002 w
.15962 .14322 m
.15962 .14656 L
s
.015 w
.15962 .14489 Mdot
.002 w
.22666 .09169 m
.22666 .09395 L
s
.015 w
.22666 .09282 Mdot
.002 w
.27824 .06603 m
.27824 .06985 L
s
.015 w
.27824 .06794 Mdot
.002 w
.32809 .05478 m
.32809 .05728 L
s
.015 w
.32809 .05603 Mdot
.002 w
.347 .05095 m
.347 .05321 L
s
.015 w
.347 .05208 Mdot
.002 w
.38998 .04335 m
.38998 .04597 L
s
.015 w
.38998 .04466 Mdot
.002 w
.44843 .03648 m
.44843 .03791 L
s
.015 w
.44843 .03719 Mdot
.002 w
.53438 .02862 m
.53438 .03005 L
s
.015 w
.53438 .02934 Mdot
.002 w
.62034 .02503 m
.62034 .02658 L
s
.015 w
.62034 .0258 Mdot
.002 w
.62549 .02366 m
.62549 .02636 L
s
.015 w
.62549 .02501 Mdot
.002 w
.70629 .02093 m
.70629 .02247 L
s
.015 w
.70629 .0217 Mdot
.002 w
.84038 .01817 m
.84038 .0193 L
s
.015 w
.84038 .01874 Mdot
.002 w
.97619 .01617 m
.97619 .01761 L
s
.015 w
.97619 .01689 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{365.75, 226.063},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008T0001Db000`400?l00000o`00003oo`3ooooo
0?ooobH0oooo003o0?ooool0oooo9P3oool00?l0ooooo`3ooolV0?ooo`00o@3oool01@3ogkl0_moo
0?ooo`3oool0gia800@0000000<0001d0;ooo`3oool03@3oool01@3OW4P00000000000000000M;oo
00<0oooo00<0oooO09a8B02Lgol08@3oool00`3ooml0W4Q809cOo`0;0?ooo`030?oOW018B9`0gooo
00X0oooo00D0gia8000000000000000007Boo`320?ooo`00o@3oool00`3ooml0W4P0000000020000
00070000M02oool0oooo0?ooo`3ookl0M01d0;ooo`0=0?ooo`030=nLB00007@0_ooo00D0oooo00<0
_gA809cOo`3oool08P3oool00`3OW4P0B9cO0?ooo`0;0?ooo`030?oOW018M;l0oooo00/0oooo00<0
oono07@0001d_ol0`P3oool00?l0oooo00<0_g@004RLg`3oool01@3oool00`3ookl0M7Bo0?ooo`0<
0?ooo`030=nLB000B9`0gooo00@0oooo00<0gia804RLg`3oool08`3oool00`3ogi`0B7Bo0?ooo`0<
0?ooo`030=nLB018W=l0oooo00X0oooo00<0oono07@0001d_ol0`P3oool00?l0oooo0@3oool00`3O
W4P007Bo0?ooo`0C0?ooo`030=nLB000B9`0gooo00<0oooo00<0oono07@0M02oool0903oool00`3o
okl0M01d0;ooo`0<0?ooo`030?oo_`1d04P0W=oo00X0oooo00<0oono07@0001d_ol0`P3oool00?l0
oooo00<0omnL04P0000000000P0000000`00B9`0gooo0?ooo`0?0?ooo`030=nLB000B9`0gooo00<0
oooo00<0gia80018W03Oool01@3oool00`3ogi`0B00000000003000000030018W03Oool0oooo0080
oooo00D0gia8000000000000000007Boo`040?ooo`030=nLB000B9`0gooo00T0oooo00<0_g@007Bo
o`3oool00P3oool01@3OW4P000000000000004P0W=oo00@0oooo00<0oooO09a80018W=l02P3oool0
0`3ookl0M00007Boo`320?ooo`00oP3oool0103ogi`0B0000018W03Oool20?ooo`030;md00000000
B9cO00l0oooo00<0gia80018W03Oool00`3oool00`2oM00007Bo0?ooo`040?ooo`030;md00000000
B9cO00<0oooo00<0omnL04P0001d_ol00P3oool00`3ogi`0B00007Boo`020?ooo`030;md0000M;l0
oooo0080oooo00<0okmd00000018W=l02@3oool02`3o_g@007Bo0?ooo`3oool0okmd0000001d_ol0
oooo0?oog`2LB000B9cO00@0oooo00<0gia80018W03Oool02@3oool00`3ookl0M00007Boo`320?oo
o`00oP3oool00`2oM000001809cOo`040?ooo`030?noM000000007Bo00h0oooo00<0gia80018W03O
ool00P3oool00`3ookl0M00004RLg`040?ooo`030?noM0000000B9cO00@0oooo00<0oooO09a80000
M;l00P3oool00`3OW4P004RL0=ooo`030?ooo`080=nLW02oool0oooo0?ooo`3ogi`0B00004Q8B01d
_ol80?ooo`070?oo_`1dB9`0gooo0?ooo`3OW4P004RL0=ooo`030?ooo`030=oOg`3oool0oooo0080
oooo00<0gia80000M02oool02@3oool00`3ookl0M00007Boo`320?ooo`00o@3oool0103ooml0W4P0
0018W03Oool50?ooo`030=nLB0000000M;oo00d0oooo00<0gia80018W03Oool00P3oool00`3ookl0
M00007Boo`030?ooo`040?oog`2LB000001d0;ooo`D0oooo00D0_g@00000M02oool0oooo0?noM003
0000000304RLg`3oool0oooo00<0oooo00D0oono07@00018W=l0_gA809cOo`070?ooo`070?oog`2L
B4P0W=oo0?ooo`3o_g@0001d0;ooo`080?ooo`030?noM00007@0_ooo00T0oooo00<0oono07@0001d
_ol0`P3oool00?d0oooo00@0oooO09a8000007@0_ooo1@3oool0103ooml0W4P00000M02oool<0?oo
o`030=nLB000B9`0gooo0080oooo00<0oono07@0001d_ol00`3oool0103ookl0M0000000M02oool5
0?ooo`0;0?noM0000000B9cO0?ooo`3ookl0M01809cOo`3oool0gia800000018W=l00`3oool01P3o
oml0W4P004RLg`3ookl0M4RL0=ooo`L0oooo00H0gia804RLg`3oool0oono07@0001d_ol80?ooo`03
0?noM00004P0W=oo00T0oooo00<0oono07@0001d_ol0`P3oool00?h0oooo00<0_g@00000B02Lgol0
1P3oool00`3OW4P0000004RLg`0<0?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@0001d_ol0
0`3oool0103ooml0W4P00000M02oool30?ooo`030?oo_`1d000000000080000000T0001d0;ooo`3o
_g@004RL0=ooo`3oool0oono07@00018W=l00P3oool01P3ooml0W4P0001d_`3oool0okmd001d_`L0
oooo00/0omnL04Q8W03Oool0oooo0=nLB00007@0_ooo0?ooo`3ookl0M00007Boo`030?ooo`030?no
M00007@0_ooo00T0oooo00<0oono07@0001d_ol0`P3oool00?h0oooo00<0gia800000018W=l01P3o
ool0103ogi`0B0000018W03Oool;0?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@00018W=l0
103oool00`2oM000001809cOo`0;0?ooo`070=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`020?oo
o`060;md0000B9`0gooo0?ooo`2oM000M;oo1P3oool00`3ooml0W4Qd0;ooo`020?ooo`060;md0000
07@0_ooo0?ooo`3OW4P0B9cO0`3oool00`3OW4P0001d0;ooo`090?ooo`030?oo_`1d0000M;oo0<80
oooo003n0?ooo`040?oo_`1d000004RL0=ooo`D0oooo00@0oooO09a80000B9`0Yg@00`0000000`00
07@0_ooo0?ooo`050?ooo`030=nLB000B9`0gooo0080oooo00<0oooO09a80018W=l0103oool00`3o
_g@0000004RLg`0<0?ooo`050?oOW018000000000000001d_ol00`3oool01`3OW4P0001d0;ooo`3o
ool0oooO09a8B02Lgol01P3oool00`2oM000M;oo0?ooo`020?ooo`050?oOW018000000000000B02L
gol00`3oool0103OW4P004RL0=ooo`2oM003000000030000M02oool0oooo0080oooo00<0oono07@0
001d_ol0`P3oool00?l0oooo00<0okmd00000018W=l01@3oool02@3ooml0W4P00018W03Oool0okmd
0018W03Oool0omnL09boo`060?ooo`030=nLB000B9`0gooo00<0oooo00<0gia80018W03Oool00`3o
ool0103ooml0W4P00000B02Lgol60?ooo`030=nLW02oool0oooo00X0oooo00<0okmd0000B02Lgol0
0P3oool00`3ogi`0B01d0;ooo`050?ooo`030?noM000M;l0oooo00T0oooo00<0oooO09a80018W=l0
0P3oool01@3o_g@004RL0=ooo`3ogi`0W;oo00<0oooo00<0oono07@0001d_ol0`P3oool00?l0oooo
0@3oool00`3OW4P0000007Boo`040?ooo`030?oo_`1d000007Bo0080oooo00<0omnL04Q8W03Oool0
1`3oool00`3OW4P004RL0=ooo`030?ooo`030?oo_`1d04P0W=oo00@0oooo00@0oono07@0000007@0
_ooo1@3oool00`2oM000M;oo0?ooo`0:0?ooo`030?oOW0180000M;oo00<0oooo00<0okmd001d_`3o
ool0103oool00`3ookl0M01d0;ooo`090?ooo`030?oo_`1d04P0W=oo00<0oooo00<0omnL04Q8W03O
ool0103oool00`3ookl0M00007Boo`320?ooo`00o`3oool20?ooo`030?noM000000007Bo0080oooo
00@0oooO09a8000004P0W=oo0`3oool00`3ogi`0B4RL0=ooo`060?ooo`030=nLB000B9`0gooo00@0
oooo00<0gia804RLg`3oool0103oool0103ooml0W4P00000001d_ol30?ooo`030?noM0000000B9cO
00/0oooo00<0omnL04P00018W=l0103oool00`2oM00007Bo0?ooo`040?ooo`030;md001d_ol0oooo
00P0oooo00<0gia804RLg`3oool0103oool00`3ogi`0B4RL0=ooo`030?ooo`030?oo_`1d0000M;oo
0<80oooo003o0?ooo`@0oooo00H0_g@000000000000000000000M02oool50?ooo`030?oOW01dW=l0
oooo00D0oooo00<0gia80000M02oool01@3oool00`2oM000B9cO0?ooo`050?ooo`030?oo_`1d0000
00000080000000@007Bo0?oog`2LM9`0gooo2@3oool0202oM0000000000000000000001d0;ooo`3o
ool0gia80P0000000`0007@0_ooo0?ooo`020?ooo`030=nLB018W=l0oooo00L0oooo00<0okmd001d
_`3oool01P3oool00`3ogi`0M9cO0?ooo`020?ooo`030?oo_`1d0000M;oo0<80oooo003o0?oooa00
oooo00<0_gA809cOo`3oool0103oool01@3OW4P00000000000000000M;oo00@0oooo00<0giad09cO
o`3oool09@3oool00`3ogi`0B4RL0=ooo`060?ooo`030?oo_`1dM;l0oooo00P0oooo00<0_gA809cO
o`3OW4P00`0000000`1d_ol0oooo0?ooo`300?ooo`00o`3oool<0?ooo`060?oOg`3Oool0oooo0?oo
o`3OW4P0B9cOAP3oool01P3ogml0gooo0?ooo`3oool0gia804RLglL0oooo003o0?ooo``0oooo00H0
oono07A8W03Oool0oooO09a80018W=m60?ooo`060?oo_`1dB9`0gooo0?oog`2LB000B9cOa`3oool0
0?l0oooo3@3oool01@3ogi`0B00000000000B9`0gooo04L0oooo00D0omnL04P00000000004RL0=oo
o`370?ooo`00o`3ooooo0?ooobH0oooo003o0?ooool0oooo9P3oool00?l0ooooo`3ooolV0?ooo`00
o`3ooooo0?ooobH0oooo003o0?ooool0oooo9P3oool00?l0ooooo`3ooolV0?ooo`00o`3ooooo0?oo
obH0oooo003o0?ooool0oooo9P3oool006L0oooo00@0_g@0000000000000M;oo@P3oool00`3ookl0
M000000000020000000304RLg`3oool0oooo03l0oooo00<0gia80000000000000`0000000`18W=l0
oooo0?ooo`0o0?ooo`030?oo_`1d000000000080000000<0M;oo0?ooo`3oool0@P3oool00`3ooml0
W4P004RLg`120?ooo`060?oOW0180000000000000000B9`0goooD`3oool006H0oooo00H0gia8001d
_`3oool0omnL04P00018W=m30?ooo`030?noM0000000M;oo04<0oooo00<0okmd001d_`3oool00P3o
ool00`2oM7@0_ooo0?ooo`0o0?ooo`030?noM000000007Bo0080oooo00<0_gA809cOo`3oool0@`3o
ool00`3OW4P004RL0=ooo`110?ooo`060?noM0000000M;oo0?ooo`3ogi`0B7BoD`3oool006D0oooo
00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`120?ooo`030?oog`2LB00007Bo04@0oooo
00<0okmd001d_`3oool0B03oool00`3ogi`0B4RL0=ooo`130?ooo`030?oOW01804P0W=oo04H0oooo
00<0okmd04RLg`3oool0D@3oool006D0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018
W03Ooom20?ooo`030=nLB000B9`0gooo04@0oooo00<0okmd001d_`3oool0A`3oool00`3ookl0M000
07Boo`0o0?ooo`030?noM0000000000000@0000000<004RL0=ooo`3oool0@`3oool00`3ookl0M01d
0;ooo`1A0?ooo`00I@3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo0480oooo00<0
omnL04P0B02Lgol0A@3oool00`3o_g@0B9cO0?ooo`160?ooo`030?oOW0180000M;oo03l0oooo00<0
oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`150?ooo`030?oog`2LB000M;oo0540oooo001V
0?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l0@P3oool00`3ookl0M00007Boo`16
0?ooo`030=nLB018W=l0oooo04@0oooo00@0omnL04P0000004P0W=oo@03oool01`3ookl0M01d0;oo
o`3oool0okmd0018W03Oool0A@3oool00`3ookl0M00004RLg`1A0?ooo`00IP3oool00`3OW4P00018
09cOo`020?ooo`030?oog`2LB00007Bo04<0oooo00<0_g@0001d_`3oool0AP3oool00`2oM000M;oo
0?ooo`110?ooo`050?oOW018000000000000001d_ol0@P3oool01P3ogi`0B4RL0=ooo`3ogi`0B018
09cOodD0oooo00<0okmd0000001d_ol0D@3oool006H0oooo00<0omnL04P00018W=l00`3oool00`2o
M00004RL0=ooo`120?ooo`030=nLB000B9`0gooo04H0oooo00<0omnL04P0M02oool0@`3oool00`3o
okl0M000001d_`130?ooo`050?noM000M;l0oooO09a80018W=l0@P3oool01P3ogml0gooo0?noM000
0000001d0;oooe40oooo001V0?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0
@P3oool00`3ogi`0B01809cOo`160?ooo`030?oog`2LB000M;oo04D0oooo00<0gia8001d_`3oool0
@P3oool0103OW4P0B9cO0;md0000M;m20?ooo`050?oo_`1d000000000000B02Lgol0DP3oool006L0
oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`120?ooo`030?oo_`1d0000M;oo04H0
oooo00<0oono07@00018W=l0A@3oool00`3ogi`0B01809cOo`130?ooo`040=nLB01dM7@0001d0;oo
od80oooo00<0gibL0;ooo`3oool0D`3oool006P0oooo00H0gia80000M02oool0oooO09a80018W=m4
0?ooo`030;md0000M;l0oooo0440oooo00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo0480oooo
00H0_g@007Boo`3oool0gia80000B02Lgom40?ooo`030;md000004P0W=oo0480oooo00<0omnL07BL
g`3oool0D`3oool006T0oooo00@0gia800000000000007Bo@`3oool01@3ookl0M0000000000007@0
_ooo0480oooo00D0omnL04P000000000000007Boo`140?ooo`040;md00000000000004RLgdD0oooo
00<0oooO09a80018W=l0@`3oool01@2oM0000000000000000000B9cO0500oooo003o0?ooool0oooo
9P3oool00?l0ooooo`3ooolV0?ooo`00o`3ooooo0?ooobH0oooo003o0?ooool0oooo9P3oool00?l0
ooooo`3ooolV0?ooo`00o`3ooooo0?ooobH0oooo003o0?ooool0oooo9P3oool00?l0ooooo`3ooolV
0?ooo`00o`3ooooo0?ooobH0oooo003o0?ooool0oooo9P3oool00600ooooo`00002R000002<0oooo
001A0?ooo`040;md00000000000007Boo`/0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool08@3oool00500oooo00H0
gia8001d_`3oool0omnL04P00018W=l:0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0A03oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3o
ool0A@3oool00`000000oooo0?ooo`140?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3o
ool08@3oool004l0oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`090?ooo`030000
003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00C`3oool00`3ooml0W4P004RL
g`020?ooo`040?oog`2LB00004RL0=ooo`P0oooo00<000000?ooo`3oool0o`3ooonC0?ooo`800000
2@3oool00`000000oooo0?ooo`0Q0?ooo`00C`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB000
04P0W=oo00P0oooo00<000000?ooo`3oool0o`3ooomK0?ooo`800000=@3oool4000000P0oooo00<0
00000?ooo`3oool08@3oool00500oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`08
0?ooo`@00000o`3ooolR0?ooo`800000=@3oool4000003<0oooo1P0000040?ooo`@000008`3oool0
0500oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`080?ooo`030000003oool0oooo
0?l0oooo8P3oool4000003<0oooo1P00000b0?ooo`H000001`3oool00`000000oooo0?ooo`0Q0?oo
o`00D03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00L0oooo00<000000?ooo`3o
ool0oP3oool2000000030?ooo`000000000001d0oooo1P00000b0?ooo`H00000<`3oool4000000P0
oooo00<000000?ooo`3oool08@3oool00500oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P0
04RL0=ooo`070?ooo`030000003oool0oooo0=/0oooo0P00000P0?ooo`L00000703oool6000003<0
oooo1000000e0?ooo`8000002@3oool00`000000oooo0?ooo`0Q0?ooo`00D@3oool00`3ogi`0B000
07Boo`020?ooo`030=nLB000M;l0oooo00L0oooo00<000000?ooo`3oool0fP3oool4000001h0oooo
2@00000L0?ooo`@00000=@3oool200000440oooo00<000000?ooo`3oool08@3oool00580oooo00H0
gia80000M02oool0oooO09a80018W=l80?ooo`030000003oool0oooo0=T0oooo1P00000M0?ooo`T0
00007@3oool2000007T0oooo00<000000?ooo`3oool08@3oool005<0oooo00@0gia8000000000000
07Bo2@3oool00`000000oooo0?ooo`3I0?ooo`H000007P3oool7000009T0oooo00<000000?ooo`3o
ool08@3oool00600oooo00<000000?ooo`3oool0]`3oool200000240oooo1000000P0?ooo`800000
00<0oooo000000000000VP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?oo
o`2f0?ooo`@000008@3oool200000<00oooo00<000000?ooo`3oool08@3oool00600oooo00<00000
0?ooo`3oool0]@3oool600000>80oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?oo
o`3oool0W`3oool2000001@0oooo1P00003R0?ooo`030000003oool0oooo0240oooo001P0?ooo`03
0000003oool0oooo09h0oooo1000000D0?ooo`@00000h`3oool00`000000oooo0?ooo`0Q0?ooo`00
H03oool2000009h0oooo1P00000D0?ooo`800000hP3oool3000002<0oooo001P0?ooo`030000003o
ool0oooo08d0oooo0P00000>0?ooo`H00000nP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool0
0`000000oooo0?ooo`2<0?ooo`@000003P3oool400000?/0oooo00<000000?ooo`3oool08@3oool0
0600oooo00<000000?ooo`3oool0P`3oool4000000@0oooo1P00000>0?ooo`800000o03oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`220?ooo`H000000`3oool600000?l0
oooo3@3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`210?ooo`P00000
0`3oool400000?l0oooo3P3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?oo
o`210?ooo`P00000103oool200000?l0oooo3`3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool0
0`000000oooo0?ooo`220?ooo`H00000o`3ooolF0?ooo`030000003oool0oooo0240oooo001P0?oo
o`030000003oool0oooo0700oooo0P00000A0?ooo`@00000o`3ooolG0?ooo`030000003oool0oooo
0240oooo001P0?ooo`030000003oool0oooo06l0oooo1000003o0?ooob/0oooo00<000000?ooo`3o
ool08@3oool00600oooo00<000000?ooo`3oool0KP3oool600000?l0oooo:P3oool00`000000oooo
0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`1^0?ooo`H00000o`3ooolZ0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo06l0oooo1000003o0?ooob/0oooo00<00000
0?ooo`3oool08@3oool00600oooo0P00001a0?ooo`800000o`3ooolZ0?ooo`<000008`3oool00600
oooo00<000000?ooo`3oool0I`3oool200000?l0oooo=@3oool00`000000oooo0?ooo`0Q0?ooo`00
H03oool00`000000oooo0?ooo`1V0?ooo`@00000o`3ooold0?ooo`030000003oool0oooo0240oooo
001P0?ooo`030000003oool0oooo06D0oooo1P00003o0?oooc<0oooo00<000000?ooo`3oool08@3o
ool00600oooo00<000000?ooo`3oool0I@3oool600000?l0oooo<`3oool00`000000oooo0?ooo`0Q
0?ooo`00H03oool00`000000oooo0?ooo`1K0?ooo`8000002@3oool400000?l0oooo=03oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`1J0?ooo`@000002@3oool200000?l0
oooo=@3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`1E0?ooo`800000
0P3oool600000?l0oooo?`3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?oo
o`1D0?ooo`@0000000<0oooo0000000000001000003o0?ooocl0oooo00<000000?ooo`3oool08@3o
ool00600oooo00<000000?ooo`3oool0D`3oool6000000030?ooo`000000000000800000o`3ooom0
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo05<0oooo1P0000020?oo
o`800000o`3ooom10?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo05@0
oooo1000003o0?ooodH0oooo00<000000?ooo`3oool08@3oool00600oooo0P00001F0?ooo`800000
o`3ooom50?ooo`<000008`3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q
0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600
oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000
003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?oo
o`3o0?oooih0oooo00<000000?ooo`3oool08@3oool004<0oooo00@0_g@0000000000000M;oo0`3o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB0000000000000<0000000<0B9cO0?ooo`3oool0203o
ool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00480oooo00H0gia8
001d_`3oool0omnL04P00018W=l20?ooo`030?oOW0180000M;oo00<0oooo00<0okmd001d_`3oool0
0P3oool00`2oM7@0_ooo0?ooo`080?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo
0?ooo`0Q0?ooo`00@@3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0
okmd001d_`3oool0303oool00`000000oooo0?ooo`190?ooo`800000o`3ooomC0?ooo`030000003o
ool0oooo0240oooo00110?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo203o
ool00`3o_g@007Bo0?ooo`0;0?ooo`030000003oool0oooo04P0oooo1000003o0?oooe80oooo00<0
00000?ooo`3oool08@3oool00440oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`09
0?ooo`030?noM018W=l0oooo00X0oooo00<000000?ooo`3oool0A`3oool600000?l0ooooD@3oool0
0`000000oooo0?ooo`0Q0?ooo`00@P3oool00`2oM00004RL0=ooo`020?ooo`030?oOW0180000B9cO
00X0oooo00<0gia804RLg`3oool02@3oool4000004H0oooo1P00003o0?ooodh0oooo1000000S0?oo
o`00@P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00/0oooo00<0_g@007Boo`3o
ool0203oool00`000000oooo0?ooo`180?ooo`@00000o`3ooomB0?ooo`030000003oool0oooo0240
oooo00120?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool02P3oool00`3ogi`0B01d
0;ooo`080?ooo`030000003oool0oooo03h0oooo0P0000090?ooo`800000o`3ooomC0?ooo`030000
003oool0oooo0240oooo00120?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0
2P3oool00`3ooml0W4P007Boo`080?ooo`030000003oool0oooo03d0oooo1000003o0?oooed0oooo
00<000000?ooo`3oool08@3oool004<0oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?oo
o`0:0?ooo`030?oo_`1d0000B9cO00P0oooo00<000000?ooo`3oool0?03oool600000?l0ooooG03o
ool00`000000oooo0?ooo`0Q0?ooo`00A03oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo
00L0okmd0018W03Oool0oooo0=nLB00004P0W=oo00P0oooo00<000000?ooo`3oool0?03oool70000
0?l0ooooF`3oool00`000000oooo0?ooo`0Q0?ooo`00A@3oool0103OW4P0000000000000M;l90?oo
o`050?oOW018000000000000001d_ol02@3oool00`000000oooo0?ooo`0m0?ooo`H00000o`3ooomK
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo03h0oooo1000003o0?oo
oe`0oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0?`3oool200000?l0
ooooG@3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo
00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q
0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600
oooo0P00000k0?ooo`030000003oool0oooo0?l0ooooG`3oool3000002<0oooo001P0?ooo`030000
003oool0oooo03D0oooo0P0000030?ooo`030000003oool0oooo0?l0ooooH@3oool00`000000oooo
0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0d0?ooo`@000000P3oool200000?l0ooooHP3o
ool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0c0?ooo`X00000o`3ooomQ
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo03<0oooo2`00003o0?oo
of00oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0=03oool:00000?l0
ooooH03oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0e0?ooo`800000
0P3oool400000?l0ooooH@3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?oo
o`0j0?ooo`800000o`3ooomR0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0
oooo0300oooo0P0000080?ooo`030000003oool0oooo0?l0ooooH@3oool00`000000oooo0?ooo`0Q
0?ooo`00H03oool00`000000oooo0?ooo`0_0?ooo`@000001`3oool00`000000oooo0?ooo`3o0?oo
of40oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0;P3oool600000?l0
ooooJP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0^0?ooo`H00000
o`3ooomZ0?ooo`030000003oool0oooo0240oooo001P0?ooo`800000<03oool400000?l0ooooJ@3o
ool3000002<0oooo001P0?ooo`030000003oool0oooo0300oooo0P00003o0?ooof`0oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo
0240oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00
H03oool00`000000oooo0?ooo`0/0?ooo`800000o`3ooom`0?ooo`030000003oool0oooo0240oooo
001P0?ooo`030000003oool0oooo02/0oooo1000003o0?ooofl0oooo00<000000?ooo`3oool08@3o
ool00600oooo00<000000?ooo`3oool0:P3oool600000?l0ooooKP3oool00`000000oooo0?ooo`0Q
0?ooo`00H03oool00`000000oooo0?ooo`0Z0?ooo`H00000o`3ooom^0?ooo`030000003oool0oooo
0240oooo001P0?ooo`030000003oool0oooo02T0oooo1P00003o0?ooofl0oooo00<000000?ooo`3o
ool08@3oool00600oooo00<000000?ooo`3oool0:03oool600000?l0ooooL03oool00`000000oooo
0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0V0?ooo`P00000o`3ooom`0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo02H0oooo1`00003o0?ooog40oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool09@3oool900000?l0ooooL03oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool2000002D0oooo2P00003o0?ooofh0oooo0`00000S0?ooo`00
H03oool00`000000oooo0?ooo`0T0?ooo`T00000o`3oooma0?ooo`030000003oool0oooo0240oooo
001P0?ooo`030000003oool0oooo02D0oooo100000000`3oool000000000003o0?ooog80oooo00<0
00000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool09P3oool2000000<0oooo00<00000
0?ooo`3oool0o`3ooom`0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo
02D0oooo100000020?ooo`030000003oool0oooo0?l0ooooL03oool00`000000oooo0?ooo`0Q0?oo
o`00H03oool00`000000oooo0?ooo`0T0?ooo`H00000o`3ooomd0?ooo`030000003oool0oooo0240
oooo001P0?ooo`030000003oool0oooo0280oooo00<000000?ooo`0000001@00003o0?ooog@0oooo
00<000000?ooo`3oool08@3oool004<0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0B000
07Boo`050?ooo`030?oog`2LB000B9cO00/0oooo00<000000?ooo`3oool08P3oool00`000000oooo
0?ooo`0400000?l0ooooM@3oool00`000000oooo0?ooo`0Q0?ooo`00@P3oool01P3OW4P007Bo0?oo
o`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol01P3oool00`3OW4P004RL0=ooo`0:0?ooo`03
0000003oool0oooo0280oooo00@000000?ooo`3oool0oooo0P00003o0?ooogH0oooo00<000000?oo
o`3oool08@3oool00440oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`0:0?ooo`03
0?oOW01804P0W=oo00X0oooo00<000000?ooo`3oool08@3oool200000?l0ooooN`3oool00`000000
oooo0?ooo`0Q0?ooo`00@@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0
oooo00<0okmd000000000000100000000`00B9`0gooo0?ooo`070?ooo`030000003oool0oooo0200
oooo1000003o0?ooogX0oooo00<000000?ooo`3oool08@3oool00440oooo00<0oooO09a80000M;l0
0`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB4P0W=oo0080oooo00<0_g@0001d_`3oool0
2@3oool00`000000oooo0?ooo`0O0?ooo`H00000o`3ooomi0?ooo`030000003oool0oooo0240oooo
00120?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01P3oool01`3ookl0M01d0;oo
o`3oool0okmd0018W03Oool02@3oool4000001h0oooo1`00003o0?ooogD0oooo1000000S0?ooo`00
@P3oool00`3OW4P0001809cOo`020?ooo`030?oog`2LB00007Bo00L0oooo00H0omnL04Q8W03Oool0
omnL04P0B02Lgol90?ooo`030000003oool0oooo0200oooo1`00003o0?ooogL0oooo00<000000?oo
o`3oool08@3oool00480oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?ooo`05
0?noM000M;l0oooO09a80018W=l02@3oool00`000000oooo0?ooo`0Q0?ooo`H00000o`3ooomg0?oo
o`030000003oool0oooo0240oooo00120?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018
W03Oool0203oool0103OW4P0B9cO0;md0000M;l90?ooo`030000003oool0oooo0200oooo1P00003o
0?ooogP0oooo00<000000?ooo`3oool08@3oool004<0oooo00<0omnL04P0001d_ol00P3oool00`3O
W4P007Bo0?ooo`090?ooo`040=nLB01dM7@0001d0;ooo`P0oooo00<000000?ooo`3oool0803oool6
00000?l0ooooN03oool00`000000oooo0?ooo`0Q0?ooo`00A03oool01P3OW4P0001d0;ooo`3ooml0
W4P004RLg`/0oooo00<0_g@00000B02Lgol0203oool00`000000oooo0?ooo`0O0?ooo`H00000o`3o
oomi0?ooo`030000003oool0oooo0240oooo00150?ooo`040=nLB00000000000001d_``0oooo00<0
oooO09a80018W=l0203oool00`000000oooo0?ooo`0O0?ooo`H00000o`3ooomi0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo0200oooo1000003o0?ooogX0oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool08@3oool200000?l0ooooN`3oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0N0?ooo`80000000<0oooo0000003o
ool0o`3ooomk0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo01d0oooo
1000003o0?ooogd0oooo00<000000?ooo`3oool08@3oool002`0oooo00<0gia80000000000000P00
00000`18W=l0oooo0?ooo`0/0?ooo`030000003oool0oooo01`0oooo1P00003o0?ooog`0oooo00<0
00000?ooo`3oool08@3oool002X0oooo00<0okmd00000018W=l0103oool00`3OW4P007Bo0?ooo`0/
0?ooo`8000007@3oool600000?l0ooooNP3oool3000002<0oooo000Y0?ooo`030=nLB000M;l0oooo
00H0oooo00<0okmd07Boo`3oool0:`3oool00`000000oooo0?ooo`0M0?ooo`@00000o`3ooomm0?oo
o`030000003oool0oooo0240oooo000X0?ooo`030?noM018W=l0oooo00P0oooo00<0_ibL0=ooo`3o
ool0:P3oool00`000000oooo0?ooo`0N0?ooo`800000o`3ooomn0?ooo`030000003oool0oooo0240
oooo000W0?ooo`030?oog`2LW;l0oooo00T0oooo00<0gibL0;ooo`3oool0:P3oool00`000000oooo
0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool002L0oooo00<0omnL09boo`3oool02@3o
ool00`3OW9`0_ooo0?ooo`0Z0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?oo
o`0Q0?ooo`009`3oool00`3o_i`0W=oo0?ooo`090?ooo`030;mdM02oool0oooo02X0oooo00<00000
0?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo000W0?ooo`030=nLW02oool0oooo
00<0oooo00<0okmd07Boo`3oool00P3oool00`3ooml0W7BL0=ooo`0[0?ooo`030000003oool0oooo
0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`009`3oool00`3o_g@0M;oo0?ooo`030?ooo`04
0?noM018W=l0oooO09a80080000000<0B9cO0?ooo`3oool0:P3oool00`000000oooo0?ooo`3o0?oo
oih0oooo00<000000?ooo`3oool08@3oool002L0oooo00<0oono07@0M02oool00`3oool01P3o_g@0
0000000000000000001d0;ooobd0oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0
oooo0240oooo000W0?ooo`090?oo_`1d000000000000001d_ol0oooo0?noM000B9`0gooo0300oooo
00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo000W0?ooo`030=nLB000
0000M;oo00<0oooo00<0okmd07Boo`3oool0<03oool00`000000oooo0?ooo`3o0?oooih0oooo00<0
00000?ooo`3oool08@3oool003H0oooo00<0_gAd0;ooo`3oool09`3oool00`000000oooo0?ooo`0J
0?ooo`800000o`3ooon20?ooo`030000003oool0oooo0240oooo000c0?ooo`050?noM00000000000
0000M02oool0:03oool00`000000oooo0?ooo`0I0?ooo`@00000o`3ooon10?ooo`030000003oool0
oooo0240oooo000^0?ooo`030=nLB000000000000080000000D007Bo0?ooo`3oool0gibL0;ooo`0X
0?ooo`8000006@3oool600000?l0ooooOP3oool3000002<0oooo000^0?ooo`060=nLB01d_ol0oooo
0?oog`2LM9`0gooo0P3oool00`3OW9`0_ooo0?ooo`0W0?ooo`030000003oool0oooo01P0oooo1P00
003o0?oooh00oooo00<000000?ooo`3oool08@3oool002h0oooo00X0gibL0;ooo`3oool0oooO09a8
00000000M;oo0?oog`2LB7@0_ooo:03oool00`000000oooo0?ooo`0I0?ooo`@00000o`3ooon10?oo
o`030000003oool0oooo0240oooo000^0?ooo`050=nLW02oool0oono07@00000M;l00P3oool00`2o
M000001d0;ooo`0X0?ooo`030000003oool0oooo01X0oooo0P00003o0?oooh80oooo00<000000?oo
o`3oool08@3oool002h0oooo00<0gibL0;ooo`3oool00`3oool00`3OW4P004RL0=ooo`020?ooo`03
0;nLW03Oool0oooo02@0oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240
oooo000^0?ooo`030=nLB0000000B9cO00L0oooo00<0omnL04Q8W03Oool09@3oool00`000000oooo
0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00340oooo00@0_ibL0=ooo`3oool0okmd
100000000`00B9`0gooo0?ooo`0T0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo
0?ooo`0Q0?ooo`00<@3oool01`2oM000000000000000B9`0gooo0;mdM02oool0:03oool00`000000
oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00340oooo00<0oono07Ad_`3oool0
0P3oool00`3OW9`0_ooo0?ooo`0W0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo
0?ooo`0Q0?ooo`00<@3oool01P3OW9`0_ooo0?ooo`3oool0oono07Ad_bT0oooo00<000000?ooo`3o
ool0o`3ooonN0?ooo`030000003oool0oooo0240oooo000`0?ooo`070?oog`2LM9`0gooo0?ooo`3o
_g@0001d0;ooo`0Y0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?oo
o`00<@3oool0103OW4P0000000000000M;l[0?ooo`030000003oool0oooo0?l0ooooWP3oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool200000?l0ooooW@3oool3000002<0oooo001P0?ooo`030000
003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?oo
o`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3o
ool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo
001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00@`3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo00<0oooo00D0oooO09a8000000000000
001d_`0;0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00@P3o
ool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL04P0001d_ol00`3oool01`3OW4P0
07Bo0?ooo`3ookl0M0000018W03Oool02@3oool00`000000oooo0?ooo`3o0?oooih0oooo00<00000
0?ooo`3oool08@3oool00440oooo00<0oooO09a80018W=l00P3oool00`3o_g@0001809cOo`060?oo
o`030?oog`2LB000M;oo0080oooo00<0omnL04P00018W=l02@3oool00`000000oooo0?ooo`3o0?oo
oih0oooo00<000000?ooo`3oool08@3oool00440oooo00<0oooO09a80018W=l00P3oool0103ooml0
W4P00018W03Oool50?ooo`030?oo_`1d0000M;oo00<0oooo00<0_g@00000M02oool0203oool00`00
0000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00440oooo00<0oooO09a80000
M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB00007Bo00<0oooo00<0gia80000B02L
gol0203oool00`000000oooo0?ooo`0D0?ooo`800000o`3ooon80?ooo`030000003oool0oooo0240
oooo00120?ooo`030;md0000B9`0gooo0080oooo00<0omnL04P00018W=l01P3oool00`2oM000001d
0;ooo`020?ooo`030=nLB00004P0W=oo00P0oooo1000000B0?ooo`@00000o`3ooon40?ooo`@00000
8`3oool00480oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`060?ooo`080?noM000
0000B9cO0?ooo`3ooml0W4P00018W03Oool80?ooo`030000003oool0oooo0180oooo1P00003o0?oo
ohH0oooo00<000000?ooo`3oool08@3oool00480oooo00<0omnL04P00018W=l00`3oool00`2oM000
04RL0=ooo`050?ooo`030?oog`2LB00000000080000000<0001d0;ooo`3oool0203oool00`000000
oooo0?ooo`0B0?ooo`H00000o`3ooon60?ooo`030000003oool0oooo0240oooo00120?ooo`040?oo
g`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool01P3oool00`3ookl0M000001d_`0<0?ooo`03
0000003oool0oooo01<0oooo1000003o0?ooohL0oooo00<000000?ooo`3oool08@3oool004<0oooo
00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`070?ooo`030?oOW0180000B9cO00/0oooo
00<000000?ooo`3oool0503oool200000?l0ooooR03oool00`000000oooo0?ooo`0Q0?ooo`00A03o
ool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`T0oooo00@0oooO09a8000007@0_ooo2@3oool00`00
0000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool004D0oooo00@0gia800000000
000007Bo303oool0103ookl0M00000000018W=l70?ooo`030000003oool0oooo0?l0ooooWP3oool0
0`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?oo
o`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240
oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03o
ool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600oooo00<00000
0?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`800000o`3ooonM0?oo
o`<000008`3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240
oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03o
ool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600oooo00<00000
0?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo
0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0
oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000
003oool0oooo0240oooo001P0?ooo`030000003oool0oooo0100oooo0P00003o0?oooh`0oooo00<0
00000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool03`3oool400000?l0ooooR`3oool0
0`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`0>0?ooo`H00000o`3ooon:0?oo
o`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00h0oooo1P00003o0?ooohX0
oooo00<000000?ooo`3oool08@3oool00600oooo0P00000@0?ooo`@00000o`3ooon90?ooo`<00000
8`3oool00600oooo00<000000?ooo`3oool0403oool200000?l0ooooS03oool00`000000oooo0?oo
o`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool0
0600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`03
0000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo
0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3o
oonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool0
0`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?oo
o`3oool08@3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240
oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03o
ool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600oooo00<00000
0?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`800000o`3ooonM0?oo
o`<000008`3oool00600oooo00<000000?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240
oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03o
ool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool00600oooo00<00000
0?ooo`3oool0o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo
0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0
oooo00<000000?ooo`3oool08@3oool004<0oooo00@0_g@0000000000000M;oo0`3oool00`3ogi`0
B00007Boo`030?ooo`060?oo_`1d0000000000000000B9`0gooo2P3oool00`000000oooo0?ooo`3o
0?oooih0oooo00<000000?ooo`3oool08@3oool00480oooo00H0gia8001d_`3oool0omnL04P00018
W=l20?ooo`030?oOW0180000M;oo0080oooo00<0oono07@00018W=l00P3oool00`3OW4P0001d0;oo
o`090?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00@@3oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0okmd0018W03Oool00`3oool0
0`2oM000B9cO0?ooo`080?ooo`030000003oool0oooo0?l0ooooWP3oool00`000000oooo0?ooo`0Q
0?ooo`00@@3oool00`3ooml0W4P004RLg`020?ooo`040?oog`2LB00004RL0=ooo`D0oooo00<0okmd
0018W03Oool00`3oool00`2oM00007Bo0?ooo`080?ooo`030000003oool0oooo0?l0ooooWP3oool0
0`000000oooo0?ooo`0Q0?ooo`00@@3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo
00D0oooo00<0oooO09a8001d_ol00P3oool00`3ookl0M00004RLg`090?ooo`030000003oool0oooo
0?l0ooooWP3oool00`000000oooo0?ooo`0Q0?ooo`00@P3oool00`2oM00004RL0=ooo`020?ooo`03
0?oOW0180000B9cO00L0oooo00H0_g@00000B02Lgml0W4P00000M02oool90?ooo`@00000o`3ooonJ
0?ooo`@000008`3oool00480oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0001d_`080?oo
o`050?oog`2LB00000000018W03Oool02@3oool00`000000oooo0?ooo`3o0?oooih0oooo00<00000
0?ooo`3oool08@3oool00480oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?oo
o`060;md000004P0W=oo0?oOW018B9`0gooo203oool00`000000oooo0?ooo`3o0?oooih0oooo00<0
00000?ooo`3oool08@3oool00480oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=oo
o`060?ooo`030?oOW0180000M;oo0080oooo00<0okmd04RLg`3oool01`3oool00`000000oooo0?oo
o`3o0?oooih0oooo00<000000?ooo`3oool08@3oool004<0oooo00<0omnL04P0001d_ol00P3oool0
0`3OW4P007Bo0?ooo`060?ooo`030?oOW01804P0W=oo0080oooo00<0okmd0018W03Oool01`3oool0
0`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool004@0oooo00H0gia80000
M02oool0oooO09a80018W=l70?ooo`030?oog`2LB000B9cO0080oooo00<0_g@0001d_`3oool01`3o
ool00`000000oooo0?ooo`3o0?oooih0oooo00<000000?ooo`3oool08@3oool004D0oooo00@0gia8
00000000000007Bo2@3oool01@3ooml0W4P000000000000007Bo00T0oooo00<000000?ooo`3oool0
o`3ooonN0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo0?l0ooooWP3o
ool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`3o0?oooih0oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool02P3oool200000?l0ooooTP3oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`090?ooo`@00000o`3ooonA0?ooo`03
0000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00P0oooo1P00003o0?oooi00oooo
00<000000?ooo`3oool08@3oool00600oooo0P0000090?ooo`H00000o`3ooon>0?ooo`<000008`3o
ool00600oooo00<000000?ooo`3oool02@3oool400000?l0ooooT@3oool00`000000oooo0?ooo`0Q
0?ooo`00H03oool00`000000oooo0?ooo`090?ooo`@00000o`3ooonA0?ooo`030000003oool0oooo
0240oooo001P0?ooo`030000003oool0oooo00P0oooo1P00003o0?oooi00oooo00<000000?ooo`3o
ool08@3oool00600oooo00<000000?ooo`3oool0203oool600000?l0ooooT03oool00`000000oooo
0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`090?ooo`@00000o`3ooonA0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo00T0oooo0`00003o0?oooi80oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0203oool400000?l0ooooTP3oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`070?ooo`H00000o`3ooonA0?ooo`03
0000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00L0oooo1P00003o0?oooi40oooo
00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool0203oool400000?l0ooooTP3o
ool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`090?ooo`800000o`3ooonC
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00P0oooo1000003o0?oo
oi80oooo00<000000?ooo`3oool08@3oool00600oooo0P0000080?ooo`H00000o`3ooon?0?ooo`<0
00008`3oool00600oooo00<000000?ooo`3oool01P3oool700000?l0ooooT@3oool00`000000oooo
0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`060?ooo`H00000o`3ooonB0?ooo`030000003o
ool0oooo0240oooo001P0?ooo`030000003oool0oooo00H0oooo1P00003o0?oooi80oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool01P3oool600000?l0ooooTP3oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`070?ooo`@00000o`3ooonC0?ooo`03
0000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00H0oooo1P00003o0?oooi80oooo
00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool01P3oool600000?l0ooooTP3o
ool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`060?ooo`H00000o`3ooonB
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00H0oooo1P00003o0?oo
oi80oooo00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool01P3oool600000?l0
ooooTP3oool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`060?ooo`H00000
o`3ooonB0?ooo`030000003oool0oooo0240oooo001P0?ooo`8000001P3oool700000?l0ooooT03o
ool3000002<0oooo001P0?ooo`030000003oool0oooo00D0oooo1P00003o0?oooi<0oooo00<00000
0?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool01P3oool400000?l0ooooU03oool00`00
0000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`050?ooo`H00000o`3ooonC0?ooo`03
0000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00D0oooo1P00003o0?oooi<0oooo
00<000000?ooo`3oool08@3oool00600oooo00<000000?ooo`3oool01@3oool600000?l0ooooT`3o
ool00`000000oooo0?ooo`0Q0?ooo`00H03oool00`000000oooo0?ooo`050?ooo`H00000o`3ooonC
0?ooo`030000003oool0oooo0240oooo001P0?ooo`030000003oool0oooo00D0oooo1P00003o0?oo
oi<0oooo00<000000?ooo`3oool08@3oool004l0oooo00<0oono07@0000000000P0000000`18W=l0
oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo1000003o0?oooi@0oooo00<000000?ooo`3o
ool08@3oool00540oooo00<0okmd0000001d_ol0303oool00`000000oooo0?ooo`050?ooo`H00000
o`3ooonC0?ooo`030000003oool0oooo0240oooo001A0?ooo`030?oog`2LB00007Bo00`0oooo00<0
00000?ooo`3oool01@3oool600000?l0ooooT`3oool00`000000oooo0?ooo`0Q0?ooo`00DP3oool0
0`3OW4P004RL0=ooo`0;0?ooo`030000003oool0oooo00H0oooo1000003o0?oooi@0oooo00<00000
0?ooo`3oool08@3oool00580oooo00<0omnL04P0B02Lgol02`3oool00`000000oooo0?ooo`070?oo
o`800000o`3ooonE0?ooo`030000003oool0oooo0240oooo001B0?ooo`030?oo_`1d0000M;oo00/0
oooo100000060?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0A@3oool00`000000
oooo0?ooo`140?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0A03oool00`000000
oooo0?ooo`0[0?ooo`@000008`3oool005<0oooo00<0_g@0001d_`3oool02P3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0Q0?ooo`00D`3oool00`3OW4P004RL0=ooo`0:0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00`0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0240oooo001C
0?ooo`030?oOW01804P0W=oo00X0ooooo`00002R000002<0oooo001C0?ooo`030?oo_`1d0000M;oo
0?l0ooooc`3oool005@0oooo00<0_g@0001d_`3oool0o`3oooo>0?ooo`00DP3oool01@3ookl0M000
0000000007@0_ooo0?l0oooocP3oool00?l0ooooo`3ooolV0?ooo`00o`3ooooo0?ooobH0oooo003o
0?ooool0oooo9P3oool00?l0ooooo`3ooolV0?ooo`00o`3ooooo0?ooobH0oooo003o0?ooool0oooo
9P3oool00?l0ooooo`3ooolV0?ooo`00o`3ooooo0?ooobH0oooo00360?ooo`060=nLB00000000000
0000000004P0W=oo1P3oool00`2oM00007Bo0?ooo`050?ooo`050?noM000000000000000B02Lgol0
0`3oool0103o_g@000000000B02Lgol20?ooo`050?noM000000000000000B02Lgol00`3oool00`3O
W4P004RL0=ooo`020?ooo`030;md0000000007Bo00D0oooo00<0okmd0000000000001`0000001P1d
_ol0oooo0?noM0000000001d0;ooo`80oooo00D0gia8000000000000000007Boo`030?ooo`050=nL
B000000000000000B02Lgol00P3oool01`3o_g@000000000B02Lgol0oooo0;md0000M;l01P3oool0
102oM00000000000B02Lgol20?ooo`050=nLB000000000000000B02Lgol0103oool01P3OW4P00000
000000000000001809cOo`L0oooo00D0okmd000000000000001809cOo`030?ooo`030;md0000M;l0
oooo00@0oooo00`0gia80018W03Oool0oooo0?noM00007@0_ooo0?ooo`3ooml0W4P00000001d_ol5
0?ooo`060=nLB000000000000000000004P0W=oo1P3oool02`3ogi`0B000000000000000M;oo0=nL
B000000004RL0=ooo`3oool0gia80080000000P0001809cOo`3oool0oooo0?noM0000000001809cO
o`80oooo00D0okmd000000000000001809cOo`030?ooo`030;md0000M;l0oooo08l0oooo00370?oo
o`040?oog`2LB00004RL0=ooo`L0oooo00<0okmd0000M02oool0103oool01`3ogi`0B01809cOo`3o
ool0omnL04P00000M;l00P3oool0303o_g@0001809cOg`2LM9`0gooo0?oOW01804P0W=oo0?ooo`3o
gi`0B000001d_`80oooo00<0okmd0000M02oool00P3oool0103OW4P004RL0=noM01d_ol50?ooo`04
0?oog`2LB00004RL0=ooo`<0oooo00`0oooO09a80000B9`0gooo0?oOW01807@0_ono09bLg`3oool0
omnL04P0001d_ol20?ooo`0C0;md0000M;l0oooo0?noM0000000M;oo0?ooo`3ooml0W4P004RLg`3o
ool0okmd0000B02Lgml0W7BL0=ooo`3o_g@0001d0;ooo`050?ooo`0<0=nLB000B9`0gooO09adW03O
ool0okmd0000001d_ol0oooo0?oog`2LB000B9cO103oool0103ooml0W4P00018W03Oool70?ooo`07
0?oOW01804P0W=oo0?ooo`3ogi`0B000001d_`020?ooo`030?noM00007@0_ooo00@0oooo00L0okmd
0000B02Lgol0oooo0?oo_`1d0000M;oo0080oooo00@0_g@0001d_`2oM7@0_ooo1@3oool0103ooml0
W4P00018W03Oool70?ooo`0M0;md0000B9`0gooo0?ooo`2oM000M7Ad0000M02ogi`0W;oo0?noM000
0000M;oo0?ooo`3ooml0W4P004RLg`3oool0okmd0000B02Lgml0W7BL0=ooo`3ogi`0B01809cOo`3o
ool0omnL04P00000M;l00P3oool00`3o_g@0001d0;ooo`2?0?ooo`00b03oool00`3o_g@0001809cO
o`070?ooo`030?oOW0180000M;oo00@0oooo00<0gia80018W03Oool00P3oool01`3ogi`0B00004RL
g`3oool0oono07@0001d_ol00P3oool00`3OW4P004RL0=ooo`020?ooo`070?oOW0180000B9cO0?oo
o`3ookl0M00004RLg`020?ooo`030?oOW01804P0W=oo00L0oooo00<0okmd0000B02Lgol01@3oool0
2P3OW7@0W=oo0?oo_`1d0000M;oo0=nLW02oool0gia80018W03Oool30?ooo`050=nLW02oool0gia8
0018W03Oool00`3oool01@3Ogml0oooo0?oo_`1d0000M;oo0080oooo00<0omnL04P0001d_ol01@3o
ool00`3o_g@0001809cOo`020?ooo`030=nLB000B9`0gooo00<0oooo00<0gmoO0?ooo`3oool00`3o
ool00`3o_g@0001809cOo`070?ooo`030=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo
0?oOW0180000M;oo00@0oooo00L0oono07@00018W=l0oooo0?oog`2LB00007Bo0080oooo00<0gia8
0018W03Oool01`3oool00`3o_g@0001809cOo`070?ooo`030;md0000M;l0oooo0080oooo00L0_g@0
0000001d_ol0oooo0=nLB000B9`0gooo00<0oooo00D0gmoO0?ooo`3ookl0M00007Boo`020?ooo`03
0=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0oooo0?oOW0180000M;oo08l0oooo00380?oo
o`030?oo_`1d0000B9cO00L0oooo00@0oooO09a80000B9`0gooo0`3oool00`3o_g@0001d0;ooo`03
0?ooo`060=nLB00007@0_ooo0?ooo`2oM00007Bo0P3oool00`3o_g@0001d0;ooo`030?ooo`0;0=nL
B00007@0_ooo0?oog`2LB000001d0;ooo`3oool0oooO09a80018W=l01`3oool00`3ookl0M00004RL
g`080?ooo`050;md0018W=l0oooo0?ooo`3o_g@00`0000001`18W=l0oooo0?ooo`3oool0okmd0000
M02oool01P3oool0202oM00007Bo0?ooo`3oool0oooO09a80000B9`0gooo103oool00`3ookl0M000
04RLg`020?ooo`030?noM00007@0_ooo00T0oooo00<0omnL04P0001d_ol01`3oool00`3o_g@0001d
0;ooo`030?ooo`070=nLB00007@0_ooo0?oog`2LB00004RL0=ooo`030?ooo`0<0?oog`2LB00004RL
0=ooo`3oool0gia80000001d_ol0oooo0?oOW01804P0W=oo1`3oool00`3ogi`0B00007Boo`070?oo
o`030?noM00007@0_ooo0080oooo00L0oono07@00018W=l0oooo0?noM00007@0_ooo00H0oooo00L0
_g@0001d_`3oool0oooo0?noM00007@0_ooo00<0oooo00L0gia80000M02oool0oooO09a80000B9`0
gooo08h0oooo00380?ooo`030?oog`2LB00007Bo00P0oooo00<0_g@004QdM01dW=l00`3oool00`3o
okl0M00007Boo`030?ooo`0;0?oOW01804P0W=oo0?ooo`3OW4P004RL0=ooo`3oool0oono07@0001d
_ol00`3oool01`3ogi`0B01809cOo`3oool0gia80018B018M;l00P3oool00`2oM00007Bo0?ooo`06
0?ooo`030?oog`2LB00007Bo0080oooo00<0oono09bLg`3oool00`3oool03`3OW4P004RL0=ooo`3o
ool0oono07@0B02Lgol0oooo0=nLB0000000B9cO0?ooo`3ookl0M00007Boo`060?ooo`030=nLB000
B9`0gooo0080oooo00<0_g@004QdM01dW=l01@3oool01`2oM00004RL0=ooo`3oool0oono07@0001d
_ol02@3oool00`3ooml0W4P0001d_`030?ooo`070;nLW03Oool0oooo0?ooo`3ookl0M00007Boo`03
0?ooo`070?oOW01804P0W=oo0?ooo`2oM000B7Ad07BLg`040?ooo`0;0=nLB000B4P0M9cO0?ooo`3o
_g@0001d07AdM02oool0oooO09a80018W=l01`3oool00`3ooml0W4P0001d_`030?ooo`070;nLW03O
ool0oooo0?ooo`3ookl0M00007Boo`030?ooo`060;md0000B9`0gooo0?oo_`1d0000M;oo1P3oool0
1`3OW4P004RL0=ooo`3oool0oono07@0001d_ol00`3oool01`3ogi`0B01809cOo`3oool0_g@004Qd
M01dW=l0SP3oool00<T0oooo00<0gia80000B018B0000`0000002018W=l0oooo0?ooo`3oool0okmd
0018W02oW4P0M;oo0`3oool00`3OW4P0001d0;ooo`020?ooo`070?oo_`1d0000M;oo0?ooo`3ogi`0
B01809cOo`020?ooo`030=nLB00007@0_ooo0080oooo00`0oono07@0001d_ol0oooo0?noM00007@0
_kmd04RLg`3oool0okmd0000M02oool70?ooo`060=nLB00007@0_ooo0?ooo`3ogi`0B7Bo103oool0
0`3ogi`0B01d0;ooo`020?ooo`0C0?noM000B9`0gooo0?ooo`3ookl0M00004RLg`3oool0gia80000
M02oool0oooo0?oo_`1d0000M;oo0?ooo`3ogi`0B01809cOo`020?ooo`040?noM000B9`0_ia807Bo
o`@0oooo00<0gia80000B02Lgol00P3oool01`3OW4P0001d0;ooo`3oool0oono07@0001d_ol01@3o
ool01`2oM00004RL0=ooo`3oool0oooO09a8B02Lgol00`3oool00`3OW4P0001d0;ooo`020?ooo`08
0?oo_`1d0000M;oo0?ooo`3o_g@004RL0;nLB01d_ol30?ooo`0<0?noM00007@0W9ad09cOo`3ookl0
M00007Bo_`1dM;l0oooo0;md0000B9`0gooo1`3oool01`2oM00004RL0=ooo`3oool0oooO09a8B02L
gol00`3oool00`3o_g@0001d0;ooo`020?ooo`0?0=nLB00007@0_ooo0?ooo`3OW4P0001d0;ooo`3o
ool0oono07@0001d_ol0oooo0?oOW01804P0W=oo0080oooo00<0gia80000M02oool00P3oool0203o
okl0M00007Boo`3oool0okmd0018W02oW4P0M;ooS@3oool00<T0oooo00<0okmd0000B02Lgol00P3o
ool0603ogi`0B0000000M02oool0oooo0?oo_`1d04P0W=oo0;md001dM7@004RL0=ooo`3oool0_g@0
0000M02oool0oooo0?noM00007@0_ooo0?ooo`3ooml0W4P004RLg`<0oooo0100_g@00000M02oool0
oooo0?noM00007@0_ooo0?ooo`3ooml0W4P007Boo`3OW4P007Bo0?oOW01804P0W=oo1`3oool00`3o
_g@0000000000002000000030000M02oool0oooo0080oooo00<0oono07@0001d_ol00`3oool04P3O
W4P0001d0;ooo`3oool0gia80018W03Oool0oooo0;md000007@0_ooo0?ooo`3OW4P0B9cO0?ooo`3o
oml0W4P004RLg`80oooo00/0oono07@0B02Lgol0_g@007AdM000B9`0gooo0?ooo`3ogi`0B00004RL
g`030?ooo`060;md000007@0_ooo0?ooo`3OW4P0B9cO1@3oool00`3o_g@000000000000300000003
07Boo`3oool0oooo0080oooo01/0_g@00000M02oool0oooo0?noM00007@0_ooo0?ooo`3ookl0M018
09cOo`2oM000M7Ad0018W03Oool0oono07@0B02Lgml0W4Q809cOg`2LB000B9cO0?noM000M;l0okmd
0000M02oool01`3oool00`3o_g@00000000000030000000307Boo`3oool0oooo0080oooo00L0gia8
0000M02oool0oooo0;md000004P0W=oo0080oooo00X0_g@00000M02oool0oooo0=nLB018W=l0oooo
0?oog`2LB000B9cO0`3oool03`2oM000001d0;ooo`3oool0okmd0000M02oool0oooo0?oo_`1d04P0
W=oo0;md001dM7@004RL0=ooo`2:0?ooo`00b@3oool00`3ookl0M00004RLg`030?ooo`0;0?oOW018
000007Bo0?oOW0180000000007Boo`3oool0_g@00000M02oool00P3oool0203o_g@0000000000000
0000M;oo0?ooo`3ookl0M0000`000000101d_ol0oooo0?ooo`3o_g@30000000;07Boo`3oool0oono
07@000000000B9cO0?ooo`3o_g@000000000B02Lgol01`3oool00`3ookl0M00007Boo`020?ooo`03
0;mdB02Lgol0oooo00<0oooo00<0_g@004RLg`3oool00`3oool01@3ogi`0B000000000000000M;oo
00<0oooo00L0omnL04P000000000001809cOo`3ookl0M00000<0000000T0M9bL04P000000000M;oo
0?ooo`2oM000001d0;ooo`3o_g@00P0000000`00B9`0gooo0?ooo`020?ooo`050?oOW01800000000
0000B02Lgol01@3oool00`3ookl0M00007Boo`020?ooo`030=nLB018W=l0oooo00@0oooo01X0okmd
000000000000000007Boo`3oool0omnL04P000000000M;oo0?ooo`2oM000001d0;oo_`1d00000000
04RLg`3ooml0W4P000000018W=l0oooo0?noM000000004RL0=ooo`L0oooo00<0oono07@0001d_ol0
0P3oool00`3OW4P0B9cO0?ooo`040?ooo`060?oOW0180000000004RLW01d0000B9cO0`3oool01`3o
gi`0B0000000000004P0W=oo0?oo_`1d00000`000000101d_ol0oooo0?ooo`3o_g@30000000:07Bo
o`3oool0omnL04P000000000M;oo0?ooo`2oM000001d0;ooohX0oooo003:0?ooo`030;md0000B9`0
gooo00<0oooo00<0_g@00000M02oool04@3oool00`3o_g@0001d0;ooo`0J0?ooo`060;md0000M;l0
oooo0?ooo`3ogi`0M9cO103oool00`3OW4P004RL0=ooo`0B0?ooo`030?noM00007@0_ooo01X0oooo
00<0oooO09a80000M;l00P3oool00`3ooml0W7BL0=ooo`0U0?ooo`030?oog`2LB00007Bo0080oooo
00<0oooO09adW03Oool0503oool00`3o_g@0001d0;ooo`2K0?ooo`00bP3oool00`3OW4P0001d0;oo
o`030?ooo`030;md000007@0_ooo0180oooo00<0giad09cOo`3oool06@3oool00`3OW4P0001d0;oo
o`040?ooo`060=nLW02oool0oooo0?oOW01807@0_ooo4`3oool00`3OW7@0W=oo0?ooo`0J0?ooo`03
0=nLB000B9`0gooo00@0oooo00<0omnL09boo`3oool08`3oool00`3OW4P004RL0=ooo`040?ooo`03
0?oOW02L_ol0oooo0180oooo00<0giad09cOo`3oool0VP3oool00<X0oooo00<0okmd0000001d_ol0
0P3oool00`3o_g@00000001d_`0_0?ooo`030?noM00004P0W=oo00<0oooo00L0oono07@0B02Lgol0
oooo0?oog`2LB000M;oo01l0oooo00@0oono07@00000B9`0gooo3@3oool00`3o_g@0001809cOo`04
0?ooo`030=nLB000M;l0oooo02<0oooo00<0okmd0000B02Lgol0103oool00`3OW4P007Bo0?ooo`2_
0?ooo`00b@3oool00`2oM00000000000000400000003001d_`3oool0oooo02d0oooo00<0_g@00000
000000001P0000001@1d_ol0oooo0?ooo`3OW4P0B9cO02l0oooo00<0_g@00000000000001P000000
0`0007@0_ooo0?ooo`0Q0?ooo`030;md00000000000000H0000000<0001d0;ooo`3oool0[P3oool0
0?l0oooo2`3oool0103ooml0W4P000000000M;oo0?oooaL0oooo003o0?ooool0oooo9P3oool00?l0
ooooo`3ooolV0?ooo`00o`3ooooo0?ooobH0oooo003o0?ooool0oooo9P3oool00001\
\>"],
  ImageRangeCache->{{{0, 547.625}, {338.063, 0}} -> {-1.4868, -0.2356, \
0.0209774, 0.00605305}}]
}, Closed]],

Cell[TextData[{
  "More detailed instructions on reading files with more general formats can \
be found in ",
  ButtonBox["files.nb",
    ButtonData:>{
      URL[ 
      "www.physics.umd.edu/courses/CourseWare/EssentialMathematica/Additional \
topics/files.nb"], None},
    ButtonStyle->"Hyperlink"],
  " or in the ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " reference guide."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Fitting data", "Section"],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides several functions for fitting data, but here we discuss only ",
  ButtonBox["NonlinearRegress",
    ButtonStyle->"RefGuideLink"],
  " because it is the most general and subsumes the other functions as \
special cases.  The basic syntax and list of options are given in the \
reference guide, obtained by following the highlighted link.  For use in this \
function, we provide data in the form ",
  Cell[BoxData[
      \(TraditionalForm\`{{x\_1, y\_1}, \[CenterEllipsis], {x\_N, y\_N}}\)]],
  " and a separate list of weights ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Sigma]\_1\^\(-2\), \[CenterEllipsis], \[Sigma]\_N\
\^\(-2\)}\)]],
  ".  The fitting function is given explicitly, followed by a list of \
independent variables and a list of parameters.  The output is returned as a \
list of replacement rules.  Note that when the fitting function is a \
nonlinear function of the fitting parameters, it is often important to begin \
the search with reasonable initial guesses for those parameters.  Initial \
values are provided by giving the list of parameters in the form ",
  Cell[BoxData[
      \(TraditionalForm\`{{a, a\_0}, {b, b\_0}, \[CenterEllipsis]}\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`a, b, \[CenterEllipsis]\)]],
  " are the parameters and ",
  Cell[BoxData[
      \(TraditionalForm\`a\_0, b\_0, \[CenterEllipsis]\)]],
  " are their initial values."
}], "Text"],

Cell[TextData[{
  "We illustrate this function by fitting the data for ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep[Q\^2]\)]],
  " read in the previous section.  A simple parametrization is the rational \
function"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`G\_Ep[
        Q\^2]\  = \ \(\[Sum]\+\(\[Mu] = 0\)\%m\( a\_\[Mu]\) Q\^\(2  \[Mu]\)\)\
\/\(\[Sum]\+\(\[Nu] = 0\)\%\(m + 2\)\(b\_\[Nu]\) Q\^\(2  \[Nu]\)\)\)], \
"DisplayFormula"],

Cell[TextData[{
  "where the symmetry of the problem limits the expansions to even powers.  A \
rational function of this kind is sometimes called a ",
  StyleBox["Pad\[EAcute]",
    FontSlant->"Italic"],
  " approximation.  Recognizing that ",
  Cell[BoxData[
      \(TraditionalForm\`G\_Ep[0] = 1\)]],
  " because the charge must be one elementary unit, we can constrain ",
  Cell[BoxData[
      \(TraditionalForm\`a\_0 = \(b\_0 = 1\)\)]],
  ", but it may be useful to vary ",
  Cell[BoxData[
      \(TraditionalForm\`a\_0\)]],
  " if the systematic normalization uncertainties are nonnegligible.  \
Finally, there are theoretical reasons based upon the quark model to require"
}], "Text"],

Cell[BoxData[
    \(TraditionalForm\`Q\^2 \[Rule] \[Infinity]\ \ \[DoubleLongRightArrow]\ \ \
G\_Ep \[Proportional] Q\^\(-4\)\)], "DisplayFormula"],

Cell["\<\
This is the reason that we include two more terms in the denominator than in \
the numerator. \
\>", "Text"],

Cell[TextData[{
  "The cells below perform such a fit.  First we define the fitting function \
and then we define a list of parameters and their initial values; note that \
the initial guesses do not have to be especially precise for this problem.  \
The uncertainties are then extracted from the input data and listability is \
used to specify the weights ",
  Cell[BoxData[
      \(TraditionalForm\`w\_i = 1/\[Sigma]\_i\^2\)]],
  " succinctly.  The output of ",
  StyleBox["NonlinearRegress", "InlineInput"],
  " is a fairly detailed report containing the fitted parameters, their \
standard errors (SE), and other statistics in the form of replacement rules."
}], "Text"],

Cell[BoxData[{
    \(\(pade[
          Q2_] = \(1 + a[1]\ Q2\)\/\(1 + b[1]\ Q2\  + \ b[2]\ Q2\^2 + b[3]\ \
Q2\^3\);\)\), "\[IndentingNewLine]", 
    \(\(start = {{a[1], 0}, {b[1], 2}, {b[2], 0}, {b[3], 0}};\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\({x, y, dy} = Transpose[gepdata];\)\), "\[IndentingNewLine]", 
    \(gepresults = 
      NonlinearRegress[Transpose[{x, y}], pade[Q2], Q2, start, 
        Weights \[Rule] dy\^\(-2\)]\)}], "Input"],

Cell[BoxData[
    RowBox[{"{", 
      RowBox[{\(BestFitParameters \[Rule] {a[
                1] \[Rule] \(-0.06703155763057679`\), 
            b[1] \[Rule] 3.1176732450061886`, 
            b[2] \[Rule] 1.034075071571788`, 
            b[3] \[Rule] 0.5047708157066234`}\), ",", 
        RowBox[{"ParameterCITable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Estimate\"\>", "\<\"Asymptotic SE\"\>", "\<\
\"CI\"\>"},
                {\(a[1]\), \(-0.06703155763057679`\), 
                  "0.032680190394446064`", \({\(-0.1327394843234275`\), \
\(-0.0013236309377260835`\)}\)},
                {\(b[1]\), "3.1176732450061886`", 
                  "0.0493449914086023`", \({3.0184584901653837`, 
                    3.2168879998469935`}\)},
                {\(b[2]\), "1.034075071571788`", 
                  "0.0834244486518478`", \({0.866338975476745`, 
                    1.2018111676668308`}\)},
                {\(b[3]\), "0.5047708157066234`", 
                  "0.14493162086350037`", \({0.21336626131965275`, 
                    0.7961753700935941`}\)}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ BoxForm`e$, TableDepth -> 2, TableHeadings -> {{
                a[ 1], 
                b[ 1], 
                b[ 2], 
                b[ 3]}, {"Estimate", "Asymptotic SE", "CI"}}]]]}], 
        ",", \(EstimatedVariance \[Rule] 0.8454699647048624`\), ",", 
        RowBox[{"ANOVATable", "\[Rule]", 
          InterpretationBox[GridBox[{
                {"\<\"\"\>", "\<\"DF\"\>", "\<\"SumOfSq\"\>", \
"\<\"MeanSq\"\>"},
                {"\<\"Model\"\>", "4", "2.7544198845825577`*^6", 
                  "688604.9711456394`"},
                {"\<\"Error\"\>", "48", "40.582558305833395`", 
                  "0.8454699647048624`"},
                {"\<\"Uncorrected Total\"\>", "52", 
                  "2.7544604671408636`*^6", "\<\"\"\>"},
                {"\<\"Corrected Total\"\>", "51", 
                  "1.6582070697390493`*^6", "\<\"\"\>"}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            
            TableForm[ {{4, 2754419.8845825577, 688604.97114563943}, {48, 
              40.582558305833395, .84546996470486235}, {52, 
              2754460.4671408636}, {51, 1658207.0697390493}}, 
              TableHeadings -> {{"Model", "Error", "Uncorrected Total", 
              "Corrected Total"}, {"DF", "SumOfSq", "MeanSq"}}]]}], ",", 
        RowBox[{"AsymptoticCorrelationMatrix", "\[Rule]", 
          TagBox[
            RowBox[{"(", "\[NoBreak]", GridBox[{
                  {"1.`", "0.8543384249285648`", "0.21199850147095156`", 
                    "0.9550277111745461`"},
                  {"0.8543384249285647`", 
                    "0.9999999999999999`", \(-0.1705854400846496`\), 
                    "0.8900235620669643`"},
                  {"0.21199850147095156`", \(-0.17058544008464957`\), 
                    "1.0000000000000002`", \(-0.05211400125457213`\)},
                  {"0.955027711174546`", 
                    "0.8900235620669642`", \(-0.05211400125457211`\), 
                    "0.9999999999999999`"}
                  },
                RowSpacings->1,
                ColumnSpacings->1,
                ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
            Function[ BoxForm`e$, 
              MatrixForm[ BoxForm`e$]]]}], ",", 
        RowBox[{"FitCurvatureTable", "\[Rule]", 
          TagBox[GridBox[{
                {"\<\"\"\>", "\<\"Curvature\"\>"},
                {"\<\"Max Intrinsic\"\>", "91.71315925310532`"},
                {"\<\"Max Parameter-Effects\"\>", "240.27747858380698`"},
                {"\<\"95. % Confidence Region\"\>", "0.6243612699707356`"}
                },
              RowSpacings->1,
              ColumnSpacings->3,
              RowAlignments->Baseline,
              ColumnAlignments->{Left}],
            Function[ BoxForm`e$, 
              TableForm[ 
              BoxForm`e$, TableDepth -> 2, TableHeadings -> {{"Max Intrinsic",
                 "Max Parameter-Effects", "95. % Confidence Region"}, {
                "Curvature"}}]]]}]}], "}"}]], "Output"]
}, Closed]],

Cell["\<\
Next we compare the fit with the data using the replacements rules to \
substitute the best-fit parameters into the fitting function.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(DisplayTogether[\[IndentingNewLine]ErrorListPlot[gepdata, 
          Frame \[Rule] True, 
          FrameLabel \[Rule] {\*"\"\<\!\(Q\^2\) [(GeV/c\!\(\()\^2\)\)]\>\"", \
\ \*"\"\<\!\(G\_Ep\)\>\""}, 
          PlotLabel \[Rule] "\<Proton Electric Form Factor\>"], \
\[IndentingNewLine]Plot[
          Evaluate[pade[Q2] /. \((BestFitParameters /. gepresults)\)], {Q2, 
            0, 6}]\[IndentingNewLine]];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Times
%%IncludeFont: Times
/Times findfont 12  scalefont  setfont
% Scaling calculations
0.0238095 0.15873 0.0147151 0.588604 [
[.02381 -0.0125 -5 -15.6875 ]
[.02381 -0.0125 5 0 ]
[.18254 -0.0125 -5 -15.6875 ]
[.18254 -0.0125 5 0 ]
[.34127 -0.0125 -5 -15.6875 ]
[.34127 -0.0125 5 0 ]
[.5 -0.0125 -5 -15.6875 ]
[.5 -0.0125 5 0 ]
[.65873 -0.0125 -5 -15.6875 ]
[.65873 -0.0125 5 0 ]
[.81746 -0.0125 -5 -15.6875 ]
[.81746 -0.0125 5 0 ]
[.97619 -0.0125 -5.125 -15.6875 ]
[.97619 -0.0125 5.125 0 ]
[ 0 0 -0.125 0 ]
[.5 -0.0125 -36.1563 -31.375 ]
[.5 -0.0125 36.1563 -15.6875 ]
[-0.0125 .01472 -10 -7.84375 ]
[-0.0125 .01472 0 7.84375 ]
[-0.0125 .13244 -19 -7.84375 ]
[-0.0125 .13244 0 7.84375 ]
[-0.0125 .25016 -19 -7.84375 ]
[-0.0125 .25016 0 7.84375 ]
[-0.0125 .36788 -19.25 -7.84375 ]
[-0.0125 .36788 0 7.84375 ]
[-0.0125 .4856 -19 -7.84375 ]
[-0.0125 .4856 0 7.84375 ]
[-0.0125 .60332 -10 -7.84375 ]
[-0.0125 .60332 0 7.84375 ]
[ 0 0 -0.125 0 ]
[-0.01875 .30902 -35.6875 -11.3125 ]
[-0.01875 .30902 -19.25 11.3125 ]
[ 0 .61803 .125 0 ]
[ 1 0 .125 0 ]
[.5 .63053 -71.7188 0 ]
[.5 .63053 71.7188 15.6875 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 0 m
.02381 .00625 L
s
gsave
.02381 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
.18254 0 m
.18254 .00625 L
s
gsave
.18254 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.34127 0 m
.34127 .00625 L
s
gsave
.34127 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
1.000 setlinewidth
grestore
.5 0 m
.5 .00625 L
s
gsave
.5 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(3) show
1.000 setlinewidth
grestore
.65873 0 m
.65873 .00625 L
s
gsave
.65873 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(4) show
1.000 setlinewidth
grestore
.81746 0 m
.81746 .00625 L
s
gsave
.81746 -0.0125 -66 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(5) show
1.000 setlinewidth
grestore
.97619 0 m
.97619 .00625 L
s
gsave
.97619 -0.0125 -66.125 -19.6875 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(6) show
1.000 setlinewidth
grestore
.125 Mabswid
.05556 0 m
.05556 .00375 L
s
.0873 0 m
.0873 .00375 L
s
.11905 0 m
.11905 .00375 L
s
.15079 0 m
.15079 .00375 L
s
.21429 0 m
.21429 .00375 L
s
.24603 0 m
.24603 .00375 L
s
.27778 0 m
.27778 .00375 L
s
.30952 0 m
.30952 .00375 L
s
.37302 0 m
.37302 .00375 L
s
.40476 0 m
.40476 .00375 L
s
.43651 0 m
.43651 .00375 L
s
.46825 0 m
.46825 .00375 L
s
.53175 0 m
.53175 .00375 L
s
.56349 0 m
.56349 .00375 L
s
.59524 0 m
.59524 .00375 L
s
.62698 0 m
.62698 .00375 L
s
.69048 0 m
.69048 .00375 L
s
.72222 0 m
.72222 .00375 L
s
.75397 0 m
.75397 .00375 L
s
.78571 0 m
.78571 .00375 L
s
.84921 0 m
.84921 .00375 L
s
.88095 0 m
.88095 .00375 L
s
.9127 0 m
.9127 .00375 L
s
.94444 0 m
.94444 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
gsave
.5 -0.0125 -97.1563 -35.375 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Q) show
71.625 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
79.438 15.813 moveto
(@) show
83.375 15.813 moveto
(H) show
87.188 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(GeV) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
109.688 15.813 moveto
(\\220) show
113.500 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(c) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
118.750 15.813 moveto
(L) show
122.563 11.000 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(2) show
%%IncludeResource: font Mathematica2
%%IncludeFont: Mathematica2
/Mathematica2 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
127.375 15.813 moveto
(D) show
131.313 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 .01472 m
.00625 .01472 L
s
gsave
-0.0125 .01472 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0) show
1.000 setlinewidth
grestore
0 .13244 m
.00625 .13244 L
s
gsave
-0.0125 .13244 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.2) show
1.000 setlinewidth
grestore
0 .25016 m
.00625 .25016 L
s
gsave
-0.0125 .25016 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.4) show
1.000 setlinewidth
grestore
0 .36788 m
.00625 .36788 L
s
gsave
-0.0125 .36788 -80.25 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.6) show
1.000 setlinewidth
grestore
0 .4856 m
.00625 .4856 L
s
gsave
-0.0125 .4856 -80 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(0.8) show
1.000 setlinewidth
grestore
0 .60332 m
.00625 .60332 L
s
gsave
-0.0125 .60332 -71 -11.8438 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(1) show
1.000 setlinewidth
grestore
.125 Mabswid
0 .04415 m
.00375 .04415 L
s
0 .07358 m
.00375 .07358 L
s
0 .10301 m
.00375 .10301 L
s
0 .16187 m
.00375 .16187 L
s
0 .1913 m
.00375 .1913 L
s
0 .22073 m
.00375 .22073 L
s
0 .27959 m
.00375 .27959 L
s
0 .30902 m
.00375 .30902 L
s
0 .33845 m
.00375 .33845 L
s
0 .39731 m
.00375 .39731 L
s
0 .42674 m
.00375 .42674 L
s
0 .45617 m
.00375 .45617 L
s
0 .51503 m
.00375 .51503 L
s
0 .54446 m
.00375 .54446 L
s
0 .57389 m
.00375 .57389 L
s
.25 Mabswid
0 0 m
0 .61803 L
s
gsave
-0.01875 .30902 -99.7813 -12.2188 Mabsadd m
1 1 Mabs scale
currentpoint translate
72.3125 12.2188 translate
90 rotate
-72.3125 -12.2188 translate
0 24.4375 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding WindowsANSIEncoding def
        currentdict
    end

    newfontname exch definefont pop
} def

63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
%%BeginResource: font Times-Italic-MISO
%%BeginFont: Times-Italic-MISO
/Times-Italic /Times-Italic-MISO MISOfy
%%EndFont
%%EndResource
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(G) show
71.688 17.688 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 8.500 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Ep) show
81.625 15.813 moveto
%%IncludeResource: font Times-Italic-MISO
%%IncludeFont: Times-Italic-MISO
/Times-Italic-MISO findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
.02381 .61178 m
.02381 .61803 L
s
.18254 .61178 m
.18254 .61803 L
s
.34127 .61178 m
.34127 .61803 L
s
.5 .61178 m
.5 .61803 L
s
.65873 .61178 m
.65873 .61803 L
s
.81746 .61178 m
.81746 .61803 L
s
.97619 .61178 m
.97619 .61803 L
s
.125 Mabswid
.05556 .61428 m
.05556 .61803 L
s
.0873 .61428 m
.0873 .61803 L
s
.11905 .61428 m
.11905 .61803 L
s
.15079 .61428 m
.15079 .61803 L
s
.21429 .61428 m
.21429 .61803 L
s
.24603 .61428 m
.24603 .61803 L
s
.27778 .61428 m
.27778 .61803 L
s
.30952 .61428 m
.30952 .61803 L
s
.37302 .61428 m
.37302 .61803 L
s
.40476 .61428 m
.40476 .61803 L
s
.43651 .61428 m
.43651 .61803 L
s
.46825 .61428 m
.46825 .61803 L
s
.53175 .61428 m
.53175 .61803 L
s
.56349 .61428 m
.56349 .61803 L
s
.59524 .61428 m
.59524 .61803 L
s
.62698 .61428 m
.62698 .61803 L
s
.69048 .61428 m
.69048 .61803 L
s
.72222 .61428 m
.72222 .61803 L
s
.75397 .61428 m
.75397 .61803 L
s
.78571 .61428 m
.78571 .61803 L
s
.84921 .61428 m
.84921 .61803 L
s
.88095 .61428 m
.88095 .61803 L
s
.9127 .61428 m
.9127 .61803 L
s
.94444 .61428 m
.94444 .61803 L
s
.25 Mabswid
0 .61803 m
1 .61803 L
s
.99375 .01472 m
1 .01472 L
s
.99375 .13244 m
1 .13244 L
s
.99375 .25016 m
1 .25016 L
s
.99375 .36788 m
1 .36788 L
s
.99375 .4856 m
1 .4856 L
s
.99375 .60332 m
1 .60332 L
s
.125 Mabswid
.99625 .04415 m
1 .04415 L
s
.99625 .07358 m
1 .07358 L
s
.99625 .10301 m
1 .10301 L
s
.99625 .16187 m
1 .16187 L
s
.99625 .1913 m
1 .1913 L
s
.99625 .22073 m
1 .22073 L
s
.99625 .27959 m
1 .27959 L
s
.99625 .30902 m
1 .30902 L
s
.99625 .33845 m
1 .33845 L
s
.99625 .39731 m
1 .39731 L
s
.99625 .42674 m
1 .42674 L
s
.99625 .45617 m
1 .45617 L
s
.99625 .51503 m
1 .51503 L
s
.99625 .54446 m
1 .54446 L
s
.99625 .57389 m
1 .57389 L
s
.25 Mabswid
1 0 m
1 .61803 L
s
gsave
.5 .63053 -132.719 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 23.6875 translate 1 -1 scale
/g { setgray} bind def
/k { setcmykcolor} bind def
/p { gsave} bind def
/r { setrgbcolor} bind def
/w { setlinewidth} bind def
/C { curveto} bind def
/F { fill} bind def
/L { lineto} bind def
/rL { rlineto} bind def
/P { grestore} bind def
/s { stroke} bind def
/S { show} bind def
/N {currentpoint 3 -1 roll show moveto} bind def
/Msf { findfont exch scalefont [1 0 0 -1 0 0 ] makefont setfont} bind def
/m { moveto} bind def
/Mr { rmoveto} bind def
/Mx {currentpoint exch pop moveto} bind def
/My {currentpoint pop exch moveto} bind def
/X {0 rmoveto} bind def
/Y {0 exch rmoveto} bind def
63.000 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
63.063 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
(Proton) show
99.438 15.813 moveto
(Electric) show
140.188 15.813 moveto
(Form) show
169.875 15.813 moveto
(Factor) show
202.438 15.813 moveto
%%IncludeResource: font Times-Italic
%%IncludeFont: Times-Italic
/Times-Italic findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 0.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.002 w
.02467 .59368 m
.02467 .59624 L
s
.015 w
.02467 .59496 Mdot
.002 w
.02505 .58817 m
.02505 .59104 L
s
.015 w
.02505 .5896 Mdot
.002 w
.02535 .5857 m
.02535 .58822 L
s
.015 w
.02535 .58696 Mdot
.002 w
.02567 .57846 m
.02567 .58333 L
s
.015 w
.02567 .58089 Mdot
.002 w
.02597 .57924 m
.02597 .58172 L
s
.015 w
.02597 .58048 Mdot
.002 w
.02629 .57173 m
.02629 .57734 L
s
.015 w
.02629 .57454 Mdot
.002 w
.02659 .57082 m
.02659 .5726 L
s
.015 w
.02659 .57171 Mdot
.002 w
.0269 .56793 m
.0269 .57137 L
s
.015 w
.0269 .56965 Mdot
.002 w
.02721 .56435 m
.02721 .56612 L
s
.015 w
.02721 .56524 Mdot
.002 w
.02752 .56165 m
.02752 .56494 L
s
.015 w
.02752 .56329 Mdot
.002 w
.02783 .55796 m
.02783 .55992 L
s
.015 w
.02783 .55894 Mdot
.002 w
.02814 .55228 m
.02814 .55606 L
s
.015 w
.02814 .55417 Mdot
.002 w
.02876 .54678 m
.02876 .54967 L
s
.015 w
.02876 .54823 Mdot
.002 w
.02906 .54529 m
.02906 .54869 L
s
.015 w
.02906 .54699 Mdot
.002 w
.02937 .54385 m
.02937 .54778 L
s
.015 w
.02937 .54581 Mdot
.002 w
.02998 .53082 m
.02998 .53914 L
s
.015 w
.02998 .53498 Mdot
.002 w
.03122 .52004 m
.03122 .52614 L
s
.015 w
.03122 .52309 Mdot
.002 w
.03246 .51103 m
.03246 .51502 L
s
.015 w
.03246 .51303 Mdot
.002 w
.04444 .41854 m
.04444 .42787 L
s
.015 w
.04444 .42321 Mdot
.002 w
.05397 .3685 m
.05397 .37432 L
s
.015 w
.05397 .37141 Mdot
.002 w
.06667 .30669 m
.06667 .31488 L
s
.015 w
.06667 .31078 Mdot
.002 w
.07619 .27492 m
.07619 .28072 L
s
.015 w
.07619 .27782 Mdot
.002 w
.08571 .25548 m
.08571 .26131 L
s
.015 w
.08571 .2584 Mdot
.002 w
.09524 .22987 m
.09524 .24102 L
s
.015 w
.09524 .23544 Mdot
.002 w
.10794 .20054 m
.10794 .21384 L
s
.015 w
.10794 .20719 Mdot
.002 w
.11587 .18526 m
.11587 .19027 L
s
.015 w
.11587 .18776 Mdot
.002 w
.12698 .16393 m
.12698 .17746 L
s
.015 w
.12698 .1707 Mdot
.002 w
.1381 .16351 m
.1381 .18376 L
s
.015 w
.1381 .17364 Mdot
.002 w
.14762 .13865 m
.14762 .14624 L
s
.015 w
.14762 .14244 Mdot
.002 w
.17302 .12557 m
.17302 .13459 L
s
.015 w
.17302 .13008 Mdot
.002 w
.19841 .09478 m
.19841 .10063 L
s
.015 w
.19841 .09771 Mdot
.002 w
.2381 .07872 m
.2381 .08491 L
s
.015 w
.2381 .08182 Mdot
.002 w
.30159 .05293 m
.30159 .06044 L
s
.015 w
.30159 .05668 Mdot
.002 w
.08302 .25663 m
.08302 .27076 L
s
.015 w
.08302 .26369 Mdot
.002 w
.08746 .2478 m
.08746 .26193 L
s
.015 w
.08746 .25487 Mdot
.002 w
.09381 .21543 m
.09381 .22956 L
s
.015 w
.09381 .22249 Mdot
.002 w
.08413 .23603 m
.08413 .26075 L
s
.015 w
.08413 .24839 Mdot
.002 w
.10317 .21072 m
.10317 .23073 L
s
.015 w
.10317 .22073 Mdot
.002 w
.10159 .21243 m
.10159 .21902 L
s
.015 w
.10159 .21572 Mdot
.002 w
.14921 .14168 m
.14921 .14497 L
s
.015 w
.14921 .14333 Mdot
.002 w
.21111 .09076 m
.21111 .093 L
s
.015 w
.21111 .09188 Mdot
.002 w
.25873 .06542 m
.25873 .06918 L
s
.015 w
.25873 .0673 Mdot
.002 w
.30476 .0543 m
.30476 .05677 L
s
.015 w
.30476 .05553 Mdot
.002 w
.32222 .05051 m
.32222 .05275 L
s
.015 w
.32222 .05163 Mdot
.002 w
.3619 .043 m
.3619 .04559 L
s
.015 w
.3619 .0443 Mdot
.002 w
.41587 .03622 m
.41587 .03763 L
s
.015 w
.41587 .03692 Mdot
.002 w
.49524 .02845 m
.49524 .02987 L
s
.015 w
.49524 .02916 Mdot
.002 w
.5746 .0249 m
.5746 .02643 L
s
.015 w
.5746 .02567 Mdot
.002 w
.57937 .02355 m
.57937 .02622 L
s
.015 w
.57937 .02489 Mdot
.002 w
.65397 .02086 m
.65397 .02237 L
s
.015 w
.65397 .02162 Mdot
.002 w
.77778 .01813 m
.77778 .01925 L
s
.015 w
.77778 .01869 Mdot
.002 w
.90317 .01615 m
.90317 .01758 L
s
.015 w
.90317 .01686 Mdot
.5 Mabswid
.02381 .60332 m
.03279 .51177 L
.04262 .4364 L
.05288 .37609 L
.06244 .33156 L
.0828 .26131 L
.09421 .23186 L
.10458 .20937 L
.12534 .17344 L
.14451 .14801 L
.18292 .1115 L
.20165 .09846 L
.22227 .0866 L
.26256 .06902 L
.30133 .05693 L
.34256 .04754 L
.38227 .04083 L
.42046 .03591 L
.46111 .03186 L
.50023 .02883 L
.54182 .02629 L
.58188 .02435 L
.62043 .02285 L
.66144 .02155 L
.70093 .02053 L
.7389 .01971 L
.77932 .01899 L
.81823 .01841 L
.85959 .01789 L
.89944 .01748 L
.93777 .01713 L
.97619 .01684 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006`00012R000`400?l00000o`00003oo`3ooona
0?ooo`00o`3ooona0?ooo`00`@3oool01@3ogkl0_moo0?ooo`3oool0gia800@0000000<0001d0;oo
o`3oool03P3oool01@3OW4P00000000000000000M;oo00<0oooo00<0oooO09a8B02Lgol0803oool0
0`3ooml0W4Q809cOo`0<0?ooo`030?oOW018B9`0gooo00X0oooo00D0gia8000000000000000007Bo
o`290?ooo`00`@3oool00`3ooml0W4P000000002000000070000M02oool0oooo0?ooo`3ookl0M01d
0;ooo`0>0?ooo`030=nLB00007@0_ooo00D0oooo00<0_gA809cOo`3oool08@3oool00`3OW4P0B9cO
0?ooo`0<0?ooo`030?oOW018M;l0oooo00/0oooo00<0oono07@0001d_ol0R@3oool00<<0oooo00<0
_g@004RLg`3oool01@3oool00`3ookl0M7Bo0?ooo`0=0?ooo`030=nLB000B9`0gooo00@0oooo00<0
gia804RLg`3oool08P3oool00`3ogi`0B7Bo0?ooo`0=0?ooo`030=nLB018W=l0oooo00X0oooo00<0
oono07@0001d_ol0R@3oool00<@0oooo00<0gia8001d_`3oool0503oool00`3OW4P004RL0=ooo`03
0?ooo`030?oo_`1d07@0_ooo02<0oooo00<0oono07@0M02oool03@3oool00`3ookl0M01809cOo`0:
0?ooo`030?oo_`1d0000M;oo08T0oooo00330?ooo`030?oOW018000000000080000000<004RL0=oo
o`3oool0403oool00`3OW4P004RL0=ooo`030?ooo`030=nLB000B9`0gooo00@0oooo00<0omnL04P0
000000000`0000000`00B9`0gooo0?ooo`020?ooo`050=nLB000000000000000001d_ol0103oool0
0`3OW4P004RL0=ooo`090?ooo`030;md001d_ol0oooo00<0oooo00D0gia8000000000000001809cO
o`040?ooo`030?oog`2LB000B9cO00X0oooo00<0oono07@0001d_ol0R@3oool00<80oooo00@0omnL
04P00000B9`0gooo0P3oool00`2oM000000004RLg`0@0?ooo`030=nLB000B9`0gooo00<0oooo00<0
_g@0001d_`3oool00`3oool00`2oM000000004RLg`030?ooo`030?oOW0180000M;oo0080oooo00<0
omnL04P0001d_ol00P3oool00`2oM00007Bo0?ooo`020?ooo`030?noM0000000B9cO00T0oooo00<0
okmd001d_`3oool00P3oool01`3o_g@0000007Boo`3oool0oooO09a80018W=l0103oool00`3OW4P0
04RL0=ooo`090?ooo`030?oo_`1d0000M;oo08T0oooo00320?ooo`030;md000004P0W=oo00@0oooo
00<0okmd00000000M;l03`3oool00`3OW4P004RL0=ooo`020?ooo`030?oo_`1d0000B9cO00<0oooo
00<0okmd00000018W=l0103oool00`3ooml0W4P0001d_`020?ooo`030=nLB000B9`0gooo00<0oooo
00P0gibL0;ooo`3oool0oooo0?oOW0180000B4Q807Boo`P0oooo00<0oono07A8W03Oool00P3oool0
0`3OW4P004RL0=ooo`030?ooo`030=oOg`3oool0oooo0080oooo00<0gia80000M02oool02@3oool0
0`3ookl0M00007Boo`290?ooo`00`@3oool0103ooml0W4P00018W03Oool50?ooo`030=nLB0000000
M;oo00h0oooo00<0gia80018W03Oool00P3oool00`3ookl0M00007Boo`020?ooo`040?oog`2LB000
001d0;ooo`D0oooo00D0_g@00000M02oool0oooo0?noM0030000000304RLg`3oool0oooo00<0oooo
00D0oono07@00018W=l0_gA809cOo`070?ooo`030?oog`2LB4P0W=oo0080oooo00<0okmd0000M02o
ool0203oool00`3o_g@0001d0;ooo`090?ooo`030?oo_`1d0000M;oo08T0oooo00310?ooo`040?oo
g`2LB000001d0;ooo`D0oooo00@0oooO09a8000007@0_ooo3@3oool00`3OW4P004RL0=ooo`020?oo
o`030?oo_`1d0000M;oo0080oooo00@0oono07@0000007@0_ooo1@3oool02`3o_g@0000004RLg`3o
ool0oono07@0B02Lgol0oooo0=nLB0000000B9cO00<0oooo00H0oooO09a80018W=l0oono07A8W03O
ool70?ooo`070=nLB018W=l0oooo0?ooo`3ookl0M00007Boo`080?ooo`030?noM00004P0W=oo00T0
oooo00<0oono07@0001d_ol0R@3oool00<80oooo00<0_g@00000B02Lgol01P3oool00`3OW4P00000
04RLg`0=0?ooo`030=nLB000B9`0gooo0080oooo00<0oono07@0001d_ol00P3oool0103ooml0W4P0
0000M02oool30?ooo`030?oo_`1d000000000080000000T0001d0;ooo`3o_g@004RL0=ooo`3oool0
oono07@00018W=l00P3oool01P3ooml0W4P0001d_`3oool0okmd001d_`L0oooo00<0omnL04Q8W03O
ool00P3oool01`3OW4P0001d0;ooo`3oool0oono07@0001d_ol00`3oool00`3o_g@0001d0;ooo`09
0?ooo`030?oo_`1d0000M;oo08T0oooo00320?ooo`030=nLB0000000B9cO00H0oooo00@0omnL04P0
0000B9`0gooo303oool00`3OW4P004RL0=ooo`020?ooo`030?oo_`1d0000B9cO00<0oooo00<0_g@0
0000B02Lgol02`3oool01`3OW4P0001d0;ooo`3oool0gia80018W03Oool00P3oool01P2oM00004RL
0=ooo`3oool0_g@007Boo`H0oooo00<0oooO09a8M02oool00`3oool01P2oM000001d0;ooo`3oool0
gia804RLg`<0oooo00<0gia80000M02oool02@3oool00`3ookl0M00007Boo`290?ooo`00`P3oool0
103ookl0M0000018W03Oool50?ooo`040?oog`2LB00004RL0=ooo``0oooo00<0gia80018W03Oool0
0P3oool00`3ooml0W4P004RLg`030?ooo`030?noM0000000B9cO00`0oooo00D0omnL04P000000000
000007Boo`030?ooo`070=nLB00007@0_ooo0?ooo`3ooml0W4Q809cOo`060?ooo`030;md001d_ol0
oooo00<0oooo00D0omnL04P000000000001809cOo`030?ooo`030=nLB000B9`0gooo00T0oooo00<0
oono07@0001d_ol0R@3oool00<<0oooo00<0okmd00000018W=l01@3oool0103ooml0W4P00018W02W
M003000000030000M02oool0oooo00H0oooo00<0gia80018W03Oool00`3oool00`3OW4P004RL0=oo
o`020?ooo`040?oog`2LB000001809cOo`H0oooo00<0gibL0;ooo`3oool02P3oool00`3o_g@00018
09cOo`020?ooo`030?oOW01807@0_ooo00D0oooo00<0okmd001d_`3oool02P3oool0103ooml0W4P0
04RLg`2oM003000000030000M02oool0oooo00<0oooo00<0oono07@0001d_ol0R@3oool00<@0oooo
00<0gia80000001d_ol0103oool02@3ookl0M000001d_`3oool0okmd0018W03Oool0omnL09boo`07
0?ooo`030=nLB000B9`0gooo00<0oooo00<0oono07@0B02Lgol00`3oool0103ookl0M0000000M02o
ool50?ooo`030;md001d_ol0oooo00X0oooo00<0omnL04P0001d_ol00`3oool00`3o_g@007Bo0?oo
o`040?ooo`030?oo_`1d07@0_ooo00X0oooo00T0oono07@0B02Lgol0oooo0?noM000B9`0gooo0?oO
W02L_ol0103oool00`3ookl0M00007Boo`290?ooo`00a@3oool00`3o_g@00000001d_`020?ooo`04
0?oog`2LB000001809cOo`80oooo00<0omnL04Q8W03Oool0203oool00`3OW4P004RL0=ooo`040?oo
o`030=nLB018W=l0oooo00<0oooo00@0oooO09a800000000M;oo0`3oool00`3o_g@0000004RLg`0;
0?ooo`030?oOW0180000B9cO00@0oooo00<0_g@0001d_`3oool0103oool00`2oM000M;oo0?ooo`09
0?ooo`030=nLB018W=l0oooo0080oooo00<0omnL04Q8W03Oool01@3oool00`3ookl0M00007Boo`29
0?ooo`00a`3oool01P2oM0000000000000000000001d0;ooo`@0oooo00<0omnL04Q8W03Oool01`3o
ool00`3OW4P0001d0;ooo`050?ooo`030;md0018W=l0oooo00@0oooo00<0oono07@0000000000P00
00001000M;l0oooO09adW03Oool90?ooo`080;md0000000000000000000007@0_ooo0?ooo`3OW4P2
000000030000M02oool0oooo0080oooo00<0gia804RLg`3oool0203oool00`3o_g@007Bo0?ooo`04
0?ooo`030?oOW018B9`0gooo00@0oooo00<0oono07@0001d_ol0R@3oool00=80oooo00<0omnL07BL
g`3oool01P3oool01@3OW4P00000000000000000M;oo00@0oooo00<0giad09cOo`3oool0903oool0
0`3ogi`0B4RL0=ooo`070?ooo`030?oo_`1dM;l0oooo00H0oooo00D0omnL07BLg`3oool0oooo0=nL
B0030000000307Boo`3oool0oooo08L0oooo003C0?ooo`030;mdB02Lgol0oooo04T0oooo00<0_gA8
09cOo`3oool0SP3oool00<l0oooo00H0omoO0=ooo`3oool0oooo0=nLB018W=m60?ooo`060?oOg`3O
ool0oooo0?ooo`3OW4P0B9cOS`3oool00<l0oooo00H0oono07A8W03Oool0oooO09a80018W=m60?oo
o`060?oo_`1dB9`0gooo0?oog`2LB000B9cOS`3oool00=00oooo00D0omnL04P00000000004RL0=oo
o`170?ooo`050?oOW018000000000018W03Oool0S`3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool0
0640oooo00@0_g@0000000000000M;oo:P3oool00`3ookl0M000000000020000000304RLg`3oool0
oooo02T0oooo00<0gia80000000000000`0000000`18W=l0oooo0?ooo`0W0?ooo`030?oo_`1d0000
00000080000000<0M;oo0?ooo`3oool0:`3oool00`3ooml0W4P004RLg`0Z0?ooo`060?oOW0180000
000000000000B9`0gooo:`3oool01@3ooml0W4P000000000000007Bo02X0oooo001P0?ooo`060=nL
B000M;l0oooo0?oOW0180000B9cO:`3oool00`3o_g@0000007Boo`0]0?ooo`030?noM000M;l0oooo
0080oooo00<0_gAd0;ooo`3oool09`3oool00`3o_g@00000001d_`020?ooo`030;mdB02Lgol0oooo
02`0oooo00<0gia80018W03Oool0:@3oool01P3o_g@0000007Boo`3oool0omnL04Qd_b/0oooo00L0
gia8001d_`3oool0oono07@00000B9`0gooo02P0oooo001O0?ooo`030?oog`2LB000B9cO0080oooo
00<0okmd0000B02Lgol0:P3oool00`3ooml0W4P0001d_`0^0?ooo`030?noM000M;l0oooo0300oooo
00<0omnL04Q8W03Oool0;03oool00`3ogi`0B01809cOo`0^0?ooo`030?noM018W=l0oooo02P0oooo
00<0oooO09a8001d_ol00P3oool00`3ogi`0B00004RLg`0X0?ooo`00G`3oool00`3ooml0W4P004RL
g`020?ooo`040?oog`2LB00004RL0=ooobX0oooo00<0gia80018W03Oool0;P3oool00`3o_g@007Bo
0?ooo`0_0?ooo`030?oo_`1d0000M;oo02P0oooo00<0okmd000000000000100000000`00B9`0gooo
0?ooo`0[0?ooo`030?oo_`1d07@0_ooo02P0oooo00<0oono07@0001d_ol00`3oool00`2oM000001d
0;ooo`0W0?ooo`00G`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo02X0oooo00<0
omnL04P0B02Lgol0;`3oool00`3o_g@0B9cO0?ooo`0^0?ooo`030?oOW0180000M;oo02P0oooo00<0
oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`0]0?ooo`030?oog`2LB000M;oo02P0oooo00<0
oooO09a80000M;l00`3oool00`3OW4P0001809cOo`0W0?ooo`00H03oool00`2oM00004RL0=ooo`02
0?ooo`030?oOW0180000B9cO02X0oooo00<0oono07@0001d_ol0<03oool00`3OW4P0B9cO0?ooo`0/
0?ooo`040?oOW0180000001809cOobT0oooo00L0oono07@0M02oool0oooo0?noM000B9`0gooo02d0
oooo00<0oono07@00018W=l0:@3oool00`2oM000001d0;ooo`020?ooo`030=nLB00004P0W=oo02L0
oooo001P0?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l0:`3oool00`2oM00007Bo
0?ooo`0`0?ooo`030;md001d_ol0oooo02T0oooo00D0omnL04P000000000000007Boo`0[0?ooo`06
0?oOW018B9`0gooo0?oOW01804P0W=oo;@3oool00`3o_g@0000007Boo`0Y0?ooo`080?noM0000000
B9cO0?ooo`3ooml0W4P00018W03OoolW0?ooo`00H03oool00`3ogi`0B00004RLg`030?ooo`030;md
0000B9`0gooo02X0oooo00<0gia80018W03Oool0<03oool00`3ogi`0B01d0;ooo`0[0?ooo`030?oo
_`1d000007Bo02`0oooo00D0okmd001d_`3ooml0W4P004RLg`0Z0?ooo`060?oOg`3Oool0okmd0000
000007@0_ooo:@3oool00`3ooml0W4P000000002000000030000M02oool0oooo02L0oooo001P0?oo
o`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool0:P3oool00`3ogi`0B01809cOo`0`
0?ooo`030?oog`2LB000M;oo02d0oooo00<0gia8001d_`3oool0:`3oool0103OW4P0B9cO0;md0000
M;lZ0?ooo`050?oo_`1d000000000000B02Lgol0:`3oool00`3ookl0M000001d_`0[0?ooo`00H@3o
ool00`3ogi`0B00007Boo`020?ooo`030=nLB000M;l0oooo02X0oooo00<0oono07@0001d_ol0<03o
ool00`3ookl0M00004RLg`0]0?ooo`030?oOW01804P0W=oo02`0oooo00@0gia807AdM00007@0_ooo
:P3oool00`3OW9`0_ooo0?ooo`0]0?ooo`030?oOW0180000B9cO02X0oooo001R0?ooo`060=nLB000
07@0_ooo0?oog`2LB000B9cO;03oool00`2oM00007Bo0?ooo`0[0?ooo`070?noM000B9`0gooo0?oo
o`3OW4P0001809cOo`0Z0?ooo`060;md001d_ol0oooo0=nLB00004P0W=oo;@3oool00`2oM0000018
09cOo`0Z0?ooo`030?oOW01dW=l0oooo02h0oooo00@0oooO09a8000007@0_ooo:03oool006<0oooo
00@0gia800000000000007Bo:`3oool01@3ookl0M0000000000007@0_ooo02`0oooo00D0omnL04P0
00000000000007Boo`0/0?ooo`040;md00000000000004RLgbh0oooo00<0oooO09a80018W=l0:`3o
ool01@2oM0000000000000000000B9cO02d0oooo00@0oono07@000000000B9cO9P3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool004l0oooo00@0_g@0000000000000M;oo2P3ooooo
0000034000008`3oool004h0oooo00H0gia8001d_`3oool0omnL04P00018W=l90?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0Q0?ooo`00C@3oool0
0`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo0240oooo001=0?ooo`030?oog`2LB000B9cO
0080oooo00@0oooO09a80000B9`0gooo1`3oool00`000000oooo0?ooo`3o0?ooobd0oooo00<00000
0?ooo`3oool08@3oool004d0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`070?oo
o`030000003oool0oooo0?l0oooo3`3oool2000001`0oooo00<000000?ooo`3oool08@3oool004h0
oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`070?ooo`800000j@3oool2000002<0
oooo1000000I0?ooo`<000008`3oool004h0oooo00<0gia80000B02Lgol00P3oool00`3ooml0W4P0
001d_`070?ooo`030000003oool0oooo0<80oooo0P00000D0?ooodh000001P3oool00`000000oooo
0?ooo`0Q0?ooo`00CP3oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00H0oooo00<0
00000?ooo`3oool0ZP3oool4000000H0oooo9000000?0?ooo`@000008`3oool2000001`0oooo00<0
00000?ooo`3oool08@3oool004h0oooo00@0oooO09a80000B9`0gooo0P3oool00`3OW4P004RL0=oo
o`060?ooo`030000003oool0oooo0980oooo0P0000070?oooaT000003@3oool4000002<0oooo0P00
00120?ooo`030000003oool0oooo0240oooo001?0?ooo`030?oOW0180000M;oo0080oooo00<0gia8
001d_`3oool01P3oool00`000000oooo0?ooo`2?0?ooo``000003P3oool6000001<0oooo0P00001X
0?ooo`030000003oool0oooo0240oooo001@0?ooo`060=nLB00007@0_ooo0?oog`2LB000B9cO1`3o
ool00`000000oooo0?ooo`1j0?ooo`8000001`3oool<00000080oooo1000000E0?ooo`@00000OP3o
ool00`000000oooo0?ooo`0Q0?ooo`00D@3oool0103OW4P0000000000000M;l80?ooo`030000003o
ool0oooo07T0oooo2P00000?0?ooo`800000V03oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool0
0`000000oooo0?ooo`1d0?ooo`T00000[`3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`00
0000oooo0?ooo`1Z0?ooo`8000000`3oool5000000H0oooo0P00002`0?ooo`030000003oool0oooo
0240oooo001M0?ooo`800000JP3oool600000;`0oooo0P00000S0?ooo`00G@3oool00`000000oooo
0?ooo`1N0?ooo`8000000`3oool:00000;l0oooo00<000000?ooo`3oool08@3oool005d0oooo00<0
00000?ooo`3oool0E`3oool3000000<0oooo1P0000070?ooo`800000`03oool00`000000oooo0?oo
o`0Q0?ooo`00G@3oool00`000000oooo0?ooo`1F0?ooo`/00000b`3oool00`000000oooo0?ooo`0Q
0?ooo`00G@3oool00`000000oooo0?ooo`1F0?ooo`D000000`3oool200000<`0oooo00<000000?oo
o`3oool08@3oool005d0oooo00<000000?ooo`3oool0D`3oool700000=80oooo00<000000?ooo`3o
ool08@3oool005d0oooo00<000000?ooo`3oool0BP3oool2000000<0oooo1000003I0?ooo`030000
003oool0oooo0240oooo001M0?ooo`030000003oool0oooo04T0oooo1P00003M0?ooo`030000003o
ool0oooo0240oooo001M0?ooo`030000003oool0oooo04T0oooo1000003O0?ooo`030000003oool0
oooo0240oooo001M0?ooo`800000B@3oool400000=l0oooo0P00000S0?ooo`00G@3oool00`000000
oooo0?ooo`140?ooo`@00000i03oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo
0?ooo`130?ooo`@00000i@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?oo
o`120?ooo`D00000i@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?ooo`0l
0?ooo`8000000P3oool200000080oooo0P00003V0?ooo`030000003oool0oooo0240oooo001M0?oo
o`030000003oool0oooo03/0oooo1@00003/0?ooo`030000003oool0oooo0240oooo001M0?ooo`03
0000003oool0oooo03P0oooo0P0000000`3oool000000000000200000>d0oooo00<000000?ooo`3o
ool08@3oool005d0oooo00<000000?ooo`3oool0=`3oool700000>h0oooo00<000000?ooo`3oool0
8@3oool005d0oooo00<000000?ooo`3oool0=`3oool400000?40oooo00<000000?ooo`3oool08@3o
ool005d0oooo0P00000h0?ooo`<00000l@3oool2000002<0oooo001M0?ooo`030000003oool0oooo
03D0oooo0P00003e0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo03@0
oooo00<000000?ooo`3oool0m@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo
0?ooo`0c0?ooo`030000003oool0oooo0?H0oooo00<000000?ooo`3oool08@3oool00440oooo00@0
_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`020?ooo`030=nLB0000000000000<00000
00<0B9cO0?ooo`3oool01`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo0?L0oooo
00<000000?ooo`3oool08@3oool00400oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`03
0?oOW0180000M;oo00<0oooo00<0okmd001d_`3oool00P3oool00`2oM7@0_ooo0?ooo`070?ooo`03
0000003oool0oooo0340oooo00<000000?ooo`3oool0n03oool00`000000oooo0?ooo`0Q0?ooo`00
?`3oool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00P0oooo00<0okmd001d_`3oool0
2`3oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo0?T0oooo00<000000?ooo`3oool0
8@3oool003l0oooo00<0oooO09a80018W=l00P3oool0103ooml0W4P00018W03Oool80?ooo`030?no
M000M;l0oooo00X0oooo00<000000?ooo`3oool0;`3oool300000?X0oooo00<000000?ooo`3oool0
8@3oool003l0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`090?ooo`030?noM018
W=l0oooo00T0oooo00<000000?ooo`3oool0;@3oool600000?T0oooo00<000000?ooo`3oool08@3o
ool00400oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`0:0?ooo`030=nLB018W=l0
oooo00P0oooo0P00000]0?ooo`030000003oool0oooo00@00000m`3oool3000002<0oooo00100?oo
o`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l02`3oool00`2oM000M;oo0?ooo`070?oo
o`030000003oool0oooo02T0oooo0`0000040?ooo`800000nP3oool00`000000oooo0?ooo`0Q0?oo
o`00@03oool00`3ogi`0B00004RLg`030?ooo`030;md0000B9`0gooo00X0oooo00<0omnL04P0M02o
ool01`3oool00`000000oooo0?ooo`0X0?ooo`@00000o`3oool10?ooo`030000003oool0oooo0240
oooo00100?ooo`040?oog`2LB00004RL0=ooo`80oooo00<0gia80018W03Oool02P3oool00`3ooml0
W4P007Boo`070?ooo`030000003oool0oooo02P0oooo1000003o0?ooo`40oooo00<000000?ooo`3o
ool08@3oool00440oooo00<0omnL04P0001d_ol00P3oool00`3OW4P007Bo0?ooo`0:0?ooo`030?oo
_`1d0000B9cO00L0oooo00<000000?ooo`3oool0:03oool400000?l0oooo0@3oool00`000000oooo
0?ooo`0Q0?ooo`00@P3oool01P3OW4P0001d0;ooo`3ooml0W4P004RLg`L0oooo00L0okmd0018W03O
ool0oooo0=nLB00004P0W=oo00L0oooo00<000000?ooo`3oool09`3oool010000000oooo00000000
003o0?ooo`80oooo00<000000?ooo`3oool08@3oool004<0oooo00@0gia800000000000007Bo2@3o
ool01@3ogi`0B000000000000000M;oo00P0oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`3o0?ooo`<0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool09P3o
ool00`000000oooo0?ooo`3o0?ooo`@0oooo00<000000?ooo`3oool08@3oool005d0oooo00<00000
0?ooo`3oool09@3oool00`000000oooo0?ooo`3o0?ooo`D0oooo00<000000?ooo`3oool08@3oool0
05d0oooo0P00000T0?ooo`040000003oool0000000000?l0oooo1@3oool2000002<0oooo001M0?oo
o`030000003oool0oooo0280oooo0`0000000`3oool000000?ooo`3o0?ooo`D0oooo00<000000?oo
o`3oool08@3oool005d0oooo00<000000?ooo`3oool08@3oool4000000030?ooo`00000000000?l0
oooo1@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?ooo`0Q0?ooo`P00000
o`3oool40?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo0280oooo0P00
00000`3oool000000000000200000?l0oooo103oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool0
0`000000oooo0?ooo`0R0?ooo`040000003oool0oooo0?ooo`800000o`3oool50?ooo`030000003o
ool0oooo0240oooo001M0?ooo`030000003oool0oooo01l0oooo0`0000040?ooo`030000003oool0
oooo0?l0oooo103oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?ooo`0N0?oo
o`@00000o`3oool;0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo01h0
oooo1000003o0?ooo`/0oooo00<000000?ooo`3oool08@3oool005d0oooo0P00000P0?ooo`800000
o`3oool;0?ooo`8000008`3oool005d0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?oo
o`3o0?ooo`/0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool07@3oool2
00000?l0oooo3P3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?ooo`0M0?oo
o`800000o`3oool>0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo01`0
oooo1000003o0?ooo`d0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool0
703oool400000?l0oooo3@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?oo
o`0K0?ooo`@00000o`3oool>0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0
oooo01T0oooo1@00003o0?ooo`l0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?oo
o`3oool0603oool600000?l0oooo3`3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool2000001P0
oooo1`00003o0?ooo`h0oooo0P00000S0?ooo`00G@3oool00`000000oooo0?ooo`0G0?ooo`H00000
o`3oool@0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo01P0oooo0P00
00020?ooo`030000003oool0oooo0?l0oooo3P3oool00`000000oooo0?ooo`0Q0?ooo`00@@3oool0
102oM00000000000001d_ol30?ooo`030?oOW0180000M;oo00D0oooo00<0oooO09a80018W=l02P3o
ool00`000000oooo0?ooo`0I0?ooo`80000000<0oooo0000003oool0o`3oool?0?ooo`030000003o
ool0oooo0240oooo00100?ooo`060=nLB000M;l0oooo0?oOW0180000B9cO0P3oool00`3ogi`0B000
07Boo`060?ooo`030=nLB000B9`0gooo00T0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0000000300000?l0oooo4@3oool00`000000oooo0?ooo`0Q0?ooo`00?`3oool00`3ooml0W4P004RL
g`020?ooo`030?noM00004P0W=oo00X0oooo00<0omnL04P0B02Lgol02@3oool00`000000oooo0?oo
o`0F0?ooo`030000003oool0000000<00000o`3ooolA0?ooo`030000003oool0oooo0240oooo000o
0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO09a80000B9`0gooo1@3oool00`3o_g@000000000
0004000000030018W03Oool0oooo00H0oooo00<000000?ooo`3oool05P3oool00`000000oooo0000
000200000?l0oooo4P3oool00`000000oooo0?ooo`0Q0?ooo`00?`3oool00`3ooml0W4P0001d_`03
0?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a8B02Lgol00P3oool00`2oM00007Bo0?ooo`08
0?ooo`030000003oool0oooo01D0oooo0P0000000`3oool000000?ooo`3o0?oooa<0oooo00<00000
0?ooo`3oool08@3oool00400oooo00<0_g@00018W03Oool00P3oool00`3ogi`0B00004RLg`060?oo
o`070?oo_`1d07@0_ooo0?ooo`3o_g@004RL0=ooo`080?ooo`8000005@3oool400000?l0oooo4`3o
ool3000002<0oooo00100?ooo`030=nLB00004P0W=oo0080oooo00<0oooO09a80000M;l01`3oool0
1P3ogi`0B4RL0=ooo`3ogi`0B01809cOo`P0oooo00<000000?ooo`3oool0503oool400000?l0oooo
5@3oool00`000000oooo0?ooo`0Q0?ooo`00@03oool00`3ogi`0B00004RLg`030?ooo`030;md0000
B9`0gooo00L0oooo00D0okmd001d_`3ooml0W4P004RLg`080?ooo`030000003oool0oooo01D0oooo
1000003o0?oooa@0oooo00<000000?ooo`3oool08@3oool00400oooo00@0oooO09a80000B9`0gooo
0P3oool00`3OW4P004RL0=ooo`080?ooo`040=nLB018W=l0_g@0001d_`P0oooo00<000000?ooo`3o
ool05@3oool400000?l0oooo503oool00`000000oooo0?ooo`0Q0?ooo`00@@3oool00`3ogi`0B000
07Boo`020?ooo`030=nLB000M;l0oooo00T0oooo00@0gia807AdM00007@0_ooo1`3oool00`000000
oooo0?ooo`0D0?ooo`D00000o`3ooolD0?ooo`030000003oool0oooo0240oooo00120?ooo`060=nL
B00007@0_ooo0?oog`2LB000B9cO2`3oool00`2oM000001809cOo`070?ooo`030000003oool0oooo
01@0oooo1000003o0?oooaD0oooo00<000000?ooo`3oool08@3oool002/0oooo00<0gia800000000
00000P0000000`18W=l0oooo0?ooo`0@0?ooo`040=nLB00000000000001d_``0oooo00<0oooO09a8
0018W=l01`3oool00`000000oooo0?ooo`0E0?ooo`800000o`3ooolF0?ooo`030000003oool0oooo
0240oooo000Y0?ooo`030?noM0000000B9cO00@0oooo00<0gia8001d_`3oool0:P3oool00`000000
oooo0?ooo`0E0?ooo`800000o`3ooolF0?ooo`030000003oool0oooo0240oooo000X0?ooo`030=nL
B000M;l0oooo00H0oooo00<0okmd07Boo`3oool0:@3oool00`000000oooo0?ooo`0C0?ooo`<00000
o`3ooolG0?ooo`030000003oool0oooo0240oooo000W0?ooo`030?noM018W=l0oooo00P0oooo00<0
_ibL0=ooo`3oool0:03oool2000001<0oooo1000003o0?oooaH0oooo0P00000S0?ooo`009P3oool0
0`3ooml0W9bo0?ooo`090?ooo`030=nLW02oool0oooo02P0oooo00<000000?ooo`3oool04P3oool4
00000?l0oooo5`3oool00`000000oooo0?ooo`0Q0?ooo`009P3oool00`3ogi`0W;oo0?ooo`090?oo
o`030=nLW02oool0oooo02P0oooo00<000000?ooo`3oool04`3oool200000?l0oooo603oool00`00
0000oooo0?ooo`0Q0?ooo`009P3oool00`3o_i`0W=oo0?ooo`090?ooo`030;mdM02oool0oooo02P0
oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`3o0?oooaH0oooo00<000000?ooo`3o
ool08@3oool002H0oooo00<0gibL0;ooo`3oool00`3oool00`3o_g@0M;oo0?ooo`020?ooo`030?oo
g`2LM9`0gooo02T0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`3o0?oooaL0oooo
00<000000?ooo`3oool08@3oool002H0oooo00<0okmd07Boo`3oool00`3oool0103o_g@0B9cO0?oo
g`2LB0020000000304RLg`3oool0oooo02P0oooo00<000000?ooo`3oool04`3oool00`000000oooo
0?ooo`3o0?oooaL0oooo00<000000?ooo`3oool08@3oool002H0oooo00<0oono07@0M02oool00`3o
ool01P3o_g@00000000000000000001d0;ooob/0oooo00<000000?ooo`3oool04`3oool00`000000
oooo0?ooo`3o0?oooaL0oooo00<000000?ooo`3oool08@3oool002H0oooo00T0oono07@000000000
000007Boo`3oool0okmd0018W03Oool0;P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0
oooo0?l0oooo5`3oool00`000000oooo0?ooo`0Q0?ooo`009P3oool00`3OW4P0000007Boo`030?oo
o`030?noM01d_ol0oooo02h0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`3o0?oo
oaP0oooo00<000000?ooo`3oool08@3oool003D0oooo00<0_gAd0;ooo`3oool09@3oool200000140
oooo0`00003o0?oooaT0oooo0P00000S0?ooo`00<P3oool01@3o_g@000000000000007@0_ooo02H0
oooo00<000000?ooo`3oool03`3oool400000?l0oooo6P3oool00`000000oooo0?ooo`0Q0?ooo`00
;@3oool00`3OW4P000000000000200000005001d_`3oool0oooo0=nLW02oool09P3oool00`000000
oooo0?ooo`0?0?ooo`@00000o`3ooolJ0?ooo`030000003oool0oooo0240oooo000]0?ooo`060=nL
B01d_ol0oooo0?oog`2LM9`0gooo0P3oool00`3OW9`0_ooo0?ooo`0U0?ooo`030000003oool0oooo
0100oooo0P00003o0?oooa/0oooo00<000000?ooo`3oool08@3oool002d0oooo00X0gibL0;ooo`3o
ool0oooO09a800000000M;oo0?oog`2LB7@0_ooo9P3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0?l0oooo6@3oool00`000000oooo0?ooo`0Q0?ooo`00;@3oool01@3OW9`0_ooo0?oo
_`1d000007Bo0080oooo00<0_g@00000M02oool09P3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0?l0oooo6@3oool00`000000oooo0?ooo`0Q0?ooo`00;@3oool00`3OW9`0_ooo0?oo
o`030?ooo`030=nLB000B9`0gooo0080oooo00<0_ibL0=ooo`3oool08P3oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0?l0oooo6P3oool00`000000oooo0?ooo`0Q0?ooo`00;@3oool0
0`3OW4P0000004RLg`070?ooo`030?oOW018B9`0gooo02<0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`3o0?oooaX0oooo00<000000?ooo`3oool08@3oool00300oooo00@0_ibL0=oo
o`3oool0okmd100000000`00B9`0gooo0?ooo`0R0?ooo`030000003oool0oooo0100oooo00<00000
0?ooo`3oool0o`3ooolJ0?ooo`030000003oool0oooo0240oooo000`0?ooo`070;md000000000000
0018W03Oool0_gAd0;ooo`0V0?ooo`8000004@3oool00`000000oooo0?ooo`3o0?oooaT0oooo0P00
000S0?ooo`00<03oool00`3ookl0M7Bo0?ooo`020?ooo`030=nLW02oool0oooo02D0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`3o0?oooa/0oooo00<000000?ooo`3oool08@3oool0
0300oooo00H0gibL0;ooo`3oool0oooo0?oo_`1dM;lW0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool0o`3ooolK0?ooo`030000003oool0oooo0240oooo000_0?ooo`070?oog`2LM9`0
gooo0?ooo`3o_g@0001d0;ooo`0W0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0
o`3ooolK0?ooo`030000003oool0oooo0240oooo000`0?ooo`040=nLB00000000000001d_`d0oooo
00@0_g@0000000000000M;oo0`3oool00`3ogi`0B00007Boo`030?ooo`050?oog`2LB00000000000
0000M;l02P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0?l0oooo6`3oool00`00
0000oooo0?ooo`0Q0?ooo`00@03oool01P3OW4P007Bo0?ooo`3ogi`0B00004RLg`80oooo00<0omnL
04P0001d_ol00`3oool01`3OW4P007Bo0?ooo`3ookl0M0000018W03Oool0203oool00`000000oooo
0?ooo`0>0?ooo`030000003oool0oooo0?l0oooo703oool00`000000oooo0?ooo`0Q0?ooo`00?`3o
ool00`3ooml0W4P004RLg`020?ooo`030?noM00004P0W=oo00H0oooo00<0oooO09a8001d_ol00P3o
ool00`3ogi`0B00004RLg`080?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0o`3o
oolL0?ooo`030000003oool0oooo0240oooo000o0?ooo`030?oog`2LB000B9cO0080oooo00@0oooO
09a80000B9`0gooo1@3oool00`3ookl0M00007Boo`030?ooo`030;md000007@0_ooo00L0oooo00<0
00000?ooo`3oool03P3oool00`000000oooo0?ooo`3o0?oooa`0oooo00<000000?ooo`3oool08@3o
ool003l0oooo00<0oooO09a80000M;l00`3oool00`3OW4P0001809cOo`050?ooo`030?oog`2LB000
07Bo00<0oooo00<0gia80000B02Lgol01`3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0
oooo0?l0oooo703oool00`000000oooo0?ooo`0Q0?ooo`00@03oool00`2oM00004RL0=ooo`020?oo
o`030?oOW0180000B9cO00H0oooo00<0_g@00000M02oool00P3oool00`3OW4P0001809cOo`070?oo
o`8000003@3oool300000?l0oooo703oool3000002<0oooo00100?ooo`030=nLB00004P0W=oo0080
oooo00<0oooO09a80000M;l01P3oool0203o_g@0000004RLg`3oool0oooO09a80000B9`0gooo1`3o
ool00`000000oooo0?ooo`0;0?ooo`@00000o`3ooolN0?ooo`030000003oool0oooo0240oooo0010
0?ooo`030?oOW0180000B9cO00<0oooo00<0_g@00018W03Oool01@3oool00`3ooml0W4P000000002
000000030000M02oool0oooo00L0oooo00<000000?ooo`3oool02`3oool400000?l0oooo7P3oool0
0`000000oooo0?ooo`0Q0?ooo`00@03oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000B9`0
gooo00H0oooo00<0oono07@00000M;l02`3oool00`000000oooo0?ooo`0<0?ooo`800000o`3ooolO
0?ooo`030000003oool0oooo0240oooo00110?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d
_`3oool01`3oool00`3ogi`0B00004RLg`0:0?ooo`030000003oool0oooo00d0oooo00<000000?oo
o`3oool0o`3ooolM0?ooo`030000003oool0oooo0240oooo00120?ooo`060=nLB00007@0_ooo0?oo
g`2LB000B9cO2@3oool0103ooml0W4P00000M02oool80?ooo`030000003oool0oooo00d0oooo00<0
00000?ooo`3oool0o`3ooolM0?ooo`030000003oool0oooo0240oooo00130?ooo`040=nLB0000000
0000001d_``0oooo00@0oono07@000000000B9cO1P3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo0?l0oooo7P3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo0?l0oooo7P3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0?l0oooo7P3oool00`000000oooo0?ooo`0Q
0?ooo`00G@3oool2000000d0oooo00<000000?ooo`3oool0o`3ooolM0?ooo`8000008`3oool005d0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3o0?oooal0oooo00<000000?ooo`3o
ool08@3oool005d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`3o0?oooal0oooo
00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`3o0?oooal0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`3o0?oooal0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?oo
o`3oool02`3oool00`000000oooo0?ooo`3o0?oooal0oooo00<000000?ooo`3oool08@3oool005d0
oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`3o0?ooob00oooo00<000000?ooo`3o
ool08@3oool005d0oooo00<000000?ooo`3oool02@3oool200000?l0oooo8P3oool00`000000oooo
0?ooo`0Q0?ooo`00G@3oool00`000000oooo0?ooo`080?ooo`@00000o`3ooolQ0?ooo`030000003o
ool0oooo0240oooo001M0?ooo`8000002@3oool400000?l0oooo803oool2000002<0oooo001M0?oo
o`030000003oool0oooo00T0oooo0P00003o0?ooob80oooo00<000000?ooo`3oool08@3oool005d0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?ooob40oooo00<000000?ooo`3o
ool08@3oool005d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?ooob40oooo
00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`3o0?ooob40oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`3o0?ooob40oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`3o0?ooob40oooo00<000000?ooo`3oool08@3oool005d0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?ooob40oooo00<000000?ooo`3o
ool08@3oool005d0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`3o0?ooob40oooo
00<000000?ooo`3oool08@3oool005d0oooo0P0000090?ooo`030000003oool0oooo0?l0oooo8@3o
ool2000002<0oooo001M0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3ooolR
0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0o`3ooolR0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0o`3ooolR0?ooo`030000003oool0oooo0240oooo00110?ooo`040;md
00000000000007Boo`<0oooo00<0omnL04P0001d_ol00`3oool01P3ookl0M00000000000000004RL
0=ooo`T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`3o0?ooob80oooo00<00000
0?ooo`3oool08@3oool00400oooo00H0gia8001d_`3oool0omnL04P00018W=l20?ooo`030?oOW018
0000M;oo0080oooo00<0oono07@00018W=l00P3oool00`3OW4P0001d0;ooo`080?ooo`030000003o
ool0oooo00P0oooo00<000000?ooo`3oool0o`3ooolR0?ooo`030000003oool0oooo0240oooo000o
0?ooo`030?oog`2LB000B9cO0080oooo00<0okmd0000B02Lgol01P3oool00`3o_g@004RL0=ooo`03
0?ooo`030;md0018W=l0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`3o
0?ooob80oooo00<000000?ooo`3oool08@3oool003l0oooo00<0oooO09a80018W=l00P3oool0103o
oml0W4P00018W03Oool50?ooo`030?noM000B9`0gooo00<0oooo00<0_g@0001d_`3oool01`3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo0?l0oooo8P3oool00`000000oooo0?ooo`0Q
0?ooo`00?`3oool00`3ooml0W4P0001d_`030?ooo`030=nLB00004P0W=oo00D0oooo00<0oooO09a8
001d_ol00P3oool00`3ookl0M00004RLg`080?ooo`030000003oool0oooo00P0oooo00<000000?oo
o`3oool0o`3ooolR0?ooo`030000003oool0oooo0240oooo00100?ooo`030;md0000B9`0gooo0080
oooo00<0omnL04P00018W=l01`3oool01P2oM000001809cOg`2LB000001d0;ooo`P0oooo0P000009
0?ooo`030000003oool0oooo0?l0oooo803oool3000002<0oooo00100?ooo`030=nLB00004P0W=oo
0080oooo00<0oooO09a80000M;l0203oool01@3ooml0W4P000000000B9`0gooo00P0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`3o0?ooob80oooo00<000000?ooo`3oool08@3oool0
0400oooo00<0omnL04P00018W=l00`3oool00`2oM00004RL0=ooo`070?ooo`060;md000004P0W=oo
0?oOW018B9`0gooo1`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0?l0oooo8P3o
ool00`000000oooo0?ooo`0Q0?ooo`00@03oool0103ooml0W4P00018W03Oool20?ooo`030=nLB000
B9`0gooo00H0oooo00<0omnL04P0001d_ol00P3oool00`3o_g@0B9cO0?ooo`060?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool0o`3ooolS0?ooo`030000003oool0oooo0240oooo0011
0?ooo`030?oOW0180000M;oo0080oooo00<0gia8001d_`3oool01P3oool00`3ogi`0B01809cOo`02
0?ooo`030?noM000B9`0gooo00H0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`3o
0?ooob<0oooo00<000000?ooo`3oool08@3oool00480oooo00H0gia80000M02oool0oooO09a80018
W=l70?ooo`030?oog`2LB000B9cO0080oooo00<0_g@0001d_`3oool01P3oool00`000000oooo0?oo
o`070?ooo`030000003oool0oooo0?l0oooo8`3oool00`000000oooo0?ooo`0Q0?ooo`00@`3oool0
103OW4P0000000000000M;l90?ooo`050?oog`2LB000000000000000M;l0203oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo0?l0oooo8`3oool00`000000oooo0?ooo`0Q0?ooo`00G@3o
ool00`000000oooo0?ooo`060?ooo`800000o`3ooolU0?ooo`030000003oool0oooo0240oooo001M
0?ooo`030000003oool0oooo00D0oooo1000003o0?ooob@0oooo00<000000?ooo`3oool08@3oool0
05d0oooo0P0000060?ooo`@00000o`3ooolS0?ooo`8000008`3oool005d0oooo00<000000?ooo`3o
ool01@3oool300000?l0oooo9@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo
0?ooo`040?ooo`@00000o`3ooolU0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003o
ool0oooo00@0oooo1000003o0?ooobD0oooo00<000000?ooo`3oool08@3oool005d0oooo00<00000
0?ooo`3oool01@3oool200000?l0oooo9P3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`00
0000oooo0?ooo`050?ooo`800000o`3ooolV0?ooo`030000003oool0oooo0240oooo001M0?ooo`03
0000003oool0oooo00@0oooo1000003o0?ooobD0oooo00<000000?ooo`3oool08@3oool005d0oooo
00<000000?ooo`3oool0103oool400000?l0oooo9@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3o
ool00`000000oooo0?ooo`050?ooo`800000o`3ooolV0?ooo`030000003oool0oooo0240oooo001M
0?ooo`8000001@3oool400000?l0oooo903oool2000002<0oooo001M0?ooo`030000003oool0oooo
00@0oooo1000003o0?ooobD0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3o
ool0103oool400000?l0oooo9@3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`000000oooo
0?ooo`030?ooo`@00000o`3ooolV0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003o
ool0oooo00<0oooo1000003o0?ooobH0oooo00<000000?ooo`3oool08@3oool005d0oooo00<00000
0?ooo`3oool00`3oool400000?l0oooo9P3oool00`000000oooo0?ooo`0Q0?ooo`00G@3oool00`00
0000oooo0?ooo`030?ooo`@00000o`3ooolV0?ooo`030000003oool0oooo0240oooo001M0?ooo`03
0000003oool0oooo00<0oooo1000003o0?ooobH0oooo00<000000?ooo`3oool08@3oool005d0oooo
00<000000?ooo`3oool00`3oool400000?l0oooo9P3oool00`000000oooo0?ooo`0Q0?ooo`00G@3o
ool2000000@0oooo1000003o0?ooobD0oooo0P00000S0?ooo`00G@3oool00`000000oooo0?ooo`03
0?ooo`@00000o`3ooolV0?ooo`030000003oool0oooo0240oooo001M0?ooo`030000003oool0oooo
00<0oooo1000003o0?ooobH0oooo00<000000?ooo`3oool08@3oool005d0oooo00<000000?ooo`3o
ool00P3oool500000?l0oooo9P3oool00`000000oooo0?ooo`0Q0?ooo`00C@3oool00`3ookl0M000
000000020000000304RLg`3oool0oooo00P0oooo00<000000?ooo`3oool00P3oool400000?l0oooo
9`3oool00`000000oooo0?ooo`0Q0?ooo`00C`3oool00`3o_g@0000007Boo`0;0?ooo`030000003o
ool0oooo0080oooo1000003o0?ooobL0oooo00<000000?ooo`3oool08@3oool004l0oooo00<0oooO
09a80000M;l02`3oool00`000000oooo0?ooo`020?ooo`@00000o`3ooolW0?ooo`030000003oool0
oooo0240oooo001@0?ooo`030=nLB000B9`0gooo00X0oooo00<000000?ooo`3oool00P3oool40000
0?l0oooo9`3oool00`000000oooo0?ooo`0Q0?ooo`00D03oool00`3ogi`0B01809cOo`0:0?ooo`03
0000003oool0oooo00<0oooo0P00003o0?ooobP0oooo00<000000?ooo`3oool08@3oool00500oooo
00<0oono07@0001d_ol02P3oool2000000D0oooo00<000000?ooo`3oool0o`3ooolT0?ooo`<00000
8`3oool00540oooo00<0_g@0001d_`3oool02@3oool00`000000oooo0?ooo`3o0?ooobd0oooo00<0
00000?ooo`3oool08@3oool00540oooo00<0gia80018W03Oool02@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0]
0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0]
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool08@3oool00540oooo00<0omnL04P0
B02Lgol02@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool08@3oool00540oooo00<0oono07@0001d_ol02@3ooooo0000034000008`3oool00580oooo
00<0_g@0001d_`3oool0o`3ooomL0?ooo`00D03oool01@3ookl0M0000000000007@0_ooo0?l0oooo
G03oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo
/@3oool00?l0oooo/@3oool00?l0oooo/@3oool00?l0oooo/@3oool008/0oooo00H0gia800000000
000000000000B02Lgol60?ooo`030;md0000M;l0oooo00D0oooo00D0okmd000000000000001809cO
o`030?ooo`040?noM0000000001809cOo`80oooo00D0okmd000000000000001809cOo`030?ooo`03
0=nLB000B9`0gooo0080oooo00<0_g@000000000M;l0103oool00`3o_g@000000000000700000006
07Boo`3oool0okmd0000000007@0_ooo0P3oool01@3OW4P00000000000000000M;oo00<0oooo00D0
gia8000000000000001809cOo`020?ooo`070?noM0000000001809cOo`3oool0_g@0001d_`060?oo
o`040;md00000000001809cOo`80oooo00D0gia8000000000000001809cOo`050?ooo`060=nLB000
000000000000000004P0W=oo1`3oool01@3o_g@000000000000004P0W=oo00<0oooo00<0_g@0001d
_`3oool0103oool0303OW4P004RL0=ooo`3oool0okmd0000M02oool0oooo0?oog`2LB000000007Bo
o`@0oooo00H0gia800000000000000000000B02Lgol60?ooo`0;0?oOW018000000000000001d_ol0
gia800000000B9`0gooo0?ooo`3OW4P00P000000200004P0W=oo0?ooo`3oool0okmd0000000004P0
W=oo0P3oool01@3o_g@000000000000004P0W=oo00<0oooo00<0_g@0001d_`3oool0E`3oool008`0
oooo00@0oooO09a80000B9`0gooo1`3oool00`3o_g@0001d0;ooo`040?ooo`070?oOW01804P0W=oo
0?ooo`3ogi`0B000001d_`020?ooo`0<0?noM00004P0W=oO09adW03Oool0omnL04P0B02Lgol0oooo
0?oOW018000007Bo0P3oool00`3o_g@0001d0;ooo`020?ooo`040=nLB000B9`0gkmd07Boo`@0oooo
00@0oooO09a80000B9`0gooo0`3oool0303ooml0W4P00018W03Oool0omnL04P0M02ookl0W9cO0?oo
o`3ogi`0B00007Boo`80oooo01<0_g@0001d_`3oool0okmd0000001d_ol0oooo0?oog`2LB000B9cO
0?ooo`3o_g@0001809cOg`2LM9`0gooo0?noM00007@0_ooo00D0oooo00`0gia80018W03Ooml0W7BL
0=ooo`3o_g@0000007Boo`3oool0oooO09a80018W=l50?ooo`040?oog`2LB00004RL0=ooo`L0oooo
00L0omnL04P0B02Lgol0oooo0?oOW018000007Bo0080oooo00<0okmd0000M02oool0103oool01`3o
_g@0001809cOo`3oool0oono07@0001d_ol00P3oool0102oM00007Bo0;mdM02oool40?ooo`040?oo
g`2LB00004RL0=ooo`L0oooo01d0_g@00018W03Oool0oooo0;md001dM7@0001d0;oOW02L_ol0okmd
0000001d_ol0oooo0?oog`2LB000B9cO0?ooo`3o_g@0001809cOg`2LM9`0gooo0?oOW01804P0W=oo
0?ooo`3ogi`0B000001d_`020?ooo`030?noM00007@0_ooo05L0oooo002=0?ooo`030?noM00004P0
W=oo00L0oooo00<0omnL04P0001d_ol0103oool00`3OW4P004RL0=ooo`020?ooo`070?oOW0180000
B9cO0?ooo`3ookl0M00007Boo`020?ooo`030=nLB000B9`0gooo0080oooo00L0omnL04P00018W=l0
oooo0?oo_`1d0000B9cO0080oooo00<0omnL04P0B02Lgol01P3oool00`3o_g@0001809cOo`050?oo
o`0:0=nLM02Lgol0oono07@0001d_ol0gibL0;ooo`3OW4P004RL0=ooo`<0oooo00D0gibL0;ooo`3O
W4P004RL0=ooo`030?ooo`050=oOg`3oool0oono07@0001d_ol00P3oool00`3ogi`0B00007Boo`05
0?ooo`030?noM00004P0W=oo0080oooo00<0gia80018W03Oool00`3oool00`3Ogml0oooo0?ooo`04
0?ooo`030?noM00004P0W=oo00L0oooo00<0gia80018W03Oool00P3oool01`3ogi`0B00004RLg`3o
ool0omnL04P0001d_ol0103oool01`3ookl0M00004RLg`3oool0oooO09a80000M;l00P3oool00`3O
W4P004RL0=ooo`060?ooo`030?noM00004P0W=oo00L0oooo00<0_g@0001d_`3oool00P3oool01`2o
M000000007Boo`3oool0gia80018W03Oool00`3oool01@3Ogml0oooo0?oo_`1d0000M;oo0080oooo
00<0gia80018W03Oool00P3oool01`3ogi`0B00004RLg`3oool0omnL04P0001d_ol0E`3oool008d0
oooo00<0oono07@00018W=l01`3oool0103ooml0W4P00018W03Oool30?ooo`030?noM00007@0_ooo
00<0oooo00H0gia80000M02oool0oooo0;md0000M;l20?ooo`030?noM00007@0_ooo00<0oooo00/0
gia80000M02oool0oooO09a8000007@0_ooo0?ooo`3ooml0W4P004RLg`060?ooo`030?oo_`1d0000
B9cO00P0oooo00D0_g@004RLg`3oool0oooo0?noM0030000000704RLg`3oool0oooo0?ooo`3o_g@0
001d0;ooo`060?ooo`080;md0000M;l0oooo0?ooo`3ooml0W4P00018W03Oool40?ooo`030?oo_`1d
0000B9cO0080oooo00<0okmd0000M02oool02P3oool00`3ogi`0B00007Boo`070?ooo`030?noM000
07@0_ooo00<0oooo00L0gia80000M02oool0oooO09a80000B9`0gooo00<0oooo00`0oooO09a80000
B9`0gooo0?ooo`3OW4P0000007Boo`3oool0omnL04P0B02Lgol60?ooo`030?oOW0180000M;oo00L0
oooo00<0okmd0000M02oool00P3oool01`3ookl0M00004RLg`3oool0okmd0000M02oool01P3oool0
1`2oM00007Bo0?ooo`3oool0okmd0000M02oool00`3oool01`3OW4P0001d0;ooo`3ooml0W4P00018
W03Oool0EP3oool008d0oooo00<0oooO09a80000M;l0203oool00`2oM000B7Ad07BLg`030?ooo`03
0?oo_`1d0000M;oo00<0oooo00/0omnL04P0B02Lgol0oooo0=nLB000B9`0gooo0?ooo`3ookl0M000
07Boo`030?ooo`070?oOW01804P0W=oo0?ooo`3OW4P004Q804Qd_`020?ooo`030;md0000M;l0oooo
00D0oooo00<0oooO09a80000M;l00P3oool00`3ookl0W9cO0?ooo`030?ooo`0?0=nLB000B9`0gooo
0?ooo`3ookl0M01809cOo`3oool0gia800000018W=l0oooo0?oo_`1d0000M;oo00H0oooo00<0gia8
0018W03Oool00P3oool00`2oM000B7Ad07BLg`050?ooo`070;md0000B9`0gooo0?ooo`3ookl0M000
07Boo`0:0?ooo`030?oog`2LB00007Bo00<0oooo00L0_ibL0=ooo`3oool0oooo0?oo_`1d0000M;oo
00<0oooo00L0omnL04P0B02Lgol0oooo0;md0018M7@0M9cO00@0oooo00/0gia80018B01dW=l0oooo
0?noM00007@0M7Ad0;ooo`3ooml0W4P004RLg`060?ooo`030?oog`2LB00007Bo00<0oooo00L0_ibL
0=ooo`3oool0oooo0?oo_`1d0000M;oo00<0oooo00H0_g@00018W03Oool0oono07@0001d_ol60?oo
o`070=nLB000B9`0gooo0?ooo`3ookl0M00007Boo`030?ooo`070?oOW01804P0W=oo0?ooo`2oM000
B7Ad07BLg`1F0?ooo`00SP3oool00`3OW4P0001804Q800030000000804RLg`3oool0oooo0?ooo`3o
_g@004RL0;nLB01d_ol30?ooo`030=nLB00007@0_ooo0080oooo00L0oono07@0001d_ol0oooo0?oO
W01804P0W=oo0080oooo00<0gia80000M02oool00P3oool0303ookl0M00007Boo`3oool0okmd0000
M02o_g@0B9cO0?ooo`3o_g@0001d0;ooo`H0oooo00H0gia80000M02oool0oooo0?oOW018M;l40?oo
o`030?oOW01807@0_ooo0080oooo01<0okmd0018W03Oool0oooo0?oo_`1d0000B9cO0?ooo`3OW4P0
001d0;ooo`3oool0oono07@0001d_ol0oooo0?oOW01804P0W=oo0080oooo00@0okmd0018W02oW4P0
M;oo103oool00`3OW4P0001809cOo`020?ooo`070=nLB00007@0_ooo0?ooo`3ookl0M00007Boo`06
0?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4Q809cOo`030?ooo`030=nLB00007@0_ooo0080oooo
00P0oono07@0001d_ol0oooo0?noM000B9`0_ia807Boo`<0oooo00`0okmd0000M02LW7@0W=oo0?oo
_`1d0000M;no07Ad_`3oool0_g@00018W03Oool60?ooo`070;md0000B9`0gooo0?ooo`3ooml0W4Q8
09cOo`030?ooo`030?noM00007@0_ooo0080oooo00l0gia80000M02oool0oooo0=nLB00007@0_ooo
0?ooo`3ookl0M00007Boo`3oool0omnL04P0B02Lgol00P3oool00`3OW4P0001d0;ooo`020?ooo`08
0?oo_`1d0000M;oo0?ooo`3o_g@004RL0;nLB01d_omE0?ooo`00SP3oool00`3o_g@0001809cOo`02
0?ooo`0H0?oOW0180000001d0;ooo`3oool0oono07@0B02Lgol0_g@007AdM000B9`0gooo0?ooo`2o
M000001d0;ooo`3oool0okmd0000M02oool0oooo0?oog`2LB000B9cO0`3oool0402oM000001d0;oo
o`3oool0okmd0000M02oool0oooo0?oog`2LB000M;oo0=nLB000M;l0omnL04P0B02Lgol60?ooo`03
0?noM000000000000080000000<0001d0;ooo`3oool00P3oool00`3ookl0M00007Boo`030?ooo`0B
0=nLB00007@0_ooo0?ooo`3OW4P004RL0=ooo`3oool0_g@00000M02oool0oooo0=nLB018W=l0oooo
0?oog`2LB000B9cO0P3oool02`3ookl0M01809cOo`2oM000M7Ad0018W03Oool0oooo0?oOW0180000
B9cO00<0oooo00H0_g@00000M02oool0oooo0=nLB018W=l60?ooo`030?noM0000000000000<00000
00<0M;oo0?ooo`3oool00P3oool06`2oM000001d0;ooo`3oool0okmd0000M02oool0oooo0?oo_`1d
04P0W=oo0;md001dM7@004RL0=ooo`3ookl0M01809cOg`2LB4P0W=oO09a80018W=l0okmd001d_`3o
_g@0001d0;ooo`060?ooo`030?noM0000000000000<0000000<0M;oo0?ooo`3oool00P3oool01`3O
W4P0001d0;ooo`3oool0_g@00000B02Lgol00P3oool02P2oM000001d0;ooo`3oool0gia804RLg`3o
ool0oooO09a80018W=l30?ooo`0?0;md000007@0_ooo0?ooo`3o_g@0001d0;ooo`3oool0oono07@0
B02Lgol0_g@007AdM000B9`0gooo0580oooo002>0?ooo`030?oo_`1d0000B9cO00<0oooo00/0omnL
04P00000M;l0omnL04P000000000M;oo0?ooo`2oM000001d0;ooo`020?ooo`080?noM00000000000
0000001d_ol0oooo0?oo_`1d00030000000407Boo`3oool0oooo0?noM0<0000000/0M;oo0?ooo`3o
okl0M00000000018W=l0oooo0?noM0000000001809cOo`060?ooo`030?oo_`1d0000M;oo0080oooo
00<0_gA809cOo`3oool00`3oool00`2oM000B9cO0?ooo`030?ooo`050?oOW018000000000000001d
_ol00`3oool01`3ogi`0B0000000000004P0W=oo0?oo_`1d00000`0000002@1dW9`0B0000000001d
_ol0oooo0;md000007@0_ooo0?noM002000000030018W03Oool0oooo0080oooo00D0omnL04P00000
0000001809cOo`060?ooo`030?oo_`1d0000M;oo0080oooo00<0gia804RLg`3oool0103oool06P3o
_g@00000000000000000M;oo0?ooo`3ogi`0B0000000001d_ol0oooo0;md000007@0_ono07@00000
0000B9cO0?oog`2LB000000004RLg`3oool0okmd00000000B9`0gooo1P3oool00`3ookl0M00007Bo
o`020?ooo`030=nLB018W=l0oooo00@0oooo00H0omnL04P000000000B9bL07@00018W=l30?ooo`07
0?oOW018000000000000B02Lgol0oono07@000030000000407Boo`3oool0oooo0?noM0<0000000X0
M;oo0?ooo`3ogi`0B0000000001d_ol0oooo0;md000007@0_oooDP3oool008l0oooo00<0_g@00018
W03Oool00`3oool00`2oM000001d0;ooo`0A0?ooo`030?noM00007@0_ooo01T0oooo00H0_g@0001d
_`3oool0oooo0?oOW01dW=l40?ooo`030=nLB000B9`0gooo0180oooo00<0okmd0000M02oool06`3o
ool00`3ooml0W4P0001d_`020?ooo`030?oog`2LM9`0gooo02@0oooo00<0oooO09a80000M;l00P3o
ool00`3ooml0W7BL0=ooo`0D0?ooo`030?noM00007@0_ooo06<0oooo002?0?ooo`030=nLB00007@0
_ooo00<0oooo00<0_g@00000M02oool04P3oool00`3OW7@0W=oo0?ooo`0H0?ooo`030=nLB00007@0
_ooo00@0oooo00H0gibL0;ooo`3oool0omnL04P0M02ooolC0?ooo`030=nLM02Lgol0oooo01/0oooo
00<0gia80018W03Oool0103oool00`3ogi`0W;oo0?ooo`0R0?ooo`030=nLB000B9`0gooo00@0oooo
00<0omnL09boo`3oool04P3oool00`3OW7@0W=oo0?ooo`1R0?ooo`00S`3oool00`3o_g@0000007Bo
o`020?ooo`030?noM000000007Bo02h0oooo00<0okmd0000B02Lgol00`3oool01`3ookl0M01809cO
o`3oool0oooO09a8001d_ol07`3oool0103ookl0M0000018W03Oool>0?ooo`030?noM00004P0W=oo
00@0oooo00<0gia8001d_`3oool08P3oool00`3o_g@0001809cOo`040?ooo`030=nLB000M;l0oooo
07L0oooo002>0?ooo`030;md00000000000000@0000000<007Bo0?ooo`3oool0;03oool00`2oM000
0000000000060000000507Boo`3oool0oooo0=nLB018W=l0<03oool00`2oM0000000000000060000
00030000M02oool0oooo0200oooo00<0_g@00000000000001P0000000`0007@0_ooo0?ooo`1f0?oo
o`00cP3oool0103ooml0W4P000000000M;oN0?ooo`00o`3ooona0?ooo`00o`3ooona0?ooo`00o`3o
oona0?ooo`00o`3ooona0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 431}, {265.875, 0}} -> {-2.08015, -0.309528, \
0.0311177, 0.00839157}}]
}, Closed]],

Cell[TextData[{
  "The theory of nonlinear least-squares is explained in a separate notebook, \
",
  ButtonBox["DataAnalysis.nb",
    ButtonData:>{
      URL[ 
      "www.physics.umd.edu/courses/CourseWare/EssentialMathematica/\
Applications/DataAnalysis.nb"], None},
    ButtonStyle->"Hyperlink"],
  ", which also develops and tests several data analysis programs.  Below we \
load the package containing those programs, request help for the syntax, and \
repeat the fit using ",
  StyleBox["Curfit", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[BoxData[
    \(Needs["\<Curfit`\>"]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(?*fit*\)\)], "Input"],

Cell[BoxData[GridBox[{
        {
          StyleBox["Global`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["fit",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438504-1764354", "Global`fit"},
                  ButtonFrame->"None",
                  ButtonNote->"Global`"], "", "", "", "", "", "", "", "", "", 
                "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", 
                "", "", "", ""}
              },
            RowMinHeight->1.05]},
        {
          StyleBox["Curfit`",
            FontFamily->"Helvetica",
            FontSize->12,
            FontWeight->"Bold"]},
        {GridBox[{
              {
                ButtonBox[
                  StyleBox["Curfit",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438504-1764354", "Curfit`Curfit"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], 
                ButtonBox[
                  StyleBox["Linfit",
                    "Hyperlink"],
                  
                  ButtonFunction:>(
                    Internal`PutInformation[ #, LongForm -> False]&),
                  ButtonEvaluator->Automatic,
                  ButtonData:>{"Info3351438504-1764354", "Curfit`Linfit"},
                  ButtonFrame->"None",
                  ButtonNote->"Curfit`"], "", "", "", "", "", "", "", "", "", 
                "", "", "", "", "", ""}
              },
            RowMinHeight->1.05]}
        },
      RowSpacings->{2, 3, 2, 3},
      ColumnsEqual->True]], "Print",
  CellMargins->{{20, Inherited}, {Inherited, Inherited}},
  ShowCellLabel->False,
  CellFrameMargins->{{Inherited, Inherited}, {14, 14}},
  Background->RGBColor[0.964706, 0.929412, 0.839216],
  ButtonBoxOptions->{Active->True},
  GridBoxOptions->{ColumnAlignments->{Left}}],

Cell[BoxData["\<\"Curfit[data,f,var,par] performs nonlinear least-squares \
fitting using the Levenberg-Marquardt algorithm. The data array should be a \
list of the form \!\(\* \
StyleBox[\\\"{\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"x\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"y\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"{\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"x\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"y\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"dy\\\",\
\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"or\\\",\\nFontSlant->\
\\\"Italic\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"{\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"x\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"dx\\\",\
\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"y\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\"dy\\\",\
\\nFontSlant->\\\"Italic\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Italic\\\"]\) for each measurement. The \
fitting function \!\(\* StyleBox[\\\"f\\\",\\nFontSlant->\\\"Italic\\\"]\) is \
expressed in terms of the independent variable \!\(\* \
StyleBox[\\\"var\\\",\\nFontSlant->\\\"Italic\\\"]\) and the parameters \
\!\(\* StyleBox[\\\"par\\\",\\nFontSlant->\\\"Italic\\\"]\)\!\(\* StyleBox[\\\
\" \\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"are\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \\\",\
\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"given\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"as\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\"a\\\",\\nFontSlant->\\\
\"Plain\\\"]\)\!\(\* StyleBox[\\\" \\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"list\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"containing\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\" \\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"{\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"name\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\",\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"initial\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"value\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"}\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\"for\\\",\\nFontSlant->\
\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"each\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* StyleBox[\\\" \
\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"adjustable\\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\" \\\",\\nFontSlant->\\\"Plain\\\"]\)\!\(\* \
StyleBox[\\\"parameter\\\",\\nFontSlant->\\\"Plain\\\"]\).\"\>"], "Print",
  CellTags->"Info3351438504-1764354"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(gepresults2 = Curfit[gepdata, pade[Q2], Q2, start]\)], "Input"],

Cell[BoxData[
    \({Parameters \[Rule] {a[1] \[Rule] \(-0.06742571662844726`\), 
          b[1] \[Rule] 3.1171356830209787`, b[2] \[Rule] 1.0340705843594717`, 
          b[3] \[Rule] 0.5029653885103269`}, 
      Uncertainties \[Rule] {0.03502186618302049`, 0.05317185141355976`, 
          0.09031748745075263`, 0.15572004989204688`}, 
      ReducedChiSquare \[Rule] 0.8454722225629819`, 
      ChiSquare \[Rule] 40.582666683023135`, NITS \[Rule] 7, 
      Correlations \[Rule] {{1.0000000000000002`, 0.8515743691308576`, 
            0.20371063702294415`, 0.9543891469650041`}, {0.85157436913086`, 
            1.`, \(-0.1819870871583465`\), 
            0.8880548685730124`}, {0.20371063702293665`, \
\(-0.18198708715835318`\), 
            1.`, \(-0.062439650110507394`\)}, {0.9543891469650064`, 
            0.8880548685730122`, \(-0.06243965011049992`\), 
            1.`}}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "These results are equivalent to those from the built-in function, but the \
availability of a custom program offers us the opportunity to understand its \
algorithm in detail and to modify it as the need arises.  For example, the \
fitting functions provided by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " do not permit uncertainties in the independent variable whereas ",
  StyleBox["Linfit", "InlineInput"],
  " and ",
  StyleBox["Curfit", "InlineInput"],
  " from this user package do.  On the other hand, ",
  StyleBox["NonlinearRegress", "InlineInput"],
  " provides more detailed statistical analyses and can handle data with \
several independent variables."
}], "Text"]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
ScreenStyleEnvironment->"Working",
WindowToolbars->"EditBar",
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
ShowSelection->True,
ShowCellLabel->False,
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Open  ]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "Info3351235691-3603827"->{
    Cell[12139, 324, 379, 9, 60, "Print",
      CellTags->"Info3351235691-3603827"]},
  "Info3351235691-3662103"->{
    Cell[12603, 341, 379, 9, 60, "Print",
      CellTags->"Info3351235691-3662103"],
    Cell[12985, 352, 350, 9, 35, "Print",
      CellTags->"Info3351235691-3662103"]},
  "Info3303546715-7418521"->{
    Cell[48798, 1447, 63, 1, 34, "Print",
      CellTags->"Info3303546715-7418521"],
    Cell[48864, 1450, 527, 15, 98, "Print",
      CellTags->"Info3303546715-7418521"]},
  "Info3351237306-3667888"->{
    Cell[86747, 2748, 462, 11, 60, "Print",
      CellTags->"Info3351237306-3667888"]},
  "Info3351438504-1764354"->{
    Cell[791103, 19901, 3797, 61, 110, "Print",
      CellTags->"Info3351438504-1764354"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"Info3351235691-3603827", 830614, 21225},
  {"Info3351235691-3662103", 830730, 21228},
  {"Info3303546715-7418521", 830931, 21233},
  {"Info3351237306-3667888", 831134, 21238},
  {"Info3351438504-1764354", 831252, 21241}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 196, 7, 139, "Title"],
Cell[1975, 62, 455, 15, 85, "Author"],

Cell[CellGroupData[{
Cell[2455, 81, 31, 0, 83, "Section"],
Cell[2489, 83, 1154, 29, 121, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[3680, 117, 41, 0, 51, "Section"],
Cell[3724, 119, 654, 12, 119, "Text"],
Cell[4381, 133, 416, 8, 67, "Text"],
Cell[4800, 143, 1074, 21, 145, "Text"],
Cell[5877, 166, 802, 13, 119, "Text"],
Cell[6682, 181, 1406, 23, 197, "Text"],
Cell[8091, 206, 175, 3, 58, "Exercise"],
Cell[8269, 211, 765, 14, 133, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[9071, 230, 24, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[9120, 234, 34, 0, 68, "Subsection"],
Cell[9157, 236, 904, 21, 119, "Text"],
Cell[10064, 259, 250, 4, 83, "Exercise"],
Cell[10317, 265, 123, 3, 58, "Exercise"],
Cell[10443, 270, 1245, 31, 147, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[11725, 306, 40, 0, 68, "Subsection"],
Cell[11768, 308, 296, 9, 42, "Text"],

Cell[CellGroupData[{
Cell[12089, 321, 47, 1, 73, "Input"],
Cell[12139, 324, 379, 9, 60, "Print",
  CellTags->"Info3351235691-3603827"]
}, Closed]],

Cell[CellGroupData[{
Cell[12555, 338, 45, 1, 73, "Input"],
Cell[12603, 341, 379, 9, 60, "Print",
  CellTags->"Info3351235691-3662103"],
Cell[12985, 352, 350, 9, 35, "Print",
  CellTags->"Info3351235691-3662103"]
}, Closed]],
Cell[13350, 364, 358, 6, 67, "Text"],
Cell[13711, 372, 349, 6, 68, "Text"],

Cell[CellGroupData[{
Cell[14085, 382, 43, 1, 73, "Input"],
Cell[14131, 385, 5102, 126, 128, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[19270, 516, 44, 1, 73, "Input"],
Cell[19317, 519, 9776, 242, 150, "Print"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[29154, 768, 36, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[29215, 772, 29, 0, 68, "Subsection"],
Cell[29247, 774, 1144, 27, 146, "Text"],
Cell[30394, 803, 354, 8, 67, "Text"],
Cell[30751, 813, 764, 26, 237, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[31552, 844, 33, 0, 68, "Subsection"],
Cell[31588, 846, 417, 10, 68, "Text"],

Cell[CellGroupData[{
Cell[32030, 860, 55, 1, 73, "Input"],
Cell[32088, 863, 39, 1, 59, "Output"]
}, Closed]],
Cell[32142, 867, 637, 15, 95, "Text"],

Cell[CellGroupData[{
Cell[32804, 886, 37, 1, 73, "Input"],
Cell[32844, 889, 52, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[32933, 895, 42, 1, 73, "Input"],
Cell[32978, 898, 159, 2, 59, "Output"]
}, Closed]],
Cell[33152, 903, 41, 0, 41, "Text"],
Cell[33196, 905, 414, 8, 67, "Text"],

Cell[CellGroupData[{
Cell[33635, 917, 84, 1, 74, "Input"],
Cell[33722, 920, 111, 2, 61, "Output"]
}, Closed]],
Cell[33848, 925, 201, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[34074, 933, 127, 2, 99, "Input"],
Cell[34204, 937, 150, 2, 84, "Output"]
}, Closed]],
Cell[34369, 942, 73, 0, 71, "Advice"],
Cell[34445, 944, 504, 13, 68, "Text"],

Cell[CellGroupData[{
Cell[34974, 961, 133, 3, 125, "Input"],
Cell[35110, 966, 42, 1, 59, "Output"]
}, Closed]],
Cell[35167, 970, 417, 10, 68, "Text"],
Cell[35587, 982, 44, 1, 73, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[35668, 988, 34, 0, 68, "Subsection"],
Cell[35705, 990, 787, 16, 121, "Text"],

Cell[CellGroupData[{
Cell[36517, 1010, 105, 2, 74, "Input"],
Cell[36625, 1014, 150, 2, 84, "Output"]
}, Closed]],
Cell[36790, 1019, 177, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[36992, 1029, 72, 1, 74, "Input"],
Cell[37067, 1032, 40, 1, 59, "Output"]
}, Closed]],
Cell[37122, 1036, 276, 5, 67, "Text"],
Cell[37401, 1043, 69, 1, 74, "Input"],
Cell[37473, 1046, 57, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[37555, 1050, 113, 2, 73, "Input"],
Cell[37671, 1054, 38, 1, 59, "Output"]
}, Closed]],
Cell[37724, 1058, 282, 9, 42, "Text"],
Cell[38009, 1069, 168, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[38202, 1078, 49, 1, 74, "Input"],
Cell[38254, 1081, 42, 1, 61, "Output"]
}, Closed]],
Cell[38311, 1085, 173, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[38509, 1092, 80, 1, 74, "Input"],
Cell[38592, 1095, 87, 1, 59, "Output"]
}, Closed]],
Cell[38694, 1099, 422, 10, 68, "Text"],

Cell[CellGroupData[{
Cell[39141, 1113, 48, 1, 73, "Input"],
Cell[39192, 1116, 136, 2, 59, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[39377, 1124, 32, 0, 68, "Subsection"],
Cell[39412, 1126, 863, 24, 93, "Text"],

Cell[CellGroupData[{
Cell[40300, 1154, 46, 1, 73, "Input"],
Cell[40349, 1157, 48, 1, 59, "Output"]
}, Closed]],
Cell[40412, 1161, 219, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[40656, 1171, 54, 1, 73, "Input"],
Cell[40713, 1174, 45, 1, 59, "Output"]
}, Closed]],
Cell[40773, 1178, 140, 3, 41, "Text"],
Cell[40916, 1183, 2665, 71, 263, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[43630, 1260, 28, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[43683, 1264, 53, 0, 68, "Subsection"],
Cell[43739, 1266, 796, 15, 120, "Text"],
Cell[44538, 1283, 1553, 39, 257, "Text"],
Cell[46094, 1324, 132, 5, 58, "Exercise"],
Cell[46229, 1331, 223, 8, 59, "Exercise"],
Cell[46455, 1341, 347, 8, 68, "Text"],

Cell[CellGroupData[{
Cell[46827, 1353, 48, 1, 73, "Input"],
Cell[46878, 1356, 38, 1, 75, "Output"]
}, Closed]],
Cell[46931, 1360, 82, 1, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[47050, 1366, 40, 0, 68, "Subsection"],
Cell[47093, 1368, 331, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[47449, 1379, 65, 1, 74, "Input"],
Cell[47517, 1382, 73, 1, 74, "Output"]
}, Closed]],
Cell[47605, 1386, 89, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[47719, 1392, 37, 1, 73, "Input"],
Cell[47759, 1395, 38, 1, 59, "Output"]
}, Closed]],
Cell[47812, 1399, 50, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[47887, 1403, 54, 1, 73, "Input"],
Cell[47944, 1406, 53, 1, 59, "Output"]
}, Closed]],
Cell[48012, 1410, 196, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[48233, 1419, 37, 1, 73, "Input"],
Cell[48273, 1422, 38, 1, 59, "Output"]
}, Closed]],
Cell[48326, 1426, 149, 3, 41, "Text"],
Cell[48478, 1431, 58, 1, 74, "Input"],
Cell[48539, 1434, 192, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[48756, 1444, 39, 1, 73, "Input"],
Cell[48798, 1447, 63, 1, 34, "Print",
  CellTags->"Info3303546715-7418521"],
Cell[48864, 1450, 527, 15, 98, "Print",
  CellTags->"Info3303546715-7418521"]
}, Closed]],
Cell[49406, 1468, 278, 6, 67, "Text"],
Cell[49687, 1476, 71, 1, 74, "Input"],
Cell[49761, 1479, 50, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[49836, 1483, 39, 1, 73, "Input"],
Cell[49878, 1486, 53, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[49968, 1492, 37, 1, 73, "Input"],
Cell[50008, 1495, 73, 1, 75, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[50118, 1501, 51, 1, 73, "Input"],
Cell[50172, 1504, 114, 2, 69, "Output"]
}, Closed]],
Cell[50301, 1509, 397, 11, 68, "Text"],
Cell[50701, 1522, 156, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[50882, 1529, 54, 1, 73, "Input"],
Cell[50939, 1532, 35, 1, 59, "Output"]
}, Closed]],
Cell[50989, 1536, 61, 0, 41, "Text"],
Cell[51053, 1538, 122, 2, 64, "DisplayFormula"],
Cell[51178, 1542, 438, 14, 42, "Text"],
Cell[51619, 1558, 55, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[51699, 1563, 45, 1, 73, "Input"],
Cell[51747, 1566, 40, 1, 59, "Output"]
}, Closed]],
Cell[51802, 1570, 301, 9, 42, "Text"],
Cell[52106, 1581, 99, 2, 71, "Advice"]
}, Closed]],

Cell[CellGroupData[{
Cell[52242, 1588, 44, 0, 68, "Subsection"],
Cell[52289, 1590, 823, 16, 120, "Text"],
Cell[53115, 1608, 146, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[53286, 1617, 170, 3, 99, "Input"],
Cell[53459, 1622, 228, 3, 61, "Output"]
}, Closed]],
Cell[53702, 1628, 268, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[53995, 1637, 143, 3, 74, "Input"],
Cell[54141, 1642, 84, 1, 59, "Output"]
}, Closed]],
Cell[54240, 1646, 795, 24, 95, "Text"],

Cell[CellGroupData[{
Cell[55060, 1674, 132, 2, 74, "Input"],
Cell[55195, 1678, 84, 1, 59, "Output"]
}, Closed]],
Cell[55294, 1682, 486, 8, 94, "Text"],
Cell[55783, 1692, 425, 13, 68, "Text"],

Cell[CellGroupData[{
Cell[56233, 1709, 148, 2, 74, "Input"],
Cell[56384, 1713, 87, 1, 61, "Output"]
}, Closed]],
Cell[56486, 1717, 297, 5, 67, "Text"],
Cell[56786, 1724, 41, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[56852, 1729, 43, 1, 73, "Input"],
Cell[56898, 1732, 42, 1, 59, "Output"]
}, Closed]],
Cell[56955, 1736, 294, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[57274, 1745, 44, 1, 73, "Input"],
Cell[57321, 1748, 42, 1, 59, "Output"]
}, Closed]],
Cell[57378, 1752, 587, 12, 93, "Text"],
Cell[57968, 1766, 72, 0, 71, "Advice"],

Cell[CellGroupData[{
Cell[58065, 1770, 46, 1, 73, "Input"],
Cell[58114, 1773, 42, 1, 59, "Output"]
}, Closed]],
Cell[58171, 1777, 291, 10, 42, "Text"],

Cell[CellGroupData[{
Cell[58487, 1791, 38, 1, 73, "Input"],
Cell[58528, 1794, 41, 1, 59, "Output"]
}, Closed]],
Cell[58584, 1798, 137, 3, 42, "Text"],

Cell[CellGroupData[{
Cell[58746, 1805, 46, 1, 73, "Input"],
Cell[58795, 1808, 49, 1, 59, "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[58905, 1816, 24, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[58954, 1820, 53, 0, 68, "Subsection"],
Cell[59010, 1822, 352, 11, 42, "Text"],

Cell[CellGroupData[{
Cell[59387, 1837, 57, 1, 73, "Input"],
Cell[59447, 1840, 149, 2, 59, "Output"]
}, Closed]],
Cell[59611, 1845, 69, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[59705, 1849, 45, 1, 73, "Input"],
Cell[59753, 1852, 35, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[59825, 1858, 44, 1, 73, "Input"],
Cell[59872, 1861, 37, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[59946, 1867, 46, 1, 73, "Input"],
Cell[59995, 1870, 36, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[60068, 1876, 47, 1, 73, "Input"],
Cell[60118, 1879, 140, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[60295, 1886, 52, 1, 73, "Input"],
Cell[60350, 1889, 138, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[60525, 1896, 47, 1, 73, "Input"],
Cell[60575, 1899, 149, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[60761, 1906, 52, 1, 73, "Input"],
Cell[60816, 1909, 74, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[60927, 1915, 52, 1, 73, "Input"],
Cell[60982, 1918, 109, 2, 59, "Output"]
}, Closed]],
Cell[61106, 1923, 277, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[61408, 1933, 47, 1, 73, "Input"],
Cell[61458, 1936, 38, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[61533, 1942, 80, 1, 73, "Input"],
Cell[61616, 1945, 38, 1, 59, "Output"]
}, Closed]],
Cell[61669, 1949, 144, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[61838, 1958, 78, 1, 74, "Input"],
Cell[61919, 1961, 280, 4, 84, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[62236, 1970, 50, 1, 73, "Input"],
Cell[62289, 1973, 41, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[62367, 1979, 83, 1, 73, "Input"],
Cell[62453, 1982, 36, 1, 59, "Output"]
}, Closed]],
Cell[62504, 1986, 153, 4, 42, "Text"],

Cell[CellGroupData[{
Cell[62682, 1994, 50, 1, 73, "Input"],
Cell[62735, 1997, 709, 18, 252, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[63481, 2020, 63, 1, 73, "Input"],
Cell[63547, 2023, 654, 14, 142, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[64250, 2043, 40, 0, 68, "Subsection"],
Cell[64293, 2045, 809, 20, 120, "Text"],
Cell[65105, 2067, 173, 3, 58, "Exercise"],
Cell[65281, 2072, 125, 2, 99, "Input"],

Cell[CellGroupData[{
Cell[65431, 2078, 40, 1, 73, "Input"],
Cell[65474, 2081, 61, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[65572, 2087, 44, 1, 73, "Input"],
Cell[65619, 2090, 61, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[65717, 2096, 44, 1, 73, "Input"],
Cell[65764, 2099, 61, 1, 75, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[65862, 2105, 41, 1, 74, "Input"],
Cell[65906, 2108, 51, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[65994, 2114, 40, 1, 78, "Input"],
Cell[66037, 2117, 59, 1, 64, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[66133, 2123, 45, 1, 74, "Input"],
Cell[66181, 2126, 61, 1, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[66279, 2132, 55, 1, 74, "Input"],
Cell[66337, 2135, 116, 2, 61, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[66502, 2143, 57, 0, 68, "Subsection"],
Cell[66562, 2145, 304, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[66891, 2157, 56, 1, 74, "Input"],
Cell[66950, 2160, 36, 1, 59, "Output"]
}, Closed]],
Cell[67001, 2164, 310, 9, 42, "Text"],

Cell[CellGroupData[{
Cell[67336, 2177, 49, 1, 73, "Input"],
Cell[67388, 2180, 55, 1, 64, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[67492, 2187, 33, 0, 68, "Subsection"],
Cell[67528, 2189, 160, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[67713, 2198, 50, 1, 73, "Input"],
Cell[67766, 2201, 63, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[67866, 2207, 46, 1, 73, "Input"],
Cell[67915, 2210, 63, 1, 59, "Output"]
}, Closed]],
Cell[67993, 2214, 422, 10, 68, "Text"],

Cell[CellGroupData[{
Cell[68440, 2228, 60, 1, 73, "Input"],
Cell[68503, 2231, 59, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[68599, 2237, 67, 1, 73, "Input"],
Cell[68669, 2240, 74, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[68780, 2246, 64, 1, 73, "Input"],
Cell[68847, 2249, 73, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[68957, 2255, 42, 0, 55, "Subsubsection"],
Cell[69002, 2257, 304, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[69331, 2268, 900, 18, 333, "Input"],
Cell[70234, 2288, 531, 8, 84, "Output"]
}, Closed]],
Cell[70780, 2299, 42, 0, 41, "Text"],
Cell[70825, 2301, 142, 3, 71, "DisplayFormula"],
Cell[70970, 2306, 132, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[71127, 2315, 152, 2, 99, "Input"],
Cell[71282, 2319, 56, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[71375, 2325, 54, 1, 78, "Input"],
Cell[71432, 2328, 55, 1, 59, "Output"]
}, Closed]],
Cell[71502, 2332, 827, 20, 122, "Text"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[72378, 2358, 38, 0, 68, "Subsection"],
Cell[72419, 2360, 711, 26, 68, "Text"],

Cell[CellGroupData[{
Cell[73155, 2390, 65, 1, 73, "Input"],
Cell[73223, 2393, 910, 15, 135, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[74170, 2413, 49, 1, 73, "Input"],
Cell[74222, 2416, 41, 1, 59, "Output"]
}, Closed]],
Cell[74278, 2420, 103, 3, 41, "Text"],
Cell[74384, 2425, 62, 1, 73, "Input"],
Cell[74449, 2428, 74, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[74548, 2432, 73, 1, 73, "Input"],
Cell[74624, 2435, 38, 1, 59, "Output"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[74723, 2443, 27, 0, 51, "Section"],
Cell[74753, 2445, 1141, 25, 146, "Text"],
Cell[75897, 2472, 754, 14, 121, "Text"],
Cell[76654, 2488, 160, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[76839, 2495, 44, 1, 73, "Input"],
Cell[76886, 2498, 9776, 242, 150, "Print"]
}, Closed]],

Cell[CellGroupData[{
Cell[86699, 2745, 45, 1, 73, "Input"],
Cell[86747, 2748, 462, 11, 60, "Print",
  CellTags->"Info3351237306-3667888"]
}, Closed]],
Cell[87224, 2762, 251, 4, 67, "Text"],
Cell[87478, 2768, 161, 3, 99, "Input",
  Evaluatable->False],
Cell[87642, 2773, 67, 0, 41, "Text"],
Cell[87712, 2775, 125, 2, 99, "Input"],
Cell[87840, 2779, 188, 4, 41, "Text"],

Cell[CellGroupData[{
Cell[88053, 2787, 44, 1, 73, "Input"],
Cell[88100, 2790, 41094, 1022, 1216, "Print"]
}, Closed]],
Cell[129209, 3815, 45, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[129279, 3819, 45, 1, 73, "Input"],
Cell[129327, 3822, 23134, 550, 598, "Print"]
}, Closed]],
Cell[152476, 4375, 278, 5, 67, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[152791, 4385, 41, 0, 51, "Section"],
Cell[152835, 4387, 1499, 29, 198, "Text"],
Cell[154337, 4418, 117, 2, 99, "Input"],
Cell[154457, 4422, 150, 2, 73, "Input"],
Cell[154610, 4426, 203, 3, 99, "Input"],
Cell[154816, 4431, 125, 2, 99, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[154978, 4438, 27, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[155030, 4442, 28, 0, 68, "Subsection"],
Cell[155061, 4444, 543, 13, 68, "Text"],
Cell[155607, 4459, 182, 6, 42, "Text"],
Cell[155792, 4467, 442, 13, 68, "Text"],

Cell[CellGroupData[{
Cell[156259, 4484, 69, 1, 73, "Input"],
Cell[156331, 4487, 36240, 922, 279, 13862, 641, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[192586, 5412, 43, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[192654, 5416, 246, 4, 99, "Input"],
Cell[192903, 5422, 70376, 1716, 279, 23412, 1132, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[263294, 7141, 148, 3, 41, "Text"],
Cell[263445, 7146, 175, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[263645, 7155, 75, 1, 73, "Input"],
Cell[263723, 7158, 291, 5, 35, "Message"],
Cell[264017, 7165, 291, 5, 35, "Message"],
Cell[264311, 7172, 291, 5, 35, "Message"],
Cell[264605, 7179, 331, 5, 35, "Message"],
Cell[264939, 7186, 370, 5, 35, "Message"],
Cell[265312, 7193, 370, 5, 35, "Message"],
Cell[265685, 7200, 370, 5, 35, "Message"],
Cell[266058, 7207, 328, 5, 35, "Message"],
Cell[266389, 7214, 33042, 794, 279, 11953, 529, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[299446, 8011, 141, 4, 42, "Text"],

Cell[CellGroupData[{
Cell[299612, 8019, 85, 1, 73, "Input"],
Cell[299700, 8022, 40214, 959, 279, 12947, 618, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[339929, 8984, 552, 10, 94, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[340518, 8999, 26, 0, 68, "Subsection"],
Cell[340547, 9001, 782, 16, 121, "Text"],
Cell[341332, 9019, 376, 8, 152, "Input"],

Cell[CellGroupData[{
Cell[341733, 9031, 87, 1, 73, "Input"],
Cell[341823, 9034, 45726, 1133, 279, 17265, 777, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[387564, 10170, 91, 2, 41, "Text"],

Cell[CellGroupData[{
Cell[387680, 10176, 164, 2, 73, "Input"],
Cell[387847, 10180, 42794, 1057, 279, 14955, 709, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[430656, 11240, 467, 13, 42, "Text"],

Cell[CellGroupData[{
Cell[431148, 11257, 102, 2, 73, "Input"],
Cell[431253, 11261, 1084, 17, 160, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[432374, 11283, 135, 2, 73, "Input"],
Cell[432512, 11287, 54000, 1282, 279, 15846, 806, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[486527, 12572, 1695, 46, 123, "Text"],
Cell[488225, 12620, 898, 23, 120, "Text"],
Cell[489126, 12645, 147, 2, 99, "Input"],
Cell[489276, 12649, 327, 9, 99, "Input"],
Cell[489606, 12660, 328, 9, 99, "Input"],
Cell[489937, 12671, 149, 2, 99, "Input"],
Cell[490089, 12675, 208, 4, 99, "Input"],
Cell[490300, 12681, 145, 2, 99, "Input"],

Cell[CellGroupData[{
Cell[490470, 12687, 554, 10, 177, "Input"],
Cell[491027, 12699, 137652, 2827, 429, 32111, 1520, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[628740, 15533, 38, 0, 51, "Section"],
Cell[628781, 15535, 1129, 23, 147, "Text"],
Cell[629913, 15560, 69, 1, 73, "Input"],
Cell[629985, 15563, 517, 12, 68, "Text"],

Cell[CellGroupData[{
Cell[630527, 15579, 248, 4, 101, "Input"],
Cell[630778, 15585, 74981, 1977, 352, 26552, 1375, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[705774, 17565, 403, 13, 42, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[706214, 17583, 31, 0, 51, "Section"],
Cell[706248, 17585, 1483, 31, 172, "Text"],
Cell[707734, 17618, 231, 6, 41, "Text"],
Cell[707968, 17626, 213, 4, 75, "DisplayFormula"],
Cell[708184, 17632, 691, 16, 93, "Text"],
Cell[708878, 17650, 147, 2, 46, "DisplayFormula"],
Cell[709028, 17654, 118, 3, 41, "Text"],
Cell[709149, 17659, 674, 12, 94, "Text"],
Cell[709826, 17673, 222, 4, 128, "Input"],

Cell[CellGroupData[{
Cell[710073, 17681, 218, 4, 99, "Input"],
Cell[710294, 17687, 4394, 92, 533, "Output"]
}, Closed]],
Cell[714703, 17782, 157, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[714885, 17789, 442, 8, 179, "Input"],
Cell[715330, 17799, 72754, 2012, 279, 28176, 1457, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[788099, 19814, 559, 15, 69, "Text"],
Cell[788661, 19831, 53, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[788739, 19836, 43, 1, 73, "Input"],
Cell[788785, 19839, 2315, 60, 178, "Print"],
Cell[791103, 19901, 3797, 61, 110, "Print",
  CellTags->"Info3351438504-1764354"]
}, Closed]],

Cell[CellGroupData[{
Cell[794937, 19967, 83, 1, 73, "Input"],
Cell[795023, 19970, 894, 15, 134, "Output"]
}, Closed]],
Cell[795932, 19988, 711, 15, 95, "Text"]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

