(************** Content-type: application/mathematica **************
                     CreatedBy='Mathematica 5.2'

                    Mathematica-Compatible Notebook

This notebook can be used with any Mathematica-compatible
application, such as Mathematica, MathReader or Publicon. The data
for the notebook starts with the line containing stars above.

To get the notebook into a Mathematica-compatible application, do
one of the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the
  application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing
the word CacheID, otherwise Mathematica-compatible applications may
try to use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
*******************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[    381521,      10890]*)
(*NotebookOutlinePosition[    417937,      12178]*)
(*  CellTagsIndexPosition[    417825,      12171]*)
(*WindowFrame->Normal*)



Notebook[{

Cell[CellGroupData[{
Cell["Algebra", "Title",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "from ",
  StyleBox[ButtonBox["Essential Mathematica for Students of Science",
    Active->False,
    ButtonData:>{
      URL[ 
      "http://www.physics.umd.edu/courses/CourseWare/EssentialMathematica/"], 
      None},
    Active->False,
    ButtonStyle->"Hyperlink"],
    Active->False],
  "\n\[Copyright] James J. Kelly, 1998\nLast revised: January, 2006"
}], "Author",
  Active->True,
  TextAlignment->Center,
  TextJustification->0],

Cell["\<\
In this notebook we introduce the techniques for manipulating algebraic \
expressions, solving equations, and finding eigenvalues and eigenvectors.\
\>", "Abstract"],

Cell[CellGroupData[{

Cell["Initialization", "Section"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(Off[General::spell, \ General::spell1]\)}], "Input"],

Cell[BoxData[
    \(Needs["\<Utilities`Notation`\>"]\)], "Input"]
}, Closed]],

Cell[CellGroupData[{

Cell["Manipulation of algebraic expressions", "Section"],

Cell[CellGroupData[{

Cell["Rational expressions", "Subsection"],

Cell[TextData[{
  ButtonBox["Together",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " collects terms over a common denominator and cancels common factors.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(a\/\(b + c\) - \ d\/\(e - f\) // Together\)], "Input"],

Cell[BoxData[
    \(\(\(-b\)\ d - c\ d + a\ e - a\ f\)\/\(\((b + c)\)\ \((e - f)\)\)\)], \
"Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["Apart",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " separates an expression into a sum of terms with simpler denominators.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // Apart\)], "Input"],

Cell[BoxData[
    \(a\/\(b + c\) + d\/\(\(-e\) + f\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["Cancel",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " cancels common factors between the numerator and denominator of a \
rational expression.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(2\ a\^2\ b\)\/\(\((b + c)\) a\)\  // Cancel\)], "Input"],

Cell[BoxData[
    \(\(2\ a\ b\)\/\(b + c\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The ",
  ButtonBox["Numerator",
    ButtonStyle->"RefGuideLink"],
  " or ",
  ButtonBox["Denominator",
    ButtonStyle->"RefGuideLink"],
  " of a rational function of polynomials can be extracted using functions of \
the same name.  Note that the numerator is considered the part of the \
expression which does not have a \"superficially\" negative power.  "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(expr\  = \ \(a\ b\^\(-2\)\ g\^\(-n\)\)\/\(\((c - d)\) \((e - \
f)\)\^\(-1\)\);\)\), "\n", 
    \({Numerator[expr], Denominator[expr]}\)}], "Input"],

Cell[BoxData[
    \({a\ \((e - f)\), b\^2\ \((c - d)\)\ g\^n}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Note that the option ",
  StyleBox["Trig\[Rule]True", "InlineInput",
    FontWeight->"Bold"],
  " is needed for use with trigonometric expressions to avoid automatic \
conversions of expressions of the type ",
  Cell[BoxData[
      \(TraditionalForm\`1/Cos[x]\)]],
  "\[Rule]",
  Cell[BoxData[
      \(TraditionalForm\`Sec[x]\)]],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(expr = Sin[x\^2]\/Cos[x\^3];\)\), "\n", 
    \({Numerator[expr], Denominator[expr], 
      Denominator[expr, Trig \[Rule] True]}\)}], "Input"],

Cell[BoxData[
    \({Sec[x\^3]\ Sin[x\^2], 1, Cos[x\^3]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Combine: ",
  Cell[BoxData[
      \(TraditionalForm\`x\/\(1 - x\^2\) - 1\/\(1 + x\)\)]]
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Expanding products", "Subsection"],

Cell[TextData[{
  ButtonBox["ExpandNumerator",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" expands numerators while ",
    FontWeight->"Plain"],
  ButtonBox["ExpandDenominator",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" expands denominators.  (Duh!)",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((a + b)\) \((b - 2  a)\)\)\/\(\((a - b)\) a\)\  // 
      ExpandNumerator\)], "Input"],

Cell[BoxData[
    \(\(\(-2\)\ a\^2 - a\ b + b\^2\)\/\(a\ \((a - b)\)\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((a + b)\) \((b - 2  a)\)\)\/\(\((a - b)\) a\)\  // 
      ExpandDenominator\)], "Input"],

Cell[BoxData[
    \(\(\((\(-2\)\ a + b)\)\ \((a + b)\)\)\/\(a\^2 - a\ b\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["Expand",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " expands numerators leaving denominators in factored form, cancelling \
common factors where possible.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((a + b)\) \((b - 2  a)\)\)\/\(\((a - b)\) a\)\  // 
      Expand\)], "Input"],

Cell[BoxData[
    \(\(-\(\(2\ a\)\/\(a - b\)\)\) - b\/\(a - b\) + 
      b\^2\/\(a\ \((a - b)\)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["ExpandAll",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " expands both numerators and denominators.  Note that common factors are \
not cancelled.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((a + b)\) \((b - 2  a)\)\)\/\(\((a - b)\) a\)\  // 
      ExpandAll\)], "Input"],

Cell[BoxData[
    \(\(-\(\(2\ a\^2\)\/\(a\^2 - a\ b\)\)\) - \(a\ b\)\/\(a\^2 - a\ b\) + 
      b\^2\/\(a\^2 - a\ b\)\)], "Output"]
}, Closed]],

Cell["Cancel common factors from the expression above.", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Factoring sums", "Subsection"],

Cell[TextData[{
  ButtonBox["Factor",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" reduces an expression to a product of factors.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((a + b)\) \((2  a - b)\) \((c - a)\) \((d + b)\)\  // 
      Expand\)], "Input"],

Cell[BoxData[
    \(\(-2\)\ a\^3\ b - a\^2\ b\^2 + a\ b\^3 + 2\ a\^2\ b\ c + a\ b\^2\ c - 
      b\^3\ c - 2\ a\^3\ d - a\^2\ b\ d + a\ b\^2\ d + 2\ a\^2\ c\ d + 
      a\ b\ c\ d - b\^2\ c\ d\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // Factor\)], "Input"],

Cell[BoxData[
    \(\(-\((2\ a - b)\)\)\ \((a + b)\)\ \((a - c)\)\ \((b + d)\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((a + b)\) \((2  a - b)\)\)\/\(\((c - a)\) \((d + b)\)\)\  // 
      ExpandAll\)], "Input"],

Cell[BoxData[
    \(\(2\ a\^2\)\/\(\(-a\)\ b + b\ c - a\ d + c\ d\) + \(a\ b\)\/\(\(-a\)\ b \
+ b\ c - a\ d + c\ d\) - b\^2\/\(\(-a\)\ b + b\ c - a\ d + c\ d\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(% // Factor\)], "Input"],

Cell[BoxData[
    \(\(-\(\(\((2\ a - b)\)\ \((a + b)\)\)\/\(\((a - c)\)\ \((b + 
                d)\)\)\)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "By default, ",
  StyleBox["Factor", "InlineInput",
    FontWeight->"Bold"],
  " is limited to integers and thus will not factor"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(x\^2 + 1\  // Factor\)], "Input"],

Cell[BoxData[
    \(1 + x\^2\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The option ",
  StyleBox["GaussianIntegers\[Rule]True", "InlineInput",
    FontWeight->"Bold"],
  " can be used to extend ",
  StyleBox["Factor", "InlineInput",
    FontWeight->"Bold"],
  " to complex numbers."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[x\^2 + 1, GaussianIntegers \[Rule] True]\)], "Input"],

Cell[BoxData[
    \(\((\(-\[ImaginaryI]\) + x)\)\ \((\[ImaginaryI] + x)\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Options[Factor]\)], "Input"],

Cell[BoxData[
    \({Extension \[Rule] None, GaussianIntegers \[Rule] False, 
      Modulus \[Rule] 0, Trig \[Rule] False}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Similarly, with the option ",
  StyleBox["Trig\[Rule]True", "InlineInput",
    FontWeight->"Bold"],
  " multiple-angle formulae are used to factor expressions involving \
trigonometric and hyperbolic functions."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[Sin[2  x]\ Cos[2  x], Trig \[Rule] True]\)], "Input"],

Cell[BoxData[
    \(2\ Cos[x]\ \((Cos[x] - Sin[x])\)\ Sin[
        x]\ \((Cos[x] + Sin[x])\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Factor[Sinh[2  x]\ Cosh[2  x], Trig \[Rule] True]\)], "Input"],

Cell[BoxData[
    \(2\ Cosh[x]\ Cosh[2\ x]\ Sinh[x]\)], "Output"]
}, Closed]],

Cell[BoxData[
    \(TextForm\`Factor : \(-24\)\ x + 12\ x\^2 - 24\ y + 4\ x\ y + 
        10\ x\^2\ y - 3\ x\^3\ y - 8\ y\^2 + 10\ x\ y\^2 - x\^2\ y\^2 - 
        x\^3\ y\^2 + 2\ x\ y\^3 - x\^2\ y\^3\)], "Exercise"],

Cell["\<\
Factor: 
a) Cos[x]+Cos[y]\tb) Cosh[x]+Cosh[y] \
\>", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Collecting terms", "Subsection"],

Cell[TextData[{
  ButtonBox["Collect",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[expr,x]", "InlineInput"],
  StyleBox[" collects terms involving the same powers of ",
    FontWeight->"Plain"],
  StyleBox["x", "InlineInput"],
  StyleBox[".  ",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((a\  + \ b\ x)\) \((c\ x\  - d\ x\^2)\)\  // 
      Collect[#, x] &\)], "Input"],

Cell[BoxData[
    \(a\ c\ x + \((b\ c - a\ d)\)\ x\^2 - b\ d\ x\^3\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Collect[expr,x,f]", "InlineInput"],
  StyleBox[" applies the function ",
    FontWeight->"Plain"],
  StyleBox["f", "InlineInput"],
  StyleBox[" ",
    FontWeight->"Plain"],
  StyleBox[" ",
    FontWeight->"Plain",
    FontSlant->"Italic"],
  StyleBox["to each coefficient separately. Thus, ",
    FontWeight->"Plain"],
  StyleBox["Collect[", "InlineInput"],
  StyleBox["expr,x,Simplify", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" can be used to simplify the coefficients or ",
    FontWeight->"Plain"],
  StyleBox["Collect[", "InlineInput"],
  StyleBox["expr,x,Factor", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" to factor them.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((a\  + \ b\ x)\) \((a\ c\ x\  - d\ x\^2)\)\  // 
      Collect[#, x, Factor] &\)], "Input"],

Cell[BoxData[
    \(a\^2\ c\ x + a\ \((b\ c - d)\)\ x\^2 - b\ d\ x\^3\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Collect[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr,{x1,x2}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" collects terms involving the same powers from a list of \
objects.",
    FontWeight->"Plain"]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((a\ y\  + \ b\ x + f)\) \((c\ y\^2\  - d\ x\^2\  - g)\)\  // 
      Collect[#, {x, y}] &\)], "Input"],

Cell[BoxData[
    \(\(-f\)\ g - b\ d\ x\^3 - a\ g\ y + c\ f\ y\^2 + a\ c\ y\^3 + 
      x\^2\ \((\(-d\)\ f - a\ d\ y)\) + 
      x\ \((\(-b\)\ g + b\ c\ y\^2)\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((a\ y\  + \ b\ x + f)\) \((c\ y\^2\  - d\ x\^2\  - g)\)\  // 
      Collect[#, {y, x}] &\)], "Input"],

Cell[BoxData[
    \(\(-f\)\ g - b\ g\ x - d\ f\ x\^2 - 
      b\ d\ x\^3 + \((\(-a\)\ g - a\ d\ x\^2)\)\ y + \((c\ f + 
            b\ c\ x)\)\ y\^2 + a\ c\ y\^3\)], "Output"]
}, Closed]],

Cell["\<\
Note that collection priority is ordered according to the listed order of the \
variables.\
\>", "Text"],

Cell[TextData[{
  ButtonBox["FactorTerms",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput"],
  StyleBox["expr,x", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" extracts any that are independent of ",
    FontWeight->"Plain"],
  StyleBox["x", "InlineInput"],
  StyleBox[".  ",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(-a\)\ d\ x - b\ d\ y + a\ c\ x\ y + b\ c\ y\^2\  // 
      FactorTerms[#, x] &\)], "Input"],

Cell[BoxData[
    \(\(-\((a\ x + b\ y)\)\)\ \((d - c\ y)\)\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(-a\)\ d\ x - b\ d\ y + a\ c\ x\ y + b\ c\ y\^2\  // 
      FactorTerms[#, {x, y}] &\)], "Input"],

Cell[BoxData[
    \(\((a\ x + b\ y)\)\ \((\(-d\) + c\ y)\)\)], "Output"]
}, Closed]],

Cell[TextData[StyleBox[
"With only one argument an overall rational factor is pulled out of a \
polynomial.",
  FontWeight->"Plain"]], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(2\ x\  + \ 1\/2\  + \ x\^2\/3\  // FactorTerms\)], "Input"],

Cell[BoxData[
    \(1\/6\ \((3 + 12\ x + 2\ x\^2)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Coefficients are obtained using ",
  ButtonBox["Coefficient",
    ButtonStyle->"RefGuideLink"],
  " or ",
  ButtonBox["CoefficientList",
    ButtonStyle->"RefGuideLink"],
  ".  ",
  StyleBox["Coefficient[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr,x", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " reports the coefficient of ",
  StyleBox["x", "InlineInput"],
  " found in expression ",
  StyleBox["expr", "InlineInput",
    FontWeight->"Bold"],
  "; note that for this purpose ",
  Cell[BoxData[
      \(TraditionalForm\`x\^2\)]],
  " is considered distinct from ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  ", such that the coefficient returned will not contain any power of ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  ".  ",
  StyleBox["Coefficient[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr,x,n", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " returns the coefficient of ",
  Cell[BoxData[
      \(TraditionalForm\`x\^n\)]],
  ".  "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(expr\  = \ \((a\  + 3  x\ y\  + b\/x)\)\^15;\)\), "\n", 
    \({Coefficient[expr, x\^7], Coefficient[expr, x\^2, 7], 
      Coefficient[expr, x, \(-3\)]}\)}], "Input"],

Cell[BoxData[
    \({14073345\ a\^8\ y\^7 + 295540245\ a\^6\ b\ y\^8 + 
        1477701225\ a\^4\ b\^2\ y\^9 + 1773241470\ a\^2\ b\^3\ y\^10 + 
        241805655\ b\^4\ y\^11, 71744535\ a\ y\^14, 
      455\ a\^12\ b\^3 + 45045\ a\^10\ b\^4\ y + 1216215\ a\^8\ b\^5\ y\^2 + 
        11351340\ a\^6\ b\^6\ y\^3 + 36486450\ a\^4\ b\^7\ y\^4 + 
        32837805\ a\^2\ b\^8\ y\^5 + 3648645\ b\^9\ y\^6}\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["CoefficientList[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr,x", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " returns a list of coefficients beginning with ",
  Cell[BoxData[
      \(TraditionalForm\`x\^0\)]],
  ".  ",
  StyleBox["Coefficient[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr,{x1,x2,...}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " returns a rectangular array.  Note that since both ",
  StyleBox["Coefficient", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["CoefficientList", "InlineInput",
    FontWeight->"Bold"],
  " are both intended for use primarily with polynomials, functions other \
than powers may appear in the coefficients.  Also, ",
  StyleBox["CoefficientList", "InlineInput",
    FontWeight->"Bold"],
  " does not handle negative powers."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(expr\  = \ \((x\  + \ y\ Cos[y])\)\^5;\)\), "\n", 
    \(CoefficientList[expr, x]\)}], "Input"],

Cell[BoxData[
    \({y\^5\ Cos[y]\^5, 5\ y\^4\ Cos[y]\^4, 10\ y\^3\ Cos[y]\^3, 
      10\ y\^2\ Cos[y]\^2, 5\ y\ Cos[y], 1}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(CoefficientList[\((x + y)\)\^5, {x, y}]\  // MatrixForm\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {"0", "0", "0", "0", "0", "1"},
            {"0", "0", "0", "0", "5", "0"},
            {"0", "0", "0", "10", "0", "0"},
            {"0", "0", "10", "0", "0", "0"},
            {"0", "5", "0", "0", "0", "0"},
            {"1", "0", "0", "0", "0", "0"}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell[TextData[{
  "Let expr = Nest[\[Mu] #(1-#)&,x,4].\na) Use Collect to produce an explicit \
polynomial in \[Mu] and to factor the coefficients.\nb) Alternatively, use \
CoefficientList to list the coefficients of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Mu]\^n\)]],
  " and then factor those coefficients."
}], "Exercise"],

Cell[TextData[{
  "Deduce the linear transformation ",
  Cell[BoxData[
      \(TraditionalForm\`x \[Rule] a\ z + b\)]],
  " that takes an arbitrary quadratic function ",
  Cell[BoxData[
      \(TraditionalForm\`\[Alpha]\ x\^2 + \[Beta]\ x + \[Gamma]\)]],
  " to the simpler form ",
  Cell[BoxData[
      \(TraditionalForm\`z\^2 + c\)]],
  " by determining ",
  Cell[BoxData[
      \(TraditionalForm\`{a, b, c}\)]],
  " in terms of ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Alpha], \[Beta], \[Gamma]}\)]],
  "."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Using the AlgebraicManipulation palette", "Subsection"],

Cell[TextData[{
  "There exists an ",
  StyleBox["AlgebraicManipulation", "InlineInput",
    FontWeight->"Bold"],
  " palette which can be used to perform many of the operations by ",
  StyleBox["point-and-click",
    FontSlant->"Italic"],
  " rather than by entering commands directly.  If you are unsure of whether \
that palette is active, use the Window submenu of the main toolbar to display \
a list of the active windows.  If ",
  StyleBox["AlgebraicManipulation", "InlineInput",
    FontWeight->"Bold"],
  " is present, clicking on it will bring it to the fore; otherwise use File \
\[RightTeeArrow] Palettes to activate it.  The palette contains a list of \
operations that can be applied to a highlighted expression or subexpression.  \
Highlight the following expression or a portion thereof. Then apply one of \
the transformations, such as ",
  StyleBox["Expand", "InlineInput",
    FontWeight->"Bold"],
  ", by clicking on the palette.  That operation is performed ",
  StyleBox["in situ",
    FontSlant->"Italic"],
  ", leaving the result highlighted.  Clicking on ",
  StyleBox["Factor", "InlineInput",
    FontWeight->"Bold"],
  " undoes the previous the operation.  ",
  StyleBox["Expand", "InlineInput",
    FontWeight->"Bold"],
  " followed by ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " produces another variation.  Experimenting with various transformation \
sequences can produce the desired result more quickly than typing and editing \
commands.  Of course, if you expect to repeat these evaluations later, you \
should record the best sequence of transformations in the notebook itself."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(-\((2\ a - b)\)\)\ \((a + b)\)\ \((a - c)\)\ \((b + d)\)\)], "Input"],

Cell[BoxData[
    \(\((\(-2\)\ a + b)\)\ \((a + b)\)\ \((a - c)\)\ \((b + d)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "This method can be applied to parts of expressions also.  To get some \
familiarity with this technique, perform the following operations on a ",
  StyleBox["copy",
    FontSlant->"Italic"],
  " of the expression below.\n\t1) put the first term in the numerator over a \
common denominator\n\t2) factor the second term of the numerator\n\t3) \
simplify the numerator\n\t4) try simplifying the denominator \[LongDash] the \
result is not likely to be satisfactory\n\t5) try other sequences on the \
whole or parts of the denominator"
}], "Text"],

Cell[BoxData[
    \(\(\(\((a\  + b\ x\/\(c - d\))\) \((c\^2 - d\^2)\)\)\/\(a + b\/\(c + d\/\
\(x - d\)\) + 1\/\(c\ x\  - d\ c + d\)\);\)\)], "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Manipulation of trigonometric expressions", "Section"],

Cell["\<\
Many functions are also available for manipulation of expressions involving \
trigonometric functions, often appearing as counterparts to functions that \
transform algebraic expressions in similar ways.  Both circular and \
hyperbolic trigonometry are handled by these functions.  In addition, there \
are functions which transform between trigonometric and exponential \
representations.\
\>", "Text"],

Cell[TextData[{
  ButtonBox["TrigExpand",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " expands products of trigonometric functions into a sum of terms and \
applies multiple-angle formulas to reduce the arguments to their most basic \
form.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cos[2  x]\ Sin[2  x]\  // TrigExpand\)], "Input"],

Cell[BoxData[
    \(2\ Cos[x]\^3\ Sin[x] - 2\ Cos[x]\ Sin[x]\^3\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cosh[2  x]\ Sinh[2  x]\  // TrigExpand\)], "Input"],

Cell[BoxData[
    \(2\ Cosh[x]\^3\ Sinh[x] + 2\ Cosh[x]\ Sinh[x]\^3\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["TrigFactor",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" factors expressions into products of terms.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(2\ Cosh[x]\^3\ Sinh[x] + 2\ Cosh[x]\ Sinh[x]\^3 // 
      TrigFactor\)], "Input"],

Cell[BoxData[
    \(2\ Cosh[
        x]\ \((Cosh[x] - \[ImaginaryI]\ Sinh[x])\)\ \((Cosh[
            x] + \[ImaginaryI]\ Sinh[x])\)\ Sinh[x]\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["TrigReduce",
    ButtonStyle->"RefGuideLink"],
  StyleBox[
  " uses trigonometric identities to simplify expressions, generally \
attempting to reduce the number of trigonometric functions involved, often \
using multiple-angle formulas.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Sin[x]\^2 + Cos[x]\^\(\(2\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \(Cos[x]\^2 + Sin[x]\^2\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Sin[x]\^2 + Cos[x]\^2 // TrigReduce\)], "Input"],

Cell[BoxData[
    \(1\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(2\ Cos[x]\^3\ Sin[x] - 2\ Cos[x]\ Sin[x]\^3 // TrigReduce\)], "Input"],

Cell[BoxData[
    \(1\/2\ Sin[4\ x]\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["TrigToExp",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" converts trigonometric functions to exponential form.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(stuff = Cosh[2  x]\^2\  - \ Sinh[x]\  // TrigToExp\)], "Input"],

Cell[BoxData[
    \(1\/2 + \[ExponentialE]\^\(\(-4\)\ x\)\/4 + \[ExponentialE]\^\(-x\)\/2 - \
\[ExponentialE]\^x\/2 + \[ExponentialE]\^\(4\ x\)\/4\)], "Output"]
}, Closed]],

Cell[TextData[{
  ButtonBox["ExpToTrig",
    ButtonStyle->"RefGuideLink"],
  StyleBox[" converts from exponential to trigonometric functions.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(stuff // ExpToTrig\)], "Input"],

Cell[BoxData[
    \(1\/2 + 1\/2\ Cosh[4\ x] - Sinh[x]\)], "Output"]
}, Closed]],

Cell["Often several transformations can be profitably combined.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\[Sum]\+\(m = \(-j\)\)\%j Exp[\(m\ y\)\/j] // ExpToTrig\) // 
      TrigFactor\)], "Input"],

Cell[BoxData[
    \(Csch[y\/\(2\ j\)]\ Sinh[y + y\/\(2\ j\)]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Identities can also be checked, but it will often be necessary to use ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " to obtain comparable expressions in both arguments of ",
  StyleBox["Equal", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Cos[x + y]\^2 \[Equal] 1\/2\ \((1 + Cos[2\ x + 2\ y])\) // 
      Simplify\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Closed]],

Cell["\<\
Expand:
a) Cos[4x]\tb) Tanh[2x]\
\>", "Exercise"],

Cell[TextData[{
  "Simplify using TrigReduce: ",
  Cell[BoxData[
      \(TraditionalForm
      \`2 \(\( cosh(x)\)\^3\) \(sinh(x)\) + 
        2 \( cosh(x)\) \(sinh(x)\)\^3\)]],
  "."
}], "Exercise"],

Cell[TextData[{
  "Expand and place over common denominator: \na) ",
  Cell[BoxData[
      \(TraditionalForm\`cos(x + y)\)]],
  "\t\tb) ",
  Cell[BoxData[
      \(TraditionalForm\`tanh(x - y)\)]]
}], "Exercise"],

Cell[TextData[{
  "Expand the following product and use TrigReduce to simplify the result: ",
  Cell[BoxData[
      \(TraditionalForm
      \`\(\((\(cos(x)\) \(cos(y)\) - \(sin(x)\) \(sin(y)\))\)\^2 . \)\)]]
}], "Exercise"],

Cell[TextData[{
  "Use ExpToTrig and TrigFactor to simplify ",
  Cell[BoxData[
      \(TraditionalForm\`\(1 + \[ExponentialE]\^x\)\/\(1 - \[ExponentialE]\^x\
\)\)]],
  ".  Compare with the effect of Simplify."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Watch out for that branch!", "Section"],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " tends to be more careful about its algebra than the average user.  For \
example, many new users are puzzled by the \"failure\" of ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " to reduce the following expression "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\@a\^2 // Simplify\)], "Input"],

Cell[BoxData[
    \(\@a\^2\)], "Output"]
}, Closed]],

Cell[TextData[{
  "or ",
  StyleBox["Expand", "InlineInput",
    FontWeight->"Bold"],
  " to alter"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[a\ b]\  // Expand\)], "Input"],

Cell[BoxData[
    \(Log[a\ b]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "but ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " does not automatically assume that the positive branch of the square root \
is desired or that variables are positive or even that they are real.  If you \
are confident that all relevant quantities are positive, ",
  ButtonBox["PowerExpand",
    ButtonStyle->"RefGuideLink"],
  " can be used to simplify powers and products, expand logarithms, and \
perform related transformations that are valid for positive quantities.  \
Thus,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\@a\^2 // PowerExpand\)], "Input"],

Cell[BoxData[
    \(a\)], "Output"]
}, Closed]],

Cell["gives the positive branch of the square root and", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[a\^2\ b\^3] // PowerExpand\)], "Input"],

Cell[BoxData[
    \(2\ Log[a] + 3\ Log[b]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "separates the log of a product into the sum of their logs assuming that \
both factors are positive and extracts their exponents properly.  However, \
although ",
  StyleBox["PowerExpand", "InlineInput",
    FontWeight->"Bold"],
  " can be very useful, you must ensure that all affected expressions satisfy \
its assumptions.  Thus, if ",
  Cell[BoxData[
      \(TraditionalForm\`a > b\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`c > 0\)]],
  " and all are real, then"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[\((a - b)\)\ c] // PowerExpand\)], "Input"],

Cell[BoxData[
    \(Log[a - b] + Log[c]\)], "Output"]
}, Closed]],

Cell["\<\
is handled correctly, but otherwise you must determine the appropriate branch \
of the complex logarithm function.\
\>", "Text"],

Cell[TextData[{
  "Similarly, ",
  ButtonBox["ComplexExpand",
    ButtonStyle->"RefGuideLink"],
  " assumes that all variables are real, but not necessarily positive.  Thus, \
"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\@a\^2 // ComplexExpand\)], "Input"],

Cell[BoxData[
    \(\@a\^2\)], "Output"]
}, Closed]],

Cell[TextData[{
  "still cannot alter this expression because ",
  Cell[BoxData[
      \(TraditionalForm\`\@a\^2\)]],
  ", though real, could still be ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[PlusMinus]a\)\)]],
  ".  Similarly,"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[a\ b]\  // ComplexExpand\)], "Input"],

Cell[BoxData[
    \(\[ImaginaryI]\ Arg[a\ b] + Log[\@a\^2\ \@b\^2]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "involves the argument function ",
  StyleBox["Arg", "InlineInput",
    FontWeight->"Bold"],
  " because the sign of ",
  Cell[BoxData[
      \(TraditionalForm\`a\ b\)]],
  " is not known ",
  StyleBox["a priori",
    FontSlant->"Italic"],
  ".  Nevertheless, ",
  StyleBox["ComplexExpand", "InlineInput",
    FontWeight->"Bold"],
  " is helpful for expressions like"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Sin[x + \[ImaginaryI]\ y]\  // ComplexExpand\)], "Input"],

Cell[BoxData[
    \(Cosh[y]\ Sin[x] + \[ImaginaryI]\ Cos[x]\ Sinh[y]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "where explicit real and imaginary parts are expressed in terms of real \
variables.  ",
  StyleBox["ComplexExpand[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr,{x1,x2,\[CenterEllipsis]}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " assumes that ",
  StyleBox["{x1,x2,\[CenterEllipsis]}", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontSlant->"Italic"],
  "are complex but that all other variables are real."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Exp[z\^2\ \((x\  + \ \[ImaginaryI]\ y)\)]\  // 
      ComplexExpand[#, {z}] &\)], "Input"],

Cell[BoxData[
    \(\[ExponentialE]\^\(\(-2\)\ y\ Im[z]\ Re[z] + x\ \((\(-Im[z]\^2\) + \
Re[z]\^2)\)\)\ Cos[
          2\ x\ Im[z]\ Re[z] + 
            y\ \((\(-Im[z]\^2\) + 
                  Re[z]\^2)\)] + \[ImaginaryI]\ \[ExponentialE]\^\(\(-2\)\ y\ \
Im[z]\ Re[z] + x\ \((\(-Im[z]\^2\) + Re[z]\^2)\)\)\ Sin[
          2\ x\ Im[z]\ Re[z] + y\ \((\(-Im[z]\^2\) + Re[z]\^2)\)]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "More information about the algebra of complex variables can found in ",
  ButtonBox["complex.nb",
    ButtonData:>{"Additional topics/complex.nb", None},
    ButtonStyle->"Hyperlink"],
  "."
}], "Text"],

Cell["\<\
Caveat emptor: use PowerExpand and ComplexExpand cautiously! You are \
responsible for ensuring that all affected expressions satisfy the \
assumptions made by these functions.\
\>", "Caution",
  FontSlant->"Italic"],

Cell[TextData[{
  "Expand and simplify ",
  Cell[BoxData[
      \(TraditionalForm\`coth(x + \[ImaginaryI]\ y)\)]],
  " assuming that ",
  StyleBox["x",
    FontSlant->"Italic"],
  " and ",
  StyleBox["y",
    FontSlant->"Italic"],
  " are real."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Simplicity is in the eye of the beholder", "Section"],

Cell[TextData[{
  "Perhaps the most useful, but often the most frustrating, ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " function for manipulation of symbolic expressions is ",
  ButtonBox["Simplify",
    ButtonStyle->"RefGuideLink"],
  ", which attempts to reduce an expression to a simpler form.  ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " includes expansion, factorization, and many other algebraic \
transformations.  With the option ",
  StyleBox["Trig\[Rule]True", "InlineInput",
    FontWeight->"Bold"],
  ", which is the default, it applies trigonometric identities also.  Thus, \
because ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " includes all of the functions described above, plus others, the first \
step in simplifying an expression is usually to append a simplification \
command in postfix notation, namely ",
  StyleBox["expr//Simplify", "InlineInput",
    FontWeight->"Bold"],
  ", and seeing what you get.  "
}], "Text"],

Cell["\<\
In the contrived example below this simple operation produces immediate \
gratification\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((4\ \((Cosh[x]\^6\ Sinh[x]\^2 + 2\ Cosh[x]\^4\ Sinh[x]\^4 + 
                Cosh[x]\^2\ Sinh[x]\^6)\))\)/\((8 + 12\ x + 6\ x\^2 + 
            x\^3)\)\  // Simplify\)], "Input"],

Cell[BoxData[
    \(Sinh[4\ x]\^2\/\(4\ \((2 + x)\)\^3\)\)], "Output"]
}, Closed]],

Cell["\<\
whereas for this next expression it is useful to convert from exponential to \
trig before simplifying: \
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(\(y\^2\) \[ExponentialE]\^\(\(y\)\(\ \ \)\)\)\/\((1 + \
\[ExponentialE]\^y)\)\^2\  // ExpToTrig\) // Simplify\)], "Input"],

Cell[BoxData[
    \(1\/4\ y\^2\ Sech[y\/2]\^2\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Sometimes when ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " is stumped, ",
  ButtonBox["FullSimplify",
    ButtonStyle->"RefGuideLink"],
  " can help.  ",
  StyleBox["FullSimplify", "InlineInput",
    FontWeight->"Bold"],
  " tries a much wider range of transformations and includes rules for many \
special functions.  Thus, ",
  StyleBox["FullSimplify", "InlineInput",
    FontWeight->"Bold"],
  " is useful for the next expression where ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " is not."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(2  n\ \ BesselJ[n, x]\  - x\ BesselJ[n + 1, x] // 
      FullSimplify\)], "Input"],

Cell[BoxData[
    \(x\ BesselJ[\(-1\) + n, x]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "However, because ",
  StyleBox["FullSimplify", "InlineInput",
    FontWeight->"Bold"],
  " must test a broader range of possibilities, it can become extremely time \
consuming and can exhaust the memory of your computer.  If you hear the hard \
disk churning, your expression is probably too complicated to handle without \
human guidance.  It is a good idea to save your work before attempting to \
apply ",
  StyleBox["FullSimplify", "InlineInput",
    FontWeight->"Bold"],
  " to an unwieldy expression in case it becomes necessary to abort the \
evaluation, which is not always without risk!"
}], "Text"],

Cell["Use FullSimplify sparingly and do not abuse Simplify either.", "Caution"],

Cell[TextData[{
  "Furthermore, one often has an aesthetic sense of the form that is desired \
and there is no guarantee that ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " shares your notion of simplicity or aesthetics.  You can then attempt to \
guide the simplification process by applying transformation functions, or \
explicit replacement rules, in the order you believe will lead to the desired \
form.  Such a calculation is generally developed by accretion, adding \
transformations sequentially in postfix form until the goal is reached.  \
Sometimes the same transformation has to be applied several times at \
different stages of the calculation, perhaps interspersed with simplification \
commands, with intermediate stages organized using parentheses.  \
Unfortunately, the appropriate order for these steps can be difficult to work \
out and  sometimes changes from one release of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " to the next.  There are few general rules we can impart \[LongDash] \
simplification guided by aesthetics is a fine art acquired only by \
experience.  On the other hand, unless you have a good reason to insist upon \
the simplest possible expressions, simplification is often superfluous.  One \
can calculate and plot functions or investigate many of their properties \
without getting bogged down in unnecessary manipulations."
}], "Text"],

Cell[TextData[{
  "Simplify: \na) ",
  Cell[BoxData[
      \(TraditionalForm\`3 \( sin(x)\) - sin(3  x)\)]],
  "\t\tb) ",
  Cell[BoxData[
      \(TraditionalForm\`cos(\(sin\^\(-1\)\)(x))\)]],
  " \tc) ",
  Cell[BoxData[
      \(TraditionalForm\`\(3 \( tanh(x)\) + \(tanh\^3\)(x)\)\/\(1 + 3 \(\( \
tanh\^2\)(x)\)\)\)]],
  "\t"
}], "Exercise"],

Cell[TextData[{
  "For the expression 2 ",
  Cell[BoxData[
      \(TraditionalForm\`cosh\^3\)]],
  "(x) sinh(x) + 2 cosh(x) ",
  Cell[BoxData[
      \(TraditionalForm\`\(sinh\^3\)(x)\)]],
  ", compare the following transformations:\na) Factor\tb) Factor[expr,Trig\
\[Rule]True]\tc) TrigFactor\t\nd) TrigReduce\te) Simplify\tf) \
Simplify[expr,Trig\[Rule]False]"
}], "Exercise"],

Cell[TextData[{
  "Express ",
  Cell[BoxData[
      \(TraditionalForm\`ln(\(x + 1\)\/\(x - 1\))\)]],
  " in terms of hyperbolic trigonometric functions.  [Hint: you might need \
the substitution ",
  Cell[BoxData[
      \(TraditionalForm\`x \[Rule] E\^\(2  y\)\)]],
  " and several steps.]   I leave it to you to judge which form is simplest \
\[LongDash] most statistical physics textbooks express this formula in terms \
of hyperbolic trigonometric functions, but rational expressions are pretty \
simple also.  Often one's preference depends upon context."
}], "Exercise"],

Cell[CellGroupData[{

Cell["Example", "Subsubsection"],

Cell["\<\
The entropy for a system is proportional to the logarithm of the total number \
of states with the same energy.  Suppose that a system is described by a \
multiplicity function of the form\
\>", "Text"],

Cell[BoxData[
    \(\(multiplicity = \ \((\(\(n!\)\(\ \)\)\/\(\(m!\)\ \(\((n - m)\)!\)\))\)\
\^2;\)\)], "Input"],

Cell[TextData[{
  "where ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " are very large positive integers, of order ",
  Cell[BoxData[
      \(TraditionalForm\`10\^23\)]],
  ", with ",
  Cell[BoxData[
      FormBox[
        RowBox[{"m", "\[LessLess]", 
          StyleBox["n",
            FontSlant->"Italic"]}], TraditionalForm]]],
  ".  Under those circumstances it is useful to apply the Stirling \
approximation for ",
  Cell[BoxData[
      \(TraditionalForm\`Log[\(n!\)]\)]],
  " in the form"
}], "Text"],

Cell[BoxData[
    \(\(StirlingApprox = {Log[\(x_!\)] \[Rule] x\ Log[x] - x};\)\)], "Input"],

Cell[TextData[{
  "First, note that ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  " is ineffective because it does not recognize that the variables in this \
expression are positive and that the logarithm can be reduced without \
specifying the appropriate branch, but ",
  StyleBox["PowerExpand", "InlineInput",
    FontWeight->"Bold"],
  " does help."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[multiplicity]\  // Simplify\)], "Input"],

Cell[BoxData[
    \(Log[\((\(n!\))\)\^2\/\(\((\(m!\))\)\^2\ \((\(\((\(-m\) + \
n)\)!\))\)\^2\)]\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Log[multiplicity]\  // PowerExpand\)], "Input"],

Cell[BoxData[
    \(\(-2\)\ Log[\(m!\)] + 2\ Log[\(n!\)] - 
      2\ Log[\(\((\(-m\) + n)\)!\)]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Next we apply the Stirling approximation.  Note that parentheses are \
included to ensure that ",
  StyleBox["PowerExpand", "InlineInput",
    FontWeight->"Bold"],
  " is executed first.  (Try to do without the parentheses.)"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((Log[multiplicity]\  // PowerExpand)\)\  /. StirlingApprox\)], "Input"],

Cell[BoxData[
    \(\(-2\)\ \((\(-m\) + m\ Log[m])\) + 2\ \((\(-n\) + n\ Log[n])\) - 
      2\ \((m - n + \((\(-m\) + n)\)\ Log[\(-m\) + n])\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Because we know that ",
  Cell[BoxData[
      \(TraditionalForm\`m < n\)]],
  ", it is convenient to make the substitution ",
  Cell[BoxData[
      \(TraditionalForm\`m \[Rule] x\ n\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`0 \[LessEqual] x \[LessEqual] 1\)]],
  " is the ratio between ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  ".  (The physics would also be discussed in terms of ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " if this were a course in statistical physics.)  Therefore, we append this \
replacement rule and another simplification step."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\((Log[multiplicity]\  // PowerExpand)\)\  /. StirlingApprox\)\  /. 
        m \[Rule] x\ n\  // Simplify\)], "Input"],

Cell[BoxData[
    \(2\ n\ \((Log[n] - 
          x\ Log[n\ x] + \((\(-1\) + x)\)\ Log[n - n\ x])\)\)], "Output"]
}, Closed]],

Cell["\<\
Finally, although we probably could use built-in functions, it is easier to \
simplify these logarithms with explicit rules, using cut-and-paste editing.  \
It is again necessary to use parentheses to ensure proper order of \
operations.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((\(\((Log[multiplicity]\  // PowerExpand)\)\  /. StirlingApprox\)\  /. 
              m \[Rule] x\ n\  // Simplify)\)\  /. {Log[n\ x] \[Rule] 
            Log[n]\  + \ Log[x], 
          Log[n - n\ x] \[Rule] Log[n]\  + \ Log[1\  - x]}\  // 
      Simplify\)], "Input"],

Cell[BoxData[
    \(2\ n\ \((\((\(-1\) + x)\)\ Log[1 - x] - x\ Log[x])\)\)], "Output"]
}, Closed]],

Cell["\<\
We have now obtained a tidy expression, and could transform it in other ways \
if desired.  Rather than retaining each step of such a calculation, usually \
only the last pair of input/output cells would be kept in a notebook.  \
Although the fairly complicated input cell above would then appear to have \
sprung from the author's forehead fully armed in Athenian splendor, it would \
actually have been assembled like an oyster's shell one layer at a time.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["FunctionExpand", "Subsection"],

Cell[TextData[{
  "Expressions involving special functions with complicated arguments can \
often be simplified using ",
  ButtonBox["FunctionExpand",
    ButtonStyle->"RefGuideLink"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(FunctionExpand[Sinh[ArcCosh[x]\/2]]\)], "Input",
  CellTags->"FunctionExpand"],

Cell[BoxData[
    \(\@\(\(-1\) + x\)\/\@2\)], "Output",
  CellTags->"FunctionExpand"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(FunctionExpand[Cos[4  ArcTan[x]]] // Simplify\)], "Input",
  CellTags->"FunctionExpand"],

Cell[BoxData[
    \(\(1 - 6\ x\^2 + x\^4\)\/\((1 + x\^2)\)\^2\)], "Output",
  CellTags->"FunctionExpand"]
}, Closed]],

Cell[TextData[{
  "Simplify: Tan[",
  Cell[BoxData[
      \(TraditionalForm\`ArcCos[x]\/2\)]],
  "]"
}], "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Simplification with assumptions", "Section"],

Cell[TextData[{
  "A powerful new feature of simplification became available in ",
  StyleBox["Version 4.0",
    FontSlant->"Italic"],
  ".  It is often possible to simplify symbolic expressions much further when \
the properties of the various symbols, such as their numerical ranges, are \
known in advance.  The syntax ",
  StyleBox["Simplify[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expression,assumptions", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " permits a list of ",
  StyleBox["assumptions", "InlineInput",
    FontWeight->"Bold"],
  " to be employed during the simplification of ",
  StyleBox["expression", "InlineInput",
    FontWeight->"Bold"],
  ".  The assumptions are specified either as a list, ",
  StyleBox["{assumption1,assumption2,\[CenterEllipsis]}", "InlineInput",
    FontWeight->"Bold"],
  " or as a logical expression, such as ",
  StyleBox["assumption1&&assumption2&&assumption3", "InlineInput",
    FontWeight->"Bold"],
  ".  Assumptions can specify the type (domain) for various variables, \
allowed ranges for values, or relationships between variables.  Assumptions \
can be employed with ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["FullSimplify", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["FunctionExpand", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Refine", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Limit", "InlineInput",
    FontWeight->"Bold"],
  ", or ",
  StyleBox["Integrate", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[".",
    FontWeight->"Bold"]
}], "Text"],

Cell["\<\
Consider the following expression.\
\>", "Text"],

Cell[BoxData[
    \(\(expr = Log[a - b] + Log[a + b];\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[expr]\)], "Input"],

Cell[BoxData[
    \(Log[a - b] + Log[a + b]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Simplification is ineffective without further information about ",
  StyleBox["a",
    FontSlant->"Italic"],
  " and ",
  StyleBox["b",
    FontSlant->"Italic"],
  ", but if we know that ",
  Cell[BoxData[
      \(TraditionalForm\`a > b\)]]
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[expr, {a > b}]\)], "Input"],

Cell[BoxData[
    \(Log[a\^2 - b\^2]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "then the two functions can be combined.  Note that the assumption ",
  Cell[BoxData[
      \(TraditionalForm\`a > b\)]],
  " implicitly contains the additional assumption that both ",
  Cell[BoxData[
      \(TraditionalForm\`a\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  " are real numbers, so that their magnitudes can be compared directly.  \
Similarly, the following expressions can be simplified by employing \
inequalities to define ranges."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({Simplify[\@\((x - y)\)\^2], 
      Simplify[\@\((x - y)\)\^2, x > y]}\)], "Input"],

Cell[BoxData[
    \({\@\((x - y)\)\^2, x - y}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[
      2 \@\( x\ y\) \[LessEqual] x + y, {x \[GreaterEqual] 0, 
        y \[GreaterEqual] 0}]\)], "Input"],

Cell[BoxData[
    \(True\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Limit[Exp[x\ y], y \[Rule] \[Infinity], 
      Assumptions \[Rule] {x < 0}]\)], "Input"],

Cell[BoxData[
    \(0\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Often it is sufficient to specify a domain (",
  StyleBox["Integers", "InlineInput"],
  ", ",
  StyleBox["Rationals", "InlineInput"],
  ", ",
  StyleBox["Reals", "InlineInput"],
  ", ",
  StyleBox["Algebraics", "InlineInput"],
  ", ",
  StyleBox["Complexes", "InlineInput"],
  ", ",
  StyleBox["Booleans", "InlineInput"],
  ", ",
  StyleBox["Primes", "InlineInput"],
  ") for one or more variables.  The function ",
  ButtonBox["Element",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["x,domain", "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " specifies that ",
  StyleBox["x\[Element]domain", "InlineInput"],
  " is an element of the specified domain.  The element operator ",
  StyleBox["\[Element]", "InlineInput"],
  " can be entered with the keystroke sequence \[EscapeKey]elem\[EscapeKey].  \
The variations ",
  StyleBox["{x,y,z}\[Element]domain", "InlineInput"],
  " or ",
  StyleBox["(x|y|z)\[Element]domain", "InlineInput"],
  " assign each member of a list to the specified domain.  Domains can also \
be specified for patterns according to ",
  StyleBox["patt\[Element]domain", "InlineInput"],
  ". "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({Simplify[Cos[n\ \[Pi]]], 
      Simplify[Cos[n\ \[Pi]], {n \[Element] Integers}]}\)], "Input"],

Cell[BoxData[
    \({Cos[n\ \[Pi]], \((\(-1\))\)\^n}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Simplify[Cos[n\ \[Pi]], {\(n + 1\)\/2 \[Element] Integers}]\)], "Input"],

Cell[BoxData[
    \(\(-1\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "A related function is ",
  ButtonBox["Refine",
    ButtonStyle->"RefGuideLink"],
  ", which uses assumptions to produce output that more closely approximates \
the form that you might use when its symbols satisfy explicit numerical \
assumptions.  The result is often better than produced by ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  ", or at least more specific."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({Simplify[Abs[x], x < 0], Refine[Abs[x], x < 0]}\)], "Input"],

Cell[BoxData[
    \({Abs[x], \(-x\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \({Simplify[Log[x], x < 0], Refine[Log[x], x < 0]}\)], "Input"],

Cell[BoxData[
    \({Log[x], \[ImaginaryI]\ \[Pi] + Log[\(-x\)]}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \({Simplify[Cos[x + \[Pi]  y], y \[Element] Integers], 
      Refine[Cos[x + \[Pi]  y], y \[Element] Integers]}\)], "Input",
  CellTags->"Refine"],

Cell[BoxData[
    \({Cos[x + \[Pi]\ y], \((\(-1\))\)\^y\ Cos[x]}\)], "Output",
  CellTags->"Refine"]
}, Closed]],

Cell["\<\
If one is performing a long derivation involving many symbols, it can become \
cumbersome to specify a long list of assumptions many times.  The following \
example illustrates a useful technique for performing such calculations.  \
First we create a list a assumptions that apply to the variables of interest. \
 Then we define our own versions of the simplification functions using these \
assumptions.  Two examples using these functions follow.\
\>", "Text"],

Cell[BoxData[{
    \(\(MyAssumptions = {0 < \[Theta]\_e < \[Pi], \[Omega] > 0, q > Q, Q > 0, 
          0 < \[Epsilon] < 1, \[CurlyEpsilon]\_f > 
            0, \[CurlyEpsilon]\_i > \[CurlyEpsilon]\_f};\)\), "\n", 
    \(\(MySimplify = Simplify[#, MyAssumptions] &;\)\), "\n", 
    \(\(MyFullSimplify = FullSimplify[#, MyAssumptions] &;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol1 = \(Solve[\[Epsilon] \[Equal] \((1 + 2\ \(q\^2\/Q\^2\) \
Tan[\[Theta]\_e\/2]\^2)\)\^\(-1\), \[Theta]\_e]\)\[LeftDoubleBracket]2\
\[RightDoubleBracket] // MySimplify\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"ifun"\), \(\(:\)\(\ \)\), "\<\"Inverse functions are \
being used by \\!\\(Solve\\), so some solutions may not be found; use Reduce \
for complete solution information. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({\[Theta]\_e \[Rule] \(-2\)\ ArcTan[\(Q\ \@\(\(-1\) + 1\/\[Epsilon]\)\)\
\/\(\@2\ q\)]}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(Solve[
              Q\^2 \[Equal] 
                4\ \(\[CurlyEpsilon]\_i\) \((\[CurlyEpsilon]\_i - \[Omega])\)\
\ Sin[\[Theta]\_e\/2]\^2, \[CurlyEpsilon]\_i]\)\[LeftDoubleBracket]2\
\[RightDoubleBracket] /. \[Omega] \[Rule] \@\(q\^2 - Q\^2\)\) /. sol1 // 
      MyFullSimplify\)], "Input"],

Cell[BoxData[
    \({\[CurlyEpsilon]\_i \[Rule] 
        1\/2\ \((\@\(\((q - Q)\)\ \((q + Q)\)\) + \(q\ \((1 + \[Epsilon])\)\)\
\/\@\(1 - \[Epsilon]\^2\))\)}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Although the formatting procedure used by ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " still has some peculiarities (",
  StyleBox["e.g.",
    FontSlant->"Italic"],
  ", too many minus signs), these results were simplified fairly completely \
with relatively little effort, whereas much more complicated expressions \
would have been produced had we not told ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " what assumptions it could employ  (Try it!).  In a real problem we would \
probably use our simplification routines many more times, perhaps adding \
additional assumptions as the solution is developed."
}], "Text"],

Cell[TextData[{
  "Under what assumptions will ",
  Cell[BoxData[
      \(TraditionalForm\`\((x\^m)\)\^n\)]],
  " reduce to ",
  Cell[BoxData[
      \(TraditionalForm\`x\^\(m\ n\)\)]],
  "?  Verify.  List a few examples which show that uncritical use of the \
proposed replacement rule leads to incorrect results."
}], "Exercise"],

Cell[TextData[{
  "Prove ",
  Cell[BoxData[
      \(TraditionalForm\`a\^n - 
          b\^n = \((a - 
              b)\) \((a\^\(n - 1\) + \(a\^\(n - 2\)\) 
                b\  + \ \[CenterEllipsis] + \ b\^\(n - 1\))\)\)]],
  " for any positive integer ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " and real ",
  Cell[BoxData[
      \(TraditionalForm\`a, b\)]],
  "."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Symbolic solution of equations", "Section"],

Cell[CellGroupData[{

Cell["Basic syntax", "Subsection"],

Cell[TextData[{
  "The basic tools for symbolic solution of algebraic equations are ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  ", ",
  StyleBox["Reduce", "InlineInput",
    FontWeight->"Bold"],
  ", and ",
  StyleBox["Eliminate", "InlineInput",
    FontWeight->"Bold"],
  ".  ",
  ButtonBox["Solve",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["eqs,vars", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " attempts to solve an equation or list of equations, ",
  StyleBox["eqs", "InlineInput",
    FontWeight->"Bold"],
  ", for the variable or list of variables, ",
  StyleBox["vars", "InlineInput",
    FontWeight->"Bold"],
  ".  Equations must be expressed in the form ",
  Cell[BoxData[
      FormBox[
        RowBox[{"lhs", 
          StyleBox["\[Equal]",
            FontSlant->"Plain"], "rhs"}], TraditionalForm]], "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[BoxData[
    \(\(eq1\  = \ a\ x\^2 + \ b\ x\  + \ c\  \[Equal] 0;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol1 = \ Solve[eq1, x]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(\(-b\) - \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)}, {x \[Rule] \(\
\(-b\) + \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)}}\)], "Output"]
}, Closed]],

Cell["\<\
Solutions are returned as sets of replacement rules.  In this case there are \
two possible solutions, each containing a single replacement rule for the \
only independent variable.  A particular solution can then be selected by \
using the replacement rule of your choice to make an assignment\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(x1\  = \ x /. sol1\[LeftDoubleBracket]1\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \(\(\(-b\) - \@\(b\^2 - 4\ a\ c\)\)\/\(2\ a\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Caution:",
    FontColor->RGBColor[1, 0, 0]],
  " if you use assignments of the form ",
  StyleBox["x=x/.solution", "InlineInput",
    FontWeight->"Bold"],
  " you will contaminate the symbol ",
  StyleBox["x", "InlineInput",
    FontWeight->"Bold"],
  " and may have difficulty revising your equation later."
}], "Text"],

Cell["\<\
Sometimes it is useful to verify that solutions are correct by substitution \
into the original equations.  It is usually necessary to perform some \
simplification steps before the veracity of the equations becomes manifest.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq1 /. sol1 // Simplify\)], "Input"],

Cell[BoxData[
    \({True, True}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "A system of equations is expressed in the form of a list of equations, ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          StyleBox["{",
            FontSlant->"Italic"], 
          RowBox[{
            RowBox[{
              StyleBox["lhs1",
                FontSlant->"Italic"], 
              StyleBox["\[Equal]",
                FontSlant->"Plain"], 
              StyleBox["rhs1",
                FontSlant->"Italic"]}], 
            StyleBox[",",
              FontSlant->"Italic"], 
            StyleBox[" ",
              FontSlant->"Italic"], 
            StyleBox[
              RowBox[{
                StyleBox["lhs2",
                  FontSlant->"Italic"], 
                StyleBox["\[Equal]",
                  FontSlant->"Plain"], 
                StyleBox["rhs2",
                  FontSlant->"Italic"]}],
              FontSlant->"Italic"], 
            StyleBox[",",
              FontSlant->"Italic"], 
            StyleBox[" ",
              FontSlant->"Italic"], 
            StyleBox["\[CenterEllipsis]",
              FontSlant->"Italic"]}], 
          StyleBox["}",
            FontSlant->"Italic"]}], TraditionalForm]]],
  " or as an expression in which equations are combined using ",
  StyleBox["And", "InlineInput",
    FontWeight->"Bold"],
  " (",
  StyleBox["&&", "InlineInput"],
  ") in the form ",
  StyleBox["lhs1",
    FontSlant->"Italic"],
  "\[Equal] ",
  StyleBox["rhs1 && lhs2",
    FontSlant->"Italic"],
  "\[Equal]",
  StyleBox[" rhs2 \[CenterEllipsis]",
    FontSlant->"Italic"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(Solve[{3  x\^2\  - 2  y\^2\  \[Equal] \ 1, \ 
        x\^2\  + \ 4  y\^2 \[Equal] 3}, {x, y}]\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(-\@\(5\/7\)\), 
        y \[Rule] \(-\(2\/\@7\)\)}, {x \[Rule] \(-\@\(5\/7\)\), 
        y \[Rule] 2\/\@7}, {x \[Rule] \@\(5\/7\), 
        y \[Rule] \(-\(2\/\@7\)\)}, {x \[Rule] \@\(5\/7\), 
        y \[Rule] 2\/\@7}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Now ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  " returns four solutions, each consisting of two replacement rules.  In \
other cases, degenerate solutions are duplicated."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[36 + 12\ x - 11\ x\^2 - 2\ x\^3 + x\^4 \[Equal] 0, x]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(-2\)}, {x \[Rule] \(-2\)}, {x \[Rule] 3}, {x \[Rule] 
          3}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Solve", "InlineInput"],
  StyleBox[" does not check for special cases that might arise for some \
choices of parameters.  For example, the solution proposed for the following \
equation",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[a\ x\^2 + b\ x\  \[Equal] 1, x]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(\(-b\) - \@\(4\ a + b\^2\)\)\/\(2\ a\)}, {x \[Rule] \
\(\(-b\) + \@\(4\ a + b\^2\)\)\/\(2\ a\)}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "does not accommodate the special case ",
  Cell[BoxData[
      \(TraditionalForm\`a \[Rule] 0\)]],
  ".  More complete solutions are provided by ",
  ButtonBox["Reduce",
    ButtonStyle->"RefGuideLink"],
  ", which checks special cases and returns solutions in the form of logical \
conditions upon both variables and parameters."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Reduce[a\ x\^2 + b\ x\  \[Equal] 1, x]\)], "Input"],

Cell[BoxData[
    \(\((a \[NotEqual] 
            0 && \((x \[Equal] \(\(-b\) - \@\(4\ a + b\^2\)\)\/\(2\ a\) || 
              x \[Equal] \(\(-b\) + \@\(4\ a + b\^2\)\)\/\(2\ a\))\))\) || \
\((a \[Equal] 0 && b \[NotEqual] 0 && x \[Equal] 1\/b)\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Thus, we recover the original solutions when ",
  Cell[BoxData[
      \(TraditionalForm\`a \[NotEqual] 0\)]],
  ", but in addition find a solution for the special case of ",
  Cell[BoxData[
      \(TraditionalForm\`a = 0\)]],
  " also.  (Actually, the solution for ",
  Cell[BoxData[
      \(TraditionalForm\`a \[Rule] 0\)]],
  " can be obtained by Taylor expansion of the general case but ",
  StyleBox["Reduce", "InlineInput",
    FontWeight->"Bold"],
  " requires less work.)"
}], "Text"],

Cell[TextData[{
  ButtonBox["Eliminate",
    ButtonStyle->"RefGuideLink"],
  " eliminates a variable or set of variables from a system of equations and \
returns a smaller system combined with logical operators.  Often this is \
useful when it is convenient to formulate a system of equations in terms of \
one or more intermediate quantities that help clarify the relationships \
between the independent variables.  Sometimes it is useful for transformation \
of variables.  For example, the following expression transforms a hyperbola \
and a line from Cartesian to polar coordinates."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Eliminate[{x\^2 - y\^2 \[Equal] a\^2, \ x\  + y\  \[Equal] b, 
        x \[Equal] r\ Cos[\[Theta]], y \[Equal] r\ Sin[\[Theta]]}, {x, 
        y}]\)], "Input"],

Cell[BoxData[
    \(b \[Equal] r\ \((Cos[\[Theta]] + Sin[\[Theta]])\) && 
      r\^2\ Cos[\[Theta]]\^2 \[Equal] 
        a\^2 + r\^2\ Sin[\[Theta]]\^2\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Solve[", "InlineInput"],
  StyleBox["eqs,vars,elims", "InlineInput"],
  StyleBox["]", "InlineInput"],
  StyleBox[" first eliminates variables in ",
    FontWeight->"Plain"],
  StyleBox["elims", "InlineInput"],
  " ",
  StyleBox["and then solves the resulting equations for ",
    FontWeight->"Plain"],
  StyleBox["vars", "InlineInput"],
  StyleBox[", thereby combining the functionality of ",
    FontWeight->"Plain"],
  StyleBox["Solve", "InlineInput"],
  StyleBox[" and ",
    FontWeight->"Plain"],
  StyleBox["Eliminate", "InlineInput"],
  StyleBox[" in a single expression.",
    FontWeight->"Plain"]
}], "Text",
  FontWeight->"Bold"],

Cell[TextData[
"Solve the preceding system of equations for {r,\[Theta]} and simplify \
assuming that {a,b,r,\[Theta]} are all positive.  "], "Exercise"],

Cell["\<\
Find the roots of LegendreP[6,x] symbolically and verify that these roots \
are, in fact, real.\
\>", "Exercise"],

Cell[TextData[{
  "A parabola is the locus of all points in a plane that are equidistant from \
a fixed point, called the ",
  StyleBox["focus",
    FontSlant->"Italic"],
  ", and a fixed line, called the ",
  StyleBox["directrix",
    FontSlant->"Italic"],
  ".  Construct the parabola whose focus is given by ",
  StyleBox["focus={a,b}",
    FontSlant->"Italic"],
  " and whose directrix is the ",
  StyleBox["x",
    FontSlant->"Italic"],
  "-axis.  Then determine the coordinates of the ",
  StyleBox["vertex",
    FontSlant->"Italic"],
  ", which is the point closest to the directrix."
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["High-order polynomial equations", "Subsection"],

Cell[TextData[{
  StyleBox["Solve",
    FontWeight->"Bold"],
  " produces symbolic or exact numerical solutions to polynomial equations of \
degree 4 or less, but will also return symbolic roots to higher-order \
polynomials as well.  Suppose that we wish to determine the fixed points for \
the ",
  Cell[BoxData[
      \(TraditionalForm\`4\^th\)]],
  " iteration of the logistic map, which are defined by the following \
equation."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq = 
      x \[Equal] \((Nest[\[Mu]\ #\ \((1 - #)\) &, x, 4] // 
            Expand)\)\)], "Input"],

Cell[BoxData[
    \(x \[Equal] 
      x\ \[Mu]\^4 - x\^2\ \[Mu]\^4 - x\^2\ \[Mu]\^5 + 2\ x\^3\ \[Mu]\^5 - 
        x\^4\ \[Mu]\^5 - x\^2\ \[Mu]\^6 + 2\ x\^3\ \[Mu]\^6 - 
        x\^4\ \[Mu]\^6 - x\^2\ \[Mu]\^7 + 4\ x\^3\ \[Mu]\^7 - 
        7\ x\^4\ \[Mu]\^7 + 6\ x\^5\ \[Mu]\^7 - 2\ x\^6\ \[Mu]\^7 + 
        2\ x\^3\ \[Mu]\^8 - 7\ x\^4\ \[Mu]\^8 + 10\ x\^5\ \[Mu]\^8 - 
        8\ x\^6\ \[Mu]\^8 + 4\ x\^7\ \[Mu]\^8 - x\^8\ \[Mu]\^8 + 
        2\ x\^3\ \[Mu]\^9 - 7\ x\^4\ \[Mu]\^9 + 10\ x\^5\ \[Mu]\^9 - 
        8\ x\^6\ \[Mu]\^9 + 4\ x\^7\ \[Mu]\^9 - x\^8\ \[Mu]\^9 - 
        6\ x\^4\ \[Mu]\^10 + 24\ x\^5\ \[Mu]\^10 - 36\ x\^6\ \[Mu]\^10 + 
        24\ x\^7\ \[Mu]\^10 - 6\ x\^8\ \[Mu]\^10 - x\^4\ \[Mu]\^11 + 
        10\ x\^5\ \[Mu]\^11 - 36\ x\^6\ \[Mu]\^11 + 64\ x\^7\ \[Mu]\^11 - 
        61\ x\^8\ \[Mu]\^11 + 30\ x\^9\ \[Mu]\^11 - 6\ x\^10\ \[Mu]\^11 + 
        4\ x\^5\ \[Mu]\^12 - 22\ x\^6\ \[Mu]\^12 + 52\ x\^7\ \[Mu]\^12 - 
        70\ x\^8\ \[Mu]\^12 + 60\ x\^9\ \[Mu]\^12 - 34\ x\^10\ \[Mu]\^12 + 
        12\ x\^11\ \[Mu]\^12 - 2\ x\^12\ \[Mu]\^12 - 6\ x\^6\ \[Mu]\^13 + 
        36\ x\^7\ \[Mu]\^13 - 90\ x\^8\ \[Mu]\^13 + 120\ x\^9\ \[Mu]\^13 - 
        90\ x\^10\ \[Mu]\^13 + 36\ x\^11\ \[Mu]\^13 - 6\ x\^12\ \[Mu]\^13 + 
        4\ x\^7\ \[Mu]\^14 - 28\ x\^8\ \[Mu]\^14 + 84\ x\^9\ \[Mu]\^14 - 
        140\ x\^10\ \[Mu]\^14 + 140\ x\^11\ \[Mu]\^14 - 
        84\ x\^12\ \[Mu]\^14 + 28\ x\^13\ \[Mu]\^14 - 4\ x\^14\ \[Mu]\^14 - 
        x\^8\ \[Mu]\^15 + 8\ x\^9\ \[Mu]\^15 - 28\ x\^10\ \[Mu]\^15 + 
        56\ x\^11\ \[Mu]\^15 - 70\ x\^12\ \[Mu]\^15 + 56\ x\^13\ \[Mu]\^15 - 
        28\ x\^14\ \[Mu]\^15 + 8\ x\^15\ \[Mu]\^15 - 
        x\^16\ \[Mu]\^15\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Although there exist no general methods for solving equations of order 16, \
",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  " will nonetheless return ",
  ButtonBox["Root",
    ButtonStyle->"RefGuideLink"],
  " objects which can be evaluated numerically.  (Open to view)"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(fixedpoints = Solve[eq, x]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] 
          0}, {x \[Rule] \(\(-1\) + \[Mu]\)\/\[Mu]}, {x \[Rule] \(\[Mu] + \
\[Mu]\^2 - \[Mu]\ \@\(\(-3\) - 2\ \[Mu] + \[Mu]\^2\)\)\/\(2\ \[Mu]\^2\)}, {x \
\[Rule] \(\[Mu] + \[Mu]\^2 + \[Mu]\ \@\(\(-3\) - 2\ \[Mu] + \[Mu]\^2\)\)\/\(2\
\ \[Mu]\^2\)}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 1]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 2]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 3]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 4]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 5]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 6]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 7]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 8]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 9]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 10]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 11]}, {x \[Rule] 
          Root[1 + \[Mu]\^2 - \[Mu]\^2\ #1 - \[Mu]\^3\ #1 - \[Mu]\^4\ #1 - \
\[Mu]\^5\ #1 + 2\ \[Mu]\^3\ #1\^2 + \[Mu]\^4\ #1\^2 + 
                4\ \[Mu]\^5\ #1\^2 + \[Mu]\^6\ #1\^2 + 
                2\ \[Mu]\^7\ #1\^2 - \[Mu]\^3\ #1\^3 - 5\ \[Mu]\^5\ #1\^3 - 
                4\ \[Mu]\^6\ #1\^3 - 5\ \[Mu]\^7\ #1\^3 - 
                4\ \[Mu]\^8\ #1\^3 - \[Mu]\^9\ #1\^3 + 2\ \[Mu]\^5\ #1\^4 + 
                6\ \[Mu]\^6\ #1\^4 + 4\ \[Mu]\^7\ #1\^4 + 
                14\ \[Mu]\^8\ #1\^4 + 5\ \[Mu]\^9\ #1\^4 + 
                3\ \[Mu]\^10\ #1\^4 - 4\ \[Mu]\^6\ #1\^5 - \[Mu]\^7\ #1\^5 - 
                18\ \[Mu]\^8\ #1\^5 - 12\ \[Mu]\^9\ #1\^5 - 
                12\ \[Mu]\^10\ #1\^5 - 
                3\ \[Mu]\^11\ #1\^5 + \[Mu]\^6\ #1\^6 + 10\ \[Mu]\^8\ #1\^6 + 
                17\ \[Mu]\^9\ #1\^6 + 18\ \[Mu]\^10\ #1\^6 + 
                15\ \[Mu]\^11\ #1\^6 + \[Mu]\^12\ #1\^6 - 
                2\ \[Mu]\^8\ #1\^7 - 14\ \[Mu]\^9\ #1\^7 - 
                12\ \[Mu]\^10\ #1\^7 - 30\ \[Mu]\^11\ #1\^7 - 
                6\ \[Mu]\^12\ #1\^7 + 6\ \[Mu]\^9\ #1\^8 + 
                3\ \[Mu]\^10\ #1\^8 + 30\ \[Mu]\^11\ #1\^8 + 
                15\ \[Mu]\^12\ #1\^8 - \[Mu]\^9\ #1\^9 - 
                15\ \[Mu]\^11\ #1\^9 - 20\ \[Mu]\^12\ #1\^9 + 
                3\ \[Mu]\^11\ #1\^10 + 15\ \[Mu]\^12\ #1\^10 - 
                6\ \[Mu]\^12\ #1\^11 + \[Mu]\^12\ #1\^12 &, 12]}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Length[fixedpoints]\)], "Input"],

Cell[BoxData[
    \(16\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The list begins with roots that can be obtained by factoring and is \
followed by ",
  StyleBox["Root", "InlineInput",
    FontWeight->"Bold"],
  " objects representing the roots of the remaining irreducible polynomial.  \
Each ",
  StyleBox["Root", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontWeight->"Bold"],
  "object has two arguments: a pure function represents the irreducible part \
of the polynomial and an index identifies a particular root.  These roots can \
be displayed for a range of \[Mu] using ",
  StyleBox["Plot", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[x /. fixedpoints], {\[Mu], 2.5, 5}];\)\)], "Input"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(\\(\[Mu] + \
\[Mu]\\^2 - \\(\\(\[Mu]\\\\ \\@\\(\\(\\(-3\\)\\) - \\(\\(2\\\\ \[Mu]\\)\\) + \
\[Mu]\\^2\\)\\)\\)\\)\\/\\(2\\\\ \[Mu]\\^2\\)\\) is not a machine-size real \
number at \\!\\(\[Mu]\\) = \\!\\(2.500000104166667`\\). \\!\\(\\*ButtonBox[\\\
\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\\\", \
ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(\\(\[Mu] + \
\[Mu]\\^2 - \\(\\(\[Mu]\\\\ \\@\\(\\(\\(-3\\)\\) - \\(\\(2\\\\ \[Mu]\\)\\) + \
\[Mu]\\^2\\)\\)\\)\\)\\/\\(2\\\\ \[Mu]\\^2\\)\\) is not a machine-size real \
number at \\!\\(\[Mu]\\) = \\!\\(2.6014174789322895`\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(Plot::"plnr"\), \(\(:\)\(\ \)\), "\<\"\\!\\(\\(\[Mu] + \
\[Mu]\\^2 - \\(\\(\[Mu]\\\\ \\@\\(\\(\\(-3\\)\\) - \\(\\(2\\\\ \[Mu]\\)\\) + \
\[Mu]\\^2\\)\\)\\)\\)\\/\\(2\\\\ \[Mu]\\^2\\)\\) is not a machine-size real \
number at \\!\\(\[Mu]\\) = \\!\\(2.7120219996484343`\\). \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Plot::plnr\\\"]\\)\"\>"}]], "Message"],

Cell[BoxData[
    RowBox[{\(General::"stop"\), \(\(:\)\(\ \)\), "\<\"Further output of \
\\!\\(Plot :: \\\"plnr\\\"\\) will be suppressed during this calculation. \\!\
\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", \
ButtonStyle->\\\"RefGuideLinkText\\\", ButtonFrame->None, \
ButtonData:>\\\"General::stop\\\"]\\)\"\>"}]], "Message"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
-0.928572 0.380952 0.0147151 0.590004 [
[.02381 .00222 -9 -9 ]
[.02381 .00222 9 0 ]
[.40476 .00222 -9 -9 ]
[.40476 .00222 9 0 ]
[.59524 .00222 -3 -9 ]
[.59524 .00222 3 0 ]
[.78571 .00222 -9 -9 ]
[.78571 .00222 9 0 ]
[.97619 .00222 -3 -9 ]
[.97619 .00222 3 0 ]
[.20179 .13272 -18 -4.5 ]
[.20179 .13272 0 4.5 ]
[.20179 .25072 -18 -4.5 ]
[.20179 .25072 0 4.5 ]
[.20179 .36872 -18 -4.5 ]
[.20179 .36872 0 4.5 ]
[.20179 .48672 -18 -4.5 ]
[.20179 .48672 0 4.5 ]
[.20179 .60472 -6 -4.5 ]
[.20179 .60472 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.02381 .01472 m
.02381 .02097 L
s
[(2.5)] .02381 .00222 0 1 Mshowa
.40476 .01472 m
.40476 .02097 L
s
[(3.5)] .40476 .00222 0 1 Mshowa
.59524 .01472 m
.59524 .02097 L
s
[(4)] .59524 .00222 0 1 Mshowa
.78571 .01472 m
.78571 .02097 L
s
[(4.5)] .78571 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(5)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.0619 .01472 m
.0619 .01847 L
s
.1 .01472 m
.1 .01847 L
s
.1381 .01472 m
.1381 .01847 L
s
.17619 .01472 m
.17619 .01847 L
s
.25238 .01472 m
.25238 .01847 L
s
.29048 .01472 m
.29048 .01847 L
s
.32857 .01472 m
.32857 .01847 L
s
.36667 .01472 m
.36667 .01847 L
s
.44286 .01472 m
.44286 .01847 L
s
.48095 .01472 m
.48095 .01847 L
s
.51905 .01472 m
.51905 .01847 L
s
.55714 .01472 m
.55714 .01847 L
s
.63333 .01472 m
.63333 .01847 L
s
.67143 .01472 m
.67143 .01847 L
s
.70952 .01472 m
.70952 .01847 L
s
.74762 .01472 m
.74762 .01847 L
s
.82381 .01472 m
.82381 .01847 L
s
.8619 .01472 m
.8619 .01847 L
s
.9 .01472 m
.9 .01847 L
s
.9381 .01472 m
.9381 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.21429 .13272 m
.22054 .13272 L
s
[(0.2)] .20179 .13272 1 0 Mshowa
.21429 .25072 m
.22054 .25072 L
s
[(0.4)] .20179 .25072 1 0 Mshowa
.21429 .36872 m
.22054 .36872 L
s
[(0.6)] .20179 .36872 1 0 Mshowa
.21429 .48672 m
.22054 .48672 L
s
[(0.8)] .20179 .48672 1 0 Mshowa
.21429 .60472 m
.22054 .60472 L
s
[(1)] .20179 .60472 1 0 Mshowa
.125 Mabswid
.21429 .04422 m
.21804 .04422 L
s
.21429 .07372 m
.21804 .07372 L
s
.21429 .10322 m
.21804 .10322 L
s
.21429 .16222 m
.21804 .16222 L
s
.21429 .19172 m
.21804 .19172 L
s
.21429 .22122 m
.21804 .22122 L
s
.21429 .28022 m
.21804 .28022 L
s
.21429 .30972 m
.21804 .30972 L
s
.21429 .33922 m
.21804 .33922 L
s
.21429 .39822 m
.21804 .39822 L
s
.21429 .42772 m
.21804 .42772 L
s
.21429 .45722 m
.21804 .45722 L
s
.21429 .51622 m
.21804 .51622 L
s
.21429 .54572 m
.21804 .54572 L
s
.21429 .57522 m
.21804 .57522 L
s
.25 Mabswid
.21429 0 m
.21429 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .01472 m
.02499 .01472 L
.02605 .01472 L
.02729 .01472 L
.02846 .01472 L
.02954 .01472 L
.03053 .01472 L
.03163 .01472 L
.03279 .01472 L
.03395 .01472 L
.0352 .01472 L
.03638 .01472 L
.03746 .01472 L
.03873 .01472 L
.03944 .01472 L
.04011 .01472 L
.04132 .01472 L
.04262 .01472 L
.04327 .01472 L
.04396 .01472 L
.04519 .01472 L
.04588 .01472 L
.04652 .01472 L
.04722 .01472 L
.04797 .01472 L
.04862 .01472 L
.04933 .01472 L
.05061 .01472 L
.05186 .01472 L
.05302 .01472 L
.05414 .01472 L
.05519 .01472 L
.05634 .01472 L
.05759 .01472 L
.05882 .01472 L
.06012 .01472 L
.06123 .01472 L
.06244 .01472 L
.06315 .01472 L
.06381 .01472 L
.06453 .01472 L
.06529 .01472 L
.06657 .01472 L
.06729 .01472 L
.06797 .01472 L
.06924 .01472 L
.0704 .01472 L
.07103 .01472 L
.07173 .01472 L
.073 .01472 L
Mistroke
.07376 .01472 L
.07448 .01472 L
.07513 .01472 L
.07583 .01472 L
.0766 .01472 L
.0774 .01472 L
.07809 .01472 L
.07885 .01472 L
.07956 .01472 L
.0802 .01472 L
.08095 .01472 L
.08166 .01472 L
.08291 .01472 L
.08362 .01472 L
.08427 .01472 L
.08544 .01472 L
.08669 .01472 L
.08787 .01472 L
.08896 .01472 L
.09016 .01472 L
.09148 .01472 L
.09214 .01472 L
.09285 .01472 L
.09413 .01472 L
.09483 .01472 L
.09548 .01472 L
.09619 .01472 L
.09695 .01472 L
.09821 .01472 L
.09893 .01472 L
.09961 .01472 L
.10086 .01472 L
.10201 .01472 L
.10333 .01472 L
.10458 .01472 L
.10572 .01472 L
.10694 .01472 L
.1081 .01472 L
.10916 .01472 L
.11033 .01472 L
.11162 .01472 L
.11227 .01472 L
.11297 .01472 L
.11422 .01472 L
.11491 .01472 L
.11554 .01472 L
.11624 .01472 L
.11698 .01472 L
.11822 .01472 L
.11892 .01472 L
Mistroke
.11958 .01472 L
.12081 .01472 L
.12194 .01472 L
.12323 .01472 L
.12446 .01472 L
.12562 .01472 L
.12668 .01472 L
.12789 .01472 L
.12903 .01472 L
.13009 .01472 L
.13107 .01472 L
.13214 .01472 L
.13327 .01472 L
.13441 .01472 L
.13564 .01472 L
.1368 .01472 L
.13785 .01472 L
.13909 .01472 L
.13979 .01472 L
.14043 .01472 L
.14162 .01472 L
.14288 .01472 L
.14402 .01472 L
.14526 .01472 L
.14656 .01472 L
.14777 .01472 L
.14891 .01472 L
.14994 .01472 L
.15113 .01472 L
.15226 .01472 L
.15348 .01472 L
.15478 .01472 L
.15589 .01472 L
.15711 .01472 L
.15842 .01472 L
.15915 .01472 L
.15982 .01472 L
.16113 .01472 L
.16236 .01472 L
.16352 .01472 L
.16456 .01472 L
.16577 .01472 L
.1669 .01472 L
.16796 .01472 L
.16893 .01472 L
.16999 .01472 L
.17111 .01472 L
.17225 .01472 L
.17347 .01472 L
.17462 .01472 L
Mistroke
.17566 .01472 L
.17689 .01472 L
.17758 .01472 L
.17823 .01472 L
.1794 .01472 L
.18065 .01472 L
.18185 .01472 L
.18316 .01472 L
.18441 .01472 L
.18556 .01472 L
.1867 .01472 L
.18773 .01472 L
.18893 .01472 L
.19005 .01472 L
.19128 .01472 L
.19258 .01472 L
.19369 .01472 L
.19491 .01472 L
.19623 .01472 L
.19696 .01472 L
.19763 .01472 L
.19894 .01472 L
.20017 .01472 L
.20122 .01472 L
.20238 .01472 L
.20359 .01472 L
.20472 .01472 L
.20578 .01472 L
.20675 .01472 L
.20781 .01472 L
.20894 .01472 L
.21008 .01472 L
.21129 .01472 L
.21244 .01472 L
.21349 .01472 L
.21472 .01472 L
.21541 .01472 L
.21605 .01472 L
.21723 .01472 L
.21847 .01472 L
.21968 .01472 L
.221 .01472 L
.22224 .01472 L
.2234 .01472 L
.22454 .01472 L
.22558 .01472 L
.22677 .01472 L
.2279 .01472 L
.22914 .01472 L
.23044 .01472 L
Mistroke
.23155 .01472 L
.23277 .01472 L
.23409 .01472 L
.23482 .01472 L
.2355 .01472 L
.2368 .01472 L
.23804 .01472 L
.23909 .01472 L
.24025 .01472 L
.24146 .01472 L
.2426 .01472 L
.24365 .01472 L
.24462 .01472 L
.24569 .01472 L
.24681 .01472 L
.24795 .01472 L
.24917 .01472 L
.25032 .01472 L
.25137 .01472 L
.2526 .01472 L
.25329 .01472 L
.25393 .01472 L
.2551 .01472 L
.25635 .01472 L
.25756 .01472 L
.25888 .01472 L
.26013 .01472 L
.26129 .01472 L
.26233 .01472 L
.26348 .01472 L
.26467 .01472 L
.2658 .01472 L
.26704 .01472 L
.26835 .01472 L
.26946 .01472 L
.27068 .01472 L
.272 .01472 L
.27274 .01472 L
.27341 .01472 L
.27472 .01472 L
.27595 .01472 L
.27701 .01472 L
.27817 .01472 L
.27938 .01472 L
.28052 .01472 L
.28158 .01472 L
.28255 .01472 L
.28367 .01472 L
.28474 .01472 L
.28588 .01472 L
Mistroke
.2871 .01472 L
.28825 .01472 L
.2893 .01472 L
.29053 .01472 L
.29122 .01472 L
.29186 .01472 L
.29303 .01472 L
.29427 .01472 L
.29549 .01472 L
.29618 .01472 L
.29682 .01472 L
.29799 .01472 L
.29923 .01472 L
.30034 .01472 L
.30138 .01472 L
.30252 .01472 L
.30375 .01472 L
.30499 .01472 L
.30569 .01472 L
.30634 .01472 L
.30765 .01472 L
.30839 .01472 L
.30906 .01472 L
.31026 .01472 L
.31156 .01472 L
.31278 .01472 L
.31392 .01472 L
.31497 .01472 L
.31611 .01472 L
.31718 .01472 L
.31816 .01472 L
.3193 .01472 L
.32055 .01472 L
.32173 .01472 L
.32282 .01472 L
.32406 .01472 L
.32522 .01472 L
.32626 .01472 L
.32739 .01472 L
.32862 .01472 L
.32992 .01472 L
.33103 .01472 L
.33225 .01472 L
.33356 .01472 L
.33477 .01472 L
.33541 .01472 L
.3361 .01472 L
.33683 .01472 L
.33751 .01472 L
.33876 .01472 L
Mistroke
.33991 .01472 L
.34123 .01472 L
.34249 .01472 L
.3436 .01472 L
.3448 .01472 L
.34592 .01472 L
.34694 .01472 L
.34814 .01472 L
.34945 .01472 L
.35067 .01472 L
.35181 .01472 L
.35302 .01472 L
.35431 .01472 L
.35554 .01472 L
.35665 .01472 L
.35796 .01472 L
.3587 .01472 L
.35939 .01472 L
.36064 .01472 L
.36128 .01472 L
.36198 .01472 L
.36266 .01472 L
.36337 .01472 L
.36465 .01472 L
.36536 .01472 L
.36602 .01472 L
.36675 .01472 L
.36753 .01472 L
.3682 .01472 L
.36893 .01472 L
.36962 .01472 L
.37026 .01472 L
.37146 .01472 L
.37274 .01472 L
.37394 .01472 L
.37526 .01472 L
.37651 .01472 L
.37766 .01472 L
.3788 .01472 L
.37984 .01472 L
.38103 .01472 L
.38216 .01472 L
.38339 .01472 L
.3847 .01472 L
.38581 .01472 L
.38703 .01472 L
.38834 .01472 L
.38908 .01472 L
.38975 .01472 L
.39106 .01472 L
Mistroke
.39229 .01472 L
.39334 .01472 L
.3945 .01472 L
.39571 .01472 L
.39685 .01472 L
.39791 .01472 L
.39888 .01472 L
.39994 .01472 L
.40106 .01472 L
.40221 .01472 L
.40342 .01472 L
.40457 .01472 L
.40562 .01472 L
.40685 .01472 L
.40754 .01472 L
.40818 .01472 L
.40936 .01472 L
.4106 .01472 L
.41181 .01472 L
.41313 .01472 L
.41438 .01472 L
.41554 .01472 L
.41669 .01472 L
.41772 .01472 L
.41892 .01472 L
.42005 .01472 L
.42129 .01472 L
.42259 .01472 L
.42371 .01472 L
.42493 .01472 L
.42624 .01472 L
.42698 .01472 L
.42765 .01472 L
.42896 .01472 L
.43019 .01472 L
.43125 .01472 L
.43241 .01472 L
.43362 .01472 L
.43476 .01472 L
.43582 .01472 L
.43679 .01472 L
.43791 .01472 L
.43898 .01472 L
.44012 .01472 L
.44134 .01472 L
.44249 .01472 L
.44354 .01472 L
.44477 .01472 L
.44546 .01472 L
.4461 .01472 L
Mistroke
.44727 .01472 L
.44851 .01472 L
.44973 .01472 L
.45042 .01472 L
.45106 .01472 L
.45223 .01472 L
.45347 .01472 L
.45458 .01472 L
.45562 .01472 L
.45676 .01472 L
.45799 .01472 L
.45922 .01472 L
.45992 .01472 L
.46057 .01472 L
.46188 .01472 L
.46262 .01472 L
.46329 .01472 L
.46449 .01472 L
.46579 .01472 L
.46702 .01472 L
.46815 .01472 L
.4692 .01472 L
.47034 .01472 L
.47141 .01472 L
.47239 .01472 L
.47353 .01472 L
.47478 .01472 L
.47596 .01472 L
.47705 .01472 L
.47829 .01472 L
.47945 .01472 L
.48049 .01472 L
.48162 .01472 L
.48285 .01472 L
.48415 .01472 L
.48526 .01472 L
.48648 .01472 L
.48779 .01472 L
.489 .01472 L
.48963 .01472 L
.49033 .01472 L
.49105 .01472 L
.49174 .01472 L
.49299 .01472 L
.49413 .01472 L
.49539 .01472 L
.49671 .01472 L
.49783 .01472 L
.49902 .01472 L
.50014 .01472 L
Mistroke
.50116 .01472 L
.50236 .01472 L
.50366 .01472 L
.50489 .01472 L
.50603 .01472 L
.50723 .01472 L
.50853 .01472 L
.50975 .01472 L
.51087 .01472 L
.51218 .01472 L
.51292 .01472 L
.51361 .01472 L
.51486 .01472 L
.5155 .01472 L
.5162 .01472 L
.51687 .01472 L
.51759 .01472 L
.51886 .01472 L
.51958 .01472 L
.52024 .01472 L
.52096 .01472 L
.52174 .01472 L
.52241 .01472 L
.52314 .01472 L
.52383 .01472 L
.52447 .01472 L
.52568 .01472 L
.52695 .01472 L
.52816 .01472 L
.52947 .01472 L
.53072 .01472 L
.53187 .01472 L
.53301 .01472 L
.53405 .01472 L
.53525 .01472 L
.53637 .01472 L
.5376 .01472 L
.53891 .01472 L
.54002 .01472 L
.54124 .01472 L
.54255 .01472 L
.54329 .01472 L
.54396 .01472 L
.54526 .01472 L
.5465 .01472 L
.54755 .01472 L
.54871 .01472 L
.54992 .01472 L
.55105 .01472 L
.55211 .01472 L
Mistroke
.55308 .01472 L
.55415 .01472 L
.55527 .01472 L
.55641 .01472 L
.55763 .01472 L
.55878 .01472 L
.55982 .01472 L
.56105 .01472 L
.56175 .01472 L
.56238 .01472 L
.56356 .01472 L
.5648 .01472 L
.56601 .01472 L
.56734 .01472 L
.56858 .01472 L
.56974 .01472 L
.57088 .01472 L
.57192 .01472 L
.57312 .01472 L
.57425 .01472 L
.57548 .01472 L
.57679 .01472 L
.5779 .01472 L
.57912 .01472 L
.58044 .01472 L
.58118 .01472 L
.58185 .01472 L
.58315 .01472 L
.58439 .01472 L
.58544 .01472 L
.5866 .01472 L
.58782 .01472 L
.58895 .01472 L
.59001 .01472 L
.59098 .01472 L
.5921 .01472 L
.59317 .01472 L
.59431 .01472 L
.59553 .01472 L
.59668 .01472 L
.59773 .01472 L
.59896 .01472 L
.59965 .01472 L
.60029 .01472 L
.60146 .01472 L
.60271 .01472 L
.60392 .01472 L
.60461 .01472 L
.60525 .01472 L
.60642 .01472 L
Mistroke
.60766 .01472 L
.60876 .01472 L
.6098 .01472 L
.61094 .01472 L
.61217 .01472 L
.61341 .01472 L
.6141 .01472 L
.61475 .01472 L
.61607 .01472 L
.6168 .01472 L
.61747 .01472 L
.61867 .01472 L
.61997 .01472 L
.6212 .01472 L
.62233 .01472 L
.62339 .01472 L
.62452 .01472 L
.62559 .01472 L
.62657 .01472 L
.62771 .01472 L
.62896 .01472 L
.63014 .01472 L
.63123 .01472 L
.63247 .01472 L
.63363 .01472 L
.63467 .01472 L
.6358 .01472 L
.63703 .01472 L
.63833 .01472 L
.63944 .01472 L
.64066 .01472 L
.64197 .01472 L
.64318 .01472 L
.64381 .01472 L
.6445 .01472 L
.64523 .01472 L
.64591 .01472 L
.64716 .01472 L
.64831 .01472 L
.64956 .01472 L
.65088 .01472 L
.652 .01472 L
.65319 .01472 L
.65431 .01472 L
.65533 .01472 L
.65653 .01472 L
.65784 .01472 L
.65906 .01472 L
.6602 .01472 L
.6614 .01472 L
Mistroke
.6627 .01472 L
.66392 .01472 L
.66504 .01472 L
.66635 .01472 L
.66709 .01472 L
.66777 .01472 L
.66903 .01472 L
.66966 .01472 L
.67036 .01472 L
.67104 .01472 L
.67175 .01472 L
.67303 .01472 L
.67374 .01472 L
.6744 .01472 L
.67513 .01472 L
.67591 .01472 L
.67658 .01472 L
.67731 .01472 L
.678 .01472 L
.67864 .01472 L
.67984 .01472 L
.68112 .01472 L
.68232 .01472 L
.68364 .01472 L
.68488 .01472 L
.68603 .01472 L
.68718 .01472 L
.68821 .01472 L
.68941 .01472 L
.69053 .01472 L
.69176 .01472 L
.69307 .01472 L
.69418 .01472 L
.6954 .01472 L
.69671 .01472 L
.69745 .01472 L
.69812 .01472 L
.69942 .01472 L
.70066 .01472 L
.70171 .01472 L
.70286 .01472 L
.70407 .01472 L
.70521 .01472 L
.70627 .01472 L
.70724 .01472 L
.7083 .01472 L
.70942 .01472 L
.71056 .01472 L
.71178 .01472 L
.71293 .01472 L
Mistroke
.71398 .01472 L
.71521 .01472 L
.7159 .01472 L
.71654 .01472 L
.71771 .01472 L
.71896 .01472 L
.72017 .01472 L
.72149 .01472 L
.72273 .01472 L
.72389 .01472 L
.72503 .01472 L
.72607 .01472 L
.72727 .01472 L
.7284 .01472 L
.72963 .01472 L
.73094 .01472 L
.73205 .01472 L
.73327 .01472 L
.73458 .01472 L
.73532 .01472 L
.73599 .01472 L
.7373 .01472 L
.73853 .01472 L
.73959 .01472 L
.74075 .01472 L
.74196 .01472 L
.7431 .01472 L
.74416 .01472 L
.74513 .01472 L
.74619 .01472 L
.74731 .01472 L
.74846 .01472 L
.74967 .01472 L
.75083 .01472 L
.75187 .01472 L
.7531 .01472 L
.75379 .01472 L
.75443 .01472 L
.75561 .01472 L
.75685 .01472 L
.75806 .01472 L
.75875 .01472 L
.75939 .01472 L
.76056 .01472 L
.7618 .01472 L
.7629 .01472 L
.76394 .01472 L
.76508 .01472 L
.76631 .01472 L
.76754 .01472 L
Mistroke
.76824 .01472 L
.76889 .01472 L
.7702 .01472 L
.77094 .01472 L
.77161 .01472 L
.7728 .01472 L
.7741 .01472 L
.77533 .01472 L
.77646 .01472 L
.77752 .01472 L
.77865 .01472 L
.77972 .01472 L
.7807 .01472 L
.78184 .01472 L
.78309 .01472 L
.78427 .01472 L
.78536 .01472 L
.7866 .01472 L
.78776 .01472 L
.7888 .01472 L
.78993 .01472 L
.79116 .01472 L
.79246 .01472 L
.79357 .01472 L
.79479 .01472 L
.7961 .01472 L
.7973 .01472 L
.79793 .01472 L
.79863 .01472 L
.79936 .01472 L
.80004 .01472 L
.80129 .01472 L
.80243 .01472 L
.80369 .01472 L
.80501 .01472 L
.80612 .01472 L
.80731 .01472 L
.80843 .01472 L
.80945 .01472 L
.81066 .01472 L
.81196 .01472 L
.81318 .01472 L
.81432 .01472 L
.81552 .01472 L
.81682 .01472 L
.81804 .01472 L
.81916 .01472 L
.82047 .01472 L
.8212 .01472 L
.82189 .01472 L
Mistroke
.82315 .01472 L
.82385 .01472 L
.82448 .01472 L
.82516 .01472 L
.82587 .01472 L
.82715 .01472 L
.82786 .01472 L
.82852 .01472 L
.82925 .01472 L
.83002 .01472 L
.8307 .01472 L
.83143 .01472 L
.83212 .01472 L
.83275 .01472 L
.83396 .01472 L
.83524 .01472 L
.83644 .01472 L
.83775 .01472 L
.839 .01472 L
.84015 .01472 L
.84129 .01472 L
.84232 .01472 L
.84352 .01472 L
.84464 .01472 L
.84587 .01472 L
.84718 .01472 L
.84829 .01472 L
.8495 .01472 L
.85082 .01472 L
.85155 .01472 L
.85222 .01472 L
.85353 .01472 L
.85476 .01472 L
.85581 .01472 L
.85697 .01472 L
.85818 .01472 L
.85932 .01472 L
.86037 .01472 L
.86134 .01472 L
.86241 .01472 L
.86353 .01472 L
.86467 .01472 L
.86589 .01472 L
.86704 .01472 L
.86808 .01472 L
.86931 .01472 L
.87 .01472 L
.87064 .01472 L
.87182 .01472 L
.87306 .01472 L
Mistroke
.87427 .01472 L
.87559 .01472 L
.87684 .01472 L
.87799 .01472 L
.87914 .01472 L
.88017 .01472 L
.88137 .01472 L
.88249 .01472 L
.88373 .01472 L
.88503 .01472 L
.88615 .01472 L
.88737 .01472 L
.88868 .01472 L
.88942 .01472 L
.89009 .01472 L
.89139 .01472 L
.89263 .01472 L
.89368 .01472 L
.89484 .01472 L
.89605 .01472 L
.89719 .01472 L
.89825 .01472 L
.89922 .01472 L
.90029 .01472 L
.90141 .01472 L
.90255 .01472 L
.90377 .01472 L
.90492 .01472 L
.90596 .01472 L
.90719 .01472 L
.90789 .01472 L
.90852 .01472 L
.9097 .01472 L
.91094 .01472 L
.91216 .01472 L
.91284 .01472 L
.91348 .01472 L
.91464 .01472 L
.91589 .01472 L
.91699 .01472 L
.91803 .01472 L
.91917 .01472 L
.9204 .01472 L
.92163 .01472 L
.92233 .01472 L
.92297 .01472 L
.92429 .01472 L
.92502 .01472 L
.92569 .01472 L
.92689 .01472 L
Mistroke
.92819 .01472 L
.92941 .01472 L
.93055 .01472 L
.9316 .01472 L
.93273 .01472 L
.9338 .01472 L
.93478 .01472 L
.93592 .01472 L
.93717 .01472 L
.93835 .01472 L
.93944 .01472 L
.94068 .01472 L
.94184 .01472 L
.94288 .01472 L
.94401 .01472 L
.94524 .01472 L
.94654 .01472 L
.94765 .01472 L
.94887 .01472 L
.94975 .01472 L
.95057 .01472 L
.95146 .01472 L
.95241 .01472 L
.95324 .01472 L
.95401 .01472 L
.95486 .01472 L
.95577 .01472 L
.95653 .01472 L
.95737 .01472 L
.95824 .01472 L
.95905 .01472 L
.95981 .01472 L
.9605 .01472 L
.9613 .01472 L
.96206 .01472 L
.96287 .01472 L
.96377 .01472 L
.96471 .01472 L
.96559 .01472 L
.96641 .01472 L
.96716 .01472 L
.96802 .01472 L
.96884 .01472 L
.96972 .01472 L
.97067 .01472 L
.97155 .01472 L
.97236 .01472 L
.97331 .01472 L
.97433 .01472 L
.9753 .01472 L
Mistroke
.97619 .01472 L
Mfstroke
.02381 .36872 m
.06244 .37792 L
.10458 .38717 L
.14415 .39519 L
.18221 .40237 L
.22272 .40949 L
.26171 .41589 L
.30316 .42224 L
.34309 .42797 L
.3815 .43315 L
.42237 .43834 L
.46172 .44305 L
.49955 .44734 L
.53984 .45165 L
.57861 .45559 L
.61984 .45956 L
.65954 .46319 L
.69774 .46651 L
.73838 .46988 L
.77751 .47298 L
.81909 .47611 L
.85916 .47899 L
.89771 .48165 L
.93871 .48435 L
.97619 .48672 L
s
.21447 .40366 m
.21568 .39605 L
.2167 .39219 L
.21784 .3888 L
.22272 .37821 L
.22778 .37016 L
.23242 .36403 L
.24285 .35262 L
.26171 .33648 L
.30142 .31123 L
.33961 .29258 L
.38026 .27622 L
.41939 .26277 L
.46097 .25029 L
.50104 .23964 L
.53959 .23042 L
.58059 .22153 L
.62008 .21372 L
.65805 .20681 L
.69847 .20001 L
.73738 .19394 L
.77874 .18795 L
.81859 .18256 L
.85692 .17771 L
.8977 .17286 L
.93697 .16847 L
.97472 .16449 L
.97619 .16433 L
s
.21447 .41241 m
.21568 .41981 L
.2167 .42349 L
.21784 .42669 L
.22272 .43645 L
.22778 .44365 L
.23242 .449 L
.24285 .45868 L
.26171 .47179 L
.2807 .48193 L
.30142 .49094 L
.34206 .50482 L
.38119 .51517 L
.42278 .52407 L
.46284 .53121 L
.50139 .53706 L
.5424 .54245 L
.58188 .54697 L
.61986 .55082 L
.66028 .55448 L
.69919 .55762 L
.74055 .56063 L
.78039 .56324 L
.81872 .56552 L
.85951 .56774 L
.89877 .5697 L
.93652 .57143 L
.97619 .5731 L
s
.38575 .27048 m
.38634 .2671 L
.38764 .26276 L
.38837 .26095 L
.38903 .2595 L
.39154 .25504 L
.39638 .24865 L
.40095 .24395 L
.41124 .23563 L
.42237 .22859 L
.46172 .21094 L
.48166 .20426 L
.50272 .19816 L
.54069 .18891 L
.5602 .1848 L
.56582 .18368 L
.57114 .18265 L
.57592 .18174 L
.57723 .18149 L
.57862 .18123 L
.57981 .18101 L
.58112 .03708 L
.58233 .0367 L
.58346 .03642 L
.58603 .03592 L
.59121 .03513 L
.60046 .03407 L
.62132 .0323 L
.66036 .02995 L
.70186 .02809 L
.74184 .02667 L
.78031 .02552 L
.82123 .02449 L
.86063 .02364 L
.89852 .02292 L
.93886 .02225 L
.97619 .0217 L
s
.38575 .27814 m
.38634 .28158 L
.38764 .28608 L
.38903 .2895 L
.39154 .29425 L
.39647 .30132 L
.40095 .30643 L
.42237 .32419 L
.44263 .33644 L
.46172 .34599 L
.50027 .36183 L
.51981 .36868 L
.54127 .37553 L
.55113 .37848 L
.56169 .38152 L
.56681 .38295 L
.57167 .38428 L
.57602 .38546 L
.57721 .38578 L
.57848 .38612 L
.57956 .3864 L
.58076 .03848 L
.58197 .03885 L
.58311 .03907 L
.58567 .03944 L
.58818 .03971 L
.59085 .03993 L
.5957 .04025 L
.6001 .04046 L
.60525 .04066 L
.61083 .04083 L
.6161 .04095 L
.62096 .04104 L
.62563 .04111 L
.63066 .04116 L
.63539 .0412 L
.63972 .04123 L
.64235 .04124 L
.64479 .04125 L
.64617 .04125 L
.64746 .04126 L
.64818 .04126 L
.64896 .04126 L
.64966 .04126 L
.65031 .04126 L
.65153 .04126 L
.65217 .04126 L
.65286 .04126 L
.65358 .04126 L
.65427 .04126 L
Mistroke
.65555 .04126 L
.65675 .04126 L
.65786 .04126 L
.66036 .04125 L
.66279 .04125 L
.66501 .04124 L
.67002 .04122 L
.67907 .04117 L
.68912 .04109 L
.69968 .04099 L
.73784 .04052 L
.77846 .0399 L
.81755 .03923 L
.8591 .03848 L
.89914 .03775 L
.93766 .03704 L
.97619 .03634 L
Mfstroke
.38575 .5146 m
.38634 .51321 L
.38764 .51143 L
.38837 .51069 L
.38903 .5101 L
.39154 .50829 L
.39638 .50574 L
.40095 .50391 L
.41124 .50076 L
.42237 .49823 L
.44286 .49486 L
.46172 .49267 L
.48054 .49104 L
.5012 .48973 L
.52071 .48884 L
.53165 .48846 L
.54163 .48818 L
.55088 .48797 L
.56071 .48779 L
.56584 .48771 L
.5715 .48765 L
.57397 .48762 L
.57664 .48759 L
.57787 .48758 L
.57916 .48757 L
.58027 .10156 L
.58147 .09955 L
.58279 .09825 L
.58401 .09729 L
.58676 .09556 L
.59175 .09312 L
.60108 .08965 L
.62203 .08397 L
.66144 .07643 L
.69932 .07099 L
.73966 .06628 L
.77849 .06248 L
.81977 .059 L
.85953 .05608 L
.89777 .05358 L
.93847 .05121 L
.97619 .04922 L
s
.38575 .51778 m
.38634 .5192 L
.38764 .52107 L
.38903 .52248 L
.39154 .52444 L
.39647 .52734 L
.40095 .52943 L
.42237 .53653 L
.44263 .54127 L
.46172 .54488 L
.50027 .55068 L
.51981 .55311 L
.54127 .5555 L
.55113 .55651 L
.56169 .55755 L
.56681 .55803 L
.57167 .55848 L
.57602 .55887 L
.57721 .55898 L
.57848 .55909 L
.57956 .55919 L
.58076 .1051 L
.58197 .10649 L
.58311 .10738 L
.58567 .10889 L
.58818 .11001 L
.59085 .111 L
.6001 .11352 L
.60573 .11467 L
.61103 .11559 L
.62096 .117 L
.63068 .11811 L
.64097 .11907 L
.65943 .12037 L
.66886 .12088 L
.679 .12134 L
.69731 .12196 L
.70745 .12222 L
.71849 .12243 L
.72892 .12258 L
.73859 .12268 L
.74368 .12272 L
.74636 .12273 L
.74919 .12275 L
.75166 .12276 L
.75438 .12277 L
.75557 .12277 L
.75685 .12277 L
.75797 .12277 L
.75915 .12278 L
Mistroke
.7604 .12278 L
.76155 .12278 L
.76279 .12278 L
.7635 .12278 L
.76416 .12278 L
.76535 .12278 L
.76644 .12278 L
.76764 .12277 L
.7689 .12277 L
.7711 .12277 L
.77314 .12276 L
.77777 .12275 L
.78284 .12272 L
.78832 .12269 L
.79827 .12261 L
.81687 .12241 L
.83679 .12213 L
.85784 .12177 L
.89578 .12097 L
.93617 .11998 L
.97505 .11891 L
.97619 .11888 L
Mfstroke
.58055 .18087 m
.58121 .18075 L
.58356 .18031 L
.58883 .17936 L
.59832 .17769 L
.61984 .1741 L
.65954 .16812 L
.70055 .16265 L
.74003 .15791 L
.77801 .15375 L
.81843 .14968 L
.85734 .14607 L
.8987 .14249 L
.93855 .13928 L
.97619 .13643 L
s
.58055 .30625 m
.58121 .30353 L
.58243 .30004 L
.58356 .29757 L
.58883 .28944 L
.59345 .28426 L
.59832 .2797 L
.61984 .26478 L
.63877 .255 L
.65954 .24611 L
.69997 .23204 L
.73888 .22101 L
.78024 .21103 L
.82008 .20264 L
.85841 .19545 L
.89919 .18854 L
.93846 .1825 L
.97619 .17716 L
s
.58055 .31696 m
.58121 .31968 L
.58243 .32316 L
.58356 .32561 L
.58883 .33368 L
.59345 .33881 L
.59832 .34331 L
.61984 .35801 L
.63877 .36761 L
.65954 .37631 L
.69997 .39005 L
.73888 .40082 L
.78024 .41055 L
.82008 .41872 L
.85841 .42573 L
.89919 .43247 L
.93846 .43837 L
.97619 .44358 L
s
.58055 .38667 m
.58121 .38684 L
.58356 .38746 L
.58883 .38882 L
.59832 .39122 L
.61984 .3964 L
.65954 .40514 L
.70055 .41324 L
.74003 .42031 L
.77801 .42655 L
.81843 .43266 L
.85734 .43811 L
.8987 .44348 L
.93855 .4483 L
.97619 .45257 L
s
.58055 .48756 m
.58121 .48756 L
.58356 .48754 L
.58612 .48753 L
.58883 .48752 L
.59119 .48751 L
.59377 .4875 L
.59491 .4875 L
.59613 .48749 L
.59727 .48749 L
.59832 .48749 L
.59895 .48749 L
.59965 .48749 L
.60038 .48749 L
.60107 .48749 L
.60229 .48749 L
.60361 .48749 L
.60431 .48749 L
.60508 .48749 L
.60662 .48749 L
.60804 .4875 L
.60937 .4875 L
.61443 .48752 L
.61726 .48753 L
.61984 .48754 L
.62964 .48761 L
.64026 .48771 L
.65954 .48796 L
.67942 .48829 L
.70046 .48872 L
.73835 .48965 L
.77869 .49081 L
.81751 .49205 L
.85879 .49345 L
.89855 .49488 L
.9368 .49628 L
.97619 .49776 L
s
.58055 .55928 m
.58121 .55933 L
.58356 .55954 L
.58883 .55999 L
.59832 .56079 L
.61984 .56248 L
.65954 .56528 L
.70055 .56782 L
.74003 .56998 L
.77801 .57185 L
.81843 .57365 L
.85734 .57523 L
.8987 .57675 L
.93855 .5781 L
.97619 .57928 L
s
.58055 .59868 m
.58121 .59862 L
.58243 .59855 L
.58356 .5985 L
.58612 .59842 L
.58741 .59839 L
.58883 .59836 L
.5912 .59832 L
.59374 .59829 L
.59832 .59824 L
.60109 .59823 L
.6041 .59821 L
.60694 .5982 L
.60957 .59819 L
.61194 .59818 L
.61329 .59818 L
.61455 .59818 L
.61588 .59818 L
.61709 .59818 L
.61773 .59818 L
.61843 .59818 L
.61916 .59818 L
.61984 .59818 L
.62113 .59818 L
.62232 .59818 L
.62363 .59818 L
.62428 .59818 L
.62501 .59818 L
.62753 .59818 L
.62989 .59818 L
.6343 .59819 L
.63903 .5982 L
.64975 .59824 L
.65954 .59828 L
.6993 .5985 L
.73755 .59876 L
.77825 .59904 L
.81743 .59932 L
.85906 .59961 L
.89918 .59988 L
.93779 .60013 L
.97619 .60036 L
s
.58055 .59895 m
.58121 .59903 L
.58243 .59913 L
.58356 .5992 L
.58883 .59946 L
.59832 .59979 L
.61984 .60033 L
.63914 .6007 L
.65954 .60103 L
.69752 .60153 L
.73794 .60195 L
.77685 .60227 L
.81821 .60256 L
.85806 .6028 L
.89638 .60299 L
.93717 .60317 L
.97619 .60332 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{314, 194},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40007G00018b000`400?l00000o`00003oo`3ooooH
0?ooo`00o`3ooooH0?ooo`002@3oool00`3ookl0M000000000040000000307Boo`3oool0oooo0080
oooo00<0_g@000000000M;l01@3oool01@2oM0000000000000000000M;oo09@0oooo00D0_g@00000
00000000000004RLg`050?ooo`030;md0000000007Bo00D0oooo00D0_g@0000000000000000007Bo
o`180?ooo`050?oOW018000000000000001d_ol0A`3oool01@3ogi`0B000000000000000M;oo00@0
oooo00<0_g@000000000M;l01@3oool01@2oM0000000000000000000M;oo04P0oooo00D0_g@00000
00000000000007Boo`0C0?ooo`002P3oool00`2oM000001d0;ooo`020?ooo`030=nLM02Lgol0oooo
00<0oooo00<0_g@000000000M;l0103oool01`2oM00007Bo0?ooo`3oool0omnL04P0001d_ol0TP3o
ool01`2oM00007Bo0?ooo`3oool0oono07@00018W=l0103oool00`2oM0000000001d_`040?ooo`07
0;md0000M;l0oooo0?ooo`3ogi`0B00007Boo`190?ooo`030;md001d_ol0oooo04T0oooo00<0_g@0
07Boo`3oool0103oool00`2oM0000000001d_`040?ooo`070;md0000M;l0oooo0?ooo`3ogi`0B000
07Boo`160?ooo`070;md0000M;l0oooo0?ooo`3ogi`0B00007Boo`0B0?ooo`002`3oool00`3o_g@0
001809cOo`0C0?ooo`030?noM000M;l0oooo09H0oooo00<0okmd001d_`3oool03`3oool00`3o_g@0
07Bo0?ooo`180?ooo`030;md001d_ol0oooo04T0oooo00<0_g@007Boo`3oool0403oool00`3o_g@0
07Bo0?ooo`1:0?ooo`030?noM000M;l0oooo0140oooo000<0?ooo`030?oOW0180000B9cO0180oooo
00<0okmd001d_`3oool0U@3oool00`3ogi`0B01809cOo`0@0?ooo`030?noM000M;l0oooo04@0oooo
00<0gia80000000000000`0000000`1d_ol0oooo0?ooo`130?ooo`030=nLB0000000000000<00000
00<0M;oo0?ooo`3oool03P3oool00`3o_g@007Bo0?ooo`1:0?ooo`030?noM000M;l0oooo0140oooo
000>0?ooo`0309a80018W=l0oooo0100oooo00<0W4P007Boo`3oool0T`3oool0103OW4P000000000
B02LgolA0?ooo`0309a8001d_ol0oooo04@0oooo00H0okmd0000M02oool0oooo0;md001d_om60?oo
o`060?noM00007@0_ooo0?ooo`2oM000M;oo4@3oool00`2LB000M;oo0?ooo`1:0?ooo`0309a8001d
_ol0oooo0140oooo000?0?ooo`0309a8M02oool0oooo00X0oooo00<0oono07@0000000000P000000
0`18W=l0oooo0?ooo`2E0?ooo`030?oOW01804P0W=oo00/0oooo00<0oono07@0000000000P000000
0`18W=l0oooo0?ooo`150?ooo`050=nLB000M;l0oooo0;md001d_ol0A`3oool01@3OW4P007Bo0?oo
o`2oM000M;oo00`0oooo00<0oono07@0000000000P0000000`18W=l0oooo0?ooo`150?ooo`030?oo
_`1d000000000080000000<0B9cO0?ooo`3oool04@3oool000X0oooo00<0gia807Boo`3oool00P3o
ool00`2oM000M;oo0?ooo`0:0?ooo`030?oo_`1d07@0_ooo09/0oooo00<0_g@007Boo`3oool02P3o
ool00`3ookl0M01d0;ooo`1;0?ooo`0409a80018W=l0_g@007BoodP0oooo00@0W4P004RLg`2oM000
M;oo303oool00`3ookl0M01d0;ooo`1:0?ooo`030?oo_`1d07@0_ooo01H0oooo000:0?ooo`070?no
M00007@0_ooo0?ooo`3o_g@0001809cOo`0;0?ooo`030?oo_`1d07@0_ooo09H0oooo00L0okmd0000
B02Lgol0oooo0?oOW01804P0W=oo00/0oooo00<0oono07@0M02oool0B`3oool0103ogi`0B01804Q8
001d_om80?ooo`040?oOW01804P0B4P007Boo``0oooo00<0oono07@0M02oool0BP3oool00`3ookl0
M01d0;ooo`0F0?ooo`002`3oool01@3o_g@000000000000004P0W=oo00`0oooo00<0oono07@00000
00000P0000000`0007@0_ooo0?ooo`2B0?ooo`050?oOW018000000000000001d_ol0303oool00`3o
okl0M00000000002000000030000M02oool0oooo04L0oooo00<0okmd0000001d_ol0B@3oool00`3o
_g@0000007Boo`0<0?ooo`030?oo_`1d000000000080000000<0001d0;ooo`3oool0A@3oool00`3o
okl0M00000000002000000030000M02oool0oooo0140oooo003o0?ooomP0oooo003o0?ooomP0oooo
001/0?ooo`030000003oool0oooo0?l0ooooJ@3oool006`0oooo00<000000?ooo`3oool0o`3ooomY
0?ooo`00K03oool00`000000oooo0?ooo`3o0?ooofT0oooo001/0?ooo`030000003oool0oooo0?l0
ooooJ@3oool006`0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`00K03oool00`000000oooo0?oo
o`3o0?ooofT0oooo001/0?ooo`030000003oool0oooo0?l0ooooJ@3oool000`0ooooo`0000320000
00X0oooo000G0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03P3oool00`000000
oooo0?ooo`0>0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool04`3oool001L0oooo
00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo05<0oooo
00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo01<0oooo
001/0?ooo`030000003oool0oooo0?l0oooo:03oool/000001D0oooo001/0?ooo`030000003oool0
oooo0?80oooo=@0000110?ooo`00K03oool00`000000oooo0?ooo`3Q0?oooa400000MP3oool006`0
oooo00<000000?ooo`3oool0_@3ooolT000008L0oooo001/0?ooo`030000003oool0oooo0:l0oooo
3P00002[0?ooo`00K03oool00`000000oooo0?ooo`2Y0?ooo`H00000^@3oool006`0oooo00<00000
0?ooo`3oool0XP3oool700000:40oooo2@00000E0?ooo`00K03oool00`000000oooo0?ooo`2Q0?oo
o`800000LP3ooole000001h0oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`00
0000L@00001C0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0<@0oooo
001/0?ooo`<00000X@3oool00`000000oooo0?ooo`340?ooo`00K03oool00`000000oooo0?ooo`2Q
0?ooo`030000003oool0oooo0<@0oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?oo
o`3oool0YP3oool9000001D0oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0U@3ooolA000001h0oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0
P`3ooolB000002l0oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0LP3o
oolA00000440oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0I03oool>
00000580oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0F`3oool90000
0600oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0DP3oool9000006T0
oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0B@3oool900000780oooo
001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0@03oool9000007/0oooo001/
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0=`3oool9000008@0oooo001/0?oo
o`030000003oool0oooo0:40oooo00<000000?ooo`3oool0;P3oool9000008d0oooo001/0?ooo`<0
0000X@3oool00`000000oooo0?ooo`0V0?ooo`P00000UP3oool006`0oooo00<000000?ooo`3oool0
X@3oool00`000000oooo0?ooo`0O0?ooo`L00000WP3oool006`0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`0K0?ooo`@00000Y@3oool006`0oooo00<000000?ooo`3oool0X@3oool0
0`000000oooo0?ooo`0G0?ooo`@00000Z@3oool006`0oooo00<000000?ooo`3oool0X@3oool00`00
0000oooo0?ooo`0C0?ooo`@00000[@3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000
oooo0?ooo`0>0?ooo`D00000/@3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`090?ooo`D00000]P3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?oo
o`060?ooo`<00000^`3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`04
0?ooo`800000_P3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`020?oo
o`800000`03oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0200000<80
oooo001/0?ooo`030000003oool0oooo0:40oooo0`0000340?ooo`00K03oool00`000000oooo0?oo
o`2Q0?ooo`030000003oool0oooo0<@0oooo001/0?ooo`<00000X@3oool00`000000oooo0?ooo`34
0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0<@0oooo001/0?ooo`03
0000003oool0oooo0:40oooo0P0000350?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`040000
003oool0000000000<<0oooo001/0?ooo`030000003oool0oooo0:40oooo00@000000?ooo`3oool0
oooo0`0000300?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00@0oooo
1P00002j0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00X0oooo1`00
002c0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0140oooo2`00002:
0?ooo`T000005@3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`0L0?oo
oa400000E@3ooolT000001h0oooo001/0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3o
ool0;@3ooomE00000480oooo001=0?ooo`050?noM000000000000000001d_ol01@3oool00`2oM000
0000001d_`030?ooo`030?oo_`1d0000000000@0000000<0M;oo0?ooo`3oool01@3oool00`000000
oooo0?ooo`2Q0?ooo`030000003oool0oooo0<@0oooo001=0?ooo`0609a8M02oool0oooo0?oo_`1d
07@0_ooo103oool00`2oM0000000001d_`040?ooo`030;md000007@0_ooo0080oooo00<0giad09cO
o`3oool01P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0<@0oooo001<0?ooo`03
0?noM000M;l0oooo0080oooo00<0gia804RLg`3oool02`3oool00`3o_g@0001809cOo`0:0?ooo`03
0000003oool0oooo0:40oooo00<000000?ooo`3oool0a03oool004`0oooo00<0gia804RLg`3oool0
0P3oool00`3o_g@007Bo0?ooo`0<0?ooo`030?oOW0180000B9cO00T0oooo1000002P0?ooo`030000
003oool0oooo0<@0oooo001<0?ooo`030=nLB018W=l0oooo0080oooo00<0okmd001d_`3oool03P3o
ool00`2LB000B9cO0?ooo`070?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool0a03o
ool004`0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0?0?ooo`0309a8M02oool0
oooo00H0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`2V0?ooo`T000005@3oool0
04`0oooo00<0okmd04RLg`3oool00P3oool00`3OW4P0B9cO0?ooo`0:0?ooo`030=nLB01d_ol0oooo
0080oooo00<0_g@007Boo`3oool01P3oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo
09T0oooo3@00000N0?ooo`00C@3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;ooo`/0oooo00L0okmd
0000M02oool0oooo0?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0X@3oool00`000000oooo
0?ooo`2@0?ooo`T00000:`3oool004d0oooo00D0oono07@000000000001809cOo`0=0?ooo`050?no
M000000000000000B02Lgol0203oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo08<0
oooo3@00000d0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo07D0oooo
3P0000110?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo06`0oooo2@00
001?0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo06<0oooo2@00001H
0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo05X0oooo2@00001Q0?oo
o`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0540oooo2@00001Z0?ooo`00
K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo04T0oooo2000001c0?ooo`00K03o
ool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo0400oooo2@00001k0?ooo`00K03oool0
0`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo03P0oooo200000240?ooo`00K03oool30000
0:40oooo00<000000?ooo`3oool0;`3oool9000008`0oooo001/0?ooo`030000003oool0oooo0:40
oooo00<000000?ooo`3oool09@3oool:000007L0oooo2@00000E0?ooo`00K03oool00`000000oooo
0?ooo`2Q0?ooo`030000003oool0oooo01h0oooo1`00001d0?ooo`d000007P3oool006`0oooo00<0
00000?ooo`3oool0X@3oool00`000000oooo0?ooo`0H0?ooo`H00000LP3oool9000002/0oooo001/
0?ooo`030000003oool0oooo0:40oooo00<000000?ooo`3oool04P3oool600000740oooo1`00000d
0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000003oool0oooo00d0oooo1@00001a0?oo
o`H00000>`3oool006`0oooo00<000000?ooo`3oool0X@3oool00`000000oooo0?ooo`080?ooo`D0
0000L03oool6000002T0oooo0`00000E0?ooo`00K03oool00`000000oooo0?ooo`2Q0?ooo`030000
003oool0oooo00P00000K@3oool8000002T0oooo1P00000H0?ooo`00K03oool00`000000oooo0?oo
o`2O0?ooo`D00000K@3oool8000002/0oooo1P00000N0?ooo`00K03oool00`000000oooo0?ooo`2J
0?ooo`D000000P3oool00`000000oooo0?ooo`1T0?ooo`T00000;03oool7000002@0oooo001/0?oo
o`030000003oool0oooo09H0oooo100000070?ooo`030000003oool0oooo05/0oooo2@00000]0?oo
o`P00000:`3oool006`0oooo00<000000?ooo`3oool0TP3oool4000000/0oooo00<000000?ooo`3o
ool0D`3oool8000002l0oooo1`00000c0?ooo`00K03oool00`000000oooo0?ooo`2=0?ooo`D00000
3`3oool00`000000oooo0?ooo`1=0?ooo`H00000<P3oool5000003X0oooo001/0?ooo`<00000R@3o
ool4000001@0oooo00<000000?ooo`3oool0A`3oool6000003@0oooo1000000o0?ooo`00K03oool0
0`000000oooo0?ooo`250?ooo`@00000603oool00`000000oooo0?ooo`110?ooo`H00000=@3oool5
000004<0oooo001/0?ooo`030000003oool0oooo0840oooo1000000L0?ooo`030000003oool0oooo
03/0oooo1P00000f0?ooo`D00000B03oool006`0oooo00<000000?ooo`3oool0O03oool500000200
oooo00<000000?ooo`3oool0=@3oool6000003H0oooo1P00001=0?ooo`00K03oool00`000000oooo
0?ooo`1h0?ooo`@000009@3oool00`000000oooo0?ooo`0`0?ooo`D00000=P3oool6000005<0oooo
001/0?ooo`030000003oool0oooo07@0oooo1000000Y0?ooo`030000003oool0oooo02X0oooo1P00
000f0?ooo`D00000F@3oool006`0oooo00<000000?ooo`3oool0L@3oool3000002d0oooo00<00000
0?ooo`3oool0903oool6000003H0oooo1P00001N0?ooo`00K03oool00`000000oooo0?ooo`1^0?oo
o`<00000<03oool00`000000oooo0?ooo`0N0?ooo`H00000=P3oool6000006@0oooo001/0?ooo`03
0000003oool0oooo06/0oooo0`00000c0?ooo`030000003oool0oooo01P0oooo1P00000g0?ooo`D0
0000JP3oool006`0oooo00<000000?ooo`3oool0J@3oool2000003H0oooo00<000000?ooo`3oool0
4P3oool6000003T0oooo1000001_0?ooo`00K03oool00`000000oooo0?ooo`1W0?ooo`800000>03o
ool00`000000oooo0?ooo`0=0?ooo`D00000?03oool3000007<0oooo001/0?ooo`030000003oool0
oooo06D0oooo0P00000j0?ooo`030000003oool0oooo00L0oooo1P00000m0?ooo`@00000MP3oool0
06`0oooo00<000000?ooo`3oool0HP3oool3000003`0oooo00@000000?ooo`3oool0oooo1P00000o
0?ooo`@00000NP3oool006`0oooo0`00001P0?ooo`800000?@3oool600000480oooo0`00001n0?oo
o`00K03oool00`000000oooo0?ooo`1N0?ooo`800000>`3oool400000080oooo00<000000?ooo`3o
ool0?`3oool400000840oooo001/0?ooo`030000003oool0oooo05`0oooo0P00000h0?ooo`D00000
1P3oool00`000000oooo0?ooo`0l0?ooo`<00000Q@3oool006`0oooo00<000000?ooo`3oool0FP3o
ool2000003H0oooo1000000;0?ooo`030000003oool0oooo03P0oooo100000280?ooo`00K03oool0
0`000000oooo0?ooo`1I0?ooo`030000003oool0oooo0340oooo1@00000?0?ooo`030000003oool0
oooo03@0oooo1000002<0?ooo`00K03oool00`000000oooo0?ooo`1G0?ooo`800000<03oool40000
01@0oooo00<000000?ooo`3oool0<@3oool300000900oooo001/0?ooo`030000003oool0oooo05H0
oooo00<000000?ooo`3oool0:`3oool5000001P0oooo00<000000?ooo`3oool0;P3oool3000009<0
oooo001/0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0:03oool4000001d0oooo
00<000000?ooo`3oool0:`3oool3000009H0oooo001/0?ooo`030000003oool0oooo05@0oooo00<0
00000?ooo`3oool09@3oool400000240oooo00<000000?ooo`3oool0:03oool3000009T0oooo001/
0?ooo`030000003oool0oooo0580oooo0P00000T0?ooo`@000009@3oool00`000000oooo0?ooo`0U
0?ooo`<00000W03oool004d0oooo00D0okmd000000000000000007Boo`050?ooo`030;md00000000
07Bo00H0oooo00D0omnL04P000000000000007Boo`070?ooo`030000003oool0oooo0540oooo00<0
00000?ooo`3oool08@3oool3000002T0oooo00<000000?ooo`3oool08P3oool3000009l0oooo001=
0?ooo`0609a8M02oool0oooo0?oo_`1d07@0_ooo103oool00`2oM0000000001d_`080?ooo`030;md
001d_ol0oooo00L0oooo00<000000?ooo`3oool0D03oool00`000000oooo0?ooo`0N0?ooo`@00000
;03oool00`000000oooo0?ooo`0O0?ooo`<00000XP3oool004`0oooo00<0okmd001d_`3oool00P3o
ool00`3OW4P0B9cO0?ooo`0>0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0C`3o
ool00`000000oooo0?ooo`0K0?ooo`@00000<03oool00`000000oooo0?ooo`0M0?ooo`800000Y@3o
ool004`0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0:0?ooo`030=nLB0000000
000000<0000000<0M;oo0?ooo`3oool01@3oool4000004d0oooo00<000000?ooo`3oool06@3oool3
000003@0oooo00<000000?ooo`3oool06`3oool200000:L0oooo001<0?ooo`030=nLB018W=l0oooo
0080oooo00<0okmd001d_`3oool02P3oool01P3o_g@0001d0;ooo`3oool0_g@007Boo`P0oooo00<0
00000?ooo`3oool0C@3oool00`000000oooo0?ooo`0G0?ooo`<00000=`3oool00`000000oooo0?oo
o`0I0?ooo`800000Z@3oool004`0oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0;
0?ooo`050=nLB000M;l0oooo0;md001d_ol0203oool00`000000oooo0?ooo`1<0?ooo`030000003o
ool0oooo01D0oooo0`00000j0?ooo`030000003oool0oooo01L0oooo0P00002[0?ooo`00C03oool0
0`3o_g@0B9cO0?ooo`020?ooo`030=nLB018W=l0oooo00`0oooo00@0W4P004RLg`2oM000M;oo203o
ool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo01<0oooo0`00000m0?ooo`030000003o
ool0oooo01D0oooo0P00002]0?ooo`00C@3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;ooo`d0oooo
00@0omnL04P0B018B000M;oo203oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0100
oooo0`0000100?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0[@3oool004d0oooo
00D0oono07@000000000001809cOo`0?0?ooo`030?noM0000000M;oo00P0oooo00<000000?ooo`3o
ool0B`3oool00`000000oooo0?ooo`0=0?ooo`<00000@`3oool00`000000oooo0?ooo`0B0?ooo`80
0000/03oool006`0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0:0?ooo`<00000
AP3oool00`000000oooo0?ooo`0@0?ooo`800000/P3oool006`0oooo00<000000?ooo`3oool0BP3o
ool00`000000oooo0?ooo`080?ooo`<00000B@3oool00`000000oooo0?ooo`0?0?ooo`030000003o
ool0oooo0;80oooo001/0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool01@3oool3
000004`0oooo00<000000?ooo`3oool03@3oool200000;D0oooo001/0?ooo`030000003oool0oooo
04X0oooo00<000000?ooo`3oool00P3oool3000004l0oooo00<000000?ooo`3oool02`3oool20000
0;L0oooo001/0?ooo`030000003oool0oooo04`0oooo0`00001B0?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0]`3oool006`0oooo00<000000?ooo`3oool0B@3oool3000005D0oooo
00<000000?ooo`3oool0203oool200000;X0oooo001/0?ooo`030000003oool0oooo04H0oooo0`00
001H0?ooo`030000003oool0oooo00H0oooo0P00002l0?ooo`00K03oool3000004@0oooo0P000004
0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`2l
0?ooo`00K03oool00`000000oooo0?ooo`110?ooo`<000001P3oool00`000000oooo0?ooo`1D0?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0_@3oool006`0oooo00<000000?ooo`3o
ool0?`3oool2000000T0oooo00<000000?ooo`3oool0E03oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0;h0oooo001/0?ooo`030000003oool0oooo03`0oooo0`00000<0?ooo`030000
003oool0oooo05<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`2o0?ooo`00K03o
ool00`000000oooo0?ooo`0i0?ooo`<000003`3oool00`000000oooo0?ooo`1C0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0_`3oool006`0oooo00<000000?ooo`3oool0=`3oool2
00000180oooo00<000000?ooo`3oool0D`3oool01@000000oooo0?ooo`3oool000000<80oooo001/
0?ooo`030000003oool0oooo03@0oooo0`00000D0?ooo`030000003oool0oooo05<0oooo00@00000
0?ooo`3oool00000``3oool006`0oooo00<000000?ooo`3oool0<P3oool2000001P0oooo00<00000
0?ooo`3oool0D@3oool2000000030?ooo`000000oooo0<<0oooo001/0?ooo`030000003oool0oooo
0300oooo0P00000K0?ooo`030000003oool0oooo0500oooo0`0000350?ooo`00K03oool00`000000
oooo0?ooo`0^0?ooo`8000007P3oool00`000000oooo0?ooo`1?0?ooo`<00000a@3oool006`0oooo
00<000000?ooo`3oool0;03oool200000240oooo00<000000?ooo`3oool0CP3oool200000<H0oooo
001/0?ooo`030000003oool0oooo02X0oooo0P00000T0?ooo`030000003oool0oooo04d0oooo0P00
00360?ooo`00K03oool00`000000oooo0?ooo`0X0?ooo`8000009`3oool00`000000oooo0?ooo`1<
0?ooo`800000aP3oool006`0oooo00<000000?ooo`3oool09P3oool2000002X0oooo00<000000?oo
o`3oool0B`3oool200000<H0oooo001/0?ooo`<00000903oool2000002d0oooo00<000000?ooo`3o
ool0BP3oool200000<H0oooo001/0?ooo`030000003oool0oooo0280oooo0P00000`0?ooo`030000
003oool0oooo04T0oooo0P0000360?ooo`00K03oool00`000000oooo0?ooo`0Q0?ooo`030000003o
ool0oooo0340oooo0P0000190?ooo`800000aP3oool006`0oooo00<000000?ooo`3oool07`3oool2
000003H0oooo00<000000?ooo`3oool0AP3oool200000<H0oooo001/0?ooo`030000003oool0oooo
01h0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`150?ooo`800000aP3oool006`0
oooo00<000000?ooo`3oool0703oool2000003/0oooo00<000000?ooo`3oool0A03oool200000<H0
oooo001/0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0?03oool00`000000oooo
0?ooo`130?ooo`<00000a@3oool006`0oooo00<000000?ooo`3oool06@3oool200000400oooo0P00
00130?ooo`<00000a@3oool006`0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`12
0?ooo`030000003oool0oooo0400oooo0P0000000`3oool000000?ooo`330?ooo`00K03oool00`00
0000oooo0?ooo`0F0?ooo`800000AP3oool200000400oooo0P0000020?ooo`030000003oool0oooo
0<40oooo001/0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0B03oool00`000000
oooo0?ooo`0m0?ooo`8000000`3oool00`000000oooo0?ooo`300?ooo`00K03oool00`000000oooo
0?ooo`0C0?ooo`800000C03oool2000003d0oooo0P0000040?ooo`030000003oool0oooo0;l0oooo
001/0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0CP3oool2000003/0oooo0P00
00040?ooo`030000003oool0oooo0;l0oooo001/0?ooo`<000004@3oool00`000000oooo0?ooo`1A
0?ooo`800000>@3oool2000000D0oooo00<000000?ooo`3oool0_P3oool006`0oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`1D0?ooo`800000=`3oool2000000H0oooo00<000000?oo
o`3oool0_@3oool006`0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`1G0?ooo`80
0000=@3oool2000000L0oooo00<000000?ooo`3oool0_03oool006`0oooo00<000000?ooo`3oool0
3@3oool2000005`0oooo0`00000b0?ooo`800000203oool200000;`0oooo001/0?ooo`030000003o
ool0oooo00`0oooo00<000000?ooo`3oool0G`3oool200000300oooo0P00000:0?ooo`800000^P3o
ool006`0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1R0?ooo`<00000;@3oool2
000000`0oooo00<000000?ooo`3oool0]`3oool006`0oooo00<000000?ooo`3oool02P3oool00`00
0000oooo0?ooo`1V0?ooo`800000:`3oool2000000d0oooo0P00002g0?ooo`00K03oool00`000000
oooo0?ooo`090?ooo`030000003oool0oooo06T0oooo0P00000Y0?ooo`8000003`3oool200000;D0
oooo001/0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0K03oool3000002H0oooo
0P00000A0?ooo`030000003oool0oooo0;80oooo001/0?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0L03oool2000002@0oooo0P00000B0?ooo`800000/P3oool004d0oooo00D0okmd
000000000000000007Boo`050?ooo`030;md0000000007Bo00H0oooo00D0_g@0000000000000001d
0;ooo`070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0L`3oool300000240oooo
0P00000D0?ooo`030000003oool0oooo0:l0oooo001=0?ooo`0609a8M02oool0oooo0?oo_`1d07@0
_ooo103oool00`2oM0000000001d_`050?ooo`060=nLB000M;l0oooo0?ooo`3o_g@007Bo1`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo07L0oooo0`00000N0?ooo`8000005@3oool2
00000:l0oooo001<0?ooo`030?noM000M;l0oooo0080oooo00<0gia804RLg`3oool02P3oool00`3o
okl0M00007Boo`020?ooo`030?oo_`1dB9`0oooo00H0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`1k0?ooo`<000006`3oool2000001L0oooo0P00002]0?ooo`005`3oool300000380
oooo00<0gia804RLg`3oool00P3oool00`3o_g@007Bo0?ooo`0:0?ooo`070?noM000B4P0M01d0;oo
o`3oool0gia8001d_`070?ooo`@000000`3oool00`000000oooo0?ooo`1n0?ooo`<00000603oool2
000001T0oooo0P00002[0?ooo`006P3oool4000002h0oooo00<0gia804RLg`3oool00P3oool00`3o
_g@007Bo0?ooo`0:0?ooo`030?oOW018B9`0okmd0080000000<004RL0?ooo`3oool01P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0880oooo0`00000E0?ooo`8000006`3oool30000
0:P0oooo000N0?ooo`@00000:P3oool00`3OW4P0B9cO0?ooo`020?ooo`030?noM000M;l0oooo00X0
oooo00<0oono07@0M02oool02`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo08H0
oooo0`00000B0?ooo`8000007P3oool300000:D0oooo000R0?ooo`@000009P3oool00`3o_g@0B9cO
0?ooo`020?ooo`030=nLB018W=l0oooo00/0oooo00<0_g@004RLg`3oool02P3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo08T0oooo1000000>0?ooo`8000008@3oool300000:80oooo
000V0?ooo`D000008P3oool01P2LB000M;oo0?ooo`3ogi`0B01d0;ooo``0oooo00<0oono07@00000
M;l02P3oool01@000000oooo0?ooo`3oool000000900oooo1000000:0?ooo`800000903oool20000
0:00oooo000[0?ooo`D000007@3oool01@3ookl0M0000000000004P0W=oo00l0oooo00@0_g@00000
0000000004RL1`3oool01@000000oooo0?ooo`3oool0000009@0oooo0P0000080?ooo`8000009P3o
ool3000009d0oooo000`0?ooo`@00000>03oool010000000oooo0?ooo`00002G0?ooo`@00000103o
ool2000002T0oooo0P00002K0?ooo`00=03oool5000003<0oooo00@000000?ooo`3oool00000V`3o
ool200000080oooo0P00000[0?ooo`<00000V03oool003T0oooo1@00000^0?ooo`030000003oool0
000009h0oooo1P00000/0?ooo`<00000U@3oool003h0oooo1000000Z0?ooo`030000003oool00000
0:00oooo0P0000020?ooo`<00000;03oool2000009<0oooo00120?ooo`D000009@3oool200000:40
oooo0P0000050?ooo`D00000:@3oool4000008l0oooo00170?ooo`@000008@3oool200000:40oooo
0P00000:0?ooo`D00000:03oool4000008/0oooo001;0?ooo`D00000703oool200000:40oooo0P00
000?0?ooo`D000009`3oool5000008H0oooo001@0?ooo`H000005P3oool300000:00oooo0P00000D
0?ooo`@00000:03oool400000880oooo001F0?ooo`H00000403oool00`000000oooo0?ooo`2P0?oo
o`800000603oool4000002P0oooo1000001n0?ooo`00G03oool5000000/0oooo00<000000?ooo`3o
ool0X03oool2000001`0oooo1000000X0?ooo`@00000NP3oool00640oooo1P0000050?ooo`030000
003oool0oooo0:00oooo0P00000P0?ooo`D000009`3oool3000007L0oooo001W0?ooo`H00000XP3o
ool2000002D0oooo1@00000U0?ooo`@00000L`3oool006`0oooo1`00002L0?ooo`800000:P3oool4
000002D0oooo1000001_0?ooo`00K03oool00`000000oooo0?ooo`040?ooo`H00000UP3oool20000
02h0oooo1@00000T0?ooo`@00000J`3oool006`0oooo00<000000?ooo`3oool02P3oool600000900
oooo0P00000c0?ooo`D000008`3oool6000006D0oooo001/0?ooo`030000003oool0oooo0100oooo
1P00002:0?ooo`800000>03oool6000002<0oooo1P00001O0?ooo`00K03oool2000001L0oooo1P00
00240?ooo`800000?P3oool6000002<0oooo1P00001I0?ooo`00K03oool2000001d0oooo1P00001n
0?ooo`800000A03oool5000002@0oooo1000001E0?ooo`00K03oool2000002<0oooo1P00001h0?oo
o`800000B@3oool600000280oooo1000001A0?ooo`00K03oool2000002T0oooo1P00001b0?ooo`80
0000C`3oool600000200oooo1000001=0?ooo`00K03oool00`000000oooo0000000^0?ooo`H00000
K03oool2000005D0oooo1@00000O0?ooo`H00000A`3oool006`0oooo0`00000d0?ooo`P00000I03o
ool2000005X0oooo1P00000O0?ooo`H00000@@3oool006`0oooo00@000000?ooo`3oool00000>`3o
ool8000005`0oooo0P00001P0?ooo`H000007`3oool6000003/0oooo001/0?ooo`040000003oool0
oooo000004<0oooo2@00001C0?ooo`800000IP3oool8000001d0oooo2000000c0?ooo`00K03oool0
1@000000oooo0?ooo`3oool0000004/0oooo2@00001:0?ooo`800000KP3oool9000001`0oooo2000
000[0?ooo`00K03oool01@000000oooo0?ooo`3oool0000005@0oooo2@0000110?ooo`800000M`3o
ool7000001d0oooo1`00000T0?ooo`00K03oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo05X0oooo2@00000h0?ooo`800000OP3oool6000001h0oooo1P00000N0?ooo`00K03oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo06<0oooo2@00000_0?ooo`800000Q03oool60000
01h0oooo1P00000H0?ooo`00K03oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06/0
oooo2000000W0?ooo`800000RP3oool8000001`0oooo0`00000E0?ooo`00K03oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0780oooo2@00000N0?ooo`800000TP3oool9000002/0oooo
001/0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0NP3oool9000001D0oooo0P00
002K0?ooo`T000008P3oool006`0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`22
0?ooo`T00000303oool200000:@0oooo2000000J0?ooo`00K03oool00`000000oooo0?ooo`070?oo
o`800000R`3oool8000000@0oooo0P00002/0?ooo`D000005@3oool006`0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`2@0?ooo`T00000``3oool006`0oooo0`00000:0?ooo`030000
003oool0oooo09<0oooo0P0000030?ooo`X00000^@3oool006`0oooo00<000000?ooo`3oool02`3o
ool00`000000oooo0?ooo`2B0?ooo`8000003@3oool=00000:`0oooo001/0?ooo`030000003oool0
oooo00`0oooo0P00002B0?ooo`8000006P3oool=000009l0oooo001/0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool0S`3oool2000002L0oooo2000002G0?ooo`00K03oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo08h0oooo0P00000_0?ooo`h00000R@3oool006`0oooo
00<000000?ooo`3oool0403oool2000008h0oooo0P00000m0?ooo`h00000N`3oool006`0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`2;0?ooo`800000B`3oool8000007<0oooo001/
0?ooo`030000003oool0oooo01<0oooo0P00002;0?ooo`800000D`3oool>000006D0oooo001/0?oo
o`030000003oool0oooo01D0oooo0P0000290?ooo`800000H@3ooolB000005<0oooo001=0?ooo`05
0?noM000000000000000001d_ol01@3oool00`2oM0000000001d_`050?ooo`050;md000000000000
00000000M;l0203oool00`000000oooo0?ooo`0G0?ooo`800000Q`3oool2000007<0oooo3P000015
0?ooo`00C@3oool01P2LB7@0_ooo0?ooo`3ookl0M01d0;ooo`@0oooo00<0_g@000000000M;l0103o
ool00`3o_g@007Bo0?ooo`020?ooo`030;mdB02Lgol0oooo00H0oooo00<000000?ooo`3oool06@3o
ool2000008D0oooo0P0000210?ooo`P00000?@3oool004`0oooo00<0okmd001d_`3oool00P3oool0
0`3OW4P0B9cO0?ooo`0:0?ooo`030=nLB018W=l0oooo0080oooo00<0_g@007Boo`3oool01P3oool0
0`000000oooo0?ooo`0K0?ooo`800000P`3oool2000008T0oooo3@00000`0?ooo`00C03oool00`3O
W4P0B9cO0?ooo`020?ooo`030?noM000M;l0oooo00X0oooo00<0okmd0018W03oool00P3oool00`2L
B4P0W=oo0?ooo`060?ooo`030000003oool0oooo01d0oooo0P0000210?ooo`800000UP3ooolB0000
01h0oooo001<0?ooo`030=nLB018W=l0oooo0080oooo00<0okmd001d_`3oool02`3oool01@3OW4P0
0000000000000000B9cO00P0oooo1000000N0?ooo`800000LP3ooolf00000840oooo2@00000E0?oo
o`00C03oool00`3OW4P0B9cO0?ooo`020?ooo`030?noM000M;l0oooo00X0oooo00<0oono07A8W03o
ool00P3oool00`2LB7@0_ooo0?ooo`060?ooo`030000003oool0oooo0240oooo0P00001F0?oooaX0
00003P3oool00`000000oooo0?ooo`0U0?ooobP00000M`3oool004`0oooo00<0okmd04RLg`3oool0
0P3oool00`3OW4P0B9cO0?ooo`0:0?ooo`030?noM018W=l0oooo0080oooo00<0_g@007Boo`3oool0
1P3oool00`000000oooo0?ooo`0S0?ooo`<00000B`3oool8000002P0oooo00<000000?ooo`3oool0
C@3ooolT000005<0oooo001=0?ooo`0609a8001d_ol0oooo0?oOW01807@0_ooo2`3oool01`3ogi`0
B01d0;ooo`3oool0omnL04P0M02oool01`3oool00`000000oooo0?ooo`0V0?ooo`<00000?`3oool9
00000300oooo00<000000?ooo`3oool0L@3ooolB00000440oooo001=0?ooo`050?oo_`1d00000000
0000B02Lgol03@3oool01@3o_g@000000000000004P0W=oo00P0oooo00<000000?ooo`3oool0:@3o
ool3000003<0oooo2@00000i0?ooo`030000003oool0oooo08<0oooo8P00000O0?ooo`00K03oool0
0`000000oooo0?ooo`0/0?ooo`<00000:@3oool700000480oooo00<000000?ooo`3oool0Y@3oool:
000001D0oooo001/0?ooo`030000003oool0oooo02l0oooo0`00000S0?ooo`<00000B@3oool00`00
0000oooo0?ooo`340?ooo`00K03oool00`000000oooo0?ooo`0b0?ooo`<000007P3oool2000004`0
oooo00<000000?ooo`3oool0a03oool006`0oooo00<000000?ooo`3oool0=@3oool3000001L0oooo
1000001>0?ooo`030000003oool0oooo0<@0oooo001/0?ooo`030000003oool0oooo03P0oooo1000
000A0?ooo`800000DP3oool00`000000oooo0?ooo`340?ooo`00K03oool00`000000oooo0?ooo`0l
0?ooo`<00000303oool2000005@0oooo00<000000?ooo`3oool0a03oool006`0oooo00<000000?oo
o`3oool0?`3oool4000000P0oooo00<000000?ooo`3oool0D`3oool00`000000oooo0?ooo`340?oo
o`00K03oool00`000000oooo0?ooo`130?ooo`@000000`3oool00`000000oooo0?ooo`1D0?ooo`03
0000003oool0oooo0<@0oooo001/0?ooo`<00000A`3oool4000005H0oooo00<000000?ooo`3oool0
a03oool006`0oooo00<000000?ooo`3oool0BP3oool500000580oooo00<000000?ooo`3oool0a03o
ool006`0oooo00<000000?ooo`3oool0B`3oool010000000oooo0?ooo`3oool5000004d0oooo00<0
00000?ooo`3oool0a03oool006`0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`06
0?ooo`@00000B@3oool00`000000oooo0?ooo`340?ooo`00K03oool00`000000oooo0?ooo`1<0?oo
o`8000002P3oool5000004@0oooo00<000000?ooo`3oool0a03oool006`0oooo00<000000?ooo`3o
ool0CP3oool2000000d0oooo1P00000n0?ooo`030000003oool0oooo0<@0oooo001/0?ooo`030000
003oool0oooo0500oooo0`00000@0?ooo`H00000>03oool00`000000oooo0?ooo`340?ooo`00K03o
ool00`000000oooo0?ooo`1C0?ooo`<000004`3oool800000300oooo00<000000?ooo`3oool0a03o
ool006`0oooo00<000000?ooo`3oool0EP3oool3000001P0oooo2@00000W0?ooo`030000003oool0
oooo0<@0oooo001/0?ooo`030000003oool0oooo05T0oooo1000000M0?ooo`L00000803oool00`00
0000oooo0?ooo`340?ooo`00K03oool00`000000oooo0?ooo`1M0?ooo`@00000803oool6000001X0
oooo00<000000?ooo`3oool0a03oool006`0oooo00<000000?ooo`3oool0H@3oool500000240oooo
1P00000D0?ooo`030000003oool0oooo0<@0oooo001/0?ooo`030000003oool0oooo06H0oooo1000
000S0?ooo`P00000303oool00`000000oooo0?ooo`340?ooo`00K03oool3000006X0oooo1@00000V
0?ooo`T000000`3oool00`000000oooo0?ooo`340?ooo`00K03oool00`000000oooo0?ooo`1_0?oo
o`H00000:@3oool900000<40oooo001/0?ooo`030000003oool0oooo07D0oooo1P00000V0?ooo`03
0000003oool0oooo00<0oooo2000002i0?ooo`00K03oool00`000000oooo0?ooo`1k0?ooo`L00000
7`3oool00`000000oooo0?ooo`0;0?ooo`d00000[03oool006`0oooo00<000000?ooo`3oool0PP3o
ool9000001H0oooo00<000000?ooo`3oool0603oool>000009h0oooo001/0?ooo`030000003oool0
oooo08/0oooo3000000:0?ooo`030000003oool0oooo02H0oooo2000002F0?ooo`00K03oool00`00
0000oooo0?ooo`2G0?ooo`/00000<03oool>000008P0oooo001/0?ooo`030000003oool0oooo0:80
oooo3000000b0?oooa800000MP3oool006`0oooo00<000000?ooo`3oool0[P3oool>000003H0oooo
4@00001U0?ooo`00K03oool00`000000oooo0?ooo`2l0?oooa800000=@3ooolB000005<0oooo001/
0?ooo`030000003oool0oooo0<h0oooo4P00000e0?oooa800000@@3oool006`0oooo00<000000?oo
o`3oool0h03ooolA000003H0oooo4@00000`0?ooo`00K03oool00`000000oooo0?ooo`3a0?oooa40
0000=P3ooolA000001l0oooo001/0?ooo`030000003oool0oooo0?l0oooo0`3ooolB000003D0oooo
2P00000E0?ooo`00K03oool300000?l0oooo5@3ooolT00000300oooo001/0?ooo`030000003oool0
oooo0?l0oooo>@3ooolK000001D0oooo001/0?ooo`030000003oool0oooo0?l0ooooJ@3oool006`0
oooo00<000000?ooo`3oool0o`3ooomY0?ooo`00K03oool00`000000oooo0?ooo`3o0?ooofT0oooo
001/0?ooo`030000003oool0oooo0?l0ooooJ@3oool006`0oooo00<000000?ooo`3oool0o`3ooomY
0?ooo`00K03oool00`000000oooo0?ooo`3o0?ooofT0oooo001/0?ooo`030000003oool0oooo0?l0
ooooJ@3oool005h0oooo00<0okmd0000000000000`0000000`00M;l0oooo0?ooo`050?ooo`030000
003oool0oooo0?l0ooooJ@3oool00640oooo00<0W4Q809cOo`3oool0203oool00`000000oooo0?oo
o`2Q0?ooohH00000@@3oool00640oooo00<0W4Q809cOo`3oool0203oool00`000000oooo0?ooo`2W
0?ooocL00000B@3oool/000001D0oooo001Q0?ooo`0309a8B02Lgol0oooo00P0oooo00<000000?oo
o`3oool0gP3ooome000001D0oooo001Q0?ooo`0309a8B02Lgol0oooo00P0oooo1000003o0?ooofP0
oooo001Q0?ooo`0309a8B02Lgol0oooo00P0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`00H@3o
ool00`2LB4P0W=oo0?ooo`080?ooo`030000003oool0oooo0?l0ooooJ@3oool005h0oooo00D0gia8
00000000B9`0W4Q809cOo`090?ooo`030000003oool0oooo0?l0ooooJ@3oool00600oooo00<0gia8
0000B02Lgol02@3oool00`000000oooo0?ooo`3o0?ooofT0oooo001/0?ooo`030000003oool0oooo
0?l0ooooJ@3oool006`0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 470}, {290, 0}} -> {2.36384, -0.0722899, 0.0087947, \
0.00567853}}]
}, Closed]],

Cell[TextData[{
  "The error messages are generated for roots that are complex for some range \
of \[Mu] and are no cause for alarm.  As \[Mu] increases more of these roots \
become real and gradually assemble the pitchfork bifurcation pattern for the \
logistic map.  More detail about the nature of period doubling and the onset \
of chaos within deterministic systems can be found in ",
  StyleBox[ButtonBox["cobweb.nb",
    ButtonData:>{"Applications/cobweb.nb", None},
    ButtonStyle->"Hyperlink"],
    FontWeight->"Bold",
    FontSlant->"Italic"],
  StyleBox[".",
    FontSlant->"Italic"]
}], "Text"],

Cell[TextData[{
  "Solve the following equation for the quantity r=y/x and plot the real \
solutions for {a,-1,1}:\n\t\t",
  Cell[BoxData[
      \(TraditionalForm\`2\/y\^4 == 1\/x\^4 + a\/\((x + y)\)\^4\)]],
  "\n[Hint: substitute y\[Rule]r x first and then simplify the equation as \
much as possible.  (You may need to use Map to perform the same \
transformations to both sides of the equation.)]"
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Algebraic equations in transcendental disguise", "Subsection"],

Cell[TextData[{
  StyleBox["Solve",
    FontWeight->"Bold"],
  " is designed to handle polynomial equations, but will often attempt to \
find solutions for more general equations.  For example, in the trivial \
example"
}], "Text",
  FontWeight->"Plain"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[Cos[x] \[Equal] Sin[x], x]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"ifun"\), \(\(:\)\(\ \)\), "\<\"Inverse functions are \
being used by \\!\\(Solve\\), so some solutions may not be found; use Reduce \
for complete solution information. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({{x \[Rule] \(-\(\(3\ \[Pi]\)\/4\)\)}, {x \[Rule] \[Pi]\/4}}\)], \
"Output"]
}, Closed]],

Cell[TextData[
"two solutions are reported, but the 2\[Pi] periodicity is overlooked when \
only the principal branches of inverse trigonometric functions are used.  \
However, in other cases inverse functions do not any provide solutions at all \
and numerical solutions must be sought (see below)."], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[x \[Equal] Tan[x], x]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"tdep"\), \(\(:\)\(\ \)\), "\<\"The equations appear to \
involve the variables to be solved for in an essentially non-algebraic way. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::tdep\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \(Solve[x \[Equal] Tan[x], x]\)], "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Examples", "Subsection"],

Cell[CellGroupData[{

Cell["Example: y-scaling for quasielastic scattering", "Subsubsection"],

Cell[TextData[{
  "Suppose that a high-energy electron scatters from one of the nucleons in a \
nucleus with mass ",
  Cell[BoxData[
      \(TraditionalForm\`m\_A\)]],
  ", delivering energy transfer ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\)]],
  " and momentum transfer ",
  Cell[BoxData[
      \(TraditionalForm\`q\)]],
  ", such that a nucleon of mass ",
  Cell[BoxData[
      \(TraditionalForm\`m\_N\)]],
  " is ejected leaving a residual nucleus of mass ",
  Cell[BoxData[
      \(TraditionalForm\`m\_B\)]],
  ".  In the spectator model, the target nucleus is represented at the time \
of the interaction as the struck nucleon with initial momentum ",
  Cell[BoxData[
      \(TraditionalForm\`y\)]],
  " plus spectator ",
  Cell[BoxData[
      \(TraditionalForm\`m\_B\)]],
  " with momentum ",
  Cell[BoxData[
      \(TraditionalForm\`\(-y\)\)]],
  ", such that the ejectile emerges with final momentum ",
  Cell[BoxData[
      \(TraditionalForm\`y + q\)]],
  ".  Conservation of energy (relativistic) then requires"
}], "Text"],

Cell[BoxData[
    \(\(eq1 = \[Omega] + 
            m\_A \[Equal] \@\(m\_N\^2 + \((y + q)\)\^2\) + \@\(m\_B\^2 + y\^2\
\);\)\)], "Input"],

Cell[TextData[{
  "Don't worry if you are unfamiliar with the physics behind this equation \
\[LongDash] we only pose it as an example of a nontrivial algebraic equation \
which we would rather solve using ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " than by hand."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol1 = Solve[eq1, y] // Simplify\)], "Input"],

Cell[BoxData[
    \({{y \[Rule] \(-\(\(1\/\(2\ \((\(-q\^2\) + \[Omega]\^2 + 
                      2\ \[Omega]\ m\_A + m\_A\%2)\)\)\)\((\(-q\^3\) + 
                q\ \[Omega]\^2 + 2\ q\ \[Omega]\ m\_A + q\ m\_A\%2 + 
                q\ m\_B\%2 - 
                q\ m\_N\%2 + \@\(\((\[Omega] + m\_A)\)\^2\ \((4\ \[Omega]\ \
m\_A\%3 + m\_A\%4 + m\_B\%4 + 2\ m\_B\%2\ \((q\^2 - \[Omega]\^2 - m\_N\%2)\) \
+ \((q\^2 - \[Omega]\^2 + m\_N\%2)\)\^2 - 2\ m\_A\%2\ \((q\^2 - 3\ \
\[Omega]\^2 + m\_B\%2 + m\_N\%2)\) - 4\ \[Omega]\ m\_A\ \((q\^2 - \[Omega]\^2 \
+ m\_B\%2 + m\_N\%2)\))\)\))\)\)\)}, {y \[Rule] \(-\(\(1\/\(2\ \((\(-q\^2\) + \
\[Omega]\^2 + 2\ \[Omega]\ m\_A + m\_A\%2)\)\)\)\((\(-q\^3\) + 
                q\ \[Omega]\^2 + 2\ q\ \[Omega]\ m\_A + q\ m\_A\%2 + 
                q\ m\_B\%2 - 
                q\ m\_N\%2 - \@\(\((\[Omega] + m\_A)\)\^2\ \((4\ \[Omega]\ \
m\_A\%3 + m\_A\%4 + m\_B\%4 + 2\ m\_B\%2\ \((q\^2 - \[Omega]\^2 - m\_N\%2)\) \
+ \((q\^2 - \[Omega]\^2 + m\_N\%2)\)\^2 - 2\ m\_A\%2\ \((q\^2 - 3\ \
\[Omega]\^2 + m\_B\%2 + m\_N\%2)\) - 4\ \[Omega]\ m\_A\ \((q\^2 - \[Omega]\^2 \
+ m\_B\%2 + m\_N\%2)\))\)\))\)\)\)}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " produces two solutions, but the ordering of these solutions is to some \
degree arbitrary and varies from version to version.  We must select the \
appropriate solution using additional information, such as a physical \
interpretation or constraint.  For this problem, it is useful to specialize \
to a target which consists only of a single nucleon without any residual \
spectator.  Furthermore, we know that the momentum transfer must be greater \
than or equal to the energy transfer.  Hence, it is useful to solve for ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\)]],
  " in terms of ",
  Cell[BoxData[
      \(TraditionalForm\`q\)]],
  " for this special case.  "
}], "Text"],

Cell[TextData[{
  "We use ",
  StyleBox["Thread", "InlineInput",
    FontWeight->"Bold"],
  " to produce new equations for this special case from the list of possible \
solutions for the general case.  "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(newEqs\  = 
      Thread[\(0 \[Equal] y /. sol1\) /. {m\_A \[Rule] m\_N, 
            m\_B \[Rule] 0}]\)], "Input"],

Cell[BoxData[
    \({0 \[Equal] \(-\(\(\(-q\^3\) + q\ \[Omega]\^2 + 
                2\ q\ \[Omega]\ m\_N + \@\(\((\[Omega] + m\_N)\)\^2\ \((4\ \
\[Omega]\ m\_N\%3 + m\_N\%4 - 2\ m\_N\%2\ \((q\^2 - 3\ \[Omega]\^2 + m\_N\%2)\
\) - 4\ \[Omega]\ m\_N\ \((q\^2 - \[Omega]\^2 + m\_N\%2)\) + \((q\^2 - \
\[Omega]\^2 + m\_N\%2)\)\^2)\)\)\)\/\(2\ \((\(-q\^2\) + \[Omega]\^2 + 
                    2\ \[Omega]\ m\_N + m\_N\%2)\)\)\)\), 
      0 \[Equal] \(-\(\(\(-q\^3\) + q\ \[Omega]\^2 + 
                2\ q\ \[Omega]\ m\_N - \@\(\((\[Omega] + m\_N)\)\^2\ \((4\ \
\[Omega]\ m\_N\%3 + m\_N\%4 - 2\ m\_N\%2\ \((q\^2 - 3\ \[Omega]\^2 + m\_N\%2)\
\) - 4\ \[Omega]\ m\_N\ \((q\^2 - \[Omega]\^2 + m\_N\%2)\) + \((q\^2 - \
\[Omega]\^2 + m\_N\%2)\)\^2)\)\)\)\/\(2\ \((\(-q\^2\) + \[Omega]\^2 + 
                    2\ \[Omega]\ m\_N + m\_N\%2)\)\)\)\)}\)], "Output"]
}, Closed]],

Cell["\<\
Nested expansion and simplification steps are needed to print attractive \
versions of these equations, but that does not affect their solution.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(newEqs\  = \((\(newEqs // Simplify\) // PowerExpand)\) // 
        Simplify\)], "Input"],

Cell[BoxData[
    \({\(q\^2 - \[Omega]\^2 - 2\ \[Omega]\ m\_N\)\/\(q - \[Omega] - m\_N\) \
\[Equal] 0, \(\(-q\^2\) + \[Omega]\^2 + 2\ \[Omega]\ m\_N\)\/\(q + \[Omega] + \
m\_N\) \[Equal] 0}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Solve[newEqs, \[Omega]]\)], "Input"],

Cell[BoxData[
    \({{\[Omega] \[Rule] \(-m\_N\) - \@\(q\^2 + m\_N\%2\)}, {\[Omega] \[Rule] \
\(-m\_N\) + \@\(q\^2 + m\_N\%2\)}}\)], "Output"]
}, Closed]],

Cell["\<\
It is now clear that the second solution is the one that we want.\
\>", "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Example: nonrelativistic binary collisions", "Subsubsection"],

Cell[TextData[{
  "A nonrelativistic binary collision is described by a mass-balance equation \
of the form ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          StyleBox[\(m1 + m2\),
            FontSlant->"Italic"], 
          StyleBox["\[Equal]",
            FontSlant->"Plain"], 
          StyleBox[\(m3 + m4\),
            FontSlant->"Italic"]}], TraditionalForm]]],
  ", where we consider ",
  Cell[BoxData[
      FormBox[
        StyleBox["m1",
          FontSlant->"Italic"], TraditionalForm]]],
  " to be the projectile, ",
  StyleBox["m2",
    FontSlant->"Italic"],
  " the target, ",
  StyleBox["m3",
    FontSlant->"Italic"],
  " the scattered particle, and ",
  Cell[BoxData[
      FormBox[
        StyleBox["m4",
          FontSlant->"Italic"], TraditionalForm]]],
  " the scattered target, but it is also possible that the reactants exchange \
mass during the collision.  The kinematics of the reaction are governed by \
conservation of momentum and of energy, which are expressed by the following \
equations."
}], "Text"],

Cell[BoxData[
    \(\(collisionEqs = {m1 + m2 \[Equal] m3 + m4, p1 + p2 \[Equal] p3 + p4, 
          p1 . p1\/\(2\ m1\) + p2 . p2\/\(2\ m2\) + qvalue \[Equal] 
            p3 . p3\/\(2\ m3\) + p4 . p4\/\(2\ m4\)};\)\)], "Input"],

Cell[TextData[{
  "Here the ",
  StyleBox["qvalue",
    FontSlant->"Italic"],
  " measures the inelasticity of the reaction and vanishes for elastic \
scattering.  Note that we use ",
  ButtonBox["Dot",
    ButtonStyle->"RefGuideLink"],
  ", with operator notation ., to indicate the scalar product of two vectors. \
 The momenta are defined using a set of replacement rules.  Note that the \
since the reaction is confined to a plane, we can omit the z-components and \
work with vectors in two dimensions."
}], "Text"],

Cell[BoxData[
    \(\(momentumComponentRules = {p1 \[Rule] {p1x, p1y}, 
          p2 \[Rule] {p2x, p2y}, p3 \[Rule] {p3x, p3y}, 
          p4 \[Rule] {p4x, p4y}, p1x \[Rule] p1r\ Cos[theta1], 
          p1y \[Rule] p1r\ Sin[theta1], p2x \[Rule] p2r\ Cos[theta2], 
          p2y \[Rule] p2r\ Sin[theta2], p3x \[Rule] p3r\ Cos[theta3], 
          p3y \[Rule] p3r\ Sin[theta3]};\)\)], "Input"],

Cell[TextData[{
  "Because there are 4 independent equations, we can solve this system for 4 \
variables.  It is convenient to specify three masses ",
  StyleBox["{m1,m2,m3}",
    FontSlant->"Italic"],
  ", the projectile momentum ",
  StyleBox["{p1x,p1y}",
    FontSlant->"Italic"],
  " in terms of magnitude ",
  StyleBox["p1r",
    FontSlant->"Italic"],
  " and angle ",
  StyleBox["theta1",
    FontSlant->"Italic"],
  ", the target momentum {",
  StyleBox["p2x,p2y}",
    FontSlant->"Italic"],
  " in terms of magnitude ",
  StyleBox["p2r",
    FontSlant->"Italic"],
  " and angle ",
  StyleBox["theta2",
    FontSlant->"Italic"],
  ", and the projectile scattering angle ",
  StyleBox["theta3",
    FontSlant->"Italic"],
  ". ",
  StyleBox[" Mathematica",
    FontSlant->"Italic"],
  " will then happily report a general solution for the remaining mass ",
  StyleBox["m4",
    FontSlant->"Italic"],
  ", the momentum ",
  StyleBox["p3r",
    FontSlant->"Italic"],
  " for the scattered particle, and the momentum ",
  StyleBox["{p4x,p4y}",
    FontSlant->"Italic"],
  " of the recoil particle.  Although we could use ",
  StyleBox["Simplify", "InlineInput",
    FontWeight->"Bold"],
  ", the minimal return is not worth the considerable investment of time \
required to simplify expressions of this complexity!"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(\(gensol\)\(=\)\(Solve[
        collisionEqs //. momentumComponentRules, {m4, p3r, p4x, 
          p4y}]\)\(\ \)\)\)], "Input"],

Cell[BoxData[
    \({{p4x \[Rule] 
          p1r\ Cos[theta1] + 
            p2r\ Cos[
                theta2] - \(m1\ m2\ m3\ p1r\ Cos[theta1]\ \
Cos[theta3]\^2\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ \
Cos[theta2]\ Cos[theta3]\^2\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ \
Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \
\(m1\ m2\ m3\ p1r\ Cos[theta3]\ Sin[theta1]\ Sin[theta3]\)\/\(m1\^2\ m2\ \
Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ \
m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ Cos[theta3]\ Sin[theta2]\ \
Sin[theta3]\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \((Cos[
                    theta3]\ \[Sqrt]\((\((\(-2\)\ m1\ m2\ m3\ p1r\ \
Cos[theta1]\ Cos[theta3] - 2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] - 2\ \
m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] - 2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3])\)\^2 - 4\ \((\(-2\)\ m1\^2\ m2\ m3\ qvalue - 
                              2\ m1\ m2\^2\ m3\ qvalue + 
                              2\ m1\ m2\ m3\^2\ qvalue - 
                              m2\^2\ m3\ p1r\^2\ Cos[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Cos[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Cos[theta1]\ Cos[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Cos[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Cos[theta2]\^2 - 
                              m2\^2\ m3\ p1r\^2\ Sin[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Sin[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Sin[theta1]\ Sin[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Sin[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Sin[theta2]\^2)\)\ \((m1\^2\ \
m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + 
                              m1\ m2\^2\ Sin[theta3]\^2)\))\))\)/\((2\ \
\((m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + 
                      m1\^2\ m2\ Sin[theta3]\^2 + 
                      m1\ m2\^2\ Sin[theta3]\^2)\))\), 
        p4y \[Rule] 
          p1r\ Sin[theta1] + 
            p2r\ Sin[
                theta2] - \(m1\ m2\ m3\ p1r\ Cos[theta1]\ Cos[theta3]\ \
Sin[theta3]\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ \
Cos[theta2]\ Cos[theta3]\ Sin[theta3]\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\
\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) \
- \(m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3]\^2\)\/\(m1\^2\ m2\ \
Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ \
m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3]\^2\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \((Sin[
                    theta3]\ \[Sqrt]\((\((\(-2\)\ m1\ m2\ m3\ p1r\ \
Cos[theta1]\ Cos[theta3] - 2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] - 2\ \
m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] - 2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3])\)\^2 - 4\ \((\(-2\)\ m1\^2\ m2\ m3\ qvalue - 
                              2\ m1\ m2\^2\ m3\ qvalue + 
                              2\ m1\ m2\ m3\^2\ qvalue - 
                              m2\^2\ m3\ p1r\^2\ Cos[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Cos[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Cos[theta1]\ Cos[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Cos[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Cos[theta2]\^2 - 
                              m2\^2\ m3\ p1r\^2\ Sin[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Sin[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Sin[theta1]\ Sin[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Sin[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Sin[theta2]\^2)\)\ \((m1\^2\ \
m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + 
                              m1\ m2\^2\ Sin[theta3]\^2)\))\))\)/\((2\ \
\((m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + 
                      m1\^2\ m2\ Sin[theta3]\^2 + 
                      m1\ m2\^2\ Sin[theta3]\^2)\))\), 
        p3r \[Rule] \((2\ m1\ m2\ m3\ p1r\ Cos[theta1]\ Cos[theta3] + 
                2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] + 
                2\ m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] + 
                2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ Sin[
                    theta3] + \[Sqrt]\((\((\(-2\)\ m1\ m2\ m3\ p1r\ \
Cos[theta1]\ Cos[theta3] - 2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] - 2\ \
m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] - 2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3])\)\^2 - 4\ \((\(-2\)\ m1\^2\ m2\ m3\ qvalue - 
                            2\ m1\ m2\^2\ m3\ qvalue + 
                            2\ m1\ m2\ m3\^2\ qvalue - 
                            m2\^2\ m3\ p1r\^2\ Cos[theta1]\^2 + 
                            m2\ m3\^2\ p1r\^2\ Cos[theta1]\^2 + 
                            2\ m1\ m2\ m3\ p1r\ p2r\ Cos[theta1]\ Cos[
                                theta2] - m1\^2\ m3\ p2r\^2\ Cos[theta2]\^2 + 
                            m1\ m3\^2\ p2r\^2\ Cos[theta2]\^2 - 
                            m2\^2\ m3\ p1r\^2\ Sin[theta1]\^2 + 
                            m2\ m3\^2\ p1r\^2\ Sin[theta1]\^2 + 
                            2\ m1\ m2\ m3\ p1r\ p2r\ Sin[theta1]\ Sin[
                                theta2] - m1\^2\ m3\ p2r\^2\ Sin[theta2]\^2 + 
                            m1\ m3\^2\ p2r\^2\ Sin[theta2]\^2)\)\ \((m1\^2\ \
m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + 
                            m1\ m2\^2\ Sin[theta3]\^2)\))\))\)/\((2\ \((m1\^2\
\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + 
                    m1\^2\ m2\ Sin[theta3]\^2 + 
                    m1\ m2\^2\ Sin[theta3]\^2)\))\), 
        m4 \[Rule] m1 + m2 - m3}, {p4x \[Rule] 
          p1r\ Cos[theta1] + 
            p2r\ Cos[
                theta2] - \(m1\ m2\ m3\ p1r\ Cos[theta1]\ \
Cos[theta3]\^2\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ \
Cos[theta2]\ Cos[theta3]\^2\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ \
Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \
\(m1\ m2\ m3\ p1r\ Cos[theta3]\ Sin[theta1]\ Sin[theta3]\)\/\(m1\^2\ m2\ \
Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ \
m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ Cos[theta3]\ Sin[theta2]\ \
Sin[theta3]\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) + \((Cos[
                    theta3]\ \[Sqrt]\((\((\(-2\)\ m1\ m2\ m3\ p1r\ \
Cos[theta1]\ Cos[theta3] - 2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] - 2\ \
m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] - 2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3])\)\^2 - 4\ \((\(-2\)\ m1\^2\ m2\ m3\ qvalue - 
                              2\ m1\ m2\^2\ m3\ qvalue + 
                              2\ m1\ m2\ m3\^2\ qvalue - 
                              m2\^2\ m3\ p1r\^2\ Cos[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Cos[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Cos[theta1]\ Cos[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Cos[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Cos[theta2]\^2 - 
                              m2\^2\ m3\ p1r\^2\ Sin[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Sin[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Sin[theta1]\ Sin[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Sin[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Sin[theta2]\^2)\)\ \((m1\^2\ \
m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + 
                              m1\ m2\^2\ Sin[theta3]\^2)\))\))\)/\((2\ \
\((m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + 
                      m1\^2\ m2\ Sin[theta3]\^2 + 
                      m1\ m2\^2\ Sin[theta3]\^2)\))\), 
        p4y \[Rule] 
          p1r\ Sin[theta1] + 
            p2r\ Sin[
                theta2] - \(m1\ m2\ m3\ p1r\ Cos[theta1]\ Cos[theta3]\ \
Sin[theta3]\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ \
Cos[theta2]\ Cos[theta3]\ Sin[theta3]\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\
\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) \
- \(m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3]\^2\)\/\(m1\^2\ m2\ \
Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + m1\ \
m2\^2\ Sin[theta3]\^2\) - \(m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3]\^2\)\/\(m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + \
m1\^2\ m2\ Sin[theta3]\^2 + m1\ m2\^2\ Sin[theta3]\^2\) + \((Sin[
                    theta3]\ \[Sqrt]\((\((\(-2\)\ m1\ m2\ m3\ p1r\ \
Cos[theta1]\ Cos[theta3] - 2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] - 2\ \
m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] - 2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3])\)\^2 - 4\ \((\(-2\)\ m1\^2\ m2\ m3\ qvalue - 
                              2\ m1\ m2\^2\ m3\ qvalue + 
                              2\ m1\ m2\ m3\^2\ qvalue - 
                              m2\^2\ m3\ p1r\^2\ Cos[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Cos[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Cos[theta1]\ Cos[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Cos[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Cos[theta2]\^2 - 
                              m2\^2\ m3\ p1r\^2\ Sin[theta1]\^2 + 
                              m2\ m3\^2\ p1r\^2\ Sin[theta1]\^2 + 
                              2\ m1\ m2\ m3\ p1r\ p2r\ Sin[theta1]\ Sin[
                                  theta2] - 
                              m1\^2\ m3\ p2r\^2\ Sin[theta2]\^2 + 
                              m1\ m3\^2\ p2r\^2\ Sin[theta2]\^2)\)\ \((m1\^2\ \
m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + 
                              m1\ m2\^2\ Sin[theta3]\^2)\))\))\)/\((2\ \
\((m1\^2\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + 
                      m1\^2\ m2\ Sin[theta3]\^2 + 
                      m1\ m2\^2\ Sin[theta3]\^2)\))\), 
        p3r \[Rule] \((2\ m1\ m2\ m3\ p1r\ Cos[theta1]\ Cos[theta3] + 
                2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] + 
                2\ m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] + 
                2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ Sin[
                    theta3] - \[Sqrt]\((\((\(-2\)\ m1\ m2\ m3\ p1r\ \
Cos[theta1]\ Cos[theta3] - 2\ m1\ m2\ m3\ p2r\ Cos[theta2]\ Cos[theta3] - 2\ \
m1\ m2\ m3\ p1r\ Sin[theta1]\ Sin[theta3] - 2\ m1\ m2\ m3\ p2r\ Sin[theta2]\ \
Sin[theta3])\)\^2 - 4\ \((\(-2\)\ m1\^2\ m2\ m3\ qvalue - 
                            2\ m1\ m2\^2\ m3\ qvalue + 
                            2\ m1\ m2\ m3\^2\ qvalue - 
                            m2\^2\ m3\ p1r\^2\ Cos[theta1]\^2 + 
                            m2\ m3\^2\ p1r\^2\ Cos[theta1]\^2 + 
                            2\ m1\ m2\ m3\ p1r\ p2r\ Cos[theta1]\ Cos[
                                theta2] - m1\^2\ m3\ p2r\^2\ Cos[theta2]\^2 + 
                            m1\ m3\^2\ p2r\^2\ Cos[theta2]\^2 - 
                            m2\^2\ m3\ p1r\^2\ Sin[theta1]\^2 + 
                            m2\ m3\^2\ p1r\^2\ Sin[theta1]\^2 + 
                            2\ m1\ m2\ m3\ p1r\ p2r\ Sin[theta1]\ Sin[
                                theta2] - m1\^2\ m3\ p2r\^2\ Sin[theta2]\^2 + 
                            m1\ m3\^2\ p2r\^2\ Sin[theta2]\^2)\)\ \((m1\^2\ \
m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + m1\^2\ m2\ Sin[theta3]\^2 + 
                            m1\ m2\^2\ Sin[theta3]\^2)\))\))\)/\((2\ \((m1\^2\
\ m2\ Cos[theta3]\^2 + m1\ m2\^2\ Cos[theta3]\^2 + 
                    m1\^2\ m2\ Sin[theta3]\^2 + 
                    m1\ m2\^2\ Sin[theta3]\^2)\))\), 
        m4 \[Rule] m1 + m2 - m3}}\)], "Output"]
}, Closed]],

Cell["\<\
Familiar special cases can be examined by choosing appropriate parameters.  \
For example, consider the elastic scattering of two particles of equal mass \
assuming that the target is at rest.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\((\(gensol\  /. {m1 \[Rule] m, m2 \[Rule] m, m3 \[Rule] m, 
                qvalue \[Rule] 0, theta1 \[Rule] 0, p2r \[Rule] 0}\  // 
            Simplify\) // PowerExpand)\)\  // Simplify\)], "Input"],

Cell[BoxData[
    \({{p4x \[Rule] p1r\ Sin[theta3]\^2, 
        p4y \[Rule] \(-p1r\)\ Cos[theta3]\ Sin[theta3], 
        p3r \[Rule] p1r\ Cos[theta3], m4 \[Rule] m}, {p4x \[Rule] p1r, 
        p4y \[Rule] 0, p3r \[Rule] 0, m4 \[Rule] m}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Clearly the first solution is the one we seek while the second solution is \
degenerate, describing only the special case of a head-on collision in which \
the projectile transfers all of its momentum to the struck particle.  More \
general situations can now be investigated by varying the parameters in the \
first solution.  A detailed exploration of these possibilities is beyond the \
scope of this notebook, but can be found in ",
  ButtonBox["collisions.nb",
    ButtonData:>{"Applications/collisions.nb", None},
    ButtonStyle->"Hyperlink"],
  ", which also develops functions which produce a variety of useful vector \
diagrams."
}], "Text"]
}, Closed]],

Cell[CellGroupData[{

Cell["Example: entropy for one-dimensional Ising model", "Subsubsection"],

Cell[TextData[{
  "The following equations arise in the one-dimensional Ising model of \
magnetic systems.  (Don't worry about the physics, we just need a nontrivial \
system of equations for illustration of ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions.)"
}], "Text"],

Cell[BoxData[
    \(Needs["\<Utilities`Notation`\>"]\)], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(N\_\[UpArrow]\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(N\_\[DownArrow]\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(N\_\[UpDownArrow]\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "\n", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(N\_T\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Lambda]\_1\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Lambda]\_2\),
          NotationBoxTag,
          Editable->True], "]"}]}]], "Input"],

Cell[BoxData[
    \(\(eqs\  = \ {0 \[Equal] \[Lambda]\_2 - B\ \[Lambda]\_1\ \[Mu] + 
              Log[N\_\[UpArrow]] - Log[N\_\[UpArrow] - N\_\[UpDownArrow]\/2], 
          0 \[Equal] \[Lambda]\_2 + B\ \[Lambda]\_1\ \[Mu] + 
              Log[N\_\[DownArrow]] - 
              Log[N\_\[DownArrow] - N\_\[UpDownArrow]\/2], 
          0 \[Equal] 
            2\ J\ \[Lambda]\_1 + 
              1\/2\ \((Log[4] + Log[N\_\[DownArrow] - N\_\[UpDownArrow]\/2] + 
                    Log[N\_\[UpArrow] - N\_\[UpDownArrow]\/2] - 
                    2\ Log[N\_\[UpDownArrow]])\)\ , 
          N\_T\  \[Equal] \ 
            N\_\[UpArrow]\  + \ N\_\[DownArrow]}\ ;\)\)], "Input"],

Cell[TextData[{
  "Obviously, we can eliminate ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda]\_2\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`N\_\[DownArrow]\)]],
  " immediately. "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(newEqs = 
      Eliminate[eqs, {\[Lambda]\_2, N\_\[DownArrow]}] // Simplify\)], "Input"],

Cell[BoxData[
    \(N\_T \[Equal] \(N\_\[UpArrow]\) + \(\[ExponentialE]\^\(\(-2\)\ \
\[Lambda]\_1\ \((2\ J + B\ \[Mu])\)\)\ \(N\_\[UpArrow]\)\ \
N\_\[UpDownArrow]\%2\)\/\((\(-2\)\ \(N\_\[UpArrow]\) + \(N\_\[UpDownArrow]\))\
\)\^2 && 4\ N\_T \[Equal] 
        4\ \(N\_\[UpArrow]\) + \(N\_\[UpDownArrow]\)\ \((2 + \
\(\[ExponentialE]\^\(\(-4\)\ J\ \[Lambda]\_1\)\ \
\(N\_\[UpDownArrow]\)\)\/\(\(N\_\[UpArrow]\) - \
\(N\_\[UpDownArrow]\)\/2\))\)\)], "Output"]
}, Closed]],

Cell["\<\
We find that there are two basic quantities for which substitutions can be \
used to help simplify the algebra.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(newEqs = 
      newEqs\  //. {Exp[
                a_.  \((B\ \[Mu] + 
                      b_. )\)] \[Rule] \(\[Gamma]\^\(\(-a\)/\((2\ \
\[Lambda]\_1)\)\)\) Exp[a\ b], 
            Exp[a_. \ \((J\  + \ 
                      b_. )\)\ ] \[Rule] \(\[CurlyPhi]\^\(\(-a\)/\((4  \
\[Lambda]\_1)\)\)\) Exp[a\ b]} // Simplify\)], "Input"],

Cell[BoxData[
    \(N\_T \[Equal] \(N\_\[UpArrow]\) + \(\(N\_\[UpArrow]\)\ \
N\_\[UpDownArrow]\%2\ \[Gamma]\ \[CurlyPhi]\)\/\((\(-2\)\ \(N\_\[UpArrow]\) + \
\(N\_\[UpDownArrow]\))\)\^2 && 
      4\ N\_T \[Equal] 
        4\ \(N\_\[UpArrow]\) + \(N\_\[UpDownArrow]\)\ \((2 + \(\(N\_\
\[UpDownArrow]\)\ \[CurlyPhi]\)\/\(\(N\_\[UpArrow]\) - \
\(N\_\[UpDownArrow]\)\/2\))\)\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " can now produce relatively simple solutions to these equations. (I \
preferred the expressions produced by version  4, but these are equivalent.)"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol = 
      Solve[newEqs, {N\_\[UpArrow], N\_\[UpDownArrow]}] // 
        FullSimplify[#, {\[Gamma] > 0, \[CurlyPhi] > 0, 
              N\_T > 0}] &\)], "Input"],

Cell[BoxData[
    \({{\(N\_\[UpArrow]\) \[Rule] N\_T, \(N\_\[UpDownArrow]\) \[Rule] 
          0}, {\(N\_\[UpArrow]\) \[Rule] \(N\_T\ \((\[CurlyPhi] + \@\(\
\[CurlyPhi]\ \((4\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi])\)\) + \
\[Gamma]\ \((4 + \((\(-2\) + \[Gamma])\)\ \[CurlyPhi] - \@\(\[CurlyPhi]\ \((4\
\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi])\)\))\))\)\)\/\(8\ \
\[Gamma] + 2\ \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi]\), \
\(N\_\[UpDownArrow]\) \[Rule] \(N\_T\ \((\(-4\)\ \[Gamma] - \[CurlyPhi] + 2\ \
\[Gamma]\ \[CurlyPhi] - \[Gamma]\^2\ \[CurlyPhi] - \((1 + \[Gamma])\)\ \@\(\
\[CurlyPhi]\ \((4\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi])\)\))\)\
\)\/\(\((\(-1\) + \[CurlyPhi])\)\ \((4\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\
\ \[CurlyPhi])\)\)}, {\(N\_\[UpArrow]\) \[Rule] \(N\_T\ \((\[CurlyPhi] - \@\(\
\[CurlyPhi]\ \((4\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi])\)\) + \
\[Gamma]\ \((4 + \((\(-2\) + \[Gamma])\)\ \[CurlyPhi] + \@\(\[CurlyPhi]\ \((4\
\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi])\)\))\))\)\)\/\(8\ \
\[Gamma] + 2\ \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi]\), \
\(N\_\[UpDownArrow]\) \[Rule] \(N\_T\ \((\(-4\)\ \[Gamma] - \[CurlyPhi] + 2\ \
\[Gamma]\ \[CurlyPhi] - \[Gamma]\^2\ \[CurlyPhi] + \((1 + \[Gamma])\)\ \@\(\
\[CurlyPhi]\ \((4\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\ \[CurlyPhi])\)\))\)\
\)\/\(\((\(-1\) + \[CurlyPhi])\)\ \((4\ \[Gamma] + \((\(-1\) + \[Gamma])\)\^2\
\ \[CurlyPhi])\)\)}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "At this point we would need to use physical arguments to select the \
appropriate solution, but that is beyond the scope of this course.  \
Nevertheless, it should now be more clear how some of these ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions can be used to simplify and solve systems of equations."
}], "Text"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Numerical solution of equations", "Section"],

Cell[CellGroupData[{

Cell["Polynomial equations", "Subsection"],

Cell[TextData[{
  "An equation or system of equations involving only polynomials can be \
solved numerically using ",
  ButtonBox["NSolve",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[eqs,vars,n]", "InlineInput",
    FontWeight->"Bold"],
  " where ",
  StyleBox["eqs", "InlineInput",
    FontWeight->"Bold"],
  " is an equation or list of equations, ",
  StyleBox["vars", "InlineInput",
    FontWeight->"Bold"],
  " is the variable or list of variables, and ",
  StyleBox["n", "InlineInput"],
  " is an optional parameter specifying the precision sought in terms of \
number of decimal digits.  The result is returned as a list of sets of \
replacement rules."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(NSolve[x\^7 - 3  x\^3 \[Equal] x - 2, x]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(-1.4293737399776678`\)}, {x \[Rule] \
\(-0.4258430014826768`\) - 
            0.8122523707186996`\ \[ImaginaryI]}, {x \[Rule] \
\(-0.4258430014826768`\) + 
            0.8122523707186996`\ \[ImaginaryI]}, {x \[Rule] \
\(\(0.11535124648122039`\)\(\[InvisibleSpace]\)\) - 
            1.2924361744793416`\ \[ImaginaryI]}, {x \[Rule] \
\(\(0.11535124648122039`\)\(\[InvisibleSpace]\)\) + 
            1.2924361744793416`\ \[ImaginaryI]}, {x \[Rule] 
          0.7742779485094532`}, {x \[Rule] 1.2760793014711274`}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(NSolve[{\(x\^2\) y - x \[Equal] 1, y\^2 + x\ y \[Equal] \(-2\)}, {x, 
        y}]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] \(\(0.4019802320146253`\)\(\[InvisibleSpace]\)\) - 
            0.8357928299565638`\ \[ImaginaryI], 
        y \[Rule] \(-0.25843960059643717`\) + 
            1.879929530266246`\ \[ImaginaryI]}, {x \[Rule] \
\(\(0.4019802320146253`\)\(\[InvisibleSpace]\)\) + 
            0.8357928299565638`\ \[ImaginaryI], 
        y \[Rule] \(-0.25843960059643717`\) - 
            1.879929530266246`\ \[ImaginaryI]}, {x \[Rule] \
\(-0.5686468986812908`\) + 0.25332888797245895`\ \[ImaginaryI], 
        y \[Rule] \(\(0.25843960059643734`\)\(\[InvisibleSpace]\)\) + 
            1.2646924588791695`\ \[ImaginaryI]}, {x \[Rule] \
\(-0.5686468986812908`\) - 0.25332888797245895`\ \[ImaginaryI], 
        y \[Rule] \(\(0.25843960059643734`\)\(\[InvisibleSpace]\)\) - 
            1.2646924588791695`\ \[ImaginaryI]}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  StyleBox["NSolve", "InlineInput",
    FontWeight->"Bold"],
  " will also handle equations which are really polynomials in disguise"
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(NSolve[{\@\(\(x\^2\) y\  - \ \(y\^2\) x\) \[Equal] \@\(1 + x\ y\)\%4, 
        x \@\( x\  - \ y\) \[Equal] 2}, {x, y}]\)], "Input"],

Cell[BoxData[
    \({{x \[Rule] 1.862510719108471`, 
        y \[Rule] 0.7094227800529718`}, {x \[Rule] 3.6937618996312302`, 
        y \[Rule] 3.4005900974776084`}}\)], "Output"]
}, Closed]],

Cell["and many others for which inverse functions are available", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(NSolve[Cos[x]\^2 - Sin[x]\^2 \[Equal] 0.5, x]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"ifun"\), \(\(:\)\(\ \)\), "\<\"Inverse functions are \
being used by \\!\\(Solve\\), so some solutions may not be found; use Reduce \
for complete solution information. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({{x \[Rule] \(-2.6179938779914944`\)}, {x \[Rule] \
\(-0.5235987755982989`\)}, {x \[Rule] 0.5235987755982989`}, {x \[Rule] 
          2.6179938779914944`}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(NSolve[\[ExponentialE]\^\(2  x\) - \[ExponentialE]\^x \[Equal] 1, 
      x]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"ifun"\), \(\(:\)\(\ \)\), "\<\"Inverse functions are \
being used by \\!\\(Solve\\), so some solutions may not be found; use Reduce \
for complete solution information. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::ifun\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \({{x \[Rule] \(-0.48121182505960347`\) + 
            3.141592653589793`\ \[ImaginaryI]}, {x \[Rule] 
          0.48121182505960347`}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The warning message about inverse functions is usually nothing to worry \
about, and can be disabled using ",
  StyleBox["Off", "InlineInput",
    FontWeight->"Bold"],
  ", but after solving a problem one should go back and determine whether \
alternative solutions have been overlooked."
}], "Text"],

Cell[TextData[{
  "However, ",
  StyleBox["NSolve", "InlineInput",
    FontWeight->"Bold"],
  " is unable to solve equations which involve transcendental functions in an \
essentially nonalgebraic manner that makes transformation to polynomial form \
impossible."
}], "Text"],

Cell[BoxData[{
    \(lhs\  = \ \[Xi]\/\@\(1 - \ \[Xi]\^2\); \ \ rhs\  = \ 
      Cot[\[Xi]];\), "\n", 
    \(\(eq1 = \ lhs \[Equal] rhs;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(NSolve[eq1, \[Xi]]\)], "Input"],

Cell[BoxData[
    RowBox[{\(Solve::"tdep"\), \(\(:\)\(\ \)\), "\<\"The equations appear to \
involve the variables to be solved for in an essentially non-algebraic way. \
\\!\\(\\*ButtonBox[\\\"More\[Ellipsis]\\\", ButtonStyle->\\\"RefGuideLinkText\
\\\", ButtonFrame->None, ButtonData:>\\\"Solve::tdep\\\"]\\)\"\>"}]], \
"Message"],

Cell[BoxData[
    \(NSolve[\[Xi]\/\@\(1 - \[Xi]\^2\) \[Equal] 
        Cot[\[Xi]], \[Xi]]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Write a function using NSolve and Cases which returns the real solutions \
to  the following equation in terms of the parameter ",
  StyleBox["a",
    FontSlant->"Italic"],
  " :\n\t\t",
  Cell[BoxData[
      \(TraditionalForm\`2 == r\^4 + \(a\ r\^4\)\/\((1 + r)\)\^4\)]],
  "\nPlot these solutions for {a,-1,1}.  [Hint: it is probably easiest to use \
ListPlot for this purpose.]"
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Transcendental equations", "Subsection"],

Cell[TextData[{
  "The more general function ",
  ButtonBox["FindRoot",
    ButtonStyle->"RefGuideLink"],
  " can be used to find numerical solutions to equations or systems of \
equations given appropriate starting conditions.  For a single equation, ",
  StyleBox["FindRoot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["eq", "InlineInput"],
  StyleBox[",", "InlineInput",
    FontSlant->"Italic"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`{x, x\_0}\)], "InlineInput"], "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " uses a variant of the secant method to search for a solution in the \
vicinity of ",
  Cell[BoxData[
      \(TraditionalForm\`x\_0\)]],
  " while ",
  StyleBox["FindRoot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["eq,", "InlineInput",
    FontSlant->"Italic"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`{x, {x\_0, x\_1}}\)], "InlineInput"], 
    "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " uses Newton's method based upon two starting values ",
  Cell[BoxData[
      \(TraditionalForm\`{x\_0, x\_1}\)]],
  ".  The secant method can be used when symbolic derivatives are possible, \
but Newton's method must be used otherwise.  Also note that ",
  StyleBox["FindRoot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr", "InlineInput"],
  StyleBox[",", "InlineInput",
    FontSlant->"Italic"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`{x, x\_0}\)], "InlineInput"], "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  " where ",
  StyleBox["expr",
    FontSlant->"Italic"],
  " is an expression rather than an equation is equivalent to ",
  StyleBox["FindRoot[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["expr\[Equal]0", "InlineInput"],
  StyleBox[",", "InlineInput",
    FontSlant->"Italic"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`{x, x\_0}\)], "InlineInput"], "InlineInput"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  ". "
}], "Text"],

Cell[TextData[{
  "For the simple equation below we recognize by inspection that there is \
only one root for ",
  Cell[BoxData[
      \(TraditionalForm\`y > 0\)]],
  " and, hence, need not be too careful in selecting the starting value for \
the secant method."
}], "Text"],

Cell[BoxData[
    \(\(eq1\  = \ y\ Tanh[y] \[Equal] 1;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(sol1\  = \ FindRoot[eq1, {y, 0.1}]\)], "Input"],

Cell[BoxData[
    \({y \[Rule] 1.1996786402577337`}\)], "Output"]
}, Closed]],

Cell["\<\
To test the accuracy of the solution, we change the equation to a difference \
between its two sides and then substitute the solution.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(eq1 /. Equal \[Rule] Subtract\) /. sol1\)], "Input"],

Cell[BoxData[
    \(\(-1.1102230246251565`*^-16\)\)], "Output"]
}, Closed]],

Cell["\<\
Even though our starting value is poor, the accuracy of the solution is quite \
good; note that the solution contains more significant figures than are \
routinely printed.  \
\>", "Text"],

Cell["\<\
For more complicated equations with several roots, we should compare the \
left-hand and right-hand sides graphically to choose the appropriate starting \
conditions.\
\>", "Text"],

Cell[BoxData[{
    \(lhs\  = \ \[Xi]\/\@\(16 - \ \[Xi]\^2\); \ \ rhs\  = \ 
      Cot[\[Xi]];\), "\n", 
    \(\(eq1 = \ lhs \[Equal] rhs;\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(Plot[Evaluate[{lhs, rhs}], {\[Xi], 0, 4}, 
        PlotRange \[Rule] {Automatic, {\(-5\), 5}}];\)\)], "Input"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.238095 0.309017 0.0618034 [
[.2619 .29652 -3 -9 ]
[.2619 .29652 3 0 ]
[.5 .29652 -3 -9 ]
[.5 .29652 3 0 ]
[.7381 .29652 -3 -9 ]
[.7381 .29652 3 0 ]
[.97619 .29652 -3 -9 ]
[.97619 .29652 3 0 ]
[.01131 .0618 -12 -4.5 ]
[.01131 .0618 0 4.5 ]
[.01131 .18541 -12 -4.5 ]
[.01131 .18541 0 4.5 ]
[.01131 .43262 -6 -4.5 ]
[.01131 .43262 0 4.5 ]
[.01131 .55623 -6 -4.5 ]
[.01131 .55623 0 4.5 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.2619 .30902 m
.2619 .31527 L
s
[(1)] .2619 .29652 0 1 Mshowa
.5 .30902 m
.5 .31527 L
s
[(2)] .5 .29652 0 1 Mshowa
.7381 .30902 m
.7381 .31527 L
s
[(3)] .7381 .29652 0 1 Mshowa
.97619 .30902 m
.97619 .31527 L
s
[(4)] .97619 .29652 0 1 Mshowa
.125 Mabswid
.07143 .30902 m
.07143 .31277 L
s
.11905 .30902 m
.11905 .31277 L
s
.16667 .30902 m
.16667 .31277 L
s
.21429 .30902 m
.21429 .31277 L
s
.30952 .30902 m
.30952 .31277 L
s
.35714 .30902 m
.35714 .31277 L
s
.40476 .30902 m
.40476 .31277 L
s
.45238 .30902 m
.45238 .31277 L
s
.54762 .30902 m
.54762 .31277 L
s
.59524 .30902 m
.59524 .31277 L
s
.64286 .30902 m
.64286 .31277 L
s
.69048 .30902 m
.69048 .31277 L
s
.78571 .30902 m
.78571 .31277 L
s
.83333 .30902 m
.83333 .31277 L
s
.88095 .30902 m
.88095 .31277 L
s
.92857 .30902 m
.92857 .31277 L
s
.25 Mabswid
0 .30902 m
1 .30902 L
s
.02381 .0618 m
.03006 .0618 L
s
[(-4)] .01131 .0618 1 0 Mshowa
.02381 .18541 m
.03006 .18541 L
s
[(-2)] .01131 .18541 1 0 Mshowa
.02381 .43262 m
.03006 .43262 L
s
[(2)] .01131 .43262 1 0 Mshowa
.02381 .55623 m
.03006 .55623 L
s
[(4)] .01131 .55623 1 0 Mshowa
.125 Mabswid
.02381 .09271 m
.02756 .09271 L
s
.02381 .12361 m
.02756 .12361 L
s
.02381 .15451 m
.02756 .15451 L
s
.02381 .21631 m
.02756 .21631 L
s
.02381 .24721 m
.02756 .24721 L
s
.02381 .27812 m
.02756 .27812 L
s
.02381 .33992 m
.02756 .33992 L
s
.02381 .37082 m
.02756 .37082 L
s
.02381 .40172 m
.02756 .40172 L
s
.02381 .46353 m
.02756 .46353 L
s
.02381 .49443 m
.02756 .49443 L
s
.02381 .52533 m
.02756 .52533 L
s
.02381 .0309 m
.02756 .0309 L
s
.02381 .58713 m
.02756 .58713 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.5 Mabswid
.02381 .30902 m
.06244 .31153 L
.10458 .31428 L
.14415 .31689 L
.18221 .31944 L
.22272 .32222 L
.26171 .32496 L
.30316 .32798 L
.34309 .33101 L
.3815 .33406 L
.42237 .33749 L
.46172 .34102 L
.49955 .34465 L
.53984 .34886 L
.57861 .35331 L
.61984 .35861 L
.65954 .36442 L
.69774 .37091 L
.73838 .37916 L
.77751 .38903 L
.79749 .39511 L
.81909 .40282 L
.85916 .42189 L
.8791 .43519 L
.88927 .44357 L
.90016 .45427 L
.91044 .4666 L
.9198 .48054 L
.92922 .49844 L
.93814 .52103 L
.94269 .53584 L
.94772 .55607 L
.95297 .58376 L
.95537 .59974 L
s
.95537 .59974 m
.95768 .61803 L
s
.07094 .61803 m
.07287 .60472 L
.07824 .57463 L
.08414 .54768 L
.09408 .51231 L
.10458 .48416 L
.12409 .44698 L
.13503 .43156 L
.145 .41976 L
.1657 .40015 L
.18485 .38601 L
.22318 .36471 L
.26245 .3485 L
.3002 .33588 L
.3404 .32421 L
.37909 .31389 L
.42023 .30318 L
.45985 .29254 L
.49796 .28137 L
.53852 .26755 L
.57757 .25086 L
.59748 .24026 L
.61906 .22627 L
.63987 .20915 L
.65904 .18843 L
.67781 .16069 L
.68847 .13971 L
.69845 .11482 L
.70886 .08074 L
.71874 .03636 L
s
.71874 .03636 m
.72445 0 L
s
.77066 0 m
.77066 .61803 L
s
.81904 .61803 m
.82695 .57107 L
.83229 .54706 L
.83714 .52856 L
.8555 .47754 L
.86547 .45794 L
.87466 .44307 L
.89545 .41713 L
.91419 .39974 L
.93482 .38472 L
.97513 .36288 L
.97619 .3624 L
s
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{353.125, 218.25},
  ImageMargins->{{16, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40008A0001Ab000`400?l00000o`00003oo`3ooooo
0?oooa<0oooo003o0?ooool0oooo4`3oool00?l0ooooo`3ooolC0?ooo`00o`3ooooo0?oooa<0oooo
000H0?ooo`030000003oool0oooo0?l0ooooJ03oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooJ03oool00`000000oooo0?ooo`0E
0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooJ03oool00`00
0000oooo0?ooo`0E0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0
ooooI`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000
003oool0oooo0?l0ooooI`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo07@0oooo
000H0?ooo`030000003oool0oooo0?l0ooooI`3oool00`000000oooo0?ooo`0F0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooI`3oool00`000000oooo0?ooo`0F
0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooI`3oool00`00
0000oooo0?ooo`0F0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0
ooooI`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000
003oool0oooo0?l0ooooIP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo07@0oooo
000H0?ooo`030000003oool0oooo0?l0ooooIP3oool00`000000oooo0?ooo`0G0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooIP3oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooIP3oool00`00
0000oooo0?ooo`0G0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0
ooooIP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000
003oool0oooo0?l0ooooIP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo07@0oooo
000H0?ooo`030000003oool0oooo0?l0ooooI@3oool00`000000oooo0?ooo`0H0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`<00000o`3ooomU0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomU0?ooo`030000003oool0oooo
01P0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomU0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0
o`3ooomU0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0M03oool001P0oooo00<0
00000?ooo`3oool0o`3ooomU0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0M03o
ool001P0oooo00<000000?ooo`3oool0o`3ooomT0?ooo`030000003oool0oooo01T0oooo00<00000
0?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomT0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomT0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0
o`3ooomT0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0M03oool001P0oooo00<0
00000?ooo`3oool0o`3ooomS0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0M03o
ool001P0oooo00<000000?ooo`3oool0o`3ooomS0?ooo`030000003oool0oooo01X0oooo00<00000
0?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomS0?ooo`030000003oool0oooo
01X0oooo00<000000?ooo`3oool0M03oool000`0oooo00D0omnL04P000000000000007Boo`070?oo
o`030000003oool0oooo0?l0ooooH`3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
07@0oooo000>0?ooo`030;md001d_ol0oooo00L0oooo00<000000?ooo`3oool0o`3ooomS0?ooo`03
0000003oool0oooo01X0oooo00<000000?ooo`3oool0M03oool000h0oooo00<0_g@007Boo`3oool0
1`3oool00`000000oooo0?ooo`3o0?ooof80oooo00<000000?ooo`3oool06`3oool00`000000oooo
0?ooo`1d0?ooo`002P3oool00`3OW4P00000000000030000000307Boo`3oool0oooo00D0oooo00<0
00000?ooo`3oool0o`3ooomR0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0M03o
ool000030?ooo`2LB000000000@0000000T004RL0?ooo`3oool0okmd0000M02oool0oooo0;md001d
_ol0203oool400000?l0ooooH@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo07@0
oooo000;0?ooo`050=nLB000M;l0oooo0;md001d_ol0203oool00`000000oooo0?ooo`3o0?ooof80
oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1d0?ooo`00303oool0102LB000B9cO
0;md001d_ol80?ooo`030000003oool0oooo0?l0ooooHP3oool00`000000oooo0?ooo`0K0?ooo`03
0000003oool0oooo07@0oooo000<0?ooo`040?oOW01804P0B4P007Boo`P0oooo00<000000?ooo`3o
ool0o`3ooomQ0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0M03oool000d0oooo
00<0okmd0000001d_ol0203oool00`000000oooo0?ooo`3o0?ooof40oooo00<000000?ooo`3oool0
703oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooof40oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o
0?ooof40oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`1d0?ooo`00603oool00`00
0000oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`1d0?oo
o`00603oool00`000000oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool07@3oool00`000000
oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooof00oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o
0?oooel0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`00
0000oooo0?ooo`3o0?oooel0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`1d0?oo
o`00603oool00`000000oooo0?ooo`3o0?oooel0oooo00<000000?ooo`3oool07P3oool00`000000
oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oooeh0oooo00<000000?ooo`3oool0
7`3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oooeh0oooo00<0
00000?ooo`3oool07`3oool00`000000oooo0?ooo`1d0?ooo`00603oool300000?l0ooooGP3oool0
0`000000oooo0?ooo`0O0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo
0?l0ooooG@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo07@0oooo000H0?ooo`03
0000003oool0oooo0?l0ooooG@3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo07@0
oooo000H0?ooo`030000003oool0oooo0?l0ooooG@3oool00`000000oooo0?ooo`0P0?ooo`030000
003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooG03oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooG03oool0
0`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo
0?l0ooooG03oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo07@0oooo000H0?ooo`03
0000003oool0oooo0?l0ooooF`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo07@0
oooo000H0?ooo`030000003oool0oooo0?l0ooooF`3oool00`000000oooo0?ooo`0R0?ooo`030000
003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooF`3oool00`000000oooo0?oo
o`0R0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooFP3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo
0?l0ooooFP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo07@0oooo000H0?ooo`03
0000003oool0oooo0?l0ooooFP3oool00`000000oooo0?ooo`0S0?ooo`030000003oool0oooo07@0
oooo000H0?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?ooo`0T0?ooo`030000
003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000oooo0?oo
o`0T0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0ooooF03oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo07@0oooo000H0?ooo`<00000o`3ooomH0?oo
o`030000003oool0oooo02D0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3o
ool0o`3ooomH0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0M03oool001P0oooo
00<000000?ooo`3oool0o`3ooomG0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0
M03oool001P0oooo00<000000?ooo`3oool0o`3ooomG0?ooo`030000003oool0oooo02H0oooo00<0
00000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomG0?ooo`030000003oool0
oooo02H0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomF0?oo
o`030000003oool0oooo02L0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3o
ool0o`3ooomF0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0M03oool001P0oooo
00<000000?ooo`3oool0o`3ooomE0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0
M03oool001P0oooo00<000000?ooo`3oool0o`3ooomE0?ooo`030000003oool0oooo02P0oooo00<0
00000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomE0?ooo`030000003oool0
oooo02P0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomD0?oo
o`030000003oool0oooo02T0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3o
ool0o`3ooomD0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0M03oool001P0oooo
00<000000?ooo`3oool0o`3ooomC0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0
M03oool001P0oooo00<000000?ooo`3oool0o`3ooomC0?ooo`030000003oool0oooo02X0oooo00<0
00000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomB0?ooo`030000003oool0
oooo02/0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooomB0?oo
o`030000003oool0oooo02/0oooo00<000000?ooo`3oool0M03oool001P0oooo0`00003o0?oooe40
oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo
0?ooo`3o0?oooe40oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`1d0?ooo`00603o
ool00`000000oooo0?ooo`3o0?oooe00oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oooe00oooo00<000000?ooo`3oool0;@3oool0
0`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooodl0oooo00<000000?oo
o`3oool0;P3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooodl0
oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo
0?ooo`3o0?ooodh0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`1d0?ooo`00603o
ool00`000000oooo0?ooo`3o0?ooodd0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?oo
o`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oood`0oooo00<000000?ooo`3oool0<@3oool0
0`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oood`0oooo00<000000?oo
o`3oool0<@3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oood/0
oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo
0?ooo`3o0?ooodX0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`1d0?ooo`002@3o
ool00`3ookl0M000000000040000000307Boo`3oool0oooo00D0oooo00<000000?ooo`3oool0o`3o
oom:0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0M03oool000X0oooo00<0_g@0
0000M02oool00P3oool00`3OW7@0W=oo0?ooo`060?ooo`030000003oool0oooo0?l0ooooB@3oool0
0`000000oooo0?ooo`0d0?ooo`030000003oool0oooo07@0oooo000;0?ooo`030?noM00004P0W=oo
00X0oooo00<000000?ooo`3oool0o`3ooom80?ooo`030000003oool0oooo03D0oooo00<000000?oo
o`3oool0M03oool000`0oooo00<0omnL04P00018W=l02@3oool400000?l0ooooAP3oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo07@0oooo00000`3oool0W4P000000004000000030018
W03oool0oooo00@0oooo00<0W4P004RLg`3oool01`3oool00`000000oooo0?ooo`3o0?ooodL0oooo
00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`1d0?ooo`003`3oool00`2LB7@0_ooo0?oo
o`060?ooo`030000003oool0oooo0?l0ooooAP3oool00`000000oooo0?ooo`0g0?ooo`030000003o
ool0oooo07@0oooo000:0?ooo`030=nLB01d_ol0oooo0080oooo00<0_g@007Boo`3oool01P3oool0
0`000000oooo0?ooo`3o0?ooodD0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`1d
0?ooo`002P3oool01`3o_g@0001d0;ooo`3oool0okmd0000B02Lgol01`3oool00`000000oooo0?oo
o`3o0?oood@0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`1d0?ooo`002`3oool0
1@3o_g@000000000000004P0W=oo00P0oooo00<000000?ooo`3oool0o`3ooom30?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooom2
0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?oo
o`3oool0o`3ooom10?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0M03oool001P0
oooo00<000000?ooo`3oool0o`3ooom10?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3o
ool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooom00?ooo`030000003oool0oooo03d0oooo
00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooolo0?ooo`030000003o
ool0oooo03h0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3oooln
0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?oo
o`3oool0o`3ooolm0?ooo`030000003oool0oooo0400oooo00<000000?ooo`3oool0M03oool001P0
oooo00<000000?ooo`3oool0o`3oooll0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3o
ool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooolk0?ooo`030000003oool0oooo0480oooo
00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3oooli0?ooo`800000A@3o
ool00`000000oooo0?ooo`1d0?ooo`00603oool300000?l0oooo>03oool00`000000oooo0?ooo`15
0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo=`3oool00`00
0000oooo0?ooo`160?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0
oooo=P3oool00`000000oooo0?ooo`170?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000
003oool0oooo0?l0oooo=@3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo07@0oooo
000H0?ooo`030000003oool0oooo0?l0oooo<`3oool2000004/0oooo00<000000?ooo`3oool0M03o
ool001P0oooo00<000000?ooo`3oool0o`3ooolb0?ooo`030000003oool0oooo04/0oooo00<00000
0?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooola0?ooo`030000003oool0oooo
04`0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3oool_0?ooo`80
0000C`3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooobh0oooo
00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?oo
o`3o0?ooob`0oooo0P00001B0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0
oooo0?l0oooo:P3oool2000005@0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?oo
o`3oool0o`3ooolY0?ooo`030000003oool0oooo05@0oooo00<000000?ooo`3oool0M03oool001P0
oooo00<000000?ooo`3oool0o`3ooolW0?ooo`800000E`3oool00`000000oooo0?ooo`1d0?ooo`00
603oool00`000000oooo0?ooo`3o0?ooobH0oooo00<000000?ooo`3oool0E`3oool00`000000oooo
0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooob@0oooo0P00001J0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo8P3oool2000005`0oooo00<00000
0?ooo`3oool0M03oool001P0oooo0`00003o0?ooob00oooo0P00001N0?ooo`030000003oool0oooo
07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo7P3oool200000600oooo00<000000?ooo`3o
ool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooolK0?ooo`<00000HP3oool00`000000oooo
0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oooaT0oooo0P00001U0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo5`3oool2000006L0oooo00<00000
0?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3ooolE0?ooo`800000J@3oool00`00
0000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?oooa<0oooo0P00001[0?ooo`03
0000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo403oool3000006d0oooo
00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3oool>0?ooo`800000L03o
ool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooo``0oooo0P00001b
0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo2P3oool20000
07@0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0o`3oool70?ooo`<0
0000MP3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3o0?ooo`@0oooo
0`00001i0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0?l0oooo0@3o
ool3000007`0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0M03oool0
0`3o_g@000000000000300000003001d_`3oool0oooo0740oooo00<0oono07@00000000010000000
0`1d_ol0oooo0?ooo`050?ooo`<00000J`3oool01@2oM0000000000000000000B9cO00l0oooo00<0
00000?ooo`3oool0I@3oool01@3ogi`0B000000000000000M;oo00X0oooo000H0?ooo`030000003o
ool0oooo07L0oooo00<0W4Q809cOo`3oool0M@3oool00`2oM000001d0;ooo`020?ooo`030=nLM02L
gol0oooo00<0oooo0`00001]0?ooo`070;md0000M;l0oooo0?ooo`3ookl0M00004RLg`0>0?ooo`03
0000003oool0oooo06L0oooo00<0_g@007Boo`3oool02P3oool001P0oooo0`00001g0?ooo`0309a8
B02Lgol0oooo07H0oooo00<0okmd0000B02Lgol0103oool3000007D0oooo00<0okmd001d_`3oool0
3@3oool00`000000oooo0?ooo`1W0?ooo`030;md001d_ol0oooo00X0oooo000H0?ooo`030000003o
ool0oooo07L0oooo00<0W4Q809cOo`3oool0M`3oool00`3ogi`0B00004RLg`03000007L0oooo00<0
omnL04P0B02Lgol03P3oool00`000000oooo0?ooo`1S0?ooo`030=nLB0000000000000<0000000<0
M;oo0?ooo`3oool0203oool001P0oooo00<000000?ooo`3oool0M`3oool00`2LB4P0W=oo0?ooo`1g
0?ooo`<0000000<0B9cO0?ooo`3oool0M@3oool0103OW4P000000000B02Lgol?0?ooo`030000003o
ool0oooo06<0oooo00H0okmd0000M02oool0oooo0;md001d_ol;0?ooo`00603oool00`000000oooo
0?ooo`1g0?ooo`0309a8B02Lgol0oooo07@0oooo0`0000030?ooo`0309a8M02oool0oooo07L0oooo
00<0omnL04P0B02Lgol03P3oool00`000000oooo0?ooo`1T0?ooo`050=nLB000M;l0oooo0;md001d
_ol02`3oool001P0oooo00<000000?ooo`3oool0M`3oool00`2LB4P0W=oo0?ooo`1`0?ooo`@00000
00<0oooo0=nLB01d_ol00`3oool00`2oM000M;oo0?ooo`1h0?ooo`030;md001d_ol0oooo00d0oooo
00<000000?ooo`3oool0I@3oool0102LB000B9cO0;md001d_ol;0?ooo`00603oool00`000000oooo
0?ooo`1d0?ooo`050=nLB000000004RL09a8B02Lgol0KP3oool3000000D0oooo00L0okmd0000M02o
ool0oooo0?noM00004P0W=oo07@0oooo00L0okmd0000B02Lgol0oooo0?oOW01804P0W=oo00h0oooo
00<000000?ooo`3oool0I@3oool0103ogi`0B01804Q8001d_ol;0?ooo`00603oool00`000000oooo
0?ooo`1f0?ooo`030=nLB00004P0W=oo06/0oooo0`0000090?ooo`050?noM000000000000000B02L
gol0MP3oool01@3ogi`0B000000000000000M;oo00l0oooo00<000000?ooo`3oool0IP3oool00`3o
_g@0000007Boo`0;0?ooo`00603oool00`000000oooo0?ooo`3P0?ooo`@00000V`3oool00`000000
oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3M0?ooo`<00000W`3oool00`000000oooo
0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`3I0?ooo`@00000XP3oool00`000000oooo0?oo
o`1d0?ooo`00603oool00`000000oooo0?ooo`3E0?ooo`@00000YP3oool00`000000oooo0?ooo`1d
0?ooo`00603oool00`000000oooo0?ooo`3A0?ooo`@00000ZP3oool00`000000oooo0?ooo`1d0?oo
o`00603oool00`000000oooo0?ooo`3=0?ooo`@00000[P3oool00`000000oooo0?ooo`1d0?ooo`00
603oool00`000000oooo0?ooo`390?ooo`@00000/P3oool00`000000oooo0?ooo`1d0?ooo`00603o
ool00`000000oooo0?ooo`350?ooo`@00000]P3oool00`000000oooo0?ooo`1d0?ooo`00603oool0
0`000000oooo0?ooo`310?ooo`@00000^P3oool00`000000oooo0?ooo`1d0?ooo`00303ooooo0000
0?l000001`00000001P0oooo00<000000?ooo`3oool00P3oool:000000X0oooo00<000000?ooo`3o
ool05@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool05P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3o
ool03P3oool4000000@0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo01H0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0>0?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo01D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0:0?ooo`00
603oool00`000000oooo0?ooo`0<0?oooa0000004P3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0F0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool02P3oool4000000P0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
5P3oool00`000000oooo0?ooo`0:0?ooo`00603oool00`000000oooo0?ooo`0L0?oooaD00000A`3o
ool00`000000oooo0?ooo`0h0?ooo`<00000?@3oool00`000000oooo0?ooo`1h0?ooo`030000003o
ool0oooo00h0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`0:0?ooo`00603oool0
0`000000oooo0?ooo`0a0?ooo`l00000>03oool00`000000oooo0?ooo`0d0?ooo`@00000@03oool0
0`000000oooo0?ooo`1h0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0I`3oool0
0`000000oooo0?ooo`0:0?ooo`00603oool00`000000oooo0?ooo`100?ooo`X00000H@3oool40000
0=00oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0BP3oool@000004d0
oooo1000003D0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo05X0oooo
3`00000j0?ooo`@00000f03oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?oo
o`1Y0?ooo`X00000;03oool400000=`0oooo00<000000?ooo`3oool0M03oool001P0oooo00<00000
0?ooo`3oool0L`3oool@000001T0oooo0`00003P0?ooo`030000003oool0oooo07@0oooo000H0?oo
o`030000003oool0oooo08<0oooo400000050?ooo`@00000h`3oool00`000000oooo0?ooo`1d0?oo
o`00603oool00`000000oooo0?ooo`2C0?ooo`X00000hP3oool00`000000oooo0?ooo`1d0?ooo`00
603oool00`000000oooo0?ooo`2B0?ooo`<00000203oool?00000=<0oooo00<000000?ooo`3oool0
M03oool001P0oooo00<000000?ooo`3oool0SP3oool4000001X0oooo400000330?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo08/0oooo0`00000^0?ooo`X00000^@3oool0
0`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`280?ooo`<00000>`3oool:0000
0:l0oooo00<000000?ooo`3oool0M03oool001P0oooo0`0000250?ooo`<00000B03oool:00000:D0
oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0P`3oool2000005D0oooo
2P00002K0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo0800oooo0`00
001Q0?ooo`X00000T@3oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`1m
0?ooo`<00000KP3oool;000008H0oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?oo
o`3oool0NP3oool3000007`0oooo2P00001l0?ooo`030000003oool0oooo07@0oooo000H0?ooo`03
0000003oool0oooo07L0oooo0`0000290?ooo`X00000LP3oool00`000000oooo0?ooo`1d0?ooo`00
603oool00`000000oooo0?ooo`1e0?ooo`800000UP3oool:000006P0oooo00<000000?ooo`3oool0
M03oool001P0oooo00<000000?ooo`3oool0LP3oool300000:80oooo2@00001O0?ooo`030000003o
ool0oooo07@0oooo000H0?ooo`030000003oool0oooo0700oooo0P00002^0?ooo`L00000F03oool0
0`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?ooo`1]0?ooo`<00000]`3oool70000
0540oooo00<000000?ooo`3oool0M03oool001P0oooo00<000000?ooo`3oool0J`3oool200000<40
oooo1`00001:0?ooo`030000003oool0oooo07@0oooo000H0?ooo`030000003oool0oooo06P0oooo
0`00003:0?ooo`H00000A03oool00`000000oooo0?ooo`1d0?ooo`00603oool00`000000oooo0?oo
o`1V0?ooo`800000d`3oool7000003d0oooo00<000000?ooo`3oool0IP3oool2000000`0oooo000H
0?ooo`030000003oool0oooo06@0oooo0P00003L0?ooo`L00000=P3oool00`000000oooo0?ooo`1T
0?ooo`8000003P3oool001P0oooo00<000000?ooo`3oool0HP3oool200000>D0oooo1P00000`0?oo
o`030000003oool0oooo0680oooo0P00000@0?ooo`00603oool00`000000oooo0?ooo`1P0?ooo`80
0000k@3oool7000002T0oooo00<000000?ooo`3oool0H03oool200000180oooo000H0?ooo`<00000
GP3oool200000?H0oooo1P00000S0?ooo`030000003oool0oooo05h0oooo0P00000D0?ooo`00603o
ool00`000000oooo0?ooo`1L0?ooo`800000oP3oool4000001l0oooo00<000000?ooo`3oool0G@3o
ool00`000000oooo0?ooo`0D0?ooo`00603oool00`000000oooo0?ooo`1K0?ooo`030000003oool0
oooo0?l0oooo0`3oool4000001/0oooo00<000000?ooo`3oool0F`3oool2000001L0oooo000H0?oo
o`030000003oool0oooo05T0oooo0P00003o0?ooo`X0oooo1000000G0?ooo`030000003oool0oooo
05T0oooo0P00000I0?ooo`00603oool00`000000oooo0?ooo`1G0?ooo`800000o`3oool@0?ooo`@0
00004`3oool00`000000oooo0?ooo`1G0?ooo`8000006`3oool001P0oooo00<000000?ooo`3oool0
E@3oool200000?l0oooo5P3oool4000000l0oooo00<000000?ooo`3oool0E@3oool2000001d0oooo
000H0?ooo`030000003oool0oooo05<0oooo0P00003o0?oooa`0oooo1000000;0?ooo`030000003o
ool0oooo05<0oooo0P00000O0?ooo`00603oool00`000000oooo0?ooo`1A0?ooo`800000o`3ooolR
0?ooo`@000001`3oool00`000000oooo0?ooo`1B0?ooo`030000003oool0oooo01l0oooo000H0?oo
o`030000003oool0oooo0500oooo00<000000?ooo`3oool0o`3ooolV0?ooo`@000000`3oool00`00
0000oooo0?ooo`1@0?ooo`8000008P3oool001P0oooo00<000000?ooo`3oool0CP3oool200000?l0
oooo;@3oool400000540oooo00<000000?ooo`3oool08P3oool001P0oooo00<000000?ooo`3oool0
C@3oool00`000000oooo0?ooo`3o0?oooc00oooo1@00001<0?ooo`030000003oool0oooo02<0oooo
000H0?ooo`030000003oool0oooo04`0oooo00<000000?ooo`3oool0o`3ooola0?ooo`030000003o
ool0oooo0080oooo100000160?ooo`8000009P3oool001P0oooo00<000000?ooo`3oool0BP3oool2
00000?l0oooo=03oool00`000000oooo0?ooo`060?ooo`@00000@@3oool00`000000oooo0?ooo`0V
0?ooo`00603oool00`000000oooo0?ooo`190?ooo`030000003oool0oooo0?l0oooo=03oool00`00
0000oooo0?ooo`0:0?ooo`<00000?@3oool00`000000oooo0?ooo`0W0?ooo`00603oool00`000000
oooo0?ooo`170?ooo`800000o`3ooolg0?ooo`030000003oool0oooo00d0oooo0`00000h0?ooo`80
0000:P3oool001P0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`3o0?ooocL0oooo
00<000000?ooo`3oool0403oool2000003D0oooo00<000000?ooo`3oool0:P3oool001P0oooo0`00
00150?ooo`030000003oool0oooo0?l0oooo>03oool00`000000oooo0?ooo`0B0?ooo`<00000<@3o
ool00`000000oooo0?ooo`0[0?ooo`00603oool00`000000oooo0?ooo`140?ooo`030000003oool0
oooo0?l0oooo>@3oool00`000000oooo0?ooo`0E0?ooo`800000;P3oool00`000000oooo0?ooo`0/
0?ooo`00603oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0?l0oooo>P3oool00`00
0000oooo0?ooo`0G0?ooo`800000:`3oool00`000000oooo0?ooo`0]0?ooo`00603oool00`000000
oooo0?ooo`120?ooo`030000003oool0oooo0?l0oooo>`3oool00`000000oooo0?ooo`0I0?ooo`80
0000:03oool00`000000oooo0?ooo`0^0?ooo`00603oool00`000000oooo0?ooo`110?ooo`030000
003oool0oooo0?l0oooo?03oool00`000000oooo0?ooo`0K0?ooo`800000903oool200000340oooo
000H0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0o`3oooll0?ooo`030000003o
ool0oooo01d0oooo0P00000Q0?ooo`030000003oool0oooo0340oooo000H0?ooo`030000003oool0
oooo0400oooo00<000000?ooo`3oool0o`3ooolm0?ooo`030000003oool0oooo01l0oooo0P00000N
0?ooo`030000003oool0oooo0380oooo000H0?ooo`030000003oool0oooo03l0oooo00<000000?oo
o`3oool0o`3oooln0?ooo`030000003oool0oooo0240oooo0P00000K0?ooo`030000003oool0oooo
03<0oooo000H0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0o`3ooolo0?ooo`03
0000003oool0oooo02<0oooo0P00000H0?ooo`030000003oool0oooo03@0oooo000H0?ooo`030000
003oool0oooo03d0oooo00<000000?ooo`3oool0o`3ooom00?ooo`030000003oool0oooo02D0oooo
0P00000E0?ooo`030000003oool0oooo03D0oooo000H0?ooo`030000003oool0oooo03`0oooo00<0
00000?ooo`3oool0o`3ooom10?ooo`030000003oool0oooo02L0oooo0P00000B0?ooo`030000003o
ool0oooo03H0oooo000H0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0o`3ooom2
0?ooo`030000003oool0oooo02T0oooo0P00000@0?ooo`030000003oool0oooo03H0oooo000H0?oo
o`030000003oool0oooo03X0oooo00<000000?ooo`3oool0o`3ooom30?ooo`030000003oool0oooo
02/0oooo0P00000=0?ooo`030000003oool0oooo03L0oooo00090?ooo`030?oo_`1d0000000000@0
000000<0M;oo0?ooo`3oool01@3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0?l0
ooooA03oool00`000000oooo0?ooo`0]0?ooo`8000002P3oool00`000000oooo0?ooo`0h0?ooo`00
2P3oool00`2oM000001d0;ooo`020?ooo`030=nLM02Lgol0oooo00H0oooo00<000000?ooo`3oool0
>03oool00`000000oooo0?ooo`3o0?ooodD0oooo00<000000?ooo`3oool0;`3oool2000000L0oooo
00<000000?ooo`3oool0>@3oool000/0oooo00<0okmd0000B02Lgol02P3oool00`000000oooo0?oo
o`0g0?ooo`030000003oool0oooo0?l0ooooAP3oool00`000000oooo0?ooo`0a0?ooo`800000103o
ool00`000000oooo0?ooo`0j0?ooo`00303oool00`3ogi`0B00004RLg`090?ooo`@00000=@3oool0
0`000000oooo0?ooo`3o0?ooodL0oooo00<000000?ooo`3oool0<`3oool200000080oooo00<00000
0?ooo`3oool0>P3oool000h0oooo00<0W4P004RLg`3oool01`3oool00`000000oooo0?ooo`0e0?oo
o`030000003oool0oooo0?l0ooooB03oool00`000000oooo0?ooo`0e0?ooo`800000?@3oool000l0
oooo00<0W4Qd0;ooo`3oool01P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo0?l0
ooooB03oool00`000000oooo0?ooo`0e0?ooo`800000?@3oool000X0oooo00<0gia807Boo`3oool0
0P3oool00`2oM000M;oo0?ooo`060?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool0
o`3ooom90?ooo`030000003oool0oooo03@0oooo00@000000?ooo`3oool00000?03oool000X0oooo
00L0okmd0000M02oool0oooo0?noM00004P0W=oo00L0oooo00<000000?ooo`3oool0<`3oool00`00
0000oooo0?ooo`3o0?ooodX0oooo00<000000?ooo`3oool0=03oool01@000000oooo0?ooo`3oool0
000003/0oooo000;0?ooo`050?noM000000000000000B02Lgol0203oool00`000000oooo0?ooo`0c
0?ooo`030000003oool0oooo0?l0ooooBP3oool00`000000oooo0?ooo`0c0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0>03oool001P0oooo00<000000?ooo`3oool0<P3oool00`00
0000oooo0?ooo`3o0?oood/0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`050?oo
o`030000003oool0oooo03L0oooo000H0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3o
ool0o`3ooom;0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0f0?ooo`00603oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo0?l0oooo
C03oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
=@3oool001P0oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`3o0?ooodd0oooo00<0
00000?ooo`3oool0<03oool00`000000oooo0?ooo`0:0?ooo`800000=@3oool001P0oooo00<00000
0?ooo`3oool0<03oool00`000000oooo0?ooo`3o0?ooodd0oooo00<000000?ooo`3oool0<03oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0380oooo000H0?ooo`030000003oool0oooo
02l0oooo00<000000?ooo`3oool0o`3ooom>0?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool03P3oool00`000000oooo0?ooo`0a0?ooo`00603oool00`000000oooo0?ooo`0_0?ooo`03
0000003oool0oooo0?l0ooooCP3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo00l0
oooo00<000000?ooo`3oool0<03oool001P0oooo00<000000?ooo`3oool0;P3oool00`000000oooo
0?ooo`3o0?ooodl0oooo00<000000?ooo`3oool0;P3oool00`000000oooo0?ooo`0@0?ooo`030000
003oool0oooo0300oooo000H0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0o`3o
oom?0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0_0?ooo`00603oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo0?l0ooooD03oool0
0`000000oooo0?ooo`0]0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0;P3oool0
01P0oooo0`00000]0?ooo`030000003oool0oooo0?l0ooooD03oool00`000000oooo0?ooo`0/0?oo
o`030000003oool0oooo01D0oooo00<000000?ooo`3oool0;@3oool001P0oooo00<000000?ooo`3o
ool0;03oool00`000000oooo0?ooo`3o0?oooe40oooo00<000000?ooo`3oool0;03oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo02d0oooo000H0?ooo`030000003oool0oooo02`0oooo
00<000000?ooo`3oool0o`3ooomA0?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0
5`3oool00`000000oooo0?ooo`0/0?ooo`00603oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo0?l0ooooDP3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo01P0oooo00<0
00000?ooo`3oool0:`3oool001P0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`3o
0?oooe80oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0
oooo02/0oooo000H0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool0o`3ooomC0?oo
o`030000003oool0oooo02X0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`0Z0?oo
o`00603oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo0?l0ooooD`3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0:@3oool001P0oooo
00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`3o0?oooe@0oooo00<000000?ooo`3oool0
:@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo02P0oooo000H0?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`3oool0o`3ooomD0?ooo`030000003oool0oooo02T0oooo00<0
00000?ooo`3oool07@3oool00`000000oooo0?ooo`0X0?ooo`00603oool00`000000oooo0?ooo`0X
0?ooo`030000003oool0oooo0?l0ooooE@3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool09`3oool001P0oooo00<000000?ooo`3oool0:03oool00`00
0000oooo0?ooo`3o0?oooeD0oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`0P0?oo
o`030000003oool0oooo02H0oooo000H0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3o
ool0o`3ooomF0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0803oool00`000000
oooo0?ooo`0V0?ooo`00603oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0?l0oooo
EP3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0
9@3oool001P0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`3o0?oooeL0oooo00<0
00000?ooo`3oool09`3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo02D0oooo000H
0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0o`3ooomG0?ooo`030000003oool0
oooo02H0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0T0?ooo`00603oool00`00
0000oooo0?ooo`0U0?ooo`030000003oool0oooo0?l0ooooF03oool00`000000oooo0?ooo`0V0?oo
o`030000003oool0oooo02@0oooo00<000000?ooo`3oool0903oool001P0oooo0`00000U0?ooo`03
0000003oool0oooo0?l0ooooF03oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo02D0
oooo00<000000?ooo`3oool08`3oool001P0oooo00<000000?ooo`3oool0903oool00`000000oooo
0?ooo`3o0?oooeT0oooo00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`0V0?ooo`030000
003oool0oooo02<0oooo000H0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool0o`3o
oomI0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?oo
o`0R0?ooo`00603oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo0?l0ooooF@3oool0
0`000000oooo0?ooo`0U0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool08P3oool0
01P0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`3o0?oooeX0oooo00<000000?oo
o`3oool0903oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0240oooo000H0?ooo`03
0000003oool0oooo02<0oooo00<000000?ooo`3oool0o`3ooomJ0?ooo`030000003oool0oooo02@0
oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`0Q0?ooo`00603oool00`000000oooo
0?ooo`0R0?ooo`030000003oool0oooo0?l0ooooF`3oool00`000000oooo0?ooo`0S0?ooo`030000
003oool0oooo02X0oooo00<000000?ooo`3oool08@3oool001P0oooo00<000000?ooo`3oool08P3o
ool00`000000oooo0?ooo`3o0?oooe/0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`0[0?ooo`030000003oool0oooo0200oooo000H0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool0o`3ooomL0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0:`3oool0
0`000000oooo0?ooo`0P0?ooo`00603oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo
0?l0ooooG03oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool0803oool001P0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`3o0?oooe`0
oooo00<000000?ooo`3oool08P3oool00`000000oooo0?ooo`0]0?ooo`030000003oool0oooo01l0
oooo000H0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool0o`3ooomM0?ooo`030000
003oool0oooo0280oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0O0?ooo`00603o
ool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0?l0ooooG@3oool00`000000oooo0?oo
o`0Q0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool07`3oool001P0oooo00<00000
0?ooo`3oool0803oool00`000000oooo0?ooo`3o0?oooed0oooo00<000000?ooo`3oool08@3oool0
0`000000oooo0?ooo`0_0?ooo`030000003oool0oooo01h0oooo000H0?ooo`030000003oool0oooo
0200oooo00<000000?ooo`3oool0o`3ooomM0?ooo`030000003oool0oooo0200oooo00<000000?oo
o`3oool0<03oool00`000000oooo0?ooo`0N0?ooo`00603oool00`000000oooo0?ooo`0O0?ooo`03
0000003oool0oooo0?l0ooooGP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0300
oooo00<000000?ooo`3oool07P3oool001P0oooo0`00000O0?ooo`030000003oool0oooo0?l0oooo
GP3oool00`000000oooo0?ooo`0P0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0
7@3oool001P0oooo00<000000?ooo`3oool07`3oool00`000000oooo0?ooo`3o0?oooeh0oooo00<0
00000?ooo`3oool07`3oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo01d0oooo000H
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0o`3ooomN0?ooo`030000003oool0
oooo01l0oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`0L0?ooo`00603oool00`00
0000oooo0?ooo`0N0?ooo`030000003oool0oooo0?l0ooooG`3oool00`000000oooo0?ooo`0O0?oo
o`030000003oool0oooo03<0oooo00<000000?ooo`3oool0703oool001P0oooo00<000000?ooo`3o
ool07P3oool00`000000oooo0?ooo`3o0?oooel0oooo00<000000?ooo`3oool07`3oool00`000000
oooo0?ooo`0d0?ooo`030000003oool0oooo01/0oooo000H0?ooo`030000003oool0oooo01h0oooo
00<000000?ooo`3oool0o`3ooomO0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0
=@3oool00`000000oooo0?ooo`0K0?ooo`00603oool00`000000oooo0?ooo`0N0?ooo`030000003o
ool0oooo0?l0ooooG`3oool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo03D0oooo00<0
00000?ooo`3oool06`3oool001P0oooo00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`3o
0?ooof00oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0
oooo01/0oooo000H0?ooo`030000003oool0oooo01d0oooo00<000000?ooo`3oool0o`3ooomP0?oo
o`030000003oool0oooo01h0oooo00<000000?ooo`3oool0=P3oool00`000000oooo0?ooo`0J0?oo
o`00603oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo0?l0ooooH03oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool06P3oool001P0oooo
00<000000?ooo`3oool07@3oool00`000000oooo0?ooo`3o0?ooof00oooo00<000000?ooo`3oool0
7@3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo01X0oooo000H0?ooo`030000003o
ool0oooo01`0oooo00<000000?ooo`3oool0o`3ooomQ0?ooo`030000003oool0oooo01d0oooo00<0
00000?ooo`3oool0=`3oool00`000000oooo0?ooo`0J0?ooo`00603oool00`000000oooo0?ooo`0L
0?ooo`030000003oool0oooo0?l0ooooH@3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0
oooo03L0oooo00<000000?ooo`3oool06P3oool000`0oooo00D0omnL04P000000000000007Boo`07
0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0o`3ooomQ0?ooo`030000003oool0
oooo01d0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0J0?ooo`003P3oool00`2o
M000M;oo0?ooo`070?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0o`3ooomQ0?oo
o`030000003oool0oooo01`0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0I0?oo
o`003P3oool00`2oM000M;oo0?ooo`070?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0o`3ooomR0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0>@3oool00`000000
oooo0?ooo`0I0?ooo`002P3oool00`3OW4P00000000000030000000307Boo`3oool0oooo00D0oooo
1000000J0?ooo`030000003oool0oooo0?l0ooooHP3oool00`000000oooo0?ooo`0L0?ooo`030000
003oool0oooo03T0oooo00<000000?ooo`3oool06@3oool000X0oooo00H0okmd0000M02oool0oooo
0;md001d_ol80?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0o`3ooomR0?ooo`03
0000003oool0oooo01`0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`0I0?ooo`00
2`3oool01@3OW4P007Bo0?ooo`2oM000M;oo00P0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`3o0?ooof80oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0j0?ooo`03
0000003oool0oooo01T0oooo000<0?ooo`0409a80018W=l0_g@007Boo`P0oooo00<000000?ooo`3o
ool06P3oool00`000000oooo0?ooo`3o0?ooof<0oooo00<000000?ooo`3oool06`3oool00`000000
oooo0?ooo`0k0?ooo`030000003oool0oooo01P0oooo000<0?ooo`040?oOW01804P0B4P007Boo`P0
oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?ooof<0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo01P0oooo000=0?ooo`030?no
M0000000M;oo00P0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`3o0?ooof<0oooo
00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0k0?ooo`030000003oool0oooo01P0oooo
000H0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0o`3ooomS0?ooo`030000003o
ool0oooo01X0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0H0?ooo`00603oool0
0`000000oooo0?ooo`0J0?ooo`030000003oool0oooo0?l0ooooH`3oool00`000000oooo0?ooo`0J
0?ooo`030000003oool0oooo03`0oooo00<000000?ooo`3oool0603oool001P0oooo00<000000?oo
o`3oool06@3oool00`000000oooo0?ooo`3o0?ooof@0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`0m0?ooo`030000003oool0oooo01L0oooo000H0?ooo`030000003oool0oooo01T0
oooo00<000000?ooo`3oool0o`3ooomT0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3o
ool0?@3oool00`000000oooo0?ooo`0G0?ooo`00603oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo0?l0ooooI03oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo03d0oooo
00<000000?ooo`3oool05`3oool001P0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?oo
o`3o0?ooof@0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0n0?ooo`030000003o
ool0oooo01L0oooo000H0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0o`3ooomT
0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0F
0?ooo`00603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0?l0ooooI@3oool00`00
0000oooo0?ooo`0I0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool05P3oool001P0
oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`3o0?ooofD0oooo00<000000?ooo`3o
ool06@3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo01H0oooo000H0?ooo`030000
003oool0oooo01P0oooo00<000000?ooo`3oool0o`3ooomU0?ooo`030000003oool0oooo01T0oooo
00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0F0?ooo`00603oool3000001P0oooo00<0
00000?ooo`3oool0o`3ooomU0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0?`3o
ool00`000000oooo0?ooo`0F0?ooo`00603oool00`000000oooo0?ooo`0H0?ooo`030000003oool0
oooo0?l0ooooI@3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo0400oooo00<00000
0?ooo`3oool05P3oool001P0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`3o0?oo
ofH0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo
01H0oooo000H0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0o`3ooomV0?ooo`03
0000003oool0oooo01P0oooo00<000000?ooo`3oool0@@3oool00`000000oooo0?ooo`0E0?ooo`00
603oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0?l0ooooIP3oool00`000000oooo
0?ooo`0H0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool05@3oool001P0oooo00<0
00000?ooo`3oool05`3oool00`000000oooo0?ooo`3o0?ooofH0oooo00<000000?ooo`3oool0603o
ool00`000000oooo0?ooo`110?ooo`030000003oool0oooo01D0oooo000H0?ooo`030000003oool0
oooo01L0oooo00<000000?ooo`3oool0o`3ooomV0?ooo`030000003oool0oooo01P0oooo00<00000
0?ooo`3oool0@@3oool00`000000oooo0?ooo`0E0?ooo`00603oool00`000000oooo0?ooo`0F0?oo
o`030000003oool0oooo0?l0ooooI`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo
0480oooo00<000000?ooo`3oool05@3oool001P0oooo00<000000?ooo`3oool05P3oool00`000000
oooo0?ooo`3o0?ooofL0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`120?ooo`03
0000003oool0oooo01D0oooo000H0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
o`3ooomW0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0@P3oool00`000000oooo
0?ooo`0E0?ooo`00603oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0?l0ooooI`3o
ool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool05@3o
ool001P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`3o0?ooofL0oooo00<00000
0?ooo`3oool05`3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo01D0oooo000H0?oo
o`030000003oool0oooo01H0oooo00<000000?ooo`3oool0o`3ooomW0?ooo`030000003oool0oooo
01L0oooo00<000000?ooo`3oool0@`3oool00`000000oooo0?ooo`0D0?ooo`00603oool00`000000
oooo0?ooo`0E0?ooo`030000003oool0oooo0?l0ooooJ03oool00`000000oooo0?ooo`0F0?ooo`03
0000003oool0oooo04@0oooo00<000000?ooo`3oool0503oool001P0oooo00<000000?ooo`3oool0
5@3oool00`000000oooo0?ooo`3o0?ooofP0oooo00<000000?ooo`3oool05P3oool00`000000oooo
0?ooo`140?ooo`030000003oool0oooo01@0oooo000H0?ooo`030000003oool0oooo01D0oooo00<0
00000?ooo`3oool0o`3ooomX0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0A03o
ool00`000000oooo0?ooo`0D0?ooo`00603oool00`000000oooo0?ooo`0E0?ooo`030000003oool0
oooo0?l0ooooJ03oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo04@0oooo00<00000
0?ooo`3oool0503oool00?l0ooooo`3ooolC0?ooo`00o`3ooooo0?oooa<0oooo003o0?ooool0oooo
4`3oool00001\
\>"],
  ImageRangeCache->{{{0, 528.688}, {326.375, 0}} -> {-0.198882, -5.11195, \
0.0121969, 0.0469881}}]
}, Closed]],

Cell["\<\
Here we find two positive roots by supplying two sets of starting conditions \
based upon the figure.\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(FindRoot[eq1, {\[Xi], #}] &\)\  /@ {1, 3.5}\)], "Input"],

Cell[BoxData[
    \({{\[Xi] \[Rule] 1.2523532340025887`}, {\[Xi] \[Rule] 
          3.595304867161548`}}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Find the first 3 positive roots of ",
  Cell[BoxData[
      \(TraditionalForm\`tan(x) = 1/x\)]],
  "."
}], "Exercise"],

Cell["\<\
Find the first 3 positive roots of BesselJ[2,x].  Then explore the \
sensitivity to starting values; for example, compare results for starting \
points of 6.8, 7.0, 7.2.\
\>", "Exercise"],

Cell[TextData[{
  "Find all real solutions to the pair of equations ",
  Cell[BoxData[
      \(TraditionalForm\`x\^4 + y\^4 = 1\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`e\^x - e\^y = \ 1\)]],
  ".  [Hint: use ImplicitPlot to display the two equations and to locate \
appropriate starting values.]"
}], "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Linear algebra", "Section"],

Cell[CellGroupData[{

Cell["Matrix manipulation", "Subsection"],

Cell[TextData[{
  "A matrix is represented as a list of sublists in which each sublist has \
the same size.  ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " provides many built-in functions with obvious names for manipulating and \
analyzing matrices.  The most important matrix functions are summarized in \
the accompanying table."
}], "Text"],

Cell[BoxData[
    StyleBox[GridBox[{
          {
            StyleBox["function",
              "Text"], 
            StyleBox["output",
              "Text"]},
          {
            StyleBox[
              RowBox[{"Array", "[", 
                RowBox[{
                  StyleBox["mat",
                    "Text",
                    FontSlant->"Italic"], 
                  StyleBox[",",
                    "Text"], 
                  RowBox[{
                    StyleBox["{",
                      "Text"], 
                    StyleBox[\(m, n\),
                      "Text",
                      FontSlant->"Italic"], 
                    StyleBox["}",
                      "Text",
                      FontSlant->"Italic"]}]}], "]"}],
              "InlineInput"], 
            RowBox[{
              StyleBox["m",
                "Text",
                FontSlant->"Italic"], 
              StyleBox["\[Times]",
                "Text",
                FontSlant->"Italic"], 
              StyleBox["n",
                "Text",
                FontSlant->"Italic"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["matrix",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["with",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["elements",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox[\(mat[i, j]\),
                "Text",
                FontSlant->"Italic"]}]},
          {
            StyleBox[
              RowBox[{
                StyleBox["DiagonalMatrix",
                  "InlineInput"], 
                StyleBox["[",
                  "InlineInput"], 
                StyleBox["list",
                  "InlineInput",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "InlineInput"]}],
              "InlineInput"], 
            RowBox[{
              StyleBox["diagonal",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["matrix",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["using",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["elements",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["from",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["list",
                "Text",
                FontSlant->"Italic"]}]},
          {
            StyleBox[
              RowBox[{
                StyleBox["IdentityMatrix",
                  "InlineInput"], 
                StyleBox["[",
                  "InlineInput"], 
                StyleBox["m",
                  "InlineInput",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "InlineInput",
                  FontSlant->"Italic"]}],
              "InlineInput"], 
            RowBox[{
              StyleBox["m",
                "Text",
                FontSlant->"Italic"], 
              StyleBox["-",
                "Text"], 
              StyleBox[\(dimensional\ unit\ matrix\),
                "Text"]}]},
          {
            StyleBox["MatrixForm",
              "InlineInput"], 
            StyleBox[\(print\ matrix\ in\ traditional\ form\),
              "Text"]},
          {
            StyleBox["Transpose",
              "InlineInput"], 
            StyleBox[\(interchange\ rows\ and\ columns\),
              "Text"]},
          {
            StyleBox[
              RowBox[{
                StyleBox["Inverse",
                  "InlineInput"], 
                StyleBox["[",
                  "InlineInput"], 
                StyleBox["A",
                  "InlineInput",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "InlineInput",
                  FontSlant->"Italic"]}],
              "InlineInput"], 
            RowBox[{
              StyleBox["inverse",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["matrix",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox[\(A\^\(-1\)\),
                FontSlant->"Italic"]}]},
          {
            StyleBox[
              RowBox[{"Dot", "[", 
                StyleBox[\(A, B\),
                  FontSlant->"Italic"], "]"}],
              "InlineInput"], 
            RowBox[{
              StyleBox["product",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox[\(A . B\),
                "Text",
                FontSlant->"Italic"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["for",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["compatible",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["matrices",
                "Text"]}]},
          {
            StyleBox["Det",
              "InlineInput"], 
            StyleBox[\(determinant\ of\ square\ matrix\),
              "Text"]},
          {
            StyleBox[
              RowBox[{
                RowBox[{
                  StyleBox["Eigenvalues",
                    "InlineInput"], 
                  StyleBox["[",
                    "InlineInput"], 
                  StyleBox["A",
                    "InlineInput",
                    FontSlant->"Italic"], 
                  StyleBox["]",
                    "InlineInput"]}], 
                StyleBox[" ",
                  "InlineInput"]}],
              "InlineInput"], 
            
            StyleBox[\(list\ of\ eigenvalues\ \[Lambda], \ 
              such\ that\ A . x = \[Lambda]\ x\),
              "Text"]},
          {
            StyleBox[
              RowBox[{
                StyleBox["Eigenvectors",
                  "InlineInput"], 
                StyleBox["[",
                  "InlineInput"], 
                StyleBox["A",
                  "InlineInput",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "InlineInput",
                  FontSlant->"Italic"]}],
              "InlineInput"], 
            StyleBox[
              RowBox[{\(list\ of\ eigenvectors\ x\), ",", 
                StyleBox[
                  RowBox[{
                    StyleBox[" ",
                      "Text"], " "}]], \(such\ that\ A . x = \[Lambda]\ x\)}],
              
              "Text"]},
          {
            StyleBox["Eigensystem",
              "InlineInput"], 
            StyleBox[\(list\ containing\ both\ list\ of\ eigenvalues\ and\ \n
              list\ of\ eigenvectors\),
              "Text"]},
          {
            StyleBox[
              RowBox[{"LinearSolve", "[", 
                StyleBox[\(A, rhs\),
                  FontSlant->"Italic"], 
                StyleBox["]",
                  FontSlant->"Italic"]}],
              "InlineInput"], 
            RowBox[{
              RowBox[{
                StyleBox["solution",
                  "Text"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["vector",
                  "Text"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["x",
                  "Text",
                  FontSlant->"Italic"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["to",
                  "Text"], 
                StyleBox[" ",
                  "Text"], \(A . x\)}], "==", "rhs"}]},
          {
            StyleBox[
              RowBox[{"NullSpace", "[", 
                StyleBox["A",
                  FontSlant->"Italic"], "]"}],
              "InlineInput"], 
            RowBox[{
              RowBox[{
                RowBox[{
                  StyleBox["list",
                    "Text"], 
                  StyleBox[" ",
                    "Text"], 
                  StyleBox["of",
                    "Text"], 
                  StyleBox[" ",
                    "Text"], 
                  StyleBox["vectors",
                    "Text"], 
                  StyleBox[" ",
                    "Text"], 
                  StyleBox["x",
                    "Text",
                    FontSlant->"Italic"], 
                  StyleBox[" ",
                    "Text"], 
                  StyleBox["for",
                    "Text"], 
                  StyleBox[" ",
                    "Text"], 
                  StyleBox["which",
                    "Text"], 
                  StyleBox[" ",
                    "Text"], 
                  StyleBox[\(A . x\),
                    "Text"]}], 
                StyleBox["==",
                  "Text"], 
                StyleBox["0",
                  "Text"]}], 
              StyleBox[" ",
                "Text"]}]},
          {
            StyleBox[
              RowBox[{
                StyleBox["MatrixPower",
                  "InlineInput"], 
                StyleBox["[",
                  "InlineInput"], 
                StyleBox[\(A, b\),
                  "InlineInput",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "InlineInput",
                  FontSlant->"Italic"]}],
              "InlineInput"], 
            RowBox[{
              StyleBox[
                SuperscriptBox[
                  StyleBox["A",
                    FontSlant->"Italic"], "b"],
                "Text"], 
              StyleBox["=",
                "Text"], 
              RowBox[{
                StyleBox["matrix",
                  "Text"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["A",
                  "Text",
                  FontSlant->"Italic"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["raised",
                  "Text"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["to",
                  "Text"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["power",
                  "Text"], 
                StyleBox[" ",
                  "Text"], 
                StyleBox["b",
                  FontSlant->"Italic"]}]}]},
          {
            StyleBox[
              RowBox[{
                StyleBox["MatrixExp",
                  "InlineInput"], 
                StyleBox["[",
                  "InlineInput"], 
                StyleBox["A",
                  "InlineInput",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "InlineInput"]}],
              "InlineInput"], 
            RowBox[{
              RowBox[{
                StyleBox["Exp",
                  "Text"], 
                StyleBox["[",
                  "Text"], 
                StyleBox["A",
                  FontSlant->"Italic"], 
                StyleBox["]",
                  "Text"]}], 
              StyleBox[" ",
                "Text"], 
              StyleBox["as",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["defined",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["by",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["power",
                "Text"], 
              StyleBox[" ",
                "Text"], 
              StyleBox["series",
                "Text"]}]}
          },
        GridFrame->True],
      Background->RGBColor[0.500008, 0.996109, 0.996109]]], "Input",
  CellFrame->False,
  Evaluatable->False,
  CellFrameLabels->{{Inherited, Inherited}, {Inherited, None}},
  TextAlignment->Center,
  TextJustification->0,
  FontColor->GrayLevel[0],
  Background->None,
  PrivateFontOptions->{"FontType"->"Bitmap"},
  GridBoxOptions->{ColumnAlignments->{Left},
  RowLines->{True, False}}],

Cell["A few examples of some of these functions are given below.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(matrix\  = Array[m, {3, 3}]\)], "Input"],

Cell[BoxData[
    \({{m[1, 1], m[1, 2], m[1, 3]}, {m[2, 1], m[2, 2], m[2, 3]}, {m[3, 1], 
        m[3, 2], m[3, 3]}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(matrix\  // MatrixForm\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(m[1, 1]\), \(m[1, 2]\), \(m[1, 3]\)},
            {\(m[2, 1]\), \(m[2, 2]\), \(m[2, 3]\)},
            {\(m[3, 1]\), \(m[3, 2]\), \(m[3, 3]\)}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(\(matrix // Transpose\) // MatrixForm\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(m[1, 1]\), \(m[2, 1]\), \(m[3, 1]\)},
            {\(m[1, 2]\), \(m[2, 2]\), \(m[3, 2]\)},
            {\(m[1, 3]\), \(m[2, 3]\), \(m[3, 3]\)}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell[TextData[{
  "The product of two matrices with compatible dimensions is formed by ",
  ButtonBox["Dot",
    ButtonStyle->"RefGuideLink"],
  " with operator form (.)."
}], "Text"],

Cell[BoxData[
    \(\(\(var\  = \ Array[v, {3}];\)\(\ \)\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(matrix . var\)], "Input"],

Cell[BoxData[
    \({m[1, 1]\ v[1] + m[1, 2]\ v[2] + m[1, 3]\ v[3], 
      m[2, 1]\ v[1] + m[2, 2]\ v[2] + m[2, 3]\ v[3], 
      m[3, 1]\ v[1] + m[3, 2]\ v[2] + m[3, 3]\ v[3]}\)], "Output"]
}, Closed]],

Cell["\<\
Evaluate MatrixExp[{{0,\[Theta]},{-\[Theta],0}}] and interpret the result.\
\>", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Matrix form of linear equations", "Subsection"],

Cell[TextData[{
  "A system of equations of the form ",
  Cell[BoxData[
      \(TraditionalForm\`A . x \[Equal] b\)]],
  " can be formed using ",
  StyleBox["Thread", "InlineInput",
    FontWeight->"Bold"],
  " to move the function ",
  StyleBox["Equal", "InlineInput",
    FontWeight->"Bold"],
  " through the head ",
  StyleBox["List", "InlineInput",
    FontWeight->"Bold"],
  ".  (What happens if ",
  StyleBox["Thread", "InlineInput",
    FontWeight->"Bold"],
  " is omitted?)"
}], "Text"],

Cell[BoxData[{
    \(\(matrix\  = Array[A, {3, 3}];\)\), "\n", 
    \(\(var = Array[x, {3}];\)\), "\n", 
    \(\(rhs = \ Array[b, {3}];\)\)}], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Thread[matrix . var \[Equal] rhs]\)], "Input"],

Cell[BoxData[
    \({A[1, 1]\ x[1] + A[1, 2]\ x[2] + A[1, 3]\ x[3] \[Equal] b[1], 
      A[2, 1]\ x[1] + A[2, 2]\ x[2] + A[2, 3]\ x[3] \[Equal] b[2], 
      A[3, 1]\ x[1] + A[3, 2]\ x[2] + A[3, 3]\ x[3] \[Equal] 
        b[3]}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The result is then a form suitable for ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  ".  Alternatively, ",
  StyleBox["LogicalExpand", "InlineInput",
    FontWeight->"Bold"],
  " "
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(LogicalExpand[matrix . var \[Equal] rhs]\)], "Input"],

Cell[BoxData[
    \(A[1, 1]\ x[1] + A[1, 2]\ x[2] + A[1, 3]\ x[3] \[Equal] b[1] && 
      A[2, 1]\ x[1] + A[2, 2]\ x[2] + A[2, 3]\ x[3] \[Equal] b[2] && 
      A[3, 1]\ x[1] + A[3, 2]\ x[2] + A[3, 3]\ x[3] \[Equal] b[3]\)], "Output"]
}, Closed]],

Cell[TextData[{
  "also produces an expression suitable for input to ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  ".  Although these methods are most often used for systems of linear \
equations, they can be employed for more general problems if the coefficient \
matrix and variable vector are constructed properly."
}], "Text"],

Cell[TextData[{
  "Alternatively, ",
  ButtonBox["LinearSolve",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[coeff,rhs]", "InlineInput",
    FontWeight->"Bold"],
  " is a specialized version for linear equations of the form ",
  StyleBox["coeff.x\[Equal]rhs", "InlineInput",
    FontWeight->"Bold"],
  ".  The solution is a vector ",
  StyleBox["x", "InlineInput"],
  " which is returned as a simple list, rather than as a set of replacement \
rules; note that it is not necessary to supply variable names.  ",
  StyleBox["LinearSolve", "InlineInput",
    FontWeight->"Bold"],
  " works with either symbolic or numerical expressions.  For underdetermined \
systems ",
  StyleBox["LinearSolve", "InlineInput",
    FontWeight->"Bold"],
  " returns only one of the possible solutions, whereas ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  " returns the general solution.  Also note that for sparse matrices it is \
usually more efficient to use ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(coeff\  = \ {\ {1, \(-2\), \ 3\/8}, {2, 2, \(-3\)}, {1\/2, \(-1\), 
            4}};\)\), "\n", 
    \(\(rhs = {\(-1\), 3, 0};\)\), "\n", 
    \(LinearSolve[coeff, rhs]\)}], "Input"],

Cell[BoxData[
    \({143\/183, 335\/366, 8\/61}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "The null space of a matrix is defined by a linear combination of basis \
vectors satisfying ",
  StyleBox["matrix.m\[Equal]0", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[" ",
    FontSlant->"Italic"],
  "and is obtained using ",
  ButtonBox["NullSpace",
    ButtonStyle->"RefGuideLink"],
  StyleBox["[", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["matrix", "InlineInput",
    FontWeight->"Bold"],
  StyleBox["]", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(a\  = \ {{1, 1, 0}, {\(-1\), 0, 0}, {1, 1, 0}};\)\), "\n", 
    \(b\  = NullSpace[a]\)}], "Input"],

Cell[BoxData[
    \({{0, 0, 1}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(a . b\[LeftDoubleBracket]1\[RightDoubleBracket]\)], "Input"],

Cell[BoxData[
    \({0, 0, 0}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Note that ",
  StyleBox["LinearSolve", "InlineInput",
    FontWeight->"Bold"],
  " and ",
  StyleBox["NullSpace", "InlineInput",
    FontWeight->"Bold"],
  " can be used for rectangular as well as square matrices, as illustrated by \
the following exercise."
}], "Text"],

Cell[TextData[{
  "Let a={{1,2,1},{-1,-1,2}} and b={2,3}.  Determine vectors ",
  StyleBox["x",
    FontSlant->"Italic"],
  " and ",
  StyleBox["c",
    FontSlant->"Italic"],
  " such that ",
  StyleBox["a.(x+\[Lambda] c)==b",
    FontSlant->"Italic"],
  " for any value of \[Lambda].  [Hint: use both LinearSolve and NullSpace.]"
}], "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["Eigensystems", "Subsection"],

Cell[TextData[{
  "The determinant of a square matrix is given by ",
  ButtonBox["Det",
    ButtonStyle->"RefGuideLink"],
  ".  ",
  ButtonBox["Eigenvalues",
    ButtonStyle->"RefGuideLink"],
  " finds the eigenvalues and ",
  ButtonBox["Eigenvectors",
    ButtonStyle->"RefGuideLink"],
  " the eigenvectors of a matrix, either symbolically or numerically.  \
Repeated eigenvalues appear with the requisite multiplicity.  ",
  ButtonBox["Eigensystem",
    ButtonStyle->"RefGuideLink"],
  " returns both in the form of a list with two sublists, the first being the \
eigenvalues and the second the corresponding eigenvectors.  Eigenvectors are \
not normalized automatically.  The package ",
  StyleBox["LinearAlgebra`Orthogonalization`", "InlineInput",
    FontWeight->"Bold"],
  " provides tools to create orthonormal bases.  Rather than attempt to \
describe all available functions in general terms we prefer to use an \
example."
}], "Text"],

Cell[CellGroupData[{

Cell["Example: linear triatomic molecule", "Subsubsection"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .20833 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.579365 0.396825 0.124008 0.396825 [
[.10317 .12401 -12.5 -6.625 ]
[.10317 .12401 12.5 6.625 ]
[.57937 .12401 -12.5 -6.625 ]
[.57937 .12401 12.5 6.625 ]
[.89683 .12401 -12.5 -6.65625 ]
[.89683 .12401 12.5 6.65625 ]
[.34127 .00496 -14.9375 -9.5625 ]
[.34127 .00496 14.9375 9.5625 ]
[.7381 .00496 -14.9375 -9.5625 ]
[.7381 .00496 14.9375 9.5625 ]
[ 0 0 0 0 ]
[ 1 .20833 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.5 Mabswid
newpath
.10317 .12401 .07937 0 365.73 arc
s
newpath
.57937 .12401 .07937 0 365.73 arc
s
newpath
.89683 .12401 .07937 0 365.73 arc
s
.18254 .12401 m
.1832 .12665 L
.18408 .12947 L
.18611 .13446 L
.18924 .14016 L
.19274 .14503 L
.19701 .14968 L
.20227 .15411 L
.20775 .15761 L
.21319 .16018 L
.21615 .16126 L
.2194 .1622 L
.22099 .16258 L
.22252 .16289 L
.22403 .16314 L
.22541 .16333 L
.22707 .16351 L
.22798 .16358 L
.22885 .16363 L
.23039 .16368 L
.23205 .16368 L
.23293 .16366 L
.23389 .16362 L
.23489 .16356 L
.23583 .16348 L
.2375 .16329 L
.23928 .16302 L
.24283 .16229 L
.24662 .1612 L
.25011 .15989 L
.25324 .15844 L
.25886 .15514 L
.2643 .15087 L
.26872 .14637 L
.27225 .14182 L
.27548 .13647 L
.27686 .13364 L
.27812 .13055 L
.27905 .12778 L
.27976 .12525 L
.28032 .1227 L
.28071 .1203 L
.28088 .11895 L
.28101 .11752 L
.28109 .1163 L
.28114 .11498 L
.28114 .11356 L
.28113 .11278 L
.2811 .11206 L
.28101 .11069 L
Mistroke
.28089 .1094 L
.28071 .10804 L
.28047 .1066 L
.27991 .10406 L
.27921 .10168 L
.27844 .0996 L
.27626 .09526 L
.27496 .09331 L
.27339 .09136 L
.27048 .08859 L
.26872 .08733 L
.26704 .08636 L
.26544 .08563 L
.26457 .08531 L
.26364 .08501 L
.26283 .0848 L
.26192 .08461 L
.2611 .08448 L
.26032 .08439 L
.2594 .08434 L
.25841 .08433 L
.25749 .08437 L
.25665 .08446 L
.25566 .08461 L
.25461 .08483 L
.25361 .08511 L
.25271 .08541 L
.25093 .08616 L
.24904 .08721 L
.24712 .08857 L
.24547 .09001 L
.24274 .09313 L
.2415 .09496 L
.24027 .09712 L
.23928 .09925 L
.23848 .10133 L
.23783 .10343 L
.23728 .10572 L
.23704 .10693 L
.23683 .10828 L
.23667 .10953 L
.23656 .11071 L
.23647 .11212 L
.23643 .11342 L
.23643 .11414 L
.23644 .11493 L
.23649 .11637 L
.23657 .11763 L
.2367 .11902 L
.23689 .12048 L
Mistroke
.2371 .12186 L
.2376 .12432 L
.23828 .12696 L
.23927 .13005 L
.2406 .13337 L
.24365 .13923 L
.24738 .14459 L
.25144 .14914 L
.2564 .15347 L
.26236 .1574 L
.26528 .15891 L
.26846 .16029 L
.27127 .1613 L
.27436 .1622 L
.27741 .16287 L
.28022 .16331 L
.28158 .16346 L
.28306 .16358 L
.28446 .16366 L
.28576 .16369 L
.28727 .16368 L
.28868 .16363 L
.29023 .16353 L
.29189 .16336 L
.29353 .16314 L
.29525 .16285 L
.29829 .16217 L
.3015 .16123 L
.30447 .16015 L
.30981 .15761 L
.31518 .1542 L
.32046 .14978 L
.32454 .14539 L
.32824 .1403 L
.33087 .13568 L
.33317 .13039 L
.33414 .12747 L
.33498 .12427 L
.33528 .1228 L
.33556 .12124 L
.33577 .11979 L
.33592 .11845 L
.33604 .11705 L
.33611 .11554 L
.33613 .11422 L
.33612 .11284 L
.33606 .11162 L
.33598 .11049 L
.33586 .10926 L
.33569 .10796 L
Mistroke
.33516 .10511 L
.33477 .10356 L
.33435 .10216 L
.33348 .09973 L
.33234 .09722 L
.32995 .09331 L
.32845 .09143 L
.32692 .08985 L
.3236 .08726 L
.32158 .08614 L
.32059 .0857 L
.31965 .08534 L
.31874 .08504 L
.31774 .08478 L
.31683 .0846 L
.31599 .08447 L
.31507 .08437 L
.31406 .08433 L
.31356 .08433 L
.31301 .08434 L
.31203 .08441 L
.31107 .08454 L
.31018 .0847 L
.30922 .08493 L
.30821 .08523 L
.30653 .08588 L
.30473 .08679 L
.3016 .08899 L
.29893 .09163 L
.29638 .09513 L
.29514 .09737 L
.29412 .0996 L
.29323 .10206 L
.29254 .10449 L
.29223 .10589 L
.29195 .10743 L
.29174 .10886 L
.29159 .1102 L
.2915 .11138 L
.29144 .11267 L
.29142 .1139 L
.29143 .11502 L
.29147 .11635 L
.29157 .11782 L
.29171 .1192 L
.29188 .1205 L
.29237 .12334 L
.29298 .12592 L
.29379 .12867 L
.29487 .13165 L
Mistroke
.29748 .1372 L
.30098 .14276 L
.30487 .14751 L
.309 .15152 L
.31446 .15564 L
.31987 .15872 L
.32302 .16012 L
.32655 .1614 L
.32996 .16235 L
.33151 .16269 L
.33316 .163 L
.33465 .16323 L
.33628 .16343 L
.33777 .16356 L
.33916 .16364 L
.3408 .16369 L
.3423 .16368 L
.34313 .16365 L
.34402 .16361 L
.34565 .16349 L
.34706 .16334 L
.3486 .16313 L
.35169 .16255 L
.35452 .16184 L
.35712 .16102 L
.36334 .15838 L
.36865 .15526 L
.37425 .15089 L
.37894 .14609 L
.3825 .14142 L
.38576 .13589 L
.38711 .13301 L
.38835 .12983 L
.38925 .12706 L
.39002 .12401 L
.39031 .12263 L
.39056 .12116 L
.39074 .11989 L
.3909 .11852 L
.39097 .11772 L
.39103 .11698 L
.3911 .11553 L
.39112 .11422 L
.39111 .11285 L
.39106 .11168 L
.39097 .11043 L
.39085 .10926 L
.39071 .10821 L
.39027 .10569 L
.38971 .10338 L
Mistroke
.38889 .10084 L
.38797 .09857 L
.38588 .09469 L
.38315 .09112 L
.38143 .08941 L
.37973 .08804 L
.37672 .08621 L
.37493 .08544 L
.37321 .0849 L
.37227 .08468 L
.37138 .08452 L
.37057 .08442 L
.36969 .08435 L
.36873 .08433 L
.36771 .08436 L
.36684 .08443 L
.36589 .08456 L
.36495 .08475 L
.36408 .08496 L
.36214 .08562 L
.3603 .08647 L
.35841 .08761 L
.3552 .09025 L
.35262 .09326 L
.35009 .09744 L
.34892 .10008 L
.34803 .10267 L
.3474 .10504 L
.34688 .10776 L
.34669 .10917 L
.34661 .10995 L
.34654 .1107 L
.34645 .11217 L
.34641 .11353 L
.34641 .11496 L
.34644 .11577 L
.34647 .11651 L
.34657 .1179 L
.34672 .11936 L
.34689 .12068 L
.34709 .12192 L
.34767 .12473 L
.3485 .12778 L
.34942 .13053 L
.35212 .13657 L
.35509 .14151 L
.35895 .14651 L
.36284 .15049 L
.3675 .1543 L
.37284 .15767 L
Mistroke
.37821 .1602 L
.38161 .16142 L
.38492 .16234 L
.38794 .16297 L
.38967 .16324 L
.39128 .16343 L
.39299 .16358 L
.39395 .16364 L
.39484 .16367 L
.39647 .16369 L
.3973 .16368 L
.3982 .16365 L
.39977 .16356 L
.40144 .16341 L
.40302 .16322 L
.40445 .16299 L
.40784 .16228 L
.41147 .16124 L
.41481 .16 L
.41782 .15864 L
.42304 .15566 L
.42823 .15178 L
.43262 .14756 L
.43617 .14328 L
.43951 .13817 L
.44242 .13225 L
.44357 .12918 L
.44456 .12587 L
.44528 .12272 L
.44554 .12125 L
.44573 .11987 L
.44588 .11857 L
.446 .11718 L
.44605 .1164 L
.44608 .11565 L
.44611 .11426 L
.44609 .11282 L
.44603 .11149 L
.44592 .11005 L
.44584 .10926 L
.44575 .10854 L
.44534 .10611 L
.44475 .10357 L
.44404 .10128 L
.44326 .09926 L
.44112 .09509 L
.43968 .09298 L
.43821 .09119 L
.43491 .08817 L
.43289 .08685 L
Mistroke
.43098 .08587 L
.42927 .08522 L
.42828 .08492 L
.42734 .0847 L
.42635 .08452 L
.42543 .08441 L
.42443 .08434 L
.42392 .08433 L
.42337 .08433 L
.42243 .08438 L
.42158 .08446 L
.4206 .08461 L
.41968 .08481 L
.41807 .08527 L
.41635 .08595 L
.41454 .08689 L
.41262 .08816 L
.41085 .08962 L
.40908 .09144 L
.40755 .09335 L
.40627 .09527 L
.4051 .0974 L
.40416 .09945 L
.40334 .10164 L
.40261 .10412 L
.40228 .10553 L
.40202 .10685 L
.4018 .10822 L
.40164 .10952 L
.40153 .11065 L
.40145 .11188 L
.4014 .11326 L
.4014 .11452 L
.40143 .11585 L
.40151 .11731 L
.40163 .11869 L
.40178 .11997 L
.40226 .12288 L
.40287 .12561 L
.40374 .1286 L
.4049 .13178 L
.40794 .13808 L
.41152 .14352 L
.41548 .1482 L
.42032 .15265 L
.4254 .15625 L
.43036 .15896 L
.43333 .16025 L
.43655 .1614 L
.44017 .1624 L
Mistroke
.4418 .16276 L
.44356 .16307 L
.44522 .16331 L
.44675 .16348 L
.44829 .1636 L
.44975 .16367 L
.45145 .16369 L
.45302 .16366 L
.45384 .16362 L
.45475 .16356 L
.45569 .16349 L
.45657 .1634 L
.45809 .1632 L
.4597 .16294 L
.46258 .16234 L
.46595 .1614 L
.46894 .16034 L
.4752 .15738 L
.48067 .15381 L
.48524 .14995 L
.48903 .14596 L
.49264 .14117 L
.4958 .13575 L
.49714 .13288 L
.49836 .12975 L
.5 .12401 L
Mfstroke
.65873 .12401 m
.65906 .12665 L
.6595 .12947 L
.66051 .13446 L
.66208 .14016 L
.66383 .14503 L
.66596 .14968 L
.66859 .15411 L
.67134 .15761 L
.67405 .16018 L
.67553 .16126 L
.67716 .1622 L
.67796 .16258 L
.67872 .16289 L
.67948 .16314 L
.68016 .16333 L
.68099 .16351 L
.68145 .16358 L
.68188 .16363 L
.68265 .16368 L
.68349 .16368 L
.68392 .16366 L
.68441 .16362 L
.68491 .16356 L
.68538 .16348 L
.68621 .16329 L
.6871 .16302 L
.68888 .16229 L
.69077 .1612 L
.69252 .15989 L
.69408 .15844 L
.69689 .15514 L
.69961 .15087 L
.70182 .14637 L
.70358 .14182 L
.7052 .13647 L
.70589 .13364 L
.70652 .13055 L
.70699 .12778 L
.70734 .12525 L
.70762 .1227 L
.70782 .1203 L
.7079 .11895 L
.70797 .11752 L
.70801 .1163 L
.70803 .11498 L
.70803 .11356 L
.70802 .11278 L
.70801 .11206 L
.70797 .11069 L
Mistroke
.7079 .1094 L
.70782 .10804 L
.7077 .1066 L
.70742 .10406 L
.70707 .10168 L
.70668 .0996 L
.70559 .09526 L
.70494 .09331 L
.70416 .09136 L
.7027 .08859 L
.70182 .08733 L
.70098 .08636 L
.70018 .08563 L
.69975 .08531 L
.69928 .08501 L
.69887 .0848 L
.69842 .08461 L
.69801 .08448 L
.69762 .08439 L
.69716 .08434 L
.69667 .08433 L
.6962 .08437 L
.69579 .08446 L
.69529 .08461 L
.69476 .08483 L
.69427 .08511 L
.69382 .08541 L
.69292 .08616 L
.69198 .08721 L
.69102 .08857 L
.6902 .09001 L
.68883 .09313 L
.68821 .09496 L
.6876 .09712 L
.6871 .09925 L
.6867 .10133 L
.68638 .10343 L
.6861 .10572 L
.68598 .10693 L
.68587 .10828 L
.6858 .10953 L
.68574 .11071 L
.6857 .11212 L
.68568 .11342 L
.68567 .11414 L
.68568 .11493 L
.6857 .11637 L
.68575 .11763 L
.68581 .11902 L
.6859 .12048 L
Mistroke
.68601 .12186 L
.68626 .12432 L
.6866 .12696 L
.6871 .13005 L
.68776 .13337 L
.68929 .13923 L
.69115 .14459 L
.69318 .14914 L
.69566 .15347 L
.69864 .1574 L
.7001 .15891 L
.70169 .16029 L
.7031 .1613 L
.70464 .1622 L
.70616 .16287 L
.70757 .16331 L
.70825 .16346 L
.70899 .16358 L
.70969 .16366 L
.71034 .16369 L
.7111 .16368 L
.7118 .16363 L
.71257 .16353 L
.7134 .16336 L
.71422 .16314 L
.71508 .16285 L
.71661 .16217 L
.71821 .16123 L
.7197 .16015 L
.72237 .15761 L
.72505 .1542 L
.72769 .14978 L
.72973 .14539 L
.73158 .1403 L
.7329 .13568 L
.73404 .13039 L
.73453 .12747 L
.73495 .12427 L
.7351 .1228 L
.73524 .12124 L
.73534 .11979 L
.73542 .11845 L
.73548 .11705 L
.73552 .11554 L
.73553 .11422 L
.73552 .11284 L
.73549 .11162 L
.73545 .11049 L
.73539 .10926 L
.7353 .10796 L
Mistroke
.73504 .10511 L
.73484 .10356 L
.73464 .10216 L
.7342 .09973 L
.73363 .09722 L
.73244 .09331 L
.73168 .09143 L
.73092 .08985 L
.72926 .08726 L
.72825 .08614 L
.72776 .0857 L
.72728 .08534 L
.72683 .08504 L
.72633 .08478 L
.72588 .0846 L
.72546 .08447 L
.725 .08437 L
.72449 .08433 L
.72424 .08433 L
.72397 .08434 L
.72347 .08441 L
.72299 .08454 L
.72255 .0847 L
.72207 .08493 L
.72156 .08523 L
.72073 .08588 L
.71983 .08679 L
.71826 .08899 L
.71693 .09163 L
.71565 .09513 L
.71503 .09737 L
.71452 .0996 L
.71408 .10206 L
.71373 .10449 L
.71357 .10589 L
.71343 .10743 L
.71333 .10886 L
.71326 .1102 L
.71321 .11138 L
.71318 .11267 L
.71317 .1139 L
.71317 .11502 L
.7132 .11635 L
.71325 .11782 L
.71332 .1192 L
.7134 .1205 L
.71365 .12334 L
.71395 .12592 L
.71435 .12867 L
.71489 .13165 L
Mistroke
.7162 .1372 L
.71795 .14276 L
.71989 .14751 L
.72196 .15152 L
.72469 .15564 L
.7274 .15872 L
.72897 .16012 L
.73073 .1614 L
.73244 .16235 L
.73322 .16269 L
.73404 .163 L
.73478 .16323 L
.7356 .16343 L
.73634 .16356 L
.73704 .16364 L
.73786 .16369 L
.73861 .16368 L
.73902 .16365 L
.73947 .16361 L
.74028 .16349 L
.74099 .16334 L
.74176 .16313 L
.74331 .16255 L
.74472 .16184 L
.74602 .16102 L
.74913 .15838 L
.75179 .15526 L
.75459 .15089 L
.75693 .14609 L
.75871 .14142 L
.76034 .13589 L
.76102 .13301 L
.76164 .12983 L
.76208 .12706 L
.76247 .12401 L
.76261 .12263 L
.76274 .12116 L
.76283 .11989 L
.76291 .11852 L
.76295 .11772 L
.76297 .11698 L
.76301 .11553 L
.76302 .11422 L
.76301 .11285 L
.76299 .11168 L
.76294 .11043 L
.76288 .10926 L
.76282 .10821 L
.7626 .10569 L
.76231 .10338 L
Mistroke
.76191 .10084 L
.76145 .09857 L
.7604 .09469 L
.75904 .09112 L
.75817 .08941 L
.75733 .08804 L
.75582 .08621 L
.75492 .08544 L
.75407 .0849 L
.7536 .08468 L
.75315 .08452 L
.75275 .08442 L
.7523 .08435 L
.75183 .08433 L
.75132 .08436 L
.75088 .08443 L
.7504 .08456 L
.74994 .08475 L
.7495 .08496 L
.74853 .08562 L
.74761 .08647 L
.74666 .08761 L
.74506 .09025 L
.74377 .09326 L
.7425 .09744 L
.74192 .10008 L
.74147 .10267 L
.74116 .10504 L
.7409 .10776 L
.74081 .10917 L
.74076 .10995 L
.74073 .1107 L
.74068 .11217 L
.74066 .11353 L
.74067 .11496 L
.74068 .11577 L
.7407 .11651 L
.74074 .1179 L
.74082 .11936 L
.74091 .12068 L
.74101 .12192 L
.7413 .12473 L
.74171 .12778 L
.74217 .13053 L
.74352 .13657 L
.74501 .14151 L
.74694 .14651 L
.74888 .15049 L
.75121 .1543 L
.75388 .15767 L
Mistroke
.75656 .1602 L
.75827 .16142 L
.75992 .16234 L
.76143 .16297 L
.7623 .16324 L
.7631 .16343 L
.76395 .16358 L
.76444 .16364 L
.76488 .16367 L
.7657 .16369 L
.76611 .16368 L
.76656 .16365 L
.76734 .16356 L
.76818 .16341 L
.76897 .16322 L
.76968 .16299 L
.77138 .16228 L
.77319 .16124 L
.77487 .16 L
.77637 .15864 L
.77898 .15566 L
.78157 .15178 L
.78377 .14756 L
.78555 .14328 L
.78722 .13817 L
.78867 .13225 L
.78924 .12918 L
.78974 .12587 L
.7901 .12272 L
.79023 .12125 L
.79033 .11987 L
.7904 .11857 L
.79046 .11718 L
.79049 .1164 L
.7905 .11565 L
.79052 .11426 L
.79051 .11282 L
.79048 .11149 L
.79042 .11005 L
.79038 .10926 L
.79033 .10854 L
.79013 .10611 L
.78983 .10357 L
.78948 .10128 L
.78909 .09926 L
.78802 .09509 L
.7873 .09298 L
.78657 .09119 L
.78491 .08817 L
.78391 .08685 L
Mistroke
.78295 .08587 L
.78209 .08522 L
.7816 .08492 L
.78113 .0847 L
.78063 .08452 L
.78018 .08441 L
.77967 .08434 L
.77942 .08433 L
.77915 .08433 L
.77868 .08438 L
.77825 .08446 L
.77776 .08461 L
.7773 .08481 L
.77649 .08527 L
.77564 .08595 L
.77473 .08689 L
.77377 .08816 L
.77289 .08962 L
.772 .09144 L
.77123 .09335 L
.7706 .09527 L
.77001 .0974 L
.76954 .09945 L
.76913 .10164 L
.76877 .10412 L
.7686 .10553 L
.76847 .10685 L
.76836 .10822 L
.76828 .10952 L
.76823 .11065 L
.76819 .11188 L
.76816 .11326 L
.76816 .11452 L
.76817 .11585 L
.76822 .11731 L
.76828 .11869 L
.76835 .11997 L
.76859 .12288 L
.7689 .12561 L
.76933 .1286 L
.76991 .13178 L
.77143 .13808 L
.77322 .14352 L
.7752 .1482 L
.77762 .15265 L
.78016 .15625 L
.78264 .15896 L
.78412 .16025 L
.78574 .1614 L
.78755 .1624 L
Mistroke
.78836 .16276 L
.78924 .16307 L
.79007 .16331 L
.79084 .16348 L
.79161 .1636 L
.79234 .16367 L
.79319 .16369 L
.79397 .16366 L
.79438 .16362 L
.79483 .16356 L
.79531 .16349 L
.79575 .1634 L
.7965 .1632 L
.79731 .16294 L
.79875 .16234 L
.80043 .1614 L
.80193 .16034 L
.80506 .15738 L
.80779 .15381 L
.81008 .14995 L
.81197 .14596 L
.81378 .14117 L
.81536 .13575 L
.81603 .13288 L
.81664 .12975 L
.81746 .12401 L
Mfstroke
gsave
.10317 .12401 -73.5 -10.625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 21.25 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 17.250000 moveto
 460.000000 17.250000 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 17.250000 lineto
closepath
clip newpath
63.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(m) show
77.062500 16.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
84.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-BoldItalic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-BoldItalic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-BoldItalic
grestore
gsave
.57937 .12401 -73.5 -10.625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 21.25 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 17.250000 moveto
 460.000000 17.250000 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 17.250000 lineto
closepath
clip newpath
63.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(m) show
77.062500 16.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
84.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-BoldItalic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-BoldItalic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-BoldItalic
grestore
gsave
.89683 .12401 -73.5 -10.6563 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 21.3125 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 17.312500 moveto
 460.000000 17.312500 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 17.312500 lineto
closepath
clip newpath
63.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(m) show
77.062500 16.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
84.000000 13.250000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-BoldItalic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-BoldItalic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-BoldItalic
grestore
gsave
.34127 .00496 -75.9375 -13.5625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.125 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 23.125000 moveto
 460.000000 23.125000 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 23.125000 lineto
closepath
clip newpath
63.000000 17.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 17.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(k) show
72.000000 20.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(1) show
78.750000 20.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
81.937500 20.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
88.875000 17.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-BoldItalic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-BoldItalic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-BoldItalic
grestore
gsave
.7381 .00496 -75.9375 -13.5625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 27.125 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 23.125000 moveto
 460.000000 23.125000 lineto
 460.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 23.125000 lineto
closepath
clip newpath
63.000000 17.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 17.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(k) show
72.000000 20.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(2) show
78.750000 20.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(,) show
81.937500 20.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 12.750000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(3) show
88.875000 17.125000 moveto
%%IncludeResource: font Times-BoldItalic
%%IncludeFont: Times-BoldItalic
/Times-BoldItalic findfont 18.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Times-BoldItalic
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Times-BoldItalic
%%DocumentSuppliedFonts:
%%DocumentFonts: font Times-BoldItalic
grestore
0 0 m
1 0 L
1 .20833 L
0 .20833 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{305.25, 63.5},
  ImageMargins->{{0, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000790000Gb000`400?l00000o`00003oo`3oooo:
0?ooo`00o`3oooo:0?ooo`00o`3oooo:0?ooo`00Z@3oool00`3o_g@0001d0;ooo`2I0?ooo`030?no
M00007@0_ooo0840oooo002Z0?ooo`030=nLB0000000B9cO09T0oooo00<0gia800000018W=l0P03o
ool00:/0oooo00<0omnL04P0001d_ol0V@3oool00`3ogi`0B00007Boo`1o0?ooo`00X03oool00`2o
M0000000000000030000000:001d_`3oool0oooo0?ooo`3ooml0W4P000000000M;l0oooO09a800H0
000000<0M;oo0?ooo`3oool0PP3oool00`3ooml0W4P0000000050000000807Boo`3oool0oooo0?oo
g`2LB0000000001d_`3OW4P3000000030000B02Lgol0oooo07P0oooo002R0?ooo`040=nLB0000000
001d0;ooo`@0oooo00L0gia800000000000004RL0=ooo`3oool0gia800D0000000<004RL0=ooo`3o
ool0Q03oool00`3OW4P0000000000003000000040018W03Oool0oooo0=nLB080000000T004Qd06D0
00000000001809cOo`3ooml0W4P00018W03Oool0M`3oool00:80oooo00@0oono07@000000000M;oo
103oool0103ogi`0B0000000001d_ol30?ooo`030=nLB018W=l0oooo0080oooo00<0okmd04RLg`3o
ool0Q@3oool00`3OW4P0B9cO0?ooo`020?ooo`0:0?noM018W=l0oooo0?oOW0180000000007Boo`2o
M000001d0;ooo`80oooo00<0okmd00000000M;l0M`3oool00:<0oooo00<0_g@000000000M;l0303o
ool00`3OW4P007Bo0?ooo`2:0?ooo`030=nLB000M;l0oooo00d0oooo00@0omnL04P0000004P0W=oo
MP3oool00980oooo00D0gia800000000000004RL0=ooo`020?ooo`040?oog`2LB0000000001d_`H0
oooo00@0gia800000000B9`0gooo303oool00`3OW4P0001d0;ooo`1h0?ooo`050=nLB00000000000
0018W03Oool00P3oool0103ooml0W4P000000000M;l70?ooo`030=nLB00007@0_ooo00`0oooo00@0
gia80000000004P0W=ooMP3oool00980oooo00D0omnL04P000000000001809cOo`020?ooo`060?no
M0000000000000000000B9`0gooo103oool0103ogi`0B0000000B02Lgol=0?ooo`030=nLB00007@0
_ooo07L0oooo00D0omnL04P000000000001809cOo`020?ooo`060?noM0000000000000000000B9`0
gooo1P3oool00`3OW4P0001d0;ooo`0:0?ooo`040=nLB00000000000001d_gL0oooo002B0?ooo`05
0?oog`2LB000000000000018W=l00P3oool01P2oM000000000000000B9`0W7@004RLg`@0oooo00@0
oono07@000000000M;oo3P3oool00`3OW4P0001809cOo`1f0?ooo`050?oog`2LB000000000000018
W=l00P3oool01P2oM000000000000000B9`0W7@004RLg`L0oooo00<0gia80000B02Lgol01`3oool0
1@3o_g@00000000000000000M;oo07P0oooo002C0?ooo`070;md0000000000000018W03Oool0oono
07@000020000000407Boo`3oool0giad09cOo`@0oooo00<0_g@000000000M;l03P3oool0103ookl0
M0000000M02ooomf0?ooo`070;md0000000000000018W03Oool0oono07@000020000000407Boo`3o
ool0giad09cOo`H0oooo00@0oono07@0000007@0_ooo2P3oool00`2oM000001809cOo`1f0?ooo`00
T`3oool02P3o_g@000000000000004P0W=oo0?oOW018000000000000M02oool70?ooo`040=nLB000
000004RL0=ooo`d0oooo00@0oooO09a800000000B9cOMP3oool02P3o_g@000000000000004P0W=oo
0?oOW018000000000000M02oool90?ooo`040?oog`2LB000000004RLg`H0oooo00P0oooO0=oOo`3o
ool0oooo0?oo_`1d0000001809cOogD0oooo002C0?ooo`060?oo_`1d000000000000000007@0_kmd
0P0000000`00B9`0gooo0?ooo`060?ooo`040?oOW0180000001809cOo`T0oooo00D0okmd001d_`3o
ool0oooo0?noM00200000003001d_`3oool0oooo07@0oooo00H0oono07@00000000000000000M02o
_g@2000000030018W03Oool0oooo00@0oooo00D0okmd001d_`3oool0oooo0?noM00200000003001d
_`3oool0oooo00D0oooo00@0_gA809cOo`3oool0okmd0P0000000`1d_ol0oooo0?ooo`1c0?ooo`00
T`3oool01P3ooml0W4P000000000000007Ad07@00080000000<0B9cO0?ooo`3oool0103oool00`3o
gi`0B000000000020000000307Boo`3oool0oooo00P0oooo00<0_g@00000000000000`0000000`1d
_ol0oooo0?ooo`1d0?ooo`060?oog`2LB000000000000000M7@0M0000P0000000`18W=l0oooo0?oo
o`050?ooo`030;md00000000000000<0000000<0M;oo0?ooo`3oool01@3oool00`3ookl0M0000000
0002000000030000B02Lgol0oooo07@0oooo002D0?ooo`080=nLB000000000000000M02ookl0M000
0000B02Lgol80?ooo`040?oog`2LB0000000001d_`/0oooo00D0gia8000000000000000007Boo`1h
0?ooo`080=nLB000000000000000M02ookl0M0000000B02Lgol80?ooo`050=nLB000000000000000
001d_ol02@3oool01@3ogi`0B000000000000000M;oo07H0oooo002D0?ooo`080?oOW01800000000
0000B02Lgol0oono07@00018W=nD0?ooo`080?oOW018000000000000B02Lgol0oono07@00018W=nA
0?ooo`00U03oool01@3ooml0W4P0000000000000B9cO0080oooo00<0gia80018W03Oool0TP3oool0
1@3ooml0W4P0000000000000B9cO0080oooo00<0gia80018W03Oool0S`3oool009D0oooo00X0_g@0
00000000000007Bo0?ooo`3ogml0gooO09a80000000007BoTP3oool02P2oM000000000000000M;l0
oooo0?oOg`3Ooml0W4P000000000M;n>0?ooo`00U@3oool01`3o_g@000000000000007@0_ooo0?oO
W0180000100000000`18W=l0oooo0?ooo`2>0?ooo`070?noM000000000000000M02oool0omnL04P0
00040000000304RLg`3oool0oooo08X0oooo002E0?ooo`050?oo_`1d000000000000001d_ol0U`3o
ool01@3ookl0M000000000000000M;oo09<0oooo002E0?ooo`050?oog`2LB000000000000018W=l0
U`3oool01@3ooml0W4P0000000000000B9cO09<0oooo002F0?ooo`050=nLB000000000000018W03O
ool0U`3oool01@3OW4P0000000000000B9`0gooo0980oooo002F0?ooo`050?oOW018000000000000
B02Lgol0U`3oool01@3ogi`0B0000000000004P0W=oo0980oooo002E0?ooo`030?oOW01800000000
0080000000<0M;oo0?ooo`3oool0U03oool00`3ogi`0B000000000020000000307Boo`3oool0oooo
0900oooo002G0?ooo`040?oo_`1d00000000001d_iP0oooo00@0oono07@00000000007BoTP3oool0
0?l0oooobP3oool00?l0oooobP3oool004@0oooo2@00002b0?ooo`T00000M03oool9000004@0oooo
00100?ooo`@000002@3oool500000:T0oooo100000090?ooo`H00000JP3oool4000000T0oooo1@00
000o0?ooo`00?@3oool300000180oooo0P00002T0?ooo`<000004`3oool2000006D0oooo0`00000B
0?ooo`800000?@3oool003X0oooo0`00000G0?ooo`800000X03oool2000001P0oooo00<000000?oo
o`3oool0G`3oool3000001L0oooo0P00000k0?ooo`00>03oool2000001`0oooo0P00002L0?ooo`80
00006`3oool2000005d0oooo0P00000L0?ooo`800000>@3oool003L0oooo00<000000?ooo`3oool0
7P3oool2000009P0oooo0P00000O0?ooo`800000FP3oool00`000000oooo0?ooo`0N0?ooo`800000
=`3oool003H0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`2D0?ooo`030000003o
ool0oooo0240oooo0P00001G0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0=03o
ool003@0oooo0P00000U0?ooo`030000003oool0oooo0940oooo0P00000V0?ooo`030000003oool0
oooo0580oooo0P00000U0?ooo`030000003oool0oooo03<0oooo000c0?ooo`030000003oool0oooo
02H0oooo0P00002@0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0D03oool00`00
0000oooo0?ooo`0V0?ooo`800000<`3oool00380oooo00<000000?ooo`3oool0:@3oool00`000000
oooo0?ooo`2<0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0CP3oool00`000000
oooo0?ooo`0Y0?ooo`030000003oool0oooo0300oooo000a0?ooo`030000003oool0oooo02/0oooo
00<000000?ooo`3oool0RP3oool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo04`0oooo
00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0_0?ooo`00<03oool00`000000oooo0?oo
o`0]0?ooo`030000003oool0oooo08P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?oo
o`1:0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0;P3oool00300oooo00<00000
0?ooo`3oool0;P3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0B@3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo02d0oooo000_0?oo
o`030000003oool0oooo02l0oooo00<000000?ooo`3oool0QP3oool00`000000oooo0?ooo`0_0?oo
o`030000003oool0oooo04P0oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`0]0?oo
o`00;P3oool00`000000oooo0?ooo`0a0?ooo`030000003oool0oooo08@0oooo00<000000?ooo`3o
ool0<@3oool00`000000oooo0?ooo`160?ooo`030000003oool0oooo0340oooo00<000000?ooo`3o
ool0;03oool002d0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`240?ooo`030000
003oool0oooo0380oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`0b0?ooo`030000
003oool0oooo02`0oooo000]0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0703o
ool7000000l0oooo1`00000?0?ooo`H000003`3oool7000001h0oooo00<000000?ooo`3oool0=03o
ool00`000000oooo0?ooo`0>0?ooo`@000001`3oool4000000L0oooo0`0000070?ooo`@000004@3o
ool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo02/0oooo000/0?ooo`030000003oool0
oooo03D0oooo00<000000?ooo`3oool06P3oool00`000000oooo0?ooo`040?ooo`<00000303oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00X0oooo0P0000060?ooo`800000303oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0=03oool0
0`000000oooo0?ooo`0=0?ooo`040000003oool0oooo0?ooo`8000001@3oool01@000000oooo0?oo
o`3oool0000000H0oooo00<000000?ooo`3oool00P0000060?ooo`030000003oool0oooo00800000
3`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo02X0oooo000/0?ooo`030000003o
ool0oooo03D0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003o
ool0oooo00L0oooo0P00000:0?ooo`030000003oool0oooo00L0oooo0P00000J0?ooo`030000003o
ool0oooo03H0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo0080oooo0P0000040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0303o
ool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo02X0oooo000/0?ooo`030000003oool0
oooo03H0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`090?ooo`800000203oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0603oool0
0`000000oooo0?ooo`0f0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`0<0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0:@3oool0
02/0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool0000000D0oooo00D00000
0?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0:@3oool002/0oooo00<0
00000?ooo`3oool0>03oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00/0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo00X0oooo00<0
00000?ooo`3oool0103oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo000000L0
oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool0:03oool002/0oooo00<000000?ooo`3oool0>03oool00`000000
oooo0?ooo`0D0?ooo`8000003P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo03P0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`050?ooo`040000003oool0oooo0000
00L0oooo00@000000?ooo`3oool000001`3oool010000000oooo0?ooo`0000070?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool0>03oool00`000000oooo0?ooo`0X0?ooo`00:P3oool0
0`000000oooo0?ooo`0i0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`0>0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03@3oool0
0`000000oooo0?ooo`0D0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`050?ooo`040000003oool0oooo000000L0oooo00@000000?ooo`3oool00000
1`3oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0>@3oool00`000000oooo0?ooo`0X0?ooo`00:P3oool00`000000oooo0?ooo`0P0?ooo`030;md
00000000000000<0000000<007Bo0?ooo`3oool04@3oool00`000000oooo0?ooo`0C0?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00d0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0O0?ooo`030?oo
g`2LB000000000D0000000<0M;oo0?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000
003oool0oooo00D0oooo00@000000?ooo`3oool000001`3oool010000000oooo0?ooo`0000070?oo
o`040000003oool00000000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0P
0?ooo`060=nLB000000000000000000004P0W=oo4`3oool00`000000oooo0?ooo`0X0?ooo`00:P3o
ool00`000000oooo0?ooo`0R0?ooo`040=nLB0000000001d0;oooa@0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`000000403oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo00h0oooo00D000000?ooo`3oool0oooo0000000A0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool08@3oool00`3OW4P00000000000030000
00030018W03Oool0oooo0100oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?oo
o`030000003oool0000000L0oooo00@000000?ooo`3oool000001`3oool00`000000oooo00000008
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool07`3oool02@3ookl0M00000000000
04P0W=oo0?oog`2LB00004RL0=ooo`0A0?ooo`030000003oool0oooo02P0oooo000Y0?ooo`030000
003oool0oooo02<0oooo00@0oono07@000000000M;oo503oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo00l0oooo00D000000?ooo`3oool0oooo0000000@0?ooo`8000000`3oool00`00
0000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0B
0?ooo`030000003oool0oooo02<0oooo00<0gia804RLg`3oool00P3oool00`3o_g@0B9cO0?ooo`0@
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo00000007
0?ooo`040000003oool0oooo000000L0oooo00<000000?ooo`000000203oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo0240oooo00<0_g@00000M02oool00P3oool00`3o_g@00000001d
_`0A0?ooo`030000003oool0oooo02P0oooo000Y0?ooo`030000003oool0oooo02@0oooo00<0_g@0
00000000M;l0503oool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00l0oooo00D00000
0?ooo`3oool0oooo0000000A0?ooo`050000003oool0oooo0?ooo`000000403oool01@000000oooo
0?ooo`3oool000000140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0T0?ooo`03
0=nLB000M;l0oooo01@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`03
0000003oool0000000L0oooo00@000000?ooo`00000000001`3oool00`000000oooo000000080?oo
o`030000003oool0oooo00P0oooo00<000000?ooo`3oool09P3oool0103ogi`0B0000000B02LgolA
0?ooo`030000003oool0oooo02L0oooo000Y0?ooo`030000003oool0oooo00`0oooo00/0gia80000
0000000004RL0=ooo`3oool0oono07@000000000001d0;ooo`030?ooo`050?noM000000000000018
W03Oool01@3oool0103OW4P000000018W03OoolC0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool03`3oool01@000000oooo0?ooo`3oool000000140oooo00D000000?ooo`3oool0oooo
0000000@0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000oooo0?ooo`0B0?ooo`03
0000003oool0oooo00`0oooo00/0gia800000000000004RL0=ooo`3oool0oono07@000000000001d
0;ooo`030?ooo`050?noM000000000000018W03Oool01P3oool00`3OW4P0001d0;ooo`0C0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool00`000000oooo000000070?ooo`04
0000003oool0oooo000000L0oooo00<000000?ooo`000000203oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo02H0oooo00@0gia80000000004P0W=oo4@3oool00`000000oooo0?ooo`0W
0?ooo`00:@3oool00`000000oooo0?ooo`0<0?ooo`080?oOW018000000000000B02Lgol0oooo0?oo
g`2LB0020000000507Boo`3oool0oooo0?ooo`2oM0000`0000000`0004P0W=oo0?ooo`030?ooo`04
0?oOW0180000001809cOoa<0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`0?0?oo
o`050000003oool0oooo0?ooo`000000403oool2000000<0oooo00<000000?ooo`3oool03P3oool0
1@000000oooo0?ooo`3oool000000140oooo00<000000?ooo`3oool04P3oool00`000000oooo0?oo
o`0<0?ooo`080?oOW018000000000000B02Lgol0oooo0?oog`2LB0020000000507Boo`3oool0oooo
0?ooo`2oM0000`0000000`0004P0W=oo0?ooo`050?ooo`030=nLB00007@0_ooo0180oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`030000003oool0000000L0oooo00@00000
0?ooo`3oool000001`3oool00`000000oooo000000080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0303oool02`3OW4P0000000000000B9`0gooo0?ooo`3ookl0M0000000000007@0
_ooo00<0oooo00D0okmd00000000000004RL0=ooo`060?ooo`040=nLB00000000000001d_a80oooo
00<000000?ooo`3oool09`3oool002T0oooo00<000000?ooo`3oool0303oool01@3ooml0W4P00000
00000000M;oo0080oooo00@0gia800000000000007Bo0`3oool01`3OW4P0000000000000M;l0oooO
09a8M02oool00`3oool0103ookl0M0000000001d_olC0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool03P3oool01@000000oooo0?ooo`3oool000000100oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`0000004@3oool00`000000
oooo0?ooo`0B0?ooo`030000003oool0oooo00`0oooo00D0oooO09a800000000000007Boo`020?oo
o`040=nLB00000000000001d_`<0oooo00L0gia800000000000007Bo0?oog`2LB7@0_ooo00H0oooo
00<0gia80000B02Lgol04@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`0000001`3oool010000000oooo0?ooo`0000070?ooo`030000003oool0000000P0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0<0?ooo`080?oOW018000000000000
B02Lgol0oooo0?oog`2LB0020000000507Boo`3oool0oooo0?ooo`2oM0000`0000000`0004P0W=oo
0?ooo`020?ooo`050?noM000000000000000001d_ol04`3oool00`000000oooo0?ooo`0W0?ooo`00
:@3oool00`000000oooo0?ooo`0=0?ooo`040=nLB00000000000001d_`80oooo00D0omnL04P00000
0000001d0;ooo`020?ooo`050?oOW018000000000000M02oool01P3oool00`2oM0000000001d_`0C
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool03P3oool01@000000oooo0?ooo`3o
ool000000100oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0>0?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo00d0oooo00@0gia800000000000007Bo0P3oool01@3ogi`0B0000000000007@0_ooo0080
oooo00D0omnL04P000000000001d0;ooo`080?ooo`040?oo_`1d0000001d0;oooa00oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`030000003oool0000000L0oooo00@00000
0?ooo`3oool000001`3oool00`000000oooo000000080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0303oool01@3ooml0W4P0000000000000M;oo0080oooo00@0gia8000000000000
07Bo0`3oool01`3OW4P0000000000000M;l0oooO09a8M02oool01P3oool00`2oM000001809cOo`0A
0?ooo`030000003oool0oooo02L0oooo000Y0?ooo`030000003oool0oooo00d0oooo00P0okmd0000
00000000001809cOo`3oool0oono07@00080000000D0M;oo0?ooo`3oool0oooO09a8000200000003
07Boo`3oool0oooo00@0oooo00@0gia800000000B9`0gooo4P3oool00`000000oooo0?ooo`0C0?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0=0?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00d0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0=0?oo
o`080?noM000000000000000B02Lgol0oooo0?oo_`1d00020000000507Boo`3oool0oooo0?oog`2L
B0000P0000000`1d_ol0oooo0?ooo`060?ooo`040?oog`2LB000000004RLga00oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`050?ooo`040000003oool0oooo000000L0oooo00@00000
0?ooo`3oool000001`3oool010000000oooo0?ooo`0000070?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool03@3oool0103OW4P0000000000000M;l20?ooo`050?oOW018000000000000
M02oool00P3oool01@3ogi`0B0000000000007@0_ooo00@0oooo00P0oooO0=oOo`3oool0oooo0?oo
_`1d0000001809cOoa00oooo00<000000?ooo`3oool09`3oool002T0oooo00<000000?ooo`3oool0
3@3oool01@3ookl0M00000000000000007Bo0080oooo00D0_g@000000000000004RL0=ooo`020?oo
o`040;md000000000000001d_`H0oooo00@0omnL04P0000004P0W=oo4P3oool00`000000oooo0?oo
o`0C0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool04P3oool2000000l0oooo00D0
oono07@0000000000000001d_`020?ooo`050;md0000000000000018W03Oool00P3oool0102oM000
000000000000M;l40?ooo`050?noM000M;l0oooo0?ooo`3o_g@00P0000000`00M;l0oooo0?ooo`0>
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`00
00070?ooo`040000003oool0oooo000000L0oooo00@000000?ooo`3oool000001`3oool00`000000
oooo0?ooo`070?ooo`8000003`3oool0203o_g@000000000000004P0W=oo0?ooo`3ookl0M0000P00
00001@1d_ol0oooo0?ooo`3ooml0W4P00080000000<0M;oo0?ooo`3oool00`3oool0102oM4P0W=oo
0?ooo`3o_g@20000000307Boo`3oool0oooo00h0oooo00<000000?ooo`3oool09`3oool002T0oooo
00<000000?ooo`3oool03P3oool01`2oM00000000000000004P0W=oo0?ooo`3o_g@00`0000001000
M;l0oooo0?ooo`3o_g@2000000030000M02oool0oooo0080oooo00<0omnL04P0000000000P000000
0`1d_ol0oooo0?ooo`0@0?ooo`800000503oool00`000000oooo0?ooo`0=0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00`0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0;0?ooo`030000003oool0
oooo01<0oooo0P00000@0?ooo`070;md0000000000000000B02Lgol0oooo0?noM00300000004001d
_`3oool0oooo0?noM080000000<0001d0;ooo`3oool00`3oool00`2oM00000000000000300000003
07Boo`3oool0oooo00h0oooo0P0000090?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3o
ool000001`3oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo000000H0oooo00<0
00000?ooo`3oool0203oool2000000l0oooo00D0oono07@0000000000000001d_`020?ooo`050;md
0000000000000018W03Oool00P3oool0102oM000000000000000M;l50?ooo`030?oo_`1d00000000
0080000000<0001809cOo`3oool03`3oool00`000000oooo0?ooo`0W0?ooo`00:@3oool00`000000
oooo0?ooo`0>0?ooo`0?0?noM000000000000018M01dM9`0gooo0?oOW018000000000000B02LW7@0
M;oo0?ooo`3ookl0M0000080000000<0M;oo0?ooo`3oool00`3oool0103ooml0W4P000000000M;lB
0?ooo`030000003oool0000001@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0C
0?ooo`800000403oool03`3o_g@0000000000000B7@0M7BL0=ooo`3ogi`0B0000000000004P0W9ad
07Boo`3oool0oono07@000020000000307Boo`3oool0oooo00<0oooo00D0gia80000000000000000
07Boo`0A0?ooo`8000002P3oool00`000000oooo0?ooo`040?ooo`040000003oool0oooo000000H0
oooo00D000000?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000001P3oool0
0`000000oooo0?ooo`080?ooo`800000403oool01`2oM00000000000000004P0W=oo0?ooo`3o_g@0
0`0000001000M;l0oooo0?ooo`3o_g@2000000030000M02oool0oooo00@0oooo00D0omnL04P00000
0000000007Boo`0A0?ooo`030000003oool0oooo02L0oooo000Y0?ooo`030000003oool0oooo00h0
oooo00P0oono07@000000000001809boW018M;l0oooO09a80080000000D0M;oo0;mdB02Lgol0oooo
0;md000200000003001d_`3oool0oooo01T0oooo00<000000?ooo`000000503oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo01<0oooo00<000000?ooo`000000403oool0203ookl0M000
0000000004P0W;nL04Qd_`3ooml0W4P00P0000001@1d_ol0_gA809cOo`3oool0_g@00080000000<0
07Bo0?ooo`3oool06@3oool2000000X0oooo00<000000?ooo`3oool0103oool01@000000oooo0?oo
o`3oool0000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`080?ooo`800000
403oool03`3o_g@0000000000000B7@0M7BL0=ooo`3ogi`0B0000000000004P0W9ad07Boo`3oool0
oono07@000020000000307Boo`3oool0oooo01T0oooo00<000000?ooo`3oool09`3oool002T0oooo
00<000000?ooo`3oool03P3oool0203ooml0W4P0000000000000M;oo0?noM018W=l0gia80P000000
1@00M;l0oooO09a8B02Lgol0gia80080000000<004RL0=ooo`3oool06@3oool010000000oooo0?oo
o`00000D0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool04P3oool010000000oooo
0?ooo`00000@0?ooo`080?oog`2LB00000000000001d_ol0okmd04RLg`3OW4P200000005001d_`3o
oml0W4Q809cOo`3OW4P00P0000000`00B9`0gooo0?ooo`0I0?ooo`030000003oool0000000T0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool000000100oooo
00P0oono07@000000000001809boW018M;l0oooO09a80080000000D0M;oo0;mdB02Lgol0oooo0;md
000200000003001d_`3oool0oooo01P0oooo00<000000?ooo`3oool0:03oool002T0oooo00<00000
0?ooo`3oool03@3oool00`3ogi`0B0000000000200000003001d_`3oool0gia800<0000000@004RL
0=ooo`3ooml0W4P00`0000000`0007@0_ooo0?ooo`0I0?ooo`040000003oool0oooo000001D0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0C0?ooo`050000003oool0oooo0?ooo`00
00003P3oool00`3ogi`0B0000000000200000003001d_`3oool0gia800<0000000@004RL0=ooo`3o
oml0W4P00`0000000`0007@0_ooo0?ooo`0I0?ooo`030000003oool0000000T0oooo00<000000?oo
o`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`080?ooo`030000003oool000000100oooo00P0oooO09a8
00000000000007Boo`3o_g@0B9cO0=nLB080000000D007Bo0?oog`2LB4P0W=oo0=nLB00200000003
0018W03Oool0oooo01P0oooo00<000000?ooo`3oool0:03oool002X0oooo00<000000?ooo`3oool0
3P3oool01`3ooml0W4P00000000007@0_ooo0?ooo`3OW4P00P0000001@18W=l0oooo0?ooo`3oool0
gia80080000000<0B9cO0?ooo`3oool06@3oool010000000oooo0?ooo`00000F0?ooo`030000003o
ool0oooo00D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`060?ooo`030000003o
ool0oooo00X0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`090?ooo`030000003o
ool0oooo00L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00h0oooo00L0oooO09a800000000001d0;ooo`3oool0gia80080000000D0B9cO0?ooo`3o
ool0oooo0=nLB0020000000304RLg`3oool0oooo01T0oooo00<000000?ooo`0000002P3oool01@00
0000oooo0?ooo`3oool0000000H0oooo00D000000?ooo`3oool0oooo000000050?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00P0oooo00@000000?ooo`3oool000003P3oool00`3ogi`0B0000000000200000003001d
_`3oool0gia800<0000000@004RL0=ooo`3ooml0W4P00`0000000`0007@0_ooo0?ooo`0H0?ooo`03
0000003oool0oooo02P0oooo000Z0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0
103oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
4`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03T0oooo00@000000?ooo`3oool0
00002P3oool01@000000oooo0?ooo`3oool0000000H0oooo00D000000?ooo`3oool0oooo00000006
0?ooo`050000003oool0oooo0?ooo`0000001@3oool01@000000oooo0?ooo`3oool0000000/0oooo
00@000000?ooo`3oool00000403oool01`3ooml0W4P00000000007@0_ooo0?ooo`3OW4P00P000000
1@18W=l0oooo0?ooo`3oool0gia80080000000<0B9cO0?ooo`3oool0603oool00`000000oooo0?oo
o`0X0?ooo`00:`3oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo00<0oooo0P00000D
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`03
0?ooo`8000003P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00d0oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0>03oool01@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`0000001`3o
ool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo000000L0oooo00@000000?ooo`3o
ool000002P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool0:03oool002/0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo0180oooo00D000000?ooo`3oool0oooo0000000@0?ooo`8000000P3oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool03P3oool01@00
0000oooo0?ooo`3oool0000001D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0g
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0203oool00`000000oooo00000008
0?ooo`8000002@3oool00`000000oooo000000070?ooo`040000003oool0oooo000000X0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo02T0oooo000[
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0B
0?ooo`<000004`3oool2000001<0oooo1000000B0?ooo`030000003oool0000001@0oooo0P000008
0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00L0oooo0P00000:0?ooo`030000003oool0oooo00L0oooo0P00000:
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`0Y
0?ooo`00;03oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00T0oooo00<000000?oo
o`3oool04P3oool300000180oooo0P00000D0?ooo`800000503oool200000180oooo0P00000;0?oo
o`030000003oool0oooo03H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`080?oo
o`800000203oool2000000X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?oo
o`030000003oool0oooo00D0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0Z0?oo
o`00;03oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3o
ool03P3oool3000000<0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0200000100
oooo0P0000020?ooo`030000003oool0oooo00l0oooo0P0000020?ooo`8000003`3oool00`000000
oooo0?ooo`0;0?ooo`030000003oool0oooo03D0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`060?ooo`040000003oool0oooo000000L0oooo00<000000?ooo`000000203oool00`00
0000oooo000000080?ooo`030000003oool0000000L0oooo00<000000?ooo`3oool01P3oool00`00
0000oooo0?ooo`0e0?ooo`030000003oool0oooo02X0oooo000/0?ooo`030000003oool0oooo03@0
oooo00<000000?ooo`3oool0303oool5000000P0oooo0`0000070?ooo`<000002@3oool4000000D0
oooo100000090?ooo`<000001@3oool5000000P0oooo100000060?ooo`<00000203oool4000000l0
oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`060?ooo`<000000`3oool2000000<0
oooo0P0000040?ooo`8000000`3oool2000000@0oooo0P0000030?ooo`800000103oool2000000<0
oooo0P0000030?ooo`<00000203oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo02/0
oooo000]0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0403oool9000000d0oooo
2@00000=0?ooo`T000003@3oool8000000d0oooo2000000C0?ooo`030000003oool0oooo03<0oooo
00<000000?ooo`3oool02@3oool4000000L0oooo100000070?ooo`D000001P3oool4000000H0oooo
1@00000;0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0:`3oool002d0oooo00<0
00000?ooo`3oool0<P3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo0380oooo00<0
00000?ooo`3oool0A03oool00`000000oooo0?ooo`0b0?ooo`030000003oool0oooo02`0oooo000^
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0Q03oool00`000000oooo0?ooo`0a
0?ooo`030000003oool0oooo04H0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0/
0?ooo`00;`3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo08H0oooo00<000000?oo
o`3oool0;`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo02l0oooo00<000000?oo
o`3oool0;@3oool00300oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`280?ooo`03
0000003oool0oooo02d0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`0]0?ooo`03
0000003oool0oooo02h0oooo000`0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0
R@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool0
;03oool00`000000oooo0?ooo`0_0?ooo`00<@3oool00`000000oooo0?ooo`0[0?ooo`030000003o
ool0oooo08X0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1<0?ooo`030000003o
ool0oooo02/0oooo00<000000?ooo`3oool0;`3oool00380oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`2<0?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0CP3oool0
0`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0300oooo000c0?ooo`030000003oool0oooo
02H0oooo0P00002@0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0D03oool00`00
0000oooo0?ooo`0V0?ooo`800000<`3oool003@0oooo0P00000U0?ooo`030000003oool0oooo0940
oooo0P00000V0?ooo`030000003oool0oooo0580oooo0P00000U0?ooo`030000003oool0oooo03<0
oooo000f0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0U03oool200000280oooo
0P00001G0?ooo`030000003oool0oooo0240oooo00<000000?ooo`3oool0=03oool003L0oooo0P00
000O0?ooo`800000V@3oool2000001h0oooo0P00001J0?ooo`8000007`3oool2000003L0oooo000i
0?ooo`8000006P3oool3000009d0oooo00<000000?ooo`3oool06@3oool2000005h0oooo0P00000J
0?ooo`<00000>@3oool003/0oooo0P00000E0?ooo`<00000X@3oool2000001H0oooo0`00001R0?oo
o`8000005@3oool3000003`0oooo000m0?ooo`<000003`3oool300000:H0oooo0`00000@0?ooo`<0
0000I`3oool3000000l0oooo0`00000o0?ooo`00@03oool7000000<0oooo1@00002/0?ooo`L00000
0`3oool6000006d0oooo1`0000030?ooo`D00000@P3oool004L0oooo0`00002h0?ooo`<00000NP3o
ool3000004L0oooo003o0?ooolX0oooo0000\
\>"],
  ImageRangeCache->{{{0, 456.875}, {94.25, 0}} -> {-1.66514, -0.392001, \
0.00962061, 0.00962061}}],

Cell[TextData[{
  "Suppose that three masses move along a line subject to pairwise forces \
represented by springs.  The figure above displays the interactions between \
neighboring masses, but omits a possible interaction between the outer masses \
with spring constant ",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 3\)\)]],
  ".  One can assume, without loss of generality, that the resting lengths of \
both interior springs are equal to ",
  Cell[BoxData[
      \(TraditionalForm\`b\)]],
  ".  For simplicity we begin by assuming that the two outer masses are \
equal, ",
  Cell[BoxData[
      \(TraditionalForm\`m\_3 = m\_1\)]],
  ", but not necessarily equal to the central mass ",
  Cell[BoxData[
      \(TraditionalForm\`m\_2\)]],
  ", and that ",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 2\) = k\_\(2, 3\)\)]],
  " while ",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 3\) = 0\)]],
  ".  With these simplifications we can obtain simple but nontrivial symbolic \
expressions for the frequencies and normal modes of vibration; two different \
methods are used to illustrate several ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " functions.  Numerical results can also be obtained easily for more \
general systems, but symbolic expressions become unwieldy."
}], "Text"],

Cell[BoxData[{
    \(\(ClearAll["\<Global`*\>"];\)\), "\n", 
    \(\(Needs["\<Utilities`Notation`\>"];\)\)}], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(m\_1\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(m\_2\),
          NotationBoxTag,
          Editable->True], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(m\_3\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "\n", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(x\&\[DoubleDot]\_1\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(x\&\[DoubleDot]\_2\),
          NotationBoxTag,
          Editable->True], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(x\&\[DoubleDot]\_3\),
          NotationBoxTag,
          Editable->True], "]"}]}]], "Input"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(k\_1\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(k\_2\),
          NotationBoxTag,
          Editable->True], "]"}], ";", "\n", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(k\_\(1, 2\)\),
          NotationBoxTag,
          Editable->True], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(k\_\(2, 3\)\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(k\_\(1, 3\)\),
          NotationBoxTag,
          Editable->True], "]"}]}]], "Input"],

Cell["\<\
The equations of motion for the general system, assuming that the spring \
force is proportional to its length change, are obtained from Newton's second \
law as follows.\
\>", "Text"],

Cell[BoxData[
    \(\(equations = \ {\n\t\t\(m\_1\) x\&\[DoubleDot]\_1 \[Equal] 
            k\_\(1, 2\)\ \((x\_2 - x\_1 - b)\) + \(k\_\(1, 3\)\) \((x\_3 - 
                    x\_1 - 2  b)\), \n\t\t\(m\_2\) 
              x\&\[DoubleDot]\_2 \[Equal] \(k\_\(2, 3\)\) \((x\_3 - x\_2 - 
                    b)\) - \(k\_\(1, 2\)\) \((x\_2 - x\_1 - 
                    b)\), \n\t\t\(m\_3\) 
              x\&\[DoubleDot]\_3 \[Equal] \(k\_\(2, 3\)\) \((x\_2 - x\_3 + 
                    b)\) - \(k\_\(1, 3\)\) \((x\_3 - x\_1 - 
                    2  b)\)\ };\)\)], "Input"],

Cell["For our initial special case, these equations reduce to: ", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq1 = 
      equations /. {m\_3 \[Rule] m\_1, k\_\(1, 2\) \[Rule] k, 
          k\_\(2, 3\) \[Rule] k, k\_\(1, 3\) \[Rule] 0}\)], "Input"],

Cell[BoxData[
    \({m\_1\ x\&\[DoubleDot]\_1 \[Equal] k\ \((\(-b\) - x\_1 + x\_2)\), 
      m\_2\ x\&\[DoubleDot]\_2 \[Equal] \(-k\)\ \((\(-b\) - x\_1 + x\_2)\) + 
          k\ \((\(-b\) - x\_2 + x\_3)\), 
      m\_1\ x\&\[DoubleDot]\_3 \[Equal] k\ \((b + x\_2 - x\_3)\)}\)], "Output"]
}, Closed]],

Cell["\<\
It is useful to eliminate the resting length with the following change of \
variables.\
\>", "Text"],

Cell[BoxData[
    RowBox[{
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Eta]\&\[DoubleDot]\_1\),
          NotationBoxTag,
          Editable->True], "]"}], ";", " ", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Eta]\&\[DoubleDot]\_2\),
          NotationBoxTag,
          Editable->True], "]"}], ";", 
      RowBox[{"Symbolize", "[", 
        TagBox[\(\[Eta]\&\[DoubleDot]\_3\),
          NotationBoxTag,
          Editable->True], "]"}]}]], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq2\  = 
      eq1\  /. {x\_1 \[Rule] \[Eta]\_1 - b, x\_2 \[Rule] \[Eta]\_2, 
          x\_3 \[Rule] \[Eta]\_3 + b, 
          x\&\[DoubleDot]\_1 \[Rule] \[Eta]\&\[DoubleDot]\_1, 
          x\&\[DoubleDot]\_2 \[Rule] \[Eta]\&\[DoubleDot]\_2, 
          x\&\[DoubleDot]\_3 \[Rule] \[Eta]\&\[DoubleDot]\_3}\)], "Input"],

Cell[BoxData[
    \({m\_1\ \[Eta]\&\[DoubleDot]\_1 \[Equal] 
        k\ \((\(-\[Eta]\_1\) + \[Eta]\_2)\), 
      m\_2\ \[Eta]\&\[DoubleDot]\_2 \[Equal] \(-k\)\ \((\(-\[Eta]\_1\) + \
\[Eta]\_2)\) + k\ \((\(-\[Eta]\_2\) + \[Eta]\_3)\), 
      m\_1\ \[Eta]\&\[DoubleDot]\_3 \[Equal] 
        k\ \((\[Eta]\_2 - \[Eta]\_3)\)}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Normal modes are defined to be solution vectors of the form ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta](t) = \(\[Eta](0)\) 
          Exp[\(-\[ImaginaryI]\)\ \[Omega]\ t]\)]],
  ".  Thus, if we make the substitution ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\&\[DoubleDot] \[Rule] \(-\[Omega]\^2\) \[Eta]\
\)]],
  " for each coordinate, the time dependence divides out.  It is also \
convenient to replace the equations by differences so that we can extract the \
coefficient matrix using ",
  StyleBox["Coefficient", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(eq3 = 
      eq2\  /. {\[Eta]\&\[DoubleDot]\_1 \[Rule] \ \(-\[Omega]\^2\) \[Eta]\_1, \
\[Eta]\&\[DoubleDot]\_2 \[Rule] \ \(-\[Omega]\^2\) \[Eta]\_2, \[Eta]\&\
\[DoubleDot]\_3 \[Rule] \ \(-\[Omega]\^2\) \[Eta]\_3}\)], "Input"],

Cell[BoxData[
    \({\(-m\_1\)\ \[Omega]\^2\ \[Eta]\_1 \[Equal] 
        k\ \((\(-\[Eta]\_1\) + \[Eta]\_2)\), \(-m\_2\)\ \[Omega]\^2\ \
\[Eta]\_2 \[Equal] \(-k\)\ \((\(-\[Eta]\_1\) + \[Eta]\_2)\) + 
          k\ \((\(-\[Eta]\_2\) + \[Eta]\_3)\), \(-m\_1\)\ \[Omega]\^2\ \[Eta]\
\_3 \[Equal] k\ \((\[Eta]\_2 - \[Eta]\_3)\)}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(\(matrix\  = \(Coefficient[#, {\[Eta]\_1, \[Eta]\_2, \[Eta]\_3}] &\)\  \
/@ \ \((eq3 /. Equal \[Rule] Subtract)\);\)\), "\n", 
    \(matrix // MatrixForm\)}], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(k - m\_1\ \[Omega]\^2\), \(-k\), "0"},
            {\(-k\), \(2\ k - m\_2\ \[Omega]\^2\), \(-k\)},
            {"0", \(-k\), \(k - m\_1\ \[Omega]\^2\)}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell[TextData[{
  "Therefore, the vibrational frequencies for the normal modes are solutions \
to the equations ",
  StyleBox["matrix.\[Eta]\[Equal]0", "InlineInput"],
  ", which is equivalent to the condition ",
  StyleBox["Det[matrix]\[Equal]0", "InlineInput"],
  ".  For an ",
  StyleBox["n",
    FontSlant->"Italic"],
  "-dimensional system, this condition yields an ",
  Cell[BoxData[
      \(TraditionalForm\`n\^th\)]],
  " degree polynomial equation known as the ",
  StyleBox["secular",
    FontSlant->"Italic"],
  " or ",
  StyleBox["characteristic equation",
    FontSlant->"Italic"],
  " for the system.  To avoid negative frequencies in the present problem, it \
is easiest to solve for ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\^2\)]],
  " instead of ",
  Cell[BoxData[
      \(TraditionalForm\`\[Omega]\)]],
  " itself and to take the positive root later."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(freqRule2\  = 
      Solve[Det[matrix /. \[Omega]\^2 \[Rule] \[Omega]2] \[Equal] 
            0, \[Omega]2]\  // Simplify\)], "Input"],

Cell[BoxData[
    \({{\[Omega]2 \[Rule] 0}, {\[Omega]2 \[Rule] k\/m\_1}, {\[Omega]2 \[Rule] 
          k\ \((1\/m\_1 + 2\/m\_2)\)}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(freqRule\  = \ \[Omega] \[Rule] \@\[Omega]2 /. freqRule2\)], "Input"],

Cell[BoxData[
    \({\[Omega] \[Rule] 
        0, \[Omega] \[Rule] \@\(k\/m\_1\), \[Omega] \[Rule] \@\(k\ \((1\/m\_1 \
+ 2\/m\_2)\)\)}\)], "Output"]
}, Closed]],

Cell["\<\
The modes are then found by substituting the frequencies back into the \
equations of motion and solving for the amplitudes.  Because the overall \
amplitude is arbitrary for linear systems, it is useful to set one of the \
amplitudes to unity (but not the central one because of the symmetry of the \
problem).\
\>", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[
      Solve[\(eq3 /. 
            freqRule\[LeftDoubleBracket]
              i\[RightDoubleBracket]\) /. \[Eta]\_1 \[Rule] 
            1, {\[Eta]\_2, \[Eta]\_3}], {i, 1, 3}]\)], "Input"],

Cell[BoxData[
    \({{{\[Eta]\_3 \[Rule] 1, \[Eta]\_2 \[Rule] 
            1}}, {{\[Eta]\_3 \[Rule] \(-1\), \[Eta]\_2 \[Rule] 
            0}}, {{\[Eta]\_3 \[Rule] 
            1, \[Eta]\_2 \[Rule] \(-\(\(2\ m\_1\)\/m\_2\)\)}}}\)], "Output"]
}, Closed]],

Cell["\<\
The nature of these solutions should now be obvious.  The zero-frequency mode \
represents the motion of the entire system together without any internal \
vibration.  In the second mode the central mass remains stationary while the \
two outer mass vibrate in opposite directions symmetrically.  For the third \
mode the two outer masses move together while the central mass moves in the \
opposite direction, with the relative amplitudes proportional to the ratio of \
masses.\
\>", "Text"],

Cell[TextData[{
  "There is a more elegant way to obtain the eigenvalues and eigenvectors.  \
We can transform the system of equations to the form ",
  Cell[BoxData[
      \(TraditionalForm\`h . \[Eta] \[Equal] \(-\[Omega]\^2\)\ \[Eta]\)]],
  " and then use ",
  StyleBox["Eigensystem", "InlineInput",
    FontWeight->"Bold"],
  " to obtain both the eigenvalues ",
  Cell[BoxData[
      \(TraditionalForm\`\[Lambda] = \(-\[Omega]\^2\)\)]],
  " and the corresponding eigenvectors directly.  First we extract the \
right-hand sides of ",
  StyleBox["eq2", "InlineInput"],
  " and divide out the masses."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(rhs = \(#\[LeftDoubleBracket]2\[RightDoubleBracket] &\)\  /@ 
            eq2\ /{m\_1, m\_2, m\_1} // Simplify\)], "Input"],

Cell[BoxData[
    \({\(k\ \((\(-\[Eta]\_1\) + \[Eta]\_2)\)\)\/m\_1, \(k\ \((\[Eta]\_1 - 2\ \
\[Eta]\_2 + \[Eta]\_3)\)\)\/m\_2, \(k\ \((\[Eta]\_2 - \
\[Eta]\_3)\)\)\/m\_1}\)], "Output"]
}, Closed]],

Cell["Then we construct the coefficient matrix.  ", "Text"],

Cell[BoxData[
    \(\(h = \(Coefficient[#, {\[Eta]\_1, \[Eta]\_2, \[Eta]\_3}] &\) /@ 
          rhs;\)\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[
    \(h // MatrixForm\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(-\(k\/m\_1\)\), \(k\/m\_1\), "0"},
            {\(k\/m\_2\), \(-\(\(2\ k\)\/m\_2\)\), \(k\/m\_2\)},
            {"0", \(k\/m\_1\), \(-\(k\/m\_1\)\)}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell["Finally, we evaluate the eigenvalues and eigenvectors", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \({\[Lambda], vectors} = Eigensystem[h]\)], "Input"],

Cell[BoxData[
    \({{0, \(-\(k\/m\_1\)\), \(\(-2\)\ k\ m\_1 - k\ m\_2\)\/\(m\_1\ m\_2\)}, \
{{1, 1, 1}, {\(-1\), 0, 1}, {1, \(-\(\(2\ m\_1\)\/m\_2\)\), 1}}}\)], "Output"]
}, Closed]],

Cell["and compute the frequencies.", "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(frequencies = Sqrt[\(-\[Lambda]\)]\)], "Input"],

Cell[BoxData[
    \({0, \@\(k\/m\_1\), \@\(-\(\(\(-2\)\ k\ m\_1 - k\ m\_2\)\/\(m\_1\ m\_2\)\
\)\)}\)], "Output"]
}, Closed]],

Cell[TextData[{
  "Note that ",
  StyleBox["Eigensystem", "InlineInput",
    FontWeight->"Bold"],
  " does not normalize the eigenvectors.  It is often useful to construct a \
matrix containing the normalized eigenvectors in column form."
}], "Text"],

Cell[CellGroupData[{

Cell[BoxData[
    \(modes\  = \ vectors/\((\(Sqrt[# . #] &\)\  /@ vectors)\)\)], "Input"],

Cell[BoxData[
    \({{1\/\@3, 1\/\@3, 1\/\@3}, {\(-\(1\/\@2\)\), 0, 
        1\/\@2}, {1\/\@\(2 + \(4\ m\_1\%2\)\/m\_2\%2\), \(-\(\(2\ \
m\_1\)\/\(\@\(2 + \(4\ m\_1\%2\)\/m\_2\%2\)\ m\_2\)\)\), 
        1\/\@\(2 + \(4\ m\_1\%2\)\/m\_2\%2\)}}\)], "Output"]
}, Closed]],

Cell[CellGroupData[{

Cell[BoxData[
    \(Transpose[modes] // MatrixForm\)], "Input"],

Cell[BoxData[
    TagBox[
      RowBox[{"(", "\[NoBreak]", GridBox[{
            {\(1\/\@3\), \(-\(1\/\@2\)\), \(1\/\@\(2 + \(4\ \
m\_1\%2\)\/m\_2\%2\)\)},
            {\(1\/\@3\), 
              "0", \(-\(\(2\ m\_1\)\/\(\@\(2 + \(4\ m\_1\%2\)\/m\_2\%2\)\ \
m\_2\)\)\)},
            {\(1\/\@3\), \(1\/\@2\), \(1\/\@\(2 + \(4\ \
m\_1\%2\)\/m\_2\%2\)\)}
            },
          RowSpacings->1,
          ColumnSpacings->1,
          ColumnAlignments->{Left}], "\[NoBreak]", ")"}],
      Function[ BoxForm`e$, 
        MatrixForm[ BoxForm`e$]]]], "Output"]
}, Closed]],

Cell["\<\
Of course, this procedure can be generalized and streamlined considerably if \
we formulate the problem in terms of the appropriate matrices from the \
beginning!\
\>", "Text"],

Cell["\<\
Develop the coupling matrix for arbitrary parameters in the linear triatomic \
model.  It is more convenient for this and later exercises not to divide out \
the masses.\
\>", "Exercise"],

Cell[TextData[{
  "Write a function which returns the eigenfrequencies and normal modes given \
a coupling matrix, ",
  Cell[BoxData[
      \(TraditionalForm\`h\)]],
  ", for which ",
  Cell[BoxData[
      \(TraditionalForm\`h . \[Eta] == 
        m . \[Eta]\&\[DoubleDot] == \(-\[Omega]\^2\) m . \[Eta]\)]],
  "  where ",
  Cell[BoxData[
      \(TraditionalForm\`\[Eta]\)]],
  " is a coordinate vector and ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " is a diagonal matrix carrying the masses. "
}], "Exercise"],

Cell[TextData[{
  "Find the eigenfrequencies and normal modes assuming {",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 2\) = \(k\_\(2, 3\) = k\)\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 3 = 0\)\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1 = m\_3\)]],
  "}.  Verify that the previous results are recovered.  Plot the frequency \
and ratio of amplitudes for the mode in which the central mass vibrates \
against the outer masses as a function of the ratio of masses."
}], "Exercise"],

Cell[TextData[{
  "Find the eigenfrequencies and normal modes assuming {",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 2\) = \(k\_\(2, 3\) = k\_1\)\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`m\_1 = m\_3\)]],
  "} and including the interaction ",
  Cell[BoxData[
      \(TraditionalForm\`k\_\(1, 3\) = k\_2\)]],
  " between the outer masses.  Verify that the previous results are recovered \
when the spring constant for the new force is set to zero.  What is the \
effect of ",
  Cell[BoxData[
      \(TraditionalForm\`k\_2\)]],
  "?  "
}], "Exercise"],

Cell["\<\
Generalize to the case of three arbitrary masses.  You will probably find \
that symbolic expressions for the eigenvalues are too lengthy to display \
comfortably and are difficult to simplify.  Nevertheless, you can still plot \
interesting quantities.  For example, plot the two nonzero frequencies \
together as functions of the mass ratios (either in 3D or in 2D as functions \
of one ratio for several values of the other).  \
\>", "Exercise"]
}, Closed]],

Cell[CellGroupData[{

Cell["More examples", "Subsubsection"],

Cell[TextData[{
  ButtonBox["Paschen.nb",
    ButtonData:>{"Applications/Paschen.nb", None},
    ButtonStyle->"Hyperlink"],
  " \[LongDash] investigates the energy levels of the hydrogen atom in \
external magnetic fields, including both weak and strong-field limits."
}], "Text"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{

Cell["Exercises", "ExerciseMain"],

Cell[CellGroupData[{

Cell["Forgotten form of quadratic theorem", "ExerciseTitle"],

Cell[TextData[{
  "Everyone knows that the solutions to the quadratic equation ",
  Cell[BoxData[
      \(TraditionalForm\`a\ x\^2 + b\ x + c \[Equal] 0\)]],
  " are ",
  Cell[BoxData[
      \(TraditionalForm\`x = \(\(-b\) \[PlusMinus] \@\(b\^2 - 4  a\ \
c\)\)\/\(2  a\)\)]],
  ", but there is another form that is sometimes useful for hand calculation. \
 Prove that the solutions can also be expressed in the form ",
  Cell[BoxData[
      \(TraditionalForm\`x = \(2  c\)\/\(\(\[MinusPlus]\@\(b\^2 - 4  a\ c\)\) \
- b\)\)]],
  "."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Amusing factorization trivia", "ExerciseTitle"],

Cell[TextData[{
  "An interesting mathematical trivia fact is that only coefficients of ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[PlusMinus]1\)\)]],
  " appear in factorizations of ",
  Cell[BoxData[
      \(TraditionalForm\`x\^n - 1\)]],
  " until ",
  Cell[BoxData[
      \(TraditionalForm\`n\)]],
  " becomes surprisingly large.  Inspection of a table for ",
  Cell[BoxData[
      \(TraditionalForm\`n \[LessEqual] 10\)]],
  " "
}], "ExerciseText"],

Cell[CellGroupData[{

Cell[BoxData[
    \(Table[Factor[x\^n - 1], {n, 1, 10}] // ColumnForm\)], "Input"],

Cell[BoxData[
    InterpretationBox[GridBox[{
          {\(\(-1\) + x\)},
          {\(\((\(-1\) + x)\)\ \((1 + x)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x + x\^2)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x)\)\ \((1 + x\^2)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x + x\^2 + x\^3 + x\^4)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x)\)\ \((1 - x + x\^2)\)\ \((1 + x + 
                  x\^2)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x + x\^2 + x\^3 + x\^4 + x\^5 + 
                  x\^6)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x)\)\ \((1 + x\^2)\)\ \((1 + 
                  x\^4)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x + x\^2)\)\ \((1 + x\^3 + x\^6)\)\)},
          {\(\((\(-1\) + x)\)\ \((1 + x)\)\ \((1 - x + x\^2 - x\^3 + 
                  x\^4)\)\ \((1 + x + x\^2 + x\^3 + x\^4)\)\)}
          },
        GridBaseline->{Baseline, {1, 1}},
        ColumnAlignments->{Left}],
      ColumnForm[ {
        Plus[ -1, x], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x, 
            Power[ x, 2]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x], 
          Plus[ 1, 
            Power[ x, 2]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x, 
            Power[ x, 2], 
            Power[ x, 3], 
            Power[ x, 4]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x], 
          Plus[ 1, 
            Times[ -1, x], 
            Power[ x, 2]], 
          Plus[ 1, x, 
            Power[ x, 2]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x, 
            Power[ x, 2], 
            Power[ x, 3], 
            Power[ x, 4], 
            Power[ x, 5], 
            Power[ x, 6]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x], 
          Plus[ 1, 
            Power[ x, 2]], 
          Plus[ 1, 
            Power[ x, 4]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x, 
            Power[ x, 2]], 
          Plus[ 1, 
            Power[ x, 3], 
            Power[ x, 6]]], 
        Times[ 
          Plus[ -1, x], 
          Plus[ 1, x], 
          Plus[ 1, 
            Times[ -1, x], 
            Power[ x, 2], 
            Times[ -1, 
              Power[ x, 3]], 
            Power[ x, 4]], 
          Plus[ 1, x, 
            Power[ x, 2], 
            Power[ x, 3], 
            Power[ x, 4]]]}],
      Editable->False]], "Output"]
}, Closed]],

Cell[TextData[{
  "suggests the conjecture that factorizations of ",
  Cell[BoxData[
      \(TraditionalForm\`x\^n - 1\)]],
  " for any ",
  StyleBox["n",
    FontSlant->"Italic"],
  " involve only coefficients of unit magnitude.  Write a short program that \
disproves this hypothesis by determining the first value of ",
  StyleBox["n",
    FontSlant->"Italic"],
  " for which a different coefficient appears."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Law of Cosines", "ExerciseTitle"],

Cell[TextData[{
  "The figure below shows a triangle with sides of length ",
  Cell[BoxData[
      \(TraditionalForm\`{a, b, c}\)]],
  " opposite angles ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Alpha], \[Beta], \[Gamma]}\)]],
  ".  An altitude ",
  Cell[BoxData[
      \(TraditionalForm\`h\)]],
  " subdivides the base into segments ",
  Cell[BoxData[
      \(TraditionalForm\`x\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`b - x\)]],
  ", forming two right triangles.  Apply the Pythagorean theorem to these \
right triangles together with an equation for ",
  Cell[BoxData[
      \(TraditionalForm\`Cos[\[Gamma]]\)]],
  " to form a system of 3 simultaneous equations that can be used to solve \
for ",
  Cell[BoxData[
      \(TraditionalForm\`c\)]],
  " in terms of ",
  Cell[BoxData[
      \(TraditionalForm\`a, b, \[Gamma]\)]],
  ".  Note that two variables should be eliminated using the third (optional) \
argument to ",
  StyleBox["Solve", "InlineInput",
    FontWeight->"Bold"],
  "."
}], "ExerciseText"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .64151 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0507637 0.898473 0.0601977 0.898473 [
[.32031 .32974 -5.34375 -7.71875 ]
[.32031 .32974 5.34375 7.71875 ]
[.36523 .01527 -12.125 -7.71875 ]
[.36523 .01527 12.125 7.71875 ]
[.81447 .01527 -4.9375 -7.71875 ]
[.81447 .01527 4.9375 7.71875 ]
[.85939 .32974 -5 -7.71875 ]
[.85939 .32974 5 7.71875 ]
[.63477 .32974 -5.34375 -7.71875 ]
[.63477 .32974 5.34375 7.71875 ]
[.97619 .03324 -5.53125 -7.71875 ]
[.97619 .03324 5.53125 7.71875 ]
[.67969 .62624 -5.65625 -7.71875 ]
[.67969 .62624 5.65625 7.71875 ]
[.02381 .03324 -5.21875 -7.71875 ]
[.02381 .03324 5.21875 7.71875 ]
[ 0 0 0 0 ]
[ 1 .64151 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.5 Mabswid
[ ] 0 setdash
.05076 .0602 m
.94924 .0602 L
.67969 .59928 L
.05076 .0602 L
s
[ .02 .01 ] 0 setdash
.67969 .0602 m
.67969 .59928 L
s
gsave
.32031 .32974 -66.3438 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(a) show
69.688 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.36523 .01527 -73.125 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(b) show
69.688 15.813 moveto
%%IncludeResource: font Math1
%%IncludeFont: Math1
/Math1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(-) show
77.375 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(x) show
83.250 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.81447 .01527 -65.9375 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(x) show
68.875 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.85939 .32974 -66 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(c) show
69.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.63477 .32974 -66.3438 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(h) show
69.688 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.97619 .03324 -66.5313 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Math1
%%IncludeFont: Math1
/Math1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(a) show
70.063 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.67969 .62624 -66.6563 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
64.188 15.813 moveto
%%IncludeResource: font Math1
%%IncludeFont: Math1
/Math1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(b) show
70.313 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
gsave
.02381 .03324 -66.2188 -11.7188 Mabsadd m
1 1 Mabs scale
currentpoint translate
/MISOfy
{
    /newfontname exch def
    /oldfontname exch def

    oldfontname findfont
    dup length dict begin
        {1 index /FID ne {def} {pop pop} ifelse} forall
        /Encoding ISOLatin1Encoding def
        currentdict
    end

    newfontname exch definefont pop
} def

0 23.4375 translate 1 -1 scale
63.000 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
63.000 15.813 moveto
%%IncludeResource: font Math1
%%IncludeFont: Math1
/Math1 findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
(g) show
69.438 15.813 moveto
%%IncludeResource: font Helvetica
%%IncludeFont: Helvetica
/Helvetica findfont 12.000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000 0.000 1.000 setrgbcolor
0.000 0.000 rmoveto
1.000 setlinewidth
grestore
0 0 m
1 0 L
1 .64151 L
0 .64151 L
closepath
clip
newpath
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{282.25, 181},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006W00013b000`400?l00000o`00003oo`3ooonX
0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX
0?ooo`00SP3oool0102LB?l004So0=ooo`1d0?l20000o`0309cOo`3oool0oooo00`0oooo00<0_gCo
0000o`2Lgol00P3oool00`3ogol0B03o001do`2S0?ooo`030;mdo`000?l0W=oo0080oooo00<0omoo
04P0o`00M?l0D@3oool008h0oooo00<0W4So0018o`1dB?l010000?l00`1d_ol0oooo0?ooo`0<0?oo
o`0707@0o`00M?l0oooo0?ooo`2oM?l0003o0;ooo`2T0?ooo`0707@0o`00M?l0oooo0?ooo`2oM?l0
003o0;ooo`1A0?ooo`00SP3oool00`2LB?l0003o001do`020?ooo`040?noo`000?l004So0=ooo`d0
oooo00H0gico0000o`2Lgol0omoo04P0o`18W?nU0?ooo`060=nLo`000?l0W=oo0?oOo`180?l0B9co
DP3oool008h0oooo00<0W4So0000o`2oool00`3oool00`3OW?l0003o09cOo`0>0?ooo`0509a8o`00
B?l0W7Co0018o`3Oool0YP3oool01@2LB?l004So09ado`00B?l0gooo0580oooo000?0?ooo`0307@0
o`18W?l0oooo07`0oooo00<0W4So0018o`3Oool00`3oool01@3o_ol0003o07Boo`3oool0W4So00P0
003o00D0B9co0?ooo`3oool0oooo0?noo`020000o`0307Boo`3oool0oooo0:D0oooo00@0okoo0000
o`000?l0M;ooD`3oool000l0oooo00<0_gCo0000o`2oool0O03oool00`2LB?l004So0=ooo`030?oo
o`030?noo`000?l0M;oo00l0oooo00<0M03o0000o`2oool0Z03oool00`1d0?l0003o0;ooo`1C0?oo
o`003`3oool00`3ogol0B03o07Boo`1l0?ooo`0309a8o`000?l0_ooo00<0oooo00<0gico0000o`2L
gol03P3oool0102oM?l0003o0000o`18W?nW0?ooo`040;mdo`000?l0003o04RLoe<0oooo000@0?oo
o`030?noo`18W?l0oooo07/0oooo00T0W4So0000o`00B?l0gooo0?ooo`3ogol0B03o0000o`2oool0
3@3oool01P3ogol0B03o04RLo`3OW?l0003o09cOojD0oooo00H0omoo04P0o`18W?l0gico0000o`2L
gomB0?ooo`004@3oool00`2LB?l0_ooo0?ooo`1j0?ooo`0309a8o`00B?l0M4So00@0003o00<0M;oo
0?ooo`3oool0303oool01`2LB?l0003o0;ooo`3oool0W4So0018o`3Oool0Y03oool01`2LB?l0003o
0;ooo`3oool0W4So0018o`3Oool0>@3oool0203OW?l0003o0000o`000?l0_ooo07@0o`000?l007Co
403oool00140oooo00<0W4So001do`3oool0NP3oool0102LB?l004So0=ooo`1d0?l20000o`0307Bo
o`3oool0oooo00`0oooo00<0okoo0000o`1d_ol00P3oool00`3o_ol0003o07Boo`2S0?ooo`030?no
o`000?l0M;oo0080oooo00<0okoo0000o`1d_ol0>03oool00`3o_ol0003o07Boo`020?ooo`0507@0
o`000?l0W=oo0?noo`1d_ol03`3oool00140oooo00<0M03o07Ado`1d_ol0NP3oool00`2LB?l004So
0=ooo`3m0?ooo`030=nLo`000?l0_ooo00<0oooo00<0M03o07Boo`3oool0403oool00140oooo00@0
M03o09cOo`1d0?l0_oooN@3oool00`2LB?l004So0=ooo`3m0?ooo`030=nLo`000?l0_ooo00<0oooo
00<0_gCo0000o`2oool0403oool000d0oooo00P0M7Co0?ooo`3oool0omoo04P0o`2Lgol0gico04RL
ogT0oooo00<0W4So0018o`3Oool0o@3oool00`3ogol0B03o07Boo`030?ooo`040=nLo`00B?l0B4So
0=ooo`l0oooo000=0?ooo`0909a8o`2Lgol0oooo0;mdo`000?l0_ooo0?ooo`1d0?l0W=oo0?l0oooo
NP3oool00`2oM?l004So0=ooo`020?ooo`040=nLo`00B?l0gkoo001do`l0oooo000=0?ooo`050?oO
o`180?l0003o0000o`00M?l00P3oool00`2oM?l007Co0?ooo`3o0?ooogX0oooo00T0W4So001do`3o
ool0oooo0;mdo`00B?l0gooo0;mdo`00M?l03P3oool000h0oooo00@0okoo0000o`000?l0_ooo0P3o
ool00`3ogol0B03o09cOo`3o0?ooog/0oooo00D0gico0000o`000?l0003o09cOo`0A0?ooo`00o`3o
oonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`00o`3o
oonX0?ooo`00o`3ooonX0?ooo`00o`3ooonX0?ooo`007@3ooooo000006h000007@3oool001h0oooo
00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`1Z0?ooo`030000003oool0oooo01`0oooo
000O0?ooo`030000003oool0oooo0?T0oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?oo
o`0L0?ooo`00803oool200000?T0oooo00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0M
0?ooo`008P3oool00`000000oooo0?ooo`3f0?ooo`030000003oool0oooo06T0oooo00<000000?oo
o`3oool07@3oool002<0oooo00<000000?ooo`3oool0m@3oool00`000000oooo0?ooo`1X0?ooo`03
0000003oool0oooo01h0oooo000T0?ooo`030000003oool0oooo0?@0oooo00<000000?ooo`3oool0
J03oool00`000000oooo0?ooo`0N0?ooo`009@3oool00`000000oooo0?ooo`3c0?ooo`030000003o
ool0oooo06L0oooo00<000000?ooo`3oool07`3oool002H0oooo00<000000?ooo`3oool0lP3oool0
0`000000oooo0?ooo`1W0?ooo`030000003oool0oooo01l0oooo000W0?ooo`800000o`3ooomL0?oo
o`030000003oool0oooo0200oooo000Y0?ooo`030000003oool0oooo0?l0ooooF@3oool00`000000
oooo0?ooo`0P0?ooo`00:P3oool00`000000oooo0?ooo`3o0?oooeL0oooo00<000000?ooo`3oool0
8@3oool002/0oooo00<000000?ooo`3oool0k@3oool00`000000oooo0?ooo`1U0?ooo`030000003o
ool0oooo0240oooo000/0?ooo`030000003oool0oooo0>`0oooo00<000000?ooo`3oool0I03oool0
0`000000oooo0?ooo`0R0?ooo`00;@3oool00`000000oooo0?ooo`3[0?ooo`030000003oool0oooo
06@0oooo00<000000?ooo`3oool08P3oool002h0oooo0P00003[0?ooo`030000003oool0oooo06<0
oooo00<000000?ooo`3oool08`3oool00300oooo00<000000?ooo`3oool0j03oool00`000000oooo
0?ooo`1S0?ooo`030000003oool0oooo02<0oooo000a0?ooo`030000003oool0oooo0>L0oooo00<0
00000?ooo`3oool0HP3oool00`000000oooo0?ooo`0T0?ooo`00<P3oool00`000000oooo0?ooo`3V
0?ooo`030000003oool0oooo0680oooo00<000000?ooo`3oool0903oool003<0oooo00<000000?oo
o`3oool0i@3oool00`000000oooo0?ooo`1Q0?ooo`030000003oool0oooo02D0oooo000d0?ooo`03
0000003oool0oooo0>@0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`0U0?ooo`00
=@3oool200000?l0ooooB03oool00`000000oooo0?ooo`0V0?ooo`00=`3oool00`000000oooo0?oo
o`3o0?ooodD0oooo00<000000?ooo`3oool09P3oool003P0oooo00<000000?ooo`3oool0o`3ooom3
0?ooo`030000003oool0oooo02L0oooo000i0?ooo`030000003oool0oooo0=l0oooo00<000000?oo
o`3oool0G`3oool00`000000oooo0?ooo`0W0?ooo`00>P3oool00`000000oooo0?ooo`3N0?ooo`03
0000003oool0oooo05h0oooo00<000000?ooo`3oool0:03oool003/0oooo00<000000?ooo`3oool0
g@3oool00`000000oooo0?ooo`1N0?ooo`030000003oool0oooo02P0oooo000l0?ooo`800000g@3o
ool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo02T0oooo000n0?ooo`030000003oool0
oooo0=X0oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`0Y0?ooo`00?`3oool00`00
0000oooo0?ooo`3I0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3oool0:P3oool00400
oooo00<000000?ooo`3oool0f03oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo02X0
oooo00110?ooo`030000003oool0oooo0=L0oooo00<000000?ooo`3oool0F`3oool00`000000oooo
0?ooo`0[0?ooo`00@P3oool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo05/0oooo00<0
00000?ooo`3oool0:`3oool004<0oooo0P00003o0?oooc@0oooo00<000000?ooo`3oool0;03oool0
04D0oooo00<000000?ooo`3oool0o`3ooola0?ooo`030000003oool0oooo02`0oooo00160?ooo`03
0000003oool0oooo0?l0oooo;`3oool00`000000oooo0?ooo`0]0?ooo`00A`3oool00`000000oooo
0?ooo`3A0?ooo`030000003oool0oooo05T0oooo00<000000?ooo`3oool0;@3oool004P0oooo00<0
00000?ooo`3oool0d03oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo02h0oooo0019
0?ooo`030000003oool0oooo0<l0oooo00<000000?ooo`3oool0F03oool00`000000oooo0?ooo`0^
0?ooo`00BP3oool200000<l0oooo00<000000?ooo`3oool0E`3oool00`000000oooo0?ooo`0_0?oo
o`00C03oool00`000000oooo0?ooo`3<0?ooo`030000003oool0oooo05L0oooo00<000000?ooo`3o
ool0;`3oool004d0oooo00<000000?ooo`3oool0b`3oool00`000000oooo0?ooo`1F0?ooo`030000
003oool0oooo0300oooo001>0?ooo`030000003oool0oooo0<X0oooo00<000000?ooo`3oool0EP3o
ool00`000000oooo0?ooo`0`0?ooo`00C`3oool00`000000oooo0?ooo`390?ooo`030000003oool0
oooo05D0oooo00<000000?ooo`3oool0<@3oool00500oooo00<000000?ooo`3oool0b03oool00`00
0000oooo0?ooo`1E0?ooo`030000003oool0oooo0340oooo001A0?ooo`800000o`3ooolP0?ooo`03
0000003oool0oooo0380oooo001C0?ooo`030000003oool0oooo0?l0oooo7@3oool00`000000oooo
0?ooo`0b0?ooo`00E03oool00`000000oooo0?ooo`3o0?oooa/0oooo00<000000?ooo`3oool0<`3o
ool005D0oooo00<000000?ooo`3oool0``3oool00`000000oooo0?ooo`1C0?ooo`030000003oool0
oooo03<0oooo001F0?ooo`030000003oool0oooo0<80oooo00<000000?ooo`3oool0DP3oool00`00
0000oooo0?ooo`0d0?ooo`00E`3oool00`000000oooo0?ooo`310?ooo`030000003oool0oooo0580
oooo00<000000?ooo`3oool0=03oool005P0oooo0P0000310?ooo`030000003oool0oooo0540oooo
00<000000?ooo`3oool0=@3oool005X0oooo00<000000?ooo`3oool0_P3oool00`000000oooo0?oo
o`1A0?ooo`030000003oool0oooo03D0oooo001K0?ooo`030000003oool0oooo0;d0oooo00<00000
0?ooo`3oool0D03oool00`000000oooo0?ooo`0f0?ooo`00G03oool00`000000oooo0?ooo`2l0?oo
o`030000003oool0oooo0500oooo00<000000?ooo`3oool0=P3oool005d0oooo00<000000?ooo`3o
ool0^`3oool00`000000oooo0?ooo`1?0?ooo`030000003oool0oooo03L0oooo001N0?ooo`030000
003oool0oooo0;X0oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`0g0?ooo`00G`3o
ool00`000000oooo0?ooo`3o0?ooo`/0oooo00<000000?ooo`3oool0>03oool00600oooo0P00003o
0?ooo`/0oooo00<000000?ooo`3oool0>03oool00680oooo00<000000?ooo`3oool0o`3oool70?oo
o`030000003oool0oooo03T0oooo001S0?ooo`030000003oool0oooo0?l0oooo1P3oool00`000000
oooo0?ooo`0i0?ooo`00I03oool00`000000oooo0?ooo`2d0?ooo`030000003oool0oooo04`0oooo
00<000000?ooo`3oool0>P3oool006D0oooo00<000000?ooo`3oool0/`3oool00`000000oooo0?oo
o`1<0?ooo`030000003oool0oooo03X0oooo001V0?ooo`030000003oool0oooo0;80oooo00<00000
0?ooo`3oool0B`3oool00`000000oooo0?ooo`0k0?ooo`00I`3oool200000;80oooo00<000000?oo
o`3oool0B`3oool00`000000oooo0?ooo`0k0?ooo`00J@3oool00`000000oooo0?ooo`2_0?ooo`03
0000003oool0oooo04X0oooo00<000000?ooo`3oool0?03oool006X0oooo00<000000?ooo`3oool0
[P3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0oooo03`0oooo001[0?ooo`030000003o
ool0oooo0:d0oooo00<000000?ooo`3oool0B@3oool00`000000oooo0?ooo`0m0?ooo`00K03oool0
0`000000oooo0?ooo`2/0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0?@3oool0
06d0oooo00<000000?ooo`3oool0Z`3oool00`000000oooo0?ooo`180?ooo`030000003oool0oooo
03h0oooo001^0?ooo`800000mP3oool00`000000oooo0?ooo`0n0?ooo`00L03oool00`000000oooo
0?ooo`3b0?ooo`030000003oool0oooo03l0oooo001a0?ooo`030000003oool0oooo0?40oooo00<0
00000?ooo`3oool0?`3oool00780oooo00<000000?ooo`3oool0YP3oool00`000000oooo0?ooo`16
0?ooo`030000003oool0oooo0400oooo001c0?ooo`030000003oool0oooo0:D0oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`100?ooo`00M03oool00`000000oooo0?ooo`2T0?ooo`03
0000003oool0oooo04D0oooo00<000000?ooo`3oool0@@3oool007D0oooo0P00002T0?ooo`030000
003oool0oooo04D0oooo00<000000?ooo`3oool0@@3oool007L0oooo00<000000?ooo`3oool0X@3o
ool00`000000oooo0?ooo`140?ooo`030000003oool0oooo0480oooo001h0?ooo`030000003oool0
oooo0:00oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`120?ooo`00N@3oool00`00
0000oooo0?ooo`2O0?ooo`030000003oool0oooo04<0oooo00<000000?ooo`3oool0@`3oool007X0
oooo00<000000?ooo`3oool0WP3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo04<0
oooo001k0?ooo`030000003oool0oooo09d0oooo00<000000?ooo`3oool0@P3oool00`000000oooo
0?ooo`140?ooo`00O03oool200000>80oooo00<000000?ooo`3oool0A03oool007h0oooo00<00000
0?ooo`3oool0gP3oool00`000000oooo0?ooo`150?ooo`00O`3oool00`000000oooo0?ooo`3M0?oo
o`030000003oool0oooo04D0oooo00200?ooo`030000003oool0oooo09P0oooo00<000000?ooo`3o
ool0@03oool00`000000oooo0?ooo`160?ooo`00P@3oool00`000000oooo0?ooo`2G0?ooo`030000
003oool0oooo0400oooo00<000000?ooo`3oool0AP3oool00880oooo00<000000?ooo`3oool0UP3o
ool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo04L0oooo00230?ooo`800000UP3oool0
0`000000oooo0?ooo`0o0?ooo`030000003oool0oooo04L0oooo00250?ooo`030000003oool0oooo
09<0oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`180?ooo`00QP3oool00`000000
oooo0?ooo`2B0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool0B03oool008L0oooo
00<000000?ooo`3oool0T@3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo04T0oooo
00280?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0?@3oool00`000000oooo0?oo
o`190?ooo`00R@3oool00`000000oooo0?ooo`2?0?ooo`030000003oool0oooo03`0oooo00<00000
0?ooo`3oool0BP3oool008X0oooo0P00003>0?ooo`030000003oool0oooo04X0oooo002<0?ooo`03
0000003oool0oooo0<X0oooo00<000000?ooo`3oool0B`3oool008d0oooo00<000000?ooo`3oool0
b@3oool00`000000oooo0?ooo`1;0?ooo`00SP3oool00`000000oooo0?ooo`2:0?ooo`030000003o
ool0oooo03X0oooo00<000000?ooo`3oool0C03oool008l0oooo00<000000?ooo`3oool0R@3oool0
0`000000oooo0?ooo`0j0?ooo`030000003oool0oooo04`0oooo002@0?ooo`030000003oool0oooo
08P0oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`1=0?ooo`00T@3oool2000008P0
oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`1=0?ooo`00T`3oool00`000000oooo
0?ooo`250?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0CP3oool009@0oooo00<0
00000?ooo`3oool0Q03oool00`000000oooo0?ooo`0h0?ooo`030000003oool0oooo04h0oooo002E
0?ooo`030000003oool0oooo08<0oooo00<000000?ooo`3oool0=`3oool00`000000oooo0?ooo`1?
0?ooo`00Q`3oool0201d0?l0003o0000o`000?l0W=oo0;mdo`000?l0_ooo1`3oool00`000000oooo
0?ooo`1/0?ooo`0309a8o`00B?l0gooo00<0oooo00<0W4So001do`3oool03@3oool00`000000oooo
0?ooo`0g0?ooo`030000003oool0oooo00X0oooo00D0okoo0000o`000?l0003o07Boo`100?ooo`00
QP3oool00`2LB?l0003o0000o`040000o`030018o`3Oool0oooo00L0oooo00<000000?ooo`3oool0
J`3oool00`2LB?l004So0=ooo`030?ooo`0309a8o`00M?l0oooo00d0oooo00<000000?ooo`3oool0
=P3oool00`000000oooo0?ooo`0:0?ooo`030?noo`000?l0003o00<0003o00<0M;oo0?ooo`3oool0
?@3oool008D0oooo00<0omoo04P0o`00M?l00`3oool0101d0?l0003o0018o`3Oool90?ooo`800000
J`3oool00`2LB?l004So0=ooo`030?ooo`0309a8o`00M?l0oooo04H0oooo00<000000?ooo`3oool0
2P3oool00`2LB?l0003o09cOo`020?ooo`0307@0o`00B?l0gooo03h0oooo00250?ooo`030?oOo`18
0?l0M;oo00<0oooo00<0omoo04P0o`00M?l0303oool00`000000oooo0?ooo`1X0?ooo`0309a8o`00
B?l0gooo00<0oooo00<0W4So001do`3oool0A@3oool00`000000oooo0?ooo`0:0?ooo`030?oOo`18
0?l0B9co04@0oooo00260?ooo`0407@0o`000?l0003o07Boo`80oooo00<0M03o001do`3oool0303o
ool00`000000oooo0?ooo`1W0?ooo`0309a8o`00B?l0gooo00<0oooo00<0W4So001do`3oool0A@3o
ool00`000000oooo0?ooo`0:0?ooo`030?oOo`180?l0M;oo04@0oooo00260?ooo`030?oOo`180?l0
003o00@0003o00<007Co0?ooo`3oool0303oool00`000000oooo0?ooo`1V0?ooo`0309a8o`00B?l0
gooo00<0oooo00<0W4So001do`3oool03@3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0
oooo00/0oooo00<0omoo04P0o`1d_ol0A03oool008X0oooo00@0M03o0000o`000?l007Co3`3oool0
0`000000oooo0?ooo`1U0?ooo`0309a8o`000?l0_ooo00<0oooo00<0M03o001do`3oool03@3oool0
0`000000oooo0?ooo`0c0?ooo`030000003oool0oooo00`0oooo00<0omoo04P0o`18W?l0A03oool0
08H0oooo00<0W4So0000o`2oool00P3oool00`3ogol0B03o001do`0@0?ooo`030000003oool0oooo
06@0oooo00P0W4So0000o`00B?l0gooo0?ooo`3o_ol0003o001do`h0oooo00<000000?ooo`3oool0
<`3oool00`000000oooo0?ooo`0=0?ooo`0709a8o`000?l0W=oo0?ooo`3o_ol0003o04RLo`0o0?oo
o`00QP3oool00`3OW?l0003o0000o`040000o`0307Boo`3oool0oooo00l0oooo0P00001T0?ooo`03
09a8o`00B?l0M4So00@0003o00<0M;oo0?ooo`3oool0303oool00`000000oooo0?ooo`0b0?ooo`03
0000003oool0oooo00h0oooo00<0gico0000o`000?l00`000?l00`2Lgol0oooo0?ooo`0m0?ooo`00
Q`3oool00`3o_ol0003o0000o`020000o`0304RLo`3oool0oooo0180oooo00<000000?ooo`3oool0
H@3oool0102LB?l004So0=ooo`1d0?l20000o`0304RLo`3oool0oooo00d0oooo00<000000?ooo`3o
ool0<P3oool00`000000oooo0?ooo`0?0?ooo`050?noo`000?l0003o0000o`2Lgol0@03oool00:80
oooo00<000000?ooo`3oool0H03oool00`2LB?l004So0=ooo`0C0?ooo`030000003oool0oooo0340
oooo00<000000?ooo`3oool0E@3oool00:<0oooo00<000000?ooo`3oool0G`3oool00`2LB?l004So
0=ooo`0C0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0E@3oool00:@0oooo00<0
00000?ooo`3oool0GP3oool00`2LB?l004So0=ooo`0C0?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool0EP3oool00:D0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`0`
0?ooo`030000003oool0oooo05H0oooo002V0?ooo`800000Y@3oool00`000000oooo0?ooo`1G0?oo
o`00Z03oool00`000000oooo0?ooo`2R0?ooo`030000003oool0oooo05L0oooo002Y0?ooo`030000
003oool0oooo0:00oooo00<000000?ooo`3oool0F03oool00:X0oooo00<000000?ooo`3oool0KP3o
ool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo05P0oooo002[0?ooo`030000003oool0
oooo06d0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`1I0?ooo`00[03oool00`00
0000oooo0?ooo`1/0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0F@3oool00:d0
oooo0P00001/0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3oool0FP3oool00:l0oooo
00<000000?ooo`3oool0J@3oool00`000000oooo0?ooo`0/0?ooo`030000003oool0oooo05X0oooo
002`0?ooo`030000003oool0oooo06P0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?oo
o`1K0?ooo`00/@3oool00`000000oooo0?ooo`1W0?ooo`030000003oool0oooo02/0oooo00<00000
0?ooo`3oool0F`3oool00;80oooo00<000000?ooo`3oool0IP3oool00`000000oooo0?ooo`0Z0?oo
o`030000003oool0oooo05`0oooo002c0?ooo`030000003oool0oooo06D0oooo00<000000?ooo`3o
ool0:P3oool00`000000oooo0?ooo`1L0?ooo`00]03oool200000940oooo00<000000?ooo`3oool0
G@3oool00;H0oooo00<000000?ooo`3oool0SP3oool00`000000oooo0?ooo`1M0?ooo`00]`3oool0
0`000000oooo0?ooo`2<0?ooo`030000003oool0oooo05h0oooo002h0?ooo`030000003oool0oooo
0600oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`1N0?ooo`00^@3oool00`000000
oooo0?ooo`1O0?ooo`030000003oool0oooo02L0oooo00<000000?ooo`3oool0G`3oool00;X0oooo
00<000000?ooo`3oool0GP3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo05l0oooo
002k0?ooo`800000GP3oool00`000000oooo0?ooo`0V0?ooo`030000003oool0oooo0600oooo002m
0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`1P
0?ooo`00_P3oool00`000000oooo0?ooo`1J0?ooo`030000003oool0oooo02D0oooo00<000000?oo
o`3oool0H@3oool00;l0oooo00<000000?ooo`3oool0F@3oool00`000000oooo0?ooo`0U0?ooo`03
0000003oool0oooo0640oooo00300?ooo`030000003oool0oooo05P0oooo00<000000?ooo`3oool0
903oool00`000000oooo0?ooo`1R0?ooo`00`@3oool00`000000oooo0?ooo`1G0?ooo`030000003o
ool0oooo02@0oooo00<000000?ooo`3oool0HP3oool00<80oooo0P00001G0?ooo`030000003oool0
oooo02<0oooo00<000000?ooo`3oool0H`3oool00<@0oooo00<000000?ooo`3oool0NP3oool00`00
0000oooo0?ooo`1S0?ooo`00a@3oool00`000000oooo0?ooo`1h0?ooo`030000003oool0oooo06@0
oooo00360?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3oool0I03oool00<L0oooo00<0
00000?ooo`3oool0D@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo06D0oooo0038
0?ooo`030000003oool0oooo0500oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`1U
0?ooo`00b@3oool200000500oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`1V0?oo
o`00b`3oool00`000000oooo0?ooo`1=0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3o
ool0IP3oool00<`0oooo00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0O0?ooo`030000
003oool0oooo06L0oooo003=0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool07`3o
ool00`000000oooo0?ooo`1W0?ooo`00cP3oool00`000000oooo0?ooo`1:0?ooo`030000003oool0
oooo01h0oooo00<000000?ooo`3oool0J03oool00<l0oooo00<000000?ooo`3oool0B@3oool00`00
0000oooo0?ooo`0N0?ooo`030000003oool0oooo06P0oooo003@0?ooo`800000B@3oool00`000000
oooo0?ooo`0M0?ooo`030000003oool0oooo06T0oooo003B0?ooo`030000003oool0oooo06H0oooo
00<000000?ooo`3oool0J@3oool00=<0oooo00<000000?ooo`3oool0I03oool00`000000oooo0?oo
o`1Z0?ooo`00e03oool00`000000oooo0?ooo`1S0?ooo`030000003oool0oooo06X0oooo003E0?oo
o`030000003oool0oooo04<0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`1[0?oo
o`00eP3oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3o
ool0J`3oool00=L0oooo0P0000120?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0
K03oool00=T0oooo00<000000?ooo`3oool0?`3oool00`000000oooo0?ooo`0J0?ooo`030000003o
ool0oooo06`0oooo003J0?ooo`030000003oool0oooo03h0oooo00<000000?ooo`3oool06@3oool0
0`000000oooo0?ooo`1]0?ooo`00f`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0oooo
01T0oooo00<000000?ooo`3oool0K@3oool00=`0oooo00<000000?ooo`3oool0?03oool00`000000
oooo0?ooo`0H0?ooo`030000003oool0oooo06h0oooo003M0?ooo`030000003oool0oooo03/0oooo
00<000000?ooo`3oool0603oool00`000000oooo0?ooo`1^0?ooo`00gP3oool00`000000oooo0?oo
o`0j0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0K`3oool00=l0oooo0P00001D
0?ooo`030000003oool0oooo06l0oooo003Q0?ooo`030000003oool0oooo0500oooo00<000000?oo
o`3oool0L03oool00>80oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`1`0?ooo`00
h`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
L@3oool00>@0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`0E0?ooo`030000003o
ool0oooo0740oooo003U0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0503oool0
0`000000oooo0?ooo`1b0?ooo`00iP3oool2000003<0oooo00<000000?ooo`3oool0503oool00`00
0000oooo0?ooo`1b0?ooo`00j03oool00`000000oooo0?ooo`0`0?ooo`030000003oool0oooo01<0
oooo00<000000?ooo`3oool0L`3oool00>T0oooo00<000000?ooo`3oool0;`3oool00`000000oooo
0?ooo`0C0?ooo`030000003oool0oooo07<0oooo003Z0?ooo`030000003oool0oooo02h0oooo00<0
00000?ooo`3oool04P3oool00`000000oooo0?ooo`1d0?ooo`00j`3oool00`000000oooo0?ooo`0]
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0M03oool00>`0oooo00<000000?oo
o`3oool0;03oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo07D0oooo003]0?ooo`80
0000@03oool00`000000oooo0?ooo`1e0?ooo`00k`3oool00`000000oooo0?ooo`0l0?ooo`030000
003oool0oooo07H0oooo003`0?ooo`030000003oool0oooo03/0oooo00<000000?ooo`3oool0MP3o
ool00?40oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0
oooo07L0oooo003b0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`1g0?ooo`00l`3oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo00h0
oooo00<000000?ooo`3oool0N03oool00?@0oooo0P00000U0?ooo`030000003oool0oooo00h0oooo
00<000000?ooo`3oool0N03oool00?H0oooo00<000000?ooo`3oool08P3oool00`000000oooo0?oo
o`0=0?ooo`030000003oool0oooo07T0oooo003g0?ooo`030000003oool0oooo0240oooo00<00000
0?ooo`3oool03@3oool00`000000oooo0?ooo`1i0?ooo`00n03oool00`000000oooo0?ooo`0P0?oo
o`030000003oool0oooo00`0oooo00<000000?ooo`3oool0NP3oool00?T0oooo00<000000?ooo`3o
ool07`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo07X0oooo003j0?ooo`030000
003oool0oooo01h0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`1k0?ooo`00n`3o
ool2000002`0oooo00<000000?ooo`3oool0N`3oool00?d0oooo00<000000?ooo`3oool0:03oool0
0`000000oooo0?ooo`1l0?ooo`00oP3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo
07`0oooo003o0?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`1m0?ooo`00o`3oool10?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`1m0?ooo`00o`3oool20?ooo`030000003oool0oooo01L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`1n0?ooo`00o`3oool30?ooo`8000005`3oool0
0`000000oooo0?ooo`080?ooo`030000003oool0oooo07h0oooo003o0?ooo`D0oooo00<000000?oo
o`3oool0503oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo07l0oooo003o0?ooo`H0
oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo07l0
oooo003o0?ooo`L0oooo00<000000?ooo`3oool04P3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo0800oooo003o0?ooo`P0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo0800oooo003o0?ooo`T0oooo00<000000?ooo`3oool0403oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo0840oooo003o0?ooo`X0oooo0P00000H0?oo
o`030000003oool0oooo0840oooo003o0?ooo``0oooo00<000000?ooo`3oool0503oool00`000000
oooo0?ooo`220?ooo`00o`3oool=0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0
PP3oool00?l0oooo3P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0P`3oool00?l0oooo3`3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0P`3oool00?l0oooo403oool00`000000oooo0?ooo`090?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool0Q03oool00?l0oooo4@3oool2000000T0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`240?ooo`00o`3ooolC0?ooo`030000
003oool0oooo00H0oooo00D000000?ooo`3oool0oooo000000270?ooo`00o`3ooolD0?ooo`030000
003oool0oooo00D0oooo00D000000?ooo`3oool0oooo000000270?ooo`00o`3ooolE0?ooo`030000
003oool0oooo00@0oooo00@000000?ooo`3oool00000R03oool00?l0oooo5P3oool00`000000oooo
0?ooo`030?ooo`040000003oool0oooo000008P0oooo003o0?oooaL0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo000000290?ooo`00o`3ooolH0?ooo`800000103oool00`000000oooo0?oo
o`270?ooo`00o`3ooolJ0?ooo`040000003oool0oooo000008X0oooo003o0?oooa/0oooo00<00000
0?ooo`000000RP3oool00?l0oooo703oool00`000000oooo0?ooo`290?ooo`00o`3ooonX0?ooo`00
o`3ooolG0?ooo`030;mdo`000?l0W=oo08h0oooo003o0?oooaP0oooo00<0M03o04RLo`3oool0S@3o
ool00?l0oooo603oool00`2oM?l007Co0?ooo`2=0?ooo`00o`3ooolH0?ooo`030=nLo`00B?l0gooo
08d0oooo003o0?oooaP0oooo00@0omoo04P0o`2Lgol0W4So0P000?l00`00B?l0gooo0?ooo`270?oo
o`00o`3ooolI0?ooo`0307@0o`000?l0M;oo0080oooo00<0_gCo0000o`2Lgol0Q`3oool00?l0oooo
6@3oool00`2oM?l004So0=ooo`030?ooo`0309a8o`00B?l0gooo08H0oooo003o0?oooaT0oooo00<0
gico0018o`3Oool00`3oool00`3OW?l0003o09cOo`260?ooo`00o`3ooolI0?ooo`030?oOo`180?l0
W=oo00<0oooo00<0okoo0000o`1d_ol0QP3oool00?l0oooo6P3oool00`1d0?l0M;oo0?ooo`020?oo
o`030?noo`000?l0W=oo08H0oooo003o0?oooaX0oooo00<0_gCo001do`3oool00P3oool00`2oM?l0
07Co0?ooo`260?ooo`00o`3ooolJ0?ooo`070?noo`00B?l0gooo07@0o`000?l004So0=ooo`270?oo
o`00o`3ooolJ0?ooo`030?oOo`180?l0W=oo0080oooo00<0omoo04P0o`2oool0QP3oool00?l0oooo
6`3oool00`2LB?l0M;oo0?ooo`020?ooo`030;mdo`00M?l0oooo08D0oooo003o0?oooa/0oooo00<0
gico0018o`3Oool00P3oool00`3OW?l0003o0;ooo`250?ooo`00o`3ooolL0?ooo`0709a8o`00M?l0
oooo0?ooo`2oM?l0003o0;ooo`250?ooo`00o`3ooolM0?ooo`050;mdo`000?l0003o0000o`2Lgol0
QP3oool00?l0ooooZ03oool00?l0ooooZ03oool00?l0ooooZ03oool00?l0ooooZ03oool00001\
\>"],
  ImageRangeCache->{{{0, 422.375}, {270.5, 0}} -> {-0.0806128, -0.0818394, \
0.00412391, 0.00412391}}]
}, Closed]],

Cell[CellGroupData[{

Cell["Chadwick measurement of neutron mass", "ExerciseTitle"],

Cell[TextData[{
  "Chadwick performed the first measurement of the neutron mass, ",
  Cell[BoxData[
      \(TraditionalForm\`m\_n\)]],
  ", in 1932 by comparing the recoil velocities, ",
  Cell[BoxData[
      \(TraditionalForm\`v\_H\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`v\_N\)]],
  ", imparted to either a hydrogen nucleus or to a nitrogen nucleus in a \
head-on elastic collision.  Let ",
  Cell[BoxData[
      \(TraditionalForm\`m\_H = \ 1\ u\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`m\_N = 14\ u\)]],
  ", where ",
  Cell[BoxData[
      \(TraditionalForm\`u\)]],
  " represents an atomic mass unit.  He employed a nuclear reaction to \
prepare a beam of neutrons with velocity ",
  Cell[BoxData[
      \(TraditionalForm\`v\_n\)]],
  ", but that velocity was unknown.    Let ",
  Cell[BoxData[
      \(TraditionalForm\`v\_1\)]],
  "and ",
  Cell[BoxData[
      \(TraditionalForm\`v\_2\)]],
  " represent the neutron recoil velocities after colliding with hydrogen or \
with nitrogen, respectively, which were also unknown.  Thus, there are four \
unknown quantities {",
  Cell[BoxData[
      \(TraditionalForm\`m\_n\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`v\_n\)]],
  ",  ",
  Cell[BoxData[
      \(TraditionalForm\`v\_1\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`v\_2\)]],
  "}.  Write a system of four equations which express the conservation of \
energy and of momentum for the two head-on elastic scattering reactions \
(hydrogen or nitrogen targets) and solve for ",
  Cell[BoxData[
      \(TraditionalForm\`m\_n\)]],
  ".  Evaluate ",
  Cell[BoxData[
      \(TraditionalForm\`m\_n\)]],
  " (in atomic mass units) given Chadwick's data: {",
  Cell[BoxData[
      \(TraditionalForm\`v\_H \[Rule] 3.3\ \[Times]\ 10\^9\)]],
  "cm/s, ",
  Cell[BoxData[
      \(TraditionalForm\`v\_N \[Rule] 0.47\ \[Times]\ 10\^9\)]],
  "cm/s} and compare with the modern value of 1.00867 u."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Cylinder in a trough", "ExerciseTitle"],

Cell[TextData[{
  "Suppose that a cylinder is dropped into a trough and that we need to know \
how high above the bottom of the trough it will become wedged.  Let the \
radius of the cylinder be ",
  StyleBox["R",
    FontSlant->"Italic"],
  " and assume that the trough has a parabolic shape ",
  Cell[BoxData[
      \(TraditionalForm\`y = a\ x\^2\)]],
  ".  Determine the height of the center of the cylinder above the bottom of \
the trough and draw the final configuration."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Parabola for any directrix and focus", "ExerciseTitle"],

Cell["\<\
Construct the equation for a parabola given any linear equation for the \
directrix and any set of coordinates for the focus.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Quadratic form in two dimensions ", "ExerciseTitle"],

Cell[TextData[{
  "A general two-dimensional quadratic form is represented by the expression \
",
  Cell[BoxData[
      \(TraditionalForm\`a\ x\^2\  + \ b\ x\ y\  + \ c\ y\^2\  + \ d\ x\  + \ 
        e\ y\  + \ f\)]],
  ".  Coordinate transformations can be used to eliminate the linear terms \
and the cross terms, thereby simplifying the quadratic form in the new \
coordinate system.\na) A translation of the origin is accomplished by the \
replacements ",
  Cell[BoxData[
      \(TraditionalForm\`{x \[Rule] x - x\_0, y \[Rule] y - y\_0}\)]],
  ".  Perform this transformation and isolate the coefficient of the linear \
terms, x and y.  [Note that the desired coefficient of x is independent of y, \
and vice versa, because we will handle the cross term separately.\nb) \
Determine the values of ",
  Cell[BoxData[
      \(TraditionalForm\`{a, b}\)]],
  " which eliminate the linear terms and produce a transformed expression in \
the new coordinate system.  [Hint: solve a pair of equations.]\nc) A rotation \
of the coordinate system is accomplished by the replacements ",
  Cell[BoxData[
      \(TraditionalForm\`{x \[Rule] x\ Cos[\[Theta]]\  + \ y\ Sin[\[Theta]], 
        y \[Rule] \(-x\)\ Sin[\[Theta]] + y\ Cos[\[Theta]]}\)]],
  ".  Perform this transformation and isolate the coefficient of the cross \
term.  Find a set of angles \[Theta] for which the cross term is absent in \
the rotated coordinate system.  Compare the new expressions for each \[Theta] \
solution."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Diagonalization of a quadratic form", "ExerciseTitle"],

Cell[TextData[{
  "A general two-dimensional quadratic form ",
  Cell[BoxData[
      \(TraditionalForm
      \`a\ x\^2\  + \ b\ x\ y\  + \ c\ y\^2\  + \ d\ x\  + \ e\ y\  + \ f
        \)]],
  "  can be expressed as ",
  Cell[BoxData[
      \(TraditionalForm\`X . A . X\  + \ B\ X\  + \ C\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`X = {x, y}\)]],
  " is a vector, ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " is a symmetric matrix, ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  " is a vector, and ",
  Cell[BoxData[
      \(TraditionalForm\`C\)]],
  " is a constant.\na) Determine ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`B\)]],
  ", and ",
  Cell[BoxData[
      \(TraditionalForm\`C\)]],
  " in terms of ",
  Cell[BoxData[
      \(TraditionalForm\`{a, b, c, d, e, f}\)]],
  ".  \nb) A general theorem of linear algebra shows that the matrix ",
  Cell[BoxData[
      \(TraditionalForm\`R\)]],
  " whose columns are the normalized eigenvectors of ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " effects a coordinate transformation ",
  Cell[BoxData[
      \(TraditionalForm\`R\^T . A . R \[Rule] D\)]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`R\^T\)]],
  " is the transpose of ",
  Cell[BoxData[
      \(TraditionalForm\`R\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`D\)]],
  " is a diagonal matrix carrying the eigenvalues of ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  " along the diagonal in the order given by ",
  Cell[BoxData[
      \(TraditionalForm\`R\)]],
  ".  Construct ",
  Cell[BoxData[
      \(TraditionalForm\`R\)]],
  " and evaluate the quadratic form in the new coordinate system obtained by \
the transformation ",
  Cell[BoxData[
      \(TraditionalForm\`X \[Rule] R\ X\)]],
  ".  Simplification should reveal the absence of cross terms.\nc) Determine \
a translation ",
  Cell[BoxData[
      \(TraditionalForm\`X \[Rule] X - X\_0\)]],
  " which eliminates terms linear in x and y, such that the new quadratic \
form becomes ",
  Cell[BoxData[
      \(TraditionalForm\`X\^\[Prime] . D . X\^\[Prime] + C\^\[Prime]\)]],
  ".\nd) Classify the quadratic has a circle, ellipse, parabola, hyperbola, \
or line according to the eigenvalues of ",
  Cell[BoxData[
      \(TraditionalForm\`A\)]],
  ".\ne) Challenge: generalize this procedure to three dimensions or higher."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Attack from a safe distance!", "ExerciseTitle"],

Cell[TextData[{
  "The following is based upon a problem formulated by W.M. MacDonald.  A ",
  StyleBox["ballista",
    FontSlant->"Italic"],
  " was a siege engine used by the Romans that could launch rocks against the \
wall of a fortified town or throw objects over the wall.  Although a ballista \
could easily throw an object over a wall of reasonable height if it could be \
placed close enough, defenders on top of the wall would rain arrows upon the \
attackers if they get too close.  Consequently, the attackers need to \
determine the maximum distance for which a projectile will surmount a wall of \
specified height.  Suppose that the ballista is located a distance ",
  Cell[BoxData[
      \(TraditionalForm\`d\)]],
  " from a wall of height ",
  Cell[BoxData[
      \(TraditionalForm\`h\)]],
  ".  Neglecting air resistance, the equations of motion for a projectile \
launched with initial velocity ",
  Cell[BoxData[
      \(TraditionalForm\`v\_0\)]],
  " at angle ",
  Cell[BoxData[
      \(TraditionalForm\`\[Theta]\)]],
  " above horizontal are: {",
  Cell[BoxData[
      \(TraditionalForm\`x = x\_0 + \(v\_0\) t\ cos\[Theta]\)]],
  ", ",
  Cell[BoxData[
      \(TraditionalForm\`y = 
        y\_0 + \(v\_0\) t\ sin\[Theta]\  - \ \(1\/2\) g\ t\^2\)]],
  "} where {",
  Cell[BoxData[
      \(TraditionalForm\`x\_0\)]],
  ",",
  Cell[BoxData[
      \(TraditionalForm\`y\_0\)]],
  "} are the initial coordinates.  Determine the relationship between ",
  Cell[BoxData[
      \(TraditionalForm\`d\_max\)]],
  " and ",
  Cell[BoxData[
      \(TraditionalForm\`h\)]],
  "."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Prank with remarkable consequences", "ExerciseTitle"],

Cell[TextData[{
  "Acton begins his book on ",
  StyleBox["Numerical Methods that Work ",
    FontSlant->"Italic"],
  "with the following problem.  A railroad track 1 mile long is firmly fixed \
at both ends.  During the night some prankster cuts the rail and welds in an \
additional foot, causing the rail to bow up in a circular arc.  What is the \
maximum height that this rail now achieves over its former position, assuming \
that the original track was a perfectly straight line?  Solve the problem \
using ",
  StyleBox["Mathematica",
    FontSlant->"Italic"],
  " and check the numerical accuracy of your result.  It is also worth \
considering how one would obtain an accurate answer by hand."
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Hilbert matrix", "ExerciseTitle"],

Cell["A Hilbert matrix is defined by", "ExerciseText"],

Cell[BoxData[
    \(hilbert[n_] := \ 
      Table[1\/\(i + j - 1\), {i, 1, n}, {j, 1, n}]\)], "InlineInput",
  Evaluatable->False,
  TextAlignment->Center,
  TextJustification->0],

Cell["\<\
a) Display the Hilbert matrix for n=3 and its inverse in conventional form.  \
Check the inverse by multiplying the matrices.  \
\>", "ExerciseText",
  CellDingbat->None],

Cell[TextData[{
  "b) Check the inversion for n=10 without attempting to display the \
matrices.  Find the largest element in the inverse of the n=10 Hilbert \
matrix.\nc) Repeat part b) using approximate real numbers, obtained with ",
  StyleBox["N", "InlineInput",
    FontWeight->"Bold"],
  ", instead of rational numbers.  Determine the largest difference between \
the identity matrix and the product of the hilbert matrix and its inverse.  \
Explain."
}], "ExerciseText",
  CellDingbat->None]
}, Closed]],

Cell[CellGroupData[{

Cell[TextData[{
  "Rotation matrix for spin-",
  Cell[BoxData[
      \(TraditionalForm\`1\/2\)]]
}], "ExerciseTitle"],

Cell[TextData[{
  "The Pauli spin vector ",
  Cell[BoxData[
      \(TraditionalForm\`\(\[Sigma]\& \[RightVector] \)\)]],
  " is defined as ",
  Cell[BoxData[
      \(TraditionalForm\`{\[Sigma]\_x, \[Sigma]\_y, \[Sigma]\_z}\)]],
  " where the customary representation of the Pauli matrices is given by"
}], "ExerciseText"],

Cell[TextData[{
  Cell[BoxData[
      FormBox[
        RowBox[{\(\[Sigma]\_x\), "=", 
          RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                {"0", "1"},
                {"1", "0"}
                }], "\[NegativeThinSpace]", ")"}]}], TraditionalForm]]],
  ",   ",
  Cell[BoxData[
      FormBox[
        RowBox[{\(\[Sigma]\_y\), "=", 
          RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                {"0", \(-I\)},
                {"I", "0"}
                }], "\[NegativeThinSpace]", ")"}]}], TraditionalForm]]],
  ",  ",
  Cell[BoxData[
      FormBox[
        RowBox[{\(\[Sigma]\_z\), "=", 
          RowBox[{"(", "\[NegativeThinSpace]", GridBox[{
                {"1", "0"},
                {"0", \(-1\)}
                }], "\[NegativeThinSpace]", ")"}]}], TraditionalForm]]]
}], "ExerciseText",
  TextAlignment->Center,
  TextJustification->0],

Cell[TextData[{
  "Let ",
  Cell[BoxData[
      \(TraditionalForm\`n\&^ = {n\_x, n\_y, n\_z}\)]],
  " represent an arbitrary unit vector.  Prove that ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          RowBox[{
            StyleBox["MatrixExp",
              FontWeight->"Bold"], 
            "[", \(\(-\(I\/2\)\) \[Theta]\ n\&^\[CenterDot]\[Sigma]\&\
\[RightVector]\), "]"}], 
          "==", \(\[DoubleStruckCapitalI]\ Cos[\[Theta]\/2] - 
            I\ n\&^\[CenterDot]\[Sigma]\&\[RightVector]\ Sin[\[Theta]\/2]\)}],
         TraditionalForm]]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`\[DoubleStruckCapitalI]\)]],
  " is the identity matrix.  [Hint: use the fact that ",
  Cell[BoxData[
      \(TraditionalForm\`n\_x\^2 + n\_y\^2 + n\_z\^2 = 1\)]],
  " for a unit vector and use ",
  StyleBox["ExpToTrig",
    FontWeight->"Bold"],
  ".]"
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Double pendulum", "ExerciseTitle"],

Cell["\<\
Analyze the normal modes of the double pendulum.  (Recall that the double \
pendulum consists of two similar pendulums with one suspended from a fixed \
pivot and the second suspended from the bob of the first.)  Assume that the \
amplitudes are sufficiently small to linearize the equations of motion.\
\>", "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Interpolating polynomials", "ExerciseTitle"],

Cell[TextData[{
  "Suppose that we are given the values of a function as a list of",
  StyleBox[" ", "InlineInput"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`x = {x\_0, x\_1, \ \[CenterEllipsis], \ x\_n}\)], 
    "InlineInput"], "InlineInput"],
  StyleBox[" ", "InlineInput"],
  StyleBox["and a list of corresponding values ", "Text"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`y = {y\_0, y\_1, \ \[CenterEllipsis], \ y\_n}\)], 
    "InlineInput"], "InlineInput"],
  StyleBox[" such that ", "Text"],
  Cell[BoxData[
      FormBox[
        StyleBox[\(y\_i == f[x\_i]\),
          "InlineInput"], TraditionalForm]]],
  ".  We wish to construct an interpolating polynomial ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(P[x] = 
            a\_0 + \(a\_1\) x\  + \ \(a\_2\) 
                x\^2 + \[CenterEllipsis]\  + \ \(a\_n\) x\^n\),
          "InlineInput"], TraditionalForm]]],
  " for which ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(P[x\_i] == y\_i\),
          "InlineInput"], TraditionalForm]]],
  " reproduces the specified function values at the points given.  Although \
there is no guarantee that this interpolation function will reproduce the \
function at intermediate points, it is likely to be a good approximation \
provided that the function is sufficiently smooth.\na) Show that the system \
of equations ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(P[x\_i] == y\_i\),
          "InlineInput"], TraditionalForm]]],
  " can be expressed as a matrix equation of the form ",
  StyleBox["V.a==x", "InlineInput"],
  " where ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(a = {a\_0, a\_\(\(1\)\(\ \)\), \[CenterEllipsis], \ 
              a\_n}\),
          "InlineInput"], TraditionalForm]]],
  " is a coefficient vector and ",
  StyleBox["V", "InlineInput"],
  " is a matrix known as the ",
  StyleBox["Vandermonde",
    FontSlant->"Italic"],
  " matrix for this system.  Produce a symbolic expression for ",
  StyleBox["V", "InlineInput"],
  ".\nb) Produce a list of 11 equally spaced values for the function ",
  StyleBox["f[x]=Exp[-x]", "InlineInput"],
  StyleBox[Cell[BoxData[
      \(TraditionalForm\`Cos[x\^2]\)], "InlineInput"], "InlineInput"],
  StyleBox[" ", "InlineInput"],
  StyleBox["over the range ", "Text"],
  StyleBox["{x,-1,1}", "InlineInput"],
  StyleBox[".  Construct the interpolating polynomial for these data and \
estimate the largest deviation between the function and the interpolating \
polynomial over this range.\nc) Compare the interpolating polynomial with the \
function it is supposed to represent over a range of ", "Text"],
  StyleBox["x", "Text",
    FontWeight->"Bold"],
  StyleBox[" that is larger than the range covered by the data used to \
construct the polynomial approximation.  What conclusions can you draw?", 
    "Text"]
}], "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{

Cell["Least-squares polynomials", "ExerciseTitle"],

Cell[TextData[{
  "Suppose that we are given the values of a function as a list of",
  StyleBox[" ", "InlineInput"],
  Cell[BoxData[
      \(TraditionalForm\`x = {x\_0, x\_1, \ \[CenterEllipsis], \ x\_n}\)], 
    "InlineInput"],
  StyleBox[" ", "InlineInput"],
  StyleBox["and a list of corresponding values ", "Text"],
  Cell[BoxData[
      \(TraditionalForm\`y = {y\_0, y\_1, \ \[CenterEllipsis], \ y\_n}\)], 
    "InlineInput"],
  StyleBox[" such that ", "Text"],
  Cell[BoxData[
      FormBox[
        StyleBox[\(y\_i == f[x\_i]\),
          "InlineInput"], TraditionalForm]]],
  ".  We wish to construct an ",
  Cell[BoxData[
      \(TraditionalForm\`m\^th\)]],
  "-order polynomial approximation ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          RowBox[{
            SubscriptBox[
              StyleBox["P",
                "InlineInput"], "m"], 
            StyleBox["[",
              "InlineInput"], 
            StyleBox["x",
              "InlineInput"], 
            StyleBox["]",
              "InlineInput"]}], 
          StyleBox["=",
            "InlineInput"], 
          
          StyleBox[\(a\_0 + \(a\_1\) x\  + \ \(a\_2\) 
                x\^2 + \[CenterEllipsis]\  + \ \(a\_m\) x\^m\),
            "InlineInput"]}], TraditionalForm]]],
  " where ",
  Cell[BoxData[
      \(TraditionalForm\`m \[LessEqual] n\)]],
  "  by minimizing the function ",
  Cell[BoxData[
      FormBox[
        RowBox[{
          SubscriptBox[
            StyleBox["S",
              "InlineInput"], "m"], 
          StyleBox["==",
            "InlineInput"], 
          RowBox[{\(1\/\(n - m\)\), 
            RowBox[{
              StyleBox[\(\[Sum]\+\(i = 0\)\%n\),
                "InlineInput"], 
              SuperscriptBox[
                RowBox[{
                  StyleBox["(",
                    "InlineInput"], \(y\_i - P\_m[x\_i]\), ")"}], "2"]}]}]}], 
        TraditionalForm]], "InlineInput"],
  StyleBox[".  ", "Text"],
  "Although there is no guarantee that this ",
  StyleBox["least-squares polynomial",
    FontSlant->"Italic"],
  " will reproduce the function at intermediate points, it is likely to be a \
good approximation provided that the function is sufficiently smooth.  This \
technique is also suitable for analyzing experimental data for which the \
measured values include random errors.\na) Show that the requirement that ",
  StyleBox["S",
    FontWeight->"Bold",
    FontSlant->"Italic"],
  " be minimized leads to a system of equations which can be expressed in the \
form ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(\((V\&~ . V)\) . a == V\&~ . y\),
          "InlineInput"], TraditionalForm]]],
  "  where ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(a = {a\_0, a\_\(\(1\)\(\ \)\), \[CenterEllipsis], \ 
              a\_m}\),
          "InlineInput"], TraditionalForm]]],
  " is the coefficient vector, ",
  StyleBox["V", "InlineInput"],
  " is an ",
  StyleBox["n\[Times]m",
    FontSlant->"Italic"],
  " matrix, and ",
  Cell[BoxData[
      FormBox[
        StyleBox[\(V\&~\),
          "InlineInput"], InputForm]]],
  " is the transpose of ",
  StyleBox["V", "InlineInput"],
  ".  Produce a symbolic expression for ",
  StyleBox["V", "InlineInput"],
  StyleBox[", which is sometimes known as the Vandermonde matrix.", "Text"],
  "\nb) Produce a list of 11 equally spaced values for the function ",
  StyleBox["f[x]=Exp[-x]", "InlineInput"],
  Cell[BoxData[
      \(TraditionalForm\`Cos[x\^2]\)], "InlineInput"],
  StyleBox[" ", "InlineInput"],
  StyleBox["over the range ", "Text"],
  StyleBox["{x,-1,1}", "InlineInput"],
  StyleBox[".  Write a function which constructs the least-squares polynomial \
for any ", "Text"],
  Cell[BoxData[
      \(TraditionalForm\`m \[LessEqual] n\)]],
  " and reports the associated variance.  Compare least-squares polynomials \
for several values of ",
  Cell[BoxData[
      \(TraditionalForm\`m\)]],
  " with the original function within the range fitted and comment.",
  StyleBox["\nc) Compare the least-squares polynomial with the function it is \
supposed to represent over a range of ", "Text"],
  StyleBox["x", "Text",
    FontWeight->"Bold"],
  StyleBox[" that is larger than the range covered by the data.  What \
conclusions can you draw?\nd) Repeat part b) using data in which small random \
errors are included.  An easy way to introduce such fluctuations is to \
multiply each data point by ", "Text"],
  StyleBox["Random[Real,{0.9,1.1}]", "InlineInput",
    FontWeight->"Bold"],
  StyleBox[".", "Text"]
}], "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
},
FrontEndVersion->"5.2 for Microsoft Windows",
ScreenRectangle->{{0, 1400}, {0, 952}},
ScreenStyleEnvironment->"Working",
WindowToolbars->"EditBar",
WindowSize->{1392, 912},
WindowMargins->{{0, Automatic}, {Automatic, 0}},
Visible->True,
ShowCellLabel->False,
InputAliases->{"notation"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftRightArrow]", " ",
           
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation>"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongRightArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "notation<"->RowBox[ {"Notation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], " ", "\[DoubleLongLeftArrow]", " ", 
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"]}], "]"}], 
    "symb"->RowBox[ {"Symbolize", "[", 
        TagBox[ 
        "\[Placeholder]", NotationBoxTag, TagStyle -> 
          "NotationTemplateStyle"], "]"}], 
    "infixnotation"->RowBox[ {"InfixNotation", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "addia"->RowBox[ {"AddInputAlias", "[", 
        RowBox[ {
          TagBox[ 
          "\[Placeholder]", NotationBoxTag, TagStyle -> 
            "NotationTemplateStyle"], ",", "\[Placeholder]"}], "]"}], 
    "pattwraper"->TagBox[ 
      "\[Placeholder]", NotationPatternTag, TagStyle -> 
        "NotationPatternWrapperStyle"], 
    "madeboxeswraper"->TagBox[ 
      "\[Placeholder]", NotationMadeBoxesTag, TagStyle -> 
        "NotationMadeBoxesWrapperStyle"]},
Magnification->1.5,
StyleDefinitions -> Notebook[{
  
  Cell[CellGroupData[{
  Cell["Style Definitions", "Subtitle"],
  
  Cell["\<\
Modify the definitions below to change the default appearance of all cells in \
a given style.  Make modifications to any definition using commands in the \
Format menu.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell["Style Environment Names", "Section"],
  
  Cell[StyleData[All, "Working"],
    PageWidth->WindowWidth,
    ScriptMinSize->9],
  
  Cell[StyleData[All, "Presentation"],
    PageWidth->WindowWidth,
    ScriptMinSize->12,
    FontSize->16],
  
  Cell[StyleData[All, "Condensed"],
    PageWidth->PaperWidth,
    CellBracketOptions->{"Margins"->{1, 1},
    "Widths"->{0, 5}},
    ScriptMinSize->7,
    FontSize->10],
  
  Cell[StyleData[All, "Printout"],
    PageWidth->PaperWidth,
    ScriptMinSize->7,
    FontSize->10,
    PrivateFontOptions->{"FontType"->"Outline"}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Notebook Options", "Section"],
  
  Cell["\<\
The options defined for the style below will be used at the Notebook level.\
\>", "Text"],
  
  Cell[StyleData["Notebook"],
    PageHeaders->{{Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"], None, Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"]}, {Cell[ 
            TextData[ {
              ValueBox[ "FileName"]}], "Header"], None, Cell[ 
            TextData[ {
              CounterBox[ "Page"]}], "PageNumber"]}},
    PageFooters->{{Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}, {
        Cell[ 
            TextData[ {"Essential Mathematica for Students of Science"}], 
            "Footer"], None, Cell[ 
            TextData[ {"\[Copyright] James J. Kelly, 1998"}], "Footer"]}},
    PageHeaderLines->{True, True},
    PageFooterLines->{True, True},
    PrintingOptions->{"FirstPageHeader"->False,
    "FacingPages"->True},
    CellFrameLabelMargins->6,
    StyleMenuListing->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headings", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Title"],
    ShowCellBracket->True,
    CellMargins->{{0, 0}, {0, 0}},
    CellGroupingRules->"TitleGrouping",
    PageBreakBelow->False,
    LineSpacing->{1, 0},
    CounterIncrements->"Title",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subtitle", 0}, {"Subsubtitle", 0}},
    FontSize->34,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Title", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 0},
    FontSize->44],
  
  Cell[StyleData["Title", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24],
  
  Cell[StyleData["Title", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    FontSize->24,
    FontTracking->"Plain",
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subtitle"],
    CellMargins->{{0, 0}, {0, 0}},
    PageBreakBelow->False,
    LineSpacing->{1, 3},
    ParagraphIndent->-96,
    CounterIncrements->"Subtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}, {
        "Subsubtitle", 0}},
    FontFamily->"Helvetica",
    FontSize->18,
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Subtitle", "Presentation"],
    CellMargins->{{0, 0}, {0, 0}},
    LineSpacing->{1, 6},
    ParagraphIndent->-157,
    FontSize->30],
  
  Cell[StyleData["Subtitle", "Condensed"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-78,
    FontSize->14],
  
  Cell[StyleData["Subtitle", "Printout"],
    CellMargins->{{0, 0}, {0, 0}},
    ParagraphIndent->-85,
    FontSize->16,
    FontColor->GrayLevel[0],
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubtitle"],
    ShowCellBracket->True,
    CellMargins->{{10, 4}, {30, 10}},
    PageBreakBelow->False,
    CounterIncrements->"Subsubtitle",
    CounterAssignments->{{"Section", 0}, {"Equation", 0}, {"Figure", 0}},
    FontFamily->"Helvetica",
    FontSize->14,
    FontSlant->"Italic"],
  
  Cell[StyleData["Subsubtitle", "Presentation"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {40, 20}},
    LineSpacing->{1, 0},
    FontSize->24],
  
  Cell[StyleData["Subsubtitle", "Condensed"],
    ShowCellBracket->True,
    CellMargins->{{8, 10}, {12, 8}},
    FontSize->12],
  
  Cell[StyleData["Subsubtitle", "Printout"],
    ShowCellBracket->True,
    CellMargins->{{9, 10}, {50, 10}},
    FontSize->14]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Section"],
    CellFrame->{{0, 0}, {0, 1}},
    CellDingbat->None,
    CellMargins->{{12, Inherited}, {4, 24}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameMargins->6,
    CounterIncrements->"Section",
    CounterAssignments->{{"Subsection", 0}, {"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Section", "Presentation"],
    CellMargins->{{10, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["Section", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["Section", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsection"],
    CellMargins->{{12, Inherited}, {8, 20}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsection", "Presentation"],
    CellMargins->{{11, 10}, {8, 32}},
    LineSpacing->{1, 0},
    FontSize->22],
  
  Cell[StyleData["Subsection", "Condensed"],
    CellMargins->{{8, Inherited}, {2, 12}},
    FontSize->12],
  
  Cell[StyleData["Subsection", "Printout"],
    CellMargins->{{9, 0}, {4, 40}},
    FontSize->12]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Subsubsection"],
    CellDingbat->"\[FilledSquare]",
    CellMargins->{{25, Inherited}, {8, 12}},
    CellGroupingRules->{"SectionGrouping", 50},
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold"],
  
  Cell[StyleData["Subsubsection", "Presentation"],
    CellMargins->{{29, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Subsubsection", "Condensed"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11],
  
  Cell[StyleData["Subsubsection", "Printout"],
    CellMargins->{{21, 0}, {4, 20}},
    FontSize->11]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Body Text", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Text"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Text",
    FontFamily->"Times"],
  
  Cell[StyleData["Text", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Text", "Condensed"],
    CellMargins->{{9, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["Text", "Printout"],
    CellMargins->{{9, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["SmallText"],
    CellMargins->{{12, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 6},
    CounterIncrements->"SmallText",
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["SmallText", "Presentation"],
    CellMargins->{{13, 10}, {8, 8}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["SmallText", "Condensed"],
    CellMargins->{{8, 10}, {2, 2}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["SmallText", "Printout"],
    CellMargins->{{9, 0}, {4, 4}}]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Input/Output", "Section"],
  
  Cell["\<\
The cells in this section define styles used for input and output to the \
kernel.  Be careful when modifying, renaming, or removing these styles, \
because the front end associates special meanings with these style names.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Input"],
    CellFrame->0.5,
    CellMargins->{{25, 10}, {8, 8}},
    Evaluatable->True,
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{5, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultInputFormatType,
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold",
    Background->RGBColor[1, 0.700008, 0.4]],
  
  Cell[StyleData["Input", "Presentation"],
    CellMargins->{{25, Inherited}, {10, 10}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Input", "Condensed"],
    CellMargins->{{25, 0}, {6, 6}}],
  
  Cell[StyleData["Input", "Printout"],
    CellFrame->0.5,
    CellMargins->{{25, 0}, {6, 6}},
    Background->None]
  }, Closed]],
  
  Cell[StyleData["InlineInput"],
    CellGroupingRules->"InputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    DefaultFormatType->DefaultInputFormatType,
    AutoItalicWords->{},
    FormatType->InputForm,
    ShowStringCharacters->True,
    NumberMarks->True,
    CounterIncrements->"Input",
    FontWeight->"Bold"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Output"],
    CellFrame->True,
    CellMargins->{{25, 10}, {8, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    CellEditDuplicate->True,
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    CellLabelMargins->{{3, Inherited}, {Inherited, Inherited}},
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Output",
    Background->RGBColor[1, 0.900008, 0.900008]],
  
  Cell[StyleData["Output", "Presentation"],
    CellMargins->{{25, Inherited}, {12, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Output", "Condensed"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4}],
  
  Cell[StyleData["Output", "Printout"],
    CellMargins->{{25, 0}, {6, 0}},
    CellElementSpacings->{"ClosedGroupTopMargin"->4},
    Background->None]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Message"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Message",
    StyleMenuListing->None,
    FontColor->RGBColor[1, 0, 0]],
  
  Cell[StyleData["Message", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Message", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Message", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Print"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"OutputGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Print",
    StyleMenuListing->None],
  
  Cell[StyleData["Print", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Print", "Condensed"],
    CellMargins->{{50, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Print", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Graphics"],
    CellMargins->{{62, Inherited}, {Inherited, Inherited}},
    CellGroupingRules->"GraphicsGrouping",
    CellHorizontalScrolling->True,
    PageBreakWithin->False,
    GeneratedCell->True,
    CellAutoOverwrite->True,
    ShowCellLabel->False,
    DefaultFormatType->DefaultOutputFormatType,
    FormatType->InputForm,
    CounterIncrements->"Graphics",
    StyleMenuListing->None],
  
  Cell[StyleData["Graphics", "Presentation"],
    CellMargins->{{74, Inherited}, {Inherited, Inherited}}],
  
  Cell[StyleData["Graphics", "Condensed"],
    CellMargins->{{52, Inherited}, {Inherited, Inherited}},
    ImageSize->{175, 175}],
  
  Cell[StyleData["Graphics", "Printout"],
    CellMargins->{{54, Inherited}, {Inherited, Inherited}},
    ImageSize->{250, 250}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["CellLabel"],
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->11,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["CellLabel", "Presentation"],
    FontSize->12],
  
  Cell[StyleData["CellLabel", "Condensed"],
    FontSize->8],
  
  Cell[StyleData["CellLabel", "Printout"],
    FontSize->8,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Unique Styles", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Author"],
    CellMargins->{{10, 4}, {2, 10}},
    FontSize->12,
    FontSlant->"Italic"],
  
  Cell[StyleData["Author", "Presentation"],
    CellMargins->{{12, 10}, {2, 12}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12},
    FontSize->16],
  
  Cell[StyleData["Author", "Condensed"],
    CellMargins->{{8, 10}, {1, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4},
    FontSize->9],
  
  Cell[StyleData["Author", "Printout"],
    CellMargins->{{9, 0}, {4, 12}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Copyright"],
    CellMargins->{{10, 10}, {40, 2}},
    FontFamily->"Helvetica",
    FontSize->9],
  
  Cell[StyleData["Copyright", "Presentation"],
    CellMargins->{{12, 10}, {50, 2}},
    LineSpacing->{1, 5},
    FontSize->12],
  
  Cell[StyleData["Copyright", "Condensed"],
    CellMargins->{{8, 10}, {12, 1}},
    LineSpacing->{1, 2},
    FontSize->9],
  
  Cell[StyleData["Copyright", "Printout"],
    CellMargins->{{9, 0}, {72, 4}},
    FontSize->7]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Abstract"],
    CellMargins->{{45, 75}, {Inherited, 30}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Abstract", "Printout"],
    CellMargins->{{36, 67}, {Inherited, 50}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Example"],
    CellMargins->{{12, 10}, {5, 12}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 12},
    CounterIncrements->"Example",
    FontFamily->"Times",
    FontWeight->"Bold"],
  
  Cell[StyleData["Example", "Presentation"],
    CellMargins->{{18, 10}, {8, 20}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["Example", "Condensed"],
    CellMargins->{{8, 10}, {4, 8}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 4}],
  
  Cell[StyleData["Example", "Printout"],
    CellMargins->{{9, 0}, {4, 10}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Definition"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Definition",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.2, 0.700008, 0.700008]],
  
  Cell[StyleData["Definition", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Definition", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Definition", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.6]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Theorem"],
    CellFrame->{{3, 0}, {0, 0}},
    CellMargins->{{52, 10}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Theorem",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontColor->GrayLevel[1],
    Background->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["Theorem", "Presentation"],
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Theorem", "Condensed"],
    CellMargins->{{40, 10}, {1, 4}}],
  
  Cell[StyleData["Theorem", "Printout"],
    CellMargins->{{44, 0}, {2, 6}},
    Background->GrayLevel[0.4]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Caution"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Caution",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0.996109, 0, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Caution", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Caution", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Caution", "Printout"],
    CellFrame->False,
    CellDingbat->"\[WarningSign]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Advice"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{52, 25}, {8, 8}},
    PageBreakWithin->False,
    GroupPageBreakWithin->False,
    CellLabelMargins->{{23, Inherited}, {Inherited, Inherited}},
    ShowStringCharacters->True,
    CounterIncrements->"Advice",
    FontFamily->"Helvetica",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->GrayLevel[0],
    Background->RGBColor[0, 1, 0],
    FontVariations->{"Underline"->False}],
  
  Cell[StyleData["Advice", "Presentation"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{62, Inherited}, {5, 12}},
    LineSpacing->{1, 0}],
  
  Cell[StyleData["Advice", "Condensed"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{40, 25}, {1, 4}}],
  
  Cell[StyleData["Advice", "Printout"],
    CellFrame->False,
    CellDingbat->"\[LightBulb]",
    CellMargins->{{44, 25}, {2, 6}},
    Background->GrayLevel[0.900008]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Exercise"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->"NormalGrouping",
    PageBreakBelow->False,
    CounterIncrements->"Subsubsection",
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0, 0, 0.996109]],
  
  Cell[StyleData["Exercise", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Exercise", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Exercise", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseMain"],
    CellFrame->{{6, 0}, {0, 1}},
    CellMargins->{{12, Inherited}, {4, 20}},
    CellGroupingRules->{"SectionGrouping", 30},
    PageBreakBelow->False,
    CellFrameColor->RGBColor[0.571389, 0.19675, 0.570504],
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Helvetica",
    FontSize->16,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseMain", "Presentation"],
    CellMargins->{{18, 10}, {8, 32}},
    LineSpacing->{1, 2},
    FontSize->24,
    FontTracking->"Condensed"],
  
  Cell[StyleData["ExerciseMain", "Condensed"],
    CellMargins->{{9, 0}, {2, 50}},
    FontSize->14],
  
  Cell[StyleData["ExerciseMain", "Printout"],
    CellMargins->{{9, 0}, {2, 50}},
    CellFrameColor->GrayLevel[0.500008],
    FontSize->14,
    FontTracking->"Plain",
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseTitle"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Bold",
    FontColor->RGBColor[0.571389, 0.19675, 0.570504]],
  
  Cell[StyleData["ExerciseTitle", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["ExerciseTitle", "Condensed"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["ExerciseTitle", "Printout"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ExerciseText"],
    CellMargins->{{32, 10}, {5, 5}},
    LineSpacing->{1, 3},
    ParagraphSpacing->{0, 8},
    CounterIncrements->"ExcersiceText",
    FontFamily->"Times"],
  
  Cell[StyleData["ExerciseText", "Presentation"],
    CellMargins->{{32, 10}, {8, 8}},
    LineSpacing->{1, 5},
    ParagraphSpacing->{0, 12}],
  
  Cell[StyleData["ExerciseText", "Condensed"],
    CellMargins->{{32, 0}, {4, 4}},
    LineSpacing->{1, 1},
    ParagraphSpacing->{0, 6}],
  
  Cell[StyleData["ExerciseText", "Printout"],
    CellMargins->{{32, 0}, {4, 4}},
    ParagraphSpacing->{0, 6}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Problem"],
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{23, Inherited}, {4, 18}},
    CellGroupingRules->{"SectionGrouping", 40},
    PageBreakBelow->False,
    CounterIncrements->"Subsection",
    CounterAssignments->{{"Subsubsection", 0}},
    FontFamily->"Times",
    FontSize->13,
    FontWeight->"Plain",
    FontColor->GrayLevel[0]],
  
  Cell[StyleData["Problem", "Presentation"],
    CellMargins->{{33, 10}, {8, 26}},
    LineSpacing->{1, 0},
    FontSize->18],
  
  Cell[StyleData["Problem", "Condensed"],
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11],
  
  Cell[StyleData["Problem", "Printout"],
    CellFrame->False,
    CellDingbat->"\[FilledDownTriangle]",
    CellMargins->{{9, 0}, {6, 20}},
    FontSize->11,
    FontColor->GrayLevel[0]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Formulas and Programming", "Section"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["DisplayFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    CellHorizontalScrolling->True,
    DefaultFormatType->DefaultInputFormatType,
    ScriptLevel->0,
    SingleLetterItalics->True,
    UnderoverscriptBoxOptions->{LimitsPositioning->True}],
  
  Cell[StyleData["DisplayFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}}],
  
  Cell[StyleData["DisplayFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["ChemicalFormula"],
    CellMargins->{{62, 10}, {2, 10}},
    DefaultFormatType->DefaultInputFormatType,
    AutoSpacing->False,
    ScriptLevel->1,
    ScriptBaselineShifts->{0.6, Automatic},
    SingleLetterItalics->False,
    ZeroWidthTimes->True],
  
  Cell[StyleData["ChemicalFormula", "Presentation"],
    CellMargins->{{74, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Condensed"],
    CellMargins->{{52, 10}, {2, 10}},
    FontSize->10],
  
  Cell[StyleData["ChemicalFormula", "Printout"],
    CellMargins->{{54, 10}, {2, 10}},
    FontSize->10]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Program"],
    CellMargins->{{12, 10}, {Inherited, 6}},
    FontFamily->"Courier"],
  
  Cell[StyleData["Program", "Presentation"],
    CellMargins->{{13, 30}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Condensed"],
    CellMargins->{{8, 10}, {Inherited, 4}},
    FontSize->9.5],
  
  Cell[StyleData["Program", "Printout"],
    CellMargins->{{9, 0}, {Inherited, 4}},
    FontSize->9.5]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Automatic Numbering", "Section"],
  
  Cell["\<\
The following styles are useful for numbered equations, figures, etc.  They \
automatically give the cell a FrameLabel containing a reference to a \
particular counter, and also increment that counter.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedEquation"],
    CellMargins->{{62, 10}, {Inherited, Inherited}},
    CellFrameLabels->{{None, Cell[ 
            TextData[ {"(", 
              CounterBox[ "NumberedEquation"], ")"}]]}, {None, None}},
    DefaultFormatType->DefaultInputFormatType,
    CounterIncrements->"NumberedEquation",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedEquation", "Presentation"],
    CellMargins->{{74, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Condensed"],
    CellMargins->{{52, 10}, {Inherited, Inherited}}],
  
  Cell[StyleData["NumberedEquation", "Printout"],
    CellMargins->{{54, 0}, {Inherited, Inherited}}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedFigure"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Figure ", 
              CounterBox[ "NumberedFigure"]}]], None}},
    CounterIncrements->"NumberedFigure",
    ImageMargins->{{43, Inherited}, {Inherited, 0}},
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedFigure", "Presentation"]],
  
  Cell[StyleData["NumberedFigure", "Condensed"]],
  
  Cell[StyleData["NumberedFigure", "Printout"]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["NumberedTable"],
    CellMargins->{{62, 145}, {Inherited, Inherited}},
    CellFrameLabels->{{None, None}, {Cell[ 
            TextData[ {"Table ", 
              CounterBox[ "NumberedTable"]}]], None}},
    TextAlignment->Center,
    CounterIncrements->"NumberedTable",
    FormatTypeAutoConvert->False],
  
  Cell[StyleData["NumberedTable", "Presentation"]],
  
  Cell[StyleData["NumberedTable", "Condensed"]],
  
  Cell[StyleData["NumberedTable", "Printout"]]
  }, Closed]]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Styles for Headers and Footers", "Section"],
  
  Cell[StyleData["Header"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["Footer"],
    CellMargins->{{0, 0}, {0, 4}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontSlant->"Italic"],
  
  Cell[StyleData["PageNumber"],
    CellMargins->{{0, 0}, {4, 1}},
    StyleMenuListing->None,
    FontFamily->"Helvetica",
    FontSize->9,
    FontWeight->"Bold"]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell["Hyperlink Styles", "Section"],
  
  Cell["\<\
The cells below define styles useful for making hypertext ButtonBoxes.  The \
\"Hyperlink\" style is for links within the same Notebook, or between \
Notebooks.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["Hyperlink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`NotebookLocate[ #2]}]&),
    Active->True,
    ButtonFrame->"None",
    ButtonNote->ButtonData}],
  
  Cell[StyleData["Hyperlink", "Presentation"]],
  
  Cell[StyleData["Hyperlink", "Condensed"]],
  
  Cell[StyleData["Hyperlink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell["\<\
The following styles are for linking automatically to the on-line help \
system.\
\>", "Text"],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["MainBookLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "MainBook", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["MainBookLink", "Presentation"]],
  
  Cell[StyleData["MainBookLink", "Condensed"]],
  
  Cell[StyleData["MainBookLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["AddOnsLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "AddOns", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["AddOnsLink", "Presentation"]],
  
  Cell[StyleData["AddOnsLink", "Condensed"]],
  
  Cell[StyleData["AddOnLink", "Printout"],
    FontWeight->"Bold",
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["RefGuideLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontFamily->"Courier",
    FontWeight->"Bold",
    FontSlant->"Plain",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "RefGuideLink", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["RefGuideLink", "Presentation"]],
  
  Cell[StyleData["RefGuideLink", "Condensed"]],
  
  Cell[StyleData["RefGuideLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["GettingStartedLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "GettingStarted", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["GettingStartedLink", "Presentation"]],
  
  Cell[StyleData["GettingStartedLink", "Condensed"]],
  
  Cell[StyleData["GettingStartedLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]],
  
  Cell[CellGroupData[{
  
  Cell[StyleData["OtherInformationLink"],
    StyleMenuListing->None,
    ButtonStyleMenuListing->Automatic,
    FontWeight->"Bold",
    FontSlant->"Italic",
    FontColor->RGBColor[0, 0, 1],
    Background->None,
    FontVariations->{"Underline"->True},
    ButtonBoxOptions->{ButtonFunction:>(FrontEndExecute[ {
        FrontEnd`HelpBrowserLookup[ "OtherInformation", #]}]&),
    Active->True,
    ButtonFrame->"None"}],
  
  Cell[StyleData["OtherInformationLink", "Presentation"]],
  
  Cell[StyleData["OtherInformationLink", "Condensed"]],
  
  Cell[StyleData["OtherInformationLink", "Printout"],
    FontColor->GrayLevel[0],
    Background->GrayLevel[1],
    FontVariations->{"Underline"->True}]
  }, Closed]]
  }, Closed]]
  }, Open  ]]
  }]
]

(*******************************************************************
Cached data follows.  If you edit this Notebook file directly, not
using Mathematica, you must remove the line containing CacheID at
the top of  the file.  The cache data will then be recreated when
you save this file from within Mathematica.
*******************************************************************)

(*CellTagsOutline
CellTagsIndex->{
  "FunctionExpand"->{
    Cell[43604, 1568, 98, 2, 95, "Input",
      CellTags->"FunctionExpand"],
    Cell[43705, 1572, 85, 2, 89, "Output",
      CellTags->"FunctionExpand"],
    Cell[43827, 1579, 108, 2, 73, "Input",
      CellTags->"FunctionExpand"],
    Cell[43938, 1583, 105, 2, 85, "Output",
      CellTags->"FunctionExpand"]},
  "Refine"->{
    Cell[50186, 1822, 164, 3, 73, "Input",
      CellTags->"Refine"],
    Cell[50353, 1827, 100, 2, 59, "Output",
      CellTags->"Refine"]}
  }
*)

(*CellTagsIndex
CellTagsIndex->{
  {"FunctionExpand", 417331, 12153},
  {"Refine", 417658, 12162}
  }
*)

(*NotebookFileOutline
Notebook[{

Cell[CellGroupData[{
Cell[1776, 53, 73, 2, 81, "Title"],
Cell[1852, 57, 455, 15, 85, "Author"],
Cell[2310, 74, 175, 3, 71, "Abstract"],

Cell[CellGroupData[{
Cell[2510, 81, 33, 0, 83, "Section"],
Cell[2546, 83, 119, 2, 99, "Input"],
Cell[2668, 87, 65, 1, 73, "Input"]
}, Closed]],

Cell[CellGroupData[{
Cell[2770, 93, 56, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[2851, 97, 42, 0, 68, "Subsection"],
Cell[2896, 99, 219, 7, 42, "Text"],

Cell[CellGroupData[{
Cell[3140, 110, 74, 1, 95, "Input"],
Cell[3217, 113, 100, 2, 78, "Output"]
}, Closed]],
Cell[3332, 118, 218, 7, 42, "Text"],

Cell[CellGroupData[{
Cell[3575, 129, 43, 1, 73, "Input"],
Cell[3621, 132, 66, 1, 75, "Output"]
}, Closed]],
Cell[3702, 136, 237, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[3964, 148, 78, 1, 103, "Input"],
Cell[4045, 151, 56, 1, 75, "Output"]
}, Closed]],
Cell[4116, 155, 389, 10, 68, "Text"],

Cell[CellGroupData[{
Cell[4530, 169, 170, 3, 133, "Input"],
Cell[4703, 174, 75, 1, 61, "Output"]
}, Closed]],
Cell[4793, 178, 367, 12, 42, "Text"],

Cell[CellGroupData[{
Cell[5185, 194, 165, 3, 133, "Input"],
Cell[5353, 199, 70, 1, 61, "Output"]
}, Closed]],
Cell[5438, 203, 122, 4, 66, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[5597, 212, 40, 0, 68, "Subsection"],
Cell[5640, 214, 318, 10, 42, "Text"],

Cell[CellGroupData[{
Cell[5983, 228, 110, 2, 99, "Input"],
Cell[6096, 232, 84, 1, 83, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[6217, 238, 112, 2, 99, "Input"],
Cell[6332, 242, 87, 1, 76, "Output"]
}, Closed]],
Cell[6434, 246, 250, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[6709, 258, 101, 2, 99, "Input"],
Cell[6813, 262, 111, 2, 83, "Output"]
}, Closed]],
Cell[6939, 267, 240, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[7204, 279, 104, 2, 99, "Input"],
Cell[7311, 283, 130, 2, 81, "Output"]
}, Closed]],
Cell[7456, 288, 68, 0, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[7561, 293, 36, 0, 68, "Subsection"],
Cell[7600, 295, 192, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[7817, 305, 102, 2, 73, "Input"],
Cell[7922, 309, 206, 3, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[8165, 317, 44, 1, 73, "Input"],
Cell[8212, 320, 92, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[8341, 326, 114, 2, 99, "Input"],
Cell[8458, 330, 173, 2, 81, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[8668, 337, 44, 1, 73, "Input"],
Cell[8715, 340, 121, 2, 78, "Output"]
}, Closed]],
Cell[8851, 345, 159, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[9035, 354, 53, 1, 74, "Input"],
Cell[9091, 357, 42, 1, 61, "Output"]
}, Closed]],
Cell[9148, 361, 241, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[9414, 373, 80, 1, 74, "Input"],
Cell[9497, 376, 87, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[9621, 382, 48, 1, 73, "Input"],
Cell[9672, 385, 136, 2, 59, "Output"]
}, Closed]],
Cell[9823, 390, 242, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[10090, 400, 80, 1, 73, "Input"],
Cell[10173, 403, 106, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[10316, 410, 82, 1, 73, "Input"],
Cell[10401, 413, 65, 1, 59, "Output"]
}, Closed]],
Cell[10481, 417, 215, 3, 60, "Exercise"],
Cell[10699, 422, 74, 3, 83, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[10810, 430, 38, 0, 68, "Subsection"],
Cell[10851, 432, 306, 10, 42, "Text"],

Cell[CellGroupData[{
Cell[11182, 446, 102, 2, 74, "Input"],
Cell[11287, 450, 80, 1, 61, "Output"]
}, Closed]],
Cell[11382, 454, 754, 23, 69, "Text"],

Cell[CellGroupData[{
Cell[12161, 481, 113, 2, 74, "Input"],
Cell[12277, 485, 83, 1, 61, "Output"]
}, Closed]],
Cell[12375, 489, 320, 10, 42, "Text"],

Cell[CellGroupData[{
Cell[12720, 503, 123, 2, 74, "Input"],
Cell[12846, 507, 174, 3, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[13057, 515, 123, 2, 74, "Input"],
Cell[13183, 519, 175, 3, 61, "Output"]
}, Closed]],
Cell[13373, 525, 114, 3, 41, "Text"],
Cell[13490, 530, 365, 12, 42, "Text"],

Cell[CellGroupData[{
Cell[13880, 546, 113, 2, 74, "Input"],
Cell[13996, 550, 72, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[14105, 556, 118, 2, 74, "Input"],
Cell[14226, 560, 72, 1, 59, "Output"]
}, Closed]],
Cell[14313, 564, 142, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[14480, 571, 79, 1, 100, "Input"],
Cell[14562, 574, 65, 1, 75, "Output"]
}, Closed]],
Cell[14642, 578, 1124, 39, 69, "Text"],

Cell[CellGroupData[{
Cell[15791, 621, 190, 3, 126, "Input"],
Cell[15984, 626, 413, 6, 87, "Output"]
}, Closed]],
Cell[16412, 635, 965, 28, 96, "Text"],

Cell[CellGroupData[{
Cell[17402, 667, 118, 2, 100, "Input"],
Cell[17523, 671, 137, 2, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[17697, 678, 88, 1, 74, "Input"],
Cell[17788, 681, 536, 14, 164, "Output"]
}, Closed]],
Cell[18339, 698, 331, 7, 108, "Exercise"],
Cell[18673, 707, 531, 17, 83, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[19241, 729, 61, 0, 68, "Subsection"],
Cell[19305, 731, 1651, 35, 201, "Text"],

Cell[CellGroupData[{
Cell[20981, 770, 91, 1, 73, "Input"],
Cell[21075, 773, 92, 1, 59, "Output"]
}, Closed]],
Cell[21182, 777, 563, 10, 231, "Text"],
Cell[21748, 789, 149, 2, 132, "Input"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[21946, 797, 60, 0, 51, "Section"],
Cell[22009, 799, 413, 7, 93, "Text"],
Cell[22425, 808, 306, 9, 42, "Text"],

Cell[CellGroupData[{
Cell[22756, 821, 69, 1, 73, "Input"],
Cell[22828, 824, 77, 1, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[22942, 830, 71, 1, 73, "Input"],
Cell[23016, 833, 81, 1, 61, "Output"]
}, Closed]],
Cell[23112, 837, 193, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[23330, 847, 101, 2, 74, "Input"],
Cell[23434, 851, 155, 3, 59, "Output"]
}, Closed]],
Cell[23604, 857, 327, 9, 68, "Text"],

Cell[CellGroupData[{
Cell[23956, 870, 68, 1, 74, "Input"],
Cell[24027, 873, 55, 1, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[24119, 879, 68, 1, 74, "Input"],
Cell[24190, 882, 35, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[24262, 888, 90, 1, 74, "Input"],
Cell[24355, 891, 49, 1, 75, "Output"]
}, Closed]],
Cell[24419, 895, 202, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[24646, 905, 83, 1, 74, "Input"],
Cell[24732, 908, 160, 2, 80, "Output"]
}, Closed]],
Cell[24907, 913, 202, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[25134, 923, 51, 1, 73, "Input"],
Cell[25188, 926, 67, 1, 75, "Output"]
}, Closed]],
Cell[25270, 930, 73, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[25368, 934, 113, 2, 126, "Input"],
Cell[25484, 938, 74, 1, 75, "Output"]
}, Closed]],
Cell[25573, 942, 293, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[25891, 954, 107, 2, 96, "Input"],
Cell[26001, 958, 38, 1, 59, "Output"]
}, Closed]],
Cell[26054, 962, 59, 3, 83, "Exercise"],
Cell[26116, 967, 198, 7, 58, "Exercise"],
Cell[26317, 976, 211, 7, 83, "Exercise"],
Cell[26531, 985, 223, 5, 58, "Exercise"],
Cell[26757, 992, 225, 6, 69, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[27019, 1003, 45, 0, 51, "Section"],
Cell[27067, 1005, 321, 8, 68, "Text"],

Cell[CellGroupData[{
Cell[27413, 1017, 51, 1, 85, "Input"],
Cell[27467, 1020, 40, 1, 64, "Output"]
}, Closed]],
Cell[27522, 1024, 111, 5, 42, "Text"],

Cell[CellGroupData[{
Cell[27658, 1033, 54, 1, 73, "Input"],
Cell[27715, 1036, 43, 1, 59, "Output"]
}, Closed]],
Cell[27773, 1040, 526, 12, 94, "Text"],

Cell[CellGroupData[{
Cell[28324, 1056, 54, 1, 85, "Input"],
Cell[28381, 1059, 35, 1, 59, "Output"]
}, Closed]],
Cell[28431, 1063, 64, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[28520, 1067, 63, 1, 74, "Input"],
Cell[28586, 1070, 55, 1, 59, "Output"]
}, Closed]],
Cell[28656, 1074, 512, 14, 68, "Text"],

Cell[CellGroupData[{
Cell[29193, 1092, 67, 1, 73, "Input"],
Cell[29263, 1095, 53, 1, 59, "Output"]
}, Closed]],
Cell[29331, 1099, 138, 3, 41, "Text"],
Cell[29472, 1104, 189, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[29686, 1114, 56, 1, 85, "Input"],
Cell[29745, 1117, 40, 1, 64, "Output"]
}, Closed]],
Cell[29800, 1121, 247, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[30072, 1133, 61, 1, 73, "Input"],
Cell[30136, 1136, 80, 1, 66, "Output"]
}, Closed]],
Cell[30231, 1140, 397, 14, 42, "Text"],

Cell[CellGroupData[{
Cell[30653, 1158, 77, 1, 73, "Input"],
Cell[30733, 1161, 82, 1, 59, "Output"]
}, Closed]],
Cell[30830, 1165, 525, 15, 68, "Text"],

Cell[CellGroupData[{
Cell[31380, 1184, 110, 2, 74, "Input"],
Cell[31493, 1188, 392, 7, 97, "Output"]
}, Closed]],
Cell[31900, 1198, 221, 6, 42, "Text"],
Cell[32124, 1206, 226, 5, 94, "Caution"],
Cell[32353, 1213, 261, 11, 59, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[32651, 1229, 59, 0, 51, "Section"],
Cell[32713, 1231, 1006, 24, 123, "Text"],
Cell[33722, 1257, 111, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[33858, 1264, 200, 3, 74, "Input"],
Cell[34061, 1269, 70, 1, 83, "Output"]
}, Closed]],
Cell[34146, 1273, 128, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[34299, 1280, 146, 2, 103, "Input"],
Cell[34448, 1284, 59, 1, 75, "Output"]
}, Closed]],
Cell[34522, 1288, 574, 18, 69, "Text"],

Cell[CellGroupData[{
Cell[35121, 1310, 102, 2, 73, "Input"],
Cell[35226, 1314, 59, 1, 59, "Output"]
}, Closed]],
Cell[35300, 1318, 627, 13, 95, "Text"],
Cell[35930, 1333, 79, 0, 71, "Caution"],
Cell[36012, 1335, 1420, 23, 198, "Text"],
Cell[37435, 1360, 341, 12, 101, "Exercise"],
Cell[37779, 1374, 377, 10, 108, "Exercise"],
Cell[38159, 1386, 575, 12, 116, "Exercise"],

Cell[CellGroupData[{
Cell[38759, 1402, 32, 0, 55, "Subsubsection"],
Cell[38794, 1404, 212, 4, 67, "Text"],
Cell[39009, 1410, 112, 2, 100, "Input"],
Cell[39124, 1414, 575, 21, 41, "Text"],
Cell[39702, 1437, 91, 1, 73, "Input"],
Cell[39796, 1440, 387, 10, 69, "Text"],

Cell[CellGroupData[{
Cell[40208, 1454, 64, 1, 73, "Input"],
Cell[40275, 1457, 109, 2, 83, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[40421, 1464, 67, 1, 73, "Input"],
Cell[40491, 1467, 109, 2, 59, "Output"]
}, Closed]],
Cell[40615, 1472, 256, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[40896, 1482, 93, 1, 73, "Input"],
Cell[40992, 1485, 156, 2, 59, "Output"]
}, Closed]],
Cell[41163, 1490, 680, 21, 67, "Text"],

Cell[CellGroupData[{
Cell[41868, 1515, 140, 2, 73, "Input"],
Cell[42011, 1519, 112, 2, 59, "Output"]
}, Closed]],
Cell[42138, 1524, 261, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[42424, 1533, 291, 5, 99, "Input"],
Cell[42718, 1540, 86, 1, 59, "Output"]
}, Closed]],
Cell[42819, 1544, 482, 7, 93, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[43338, 1556, 36, 0, 68, "Subsection"],
Cell[43377, 1558, 202, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[43604, 1568, 98, 2, 95, "Input",
  CellTags->"FunctionExpand"],
Cell[43705, 1572, 85, 2, 89, "Output",
  CellTags->"FunctionExpand"]
}, Closed]],

Cell[CellGroupData[{
Cell[43827, 1579, 108, 2, 73, "Input",
  CellTags->"FunctionExpand"],
Cell[43938, 1583, 105, 2, 85, "Output",
  CellTags->"FunctionExpand"]
}, Closed]],
Cell[44058, 1588, 116, 5, 67, "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[44223, 1599, 50, 0, 51, "Section"],
Cell[44276, 1601, 1672, 47, 149, "Text"],
Cell[45951, 1650, 58, 2, 41, "Text"],
Cell[46012, 1654, 68, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[46105, 1659, 47, 1, 73, "Input"],
Cell[46155, 1662, 57, 1, 59, "Output"]
}, Closed]],
Cell[46227, 1666, 271, 10, 41, "Text"],

Cell[CellGroupData[{
Cell[46523, 1680, 56, 1, 73, "Input"],
Cell[46582, 1683, 50, 1, 61, "Output"]
}, Closed]],
Cell[46647, 1687, 499, 13, 67, "Text"],

Cell[CellGroupData[{
Cell[47171, 1704, 103, 2, 86, "Input"],
Cell[47277, 1708, 59, 1, 66, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[47373, 1714, 135, 3, 80, "Input"],
Cell[47511, 1719, 38, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[47586, 1725, 108, 2, 73, "Input"],
Cell[47697, 1729, 35, 1, 59, "Output"]
}, Closed]],
Cell[47747, 1733, 1227, 36, 123, "Text"],

Cell[CellGroupData[{
Cell[48999, 1773, 115, 2, 73, "Input"],
Cell[49117, 1777, 66, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[49220, 1783, 92, 1, 96, "Input"],
Cell[49315, 1786, 40, 1, 59, "Output"]
}, Closed]],
Cell[49370, 1790, 421, 10, 69, "Text"],

Cell[CellGroupData[{
Cell[49816, 1804, 81, 1, 73, "Input"],
Cell[49900, 1807, 50, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[49987, 1813, 81, 1, 73, "Input"],
Cell[50071, 1816, 78, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[50186, 1822, 164, 3, 73, "Input",
  CellTags->"Refine"],
Cell[50353, 1827, 100, 2, 59, "Output",
  CellTags->"Refine"]
}, Closed]],
Cell[50468, 1832, 472, 7, 93, "Text"],
Cell[50943, 1841, 352, 5, 125, "Input"],

Cell[CellGroupData[{
Cell[51320, 1850, 202, 3, 107, "Input"],
Cell[51525, 1855, 368, 6, 60, "Message"],
Cell[51896, 1863, 121, 2, 97, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[52054, 1870, 316, 6, 96, "Input"],
Cell[52373, 1878, 171, 3, 86, "Output"]
}, Closed]],
Cell[52559, 1884, 666, 15, 93, "Text"],
Cell[53228, 1901, 330, 9, 83, "Exercise"],
Cell[53561, 1912, 392, 14, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[53990, 1931, 49, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[54064, 1935, 34, 0, 68, "Subsection"],
Cell[54101, 1937, 1007, 33, 69, "Text"],
Cell[55111, 1972, 85, 1, 74, "Input"],

Cell[CellGroupData[{
Cell[55221, 1977, 55, 1, 73, "Input"],
Cell[55279, 1980, 150, 2, 84, "Output"]
}, Closed]],
Cell[55444, 1985, 318, 5, 67, "Text"],

Cell[CellGroupData[{
Cell[55787, 1994, 93, 1, 73, "Input"],
Cell[55883, 1997, 77, 1, 84, "Output"]
}, Closed]],
Cell[55975, 2001, 349, 10, 42, "Text"],
Cell[56327, 2013, 249, 4, 67, "Text"],

Cell[CellGroupData[{
Cell[56601, 2021, 56, 1, 73, "Input"],
Cell[56660, 2024, 46, 1, 59, "Output"]
}, Closed]],
Cell[56721, 2028, 1568, 51, 68, "Text"],

Cell[CellGroupData[{
Cell[58314, 2083, 140, 2, 74, "Input"],
Cell[58457, 2087, 266, 5, 93, "Output"]
}, Closed]],
Cell[58738, 2095, 224, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[58987, 2105, 92, 1, 74, "Input"],
Cell[59082, 2108, 115, 2, 59, "Output"]
}, Closed]],
Cell[59212, 2113, 274, 7, 42, "Text"],

Cell[CellGroupData[{
Cell[59511, 2124, 70, 1, 74, "Input"],
Cell[59584, 2127, 144, 2, 84, "Output"]
}, Closed]],
Cell[59743, 2132, 361, 9, 68, "Text"],

Cell[CellGroupData[{
Cell[60129, 2145, 71, 1, 74, "Input"],
Cell[60203, 2148, 261, 4, 92, "Output"]
}, Closed]],
Cell[60479, 2155, 510, 14, 68, "Text"],
Cell[60992, 2171, 599, 10, 94, "Text"],

Cell[CellGroupData[{
Cell[61616, 2185, 179, 3, 74, "Input"],
Cell[61798, 2190, 164, 3, 61, "Output"]
}, Closed]],
Cell[61977, 2196, 666, 20, 69, "Text"],
Cell[62646, 2218, 153, 2, 58, "Exercise"],
Cell[62802, 2222, 123, 3, 58, "Exercise"],
Cell[62928, 2227, 607, 18, 110, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[63572, 2250, 53, 0, 68, "Subsection"],
Cell[63628, 2252, 445, 11, 67, "Text"],

Cell[CellGroupData[{
Cell[64098, 2267, 120, 3, 73, "Input"],
Cell[64221, 2272, 1695, 25, 191, "Output"]
}, Closed]],
Cell[65931, 2300, 320, 9, 42, "Text"],

Cell[CellGroupData[{
Cell[66276, 2313, 59, 1, 73, "Input"],
Cell[66338, 2316, 17170, 269, 1652, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[83545, 2590, 52, 1, 73, "Input"],
Cell[83600, 2593, 36, 1, 59, "Output"]
}, Closed]],
Cell[83651, 2597, 628, 17, 69, "Text"],

Cell[CellGroupData[{
Cell[84304, 2618, 87, 1, 73, "Input"],
Cell[84394, 2621, 458, 6, 62, "Message"],
Cell[84855, 2629, 459, 6, 62, "Message"],
Cell[85317, 2637, 459, 6, 62, "Message"],
Cell[85779, 2645, 328, 5, 35, "Message"],
Cell[86110, 2652, 56371, 2018, 303, 25134, 1628, "GraphicsData", \
"PostScript", "Graphics"]
}, Closed]],
Cell[142496, 4673, 607, 13, 94, "Text"],
Cell[143106, 4688, 416, 8, 144, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[143559, 4701, 68, 0, 68, "Subsection"],
Cell[143630, 4703, 254, 7, 41, "Text"],

Cell[CellGroupData[{
Cell[143909, 4714, 65, 1, 73, "Input"],
Cell[143977, 4717, 368, 6, 60, "Message"],
Cell[144348, 4725, 96, 2, 75, "Output"]
}, Closed]],
Cell[144459, 4730, 310, 4, 67, "Text"],

Cell[CellGroupData[{
Cell[144794, 4738, 60, 1, 73, "Input"],
Cell[144857, 4741, 332, 5, 60, "Message"],
Cell[145192, 4748, 61, 1, 59, "Output"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[145302, 4755, 30, 0, 68, "Subsection"],

Cell[CellGroupData[{
Cell[145357, 4759, 71, 0, 55, "Subsubsection"],
Cell[145431, 4761, 1046, 31, 93, "Text"],
Cell[146480, 4794, 137, 3, 85, "Input"],
Cell[146620, 4799, 291, 7, 67, "Text"],

Cell[CellGroupData[{
Cell[146936, 4810, 65, 1, 73, "Input"],
Cell[147004, 4813, 1153, 17, 216, "Output"]
}, Closed]],
Cell[148172, 4833, 761, 16, 119, "Text"],
Cell[148936, 4851, 215, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[149176, 4861, 135, 3, 73, "Input"],
Cell[149314, 4866, 853, 12, 165, "Output"]
}, Closed]],
Cell[150182, 4881, 168, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[150375, 4888, 108, 2, 73, "Input"],
Cell[150486, 4892, 203, 3, 83, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[150726, 4900, 56, 1, 73, "Input"],
Cell[150785, 4903, 142, 2, 70, "Output"]
}, Closed]],
Cell[150942, 4908, 89, 2, 41, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[151068, 4915, 67, 0, 55, "Subsubsection"],
Cell[151138, 4917, 1045, 32, 93, "Text"],
Cell[152186, 4951, 228, 3, 97, "Input"],
Cell[152417, 4956, 520, 12, 94, "Text"],
Cell[152940, 4970, 390, 6, 99, "Input"],
Cell[153333, 4978, 1329, 43, 120, "Text"],

Cell[CellGroupData[{
Cell[154687, 5025, 149, 3, 73, "Input"],
Cell[154839, 5030, 12678, 195, 2479, "Output"]
}, Closed]],
Cell[167532, 5228, 216, 4, 67, "Text"],

Cell[CellGroupData[{
Cell[167773, 5236, 221, 3, 73, "Input"],
Cell[167997, 5241, 251, 4, 86, "Output"]
}, Closed]],
Cell[168263, 5248, 670, 12, 94, "Text"]
}, Closed]],

Cell[CellGroupData[{
Cell[168970, 5265, 73, 0, 55, "Subsubsection"],
Cell[169046, 5267, 291, 7, 67, "Text"],
Cell[169340, 5276, 65, 1, 73, "Input"],
Cell[169408, 5279, 851, 25, 99, "Input"],
Cell[170262, 5306, 672, 12, 149, "Input"],
Cell[170937, 5320, 212, 8, 41, "Text"],

Cell[CellGroupData[{
Cell[171174, 5332, 108, 2, 73, "Input"],
Cell[171285, 5336, 456, 8, 101, "Output"]
}, Closed]],
Cell[171756, 5347, 135, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[171916, 5354, 353, 8, 74, "Input"],
Cell[172272, 5364, 383, 7, 101, "Output"]
}, Closed]],
Cell[172670, 5374, 230, 5, 41, "Text"],

Cell[CellGroupData[{
Cell[172925, 5383, 183, 4, 73, "Input"],
Cell[173111, 5389, 1480, 20, 276, "Output"]
}, Closed]],
Cell[174606, 5412, 356, 7, 67, "Text"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[175023, 5426, 50, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[175098, 5430, 42, 0, 68, "Subsection"],
Cell[175143, 5432, 675, 18, 95, "Text"],

Cell[CellGroupData[{
Cell[175843, 5454, 73, 1, 74, "Input"],
Cell[175919, 5457, 559, 10, 84, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[176515, 5472, 114, 2, 74, "Input"],
Cell[176632, 5476, 846, 14, 84, "Output"]
}, Closed]],
Cell[177493, 5493, 161, 4, 42, "Text"],

Cell[CellGroupData[{
Cell[177679, 5501, 151, 2, 86, "Input"],
Cell[177833, 5505, 179, 3, 59, "Output"]
}, Closed]],
Cell[178027, 5511, 73, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[178125, 5515, 78, 1, 74, "Input"],
Cell[178206, 5518, 368, 6, 60, "Message"],
Cell[178577, 5526, 189, 3, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[178803, 5534, 108, 2, 74, "Input"],
Cell[178914, 5538, 368, 6, 60, "Message"],
Cell[179285, 5546, 167, 3, 59, "Output"]
}, Closed]],
Cell[179467, 5552, 319, 7, 68, "Text"],
Cell[179789, 5561, 275, 7, 68, "Text"],
Cell[180067, 5570, 152, 3, 136, "Input"],

Cell[CellGroupData[{
Cell[180244, 5577, 51, 1, 73, "Input"],
Cell[180298, 5580, 332, 5, 60, "Message"],
Cell[180633, 5587, 103, 2, 84, "Output"]
}, Closed]],
Cell[180751, 5592, 416, 10, 126, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[181204, 5607, 46, 0, 68, "Subsection"],
Cell[181253, 5609, 2027, 57, 122, "Text"],
Cell[183283, 5668, 274, 7, 67, "Text"],
Cell[183560, 5677, 69, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[183654, 5682, 67, 1, 73, "Input"],
Cell[183724, 5685, 65, 1, 59, "Output"]
}, Closed]],
Cell[183804, 5689, 158, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[183987, 5696, 74, 1, 73, "Input"],
Cell[184064, 5699, 63, 1, 61, "Output"]
}, Closed]],
Cell[184142, 5703, 198, 4, 41, "Text"],
Cell[184343, 5709, 190, 4, 41, "Text"],
Cell[184536, 5715, 153, 3, 136, "Input"],

Cell[CellGroupData[{
Cell[184714, 5722, 132, 2, 73, "Input"],
Cell[184849, 5726, 44357, 780, 340, 3919, 276, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],
Cell[229221, 6509, 125, 3, 41, "Text"],

Cell[CellGroupData[{
Cell[229371, 6516, 78, 1, 73, "Input"],
Cell[229452, 6519, 118, 2, 59, "Output"]
}, Closed]],
Cell[229585, 6524, 137, 5, 58, "Exercise"],
Cell[229725, 6531, 197, 4, 83, "Exercise"],
Cell[229925, 6537, 329, 9, 83, "Exercise"]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[230303, 6552, 33, 0, 51, "Section"],

Cell[CellGroupData[{
Cell[230361, 6556, 41, 0, 68, "Subsection"],
Cell[230405, 6558, 356, 8, 67, "Text"],
Cell[230764, 6568, 12178, 396, 453, "Input",
  Evaluatable->False],
Cell[242945, 6966, 74, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[243044, 6970, 60, 1, 73, "Input"],
Cell[243107, 6973, 130, 2, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[243274, 6980, 55, 1, 73, "Input"],
Cell[243332, 6983, 429, 11, 98, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[243798, 6999, 70, 1, 73, "Input"],
Cell[243871, 7002, 429, 11, 98, "Output"]
}, Closed]],
Cell[244315, 7016, 183, 5, 42, "Text"],
Cell[244501, 7023, 71, 1, 73, "Input"],

Cell[CellGroupData[{
Cell[244597, 7028, 45, 1, 73, "Input"],
Cell[244645, 7031, 189, 3, 84, "Output"]
}, Closed]],
Cell[244849, 7037, 102, 2, 58, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[244988, 7044, 53, 0, 68, "Subsection"],
Cell[245044, 7046, 494, 17, 42, "Text"],
Cell[245541, 7065, 151, 3, 125, "Input"],

Cell[CellGroupData[{
Cell[245717, 7072, 66, 1, 73, "Input"],
Cell[245786, 7075, 240, 4, 84, "Output"]
}, Closed]],
Cell[246041, 7082, 230, 8, 42, "Text"],

Cell[CellGroupData[{
Cell[246296, 7094, 73, 1, 73, "Input"],
Cell[246372, 7097, 233, 3, 84, "Output"]
}, Closed]],
Cell[246620, 7103, 349, 7, 68, "Text"],
Cell[246972, 7112, 1043, 27, 123, "Text"],

Cell[CellGroupData[{
Cell[248040, 7143, 205, 4, 151, "Input"],
Cell[248248, 7149, 61, 1, 75, "Output"]
}, Closed]],
Cell[248324, 7153, 505, 17, 42, "Text"],

Cell[CellGroupData[{
Cell[248854, 7174, 121, 2, 99, "Input"],
Cell[248978, 7178, 45, 1, 59, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[249060, 7184, 80, 1, 73, "Input"],
Cell[249143, 7187, 43, 1, 59, "Output"]
}, Closed]],
Cell[249201, 7191, 289, 9, 42, "Text"],
Cell[249493, 7202, 346, 11, 59, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[249876, 7218, 34, 0, 68, "Subsection"],
Cell[249913, 7220, 945, 22, 122, "Text"],

Cell[CellGroupData[{
Cell[250883, 7246, 59, 0, 55, "Subsubsection"],
Cell[250945, 7248, 54207, 1612, 108, 23226, 1225, "GraphicsData", \
"PostScript", "Graphics"],
Cell[305155, 8862, 1306, 31, 145, "Text"],
Cell[306464, 8895, 118, 2, 99, "Input"],
Cell[306585, 8899, 834, 25, 99, "Input"],
Cell[307422, 8926, 689, 21, 99, "Input"],
Cell[308114, 8949, 194, 4, 41, "Text"],
Cell[308311, 8955, 571, 9, 151, "Input"],
Cell[308885, 8966, 73, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[308983, 8970, 158, 3, 73, "Input"],
Cell[309144, 8975, 286, 4, 59, "Output"]
}, Closed]],
Cell[309445, 8982, 110, 3, 41, "Text"],
Cell[309558, 8987, 466, 13, 73, "Input"],

Cell[CellGroupData[{
Cell[310049, 9004, 337, 6, 73, "Input"],
Cell[310389, 9012, 334, 6, 59, "Output"]
}, Closed]],
Cell[310738, 9021, 609, 15, 68, "Text"],

Cell[CellGroupData[{
Cell[311372, 9040, 245, 4, 74, "Input"],
Cell[311620, 9046, 336, 5, 61, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[311993, 9056, 188, 3, 99, "Input"],
Cell[312184, 9061, 438, 11, 112, "Output"]
}, Closed]],
Cell[312637, 9075, 893, 26, 94, "Text"],

Cell[CellGroupData[{
Cell[313555, 9105, 154, 3, 74, "Input"],
Cell[313712, 9110, 145, 2, 78, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[313894, 9117, 89, 1, 78, "Input"],
Cell[313986, 9120, 148, 3, 91, "Output"]
}, Closed]],
Cell[314149, 9126, 335, 6, 67, "Text"],

Cell[CellGroupData[{
Cell[314509, 9136, 214, 5, 73, "Input"],
Cell[314726, 9143, 241, 4, 78, "Output"]
}, Closed]],
Cell[314982, 9150, 501, 8, 93, "Text"],
Cell[315486, 9160, 613, 15, 69, "Text"],

Cell[CellGroupData[{
Cell[316124, 9179, 143, 2, 73, "Input"],
Cell[316270, 9183, 184, 3, 78, "Output"]
}, Closed]],
Cell[316469, 9189, 59, 0, 41, "Text"],
Cell[316531, 9191, 115, 2, 73, "Input"],

Cell[CellGroupData[{
Cell[316671, 9197, 48, 1, 73, "Input"],
Cell[316722, 9200, 435, 11, 148, "Output"]
}, Closed]],
Cell[317172, 9214, 69, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[317266, 9218, 70, 1, 73, "Input"],
Cell[317339, 9221, 171, 2, 78, "Output"]
}, Closed]],
Cell[317525, 9226, 44, 0, 41, "Text"],

Cell[CellGroupData[{
Cell[317594, 9230, 67, 1, 73, "Input"],
Cell[317664, 9233, 112, 2, 91, "Output"]
}, Closed]],
Cell[317791, 9238, 250, 6, 42, "Text"],

Cell[CellGroupData[{
Cell[318066, 9248, 89, 1, 73, "Input"],
Cell[318158, 9251, 255, 4, 113, "Output"]
}, Closed]],

Cell[CellGroupData[{
Cell[318450, 9260, 63, 1, 73, "Input"],
Cell[318516, 9263, 554, 15, 266, "Output"]
}, Closed]],
Cell[319085, 9281, 186, 4, 41, "Text"],
Cell[319274, 9287, 197, 4, 83, "Exercise"],
Cell[319474, 9293, 521, 16, 83, "Exercise"],
Cell[319998, 9311, 521, 13, 83, "Exercise"],
Cell[320522, 9326, 568, 16, 83, "Exercise"],
Cell[321093, 9344, 458, 7, 108, "Exercise"]
}, Closed]],

Cell[CellGroupData[{
Cell[321588, 9356, 38, 0, 55, "Subsubsection"],
Cell[321629, 9358, 280, 6, 42, "Text"]
}, Closed]]
}, Closed]]
}, Closed]],

Cell[CellGroupData[{
Cell[321970, 9371, 33, 0, 54, "ExerciseMain"],

Cell[CellGroupData[{
Cell[322028, 9375, 60, 0, 58, "ExerciseTitle"],
Cell[322091, 9377, 551, 14, 98, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[322679, 9396, 53, 0, 58, "ExerciseTitle"],
Cell[322735, 9398, 457, 14, 41, "ExerciseText"],

Cell[CellGroupData[{
Cell[323217, 9416, 82, 1, 73, "Input"],
Cell[323302, 9419, 2461, 82, 295, "Output"]
}, Closed]],
Cell[325778, 9504, 432, 12, 67, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[326247, 9521, 39, 0, 58, "ExerciseTitle"],
Cell[326289, 9523, 1030, 32, 94, "ExerciseText"],
Cell[327322, 9557, 33378, 718, 284, 9725, 422, "GraphicsData", "PostScript", \
"Graphics"]
}, Closed]],

Cell[CellGroupData[{
Cell[360737, 10280, 61, 0, 58, "ExerciseTitle"],
Cell[360801, 10282, 1959, 59, 171, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[362797, 10346, 45, 0, 58, "ExerciseTitle"],
Cell[362845, 10348, 498, 11, 67, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[363380, 10364, 61, 0, 58, "ExerciseTitle"],
Cell[363444, 10366, 157, 3, 41, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[363638, 10374, 58, 0, 58, "ExerciseTitle"],
Cell[363699, 10376, 1504, 28, 221, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[365240, 10409, 60, 0, 58, "ExerciseTitle"],
Cell[365303, 10411, 2438, 77, 289, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[367778, 10493, 53, 0, 58, "ExerciseTitle"],
Cell[367834, 10495, 1605, 43, 176, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[369476, 10543, 59, 0, 58, "ExerciseTitle"],
Cell[369538, 10545, 723, 14, 119, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[370298, 10564, 39, 0, 58, "ExerciseTitle"],
Cell[370340, 10566, 54, 0, 41, "ExerciseText"],
Cell[370397, 10568, 179, 5, 62, "InlineInput",
  Evaluatable->False],
Cell[370579, 10575, 180, 4, 41, "ExerciseText"],
Cell[370762, 10581, 498, 10, 102, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[371297, 10596, 117, 4, 66, "ExerciseTitle"],
Cell[371417, 10602, 321, 8, 41, "ExerciseText"],
Cell[371741, 10612, 873, 26, 58, "ExerciseText"],
Cell[372617, 10640, 884, 26, 72, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[373538, 10671, 40, 0, 58, "ExerciseTitle"],
Cell[373581, 10673, 334, 5, 67, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[373952, 10683, 50, 0, 58, "ExerciseTitle"],
Cell[374005, 10685, 2847, 68, 303, "ExerciseText"]
}, Closed]],

Cell[CellGroupData[{
Cell[376889, 10758, 50, 0, 58, "ExerciseTitle"],
Cell[376942, 10760, 4539, 125, 397, "ExerciseText"]
}, Closed]]
}, Closed]]
}, Open  ]]
}
]
*)



(*******************************************************************
End of Mathematica Notebook file.
*******************************************************************)

